// The Module object: Our interface to the outside world. We import
// and export values on it, and do the work to get that through
// closure compiler if necessary. There are various ways Module can be used:
// 1. Not defined. We create it here
// 2. A function parameter, function(Module) { ..generated code.. }
// 3. pre-run appended it, var Module = {}; ..generated code..
// 4. External script tag defines var Module.
// We need to do an eval in order to handle the closure compiler
// case, where this code here is minified but Module was defined
// elsewhere (e.g. case 4 above). We also need to check if Module
// already exists (e.g. case 3 above).
// Note that if you want to run closure, and also to use Module
// after the generated code, you will need to define   var Module = {};
// before the code. Then that object will be used in the code, and you
// can continue to use Module afterwards as well.
var Module;
if (!Module) Module = (typeof Module !== 'undefined' ? Module : null) || {};

// Sometimes an existing Module object exists with properties
// meant to overwrite the default module functionality. Here
// we collect those properties and reapply _after_ we configure
// the current environment's defaults to avoid having to be so
// defensive during initialization.
var moduleOverrides = {};
for (var key in Module) {
  if (Module.hasOwnProperty(key)) {
    moduleOverrides[key] = Module[key];
  }
}

// The environment setup code below is customized to use Module.
// *** Environment setup code ***
var ENVIRONMENT_IS_WEB = false;
var ENVIRONMENT_IS_WORKER = false;
var ENVIRONMENT_IS_NODE = false;
var ENVIRONMENT_IS_SHELL = false;

// Three configurations we can be running in:
// 1) We could be the application main() thread running in the main JS UI thread. (ENVIRONMENT_IS_WORKER == false and ENVIRONMENT_IS_PTHREAD == false)
// 2) We could be the application main() thread proxied to worker. (with Emscripten -s PROXY_TO_WORKER=1) (ENVIRONMENT_IS_WORKER == true, ENVIRONMENT_IS_PTHREAD == false)
// 3) We could be an application pthread running in a worker. (ENVIRONMENT_IS_WORKER == true and ENVIRONMENT_IS_PTHREAD == true)

if (Module['ENVIRONMENT']) {
  if (Module['ENVIRONMENT'] === 'WEB') {
    ENVIRONMENT_IS_WEB = true;
  } else if (Module['ENVIRONMENT'] === 'WORKER') {
    ENVIRONMENT_IS_WORKER = true;
  } else if (Module['ENVIRONMENT'] === 'NODE') {
    ENVIRONMENT_IS_NODE = true;
  } else if (Module['ENVIRONMENT'] === 'SHELL') {
    ENVIRONMENT_IS_SHELL = true;
  } else {
    throw new Error('The provided Module[\'ENVIRONMENT\'] value is not valid. It must be one of: WEB|WORKER|NODE|SHELL.');
  }
} else {
  ENVIRONMENT_IS_WEB = typeof window === 'object';
  ENVIRONMENT_IS_WORKER = typeof importScripts === 'function';
  ENVIRONMENT_IS_NODE = typeof process === 'object' && typeof require === 'function' && !ENVIRONMENT_IS_WEB && !ENVIRONMENT_IS_WORKER;
  ENVIRONMENT_IS_SHELL = !ENVIRONMENT_IS_WEB && !ENVIRONMENT_IS_NODE && !ENVIRONMENT_IS_WORKER;
}


if (ENVIRONMENT_IS_NODE) {
  // Expose functionality in the same simple way that the shells work
  // Note that we pollute the global namespace here, otherwise we break in node
  if (!Module['print']) Module['print'] = console.log;
  if (!Module['printErr']) Module['printErr'] = console.warn;

  var nodeFS;
  var nodePath;

  Module['read'] = function read(filename, binary) {
    if (!nodeFS) nodeFS = require('fs');
    if (!nodePath) nodePath = require('path');

    filename = nodePath['normalize'](filename);
    var ret = nodeFS['readFileSync'](filename);
    // The path is absolute if the normalized version is the same as the resolved.
    if (!ret && filename != nodePath['resolve'](filename)) {
      filename = path.join(__dirname, '..', 'src', filename);
      ret = nodeFS['readFileSync'](filename);
    }
    if (ret && !binary) ret = ret.toString();
    return ret;
  };

  Module['readBinary'] = function readBinary(filename) {
    var ret = Module['read'](filename, true);
    if (!ret.buffer) {
      ret = new Uint8Array(ret);
    }
    assert(ret.buffer);
    return ret;
  };

  Module['load'] = function load(f) {
    globalEval(read(f));
  };

  if (!Module['thisProgram']) {
    if (process['argv'].length > 1) {
      Module['thisProgram'] = process['argv'][1].replace(/\\/g, '/');
    } else {
      Module['thisProgram'] = 'unknown-program';
    }
  }

  Module['arguments'] = process['argv'].slice(2);

  if (typeof module !== 'undefined') {
    module['exports'] = Module;
  }

  process['on']('uncaughtException', function(ex) {
    // suppress ExitStatus exceptions from showing an error
    if (!(ex instanceof ExitStatus)) {
      throw ex;
    }
  });

  Module['inspect'] = function () { return '[Emscripten Module object]'; };
}
else if (ENVIRONMENT_IS_SHELL) {
  if (!Module['print']) Module['print'] = print;
  if (typeof printErr != 'undefined') Module['printErr'] = printErr; // not present in v8 or older sm

  if (typeof read != 'undefined') {
    Module['read'] = read;
  } else {
    Module['read'] = function read() { throw 'no read() available (jsc?)' };
  }

  Module['readBinary'] = function readBinary(f) {
    if (typeof readbuffer === 'function') {
      return new Uint8Array(readbuffer(f));
    }
    var data = read(f, 'binary');
    assert(typeof data === 'object');
    return data;
  };

  if (typeof scriptArgs != 'undefined') {
    Module['arguments'] = scriptArgs;
  } else if (typeof arguments != 'undefined') {
    Module['arguments'] = arguments;
  }

}
else if (ENVIRONMENT_IS_WEB || ENVIRONMENT_IS_WORKER) {
  Module['read'] = function read(url) {
    var xhr = new XMLHttpRequest();
    xhr.open('GET', url, false);
    xhr.send(null);
    return xhr.responseText;
  };

  Module['readAsync'] = function readAsync(url, onload, onerror) {
    var xhr = new XMLHttpRequest();
    xhr.open('GET', url, true);
    xhr.responseType = 'arraybuffer';
    xhr.onload = function xhr_onload() {
      if (xhr.status == 200 || (xhr.status == 0 && xhr.response)) { // file URLs can return 0
        onload(xhr.response);
      } else {
        onerror();
      }
    };
    xhr.onerror = onerror;
    xhr.send(null);
  };

  if (typeof arguments != 'undefined') {
    Module['arguments'] = arguments;
  }

  if (typeof console !== 'undefined') {
    if (!Module['print']) Module['print'] = function print(x) {
      console.log(x);
    };
    if (!Module['printErr']) Module['printErr'] = function printErr(x) {
      console.warn(x);
    };
  } else {
    // Probably a worker, and without console.log. We can do very little here...
    var TRY_USE_DUMP = false;
    if (!Module['print']) Module['print'] = (TRY_USE_DUMP && (typeof(dump) !== "undefined") ? (function(x) {
      dump(x);
    }) : (function(x) {
      // self.postMessage(x); // enable this if you want stdout to be sent as messages
    }));
  }

  if (ENVIRONMENT_IS_WORKER) {
    Module['load'] = importScripts;
  }

  if (typeof Module['setWindowTitle'] === 'undefined') {
    Module['setWindowTitle'] = function(title) { document.title = title };
  }
}
else {
  // Unreachable because SHELL is dependant on the others
  throw 'Unknown runtime environment. Where are we?';
}

function globalEval(x) {
  eval.call(null, x);
}
if (!Module['load'] && Module['read']) {
  Module['load'] = function load(f) {
    globalEval(Module['read'](f));
  };
}
if (!Module['print']) {
  Module['print'] = function(){};
}
if (!Module['printErr']) {
  Module['printErr'] = Module['print'];
}
if (!Module['arguments']) {
  Module['arguments'] = [];
}
if (!Module['thisProgram']) {
  Module['thisProgram'] = './this.program';
}

// *** Environment setup code ***

// Closure helpers
Module.print = Module['print'];
Module.printErr = Module['printErr'];

// Callbacks
Module['preRun'] = [];
Module['postRun'] = [];

// Merge back in the overrides
for (var key in moduleOverrides) {
  if (moduleOverrides.hasOwnProperty(key)) {
    Module[key] = moduleOverrides[key];
  }
}
// Free the object hierarchy contained in the overrides, this lets the GC
// reclaim data used e.g. in memoryInitializerRequest, which is a large typed array.
moduleOverrides = undefined;



// {{PREAMBLE_ADDITIONS}}

// === Preamble library stuff ===

// Documentation for the public APIs defined in this file must be updated in: 
//    site/source/docs/api_reference/preamble.js.rst
// A prebuilt local version of the documentation is available at: 
//    site/build/text/docs/api_reference/preamble.js.txt
// You can also build docs locally as HTML or other formats in site/
// An online HTML version (which may be of a different version of Emscripten)
//    is up at http://kripken.github.io/emscripten-site/docs/api_reference/preamble.js.html

//========================================
// Runtime code shared with compiler
//========================================

var Runtime = {
  setTempRet0: function (value) {
    tempRet0 = value;
  },
  getTempRet0: function () {
    return tempRet0;
  },
  stackSave: function () {
    return STACKTOP;
  },
  stackRestore: function (stackTop) {
    STACKTOP = stackTop;
  },
  getNativeTypeSize: function (type) {
    switch (type) {
      case 'i1': case 'i8': return 1;
      case 'i16': return 2;
      case 'i32': return 4;
      case 'i64': return 8;
      case 'float': return 4;
      case 'double': return 8;
      default: {
        if (type[type.length-1] === '*') {
          return Runtime.QUANTUM_SIZE; // A pointer
        } else if (type[0] === 'i') {
          var bits = parseInt(type.substr(1));
          assert(bits % 8 === 0);
          return bits/8;
        } else {
          return 0;
        }
      }
    }
  },
  getNativeFieldSize: function (type) {
    return Math.max(Runtime.getNativeTypeSize(type), Runtime.QUANTUM_SIZE);
  },
  STACK_ALIGN: 16,
  prepVararg: function (ptr, type) {
    if (type === 'double' || type === 'i64') {
      // move so the load is aligned
      if (ptr & 7) {
        assert((ptr & 7) === 4);
        ptr += 4;
      }
    } else {
      assert((ptr & 3) === 0);
    }
    return ptr;
  },
  getAlignSize: function (type, size, vararg) {
    // we align i64s and doubles on 64-bit boundaries, unlike x86
    if (!vararg && (type == 'i64' || type == 'double')) return 8;
    if (!type) return Math.min(size, 8); // align structures internally to 64 bits
    return Math.min(size || (type ? Runtime.getNativeFieldSize(type) : 0), Runtime.QUANTUM_SIZE);
  },
  dynCall: function (sig, ptr, args) {
    if (args && args.length) {
      assert(args.length == sig.length-1);
      if (!args.splice) args = Array.prototype.slice.call(args);
      args.splice(0, 0, ptr);
      assert(('dynCall_' + sig) in Module, 'bad function pointer type - no table for sig \'' + sig + '\'');
      return Module['dynCall_' + sig].apply(null, args);
    } else {
      assert(sig.length == 1);
      assert(('dynCall_' + sig) in Module, 'bad function pointer type - no table for sig \'' + sig + '\'');
      return Module['dynCall_' + sig].call(null, ptr);
    }
  },
  functionPointers: [],
  addFunction: function (func) {
    for (var i = 0; i < Runtime.functionPointers.length; i++) {
      if (!Runtime.functionPointers[i]) {
        Runtime.functionPointers[i] = func;
        return 2*(1 + i);
      }
    }
    throw 'Finished up all reserved function pointers. Use a higher value for RESERVED_FUNCTION_POINTERS.';
  },
  removeFunction: function (index) {
    Runtime.functionPointers[(index-2)/2] = null;
  },
  warnOnce: function (text) {
    if (!Runtime.warnOnce.shown) Runtime.warnOnce.shown = {};
    if (!Runtime.warnOnce.shown[text]) {
      Runtime.warnOnce.shown[text] = 1;
      Module.printErr(text);
    }
  },
  funcWrappers: {},
  getFuncWrapper: function (func, sig) {
    assert(sig);
    if (!Runtime.funcWrappers[sig]) {
      Runtime.funcWrappers[sig] = {};
    }
    var sigCache = Runtime.funcWrappers[sig];
    if (!sigCache[func]) {
      sigCache[func] = function dynCall_wrapper() {
        return Runtime.dynCall(sig, func, arguments);
      };
    }
    return sigCache[func];
  },
  getCompilerSetting: function (name) {
    throw 'You must build with -s RETAIN_COMPILER_SETTINGS=1 for Runtime.getCompilerSetting or emscripten_get_compiler_setting to work';
  },
  stackAlloc: function (size) { var ret = STACKTOP;STACKTOP = (STACKTOP + size)|0;STACKTOP = (((STACKTOP)+15)&-16);(assert((((STACKTOP|0) < (STACK_MAX|0))|0))|0); return ret; },
  staticAlloc: function (size) { var ret = STATICTOP;STATICTOP = (STATICTOP + (assert(!staticSealed),size))|0;STATICTOP = (((STATICTOP)+15)&-16); return ret; },
  dynamicAlloc: function (size) { var ret = DYNAMICTOP;DYNAMICTOP = (DYNAMICTOP + (assert(DYNAMICTOP > 0),size))|0;DYNAMICTOP = (((DYNAMICTOP)+15)&-16); if (DYNAMICTOP >= TOTAL_MEMORY) { var success = enlargeMemory(); if (!success) { DYNAMICTOP = ret;  return 0; } }; return ret; },
  alignMemory: function (size,quantum) { var ret = size = Math.ceil((size)/(quantum ? quantum : 16))*(quantum ? quantum : 16); return ret; },
  makeBigInt: function (low,high,unsigned) { var ret = (unsigned ? ((+((low>>>0)))+((+((high>>>0)))*4294967296.0)) : ((+((low>>>0)))+((+((high|0)))*4294967296.0))); return ret; },
  GLOBAL_BASE: 8,
  QUANTUM_SIZE: 4,
  __dummy__: 0
}



Module["Runtime"] = Runtime;



//========================================
// Runtime essentials
//========================================

var ABORT = false; // whether we are quitting the application. no code should run after this. set in exit() and abort()
var EXITSTATUS = 0;

function assert(condition, text) {
  if (!condition) {
    abort('Assertion failed: ' + text);
  }
}

var globalScope = this;

// Returns the C function with a specified identifier (for C++, you need to do manual name mangling)
function getCFunc(ident) {
  var func = Module['_' + ident]; // closure exported function
  if (!func) {
    try { func = eval('_' + ident); } catch(e) {}
  }
  assert(func, 'Cannot call unknown function ' + ident + ' (perhaps LLVM optimizations or closure removed it?)');
  return func;
}

var cwrap, ccall;
(function(){
  var JSfuncs = {
    // Helpers for cwrap -- it can't refer to Runtime directly because it might
    // be renamed by closure, instead it calls JSfuncs['stackSave'].body to find
    // out what the minified function name is.
    'stackSave': function() {
      Runtime.stackSave()
    },
    'stackRestore': function() {
      Runtime.stackRestore()
    },
    // type conversion from js to c
    'arrayToC' : function(arr) {
      var ret = Runtime.stackAlloc(arr.length);
      writeArrayToMemory(arr, ret);
      return ret;
    },
    'stringToC' : function(str) {
      var ret = 0;
      if (str !== null && str !== undefined && str !== 0) { // null string
        // at most 4 bytes per UTF-8 code point, +1 for the trailing '\0'
        ret = Runtime.stackAlloc((str.length << 2) + 1);
        writeStringToMemory(str, ret);
      }
      return ret;
    }
  };
  // For fast lookup of conversion functions
  var toC = {'string' : JSfuncs['stringToC'], 'array' : JSfuncs['arrayToC']};

  // C calling interface. 
  ccall = function ccallFunc(ident, returnType, argTypes, args, opts) {
    var func = getCFunc(ident);
    var cArgs = [];
    var stack = 0;
    assert(returnType !== 'array', 'Return type should not be "array".');
    if (args) {
      for (var i = 0; i < args.length; i++) {
        var converter = toC[argTypes[i]];
        if (converter) {
          if (stack === 0) stack = Runtime.stackSave();
          cArgs[i] = converter(args[i]);
        } else {
          cArgs[i] = args[i];
        }
      }
    }
    var ret = func.apply(null, cArgs);
    if ((!opts || !opts.async) && typeof EmterpreterAsync === 'object') {
      assert(!EmterpreterAsync.state, 'cannot start async op with normal JS calling ccall');
    }
    if (opts && opts.async) assert(!returnType, 'async ccalls cannot return values');
    if (returnType === 'string') ret = Pointer_stringify(ret);
    if (stack !== 0) {
      if (opts && opts.async) {
        EmterpreterAsync.asyncFinalizers.push(function() {
          Runtime.stackRestore(stack);
        });
        return;
      }
      Runtime.stackRestore(stack);
    }
    return ret;
  }

  var sourceRegex = /^function\s*\(([^)]*)\)\s*{\s*([^*]*?)[\s;]*(?:return\s*(.*?)[;\s]*)?}$/;
  function parseJSFunc(jsfunc) {
    // Match the body and the return value of a javascript function source
    var parsed = jsfunc.toString().match(sourceRegex).slice(1);
    return {arguments : parsed[0], body : parsed[1], returnValue: parsed[2]}
  }

  // sources of useful functions. we create this lazily as it can trigger a source decompression on this entire file
  var JSsource = null;
  function ensureJSsource() {
    if (!JSsource) {
      JSsource = {};
      for (var fun in JSfuncs) {
        if (JSfuncs.hasOwnProperty(fun)) {
          // Elements of toCsource are arrays of three items:
          // the code, and the return value
          JSsource[fun] = parseJSFunc(JSfuncs[fun]);
        }
      }
    }
  }
  
  cwrap = function cwrap(ident, returnType, argTypes) {
    argTypes = argTypes || [];
    var cfunc = getCFunc(ident);
    // When the function takes numbers and returns a number, we can just return
    // the original function
    var numericArgs = argTypes.every(function(type){ return type === 'number'});
    var numericRet = (returnType !== 'string');
    if ( numericRet && numericArgs) {
      return cfunc;
    }
    // Creation of the arguments list (["$1","$2",...,"$nargs"])
    var argNames = argTypes.map(function(x,i){return '$'+i});
    var funcstr = "(function(" + argNames.join(',') + ") {";
    var nargs = argTypes.length;
    if (!numericArgs) {
      // Generate the code needed to convert the arguments from javascript
      // values to pointers
      ensureJSsource();
      funcstr += 'var stack = ' + JSsource['stackSave'].body + ';';
      for (var i = 0; i < nargs; i++) {
        var arg = argNames[i], type = argTypes[i];
        if (type === 'number') continue;
        var convertCode = JSsource[type + 'ToC']; // [code, return]
        funcstr += 'var ' + convertCode.arguments + ' = ' + arg + ';';
        funcstr += convertCode.body + ';';
        funcstr += arg + '=(' + convertCode.returnValue + ');';
      }
    }

    // When the code is compressed, the name of cfunc is not literally 'cfunc' anymore
    var cfuncname = parseJSFunc(function(){return cfunc}).returnValue;
    // Call the function
    funcstr += 'var ret = ' + cfuncname + '(' + argNames.join(',') + ');';
    if (!numericRet) { // Return type can only by 'string' or 'number'
      // Convert the result to a string
      var strgfy = parseJSFunc(function(){return Pointer_stringify}).returnValue;
      funcstr += 'ret = ' + strgfy + '(ret);';
    }
    funcstr += "if (typeof EmterpreterAsync === 'object') { assert(!EmterpreterAsync.state, 'cannot start async op with normal JS calling cwrap') }";
    if (!numericArgs) {
      // If we had a stack, restore it
      ensureJSsource();
      funcstr += JSsource['stackRestore'].body.replace('()', '(stack)') + ';';
    }
    funcstr += 'return ret})';
    return eval(funcstr);
  };
})();
Module["ccall"] = ccall;
Module["cwrap"] = cwrap;

function setValue(ptr, value, type, noSafe) {
  type = type || 'i8';
  if (type.charAt(type.length-1) === '*') type = 'i32'; // pointers are 32-bit
    switch(type) {
      case 'i1': HEAP8[((ptr)>>0)]=value; break;
      case 'i8': HEAP8[((ptr)>>0)]=value; break;
      case 'i16': HEAP16[((ptr)>>1)]=value; break;
      case 'i32': HEAP32[((ptr)>>2)]=value; break;
      case 'i64': (tempI64 = [value>>>0,(tempDouble=value,(+(Math_abs(tempDouble))) >= 1.0 ? (tempDouble > 0.0 ? ((Math_min((+(Math_floor((tempDouble)/4294967296.0))), 4294967295.0))|0)>>>0 : (~~((+(Math_ceil((tempDouble - +(((~~(tempDouble)))>>>0))/4294967296.0)))))>>>0) : 0)],HEAP32[((ptr)>>2)]=tempI64[0],HEAP32[(((ptr)+(4))>>2)]=tempI64[1]); break;
      case 'float': HEAPF32[((ptr)>>2)]=value; break;
      case 'double': HEAPF64[((ptr)>>3)]=value; break;
      default: abort('invalid type for setValue: ' + type);
    }
}
Module["setValue"] = setValue;


function getValue(ptr, type, noSafe) {
  type = type || 'i8';
  if (type.charAt(type.length-1) === '*') type = 'i32'; // pointers are 32-bit
    switch(type) {
      case 'i1': return HEAP8[((ptr)>>0)];
      case 'i8': return HEAP8[((ptr)>>0)];
      case 'i16': return HEAP16[((ptr)>>1)];
      case 'i32': return HEAP32[((ptr)>>2)];
      case 'i64': return HEAP32[((ptr)>>2)];
      case 'float': return HEAPF32[((ptr)>>2)];
      case 'double': return HEAPF64[((ptr)>>3)];
      default: abort('invalid type for setValue: ' + type);
    }
  return null;
}
Module["getValue"] = getValue;

var ALLOC_NORMAL = 0; // Tries to use _malloc()
var ALLOC_STACK = 1; // Lives for the duration of the current function call
var ALLOC_STATIC = 2; // Cannot be freed
var ALLOC_DYNAMIC = 3; // Cannot be freed except through sbrk
var ALLOC_NONE = 4; // Do not allocate
Module["ALLOC_NORMAL"] = ALLOC_NORMAL;
Module["ALLOC_STACK"] = ALLOC_STACK;
Module["ALLOC_STATIC"] = ALLOC_STATIC;
Module["ALLOC_DYNAMIC"] = ALLOC_DYNAMIC;
Module["ALLOC_NONE"] = ALLOC_NONE;

// allocate(): This is for internal use. You can use it yourself as well, but the interface
//             is a little tricky (see docs right below). The reason is that it is optimized
//             for multiple syntaxes to save space in generated code. So you should
//             normally not use allocate(), and instead allocate memory using _malloc(),
//             initialize it with setValue(), and so forth.
// @slab: An array of data, or a number. If a number, then the size of the block to allocate,
//        in *bytes* (note that this is sometimes confusing: the next parameter does not
//        affect this!)
// @types: Either an array of types, one for each byte (or 0 if no type at that position),
//         or a single type which is used for the entire block. This only matters if there
//         is initial data - if @slab is a number, then this does not matter at all and is
//         ignored.
// @allocator: How to allocate memory, see ALLOC_*
function allocate(slab, types, allocator, ptr) {
  var zeroinit, size;
  if (typeof slab === 'number') {
    zeroinit = true;
    size = slab;
  } else {
    zeroinit = false;
    size = slab.length;
  }

  var singleType = typeof types === 'string' ? types : null;

  var ret;
  if (allocator == ALLOC_NONE) {
    ret = ptr;
  } else {
    ret = [typeof _malloc === 'function' ? _malloc : Runtime.staticAlloc, Runtime.stackAlloc, Runtime.staticAlloc, Runtime.dynamicAlloc][allocator === undefined ? ALLOC_STATIC : allocator](Math.max(size, singleType ? 1 : types.length));
  }

  if (zeroinit) {
    var ptr = ret, stop;
    assert((ret & 3) == 0);
    stop = ret + (size & ~3);
    for (; ptr < stop; ptr += 4) {
      HEAP32[((ptr)>>2)]=0;
    }
    stop = ret + size;
    while (ptr < stop) {
      HEAP8[((ptr++)>>0)]=0;
    }
    return ret;
  }

  if (singleType === 'i8') {
    if (slab.subarray || slab.slice) {
      HEAPU8.set(slab, ret);
    } else {
      HEAPU8.set(new Uint8Array(slab), ret);
    }
    return ret;
  }

  var i = 0, type, typeSize, previousType;
  while (i < size) {
    var curr = slab[i];

    if (typeof curr === 'function') {
      curr = Runtime.getFunctionIndex(curr);
    }

    type = singleType || types[i];
    if (type === 0) {
      i++;
      continue;
    }
    assert(type, 'Must know what type to store in allocate!');

    if (type == 'i64') type = 'i32'; // special case: we have one i32 here, and one i32 later

    setValue(ret+i, curr, type);

    // no need to look up size unless type changes, so cache it
    if (previousType !== type) {
      typeSize = Runtime.getNativeTypeSize(type);
      previousType = type;
    }
    i += typeSize;
  }

  return ret;
}
Module["allocate"] = allocate;

// Allocate memory during any stage of startup - static memory early on, dynamic memory later, malloc when ready
function getMemory(size) {
  if (!staticSealed) return Runtime.staticAlloc(size);
  if ((typeof _sbrk !== 'undefined' && !_sbrk.called) || !runtimeInitialized) return Runtime.dynamicAlloc(size);
  return _malloc(size);
}
Module["getMemory"] = getMemory;

function Pointer_stringify(ptr, /* optional */ length) {
  if (length === 0 || !ptr) return '';
  // TODO: use TextDecoder
  // Find the length, and check for UTF while doing so
  var hasUtf = 0;
  var t;
  var i = 0;
  while (1) {
    assert(ptr + i < TOTAL_MEMORY);
    t = HEAPU8[(((ptr)+(i))>>0)];
    hasUtf |= t;
    if (t == 0 && !length) break;
    i++;
    if (length && i == length) break;
  }
  if (!length) length = i;

  var ret = '';

  if (hasUtf < 128) {
    var MAX_CHUNK = 1024; // split up into chunks, because .apply on a huge string can overflow the stack
    var curr;
    while (length > 0) {
      curr = String.fromCharCode.apply(String, HEAPU8.subarray(ptr, ptr + Math.min(length, MAX_CHUNK)));
      ret = ret ? ret + curr : curr;
      ptr += MAX_CHUNK;
      length -= MAX_CHUNK;
    }
    return ret;
  }
  return Module['UTF8ToString'](ptr);
}
Module["Pointer_stringify"] = Pointer_stringify;

// Given a pointer 'ptr' to a null-terminated ASCII-encoded string in the emscripten HEAP, returns
// a copy of that string as a Javascript String object.

function AsciiToString(ptr) {
  var str = '';
  while (1) {
    var ch = HEAP8[((ptr++)>>0)];
    if (!ch) return str;
    str += String.fromCharCode(ch);
  }
}
Module["AsciiToString"] = AsciiToString;

// Copies the given Javascript String object 'str' to the emscripten HEAP at address 'outPtr',
// null-terminated and encoded in ASCII form. The copy will require at most str.length+1 bytes of space in the HEAP.

function stringToAscii(str, outPtr) {
  return writeAsciiToMemory(str, outPtr, false);
}
Module["stringToAscii"] = stringToAscii;

// Given a pointer 'ptr' to a null-terminated UTF8-encoded string in the given array that contains uint8 values, returns
// a copy of that string as a Javascript String object.

function UTF8ArrayToString(u8Array, idx) {
  var u0, u1, u2, u3, u4, u5;

  var str = '';
  while (1) {
    // For UTF8 byte structure, see http://en.wikipedia.org/wiki/UTF-8#Description and https://www.ietf.org/rfc/rfc2279.txt and https://tools.ietf.org/html/rfc3629
    u0 = u8Array[idx++];
    if (!u0) return str;
    if (!(u0 & 0x80)) { str += String.fromCharCode(u0); continue; }
    u1 = u8Array[idx++] & 63;
    if ((u0 & 0xE0) == 0xC0) { str += String.fromCharCode(((u0 & 31) << 6) | u1); continue; }
    u2 = u8Array[idx++] & 63;
    if ((u0 & 0xF0) == 0xE0) {
      u0 = ((u0 & 15) << 12) | (u1 << 6) | u2;
    } else {
      u3 = u8Array[idx++] & 63;
      if ((u0 & 0xF8) == 0xF0) {
        u0 = ((u0 & 7) << 18) | (u1 << 12) | (u2 << 6) | u3;
      } else {
        u4 = u8Array[idx++] & 63;
        if ((u0 & 0xFC) == 0xF8) {
          u0 = ((u0 & 3) << 24) | (u1 << 18) | (u2 << 12) | (u3 << 6) | u4;
        } else {
          u5 = u8Array[idx++] & 63;
          u0 = ((u0 & 1) << 30) | (u1 << 24) | (u2 << 18) | (u3 << 12) | (u4 << 6) | u5;
        }
      }
    }
    if (u0 < 0x10000) {
      str += String.fromCharCode(u0);
    } else {
      var ch = u0 - 0x10000;
      str += String.fromCharCode(0xD800 | (ch >> 10), 0xDC00 | (ch & 0x3FF));
    }
  }
}
Module["UTF8ArrayToString"] = UTF8ArrayToString;

// Given a pointer 'ptr' to a null-terminated UTF8-encoded string in the emscripten HEAP, returns
// a copy of that string as a Javascript String object.

function UTF8ToString(ptr) {
  return UTF8ArrayToString(HEAPU8,ptr);
}
Module["UTF8ToString"] = UTF8ToString;

// Copies the given Javascript String object 'str' to the given byte array at address 'outIdx',
// encoded in UTF8 form and null-terminated. The copy will require at most str.length*4+1 bytes of space in the HEAP.
// Use the function lengthBytesUTF8() to compute the exact number of bytes (excluding null terminator) that this function will write.
// Parameters:
//   str: the Javascript string to copy.
//   outU8Array: the array to copy to. Each index in this array is assumed to be one 8-byte element.
//   outIdx: The starting offset in the array to begin the copying.
//   maxBytesToWrite: The maximum number of bytes this function can write to the array. This count should include the null 
//                    terminator, i.e. if maxBytesToWrite=1, only the null terminator will be written and nothing else.
//                    maxBytesToWrite=0 does not write any bytes to the output, not even the null terminator.
// Returns the number of bytes written, EXCLUDING the null terminator.

function stringToUTF8Array(str, outU8Array, outIdx, maxBytesToWrite) {
  if (!(maxBytesToWrite > 0)) // Parameter maxBytesToWrite is not optional. Negative values, 0, null, undefined and false each don't write out any bytes.
    return 0;

  var startIdx = outIdx;
  var endIdx = outIdx + maxBytesToWrite - 1; // -1 for string null terminator.
  for (var i = 0; i < str.length; ++i) {
    // Gotcha: charCodeAt returns a 16-bit word that is a UTF-16 encoded code unit, not a Unicode code point of the character! So decode UTF16->UTF32->UTF8.
    // See http://unicode.org/faq/utf_bom.html#utf16-3
    // For UTF8 byte structure, see http://en.wikipedia.org/wiki/UTF-8#Description and https://www.ietf.org/rfc/rfc2279.txt and https://tools.ietf.org/html/rfc3629
    var u = str.charCodeAt(i); // possibly a lead surrogate
    if (u >= 0xD800 && u <= 0xDFFF) u = 0x10000 + ((u & 0x3FF) << 10) | (str.charCodeAt(++i) & 0x3FF);
    if (u <= 0x7F) {
      if (outIdx >= endIdx) break;
      outU8Array[outIdx++] = u;
    } else if (u <= 0x7FF) {
      if (outIdx + 1 >= endIdx) break;
      outU8Array[outIdx++] = 0xC0 | (u >> 6);
      outU8Array[outIdx++] = 0x80 | (u & 63);
    } else if (u <= 0xFFFF) {
      if (outIdx + 2 >= endIdx) break;
      outU8Array[outIdx++] = 0xE0 | (u >> 12);
      outU8Array[outIdx++] = 0x80 | ((u >> 6) & 63);
      outU8Array[outIdx++] = 0x80 | (u & 63);
    } else if (u <= 0x1FFFFF) {
      if (outIdx + 3 >= endIdx) break;
      outU8Array[outIdx++] = 0xF0 | (u >> 18);
      outU8Array[outIdx++] = 0x80 | ((u >> 12) & 63);
      outU8Array[outIdx++] = 0x80 | ((u >> 6) & 63);
      outU8Array[outIdx++] = 0x80 | (u & 63);
    } else if (u <= 0x3FFFFFF) {
      if (outIdx + 4 >= endIdx) break;
      outU8Array[outIdx++] = 0xF8 | (u >> 24);
      outU8Array[outIdx++] = 0x80 | ((u >> 18) & 63);
      outU8Array[outIdx++] = 0x80 | ((u >> 12) & 63);
      outU8Array[outIdx++] = 0x80 | ((u >> 6) & 63);
      outU8Array[outIdx++] = 0x80 | (u & 63);
    } else {
      if (outIdx + 5 >= endIdx) break;
      outU8Array[outIdx++] = 0xFC | (u >> 30);
      outU8Array[outIdx++] = 0x80 | ((u >> 24) & 63);
      outU8Array[outIdx++] = 0x80 | ((u >> 18) & 63);
      outU8Array[outIdx++] = 0x80 | ((u >> 12) & 63);
      outU8Array[outIdx++] = 0x80 | ((u >> 6) & 63);
      outU8Array[outIdx++] = 0x80 | (u & 63);
    }
  }
  // Null-terminate the pointer to the buffer.
  outU8Array[outIdx] = 0;
  return outIdx - startIdx;
}
Module["stringToUTF8Array"] = stringToUTF8Array;

// Copies the given Javascript String object 'str' to the emscripten HEAP at address 'outPtr',
// null-terminated and encoded in UTF8 form. The copy will require at most str.length*4+1 bytes of space in the HEAP.
// Use the function lengthBytesUTF8() to compute the exact number of bytes (excluding null terminator) that this function will write.
// Returns the number of bytes written, EXCLUDING the null terminator.

function stringToUTF8(str, outPtr, maxBytesToWrite) {
  assert(typeof maxBytesToWrite == 'number', 'stringToUTF8(str, outPtr, maxBytesToWrite) is missing the third parameter that specifies the length of the output buffer!');
  return stringToUTF8Array(str, HEAPU8,outPtr, maxBytesToWrite);
}
Module["stringToUTF8"] = stringToUTF8;

// Returns the number of bytes the given Javascript string takes if encoded as a UTF8 byte array, EXCLUDING the null terminator byte.

function lengthBytesUTF8(str) {
  var len = 0;
  for (var i = 0; i < str.length; ++i) {
    // Gotcha: charCodeAt returns a 16-bit word that is a UTF-16 encoded code unit, not a Unicode code point of the character! So decode UTF16->UTF32->UTF8.
    // See http://unicode.org/faq/utf_bom.html#utf16-3
    var u = str.charCodeAt(i); // possibly a lead surrogate
    if (u >= 0xD800 && u <= 0xDFFF) u = 0x10000 + ((u & 0x3FF) << 10) | (str.charCodeAt(++i) & 0x3FF);
    if (u <= 0x7F) {
      ++len;
    } else if (u <= 0x7FF) {
      len += 2;
    } else if (u <= 0xFFFF) {
      len += 3;
    } else if (u <= 0x1FFFFF) {
      len += 4;
    } else if (u <= 0x3FFFFFF) {
      len += 5;
    } else {
      len += 6;
    }
  }
  return len;
}
Module["lengthBytesUTF8"] = lengthBytesUTF8;

// Given a pointer 'ptr' to a null-terminated UTF16LE-encoded string in the emscripten HEAP, returns
// a copy of that string as a Javascript String object.

function UTF16ToString(ptr) {
  var i = 0;

  var str = '';
  while (1) {
    var codeUnit = HEAP16[(((ptr)+(i*2))>>1)];
    if (codeUnit == 0)
      return str;
    ++i;
    // fromCharCode constructs a character from a UTF-16 code unit, so we can pass the UTF16 string right through.
    str += String.fromCharCode(codeUnit);
  }
}


// Copies the given Javascript String object 'str' to the emscripten HEAP at address 'outPtr',
// null-terminated and encoded in UTF16 form. The copy will require at most str.length*4+2 bytes of space in the HEAP.
// Use the function lengthBytesUTF16() to compute the exact number of bytes (excluding null terminator) that this function will write.
// Parameters:
//   str: the Javascript string to copy.
//   outPtr: Byte address in Emscripten HEAP where to write the string to.
//   maxBytesToWrite: The maximum number of bytes this function can write to the array. This count should include the null 
//                    terminator, i.e. if maxBytesToWrite=2, only the null terminator will be written and nothing else.
//                    maxBytesToWrite<2 does not write any bytes to the output, not even the null terminator.
// Returns the number of bytes written, EXCLUDING the null terminator.

function stringToUTF16(str, outPtr, maxBytesToWrite) {
  assert(typeof maxBytesToWrite == 'number', 'stringToUTF16(str, outPtr, maxBytesToWrite) is missing the third parameter that specifies the length of the output buffer!');
  // Backwards compatibility: if max bytes is not specified, assume unsafe unbounded write is allowed.
  if (maxBytesToWrite === undefined) {
    maxBytesToWrite = 0x7FFFFFFF;
  }
  if (maxBytesToWrite < 2) return 0;
  maxBytesToWrite -= 2; // Null terminator.
  var startPtr = outPtr;
  var numCharsToWrite = (maxBytesToWrite < str.length*2) ? (maxBytesToWrite / 2) : str.length;
  for (var i = 0; i < numCharsToWrite; ++i) {
    // charCodeAt returns a UTF-16 encoded code unit, so it can be directly written to the HEAP.
    var codeUnit = str.charCodeAt(i); // possibly a lead surrogate
    HEAP16[((outPtr)>>1)]=codeUnit;
    outPtr += 2;
  }
  // Null-terminate the pointer to the HEAP.
  HEAP16[((outPtr)>>1)]=0;
  return outPtr - startPtr;
}


// Returns the number of bytes the given Javascript string takes if encoded as a UTF16 byte array, EXCLUDING the null terminator byte.

function lengthBytesUTF16(str) {
  return str.length*2;
}


function UTF32ToString(ptr) {
  var i = 0;

  var str = '';
  while (1) {
    var utf32 = HEAP32[(((ptr)+(i*4))>>2)];
    if (utf32 == 0)
      return str;
    ++i;
    // Gotcha: fromCharCode constructs a character from a UTF-16 encoded code (pair), not from a Unicode code point! So encode the code point to UTF-16 for constructing.
    // See http://unicode.org/faq/utf_bom.html#utf16-3
    if (utf32 >= 0x10000) {
      var ch = utf32 - 0x10000;
      str += String.fromCharCode(0xD800 | (ch >> 10), 0xDC00 | (ch & 0x3FF));
    } else {
      str += String.fromCharCode(utf32);
    }
  }
}


// Copies the given Javascript String object 'str' to the emscripten HEAP at address 'outPtr',
// null-terminated and encoded in UTF32 form. The copy will require at most str.length*4+4 bytes of space in the HEAP.
// Use the function lengthBytesUTF32() to compute the exact number of bytes (excluding null terminator) that this function will write.
// Parameters:
//   str: the Javascript string to copy.
//   outPtr: Byte address in Emscripten HEAP where to write the string to.
//   maxBytesToWrite: The maximum number of bytes this function can write to the array. This count should include the null 
//                    terminator, i.e. if maxBytesToWrite=4, only the null terminator will be written and nothing else.
//                    maxBytesToWrite<4 does not write any bytes to the output, not even the null terminator.
// Returns the number of bytes written, EXCLUDING the null terminator.

function stringToUTF32(str, outPtr, maxBytesToWrite) {
  assert(typeof maxBytesToWrite == 'number', 'stringToUTF32(str, outPtr, maxBytesToWrite) is missing the third parameter that specifies the length of the output buffer!');
  // Backwards compatibility: if max bytes is not specified, assume unsafe unbounded write is allowed.
  if (maxBytesToWrite === undefined) {
    maxBytesToWrite = 0x7FFFFFFF;
  }
  if (maxBytesToWrite < 4) return 0;
  var startPtr = outPtr;
  var endPtr = startPtr + maxBytesToWrite - 4;
  for (var i = 0; i < str.length; ++i) {
    // Gotcha: charCodeAt returns a 16-bit word that is a UTF-16 encoded code unit, not a Unicode code point of the character! We must decode the string to UTF-32 to the heap.
    // See http://unicode.org/faq/utf_bom.html#utf16-3
    var codeUnit = str.charCodeAt(i); // possibly a lead surrogate
    if (codeUnit >= 0xD800 && codeUnit <= 0xDFFF) {
      var trailSurrogate = str.charCodeAt(++i);
      codeUnit = 0x10000 + ((codeUnit & 0x3FF) << 10) | (trailSurrogate & 0x3FF);
    }
    HEAP32[((outPtr)>>2)]=codeUnit;
    outPtr += 4;
    if (outPtr + 4 > endPtr) break;
  }
  // Null-terminate the pointer to the HEAP.
  HEAP32[((outPtr)>>2)]=0;
  return outPtr - startPtr;
}


// Returns the number of bytes the given Javascript string takes if encoded as a UTF16 byte array, EXCLUDING the null terminator byte.

function lengthBytesUTF32(str) {
  var len = 0;
  for (var i = 0; i < str.length; ++i) {
    // Gotcha: charCodeAt returns a 16-bit word that is a UTF-16 encoded code unit, not a Unicode code point of the character! We must decode the string to UTF-32 to the heap.
    // See http://unicode.org/faq/utf_bom.html#utf16-3
    var codeUnit = str.charCodeAt(i);
    if (codeUnit >= 0xD800 && codeUnit <= 0xDFFF) ++i; // possibly a lead surrogate, so skip over the tail surrogate.
    len += 4;
  }

  return len;
}


function demangle(func) {
  var hasLibcxxabi = !!Module['___cxa_demangle'];
  if (hasLibcxxabi) {
    try {
      var buf = _malloc(func.length);
      writeStringToMemory(func.substr(1), buf);
      var status = _malloc(4);
      var ret = Module['___cxa_demangle'](buf, 0, 0, status);
      if (getValue(status, 'i32') === 0 && ret) {
        return Pointer_stringify(ret);
      }
      // otherwise, libcxxabi failed, we can try ours which may return a partial result
    } catch(e) {
      // failure when using libcxxabi, we can try ours which may return a partial result
      return func;
    } finally {
      if (buf) _free(buf);
      if (status) _free(status);
      if (ret) _free(ret);
    }
  }
  Runtime.warnOnce('warning: build with  -s DEMANGLE_SUPPORT=1  to link in libcxxabi demangling');
  return func;
}

function demangleAll(text) {
  return text.replace(/__Z[\w\d_]+/g, function(x) { var y = demangle(x); return x === y ? x : (x + ' [' + y + ']') });
}

function jsStackTrace() {
  var err = new Error();
  if (!err.stack) {
    // IE10+ special cases: It does have callstack info, but it is only populated if an Error object is thrown,
    // so try that as a special-case.
    try {
      throw new Error(0);
    } catch(e) {
      err = e;
    }
    if (!err.stack) {
      return '(no stack trace available)';
    }
  }
  return err.stack.toString();
}

function stackTrace() {
  return demangleAll(jsStackTrace());
}
Module["stackTrace"] = stackTrace;

// Memory management

var PAGE_SIZE = 4096;

function alignMemoryPage(x) {
  if (x % 4096 > 0) {
    x += (4096 - (x % 4096));
  }
  return x;
}

var HEAP;
var buffer;
var HEAP8, HEAPU8, HEAP16, HEAPU16, HEAP32, HEAPU32, HEAPF32, HEAPF64;

function updateGlobalBuffer(buf) {
  Module['buffer'] = buffer = buf;
}

function updateGlobalBufferViews() {
  Module['HEAP8'] = HEAP8 = new Int8Array(buffer);
  Module['HEAP16'] = HEAP16 = new Int16Array(buffer);
  Module['HEAP32'] = HEAP32 = new Int32Array(buffer);
  Module['HEAPU8'] = HEAPU8 = new Uint8Array(buffer);
  Module['HEAPU16'] = HEAPU16 = new Uint16Array(buffer);
  Module['HEAPU32'] = HEAPU32 = new Uint32Array(buffer);
  Module['HEAPF32'] = HEAPF32 = new Float32Array(buffer);
  Module['HEAPF64'] = HEAPF64 = new Float64Array(buffer);
}

var STATIC_BASE = 0, STATICTOP = 0, staticSealed = false; // static area
var STACK_BASE = 0, STACKTOP = 0, STACK_MAX = 0; // stack area
var DYNAMIC_BASE = 0, DYNAMICTOP = 0; // dynamic area handled by sbrk


if (!Module['reallocBuffer']) Module['reallocBuffer'] = function(size) {
  var ret;
  try {
    if (ArrayBuffer.transfer) {
      ret = ArrayBuffer.transfer(buffer, size);
    } else {
      var oldHEAP8 = HEAP8;
      ret = new ArrayBuffer(size);
      var temp = new Int8Array(ret);
      temp.set(oldHEAP8);
    }
  } catch(e) {
    return false;
  }
  var success = _emscripten_replace_memory(ret);
  if (!success) return false;
  return ret;
};

function enlargeMemory() {
  // TOTAL_MEMORY is the current size of the actual array, and DYNAMICTOP is the new top.
  assert(DYNAMICTOP >= TOTAL_MEMORY);
  assert(TOTAL_MEMORY > 4); // So the loop below will not be infinite

  var OLD_TOTAL_MEMORY = TOTAL_MEMORY;


  var LIMIT = Math.pow(2, 31); // 2GB is a practical maximum, as we use signed ints as pointers
                               // and JS engines seem unhappy to give us 2GB arrays currently
  if (DYNAMICTOP >= LIMIT) return false;

  while (TOTAL_MEMORY <= DYNAMICTOP) { // Simple heuristic.
    if (TOTAL_MEMORY < LIMIT/2) {
      TOTAL_MEMORY = alignMemoryPage(2*TOTAL_MEMORY); // double until 1GB
    } else {
      var last = TOTAL_MEMORY;
      TOTAL_MEMORY = alignMemoryPage((3*TOTAL_MEMORY + LIMIT)/4); // add smaller increments towards 2GB, which we cannot reach
      if (TOTAL_MEMORY <= last) return false;
    }
  }

  TOTAL_MEMORY = Math.max(TOTAL_MEMORY, 16*1024*1024);

  if (TOTAL_MEMORY >= LIMIT) return false;

  Module.printErr('Warning: Enlarging memory arrays, this is not fast! ' + [OLD_TOTAL_MEMORY, TOTAL_MEMORY]);


  var start = Date.now();

  var replacement = Module['reallocBuffer'](TOTAL_MEMORY);
  if (!replacement) return false;

  // everything worked

  updateGlobalBuffer(replacement);
  updateGlobalBufferViews();

  Module.printErr('enlarged memory arrays from ' + OLD_TOTAL_MEMORY + ' to ' + TOTAL_MEMORY + ', took ' + (Date.now() - start) + ' ms (has ArrayBuffer.transfer? ' + (!!ArrayBuffer.transfer) + ')');

  return true;
}

var byteLength;
try {
  byteLength = Function.prototype.call.bind(Object.getOwnPropertyDescriptor(ArrayBuffer.prototype, 'byteLength').get);
  byteLength(new ArrayBuffer(4)); // can fail on older ie
} catch(e) { // can fail on older node/v8
  byteLength = function(buffer) { return buffer.byteLength; };
}

var TOTAL_STACK = Module['TOTAL_STACK'] || 5242880;
var TOTAL_MEMORY = Module['TOTAL_MEMORY'] || 16777216;

var totalMemory = 64*1024;
while (totalMemory < TOTAL_MEMORY || totalMemory < 2*TOTAL_STACK) {
  if (totalMemory < 16*1024*1024) {
    totalMemory *= 2;
  } else {
    totalMemory += 16*1024*1024
  }
}
totalMemory = Math.max(totalMemory, 16*1024*1024);
if (totalMemory !== TOTAL_MEMORY) {
  Module.printErr('increasing TOTAL_MEMORY to ' + totalMemory + ' to be compliant with the asm.js spec (and given that TOTAL_STACK=' + TOTAL_STACK + ')');
  TOTAL_MEMORY = totalMemory;
}

// Initialize the runtime's memory
// check for full engine support (use string 'subarray' to avoid closure compiler confusion)
assert(typeof Int32Array !== 'undefined' && typeof Float64Array !== 'undefined' && !!(new Int32Array(1)['subarray']) && !!(new Int32Array(1)['set']),
       'JS engine does not provide full typed array support');



// Use a provided buffer, if there is one, or else allocate a new one
if (Module['buffer']) {
  buffer = Module['buffer'];
  assert(buffer.byteLength === TOTAL_MEMORY, 'provided buffer should be ' + TOTAL_MEMORY + ' bytes, but it is ' + buffer.byteLength);
} else {
  buffer = new ArrayBuffer(TOTAL_MEMORY);
}
updateGlobalBufferViews();


// Endianness check (note: assumes compiler arch was little-endian)
HEAP32[0] = 255;
if (HEAPU8[0] !== 255 || HEAPU8[3] !== 0) throw 'Typed arrays 2 must be run on a little-endian system';

Module['HEAP'] = HEAP;
Module['buffer'] = buffer;
Module['HEAP8'] = HEAP8;
Module['HEAP16'] = HEAP16;
Module['HEAP32'] = HEAP32;
Module['HEAPU8'] = HEAPU8;
Module['HEAPU16'] = HEAPU16;
Module['HEAPU32'] = HEAPU32;
Module['HEAPF32'] = HEAPF32;
Module['HEAPF64'] = HEAPF64;

function callRuntimeCallbacks(callbacks) {
  while(callbacks.length > 0) {
    var callback = callbacks.shift();
    if (typeof callback == 'function') {
      callback();
      continue;
    }
    var func = callback.func;
    if (typeof func === 'number') {
      if (callback.arg === undefined) {
        Runtime.dynCall('v', func);
      } else {
        Runtime.dynCall('vi', func, [callback.arg]);
      }
    } else {
      func(callback.arg === undefined ? null : callback.arg);
    }
  }
}

var __ATPRERUN__  = []; // functions called before the runtime is initialized
var __ATINIT__    = []; // functions called during startup
var __ATMAIN__    = []; // functions called when main() is to be run
var __ATEXIT__    = []; // functions called during shutdown
var __ATPOSTRUN__ = []; // functions called after the runtime has exited

var runtimeInitialized = false;
var runtimeExited = false;


function preRun() {
  // compatibility - merge in anything from Module['preRun'] at this time
  if (Module['preRun']) {
    if (typeof Module['preRun'] == 'function') Module['preRun'] = [Module['preRun']];
    while (Module['preRun'].length) {
      addOnPreRun(Module['preRun'].shift());
    }
  }
  callRuntimeCallbacks(__ATPRERUN__);
}

function ensureInitRuntime() {
  if (runtimeInitialized) return;
  runtimeInitialized = true;
  callRuntimeCallbacks(__ATINIT__);
}

function preMain() {
  callRuntimeCallbacks(__ATMAIN__);
}

function exitRuntime() {
  callRuntimeCallbacks(__ATEXIT__);
  runtimeExited = true;
}

function postRun() {
  // compatibility - merge in anything from Module['postRun'] at this time
  if (Module['postRun']) {
    if (typeof Module['postRun'] == 'function') Module['postRun'] = [Module['postRun']];
    while (Module['postRun'].length) {
      addOnPostRun(Module['postRun'].shift());
    }
  }
  callRuntimeCallbacks(__ATPOSTRUN__);
}

function addOnPreRun(cb) {
  __ATPRERUN__.unshift(cb);
}
Module["addOnPreRun"] = addOnPreRun;

function addOnInit(cb) {
  __ATINIT__.unshift(cb);
}
Module["addOnInit"] = addOnInit;

function addOnPreMain(cb) {
  __ATMAIN__.unshift(cb);
}
Module["addOnPreMain"] = addOnPreMain;

function addOnExit(cb) {
  __ATEXIT__.unshift(cb);
}
Module["addOnExit"] = addOnExit;

function addOnPostRun(cb) {
  __ATPOSTRUN__.unshift(cb);
}
Module["addOnPostRun"] = addOnPostRun;

// Tools


function intArrayFromString(stringy, dontAddNull, length /* optional */) {
  var len = length > 0 ? length : lengthBytesUTF8(stringy)+1;
  var u8array = new Array(len);
  var numBytesWritten = stringToUTF8Array(stringy, u8array, 0, u8array.length);
  if (dontAddNull) u8array.length = numBytesWritten;
  return u8array;
}
Module["intArrayFromString"] = intArrayFromString;

function intArrayToString(array) {
  var ret = [];
  for (var i = 0; i < array.length; i++) {
    var chr = array[i];
    if (chr > 0xFF) {
      assert(false, 'Character code ' + chr + ' (' + String.fromCharCode(chr) + ')  at offset ' + i + ' not in 0x00-0xFF.');
      chr &= 0xFF;
    }
    ret.push(String.fromCharCode(chr));
  }
  return ret.join('');
}
Module["intArrayToString"] = intArrayToString;

function writeStringToMemory(string, buffer, dontAddNull) {
  var array = intArrayFromString(string, dontAddNull);
  var i = 0;
  while (i < array.length) {
    var chr = array[i];
    HEAP8[(((buffer)+(i))>>0)]=chr;
    i = i + 1;
  }
}
Module["writeStringToMemory"] = writeStringToMemory;

function writeArrayToMemory(array, buffer) {
  for (var i = 0; i < array.length; i++) {
    HEAP8[((buffer++)>>0)]=array[i];
  }
}
Module["writeArrayToMemory"] = writeArrayToMemory;

function writeAsciiToMemory(str, buffer, dontAddNull) {
  for (var i = 0; i < str.length; ++i) {
    assert(str.charCodeAt(i) === str.charCodeAt(i)&0xff);
    HEAP8[((buffer++)>>0)]=str.charCodeAt(i);
  }
  // Null-terminate the pointer to the HEAP.
  if (!dontAddNull) HEAP8[((buffer)>>0)]=0;
}
Module["writeAsciiToMemory"] = writeAsciiToMemory;

function unSign(value, bits, ignore) {
  if (value >= 0) {
    return value;
  }
  return bits <= 32 ? 2*Math.abs(1 << (bits-1)) + value // Need some trickery, since if bits == 32, we are right at the limit of the bits JS uses in bitshifts
                    : Math.pow(2, bits)         + value;
}
function reSign(value, bits, ignore) {
  if (value <= 0) {
    return value;
  }
  var half = bits <= 32 ? Math.abs(1 << (bits-1)) // abs is needed if bits == 32
                        : Math.pow(2, bits-1);
  if (value >= half && (bits <= 32 || value > half)) { // for huge values, we can hit the precision limit and always get true here. so don't do that
                                                       // but, in general there is no perfect solution here. With 64-bit ints, we get rounding and errors
                                                       // TODO: In i64 mode 1, resign the two parts separately and safely
    value = -2*half + value; // Cannot bitshift half, as it may be at the limit of the bits JS uses in bitshifts
  }
  return value;
}


// check for imul support, and also for correctness ( https://bugs.webkit.org/show_bug.cgi?id=126345 )
if (!Math['imul'] || Math['imul'](0xffffffff, 5) !== -5) Math['imul'] = function imul(a, b) {
  var ah  = a >>> 16;
  var al = a & 0xffff;
  var bh  = b >>> 16;
  var bl = b & 0xffff;
  return (al*bl + ((ah*bl + al*bh) << 16))|0;
};
Math.imul = Math['imul'];


if (!Math['clz32']) Math['clz32'] = function(x) {
  x = x >>> 0;
  for (var i = 0; i < 32; i++) {
    if (x & (1 << (31 - i))) return i;
  }
  return 32;
};
Math.clz32 = Math['clz32']

var Math_abs = Math.abs;
var Math_cos = Math.cos;
var Math_sin = Math.sin;
var Math_tan = Math.tan;
var Math_acos = Math.acos;
var Math_asin = Math.asin;
var Math_atan = Math.atan;
var Math_atan2 = Math.atan2;
var Math_exp = Math.exp;
var Math_log = Math.log;
var Math_sqrt = Math.sqrt;
var Math_ceil = Math.ceil;
var Math_floor = Math.floor;
var Math_pow = Math.pow;
var Math_imul = Math.imul;
var Math_fround = Math.fround;
var Math_min = Math.min;
var Math_clz32 = Math.clz32;

// A counter of dependencies for calling run(). If we need to
// do asynchronous work before running, increment this and
// decrement it. Incrementing must happen in a place like
// PRE_RUN_ADDITIONS (used by emcc to add file preloading).
// Note that you can add dependencies in preRun, even though
// it happens right before run - run will be postponed until
// the dependencies are met.
var runDependencies = 0;
var runDependencyWatcher = null;
var dependenciesFulfilled = null; // overridden to take different actions when all run dependencies are fulfilled
var runDependencyTracking = {};

function getUniqueRunDependency(id) {
  var orig = id;
  while (1) {
    if (!runDependencyTracking[id]) return id;
    id = orig + Math.random();
  }
  return id;
}

function addRunDependency(id) {
  runDependencies++;
  if (Module['monitorRunDependencies']) {
    Module['monitorRunDependencies'](runDependencies);
  }
  if (id) {
    assert(!runDependencyTracking[id]);
    runDependencyTracking[id] = 1;
    if (runDependencyWatcher === null && typeof setInterval !== 'undefined') {
      // Check for missing dependencies every few seconds
      runDependencyWatcher = setInterval(function() {
        if (ABORT) {
          clearInterval(runDependencyWatcher);
          runDependencyWatcher = null;
          return;
        }
        var shown = false;
        for (var dep in runDependencyTracking) {
          if (!shown) {
            shown = true;
            Module.printErr('still waiting on run dependencies:');
          }
          Module.printErr('dependency: ' + dep);
        }
        if (shown) {
          Module.printErr('(end of list)');
        }
      }, 10000);
    }
  } else {
    Module.printErr('warning: run dependency added without ID');
  }
}
Module["addRunDependency"] = addRunDependency;

function removeRunDependency(id) {
  runDependencies--;
  if (Module['monitorRunDependencies']) {
    Module['monitorRunDependencies'](runDependencies);
  }
  if (id) {
    assert(runDependencyTracking[id]);
    delete runDependencyTracking[id];
  } else {
    Module.printErr('warning: run dependency removed without ID');
  }
  if (runDependencies == 0) {
    if (runDependencyWatcher !== null) {
      clearInterval(runDependencyWatcher);
      runDependencyWatcher = null;
    }
    if (dependenciesFulfilled) {
      var callback = dependenciesFulfilled;
      dependenciesFulfilled = null;
      callback(); // can add another dependenciesFulfilled
    }
  }
}
Module["removeRunDependency"] = removeRunDependency;

Module["preloadedImages"] = {}; // maps url to image data
Module["preloadedAudios"] = {}; // maps url to audio data



var memoryInitializer = null;




// === Body ===

var ASM_CONSTS = [function() { FS.syncfs(false, function (err) { if (err) { console.log(err); } }); },
 function() { { return screen.width; } },
 function() { { return screen.height; } },
 function($0) { { if (Module['canvas']) { Module['canvas'].style['cursor'] = Module['Pointer_stringify']($0); } return 0; } },
 function() { if (Module['canvas']) { Module['canvas'].style['cursor'] = 'none'; } },
 function($0) { { if (typeof Module['setWindowTitle'] !== 'undefined') { Module['setWindowTitle'](Module['Pointer_stringify']($0)); } return 0; } },
 function($0, $1, $2) { { var w = $0; var h = $1; var pixels = $2; if (!Module['SDL2']) Module['SDL2'] = {}; var SDL2 = Module['SDL2']; if (SDL2.ctxCanvas !== Module['canvas']) { SDL2.ctx = Module['createContext'](Module['canvas'], false, true); SDL2.ctxCanvas = Module['canvas']; } if (SDL2.w !== w || SDL2.h !== h || SDL2.imageCtx !== SDL2.ctx) { SDL2.image = SDL2.ctx.createImageData(w, h); SDL2.w = w; SDL2.h = h; SDL2.imageCtx = SDL2.ctx; } var data = SDL2.image.data; var src = pixels >> 2; var dst = 0; var num; if (typeof CanvasPixelArray !== 'undefined' && data instanceof CanvasPixelArray) { num = data.length; while (dst < num) { var val = HEAP32[src]; data[dst ] = val & 0xff; data[dst+1] = (val >> 8) & 0xff; data[dst+2] = (val >> 16) & 0xff; data[dst+3] = 0xff; src++; dst += 4; } } else { if (SDL2.data32Data !== data) { SDL2.data32 = new Int32Array(data.buffer); SDL2.data8 = new Uint8Array(data.buffer); } var data32 = SDL2.data32; num = data32.length; data32.set(HEAP32.subarray(src, src + num)); var data8 = SDL2.data8; var i = 3; var j = i + 4*num; if (num % 8 == 0) { while (i < j) { data8[i] = 0xff; i = i + 4 | 0; data8[i] = 0xff; i = i + 4 | 0; data8[i] = 0xff; i = i + 4 | 0; data8[i] = 0xff; i = i + 4 | 0; data8[i] = 0xff; i = i + 4 | 0; data8[i] = 0xff; i = i + 4 | 0; data8[i] = 0xff; i = i + 4 | 0; data8[i] = 0xff; i = i + 4 | 0; } } else { while (i < j) { data8[i] = 0xff; i = i + 4 | 0; } } } SDL2.ctx.putImageData(SDL2.image, 0, 0); return 0; } },
 function() { { if (typeof(AudioContext) !== 'undefined') { return 1; } else if (typeof(webkitAudioContext) !== 'undefined') { return 1; } return 0; } },
 function() { { if(typeof(SDL2) === 'undefined') SDL2 = {}; if(typeof(SDL2.audio) === 'undefined') SDL2.audio = {}; if (!SDL2.audioContext) { if (typeof(AudioContext) !== 'undefined') { SDL2.audioContext = new AudioContext(); } else if (typeof(webkitAudioContext) !== 'undefined') { SDL2.audioContext = new webkitAudioContext(); } else { return -1; } } return 0; } },
 function() { { return SDL2.audioContext['sampleRate']; } },
 function($0, $1, $2, $3) { { SDL2.audio.scriptProcessorNode = SDL2.audioContext['createScriptProcessor']($1, 0, $0); SDL2.audio.scriptProcessorNode['onaudioprocess'] = function (e) { SDL2.audio.currentOutputBuffer = e['outputBuffer']; Runtime.dynCall('vi', $2, [$3]); }; SDL2.audio.scriptProcessorNode['connect'](SDL2.audioContext['destination']); } },
 function($0, $1) { { var numChannels = SDL2.audio.currentOutputBuffer['numberOfChannels']; for (var c = 0; c < numChannels; ++c) { var channelData = SDL2.audio.currentOutputBuffer['getChannelData'](c); if (channelData.length != $1) { throw 'Web Audio output buffer length mismatch! Destination size: ' + channelData.length + ' samples vs expected ' + $1 + ' samples!'; } for (var j = 0; j < $1; ++j) { channelData[j] = HEAPF32[$0 + ((j*numChannels + c) << 2) >> 2]; } } } },
 function($0, $1) { { Module.printErr('bad name in getProcAddress: ' + [Pointer_stringify($0), Pointer_stringify($1)]); } }];

function _emscripten_asm_const_ii(code, a0) {
 return ASM_CONSTS[code](a0);
}

function _emscripten_asm_const_i(code) {
 return ASM_CONSTS[code]();
}

function _emscripten_asm_const_iiiii(code, a0, a1, a2, a3) {
 return ASM_CONSTS[code](a0, a1, a2, a3);
}

function _emscripten_asm_const_iiii(code, a0, a1, a2) {
 return ASM_CONSTS[code](a0, a1, a2);
}

function _emscripten_asm_const_v(code) {
 return ASM_CONSTS[code]();
}

function _emscripten_asm_const_iii(code, a0, a1) {
 return ASM_CONSTS[code](a0, a1);
}



STATIC_BASE = 8;

STATICTOP = STATIC_BASE + 1282480;
  /* global initializers */  __ATINIT__.push({ func: function() { __GLOBAL__I_000101() } }, { func: function() { __GLOBAL__sub_I_Shape_cpp() } }, { func: function() { __GLOBAL__sub_I_types_cpp() } }, { func: function() { __GLOBAL__sub_I_FileData_cpp() } }, { func: function() { __GLOBAL__sub_I_Graphics_cpp() } }, { func: function() { __GLOBAL__sub_I_ParticleSystem_cpp() } }, { func: function() { __GLOBAL__sub_I_Texture_cpp() } }, { func: function() { __GLOBAL__sub_I_CompressedImageData_cpp() } }, { func: function() { __GLOBAL__sub_I_Compressor_cpp() } }, { func: function() { __GLOBAL__sub_I_MathModule_cpp() } }, { func: function() { __GLOBAL__sub_I_Cursor_cpp() } }, { func: function() { __GLOBAL__sub_I_Joint_cpp() } }, { func: function() { __GLOBAL__sub_I_System_cpp() } }, { func: function() { __GLOBAL__sub_I_System_cpp_2273() } }, { func: function() { __GLOBAL__sub_I_Window_cpp() } }, { func: function() { __GLOBAL__sub_I_Memoizer_cpp() } }, { func: function() { __GLOBAL__sub_I_Audio_cpp() } }, { func: function() { __GLOBAL__sub_I_Source_cpp() } }, { func: function() { __GLOBAL__sub_I_Event_cpp() } }, { func: function() { __GLOBAL__sub_I_File_cpp() } }, { func: function() { __GLOBAL__sub_I_GlyphData_cpp() } }, { func: function() { __GLOBAL__sub_I_TrueTypeRasterizer_cpp() } }, { func: function() { __GLOBAL__sub_I_Canvas_cpp() } }, { func: function() { __GLOBAL__sub_I_Font_cpp() } }, { func: function() { __GLOBAL__sub_I_Image_cpp() } }, { func: function() { __GLOBAL__sub_I_Mesh_cpp() } }, { func: function() { __GLOBAL__sub_I_OpenGL_cpp() } }, { func: function() { __GLOBAL__sub_I_Shader_cpp() } }, { func: function() { __GLOBAL__sub_I_Volatile_cpp() } }, { func: function() { __GLOBAL__sub_I_ImageData_cpp() } }, { func: function() { __GLOBAL__sub_I_Joystick_cpp() } }, { func: function() { __GLOBAL__sub_I_Joystick_cpp_4336() } }, { func: function() { __GLOBAL__sub_I_Keyboard_cpp() } }, { func: function() { __GLOBAL__sub_I_Keyboard_cpp_4545() } }, { func: function() { __GLOBAL__sub_I_Cursor_cpp_4586() } }, { func: function() { __GLOBAL__sub_I_Channel_cpp() } }, { func: function() { __GLOBAL__sub_I_fastmix_cpp() } }, { func: function() { __GLOBAL__sub_I_Body_cpp() } }, { func: function() { __GLOBAL__sub_I_iostream_cpp() } });
  

/* memory initializer */ allocate([92,104,2,0,139,191,2,0,24,0,0,0,0,0,0,0,52,104,2,0,14,192,2,0,92,104,2,0,174,192,2,0,80,30,0,0,0,0,0,0,92,104,2,0,243,192,2,0,224,29,0,0,0,0,0,0,92,104,2,0,128,253,3,0,24,0,0,0,0,0,0,0,92,104,2,0,160,253,3,0,24,0,0,0,0,0,0,0,92,104,2,0,139,253,4,0,32,35,0,0,0,0,0,0,92,104,2,0,86,2,5,0,128,0,0,0,0,0,0,0,92,104,2,0,115,2,5,0,24,0,0,0,0,0,0,0,92,104,2,0,134,76,5,0,8,0,0,0,0,0,0,0,92,104,2,0,160,76,5,0,176,0,0,0,0,0,0,0,92,104,2,0,193,76,5,0,24,0,0,0,0,0,0,0,92,104,2,0,53,77,5,0,24,0,0,0,0,0,0,0,92,104,2,0,75,77,5,0,176,0,0,0,0,0,0,0,92,104,2,0,4,78,5,0,128,0,0,0,0,0,0,0,92,104,2,0,99,107,5,0,128,0,0,0,0,0,0,0,92,104,2,0,142,107,5,0,16,1,0,0,0,0,0,0,52,104,2,0,170,107,5,0,92,104,2,0,113,108,5,0,16,1,0,0,0,0,0,0,56,159,9,0,60,159,9,0,92,104,2,0,9,109,5,0,8,0,0,0,0,0,0,0,92,104,2,0,107,109,5,0,24,0,0,0,0,0,0,0,92,104,2,0,147,110,5,0,96,1,0,0,0,0,0,0,92,104,2,0,171,110,5,0,8,0,0,0,0,0,0,0,92,104,2,0,24,111,5,0,24,0,0,0,0,0,0,0,92,104,2,0,73,111,5,0,144,2,0,0,0,0,0,0,92,104,2,0,101,111,5,0,8,0,0,0,0,0,0,0,92,104,2,0,63,112,5,0,240,1,0,0,0,0,0,0,92,104,2,0,99,112,5,0,128,1,0,0,0,0,0,0,92,104,2,0,137,112,5,0,128,1,0,0,0,0,0,0,92,104,2,0,172,112,5,0,128,1,0,0,0,0,0,0,92,104,2,0,209,112,5,0,128,1,0,0,0,0,0,0,92,104,2,0,241,112,5,0,80,0,0,0,0,0,0,0,92,104,2,0,13,113,5,0,128,1,0,0,0,0,0,0,92,104,2,0,45,113,5,0,128,1,0,0,0,0,0,0,92,104,2,0,79,113,5,0,48,2,0,0,0,0,0,0,52,104,2,0,122,113,5,0,92,104,2,0,140,113,5,0,72,2,0,0,0,0,0,0,52,104,2,0,185,113,5,0,40,105,2,0,246,113,5,0,0,0,0,0,4,0,0,0,24,0,0,0,2,0,0,0,128,2,0,0,2,8,0,0,96,4,0,0,2,12,0,0,136,2,0,0,2,16,0,0,52,104,2,0,18,114,5,0,52,104,2,0,38,114,5,0,92,104,2,0,51,122,5,0,24,0,0,0,0,0,0,0,92,104,2,0,73,122,5,0,208,2,0,0,0,0,0,0,92,104,2,0,120,122,5,0,96,10,0,0,0,0,0,0,92,104,2,0,198,122,5,0,96,10,0,0,0,0,0,0,92,104,2,0,98,123,5,0,8,0,0,0,0,0,0,0,92,104,2,0,118,123,5,0,128,0,0,0,0,0,0,0,92,104,2,0,149,124,5,0,0,3,0,0,0,0,0,0,92,104,2,0,191,124,5,0,8,0,0,0,0,0,0,0,92,104,2,0,252,124,5,0,8,0,0,0,0,0,0,0,52,104,2,0,41,125,5,0,92,104,2,0,212,125,5,0,56,3,0,0,0,0,0,0,92,104,2,0,251,125,5,0,8,0,0,0,0,0,0,0,92,104,2,0,15,126,5,0,88,3,0,0,0,0,0,0,92,104,2,0,39,126,5,0,8,0,0,0,0,0,0,0,92,104,2,0,87,126,5,0,120,3,0,0,0,0,0,0,92,104,2,0,114,126,5,0,8,0,0,0,0,0,0,0,92,104,2,0,152,126,5,0,32,3,0,0,0,0,0,0,92,104,2,0,192,126,5,0,216,3,0,0,0,0,0,0,92,104,2,0,7,127,5,0,184,3,0,0,0,0,0,0,92,104,2,0,44,127,5,0,24,0,0,0,0,0,0,0,92,104,2,0,81,127,5,0,184,3,0,0,0,0,0,0,92,104,2,0,120,127,5,0,232,3,0,0,0,0,0,0,92,104,2,0,147,127,5,0,24,0,0,0,0,0,0,0,92,104,2,0,162,127,5,0,8,4,0,0,0,0,0,0,92,104,2,0,206,129,5,0,8,0,0,0,0,0,0,0,92,104,2,0,27,130,5,0,40,4,0,0,0,0,0,0,52,104,2,0,42,130,5,0,92,104,2,0,32,131,5,0,40,4,0,0,0,0,0,0,92,104,2,0,48,131,5,0,40,4,0,0,0,0,0,0,92,104,2,0,62,131,5,0,40,4,0,0,0,0,0,0,52,104,2,0,209,133,5,0,52,104,2,0,122,134,5,0,92,104,2,0,133,134,5,0,104,4,0,0,0,0,0,0,92,104,2,0,158,134,5,0,104,4,0,0,0,0,0,0,92,104,2,0,225,134,5,0,104,4,0,0,0,0,0,0,92,104,2,0,253,134,5,0,104,4,0,0,0,0,0,0,52,104,2,0,60,135,5,0,92,104,2,0,141,135,5,0,176,4,0,0,0,0,0,0,92,104,2,0,84,136,5,0,176,4,0,0,0,0,0,0,92,104,2,0,91,137,5,0,176,4,0,0,0,0,0,0,92,104,2,0,236,137,5,0,176,4,0,0,0,0,0,0,92,104,2,0,194,138,5,0,176,4,0,0,0,0,0,0,92,104,2,0,83,139,5,0,176,4,0,0,0,0,0,0,92,104,2,0,146,139,5,0,176,4,0,0,0,0,0,0,92,104,2,0,214,139,5,0,176,4,0,0,0,0,0,0,92,104,2,0,206,143,5,0,96,10,0,0,0,0,0,0,52,104,2,0,137,144,5,0,92,104,2,0,171,144,5,0,24,0,0,0,0,0,0,0,92,104,2,0,187,144,5,0,112,5,0,0,0,0,0,0,92,104,2,0,212,144,5,0,8,0,0,0,0,0,0,0,92,104,2,0,248,144,5,0,240,5,0,0,0,0,0,0,92,104,2,0,18,145,5,0,32,3,0,0,0,0,0,0,92,104,2,0,57,145,5,0,112,5,0,0,0,0,0,0,92,104,2,0,221,145,5,0,24,0,0,0,0,0,0,0,92,104,2,0,4,146,5,0,240,5,0,0,0,0,0,0,92,104,2,0,170,146,5,0,96,0,0,0,0,0,0,0,92,104,2,0,22,147,5,0,96,0,0,0,0,0,0,0,92,104,2,0,129,147,5,0,24,0,0,0,0,0,0,0,92,104,2,0,166,147,5,0,32,6,0,0,0,0,0,0,92,104,2,0,93,149,5,0,24,0,0,0,0,0,0,0,92,104,2,0,115,149,5,0,8,0,0,0,0,0,0,0,92,104,2,0,216,149,5,0,96,6,0,0,0,0,0,0,92,104,2,0,137,150,5,0,80,6,0,0,0,0,0,0,92,104,2,0,173,150,5,0,24,0,0,0,0,0,0,0,92,104,2,0,222,150,5,0,8,0,0,0,0,0,0,0,92,104,2,0,145,152,5,0,176,6,0,0,0,0,0,0,92,104,2,0,9,153,5,0,144,6,0,0,0,0,0,0,92,104,2,0,51,153,5,0,48,7,0,0,0,0,0,0,92,104,2,0,220,154,5,0,48,7,0,0,0,0,0,0,92,104,2,0,51,158,6,0,8,0,0,0,0,0,0,0,92,104,2,0,69,158,6,0,128,0,0,0,0,0,0,0,92,104,2,0,99,158,6,0,32,35,0,0,0,0,0,0,92,104,2,0,117,158,6,0,208,6,0,0,0,0,0,0,92,104,2,0,159,158,6,0,208,6,0,0,0,0,0,0,92,104,2,0,195,158,6,0,208,6,0,0,0,0,0,0,92,104,2,0,242,158,6,0,128,0,0,0,0,0,0,0,92,104,2,0,29,159,6,0,48,7,0,0,0,0,0,0,92,104,2,0,59,159,6,0,24,0,0,0,0,0,0,0,40,105,2,0,95,159,6,0,0,0,0,0,2,0,0,0,208,0,0,0,2,0,0,0,136,8,0,0,2,120,0,0,40,105,2,0,10,161,6,0,0,0,0,0,2,0,0,0,24,0,0,0,2,0,0,0,136,8,0,0,2,8,0,0,92,104,2,0,198,161,6,0,240,29,0,0,0,0,0,0,92,104,2,0,58,162,6,0,136,8,0,0,0,0,0,0,92,104,2,0,116,162,6,0,144,0,0,0,0,0,0,0,40,105,2,0,32,167,6,0,0,0,0,0,2,0,0,0,208,0,0,0,2,0,0,0,136,8,0,0,2,120,0,0,92,104,2,0,235,169,6,0,176,0,0,0,0,0,0,0,92,104,2,0,214,173,6,0,160,0,0,0,0,0,0,0,52,104,2,0,254,173,6,0,92,104,2,0,31,174,6,0,240,7,0,0,0,0,0,0,92,104,2,0,73,174,6,0,240,7,0,0,0,0,0,0,92,104,2,0,116,174,6,0,240,7,0,0,0,0,0,0,40,105,2,0,159,174,6,0,0,0,0,0,2,0,0,0,24,0,0,0,2,0,0,0,136,8,0,0,2,8,0,0,92,104,2,0,135,178,6,0,176,0,0,0,0,0,0,0,92,104,2,0,177,179,6,0,176,0,0,0,0,0,0,0,40,105,2,0,206,179,6,0,0,0,0,0,2,0,0,0,176,0,0,0,2,0,0,0,136,8,0,0,2,8,0,0,52,104,2,0,234,180,6,0,92,104,2,0,4,181,6,0,160,8,0,0,0,0,0,0,92,104,2,0,31,181,6,0,8,0,0,0,0,0,0,0,92,104,2,0,69,181,6,0,32,9,0,0,0,0,0,0,92,104,2,0,231,181,6,0,208,8,0,0,0,0,0,0,92,104,2,0,8,182,6,0,24,0,0,0,0,0,0,0,92,104,2,0,81,183,6,0,208,8,0,0,0,0,0,0,92,104,2,0,223,183,6,0,168,11,0,0,0,0,0,0,92,104,2,0,109,184,6,0,208,8,0,0,0,0,0,0,92,104,2,0,99,185,6,0,168,11,0,0,0,0,0,0,92,104,2,0,223,188,6,0,128,0,0,0,0,0,0,0,92,104,2,0,69,189,6,0,64,9,0,0,0,0,0,0,92,104,2,0,99,189,6,0,24,0,0,0,0,0,0,0,92,104,2,0,125,189,6,0,96,9,0,0,0,0,0,0,92,104,2,0,162,189,6,0,8,0,0,0,0,0,0,0,92,104,2,0,238,191,6,0,128,9,0,0,0,0,0,0,92,104,2,0,12,192,6,0,8,0,0,0,0,0,0,0,92,104,2,0,246,197,6,0,24,0,0,0,0,0,0,0,92,104,2,0,7,199,6,0,112,1,0,0,0,0,0,0,92,104,2,0,169,199,6,0,64,0,0,0,0,0,0,0,92,104,2,0,196,199,6,0,240,1,0,0,0,0,0,0,92,104,2,0,135,200,6,0,240,1,0,0,0,0,0,0,92,104,2,0,170,200,6,0,24,0,0,0,0,0,0,0,92,104,2,0,232,200,6,0,128,1,0,0,0,0,0,0,92,104,2,0,13,201,6,0,240,1,0,0,0,0,0,0,92,104,2,0,45,201,6,0,24,0,0,0,0,0,0,0,92,104,2,0,221,201,6,0,128,1,0,0,0,0,0,0,92,104,2,0,2,202,6,0,128,1,0,0,0,0,0,0,92,104,2,0,64,202,6,0,128,1,0,0,0,0,0,0,92,104,2,0,98,202,6,0,128,1,0,0,0,0,0,0,92,104,2,0,132,202,6,0,112,10,0,0,0,0,0,0,92,104,2,0,162,202,6,0,24,0,0,0,0,0,0,0,92,104,2,0,184,202,6,0,96,10,0,0,0,0,0,0,92,104,2,0,90,203,6,0,160,10,0,0,0,0,0,0,52,104,2,0,115,203,6,0,92,104,2,0,136,203,6,0,184,10,0,0,0,0,0,0,52,104,2,0,168,203,6,0,92,104,2,0,196,203,6,0,24,0,0,0,0,0,0,0,40,105,2,0,219,203,6,0,0,0,0,0,2,0,0,0,24,0,0,0,2,0,0,0,32,3,0,0,2,8,0,0,92,104,2,0,106,208,6,0,104,4,0,0,0,0,0,0,92,104,2,0,132,208,6,0,104,4,0,0,0,0,0,0,92,104,2,0,245,208,6,0,104,4,0,0,0,0,0,0,92,104,2,0,215,209,6,0,176,4,0,0,0,0,0,0,92,104,2,0,86,210,6,0,176,4,0,0,0,0,0,0,92,104,2,0,81,211,6,0,176,4,0,0,0,0,0,0,92,104,2,0,165,171,7,0,96,11,0,0,0,0,0,0,52,104,2,0,182,171,7,0,92,104,2,0,203,188,7,0,80,6,0,0,0,0,0,0,92,104,2,0,21,190,7,0,208,8,0,0,0,0,0,0,92,104,2,0,233,190,7,0,224,0,0,0,0,0,0,0,92,104,2,0,149,191,7,0,208,8,0,0,0,0,0,0,92,104,2,0,89,192,7,0,24,0,0,0,0,0,0,0,92,104,2,0,243,192,7,0,200,11,0,0,0,0,0,0,52,104,2,0,13,193,7,0,0,0,0,0,0,0,0,0,85,85,85,85,85,85,229,191,85,85,85,85,85,85,229,63,146,36,73,146,36,73,210,63,17,17,17,17,17,17,193,63,132,16,66,8,33,132,176,63,16,4,65,16,4,65,160,63,8,4,2,129,64,32,144,63,16,16,16,16,16,16,128,63,128,0,1,2,4,8,112,63,4,16,64,0,1,4,96,63,0,1,8,64,0,2,80,63,16,0,1,16,0,1,64,63,1,32,0,4,128,0,48,63,0,4,0,1,64,0,32,63,128,0,64,0,32,0,16,63,16,0,16,0,16,0,0,63,154,153,153,153,153,153,233,191,154,153,153,153,153,153,217,191,154,153,153,153,153,153,217,63,154,153,153,153,153,153,233,63,28,199,113,28,199,113,236,191,28,199,113,28,199,113,220,191,28,199,113,28,199,113,204,191,28,199,113,28,199,113,204,63,28,199,113,28,199,113,220,63,28,199,113,28,199,113,236,63,51,51,51,51,51,51,227,191,31,133,235,81,184,30,225,191,31,133,235,81,184,30,213,191,174,71,225,122,20,174,199,191,82,184,30,133,235,81,184,191,203,161,69,182,243,253,164,191,134,90,211,188,227,20,141,191,172,173,216,95,118,79,110,191,0,0,0,0,0,0,0,0,0,0,0,0,0,0,224,63,0,0,0,0,0,0,240,63,0,0,0,0,0,0,248,63,0,0,0,0,0,0,4,64,0,0,0,0,0,0,18,64,0,0,0,0,0,0,33,64,0,0,0,0,0,128,48,64,0,0,0,4,107,244,52,66,0,0,0,0,0,0,0,0,0,0,0,0,0,0,224,63,0,0,0,0,0,0,240,63,0,0,0,0,0,0,248,63,0,0,0,0,0,0,0,64,0,0,0,0,0,0,4,64,0,0,0,0,0,0,18,64,0,0,0,0,0,0,33,64,0,0,0,4,107,244,52,66,93,61,127,102,158,160,230,63,0,0,0,0,0,136,57,61,68,23,117,250,82,176,230,63,0,0,0,0,0,0,216,60,254,217,11,117,18,192,230,63,0,0,0,0,0,120,40,189,191,118,212,221,220,207,230,63,0,0,0,0,0,192,30,61,41,26,101,60,178,223,230,63,0,0,0,0,0,0,216,188,227,58,89,152,146,239,230,63,0,0,0,0,0,0,188,188,134,147,81,249,125,255,230,63,0,0,0,0,0,216,47,189,163,45,244,102,116,15,231,63,0,0,0,0,0,136,44,189,195,95,236,232,117,31,231,63,0,0,0,0,0,192,19,61,5,207,234,134,130,47,231,63,0,0,0,0,0,48,56,189,82,129,165,72,154,63,231,63,0,0,0,0,0,192,0,189,252,204,215,53,189,79,231,63,0,0,0,0,0,136,47,61,241,103,66,86,235,95,231,63,0,0,0,0,0,224,3,61,72,109,171,177,36,112,231,63,0,0,0,0,0,208,39,189,56,93,222,79,105,128,231,63,0,0,0,0,0,0,221,188,0,29,172,56,185,144,231,63,0,0,0,0,0,0,227,60,120,1,235,115,20,161,231,63,0,0,0,0,0,0,237,188,96,208,118,9,123,177,231,63,0,0,0,0,0,64,32,61,51,193,48,1,237,193,231,63,0,0,0,0,0,0,160,60,54,134,255,98,106,210,231,63,0,0,0,0,0,144,38,189,59,78,207,54,243,226,231,63,0,0,0,0,0,224,2,189,232,195,145,132,135,243,231,63,0,0,0,0,0,88,36,189,78,27,62,84,39,4,232,63,0,0,0,0,0,0,51,61,26,7,209,173,210,20,232,63,0,0,0,0,0,0,15,61,126,205,76,153,137,37,232,63,0,0,0,0,0,192,33,189,208,66,185,30,76,54,232,63,0,0,0,0,0,208,41,61,181,202,35,70,26,71,232,63,0,0,0,0,0,16,71,61,188,91,159,23,244,87,232,63,0,0,0,0,0,96,34,61,175,145,68,155,217,104,232,63,0,0,0,0,0,196,50,189,149,163,49,217,202,121,232,63,0,0,0,0,0,0,35,189,184,101,138,217,199,138,232,63,0,0,0,0,0,128,42,189,0,88,120,164,208,155,232,63,0,0,0,0,0,0,237,188,35,162,42,66,229,172,232,63,0,0,0,0,0,40,51,61,250,25,214,186,5,190,232,63,0,0,0,0,0,180,66,61,131,67,181,22,50,207,232,63,0,0,0,0,0,208,46,189,76,102,8,94,106,224,232,63,0,0,0,0,0,80,32,189,7,120,21,153,174,241,232,63,0,0,0,0,0,40,40,61,14,44,40,208,254,2,233,63,0,0,0,0,0,176,28,189,150,255,145,11,91,20,233,63,0,0,0,0,0,224,5,189,249,47,170,83,195,37,233,63,0,0,0,0,0,64,245,60,74,198,205,176,55,55,233,63,0,0,0,0,0,32,23,61,174,152,95,43,184,72,233,63,0,0,0,0,0,0,9,189,203,82,200,203,68,90,233,63,0,0,0,0,0,104,37,61,33,111,118,154,221,107,233,63,0,0,0,0,0,208,54,189,42,78,222,159,130,125,233,63,0,0,0,0,0,0,1,189,163,35,122,228,51,143,233,63,0,0,0,0,0,0,45,61,4,6,202,112,241,160,233,63,0,0,0,0,0,164,56,189,137,255,83,77,187,178,233,63,0,0,0,0,0,92,53,61,91,241,163,130,145,196,233,63,0,0,0,0,0,184,38,61,197,184,75,25,116,214,233,63,0,0,0,0,0,0,236,188,142,35,227,25,99,232,233,63,0,0,0,0,0,208,23,61,2,243,7,141,94,250,233,63,0,0,0,0,0,64,22,61,77,229,93,123,102,12,234,63,0,0,0,0,0,0,245,188,246,184,142,237,122,30,234,63,0,0,0,0,0,224,9,61,39,46,74,236,155,48,234,63,0,0,0,0,0,216,42,61,93,10,70,128,201,66,234,63,0,0,0,0,0,240,26,189,155,37,62,178,3,85,234,63,0,0,0,0,0,96,11,61,19,98,244,138,74,103,234,63,0,0,0,0,0,136,56,61,167,179,48,19,158,121,234,63,0,0,0,0,0,32,17,61,141,46,193,83,254,139,234,63,0,0,0,0,0,192,6,61,210,252,121,85,107,158,234,63,0,0,0,0,0,184,41,189,184,111,53,33,229,176,234,63,0,0,0,0,0,112,43,61,129,243,211,191,107,195,234,63,0,0,0,0,0,0,217,60,128,39,60,58,255,213,234,63,0,0,0,0,0,0,228,60,163,210,90,153,159,232,234,63,0,0,0,0,0,144,44,189,103,243,34,230,76,251,234,63,0,0,0,0,0,80,22,61,144,183,141,41,7,14,235,63,0,0,0,0,0,212,47,61,169,137,154,108,206,32,235,63,0,0,0,0,0,112,18,61,75,26,79,184,162,51,235,63,0,0,0,0,0,71,77,61,231,71,183,21,132,70,235,63,0,0,0,0,0,56,56,189,58,89,229,141,114,89,235,63,0,0,0,0,0,0,152,60,106,197,241,41,110,108,235,63,0,0,0,0,0,208,10,61,80,94,251,242,118,127,235,63,0,0,0,0,0,128,222,60,178,73,39,242,140,146,235,63,0,0,0,0,0,192,4,189,3,6,161,48,176,165,235,63,0,0,0,0,0,112,13,189,102,111,154,183,224,184,235,63,0,0,0,0,0,144,13,61,255,193,75,144,30,204,235,63,0,0,0,0,0,160,2,61,111,161,243,195,105,223,235,63,0,0,0,0,0,120,31,189,184,29,215,91,194,242,235,63,0,0,0,0,0,160,16,189,233,178,65,97,40,6,236,63,0,0,0,0,0,64,17,189,224,82,133,221,155,25,236,63,0,0,0,0,0,224,11,61,238,100,250,217,28,45,236,63,0,0,0,0,0,64,9,189,47,208,255,95,171,64,236,63,0,0,0,0,0,208,14,189,21,253,250,120,71,84,236,63,0,0,0,0,0,102,57,61,203,208,87,46,241,103,236,63,0,0,0,0,0,16,26,189,182,193,136,137,168,123,236,63,0,0,0,0,128,69,88,189,51,231,6,148,109,143,236,63,0,0,0,0,0,72,26,189,223,196,81,87,64,163,236,63,0,0,0,0,0,0,203,60,148,144,239,220,32,183,236,63,0,0,0,0,0,64,1,61,137,22,109,46,15,203,236,63,0,0,0,0,0,32,240,60,18,196,93,85,11,223,236,63,0,0,0,0,0,96,243,60,59,171,91,91,21,243,236,63,0,0,0,0,0,144,6,189,188,137,7,74,45,7,237,63,0,0,0,0,0,160,9,61,250,200,8,43,83,27,237,63,0,0,0,0,0,224,21,189,133,138,13,8,135,47,237,63,0,0,0,0,0,40,29,61,3,162,202,234,200,67,237,63,0,0,0,0,0,160,1,61,145,164,251,220,24,88,237,63,0,0,0,0,0,0,223,60,161,230,98,232,118,108,237,63,0,0,0,0,0,160,3,189,78,131,201,22,227,128,237,63,0,0,0,0,0,216,12,189,144,96,255,113,93,149,237,63,0,0,0,0,0,192,244,60,174,50,219,3,230,169,237,63,0,0,0,0,0,144,255,60,37,131,58,214,124,190,237,63,0,0,0,0,0,128,233,60,69,180,1,243,33,211,237,63,0,0,0,0,0,32,245,188,191,5,28,100,213,231,237,63,0,0,0,0,0,112,29,189,236,154,123,51,151,252,237,63,0,0,0,0,0,20,22,189,94,125,25,107,103,17,238,63,0,0,0,0,0,72,11,61,231,163,245,20,70,38,238,63,0,0,0,0,0,206,64,61,92,238,22,59,51,59,238,63,0,0,0,0,0,104,12,61,180,63,139,231,46,80,238,63,0,0,0,0,0,48,9,189,104,109,103,36,57,101,238,63,0,0,0,0,0,0,229,188,68,76,199,251,81,122,238,63,0,0,0,0,0,248,7,189,38,183,205,119,121,143,238,63,0,0,0,0,0,112,243,188,232,144,164,162,175,164,238,63,0,0,0,0,0,208,229,60,228,202,124,134,244,185,238,63,0,0,0,0,0,26,22,61,13,104,142,45,72,207,238,63,0,0,0,0,0,80,245,60,20,133,24,162,170,228,238,63,0,0,0,0,0,64,198,60,19,90,97,238,27,250,238,63,0,0,0,0,0,128,238,188,6,65,182,28,156,15,239,63,0,0,0,0,0,136,250,188,99,185,107,55,43,37,239,63,0,0,0,0,0,144,44,189,117,114,221,72,201,58,239,63,0,0,0,0,0,0,170,60,36,69,110,91,118,80,239,63,0,0,0,0,0,240,244,188,253,68,136,121,50,102,239,63,0,0,0,0,0,128,202,60,56,190,156,173,253,123,239,63,0,0,0,0,0,188,250,60,130,60,36,2,216,145,239,63,0,0,0,0,0,96,212,188,142,144,158,129,193,167,239,63,0,0,0,0,0,12,11,189,17,213,146,54,186,189,239,63,0,0,0,0,0,224,192,188,148,113,143,43,194,211,239,63,0,0,0,0,128,222,16,189,238,35,42,107,217,233,239,63,0,0,0,0,0,67,238,60,0,0,0,0,0,0,240,63,0,0,0,0,0,0,0,0,190,188,90,250,26,11,240,63,0,0,0,0,0,64,179,188,3,51,251,169,61,22,240,63,0,0,0,0,0,23,18,189,130,2,59,20,104,33,240,63,0,0,0,0,0,64,186,60,108,128,119,62,154,44,240,63,0,0,0,0,0,152,239,60,202,187,17,46,212,55,240,63,0,0,0,0,0,64,199,188,137,127,110,232,21,67,240,63,0,0,0,0,0,48,216,60,103,84,246,114,95,78,240,63,0,0,0,0,0,63,26,189,90,133,21,211,176,89,240,63,0,0,0,0,0,132,2,189,149,31,60,14,10,101,240,63,0,0,0,0,0,96,241,60,26,247,221,41,107,112,240,63,0,0,0,0,0,36,21,61,45,168,114,43,212,123,240,63,0,0,0,0,0,160,233,188,208,155,117,24,69,135,240,63,0,0,0,0,0,64,230,60,200,7,102,246,189,146,240,63,0,0,0,0,0,120,0,189,131,243,198,202,62,158,240,63,0,0,0,0,0,0,152,188,48,57,31,155,199,169,240,63,0,0,0,0,0,160,255,60,252,136,249,108,88,181,240,63,0,0,0,0,0,200,250,188,138,108,228,69,241,192,240,63,0,0,0,0,0,192,217,60,22,72,114,43,146,204,240,63,0,0,0,0,0,32,5,61,216,93,57,35,59,216,240,63,0,0,0,0,0,208,250,188,243,209,211,50,236,227,240,63,0,0,0,0,0,172,27,61,166,169,223,95,165,239,240,63,0,0,0,0,0,232,4,189,240,210,254,175,102,251,240,63,0,0,0,0,0,48,13,189,75,35,215,40,48,7,241,63,0,0,0,0,0,80,241,60,91,91,18,208,1,19,241,63,0,0,0,0,0,0,236,60,249,42,94,171,219,30,241,63,0,0,0,0,0,188,22,61,213,49,108,192,189,42,241,63,0,0,0,0,0,64,232,60,125,4,242,20,168,54,241,63,0,0,0,0,0,208,14,189,233,45,169,174,154,66,241,63,0,0,0,0,0,224,232,60,56,49,79,147,149,78,241,63,0,0,0,0,0,64,235,60,113,142,165,200,152,90,241,63,0,0,0,0,0,48,5,61,223,195,113,84,164,102,241,63,0,0,0,0,0,56,3,61,17,82,125,60,184,114,241,63,0,0,0,0,0,212,40,61,159,187,149,134,212,126,241,63,0,0,0,0,0,208,5,189,147,141,140,56,249,138,241,63,0,0,0,0,0,136,28,189,102,93,55,88,38,151,241,63,0,0,0,0,0,240,17,61,167,203,111,235,91,163,241,63,0,0,0,0,0,72,16,61,227,135,19,248,153,175,241,63,0,0,0,0,0,57,71,189,84,93,4,132,224,187,241,63,0,0,0,0,0,228,36,61,67,28,40,149,47,200,241,63,0,0,0,0,0,32,10,189,178,185,104,49,135,212,241,63,0,0,0,0,0,128,227,60,49,64,180,94,231,224,241,63,0,0,0,0,0,192,234,60,56,217,252,34,80,237,241,63,0,0,0,0,0,144,1,61,247,205,56,132,193,249,241,63,0,0,0,0,0,120,27,189,143,141,98,136,59,6,242,63,0,0,0,0,0,148,45,61,30,168,120,53,190,18,242,63,0,0,0,0,0,0,216,60,65,221,125,145,73,31,242,63,0,0,0,0,0,52,43,61,35,19,121,162,221,43,242,63,0,0,0,0,0,248,25,61,231,97,117,110,122,56,242,63,0,0,0,0,0,200,25,189,39,20,130,251,31,69,242,63,0,0,0,0,0,48,2,61,2,166,178,79,206,81,242,63,0,0,0,0,0,72,19,189,176,206,30,113,133,94,242,63,0,0,0,0,0,112,18,61,22,125,226,101,69,107,242,63,0,0,0,0,0,208,17,61,15,224,29,52,14,120,242,63,0,0,0,0,0,238,49,61,62,99,245,225,223,132,242,63,0,0,0,0,0,192,20,189,48,187,145,117,186,145,242,63,0,0,0,0,0,216,19,189,9,223,31,245,157,158,242,63,0,0,0,0,0,176,8,61,155,14,209,102,138,171,242,63,0,0,0,0,0,124,34,189,58,218,218,208,127,184,242,63,0,0,0,0,0,52,42,61,249,26,119,57,126,197,242,63,0,0,0,0,0,128,16,189,217,2,228,166,133,210,242,63,0,0,0,0,0,208,14,189,121,21,100,31,150,223,242,63,0,0,0,0,0,32,244,188,207,46,62,169,175,236,242,63,0,0,0,0,0,152,36,189,34,136,189,74,210,249,242,63,0,0,0,0,0,48,22,189,37,182,49,10,254,6,243,63,0,0,0,0,0,54,50,189,11,165,238,237,50,20,243,63,0,0,0,0,128,223,112,189,184,215,76,252,112,33,243,63,0,0,0,0,0,72,34,189,162,233,168,59,184,46,243,63,0,0,0,0,0,152,37,189,102,23,100,178,8,60,243,63,0,0,0,0,0,208,30,61,39,250,227,102,98,73,243,63,0,0,0,0,0,0,220,188,15,159,146,95,197,86,243,63,0,0,0,0,0,216,48,189,185,136,222,162,49,100,243,63,0,0,0,0,0,200,34,61,57,170,58,55,167,113,243,63,0,0,0,0,0,96,32,61,254,116,30,35,38,127,243,63,0,0,0,0,0,96,22,189,56,216,5,109,174,140,243,63,0,0,0,0,0,224,10,189,195,62,113,27,64,154,243,63,0,0,0,0,0,114,68,189,32,160,229,52,219,167,243,63,0,0,0,0,0,32,8,61,149,110,236,191,127,181,243,63,0,0,0,0,0,128,62,61,242,168,19,195,45,195,243,63,0,0,0,0,0,128,239,60,34,225,237,68,229,208,243,63,0,0,0,0,0,160,23,189,187,52,18,76,166,222,243,63,0,0,0,0,0,48,38,61,204,78,28,223,112,236,243,63,0,0,0,0,0,166,72,189,140,126,172,4,69,250,243,63,0,0,0,0,0,220,60,189,187,160,103,195,34,8,244,63,0,0,0,0,0,184,37,61,149,46,247,33,10,22,244,63,0,0,0,0,0,192,30,61,70,70,9,39,251,35,244,63,0,0,0,0,0,96,19,189,32,169,80,217,245,49,244,63,0,0,0,0,0,152,35,61,235,185,132,63,250,63,244,63,0,0,0,0,0,0,250,60,25,137,97,96,8,78,244,63,0,0,0,0,0,192,246,188,1,210,167,66,32,92,244,63,0,0,0,0,0,192,11,189,22,0,29,237,65,106,244,63,0,0,0,0,0,128,18,189,38,51,139,102,109,120,244,63,0,0,0,0,0,224,48,61,0,60,193,181,162,134,244,63,0,0,0,0,0,64,45,189,4,175,146,225,225,148,244,63,0,0,0,0,0,32,12,61,114,211,215,240,42,163,244,63,0,0,0,0,0,80,30,189,1,184,109,234,125,177,244,63,0,0,0,0,0,128,7,61,225,41,54,213,218,191,244,63,0,0,0,0,0,128,19,189,50,193,23,184,65,206,244,63,0,0,0,0,0,128,0,61,219,221,253,153,178,220,244,63,0,0,0,0,0,112,44,61,150,171,216,129,45,235,244,63,0,0,0,0,0,224,28,189,2,45,157,118,178,249,244,63,0,0,0,0,0,32,25,61,193,49,69,127,65,8,245,63,0,0,0,0,0,192,8,189,42,102,207,162,218,22,245,63,0,0,0,0,0,0,250,188,234,81,63,232,125,37,245,63,0,0,0,0,0,8,74,61,218,78,157,86,43,52,245,63,0,0,0,0,0,216,38,189,26,172,246,244,226,66,245,63,0,0,0,0,0,68,50,189,219,148,93,202,164,81,245,63,0,0,0,0,0,60,72,61,107,17,233,221,112,96,245,63,0,0,0,0,0,176,36,61,222,41,181,54,71,111,245,63,0,0,0,0,0,90,65,61,14,196,226,219,39,126,245,63,0,0,0,0,0,224,41,189,111,199,151,212,18,141,245,63,0,0,0,0,0,8,35,189,76,11,255,39,8,156,245,63,0,0,0,0,0,236,77,61,39,84,72,221,7,171,245,63,0,0,0,0,0,0,196,188,244,122,168,251,17,186,245,63,0,0,0,0,0,8,48,61,11,70,89,138,38,201,245,63,0,0,0,0,0,200,38,189,63,142,153,144,69,216,245,63,0,0,0,0,0,154,70,61,225,32,173,21,111,231,245,63,0,0,0,0,0,64,27,189,202,235,220,32,163,246,245,63,0,0,0,0,0,112,23,61,184,220,118,185,225,5,246,63,0,0,0,0,0,248,38,61,21,247,205,230,42,21,246,63,0,0,0,0,0,0,1,61,49,85,58,176,126,36,246,63,0,0,0,0,0,208,21,189,181,41,25,29,221,51,246,63,0,0,0,0,0,208,18,189,19,195,204,52,70,67,246,63,0,0,0,0,0,128,234,188,250,142,188,254,185,82,246,63,0,0,0,0,0,96,40,189,151,51,85,130,56,98,246,63,0,0,0,0,0,254,113,61,142,50,8,199,193,113,246,63,0,0,0,0,0,32,55,189,126,169,76,212,85,129,246,63,0,0,0,0,0,128,230,60,113,148,158,177,244,144,246,63,0,0,0,0,0,120,41,189,92,104,2,0,65,116,9,0,224,29,0,0,0,0,0,0,92,104,2,0,127,116,9,0,224,29,0,0,0,0,0,0,92,104,2,0,152,116,9,0,232,29,0,0,0,0,0,0,92,104,2,0,176,116,9,0,232,29,0,0,0,0,0,0,92,104,2,0,236,116,9,0,200,29,0,0,0,0,0,0,52,104,2,0,22,117,9,0,92,104,2,0,40,117,9,0,200,29,0,0,0,0,0,0,52,104,2,0,82,117,9,0,52,104,2,0,131,117,9,0,40,105,2,0,180,117,9,0,0,0,0,0,1,0,0,0,184,29,0,0,3,244,255,255,40,105,2,0,227,117,9,0,0,0,0,0,1,0,0,0,208,29,0,0,3,244,255,255,40,105,2,0,18,118,9,0,0,0,0,0,1,0,0,0,184,29,0,0,3,244,255,255,40,105,2,0,65,118,9,0,0,0,0,0,1,0,0,0,208,29,0,0,3,244,255,255,40,105,2,0,112,118,9,0,3,0,0,0,2,0,0,0,240,29,0,0,2,0,0,0,32,30,0,0,2,8,0,0,92,104,2,0,160,118,9,0,0,35,0,0,0,0,0,0,92,104,2,0,186,118,9,0,16,35,0,0,0,0,0,0,92,104,2,0,21,119,9,0,160,30,0,0,0,0,0,0,92,104,2,0,41,119,9,0,240,34,0,0,0,0,0,0,92,104,2,0,63,119,9,0,160,30,0,0,0,0,0,0,40,105,2,0,83,119,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,224,30,0,0,0,0,0,0,40,105,2,0,151,119,9,0,0,0,0,0,1,0,0,0,248,30,0,0,0,0,0,0,52,104,2,0,173,119,9,0,40,105,2,0,236,119,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,32,31,0,0,0,0,0,0,40,105,2,0,48,120,9,0,0,0,0,0,1,0,0,0,248,30,0,0,0,0,0,0,40,105,2,0,70,120,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,88,31,0,0,0,0,0,0,40,105,2,0,138,120,9,0,0,0,0,0,1,0,0,0,112,31,0,0,0,0,0,0,52,104,2,0,160,120,9,0,40,105,2,0,204,120,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,152,31,0,0,0,0,0,0,40,105,2,0,16,121,9,0,0,0,0,0,1,0,0,0,112,31,0,0,0,0,0,0,40,105,2,0,38,121,9,0,0,0,0,0,3,0,0,0,160,30,0,0,2,0,0,0,216,31,0,0,2,0,0,0,224,31,0,0,0,8,0,0,52,104,2,0,107,121,9,0,52,104,2,0,126,121,9,0,40,105,2,0,224,122,9,0,0,0,0,0,3,0,0,0,160,30,0,0,2,0,0,0,216,31,0,0,2,0,0,0,16,32,0,0,0,8,0,0,52,104,2,0,37,123,9,0,40,105,2,0,71,123,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,56,32,0,0,0,8,0,0,52,104,2,0,140,123,9,0,40,105,2,0,161,123,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,56,32,0,0,0,8,0,0,40,105,2,0,251,123,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,128,32,0,0,2,0,0,0,52,104,2,0,23,124,9,0,40,105,2,0,44,124,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,128,32,0,0,2,0,0,0,40,105,2,0,72,124,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,128,32,0,0,2,0,0,0,40,105,2,0,100,124,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,128,32,0,0,2,0,0,0,40,105,2,0,128,124,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,8,33,0,0,0,0,0,0,52,104,2,0,198,124,9,0,40,105,2,0,254,124,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,48,33,0,0,0,0,0,0,52,104,2,0,68,125,9,0,40,105,2,0,104,125,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,88,33,0,0,0,0,0,0,52,104,2,0,174,125,9,0,40,105,2,0,205,125,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,128,33,0,0,0,0,0,0,52,104,2,0,19,126,9,0,40,105,2,0,44,126,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,168,33,0,0,2,0,0,0,52,104,2,0,65,126,9,0,40,105,2,0,89,126,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,168,33,0,0,2,0,0,0,92,104,2,0,110,126,9,0,224,33,0,0,0,0,0,0,40,105,2,0,144,126,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,0,34,0,0,2,0,0,0,52,104,2,0,179,126,9,0,92,104,2,0,202,126,9,0,224,33,0,0,0,0,0,0,40,105,2,0,237,126,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,0,34,0,0,2,0,0,0,40,105,2,0,15,127,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,0,34,0,0,2,0,0,0,40,105,2,0,49,127,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,0,34,0,0,2,0,0,0,92,104,2,0,84,127,9,0,160,30,0,0,0,0,0,0,40,105,2,0,106,127,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,168,34,0,0,2,0,0,0,52,104,2,0,124,127,9,0,40,105,2,0,145,127,9,0,0,0,0,0,2,0,0,0,160,30,0,0,2,0,0,0,168,34,0,0,2,0,0,0,92,104,2,0,163,127,9,0,160,30,0,0,0,0,0,0,92,104,2,0,195,127,9,0,160,30,0,0,0,0,0,0,52,104,2,0,223,127,9,0,52,104,2,0,248,127,9,0,92,104,2,0,17,128,9,0,128,35,0,0,0,0,0,0,92,104,2,0,40,128,9,0,248,34,0,0,0,0,0,0,52,104,2,0,79,129,9,0,92,104,2,0,45,129,9,0,72,35,0,0,0,0,0,0,92,104,2,0,218,128,9,0,40,35,0,0,0,0,0,0,92,104,2,0,255,128,9,0,88,35,0,0,0,0,0,0,52,104,2,0,32,129,9,0,92,104,2,0,71,130,9,0,32,35,0,0,0,0,0,0,92,104,2,0,99,130,9,0,32,35,0,0,0,0,0,0,92,104,2,0,115,130,9,0,32,35,0,0,0,0,0,0,92,104,2,0,133,130,9,0,112,35,0,0,0,0,0,0,92,104,2,0,150,130,9,0,112,35,0,0,0,0,0,0,92,104,2,0,167,130,9,0,128,35,0,0,0,0,0,0,92,104,2,0,186,130,9,0,32,35,0,0,0,0,0,0,92,104,2,0,212,130,9,0,32,35,0,0,0,0,0,0,92,104,2,0,23,131,9,0,72,35,0,0,0,0,0,0,92,104,2,0,243,130,9,0,224,35,0,0,0,0,0,0,92,104,2,0,57,131,9,0,40,35,0,0,0,0,0,0,0,0,0,0,8,0,0,0,1,0,0,0,2,0,0,0,3,0,0,0,4,0,0,0,5,0,0,0,5,0,0,0,0,0,0,0,24,0,0,0,5,0,0,0,5,0,0,0,3,0,0,0,4,0,0,0,64,0,0,0,0,0,0,0,32,0,0,0,6,0,0,0,7,0,0,0,56,0,0,0,248,255,255,255,32,0,0,0,8,0,0,0,9,0,0,0,192,255,255,255,192,255,255,255,32,0,0,0,10,0,0,0,11,0,0,0,0,0,0,0,48,0,0,0,12,0,0,0,13,0,0,0,14,0,0,0,15,0,0,0,16,0,0,0,17,0,0,0,18,0,0,0,19,0,0,0,20,0,0,0,21,0,0,0,22,0,0,0,23,0,0,0,24,0,0,0,25,0,0,0,84,36,0,0,248,36,0,0,52,37,0,0,72,37,0,0,92,37,0,0,112,37,0,0,32,37,0,0,12,37,0,0,124,36,0,0,104,36,0,0,64,0,0,0,0,0,0,0,80,30,0,0,26,0,0,0,27,0,0,0,56,0,0,0,248,255,255,255,80,30,0,0,28,0,0,0,29,0,0,0,192,255,255,255,192,255,255,255,80,30,0,0,30,0,0,0,31,0,0,0,64,0,0,0,0,0,0,0,240,29,0,0,32,0,0,0,33,0,0,0,192,255,255,255,192,255,255,255,240,29,0,0,34,0,0,0,35,0,0,0,56,0,0,0,0,0,0,0,32,30,0,0,36,0,0,0,37,0,0,0,200,255,255,255,200,255,255,255,32,30,0,0,38,0,0,0,39,0,0,0,82,193,2,0,40,0,0,0,93,193,2,0,41,0,0,0,5,204,6,0,42,0,0,0,104,193,2,0,43,0,0,0,114,193,2,0,44,0,0,0,128,193,2,0,45,0,0,0,139,193,2,0,46,0,0,0,153,193,2,0,47,0,0,0,27,109,5,0,48,0,0,0,167,193,2,0,49,0,0,0,178,193,2,0,50,0,0,0,191,193,2,0,51,0,0,0,202,193,2,0,52,0,0,0,214,193,2,0,53,0,0,0,249,203,6,0,54,0,0,0,225,193,2,0,55,0,0,0,236,193,2,0,56,0,0,0,247,193,2,0,57,0,0,0,3,194,2,0,58,0,0,0,118,156,3,0,59,0,0,0,0,0,0,0,0,0,0,0,67,224,3,0,60,0,0,0,87,224,3,0,61,0,0,0,106,224,3,0,62,0,0,0,121,224,3,0,63,0,0,0,133,224,3,0,64,0,0,0,142,224,3,0,65,0,0,0,156,224,3,0,66,0,0,0,170,224,3,0,67,0,0,0,176,224,3,0,68,0,0,0,185,224,3,0,69,0,0,0,0,0,0,0,0,0,0,0,70,0,0,0,71,0,0,0,72,0,0,0,0,0,0,0,73,0,0,0,74,0,0,0,75,0,0,0,76,0,0,0,77,0,0,0,78,0,0,0,88,238,3,0,79,0,0,0,96,238,3,0,80,0,0,0,109,238,3,0,81,0,0,0,117,238,3,0,82,0,0,0,125,238,3,0,83,0,0,0,134,238,3,0,84,0,0,0,0,0,0,0,0,0,0,0,85,0,0,0,113,248,3,0,86,0,0,0,123,248,3,0,87,0,0,0,139,248,3,0,88,0,0,0,149,248,3,0,89,0,0,0,159,248,3,0,90,0,0,0,155,114,5,0,91,0,0,0,160,114,5,0,92,0,0,0,129,115,5,0,93,0,0,0,134,115,5,0,94,0,0,0,185,115,5,0,95,0,0,0,109,191,6,0,96,0,0,0,169,248,3,0,97,0,0,0,59,17,4,0,98,0,0,0,174,114,5,0,99,0,0,0,180,248,3,0,100,0,0,0,191,248,3,0,101,0,0,0,201,248,3,0,102,0,0,0,217,248,3,0,103,0,0,0,0,0,0,0,0,0,0,0,104,0,0,0,0,0,0,0,38,249,3,0,105,0,0,0,47,249,3,0,106,0,0,0,55,249,3,0,107,0,0,0,66,249,3,0,108,0,0,0,81,249,3,0,109,0,0,0,99,249,3,0,110,0,0,0,115,249,3,0,111,0,0,0,128,249,3,0,112,0,0,0,142,249,3,0,113,0,0,0,159,249,3,0,114,0,0,0,173,249,3,0,115,0,0,0,190,249,3,0,116,0,0,0,208,249,3,0,117,0,0,0,223,249,3,0,118,0,0,0,236,249,3,0,119,0,0,0,253,249,3,0,120,0,0,0,10,250,3,0,121,0,0,0,24,250,3,0,122,0,0,0,37,250,3,0,123,0,0,0,51,250,3,0,124,0,0,0,63,250,3,0,125,0,0,0,72,250,3,0,126,0,0,0,0,0,0,0,0,0,0,0,127,0,0,0,128,0,0,0,129,0,0,0,130,0,0,0,131,0,0,0,132,0,0,0,133,0,0,0,134,0,0,0,135,0,0,0,136,0,0,0,137,0,0,0,138,0,0,0,139,0,0,0,140,0,0,0,141,0,0,0,142,0,0,0,143,0,0,0,144,0,0,0,145,0,0,0,146,0,0,0,147,0,0,0,0,0,0,0,219,118,5,0,148,0,0,0,94,250,3,0,149,0,0,0,0,0,0,0,0,0,0,0,187,251,3,0,150,0,0,0,207,251,3,0,151,0,0,0,221,251,3,0,152,0,0,0,237,251,3,0,153,0,0,0,118,250,3,0,154,0,0,0,252,251,3,0,155,0,0,0,10,252,3,0,156,0,0,0,135,250,3,0,157,0,0,0,149,250,3,0,158,0,0,0,245,250,3,0,159,0,0,0,6,251,3,0,160,0,0,0,23,251,3,0,161,0,0,0,37,251,3,0,162,0,0,0,51,251,3,0,163,0,0,0,61,251,3,0,164,0,0,0,75,251,3,0,165,0,0,0,89,251,3,0,166,0,0,0,0,0,0,0,0,0,0,0,178,250,3,0,167,0,0,0,195,250,3,0,168,0,0,0,206,250,3,0,169,0,0,0,45,120,5,0,170,0,0,0,0,0,0,0,0,0,0,0,231,250,3,0,171,0,0,0,207,251,3,0,172,0,0,0,221,251,3,0,173,0,0,0,237,251,3,0,174,0,0,0,24,252,3,0,175,0,0,0,252,251,3,0,176,0,0,0,10,252,3,0,177,0,0,0,60,252,3,0,178,0,0,0,42,252,3,0,179,0,0,0,245,250,3,0,180,0,0,0,6,251,3,0,181,0,0,0,23,251,3,0,182,0,0,0,37,251,3,0,183,0,0,0,51,251,3,0,184,0,0,0,61,251,3,0,185,0,0,0,75,251,3,0,186,0,0,0,89,251,3,0,187,0,0,0,0,0,0,0,0,0,0,0,109,251,3,0,188,0,0,0,0,0,0,0,0,0,0,0,64,120,5,0,189,0,0,0,209,117,5,0,190,0,0,0,122,251,3,0,191,0,0,0,127,119,5,0,192,0,0,0,137,119,5,0,193,0,0,0,136,251,3,0,194,0,0,0,148,251,3,0,195,0,0,0,0,0,0,0,0,0,0,0,172,121,5,0,196,0,0,0,185,121,5,0,197,0,0,0,198,121,5,0,198,0,0,0,214,121,5,0,199,0,0,0,0,0,0,0,0,0,0,0,187,251,3,0,200,0,0,0,207,251,3,0,201,0,0,0,221,251,3,0,202,0,0,0,237,251,3,0,203,0,0,0,252,251,3,0,204,0,0,0,10,252,3,0,205,0,0,0,24,252,3,0,206,0,0,0,42,252,3,0,207,0,0,0,60,252,3,0,208,0,0,0,75,252,3,0,209,0,0,0,94,252,3,0,210,0,0,0,113,252,3,0,211,0,0,0,135,252,3,0,212,0,0,0,0,0,0,0,0,0,0,0,251,65,5,0,213,0,0,0,195,252,3,0,214,0,0,0,208,252,3,0,215,0,0,0,221,252,3,0,216,0,0,0,238,252,3,0,217,0,0,0,255,252,3,0,218,0,0,0,10,253,3,0,219,0,0,0,21,253,3,0,220,0,0,0,218,116,5,0,221,0,0,0,237,116,5,0,222,0,0,0,37,253,3,0,223,0,0,0,46,253,3,0,224,0,0,0,59,253,3,0,225,0,0,0,73,253,3,0,226,0,0,0,89,253,3,0,227,0,0,0,75,117,5,0,228,0,0,0,88,117,5,0,229,0,0,0,101,253,3,0,230,0,0,0,137,119,5,0,231,0,0,0,172,120,5,0,232,0,0,0,180,120,5,0,233,0,0,0,0,0,0,0,0,0,0,0,87,170,6,0,1,0,0,0,79,170,6,0,2,0,0,0,118,253,3,0,3,0,0,0,220,14,5,0,1,0,0,0,239,14,5,0,2,0,0,0,149,253,3,0,3,0,0,0,154,253,3,0,4,0,0,0,87,255,4,0,234,0,0,0,122,255,4,0,235,0,0,0,0,0,0,0,0,0,0,0,236,0,0,0,237,0,0,0,0,0,0,0,99,124,5,0,238,0,0,0,111,124,5,0,239,0,0,0,123,124,5,0,240,0,0,0,222,253,3,0,241,0,0,0,137,124,5,0,242,0,0,0,237,253,3,0,243,0,0,0,247,253,3,0,244,0,0,0,0,0,0,0,0,0,0,0,141,14,4,0,245,0,0,0,147,14,4,0,246,0,0,0,165,14,4,0,247,0,0,0,182,14,4,0,248,0,0,0,199,14,4,0,249,0,0,0,212,14,4,0,250,0,0,0,220,14,4,0,251,0,0,0,0,0,0,0,0,0,0,0,240,14,4,0,252,0,0,0,250,14,4,0,253,0,0,0,5,15,4,0,254,0,0,0,0,0,0,0,0,0,0,0,255,0,0,0,0,1,0,0,0,0,0,0,16,240,7,0,1,1,0,0,22,15,4,0,2,1,0,0,31,15,4,0,3,1,0,0,38,15,4,0,4,1,0,0,161,195,6,0,5,1,0,0,54,15,4,0,6,1,0,0,0,0,0,0,0,0,0,0,77,15,4,0,7,1,0,0,174,114,5,0,8,1,0,0,88,15,4,0,9,1,0,0,0,0,0,0,0,0,0,0,183,9,5,0,10,1,0,0,0,0,0,0,0,0,0,0,11,1,0,0,0,0,0,0,179,15,4,0,12,1,0,0,120,152,5,0,13,1,0,0,80,1,5,0,14,1,0,0,238,192,6,0,15,1,0,0,147,245,7,0,16,1,0,0,85,1,5,0,17,1,0,0,134,151,5,0,18,1,0,0,214,1,5,0,19,1,0,0,0,0,0,0,0,0,0,0,106,16,4,0,20,1,0,0,122,16,4,0,21,1,0,0,137,16,4,0,22,1,0,0,79,152,5,0,23,1,0,0,145,16,4,0,24,1,0,0,164,16,4,0,25,1,0,0,178,16,4,0,26,1,0,0,97,151,5,0,27,1,0,0,108,14,5,0,27,1,0,0,11,245,7,0,28,1,0,0,192,16,4,0,29,1,0,0,185,115,5,0,30,1,0,0,174,114,5,0,31,1,0,0,213,16,4,0,32,1,0,0,221,16,4,0,33,1,0,0,229,16,4,0,34,1,0,0,252,16,4,0,35,1,0,0,18,17,4,0,36,1,0,0,27,17,4,0,37,1,0,0,36,17,4,0,38,1,0,0,45,17,4,0,39,1,0,0,59,17,4,0,40,1,0,0,69,17,4,0,41,1,0,0,83,17,4,0,42,1,0,0,92,17,4,0,43,1,0,0,103,17,4,0,44,1,0,0,112,17,4,0,45,1,0,0,121,17,4,0,46,1,0,0,136,17,4,0,47,1,0,0,0,0,0,0,0,0,0,0,163,17,4,0,48,1,0,0,0,18,4,0,49,1,0,0,34,18,4,0,50,1,0,0,186,141,5,0,51,1,0,0,54,245,7,0,52,1,0,0,225,142,5,0,53,1,0,0,0,19,4,0,54,1,0,0,14,19,4,0,55,1,0,0,30,19,4,0,56,1,0,0,64,19,4,0,56,1,0,0,172,120,5,0,57,1,0,0,83,19,4,0,58,1,0,0,99,19,4,0,59,1,0,0,119,19,4,0,60,1,0,0,132,19,4,0,61,1,0,0,143,19,4,0,62,1,0,0,0,0,0,0,0,0,0,0,245,17,4,0,63,1,0,0,171,19,4,0,64,1,0,0,186,19,4,0,65,1,0,0,221,65,5,0,66,1,0,0,203,19,4,0,67,1,0,0,123,142,5,0,68,1,0,0,168,142,5,0,69,1,0,0,208,19,4,0,70,1,0,0,227,19,4,0,71,1,0,0,186,187,7,0,72,1,0,0,51,236,7,0,73,1,0,0,15,20,4,0,74,1,0,0,163,20,4,0,75,1,0,0,174,20,4,0,76,1,0,0,190,20,4,0,77,1,0,0,0,0,0,0,0,0,0,0,216,20,4,0,78,1,0,0,21,21,4,0,79,1,0,0,0,0,0,0,0,0,0,0,80,1,0,0,81,1,0,0,82,1,0,0,0,0,0,0,8,0,0,0,48,0,0,0,44,0,0,0,8,0,0,0,4,0,0,0,6,0,0,0,8,0,0,0,24,0,0,0,8,0,0,0,12,0,0,0,16,0,0,0,24,0,0,0,170,248,4,0,83,1,0,0,239,248,4,0,84,1,0,0,27,249,7,0,85,1,0,0,51,251,7,0,86,1,0,0,104,249,4,0,87,1,0,0,110,249,4,0,0,0,0,0,0,0,0,0,0,0,0,0,255,0,0,0,127,0,0,0,255,7,0,0,255,255,0,0,0,0,0,0,96,0,0,0,88,1,0,0,89,1,0,0,90,1,0,0,55,255,4,0,91,1,0,0,65,255,4,0,92,1,0,0,10,151,5,0,93,1,0,0,0,0,0,0,0,0,0,0,222,255,4,0,94,1,0,0,237,255,4,0,95,1,0,0,80,1,5,0,96,1,0,0,3,240,7,0,97,1,0,0,238,192,6,0,98,1,0,0,95,242,7,0,99,1,0,0,85,1,5,0,100,1,0,0,110,1,5,0,101,1,0,0,120,1,5,0,102,1,0,0,185,115,5,0,103,1,0,0,174,114,5,0,104,1,0,0,247,255,4,0,105,1,0,0,6,0,5,0,106,1,0,0,130,1,5,0,107,1,0,0,142,1,5,0,108,1,0,0,21,0,5,0,109,1,0,0,37,0,5,0,110,1,0,0,53,0,5,0,111,1,0,0,70,0,5,0,112,1,0,0,0,0,0,0,0,0,0,0,113,1,0,0,0,0,0,0,126,63,5,0,114,1,0,0,80,1,5,0,115,1,0,0,3,240,7,0,116,1,0,0,238,192,6,0,117,1,0,0,95,242,7,0,118,1,0,0,85,1,5,0,119,1,0,0,92,1,5,0,120,1,0,0,101,1,5,0,121,1,0,0,110,1,5,0,122,1,0,0,120,1,5,0,123,1,0,0,147,245,7,0,124,1,0,0,134,151,5,0,125,1,0,0,137,124,5,0,126,1,0,0,185,115,5,0,127,1,0,0,174,114,5,0,128,1,0,0,130,1,5,0,129,1,0,0,142,1,5,0,130,1,0,0,49,64,5,0,131,1,0,0,62,64,5,0,132,1,0,0,154,1,5,0,133,1,0,0,162,1,5,0,134,1,0,0,170,1,5,0,135,1,0,0,182,1,5,0,136,1,0,0,193,1,5,0,137,1,0,0,204,1,5,0,138,1,0,0,241,65,5,0,139,1,0,0,232,65,5,0,140,1,0,0,214,1,5,0,141,1,0,0,224,1,5,0,142,1,0,0,240,1,5,0,143,1,0,0,0,2,5,0,144,1,0,0,24,2,5,0,145,1,0,0,48,2,5,0,146,1,0,0,59,2,5,0,147,1,0,0,99,124,5,0,148,1,0,0,64,120,5,0,149,1,0,0,0,0,0,0,0,0,0,0,76,2,5,0,150,1,0,0,81,2,5,0,151,1,0,0,231,58,9,0,152,1,0,0,76,125,5,0,153,1,0,0,8,194,6,0,154,1,0,0,31,149,5,0,155,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,112,0,0,0,156,1,0,0,157,1,0,0,3,0,0,0,4,0,0,0,158,1,0,0,159,1,0,0,74,246,7,0,0,0,0,0,128,2,5,0,1,0,0,0,195,251,7,0,160,1,0,0,192,3,5,0,161,1,0,0,201,3,5,0,162,1,0,0,209,3,5,0,163,1,0,0,233,3,5,0,164,1,0,0,245,3,5,0,165,1,0,0,1,4,5,0,166,1,0,0,180,68,5,0,167,1,0,0,11,4,5,0,168,1,0,0,17,4,5,0,169,1,0,0,194,2,5,0,170,1,0,0,25,4,5,0,171,1,0,0,45,4,5,0,172,1,0,0,62,4,5,0,173,1,0,0,82,4,5,0,174,1,0,0,99,4,5,0,175,1,0,0,122,4,5,0,176,1,0,0,139,4,5,0,177,1,0,0,157,4,5,0,178,1,0,0,164,4,5,0,179,1,0,0,34,106,5,0,180,1,0,0,176,4,5,0,181,1,0,0,186,4,5,0,182,1,0,0,178,248,7,0,183,1,0,0,89,245,7,0,184,1,0,0,182,245,7,0,185,1,0,0,202,4,5,0,186,1,0,0,209,4,5,0,187,1,0,0,60,245,7,0,188,1,0,0,173,240,7,0,189,1,0,0,227,4,5,0,190,1,0,0,10,151,5,0,191,1,0,0,243,4,5,0,192,1,0,0,6,5,5,0,193,1,0,0,25,5,5,0,194,1,0,0,37,5,5,0,195,1,0,0,52,5,5,0,196,1,0,0,0,0,0,0,0,0,0,0,197,1,0,0,198,1,0,0,199,1,0,0,0,0,0,0,26,7,5,0,200,1,0,0,106,5,5,0,201,1,0,0,128,5,5,0,202,1,0,0,70,7,5,0,203,1,0,0,148,5,5,0,204,1,0,0,0,0,0,0,0,0,0,0,205,1,0,0,206,1,0,0,0,0,0,0,113,180,6,0,207,1,0,0,103,180,6,0,208,1,0,0,89,80,5,0,209,1,0,0,197,5,5,0,210,1,0,0,206,5,5,0,211,1,0,0,243,5,5,0,212,1,0,0,221,5,5,0,213,1,0,0,213,119,5,0,214,1,0,0,56,180,6,0,215,1,0,0,0,0,0,0,0,0,0,0,103,180,6,0,216,1,0,0,243,5,5,0,217,1,0,0,178,180,6,0,218,1,0,0,188,180,6,0,219,1,0,0,144,180,6,0,220,1,0,0,254,5,5,0,221,1,0,0,11,6,5,0,222,1,0,0,211,180,6,0,223,1,0,0,0,0,0,0,0,0,0,0,221,65,5,0,224,1,0,0,8,194,6,0,225,1,0,0,250,11,5,0,226,1,0,0,2,12,5,0,227,1,0,0,10,12,5,0,228,1,0,0,19,12,5,0,229,1,0,0,27,12,5,0,230,1,0,0,35,12,5,0,231,1,0,0,48,12,5,0,232,1,0,0,63,12,5,0,233,1,0,0,81,12,5,0,234,1,0,0,91,12,5,0,235,1,0,0,101,12,5,0,236,1,0,0,109,12,5,0,237,1,0,0,117,12,5,0,238,1,0,0,48,68,5,0,239,1,0,0,57,68,5,0,240,1,0,0,127,12,5,0,241,1,0,0,146,12,5,0,242,1,0,0,165,12,5,0,243,1,0,0,148,68,5,0,244,1,0,0,156,68,5,0,245,1,0,0,176,12,5,0,246,1,0,0,189,12,5,0,247,1,0,0,202,12,5,0,248,1,0,0,215,12,5,0,249,1,0,0,228,12,5,0,250,1,0,0,245,12,5,0,251,1,0,0,6,13,5,0,252,1,0,0,29,13,5,0,253,1,0,0,52,13,5,0,254,1,0,0,65,13,5,0,255,1,0,0,78,13,5,0,0,2,0,0,90,13,5,0,1,2,0,0,103,13,5,0,2,2,0,0,116,13,5,0,3,2,0,0,128,13,5,0,4,2,0,0,141,13,5,0,5,2,0,0,154,13,5,0,6,2,0,0,167,13,5,0,7,2,0,0,179,13,5,0,8,2,0,0,193,13,5,0,9,2,0,0,203,13,5,0,10,2,0,0,213,13,5,0,11,2,0,0,223,13,5,0,12,2,0,0,233,13,5,0,13,2,0,0,255,13,5,0,14,2,0,0,12,14,5,0,15,2,0,0,29,14,5,0,16,2,0,0,55,14,5,0,17,2,0,0,71,14,5,0,18,2,0,0,87,14,5,0,19,2,0,0,96,14,5,0,20,2,0,0,15,241,7,0,21,2,0,0,101,14,5,0,22,2,0,0,108,14,5,0,23,2,0,0,201,116,5,0,24,2,0,0,118,14,5,0,25,2,0,0,113,180,6,0,26,2,0,0,103,180,6,0,27,2,0,0,89,80,5,0,28,2,0,0,133,14,5,0,29,2,0,0,144,14,5,0,30,2,0,0,161,14,5,0,31,2,0,0,172,14,5,0,32,2,0,0,180,14,5,0,33,2,0,0,195,14,5,0,34,2,0,0,114,25,9,0,35,2,0,0,177,245,7,0,36,2,0,0,210,14,5,0,37,2,0,0,220,14,5,0,38,2,0,0,227,14,5,0,39,2,0,0,235,14,5,0,40,2,0,0,239,14,5,0,41,2,0,0,76,125,5,0,42,2,0,0,88,125,5,0,43,2,0,0,46,110,5,0,44,2,0,0,53,110,5,0,45,2,0,0,36,110,5,0,46,2,0,0,247,14,5,0,47,2,0,0,253,14,5,0,48,2,0,0,0,0,0,0,0,0,0,0,49,2,0,0,50,2,0,0,51,2,0,0,52,2,0,0,53,2,0,0,54,2,0,0,55,2,0,0,56,2,0,0,57,2,0,0,58,2,0,0,59,2,0,0,60,2,0,0,0,0,0,0,78,60,5,0,61,2,0,0,94,60,5,0,62,2,0,0,231,79,5,0,63,2,0,0,110,60,5,0,64,2,0,0,118,60,5,0,65,2,0,0,126,60,5,0,66,2,0,0,0,0,0,0,0,0,0,0,73,61,5,0,67,2,0,0,85,61,5,0,68,2,0,0,95,61,5,0,69,2,0,0,105,61,5,0,70,2,0,0,124,61,5,0,71,2,0,0,182,117,5,0,72,2,0,0,143,61,5,0,73,2,0,0,159,61,5,0,74,2,0,0,179,61,5,0,75,2,0,0,94,68,5,0,76,2,0,0,54,245,7,0,77,2,0,0,198,61,5,0,78,2,0,0,211,61,5,0,79,2,0,0,26,68,5,0,80,2,0,0,37,68,5,0,81,2,0,0,224,61,5,0,82,2,0,0,236,61,5,0,83,2,0,0,248,61,5,0,84,2,0,0,5,62,5,0,85,2,0,0,0,0,0,0,0,0,0,0,126,63,5,0,86,2,0,0,26,68,5,0,87,2,0,0,37,68,5,0,88,2,0,0,66,68,5,0,89,2,0,0,80,68,5,0,90,2,0,0,132,63,5,0,91,2,0,0,146,63,5,0,92,2,0,0,160,63,5,0,93,2,0,0,176,63,5,0,94,2,0,0,192,63,5,0,95,2,0,0,211,63,5,0,96,2,0,0,230,63,5,0,97,2,0,0,250,63,5,0,98,2,0,0,185,115,5,0,99,2,0,0,174,114,5,0,100,2,0,0,14,64,5,0,101,2,0,0,21,64,5,0,102,2,0,0,35,64,5,0,103,2,0,0,49,64,5,0,104,2,0,0,62,64,5,0,105,2,0,0,75,64,5,0,106,2,0,0,85,64,5,0,107,2,0,0,95,64,5,0,108,2,0,0,104,64,5,0,109,2,0,0,113,64,5,0,110,2,0,0,135,64,5,0,111,2,0,0,157,64,5,0,112,2,0,0,179,64,5,0,113,2,0,0,201,64,5,0,114,2,0,0,227,64,5,0,115,2,0,0,4,116,5,0,116,2,0,0,34,115,5,0,117,2,0,0,253,64,5,0,118,2,0,0,6,65,5,0,119,2,0,0,15,65,5,0,120,2,0,0,32,65,5,0,121,2,0,0,49,65,5,0,122,2,0,0,61,65,5,0,123,2,0,0,73,65,5,0,124,2,0,0,81,65,5,0,125,2,0,0,89,65,5,0,126,2,0,0,106,65,5,0,127,2,0,0,123,65,5,0,128,2,0,0,133,65,5,0,129,2,0,0,143,65,5,0,130,2,0,0,152,65,5,0,131,2,0,0,161,65,5,0,132,2,0,0,171,65,5,0,133,2,0,0,181,65,5,0,134,2,0,0,201,65,5,0,135,2,0,0,104,125,5,0,136,2,0,0,52,82,9,0,137,2,0,0,3,240,7,0,138,2,0,0,238,192,6,0,139,2,0,0,221,65,5,0,140,2,0,0,227,65,5,0,141,2,0,0,201,116,5,0,142,2,0,0,232,65,5,0,143,2,0,0,241,65,5,0,144,2,0,0,251,65,5,0,145,2,0,0,0,0,0,0,0,0,0,0,143,67,5,0,146,2,0,0,155,67,5,0,147,2,0,0,163,67,5,0,147,2,0,0,175,67,5,0,148,2,0,0,185,67,5,0,149,2,0,0,195,67,5,0,150,2,0,0,206,67,5,0,151,2,0,0,168,142,5,0,152,2,0,0,218,67,5,0,153,2,0,0,0,0,0,0,0,0,0,0,53,75,5,0,154,2,0,0,156,245,7,0,155,2,0,0,8,194,6,0,156,2,0,0,54,245,7,0,157,2,0,0,26,68,5,0,158,2,0,0,37,68,5,0,159,2,0,0,48,68,5,0,160,2,0,0,57,68,5,0,161,2,0,0,104,125,5,0,162,2,0,0,66,68,5,0,163,2,0,0,80,68,5,0,164,2,0,0,94,68,5,0,165,2,0,0,0,0,0,0,0,0,0,0,156,245,7,0,166,2,0,0,138,68,5,0,167,2,0,0,53,75,5,0,168,2,0,0,143,68,5,0,169,2,0,0,8,194,6,0,170,2,0,0,148,68,5,0,171,2,0,0,156,68,5,0,172,2,0,0,113,180,6,0,173,2,0,0,103,180,6,0,174,2,0,0,89,80,5,0,175,2,0,0,0,0,0,0,0,0,0,0,170,68,5,0,176,2,0,0,180,68,5,0,177,2,0,0,190,68,5,0,178,2,0,0,113,180,6,0,179,2,0,0,103,180,6,0,180,2,0,0,89,80,5,0,181,2,0,0,158,180,6,0,182,2,0,0,168,180,6,0,183,2,0,0,0,0,0,0,0,0,0,0,177,245,7,0,0,0,0,0,38,75,5,0,1,0,0,0,16,246,7,0,0,0,0,0,106,187,7,0,1,0,0,0,43,75,5,0,2,0,0,0,47,75,5,0,0,0,0,0,53,75,5,0,1,0,0,0,57,75,5,0,2,0,0,0,66,75,5,0,3,0,0,0,75,75,5,0,4,0,0,0,83,75,5,0,5,0,0,0,90,75,5,0,6,0,0,0,97,75,5,0,7,0,0,0,105,75,5,0,0,0,0,0,119,75,5,0,1,0,0,0,183,12,9,0,1,0,0,0,133,75,5,0,0,0,0,0,83,3,8,0,0,0,0,0,139,75,5,0,1,0,0,0,145,75,5,0,2,0,0,0,97,75,5,0,0,0,0,0,151,75,5,0,1,0,0,0,161,75,5,0,2,0,0,0,171,75,5,0,3,0,0,0,185,75,5,0,4,0,0,0,199,75,5,0,5,0,0,0,206,75,5,0,0,0,0,0,211,75,5,0,1,0,0,0,218,75,5,0,2,0,0,0,224,75,5,0,3,0,0,0,231,75,5,0,4,0,0,0,239,75,5,0,5,0,0,0,248,75,5,0,6,0,0,0,255,75,5,0,0,0,0,0,106,77,5,0,1,0,0,0,75,75,5,0,2,0,0,0,18,76,5,0,0,0,0,0,28,76,5,0,1,0,0,0,40,76,5,0,2,0,0,0,52,76,5,0,3,0,0,0,254,13,8,0,0,0,0,0,63,76,5,0,1,0,0,0,73,76,5,0,0,0,0,0,83,76,5,0,1,0,0,0,98,76,5,0,2,0,0,0,107,76,5,0,3,0,0,0,114,76,5,0,4,0,0,0,120,76,5,0,5,0,0,0,0,0,0,0,160,0,0,0,184,2,0,0,185,2,0,0,3,0,0,0,4,0,0,0,5,0,0,0,5,0,0,0,186,2,0,0,83,3,8,0,0,0,0,0,239,76,5,0,1,0,0,0,168,242,7,0,2,0,0,0,247,76,5,0,0,0,0,0,251,76,5,0,1,0,0,0,228,246,7,0,2,0,0,0,0,0,0,0,192,0,0,0,187,2,0,0,188,2,0,0,3,0,0,0,4,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,128,63,0,0,0,0,208,0,0,0,189,2,0,0,190,2,0,0,3,0,0,0,4,0,0,0,5,0,0,0,5,0,0,0,191,2,0,0,192,2,0,0,5,0,0,0,193,2,0,0,5,0,0,0,194,2,0,0,195,2,0,0,5,0,0,0,110,168,6,0,1,0,0,0,194,56,9,0,2,0,0,0,83,3,8,0,0,0,0,0,100,77,5,0,0,0,0,0,106,77,5,0,1,0,0,0,134,238,7,0,2,0,0,0,116,77,5,0,3,0,0,0,136,77,5,0,196,2,0,0,148,77,5,0,197,2,0,0,0,0,0,0,0,0,0,0,113,180,6,0,198,2,0,0,103,180,6,0,199,2,0,0,89,80,5,0,200,2,0,0,158,180,6,0,201,2,0,0,168,180,6,0,202,2,0,0,181,77,5,0,203,2,0,0,122,180,6,0,204,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,224,0,0,0,205,2,0,0,206,2,0,0,3,0,0,0,4,0,0,0,207,2,0,0,208,2,0,0,56,192,6,0,0,0,0,0,39,78,5,0,1,0,0,0,44,78,5,0,2,0,0,0,49,78,5,0,3,0,0,0,54,78,5,0,4,0,0,0,58,78,5,0,5,0,0,0,63,78,5,0,6,0,0,0,67,78,5,0,7,0,0,0,72,78,5,0,8,0,0,0,77,78,5,0,9,0,0,0,83,78,5,0,10,0,0,0,87,78,5,0,11,0,0,0,96,78,5,0,12,0,0,0,105,78,5,0,13,0,0,0,115,78,5,0,14,0,0,0,125,78,5,0,15,0,0,0,130,78,5,0,16,0,0,0,138,78,5,0,17,0,0,0,147,78,5,0,18,0,0,0,157,78,5,0,19,0,0,0,162,78,5,0,20,0,0,0,168,78,5,0,21,0,0,0,174,78,5,0,22,0,0,0,181,78,5,0,23,0,0,0,189,78,5,0,24,0,0,0,197,78,5,0,25,0,0,0,205,78,5,0,26,0,0,0,213,78,5,0,27,0,0,0,221,78,5,0,28,0,0,0,229,78,5,0,29,0,0,0,237,78,5,0,30,0,0,0,245,78,5,0,31,0,0,0,254,78,5,0,32,0,0,0,7,79,5,0,33,0,0,0,16,79,5,0,34,0,0,0,26,79,5,0,35,0,0,0,36,79,5,0,36,0,0,0,113,180,6,0,209,2,0,0,103,180,6,0,210,2,0,0,89,80,5,0,211,2,0,0,97,79,5,0,212,2,0,0,56,180,6,0,213,2,0,0,0,0,0,0,0,0,0,0,43,180,6,0,214,2,0,0,213,79,5,0,215,2,0,0,231,79,5,0,216,2,0,0,0,0,0,0,0,0,0,0,217,2,0,0,218,2,0,0,0,0,0,0,113,180,6,0,219,2,0,0,103,180,6,0,220,2,0,0,89,80,5,0,221,2,0,0,103,80,5,0,222,2,0,0,112,80,5,0,223,2,0,0,215,193,6,0,224,2,0,0,121,80,5,0,225,2,0,0,128,80,5,0,226,2,0,0,144,80,5,0,227,2,0,0,0,0,0,0,0,0,0,0,228,2,0,0,229,2,0,0,50,106,5,0,230,2,0,0,63,106,5,0,231,2,0,0,80,106,5,0,232,2,0,0,220,191,6,0,233,2,0,0,98,106,5,0,234,2,0,0,118,106,5,0,235,2,0,0,0,0,0,0,0,0,0,0,236,2,0,0,0,0,0,0,232,106,5,0,237,2,0,0,245,106,5,0,238,2,0,0,2,107,5,0,239,2,0,0,15,107,5,0,240,2,0,0,28,107,5,0,241,2,0,0,109,191,6,0,242,2,0,0,46,107,5,0,243,2,0,0,61,107,5,0,244,2,0,0,80,107,5,0,245,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,240,0,0,0,246,2,0,0,247,2,0,0,3,0,0,0,4,0,0,0,248,2,0,0,249,2,0,0,128,107,5,0,0,0,0,0,132,107,5,0,1,0,0,0,137,107,5,0,2,0,0,0,0,0,0,0,0,1,0,0,250,2,0,0,251,2,0,0,252,2,0,0,253,2,0,0,254,2,0,0,0,0,0,0,24,1,0,0,250,2,0,0,255,2,0,0,0,3,0,0,1,3,0,0,2,3,0,0,0,0,0,0,48,1,0,0,3,3,0,0,4,3,0,0,3,0,0,0,4,0,0,0,5,3,0,0,6,3,0,0,0,0,0,0,64,1,0,0,7,3,0,0,8,3,0,0,3,0,0,0,4,0,0,0,177,109,5,0,9,3,0,0,187,109,5,0,10,3,0,0,201,109,5,0,11,3,0,0,217,109,5,0,12,3,0,0,233,109,5,0,13,3,0,0,252,109,5,0,14,3,0,0,15,110,5,0,15,3,0,0,36,110,5,0,16,3,0,0,46,110,5,0,17,3,0,0,53,110,5,0,18,3,0,0,59,110,5,0,19,3,0,0,68,110,5,0,20,3,0,0,79,110,5,0,21,3,0,0,86,110,5,0,22,3,0,0,0,0,0,0,0,0,0,0,56,180,6,0,23,3,0,0,0,0,0,0,0,0,0,0,0,0,0,0,80,1,0,0,24,3,0,0,25,3,0,0,3,0,0,0,4,0,0,0,26,3,0,0,27,3,0,0,28,3,0,0,29,3,0,0,30,3,0,0,31,3,0,0,32,3,0,0,33,3,0,0,34,3,0,0,35,3,0,0,36,3,0,0,37,3,0,0,38,3,0,0,39,3,0,0,40,3,0,0,41,3,0,0,42,3,0,0,43,3,0,0,44,3,0,0,45,3,0,0,46,3,0,0,206,110,5,0,0,0,0,0,212,110,5,0,1,0,0,0,231,58,9,0,2,0,0,0,221,58,9,0,3,0,0,0,218,110,5,0,4,0,0,0,228,110,5,0,5,0,0,0,237,110,5,0,6,0,0,0,246,110,5,0,7,0,0,0,253,110,5,0,8,0,0,0,4,111,5,0,9,0,0,0,25,18,8,0,10,0,0,0,12,111,5,0,11,0,0,0,17,111,5,0,0,0,0,0,215,175,6,0,1,0,0,0,64,120,5,0,47,3,0,0,0,0,0,0,0,0,0,0,0,0,0,0,128,1,0,0,48,3,0,0,49,3,0,0,3,0,0,0,4,0,0,0,30,0,0,0,0,0,0,0,144,1,0,0,1,0,0,0,50,3,0,0,3,0,0,0,4,0,0,0,51,3,0,0,52,3,0,0,0,0,0,0,160,1,0,0,53,3,0,0,54,3,0,0,3,0,0,0,4,0,0,0,0,0,0,0,176,1,0,0,55,3,0,0,56,3,0,0,3,0,0,0,4,0,0,0,0,0,0,0,192,1,0,0,57,3,0,0,58,3,0,0,3,0,0,0,4,0,0,0,0,0,0,0,208,1,0,0,59,3,0,0,60,3,0,0,3,0,0,0,4,0,0,0,0,0,0,0,224,1,0,0,61,3,0,0,62,3,0,0,3,0,0,0,4,0,0,0,0,0,0,0,240,1,0,0,63,3,0,0,64,3,0,0,3,0,0,0,4,0,0,0,0,0,0,0,0,2,0,0,65,3,0,0,66,3,0,0,3,0,0,0,4,0,0,0,0,0,0,0,16,2,0,0,67,3,0,0,68,3,0,0,3,0,0,0,4,0,0,0,0,0,0,0,32,2,0,0,69,3,0,0,70,3,0,0,71,3,0,0,0,0,0,0,56,2,0,0,72,3,0,0,73,3,0,0,74,3,0,0,0,0,0,0,80,2,0,0,75,3,0,0,76,3,0,0,3,0,0,0,4,0,0,0,77,3,0,0,78,3,0,0,79,3,0,0,80,3,0,0,81,3,0,0,82,3,0,0,83,3,0,0,248,255,255,255,80,2,0,0,84,3,0,0,85,3,0,0,86,3,0,0,87,3,0,0,88,3,0,0,89,3,0,0,244,255,255,255,80,2,0,0,90,3,0,0,91,3,0,0,92,3,0,0,240,255,255,255,80,2,0,0,93,3,0,0,94,3,0,0,95,3,0,0,96,3,0,0,155,114,5,0,97,3,0,0,160,114,5,0,98,3,0,0,165,114,5,0,99,3,0,0,174,114,5,0,100,3,0,0,186,114,5,0,101,3,0,0,204,114,5,0,102,3,0,0,219,114,5,0,103,3,0,0,234,114,5,0,104,3,0,0,253,114,5,0,105,3,0,0,5,115,5,0,106,3,0,0,228,119,5,0,107,3,0,0,16,115,5,0,108,3,0,0,34,115,5,0,109,3,0,0,51,115,5,0,110,3,0,0,64,120,5,0,111,3,0,0,67,115,5,0,112,3,0,0,86,115,5,0,113,3,0,0,106,115,5,0,114,3,0,0,118,115,5,0,115,3,0,0,129,115,5,0,116,3,0,0,134,115,5,0,117,3,0,0,139,115,5,0,118,3,0,0,157,115,5,0,119,3,0,0,166,115,5,0,120,3,0,0,185,115,5,0,121,3,0,0,197,115,5,0,122,3,0,0,211,115,5,0,123,3,0,0,223,115,5,0,124,3,0,0,231,115,5,0,125,3,0,0,242,115,5,0,126,3,0,0,4,116,5,0,127,3,0,0,21,116,5,0,128,3,0,0,37,116,5,0,129,3,0,0,45,116,5,0,130,3,0,0,59,116,5,0,131,3,0,0,74,116,5,0,132,3,0,0,89,116,5,0,133,3,0,0,103,116,5,0,134,3,0,0,118,116,5,0,135,3,0,0,150,116,5,0,136,3,0,0,182,116,5,0,137,3,0,0,191,116,5,0,138,3,0,0,201,116,5,0,139,3,0,0,210,116,5,0,140,3,0,0,218,116,5,0,141,3,0,0,237,116,5,0,142,3,0,0,255,116,5,0,143,3,0,0,9,117,5,0,144,3,0,0,18,117,5,0,145,3,0,0,35,117,5,0,146,3,0,0,51,117,5,0,147,3,0,0,60,117,5,0,148,3,0,0,75,117,5,0,149,3,0,0,88,117,5,0,150,3,0,0,172,120,5,0,151,3,0,0,180,120,5,0,152,3,0,0,148,120,5,0,153,3,0,0,160,120,5,0,154,3,0,0,0,0,0,0,0,0,0,0,137,117,5,0,155,3,0,0,151,117,5,0,156,3,0,0,169,117,5,0,157,3,0,0,182,117,5,0,158,3,0,0,229,117,5,0,159,3,0,0,219,118,5,0,160,3,0,0,0,0,0,0,0,0,0,0,209,117,5,0,161,3,0,0,219,117,5,0,162,3,0,0,229,117,5,0,163,3,0,0,238,117,5,0,164,3,0,0,0,0,0,0,0,0,0,0,33,118,5,0,165,3,0,0,46,118,5,0,166,3,0,0,63,119,5,0,167,3,0,0,75,119,5,0,168,3,0,0,56,118,5,0,169,3,0,0,66,118,5,0,170,3,0,0,15,119,5,0,171,3,0,0,27,119,5,0,172,3,0,0,77,118,5,0,173,3,0,0,88,118,5,0,174,3,0,0,102,118,5,0,175,3,0,0,119,118,5,0,176,3,0,0,135,118,5,0,177,3,0,0,151,118,5,0,178,3,0,0,163,118,5,0,179,3,0,0,180,120,5,0,180,3,0,0,0,0,0,0,0,0,0,0,189,118,5,0,181,3,0,0,199,118,5,0,182,3,0,0,172,121,5,0,183,3,0,0,185,121,5,0,184,3,0,0,198,121,5,0,185,3,0,0,214,121,5,0,186,3,0,0,0,0,0,0,0,0,0,0,219,118,5,0,187,3,0,0,0,0,0,0,0,0,0,0,64,120,5,0,188,3,0,0,15,119,5,0,189,3,0,0,27,119,5,0,190,3,0,0,42,119,5,0,191,3,0,0,53,119,5,0,192,3,0,0,63,119,5,0,193,3,0,0,75,119,5,0,194,3,0,0,90,119,5,0,195,3,0,0,101,119,5,0,196,3,0,0,109,119,5,0,197,3,0,0,118,119,5,0,198,3,0,0,127,119,5,0,199,3,0,0,137,119,5,0,200,3,0,0,145,119,5,0,201,3,0,0,159,119,5,0,202,3,0,0,173,119,5,0,203,3,0,0,185,119,5,0,204,3,0,0,197,119,5,0,205,3,0,0,205,119,5,0,206,3,0,0,148,120,5,0,207,3,0,0,160,120,5,0,208,3,0,0,213,119,5,0,209,3,0,0,228,119,5,0,210,3,0,0,240,119,5,0,211,3,0,0,254,119,5,0,212,3,0,0,172,120,5,0,213,3,0,0,180,120,5,0,214,3,0,0,0,0,0,0,0,0,0,0,148,121,5,0,215,3,0,0,160,121,5,0,216,3,0,0,19,121,5,0,217,3,0,0,32,121,5,0,218,3,0,0,0,0,0,0,0,0,0,0,36,120,5,0,219,3,0,0,45,120,5,0,220,3,0,0,54,120,5,0,221,3,0,0,0,0,0,0,0,0,0,0,64,120,5,0,222,3,0,0,72,120,5,0,223,3,0,0,82,120,5,0,224,3,0,0,93,120,5,0,225,3,0,0,110,120,5,0,226,3,0,0,128,120,5,0,227,3,0,0,148,120,5,0,228,3,0,0,160,120,5,0,229,3,0,0,172,120,5,0,230,3,0,0,180,120,5,0,231,3,0,0,0,0,0,0,0,0,0,0,209,120,5,0,232,3,0,0,225,120,5,0,233,3,0,0,241,120,5,0,234,3,0,0,2,121,5,0,235,3,0,0,148,121,5,0,236,3,0,0,160,121,5,0,237,3,0,0,19,121,5,0,238,3,0,0,32,121,5,0,239,3,0,0,45,121,5,0,240,3,0,0,65,121,5,0,241,3,0,0,0,0,0,0,0,0,0,0,128,121,5,0,242,3,0,0,138,121,5,0,243,3,0,0,148,121,5,0,244,3,0,0,160,121,5,0,245,3,0,0,172,121,5,0,246,3,0,0,185,121,5,0,247,3,0,0,198,121,5,0,248,3,0,0,214,121,5,0,249,3,0,0,0,0,0,0,0,0,0,0,230,121,5,0,1,0,0,0,239,121,5,0,2,0,0,0,248,121,5,0,3,0,0,0,2,122,5,0,4,0,0,0,8,122,5,0,5,0,0,0,15,122,5,0,6,0,0,0,20,122,5,0,7,0,0,0,29,122,5,0,8,0,0,0,34,122,5,0,9,0,0,0,40,122,5,0,10,0,0,0,45,122,5,0,11,0,0,0,0,0,0,0,160,2,0,0,250,3,0,0,251,3,0,0,3,0,0,0,4,0,0,0,252,3,0,0,253,3,0,0,254,3,0,0,0,0,0,0,176,2,0,0,255,3,0,0,0,4,0,0,3,0,0,0,4,0,0,0,1,4,0,0,2,4,0,0,3,4,0,0,4,4,0,0,5,4,0,0,6,4,0,0,7,4,0,0,8,4,0,0,9,4,0,0,10,4,0,0,11,4,0,0,12,4,0,0,13,4,0,0,14,4,0,0,15,4,0,0,0,0,0,0,192,2,0,0,16,4,0,0,17,4,0,0,3,0,0,0,4,0,0,0,18,4,0,0,19,4,0,0,3,4,0,0,4,4,0,0,20,4,0,0,21,4,0,0,22,4,0,0,8,4,0,0,23,4,0,0,24,4,0,0,25,4,0,0,26,4,0,0,0,0,0,0,224,2,0,0,27,4,0,0,28,4,0,0,3,0,0,0,4,0,0,0,29,4,0,0,30,4,0,0,31,4,0,0,32,4,0,0,33,4,0,0,34,4,0,0,35,4,0,0,99,124,5,0,36,4,0,0,111,124,5,0,37,4,0,0,123,124,5,0,38,4,0,0,137,124,5,0,39,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,240,2,0,0,1,0,0,0,40,4,0,0,3,0,0,0,4,0,0,0,41,4,0,0,42,4,0,0,43,4,0,0,44,4,0,0,45,4,0,0,46,4,0,0,47,4,0,0,48,4,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,2,0,0,0,2,0,0,0,3,0,0,0,3,0,0,0,4,0,0,0,4,0,0,0,0,0,0,0,0,3,0,0,1,0,0,0,49,4,0,0,3,0,0,0,4,0,0,0,41,4,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,47,4,0,0,48,4,0,0,56,192,6,0,0,0,0,0,213,124,5,0,1,0,0,0,221,124,5,0,2,0,0,0,231,124,5,0,3,0,0,0,240,124,5,0,4,0,0,0,0,0,0,0,16,3,0,0,1,0,0,0,50,4,0,0,3,0,0,0,4,0,0,0,51,4,0,0,52,4,0,0,53,4,0,0,54,4,0,0,55,4,0,0,0,0,0,0,32,3,0,0,56,4,0,0,57,4,0,0,5,0,0,0,76,125,5,0,58,4,0,0,81,125,5,0,59,4,0,0,88,125,5,0,60,4,0,0,92,125,5,0,61,4,0,0,99,125,5,0,62,4,0,0,104,125,5,0,63,4,0,0,8,194,6,0,64,4,0,0,113,125,5,0,65,4,0,0,0,0,0,0,0,0,0,0,52,82,9,0,66,4,0,0,231,58,9,0,67,4,0,0,193,125,5,0,68,4,0,0,202,125,5,0,69,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,40,3,0,0,70,4,0,0,71,4,0,0,3,0,0,0,4,0,0,0,72,4,0,0,73,4,0,0,74,4,0,0,75,4,0,0,76,4,0,0,77,4,0,0,78,4,0,0,79,4,0,0,0,0,0,0,56,3,0,0,1,0,0,0,80,4,0,0,3,0,0,0,4,0,0,0,72,4,0,0,5,0,0,0,74,4,0,0,5,0,0,0,76,4,0,0,77,4,0,0,78,4,0,0,79,4,0,0,0,0,0,0,72,3,0,0,81,4,0,0,82,4,0,0,3,0,0,0,4,0,0,0,83,4,0,0,84,4,0,0,85,4,0,0,86,4,0,0,0,0,0,0,104,3,0,0,87,4,0,0,88,4,0,0,3,0,0,0,4,0,0,0,89,4,0,0,90,4,0,0,91,4,0,0,0,0,0,0,136,3,0,0,92,4,0,0,93,4,0,0,94,4,0,0,0,0,0,0,152,3,0,0,95,4,0,0,96,4,0,0,3,0,0,0,4,0,0,0,97,4,0,0,98,4,0,0,99,4,0,0,100,4,0,0,101,4,0,0,102,4,0,0,103,4,0,0,104,4,0,0,105,4,0,0,106,4,0,0,107,4,0,0,108,4,0,0,109,4,0,0,110,4,0,0,0,0,0,0,168,3,0,0,111,4,0,0,112,4,0,0,3,0,0,0,4,0,0,0,113,4,0,0,114,4,0,0,115,4,0,0,116,4,0,0,117,4,0,0,118,4,0,0,119,4,0,0,0,0,0,0,200,3,0,0,120,4,0,0,121,4,0,0,3,0,0,0,4,0,0,0,122,4,0,0,123,4,0,0,124,4,0,0,125,4,0,0,126,4,0,0,118,4,0,0,127,4,0,0,0,0,0,0,216,3,0,0,128,4,0,0,129,4,0,0,3,0,0,0,4,0,0,0,130,4,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,104,4,0,0,105,4,0,0,106,4,0,0,107,4,0,0,131,4,0,0,132,4,0,0,110,4,0,0,0,0,0,0,248,3,0,0,133,4,0,0,134,4,0,0,3,0,0,0,4,0,0,0,135,4,0,0,136,4,0,0,137,4,0,0,138,4,0,0,139,4,0,0,140,4,0,0,141,4,0,0,142,4,0,0,143,4,0,0,144,4,0,0,145,4,0,0,146,4,0,0,147,4,0,0,148,4,0,0,149,4,0,0,150,4,0,0,151,4,0,0,152,4,0,0,153,4,0,0,154,4,0,0,155,4,0,0,156,4,0,0,157,4,0,0,158,4,0,0,159,4,0,0,160,4,0,0,161,4,0,0,162,4,0,0,163,4,0,0,164,4,0,0,165,4,0,0,166,4,0,0,167,4,0,0,168,4,0,0,169,4,0,0,170,4,0,0,171,4,0,0,172,4,0,0,173,4,0,0,174,4,0,0,175,4,0,0,176,4,0,0,177,4,0,0,66,129,5,0,0,0,0,0,77,129,5,0,1,0,0,0,92,129,5,0,2,0,0,0,98,129,5,0,3,0,0,0,103,129,5,0,4,0,0,0,113,129,5,0,5,0,0,0,122,129,5,0,6,0,0,0,132,129,5,0,7,0,0,0,143,129,5,0,8,0,0,0,152,129,5,0,9,0,0,0,160,129,5,0,10,0,0,0,168,129,5,0,11,0,0,0,37,82,9,0,12,0,0,0,39,82,9,0,13,0,0,0,180,129,5,0,0,0,0,0,190,129,5,0,1,0,0,0,48,240,7,0,0,0,0,0,198,129,5,0,1,0,0,0,251,154,5,0,2,0,0,0,0,0,0,0,24,4,0,0,178,4,0,0,179,4,0,0,180,4,0,0,181,4,0,0,182,4,0,0,183,4,0,0,184,4,0,0,185,4,0,0,0,0,0,0,48,4,0,0,186,4,0,0,187,4,0,0,188,4,0,0,189,4,0,0,190,4,0,0,191,4,0,0,192,4,0,0,193,4,0,0,0,0,0,0,64,4,0,0,186,4,0,0,194,4,0,0,195,4,0,0,196,4,0,0,197,4,0,0,198,4,0,0,199,4,0,0,200,4,0,0,0,0,0,0,80,4,0,0,186,4,0,0,201,4,0,0,202,4,0,0,203,4,0,0,204,4,0,0,205,4,0,0,206,4,0,0,207,4,0,0,16,0,0,0,32,0,0,0,64,0,0,0,96,0,0,0,128,0,0,0,160,0,0,0,192,0,0,0,224,0,0,0,0,1,0,0,64,1,0,0,128,1,0,0,192,1,0,0,0,2,0,0,128,2,0,0,0,0,0,0,96,4,0,0,208,4,0,0,209,4,0,0,210,4,0,0,0,0,0,0,104,4,0,0,5,0,0,0,211,4,0,0,212,4,0,0,0,0,0,0,112,4,0,0,213,4,0,0,211,4,0,0,214,4,0,0,0,0,0,0,128,4,0,0,215,4,0,0,211,4,0,0,216,4,0,0,0,0,0,0,144,4,0,0,217,4,0,0,211,4,0,0,218,4,0,0,0,0,0,0,160,4,0,0,219,4,0,0,211,4,0,0,220,4,0,0,0,0,0,0,176,4,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,221,4,0,0,222,4,0,0,223,4,0,0,224,4,0,0,5,0,0,0,5,0,0,0,5,0,0,0,0,0,0,0,184,4,0,0,225,4,0,0,226,4,0,0,227,4,0,0,228,4,0,0,229,4,0,0,222,4,0,0,223,4,0,0,230,4,0,0,231,4,0,0,232,4,0,0,233,4,0,0,0,0,0,0,200,4,0,0,234,4,0,0,235,4,0,0,236,4,0,0,237,4,0,0,238,4,0,0,239,4,0,0,223,4,0,0,240,4,0,0,241,4,0,0,242,4,0,0,243,4,0,0,0,0,0,0,216,4,0,0,244,4,0,0,245,4,0,0,246,4,0,0,247,4,0,0,248,4,0,0,222,4,0,0,223,4,0,0,249,4,0,0,250,4,0,0,251,4,0,0,252,4,0,0,0,0,0,0,232,4,0,0,253,4,0,0,254,4,0,0,255,4,0,0,0,5,0,0,1,5,0,0,2,5,0,0,223,4,0,0,3,5,0,0,4,5,0,0,5,5,0,0,6,5,0,0,0,0,0,0,248,4,0,0,7,5,0,0,8,5,0,0,9,5,0,0,10,5,0,0,11,5,0,0,222,4,0,0,223,4,0,0,12,5,0,0,13,5,0,0,14,5,0,0,15,5,0,0,0,0,0,0,8,5,0,0,16,5,0,0,17,5,0,0,18,5,0,0,19,5,0,0,20,5,0,0,222,4,0,0,223,4,0,0,21,5,0,0,22,5,0,0,23,5,0,0,24,5,0,0,0,0,0,0,24,5,0,0,25,5,0,0,26,5,0,0,27,5,0,0,28,5,0,0,29,5,0,0,222,4,0,0,223,4,0,0,30,5,0,0,31,5,0,0,32,5,0,0,33,5,0,0,0,0,0,0,40,5,0,0,34,5,0,0,35,5,0,0,36,5,0,0,37,5,0,0,38,5,0,0,222,4,0,0,223,4,0,0,39,5,0,0,40,5,0,0,41,5,0,0,42,5,0,0,153,140,5,0,43,5,0,0,158,140,5,0,44,5,0,0,0,0,0,0,0,0,0,0,212,140,5,0,45,5,0,0,221,140,5,0,46,5,0,0,186,187,7,0,47,5,0,0,228,140,5,0,48,5,0,0,235,140,5,0,49,5,0,0,30,142,5,0,50,5,0,0,109,143,5,0,51,5,0,0,96,241,7,0,52,5,0,0,0,0,0,0,0,0,0,0,245,140,5,0,53,5,0,0,249,140,5,0,54,5,0,0,66,141,5,0,55,5,0,0,73,141,5,0,56,5,0,0,97,141,5,0,57,5,0,0,103,141,5,0,58,5,0,0,109,141,5,0,59,5,0,0,114,141,5,0,60,5,0,0,0,0,0,0,0,0,0,0,96,241,7,0,61,5,0,0,0,0,0,0,0,0,0,0,98,251,7,0,62,5,0,0,194,245,7,0,63,5,0,0,150,141,5,0,64,5,0,0,181,141,5,0,65,5,0,0,11,245,7,0,62,5,0,0,186,141,5,0,66,5,0,0,87,142,5,0,67,5,0,0,93,142,5,0,68,5,0,0,99,142,5,0,69,5,0,0,111,142,5,0,70,5,0,0,194,141,5,0,71,5,0,0,203,141,5,0,72,5,0,0,212,141,5,0,73,5,0,0,123,142,5,0,74,5,0,0,168,142,5,0,75,5,0,0,199,142,5,0,76,5,0,0,205,142,5,0,77,5,0,0,61,143,5,0,66,5,0,0,73,143,5,0,65,5,0,0,85,143,5,0,78,5,0,0,248,141,5,0,79,5,0,0,0,0,0,0,0,0,0,0,219,141,5,0,80,5,0,0,235,142,5,0,81,5,0,0,229,141,5,0,82,5,0,0,241,141,5,0,83,5,0,0,0,0,0,0,0,0,0,0,30,142,5,0,84,5,0,0,0,0,0,0,0,0,0,0,34,142,5,0,85,5,0,0,161,195,6,0,86,5,0,0,0,0,0,0,0,0,0,0,98,251,7,0,87,5,0,0,194,245,7,0,63,5,0,0,11,245,7,0,87,5,0,0,87,142,5,0,88,5,0,0,93,142,5,0,89,5,0,0,99,142,5,0,90,5,0,0,111,142,5,0,91,5,0,0,123,142,5,0,92,5,0,0,139,142,5,0,93,5,0,0,168,142,5,0,94,5,0,0,173,142,5,0,95,5,0,0,199,142,5,0,96,5,0,0,205,142,5,0,97,5,0,0,61,143,5,0,98,5,0,0,73,143,5,0,99,5,0,0,85,143,5,0,100,5,0,0,0,0,0,0,0,0,0,0,215,142,5,0,101,5,0,0,225,142,5,0,102,5,0,0,235,142,5,0,81,5,0,0,245,142,5,0,103,5,0,0,6,143,5,0,104,5,0,0,24,143,5,0,105,5,0,0,42,143,5,0,106,5,0,0,0,0,0,0,0,0,0,0,109,143,5,0,107,5,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,255,255,255,255,0,0,0,0,1,0,0,0,2,0,0,0,3,0,0,0,4,0,0,0,1,0,0,0,2,0,0,0,1,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,1,0,0,0,0,0,0,0,56,5,0,0,108,5,0,0,109,5,0,0,3,0,0,0,4,0,0,0,110,5,0,0,111,5,0,0,3,4,0,0,4,4,0,0,112,5,0,0,113,5,0,0,114,5,0,0,8,4,0,0,115,5,0,0,116,5,0,0,117,5,0,0,118,5,0,0,0,0,0,0,72,5,0,0,119,5,0,0,120,5,0,0,0,0,0,0,80,5,0,0,121,5,0,0,122,5,0,0,3,0,0,0,4,0,0,0,0,0,0,0,96,5,0,0,123,5,0,0,124,5,0,0,3,0,0,0,4,0,0,0,125,5,0,0,126,5,0,0,127,5,0,0,128,5,0,0,129,5,0,0,130,5,0,0,131,5,0,0,132,5,0,0,133,5,0,0,134,5,0,0,135,5,0,0,136,5,0,0,137,5,0,0,138,5,0,0,139,5,0,0,140,5,0,0,141,5,0,0,142,5,0,0,143,5,0,0,144,5,0,0,145,5,0,0,146,5,0,0,147,5,0,0,148,5,0,0,149,5,0,0,150,5,0,0,151,5,0,0,152,5,0,0,153,5,0,0,154,5,0,0,155,5,0,0,156,5,0,0,0,0,0,0,128,5,0,0,157,5,0,0,158,5,0,0,3,0,0,0,4,0,0,0,159,5,0,0,160,5,0,0,161,5,0,0,162,5,0,0,163,5,0,0,164,5,0,0,165,5,0,0,166,5,0,0,167,5,0,0,168,5,0,0,169,5,0,0,170,5,0,0,171,5,0,0,172,5,0,0,173,5,0,0,174,5,0,0,175,5,0,0,176,5,0,0,177,5,0,0,178,5,0,0,179,5,0,0,180,5,0,0,181,5,0,0,182,5,0,0,183,5,0,0,184,5,0,0,185,5,0,0,186,5,0,0,187,5,0,0,188,5,0,0,189,5,0,0,190,5,0,0,191,5,0,0,192,5,0,0,193,5,0,0,194,5,0,0,195,5,0,0,196,5,0,0,197,5,0,0,198,5,0,0,199,5,0,0,0,0,0,0,144,5,0,0,200,5,0,0,201,5,0,0,202,5,0,0,0,0,0,0,160,5,0,0,203,5,0,0,204,5,0,0,3,0,0,0,4,0,0,0,125,5,0,0,205,5,0,0,206,5,0,0,207,5,0,0,208,5,0,0,209,5,0,0,210,5,0,0,211,5,0,0,212,5,0,0,213,5,0,0,214,5,0,0,215,5,0,0,216,5,0,0,217,5,0,0,218,5,0,0,219,5,0,0,220,5,0,0,221,5,0,0,222,5,0,0,223,5,0,0,224,5,0,0,225,5,0,0,226,5,0,0,227,5,0,0,228,5,0,0,229,5,0,0,230,5,0,0,231,5,0,0,232,5,0,0,233,5,0,0,234,5,0,0,235,5,0,0,0,0,0,0,176,5,0,0,236,5,0,0,237,5,0,0,3,0,0,0,4,0,0,0,0,0,0,0,192,5,0,0,238,5,0,0,239,5,0,0,3,0,0,0,4,0,0,0,240,5,0,0,241,5,0,0,242,5,0,0,243,5,0,0,244,5,0,0,245,5,0,0,246,5,0,0,247,5,0,0,248,5,0,0,249,5,0,0,250,5,0,0,251,5,0,0,252,5,0,0,253,5,0,0,254,5,0,0,255,5,0,0,0,6,0,0,1,6,0,0,2,6,0,0,3,6,0,0,4,6,0,0,5,6,0,0,6,6,0,0,7,6,0,0,8,6,0,0,9,6,0,0,10,6,0,0,11,6,0,0,12,6,0,0,13,6,0,0,14,6,0,0,15,6,0,0,16,6,0,0,17,6,0,0,18,6,0,0,19,6,0,0,20,6,0,0,21,6,0,0,22,6,0,0,23,6,0,0,199,5,0,0,24,6,0,0,25,6,0,0,26,6,0,0,0,0,0,0,208,5,0,0,88,1,0,0,27,6,0,0,90,1,0,0,0,0,0,0,224,5,0,0,88,1,0,0,28,6,0,0,90,1,0,0,83,3,8,0,0,0,0,0,67,147,5,0,1,0,0,0,75,147,5,0,2,0,0,0,110,168,6,0,3,0,0,0,90,147,5,0,4,0,0,0,104,147,5,0,5,0,0,0,113,147,5,0,6,0,0,0,0,0,0,0,240,5,0,0,29,6,0,0,30,6,0,0,3,0,0,0,4,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,199,5,0,0,87,170,6,0,0,0,0,0,72,170,6,0,1,0,0,0,150,147,5,0,0,0,0,0,158,147,5,0,1,0,0,0,0,0,0,0,0,6,0,0,31,6,0,0,32,6,0,0,3,0,0,0,4,0,0,0,33,6,0,0,34,6,0,0,35,6,0,0,36,6,0,0,37,6,0,0,0,0,0,0,16,6,0,0,38,6,0,0,39,6,0,0,3,0,0,0,4,0,0,0,0,0,0,0,32,6,0,0,40,6,0,0,41,6,0,0,3,0,0,0,4,0,0,0,33,6,0,0,5,0,0,0,42,6,0,0,5,0,0,0,5,0,0,0,0,0,0,0,48,6,0,0,43,6,0,0,44,6,0,0,3,0,0,0,4,0,0,0,45,6,0,0,46,6,0,0,47,6,0,0,48,6,0,0,49,6,0,0,50,6,0,0,51,6,0,0,52,6,0,0,53,6,0,0,54,6,0,0,55,6,0,0,56,6,0,0,57,6,0,0,58,6,0,0,59,6,0,0,60,6,0,0,61,6,0,0,62,6,0,0,63,6,0,0,64,6,0,0,65,6,0,0,66,6,0,0,67,6,0,0,68,6,0,0,69,6,0,0,70,6,0,0,71,6,0,0,72,6,0,0,73,6,0,0,74,6,0,0,75,6,0,0,76,6,0,0,77,6,0,0,78,6,0,0,79,6,0,0,80,6,0,0,81,6,0,0,82,6,0,0,83,6,0,0,84,6,0,0,85,6,0,0,0,0,0,0,64,6,0,0,86,6,0,0,87,6,0,0,3,0,0,0,4,0,0,0,88,6,0,0,89,6,0,0,90,6,0,0,91,6,0,0,92,6,0,0,93,6,0,0,94,6,0,0,95,6,0,0,96,6,0,0,97,6,0,0,98,6,0,0,99,6,0,0,100,6,0,0,101,6,0,0,102,6,0,0,103,6,0,0,104,6,0,0,139,192,6,0,0,0,0,0,39,11,8,0,1,0,0,0,5,88,9,0,2,0,0,0,33,82,9,0,3,0,0,0,83,3,8,0,0,0,0,0,177,245,7,0,1,0,0,0,172,245,7,0,2,0,0,0,0,0,0,0,96,6,0,0,105,6,0,0,106,6,0,0,3,0,0,0,4,0,0,0,45,6,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,51,6,0,0,52,6,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,84,6,0,0,85,6,0,0,10,151,5,0,107,6,0,0,16,246,7,0,108,6,0,0,11,245,7,0,109,6,0,0,97,151,5,0,110,6,0,0,89,245,7,0,111,6,0,0,182,245,7,0,112,6,0,0,54,245,7,0,113,6,0,0,128,151,5,0,114,6,0,0,134,151,5,0,115,6,0,0,147,245,7,0,116,6,0,0,60,245,7,0,117,6,0,0,2,152,5,0,118,6,0,0,43,152,5,0,119,6,0,0,79,152,5,0,120,6,0,0,120,152,5,0,121,6,0,0,132,152,5,0,122,6,0,0,0,0,0,0,0,0,0,0,120,152,5,0,123,6,0,0,132,152,5,0,124,6,0,0,0,0,0,0,0,0,0,0,0,0,0,0,112,6,0,0,125,6,0,0,126,6,0,0,3,0,0,0,4,0,0,0,127,6,0,0,128,6,0,0,129,6,0,0,130,6,0,0,131,6,0,0,132,6,0,0,133,6,0,0,134,6,0,0,135,6,0,0,136,6,0,0,0,0,0,0,128,6,0,0,137,6,0,0,138,6,0,0,3,0,0,0,4,0,0,0,139,6,0,0,140,6,0,0,141,6,0,0,142,6,0,0,143,6,0,0,144,6,0,0,145,6,0,0,146,6,0,0,147,6,0,0,148,6,0,0,149,6,0,0,0,0,0,0,160,6,0,0,150,6,0,0,151,6,0,0,3,0,0,0,4,0,0,0,139,6,0,0,140,6,0,0,141,6,0,0,142,6,0,0,152,6,0,0,153,6,0,0,145,6,0,0,154,6,0,0,155,6,0,0,148,6,0,0,156,6,0,0,0,0,0,0,192,6,0,0,7,3,0,0,157,6,0,0,3,0,0,0,4,0,0,0,158,6,0,0,159,6,0,0,0,0,0,0,224,6,0,0,160,6,0,0,161,6,0,0,162,6,0,0,0,0,0,0,240,6,0,0,160,6,0,0,163,6,0,0,164,6,0,0,0,0,0,0,0,7,0,0,160,6,0,0,165,6,0,0,166,6,0,0,0,0,0,0,16,7,0,0,167,6,0,0,168,6,0,0,3,0,0,0,4,0,0,0,169,6,0,0,170,6,0,0,171,6,0,0,172,6,0,0,9,159,6,0,0,0,0,0,24,159,6,0,1,0,0,0,0,0,0,0,32,7,0,0,173,6,0,0,174,6,0,0,3,0,0,0,4,0,0,0,139,6,0,0,140,6,0,0,141,6,0,0,142,6,0,0,175,6,0,0,176,6,0,0,145,6,0,0,177,6,0,0,178,6,0,0,148,6,0,0,179,6,0,0,168,242,7,0,0,0,0,0,84,159,6,0,1,0,0,0,90,159,6,0,2,0,0,0,83,3,8,0,3,0,0,0,0,0,0,0,64,7,0,0,180,6,0,0,181,6,0,0,3,0,0,0,4,0,0,0,182,6,0,0,183,6,0,0,191,2,0,0,192,2,0,0,184,6,0,0,193,2,0,0,185,6,0,0,194,2,0,0,195,2,0,0,186,6,0,0,187,6,0,0,188,6,0,0,136,255,255,255,64,7,0,0,189,6,0,0,190,6,0,0,191,6,0,0,192,6,0,0,168,242,7,0,0,0,0,0,158,159,6,0,1,0,0,0,162,159,6,0,2,0,0,0,168,159,6,0,3,0,0,0,175,159,6,0,4,0,0,0,182,159,6,0,5,0,0,0,185,159,6,0,6,0,0,0,88,164,6,0,7,0,0,0,189,159,6,0,8,0,0,0,197,159,6,0,9,0,0,0,206,159,6,0,10,0,0,0,211,159,6,0,11,0,0,0,217,159,6,0,12,0,0,0,225,159,6,0,13,0,0,0,230,159,6,0,14,0,0,0,236,159,6,0,15,0,0,0,244,159,6,0,16,0,0,0,0,0,0,0,96,7,0,0,193,6,0,0,194,6,0,0,3,0,0,0,4,0,0,0,195,6,0,0,196,6,0,0,248,255,255,255,96,7,0,0,197,6,0,0,198,6,0,0,199,6,0,0,200,6,0,0,22,193,6,0,0,0,0,0,16,193,6,0,2,0,0,0,70,161,6,0,1,0,0,0,77,161,6,0,3,0,0,0,80,96,0,0,120,96,0,0,140,96,0,0,100,96,0,0,60,0,0,0,0,0,0,0,128,7,0,0,201,6,0,0,202,6,0,0,196,255,255,255,196,255,255,255,128,7,0,0,203,6,0,0,204,6,0,0,60,0,0,0,0,0,0,0,240,29,0,0,32,0,0,0,33,0,0,0,196,255,255,255,196,255,255,255,240,29,0,0,34,0,0,0,35,0,0,0,0,0,0,0,144,7,0,0,205,6,0,0,206,6,0,0,207,6,0,0,208,6,0,0,3,20,0,0,5,20,0,0,0,0,0,0,160,7,0,0,209,6,0,0,210,6,0,0,3,0,0,0,4,0,0,0,211,6,0,0,212,6,0,0,213,6,0,0,214,6,0,0,215,6,0,0,216,6,0,0,217,6,0,0,1,0,0,0,0,0,0,0,176,7,0,0,218,6,0,0,219,6,0,0,3,0,0,0,4,0,0,0,220,6,0,0,221,6,0,0,191,2,0,0,192,2,0,0,222,6,0,0,193,2,0,0,223,6,0,0,194,2,0,0,195,2,0,0,224,6,0,0,225,6,0,0,226,6,0,0,136,255,255,255,176,7,0,0,227,6,0,0,228,6,0,0,229,6,0,0,230,6,0,0,102,168,6,0,0,0,0,0,110,168,6,0,1,0,0,0,0,0,0,0,208,7,0,0,231,6,0,0,232,6,0,0,3,0,0,0,4,0,0,0,233,6,0,0,72,170,6,0,0,0,0,0,79,170,6,0,1,0,0,0,87,170,6,0,2,0,0,0,94,170,6,0,0,0,0,0,98,170,6,0,1,0,0,0,104,170,6,0,2,0,0,0,114,25,9,0,3,0,0,0,0,249,7,0,0,0,0,0,200,175,6,0,1,0,0,0,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,0,0,224,7,0,0,234,6,0,0,235,6,0,0,3,0,0,0,4,0,0,0,236,6,0,0,237,6,0,0,238,6,0,0,0,0,0,0,240,7,0,0,239,6,0,0,240,6,0,0,241,6,0,0,242,6,0,0,243,6,0,0,5,0,0,0,0,0,0,0,248,7,0,0,239,6,0,0,244,6,0,0,245,6,0,0,246,6,0,0,247,6,0,0,248,6,0,0,0,0,0,0,8,8,0,0,239,6,0,0,249,6,0,0,241,6,0,0,242,6,0,0,243,6,0,0,250,6,0,0,0,0,0,0,24,8,0,0,239,6,0,0,251,6,0,0,241,6,0,0,242,6,0,0,243,6,0,0,252,6,0,0,0,0,0,0,40,8,0,0,253,6,0,0,254,6,0,0,3,0,0,0,4,0,0,0,255,6,0,0,0,7,0,0,248,255,255,255,40,8,0,0,1,7,0,0,2,7,0,0,3,7,0,0,4,7,0,0,187,175,6,0,0,0,0,0,194,175,6,0,1,0,0,0,200,175,6,0,0,0,0,0,206,175,6,0,1,0,0,0,210,175,6,0,2,0,0,0,215,175,6,0,3,0,0,0,56,192,6,0,4,0,0,0,221,175,6,0,0,0,0,0,236,175,6,0,1,0,0,0,251,175,6,0,2,0,0,0,7,176,6,0,3,0,0,0,21,176,6,0,0,0,0,0,37,176,6,0,1,0,0,0,54,176,6,0,2,0,0,0,80,176,6,0,3,0,0,0,93,176,6,0,4,0,0,0,108,176,6,0,5,0,0,0,124,176,6,0,6,0,0,0,143,176,6,0,7,0,0,0,163,176,6,0,8,0,0,0,0,0,0,0,72,8,0,0,5,7,0,0,6,7,0,0,3,0,0,0,4,0,0,0,7,7,0,0,0,0,0,0,88,8,0,0,8,7,0,0,9,7,0,0,3,0,0,0,4,0,0,0,10,7,0,0,0,0,0,0,104,8,0,0,11,7,0,0,12,7,0,0,3,0,0,0,4,0,0,0,13,7,0,0,14,7,0,0,15,7,0,0,248,255,255,255,104,8,0,0,16,7,0,0,17,7,0,0,18,7,0,0,19,7,0,0,0,0,0,0,176,0,0,0,7,3,0,0,20,7,0,0,3,0,0,0,4,0,0,0,5,0,0,0,34,180,6,0,21,7,0,0,43,180,6,0,22,7,0,0,56,180,6,0,23,7,0,0,66,180,6,0,24,7,0,0,0,0,0,0,0,0,0,0,103,180,6,0,25,7,0,0,113,180,6,0,26,7,0,0,122,180,6,0,27,7,0,0,130,180,6,0,28,7,0,0,144,180,6,0,29,7,0,0,158,180,6,0,30,7,0,0,168,180,6,0,31,7,0,0,178,180,6,0,32,7,0,0,188,180,6,0,33,7,0,0,199,180,6,0,34,7,0,0,211,180,6,0,35,7,0,0,221,180,6,0,36,7,0,0,0,0,0,0,0,0,0,0,0,0,0,0,136,8,0,0,37,7,0,0,38,7,0,0,5,0,0,0,5,0,0,0,0,0,0,0,144,8,0,0,39,7,0,0,40,7,0,0,3,0,0,0,4,0,0,0,41,7,0,0,42,7,0,0,43,7,0,0,44,7,0,0,45,7,0,0,46,7,0,0,47,7,0,0,0,0,0,0,176,8,0,0,48,7,0,0,49,7,0,0,3,0,0,0,4,0,0,0,50,7,0,0,51,7,0,0,52,7,0,0,0,0,0,0,192,8,0,0,7,3,0,0,53,7,0,0,3,0,0,0,4,0,0,0,54,7,0,0,55,7,0,0,0,0,0,0,224,8,0,0,7,3,0,0,56,7,0,0,3,0,0,0,4,0,0,0,57,7,0,0,58,7,0,0,0,0,0,0,240,8,0,0,59,7,0,0,60,7,0,0,3,0,0,0,4,0,0,0,61,7,0,0,62,7,0,0,63,7,0,0,64,7,0,0,65,7,0,0,0,0,0,0,0,9,0,0,7,3,0,0,66,7,0,0,3,0,0,0,4,0,0,0,67,7,0,0,68,7,0,0,0,0,0,0,16,9,0,0,59,7,0,0,69,7,0,0,3,0,0,0,4,0,0,0,70,7,0,0,71,7,0,0,72,7,0,0,73,7,0,0,74,7,0,0,0,0,0,0,255,255,255,255,253,255,255,255,249,255,255,255,241,255,255,255,225,255,255,255,193,255,255,255,129,255,255,255,1,255,255,255,1,254,255,255,1,252,255,255,1,248,255,255,1,240,255,255,1,224,255,255,1,192,255,255,1,128,255,255,0,0,0,0,32,9,0,0,75,7,0,0,76,7,0,0,3,0,0,0,4,0,0,0,50,7,0,0,51,7,0,0,5,0,0,0,247,188,6,0,0,0,0,0,251,188,6,0,1,0,0,0,0,0,0,0,48,9,0,0,77,7,0,0,78,7,0,0,3,0,0,0,4,0,0,0,79,7,0,0,80,7,0,0,81,7,0,0,82,7,0,0,83,7,0,0,84,7,0,0,85,7,0,0,86,7,0,0,87,7,0,0,88,7,0,0,89,7,0,0,90,7,0,0,91,7,0,0,92,7,0,0,93,7,0,0,94,7,0,0,95,7,0,0,96,7,0,0,97,7,0,0,98,7,0,0,99,7,0,0,100,7,0,0,101,7,0,0,1,0,0,0,0,0,0,0,2,0,0,0,1,0,0,0,3,0,0,0,2,0,0,0,4,0,0,0,4,0,0,0,5,0,0,0,8,0,0,0,6,0,0,0,3,0,0,0,7,0,0,0,6,0,0,0,8,0,0,0,9,0,0,0,9,0,0,0,12,0,0,0,1,0,0,0,0,0,0,0,2,0,0,0,1,0,0,0,3,0,0,0,2,0,0,0,4,0,0,0,3,0,0,0,5,0,0,0,4,0,0,0,6,0,0,0,5,0,0,0,1,0,0,0,0,0,0,0,2,0,0,0,1,0,0,0,3,0,0,0,2,0,0,0,4,0,0,0,3,0,0,0,5,0,0,0,4,0,0,0,6,0,0,0,5,0,0,0,7,0,0,0,6,0,0,0,8,0,0,0,7,0,0,0,9,0,0,0,8,0,0,0,10,0,0,0,9,0,0,0,11,0,0,0,10,0,0,0,12,0,0,0,11,0,0,0,13,0,0,0,12,0,0,0,14,0,0,0,13,0,0,0,15,0,0,0,14,0,0,0,0,0,0,0,80,9,0,0,102,7,0,0,103,7,0,0,3,0,0,0,4,0,0,0,104,7,0,0,105,7,0,0,106,7,0,0,107,7,0,0,108,7,0,0,109,7,0,0,110,7,0,0,111,7,0,0,112,7,0,0,113,7,0,0,114,7,0,0,115,7,0,0,139,192,6,0,1,0,0,0,161,192,6,0,2,0,0,0,39,11,8,0,3,0,0,0,1,3,8,0,4,0,0,0,191,120,9,0,5,0,0,0,146,190,6,0,6,0,0,0,149,190,6,0,7,0,0,0,152,190,6,0,8,0,0,0,155,190,6,0,9,0,0,0,133,82,9,0,1,0,0,0,139,82,9,0,2,0,0,0,145,82,9,0,3,0,0,0,152,82,9,0,4,0,0,0,158,190,6,0,5,0,0,0,170,190,6,0,6,0,0,0,33,82,9,0,1,0,0,0,35,82,9,0,2,0,0,0,37,82,9,0,3,0,0,0,39,82,9,0,4,0,0,0,41,82,9,0,5,0,0,0,46,82,9,0,6,0,0,0,52,82,9,0,7,0,0,0,58,82,9,0,8,0,0,0,68,82,9,0,9,0,0,0,79,82,9,0,10,0,0,0,92,82,9,0,11,0,0,0,106,82,9,0,12,0,0,0,111,82,9,0,13,0,0,0,118,82,9,0,14,0,0,0,125,82,9,0,15,0,0,0,183,190,6,0,0,0,0,0,188,190,6,0,1,0,0,0,195,190,6,0,2,0,0,0,12,191,6,0,116,7,0,0,24,191,6,0,117,7,0,0,32,191,6,0,118,7,0,0,38,191,6,0,119,7,0,0,46,191,6,0,120,7,0,0,59,191,6,0,121,7,0,0,74,191,6,0,122,7,0,0,86,191,6,0,123,7,0,0,94,191,6,0,124,7,0,0,102,191,6,0,125,7,0,0,109,191,6,0,126,7,0,0,116,191,6,0,127,7,0,0,126,191,6,0,128,7,0,0,141,191,6,0,129,7,0,0,155,191,6,0,130,7,0,0,176,191,6,0,131,7,0,0,189,191,6,0,132,7,0,0,202,191,6,0,133,7,0,0,220,191,6,0,233,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,112,9,0,0,1,0,0,0,134,7,0,0,3,0,0,0,4,0,0,0,135,7,0,0,136,7,0,0,137,7,0,0,138,7,0,0,139,7,0,0,140,7,0,0,141,7,0,0,142,7,0,0,143,7,0,0,144,7,0,0,145,7,0,0,146,7,0,0,0,0,0,0,0,0,0,0,1,0,0,0,4,0,0,0,2,0,0,0,5,0,0,0,3,0,0,0,6,0,0,0,4,0,0,0,7,0,0,0,5,0,0,0,8,0,0,0,6,0,0,0,9,0,0,0,7,0,0,0,10,0,0,0,8,0,0,0,11,0,0,0,9,0,0,0,12,0,0,0,10,0,0,0,13,0,0,0,11,0,0,0,14,0,0,0,12,0,0,0,15,0,0,0,13,0,0,0,16,0,0,0,14,0,0,0,17,0,0,0,15,0,0,0,18,0,0,0,16,0,0,0,19,0,0,0,17,0,0,0,20,0,0,0,18,0,0,0,21,0,0,0,19,0,0,0,22,0,0,0,20,0,0,0,23,0,0,0,21,0,0,0,24,0,0,0,22,0,0,0,25,0,0,0,23,0,0,0,26,0,0,0,24,0,0,0,27,0,0,0,25,0,0,0,28,0,0,0,26,0,0,0,29,0,0,0,27,0,0,0,30,0,0,0,28,0,0,0,31,0,0,0,29,0,0,0,32,0,0,0,30,0,0,0,33,0,0,0,31,0,0,0,34,0,0,0,32,0,0,0,35,0,0,0,33,0,0,0,36,0,0,0,34,0,0,0,37,0,0,0,35,0,0,0,38,0,0,0,36,0,0,0,39,0,0,0,37,0,0,0,40,0,0,0,38,0,0,0,41,0,0,0,39,0,0,0,42,0,0,0,40,0,0,0,43,0,0,0,41,0,0,0,44,0,0,0,42,0,0,0,45,0,0,0,43,0,0,0,46,0,0,0,44,0,0,0,47,0,0,0,45,0,0,0,48,0,0,0,46,0,0,0,49,0,0,0,47,0,0,0,50,0,0,0,48,0,0,0,51,0,0,0,49,0,0,0,52,0,0,0,50,0,0,0,53,0,0,0,51,0,0,0,54,0,0,0,52,0,0,0,55,0,0,0,53,0,0,0,56,0,0,0,54,0,0,0,57,0,0,0,55,0,0,0,58,0,0,0,56,0,0,0,59,0,0,0,57,0,0,0,60,0,0,0,58,0,0,0,61,0,0,0,59,0,0,0,62,0,0,0,60,0,0,0,63,0,0,0,61,0,0,0,64,0,0,0,62,0,0,0,65,0,0,0,63,0,0,0,66,0,0,0,64,0,0,0,67,0,0,0,65,0,0,0,68,0,0,0,66,0,0,0,69,0,0,0,67,0,0,0,70,0,0,0,68,0,0,0,71,0,0,0,69,0,0,0,72,0,0,0,70,0,0,0,73,0,0,0,71,0,0,0,74,0,0,0,72,0,0,0,75,0,0,0,73,0,0,0,76,0,0,0,74,0,0,0,77,0,0,0,75,0,0,0,78,0,0,0,76,0,0,0,79,0,0,0,77,0,0,0,80,0,0,0,78,0,0,0,81,0,0,0,79,0,0,0,82,0,0,0,80,0,0,0,83,0,0,0,81,0,0,0,84,0,0,0,82,0,0,0,85,0,0,0,83,0,0,0,86,0,0,0,84,0,0,0,87,0,0,0,85,0,0,0,88,0,0,0,86,0,0,0,89,0,0,0,87,0,0,0,90,0,0,0,88,0,0,0,91,0,0,0,89,0,0,0,92,0,0,0,90,0,0,0,93,0,0,0,91,0,0,0,94,0,0,0,92,0,0,0,95,0,0,0,93,0,0,0,96,0,0,0,94,0,0,0,97,0,0,0,95,0,0,0,98,0,0,0,96,0,0,0,99,0,0,0,97,0,0,0,100,0,0,0,98,0,0,0,101,0,0,0,99,0,0,0,102,0,0,0,100,0,0,0,103,0,0,0,101,0,0,0,104,0,0,0,102,0,0,0,105,0,0,0,103,0,0,0,106,0,0,0,104,0,0,0,107,0,0,0,105,0,0,0,108,0,0,0,106,0,0,0,109,0,0,0,107,0,0,0,110,0,0,0,108,0,0,0,111,0,0,0,109,0,0,0,112,0,0,0,110,0,0,0,113,0,0,0,111,0,0,0,114,0,0,0,112,0,0,0,115,0,0,0,113,0,0,0,116,0,0,0,114,0,0,0,117,0,0,0,115,0,0,0,118,0,0,0,116,0,0,0,119,0,0,0,117,0,0,0,120,0,0,0,118,0,0,0,121,0,0,0,119,0,0,0,122,0,0,0,120,0,0,0,123,0,0,0,121,0,0,0,124,0,0,0,122,0,0,0,125,0,0,0,123,0,0,0,126,0,0,0,124,0,0,0,127,0,0,0,125,0,0,0,128,0,0,0,126,0,0,0,129,0,0,0,127,0,0,0,133,0,0,0,128,0,0,0,134,0,0,0,129,0,0,0,135,0,0,0,130,0,0,0,136,0,0,0,131,0,0,0,137,0,0,0,132,0,0,0,138,0,0,0,133,0,0,0,139,0,0,0,134,0,0,0,140,0,0,0,135,0,0,0,141,0,0,0,136,0,0,0,142,0,0,0,137,0,0,0,143,0,0,0,138,0,0,0,144,0,0,0,139,0,0,0,145,0,0,0,140,0,0,0,146,0,0,0,141,0,0,0,147,0,0,0,142,0,0,0,148,0,0,0,143,0,0,0,149,0,0,0,144,0,0,0,150,0,0,0,145,0,0,0,151,0,0,0,146,0,0,0,152,0,0,0,147,0,0,0,153,0,0,0,148,0,0,0,154,0,0,0,149,0,0,0,155,0,0,0,150,0,0,0,156,0,0,0,151,0,0,0,157,0,0,0,152,0,0,0,158,0,0,0,153,0,0,0,159,0,0,0,154,0,0,0,160,0,0,0,155,0,0,0,161,0,0,0,156,0,0,0,162,0,0,0,157,0,0,0,163,0,0,0,158,0,0,0,164,0,0,0,159,0,0,0,176,0,0,0,160,0,0,0,177,0,0,0,161,0,0,0,178,0,0,0,162,0,0,0,179,0,0,0,163,0,0,0,180,0,0,0,164,0,0,0,181,0,0,0,165,0,0,0,182,0,0,0,166,0,0,0,183,0,0,0,167,0,0,0,184,0,0,0,168,0,0,0,185,0,0,0,169,0,0,0,186,0,0,0,170,0,0,0,187,0,0,0,171,0,0,0,188,0,0,0,172,0,0,0,189,0,0,0,173,0,0,0,190,0,0,0,174,0,0,0,191,0,0,0,175,0,0,0,192,0,0,0,176,0,0,0,193,0,0,0,177,0,0,0,194,0,0,0,178,0,0,0,195,0,0,0,179,0,0,0,196,0,0,0,180,0,0,0,197,0,0,0,181,0,0,0,198,0,0,0,182,0,0,0,199,0,0,0,183,0,0,0,200,0,0,0,184,0,0,0,201,0,0,0,185,0,0,0,202,0,0,0,186,0,0,0,203,0,0,0,187,0,0,0,204,0,0,0,188,0,0,0,205,0,0,0,189,0,0,0,206,0,0,0,190,0,0,0,207,0,0,0,191,0,0,0,208,0,0,0,192,0,0,0,209,0,0,0,193,0,0,0,210,0,0,0,194,0,0,0,211,0,0,0,195,0,0,0,212,0,0,0,196,0,0,0,213,0,0,0,197,0,0,0,214,0,0,0,198,0,0,0,215,0,0,0,199,0,0,0,216,0,0,0,200,0,0,0,217,0,0,0,201,0,0,0,218,0,0,0,202,0,0,0,219,0,0,0,203,0,0,0,220,0,0,0,204,0,0,0,221,0,0,0,205,0,0,0,224,0,0,0,206,0,0,0,225,0,0,0,207,0,0,0,226,0,0,0,208,0,0,0,227,0,0,0,209,0,0,0,228,0,0,0,210,0,0,0,229,0,0,0,211,0,0,0,230,0,0,0,212,0,0,0,231,0,0,0,213,0,0,0,1,1,0,0,214,0,0,0,2,1,0,0,215,0,0,0,3,1,0,0,216,0,0,0,4,1,0,0,217,0,0,0,5,1,0,0,218,0,0,0,6,1,0,0,219,0,0,0,7,1,0,0,220,0,0,0,8,1,0,0,221,0,0,0,9,1,0,0,222,0,0,0,10,1,0,0,223,0,0,0,11,1,0,0,224,0,0,0,12,1,0,0,225,0,0,0,13,1,0,0,226,0,0,0,14,1,0,0,227,0,0,0,15,1,0,0,228,0,0,0,16,1,0,0,229,0,0,0,17,1,0,0,230,0,0,0,18,1,0,0,231,0,0,0,19,1,0,0,232,0,0,0,20,1,0,0,233,0,0,0,21,1,0,0,234,0,0,0,22,1,0,0,235,0,0,0,23,1,0,0,236,0,0,0,24,1,0,0,237,0,0,0,25,1,0,0,238,0,0,0,26,1,0,0,239,0,0,0,27,1,0,0,240,0,0,0,28,1,0,0,56,192,6,0,0,0,0,0,59,243,7,0,1,0,0,0,64,192,6,0,2,0,0,0,71,192,6,0,3,0,0,0,81,192,6,0,4,0,0,0,85,192,6,0,5,0,0,0,91,192,6,0,6,0,0,0,93,192,6,0,7,0,0,0,95,192,6,0,8,0,0,0,143,24,8,0,9,0,0,0,97,192,6,0,10,0,0,0,99,192,6,0,11,0,0,0,101,192,6,0,12,0,0,0,103,192,6,0,13,0,0,0,105,192,6,0,14,0,0,0,89,188,7,0,15,0,0,0,87,2,8,0,16,0,0,0,107,192,6,0,17,0,0,0,37,30,8,0,18,0,0,0,185,115,9,0,19,0,0,0,39,252,7,0,20,0,0,0,130,57,9,0,21,0,0,0,75,32,8,0,22,0,0,0,109,192,6,0,23,0,0,0,111,192,6,0,24,0,0,0,113,192,6,0,25,0,0,0,159,2,8,0,26,0,0,0,78,2,8,0,27,0,0,0,65,2,8,0,28,0,0,0,115,192,6,0,29,0,0,0,106,2,8,0,30,0,0,0,117,192,6,0,31,0,0,0,119,192,6,0,32,0,0,0,121,192,6,0,33,0,0,0,123,192,6,0,34,0,0,0,125,192,6,0,35,0,0,0,71,252,7,0,36,0,0,0,127,192,6,0,37,0,0,0,129,192,6,0,38,0,0,0,131,192,6,0,39,0,0,0,133,192,6,0,40,0,0,0,135,192,6,0,41,0,0,0,140,252,7,0,42,0,0,0,137,192,6,0,43,0,0,0,33,82,9,0,44,0,0,0,35,82,9,0,45,0,0,0,139,192,6,0,46,0,0,0,1,3,8,0,47,0,0,0,141,192,6,0,48,0,0,0,255,253,7,0,49,0,0,0,143,192,6,0,50,0,0,0,7,88,9,0,51,0,0,0,145,192,6,0,52,0,0,0,147,192,6,0,53,0,0,0,149,192,6,0,54,0,0,0,191,120,9,0,55,0,0,0,63,2,8,0,56,0,0,0,255,236,7,0,57,0,0,0,151,192,6,0,58,0,0,0,153,192,6,0,59,0,0,0,155,192,6,0,60,0,0,0,39,11,8,0,61,0,0,0,157,192,6,0,62,0,0,0,159,192,6,0,63,0,0,0,161,192,6,0,64,0,0,0,163,192,6,0,65,0,0,0,5,88,9,0,66,0,0,0,37,82,9,0,67,0,0,0,39,82,9,0,68,0,0,0,165,192,6,0,69,0,0,0,167,192,6,0,70,0,0,0,176,192,6,0,71,0,0,0,179,192,6,0,72,0,0,0,182,192,6,0,73,0,0,0,185,192,6,0,74,0,0,0,188,192,6,0,75,0,0,0,191,192,6,0,76,0,0,0,194,192,6,0,77,0,0,0,197,192,6,0,78,0,0,0,200,192,6,0,79,0,0,0,203,192,6,0,80,0,0,0,207,192,6,0,81,0,0,0,211,192,6,0,82,0,0,0,215,192,6,0,83,0,0,0,227,192,6,0,84,0,0,0,238,192,6,0,85,0,0,0,133,248,7,0,86,0,0,0,244,192,6,0,87,0,0,0,249,192,6,0,88,0,0,0,0,193,6,0,89,0,0,0,160,245,7,0,90,0,0,0,7,193,6,0,91,0,0,0,16,193,6,0,92,0,0,0,22,193,6,0,93,0,0,0,27,193,6,0,94,0,0,0,32,193,6,0,95,0,0,0,35,193,6,0,96,0,0,0,43,193,6,0,97,0,0,0,47,193,6,0,98,0,0,0,51,193,6,0,99,0,0,0,55,193,6,0,100,0,0,0,59,193,6,0,101,0,0,0,67,193,6,0,111,0,0,0,71,193,6,0,102,0,0,0,75,193,6,0,103,0,0,0,79,193,6,0,104,0,0,0,83,193,6,0,105,0,0,0,87,193,6,0,106,0,0,0,91,193,6,0,107,0,0,0,95,193,6,0,108,0,0,0,99,193,6,0,109,0,0,0,103,193,6,0,110,0,0,0,107,193,6,0,112,0,0,0,111,193,6,0,113,0,0,0,115,193,6,0,114,0,0,0,119,193,6,0,115,0,0,0,131,193,6,0,116,0,0,0,137,193,6,0,117,0,0,0,141,193,6,0,118,0,0,0,145,193,6,0,119,0,0,0,149,193,6,0,120,0,0,0,153,193,6,0,121,0,0,0,157,193,6,0,122,0,0,0,161,193,6,0,123,0,0,0,165,193,6,0,124,0,0,0,169,193,6,0,125,0,0,0,173,193,6,0,126,0,0,0,177,193,6,0,127,0,0,0,181,193,6,0,128,0,0,0,138,247,7,0,129,0,0,0,185,193,6,0,130,0,0,0,190,193,6,0,131,0,0,0,96,241,7,0,132,0,0,0,3,240,7,0,133,0,0,0,195,193,6,0,134,0,0,0,201,193,6,0,135,0,0,0,206,193,6,0,136,0,0,0,210,193,6,0,137,0,0,0,215,193,6,0,138,0,0,0,81,249,7,0,139,0,0,0,96,10,8,0,140,0,0,0,221,193,6,0,141,0,0,0,230,193,6,0,142,0,0,0,241,193,6,0,143,0,0,0,250,193,6,0,144,0,0,0,1,194,6,0,145,0,0,0,8,194,6,0,146,0,0,0,14,194,6,0,147,0,0,0,20,194,6,0,148,0,0,0,28,194,6,0,149,0,0,0,38,194,6,0,150,0,0,0,42,194,6,0,151,0,0,0,47,194,6,0,152,0,0,0,58,194,6,0,153,0,0,0,78,194,6,0,154,0,0,0,95,194,6,0,155,0,0,0,108,194,6,0,156,0,0,0,124,194,6,0,157,0,0,0,130,194,6,0,158,0,0,0,137,194,6,0,159,0,0,0,142,194,6,0,160,0,0,0,147,194,6,0,161,0,0,0,153,194,6,0,162,0,0,0,160,194,6,0,163,0,0,0,165,194,6,0,164,0,0,0,154,25,9,0,165,0,0,0,170,194,6,0,166,0,0,0,180,194,6,0,167,0,0,0,190,194,6,0,168,0,0,0,200,194,6,0,169,0,0,0,210,194,6,0,170,0,0,0,220,194,6,0,171,0,0,0,232,194,6,0,172,0,0,0,236,194,6,0,173,0,0,0,241,194,6,0,174,0,0,0,252,194,6,0,175,0,0,0,5,195,6,0,176,0,0,0,15,195,6,0,177,0,0,0,23,195,6,0,178,0,0,0,31,195,6,0,179,0,0,0,42,195,6,0,180,0,0,0,50,195,6,0,181,0,0,0,61,195,6,0,182,0,0,0,74,195,6,0,183,0,0,0,89,195,6,0,184,0,0,0,102,195,6,0,185,0,0,0,116,195,6,0,186,0,0,0,131,195,6,0,187,0,0,0,144,195,6,0,188,0,0,0,155,195,6,0,189,0,0,0,161,195,6,0,190,0,0,0,56,192,6,0,0,0,0,0,33,82,9,0,1,0,0,0,35,82,9,0,2,0,0,0,139,192,6,0,3,0,0,0,1,3,8,0,4,0,0,0,141,192,6,0,5,0,0,0,255,253,7,0,6,0,0,0,143,192,6,0,7,0,0,0,7,88,9,0,8,0,0,0,145,192,6,0,9,0,0,0,147,192,6,0,10,0,0,0,149,192,6,0,11,0,0,0,191,120,9,0,12,0,0,0,63,2,8,0,13,0,0,0,255,236,7,0,14,0,0,0,151,192,6,0,15,0,0,0,153,192,6,0,16,0,0,0,155,192,6,0,17,0,0,0,39,11,8,0,18,0,0,0,157,192,6,0,19,0,0,0,159,192,6,0,20,0,0,0,161,192,6,0,21,0,0,0,163,192,6,0,22,0,0,0,5,88,9,0,23,0,0,0,37,82,9,0,24,0,0,0,39,82,9,0,25,0,0,0,165,192,6,0,26,0,0,0,75,32,8,0,27,0,0,0,109,192,6,0,28,0,0,0,111,192,6,0,29,0,0,0,113,192,6,0,30,0,0,0,159,2,8,0,31,0,0,0,78,2,8,0,32,0,0,0,65,2,8,0,33,0,0,0,115,192,6,0,34,0,0,0,106,2,8,0,35,0,0,0,130,57,9,0,36,0,0,0,59,243,7,0,37,0,0,0,64,192,6,0,38,0,0,0,71,192,6,0,39,0,0,0,81,192,6,0,40,0,0,0,85,192,6,0,41,0,0,0,37,30,8,0,42,0,0,0,123,192,6,0,43,0,0,0,129,192,6,0,44,0,0,0,133,192,6,0,45,0,0,0,131,192,6,0,46,0,0,0,167,195,6,0,47,0,0,0,119,192,6,0,48,0,0,0,101,192,6,0,49,0,0,0,137,192,6,0,50,0,0,0,107,192,6,0,51,0,0,0,185,115,9,0,52,0,0,0,39,252,7,0,53,0,0,0,167,192,6,0,54,0,0,0,176,192,6,0,55,0,0,0,179,192,6,0,56,0,0,0,182,192,6,0,57,0,0,0,185,192,6,0,58,0,0,0,188,192,6,0,59,0,0,0,191,192,6,0,60,0,0,0,194,192,6,0,61,0,0,0,197,192,6,0,62,0,0,0,200,192,6,0,63,0,0,0,203,192,6,0,64,0,0,0,207,192,6,0,65,0,0,0,211,192,6,0,66,0,0,0,215,192,6,0,67,0,0,0,227,192,6,0,68,0,0,0,238,192,6,0,69,0,0,0,133,248,7,0,70,0,0,0,244,192,6,0,71,0,0,0,249,192,6,0,72,0,0,0,0,193,6,0,73,0,0,0,160,245,7,0,74,0,0,0,7,193,6,0,75,0,0,0,16,193,6,0,76,0,0,0,22,193,6,0,77,0,0,0,27,193,6,0,78,0,0,0,32,193,6,0,79,0,0,0,35,193,6,0,80,0,0,0,43,193,6,0,81,0,0,0,47,193,6,0,82,0,0,0,51,193,6,0,83,0,0,0,55,193,6,0,84,0,0,0,59,193,6,0,85,0,0,0,71,193,6,0,86,0,0,0,75,193,6,0,87,0,0,0,79,193,6,0,88,0,0,0,83,193,6,0,89,0,0,0,87,193,6,0,90,0,0,0,91,193,6,0,91,0,0,0,95,193,6,0,92,0,0,0,99,193,6,0,93,0,0,0,103,193,6,0,94,0,0,0,67,193,6,0,95,0,0,0,107,193,6,0,96,0,0,0,174,195,6,0,97,0,0,0,119,193,6,0,98,0,0,0,131,193,6,0,99,0,0,0,115,193,6,0,100,0,0,0,137,193,6,0,101,0,0,0,141,193,6,0,102,0,0,0,145,193,6,0,103,0,0,0,149,193,6,0,104,0,0,0,153,193,6,0,105,0,0,0,157,193,6,0,106,0,0,0,161,193,6,0,107,0,0,0,165,193,6,0,108,0,0,0,169,193,6,0,109,0,0,0,173,193,6,0,110,0,0,0,177,193,6,0,111,0,0,0,181,193,6,0,112,0,0,0,138,247,7,0,113,0,0,0,185,193,6,0,114,0,0,0,190,193,6,0,115,0,0,0,96,241,7,0,116,0,0,0,3,240,7,0,117,0,0,0,195,193,6,0,118,0,0,0,201,193,6,0,119,0,0,0,206,193,6,0,120,0,0,0,210,193,6,0,121,0,0,0,215,193,6,0,122,0,0,0,81,249,7,0,123,0,0,0,96,10,8,0,124,0,0,0,221,193,6,0,125,0,0,0,230,193,6,0,126,0,0,0,111,193,6,0,127,0,0,0,189,195,6,0,128,0,0,0,196,195,6,0,129,0,0,0,211,195,6,0,130,0,0,0,226,195,6,0,131,0,0,0,241,195,6,0,132,0,0,0,0,196,6,0,133,0,0,0,15,196,6,0,134,0,0,0,30,196,6,0,135,0,0,0,45,196,6,0,136,0,0,0,60,196,6,0,137,0,0,0,75,196,6,0,138,0,0,0,81,196,6,0,139,0,0,0,87,196,6,0,140,0,0,0,93,196,6,0,141,0,0,0,99,196,6,0,142,0,0,0,105,196,6,0,143,0,0,0,111,196,6,0,144,0,0,0,117,196,6,0,145,0,0,0,123,196,6,0,146,0,0,0,241,193,6,0,147,0,0,0,250,193,6,0,148,0,0,0,1,194,6,0,149,0,0,0,8,194,6,0,150,0,0,0,14,194,6,0,151,0,0,0,20,194,6,0,152,0,0,0,28,194,6,0,153,0,0,0,38,194,6,0,154,0,0,0,42,194,6,0,155,0,0,0,47,194,6,0,156,0,0,0,129,196,6,0,157,0,0,0,135,196,6,0,158,0,0,0,141,196,6,0,159,0,0,0,146,196,6,0,160,0,0,0,58,194,6,0,161,0,0,0,78,194,6,0,162,0,0,0,95,194,6,0,163,0,0,0,108,194,6,0,164,0,0,0,152,196,6,0,165,0,0,0,156,196,6,0,166,0,0,0,160,196,6,0,167,0,0,0,164,196,6,0,168,0,0,0,168,196,6,0,169,0,0,0,174,196,6,0,170,0,0,0,186,196,6,0,171,0,0,0,190,196,6,0,172,0,0,0,194,196,6,0,173,0,0,0,198,196,6,0,174,0,0,0,202,196,6,0,175,0,0,0,206,196,6,0,176,0,0,0,210,196,6,0,177,0,0,0,216,196,6,0,178,0,0,0,223,196,6,0,179,0,0,0,227,196,6,0,180,0,0,0,231,196,6,0,181,0,0,0,235,196,6,0,182,0,0,0,239,196,6,0,183,0,0,0,244,196,6,0,184,0,0,0,248,196,6,0,185,0,0,0,253,196,6,0,186,0,0,0,1,197,6,0,187,0,0,0,5,197,6,0,188,0,0,0,9,197,6,0,189,0,0,0,13,197,6,0,190,0,0,0,17,197,6,0,191,0,0,0,28,197,6,0,192,0,0,0,40,197,6,0,193,0,0,0,51,197,6,0,194,0,0,0,58,197,6,0,195,0,0,0,65,197,6,0,196,0,0,0,72,197,6,0,197,0,0,0,79,197,6,0,198,0,0,0,84,197,6,0,199,0,0,0,92,197,6,0,200,0,0,0,105,197,6,0,201,0,0,0,114,197,6,0,202,0,0,0,122,197,6,0,203,0,0,0,132,197,6,0,204,0,0,0,124,194,6,0,205,0,0,0,130,194,6,0,206,0,0,0,137,194,6,0,207,0,0,0,142,194,6,0,208,0,0,0,147,194,6,0,209,0,0,0,153,194,6,0,210,0,0,0,160,194,6,0,211,0,0,0,165,194,6,0,212,0,0,0,154,25,9,0,213,0,0,0,170,194,6,0,214,0,0,0,180,194,6,0,215,0,0,0,190,194,6,0,216,0,0,0,200,194,6,0,217,0,0,0,210,194,6,0,218,0,0,0,220,194,6,0,219,0,0,0,232,194,6,0,220,0,0,0,236,194,6,0,221,0,0,0,241,194,6,0,222,0,0,0,252,194,6,0,223,0,0,0,138,197,6,0,224,0,0,0,147,197,6,0,225,0,0,0,154,197,6,0,226,0,0,0,161,197,6,0,227,0,0,0,171,197,6,0,228,0,0,0,178,197,6,0,229,0,0,0,188,197,6,0,230,0,0,0,74,195,6,0,231,0,0,0,89,195,6,0,232,0,0,0,102,195,6,0,233,0,0,0,116,195,6,0,234,0,0,0,131,195,6,0,235,0,0,0,144,195,6,0,236,0,0,0,155,195,6,0,237,0,0,0,161,195,6,0,238,0,0,0,200,197,6,0,239,0,0,0,205,197,6,0,240,0,0,0,0,0,0,0,144,9,0,0,147,7,0,0,148,7,0,0,3,0,0,0,4,0,0,0,0,0,0,0,160,9,0,0,149,7,0,0,150,7,0,0,3,0,0,0,4,0,0,0,151,7,0,0,152,7,0,0,153,7,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,2,0,0,0,2,0,0,0,3,0,0,0,3,0,0,0,4,0,0,0,4,0,0,0,5,0,0,0,5,0,0,0,6,0,0,0,6,0,0,0,7,0,0,0,7,0,0,0,8,0,0,0,8,0,0,0,9,0,0,0,9,0,0,0,10,0,0,0,10,0,0,0,11,0,0,0,11,0,0,0,0,0,0,0,176,9,0,0,154,7,0,0,155,7,0,0,3,0,0,0,4,0,0,0,0,0,0,0,192,9,0,0,156,7,0,0,157,7,0,0,3,0,0,0,4,0,0,0,0,0,0,0,208,9,0,0,158,7,0,0,159,7,0,0,3,0,0,0,4,0,0,0,0,0,0,0,224,9,0,0,160,7,0,0,161,7,0,0,3,0,0,0,4,0,0,0,0,0,0,0,240,9,0,0,162,7,0,0,163,7,0,0,3,0,0,0,4,0,0,0,0,0,0,0,0,10,0,0,164,7,0,0,165,7,0,0,3,0,0,0,4,0,0,0,0,0,0,0,16,10,0,0,166,7,0,0,167,7,0,0,3,0,0,0,4,0,0,0,0,0,0,0,32,10,0,0,168,7,0,0,169,7,0,0,3,0,0,0,4,0,0,0,0,0,0,0,48,10,0,0,170,7,0,0,171,7,0,0,3,0,0,0,4,0,0,0,0,0,0,0,64,10,0,0,172,7,0,0,173,7,0,0,3,0,0,0,4,0,0,0,0,0,0,0,80,10,0,0,174,7,0,0,175,7,0,0,3,0,0,0,4,0,0,0,0,0,0,0,96,10,0,0,176,7,0,0,177,7,0,0,3,0,0,0,4,0,0,0,5,0,0,0,5,0,0,0,3,4,0,0,4,4,0,0,5,0,0,0,5,0,0,0,5,0,0,0,8,4,0,0,5,0,0,0,5,0,0,0,117,5,0,0,5,0,0,0,0,0,0,0,128,10,0,0,178,7,0,0,179,7,0,0,3,0,0,0,4,0,0,0,180,7,0,0,181,7,0,0,3,4,0,0,4,4,0,0,182,7,0,0,183,7,0,0,184,7,0,0,8,4,0,0,185,7,0,0,186,7,0,0,117,5,0,0,187,7,0,0,0,0,0,0,144,10,0,0,188,7,0,0,189,7,0,0,190,7,0,0,191,7,0,0,0,0,0,0,168,10,0,0,192,7,0,0,193,7,0,0,194,7,0,0,195,7,0,0,196,7,0,0,0,0,0,0,192,10,0,0,197,7,0,0,198,7,0,0,199,7,0,0,200,7,0,0,0,0,0,0,208,10,0,0,201,7,0,0,202,7,0,0,3,0,0,0,4,0,0,0,203,7,0,0,248,255,255,255,208,10,0,0,204,7,0,0,205,7,0,0,206,7,0,0,112,80,0,0,108,129,0,0,0,0,0,0,128,2,0,0,207,7,0,0,208,7,0,0,209,7,0,0,210,7,0,0,211,7,0,0,212,7,0,0,0,0,0,0,240,10,0,0,213,7,0,0,211,4,0,0,214,7,0,0,0,0,0,0,0,11,0,0,215,7,0,0,211,4,0,0,216,7,0,0,0,0,0,0,16,11,0,0,217,7,0,0,211,4,0,0,218,7,0,0,0,0,0,0,32,11,0,0,219,7,0,0,220,7,0,0,221,7,0,0,222,7,0,0,223,7,0,0,222,4,0,0,223,4,0,0,224,7,0,0,225,7,0,0,226,7,0,0,227,7,0,0,0,0,0,0,48,11,0,0,228,7,0,0,229,7,0,0,230,7,0,0,231,7,0,0,232,7,0,0,222,4,0,0,223,4,0,0,233,7,0,0,234,7,0,0,235,7,0,0,236,7,0,0,0,0,0,0,64,11,0,0,237,7,0,0,238,7,0,0,239,7,0,0,240,7,0,0,241,7,0,0,222,4,0,0,223,4,0,0,242,7,0,0,243,7,0,0,244,7,0,0,245,7,0,0,0,0,0,0,80,11,0,0,246,7,0,0,247,7,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,2,0,0,0,2,0,0,0,2,0,0,0,2,0,0,0,3,0,0,0,3,0,0,0,3,0,0,0,3,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,0,0,0,0,3,0,0,0,4,0,0,0,5,0,0,0,6,0,0,0,7,0,0,0,8,0,0,0,9,0,0,0,10,0,0,0,11,0,0,0,13,0,0,0,15,0,0,0,17,0,0,0,19,0,0,0,23,0,0,0,27,0,0,0,31,0,0,0,35,0,0,0,43,0,0,0,51,0,0,0,59,0,0,0,67,0,0,0,83,0,0,0,99,0,0,0,115,0,0,0,131,0,0,0,163,0,0,0,195,0,0,0,227,0,0,0,2,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,2,0,0,0,2,0,0,0,3,0,0,0,3,0,0,0,4,0,0,0,4,0,0,0,5,0,0,0,5,0,0,0,6,0,0,0,6,0,0,0,7,0,0,0,7,0,0,0,8,0,0,0,8,0,0,0,9,0,0,0,9,0,0,0,10,0,0,0,10,0,0,0,11,0,0,0,11,0,0,0,12,0,0,0,12,0,0,0,13,0,0,0,13,0,0,0,1,0,0,0,2,0,0,0,3,0,0,0,4,0,0,0,5,0,0,0,7,0,0,0,9,0,0,0,13,0,0,0,17,0,0,0,25,0,0,0,33,0,0,0,49,0,0,0,65,0,0,0,97,0,0,0,129,0,0,0,193,0,0,0,1,1,0,0,129,1,0,0,1,2,0,0,1,3,0,0,1,4,0,0,1,6,0,0,1,8,0,0,1,12,0,0,1,16,0,0,1,24,0,0,1,32,0,0,1,48,0,0,1,64,0,0,1,96,0,0,0,0,0,0,150,48,7,119,44,97,14,238,186,81,9,153,25,196,109,7,143,244,106,112,53,165,99,233,163,149,100,158,50,136,219,14,164,184,220,121,30,233,213,224,136,217,210,151,43,76,182,9,189,124,177,126,7,45,184,231,145,29,191,144,100,16,183,29,242,32,176,106,72,113,185,243,222,65,190,132,125,212,218,26,235,228,221,109,81,181,212,244,199,133,211,131,86,152,108,19,192,168,107,100,122,249,98,253,236,201,101,138,79,92,1,20,217,108,6,99,99,61,15,250,245,13,8,141,200,32,110,59,94,16,105,76,228,65,96,213,114,113,103,162,209,228,3,60,71,212,4,75,253,133,13,210,107,181,10,165,250,168,181,53,108,152,178,66,214,201,187,219,64,249,188,172,227,108,216,50,117,92,223,69,207,13,214,220,89,61,209,171,172,48,217,38,58,0,222,81,128,81,215,200,22,97,208,191,181,244,180,33,35,196,179,86,153,149,186,207,15,165,189,184,158,184,2,40,8,136,5,95,178,217,12,198,36,233,11,177,135,124,111,47,17,76,104,88,171,29,97,193,61,45,102,182,144,65,220,118,6,113,219,1,188,32,210,152,42,16,213,239,137,133,177,113,31,181,182,6,165,228,191,159,51,212,184,232,162,201,7,120,52,249,0,15,142,168,9,150,24,152,14,225,187,13,106,127,45,61,109,8,151,108,100,145,1,92,99,230,244,81,107,107,98,97,108,28,216,48,101,133,78,0,98,242,237,149,6,108,123,165,1,27,193,244,8,130,87,196,15,245,198,217,176,101,80,233,183,18,234,184,190,139,124,136,185,252,223,29,221,98,73,45,218,21,243,124,211,140,101,76,212,251,88,97,178,77,206,81,181,58,116,0,188,163,226,48,187,212,65,165,223,74,215,149,216,61,109,196,209,164,251,244,214,211,106,233,105,67,252,217,110,52,70,136,103,173,208,184,96,218,115,45,4,68,229,29,3,51,95,76,10,170,201,124,13,221,60,113,5,80,170,65,2,39,16,16,11,190,134,32,12,201,37,181,104,87,179,133,111,32,9,212,102,185,159,228,97,206,14,249,222,94,152,201,217,41,34,152,208,176,180,168,215,199,23,61,179,89,129,13,180,46,59,92,189,183,173,108,186,192,32,131,184,237,182,179,191,154,12,226,182,3,154,210,177,116,57,71,213,234,175,119,210,157,21,38,219,4,131,22,220,115,18,11,99,227,132,59,100,148,62,106,109,13,168,90,106,122,11,207,14,228,157,255,9,147,39,174,0,10,177,158,7,125,68,147,15,240,210,163,8,135,104,242,1,30,254,194,6,105,93,87,98,247,203,103,101,128,113,54,108,25,231,6,107,110,118,27,212,254,224,43,211,137,90,122,218,16,204,74,221,103,111,223,185,249,249,239,190,142,67,190,183,23,213,142,176,96,232,163,214,214,126,147,209,161,196,194,216,56,82,242,223,79,241,103,187,209,103,87,188,166,221,6,181,63,75,54,178,72,218,43,13,216,76,27,10,175,246,74,3,54,96,122,4,65,195,239,96,223,85,223,103,168,239,142,110,49,121,190,105,70,140,179,97,203,26,131,102,188,160,210,111,37,54,226,104,82,149,119,12,204,3,71,11,187,185,22,2,34,47,38,5,85,190,59,186,197,40,11,189,178,146,90,180,43,4,106,179,92,167,255,215,194,49,207,208,181,139,158,217,44,29,174,222,91,176,194,100,155,38,242,99,236,156,163,106,117,10,147,109,2,169,6,9,156,63,54,14,235,133,103,7,114,19,87,0,5,130,74,191,149,20,122,184,226,174,43,177,123,56,27,182,12,155,142,210,146,13,190,213,229,183,239,220,124,33,223,219,11,212,210,211,134,66,226,212,241,248,179,221,104,110,131,218,31,205,22,190,129,91,38,185,246,225,119,176,111,119,71,183,24,230,90,8,136,112,106,15,255,202,59,6,102,92,11,1,17,255,158,101,143,105,174,98,248,211,255,107,97,69,207,108,22,120,226,10,160,238,210,13,215,84,131,4,78,194,179,3,57,97,38,103,167,247,22,96,208,77,71,105,73,219,119,110,62,74,106,209,174,220,90,214,217,102,11,223,64,240,59,216,55,83,174,188,169,197,158,187,222,127,207,178,71,233,255,181,48,28,242,189,189,138,194,186,202,48,147,179,83,166,163,180,36,5,54,208,186,147,6,215,205,41,87,222,84,191,103,217,35,46,122,102,179,184,74,97,196,2,27,104,93,148,43,111,42,55,190,11,180,161,142,12,195,27,223,5,90,141,239,2,45,8,0,0,0,8,0,0,0,4,0,0,0,4,0,0,0,2,0,0,0,2,0,0,0,1,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,2,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,4,0,0,0,4,0,0,0,2,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,2,0,0,0,0,0,0,0,1,0,0,0,249,186,7,0,248,7,0,0,0,187,7,0,249,7,0,0,0,0,0,0,0,0,0,0,12,187,7,0,250,7,0,0,26,187,7,0,251,7,0,0,37,187,7,0,252,7,0,0,0,0,0,0,0,0,0,0,0,0,0,0,104,11,0,0,253,7,0,0,254,7,0,0,3,0,0,0,4,0,0,0,255,7,0,0,0,8,0,0,1,8,0,0,2,8,0,0,92,6,0,0,3,8,0,0,4,8,0,0,95,6,0,0,5,8,0,0,6,8,0,0,7,8,0,0,8,8,0,0,9,8,0,0,10,8,0,0,11,8,0,0,12,8,0,0,104,6,0,0,0,0,0,0,120,11,0,0,7,3,0,0,13,8,0,0,3,0,0,0,4,0,0,0,14,8,0,0,15,8,0,0,0,0,0,0,136,11,0,0,16,8,0,0,17,8,0,0,3,0,0,0,4,0,0,0,207,2,0,0,208,2,0,0,0,0,0,0,152,11,0,0,7,3,0,0,18,8,0,0,3,0,0,0,4,0,0,0,19,8,0,0,20,8,0,0,0,0,0,0,168,11,0,0,59,7,0,0,21,8,0,0,3,0,0,0,4,0,0,0,22,8,0,0,23,8,0,0,24,8,0,0,25,8,0,0,26,8,0,0,0,0,0,0,184,11,0,0,27,8,0,0,28,8,0,0,29,8,0,0,30,8,0,0,31,8,0,0,140,45,10,0,204,45,10,0,236,45,10,0,252,45,10,0,4,46,10,0,0,0,0,0,32,0,0,0,64,0,0,0,96,0,0,0,128,0,0,0,160,0,0,0,192,0,0,0,224,0,0,0,0,1,0,0,32,1,0,0,64,1,0,0,96,1,0,0,128,1,0,0,160,1,0,0,192,1,0,0,0,0,0,0,0,0,0,0,32,0,0,0,48,0,0,0,56,0,0,0,64,0,0,0,80,0,0,0,96,0,0,0,112,0,0,0,128,0,0,0,160,0,0,0,192,0,0,0,224,0,0,0,0,1,0,0,64,1,0,0,128,1,0,0,0,0,0,0,0,0,0,0,32,0,0,0,40,0,0,0,48,0,0,0,56,0,0,0,64,0,0,0,80,0,0,0,96,0,0,0,112,0,0,0,128,0,0,0,160,0,0,0,192,0,0,0,224,0,0,0,0,1,0,0,64,1,0,0,0,0,0,0,0,0,0,0,32,0,0,0,48,0,0,0,56,0,0,0,64,0,0,0,80,0,0,0,96,0,0,0,112,0,0,0,128,0,0,0,144,0,0,0,160,0,0,0,176,0,0,0,192,0,0,0,224,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,8,0,0,0,16,0,0,0,24,0,0,0,32,0,0,0,40,0,0,0,48,0,0,0,56,0,0,0,64,0,0,0,80,0,0,0,96,0,0,0,112,0,0,0,128,0,0,0,144,0,0,0,160,0,0,0,0,0,0,0,0,0,0,0,8,0,0,0,16,0,0,0,24,0,0,0,32,0,0,0,40,0,0,0,48,0,0,0,56,0,0,0,64,0,0,0,80,0,0,0,96,0,0,0,112,0,0,0,128,0,0,0,144,0,0,0,160,0,0,0,0,0,0,0,68,172,0,0,128,187,0,0,0,125,0,0,34,86,0,0,192,93,0,0,128,62,0,0,17,43,0,0,224,46,0,0,64,31,0,0,110,195,7,0,114,195,7,0,118,195,7,0,122,195,7,0,126,195,7,0,130,195,7,0,134,195,7,0,138,195,7,0,142,195,7,0,146,195,7,0,150,195,7,0,154,195,7,0,158,195,7,0,162,195,7,0,166,195,7,0,170,195,7,0,174,195,7,0,178,195,7,0,182,195,7,0,186,195,7,0,190,195,7,0,194,195,7,0,198,195,7,0,202,195,7,0,206,195,7,0,210,195,7,0,214,195,7,0,134,195,7,0,218,195,7,0,222,195,7,0,226,195,7,0,230,195,7,0,234,195,7,0,238,195,7,0,242,195,7,0,246,195,7,0,250,195,7,0,254,195,7,0,3,196,7,0,8,196,7,0,13,196,7,0,18,196,7,0,23,196,7,0,28,196,7,0,33,196,7,0,38,196,7,0,43,196,7,0,48,196,7,0,53,196,7,0,58,196,7,0,63,196,7,0,68,196,7,0,73,196,7,0,78,196,7,0,83,196,7,0,88,196,7,0,93,196,7,0,98,196,7,0,103,196,7,0,108,196,7,0,113,196,7,0,118,196,7,0,123,196,7,0,128,196,7,0,133,196,7,0,138,196,7,0,143,196,7,0,148,196,7,0,153,196,7,0,158,196,7,0,163,196,7,0,168,196,7,0,173,196,7,0,178,196,7,0,1,0,0,0,2,0,0,0,2,0,0,0,1,0,0,0,32,8,0,0,33,8,0,0,33,8,0,0,34,8,0,0,0,0,0,0,2,0,0,0,2,0,0,0,2,0,0,0,2,0,0,0,2,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,2,0,0,0,2,0,0,0,2,0,0,0,2,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,3,0,0,0,3,0,0,0,3,0,0,0,3,0,0,0,3,0,0,0,3,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,3,0,0,0,3,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,27,0,0,0,30,0,0,0,8,0,0,0,12,0,0,0,30,0,0,0,72,105,2,0,200,109,2,0,120,114,2,0,184,115,2,0,120,117,2,0,2,0,0,0,9,0,0,0,5,0,0,0,3,0,0,0,4,0,0,0,1,0,0,0,8,0,0,0,1,0,0,0,2,0,0,0,9,0,0,0,0,0,0,0,160,116,18,0,0,0,0,0,102,124,2,0,0,0,0,0,116,124,2,0,0,0,0,0,150,124,2,0,0,0,0,0,160,116,18,0,0,0,0,0,184,124,2,0,0,0,0,0,246,124,2,0,0,0,0,0,52,125,2,0,0,0,0,0,194,125,2,0,0,0,0,0,80,126,2,0,0,0,0,0,222,126,2,0,0,0,0,0,220,127,2,0,0,0,0,0,218,128,2,0,0,0,0,0,216,129,2,0,0,0,0,0,160,116,18,0,0,0,0,0,214,133,2,0,1,0,0,0,212,137,2,0,2,0,0,0,212,137,2,0,3,0,0,0,212,137,2,0,4,0,0,0,212,137,2,0,6,0,0,0,212,137,2,0,8,0,0,0,212,137,2,0,10,0,0,0,212,137,2,0,13,0,0,0,212,137,2,0,4,0,0,0,210,141,2,0,5,0,0,0,210,141,2,0,6,0,0,0,210,141,2,0,7,0,0,0,210,141,2,0,8,0,0,0,210,141,2,0,9,0,0,0,210,141,2,0,11,0,0,0,210,141,2,0,13,0,0,0,210,141,2,0,0,0,0,0,208,145,2,0,0,0,0,0,14,146,2,0,228,242,10,0,36,243,10,0,100,243,10,0,164,243,10,0,228,243,10,0,100,244,10,0,228,244,10,0,100,245,10,0,36,244,10,0,164,244,10,0,36,245,10,0,164,245,10,0,35,8,0,0,36,8,0,0,37,8,0,0,38,8,0,0,39,8,0,0,40,8,0,0,41,8,0,0,42,8,0,0,43,8,0,0,44,8,0,0,45,8,0,0,0,0,0,0,64,31,0,0,17,43,0,0,224,46,0,0,128,62,0,0,34,86,0,0,192,93,0,0,0,125,0,0,68,172,0,0,128,187,0,0,208,0,0,0,96,0,0,0,128,17,0,0,0,33,0,0,128,80,0,0,0,96,0,0,0,2,0,0,0,4,0,0,130,0,0,0,1,0,0,0,4,0,0,0,8,0,0,0,0,0,0,0,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,254,255,255,255,254,255,255,255,254,255,255,255,254,255,255,255,253,255,255,255,253,255,255,255,252,255,255,255,252,255,255,255,251,255,255,255,251,255,255,255,250,255,255,255,249,255,255,255,249,255,255,255,248,255,255,255,247,255,255,255,246,255,255,255,245,255,255,255,243,255,255,255,242,255,255,255,240,255,255,255,239,255,255,255,237,255,255,255,235,255,255,255,232,255,255,255,230,255,255,255,227,255,255,255,225,255,255,255,221,255,255,255,218,255,255,255,215,255,255,255,211,255,255,255,207,255,255,255,203,255,255,255,198,255,255,255,193,255,255,255,188,255,255,255,183,255,255,255,177,255,255,255,171,255,255,255,165,255,255,255,159,255,255,255,152,255,255,255,145,255,255,255,139,255,255,255,131,255,255,255,124,255,255,255,117,255,255,255,109,255,255,255,102,255,255,255,95,255,255,255,87,255,255,255,80,255,255,255,73,255,255,255,66,255,255,255,60,255,255,255,54,255,255,255,48,255,255,255,43,255,255,255,38,255,255,255,34,255,255,255,31,255,255,255,29,255,255,255,28,255,255,255,28,255,255,255,29,255,255,255,32,255,255,255,35,255,255,255,41,255,255,255,48,255,255,255,56,255,255,255,67,255,255,255,79,255,255,255,93,255,255,255,110,255,255,255,129,255,255,255,150,255,255,255,173,255,255,255,199,255,255,255,227,255,255,255,2,0,0,0,36,0,0,0,72,0,0,0,111,0,0,0,153,0,0,0,197,0,0,0,244,0,0,0,38,1,0,0,91,1,0,0,145,1,0,0,203,1,0,0,7,2,0,0,69,2,0,0,133,2,0,0,199,2,0,0,11,3,0,0,80,3,0,0,151,3,0,0,223,3,0,0,40,4,0,0,113,4,0,0,186,4,0,0,3,5,0,0,76,5,0,0,148,5,0,0,218,5,0,0,31,6,0,0,98,6,0,0,162,6,0,0,223,6,0,0,25,7,0,0,78,7,0,0,127,7,0,0,170,7,0,0,209,7,0,0,240,7,0,0,9,8,0,0,27,8,0,0,37,8,0,0,39,8,0,0,32,8,0,0,15,8,0,0,245,7,0,0,208,7,0,0,160,7,0,0,101,7,0,0,30,7,0,0,203,6,0,0,108,6,0,0,255,5,0,0,134,5,0,0,0,5,0,0,107,4,0,0,202,3,0,0,26,3,0,0,93,2,0,0,146,1,0,0,185,0,0,0,211,255,255,255,224,254,255,255,223,253,255,255,210,252,255,255,185,251,255,255,148,250,255,255,100,249,255,255,42,248,255,255,230,246,255,255,153,245,255,255,68,244,255,255,233,242,255,255,135,241,255,255,33,240,255,255,183,238,255,255,76,237,255,255,223,235,255,255,115,234,255,255,9,233,255,255,163,231,255,255,67,230,255,255,233,228,255,255,153,227,255,255,83,226,255,255,26,225,255,255,239,223,255,255,213,222,255,255,205,221,255,255,218,220,255,255,253,219,255,255,56,219,255,255,143,218,255,255,1,218,255,255,146,217,255,255,68,217,255,255,25,217,255,255,18,217,255,255,49,217,255,255,121,217,255,255,234,217,255,255,136,218,255,255,83,219,255,255,77,220,255,255,120,221,255,255,212,222,255,255,100,224,255,255,40,226,255,255,34,228,255,255,82,230,255,255,185,232,255,255,88,235,255,255,47,238,255,255,64,241,255,255,137,244,255,255,11,248,255,255,198,251,255,255,186,255,255,255,230,3,0,0,74,8,0,0,228,12,0,0,181,17,0,0,186,22,0,0,242,27,0,0,92,33,0,0,247,38,0,0,191,44,0,0,180,50,0,0,212,56,0,0,27,63,0,0,135,69,0,0,22,76,0,0,197,82,0,0,145,89,0,0,118,96,0,0,114,103,0,0,129,110,0,0,160,117,0,0,203,124,0,0,255,131,0,0,56,139,0,0,113,146,0,0,168,153,0,0,216,160,0,0,254,167,0,0,21,175,0,0,25,182,0,0,6,189,0,0,217,195,0,0,141,202,0,0,30,209,0,0,138,215,0,0,202,221,0,0,221,227,0,0,190,233,0,0,105,239,0,0,220,244,0,0,19,250,0,0,10,255,0,0,190,3,1,0,45,8,1,0,84,12,1,0,47,16,1,0,190,19,1,0,252,22,1,0,233,25,1,0,131,28,1,0,199,30,1,0,180,32,1,0,73,34,1,0,134,35,1,0,104,36,1,0,240,36,1,0,30,37,1,0,46,8,0,0,47,8,0,0,48,8,0,0,49,8,0,0,50,8,0,0,51,8,0,0,52,8,0,0,53,8,0,0,54,8,0,0,55,8,0,0,56,8,0,0,57,8,0,0,58,8,0,0,59,8,0,0,60,8,0,0,61,8,0,0,62,8,0,0,62,8,0,0,62,8,0,0,62,8,0,0,62,8,0,0,62,8,0,0,62,8,0,0,62,8,0,0,62,8,0,0,62,8,0,0,62,8,0,0,62,8,0,0,62,8,0,0,62,8,0,0,62,8,0,0,62,8,0,0,63,8,0,0,64,8,0,0,65,8,0,0,66,8,0,0,67,8,0,0,68,8,0,0,69,8,0,0,70,8,0,0,71,8,0,0,72,8,0,0,73,8,0,0,74,8,0,0,75,8,0,0,76,8,0,0,77,8,0,0,78,8,0,0,79,8,0,0,80,8,0,0,81,8,0,0,82,8,0,0,83,8,0,0,84,8,0,0,85,8,0,0,86,8,0,0,87,8,0,0,88,8,0,0,89,8,0,0,90,8,0,0,91,8,0,0,92,8,0,0,93,8,0,0,94,8,0,0,95,8,0,0,96,8,0,0,97,8,0,0,98,8,0,0,99,8,0,0,100,8,0,0,101,8,0,0,102,8,0,0,103,8,0,0,104,8,0,0,105,8,0,0,0,0,0,0,95,8,0,0,96,8,0,0,106,8,0,0,98,8,0,0,99,8,0,0,100,8,0,0,101,8,0,0,102,8,0,0,103,8,0,0,104,8,0,0,105,8,0,0,0,0,0,0,107,8,0,0,96,8,0,0,108,8,0,0,98,8,0,0,99,8,0,0,109,8,0,0,101,8,0,0,110,8,0,0,111,8,0,0,104,8,0,0,105,8,0,0,112,8,0,0,95,8,0,0,96,8,0,0,113,8,0,0,98,8,0,0,99,8,0,0,100,8,0,0,101,8,0,0,102,8,0,0,103,8,0,0,104,8,0,0,105,8,0,0,112,8,0,0,95,8,0,0,96,8,0,0,113,8,0,0,98,8,0,0,99,8,0,0,100,8,0,0,101,8,0,0,102,8,0,0,103,8,0,0,104,8,0,0,105,8,0,0,112,8,0,0,1,0,0,0,0,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,17,0,0,0,18,0,0,0,0,0,0,0,19,0,0,0,20,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,21,0,0,0,1,0,0,0,22,0,0,0,23,0,0,0,0,0,0,0,24,0,0,0,25,0,0,0,2,0,0,0,26,0,0,0,4,151,0,0,104,151,0,0,0,0,0,0,39,252,7,0,162,145,19,0,44,228,7,0,116,228,7,0,153,228,7,0,244,150,0,0,114,8,0,0,115,8,0,0,116,8,0,0,117,8,0,0,118,8,0,0,119,8,0,0,120,8,0,0,121,8,0,0,122,8,0,0,123,8,0,0,124,8,0,0,125,8,0,0,126,8,0,0,127,8,0,0,128,8,0,0,129,8,0,0,130,8,0,0,131,8,0,0,132,8,0,0,133,8,0,0,79,228,7,0,83,228,7,0,116,228,7,0,153,228,7,0,88,151,0,0,134,8,0,0,135,8,0,0,136,8,0,0,137,8,0,0,138,8,0,0,139,8,0,0,140,8,0,0,141,8,0,0,142,8,0,0,143,8,0,0,144,8,0,0,145,8,0,0,146,8,0,0,147,8,0,0,148,8,0,0,149,8,0,0,150,8,0,0,151,8,0,0,152,8,0,0,153,8,0,0,185,241,7,0,55,235,7,0,63,235,7,0,72,235,7,0,249,248,7,0,40,248,7,0,106,238,7,0,63,235,7,0,79,235,7,0,86,235,7,0,92,235,7,0,207,251,7,0,98,235,7,0,98,251,7,0,14,242,7,0,109,235,7,0,114,235,7,0,120,235,7,0,126,235,7,0,132,235,7,0,138,235,7,0,144,235,7,0,150,235,7,0,156,235,7,0,162,235,7,0,167,235,7,0,172,235,7,0,181,235,7,0,77,238,7,0,81,238,7,0,87,238,7,0,90,238,7,0,95,238,7,0,160,245,7,0,189,127,9,0,102,238,7,0,106,238,7,0,115,238,7,0,118,238,7,0,121,238,7,0,185,241,7,0,127,238,7,0,131,238,7,0,134,238,7,0,59,243,7,0,141,238,7,0,184,127,9,0,146,238,7,0,152,238,7,0,158,238,7,0,161,238,7,0,165,238,7,0,168,238,7,0,171,238,7,0,174,238,7,0,177,238,7,0,186,238,7,0,193,238,7,0,202,238,7,0,0,0,0,0,211,239,7,0,154,8,0,0,236,239,7,0,155,8,0,0,41,240,7,0,156,8,0,0,48,240,7,0,157,8,0,0,54,240,7,0,158,8,0,0,37,243,7,0,159,8,0,0,7,244,7,0,160,8,0,0,164,240,7,0,161,8,0,0,173,240,7,0,162,8,0,0,249,240,7,0,163,8,0,0,4,241,7,0,164,8,0,0,9,241,7,0,165,8,0,0,15,241,7,0,166,8,0,0,73,241,7,0,167,8,0,0,82,241,7,0,168,8,0,0,89,241,7,0,169,8,0,0,96,241,7,0,170,8,0,0,31,244,7,0,171,8,0,0,108,244,7,0,172,8,0,0,158,241,7,0,173,8,0,0,21,241,7,0,174,8,0,0,57,246,7,0,175,8,0,0,196,241,7,0,176,8,0,0,230,241,7,0,177,8,0,0,0,0,0,0,0,0,0,0,3,240,7,0,8,240,7,0,251,239,7,0,121,25,9,0,16,240,7,0,21,240,7,0,30,240,7,0,0,0,0,0,0,0,0,0,1,0,0,0,2,0,0,0,3,0,0,0,5,0,0,0,6,0,0,0,7,0,0,0,66,242,7,0,178,8,0,0,95,242,7,0,179,8,0,0,150,242,7,0,180,8,0,0,234,242,7,0,181,8,0,0,241,242,7,0,182,8,0,0,246,242,7,0,183,8,0,0,0,0,0,0,0,0,0,0,150,242,7,0,158,242,7,0,168,242,7,0,175,242,7,0,252,242,7,0,184,8,0,0,37,243,7,0,185,8,0,0,45,243,7,0,186,8,0,0,92,243,7,0,187,8,0,0,223,243,7,0,188,8,0,0,251,243,7,0,189,8,0,0,7,244,7,0,190,8,0,0,20,244,7,0,191,8,0,0,31,244,7,0,192,8,0,0,91,244,7,0,193,8,0,0,99,244,7,0,194,8,0,0,108,244,7,0,195,8,0,0,143,244,7,0,196,8,0,0,154,244,7,0,197,8,0,0,0,0,0,0,0,0,0,0,54,243,7,0,59,243,7,0,177,245,7,0,121,25,9,0,66,243,7,0,11,245,7,0,198,8,0,0,54,245,7,0,199,8,0,0,60,245,7,0,200,8,0,0,89,245,7,0,201,8,0,0,147,245,7,0,202,8,0,0,164,245,7,0,203,8,0,0,182,245,7,0,204,8,0,0,98,251,7,0,205,8,0,0,194,245,7,0,206,8,0,0,0,0,0,0,0,0,0,0,156,245,7,0,152,245,7,0,160,245,7,0,0,0,0,0,0,0,0,0,1,0,0,0,2,0,0,0,25,18,8,0,172,245,7,0,177,245,7,0,0,0,0,0,2,0,0,0,0,0,0,0,1,0,0,0,11,245,7,0,198,8,0,0,54,245,7,0,207,8,0,0,10,246,7,0,208,8,0,0,60,245,7,0,209,8,0,0,16,246,7,0,210,8,0,0,3,246,7,0,211,8,0,0,21,246,7,0,212,8,0,0,89,245,7,0,213,8,0,0,49,246,7,0,214,8,0,0,57,246,7,0,215,8,0,0,182,245,7,0,216,8,0,0,0,0,0,0,0,0,0,0,131,246,7,0,217,8,0,0,135,246,7,0,218,8,0,0,140,246,7,0,219,8,0,0,145,246,7,0,220,8,0,0,151,246,7,0,221,8,0,0,156,246,7,0,222,8,0,0,161,246,7,0,223,8,0,0,166,246,7,0,224,8,0,0,170,246,7,0,225,8,0,0,174,246,7,0,226,8,0,0,178,246,7,0,227,8,0,0,184,246,7,0,228,8,0,0,189,246,7,0,229,8,0,0,195,246,7,0,230,8,0,0,201,246,7,0,231,8,0,0,207,246,7,0,232,8,0,0,211,246,7,0,233,8,0,0,121,4,8,0,234,8,0,0,215,246,7,0,235,8,0,0,220,246,7,0,236,8,0,0,224,246,7,0,237,8,0,0,228,246,7,0,238,8,0,0,23,247,7,0,239,8,0,0,34,247,7,0,240,8,0,0,39,247,7,0,241,8,0,0,43,247,7,0,242,8,0,0,48,247,7,0,243,8,0,0,53,247,7,0,244,8,0,0,0,0,0,0,0,0,0,0,72,247,7,0,245,8,0,0,78,247,7,0,246,8,0,0,129,247,7,0,247,8,0,0,138,247,7,0,248,8,0,0,146,247,7,0,249,8,0,0,151,247,7,0,250,8,0,0,178,248,7,0,251,8,0,0,165,247,7,0,252,8,0,0,172,247,7,0,253,8,0,0,213,247,7,0,254,8,0,0,251,247,7,0,255,8,0,0,0,0,0,0,0,0,0,0,254,13,8,0,182,247,7,0,190,247,7,0,196,247,7,0,205,247,7,0,213,247,7,0,0,0,0,0,6,0,0,0,3,0,0,0,0,0,0,0,4,0,0,0,1,0,0,0,2,0,0,0,46,248,7,0,0,9,0,0,106,248,7,0,1,9,0,0,114,248,7,0,2,9,0,0,123,248,7,0,3,9,0,0,128,248,7,0,4,9,0,0,133,248,7,0,5,9,0,0,178,248,7,0,6,9,0,0,185,248,7,0,7,9,0,0,209,248,7,0,8,9,0,0,0,0,0,0,0,0,0,0,0,249,7,0,9,9,0,0,27,249,7,0,10,9,0,0,46,249,7,0,11,9,0,0,81,249,7,0,12,9,0,0,79,30,9,0,13,9,0,0,179,250,7,0,14,9,0,0,231,250,7,0,15,9,0,0,238,250,7,0,16,9,0,0,51,251,7,0,17,9,0,0,55,251,7,0,18,9,0,0,61,251,7,0,19,9,0,0,67,251,7,0,20,9,0,0,71,251,7,0,21,9,0,0,79,251,7,0,22,9,0,0,83,251,7,0,23,9,0,0,0,0,0,0,0,0,0,0,111,251,7,0,24,9,0,0,200,251,7,0,25,9,0,0,0,0,0,0,0,0,0,0,200,253,7,0,26,9,0,0,41,254,7,0,27,9,0,0,0,0,0,0,0,0,0,0,128,0,0,0,32,0,0,0,1,0,0,0,68,172,0,0,16,0,0,0,0,2,0,0,64,0,0,0,6,0,0,0,14,0,0,0,1,0,0,0,100,0,0,0,12,0,0,0,20,0,0,0,3,0,0,0,2,0,0,0,16,0,0,0,68,172,0,0,1,0,0,0,128,0,0,0,32,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,96,0,0,0,96,0,0,0,96,0,0,0,112,0,0,0,128,0,0,0,136,0,0,0,144,0,0,0,152,0,0,0,160,0,0,0,164,0,0,0,168,0,0,0,176,0,0,0,180,0,0,0,184,0,0,0,188,0,0,0,192,0,0,0,96,0,0,0,96,0,0,0,96,0,0,0,96,0,0,0,104,0,0,0,112,0,0,0,120,0,0,0,128,0,0,0,132,0,0,0,136,0,0,0,140,0,0,0,144,0,0,0,148,0,0,0,152,0,0,0,156,0,0,0,160,0,0,0,0,0,1,0,237,0,1,0,218,1,1,0,201,2,1,0,185,3,1,0,169,4,1,0,155,5,1,0,141,6,1,0,128,7,1,0,116,8,1,0,105,9,1,0,94,10,1,0,85,11,1,0,77,12,1,0,69,13,1,0,62,14,1,0,56,15,1,0,52,16,1,0,48,17,1,0,45,18,1,0,42,19,1,0,41,20,1,0,41,21,1,0,41,22,1,0,43,23,1,0,45,24,1,0,49,25,1,0,53,26,1,0,59,27,1,0,65,28,1,0,72,29,1,0,80,30,1,0,89,31,1,0,99,32,1,0,110,33,1,0,122,34,1,0,135,35,1,0,149,36,1,0,164,37,1,0,180,38,1,0,197,39,1,0,215,40,1,0,233,41,1,0,253,42,1,0,18,44,1,0,40,45,1,0,63,46,1,0,87,47,1,0,111,48,1,0,137,49,1,0,164,50,1,0,192,51,1,0,221,52,1,0,251,53,1,0,26,55,1,0,58,56,1,0,91,57,1,0,125,58,1,0,160,59,1,0,197,60,1,0,234,61,1,0,16,63,1,0,56,64,1,0,96,65,1,0,138,66,1,0,180,67,1,0,224,68,1,0,13,70,1,0,59,71,1,0,106,72,1,0,154,73,1,0,203,74,1,0,253,75,1,0,49,77,1,0,101,78,1,0,155,79,1,0,209,80,1,0,9,82,1,0,66,83,1,0,124,84,1,0,184,85,1,0,244,86,1,0,49,88,1,0,112,89,1,0,176,90,1,0,241,91,1,0,51,93,1,0,118,94,1,0,187,95,1,0,1,97,1,0,71,98,1,0,143,99,1,0,217,100,1,0,35,102,1,0,111,103,1,0,187,104,1,0,9,106,1,0,89,107,1,0,169,108,1,0,251,109,1,0,77,111,1,0,162,112,1,0,247,113,1,0,77,115,1,0,165,116,1,0,254,117,1,0,88,119,1,0,180,120,1,0,17,122,1,0,111,123,1,0,206,124,1,0,47,126,1,0,145,127,1,0,244,128,1,0,88,130,1,0,190,131,1,0,37,133,1,0,141,134,1,0,247,135,1,0,98,137,1,0,206,138,1,0,59,140,1,0,170,141,1,0,26,143,1,0,140,144,1,0,255,145,1,0,115,147,1,0,233,148,1,0,95,150,1,0,216,151,1,0,81,153,1,0,204,154,1,0,73,156,1,0,198,157,1,0,69,159,1,0,198,160,1,0,72,162,1,0,203,163,1,0,80,165,1,0,214,166,1,0,93,168,1,0,230,169,1,0,113,171,1,0,252,172,1,0,137,174,1,0,24,176,1,0,168,177,1,0,58,179,1,0,205,180,1,0,97,182,1,0,247,183,1,0,142,185,1,0,39,187,1,0,193,188,1,0,93,190,1,0,250,191,1,0,153,193,1,0,58,195,1,0,219,196,1,0,127,198,1,0,35,200,1,0,202,201,1,0,114,203,1,0,27,205,1,0,198,206,1,0,114,208,1,0,32,210,1,0,208,211,1,0,129,213,1,0,52,215,1,0,232,216,1,0,158,218,1,0,85,220,1,0,14,222,1,0,201,223,1,0,133,225,1,0,67,227,1,0,2,229,1,0,195,230,1,0,134,232,1,0,74,234,1,0,16,236,1,0,216,237,1,0,161,239,1,0,108,241,1,0,57,243,1,0,7,245,1,0,215,246,1,0,168,248,1,0,124,250,1,0,81,252,1,0,39,254,1,0,0,0,2,0,218,1,2,0,181,3,2,0,147,5,2,0,114,7,2,0,83,9,2,0,54,11,2,0,26,13,2,0,0,15,2,0,232,16,2,0,210,18,2,0,189,20,2,0,171,22,2,0,154,24,2,0,138,26,2,0,125,28,2,0,113,30,2,0,104,32,2,0,96,34,2,0,90,36,2,0,85,38,2,0,83,40,2,0,82,42,2,0,83,44,2,0,87,46,2,0,91,48,2,0,98,50,2,0,107,52,2,0,118,54,2,0,130,56,2,0,144,58,2,0,161,60,2,0,179,62,2,0,199,64,2,0,221,66,2,0,245,68,2,0,15,71,2,0,43,73,2,0,72,75,2,0,104,77,2,0,138,79,2,0,174,81,2,0,211,83,2,0,251,85,2,0,37,88,2,0,80,90,2,0,126,92,2,0,174,94,2,0,223,96,2,0,19,99,2,0,73,101,2,0,129,103,2,0,186,105,2,0,246,107,2,0,52,110,2,0,116,112,2,0,183,114,2,0,251,116,2,0,65,119,2,0,138,121,2,0,212,123,2,0,33,126,2,0,112,128,2,0,193,130,2,0,0,0,1,0,19,255,0,0,40,254,0,0,62,253,0,0,84,252,0,0,107,251,0,0,131,250,0,0,156,249,0,0,182,248,0,0,208,247,0,0,236,246,0,0,8,246,0,0,37,245,0,0,67,244,0,0,97,243,0,0,129,242,0,0,161,241,0,0,194,240,0,0,228,239,0,0,7,239,0,0,42,238,0,0,79,237,0,0,116,236,0,0,154,235,0,0,192,234,0,0,232,233,0,0,16,233,0,0,57,232,0,0,99,231,0,0,141,230,0,0,185,229,0,0,229,228,0,0,17,228,0,0,63,227,0,0,109,226,0,0,157,225,0,0,204,224,0,0,253,223,0,0,46,223,0,0,96,222,0,0,147,221,0,0,199,220,0,0,251,219,0,0,48,219,0,0,102,218,0,0,157,217,0,0,212,216,0,0,12,216,0,0,68,215,0,0,126,214,0,0,184,213,0,0,243,212,0,0,46,212,0,0,107,211,0,0,168,210,0,0,229,209,0,0,36,209,0,0,99,208,0,0,162,207,0,0,227,206,0,0,36,206,0,0,102,205,0,0,168,204,0,0,236,203,0,0,47,203,0,0,116,202,0,0,185,201,0,0,255,200,0,0,70,200,0,0,141,199,0,0,213,198,0,0,29,198,0,0,103,197,0,0,177,196,0,0,251,195,0,0,70,195,0,0,146,194,0,0,223,193,0,0,44,193,0,0,122,192,0,0,200,191,0,0,23,191,0,0,103,190,0,0,183,189,0,0,8,189,0,0,90,188,0,0,172,187,0,0,255,186,0,0,82,186,0,0,166,185,0,0,251,184,0,0,81,184,0,0,166,183,0,0,253,182,0,0,84,182,0,0,172,181,0,0,4,181,0,0,93,180,0,0,183,179,0,0,17,179,0,0,108,178,0,0,199,177,0,0,35,177,0,0,128,176,0,0,221,175,0,0,59,175,0,0,153,174,0,0,248,173,0,0,88,173,0,0,184,172,0,0,24,172,0,0,122,171,0,0,220,170,0,0,62,170,0,0,161,169,0,0,4,169,0,0,104,168,0,0,205,167,0,0,50,167,0,0,152,166,0,0,254,165,0,0,101,165,0,0,205,164,0,0,53,164,0,0,157,163,0,0,6,163,0,0,112,162,0,0,218,161,0,0,69,161,0,0,176,160,0,0,28,160,0,0,136,159,0,0,245,158,0,0,98,158,0,0,208,157,0,0,62,157,0,0,173,156,0,0,29,156,0,0,141,155,0,0,253,154,0,0,110,154,0,0,224,153,0,0,82,153,0,0,196,152,0,0,55,152,0,0,171,151,0,0,31,151,0,0,148,150,0,0,9,150,0,0,126,149,0,0,244,148,0,0,107,148,0,0,226,147,0,0,90,147,0,0,210,146,0,0,74,146,0,0,195,145,0,0,61,145,0,0,183,144,0,0,49,144,0,0,172,143,0,0,40,143,0,0,164,142,0,0,32,142,0,0,157,141,0,0,26,141,0,0,152,140,0,0,22,140,0,0,149,139,0,0,20,139,0,0,148,138,0,0,20,138,0,0,149,137,0,0,22,137,0,0,152,136,0,0,26,136,0,0,156,135,0,0,31,135,0,0,162,134,0,0,38,134,0,0,170,133,0,0,47,133,0,0,180,132,0,0,58,132,0,0,192,131,0,0,70,131,0,0,205,130,0,0,84,130,0,0,220,129,0,0,100,129,0,0,237,128,0,0,118,128,0,0,0,128,0,0,137,127,0,0,20,127,0,0,159,126,0,0,42,126,0,0,181,125,0,0,65,125,0,0,206,124,0,0,91,124,0,0,232,123,0,0,118,123,0,0,4,123,0,0,146,122,0,0,33,122,0,0,176,121,0,0,64,121,0,0,208,120,0,0,97,120,0,0,242,119,0,0,131,119,0,0,21,119,0,0,167,118,0,0,58,118,0,0,205,117,0,0,96,117,0,0,244,116,0,0,136,116,0,0,28,116,0,0,177,115,0,0,70,115,0,0,220,114,0,0,114,114,0,0,8,114,0,0,159,113,0,0,54,113,0,0,206,112,0,0,102,112,0,0,254,111,0,0,151,111,0,0,48,111,0,0,201,110,0,0,99,110,0,0,253,109,0,0,152,109,0,0,51,109,0,0,206,108,0,0,106,108,0,0,6,108,0,0,162,107,0,0,63,107,0,0,220,106,0,0,121,106,0,0,23,106,0,0,181,105,0,0,84,105,0,0,242,104,0,0,146,104,0,0,49,104,0,0,209,103,0,0,113,103,0,0,18,103,0,0,179,102,0,0,84,102,0,0,246,101,0,0,255,255,0,0,197,255,0,0,138,255,0,0,79,255,0,0,20,255,0,0,217,254,0,0,158,254,0,0,99,254,0,0,41,254,0,0,238,253,0,0,179,253,0,0,120,253,0,0,62,253,0,0,4,253,0,0,201,252,0,0,133,252,0,0,0,0,1,0,59,0,1,0,118,0,1,0,178,0,1,0,237,0,1,0,40,1,1,0,100,1,1,0,159,1,1,0,219,1,1,0,23,2,1,0,82,2,1,0,142,2,1,0,202,2,1,0,5,3,1,0,65,3,1,0,125,3,1,0,192,42,8,0,221,40,8,0,250,38,8,0,24,37,8,0,55,35,8,0,86,33,8,0,117,31,8,0,149,29,8,0,181,27,8,0,214,25,8,0,247,23,8,0,24,22,8,0,58,20,8,0,92,18,8,0,127,16,8,0,162,14,8,0,198,12,8,0,234,10,8,0,15,9,8,0,51,7,8,0,89,5,8,0,127,3,8,0,165,1,8,0,204,255,7,0,243,253,7,0,26,252,7,0,66,250,7,0,106,248,7,0,147,246,7,0,188,244,7,0,230,242,7,0,16,241,7,0,58,239,7,0,101,237,7,0,145,235,7,0,188,233,7,0,233,231,7,0,21,230,7,0,66,228,7,0,112,226,7,0,157,224,7,0,204,222,7,0,250,220,7,0,42,219,7,0,89,217,7,0,137,215,7,0,185,213,7,0,234,211,7,0,27,210,7,0,77,208,7,0,127,206,7,0,178,204,7,0,228,202,7,0,24,201,7,0,75,199,7,0,128,197,7,0,180,195,7,0,233,193,7,0,30,192,7,0,84,190,7,0,138,188,7,0,193,186,7,0,248,184,7,0,47,183,7,0,103,181,7,0,159,179,7,0,216,177,7,0,17,176,7,0,75,174,7,0,133,172,7,0,191,170,7,0,250,168,7,0,53,167,7,0,112,165,7,0,172,163,7,0,233,161,7,0,37,160,7,0,98,158,7,0,160,156,7,0,222,154,7,0,28,153,7,0,91,151,7,0,154,149,7,0,218,147,7,0,26,146,7,0,90,144,7,0,155,142,7,0,220,140,7,0,30,139,7,0,96,137,7,0,162,135,7,0,229,133,7,0,41,132,7,0,108,130,7,0,176,128,7,0,245,126,7,0,57,125,7,0,127,123,7,0,196,121,7,0,10,120,7,0,81,118,7,0,152,116,7,0,223,114,7,0,38,113,7,0,110,111,7,0,183,109,7,0,0,108,7,0,73,106,7,0,146,104,7,0,220,102,7,0,39,101,7,0,114,99,7,0,189,97,7,0,8,96,7,0,84,94,7,0,161,92,7,0,237,90,7,0,59,89,7,0,136,87,7,0,214,85,7,0,36,84,7,0,115,82,7,0,194,80,7,0,18,79,7,0,98,77,7,0,178,75,7,0,3,74,7,0,84,72,7,0,165,70,7,0,247,68,7,0,73,67,7,0,156,65,7,0,239,63,7,0,66,62,7,0,150,60,7,0,234,58,7,0,63,57,7,0,147,55,7,0,233,53,7,0,62,52,7,0,149,50,7,0,235,48,7,0,66,47,7,0,153,45,7,0,241,43,7,0,73,42,7,0,161,40,7,0,250,38,7,0,83,37,7,0,172,35,7,0,6,34,7,0,97,32,7,0,187,30,7,0,22,29,7,0,114,27,7,0,206,25,7,0,42,24,7,0,134,22,7,0,227,20,7,0,65,19,7,0,158,17,7,0,252,15,7,0,91,14,7,0,186,12,7,0,25,11,7,0,121,9,7,0,216,7,7,0,57,6,7,0,154,4,7,0,251,2,7,0,92,1,7,0,190,255,6,0,32,254,6,0,131,252,6,0,230,250,6,0,73,249,6,0,173,247,6,0,17,246,6,0,117,244,6,0,218,242,6,0,63,241,6,0,165,239,6,0,11,238,6,0,113,236,6,0,216,234,6,0,63,233,6,0,166,231,6,0,14,230,6,0,118,228,6,0,222,226,6,0,71,225,6,0,177,223,6,0,26,222,6,0,132,220,6,0,238,218,6,0,89,217,6,0,196,215,6,0,48,214,6,0,156,212,6,0,8,211,6,0,116,209,6,0,225,207,6,0,78,206,6,0,188,204,6,0,42,203,6,0,152,201,6,0,7,200,6,0,118,198,6,0,230,196,6,0,85,195,6,0,197,193,6,0,54,192,6,0,167,190,6,0,24,189,6,0,138,187,6,0,252,185,6,0,110,184,6,0,225,182,6,0,84,181,6,0,199,179,6,0,59,178,6,0,175,176,6,0,35,175,6,0,152,173,6,0,13,172,6,0,131,170,6,0,249,168,6,0,111,167,6,0,230,165,6,0,93,164,6,0,212,162,6,0,76,161,6,0,196,159,6,0,60,158,6,0,181,156,6,0,46,155,6,0,167,153,6,0,33,152,6,0,155,150,6,0,22,149,6,0,144,147,6,0,12,146,6,0,135,144,6,0,3,143,6,0,127,141,6,0,252,139,6,0,121,138,6,0,246,136,6,0,116,135,6,0,242,133,6,0,112,132,6,0,239,130,6,0,110,129,6,0,237,127,6,0,109,126,6,0,237,124,6,0,109,123,6,0,238,121,6,0,111,120,6,0,241,118,6,0,115,117,6,0,245,115,6,0,119,114,6,0,250,112,6,0,125,111,6,0,1,110,6,0,133,108,6,0,9,107,6,0,141,105,6,0,18,104,6,0,151,102,6,0,29,101,6,0,163,99,6,0,41,98,6,0,176,96,6,0,55,95,6,0,190,93,6,0,70,92,6,0,206,90,6,0,86,89,6,0,223,87,6,0,104,86,6,0,241,84,6,0,122,83,6,0,4,82,6,0,143,80,6,0,25,79,6,0,164,77,6,0,48,76,6,0,187,74,6,0,71,73,6,0,212,71,6,0,96,70,6,0,238,68,6,0,123,67,6,0,9,66,6,0,151,64,6,0,37,63,6,0,180,61,6,0,67,60,6,0,210,58,6,0,98,57,6,0,242,55,6,0,130,54,6,0,19,53,6,0,164,51,6,0,53,50,6,0,199,48,6,0,89,47,6,0,235,45,6,0,126,44,6,0,17,43,6,0,164,41,6,0,56,40,6,0,204,38,6,0,96,37,6,0,245,35,6,0,138,34,6,0,31,33,6,0,181,31,6,0,74,30,6,0,225,28,6,0,119,27,6,0,14,26,6,0,165,24,6,0,61,23,6,0,213,21,6,0,109,20,6,0,6,19,6,0,159,17,6,0,56,16,6,0,209,14,6,0,107,13,6,0,5,12,6,0,160,10,6,0,59,9,6,0,214,7,6,0,113,6,6,0,13,5,6,0,169,3,6,0,70,2,6,0,226,0,6,0,127,255,5,0,29,254,5,0,186,252,5,0,88,251,5,0,247,249,5,0,149,248,5,0,52,247,5,0,212,245,5,0,115,244,5,0,19,243,5,0,180,241,5,0,84,240,5,0,245,238,5,0,150,237,5,0,56,236,5,0,218,234,5,0,124,233,5,0,30,232,5,0,193,230,5,0,100,229,5,0,8,228,5,0,171,226,5,0,79,225,5,0,244,223,5,0,152,222,5,0,61,221,5,0,227,219,5,0,136,218,5,0,46,217,5,0,212,215,5,0,123,214,5,0,34,213,5,0,201,211,5,0,112,210,5,0,24,209,5,0,192,207,5,0,105,206,5,0,17,205,5,0,186,203,5,0,100,202,5,0,13,201,5,0,183,199,5,0,98,198,5,0,12,197,5,0,183,195,5,0,98,194,5,0,14,193,5,0,186,191,5,0,102,190,5,0,18,189,5,0,191,187,5,0,108,186,5,0,25,185,5,0,199,183,5,0,117,182,5,0,35,181,5,0,210,179,5,0,129,178,5,0,48,177,5,0,223,175,5,0,143,174,5,0,63,173,5,0,239,171,5,0,160,170,5,0,81,169,5,0,2,168,5,0,180,166,5,0,102,165,5,0,24,164,5,0,203,162,5,0,125,161,5,0,48,160,5,0,228,158,5,0,151,157,5,0,75,156,5,0,0,155,5,0,180,153,5,0,105,152,5,0,30,151,5,0,212,149,5,0,138,148,5,0,64,147,5,0,246,145,5,0,173,144,5,0,100,143,5,0,27,142,5,0,211,140,5,0,139,139,5,0,67,138,5,0,251,136,5,0,180,135,5,0,109,134,5,0,38,133,5,0,224,131,5,0,154,130,5,0,84,129,5,0,15,128,5,0,201,126,5,0,133,125,5,0,64,124,5,0,252,122,5,0,184,121,5,0,116,120,5,0,49,119,5,0,237,117,5,0,171,116,5,0,104,115,5,0,38,114,5,0,228,112,5,0,162,111,5,0,97,110,5,0,32,109,5,0,223,107,5,0,158,106,5,0,94,105,5,0,30,104,5,0,222,102,5,0,159,101,5,0,96,100,5,0,33,99,5,0,227,97,5,0,164,96,5,0,103,95,5,0,41,94,5,0,236,92,5,0,174,91,5,0,114,90,5,0,53,89,5,0,249,87,5,0,189,86,5,0,129,85,5,0,70,84,5,0,11,83,5,0,208,81,5,0,150,80,5,0,91,79,5,0,34,78,5,0,232,76,5,0,175,75,5,0,117,74,5,0,61,73,5,0,4,72,5,0,204,70,5,0,148,69,5,0,92,68,5,0,37,67,5,0,238,65,5,0,183,64,5,0,128,63,5,0,74,62,5,0,20,61,5,0,222,59,5,0,169,58,5,0,116,57,5,0,63,56,5,0,10,55,5,0,214,53,5,0,162,52,5,0,110,51,5,0,59,50,5,0,7,49,5,0,212,47,5,0,162,46,5,0,111,45,5,0,61,44,5,0,12,43,5,0,218,41,5,0,169,40,5,0,120,39,5,0,71,38,5,0,23,37,5,0,230,35,5,0,182,34,5,0,135,33,5,0,87,32,5,0,40,31,5,0,250,29,5,0,203,28,5,0,157,27,5,0,111,26,5,0,65,25,5,0,20,24,5,0,230,22,5,0,186,21,5,0,141,20,5,0,97,19,5,0,53,18,5,0,9,17,5,0,221,15,5,0,178,14,5,0,135,13,5,0,92,12,5,0,50,11,5,0,7,10,5,0,222,8,5,0,180,7,5,0,139,6,5,0,97,5,5,0,57,4,5,0,16,3,5,0,232,1,5,0,192,0,5,0,152,255,4,0,112,254,4,0,73,253,4,0,34,252,4,0,251,250,4,0,213,249,4,0,175,248,4,0,137,247,4,0,99,246,4,0,62,245,4,0,25,244,4,0,244,242,4,0,207,241,4,0,171,240,4,0,135,239,4,0,99,238,4,0,63,237,4,0,28,236,4,0,249,234,4,0,214,233,4,0,180,232,4,0,146,231,4,0,112,230,4,0,78,229,4,0,45,228,4,0,11,227,4,0,234,225,4,0,202,224,4,0,169,223,4,0,137,222,4,0,105,221,4,0,74,220,4,0,42,219,4,0,11,218,4,0,236,216,4,0,206,215,4,0,175,214,4,0,145,213,4,0,116,212,4,0,86,211,4,0,57,210,4,0,28,209,4,0,255,207,4,0,226,206,4,0,198,205,4,0,170,204,4,0,142,203,4,0,115,202,4,0,88,201,4,0,61,200,4,0,34,199,4,0,7,198,4,0,237,196,4,0,211,195,4,0,186,194,4,0,160,193,4,0,135,192,4,0,110,191,4,0,85,190,4,0,61,189,4,0,37,188,4,0,13,187,4,0,245,185,4,0,222,184,4,0,198,183,4,0,175,182,4,0,153,181,4,0,130,180,4,0,108,179,4,0,86,178,4,0,65,177,4,0,43,176,4,0,22,175,4,0,1,174,4,0,236,172,4,0,216,171,4,0,196,170,4,0,176,169,4,0,156,168,4,0,137,167,4,0,118,166,4,0,99,165,4,0,80,164,4,0,62,163,4,0,43,162,4,0,25,161,4,0,8,160,4,0,246,158,4,0,229,157,4,0,212,156,4,0,195,155,4,0,179,154,4,0,163,153,4,0,147,152,4,0,131,151,4,0,116,150,4,0,100,149,4,0,85,148,4,0,71,147,4,0,56,146,4,0,42,145,4,0,28,144,4,0,14,143,4,0,1,142,4,0,243,140,4,0,230,139,4,0,217,138,4,0,205,137,4,0,193,136,4,0,180,135,4,0,169,134,4,0,157,133,4,0,146,132,4,0,135,131,4,0,124,130,4,0,113,129,4,0,103,128,4,0,93,127,4,0,83,126,4,0,73,125,4,0,64,124,4,0,54,123,4,0,45,122,4,0,37,121,4,0,28,120,4,0,20,119,4,0,12,118,4,0,4,117,4,0,253,115,4,0,246,114,4,0,239,113,4,0,232,112,4,0,225,111,4,0,219,110,4,0,213,109,4,0,207,108,4,0,201,107,4,0,196,106,4,0,191,105,4,0,186,104,4,0,181,103,4,0,177,102,4,0,173,101,4,0,169,100,4,0,165,99,4,0,161,98,4,0,158,97,4,0,155,96,4,0,152,95,4,0,150,94,4,0,147,93,4,0,145,92,4,0,143,91,4,0,142,90,4,0,140,89,4,0,139,88,4,0,138,87,4,0,138,86,4,0,137,85,4,0,137,84,4,0,137,83,4,0,137,82,4,0,137,81,4,0,138,80,4,0,139,79,4,0,140,78,4,0,142,77,4,0,143,76,4,0,145,75,4,0,147,74,4,0,149,73,4,0,152,72,4,0,155,71,4,0,158,70,4,0,161,69,4,0,164,68,4,0,168,67,4,0,172,66,4,0,176,65,4,0,180,64,4,0,185,63,4,0,190,62,4,0,195,61,4,0,200,60,4,0,205,59,4,0,211,58,4,0,217,57,4,0,223,56,4,0,229,55,4,0,236,54,4,0,243,53,4,0,250,52,4,0,1,52,4,0,9,51,4,0,16,50,4,0,24,49,4,0,33,48,4,0,41,47,4,0,50,46,4,0,58,45,4,0,68,44,4,0,77,43,4,0,86,42,4,0,96,41,4,0,106,40,4,0,116,39,4,0,127,38,4,0,137,37,4,0,148,36,4,0,159,35,4,0,170,34,4,0,182,33,4,0,194,32,4,0,206,31,4,0,218,30,4,0,230,29,4,0,243,28,4,0,0,28,4,0,13,27,4,0,26,26,4,0,39,25,4,0,53,24,4,0,67,23,4,0,81,22,4,0,28,9,0,0,29,9,0,0,30,9,0,0,31,9,0,0,32,9,0,0,33,9,0,0,34,9,0,0,35,9,0,0,36,9,0,0,37,9,0,0,38,9,0,0,39,9,0,0,40,9,0,0,41,9,0,0,42,9,0,0,43,9,0,0,44,9,0,0,45,9,0,0,46,9,0,0,47,9,0,0,48,9,0,0,49,9,0,0,50,9,0,0,51,9,0,0,52,9,0,0,53,9,0,0,54,9,0,0,55,9,0,0,56,9,0,0,57,9,0,0,58,9,0,0,59,9,0,0,60,9,0,0,61,9,0,0,62,9,0,0,63,9,0,0,64,9,0,0,65,9,0,0,66,9,0,0,67,9,0,0,68,9,0,0,69,9,0,0,70,9,0,0,71,9,0,0,72,9,0,0,73,9,0,0,74,9,0,0,75,9,0,0,76,9,0,0,77,9,0,0,78,9,0,0,79,9,0,0,80,9,0,0,81,9,0,0,82,9,0,0,83,9,0,0,84,9,0,0,85,9,0,0,86,9,0,0,87,9,0,0,88,9,0,0,89,9,0,0,90,9,0,0,91,9,0,0,92,9,0,0,93,9,0,0,30,9,0,0,31,9,0,0,94,9,0,0,95,9,0,0,34,9,0,0,35,9,0,0,36,9,0,0,37,9,0,0,38,9,0,0,39,9,0,0,40,9,0,0,41,9,0,0,42,9,0,0,43,9,0,0,96,9,0,0,97,9,0,0,46,9,0,0,47,9,0,0,98,9,0,0,99,9,0,0,50,9,0,0,51,9,0,0,52,9,0,0,53,9,0,0,54,9,0,0,55,9,0,0,56,9,0,0,57,9,0,0,58,9,0,0,59,9,0,0,60,9,0,0,61,9,0,0,62,9,0,0,63,9,0,0,64,9,0,0,65,9,0,0,66,9,0,0,67,9,0,0,68,9,0,0,69,9,0,0,70,9,0,0,71,9,0,0,72,9,0,0,73,9,0,0,74,9,0,0,75,9,0,0,76,9,0,0,77,9,0,0,78,9,0,0,79,9,0,0,80,9,0,0,81,9,0,0,82,9,0,0,83,9,0,0,84,9,0,0,85,9,0,0,86,9,0,0,87,9,0,0,88,9,0,0,89,9,0,0,90,9,0,0,91,9,0,0,1,0,0,0,3,0,0,0,7,0,0,0,15,0,0,0,30,0,0,0,60,0,0,0,120,0,0,0,240,0,0,0,240,1,0,0,240,3,0,0,240,7,0,0,240,15,0,0,240,31,0,0,240,63,0,0,240,127,0,0,240,255,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,3,0,0,0,2,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,3,0,0,0,7,0,0,0,15,0,0,0,30,0,0,0,60,0,0,0,120,0,0,0,248,0,0,0,3,0,0,0,3,0,0,0,3,0,0,0,3,0,0,0,2,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,7,0,0,0,0,0,0,0,3,0,0,0,7,0,0,0,0,0,0,0,4,0,0,0,7,0,0,0,10,0,0,0,0,0,0,0,3,0,0,0,7,0,0,0,10,0,0,0,0,0,0,0,4,0,0,0,7,0,0,0,11,0,0,0,0,0,0,0,4,0,0,0,7,0,0,0,11,0,0,0,0,0,0,0,4,0,0,0,7,0,0,0,9,0,0,0,0,0,0,0,3,0,0,0,7,0,0,0,9,0,0,0,0,0,0,0,4,0,0,0,8,0,0,0,0,0,0,0,4,0,0,0,8,0,0,0,0,0,0,0,4,0,0,0,8,0,0,0,10,0,0,0,0,0,0,0,3,0,0,0,6,0,0,0,0,0,0,0,3,0,0,0,6,0,0,0,9,0,0,0,0,0,0,0,4,0,0,0,7,0,0,0,10,0,0,0,2,0,0,0,0,0,0,0,3,0,0,0,7,0,0,0,10,0,0,0,2,0,0,0,0,0,0,0,4,0,0,0,7,0,0,0,11,0,0,0,2,0,0,0,0,0,0,0,4,0,0,0,7,0,0,0,11,0,0,0,2,0,0,0,0,0,0,0,4,0,0,0,7,0,0,0,10,0,0,0,2,0,0,0,5,0,0,0,0,0,0,0,4,0,0,0,7,0,0,0,10,0,0,0,13,0,0,0,0,0,0,0,5,0,0,0,7,0,0,0,0,0,0,0,2,0,0,0,7,0,0,0,0,0,0,0,5,0,0,0,7,0,0,0,10,0,0,0,0,0,0,0,5,0,0,0,7,0,0,0,10,0,0,0,0,0,0,0,2,0,0,0,7,0,0,0,10,0,0,0,0,0,0,0,5,0,0,0,10,0,0,0,14,0,0,0,19,0,0,0,0,0,0,0,7,0,0,0,0,0,0,0,4,0,0,0,7,0,0,0,10,0,0,0,16,0,0,0,21,0,0,0,150,4,8,0,203,4,8,0,0,5,8,0,53,5,8,0,106,5,8,0,159,5,8,0,212,5,8,0,9,6,8,0,62,6,8,0,115,6,8,0,168,6,8,0,221,6,8,0,18,7,8,0,71,7,8,0,124,7,8,0,0,0,0,0,42,3,8,0,211,7,8,0,237,7,8,0,7,8,8,0,33,8,8,0,59,8,8,0,85,8,8,0,42,3,8,0,211,7,8,0,237,7,8,0,7,8,8,0,33,8,8,0,59,8,8,0,111,8,8,0,137,8,8,0,100,9,0,0,101,9,0,0,102,9,0,0,0,0,0,0,0,16,0,0,0,0,160,1,0,0,144,1,0,0,160,2,0,0,144,2,0,0,160,4,0,0,144,4,0,0,160,8,0,0,160,40,0,0,144,8,0,0,144,40,10,0,32,0,10,0,224,255,0,0,240,255,0,1,0,0,0,2,0,0,0,3,0,0,1,0,32,0,1,0,224,255,2,0,32,0,2,0,224,255,3,0,32,0,3,0,224,255,4,0,32,0,4,0,224,255,5,0,32,0,5,0,224,255,1,0,64,0,1,0,96,0,1,0,192,255,1,0,160,255,6,0,32,0,7,0,32,0,8,0,32,0,9,0,32,0,6,0,224,255,7,0,224,255,8,0,224,255,9,0,224,255,2,0,64,0,3,0,64,0,2,0,96,0,3,0,96,0,2,0,192,255,3,0,192,255,2,0,160,255,3,0,160,255,0,0,240,255,1,0,0,0,2,0,0,0,3,0,0,0,4,0,0,0,5,0,0,0,6,0,0,0,7,0,0,0,0,0,32,0,0,0,224,255,0,0,64,0,0,0,192,255,0,0,96,0,0,0,160,255,0,0,128,0,0,0,128,255,0,0,160,0,0,0,96,255,0,0,192,0,0,0,64,255,0,0,224,0,0,0,0,1,0,0,32,255,0,0,0,255,0,0,32,1,0,0,64,1,0,0,96,1,0,0,128,1,0,0,224,254,0,0,192,254,0,0,160,254,0,0,128,254,0,8,0,0,0,9,0,0,0,10,0,0,0,11,0,0,0,12,0,0,0,13,0,0,0,14,0,0,0,15,0,0,0,4,0,0,0,5,0,0,0,6,0,0,0,7,0,0,103,9,0,0,104,9,0,0,105,9,0,0,106,9,0,0,16,0,0,0,32,0,0,0,8,0,0,0,16,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,3,0,0,0,2,0,0,0,3,0,0,0,3,0,0,0,0,0,0,0,3,0,0,0,0,0,0,0,2,0,0,0,3,0,0,0,1,0,0,0,3,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,3,0,0,0,2,0,0,0,0,0,0,0,2,0,0,0,3,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,2,0,0,0,2,0,0,0,1,0,0,0,2,0,0,0,2,0,0,0,107,9,0,0,108,9,0,0,109,9,0,0,110,9,0,0,52,0,0,0,115,116,105,98,111,9,0,0,112,9,0,0,113,9,0,0,0,0,0,0,114,9,0,0,0,0,0,0,40,0,0,0,108,116,117,111,115,9,0,0,116,9,0,0,117,9,0,0,118,9,0,0,119,9,0,0,120,9,0,0,113,17,9,0,0,0,0,0,121,9,0,0,0,0,0,0,122,9,0,0,1,0,0,0,123,9,0,0,2,0,0,0,124,9,0,0,3,0,0,0,125,9,0,0,4,0,0,0,126,9,0,0,5,0,0,0,127,9,0,0,6,0,0,0,128,9,0,0,7,0,0,0,129,9,0,0,8,0,0,0,1,2,0,0,28,0,0,0,136,12,9,0,0,0,1,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,130,9,0,0,168,0,0,0,44,0,0,0,160,0,0,0,131,9,0,0,132,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,133,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,134,9,0,0,135,9,0,0,136,12,9,0,120,187,0,0,255,18,9,0,162,24,8,0,0,0,0,0,0,0,0,0,136,9,0,0,137,9,0,0,231,31,8,0,1,0,0,0,1,0,0,0,0,0,0,0,2,32,8,0,2,0,0,0,1,0,0,0,0,0,0,0,184,24,8,0,2,0,0,0,1,0,0,0,0,0,0,0,202,24,8,0,2,0,0,0,1,0,0,0,0,0,0,0,222,24,8,0,2,0,0,0,1,0,0,0,0,0,0,0,233,24,8,0,1,0,0,0,1,0,0,0,0,0,0,0,157,12,9,0,1,0,0,0,1,0,0,0,0,0,0,0,140,12,9,0,1,0,0,0,1,0,0,0,0,0,0,0,166,24,8,0,1,0,0,0,1,0,0,0,0,0,0,0,253,24,8,0,1,0,0,0,1,0,0,0,0,0,0,0,7,25,8,0,3,0,0,0,1,0,0,0,0,0,0,0,20,25,8,0,3,0,0,0,1,0,0,0,0,0,0,0,32,25,8,0,1,0,0,0,1,0,0,0,0,0,0,0,49,25,8,0,2,0,0,0,1,0,0,0,0,0,0,0,59,25,8,0,1,0,0,0,1,0,0,0,0,0,0,0,246,31,8,0,1,0,0,0,1,0,0,0,0,0,0,0,69,25,8,0,2,0,0,0,1,0,0,0,0,0,0,0,82,25,8,0,1,0,0,0,1,0,0,0,0,0,0,0,87,25,8,0,1,0,0,0,1,0,0,0,0,0,0,0,105,25,8,0,2,0,0,0,1,0,0,0,0,0,0,0,117,25,8,0,2,0,0,0,1,0,0,0,0,0,0,0,130,25,8,0,1,0,0,0,1,0,0,0,0,0,0,0,138,25,8,0,1,0,0,0,1,0,0,0,0,0,0,0,148,25,8,0,2,0,0,0,1,0,0,0,0,0,0,0,161,25,8,0,2,0,0,0,1,0,0,0,0,0,0,0,171,25,8,0,2,0,0,0,1,0,0,0,0,0,0,0,181,25,8,0,2,0,0,0,1,0,0,0,0,0,0,0,192,25,8,0,1,0,0,0,1,0,0,0,0,0,0,0,27,32,8,0,2,0,0,0,1,0,0,0,0,0,0,0,16,32,8,0,2,0,0,0,1,0,0,0,0,0,0,0,199,25,8,0,2,0,0,0,1,0,0,0,0,0,0,0,210,25,8,0,2,0,0,0,1,0,0,0,0,0,0,0,221,25,8,0,2,0,0,0,1,0,0,0,0,0,0,0,239,25,8,0,2,0,0,0,1,0,0,0,0,0,0,0,5,26,8,0,2,0,0,0,1,0,0,0,0,0,0,0,29,26,8,0,2,0,0,0,1,0,0,0,0,0,0,0,44,26,8,0,2,0,0,0,1,0,0,0,0,0,0,0,56,26,8,0,2,0,0,0,1,0,0,0,0,0,0,0,70,26,8,0,2,0,0,0,1,0,0,0,0,0,0,0,87,26,8,0,2,0,0,0,1,0,0,0,0,0,0,0,101,26,8,0,2,0,0,0,1,0,0,0,0,0,0,0,115,26,8,0,2,0,0,0,1,0,0,0,0,0,0,0,130,26,8,0,2,0,0,0,1,0,0,0,0,0,0,0,145,26,8,0,2,0,0,0,1,0,0,0,0,0,0,0,160,26,8,0,2,0,0,0,1,0,0,0,0,0,0,0,174,26,8,0,2,0,0,0,1,0,0,0,0,0,0,0,188,26,8,0,2,0,0,0,1,0,0,0,0,0,0,0,203,26,8,0,2,0,0,0,1,0,0,0,0,0,0,0,222,26,8,0,2,0,0,0,1,0,0,0,0,0,0,0,243,26,8,0,2,0,0,0,1,0,0,0,0,0,0,0,9,27,8,0,2,0,0,0,1,0,0,0,0,0,0,0,28,27,8,0,2,0,0,0,1,0,0,0,0,0,0,0,44,27,8,0,2,0,0,0,1,0,0,0,0,0,0,0,60,27,8,0,2,0,0,0,1,0,0,0,0,0,0,0,81,27,8,0,2,0,0,0,1,0,0,0,0,0,0,0,99,27,8,0,2,0,0,0,1,0,0,0,0,0,0,0,117,27,8,0,2,0,0,0,1,0,0,0,0,0,0,0,140,27,8,0,2,0,0,0,1,0,0,0,0,0,0,0,164,27,8,0,2,0,0,0,1,0,0,0,0,0,0,0,177,27,8,0,3,0,0,0,1,0,0,0,0,0,0,0,195,27,8,0,3,0,0,0,1,0,0,0,0,0,0,0,211,27,8,0,2,0,0,0,1,0,0,0,0,0,0,0,38,32,8,0,3,0,0,0,1,0,0,0,0,0,0,0,51,32,8,0,3,0,0,0,1,0,0,0,0,0,0,0,217,31,8,0,1,0,0,0,1,0,0,0,0,0,0,0,191,31,8,0,1,0,0,0,1,0,0,0,0,0,0,0,222,27,8,0,2,0,0,0,1,0,0,0,0,0,0,0,237,27,8,0,1,0,0,0,1,0,0,0,0,0,0,0,245,27,8,0,2,0,0,0,1,0,0,0,0,0,0,0,6,28,8,0,2,0,0,0,1,0,0,0,0,0,0,0,24,28,8,0,2,0,0,0,1,0,0,0,0,0,0,0,39,28,8,0,2,0,0,0,1,0,0,0,0,0,0,0,51,28,8,0,2,0,0,0,1,0,0,0,0,0,0,0,63,28,8,0,2,0,0,0,1,0,0,0,0,0,0,0,80,28,8,0,2,0,0,0,1,0,0,0,0,0,0,0,94,28,8,0,2,0,0,0,1,0,0,0,0,0,0,0,108,28,8,0,2,0,0,0,1,0,0,0,0,0,0,0,127,28,8,0,2,0,0,0,1,0,0,0,0,0,0,0,147,28,8,0,3,0,0,0,1,0,0,0,0,0,0,0,205,31,8,0,1,0,0,0,1,0,0,0,0,0,0,0,154,28,8,0,2,0,0,0,1,0,0,0,0,0,0,0,163,28,8,0,2,0,0,0,1,0,0,0,0,0,0,0,185,28,8,0,2,0,0,0,1,0,0,0,0,0,0,0,24,0,0,0,138,9,0,0,139,9,0,0,140,9,0,0,141,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,5,0,0,68,0,0,0,44,30,8,0,0,0,1,0,0,0,2,0,0,0,0,0,142,9,0,0,143,9,0,0,144,9,0,0,24,3,0,0,48,0,0,0,172,0,0,0,145,9,0,0,146,9,0,0,147,9,0,0,148,9,0,0,149,9,0,0,150,9,0,0,151,9,0,0,152,9,0,0,0,0,0,0,153,9,0,0,154,9,0,0,155,9,0,0,255,18,9,0,48,30,8,0,161,17,9,0,120,193,0,0,140,17,9,0,140,193,0,0,129,17,9,0,144,193,0,0,174,12,9,0,152,193,0,0,171,30,8,0,236,193,0,0,57,22,9,0,248,193,0,0,0,0,0,0,0,0,0,0,156,9,0,0,0,0,0,0,157,9,0,0,0,0,0,0,0,0,0,0,158,9,0,0,159,9,0,0,160,9,0,0,161,9,0,0,20,0,0,0,162,9,0,0,163,9,0,0,164,9,0,0,165,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,24,0,0,0,166,9,0,0,167,9,0,0,168,9,0,0,169,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,170,9,0,0,171,9,0,0,172,9,0,0,173,9,0,0,174,9,0,0,4,0,0,0,0,16,0,0,0,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,1,16,0,0,4,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,0,17,0,0,8,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,2,16,0,0,12,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,3,16,0,0,16,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,4,16,0,0,20,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,5,0,0,0,1,17,0,0,24,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,2,17,0,0,28,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,3,17,0,0,32,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,4,17,0,0,36,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,5,17,0,0,40,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,6,17,0,0,44,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,7,0,0,0,7,17,0,0,0,0,0,0,0,0,0,0,175,9,0,0,0,0,0,0,0,0,0,0,1,0,0,0,13,16,0,0,80,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,7,0,0,0,5,16,0,0,0,0,0,0,0,0,0,0,176,9,0,0,0,0,0,0,0,0,0,0,1,0,0,0,8,17,0,0,100,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,15,16,0,0,104,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,16,16,0,0,108,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,17,16,0,0,112,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,7,0,0,0,18,16,0,0,0,0,0,0,0,0,0,0,177,9,0,0,0,0,0,0,0,0,0,0,1,0,0,0,20,17,0,0,124,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,21,17,0,0,128,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,7,0,0,0,30,17,0,0,0,0,0,0,0,0,0,0,178,9,0,0,0,0,0,0,0,0,0,0,1,0,0,0,31,17,0,0,144,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,32,17,0,0,148,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,33,17,0,0,152,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,34,17,0,0,156,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,35,17,0,0,160,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,36,17,0,0,164,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,37,17,0,0,168,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,38,17,0,0,172,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,6,0,0,0,6,32,0,0,4,0,0,0,4,0,0,0,0,0,0,0,14,0,0,0,0,0,0,0,6,0,0,0,7,32,0,0,60,0,0,0,4,0,0,0,0,0,0,0,10,0,0,0,1,0,0,0,6,0,0,0,8,32,0,0,100,0,0,0,4,0,0,0,0,0,0,0,14,0,0,0,2,0,0,0,6,0,0,0,9,32,0,0,156,0,0,0,4,0,0,0,0,0,0,0,10,0,0,0,3,0,0,0,3,0,0,0,9,33,0,0,196,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,10,33,0,0,200,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,11,33,0,0,204,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,10,32,0,0,208,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,11,32,0,0,212,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,6,0,0,0,12,33,0,0,220,0,0,0,4,0,0,0,0,0,0,0,13,0,0,0,216,0,0,0,6,0,0,0,13,33,0,0,16,1,0,0,4,0,0,0,0,0,0,0,13,0,0,0,217,0,0,0,5,0,0,0,14,33,0,0,68,1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,15,33,0,0,72,1,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,16,33,0,0,76,1,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,17,33,0,0,80,1,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,18,33,0,0,84,1,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,19,33,0,0,88,1,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,19,32,0,0,92,1,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,20,32,0,0,96,1,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,21,32,0,0,100,1,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,10,0,0,0,100,0,0,0,232,3,0,0,16,39,0,0,160,134,1,0,64,66,15,0,128,150,152,0,0,225,245,5,0,202,154,59,1,5,0,0,28,0,0,0,154,30,8,0,0,0,1,0,0,0,2,0,0,0,0,0,179,9,0,0,180,9,0,0,181,9,0,0,76,1,0,0,48,0,0,0,172,0,0,0,182,9,0,0,183,9,0,0,184,9,0,0,185,9,0,0,186,9,0,0,187,9,0,0,188,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,189,9,0,0,0,0,0,0,255,18,9,0,160,30,8,0,140,17,9,0,96,200,0,0,161,17,9,0,100,200,0,0,171,30,8,0,120,200,0,0,0,0,0,0,0,0,0,0,190,9,0,0,191,9,0,0,192,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,193,9,0,0,194,9,0,0,195,9,0,0,248,30,8,0,0,0,0,0,6,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,31,8,0,0,0,0,0,3,0,0,0,0,0,0,0,4,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,19,31,8,0,0,0,0,0,2,0,0,0,0,0,0,0,8,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,31,31,8,0,0,0,0,0,5,0,0,0,0,0,0,0,12,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,40,31,8,0,0,0,0,0,5,0,0,0,0,0,0,0,16,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,49,31,8,0,0,0,0,0,2,0,0,0,0,0,0,0,20,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,60,31,8,0,0,0,0,0,2,0,0,0,0,0,0,0,72,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,68,31,8,0,0,0,0,0,2,0,0,0,0,0,0,0,144,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,81,31,8,0,0,0,0,0,2,0,0,0,0,0,0,0,148,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,89,31,8,0,0,0,0,0,2,0,0,0,0,0,0,0,152,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,97,31,8,0,0,0,0,0,2,0,0,0,0,0,0,0,156,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,246,17,9,0,3,0,0,0,5,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,254,17,9,0,3,0,0,0,5,0,0,0,0,0,0,0,4,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,5,18,9,0,3,0,0,0,5,0,0,0,0,0,0,0,8,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,14,18,9,0,3,0,0,0,5,0,0,0,0,0,0,0,12,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,25,18,9,0,3,0,0,0,5,0,0,0,0,0,0,0,16,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,32,18,9,0,3,0,0,0,2,0,0,0,0,0,0,0,20,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,44,18,9,0,3,0,0,0,1,0,0,0,0,0,0,0,24,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,57,18,9,0,3,0,0,0,2,0,0,0,0,0,0,0,26,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,75,18,9,0,3,0,0,0,2,0,0,0,0,0,0,0,28,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,94,18,9,0,2,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,110,18,9,0,1,0,0,0,2,0,0,0,0,0,0,0,212,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,120,18,9,0,1,0,0,0,2,0,0,0,0,0,0,0,213,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,106,31,8,0,1,0,0,0,2,0,0,0,0,0,0,0,244,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,120,31,8,0,1,0,0,0,2,0,0,0,0,0,0,0,248,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,128,31,8,0,1,0,0,0,2,0,0,0,0,0,0,0,240,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,138,31,8,0,1,0,0,0,2,0,0,0,0,0,0,0,196,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,156,31,8,0,1,0,0,0,3,0,0,0,0,0,0,0,200,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,129,18,9,0,1,0,0,0,3,0,0,0,0,0,0,0,204,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,49,16,9,0,4,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,58,16,9,0,4,0,0,0,2,0,0,0,0,0,0,0,4,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,64,16,9,0,4,0,0,0,2,0,0,0,0,0,0,0,184,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,78,16,9,0,4,0,0,0,2,0,0,0,0,0,0,0,188,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,87,16,9,0,4,0,0,0,4,0,0,0,0,0,0,0,108,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,97,16,9,0,4,0,0,0,2,0,0,0,0,0,0,0,112,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,107,16,9,0,4,0,0,0,2,0,0,0,0,0,0,0,116,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,116,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,12,0,0,0,2,0,0,0,14,0,0,0,8,0,0,0,0,0,0,0,127,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,40,0,0,0,2,0,0,0,10,0,0,0,9,0,0,0,0,0,0,0,138,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,60,0,0,0,2,0,0,0,14,0,0,0,10,0,0,0,0,0,0,0,150,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,88,0,0,0,2,0,0,0,10,0,0,0,11,0,0,0,0,0,0,0,167,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,120,0,0,0,2,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,173,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,122,0,0,0,2,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,179,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,192,0,0,0,2,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,190,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,128,0,0,0,2,0,0,0,12,0,0,0,124,0,0,0,0,0,0,0,200,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,154,0,0,0,2,0,0,0,12,0,0,0,125,0,0,0,0,0,0,0,226,16,9,0,4,0,0,0,1,0,0,0,0,0,0,0,126,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,141,18,9,0,5,0,0,0,7,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,175,31,8,0,5,0,0,0,11,0,0,0,196,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,150,18,9,0,5,0,0,0,11,0,0,0,197,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,210,16,9,0,5,0,0,0,11,0,0,0,198,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,2,0,0,28,0,0,0,183,31,8,0,0,0,1,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,199,9,0,0,72,1,0,0,44,0,0,0,160,0,0,0,200,9,0,0,201,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,202,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,203,9,0,0,204,9,0,0,136,12,9,0,40,208,0,0,255,18,9,0,187,31,8,0,0,0,0,0,0,0,0,0,205,9,0,0,206,9,0,0,24,0,0,0,207,9,0,0,208,9,0,0,209,9,0,0,210,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,24,0,0,0,211,9,0,0,212,9,0,0,213,9,0,0,214,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,28,0,0,0,77,32,8,0,0,0,1,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,215,9,0,0,184,1,0,0,44,0,0,0,196,0,0,0,216,9,0,0,217,9,0,0,0,0,0,0,0,0,0,0,218,9,0,0,219,9,0,0,220,9,0,0,221,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,81,32,8,0,248,208,0,0,255,18,9,0,93,32,8,0,0,0,0,0,0,0,0,0,222,9,0,0,223,9,0,0,224,9,0,0,225,9,0,0,226,9,0,0,227,9,0,0,228,9,0,0,229,9,0,0,230,9,0,0,231,9,0,0,232,9,0,0,233,9,0,0,234,9,0,0,235,9,0,0,236,9,0,0,237,9,0,0,238,9,0,0,239,9,0,0,240,9,0,0,241,9,0,0,242,9,0,0,243,9,0,0,244,9,0,0,245,9,0,0,246,9,0,0,247,9,0,0,248,9,0,0,249,9,0,0,250,9,0,0,251,9,0,0,252,9,0,0,0,0,0,0,0,0,0,0,2,0,0,0,5,0,0,0,4,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,4,0,0,0,2,0,0,0,2,0,0,0,6,0,0,0,4,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,2,0,0,0,6,0,0,0,2,0,0,0,6,0,0,0,2,0,0,0,255,255,255,255,1,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,2,0,0,0,253,9,0,0,254,9,0,0,255,9,0,0,242,32,8,0,251,32,8,0,5,33,8,0,163,27,9,0,64,17,9,0,49,17,9,0,28,17,9,0,231,119,9,0,14,33,8,0,17,33,8,0,20,33,8,0,30,33,8,0,40,33,8,0,53,33,8,0,64,33,8,0,74,33,8,0,89,33,8,0,97,33,8,0,112,33,8,0,126,33,8,0,139,33,8,0,154,33,8,0,166,33,8,0,179,33,8,0,192,33,8,0,14,18,9,0,141,18,9,0,101,18,9,0,5,18,9,0,200,33,8,0,211,33,8,0,221,33,8,0,234,33,8,0,32,18,9,0,243,33,8,0,246,33,8,0,250,33,8,0,254,33,8,0,196,120,9,0,2,34,8,0,16,34,8,0,28,34,8,0,254,17,9,0,30,34,8,0,34,34,8,0,44,34,8,0,61,34,8,0,77,34,8,0,225,32,8,0,92,34,8,0,106,34,8,0,121,34,8,0,137,34,8,0,153,34,8,0,167,16,9,0,173,16,9,0,168,34,8,0,57,18,9,0,75,18,9,0,178,34,8,0,181,34,8,0,189,34,8,0,197,34,8,0,199,34,8,0,202,34,8,0,206,34,8,0,210,34,8,0,213,34,8,0,217,34,8,0,221,34,8,0,224,34,8,0,25,18,9,0,79,17,9,0,227,34,8,0,32,0,0,0,0,10,0,0,1,10,0,0,2,10,0,0,3,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,32,0,0,0,4,10,0,0,1,10,0,0,2,10,0,0,3,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,28,0,0,0,5,10,0,0,6,10,0,0,7,10,0,0,8,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,24,0,0,0,9,10,0,0,10,10,0,0,11,10,0,0,12,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,20,211,0,0,60,211,0,0,100,211,0,0,140,211,0,0,0,0,0,0,12,0,0,0,202,17,9,0,0,0,2,0,0,0,2,0,232,211,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,209,0,0,20,209,0,0,72,209,0,0,104,209,0,0,13,10,0,0,180,211,0,0,224,209,0,0,0,0,0,0,168,0,0,0,254,15,9,0,0,0,1,0,0,0,2,0,40,212,0,0,14,10,0,0,15,10,0,0,0,0,0,0,16,10,0,0,17,10,0,0,18,10,0,0,0,0,0,0,12,0,0,0,235,34,8,0,0,0,1,0,0,0,2,0,88,212,0,0,0,0,0,0,0,0,0,0,19,10,0,0,20,10,0,0,21,10,0,0,22,10,0,0,23,10,0,0,24,10,0,0,25,10,0,0,72,155,2,0,226,165,2,0,185,17,9,0,88,212,0,0,0,0,0,0,0,0,0,0,108,116,117,111,26,10,0,0,27,10,0,0,28,10,0,0,29,10,0,0,30,10,0,0,2,0,0,0,64,0,0,0,117,12,9,0,0,0,1,0,0,0,2,0,0,0,0,0,31,10,0,0,0,0,0,0,0,0,0,0,108,116,117,111,32,10,0,0,33,10,0,0,34,10,0,0,35,10,0,0,136,212,0,0,24,0,0,0,36,10,0,0,0,0,0,0,37,10,0,0,38,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,39,10,0,0,40,10,0,0,24,0,0,0,36,10,0,0,0,0,0,0,41,10,0,0,42,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,43,10,0,0,44,10,0,0,56,0,0,0,45,10,0,0,0,0,0,0,46,10,0,0,47,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,48,10,0,0,49,10,0,0,24,0,0,0,36,10,0,0,0,0,0,0,50,10,0,0,51,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,6,0,0,0,52,10,0,0,53,10,0,0,24,0,0,0,36,10,0,0,0,0,0,0,54,10,0,0,55,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,8,0,0,0,56,10,0,0,57,10,0,0,24,0,0,0,36,10,0,0,0,0,0,0,58,10,0,0,59,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,10,0,0,0,60,10,0,0,61,10,0,0,44,0,0,0,62,10,0,0,0,0,0,0,63,10,0,0,64,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,12,0,0,0,65,10,0,0,66,10,0,0,44,0,0,0,67,10,0,0,0,0,0,0,68,10,0,0,69,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,13,0,0,0,70,10,0,0,71,10,0,0,40,0,0,0,72,10,0,0,73,10,0,0,74,10,0,0,75,10,0,0,76,10,0,0,77,10,0,0,78,10,0,0,79,10,0,0,80,10,0,0,14,0,0,0,81,10,0,0,82,10,0,0,0,0,0,0,12,0,0,0,16,13,9,0,0,0,1,0,0,0,2,0,212,214,0,0,0,0,0,0,0,0,0,0,83,10,0,0,84,10,0,0,85,10,0,0,86,10,0,0,87,10,0,0,83,10,0,0,88,10,0,0,89,10,0,0,90,10,0,0,91,10,0,0,92,10,0,0,93,10,0,0,94,10,0,0,95,10,0,0,96,10,0,0,97,10,0,0,98,10,0,0,99,10,0,0,100,10,0,0,101,10,0,0,102,10,0,0,103,10,0,0,104,10,0,0,105,10,0,0,106,10,0,0,107,10,0,0,108,10,0,0,109,10,0,0,110,10,0,0,111,10,0,0,220,212,0,0,16,213,0,0,68,213,0,0,120,213,0,0,172,213,0,0,224,213,0,0,20,214,0,0,72,214,0,0,124,214,0,0,0,0,0,0,2,0,0,0,255,255,255,255,99,105,110,117,0,0,0,0,255,255,255,255,99,105,110,117,1,0,0,0,0,0,0,0,110,109,114,97,3,0,0,0,0,0,0,0,98,109,121,115,3,0,0,0,10,0,0,0,99,105,110,117,3,0,0,0,1,0,0,0,99,105,110,117,3,0,0,0,2,0,0,0,115,105,106,115,3,0,0,0,3,0,0,0,32,32,98,103,3,0,0,0,4,0,0,0,53,103,105,98,3,0,0,0,5,0,0,0,115,110,97,119,3,0,0,0,6,0,0,0,97,104,111,106,125,12,9,0,36,216,0,0,140,17,9,0,48,216,0,0,129,17,9,0,52,216,0,0,136,12,9,0,60,216,0,0,174,12,9,0,68,216,0,0,0,0,0,0,0,0,0,0,88,10,0,0,112,10,0,0,113,10,0,0,114,10,0,0,115,10,0,0,116,10,0,0,117,10,0,0,118,10,0,0,119,10,0,0,108,116,117,111,120,10,0,0,121,10,0,0,122,10,0,0,123,10,0,0,124,10,0,0,125,10,0,0,126,10,0,0,127,10,0,0,128,10,0,0,0,0,0,0,0,0,0,0,2,0,0,0,64,0,0,0,183,12,9,0,0,0,1,0,0,0,2,0,0,0,0,0,129,10,0,0,0,0,0,0,0,0,0,0,108,116,117,111,130,10,0,0,131,10,0,0,132,10,0,0,133,10,0,0,72,216,0,0,2,0,0,0,64,0,0,0,190,12,9,0,0,0,1,0,0,0,2,0,0,0,0,0,129,10,0,0,0,0,0,0,0,0,0,0,108,116,117,111,134,10,0,0,131,10,0,0,132,10,0,0,133,10,0,0,72,216,0,0,2,0,0,0,64,0,0,0,201,12,9,0,0,0,1,0,0,0,2,0,0,0,0,0,129,10,0,0,0,0,0,0,0,0,0,0,108,116,117,111,135,10,0,0,131,10,0,0,132,10,0,0,133,10,0,0,72,216,0,0,1,5,0,0,68,0,0,0,120,17,9,0,0,0,1,0,0,0,2,0,0,0,0,0,136,10,0,0,137,10,0,0,138,10,0,0,24,3,0,0,48,1,0,0,160,0,0,0,139,10,0,0,140,10,0,0,141,10,0,0,142,10,0,0,143,10,0,0,0,0,0,0,144,10,0,0,145,10,0,0,0,0,0,0,146,10,0,0,147,10,0,0,148,10,0,0,255,18,9,0,213,12,9,0,222,15,9,0,188,217,0,0,228,12,9,0,208,217,0,0,244,12,9,0,212,217,0,0,57,22,9,0,216,217,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,149,10,0,0,150,10,0,0,151,10,0,0,2,0,0,0,152,10,0,0,153,10,0,0,154,10,0,0,88,240,188,5,228,2,0,0,241,59,35,40,196,135,0,0,234,161,68,163,225,1,0,0,88,240,188,5,228,2,0,0,241,59,35,40,196,135,0,0,235,161,68,163,225,1,0,0,212,234,229,17,80,3,0,0,59,202,48,90,99,144,0,0,2,38,164,19,126,0,0,0,252,255,251,255,8,0,0,0,184,72,158,156,162,190,0,0,18,1,2,112,8,0,0,0,252,255,251,255,8,0,0,0,131,4,90,10,57,124,1,0,18,1,2,112,8,0,0,0,0,0,0,0,0,0,0,0,85,37,201,64,229,0,0,0,227,88,155,163,124,17,0,0,0,0,0,0,0,0,0,0,82,22,196,51,229,0,0,0,42,197,214,38,106,15,0,0,0,0,0,0,0,0,0,0,29,101,177,109,157,1,0,0,3,75,110,108,146,36,0,0,0,0,0,0,0,0,0,0,85,37,201,64,229,0,0,0,208,250,81,222,124,17,0,0,0,0,0,0,0,0,0,0,100,118,228,133,229,0,0,0,49,40,198,166,170,28,0,0,0,0,0,0,0,0,0,0,253,28,137,45,157,1,0,0,51,70,96,160,232,29,0,0,0,0,0,0,0,0,0,0,76,119,170,64,203,1,0,0,150,170,92,155,154,31,0,0,0,0,0,0,0,0,0,0,203,233,61,13,65,1,0,0,102,119,18,212,128,34,0,0,0,0,0,0,0,0,0,0,152,38,105,74,240,1,0,0,70,67,13,52,202,31,0,0,0,0,0,0,0,0,0,0,4,198,52,205,102,1,0,0,70,16,243,108,176,34,0,0,0,0,0,0,0,0,0,0,21,83,167,93,157,1,0,0,95,90,116,64,224,34,0,0,0,0,0,0,0,0,0,0,72,252,85,240,194,1,0,0,211,222,0,57,24,30,0,0,0,0,0,0,0,0,0,64,0,0,0,64,0,0,0,64,0,0,0,0,1,0,0,0,64,0,0,0,1,0,0,0,1,0,0,0,68,0,0,0,0,0,0,0,0,0,0,0,9,0,3,0,0,0,0,0,0,0,0,0,1,0,1,0,1,0,0,0,1,5,0,0,28,0,0,0,201,15,9,0,0,0,1,0,0,0,2,0,0,0,0,0,155,10,0,0,156,10,0,0,157,10,0,0,40,2,0,0,44,0,0,0,180,0,0,0,158,10,0,0,159,10,0,0,160,10,0,0,161,10,0,0,162,10,0,0,163,10,0,0,164,10,0,0,165,10,0,0,166,10,0,0,167,10,0,0,168,10,0,0,0,0,0,0,140,17,9,0,84,220,0,0,129,17,9,0,88,220,0,0,255,18,9,0,207,15,9,0,161,17,9,0,96,220,0,0,214,15,9,0,116,220,0,0,222,15,9,0,120,220,0,0,0,0,0,0,0,0,0,0,169,10,0,0,170,10,0,0,171,10,0,0,172,10,0,0,173,10,0,0,174,10,0,0,175,10,0,0,176,10,0,0,177,10,0,0,178,10,0,0,179,10,0,0,180,10,0,0,181,10,0,0,182,10,0,0,246,17,9,0,3,0,0,0,5,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,254,17,9,0,3,0,0,0,5,0,0,0,0,0,0,0,4,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,5,18,9,0,3,0,0,0,5,0,0,0,0,0,0,0,8,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,14,18,9,0,3,0,0,0,5,0,0,0,0,0,0,0,12,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,25,18,9,0,3,0,0,0,5,0,0,0,0,0,0,0,16,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,32,18,9,0,3,0,0,0,2,0,0,0,0,0,0,0,20,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,44,18,9,0,3,0,0,0,1,0,0,0,0,0,0,0,24,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,57,18,9,0,3,0,0,0,2,0,0,0,0,0,0,0,26,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,75,18,9,0,3,0,0,0,2,0,0,0,0,0,0,0,28,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,94,18,9,0,2,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,49,16,9,0,4,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,3,0,0,0,58,16,9,0,4,0,0,0,2,0,0,0,0,0,0,0,4,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,64,16,9,0,4,0,0,0,2,0,0,0,0,0,0,0,184,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,78,16,9,0,4,0,0,0,2,0,0,0,0,0,0,0,188,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,87,16,9,0,4,0,0,0,4,0,0,0,0,0,0,0,108,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,97,16,9,0,4,0,0,0,2,0,0,0,0,0,0,0,112,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,107,16,9,0,4,0,0,0,2,0,0,0,0,0,0,0,116,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,116,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,12,0,0,0,2,0,0,0,14,0,0,0,8,0,0,0,2,0,0,0,127,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,40,0,0,0,2,0,0,0,10,0,0,0,9,0,0,0,2,0,0,0,138,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,60,0,0,0,2,0,0,0,14,0,0,0,10,0,0,0,2,0,0,0,150,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,88,0,0,0,2,0,0,0,10,0,0,0,11,0,0,0,2,0,0,0,167,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,120,0,0,0,2,0,0,0,1,0,0,0,0,0,0,0,2,0,0,0,173,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,122,0,0,0,2,0,0,0,1,0,0,0,0,0,0,0,2,0,0,0,179,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,192,0,0,0,2,0,0,0,2,0,0,0,0,0,0,0,2,0,0,0,190,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,128,0,0,0,2,0,0,0,12,0,0,0,124,0,0,0,2,0,0,0,200,16,9,0,4,0,0,0,9,0,0,0,0,0,0,0,154,0,0,0,2,0,0,0,12,0,0,0,125,0,0,0,2,0,0,0,210,16,9,0,4,0,0,0,3,0,0,0,0,0,0,0,180,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,226,16,9,0,4,0,0,0,1,0,0,0,0,0,0,0,126,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,101,18,9,0,1,0,0,0,6,0,0,0,0,0,0,0,232,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,110,18,9,0,1,0,0,0,2,0,0,0,0,0,0,0,44,1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,120,18,9,0,1,0,0,0,2,0,0,0,0,0,0,0,45,1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,129,18,9,0,1,0,0,0,3,0,0,0,0,0,0,0,92,1,0,0,4,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,141,18,9,0,5,0,0,0,7,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,236,16,9,0,7,0,0,0,2,0,0,0,0,0,0,0,20,2,0,0,4,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,240,16,9,0,7,0,0,0,2,0,0,0,0,0,0,0,24,2,0,0,4,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,244,16,9,0,8,0,0,0,9,0,0,0,0,0,0,0,96,1,0,0,4,0,0,0,16,0,0,0,160,1,0,0,1,0,0,0,150,18,9,0,8,0,0,0,11,0,0,0,183,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,161,18,9,0,8,0,0,0,11,0,0,0,184,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,17,9,0,8,0,0,0,11,0,0,0,185,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,228,18,9,0,8,0,0,0,11,0,0,0,186,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,20,17,9,0,8,0,0,0,11,0,0,0,187,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,28,17,9,0,8,0,0,0,11,0,0,0,188,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,49,17,9,0,8,0,0,0,11,0,0,0,189,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,64,17,9,0,8,0,0,0,11,0,0,0,190,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,79,17,9,0,8,0,0,0,11,0,0,0,191,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,92,17,9,0,8,0,0,0,11,0,0,0,192,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,5,0,0,32,0,0,0,113,17,9,0,0,0,1,0,0,0,2,0,0,0,0,0,193,10,0,0,194,10,0,0,195,10,0,0,48,2,0,0,48,0,0,0,164,0,0,0,196,10,0,0,197,10,0,0,198,10,0,0,199,10,0,0,200,10,0,0,201,10,0,0,202,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,203,10,0,0,204,10,0,0,129,17,9,0,176,227,0,0,140,17,9,0,184,227,0,0,161,17,9,0,188,227,0,0,255,18,9,0,177,17,9,0,0,0,0,0,0,0,0,0,205,10,0,0,206,10,0,0,207,10,0,0,208,10,0,0,209,10,0,0,210,10,0,0,211,10,0,0,0,0,0,0,246,17,9,0,3,0,0,0,5,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,254,17,9,0,3,0,0,0,5,0,0,0,0,0,0,0,4,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,5,18,9,0,3,0,0,0,5,0,0,0,0,0,0,0,8,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,14,18,9,0,3,0,0,0,5,0,0,0,0,0,0,0,12,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,25,18,9,0,3,0,0,0,5,0,0,0,0,0,0,0,16,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,32,18,9,0,3,0,0,0,2,0,0,0,0,0,0,0,20,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,44,18,9,0,3,0,0,0,1,0,0,0,0,0,0,0,24,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,57,18,9,0,3,0,0,0,2,0,0,0,0,0,0,0,26,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,75,18,9,0,3,0,0,0,2,0,0,0,0,0,0,0,28,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,94,18,9,0,2,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,101,18,9,0,1,0,0,0,6,0,0,0,0,0,0,0,232,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,110,18,9,0,1,0,0,0,2,0,0,0,0,0,0,0,44,1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,120,18,9,0,1,0,0,0,2,0,0,0,0,0,0,0,45,1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,129,18,9,0,1,0,0,0,3,0,0,0,0,0,0,0,92,1,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,141,18,9,0,5,0,0,0,7,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,150,18,9,0,5,0,0,0,11,0,0,0,212,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161,18,9,0,5,0,0,0,11,0,0,0,213,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,228,18,9,0,5,0,0,0,11,0,0,0,214,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,240,18,9,0,5,0,0,0,11,0,0,0,215,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,2,0,0,28,0,0,0,246,18,9,0,0,0,1,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,216,10,0,0,152,0,0,0,44,0,0,0,160,0,0,0,217,10,0,0,218,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,219,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,220,10,0,0,221,10,0,0,255,18,9,0,11,19,9,0,246,18,9,0,24,231,0,0,0,0,0,0,0,0,0,0,222,10,0,0,24,0,0,0,223,10,0,0,0,0,0,0,224,10,0,0,225,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,13,0,0,0,0,0,0,0,247,2,0,0,0,0,0,0,24,0,0,0,1,0,0,0,41,0,0,0,0,0,0,0,58,0,0,0,5,0,0,0,58,0,0,0,0,0,0,0,75,0,0,0,0,0,0,0,247,2,0,0,0,0,0,0,107,0,0,0,1,0,0,0,132,0,0,0,1,0,0,0,82,0,0,0,7,0,0,0,82,0,0,0,0,0,0,0,157,0,0,0,0,0,0,0,247,2,0,0,0,0,0,0,164,0,0,0,1,0,0,0,179,0,0,0,0,0,0,0,192,0,0,0,1,0,0,0,205,0,0,0,5,0,0,0,205,0,0,0,0,0,0,0,222,0,0,0,0,0,0,0,247,2,0,0,0,0,0,0,239,0,0,0,9,0,0,0,0,1,0,0,0,0,0,0,13,1,0,0,0,0,0,0,247,2,0,0,0,0,0,0,24,1,0,0,1,0,0,0,33,1,0,0,0,0,0,0,42,1,0,0,1,0,0,0,50,1,0,0,5,0,0,0,50,1,0,0,0,0,0,0,58,1,0,0,0,0,0,0,247,2,0,0,0,0,0,0,64,1,0,0,1,0,0,0,86,1,0,0,0,0,0,0,247,2,0,0,0,0,0,0,108,1,0,0,5,0,0,0,127,1,0,0,0,0,0,0,149,1,0,0,1,0,0,0,159,1,0,0,1,0,0,0,169,1,0,0,0,0,0,0,182,1,0,0,0,0,0,0,189,1,0,0,0,0,0,0,247,2,0,0,0,0,0,0,199,1,0,0,1,0,0,0,95,2,0,0,0,0,0,0,247,2,0,0,0,0,0,0,0,0,0,0,208,232,0,0,68,6,0,0,45,6,0,0,64,6,0,0,0,6,0,0,255,6,0,0,80,7,0,0,255,7,0,0,160,8,0,0,255,8,0,0,80,251,0,0,255,253,0,0,112,254,0,0,255,254,0,0,0,238,1,0,255,238,1,0,0,0,0,0,0,0,0,0,1,0,0,0,28,233,0,0,62,4,0,0,30,4,0,0,0,0,0,0,0,4,0,0,255,4,0,0,0,5,0,0,47,5,0,0,224,45,0,0,255,45,0,0,64,166,0,0,159,166,0,0,0,0,0,0,0,0,0,0,2,0,0,0,88,233,0,0,32,9,0,0,53,9,0,0,31,9,0,0,0,9,0,0,59,9,0,0,61,9,0,0,80,9,0,0,83,9,0,0,99,9,0,0,102,9,0,0,127,9,0,0,185,32,0,0,185,32,0,0,0,0,0,0,0,0,0,0,3,0,0,0,156,233,0,0,191,3,0,0,159,3,0,0,0,0,0,0,112,3,0,0,255,3,0,0,0,31,0,0,255,31,0,0,0,0,0,0,0,0,0,0,4,0,0,0,200,233,0,0,221,5,0,0,0,0,0,0,0,0,0,0,144,5,0,0,255,5,0,0,29,251,0,0,79,251,0,0,0,0,0,0,0,0,0,0,5,0,0,0,244,233,0,0,111,0,0,0,79,0,0,0,48,0,0,0,32,0,0,0,127,0,0,0,160,0,0,0,255,0,0,0,0,1,0,0,127,1,0,0,128,1,0,0,79,2,0,0,80,2,0,0,175,2,0,0,176,2,0,0,255,2,0,0,0,3,0,0,111,3,0,0,0,29,0,0,127,29,0,0,128,29,0,0,191,29,0,0,192,29,0,0,255,29,0,0,0,30,0,0,255,30,0,0,0,32,0,0,111,32,0,0,112,32,0,0,159,32,0,0,160,32,0,0,184,32,0,0,186,32,0,0,207,32,0,0,80,33,0,0,143,33,0,0,96,36,0,0,255,36,0,0,96,44,0,0,127,44,0,0,0,46,0,0,127,46,0,0,32,167,0,0,255,167,0,0,0,251,0,0,6,251,0,0,0,212,1,0,255,215,1,0,0,241,1,0,255,241,1,0,0,0,0,0,0,0,0,0,6,0,0,0,68,86,18,0,0,0,0,0,0,0,0,0,0,0,0,0,7,0,0,0,220,234,0,0,102,12,0,0,103,12,0,0,0,0,0,0,0,12,0,0,127,12,0,0,0,0,0,0,0,0,0,0,8,0,0,0,0,235,0,0,50,14,0,0,69,14,0,0,80,14,0,0,0,14,0,0,127,14,0,0,0,0,0,0,0,0,0,0,9,0,0,0,36,235,0,0,111,0,0,0,0,0,0,0,0,0,0,0,128,9,0,0,255,9,0,0,0,0,0,0,0,0,0,0,10,0,0,0,72,235,0,0,111,0,0,0,0,0,0,0,0,0,0,0,128,10,0,0,255,10,0,0,0,0,0,0,0,0,0,0,11,0,0,0,108,235,0,0,111,0,0,0,0,0,0,0,0,0,0,0,0,10,0,0,127,10,0,0,0,0,0,0,0,0,0,0,12,0,0,0,144,235,0,0,111,0,0,0,0,0,0,0,0,0,0,0,128,12,0,0,255,12,0,0,0,0,0,0,0,0,0,0,13,0,0,0,180,235,0,0,111,0,0,0,0,0,0,0,0,0,0,0,0,25,0,0,79,25,0,0,0,0,0,0,0,0,0,0,14,0,0,0,216,235,0,0,111,0,0,0,0,0,0,0,0,0,0,0,0,13,0,0,127,13,0,0,0,0,0,0,0,0,0,0,15,0,0,0,252,235,0,0,111,0,0,0,0,0,0,0,0,0,0,0,0,11,0,0,127,11,0,0,0,0,0,0,0,0,0,0,16,0,0,0,32,236,0,0,111,0,0,0,0,0,0,0,0,0,0,0,128,13,0,0,255,13,0,0,0,0,0,0,0,0,0,0,17,0,0,0,68,236,0,0,111,0,0,0,0,0,0,0,0,0,0,0,128,27,0,0,191,27,0,0,0,0,0,0,0,0,0,0,18,0,0,0,104,236,0,0,111,0,0,0,0,0,0,0,0,0,0,0,0,168,0,0,47,168,0,0,0,0,0,0,0,0,0,0,19,0,0,0,140,236,0,0,111,0,0,0,0,0,0,0,0,0,0,0,128,11,0,0,255,11,0,0,0,0,0,0,0,0,0,0,20,0,0,0,176,236,0,0,111,0,0,0,0,0,0,0,0,0,0,0,0,15,0,0,255,15,0,0,0,0,0,0,0,0,0,0,21,0,0,0,212,236,0,0,48,117,0,0,215,86,0,0,0,0,0,0,0,17,0,0,255,17,0,0,128,46,0,0,255,46,0,0,0,47,0,0,223,47,0,0,240,47,0,0,255,47,0,0,0,48,0,0,63,48,0,0,64,48,0,0,159,48,0,0,160,48,0,0,255,48,0,0,0,49,0,0,47,49,0,0,48,49,0,0,143,49,0,0,144,49,0,0,159,49,0,0,160,49,0,0,191,49,0,0,192,49,0,0,239,49,0,0,240,49,0,0,255,49,0,0,0,50,0,0,255,50,0,0,0,51,0,0,255,51,0,0,0,52,0,0,191,77,0,0,192,77,0,0,255,77,0,0,0,78,0,0,255,159,0,0,96,169,0,0,127,169,0,0,0,172,0,0,175,215,0,0,176,215,0,0,255,215,0,0,0,249,0,0,255,250,0,0,16,254,0,0,31,254,0,0,48,254,0,0,79,254,0,0,0,255,0,0,239,255,0,0,0,176,1,0,255,176,1,0,0,211,1,0,95,211,1,0,0,242,1,0,255,242,1,0,0,0,2,0,223,166,2,0,0,167,2,0,63,183,2,0,64,183,2,0,31,184,2,0,0,248,2,0,31,250,2,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,10,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,3,0,0,0,0,0,0,0,2,0,0,0,1,0,0,0,1,0,0,0,3,0,0,0,1,0,0,0,3,0,0,0,1,0,0,0,1,0,0,0,3,0,0,0,2,0,0,0,4,0,0,0,1,0,0,0,1,0,0,0,3,0,0,0,3,0,0,0,5,0,0,0,1,0,0,0,1,0,0,0,3,0,0,0,5,0,0,0,6,0,0,0,1,0,0,0,1,0,0,0,3,0,0,0,6,0,0,0,7,0,0,0,1,0,0,0,1,0,0,0,3,0,0,0,7,0,0,0,8,0,0,0,1,0,0,0,1,0,0,0,3,0,0,0,8,0,0,0,9,0,0,0,1,0,0,0,1,0,0,0,3,0,0,0,9,0,0,0,10,0,0,0,1,0,0,0,1,0,0,0,3,0,0,0,10,0,0,0,11,0,0,0,1,0,0,0,3,0,0,0,15,0,0,0,0,0,0,0,12,0,0,0,1,0,0,0,3,0,0,0,15,0,0,0,1,0,0,0,13,0,0,0,1,0,0,0,3,0,0,0,15,0,0,0,2,0,0,0,14,0,0,0,1,0,0,0,3,0,0,0,15,0,0,0,3,0,0,0,15,0,0,0,1,0,0,0,3,0,0,0,15,0,0,0,5,0,0,0,16,0,0,0,1,0,0,0,3,0,0,0,15,0,0,0,6,0,0,0,17,0,0,0,1,0,0,0,3,0,0,0,15,0,0,0,7,0,0,0,18,0,0,0,1,0,0,0,3,0,0,0,15,0,0,0,8,0,0,0,19,0,0,0,1,0,0,0,3,0,0,0,15,0,0,0,9,0,0,0,20,0,0,0,1,0,0,0,3,0,0,0,15,0,0,0,10,0,0,0,21,0,0,0,1,0,0,0,4,0,0,0,22,0,0,0,10,0,0,0,22,0,0,0,1,0,0,0,5,0,0,0,26,0,0,0,0,0,0,0,23,0,0,0,1,0,0,0,5,0,0,0,26,0,0,0,1,0,0,0,24,0,0,0,1,0,0,0,5,0,0,0,26,0,0,0,2,0,0,0,25,0,0,0,1,0,0,0,5,0,0,0,26,0,0,0,3,0,0,0,26,0,0,0,1,0,0,0,5,0,0,0,26,0,0,0,5,0,0,0,27,0,0,0,1,0,0,0,5,0,0,0,26,0,0,0,6,0,0,0,28,0,0,0,1,0,0,0,5,0,0,0,26,0,0,0,7,0,0,0,29,0,0,0,1,0,0,0,5,0,0,0,26,0,0,0,8,0,0,0,30,0,0,0,1,0,0,0,5,0,0,0,26,0,0,0,9,0,0,0,31,0,0,0,1,0,0,0,5,0,0,0,26,0,0,0,10,0,0,0,32,0,0,0,1,0,0,0,2,0,0,0,9,0,0,0,10,0,0,0,33,0,0,0,0,0,0,0,6,0,0,0,0,0,0,0,10,0,0,0,34,0,0,0,1,0,0,0,7,0,0,0,33,0,0,0,10,0,0,0,35,0,0,0,1,0,0,0,8,0,0,0,36,0,0,0,10,0,0,0,36,0,0,0,3,0,0,0,9,0,0,0,0,0,0,0,10,0,0,0,37,0,0,0,3,0,0,0,10,0,0,0,0,0,0,0,10,0,0,0,38,0,0,0,3,0,0,0,11,0,0,0,0,0,0,0,10,0,0,0,39,0,0,0,3,0,0,0,12,0,0,0,0,0,0,0,10,0,0,0,40,0,0,0,3,0,0,0,13,0,0,0,0,0,0,0,10,0,0,0,41,0,0,0,3,0,0,0,14,0,0,0,0,0,0,0,10,0,0,0,42,0,0,0,3,0,0,0,15,0,0,0,0,0,0,0,10,0,0,0,43,0,0,0,3,0,0,0,16,0,0,0,0,0,0,0,10,0,0,0,44,0,0,0,3,0,0,0,17,0,0,0,0,0,0,0,10,0,0,0,45,0,0,0,3,0,0,0,18,0,0,0,0,0,0,0,10,0,0,0,46,0,0,0,3,0,0,0,19,0,0,0,0,0,0,0,10,0,0,0,47,0,0,0,3,0,0,0,20,0,0,0,0,0,0,0,10,0,0,0,48,0,0,0,2,0,0,0,21,0,0,0,44,0,0,0,10,0,0,0,0,0,0,0,40,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,226,10,0,0,227,10,0,0,1,0,0,0,116,12,0,0,228,10,0,0,229,10,0,0,0,0,0,0,230,10,0,0,231,10,0,0,188,232,0,0,8,233,0,0,68,233,0,0,136,233,0,0,180,233,0,0,224,233,0,0,180,234,0,0,200,234,0,0,236,234,0,0,16,235,0,0,52,235,0,0,88,235,0,0,124,235,0,0,160,235,0,0,196,235,0,0,232,235,0,0,12,236,0,0,48,236,0,0,84,236,0,0,120,236,0,0,156,236,0,0,192,236,0,0,0,0,0,0,35,0,0,0,32,0,0,0,30,0,0,0,25,0,0,0,20,0,0,0,15,0,0,0,12,0,0,0,10,0,0,0,5,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,255,255,255,255,254,255,255,255,251,255,255,255,248,255,255,255,246,255,255,255,246,255,255,255,236,255,255,255,236,255,255,255,226,255,255,255,226,255,255,255,226,255,255,255,226,255,255,255,236,255,255,255,236,255,255,255,246,255,255,255,246,255,255,255,248,255,255,255,251,255,255,255,254,255,255,255,255,255,255,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,5,0,0,0,10,0,0,0,12,0,0,0,15,0,0,0,20,0,0,0,25,0,0,0,30,0,0,0,32,0,0,0,2,0,0,0,116,12,0,0,232,10,0,0,233,10,0,0,0,0,0,0,234,10,0,0,235,10,0,0,3,0,0,0,116,12,0,0,236,10,0,0,237,10,0,0,0,0,0,0,238,10,0,0,239,10,0,0,176,241,0,0,204,241,0,0,68,243,0,0,96,243,0,0,0,0,0,0,220,237,0,0,240,237,0,0,4,238,0,0,24,238,0,0,44,238,0,0,64,238,0,0,84,238,0,0,104,238,0,0,124,238,0,0,144,238,0,0,164,238,0,0,184,238,0,0,204,238,0,0,224,238,0,0,244,238,0,0,8,239,0,0,28,239,0,0,48,239,0,0,68,239,0,0,88,239,0,0,108,239,0,0,128,239,0,0,148,239,0,0,168,239,0,0,188,239,0,0,208,239,0,0,228,239,0,0,248,239,0,0,12,240,0,0,32,240,0,0,52,240,0,0,72,240,0,0,92,240,0,0,112,240,0,0,132,240,0,0,152,240,0,0,172,240,0,0,192,240,0,0,212,240,0,0,232,240,0,0,252,240,0,0,16,241,0,0,36,241,0,0,56,241,0,0,76,241,0,0,96,241,0,0,116,241,0,0,136,241,0,0,156,241,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,240,10,0,0,4,0,0,0,24,0,0,0,46,22,9,0,0,0,1,0,0,0,2,0,88,244,0,0,241,10,0,0,242,10,0,0,243,10,0,0,57,22,9,0,156,244,0,0,0,0,0,0,0,0,0,0,244,10,0,0,245,10,0,0,96,7,0,0,0,1,0,0,0,8,0,0,80,0,0,0,0,8,0,0,16,0,0,0,84,8,0,0,115,0,0,0,82,7,0,0,31,0,0,0,0,8,0,0,112,0,0,0,0,8,0,0,48,0,0,0,0,9,0,0,192,0,0,0,80,7,0,0,10,0,0,0,0,8,0,0,96,0,0,0,0,8,0,0,32,0,0,0,0,9,0,0,160,0,0,0,0,8,0,0,0,0,0,0,0,8,0,0,128,0,0,0,0,8,0,0,64,0,0,0,0,9,0,0,224,0,0,0,80,7,0,0,6,0,0,0,0,8,0,0,88,0,0,0,0,8,0,0,24,0,0,0,0,9,0,0,144,0,0,0,83,7,0,0,59,0,0,0,0,8,0,0,120,0,0,0,0,8,0,0,56,0,0,0,0,9,0,0,208,0,0,0,81,7,0,0,17,0,0,0,0,8,0,0,104,0,0,0,0,8,0,0,40,0,0,0,0,9,0,0,176,0,0,0,0,8,0,0,8,0,0,0,0,8,0,0,136,0,0,0,0,8,0,0,72,0,0,0,0,9,0,0,240,0,0,0,80,7,0,0,4,0,0,0,0,8,0,0,84,0,0,0,0,8,0,0,20,0,0,0,85,8,0,0,227,0,0,0,83,7,0,0,43,0,0,0,0,8,0,0,116,0,0,0,0,8,0,0,52,0,0,0,0,9,0,0,200,0,0,0,81,7,0,0,13,0,0,0,0,8,0,0,100,0,0,0,0,8,0,0,36,0,0,0,0,9,0,0,168,0,0,0,0,8,0,0,4,0,0,0,0,8,0,0,132,0,0,0,0,8,0,0,68,0,0,0,0,9,0,0,232,0,0,0,80,7,0,0,8,0,0,0,0,8,0,0,92,0,0,0,0,8,0,0,28,0,0,0,0,9,0,0,152,0,0,0,84,7,0,0,83,0,0,0,0,8,0,0,124,0,0,0,0,8,0,0,60,0,0,0,0,9,0,0,216,0,0,0,82,7,0,0,23,0,0,0,0,8,0,0,108,0,0,0,0,8,0,0,44,0,0,0,0,9,0,0,184,0,0,0,0,8,0,0,12,0,0,0,0,8,0,0,140,0,0,0,0,8,0,0,76,0,0,0,0,9,0,0,248,0,0,0,80,7,0,0,3,0,0,0,0,8,0,0,82,0,0,0,0,8,0,0,18,0,0,0,85,8,0,0,163,0,0,0,83,7,0,0,35,0,0,0,0,8,0,0,114,0,0,0,0,8,0,0,50,0,0,0,0,9,0,0,196,0,0,0,81,7,0,0,11,0,0,0,0,8,0,0,98,0,0,0,0,8,0,0,34,0,0,0,0,9,0,0,164,0,0,0,0,8,0,0,2,0,0,0,0,8,0,0,130,0,0,0,0,8,0,0,66,0,0,0,0,9,0,0,228,0,0,0,80,7,0,0,7,0,0,0,0,8,0,0,90,0,0,0,0,8,0,0,26,0,0,0,0,9,0,0,148,0,0,0,84,7,0,0,67,0,0,0,0,8,0,0,122,0,0,0,0,8,0,0,58,0,0,0,0,9,0,0,212,0,0,0,82,7,0,0,19,0,0,0,0,8,0,0,106,0,0,0,0,8,0,0,42,0,0,0,0,9,0,0,180,0,0,0,0,8,0,0,10,0,0,0,0,8,0,0,138,0,0,0,0,8,0,0,74,0,0,0,0,9,0,0,244,0,0,0,80,7,0,0,5,0,0,0,0,8,0,0,86,0,0,0,0,8,0,0,22,0,0,0,192,8,0,0,0,0,0,0,83,7,0,0,51,0,0,0,0,8,0,0,118,0,0,0,0,8,0,0,54,0,0,0,0,9,0,0,204,0,0,0,81,7,0,0,15,0,0,0,0,8,0,0,102,0,0,0,0,8,0,0,38,0,0,0,0,9,0,0,172,0,0,0,0,8,0,0,6,0,0,0,0,8,0,0,134,0,0,0,0,8,0,0,70,0,0,0,0,9,0,0,236,0,0,0,80,7,0,0,9,0,0,0,0,8,0,0,94,0,0,0,0,8,0,0,30,0,0,0,0,9,0,0,156,0,0,0,84,7,0,0,99,0,0,0,0,8,0,0,126,0,0,0,0,8,0,0,62,0,0,0,0,9,0,0,220,0,0,0,82,7,0,0,27,0,0,0,0,8,0,0,110,0,0,0,0,8,0,0,46,0,0,0,0,9,0,0,188,0,0,0,0,8,0,0,14,0,0,0,0,8,0,0,142,0,0,0,0,8,0,0,78,0,0,0,0,9,0,0,252,0,0,0,96,7,0,0,0,1,0,0,0,8,0,0,81,0,0,0,0,8,0,0,17,0,0,0,85,8,0,0,131,0,0,0,82,7,0,0,31,0,0,0,0,8,0,0,113,0,0,0,0,8,0,0,49,0,0,0,0,9,0,0,194,0,0,0,80,7,0,0,10,0,0,0,0,8,0,0,97,0,0,0,0,8,0,0,33,0,0,0,0,9,0,0,162,0,0,0,0,8,0,0,1,0,0,0,0,8,0,0,129,0,0,0,0,8,0,0,65,0,0,0,0,9,0,0,226,0,0,0,80,7,0,0,6,0,0,0,0,8,0,0,89,0,0,0,0,8,0,0,25,0,0,0,0,9,0,0,146,0,0,0,83,7,0,0,59,0,0,0,0,8,0,0,121,0,0,0,0,8,0,0,57,0,0,0,0,9,0,0,210,0,0,0,81,7,0,0,17,0,0,0,0,8,0,0,105,0,0,0,0,8,0,0,41,0,0,0,0,9,0,0,178,0,0,0,0,8,0,0,9,0,0,0,0,8,0,0,137,0,0,0,0,8,0,0,73,0,0,0,0,9,0,0,242,0,0,0,80,7,0,0,4,0,0,0,0,8,0,0,85,0,0,0,0,8,0,0,21,0,0,0,80,8,0,0,2,1,0,0,83,7,0,0,43,0,0,0,0,8,0,0,117,0,0,0,0,8,0,0,53,0,0,0,0,9,0,0,202,0,0,0,81,7,0,0,13,0,0,0,0,8,0,0,101,0,0,0,0,8,0,0,37,0,0,0,0,9,0,0,170,0,0,0,0,8,0,0,5,0,0,0,0,8,0,0,133,0,0,0,0,8,0,0,69,0,0,0,0,9,0,0,234,0,0,0,80,7,0,0,8,0,0,0,0,8,0,0,93,0,0,0,0,8,0,0,29,0,0,0,0,9,0,0,154,0,0,0,84,7,0,0,83,0,0,0,0,8,0,0,125,0,0,0,0,8,0,0,61,0,0,0,0,9,0,0,218,0,0,0,82,7,0,0,23,0,0,0,0,8,0,0,109,0,0,0,0,8,0,0,45,0,0,0,0,9,0,0,186,0,0,0,0,8,0,0,13,0,0,0,0,8,0,0,141,0,0,0,0,8,0,0,77,0,0,0,0,9,0,0,250,0,0,0,80,7,0,0,3,0,0,0,0,8,0,0,83,0,0,0,0,8,0,0,19,0,0,0,85,8,0,0,195,0,0,0,83,7,0,0,35,0,0,0,0,8,0,0,115,0,0,0,0,8,0,0,51,0,0,0,0,9,0,0,198,0,0,0,81,7,0,0,11,0,0,0,0,8,0,0,99,0,0,0,0,8,0,0,35,0,0,0,0,9,0,0,166,0,0,0,0,8,0,0,3,0,0,0,0,8,0,0,131,0,0,0,0,8,0,0,67,0,0,0,0,9,0,0,230,0,0,0,80,7,0,0,7,0,0,0,0,8,0,0,91,0,0,0,0,8,0,0,27,0,0,0,0,9,0,0,150,0,0,0,84,7,0,0,67,0,0,0,0,8,0,0,123,0,0,0,0,8,0,0,59,0,0,0,0,9,0,0,214,0,0,0,82,7,0,0,19,0,0,0,0,8,0,0,107,0,0,0,0,8,0,0,43,0,0,0,0,9,0,0,182,0,0,0,0,8,0,0,11,0,0,0,0,8,0,0,139,0,0,0,0,8,0,0,75,0,0,0,0,9,0,0,246,0,0,0,80,7,0,0,5,0,0,0,0,8,0,0,87,0,0,0,0,8,0,0,23,0,0,0,192,8,0,0,0,0,0,0,83,7,0,0,51,0,0,0,0,8,0,0,119,0,0,0,0,8,0,0,55,0,0,0,0,9,0,0,206,0,0,0,81,7,0,0,15,0,0,0,0,8,0,0,103,0,0,0,0,8,0,0,39,0,0,0,0,9,0,0,174,0,0,0,0,8,0,0,7,0,0,0,0,8,0,0,135,0,0,0,0,8,0,0,71,0,0,0,0,9,0,0,238,0,0,0,80,7,0,0,9,0,0,0,0,8,0,0,95,0,0,0,0,8,0,0,31,0,0,0,0,9,0,0,158,0,0,0,84,7,0,0,99,0,0,0,0,8,0,0,127,0,0,0,0,8,0,0,63,0,0,0,0,9,0,0,222,0,0,0,82,7,0,0,27,0,0,0,0,8,0,0,111,0,0,0,0,8,0,0,47,0,0,0,0,9,0,0,190,0,0,0,0,8,0,0,15,0,0,0,0,8,0,0,143,0,0,0,0,8,0,0,79,0,0,0,0,9,0,0,254,0,0,0,96,7,0,0,0,1,0,0,0,8,0,0,80,0,0,0,0,8,0,0,16,0,0,0,84,8,0,0,115,0,0,0,82,7,0,0,31,0,0,0,0,8,0,0,112,0,0,0,0,8,0,0,48,0,0,0,0,9,0,0,193,0,0,0,80,7,0,0,10,0,0,0,0,8,0,0,96,0,0,0,0,8,0,0,32,0,0,0,0,9,0,0,161,0,0,0,0,8,0,0,0,0,0,0,0,8,0,0,128,0,0,0,0,8,0,0,64,0,0,0,0,9,0,0,225,0,0,0,80,7,0,0,6,0,0,0,0,8,0,0,88,0,0,0,0,8,0,0,24,0,0,0,0,9,0,0,145,0,0,0,83,7,0,0,59,0,0,0,0,8,0,0,120,0,0,0,0,8,0,0,56,0,0,0,0,9,0,0,209,0,0,0,81,7,0,0,17,0,0,0,0,8,0,0,104,0,0,0,0,8,0,0,40,0,0,0,0,9,0,0,177,0,0,0,0,8,0,0,8,0,0,0,0,8,0,0,136,0,0,0,0,8,0,0,72,0,0,0,0,9,0,0,241,0,0,0,80,7,0,0,4,0,0,0,0,8,0,0,84,0,0,0,0,8,0,0,20,0,0,0,85,8,0,0,227,0,0,0,83,7,0,0,43,0,0,0,0,8,0,0,116,0,0,0,0,8,0,0,52,0,0,0,0,9,0,0,201,0,0,0,81,7,0,0,13,0,0,0,0,8,0,0,100,0,0,0,0,8,0,0,36,0,0,0,0,9,0,0,169,0,0,0,0,8,0,0,4,0,0,0,0,8,0,0,132,0,0,0,0,8,0,0,68,0,0,0,0,9,0,0,233,0,0,0,80,7,0,0,8,0,0,0,0,8,0,0,92,0,0,0,0,8,0,0,28,0,0,0,0,9,0,0,153,0,0,0,84,7,0,0,83,0,0,0,0,8,0,0,124,0,0,0,0,8,0,0,60,0,0,0,0,9,0,0,217,0,0,0,82,7,0,0,23,0,0,0,0,8,0,0,108,0,0,0,0,8,0,0,44,0,0,0,0,9,0,0,185,0,0,0,0,8,0,0,12,0,0,0,0,8,0,0,140,0,0,0,0,8,0,0,76,0,0,0,0,9,0,0,249,0,0,0,80,7,0,0,3,0,0,0,0,8,0,0,82,0,0,0,0,8,0,0,18,0,0,0,85,8,0,0,163,0,0,0,83,7,0,0,35,0,0,0,0,8,0,0,114,0,0,0,0,8,0,0,50,0,0,0,0,9,0,0,197,0,0,0,81,7,0,0,11,0,0,0,0,8,0,0,98,0,0,0,0,8,0,0,34,0,0,0,0,9,0,0,165,0,0,0,0,8,0,0,2,0,0,0,0,8,0,0,130,0,0,0,0,8,0,0,66,0,0,0,0,9,0,0,229,0,0,0,80,7,0,0,7,0,0,0,0,8,0,0,90,0,0,0,0,8,0,0,26,0,0,0,0,9,0,0,149,0,0,0,84,7,0,0,67,0,0,0,0,8,0,0,122,0,0,0,0,8,0,0,58,0,0,0,0,9,0,0,213,0,0,0,82,7,0,0,19,0,0,0,0,8,0,0,106,0,0,0,0,8,0,0,42,0,0,0,0,9,0,0,181,0,0,0,0,8,0,0,10,0,0,0,0,8,0,0,138,0,0,0,0,8,0,0,74,0,0,0,0,9,0,0,245,0,0,0,80,7,0,0,5,0,0,0,0,8,0,0,86,0,0,0,0,8,0,0,22,0,0,0,192,8,0,0,0,0,0,0,83,7,0,0,51,0,0,0,0,8,0,0,118,0,0,0,0,8,0,0,54,0,0,0,0,9,0,0,205,0,0,0,81,7,0,0,15,0,0,0,0,8,0,0,102,0,0,0,0,8,0,0,38,0,0,0,0,9,0,0,173,0,0,0,0,8,0,0,6,0,0,0,0,8,0,0,134,0,0,0,0,8,0,0,70,0,0,0,0,9,0,0,237,0,0,0,80,7,0,0,9,0,0,0,0,8,0,0,94,0,0,0,0,8,0,0,30,0,0,0,0,9,0,0,157,0,0,0,84,7,0,0,99,0,0,0,0,8,0,0,126,0,0,0,0,8,0,0,62,0,0,0,0,9,0,0,221,0,0,0,82,7,0,0,27,0,0,0,0,8,0,0,110,0,0,0,0,8,0,0,46,0,0,0,0,9,0,0,189,0,0,0,0,8,0,0,14,0,0,0,0,8,0,0,142,0,0,0,0,8,0,0,78,0,0,0,0,9,0,0,253,0,0,0,96,7,0,0,0,1,0,0,0,8,0,0,81,0,0,0,0,8,0,0,17,0,0,0,85,8,0,0,131,0,0,0,82,7,0,0,31,0,0,0,0,8,0,0,113,0,0,0,0,8,0,0,49,0,0,0,0,9,0,0,195,0,0,0,80,7,0,0,10,0,0,0,0,8,0,0,97,0,0,0,0,8,0,0,33,0,0,0,0,9,0,0,163,0,0,0,0,8,0,0,1,0,0,0,0,8,0,0,129,0,0,0,0,8,0,0,65,0,0,0,0,9,0,0,227,0,0,0,80,7,0,0,6,0,0,0,0,8,0,0,89,0,0,0,0,8,0,0,25,0,0,0,0,9,0,0,147,0,0,0,83,7,0,0,59,0,0,0,0,8,0,0,121,0,0,0,0,8,0,0,57,0,0,0,0,9,0,0,211,0,0,0,81,7,0,0,17,0,0,0,0,8,0,0,105,0,0,0,0,8,0,0,41,0,0,0,0,9,0,0,179,0,0,0,0,8,0,0,9,0,0,0,0,8,0,0,137,0,0,0,0,8,0,0,73,0,0,0,0,9,0,0,243,0,0,0,80,7,0,0,4,0,0,0,0,8,0,0,85,0,0,0,0,8,0,0,21,0,0,0,80,8,0,0,2,1,0,0,83,7,0,0,43,0,0,0,0,8,0,0,117,0,0,0,0,8,0,0,53,0,0,0,0,9,0,0,203,0,0,0,81,7,0,0,13,0,0,0,0,8,0,0,101,0,0,0,0,8,0,0,37,0,0,0,0,9,0,0,171,0,0,0,0,8,0,0,5,0,0,0,0,8,0,0,133,0,0,0,0,8,0,0,69,0,0,0,0,9,0,0,235,0,0,0,80,7,0,0,8,0,0,0,0,8,0,0,93,0,0,0,0,8,0,0,29,0,0,0,0,9,0,0,155,0,0,0,84,7,0,0,83,0,0,0,0,8,0,0,125,0,0,0,0,8,0,0,61,0,0,0,0,9,0,0,219,0,0,0,82,7,0,0,23,0,0,0,0,8,0,0,109,0,0,0,0,8,0,0,45,0,0,0,0,9,0,0,187,0,0,0,0,8,0,0,13,0,0,0,0,8,0,0,141,0,0,0,0,8,0,0,77,0,0,0,0,9,0,0,251,0,0,0,80,7,0,0,3,0,0,0,0,8,0,0,83,0,0,0,0,8,0,0,19,0,0,0,85,8,0,0,195,0,0,0,83,7,0,0,35,0,0,0,0,8,0,0,115,0,0,0,0,8,0,0,51,0,0,0,0,9,0,0,199,0,0,0,81,7,0,0,11,0,0,0,0,8,0,0,99,0,0,0,0,8,0,0,35,0,0,0,0,9,0,0,167,0,0,0,0,8,0,0,3,0,0,0,0,8,0,0,131,0,0,0,0,8,0,0,67,0,0,0,0,9,0,0,231,0,0,0,80,7,0,0,7,0,0,0,0,8,0,0,91,0,0,0,0,8,0,0,27,0,0,0,0,9,0,0,151,0,0,0,84,7,0,0,67,0,0,0,0,8,0,0,123,0,0,0,0,8,0,0,59,0,0,0,0,9,0,0,215,0,0,0,82,7,0,0,19,0,0,0,0,8,0,0,107,0,0,0,0,8,0,0,43,0,0,0,0,9,0,0,183,0,0,0,0,8,0,0,11,0,0,0,0,8,0,0,139,0,0,0,0,8,0,0,75,0,0,0,0,9,0,0,247,0,0,0,80,7,0,0,5,0,0,0,0,8,0,0,87,0,0,0,0,8,0,0,23,0,0,0,192,8,0,0,0,0,0,0,83,7,0,0,51,0,0,0,0,8,0,0,119,0,0,0,0,8,0,0,55,0,0,0,0,9,0,0,207,0,0,0,81,7,0,0,15,0,0,0,0,8,0,0,103,0,0,0,0,8,0,0,39,0,0,0,0,9,0,0,175,0,0,0,0,8,0,0,7,0,0,0,0,8,0,0,135,0,0,0,0,8,0,0,71,0,0,0,0,9,0,0,239,0,0,0,80,7,0,0,9,0,0,0,0,8,0,0,95,0,0,0,0,8,0,0,31,0,0,0,0,9,0,0,159,0,0,0,84,7,0,0,99,0,0,0,0,8,0,0,127,0,0,0,0,8,0,0,63,0,0,0,0,9,0,0,223,0,0,0,82,7,0,0,27,0,0,0,0,8,0,0,111,0,0,0,0,8,0,0,47,0,0,0,0,9,0,0,191,0,0,0,0,8,0,0,15,0,0,0,0,8,0,0,143,0,0,0,0,8,0,0,79,0,0,0,0,9,0,0,255,0,0,0,80,5,0,0,1,0,0,0,87,5,0,0,1,1,0,0,83,5,0,0,17,0,0,0,91,5,0,0,1,16,0,0,81,5,0,0,5,0,0,0,89,5,0,0,1,4,0,0,85,5,0,0,65,0,0,0,93,5,0,0,1,64,0,0,80,5,0,0,3,0,0,0,88,5,0,0,1,2,0,0,84,5,0,0,33,0,0,0,92,5,0,0,1,32,0,0,82,5,0,0,9,0,0,0,90,5,0,0,1,8,0,0,86,5,0,0,129,0,0,0,192,5,0,0,1,96,0,0,80,5,0,0,2,0,0,0,87,5,0,0,129,1,0,0,83,5,0,0,25,0,0,0,91,5,0,0,1,24,0,0,81,5,0,0,7,0,0,0,89,5,0,0,1,6,0,0,85,5,0,0,97,0,0,0,93,5,0,0,1,96,0,0,80,5,0,0,4,0,0,0,88,5,0,0,1,3,0,0,84,5,0,0,49,0,0,0,92,5,0,0,1,48,0,0,82,5,0,0,13,0,0,0,90,5,0,0,1,12,0,0,86,5,0,0,193,0,0,0,192,5,0,0,1,96,0,0,16,0,0,0,17,0,0,0,18,0,0,0,0,0,0,0,8,0,0,0,7,0,0,0,9,0,0,0,6,0,0,0,10,0,0,0,5,0,0,0,11,0,0,0,4,0,0,0,12,0,0,0,3,0,0,0,13,0,0,0,2,0,0,0,14,0,0,0,1,0,0,0,15,0,0,0,0,0,0,0,1,0,0,0,3,0,0,0,7,0,0,0,15,0,0,0,31,0,0,0,63,0,0,0,127,0,0,0,255,0,0,0,255,1,0,0,255,3,0,0,255,7,0,0,255,15,0,0,255,31,0,0,255,63,0,0,255,127,0,0,255,255,0,0,3,0,0,0,4,0,0,0,5,0,0,0,6,0,0,0,7,0,0,0,8,0,0,0,9,0,0,0,10,0,0,0,11,0,0,0,13,0,0,0,15,0,0,0,17,0,0,0,19,0,0,0,23,0,0,0,27,0,0,0,31,0,0,0,35,0,0,0,43,0,0,0,51,0,0,0,59,0,0,0,67,0,0,0,83,0,0,0,99,0,0,0,115,0,0,0,131,0,0,0,163,0,0,0,195,0,0,0,227,0,0,0,2,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,2,0,0,0,2,0,0,0,2,0,0,0,2,0,0,0,3,0,0,0,3,0,0,0,3,0,0,0,3,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,0,0,0,0,112,0,0,0,112,0,0,0,1,0,0,0,2,0,0,0,3,0,0,0,4,0,0,0,5,0,0,0,7,0,0,0,9,0,0,0,13,0,0,0,17,0,0,0,25,0,0,0,33,0,0,0,49,0,0,0,65,0,0,0,97,0,0,0,129,0,0,0,193,0,0,0,1,1,0,0,129,1,0,0,1,2,0,0,1,3,0,0,1,4,0,0,1,6,0,0,1,8,0,0,1,12,0,0,1,16,0,0,1,24,0,0,1,32,0,0,1,48,0,0,1,64,0,0,1,96,0,0,0,0,0,0,246,10,0,0,247,10,0,0,248,10,0,0,249,10,0,0,250,10,0,0,251,10,0,0,252,10,0,0,253,10,0,0,254,10,0,0,255,10,0,0,0,11,0,0,1,11,0,0,2,11,0,0,3,11,0,0,4,11,0,0,5,11,0,0,6,11,0,0,7,11,0,0,16,8,1,0,144,8,1,0,144,9,1,0,144,11,1,0,144,15,1,0,144,23,1,0,144,39,1,0,144,71,1,0,24,0,120,58,76,70,11,60,242,204,192,60,116,252,59,61,86,73,154,61,241,93,228,61,248,163,29,62,180,231,78,62,54,157,130,62,78,220,159,62,193,174,190,62,65,132,222,62,173,194,254,62,186,101,15,63,248,0,31,63,29,233,45,63,249,219,59,63,45,162,72,63,160,17,84,63,38,15,94,63,46,143,102,63,112,149,109,63,174,51,115,63,159,135,119,63,66,184,122,63,196,242,124,63,75,103,126,63,196,69,127,63,241,186,127,63,217,237,127,63,162,253,127,63,248,255,127,63,168,9,120,57,17,119,11,59,135,139,193,59,74,113,61,60,148,82,156,60,94,8,233,60,42,83,34,61,74,118,87,61,138,227,137,61,7,140,171,61,34,154,208,61,108,239,248,61,164,52,18,62,100,112,41,62,65,21,66,62,67,11,92,62,47,56,119,62,197,191,137,62,92,97,152,62,135,112,167,62,4,220,182,62,188,145,198,62,231,126,214,62,48,144,230,62,227,177,246,62,13,104,3,63,121,107,11,63,98,89,19,63,42,40,27,63,137,206,34,63,166,67,42,63,49,127,49,63,126,121,56,63,153,43,63,63,92,143,69,63,127,159,75,63,165,87,81,63,104,180,86,63,89,179,91,63,8,83,96,63,252,146,100,63,177,115,104,63,138,246,107,63,198,29,111,63,109,236,113,63,62,102,116,63,154,143,118,63,104,109,120,63,3,5,122,63,26,92,123,63,153,120,124,63,143,96,125,63,17,26,126,63,39,171,126,63,176,25,127,63,74,107,127,63,68,165,127,63,132,204,127,63,123,229,127,63,17,244,127,63,158,251,127,63,219,254,127,63,218,255,127,63,0,0,128,63,5,12,120,56,50,131,11,58,118,186,193,58,226,203,61,59,38,207,156,59,139,32,234,59,245,102,35,60,63,100,89,60,184,127,139,60,59,23,174,60,239,114,212,60,96,140,254,60,45,46,22,61,114,237,46,61,155,127,73,61,220,223,101,61,123,4,130,61,159,250,145,61,71,207,162,61,38,127,180,61,173,6,199,61,16,98,218,61,63,141,238,61,244,193,1,62,185,160,12,62,128,224,23,62,182,126,35,62,166,120,47,62,116,203,59,62,34,116,72,62,141,111,85,62,107,186,98,62,83,81,112,62,180,48,126,62,110,42,134,62,252,92,141,62,9,174,148,62,138,27,156,62,100,163,163,62,112,67,171,62,119,249,178,62,54,195,186,62,93,158,194,62,147,136,202,62,118,127,210,62,154,128,218,62,142,137,226,62,217,151,234,62,2,169,242,62,139,186,250,62,251,100,1,63,99,106,5,63,65,108,9,63,89,105,13,63,116,96,17,63,94,80,21,63,231,55,25,63,231,21,29,63,58,233,32,63,197,176,36,63,116,107,40,63,62,24,44,63,35,182,47,63,43,68,51,63,109,193,54,63,10,45,58,63,48,134,61,63,26,204,64,63,17,254,67,63,107,27,71,63,142,35,74,63,238,21,77,63,15,242,79,63,132,183,82,63,239,101,85,63,3,253,87,63,129,124,90,63,60,228,92,63,21,52,95,63,254,107,97,63,246,139,99,63,14,148,101,63,98,132,103,63,33,93,105,63,133,30,107,63,213,200,108,63,103,92,110,63,155,217,111,63,224,64,113,63,172,146,114,63,131,207,115,63,241,247,116,63,139,12,118,63,239,13,119,63,193,252,119,63,172,217,120,63,99,165,121,63,155,96,122,63,15,12,123,63,124,168,123,63,163,54,124,63,71,183,124,63,41,43,125,63,13,147,125,63,183,239,125,63,229,65,126,63,89,138,126,63,205,201,126,63,251,0,127,63,150,48,127,63,78,89,127,63,205,123,127,63,182,152,127,63,167,176,127,63,53,196,127,63,239,211,127,63,91,224,127,63,245,233,127,63,51,241,127,63,127,246,127,63,59,250,127,63,190,252,127,63,84,254,127,63,64,255,127,63,186,255,127,63,238,255,127,63,254,255,127,63,0,0,128,63,169,12,120,55,54,134,11,57,38,198,193,57,94,226,61,58,234,237,156,58,85,101,234,58,56,170,35,59,207,219,89,59,169,226,139,59,42,178,174,59,13,91,213,59,204,219,255,59,91,25,23,60,250,46,48,60,194,45,75,60,156,20,104,60,46,113,131,60,225,202,147,60,185,22,165,60,1,84,183,60,245,129,202,60,198,159,222,60,155,172,243,60,199,211,4,61,213,71,16,61,250,49,28,61,174,145,40,61,101,102,53,61,141,175,66,61,140,108,80,61,193,156,94,61,133,63,109,61,41,84,124,61,252,236,133,61,26,232,141,61,13,27,150,61,110,133,158,61,212,38,167,61,210,254,175,61,245,12,185,61,200,80,194,61,209,201,203,61,146,119,213,61,139,89,223,61,51,111,233,61,2,184,243,61,105,51,254,61,106,112,4,62,214,223,9,62,171,103,15,62,153,7,21,62,77,191,26,62,116,142,32,62,181,116,38,62,184,113,44,62,34,133,50,62,149,174,56,62,178,237,62,62,21,66,69,62,92,171,75,62,30,41,82,62,243,186,88,62,112,96,95,62,40,25,102,62,170,228,108,62,132,194,115,62,68,178,122,62,185,217,128,62,203,98,132,62,26,244,135,62,105,141,139,62,120,46,143,62,6,215,146,62,211,134,150,62,156,61,154,62,29,251,157,62,19,191,161,62,57,137,165,62,71,89,169,62,249,46,173,62,5,10,177,62,36,234,180,62,13,207,184,62,117,184,188,62,18,166,192,62,153,151,196,62,190,140,200,62,52,133,204,62,175,128,208,62,225,126,212,62,125,127,216,62,52,130,220,62,184,134,224,62,185,140,228,62,233,147,232,62,248,155,236,62,150,164,240,62,117,173,244,62,67,182,248,62,178,190,252,62,57,99,0,63,153,102,2,63,82,105,4,63,60,107,6,63,48,108,8,63,6,108,10,63,151,106,12,63,188,103,14,63,78,99,16,63,39,93,18,63,33,85,20,63,21,75,22,63,222,62,24,63,87,48,26,63,92,31,28,63,199,11,30,63,117,245,31,63,66,220,33,63,12,192,35,63,176,160,37,63,12,126,39,63,254,87,41,63,104,46,43,63,39,1,45,63,29,208,46,63,43,155,48,63,51,98,50,63,23,37,52,63,188,227,53,63,4,158,55,63,214,83,57,63,23,5,59,63,173,177,60,63,128,89,62,63,120,252,63,63,126,154,65,63,124,51,67,63,93,199,68,63,12,86,70,63,119,223,71,63,138,99,73,63,54,226,74,63,104,91,76,63,17,207,77,63,35,61,79,63,145,165,80,63,76,8,82,63,75,101,83,63,130,188,84,63,231,13,86,63,114,89,87,63,26,159,88,63,218,222,89,63,172,24,91,63,138,76,92,63,113,122,93,63,93,162,94,63,78,196,95,63,67,224,96,63,58,246,97,63,54,6,99,63,56,16,100,63,67,20,101,63,92,18,102,63,133,10,103,63,198,252,103,63,37,233,104,63,168,207,105,63,89,176,106,63,64,139,107,63,102,96,108,63,216,47,109,63,159,249,109,63,201,189,110,63,97,124,111,63,118,53,112,63,23,233,112,63,81,151,113,63,53,64,114,63,212,227,114,63,61,130,115,63,131,27,116,63,184,175,116,63,238,62,117,63,56,201,117,63,171,78,118,63,90,207,118,63,90,75,119,63,192,194,119,63,162,53,120,63,21,164,120,63,48,14,121,63,8,116,121,63,182,213,121,63,79,51,122,63,235,140,122,63,162,226,122,63,139,52,123,63,191,130,123,63,85,205,123,63,102,20,124,63,9,88,124,63,88,152,124,63,106,213,124,63,88,15,125,63,58,70,125,63,41,122,125,63,62,171,125,63,143,217,125,63,54,5,126,63,75,46,126,63,228,84,126,63,27,121,126,63,7,155,126,63,190,186,126,63,88,216,126,63,236,243,126,63,144,13,127,63,91,37,127,63,99,59,127,63,188,79,127,63,125,98,127,63,185,115,127,63,135,131,127,63,249,145,127,63,36,159,127,63,26,171,127,63,238,181,127,63,179,191,127,63,122,200,127,63,85,208,127,63,84,215,127,63,136,221,127,63,0,227,127,63,204,231,127,63,249,235,127,63,150,239,127,63,177,242,127,63,85,245,127,63,144,247,127,63,109,249,127,63,246,250,127,63,54,252,127,63,55,253,127,63,1,254,127,63,156,254,127,63,18,255,127,63,103,255,127,63,163,255,127,63,204,255,127,63,229,255,127,63,244,255,127,63,252,255,127,63,255,255,127,63,0,0,128,63,0,0,128,63,60,12,120,54,253,134,11,56,19,201,193,56,248,231,61,57,148,245,156,57,115,118,234,57,238,186,35,58,113,249,89,58,32,251,139,58,96,216,174,58,34,148,213,58,3,23,0,59,209,82,23,59,65,125,48,59,21,150,75,59,8,157,104,59,233,200,131,59,20,58,148,59,218,161,165,59,16,0,184,59,136,84,203,59,16,159,223,59,118,223,244,59,194,138,5,60,128,32,17,60,217,48,29,60,172,187,41,60,219,192,54,60,67,64,68,60,194,57,82,60,52,173,96,60,115,154,111,60,88,1,127,60,222,112,135,60,186,157,143,60,42,7,152,60,25,173,160,60,112,143,169,60,23,174,178,60,246,8,188,60,243,159,197,60,245,114,207,60,225,129,217,60,156,204,227,60,10,83,238,60,14,21,249,60,70,9,2,61,177,165,7,61,187,95,13,61,81,55,19,61,102,44,25,61,230,62,31,61,195,110,37,61,233,187,43,61,71,38,50,61,202,173,56,61,97,82,63,61,247,19,70,61,121,242,76,61,210,237,83,61,240,5,91,61,187,58,98,61,32,140,105,61,8,250,112,61,93,132,120,61,132,21,128,61,249,246,131,61,130,230,135,61,19,228,139,61,159,239,143,61,26,9,148,61,119,48,152,61,169,101,156,61,163,168,160,61,88,249,164,61,186,87,169,61,186,195,173,61,76,61,178,61,95,196,182,61,230,88,187,61,209,250,191,61,18,170,196,61,152,102,201,61,85,48,206,61,56,7,211,61,48,235,215,61,47,220,220,61,34,218,225,61,248,228,230,61,161,252,235,61,11,33,241,61,35,82,246,61,217,143,251,61,13,109,0,62,105,24,3,62,247,201,5,62,174,129,8,62,133,63,11,62,113,3,14,62,104,205,16,62,96,157,19,62,79,115,22,62,42,79,25,62,232,48,28,62,124,24,31,62,221,5,34,62,255,248,36,62,215,241,39,62,90,240,42,62,125,244,45,62,51,254,48,62,114,13,52,62,45,34,55,62,88,60,58,62,232,91,61,62,208,128,64,62,3,171,67,62,118,218,70,62,26,15,74,62,229,72,77,62,199,135,80,62,181,203,83,62,162,20,87,62,127,98,90,62,63,181,93,62,213,12,97,62,50,105,100,62,73,202,103,62,12,48,107,62,108,154,110,62,92,9,114,62,203,124,117,62,173,244,120,62,241,112,124,62,138,241,127,62,52,187,129,62,190,127,131,62,91,70,133,62,4,15,135,62,176,217,136,62,89,166,138,62,245,116,140,62,126,69,142,62,234,23,144,62,50,236,145,62,78,194,147,62,54,154,149,62,224,115,151,62,70,79,153,62,93,44,155,62,31,11,157,62,130,235,158,62,127,205,160,62,11,177,162,62,31,150,164,62,177,124,166,62,186,100,168,62,47,78,170,62,9,57,172,62,62,37,174,62,198,18,176,62,150,1,178,62,167,241,179,62,238,226,181,62,100,213,183,62,254,200,185,62,179,189,187,62,122,179,189,62,74,170,191,62,25,162,193,62,221,154,195,62,142,148,197,62,34,143,199,62,142,138,201,62,203,134,203,62,205,131,205,62,140,129,207,62,253,127,209,62,24,127,211,62,210,126,213,62,33,127,215,62,252,127,217,62,88,129,219,62,45,131,221,62,112,133,223,62,23,136,225,62,25,139,227,62,108,142,229,62,5,146,231,62,219,149,233,62,228,153,235,62,21,158,237,62,102,162,239,62,203,166,241,62,59,171,243,62,173,175,245,62,21,180,247,62,107,184,249,62,164,188,251,62,181,192,253,62,150,196,255,62,30,228,0,63,207,229,1,63,88,231,2,63,182,232,3,63,226,233,4,63,215,234,5,63,146,235,6,63,12,236,7,63,66,236,8,63,45,236,9,63,202,235,10,63,19,235,11,63,4,234,12,63,151,232,13,63,200,230,14,63,145,228,15,63,239,225,16,63,220,222,17,63,84,219,18,63,81,215,19,63,208,210,20,63,202,205,21,63,61,200,22,63,34,194,23,63,117,187,24,63,50,180,25,63,85,172,26,63,215,163,27,63,182,154,28,63,236,144,29,63,117,134,30,63,77,123,31,63,110,111,32,63,214,98,33,63,126,85,34,63,100,71,35,63,130,56,36,63,212,40,37,63,87,24,38,63,5,7,39,63,219,244,39,63,213,225,40,63,239,205,41,63,36,185,42,63,113,163,43,63,209,140,44,63,64,117,45,63,188,92,46,63,63,67,47,63,199,40,48,63,78,13,49,63,211,240,49,63,80,211,50,63,195,180,51,63,39,149,52,63,122,116,53,63,184,82,54,63,220,47,55,63,229,11,56,63,206,230,56,63,149,192,57,63,54,153,58,63,174,112,59,63,249,70,60,63,21,28,61,63,255,239,61,63,179,194,62,63,48,148,63,63,113,100,64,63,116,51,65,63,55,1,66,63,182,205,66,63,239,152,67,63,224,98,68,63,134,43,69,63,222,242,69,63,230,184,70,63,156,125,71,63,253,64,72,63,7,3,73,63,184,195,73,63,14,131,74,63,6,65,75,63,159,253,75,63,215,184,76,63,172,114,77,63,28,43,78,63,38,226,78,63,199,151,79,63,253,75,80,63,201,254,80,63,39,176,81,63,22,96,82,63,150,14,83,63,164,187,83,63,63,103,84,63,103,17,85,63,26,186,85,63,86,97,86,63,28,7,87,63,105,171,87,63,62,78,88,63,152,239,88,63,120,143,89,63,221,45,90,63,198,202,90,63,50,102,91,63,33,0,92,63,147,152,92,63,134,47,93,63,251,196,93,63,242,88,94,63,105,235,94,63,98,124,95,63,219,11,96,63,213,153,96,63,80,38,97,63,76,177,97,63,201,58,98,63,199,194,98,63,70,73,99,63,71,206,99,63,202,81,100,63,208,211,100,63,88,84,101,63,100,211,101,63,244,80,102,63,9,205,102,63,163,71,103,63,195,192,103,63,107,56,104,63,154,174,104,63,82,35,105,63,147,150,105,63,96,8,106,63,184,120,106,63,157,231,106,63,16,85,107,63,19,193,107,63,166,43,108,63,203,148,108,63,132,252,108,63,209,98,109,63,180,199,109,63,48,43,110,63,68,141,110,63,244,237,110,63,64,77,111,63,42,171,111,63,181,7,112,63,225,98,112,63,177,188,112,63,38,21,113,63,67,108,113,63,10,194,113,63,123,22,114,63,155,105,114,63,106,187,114,63,234,11,115,63,31,91,115,63,9,169,115,63,172,245,115,63,9,65,116,63,35,139,116,63,252,211,116,63,151,27,117,63,245,97,117,63,26,167,117,63,8,235,117,63,193,45,118,63,72,111,118,63,159,175,118,63,202,238,118,63,201,44,119,63,161,105,119,63,84,165,119,63,228,223,119,63,85,25,120,63,168,81,120,63,226,136,120,63,3,191,120,63,16,244,120,63,11,40,121,63,247,90,121,63,215,140,121,63,173,189,121,63,125,237,121,63,73,28,122,63,20,74,122,63,226,118,122,63,181,162,122,63,144,205,122,63,118,247,122,63,107,32,123,63,112,72,123,63,138,111,123,63,186,149,123,63,5,187,123,63,109,223,123,63,245,2,124,63,160,37,124,63,113,71,124,63,108,104,124,63,147,136,124,63,233,167,124,63,114,198,124,63,48,228,124,63,38,1,125,63,89,29,125,63,201,56,125,63,124,83,125,63,115,109,125,63,178,134,125,63,60,159,125,63,19,183,125,63,60,206,125,63,184,228,125,63,139,250,125,63,184,15,126,63,66,36,126,63,44,56,126,63,120,75,126,63,43,94,126,63,70,112,126,63,204,129,126,63,194,146,126,63,41,163,126,63,4,179,126,63,86,194,126,63,35,209,126,63,109,223,126,63,55,237,126,63,131,250,126,63,85,7,127,63,175,19,127,63,148,31,127,63,7,43,127,63,10,54,127,63,160,64,127,63,205,74,127,63,146,84,127,63,242,93,127,63,239,102,127,63,141,111,127,63,206,119,127,63,181,127,127,63,67,135,127,63,124,142,127,63,98,149,127,63,247,155,127,63,61,162,127,63,56,168,127,63,233,173,127,63,83,179,127,63,120,184,127,63,90,189,127,63,252,193,127,63,95,198,127,63,134,202,127,63,116,206,127,63,41,210,127,63,168,213,127,63,244,216,127,63,13,220,127,63,247,222,127,63,179,225,127,63,67,228,127,63,168,230,127,63,229,232,127,63,252,234,127,63,237,236,127,63,188,238,127,63,105,240,127,63,246,241,127,63,101,243,127,63,183,244,127,63,238,245,127,63,11,247,127,63,16,248,127,63,254,248,127,63,214,249,127,63,155,250,127,63,76,251,127,63,236,251,127,63,124,252,127,63,252,252,127,63,110,253,127,63,211,253,127,63,44,254,127,63,121,254,127,63,189,254,127,63,247,254,127,63,42,255,127,63,84,255,127,63,120,255,127,63,150,255,127,63,175,255,127,63,195,255,127,63,211,255,127,63,224,255,127,63,234,255,127,63,241,255,127,63,246,255,127,63,250,255,127,63,253,255,127,63,254,255,127,63,255,255,127,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,171,15,120,53,24,135,11,55,225,201,193,55,107,233,61,56,128,247,156,56,187,122,234,56,24,191,35,57,213,0,90,57,56,1,140,57,229,225,174,57,88,162,213,57,60,33,0,58,24,97,23,58,175,144,48,58,243,175,75,58,212,190,104,58,159,222,131,58,143,85,148,58,48,196,165,58,119,42,184,58,90,136,203,58,204,221,223,58,191,42,245,58,148,183,5,59,124,85,17,59,16,111,29,59,73,4,42,59,31,21,55,59,138,161,68,59,129,169,82,59,252,44,97,59,241,43,112,59,88,166,127,59,19,206,135,59,169,6,144,59,233,124,152,59,204,48,161,59,79,34,170,59,106,81,179,59,26,190,188,59,86,104,198,59,26,80,208,59,95,117,218,59,31,216,228,59,83,120,239,59,244,85,250,59,126,184,2,60,177,100,8,60,145,47,14,60,25,25,20,60,70,33,26,60,19,72,32,60,126,141,38,60,129,241,44,60,25,116,51,60,65,21,58,60,246,212,64,60,50,179,71,60,243,175,78,60,50,203,85,60,235,4,93,60,26,93,100,60,186,211,107,60,198,104,115,60,58,28,123,60,7,119,129,60,33,111,133,60,102,118,137,60,212,140,141,60,105,178,145,60,33,231,149,60,251,42,154,60,243,125,158,60,6,224,162,60,50,81,167,60,115,209,171,60,199,96,176,60,43,255,180,60,154,172,185,60,19,105,190,60,146,52,195,60,20,15,200,60,149,248,204,60,19,241,209,60,137,248,214,60,245,14,220,60,83,52,225,60,160,104,230,60,215,171,235,60,246,253,240,60,249,94,246,60,220,206,251,60,205,166,0,61,153,109,3,61,207,59,6,61,109,17,9,61,114,238,11,61,220,210,14,61,167,190,17,61,211,177,20,61,94,172,23,61,68,174,26,61,133,183,29,61,30,200,32,61,12,224,35,61,78,255,38,61,225,37,42,61,196,83,45,61,243,136,48,61,109,197,51,61,47,9,55,61,55,84,58,61,130,166,61,61,15,0,65,61,218,96,68,61,226,200,71,61,35,56,75,61,156,174,78,61,73,44,82,61,40,177,85,61,55,61,89,61,115,208,92,61,217,106,96,61,103,12,100,61,25,181,103,61,238,100,107,61,227,27,111,61,244,217,114,61,30,159,118,61,96,107,122,61,182,62,126,61,143,12,129,61,73,253,130,61,138,241,132,61,79,233,134,61,150,228,136,61,94,227,138,61,167,229,140,61,109,235,142,61,175,244,144,61,109,1,147,61,164,17,149,61,83,37,151,61,120,60,153,61,17,87,155,61,30,117,157,61,155,150,159,61,136,187,161,61,226,227,163,61,169,15,166,61,218,62,168,61,116,113,170,61,116,167,172,61,218,224,174,61,162,29,177,61,205,93,179,61,87,161,181,61,62,232,183,61,130,50,186,61,32,128,188,61,22,209,190,61,98,37,193,61,2,125,195,61,245,215,197,61,57,54,200,61,203,151,202,61,169,252,204,61,211,100,207,61,68,208,209,61,252,62,212,61,249,176,214,61,56,38,217,61,184,158,219,61,117,26,222,61,111,153,224,61,163,27,227,61,14,161,229,61,175,41,232,61,132,181,234,61,138,68,237,61,191,214,239,61,33,108,242,61,174,4,245,61,99,160,247,61,62,63,250,61,61,225,252,61,93,134,255,61,78,23,1,62,252,108,2,62,56,196,3,62,255,28,5,62,81,119,6,62,45,211,7,62,145,48,9,62,125,143,10,62,238,239,11,62,228,81,13,62,94,181,14,62,89,26,16,62,214,128,17,62,210,232,18,62,77,82,20,62,69,189,21,62,184,41,23,62,166,151,24,62,13,7,26,62,236,119,27,62,65,234,28,62,11,94,30,62,73,211,31,62,250,73,33,62,28,194,34,62,173,59,36,62,172,182,37,62,24,51,39,62,240,176,40,62,50,48,42,62,220,176,43,62,238,50,45,62,101,182,46,62,64,59,48,62,126,193,49,62,30,73,51,62,29,210,52,62,123,92,54,62,54,232,55,62,76,117,57,62,187,3,59,62,131,147,60,62,162,36,62,62,22,183,63,62,222,74,65,62,248,223,66,62,98,118,68,62,28,14,70,62,35,167,71,62,117,65,73,62,18,221,74,62,247,121,76,62,35,24,78,62,149,183,79,62,74,88,81,62,66,250,82,62,121,157,84,62,240,65,86,62,163,231,87,62,146,142,89,62,186,54,91,62,26,224,92,62,177,138,94,62,124,54,96,62,122,227,97,62,169,145,99,62,7,65,101,62,147,241,102,62,75,163,104,62,44,86,106,62,54,10,108,62,102,191,109,62,187,117,111,62,51,45,113,62,204,229,114,62,132,159,116,62,90,90,118,62,75,22,120,62,85,211,121,62,120,145,123,62,176,80,125,62,253,16,127,62,46,105,128,62,101,74,129,62,36,44,130,62,105,14,131,62,52,241,131,62,130,212,132,62,84,184,133,62,169,156,134,62,127,129,135,62,213,102,136,62,171,76,137,62,255,50,138,62,209,25,139,62,32,1,140,62,233,232,140,62,46,209,141,62,236,185,142,62,34,163,143,62,208,140,144,62,244,118,145,62,142,97,146,62,156,76,147,62,29,56,148,62,17,36,149,62,118,16,150,62,76,253,150,62,144,234,151,62,67,216,152,62,99,198,153,62,239,180,154,62,230,163,155,62,71,147,156,62,17,131,157,62,67,115,158,62,219,99,159,62,218,84,160,62,60,70,161,62,3,56,162,62,43,42,163,62,181,28,164,62,160,15,165,62,233,2,166,62,145,246,166,62,149,234,167,62,245,222,168,62,176,211,169,62,197,200,170,62,50,190,171,62,246,179,172,62,17,170,173,62,129,160,174,62,69,151,175,62,91,142,176,62,196,133,177,62,125,125,178,62,133,117,179,62,220,109,180,62,128,102,181,62,112,95,182,62,171,88,183,62,47,82,184,62,252,75,185,62,17,70,186,62,108,64,187,62,11,59,188,62,239,53,189,62,22,49,190,62,126,44,191,62,38,40,192,62,13,36,193,62,51,32,194,62,150,28,195,62,52,25,196,62,12,22,197,62,30,19,198,62,104,16,199,62,233,13,200,62,159,11,201,62,138,9,202,62,169,7,203,62,249,5,204,62,123,4,205,62,44,3,206,62,11,2,207,62,24,1,208,62,81,0,209,62,181,255,209,62,66,255,210,62,248,254,211,62,213,254,212,62,216,254,213,62,255,254,214,62,75,255,215,62,184,255,216,62,71,0,218,62,245,0,219,62,195,1,220,62,173,2,221,62,180,3,222,62,214,4,223,62,17,6,224,62,101,7,225,62,208,8,226,62,81,10,227,62,231,11,228,62,144,13,229,62,76,15,230,62,25,17,231,62,245,18,232,62,224,20,233,62,217,22,234,62,221,24,235,62,236,26,236,62,5,29,237,62,39,31,238,62,79,33,239,62,125,35,240,62,176,37,241,62,230,39,242,62,31,42,243,62,88,44,244,62,145,46,245,62,200,48,246,62,253,50,247,62,45,53,248,62,88,55,249,62,124,57,250,62,153,59,251,62,172,61,252,62,181,63,253,62,179,65,254,62,163,67,255,62,195,34,0,63,173,163,0,63,142,36,1,63,102,165,1,63,53,38,2,63,250,166,2,63,180,39,3,63,99,168,3,63,5,41,4,63,155,169,4,63,36,42,5,63,159,170,5,63,12,43,6,63,105,171,6,63,183,43,7,63,244,171,7,63,32,44,8,63,59,172,8,63,68,44,9,63,58,172,9,63,28,44,10,63,235,171,10,63,164,43,11,63,73,171,11,63,216,42,12,63,80,170,12,63,177,41,13,63,251,168,13,63,44,40,14,63,69,167,14,63,68,38,15,63,41,165,15,63,243,35,16,63,162,162,16,63,53,33,17,63,172,159,17,63,5,30,18,63,65,156,18,63,95,26,19,63,94,152,19,63,61,22,20,63,252,147,20,63,155,17,21,63,24,143,21,63,116,12,22,63,173,137,22,63,195,6,23,63,182,131,23,63,133,0,24,63,46,125,24,63,179,249,24,63,18,118,25,63,74,242,25,63,91,110,26,63,69,234,26,63,6,102,27,63,159,225,27,63,14,93,28,63,84,216,28,63,111,83,29,63,95,206,29,63,36,73,30,63,188,195,30,63,40,62,31,63,102,184,31,63,119,50,32,63,90,172,32,63,14,38,33,63,146,159,33,63,230,24,34,63,10,146,34,63,253,10,35,63,190,131,35,63,77,252,35,63,169,116,36,63,211,236,36,63,200,100,37,63,138,220,37,63,22,84,38,63,110,203,38,63,143,66,39,63,122,185,39,63,47,48,40,63,172,166,40,63,241,28,41,63,254,146,41,63,210,8,42,63,108,126,42,63,205,243,42,63,243,104,43,63,223,221,43,63,143,82,44,63,3,199,44,63,59,59,45,63,54,175,45,63,244,34,46,63,116,150,46,63,182,9,47,63,185,124,47,63,125,239,47,63,1,98,48,63,69,212,48,63,72,70,49,63,10,184,49,63,139,41,50,63,202,154,50,63,198,11,51,63,127,124,51,63,246,236,51,63,40,93,52,63,22,205,52,63,191,60,53,63,36,172,53,63,66,27,54,63,27,138,54,63,174,248,54,63,249,102,55,63,254,212,55,63,187,66,56,63,47,176,56,63,91,29,57,63,63,138,57,63,217,246,57,63,41,99,58,63,48,207,58,63,236,58,59,63,93,166,59,63,130,17,60,63,93,124,60,63,235,230,60,63,44,81,61,63,33,187,61,63,201,36,62,63,35,142,62,63,48,247,62,63,238,95,63,63,94,200,63,63,126,48,64,63,80,152,64,63,209,255,64,63,3,103,65,63,228,205,65,63,117,52,66,63,181,154,66,63,163,0,67,63,64,102,67,63,139,203,67,63,131,48,68,63,41,149,68,63,124,249,68,63,123,93,69,63,39,193,69,63,127,36,70,63,132,135,70,63,51,234,70,63,142,76,71,63,148,174,71,63,68,16,72,63,159,113,72,63,164,210,72,63,83,51,73,63,172,147,73,63,174,243,73,63,89,83,74,63,173,178,74,63,169,17,75,63,77,112,75,63,154,206,75,63,143,44,76,63,43,138,76,63,110,231,76,63,89,68,77,63,234,160,77,63,34,253,77,63,0,89,78,63,133,180,78,63,176,15,79,63,128,106,79,63,246,196,79,63,18,31,80,63,210,120,80,63,56,210,80,63,66,43,81,63,242,131,81,63,69,220,81,63,61,52,82,63,217,139,82,63,24,227,82,63,252,57,83,63,131,144,83,63,174,230,83,63,123,60,84,63,236,145,84,63,0,231,84,63,183,59,85,63,16,144,85,63,12,228,85,63,170,55,86,63,235,138,86,63,206,221,86,63,83,48,87,63,121,130,87,63,66,212,87,63,172,37,88,63,184,118,88,63,101,199,88,63,180,23,89,63,164,103,89,63,53,183,89,63,104,6,90,63,59,85,90,63,175,163,90,63,197,241,90,63,123,63,91,63,210,140,91,63,201,217,91,63,97,38,92,63,154,114,92,63,115,190,92,63,237,9,93,63,7,85,93,63,194,159,93,63,29,234,93,63,24,52,94,63,179,125,94,63,239,198,94,63,203,15,95,63,72,88,95,63,100,160,95,63,33,232,95,63,126,47,96,63,123,118,96,63,24,189,96,63,85,3,97,63,51,73,97,63,177,142,97,63,207,211,97,63,141,24,98,63,236,92,98,63,235,160,98,63,138,228,98,63,202,39,99,63,170,106,99,63,42,173,99,63,75,239,99,63,13,49,100,63,111,114,100,63,114,179,100,63,21,244,100,63,90,52,101,63,63,116,101,63,197,179,101,63,236,242,101,63,180,49,102,63,29,112,102,63,39,174,102,63,211,235,102,63,32,41,103,63,15,102,103,63,159,162,103,63,209,222,103,63,164,26,104,63,26,86,104,63,49,145,104,63,235,203,104,63,71,6,105,63,69,64,105,63,230,121,105,63,42,179,105,63,16,236,105,63,153,36,106,63,197,92,106,63,148,148,106,63,7,204,106,63,29,3,107,63,214,57,107,63,52,112,107,63,53,166,107,63,218,219,107,63,36,17,108,63,18,70,108,63,164,122,108,63,220,174,108,63,184,226,108,63,57,22,109,63,96,73,109,63,44,124,109,63,157,174,109,63,181,224,109,63,115,18,110,63,214,67,110,63,225,116,110,63,146,165,110,63,233,213,110,63,232,5,111,63,142,53,111,63,219,100,111,63,209,147,111,63,110,194,111,63,179,240,111,63,160,30,112,63,54,76,112,63,117,121,112,63,93,166,112,63,239,210,112,63,41,255,112,63,14,43,113,63,156,86,113,63,213,129,113,63,184,172,113,63,70,215,113,63,127,1,114,63,99,43,114,63,243,84,114,63,46,126,114,63,21,167,114,63,169,207,114,63,233,247,114,63,214,31,115,63,113,71,115,63,184,110,115,63,173,149,115,63,80,188,115,63,162,226,115,63,161,8,116,63,80,46,116,63,174,83,116,63,187,120,116,63,119,157,116,63,228,193,116,63,1,230,116,63,206,9,117,63,76,45,117,63,123,80,117,63,92,115,117,63,238,149,117,63,51,184,117,63,42,218,117,63,211,251,117,63,48,29,118,63,64,62,118,63,3,95,118,63,122,127,118,63,166,159,118,63,134,191,118,63,27,223,118,63,101,254,118,63,101,29,119,63,27,60,119,63,135,90,119,63,169,120,119,63,131,150,119,63,19,180,119,63,91,209,119,63,91,238,119,63,20,11,120,63,132,39,120,63,174,67,120,63,145,95,120,63,46,123,120,63,132,150,120,63,149,177,120,63,96,204,120,63,231,230,120,63,41,1,121,63,38,27,121,63,223,52,121,63,85,78,121,63,136,103,121,63,120,128,121,63,37,153,121,63,144,177,121,63,185,201,121,63,161,225,121,63,72,249,121,63,174,16,122,63,212,39,122,63,185,62,122,63,96,85,122,63,198,107,122,63,238,129,122,63,216,151,122,63,131,173,122,63,241,194,122,63,33,216,122,63,20,237,122,63,202,1,123,63,68,22,123,63,130,42,123,63,133,62,123,63,77,82,123,63,217,101,123,63,43,121,123,63,68,140,123,63,34,159,123,63,200,177,123,63,52,196,123,63,104,214,123,63,99,232,123,63,39,250,123,63,180,11,124,63,9,29,124,63,40,46,124,63,17,63,124,63,196,79,124,63,65,96,124,63,137,112,124,63,156,128,124,63,124,144,124,63,39,160,124,63,158,175,124,63,226,190,124,63,244,205,124,63,211,220,124,63,128,235,124,63,251,249,124,63,69,8,125,63,94,22,125,63,71,36,125,63,255,49,125,63,136,63,125,63,225,76,125,63,11,90,125,63,7,103,125,63,212,115,125,63,115,128,125,63,229,140,125,63,42,153,125,63,66,165,125,63,46,177,125,63,238,188,125,63,130,200,125,63,235,211,125,63,41,223,125,63,61,234,125,63,38,245,125,63,230,255,125,63,124,10,126,63,234,20,126,63,47,31,126,63,75,41,126,63,64,51,126,63,13,61,126,63,180,70,126,63,51,80,126,63,140,89,126,63,191,98,126,63,205,107,126,63,181,116,126,63,120,125,126,63,23,134,126,63,146,142,126,63,233,150,126,63,28,159,126,63,44,167,126,63,26,175,126,63,229,182,126,63,142,190,126,63,22,198,126,63,124,205,126,63,194,212,126,63,231,219,126,63,235,226,126,63,208,233,126,63,149,240,126,63,59,247,126,63,195,253,126,63,44,4,127,63,118,10,127,63,163,16,127,63,179,22,127,63,165,28,127,63,123,34,127,63,52,40,127,63,210,45,127,63,83,51,127,63,186,56,127,63,5,62,127,63,53,67,127,63,75,72,127,63,72,77,127,63,42,82,127,63,243,86,127,63,163,91,127,63,58,96,127,63,185,100,127,63,32,105,127,63,111,109,127,63,166,113,127,63,199,117,127,63,208,121,127,63,196,125,127,63,161,129,127,63,104,133,127,63,25,137,127,63,182,140,127,63,61,144,127,63,176,147,127,63,14,151,127,63,89,154,127,63,143,157,127,63,179,160,127,63,195,163,127,63,192,166,127,63,171,169,127,63,132,172,127,63,74,175,127,63,255,177,127,63,163,180,127,63,53,183,127,63,183,185,127,63,40,188,127,63,137,190,127,63,217,192,127,63,26,195,127,63,76,197,127,63,111,199,127,63,130,201,127,63,135,203,127,63,126,205,127,63,102,207,127,63,65,209,127,63,14,211,127,63,205,212,127,63,128,214,127,63,38,216,127,63,191,217,127,63,76,219,127,63,204,220,127,63,65,222,127,63,170,223,127,63,8,225,127,63,91,226,127,63,163,227,127,63,224,228,127,63,19,230,127,63,59,231,127,63,90,232,127,63,110,233,127,63,122,234,127,63,124,235,127,63,116,236,127,63,100,237,127,63,75,238,127,63,42,239,127,63,1,240,127,63,207,240,127,63,149,241,127,63,84,242,127,63,12,243,127,63,188,243,127,63,101,244,127,63,7,245,127,63,162,245,127,63,55,246,127,63,198,246,127,63,78,247,127,63,209,247,127,63,77,248,127,63,196,248,127,63,54,249,127,63,162,249,127,63,9,250,127,63,108,250,127,63,201,250,127,63,34,251,127,63,118,251,127,63,198,251,127,63,18,252,127,63,89,252,127,63,157,252,127,63,221,252,127,63,26,253,127,63,83,253,127,63,136,253,127,63,187,253,127,63,234,253,127,63,22,254,127,63,64,254,127,63,103,254,127,63,139,254,127,63,173,254,127,63,204,254,127,63,234,254,127,63,5,255,127,63,30,255,127,63,53,255,127,63,74,255,127,63,94,255,127,63,112,255,127,63,128,255,127,63,143,255,127,63,157,255,127,63,169,255,127,63,180,255,127,63,191,255,127,63,200,255,127,63,208,255,127,63,215,255,127,63,221,255,127,63,227,255,127,63,232,255,127,63,236,255,127,63,239,255,127,63,243,255,127,63,245,255,127,63,248,255,127,63,249,255,127,63,251,255,127,63,252,255,127,63,253,255,127,63,254,255,127,63,255,255,127,63,255,255,127,63,255,255,127,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,204,8,120,52,171,134,11,54,79,202,193,54,190,233,61,55,238,247,156,55,192,123,234,55,43,192,35,56,161,2,90,56,189,2,140,56,76,228,174,56,227,165,213,56,199,35,0,57,168,100,23,57,134,149,48,57,104,182,75,57,64,199,104,57,7,228,131,57,105,92,148,57,191,204,165,57,6,53,184,57,65,149,203,57,105,237,223,57,120,61,245,57,184,194,5,58,166,98,17,58,134,126,29,58,81,22,42,58,9,42,55,58,172,185,68,58,54,197,82,58,165,76,97,58,250,79,112,58,47,207,127,58,34,229,135,58,154,32,144,58,255,153,152,58,80,81,161,58,139,70,170,58,174,121,179,58,186,234,188,58,171,153,198,58,129,134,208,58,58,177,218,58,212,25,229,58,79,192,239,58,167,164,250,58,109,227,2,59,117,147,8,59,105,98,14,59,73,80,20,59,19,93,26,59,199,136,32,59,100,211,38,59,232,60,45,59,83,197,51,59,164,108,58,59,218,50,65,59,243,23,72,59,239,27,79,59,204,62,86,59,138,128,93,59,38,225,100,59,161,96,108,59,249,254,115,59,45,188,123,59,29,204,129,59,145,201,133,59,113,214,137,59,188,242,141,59,113,30,146,59,145,89,150,59,26,164,154,59,12,254,158,59,102,103,163,59,40,224,167,59,80,104,172,59,222,255,176,59,209,166,181,59,40,93,186,59,228,34,191,59,2,248,195,59,131,220,200,59,101,208,205,59,168,211,210,59,74,230,215,59,76,8,221,59,172,57,226,59,105,122,231,59,131,202,236,59,249,41,242,59,202,152,247,59,245,22,253,59,60,82,1,60,170,32,4,60,196,246,6,60,137,212,9,60,249,185,12,60,19,167,15,60,216,155,18,60,69,152,21,60,92,156,24,60,26,168,27,60,129,187,30,60,143,214,33,60,69,249,36,60,160,35,40,60,162,85,43,60,73,143,46,60,149,208,49,60,133,25,53,60,26,106,56,60,81,194,59,60,44,34,63,60,168,137,66,60,199,248,69,60,134,111,73,60,230,237,76,60,231,115,80,60,134,1,84,60,197,150,87,60,162,51,91,60,28,216,94,60,52,132,98,60,232,55,102,60,56,243,105,60,35,182,109,60,170,128,113,60,202,82,117,60,131,44,121,60,214,13,125,60,96,123,128,60,161,115,130,60,174,111,132,60,134,111,134,60,40,115,136,60,149,122,138,60,205,133,140,60,206,148,142,60,152,167,144,60,44,190,146,60,136,216,148,60,173,246,150,60,154,24,153,60,78,62,155,60,202,103,157,60,13,149,159,60,23,198,161,60,231,250,163,60,125,51,166,60,217,111,168,60,249,175,170,60,223,243,172,60,137,59,175,60,247,134,177,60,40,214,179,60,29,41,182,60,213,127,184,60,80,218,186,60,140,56,189,60,138,154,191,60,74,0,194,60,202,105,196,60,11,215,198,60,12,72,201,60,205,188,203,60,77,53,206,60,140,177,208,60,137,49,211,60,69,181,213,60,189,60,216,60,243,199,218,60,230,86,221,60,149,233,223,60,0,128,226,60,39,26,229,60,8,184,231,60,164,89,234,60,250,254,236,60,9,168,239,60,210,84,242,60,83,5,245,60,141,185,247,60,126,113,250,60,39,45,253,60,134,236,255,60,206,87,1,61,52,187,2,61,117,32,4,61,144,135,5,61,133,240,6,61,84,91,8,61,253,199,9,61,128,54,11,61,219,166,12,61,16,25,14,61,29,141,15,61,3,3,17,61,193,122,18,61,87,244,19,61,197,111,21,61,10,237,22,61,39,108,24,61,26,237,25,61,228,111,27,61,132,244,28,61,251,122,30,61,71,3,32,61,105,141,33,61,96,25,35,61,45,167,36,61,206,54,38,61,67,200,39,61,141,91,41,61,171,240,42,61,156,135,44,61,96,32,46,61,248,186,47,61,99,87,49,61,160,245,50,61,175,149,52,61,144,55,54,61,67,219,55,61,199,128,57,61,28,40,59,61,65,209,60,61,56,124,62,61,254,40,64,61,148,215,65,61,250,135,67,61,47,58,69,61,51,238,70,61,5,164,72,61,166,91,74,61,20,21,76,61,80,208,77,61,90,141,79,61,49,76,81,61,212,12,83,61,68,207,84,61,128,147,86,61,135,89,88,61,90,33,90,61,248,234,91,61,97,182,93,61,148,131,95,61,145,82,97,61,88,35,99,61,232,245,100,61,65,202,102,61,100,160,104,61,78,120,106,61,1,82,108,61,123,45,110,61,188,10,112,61,197,233,113,61,148,202,115,61,41,173,117,61,133,145,119,61,166,119,121,61,140,95,123,61,55,73,125,61,166,52,127,61,237,144,128,61,105,136,129,61,198,128,130,61,5,122,131,61,37,116,132,61,39,111,133,61,9,107,134,61,204,103,135,61,112,101,136,61,244,99,137,61,88,99,138,61,157,99,139,61,193,100,140,61,196,102,141,61,167,105,142,61,106,109,143,61,11,114,144,61,139,119,145,61,234,125,146,61,40,133,147,61,67,141,148,61,61,150,149,61,20,160,150,61,201,170,151,61,92,182,152,61,203,194,153,61,24,208,154,61,66,222,155,61,72,237,156,61,42,253,157,61,233,13,159,61,132,31,160,61,250,49,161,61,76,69,162,61,122,89,163,61,130,110,164,61,101,132,165,61,35,155,166,61,188,178,167,61,47,203,168,61,124,228,169,61,162,254,170,61,163,25,172,61,124,53,173,61,47,82,174,61,187,111,175,61,31,142,176,61,92,173,177,61,113,205,178,61,94,238,179,61,35,16,181,61,192,50,182,61,52,86,183,61,127,122,184,61,160,159,185,61,153,197,186,61,104,236,187,61,13,20,189,61,136,60,190,61,217,101,191,61,255,143,192,61,250,186,193,61,202,230,194,61,111,19,196,61,233,64,197,61,55,111,198,61,89,158,199,61,78,206,200,61,23,255,201,61,179,48,203,61,35,99,204,61,101,150,205,61,121,202,206,61,96,255,207,61,25,53,209,61,164,107,210,61,0,163,211,61,45,219,212,61,44,20,214,61,251,77,215,61,154,136,216,61,10,196,217,61,74,0,219,61,89,61,220,61,56,123,221,61,230,185,222,61,99,249,223,61,174,57,225,61,200,122,226,61,176,188,227,61,102,255,228,61,233,66,230,61,58,135,231,61,88,204,232,61,66,18,234,61,249,88,235,61,124,160,236,61,203,232,237,61,230,49,239,61,204,123,240,61,125,198,241,61,249,17,243,61,63,94,244,61,79,171,245,61,42,249,246,61,206,71,248,61,60,151,249,61,114,231,250,61,114,56,252,61,58,138,253,61,202,220,254,61,17,24,0,62,33,194,0,62,149,108,1,62,108,23,2,62,166,194,2,62,68,110,3,62,69,26,4,62,168,198,4,62,111,115,5,62,152,32,6,62,35,206,6,62,17,124,7,62,98,42,8,62,20,217,8,62,40,136,9,62,157,55,10,62,117,231,10,62,173,151,11,62,71,72,12,62,66,249,12,62,158,170,13,62,91,92,14,62,120,14,15,62,246,192,15,62,213,115,16,62,19,39,17,62,177,218,17,62,175,142,18,62,13,67,19,62,202,247,19,62,231,172,20,62,99,98,21,62,62,24,22,62,120,206,22,62,16,133,23,62,7,60,24,62,92,243,24,62,16,171,25,62,33,99,26,62,145,27,27,62,94,212,27,62,137,141,28,62,17,71,29,62,246,0,30,62,56,187,30,62,215,117,31,62,211,48,32,62,43,236,32,62,224,167,33,62,241,99,34,62,93,32,35,62,38,221,35,62,74,154,36,62,202,87,37,62,165,21,38,62,219,211,38,62,108,146,39,62,88,81,40,62,159,16,41,62,64,208,41,62,59,144,42,62,144,80,43,62,63,17,44,62,72,210,44,62,170,147,45,62,102,85,46,62,122,23,47,62,232,217,47,62,175,156,48,62,206,95,49,62,69,35,50,62,21,231,50,62,61,171,51,62,189,111,52,62,148,52,53,62,195,249,53,62,73,191,54,62,38,133,55,62,91,75,56,62,230,17,57,62,199,216,57,62,255,159,58,62,141,103,59,62,113,47,60,62,171,247,60,62,59,192,61,62,31,137,62,62,89,82,63,62,232,27,64,62,204,229,64,62,5,176,65,62,146,122,66,62,115,69,67,62,168,16,68,62,49,220,68,62,14,168,69,62,62,116,70,62,194,64,71,62,152,13,72,62,193,218,72,62,61,168,73,62,12,118,74,62,44,68,75,62,159,18,76,62,100,225,76,62,122,176,77,62,225,127,78,62,154,79,79,62,164,31,80,62,255,239,80,62,170,192,81,62,166,145,82,62,242,98,83,62,141,52,84,62,121,6,85,62,180,216,85,62,63,171,86,62,25,126,87,62,65,81,88,62,185,36,89,62,126,248,89,62,147,204,90,62,245,160,91,62,165,117,92,62,163,74,93,62,238,31,94,62,135,245,94,62,109,203,95,62,159,161,96,62,30,120,97,62,233,78,98,62,1,38,99,62,100,253,99,62,19,213,100,62,14,173,101,62,84,133,102,62,229,93,103,62,193,54,104,62,231,15,105,62,88,233,105,62,19,195,106,62,24,157,107,62,103,119,108,62,255,81,109,62,224,44,110,62,11,8,111,62,126,227,111,62,58,191,112,62,62,155,113,62,139,119,114,62,31,84,115,62,251,48,116,62,31,14,117,62,138,235,117,62,59,201,118,62,52,167,119,62,115,133,120,62,248,99,121,62,196,66,122,62,213,33,123,62,44,1,124,62,200,224,124,62,170,192,125,62,208,160,126,62,59,129,127,62,245,48,128,62,111,161,128,62,11,18,129,62,201,130,129,62,168,243,129,62,169,100,130,62,204,213,130,62,15,71,131,62,117,184,131,62,251,41,132,62,162,155,132,62,107,13,133,62,84,127,133,62,93,241,133,62,136,99,134,62,210,213,134,62,61,72,135,62,200,186,135,62,116,45,136,62,63,160,136,62,42,19,137,62,52,134,137,62,94,249,137,62,168,108,138,62,17,224,138,62,153,83,139,62,64,199,139,62,6,59,140,62,235,174,140,62,239,34,141,62,17,151,141,62,82,11,142,62,177,127,142,62,46,244,142,62,201,104,143,62,130,221,143,62,89,82,144,62,78,199,144,62,96,60,145,62,143,177,145,62,220,38,146,62,70,156,146,62,205,17,147,62,113,135,147,62,50,253,147,62,16,115,148,62,9,233,148,62,32,95,149,62,82,213,149,62,161,75,150,62,12,194,150,62,146,56,151,62,53,175,151,62,243,37,152,62,204,156,152,62,193,19,153,62,209,138,153,62,252,1,154,62,66,121,154,62,163,240,154,62,31,104,155,62,181,223,155,62,101,87,156,62,48,207,156,62,21,71,157,62,20,191,157,62,45,55,158,62,96,175,158,62,172,39,159,62,18,160,159,62,145,24,160,62,41,145,160,62,218,9,161,62,165,130,161,62,136,251,161,62,132,116,162,62,152,237,162,62,197,102,163,62,10,224,163,62,103,89,164,62,220,210,164,62,105,76,165,62,14,198,165,62,202,63,166,62,158,185,166,62,137,51,167,62,139,173,167,62,164,39,168,62,213,161,168,62,27,28,169,62,121,150,169,62,237,16,170,62,119,139,170,62,24,6,171,62,206,128,171,62,155,251,171,62,125,118,172,62,117,241,172,62,130,108,173,62,165,231,173,62,221,98,174,62,42,222,174,62,140,89,175,62,2,213,175,62,142,80,176,62,46,204,176,62,226,71,177,62,170,195,177,62,135,63,178,62,119,187,178,62,124,55,179,62,148,179,179,62,191,47,180,62,254,171,180,62,80,40,181,62,181,164,181,62,45,33,182,62,184,157,182,62,85,26,183,62,5,151,183,62,199,19,184,62,156,144,184,62,130,13,185,62,123,138,185,62,133,7,186,62,161,132,186,62,206,1,187,62,13,127,187,62,93,252,187,62,190,121,188,62,48,247,188,62,178,116,189,62,70,242,189,62,233,111,190,62,157,237,190,62,98,107,191,62,54,233,191,62,26,103,192,62,14,229,192,62,17,99,193,62,36,225,193,62,70,95,194,62,119,221,194,62,184,91,195,62,7,218,195,62,100,88,196,62,209,214,196,62,75,85,197,62,212,211,197,62,107,82,198,62,16,209,198,62,195,79,199,62,132,206,199,62,82,77,200,62,45,204,200,62,21,75,201,62,11,202,201,62,13,73,202,62,29,200,202,62,56,71,203,62,97,198,203,62,149,69,204,62,214,196,204,62,34,68,205,62,123,195,205,62,223,66,206,62,79,194,206,62,202,65,207,62,81,193,207,62,226,64,208,62,127,192,208,62,38,64,209,62,216,191,209,62,148,63,210,62,91,191,210,62,44,63,211,62,7,191,211,62,235,62,212,62,218,190,212,62,210,62,213,62,211,190,213,62,222,62,214,62,242,190,214,62,15,63,215,62,53,191,215,62,99,63,216,62,154,191,216,62,217,63,217,62,32,192,217,62,112,64,218,62,199,192,218,62,38,65,219,62,140,193,219,62,250,65,220,62,112,194,220,62,236,66,221,62,112,195,221,62,250,67,222,62,139,196,222,62,34,69,223,62,192,197,223,62,100,70,224,62,14,199,224,62,189,71,225,62,115,200,225,62,46,73,226,62,239,201,226,62,181,74,227,62,127,203,227,62,79,76,228,62,36,205,228,62,253,77,229,62,219,206,229,62,190,79,230,62,164,208,230,62,142,81,231,62,125,210,231,62,111,83,232,62,100,212,232,62,93,85,233,62,89,214,233,62,89,87,234,62,91,216,234,62,96,89,235,62,104,218,235,62,114,91,236,62,126,220,236,62,141,93,237,62,158,222,237,62,176,95,238,62,196,224,238,62,218,97,239,62,241,226,239,62,10,100,240,62,35,229,240,62,62,102,241,62,89,231,241,62,116,104,242,62,145,233,242,62,173,106,243,62,202,235,243,62,230,108,244,62,3,238,244,62,31,111,245,62,59,240,245,62,86,113,246,62,112,242,246,62,137,115,247,62,161,244,247,62,184,117,248,62,206,246,248,62,226,119,249,62,244,248,249,62,4,122,250,62,18,251,250,62,30,124,251,62,40,253,251,62,47,126,252,62,52,255,252,62,54,128,253,62,52,1,254,62,48,130,254,62,40,3,255,62,29,132,255,62,135,2,0,63,254,66,0,63,115,131,0,63,230,195,0,63,86,4,1,63,197,68,1,63,49,133,1,63,155,197,1,63,3,6,2,63,103,70,2,63,202,134,2,63,42,199,2,63,135,7,3,63,225,71,3,63,56,136,3,63,141,200,3,63,222,8,4,63,44,73,4,63,119,137,4,63,191,201,4,63,3,10,5,63,68,74,5,63,130,138,5,63,188,202,5,63,242,10,6,63,36,75,6,63,83,139,6,63,126,203,6,63,165,11,7,63,199,75,7,63,230,139,7,63,1,204,7,63,23,12,8,63,41,76,8,63,54,140,8,63,63,204,8,63,67,12,9,63,67,76,9,63,62,140,9,63,52,204,9,63,37,12,10,63,18,76,10,63,249,139,10,63,219,203,10,63,184,11,11,63,144,75,11,63,98,139,11,63,47,203,11,63,246,10,12,63,184,74,12,63,116,138,12,63,43,202,12,63,219,9,13,63,134,73,13,63,43,137,13,63,202,200,13,63,98,8,14,63,245,71,14,63,129,135,14,63,7,199,14,63,135,6,15,63,0,70,15,63,114,133,15,63,222,196,15,63,67,4,16,63,161,67,16,63,249,130,16,63,73,194,16,63,147,1,17,63,213,64,17,63,17,128,17,63,69,191,17,63,114,254,17,63,151,61,18,63,181,124,18,63,203,187,18,63,218,250,18,63,225,57,19,63,225,120,19,63,216,183,19,63,200,246,19,63,176,53,20,63,143,116,20,63,103,179,20,63,54,242,20,63,253,48,21,63,188,111,21,63,114,174,21,63,32,237,21,63,197,43,22,63,98,106,22,63,246,168,22,63,129,231,22,63,3,38,23,63,125,100,23,63,237,162,23,63,84,225,23,63,178,31,24,63,7,94,24,63,83,156,24,63,149,218,24,63,206,24,25,63,253,86,25,63,35,149,25,63,63,211,25,63,82,17,26,63,90,79,26,63,89,141,26,63,78,203,26,63,57,9,27,63,25,71,27,63,240,132,27,63,188,194,27,63,126,0,28,63,54,62,28,63,227,123,28,63,134,185,28,63,30,247,28,63,172,52,29,63,47,114,29,63,167,175,29,63,20,237,29,63,118,42,30,63,206,103,30,63,26,165,30,63,91,226,30,63,145,31,31,63,188,92,31,63,219,153,31,63,239,214,31,63,247,19,32,63,244,80,32,63,230,141,32,63,203,202,32,63,165,7,33,63,115,68,33,63,53,129,33,63,235,189,33,63,150,250,33,63,52,55,34,63,198,115,34,63,75,176,34,63,197,236,34,63,50,41,35,63,146,101,35,63,230,161,35,63,46,222,35,63,105,26,36,63,151,86,36,63,185,146,36,63,205,206,36,63,213,10,37,63,208,70,37,63,190,130,37,63,158,190,37,63,114,250,37,63,56,54,38,63,241,113,38,63,157,173,38,63,59,233,38,63,204,36,39,63,79,96,39,63,197,155,39,63,45,215,39,63,135,18,40,63,211,77,40,63,18,137,40,63,66,196,40,63,101,255,40,63,121,58,41,63,128,117,41,63,120,176,41,63,98,235,41,63,62,38,42,63,11,97,42,63,202,155,42,63,122,214,42,63,28,17,43,63,175,75,43,63,52,134,43,63,170,192,43,63,16,251,43,63,105,53,44,63,178,111,44,63,236,169,44,63,23,228,44,63,51,30,45,63,64,88,45,63,61,146,45,63,43,204,45,63,10,6,46,63,218,63,46,63,154,121,46,63,74,179,46,63,235,236,46,63,124,38,47,63,254,95,47,63,112,153,47,63,210,210,47,63,36,12,48,63,102,69,48,63,152,126,48,63,186,183,48,63,204,240,48,63,205,41,49,63,191,98,49,63,160,155,49,63,113,212,49,63,49,13,50,63,225,69,50,63,128,126,50,63,15,183,50,63,141,239,50,63,251,39,51,63,87,96,51,63,163,152,51,63,222,208,51,63,8,9,52,63,34,65,52,63,42,121,52,63,33,177,52,63,7,233,52,63,219,32,53,63,159,88,53,63,81,144,53,63,242,199,53,63,129,255,53,63,255,54,54,63,108,110,54,63,198,165,54,63,16,221,54,63,71,20,55,63,109,75,55,63,129,130,55,63,131,185,55,63,116,240,55,63,82,39,56,63,30,94,56,63,217,148,56,63,129,203,56,63,23,2,57,63,155,56,57,63,13,111,57,63,108,165,57,63,185,219,57,63,244,17,58,63,28,72,58,63,50,126,58,63,53,180,58,63,38,234,58,63,4,32,59,63,207,85,59,63,135,139,59,63,45,193,59,63,192,246,59,63,64,44,60,63,173,97,60,63,7,151,60,63,78,204,60,63,130,1,61,63,163,54,61,63,177,107,61,63,171,160,61,63,146,213,61,63,102,10,62,63,39,63,62,63,212,115,62,63,110,168,62,63,244,220,62,63,103,17,63,63,198,69,63,63,17,122,63,63,73,174,63,63,109,226,63,63,126,22,64,63,122,74,64,63,99,126,64,63,56,178,64,63,248,229,64,63,165,25,65,63,62,77,65,63,195,128,65,63,52,180,65,63,144,231,65,63,216,26,66,63,13,78,66,63,44,129,66,63,56,180,66,63,47,231,66,63,18,26,67,63,224,76,67,63,154,127,67,63,64,178,67,63,208,228,67,63,77,23,68,63,180,73,68,63,7,124,68,63,69,174,68,63,111,224,68,63,131,18,69,63,131,68,69,63,110,118,69,63,68,168,69,63,5,218,69,63,177,11,70,63,72,61,70,63,202,110,70,63,55,160,70,63,143,209,70,63,210,2,71,63,255,51,71,63,23,101,71,63,26,150,71,63,8,199,71,63,224,247,71,63,163,40,72,63,81,89,72,63,233,137,72,63,107,186,72,63,216,234,72,63,48,27,73,63,114,75,73,63,158,123,73,63,181,171,73,63,181,219,73,63,161,11,74,63,118,59,74,63,54,107,74,63,224,154,74,63,116,202,74,63,242,249,74,63,90,41,75,63,173,88,75,63,233,135,75,63,15,183,75,63,32,230,75,63,26,21,76,63,254,67,76,63,204,114,76,63,132,161,76,63,38,208,76,63,177,254,76,63,38,45,77,63,133,91,77,63,206,137,77,63,0,184,77,63,28,230,77,63,34,20,78,63,17,66,78,63,234,111,78,63,172,157,78,63,88,203,78,63,238,248,78,63,108,38,79,63,213,83,79,63,38,129,79,63,97,174,79,63,134,219,79,63,147,8,80,63,138,53,80,63,107,98,80,63,52,143,80,63,231,187,80,63,131,232,80,63,8,21,81,63,119,65,81,63,206,109,81,63,15,154,81,63,57,198,81,63,76,242,81,63,71,30,82,63,44,74,82,63,250,117,82,63,177,161,82,63,81,205,82,63,218,248,82,63,76,36,83,63,166,79,83,63,234,122,83,63,22,166,83,63,44,209,83,63,42,252,83,63,17,39,84,63,224,81,84,63,153,124,84,63,58,167,84,63,196,209,84,63,54,252,84,63,146,38,85,63,214,80,85,63,2,123,85,63,24,165,85,63,22,207,85,63,252,248,85,63,204,34,86,63,131,76,86,63,36,118,86,63,172,159,86,63,30,201,86,63,120,242,86,63,186,27,87,63,229,68,87,63,248,109,87,63,244,150,87,63,216,191,87,63,165,232,87,63,90,17,88,63,248,57,88,63,126,98,88,63,236,138,88,63,67,179,88,63,130,219,88,63,169,3,89,63,185,43,89,63,177,83,89,63,145,123,89,63,90,163,89,63,11,203,89,63,164,242,89,63,37,26,90,63,143,65,90,63,225,104,90,63,27,144,90,63,62,183,90,63,72,222,90,63,59,5,91,63,22,44,91,63,217,82,91,63,133,121,91,63,24,160,91,63,148,198,91,63,248,236,91,63,68,19,92,63,120,57,92,63,149,95,92,63,153,133,92,63,134,171,92,63,91,209,92,63,24,247,92,63,189,28,93,63,74,66,93,63,191,103,93,63,28,141,93,63,98,178,93,63,143,215,93,63,165,252,93,63,162,33,94,63,136,70,94,63,86,107,94,63,11,144,94,63,169,180,94,63,47,217,94,63,157,253,94,63,243,33,95,63,49,70,95,63,88,106,95,63,102,142,95,63,92,178,95,63,59,214,95,63,1,250,95,63,175,29,96,63,70,65,96,63,196,100,96,63,43,136,96,63,122,171,96,63,176,206,96,63,207,241,96,63,214,20,97,63,197,55,97,63,155,90,97,63,90,125,97,63,1,160,97,63,144,194,97,63,8,229,97,63,103,7,98,63,174,41,98,63,221,75,98,63,245,109,98,63,244,143,98,63,220,177,98,63,171,211,98,63,99,245,98,63,3,23,99,63,139,56,99,63,251,89,99,63,83,123,99,63,147,156,99,63,188,189,99,63,204,222,99,63,197,255,99,63,166,32,100,63,110,65,100,63,32,98,100,63,185,130,100,63,58,163,100,63,164,195,100,63,245,227,100,63,47,4,101,63,82,36,101,63,92,68,101,63,78,100,101,63,41,132,101,63,236,163,101,63,151,195,101,63,43,227,101,63,167,2,102,63,11,34,102,63,87,65,102,63,139,96,102,63,168,127,102,63,174,158,102,63,155,189,102,63,113,220,102,63,47,251,102,63,214,25,103,63,101,56,103,63,220,86,103,63,59,117,103,63,132,147,103,63,180,177,103,63,205,207,103,63,206,237,103,63,184,11,104,63,138,41,104,63,69,71,104,63,233,100,104,63,116,130,104,63,233,159,104,63,69,189,104,63,139,218,104,63,185,247,104,63,207,20,105,63,207,49,105,63,182,78,105,63,135,107,105,63,64,136,105,63,225,164,105,63,108,193,105,63,223,221,105,63,59,250,105,63,127,22,106,63,172,50,106,63,195,78,106,63,193,106,106,63,169,134,106,63,121,162,106,63,51,190,106,63,213,217,106,63,96,245,106,63,212,16,107,63,48,44,107,63,118,71,107,63,165,98,107,63,188,125,107,63,189,152,107,63,167,179,107,63,121,206,107,63,53,233,107,63,218,3,108,63,104,30,108,63,223,56,108,63,63,83,108,63,136,109,108,63,187,135,108,63,214,161,108,63,219,187,108,63,201,213,108,63,161,239,108,63,97,9,109,63,11,35,109,63,159,60,109,63,27,86,109,63,129,111,109,63,209,136,109,63,9,162,109,63,44,187,109,63,56,212,109,63,45,237,109,63,12,6,110,63,212,30,110,63,134,55,110,63,33,80,110,63,166,104,110,63,21,129,110,63,110,153,110,63,176,177,110,63,220,201,110,63,241,225,110,63,241,249,110,63,218,17,111,63,173,41,111,63,106,65,111,63,16,89,111,63,161,112,111,63,28,136,111,63,128,159,111,63,207,182,111,63,7,206,111,63,42,229,111,63,54,252,111,63,45,19,112,63,14,42,112,63,217,64,112,63,142,87,112,63,46,110,112,63,184,132,112,63,43,155,112,63,138,177,112,63,210,199,112,63,5,222,112,63,35,244,112,63,42,10,113,63,29,32,113,63,249,53,113,63,193,75,113,63,114,97,113,63,15,119,113,63,150,140,113,63,7,162,113,63,99,183,113,63,170,204,113,63,220,225,113,63,249,246,113,63,0,12,114,63,242,32,114,63,207,53,114,63,151,74,114,63,73,95,114,63,231,115,114,63,112,136,114,63,227,156,114,63,66,177,114,63,140,197,114,63,193,217,114,63,225,237,114,63,236,1,115,63,227,21,115,63,197,41,115,63,146,61,115,63,74,81,115,63,238,100,115,63,125,120,115,63,248,139,115,63,94,159,115,63,175,178,115,63,236,197,115,63,21,217,115,63,41,236,115,63,41,255,115,63,21,18,116,63,236,36,116,63,175,55,116,63,94,74,116,63,248,92,116,63,127,111,116,63,241,129,116,63,80,148,116,63,154,166,116,63,208,184,116,63,242,202,116,63,1,221,116,63,251,238,116,63,226,0,117,63,181,18,117,63,116,36,117,63,31,54,117,63,183,71,117,63,59,89,117,63,171,106,117,63,8,124,117,63,81,141,117,63,135,158,117,63,169,175,117,63,184,192,117,63,179,209,117,63,155,226,117,63,112,243,117,63,50,4,118,63,224,20,118,63,123,37,118,63,3,54,118,63,120,70,118,63,217,86,118,63,40,103,118,63,100,119,118,63,140,135,118,63,162,151,118,63,165,167,118,63,149,183,118,63,114,199,118,63,61,215,118,63,245,230,118,63,154,246,118,63,44,6,119,63,172,21,119,63,26,37,119,63,117,52,119,63,189,67,119,63,243,82,119,63,22,98,119,63,40,113,119,63,39,128,119,63,19,143,119,63,238,157,119,63,182,172,119,63,108,187,119,63,16,202,119,63,162,216,119,63,34,231,119,63,144,245,119,63,236,3,120,63,55,18,120,63,111,32,120,63,150,46,120,63,170,60,120,63,174,74,120,63,159,88,120,63,127,102,120,63,77,116,120,63,10,130,120,63,181,143,120,63,79,157,120,63,215,170,120,63,78,184,120,63,180,197,120,63,8,211,120,63,76,224,120,63,126,237,120,63,158,250,120,63,174,7,121,63,173,20,121,63,155,33,121,63,119,46,121,63,67,59,121,63,254,71,121,63,168,84,121,63,66,97,121,63,202,109,121,63,66,122,121,63,169,134,121,63,0,147,121,63,70,159,121,63,124,171,121,63,161,183,121,63,181,195,121,63,186,207,121,63,173,219,121,63,145,231,121,63,100,243,121,63,40,255,121,63,219,10,122,63,126,22,122,63,16,34,122,63,147,45,122,63,6,57,122,63,105,68,122,63,188,79,122,63,255,90,122,63,51,102,122,63,86,113,122,63,106,124,122,63,111,135,122,63,99,146,122,63,72,157,122,63,30,168,122,63,228,178,122,63,155,189,122,63,66,200,122,63,218,210,122,63,99,221,122,63,221,231,122,63,71,242,122,63,162,252,122,63,238,6,123,63,43,17,123,63,89,27,123,63,120,37,123,63,137,47,123,63,138,57,123,63,124,67,123,63,96,77,123,63,53,87,123,63,252,96,123,63,179,106,123,63,92,116,123,63,247,125,123,63,131,135,123,63,1,145,123,63,112,154,123,63,209,163,123,63,36,173,123,63,104,182,123,63,158,191,123,63,198,200,123,63,224,209,123,63,236,218,123,63,234,227,123,63,218,236,123,63,188,245,123,63,144,254,123,63,86,7,124,63,14,16,124,63,185,24,124,63,86,33,124,63,230,41,124,63,104,50,124,63,220,58,124,63,67,67,124,63,156,75,124,63,232,83,124,63,39,92,124,63,88,100,124,63,124,108,124,63,147,116,124,63,157,124,124,63,153,132,124,63,137,140,124,63,107,148,124,63,65,156,124,63,9,164,124,63,197,171,124,63,116,179,124,63,22,187,124,63,172,194,124,63,52,202,124,63,176,209,124,63,32,217,124,63,131,224,124,63,217,231,124,63,35,239,124,63,97,246,124,63,146,253,124,63,183,4,125,63,208,11,125,63,221,18,125,63,221,25,125,63,209,32,125,63,185,39,125,63,150,46,125,63,102,53,125,63,42,60,125,63,227,66,125,63,143,73,125,63,48,80,125,63,197,86,125,63,78,93,125,63,204,99,125,63,62,106,125,63,165,112,125,63,0,119,125,63,80,125,125,63,148,131,125,63,205,137,125,63,251,143,125,63,29,150,125,63,52,156,125,63,64,162,125,63,65,168,125,63,55,174,125,63,34,180,125,63,2,186,125,63,215,191,125,63,161,197,125,63,96,203,125,63,21,209,125,63,190,214,125,63,93,220,125,63,242,225,125,63,124,231,125,63,251,236,125,63,112,242,125,63,218,247,125,63,58,253,125,63,143,2,126,63,219,7,126,63,28,13,126,63,82,18,126,63,127,23,126,63,161,28,126,63,186,33,126,63,200,38,126,63,204,43,126,63,199,48,126,63,183,53,126,63,158,58,126,63,123,63,126,63,78,68,126,63,23,73,126,63,215,77,126,63,141,82,126,63,58,87,126,63,221,91,126,63,118,96,126,63,6,101,126,63,141,105,126,63,10,110,126,63,126,114,126,63,233,118,126,63,75,123,126,63,164,127,126,63,243,131,126,63,57,136,126,63,119,140,126,63,171,144,126,63,214,148,126,63,249,152,126,63,18,157,126,63,35,161,126,63,44,165,126,63,43,169,126,63,34,173,126,63,16,177,126,63,246,180,126,63,211,184,126,63,167,188,126,63,115,192,126,63,55,196,126,63,243,199,126,63,166,203,126,63,81,207,126,63,243,210,126,63,142,214,126,63,32,218,126,63,171,221,126,63,45,225,126,63,167,228,126,63,26,232,126,63,132,235,126,63,231,238,126,63,66,242,126,63,149,245,126,63,224,248,126,63,36,252,126,63,96,255,126,63,148,2,127,63,193,5,127,63,230,8,127,63,4,12,127,63,27,15,127,63,42,18,127,63,50,21,127,63,50,24,127,63,43,27,127,63,29,30,127,63,8,33,127,63,236,35,127,63,201,38,127,63,158,41,127,63,109,44,127,63,53,47,127,63,246,49,127,63,175,52,127,63,99,55,127,63,15,58,127,63,181,60,127,63,83,63,127,63,236,65,127,63,125,68,127,63,8,71,127,63,141,73,127,63,11,76,127,63,131,78,127,63,244,80,127,63,95,83,127,63,195,85,127,63,33,88,127,63,121,90,127,63,203,92,127,63,23,95,127,63,92,97,127,63,155,99,127,63,213,101,127,63,8,104,127,63,54,106,127,63,93,108,127,63,127,110,127,63,155,112,127,63,177,114,127,63,193,116,127,63,203,118,127,63,208,120,127,63,207,122,127,63,201,124,127,63,189,126,127,63,171,128,127,63,148,130,127,63,120,132,127,63,86,134,127,63,47,136,127,63,2,138,127,63,209,139,127,63,153,141,127,63,93,143,127,63,28,145,127,63,213,146,127,63,137,148,127,63,57,150,127,63,227,151,127,63,136,153,127,63,40,155,127,63,196,156,127,63,90,158,127,63,236,159,127,63,121,161,127,63,1,163,127,63,132,164,127,63,3,166,127,63,125,167,127,63,242,168,127,63,99,170,127,63,207,171,127,63,55,173,127,63,154,174,127,63,249,175,127,63,84,177,127,63,170,178,127,63,251,179,127,63,73,181,127,63,146,182,127,63,215,183,127,63,24,185,127,63,85,186,127,63,141,187,127,63,193,188,127,63,242,189,127,63,30,191,127,63,71,192,127,63,107,193,127,63,140,194,127,63,168,195,127,63,193,196,127,63,214,197,127,63,231,198,127,63,245,199,127,63,255,200,127,63,5,202,127,63,7,203,127,63,6,204,127,63,1,205,127,63,249,205,127,63,237,206,127,63,222,207,127,63,203,208,127,63,181,209,127,63,156,210,127,63,127,211,127,63,95,212,127,63,59,213,127,63,20,214,127,63,234,214,127,63,189,215,127,63,141,216,127,63,90,217,127,63,35,218,127,63,233,218,127,63,173,219,127,63,109,220,127,63,43,221,127,63,229,221,127,63,156,222,127,63,81,223,127,63,3,224,127,63,178,224,127,63,94,225,127,63,7,226,127,63,174,226,127,63,82,227,127,63,243,227,127,63,146,228,127,63,46,229,127,63,199,229,127,63,94,230,127,63,242,230,127,63,132,231,127,63,19,232,127,63,160,232,127,63,42,233,127,63,178,233,127,63,56,234,127,63,187,234,127,63,60,235,127,63,187,235,127,63,55,236,127,63,177,236,127,63,41,237,127,63,159,237,127,63,18,238,127,63,132,238,127,63,243,238,127,63,96,239,127,63,204,239,127,63,53,240,127,63,156,240,127,63,1,241,127,63,101,241,127,63,198,241,127,63,37,242,127,63,131,242,127,63,222,242,127,63,56,243,127,63,144,243,127,63,231,243,127,63,59,244,127,63,142,244,127,63,223,244,127,63,46,245,127,63,124,245,127,63,200,245,127,63,19,246,127,63,91,246,127,63,163,246,127,63,233,246,127,63,45,247,127,63,111,247,127,63,177,247,127,63,240,247,127,63,47,248,127,63,108,248,127,63,167,248,127,63,225,248,127,63,26,249,127,63,82,249,127,63,136,249,127,63,188,249,127,63,240,249,127,63,34,250,127,63,83,250,127,63,131,250,127,63,178,250,127,63,224,250,127,63,12,251,127,63,55,251,127,63,97,251,127,63,138,251,127,63,178,251,127,63,217,251,127,63,255,251,127,63,36,252,127,63,72,252,127,63,107,252,127,63,141,252,127,63,173,252,127,63,205,252,127,63,237,252,127,63,11,253,127,63,40,253,127,63,69,253,127,63,96,253,127,63,123,253,127,63,149,253,127,63,174,253,127,63,199,253,127,63,222,253,127,63,245,253,127,63,12,254,127,63,33,254,127,63,54,254,127,63,74,254,127,63,93,254,127,63,112,254,127,63,130,254,127,63,148,254,127,63,165,254,127,63,181,254,127,63,197,254,127,63,212,254,127,63,227,254,127,63,241,254,127,63,254,254,127,63,11,255,127,63,24,255,127,63,36,255,127,63,47,255,127,63,59,255,127,63,69,255,127,63,79,255,127,63,89,255,127,63,99,255,127,63,108,255,127,63,116,255,127,63,124,255,127,63,132,255,127,63,140,255,127,63,147,255,127,63,154,255,127,63,160,255,127,63,166,255,127,63,172,255,127,63,178,255,127,63,183,255,127,63,188,255,127,63,193,255,127,63,197,255,127,63,202,255,127,63,206,255,127,63,209,255,127,63,213,255,127,63,216,255,127,63,220,255,127,63,223,255,127,63,225,255,127,63,228,255,127,63,230,255,127,63,233,255,127,63,235,255,127,63,237,255,127,63,239,255,127,63,240,255,127,63,242,255,127,63,243,255,127,63,245,255,127,63,246,255,127,63,247,255,127,63,248,255,127,63,249,255,127,63,250,255,127,63,251,255,127,63,251,255,127,63,252,255,127,63,252,255,127,63,253,255,127,63,253,255,127,63,254,255,127,63,254,255,127,63,254,255,127,63,255,255,127,63,255,255,127,63,255,255,127,63,255,255,127,63,255,255,127,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,198,63,120,51,98,136,11,53,151,200,193,53,80,233,61,54,183,247,156,54,46,124,234,54,153,192,35,55,244,2,90,55,56,3,140,55,227,228,174,55,177,166,213,55,108,36,0,56,146,101,23,56,201,150,48,56,18,184,75,56,81,201,104,56,94,229,131,56,29,94,148,56,229,206,165,56,167,55,184,56,128,152,203,56,85,241,223,56,36,66,245,56,126,197,5,57,238,101,17,57,99,130,29,57,207,26,42,57,63,47,55,57,179,191,68,57,30,204,82,57,141,84,97,57,243,88,112,57,94,217,127,57,227,234,135,57,18,39,144,57,64,161,152,57,105,89,161,57,146,79,170,57,181,131,179,57,215,245,188,57,245,165,198,57,14,148,208,57,34,192,218,57,46,42,229,57,57,210,239,57,60,184,250,57,27,238,2,58,22,159,8,58,13,111,14,58,0,94,20,58,239,107,26,58,218,152,32,58,192,228,38,58,161,79,45,58,124,217,51,58,83,130,58,58,37,74,65,58,240,48,72,58,182,54,79,58,116,91,86,58,45,159,93,58,222,1,101,58,136,131,108,58,42,36,116,58,196,227,123,58,44,225,129,58,241,223,133,58,49,238,137,58,238,11,142,58,37,57,146,58,215,117,150,58,5,194,154,58,174,29,159,58,209,136,163,58,110,3,168,58,134,141,172,58,24,39,177,58,36,208,181,58,169,136,186,58,169,80,191,58,33,40,196,58,19,15,201,58,126,5,206,58,98,11,211,58,191,32,216,58,148,69,221,58,225,121,226,58,166,189,231,58,227,16,237,58,152,115,242,58,196,229,247,58,103,103,253,58,65,124,1,59,137,76,4,59,141,36,7,59,76,4,10,59,198,235,12,59,251,218,15,59,235,209,18,59,149,208,21,59,251,214,24,59,26,229,27,59,244,250,30,59,136,24,34,59,215,61,37,59,223,106,40,59,161,159,43,59,29,220,46,59,83,32,50,59,66,108,53,59,234,191,56,59,76,27,60,59,103,126,63,59,59,233,66,59,199,91,70,59,12,214,73,59,10,88,77,59,193,225,80,59,48,115,84,59,86,12,88,59,53,173,91,59,204,85,95,59,26,6,99,59,32,190,102,59,222,125,106,59,82,69,110,59,127,20,114,59,97,235,117,59,251,201,121,59,76,176,125,59,41,207,128,59,8,202,130,59,194,200,132,59,87,203,134,59,198,209,136,59,17,220,138,59,55,234,140,59,55,252,142,59,18,18,145,59,199,43,147,59,87,73,149,59,194,106,151,59,6,144,153,59,37,185,155,59,30,230,157,59,241,22,160,59,158,75,162,59,37,132,164,59,134,192,166,59,192,0,169,59,212,68,171,59,193,140,173,59,137,216,175,59,41,40,178,59,163,123,180,59,245,210,182,59,33,46,185,59,38,141,187,59,4,240,189,59,186,86,192,59,73,193,194,59,177,47,197,59,242,161,199,59,10,24,202,59,251,145,204,59,196,15,207,59,102,145,209,59,223,22,212,59,49,160,214,59,90,45,217,59,91,190,219,59,51,83,222,59,227,235,224,59,107,136,227,59,201,40,230,59,255,204,232,59,12,117,235,59,240,32,238,59,171,208,240,59,61,132,243,59,165,59,246,59,228,246,248,59,250,181,251,59,229,120,254,59,212,159,0,60,32,5,2,60,87,108,3,60,121,213,4,60,134,64,6,60,126,173,7,60,96,28,9,60,45,141,10,60,229,255,11,60,136,116,13,60,21,235,14,60,141,99,16,60,239,221,17,60,59,90,19,60,114,216,20,60,147,88,22,60,158,218,23,60,147,94,25,60,115,228,26,60,60,108,28,60,240,245,29,60,141,129,31,60,20,15,33,60,133,158,34,60,224,47,36,60,36,195,37,60,82,88,39,60,105,239,40,60,106,136,42,60,84,35,44,60,40,192,45,60,229,94,47,60,139,255,48,60,26,162,50,60,146,70,52,60,243,236,53,60,61,149,55,60,112,63,57,60,140,235,58,60,145,153,60,60,126,73,62,60,84,251,63,60,18,175,65,60,185,100,67,60,72,28,69,60,192,213,70,60,31,145,72,60,103,78,74,60,151,13,76,60,175,206,77,60,176,145,79,60,152,86,81,60,103,29,83,60,31,230,84,60,190,176,86,60,69,125,88,60,179,75,90,60,9,28,92,60,71,238,93,60,107,194,95,60,119,152,97,60,106,112,99,60,68,74,101,60,5,38,103,60,173,3,105,60,60,227,106,60,178,196,108,60,14,168,110,60,81,141,112,60,123,116,114,60,139,93,116,60,130,72,118,60,95,53,120,60,34,36,122,60,203,20,124,60,90,7,126,60,208,251,127,60,22,249,128,60,54,245,129,60,74,242,130,60,80,240,131,60,73,239,132,60,53,239,133,60,19,240,134,60,229,241,135,60,169,244,136,60,95,248,137,60,8,253,138,60,164,2,140,60,50,9,141,60,178,16,142,60,37,25,143,60,139,34,144,60,226,44,145,60,44,56,146,60,104,68,147,60,150,81,148,60,182,95,149,60,201,110,150,60,205,126,151,60,196,143,152,60,172,161,153,60,135,180,154,60,83,200,155,60,17,221,156,60,193,242,157,60,98,9,159,60,245,32,160,60,122,57,161,60,241,82,162,60,89,109,163,60,178,136,164,60,253,164,165,60,57,194,166,60,103,224,167,60,134,255,168,60,151,31,170,60,152,64,171,60,139,98,172,60,111,133,173,60,68,169,174,60,10,206,175,60,193,243,176,60,105,26,178,60,2,66,179,60,139,106,180,60,6,148,181,60,113,190,182,60,205,233,183,60,26,22,185,60,87,67,186,60,133,113,187,60,163,160,188,60,177,208,189,60,177,1,191,60,160,51,192,60,128,102,193,60,80,154,194,60,16,207,195,60,193,4,197,60,97,59,198,60,242,114,199,60,114,171,200,60,227,228,201,60,67,31,203,60,147,90,204,60,211,150,205,60,3,212,206,60,34,18,208,60,49,81,209,60,48,145,210,60,30,210,211,60,252,19,213,60,201,86,214,60,133,154,215,60,49,223,216,60,204,36,218,60,86,107,219,60,208,178,220,60,56,251,221,60,144,68,223,60,214,142,224,60,12,218,225,60,48,38,227,60,67,115,228,60,69,193,229,60,54,16,231,60,21,96,232,60,227,176,233,60,160,2,235,60,75,85,236,60,228,168,237,60,108,253,238,60,226,82,240,60,70,169,241,60,153,0,243,60,218,88,244,60,8,178,245,60,37,12,247,60,48,103,248,60,41,195,249,60,15,32,251,60,228,125,252,60,166,220,253,60,85,60,255,60,121,78,0,61,63,255,0,61,123,176,1,61,46,98,2,61,88,20,3,61,248,198,3,61,15,122,4,61,156,45,5,61,161,225,5,61,27,150,6,61,12,75,7,61,116,0,8,61,82,182,8,61,167,108,9,61,113,35,10,61,179,218,10,61,106,146,11,61,152,74,12,61,60,3,13,61,87,188,13,61,231,117,14,61,238,47,15,61,107,234,15,61,94,165,16,61,199,96,17,61,166,28,18,61,251,216,18,61,198,149,19,61,7,83,20,61,190,16,21,61,234,206,21,61,141,141,22,61,165,76,23,61,52,12,24,61,56,204,24,61,177,140,25,61,161,77,26,61,6,15,27,61,224,208,27,61,48,147,28,61,246,85,29,61,49,25,30,61,226,220,30,61,8,161,31,61,164,101,32,61,181,42,33,61,59,240,33,61,55,182,34,61,168,124,35,61,142,67,36,61,233,10,37,61,186,210,37,61,255,154,38,61,186,99,39,61,234,44,40,61,143,246,40,61,168,192,41,61,55,139,42,61,59,86,43,61,180,33,44,61,161,237,44,61,4,186,45,61,219,134,46,61,38,84,47,61,231,33,48,61,28,240,48,61,198,190,49,61,229,141,50,61,120,93,51,61,127,45,52,61,251,253,52,61,236,206,53,61,81,160,54,61,42,114,55,61,120,68,56,61,58,23,57,61,112,234,57,61,27,190,58,61,58,146,59,61,204,102,60,61,211,59,61,61,79,17,62,61,62,231,62,61,161,189,63,61,120,148,64,61,195,107,65,61,130,67,66,61,181,27,67,61,92,244,67,61,118,205,68,61,4,167,69,61,6,129,70,61,124,91,71,61,101,54,72,61,194,17,73,61,146,237,73,61,214,201,74,61,141,166,75,61,184,131,76,61,86,97,77,61,104,63,78,61,236,29,79,61,229,252,79,61,80,220,80,61,46,188,81,61,128,156,82,61,69,125,83,61,125,94,84,61,40,64,85,61,69,34,86,61,214,4,87,61,218,231,87,61,81,203,88,61,58,175,89,61,150,147,90,61,101,120,91,61,167,93,92,61,91,67,93,61,130,41,94,61,28,16,95,61,40,247,95,61,167,222,96,61,152,198,97,61,251,174,98,61,209,151,99,61,25,129,100,61,212,106,101,61,0,85,102,61,159,63,103,61,176,42,104,61,51,22,105,61,41,2,106,61,144,238,106,61,105,219,107,61,180,200,108,61,113,182,109,61,160,164,110,61,65,147,111,61,84,130,112,61,216,113,113,61,206,97,114,61,54,82,115,61,15,67,116,61,89,52,117,61,22,38,118,61,67,24,119,61,226,10,120,61,243,253,120,61,117,241,121,61,104,229,122,61,204,217,123,61,162,206,124,61,232,195,125,61,160,185,126,61,201,175,127,61,49,83,128,61,183,206,128,61,117,74,129,61,107,198,129,61,154,66,130,61,1,191,130,61,160,59,131,61,120,184,131,61,136,53,132,61,209,178,132,61,81,48,133,61,10,174,133,61,251,43,134,61,37,170,134,61,134,40,135,61,32,167,135,61,242,37,136,61,252,164,136,61,62,36,137,61,184,163,137,61,106,35,138,61,84,163,138,61,118,35,139,61,209,163,139,61,99,36,140,61,45,165,140,61,46,38,141,61,104,167,141,61,218,40,142,61,131,170,142,61,100,44,143,61,125,174,143,61,206,48,144,61,86,179,144,61,23,54,145,61,14,185,145,61,62,60,146,61,165,191,146,61,67,67,147,61,26,199,147,61,39,75,148,61,109,207,148,61,234,83,149,61,158,216,149,61,138,93,150,61,173,226,150,61,7,104,151,61,153,237,151,61,98,115,152,61,99,249,152,61,155,127,153,61,10,6,154,61,176,140,154,61,142,19,155,61,163,154,155,61,239,33,156,61,114,169,156,61,44,49,157,61,29,185,157,61,69,65,158,61,165,201,158,61,59,82,159,61,8,219,159,61,13,100,160,61,72,237,160,61,186,118,161,61,99,0,162,61,67,138,162,61,90,20,163,61,167,158,163,61,43,41,164,61,230,179,164,61,216,62,165,61,0,202,165,61,95,85,166,61,245,224,166,61,193,108,167,61,196,248,167,61,254,132,168,61,110,17,169,61,20,158,169,61,241,42,170,61,4,184,170,61,78,69,171,61,206,210,171,61,133,96,172,61,113,238,172,61,149,124,173,61,238,10,174,61,126,153,174,61,67,40,175,61,63,183,175,61,114,70,176,61,218,213,176,61,120,101,177,61,77,245,177,61,88,133,178,61,152,21,179,61,15,166,179,61,187,54,180,61,158,199,180,61,182,88,181,61,4,234,181,61,137,123,182,61,67,13,183,61,50,159,183,61,88,49,184,61,179,195,184,61,68,86,185,61,11,233,185,61,7,124,186,61,57,15,187,61,160,162,187,61,61,54,188,61,16,202,188,61,24,94,189,61,85,242,189,61,200,134,190,61,112,27,191,61,78,176,191,61,97,69,192,61,170,218,192,61,39,112,193,61,218,5,194,61,194,155,194,61,224,49,195,61,50,200,195,61,186,94,196,61,119,245,196,61,104,140,197,61,143,35,198,61,235,186,198,61,124,82,199,61,66,234,199,61,61,130,200,61,108,26,201,61,209,178,201,61,106,75,202,61,57,228,202,61,59,125,203,61,115,22,204,61,224,175,204,61,129,73,205,61,86,227,205,61,97,125,206,61,159,23,207,61,19,178,207,61,187,76,208,61,151,231,208,61,168,130,209,61,237,29,210,61,103,185,210,61,21,85,211,61,248,240,211,61,14,141,212,61,89,41,213,61,216,197,213,61,140,98,214,61,115,255,214,61,143,156,215,61,223,57,216,61,99,215,216,61,27,117,217,61,7,19,218,61,38,177,218,61,122,79,219,61,2,238,219,61,189,140,220,61,173,43,221,61,208,202,221,61,39,106,222,61,178,9,223,61,112,169,223,61,98,73,224,61,136,233,224,61,226,137,225,61,111,42,226,61,47,203,226,61,35,108,227,61,74,13,228,61,165,174,228,61,52,80,229,61,245,241,229,61,234,147,230,61,19,54,231,61,110,216,231,61,253,122,232,61,191,29,233,61,180,192,233,61,221,99,234,61,56,7,235,61,199,170,235,61,136,78,236,61,125,242,236,61,164,150,237,61,255,58,238,61,140,223,238,61,76,132,239,61,63,41,240,61,101,206,240,61,189,115,241,61,73,25,242,61,7,191,242,61,247,100,243,61,26,11,244,61,112,177,244,61,248,87,245,61,179,254,245,61,160,165,246,61,192,76,247,61,18,244,247,61,151,155,248,61,77,67,249,61,55,235,249,61,82,147,250,61,159,59,251,61,31,228,251,61,209,140,252,61,181,53,253,61,203,222,253,61,19,136,254,61,141,49,255,61,57,219,255,61,140,66,0,62,148,151,0,62,181,236,0,62,238,65,1,62,65,151,1,62,173,236,1,62,49,66,2,62,206,151,2,62,132,237,2,62,83,67,3,62,59,153,3,62,59,239,3,62,84,69,4,62,134,155,4,62,209,241,4,62,52,72,5,62,176,158,5,62,68,245,5,62,242,75,6,62,183,162,6,62,150,249,6,62,141,80,7,62,156,167,7,62,196,254,7,62,5,86,8,62,94,173,8,62,207,4,9,62,89,92,9,62,252,179,9,62,183,11,10,62,138,99,10,62,118,187,10,62,122,19,11,62,150,107,11,62,203,195,11,62,24,28,12,62,125,116,12,62,250,204,12,62,144,37,13,62,62,126,13,62,4,215,13,62,227,47,14,62,217,136,14,62,232,225,14,62,15,59,15,62,78,148,15,62,165,237,15,62,20,71,16,62,155,160,16,62,58,250,16,62,241,83,17,62,193,173,17,62,168,7,18,62,167,97,18,62,190,187,18,62,237,21,19,62,51,112,19,62,146,202,19,62,9,37,20,62,151,127,20,62,61,218,20,62,251,52,21,62,209,143,21,62,190,234,21,62,195,69,22,62,224,160,22,62,21,252,22,62,97,87,23,62,197,178,23,62,64,14,24,62,211,105,24,62,126,197,24,62,64,33,25,62,26,125,25,62,11,217,25,62,20,53,26,62,52,145,26,62,108,237,26,62,187,73,27,62,34,166,27,62,160,2,28,62,53,95,28,62,226,187,28,62,166,24,29,62,129,117,29,62,116,210,29,62,126,47,30,62,159,140,30,62,215,233,30,62,39,71,31,62,141,164,31,62,11,2,32,62,160,95,32,62,76,189,32,62,16,27,33,62,234,120,33,62,219,214,33,62,228,52,34,62,3,147,34,62,58,241,34,62,135,79,35,62,235,173,35,62,103,12,36,62,249,106,36,62,162,201,36,62,98,40,37,62,56,135,37,62,38,230,37,62,42,69,38,62,69,164,38,62,119,3,39,62,192,98,39,62,31,194,39,62,149,33,40,62,33,129,40,62,197,224,40,62,126,64,41,62,79,160,41,62,54,0,42,62,51,96,42,62,72,192,42,62,114,32,43,62,179,128,43,62,11,225,43,62,121,65,44,62,253,161,44,62,152,2,45,62,73,99,45,62,16,196,45,62,238,36,46,62,226,133,46,62,237,230,46,62,13,72,47,62,68,169,47,62,145,10,48,62,245,107,48,62,110,205,48,62,254,46,49,62,163,144,49,62,95,242,49,62,49,84,50,62,25,182,50,62,23,24,51,62,43,122,51,62,85,220,51,62,148,62,52,62,234,160,52,62,86,3,53,62,216,101,53,62,111,200,53,62,28,43,54,62,223,141,54,62,184,240,54,62,167,83,55,62,171,182,55,62,197,25,56,62,245,124,56,62,59,224,56,62,150,67,57,62,7,167,57,62,141,10,58,62,41,110,58,62,219,209,58,62,162,53,59,62,126,153,59,62,112,253,59,62,120,97,60,62,149,197,60,62,199,41,61,62,15,142,61,62,108,242,61,62,222,86,62,62,102,187,62,62,3,32,63,62,181,132,63,62,125,233,63,62,90,78,64,62,75,179,64,62,83,24,65,62,111,125,65,62,160,226,65,62,231,71,66,62,66,173,66,62,179,18,67,62,57,120,67,62,211,221,67,62,131,67,68,62,71,169,68,62,33,15,69,62,15,117,69,62,18,219,69,62,42,65,70,62,87,167,70,62,153,13,71,62,240,115,71,62,91,218,71,62,219,64,72,62,111,167,72,62,25,14,73,62,215,116,73,62,169,219,73,62,144,66,74,62,140,169,74,62,157,16,75,62,193,119,75,62,251,222,75,62,73,70,76,62,171,173,76,62,34,21,77,62,173,124,77,62,76,228,77,62,0,76,78,62,200,179,78,62,164,27,79,62,149,131,79,62,154,235,79,62,179,83,80,62,225,187,80,62,34,36,81,62,120,140,81,62,225,244,81,62,95,93,82,62,241,197,82,62,151,46,83,62,81,151,83,62,31,0,84,62,1,105,84,62,247,209,84,62,0,59,85,62,30,164,85,62,79,13,86,62,149,118,86,62,238,223,86,62,91,73,87,62,219,178,87,62,112,28,88,62,24,134,88,62,211,239,88,62,163,89,89,62,134,195,89,62,124,45,90,62,134,151,90,62,164,1,91,62,213,107,91,62,26,214,91,62,114,64,92,62,221,170,92,62,92,21,93,62,239,127,93,62,148,234,93,62,77,85,94,62,26,192,94,62,249,42,95,62,236,149,95,62,242,0,96,62,11,108,96,62,55,215,96,62,119,66,97,62,202,173,97,62,47,25,98,62,168,132,98,62,52,240,98,62,210,91,99,62,132,199,99,62,73,51,100,62,32,159,100,62,11,11,101,62,8,119,101,62,24,227,101,62,59,79,102,62,113,187,102,62,186,39,103,62,21,148,103,62,131,0,104,62,3,109,104,62,151,217,104,62,60,70,105,62,245,178,105,62,192,31,106,62,157,140,106,62,141,249,106,62,144,102,107,62,165,211,107,62,204,64,108,62,6,174,108,62,82,27,109,62,176,136,109,62,33,246,109,62,164,99,110,62,57,209,110,62,225,62,111,62,154,172,111,62,102,26,112,62,68,136,112,62,52,246,112,62,55,100,113,62,75,210,113,62,113,64,114,62,169,174,114,62,243,28,115,62,80,139,115,62,190,249,115,62,61,104,116,62,207,214,116,62,115,69,117,62,40,180,117,62,239,34,118,62,200,145,118,62,179,0,119,62,175,111,119,62,189,222,119,62,221,77,120,62,14,189,120,62,80,44,121,62,165,155,121,62,10,11,122,62,130,122,122,62,10,234,122,62,164,89,123,62,80,201,123,62,13,57,124,62,219,168,124,62,186,24,125,62,171,136,125,62,173,248,125,62,192,104,126,62,228,216,126,62,26,73,127,62,96,185,127,62,220,20,128,62,16,77,128,62,77,133,128,62,147,189,128,62,225,245,128,62,55,46,129,62,150,102,129,62,253,158,129,62,109,215,129,62,229,15,130,62,102,72,130,62,238,128,130,62,128,185,130,62,25,242,130,62,187,42,131,62,102,99,131,62,24,156,131,62,211,212,131,62,150,13,132,62,98,70,132,62,53,127,132,62,17,184,132,62,245,240,132,62,226,41,133,62,214,98,133,62,211,155,133,62,216,212,133,62,229,13,134,62,250,70,134,62,23,128,134,62,61,185,134,62,106,242,134,62,160,43,135,62,221,100,135,62,35,158,135,62,112,215,135,62,198,16,136,62,35,74,136,62,137,131,136,62,247,188,136,62,108,246,136,62,233,47,137,62,111,105,137,62,252,162,137,62,145,220,137,62,46,22,138,62,211,79,138,62,127,137,138,62,52,195,138,62,240,252,138,62,180,54,139,62,128,112,139,62,84,170,139,62,47,228,139,62,18,30,140,62,253,87,140,62,239,145,140,62,233,203,140,62,235,5,141,62,245,63,141,62,6,122,141,62,31,180,141,62,63,238,141,62,103,40,142,62,150,98,142,62,205,156,142,62,12,215,142,62,82,17,143,62,159,75,143,62,245,133,143,62,81,192,143,62,181,250,143,62,33,53,144,62,147,111,144,62,14,170,144,62,143,228,144,62,25,31,145,62,169,89,145,62,65,148,145,62,224,206,145,62,134,9,146,62,52,68,146,62,233,126,146,62,165,185,146,62,105,244,146,62,52,47,147,62,6,106,147,62,223,164,147,62,191,223,147,62,167,26,148,62,150,85,148,62,139,144,148,62,136,203,148,62,140,6,149,62,152,65,149,62,170,124,149,62,195,183,149,62,227,242,149,62,11,46,150,62,57,105,150,62,111,164,150,62,171,223,150,62,238,26,151,62,56,86,151,62,138,145,151,62,226,204,151,62,65,8,152,62,167,67,152,62,19,127,152,62,135,186,152,62,1,246,152,62,130,49,153,62,10,109,153,62,153,168,153,62,47,228,153,62,203,31,154,62,110,91,154,62,24,151,154,62,200,210,154,62,127,14,155,62,61,74,155,62,2,134,155,62,205,193,155,62,158,253,155,62,119,57,156,62,85,117,156,62,59,177,156,62,39,237,156,62,25,41,157,62,18,101,157,62,18,161,157,62,24,221,157,62,36,25,158,62,55,85,158,62,80,145,158,62,112,205,158,62,150,9,159,62,195,69,159,62,246,129,159,62,47,190,159,62,111,250,159,62,180,54,160,62,1,115,160,62,83,175,160,62,172,235,160,62,11,40,161,62,112,100,161,62,219,160,161,62,77,221,161,62,196,25,162,62,66,86,162,62,198,146,162,62,81,207,162,62,225,11,163,62,119,72,163,62,20,133,163,62,182,193,163,62,95,254,163,62,13,59,164,62,194,119,164,62,125,180,164,62,61,241,164,62,4,46,165,62,208,106,165,62,162,167,165,62,123,228,165,62,89,33,166,62,61,94,166,62,39,155,166,62,23,216,166,62,12,21,167,62,7,82,167,62,8,143,167,62,15,204,167,62,28,9,168,62,46,70,168,62,70,131,168,62,100,192,168,62,136,253,168,62,177,58,169,62,223,119,169,62,20,181,169,62,78,242,169,62,141,47,170,62,211,108,170,62,29,170,170,62,109,231,170,62,195,36,171,62,31,98,171,62,127,159,171,62,230,220,171,62,81,26,172,62,194,87,172,62,57,149,172,62,181,210,172,62,54,16,173,62,189,77,173,62,73,139,173,62,218,200,173,62,113,6,174,62,13,68,174,62,174,129,174,62,85,191,174,62,0,253,174,62,177,58,175,62,103,120,175,62,35,182,175,62,227,243,175,62,169,49,176,62,116,111,176,62,68,173,176,62,25,235,176,62,243,40,177,62,210,102,177,62,182,164,177,62,160,226,177,62,142,32,178,62,129,94,178,62,121,156,178,62,119,218,178,62,121,24,179,62,128,86,179,62,140,148,179,62,157,210,179,62,178,16,180,62,205,78,180,62,236,140,180,62,16,203,180,62,57,9,181,62,103,71,181,62,154,133,181,62,209,195,181,62,13,2,182,62,78,64,182,62,147,126,182,62,221,188,182,62,44,251,182,62,127,57,183,62,215,119,183,62,52,182,183,62,149,244,183,62,251,50,184,62,101,113,184,62,212,175,184,62,71,238,184,62,191,44,185,62,59,107,185,62,188,169,185,62,65,232,185,62,202,38,186,62,88,101,186,62,235,163,186,62,129,226,186,62,28,33,187,62,188,95,187,62,95,158,187,62,7,221,187,62,180,27,188,62,100,90,188,62,25,153,188,62,210,215,188,62,143,22,189,62,80,85,189,62,22,148,189,62,223,210,189,62,173,17,190,62,127,80,190,62,85,143,190,62,47,206,190,62,13,13,191,62,239,75,191,62,213,138,191,62,191,201,191,62,173,8,192,62,159,71,192,62,149,134,192,62,143,197,192,62,141,4,193,62,143,67,193,62,148,130,193,62,158,193,193,62,171,0,194,62,188,63,194,62,209,126,194,62,234,189,194,62,6,253,194,62,38,60,195,62,74,123,195,62,113,186,195,62,157,249,195,62,204,56,196,62,254,119,196,62,52,183,196,62,110,246,196,62,171,53,197,62,236,116,197,62,49,180,197,62,121,243,197,62,196,50,198,62,19,114,198,62,102,177,198,62,188,240,198,62,21,48,199,62,114,111,199,62,210,174,199,62,54,238,199,62,157,45,200,62,7,109,200,62,117,172,200,62,230,235,200,62,90,43,201,62,209,106,201,62,76,170,201,62,202,233,201,62,75,41,202,62,208,104,202,62,88,168,202,62,226,231,202,62,112,39,203,62,1,103,203,62,149,166,203,62,45,230,203,62,199,37,204,62,100,101,204,62,4,165,204,62,168,228,204,62,78,36,205,62,248,99,205,62,164,163,205,62,83,227,205,62,5,35,206,62,186,98,206,62,114,162,206,62,45,226,206,62,234,33,207,62,171,97,207,62,110,161,207,62,52,225,207,62,253,32,208,62,200,96,208,62,150,160,208,62,103,224,208,62,59,32,209,62,17,96,209,62,234,159,209,62,198,223,209,62,164,31,210,62,133,95,210,62,104,159,210,62,78,223,210,62,55,31,211,62,33,95,211,62,15,159,211,62,255,222,211,62,241,30,212,62,230,94,212,62,221,158,212,62,215,222,212,62,211,30,213,62,209,94,213,62,210,158,213,62,213,222,213,62,219,30,214,62,226,94,214,62,236,158,214,62,248,222,214,62,7,31,215,62,24,95,215,62,42,159,215,62,63,223,215,62,87,31,216,62,112,95,216,62,139,159,216,62,169,223,216,62,200,31,217,62,234,95,217,62,14,160,217,62,51,224,217,62,91,32,218,62,133,96,218,62,176,160,218,62,222,224,218,62,13,33,219,62,63,97,219,62,114,161,219,62,167,225,219,62,222,33,220,62,23,98,220,62,82,162,220,62,142,226,220,62,204,34,221,62,12,99,221,62,78,163,221,62,146,227,221,62,215,35,222,62,29,100,222,62,102,164,222,62,176,228,222,62,252,36,223,62,73,101,223,62,152,165,223,62,232,229,223,62,58,38,224,62,142,102,224,62,227,166,224,62,57,231,224,62,145,39,225,62,234,103,225,62,69,168,225,62,161,232,225,62,255,40,226,62,94,105,226,62,190,169,226,62,32,234,226,62,131,42,227,62,231,106,227,62,76,171,227,62,179,235,227,62,27,44,228,62,132,108,228,62,238,172,228,62,90,237,228,62,199,45,229,62,52,110,229,62,163,174,229,62,19,239,229,62,133,47,230,62,247,111,230,62,106,176,230,62,222,240,230,62,83,49,231,62,202,113,231,62,65,178,231,62,185,242,231,62,50,51,232,62,172,115,232,62,38,180,232,62,162,244,232,62,31,53,233,62,156,117,233,62,26,182,233,62,153,246,233,62,25,55,234,62,153,119,234,62,26,184,234,62,156,248,234,62,31,57,235,62,162,121,235,62,38,186,235,62,170,250,235,62,47,59,236,62,181,123,236,62,59,188,236,62,194,252,236,62,73,61,237,62,209,125,237,62,89,190,237,62,226,254,237,62,107,63,238,62,245,127,238,62,127,192,238,62,10,1,239,62,149,65,239,62,32,130,239,62,171,194,239,62,55,3,240,62,196,67,240,62,80,132,240,62,221,196,240,62,106,5,241,62,247,69,241,62,132,134,241,62,18,199,241,62,160,7,242,62,45,72,242,62,187,136,242,62,74,201,242,62,216,9,243,62,102,74,243,62,244,138,243,62,131,203,243,62,17,12,244,62,159,76,244,62,46,141,244,62,188,205,244,62,74,14,245,62,216,78,245,62,102,143,245,62,244,207,245,62,129,16,246,62,15,81,246,62,156,145,246,62,41,210,246,62,182,18,247,62,67,83,247,62,207,147,247,62,91,212,247,62,231,20,248,62,115,85,248,62,254,149,248,62,136,214,248,62,19,23,249,62,157,87,249,62,38,152,249,62,175,216,249,62,56,25,250,62,192,89,250,62,72,154,250,62,207,218,250,62,86,27,251,62,220,91,251,62,97,156,251,62,230,220,251,62,106,29,252,62,238,93,252,62,113,158,252,62,243,222,252,62,117,31,253,62,245,95,253,62,118,160,253,62,245,224,253,62,116,33,254,62,241,97,254,62,110,162,254,62,235,226,254,62,102,35,255,62,224,99,255,62,90,164,255,62,211,228,255,62,165,18,0,63,225,50,0,63,27,83,0,63,86,115,0,63,144,147,0,63,201,179,0,63,2,212,0,63,58,244,0,63,114,20,1,63,169,52,1,63,224,84,1,63,22,117,1,63,76,149,1,63,129,181,1,63,181,213,1,63,233,245,1,63,28,22,2,63,78,54,2,63,128,86,2,63,178,118,2,63,226,150,2,63,18,183,2,63,65,215,2,63,112,247,2,63,157,23,3,63,203,55,3,63,247,87,3,63,35,120,3,63,78,152,3,63,120,184,3,63,161,216,3,63,202,248,3,63,242,24,4,63,25,57,4,63,63,89,4,63,101,121,4,63,137,153,4,63,173,185,4,63,208,217,4,63,243,249,4,63,20,26,5,63,52,58,5,63,84,90,5,63,115,122,5,63,145,154,5,63,173,186,5,63,202,218,5,63,229,250,5,63,255,26,6,63,24,59,6,63,48,91,6,63,72,123,6,63,94,155,6,63,116,187,6,63,136,219,6,63,155,251,6,63,174,27,7,63,191,59,7,63,208,91,7,63,223,123,7,63,237,155,7,63,250,187,7,63,7,220,7,63,18,252,7,63,28,28,8,63,37,60,8,63,44,92,8,63,51,124,8,63,57,156,8,63,61,188,8,63,64,220,8,63,67,252,8,63,68,28,9,63,68,60,9,63,66,92,9,63,64,124,9,63,60,156,9,63,55,188,9,63,49,220,9,63,41,252,9,63,33,28,10,63,23,60,10,63,12,92,10,63,255,123,10,63,242,155,10,63,227,187,10,63,211,219,10,63,193,251,10,63,174,27,11,63,154,59,11,63,133,91,11,63,110,123,11,63,86,155,11,63,60,187,11,63,33,219,11,63,5,251,11,63,231,26,12,63,200,58,12,63,168,90,12,63,134,122,12,63,98,154,12,63,62,186,12,63,23,218,12,63,240,249,12,63,199,25,13,63,156,57,13,63,112,89,13,63,66,121,13,63,19,153,13,63,227,184,13,63,176,216,13,63,125,248,13,63,72,24,14,63,17,56,14,63,216,87,14,63,159,119,14,63,99,151,14,63,38,183,14,63,232,214,14,63,167,246,14,63,101,22,15,63,34,54,15,63,221,85,15,63,150,117,15,63,78,149,15,63,4,181,15,63,184,212,15,63,106,244,15,63,27,20,16,63,202,51,16,63,120,83,16,63,36,115,16,63,206,146,16,63,118,178,16,63,28,210,16,63,193,241,16,63,100,17,17,63,6,49,17,63,165,80,17,63,67,112,17,63,223,143,17,63,121,175,17,63,17,207,17,63,167,238,17,63,60,14,18,63,206,45,18,63,95,77,18,63,238,108,18,63,123,140,18,63,7,172,18,63,144,203,18,63,23,235,18,63,157,10,19,63,32,42,19,63,162,73,19,63,34,105,19,63,159,136,19,63,27,168,19,63,149,199,19,63,13,231,19,63,131,6,20,63,247,37,20,63,104,69,20,63,216,100,20,63,70,132,20,63,178,163,20,63,27,195,20,63,131,226,20,63,233,1,21,63,76,33,21,63,174,64,21,63,13,96,21,63,106,127,21,63,197,158,21,63,31,190,21,63,117,221,21,63,202,252,21,63,29,28,22,63,109,59,22,63,188,90,22,63,8,122,22,63,82,153,22,63,153,184,22,63,223,215,22,63,34,247,22,63,100,22,23,63,162,53,23,63,223,84,23,63,26,116,23,63,82,147,23,63,136,178,23,63,187,209,23,63,237,240,23,63,28,16,24,63,73,47,24,63,115,78,24,63,155,109,24,63,193,140,24,63,228,171,24,63,6,203,24,63,36,234,24,63,65,9,25,63,91,40,25,63,115,71,25,63,136,102,25,63,155,133,25,63,171,164,25,63,185,195,25,63,197,226,25,63,206,1,26,63,213,32,26,63,217,63,26,63,219,94,26,63,218,125,26,63,215,156,26,63,210,187,26,63,202,218,26,63,191,249,26,63,178,24,27,63,162,55,27,63,144,86,27,63,123,117,27,63,100,148,27,63,74,179,27,63,46,210,27,63,15,241,27,63,237,15,28,63,201,46,28,63,162,77,28,63,121,108,28,63,77,139,28,63,31,170,28,63,237,200,28,63,185,231,28,63,131,6,29,63,74,37,29,63,14,68,29,63,207,98,29,63,142,129,29,63,74,160,29,63,3,191,29,63,186,221,29,63,110,252,29,63,31,27,30,63,205,57,30,63,121,88,30,63,34,119,30,63,200,149,30,63,107,180,30,63,12,211,30,63,170,241,30,63,69,16,31,63,221,46,31,63,114,77,31,63,5,108,31,63,148,138,31,63,33,169,31,63,171,199,31,63,50,230,31,63,182,4,32,63,56,35,32,63,182,65,32,63,50,96,32,63,170,126,32,63,32,157,32,63,147,187,32,63,3,218,32,63,112,248,32,63,218,22,33,63,65,53,33,63,165,83,33,63,6,114,33,63,100,144,33,63,191,174,33,63,23,205,33,63,108,235,33,63,190,9,34,63,13,40,34,63,89,70,34,63,162,100,34,63,232,130,34,63,43,161,34,63,107,191,34,63,167,221,34,63,225,251,34,63,24,26,35,63,75,56,35,63,123,86,35,63,168,116,35,63,211,146,35,63,249,176,35,63,29,207,35,63,62,237,35,63,91,11,36,63,118,41,36,63,141,71,36,63,161,101,36,63,177,131,36,63,191,161,36,63,201,191,36,63,208,221,36,63,212,251,36,63,213,25,37,63,210,55,37,63,204,85,37,63,195,115,37,63,183,145,37,63,167,175,37,63,148,205,37,63,126,235,37,63,101,9,38,63,72,39,38,63,40,69,38,63,4,99,38,63,221,128,38,63,179,158,38,63,134,188,38,63,85,218,38,63,33,248,38,63,233,21,39,63,174,51,39,63,112,81,39,63,46,111,39,63,233,140,39,63,160,170,39,63,84,200,39,63,4,230,39,63,178,3,40,63,91,33,40,63,1,63,40,63,164,92,40,63,67,122,40,63,223,151,40,63,120,181,40,63,12,211,40,63,158,240,40,63,43,14,41,63,182,43,41,63,60,73,41,63,192,102,41,63,63,132,41,63,187,161,41,63,52,191,41,63,169,220,41,63,26,250,41,63,136,23,42,63,242,52,42,63,89,82,42,63,188,111,42,63,28,141,42,63,119,170,42,63,208,199,42,63,36,229,42,63,117,2,43,63,194,31,43,63,12,61,43,63,82,90,43,63,148,119,43,63,211,148,43,63,14,178,43,63,69,207,43,63,120,236,43,63,168,9,44,63,212,38,44,63,252,67,44,63,33,97,44,63,66,126,44,63,95,155,44,63,120,184,44,63,142,213,44,63,159,242,44,63,173,15,45,63,184,44,45,63,190,73,45,63,193,102,45,63,191,131,45,63,186,160,45,63,177,189,45,63,165,218,45,63,148,247,45,63,128,20,46,63,103,49,46,63,75,78,46,63,43,107,46,63,7,136,46,63,224,164,46,63,180,193,46,63,132,222,46,63,81,251,46,63,26,24,47,63,222,52,47,63,159,81,47,63,92,110,47,63,21,139,47,63,202,167,47,63,123,196,47,63,40,225,47,63,209,253,47,63,118,26,48,63,23,55,48,63,180,83,48,63,77,112,48,63,226,140,48,63,115,169,48,63,0,198,48,63,137,226,48,63,14,255,48,63,142,27,49,63,11,56,49,63,132,84,49,63,248,112,49,63,105,141,49,63,214,169,49,63,62,198,49,63,162,226,49,63,2,255,49,63,95,27,50,63,182,55,50,63,10,84,50,63,90,112,50,63,166,140,50,63,237,168,50,63,48,197,50,63,111,225,50,63,170,253,50,63,225,25,51,63,19,54,51,63,66,82,51,63,108,110,51,63,146,138,51,63,180,166,51,63,209,194,51,63,234,222,51,63,0,251,51,63,16,23,52,63,29,51,52,63,37,79,52,63,41,107,52,63,41,135,52,63,37,163,52,63,28,191,52,63,15,219,52,63,253,246,52,63,232,18,53,63,206,46,53,63,176,74,53,63,141,102,53,63,102,130,53,63,59,158,53,63,11,186,53,63,215,213,53,63,159,241,53,63,98,13,54,63,33,41,54,63,220,68,54,63,146,96,54,63,68,124,54,63,241,151,54,63,154,179,54,63,63,207,54,63,223,234,54,63,123,6,55,63,18,34,55,63,165,61,55,63,52,89,55,63,190,116,55,63,67,144,55,63,196,171,55,63,65,199,55,63,185,226,55,63,45,254,55,63,156,25,56,63,7,53,56,63,109,80,56,63,207,107,56,63,44,135,56,63,133,162,56,63,217,189,56,63,40,217,56,63,115,244,56,63,186,15,57,63,252,42,57,63,57,70,57,63,114,97,57,63,166,124,57,63,214,151,57,63,1,179,57,63,40,206,57,63,74,233,57,63,103,4,58,63,128,31,58,63,148,58,58,63,163,85,58,63,174,112,58,63,180,139,58,63,182,166,58,63,179,193,58,63,171,220,58,63,159,247,58,63,142,18,59,63,120,45,59,63,94,72,59,63,63,99,59,63,27,126,59,63,243,152,59,63,197,179,59,63,148,206,59,63,93,233,59,63,34,4,60,63,226,30,60,63,157,57,60,63,84,84,60,63,5,111,60,63,178,137,60,63,91,164,60,63,254,190,60,63,157,217,60,63,55,244,60,63,204,14,61,63,93,41,61,63,232,67,61,63,111,94,61,63,241,120,61,63,110,147,61,63,231,173,61,63,91,200,61,63,201,226,61,63,51,253,61,63,152,23,62,63,249,49,62,63,84,76,62,63,171,102,62,63,252,128,62,63,73,155,62,63,145,181,62,63,212,207,62,63,19,234,62,63,76,4,63,63,128,30,63,63,176,56,63,63,219,82,63,63,0,109,63,63,33,135,63,63,61,161,63,63,84,187,63,63,102,213,63,63,115,239,63,63,123,9,64,63,127,35,64,63,125,61,64,63,118,87,64,63,106,113,64,63,90,139,64,63,68,165,64,63,42,191,64,63,10,217,64,63,229,242,64,63,188,12,65,63,141,38,65,63,90,64,65,63,33,90,65,63,228,115,65,63,161,141,65,63,89,167,65,63,13,193,65,63,187,218,65,63,100,244,65,63,8,14,66,63,167,39,66,63,65,65,66,63,214,90,66,63,102,116,66,63,241,141,66,63,119,167,66,63,248,192,66,63,115,218,66,63,234,243,66,63,91,13,67,63,199,38,67,63,47,64,67,63,145,89,67,63,238,114,67,63,69,140,67,63,152,165,67,63,230,190,67,63,46,216,67,63,113,241,67,63,175,10,68,63,232,35,68,63,28,61,68,63,75,86,68,63,116,111,68,63,153,136,68,63,184,161,68,63,210,186,68,63,230,211,68,63,246,236,68,63,0,6,69,63,5,31,69,63,5,56,69,63,0,81,69,63,245,105,69,63,230,130,69,63,209,155,69,63,182,180,69,63,151,205,69,63,114,230,69,63,72,255,69,63,25,24,70,63,229,48,70,63,171,73,70,63,108,98,70,63,40,123,70,63,222,147,70,63,143,172,70,63,59,197,70,63,226,221,70,63,131,246,70,63,31,15,71,63,182,39,71,63,71,64,71,63,211,88,71,63,90,113,71,63,220,137,71,63,88,162,71,63,207,186,71,63,64,211,71,63,172,235,71,63,19,4,72,63,116,28,72,63,209,52,72,63,39,77,72,63,121,101,72,63,197,125,72,63,11,150,72,63,77,174,72,63,137,198,72,63,191,222,72,63,240,246,72,63,28,15,73,63,66,39,73,63,99,63,73,63,127,87,73,63,149,111,73,63,166,135,73,63,177,159,73,63,183,183,73,63,183,207,73,63,178,231,73,63,168,255,73,63,152,23,74,63,131,47,74,63,104,71,74,63,72,95,74,63,34,119,74,63,247,142,74,63,199,166,74,63,145,190,74,63,85,214,74,63,20,238,74,63,206,5,75,63,130,29,75,63,49,53,75,63,218,76,75,63,126,100,75,63,28,124,75,63,181,147,75,63,72,171,75,63,213,194,75,63,93,218,75,63,224,241,75,63,93,9,76,63,213,32,76,63,71,56,76,63,179,79,76,63,26,103,76,63,124,126,76,63,216,149,76,63,46,173,76,63,127,196,76,63,202,219,76,63,16,243,76,63,80,10,77,63,139,33,77,63,192,56,77,63,240,79,77,63,26,103,77,63,62,126,77,63,93,149,77,63,118,172,77,63,137,195,77,63,151,218,77,63,160,241,77,63,163,8,78,63,160,31,78,63,151,54,78,63,137,77,78,63,118,100,78,63,93,123,78,63,62,146,78,63,25,169,78,63,239,191,78,63,192,214,78,63,138,237,78,63,79,4,79,63,15,27,79,63,201,49,79,63,125,72,79,63,43,95,79,63,212,117,79,63,119,140,79,63,21,163,79,63,172,185,79,63,63,208,79,63,203,230,79,63,82,253,79,63,211,19,80,63,79,42,80,63,197,64,80,63,53,87,80,63,159,109,80,63,4,132,80,63,99,154,80,63,189,176,80,63,16,199,80,63,94,221,80,63,167,243,80,63,233,9,81,63,38,32,81,63,93,54,81,63,143,76,81,63,187,98,81,63,225,120,81,63,1,143,81,63,28,165,81,63,48,187,81,63,64,209,81,63,73,231,81,63,77,253,81,63,75,19,82,63,67,41,82,63,53,63,82,63,34,85,82,63,9,107,82,63,234,128,82,63,198,150,82,63,155,172,82,63,107,194,82,63,53,216,82,63,250,237,82,63,185,3,83,63,113,25,83,63,37,47,83,63,210,68,83,63,121,90,83,63,27,112,83,63,183,133,83,63,77,155,83,63,222,176,83,63,104,198,83,63,237,219,83,63,108,241,83,63,230,6,84,63,89,28,84,63,199,49,84,63,46,71,84,63,145,92,84,63,237,113,84,63,67,135,84,63,148,156,84,63,223,177,84,63,35,199,84,63,99,220,84,63,156,241,84,63,207,6,85,63,253,27,85,63,37,49,85,63,71,70,85,63,99,91,85,63,121,112,85,63,138,133,85,63,149,154,85,63,153,175,85,63,152,196,85,63,146,217,85,63,133,238,85,63,114,3,86,63,90,24,86,63,60,45,86,63,24,66,86,63,238,86,86,63,190,107,86,63,136,128,86,63,76,149,86,63,11,170,86,63,196,190,86,63,118,211,86,63,35,232,86,63,203,252,86,63,108,17,87,63,7,38,87,63,156,58,87,63,44,79,87,63,182,99,87,63,58,120,87,63,183,140,87,63,47,161,87,63,162,181,87,63,14,202,87,63,116,222,87,63,213,242,87,63,47,7,88,63,132,27,88,63,211,47,88,63,28,68,88,63,95,88,88,63,156,108,88,63,211,128,88,63,4,149,88,63,47,169,88,63,85,189,88,63,116,209,88,63,142,229,88,63,162,249,88,63,175,13,89,63,183,33,89,63,185,53,89,63,181,73,89,63,171,93,89,63,155,113,89,63,134,133,89,63,106,153,89,63,72,173,89,63,33,193,89,63,243,212,89,63,192,232,89,63,135,252,89,63,71,16,90,63,2,36,90,63,183,55,90,63,102,75,90,63,15,95,90,63,178,114,90,63,79,134,90,63,230,153,90,63,119,173,90,63,3,193,90,63,136,212,90,63,7,232,90,63,129,251,90,63,244,14,91,63,98,34,91,63,201,53,91,63,43,73,91,63,135,92,91,63,220,111,91,63,44,131,91,63,118,150,91,63,186,169,91,63,248,188,91,63,47,208,91,63,97,227,91,63,141,246,91,63,179,9,92,63,212,28,92,63,238,47,92,63,2,67,92,63,16,86,92,63,24,105,92,63,26,124,92,63,23,143,92,63,13,162,92,63,253,180,92,63,232,199,92,63,204,218,92,63,171,237,92,63,131,0,93,63,86,19,93,63,34,38,93,63,233,56,93,63,169,75,93,63,100,94,93,63,24,113,93,63,199,131,93,63,112,150,93,63,18,169,93,63,175,187,93,63,70,206,93,63,215,224,93,63,97,243,93,63,230,5,94,63,101,24,94,63,222,42,94,63,81,61,94,63,190,79,94,63,36,98,94,63,133,116,94,63,224,134,94,63,53,153,94,63,132,171,94,63,205,189,94,63,16,208,94,63,77,226,94,63,132,244,94,63,181,6,95,63,224,24,95,63,5,43,95,63,36,61,95,63,61,79,95,63,80,97,95,63,93,115,95,63,101,133,95,63,102,151,95,63,97,169,95,63,86,187,95,63,69,205,95,63,46,223,95,63,18,241,95,63,239,2,96,63,198,20,96,63,151,38,96,63,98,56,96,63,40,74,96,63,231,91,96,63,160,109,96,63,84,127,96,63,1,145,96,63,168,162,96,63,73,180,96,63,229,197,96,63,122,215,96,63,10,233,96,63,147,250,96,63,22,12,97,63,148,29,97,63,11,47,97,63,125,64,97,63,232,81,97,63,77,99,97,63,173,116,97,63,6,134,97,63,90,151,97,63,167,168,97,63,239,185,97,63,48,203,97,63,108,220,97,63,162,237,97,63,209,254,97,63,251,15,98,63,30,33,98,63,60,50,98,63,84,67,98,63,101,84,98,63,113,101,98,63,119,118,98,63,119,135,98,63,112,152,98,63,100,169,98,63,82,186,98,63,58,203,98,63,28,220,98,63,247,236,98,63,205,253,98,63,157,14,99,63,103,31,99,63,43,48,99,63,233,64,99,63,161,81,99,63,83,98,99,63,255,114,99,63,165,131,99,63,69,148,99,63,224,164,99,63,116,181,99,63,2,198,99,63,138,214,99,63,13,231,99,63,137,247,99,63,255,7,100,63,112,24,100,63,218,40,100,63,62,57,100,63,157,73,100,63,246,89,100,63,72,106,100,63,149,122,100,63,219,138,100,63,28,155,100,63,87,171,100,63,140,187,100,63,186,203,100,63,227,219,100,63,6,236,100,63,35,252,100,63,58,12,101,63,75,28,101,63,86,44,101,63,91,60,101,63,91,76,101,63,84,92,101,63,71,108,101,63,53,124,101,63,28,140,101,63,254,155,101,63,217,171,101,63,175,187,101,63,126,203,101,63,72,219,101,63,12,235,101,63,202,250,101,63,130,10,102,63,52,26,102,63,224,41,102,63,134,57,102,63,38,73,102,63,193,88,102,63,85,104,102,63,227,119,102,63,108,135,102,63,238,150,102,63,107,166,102,63,226,181,102,63,83,197,102,63,190,212,102,63,35,228,102,63,130,243,102,63,219,2,103,63,46,18,103,63,124,33,103,63,195,48,103,63,5,64,103,63,64,79,103,63,118,94,103,63,166,109,103,63,208,124,103,63,244,139,103,63,18,155,103,63,42,170,103,63,61,185,103,63,73,200,103,63,80,215,103,63,80,230,103,63,75,245,103,63,64,4,104,63,47,19,104,63,24,34,104,63,251,48,104,63,217,63,104,63,176,78,104,63,130,93,104,63,78,108,104,63,20,123,104,63,212,137,104,63,142,152,104,63,66,167,104,63,240,181,104,63,153,196,104,63,60,211,104,63,217,225,104,63,112,240,104,63,1,255,104,63,140,13,105,63,17,28,105,63,145,42,105,63,11,57,105,63,127,71,105,63,237,85,105,63,85,100,105,63,183,114,105,63,20,129,105,63,106,143,105,63,187,157,105,63,6,172,105,63,75,186,105,63,139,200,105,63,196,214,105,63,248,228,105,63,38,243,105,63,78,1,106,63,112,15,106,63,141,29,106,63,163,43,106,63,180,57,106,63,191,71,106,63,196,85,106,63,196,99,106,63,189,113,106,63,177,127,106,63,159,141,106,63,135,155,106,63,106,169,106,63,70,183,106,63,29,197,106,63,238,210,106,63,186,224,106,63,127,238,106,63,63,252,106,63,249,9,107,63,173,23,107,63,91,37,107,63,4,51,107,63,167,64,107,63,68,78,107,63,219,91,107,63,109,105,107,63,249,118,107,63,127,132,107,63,255,145,107,63,122,159,107,63,238,172,107,63,94,186,107,63,199,199,107,63,42,213,107,63,136,226,107,63,224,239,107,63,51,253,107,63,128,10,108,63,198,23,108,63,8,37,108,63,67,50,108,63,121,63,108,63,169,76,108,63,211,89,108,63,248,102,108,63,23,116,108,63,48,129,108,63,68,142,108,63,82,155,108,63,90,168,108,63,92,181,108,63,89,194,108,63,80,207,108,63,65,220,108,63,45,233,108,63,19,246,108,63,243,2,109,63,206,15,109,63,163,28,109,63,114,41,109,63,60,54,109,63,0,67,109,63,190,79,109,63,119,92,109,63,42,105,109,63,215,117,109,63,127,130,109,63,33,143,109,63,189,155,109,63,84,168,109,63,229,180,109,63,113,193,109,63,247,205,109,63,119,218,109,63,242,230,109,63,103,243,109,63,214,255,109,63,64,12,110,63,164,24,110,63,3,37,110,63,91,49,110,63,175,61,110,63,253,73,110,63,69,86,110,63,135,98,110,63,196,110,110,63,252,122,110,63,45,135,110,63,90,147,110,63,128,159,110,63,161,171,110,63,189,183,110,63,211,195,110,63,227,207,110,63,238,219,110,63,243,231,110,63,243,243,110,63,237,255,110,63,226,11,111,63,209,23,111,63,186,35,111,63,158,47,111,63,125,59,111,63,85,71,111,63,41,83,111,63,247,94,111,63,191,106,111,63,130,118,111,63,63,130,111,63,247,141,111,63,169,153,111,63,86,165,111,63,253,176,111,63,159,188,111,63,59,200,111,63,210,211,111,63,99,223,111,63,239,234,111,63,117,246,111,63,246,1,112,63,114,13,112,63,231,24,112,63,88,36,112,63,195,47,112,63,40,59,112,63,137,70,112,63,227,81,112,63,56,93,112,63,136,104,112,63,210,115,112,63,23,127,112,63,87,138,112,63,145,149,112,63,197,160,112,63,244,171,112,63,30,183,112,63,66,194,112,63,97,205,112,63,123,216,112,63,143,227,112,63,157,238,112,63,167,249,112,63,171,4,113,63,169,15,113,63,162,26,113,63,150,37,113,63,132,48,113,63,109,59,113,63,81,70,113,63,47,81,113,63,8,92,113,63,219,102,113,63,170,113,113,63,114,124,113,63,54,135,113,63,244,145,113,63,173,156,113,63,96,167,113,63,14,178,113,63,183,188,113,63,91,199,113,63,249,209,113,63,146,220,113,63,37,231,113,63,179,241,113,63,60,252,113,63,192,6,114,63,62,17,114,63,183,27,114,63,43,38,114,63,154,48,114,63,3,59,114,63,103,69,114,63,197,79,114,63,31,90,114,63,115,100,114,63,194,110,114,63,11,121,114,63,79,131,114,63,143,141,114,63,200,151,114,63,253,161,114,63,44,172,114,63,87,182,114,63,123,192,114,63,155,202,114,63,182,212,114,63,203,222,114,63,219,232,114,63,230,242,114,63,235,252,114,63,236,6,115,63,231,16,115,63,221,26,115,63,206,36,115,63,186,46,115,63,160,56,115,63,130,66,115,63,94,76,115,63,53,86,115,63,7,96,115,63,212,105,115,63,155,115,115,63,94,125,115,63,27,135,115,63,211,144,115,63,134,154,115,63,52,164,115,63,221,173,115,63,128,183,115,63,31,193,115,63,184,202,115,63,77,212,115,63,220,221,115,63,102,231,115,63,235,240,115,63,107,250,115,63,230,3,116,63,92,13,116,63,204,22,116,63,56,32,116,63,159,41,116,63,0,51,116,63,93,60,116,63,180,69,116,63,6,79,116,63,84,88,116,63,156,97,116,63,223,106,116,63,29,116,116,63,87,125,116,63,139,134,116,63,186,143,116,63,228,152,116,63,9,162,116,63,41,171,116,63,68,180,116,63,91,189,116,63,108,198,116,63,120,207,116,63,127,216,116,63,129,225,116,63,127,234,116,63,119,243,116,63,106,252,116,63,89,5,117,63,66,14,117,63,38,23,117,63,6,32,117,63,225,40,117,63,182,49,117,63,135,58,117,63,83,67,117,63,26,76,117,63,220,84,117,63,153,93,117,63,81,102,117,63,4,111,117,63,179,119,117,63,92,128,117,63,1,137,117,63,160,145,117,63,59,154,117,63,209,162,117,63,98,171,117,63,239,179,117,63,118,188,117,63,249,196,117,63,118,205,117,63,239,213,117,63,99,222,117,63,210,230,117,63,61,239,117,63,162,247,117,63,3,0,118,63,95,8,118,63,182,16,118,63,8,25,118,63,86,33,118,63,159,41,118,63,227,49,118,63,34,58,118,63,92,66,118,63,146,74,118,63,195,82,118,63,239,90,118,63,22,99,118,63,57,107,118,63,86,115,118,63,112,123,118,63,132,131,118,63,148,139,118,63,158,147,118,63,165,155,118,63,166,163,118,63,163,171,118,63,155,179,118,63,142,187,118,63,125,195,118,63,103,203,118,63,76,211,118,63,45,219,118,63,9,227,118,63,224,234,118,63,178,242,118,63,128,250,118,63,74,2,119,63,14,10,119,63,206,17,119,63,137,25,119,63,64,33,119,63,242,40,119,63,160,48,119,63,72,56,119,63,237,63,119,63,140,71,119,63,39,79,119,63,190,86,119,63,79,94,119,63,220,101,119,63,101,109,119,63,233,116,119,63,105,124,119,63,228,131,119,63,90,139,119,63,204,146,119,63,57,154,119,63,162,161,119,63,6,169,119,63,101,176,119,63,192,183,119,63,23,191,119,63,105,198,119,63,182,205,119,63,255,212,119,63,68,220,119,63,132,227,119,63,191,234,119,63,246,241,119,63,41,249,119,63,87,0,120,63,129,7,120,63,166,14,120,63,198,21,120,63,227,28,120,63,250,35,120,63,14,43,120,63,28,50,120,63,39,57,120,63,45,64,120,63,46,71,120,63,44,78,120,63,36,85,120,63,25,92,120,63,9,99,120,63,244,105,120,63,219,112,120,63,190,119,120,63,156,126,120,63,118,133,120,63,76,140,120,63,29,147,120,63,234,153,120,63,179,160,120,63,119,167,120,63,55,174,120,63,242,180,120,63,169,187,120,63,92,194,120,63,11,201,120,63,181,207,120,63,91,214,120,63,252,220,120,63,154,227,120,63,51,234,120,63,199,240,120,63,88,247,120,63,228,253,120,63,108,4,121,63,240,10,121,63,111,17,121,63,234,23,121,63,97,30,121,63,211,36,121,63,66,43,121,63,172,49,121,63,18,56,121,63,116,62,121,63,209,68,121,63,42,75,121,63,127,81,121,63,208,87,121,63,29,94,121,63,101,100,121,63,170,106,121,63,234,112,121,63,38,119,121,63,93,125,121,63,145,131,121,63,193,137,121,63,236,143,121,63,19,150,121,63,54,156,121,63,85,162,121,63,112,168,121,63,134,174,121,63,153,180,121,63,167,186,121,63,178,192,121,63,184,198,121,63,186,204,121,63,184,210,121,63,178,216,121,63,168,222,121,63,154,228,121,63,135,234,121,63,113,240,121,63,87,246,121,63,56,252,121,63,22,2,122,63,239,7,122,63,197,13,122,63,150,19,122,63,100,25,122,63,45,31,122,63,243,36,122,63,180,42,122,63,113,48,122,63,43,54,122,63,224,59,122,63,146,65,122,63,63,71,122,63,233,76,122,63,142,82,122,63,48,88,122,63,206,93,122,63,103,99,122,63,253,104,122,63,143,110,122,63,29,116,122,63,167,121,122,63,45,127,122,63,175,132,122,63,45,138,122,63,168,143,122,63,30,149,122,63,145,154,122,63,255,159,122,63,106,165,122,63,209,170,122,63,52,176,122,63,147,181,122,63,239,186,122,63,70,192,122,63,154,197,122,63,234,202,122,63,54,208,122,63,126,213,122,63,194,218,122,63,3,224,122,63,64,229,122,63,121,234,122,63,174,239,122,63,223,244,122,63,13,250,122,63,55,255,122,63,93,4,123,63,127,9,123,63,157,14,123,63,184,19,123,63,207,24,123,63,227,29,123,63,242,34,123,63,254,39,123,63,6,45,123,63,10,50,123,63,11,55,123,63,8,60,123,63,1,65,123,63,247,69,123,63,233,74,123,63,215,79,123,63,193,84,123,63,168,89,123,63,139,94,123,63,107,99,123,63,71,104,123,63,31,109,123,63,243,113,123,63,196,118,123,63,146,123,123,63,91,128,123,63,33,133,123,63,228,137,123,63,163,142,123,63,94,147,123,63,22,152,123,63,202,156,123,63,122,161,123,63,39,166,123,63,208,170,123,63,118,175,123,63,24,180,123,63,183,184,123,63,82,189,123,63,233,193,123,63,125,198,123,63,14,203,123,63,155,207,123,63,36,212,123,63,170,216,123,63,45,221,123,63,172,225,123,63,39,230,123,63,159,234,123,63,19,239,123,63,132,243,123,63,242,247,123,63,92,252,123,63,195,0,124,63,38,5,124,63,133,9,124,63,226,13,124,63,58,18,124,63,144,22,124,63,226,26,124,63,48,31,124,63,123,35,124,63,195,39,124,63,7,44,124,63,72,48,124,63,134,52,124,63,192,56,124,63,247,60,124,63,42,65,124,63,90,69,124,63,135,73,124,63,176,77,124,63,214,81,124,63,249,85,124,63,24,90,124,63,52,94,124,63,77,98,124,63,98,102,124,63,116,106,124,63,131,110,124,63,142,114,124,63,150,118,124,63,155,122,124,63,157,126,124,63,155,130,124,63,150,134,124,63,142,138,124,63,130,142,124,63,116,146,124,63,98,150,124,63,77,154,124,63,52,158,124,63,24,162,124,63,249,165,124,63,215,169,124,63,178,173,124,63,137,177,124,63,94,181,124,63,47,185,124,63,253,188,124,63,199,192,124,63,143,196,124,63,83,200,124,63,20,204,124,63,211,207,124,63,141,211,124,63,69,215,124,63,250,218,124,63,171,222,124,63,90,226,124,63,5,230,124,63,173,233,124,63,82,237,124,63,244,240,124,63,147,244,124,63,46,248,124,63,199,251,124,63,93,255,124,63,239,2,125,63,127,6,125,63,11,10,125,63,148,13,125,63,27,17,125,63,158,20,125,63,30,24,125,63,155,27,125,63,21,31,125,63,140,34,125,63,0,38,125,63,114,41,125,63,224,44,125,63,75,48,125,63,179,51,125,63,24,55,125,63,122,58,125,63,217,61,125,63,54,65,125,63,143,68,125,63,229,71,125,63,56,75,125,63,137,78,125,63,214,81,125,63,33,85,125,63,104,88,125,63,173,91,125,63,239,94,125,63,46,98,125,63,106,101,125,63,163,104,125,63,217,107,125,63,12,111,125,63,61,114,125,63,106,117,125,63,149,120,125,63,189,123,125,63,226,126,125,63,4,130,125,63,36,133,125,63,64,136,125,63,90,139,125,63,112,142,125,63,133,145,125,63,150,148,125,63,164,151,125,63,176,154,125,63,185,157,125,63,191,160,125,63,194,163,125,63,194,166,125,63,192,169,125,63,187,172,125,63,179,175,125,63,168,178,125,63,155,181,125,63,139,184,125,63,120,187,125,63,99,190,125,63,74,193,125,63,48,196,125,63,18,199,125,63,241,201,125,63,206,204,125,63,169,207,125,63,128,210,125,63,85,213,125,63,39,216,125,63,247,218,125,63,196,221,125,63,142,224,125,63,85,227,125,63,26,230,125,63,220,232,125,63,156,235,125,63,89,238,125,63,19,241,125,63,203,243,125,63,128,246,125,63,51,249,125,63,227,251,125,63,144,254,125,63,59,1,126,63,227,3,126,63,137,6,126,63,44,9,126,63,204,11,126,63,106,14,126,63,6,17,126,63,158,19,126,63,53,22,126,63,200,24,126,63,90,27,126,63,232,29,126,63,116,32,126,63,254,34,126,63,133,37,126,63,10,40,126,63,140,42,126,63,12,45,126,63,137,47,126,63,4,50,126,63,124,52,126,63,242,54,126,63,101,57,126,63,214,59,126,63,68,62,126,63,176,64,126,63,26,67,126,63,129,69,126,63,230,71,126,63,72,74,126,63,168,76,126,63,5,79,126,63,96,81,126,63,185,83,126,63,15,86,126,63,99,88,126,63,181,90,126,63,4,93,126,63,81,95,126,63,155,97,126,63,227,99,126,63,41,102,126,63,108,104,126,63,173,106,126,63,236,108,126,63,40,111,126,63,98,113,126,63,154,115,126,63,208,117,126,63,3,120,126,63,51,122,126,63,98,124,126,63,142,126,126,63,184,128,126,63,224,130,126,63,5,133,126,63,40,135,126,63,73,137,126,63,104,139,126,63,132,141,126,63,159,143,126,63,183,145,126,63,204,147,126,63,224,149,126,63,241,151,126,63,0,154,126,63,13,156,126,63,24,158,126,63,32,160,126,63,38,162,126,63,42,164,126,63,44,166,126,63,44,168,126,63,41,170,126,63,37,172,126,63,30,174,126,63,21,176,126,63,10,178,126,63,253,179,126,63,238,181,126,63,220,183,126,63,201,185,126,63,179,187,126,63,155,189,126,63,129,191,126,63,101,193,126,63,71,195,126,63,39,197,126,63,5,199,126,63,224,200,126,63,186,202,126,63,145,204,126,63,103,206,126,63,58,208,126,63,12,210,126,63,219,211,126,63,168,213,126,63,115,215,126,63,61,217,126,63,4,219,126,63,201,220,126,63,140,222,126,63,77,224,126,63,12,226,126,63,202,227,126,63,133,229,126,63,62,231,126,63,245,232,126,63,170,234,126,63,94,236,126,63,15,238,126,63,190,239,126,63,108,241,126,63,23,243,126,63,193,244,126,63,104,246,126,63,14,248,126,63,178,249,126,63,84,251,126,63,243,252,126,63,145,254,126,63,46,0,127,63,200,1,127,63,96,3,127,63,247,4,127,63,139,6,127,63,30,8,127,63,175,9,127,63,62,11,127,63,203,12,127,63,86,14,127,63,223,15,127,63,103,17,127,63,237,18,127,63,112,20,127,63,242,21,127,63,115,23,127,63,241,24,127,63,110,26,127,63,233,27,127,63,98,29,127,63,217,30,127,63,78,32,127,63,194,33,127,63,52,35,127,63,164,36,127,63,18,38,127,63,127,39,127,63,234,40,127,63,83,42,127,63,186,43,127,63,32,45,127,63,131,46,127,63,230,47,127,63,70,49,127,63,165,50,127,63,2,52,127,63,93,53,127,63,182,54,127,63,14,56,127,63,100,57,127,63,185,58,127,63,12,60,127,63,93,61,127,63,172,62,127,63,250,63,127,63,70,65,127,63,145,66,127,63,217,67,127,63,33,69,127,63,102,70,127,63,170,71,127,63,236,72,127,63,45,74,127,63,108,75,127,63,169,76,127,63,229,77,127,63,31,79,127,63,88,80,127,63,143,81,127,63,196,82,127,63,248,83,127,63,42,85,127,63,91,86,127,63,138,87,127,63,184,88,127,63,228,89,127,63,14,91,127,63,55,92,127,63,94,93,127,63,132,94,127,63,169,95,127,63,203,96,127,63,237,97,127,63,12,99,127,63,42,100,127,63,71,101,127,63,98,102,127,63,124,103,127,63,148,104,127,63,171,105,127,63,192,106,127,63,212,107,127,63,230,108,127,63,247,109,127,63,6,111,127,63,20,112,127,63,33,113,127,63,44,114,127,63,53,115,127,63,61,116,127,63,68,117,127,63,73,118,127,63,77,119,127,63,79,120,127,63,80,121,127,63,80,122,127,63,78,123,127,63,75,124,127,63,70,125,127,63,64,126,127,63,57,127,127,63,48,128,127,63,38,129,127,63,27,130,127,63,14,131,127,63,0,132,127,63,240,132,127,63,223,133,127,63,205,134,127,63,185,135,127,63,164,136,127,63,142,137,127,63,118,138,127,63,93,139,127,63,67,140,127,63,40,141,127,63,11,142,127,63,237,142,127,63,205,143,127,63,173,144,127,63,139,145,127,63,103,146,127,63,67,147,127,63,29,148,127,63,246,148,127,63,205,149,127,63,164,150,127,63,121,151,127,63,77,152,127,63,31,153,127,63,241,153,127,63,193,154,127,63,144,155,127,63,93,156,127,63,42,157,127,63,245,157,127,63,191,158,127,63,136,159,127,63,79,160,127,63,22,161,127,63,219,161,127,63,159,162,127,63,98,163,127,63,36,164,127,63,228,164,127,63,163,165,127,63,98,166,127,63,31,167,127,63,219,167,127,63,149,168,127,63,79,169,127,63,7,170,127,63,190,170,127,63,117,171,127,63,42,172,127,63,221,172,127,63,144,173,127,63,66,174,127,63,242,174,127,63,162,175,127,63,80,176,127,63,253,176,127,63,169,177,127,63,85,178,127,63,254,178,127,63,167,179,127,63,79,180,127,63,246,180,127,63,156,181,127,63,64,182,127,63,228,182,127,63,134,183,127,63,40,184,127,63,200,184,127,63,103,185,127,63,6,186,127,63,163,186,127,63,63,187,127,63,219,187,127,63,117,188,127,63,14,189,127,63,166,189,127,63,61,190,127,63,212,190,127,63,105,191,127,63,253,191,127,63,144,192,127,63,34,193,127,63,180,193,127,63,68,194,127,63,211,194,127,63,98,195,127,63,239,195,127,63,123,196,127,63,7,197,127,63,145,197,127,63,27,198,127,63,163,198,127,63,43,199,127,63,178,199,127,63,56,200,127,63,189,200,127,63,65,201,127,63,196,201,127,63,70,202,127,63,199,202,127,63,71,203,127,63,199,203,127,63,69,204,127,63,195,204,127,63,64,205,127,63,187,205,127,63,54,206,127,63,177,206,127,63,42,207,127,63,162,207,127,63,26,208,127,63,144,208,127,63,6,209,127,63,123,209,127,63,239,209,127,63,98,210,127,63,213,210,127,63,70,211,127,63,183,211,127,63,39,212,127,63,150,212,127,63,4,213,127,63,114,213,127,63,222,213,127,63,74,214,127,63,181,214,127,63,32,215,127,63,137,215,127,63,242,215,127,63,89,216,127,63,192,216,127,63,39,217,127,63,140,217,127,63,241,217,127,63,85,218,127,63,184,218,127,63,27,219,127,63,124,219,127,63,221,219,127,63,61,220,127,63,157,220,127,63,251,220,127,63,89,221,127,63,183,221,127,63,19,222,127,63,111,222,127,63,202,222,127,63,36,223,127,63,126,223,127,63,215,223,127,63,47,224,127,63,134,224,127,63,221,224,127,63,51,225,127,63,137,225,127,63,221,225,127,63,49,226,127,63,133,226,127,63,215,226,127,63,41,227,127,63,122,227,127,63,203,227,127,63,27,228,127,63,106,228,127,63,185,228,127,63,7,229,127,63,84,229,127,63,161,229,127,63,237,229,127,63,56,230,127,63,131,230,127,63,205,230,127,63,23,231,127,63,96,231,127,63,168,231,127,63,239,231,127,63,54,232,127,63,125,232,127,63,195,232,127,63,8,233,127,63,76,233,127,63,144,233,127,63,212,233,127,63,23,234,127,63,89,234,127,63,154,234,127,63,219,234,127,63,28,235,127,63,92,235,127,63,155,235,127,63,218,235,127,63,24,236,127,63,86,236,127,63,147,236,127,63,207,236,127,63,11,237,127,63,71,237,127,63,130,237,127,63,188,237,127,63,246,237,127,63,47,238,127,63,104,238,127,63,160,238,127,63,216,238,127,63,15,239,127,63,69,239,127,63,123,239,127,63,177,239,127,63,230,239,127,63,27,240,127,63,79,240,127,63,130,240,127,63,182,240,127,63,232,240,127,63,26,241,127,63,76,241,127,63,125,241,127,63,174,241,127,63,222,241,127,63,14,242,127,63,61,242,127,63,108,242,127,63,154,242,127,63,200,242,127,63,245,242,127,63,34,243,127,63,79,243,127,63,123,243,127,63,166,243,127,63,209,243,127,63,252,243,127,63,38,244,127,63,80,244,127,63,121,244,127,63,162,244,127,63,203,244,127,63,243,244,127,63,27,245,127,63,66,245,127,63,105,245,127,63,143,245,127,63,181,245,127,63,219,245,127,63,0,246,127,63,37,246,127,63,73,246,127,63,109,246,127,63,145,246,127,63,180,246,127,63,215,246,127,63,250,246,127,63,28,247,127,63,62,247,127,63,95,247,127,63,128,247,127,63,160,247,127,63,193,247,127,63,225,247,127,63,0,248,127,63,31,248,127,63,62,248,127,63,93,248,127,63,123,248,127,63,152,248,127,63,182,248,127,63,211,248,127,63,240,248,127,63,12,249,127,63,40,249,127,63,68,249,127,63,95,249,127,63,122,249,127,63,149,249,127,63,175,249,127,63,202,249,127,63,227,249,127,63,253,249,127,63,22,250,127,63,47,250,127,63,71,250,127,63,96,250,127,63,120,250,127,63,143,250,127,63,166,250,127,63,190,250,127,63,212,250,127,63,235,250,127,63,1,251,127,63,23,251,127,63,44,251,127,63,66,251,127,63,87,251,127,63,108,251,127,63,128,251,127,63,148,251,127,63,168,251,127,63,188,251,127,63,208,251,127,63,227,251,127,63,246,251,127,63,8,252,127,63,27,252,127,63,45,252,127,63,63,252,127,63,81,252,127,63,98,252,127,63,115,252,127,63,132,252,127,63,149,252,127,63,165,252,127,63,182,252,127,63,198,252,127,63,213,252,127,63,229,252,127,63,244,252,127,63,3,253,127,63,18,253,127,63,33,253,127,63,47,253,127,63,62,253,127,63,76,253,127,63,89,253,127,63,103,253,127,63,116,253,127,63,130,253,127,63,143,253,127,63,155,253,127,63,168,253,127,63,181,253,127,63,193,253,127,63,205,253,127,63,217,253,127,63,228,253,127,63,240,253,127,63,251,253,127,63,6,254,127,63,17,254,127,63,28,254,127,63,38,254,127,63,49,254,127,63,59,254,127,63,69,254,127,63,79,254,127,63,89,254,127,63,98,254,127,63,108,254,127,63,117,254,127,63,126,254,127,63,135,254,127,63,144,254,127,63,152,254,127,63,161,254,127,63,169,254,127,63,177,254,127,63,185,254,127,63,193,254,127,63,201,254,127,63,208,254,127,63,216,254,127,63,223,254,127,63,230,254,127,63,237,254,127,63,244,254,127,63,251,254,127,63,2,255,127,63,8,255,127,63,14,255,127,63,21,255,127,63,27,255,127,63,33,255,127,63,39,255,127,63,45,255,127,63,50,255,127,63,56,255,127,63,61,255,127,63,67,255,127,63,72,255,127,63,77,255,127,63,82,255,127,63,87,255,127,63,92,255,127,63,96,255,127,63,101,255,127,63,105,255,127,63,110,255,127,63,114,255,127,63,118,255,127,63,122,255,127,63,126,255,127,63,130,255,127,63,134,255,127,63,138,255,127,63,142,255,127,63,145,255,127,63,149,255,127,63,152,255,127,63,155,255,127,63,159,255,127,63,162,255,127,63,165,255,127,63,168,255,127,63,171,255,127,63,174,255,127,63,176,255,127,63,179,255,127,63,182,255,127,63,184,255,127,63,187,255,127,63,189,255,127,63,192,255,127,63,194,255,127,63,196,255,127,63,198,255,127,63,201,255,127,63,203,255,127,63,205,255,127,63,207,255,127,63,209,255,127,63,210,255,127,63,212,255,127,63,214,255,127,63,216,255,127,63,217,255,127,63,219,255,127,63,220,255,127,63,222,255,127,63,223,255,127,63,225,255,127,63,226,255,127,63,227,255,127,63,229,255,127,63,230,255,127,63,231,255,127,63,232,255,127,63,233,255,127,63,234,255,127,63,235,255,127,63,236,255,127,63,237,255,127,63,238,255,127,63,239,255,127,63,240,255,127,63,241,255,127,63,241,255,127,63,242,255,127,63,243,255,127,63,244,255,127,63,244,255,127,63,245,255,127,63,246,255,127,63,246,255,127,63,247,255,127,63,247,255,127,63,248,255,127,63,248,255,127,63,249,255,127,63,249,255,127,63,250,255,127,63,250,255,127,63,250,255,127,63,251,255,127,63,251,255,127,63,251,255,127,63,252,255,127,63,252,255,127,63,252,255,127,63,253,255,127,63,253,255,127,63,253,255,127,63,253,255,127,63,254,255,127,63,254,255,127,63,254,255,127,63,254,255,127,63,254,255,127,63,254,255,127,63,255,255,127,63,255,255,127,63,255,255,127,63,255,255,127,63,255,255,127,63,255,255,127,63,255,255,127,63,255,255,127,63,255,255,127,63,255,255,127,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,62,180,228,51,9,145,243,51,139,178,1,52,60,32,10,52,35,26,19,52,96,169,28,52,167,215,38,52,75,175,49,52,80,59,61,52,112,135,73,52,35,160,86,52,184,146,100,52,85,109,115,52,136,159,129,52,252,11,138,52,147,4,147,52,105,146,156,52,50,191,166,52,63,149,177,52,147,31,189,52,228,105,201,52,173,128,214,52,54,113,228,52,166,73,243,52,136,140,1,53,192,247,9,53,6,239,18,53,118,123,28,53,192,166,38,53,55,123,49,53,218,3,61,53,94,76,73,53,59,97,86,53,185,79,100,53,252,37,115,53,138,121,129,53,134,227,137,53,124,217,146,53,133,100,156,53,82,142,166,53,51,97,177,53,37,232,188,53,220,46,201,53,206,65,214,53,65,46,228,53,87,2,243,53,143,102,1,54,79,207,9,54,245,195,18,54,152,77,28,54,232,117,38,54,50,71,49,54,116,204,60,54,94,17,73,54,101,34,86,54,206,12,100,54,184,222,114,54,151,83,129,54,28,187,137,54,114,174,146,54,175,54,156,54,129,93,166,54,53,45,177,54,199,176,188,54,228,243,200,54,1,3,214,54,96,235,227,54,30,187,242,54,162,64,1,55,235,166,9,55,241,152,18,55,201,31,28,55,30,69,38,55,61,19,49,55,30,149,60,55,111,214,72,55,162,227,85,55,247,201,99,55,137,151,114,55,175,45,129,55,190,146,137,55,116,131,146,55,230,8,156,55,190,44,166,55,71,249,176,55,121,121,188,55,254,184,200,55,71,196,213,55,146,168,227,55,248,115,242,55,192,26,1,56,147,126,9,56,249,109,18,56,6,242,27,56,98,20,38,56,86,223,48,56,216,93,60,56,146,155,72,56,242,164,85,56,51,135,99,56,110,80,114,56,211,7,129,56,107,106,137,56,130,88,146,56,42,219,155,56,9,252,165,56,104,197,176,56,59,66,188,56,41,126,200,56,160,133,213,56,217,101,227,56,232,44,242,56,233,244,0,57,70,86,9,57,14,67,18,57,81,196,27,57,181,227,37,57,127,171,48,57,162,38,60,57,197,96,72,57,83,102,85,57,131,68,99,57,104,9,114,57,1,226,128,57,36,66,137,57,157,45,146,57,123,173,155,57,99,203,165,57,153,145,176,57,13,11,188,57,102,67,200,57,11,71,213,57,50,35,227,57,237,229,241,57,29,207,0,58,5,46,9,58,48,24,18,58,169,150,27,58,21,179,37,58,183,119,48,58,124,239,59,58,10,38,72,58,199,39,85,58,230,1,99,58,120,194,113,58,59,188,128,58,233,25,137,58,198,2,146,58,219,127,155,58,203,154,165,58,216,93,176,58,239,211,187,58,179,8,200,58,136,8,213,58,159,224,226,58,7,159,241,58,92,169,0,59,208,5,9,59,94,237,17,59,15,105,27,59,132,130,37,59,253,67,48,59,103,184,59,59,97,235,71,59,77,233,84,59,93,191,98,59,156,123,113,59,127,150,128,59,186,241,136,59,249,215,145,59,71,82,155,59,65,106,165,59,39,42,176,59,226,156,187,59,18,206,199,59,23,202,212,59,32,158,226,59,53,88,241,59,166,131,0,60,167,221,8,60,152,194,17,60,130,59,27,60,1,82,37,60,84,16,48,60,97,129,59,60,200,176,71,60,229,170,84,60,232,124,98,60,212,52,113,60,207,112,128,60,150,201,136,60,58,173,145,60,192,36,155,60,197,57,165,60,133,246,175,60,229,101,187,60,130,147,199,60,185,139,212,60,180,91,226,60,121,17,241,60,251,93,0,61,137,181,8,61,223,151,17,61,2,14,27,61,141,33,37,61,185,220,47,61,109,74,59,61,64,118,71,61,145,108,84,61,133,58,98,61,34,238,112,61,42,75,128,61,127,161,136,61,136,130,145,61,72,247,154,61,88,9,165,61,242,194,175,61,248,46,187,61,3,89,199,61,109,77,212,61,92,25,226,61,209,202,240,61,91,56,0,62,119,141,8,62,51,109,17,62,144,224,26,62,39,241,36,62,46,169,47,62,135,19,59,62,202,59,71,62,77,46,84,62,55,248,97,62,132,167,112,62,143,37,128,62,115,121,136,62,226,87,145,62,220,201,154,62,249,216,164,62,109,143,175,62,27,248,186,62,149,30,199,62,51,15,212,62,23,215,225,62,61,132,240,62,198,18,0,63,114,101,8,63,147,66,17,63,43,179,26,63,206,192,36,63,177,117,47,63,178,220,58,63,101,1,71,63,29,240,83,63,251,181,97,63,251,96,112,63,0,0,128,63,156,139,1,0,188,139,1,0,220,139,1,0,0,0,0,0,8,11,0,0,9,11,0,0,10,11,0,0,11,11,0,0,0,0,0,0,0,0,0,0,12,11,0,0,13,11,0,0,8,11,0,0,9,11,0,0,10,11,0,0,11,11,0,0,14,11,0,0,15,11,0,0,16,11,0,0,13,11,0,0,8,11,0,0,9,11,0,0,10,11,0,0,11,11,0,0,17,11,0,0,18,11,0,0,19,11,0,0,164,7,1,0,192,7,1,0,220,7,1,0,4,0,0,0,2,0,0,0,3,0,0,0,5,0,0,0,0,0,76,194,0,0,80,194,0,0,84,194,0,0,88,194,0,0,92,194,0,0,96,194,0,0,100,194,0,0,104,194,0,0,108,194,0,0,112,194,0,0,116,194,0,0,120,194,0,0,124,194,0,0,128,194,0,0,130,194,0,0,132,194,0,0,134,194,0,0,136,194,0,0,138,194,0,0,140,194,0,0,142,194,0,0,144,194,0,0,146,194,0,0,148,194,0,0,150,194,0,0,152,194,0,0,154,194,0,0,156,194,0,0,160,194,0,0,162,194,0,0,164,194,0,0,166,194,0,0,168,194,0,0,170,194,0,0,172,194,0,0,174,194,0,0,176,194,0,0,176,194,0,0,178,194,0,0,178,194,0,0,180,194,0,0,182,194,0,0,182,194,0,0,184,194,0,0,186,194,0,0,188,194,0,0,190,194,0,0,192,194,0,0,192,194,0,0,194,194,0,0,196,194,0,0,196,194,0,0,198,194,0,0,198,194,0,0,200,194,0,0,200,194,0,0,202,194,0,0,204,194,0,0,206,194,0,0,208,194,0,0,212,194,0,0,214,194,0,0,214,194,0,0,214,194,0,0,214,194,0,0,210,194,0,0,206,194,0,0,204,194,0,0,202,194,0,0,198,194,0,0,196,194,0,0,192,194,0,0,190,194,0,0,190,194,0,0,192,194,0,0,194,194,0,0,192,194,0,0,190,194,0,0,186,194,0,0,180,194,0,0,160,194,0,0,140,194,0,0,72,194,0,0,32,194,0,0,240,193,0,0,240,193,0,0,240,193,0,0,240,193,0,0,112,194,0,0,112,194,0,0,112,194,0,0,112,194,0,0,112,194,0,0,112,194,0,0,112,194,0,0,112,194,0,0,112,194,0,0,112,194,0,0,112,194,0,0,112,194,0,0,120,194,0,0,120,194,0,0,130,194,0,0,146,194,0,0,138,194,0,0,136,194,0,0,136,194,0,0,134,194,0,0,140,194,0,0,140,194,0,0,144,194,0,0,148,194,0,0,150,194,0,0,158,194,0,0,158,194,0,0,160,194,0,0,166,194,0,0,176,194,0,0,186,194,0,0,200,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,64,194,0,0,64,194,0,0,64,194,0,0,64,194,0,0,64,194,0,0,64,194,0,0,64,194,0,0,64,194,0,0,64,194,0,0,64,194,0,0,64,194,0,0,64,194,0,0,64,194,0,0,84,194,0,0,116,194,0,0,132,194,0,0,132,194,0,0,136,194,0,0,134,194,0,0,140,194,0,0,152,194,0,0,152,194,0,0,144,194,0,0,146,194,0,0,150,194,0,0,152,194,0,0,156,194,0,0,158,194,0,0,166,194,0,0,176,194,0,0,186,194,0,0,200,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,20,194,0,0,20,194,0,0,20,194,0,0,20,194,0,0,20,194,0,0,20,194,0,0,20,194,0,0,20,194,0,0,24,194,0,0,32,194,0,0,40,194,0,0,56,194,0,0,64,194,0,0,84,194,0,0,92,194,0,0,120,194,0,0,130,194,0,0,104,194,0,0,96,194,0,0,96,194,0,0,116,194,0,0,112,194,0,0,130,194,0,0,134,194,0,0,138,194,0,0,142,194,0,0,154,194,0,0,154,194,0,0,156,194,0,0,160,194,0,0,164,194,0,0,168,194,0,0,176,194,0,0,186,194,0,0,196,194,0,0,212,194,0,0,224,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,200,193,0,0,200,193,0,0,200,193,0,0,200,193,0,0,200,193,0,0,200,193,0,0,200,193,0,0,200,193,0,0,200,193,0,0,208,193,0,0,216,193,0,0,232,193,0,0,0,194,0,0,24,194,0,0,64,194,0,0,80,194,0,0,80,194,0,0,72,194,0,0,64,194,0,0,64,194,0,0,76,194,0,0,80,194,0,0,88,194,0,0,112,194,0,0,134,194,0,0,134,194,0,0,132,194,0,0,136,194,0,0,138,194,0,0,146,194,0,0,146,194,0,0,152,194,0,0,160,194,0,0,162,194,0,0,162,194,0,0,170,194,0,0,170,194,0,0,172,194,0,0,176,194,0,0,186,194,0,0,200,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,128,193,0,0,128,193,0,0,128,193,0,0,128,193,0,0,128,193,0,0,128,193,0,0,128,193,0,0,128,193,0,0,136,193,0,0,152,193,0,0,160,193,0,0,176,193,0,0,208,193,0,0,224,193,0,0,248,193,0,0,32,194,0,0,60,194,0,0,28,194,0,0,28,194,0,0,32,194,0,0,40,194,0,0,44,194,0,0,60,194,0,0,76,194,0,0,100,194,0,0,80,194,0,0,92,194,0,0,92,194,0,0,112,194,0,0,104,194,0,0,120,194,0,0,124,194,0,0,140,194,0,0,134,194,0,0,138,194,0,0,144,194,0,0,146,194,0,0,154,194,0,0,160,194,0,0,164,194,0,0,166,194,0,0,174,194,0,0,180,194,0,0,188,194,0,0,196,194,0,0,208,194,0,0,230,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,0,193,0,0,0,193,0,0,0,193,0,0,0,193,0,0,0,193,0,0,0,193,0,0,0,193,0,0,0,193,0,0,0,193,0,0,0,193,0,0,32,193,0,0,48,193,0,0,112,193,0,0,152,193,0,0,200,193,0,0,240,193,0,0,8,194,0,0,248,193,0,0,240,193,0,0,248,193,0,0,232,193,0,0,0,194,0,0,12,194,0,0,40,194,0,0,64,194,0,0,40,194,0,0,48,194,0,0,56,194,0,0,72,194,0,0,72,194,0,0,76,194,0,0,80,194,0,0,108,194,0,0,88,194,0,0,92,194,0,0,92,194,0,0,104,194,0,0,120,194,0,0,124,194,0,0,132,194,0,0,144,194,0,0,146,194,0,0,152,194,0,0,150,194,0,0,156,194,0,0,160,194,0,0,160,194,0,0,162,194,0,0,168,194,0,0,176,194,0,0,180,194,0,0,188,194,0,0,196,194,0,0,202,194,0,0,212,194,0,0,220,194,0,0,132,194,0,0,132,194,0,0,132,194,0,0,132,194,0,0,132,194,0,0,132,194,0,0,132,194,0,0,132,194,0,0,132,194,0,0,132,194,0,0,132,194,0,0,132,194,0,0,132,194,0,0,134,194,0,0,134,194,0,0,134,194,0,0,152,194,0,0,144,194,0,0,142,194,0,0,148,194,0,0,152,194,0,0,152,194,0,0,150,194,0,0,156,194,0,0,158,194,0,0,158,194,0,0,162,194,0,0,166,194,0,0,172,194,0,0,178,194,0,0,186,194,0,0,194,194,0,0,200,194,0,0,210,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,60,194,0,0,60,194,0,0,60,194,0,0,60,194,0,0,60,194,0,0,60,194,0,0,60,194,0,0,60,194,0,0,60,194,0,0,60,194,0,0,60,194,0,0,64,194,0,0,76,194,0,0,92,194,0,0,108,194,0,0,132,194,0,0,132,194,0,0,132,194,0,0,134,194,0,0,132,194,0,0,136,194,0,0,138,194,0,0,140,194,0,0,148,194,0,0,158,194,0,0,154,194,0,0,154,194,0,0,156,194,0,0,160,194,0,0,162,194,0,0,164,194,0,0,168,194,0,0,172,194,0,0,176,194,0,0,182,194,0,0,190,194,0,0,200,194,0,0,216,194,0,0,232,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,16,194,0,0,16,194,0,0,16,194,0,0,16,194,0,0,16,194,0,0,16,194,0,0,16,194,0,0,16,194,0,0,16,194,0,0,20,194,0,0,20,194,0,0,36,194,0,0,48,194,0,0,64,194,0,0,76,194,0,0,104,194,0,0,120,194,0,0,112,194,0,0,100,194,0,0,108,194,0,0,108,194,0,0,112,194,0,0,124,194,0,0,130,194,0,0,144,194,0,0,142,194,0,0,140,194,0,0,144,194,0,0,148,194,0,0,154,194,0,0,152,194,0,0,156,194,0,0,162,194,0,0,162,194,0,0,160,194,0,0,166,194,0,0,172,194,0,0,182,194,0,0,192,194,0,0,200,194,0,0,210,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,224,193,0,0,224,193,0,0,224,193,0,0,224,193,0,0,224,193,0,0,224,193,0,0,224,193,0,0,224,193,0,0,224,193,0,0,240,193,0,0,0,194,0,0,0,194,0,0,4,194,0,0,12,194,0,0,36,194,0,0,68,194,0,0,72,194,0,0,68,194,0,0,60,194,0,0,64,194,0,0,64,194,0,0,80,194,0,0,76,194,0,0,100,194,0,0,130,194,0,0,116,194,0,0,108,194,0,0,116,194,0,0,128,194,0,0,138,194,0,0,140,194,0,0,148,194,0,0,154,194,0,0,154,194,0,0,156,194,0,0,162,194,0,0,168,194,0,0,170,194,0,0,174,194,0,0,180,194,0,0,184,194,0,0,192,194,0,0,200,194,0,0,214,194,0,0,224,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,152,193,0,0,152,193,0,0,152,193,0,0,152,193,0,0,152,193,0,0,152,193,0,0,152,193,0,0,152,193,0,0,160,193,0,0,168,193,0,0,184,193,0,0,216,193,0,0,240,193,0,0,12,194,0,0,16,194,0,0,36,194,0,0,56,194,0,0,48,194,0,0,40,194,0,0,32,194,0,0,36,194,0,0,36,194,0,0,44,194,0,0,64,194,0,0,92,194,0,0,84,194,0,0,80,194,0,0,84,194,0,0,96,194,0,0,108,194,0,0,104,194,0,0,112,194,0,0,134,194,0,0,132,194,0,0,138,194,0,0,142,194,0,0,144,194,0,0,150,194,0,0,158,194,0,0,162,194,0,0,168,194,0,0,174,194,0,0,180,194,0,0,186,194,0,0,194,194,0,0,202,194,0,0,214,194,0,0,228,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,16,193,0,0,16,193,0,0,16,193,0,0,16,193,0,0,16,193,0,0,16,193,0,0,16,193,0,0,16,193,0,0,48,193,0,0,64,193,0,0,64,193,0,0,112,193,0,0,128,193,0,0,160,193,0,0,184,193,0,0,240,193,0,0,20,194,0,0,8,194,0,0,4,194,0,0,8,194,0,0,248,193,0,0,0,194,0,0,0,194,0,0,24,194,0,0,60,194,0,0,48,194,0,0,36,194,0,0,32,194,0,0,60,194,0,0,68,194,0,0,56,194,0,0,56,194,0,0,104,194,0,0,72,194,0,0,72,194,0,0,88,194,0,0,104,194,0,0,120,194,0,0,128,194,0,0,134,194,0,0,134,194,0,0,140,194,0,0,144,194,0,0,152,194,0,0,158,194,0,0,166,194,0,0,174,194,0,0,182,194,0,0,192,194,0,0,200,194,0,0,208,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,120,194,0,0,120,194,0,0,120,194,0,0,120,194,0,0,120,194,0,0,120,194,0,0,120,194,0,0,120,194,0,0,120,194,0,0,120,194,0,0,124,194,0,0,128,194,0,0,132,194,0,0,134,194,0,0,132,194,0,0,136,194,0,0,150,194,0,0,144,194,0,0,152,194,0,0,150,194,0,0,152,194,0,0,156,194,0,0,158,194,0,0,164,194,0,0,168,194,0,0,170,194,0,0,180,194,0,0,188,194,0,0,202,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,108,194,0,0,108,194,0,0,108,194,0,0,108,194,0,0,108,194,0,0,108,194,0,0,108,194,0,0,108,194,0,0,108,194,0,0,108,194,0,0,108,194,0,0,112,194,0,0,112,194,0,0,116,194,0,0,124,194,0,0,132,194,0,0,142,194,0,0,136,194,0,0,140,194,0,0,140,194,0,0,142,194,0,0,144,194,0,0,144,194,0,0,150,194,0,0,162,194,0,0,156,194,0,0,158,194,0,0,164,194,0,0,166,194,0,0,172,194,0,0,180,194,0,0,194,194,0,0,206,194,0,0,226,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,84,194,0,0,84,194,0,0,84,194,0,0,84,194,0,0,84,194,0,0,84,194,0,0,84,194,0,0,84,194,0,0,84,194,0,0,88,194,0,0,92,194,0,0,100,194,0,0,96,194,0,0,100,194,0,0,92,194,0,0,116,194,0,0,130,194,0,0,112,194,0,0,112,194,0,0,120,194,0,0,124,194,0,0,124,194,0,0,132,194,0,0,136,194,0,0,148,194,0,0,146,194,0,0,150,194,0,0,150,194,0,0,156,194,0,0,160,194,0,0,160,194,0,0,164,194,0,0,170,194,0,0,180,194,0,0,192,194,0,0,202,194,0,0,216,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,56,194,0,0,56,194,0,0,56,194,0,0,56,194,0,0,56,194,0,0,56,194,0,0,56,194,0,0,56,194,0,0,56,194,0,0,56,194,0,0,60,194,0,0,60,194,0,0,60,194,0,0,60,194,0,0,64,194,0,0,76,194,0,0,100,194,0,0,76,194,0,0,68,194,0,0,72,194,0,0,76,194,0,0,84,194,0,0,88,194,0,0,108,194,0,0,132,194,0,0,112,194,0,0,120,194,0,0,134,194,0,0,134,194,0,0,140,194,0,0,144,194,0,0,150,194,0,0,152,194,0,0,156,194,0,0,162,194,0,0,170,194,0,0,176,194,0,0,188,194,0,0,194,194,0,0,208,194,0,0,224,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,16,194,0,0,16,194,0,0,16,194,0,0,16,194,0,0,16,194,0,0,16,194,0,0,16,194,0,0,16,194,0,0,28,194,0,0,36,194,0,0,40,194,0,0,40,194,0,0,28,194,0,0,24,194,0,0,36,194,0,0,44,194,0,0,80,194,0,0,48,194,0,0,32,194,0,0,28,194,0,0,20,194,0,0,20,194,0,0,32,194,0,0,60,194,0,0,88,194,0,0,72,194,0,0,64,194,0,0,72,194,0,0,92,194,0,0,116,194,0,0,108,194,0,0,120,194,0,0,132,194,0,0,132,194,0,0,132,194,0,0,138,194,0,0,138,194,0,0,146,194,0,0,148,194,0,0,148,194,0,0,150,194,0,0,154,194,0,0,158,194,0,0,164,194,0,0,174,194,0,0,182,194,0,0,190,194,0,0,200,194,0,0,216,194,0,0,230,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,224,193,0,0,208,193,0,0,192,193,0,0,176,193,0,0,160,193,0,0,160,193,0,0,184,193,0,0,232,193,0,0,240,193,0,0,248,193,0,0,224,193,0,0,216,193,0,0,224,193,0,0,224,193,0,0,224,193,0,0,12,194,0,0,32,194,0,0,4,194,0,0,0,194,0,0,232,193,0,0,240,193,0,0,240,193,0,0,240,193,0,0,20,194,0,0,52,194,0,0,36,194,0,0,20,194,0,0,24,194,0,0,52,194,0,0,60,194,0,0,60,194,0,0,64,194,0,0,84,194,0,0,68,194,0,0,64,194,0,0,72,194,0,0,68,194,0,0,68,194,0,0,76,194,0,0,80,194,0,0,104,194,0,0,96,194,0,0,100,194,0,0,96,194,0,0,112,194,0,0,116,194,0,0,120,194,0,0,140,194,0,0,144,194,0,0,148,194,0,0,156,194,0,0,166,194,0,0,176,194,0,0,186,194,0,0,200,194,0,0,212,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,210,194,0,0,200,194,0,0,190,194,0,0,182,194,0,0,174,194,0,0,166,194,0,0,160,194,0,0,156,194,0,0,152,194,0,0,156,194,0,0,156,194,0,0,162,194,0,0,166,194,0,0,170,194,0,0,172,194,0,0,170,194,0,0,172,194,0,0,174,194,0,0,180,194,0,0,194,194,0,0,214,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,210,194,0,0,200,194,0,0,190,194,0,0,180,194,0,0,170,194,0,0,162,194,0,0,154,194,0,0,146,194,0,0,140,194,0,0,134,194,0,0,134,194,0,0,136,194,0,0,150,194,0,0,146,194,0,0,140,194,0,0,138,194,0,0,140,194,0,0,144,194,0,0,150,194,0,0,158,194,0,0,168,194,0,0,166,194,0,0,168,194,0,0,172,194,0,0,176,194,0,0,178,194,0,0,178,194,0,0,186,194,0,0,196,194,0,0,210,194,0,0,224,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,210,194,0,0,200,194,0,0,190,194,0,0,180,194,0,0,170,194,0,0,160,194,0,0,152,194,0,0,142,194,0,0,136,194,0,0,136,194,0,0,130,194,0,0,124,194,0,0,124,194,0,0,120,194,0,0,120,194,0,0,128,194,0,0,130,194,0,0,128,194,0,0,116,194,0,0,120,194,0,0,124,194,0,0,128,194,0,0,132,194,0,0,136,194,0,0,146,194,0,0,146,194,0,0,148,194,0,0,150,194,0,0,152,194,0,0,162,194,0,0,166,194,0,0,170,194,0,0,176,194,0,0,178,194,0,0,184,194,0,0,190,194,0,0,200,194,0,0,216,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,160,194,0,0,150,194,0,0,142,194,0,0,136,194,0,0,130,194,0,0,124,194,0,0,120,194,0,0,116,194,0,0,116,194,0,0,116,194,0,0,116,194,0,0,108,194,0,0,96,194,0,0,100,194,0,0,84,194,0,0,72,194,0,0,104,194,0,0,80,194,0,0,72,194,0,0,72,194,0,0,80,194,0,0,84,194,0,0,88,194,0,0,104,194,0,0,134,194,0,0,124,194,0,0,134,194,0,0,136,194,0,0,144,194,0,0,150,194,0,0,156,194,0,0,160,194,0,0,162,194,0,0,162,194,0,0,164,194,0,0,170,194,0,0,178,194,0,0,180,194,0,0,186,194,0,0,194,194,0,0,202,194,0,0,214,194,0,0,228,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,130,194,0,0,116,194,0,0,108,194,0,0,100,194,0,0,96,194,0,0,92,194,0,0,92,194,0,0,96,194,0,0,96,194,0,0,100,194,0,0,92,194,0,0,84,194,0,0,80,194,0,0,60,194,0,0,48,194,0,0,48,194,0,0,72,194,0,0,48,194,0,0,36,194,0,0,28,194,0,0,28,194,0,0,40,194,0,0,32,194,0,0,56,194,0,0,76,194,0,0,68,194,0,0,72,194,0,0,84,194,0,0,88,194,0,0,124,194,0,0,112,194,0,0,116,194,0,0,120,194,0,0,132,194,0,0,132,194,0,0,132,194,0,0,140,194,0,0,146,194,0,0,148,194,0,0,150,194,0,0,152,194,0,0,150,194,0,0,158,194,0,0,170,194,0,0,178,194,0,0,182,194,0,0,192,194,0,0,204,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,80,194,0,0,72,194,0,0,68,194,0,0,68,194,0,0,64,194,0,0,64,194,0,0,64,194,0,0,68,194,0,0,72,194,0,0,72,194,0,0,68,194,0,0,56,194,0,0,44,194,0,0,28,194,0,0,12,194,0,0,4,194,0,0,24,194,0,0,16,194,0,0,0,194,0,0,232,193,0,0,0,194,0,0,0,194,0,0,0,194,0,0,12,194,0,0,48,194,0,0,28,194,0,0,24,194,0,0,24,194,0,0,56,194,0,0,72,194,0,0,52,194,0,0,56,194,0,0,84,194,0,0,72,194,0,0,72,194,0,0,72,194,0,0,88,194,0,0,88,194,0,0,84,194,0,0,84,194,0,0,96,194,0,0,100,194,0,0,108,194,0,0,132,194,0,0,140,194,0,0,144,194,0,0,148,194,0,0,158,194,0,0,166,194,0,0,170,194,0,0,180,194,0,0,194,194,0,0,228,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,210,194,0,0,200,194,0,0,190,194,0,0,180,194,0,0,172,194,0,0,160,194,0,0,150,194,0,0,150,194,0,0,158,194,0,0,160,194,0,0,158,194,0,0,160,194,0,0,162,194,0,0,164,194,0,0,176,194,0,0,190,194,0,0,206,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,216,194,0,0,206,194,0,0,196,194,0,0,186,194,0,0,176,194,0,0,166,194,0,0,158,194,0,0,156,194,0,0,150,194,0,0,142,194,0,0,134,194,0,0,136,194,0,0,146,194,0,0,146,194,0,0,144,194,0,0,146,194,0,0,150,194,0,0,154,194,0,0,160,194,0,0,164,194,0,0,176,194,0,0,186,194,0,0,200,194,0,0,214,194,0,0,228,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,210,194,0,0,202,194,0,0,192,194,0,0,180,194,0,0,172,194,0,0,162,194,0,0,154,194,0,0,146,194,0,0,138,194,0,0,132,194,0,0,116,194,0,0,120,194,0,0,132,194,0,0,128,194,0,0,120,194,0,0,130,194,0,0,132,194,0,0,140,194,0,0,144,194,0,0,152,194,0,0,162,194,0,0,160,194,0,0,168,194,0,0,180,194,0,0,190,194,0,0,204,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,214,194,0,0,206,194,0,0,194,194,0,0,184,194,0,0,176,194,0,0,166,194,0,0,158,194,0,0,148,194,0,0,140,194,0,0,132,194,0,0,108,194,0,0,84,194,0,0,104,194,0,0,120,194,0,0,92,194,0,0,88,194,0,0,88,194,0,0,88,194,0,0,104,194,0,0,116,194,0,0,120,194,0,0,144,194,0,0,140,194,0,0,144,194,0,0,150,194,0,0,156,194,0,0,160,194,0,0,162,194,0,0,160,194,0,0,166,194,0,0,166,194,0,0,176,194,0,0,186,194,0,0,200,194,0,0,214,194,0,0,230,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,210,194,0,0,200,194,0,0,190,194,0,0,180,194,0,0,170,194,0,0,160,194,0,0,150,194,0,0,140,194,0,0,132,194,0,0,120,194,0,0,96,194,0,0,64,194,0,0,48,194,0,0,64,194,0,0,56,194,0,0,56,194,0,0,44,194,0,0,56,194,0,0,64,194,0,0,64,194,0,0,76,194,0,0,104,194,0,0,104,194,0,0,108,194,0,0,112,194,0,0,120,194,0,0,120,194,0,0,116,194,0,0,116,194,0,0,130,194,0,0,128,194,0,0,130,194,0,0,136,194,0,0,140,194,0,0,148,194,0,0,150,194,0,0,156,194,0,0,162,194,0,0,172,194,0,0,190,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,210,194,0,0,200,194,0,0,190,194,0,0,180,194,0,0,170,194,0,0,160,194,0,0,150,194,0,0,140,194,0,0,130,194,0,0,116,194,0,0,92,194,0,0,68,194,0,0,28,194,0,0,4,194,0,0,32,194,0,0,12,194,0,0,0,194,0,0,24,194,0,0,32,194,0,0,4,194,0,0,12,194,0,0,20,194,0,0,56,194,0,0,36,194,0,0,52,194,0,0,48,194,0,0,56,194,0,0,40,194,0,0,52,194,0,0,56,194,0,0,80,194,0,0,72,194,0,0,72,194,0,0,72,194,0,0,88,194,0,0,88,194,0,0,92,194,0,0,100,194,0,0,120,194,0,0,128,194,0,0,132,194,0,0,136,194,0,0,140,194,0,0,152,194,0,0,162,194,0,0,180,194,0,0,200,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,210,194,0,0,196,194,0,0,180,194,0,0,170,194,0,0,164,194,0,0,166,194,0,0,160,194,0,0,156,194,0,0,168,194,0,0,158,194,0,0,160,194,0,0,166,194,0,0,174,194,0,0,178,194,0,0,182,194,0,0,186,194,0,0,198,194,0,0,212,194,0,0,234,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,210,194,0,0,196,194,0,0,180,194,0,0,170,194,0,0,160,194,0,0,150,194,0,0,140,194,0,0,136,194,0,0,148,194,0,0,144,194,0,0,148,194,0,0,154,194,0,0,160,194,0,0,164,194,0,0,170,194,0,0,174,194,0,0,184,194,0,0,178,194,0,0,182,194,0,0,190,194,0,0,200,194,0,0,212,194,0,0,224,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,210,194,0,0,196,194,0,0,180,194,0,0,166,194,0,0,150,194,0,0,142,194,0,0,124,194,0,0,128,194,0,0,134,194,0,0,120,194,0,0,128,194,0,0,134,194,0,0,140,194,0,0,146,194,0,0,154,194,0,0,162,194,0,0,168,194,0,0,166,194,0,0,170,194,0,0,178,194,0,0,180,194,0,0,186,194,0,0,196,194,0,0,208,194,0,0,218,194,0,0,228,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,206,194,0,0,192,194,0,0,176,194,0,0,162,194,0,0,150,194,0,0,136,194,0,0,104,194,0,0,88,194,0,0,96,194,0,0,88,194,0,0,96,194,0,0,96,194,0,0,104,194,0,0,112,194,0,0,124,194,0,0,132,194,0,0,148,194,0,0,138,194,0,0,144,194,0,0,144,194,0,0,150,194,0,0,148,194,0,0,154,194,0,0,162,194,0,0,162,194,0,0,164,194,0,0,168,194,0,0,174,194,0,0,186,194,0,0,192,194,0,0,198,194,0,0,208,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,216,194,0,0,204,194,0,0,192,194,0,0,182,194,0,0,170,194,0,0,160,194,0,0,148,194,0,0,136,194,0,0,112,194,0,0,76,194,0,0,56,194,0,0,64,194,0,0,56,194,0,0,44,194,0,0,52,194,0,0,60,194,0,0,60,194,0,0,68,194,0,0,64,194,0,0,96,194,0,0,84,194,0,0,92,194,0,0,104,194,0,0,100,194,0,0,124,194,0,0,104,194,0,0,112,194,0,0,132,194,0,0,128,194,0,0,134,194,0,0,140,194,0,0,140,194,0,0,148,194,0,0,154,194,0,0,168,194,0,0,172,194,0,0,178,194,0,0,182,194,0,0,186,194,0,0,188,194,0,0,202,194,0,0,218,194,0,0,236,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,216,194,0,0,206,194,0,0,196,194,0,0,186,194,0,0,176,194,0,0,166,194,0,0,156,194,0,0,146,194,0,0,136,194,0,0,112,194,0,0,84,194,0,0,48,194,0,0,12,194,0,0,24,194,0,0,24,194,0,0,8,194,0,0,8,194,0,0,16,194,0,0,32,194,0,0,36,194,0,0,48,194,0,0,76,194,0,0,52,194,0,0,56,194,0,0,60,194,0,0,56,194,0,0,88,194,0,0,72,194,0,0,68,194,0,0,72,194,0,0,72,194,0,0,72,194,0,0,76,194,0,0,88,194,0,0,100,194,0,0,104,194,0,0,112,194,0,0,132,194,0,0,132,194,0,0,132,194,0,0,128,194,0,0,130,194,0,0,136,194,0,0,154,194,0,0,164,194,0,0,174,194,0,0,190,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,214,194,0,0,204,194,0,0,194,194,0,0,184,194,0,0,174,194,0,0,166,194,0,0,156,194,0,0,150,194,0,0,164,194,0,0,158,194,0,0,166,194,0,0,170,194,0,0,178,194,0,0,184,194,0,0,190,194,0,0,196,194,0,0,202,194,0,0,210,194,0,0,218,194,0,0,226,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,212,194,0,0,200,194,0,0,190,194,0,0,180,194,0,0,172,194,0,0,162,194,0,0,156,194,0,0,148,194,0,0,138,194,0,0,148,194,0,0,148,194,0,0,152,194,0,0,158,194,0,0,166,194,0,0,168,194,0,0,172,194,0,0,178,194,0,0,184,194,0,0,194,194,0,0,186,194,0,0,200,194,0,0,206,194,0,0,214,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,212,194,0,0,200,194,0,0,190,194,0,0,180,194,0,0,174,194,0,0,166,194,0,0,160,194,0,0,150,194,0,0,138,194,0,0,112,194,0,0,132,194,0,0,132,194,0,0,136,194,0,0,140,194,0,0,148,194,0,0,156,194,0,0,158,194,0,0,162,194,0,0,162,194,0,0,166,194,0,0,168,194,0,0,174,194,0,0,186,194,0,0,192,194,0,0,198,194,0,0,206,194,0,0,214,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,216,194,0,0,206,194,0,0,196,194,0,0,186,194,0,0,178,194,0,0,170,194,0,0,164,194,0,0,156,194,0,0,142,194,0,0,120,194,0,0,92,194,0,0,104,194,0,0,104,194,0,0,88,194,0,0,88,194,0,0,92,194,0,0,108,194,0,0,116,194,0,0,120,194,0,0,140,194,0,0,132,194,0,0,132,194,0,0,134,194,0,0,140,194,0,0,144,194,0,0,150,194,0,0,156,194,0,0,168,194,0,0,168,194,0,0,168,194,0,0,176,194,0,0,182,194,0,0,180,194,0,0,190,194,0,0,196,194,0,0,204,194,0,0,206,194,0,0,212,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,216,194,0,0,206,194,0,0,196,194,0,0,188,194,0,0,180,194,0,0,174,194,0,0,164,194,0,0,158,194,0,0,146,194,0,0,134,194,0,0,104,194,0,0,60,194,0,0,72,194,0,0,52,194,0,0,36,194,0,0,52,194,0,0,64,194,0,0,48,194,0,0,48,194,0,0,68,194,0,0,88,194,0,0,76,194,0,0,64,194,0,0,60,194,0,0,68,194,0,0,72,194,0,0,76,194,0,0,100,194,0,0,104,194,0,0,112,194,0,0,124,194,0,0,138,194,0,0,140,194,0,0,138,194,0,0,142,194,0,0,148,194,0,0,156,194,0,0,164,194,0,0,180,194,0,0,190,194,0,0,202,194,0,0,210,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,210,194,0,0,202,194,0,0,194,194,0,0,186,194,0,0,180,194,0,0,170,194,0,0,160,194,0,0,154,194,0,0,144,194,0,0,130,194,0,0,96,194,0,0,64,194,0,0,20,194,0,0,32,194,0,0,16,194,0,0,8,194,0,0,32,194,0,0,72,194,0,0,60,194,0,0,24,194,0,0,36,194,0,0,60,194,0,0,24,194,0,0,12,194,0,0,28,194,0,0,24,194,0,0,44,194,0,0,32,194,0,0,52,194,0,0,72,194,0,0,52,194,0,0,48,194,0,0,60,194,0,0,72,194,0,0,92,194,0,0,64,194,0,0,64,194,0,0,80,194,0,0,132,194,0,0,140,194,0,0,152,194,0,0,164,194,0,0,180,194,0,0,194,194,0,0,210,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,216,194,0,0,206,194,0,0,196,194,0,0,186,194,0,0,172,194,0,0,158,194,0,0,152,194,0,0,166,194,0,0,162,194,0,0,170,194,0,0,174,194,0,0,178,194,0,0,186,194,0,0,196,194,0,0,204,194,0,0,214,194,0,0,224,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,216,194,0,0,206,194,0,0,196,194,0,0,186,194,0,0,172,194,0,0,158,194,0,0,142,194,0,0,154,194,0,0,148,194,0,0,154,194,0,0,158,194,0,0,162,194,0,0,168,194,0,0,170,194,0,0,180,194,0,0,184,194,0,0,186,194,0,0,184,194,0,0,196,194,0,0,202,194,0,0,216,194,0,0,224,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,216,194,0,0,206,194,0,0,196,194,0,0,186,194,0,0,174,194,0,0,156,194,0,0,136,194,0,0,130,194,0,0,132,194,0,0,120,194,0,0,130,194,0,0,134,194,0,0,140,194,0,0,146,194,0,0,150,194,0,0,156,194,0,0,164,194,0,0,164,194,0,0,166,194,0,0,168,194,0,0,182,194,0,0,186,194,0,0,196,194,0,0,204,194,0,0,212,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,210,194,0,0,200,194,0,0,190,194,0,0,180,194,0,0,164,194,0,0,148,194,0,0,120,194,0,0,100,194,0,0,104,194,0,0,96,194,0,0,76,194,0,0,80,194,0,0,80,194,0,0,88,194,0,0,88,194,0,0,104,194,0,0,132,194,0,0,108,194,0,0,112,194,0,0,124,194,0,0,132,194,0,0,138,194,0,0,146,194,0,0,158,194,0,0,166,194,0,0,168,194,0,0,160,194,0,0,162,194,0,0,162,194,0,0,164,194,0,0,176,194,0,0,184,194,0,0,196,194,0,0,210,194,0,0,226,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,214,194,0,0,204,194,0,0,194,194,0,0,184,194,0,0,168,194,0,0,158,194,0,0,138,194,0,0,100,194,0,0,60,194,0,0,80,194,0,0,60,194,0,0,48,194,0,0,52,194,0,0,72,194,0,0,80,194,0,0,40,194,0,0,40,194,0,0,84,194,0,0,44,194,0,0,44,194,0,0,64,194,0,0,76,194,0,0,96,194,0,0,92,194,0,0,80,194,0,0,100,194,0,0,108,194,0,0,116,194,0,0,120,194,0,0,134,194,0,0,142,194,0,0,156,194,0,0,166,194,0,0,172,194,0,0,188,194,0,0,196,194,0,0,206,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,210,194,0,0,200,194,0,0,190,194,0,0,180,194,0,0,168,194,0,0,156,194,0,0,140,194,0,0,116,194,0,0,76,194,0,0,36,194,0,0,32,194,0,0,24,194,0,0,32,194,0,0,56,194,0,0,80,194,0,0,76,194,0,0,36,194,0,0,32,194,0,0,56,194,0,0,32,194,0,0,24,194,0,0,24,194,0,0,36,194,0,0,56,194,0,0,36,194,0,0,56,194,0,0,60,194,0,0,44,194,0,0,44,194,0,0,52,194,0,0,36,194,0,0,52,194,0,0,96,194,0,0,134,194,0,0,136,194,0,0,166,194,0,0,174,194,0,0,180,194,0,0,190,194,0,0,204,194,0,0,214,194,0,0,226,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,218,194,0,0,210,194,0,0,202,194,0,0,192,194,0,0,182,194,0,0,168,194,0,0,154,194,0,0,164,194,0,0,164,194,0,0,170,194,0,0,178,194,0,0,188,194,0,0,200,194,0,0,212,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,212,194,0,0,206,194,0,0,196,194,0,0,184,194,0,0,170,194,0,0,160,194,0,0,142,194,0,0,150,194,0,0,144,194,0,0,152,194,0,0,160,194,0,0,168,194,0,0,172,194,0,0,178,194,0,0,186,194,0,0,200,194,0,0,214,194,0,0,226,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,214,194,0,0,208,194,0,0,202,194,0,0,194,194,0,0,184,194,0,0,176,194,0,0,168,194,0,0,160,194,0,0,128,194,0,0,132,194,0,0,124,194,0,0,128,194,0,0,132,194,0,0,138,194,0,0,146,194,0,0,154,194,0,0,166,194,0,0,166,194,0,0,172,194,0,0,182,194,0,0,196,194,0,0,208,194,0,0,222,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,214,194,0,0,208,194,0,0,202,194,0,0,194,194,0,0,184,194,0,0,180,194,0,0,168,194,0,0,148,194,0,0,100,194,0,0,104,194,0,0,80,194,0,0,92,194,0,0,88,194,0,0,72,194,0,0,80,194,0,0,72,194,0,0,80,194,0,0,124,194,0,0,120,194,0,0,138,194,0,0,152,194,0,0,154,194,0,0,156,194,0,0,156,194,0,0,158,194,0,0,164,194,0,0,176,194,0,0,188,194,0,0,200,194,0,0,212,194,0,0,222,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,212,194,0,0,204,194,0,0,196,194,0,0,190,194,0,0,180,194,0,0,170,194,0,0,166,194,0,0,156,194,0,0,140,194,0,0,72,194,0,0,72,194,0,0,36,194,0,0,48,194,0,0,68,194,0,0,60,194,0,0,72,194,0,0,72,194,0,0,48,194,0,0,92,194,0,0,56,194,0,0,60,194,0,0,64,194,0,0,64,194,0,0,88,194,0,0,68,194,0,0,68,194,0,0,104,194,0,0,120,194,0,0,142,194,0,0,162,194,0,0,174,194,0,0,184,194,0,0,194,194,0,0,204,194,0,0,216,194,0,0,228,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,212,194,0,0,204,194,0,0,196,194,0,0,190,194,0,0,180,194,0,0,170,194,0,0,166,194,0,0,156,194,0,0,140,194,0,0,52,194,0,0,44,194,0,0,36,194,0,0,60,194,0,0,72,194,0,0,76,194,0,0,72,194,0,0,68,194,0,0,52,194,0,0,60,194,0,0,36,194,0,0,48,194,0,0,36,194,0,0,28,194,0,0,44,194,0,0,24,194,0,0,20,194,0,0,32,194,0,0,36,194,0,0,48,194,0,0,72,194,0,0,104,194,0,0,130,194,0,0,146,194,0,0,158,194,0,0,170,194,0,0,184,194,0,0,194,194,0,0,202,194,0,0,210,194,0,0,218,194,0,0,226,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,214,194,0,0,200,194,0,0,190,194,0,0,174,194,0,0,162,194,0,0,170,194,0,0,166,194,0,0,176,194,0,0,186,194,0,0,200,194,0,0,214,194,0,0,228,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,214,194,0,0,202,194,0,0,190,194,0,0,176,194,0,0,166,194,0,0,152,194,0,0,146,194,0,0,144,194,0,0,158,194,0,0,168,194,0,0,180,194,0,0,190,194,0,0,200,194,0,0,210,194,0,0,220,194,0,0,230,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,208,194,0,0,196,194,0,0,184,194,0,0,174,194,0,0,162,194,0,0,140,194,0,0,130,194,0,0,120,194,0,0,134,194,0,0,142,194,0,0,148,194,0,0,160,194,0,0,170,194,0,0,182,194,0,0,190,194,0,0,198,194,0,0,206,194,0,0,216,194,0,0,222,194,0,0,228,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,206,194,0,0,194,194,0,0,180,194,0,0,170,194,0,0,152,194,0,0,112,194,0,0,96,194,0,0,88,194,0,0,112,194,0,0,120,194,0,0,116,194,0,0,96,194,0,0,124,194,0,0,130,194,0,0,146,194,0,0,148,194,0,0,154,194,0,0,150,194,0,0,156,194,0,0,162,194,0,0,172,194,0,0,174,194,0,0,176,194,0,0,182,194,0,0,188,194,0,0,196,194,0,0,206,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,210,194,0,0,200,194,0,0,194,194,0,0,184,194,0,0,172,194,0,0,162,194,0,0,158,194,0,0,140,194,0,0,100,194,0,0,76,194,0,0,60,194,0,0,76,194,0,0,104,194,0,0,112,194,0,0,96,194,0,0,84,194,0,0,72,194,0,0,104,194,0,0,80,194,0,0,72,194,0,0,72,194,0,0,84,194,0,0,92,194,0,0,128,194,0,0,138,194,0,0,142,194,0,0,170,194,0,0,164,194,0,0,156,194,0,0,162,194,0,0,170,194,0,0,190,194,0,0,204,194,0,0,224,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,210,194,0,0,200,194,0,0,194,194,0,0,184,194,0,0,170,194,0,0,166,194,0,0,158,194,0,0,144,194,0,0,68,194,0,0,32,194,0,0,44,194,0,0,44,194,0,0,88,194,0,0,96,194,0,0,76,194,0,0,72,194,0,0,32,194,0,0,44,194,0,0,24,194,0,0,16,194,0,0,12,194,0,0,20,194,0,0,24,194,0,0,20,194,0,0,48,194,0,0,88,194,0,0,112,194,0,0,100,194,0,0,112,194,0,0,140,194,0,0,150,194,0,0,168,194,0,0,184,194,0,0,206,194,0,0,224,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,204,194,0,0,190,194,0,0,178,194,0,0,164,194,0,0,166,194,0,0,168,194,0,0,180,194,0,0,184,194,0,0,198,194,0,0,214,194,0,0,226,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,214,194,0,0,202,194,0,0,190,194,0,0,178,194,0,0,166,194,0,0,144,194,0,0,148,194,0,0,156,194,0,0,170,194,0,0,176,194,0,0,176,194,0,0,180,194,0,0,184,194,0,0,196,194,0,0,210,194,0,0,222,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,218,194,0,0,206,194,0,0,194,194,0,0,186,194,0,0,174,194,0,0,162,194,0,0,140,194,0,0,140,194,0,0,134,194,0,0,150,194,0,0,146,194,0,0,152,194,0,0,158,194,0,0,162,194,0,0,166,194,0,0,176,194,0,0,178,194,0,0,194,194,0,0,206,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,214,194,0,0,200,194,0,0,188,194,0,0,176,194,0,0,166,194,0,0,150,194,0,0,124,194,0,0,108,194,0,0,108,194,0,0,124,194,0,0,132,194,0,0,112,194,0,0,120,194,0,0,134,194,0,0,134,194,0,0,154,194,0,0,152,194,0,0,162,194,0,0,176,194,0,0,172,194,0,0,184,194,0,0,192,194,0,0,204,194,0,0,218,194,0,0,232,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,210,194,0,0,196,194,0,0,184,194,0,0,172,194,0,0,162,194,0,0,146,194,0,0,96,194,0,0,80,194,0,0,60,194,0,0,92,194,0,0,112,194,0,0,104,194,0,0,80,194,0,0,76,194,0,0,52,194,0,0,68,194,0,0,72,194,0,0,84,194,0,0,88,194,0,0,116,194,0,0,142,194,0,0,140,194,0,0,138,194,0,0,156,194,0,0,158,194,0,0,174,194,0,0,180,194,0,0,192,194,0,0,208,194,0,0,224,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,206,194,0,0,192,194,0,0,180,194,0,0,172,194,0,0,156,194,0,0,140,194,0,0,76,194,0,0,40,194,0,0,60,194,0,0,64,194,0,0,92,194,0,0,88,194,0,0,88,194,0,0,84,194,0,0,40,194,0,0,12,194,0,0,224,193,0,0,4,194,0,0,24,194,0,0,20,194,0,0,48,194,0,0,60,194,0,0,68,194,0,0,88,194,0,0,124,194,0,0,136,194,0,0,156,194,0,0,164,194,0,0,178,194,0,0,188,194,0,0,198,194,0,0,208,194,0,0,218,194,0,0,228,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,200,194,0,0,180,194,0,0,158,194,0,0,170,194,0,0,162,194,0,0,164,194,0,0,164,194,0,0,178,194,0,0,188,194,0,0,198,194,0,0,206,194,0,0,218,194,0,0,230,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,210,194,0,0,194,194,0,0,170,194,0,0,144,194,0,0,148,194,0,0,140,194,0,0,140,194,0,0,140,194,0,0,152,194,0,0,170,194,0,0,182,194,0,0,186,194,0,0,194,194,0,0,206,194,0,0,218,194,0,0,230,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,224,194,0,0,186,194,0,0,162,194,0,0,136,194,0,0,120,194,0,0,112,194,0,0,112,194,0,0,100,194,0,0,124,194,0,0,140,194,0,0,154,194,0,0,164,194,0,0,180,194,0,0,186,194,0,0,196,194,0,0,208,194,0,0,218,194,0,0,226,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,226,194,0,0,200,194,0,0,186,194,0,0,168,194,0,0,124,194,0,0,104,194,0,0,64,194,0,0,84,194,0,0,88,194,0,0,80,194,0,0,80,194,0,0,100,194,0,0,128,194,0,0,132,194,0,0,152,194,0,0,166,194,0,0,162,194,0,0,170,194,0,0,170,194,0,0,180,194,0,0,190,194,0,0,196,194,0,0,202,194,0,0,206,194,0,0,212,194,0,0,216,194,0,0,222,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,210,194,0,0,190,194,0,0,172,194,0,0,148,194,0,0,84,194,0,0,72,194,0,0,24,194,0,0,44,194,0,0,68,194,0,0,44,194,0,0,40,194,0,0,28,194,0,0,28,194,0,0,56,194,0,0,80,194,0,0,100,194,0,0,96,194,0,0,144,194,0,0,138,194,0,0,148,194,0,0,162,194,0,0,174,194,0,0,184,194,0,0,188,194,0,0,194,194,0,0,198,194,0,0,204,194,0,0,210,194,0,0,216,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,216,194,0,0,198,194,0,0,180,194,0,0,152,194,0,0,132,194,0,0,52,194,0,0,44,194,0,0,36,194,0,0,48,194,0,0,60,194,0,0,44,194,0,0,60,194,0,0,32,194,0,0,240,193,0,0,248,193,0,0,248,193,0,0,28,194,0,0,4,194,0,0,32,194,0,0,36,194,0,0,44,194,0,0,84,194,0,0,108,194,0,0,140,194,0,0,146,194,0,0,154,194,0,0,158,194,0,0,164,194,0,0,168,194,0,0,174,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,182,194,0,0,152,194,0,0,150,194,0,0,170,194,0,0,186,194,0,0,196,194,0,0,208,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,182,194,0,0,140,194,0,0,140,194,0,0,150,194,0,0,172,194,0,0,178,194,0,0,188,194,0,0,196,194,0,0,202,194,0,0,212,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,190,194,0,0,160,194,0,0,112,194,0,0,130,194,0,0,128,194,0,0,148,194,0,0,166,194,0,0,176,194,0,0,182,194,0,0,190,194,0,0,198,194,0,0,206,194,0,0,214,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,190,194,0,0,160,194,0,0,104,194,0,0,92,194,0,0,68,194,0,0,132,194,0,0,136,194,0,0,142,194,0,0,156,194,0,0,156,194,0,0,160,194,0,0,176,194,0,0,170,194,0,0,178,194,0,0,194,194,0,0,200,194,0,0,210,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,190,194,0,0,160,194,0,0,84,194,0,0,80,194,0,0,36,194,0,0,108,194,0,0,108,194,0,0,68,194,0,0,104,194,0,0,96,194,0,0,124,194,0,0,172,194,0,0,158,194,0,0,180,194,0,0,186,194,0,0,196,194,0,0,206,194,0,0,214,194,0,0,224,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,194,194,0,0,182,194,0,0,146,194,0,0,52,194,0,0,32,194,0,0,4,194,0,0,84,194,0,0,116,194,0,0,68,194,0,0,88,194,0,0,72,194,0,0,72,194,0,0,112,194,0,0,80,194,0,0,134,194,0,0,148,194,0,0,162,194,0,0,184,194,0,0,192,194,0,0,200,194,0,0,210,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,226,194,0,0,212,194,0,0,198,194,0,0,184,194,0,0,154,194,0,0,160,194,0,0,176,194,0,0,194,194,0,0,212,194,0,0,230,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,232,194,0,0,218,194,0,0,204,194,0,0,190,194,0,0,178,194,0,0,148,194,0,0,144,194,0,0,176,194,0,0,174,194,0,0,190,194,0,0,204,194,0,0,218,194,0,0,232,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,232,194,0,0,218,194,0,0,204,194,0,0,190,194,0,0,178,194,0,0,150,194,0,0,132,194,0,0,148,194,0,0,154,194,0,0,156,194,0,0,172,194,0,0,174,194,0,0,180,194,0,0,192,194,0,0,210,194,0,0,230,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,230,194,0,0,216,194,0,0,202,194,0,0,188,194,0,0,176,194,0,0,132,194,0,0,96,194,0,0,116,194,0,0,140,194,0,0,130,194,0,0,156,194,0,0,144,194,0,0,166,194,0,0,168,194,0,0,186,194,0,0,196,194,0,0,210,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,210,194,0,0,190,194,0,0,178,194,0,0,164,194,0,0,100,194,0,0,80,194,0,0,80,194,0,0,108,194,0,0,96,194,0,0,108,194,0,0,104,194,0,0,138,194,0,0,134,194,0,0,176,194,0,0,164,194,0,0,164,194,0,0,178,194,0,0,188,194,0,0,200,194,0,0,216,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,202,194,0,0,192,194,0,0,180,194,0,0,166,194,0,0,154,194,0,0,88,194,0,0,44,194,0,0,24,194,0,0,72,194,0,0,64,194,0,0,80,194,0,0,64,194,0,0,40,194,0,0,40,194,0,0,76,194,0,0,80,194,0,0,84,194,0,0,108,194,0,0,130,194,0,0,142,194,0,0,156,194,0,0,170,194,0,0,190,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,240,194,0,0,210,194,0,0,172,194,0,0,136,194,0,0,156,194,0,0,158,194,0,0,180,194,0,0,200,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,240,194,0,0,210,194,0,0,172,194,0,0,132,194,0,0,146,194,0,0,154,194,0,0,176,194,0,0,192,194,0,0,210,194,0,0,230,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,240,194,0,0,210,194,0,0,184,194,0,0,160,194,0,0,116,194,0,0,128,194,0,0,136,194,0,0,160,194,0,0,174,194,0,0,184,194,0,0,200,194,0,0,220,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,240,194,0,0,208,194,0,0,182,194,0,0,158,194,0,0,80,194,0,0,112,194,0,0,88,194,0,0,128,194,0,0,138,194,0,0,154,194,0,0,160,194,0,0,164,194,0,0,168,194,0,0,170,194,0,0,174,194,0,0,176,194,0,0,180,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,236,194,0,0,200,194,0,0,174,194,0,0,154,194,0,0,68,194,0,0,72,194,0,0,48,194,0,0,104,194,0,0,116,194,0,0,116,194,0,0,134,194,0,0,130,194,0,0,120,194,0,0,120,194,0,0,120,194,0,0,130,194,0,0,136,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,230,194,0,0,196,194,0,0,168,194,0,0,120,194,0,0,68,194,0,0,48,194,0,0,24,194,0,0,56,194,0,0,68,194,0,0,68,194,0,0,56,194,0,0,28,194,0,0,20,194,0,0,28,194,0,0,32,194,0,0,40,194,0,0,44,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,176,194,0,0,148,194,0,0,154,194,0,0,164,194,0,0,164,194,0,0,170,194,0,0,180,194,0,0,188,194,0,0,198,194,0,0,208,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,176,194,0,0,132,194,0,0,140,194,0,0,162,194,0,0,160,194,0,0,162,194,0,0,168,194,0,0,176,194,0,0,182,194,0,0,186,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,176,194,0,0,116,194,0,0,124,194,0,0,140,194,0,0,142,194,0,0,148,194,0,0,154,194,0,0,160,194,0,0,166,194,0,0,170,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,172,194,0,0,120,194,0,0,124,194,0,0,120,194,0,0,120,194,0,0,104,194,0,0,80,194,0,0,72,194,0,0,72,194,0,0,80,194,0,0,88,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,236,194,0,0,216,194,0,0,168,194,0,0,84,194,0,0,72,194,0,0,72,194,0,0,72,194,0,0,92,194,0,0,60,194,0,0,52,194,0,0,32,194,0,0,32,194,0,0,32,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,236,194,0,0,200,194,0,0,146,194,0,0,44,194,0,0,20,194,0,0,40,194,0,0,44,194,0,0,84,194,0,0,24,194,0,0,20,194,0,0,12,194,0,0,12,194,0,0,24,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,200,194,0,0,182,194,0,0,168,194,0,0,148,194,0,0,160,194,0,0,160,194,0,0,160,194,0,0,160,194,0,0,160,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,200,194,0,0,182,194,0,0,168,194,0,0,148,194,0,0,136,194,0,0,136,194,0,0,136,194,0,0,136,194,0,0,136,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,200,194,0,0,172,194,0,0,156,194,0,0,140,194,0,0,112,194,0,0,52,194,0,0,240,193,0,0,168,193,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,200,194,0,0,174,194,0,0,156,194,0,0,134,194,0,0,64,194,0,0,24,194,0,0,232,193,0,0,168,193,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,200,194,0,0,172,194,0,0,138,194,0,0,96,194,0,0,52,194,0,0,12,194,0,0,4,194,0,0,232,193,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,220,194,0,0,200,194,0,0,166,194,0,0,142,194,0,0,64,194,0,0,216,193,0,0,24,194,0,0,20,194,0,0,8,194,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,192,121,196,0,0,0,0,1,0,0,0,3,0,0,0,7,0,0,0,15,0,0,0,31,0,0,0,63,0,0,0,127,0,0,0,255,0,0,0,255,1,0,0,255,3,0,0,255,7,0,0,255,15,0,0,255,31,0,0,255,63,0,0,255,127,0,0,255,255,0,0,255,255,1,0,255,255,3,0,255,255,7,0,255,255,15,0,255,255,31,0,255,255,63,0,255,255,127,0,255,255,255,0,255,255,255,1,255,255,255,3,255,255,255,7,255,255,255,15,255,255,255,31,255,255,255,63,255,255,255,127,255,255,255,255,0,0,0,0,183,29,193,4,110,59,130,9,217,38,67,13,220,118,4,19,107,107,197,23,178,77,134,26,5,80,71,30,184,237,8,38,15,240,201,34,214,214,138,47,97,203,75,43,100,155,12,53,211,134,205,49,10,160,142,60,189,189,79,56,112,219,17,76,199,198,208,72,30,224,147,69,169,253,82,65,172,173,21,95,27,176,212,91,194,150,151,86,117,139,86,82,200,54,25,106,127,43,216,110,166,13,155,99,17,16,90,103,20,64,29,121,163,93,220,125,122,123,159,112,205,102,94,116,224,182,35,152,87,171,226,156,142,141,161,145,57,144,96,149,60,192,39,139,139,221,230,143,82,251,165,130,229,230,100,134,88,91,43,190,239,70,234,186,54,96,169,183,129,125,104,179,132,45,47,173,51,48,238,169,234,22,173,164,93,11,108,160,144,109,50,212,39,112,243,208,254,86,176,221,73,75,113,217,76,27,54,199,251,6,247,195,34,32,180,206,149,61,117,202,40,128,58,242,159,157,251,246,70,187,184,251,241,166,121,255,244,246,62,225,67,235,255,229,154,205,188,232,45,208,125,236,119,112,134,52,192,109,71,48,25,75,4,61,174,86,197,57,171,6,130,39,28,27,67,35,197,61,0,46,114,32,193,42,207,157,142,18,120,128,79,22,161,166,12,27,22,187,205,31,19,235,138,1,164,246,75,5,125,208,8,8,202,205,201,12,7,171,151,120,176,182,86,124,105,144,21,113,222,141,212,117,219,221,147,107,108,192,82,111,181,230,17,98,2,251,208,102,191,70,159,94,8,91,94,90,209,125,29,87,102,96,220,83,99,48,155,77,212,45,90,73,13,11,25,68,186,22,216,64,151,198,165,172,32,219,100,168,249,253,39,165,78,224,230,161,75,176,161,191,252,173,96,187,37,139,35,182,146,150,226,178,47,43,173,138,152,54,108,142,65,16,47,131,246,13,238,135,243,93,169,153,68,64,104,157,157,102,43,144,42,123,234,148,231,29,180,224,80,0,117,228,137,38,54,233,62,59,247,237,59,107,176,243,140,118,113,247,85,80,50,250,226,77,243,254,95,240,188,198,232,237,125,194,49,203,62,207,134,214,255,203,131,134,184,213,52,155,121,209,237,189,58,220,90,160,251,216,238,224,12,105,89,253,205,109,128,219,142,96,55,198,79,100,50,150,8,122,133,139,201,126,92,173,138,115,235,176,75,119,86,13,4,79,225,16,197,75,56,54,134,70,143,43,71,66,138,123,0,92,61,102,193,88,228,64,130,85,83,93,67,81,158,59,29,37,41,38,220,33,240,0,159,44,71,29,94,40,66,77,25,54,245,80,216,50,44,118,155,63,155,107,90,59,38,214,21,3,145,203,212,7,72,237,151,10,255,240,86,14,250,160,17,16,77,189,208,20,148,155,147,25,35,134,82,29,14,86,47,241,185,75,238,245,96,109,173,248,215,112,108,252,210,32,43,226,101,61,234,230,188,27,169,235,11,6,104,239,182,187,39,215,1,166,230,211,216,128,165,222,111,157,100,218,106,205,35,196,221,208,226,192,4,246,161,205,179,235,96,201,126,141,62,189,201,144,255,185,16,182,188,180,167,171,125,176,162,251,58,174,21,230,251,170,204,192,184,167,123,221,121,163,198,96,54,155,113,125,247,159,168,91,180,146,31,70,117,150,26,22,50,136,173,11,243,140,116,45,176,129,195,48,113,133,153,144,138,93,46,141,75,89,247,171,8,84,64,182,201,80,69,230,142,78,242,251,79,74,43,221,12,71,156,192,205,67,33,125,130,123,150,96,67,127,79,70,0,114,248,91,193,118,253,11,134,104,74,22,71,108,147,48,4,97,36,45,197,101,233,75,155,17,94,86,90,21,135,112,25,24,48,109,216,28,53,61,159,2,130,32,94,6,91,6,29,11,236,27,220,15,81,166,147,55,230,187,82,51,63,157,17,62,136,128,208,58,141,208,151,36,58,205,86,32,227,235,21,45,84,246,212,41,121,38,169,197,206,59,104,193,23,29,43,204,160,0,234,200,165,80,173,214,18,77,108,210,203,107,47,223,124,118,238,219,193,203,161,227,118,214,96,231,175,240,35,234,24,237,226,238,29,189,165,240,170,160,100,244,115,134,39,249,196,155,230,253,9,253,184,137,190,224,121,141,103,198,58,128,208,219,251,132,213,139,188,154,98,150,125,158,187,176,62,147,12,173,255,151,177,16,176,175,6,13,113,171,223,43,50,166,104,54,243,162,109,102,180,188,218,123,117,184,3,93,54,181,180,64,247,177,20,11,0,0,210,52,9,0,14,0,0,0,4,0,0,0,4,32,54,22,4,32,118,22,4,24,22,22,4,24,86,22,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,21,11,0,0,0,0,0,0,158,23,9,0,166,23,9,0,172,23,9,0,177,23,9,0,182,23,9,0,188,23,9,0,3,0,0,0,4,0,0,0,6,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,214,27,9,0,214,28,9,0,86,29,9,0,150,29,9,0,182,29,9,0,198,29,9,0,206,29,9,0,210,29,9,0,212,29,9,0,22,11,0,0,23,11,0,0,24,11,0,0,25,11,0,0,1,0,0,0,44,236,1,0,0,0,0,0,0,0,0,0,0,0,0,0,48,139,0,0,255,255,255,255,58,2,0,0,246,33,9,0,1,0,0,0,80,236,1,0,0,0,0,0,0,0,0,0,0,0,0,0,48,139,0,0,255,255,255,255,58,2,0,0,48,36,9,0,1,0,0,0,116,236,1,0,0,0,0,0,0,0,0,0,0,0,0,0,48,139,0,0,255,255,255,255,151,2,0,0,106,38,9,0,1,0,0,0,152,236,1,0,0,0,0,0,0,0,0,0,0,0,0,0,48,139,0,0,255,255,255,255,51,1,0,0,1,41,9,0,1,0,0,0,152,236,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,152,236,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,228,236,1,0,0,0,0,0,0,0,0,0,0,0,0,0,48,139,0,0,255,255,255,255,117,1,0,0,52,42,9,0,1,0,0,0,228,236,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,228,236,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,48,237,1,0,0,0,0,0,0,0,0,0,0,0,0,0,48,139,0,0,255,255,255,255,87,1,0,0,169,43,9,0,1,0,0,0,48,237,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,48,237,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,124,237,1,0,0,0,0,0,0,0,0,0,0,0,0,0,48,139,0,0,255,255,255,255,251,0,0,0,0,45,9,0,1,0,0,0,124,237,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,124,237,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,200,237,1,0,0,0,0,0,0,0,0,0,0,0,0,0,48,139,0,0,255,255,255,255,121,0,0,0,251,45,9,0,1,0,0,0,200,237,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,200,237,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,20,238,1,0,0,0,0,0,0,0,0,0,0,0,0,0,49,139,0,0,255,255,255,255,51,2,0,0,116,46,9,0,1,0,0,0,200,237,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,124,237,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,48,237,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,228,236,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,152,236,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,26,11,0,0,27,11,0,0,28,11,0,0,29,11,0,0,0,0,0,0,30,11,0,0,31,11,0,0,32,11,0,0,33,11,0,0,0,0,0,0,34,11,0,0,35,11,0,0,36,11,0,0,37,11,0,0,0,0,0,0,38,11,0,0,39,11,0,0,40,11,0,0,41,11,0,0,232,238,1,0,16,239,1,0,216,15,2,0,0,16,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,42,11,0,0,0,0,0,0,0,248,0,0,224,7,0,0,31,0,0,0,4,0,0,0,0,0,255,0,0,255,0,0,255,0,0,0,0,0,0,0,43,11,0,0,7,0,0,0,0,248,0,0,224,7,0,0,31,0,0,0,4,0,0,0,255,0,0,0,0,255,0,0,0,0,255,0,0,0,0,0,44,11,0,0,7,0,0,0,0,248,0,0,224,7,0,0,31,0,0,0,4,0,0,0,0,0,0,255,0,0,255,0,0,255,0,0,0,0,0,0,45,11,0,0,7,0,0,0,0,248,0,0,224,7,0,0,31,0,0,0,4,0,0,0,0,255,0,0,0,0,255,0,0,0,0,255,0,0,0,0,46,11,0,0,7,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,42,11,0,0,0,0,0,0,0,0,0,0,0,0,0,255,8,0,0,0,0,32,0,255,16,0,0,0,0,64,0,255,24,0,0,0,0,97,0,255,32,0,0,0,0,129,0,255,41,0,0,0,0,161,0,255,49,0,0,0,0,194,0,255,57,0,0,0,0,226,0,255,65,0,0,0,0,0,8,255,74,0,0,0,0,32,8,255,82,0,0,0,0,64,8,255,90,0,0,0,0,97,8,255,98,0,0,0,0,129,8,255,106,0,0,0,0,161,8,255,115,0,0,0,0,194,8,255,123,0,0,0,0,226,8,255,131,0,0,0,0,0,16,255,139,0,0,0,0,32,16,255,148,0,0,0,0,64,16,255,156,0,0,0,0,97,16,255,164,0,0,0,0,129,16,255,172,0,0,0,0,161,16,255,180,0,0,0,0,194,16,255,189,0,0,0,0,226,16,255,197,0,0,0,0,0,24,255,205,0,0,0,0,32,24,255,213,0,0,0,0,64,24,255,222,0,0,0,0,97,24,255,230,0,0,0,0,129,24,255,238,0,0,0,0,161,24,255,246,0,0,0,0,194,24,255,255,0,0,0,0,226,24,255,0,4,0,0,0,0,32,255,8,4,0,0,0,32,32,255,16,4,0,0,0,64,32,255,24,4,0,0,0,97,32,255,32,4,0,0,0,129,32,255,41,4,0,0,0,161,32,255,49,4,0,0,0,194,32,255,57,4,0,0,0,226,32,255,65,4,0,0,0,0,41,255,74,4,0,0,0,32,41,255,82,4,0,0,0,64,41,255,90,4,0,0,0,97,41,255,98,4,0,0,0,129,41,255,106,4,0,0,0,161,41,255,115,4,0,0,0,194,41,255,123,4,0,0,0,226,41,255,131,4,0,0,0,0,49,255,139,4,0,0,0,32,49,255,148,4,0,0,0,64,49,255,156,4,0,0,0,97,49,255,164,4,0,0,0,129,49,255,172,4,0,0,0,161,49,255,180,4,0,0,0,194,49,255,189,4,0,0,0,226,49,255,197,4,0,0,0,0,57,255,205,4,0,0,0,32,57,255,213,4,0,0,0,64,57,255,222,4,0,0,0,97,57,255,230,4,0,0,0,129,57,255,238,4,0,0,0,161,57,255,246,4,0,0,0,194,57,255,255,4,0,0,0,226,57,255,0,8,0,0,0,0,65,255,8,8,0,0,0,32,65,255,16,8,0,0,0,64,65,255,24,8,0,0,0,97,65,255,32,8,0,0,0,129,65,255,41,8,0,0,0,161,65,255,49,8,0,0,0,194,65,255,57,8,0,0,0,226,65,255,65,8,0,0,0,0,74,255,74,8,0,0,0,32,74,255,82,8,0,0,0,64,74,255,90,8,0,0,0,97,74,255,98,8,0,0,0,129,74,255,106,8,0,0,0,161,74,255,115,8,0,0,0,194,74,255,123,8,0,0,0,226,74,255,131,8,0,0,0,0,82,255,139,8,0,0,0,32,82,255,148,8,0,0,0,64,82,255,156,8,0,0,0,97,82,255,164,8,0,0,0,129,82,255,172,8,0,0,0,161,82,255,180,8,0,0,0,194,82,255,189,8,0,0,0,226,82,255,197,8,0,0,0,0,90,255,205,8,0,0,0,32,90,255,213,8,0,0,0,64,90,255,222,8,0,0,0,97,90,255,230,8,0,0,0,129,90,255,238,8,0,0,0,161,90,255,246,8,0,0,0,194,90,255,255,8,0,0,0,226,90,255,0,12,0,0,0,0,98,255,8,12,0,0,0,32,98,255,16,12,0,0,0,64,98,255,24,12,0,0,0,97,98,255,32,12,0,0,0,129,98,255,41,12,0,0,0,161,98,255,49,12,0,0,0,194,98,255,57,12,0,0,0,226,98,255,65,12,0,0,0,0,106,255,74,12,0,0,0,32,106,255,82,12,0,0,0,64,106,255,90,12,0,0,0,97,106,255,98,12,0,0,0,129,106,255,106,12,0,0,0,161,106,255,115,12,0,0,0,194,106,255,123,12,0,0,0,226,106,255,131,12,0,0,0,0,115,255,139,12,0,0,0,32,115,255,148,12,0,0,0,64,115,255,156,12,0,0,0,97,115,255,164,12,0,0,0,129,115,255,172,12,0,0,0,161,115,255,180,12,0,0,0,194,115,255,189,12,0,0,0,226,115,255,197,12,0,0,0,0,123,255,205,12,0,0,0,32,123,255,213,12,0,0,0,64,123,255,222,12,0,0,0,97,123,255,230,12,0,0,0,129,123,255,238,12,0,0,0,161,123,255,246,12,0,0,0,194,123,255,255,12,0,0,0,226,123,255,0,16,0,0,0,0,131,255,8,16,0,0,0,32,131,255,16,16,0,0,0,64,131,255,24,16,0,0,0,97,131,255,32,16,0,0,0,129,131,255,41,16,0,0,0,161,131,255,49,16,0,0,0,194,131,255,57,16,0,0,0,226,131,255,65,16,0,0,0,0,139,255,74,16,0,0,0,32,139,255,82,16,0,0,0,64,139,255,90,16,0,0,0,97,139,255,98,16,0,0,0,129,139,255,106,16,0,0,0,161,139,255,115,16,0,0,0,194,139,255,123,16,0,0,0,226,139,255,131,16,0,0,0,0,148,255,139,16,0,0,0,32,148,255,148,16,0,0,0,64,148,255,156,16,0,0,0,97,148,255,164,16,0,0,0,129,148,255,172,16,0,0,0,161,148,255,180,16,0,0,0,194,148,255,189,16,0,0,0,226,148,255,197,16,0,0,0,0,156,255,205,16,0,0,0,32,156,255,213,16,0,0,0,64,156,255,222,16,0,0,0,97,156,255,230,16,0,0,0,129,156,255,238,16,0,0,0,161,156,255,246,16,0,0,0,194,156,255,255,16,0,0,0,226,156,255,0,20,0,0,0,0,164,255,8,20,0,0,0,32,164,255,16,20,0,0,0,64,164,255,24,20,0,0,0,97,164,255,32,20,0,0,0,129,164,255,41,20,0,0,0,161,164,255,49,20,0,0,0,194,164,255,57,20,0,0,0,226,164,255,65,20,0,0,0,0,172,255,74,20,0,0,0,32,172,255,82,20,0,0,0,64,172,255,90,20,0,0,0,97,172,255,98,20,0,0,0,129,172,255,106,20,0,0,0,161,172,255,115,20,0,0,0,194,172,255,123,20,0,0,0,226,172,255,131,20,0,0,0,0,180,255,139,20,0,0,0,32,180,255,148,20,0,0,0,64,180,255,156,20,0,0,0,97,180,255,164,20,0,0,0,129,180,255,172,20,0,0,0,161,180,255,180,20,0,0,0,194,180,255,189,20,0,0,0,226,180,255,197,20,0,0,0,0,189,255,205,20,0,0,0,32,189,255,213,20,0,0,0,64,189,255,222,20,0,0,0,97,189,255,230,20,0,0,0,129,189,255,238,20,0,0,0,161,189,255,246,20,0,0,0,194,189,255,255,20,0,0,0,226,189,255,0,24,0,0,0,0,197,255,8,24,0,0,0,32,197,255,16,24,0,0,0,64,197,255,24,24,0,0,0,97,197,255,32,24,0,0,0,129,197,255,41,24,0,0,0,161,197,255,49,24,0,0,0,194,197,255,57,24,0,0,0,226,197,255,65,24,0,0,0,0,205,255,74,24,0,0,0,32,205,255,82,24,0,0,0,64,205,255,90,24,0,0,0,97,205,255,98,24,0,0,0,129,205,255,106,24,0,0,0,161,205,255,115,24,0,0,0,194,205,255,123,24,0,0,0,226,205,255,131,24,0,0,0,0,213,255,139,24,0,0,0,32,213,255,148,24,0,0,0,64,213,255,156,24,0,0,0,97,213,255,164,24,0,0,0,129,213,255,172,24,0,0,0,161,213,255,180,24,0,0,0,194,213,255,189,24,0,0,0,226,213,255,197,24,0,0,0,0,222,255,205,24,0,0,0,32,222,255,213,24,0,0,0,64,222,255,222,24,0,0,0,97,222,255,230,24,0,0,0,129,222,255,238,24,0,0,0,161,222,255,246,24,0,0,0,194,222,255,255,24,0,0,0,226,222,255,0,28,0,0,0,0,230,255,8,28,0,0,0,32,230,255,16,28,0,0,0,64,230,255,24,28,0,0,0,97,230,255,32,28,0,0,0,129,230,255,41,28,0,0,0,161,230,255,49,28,0,0,0,194,230,255,57,28,0,0,0,226,230,255,65,28,0,0,0,0,238,255,74,28,0,0,0,32,238,255,82,28,0,0,0,64,238,255,90,28,0,0,0,97,238,255,98,28,0,0,0,129,238,255,106,28,0,0,0,161,238,255,115,28,0,0,0,194,238,255,123,28,0,0,0,226,238,255,131,28,0,0,0,0,246,255,139,28,0,0,0,32,246,255,148,28,0,0,0,64,246,255,156,28,0,0,0,97,246,255,164,28,0,0,0,129,246,255,172,28,0,0,0,161,246,255,180,28,0,0,0,194,246,255,189,28,0,0,0,226,246,255,197,28,0,0,0,0,255,255,205,28,0,0,0,32,255,255,213,28,0,0,0,64,255,255,222,28,0,0,0,97,255,255,230,28,0,0,0,129,255,255,238,28,0,0,0,161,255,255,246,28,0,0,0,194,255,255,255,28,0,0,0,226,255,255,0,0,0,255,0,0,0,0,0,0,8,255,0,32,0,0,0,0,16,255,0,64,0,0,0,0,24,255,0,97,0,0,0,0,32,255,0,129,0,0,0,0,41,255,0,161,0,0,0,0,49,255,0,194,0,0,0,0,57,255,0,226,0,0,0,0,65,255,8,0,0,0,0,0,74,255,8,32,0,0,0,0,82,255,8,64,0,0,0,0,90,255,8,97,0,0,0,0,98,255,8,129,0,0,0,0,106,255,8,161,0,0,0,0,115,255,8,194,0,0,0,0,123,255,8,226,0,0,0,0,131,255,16,0,0,0,0,0,139,255,16,32,0,0,0,0,148,255,16,64,0,0,0,0,156,255,16,97,0,0,0,0,164,255,16,129,0,0,0,0,172,255,16,161,0,0,0,0,180,255,16,194,0,0,0,0,189,255,16,226,0,0,0,0,197,255,24,0,0,0,0,0,205,255,24,32,0,0,0,0,213,255,24,64,0,0,0,0,222,255,24,97,0,0,0,0,230,255,24,129,0,0,0,0,238,255,24,161,0,0,0,0,246,255,24,194,0,0,0,0,255,255,24,226,0,0,0,4,0,255,32,0,0,0,0,4,8,255,32,32,0,0,0,4,16,255,32,64,0,0,0,4,24,255,32,97,0,0,0,4,32,255,32,129,0,0,0,4,41,255,32,161,0,0,0,4,49,255,32,194,0,0,0,4,57,255,32,226,0,0,0,4,65,255,41,0,0,0,0,4,74,255,41,32,0,0,0,4,82,255,41,64,0,0,0,4,90,255,41,97,0,0,0,4,98,255,41,129,0,0,0,4,106,255,41,161,0,0,0,4,115,255,41,194,0,0,0,4,123,255,41,226,0,0,0,4,131,255,49,0,0,0,0,4,139,255,49,32,0,0,0,4,148,255,49,64,0,0,0,4,156,255,49,97,0,0,0,4,164,255,49,129,0,0,0,4,172,255,49,161,0,0,0,4,180,255,49,194,0,0,0,4,189,255,49,226,0,0,0,4,197,255,57,0,0,0,0,4,205,255,57,32,0,0,0,4,213,255,57,64,0,0,0,4,222,255,57,97,0,0,0,4,230,255,57,129,0,0,0,4,238,255,57,161,0,0,0,4,246,255,57,194,0,0,0,4,255,255,57,226,0,0,0,8,0,255,65,0,0,0,0,8,8,255,65,32,0,0,0,8,16,255,65,64,0,0,0,8,24,255,65,97,0,0,0,8,32,255,65,129,0,0,0,8,41,255,65,161,0,0,0,8,49,255,65,194,0,0,0,8,57,255,65,226,0,0,0,8,65,255,74,0,0,0,0,8,74,255,74,32,0,0,0,8,82,255,74,64,0,0,0,8,90,255,74,97,0,0,0,8,98,255,74,129,0,0,0,8,106,255,74,161,0,0,0,8,115,255,74,194,0,0,0,8,123,255,74,226,0,0,0,8,131,255,82,0,0,0,0,8,139,255,82,32,0,0,0,8,148,255,82,64,0,0,0,8,156,255,82,97,0,0,0,8,164,255,82,129,0,0,0,8,172,255,82,161,0,0,0,8,180,255,82,194,0,0,0,8,189,255,82,226,0,0,0,8,197,255,90,0,0,0,0,8,205,255,90,32,0,0,0,8,213,255,90,64,0,0,0,8,222,255,90,97,0,0,0,8,230,255,90,129,0,0,0,8,238,255,90,161,0,0,0,8,246,255,90,194,0,0,0,8,255,255,90,226,0,0,0,12,0,255,98,0,0,0,0,12,8,255,98,32,0,0,0,12,16,255,98,64,0,0,0,12,24,255,98,97,0,0,0,12,32,255,98,129,0,0,0,12,41,255,98,161,0,0,0,12,49,255,98,194,0,0,0,12,57,255,98,226,0,0,0,12,65,255,106,0,0,0,0,12,74,255,106,32,0,0,0,12,82,255,106,64,0,0,0,12,90,255,106,97,0,0,0,12,98,255,106,129,0,0,0,12,106,255,106,161,0,0,0,12,115,255,106,194,0,0,0,12,123,255,106,226,0,0,0,12,131,255,115,0,0,0,0,12,139,255,115,32,0,0,0,12,148,255,115,64,0,0,0,12,156,255,115,97,0,0,0,12,164,255,115,129,0,0,0,12,172,255,115,161,0,0,0,12,180,255,115,194,0,0,0,12,189,255,115,226,0,0,0,12,197,255,123,0,0,0,0,12,205,255,123,32,0,0,0,12,213,255,123,64,0,0,0,12,222,255,123,97,0,0,0,12,230,255,123,129,0,0,0,12,238,255,123,161,0,0,0,12,246,255,123,194,0,0,0,12,255,255,123,226,0,0,0,16,0,255,131,0,0,0,0,16,8,255,131,32,0,0,0,16,16,255,131,64,0,0,0,16,24,255,131,97,0,0,0,16,32,255,131,129,0,0,0,16,41,255,131,161,0,0,0,16,49,255,131,194,0,0,0,16,57,255,131,226,0,0,0,16,65,255,139,0,0,0,0,16,74,255,139,32,0,0,0,16,82,255,139,64,0,0,0,16,90,255,139,97,0,0,0,16,98,255,139,129,0,0,0,16,106,255,139,161,0,0,0,16,115,255,139,194,0,0,0,16,123,255,139,226,0,0,0,16,131,255,148,0,0,0,0,16,139,255,148,32,0,0,0,16,148,255,148,64,0,0,0,16,156,255,148,97,0,0,0,16,164,255,148,129,0,0,0,16,172,255,148,161,0,0,0,16,180,255,148,194,0,0,0,16,189,255,148,226,0,0,0,16,197,255,156,0,0,0,0,16,205,255,156,32,0,0,0,16,213,255,156,64,0,0,0,16,222,255,156,97,0,0,0,16,230,255,156,129,0,0,0,16,238,255,156,161,0,0,0,16,246,255,156,194,0,0,0,16,255,255,156,226,0,0,0,20,0,255,164,0,0,0,0,20,8,255,164,32,0,0,0,20,16,255,164,64,0,0,0,20,24,255,164,97,0,0,0,20,32,255,164,129,0,0,0,20,41,255,164,161,0,0,0,20,49,255,164,194,0,0,0,20,57,255,164,226,0,0,0,20,65,255,172,0,0,0,0,20,74,255,172,32,0,0,0,20,82,255,172,64,0,0,0,20,90,255,172,97,0,0,0,20,98,255,172,129,0,0,0,20,106,255,172,161,0,0,0,20,115,255,172,194,0,0,0,20,123,255,172,226,0,0,0,20,131,255,180,0,0,0,0,20,139,255,180,32,0,0,0,20,148,255,180,64,0,0,0,20,156,255,180,97,0,0,0,20,164,255,180,129,0,0,0,20,172,255,180,161,0,0,0,20,180,255,180,194,0,0,0,20,189,255,180,226,0,0,0,20,197,255,189,0,0,0,0,20,205,255,189,32,0,0,0,20,213,255,189,64,0,0,0,20,222,255,189,97,0,0,0,20,230,255,189,129,0,0,0,20,238,255,189,161,0,0,0,20,246,255,189,194,0,0,0,20,255,255,189,226,0,0,0,24,0,255,197,0,0,0,0,24,8,255,197,32,0,0,0,24,16,255,197,64,0,0,0,24,24,255,197,97,0,0,0,24,32,255,197,129,0,0,0,24,41,255,197,161,0,0,0,24,49,255,197,194,0,0,0,24,57,255,197,226,0,0,0,24,65,255,205,0,0,0,0,24,74,255,205,32,0,0,0,24,82,255,205,64,0,0,0,24,90,255,205,97,0,0,0,24,98,255,205,129,0,0,0,24,106,255,205,161,0,0,0,24,115,255,205,194,0,0,0,24,123,255,205,226,0,0,0,24,131,255,213,0,0,0,0,24,139,255,213,32,0,0,0,24,148,255,213,64,0,0,0,24,156,255,213,97,0,0,0,24,164,255,213,129,0,0,0,24,172,255,213,161,0,0,0,24,180,255,213,194,0,0,0,24,189,255,213,226,0,0,0,24,197,255,222,0,0,0,0,24,205,255,222,32,0,0,0,24,213,255,222,64,0,0,0,24,222,255,222,97,0,0,0,24,230,255,222,129,0,0,0,24,238,255,222,161,0,0,0,24,246,255,222,194,0,0,0,24,255,255,222,226,0,0,0,28,0,255,230,0,0,0,0,28,8,255,230,32,0,0,0,28,16,255,230,64,0,0,0,28,24,255,230,97,0,0,0,28,32,255,230,129,0,0,0,28,41,255,230,161,0,0,0,28,49,255,230,194,0,0,0,28,57,255,230,226,0,0,0,28,65,255,238,0,0,0,0,28,74,255,238,32,0,0,0,28,82,255,238,64,0,0,0,28,90,255,238,97,0,0,0,28,98,255,238,129,0,0,0,28,106,255,238,161,0,0,0,28,115,255,238,194,0,0,0,28,123,255,238,226,0,0,0,28,131,255,246,0,0,0,0,28,139,255,246,32,0,0,0,28,148,255,246,64,0,0,0,28,156,255,246,97,0,0,0,28,164,255,246,129,0,0,0,28,172,255,246,161,0,0,0,28,180,255,246,194,0,0,0,28,189,255,246,226,0,0,0,28,197,255,255,0,0,0,0,28,205,255,255,32,0,0,0,28,213,255,255,64,0,0,0,28,222,255,255,97,0,0,0,28,230,255,255,129,0,0,0,28,238,255,255,161,0,0,0,28,246,255,255,194,0,0,0,28,255,255,255,226,0,0,255,0,0,0,0,0,0,0,255,8,0,0,0,0,32,0,255,16,0,0,0,0,64,0,255,24,0,0,0,0,97,0,255,32,0,0,0,0,129,0,255,41,0,0,0,0,161,0,255,49,0,0,0,0,194,0,255,57,0,0,0,0,226,0,255,65,0,0,0,0,0,8,255,74,0,0,0,0,32,8,255,82,0,0,0,0,64,8,255,90,0,0,0,0,97,8,255,98,0,0,0,0,129,8,255,106,0,0,0,0,161,8,255,115,0,0,0,0,194,8,255,123,0,0,0,0,226,8,255,131,0,0,0,0,0,16,255,139,0,0,0,0,32,16,255,148,0,0,0,0,64,16,255,156,0,0,0,0,97,16,255,164,0,0,0,0,129,16,255,172,0,0,0,0,161,16,255,180,0,0,0,0,194,16,255,189,0,0,0,0,226,16,255,197,0,0,0,0,0,24,255,205,0,0,0,0,32,24,255,213,0,0,0,0,64,24,255,222,0,0,0,0,97,24,255,230,0,0,0,0,129,24,255,238,0,0,0,0,161,24,255,246,0,0,0,0,194,24,255,255,0,0,0,0,226,24,255,0,4,0,0,0,0,32,255,8,4,0,0,0,32,32,255,16,4,0,0,0,64,32,255,24,4,0,0,0,97,32,255,32,4,0,0,0,129,32,255,41,4,0,0,0,161,32,255,49,4,0,0,0,194,32,255,57,4,0,0,0,226,32,255,65,4,0,0,0,0,41,255,74,4,0,0,0,32,41,255,82,4,0,0,0,64,41,255,90,4,0,0,0,97,41,255,98,4,0,0,0,129,41,255,106,4,0,0,0,161,41,255,115,4,0,0,0,194,41,255,123,4,0,0,0,226,41,255,131,4,0,0,0,0,49,255,139,4,0,0,0,32,49,255,148,4,0,0,0,64,49,255,156,4,0,0,0,97,49,255,164,4,0,0,0,129,49,255,172,4,0,0,0,161,49,255,180,4,0,0,0,194,49,255,189,4,0,0,0,226,49,255,197,4,0,0,0,0,57,255,205,4,0,0,0,32,57,255,213,4,0,0,0,64,57,255,222,4,0,0,0,97,57,255,230,4,0,0,0,129,57,255,238,4,0,0,0,161,57,255,246,4,0,0,0,194,57,255,255,4,0,0,0,226,57,255,0,8,0,0,0,0,65,255,8,8,0,0,0,32,65,255,16,8,0,0,0,64,65,255,24,8,0,0,0,97,65,255,32,8,0,0,0,129,65,255,41,8,0,0,0,161,65,255,49,8,0,0,0,194,65,255,57,8,0,0,0,226,65,255,65,8,0,0,0,0,74,255,74,8,0,0,0,32,74,255,82,8,0,0,0,64,74,255,90,8,0,0,0,97,74,255,98,8,0,0,0,129,74,255,106,8,0,0,0,161,74,255,115,8,0,0,0,194,74,255,123,8,0,0,0,226,74,255,131,8,0,0,0,0,82,255,139,8,0,0,0,32,82,255,148,8,0,0,0,64,82,255,156,8,0,0,0,97,82,255,164,8,0,0,0,129,82,255,172,8,0,0,0,161,82,255,180,8,0,0,0,194,82,255,189,8,0,0,0,226,82,255,197,8,0,0,0,0,90,255,205,8,0,0,0,32,90,255,213,8,0,0,0,64,90,255,222,8,0,0,0,97,90,255,230,8,0,0,0,129,90,255,238,8,0,0,0,161,90,255,246,8,0,0,0,194,90,255,255,8,0,0,0,226,90,255,0,12,0,0,0,0,98,255,8,12,0,0,0,32,98,255,16,12,0,0,0,64,98,255,24,12,0,0,0,97,98,255,32,12,0,0,0,129,98,255,41,12,0,0,0,161,98,255,49,12,0,0,0,194,98,255,57,12,0,0,0,226,98,255,65,12,0,0,0,0,106,255,74,12,0,0,0,32,106,255,82,12,0,0,0,64,106,255,90,12,0,0,0,97,106,255,98,12,0,0,0,129,106,255,106,12,0,0,0,161,106,255,115,12,0,0,0,194,106,255,123,12,0,0,0,226,106,255,131,12,0,0,0,0,115,255,139,12,0,0,0,32,115,255,148,12,0,0,0,64,115,255,156,12,0,0,0,97,115,255,164,12,0,0,0,129,115,255,172,12,0,0,0,161,115,255,180,12,0,0,0,194,115,255,189,12,0,0,0,226,115,255,197,12,0,0,0,0,123,255,205,12,0,0,0,32,123,255,213,12,0,0,0,64,123,255,222,12,0,0,0,97,123,255,230,12,0,0,0,129,123,255,238,12,0,0,0,161,123,255,246,12,0,0,0,194,123,255,255,12,0,0,0,226,123,255,0,16,0,0,0,0,131,255,8,16,0,0,0,32,131,255,16,16,0,0,0,64,131,255,24,16,0,0,0,97,131,255,32,16,0,0,0,129,131,255,41,16,0,0,0,161,131,255,49,16,0,0,0,194,131,255,57,16,0,0,0,226,131,255,65,16,0,0,0,0,139,255,74,16,0,0,0,32,139,255,82,16,0,0,0,64,139,255,90,16,0,0,0,97,139,255,98,16,0,0,0,129,139,255,106,16,0,0,0,161,139,255,115,16,0,0,0,194,139,255,123,16,0,0,0,226,139,255,131,16,0,0,0,0,148,255,139,16,0,0,0,32,148,255,148,16,0,0,0,64,148,255,156,16,0,0,0,97,148,255,164,16,0,0,0,129,148,255,172,16,0,0,0,161,148,255,180,16,0,0,0,194,148,255,189,16,0,0,0,226,148,255,197,16,0,0,0,0,156,255,205,16,0,0,0,32,156,255,213,16,0,0,0,64,156,255,222,16,0,0,0,97,156,255,230,16,0,0,0,129,156,255,238,16,0,0,0,161,156,255,246,16,0,0,0,194,156,255,255,16,0,0,0,226,156,255,0,20,0,0,0,0,164,255,8,20,0,0,0,32,164,255,16,20,0,0,0,64,164,255,24,20,0,0,0,97,164,255,32,20,0,0,0,129,164,255,41,20,0,0,0,161,164,255,49,20,0,0,0,194,164,255,57,20,0,0,0,226,164,255,65,20,0,0,0,0,172,255,74,20,0,0,0,32,172,255,82,20,0,0,0,64,172,255,90,20,0,0,0,97,172,255,98,20,0,0,0,129,172,255,106,20,0,0,0,161,172,255,115,20,0,0,0,194,172,255,123,20,0,0,0,226,172,255,131,20,0,0,0,0,180,255,139,20,0,0,0,32,180,255,148,20,0,0,0,64,180,255,156,20,0,0,0,97,180,255,164,20,0,0,0,129,180,255,172,20,0,0,0,161,180,255,180,20,0,0,0,194,180,255,189,20,0,0,0,226,180,255,197,20,0,0,0,0,189,255,205,20,0,0,0,32,189,255,213,20,0,0,0,64,189,255,222,20,0,0,0,97,189,255,230,20,0,0,0,129,189,255,238,20,0,0,0,161,189,255,246,20,0,0,0,194,189,255,255,20,0,0,0,226,189,255,0,24,0,0,0,0,197,255,8,24,0,0,0,32,197,255,16,24,0,0,0,64,197,255,24,24,0,0,0,97,197,255,32,24,0,0,0,129,197,255,41,24,0,0,0,161,197,255,49,24,0,0,0,194,197,255,57,24,0,0,0,226,197,255,65,24,0,0,0,0,205,255,74,24,0,0,0,32,205,255,82,24,0,0,0,64,205,255,90,24,0,0,0,97,205,255,98,24,0,0,0,129,205,255,106,24,0,0,0,161,205,255,115,24,0,0,0,194,205,255,123,24,0,0,0,226,205,255,131,24,0,0,0,0,213,255,139,24,0,0,0,32,213,255,148,24,0,0,0,64,213,255,156,24,0,0,0,97,213,255,164,24,0,0,0,129,213,255,172,24,0,0,0,161,213,255,180,24,0,0,0,194,213,255,189,24,0,0,0,226,213,255,197,24,0,0,0,0,222,255,205,24,0,0,0,32,222,255,213,24,0,0,0,64,222,255,222,24,0,0,0,97,222,255,230,24,0,0,0,129,222,255,238,24,0,0,0,161,222,255,246,24,0,0,0,194,222,255,255,24,0,0,0,226,222,255,0,28,0,0,0,0,230,255,8,28,0,0,0,32,230,255,16,28,0,0,0,64,230,255,24,28,0,0,0,97,230,255,32,28,0,0,0,129,230,255,41,28,0,0,0,161,230,255,49,28,0,0,0,194,230,255,57,28,0,0,0,226,230,255,65,28,0,0,0,0,238,255,74,28,0,0,0,32,238,255,82,28,0,0,0,64,238,255,90,28,0,0,0,97,238,255,98,28,0,0,0,129,238,255,106,28,0,0,0,161,238,255,115,28,0,0,0,194,238,255,123,28,0,0,0,226,238,255,131,28,0,0,0,0,246,255,139,28,0,0,0,32,246,255,148,28,0,0,0,64,246,255,156,28,0,0,0,97,246,255,164,28,0,0,0,129,246,255,172,28,0,0,0,161,246,255,180,28,0,0,0,194,246,255,189,28,0,0,0,226,246,255,197,28,0,0,0,0,255,255,205,28,0,0,0,32,255,255,213,28,0,0,0,64,255,255,222,28,0,0,0,97,255,255,230,28,0,0,0,129,255,255,238,28,0,0,0,161,255,255,246,28,0,0,0,194,255,255,255,28,0,0,0,226,255,0,0,0,0,255,0,0,0,0,0,0,8,255,0,32,0,0,0,0,16,255,0,64,0,0,0,0,24,255,0,97,0,0,0,0,32,255,0,129,0,0,0,0,41,255,0,161,0,0,0,0,49,255,0,194,0,0,0,0,57,255,0,226,0,0,0,0,65,255,8,0,0,0,0,0,74,255,8,32,0,0,0,0,82,255,8,64,0,0,0,0,90,255,8,97,0,0,0,0,98,255,8,129,0,0,0,0,106,255,8,161,0,0,0,0,115,255,8,194,0,0,0,0,123,255,8,226,0,0,0,0,131,255,16,0,0,0,0,0,139,255,16,32,0,0,0,0,148,255,16,64,0,0,0,0,156,255,16,97,0,0,0,0,164,255,16,129,0,0,0,0,172,255,16,161,0,0,0,0,180,255,16,194,0,0,0,0,189,255,16,226,0,0,0,0,197,255,24,0,0,0,0,0,205,255,24,32,0,0,0,0,213,255,24,64,0,0,0,0,222,255,24,97,0,0,0,0,230,255,24,129,0,0,0,0,238,255,24,161,0,0,0,0,246,255,24,194,0,0,0,0,255,255,24,226,0,0,0,4,0,255,32,0,0,0,0,4,8,255,32,32,0,0,0,4,16,255,32,64,0,0,0,4,24,255,32,97,0,0,0,4,32,255,32,129,0,0,0,4,41,255,32,161,0,0,0,4,49,255,32,194,0,0,0,4,57,255,32,226,0,0,0,4,65,255,41,0,0,0,0,4,74,255,41,32,0,0,0,4,82,255,41,64,0,0,0,4,90,255,41,97,0,0,0,4,98,255,41,129,0,0,0,4,106,255,41,161,0,0,0,4,115,255,41,194,0,0,0,4,123,255,41,226,0,0,0,4,131,255,49,0,0,0,0,4,139,255,49,32,0,0,0,4,148,255,49,64,0,0,0,4,156,255,49,97,0,0,0,4,164,255,49,129,0,0,0,4,172,255,49,161,0,0,0,4,180,255,49,194,0,0,0,4,189,255,49,226,0,0,0,4,197,255,57,0,0,0,0,4,205,255,57,32,0,0,0,4,213,255,57,64,0,0,0,4,222,255,57,97,0,0,0,4,230,255,57,129,0,0,0,4,238,255,57,161,0,0,0,4,246,255,57,194,0,0,0,4,255,255,57,226,0,0,0,8,0,255,65,0,0,0,0,8,8,255,65,32,0,0,0,8,16,255,65,64,0,0,0,8,24,255,65,97,0,0,0,8,32,255,65,129,0,0,0,8,41,255,65,161,0,0,0,8,49,255,65,194,0,0,0,8,57,255,65,226,0,0,0,8,65,255,74,0,0,0,0,8,74,255,74,32,0,0,0,8,82,255,74,64,0,0,0,8,90,255,74,97,0,0,0,8,98,255,74,129,0,0,0,8,106,255,74,161,0,0,0,8,115,255,74,194,0,0,0,8,123,255,74,226,0,0,0,8,131,255,82,0,0,0,0,8,139,255,82,32,0,0,0,8,148,255,82,64,0,0,0,8,156,255,82,97,0,0,0,8,164,255,82,129,0,0,0,8,172,255,82,161,0,0,0,8,180,255,82,194,0,0,0,8,189,255,82,226,0,0,0,8,197,255,90,0,0,0,0,8,205,255,90,32,0,0,0,8,213,255,90,64,0,0,0,8,222,255,90,97,0,0,0,8,230,255,90,129,0,0,0,8,238,255,90,161,0,0,0,8,246,255,90,194,0,0,0,8,255,255,90,226,0,0,0,12,0,255,98,0,0,0,0,12,8,255,98,32,0,0,0,12,16,255,98,64,0,0,0,12,24,255,98,97,0,0,0,12,32,255,98,129,0,0,0,12,41,255,98,161,0,0,0,12,49,255,98,194,0,0,0,12,57,255,98,226,0,0,0,12,65,255,106,0,0,0,0,12,74,255,106,32,0,0,0,12,82,255,106,64,0,0,0,12,90,255,106,97,0,0,0,12,98,255,106,129,0,0,0,12,106,255,106,161,0,0,0,12,115,255,106,194,0,0,0,12,123,255,106,226,0,0,0,12,131,255,115,0,0,0,0,12,139,255,115,32,0,0,0,12,148,255,115,64,0,0,0,12,156,255,115,97,0,0,0,12,164,255,115,129,0,0,0,12,172,255,115,161,0,0,0,12,180,255,115,194,0,0,0,12,189,255,115,226,0,0,0,12,197,255,123,0,0,0,0,12,205,255,123,32,0,0,0,12,213,255,123,64,0,0,0,12,222,255,123,97,0,0,0,12,230,255,123,129,0,0,0,12,238,255,123,161,0,0,0,12,246,255,123,194,0,0,0,12,255,255,123,226,0,0,0,16,0,255,131,0,0,0,0,16,8,255,131,32,0,0,0,16,16,255,131,64,0,0,0,16,24,255,131,97,0,0,0,16,32,255,131,129,0,0,0,16,41,255,131,161,0,0,0,16,49,255,131,194,0,0,0,16,57,255,131,226,0,0,0,16,65,255,139,0,0,0,0,16,74,255,139,32,0,0,0,16,82,255,139,64,0,0,0,16,90,255,139,97,0,0,0,16,98,255,139,129,0,0,0,16,106,255,139,161,0,0,0,16,115,255,139,194,0,0,0,16,123,255,139,226,0,0,0,16,131,255,148,0,0,0,0,16,139,255,148,32,0,0,0,16,148,255,148,64,0,0,0,16,156,255,148,97,0,0,0,16,164,255,148,129,0,0,0,16,172,255,148,161,0,0,0,16,180,255,148,194,0,0,0,16,189,255,148,226,0,0,0,16,197,255,156,0,0,0,0,16,205,255,156,32,0,0,0,16,213,255,156,64,0,0,0,16,222,255,156,97,0,0,0,16,230,255,156,129,0,0,0,16,238,255,156,161,0,0,0,16,246,255,156,194,0,0,0,16,255,255,156,226,0,0,0,20,0,255,164,0,0,0,0,20,8,255,164,32,0,0,0,20,16,255,164,64,0,0,0,20,24,255,164,97,0,0,0,20,32,255,164,129,0,0,0,20,41,255,164,161,0,0,0,20,49,255,164,194,0,0,0,20,57,255,164,226,0,0,0,20,65,255,172,0,0,0,0,20,74,255,172,32,0,0,0,20,82,255,172,64,0,0,0,20,90,255,172,97,0,0,0,20,98,255,172,129,0,0,0,20,106,255,172,161,0,0,0,20,115,255,172,194,0,0,0,20,123,255,172,226,0,0,0,20,131,255,180,0,0,0,0,20,139,255,180,32,0,0,0,20,148,255,180,64,0,0,0,20,156,255,180,97,0,0,0,20,164,255,180,129,0,0,0,20,172,255,180,161,0,0,0,20,180,255,180,194,0,0,0,20,189,255,180,226,0,0,0,20,197,255,189,0,0,0,0,20,205,255,189,32,0,0,0,20,213,255,189,64,0,0,0,20,222,255,189,97,0,0,0,20,230,255,189,129,0,0,0,20,238,255,189,161,0,0,0,20,246,255,189,194,0,0,0,20,255,255,189,226,0,0,0,24,0,255,197,0,0,0,0,24,8,255,197,32,0,0,0,24,16,255,197,64,0,0,0,24,24,255,197,97,0,0,0,24,32,255,197,129,0,0,0,24,41,255,197,161,0,0,0,24,49,255,197,194,0,0,0,24,57,255,197,226,0,0,0,24,65,255,205,0,0,0,0,24,74,255,205,32,0,0,0,24,82,255,205,64,0,0,0,24,90,255,205,97,0,0,0,24,98,255,205,129,0,0,0,24,106,255,205,161,0,0,0,24,115,255,205,194,0,0,0,24,123,255,205,226,0,0,0,24,131,255,213,0,0,0,0,24,139,255,213,32,0,0,0,24,148,255,213,64,0,0,0,24,156,255,213,97,0,0,0,24,164,255,213,129,0,0,0,24,172,255,213,161,0,0,0,24,180,255,213,194,0,0,0,24,189,255,213,226,0,0,0,24,197,255,222,0,0,0,0,24,205,255,222,32,0,0,0,24,213,255,222,64,0,0,0,24,222,255,222,97,0,0,0,24,230,255,222,129,0,0,0,24,238,255,222,161,0,0,0,24,246,255,222,194,0,0,0,24,255,255,222,226,0,0,0,28,0,255,230,0,0,0,0,28,8,255,230,32,0,0,0,28,16,255,230,64,0,0,0,28,24,255,230,97,0,0,0,28,32,255,230,129,0,0,0,28,41,255,230,161,0,0,0,28,49,255,230,194,0,0,0,28,57,255,230,226,0,0,0,28,65,255,238,0,0,0,0,28,74,255,238,32,0,0,0,28,82,255,238,64,0,0,0,28,90,255,238,97,0,0,0,28,98,255,238,129,0,0,0,28,106,255,238,161,0,0,0,28,115,255,238,194,0,0,0,28,123,255,238,226,0,0,0,28,131,255,246,0,0,0,0,28,139,255,246,32,0,0,0,28,148,255,246,64,0,0,0,28,156,255,246,97,0,0,0,28,164,255,246,129,0,0,0,28,172,255,246,161,0,0,0,28,180,255,246,194,0,0,0,28,189,255,246,226,0,0,0,28,197,255,255,0,0,0,0,28,205,255,255,32,0,0,0,28,213,255,255,64,0,0,0,28,222,255,255,97,0,0,0,28,230,255,255,129,0,0,0,28,238,255,255,161,0,0,0,28,246,255,255,194,0,0,0,28,255,255,255,226,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,42,11,0,0,0,0,0,0,0,0,255,0,0,255,0,0,255,0,0,0,2,0,0,0,0,248,0,0,224,7,0,0,31,0,0,0,0,0,0,0,47,11,0,0,1,0,0,0,0,0,255,0,0,255,0,0,255,0,0,0,2,0,0,0,0,124,0,0,224,3,0,0,31,0,0,0,0,0,0,0,48,11,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,42,11,0,0,0,0,0,0,255,255,255,255,255,255,255,255,4,24,22,22,4,24,22,22,0,2,0,0,0,0,0,0,49,11,0,0,4,24,22,22,4,24,22,22,112,0,0,0,0,0,0,0,50,11,0,0,4,24,22,22,4,24,22,22,112,2,0,0,0,0,0,0,51,11,0,0,4,24,22,22,4,24,22,22,3,0,0,0,0,0,0,0,52,11,0,0,4,24,22,22,4,24,22,22,3,2,0,0,0,0,0,0,53,11,0,0,4,24,22,22,4,24,22,22,115,0,0,0,0,0,0,0,54,11,0,0,4,24,22,22,4,24,22,22,115,2,0,0,0,0,0,0,55,11,0,0,4,24,22,22,4,24,86,22,0,2,0,0,0,0,0,0,56,11,0,0,4,24,22,22,4,24,86,22,112,0,0,0,0,0,0,0,57,11,0,0,4,24,22,22,4,24,86,22,112,2,0,0,0,0,0,0,58,11,0,0,4,24,22,22,4,24,86,22,3,0,0,0,0,0,0,0,59,11,0,0,4,24,22,22,4,24,86,22,3,2,0,0,0,0,0,0,60,11,0,0,4,24,22,22,4,24,86,22,115,0,0,0,0,0,0,0,61,11,0,0,4,24,22,22,4,24,86,22,115,2,0,0,0,0,0,0,62,11,0,0,4,24,22,22,4,32,54,22,0,2,0,0,0,0,0,0,63,11,0,0,4,24,22,22,4,32,54,22,112,0,0,0,0,0,0,0,64,11,0,0,4,24,22,22,4,32,54,22,112,2,0,0,0,0,0,0,65,11,0,0,4,24,22,22,4,32,54,22,3,0,0,0,0,0,0,0,66,11,0,0,4,24,22,22,4,32,54,22,3,2,0,0,0,0,0,0,67,11,0,0,4,24,22,22,4,32,54,22,115,0,0,0,0,0,0,0,68,11,0,0,4,24,22,22,4,32,54,22,115,2,0,0,0,0,0,0,69,11,0,0,4,24,86,22,4,24,22,22,0,2,0,0,0,0,0,0,70,11,0,0,4,24,86,22,4,24,22,22,112,0,0,0,0,0,0,0,71,11,0,0,4,24,86,22,4,24,22,22,112,2,0,0,0,0,0,0,72,11,0,0,4,24,86,22,4,24,22,22,3,0,0,0,0,0,0,0,73,11,0,0,4,24,86,22,4,24,22,22,3,2,0,0,0,0,0,0,74,11,0,0,4,24,86,22,4,24,22,22,115,0,0,0,0,0,0,0,75,11,0,0,4,24,86,22,4,24,22,22,115,2,0,0,0,0,0,0,76,11,0,0,4,24,86,22,4,24,86,22,0,2,0,0,0,0,0,0,77,11,0,0,4,24,86,22,4,24,86,22,112,0,0,0,0,0,0,0,78,11,0,0,4,24,86,22,4,24,86,22,112,2,0,0,0,0,0,0,79,11,0,0,4,24,86,22,4,24,86,22,3,0,0,0,0,0,0,0,80,11,0,0,4,24,86,22,4,24,86,22,3,2,0,0,0,0,0,0,81,11,0,0,4,24,86,22,4,24,86,22,115,0,0,0,0,0,0,0,82,11,0,0,4,24,86,22,4,24,86,22,115,2,0,0,0,0,0,0,83,11,0,0,4,24,86,22,4,32,54,22,0,2,0,0,0,0,0,0,84,11,0,0,4,24,86,22,4,32,54,22,112,0,0,0,0,0,0,0,85,11,0,0,4,24,86,22,4,32,54,22,112,2,0,0,0,0,0,0,86,11,0,0,4,24,86,22,4,32,54,22,3,0,0,0,0,0,0,0,87,11,0,0,4,24,86,22,4,32,54,22,3,2,0,0,0,0,0,0,88,11,0,0,4,24,86,22,4,32,54,22,115,0,0,0,0,0,0,0,89,11,0,0,4,24,86,22,4,32,54,22,115,2,0,0,0,0,0,0,90,11,0,0,4,32,54,22,4,24,22,22,0,2,0,0,0,0,0,0,91,11,0,0,4,32,54,22,4,24,22,22,112,0,0,0,0,0,0,0,92,11,0,0,4,32,54,22,4,24,22,22,112,2,0,0,0,0,0,0,93,11,0,0,4,32,54,22,4,24,22,22,3,0,0,0,0,0,0,0,94,11,0,0,4,32,54,22,4,24,22,22,3,2,0,0,0,0,0,0,95,11,0,0,4,32,54,22,4,24,22,22,115,0,0,0,0,0,0,0,96,11,0,0,4,32,54,22,4,24,22,22,115,2,0,0,0,0,0,0,97,11,0,0,4,32,54,22,4,24,86,22,0,2,0,0,0,0,0,0,98,11,0,0,4,32,54,22,4,24,86,22,112,0,0,0,0,0,0,0,99,11,0,0,4,32,54,22,4,24,86,22,112,2,0,0,0,0,0,0,100,11,0,0,4,32,54,22,4,24,86,22,3,0,0,0,0,0,0,0,101,11,0,0,4,32,54,22,4,24,86,22,3,2,0,0,0,0,0,0,102,11,0,0,4,32,54,22,4,24,86,22,115,0,0,0,0,0,0,0,103,11,0,0,4,32,54,22,4,24,86,22,115,2,0,0,0,0,0,0,104,11,0,0,4,32,54,22,4,32,54,22,0,2,0,0,0,0,0,0,105,11,0,0,4,32,54,22,4,32,54,22,112,0,0,0,0,0,0,0,106,11,0,0,4,32,54,22,4,32,54,22,112,2,0,0,0,0,0,0,107,11,0,0,4,32,54,22,4,32,54,22,3,0,0,0,0,0,0,0,108,11,0,0,4,32,54,22,4,32,54,22,3,2,0,0,0,0,0,0,109,11,0,0,4,32,54,22,4,32,54,22,115,0,0,0,0,0,0,0,110,11,0,0,4,32,54,22,4,32,54,22,115,2,0,0,0,0,0,0,111,11,0,0,4,32,70,22,4,24,22,22,0,2,0,0,0,0,0,0,112,11,0,0,4,32,70,22,4,24,22,22,112,0,0,0,0,0,0,0,113,11,0,0,4,32,70,22,4,24,22,22,112,2,0,0,0,0,0,0,114,11,0,0,4,32,70,22,4,24,22,22,3,0,0,0,0,0,0,0,115,11,0,0,4,32,70,22,4,24,22,22,3,2,0,0,0,0,0,0,116,11,0,0,4,32,70,22,4,24,22,22,115,0,0,0,0,0,0,0,117,11,0,0,4,32,70,22,4,24,22,22,115,2,0,0,0,0,0,0,118,11,0,0,4,32,70,22,4,24,86,22,0,2,0,0,0,0,0,0,119,11,0,0,4,32,70,22,4,24,86,22,112,0,0,0,0,0,0,0,120,11,0,0,4,32,70,22,4,24,86,22,112,2,0,0,0,0,0,0,121,11,0,0,4,32,70,22,4,24,86,22,3,0,0,0,0,0,0,0,122,11,0,0,4,32,70,22,4,24,86,22,3,2,0,0,0,0,0,0,123,11,0,0,4,32,70,22,4,24,86,22,115,0,0,0,0,0,0,0,124,11,0,0,4,32,70,22,4,24,86,22,115,2,0,0,0,0,0,0,125,11,0,0,4,32,70,22,4,32,54,22,0,2,0,0,0,0,0,0,126,11,0,0,4,32,70,22,4,32,54,22,112,0,0,0,0,0,0,0,127,11,0,0,4,32,70,22,4,32,54,22,112,2,0,0,0,0,0,0,128,11,0,0,4,32,70,22,4,32,54,22,3,0,0,0,0,0,0,0,129,11,0,0,4,32,70,22,4,32,54,22,3,2,0,0,0,0,0,0,130,11,0,0,4,32,70,22,4,32,54,22,115,0,0,0,0,0,0,0,131,11,0,0,4,32,70,22,4,32,54,22,115,2,0,0,0,0,0,0,132,11,0,0,4,32,118,22,4,24,22,22,0,2,0,0,0,0,0,0,133,11,0,0,4,32,118,22,4,24,22,22,112,0,0,0,0,0,0,0,134,11,0,0,4,32,118,22,4,24,22,22,112,2,0,0,0,0,0,0,135,11,0,0,4,32,118,22,4,24,22,22,3,0,0,0,0,0,0,0,136,11,0,0,4,32,118,22,4,24,22,22,3,2,0,0,0,0,0,0,137,11,0,0,4,32,118,22,4,24,22,22,115,0,0,0,0,0,0,0,138,11,0,0,4,32,118,22,4,24,22,22,115,2,0,0,0,0,0,0,139,11,0,0,4,32,118,22,4,24,86,22,0,2,0,0,0,0,0,0,140,11,0,0,4,32,118,22,4,24,86,22,112,0,0,0,0,0,0,0,141,11,0,0,4,32,118,22,4,24,86,22,112,2,0,0,0,0,0,0,142,11,0,0,4,32,118,22,4,24,86,22,3,0,0,0,0,0,0,0,143,11,0,0,4,32,118,22,4,24,86,22,3,2,0,0,0,0,0,0,144,11,0,0,4,32,118,22,4,24,86,22,115,0,0,0,0,0,0,0,145,11,0,0,4,32,118,22,4,24,86,22,115,2,0,0,0,0,0,0,146,11,0,0,4,32,118,22,4,32,54,22,0,2,0,0,0,0,0,0,147,11,0,0,4,32,118,22,4,32,54,22,112,0,0,0,0,0,0,0,148,11,0,0,4,32,118,22,4,32,54,22,112,2,0,0,0,0,0,0,149,11,0,0,4,32,118,22,4,32,54,22,3,0,0,0,0,0,0,0,150,11,0,0,4,32,118,22,4,32,54,22,3,2,0,0,0,0,0,0,151,11,0,0,4,32,118,22,4,32,54,22,115,0,0,0,0,0,0,0,152,11,0,0,4,32,118,22,4,32,54,22,115,2,0,0,0,0,0,0,153,11,0,0,4,32,134,22,4,24,22,22,0,2,0,0,0,0,0,0,154,11,0,0,4,32,134,22,4,24,22,22,112,0,0,0,0,0,0,0,155,11,0,0,4,32,134,22,4,24,22,22,112,2,0,0,0,0,0,0,156,11,0,0,4,32,134,22,4,24,22,22,3,0,0,0,0,0,0,0,157,11,0,0,4,32,134,22,4,24,22,22,3,2,0,0,0,0,0,0,158,11,0,0,4,32,134,22,4,24,22,22,115,0,0,0,0,0,0,0,159,11,0,0,4,32,134,22,4,24,22,22,115,2,0,0,0,0,0,0,160,11,0,0,4,32,134,22,4,24,86,22,0,2,0,0,0,0,0,0,161,11,0,0,4,32,134,22,4,24,86,22,112,0,0,0,0,0,0,0,162,11,0,0,4,32,134,22,4,24,86,22,112,2,0,0,0,0,0,0,163,11,0,0,4,32,134,22,4,24,86,22,3,0,0,0,0,0,0,0,164,11,0,0,4,32,134,22,4,24,86,22,3,2,0,0,0,0,0,0,165,11,0,0,4,32,134,22,4,24,86,22,115,0,0,0,0,0,0,0,166,11,0,0,4,32,134,22,4,24,86,22,115,2,0,0,0,0,0,0,167,11,0,0,4,32,134,22,4,32,54,22,0,2,0,0,0,0,0,0,168,11,0,0,4,32,134,22,4,32,54,22,112,0,0,0,0,0,0,0,169,11,0,0,4,32,134,22,4,32,54,22,112,2,0,0,0,0,0,0,170,11,0,0,4,32,134,22,4,32,54,22,3,0,0,0,0,0,0,0,171,11,0,0,4,32,134,22,4,32,54,22,3,2,0,0,0,0,0,0,172,11,0,0,4,32,134,22,4,32,54,22,115,0,0,0,0,0,0,0,173,11,0,0,4,32,134,22,4,32,54,22,115,2,0,0,0,0,0,0,174,11,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,175,11,0,0,73,76,9,0,9,0,0,0,8,0,0,0,4,32,54,22,4,32,118,22,4,32,70,22,4,32,134,22,4,24,22,22,4,24,86,22,2,16,21,21,2,15,19,21,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,232,81,9,0,69,57,9,0,176,11,0,0,177,11,0,0,244,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,155,0,0,0,0,0,0,0,0,0,0,0,117,0,0,0,0,0,0,0,42,0,0,0,43,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,40,0,0,0,0,0,0,0,0,0,0,0,225,0,0,0,224,0,0,0,226,0,0,0,72,0,0,0,57,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,41,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,44,0,0,0,75,0,0,0,78,0,0,0,77,0,0,0,74,0,0,0,80,0,0,0,82,0,0,0,79,0,0,0,81,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,73,0,0,0,76,0,0,0,0,0,0,0,39,0,0,0,30,0,0,0,31,0,0,0,32,0,0,0,33,0,0,0,34,0,0,0,35,0,0,0,36,0,0,0,37,0,0,0,38,0,0,0,0,0,0,0,51,0,0,0,0,0,0,0,46,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,5,0,0,0,6,0,0,0,7,0,0,0,8,0,0,0,9,0,0,0,10,0,0,0,11,0,0,0,12,0,0,0,13,0,0,0,14,0,0,0,15,0,0,0,16,0,0,0,17,0,0,0,18,0,0,0,19,0,0,0,20,0,0,0,21,0,0,0,22,0,0,0,23,0,0,0,24,0,0,0,25,0,0,0,26,0,0,0,27,0,0,0,28,0,0,0,29,0,0,0,227,0,0,0,0,0,0,0,101,0,0,0,0,0,0,0,0,0,0,0,98,0,0,0,89,0,0,0,90,0,0,0,91,0,0,0,92,0,0,0,93,0,0,0,94,0,0,0,95,0,0,0,96,0,0,0,97,0,0,0,85,0,0,0,87,0,0,0,0,0,0,0,86,0,0,0,99,0,0,0,84,0,0,0,58,0,0,0,59,0,0,0,60,0,0,0,61,0,0,0,62,0,0,0,63,0,0,0,64,0,0,0,65,0,0,0,66,0,0,0,67,0,0,0,68,0,0,0,69,0,0,0,104,0,0,0,105,0,0,0,106,0,0,0,107,0,0,0,108,0,0,0,109,0,0,0,110,0,0,0,111,0,0,0,112,0,0,0,113,0,0,0,114,0,0,0,115,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,83,0,0,0,71,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,45,0,0,0,129,0,0,0,128,0,0,0,2,1,0,0,3,1,0,0,0,0,0,0,5,1,0,0,0,0,0,0,6,1,0,0,129,0,0,0,128,0,0,0,0,0,0,0,0,0,0,0,51,0,0,0,46,0,0,0,54,0,0,0,45,0,0,0,55,0,0,0,56,0,0,0,53,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,47,0,0,0,49,0,0,0,48,0,0,0,52,0,0,0,232,81,9,0,147,67,9,0,178,11,0,0,0,0,0,0,232,74,9,0,69,73,9,0,179,11,0,0,180,11,0,0,72,74,9,0,77,74,9,0,181,11,0,0,1,0,0,0,232,74,9,0,238,74,9,0,182,11,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,97,0,0,0,98,0,0,0,99,0,0,0,100,0,0,0,101,0,0,0,102,0,0,0,103,0,0,0,104,0,0,0,105,0,0,0,106,0,0,0,107,0,0,0,108,0,0,0,109,0,0,0,110,0,0,0,111,0,0,0,112,0,0,0,113,0,0,0,114,0,0,0,115,0,0,0,116,0,0,0,117,0,0,0,118,0,0,0,119,0,0,0,120,0,0,0,121,0,0,0,122,0,0,0,49,0,0,0,50,0,0,0,51,0,0,0,52,0,0,0,53,0,0,0,54,0,0,0,55,0,0,0,56,0,0,0,57,0,0,0,48,0,0,0,13,0,0,0,27,0,0,0,8,0,0,0,9,0,0,0,32,0,0,0,45,0,0,0,61,0,0,0,91,0,0,0,93,0,0,0,92,0,0,0,35,0,0,0,59,0,0,0,39,0,0,0,96,0,0,0,44,0,0,0,46,0,0,0,47,0,0,0,57,0,0,64,58,0,0,64,59,0,0,64,60,0,0,64,61,0,0,64,62,0,0,64,63,0,0,64,64,0,0,64,65,0,0,64,66,0,0,64,67,0,0,64,68,0,0,64,69,0,0,64,70,0,0,64,71,0,0,64,72,0,0,64,73,0,0,64,74,0,0,64,75,0,0,64,127,0,0,0,77,0,0,64,78,0,0,64,79,0,0,64,80,0,0,64,81,0,0,64,82,0,0,64,83,0,0,64,84,0,0,64,85,0,0,64,86,0,0,64,87,0,0,64,88,0,0,64,89,0,0,64,90,0,0,64,91,0,0,64,92,0,0,64,93,0,0,64,94,0,0,64,95,0,0,64,96,0,0,64,97,0,0,64,98,0,0,64,99,0,0,64,0,0,0,0,101,0,0,64,102,0,0,64,103,0,0,64,104,0,0,64,105,0,0,64,106,0,0,64,107,0,0,64,108,0,0,64,109,0,0,64,110,0,0,64,111,0,0,64,112,0,0,64,113,0,0,64,114,0,0,64,115,0,0,64,116,0,0,64,117,0,0,64,118,0,0,64,119,0,0,64,120,0,0,64,121,0,0,64,122,0,0,64,123,0,0,64,124,0,0,64,125,0,0,64,126,0,0,64,127,0,0,64,128,0,0,64,129,0,0,64,0,0,0,0,0,0,0,0,0,0,0,0,133,0,0,64,134,0,0,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,153,0,0,64,154,0,0,64,155,0,0,64,156,0,0,64,157,0,0,64,158,0,0,64,159,0,0,64,160,0,0,64,161,0,0,64,162,0,0,64,163,0,0,64,164,0,0,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,176,0,0,64,177,0,0,64,178,0,0,64,179,0,0,64,180,0,0,64,181,0,0,64,182,0,0,64,183,0,0,64,184,0,0,64,185,0,0,64,186,0,0,64,187,0,0,64,188,0,0,64,189,0,0,64,190,0,0,64,191,0,0,64,192,0,0,64,193,0,0,64,194,0,0,64,195,0,0,64,196,0,0,64,197,0,0,64,198,0,0,64,199,0,0,64,200,0,0,64,201,0,0,64,202,0,0,64,203,0,0,64,204,0,0,64,205,0,0,64,206,0,0,64,207,0,0,64,208,0,0,64,209,0,0,64,210,0,0,64,211,0,0,64,212,0,0,64,213,0,0,64,214,0,0,64,215,0,0,64,216,0,0,64,217,0,0,64,218,0,0,64,219,0,0,64,220,0,0,64,221,0,0,64,0,0,0,0,0,0,0,0,224,0,0,64,225,0,0,64,226,0,0,64,227,0,0,64,228,0,0,64,229,0,0,64,230,0,0,64,231,0,0,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,64,2,1,0,64,3,1,0,64,4,1,0,64,5,1,0,64,6,1,0,64,7,1,0,64,8,1,0,64,9,1,0,64,10,1,0,64,11,1,0,64,12,1,0,64,13,1,0,64,14,1,0,64,15,1,0,64,16,1,0,64,17,1,0,64,18,1,0,64,19,1,0,64,20,1,0,64,21,1,0,64,22,1,0,64,23,1,0,64,24,1,0,64,25,1,0,64,26,1,0,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,196,26,2,0,100,30,2,0,0,0,0,0,60,235,1,0,108,26,2,0,116,30,2,0,132,30,2,0,84,30,2,0,0,0,0,0,0,6,0,0,1,6,0,0,2,6,0,0,3,6,0,0,4,6,0,0,5,6,0,0,6,6,0,0,133,82,9,0,139,82,9,0,145,82,9,0,152,82,9,0,159,82,9,0,171,82,9,0,0,0,0,0,33,82,9,0,35,82,9,0,37,82,9,0,39,82,9,0,41,82,9,0,46,82,9,0,52,82,9,0,58,82,9,0,68,82,9,0,79,82,9,0,92,82,9,0,106,82,9,0,111,82,9,0,118,82,9,0,125,82,9,0,0,0,0,0,80,6,0,0,81,6,0,0,82,6,0,0,83,6,0,0,84,6,0,0,85,6,0,0,0,0,0,0,0,0,0,0,183,11,0,0,4,0,4,0,8,0,4,0,184,11,0,0,4,0,5,0,16,0,8,0,184,11,0,0,4,0,6,0,32,0,32,0,184,11,0,0,4,0,4,0,16,0,16,0,185,11,0,0,8,0,16,0,32,0,32,0,185,11,0,0,8,0,16,0,128,0,128,0,185,11,0,0,8,0,32,0,128,0,0,1,185,11,0,0,32,0,128,0,2,1,0,4,185,11,0,0,32,0,2,1,2,1,0,16,185,11,0,0,8,183,2,0,108,130,0,0,1,1,0,0,30,1,0,0,15,0,0,0,136,187,2,0,232,39,2,0,0,0,0,0,30,0,0,0,15,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,2,0,0,0,2,0,0,0,3,0,0,0,3,0,0,0,4,0,0,0,4,0,0,0,5,0,0,0,5,0,0,0,6,0,0,0,6,0,0,0,7,0,0,0,7,0,0,0,8,0,0,0,8,0,0,0,9,0,0,0,9,0,0,0,10,0,0,0,10,0,0,0,11,0,0,0,11,0,0,0,12,0,0,0,12,0,0,0,13,0,0,0,13,0,0,0,0,0,0,0,116,40,2,0,0,0,0,0,19,0,0,0,7,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,3,0,0,0,7,0,0,0,0,0,0,0,1,0,0,0,2,0,0,0,3,0,0,0,4,0,0,0,5,0,0,0,6,0,0,0,7,0,0,0,8,0,0,0,10,0,0,0,12,0,0,0,14,0,0,0,16,0,0,0,20,0,0,0,24,0,0,0,28,0,0,0,32,0,0,0,40,0,0,0,48,0,0,0,56,0,0,0,64,0,0,0,80,0,0,0,96,0,0,0,112,0,0,0,128,0,0,0,160,0,0,0,192,0,0,0,224,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,2,0,0,0,3,0,0,0,4,0,0,0,6,0,0,0,8,0,0,0,12,0,0,0,16,0,0,0,24,0,0,0,32,0,0,0,48,0,0,0,64,0,0,0,96,0,0,0,128,0,0,0,192,0,0,0,0,1,0,0,128,1,0,0,0,2,0,0,0,3,0,0,0,4,0,0,0,6,0,0,0,8,0,0,0,12,0,0,0,16,0,0,0,24,0,0,0,32,0,0,0,48,0,0,0,64,0,0,0,96,0,0,207,92,9,0,223,92,9,0,162,145,19,0,234,92,9,0,245,92,9,0,2,93,9,0,13,93,9,0,33,93,9,0,46,93,9,0,162,145,19,0,0,0,0,0,150,48,7,119,44,97,14,238,186,81,9,153,25,196,109,7,143,244,106,112,53,165,99,233,163,149,100,158,50,136,219,14,164,184,220,121,30,233,213,224,136,217,210,151,43,76,182,9,189,124,177,126,7,45,184,231,145,29,191,144,100,16,183,29,242,32,176,106,72,113,185,243,222,65,190,132,125,212,218,26,235,228,221,109,81,181,212,244,199,133,211,131,86,152,108,19,192,168,107,100,122,249,98,253,236,201,101,138,79,92,1,20,217,108,6,99,99,61,15,250,245,13,8,141,200,32,110,59,94,16,105,76,228,65,96,213,114,113,103,162,209,228,3,60,71,212,4,75,253,133,13,210,107,181,10,165,250,168,181,53,108,152,178,66,214,201,187,219,64,249,188,172,227,108,216,50,117,92,223,69,207,13,214,220,89,61,209,171,172,48,217,38,58,0,222,81,128,81,215,200,22,97,208,191,181,244,180,33,35,196,179,86,153,149,186,207,15,165,189,184,158,184,2,40,8,136,5,95,178,217,12,198,36,233,11,177,135,124,111,47,17,76,104,88,171,29,97,193,61,45,102,182,144,65,220,118,6,113,219,1,188,32,210,152,42,16,213,239,137,133,177,113,31,181,182,6,165,228,191,159,51,212,184,232,162,201,7,120,52,249,0,15,142,168,9,150,24,152,14,225,187,13,106,127,45,61,109,8,151,108,100,145,1,92,99,230,244,81,107,107,98,97,108,28,216,48,101,133,78,0,98,242,237,149,6,108,123,165,1,27,193,244,8,130,87,196,15,245,198,217,176,101,80,233,183,18,234,184,190,139,124,136,185,252,223,29,221,98,73,45,218,21,243,124,211,140,101,76,212,251,88,97,178,77,206,81,181,58,116,0,188,163,226,48,187,212,65,165,223,74,215,149,216,61,109,196,209,164,251,244,214,211,106,233,105,67,252,217,110,52,70,136,103,173,208,184,96,218,115,45,4,68,229,29,3,51,95,76,10,170,201,124,13,221,60,113,5,80,170,65,2,39,16,16,11,190,134,32,12,201,37,181,104,87,179,133,111,32,9,212,102,185,159,228,97,206,14,249,222,94,152,201,217,41,34,152,208,176,180,168,215,199,23,61,179,89,129,13,180,46,59,92,189,183,173,108,186,192,32,131,184,237,182,179,191,154,12,226,182,3,154,210,177,116,57,71,213,234,175,119,210,157,21,38,219,4,131,22,220,115,18,11,99,227,132,59,100,148,62,106,109,13,168,90,106,122,11,207,14,228,157,255,9,147,39,174,0,10,177,158,7,125,68,147,15,240,210,163,8,135,104,242,1,30,254,194,6,105,93,87,98,247,203,103,101,128,113,54,108,25,231,6,107,110,118,27,212,254,224,43,211,137,90,122,218,16,204,74,221,103,111,223,185,249,249,239,190,142,67,190,183,23,213,142,176,96,232,163,214,214,126,147,209,161,196,194,216,56,82,242,223,79,241,103,187,209,103,87,188,166,221,6,181,63,75,54,178,72,218,43,13,216,76,27,10,175,246,74,3,54,96,122,4,65,195,239,96,223,85,223,103,168,239,142,110,49,121,190,105,70,140,179,97,203,26,131,102,188,160,210,111,37,54,226,104,82,149,119,12,204,3,71,11,187,185,22,2,34,47,38,5,85,190,59,186,197,40,11,189,178,146,90,180,43,4,106,179,92,167,255,215,194,49,207,208,181,139,158,217,44,29,174,222,91,176,194,100,155,38,242,99,236,156,163,106,117,10,147,109,2,169,6,9,156,63,54,14,235,133,103,7,114,19,87,0,5,130,74,191,149,20,122,184,226,174,43,177,123,56,27,182,12,155,142,210,146,13,190,213,229,183,239,220,124,33,223,219,11,212,210,211,134,66,226,212,241,248,179,221,104,110,131,218,31,205,22,190,129,91,38,185,246,225,119,176,111,119,71,183,24,230,90,8,136,112,106,15,255,202,59,6,102,92,11,1,17,255,158,101,143,105,174,98,248,211,255,107,97,69,207,108,22,120,226,10,160,238,210,13,215,84,131,4,78,194,179,3,57,97,38,103,167,247,22,96,208,77,71,105,73,219,119,110,62,74,106,209,174,220,90,214,217,102,11,223,64,240,59,216,55,83,174,188,169,197,158,187,222,127,207,178,71,233,255,181,48,28,242,189,189,138,194,186,202,48,147,179,83,166,163,180,36,5,54,208,186,147,6,215,205,41,87,222,84,191,103,217,35,46,122,102,179,184,74,97,196,2,27,104,93,148,43,111,42,55,190,11,180,161,142,12,195,27,223,5,90,141,239,2,45,0,0,0,0,65,49,27,25,130,98,54,50,195,83,45,43,4,197,108,100,69,244,119,125,134,167,90,86,199,150,65,79,8,138,217,200,73,187,194,209,138,232,239,250,203,217,244,227,12,79,181,172,77,126,174,181,142,45,131,158,207,28,152,135,81,18,194,74,16,35,217,83,211,112,244,120,146,65,239,97,85,215,174,46,20,230,181,55,215,181,152,28,150,132,131,5,89,152,27,130,24,169,0,155,219,250,45,176,154,203,54,169,93,93,119,230,28,108,108,255,223,63,65,212,158,14,90,205,162,36,132,149,227,21,159,140,32,70,178,167,97,119,169,190,166,225,232,241,231,208,243,232,36,131,222,195,101,178,197,218,170,174,93,93,235,159,70,68,40,204,107,111,105,253,112,118,174,107,49,57,239,90,42,32,44,9,7,11,109,56,28,18,243,54,70,223,178,7,93,198,113,84,112,237,48,101,107,244,247,243,42,187,182,194,49,162,117,145,28,137,52,160,7,144,251,188,159,23,186,141,132,14,121,222,169,37,56,239,178,60,255,121,243,115,190,72,232,106,125,27,197,65,60,42,222,88,5,79,121,240,68,126,98,233,135,45,79,194,198,28,84,219,1,138,21,148,64,187,14,141,131,232,35,166,194,217,56,191,13,197,160,56,76,244,187,33,143,167,150,10,206,150,141,19,9,0,204,92,72,49,215,69,139,98,250,110,202,83,225,119,84,93,187,186,21,108,160,163,214,63,141,136,151,14,150,145,80,152,215,222,17,169,204,199,210,250,225,236,147,203,250,245,92,215,98,114,29,230,121,107,222,181,84,64,159,132,79,89,88,18,14,22,25,35,21,15,218,112,56,36,155,65,35,61,167,107,253,101,230,90,230,124,37,9,203,87,100,56,208,78,163,174,145,1,226,159,138,24,33,204,167,51,96,253,188,42,175,225,36,173,238,208,63,180,45,131,18,159,108,178,9,134,171,36,72,201,234,21,83,208,41,70,126,251,104,119,101,226,246,121,63,47,183,72,36,54,116,27,9,29,53,42,18,4,242,188,83,75,179,141,72,82,112,222,101,121,49,239,126,96,254,243,230,231,191,194,253,254,124,145,208,213,61,160,203,204,250,54,138,131,187,7,145,154,120,84,188,177,57,101,167,168,75,152,131,59,10,169,152,34,201,250,181,9,136,203,174,16,79,93,239,95,14,108,244,70,205,63,217,109,140,14,194,116,67,18,90,243,2,35,65,234,193,112,108,193,128,65,119,216,71,215,54,151,6,230,45,142,197,181,0,165,132,132,27,188,26,138,65,113,91,187,90,104,152,232,119,67,217,217,108,90,30,79,45,21,95,126,54,12,156,45,27,39,221,28,0,62,18,0,152,185,83,49,131,160,144,98,174,139,209,83,181,146,22,197,244,221,87,244,239,196,148,167,194,239,213,150,217,246,233,188,7,174,168,141,28,183,107,222,49,156,42,239,42,133,237,121,107,202,172,72,112,211,111,27,93,248,46,42,70,225,225,54,222,102,160,7,197,127,99,84,232,84,34,101,243,77,229,243,178,2,164,194,169,27,103,145,132,48,38,160,159,41,184,174,197,228,249,159,222,253,58,204,243,214,123,253,232,207,188,107,169,128,253,90,178,153,62,9,159,178,127,56,132,171,176,36,28,44,241,21,7,53,50,70,42,30,115,119,49,7,180,225,112,72,245,208,107,81,54,131,70,122,119,178,93,99,78,215,250,203,15,230,225,210,204,181,204,249,141,132,215,224,74,18,150,175,11,35,141,182,200,112,160,157,137,65,187,132,70,93,35,3,7,108,56,26,196,63,21,49,133,14,14,40,66,152,79,103,3,169,84,126,192,250,121,85,129,203,98,76,31,197,56,129,94,244,35,152,157,167,14,179,220,150,21,170,27,0,84,229,90,49,79,252,153,98,98,215,216,83,121,206,23,79,225,73,86,126,250,80,149,45,215,123,212,28,204,98,19,138,141,45,82,187,150,52,145,232,187,31,208,217,160,6,236,243,126,94,173,194,101,71,110,145,72,108,47,160,83,117,232,54,18,58,169,7,9,35,106,84,36,8,43,101,63,17,228,121,167,150,165,72,188,143,102,27,145,164,39,42,138,189,224,188,203,242,161,141,208,235,98,222,253,192,35,239,230,217,189,225,188,20,252,208,167,13,63,131,138,38,126,178,145,63,185,36,208,112,248,21,203,105,59,70,230,66,122,119,253,91,181,107,101,220,244,90,126,197,55,9,83,238,118,56,72,247,177,174,9,184,240,159,18,161,51,204,63,138,114,253,36,147,0,0,0,0,55,106,194,1,110,212,132,3,89,190,70,2,220,168,9,7,235,194,203,6,178,124,141,4,133,22,79,5,184,81,19,14,143,59,209,15,214,133,151,13,225,239,85,12,100,249,26,9,83,147,216,8,10,45,158,10,61,71,92,11,112,163,38,28,71,201,228,29,30,119,162,31,41,29,96,30,172,11,47,27,155,97,237,26,194,223,171,24,245,181,105,25,200,242,53,18,255,152,247,19,166,38,177,17,145,76,115,16,20,90,60,21,35,48,254,20,122,142,184,22,77,228,122,23,224,70,77,56,215,44,143,57,142,146,201,59,185,248,11,58,60,238,68,63,11,132,134,62,82,58,192,60,101,80,2,61,88,23,94,54,111,125,156,55,54,195,218,53,1,169,24,52,132,191,87,49,179,213,149,48,234,107,211,50,221,1,17,51,144,229,107,36,167,143,169,37,254,49,239,39,201,91,45,38,76,77,98,35,123,39,160,34,34,153,230,32,21,243,36,33,40,180,120,42,31,222,186,43,70,96,252,41,113,10,62,40,244,28,113,45,195,118,179,44,154,200,245,46,173,162,55,47,192,141,154,112,247,231,88,113,174,89,30,115,153,51,220,114,28,37,147,119,43,79,81,118,114,241,23,116,69,155,213,117,120,220,137,126,79,182,75,127,22,8,13,125,33,98,207,124,164,116,128,121,147,30,66,120,202,160,4,122,253,202,198,123,176,46,188,108,135,68,126,109,222,250,56,111,233,144,250,110,108,134,181,107,91,236,119,106,2,82,49,104,53,56,243,105,8,127,175,98,63,21,109,99,102,171,43,97,81,193,233,96,212,215,166,101,227,189,100,100,186,3,34,102,141,105,224,103,32,203,215,72,23,161,21,73,78,31,83,75,121,117,145,74,252,99,222,79,203,9,28,78,146,183,90,76,165,221,152,77,152,154,196,70,175,240,6,71,246,78,64,69,193,36,130,68,68,50,205,65,115,88,15,64,42,230,73,66,29,140,139,67,80,104,241,84,103,2,51,85,62,188,117,87,9,214,183,86,140,192,248,83,187,170,58,82,226,20,124,80,213,126,190,81,232,57,226,90,223,83,32,91,134,237,102,89,177,135,164,88,52,145,235,93,3,251,41,92,90,69,111,94,109,47,173,95,128,27,53,225,183,113,247,224,238,207,177,226,217,165,115,227,92,179,60,230,107,217,254,231,50,103,184,229,5,13,122,228,56,74,38,239,15,32,228,238,86,158,162,236,97,244,96,237,228,226,47,232,211,136,237,233,138,54,171,235,189,92,105,234,240,184,19,253,199,210,209,252,158,108,151,254,169,6,85,255,44,16,26,250,27,122,216,251,66,196,158,249,117,174,92,248,72,233,0,243,127,131,194,242,38,61,132,240,17,87,70,241,148,65,9,244,163,43,203,245,250,149,141,247,205,255,79,246,96,93,120,217,87,55,186,216,14,137,252,218,57,227,62,219,188,245,113,222,139,159,179,223,210,33,245,221,229,75,55,220,216,12,107,215,239,102,169,214,182,216,239,212,129,178,45,213,4,164,98,208,51,206,160,209,106,112,230,211,93,26,36,210,16,254,94,197,39,148,156,196,126,42,218,198,73,64,24,199,204,86,87,194,251,60,149,195,162,130,211,193,149,232,17,192,168,175,77,203,159,197,143,202,198,123,201,200,241,17,11,201,116,7,68,204,67,109,134,205,26,211,192,207,45,185,2,206,64,150,175,145,119,252,109,144,46,66,43,146,25,40,233,147,156,62,166,150,171,84,100,151,242,234,34,149,197,128,224,148,248,199,188,159,207,173,126,158,150,19,56,156,161,121,250,157,36,111,181,152,19,5,119,153,74,187,49,155,125,209,243,154,48,53,137,141,7,95,75,140,94,225,13,142,105,139,207,143,236,157,128,138,219,247,66,139,130,73,4,137,181,35,198,136,136,100,154,131,191,14,88,130,230,176,30,128,209,218,220,129,84,204,147,132,99,166,81,133,58,24,23,135,13,114,213,134,160,208,226,169,151,186,32,168,206,4,102,170,249,110,164,171,124,120,235,174,75,18,41,175,18,172,111,173,37,198,173,172,24,129,241,167,47,235,51,166,118,85,117,164,65,63,183,165,196,41,248,160,243,67,58,161,170,253,124,163,157,151,190,162,208,115,196,181,231,25,6,180,190,167,64,182,137,205,130,183,12,219,205,178,59,177,15,179,98,15,73,177,85,101,139,176,104,34,215,187,95,72,21,186,6,246,83,184,49,156,145,185,180,138,222,188,131,224,28,189,218,94,90,191,237,52,152,190,0,0,0,0,101,103,188,184,139,200,9,170,238,175,181,18,87,151,98,143,50,240,222,55,220,95,107,37,185,56,215,157,239,40,180,197,138,79,8,125,100,224,189,111,1,135,1,215,184,191,214,74,221,216,106,242,51,119,223,224,86,16,99,88,159,87,25,80,250,48,165,232,20,159,16,250,113,248,172,66,200,192,123,223,173,167,199,103,67,8,114,117,38,111,206,205,112,127,173,149,21,24,17,45,251,183,164,63,158,208,24,135,39,232,207,26,66,143,115,162,172,32,198,176,201,71,122,8,62,175,50,160,91,200,142,24,181,103,59,10,208,0,135,178,105,56,80,47,12,95,236,151,226,240,89,133,135,151,229,61,209,135,134,101,180,224,58,221,90,79,143,207,63,40,51,119,134,16,228,234,227,119,88,82,13,216,237,64,104,191,81,248,161,248,43,240,196,159,151,72,42,48,34,90,79,87,158,226,246,111,73,127,147,8,245,199,125,167,64,213,24,192,252,109,78,208,159,53,43,183,35,141,197,24,150,159,160,127,42,39,25,71,253,186,124,32,65,2,146,143,244,16,247,232,72,168,61,88,20,155,88,63,168,35,182,144,29,49,211,247,161,137,106,207,118,20,15,168,202,172,225,7,127,190,132,96,195,6,210,112,160,94,183,23,28,230,89,184,169,244,60,223,21,76,133,231,194,209,224,128,126,105,14,47,203,123,107,72,119,195,162,15,13,203,199,104,177,115,41,199,4,97,76,160,184,217,245,152,111,68,144,255,211,252,126,80,102,238,27,55,218,86,77,39,185,14,40,64,5,182,198,239,176,164,163,136,12,28,26,176,219,129,127,215,103,57,145,120,210,43,244,31,110,147,3,247,38,59,102,144,154,131,136,63,47,145,237,88,147,41,84,96,68,180,49,7,248,12,223,168,77,30,186,207,241,166,236,223,146,254,137,184,46,70,103,23,155,84,2,112,39,236,187,72,240,113,222,47,76,201,48,128,249,219,85,231,69,99,156,160,63,107,249,199,131,211,23,104,54,193,114,15,138,121,203,55,93,228,174,80,225,92,64,255,84,78,37,152,232,246,115,136,139,174,22,239,55,22,248,64,130,4,157,39,62,188,36,31,233,33,65,120,85,153,175,215,224,139,202,176,92,51,59,182,89,237,94,209,229,85,176,126,80,71,213,25,236,255,108,33,59,98,9,70,135,218,231,233,50,200,130,142,142,112,212,158,237,40,177,249,81,144,95,86,228,130,58,49,88,58,131,9,143,167,230,110,51,31,8,193,134,13,109,166,58,181,164,225,64,189,193,134,252,5,47,41,73,23,74,78,245,175,243,118,34,50,150,17,158,138,120,190,43,152,29,217,151,32,75,201,244,120,46,174,72,192,192,1,253,210,165,102,65,106,28,94,150,247,121,57,42,79,151,150,159,93,242,241,35,229,5,25,107,77,96,126,215,245,142,209,98,231,235,182,222,95,82,142,9,194,55,233,181,122,217,70,0,104,188,33,188,208,234,49,223,136,143,86,99,48,97,249,214,34,4,158,106,154,189,166,189,7,216,193,1,191,54,110,180,173,83,9,8,21,154,78,114,29,255,41,206,165,17,134,123,183,116,225,199,15,205,217,16,146,168,190,172,42,70,17,25,56,35,118,165,128,117,102,198,216,16,1,122,96,254,174,207,114,155,201,115,202,34,241,164,87,71,150,24,239,169,57,173,253,204,94,17,69,6,238,77,118,99,137,241,206,141,38,68,220,232,65,248,100,81,121,47,249,52,30,147,65,218,177,38,83,191,214,154,235,233,198,249,179,140,161,69,11,98,14,240,25,7,105,76,161,190,81,155,60,219,54,39,132,53,153,146,150,80,254,46,46,153,185,84,38,252,222,232,158,18,113,93,140,119,22,225,52,206,46,54,169,171,73,138,17,69,230,63,3,32,129,131,187,118,145,224,227,19,246,92,91,253,89,233,73,152,62,85,241,33,6,130,108,68,97,62,212,170,206,139,198,207,169,55,126,56,65,127,214,93,38,195,110,179,137,118,124,214,238,202,196,111,214,29,89,10,177,161,225,228,30,20,243,129,121,168,75,215,105,203,19,178,14,119,171,92,161,194,185,57,198,126,1,128,254,169,156,229,153,21,36,11,54,160,54,110,81,28,142,167,22,102,134,194,113,218,62,44,222,111,44,73,185,211,148,240,129,4,9,149,230,184,177,123,73,13,163,30,46,177,27,72,62,210,67,45,89,110,251,195,246,219,233,166,145,103,81,31,169,176,204,122,206,12,116,148,97,185,102,241,6,5,222,0,0,0,0,119,7,48,150,238,14,97,44,153,9,81,186,7,109,196,25,112,106,244,143,233,99,165,53,158,100,149,163,14,219,136,50,121,220,184,164,224,213,233,30,151,210,217,136,9,182,76,43,126,177,124,189,231,184,45,7,144,191,29,145,29,183,16,100,106,176,32,242,243,185,113,72,132,190,65,222,26,218,212,125,109,221,228,235,244,212,181,81,131,211,133,199,19,108,152,86,100,107,168,192,253,98,249,122,138,101,201,236,20,1,92,79,99,6,108,217,250,15,61,99,141,8,13,245,59,110,32,200,76,105,16,94,213,96,65,228,162,103,113,114,60,3,228,209,75,4,212,71,210,13,133,253,165,10,181,107,53,181,168,250,66,178,152,108,219,187,201,214,172,188,249,64,50,216,108,227,69,223,92,117,220,214,13,207,171,209,61,89,38,217,48,172,81,222,0,58,200,215,81,128,191,208,97,22,33,180,244,181,86,179,196,35,207,186,149,153,184,189,165,15,40,2,184,158,95,5,136,8,198,12,217,178,177,11,233,36,47,111,124,135,88,104,76,17,193,97,29,171,182,102,45,61,118,220,65,144,1,219,113,6,152,210,32,188,239,213,16,42,113,177,133,137,6,182,181,31,159,191,228,165,232,184,212,51,120,7,201,162,15,0,249,52,150,9,168,142,225,14,152,24,127,106,13,187,8,109,61,45,145,100,108,151,230,99,92,1,107,107,81,244,28,108,97,98,133,101,48,216,242,98,0,78,108,6,149,237,27,1,165,123,130,8,244,193,245,15,196,87,101,176,217,198,18,183,233,80,139,190,184,234,252,185,136,124,98,221,29,223,21,218,45,73,140,211,124,243,251,212,76,101,77,178,97,88,58,181,81,206,163,188,0,116,212,187,48,226,74,223,165,65,61,216,149,215,164,209,196,109,211,214,244,251,67,105,233,106,52,110,217,252,173,103,136,70,218,96,184,208,68,4,45,115,51,3,29,229,170,10,76,95,221,13,124,201,80,5,113,60,39,2,65,170,190,11,16,16,201,12,32,134,87,104,181,37,32,111,133,179,185,102,212,9,206,97,228,159,94,222,249,14,41,217,201,152,176,208,152,34,199,215,168,180,89,179,61,23,46,180,13,129,183,189,92,59,192,186,108,173,237,184,131,32,154,191,179,182,3,182,226,12,116,177,210,154,234,213,71,57,157,210,119,175,4,219,38,21,115,220,22,131,227,99,11,18,148,100,59,132,13,109,106,62,122,106,90,168,228,14,207,11,147,9,255,157,10,0,174,39,125,7,158,177,240,15,147,68,135,8,163,210,30,1,242,104,105,6,194,254,247,98,87,93,128,101,103,203,25,108,54,113,110,107,6,231,254,212,27,118,137,211,43,224,16,218,122,90,103,221,74,204,249,185,223,111,142,190,239,249,23,183,190,67,96,176,142,213,214,214,163,232,161,209,147,126,56,216,194,196,79,223,242,82,209,187,103,241,166,188,87,103,63,181,6,221,72,178,54,75,216,13,43,218,175,10,27,76,54,3,74,246,65,4,122,96,223,96,239,195,168,103,223,85,49,110,142,239,70,105,190,121,203,97,179,140,188,102,131,26,37,111,210,160,82,104,226,54,204,12,119,149,187,11,71,3,34,2,22,185,85,5,38,47,197,186,59,190,178,189,11,40,43,180,90,146,92,179,106,4,194,215,255,167,181,208,207,49,44,217,158,139,91,222,174,29,155,100,194,176,236,99,242,38,117,106,163,156,2,109,147,10,156,9,6,169,235,14,54,63,114,7,103,133,5,0,87,19,149,191,74,130,226,184,122,20,123,177,43,174,12,182,27,56,146,210,142,155,229,213,190,13,124,220,239,183,11,219,223,33,134,211,210,212,241,212,226,66,104,221,179,248,31,218,131,110,129,190,22,205,246,185,38,91,111,176,119,225,24,183,71,119,136,8,90,230,255,15,106,112,102,6,59,202,17,1,11,92,143,101,158,255,248,98,174,105,97,107,255,211,22,108,207,69,160,10,226,120,215,13,210,238,78,4,131,84,57,3,179,194,167,103,38,97,208,96,22,247,73,105,71,77,62,110,119,219,174,209,106,74,217,214,90,220,64,223,11,102,55,216,59,240,169,188,174,83,222,187,158,197,71,178,207,127,48,181,255,233,189,189,242,28,202,186,194,138,83,179,147,48,36,180,163,166,186,208,54,5,205,215,6,147,84,222,87,41,35,217,103,191,179,102,122,46,196,97,74,184,93,104,27,2,42,111,43,148,180,11,190,55,195,12,142,161,90,5,223,27,45,2,239,141,0,0,0,0,25,27,49,65,50,54,98,130,43,45,83,195,100,108,197,4,125,119,244,69,86,90,167,134,79,65,150,199,200,217,138,8,209,194,187,73,250,239,232,138,227,244,217,203,172,181,79,12,181,174,126,77,158,131,45,142,135,152,28,207,74,194,18,81,83,217,35,16,120,244,112,211,97,239,65,146,46,174,215,85,55,181,230,20,28,152,181,215,5,131,132,150,130,27,152,89,155,0,169,24,176,45,250,219,169,54,203,154,230,119,93,93,255,108,108,28,212,65,63,223,205,90,14,158,149,132,36,162,140,159,21,227,167,178,70,32,190,169,119,97,241,232,225,166,232,243,208,231,195,222,131,36,218,197,178,101,93,93,174,170,68,70,159,235,111,107,204,40,118,112,253,105,57,49,107,174,32,42,90,239,11,7,9,44,18,28,56,109,223,70,54,243,198,93,7,178,237,112,84,113,244,107,101,48,187,42,243,247,162,49,194,182,137,28,145,117,144,7,160,52,23,159,188,251,14,132,141,186,37,169,222,121,60,178,239,56,115,243,121,255,106,232,72,190,65,197,27,125,88,222,42,60,240,121,79,5,233,98,126,68,194,79,45,135,219,84,28,198,148,21,138,1,141,14,187,64,166,35,232,131,191,56,217,194,56,160,197,13,33,187,244,76,10,150,167,143,19,141,150,206,92,204,0,9,69,215,49,72,110,250,98,139,119,225,83,202,186,187,93,84,163,160,108,21,136,141,63,214,145,150,14,151,222,215,152,80,199,204,169,17,236,225,250,210,245,250,203,147,114,98,215,92,107,121,230,29,64,84,181,222,89,79,132,159,22,14,18,88,15,21,35,25,36,56,112,218,61,35,65,155,101,253,107,167,124,230,90,230,87,203,9,37,78,208,56,100,1,145,174,163,24,138,159,226,51,167,204,33,42,188,253,96,173,36,225,175,180,63,208,238,159,18,131,45,134,9,178,108,201,72,36,171,208,83,21,234,251,126,70,41,226,101,119,104,47,63,121,246,54,36,72,183,29,9,27,116,4,18,42,53,75,83,188,242,82,72,141,179,121,101,222,112,96,126,239,49,231,230,243,254,254,253,194,191,213,208,145,124,204,203,160,61,131,138,54,250,154,145,7,187,177,188,84,120,168,167,101,57,59,131,152,75,34,152,169,10,9,181,250,201,16,174,203,136,95,239,93,79,70,244,108,14,109,217,63,205,116,194,14,140,243,90,18,67,234,65,35,2,193,108,112,193,216,119,65,128,151,54,215,71,142,45,230,6,165,0,181,197,188,27,132,132,113,65,138,26,104,90,187,91,67,119,232,152,90,108,217,217,21,45,79,30,12,54,126,95,39,27,45,156,62,0,28,221,185,152,0,18,160,131,49,83,139,174,98,144,146,181,83,209,221,244,197,22,196,239,244,87,239,194,167,148,246,217,150,213,174,7,188,233,183,28,141,168,156,49,222,107,133,42,239,42,202,107,121,237,211,112,72,172,248,93,27,111,225,70,42,46,102,222,54,225,127,197,7,160,84,232,84,99,77,243,101,34,2,178,243,229,27,169,194,164,48,132,145,103,41,159,160,38,228,197,174,184,253,222,159,249,214,243,204,58,207,232,253,123,128,169,107,188,153,178,90,253,178,159,9,62,171,132,56,127,44,28,36,176,53,7,21,241,30,42,70,50,7,49,119,115,72,112,225,180,81,107,208,245,122,70,131,54,99,93,178,119,203,250,215,78,210,225,230,15,249,204,181,204,224,215,132,141,175,150,18,74,182,141,35,11,157,160,112,200,132,187,65,137,3,35,93,70,26,56,108,7,49,21,63,196,40,14,14,133,103,79,152,66,126,84,169,3,85,121,250,192,76,98,203,129,129,56,197,31,152,35,244,94,179,14,167,157,170,21,150,220,229,84,0,27,252,79,49,90,215,98,98,153,206,121,83,216,73,225,79,23,80,250,126,86,123,215,45,149,98,204,28,212,45,141,138,19,52,150,187,82,31,187,232,145,6,160,217,208,94,126,243,236,71,101,194,173,108,72,145,110,117,83,160,47,58,18,54,232,35,9,7,169,8,36,84,106,17,63,101,43,150,167,121,228,143,188,72,165,164,145,27,102,189,138,42,39,242,203,188,224,235,208,141,161,192,253,222,98,217,230,239,35,20,188,225,189,13,167,208,252,38,138,131,63,63,145,178,126,112,208,36,185,105,203,21,248,66,230,70,59,91,253,119,122,220,101,107,181,197,126,90,244,238,83,9,55,247,72,56,118,184,9,174,177,161,18,159,240,138,63,204,51,147,36,253,114,0,0,0,0,1,194,106,55,3,132,212,110,2,70,190,89,7,9,168,220,6,203,194,235,4,141,124,178,5,79,22,133,14,19,81,184,15,209,59,143,13,151,133,214,12,85,239,225,9,26,249,100,8,216,147,83,10,158,45,10,11,92,71,61,28,38,163,112,29,228,201,71,31,162,119,30,30,96,29,41,27,47,11,172,26,237,97,155,24,171,223,194,25,105,181,245,18,53,242,200,19,247,152,255,17,177,38,166,16,115,76,145,21,60,90,20,20,254,48,35,22,184,142,122,23,122,228,77,56,77,70,224,57,143,44,215,59,201,146,142,58,11,248,185,63,68,238,60,62,134,132,11,60,192,58,82,61,2,80,101,54,94,23,88,55,156,125,111,53,218,195,54,52,24,169,1,49,87,191,132,48,149,213,179,50,211,107,234,51,17,1,221,36,107,229,144,37,169,143,167,39,239,49,254,38,45,91,201,35,98,77,76,34,160,39,123,32,230,153,34,33,36,243,21,42,120,180,40,43,186,222,31,41,252,96,70,40,62,10,113,45,113,28,244,44,179,118,195,46,245,200,154,47,55,162,173,112,154,141,192,113,88,231,247,115,30,89,174,114,220,51,153,119,147,37,28,118,81,79,43,116,23,241,114,117,213,155,69,126,137,220,120,127,75,182,79,125,13,8,22,124,207,98,33,121,128,116,164,120,66,30,147,122,4,160,202,123,198,202,253,108,188,46,176,109,126,68,135,111,56,250,222,110,250,144,233,107,181,134,108,106,119,236,91,104,49,82,2,105,243,56,53,98,175,127,8,99,109,21,63,97,43,171,102,96,233,193,81,101,166,215,212,100,100,189,227,102,34,3,186,103,224,105,141,72,215,203,32,73,21,161,23,75,83,31,78,74,145,117,121,79,222,99,252,78,28,9,203,76,90,183,146,77,152,221,165,70,196,154,152,71,6,240,175,69,64,78,246,68,130,36,193,65,205,50,68,64,15,88,115,66,73,230,42,67,139,140,29,84,241,104,80,85,51,2,103,87,117,188,62,86,183,214,9,83,248,192,140,82,58,170,187,80,124,20,226,81,190,126,213,90,226,57,232,91,32,83,223,89,102,237,134,88,164,135,177,93,235,145,52,92,41,251,3,94,111,69,90,95,173,47,109,225,53,27,128,224,247,113,183,226,177,207,238,227,115,165,217,230,60,179,92,231,254,217,107,229,184,103,50,228,122,13,5,239,38,74,56,238,228,32,15,236,162,158,86,237,96,244,97,232,47,226,228,233,237,136,211,235,171,54,138,234,105,92,189,253,19,184,240,252,209,210,199,254,151,108,158,255,85,6,169,250,26,16,44,251,216,122,27,249,158,196,66,248,92,174,117,243,0,233,72,242,194,131,127,240,132,61,38,241,70,87,17,244,9,65,148,245,203,43,163,247,141,149,250,246,79,255,205,217,120,93,96,216,186,55,87,218,252,137,14,219,62,227,57,222,113,245,188,223,179,159,139,221,245,33,210,220,55,75,229,215,107,12,216,214,169,102,239,212,239,216,182,213,45,178,129,208,98,164,4,209,160,206,51,211,230,112,106,210,36,26,93,197,94,254,16,196,156,148,39,198,218,42,126,199,24,64,73,194,87,86,204,195,149,60,251,193,211,130,162,192,17,232,149,203,77,175,168,202,143,197,159,200,201,123,198,201,11,17,241,204,68,7,116,205,134,109,67,207,192,211,26,206,2,185,45,145,175,150,64,144,109,252,119,146,43,66,46,147,233,40,25,150,166,62,156,151,100,84,171,149,34,234,242,148,224,128,197,159,188,199,248,158,126,173,207,156,56,19,150,157,250,121,161,152,181,111,36,153,119,5,19,155,49,187,74,154,243,209,125,141,137,53,48,140,75,95,7,142,13,225,94,143,207,139,105,138,128,157,236,139,66,247,219,137,4,73,130,136,198,35,181,131,154,100,136,130,88,14,191,128,30,176,230,129,220,218,209,132,147,204,84,133,81,166,99,135,23,24,58,134,213,114,13,169,226,208,160,168,32,186,151,170,102,4,206,171,164,110,249,174,235,120,124,175,41,18,75,173,111,172,18,172,173,198,37,167,241,129,24,166,51,235,47,164,117,85,118,165,183,63,65,160,248,41,196,161,58,67,243,163,124,253,170,162,190,151,157,181,196,115,208,180,6,25,231,182,64,167,190,183,130,205,137,178,205,219,12,179,15,177,59,177,73,15,98,176,139,101,85,187,215,34,104,186,21,72,95,184,83,246,6,185,145,156,49,188,222,138,180,189,28,224,131,191,90,94,218,190,152,52,237,0,0,0,0,184,188,103,101,170,9,200,139,18,181,175,238,143,98,151,87,55,222,240,50,37,107,95,220,157,215,56,185,197,180,40,239,125,8,79,138,111,189,224,100,215,1,135,1,74,214,191,184,242,106,216,221,224,223,119,51,88,99,16,86,80,25,87,159,232,165,48,250,250,16,159,20,66,172,248,113,223,123,192,200,103,199,167,173,117,114,8,67,205,206,111,38,149,173,127,112,45,17,24,21,63,164,183,251,135,24,208,158,26,207,232,39,162,115,143,66,176,198,32,172,8,122,71,201,160,50,175,62,24,142,200,91,10,59,103,181,178,135,0,208,47,80,56,105,151,236,95,12,133,89,240,226,61,229,151,135,101,134,135,209,221,58,224,180,207,143,79,90,119,51,40,63,234,228,16,134,82,88,119,227,64,237,216,13,248,81,191,104,240,43,248,161,72,151,159,196,90,34,48,42,226,158,87,79,127,73,111,246,199,245,8,147,213,64,167,125,109,252,192,24,53,159,208,78,141,35,183,43,159,150,24,197,39,42,127,160,186,253,71,25,2,65,32,124,16,244,143,146,168,72,232,247,155,20,88,61,35,168,63,88,49,29,144,182,137,161,247,211,20,118,207,106,172,202,168,15,190,127,7,225,6,195,96,132,94,160,112,210,230,28,23,183,244,169,184,89,76,21,223,60,209,194,231,133,105,126,128,224,123,203,47,14,195,119,72,107,203,13,15,162,115,177,104,199,97,4,199,41,217,184,160,76,68,111,152,245,252,211,255,144,238,102,80,126,86,218,55,27,14,185,39,77,182,5,64,40,164,176,239,198,28,12,136,163,129,219,176,26,57,103,215,127,43,210,120,145,147,110,31,244,59,38,247,3,131,154,144,102,145,47,63,136,41,147,88,237,180,68,96,84,12,248,7,49,30,77,168,223,166,241,207,186,254,146,223,236,70,46,184,137,84,155,23,103,236,39,112,2,113,240,72,187,201,76,47,222,219,249,128,48,99,69,231,85,107,63,160,156,211,131,199,249,193,54,104,23,121,138,15,114,228,93,55,203,92,225,80,174,78,84,255,64,246,232,152,37,174,139,136,115,22,55,239,22,4,130,64,248,188,62,39,157,33,233,31,36,153,85,120,65,139,224,215,175,51,92,176,202,237,89,182,59,85,229,209,94,71,80,126,176,255,236,25,213,98,59,33,108,218,135,70,9,200,50,233,231,112,142,142,130,40,237,158,212,144,81,249,177,130,228,86,95,58,88,49,58,167,143,9,131,31,51,110,230,13,134,193,8,181,58,166,109,189,64,225,164,5,252,134,193,23,73,41,47,175,245,78,74,50,34,118,243,138,158,17,150,152,43,190,120,32,151,217,29,120,244,201,75,192,72,174,46,210,253,1,192,106,65,102,165,247,150,94,28,79,42,57,121,93,159,150,151,229,35,241,242,77,107,25,5,245,215,126,96,231,98,209,142,95,222,182,235,194,9,142,82,122,181,233,55,104,0,70,217,208,188,33,188,136,223,49,234,48,99,86,143,34,214,249,97,154,106,158,4,7,189,166,189,191,1,193,216,173,180,110,54,21,8,9,83,29,114,78,154,165,206,41,255,183,123,134,17,15,199,225,116,146,16,217,205,42,172,190,168,56,25,17,70,128,165,118,35,216,198,102,117,96,122,1,16,114,207,174,254,202,115,201,155,87,164,241,34,239,24,150,71,253,173,57,169,69,17,94,204,118,77,238,6,206,241,137,99,220,68,38,141,100,248,65,232,249,47,121,81,65,147,30,52,83,38,177,218,235,154,214,191,179,249,198,233,11,69,161,140,25,240,14,98,161,76,105,7,60,155,81,190,132,39,54,219,150,146,153,53,46,46,254,80,38,84,185,153,158,232,222,252,140,93,113,18,52,225,22,119,169,54,46,206,17,138,73,171,3,63,230,69,187,131,129,32,227,224,145,118,91,92,246,19,73,233,89,253,241,85,62,152,108,130,6,33,212,62,97,68,198,139,206,170,126,55,169,207,214,127,65,56,110,195,38,93,124,118,137,179,196,202,238,214,89,29,214,111,225,161,177,10,243,20,30,228,75,168,121,129,19,203,105,215,171,119,14,178,185,194,161,92,1,126,198,57,156,169,254,128,36,21,153,229,54,160,54,11,142,28,81,110,134,102,22,167,62,218,113,194,44,111,222,44,148,211,185,73,9,4,129,240,177,184,230,149,163,13,73,123,27,177,46,30,67,210,62,72,251,110,89,45,233,219,246,195,81,103,145,166,204,176,169,31,116,12,206,122,102,185,97,148,222,5,6,241,2,0,0,192,3,0,0,192,4,0,0,192,5,0,0,192,6,0,0,192,7,0,0,192,8,0,0,192,9,0,0,192,10,0,0,192,11,0,0,192,12,0,0,192,13,0,0,192,14,0,0,192,15,0,0,192,16,0,0,192,17,0,0,192,18,0,0,192,19,0,0,192,20,0,0,192,21,0,0,192,22,0,0,192,23,0,0,192,24,0,0,192,25,0,0,192,26,0,0,192,27,0,0,192,28,0,0,192,29,0,0,192,30,0,0,192,31,0,0,192,0,0,0,179,1,0,0,195,2,0,0,195,3,0,0,195,4,0,0,195,5,0,0,195,6,0,0,195,7,0,0,195,8,0,0,195,9,0,0,195,10,0,0,195,11,0,0,195,12,0,0,195,13,0,0,211,14,0,0,195,15,0,0,195,0,0,12,187,1,0,12,195,2,0,12,195,3,0,12,195,4,0,12,211,164,74,2,0,5,0,0,0,0,0,0,0,0,0,0,0,186,11,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,187,11,0,0,188,11,0,0,109,137,19,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,255,255,255,255,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,24,75,2,0,9,0,0,0,0,0,0,0,0,0,0,0,186,11,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,189,11,0,0,0,0,0,0,188,11,0,0,117,137,19,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,255,255,255,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,140,75,2,0,5,0,0,0,0,0,0,0,0,0,0,0,186,11,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,190,11,0,0,188,11,0,0,125,141,19,0,0,4,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,10,255,255,255,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,140,75,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,191,11,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,255,255,255,255,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,189,2,0,120,78,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,2,0,0,0,3,0,0,0,4,0,0,0,5,0,0,0,6,0,0,0,7,0,0,0,8,0,0,0,9,0,0,0,10,0,0,0,11,0,0,0,12,0,0,0,13,0,0,0,14,0,0,0,15,0,0,0,16,0,0,0,17,0,0,0,18,0,0,0,19,0,0,0,20,0,0,0,21,0,0,0,22,0,0,0,23,0,0,0,24,0,0,0,25,0,0,0,26,0,0,0,27,0,0,0,28,0,0,0,29,0,0,0,30,0,0,0,31,0,0,0,32,0,0,0,33,0,0,0,34,0,0,0,35,0,0,0,36,0,0,0,37,0,0,0,38,0,0,0,39,0,0,0,40,0,0,0,41,0,0,0,42,0,0,0,43,0,0,0,44,0,0,0,45,0,0,0,46,0,0,0,47,0,0,0,48,0,0,0,49,0,0,0,50,0,0,0,51,0,0,0,52,0,0,0,53,0,0,0,54,0,0,0,55,0,0,0,56,0,0,0,57,0,0,0,58,0,0,0,59,0,0,0,60,0,0,0,61,0,0,0,62,0,0,0,63,0,0,0,64,0,0,0,97,0,0,0,98,0,0,0,99,0,0,0,100,0,0,0,101,0,0,0,102,0,0,0,103,0,0,0,104,0,0,0,105,0,0,0,106,0,0,0,107,0,0,0,108,0,0,0,109,0,0,0,110,0,0,0,111,0,0,0,112,0,0,0,113,0,0,0,114,0,0,0,115,0,0,0,116,0,0,0,117,0,0,0,118,0,0,0,119,0,0,0,120,0,0,0,121,0,0,0,122,0,0,0,91,0,0,0,92,0,0,0,93,0,0,0,94,0,0,0,95,0,0,0,96,0,0,0,97,0,0,0,98,0,0,0,99,0,0,0,100,0,0,0,101,0,0,0,102,0,0,0,103,0,0,0,104,0,0,0,105,0,0,0,106,0,0,0,107,0,0,0,108,0,0,0,109,0,0,0,110,0,0,0,111,0,0,0,112,0,0,0,113,0,0,0,114,0,0,0,115,0,0,0,116,0,0,0,117,0,0,0,118,0,0,0,119,0,0,0,120,0,0,0,121,0,0,0,122,0,0,0,123,0,0,0,124,0,0,0,125,0,0,0,126,0,0,0,127,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,124,84,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,2,0,0,0,3,0,0,0,4,0,0,0,5,0,0,0,6,0,0,0,7,0,0,0,8,0,0,0,9,0,0,0,10,0,0,0,11,0,0,0,12,0,0,0,13,0,0,0,14,0,0,0,15,0,0,0,16,0,0,0,17,0,0,0,18,0,0,0,19,0,0,0,20,0,0,0,21,0,0,0,22,0,0,0,23,0,0,0,24,0,0,0,25,0,0,0,26,0,0,0,27,0,0,0,28,0,0,0,29,0,0,0,30,0,0,0,31,0,0,0,32,0,0,0,33,0,0,0,34,0,0,0,35,0,0,0,36,0,0,0,37,0,0,0,38,0,0,0,39,0,0,0,40,0,0,0,41,0,0,0,42,0,0,0,43,0,0,0,44,0,0,0,45,0,0,0,46,0,0,0,47,0,0,0,48,0,0,0,49,0,0,0,50,0,0,0,51,0,0,0,52,0,0,0,53,0,0,0,54,0,0,0,55,0,0,0,56,0,0,0,57,0,0,0,58,0,0,0,59,0,0,0,60,0,0,0,61,0,0,0,62,0,0,0,63,0,0,0,64,0,0,0,65,0,0,0,66,0,0,0,67,0,0,0,68,0,0,0,69,0,0,0,70,0,0,0,71,0,0,0,72,0,0,0,73,0,0,0,74,0,0,0,75,0,0,0,76,0,0,0,77,0,0,0,78,0,0,0,79,0,0,0,80,0,0,0,81,0,0,0,82,0,0,0,83,0,0,0,84,0,0,0,85,0,0,0,86,0,0,0,87,0,0,0,88,0,0,0,89,0,0,0,90,0,0,0,91,0,0,0,92,0,0,0,93,0,0,0,94,0,0,0,95,0,0,0,96,0,0,0,65,0,0,0,66,0,0,0,67,0,0,0,68,0,0,0,69,0,0,0,70,0,0,0,71,0,0,0,72,0,0,0,73,0,0,0,74,0,0,0,75,0,0,0,76,0,0,0,77,0,0,0,78,0,0,0,79,0,0,0,80,0,0,0,81,0,0,0,82,0,0,0,83,0,0,0,84,0,0,0,85,0,0,0,86,0,0,0,87,0,0,0,88,0,0,0,89,0,0,0,90,0,0,0,123,0,0,0,124,0,0,0,125,0,0,0,126,0,0,0,127,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,10,0,0,0,100,0,0,0,232,3,0,0,16,39,0,0,160,134,1,0,64,66,15,0,128,150,152,0,0,225,245,5,185,115,9,0,162,145,19,0,162,145,19,0,162,145,19,0,162,145,19,0,162,145,19,0,162,145,19,0,162,145,19,0,162,145,19,0,162,145,19,0,255,255,255,255,255,255,255,255,255,255,255,255,255,255,0,0,31,0,0,0,224,88,2,0,0,0,0,0,45,244,81,88,207,140,177,192,70,246,181,203,41,49,3,199,4,91,112,48,180,93,253,32,120,127,139,154,216,89,41,80,104,72,137,171,167,86,3,108,255,183,205,136,63,212,119,180,43,165,163,112,241,186,228,168,252,65,131,253,217,111,225,138,122,47,45,116,150,7,31,13,9,94,3,118,44,112,247,64,165,44,167,111,87,65,168,170,116,223,160,88,100,3,74,199,196,60,83,174,175,95,24,4,21,177,227,109,40,134,171,12,164,191,67,240,233,80,129,57,87,22,82,55,3,0,0,0,0,0,0,0,120,29,0,0,192,11,0,0,193,11,0,0,194,11,0,0,15,0,0,0,195,11,0,0,196,11,0,0,18,0,0,0,19,0,0,0,20,0,0,0,197,11,0,0,198,11,0,0,199,11,0,0,24,0,0,0,200,11,0,0,0,0,0,0,136,29,0,0,192,11,0,0,201,11,0,0,202,11,0,0,15,0,0,0,195,11,0,0,196,11,0,0,203,11,0,0,19,0,0,0,20,0,0,0,204,11,0,0,22,0,0,0,205,11,0,0,206,11,0,0,207,11,0,0,0,0,0,0,152,29,0,0,208,11,0,0,209,11,0,0,210,11,0,0,211,11,0,0,212,11,0,0,213,11,0,0,214,11,0,0,215,11,0,0,216,11,0,0,217,11,0,0,218,11,0,0,219,11,0,0,220,11,0,0,221,11,0,0,0,0,0,0,168,29,0,0,208,11,0,0,222,11,0,0,223,11,0,0,211,11,0,0,212,11,0,0,213,11,0,0,224,11,0,0,215,11,0,0,216,11,0,0,225,11,0,0,226,11,0,0,227,11,0,0,228,11,0,0,229,11,0,0,0,0,0,0,2,0,0,0,3,0,0,0,5,0,0,0,7,0,0,0,11,0,0,0,13,0,0,0,17,0,0,0,19,0,0,0,23,0,0,0,29,0,0,0,31,0,0,0,37,0,0,0,41,0,0,0,43,0,0,0,47,0,0,0,53,0,0,0,59,0,0,0,61,0,0,0,67,0,0,0,71,0,0,0,73,0,0,0,79,0,0,0,83,0,0,0,89,0,0,0,97,0,0,0,101,0,0,0,103,0,0,0,107,0,0,0,109,0,0,0,113,0,0,0,127,0,0,0,131,0,0,0,137,0,0,0,139,0,0,0,149,0,0,0,151,0,0,0,157,0,0,0,163,0,0,0,167,0,0,0,173,0,0,0,179,0,0,0,181,0,0,0,191,0,0,0,193,0,0,0,197,0,0,0,199,0,0,0,211,0,0,0,1,0,0,0,11,0,0,0,13,0,0,0,17,0,0,0,19,0,0,0,23,0,0,0,29,0,0,0,31,0,0,0,37,0,0,0,41,0,0,0,43,0,0,0,47,0,0,0,53,0,0,0,59,0,0,0,61,0,0,0,67,0,0,0,71,0,0,0,73,0,0,0,79,0,0,0,83,0,0,0,89,0,0,0,97,0,0,0,101,0,0,0,103,0,0,0,107,0,0,0,109,0,0,0,113,0,0,0,121,0,0,0,127,0,0,0,131,0,0,0,137,0,0,0,139,0,0,0,143,0,0,0,149,0,0,0,151,0,0,0,157,0,0,0,163,0,0,0,167,0,0,0,169,0,0,0,173,0,0,0,179,0,0,0,181,0,0,0,187,0,0,0,191,0,0,0,193,0,0,0,197,0,0,0,199,0,0,0,209,0,0,0,0,0,0,0,200,29,0,0,230,11,0,0,231,11,0,0,0,0,0,0,224,29,0,0,192,11,0,0,232,11,0,0,14,0,0,0,15,0,0,0,195,11,0,0,196,11,0,0,18,0,0,0,19,0,0,0,20,0,0,0,204,11,0,0,22,0,0,0,205,11,0,0,24,0,0,0,200,11,0,0,0,0,0,0,232,29,0,0,208,11,0,0,233,11,0,0,234,11,0,0,211,11,0,0,212,11,0,0,213,11,0,0,214,11,0,0,215,11,0,0,216,11,0,0,225,11,0,0,226,11,0,0,227,11,0,0,220,11,0,0,221,11,0,0,8,0,0,0,0,0,0,0,240,29,0,0,32,0,0,0,33,0,0,0,248,255,255,255,248,255,255,255,240,29,0,0,34,0,0,0,35,0,0,0,8,0,0,0,0,0,0,0,8,30,0,0,235,11,0,0,236,11,0,0,248,255,255,255,248,255,255,255,8,30,0,0,237,11,0,0,238,11,0,0,4,0,0,0,0,0,0,0,32,30,0,0,36,0,0,0,37,0,0,0,252,255,255,255,252,255,255,255,32,30,0,0,38,0,0,0,39,0,0,0,4,0,0,0,0,0,0,0,56,30,0,0,239,11,0,0,240,11,0,0,252,255,255,255,252,255,255,255,56,30,0,0,241,11,0,0,242,11,0,0,0,0,0,0,112,30,0,0,243,11,0,0,244,11,0,0,245,11,0,0,0,0,0,0,128,30,0,0,246,11,0,0,247,11,0,0,248,11,0,0,249,11,0,0,250,11,0,0,251,11,0,0,252,11,0,0,0,0,0,0,144,30,0,0,253,11,0,0,254,11,0,0,255,11,0,0,0,12,0,0,1,12,0,0,2,12,0,0,0,0,0,0,176,30,0,0,3,12,0,0,4,12,0,0,255,11,0,0,5,12,0,0,6,12,0,0,7,12,0,0,0,0,0,0,192,30,0,0,8,12,0,0,9,12,0,0,255,11,0,0,10,12,0,0,11,12,0,0,12,12,0,0,13,12,0,0,14,12,0,0,15,12,0,0,16,12,0,0,17,12,0,0,18,12,0,0,19,12,0,0,20,12,0,0,0,0,0,0,0,31,0,0,21,12,0,0,22,12,0,0,255,11,0,0,23,12,0,0,24,12,0,0,25,12,0,0,26,12,0,0,27,12,0,0,28,12,0,0,29,12,0,0,30,12,0,0,31,12,0,0,32,12,0,0,33,12,0,0,0,0,0,0,56,31,0,0,34,12,0,0,35,12,0,0,255,11,0,0,36,12,0,0,37,12,0,0,38,12,0,0,39,12,0,0,40,12,0,0,41,12,0,0,42,12,0,0,43,12,0,0,0,0,0,0,120,31,0,0,44,12,0,0,45,12,0,0,255,11,0,0,46,12,0,0,47,12,0,0,48,12,0,0,49,12,0,0,50,12,0,0,51,12,0,0,52,12,0,0,53,12,0,0,0,0,0,0,176,31,0,0,54,12,0,0,55,12,0,0,255,11,0,0,56,12,0,0,57,12,0,0,58,12,0,0,59,12,0,0,60,12,0,0,61,12,0,0,62,12,0,0,248,255,255,255,176,31,0,0,63,12,0,0,64,12,0,0,65,12,0,0,66,12,0,0,67,12,0,0,68,12,0,0,69,12,0,0,0,0,0,0,232,31,0,0,70,12,0,0,71,12,0,0,255,11,0,0,72,12,0,0,73,12,0,0,74,12,0,0,75,12,0,0,76,12,0,0,77,12,0,0,78,12,0,0,248,255,255,255,232,31,0,0,79,12,0,0,80,12,0,0,81,12,0,0,82,12,0,0,83,12,0,0,84,12,0,0,85,12,0,0,37,0,0,0,72,0,0,0,58,0,0,0,37,0,0,0,77,0,0,0,58,0,0,0,37,0,0,0,83,0,0,0,37,0,0,0,109,0,0,0,47,0,0,0,37,0,0,0,100,0,0,0,47,0,0,0,37,0,0,0,121,0,0,0,37,0,0,0,89,0,0,0,45,0,0,0,37,0,0,0,109,0,0,0,45,0,0,0,37,0,0,0,100,0,0,0,37,0,0,0,73,0,0,0,58,0,0,0,37,0,0,0,77,0,0,0,58,0,0,0,37,0,0,0,83,0,0,0,32,0,0,0,37,0,0,0,112,0,0,0,37,0,0,0,72,0,0,0,58,0,0,0,37,0,0,0,77,0,0,0,37,0,0,0,72,0,0,0,58,0,0,0,37,0,0,0,77,0,0,0,58,0,0,0,37,0,0,0,83,0,0,0,83,0,0,0,117,0,0,0,110,0,0,0,100,0,0,0,97,0,0,0,121,0,0,0,0,0,0,0,77,0,0,0,111,0,0,0,110,0,0,0,100,0,0,0,97,0,0,0,121,0,0,0,0,0,0,0,84,0,0,0,117,0,0,0,101,0,0,0,115,0,0,0,100,0,0,0,97,0,0,0,121,0,0,0,0,0,0,0,87,0,0,0,101,0,0,0,100,0,0,0,110,0,0,0,101,0,0,0,115,0,0,0,100,0,0,0,97,0,0,0,121,0,0,0,0,0,0,0,84,0,0,0,104,0,0,0,117,0,0,0,114,0,0,0,115,0,0,0,100,0,0,0,97,0,0,0,121,0,0,0,0,0,0,0,70,0,0,0,114,0,0,0,105,0,0,0,100,0,0,0,97,0,0,0,121,0,0,0,0,0,0,0,83,0,0,0,97,0,0,0,116,0,0,0,117,0,0,0,114,0,0,0,100,0,0,0,97,0,0,0,121,0,0,0,0,0,0,0,83,0,0,0,117,0,0,0,110,0,0,0,0,0,0,0,77,0,0,0,111,0,0,0,110,0,0,0,0,0,0,0,84,0,0,0,117,0,0,0,101,0,0,0,0,0,0,0,87,0,0,0,101,0,0,0,100,0,0,0,0,0,0,0,84,0,0,0,104,0,0,0,117,0,0,0,0,0,0,0,70,0,0,0,114,0,0,0,105,0,0,0,0,0,0,0,83,0,0,0,97,0,0,0,116,0,0,0,0,0,0,0,74,0,0,0,97,0,0,0,110,0,0,0,117,0,0,0,97,0,0,0,114,0,0,0,121,0,0,0,0,0,0,0,70,0,0,0,101,0,0,0,98,0,0,0,114,0,0,0,117,0,0,0,97,0,0,0,114,0,0,0,121,0,0,0,0,0,0,0,77,0,0,0,97,0,0,0,114,0,0,0,99,0,0,0,104,0,0,0,0,0,0,0,65,0,0,0,112,0,0,0,114,0,0,0,105,0,0,0,108,0,0,0,0,0,0,0,77,0,0,0,97,0,0,0,121,0,0,0,0,0,0,0,74,0,0,0,117,0,0,0,110,0,0,0,101,0,0,0,0,0,0,0,74,0,0,0,117,0,0,0,108,0,0,0,121,0,0,0,0,0,0,0,65,0,0,0,117,0,0,0,103,0,0,0,117,0,0,0,115,0,0,0,116,0,0,0,0,0,0,0,83,0,0,0,101,0,0,0,112,0,0,0,116,0,0,0,101,0,0,0,109,0,0,0,98,0,0,0,101,0,0,0,114,0,0,0,0,0,0,0,79,0,0,0,99,0,0,0,116,0,0,0,111,0,0,0,98,0,0,0,101,0,0,0,114,0,0,0,0,0,0,0,78,0,0,0,111,0,0,0,118,0,0,0,101,0,0,0,109,0,0,0,98,0,0,0,101,0,0,0,114,0,0,0,0,0,0,0,68,0,0,0,101,0,0,0,99,0,0,0,101,0,0,0,109,0,0,0,98,0,0,0,101,0,0,0,114,0,0,0,0,0,0,0,74,0,0,0,97,0,0,0,110,0,0,0,0,0,0,0,70,0,0,0,101,0,0,0,98,0,0,0,0,0,0,0,77,0,0,0,97,0,0,0,114,0,0,0,0,0,0,0,65,0,0,0,112,0,0,0,114,0,0,0,0,0,0,0,74,0,0,0,117,0,0,0,110,0,0,0,0,0,0,0,74,0,0,0,117,0,0,0,108,0,0,0,0,0,0,0,65,0,0,0,117,0,0,0,103,0,0,0,0,0,0,0,83,0,0,0,101,0,0,0,112,0,0,0,0,0,0,0,79,0,0,0,99,0,0,0,116,0,0,0,0,0,0,0,78,0,0,0,111,0,0,0,118,0,0,0,0,0,0,0,68,0,0,0,101,0,0,0,99,0,0,0,0,0,0,0,65,0,0,0,77,0,0,0,0,0,0,0,80,0,0,0,77,0,0,0,0,0,0,0,37,0,0,0,97,0,0,0,32,0,0,0,37,0,0,0,98,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,32,0,0,0,37,0,0,0,72,0,0,0,58,0,0,0,37,0,0,0,77,0,0,0,58,0,0,0,37,0,0,0,83,0,0,0,32,0,0,0,37,0,0,0,89,0,0,0,0,0,0,0,37,0,0,0,73,0,0,0,58,0,0,0,37,0,0,0,77,0,0,0,58,0,0,0,37,0,0,0,83,0,0,0,32,0,0,0,37,0,0,0,112,0,0,0,0,0,0,0,37,0,0,0,109,0,0,0,47,0,0,0,37,0,0,0,100,0,0,0,47,0,0,0,37,0,0,0,121,0,0,0,0,0,0,0,37,0,0,0,72,0,0,0,58,0,0,0,37,0,0,0,77,0,0,0,58,0,0,0,37,0,0,0,83,0,0,0,0,0,0,0,0,0,0,0,24,32,0,0,86,12,0,0,87,12,0,0,255,11,0,0,88,12,0,0,0,0,0,0,64,32,0,0,89,12,0,0,90,12,0,0,255,11,0,0,91,12,0,0,0,0,0,0,96,32,0,0,92,12,0,0,93,12,0,0,255,11,0,0,94,12,0,0,95,12,0,0,96,12,0,0,97,12,0,0,98,12,0,0,99,12,0,0,100,12,0,0,101,12,0,0,102,12,0,0,0,0,0,0,136,32,0,0,103,12,0,0,104,12,0,0,255,11,0,0,105,12,0,0,106,12,0,0,107,12,0,0,108,12,0,0,109,12,0,0,110,12,0,0,111,12,0,0,112,12,0,0,113,12,0,0,0,0,0,0,168,32,0,0,114,12,0,0,115,12,0,0,255,11,0,0,116,12,0,0,117,12,0,0,118,12,0,0,119,12,0,0,120,12,0,0,121,12,0,0,122,12,0,0,123,12,0,0,124,12,0,0,0,0,0,0,200,32,0,0,125,12,0,0,126,12,0,0,255,11,0,0,127,12,0,0,128,12,0,0,129,12,0,0,130,12,0,0,131,12,0,0,132,12,0,0,133,12,0,0,134,12,0,0,135,12,0,0,0,0,0,0,232,32,0,0,136,12,0,0,137,12,0,0,255,11,0,0,138,12,0,0,139,12,0,0,0,0,0,0,16,33,0,0,140,12,0,0,141,12,0,0,255,11,0,0,142,12,0,0,143,12,0,0,0,0,0,0,56,33,0,0,144,12,0,0,145,12,0,0,255,11,0,0,146,12,0,0,147,12,0,0,0,0,0,0,96,33,0,0,148,12,0,0,149,12,0,0,255,11,0,0,150,12,0,0,151,12,0,0,0,0,0,0,136,33,0,0,152,12,0,0,153,12,0,0,255,11,0,0,154,12,0,0,155,12,0,0,156,12,0,0,0,0,0,0,176,33,0,0,157,12,0,0,158,12,0,0,255,11,0,0,159,12,0,0,160,12,0,0,161,12,0,0,0,0,0,0,208,33,0,0,162,12,0,0,163,12,0,0,255,11,0,0,164,12,0,0,165,12,0,0,166,12,0,0,167,12,0,0,168,12,0,0,169,12,0,0,170,12,0,0,0,0,0,0,8,34,0,0,162,12,0,0,171,12,0,0,255,11,0,0,164,12,0,0,165,12,0,0,166,12,0,0,167,12,0,0,168,12,0,0,169,12,0,0,170,12,0,0,0,0,0,0,56,34,0,0,172,12,0,0,173,12,0,0,255,11,0,0,174,12,0,0,175,12,0,0,176,12,0,0,177,12,0,0,178,12,0,0,179,12,0,0,180,12,0,0,0,0,0,0,120,34,0,0,181,12,0,0,182,12,0,0,255,11,0,0,0,0,0,0,136,34,0,0,183,12,0,0,184,12,0,0,255,11,0,0,185,12,0,0,186,12,0,0,187,12,0,0,188,12,0,0,189,12,0,0,190,12,0,0,191,12,0,0,192,12,0,0,0,0,0,0,208,34,0,0,193,12,0,0,194,12,0,0,255,11,0,0,195,12,0,0,196,12,0,0,197,12,0,0,198,12,0,0,199,12,0,0,0,0,0,0,224,34,0,0,200,12,0,0,201,12,0,0,255,11,0,0,202,12,0,0,203,12,0,0,204,12,0,0,205,12,0,0,206,12,0,0,116,0,0,0,114,0,0,0,117,0,0,0,101,0,0,0,0,0,0,0,102,0,0,0,97,0,0,0,108,0,0,0,115,0,0,0,101,0,0,0,0,0,0,0,0,0,0,0,160,30,0,0,162,12,0,0,207,12,0,0,255,11,0,0,0,0,0,0,176,34,0,0,162,12,0,0,208,12,0,0,255,11,0,0,209,12,0,0,210,12,0,0,211,12,0,0,212,12,0,0,213,12,0,0,214,12,0,0,215,12,0,0,216,12,0,0,217,12,0,0,218,12,0,0,219,12,0,0,220,12,0,0,0,0,0,0,24,34,0,0,162,12,0,0,221,12,0,0,255,11,0,0,222,12,0,0,223,12,0,0,224,12,0,0,225,12,0,0,226,12,0,0,227,12,0,0,228,12,0,0,0,0,0,0,88,34,0,0,162,12,0,0,229,12,0,0,255,11,0,0,230,12,0,0,231,12,0,0,232,12,0,0,233,12,0,0,234,12,0,0,235,12,0,0,236,12,0,0,0,0,0,0,224,33,0,0,162,12,0,0,237,12,0,0,255,11,0,0,164,12,0,0,165,12,0,0,166,12,0,0,167,12,0,0,168,12,0,0,169,12,0,0,170,12,0,0,0,0,0,0,0,35,0,0,238,12,0,0,239,12,0,0,245,11,0,0,240,12,0,0,0,0,0,0,40,35,0,0,241,12,0,0,242,12,0,0,243,12,0,0,244,12,0,0,245,12,0,0,246,12,0,0,247,12,0,0,248,12,0,0,0,0,0,0,56,35,0,0,241,12,0,0,249,12,0,0,243,12,0,0,244,12,0,0,245,12,0,0,250,12,0,0,251,12,0,0,252,12,0,0,92,129,9,0,0,0,0,0,96,35,0,0,253,12,0,0,254,12,0,0,255,12,0,0,0,0,0,0,112,35,0,0,0,13,0,0,1,13,0,0,2,13,0,0,0,0,0,0,128,35,0,0,3,13,0,0,4,13,0,0,245,11,0,0,0,0,0,0,144,35,0,0,0,13,0,0,5,13,0,0,2,13,0,0,0,0,0,0,160,35,0,0,0,13,0,0,6,13,0,0,2,13,0,0,0,0,0,0,176,35,0,0,3,13,0,0,7,13,0,0,245,11,0,0,0,0,0,0,192,35,0,0,8,13,0,0,9,13,0,0,10,13,0,0,0,0,0,0,208,35,0,0,11,13,0,0,12,13,0,0,13,13,0,0,0,0,0,0,0,36,0,0,241,12,0,0,14,13,0,0,243,12,0,0,244,12,0,0,245,12,0,0,15,13,0,0,16,13,0,0,17,13,0,0,4,0,0,0,5,0,3,0,3,0,253,255,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,14,0,1,224,15,0,1,192,16,0,1,128,4,0,0,0,5,0,3,0,3,0,253,255,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,14,0,1,224,15,0,1,192,16,0,1,128,4,0,0,0,5,0,3,0,3,0,253,255,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,14,0,1,224,15,0,1,192,16,0,1,128,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,16,0,1,128,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,16,0,1,128,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,16,0,1,128,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,16,0,1,128,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,16,0,1,128,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,16,0,1,128,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,16,0,1,128,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,2,0,0,0,5,0,3,0,7,0,5,0,16,0,1,128,2,0,0,0,5,0,3,0,7,0,5,0,16,0,1,128,2,0,0,0,5,0,3,0,7,0,5,0,16,0,1,128,2,0,0,0,5,0,3,0,7,0,5,0,16,0,1,128,4,0,0,0,5,0,3,0,3,0,253,255,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,14,0,1,224,15,0,1,192,16,0,1,128,4,0,0,0,5,0,3,0,3,0,253,255,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,14,0,1,224,15,0,1,192,16,0,1,128,4,0,0,0,5,0,3,0,3,0,253,255,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,14,0,1,224,15,0,1,192,16,0,1,128,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,16,0,1,128,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,16,0,1,128,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,16,0,1,128,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,16,0,1,128,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,16,0,1,128,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,16,0,1,128,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,16,0,1,128,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,3,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,16,0,1,128,2,0,0,0,5,0,3,0,7,0,5,0,16,0,1,128,2,0,0,0,5,0,3,0,7,0,5,0,16,0,1,128,2,0,0,0,5,0,3,0,7,0,5,0,16,0,1,128,2,0,0,0,5,0,3,0,7,0,5,0,16,0,1,128,2,0,0,0,5,0,3,0,7,0,5,0,16,0,1,128,2,0,0,0,5,0,3,0,7,0,5,0,16,0,1,128,2,0,0,0,5,0,3,0,7,0,5,0,16,0,1,128,4,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,14,0,1,224,15,0,1,192,4,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,14,0,1,224,15,0,1,192,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,4,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,14,0,1,224,15,0,1,192,4,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,14,0,1,224,15,0,1,192,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,14,0,1,224,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,14,0,1,224,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,14,0,1,224,4,0,0,0,5,0,3,0,7,0,5,0,3,0,253,255,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,8,0,129,255,9,0,1,255,10,0,1,254,11,0,1,252,12,0,1,248,13,0,1,240,14,0,1,224,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,3,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,4,0,249,255,5,0,241,255,6,0,225,255,7,0,193,255,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,2,0,0,0,5,0,3,0,7,0,5,0,10,0,9,0,0,0,4,0,8,0,12,0,16,0,20,0,24,0,30,0,36,0,44,0,52,0,62,0,74,0,90,0,110,0,134,0,162,0,196,0,238,0,32,1,86,1,162,1,64,2,4,4,4,4,4,4,6,6,8,8,10,12,16,20,24,28,34,42,50,54,76,158,0,0,12,0,24,0,36,0,48,0,66,0,90,0,120,0,156,0,198,0,252,0,62,1,152,1,64,2,4,4,4,4,6,8,10,12,14,18,22,30,56,0,0,0,4,0,8,0,12,0,16,0,20,0,24,0,30,0,36,0,42,0,50,0,60,0,72,0,88,0,106,0,128,0,156,0,190,0,230,0,20,1,74,1,128,1,64,2,4,4,4,4,4,4,6,6,6,8,10,12,16,18,22,28,34,40,46,54,54,192,0,0,12,0,24,0,36,0,48,0,66,0,84,0,114,0,150,0,192,0,240,0,44,1,122,1,64,2,4,4,4,4,6,6,10,12,14,16,20,26,66,0,0,0,4,0,8,0,12,0,16,0,20,0,24,0,30,0,36,0,44,0,54,0,66,0,82,0,102,0,126,0,156,0,194,0,240,0,40,1,108,1,192,1,38,2,64,2,4,4,4,4,4,4,6,6,8,10,12,16,20,24,30,38,46,56,68,84,102,26,0,0,12,0,24,0,36,0,48,0,66,0,90,0,126,0,174,0,234,0,56,1,158,1,28,2,64,2,4,4,4,4,6,8,12,16,20,26,34,42,12,0,0,0,6,0,12,0,18,0,24,0,30,0,36,0,44,0,54,0,66,0,80,0,96,0,116,0,140,0,168,0,200,0,238,0,28,1,80,1,140,1,208,1,10,2,64,2,6,6,6,6,6,6,8,10,12,14,16,20,24,28,32,38,46,52,60,68,58,54,0,0,12,0,24,0,36,0,54,0,72,0,96,0,126,0,168,0,222,0,44,1,140,1,10,2,64,2,4,4,4,6,6,8,10,14,18,26,32,42,18,0,0,0,6,0,12,0,18,0,24,0,30,0,36,0,44,0,54,0,66,0,80,0,96,0,114,0,136,0,162,0,194,0,232,0,22,1,76,1,138,1,208,1,28,2,64,2,6,6,6,6,6,6,8,10,12,14,16,18,22,26,32,38,46,54,62,70,76,36,0,0,12,0,24,0,36,0,54,0,78,0,108,0,144,0,186,0,240,0,56,1,152,1,28,2,64,2,4,4,4,6,8,10,12,14,18,24,32,44,12,0,0,0,6,0,12,0,18,0,24,0,30,0,36,0,44,0,54,0,66,0,80,0,96,0,116,0,140,0,168,0,200,0,238,0,28,1,80,1,140,1,208,1,10,2,64,2,6,6,6,6,6,6,8,10,12,14,16,20,24,28,32,38,46,52,60,68,58,54,0,0,12,0,24,0,36,0,54,0,78,0,108,0,144,0,186,0,240,0,56,1,146,1,10,2,64,2,4,4,4,6,8,10,12,14,18,24,30,40,18,0,0,0,6,0,12,0,18,0,24,0,30,0,36,0,44,0,54,0,66,0,80,0,96,0,116,0,140,0,168,0,200,0,238,0,28,1,80,1,140,1,208,1,10,2,64,2,6,6,6,6,6,6,8,10,12,14,16,20,24,28,32,38,46,52,60,68,58,54,0,0,12,0,24,0,36,0,54,0,78,0,108,0,144,0,186,0,240,0,56,1,146,1,10,2,64,2,4,4,4,6,8,10,12,14,18,24,30,40,18,0,0,0,6,0,12,0,18,0,24,0,30,0,36,0,44,0,54,0,66,0,80,0,96,0,116,0,140,0,168,0,200,0,238,0,28,1,80,1,140,1,208,1,10,2,64,2,6,6,6,6,6,6,8,10,12,14,16,20,24,28,32,38,46,52,60,68,58,54,0,0,12,0,24,0,36,0,54,0,78,0,108,0,144,0,186,0,240,0,56,1,146,1,10,2,64,2,4,4,4,6,8,10,12,14,18,24,30,40,18,0,0,0,12,0,24,0,36,0,48,0,60,0,72,0,88,0,108,0,132,0,160,0,192,0,232,0,24,1,80,1,144,1,220,1,54,2,56,2,58,2,60,2,62,2,64,2,12,12,12,12,12,12,16,20,24,28,32,40,48,56,64,76,90,2,2,2,2,2,0,0,24,0,48,0,72,0,108,0,156,0,216,0,32,1,116,1,224,1,230,1,236,1,242,1,64,2,8,8,8,12,16,20,24,28,36,2,2,2,26,0,251,255,253,255,255,255,17,0,1,0,16,0,0,0,241,255,245,255,247,255,251,255,253,255,255,255,34,0,2,0,18,0,255,255,33,0,32,0,17,0,255,255,1,0,16,0,0,0,243,255,245,255,247,255,251,255,253,255,255,255,34,0,2,0,18,0,255,255,33,0,32,0,16,0,17,0,255,255,1,0,0,0,227,255,231,255,233,255,241,255,249,255,251,255,253,255,255,255,51,0,35,0,50,0,49,0,253,255,255,255,19,0,3,0,255,255,48,0,34,0,253,255,255,255,18,0,33,0,255,255,2,0,32,0,17,0,255,255,1,0,16,0,0,0,231,255,237,255,243,255,247,255,251,255,253,255,255,255,51,0,3,0,35,0,255,255,50,0,48,0,255,255,19,0,49,0,253,255,255,255,34,0,2,0,18,0,253,255,255,255,33,0,32,0,1,0,255,255,17,0,255,255,16,0,0,0,187,255,191,255,199,255,217,255,227,255,239,255,245,255,249,255,253,255,255,255,85,0,69,0,255,255,84,0,83,0,255,255,53,0,68,0,253,255,255,255,37,0,82,0,21,0,251,255,255,255,81,0,255,255,5,0,52,0,255,255,80,0,255,255,67,0,51,0,251,255,253,255,255,255,36,0,66,0,20,0,255,255,65,0,64,0,245,255,249,255,253,255,255,255,4,0,35,0,255,255,50,0,3,0,255,255,19,0,49,0,253,255,255,255,48,0,34,0,18,0,251,255,255,255,33,0,255,255,2,0,32,0,17,0,255,255,1,0,16,0,0,0,191,255,193,255,197,255,211,255,225,255,237,255,243,255,249,255,251,255,253,255,255,255,85,0,84,0,69,0,83,0,253,255,255,255,53,0,68,0,37,0,253,255,255,255,82,0,5,0,21,0,251,255,255,255,81,0,255,255,52,0,67,0,253,255,255,255,80,0,51,0,36,0,251,255,253,255,255,255,66,0,20,0,65,0,253,255,255,255,4,0,64,0,255,255,35,0,50,0,247,255,249,255,253,255,255,255,19,0,49,0,255,255,3,0,48,0,34,0,255,255,2,0,32,0,255,255,18,0,33,0,17,0,253,255,255,255,1,0,16,0,0,0,193,255,203,255,215,255,227,255,237,255,245,255,251,255,253,255,255,255,85,0,69,0,53,0,255,255,83,0,255,255,84,0,5,0,253,255,255,255,68,0,37,0,255,255,82,0,21,0,253,255,255,255,81,0,52,0,255,255,67,0,255,255,80,0,4,0,249,255,253,255,255,255,36,0,66,0,255,255,51,0,64,0,255,255,20,0,65,0,251,255,253,255,255,255,35,0,50,0,19,0,255,255,49,0,255,255,3,0,48,0,251,255,253,255,255,255,34,0,2,0,18,0,255,255,33,0,32,0,253,255,255,255,17,0,1,0,255,255,16,0,0,0,131,255,135,255,145,255,173,255,201,255,221,255,235,255,243,255,249,255,253,255,255,255,119,0,103,0,255,255,118,0,87,0,253,255,255,255,117,0,102,0,71,0,253,255,255,255,116,0,86,0,255,255,101,0,55,0,247,255,253,255,255,255,115,0,70,0,253,255,255,255,85,0,84,0,99,0,255,255,39,0,114,0,245,255,251,255,253,255,255,255,100,0,7,0,112,0,255,255,98,0,255,255,69,0,53,0,251,255,255,255,6,0,255,255,83,0,68,0,23,0,239,255,251,255,255,255,113,0,255,255,54,0,38,0,251,255,253,255,255,255,37,0,82,0,21,0,255,255,81,0,255,255,52,0,67,0,253,255,255,255,22,0,97,0,255,255,96,0,255,255,5,0,80,0,237,255,245,255,249,255,253,255,255,255,36,0,66,0,255,255,51,0,4,0,255,255,20,0,65,0,253,255,255,255,64,0,35,0,255,255,50,0,3,0,253,255,255,255,19,0,49,0,255,255,48,0,34,0,249,255,253,255,255,255,18,0,33,0,255,255,2,0,32,0,17,0,255,255,1,0,16,0,0,0,135,255,143,255,167,255,197,255,213,255,229,255,239,255,249,255,253,255,255,255,119,0,103,0,255,255,118,0,117,0,253,255,255,255,102,0,71,0,255,255,116,0,255,255,87,0,85,0,251,255,253,255,255,255,86,0,101,0,55,0,255,255,115,0,70,0,247,255,249,255,253,255,255,255,69,0,84,0,255,255,53,0,83,0,39,0,255,255,114,0,255,255,100,0,7,0,251,255,255,255,113,0,255,255,23,0,112,0,253,255,255,255,54,0,99,0,255,255,96,0,255,255,68,0,37,0,243,255,249,255,251,255,253,255,255,255,82,0,5,0,21,0,98,0,253,255,255,255,38,0,6,0,22,0,251,255,255,255,97,0,255,255,81,0,52,0,251,255,255,255,80,0,255,255,67,0,51,0,255,255,36,0,66,0,241,255,245,255,249,255,253,255,255,255,20,0,65,0,255,255,4,0,64,0,255,255,35,0,50,0,255,255,19,0,49,0,251,255,253,255,255,255,3,0,48,0,34,0,33,0,251,255,255,255,18,0,255,255,2,0,32,0,17,0,253,255,255,255,1,0,16,0,0,0,141,255,157,255,183,255,211,255,229,255,239,255,247,255,251,255,253,255,255,255,119,0,103,0,118,0,255,255,87,0,117,0,253,255,255,255,102,0,71,0,255,255,116,0,101,0,253,255,255,255,86,0,55,0,253,255,255,255,115,0,85,0,39,0,249,255,253,255,255,255,114,0,70,0,255,255,100,0,23,0,251,255,255,255,113,0,255,255,7,0,112,0,255,255,54,0,99,0,243,255,247,255,253,255,255,255,69,0,84,0,255,255,68,0,255,255,6,0,5,0,255,255,38,0,98,0,251,255,255,255,97,0,255,255,22,0,96,0,253,255,255,255,53,0,83,0,255,255,37,0,82,0,239,255,249,255,253,255,255,255,21,0,81,0,255,255,52,0,67,0,251,255,253,255,255,255,80,0,4,0,36,0,255,255,66,0,20,0,253,255,255,255,51,0,65,0,255,255,35,0,50,0,245,255,249,255,251,255,253,255,255,255,64,0,3,0,48,0,19,0,255,255,49,0,34,0,255,255,18,0,33,0,249,255,251,255,253,255,255,255,2,0,32,0,0,0,17,0,255,255,1,0,16,0,3,254,9,254,37,254,107,254,179,254,247,254,51,255,103,255,141,255,173,255,203,255,221,255,235,255,243,255,247,255,249,255,251,255,253,255,255,255,254,0,252,0,253,0,237,0,255,0,255,255,239,0,223,0,253,255,255,255,238,0,207,0,255,255,222,0,191,0,247,255,253,255,255,255,251,0,206,0,255,255,220,0,255,255,175,0,233,0,255,255,236,0,221,0,247,255,251,255,253,255,255,255,250,0,205,0,190,0,255,255,235,0,159,0,253,255,255,255,249,0,234,0,255,255,189,0,219,0,239,255,247,255,253,255,255,255,143,0,248,0,255,255,204,0,255,255,174,0,158,0,251,255,255,255,142,0,255,255,127,0,126,0,247,0,251,255,255,255,218,0,255,255,173,0,188,0,253,255,255,255,203,0,246,0,111,0,241,255,249,255,253,255,255,255,232,0,95,0,255,255,157,0,217,0,253,255,255,255,245,0,231,0,255,255,172,0,187,0,247,255,253,255,255,255,79,0,244,0,253,255,255,255,202,0,230,0,243,0,255,255,63,0,255,255,141,0,216,0,235,255,247,255,253,255,255,255,47,0,242,0,253,255,255,255,110,0,156,0,15,0,251,255,253,255,255,255,201,0,94,0,171,0,253,255,255,255,125,0,215,0,78,0,245,255,251,255,253,255,255,255,200,0,214,0,62,0,255,255,185,0,255,255,155,0,170,0,255,255,31,0,241,0,233,255,243,255,251,255,255,255,240,0,255,255,186,0,229,0,253,255,255,255,228,0,140,0,255,255,109,0,227,0,251,255,255,255,226,0,255,255,46,0,14,0,255,255,30,0,225,0,241,255,249,255,253,255,255,255,224,0,93,0,255,255,213,0,124,0,253,255,255,255,199,0,77,0,255,255,139,0,184,0,249,255,253,255,255,255,212,0,154,0,255,255,169,0,108,0,255,255,198,0,61,0,219,255,235,255,247,255,251,255,253,255,255,255,211,0,123,0,45,0,255,255,210,0,29,0,251,255,255,255,183,0,255,255,92,0,197,0,253,255,255,255,153,0,122,0,195,0,249,255,251,255,253,255,255,255,167,0,151,0,75,0,209,0,253,255,255,255,13,0,208,0,255,255,138,0,168,0,245,255,249,255,253,255,255,255,76,0,196,0,255,255,107,0,182,0,255,255,60,0,44,0,253,255,255,255,194,0,91,0,253,255,255,255,181,0,137,0,28,0,213,255,233,255,245,255,251,255,255,255,193,0,255,255,152,0,12,0,255,255,192,0,255,255,180,0,106,0,251,255,253,255,255,255,166,0,121,0,59,0,255,255,179,0,255,255,136,0,90,0,245,255,251,255,255,255,43,0,255,255,165,0,105,0,255,255,164,0,255,255,120,0,135,0,251,255,255,255,148,0,255,255,119,0,118,0,178,0,245,255,253,255,255,255,27,0,177,0,253,255,255,255,11,0,176,0,255,255,150,0,74,0,249,255,253,255,255,255,58,0,163,0,255,255,89,0,149,0,255,255,42,0,162,0,209,255,233,255,247,255,253,255,255,255,26,0,161,0,253,255,255,255,10,0,104,0,160,0,251,255,253,255,255,255,134,0,73,0,147,0,253,255,255,255,57,0,88,0,255,255,133,0,103,0,247,255,253,255,255,255,41,0,146,0,253,255,255,255,87,0,117,0,56,0,251,255,255,255,131,0,255,255,102,0,71,0,253,255,255,255,116,0,86,0,255,255,101,0,115,0,245,255,253,255,255,255,25,0,145,0,253,255,255,255,9,0,144,0,255,255,72,0,132,0,249,255,251,255,255,255,114,0,255,255,70,0,100,0,40,0,255,255,130,0,24,0,215,255,229,255,245,255,251,255,253,255,255,255,55,0,39,0,23,0,255,255,113,0,255,255,85,0,7,0,249,255,253,255,255,255,112,0,54,0,255,255,99,0,69,0,253,255,255,255,84,0,38,0,255,255,98,0,53,0,251,255,255,255,129,0,255,255,8,0,128,0,253,255,255,255,22,0,97,0,255,255,6,0,96,0,243,255,247,255,251,255,253,255,255,255,83,0,68,0,37,0,255,255,82,0,5,0,255,255,21,0,81,0,249,255,253,255,255,255,52,0,67,0,255,255,80,0,36,0,253,255,255,255,66,0,51,0,20,0,237,255,245,255,251,255,255,255,65,0,255,255,4,0,64,0,253,255,255,255,35,0,50,0,19,0,253,255,255,255,49,0,3,0,255,255,48,0,34,0,253,255,255,255,18,0,33,0,255,255,2,0,32,0,253,255,255,255,17,0,1,0,16,0,0,0,17,254,67,254,157,254,249,254,73,255,141,255,179,255,213,255,229,255,243,255,249,255,253,255,255,255,255,0,239,0,255,255,254,0,223,0,255,255,238,0,255,255,253,0,207,0,249,255,253,255,255,255,252,0,222,0,255,255,237,0,191,0,255,255,251,0,255,255,206,0,236,0,249,255,253,255,255,255,221,0,175,0,255,255,250,0,190,0,253,255,255,255,235,0,205,0,255,255,220,0,159,0,241,255,249,255,253,255,255,255,249,0,234,0,255,255,189,0,219,0,253,255,255,255,143,0,248,0,255,255,204,0,158,0,249,255,253,255,255,255,233,0,127,0,255,255,247,0,173,0,253,255,255,255,218,0,188,0,255,255,111,0,255,255,174,0,15,0,237,255,245,255,253,255,255,255,203,0,246,0,253,255,255,255,142,0,232,0,255,255,95,0,157,0,253,255,255,255,245,0,126,0,255,255,231,0,172,0,247,255,253,255,255,255,202,0,187,0,253,255,255,255,217,0,141,0,79,0,253,255,255,255,244,0,63,0,255,255,243,0,216,0,223,255,239,255,247,255,253,255,255,255,230,0,47,0,255,255,242,0,255,255,110,0,240,0,253,255,255,255,31,0,241,0,255,255,156,0,201,0,249,255,253,255,255,255,94,0,171,0,255,255,186,0,229,0,253,255,255,255,125,0,215,0,255,255,78,0,228,0,241,255,249,255,253,255,255,255,140,0,200,0,255,255,62,0,109,0,253,255,255,255,214,0,227,0,255,255,155,0,185,0,249,255,253,255,255,255,46,0,170,0,255,255,226,0,30,0,251,255,255,255,225,0,255,255,14,0,224,0,255,255,93,0,213,0,211,255,231,255,243,255,249,255,253,255,255,255,124,0,199,0,255,255,77,0,139,0,255,255,212,0,255,255,184,0,154,0,249,255,253,255,255,255,169,0,108,0,255,255,198,0,61,0,255,255,211,0,210,0,247,255,251,255,253,255,255,255,45,0,13,0,29,0,255,255,123,0,183,0,251,255,255,255,209,0,255,255,92,0,208,0,255,255,197,0,138,0,239,255,249,255,253,255,255,255,168,0,76,0,255,255,196,0,107,0,251,255,255,255,182,0,255,255,153,0,12,0,255,255,60,0,195,0,247,255,253,255,255,255,122,0,167,0,255,255,166,0,255,255,192,0,11,0,255,255,194,0,255,255,44,0,91,0,201,255,227,255,241,255,249,255,253,255,255,255,181,0,28,0,255,255,137,0,152,0,253,255,255,255,193,0,75,0,255,255,180,0,106,0,251,255,253,255,255,255,59,0,121,0,179,0,253,255,255,255,151,0,136,0,255,255,43,0,90,0,245,255,251,255,255,255,178,0,255,255,165,0,27,0,255,255,177,0,255,255,176,0,105,0,249,255,253,255,255,255,150,0,74,0,255,255,164,0,120,0,253,255,255,255,135,0,58,0,163,0,239,255,249,255,253,255,255,255,89,0,149,0,255,255,42,0,162,0,253,255,255,255,26,0,161,0,253,255,255,255,10,0,160,0,104,0,249,255,253,255,255,255,134,0,73,0,255,255,148,0,57,0,251,255,255,255,147,0,255,255,119,0,9,0,255,255,88,0,133,0,203,255,227,255,243,255,249,255,253,255,255,255,41,0,103,0,255,255,118,0,146,0,255,255,145,0,255,255,25,0,144,0,249,255,253,255,255,255,72,0,132,0,255,255,87,0,117,0,253,255,255,255,56,0,131,0,255,255,102,0,71,0,249,255,253,255,255,255,40,0,130,0,255,255,24,0,129,0,249,255,253,255,255,255,116,0,8,0,255,255,128,0,86,0,253,255,255,255,101,0,55,0,255,255,115,0,70,0,239,255,249,255,253,255,255,255,39,0,114,0,255,255,100,0,23,0,253,255,255,255,85,0,113,0,253,255,255,255,7,0,112,0,54,0,249,255,253,255,255,255,99,0,69,0,255,255,84,0,38,0,253,255,255,255,98,0,22,0,253,255,255,255,6,0,96,0,53,0,223,255,237,255,247,255,251,255,255,255,97,0,255,255,83,0,68,0,255,255,37,0,82,0,253,255,255,255,21,0,81,0,253,255,255,255,5,0,80,0,52,0,249,255,253,255,255,255,67,0,36,0,255,255,66,0,51,0,255,255,65,0,255,255,20,0,4,0,247,255,253,255,255,255,35,0,50,0,253,255,255,255,64,0,3,0,19,0,253,255,255,255,49,0,48,0,34,0,247,255,249,255,253,255,255,255,18,0,33,0,255,255,2,0,32,0,17,0,253,255,255,255,1,0,16,0,0,0,3,254,9,254,51,254,189,254,153,255,219,255,229,255,241,255,249,255,253,255,255,255,239,0,254,0,255,255,223,0,253,0,253,255,255,255,207,0,252,0,255,255,191,0,251,0,251,255,255,255,175,0,255,255,250,0,159,0,253,255,255,255,249,0,248,0,143,0,249,255,253,255,255,255,127,0,247,0,255,255,111,0,246,0,255,0,247,255,251,255,253,255,255,255,95,0,245,0,79,0,255,255,244,0,243,0,203,255,255,255,240,0,255,255,63,0,227,255,237,255,243,255,249,255,251,255,255,255,206,0,255,255,236,0,221,0,222,0,255,255,233,0,255,255,234,0,217,0,255,255,238,0,255,255,237,0,235,0,253,255,255,255,190,0,205,0,253,255,255,255,220,0,219,0,174,0,245,255,251,255,255,255,204,0,255,255,173,0,218,0,253,255,255,255,126,0,172,0,202,0,251,255,253,255,255,255,201,0,125,0,94,0,189,0,242,0,163,255,251,255,253,255,255,255,47,0,15,0,31,0,255,255,241,0,207,255,231,255,243,255,251,255,255,255,158,0,255,255,188,0,203,0,253,255,255,255,142,0,232,0,255,255,157,0,231,0,249,255,253,255,255,255,187,0,141,0,255,255,216,0,110,0,255,255,230,0,156,0,243,255,249,255,253,255,255,255,171,0,186,0,255,255,229,0,215,0,255,255,78,0,255,255,228,0,140,0,253,255,255,255,200,0,62,0,255,255,109,0,255,255,214,0,155,0,237,255,245,255,251,255,253,255,255,255,185,0,170,0,225,0,255,255,212,0,255,255,184,0,169,0,251,255,255,255,123,0,255,255,183,0,208,0,227,0,249,255,253,255,255,255,14,0,224,0,255,255,93,0,213,0,253,255,255,255,124,0,199,0,255,255,77,0,139,0,181,255,211,255,229,255,243,255,249,255,253,255,255,255,154,0,108,0,255,255,198,0,61,0,253,255,255,255,92,0,197,0,13,0,249,255,253,255,255,255,138,0,168,0,255,255,153,0,76,0,253,255,255,255,182,0,122,0,60,0,245,255,251,255,253,255,255,255,91,0,137,0,28,0,255,255,192,0,255,255,152,0,121,0,255,255,226,0,255,255,46,0,30,0,241,255,249,255,253,255,255,255,211,0,45,0,255,255,210,0,209,0,251,255,255,255,59,0,255,255,151,0,136,0,29,0,249,255,253,255,255,255,196,0,107,0,255,255,195,0,167,0,255,255,44,0,255,255,194,0,181,0,233,255,243,255,249,255,253,255,255,255,193,0,12,0,255,255,75,0,180,0,253,255,255,255,106,0,166,0,179,0,251,255,253,255,255,255,90,0,165,0,43,0,255,255,178,0,27,0,243,255,251,255,255,255,177,0,255,255,11,0,176,0,253,255,255,255,105,0,150,0,255,255,74,0,164,0,251,255,253,255,255,255,120,0,135,0,163,0,253,255,255,255,58,0,89,0,42,0,159,255,199,255,223,255,237,255,245,255,251,255,253,255,255,255,149,0,104,0,161,0,253,255,255,255,134,0,119,0,148,0,251,255,253,255,255,255,73,0,87,0,103,0,162,0,251,255,255,255,26,0,255,255,10,0,160,0,253,255,255,255,57,0,147,0,255,255,88,0,133,0,247,255,253,255,255,255,41,0,146,0,253,255,255,255,118,0,9,0,25,0,251,255,255,255,145,0,255,255,144,0,72,0,253,255,255,255,132,0,117,0,255,255,56,0,131,0,235,255,245,255,251,255,253,255,255,255,102,0,40,0,130,0,253,255,255,255,71,0,116,0,24,0,253,255,255,255,129,0,128,0,253,255,255,255,8,0,86,0,55,0,247,255,251,255,255,255,115,0,255,255,101,0,70,0,255,255,39,0,114,0,251,255,253,255,255,255,100,0,85,0,7,0,23,0,233,255,243,255,251,255,255,255,113,0,255,255,112,0,54,0,253,255,255,255,99,0,69,0,255,255,84,0,38,0,253,255,255,255,98,0,22,0,255,255,97,0,255,255,6,0,96,0,247,255,251,255,255,255,83,0,255,255,53,0,68,0,255,255,37,0,82,0,255,255,81,0,255,255,21,0,5,0,223,255,233,255,243,255,249,255,253,255,255,255,52,0,67,0,255,255,80,0,36,0,253,255,255,255,66,0,51,0,20,0,251,255,255,255,65,0,255,255,4,0,64,0,255,255,35,0,50,0,253,255,255,255,19,0,49,0,253,255,255,255,3,0,48,0,34,0,253,255,255,255,18,0,33,0,255,255,2,0,32,0,253,255,255,255,17,0,1,0,16,0,0,0,61,254,139,255,213,255,231,255,241,255,249,255,253,255,255,255,239,0,254,0,255,255,223,0,253,0,253,255,255,255,207,0,252,0,255,255,191,0,251,0,251,255,255,255,250,0,255,255,175,0,159,0,255,255,249,0,248,0,247,255,251,255,253,255,255,255,143,0,127,0,247,0,255,255,111,0,246,0,253,255,255,255,95,0,245,0,255,255,79,0,244,0,185,255,249,255,253,255,255,255,63,0,243,0,255,255,47,0,242,0,251,255,255,255,241,0,255,255,31,0,240,0,231,255,247,255,255,255,15,0,253,255,255,255,238,0,222,0,255,255,237,0,206,0,249,255,253,255,255,255,236,0,221,0,255,255,190,0,235,0,253,255,255,255,205,0,220,0,255,255,174,0,234,0,241,255,249,255,253,255,255,255,189,0,219,0,255,255,204,0,158,0,253,255,255,255,233,0,173,0,255,255,218,0,188,0,249,255,253,255,255,255,203,0,142,0,255,255,232,0,157,0,253,255,255,255,217,0,126,0,255,255,231,0,172,0,255,0,21,255,113,255,179,255,211,255,231,255,241,255,249,255,253,255,255,255,202,0,187,0,255,255,141,0,216,0,251,255,253,255,255,255,14,0,224,0,13,0,230,0,251,255,253,255,255,255,110,0,156,0,201,0,255,255,94,0,186,0,247,255,251,255,255,255,229,0,255,255,171,0,125,0,255,255,215,0,228,0,253,255,255,255,140,0,200,0,253,255,255,255,78,0,46,0,62,0,241,255,249,255,253,255,255,255,109,0,214,0,255,255,227,0,155,0,253,255,255,255,185,0,170,0,255,255,226,0,30,0,249,255,253,255,255,255,225,0,93,0,255,255,213,0,124,0,253,255,255,255,199,0,77,0,255,255,139,0,184,0,225,255,241,255,249,255,253,255,255,255,212,0,154,0,255,255,169,0,108,0,253,255,255,255,198,0,61,0,255,255,211,0,45,0,249,255,253,255,255,255,210,0,29,0,255,255,123,0,183,0,253,255,255,255,209,0,92,0,255,255,197,0,138,0,239,255,249,255,253,255,255,255,168,0,153,0,255,255,76,0,196,0,253,255,255,255,107,0,182,0,253,255,255,255,208,0,12,0,60,0,249,255,253,255,255,255,195,0,122,0,255,255,167,0,44,0,253,255,255,255,194,0,91,0,255,255,181,0,28,0,199,255,221,255,237,255,249,255,253,255,255,255,137,0,152,0,255,255,193,0,75,0,251,255,253,255,255,255,192,0,11,0,59,0,253,255,255,255,176,0,10,0,26,0,251,255,255,255,180,0,255,255,106,0,166,0,253,255,255,255,121,0,151,0,253,255,255,255,160,0,9,0,144,0,247,255,253,255,255,255,179,0,136,0,253,255,255,255,43,0,90,0,178,0,249,255,253,255,255,255,165,0,27,0,255,255,177,0,105,0,255,255,150,0,164,0,239,255,247,255,251,255,253,255,255,255,74,0,120,0,135,0,255,255,58,0,163,0,253,255,255,255,89,0,149,0,255,255,42,0,162,0,249,255,253,255,255,255,161,0,104,0,255,255,134,0,119,0,253,255,255,255,73,0,148,0,255,255,57,0,147,0,193,255,225,255,241,255,249,255,253,255,255,255,88,0,133,0,255,255,41,0,103,0,253,255,255,255,118,0,146,0,255,255,25,0,145,0,249,255,253,255,255,255,72,0,132,0,255,255,87,0,117,0,253,255,255,255,56,0,131,0,255,255,102,0,40,0,239,255,249,255,253,255,255,255,130,0,24,0,255,255,71,0,116,0,251,255,255,255,129,0,255,255,8,0,128,0,255,255,86,0,101,0,249,255,251,255,255,255,23,0,255,255,7,0,112,0,115,0,253,255,255,255,55,0,39,0,114,0,241,255,249,255,253,255,255,255,70,0,100,0,255,255,85,0,113,0,253,255,255,255,54,0,99,0,255,255,69,0,84,0,249,255,253,255,255,255,38,0,98,0,255,255,22,0,97,0,251,255,253,255,255,255,6,0,96,0,53,0,255,255,83,0,68,0,205,255,219,255,233,255,241,255,247,255,253,255,255,255,37,0,82,0,255,255,21,0,255,255,5,0,80,0,255,255,81,0,255,255,52,0,67,0,253,255,255,255,36,0,66,0,255,255,51,0,20,0,247,255,251,255,255,255,65,0,255,255,4,0,64,0,255,255,35,0,50,0,255,255,19,0,49,0,249,255,251,255,253,255,255,255,3,0,48,0,34,0,18,0,255,255,33,0,255,255,2,0,32,0,253,255,255,255,17,0,1,0,255,255,16,0,0,0,227,255,235,255,243,255,249,255,253,255,255,255,11,0,15,0,255,255,13,0,14,0,253,255,255,255,7,0,5,0,9,0,253,255,255,255,6,0,3,0,255,255,10,0,12,0,253,255,255,255,2,0,1,0,255,255,4,0,8,0,0,0,241,255,249,255,253,255,255,255,15,0,14,0,255,255,13,0,12,0,253,255,255,255,11,0,10,0,255,255,9,0,8,0,249,255,253,255,255,255,7,0,6,0,255,255,5,0,4,0,253,255,255,255,3,0,2,0,255,255,1,0,0,0,176,6,80,6,244,5,160,5,76,5,0,5,184,4,116,4,52,4,248,3,192,3,139,3,0,0,0,0,0,0,0,0,139,3,132,3,126,3,119,3,113,3,107,3,100,3,94,3,88,3,82,3,76,3,70,3,64,3,58,3,52,3,46,3,40,3,34,3,28,3,23,3,17,3,11,3,6,3,0,3,250,2,245,2,240,2,234,2,229,2,224,2,218,2,213,2,208,2,203,2,197,2,192,2,187,2,182,2,177,2,172,2,166,2,163,2,158,2,153,2,148,2,143,2,139,2,134,2,128,2,124,2,120,2,116,2,111,2,107,2,102,2,98,2,92,2,89,2,85,2,80,2,76,2,72,2,68,2,63,2,58,2,55,2,51,2,47,2,43,2,39,2,35,2,31,2,26,2,23,2,20,2,16,2,12,2,8,2,4,2,1,2,252,1,249,1,246,1,242,1,238,1,235,1,231,1,228,1,224,1,221,1,218,1,214,1,211,1,207,1,204,1,201,1,197,1,194,1,191,1,187,1,184,1,181,1,178,1,175,1,176,6,80,6,244,5,160,5,76,5,0,5,184,4,116,4,52,4,248,3,192,3,139,3,164,6,68,6,234,5,150,5,68,5,250,4,178,4,110,4,46,4,242,3,186,3,132,3,152,6,56,6,224,5,138,5,60,5,240,4,170,4,102,4,40,4,236,3,180,3,126,3,140,6,46,6,212,5,128,5,50,5,232,4,160,4,94,4,32,4,228,3,172,3,120,3,128,6,34,6,202,5,118,5,40,5,222,4,152,4,86,4,24,4,222,3,166,3,114,3,116,6,22,6,192,5,108,5,30,5,214,4,144,4,78,4,16,4,214,3,158,3,106,3,104,6,12,6,180,5,98,5,22,5,204,4,136,4,70,4,8,4,206,3,152,3,100,3,92,6,0,6,170,5,88,5,12,5,196,4,126,4,62,4,2,4,200,3,146,3,94,3,22,7,176,6,80,6,244,5,160,5,76,5,0,5,184,4,116,4,52,4,248,3,192,3,8,7,164,6,68,6,234,5,150,5,70,5,248,4,178,4,110,4,46,4,242,3,186,3,252,6,152,6,56,6,224,5,138,5,60,5,240,4,170,4,102,4,40,4,236,3,180,3,238,6,140,6,46,6,212,5,128,5,50,5,232,4,160,4,94,4,32,4,228,3,172,3,226,6,128,6,34,6,202,5,118,5,40,5,222,4,152,4,86,4,24,4,220,3,166,3,214,6,116,6,22,6,192,5,108,5,30,5,214,4,144,4,78,4,16,4,214,3,158,3,200,6,104,6,12,6,180,5,98,5,22,5,204,4,136,4,70,4,8,4,206,3,152,3,188,6,92,6,0,6,170,5,88,5,12,5,196,4,126,4,62,4,2,4,200,3,146,3,176,6,80,6,244,5,160,5,76,5,0,5,184,4,116,4,52,4,248,3,192,3,139,3,88,3,40,3,250,2,208,2,166,2,128,2,92,2,58,2,26,2,252,1,224,1,197,1,172,1,148,1,125,1,104,1,83,1,64,1,46,1,29,1,13,1,254,0,240,0,226,0,214,0,202,0,190,0,180,0,170,0,160,0,151,0,143,0,135,0,127,0,120,0,113,0,107,0,101,0,95,0,90,0,85,0,80,0,75,0,71,0,67,0,63,0,60,0,56,0,53,0,50,0,47,0,45,0,42,0,40,0,37,0,35,0,33,0,31,0,30,0,28,0,215,30,5,31,49,31,110,31,171,31,233,31,40,32,88,32,171,32,221,32,15,33,81,33,133,33,203,33,19,34,53,34,0,0,252,255,248,255,244,255,240,255,236,255,232,255,228,255,224,255,220,255,216,255,212,255,208,255,204,255,200,255,196,255,192,255,188,255,184,255,180,255,176,255,172,255,168,255,164,255,160,255,156,255,152,255,148,255,144,255,140,255,136,255,132,255,127,0,123,0,119,0,115,0,111,0,107,0,103,0,99,0,95,0,91,0,87,0,83,0,79,0,75,0,71,0,67,0,63,0,59,0,55,0,51,0,47,0,43,0,39,0,35,0,31,0,27,0,23,0,19,0,15,0,11,0,7,0,3,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,127,0,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,129,255,98,0,129,255,213,255,88,0,102,0,41,0,191,255,162,255,125,0,20,0,185,255,170,255,186,255,224,255,240,255,160,255,17,0,72,0,107,0,251,255,116,0,187,255,194,255,216,255,10,0,195,255,65,0,109,0,238,255,218,255,243,255,180,255,233,255,88,0,21,0,162,255,8,0,106,0,21,0,144,255,6,0,109,0,20,0,168,255,226,255,9,0,129,255,118,0,42,0,222,255,89,0,252,255,205,255,184,255,21,0,227,255,112,0,123,0,84,0,155,255,164,255,98,0,202,255,161,255,0,0,12,0,25,0,37,0,49,0,60,0,71,0,81,0,90,0,98,0,106,0,112,0,117,0,122,0,125,0,126,0,127,0,126,0,125,0,122,0,117,0,112,0,106,0,98,0,90,0,81,0,71,0,60,0,49,0,37,0,25,0,12,0,0,0,244,255,231,255,219,255,207,255,196,255,185,255,175,255,166,255,158,255,150,255,144,255,139,255,134,255,131,255,130,255,129,255,130,255,131,255,134,255,139,255,144,255,150,255,158,255,166,255,175,255,185,255,196,255,207,255,219,255,231,255,244,255,4,0,4,0,8,1,16,0,8,1,17,0,8,1,18,0,8,1,19,0,0,0,0,0,0,0,1,0,2,0,3,0,4,0,5,0,6,0,7,0,8,0,9,0,10,0,11,0,12,0,13,0,14,0,15,0,16,0,17,0,18,0,19,0,20,0,21,0,22,0,23,0,24,0,25,0,26,0,27,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,40,0,41,0,42,0,43,0,44,0,45,0,46,0,47,0,48,0,49,0,50,0,51,0,52,0,53,0,54,0,55,0,56,0,57,0,58,0,59,0,60,0,61,0,62,0,63,0,64,0,65,0,66,0,67,0,68,0,69,0,70,0,71,0,72,0,73,0,74,0,75,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,88,0,89,0,90,0,91,0,92,0,93,0,94,0,95,0,96,0,97,0,98,0,99,0,100,0,101,0,102,0,103,0,104,0,105,0,106,0,107,0,108,0,109,0,110,0,111,0,112,0,113,0,114,0,115,0,116,0,117,0,118,0,119,0,120,0,121,0,122,0,123,0,124,0,125,0,126,0,127,0,128,0,129,0,130,0,131,0,132,0,133,0,134,0,135,0,136,0,137,0,138,0,139,0,140,0,141,0,142,0,143,0,144,0,145,0,146,0,147,0,148,0,149,0,150,0,151,0,152,0,153,0,154,0,155,0,156,0,157,0,158,0,159,0,160,0,161,0,162,0,163,0,164,0,165,0,166,0,167,0,168,0,169,0,170,0,171,0,172,0,173,0,174,0,175,0,176,0,177,0,178,0,179,0,180,0,181,0,182,0,183,0,184,0,185,0,186,0,187,0,188,0,189,0,190,0,191,0,192,0,193,0,194,0,195,0,196,0,197,0,198,0,199,0,200,0,201,0,202,0,203,0,204,0,205,0,206,0,207,0,208,0,209,0,210,0,211,0,212,0,213,0,214,0,215,0,216,0,217,0,218,0,219,0,220,0,221,0,222,0,223,0,224,0,225,0,226,0,227,0,228,0,0,0,1,0,229,0,230,0,231,0,232,0,233,0,234,0,235,0,236,0,237,0,238,0,13,0,14,0,15,0,99,0,239,0,240,0,241,0,242,0,243,0,244,0,245,0,246,0,247,0,248,0,27,0,28,0,249,0,250,0,251,0,252,0,253,0,254,0,255,0,0,1,1,1,2,1,3,1,4,1,5,1,6,1,7,1,8,1,9,1,10,1,109,0,110,0,11,1,12,1,13,1,14,1,15,1,16,1,17,1,18,1,19,1,20,1,21,1,22,1,23,1,24,1,25,1,26,1,27,1,28,1,29,1,30,1,31,1,32,1,33,1,34,1,35,1,36,1,37,1,38,1,39,1,40,1,41,1,42,1,43,1,44,1,45,1,46,1,47,1,48,1,49,1,50,1,51,1,52,1,53,1,54,1,55,1,56,1,57,1,58,1,59,1,60,1,61,1,62,1,158,0,155,0,163,0,63,1,64,1,65,1,66,1,67,1,68,1,69,1,70,1,150,0,164,0,169,0,71,1,72,1,73,1,74,1,75,1,76,1,77,1,78,1,79,1,80,1,81,1,82,1,83,1,84,1,85,1,86,1,87,1,88,1,89,1,90,1,91,1,92,1,93,1,94,1,95,1,96,1,97,1,98,1,99,1,100,1,101,1,102,1,103,1,104,1,105,1,106,1,107,1,108,1,109,1,110,1,111,1,112,1,113,1,114,1,115,1,116,1,117,1,118,1,119,1,120,1,121,1,122,1,0,0,1,0,231,0,232,0,235,0,236,0,237,0,238,0,13,0,14,0,15,0,99,0,239,0,240,0,241,0,242,0,243,0,244,0,245,0,246,0,247,0,248,0,27,0,28,0,249,0,250,0,251,0,253,0,254,0,255,0,0,1,1,1,2,1,3,1,4,1,5,1,6,1,7,1,8,1,9,1,10,1,109,0,110,0,11,1,12,1,13,1,14,1,16,1,44,1,45,1,46,1,49,1,58,1,59,1,158,0,155,0,163,0,64,1,65,1,66,1,67,1,68,1,69,1,70,1,150,0,164,0,169,0,71,1,72,1,73,1,74,1,75,1,76,1,77,1,78,1,79,1,80,1,81,1,82,1,83,1,84,1,85,1,86,1,87,1,88,1,89,1,90,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,2,0,3,0,4,0,5,0,6,0,7,0,8,0,9,0,10,0,11,0,12,0,13,0,14,0,15,0,16,0,17,0,18,0,19,0,20,0,21,0,22,0,23,0,24,0,25,0,26,0,27,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,40,0,41,0,42,0,43,0,44,0,45,0,46,0,47,0,48,0,49,0,50,0,51,0,52,0,53,0,54,0,55,0,56,0,57,0,58,0,59,0,60,0,61,0,62,0,63,0,64,0,65,0,66,0,67,0,68,0,69,0,70,0,71,0,72,0,73,0,74,0,75,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,88,0,89,0,90,0,91,0,92,0,93,0,94,0,95,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,96,0,97,0,98,0,99,0,100,0,101,0,102,0,103,0,104,0,105,0,106,0,107,0,108,0,109,0,110,0,0,0,111,0,112,0,113,0,114,0,0,0,115,0,116,0,117,0,118,0,119,0,120,0,121,0,122,0,0,0,123,0,0,0,124,0,125,0,126,0,127,0,128,0,129,0,130,0,131,0,0,0,132,0,133,0,0,0,134,0,135,0,136,0,137,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,138,0,0,0,139,0,0,0,0,0,0,0,0,0,140,0,141,0,142,0,143,0,0,0,0,0,0,0,0,0,0,0,144,0,0,0,0,0,0,0,145,0,0,0,0,0,146,0,147,0,148,0,149,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,229,0,230,0,0,0,231,0,232,0,233,0,234,0,235,0,236,0,237,0,238,0,13,0,14,0,15,0,99,0,239,0,240,0,241,0,242,0,243,0,244,0,245,0,246,0,247,0,248,0,27,0,28,0,249,0,250,0,251,0,252,0,0,0,253,0,254,0,255,0,0,1,1,1,0,0,0,0,0,0,2,1,0,0,0,0,3,1,4,1,5,1,6,1,0,0,0,0,7,1,8,1,9,1,0,0,10,1,109,0,110,0,11,1,12,1,13,1,0,0,14,1,15,1,16,1,17,1,18,1,19,1,20,1,21,1,22,1,23,1,24,1,25,1,26,1,27,1,28,1,29,1,30,1,31,1,32,1,33,1,34,1,35,1,36,1,37,1,38,1,39,1,40,1,41,1,42,1,43,1,44,1,45,1,46,1,47,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,48,1,49,1,50,1,0,0,0,0,51,1,52,1,53,1,54,1,55,1,0,0,56,1,0,0,0,0,56,1,0,0,0,0,58,1,59,1,0,0,0,0,60,1,61,1,62,1,0,0,0,0,0,0,158,0,155,0,163,0,63,1,64,1,65,1,66,1,67,1,68,1,69,1,0,0,0,0,70,1,150,0,164,0,169,0,71,1,72,1,73,1,74,1,75,1,76,1,77,1,78,1,79,1,80,1,81,1,82,1,83,1,84,1,85,1,86,1,87,1,88,1,89,1,90,1,91,1,92,1,93,1,94,1,95,1,96,1,97,1,98,1,99,1,100,1,101,1,102,1,103,1,104,1,105,1,106,1,107,1,108,1,109,1,110,1,111,1,112,1,113,1,114,1,115,1,116,1,117,1,118,1,119,1,120,1,121,1,122,1,4,0,8,0,18,4,0,0,18,4,4,0,0,0,0,0,4,0,16,0,18,4,0,0,18,4,4,0,18,4,8,0,18,4,12,0,0,0,0,0,4,0,9,0,19,4,0,0,8,1,4,0,19,4,8,0,0,0,0,0,4,0,9,0,17,4,0,0,8,1,4,0,17,4,8,0,0,0,0,0,4,0,20,0,8,1,0,0,8,1,1,0,8,1,2,0,8,1,3,0,8,1,4,0,8,1,5,0,8,1,6,0,25,1,0,0,17,4,8,0,17,4,12,0,17,4,16,0,0,0,0,0,4,0,20,0,8,1,0,0,8,1,1,0,8,1,2,0,8,1,3,0,8,1,4,0,8,1,5,0,8,1,6,0,25,1,0,0,19,4,8,0,19,4,12,0,19,4,16,0,0,0,0,0,4,0,12,0,13,2,0,0,13,2,2,0,13,2,4,0,13,2,6,0,13,2,8,0,13,2,10,0,0,0,0,0,4,0,12,0,15,2,0,0,15,2,2,0,15,2,4,0,15,2,6,0,15,2,8,0,15,2,10,0,0,0,0,0,4,0,5,0,8,1,0,0,8,1,1,0,8,1,2,0,8,1,3,0,8,1,4,0,0,0,0,0,4,0,58,0,16,4,0,0,12,4,4,0,12,4,8,0,12,4,12,0,12,4,16,0,12,4,20,0,12,4,24,0,20,4,28,0,20,4,32,0,12,4,36,0,20,4,40,0,20,4,44,0,12,4,48,0,20,4,52,0,20,4,56,0,8,4,60,0,8,4,64,0,8,4,68,0,8,4,72,0,8,4,76,0,20,4,80,0,20,4,84,0,20,4,88,0,12,4,92,0,8,4,96,0,8,4,100,0,12,4,104,0,0,0,0,0,253,0,0,0,6,0,5,1,11,1,18,1,27,1,38,1,45,1,53,1,246,2,74,1,84,1,95,1,104,1,109,1,115,1,122,1,129,1,135,1,140,1,144,1,148,1,154,1,159,1,164,1,168,1,174,1,180,1,185,1,191,1,201,1,206,1,212,1,220,1,229,1,232,1,234,1,236,1,238,1,240,1,242,1,244,1,246,1,248,1,250,1,252,1,254,1,0,2,2,2,4,2,6,2,8,2,10,2,12,2,14,2,16,2,18,2,20,2,22,2,24,2,26,2,28,2,40,2,50,2,63,2,75,2,211,3,96,2,98,2,100,2,102,2,104,2,106,2,108,2,110,2,112,2,114,2,116,2,118,2,120,2,122,2,124,2,126,2,128,2,130,2,132,2,134,2,136,2,138,2,140,2,142,2,144,2,146,2,148,2,158,2,162,2,173,2,95,5,112,5,125,5,134,5,206,5,232,5,26,6,67,6,96,6,74,6,86,6,109,6,103,6,116,6,125,6,154,6,132,6,144,6,161,6,190,6,168,6,180,6,197,6,204,6,233,6,211,6,223,6,240,6,254,6,27,7,5,7,17,7,71,3,239,4,195,2,200,2,229,2,113,3,103,3,136,4,22,5,66,5,173,4,217,3,7,4,23,0,62,4,84,4,32,0,195,4,41,0,51,0,218,2,170,4,64,0,76,0,86,0,94,0,97,0,65,4,94,4,106,0,107,4,126,4,198,3,184,2,159,4,112,0,222,2,120,0,132,0,15,3,162,3,177,3,138,0,105,5,118,5,249,5,91,4,133,4,64,3,55,4,2,3,148,3,86,2,63,1,222,4,155,0,41,7,50,6,209,2,237,2,29,3,43,3,58,3,61,3,78,3,88,3,120,3,135,3,186,3,83,5,141,5,76,5,153,5,163,5,170,5,177,5,189,5,199,5,213,5,220,5,163,0,242,5,7,6,14,6,36,6,110,4,223,3,234,3,240,3,247,3,253,3,16,4,21,4,29,4,42,4,49,4,77,4,119,4,0,6,247,6,60,6,51,7,229,4,183,4,39,5,43,6,34,7,205,4,246,4,33,5,43,5,147,4,10,5,52,5,187,4,211,4,252,4,169,0,175,0,182,0,189,0,200,0,209,0,218,0,225,0,232,0,239,0,246,0,253,0,5,1,11,1,18,1,27,1,38,1,45,1,53,1,63,1,74,1,84,1,95,1,104,1,109,1,115,1,122,1,129,1,135,1,140,1,144,1,148,1,154,1,159,1,164,1,168,1,174,1,180,1,185,1,191,1,201,1,206,1,212,1,220,1,229,1,232,1,234,1,236,1,238,1,240,1,242,1,244,1,246,1,248,1,250,1,252,1,254,1,0,2,2,2,4,2,6,2,8,2,10,2,12,2,14,2,16,2,18,2,20,2,22,2,24,2,26,2,28,2,40,2,50,2,63,2,75,2,86,2,96,2,98,2,100,2,102,2,104,2,106,2,108,2,110,2,112,2,114,2,116,2,118,2,120,2,122,2,124,2,126,2,128,2,130,2,132,2,134,2,136,2,138,2,140,2,142,2,144,2,146,2,148,2,158,2,162,2,173,2,184,2,195,2,200,2,209,2,218,2,222,2,229,2,237,2,246,2,2,3,15,3,29,3,43,3,58,3,61,3,64,3,71,3,78,3,88,3,103,3,113,3,120,3,135,3,148,3,162,3,177,3,186,3,198,3,211,3,217,3,223,3,234,3,240,3,247,3,253,3,7,4,16,4,21,4,29,4,42,4,49,4,55,4,62,4,65,4,77,4,84,4,91,4,94,4,107,4,110,4,119,4,126,4,133,4,136,4,147,4,159,4,170,4,173,4,183,4,187,4,195,4,205,4,211,4,222,4,229,4,239,4,246,4,252,4,10,5,22,5,33,5,39,5,43,5,52,5,66,5,76,5,83,5,95,5,105,5,112,5,118,5,125,5,134,5,141,5,153,5,163,5,170,5,177,5,189,5,199,5,206,5,213,5,220,5,232,5,242,5,249,5,0,6,7,6,14,6,26,6,36,6,43,6,50,6,60,6,67,6,74,6,86,6,96,6,103,6,109,6,116,6,125,6,132,6,144,6,154,6,161,6,168,6,180,6,190,6,197,6,204,6,211,6,223,6,233,6,240,6,247,6,254,6,5,7,17,7,27,7,34,7,41,7,51,7,58,7,70,7,88,7,103,7,118,7,133,7,144,7,162,7,181,7,196,7,211,7,224,7,236,7,248,7,6,8,19,8,32,8,44,8,58,8,72,8,85,8,99,8,119,8,134,8,148,8,158,8,168,8,181,8,191,8,201,8,211,8,221,8,231,8,241,8,251,8,5,9,15,9,25,9,28,9,32,9,36,9,54,9,73,9,89,9,104,9,115,9,122,9,129,9,136,9,143,9,150,9,157,9,164,9,171,9,178,9,185,9,192,9,199,9,206,9,213,9,220,9,227,9,234,9,241,9,248,9,255,9,6,10,13,10,20,10,27,10,34,10,41,10,55,10,65,10,72,10,83,10,99,10,112,10,124,10,136,10,148,10,162,10,173,10,184,10,199,10,211,10,222,10,237,10,249,10,3,11,16,11,34,11,44,11,57,11,69,11,82,11,91,11,101,11,114,11,127,11,140,11,152,11,166,11,180,11,193,11,206,11,218,11,230,11,244,11,1,12,14,12,26,12,40,12,54,12,67,12,80,12,95,12,110,12,124,12,136,12,148,12,165,12,177,12,192,12,203,12,211,12,225,12,237,12,249,12,10,13,25,13,37,13,49,13,66,13,81,13,90,13,102,13,114,13,126,13,143,13,155,13,170,13,178,13,190,13,202,13,214,13,231,13,246,13,2,14,13,14,28,14,36,14,44,14,52,14,60,14,66,14,71,14,76,14,82,14,89,14,97,14,103,14,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,229,0,230,0,0,0,231,0,232,0,233,0,234,0,235,0,236,0,237,0,238,0,13,0,14,0,15,0,99,0,239,0,240,0,241,0,242,0,243,0,244,0,245,0,246,0,247,0,248,0,27,0,28,0,249,0,250,0,251,0,252,0,0,0,253,0,254,0,255,0,0,1,1,1,0,0,0,0,0,0,2,1,0,0,0,0,3,1,4,1,5,1,6,1,0,0,0,0,7,1,8,1,9,1,0,0,10,1,109,0,110,0,11,1,12,1,13,1,0,0,14,1,15,1,16,1,17,1,18,1,19,1,20,1,21,1,22,1,23,1,24,1,25,1,26,1,27,1,28,1,29,1,30,1,31,1,32,1,33,1,34,1,35,1,36,1,37,1,38,1,39,1,40,1,41,1,42,1,43,1,44,1,45,1,46,1,47,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,48,1,49,1,50,1,0,0,0,0,51,1,52,1,53,1,54,1,55,1,0,0,56,1,0,0,0,0,57,1,0,0,0,0,58,1,59,1,0,0,0,0,60,1,61,1,62,1,0,0,0,0,0,0,158,0,155,0,163,0,63,1,64,1,65,1,66,1,67,1,68,1,69,1,0,0,0,0,70,1,150,0,164,0,169,0,71,1,72,1,73,1,74,1,75,1,76,1,77,1,78,1,79,1,80,1,81,1,82,1,83,1,84,1,85,1,86,1,87,1,88,1,89,1,90,1,91,1,92,1,93,1,94,1,95,1,96,1,97,1,98,1,99,1,100,1,101,1,102,1,103,1,104,1,105,1,106,1,107,1,108,1,109,1,110,1,111,1,112,1,113,1,114,1,115,1,116,1,117,1,118,1,119,1,120,1,121,1,122,1,4,0,44,0,16,4,0,0,16,4,4,0,16,4,8,0,12,2,12,0,12,2,14,0,16,4,16,0,12,2,20,0,12,2,22,0,16,4,24,0,16,4,28,0,16,4,32,0,16,4,36,0,16,4,40,0,0,0,0,0,4,0,8,0,17,4,4,0,17,4,8,0,0,0,0,0,4,0,54,0,16,4,0,0,16,4,4,0,17,4,8,0,17,4,12,0,12,2,16,0,12,2,18,0,17,4,20,0,17,4,24,0,17,4,28,0,17,4,32,0,13,2,36,0,13,2,38,0,13,2,40,0,13,2,42,0,12,2,44,0,12,2,46,0,13,2,48,0,13,2,50,0,13,2,52,0,0,0,0,0,4,0,36,0,16,4,0,0,13,2,4,0,13,2,6,0,13,2,8,0,12,2,10,0,13,2,12,0,13,2,14,0,13,2,16,0,13,2,18,0,13,2,20,0,13,2,22,0,13,2,24,0,13,2,26,0,13,2,28,0,13,2,30,0,13,2,32,0,12,2,34,0,0,0,0,0,4,0,6,0,17,4,0,0,12,2,4,0,0,0,0,0,4,0,26,0,12,2,6,0,12,2,8,0,12,2,10,0,12,2,12,0,12,2,14,0,12,2,16,0,12,2,18,0,12,2,20,0,12,2,22,0,12,2,24,0,12,2,26,0,12,2,28,0,12,2,30,0,0,0,0,0,4,0,78,0,12,2,0,0,13,2,2,0,12,2,4,0,12,2,6,0,13,2,8,0,13,2,10,0,13,2,12,0,13,2,14,0,13,2,16,0,13,2,18,0,13,2,20,0,13,2,22,0,13,2,24,0,13,2,26,0,13,2,28,0,13,2,30,0,8,1,32,0,8,1,33,0,8,1,34,0,8,1,35,0,8,1,36,0,8,1,37,0,8,1,38,0,8,1,39,0,8,1,40,0,8,1,41,0,16,4,44,0,16,4,48,0,16,4,52,0,16,4,56,0,8,1,60,0,8,1,61,0,8,1,62,0,8,1,63,0,12,2,64,0,12,2,66,0,12,2,68,0,13,2,70,0,13,2,72,0,13,2,74,0,12,2,76,0,12,2,78,0,0,0,0,0,4,0,8,0,16,4,80,0,16,4,84,0,0,0,0,0,4,0,10,0,13,2,88,0,13,2,90,0,12,2,92,0,12,2,94,0,12,2,96,0,0,0,0,0,4,0,4,0,12,2,98,0,12,2,100,0,0,0,0,0,4,0,32,0,16,4,0,0,16,4,4,0,13,2,8,0,13,2,10,0,16,4,12,0,16,4,16,0,16,4,20,0,16,4,24,0,16,4,28,0,0,0,0,0,4,0,6,0,12,2,0,0,12,4,4,0,12,4,8,0,0,0,0,0,12,2,0,0,12,2,2,0,12,2,4,0,12,2,6,0,12,2,8,0,12,4,12,0,0,0,0,0,4,0,54,0,16,4,0,0,16,4,4,0,12,2,8,0,12,2,10,0,12,2,12,0,12,2,14,0,12,2,16,0,12,2,18,0,24,16,20,0,24,8,36,0,24,6,44,0,9,1,50,0,9,1,51,0,8,1,52,0,8,1,53,0,0,0,0,0,4,0,8,0,12,2,4,0,12,2,6,0,12,2,8,0,12,2,10,0,0,0,0,0,4,0,16,0,16,4,0,0,16,4,4,0,16,4,8,0,16,4,12,0,0,0,0,0,4,0,16,0,17,4,0,0,12,2,4,0,12,2,6,0,12,2,8,0,12,2,10,0,12,2,12,0,12,2,14,0,0,0,0,0,4,0,20,0,16,4,0,0,17,4,4,0,17,4,8,0,17,4,12,0,12,2,16,0,12,2,18,0,0,0,0,0,4,0,20,0,17,4,0,0,12,2,4,0,12,2,6,0,16,4,8,0,12,2,12,0,12,2,14,0,16,4,16,0,0,0,0,0,4,0,64,0,14,2,0,0,25,58,0,0,18,2,2,0,0,0,0,0,4,0,40,0,14,2,0,0,25,34,0,0,14,2,2,0,14,2,4,0,0,0,0,0,4,0,148,0,14,2,0,0,18,4,4,0,24,60,8,0,14,2,68,0,14,2,70,0,14,2,72,0,14,2,74,0,14,2,76,0,14,2,78,0,14,2,80,0,8,1,82,0,8,1,83,0,8,1,84,0,14,2,86,0,8,1,88,0,14,2,90,0,14,2,92,0,8,1,94,0,14,2,96,0,14,2,98,0,8,1,100,0,8,1,101,0,8,1,102,0,8,1,103,0,14,2,104,0,18,4,108,0,18,4,112,0,18,4,116,0,18,4,120,0,8,1,124,0,18,4,128,0,14,2,132,0,14,2,134,0,14,2,136,0,18,2,138,0,24,16,140,0,0,0,0,0,4,0,248,0,18,4,0,0,14,2,4,0,14,2,6,0,25,12,0,0,14,2,8,0,25,2,0,0,14,2,10,0,25,110,0,0,18,4,12,0,18,4,16,0,25,104,0,0,0,0,0,0,4,0,40,0,24,8,0,0,25,4,0,0,18,4,8,0,18,4,12,0,18,4,16,0,25,16,0,0,0,0,0,0,4,0,16,0,18,4,0,0,18,4,4,0,14,2,8,0,14,2,10,0,14,2,12,0,14,2,14,0,0,0,0,0,4,0,8,0,18,4,0,0,18,4,4,0,0,0,0,0,4,0,16,0,18,4,0,0,18,4,4,0,18,4,8,0,18,4,12,0,0,0,0,0,8,0,8,128,16,128,16,144,16,0,16,16,32,128,32,144,32,129,32,145,8,128,8,0,16,128,16,144,16,0,16,16,32,128,32,144,32,129,32,145,16,128,16,144,16,0,16,16,32,128,32,144,32,129,32,145,8,0,8,128,16,144,16,128,16,16,16,0,32,144,32,128,32,145,32,129,8,0,8,128,16,0,16,16,16,128,16,144,32,128,32,144,32,129,32,145,8,0,8,128,16,16,16,0,16,144,16,128,32,144,32,128,32,145,32,129,8,0,8,128,32,128,32,144,32,129,32,145,16,128,16,144,16,0,16,16,8,0,8,128,32,144,32,128,32,145,32,129,16,144,16,128,16,16,16,0,8,0,8,128,32,129,32,145,32,128,32,144,16,128,16,144,16,0,16,16,8,0,8,128,32,145,32,129,32,144,32,128,16,144,16,128,16,16,16,0,8,0,8,128,96,7,0,0,0,8,80,0,0,8,16,0,20,8,115,0,18,7,31,0,0,8,112,0,0,8,48,0,0,9,192,0,16,7,10,0,0,8,96,0,0,8,32,0,0,9,160,0,0,8,0,0,0,8,128,0,0,8,64,0,0,9,224,0,16,7,6,0,0,8,88,0,0,8,24,0,0,9,144,0,19,7,59,0,0,8,120,0,0,8,56,0,0,9,208,0,17,7,17,0,0,8,104,0,0,8,40,0,0,9,176,0,0,8,8,0,0,8,136,0,0,8,72,0,0,9,240,0,16,7,4,0,0,8,84,0,0,8,20,0,21,8,227,0,19,7,43,0,0,8,116,0,0,8,52,0,0,9,200,0,17,7,13,0,0,8,100,0,0,8,36,0,0,9,168,0,0,8,4,0,0,8,132,0,0,8,68,0,0,9,232,0,16,7,8,0,0,8,92,0,0,8,28,0,0,9,152,0,20,7,83,0,0,8,124,0,0,8,60,0,0,9,216,0,18,7,23,0,0,8,108,0,0,8,44,0,0,9,184,0,0,8,12,0,0,8,140,0,0,8,76,0,0,9,248,0,16,7,3,0,0,8,82,0,0,8,18,0,21,8,163,0,19,7,35,0,0,8,114,0,0,8,50,0,0,9,196,0,17,7,11,0,0,8,98,0,0,8,34,0,0,9,164,0,0,8,2,0,0,8,130,0,0,8,66,0,0,9,228,0,16,7,7,0,0,8,90,0,0,8,26,0,0,9,148,0,20,7,67,0,0,8,122,0,0,8,58,0,0,9,212,0,18,7,19,0,0,8,106,0,0,8,42,0,0,9,180,0,0,8,10,0,0,8,138,0,0,8,74,0,0,9,244,0,16,7,5,0,0,8,86,0,0,8,22,0,64,8,0,0,19,7,51,0,0,8,118,0,0,8,54,0,0,9,204,0,17,7,15,0,0,8,102,0,0,8,38,0,0,9,172,0,0,8,6,0,0,8,134,0,0,8,70,0,0,9,236,0,16,7,9,0,0,8,94,0,0,8,30,0,0,9,156,0,20,7,99,0,0,8,126,0,0,8,62,0,0,9,220,0,18,7,27,0,0,8,110,0,0,8,46,0,0,9,188,0,0,8,14,0,0,8,142,0,0,8,78,0,0,9,252,0,96,7,0,0,0,8,81,0,0,8,17,0,21,8,131,0,18,7,31,0,0,8,113,0,0,8,49,0,0,9,194,0,16,7,10,0,0,8,97,0,0,8,33,0,0,9,162,0,0,8,1,0,0,8,129,0,0,8,65,0,0,9,226,0,16,7,6,0,0,8,89,0,0,8,25,0,0,9,146,0,19,7,59,0,0,8,121,0,0,8,57,0,0,9,210,0,17,7,17,0,0,8,105,0,0,8,41,0,0,9,178,0,0,8,9,0,0,8,137,0,0,8,73,0,0,9,242,0,16,7,4,0,0,8,85,0,0,8,21,0,16,8,2,1,19,7,43,0,0,8,117,0,0,8,53,0,0,9,202,0,17,7,13,0,0,8,101,0,0,8,37,0,0,9,170,0,0,8,5,0,0,8,133,0,0,8,69,0,0,9,234,0,16,7,8,0,0,8,93,0,0,8,29,0,0,9,154,0,20,7,83,0,0,8,125,0,0,8,61,0,0,9,218,0,18,7,23,0,0,8,109,0,0,8,45,0,0,9,186,0,0,8,13,0,0,8,141,0,0,8,77,0,0,9,250,0,16,7,3,0,0,8,83,0,0,8,19,0,21,8,195,0,19,7,35,0,0,8,115,0,0,8,51,0,0,9,198,0,17,7,11,0,0,8,99,0,0,8,35,0,0,9,166,0,0,8,3,0,0,8,131,0,0,8,67,0,0,9,230,0,16,7,7,0,0,8,91,0,0,8,27,0,0,9,150,0,20,7,67,0,0,8,123,0,0,8,59,0,0,9,214,0,18,7,19,0,0,8,107,0,0,8,43,0,0,9,182,0,0,8,11,0,0,8,139,0,0,8,75,0,0,9,246,0,16,7,5,0,0,8,87,0,0,8,23,0,64,8,0,0,19,7,51,0,0,8,119,0,0,8,55,0,0,9,206,0,17,7,15,0,0,8,103,0,0,8,39,0,0,9,174,0,0,8,7,0,0,8,135,0,0,8,71,0,0,9,238,0,16,7,9,0,0,8,95,0,0,8,31,0,0,9,158,0,20,7,99,0,0,8,127,0,0,8,63,0,0,9,222,0,18,7,27,0,0,8,111,0,0,8,47,0,0,9,190,0,0,8,15,0,0,8,143,0,0,8,79,0,0,9,254,0,96,7,0,0,0,8,80,0,0,8,16,0,20,8,115,0,18,7,31,0,0,8,112,0,0,8,48,0,0,9,193,0,16,7,10,0,0,8,96,0,0,8,32,0,0,9,161,0,0,8,0,0,0,8,128,0,0,8,64,0,0,9,225,0,16,7,6,0,0,8,88,0,0,8,24,0,0,9,145,0,19,7,59,0,0,8,120,0,0,8,56,0,0,9,209,0,17,7,17,0,0,8,104,0,0,8,40,0,0,9,177,0,0,8,8,0,0,8,136,0,0,8,72,0,0,9,241,0,16,7,4,0,0,8,84,0,0,8,20,0,21,8,227,0,19,7,43,0,0,8,116,0,0,8,52,0,0,9,201,0,17,7,13,0,0,8,100,0,0,8,36,0,0,9,169,0,0,8,4,0,0,8,132,0,0,8,68,0,0,9,233,0,16,7,8,0,0,8,92,0,0,8,28,0,0,9,153,0,20,7,83,0,0,8,124,0,0,8,60,0,0,9,217,0,18,7,23,0,0,8,108,0,0,8,44,0,0,9,185,0,0,8,12,0,0,8,140,0,0,8,76,0,0,9,249,0,16,7,3,0,0,8,82,0,0,8,18,0,21,8,163,0,19,7,35,0,0,8,114,0,0,8,50,0,0,9,197,0,17,7,11,0,0,8,98,0,0,8,34,0,0,9,165,0,0,8,2,0,0,8,130,0,0,8,66,0,0,9,229,0,16,7,7,0,0,8,90,0,0,8,26,0,0,9,149,0,20,7,67,0,0,8,122,0,0,8,58,0,0,9,213,0,18,7,19,0,0,8,106,0,0,8,42,0,0,9,181,0,0,8,10,0,0,8,138,0,0,8,74,0,0,9,245,0,16,7,5,0,0,8,86,0,0,8,22,0,64,8,0,0,19,7,51,0,0,8,118,0,0,8,54,0,0,9,205,0,17,7,15,0,0,8,102,0,0,8,38,0,0,9,173,0,0,8,6,0,0,8,134,0,0,8,70,0,0,9,237,0,16,7,9,0,0,8,94,0,0,8,30,0,0,9,157,0,20,7,99,0,0,8,126,0,0,8,62,0,0,9,221,0,18,7,27,0,0,8,110,0,0,8,46,0,0,9,189,0,0,8,14,0,0,8,142,0,0,8,78,0,0,9,253,0,96,7,0,0,0,8,81,0,0,8,17,0,21,8,131,0,18,7,31,0,0,8,113,0,0,8,49,0,0,9,195,0,16,7,10,0,0,8,97,0,0,8,33,0,0,9,163,0,0,8,1,0,0,8,129,0,0,8,65,0,0,9,227,0,16,7,6,0,0,8,89,0,0,8,25,0,0,9,147,0,19,7,59,0,0,8,121,0,0,8,57,0,0,9,211,0,17,7,17,0,0,8,105,0,0,8,41,0,0,9,179,0,0,8,9,0,0,8,137,0,0,8,73,0,0,9,243,0,16,7,4,0,0,8,85,0,0,8,21,0,16,8,2,1,19,7,43,0,0,8,117,0,0,8,53,0,0,9,203,0,17,7,13,0,0,8,101,0,0,8,37,0,0,9,171,0,0,8,5,0,0,8,133,0,0,8,69,0,0,9,235,0,16,7,8,0,0,8,93,0,0,8,29,0,0,9,155,0,20,7,83,0,0,8,125,0,0,8,61,0,0,9,219,0,18,7,23,0,0,8,109,0,0,8,45,0,0,9,187,0,0,8,13,0,0,8,141,0,0,8,77,0,0,9,251,0,16,7,3,0,0,8,83,0,0,8,19,0,21,8,195,0,19,7,35,0,0,8,115,0,0,8,51,0,0,9,199,0,17,7,11,0,0,8,99,0,0,8,35,0,0,9,167,0,0,8,3,0,0,8,131,0,0,8,67,0,0,9,231,0,16,7,7,0,0,8,91,0,0,8,27,0,0,9,151,0,20,7,67,0,0,8,123,0,0,8,59,0,0,9,215,0,18,7,19,0,0,8,107,0,0,8,43,0,0,9,183,0,0,8,11,0,0,8,139,0,0,8,75,0,0,9,247,0,16,7,5,0,0,8,87,0,0,8,23,0,64,8,0,0,19,7,51,0,0,8,119,0,0,8,55,0,0,9,207,0,17,7,15,0,0,8,103,0,0,8,39,0,0,9,175,0,0,8,7,0,0,8,135,0,0,8,71,0,0,9,239,0,16,7,9,0,0,8,95,0,0,8,31,0,0,9,159,0,20,7,99,0,0,8,127,0,0,8,63,0,0,9,223,0,18,7,27,0,0,8,111,0,0,8,47,0,0,9,191,0,0,8,15,0,0,8,143,0,0,8,79,0,0,9,255,0,16,5,1,0,23,5,1,1,19,5,17,0,27,5,1,16,17,5,5,0,25,5,1,4,21,5,65,0,29,5,1,64,16,5,3,0,24,5,1,2,20,5,33,0,28,5,1,32,18,5,9,0,26,5,1,8,22,5,129,0,64,5,0,0,16,5,2,0,23,5,129,1,19,5,25,0,27,5,1,24,17,5,7,0,25,5,1,6,21,5,97,0,29,5,1,96,16,5,4,0,24,5,1,3,20,5,49,0,28,5,1,48,18,5,13,0,26,5,1,12,22,5,193,0,64,5,0,0,16,0,17,0,18,0,0,0,8,0,7,0,9,0,6,0,10,0,5,0,11,0,4,0,12,0,3,0,13,0,2,0,14,0,1,0,15,0,1,0,2,0,3,0,4,0,5,0,7,0,9,0,13,0,17,0,25,0,33,0,49,0,65,0,97,0,129,0,193,0,1,1,129,1,1,2,1,3,1,4,1,6,1,8,1,12,1,16,1,24,1,32,1,48,1,64,1,96,0,0,0,0,3,0,4,0,5,0,6,0,7,0,8,0,9,0,10,0,11,0,13,0,15,0,17,0,19,0,23,0,27,0,31,0,35,0,43,0,51,0,59,0,67,0,83,0,99,0,115,0,131,0,163,0,195,0,227,0,2,1,0,0,0,0,16,0,16,0,16,0,16,0,17,0,17,0,18,0,18,0,19,0,19,0,20,0,20,0,21,0,21,0,22,0,22,0,23,0,23,0,24,0,24,0,25,0,25,0,26,0,26,0,27,0,27,0,28,0,28,0,29,0,29,0,64,0,64,0,16,0,16,0,16,0,16,0,16,0,16,0,16,0,16,0,17,0,17,0,17,0,17,0,18,0,18,0,18,0,18,0,19,0,19,0,19,0,19,0,20,0,20,0,20,0,20,0,21,0,21,0,21,0,21,0,16,0,72,0,78,0,12,0,8,0,140,0,8,0,76,0,8,0,204,0,8,0,44,0,8,0,172,0,8,0,108,0,8,0,236,0,8,0,28,0,8,0,156,0,8,0,92,0,8,0,220,0,8,0,60,0,8,0,188,0,8,0,124,0,8,0,252,0,8,0,2,0,8,0,130,0,8,0,66,0,8,0,194,0,8,0,34,0,8,0,162,0,8,0,98,0,8,0,226,0,8,0,18,0,8,0,146,0,8,0,82,0,8,0,210,0,8,0,50,0,8,0,178,0,8,0,114,0,8,0,242,0,8,0,10,0,8,0,138,0,8,0,74,0,8,0,202,0,8,0,42,0,8,0,170,0,8,0,106,0,8,0,234,0,8,0,26,0,8,0,154,0,8,0,90,0,8,0,218,0,8,0,58,0,8,0,186,0,8,0,122,0,8,0,250,0,8,0,6,0,8,0,134,0,8,0,70,0,8,0,198,0,8,0,38,0,8,0,166,0,8,0,102,0,8,0,230,0,8,0,22,0,8,0,150,0,8,0,86,0,8,0,214,0,8,0,54,0,8,0,182,0,8,0,118,0,8,0,246,0,8,0,14,0,8,0,142,0,8,0,78,0,8,0,206,0,8,0,46,0,8,0,174,0,8,0,110,0,8,0,238,0,8,0,30,0,8,0,158,0,8,0,94,0,8,0,222,0,8,0,62,0,8,0,190,0,8,0,126,0,8,0,254,0,8,0,1,0,8,0,129,0,8,0,65,0,8,0,193,0,8,0,33,0,8,0,161,0,8,0,97,0,8,0,225,0,8,0,17,0,8,0,145,0,8,0,81,0,8,0,209,0,8,0,49,0,8,0,177,0,8,0,113,0,8,0,241,0,8,0,9,0,8,0,137,0,8,0,73,0,8,0,201,0,8,0,41,0,8,0,169,0,8,0,105,0,8,0,233,0,8,0,25,0,8,0,153,0,8,0,89,0,8,0,217,0,8,0,57,0,8,0,185,0,8,0,121,0,8,0,249,0,8,0,5,0,8,0,133,0,8,0,69,0,8,0,197,0,8,0,37,0,8,0,165,0,8,0,101,0,8,0,229,0,8,0,21,0,8,0,149,0,8,0,85,0,8,0,213,0,8,0,53,0,8,0,181,0,8,0,117,0,8,0,245,0,8,0,13,0,8,0,141,0,8,0,77,0,8,0,205,0,8,0,45,0,8,0,173,0,8,0,109,0,8,0,237,0,8,0,29,0,8,0,157,0,8,0,93,0,8,0,221,0,8,0,61,0,8,0,189,0,8,0,125,0,8,0,253,0,8,0,19,0,9,0,19,1,9,0,147,0,9,0,147,1,9,0,83,0,9,0,83,1,9,0,211,0,9,0,211,1,9,0,51,0,9,0,51,1,9,0,179,0,9,0,179,1,9,0,115,0,9,0,115,1,9,0,243,0,9,0,243,1,9,0,11,0,9,0,11,1,9,0,139,0,9,0,139,1,9,0,75,0,9,0,75,1,9,0,203,0,9,0,203,1,9,0,43,0,9,0,43,1,9,0,171,0,9,0,171,1,9,0,107,0,9,0,107,1,9,0,235,0,9,0,235,1,9,0,27,0,9,0,27,1,9,0,155,0,9,0,155,1,9,0,91,0,9,0,91,1,9,0,219,0,9,0,219,1,9,0,59,0,9,0,59,1,9,0,187,0,9,0,187,1,9,0,123,0,9,0,123,1,9,0,251,0,9,0,251,1,9,0,7,0,9,0,7,1,9,0,135,0,9,0,135,1,9,0,71,0,9,0,71,1,9,0,199,0,9,0,199,1,9,0,39,0,9,0,39,1,9,0,167,0,9,0,167,1,9,0,103,0,9,0,103,1,9,0,231,0,9,0,231,1,9,0,23,0,9,0,23,1,9,0,151,0,9,0,151,1,9,0,87,0,9,0,87,1,9,0,215,0,9,0,215,1,9,0,55,0,9,0,55,1,9,0,183,0,9,0,183,1,9,0,119,0,9,0,119,1,9,0,247,0,9,0,247,1,9,0,15,0,9,0,15,1,9,0,143,0,9,0,143,1,9,0,79,0,9,0,79,1,9,0,207,0,9,0,207,1,9,0,47,0,9,0,47,1,9,0,175,0,9,0,175,1,9,0,111,0,9,0,111,1,9,0,239,0,9,0,239,1,9,0,31,0,9,0,31,1,9,0,159,0,9,0,159,1,9,0,95,0,9,0,95,1,9,0,223,0,9,0,223,1,9,0,63,0,9,0,63,1,9,0,191,0,9,0,191,1,9,0,127,0,9,0,127,1,9,0,255,0,9,0,255,1,9,0,0,0,7,0,64,0,7,0,32,0,7,0,96,0,7,0,16,0,7,0,80,0,7,0,48,0,7,0,112,0,7,0,8,0,7,0,72,0,7,0,40,0,7,0,104,0,7,0,24,0,7,0,88,0,7,0,56,0,7,0,120,0,7,0,4,0,7,0,68,0,7,0,36,0,7,0,100,0,7,0,20,0,7,0,84,0,7,0,52,0,7,0,116,0,7,0,3,0,8,0,131,0,8,0,67,0,8,0,195,0,8,0,35,0,8,0,163,0,8,0,99,0,8,0,227,0,8,0,0,0,5,0,16,0,5,0,8,0,5,0,24,0,5,0,4,0,5,0,20,0,5,0,12,0,5,0,28,0,5,0,2,0,5,0,18,0,5,0,10,0,5,0,26,0,5,0,6,0,5,0,22,0,5,0,14,0,5,0,30,0,5,0,1,0,5,0,17,0,5,0,9,0,5,0,25,0,5,0,5,0,5,0,21,0,5,0,13,0,5,0,29,0,5,0,3,0,5,0,19,0,5,0,11,0,5,0,27,0,5,0,7,0,5,0,23,0,5,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,3,32,2,32,2,32,2,32,2,32,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,1,96,4,192,4,192,4,192,4,192,4,192,4,192,4,192,4,192,4,192,4,192,4,192,4,192,4,192,4,192,4,192,8,216,8,216,8,216,8,216,8,216,8,216,8,216,8,216,8,216,8,216,4,192,4,192,4,192,4,192,4,192,4,192,4,192,8,213,8,213,8,213,8,213,8,213,8,213,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,4,192,4,192,4,192,4,192,4,192,4,192,8,214,8,214,8,214,8,214,8,214,8,214,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,4,192,4,192,4,192,4,192,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,86,101,114,115,105,111,110,32,109,105,115,109,97,116,99,104,32,100,101,116,101,99,116,101,100,33,10,76,79,86,69,32,98,105,110,97,114,121,32,105,115,32,118,101,114,115,105,111,110,32,37,115,10,76,79,86,69,32,108,105,98,114,97,114,121,32,105,115,32,118,101,114,115,105,111,110,32,37,115,10,0,45,45,118,101,114,115,105,111,110,0,76,79,86,69,32,37,115,32,40,37,115,41,10,0,101,109,98,101,100,100,101,100,32,98,111,111,116,46,108,117,97,0,95,101,120,101,0,101,114,114,111,114,58,32,37,115,10,0,78,52,108,111,118,101,54,77,111,100,117,108,101,69,0,77,111,100,117,108,101,32,105,110,115,116,97,110,99,101,32,105,115,32,110,117,108,108,0,77,111,100,117,108,101,32,37,115,32,97,108,114,101,97,100,121,32,114,101,103,105,115,116,101,114,101,100,33,0,87,97,114,110,105,110,103,58,32,111,118,101,114,119,114,105,116,105,110,103,32,109,111,100,117,108,101,32,105,110,115,116,97,110,99,101,32,37,115,32,119,105,116,104,32,110,101,119,32,105,110,115,116,97,110,99,101,32,37,115,10,0,78,52,108,111,118,101,54,79,98,106,101,99,116,69,0,48,46,49,48,46,49,0,83,117,112,101,114,32,84,111,97,115,116,0,95,118,101,114,115,105,111,110,0,95,118,101,114,115,105,111,110,95,109,97,106,111,114,0,95,118,101,114,115,105,111,110,95,109,105,110,111,114,0,95,118,101,114,115,105,111,110,95,114,101,118,105,115,105,111,110,0,95,118,101,114,115,105,111,110,95,99,111,100,101,110,97,109,101,0,95,115,101,116,71,97,109,109,97,67,111,114,114,101,99,116,0,48,46,49,48,46,48,0,95,118,101,114,115,105,111,110,95,99,111,109,112,97,116,0,103,101,116,86,101,114,115,105,111,110,0,78,83,116,51,95,95,49,49,56,98,97,115,105,99,95,115,116,114,105,110,103,115,116,114,101,97,109,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,78,83,95,57,97,108,108,111,99,97,116,111,114,73,99,69,69,69,69,0,78,83,116,51,95,95,49,49,53,98,97,115,105,99,95,115,116,114,105,110,103,98,117,102,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,78,83,95,57,97,108,108,111,99,97,116,111,114,73,99,69,69,69,69,0,105,115,86,101,114,115,105,111,110,67,111,109,112,97,116,105,98,108,101,0,95,111,115,0,117,116,102,56,0,108,111,118,101,46,97,117,100,105,111,0,108,111,118,101,46,101,118,101,110,116,0,108,111,118,101,46,102,111,110,116,0,108,111,118,101,46,103,114,97,112,104,105,99,115,0,108,111,118,101,46,105,109,97,103,101,0,108,111,118,101,46,106,111,121,115,116,105,99,107,0,108,111,118,101,46,107,101,121,98,111,97,114,100,0,108,111,118,101,46,109,111,117,115,101,0,108,111,118,101,46,112,104,121,115,105,99,115,0,108,111,118,101,46,115,111,117,110,100,0,108,111,118,101,46,115,121,115,116,101,109,0,108,111,118,101,46,116,105,109,101,114,0,108,111,118,101,46,116,111,117,99,104,0,108,111,118,101,46,118,105,100,101,111,0,108,111,118,101,46,119,105,110,100,111,119,0,108,111,118,101,46,110,111,103,97,109,101,0,45,45,91,91,10,67,111,112,121,114,105,103,104,116,32,40,99,41,32,50,48,48,54,45,50,48,49,54,32,76,79,86,69,32,68,101,118,101,108,111,112,109,101,110,116,32,84,101,97,109,10,84,104,105,115,32,115,111,102,116,119,97,114,101,32,105,115,32,112,114,111,118,105,100,101,100,32,39,97,115,45,105,115,39,44,32,119,105,116,104,111,117,116,32,97,110,121,32,101,120,112,114,101,115,115,32,111,114,32,105,109,112,108,105,101,100,10,119,97,114,114,97,110,116,121,46,32,32,73,110,32,110,111,32,101,118,101,110,116,32,119,105,108,108,32,116,104,101,32,97,117,116,104,111,114,115,32,98,101,32,104,101,108,100,32,108,105,97,98,108,101,32,102,111,114,32,97,110,121,32,100,97,109,97,103,101,115,10,97,114,105,115,105,110,103,32,102,114,111,109,32,116,104,101,32,117,115,101,32,111,102,32,116,104,105,115,32,115,111,102,116,119,97,114,101,46,10,80,101,114,109,105,115,115,105,111,110,32,105,115,32,103,114,97,110,116,101,100,32,116,111,32,97,110,121,111,110,101,32,116,111,32,117,115,101,32,116,104,105,115,32,115,111,102,116,119,97,114,101,32,102,111,114,32,97,110,121,32,112,117,114,112,111,115,101,44,10,105,110,99,108,117,100,105,110,103,32,99,111,109,109,101,114,99,105,97,108,32,97,112,112,108,105,99,97,116,105,111,110,115,44,32,97,110,100,32,116,111,32,97,108,116,101,114,32,105,116,32,97,110,100,32,114,101,100,105,115,116,114,105,98,117,116,101,32,105,116,10,102,114,101,101,108,121,44,32,115,117,98,106,101,99,116,32,116,111,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,114,101,115,116,114,105,99,116,105,111,110,115,58,10,49,46,32,84,104,101,32,111,114,105,103,105,110,32,111,102,32,116,104,105,115,32,115,111,102,116,119,97,114,101,32,109,117,115,116,32,110,111,116,32,98,101,32,109,105,115,114,101,112,114,101,115,101,110,116,101,100,59,32,121,111,117,32,109,117,115,116,32,110,111,116,10,32,32,32,99,108,97,105,109,32,116,104,97,116,32,121,111,117,32,119,114,111,116,101,32,116,104,101,32,111,114,105,103,105,110,97,108,32,115,111,102,116,119,97,114,101,46,32,73,102,32,121,111,117,32,117,115,101,32,116,104,105,115,32,115,111,102,116,119,97,114,101,10,32,32,32,105,110,32,97,32,112,114,111,100,117,99,116,44,32,97,110,32,97,99,107,110,111,119,108,101,100,103,109,101,110,116,32,105,110,32,116,104,101,32,112,114,111,100,117,99,116,32,100,111,99,117,109,101,110,116,97,116,105,111,110,32,119,111,117,108,100,32,98,101,10,32,32,32,97,112,112,114,101,99,105,97,116,101,100,32,98,117,116,32,105,115,32,110,111,116,32,114,101,113,117,105,114,101,100,46,10,50,46,32,65,108,116,101,114,101,100,32,115,111,117,114,99,101,32,118,101,114,115,105,111,110,115,32,109,117,115,116,32,98,101,32,112,108,97,105,110,108,121,32,109,97,114,107,101,100,32,97,115,32,115,117,99,104,44,32,97,110,100,32,109,117,115,116,32,110,111,116,32,98,101,10,32,32,32,109,105,115,114,101,112,114,101,115,101,110,116,101,100,32,97,115,32,98,101,105,110,103,32,116,104,101,32,111,114,105,103,105,110,97,108,32,115,111,102,116,119,97,114,101,46,10,51,46,32,84,104,105,115,32,110,111,116,105,99,101,32,109,97,121,32,110,111,116,32,98,101,32,114,101,109,111,118,101,100,32,111,114,32,97,108,116,101,114,101,100,32,102,114,111,109,32,97,110,121,32,115,111,117,114,99,101,32,100,105,115,116,114,105,98,117,116,105,111,110,46,10,45,45,93,93,10,45,45,32,77,97,107,101,32,115,117,114,101,32,108,111,118,101,32,101,120,105,115,116,115,46,10,108,111,99,97,108,32,108,111,118,101,32,61,32,114,101,113,117,105,114,101,40,34,108,111,118,101,34,41,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,110,111,103,97,109,101,40,41,10,9,108,111,99,97,108,32,109,97,116,104,32,61,32,109,97,116,104,10,9,45,45,32,51,48,108,111,103,46,108,117,97,32,98,101,103,105,110,115,10,9,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,114,101,113,117,105,114,101,95,51,48,108,111,103,40,41,10,9,108,111,99,97,108,32,97,115,115,101,114,116,44,32,112,97,105,114,115,44,32,116,121,112,101,44,32,116,111,115,116,114,105,110,103,44,32,115,101,116,109,101,116,97,116,97,98,108,101,32,61,32,97,115,115,101,114,116,44,32,112,97,105,114,115,44,32,116,121,112,101,44,32,116,111,115,116,114,105,110,103,44,32,115,101,116,109,101,116,97,116,97,98,108,101,10,9,108,111,99,97,108,32,98,97,115,101,77,116,44,32,95,105,110,115,116,97,110,99,101,115,44,32,95,99,108,97,115,115,101,115,44,32,95,99,108,97,115,115,32,61,32,123,125,44,32,115,101,116,109,101,116,97,116,97,98,108,101,40,123,125,44,123,95,95,109,111,100,101,61,39,107,39,125,41,44,32,115,101,116,109,101,116,97,116,97,98,108,101,40,123,125,44,123,95,95,109,111,100,101,61,39,107,39,125,41,10,9,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,97,115,115,101,114,116,95,99,108,97,115,115,40,99,108,97,115,115,44,32,109,101,116,104,111,100,41,32,97,115,115,101,114,116,40,95,99,108,97,115,115,101,115,91,99,108,97,115,115,93,44,32,40,39,87,114,111,110,103,32,109,101,116,104,111,100,32,99,97,108,108,46,32,69,120,112,101,99,116,101,100,32,99,108,97,115,115,58,37,115,46,39,41,58,102,111,114,109,97,116,40,109,101,116,104,111,100,41,41,32,101,110,100,10,9,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,100,101,101,112,95,99,111,112,121,40,116,44,32,100,101,115,116,44,32,97,84,121,112,101,41,32,116,32,61,32,116,32,111,114,32,123,125,59,32,108,111,99,97,108,32,114,32,61,32,100,101,115,116,32,111,114,32,123,125,10,9,32,32,102,111,114,32,107,44,118,32,105,110,32,112,97,105,114,115,40,116,41,32,100,111,10,9,32,32,32,32,105,102,32,97,84,121,112,101,32,97,110,100,32,116,121,112,101,40,118,41,61,61,97,84,121,112,101,32,116,104,101,110,32,114,91,107,93,32,61,32,118,32,101,108,115,101,105,102,32,110,111,116,32,97,84,121,112,101,32,116,104,101,110,10,9,32,32,32,32,32,32,105,102,32,116,121,112,101,40,118,41,32,61,61,32,39,116,97,98,108,101,39,32,97,110,100,32,107,32,126,61,32,34,95,95,105,110,100,101,120,34,32,116,104,101,110,32,114,91,107,93,32,61,32,100,101,101,112,95,99,111,112,121,40,118,41,32,101,108,115,101,32,114,91,107,93,32,61,32,118,32,101,110,100,10,9,32,32,32,32,101,110,100,10,9,32,32,101,110,100,59,32,114,101,116,117,114,110,32,114,10,9,101,110,100,10,9,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,105,110,115,116,97,110,116,105,97,116,101,40,115,101,108,102,44,46,46,46,41,10,9,32,32,97,115,115,101,114,116,95,99,108,97,115,115,40,115,101,108,102,44,32,39,110,101,119,40,46,46,46,41,32,111,114,32,99,108,97,115,115,40,46,46,46,41,39,41,59,32,108,111,99,97,108,32,105,110,115,116,97,110,99,101,32,61,32,123,99,108,97,115,115,32,61,32,115,101,108,102,125,59,32,95,105,110,115,116,97,110,99,101,115,91,105,110,115,116,97,110,99,101,93,32,61,32,116,111,115,116,114,105,110,103,40,105,110,115,116,97,110,99,101,41,59,32,115,101,116,109,101,116,97,116,97,98,108,101,40,105,110,115,116,97,110,99,101,44,115,101,108,102,41,10,9,32,32,105,102,32,115,101,108,102,46,105,110,105,116,32,116,104,101,110,32,105,102,32,116,121,112,101,40,115,101,108,102,46,105,110,105,116,41,32,61,61,32,39,116,97,98,108,101,39,32,116,104,101,110,32,100,101,101,112,95,99,111,112,121,40,115,101,108,102,46,105,110,105,116,44,32,105,110,115,116,97,110,99,101,41,32,101,108,115,101,32,115,101,108,102,46,105,110,105,116,40,105,110,115,116,97,110,99,101,44,32,46,46,46,41,32,101,110,100,59,32,101,110,100,59,32,114,101,116,117,114,110,32,105,110,115,116,97,110,99,101,10,9,101,110,100,10,9,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,101,120,116,101,110,100,40,115,101,108,102,44,32,110,97,109,101,44,32,101,120,116,114,97,95,112,97,114,97,109,115,41,10,9,32,32,97,115,115,101,114,116,95,99,108,97,115,115,40,115,101,108,102,44,32,39,101,120,116,101,110,100,40,46,46,46,41,39,41,59,32,108,111,99,97,108,32,104,101,105,114,32,61,32,123,125,59,32,95,99,108,97,115,115,101,115,91,104,101,105,114,93,32,61,32,116,111,115,116,114,105,110,103,40,104,101,105,114,41,59,32,100,101,101,112,95,99,111,112,121,40,101,120,116,114,97,95,112,97,114,97,109,115,44,32,100,101,101,112,95,99,111,112,121,40,115,101,108,102,44,32,104,101,105,114,41,41,59,10,9,32,32,104,101,105,114,46,110,97,109,101,44,32,104,101,105,114,46,95,95,105,110,100,101,120,44,32,104,101,105,114,46,115,117,112,101,114,32,61,32,101,120,116,114,97,95,112,97,114,97,109,115,32,97,110,100,32,101,120,116,114,97,95,112,97,114,97,109,115,46,110,97,109,101,32,111,114,32,110,97,109,101,44,32,104,101,105,114,44,32,115,101,108,102,59,32,114,101,116,117,114,110,32,115,101,116,109,101,116,97,116,97,98,108,101,40,104,101,105,114,44,115,101,108,102,41,10,9,101,110,100,10,9,98,97,115,101,77,116,32,61,32,123,32,95,95,99,97,108,108,32,61,32,102,117,110,99,116,105,111,110,32,40,115,101,108,102,44,46,46,46,41,32,114,101,116,117,114,110,32,115,101,108,102,58,110,101,119,40,46,46,46,41,32,101,110,100,44,32,95,95,116,111,115,116,114,105,110,103,32,61,32,102,117,110,99,116,105,111,110,40,115,101,108,102,44,46,46,46,41,10,9,32,32,105,102,32,95,105,110,115,116,97,110,99,101,115,91,115,101,108,102,93,32,116,104,101,110,32,114,101,116,117,114,110,32,40,34,105,110,115,116,97,110,99,101,32,111,102,32,39,37,115,39,32,40,37,115,41,34,41,58,102,111,114,109,97,116,40,114,97,119,103,101,116,40,115,101,108,102,46,99,108,97,115,115,44,39,110,97,109,101,39,41,32,111,114,32,39,63,39,44,32,95,105,110,115,116,97,110,99,101,115,91,115,101,108,102,93,41,32,101,110,100,10,9,32,32,114,101,116,117,114,110,32,95,99,108,97,115,115,101,115,91,115,101,108,102,93,32,97,110,100,32,40,34,99,108,97,115,115,32,39,37,115,39,32,40,37,115,41,34,41,58,102,111,114,109,97,116,40,114,97,119,103,101,116,40,115,101,108,102,44,39,110,97,109,101,39,41,32,111,114,32,39,63,39,44,95,99,108,97,115,115,101,115,91,115,101,108,102,93,41,32,111,114,32,115,101,108,102,10,9,101,110,100,125,59,32,95,99,108,97,115,115,101,115,91,98,97,115,101,77,116,93,32,61,32,116,111,115,116,114,105,110,103,40,98,97,115,101,77,116,41,59,32,115,101,116,109,101,116,97,116,97,98,108,101,40,98,97,115,101,77,116,44,32,123,95,95,116,111,115,116,114,105,110,103,32,61,32,98,97,115,101,77,116,46,95,95,116,111,115,116,114,105,110,103,125,41,10,9,108,111,99,97,108,32,99,108,97,115,115,32,61,32,123,105,115,67,108,97,115,115,32,61,32,102,117,110,99,116,105,111,110,40,99,108,97,115,115,44,32,111,102,115,117,112,101,114,41,32,108,111,99,97,108,32,105,115,99,108,97,115,115,32,61,32,110,111,116,32,110,111,116,32,95,99,108,97,115,115,101,115,91,99,108,97,115,115,93,59,32,105,102,32,111,102,115,117,112,101,114,32,116,104,101,110,32,114,101,116,117,114,110,32,105,115,99,108,97,115,115,32,97,110,100,32,40,99,108,97,115,115,46,115,117,112,101,114,32,61,61,32,111,102,115,117,112,101,114,41,32,101,110,100,59,32,114,101,116,117,114,110,32,105,115,99,108,97,115,115,32,101,110,100,44,32,105,115,73,110,115,116,97,110,99,101,32,61,32,102,117,110,99,116,105,111,110,40,105,110,115,116,97,110,99,101,44,32,111,102,99,108,97,115,115,41,32,10,9,32,32,32,32,108,111,99,97,108,32,105,115,105,110,115,116,97,110,99,101,32,61,32,110,111,116,32,110,111,116,32,95,105,110,115,116,97,110,99,101,115,91,105,110,115,116,97,110,99,101,93,59,32,105,102,32,111,102,99,108,97,115,115,32,116,104,101,110,32,114,101,116,117,114,110,32,105,115,105,110,115,116,97,110,99,101,32,97,110,100,32,40,105,110,115,116,97,110,99,101,46,99,108,97,115,115,32,61,61,32,111,102,99,108,97,115,115,41,32,101,110,100,59,32,114,101,116,117,114,110,32,105,115,105,110,115,116,97,110,99,101,32,101,110,100,125,59,32,95,99,108,97,115,115,32,61,32,102,117,110,99,116,105,111,110,40,110,97,109,101,44,32,97,116,116,114,41,10,9,32,32,108,111,99,97,108,32,99,32,61,32,100,101,101,112,95,99,111,112,121,40,97,116,116,114,41,59,32,99,46,109,105,120,105,110,115,61,115,101,116,109,101,116,97,116,97,98,108,101,40,123,125,44,123,95,95,109,111,100,101,61,39,107,39,125,41,59,32,95,99,108,97,115,115,101,115,91,99,93,32,61,32,116,111,115,116,114,105,110,103,40,99,41,59,32,99,46,110,97,109,101,44,32,99,46,95,95,116,111,115,116,114,105,110,103,44,32,99,46,95,95,99,97,108,108,32,61,32,110,97,109,101,32,111,114,32,99,46,110,97,109,101,44,32,98,97,115,101,77,116,46,95,95,116,111,115,116,114,105,110,103,44,32,98,97,115,101,77,116,46,95,95,99,97,108,108,10,9,32,32,99,46,105,110,99,108,117,100,101,32,61,32,102,117,110,99,116,105,111,110,40,115,101,108,102,44,109,105,120,105,110,41,32,97,115,115,101,114,116,95,99,108,97,115,115,40,115,101,108,102,44,32,39,105,110,99,108,117,100,101,40,109,105,120,105,110,41,39,41,59,32,115,101,108,102,46,109,105,120,105,110,115,91,109,105,120,105,110,93,32,61,32,116,114,117,101,59,32,114,101,116,117,114,110,32,100,101,101,112,95,99,111,112,121,40,109,105,120,105,110,44,32,115,101,108,102,44,32,39,102,117,110,99,116,105,111,110,39,41,32,101,110,100,10,9,32,32,99,46,110,101,119,44,32,99,46,101,120,116,101,110,100,44,32,99,46,95,95,105,110,100,101,120,44,32,99,46,105,110,99,108,117,100,101,115,32,61,32,105,110,115,116,97,110,116,105,97,116,101,44,32,101,120,116,101,110,100,44,32,99,44,32,102,117,110,99,116,105,111,110,40,115,101,108,102,44,109,105,120,105,110,41,32,97,115,115,101,114,116,95,99,108,97,115,115,40,115,101,108,102,44,39,105,110,99,108,117,100,101,115,40,109,105,120,105,110,41,39,41,32,114,101,116,117,114,110,32,110,111,116,32,110,111,116,32,40,115,101,108,102,46,109,105,120,105,110,115,91,109,105,120,105,110,93,32,111,114,32,40,115,101,108,102,46,115,117,112,101,114,32,97,110,100,32,115,101,108,102,46,115,117,112,101,114,58,105,110,99,108,117,100,101,115,40,109,105,120,105,110,41,41,41,32,101,110,100,10,9,32,32,99,46,101,120,116,101,110,100,115,32,61,32,102,117,110,99,116,105,111,110,40,115,101,108,102,44,32,99,108,97,115,115,41,32,97,115,115,101,114,116,95,99,108,97,115,115,40,115,101,108,102,44,32,39,101,120,116,101,110,100,115,40,99,108,97,115,115,41,39,41,32,108,111,99,97,108,32,115,117,112,101,114,32,61,32,115,101,108,102,59,32,114,101,112,101,97,116,32,115,117,112,101,114,32,61,32,115,117,112,101,114,46,115,117,112,101,114,32,117,110,116,105,108,32,40,115,117,112,101,114,32,61,61,32,99,108,97,115,115,32,111,114,32,115,117,112,101,114,32,61,61,32,110,105,108,41,59,32,114,101,116,117,114,110,32,99,108,97,115,115,32,97,110,100,32,40,115,117,112,101,114,32,61,61,32,99,108,97,115,115,41,32,101,110,100,10,9,32,32,32,32,114,101,116,117,114,110,32,115,101,116,109,101,116,97,116,97,98,108,101,40,99,44,32,98,97,115,101,77,116,41,32,101,110,100,59,32,99,108,97,115,115,46,95,68,69,83,67,82,73,80,84,73,79,78,32,61,32,39,51,48,32,108,105,110,101,115,32,108,105,98,114,97,114,121,32,102,111,114,32,111,98,106,101,99,116,32,111,114,105,101,110,116,97,116,105,111,110,32,105,110,32,76,117,97,39,59,32,99,108,97,115,115,46,95,86,69,82,83,73,79,78,32,61,32,39,51,48,108,111,103,32,118,49,46,48,46,48,39,59,32,99,108,97,115,115,46,95,85,82,76,32,61,32,39,104,116,116,112,58,47,47,103,105,116,104,117,98,46,99,111,109,47,89,111,110,97,98,97,47,51,48,108,111,103,39,59,32,99,108,97,115,115,46,95,76,73,67,69,78,83,69,32,61,32,39,77,73,84,32,76,73,67,69,78,83,69,32,60,104,116,116,112,58,47,47,119,119,119,46,111,112,101,110,115,111,117,114,99,101,46,111,114,103,47,108,105,99,101,110,115,101,115,47,109,105,116,45,108,105,99,101,110,115,101,46,112,104,112,62,39,10,9,114,101,116,117,114,110,32,115,101,116,109,101,116,97,116,97,98,108,101,40,99,108,97,115,115,44,123,95,95,99,97,108,108,32,61,32,102,117,110,99,116,105,111,110,40,95,44,46,46,46,41,32,114,101,116,117,114,110,32,95,99,108,97,115,115,40,46,46,46,41,32,101,110,100,32,125,41,10,9,101,110,100,10,9,45,45,32,51,48,108,111,103,46,108,117,97,32,101,110,100,115,10,9,108,111,99,97,108,32,109,111,115,97,105,99,95,112,110,103,32,61,10,9,34,105,86,66,79,82,119,48,75,71,103,111,65,65,65,65,78,83,85,104,69,85,103,65,65,65,73,65,65,65,65,67,65,67,65,89,65,65,65,68,68,80,109,72,76,65,65,65,65,71,88,82,70,87,72,82,84,98,50,90,48,100,50,70,121,90,81,66,66,90,71,57,105,92,10,9,90,83,66,74,98,87,70,110,90,86,74,108,89,87,82,53,99,99,108,108,80,65,65,65,65,50,90,112,86,70,104,48,87,69,49,77,79,109,78,118,98,83,53,104,90,71,57,105,90,83,53,52,98,88,65,65,65,65,65,65,65,68,119,47,101,72,66,104,89,50,116,108,92,10,9,100,67,66,105,90,87,100,112,98,106,48,105,55,55,117,47,73,105,66,112,90,68,48,105,86,122,86,78,77,69,49,119,81,50,86,111,97,85,104,54,99,109,86,84,101,107,53,85,89,51,112,114,89,122,108,107,73,106,56,43,73,68,120,52,79,110,104,116,99,71,49,108,92,10,9,100,71,69,103,101,71,49,115,98,110,77,54,101,68,48,105,89,87,82,118,89,109,85,54,98,110,77,54,98,87,86,48,89,83,56,105,73,72,103,54,101,71,49,119,100,71,115,57,73,107,70,107,98,50,74,108,73,70,104,78,85,67,66,68,98,51,74,108,73,68,85,117,92,10,9,77,121,49,106,77,68,69,120,73,68,89,50,76,106,69,48,78,84,89,50,77,83,119,103,77,106,65,120,77,105,56,119,77,105,56,119,78,105,48,120,78,68,111,49,78,106,111,121,78,121,65,103,73,67,65,103,73,67,65,103,73,106,52,103,80,72,74,107,90,106,112,83,92,10,9,82,69,89,103,101,71,49,115,98,110,77,54,99,109,82,109,80,83,74,111,100,72,82,119,79,105,56,118,100,51,100,51,76,110,99,122,76,109,57,121,90,121,56,120,79,84,107,53,76,122,65,121,76,122,73,121,76,88,74,107,90,105,49,122,101,87,53,48,89,88,103,116,92,10,9,98,110,77,106,73,106,52,103,80,72,74,107,90,106,112,69,90,88,78,106,99,109,108,119,100,71,108,118,98,105,66,121,90,71,89,54,89,87,74,118,100,88,81,57,73,105,73,103,101,71,49,115,98,110,77,54,101,71,49,119,84,85,48,57,73,109,104,48,100,72,65,54,92,10,9,76,121,57,117,99,121,53,104,90,71,57,105,90,83,53,106,98,50,48,118,101,71,70,119,76,122,69,117,77,67,57,116,98,83,56,105,73,72,104,116,98,71,53,122,79,110,78,48,85,109,86,109,80,83,74,111,100,72,82,119,79,105,56,118,98,110,77,117,89,87,82,118,92,10,9,89,109,85,117,89,50,57,116,76,51,104,104,99,67,56,120,76,106,65,118,99,49,82,53,99,71,85,118,85,109,86,122,98,51,86,121,89,50,86,83,90,87,89,106,73,105,66,52,98,87,120,117,99,122,112,52,98,88,65,57,73,109,104,48,100,72,65,54,76,121,57,117,92,10,9,99,121,53,104,90,71,57,105,90,83,53,106,98,50,48,118,101,71,70,119,76,122,69,117,77,67,56,105,73,72,104,116,99,69,49,78,79,107,57,121,97,87,100,112,98,109,70,115,82,71,57,106,100,87,49,108,98,110,82,74,82,68,48,105,101,71,49,119,76,109,82,112,92,10,9,90,68,111,50,78,68,89,51,79,68,85,50,79,84,107,48,78,107,74,70,78,84,69,120,81,84,103,51,82,107,81,51,77,84,78,67,79,84,99,50,78,48,85,122,78,121,73,103,101,71,49,119,84,85,48,54,82,71,57,106,100,87,49,108,98,110,82,74,82,68,48,105,92,10,9,101,71,49,119,76,109,82,112,90,68,111,53,79,68,107,119,78,48,82,70,82,84,103,51,82,85,69,120,77,85,85,49,81,84,82,66,77,84,107,50,78,68,74,69,81,85,89,121,82,107,85,119,78,121,73,103,101,71,49,119,84,85,48,54,83,87,53,122,100,71,70,117,92,10,9,89,50,86,74,82,68,48,105,101,71,49,119,76,109,108,112,90,68,111,53,79,68,107,119,78,48,82,70,82,68,103,51,82,85,69,120,77,85,85,49,81,84,82,66,77,84,107,50,78,68,74,69,81,85,89,121,82,107,85,119,78,121,73,103,101,71,49,119,79,107,78,121,92,10,9,90,87,70,48,98,51,74,85,98,50,57,115,80,83,74,66,90,71,57,105,90,83,66,81,97,71,57,48,98,51,78,111,98,51,65,103,81,49,77,50,73,67,104,88,97,87,53,107,98,51,100,122,75,83,73,43,73,68,120,52,98,88,66,78,84,84,112,69,90,88,74,112,92,10,9,100,109,86,107,82,110,74,118,98,83,66,122,100,70,74,108,90,106,112,112,98,110,78,48,89,87,53,106,90,85,108,69,80,83,74,52,98,88,65,117,97,87,108,107,79,106,66,71,81,84,73,122,81,85,73,49,82,84,99,52,78,48,85,49,77,84,70,66,78,68,85,121,92,10,9,79,68,99,50,78,106,70,68,82,106,77,51,79,84,73,119,73,105,66,122,100,70,74,108,90,106,112,107,98,50,78,49,98,87,86,117,100,69,108,69,80,83,74,52,98,88,65,117,90,71,108,107,79,106,89,48,78,106,99,52,78,84,89,53,79,84,81,50,81,107,85,49,92,10,9,77,84,70,66,79,68,100,71,82,68,99,120,77,48,73,53,78,122,89,51,82,84,77,51,73,105,56,43,73,68,119,118,99,109,82,109,79,107,82,108,99,50,78,121,97,88,66,48,97,87,57,117,80,105,65,56,76,51,74,107,90,106,112,83,82,69,89,43,73,68,119,118,92,10,9,101,68,112,52,98,88,66,116,90,88,82,104,80,105,65,56,80,51,104,119,89,87,78,114,90,88,81,103,90,87,53,107,80,83,74,121,73,106,56,43,71,67,87,72,82,119,65,65,71,84,82,74,82,69,70,85,101,78,114,115,88,81,101,99,70,100,88,86,118,43,119,117,92,10,9,84,88,113,86,73,114,65,75,119,81,75,75,81,82,65,107,113,73,65,78,66,90,85,89,82,82,80,68,70,56,69,97,73,116,105,67,121,97,101,67,83,79,119,74,102,69,82,82,104,66,104,85,89,113,120,69,77,83,111,67,83,105,106,83,81,83,67,73,73,67,113,119,92,10,9,75,75,68,65,119,116,76,76,76,106,105,53,102,57,57,53,89,90,105,57,53,100,119,112,68,122,57,53,53,47,99,55,118,52,85,51,99,56,118,99,79,88,80,118,54,97,101,77,53,51,107,105,67,48,99,118,53,71,83,88,73,69,115,65,87,84,105,75,73,99,57,119,92,10,9,55,81,109,74,80,53,78,89,83,87,73,49,83,122,57,55,74,74,90,73,76,80,98,57,101,55,102,104,55,51,88,102,52,122,85,112,75,55,71,51,120,65,115,108,49,112,83,52,87,117,76,122,69,109,99,54,57,111,79,50,57,48,109,56,87,109,73,68,105,90,115,108,92,10,9,106,112,73,52,87,79,74,66,104,55,108,48,108,116,104,82,89,105,117,74,122,83,88,87,68,55,121,80,55,82,76,88,83,47,120,67,52,110,121,74,48,121,88,79,108,118,103,116,97,119,84,119,65,66,112,99,54,105,85,68,101,119,120,106,72,109,109,115,76,72,71,79,92,10,9,90,116,52,68,72,80,113,112,76,51,71,86,112,112,57,72,71,79,50,114,83,110,120,87,52,114,54,81,97,47,121,49,120,68,57,73,114,71,99,98,121,51,82,120,88,85,73,69,115,67,72,69,105,56,71,67,68,112,71,52,85,79,74,43,105,87,115,107,110,113,67,52,92,10,9,114,54,51,69,50,121,88,50,107,116,103,56,120,68,105,68,68,102,77,43,73,76,69,120,115,53,43,120,104,110,54,50,83,99,119,122,116,71,48,103,115,83,67,109,116,100,53,74,104,74,115,84,104,103,68,50,74,69,81,65,110,122,105,43,108,80,52,83,57,121,114,54,92,10,9,109,82,71,52,114,55,51,69,52,115,65,57,73,74,103,76,72,77,97,97,97,90,110,55,122,53,110,57,98,76,84,48,48,57,76,81,57,109,56,74,114,80,109,55,69,111,57,82,106,97,100,106,65,105,116,73,114,66,106,52,55,99,117,89,122,116,103,105,104,51,115,72,92,10,9,83,102,119,122,122,83,99,73,90,48,118,115,53,118,118,47,85,68,111,122,47,88,67,71,120,77,107,83,66,122,76,72,113,50,71,53,102,106,67,109,78,102,105,120,52,86,114,51,66,80,105,97,83,121,83,43,114,101,76,53,99,103,119,77,106,66,56,101,108,116,104,69,92,10,9,52,106,48,90,74,73,66,87,82,65,65,109,54,69,100,47,84,53,80,89,119,88,68,102,69,73,107,88,77,56,97,115,90,76,110,43,70,88,80,117,110,49,105,117,116,122,82,99,113,53,52,81,99,57,116,70,52,114,49,99,65,103,103,117,53,105,98,54,117,121,97,68,92,10,9,66,78,66,102,89,113,55,108,110,103,115,107,49,112,88,52,85,48,90,47,100,122,80,117,113,87,67,53,47,106,86,122,55,104,57,90,114,112,57,48,104,67,83,99,51,57,78,54,97,99,88,65,88,78,111,117,103,49,118,109,52,55,84,100,110,112,86,66,65,114,105,99,92,10,9,113,99,102,111,83,83,75,98,68,84,112,107,107,65,65,87,87,54,52,51,106,98,66,43,43,50,109,72,50,85,74,105,78,85,84,67,48,120,109,55,86,47,114,53,43,107,104,56,86,67,85,71,49,112,81,52,120,85,115,101,72,109,65,121,85,108,121,89,82,82,119,54,92,10,9,66,50,120,106,70,104,118,97,70,106,111,119,107,54,48,116,56,57,104,107,97,76,118,88,48,110,97,52,111,107,48,117,77,98,117,102,77,116,90,103,113,111,52,74,110,69,122,110,82,78,75,119,108,88,110,102,66,117,90,57,72,82,104,72,66,87,65,106,85,47,71,105,92,10,9,103,50,56,99,110,116,70,50,98,120,51,68,116,82,50,87,116,112,85,49,122,79,110,55,69,115,57,104,114,79,56,112,79,104,55,103,76,116,73,111,74,81,51,99,73,50,66,54,122,79,77,117,106,100,106,101,90,87,48,79,82,66,106,72,100,115,119,48,78,70,119,68,92,10,9,114,47,97,54,112,88,49,100,72,81,70,77,74,43,111,89,47,84,48,104,103,70,100,105,72,110,100,117,120,80,90,55,72,101,52,57,49,113,97,65,106,98,66,55,78,74,75,89,84,121,114,104,102,74,73,111,119,74,116,100,81,98,122,97,112,83,54,105,98,74,53,105,92,10,9,43,55,109,90,113,80,68,43,104,65,105,103,107,72,110,102,79,120,74,88,83,106,119,120,112,110,71,110,82,71,120,102,120,117,72,101,106,104,71,79,67,78,118,82,49,121,75,105,78,76,97,79,73,119,90,117,83,51,65,72,52,66,73,65,106,66,109,80,120,84,84,109,92,10,9,90,113,89,120,90,55,118,104,87,106,51,109,87,70,106,84,118,112,90,55,49,115,81,103,97,89,83,70,71,82,119,67,113,74,98,103,66,68,89,54,51,80,115,51,104,108,75,70,65,54,56,121,116,88,105,109,76,55,79,53,52,74,110,80,66,49,107,85,80,89,65,70,92,10,9,77,83,105,98,119,115,74,73,68,103,71,85,83,87,106,119,69,103,97,88,71,50,83,109,98,111,116,104,51,76,72,77,43,122,54,122,97,79,103,117,115,107,103,81,81,53,108,72,53,122,43,80,69,65,72,103,53,99,47,104,69,77,67,79,104,67,90,81,71,75,73,78,92,10,9,122,117,54,88,73,53,55,57,67,53,106,51,76,114,74,99,72,48,87,105,108,104,57,103,77,43,108,70,85,115,97,100,106,68,72,65,49,48,119,49,88,80,56,121,111,98,88,72,82,57,66,80,100,86,55,112,122,116,56,107,89,70,80,73,100,114,101,70,74,66,55,65,92,10,9,69,73,100,55,74,122,77,52,99,69,104,76,97,43,110,118,65,112,75,55,115,98,104,99,57,101,55,118,76,70,74,65,51,68,115,65,109,76,53,102,69,90,89,54,66,118,77,105,75,109,115,121,82,81,66,103,52,109,54,85,43,75,90,106,117,53,101,67,84,73,56,70,92,10,9,115,68,51,67,65,43,103,69,121,51,50,78,67,86,48,66,80,77,49,98,108,110,118,65,103,79,56,84,100,114,87,48,84,100,83,101,81,106,113,66,56,83,76,108,113,97,85,69,72,81,69,85,102,77,56,73,81,78,67,68,80,67,80,120,70,117,98,57,89,79,106,117,92,10,9,99,66,119,68,88,121,89,115,110,56,56,109,56,79,122,89,57,109,57,121,89,74,83,98,87,78,90,120,106,111,43,110,50,107,111,55,120,50,114,105,89,122,54,51,55,68,75,43,74,49,98,114,111,54,116,74,51,74,87,65,72,101,67,104,105,67,53,98,70,83,81,117,92,10,9,89,111,119,68,110,102,52,53,73,99,101,65,88,118,51,68,109,74,47,55,72,122,82,51,55,104,121,87,87,80,113,98,69,74,99,98,88,74,53,66,72,115,89,87,66,52,101,75,52,48,110,114,66,79,49,87,86,82,73,82,50,52,113,85,47,110,48,88,47,98,56,97,92,10,9,85,48,83,75,117,114,78,103,97,52,83,106,54,107,74,82,50,109,102,66,122,55,47,48,99,116,122,54,103,53,113,121,97,43,105,77,98,120,112,120,118,118,104,67,52,82,106,54,74,48,101,43,97,114,118,108,101,113,87,52,116,105,87,84,86,122,67,50,108,102,99,49,92,10,9,49,53,54,110,115,43,112,50,43,106,47,43,119,110,119,76,117,51,119,86,79,114,56,113,69,99,71,107,47,52,75,65,80,111,104,104,122,109,116,111,110,73,109,105,116,78,100,83,67,98,50,56,78,121,75,79,115,90,54,52,47,100,99,107,47,105,82,107,72,121,67,103,92,10,9,47,105,75,99,68,99,74,71,65,71,85,122,81,81,65,109,113,69,105,99,99,66,113,113,48,75,83,51,74,97,120,70,57,71,117,122,76,113,73,88,88,100,115,110,117,108,52,109,52,106,77,105,81,83,88,98,105,99,55,116,103,81,54,97,81,72,119,48,119,121,86,79,92,10,9,105,68,68,50,80,71,72,87,54,83,43,78,97,121,72,68,66,111,98,85,68,101,103,75,75,116,72,50,110,69,110,52,85,79,75,112,69,108,56,107,112,117,99,115,69,98,56,70,69,89,113,111,112,43,107,52,55,69,70,106,114,102,75,74,85,119,101,73,54,82,112,80,92,10,9,77,106,97,79,121,119,115,106,118,110,122,65,107,121,76,108,53,54,47,84,112,84,120,50,112,72,101,65,52,119,73,55,81,74,77,69,70,82,103,50,84,114,57,88,66,115,89,112,73,102,72,116,114,81,119,57,70,122,54,117,57,105,73,86,82,72,73,86,72,90,56,55,92,10,9,74,98,52,110,85,110,53,57,97,52,52,48,65,84,120,72,50,49,81,97,121,104,107,111,78,103,118,104,89,65,118,116,75,118,50,83,72,75,82,77,78,106,114,52,54,73,90,115,99,71,105,87,65,72,55,81,65,71,101,83,48,100,110,88,102,80,81,82,81,65,55,112,92,10,9,74,104,90,72,52,72,80,43,80,122,49,114,88,116,106,110,53,68,83,67,48,103,70,87,77,67,104,102,122,104,85,112,118,56,71,71,57,80,116,117,52,118,53,104,82,111,86,105,53,108,56,105,112,85,67,67,101,72,83,107,109,65,117,73,89,109,78,73,104,103,100,56,92,10,9,57,65,78,53,48,82,86,112,82,122,116,84,112,69,75,57,111,65,99,74,71,111,120,103,57,80,109,51,83,73,87,72,118,48,71,105,115,118,107,57,71,80,84,69,53,83,110,103,56,104,48,72,110,102,99,66,117,118,56,56,67,114,85,117,107,56,72,81,98,111,120,49,92,10,9,107,56,75,71,48,84,97,109,118,116,116,84,57,80,66,107,105,112,119,117,111,102,54,47,112,87,68,81,97,82,83,83,102,86,114,77,122,57,87,65,111,111,48,80,104,76,66,66,102,67,72,120,120,106,68,82,119,81,48,108,106,111,120,111,65,73,70,66,53,86,120,72,92,10,9,73,48,104,79,121,69,88,67,102,67,99,112,53,108,66,67,104,66,120,50,56,99,116,75,118,69,88,105,83,115,100,110,102,48,108,105,112,120,103,43,103,80,77,108,98,111,51,66,71,68,86,88,89,108,77,117,65,90,120,67,89,100,86,120,119,97,77,85,51,56,57,53,92,10,9,52,77,109,48,103,57,120,78,88,49,119,101,111,56,50,118,75,79,90,101,66,89,115,105,76,72,53,55,90,113,83,78,67,90,90,74,55,66,109,83,69,68,112,90,73,112,86,99,111,86,67,49,71,119,89,72,66,90,85,115,84,56,65,77,80,70,70,105,77,56,90,68,92,10,9,57,119,105,48,50,121,51,120,65,52,109,68,97,69,72,56,117,119,109,121,88,52,121,51,106,80,116,77,121,74,100,47,102,99,103,116,86,119,99,122,100,86,43,103,66,104,71,109,57,48,48,67,55,50,71,57,105,81,67,81,108,117,84,78,66,71,77,67,99,83,81,99,92,10,9,120,122,103,67,108,108,118,115,47,68,77,112,80,113,54,81,77,101,97,78,73,86,53,43,110,52,83,101,102,55,118,69,114,115,119,53,68,69,51,119,80,90,84,86,69,85,68,102,68,65,83,71,118,113,68,76,86,79,72,68,97,50,77,99,114,55,88,106,121,122,43,100,92,10,9,85,116,65,107,66,83,68,103,105,121,49,122,119,76,70,88,108,79,65,99,121,118,110,72,83,54,117,67,97,53,65,117,118,50,89,71,120,66,109,69,99,55,57,117,69,85,48,47,70,100,71,100,77,87,67,100,114,69,113,71,72,67,55,77,70,118,71,69,119,74,115,65,92,10,9,82,112,48,50,81,117,43,67,51,112,111,104,117,113,76,116,77,76,113,118,107,69,84,69,102,74,72,75,109,103,76,102,121,100,113,97,100,104,67,80,114,49,101,74,103,86,100,53,109,89,80,108,106,70,51,103,108,112,103,52,88,53,101,118,118,54,80,106,87,89,112,99,92,10,9,82,47,79,74,85,88,82,108,49,113,89,89,53,110,69,53,111,47,51,57,66,115,97,54,114,115,83,51,65,47,102,118,107,68,105,81,51,78,50,85,80,77,66,89,76,55,80,81,121,102,73,121,121,108,69,50,115,83,106,119,108,67,77,66,80,77,110,111,69,120,74,75,92,10,9,67,48,88,98,105,104,73,118,108,80,103,97,54,81,85,52,48,77,69,103,43,110,72,104,99,120,112,122,73,68,72,81,84,88,48,83,120,47,70,48,51,72,83,103,43,82,110,49,65,67,115,122,84,65,67,99,108,51,78,55,120,68,70,109,113,83,106,101,103,68,98,82,92,10,9,57,122,109,109,79,72,99,71,75,87,66,115,77,70,76,84,118,114,89,68,69,97,107,65,101,111,80,51,97,80,50,115,107,108,102,83,75,101,70,48,115,78,68,106,74,87,48,115,106,68,106,79,89,65,99,67,50,71,88,104,52,67,115,55,57,70,86,76,52,103,114,76,92,10,9,51,70,89,89,50,114,56,100,52,49,112,80,77,117,50,52,79,84,52,57,99,121,97,66,119,43,68,66,52,51,104,52,120,72,72,103,121,57,101,82,101,97,47,74,48,51,89,54,122,99,102,70,109,101,79,88,108,110,118,121,68,100,99,71,105,70,84,75,51,84,103,65,92,10,9,76,109,112,84,121,83,112,97,84,109,99,78,76,77,107,119,65,86,82,104,51,106,100,67,82,73,116,84,120,80,80,57,88,100,106,122,47,119,70,77,98,116,118,108,81,111,119,78,68,110,50,43,52,98,111,112,56,103,100,83,48,74,85,120,69,111,69,103,54,81,66,74,92,10,9,78,56,113,111,67,71,66,106,104,103,109,65,97,52,97,71,104,47,71,111,105,71,77,49,90,118,97,120,122,110,65,78,114,117,70,104,56,118,101,116,77,70,121,122,102,88,84,119,47,50,116,76,70,114,50,52,65,66,98,100,118,113,111,88,56,88,71,71,67,87,66,72,92,10,9,65,115,82,105,65,106,104,87,51,109,67,53,120,120,81,90,106,77,82,77,102,119,103,120,98,118,79,81,66,74,101,71,53,83,73,86,110,52,66,106,68,65,71,111,82,84,71,115,120,83,68,47,76,112,66,101,51,66,107,90,74,103,67,88,67,78,106,122,89,120,114,122,92,10,9,122,56,75,99,98,117,90,100,83,47,116,98,104,90,117,68,90,110,117,76,85,109,109,74,81,49,43,122,82,67,113,54,116,119,55,116,67,114,43,104,77,120,50,55,103,50,116,71,69,98,106,48,86,119,115,113,103,115,55,73,115,66,84,119,67,112,79,98,114,104,78,82,92,10,9,74,65,114,67,89,111,78,53,71,80,71,81,79,120,108,57,106,71,65,111,115,107,55,121,55,78,110,87,43,49,114,54,121,67,101,98,120,120,75,83,71,77,90,52,43,109,122,108,107,80,78,98,107,107,73,80,66,114,67,68,108,114,70,80,68,111,113,66,108,84,120,55,92,10,9,104,117,115,52,52,88,89,109,65,86,121,100,119,78,106,68,68,79,77,57,119,117,120,106,116,99,81,98,70,77,82,85,106,48,82,80,87,55,76,72,89,116,76,89,109,102,119,65,86,71,73,112,54,103,65,99,97,49,109,122,77,122,49,55,110,89,71,122,103,119,83,81,92,10,9,83,53,83,84,75,101,68,109,56,110,56,50,111,102,70,49,86,114,107,113,69,114,57,121,54,71,99,80,113,88,88,72,107,43,47,66,81,87,97,55,53,122,49,122,100,80,74,109,81,57,115,110,78,79,49,97,107,76,71,78,89,56,89,43,76,50,103,77,65,108,80,81,92,10,9,73,107,80,77,52,69,76,105,113,111,115,90,57,120,89,73,99,53,122,56,102,48,56,121,107,81,114,80,50,105,115,79,74,87,43,65,51,54,73,117,103,121,106,105,54,48,47,84,83,68,57,110,107,57,121,99,109,57,68,122,89,52,53,73,73,114,88,97,119,72,68,117,92,10,9,78,76,81,47,81,80,113,78,67,97,83,47,97,83,102,120,53,56,82,122,117,80,104,78,70,110,122,51,52,114,49,68,103,83,70,53,112,72,106,53,84,99,73,69,65,80,110,50,72,52,122,55,109,111,108,85,122,74,56,74,69,75,79,72,101,76,48,72,82,101,110,73,92,10,9,74,76,122,65,66,107,81,77,106,89,105,81,71,118,116,119,79,83,50,99,67,110,113,84,53,83,119,74,81,76,113,98,74,121,119,54,104,47,48,74,114,106,47,121,76,49,54,113,99,119,112,116,83,73,120,83,85,106,67,90,116,108,110,79,57,118,57,114,121,120,107,54,92,10,9,109,117,108,108,70,66,98,55,79,71,122,112,76,118,52,81,110,76,71,51,74,98,84,43,87,121,81,50,56,83,119,43,103,84,57,109,71,106,78,99,89,98,78,110,76,112,85,83,120,72,71,97,70,47,43,77,120,54,47,100,69,120,87,55,120,79,105,97,57,97,75,68,92,10,9,99,87,112,81,81,104,53,74,55,84,121,109,86,47,68,74,90,70,71,75,67,57,89,90,122,74,56,54,49,122,67,47,78,43,120,101,69,111,110,113,101,53,109,118,74,70,97,76,76,73,70,104,120,86,69,81,55,84,50,101,109,50,77,111,112,73,116,53,77,97,52,47,92,10,9,80,117,104,87,110,109,80,82,113,66,112,101,113,118,114,87,53,120,69,72,102,121,108,103,112,43,90,103,71,50,111,76,85,101,105,80,82,43,106,70,66,55,69,108,43,85,50,52,117,73,121,78,49,47,103,80,99,66,66,109,52,90,107,82,49,120,52,102,122,117,77,109,92,10,9,83,54,89,116,79,106,105,88,116,69,89,57,83,89,86,89,110,56,108,111,81,68,117,70,108,71,52,73,73,47,57,69,117,71,88,97,70,113,82,120,65,120,102,47,74,120,69,43,80,50,66,83,65,77,77,83,67,106,117,100,82,54,53,100,43,84,53,76,73,118,73,86,92,10,9,76,74,77,52,84,97,84,83,48,48,89,78,109,83,57,76,108,115,71,55,104,86,118,54,88,113,122,47,67,56,84,85,71,43,48,56,51,80,68,119,72,66,74,80,111,74,101,43,103,118,122,87,106,118,99,57,101,66,69,57,76,69,81,56,112,69,71,100,83,121,47,57,92,10,9,103,77,104,67,72,70,67,90,68,68,109,111,107,89,84,69,88,83,102,52,114,73,108,55,83,67,87,77,74,74,99,76,121,74,75,52,109,100,116,120,78,106,47,65,85,81,55,90,47,65,66,90,65,115,104,67,108,103,67,121,107,67,87,65,76,71,81,74,119,65,43,49,92,10,9,82,75,113,75,71,74,73,109,51,121,102,77,108,97,112,48,65,65,107,66,80,110,49,73,71,79,71,82,88,118,47,109,71,79,101,79,79,98,49,80,57,111,68,57,120,80,51,67,47,119,43,101,80,47,86,105,72,65,99,112,56,101,66,80,57,49,101,82,56,118,80,98,92,10,9,83,77,47,106,48,100,105,98,83,101,113,66,103,56,98,86,73,116,108,113,75,120,121,111,83,57,73,97,53,111,122,111,73,82,105,78,86,116,78,99,69,83,50,70,71,115,73,110,54,109,119,66,97,87,101,71,68,89,114,73,107,114,54,79,105,103,120,100,111,79,107,116,92,10,9,77,83,108,109,84,73,107,114,89,66,74,57,83,43,74,90,73,102,118,79,73,49,118,65,51,66,68,75,108,57,49,107,112,50,105,83,107,69,73,75,122,105,106,100,118,70,84,69,78,85,122,88,109,56,103,47,89,89,88,72,84,47,67,57,76,87,50,83,86,52,109,66,92,10,9,99,65,56,55,87,48,78,100,105,79,109,55,84,116,105,57,86,102,70,49,54,110,122,101,107,79,97,48,108,83,78,86,116,121,76,108,67,48,121,102,56,76,78,68,52,107,83,117,107,121,89,115,106,55,99,75,102,117,70,72,55,72,122,119,47,52,118,113,75,114,43,80,92,10,9,100,115,57,104,119,106,49,100,68,110,89,119,87,69,79,98,66,118,53,67,54,86,81,110,112,112,51,105,122,101,57,48,79,111,113,81,76,74,115,70,98,66,54,53,97,116,109,105,98,72,86,81,53,69,68,116,51,83,85,117,105,69,69,88,88,107,105,71,72,86,115,52,92,10,9,50,110,115,74,71,71,70,101,67,85,98,107,71,118,84,47,81,121,51,79,73,72,72,67,88,112,85,116,111,68,113,122,56,99,99,87,73,106,106,100,115,118,49,119,116,114,109,88,89,51,53,103,71,71,85,117,77,52,122,53,100,73,75,76,47,82,122,106,109,85,100,108,92,10,9,50,67,43,122,120,65,118,85,68,110,89,66,101,76,83,77,70,56,108,69,70,75,70,80,43,77,82,102,69,51,79,47,48,75,117,80,69,47,114,48,55,55,48,84,90,77,120,119,98,70,53,109,117,97,100,88,104,112,110,70,83,83,111,112,111,73,112,68,66,121,106,97,92,10,9,106,66,111,52,114,105,88,109,98,75,53,87,119,48,88,112,121,108,120,120,69,103,69,113,107,70,86,86,88,68,117,89,56,73,73,80,116,52,106,100,53,84,76,52,56,117,70,109,47,108,51,53,109,105,65,84,67,67,98,68,78,82,80,49,73,120,76,118,67,102,121,71,92,10,9,101,118,97,76,68,87,49,48,82,73,77,99,104,77,116,69,99,110,85,76,48,122,67,77,109,68,48,47,80,67,88,115,70,84,56,104,54,105,69,80,51,121,97,97,73,120,103,121,71,77,97,54,77,107,86,108,117,71,75,57,111,55,107,71,80,56,82,79,77,84,57,110,92,10,9,67,89,110,102,75,70,85,72,70,122,103,107,108,108,104,73,47,120,89,113,65,109,104,79,78,55,118,67,70,101,76,119,105,108,52,110,107,108,122,117,83,103,68,81,71,43,104,56,69,106,50,83,43,120,70,118,104,43,68,76,89,110,111,66,115,73,121,49,69,89,71,121,92,10,9,54,65,119,79,118,97,69,52,118,68,114,97,77,84,84,110,82,111,90,50,115,77,81,116,48,117,120,113,86,57,78,88,98,117,76,83,69,97,75,109,75,51,113,70,111,50,109,97,105,74,98,90,70,71,98,103,73,83,84,51,65,119,117,69,122,83,73,98,89,69,84,97,92,10,9,104,87,81,111,100,112,76,43,119,66,47,85,89,65,73,100,73,50,84,75,85,57,68,100,119,107,84,66,84,101,119,120,66,52,101,78,88,121,118,54,117,78,110,83,112,104,102,84,107,85,85,72,99,121,122,116,117,121,107,67,83,107,114,73,75,87,99,105,103,49,70,99,92,10,9,54,85,85,115,71,108,85,57,74,79,88,66,88,118,48,71,121,101,107,55,73,49,66,119,118,117,98,51,116,89,97,116,77,119,49,73,87,89,117,105,106,80,68,109,47,83,86,57,48,102,115,85,120,49,77,97,107,71,112,49,90,79,67,51,120,90,89,120,109,108,109,48,92,10,9,113,108,99,119,71,70,119,84,73,68,121,116,67,82,50,104,48,75,83,117,112,50,102,51,102,56,85,73,69,83,116,118,48,66,43,52,81,89,65,105,114,114,69,107,78,76,67,86,98,66,116,72,114,108,57,104,100,119,65,100,76,80,72,67,112,51,104,100,114,101,109,122,92,10,9,81,79,77,71,53,49,107,48,106,74,57,74,118,73,115,56,107,104,116,53,113,100,120,67,57,122,73,100,97,99,102,72,111,65,108,99,97,120,110,68,74,90,70,70,75,97,55,85,116,65,79,103,105,70,73,51,89,81,55,115,82,69,109,51,101,121,80,115,65,76,113,67,92,10,9,85,119,49,67,57,117,101,82,83,75,110,84,56,81,101,104,83,74,105,76,88,117,71,115,47,53,70,73,117,97,112,57,84,114,115,79,118,72,69,101,70,76,121,107,70,53,78,105,89,79,122,87,87,113,54,102,52,78,74,90,106,115,73,73,112,73,78,67,89,106,75,54,92,10,9,87,49,84,66,85,75,80,43,73,117,84,68,54,85,114,70,103,55,69,54,79,87,83,102,75,120,122,70,48,100,85,74,83,82,54,70,74,68,90,72,66,100,118,56,109,107,99,104,103,78,113,71,101,55,102,52,100,80,109,50,69,113,103,80,104,88,121,52,102,120,109,117,92,10,9,104,86,88,85,72,73,120,53,103,99,78,67,72,43,71,87,90,107,89,72,110,49,109,117,110,120,54,70,65,79,111,121,67,65,65,65,115,43,115,84,67,83,122,83,97,52,90,114,69,65,47,68,70,71,119,43,86,102,79,55,122,110,69,121,105,87,119,112,48,67,43,56,92,10,9,72,86,78,102,116,114,119,67,114,97,73,81,103,79,107,73,67,67,55,77,98,48,87,56,54,85,115,69,75,89,70,48,49,85,88,66,49,98,47,111,113,68,71,68,90,114,79,110,53,108,113,66,53,118,100,78,77,84,52,80,57,65,119,57,70,78,74,71,70,76,65,86,92,10,9,106,84,119,116,67,103,71,89,85,115,85,87,75,98,82,77,80,89,86,55,104,103,111,98,68,68,74,99,103,53,108,54,110,79,68,90,73,72,74,74,74,78,81,82,57,81,76,72,110,99,69,86,78,112,71,67,75,117,53,97,103,43,115,67,117,51,69,81,71,108,117,79,92,10,9,99,105,77,66,109,76,120,90,86,80,86,115,118,121,69,75,100,119,110,56,115,79,85,72,81,108,110,48,70,121,122,113,86,72,106,103,100,68,68,99,48,53,113,48,97,108,101,71,52,77,105,51,120,80,83,105,99,74,120,79,74,53,49,69,51,75,114,116,102,49,117,117,92,10,9,100,56,103,85,65,81,104,83,122,86,52,114,43,78,87,120,79,83,110,112,43,108,109,89,77,87,120,122,77,71,106,77,74,116,86,110,88,49,73,67,80,85,110,72,67,70,121,51,79,108,111,52,56,111,109,97,97,122,90,71,122,83,86,83,113,82,70,120,47,118,79,70,92,10,9,51,115,107,109,68,77,121,122,88,79,99,88,118,65,52,111,66,107,121,112,82,87,121,49,99,80,111,120,86,98,65,70,84,67,88,70,113,82,84,82,109,117,107,85,78,90,48,115,98,87,47,122,85,113,108,117,119,56,68,102,71,102,109,70,79,72,105,74,90,90,120,90,92,10,9,89,82,86,66,53,90,108,83,103,65,114,119,57,81,49,103,48,78,120,117,74,109,49,43,72,79,74,52,52,81,66,50,106,113,99,105,55,70,66,108,83,67,76,112,89,109,65,107,100,81,68,57,121,80,65,89,110,109,71,117,53,88,112,114,119,102,84,52,122,108,70,119,92,10,9,114,98,113,88,122,47,71,112,103,52,89,77,117,109,112,84,43,112,102,53,68,103,56,75,69,50,110,110,71,68,108,122,121,78,66,88,87,86,54,121,106,85,65,114,43,117,90,50,75,110,72,52,76,106,53,47,99,84,104,43,98,66,88,109,74,74,81,86,109,73,120,121,92,10,9,75,81,73,89,112,114,108,118,111,73,78,67,90,83,121,100,100,119,87,97,56,47,88,82,69,79,100,100,71,50,76,113,111,103,68,83,112,76,90,110,83,67,50,50,90,73,120,55,65,56,47,84,108,53,105,117,112,99,120,53,120,74,86,55,121,75,81,48,87,56,72,101,92,10,9,97,82,88,71,107,120,118,73,98,70,108,65,75,86,49,43,71,118,75,99,103,108,72,105,85,88,76,73,104,66,70,108,104,117,100,101,119,105,85,52,78,120,82,122,43,116,76,82,56,82,72,110,114,109,118,116,81,70,51,100,111,109,56,78,87,85,55,103,83,118,53,98,92,10,9,99,113,69,51,119,89,83,89,51,77,78,80,49,112,105,43,105,121,110,78,72,75,117,102,73,53,49,48,73,87,119,116,118,120,54,85,81,109,54,90,100,51,105,113,101,120,68,98,102,76,75,98,73,56,100,103,116,90,66,106,116,70,86,52,53,43,76,108,68,50,67,48,92,10,9,114,85,115,79,112,103,99,48,97,86,112,97,120,114,103,87,80,47,70,83,113,102,99,80,69,114,70,80,99,67,88,50,98,72,105,52,72,113,113,84,111,117,115,107,79,110,80,102,69,109,53,112,57,79,67,104,100,76,57,73,71,99,47,75,48,120,70,50,112,56,78,82,92,10,9,107,82,72,73,69,115,66,82,68,116,110,103,48,67,119,66,90,67,70,76,65,70,110,73,69,107,65,87,115,103,83,81,104,97,77,81,88,73,77,81,52,73,120,120,112,107,105,53,102,56,77,105,66,57,115,122,110,67,118,104,106,108,121,90,55,111,70,86,69,79,55,86,92,10,9,115,77,53,78,105,87,71,79,99,79,112,65,49,90,67,76,83,67,83,68,85,121,90,56,66,77,118,83,87,72,66,87,82,84,81,84,84,75,57,119,65,74,51,57,65,51,111,47,101,69,97,69,121,88,85,83,75,98,85,122,51,78,74,104,90,111,97,57,118,52,120,67,92,10,9,81,49,108,77,50,107,109,52,107,99,80,115,68,117,48,116,72,72,110,103,90,68,78,83,113,82,49,107,112,109,51,116,82,109,72,79,117,120,121,48,99,77,85,82,78,88,56,73,57,80,103,114,111,47,104,66,69,68,54,75,111,76,51,56,118,105,67,48,113,69,79,56,92,10,9,101,69,118,73,119,54,87,47,107,111,115,109,115,65,122,70,67,85,82,74,72,68,48,50,53,65,74,99,53,48,104,115,117,104,103,70,86,48,48,103,70,117,104,87,85,108,118,118,57,71,110,118,112,112,65,97,117,114,121,104,98,88,88,75,75,106,75,75,120,107,90,108,92,10,9,48,74,43,82,105,116,104,86,67,47,109,86,108,119,122,99,121,121,87,65,89,56,107,79,69,66,85,87,104,110,106,53,100,56,88,52,119,70,65,86,72,56,99,99,57,48,82,68,69,69,107,97,99,76,50,122,89,111,99,48,50,81,66,87,107,115,113,87,77,52,100,87,92,10,9,68,70,116,67,70,74,106,75,73,89,66,109,106,79,103,84,76,114,122,114,43,80,73,55,101,80,72,110,54,70,47,71,50,65,107,113,83,70,122,70,55,65,47,122,117,56,78,110,65,66,114,72,80,65,54,118,116,99,119,66,113,101,83,88,74,112,119,85,89,114,111,116,92,10,9,78,104,66,79,111,90,78,69,79,80,100,114,70,98,103,69,81,111,67,112,71,90,71,65,90,73,74,83,79,69,104,97,98,89,114,98,117,49,55,119,73,50,111,119,118,50,71,107,51,56,56,88,113,87,103,111,68,106,77,51,108,104,103,48,110,90,80,111,112,85,76,118,92,10,9,119,104,52,88,84,76,99,120,103,88,71,86,107,81,102,106,100,112,47,106,49,57,56,120,89,101,114,47,104,87,72,115,113,82,108,75,121,49,76,107,54,89,116,100,106,69,108,52,55,79,87,113,110,100,66,118,68,73,75,73,116,86,122,119,80,70,103,104,100,115,71,49,92,10,9,67,116,89,120,120,77,102,66,87,50,103,110,105,82,51,119,116,105,107,81,102,78,101,118,78,75,67,119,89,51,47,71,102,89,106,52,88,85,76,105,68,85,83,105,77,119,81,118,86,103,67,79,73,69,50,70,50,115,85,77,106,112,54,49,77,105,84,97,89,81,102,111,92,10,9,111,102,103,100,117,81,108,77,104,83,49,51,48,73,52,75,90,57,109,116,65,86,69,101,111,106,73,99,101,109,115,81,73,103,78,75,74,100,112,53,115,70,52,102,48,117,54,54,51,98,81,68,50,71,114,109,119,98,52,57,119,112,99,72,111,68,119,53,98,116,98,49,92,10,9,77,85,80,53,90,75,115,102,84,76,110,121,85,67,98,108,89,89,114,116,76,50,47,90,65,84,103,86,77,115,66,104,49,47,82,75,70,53,99,99,122,102,119,75,98,110,79,77,83,107,52,75,86,68,90,55,109,57,82,122,83,82,73,105,112,51,56,72,81,75,66,69,92,10,9,71,119,51,49,52,101,122,113,83,108,56,57,52,71,75,105,120,118,82,88,103,54,43,120,117,106,65,55,108,83,75,103,65,102,98,119,49,122,88,88,78,119,116,122,81,65,78,56,56,70,67,107,81,87,101,47,82,111,83,117,76,84,74,53,108,85,103,70,84,119,90,100,92,10,9,50,72,86,57,52,107,116,55,88,76,105,53,115,101,71,76,81,48,67,75,75,100,48,77,77,114,68,101,121,74,120,68,71,104,68,121,116,86,52,99,121,117,79,69,116,85,55,55,47,86,88,50,75,102,87,113,43,51,97,71,121,114,55,100,52,122,51,108,55,117,102,120,92,10,9,75,108,88,100,54,98,117,118,65,87,88,68,68,65,118,57,81,51,54,70,116,117,119,99,50,73,69,52,109,84,49,86,43,81,74,49,74,87,77,51,48,80,88,55,72,90,53,118,73,76,86,53,49,90,73,113,76,121,47,68,117,57,65,67,86,98,53,67,80,56,100,116,92,10,9,67,103,114,120,82,57,69,103,70,56,52,120,69,99,55,65,80,119,112,122,112,103,48,100,102,71,88,104,122,72,72,117,99,86,121,117,86,100,75,65,76,117,97,43,70,110,49,74,68,119,104,101,105,66,101,43,115,75,100,56,79,53,73,79,113,112,70,75,80,99,105,72,92,10,9,74,65,110,89,51,99,56,75,97,119,121,113,55,80,116,51,107,52,103,84,65,99,80,87,82,47,71,55,122,101,118,52,73,84,111,71,53,104,72,122,87,85,66,98,57,68,53,113,67,50,47,101,108,120,110,106,100,49,102,56,116,116,81,119,49,43,89,43,85,100,69,87,92,10,9,79,84,122,90,120,54,66,57,76,77,122,117,88,43,48,86,120,49,80,83,115,77,57,69,65,70,115,78,68,98,118,54,47,104,49,72,47,72,119,110,120,87,56,50,51,47,56,79,49,65,53,102,84,107,115,105,120,66,111,87,118,107,77,70,84,85,84,112,88,69,83,109,92,10,9,87,76,115,117,57,66,101,120,69,98,100,97,43,103,54,71,109,53,109,77,89,85,71,90,80,43,109,121,118,83,56,74,82,85,121,71,88,120,72,48,104,87,74,98,83,107,78,118,43,103,74,104,90,85,75,73,57,110,48,105,87,116,74,105,86,98,89,80,102,78,72,49,92,10,9,82,87,98,103,84,72,70,52,51,77,73,105,67,47,71,80,111,72,43,80,69,52,101,67,86,86,81,81,57,78,85,51,82,102,65,48,86,83,104,112,98,114,68,77,43,50,108,120,113,74,119,117,76,72,47,112,84,67,49,55,120,75,70,103,110,66,51,69,53,66,52,85,92,10,9,104,120,74,50,98,82,83,54,101,65,103,102,81,122,68,77,119,107,84,52,71,99,70,84,80,72,49,78,117,121,47,74,101,68,76,76,77,86,72,85,109,65,121,75,89,81,56,113,99,104,77,88,71,51,76,113,43,107,88,80,70,103,70,88,57,68,83,56,113,110,105,109,92,10,9,100,112,97,107,87,48,70,51,56,104,76,76,118,67,100,53,47,79,113,106,122,109,74,103,82,50,70,79,43,65,65,70,122,43,88,105,56,75,82,76,57,85,109,66,53,78,69,82,85,105,65,79,104,88,57,68,88,88,113,72,82,102,88,114,66,50,121,118,84,49,113,43,92,10,9,103,70,49,48,49,117,56,104,51,69,71,47,55,97,98,47,98,121,100,71,55,72,99,87,104,82,90,83,121,65,100,68,120,54,100,74,80,70,100,122,80,119,112,100,106,65,52,119,86,76,56,88,113,82,103,56,72,70,50,118,48,105,52,82,68,68,107,55,51,110,66,107,92,10,9,98,104,83,108,48,55,111,104,70,86,52,51,121,120,111,103,119,109,109,65,48,75,102,106,57,48,78,116,69,103,117,114,69,105,73,113,54,82,80,104,68,47,77,76,109,72,43,88,77,53,82,66,68,122,65,105,88,71,116,97,98,78,109,114,70,71,49,97,77,55,55,99,92,10,9,89,85,122,75,88,109,102,112,90,55,54,105,122,85,65,88,75,120,111,84,97,120,114,54,51,75,43,52,118,54,118,68,76,103,89,122,47,89,101,85,101,103,55,52,103,99,84,90,88,105,113,84,43,57,101,97,119,74,82,48,103,77,116,102,48,105,74,104,77,67,55,103,92,10,9,83,111,79,105,120,103,57,70,82,80,85,122,83,73,84,97,81,56,120,89,81,50,74,117,101,108,117,107,66,83,105,82,47,105,102,119,87,121,54,74,81,113,90,85,100,90,118,74,65,54,110,89,77,106,57,52,74,74,109,83,74,97,49,86,77,73,74,116,104,68,53,114,92,10,9,105,69,102,106,98,110,68,107,78,97,111,76,99,54,121,104,97,112,101,97,114,82,76,88,69,103,67,73,113,81,78,86,86,68,115,43,65,50,102,119,66,90,111,118,53,110,86,71,50,53,54,77,76,50,43,97,112,89,43,116,71,103,99,89,48,54,55,86,78,56,81,79,92,10,9,85,78,50,82,68,48,114,88,75,78,54,102,103,88,101,119,121,100,80,85,67,43,106,116,52,122,83,84,65,80,65,90,55,120,118,79,90,104,116,119,67,107,47,116,115,49,121,118,113,118,110,75,84,89,107,99,79,50,100,73,81,108,107,109,111,105,88,98,53,69,73,100,92,10,9,110,83,75,111,105,71,84,102,74,66,81,84,79,122,82,75,73,68,56,84,90,72,116,53,88,89,81,57,78,98,118,78,69,112,108,114,115,68,102,111,111,70,72,73,53,57,85,57,106,121,109,110,56,108,67,74,122,121,82,77,65,65,85,109,84,101,65,54,79,111,101,109,92,10,9,120,102,122,121,76,120,80,109,105,116,114,103,54,78,57,108,57,80,85,106,121,51,86,98,77,107,87,100,54,116,102,48,118,74,43,71,101,79,90,118,121,84,67,107,103,106,71,71,100,104,52,90,107,120,53,75,107,65,67,71,66,54,85,65,110,89,71,108,118,56,86,81,92,10,9,120,72,88,76,52,111,90,70,100,55,102,48,66,83,54,50,105,97,85,80,74,72,72,101,89,117,106,106,74,107,80,98,79,82,114,106,84,98,79,81,107,103,67,52,55,99,101,74,55,48,106,122,72,119,57,55,118,69,74,83,119,72,79,56,101,65,112,110,43,90,49,49,92,10,9,66,104,80,80,119,52,52,79,114,107,70,121,43,115,51,67,114,89,106,107,66,116,114,79,82,106,65,52,100,122,57,110,47,67,98,112,72,70,81,65,48,43,122,47,77,118,111,53,105,102,81,75,53,119,100,50,78,117,81,104,102,77,55,81,55,106,105,54,51,112,110,107,92,10,9,43,111,107,107,56,54,43,77,52,97,117,114,75,117,120,112,56,110,81,65,110,85,77,80,48,109,76,109,48,53,70,48,84,79,66,73,50,85,97,83,86,70,114,122,116,53,70,107,47,105,73,97,100,122,80,120,71,80,57,49,68,72,69,78,68,56,56,104,110,84,119,87,92,10,9,116,82,50,112,100,79,118,53,57,80,70,98,83,102,69,120,110,55,98,121,121,99,74,87,115,85,73,78,85,70,72,102,82,97,114,82,90,118,81,105,85,79,114,107,47,53,104,105,97,108,65,85,121,54,99,70,87,67,80,99,97,47,106,57,111,67,71,98,72,43,65,111,92,10,9,104,50,120,115,52,70,69,79,47,120,70,103,65,66,77,56,82,101,43,80,89,55,111,85,65,65,65,65,65,69,108,70,84,107,83,117,81,109,67,67,92,10,9,34,10,9,108,111,99,97,108,32,109,111,115,97,105,99,95,50,120,95,112,110,103,32,61,10,9,34,105,86,66,79,82,119,48,75,71,103,111,65,65,65,65,78,83,85,104,69,85,103,65,65,65,81,65,65,65,65,69,65,67,65,89,65,65,65,66,99,99,113,104,109,65,65,65,65,71,88,82,70,87,72,82,84,98,50,90,48,100,50,70,121,90,81,66,66,90,71,57,105,92,10,9,90,83,66,74,98,87,70,110,90,86,74,108,89,87,82,53,99,99,108,108,80,65,65,65,65,50,90,112,86,70,104,48,87,69,49,77,79,109,78,118,98,83,53,104,90,71,57,105,90,83,53,52,98,88,65,65,65,65,65,65,65,68,119,47,101,72,66,104,89,50,116,108,92,10,9,100,67,66,105,90,87,100,112,98,106,48,105,55,55,117,47,73,105,66,112,90,68,48,105,86,122,86,78,77,69,49,119,81,50,86,111,97,85,104,54,99,109,86,84,101,107,53,85,89,51,112,114,89,122,108,107,73,106,56,43,73,68,120,52,79,110,104,116,99,71,49,108,92,10,9,100,71,69,103,101,71,49,115,98,110,77,54,101,68,48,105,89,87,82,118,89,109,85,54,98,110,77,54,98,87,86,48,89,83,56,105,73,72,103,54,101,71,49,119,100,71,115,57,73,107,70,107,98,50,74,108,73,70,104,78,85,67,66,68,98,51,74,108,73,68,85,117,92,10,9,77,121,49,106,77,68,69,120,73,68,89,50,76,106,69,48,78,84,89,50,77,83,119,103,77,106,65,120,77,105,56,119,77,105,56,119,78,105,48,120,78,68,111,49,78,106,111,121,78,121,65,103,73,67,65,103,73,67,65,103,73,106,52,103,80,72,74,107,90,106,112,83,92,10,9,82,69,89,103,101,71,49,115,98,110,77,54,99,109,82,109,80,83,74,111,100,72,82,119,79,105,56,118,100,51,100,51,76,110,99,122,76,109,57,121,90,121,56,120,79,84,107,53,76,122,65,121,76,122,73,121,76,88,74,107,90,105,49,122,101,87,53,48,89,88,103,116,92,10,9,98,110,77,106,73,106,52,103,80,72,74,107,90,106,112,69,90,88,78,106,99,109,108,119,100,71,108,118,98,105,66,121,90,71,89,54,89,87,74,118,100,88,81,57,73,105,73,103,101,71,49,115,98,110,77,54,101,71,49,119,84,85,48,57,73,109,104,48,100,72,65,54,92,10,9,76,121,57,117,99,121,53,104,90,71,57,105,90,83,53,106,98,50,48,118,101,71,70,119,76,122,69,117,77,67,57,116,98,83,56,105,73,72,104,116,98,71,53,122,79,110,78,48,85,109,86,109,80,83,74,111,100,72,82,119,79,105,56,118,98,110,77,117,89,87,82,118,92,10,9,89,109,85,117,89,50,57,116,76,51,104,104,99,67,56,120,76,106,65,118,99,49,82,53,99,71,85,118,85,109,86,122,98,51,86,121,89,50,86,83,90,87,89,106,73,105,66,52,98,87,120,117,99,122,112,52,98,88,65,57,73,109,104,48,100,72,65,54,76,121,57,117,92,10,9,99,121,53,104,90,71,57,105,90,83,53,106,98,50,48,118,101,71,70,119,76,122,69,117,77,67,56,105,73,72,104,116,99,69,49,78,79,107,57,121,97,87,100,112,98,109,70,115,82,71,57,106,100,87,49,108,98,110,82,74,82,68,48,105,101,71,49,119,76,109,82,112,92,10,9,90,68,111,50,78,68,89,51,79,68,85,50,79,84,107,48,78,107,74,70,78,84,69,120,81,84,103,51,82,107,81,51,77,84,78,67,79,84,99,50,78,48,85,122,78,121,73,103,101,71,49,119,84,85,48,54,82,71,57,106,100,87,49,108,98,110,82,74,82,68,48,105,92,10,9,101,71,49,119,76,109,82,112,90,68,111,50,77,85,69,53,81,107,82,66,79,68,103,51,82,107,77,120,77,85,85,49,79,84,82,67,82,69,69,48,77,85,70,70,82,106,89,50,81,85,90,67,81,121,73,103,101,71,49,119,84,85,48,54,83,87,53,122,100,71,70,117,92,10,9,89,50,86,74,82,68,48,105,101,71,49,119,76,109,108,112,90,68,111,50,77,85,69,53,81,107,82,66,78,122,103,51,82,107,77,120,77,85,85,49,79,84,82,67,82,69,69,48,77,85,70,70,82,106,89,50,81,85,90,67,81,121,73,103,101,71,49,119,79,107,78,121,92,10,9,90,87,70,48,98,51,74,85,98,50,57,115,80,83,74,66,90,71,57,105,90,83,66,81,97,71,57,48,98,51,78,111,98,51,65,103,81,49,77,50,73,67,104,88,97,87,53,107,98,51,100,122,75,83,73,43,73,68,120,52,98,88,66,78,84,84,112,69,90,88,74,112,92,10,9,100,109,86,107,82,110,74,118,98,83,66,122,100,70,74,108,90,106,112,112,98,110,78,48,89,87,53,106,90,85,108,69,80,83,74,52,98,88,65,117,97,87,108,107,79,107,73,121,81,106,89,48,79,68,74,68,82,107,77,52,78,48,85,49,77,84,70,66,78,68,85,121,92,10,9,79,68,99,50,78,106,70,68,82,106,77,51,79,84,73,119,73,105,66,122,100,70,74,108,90,106,112,107,98,50,78,49,98,87,86,117,100,69,108,69,80,83,74,52,98,88,65,117,90,71,108,107,79,106,89,48,78,106,99,52,78,84,89,53,79,84,81,50,81,107,85,49,92,10,9,77,84,70,66,79,68,100,71,82,68,99,120,77,48,73,53,78,122,89,51,82,84,77,51,73,105,56,43,73,68,119,118,99,109,82,109,79,107,82,108,99,50,78,121,97,88,66,48,97,87,57,117,80,105,65,56,76,51,74,107,90,106,112,83,82,69,89,43,73,68,119,118,92,10,9,101,68,112,52,98,88,66,116,90,88,82,104,80,105,65,56,80,51,104,119,89,87,78,114,90,88,81,103,90,87,53,107,80,83,74,121,73,106,56,43,52,74,112,67,51,119,65,65,79,98,108,74,82,69,70,85,101,78,114,115,88,81,101,52,70,99,88,49,80,52,56,113,92,10,9,73,68,120,70,66,69,81,85,82,69,82,105,82,98,70,88,120,66,54,86,75,76,71,88,66,73,48,86,87,121,119,120,75,115,97,83,113,69,70,106,55,121,50,50,87,67,77,82,57,87,43,68,113,72,110,89,65,67,117,75,66,82,81,85,65,83,109,67,86,79,110,73,92,10,9,43,56,47,118,50,51,80,106,57,101,97,87,51,90,107,122,115,55,80,51,122,101,47,55,122,115,102,106,118,100,51,90,110,90,50,90,77,43,101,99,79,97,87,109,118,114,54,101,65,103,73,67,71,105,89,97,104,85,56,81,69,66,65,89,81,69,66,65,81,71,65,65,92,10,9,65,81,69,66,103,81,69,69,66,65,81,69,66,104,65,81,69,70,68,100,97,75,74,120,122,121,97,75,84,108,88,85,81,57,72,97,105,116,111,113,97,113,79,111,115,97,76,97,66,79,51,56,113,79,105,72,103,116,47,78,86,52,82,106,105,83,87,75,108,105,108,97,92,10,9,114,109,105,120,111,108,87,75,70,104,82,99,115,48,106,82,67,107,86,76,109,86,90,121,101,55,110,47,76,43,84,55,70,47,76,47,82,52,88,104,78,107,89,122,82,98,48,86,98,99,55,106,106,117,47,57,106,97,75,51,70,88,50,102,119,117,98,86,84,57,72,104,92,10,9,105,114,90,88,49,73,49,47,47,54,50,105,107,89,111,101,85,80,83,105,52,51,102,113,111,75,105,88,111,111,49,53,102,88,82,88,116,66,98,47,118,112,50,105,70,103,88,88,76,43,68,53,80,69,102,82,70,69,86,102,75,90,113,111,54,66,78,70,55,120,82,90,92,10,9,72,43,75,111,48,84,103,71,118,69,76,82,112,82,109,98,117,80,86,66,50,106,69,67,71,80,115,103,82,83,99,113,87,114,80,73,51,56,71,103,47,48,47,82,90,89,111,43,99,118,65,43,117,121,117,54,85,57,69,118,75,108,122,51,110,75,75,106,101,66,79,120,92,10,9,65,84,68,66,103,120,88,49,86,98,83,84,111,103,48,69,50,49,55,70,106,79,65,78,82,85,77,86,118,99,52,98,88,117,111,77,52,72,90,70,65,122,77,50,103,101,102,121,89,65,85,107,120,119,54,75,47,113,87,111,85,52,120,114,73,89,88,57,81,100,70,78,92,10,9,70,116,47,110,88,69,86,47,83,56,68,81,88,49,66,48,69,71,56,67,85,116,104,77,48,89,50,75,57,109,68,74,49,57,85,99,102,107,98,82,102,83,120,120,112,87,89,68,97,74,102,66,83,84,119,118,114,71,77,116,89,73,99,100,70,110,80,120,53,49,82,75,92,10,9,76,73,122,102,87,110,111,102,55,79,98,88,74,53,121,51,66,55,67,97,73,65,71,111,80,105,77,85,106,86,87,48,108,56,80,70,84,121,120,53,72,97,47,111,76,85,88,118,75,84,111,69,71,51,103,97,68,71,67,116,119,65,65,113,89,107,80,101,113,102,54,112,92,10,9,54,65,117,50,90,48,68,102,43,50,80,71,118,116,115,57,70,78,108,51,107,117,74,109,83,109,89,80,105,106,118,118,98,116,79,56,86,48,74,105,80,85,118,82,104,52,112,50,56,50,66,99,89,73,100,53,109,105,75,55,49,110,89,109,68,101,107,89,65,98,77,111,92,10,9,83,115,57,51,56,65,120,56,121,54,78,53,111,118,81,117,56,110,99,89,48,65,98,122,100,86,102,70,97,71,56,57,82,100,99,111,50,111,98,118,47,89,54,90,67,89,120,68,76,47,80,80,78,114,71,49,111,108,48,77,98,65,90,72,77,65,79,82,119,110,69,71,92,10,9,99,50,57,110,105,103,120,119,83,122,84,118,80,53,54,90,109,109,47,89,106,112,107,65,112,75,54,76,101,97,79,120,76,103,70,107,107,81,72,89,108,103,66,50,86,102,83,120,111,111,100,75,76,80,53,56,47,68,107,71,49,52,98,73,80,90,74,70,51,111,48,85,92,10,9,100,86,71,48,114,97,74,106,101,67,75,79,86,47,81,117,47,57,43,87,71,76,113,88,52,102,50,55,67,76,47,80,65,81,98,51,52,104,116,116,111,88,108,118,82,48,56,88,102,119,52,49,76,71,49,67,78,101,103,99,71,73,66,55,66,110,67,66,111,106,113,113,92,10,9,98,74,72,79,47,43,97,51,86,57,68,102,89,79,104,90,116,48,73,55,107,65,119,101,89,98,70,48,101,119,118,57,54,109,74,52,102,48,102,104,57,43,108,113,101,80,57,87,109,118,102,104,53,75,78,49,66,117,90,52,98,53,89,71,101,116,112,107,65,73,48,122,92,10,9,56,106,69,75,77,100,100,83,117,49,101,121,109,74,55,48,79,50,76,120,72,108,84,105,98,47,115,111,50,106,57,66,87,122,66,77,118,97,110,111,70,79,71,43,116,84,75,56,102,53,110,119,43,53,104,75,79,114,49,83,107,68,120,99,65,53,115,71,106,74,81,98,92,10,9,50,109,73,65,108,88,98,47,70,87,121,99,101,74,87,105,99,48,120,102,77,77,100,67,109,55,66,48,68,122,75,52,47,47,119,121,69,111,87,79,47,101,69,117,82,87,99,76,57,113,43,108,52,102,51,102,67,88,47,118,121,83,108,74,65,78,48,122,116,116,109,49,92,10,9,112,56,103,66,113,111,48,78,66,114,66,109,109,98,47,66,115,119,47,110,111,114,57,109,47,82,72,87,48,107,86,86,75,103,70,65,49,55,112,86,119,71,54,119,100,99,72,118,52,69,105,121,112,48,71,98,49,119,118,113,51,115,48,77,55,53,56,108,47,77,48,47,92,10,9,77,76,120,47,107,53,81,107,111,84,81,65,76,56,82,55,88,68,79,65,89,87,121,73,121,79,69,116,51,112,87,113,107,81,72,56,82,85,103,86,79,114,87,73,86,71,69,113,74,116,57,66,77,108,54,80,113,120,118,101,80,48,110,52,109,55,56,118,48,74,47,49,92,10,9,78,79,53,114,81,100,107,69,84,109,72,54,117,87,81,65,120,89,55,97,90,108,99,104,65,52,66,120,54,49,105,104,116,117,67,103,115,108,114,101,47,119,56,84,97,72,78,122,116,105,79,89,111,111,110,104,47,90,77,57,89,119,68,65,120,116,83,119,56,76,100,75,92,10,9,52,53,105,69,65,88,84,110,110,97,56,85,57,113,85,111,54,67,70,47,78,122,113,115,67,104,110,65,85,81,75,76,73,52,102,97,80,67,55,100,49,85,66,77,76,102,97,79,112,109,104,106,101,76,43,48,66,65,67,47,104,119,85,112,113,81,70,90,66,86,83,66,92,10,9,81,121,85,89,65,69,82,84,72,68,100,116,85,43,89,97,71,65,106,104,111,51,119,121,50,119,70,101,75,97,76,106,112,103,86,74,73,54,67,48,86,84,103,51,81,80,115,74,116,105,108,120,76,71,105,113,82,107,104,76,65,75,115,69,55,65,68,100,113,79,72,104,92,10,9,78,66,77,120,114,119,88,114,56,98,57,74,56,73,72,118,57,118,65,106,83,69,111,65,50,119,109,47,50,121,47,53,79,47,99,82,98,76,79,76,81,66,116,114,71,68,74,99,71,54,71,115,110,120,112,43,112,51,85,48,55,108,108,97,111,75,98,112,65,66,118,106,92,10,9,108,120,83,70,47,69,74,86,104,108,56,75,84,115,120,113,87,65,114,69,56,82,48,67,106,72,89,48,47,79,55,70,65,71,77,122,72,77,117,109,74,87,85,65,87,77,121,73,65,117,116,70,50,89,99,107,65,53,68,50,103,50,106,70,107,49,114,83,120,51,119,49,92,10,9,103,84,90,77,122,116,48,110,87,120,114,72,76,119,51,118,49,50,69,65,83,119,83,43,53,119,69,85,122,120,107,78,54,120,71,110,81,79,101,119,83,105,50,66,71,109,55,114,103,83,82,105,72,111,54,83,51,113,50,83,120,89,43,100,97,75,86,103,101,48,115,115,92,10,9,118,79,79,74,109,112,79,122,70,67,83,99,99,69,119,89,51,86,82,76,89,50,107,97,47,57,66,66,52,53,52,70,65,117,56,100,57,121,81,66,56,51,81,89,113,52,79,119,52,107,115,100,111,43,43,85,82,77,47,114,120,47,112,55,116,99,84,80,83,122,115,66,92,10,9,84,98,80,119,106,118,50,70,50,53,78,89,103,67,90,43,65,68,77,116,106,97,86,112,49,105,69,100,57,50,83,74,82,97,104,122,112,73,112,73,48,107,79,70,118,116,115,109,83,82,106,65,68,89,113,97,86,53,69,82,82,74,111,66,118,74,101,66,80,111,57,51,92,10,9,117,71,115,86,119,50,120,80,120,55,74,78,83,110,51,82,122,97,71,66,113,77,47,72,66,74,55,102,80,81,107,68,81,72,113,106,97,115,113,102,74,43,48,69,57,72,111,71,43,118,120,66,121,115,43,51,120,81,66,87,112,116,65,88,67,87,109,109,103,56,71,57,92,10,9,57,119,107,56,102,56,48,107,68,65,65,71,72,66,105,107,114,113,98,73,118,84,99,119,103,74,56,68,117,101,57,56,114,54,102,50,89,99,114,80,116,53,85,103,78,73,50,56,106,104,76,77,114,76,51,66,118,97,77,70,110,116,56,48,54,81,102,70,69,81,85,67,92,10,9,88,88,89,87,69,105,101,114,83,81,88,65,85,99,54,98,110,118,100,53,100,77,114,80,116,120,85,68,107,107,89,54,117,112,107,112,118,106,100,79,52,105,54,121,117,101,69,48,105,84,71,82,116,113,81,111,50,56,105,70,90,66,52,103,85,103,48,77,65,76,105,102,92,10,9,111,118,78,86,88,47,88,47,98,49,78,43,104,43,87,87,50,117,49,112,101,76,57,79,100,117,65,90,65,117,43,78,83,77,84,56,69,122,88,52,65,115,65,119,50,74,106,88,70,71,119,84,97,122,72,66,74,119,68,117,51,69,104,103,48,107,88,111,117,121,51,85,92,10,9,90,81,65,65,106,112,83,81,55,118,107,74,105,104,119,97,49,103,103,77,103,66,53,88,100,67,49,70,100,82,70,56,81,53,49,81,79,121,118,75,105,89,52,112,89,83,102,68,43,54,100,114,51,67,80,104,48,51,65,111,121,86,110,48,100,97,86,87,89,53,51,113,92,10,9,99,51,75,84,87,121,56,76,43,105,105,89,52,109,50,101,57,110,101,52,85,68,115,109,117,102,82,98,87,117,103,88,100,115,115,68,68,100,118,81,50,99,50,110,85,118,89,120,86,111,73,66,90,66,87,122,76,76,86,55,105,52,99,77,99,97,107,103,65,49,104,111,92,10,9,99,75,43,78,122,78,72,57,113,72,75,97,116,68,106,113,85,85,78,107,65,72,85,78,109,81,72,89,79,112,75,67,97,43,100,78,110,118,85,86,68,108,120,83,80,118,103,109,117,109,56,88,52,88,53,66,70,102,109,122,122,90,50,119,68,72,67,75,116,67,106,68,92,10,9,56,120,57,66,86,77,57,74,77,89,65,115,53,103,79,48,87,98,80,117,79,112,73,120,69,107,110,97,74,110,121,81,110,75,84,68,98,109,71,68,50,107,121,103,72,100,50,75,79,108,77,121,122,65,67,101,112,84,76,101,113,122,112,74,81,81,77,68,43,65,110,89,92,10,9,98,102,47,107,83,84,47,104,115,122,53,85,115,76,49,118,68,79,54,70,115,85,52,113,90,119,73,105,85,81,99,74,102,90,56,80,85,118,103,87,97,81,74,43,80,70,101,85,117,54,65,104,113,65,67,50,113,57,98,105,83,80,66,57,68,47,113,74,51,86,56,121,92,10,9,85,71,109,99,119,98,48,73,99,100,49,98,52,66,49,43,84,121,87,105,50,68,81,65,77,86,106,88,107,122,67,114,69,103,65,75,104,110,122,85,107,66,110,65,98,76,75,102,110,82,106,116,110,48,55,112,101,119,100,75,110,48,112,56,98,110,105,47,83,82,107,48,92,10,9,72,68,85,47,121,104,78,89,97,111,52,43,97,88,66,118,70,104,107,65,113,103,112,102,88,79,109,105,112,66,57,51,89,81,89,90,103,65,118,65,89,101,114,79,70,80,115,74,121,47,56,110,119,109,50,97,117,104,79,106,104,72,102,83,70,79,100,81,71,119,90,81,92,10,9,108,80,106,106,97,77,71,43,102,69,86,82,113,117,121,71,119,103,66,81,80,103,53,49,74,49,90,73,77,52,67,115,120,81,97,52,78,78,68,57,77,85,86,100,56,97,56,87,50,112,122,70,67,57,69,69,49,47,66,79,88,105,107,90,53,47,114,56,47,99,97,120,92,10,9,121,78,57,74,117,67,56,51,71,69,113,67,107,122,77,48,53,49,71,88,65,48,108,70,89,104,49,82,74,122,88,85,76,77,107,89,65,53,106,112,56,70,107,119,67,67,73,102,52,106,68,72,102,99,84,122,54,105,121,49,47,82,57,70,109,120,113,50,99,84,84,84,92,10,9,71,66,90,76,112,47,68,79,104,72,66,106,108,70,78,68,83,83,117,98,121,84,113,82,82,77,83,48,83,71,107,87,74,73,66,53,76,72,72,100,108,48,81,100,84,99,111,65,108,109,87,77,65,99,120,121,47,76,122,104,80,65,65,110,79,110,111,101,66,118,111,105,92,10,9,105,43,48,80,85,88,83,71,85,70,117,57,75,74,48,77,85,50,102,72,69,89,85,122,76,65,70,103,85,55,54,90,112,97,51,69,78,84,67,84,113,103,66,90,99,52,105,89,110,115,73,122,122,51,79,111,67,116,120,76,100,107,78,47,107,102,116,103,69,109,85,88,92,10,9,115,77,117,56,76,78,68,79,85,114,74,102,89,84,111,112,89,80,77,53,105,57,87,108,105,51,84,102,76,121,107,68,109,74,71,120,67,90,66,71,89,90,73,70,76,80,76,97,116,112,100,65,118,98,110,81,103,89,84,120,89,69,89,88,80,120,106,106,117,89,76,116,92,10,9,102,90,100,121,102,120,97,121,104,73,107,43,119,97,97,67,105,77,70,98,84,82,108,84,85,104,86,103,101,115,89,109,119,98,83,85,110,103,117,80,77,122,104,103,47,78,110,105,77,49,65,78,101,75,54,68,118,113,67,85,43,82,56,111,87,122,88,121,118,113,89,111,92,10,9,51,102,112,83,89,88,87,121,104,54,78,78,67,43,72,99,79,76,110,65,85,83,122,105,70,53,67,71,68,103,90,83,56,83,80,116,112,65,122,103,56,52,119,120,103,68,83,78,78,56,105,111,66,71,99,89,71,51,107,68,73,78,111,43,54,49,67,75,117,112,89,113,92,10,9,101,74,82,53,66,67,121,101,102,83,120,73,113,120,75,98,72,52,121,82,73,47,80,85,67,104,67,115,57,100,56,122,77,52,101,85,52,100,84,79,108,112,81,66,118,74,48,120,66,112,67,109,67,121,100,85,103,67,77,112,99,106,47,116,73,78,106,117,79,56,75,105,92,10,9,98,82,120,99,121,50,113,78,55,55,88,49,118,109,67,109,97,56,78,111,74,50,69,68,71,79,109,98,83,112,88,85,66,103,67,57,97,110,82,71,70,118,57,67,83,116,57,119,77,52,50,90,103,74,81,57,65,66,80,55,69,71,72,82,78,103,55,119,118,79,80,73,92,10,9,88,113,89,102,67,100,82,82,86,70,110,72,108,115,85,43,114,102,84,103,88,106,69,65,65,73,69,101,121,66,85,73,80,50,50,85,68,88,117,66,100,55,109,90,110,118,86,116,115,107,99,84,85,56,74,89,78,53,116,51,116,55,84,115,71,105,103,85,99,54,113,110,92,10,9,105,47,57,54,70,118,118,110,87,72,121,71,82,79,54,72,108,114,53,57,79,74,50,73,114,86,87,115,67,104,82,84,66,53,113,121,117,76,115,109,48,120,114,56,76,52,113,77,119,79,110,106,74,76,52,79,84,65,77,108,105,49,114,120,78,97,118,122,122,48,105,120,92,10,9,74,82,86,120,79,78,71,122,67,98,111,86,54,98,117,51,89,118,72,118,82,43,97,86,99,85,119,66,76,122,48,107,43,47,105,98,74,57,57,49,71,115,43,112,70,120,48,56,97,55,70,65,71,54,50,113,103,81,71,85,65,120,119,117,112,108,76,120,76,67,113,57,92,10,9,56,104,106,65,67,86,84,56,87,65,87,49,51,49,68,69,65,70,98,110,43,53,103,53,119,71,77,77,116,100,108,113,109,89,80,105,53,122,98,56,77,86,118,119,122,50,51,52,50,106,98,77,99,68,114,108,71,86,116,56,65,73,55,84,102,113,100,111,81,48,112,101,92,10,9,117,88,99,71,55,50,53,106,80,101,110,76,100,98,122,84,51,107,88,112,53,103,120,56,103,79,102,74,57,52,54,101,74,53,70,111,112,98,98,97,71,85,65,53,53,67,99,84,76,97,86,76,53,105,98,85,69,116,98,102,102,88,79,43,77,78,87,106,68,43,76,100,92,10,9,97,112,115,69,85,115,121,43,110,107,107,122,65,69,75,103,80,54,80,73,122,57,57,49,121,87,48,85,114,84,107,51,66,81,89,118,52,81,99,119,49,55,100,74,54,84,73,99,117,67,97,71,80,116,86,83,48,79,68,105,73,51,67,87,118,74,101,105,116,50,74,99,92,10,9,43,122,72,98,87,105,90,54,50,104,99,115,120,67,49,90,72,86,106,104,52,72,109,119,81,82,120,77,107,97,69,118,68,101,110,117,72,89,69,50,80,109,118,73,68,75,65,50,122,52,90,81,75,108,105,104,109,97,67,43,53,83,118,65,47,80,114,121,76,108,111,75,92,10,9,121,79,50,72,54,107,119,122,80,101,56,76,120,71,73,69,111,77,66,66,53,109,54,83,80,56,79,71,112,80,103,85,82,101,88,84,116,121,78,51,118,103,47,70,103,67,78,108,107,55,74,119,50,78,83,71,78,87,81,71,115,71,89,77,77,97,104,86,66,82,87,104,92,10,9,87,114,67,99,98,81,75,72,70,82,69,116,115,90,68,50,112,50,121,108,89,74,57,69,48,81,108,66,74,120,98,80,84,89,54,75,108,47,70,67,103,97,99,106,115,103,65,102,114,109,105,69,74,47,50,69,122,85,69,51,113,120,67,67,100,82,98,52,78,110,65,49,92,10,9,57,102,88,79,69,116,107,103,51,110,115,119,82,83,54,78,118,121,106,66,73,72,76,72,79,80,67,101,101,53,77,97,66,116,114,119,111,115,69,67,117,112,119,105,52,49,111,49,65,73,116,51,68,120,53,76,86,80,84,90,105,77,100,52,78,102,53,55,80,85,115,52,92,10,9,77,66,106,68,43,80,115,82,77,52,53,82,53,78,55,80,73,81,109,79,112,56,104,65,110,87,84,122,82,68,49,74,43,71,47,56,50,74,65,90,119,70,51,77,49,83,72,101,70,115,115,88,104,48,109,83,99,122,87,71,120,57,107,88,70,66,68,103,74,56,68,89,92,10,9,52,78,98,98,80,89,97,107,100,120,86,70,98,117,69,114,102,101,121,73,121,49,79,65,110,76,88,52,121,120,74,47,51,121,106,118,53,121,108,104,106,103,86,52,68,67,82,75,81,82,75,84,102,113,122,71,119,84,67,53,68,107,87,110,87,68,105,53,119,112,69,116,92,10,9,73,118,100,103,53,53,110,109,99,48,100,99,77,111,66,99,77,77,87,89,69,110,47,80,70,90,79,65,83,76,103,107,122,76,69,65,122,52,71,84,106,51,56,120,90,82,89,117,106,89,66,119,65,116,113,101,57,97,100,105,81,71,112,116,43,79,47,47,79,56,121,116,92,10,9,103,73,68,113,115,119,69,69,66,65,81,48,89,65,107,103,73,67,65,103,77,73,67,65,103,73,68,65,65,65,73,67,65,103,73,68,67,68,65,71,65,111,117,117,67,53,56,104,73,68,67,65,104,103,87,99,79,86,57,74,107,101,102,99,52,118,65,53,65,110,84,81,92,10,9,74,72,121,67,84,65,74,82,101,65,47,120,118,56,65,55,52,90,77,69,66,65,98,81,77,77,89,76,77,82,87,88,48,107,43,82,107,56,66,55,52,100,78,107,68,106,85,115,103,84,102,108,115,87,120,79,85,90,119,69,102,111,43,65,75,67,81,54,87,82,107,89,92,10,9,81,69,65,79,80,88,110,88,51,54,55,103,57,51,65,49,110,82,69,43,106,55,100,65,106,111,117,79,70,76,107,79,89,43,121,81,68,50,76,72,71,80,99,104,99,65,103,117,56,81,105,83,103,112,77,99,54,105,111,105,68,90,57,111,112,71,120,119,66,80,73,102,92,10,9,50,67,87,81,103,80,86,113,51,105,85,75,56,90,121,105,88,52,88,80,53,65,50,119,111,121,78,47,52,54,69,85,112,88,76,114,75,78,106,50,69,104,55,118,102,49,73,85,89,87,105,99,105,67,85,78,67,81,65,105,84,108,101,75,107,107,106,103,52,55,82,84,92,10,9,49,74,114,47,66,109,54,69,102,65,71,122,109,80,115,104,103,56,113,99,66,106,121,90,117,118,71,117,118,48,117,90,97,51,122,87,47,122,71,43,83,71,121,121,71,86,80,88,118,68,72,80,122,121,99,52,109,119,108,106,106,107,65,97,90,69,78,67,56,111,50,118,92,10,9,77,122,82,87,83,72,109,72,107,79,55,122,117,72,56,50,103,66,121,89,82,122,66,66,54,114,117,68,68,77,117,68,117,90,73,65,107,66,85,89,101,101,87,82,51,71,69,114,55,107,104,99,84,71,72,82,66,48,85,101,88,121,68,51,70,88,47,84,48,103,56,72,92,10,9,85,112,82,69,111,108,73,109,50,102,49,74,112,103,67,109,49,72,115,106,51,117,78,73,51,103,86,78,67,52,107,103,99,104,82,90,103,74,65,107,90,89,76,72,52,51,87,77,111,104,115,85,116,85,47,104,50,85,103,99,77,53,105,102,118,57,119,110,66,103,68,68,92,10,9,66,112,74,103,47,112,51,107,115,113,72,67,75,80,73,105,99,55,53,104,86,98,114,52,49,54,99,111,89,71,114,118,109,78,101,51,57,52,65,112,89,110,120,80,86,110,83,97,111,103,48,115,116,80,56,106,84,51,74,85,75,80,73,112,113,119,52,107,49,47,116,90,92,10,9,51,69,56,98,121,75,87,66,98,78,117,106,48,109,89,65,54,47,73,67,55,87,101,53,119,119,103,114,80,108,47,82,113,53,98,97,98,48,76,117,107,122,103,77,111,75,106,87,101,53,117,89,49,48,43,121,116,79,68,105,65,117,43,74,111,105,100,110,107,112,117,113,92,10,9,78,56,103,85,78,74,83,105,56,108,112,73,76,76,77,113,120,98,53,51,89,115,108,114,99,52,56,89,69,104,106,108,74,83,119,53,49,114,116,109,65,70,117,122,117,78,98,90,99,97,101,102,86,72,81,54,121,101,101,72,122,48,48,119,54,75,75,119,119,73,52,109,92,10,9,101,54,109,113,79,118,75,117,102,48,68,67,43,52,90,81,108,74,65,105,68,86,69,102,117,56,51,86,75,89,109,57,119,72,105,87,67,66,52,104,57,122,107,107,77,86,53,73,87,98,101,104,112,49,76,107,89,120,83,108,76,108,118,117,103,103,70,48,53,119,100,117,92,10,9,109,50,75,72,99,82,83,71,110,71,117,83,66,114,70,43,57,80,77,115,116,68,105,98,82,87,114,113,69,84,122,52,73,73,108,105,69,100,67,88,98,50,99,55,83,86,74,99,120,75,75,120,83,55,82,110,115,102,99,65,84,121,98,55,70,78,55,49,47,104,70,51,92,10,9,49,122,78,69,67,53,52,68,50,53,76,102,101,73,88,72,97,76,108,78,66,110,65,87,82,88,110,104,109,51,110,81,52,83,87,56,109,74,52,84,51,79,85,43,75,105,80,105,81,100,122,54,107,67,102,68,71,48,120,74,84,105,122,87,90,108,88,74,82,72,47,99,92,10,9,121,54,73,75,86,65,119,55,115,81,106,101,51,115,77,74,80,52,111,108,119,81,56,115,80,119,102,77,101,105,66,108,65,55,102,121,71,104,86,110,65,70,106,119,77,79,52,100,54,49,109,72,86,55,65,107,56,73,74,81,101,50,65,111,106,56,101,56,70,104,56,83,92,10,9,120,53,90,49,76,66,51,103,51,49,73,79,79,105,104,119,103,83,83,112,112,109,88,68,99,102,84,107,75,110,48,52,74,75,75,110,80,71,72,50,53,88,82,103,113,67,86,88,107,74,48,77,118,72,48,111,121,103,101,89,70,87,66,79,78,113,99,121,47,103,73,54,92,10,9,68,65,67,101,84,88,66,67,54,79,116,112,112,49,71,65,65,90,53,87,69,114,88,48,71,114,79,117,113,97,118,114,84,99,120,106,67,74,65,85,52,79,78,119,67,48,86,72,111,113,98,65,115,100,104,71,68,104,99,47,55,65,49,78,77,122,76,120,56,99,49,120,92,10,9,53,67,120,53,79,103,75,74,69,67,55,88,87,49,79,50,115,67,54,86,83,85,121,97,108,65,70,103,65,106,120,80,107,89,101,84,122,56,68,105,81,79,48,66,67,83,115,43,67,110,106,99,74,47,82,101,101,66,56,112,53,54,115,110,70,66,51,108,52,70,116,117,92,10,9,120,119,117,113,82,99,89,109,47,108,99,85,49,86,87,85,56,104,56,52,103,79,100,43,49,114,65,49,113,54,112,70,107,84,81,99,43,77,52,77,76,72,52,65,104,115,107,47,67,98,85,70,52,53,74,85,109,110,74,74,122,56,116,51,72,88,122,72,116,114,122,122,92,10,9,116,56,106,103,120,79,47,71,106,75,117,55,85,72,115,110,85,122,90,82,86,109,112,76,119,103,67,79,52,57,48,119,75,120,103,107,112,75,47,67,105,110,113,57,104,47,49,122,69,81,70,52,109,54,76,49,75,76,117,65,43,65,116,114,43,68,113,71,55,85,68,116,92,10,9,51,84,43,68,47,86,47,70,107,104,67,90,113,103,65,100,87,82,100,117,107,55,69,80,103,67,67,97,109,119,88,97,119,81,52,52,109,101,122,53,101,67,99,70,66,103,51,101,100,119,115,116,80,103,77,50,110,108,101,112,79,111,66,113,119,110,49,73,51,49,56,65,92,10,9,113,111,83,85,117,47,85,51,76,74,107,103,119,117,57,98,43,117,110,107,67,74,118,120,50,115,121,115,116,113,68,73,106,116,88,84,56,70,108,105,112,119,65,80,85,79,83,108,108,106,88,85,115,57,103,116,52,84,69,71,106,55,101,47,101,116,73,118,110,68,90,115,92,10,9,97,118,107,90,111,43,108,47,81,52,43,122,106,75,116,90,75,116,84,66,79,82,84,53,50,112,115,65,76,115,119,110,115,107,111,86,49,47,67,71,77,84,54,68,111,105,67,106,112,65,67,84,81,84,122,71,77,108,77,86,111,68,118,74,87,75,51,84,65,67,121,51,92,10,9,85,103,52,98,115,72,47,52,85,114,72,51,102,99,118,116,57,54,50,121,120,81,56,103,107,85,111,118,122,88,115,108,118,70,117,80,112,117,103,89,78,89,110,86,47,86,79,75,52,105,116,119,98,76,119,111,119,88,48,80,115,77,112,83,115,86,120,55,72,75,77,85,92,10,9,82,73,106,71,71,82,53,52,71,80,70,54,67,76,81,122,110,51,88,105,81,82,55,48,121,98,98,43,102,53,74,119,101,120,67,57,89,98,84,69,77,83,103,99,116,108,65,86,43,72,117,101,102,50,118,119,65,116,117,69,78,53,115,97,83,51,49,113,122,76,116,52,92,10,9,88,56,49,55,84,81,71,88,100,90,122,99,73,74,89,102,55,117,86,74,51,77,114,104,107,81,113,72,78,76,103,43,57,121,43,120,99,100,101,122,121,118,89,88,98,106,47,101,68,108,108,66,66,99,65,65,119,90,108,108,114,89,120,122,47,49,113,83,105,83,75,68,92,10,9,68,87,65,83,86,81,55,82,116,89,50,100,87,75,43,49,65,100,103,55,90,103,110,49,69,89,115,99,71,89,118,118,112,88,105,120,71,106,104,49,81,79,86,100,120,68,102,48,73,122,117,110,68,122,113,108,53,57,71,72,56,119,84,102,89,82,107,122,99,84,66,70,92,10,9,53,68,52,89,121,50,112,100,110,79,83,117,115,66,77,103,85,106,82,88,98,104,49,77,57,81,116,101,57,78,56,109,70,112,69,114,77,73,66,100,107,51,65,84,106,119,69,120,54,105,54,104,116,109,53,83,100,72,97,75,102,89,71,72,87,50,117,121,86,48,65,86,92,10,9,55,115,85,83,57,82,109,72,115,105,83,104,71,54,81,70,90,111,67,106,116,47,79,70,78,121,67,73,52,89,99,110,118,79,100,115,72,110,101,98,103,74,48,75,106,109,79,102,53,68,69,70,47,68,117,66,76,77,89,52,86,71,73,65,102,43,77,66,121,68,113,43,92,10,9,73,51,79,51,50,120,119,54,115,82,83,81,108,108,99,99,51,68,112,55,107,52,121,110,89,122,72,65,106,102,90,83,119,122,98,65,98,65,99,75,84,100,121,50,76,80,112,75,72,85,69,118,53,49,48,48,105,85,84,89,104,57,74,122,65,99,89,79,106,121,67,51,92,10,9,48,83,122,49,118,90,72,81,72,113,68,78,65,72,68,50,79,90,88,70,106,71,111,65,100,115,48,102,104,78,114,67,66,68,56,108,120,98,55,65,79,76,83,116,74,83,110,103,71,84,76,76,77,84,105,67,57,87,122,112,87,72,50,111,66,81,56,74,113,81,85,73,92,10,9,119,69,112,83,49,114,115,112,83,122,75,116,80,90,106,72,75,49,103,113,72,56,74,106,90,90,81,81,116,116,119,112,81,76,56,113,87,118,122,69,117,113,85,85,114,105,77,55,119,83,90,120,103,101,77,104,87,56,53,74,74,109,109,56,86,118,72,79,98,121,78,82,92,10,9,66,48,84,51,102,89,83,89,51,117,52,97,105,43,52,102,110,115,120,106,77,75,77,57,75,84,113,86,109,115,75,77,98,72,99,98,68,75,65,47,86,82,100,54,67,98,89,70,118,101,120,74,68,43,119,97,78,114,73,66,100,122,75,52,70,55,97,68,122,121,122,50,92,10,9,71,99,97,55,89,49,79,97,67,122,101,83,43,119,120,82,108,100,67,69,49,50,107,100,48,53,90,83,68,75,67,82,56,73,55,112,65,55,111,73,116,51,99,86,117,85,108,65,85,81,53,73,122,76,71,117,99,74,115,109,51,112,52,118,79,79,106,122,48,120,83,100,92,10,9,99,53,117,103,109,121,98,84,118,56,98,106,43,81,48,112,65,80,52,104,102,54,85,69,77,83,101,78,121,110,121,103,100,108,88,71,65,78,89,87,98,103,56,55,51,88,77,112,57,119,107,71,115,107,102,73,110,120,113,80,69,120,48,57,66,55,52,89,74,109,110,65,92,10,9,100,65,51,67,113,77,88,52,114,115,100,122,72,80,52,75,56,70,104,70,47,69,79,116,67,81,80,89,110,75,111,80,116,82,98,97,118,77,113,68,102,118,88,104,81,102,99,66,122,82,48,57,90,122,114,57,80,70,50,98,106,117,105,115,65,122,106,118,57,72,80,73,92,10,9,54,69,121,107,103,82,70,120,78,118,70,71,70,118,84,65,104,103,84,115,66,115,77,57,101,65,56,99,51,101,48,103,49,78,90,99,103,51,115,51,100,116,106,110,86,49,76,54,49,114,67,54,119,120,72,114,65,56,47,110,74,117,119,66,84,49,65,70,76,56,90,71,92,10,9,106,115,82,108,72,50,68,76,97,117,57,68,103,66,66,50,116,69,100,74,74,108,114,122,87,52,78,55,88,83,89,75,72,87,100,119,114,50,108,77,120,51,102,77,66,74,68,100,121,101,102,97,101,106,103,116,117,70,105,72,65,86,81,106,90,108,116,113,116,52,54,105,92,10,9,53,75,70,112,65,51,97,98,79,119,88,97,71,87,57,119,76,48,113,89,57,88,76,85,88,53,77,85,55,82,76,43,73,72,68,110,104,89,102,103,100,105,120,117,43,119,111,119,103,65,48,67,65,52,105,52,116,103,51,85,101,75,81,121,73,101,76,77,78,72,76,84,92,10,9,78,78,68,111,82,114,73,88,48,74,77,80,107,48,119,47,107,119,84,102,65,57,43,114,68,120,78,79,75,72,119,55,75,107,84,74,56,85,70,74,71,99,68,115,75,109,81,65,52,121,50,49,117,52,86,110,75,104,78,83,106,90,115,85,114,72,106,78,56,80,108,89,92,10,9,67,79,99,53,54,75,100,74,104,112,53,120,70,116,52,72,85,115,67,118,75,84,113,87,104,89,47,71,99,73,111,82,106,117,116,119,89,50,105,90,104,65,71,77,114,48,73,71,56,76,71,108,100,118,102,121,114,74,56,111,48,89,86,85,53,114,113,120,67,106,66,115,92,10,9,84,106,78,56,66,120,81,114,79,99,104,105,72,121,70,120,109,100,82,84,115,79,109,115,66,69,107,84,114,117,76,73,73,111,82,106,50,114,52,115,104,117,80,85,89,109,90,75,99,54,73,70,108,97,103,119,88,83,111,87,65,78,120,105,85,82,85,116,102,114,104,121,92,10,9,73,117,53,56,115,89,87,50,88,121,81,47,56,56,88,66,97,85,88,51,101,80,65,79,51,115,86,77,73,70,50,111,74,82,43,73,78,68,122,89,52,80,53,116,121,72,53,83,108,86,74,65,90,67,78,99,117,90,72,47,89,71,79,87,49,107,68,100,87,70,121,51,92,10,9,104,88,117,111,83,80,120,75,117,87,67,103,55,54,104,54,84,103,77,81,80,76,71,55,104,88,97,82,100,72,82,117,75,102,69,113,90,87,66,103,69,84,101,117,107,57,100,118,97,54,69,70,103,112,105,65,56,57,107,117,73,73,88,76,70,86,49,109,99,68,43,67,92,10,9,101,116,113,84,101,98,119,67,120,110,52,51,105,106,120,109,78,43,84,70,79,53,48,108,75,65,84,113,54,79,83,102,54,74,84,72,68,72,111,119,111,48,76,81,108,48,82,77,122,116,100,85,122,65,77,83,68,75,65,69,72,86,116,102,80,84,105,122,84,68,57,78,92,10,9,97,72,102,80,43,122,49,78,85,84,118,78,118,116,85,74,118,115,101,122,105,106,111,97,102,117,118,71,105,109,52,81,101,74,102,72,66,78,55,106,98,69,88,102,108,110,110,71,65,107,88,110,75,113,111,82,109,71,79,78,70,66,51,77,98,90,112,103,82,98,72,50,92,10,9,121,53,48,67,80,70,48,108,117,122,57,99,82,109,48,70,55,117,122,108,101,100,43,82,89,82,98,120,65,106,112,87,43,87,115,70,51,54,77,102,71,57,55,79,48,82,82,122,99,98,84,52,78,116,57,118,105,111,99,77,55,111,86,69,106,73,67,107,109,54,106,56,92,10,9,121,81,47,67,104,104,71,116,75,90,71,82,101,104,88,98,68,55,52,121,98,75,100,74,115,88,108,81,106,103,70,65,88,51,54,110,67,104,106,65,89,50,84,118,67,72,67,118,68,80,81,102,120,106,105,100,89,112,97,119,98,102,120,98,56,68,51,103,105,110,48,68,92,10,9,105,54,74,73,48,70,110,74,72,55,43,71,82,87,122,107,48,73,80,88,110,85,83,83,48,115,107,71,102,97,114,108,120,90,47,69,52,47,74,77,81,49,115,70,115,83,69,82,111,100,66,98,67,118,83,47,80,111,107,78,103,78,104,73,77,83,55,68,105,120,47,101,92,10,9,102,53,116,90,54,103,77,109,120,66,122,75,104,105,47,70,77,116,89,108,107,50,89,82,81,110,109,49,106,48,109,50,111,108,72,43,122,118,89,54,76,54,114,80,50,71,105,73,107,52,118,79,98,73,80,89,107,56,119,76,101,104,84,105,81,116,75,80,54,72,117,89,92,10,9,57,72,119,115,107,79,74,114,67,48,114,117,77,100,105,67,106,88,90,47,73,114,51,83,56,89,87,89,88,107,120,113,105,86,77,88,89,65,72,53,107,81,108,70,66,51,101,84,88,107,55,49,79,65,66,110,72,53,113,104,98,54,71,98,82,81,104,90,90,105,43,112,92,10,9,65,107,107,81,80,118,120,73,112,75,110,106,66,81,106,106,50,81,84,83,100,51,67,67,70,66,97,51,114,109,66,55,110,114,77,68,83,83,54,78,72,84,67,77,111,106,111,66,115,86,87,65,72,76,75,97,71,65,82,110,50,82,100,98,98,72,47,80,106,72,48,80,92,10,9,51,83,120,67,102,54,52,83,86,88,65,119,54,98,115,65,98,48,70,109,51,111,51,51,85,80,108,65,75,85,103,43,43,55,79,54,103,122,82,56,86,119,103,118,102,113,66,111,114,111,89,52,69,107,65,78,105,122,71,98,90,71,105,119,48,83,107,69,112,114,120,107,92,10,9,56,82,109,102,107,51,110,112,112,107,73,103,83,65,85,71,74,105,83,56,119,72,110,120,43,107,122,114,56,98,57,100,43,78,43,79,66,104,77,83,107,107,118,83,85,70,111,56,47,122,51,101,110,98,75,73,77,83,122,57,114,68,84,52,90,113,98,83,51,104,73,101,92,10,9,49,120,69,56,122,113,49,90,115,107,66,81,69,89,54,111,49,55,68,89,47,120,85,56,100,54,98,114,77,65,66,105,80,102,112,68,83,55,113,103,68,86,122,69,72,78,56,87,111,75,100,79,69,87,119,118,108,122,47,47,116,112,105,55,86,68,78,43,104,47,85,76,92,10,9,109,69,83,88,118,80,43,88,121,117,115,47,104,51,101,48,112,70,70,47,50,49,79,85,71,84,100,114,108,89,74,88,56,117,73,102,89,57,65,71,43,106,52,113,119,57,73,80,71,77,56,74,117,104,74,65,68,114,66,111,51,112,75,66,122,115,73,78,56,122,84,76,92,10,9,122,120,104,65,53,109,109,112,99,110,111,112,76,79,78,51,107,108,122,71,52,104,122,97,108,109,69,81,83,71,105,104,89,57,67,67,68,118,107,99,112,90,99,83,88,81,100,81,65,48,49,68,116,118,72,116,74,109,100,48,56,83,57,105,57,101,56,98,85,119,97,81,92,10,9,48,50,86,79,56,110,122,120,83,43,87,106,76,119,101,107,52,84,114,71,99,79,72,68,71,110,48,51,50,83,118,119,89,81,117,47,112,77,106,84,76,81,117,83,65,68,76,109,72,105,89,119,72,121,66,120,76,83,85,51,85,89,55,83,79,74,77,108,83,53,74,103,92,10,9,65,73,51,90,85,79,71,106,89,88,65,119,99,51,118,98,105,55,43,71,100,83,107,100,73,119,51,85,66,108,83,112,104,88,80,79,99,115,111,117,100,109,89,55,103,115,56,108,52,121,67,121,119,49,65,114,70,102,56,66,75,47,52,66,71,82,115,110,90,65,81,54,92,10,9,117,116,121,97,83,77,111,65,105,77,87,47,66,55,108,104,72,52,65,122,55,116,77,86,47,100,51,82,56,50,65,80,83,88,113,101,68,112,72,55,87,118,53,117,87,86,55,52,43,89,65,54,56,82,84,74,86,86,43,87,66,69,52,116,85,69,78,65,115,112,111,122,92,10,9,111,117,110,101,121,78,68,52,119,70,53,122,89,67,85,71,113,79,80,69,65,111,118,105,98,56,105,80,70,77,107,52,109,57,51,82,52,101,73,72,57,107,113,52,56,73,43,110,54,76,84,103,110,105,112,97,47,77,81,54,74,87,112,72,51,117,55,90,101,55,49,77,92,10,9,85,81,105,117,100,67,108,51,79,67,119,57,87,107,50,76,88,53,99,66,65,80,67,119,103,49,102,86,69,82,89,43,100,66,122,65,105,43,120,109,78,109,53,56,54,80,106,90,99,82,103,65,106,103,105,80,111,105,105,105,67,55,118,43,83,113,112,79,81,80,111,54,92,10,9,103,54,75,73,117,67,56,57,101,74,47,114,101,101,76,98,67,109,86,72,76,77,74,85,122,56,99,69,72,111,118,55,120,86,86,57,100,70,83,65,81,115,66,67,67,109,79,87,113,53,104,52,66,67,109,100,83,71,98,90,97,51,88,82,108,74,57,98,54,111,104,116,92,10,9,68,78,115,105,110,105,73,55,53,98,70,56,66,111,74,56,76,113,65,111,47,78,101,49,53,121,103,115,57,75,102,119,55,109,56,98,118,86,107,86,56,77,48,73,117,112,103,90,49,68,49,74,98,112,76,119,89,52,100,104,67,53,98,104,47,114,122,122,50,81,74,56,92,10,9,111,112,72,116,53,116,99,112,76,88,53,103,104,120,75,76,72,119,118,47,69,73,112,56,50,74,57,115,103,73,115,102,103,74,85,99,88,111,78,119,98,114,109,79,55,67,82,102,75,97,97,79,51,115,105,83,52,77,117,79,43,111,107,56,67,84,65,71,76,118,68,111,92,10,9,50,121,78,70,43,108,90,74,70,55,56,85,65,56,103,66,110,108,73,119,107,75,72,50,43,117,116,67,98,88,53,66,107,82,77,71,84,104,43,117,112,80,83,122,70,66,87,75,47,54,78,90,56,115,72,72,82,54,88,87,101,103,112,65,80,115,107,47,85,74,81,98,92,10,9,55,120,120,76,113,103,71,89,68,88,119,110,107,66,106,48,88,74,76,51,111,89,105,106,89,47,99,104,56,120,66,100,105,102,87,66,57,98,89,51,47,53,119,89,69,105,112,65,75,85,68,47,47,82,86,122,83,49,105,75,52,50,84,78,87,99,107,99,70,103,53,72,92,10,9,122,53,74,47,97,99,110,101,111,115,104,49,99,119,83,76,43,105,43,72,57,86,53,53,106,118,72,52,72,56,74,107,85,106,119,69,79,81,69,101,89,83,108,114,106,103,100,57,113,50,87,55,119,119,110,107,49,107,102,103,89,51,55,117,89,50,82,111,88,55,76,74,92,10,9,65,80,75,66,72,82,121,87,99,66,119,100,100,87,90,109,48,73,122,47,104,111,70,69,115,107,83,69,55,69,55,48,87,72,120,117,122,102,89,72,108,65,79,114,67,43,116,97,71,53,49,89,108,81,73,104,51,72,103,57,112,118,121,81,49,52,85,85,110,84,74,65,92,10,9,116,48,101,111,56,74,116,77,51,51,118,97,74,54,84,117,117,111,74,86,89,86,117,65,108,65,80,72,112,108,119,56,103,81,120,51,100,115,81,65,65,103,73,97,65,114,68,74,68,97,68,73,43,55,67,98,81,72,116,103,103,107,104,103,65,105,101,107,89,87,82,87,92,10,9,68,67,85,119,103,73,65,65,104,52,67,113,65,122,43,74,51,105,122,112,103,67,69,103,109,114,75,119,103,67,111,87,73,66,75,74,119,109,55,50,65,85,115,56,67,71,87,102,83,119,54,79,106,119,77,68,67,65,104,119,105,48,90,53,67,122,47,49,120,100,99,107,92,10,9,106,69,100,65,103,70,78,52,90,101,78,113,70,77,89,106,73,67,67,73,73,119,69,66,65,89,69,66,66,65,81,69,66,65,89,81,69,66,65,81,71,69,66,65,81,69,66,103,65,65,69,66,65,89,69,66,66,65,81,69,66,65,89,81,69,66,66,81,78,100,66,49,92,10,9,66,69,74,103,68,78,119,98,69,100,65,68,102,43,85,112,75,98,119,55,110,110,48,85,69,119,112,72,111,111,89,54,77,116,83,77,112,54,103,103,121,74,49,85,73,104,87,121,74,52,67,76,75,77,74,109,69,87,75,77,53,74,114,76,43,88,48,110,85,90,82,88,92,10,9,65,88,110,116,51,117,80,43,90,65,108,119,100,57,50,87,43,55,99,66,82,99,70,102,67,80,90,66,69,108,85,107,86,77,110,108,85,48,66,87,113,84,108,77,67,80,74,66,97,67,49,67,87,106,47,103,118,115,56,75,121,55,77,105,69,72,75,78,103,106,50,98,92,10,9,77,117,70,110,104,69,105,51,111,56,105,78,71,73,86,89,69,98,70,54,78,53,88,73,49,90,72,85,70,82,105,84,70,97,110,65,57,105,105,81,72,116,65,52,99,103,81,105,72,90,69,76,57,48,90,48,57,72,71,75,67,108,121,85,65,103,73,110,122,108,78,48,92,10,9,104,52,99,68,104,53,84,105,68,49,67,56,47,80,112,84,109,75,69,104,72,47,57,119,105,112,74,103,43,65,81,107,97,85,69,56,79,107,113,65,73,120,102,102,43,103,74,116,89,103,52,104,53,66,85,86,105,111,99,119,85,50,105,111,81,74,81,107,52,103,103,81,92,10,9,87,52,68,97,104,106,51,52,53,120,55,56,55,101,77,65,122,66,98,49,74,121,55,105,110,120,77,122,65,73,84,122,51,107,113,86,67,50,55,85,75,84,113,90,55,79,97,72,65,57,100,68,72,115,67,49,89,49,54,80,112,66,81,51,101,84,83,103,107,74,122,71,92,10,9,48,69,47,104,48,69,109,65,77,117,100,73,84,72,107,98,112,90,117,77,111,104,107,118,43,79,78,53,85,50,104,109,43,88,107,111,84,88,99,88,77,56,51,70,71,86,47,81,87,76,84,55,115,90,83,69,117,89,119,56,65,115,103,117,78,73,77,108,111,102,86,53,92,10,9,119,101,101,111,86,118,68,90,83,78,112,122,50,77,43,89,65,66,104,65,68,76,113,55,80,106,52,87,75,122,111,106,90,114,115,54,57,69,82,57,77,117,66,57,117,108,104,56,110,54,82,48,84,55,48,53,86,105,112,54,88,78,70,109,106,116,57,57,68,85,87,88,92,10,9,75,74,112,82,110,119,53,109,75,122,112,88,85,84,79,80,120,114,77,83,78,86,76,85,87,86,70,102,82,99,56,111,87,108,113,102,76,118,54,83,47,51,53,120,74,73,67,57,87,102,82,77,67,104,81,81,43,82,51,74,112,109,117,67,79,68,82,84,119,51,90,120,92,10,9,75,85,86,108,114,108,48,66,57,103,105,107,67,100,117,97,120,98,97,99,68,103,120,100,102,120,51,66,72,82,79,99,47,70,54,75,121,110,102,98,84,74,97,66,66,74,106,73,119,118,120,98,66,55,116,57,72,72,122,74,107,109,97,100,74,55,115,54,100,117,107,78,92,10,9,109,98,113,122,55,97,77,114,85,120,100,80,118,108,107,79,115,68,88,49,90,80,116,65,76,65,98,119,69,104,87,112,75,120,52,84,115,65,48,99,75,67,105,117,55,107,100,54,70,88,43,72,71,102,81,104,105,87,105,80,47,71,121,47,90,80,72,79,90,89,111,111,92,10,9,76,72,53,107,120,102,50,88,104,98,89,104,77,106,53,69,47,109,88,66,120,99,82,70,115,90,86,66,104,88,113,116,74,97,122,70,122,72,119,106,70,115,50,55,53,49,71,55,106,75,107,104,49,55,65,116,76,120,89,68,87,71,119,52,43,78,47,120,111,110,104,102,92,10,9,52,77,86,47,119,53,77,120,75,84,55,107,51,86,103,97,77,73,81,101,83,108,72,57,116,86,48,56,71,78,105,72,50,85,52,106,111,83,101,51,89,79,110,105,71,77,56,110,56,122,66,109,85,103,115,116,106,79,49,104,84,76,116,110,99,74,71,88,65,121,112,98,92,10,9,98,90,72,114,90,68,109,115,73,99,68,53,50,55,79,111,49,108,100,111,85,72,84,81,51,77,74,72,104,71,111,69,54,47,83,84,110,105,122,43,72,73,78,69,90,112,108,79,104,117,51,65,119,80,112,50,66,104,89,47,115,67,57,70,97,98,69,108,106,87,87,100,92,10,9,117,102,43,111,114,102,102,114,75,108,118,56,119,67,97,109,67,48,114,72,56,118,109,56,69,66,80,81,81,83,118,66,116,116,112,81,90,73,109,72,88,87,82,76,68,119,99,88,53,43,47,118,115,71,105,113,121,121,121,104,90,118,88,75,48,73,83,71,72,56,105,122,92,10,9,81,114,112,50,67,53,89,113,116,113,102,113,82,87,80,88,68,67,68,51,89,100,78,107,65,104,75,65,47,103,102,110,110,75,77,57,102,56,57,49,87,101,114,83,89,81,73,68,109,89,108,107,68,82,68,84,74,89,53,55,84,56,118,102,73,97,115,85,51,54,84,66,92,10,9,65,72,74,77,52,66,107,68,102,84,122,78,100,69,113,98,115,49,105,52,85,85,89,71,101,86,51,101,121,90,75,87,56,68,52,43,119,120,77,98,105,47,99,103,119,122,97,79,112,101,114,72,56,76,81,89,65,78,67,97,82,85,121,100,51,87,108,66,105,111,116,112,92,10,9,101,65,90,49,81,86,105,114,99,84,76,81,78,77,69,57,80,84,77,43,117,101,56,121,86,80,107,50,113,47,76,70,106,50,80,65,71,43,73,121,65,70,116,72,80,122,65,77,119,115,50,122,98,81,89,43,71,72,119,79,85,66,67,107,89,48,89,72,72,75,76,120,92,10,9,111,65,84,88,47,53,106,120,67,81,52,68,54,80,107,71,57,122,101,116,52,115,87,47,105,105,87,56,99,88,69,90,81,72,79,76,76,119,78,82,101,111,105,106,68,50,55,121,106,73,115,112,77,106,74,108,71,87,65,65,99,89,49,54,72,49,88,66,82,68,43,98,92,10,9,52,118,118,74,70,43,76,98,75,108,51,56,79,65,112,72,68,77,57,106,104,98,116,98,79,98,83,121,47,70,74,52,73,102,105,49,110,50,76,53,79,98,114,57,103,79,104,47,85,82,85,77,102,104,77,87,106,88,101,107,121,115,70,97,56,76,101,88,115,111,65,106,92,10,9,99,65,107,82,102,112,77,111,56,110,88,47,103,84,99,100,110,75,82,48,111,74,56,99,97,113,82,86,85,88,104,105,52,103,106,122,98,111,49,55,85,88,98,114,54,73,121,80,78,51,119,105,80,109,80,67,119,110,43,86,102,47,52,102,86,72,73,69,54,115,85,78,92,10,9,50,77,97,74,105,118,52,101,52,55,113,68,75,81,112,111,83,73,114,53,70,80,107,48,74,65,88,99,88,56,43,113,111,108,48,65,120,84,109,102,105,99,69,115,99,78,74,104,101,115,83,74,115,89,112,84,122,103,112,113,74,103,113,117,119,108,107,77,110,112,83,100,92,10,9,104,102,111,54,107,116,116,78,67,103,81,51,47,84,115,106,52,52,108,73,85,98,104,70,106,50,100,71,79,52,52,88,43,117,83,52,68,86,82,105,65,76,117,82,89,67,72,67,67,111,89,74,79,78,79,56,87,43,69,54,87,72,105,102,99,56,81,65,77,68,69,82,92,10,9,100,57,68,97,115,71,56,47,56,107,65,104,118,106,51,110,111,100,101,71,100,100,85,79,106,105,102,77,102,122,51,65,75,103,68,43,54,47,65,108,97,71,47,119,114,80,50,89,65,83,83,86,86,79,66,52,99,121,88,112,43,122,72,107,85,77,47,102,86,121,101,118,92,10,9,65,65,121,43,101,51,117,119,119,68,70,51,74,108,68,107,83,108,57,73,69,48,106,65,52,55,79,83,67,116,68,71,85,85,102,104,119,65,70,114,100,101,56,75,65,55,98,81,52,99,102,102,49,50,68,120,52,122,51,118,90,120,115,72,109,78,113,121,77,114,97,74,92,10,9,84,116,120,118,69,71,73,73,85,69,57,117,78,85,116,57,50,112,120,51,120,98,99,114,88,80,99,78,113,119,71,89,97,68,85,71,122,48,114,75,65,74,68,69,52,107,109,87,56,113,54,109,75,74,43,68,76,109,112,52,65,57,79,74,106,120,106,65,85,116,65,54,92,10,9,75,84,77,65,113,77,101,47,116,47,109,65,79,75,55,65,114,114,65,101,68,51,53,106,84,56,83,114,118,84,84,118,119,56,84,90,106,65,102,117,122,84,75,76,80,54,99,106,89,55,72,104,108,65,71,71,117,110,48,111,79,104,107,53,108,80,81,105,77,79,80,103,92,10,9,112,74,106,88,84,87,75,74,119,89,84,90,109,69,105,69,115,79,82,102,90,100,106,88,84,84,88,118,81,51,72,79,118,112,82,117,122,103,87,103,49,118,89,68,75,106,71,65,78,82,49,51,71,69,98,66,119,90,89,107,106,75,84,81,56,89,98,68,68,110,90,89,92,10,9,69,104,50,115,67,74,98,119,114,114,85,118,84,56,74,74,119,116,43,105,72,56,85,51,117,110,49,105,56,74,119,101,65,117,47,54,109,79,72,57,74,118,89,69,54,78,77,73,54,85,98,69,89,83,88,47,107,54,107,85,101,86,55,67,106,110,85,82,77,47,67,116,92,10,9,66,70,81,56,54,48,102,80,108,86,83,65,78,99,103,57,122,109,102,57,56,121,110,66,78,110,88,56,71,88,84,113,117,51,56,107,118,71,68,47,119,122,115,112,99,103,101,101,83,68,56,70,43,85,67,47,102,90,47,55,108,88,83,88,97,56,117,113,120,117,105,89,92,10,9,78,103,78,100,98,67,122,81,47,121,78,83,89,80,122,53,119,77,76,47,111,54,76,76,101,72,77,67,85,56,79,74,48,110,120,101,57,66,79,89,121,107,108,53,79,80,110,81,80,90,74,115,108,122,89,68,87,74,80,83,65,89,54,105,80,113,85,83,82,120,101,79,92,10,9,111,71,77,65,43,56,76,67,101,47,122,65,69,47,65,121,51,108,71,119,54,71,102,122,55,51,88,122,73,47,83,74,121,81,65,43,78,88,106,118,78,88,107,67,122,57,90,99,117,70,67,74,76,106,88,56,100,108,76,112,119,53,98,121,100,57,98,53,49,116,78,74,92,10,9,51,51,50,56,107,43,49,74,98,115,115,71,103,72,80,72,55,119,122,101,67,49,119,87,104,113,66,67,73,54,84,118,88,109,113,122,76,98,99,47,107,121,87,77,88,74,89,108,71,78,108,48,56,105,47,71,108,82,112,77,77,120,75,80,90,53,118,73,78,84,69,88,92,10,9,65,97,83,84,48,121,108,75,74,72,79,112,119,80,101,97,54,115,71,99,77,77,108,117,106,69,50,111,99,90,111,77,81,70,99,67,81,74,122,56,47,109,82,109,116,89,101,52,104,101,81,102,43,86,98,111,104,90,52,122,65,78,101,112,110,54,65,75,68,78,71,52,92,10,9,76,54,54,47,43,57,101,71,84,76,99,116,50,49,73,117,52,69,88,57,65,77,56,76,113,65,100,100,43,87,57,119,117,114,109,67,78,119,48,119,117,78,115,48,49,97,57,83,56,122,66,116,84,68,79,85,48,78,102,76,111,103,81,119,106,122,56,43,99,118,97,98,92,10,9,82,80,68,66,109,101,84,105,68,79,51,109,54,54,88,119,110,105,57,97,70,67,49,122,112,120,81,83,119,69,52,50,103,78,57,51,72,68,77,88,83,65,101,80,56,109,55,102,108,47,84,114,86,74,81,83,50,48,100,52,119,65,66,77,118,49,43,51,78,66,110,65,92,10,9,87,112,114,116,122,117,86,47,88,50,68,117,98,119,73,52,104,82,121,83,119,115,68,112,87,80,76,84,105,67,78,47,88,52,80,74,74,118,72,118,43,73,75,121,67,85,104,71,83,122,120,52,68,57,79,106,120,65,50,121,121,65,68,121,100,56,47,114,87,102,81,122,92,10,9,119,83,77,85,72,97,117,52,104,73,52,70,118,71,115,75,85,103,65,109,43,89,83,69,57,121,81,53,70,102,107,109,103,52,115,102,113,116,69,78,110,114,122,76,49,52,98,51,98,53,81,109,65,50,103,110,119,65,67,65,103,87,83,87,70,76,81,108,83,120,78,100,92,10,9,72,65,55,99,75,65,79,49,120,84,86,115,112,103,84,80,73,103,80,52,66,55,109,74,89,89,109,68,105,89,98,51,98,53,73,87,65,56,68,90,112,97,53,82,97,49,89,82,102,81,119,43,51,110,77,77,51,104,86,117,109,81,56,53,72,76,103,54,122,102,118,83,92,10,9,121,75,105,84,116,70,122,89,56,105,112,109,65,73,103,54,80,78,101,106,57,53,108,69,90,111,98,85,122,100,74,105,65,71,115,98,116,68,117,55,120,69,81,54,103,115,121,77,103,114,85,79,66,119,54,71,75,112,48,106,78,113,103,113,101,122,113,101,90,69,108,86,92,10,9,116,83,83,54,56,90,81,77,76,88,52,119,119,107,77,116,83,48,82,74,65,87,90,114,89,107,102,112,121,104,74,119,112,104,104,65,113,81,70,65,43,117,90,66,71,90,112,81,117,113,54,111,102,121,86,51,54,100,97,97,97,101,105,74,83,85,52,52,90,109,90,107,92,10,9,114,76,67,120,72,75,102,111,76,81,47,102,122,99,83,104,67,115,102,103,109,54,98,66,65,72,81,78,103,65,115,113,105,75,82,119,67,104,109,97,107,85,110,49,111,75,98,52,66,108,102,98,107,120,121,57,52,121,52,97,113,116,113,107,66,78,100,109,73,85,80,79,92,10,9,99,112,89,117,110,47,84,48,47,84,52,49,118,76,57,51,108,104,104,65,74,99,56,110,87,71,104,47,83,51,108,53,121,84,119,71,70,111,112,117,84,65,74,79,80,51,111,52,101,77,99,84,78,79,53,74,99,115,83,53,105,80,119,52,84,105,115,70,79,78,114,48,92,10,9,73,84,50,72,113,77,65,65,121,118,120,78,78,120,108,69,72,80,69,83,72,110,51,57,83,98,90,119,113,67,51,111,82,105,102,67,110,82,110,74,83,50,119,109,80,107,87,103,48,70,71,97,57,111,50,107,105,56,120,72,73,71,111,83,87,97,116,71,101,106,54,72,92,10,9,120,109,83,82,65,101,104,79,51,76,103,71,109,77,57,90,69,110,67,66,90,81,98,51,73,115,76,118,71,99,49,55,52,102,75,75,65,66,73,98,120,115,117,87,98,75,80,81,115,84,87,56,110,102,68,54,117,90,52,116,75,74,121,116,52,55,106,49,85,68,76,122,92,10,9,116,88,101,70,67,87,83,87,48,104,54,77,118,114,108,114,66,113,65,55,97,101,99,110,117,80,90,112,116,103,110,89,120,108,76,68,43,56,56,49,89,67,76,73,75,65,119,47,57,51,87,70,70,122,43,113,76,79,107,99,69,99,69,43,56,49,55,71,71,81,67,115,92,10,9,54,109,77,112,79,56,105,70,98,43,117,105,105,83,48,112,73,71,48,71,65,79,66,85,52,71,88,76,65,50,65,97,82,89,103,100,53,49,112,68,69,81,54,120,69,82,73,108,121,108,71,71,72,69,53,75,101,50,106,101,80,48,113,68,73,99,55,120,98,69,72,116,92,10,9,121,51,114,49,53,101,82,102,50,102,74,83,77,75,50,79,118,85,117,49,77,103,65,115,122,105,80,74,114,108,70,81,73,111,111,81,82,51,115,109,43,81,110,97,115,122,114,119,79,79,107,108,118,70,121,102,111,116,112,51,72,53,70,90,117,113,50,110,78,101,54,90,92,10,9,54,43,71,67,81,116,53,69,53,69,103,89,122,47,79,110,112,115,111,90,119,77,55,86,121,103,66,121,57,121,65,116,116,75,49,52,43,110,113,66,78,112,98,119,82,70,116,109,50,69,54,79,50,84,51,78,79,109,121,55,77,109,73,102,110,73,111,81,72,52,43,52,92,10,9,102,119,83,86,111,79,67,70,83,90,71,84,70,97,84,110,50,55,68,73,52,52,87,49,72,106,80,86,78,121,105,113,101,43,65,114,51,106,97,56,102,48,102,88,68,77,65,107,70,70,104,88,122,65,89,84,87,71,121,104,110,47,79,70,50,104,108,76,77,109,54,109,92,10,9,67,73,49,70,104,67,79,79,71,71,72,69,119,114,69,99,115,103,102,88,85,101,84,68,80,112,110,70,100,75,103,78,105,73,47,102,104,50,81,83,81,47,121,84,57,66,75,49,109,74,122,87,119,65,78,48,117,89,77,70,116,106,77,118,77,112,119,77,98,79,119,104,92,10,9,65,56,67,89,109,112,121,109,119,68,71,118,82,120,89,89,103,77,108,105,119,121,76,52,78,83,88,51,98,98,99,53,103,81,116,120,66,56,110,72,74,71,65,88,50,52,97,105,79,110,54,57,43,80,47,83,109,87,67,103,97,108,50,100,119,118,102,68,104,114,65,57,92,10,9,117,83,115,53,49,112,47,116,65,51,101,81,87,57,102,120,79,72,106,84,78,122,88,65,70,120,85,103,72,122,65,73,119,113,86,84,115,104,83,52,116,65,104,55,77,118,109,82,98,67,73,74,55,106,71,119,89,90,103,52,65,105,72,51,65,77,55,66,52,82,50,74,92,10,9,122,68,56,114,72,102,81,86,122,66,79,108,119,110,72,97,115,97,70,72,89,50,68,113,112,114,120,84,81,50,65,65,65,70,119,54,102,121,80,73,66,75,81,90,65,69,84,97,88,49,71,85,118,84,103,76,103,80,104,53,85,85,114,102,114,49,71,101,47,101,70,121,92,10,9,82,86,117,84,117,85,69,115,76,109,66,115,82,82,104,53,83,48,47,71,89,87,83,87,71,73,67,117,86,88,87,54,48,76,115,104,86,100,81,82,81,118,113,106,68,87,101,82,43,97,121,98,118,43,118,53,52,115,102,51,79,49,76,81,68,105,74,104,82,57,108,66,92,10,9,48,83,87,79,98,65,79,119,66,119,122,48,112,79,43,119,55,53,106,89,117,72,112,73,113,52,98,108,71,77,66,115,122,99,107,50,87,102,68,57,104,118,65,105,87,50,68,89,122,104,104,76,65,52,112,70,104,82,112,121,100,90,52,117,102,107,104,81,118,120,88,89,92,10,9,101,85,120,79,85,89,113,86,86,52,77,97,103,75,111,47,118,82,49,74,65,48,100,53,77,104,55,111,57,50,105,68,43,51,69,121,49,77,52,86,65,57,68,82,86,53,52,108,101,81,80,101,67,66,90,57,100,68,79,114,52,72,49,115,82,104,43,67,67,99,65,120,92,10,9,53,83,72,80,70,106,43,89,77,85,112,107,80,121,72,85,82,120,78,57,118,66,81,43,89,87,110,103,80,76,74,122,43,112,80,68,112,104,54,78,105,43,108,120,52,70,120,88,68,79,65,87,68,101,53,50,112,97,87,80,66,113,115,117,114,79,81,54,108,89,72,82,92,10,9,106,120,107,79,70,116,115,65,82,87,101,81,117,90,43,65,66,72,68,48,116,114,118,81,52,106,101,49,65,83,121,75,77,87,57,117,52,69,85,54,122,78,76,51,97,79,73,82,65,122,65,53,67,82,103,114,114,84,97,86,89,119,67,118,75,98,111,49,81,86,117,89,92,10,9,47,66,57,98,47,72,65,52,84,107,73,65,121,68,107,85,51,121,111,57,105,110,86,78,86,55,105,100,111,109,79,56,48,83,108,78,76,107,103,55,56,66,118,89,103,118,82,122,71,104,98,68,74,73,78,55,74,121,100,52,120,110,52,115,114,107,56,88,47,105,52,43,92,10,9,90,84,87,67,82,75,116,55,114,80,113,103,57,77,116,85,105,105,81,55,109,56,87,122,99,108,122,56,97,120,97,66,55,51,98,119,56,97,67,76,119,104,49,50,121,120,104,54,57,55,80,56,88,107,115,100,68,122,67,56,47,72,66,101,105,120,113,72,114,103,113,90,92,10,9,103,67,72,105,109,65,56,74,74,77,56,85,115,74,107,85,65,113,110,82,100,75,115,69,74,97,49,121,68,75,109,108,74,52,43,122,86,67,87,111,52,82,52,120,65,73,122,86,118,82,114,51,102,99,55,77,88,82,81,49,57,102,87,120,55,68,117,111,69,72,81,52,92,10,9,84,43,119,79,76,73,90,56,121,86,76,67,83,53,82,79,121,83,54,99,85,115,66,112,54,67,83,50,69,97,122,79,117,105,112,99,81,117,56,107,118,89,73,90,48,111,68,47,80,43,111,105,110,69,112,82,102,103,66,112,81,73,100,71,121,118,84,55,121,72,52,101,92,10,9,118,68,77,49,49,69,75,111,81,122,49,73,51,122,67,56,66,84,57,122,100,56,78,51,51,53,122,77,75,104,49,76,65,55,52,82,56,70,71,73,109,56,111,78,51,50,56,51,115,112,67,103,78,83,52,68,67,68,66,68,76,100,115,73,73,78,53,117,98,57,65,79,92,10,9,71,67,49,56,68,47,53,78,48,81,109,74,121,53,66,89,71,80,79,101,112,50,81,82,106,88,68,71,117,85,118,103,50,87,68,48,49,49,71,85,73,68,77,112,115,66,107,77,57,72,66,79,100,79,89,120,114,68,81,102,52,78,112,56,105,105,48,71,72,120,105,65,92,10,9,101,48,67,67,81,104,109,115,98,88,108,51,82,74,82,102,70,47,112,53,116,82,54,111,88,68,66,99,122,109,98,82,101,121,122,84,83,65,118,105,102,82,73,103,57,80,90,109,113,112,122,118,69,79,57,47,104,114,68,79,117,104,112,76,73,83,103,86,70,57,100,78,92,10,9,72,101,88,86,68,121,66,47,85,53,112,66,66,84,43,97,67,88,107,106,107,73,90,118,74,97,117,82,100,102,122,57,114,66,54,84,66,103,89,81,111,65,78,77,49,116,80,90,120,116,73,104,55,47,101,102,115,117,48,70,102,118,105,50,107,111,109,117,121,85,122,103,92,10,9,68,71,89,75,112,88,65,47,77,52,122,70,89,98,103,67,65,119,105,119,104,49,89,115,118,99,119,104,116,48,101,103,54,43,89,120,73,82,119,104,73,109,85,87,111,117,49,101,89,119,89,48,77,103,120,78,89,65,65,66,65,81,69,86,100,74,67,65,103,73,68,65,92,10,9,65,65,73,67,65,103,73,68,67,65,103,73,67,65,119,103,73,67,65,103,77,73,67,65,103,73,68,65,65,65,73,67,65,103,73,68,67,65,103,73,67,65,119,103,73,67,65,103,77,73,67,65,103,73,68,65,65,65,73,67,65,103,73,68,67,65,103,73,121,66,120,99,92,10,9,53,48,111,68,119,48,71,117,57,111,51,53,51,50,52,85,66,88,86,48,112,75,106,48,69,83,75,57,69,72,74,97,109,77,99,100,97,98,49,82,87,81,98,49,65,104,52,107,78,56,85,108,100,78,71,87,111,116,122,51,121,70,113,69,68,68,50,111,56,111,79,81,92,10,9,88,50,82,122,82,100,66,77,102,109,48,47,74,68,66,66,113,97,54,112,70,73,87,65,73,113,85,97,107,114,69,105,113,109,53,86,109,74,54,90,119,122,111,85,86,87,47,117,121,88,77,56,78,55,102,98,77,87,71,57,116,83,54,120,56,83,76,88,65,49,75,70,92,10,9,73,97,65,75,52,99,117,73,89,107,84,105,110,89,88,56,78,56,121,86,43,111,74,47,56,102,117,90,70,67,85,75,101,89,85,48,77,110,110,98,68,103,90,67,73,111,120,100,70,102,87,104,75,80,115,114,70,115,88,113,104,109,48,105,76,104,55,108,110,121,90,52,92,10,9,78,80,66,73,98,52,49,67,110,51,118,122,52,106,101,116,86,65,117,109,103,76,66,97,70,80,73,99,81,84,76,70,84,81,80,115,122,71,47,107,71,43,106,76,99,122,122,78,75,107,84,89,77,74,65,48,70,120,109,111,118,107,121,84,65,87,67,51,79,52,121,105,92,10,9,66,74,52,55,87,86,73,122,115,71,77,105,43,101,98,51,75,88,53,119,104,77,69,105,53,47,54,90,122,80,70,116,52,81,116,70,78,49,73,85,51,55,52,56,114,68,107,118,115,67,117,80,79,54,112,68,78,102,80,115,51,88,55,103,57,51,114,78,74,81,79,65,92,10,9,87,73,116,113,116,48,105,57,116,76,117,106,106,106,55,77,67,57,65,49,107,66,111,76,71,88,69,117,112,53,56,110,119,55,65,78,105,72,109,111,84,80,120,48,87,72,43,112,65,90,118,79,84,81,55,110,117,67,52,87,115,67,81,54,48,84,89,68,81,69,97,87,92,10,9,85,49,110,115,87,77,100,120,74,51,57,107,97,87,79,97,119,50,100,67,114,48,77,66,107,79,49,84,72,70,121,73,101,83,101,83,110,88,74,110,112,81,68,57,57,83,65,87,100,87,72,88,50,73,104,43,113,118,105,68,82,67,66,73,87,52,90,83,53,113,106,68,92,10,9,104,122,82,99,107,112,109,89,111,85,78,51,52,55,70,101,122,114,111,118,98,67,81,122,72,80,89,102,84,80,56,121,105,106,73,82,78,97,90,115,52,72,107,101,77,121,115,77,65,68,114,117,56,98,119,76,114,112,100,105,74,47,69,79,68,122,112,54,49,111,71,115,92,10,9,107,55,102,50,89,72,65,120,43,90,69,111,56,50,51,76,122,43,110,71,89,51,120,69,65,108,69,88,104,105,105,107,105,76,43,70,98,82,109,54,78,104,87,77,55,88,53,108,57,79,113,120,76,65,85,105,73,47,73,56,105,57,56,65,66,117,108,47,85,98,74,107,92,10,9,113,76,52,65,52,47,100,49,50,83,118,65,65,66,76,83,104,111,112,101,114,47,99,68,103,122,88,101,88,52,101,79,85,102,82,106,118,86,57,89,112,117,104,73,105,51,48,43,83,57,70,83,103,47,101,98,114,43,103,56,82,85,48,83,80,76,79,51,111,108,99,84,92,10,9,80,109,101,101,111,116,56,114,97,109,122,104,71,122,82,84,57,74,47,54,55,79,76,69,83,110,49,77,97,113,67,68,112,102,115,106,78,111,76,52,103,66,97,79,100,118,54,72,121,84,43,102,67,101,122,73,106,55,73,54,73,65,50,85,54,107,76,50,51,43,89,71,92,10,9,98,101,68,73,69,54,109,56,51,54,84,75,49,110,72,89,107,70,65,115,57,66,49,87,77,53,75,103,108,111,50,107,114,55,79,97,73,65,108,73,77,110,48,121,98,76,80,89,111,78,73,70,83,83,89,49,83,109,119,57,82,88,89,75,88,79,106,67,57,105,107,65,92,10,9,102,66,85,101,73,88,56,100,112,104,114,120,74,68,49,99,115,69,49,89,116,56,56,82,98,71,57,55,116,103,47,115,87,52,90,82,80,77,47,54,116,99,108,51,120,111,107,84,83,114,86,118,73,47,84,101,79,71,111,101,107,72,71,106,90,87,77,112,66,106,66,89,92,10,9,48,86,56,56,55,75,68,78,71,110,121,119,99,57,122,76,79,52,122,80,119,66,106,43,81,57,71,79,81,106,114,106,116,82,98,101,69,89,115,99,108,90,111,75,54,119,108,103,77,48,71,82,107,51,50,69,110,111,78,75,84,75,43,83,117,90,69,87,107,115,47,49,92,10,9,108,72,49,77,107,87,65,65,79,72,114,54,111,52,101,100,81,57,55,53,47,49,104,115,47,56,103,77,105,88,57,81,66,53,55,109,66,87,67,67,83,54,108,56,114,110,49,84,82,111,88,54,104,97,102,108,55,85,53,52,53,43,48,115,77,66,118,85,65,116,122,67,92,10,9,111,65,50,85,86,101,57,97,66,81,120,103,82,77,86,100,114,115,73,112,119,66,52,85,117,82,106,54,75,65,73,102,120,43,75,53,114,99,107,54,110,108,87,65,76,79,70,70,116,108,110,111,72,79,50,103,50,115,53,77,99,117,80,89,65,103,115,47,106,108,81,118,92,10,9,116,80,105,77,121,97,119,79,54,66,121,88,83,113,111,83,97,81,69,49,77,110,99,122,89,81,65,52,55,107,74,70,48,110,85,57,55,66,121,79,109,77,54,50,50,80,52,66,114,74,100,109,69,98,43,106,121,71,115,119,75,89,53,108,86,99,73,70,85,77,97,56,92,10,9,67,90,109,55,84,70,99,67,74,73,72,57,69,106,74,69,71,66,75,47,121,102,106,105,88,56,52,50,107,89,112,108,120,99,114,116,55,74,100,53,117,80,106,104,65,78,76,102,56,117,76,80,83,82,100,90,120,100,56,111,99,116,120,74,105,116,48,99,118,109,78,84,92,10,9,66,52,117,102,50,76,90,119,87,115,74,55,57,115,51,52,52,108,47,66,69,108,97,115,109,111,75,108,74,65,66,52,57,88,49,108,85,82,56,69,85,69,66,121,69,110,78,98,69,74,120,71,52,77,75,52,107,71,108,112,119,98,85,73,47,118,110,97,119,81,100,69,92,10,9,110,43,72,77,48,105,114,68,107,43,66,50,105,109,114,110,74,99,69,111,83,116,102,68,48,82,98,109,115,55,111,120,77,43,98,49,77,80,120,75,72,97,50,117,53,68,107,79,78,87,81,117,76,56,52,86,80,76,100,122,68,109,85,116,43,101,99,99,49,98,73,100,92,10,9,81,52,100,66,49,105,107,54,106,54,74,84,108,49,103,111,70,81,52,56,85,72,106,120,114,50,75,79,66,65,118,116,83,80,55,53,87,48,56,110,122,65,52,87,70,118,56,89,105,115,54,112,69,102,75,98,99,50,71,70,119,87,53,84,51,113,85,50,70,110,52,101,92,10,9,51,76,78,118,111,109,81,82,107,49,50,112,79,111,69,70,57,86,100,70,74,56,83,56,47,104,99,67,122,52,84,54,67,66,43,73,116,48,103,47,100,66,51,118,106,100,68,121,78,90,108,121,122,75,69,90,122,56,57,87,76,69,107,116,53,114,85,69,103,57,47,85,92,10,9,112,65,56,112,74,103,71,65,56,48,119,82,69,118,56,110,56,109,54,69,79,80,53,112,81,111,79,74,111,54,111,49,109,72,80,79,53,110,97,110,67,69,52,89,110,69,100,102,74,100,84,87,83,120,83,100,111,73,121,116,99,66,48,99,113,119,97,122,51,105,89,70,92,10,9,120,67,119,77,83,72,68,57,106,49,83,57,67,87,76,81,116,53,52,120,71,101,74,85,119,55,107,47,84,78,72,43,108,74,69,81,55,109,73,77,65,72,88,114,51,122,70,115,70,43,73,87,106,103,43,102,111,79,75,74,76,86,65,72,72,86,90,87,101,80,74,57,92,10,9,122,120,74,66,89,82,110,110,122,104,82,70,88,50,51,78,47,43,76,54,85,110,69,72,69,75,47,103,70,122,43,85,97,89,55,66,117,122,57,79,48,82,71,103,75,97,54,107,54,70,103,116,55,107,84,65,52,114,117,81,55,53,80,81,106,49,102,119,57,52,111,114,92,10,9,43,108,90,55,122,111,71,55,87,84,75,113,104,77,86,107,53,109,71,75,117,73,108,47,90,117,97,114,70,80,69,80,118,116,68,81,47,47,103,108,82,87,117,86,56,68,51,71,55,120,57,85,116,76,76,103,110,104,87,75,80,108,86,85,112,43,103,84,82,81,115,77,92,10,9,110,114,43,89,89,119,84,87,48,118,84,47,72,105,51,103,103,47,50,48,103,102,47,53,99,89,75,43,52,73,77,83,80,70,99,72,48,57,106,102,102,76,107,68,118,47,98,88,68,79,57,102,113,75,105,86,112,101,43,81,106,50,48,100,120,97,101,73,85,68,69,74,92,10,9,65,74,70,80,47,84,88,53,67,88,119,71,99,73,82,87,76,72,70,70,74,57,54,108,117,122,106,105,98,86,65,80,107,67,47,103,120,89,84,51,81,85,99,51,105,102,80,72,66,43,49,66,90,104,109,76,114,109,74,86,120,66,82,81,106,84,90,103,69,84,105,79,92,10,9,119,83,112,112,113,71,118,117,114,76,109,102,111,105,72,48,56,51,82,110,107,110,105,98,110,51,79,121,111,106,115,77,50,115,70,82,53,54,79,87,74,83,71,99,73,110,122,66,80,122,101,106,110,57,76,98,116,97,83,102,102,67,122,97,115,77,84,88,105,72,56,71,92,10,9,109,117,100,74,72,113,51,121,118,109,85,116,83,52,83,119,49,43,87,121,97,97,50,87,90,54,78,98,110,102,56,71,119,47,109,55,98,77,83,99,98,113,73,67,102,77,76,71,113,97,82,65,74,112,76,117,74,85,84,79,82,106,121,73,114,113,51,77,113,57,105,105,92,10,9,43,48,67,67,101,53,89,89,71,107,66,72,67,117,106,121,84,86,107,116,50,108,122,103,71,56,65,120,54,73,85,89,49,50,72,83,100,78,81,119,100,117,86,105,122,103,57,108,87,52,43,48,72,81,72,77,97,54,115,56,79,119,113,77,109,55,114,72,119,67,43,122,92,10,9,102,108,55,78,113,116,66,56,90,115,105,118,120,57,85,55,67,54,71,98,50,79,80,104,77,118,114,109,69,90,84,79,69,82,80,54,104,51,106,120,74,76,55,109,112,113,99,102,55,119,117,56,57,52,113,89,43,109,111,99,120,76,86,110,122,78,82,111,79,122,43,47,92,10,9,52,120,67,121,52,57,110,51,101,73,69,82,57,87,76,83,84,119,97,121,66,47,49,118,119,116,108,67,76,77,111,52,65,54,106,108,115,86,104,68,108,119,71,48,49,88,122,119,113,50,88,43,100,110,121,75,72,119,82,57,104,71,101,99,113,48,81,101,85,108,53,107,92,10,9,107,74,105,71,67,98,82,122,85,69,122,82,88,71,100,82,70,99,89,101,119,65,110,112,71,101,72,118,101,88,48,82,73,57,50,116,109,109,48,49,112,56,111,79,84,55,77,112,43,48,67,71,55,99,78,48,71,89,65,117,53,112,82,53,120,115,52,112,102,120,65,99,92,10,9,54,53,122,104,54,70,107,109,75,99,117,98,77,79,101,71,79,121,113,79,114,89,89,75,55,81,104,55,120,114,104,117,118,69,98,98,51,89,118,77,73,98,103,105,83,47,108,52,119,75,70,108,84,74,72,102,51,50,47,119,110,83,116,57,105,50,108,85,72,100,103,115,92,10,9,55,111,83,84,119,109,112,108,74,109,66,76,68,122,52,73,81,108,69,72,120,57,68,120,70,117,81,90,90,110,83,65,49,79,99,68,87,79,74,65,118,49,116,120,101,54,118,122,47,49,118,121,73,109,43,86,57,47,100,97,43,114,110,104,82,120,111,72,115,118,53,98,92,10,9,68,112,57,114,116,78,117,77,71,100,86,110,66,82,115,66,111,117,110,113,66,78,53,55,97,66,108,112,90,85,82,77,120,108,97,73,83,116,71,72,56,68,98,100,115,81,111,89,119,70,74,100,66,114,66,83,107,122,70,115,88,109,75,83,114,101,98,74,66,57,109,65,92,10,9,117,87,73,108,112,53,119,108,104,103,120,103,65,80,109,88,83,75,74,80,106,71,118,71,97,98,97,57,81,119,69,68,73,70,54,99,115,78,89,80,78,72,122,118,53,121,111,119,66,120,48,71,65,74,43,83,109,106,73,98,119,87,100,86,73,103,70,111,71,119,71,110,92,10,9,97,106,54,119,86,78,76,69,54,82,52,90,86,117,75,107,109,54,111,87,69,84,65,102,109,55,74,101,87,69,110,99,49,114,71,65,55,49,51,105,57,47,67,65,110,71,122,119,122,108,65,106,80,105,55,122,57,53,99,49,50,50,49,68,53,86,78,108,102,86,119,70,92,10,9,52,119,51,112,54,121,86,100,66,113,68,114,86,114,116,55,109,85,107,50,122,74,77,80,48,55,117,66,77,111,68,99,43,70,82,83,102,88,83,107,65,74,121,119,70,80,77,102,119,76,71,119,121,85,110,71,71,120,88,43,80,116,70,103,115,57,113,111,122,78,57,71,92,10,9,90,51,105,77,99,101,121,78,122,78,85,72,85,56,122,83,99,115,86,69,102,86,105,120,100,90,74,43,53,103,112,109,88,70,51,107,98,48,103,119,50,100,43,68,68,120,84,72,118,43,72,98,75,109,85,65,79,73,89,100,85,117,69,97,117,73,65,110,68,89,98,66,92,10,9,113,82,72,99,120,48,99,86,43,100,116,76,118,70,80,114,112,78,83,79,73,56,76,67,81,78,104,90,111,43,49,121,83,85,111,82,108,102,112,108,66,83,89,82,66,119,116,89,110,86,55,73,47,56,47,86,47,77,115,120,120,53,119,82,99,49,54,101,54,118,49,68,92,10,9,110,104,113,54,106,72,47,79,49,81,98,77,49,81,77,69,52,71,105,88,99,53,51,80,49,82,68,69,51,43,68,68,107,121,103,66,83,112,77,83,107,43,66,89,122,85,54,102,121,89,116,57,85,90,72,66,118,74,102,43,78,121,101,99,97,56,84,74,71,106,117,43,92,10,9,83,104,110,65,108,106,71,117,81,81,90,102,110,87,112,76,66,53,100,103,65,77,67,53,114,67,89,107,57,84,75,77,52,56,71,74,67,88,43,103,120,118,116,50,114,102,68,51,108,119,85,89,65,75,73,66,47,43,76,55,112,71,104,85,89,104,76,111,111,105,79,86,92,10,9,122,104,43,73,120,65,120,47,84,55,109,47,55,87,74,79,113,111,98,75,65,79,111,48,50,121,53,51,53,111,122,84,104,100,115,83,116,118,99,107,120,102,79,110,43,77,90,103,110,112,111,121,110,48,111,65,52,43,117,81,82,81,98,119,99,90,54,111,111,89,79,76,92,10,9,75,73,114,103,75,119,84,69,109,66,78,53,78,120,103,101,85,48,101,66,82,57,120,89,110,104,67,88,56,85,83,68,117,68,110,83,52,106,102,53,116,69,111,90,81,80,115,89,69,51,43,67,112,103,113,69,69,79,49,121,120,50,117,88,85,72,122,106,50,105,99,85,92,10,9,80,52,117,80,114,114,50,113,107,107,69,85,84,109,49,122,68,98,56,51,106,110,111,102,73,77,57,76,105,84,85,112,115,86,67,104,75,47,55,79,111,69,48,99,48,101,66,52,113,70,104,81,119,105,116,77,79,67,101,72,103,120,67,99,83,84,114,108,54,84,97,122,92,10,9,101,66,75,79,52,119,108,90,54,80,66,82,81,51,113,120,67,110,71,66,90,121,78,69,101,97,48,113,90,65,73,73,85,112,111,82,89,47,76,47,82,113,80,116,73,54,104,48,71,68,108,48,87,55,106,104,51,108,86,71,87,111,67,79,43,121,68,118,110,72,70,76,92,10,9,102,99,50,49,120,65,67,119,56,84,120,66,121,100,79,74,70,81,74,120,66,48,104,99,79,121,66,80,112,47,99,75,112,86,75,67,57,83,72,122,108,78,115,52,84,48,86,107,49,70,84,104,100,52,90,48,111,101,116,118,80,121,102,109,119,118,52,47,84,100,48,121,92,10,9,72,122,65,67,102,99,50,84,102,122,72,84,80,76,97,80,76,79,90,47,53,43,102,57,47,65,80,102,107,55,115,50,108,120,52,78,80,121,57,104,99,100,101,48,56,103,51,48,43,52,99,114,88,72,77,77,54,87,86,98,66,114,80,118,72,69,79,121,81,55,114,117,92,10,9,47,115,122,69,87,47,69,105,47,112,65,51,110,85,107,74,110,55,107,53,54,82,51,98,73,86,113,118,85,104,97,109,88,118,120,101,69,118,105,69,112,97,65,88,121,77,120,84,116,70,67,100,98,99,102,122,117,83,51,47,109,48,117,85,48,53,119,51,52,106,107,115,92,10,9,48,89,54,103,110,119,121,81,70,83,85,65,65,69,89,55,72,76,78,115,97,80,67,67,109,47,67,79,99,76,105,104,88,97,69,81,74,110,69,70,99,99,43,108,51,120,74,103,65,65,117,90,87,97,48,83,54,118,100,85,65,81,98,81,78,99,89,49,119,51,107,51,92,10,9,84,112,113,85,66,69,70,107,102,86,109,54,75,52,101,80,83,101,54,115,102,97,72,109,102,88,70,83,118,111,51,104,101,98,117,76,119,72,118,67,65,101,48,90,51,103,65,43,52,72,107,52,103,121,109,110,98,105,57,104,108,98,119,53,69,57,52,120,108,119,54,115,92,10,9,76,83,47,50,68,113,122,75,116,97,78,107,122,110,114,89,89,74,67,90,54,47,74,67,82,108,67,113,69,85,122,97,75,121,108,90,71,71,50,112,83,81,70,109,103,111,73,81,86,49,67,67,79,79,85,83,50,78,70,81,76,80,115,111,65,81,77,119,66,101,73,80,92,10,9,100,113,85,89,120,82,108,105,52,106,117,66,78,117,73,119,103,70,107,56,83,88,116,114,116,72,57,77,68,65,98,103,65,49,97,80,101,100,48,49,81,103,119,103,51,121,55,81,78,52,88,43,103,112,108,99,111,71,103,118,112,118,43,113,84,117,87,67,103,82,53,104,92,10,9,75,99,66,89,122,86,66,48,67,107,86,90,104,104,47,107,68,53,68,85,53,120,49,99,98,120,66,80,76,104,79,106,83,108,120,74,53,78,49,83,73,108,78,67,83,66,98,117,110,67,118,81,82,108,121,114,116,75,52,86,118,68,43,90,70,82,84,49,68,82,68,90,92,10,9,82,49,100,82,102,121,67,82,80,104,84,72,66,112,65,68,68,68,89,50,56,112,116,66,51,120,51,70,111,117,65,107,51,116,49,121,82,112,75,109,76,79,75,115,119,122,116,87,76,57,89,88,84,83,77,88,54,49,108,72,106,101,118,112,66,121,99,87,48,53,114,119,92,10,9,54,70,77,110,77,115,116,82,109,65,77,113,57,112,111,87,55,99,82,107,51,105,72,71,100,98,104,71,57,54,83,108,72,57,116,81,88,75,65,100,54,86,88,43,87,90,108,103,69,57,113,86,89,118,114,86,90,119,105,89,49,56,77,111,104,104,55,120,70,79,56,71,92,10,9,118,55,81,103,103,117,88,69,69,86,102,52,78,121,86,122,56,53,86,103,65,78,103,78,89,86,71,47,83,101,68,57,53,119,109,48,69,102,100,107,65,55,89,98,51,90,79,81,119,120,48,121,65,70,50,68,87,104,76,57,71,83,55,74,68,53,80,101,121,89,105,118,92,10,9,79,67,72,72,65,79,76,115,113,106,103,79,110,70,107,70,110,98,52,108,52,102,86,83,56,81,115,110,67,55,87,122,119,67,69,68,103,65,49,73,78,57,104,109,80,51,74,84,57,81,100,111,53,117,103,53,75,76,89,120,118,89,111,89,119,71,53,120,98,65,65,53,92,10,9,119,70,74,53,78,77,107,100,88,54,83,66,100,122,84,48,50,118,69,107,52,120,98,56,67,57,75,76,114,98,67,66,74,76,97,88,108,122,83,102,65,98,70,56,67,48,102,57,99,90,86,110,65,108,109,67,106,113,102,113,83,90,51,101,73,81,107,68,65,70,52,106,92,10,9,117,82,120,49,114,111,72,100,55,65,122,78,119,82,115,105,57,65,52,83,85,111,67,69,67,114,66,54,103,109,117,72,71,48,120,52,86,119,120,118,117,99,78,53,66,73,108,119,85,74,85,119,103,66,86,74,71,81,65,65,80,47,52,76,77,116,104,90,90,65,70,54,92,10,9,86,47,80,101,111,85,76,118,65,71,78,113,98,99,97,43,71,52,120,114,72,50,110,101,50,57,110,82,79,51,53,72,101,110,52,87,67,119,51,109,48,103,78,86,119,65,67,43,49,109,69,65,65,74,73,43,68,105,81,53,53,120,98,98,103,79,72,118,85,111,80,55,92,10,9,52,88,72,52,108,99,66,55,78,67,100,122,103,50,73,97,83,86,86,48,118,85,70,100,105,99,112,81,83,56,100,111,51,68,102,71,111,70,43,73,97,72,48,115,52,119,122,103,90,86,48,71,65,78,120,74,48,97,110,65,80,77,56,55,43,81,55,118,118,68,56,97,92,10,9,116,118,79,111,48,80,117,89,120,105,57,56,73,102,65,79,83,84,80,48,118,75,51,53,110,72,69,79,120,47,108,43,106,88,116,77,100,110,72,77,74,53,83,80,118,122,109,106,105,120,47,118,102,53,115,74,65,56,106,112,81,57,117,84,84,65,53,56,71,56,68,69,92,10,9,82,97,97,97,43,81,74,116,80,83,76,48,84,113,89,87,97,48,82,70,109,104,90,66,84,87,114,90,102,49,49,106,78,52,102,118,119,51,77,79,120,120,111,49,47,48,89,108,117,66,55,72,101,103,56,90,80,104,77,83,56,79,56,112,77,103,119,117,122,104,103,68,92,10,9,103,69,102,117,66,70,77,71,107,78,117,82,52,68,67,67,73,73,102,108,72,110,85,81,79,48,74,102,111,99,87,102,54,43,101,55,81,117,50,89,84,114,114,66,104,117,76,121,106,82,111,54,100,108,75,72,73,76,122,106,88,73,102,106,106,88,52,104,98,109,78,52,92,10,9,106,71,116,120,69,110,83,81,111,65,114,55,73,69,88,101,100,97,57,107,90,80,70,106,98,102,119,56,89,53,100,81,107,99,71,101,105,111,98,87,112,52,116,90,105,118,112,98,75,113,74,52,106,79,71,55,76,86,72,85,84,117,65,57,97,103,121,43,56,49,109,97,92,10,9,122,57,119,55,119,84,79,101,85,100,81,52,112,85,75,88,43,68,89,72,75,51,113,104,111,76,106,115,102,69,88,80,75,122,114,73,56,118,80,82,47,104,118,49,102,117,73,122,82,85,102,71,76,81,53,113,103,112,49,89,73,110,68,112,68,76,75,67,55,82,74,73,92,10,9,118,50,83,114,113,103,115,107,112,84,114,83,80,57,53,67,121,97,120,114,66,70,85,74,71,71,80,80,106,80,109,78,52,88,97,78,71,80,116,55,68,90,53,53,70,107,87,117,121,73,51,76,106,65,70,83,89,70,49,75,47,118,105,76,53,69,53,100,53,106,116,43,92,10,9,76,108,122,88,85,89,55,116,69,73,114,121,76,57,103,69,74,71,56,52,54,85,51,110,102,105,55,73,73,54,119,70,53,76,90,52,106,56,113,107,119,112,100,109,65,68,107,103,110,120,114,79,51,111,43,105,121,115,107,88,100,76,71,81,82,84,67,52,50,88,55,108,92,10,9,89,71,68,98,115,48,54,56,99,99,76,55,85,77,98,113,98,74,75,51,106,71,47,68,101,105,103,67,99,73,114,86,115,56,99,69,101,74,105,47,122,120,83,66,53,54,70,65,53,51,109,115,88,113,51,68,67,120,51,71,80,103,84,77,51,69,100,109,49,90,67,114,92,10,9,70,102,104,79,105,71,66,70,79,114,98,117,84,71,118,120,109,109,104,84,103,108,110,80,52,56,85,56,106,119,110,106,79,67,79,80,99,103,116,43,112,111,83,113,90,89,115,66,53,73,65,100,111,119,57,80,85,107,121,99,110,111,98,116,76,87,70,100,55,49,109,75,92,10,9,52,104,82,43,99,68,121,103,98,88,108,66,72,82,116,106,57,56,87,105,103,57,47,69,69,53,98,102,67,88,55,116,56,76,120,98,110,121,99,86,65,111,56,109,56,117,75,115,68,50,115,119,73,69,48,71,85,65,104,119,80,113,81,66,50,52,82,51,85,108,65,117,92,10,9,115,48,108,116,103,85,103,53,106,85,85,89,55,79,52,102,115,105,103,106,70,97,90,114,67,114,119,47,103,107,80,50,89,97,55,101,109,103,49,76,101,70,102,69,48,115,79,66,54,71,110,80,106,75,77,66,65,97,107,122,103,73,67,65,65,73,47,81,75,72,121,67,92,10,9,103,73,68,65,65,65,73,67,65,103,73,68,67,65,103,73,97,69,106,52,102,119,69,71,65,77,109,104,112,57,106,121,89,54,69,48,65,65,65,65,65,69,108,70,84,107,83,117,81,109,67,67,92,10,9,34,10,9,108,111,99,97,108,32,116,111,97,115,116,95,98,97,99,107,95,112,110,103,32,61,10,9,34,105,86,66,79,82,119,48,75,71,103,111,65,65,65,65,78,83,85,104,69,85,103,65,65,65,73,65,65,65,65,67,65,67,65,89,65,65,65,68,68,80,109,72,76,65,65,65,71,87,69,108,69,81,86,82,52,110,79,50,100,87,52,104,86,86,82,106,72,102,54,79,105,92,10,9,73,53,109,109,82,111,112,100,86,69,122,74,106,76,84,98,81,48,87,81,74,85,71,81,68,121,73,85,81,103,88,50,108,105,57,70,69,65,81,57,57,116,104,76,84,120,87,57,70,85,83,57,83,66,68,48,49,107,78,81,67,66,107,57,86,71,113,87,111,43,77,108,92,10,9,110,86,72,72,77,87,56,122,53,101,51,48,115,77,55,107,100,106,120,122,57,110,72,87,57,54,50,122,57,54,122,47,68,119,55,106,98,88,57,55,121,102,113,100,98,54,49,118,114,88,51,112,97,84,81,97,105,72,121,90,49,117,48,71,105,79,52,105,65,84,74,72,92,10,9,65,109,83,79,66,77,103,99,67,90,65,53,69,105,66,122,74,69,68,109,83,73,68,77,107,81,67,90,73,119,69,121,82,119,74,107,106,103,84,73,72,65,109,81,79,84,79,54,51,81,65,114,116,109,57,98,51,101,48,109,116,71,88,122,104,55,57,51,117,119,107,116,92,10,9,54,97,110,84,100,110,65,72,110,100,119,76,76,65,69,87,78,51,56,117,65,117,89,68,56,52,68,98,87,110,120,54,103,86,115,76,120,57,57,67,54,121,47,70,90,101,66,67,56,57,99,88,103,90,72,109,110,53,49,113,102,111,97,66,107,52,88,102,110,119,73,71,92,10,9,103,67,80,65,85,101,68,83,82,65,51,117,116,104,105,86,70,75,67,107,111,50,56,72,55,103,78,87,65,83,117,66,53,99,67,121,53,109,101,82,99,57,77,109,81,52,77,103,81,49,47,122,115,119,47,89,65,47,120,77,107,79,77,71,85,107,112,82,67,81,69,109,92,10,9,54,80,66,112,119,66,114,103,69,87,65,116,115,65,53,52,69,70,105,81,114,109,88,117,68,66,70,69,43,65,72,52,68,116,104,74,121,68,68,47,52,121,49,68,49,119,82,111,48,101,107,122,103,83,101,65,112,52,65,110,103,99,101,66,117,89,109,98,49,87,49,71,92,10,9,103,82,51,65,116,56,68,110,104,67,69,69,56,66,77,104,113,81,65,116,79,110,48,112,115,66,70,52,72,108,103,80,122,69,55,87,109,79,112,122,67,102,103,73,101,66,99,52,68,122,52,83,117,65,118,81,111,116,80,118,65,86,52,71,88,105,83,107,100,116,71,101,92,10,9,80,109,65,84,115,65,118,115,74,88,65,84,89,70,122,72,57,119,75,98,103,97,51,65,77,48,67,80,121,48,109,110,76,109,101,66,53,52,65,102,119,86,89,67,56,51,87,65,99,82,48,47,66,51,103,100,101,73,116,81,109,111,110,74,77,82,102,52,104,112,65,120,92,10,9,87,49,89,79,107,56,85,48,65,120,81,54,102,120,98,119,78,118,65,109,115,78,68,115,66,79,74,76,89,65,118,89,90,81,71,122,112,101,66,67,53,121,56,69,118,103,102,101,81,53,49,118,122,85,118,65,47,90,89,66,114,102,99,67,90,103,66,102,65,89,56,90,92,10,9,120,120,87,66,72,107,74,109,78,86,118,54,78,104,71,103,48,74,103,51,67,72,87,56,56,71,77,76,89,84,88,85,66,77,115,77,77,66,49,52,120,122,67,101,97,77,48,115,81,104,108,116,103,113,85,65,97,122,65,48,85,55,84,108,70,98,65,90,66,105,119,70,92,10,9,117,78,115,119,108,109,106,80,119,56,65,75,105,48,67,87,65,105,119,122,106,67,88,75,101,99,69,105,105,75,85,65,74,107,97,75,106,116,107,65,56,99,79,65,112,81,68,76,68,87,79,74,99,112,55,71,89,67,88,88,85,111,65,113,88,111,119,120,108,90,107,68,92,10,9,80,66,111,98,120,70,73,65,86,81,68,112,101,83,103,50,103,75,85,65,56,119,120,106,105,99,53,89,70,120,118,65,85,111,68,99,114,116,54,112,65,116,72,88,85,49,103,75,48,71,115,89,83,51,84,71,83,111,105,114,66,72,82,106,83,76,49,90,81,76,105,56,92,10,9,102,100,74,73,103,80,111,84,116,102,52,105,65,101,114,80,88,84,69,72,83,52,68,54,115,121,84,109,89,65,108,81,102,43,54,77,79,86,103,67,49,74,43,111,121,43,52,107,81,80,50,74,87,111,67,84,65,80,86,72,90,87,68,109,75,65,78,107,84,109,85,121,92,10,9,119,73,88,121,102,121,73,99,113,69,119,71,117,71,119,89,83,51,84,79,110,74,105,68,78,81,82,107,106,113,85,65,102,120,118,71,69,112,48,122,75,43,90,103,83,119,72,43,78,89,119,108,79,105,100,113,71,57,53,83,103,70,72,68,87,67,73,82,108,103,75,99,92,10,9,78,89,119,108,69,109,69,112,119,74,66,104,76,74,69,73,83,119,71,71,68,87,79,74,82,69,105,65,122,76,69,85,52,74,82,104,76,74,69,73,90,89,68,77,107,81,68,49,90,121,84,109,89,70,85,66,57,87,102,67,74,53,70,51,103,113,85,65,74,119,120,106,92,10,9,105,99,52,53,69,51,79,119,112,81,68,72,68,87,79,74,122,111,110,97,104,98,87,101,65,48,83,108,73,122,69,112,122,115,85,99,98,67,74,65,52,97,109,86,74,121,51,105,105,90,115,105,97,104,80,79,43,110,113,65,118,52,122,106,105,88,75,105,57,109,67,115,92,10,9,66,84,104,79,101,69,87,75,83,69,102,51,104,52,65,67,65,43,104,82,56,75,109,112,86,65,89,89,77,73,52,110,121,113,108,77,71,81,105,70,100,57,121,73,90,70,81,113,65,120,119,50,106,105,102,75,105,86,113,67,108,119,68,49,74,50,111,74,88,103,76,85,92,10,9,110,48,112,108,103,70,72,103,116,72,70,77,48,90,53,113,67,70,66,89,68,84,120,111,70,86,78,48,82,68,85,69,75,72,65,73,76,81,97,108,112,72,73,67,57,75,80,70,111,74,82,69,88,89,114,110,73,99,65,66,104,53,105,105,78,97,101,66,113,122,69,66,92,10,9,80,65,84,89,55,120,66,84,116,71,89,73,52,116,52,104,54,67,70,65,110,48,78,77,48,90,114,111,121,47,67,56,74,111,70,82,97,85,108,48,84,80,82,86,87,66,52,67,88,65,83,79,79,99,81,86,78,120,74,57,72,97,97,112,65,73,87,120,54,69,57,85,92,10,9,67,113,97,103,107,107,77,65,104,72,109,65,83,107,70,47,111,114,102,102,118,81,84,89,54,120,82,88,88,69,47,48,78,90,104,101,65,117,120,122,105,105,117,117,112,55,73,67,50,76,122,99,88,112,82,82,121,83,111,65,119,111,97,81,72,104,118,110,84,50,85,122,92,10,9,119,66,87,48,73,79,82,78,103,121,112,87,65,89,86,83,99,67,56,113,66,84,48,90,119,109,68,66,122,102,78,66,107,88,116,82,75,101,104,74,57,68,52,65,43,65,113,119,121,122,71,50,77,76,111,98,50,49,79,65,51,89,54,120,104,100,72,100,50,78,53,68,92,10,9,103,68,97,70,47,68,66,53,73,73,101,110,65,80,56,81,100,103,97,70,68,57,88,78,65,73,87,74,121,97,43,111,69,118,68,67,53,70,90,56,55,56,102,70,55,48,97,86,103,66,102,86,122,81,65,70,57,106,106,72,122,120,109,84,71,51,71,57,66,100,67,101,92,10,9,103,66,43,49,121,65,68,97,70,102,83,106,70,104,110,103,72,72,113,90,108,65,100,110,77,88,111,47,103,53,115,65,104,85,112,65,122,119,50,121,53,121,68,69,76,119,78,68,109,112,100,71,72,85,83,108,111,68,87,47,87,65,86,75,73,99,65,82,86,65,112,97,92,10,9,115,56,77,113,85,65,111,66,116,66,112,111,121,120,88,103,97,54,116,103,75,81,84,52,76,99,69,53,99,117,73,68,68,79,43,55,109,71,69,86,113,65,48,47,74,84,106,72,86,79,99,75,115,66,80,52,66,80,103,85,98,67,97,65,107,69,97,65,89,56,66,110,92,10,9,119,75,115,74,122,108,85,110,84,104,65,87,99,48,52,83,55,118,69,102,74,106,122,121,98,89,84,119,69,115,52,82,119,111,55,102,73,101,65,80,52,76,120,72,73,51,111,97,68,100,56,74,43,118,90,116,113,119,70,109,65,104,56,68,87,119,107,86,119,86,83,102,92,10,9,70,70,52,109,84,72,52,80,69,75,113,103,115,90,47,57,119,70,70,103,107,72,65,76,51,85,49,104,57,97,48,118,107,105,73,68,81,80,106,80,118,103,90,56,65,98,119,80,114,69,49,48,88,107,47,79,69,86,89,54,43,119,113,102,102,107,74,72,72,121,71,107,92,10,9,55,98,90,52,100,79,106,78,52,112,52,66,120,109,104,109,65,103,106,102,47,103,51,65,78,109,65,106,77,68,49,74,65,121,98,72,73,79,72,98,101,52,66,114,110,98,121,102,48,80,71,108,84,43,97,111,81,103,101,88,107,85,119,65,117,69,54,67,77,82,89,65,92,10,9,109,52,68,49,119,76,80,65,52,110,70,47,55,122,108,99,88,67,88,77,84,47,111,74,52,43,119,104,119,114,100,51,55,72,79,89,107,106,82,100,104,119,52,117,73,54,107,65,82,86,114,73,65,76,65,67,101,65,67,52,70,49,103,70,76,65,85,87,65,110,99,65,92,10,9,56,52,72,90,84,68,120,115,110,83,101,77,118,87,101,52,78,113,107,97,74,107,121,121,66,112,111,47,66,119,109,84,114,48,72,67,69,110,88,112,67,121,54,109,81,105,101,51,111,50,115,67,106,71,99,67,73,100,121,90,54,104,49,99,82,109,85,69,69,78,48,104,92,10,9,120,85,113,103,113,68,65,83,73,72,77,107,81,79,90,73,103,77,121,82,65,74,107,106,65,84,74,72,65,109,83,79,66,77,103,99,67,90,65,53,69,105,66,122,74,69,68,109,83,73,68,77,43,81,56,65,111,109,114,114,115,67,107,52,72,65,65,65,65,65,66,74,92,10,9,82,85,53,69,114,107,74,103,103,103,61,61,92,10,9,34,10,9,108,111,99,97,108,32,116,111,97,115,116,95,101,121,101,115,95,99,108,111,115,101,100,95,112,110,103,32,61,10,9,34,105,86,66,79,82,119,48,75,71,103,111,65,65,65,65,78,83,85,104,69,85,103,65,65,65,73,65,65,65,65,67,65,67,65,89,65,65,65,68,68,80,109,72,76,65,65,65,66,51,85,108,69,81,86,82,52,110,79,51,99,77,85,52,86,85,82,104,65,52,88,47,65,92,10,9,50,111,73,70,121,73,76,65,49,115,90,108,115,65,66,119,65,87,121,69,86,105,104,89,106,112,97,87,116,105,89,56,109,113,102,69,69,67,110,110,109,112,122,118,83,121,97,90,53,68,98,47,122,68,50,90,121,84,83,122,72,81,54,72,111,101,116,107,57,81,67,115,92,10,9,74,89,65,52,65,99,81,74,73,69,52,65,99,81,75,73,69,48,67,99,65,79,73,69,69,67,101,65,79,65,72,69,67,83,66,79,65,72,69,67,105,66,78,65,110,65,68,105,66,66,65,110,103,68,103,66,120,65,107,103,84,103,66,120,65,111,103,84,81,74,119,65,92,10,9,52,103,81,81,74,52,65,52,65,99,81,74,73,69,52,65,99,81,75,73,69,48,67,99,65,79,73,69,69,67,101,65,79,65,72,69,67,83,66,79,65,72,69,67,105,72,117,51,101,111,67,83,109,56,118,122,116,57,102,118,118,43,48,48,121,89,118,100,65,57,105,50,92,10,9,55,99,47,53,57,99,87,72,118,57,97,43,80,72,122,102,101,53,120,100,72,97,47,51,52,56,120,99,122,99,122,80,109,88,107,47,77,55,57,109,53,115,102,77,102,70,52,120,48,54,52,66,51,70,121,101,118,57,114,48,109,84,109,100,109,99,80,77,80,79,48,53,92,10,9,121,48,74,102,106,56,100,47,89,99,85,114,52,71,53,109,122,111,55,110,118,122,102,57,100,109,89,101,70,56,121,121,113,55,101,101,99,75,118,43,50,76,111,105,103,69,47,47,87,118,68,98,50,118,49,116,98,110,113,98,122,56,65,52,65,99,81,74,73,69,52,65,92,10,9,99,81,75,73,69,48,67,99,65,79,73,69,69,67,101,65,79,65,72,69,67,83,66,79,65,72,69,67,105,66,78,65,110,65,68,105,66,66,65,110,103,68,103,66,120,65,107,103,84,103,66,120,65,111,103,84,81,74,119,65,52,103,81,81,74,52,65,52,65,99,81,74,92,10,9,73,69,52,65,99,81,75,73,69,48,67,99,65,79,73,69,69,67,101,65,79,65,72,69,67,83,66,79,65,72,69,67,105,66,78,65,110,65,68,105,66,66,65,110,103,68,103,66,120,65,107,103,84,103,66,120,65,111,103,84,81,74,119,65,52,103,81,81,74,52,65,52,92,10,9,65,99,81,74,73,69,52,65,99,81,75,73,69,48,67,99,65,79,73,69,69,67,101,65,79,65,72,69,67,83,66,79,65,72,69,67,105,66,78,65,110,65,68,105,66,66,65,110,103,68,103,66,120,65,107,103,84,103,66,120,65,111,103,84,81,74,119,65,52,103,81,81,92,10,9,74,52,65,52,65,99,81,74,73,69,52,65,99,81,75,73,69,48,67,99,65,79,73,69,69,67,101,65,79,65,72,69,67,83,68,117,71,100,115,50,73,111,47,105,119,43,68,88,65,65,65,65,65,69,108,70,84,107,83,117,81,109,67,67,92,10,9,34,10,9,108,111,99,97,108,32,116,111,97,115,116,95,101,121,101,115,95,111,112,101,110,95,112,110,103,32,61,10,9,34,105,86,66,79,82,119,48,75,71,103,111,65,65,65,65,78,83,85,104,69,85,103,65,65,65,73,65,65,65,65,67,65,67,65,89,65,65,65,68,68,80,109,72,76,65,65,65,67,72,85,108,69,81,86,82,52,110,79,51,99,77,85,53,85,89,82,104,71,52,84,80,71,92,10,9,68,101,103,71,86,67,113,51,89,70,103,65,67,83,72,85,57,110,89,117,65,107,74,115,54,65,121,70,66,83,48,70,72,84,82,81,117,66,67,116,106,73,109,98,73,72,81,77,66,83,82,79,67,77,90,117,80,115,73,53,84,122,85,51,85,56,120,98,110,80,118,80,92,10,9,122,82,83,122,87,67,54,88,120,79,118,70,57,73,68,77,75,103,67,53,65,112,65,114,65,76,107,67,107,67,115,65,117,81,75,81,75,119,67,53,65,112,65,114,65,76,107,67,107,67,115,65,117,81,75,81,75,119,67,53,65,112,65,114,65,76,107,67,107,67,115,65,92,10,9,117,81,75,81,75,119,67,53,65,112,65,114,65,76,107,67,107,67,115,65,117,81,75,81,75,119,67,53,65,112,65,114,65,76,107,67,107,67,115,65,117,81,75,81,75,119,67,53,65,112,65,114,65,76,107,67,107,67,115,65,117,90,102,84,65,48,122,50,100,57,55,43,92,10,9,56,55,50,68,121,122,57,114,50,55,71,113,69,50,68,57,88,103,72,102,103,82,118,103,78,47,66,104,99,115,120,97,84,52,67,110,101,65,99,77,50,65,79,50,55,108,57,118,65,75,102,65,117,54,107,120,69,121,102,65,97,43,65,114,99,65,70,56,66,104,89,68,92,10,9,71,121,90,116,80,76,104,43,77,55,76,105,51,115,81,122,119,67,108,47,55,52,65,100,55,103,76,52,78,114,66,106,121,104,109,119,117,51,74,57,80,106,85,69,90,103,76,89,101,117,84,97,70,77,65,74,99,65,86,115,65,55,43,65,111,56,107,120,69,49,56,66,92,10,9,80,47,57,122,47,87,121,116,80,79,101,99,65,90,43,65,81,43,66,54,97,103,47,77,110,65,65,102,103,87,80,103,80,88,65,74,102,66,110,89,77,79,97,112,80,101,120,79,66,80,65,68,50,66,122,52,51,68,120,105,48,98,43,70,117,47,86,68,107,70,119,66,92,10,9,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,92,10,9,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,92,10,9,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,92,10,9,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,51,67,50,121,116,121,99,54,80,119,43,111,79,65,65,65,65,65,66,74,82,85,53,69,92,10,9,114,107,74,103,103,103,61,61,92,10,9,34,10,9,108,111,99,97,108,32,116,111,97,115,116,95,102,114,111,110,116,95,112,110,103,32,61,10,9,34,105,86,66,79,82,119,48,75,71,103,111,65,65,65,65,78,83,85,104,69,85,103,65,65,65,73,65,65,65,65,67,65,67,65,89,65,65,65,68,68,80,109,72,76,65,65,65,70,121,48,108,69,81,86,82,52,110,79,50,100,122,87,56,86,86,82,105,72,110,119,74,87,92,10,9,69,90,82,87,111,50,73,66,103,119,111,111,67,47,120,65,70,49,97,106,67,101,110,67,111,68,69,109,74,107,89,106,73,97,52,107,114,116,122,111,120,113,85,98,78,117,53,99,117,84,74,66,47,119,107,84,100,87,118,99,111,65,116,70,49,74,90,87,112,75,86,87,92,10,9,107,98,90,67,87,50,118,86,117,106,106,51,79,118,102,97,106,48,118,98,57,53,48,55,53,55,54,47,74,53,110,99,108,106,83,47,77,43,69,56,57,51,122,78,122,74,109,117,120,99,86,70,82,70,119,50,116,102,115,69,82,72,117,82,65,77,71,82,65,77,71,82,92,10,9,65,77,71,82,65,77,71,82,65,77,71,82,65,77,71,82,65,77,71,82,65,77,71,82,65,77,71,82,65,77,71,82,65,77,71,82,65,77,71,82,65,77,71,82,65,77,71,82,65,77,71,82,65,77,71,82,65,77,72,90,48,117,52,84,77,71,72,107,86,76,118,80,92,10,9,89,71,88,117,102,114,88,100,90,55,65,113,88,100,110,99,69,57,105,54,107,106,99,66,116,119,78,57,119,69,53,103,78,51,65,98,48,70,115,55,101,104,112,43,51,103,90,48,65,57,116,74,88,52,73,100,76,98,75,118,65,80,56,65,77,56,65,67,77,65,49,77,92,10,9,78,82,121,84,119,65,81,119,88,106,118,71,103,73,118,65,53,82,85,84,75,121,74,71,78,81,86,89,117,98,74,118,65,65,52,67,43,52,70,57,119,76,51,65,80,98,88,80,80,109,66,122,71,97,101,51,66,113,97,65,55,52,70,66,89,65,103,52,65,51,119,74,92,10,9,68,67,47,55,49,50,50,81,111,104,111,67,76,70,47,104,116,119,75,80,65,89,101,66,81,55,86,106,80,57,87,114,53,80,85,119,67,90,119,71,80,103,99,43,66,98,52,103,116,83,121,74,69,107,86,111,106,119,68,76,86,47,104,101,89,65,66,52,65,117,103,72,92,10,9,55,105,47,120,106,78,114,78,68,80,65,90,56,70,55,116,77,49,71,67,67,79,85,75,48,70,122,120,51,99,66,84,119,68,80,65,85,87,74,86,43,71,112,56,66,76,119,79,122,65,76,117,69,118,106,80,65,112,90,43,50,47,117,66,52,56,66,76,119,67,51,117,92,10,9,53,101,102,72,99,101,65,65,56,68,82,112,115,79,109,75,88,119,118,81,88,80,69,55,103,78,101,65,69,54,82,43,88,76,84,109,81,121,66,57,47,82,49,98,65,90,43,70,111,75,76,121,101,52,67,84,119,67,106,119,76,113,114,56,116,88,67,77,78,78,78,120,92,10,9,120,86,54,65,111,118,76,55,103,98,80,65,50,54,84,53,116,108,103,98,109,52,71,51,65,78,101,70,76,113,43,108,52,73,100,73,48,53,117,100,84,118,108,82,79,65,98,99,55,70,109,65,114,81,67,70,113,82,56,65,78,53,112,109,120,50,81,55,56,73,74,110,92,10,9,65,82,52,116,119,70,55,103,89,89,102,99,113,76,119,77,117,72,85,68,72,103,73,99,99,99,105,77,122,66,72,103,74,113,57,119,68,119,69,71,72,68,73,106,48,52,51,106,47,54,109,100,65,69,85,84,57,97,104,90,112,113,106,122,112,70,101,119,100,81,118,81,92,10,9,84,81,108,122,49,52,65,56,68,114,105,77,65,54,119,70,79,69,104,110,88,75,50,114,71,111,43,81,76,111,87,98,52,121,71,65,115,79,99,54,107,103,84,109,87,65,116,119,108,51,71,101,75,72,106,81,73,57,82,97,65,75,51,56,43,101,70,121,117,100,120,97,92,10,9,103,70,51,71,101,97,76,65,112,88,117,49,70,117,66,79,52,122,120,82,99,66,57,103,80,104,79,119,70,115,68,49,119,107,86,119,57,103,66,98,114,85,79,116,66,87,104,49,101,55,88,89,71,76,117,116,65,54,48,70,48,72,86,47,88,47,113,115,65,121,86,65,92,10,9,88,108,82,101,65,74,102,86,75,118,69,102,101,54,119,68,57,88,66,111,88,116,120,104,72,83,103,66,56,113,76,72,79,108,65,67,53,73,85,69,67,69,54,118,100,97,65,69,121,73,118,75,116,119,65,86,101,78,83,52,111,54,109,56,65,76,56,98,53,52,108,109,92,10,9,122,70,116,115,100,81,70,53,89,98,55,85,98,105,51,65,106,72,71,101,97,75,98,76,79,116,66,97,103,75,118,71,101,97,73,90,56,43,99,68,114,65,87,52,89,112,119,110,109,113,110,56,71,69,65,116,81,71,90,89,67,122,66,108,110,67,101,99,48,84,81,119,92,10,9,79,78,89,67,47,71,113,99,74,53,121,120,70,109,68,108,110,84,70,70,74,98,69,87,52,68,102,106,80,79,71,77,87,111,68,103,87,65,115,119,98,112,119,110,109,108,108,111,47,83,100,114,81,119,76,107,120,97,120,49,111,76,85,65,89,56,90,53,119,104,108,114,92,10,9,65,101,89,111,89,88,118,84,119,77,120,98,66,57,111,74,85,71,120,110,113,109,55,65,106,122,110,114,81,73,47,55,65,83,52,52,90,73,113,69,43,98,85,87,68,119,70,43,100,115,103,85,105,98,43,115,65,122,48,69,117,79,105,81,75,82,76,109,70,57,115,107,92,10,9,81,70,55,56,97,82,51,111,73,89,67,109,103,110,54,89,88,50,51,49,69,71,68,69,73,86,77,107,74,69,66,119,115,104,66,103,69,105,48,71,101,86,72,120,81,87,67,120,71,76,84,56,105,120,72,70,82,115,109,105,66,81,66,49,65,49,53,73,103,79,66,107,92,10,9,73,52,67,54,65,66,43,121,69,85,65,116,103,65,47,90,67,75,65,87,119,73,101,115,66,80,106,98,75,84,115,121,107,57,97,66,88,103,73,115,111,71,55,65,65,47,79,98,98,117,48,70,75,78,89,67,118,106,80,80,106,115,48,56,109,100,119,81,85,117,99,72,92,10,9,120,43,121,73,112,79,98,102,43,69,88,83,110,103,75,99,100,99,121,79,105,77,115,122,70,53,52,67,68,68,112,109,82,56,82,56,65,65,105,43,65,109,103,77,89,69,116,50,65,107,121,103,113,52,75,87,90,67,83,65,90,103,73,101,117,68,120,52,54,55,49,78,92,10,9,51,76,102,79,43,90,72,73,113,65,85,111,43,77,89,53,80,120,76,90,122,81,73,65,118,110,98,79,106,48,83,87,76,89,67,54,65,68,117,121,70,71,65,77,122,81,83,115,121,75,119,76,75,71,89,67,90,57,122,75,105,69,87,87,76,81,66,111,75,109,104,70,92,10,9,116,103,75,99,76,54,71,77,67,71,84,87,66,82,84,111,99,102,71,78,115,52,106,68,107,56,70,81,106,103,67,106,74,90,84,82,54,97,82,98,119,89,119,118,66,85,77,53,65,118,120,85,81,104,109,100,106,118,110,117,89,72,88,75,69,79,68,72,69,115,114,111,92,10,9,100,73,97,56,103,115,115,81,89,66,55,116,71,98,66,82,80,118,89,75,57,104,87,103,54,76,79,43,99,105,50,110,77,49,107,65,84,103,78,118,65,105,101,57,67,116,110,105,70,102,119,47,80,103,71,101,76,97,109,115,75,106,78,78,50,107,88,116,65,118,65,76,92,10,9,99,73,107,48,118,90,115,107,118,87,51,108,97,117,51,102,74,107,106,78,118,108,118,102,88,54,99,115,65,100,52,72,66,111,68,110,83,105,113,118,88,99,121,84,110,111,107,89,74,116,48,87,51,47,103,53,122,72,112,50,43,88,73,89,43,84,100,83,108,103,68,122,92,10,9,119,80,80,65,67,101,65,100,72,78,54,67,88,83,74,47,107,76,54,100,103,54,81,55,110,119,100,114,118,119,57,120,76,100,118,106,79,70,102,111,87,117,108,97,88,67,122,104,90,90,56,106,112,120,112,47,50,119,113,56,65,114,119,66,80,79,66,102,43,76,113,89,92,10,9,66,115,54,82,118,114,88,110,97,107,101,57,107,107,100,90,55,81,50,112,70,97,118,103,86,112,81,106,81,74,49,109,69,81,65,79,65,83,56,67,82,52,72,68,108,80,99,105,121,122,110,83,43,115,82,53,85,110,56,56,81,108,72,82,119,49,122,76,55,86,101,90,92,10,9,86,102,82,75,108,67,116,65,73,48,116,108,50,69,97,83,52,65,67,119,114,51,98,115,73,114,48,118,116,55,100,50,88,76,57,77,48,105,120,112,115,68,82,78,113,116,106,54,111,79,111,121,120,89,66,113,110,76,83,66,53,84,104,112,83,110,113,112,53,102,108,49,92,10,9,83,65,87,51,111,110,48,67,78,76,74,85,66,110,43,67,86,72,65,114,113,105,71,65,97,66,116,54,101,88,82,119,74,69,66,119,74,69,66,119,74,69,66,119,74,69,66,119,74,69,66,119,74,69,66,119,74,69,66,119,74,69,66,119,74,69,66,119,74,69,66,119,92,10,9,74,69,66,119,74,69,66,119,47,103,85,85,103,105,75,77,50,98,116,49,121,103,65,65,65,65,66,74,82,85,53,69,114,107,74,103,103,103,61,61,92,10,9,34,10,9,108,111,99,97,108,32,116,111,97,115,116,95,109,111,117,116,104,95,112,110,103,32,61,10,9,34,105,86,66,79,82,119,48,75,71,103,111,65,65,65,65,78,83,85,104,69,85,103,65,65,65,73,65,65,65,65,67,65,67,65,89,65,65,65,68,68,80,109,72,76,65,65,65,67,79,48,108,69,81,86,82,52,110,79,51,99,80,50,111,86,85,82,104,65,56,82,78,74,92,10,9,111,43,73,67,65,105,74,87,86,111,113,70,109,55,66,73,107,84,97,70,52,68,89,115,82,78,121,75,89,71,85,114,112,66,66,83,90,119,101,67,104,89,83,107,84,66,113,120,56,119,47,80,73,103,109,89,75,110,109,80,101,98,122,104,110,118,77,114,104,55,110,119,92,10,9,68,88,78,109,71,71,52,120,87,52,118,70,103,110,106,100,50,102,81,65,50,97,119,67,107,67,115,65,117,81,75,81,75,119,67,53,65,112,65,114,65,76,107,67,107,67,115,65,117,81,75,81,75,119,67,53,65,112,65,114,65,76,107,67,107,67,115,65,117,81,75,81,92,10,9,75,119,67,53,65,112,65,114,65,76,107,67,107,67,115,65,117,81,75,81,75,119,67,53,65,112,65,114,65,76,107,67,107,67,115,65,117,81,75,81,75,119,67,53,65,112,65,114,65,76,107,67,107,67,115,65,117,81,75,81,75,119,67,53,65,112,65,114,65,76,107,67,92,10,9,107,67,115,65,117,81,75,81,75,119,67,53,55,85,48,80,115,67,55,118,100,104,43,118,118,80,98,57,53,43,80,74,53,112,105,55,114,82,70,47,70,102,118,102,122,88,56,65,80,65,69,101,65,118,101,66,117,53,102,72,114,105,121,65,72,56,65,90,99,65,74,56,92,10,9,66,102,54,65,74,52,74,104,51,119,68,65,80,118,67,66,53,97,55,120,66,72,103,66,110,75,57,108,111,104,107,97,43,82,118,103,68,99,115,72,47,103,104,52,80,102,48,111,56,122,86,121,65,77,57,87,88,80,100,48,48,105,108,109,98,117,81,65,102,113,54,52,92,10,9,55,116,101,107,85,56,122,99,121,65,69,99,114,76,106,117,99,78,73,112,90,109,55,107,65,78,54,121,47,70,118,103,67,80,105,48,104,108,108,109,97,43,81,65,118,103,69,118,103,100,78,98,110,118,56,70,50,65,88,43,114,109,50,105,71,82,112,121,72,119,67,117,92,10,9,55,81,88,99,65,49,52,66,101,56,66,122,89,79,102,121,43,71,47,103,79,120,100,80,47,85,99,117,65,103,65,56,101,119,65,119,99,65,67,119,47,71,54,103,54,99,90,102,71,84,113,65,51,71,122,107,98,52,68,99,81,103,72,73,70,89,66,99,65,99,103,86,92,10,9,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,92,10,9,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,92,10,9,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,70,89,66,99,92,10,9,65,99,103,86,103,70,119,66,121,66,87,65,88,65,72,73,47,81,80,119,104,122,87,84,118,88,111,57,97,103,65,65,65,65,66,74,82,85,53,69,114,107,74,103,103,103,61,61,92,10,9,34,10,9,108,111,99,97,108,32,99,108,97,115,115,32,61,32,114,101,113,117,105,114,101,95,51,48,108,111,103,40,41,10,9,108,111,99,97,108,32,84,111,97,115,116,32,61,32,99,108,97,115,115,40,34,84,111,97,115,116,34,41,10,9,102,117,110,99,116,105,111,110,32,84,111,97,115,116,58,105,110,105,116,40,41,10,9,9,115,101,108,102,58,99,101,110,116,101,114,40,41,10,9,9,115,101,108,102,46,101,121,101,115,32,61,32,123,125,10,9,9,115,101,108,102,46,101,121,101,115,46,99,108,111,115,101,100,95,116,32,61,32,48,10,9,9,115,101,108,102,46,101,121,101,115,46,98,108,105,110,107,95,116,32,61,32,50,10,9,9,115,101,108,102,46,108,111,111,107,32,61,32,123,125,10,9,9,115,101,108,102,46,108,111,111,107,46,116,97,114,103,101,116,32,61,32,123,32,120,32,61,32,48,46,50,44,32,32,121,32,61,32,48,46,50,32,125,10,9,9,115,101,108,102,46,108,111,111,107,46,99,117,114,114,101,110,116,32,61,32,123,32,120,32,61,32,48,46,50,44,32,32,121,32,61,32,48,46,50,32,125,10,9,9,115,101,108,102,46,108,111,111,107,46,68,85,82,65,84,73,79,78,32,61,32,48,46,53,10,9,9,115,101,108,102,46,108,111,111,107,46,80,79,73,78,84,83,32,61,32,123,10,9,9,9,123,32,120,32,61,32,48,46,56,44,32,121,32,61,32,48,46,56,32,125,44,10,9,9,9,123,32,120,32,61,32,48,46,49,44,32,121,32,61,32,48,46,49,32,125,44,10,9,9,9,123,32,120,32,61,32,48,46,56,44,32,121,32,61,32,48,46,49,32,125,44,10,9,9,9,123,32,120,32,61,32,48,46,49,44,32,121,32,61,32,48,46,56,32,125,44,10,9,9,125,10,9,9,115,101,108,102,46,108,111,111,107,46,112,111,105,110,116,32,61,32,48,10,9,9,115,101,108,102,46,108,111,111,107,46,112,111,105,110,116,95,116,32,61,32,49,10,9,9,115,101,108,102,46,108,111,111,107,46,116,32,61,32,48,10,9,101,110,100,10,9,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,101,97,115,101,79,117,116,40,116,44,32,98,44,32,99,44,32,100,41,10,9,9,116,32,61,32,116,32,47,32,100,32,45,32,49,10,9,9,114,101,116,117,114,110,32,99,32,42,32,40,109,97,116,104,46,112,111,119,40,116,44,32,51,41,32,43,32,49,41,32,43,32,98,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,84,111,97,115,116,58,99,101,110,116,101,114,40,41,10,9,9,108,111,99,97,108,32,119,119,44,32,119,104,32,61,32,108,111,118,101,46,119,105,110,100,111,119,46,102,114,111,109,80,105,120,101,108,115,40,108,111,118,101,46,103,114,97,112,104,105,99,115,46,103,101,116,68,105,109,101,110,115,105,111,110,115,40,41,41,10,9,9,115,101,108,102,46,120,32,61,32,109,97,116,104,46,102,108,111,111,114,40,119,119,32,47,32,50,32,47,32,51,50,41,32,42,32,51,50,32,43,32,49,54,10,9,9,115,101,108,102,46,121,32,61,32,109,97,116,104,46,102,108,111,111,114,40,119,104,32,47,32,50,32,47,32,51,50,41,32,42,32,51,50,32,43,32,49,54,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,84,111,97,115,116,58,103,101,116,95,108,111,111,107,95,99,111,111,114,100,105,110,97,116,101,115,40,41,10,9,9,108,111,99,97,108,32,116,32,61,32,115,101,108,102,46,108,111,111,107,46,116,10,9,9,108,111,99,97,108,32,115,114,99,32,61,32,115,101,108,102,46,108,111,111,107,46,99,117,114,114,101,110,116,10,9,9,108,111,99,97,108,32,100,115,116,32,61,32,115,101,108,102,46,108,111,111,107,46,116,97,114,103,101,116,10,9,9,108,111,99,97,108,32,108,111,111,107,95,120,32,61,32,101,97,115,101,79,117,116,40,116,44,32,115,114,99,46,120,44,32,100,115,116,46,120,32,45,32,115,114,99,46,120,44,32,115,101,108,102,46,108,111,111,107,46,68,85,82,65,84,73,79,78,41,10,9,9,108,111,99,97,108,32,108,111,111,107,95,121,32,61,32,101,97,115,101,79,117,116,40,116,44,32,115,114,99,46,121,44,32,100,115,116,46,121,32,45,32,115,114,99,46,121,44,32,115,101,108,102,46,108,111,111,107,46,68,85,82,65,84,73,79,78,41,10,9,9,114,101,116,117,114,110,32,108,111,111,107,95,120,44,32,108,111,111,107,95,121,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,84,111,97,115,116,58,117,112,100,97,116,101,40,100,116,41,10,9,9,115,101,108,102,46,108,111,111,107,46,116,32,61,32,109,97,116,104,46,109,105,110,40,115,101,108,102,46,108,111,111,107,46,116,32,43,32,100,116,44,32,115,101,108,102,46,108,111,111,107,46,68,85,82,65,84,73,79,78,41,10,9,9,115,101,108,102,46,101,121,101,115,46,99,108,111,115,101,100,95,116,32,61,32,109,97,116,104,46,109,97,120,40,115,101,108,102,46,101,121,101,115,46,99,108,111,115,101,100,95,116,32,45,32,100,116,44,32,48,41,10,9,9,115,101,108,102,46,101,121,101,115,46,98,108,105,110,107,95,116,32,61,32,109,97,116,104,46,109,97,120,40,115,101,108,102,46,101,121,101,115,46,98,108,105,110,107,95,116,32,45,32,100,116,44,32,48,41,10,9,9,115,101,108,102,46,108,111,111,107,46,112,111,105,110,116,95,116,32,61,32,109,97,116,104,46,109,97,120,40,115,101,108,102,46,108,111,111,107,46,112,111,105,110,116,95,116,32,45,32,100,116,44,32,48,41,10,9,9,105,102,32,115,101,108,102,46,101,121,101,115,46,98,108,105,110,107,95,116,32,61,61,32,48,32,116,104,101,110,10,9,9,9,115,101,108,102,58,98,108,105,110,107,40,41,10,9,9,101,110,100,10,9,9,105,102,32,115,101,108,102,46,108,111,111,107,46,112,111,105,110,116,95,116,32,61,61,32,48,32,116,104,101,110,10,9,9,9,115,101,108,102,58,108,111,111,107,95,97,116,95,110,101,120,116,95,112,111,105,110,116,40,41,10,9,9,101,110,100,10,9,9,108,111,99,97,108,32,108,111,111,107,95,120,44,32,108,111,111,107,95,121,32,61,32,115,101,108,102,58,103,101,116,95,108,111,111,107,95,99,111,111,114,100,105,110,97,116,101,115,40,41,10,9,9,115,101,108,102,46,111,102,102,115,101,116,95,120,32,61,32,108,111,111,107,95,120,32,42,32,52,10,9,9,115,101,108,102,46,111,102,102,115,101,116,95,121,32,61,32,40,49,32,45,32,108,111,111,107,95,121,41,32,42,32,45,52,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,84,111,97,115,116,58,100,114,97,119,40,41,10,9,9,108,111,99,97,108,32,120,32,61,32,115,101,108,102,46,120,10,9,9,108,111,99,97,108,32,121,32,61,32,115,101,108,102,46,121,10,9,9,108,111,99,97,108,32,108,111,111,107,95,120,44,32,108,111,111,107,95,121,32,61,32,115,101,108,102,58,103,101,116,95,108,111,111,107,95,99,111,111,114,100,105,110,97,116,101,115,40,41,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,100,114,97,119,40,103,95,105,109,97,103,101,115,46,116,111,97,115,116,46,98,97,99,107,44,32,120,44,32,121,44,32,115,101,108,102,46,114,44,32,49,44,32,49,44,32,54,52,44,32,54,52,41,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,100,114,97,119,40,103,95,105,109,97,103,101,115,46,116,111,97,115,116,46,102,114,111,110,116,44,32,120,32,43,32,115,101,108,102,46,111,102,102,115,101,116,95,120,44,32,121,32,43,32,115,101,108,102,46,111,102,102,115,101,116,95,121,44,32,115,101,108,102,46,114,44,32,49,44,32,49,44,32,54,52,44,32,54,52,41,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,100,114,97,119,40,115,101,108,102,58,103,101,116,95,101,121,101,115,95,105,109,97,103,101,40,41,44,32,120,32,43,32,115,101,108,102,46,111,102,102,115,101,116,95,120,32,42,32,50,46,53,44,32,121,32,43,32,115,101,108,102,46,111,102,102,115,101,116,95,121,32,42,32,50,46,53,44,32,115,101,108,102,46,114,44,32,49,44,32,49,44,32,54,52,44,32,54,52,41,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,100,114,97,119,40,103,95,105,109,97,103,101,115,46,116,111,97,115,116,46,109,111,117,116,104,44,32,120,32,43,32,115,101,108,102,46,111,102,102,115,101,116,95,120,32,42,32,50,44,32,121,32,43,32,115,101,108,102,46,111,102,102,115,101,116,95,121,32,42,32,50,44,32,115,101,108,102,46,114,44,32,49,44,32,49,44,32,54,52,44,32,54,52,41,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,84,111,97,115,116,58,103,101,116,95,101,121,101,115,95,105,109,97,103,101,40,41,10,9,9,105,102,32,115,101,108,102,46,101,121,101,115,46,99,108,111,115,101,100,95,116,32,62,32,48,32,116,104,101,110,10,9,9,9,114,101,116,117,114,110,32,103,95,105,109,97,103,101,115,46,116,111,97,115,116,46,101,121,101,115,46,99,108,111,115,101,100,10,9,9,101,110,100,10,9,9,114,101,116,117,114,110,32,103,95,105,109,97,103,101,115,46,116,111,97,115,116,46,101,121,101,115,46,111,112,101,110,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,84,111,97,115,116,58,98,108,105,110,107,40,41,10,9,9,105,102,32,115,101,108,102,46,101,121,101,115,46,99,108,111,115,101,100,95,116,32,62,32,48,32,116,104,101,110,10,9,9,9,114,101,116,117,114,110,10,9,9,101,110,100,10,9,9,115,101,108,102,46,101,121,101,115,46,99,108,111,115,101,100,95,116,32,61,32,48,46,49,10,9,9,115,101,108,102,46,101,121,101,115,46,98,108,105,110,107,95,116,32,61,32,115,101,108,102,46,110,101,120,116,95,98,108,105,110,107,40,41,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,84,111,97,115,116,58,110,101,120,116,95,98,108,105,110,107,40,41,10,9,9,114,101,116,117,114,110,32,53,32,43,32,108,111,118,101,46,109,97,116,104,46,114,97,110,100,111,109,40,48,44,32,51,41,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,84,111,97,115,116,58,108,111,111,107,95,97,116,40,116,120,44,32,116,121,41,10,9,9,108,111,99,97,108,32,108,111,111,107,95,120,44,32,108,111,111,107,95,121,32,61,32,115,101,108,102,58,103,101,116,95,108,111,111,107,95,99,111,111,114,100,105,110,97,116,101,115,40,41,10,9,9,115,101,108,102,46,108,111,111,107,46,99,117,114,114,101,110,116,46,120,32,61,32,108,111,111,107,95,120,10,9,9,115,101,108,102,46,108,111,111,107,46,99,117,114,114,101,110,116,46,121,32,61,32,108,111,111,107,95,121,10,9,9,115,101,108,102,46,108,111,111,107,46,116,32,61,32,48,10,9,9,115,101,108,102,46,108,111,111,107,46,112,111,105,110,116,95,116,32,61,32,51,32,43,32,108,111,118,101,46,109,97,116,104,46,114,97,110,100,111,109,40,48,44,32,49,41,10,9,9,115,101,108,102,46,108,111,111,107,46,116,97,114,103,101,116,46,120,32,61,32,116,120,10,9,9,115,101,108,102,46,108,111,111,107,46,116,97,114,103,101,116,46,121,32,61,32,116,121,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,84,111,97,115,116,58,108,111,111,107,95,97,116,95,110,101,120,116,95,112,111,105,110,116,40,41,10,9,9,115,101,108,102,46,108,111,111,107,46,112,111,105,110,116,32,61,32,115,101,108,102,46,108,111,111,107,46,112,111,105,110,116,32,43,32,49,10,9,9,105,102,32,115,101,108,102,46,108,111,111,107,46,112,111,105,110,116,32,62,32,35,115,101,108,102,46,108,111,111,107,46,80,79,73,78,84,83,32,116,104,101,110,10,9,9,9,115,101,108,102,46,108,111,111,107,46,112,111,105,110,116,32,61,32,49,10,9,9,101,110,100,10,9,9,108,111,99,97,108,32,112,111,105,110,116,32,61,32,115,101,108,102,46,108,111,111,107,46,80,79,73,78,84,83,91,115,101,108,102,46,108,111,111,107,46,112,111,105,110,116,93,10,9,9,115,101,108,102,58,108,111,111,107,95,97,116,40,112,111,105,110,116,46,120,44,32,112,111,105,110,116,46,121,41,10,9,101,110,100,10,9,108,111,99,97,108,32,77,111,115,97,105,99,32,61,32,99,108,97,115,115,40,34,77,111,115,97,105,99,34,41,10,9,102,117,110,99,116,105,111,110,32,77,111,115,97,105,99,58,105,110,105,116,40,41,10,9,9,108,111,99,97,108,32,109,111,115,97,105,99,95,105,109,97,103,101,32,61,32,103,95,105,109,97,103,101,115,46,109,111,115,97,105,99,91,49,93,10,9,9,108,111,99,97,108,32,115,119,44,32,115,104,32,61,32,109,111,115,97,105,99,95,105,109,97,103,101,58,103,101,116,68,105,109,101,110,115,105,111,110,115,40,41,10,9,9,108,111,99,97,108,32,119,119,44,32,119,104,32,61,32,108,111,118,101,46,119,105,110,100,111,119,46,102,114,111,109,80,105,120,101,108,115,40,108,111,118,101,46,103,114,97,112,104,105,99,115,46,103,101,116,68,105,109,101,110,115,105,111,110,115,40,41,41,10,9,9,105,102,32,108,111,118,101,46,119,105,110,100,111,119,46,103,101,116,80,105,120,101,108,83,99,97,108,101,40,41,32,62,32,49,32,116,104,101,110,10,9,9,9,109,111,115,97,105,99,95,105,109,97,103,101,32,61,32,103,95,105,109,97,103,101,115,46,109,111,115,97,105,99,91,50,93,10,9,9,101,110,100,10,9,9,108,111,99,97,108,32,83,73,90,69,95,88,32,61,32,109,97,116,104,46,102,108,111,111,114,40,119,119,32,47,32,51,50,32,43,32,50,41,10,9,9,108,111,99,97,108,32,83,73,90,69,95,89,32,61,32,109,97,116,104,46,102,108,111,111,114,40,119,104,32,47,32,51,50,32,43,32,50,41,10,9,9,108,111,99,97,108,32,83,73,90,69,32,61,32,83,73,90,69,95,88,32,42,32,83,73,90,69,95,89,10,9,9,115,101,108,102,46,98,97,116,99,104,32,61,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,83,112,114,105,116,101,66,97,116,99,104,40,109,111,115,97,105,99,95,105,109,97,103,101,44,32,83,73,90,69,44,32,34,115,116,114,101,97,109,34,41,10,9,9,115,101,108,102,46,112,105,101,99,101,115,32,61,32,123,125,10,9,9,115,101,108,102,46,99,111,108,111,114,95,116,32,61,32,49,10,9,9,115,101,108,102,46,103,101,110,101,114,97,116,105,111,110,32,61,32,49,10,9,9,108,111,99,97,108,32,67,79,76,79,82,83,32,61,32,123,125,10,9,9,102,111,114,32,95,44,99,111,108,111,114,32,105,110,32,105,112,97,105,114,115,40,123,10,9,9,9,123,32,50,52,48,44,32,50,52,48,44,32,50,52,48,32,125,44,32,45,45,32,87,72,73,84,69,32,40,105,115,104,41,10,9,9,9,123,32,50,51,50,44,32,49,48,52,44,32,49,54,50,125,44,32,45,45,32,80,73,78,75,10,9,9,9,123,32,54,57,44,32,49,53,53,44,32,49,54,56,32,125,44,32,45,45,32,66,76,85,69,10,9,9,9,123,32,54,55,44,32,57,51,44,32,49,49,57,32,125,44,32,45,45,32,68,65,82,75,32,66,76,85,69,10,9,9,125,41,32,100,111,10,9,9,9,116,97,98,108,101,46,105,110,115,101,114,116,40,67,79,76,79,82,83,44,32,99,111,108,111,114,41,10,9,9,9,116,97,98,108,101,46,105,110,115,101,114,116,40,67,79,76,79,82,83,44,32,99,111,108,111,114,41,10,9,9,101,110,100,10,9,9,45,45,32,73,110,115,101,114,116,32,111,110,108,121,32,111,110,99,101,46,32,84,104,105,115,32,119,97,121,32,105,116,32,97,112,112,101,97,114,115,32,104,97,108,102,32,97,115,32,111,102,116,101,110,46,10,9,9,116,97,98,108,101,46,105,110,115,101,114,116,40,67,79,76,79,82,83,44,32,123,32,50,50,48,44,32,50,51,57,44,32,49,49,51,32,125,41,32,45,45,32,76,73,77,69,10,9,9,45,45,32,87,104,101,110,32,117,115,105,110,103,32,116,104,101,32,104,105,103,104,101,114,45,114,101,115,32,109,111,115,97,105,99,32,115,112,114,105,116,101,32,115,104,101,101,116,44,32,119,101,32,119,97,110,116,32,116,111,32,100,114,97,119,32,105,116,115,10,9,9,45,45,32,115,112,114,105,116,101,115,32,97,116,32,116,104,101,32,115,97,109,101,32,115,99,97,108,101,32,97,115,32,116,104,101,32,114,101,103,117,108,97,114,45,114,101,115,111,108,117,116,105,111,110,32,111,110,101,44,32,98,101,99,97,117,115,101,10,9,9,45,45,32,119,101,39,108,108,32,103,108,111,98,97,108,108,121,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,115,99,97,108,101,32,42,101,118,101,114,121,116,104,105,110,103,42,32,98,121,32,116,104,101,32,115,99,114,101,101,110,39,115,10,9,9,45,45,32,112,105,120,101,108,32,100,101,110,115,105,116,121,32,114,97,116,105,111,46,10,9,9,45,45,32,87,101,32,99,97,110,32,97,118,111,105,100,32,97,32,108,111,116,32,111,102,32,81,117,97,100,32,115,99,97,108,105,110,103,32,98,121,32,116,97,107,105,110,103,32,97,100,118,97,110,116,97,103,101,32,111,102,32,116,104,101,32,102,97,99,116,10,9,9,45,45,32,116,104,97,116,32,81,117,97,100,115,32,117,115,101,32,110,111,114,109,97,108,105,122,101,100,32,116,101,120,116,117,114,101,32,99,111,111,114,100,105,110,97,116,101,115,32,105,110,116,101,114,110,97,108,108,121,32,45,32,105,102,32,119,101,32,117,115,101,32,10,9,9,45,45,32,116,104,101,32,39,115,111,117,114,99,101,32,105,109,97,103,101,32,115,105,122,101,39,32,97,110,100,32,113,117,97,100,32,115,105,122,101,32,111,102,32,116,104,101,32,64,49,120,32,105,109,97,103,101,32,102,111,114,32,116,104,101,32,81,117,97,100,115,10,9,9,45,45,32,101,118,101,110,32,119,104,101,110,32,114,101,110,100,101,114,105,110,103,32,116,104,101,109,32,117,115,105,110,103,32,116,104,101,32,64,50,120,32,105,109,97,103,101,44,32,105,116,32,119,105,108,108,32,97,117,116,111,109,97,116,105,99,97,108,108,121,10,9,9,45,45,32,115,99,97,108,101,32,97,115,32,101,120,112,101,99,116,101,100,46,10,9,9,108,111,99,97,108,32,81,85,65,68,83,32,61,32,123,10,9,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,81,117,97,100,40,48,44,32,32,48,44,32,32,51,50,44,32,51,50,44,32,115,119,44,32,115,104,41,44,10,9,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,81,117,97,100,40,48,44,32,32,51,50,44,32,51,50,44,32,51,50,44,32,115,119,44,32,115,104,41,44,10,9,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,81,117,97,100,40,51,50,44,32,51,50,44,32,51,50,44,32,51,50,44,32,115,119,44,32,115,104,41,44,10,9,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,81,117,97,100,40,51,50,44,32,48,44,32,32,51,50,44,32,51,50,44,32,115,119,44,32,115,104,41,44,10,9,9,125,10,9,9,108,111,99,97,108,32,101,120,99,108,117,100,101,95,108,101,102,116,32,61,32,109,97,116,104,46,102,108,111,111,114,40,119,119,32,47,32,50,32,47,32,51,50,41,10,9,9,108,111,99,97,108,32,101,120,99,108,117,100,101,95,114,105,103,104,116,32,61,32,101,120,99,108,117,100,101,95,108,101,102,116,32,43,32,51,10,9,9,108,111,99,97,108,32,101,120,99,108,117,100,101,95,116,111,112,32,61,32,109,97,116,104,46,102,108,111,111,114,40,119,104,32,47,32,50,32,47,32,51,50,41,10,9,9,108,111,99,97,108,32,101,120,99,108,117,100,101,95,98,111,116,116,111,109,32,61,32,101,120,99,108,117,100,101,95,116,111,112,32,43,32,51,10,9,9,108,111,99,97,108,32,101,120,99,108,117,100,101,95,119,105,100,116,104,32,61,32,101,120,99,108,117,100,101,95,114,105,103,104,116,32,45,32,101,120,99,108,117,100,101,95,108,101,102,116,32,43,32,49,10,9,9,108,111,99,97,108,32,101,120,99,108,117,100,101,95,104,101,105,103,104,116,32,61,32,101,120,99,108,117,100,101,95,98,111,116,116,111,109,32,45,32,101,120,99,108,117,100,101,95,116,111,112,32,43,32,49,10,9,9,108,111,99,97,108,32,101,120,99,108,117,100,101,95,97,114,101,97,32,61,32,101,120,99,108,117,100,101,95,119,105,100,116,104,32,42,32,101,120,99,108,117,100,101,95,104,101,105,103,104,116,10,9,9,108,111,99,97,108,32,101,120,99,108,117,100,101,95,99,101,110,116,101,114,95,120,32,61,32,101,120,99,108,117,100,101,95,108,101,102,116,32,43,32,49,46,53,10,9,9,108,111,99,97,108,32,101,120,99,108,117,100,101,95,99,101,110,116,101,114,95,121,32,61,32,101,120,99,108,117,100,101,95,116,111,112,32,43,32,49,46,53,10,9,9,115,101,108,102,46,103,101,110,101,114,97,116,111,114,115,32,61,32,123,10,9,9,9,102,117,110,99,116,105,111,110,40,112,105,101,99,101,44,32,103,101,110,101,114,97,116,105,111,110,41,10,9,9,9,9,114,101,116,117,114,110,32,67,79,76,79,82,83,91,108,111,118,101,46,109,97,116,104,46,114,97,110,100,111,109,40,49,44,32,35,67,79,76,79,82,83,41,93,10,9,9,9,101,110,100,44,10,9,9,9,102,117,110,99,116,105,111,110,40,112,105,101,99,101,44,32,103,101,110,101,114,97,116,105,111,110,41,10,9,9,9,9,114,101,116,117,114,110,32,67,79,76,79,82,83,91,49,32,43,32,40,103,101,110,101,114,97,116,105,111,110,32,43,32,112,105,101,99,101,46,103,114,105,100,95,120,32,45,32,112,105,101,99,101,46,103,114,105,100,95,121,41,32,37,32,35,67,79,76,79,82,83,93,10,9,9,9,101,110,100,44,10,9,9,9,102,117,110,99,116,105,111,110,40,112,105,101,99,101,44,32,103,101,110,101,114,97,116,105,111,110,41,10,9,9,9,9,114,101,116,117,114,110,32,67,79,76,79,82,83,91,49,32,43,32,40,112,105,101,99,101,46,103,114,105,100,95,120,32,43,32,103,101,110,101,114,97,116,105,111,110,41,32,37,32,35,67,79,76,79,82,83,93,10,9,9,9,101,110,100,44,10,9,9,9,102,117,110,99,116,105,111,110,40,112,105,101,99,101,44,32,103,101,110,101,114,97,116,105,111,110,41,10,9,9,9,9,108,111,99,97,108,32,108,101,110,32,61,32,103,101,110,101,114,97,116,105,111,110,32,43,32,109,97,116,104,46,115,113,114,116,40,112,105,101,99,101,46,103,114,105,100,95,120,32,94,32,50,32,43,32,112,105,101,99,101,46,103,114,105,100,95,121,32,94,32,50,41,10,9,9,9,9,114,101,116,117,114,110,32,67,79,76,79,82,83,91,49,32,43,32,109,97,116,104,46,102,108,111,111,114,40,108,101,110,41,32,37,32,35,67,79,76,79,82,83,93,10,9,9,9,101,110,100,44,10,9,9,9,102,117,110,99,116,105,111,110,40,112,105,101,99,101,44,32,103,101,110,101,114,97,116,105,111,110,41,10,9,9,9,9,108,111,99,97,108,32,100,120,32,61,32,112,105,101,99,101,46,103,114,105,100,95,120,32,45,32,101,120,99,108,117,100,101,95,99,101,110,116,101,114,95,120,10,9,9,9,9,108,111,99,97,108,32,100,121,32,61,32,112,105,101,99,101,46,103,114,105,100,95,121,32,45,32,101,120,99,108,117,100,101,95,99,101,110,116,101,114,95,121,10,9,9,9,9,108,111,99,97,108,32,108,101,110,32,61,32,103,101,110,101,114,97,116,105,111,110,32,45,32,109,97,116,104,46,115,113,114,116,40,100,120,32,94,32,50,32,43,32,100,121,32,94,32,50,41,10,9,9,9,9,114,101,116,117,114,110,32,67,79,76,79,82,83,91,49,32,43,32,109,97,116,104,46,102,108,111,111,114,40,108,101,110,41,32,37,32,35,67,79,76,79,82,83,93,10,9,9,9,101,110,100,44,10,9,9,9,102,117,110,99,116,105,111,110,40,112,105,101,99,101,44,32,103,101,110,101,114,97,116,105,111,110,41,10,9,9,9,9,108,111,99,97,108,32,100,120,32,61,32,109,97,116,104,46,97,98,115,40,112,105,101,99,101,46,103,114,105,100,95,120,32,45,32,101,120,99,108,117,100,101,95,99,101,110,116,101,114,95,120,41,32,45,32,103,101,110,101,114,97,116,105,111,110,10,9,9,9,9,108,111,99,97,108,32,100,121,32,61,32,109,97,116,104,46,97,98,115,40,112,105,101,99,101,46,103,114,105,100,95,121,32,45,32,101,120,99,108,117,100,101,95,99,101,110,116,101,114,95,121,41,32,45,32,103,101,110,101,114,97,116,105,111,110,10,9,9,9,9,114,101,116,117,114,110,32,67,79,76,79,82,83,91,49,32,43,32,109,97,116,104,46,102,108,111,111,114,40,109,97,116,104,46,109,97,120,40,100,120,44,32,100,121,41,41,32,37,32,35,67,79,76,79,82,83,93,10,9,9,9,101,110,100,44,10,9,9,125,10,9,9,115,101,108,102,46,103,101,110,101,114,97,116,111,114,32,61,32,115,101,108,102,46,103,101,110,101,114,97,116,111,114,115,91,49,93,10,9,9,108,111,99,97,108,32,69,88,67,76,85,68,69,32,61,32,123,125,10,9,9,102,111,114,32,121,32,61,32,101,120,99,108,117,100,101,95,116,111,112,44,101,120,99,108,117,100,101,95,98,111,116,116,111,109,32,100,111,10,9,9,9,69,88,67,76,85,68,69,91,121,93,32,32,61,32,123,125,10,9,9,9,102,111,114,32,120,32,61,32,101,120,99,108,117,100,101,95,108,101,102,116,44,101,120,99,108,117,100,101,95,114,105,103,104,116,32,100,111,10,9,9,9,9,69,88,67,76,85,68,69,91,121,93,91,120,93,32,61,32,116,114,117,101,10,9,9,9,101,110,100,10,9,9,101,110,100,10,9,9,102,111,114,32,121,32,61,32,49,44,83,73,90,69,95,89,32,100,111,10,9,9,9,102,111,114,32,120,32,61,32,49,44,83,73,90,69,95,88,32,100,111,10,9,9,9,9,105,102,32,110,111,116,32,69,88,67,76,85,68,69,91,121,93,32,111,114,32,110,111,116,32,69,88,67,76,85,68,69,91,121,93,91,120,93,32,116,104,101,110,10,9,9,9,9,9,108,111,99,97,108,32,112,105,101,99,101,32,61,32,123,10,9,9,9,9,9,9,103,114,105,100,95,120,32,61,32,120,44,10,9,9,9,9,9,9,103,114,105,100,95,121,32,61,32,121,44,10,9,9,9,9,9,9,120,32,61,32,40,120,32,45,32,49,41,32,42,32,51,50,44,10,9,9,9,9,9,9,121,32,61,32,40,121,32,45,32,49,41,32,42,32,51,50,44,10,9,9,9,9,9,9,114,32,61,32,108,111,118,101,46,109,97,116,104,46,114,97,110,100,111,109,40,48,44,32,49,48,48,41,32,47,32,49,48,48,32,42,32,109,97,116,104,46,112,105,44,10,9,9,9,9,9,9,114,118,32,61,32,49,44,10,9,9,9,9,9,9,99,111,108,111,114,32,61,32,123,125,44,10,9,9,9,9,9,9,113,117,97,100,32,61,32,81,85,65,68,83,91,40,120,32,43,32,121,41,32,37,32,52,32,43,32,49,93,10,9,9,9,9,9,125,10,9,9,9,9,9,112,105,101,99,101,46,99,111,108,111,114,46,112,114,101,118,32,61,32,115,101,108,102,46,103,101,110,101,114,97,116,111,114,40,112,105,101,99,101,44,32,115,101,108,102,46,103,101,110,101,114,97,116,105,111,110,41,10,9,9,9,9,9,112,105,101,99,101,46,99,111,108,111,114,46,110,101,120,116,32,61,32,112,105,101,99,101,46,99,111,108,111,114,46,112,114,101,118,10,9,9,9,9,9,116,97,98,108,101,46,105,110,115,101,114,116,40,115,101,108,102,46,112,105,101,99,101,115,44,32,112,105,101,99,101,41,10,9,9,9,9,101,110,100,10,9,9,9,101,110,100,10,9,9,101,110,100,10,9,9,108,111,99,97,108,32,71,76,89,80,72,83,32,61,32,123,10,9,9,9,78,32,61,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,81,117,97,100,40,48,44,32,32,54,52,44,32,51,50,44,32,51,50,44,32,115,119,44,32,115,104,41,44,10,9,9,9,79,32,61,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,81,117,97,100,40,51,50,44,32,54,52,44,32,51,50,44,32,51,50,44,32,115,119,44,32,115,104,41,44,10,9,9,9,71,32,61,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,81,117,97,100,40,48,44,32,32,57,54,44,32,51,50,44,32,51,50,44,32,115,119,44,32,115,104,41,44,10,9,9,9,65,32,61,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,81,117,97,100,40,51,50,44,32,57,54,44,32,51,50,44,32,51,50,44,32,115,119,44,32,115,104,41,44,10,9,9,9,77,32,61,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,81,117,97,100,40,54,52,44,32,57,54,44,32,51,50,44,32,51,50,44,32,115,119,44,32,115,104,41,44,10,9,9,9,69,32,61,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,81,117,97,100,40,57,54,44,32,57,54,44,32,51,50,44,32,51,50,44,32,115,119,44,32,115,104,41,44,10,9,9,9,85,32,61,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,81,117,97,100,40,54,52,44,32,48,44,32,32,51,50,44,32,51,50,44,32,115,119,44,32,115,104,41,44,10,9,9,9,80,32,61,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,81,117,97,100,40,57,54,44,32,48,44,32,32,51,50,44,32,51,50,44,32,115,119,44,32,115,104,41,44,10,9,9,9,111,32,61,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,81,117,97,100,40,54,52,44,32,51,50,44,32,51,50,44,32,51,50,44,32,115,119,44,32,115,104,41,44,10,9,9,9,83,32,61,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,81,117,97,100,40,57,54,44,32,51,50,44,32,51,50,44,32,51,50,44,32,115,119,44,32,115,104,41,44,10,9,9,9,82,32,61,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,81,117,97,100,40,54,52,44,32,54,52,44,32,51,50,44,32,51,50,44,32,115,119,44,32,115,104,41,44,10,9,9,9,84,32,61,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,81,117,97,100,40,57,54,44,32,54,52,44,32,51,50,44,32,51,50,44,32,115,119,44,32,115,104,41,44,10,9,9,125,10,9,9,108,111,99,97,108,32,73,78,73,84,73,65,76,95,84,69,88,84,95,67,79,76,79,82,32,61,32,123,32,50,52,48,44,32,50,52,48,44,32,50,52,48,32,125,10,9,9,108,111,99,97,108,32,112,117,116,95,116,101,120,116,32,61,32,102,117,110,99,116,105,111,110,40,115,116,114,44,32,111,102,102,115,101,116,44,32,120,44,32,121,41,10,9,9,9,108,111,99,97,108,32,105,100,120,32,61,32,111,102,102,115,101,116,32,43,32,83,73,90,69,95,88,32,42,32,121,32,43,32,120,10,9,9,9,102,111,114,32,105,32,61,32,49,44,32,35,115,116,114,32,100,111,10,9,9,9,9,108,111,99,97,108,32,99,32,61,32,115,116,114,58,115,117,98,40,105,44,32,105,41,10,9,9,9,9,105,102,32,99,32,126,61,32,34,32,34,32,116,104,101,110,10,9,9,9,9,9,108,111,99,97,108,32,112,105,101,99,101,32,61,32,115,101,108,102,46,112,105,101,99,101,115,91,105,100,120,32,43,32,105,93,10,9,9,9,9,9,105,102,32,112,105,101,99,101,32,116,104,101,110,10,9,9,9,9,9,9,112,105,101,99,101,46,113,117,97,100,32,61,32,71,76,89,80,72,83,91,99,93,10,9,9,9,9,9,9,112,105,101,99,101,46,114,32,61,32,48,10,9,9,9,9,9,9,112,105,101,99,101,46,114,118,32,61,32,48,10,9,9,9,9,9,9,112,105,101,99,101,46,99,111,108,111,114,46,112,114,101,118,32,61,32,73,78,73,84,73,65,76,95,84,69,88,84,95,67,79,76,79,82,10,9,9,9,9,9,9,112,105,101,99,101,46,99,111,108,111,114,46,110,101,120,116,32,61,32,73,78,73,84,73,65,76,95,84,69,88,84,95,67,79,76,79,82,10,9,9,9,9,9,101,110,100,10,9,9,9,9,101,110,100,10,9,9,9,101,110,100,10,9,9,101,110,100,10,9,9,108,111,99,97,108,32,116,101,120,116,95,99,101,110,116,101,114,95,120,32,61,32,109,97,116,104,46,102,108,111,111,114,40,119,119,32,47,32,50,32,47,32,51,50,41,10,9,9,108,111,99,97,108,32,110,111,95,103,97,109,101,95,116,101,120,116,95,111,102,102,115,101,116,32,61,32,83,73,90,69,95,88,32,42,32,101,120,99,108,117,100,101,95,98,111,116,116,111,109,32,45,32,101,120,99,108,117,100,101,95,97,114,101,97,10,9,9,112,117,116,95,116,101,120,116,40,34,78,111,32,71,65,77,69,34,44,32,110,111,95,103,97,109,101,95,116,101,120,116,95,111,102,102,115,101,116,44,32,116,101,120,116,95,99,101,110,116,101,114,95,120,32,45,32,50,44,32,49,41,10,9,9,112,117,116,95,116,101,120,116,40,34,83,85,80,69,82,32,84,79,65,83,84,34,44,32,48,44,32,116,101,120,116,95,99,101,110,116,101,114,95,120,32,45,32,52,44,32,101,120,99,108,117,100,101,95,116,111,112,32,45,32,51,41,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,77,111,115,97,105,99,58,97,100,100,71,101,110,101,114,97,116,105,111,110,40,41,10,9,9,115,101,108,102,46,103,101,110,101,114,97,116,105,111,110,32,61,32,115,101,108,102,46,103,101,110,101,114,97,116,105,111,110,32,43,32,49,10,9,9,105,102,32,115,101,108,102,46,103,101,110,101,114,97,116,105,111,110,32,37,32,53,32,61,61,32,48,32,116,104,101,110,10,9,9,9,105,102,32,108,111,118,101,46,109,97,116,104,46,114,97,110,100,111,109,40,48,44,32,49,48,48,41,32,60,32,54,48,32,116,104,101,110,10,9,9,9,9,115,101,108,102,46,103,101,110,101,114,97,116,111,114,32,61,32,115,101,108,102,46,103,101,110,101,114,97,116,111,114,115,91,108,111,118,101,46,109,97,116,104,46,114,97,110,100,111,109,40,50,44,32,35,115,101,108,102,46,103,101,110,101,114,97,116,111,114,115,41,93,10,9,9,9,101,108,115,101,10,9,9,9,9,115,101,108,102,46,103,101,110,101,114,97,116,111,114,32,61,32,115,101,108,102,46,103,101,110,101,114,97,116,111,114,115,91,49,93,10,9,9,9,101,110,100,10,9,9,101,110,100,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,77,111,115,97,105,99,58,117,112,100,97,116,101,40,100,116,41,10,9,9,115,101,108,102,46,99,111,108,111,114,95,116,32,61,32,109,97,116,104,46,109,97,120,40,115,101,108,102,46,99,111,108,111,114,95,116,32,45,32,100,116,44,32,48,41,10,9,9,108,111,99,97,108,32,99,104,97,110,103,101,95,99,111,108,111,114,32,61,32,115,101,108,102,46,99,111,108,111,114,95,116,32,61,61,32,48,10,9,9,105,102,32,99,104,97,110,103,101,95,99,111,108,111,114,32,116,104,101,110,10,9,9,9,115,101,108,102,46,99,111,108,111,114,95,116,32,61,32,49,10,9,9,9,115,101,108,102,58,97,100,100,71,101,110,101,114,97,116,105,111,110,40,41,10,9,9,101,110,100,10,9,9,108,111,99,97,108,32,103,101,110,32,61,32,115,101,108,102,46,103,101,110,101,114,97,116,111,114,10,9,9,102,111,114,32,105,100,120,44,112,105,101,99,101,32,105,110,32,105,112,97,105,114,115,40,115,101,108,102,46,112,105,101,99,101,115,41,32,100,111,10,9,9,9,112,105,101,99,101,46,114,32,61,32,112,105,101,99,101,46,114,32,43,32,112,105,101,99,101,46,114,118,32,42,32,100,116,10,9,9,9,105,102,32,99,104,97,110,103,101,95,99,111,108,111,114,32,116,104,101,110,10,9,9,9,9,112,105,101,99,101,46,99,111,108,111,114,46,112,114,101,118,32,61,32,112,105,101,99,101,46,99,111,108,111,114,46,110,101,120,116,10,9,9,9,9,112,105,101,99,101,46,99,111,108,111,114,46,110,101,120,116,32,61,32,103,101,110,40,112,105,101,99,101,44,32,115,101,108,102,46,103,101,110,101,114,97,116,105,111,110,41,10,9,9,9,101,110,100,10,9,9,101,110,100,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,77,111,115,97,105,99,58,100,114,97,119,40,41,10,9,9,115,101,108,102,46,98,97,116,99,104,58,99,108,101,97,114,40,41,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,115,101,116,67,111,108,111,114,40,50,53,53,44,32,50,53,53,44,32,50,53,53,44,32,54,52,41,10,9,9,102,111,114,32,105,100,120,44,112,105,101,99,101,32,105,110,32,105,112,97,105,114,115,40,115,101,108,102,46,112,105,101,99,101,115,41,32,100,111,10,9,9,9,108,111,99,97,108,32,99,116,32,61,32,49,32,45,32,115,101,108,102,46,99,111,108,111,114,95,116,10,9,9,9,108,111,99,97,108,32,99,48,32,61,32,112,105,101,99,101,46,99,111,108,111,114,46,112,114,101,118,10,9,9,9,108,111,99,97,108,32,99,49,32,61,32,112,105,101,99,101,46,99,111,108,111,114,46,110,101,120,116,10,9,9,9,108,111,99,97,108,32,114,32,61,32,101,97,115,101,79,117,116,40,99,116,44,32,99,48,91,49,93,44,32,99,49,91,49,93,32,45,32,99,48,91,49,93,44,32,49,41,10,9,9,9,108,111,99,97,108,32,103,32,61,32,101,97,115,101,79,117,116,40,99,116,44,32,99,48,91,50,93,44,32,99,49,91,50,93,32,45,32,99,48,91,50,93,44,32,49,41,10,9,9,9,108,111,99,97,108,32,98,32,61,32,101,97,115,101,79,117,116,40,99,116,44,32,99,48,91,51,93,44,32,99,49,91,51,93,32,45,32,99,48,91,51,93,44,32,49,41,10,9,9,9,115,101,108,102,46,98,97,116,99,104,58,115,101,116,67,111,108,111,114,40,114,44,32,103,44,32,98,41,10,9,9,9,115,101,108,102,46,98,97,116,99,104,58,97,100,100,40,112,105,101,99,101,46,113,117,97,100,44,32,112,105,101,99,101,46,120,44,32,112,105,101,99,101,46,121,44,32,112,105,101,99,101,46,114,44,32,49,44,32,49,44,32,49,54,44,32,49,54,41,10,9,9,101,110,100,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,115,101,116,67,111,108,111,114,40,50,53,53,44,32,50,53,53,44,32,50,53,53,44,32,50,53,53,41,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,100,114,97,119,40,115,101,108,102,46,98,97,116,99,104,44,32,48,44,32,48,41,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,108,111,118,101,46,108,111,97,100,40,41,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,115,101,116,66,97,99,107,103,114,111,117,110,100,67,111,108,111,114,40,49,51,54,44,32,49,57,51,44,32,50,48,54,41,10,9,9,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,108,111,97,100,95,105,109,97,103,101,40,102,105,108,101,44,32,110,97,109,101,41,10,9,9,9,114,101,116,117,114,110,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,73,109,97,103,101,40,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,46,110,101,119,70,105,108,101,68,97,116,97,40,102,105,108,101,44,32,110,97,109,101,58,103,115,117,98,40,34,95,34,44,32,34,46,34,41,44,32,34,98,97,115,101,54,52,34,41,41,10,9,9,101,110,100,10,9,9,103,95,105,109,97,103,101,115,32,61,32,123,125,10,9,9,103,95,105,109,97,103,101,115,46,116,111,97,115,116,32,61,32,123,125,10,9,9,103,95,105,109,97,103,101,115,46,116,111,97,115,116,46,98,97,99,107,32,61,32,108,111,97,100,95,105,109,97,103,101,40,116,111,97,115,116,95,98,97,99,107,95,112,110,103,44,32,34,116,111,97,115,116,95,98,97,99,107,46,112,110,103,34,41,10,9,9,103,95,105,109,97,103,101,115,46,116,111,97,115,116,46,102,114,111,110,116,32,61,32,108,111,97,100,95,105,109,97,103,101,40,116,111,97,115,116,95,102,114,111,110,116,95,112,110,103,44,32,34,116,111,97,115,116,95,102,114,111,110,116,46,112,110,103,34,41,10,9,9,103,95,105,109,97,103,101,115,46,116,111,97,115,116,46,101,121,101,115,32,61,32,123,125,10,9,9,103,95,105,109,97,103,101,115,46,116,111,97,115,116,46,101,121,101,115,46,111,112,101,110,32,61,32,108,111,97,100,95,105,109,97,103,101,40,116,111,97,115,116,95,101,121,101,115,95,111,112,101,110,95,112,110,103,44,32,34,116,111,97,115,116,95,101,121,101,115,95,111,112,101,110,46,112,110,103,34,41,10,9,9,103,95,105,109,97,103,101,115,46,116,111,97,115,116,46,101,121,101,115,46,99,108,111,115,101,100,32,61,32,108,111,97,100,95,105,109,97,103,101,40,116,111,97,115,116,95,101,121,101,115,95,99,108,111,115,101,100,95,112,110,103,44,32,34,116,111,97,115,116,95,101,121,101,115,95,99,108,111,115,101,100,46,112,110,103,34,41,10,9,9,103,95,105,109,97,103,101,115,46,116,111,97,115,116,46,109,111,117,116,104,32,61,32,108,111,97,100,95,105,109,97,103,101,40,116,111,97,115,116,95,109,111,117,116,104,95,112,110,103,44,32,34,116,111,97,115,116,95,109,111,117,116,104,46,112,110,103,34,41,10,9,9,103,95,105,109,97,103,101,115,46,109,111,115,97,105,99,32,61,32,123,125,10,9,9,103,95,105,109,97,103,101,115,46,109,111,115,97,105,99,91,49,93,32,61,32,108,111,97,100,95,105,109,97,103,101,40,109,111,115,97,105,99,95,112,110,103,44,32,34,109,111,115,97,105,99,46,112,110,103,34,41,10,9,9,103,95,105,109,97,103,101,115,46,109,111,115,97,105,99,91,50,93,32,61,32,108,111,97,100,95,105,109,97,103,101,40,109,111,115,97,105,99,95,50,120,95,112,110,103,44,32,34,109,111,115,97,105,99,64,50,120,46,112,110,103,34,41,10,9,9,103,95,101,110,116,105,116,105,101,115,32,61,32,123,125,10,9,9,103,95,101,110,116,105,116,105,101,115,46,116,111,97,115,116,32,61,32,84,111,97,115,116,40,41,10,9,9,103,95,101,110,116,105,116,105,101,115,46,109,111,115,97,105,99,32,61,32,77,111,115,97,105,99,40,41,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,108,111,118,101,46,117,112,100,97,116,101,40,100,116,41,10,9,9,100,116,32,61,32,109,97,116,104,46,109,105,110,40,100,116,44,32,49,47,49,48,41,10,9,9,103,95,101,110,116,105,116,105,101,115,46,116,111,97,115,116,58,117,112,100,97,116,101,40,100,116,41,10,9,9,103,95,101,110,116,105,116,105,101,115,46,109,111,115,97,105,99,58,117,112,100,97,116,101,40,100,116,41,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,108,111,118,101,46,100,114,97,119,40,41,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,115,101,116,67,111,108,111,114,40,50,53,53,44,32,50,53,53,44,32,50,53,53,41,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,112,117,115,104,40,41,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,115,99,97,108,101,40,108,111,118,101,46,119,105,110,100,111,119,46,103,101,116,80,105,120,101,108,83,99,97,108,101,40,41,41,10,9,9,103,95,101,110,116,105,116,105,101,115,46,109,111,115,97,105,99,58,100,114,97,119,40,41,10,9,9,103,95,101,110,116,105,116,105,101,115,46,116,111,97,115,116,58,100,114,97,119,40,41,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,112,111,112,40,41,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,108,111,118,101,46,114,101,115,105,122,101,40,119,44,32,104,41,10,9,9,103,95,101,110,116,105,116,105,101,115,46,109,111,115,97,105,99,32,61,32,77,111,115,97,105,99,40,41,10,9,9,103,95,101,110,116,105,116,105,101,115,46,116,111,97,115,116,58,99,101,110,116,101,114,40,41,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,108,111,118,101,46,107,101,121,112,114,101,115,115,101,100,40,107,101,121,41,10,9,9,105,102,32,107,101,121,32,61,61,32,34,101,115,99,97,112,101,34,32,116,104,101,110,10,9,9,9,108,111,118,101,46,101,118,101,110,116,46,113,117,105,116,40,41,10,9,9,101,110,100,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,108,111,118,101,46,107,101,121,114,101,108,101,97,115,101,100,40,107,101,121,41,10,9,9,105,102,32,107,101,121,32,61,61,32,34,102,34,32,116,104,101,110,10,9,9,9,108,111,99,97,108,32,105,115,95,102,115,32,61,32,108,111,118,101,46,119,105,110,100,111,119,46,103,101,116,70,117,108,108,115,99,114,101,101,110,40,41,10,9,9,9,108,111,118,101,46,119,105,110,100,111,119,46,115,101,116,70,117,108,108,115,99,114,101,101,110,40,110,111,116,32,105,115,95,102,115,41,10,9,9,101,110,100,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,108,111,118,101,46,109,111,117,115,101,112,114,101,115,115,101,100,40,120,44,32,121,44,32,98,41,10,9,9,108,111,99,97,108,32,116,120,32,61,32,120,32,47,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,103,101,116,87,105,100,116,104,40,41,10,9,9,108,111,99,97,108,32,116,121,32,61,32,121,32,47,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,103,101,116,72,101,105,103,104,116,40,41,10,9,9,103,95,101,110,116,105,116,105,101,115,46,116,111,97,115,116,58,108,111,111,107,95,97,116,40,116,120,44,32,116,121,41,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,108,111,118,101,46,109,111,117,115,101,109,111,118,101,100,40,120,44,32,121,41,10,9,9,105,102,32,108,111,118,101,46,109,111,117,115,101,46,105,115,68,111,119,110,40,49,41,32,116,104,101,110,10,9,9,9,108,111,99,97,108,32,116,120,32,61,32,120,32,47,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,103,101,116,87,105,100,116,104,40,41,10,9,9,9,108,111,99,97,108,32,116,121,32,61,32,121,32,47,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,103,101,116,72,101,105,103,104,116,40,41,10,9,9,9,103,95,101,110,116,105,116,105,101,115,46,116,111,97,115,116,58,108,111,111,107,95,97,116,40,116,120,44,32,116,121,41,10,9,9,101,110,100,10,9,101,110,100,10,9,108,111,99,97,108,32,108,97,115,116,95,116,111,117,99,104,32,61,32,123,116,105,109,101,61,48,44,32,120,61,48,44,32,121,61,48,125,10,9,102,117,110,99,116,105,111,110,32,108,111,118,101,46,116,111,117,99,104,112,114,101,115,115,101,100,40,105,100,44,32,120,44,32,121,44,32,112,114,101,115,115,117,114,101,41,10,9,9,45,45,32,68,111,117,98,108,101,45,116,97,112,32,116,104,101,32,115,99,114,101,101,110,32,40,119,104,101,110,32,117,115,105,110,103,32,97,32,116,111,117,99,104,32,115,99,114,101,101,110,41,32,116,111,32,101,120,105,116,46,10,9,9,105,102,32,35,108,111,118,101,46,116,111,117,99,104,46,103,101,116,84,111,117,99,104,101,115,40,41,32,61,61,32,49,32,116,104,101,110,10,9,9,9,108,111,99,97,108,32,100,105,115,116,32,61,32,109,97,116,104,46,115,113,114,116,40,40,120,45,108,97,115,116,95,116,111,117,99,104,46,120,41,94,50,32,43,32,40,121,45,108,97,115,116,95,116,111,117,99,104,46,121,41,94,50,41,10,9,9,9,108,111,99,97,108,32,100,105,102,102,116,105,109,101,32,61,32,108,111,118,101,46,116,105,109,101,114,46,103,101,116,84,105,109,101,40,41,32,45,32,108,97,115,116,95,116,111,117,99,104,46,116,105,109,101,10,9,9,9,105,102,32,100,105,102,102,116,105,109,101,32,60,32,48,46,51,32,97,110,100,32,100,105,115,116,32,60,32,53,48,32,116,104,101,110,10,9,9,9,9,105,102,32,108,111,118,101,46,119,105,110,100,111,119,46,115,104,111,119,77,101,115,115,97,103,101,66,111,120,40,34,69,120,105,116,32,78,111,45,71,97,109,101,32,83,99,114,101,101,110,34,44,32,34,34,44,32,123,34,79,75,34,44,32,34,67,97,110,99,101,108,34,125,41,32,61,61,32,49,32,116,104,101,110,10,9,9,9,9,9,108,111,118,101,46,101,118,101,110,116,46,113,117,105,116,40,41,10,9,9,9,9,101,110,100,10,9,9,9,101,110,100,10,9,9,9,108,97,115,116,95,116,111,117,99,104,46,116,105,109,101,32,61,32,108,111,118,101,46,116,105,109,101,114,46,103,101,116,84,105,109,101,40,41,10,9,9,9,108,97,115,116,95,116,111,117,99,104,46,120,32,61,32,120,10,9,9,9,108,97,115,116,95,116,111,117,99,104,46,121,32,61,32,121,10,9,9,101,110,100,10,9,101,110,100,10,9,102,117,110,99,116,105,111,110,32,108,111,118,101,46,99,111,110,102,40,116,41,10,9,9,116,46,116,105,116,108,101,32,61,32,34,76,92,49,57,53,92,49,53,48,86,69,32,34,32,46,46,32,108,111,118,101,46,95,118,101,114,115,105,111,110,32,46,46,32,34,32,40,34,32,46,46,32,108,111,118,101,46,95,118,101,114,115,105,111,110,95,99,111,100,101,110,97,109,101,32,46,46,32,34,41,34,10,9,9,116,46,103,97,109,109,97,99,111,114,114,101,99,116,32,61,32,116,114,117,101,10,9,9,116,46,109,111,100,117,108,101,115,46,97,117,100,105,111,32,61,32,102,97,108,115,101,10,9,9,116,46,109,111,100,117,108,101,115,46,115,111,117,110,100,32,61,32,102,97,108,115,101,10,9,9,116,46,109,111,100,117,108,101,115,46,112,104,121,115,105,99,115,32,61,32,102,97,108,115,101,10,9,9,116,46,109,111,100,117,108,101,115,46,106,111,121,115,116,105,99,107,32,61,32,102,97,108,115,101,10,9,9,116,46,119,105,110,100,111,119,46,114,101,115,105,122,97,98,108,101,32,61,32,116,114,117,101,10,9,9,116,46,119,105,110,100,111,119,46,104,105,103,104,100,112,105,32,61,32,116,114,117,101,10,9,9,105,102,32,108,111,118,101,46,95,111,115,32,61,61,32,34,105,79,83,34,32,116,104,101,110,10,9,9,9,116,46,119,105,110,100,111,119,46,98,111,114,100,101,114,108,101,115,115,32,61,32,116,114,117,101,10,9,9,101,110,100,10,9,101,110,100,10,101,110,100,10,114,101,116,117,114,110,32,108,111,118,101,46,110,111,103,97,109,101,10,110,111,103,97,109,101,46,108,117,97,0,108,111,118,101,46,98,111,111,116,0,45,45,91,91,10,67,111,112,121,114,105,103,104,116,32,40,99,41,32,50,48,48,54,45,50,48,49,54,32,76,79,86,69,32,68,101,118,101,108,111,112,109,101,110,116,32,84,101,97,109,10,84,104,105,115,32,115,111,102,116,119,97,114,101,32,105,115,32,112,114,111,118,105,100,101,100,32,39,97,115,45,105,115,39,44,32,119,105,116,104,111,117,116,32,97,110,121,32,101,120,112,114,101,115,115,32,111,114,32,105,109,112,108,105,101,100,10,119,97,114,114,97,110,116,121,46,32,32,73,110,32,110,111,32,101,118,101,110,116,32,119,105,108,108,32,116,104,101,32,97,117,116,104,111,114,115,32,98,101,32,104,101,108,100,32,108,105,97,98,108,101,32,102,111,114,32,97,110,121,32,100,97,109,97,103,101,115,10,97,114,105,115,105,110,103,32,102,114,111,109,32,116,104,101,32,117,115,101,32,111,102,32,116,104,105,115,32,115,111,102,116,119,97,114,101,46,10,80,101,114,109,105,115,115,105,111,110,32,105,115,32,103,114,97,110,116,101,100,32,116,111,32,97,110,121,111,110,101,32,116,111,32,117,115,101,32,116,104,105,115,32,115,111,102,116,119,97,114,101,32,102,111,114,32,97,110,121,32,112,117,114,112,111,115,101,44,10,105,110,99,108,117,100,105,110,103,32,99,111,109,109,101,114,99,105,97,108,32,97,112,112,108,105,99,97,116,105,111,110,115,44,32,97,110,100,32,116,111,32,97,108,116,101,114,32,105,116,32,97,110,100,32,114,101,100,105,115,116,114,105,98,117,116,101,32,105,116,10,102,114,101,101,108,121,44,32,115,117,98,106,101,99,116,32,116,111,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,114,101,115,116,114,105,99,116,105,111,110,115,58,10,49,46,32,84,104,101,32,111,114,105,103,105,110,32,111,102,32,116,104,105,115,32,115,111,102,116,119,97,114,101,32,109,117,115,116,32,110,111,116,32,98,101,32,109,105,115,114,101,112,114,101,115,101,110,116,101,100,59,32,121,111,117,32,109,117,115,116,32,110,111,116,10,32,32,32,99,108,97,105,109,32,116,104,97,116,32,121,111,117,32,119,114,111,116,101,32,116,104,101,32,111,114,105,103,105,110,97,108,32,115,111,102,116,119,97,114,101,46,32,73,102,32,121,111,117,32,117,115,101,32,116,104,105,115,32,115,111,102,116,119,97,114,101,10,32,32,32,105,110,32,97,32,112,114,111,100,117,99,116,44,32,97,110,32,97,99,107,110,111,119,108,101,100,103,109,101,110,116,32,105,110,32,116,104,101,32,112,114,111,100,117,99,116,32,100,111,99,117,109,101,110,116,97,116,105,111,110,32,119,111,117,108,100,32,98,101,10,32,32,32,97,112,112,114,101,99,105,97,116,101,100,32,98,117,116,32,105,115,32,110,111,116,32,114,101,113,117,105,114,101,100,46,10,50,46,32,65,108,116,101,114,101,100,32,115,111,117,114,99,101,32,118,101,114,115,105,111,110,115,32,109,117,115,116,32,98,101,32,112,108,97,105,110,108,121,32,109,97,114,107,101,100,32,97,115,32,115,117,99,104,44,32,97,110,100,32,109,117,115,116,32,110,111,116,32,98,101,10,32,32,32,109,105,115,114,101,112,114,101,115,101,110,116,101,100,32,97,115,32,98,101,105,110,103,32,116,104,101,32,111,114,105,103,105,110,97,108,32,115,111,102,116,119,97,114,101,46,10,51,46,32,84,104,105,115,32,110,111,116,105,99,101,32,109,97,121,32,110,111,116,32,98,101,32,114,101,109,111,118,101,100,32,111,114,32,97,108,116,101,114,101,100,32,102,114,111,109,32,97,110,121,32,115,111,117,114,99,101,32,100,105,115,116,114,105,98,117,116,105,111,110,46,10,45,45,93,93,10,45,45,32,77,97,107,101,32,115,117,114,101,32,108,111,118,101,32,101,120,105,115,116,115,46,10,108,111,99,97,108,32,108,111,118,101,32,61,32,114,101,113,117,105,114,101,40,34,108,111,118,101,34,41,10,45,45,32,85,115,101,100,32,102,111,114,32,115,101,116,117,112,58,10,108,111,118,101,46,112,97,116,104,32,61,32,123,125,10,108,111,118,101,46,97,114,103,32,61,32,123,125,10,45,45,32,82,101,112,108,97,99,101,32,97,110,121,32,92,32,119,105,116,104,32,47,46,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,112,97,116,104,46,110,111,114,109,97,108,115,108,97,115,104,101,115,40,112,41,10,9,114,101,116,117,114,110,32,115,116,114,105,110,103,46,103,115,117,98,40,112,44,32,34,92,92,34,44,32,34,47,34,41,10,101,110,100,10,45,45,32,77,97,107,101,115,32,115,117,114,101,32,116,104,101,114,101,32,105,115,32,97,32,115,108,97,115,104,32,97,116,32,116,104,101,32,101,110,100,10,45,45,32,111,102,32,97,32,112,97,116,104,46,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,112,97,116,104,46,101,110,100,115,108,97,115,104,40,112,41,10,9,105,102,32,115,116,114,105,110,103,46,115,117,98,40,112,44,32,115,116,114,105,110,103,46,108,101,110,40,112,41,45,49,41,32,126,61,32,34,47,34,32,116,104,101,110,10,9,9,114,101,116,117,114,110,32,112,32,46,46,32,34,47,34,10,9,101,108,115,101,10,9,9,114,101,116,117,114,110,32,112,10,9,101,110,100,10,101,110,100,10,45,45,32,67,104,101,99,107,115,32,119,104,101,116,104,101,114,32,97,32,112,97,116,104,32,105,115,32,97,98,115,111,108,117,116,101,32,111,114,32,110,111,116,46,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,112,97,116,104,46,97,98,115,40,112,41,10,9,108,111,99,97,108,32,116,109,112,32,61,32,108,111,118,101,46,112,97,116,104,46,110,111,114,109,97,108,115,108,97,115,104,101,115,40,112,41,10,9,45,45,32,80,97,116,104,32,105,115,32,97,98,115,111,108,117,116,101,32,105,102,32,105,116,32,115,116,97,114,116,115,32,119,105,116,104,32,97,32,34,47,34,46,10,9,105,102,32,115,116,114,105,110,103,46,102,105,110,100,40,116,109,112,44,32,34,47,34,41,32,61,61,32,49,32,116,104,101,110,10,9,9,114,101,116,117,114,110,32,116,114,117,101,10,9,101,110,100,10,9,45,45,32,80,97,116,104,32,105,115,32,97,98,115,111,108,117,116,101,32,105,102,32,105,116,32,115,116,97,114,116,115,32,119,105,116,104,32,97,10,9,45,45,32,108,101,116,116,101,114,32,102,111,108,108,111,119,101,100,32,98,121,32,97,32,99,111,108,111,110,46,10,9,105,102,32,115,116,114,105,110,103,46,102,105,110,100,40,116,109,112,44,32,34,37,97,58,34,41,32,61,61,32,49,32,116,104,101,110,10,9,9,114,101,116,117,114,110,32,116,114,117,101,10,9,101,110,100,10,9,45,45,32,82,101,108,97,116,105,118,101,46,10,9,114,101,116,117,114,110,32,102,97,108,115,101,10,101,110,100,10,45,45,32,67,111,110,118,101,114,116,115,32,97,110,121,32,112,97,116,104,32,105,110,116,111,32,97,32,102,117,108,108,32,112,97,116,104,46,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,112,97,116,104,46,103,101,116,102,117,108,108,40,112,41,10,9,105,102,32,108,111,118,101,46,112,97,116,104,46,97,98,115,40,112,41,32,116,104,101,110,10,9,9,114,101,116,117,114,110,32,108,111,118,101,46,112,97,116,104,46,110,111,114,109,97,108,115,108,97,115,104,101,115,40,112,41,10,9,101,110,100,10,9,108,111,99,97,108,32,99,119,100,32,61,32,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,46,103,101,116,87,111,114,107,105,110,103,68,105,114,101,99,116,111,114,121,40,41,10,9,99,119,100,32,61,32,108,111,118,101,46,112,97,116,104,46,110,111,114,109,97,108,115,108,97,115,104,101,115,40,99,119,100,41,10,9,99,119,100,32,61,32,108,111,118,101,46,112,97,116,104,46,101,110,100,115,108,97,115,104,40,99,119,100,41,10,9,45,45,32,67,111,110,115,116,114,117,99,116,32,97,32,102,117,108,108,32,112,97,116,104,46,10,9,108,111,99,97,108,32,102,117,108,108,32,61,32,99,119,100,32,46,46,32,108,111,118,101,46,112,97,116,104,46,110,111,114,109,97,108,115,108,97,115,104,101,115,40,112,41,10,9,45,45,32,82,101,109,111,118,101,32,116,114,97,105,108,105,110,103,32,47,46,44,32,105,102,32,97,112,112,108,105,99,97,98,108,101,10,9,114,101,116,117,114,110,32,102,117,108,108,58,109,97,116,99,104,40,34,40,46,45,41,47,37,46,36,34,41,32,111,114,32,102,117,108,108,10,101,110,100,10,45,45,32,82,101,116,117,114,110,115,32,116,104,101,32,108,101,97,102,32,111,102,32,97,32,102,117,108,108,32,112,97,116,104,46,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,112,97,116,104,46,108,101,97,102,40,112,41,10,9,112,32,61,32,108,111,118,101,46,112,97,116,104,46,110,111,114,109,97,108,115,108,97,115,104,101,115,40,112,41,10,9,108,111,99,97,108,32,97,32,61,32,49,10,9,108,111,99,97,108,32,108,97,115,116,32,61,32,112,10,9,119,104,105,108,101,32,97,32,100,111,10,9,9,97,32,61,32,115,116,114,105,110,103,46,102,105,110,100,40,112,44,32,34,47,34,44,32,97,43,49,41,10,9,9,105,102,32,97,32,116,104,101,110,10,9,9,9,108,97,115,116,32,61,32,115,116,114,105,110,103,46,115,117,98,40,112,44,32,97,43,49,41,10,9,9,101,110,100,10,9,101,110,100,10,9,114,101,116,117,114,110,32,108,97,115,116,10,101,110,100,10,45,45,32,70,105,110,100,115,32,116,104,101,32,107,101,121,32,105,110,32,116,104,101,32,116,97,98,108,101,32,119,105,116,104,32,116,104,101,32,108,111,119,101,115,116,32,105,110,116,101,103,114,97,108,32,105,110,100,101,120,46,32,84,104,101,32,108,111,119,101,115,116,10,45,45,32,119,105,108,108,32,116,121,112,105,99,97,108,108,121,32,116,104,101,32,101,120,101,99,117,116,97,98,108,101,44,32,102,111,114,32,105,110,115,116,97,110,99,101,32,34,108,117,97,53,46,49,46,101,120,101,34,46,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,97,114,103,46,103,101,116,76,111,119,40,97,41,10,9,108,111,99,97,108,32,109,32,61,32,109,97,116,104,46,104,117,103,101,10,9,102,111,114,32,107,44,118,32,105,110,32,112,97,105,114,115,40,97,41,32,100,111,10,9,9,105,102,32,107,32,60,32,109,32,116,104,101,110,10,9,9,9,109,32,61,32,107,10,9,9,101,110,100,10,9,101,110,100,10,9,114,101,116,117,114,110,32,97,91,109,93,10,101,110,100,10,108,111,118,101,46,97,114,103,46,111,112,116,105,111,110,115,32,61,32,123,10,9,99,111,110,115,111,108,101,32,61,32,123,32,97,32,61,32,48,32,125,44,10,9,102,117,115,101,100,32,61,32,123,97,32,61,32,48,32,125,44,10,9,103,97,109,101,32,61,32,123,32,97,32,61,32,49,32,125,10,125,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,97,114,103,46,112,97,114,115,101,95,111,112,116,105,111,110,40,109,44,32,105,41,10,9,109,46,115,101,116,32,61,32,116,114,117,101,10,9,105,102,32,109,46,97,32,62,32,48,32,116,104,101,110,10,9,9,109,46,97,114,103,32,61,32,123,125,10,9,9,102,111,114,32,106,61,105,44,105,43,109,46,97,45,49,32,100,111,10,9,9,9,116,97,98,108,101,46,105,110,115,101,114,116,40,109,46,97,114,103,44,32,97,114,103,91,106,93,41,10,9,9,9,105,32,61,32,106,10,9,9,101,110,100,10,9,101,110,100,10,9,114,101,116,117,114,110,32,105,10,101,110,100,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,97,114,103,46,112,97,114,115,101,95,111,112,116,105,111,110,115,40,41,10,9,108,111,99,97,108,32,103,97,109,101,10,9,108,111,99,97,108,32,97,114,103,99,32,61,32,35,97,114,103,10,9,102,111,114,32,105,61,49,44,97,114,103,99,32,100,111,10,9,9,45,45,32,76,111,111,107,32,102,111,114,32,111,112,116,105,111,110,115,46,10,9,9,108,111,99,97,108,32,115,44,32,101,44,32,109,32,61,32,115,116,114,105,110,103,46,102,105,110,100,40,97,114,103,91,105,93,44,32,34,37,45,37,45,40,46,43,41,34,41,10,9,9,105,102,32,109,32,97,110,100,32,108,111,118,101,46,97,114,103,46,111,112,116,105,111,110,115,91,109,93,32,116,104,101,110,10,9,9,9,105,32,61,32,108,111,118,101,46,97,114,103,46,112,97,114,115,101,95,111,112,116,105,111,110,40,108,111,118,101,46,97,114,103,46,111,112,116,105,111,110,115,91,109,93,44,32,105,43,49,41,10,9,9,101,108,115,101,105,102,32,110,111,116,32,103,97,109,101,32,116,104,101,110,10,9,9,9,103,97,109,101,32,61,32,105,10,9,9,101,110,100,10,9,101,110,100,10,9,105,102,32,110,111,116,32,108,111,118,101,46,97,114,103,46,111,112,116,105,111,110,115,46,103,97,109,101,46,115,101,116,32,116,104,101,110,10,9,9,108,111,118,101,46,97,114,103,46,112,97,114,115,101,95,111,112,116,105,111,110,40,108,111,118,101,46,97,114,103,46,111,112,116,105,111,110,115,46,103,97,109,101,44,32,103,97,109,101,32,111,114,32,48,41,10,9,101,110,100,10,101,110,100,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,99,114,101,97,116,101,104,97,110,100,108,101,114,115,40,41,10,9,45,45,32,83,116,97,110,100,97,114,100,32,99,97,108,108,98,97,99,107,32,104,97,110,100,108,101,114,115,46,10,9,108,111,118,101,46,104,97,110,100,108,101,114,115,32,61,32,115,101,116,109,101,116,97,116,97,98,108,101,40,123,10,9,9,107,101,121,112,114,101,115,115,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,98,44,115,44,114,41,10,9,9,9,105,102,32,108,111,118,101,46,107,101,121,112,114,101,115,115,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,107,101,121,112,114,101,115,115,101,100,40,98,44,115,44,114,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,107,101,121,114,101,108,101,97,115,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,98,44,115,41,10,9,9,9,105,102,32,108,111,118,101,46,107,101,121,114,101,108,101,97,115,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,107,101,121,114,101,108,101,97,115,101,100,40,98,44,115,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,116,101,120,116,105,110,112,117,116,32,61,32,102,117,110,99,116,105,111,110,32,40,116,41,10,9,9,9,105,102,32,108,111,118,101,46,116,101,120,116,105,110,112,117,116,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,116,101,120,116,105,110,112,117,116,40,116,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,116,101,120,116,101,100,105,116,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,116,44,115,44,108,41,10,9,9,9,105,102,32,108,111,118,101,46,116,101,120,116,101,100,105,116,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,116,101,120,116,101,100,105,116,101,100,40,116,44,115,44,108,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,109,111,117,115,101,109,111,118,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,120,44,121,44,100,120,44,100,121,44,116,41,10,9,9,9,105,102,32,108,111,118,101,46,109,111,117,115,101,109,111,118,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,109,111,117,115,101,109,111,118,101,100,40,120,44,121,44,100,120,44,100,121,44,116,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,109,111,117,115,101,112,114,101,115,115,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,120,44,121,44,98,44,116,41,10,9,9,9,105,102,32,108,111,118,101,46,109,111,117,115,101,112,114,101,115,115,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,109,111,117,115,101,112,114,101,115,115,101,100,40,120,44,121,44,98,44,116,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,109,111,117,115,101,114,101,108,101,97,115,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,120,44,121,44,98,44,116,41,10,9,9,9,105,102,32,108,111,118,101,46,109,111,117,115,101,114,101,108,101,97,115,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,109,111,117,115,101,114,101,108,101,97,115,101,100,40,120,44,121,44,98,44,116,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,119,104,101,101,108,109,111,118,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,120,44,121,41,10,9,9,9,105,102,32,108,111,118,101,46,119,104,101,101,108,109,111,118,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,119,104,101,101,108,109,111,118,101,100,40,120,44,121,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,116,111,117,99,104,112,114,101,115,115,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,105,100,44,120,44,121,44,100,120,44,100,121,44,112,41,10,9,9,9,105,102,32,108,111,118,101,46,116,111,117,99,104,112,114,101,115,115,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,116,111,117,99,104,112,114,101,115,115,101,100,40,105,100,44,120,44,121,44,100,120,44,100,121,44,112,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,116,111,117,99,104,114,101,108,101,97,115,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,105,100,44,120,44,121,44,100,120,44,100,121,44,112,41,10,9,9,9,105,102,32,108,111,118,101,46,116,111,117,99,104,114,101,108,101,97,115,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,116,111,117,99,104,114,101,108,101,97,115,101,100,40,105,100,44,120,44,121,44,100,120,44,100,121,44,112,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,116,111,117,99,104,109,111,118,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,105,100,44,120,44,121,44,100,120,44,100,121,44,112,41,10,9,9,9,105,102,32,108,111,118,101,46,116,111,117,99,104,109,111,118,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,116,111,117,99,104,109,111,118,101,100,40,105,100,44,120,44,121,44,100,120,44,100,121,44,112,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,106,111,121,115,116,105,99,107,112,114,101,115,115,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,106,44,98,41,10,9,9,9,105,102,32,108,111,118,101,46,106,111,121,115,116,105,99,107,112,114,101,115,115,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,106,111,121,115,116,105,99,107,112,114,101,115,115,101,100,40,106,44,98,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,106,111,121,115,116,105,99,107,114,101,108,101,97,115,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,106,44,98,41,10,9,9,9,105,102,32,108,111,118,101,46,106,111,121,115,116,105,99,107,114,101,108,101,97,115,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,106,111,121,115,116,105,99,107,114,101,108,101,97,115,101,100,40,106,44,98,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,106,111,121,115,116,105,99,107,97,120,105,115,32,61,32,102,117,110,99,116,105,111,110,32,40,106,44,97,44,118,41,10,9,9,9,105,102,32,108,111,118,101,46,106,111,121,115,116,105,99,107,97,120,105,115,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,106,111,121,115,116,105,99,107,97,120,105,115,40,106,44,97,44,118,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,106,111,121,115,116,105,99,107,104,97,116,32,61,32,102,117,110,99,116,105,111,110,32,40,106,44,104,44,118,41,10,9,9,9,105,102,32,108,111,118,101,46,106,111,121,115,116,105,99,107,104,97,116,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,106,111,121,115,116,105,99,107,104,97,116,40,106,44,104,44,118,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,103,97,109,101,112,97,100,112,114,101,115,115,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,106,44,98,41,10,9,9,9,105,102,32,108,111,118,101,46,103,97,109,101,112,97,100,112,114,101,115,115,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,103,97,109,101,112,97,100,112,114,101,115,115,101,100,40,106,44,98,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,103,97,109,101,112,97,100,114,101,108,101,97,115,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,106,44,98,41,10,9,9,9,105,102,32,108,111,118,101,46,103,97,109,101,112,97,100,114,101,108,101,97,115,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,103,97,109,101,112,97,100,114,101,108,101,97,115,101,100,40,106,44,98,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,103,97,109,101,112,97,100,97,120,105,115,32,61,32,102,117,110,99,116,105,111,110,32,40,106,44,97,44,118,41,10,9,9,9,105,102,32,108,111,118,101,46,103,97,109,101,112,97,100,97,120,105,115,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,103,97,109,101,112,97,100,97,120,105,115,40,106,44,97,44,118,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,106,111,121,115,116,105,99,107,97,100,100,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,106,41,10,9,9,9,105,102,32,108,111,118,101,46,106,111,121,115,116,105,99,107,97,100,100,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,106,111,121,115,116,105,99,107,97,100,100,101,100,40,106,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,106,111,121,115,116,105,99,107,114,101,109,111,118,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,106,41,10,9,9,9,105,102,32,108,111,118,101,46,106,111,121,115,116,105,99,107,114,101,109,111,118,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,106,111,121,115,116,105,99,107,114,101,109,111,118,101,100,40,106,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,102,111,99,117,115,32,61,32,102,117,110,99,116,105,111,110,32,40,102,41,10,9,9,9,105,102,32,108,111,118,101,46,102,111,99,117,115,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,102,111,99,117,115,40,102,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,109,111,117,115,101,102,111,99,117,115,32,61,32,102,117,110,99,116,105,111,110,32,40,102,41,10,9,9,9,105,102,32,108,111,118,101,46,109,111,117,115,101,102,111,99,117,115,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,109,111,117,115,101,102,111,99,117,115,40,102,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,118,105,115,105,98,108,101,32,61,32,102,117,110,99,116,105,111,110,32,40,118,41,10,9,9,9,105,102,32,108,111,118,101,46,118,105,115,105,98,108,101,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,118,105,115,105,98,108,101,40,118,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,113,117,105,116,32,61,32,102,117,110,99,116,105,111,110,32,40,41,10,9,9,9,114,101,116,117,114,110,10,9,9,101,110,100,44,10,9,9,116,104,114,101,97,100,101,114,114,111,114,32,61,32,102,117,110,99,116,105,111,110,32,40,116,44,32,101,114,114,41,10,9,9,9,105,102,32,108,111,118,101,46,116,104,114,101,97,100,101,114,114,111,114,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,116,104,114,101,97,100,101,114,114,111,114,40,116,44,32,101,114,114,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,114,101,115,105,122,101,32,61,32,102,117,110,99,116,105,111,110,32,40,119,44,32,104,41,10,9,9,9,105,102,32,108,111,118,101,46,114,101,115,105,122,101,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,114,101,115,105,122,101,40,119,44,32,104,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,102,105,108,101,100,114,111,112,112,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,102,41,10,9,9,9,105,102,32,108,111,118,101,46,102,105,108,101,100,114,111,112,112,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,102,105,108,101,100,114,111,112,112,101,100,40,102,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,100,105,114,101,99,116,111,114,121,100,114,111,112,112,101,100,32,61,32,102,117,110,99,116,105,111,110,32,40,100,105,114,41,10,9,9,9,105,102,32,108,111,118,101,46,100,105,114,101,99,116,111,114,121,100,114,111,112,112,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,100,105,114,101,99,116,111,114,121,100,114,111,112,112,101,100,40,100,105,114,41,32,101,110,100,10,9,9,101,110,100,44,10,9,9,108,111,119,109,101,109,111,114,121,32,61,32,102,117,110,99,116,105,111,110,32,40,41,10,9,9,9,99,111,108,108,101,99,116,103,97,114,98,97,103,101,40,41,10,9,9,9,105,102,32,108,111,118,101,46,108,111,119,109,101,109,111,114,121,32,116,104,101,110,32,114,101,116,117,114,110,32,108,111,118,101,46,108,111,119,109,101,109,111,114,121,40,41,32,101,110,100,10,9,9,101,110,100,44,10,9,125,44,32,123,10,9,9,95,95,105,110,100,101,120,32,61,32,102,117,110,99,116,105,111,110,40,115,101,108,102,44,32,110,97,109,101,41,10,9,9,9,101,114,114,111,114,40,34,85,110,107,110,111,119,110,32,101,118,101,110,116,58,32,34,32,46,46,32,110,97,109,101,41,10,9,9,101,110,100,44,10,9,125,41,10,101,110,100,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,117,114,105,100,101,99,111,100,101,40,115,41,10,9,114,101,116,117,114,110,32,115,58,103,115,117,98,40,34,37,37,37,120,37,120,34,44,32,102,117,110,99,116,105,111,110,40,115,116,114,41,10,9,9,114,101,116,117,114,110,32,115,116,114,105,110,103,46,99,104,97,114,40,116,111,110,117,109,98,101,114,40,115,116,114,58,115,117,98,40,50,41,44,32,49,54,41,41,10,9,101,110,100,41,10,101,110,100,10,108,111,99,97,108,32,110,111,95,103,97,109,101,95,99,111,100,101,32,61,32,102,97,108,115,101,10,45,45,32,84,104,105,115,32,99,97,110,39,116,32,98,101,32,111,118,101,114,114,105,100,101,110,46,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,98,111,111,116,40,41,10,9,45,45,32,84,104,105,115,32,105,115,32,97,98,115,111,108,117,116,101,108,121,32,110,101,101,100,101,100,46,10,9,114,101,113,117,105,114,101,40,34,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,34,41,10,9,108,111,118,101,46,97,114,103,46,112,97,114,115,101,95,111,112,116,105,111,110,115,40,41,10,9,108,111,99,97,108,32,111,32,61,32,108,111,118,101,46,97,114,103,46,111,112,116,105,111,110,115,10,9,108,111,99,97,108,32,97,114,103,48,32,61,32,108,111,118,101,46,97,114,103,46,103,101,116,76,111,119,40,97,114,103,41,10,9,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,46,105,110,105,116,40,97,114,103,48,41,10,9,108,111,99,97,108,32,101,120,101,112,97,116,104,32,61,32,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,46,103,101,116,69,120,101,99,117,116,97,98,108,101,80,97,116,104,40,41,10,9,105,102,32,35,101,120,101,112,97,116,104,32,61,61,32,48,32,116,104,101,110,10,9,9,45,45,32,84,104,105,115,32,115,104,111,117,108,100,110,39,116,32,104,97,112,112,101,110,44,32,98,117,116,32,106,117,115,116,32,105,110,32,99,97,115,101,32,119,101,39,108,108,32,102,97,108,108,32,98,97,99,107,32,116,111,32,97,114,103,48,46,10,9,9,101,120,101,112,97,116,104,32,61,32,97,114,103,48,10,9,101,110,100,10,9,45,45,32,73,115,32,116,104,105,115,32,111,110,101,32,111,102,32,116,104,111,115,101,32,102,97,110,99,121,32,34,102,117,115,101,100,34,32,103,97,109,101,115,63,10,9,108,111,99,97,108,32,99,97,110,95,104,97,115,95,103,97,109,101,32,61,32,112,99,97,108,108,40,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,46,115,101,116,83,111,117,114,99,101,44,32,101,120,101,112,97,116,104,41,10,9,108,111,99,97,108,32,105,115,95,102,117,115,101,100,95,103,97,109,101,32,61,32,99,97,110,95,104,97,115,95,103,97,109,101,32,111,114,32,108,111,118,101,46,97,114,103,46,111,112,116,105,111,110,115,46,102,117,115,101,100,46,115,101,116,10,9,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,46,115,101,116,70,117,115,101,100,40,105,115,95,102,117,115,101,100,95,103,97,109,101,41,10,9,108,111,99,97,108,32,105,100,101,110,116,105,116,121,32,61,32,34,34,10,9,105,102,32,110,111,116,32,99,97,110,95,104,97,115,95,103,97,109,101,32,97,110,100,32,111,46,103,97,109,101,46,115,101,116,32,97,110,100,32,111,46,103,97,109,101,46,97,114,103,91,49,93,32,116,104,101,110,10,9,9,108,111,99,97,108,32,110,111,117,114,105,32,61,32,111,46,103,97,109,101,46,97,114,103,91,49,93,10,9,9,105,102,32,110,111,117,114,105,58,115,117,98,40,49,44,32,55,41,32,61,61,32,34,102,105,108,101,58,47,47,34,32,116,104,101,110,10,9,9,9,110,111,117,114,105,32,61,32,117,114,105,100,101,99,111,100,101,40,110,111,117,114,105,58,115,117,98,40,56,41,41,10,9,9,101,110,100,10,9,9,108,111,99,97,108,32,102,117,108,108,95,115,111,117,114,99,101,32,61,32,32,108,111,118,101,46,112,97,116,104,46,103,101,116,102,117,108,108,40,110,111,117,114,105,41,10,9,9,99,97,110,95,104,97,115,95,103,97,109,101,32,61,32,112,99,97,108,108,40,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,46,115,101,116,83,111,117,114,99,101,44,32,102,117,108,108,95,115,111,117,114,99,101,41,10,9,9,45,45,32,85,115,101,32,116,104,101,32,110,97,109,101,32,111,102,32,116,104,101,32,115,111,117,114,99,101,32,46,108,111,118,101,32,97,115,32,116,104,101,32,105,100,101,110,116,105,116,121,32,102,111,114,32,110,111,119,46,10,9,9,105,100,101,110,116,105,116,121,32,61,32,108,111,118,101,46,112,97,116,104,46,108,101,97,102,40,102,117,108,108,95,115,111,117,114,99,101,41,10,9,101,108,115,101,10,9,9,45,45,32,85,115,101,32,116,104,101,32,110,97,109,101,32,111,102,32,116,104,101,32,101,120,101,32,97,115,32,116,104,101,32,105,100,101,110,116,105,116,121,32,102,111,114,32,110,111,119,46,10,9,9,105,100,101,110,116,105,116,121,32,61,32,108,111,118,101,46,112,97,116,104,46,108,101,97,102,40,101,120,101,112,97,116,104,41,10,9,101,110,100,10,9,45,45,32,84,114,121,32,116,111,32,117,115,101,32,116,104,101,32,97,114,99,104,105,118,101,32,99,111,110,116,97,105,110,105,110,103,32,109,97,105,110,46,108,117,97,32,97,115,32,116,104,101,32,105,100,101,110,116,105,116,121,32,110,97,109,101,46,32,73,116,10,9,45,45,32,109,105,103,104,116,32,110,111,116,32,98,101,32,97,118,97,105,108,97,98,108,101,44,32,105,110,32,119,104,105,99,104,32,99,97,115,101,32,116,104,101,32,102,97,108,108,98,97,99,107,115,32,97,98,111,118,101,32,97,114,101,32,117,115,101,100,46,10,9,108,111,99,97,108,32,114,101,97,108,100,105,114,32,61,32,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,46,103,101,116,82,101,97,108,68,105,114,101,99,116,111,114,121,40,34,109,97,105,110,46,108,117,97,34,41,10,9,105,102,32,114,101,97,108,100,105,114,32,116,104,101,110,10,9,9,105,100,101,110,116,105,116,121,32,61,32,108,111,118,101,46,112,97,116,104,46,108,101,97,102,40,114,101,97,108,100,105,114,41,10,9,101,110,100,10,9,105,100,101,110,116,105,116,121,32,61,32,105,100,101,110,116,105,116,121,58,103,115,117,98,40,34,94,40,91,37,46,93,43,41,34,44,32,34,34,41,32,45,45,32,115,116,114,105,112,32,108,101,97,100,105,110,103,32,34,46,34,39,115,10,9,105,100,101,110,116,105,116,121,32,61,32,105,100,101,110,116,105,116,121,58,103,115,117,98,40,34,37,46,40,91,94,37,46,93,43,41,36,34,44,32,34,34,41,32,45,45,32,115,116,114,105,112,32,101,120,116,101,110,115,105,111,110,10,9,105,100,101,110,116,105,116,121,32,61,32,105,100,101,110,116,105,116,121,58,103,115,117,98,40,34,37,46,34,44,32,34,95,34,41,32,45,45,32,114,101,112,108,97,99,101,32,114,101,109,97,105,110,105,110,103,32,34,46,34,39,115,32,119,105,116,104,32,34,95,34,10,9,105,100,101,110,116,105,116,121,32,61,32,35,105,100,101,110,116,105,116,121,32,62,32,48,32,97,110,100,32,105,100,101,110,116,105,116,121,32,111,114,32,34,108,111,118,101,103,97,109,101,34,10,9,45,45,32,87,104,101,110,32,99,111,110,102,46,108,117,97,32,105,115,32,105,110,105,116,105,97,108,108,121,32,108,111,97,100,101,100,44,32,116,104,101,32,109,97,105,110,32,115,111,117,114,99,101,32,115,104,111,117,108,100,32,98,101,32,99,104,101,99,107,101,100,10,9,45,45,32,98,101,102,111,114,101,32,116,104,101,32,115,97,118,101,32,100,105,114,101,99,116,111,114,121,32,40,116,104,101,32,105,100,101,110,116,105,116,121,32,115,104,111,117,108,100,32,98,101,32,97,112,112,101,110,100,101,100,46,41,10,9,112,99,97,108,108,40,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,46,115,101,116,73,100,101,110,116,105,116,121,44,32,105,100,101,110,116,105,116,121,44,32,116,114,117,101,41,10,9,105,102,32,99,97,110,95,104,97,115,95,103,97,109,101,32,97,110,100,32,110,111,116,32,40,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,46,105,115,70,105,108,101,40,34,109,97,105,110,46,108,117,97,34,41,32,111,114,32,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,46,105,115,70,105,108,101,40,34,99,111,110,102,46,108,117,97,34,41,41,32,116,104,101,110,10,9,9,110,111,95,103,97,109,101,95,99,111,100,101,32,61,32,116,114,117,101,10,9,101,110,100,10,9,105,102,32,110,111,116,32,99,97,110,95,104,97,115,95,103,97,109,101,32,116,104,101,110,10,9,9,108,111,99,97,108,32,110,111,103,97,109,101,32,61,32,114,101,113,117,105,114,101,40,34,108,111,118,101,46,110,111,103,97,109,101,34,41,10,9,9,110,111,103,97,109,101,40,41,10,9,101,110,100,10,101,110,100,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,105,110,105,116,40,41,10,9,45,45,32,67,114,101,97,116,101,32,100,101,102,97,117,108,116,32,99,111,110,102,105,103,117,114,97,116,105,111,110,32,115,101,116,116,105,110,103,115,46,10,9,45,45,32,78,79,84,69,58,32,65,100,100,105,110,103,32,97,32,110,101,119,32,109,111,100,117,108,101,32,116,111,32,116,104,101,32,109,111,100,117,108,101,115,32,108,105,115,116,10,9,45,45,32,119,105,108,108,32,78,79,84,32,109,97,107,101,32,105,116,32,108,111,97,100,44,32,115,101,101,32,98,101,108,111,119,46,10,9,108,111,99,97,108,32,99,32,61,32,123,10,9,9,116,105,116,108,101,32,61,32,34,85,110,116,105,116,108,101,100,34,44,10,9,9,118,101,114,115,105,111,110,32,61,32,108,111,118,101,46,95,118,101,114,115,105,111,110,44,10,9,9,119,105,110,100,111,119,32,61,32,123,10,9,9,9,119,105,100,116,104,32,61,32,56,48,48,44,10,9,9,9,104,101,105,103,104,116,32,61,32,54,48,48,44,10,9,9,9,120,32,61,32,110,105,108,44,10,9,9,9,121,32,61,32,110,105,108,44,10,9,9,9,109,105,110,119,105,100,116,104,32,61,32,49,44,10,9,9,9,109,105,110,104,101,105,103,104,116,32,61,32,49,44,10,9,9,9,102,117,108,108,115,99,114,101,101,110,32,61,32,102,97,108,115,101,44,10,9,9,9,102,117,108,108,115,99,114,101,101,110,116,121,112,101,32,61,32,34,100,101,115,107,116,111,112,34,44,10,9,9,9,100,105,115,112,108,97,121,32,61,32,49,44,10,9,9,9,118,115,121,110,99,32,61,32,116,114,117,101,44,10,9,9,9,109,115,97,97,32,61,32,48,44,10,9,9,9,98,111,114,100,101,114,108,101,115,115,32,61,32,102,97,108,115,101,44,10,9,9,9,114,101,115,105,122,97,98,108,101,32,61,32,102,97,108,115,101,44,10,9,9,9,99,101,110,116,101,114,101,100,32,61,32,116,114,117,101,44,10,9,9,9,104,105,103,104,100,112,105,32,61,32,102,97,108,115,101,44,10,9,9,125,44,10,9,9,109,111,100,117,108,101,115,32,61,32,123,10,9,9,9,101,118,101,110,116,32,61,32,116,114,117,101,44,10,9,9,9,107,101,121,98,111,97,114,100,32,61,32,116,114,117,101,44,10,9,9,9,109,111,117,115,101,32,61,32,116,114,117,101,44,10,9,9,9,116,105,109,101,114,32,61,32,116,114,117,101,44,10,9,9,9,106,111,121,115,116,105,99,107,32,61,32,116,114,117,101,44,10,9,9,9,116,111,117,99,104,32,61,32,116,114,117,101,44,10,9,9,9,105,109,97,103,101,32,61,32,116,114,117,101,44,10,9,9,9,103,114,97,112,104,105,99,115,32,61,32,116,114,117,101,44,10,9,9,9,97,117,100,105,111,32,61,32,116,114,117,101,44,10,9,9,9,109,97,116,104,32,61,32,116,114,117,101,44,10,9,9,9,112,104,121,115,105,99,115,32,61,32,116,114,117,101,44,10,9,9,9,115,111,117,110,100,32,61,32,116,114,117,101,44,10,9,9,9,115,121,115,116,101,109,32,61,32,116,114,117,101,44,10,9,9,9,102,111,110,116,32,61,32,116,114,117,101,44,10,9,9,9,116,104,114,101,97,100,32,61,32,116,114,117,101,44,10,9,9,9,119,105,110,100,111,119,32,61,32,116,114,117,101,44,10,9,9,9,118,105,100,101,111,32,61,32,116,114,117,101,44,10,9,9,125,44,10,9,9,99,111,110,115,111,108,101,32,61,32,102,97,108,115,101,44,32,45,45,32,79,110,108,121,32,114,101,108,101,118,97,110,116,32,102,111,114,32,119,105,110,100,111,119,115,46,10,9,9,105,100,101,110,116,105,116,121,32,61,32,102,97,108,115,101,44,10,9,9,97,112,112,101,110,100,105,100,101,110,116,105,116,121,32,61,32,102,97,108,115,101,44,10,9,9,101,120,116,101,114,110,97,108,115,116,111,114,97,103,101,32,61,32,102,97,108,115,101,44,32,45,45,32,79,110,108,121,32,114,101,108,101,118,97,110,116,32,102,111,114,32,65,110,100,114,111,105,100,46,10,9,9,97,99,99,101,108,101,114,111,109,101,116,101,114,106,111,121,115,116,105,99,107,32,61,32,116,114,117,101,44,32,45,45,32,79,110,108,121,32,114,101,108,101,118,97,110,116,32,102,111,114,32,65,110,100,114,111,105,100,32,47,32,105,79,83,46,10,9,9,103,97,109,109,97,99,111,114,114,101,99,116,32,61,32,102,97,108,115,101,44,10,9,125,10,9,45,45,32,67,111,110,115,111,108,101,32,104,97,99,107,44,32,112,97,114,116,32,49,46,10,9,108,111,99,97,108,32,111,112,101,110,101,100,99,111,110,115,111,108,101,32,61,32,102,97,108,115,101,10,9,105,102,32,108,111,118,101,46,97,114,103,46,111,112,116,105,111,110,115,46,99,111,110,115,111,108,101,46,115,101,116,32,97,110,100,32,108,111,118,101,46,95,111,112,101,110,67,111,110,115,111,108,101,32,116,104,101,110,10,9,9,108,111,118,101,46,95,111,112,101,110,67,111,110,115,111,108,101,40,41,10,9,9,111,112,101,110,101,100,99,111,110,115,111,108,101,32,61,32,116,114,117,101,10,9,101,110,100,10,9,45,45,32,73,102,32,99,111,110,102,105,103,32,102,105,108,101,32,101,120,105,115,116,115,44,32,108,111,97,100,32,105,116,32,97,110,100,32,97,108,108,111,119,32,105,116,32,116,111,32,117,112,100,97,116,101,32,99,111,110,102,105,103,32,116,97,98,108,101,46,10,9,108,111,99,97,108,32,99,111,110,102,111,107,44,32,99,111,110,102,101,114,114,10,9,105,102,32,40,110,111,116,32,108,111,118,101,46,99,111,110,102,41,32,97,110,100,32,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,32,97,110,100,32,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,46,105,115,70,105,108,101,40,34,99,111,110,102,46,108,117,97,34,41,32,116,104,101,110,10,9,9,99,111,110,102,111,107,44,32,99,111,110,102,101,114,114,32,61,32,112,99,97,108,108,40,114,101,113,117,105,114,101,44,32,34,99,111,110,102,34,41,10,9,101,110,100,10,9,45,45,32,89,101,115,44,32,99,111,110,102,46,108,117,97,32,109,105,103,104,116,32,110,111,116,32,101,120,105,115,116,44,32,98,117,116,32,116,104,101,114,101,32,97,114,101,32,111,116,104,101,114,32,119,97,121,115,32,111,102,32,109,97,107,105,110,103,10,9,45,45,32,108,111,118,101,46,99,111,110,102,32,97,112,112,101,97,114,44,32,115,111,32,119,101,32,115,104,111,117,108,100,32,99,104,101,99,107,32,102,111,114,32,105,116,32,97,110,121,119,97,121,46,10,9,105,102,32,108,111,118,101,46,99,111,110,102,32,116,104,101,110,10,9,9,99,111,110,102,111,107,44,32,99,111,110,102,101,114,114,32,61,32,112,99,97,108,108,40,108,111,118,101,46,99,111,110,102,44,32,99,41,10,9,9,45,45,32,73,102,32,108,111,118,101,46,99,111,110,102,32,101,114,114,111,114,115,44,32,119,101,39,108,108,32,116,114,105,103,103,101,114,32,116,104,101,32,101,114,114,111,114,32,97,102,116,101,114,32,108,111,97,100,105,110,103,32,109,111,100,117,108,101,115,32,115,111,10,9,9,45,45,32,116,104,101,32,101,114,114,111,114,32,109,101,115,115,97,103,101,32,99,97,110,32,98,101,32,100,105,115,112,108,97,121,101,100,32,105,110,32,116,104,101,32,119,105,110,100,111,119,46,10,9,101,110,100,10,9,45,45,32,67,111,110,115,111,108,101,32,104,97,99,107,44,32,112,97,114,116,32,50,46,10,9,105,102,32,99,46,99,111,110,115,111,108,101,32,97,110,100,32,108,111,118,101,46,95,111,112,101,110,67,111,110,115,111,108,101,32,97,110,100,32,110,111,116,32,111,112,101,110,101,100,99,111,110,115,111,108,101,32,116,104,101,110,10,9,9,108,111,118,101,46,95,111,112,101,110,67,111,110,115,111,108,101,40,41,10,9,101,110,100,10,9,45,45,32,72,97,99,107,32,102,111,114,32,100,105,115,97,98,108,105,110,103,32,97,99,99,101,108,101,114,111,109,101,116,101,114,45,97,115,45,106,111,121,115,116,105,99,107,32,111,110,32,65,110,100,114,111,105,100,32,47,32,105,79,83,46,10,9,105,102,32,108,111,118,101,46,95,115,101,116,65,99,99,101,108,101,114,111,109,101,116,101,114,65,115,74,111,121,115,116,105,99,107,32,116,104,101,110,10,9,9,108,111,118,101,46,95,115,101,116,65,99,99,101,108,101,114,111,109,101,116,101,114,65,115,74,111,121,115,116,105,99,107,40,99,46,97,99,99,101,108,101,114,111,109,101,116,101,114,106,111,121,115,116,105,99,107,41,10,9,101,110,100,10,9,105,102,32,108,111,118,101,46,95,115,101,116,71,97,109,109,97,67,111,114,114,101,99,116,32,116,104,101,110,10,9,9,108,111,118,101,46,95,115,101,116,71,97,109,109,97,67,111,114,114,101,99,116,40,99,46,103,97,109,109,97,99,111,114,114,101,99,116,41,10,9,101,110,100,10,9,45,45,32,71,101,116,115,32,100,101,115,105,114,101,100,32,109,111,100,117,108,101,115,46,10,9,102,111,114,32,107,44,118,32,105,110,32,105,112,97,105,114,115,123,10,9,9,34,116,104,114,101,97,100,34,44,10,9,9,34,116,105,109,101,114,34,44,10,9,9,34,101,118,101,110,116,34,44,10,9,9,34,107,101,121,98,111,97,114,100,34,44,10,9,9,34,106,111,121,115,116,105,99,107,34,44,10,9,9,34,109,111,117,115,101,34,44,10,9,9,34,116,111,117,99,104,34,44,10,9,9,34,115,111,117,110,100,34,44,10,9,9,34,115,121,115,116,101,109,34,44,10,9,9,34,97,117,100,105,111,34,44,10,9,9,34,105,109,97,103,101,34,44,10,9,9,34,118,105,100,101,111,34,44,10,9,9,34,102,111,110,116,34,44,10,9,9,34,119,105,110,100,111,119,34,44,10,9,9,34,103,114,97,112,104,105,99,115,34,44,10,9,9,34,109,97,116,104,34,44,10,9,9,34,112,104,121,115,105,99,115,34,44,10,9,125,32,100,111,10,9,9,105,102,32,99,46,109,111,100,117,108,101,115,91,118,93,32,116,104,101,110,10,9,9,9,114,101,113,117,105,114,101,40,34,108,111,118,101,46,34,32,46,46,32,118,41,10,9,9,101,110,100,10,9,101,110,100,10,9,105,102,32,108,111,118,101,46,101,118,101,110,116,32,116,104,101,110,10,9,9,108,111,118,101,46,99,114,101,97,116,101,104,97,110,100,108,101,114,115,40,41,10,9,101,110,100,10,9,45,45,32,67,104,101,99,107,32,116,104,101,32,118,101,114,115,105,111,110,10,9,99,46,118,101,114,115,105,111,110,32,61,32,116,111,115,116,114,105,110,103,40,99,46,118,101,114,115,105,111,110,41,10,9,105,102,32,110,111,116,32,108,111,118,101,46,105,115,86,101,114,115,105,111,110,67,111,109,112,97,116,105,98,108,101,40,99,46,118,101,114,115,105,111,110,41,32,116,104,101,110,10,9,9,108,111,99,97,108,32,109,97,106,111,114,44,32,109,105,110,111,114,44,32,114,101,118,105,115,105,111,110,32,61,32,99,46,118,101,114,115,105,111,110,58,109,97,116,99,104,40,34,94,40,37,100,43,41,37,46,40,37,100,43,41,37,46,40,37,100,43,41,36,34,41,10,9,9,105,102,32,40,110,111,116,32,109,97,106,111,114,32,111,114,32,110,111,116,32,109,105,110,111,114,32,111,114,32,110,111,116,32,114,101,118,105,115,105,111,110,41,32,111,114,32,40,109,97,106,111,114,32,126,61,32,108,111,118,101,46,95,118,101,114,115,105,111,110,95,109,97,106,111,114,32,97,110,100,32,109,105,110,111,114,32,126,61,32,108,111,118,101,46,95,118,101,114,115,105,111,110,95,109,105,110,111,114,41,32,116,104,101,110,10,9,9,9,108,111,99,97,108,32,109,115,103,32,61,32,40,34,84,104,105,115,32,103,97,109,101,32,105,110,100,105,99,97,116,101,115,32,105,116,32,119,97,115,32,109,97,100,101,32,102,111,114,32,118,101,114,115,105,111,110,32,39,37,115,39,32,111,102,32,76,79,86,69,46,92,110,34,46,46,10,9,9,9,9,34,73,116,32,109,97,121,32,110,111,116,32,98,101,32,99,111,109,112,97,116,105,98,108,101,32,119,105,116,104,32,116,104,101,32,114,117,110,110,105,110,103,32,118,101,114,115,105,111,110,32,40,37,115,41,46,34,41,58,102,111,114,109,97,116,40,99,46,118,101,114,115,105,111,110,44,32,108,111,118,101,46,95,118,101,114,115,105,111,110,41,10,9,9,9,112,114,105,110,116,40,109,115,103,41,10,9,9,9,105,102,32,108,111,118,101,46,119,105,110,100,111,119,32,116,104,101,110,10,9,9,9,9,108,111,118,101,46,119,105,110,100,111,119,46,115,104,111,119,77,101,115,115,97,103,101,66,111,120,40,34,67,111,109,112,97,116,105,98,105,108,105,116,121,32,87,97,114,110,105,110,103,34,44,32,109,115,103,44,32,34,119,97,114,110,105,110,103,34,41,10,9,9,9,101,110,100,10,9,9,101,110,100,10,9,101,110,100,10,9,105,102,32,110,111,116,32,99,111,110,102,111,107,32,97,110,100,32,99,111,110,102,101,114,114,32,116,104,101,110,10,9,9,101,114,114,111,114,40,99,111,110,102,101,114,114,41,10,9,101,110,100,10,9,45,45,32,83,101,116,117,112,32,119,105,110,100,111,119,32,104,101,114,101,46,10,9,105,102,32,99,46,119,105,110,100,111,119,32,97,110,100,32,99,46,109,111,100,117,108,101,115,46,119,105,110,100,111,119,32,116,104,101,110,10,9,9,97,115,115,101,114,116,40,108,111,118,101,46,119,105,110,100,111,119,46,115,101,116,77,111,100,101,40,99,46,119,105,110,100,111,119,46,119,105,100,116,104,44,32,99,46,119,105,110,100,111,119,46,104,101,105,103,104,116,44,10,9,9,123,10,9,9,9,102,117,108,108,115,99,114,101,101,110,32,61,32,99,46,119,105,110,100,111,119,46,102,117,108,108,115,99,114,101,101,110,44,10,9,9,9,102,117,108,108,115,99,114,101,101,110,116,121,112,101,32,61,32,99,46,119,105,110,100,111,119,46,102,117,108,108,115,99,114,101,101,110,116,121,112,101,44,10,9,9,9,118,115,121,110,99,32,61,32,99,46,119,105,110,100,111,119,46,118,115,121,110,99,44,10,9,9,9,109,115,97,97,32,61,32,99,46,119,105,110,100,111,119,46,109,115,97,97,44,10,9,9,9,114,101,115,105,122,97,98,108,101,32,61,32,99,46,119,105,110,100,111,119,46,114,101,115,105,122,97,98,108,101,44,10,9,9,9,109,105,110,119,105,100,116,104,32,61,32,99,46,119,105,110,100,111,119,46,109,105,110,119,105,100,116,104,44,10,9,9,9,109,105,110,104,101,105,103,104,116,32,61,32,99,46,119,105,110,100,111,119,46,109,105,110,104,101,105,103,104,116,44,10,9,9,9,98,111,114,100,101,114,108,101,115,115,32,61,32,99,46,119,105,110,100,111,119,46,98,111,114,100,101,114,108,101,115,115,44,10,9,9,9,99,101,110,116,101,114,101,100,32,61,32,99,46,119,105,110,100,111,119,46,99,101,110,116,101,114,101,100,44,10,9,9,9,100,105,115,112,108,97,121,32,61,32,99,46,119,105,110,100,111,119,46,100,105,115,112,108,97,121,44,10,9,9,9,104,105,103,104,100,112,105,32,61,32,99,46,119,105,110,100,111,119,46,104,105,103,104,100,112,105,44,10,9,9,9,120,32,61,32,99,46,119,105,110,100,111,119,46,120,44,10,9,9,9,121,32,61,32,99,46,119,105,110,100,111,119,46,121,44,10,9,9,125,41,44,32,34,67,111,117,108,100,32,110,111,116,32,115,101,116,32,119,105,110,100,111,119,32,109,111,100,101,34,41,10,9,9,108,111,118,101,46,119,105,110,100,111,119,46,115,101,116,84,105,116,108,101,40,99,46,119,105,110,100,111,119,46,116,105,116,108,101,32,111,114,32,99,46,116,105,116,108,101,41,10,9,9,105,102,32,99,46,119,105,110,100,111,119,46,105,99,111,110,32,116,104,101,110,10,9,9,9,97,115,115,101,114,116,40,108,111,118,101,46,105,109,97,103,101,44,32,34,73,102,32,97,110,32,105,99,111,110,32,105,115,32,115,101,116,32,105,110,32,108,111,118,101,46,99,111,110,102,44,32,108,111,118,101,46,105,109,97,103,101,32,109,117,115,116,32,98,101,32,108,111,97,100,101,100,33,34,41,10,9,9,9,108,111,118,101,46,119,105,110,100,111,119,46,115,101,116,73,99,111,110,40,108,111,118,101,46,105,109,97,103,101,46,110,101,119,73,109,97,103,101,68,97,116,97,40,99,46,119,105,110,100,111,119,46,105,99,111,110,41,41,10,9,9,101,110,100,10,9,101,110,100,10,9,45,45,32,79,117,114,32,102,105,114,115,116,32,116,105,109,101,115,116,101,112,44,32,98,101,99,97,117,115,101,32,119,105,110,100,111,119,32,99,114,101,97,116,105,111,110,32,99,97,110,32,116,97,107,101,32,115,111,109,101,32,116,105,109,101,10,9,105,102,32,108,111,118,101,46,116,105,109,101,114,32,116,104,101,110,10,9,9,108,111,118,101,46,116,105,109,101,114,46,115,116,101,112,40,41,10,9,101,110,100,10,9,105,102,32,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,32,116,104,101,110,10,9,9,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,46,95,115,101,116,65,110,100,114,111,105,100,83,97,118,101,69,120,116,101,114,110,97,108,40,99,46,101,120,116,101,114,110,97,108,115,116,111,114,97,103,101,41,10,9,9,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,46,115,101,116,73,100,101,110,116,105,116,121,40,99,46,105,100,101,110,116,105,116,121,32,111,114,32,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,46,103,101,116,73,100,101,110,116,105,116,121,40,41,44,32,99,46,97,112,112,101,110,100,105,100,101,110,116,105,116,121,41,10,9,9,105,102,32,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,46,105,115,70,105,108,101,40,34,109,97,105,110,46,108,117,97,34,41,32,116,104,101,110,10,9,9,9,114,101,113,117,105,114,101,40,34,109,97,105,110,34,41,10,9,9,101,110,100,10,9,101,110,100,10,9,105,102,32,110,111,95,103,97,109,101,95,99,111,100,101,32,116,104,101,110,10,9,9,101,114,114,111,114,40,34,78,111,32,99,111,100,101,32,116,111,32,114,117,110,92,110,89,111,117,114,32,103,97,109,101,32,109,105,103,104,116,32,98,101,32,112,97,99,107,97,103,101,100,32,105,110,99,111,114,114,101,99,116,108,121,92,110,77,97,107,101,32,115,117,114,101,32,109,97,105,110,46,108,117,97,32,105,115,32,97,116,32,116,104,101,32,116,111,112,32,108,101,118,101,108,32,111,102,32,116,104,101,32,122,105,112,34,41,10,9,101,110,100,10,101,110,100,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,115,116,101,112,40,41,10,9,108,111,99,97,108,32,100,116,32,61,32,48,10,9,45,45,32,80,114,111,99,101,115,115,32,101,118,101,110,116,115,46,10,9,105,102,32,108,111,118,101,46,101,118,101,110,116,32,116,104,101,110,10,9,9,108,111,118,101,46,101,118,101,110,116,46,112,117,109,112,40,41,10,9,9,102,111,114,32,110,97,109,101,44,32,97,44,98,44,99,44,100,44,101,44,102,32,105,110,32,108,111,118,101,46,101,118,101,110,116,46,112,111,108,108,40,41,32,100,111,10,9,9,9,105,102,32,110,97,109,101,32,61,61,32,34,113,117,105,116,34,32,116,104,101,110,10,9,9,9,9,105,102,32,110,111,116,32,108,111,118,101,46,113,117,105,116,32,111,114,32,110,111,116,32,108,111,118,101,46,113,117,105,116,40,41,32,116,104,101,110,10,9,9,9,9,9,114,101,116,117,114,110,32,97,10,9,9,9,9,101,110,100,10,9,9,9,101,110,100,10,9,9,9,108,111,118,101,46,104,97,110,100,108,101,114,115,91,110,97,109,101,93,40,97,44,98,44,99,44,100,44,101,44,102,41,10,9,9,101,110,100,10,9,101,110,100,10,9,45,45,32,85,112,100,97,116,101,32,100,116,44,32,97,115,32,119,101,39,108,108,32,98,101,32,112,97,115,115,105,110,103,32,105,116,32,116,111,32,117,112,100,97,116,101,10,9,105,102,32,108,111,118,101,46,116,105,109,101,114,32,116,104,101,110,10,9,9,108,111,118,101,46,116,105,109,101,114,46,115,116,101,112,40,41,10,9,9,100,116,32,61,32,108,111,118,101,46,116,105,109,101,114,46,103,101,116,68,101,108,116,97,40,41,10,9,101,110,100,10,9,45,45,32,67,97,108,108,32,117,112,100,97,116,101,32,97,110,100,32,100,114,97,119,10,9,105,102,32,108,111,118,101,46,117,112,100,97,116,101,32,116,104,101,110,32,108,111,118,101,46,117,112,100,97,116,101,40,100,116,41,32,101,110,100,32,45,45,32,119,105,108,108,32,112,97,115,115,32,48,32,105,102,32,108,111,118,101,46,116,105,109,101,114,32,105,115,32,100,105,115,97,98,108,101,100,10,9,105,102,32,108,111,118,101,46,103,114,97,112,104,105,99,115,32,97,110,100,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,105,115,65,99,116,105,118,101,40,41,32,116,104,101,110,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,99,108,101,97,114,40,108,111,118,101,46,103,114,97,112,104,105,99,115,46,103,101,116,66,97,99,107,103,114,111,117,110,100,67,111,108,111,114,40,41,41,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,111,114,105,103,105,110,40,41,10,9,9,105,102,32,108,111,118,101,46,100,114,97,119,32,116,104,101,110,32,108,111,118,101,46,100,114,97,119,40,41,32,101,110,100,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,112,114,101,115,101,110,116,40,41,10,9,101,110,100,10,101,110,100,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,114,117,110,40,41,10,9,105,102,32,108,111,118,101,46,109,97,116,104,32,116,104,101,110,10,9,9,108,111,118,101,46,109,97,116,104,46,115,101,116,82,97,110,100,111,109,83,101,101,100,40,111,115,46,116,105,109,101,40,41,41,10,9,101,110,100,10,9,105,102,32,108,111,118,101,46,108,111,97,100,32,116,104,101,110,32,108,111,118,101,46,108,111,97,100,40,97,114,103,41,32,101,110,100,10,9,45,45,32,87,101,32,100,111,110,39,116,32,119,97,110,116,32,116,104,101,32,102,105,114,115,116,32,102,114,97,109,101,39,115,32,100,116,32,116,111,32,105,110,99,108,117,100,101,32,116,105,109,101,32,116,97,107,101,110,32,98,121,32,108,111,118,101,46,108,111,97,100,46,10,9,105,102,32,108,111,118,101,46,116,105,109,101,114,32,116,104,101,110,32,108,111,118,101,46,116,105,109,101,114,46,115,116,101,112,40,41,32,101,110,100,10,9,45,45,32,77,97,105,110,32,108,111,111,112,32,116,105,109,101,46,10,9,119,104,105,108,101,32,116,114,117,101,32,100,111,10,9,9,108,111,118,101,46,115,116,101,112,40,41,10,9,9,105,102,32,108,111,118,101,46,116,105,109,101,114,32,116,104,101,110,32,108,111,118,101,46,116,105,109,101,114,46,115,108,101,101,112,40,48,46,48,48,49,41,32,101,110,100,10,9,101,110,100,10,101,110,100,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,69,114,114,111,114,32,115,99,114,101,101,110,46,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,108,111,99,97,108,32,100,101,98,117,103,44,32,112,114,105,110,116,32,61,32,100,101,98,117,103,44,32,112,114,105,110,116,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,101,114,114,111,114,95,112,114,105,110,116,101,114,40,109,115,103,44,32,108,97,121,101,114,41,10,9,112,114,105,110,116,40,40,100,101,98,117,103,46,116,114,97,99,101,98,97,99,107,40,34,69,114,114,111,114,58,32,34,32,46,46,32,116,111,115,116,114,105,110,103,40,109,115,103,41,44,32,49,43,40,108,97,121,101,114,32,111,114,32,49,41,41,58,103,115,117,98,40,34,92,110,91,94,92,110,93,43,36,34,44,32,34,34,41,41,41,10,101,110,100,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,101,114,114,104,97,110,100,40,109,115,103,41,10,9,109,115,103,32,61,32,116,111,115,116,114,105,110,103,40,109,115,103,41,10,9,101,114,114,111,114,95,112,114,105,110,116,101,114,40,109,115,103,44,32,50,41,10,9,105,102,32,110,111,116,32,108,111,118,101,46,119,105,110,100,111,119,32,111,114,32,110,111,116,32,108,111,118,101,46,103,114,97,112,104,105,99,115,32,111,114,32,110,111,116,32,108,111,118,101,46,101,118,101,110,116,32,116,104,101,110,10,9,9,114,101,116,117,114,110,10,9,101,110,100,10,9,105,102,32,110,111,116,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,105,115,67,114,101,97,116,101,100,40,41,32,111,114,32,110,111,116,32,108,111,118,101,46,119,105,110,100,111,119,46,105,115,79,112,101,110,40,41,32,116,104,101,110,10,9,9,108,111,99,97,108,32,115,117,99,99,101,115,115,44,32,115,116,97,116,117,115,32,61,32,112,99,97,108,108,40,108,111,118,101,46,119,105,110,100,111,119,46,115,101,116,77,111,100,101,44,32,56,48,48,44,32,54,48,48,41,10,9,9,105,102,32,110,111,116,32,115,117,99,99,101,115,115,32,111,114,32,110,111,116,32,115,116,97,116,117,115,32,116,104,101,110,10,9,9,9,114,101,116,117,114,110,10,9,9,101,110,100,10,9,101,110,100,10,9,45,45,32,82,101,115,101,116,32,115,116,97,116,101,46,10,9,105,102,32,108,111,118,101,46,109,111,117,115,101,32,116,104,101,110,10,9,9,108,111,118,101,46,109,111,117,115,101,46,115,101,116,86,105,115,105,98,108,101,40,116,114,117,101,41,10,9,9,108,111,118,101,46,109,111,117,115,101,46,115,101,116,71,114,97,98,98,101,100,40,102,97,108,115,101,41,10,9,9,108,111,118,101,46,109,111,117,115,101,46,115,101,116,82,101,108,97,116,105,118,101,77,111,100,101,40,102,97,108,115,101,41,10,9,9,105,102,32,108,111,118,101,46,109,111,117,115,101,46,104,97,115,67,117,114,115,111,114,40,41,32,116,104,101,110,10,9,9,9,108,111,118,101,46,109,111,117,115,101,46,115,101,116,67,117,114,115,111,114,40,41,10,9,9,101,110,100,10,9,101,110,100,10,9,105,102,32,108,111,118,101,46,106,111,121,115,116,105,99,107,32,116,104,101,110,10,9,9,45,45,32,83,116,111,112,32,97,108,108,32,106,111,121,115,116,105,99,107,32,118,105,98,114,97,116,105,111,110,115,46,10,9,9,102,111,114,32,105,44,118,32,105,110,32,105,112,97,105,114,115,40,108,111,118,101,46,106,111,121,115,116,105,99,107,46,103,101,116,74,111,121,115,116,105,99,107,115,40,41,41,32,100,111,10,9,9,9,118,58,115,101,116,86,105,98,114,97,116,105,111,110,40,41,10,9,9,101,110,100,10,9,101,110,100,10,9,105,102,32,108,111,118,101,46,97,117,100,105,111,32,116,104,101,110,32,108,111,118,101,46,97,117,100,105,111,46,115,116,111,112,40,41,32,101,110,100,10,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,114,101,115,101,116,40,41,10,9,108,111,99,97,108,32,102,111,110,116,32,61,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,115,101,116,78,101,119,70,111,110,116,40,109,97,116,104,46,102,108,111,111,114,40,108,111,118,101,46,119,105,110,100,111,119,46,116,111,80,105,120,101,108,115,40,49,52,41,41,41,10,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,115,101,116,66,97,99,107,103,114,111,117,110,100,67,111,108,111,114,40,56,57,44,32,49,53,55,44,32,50,50,48,41,10,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,115,101,116,67,111,108,111,114,40,50,53,53,44,32,50,53,53,44,32,50,53,53,44,32,50,53,53,41,10,9,108,111,99,97,108,32,116,114,97,99,101,32,61,32,100,101,98,117,103,46,116,114,97,99,101,98,97,99,107,40,41,10,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,99,108,101,97,114,40,108,111,118,101,46,103,114,97,112,104,105,99,115,46,103,101,116,66,97,99,107,103,114,111,117,110,100,67,111,108,111,114,40,41,41,10,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,111,114,105,103,105,110,40,41,10,9,108,111,99,97,108,32,101,114,114,32,61,32,123,125,10,9,116,97,98,108,101,46,105,110,115,101,114,116,40,101,114,114,44,32,34,69,114,114,111,114,92,110,34,41,10,9,116,97,98,108,101,46,105,110,115,101,114,116,40,101,114,114,44,32,109,115,103,46,46,34,92,110,92,110,34,41,10,9,102,111,114,32,108,32,105,110,32,115,116,114,105,110,103,46,103,109,97,116,99,104,40,116,114,97,99,101,44,32,34,40,46,45,41,92,110,34,41,32,100,111,10,9,9,105,102,32,110,111,116,32,115,116,114,105,110,103,46,109,97,116,99,104,40,108,44,32,34,98,111,111,116,46,108,117,97,34,41,32,116,104,101,110,10,9,9,9,108,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,108,44,32,34,115,116,97,99,107,32,116,114,97,99,101,98,97,99,107,58,34,44,32,34,84,114,97,99,101,98,97,99,107,92,110,34,41,10,9,9,9,116,97,98,108,101,46,105,110,115,101,114,116,40,101,114,114,44,32,108,41,10,9,9,101,110,100,10,9,101,110,100,10,9,108,111,99,97,108,32,112,32,61,32,116,97,98,108,101,46,99,111,110,99,97,116,40,101,114,114,44,32,34,92,110,34,41,10,9,112,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,112,44,32,34,92,116,34,44,32,34,34,41,10,9,112,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,112,44,32,34,37,91,115,116,114,105,110,103,32,92,34,40,46,45,41,92,34,37,93,34,44,32,34,37,49,34,41,10,9,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,100,114,97,119,40,41,10,9,9,108,111,99,97,108,32,112,111,115,32,61,32,108,111,118,101,46,119,105,110,100,111,119,46,116,111,80,105,120,101,108,115,40,55,48,41,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,99,108,101,97,114,40,108,111,118,101,46,103,114,97,112,104,105,99,115,46,103,101,116,66,97,99,107,103,114,111,117,110,100,67,111,108,111,114,40,41,41,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,112,114,105,110,116,102,40,112,44,32,112,111,115,44,32,112,111,115,44,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,103,101,116,87,105,100,116,104,40,41,32,45,32,112,111,115,41,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,112,114,101,115,101,110,116,40,41,10,9,101,110,100,10,9,119,104,105,108,101,32,116,114,117,101,32,100,111,10,9,9,108,111,118,101,46,101,118,101,110,116,46,112,117,109,112,40,41,10,9,9,102,111,114,32,101,44,32,97,44,32,98,44,32,99,32,105,110,32,108,111,118,101,46,101,118,101,110,116,46,112,111,108,108,40,41,32,100,111,10,9,9,9,105,102,32,101,32,61,61,32,34,113,117,105,116,34,32,116,104,101,110,10,9,9,9,9,114,101,116,117,114,110,10,9,9,9,101,108,115,101,105,102,32,101,32,61,61,32,34,107,101,121,112,114,101,115,115,101,100,34,32,97,110,100,32,97,32,61,61,32,34,101,115,99,97,112,101,34,32,116,104,101,110,10,9,9,9,9,114,101,116,117,114,110,10,9,9,9,101,108,115,101,105,102,32,101,32,61,61,32,34,116,111,117,99,104,112,114,101,115,115,101,100,34,32,116,104,101,110,10,9,9,9,9,108,111,99,97,108,32,110,97,109,101,32,61,32,108,111,118,101,46,119,105,110,100,111,119,46,103,101,116,84,105,116,108,101,40,41,10,9,9,9,9,105,102,32,35,110,97,109,101,32,61,61,32,48,32,111,114,32,110,97,109,101,32,61,61,32,34,85,110,116,105,116,108,101,100,34,32,116,104,101,110,32,110,97,109,101,32,61,32,34,71,97,109,101,34,32,101,110,100,10,9,9,9,9,108,111,99,97,108,32,98,117,116,116,111,110,115,32,61,32,123,34,79,75,34,44,32,34,67,97,110,99,101,108,34,125,10,9,9,9,9,108,111,99,97,108,32,112,114,101,115,115,101,100,32,61,32,108,111,118,101,46,119,105,110,100,111,119,46,115,104,111,119,77,101,115,115,97,103,101,66,111,120,40,34,81,117,105,116,32,34,46,46,110,97,109,101,46,46,34,63,34,44,32,34,34,44,32,98,117,116,116,111,110,115,41,10,9,9,9,9,105,102,32,112,114,101,115,115,101,100,32,61,61,32,49,32,116,104,101,110,10,9,9,9,9,9,114,101,116,117,114,110,10,9,9,9,9,101,110,100,10,9,9,9,101,110,100,10,9,9,101,110,100,10,9,9,100,114,97,119,40,41,10,9,9,105,102,32,108,111,118,101,46,116,105,109,101,114,32,116,104,101,110,10,9,9,9,108,111,118,101,46,116,105,109,101,114,46,115,108,101,101,112,40,48,46,49,41,10,9,9,101,110,100,10,9,101,110,100,10,101,110,100,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,100,101,102,101,114,69,114,114,104,97,110,100,40,46,46,46,41,10,9,108,111,99,97,108,32,104,97,110,100,108,101,114,32,61,32,108,111,118,101,46,101,114,114,104,97,110,100,32,111,114,32,101,114,114,111,114,95,112,114,105,110,116,101,114,10,9,114,101,116,117,114,110,32,104,97,110,100,108,101,114,40,46,46,46,41,10,101,110,100,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,84,104,101,32,114,111,111,116,32,111,102,32,97,108,108,32,99,97,108,108,115,46,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,114,101,116,117,114,110,32,102,117,110,99,116,105,111,110,40,99,97,110,95,108,111,111,112,41,10,9,108,111,99,97,108,32,114,101,116,118,97,108,32,61,32,48,10,9,108,111,99,97,108,32,114,101,115,117,108,116,32,61,32,120,112,99,97,108,108,40,108,111,118,101,46,98,111,111,116,44,32,101,114,114,111,114,95,112,114,105,110,116,101,114,41,10,9,105,102,32,110,111,116,32,114,101,115,117,108,116,32,116,104,101,110,32,114,101,116,117,114,110,32,49,32,101,110,100,10,9,108,111,99,97,108,32,114,101,115,117,108,116,32,61,32,120,112,99,97,108,108,40,108,111,118,101,46,105,110,105,116,44,32,100,101,102,101,114,69,114,114,104,97,110,100,41,10,9,105,102,32,110,111,116,32,114,101,115,117,108,116,32,116,104,101,110,32,114,101,116,117,114,110,32,49,32,101,110,100,10,9,105,102,32,99,97,110,95,108,111,111,112,32,116,104,101,110,10,9,9,108,111,99,97,108,32,114,101,115,117,108,116,32,61,32,120,112,99,97,108,108,40,108,111,118,101,46,114,117,110,44,32,100,101,102,101,114,69,114,114,104,97,110,100,41,10,9,9,105,102,32,110,111,116,32,114,101,115,117,108,116,32,116,104,101,110,32,114,101,116,117,114,110,32,49,32,101,110,100,10,9,101,108,115,101,10,9,9,105,102,32,108,111,118,101,46,109,97,116,104,32,116,104,101,110,10,9,9,9,108,111,118,101,46,109,97,116,104,46,115,101,116,82,97,110,100,111,109,83,101,101,100,40,111,115,46,116,105,109,101,40,41,41,10,9,9,101,110,100,10,9,9,105,102,32,108,111,118,101,46,108,111,97,100,32,116,104,101,110,32,108,111,118,101,46,108,111,97,100,40,97,114,103,41,32,101,110,100,10,9,9,114,101,116,117,114,110,32,108,111,118,101,46,115,116,101,112,10,9,101,110,100,10,9,114,101,116,117,114,110,32,116,111,110,117,109,98,101,114,40,114,101,116,118,97,108,41,10,101,110,100,10,98,111,111,116,46,108,117,97,0,78,101,101,100,32,97,116,32,108,101,97,115,116,32,51,32,118,101,114,116,105,99,101,115,32,116,111,32,116,114,105,97,110,103,117,108,97,116,101,0,73,110,118,97,108,105,100,32,99,111,109,112,114,101,115,115,101,100,32,100,97,116,97,32,102,111,114,109,97,116,58,32,37,115,0,95,103,101,116,82,97,110,100,111,109,71,101,110,101,114,97,116,111,114,0,110,101,119,82,97,110,100,111,109,71,101,110,101,114,97,116,111,114,0,110,101,119,66,101,122,105,101,114,67,117,114,118,101,0,116,114,105,97,110,103,117,108,97,116,101,0,105,115,67,111,110,118,101,120,0,103,97,109,109,97,84,111,76,105,110,101,97,114,0,108,105,110,101,97,114,84,111,71,97,109,109,97,0,110,111,105,115,101,0,99,111,109,112,114,101,115,115,0,100,101,99,111,109,112,114,101,115,115,0,10,45,45,32,68,79,32,78,79,84,32,82,69,77,79,86,69,32,84,72,69,32,65,66,79,86,69,32,76,73,78,69,46,32,73,116,32,105,115,32,117,115,101,100,32,116,111,32,108,111,97,100,32,116,104,105,115,32,102,105,108,101,32,97,115,32,97,32,67,43,43,32,115,116,114,105,110,103,46,10,45,45,32,84,104,101,114,101,32,105,115,32,97,32,109,97,116,99,104,105,110,103,32,100,101,108,105,109,105,116,101,114,32,97,116,32,116,104,101,32,98,111,116,116,111,109,32,111,102,32,116,104,101,32,102,105,108,101,46,10,10,45,45,91,91,10,67,111,112,121,114,105,103,104,116,32,40,99,41,32,50,48,48,54,45,50,48,49,54,32,76,79,86,69,32,68,101,118,101,108,111,112,109,101,110,116,32,84,101,97,109,10,10,84,104,105,115,32,115,111,102,116,119,97,114,101,32,105,115,32,112,114,111,118,105,100,101,100,32,39,97,115,45,105,115,39,44,32,119,105,116,104,111,117,116,32,97,110,121,32,101,120,112,114,101,115,115,32,111,114,32,105,109,112,108,105,101,100,10,119,97,114,114,97,110,116,121,46,32,32,73,110,32,110,111,32,101,118,101,110,116,32,119,105,108,108,32,116,104,101,32,97,117,116,104,111,114,115,32,98,101,32,104,101,108,100,32,108,105,97,98,108,101,32,102,111,114,32,97,110,121,32,100,97,109,97,103,101,115,10,97,114,105,115,105,110,103,32,102,114,111,109,32,116,104,101,32,117,115,101,32,111,102,32,116,104,105,115,32,115,111,102,116,119,97,114,101,46,10,10,80,101,114,109,105,115,115,105,111,110,32,105,115,32,103,114,97,110,116,101,100,32,116,111,32,97,110,121,111,110,101,32,116,111,32,117,115,101,32,116,104,105,115,32,115,111,102,116,119,97,114,101,32,102,111,114,32,97,110,121,32,112,117,114,112,111,115,101,44,10,105,110,99,108,117,100,105,110,103,32,99,111,109,109,101,114,99,105,97,108,32,97,112,112,108,105,99,97,116,105,111,110,115,44,32,97,110,100,32,116,111,32,97,108,116,101,114,32,105,116,32,97,110,100,32,114,101,100,105,115,116,114,105,98,117,116,101,32,105,116,10,102,114,101,101,108,121,44,32,115,117,98,106,101,99,116,32,116,111,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,114,101,115,116,114,105,99,116,105,111,110,115,58,10,10,49,46,32,84,104,101,32,111,114,105,103,105,110,32,111,102,32,116,104,105,115,32,115,111,102,116,119,97,114,101,32,109,117,115,116,32,110,111,116,32,98,101,32,109,105,115,114,101,112,114,101,115,101,110,116,101,100,59,32,121,111,117,32,109,117,115,116,32,110,111,116,10,99,108,97,105,109,32,116,104,97,116,32,121,111,117,32,119,114,111,116,101,32,116,104,101,32,111,114,105,103,105,110,97,108,32,115,111,102,116,119,97,114,101,46,32,73,102,32,121,111,117,32,117,115,101,32,116,104,105,115,32,115,111,102,116,119,97,114,101,10,105,110,32,97,32,112,114,111,100,117,99,116,44,32,97,110,32,97,99,107,110,111,119,108,101,100,103,109,101,110,116,32,105,110,32,116,104,101,32,112,114,111,100,117,99,116,32,100,111,99,117,109,101,110,116,97,116,105,111,110,32,119,111,117,108,100,32,98,101,10,97,112,112,114,101,99,105,97,116,101,100,32,98,117,116,32,105,115,32,110,111,116,32,114,101,113,117,105,114,101,100,46,10,50,46,32,65,108,116,101,114,101,100,32,115,111,117,114,99,101,32,118,101,114,115,105,111,110,115,32,109,117,115,116,32,98,101,32,112,108,97,105,110,108,121,32,109,97,114,107,101,100,32,97,115,32,115,117,99,104,44,32,97,110,100,32,109,117,115,116,32,110,111,116,32,98,101,10,109,105,115,114,101,112,114,101,115,101,110,116,101,100,32,97,115,32,98,101,105,110,103,32,116,104,101,32,111,114,105,103,105,110,97,108,32,115,111,102,116,119,97,114,101,46,10,51,46,32,84,104,105,115,32,110,111,116,105,99,101,32,109,97,121,32,110,111,116,32,98,101,32,114,101,109,111,118,101,100,32,111,114,32,97,108,116,101,114,101,100,32,102,114,111,109,32,97,110,121,32,115,111,117,114,99,101,32,100,105,115,116,114,105,98,117,116,105,111,110,46,10,45,45,93,93,10,10,108,111,99,97,108,32,108,111,118,101,95,109,97,116,104,44,32,102,102,105,102,117,110,99,115,112,111,105,110,116,101,114,32,61,32,46,46,46,10,10,108,111,99,97,108,32,116,121,112,101,44,32,116,111,110,117,109,98,101,114,44,32,101,114,114,111,114,32,61,32,116,121,112,101,44,32,116,111,110,117,109,98,101,114,44,32,101,114,114,111,114,10,108,111,99,97,108,32,102,108,111,111,114,32,61,32,109,97,116,104,46,102,108,111,111,114,10,10,108,111,99,97,108,32,114,110,103,32,61,32,108,111,118,101,95,109,97,116,104,46,95,103,101,116,82,97,110,100,111,109,71,101,110,101,114,97,116,111,114,40,41,10,10,102,117,110,99,116,105,111,110,32,108,111,118,101,95,109,97,116,104,46,114,97,110,100,111,109,40,108,44,32,117,41,10,9,114,101,116,117,114,110,32,114,110,103,58,114,97,110,100,111,109,40,108,44,32,117,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,108,111,118,101,95,109,97,116,104,46,114,97,110,100,111,109,78,111,114,109,97,108,40,115,116,100,100,101,118,44,32,109,101,97,110,41,10,9,114,101,116,117,114,110,32,114,110,103,58,114,97,110,100,111,109,78,111,114,109,97,108,40,115,116,100,100,101,118,44,32,109,101,97,110,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,108,111,118,101,95,109,97,116,104,46,115,101,116,82,97,110,100,111,109,83,101,101,100,40,108,111,119,44,32,104,105,103,104,41,10,9,114,101,116,117,114,110,32,114,110,103,58,115,101,116,83,101,101,100,40,108,111,119,44,32,104,105,103,104,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,108,111,118,101,95,109,97,116,104,46,103,101,116,82,97,110,100,111,109,83,101,101,100,40,41,10,9,114,101,116,117,114,110,32,114,110,103,58,103,101,116,83,101,101,100,40,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,108,111,118,101,95,109,97,116,104,46,115,101,116,82,97,110,100,111,109,83,116,97,116,101,40,115,116,97,116,101,41,10,9,114,101,116,117,114,110,32,114,110,103,58,115,101,116,83,116,97,116,101,40,115,116,97,116,101,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,108,111,118,101,95,109,97,116,104,46,103,101,116,82,97,110,100,111,109,83,116,97,116,101,40,41,10,9,114,101,116,117,114,110,32,114,110,103,58,103,101,116,83,116,97,116,101,40,41,10,101,110,100,10,10,105,102,32,116,121,112,101,40,106,105,116,41,32,126,61,32,34,116,97,98,108,101,34,32,111,114,32,110,111,116,32,106,105,116,46,115,116,97,116,117,115,40,41,32,116,104,101,110,10,9,45,45,32,76,117,97,74,73,84,39,115,32,70,70,73,32,105,115,32,42,109,117,99,104,42,32,115,108,111,119,101,114,32,116,104,97,110,32,76,79,86,69,39,115,32,114,101,103,117,108,97,114,32,109,101,116,104,111,100,115,32,119,104,101,110,32,116,104,101,32,74,73,84,10,9,45,45,32,99,111,109,112,105,108,101,114,32,105,115,32,100,105,115,97,98,108,101,100,46,10,9,114,101,116,117,114,110,10,101,110,100,10,10,108,111,99,97,108,32,115,116,97,116,117,115,44,32,102,102,105,32,61,32,112,99,97,108,108,40,114,101,113,117,105,114,101,44,32,34,102,102,105,34,41,10,105,102,32,110,111,116,32,115,116,97,116,117,115,32,116,104,101,110,32,114,101,116,117,114,110,32,101,110,100,10,10,45,45,32,77,97,116,99,104,101,115,32,116,104,101,32,115,116,114,117,99,116,32,100,101,99,108,97,114,97,116,105,111,110,32,105,110,32,119,114,97,112,95,77,97,116,104,46,99,112,112,46,10,112,99,97,108,108,40,102,102,105,46,99,100,101,102,44,32,91,91,10,116,121,112,101,100,101,102,32,115,116,114,117,99,116,32,70,70,73,95,77,97,116,104,10,123,10,9,102,108,111,97,116,32,40,42,110,111,105,115,101,49,41,40,102,108,111,97,116,32,120,41,59,10,9,102,108,111,97,116,32,40,42,110,111,105,115,101,50,41,40,102,108,111,97,116,32,120,44,32,102,108,111,97,116,32,121,41,59,10,9,102,108,111,97,116,32,40,42,110,111,105,115,101,51,41,40,102,108,111,97,116,32,120,44,32,102,108,111,97,116,32,121,44,32,102,108,111,97,116,32,122,41,59,10,9,102,108,111,97,116,32,40,42,110,111,105,115,101,52,41,40,102,108,111,97,116,32,120,44,32,102,108,111,97,116,32,121,44,32,102,108,111,97,116,32,122,44,32,102,108,111,97,116,32,119,41,59,10,10,9,102,108,111,97,116,32,40,42,103,97,109,109,97,84,111,76,105,110,101,97,114,41,40,102,108,111,97,116,32,99,41,59,10,9,102,108,111,97,116,32,40,42,108,105,110,101,97,114,84,111,71,97,109,109,97,41,40,102,108,111,97,116,32,99,41,59,10,125,32,70,70,73,95,77,97,116,104,59,10,93,93,41,10,10,108,111,99,97,108,32,102,102,105,102,117,110,99,115,32,61,32,102,102,105,46,99,97,115,116,40,34,70,70,73,95,77,97,116,104,32,42,34,44,32,102,102,105,102,117,110,99,115,112,111,105,110,116,101,114,41,10,10,10,45,45,32,79,118,101,114,119,114,105,116,101,32,115,111,109,101,32,114,101,103,117,108,97,114,32,108,111,118,101,46,109,97,116,104,32,102,117,110,99,116,105,111,110,115,32,119,105,116,104,32,70,70,73,32,105,109,112,108,101,109,101,110,116,97,116,105,111,110,115,46,10,10,102,117,110,99,116,105,111,110,32,108,111,118,101,95,109,97,116,104,46,110,111,105,115,101,40,120,44,32,121,44,32,122,44,32,119,41,10,9,105,102,32,119,32,126,61,32,110,105,108,32,116,104,101,110,10,9,9,114,101,116,117,114,110,32,116,111,110,117,109,98,101,114,40,102,102,105,102,117,110,99,115,46,110,111,105,115,101,52,40,120,44,32,121,44,32,122,44,32,119,41,41,10,9,101,108,115,101,105,102,32,122,32,126,61,32,110,105,108,32,116,104,101,110,10,9,9,114,101,116,117,114,110,32,116,111,110,117,109,98,101,114,40,102,102,105,102,117,110,99,115,46,110,111,105,115,101,51,40,120,44,32,121,44,32,122,41,41,10,9,101,108,115,101,105,102,32,121,32,126,61,32,110,105,108,32,116,104,101,110,10,9,9,114,101,116,117,114,110,32,116,111,110,117,109,98,101,114,40,102,102,105,102,117,110,99,115,46,110,111,105,115,101,50,40,120,44,32,121,41,41,10,9,101,108,115,101,105,102,32,120,32,126,61,32,110,105,108,32,116,104,101,110,10,9,9,114,101,116,117,114,110,32,116,111,110,117,109,98,101,114,40,102,102,105,102,117,110,99,115,46,110,111,105,115,101,49,40,120,41,41,10,9,101,110,100,10,101,110,100,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,103,97,109,109,97,84,111,76,105,110,101,97,114,40,99,41,10,9,105,102,32,99,32,126,61,32,110,105,108,32,116,104,101,110,10,9,9,114,101,116,117,114,110,32,116,111,110,117,109,98,101,114,40,102,102,105,102,117,110,99,115,46,103,97,109,109,97,84,111,76,105,110,101,97,114,40,99,32,47,32,50,53,53,41,41,32,42,32,50,53,53,10,9,101,110,100,10,9,114,101,116,117,114,110,32,99,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,108,111,118,101,95,109,97,116,104,46,103,97,109,109,97,84,111,76,105,110,101,97,114,40,114,44,32,103,44,32,98,44,32,97,41,10,9,105,102,32,116,121,112,101,40,114,41,32,61,61,32,34,116,97,98,108,101,34,32,116,104,101,110,10,9,9,108,111,99,97,108,32,116,32,61,32,114,10,9,9,114,101,116,117,114,110,32,103,97,109,109,97,84,111,76,105,110,101,97,114,40,116,91,49,93,41,44,32,103,97,109,109,97,84,111,76,105,110,101,97,114,40,116,91,50,93,41,44,32,103,97,109,109,97,84,111,76,105,110,101,97,114,40,116,91,51,93,41,44,32,116,91,52,93,10,9,101,110,100,10,9,114,101,116,117,114,110,32,103,97,109,109,97,84,111,76,105,110,101,97,114,40,114,41,44,32,103,97,109,109,97,84,111,76,105,110,101,97,114,40,103,41,44,32,103,97,109,109,97,84,111,76,105,110,101,97,114,40,98,41,44,32,97,10,101,110,100,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,108,105,110,101,97,114,84,111,71,97,109,109,97,40,99,41,10,9,105,102,32,99,32,126,61,32,110,105,108,32,116,104,101,110,10,9,9,114,101,116,117,114,110,32,116,111,110,117,109,98,101,114,40,102,102,105,102,117,110,99,115,46,108,105,110,101,97,114,84,111,71,97,109,109,97,40,99,32,47,32,50,53,53,41,41,32,42,32,50,53,53,10,9,101,110,100,10,9,114,101,116,117,114,110,32,99,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,108,111,118,101,95,109,97,116,104,46,108,105,110,101,97,114,84,111,71,97,109,109,97,40,114,44,32,103,44,32,98,44,32,97,41,10,9,105,102,32,116,121,112,101,40,114,41,32,61,61,32,34,116,97,98,108,101,34,32,116,104,101,110,10,9,9,108,111,99,97,108,32,116,32,61,32,114,10,9,9,114,101,116,117,114,110,32,108,105,110,101,97,114,84,111,71,97,109,109,97,40,116,91,49,93,41,44,32,108,105,110,101,97,114,84,111,71,97,109,109,97,40,116,91,50,93,41,44,32,108,105,110,101,97,114,84,111,71,97,109,109,97,40,116,91,51,93,41,44,32,116,91,52,93,10,9,101,110,100,10,9,114,101,116,117,114,110,32,108,105,110,101,97,114,84,111,71,97,109,109,97,40,114,41,44,32,108,105,110,101,97,114,84,111,71,97,109,109,97,40,103,41,44,32,108,105,110,101,97,114,84,111,71,97,109,109,97,40,98,41,44,32,97,10,101,110,100,10,10,45,45,32,68,79,32,78,79,84,32,82,69,77,79,86,69,32,84,72,69,32,78,69,88,84,32,76,73,78,69,46,32,73,116,32,105,115,32,117,115,101,100,32,116,111,32,108,111,97,100,32,116,104,105,115,32,102,105,108,101,32,97,115,32,97,32,67,43,43,32,115,116,114,105,110,103,46,10,45,45,0,119,114,97,112,95,77,97,116,104,46,108,117,97,0,105,110,118,97,108,105,100,32,114,97,110,100,111,109,32,115,101,101,100,0,82,97,110,100,111,109,71,101,110,101,114,97,116,111,114,0,95,114,97,110,100,111,109,0,114,97,110,100,111,109,78,111,114,109,97,108,0,115,101,116,83,101,101,100,0,103,101,116,83,101,101,100,0,115,101,116,83,116,97,116,101,0,103,101,116,83,116,97,116,101,0,10,45,45,32,68,79,32,78,79,84,32,82,69,77,79,86,69,32,84,72,69,32,65,66,79,86,69,32,76,73,78,69,46,32,73,116,32,105,115,32,117,115,101,100,32,116,111,32,108,111,97,100,32,116,104,105,115,32,102,105,108,101,32,97,115,32,97,32,67,43,43,32,115,116,114,105,110,103,46,10,45,45,32,84,104,101,114,101,32,105,115,32,97,32,109,97,116,99,104,105,110,103,32,100,101,108,105,109,105,116,101,114,32,97,116,32,116,104,101,32,98,111,116,116,111,109,32,111,102,32,116,104,101,32,102,105,108,101,46,10,10,45,45,91,91,10,67,111,112,121,114,105,103,104,116,32,40,99,41,32,50,48,48,54,45,50,48,49,54,32,76,79,86,69,32,68,101,118,101,108,111,112,109,101,110,116,32,84,101,97,109,10,10,84,104,105,115,32,115,111,102,116,119,97,114,101,32,105,115,32,112,114,111,118,105,100,101,100,32,39,97,115,45,105,115,39,44,32,119,105,116,104,111,117,116,32,97,110,121,32,101,120,112,114,101,115,115,32,111,114,32,105,109,112,108,105,101,100,10,119,97,114,114,97,110,116,121,46,32,32,73,110,32,110,111,32,101,118,101,110,116,32,119,105,108,108,32,116,104,101,32,97,117,116,104,111,114,115,32,98,101,32,104,101,108,100,32,108,105,97,98,108,101,32,102,111,114,32,97,110,121,32,100,97,109,97,103,101,115,10,97,114,105,115,105,110,103,32,102,114,111,109,32,116,104,101,32,117,115,101,32,111,102,32,116,104,105,115,32,115,111,102,116,119,97,114,101,46,10,10,80,101,114,109,105,115,115,105,111,110,32,105,115,32,103,114,97,110,116,101,100,32,116,111,32,97,110,121,111,110,101,32,116,111,32,117,115,101,32,116,104,105,115,32,115,111,102,116,119,97,114,101,32,102,111,114,32,97,110,121,32,112,117,114,112,111,115,101,44,10,105,110,99,108,117,100,105,110,103,32,99,111,109,109,101,114,99,105,97,108,32,97,112,112,108,105,99,97,116,105,111,110,115,44,32,97,110,100,32,116,111,32,97,108,116,101,114,32,105,116,32,97,110,100,32,114,101,100,105,115,116,114,105,98,117,116,101,32,105,116,10,102,114,101,101,108,121,44,32,115,117,98,106,101,99,116,32,116,111,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,114,101,115,116,114,105,99,116,105,111,110,115,58,10,10,49,46,32,84,104,101,32,111,114,105,103,105,110,32,111,102,32,116,104,105,115,32,115,111,102,116,119,97,114,101,32,109,117,115,116,32,110,111,116,32,98,101,32,109,105,115,114,101,112,114,101,115,101,110,116,101,100,59,32,121,111,117,32,109,117,115,116,32,110,111,116,10,99,108,97,105,109,32,116,104,97,116,32,121,111,117,32,119,114,111,116,101,32,116,104,101,32,111,114,105,103,105,110,97,108,32,115,111,102,116,119,97,114,101,46,32,73,102,32,121,111,117,32,117,115,101,32,116,104,105,115,32,115,111,102,116,119,97,114,101,10,105,110,32,97,32,112,114,111,100,117,99,116,44,32,97,110,32,97,99,107,110,111,119,108,101,100,103,109,101,110,116,32,105,110,32,116,104,101,32,112,114,111,100,117,99,116,32,100,111,99,117,109,101,110,116,97,116,105,111,110,32,119,111,117,108,100,32,98,101,10,97,112,112,114,101,99,105,97,116,101,100,32,98,117,116,32,105,115,32,110,111,116,32,114,101,113,117,105,114,101,100,46,10,50,46,32,65,108,116,101,114,101,100,32,115,111,117,114,99,101,32,118,101,114,115,105,111,110,115,32,109,117,115,116,32,98,101,32,112,108,97,105,110,108,121,32,109,97,114,107,101,100,32,97,115,32,115,117,99,104,44,32,97,110,100,32,109,117,115,116,32,110,111,116,32,98,101,10,109,105,115,114,101,112,114,101,115,101,110,116,101,100,32,97,115,32,98,101,105,110,103,32,116,104,101,32,111,114,105,103,105,110,97,108,32,115,111,102,116,119,97,114,101,46,10,51,46,32,84,104,105,115,32,110,111,116,105,99,101,32,109,97,121,32,110,111,116,32,98,101,32,114,101,109,111,118,101,100,32,111,114,32,97,108,116,101,114,101,100,32,102,114,111,109,32,97,110,121,32,115,111,117,114,99,101,32,100,105,115,116,114,105,98,117,116,105,111,110,46,10,45,45,93,93,10,10,108,111,99,97,108,32,82,97,110,100,111,109,71,101,110,101,114,97,116,111,114,95,109,116,44,32,102,102,105,102,117,110,99,115,112,111,105,110,116,101,114,32,61,32,46,46,46,10,108,111,99,97,108,32,82,97,110,100,111,109,71,101,110,101,114,97,116,111,114,32,61,32,82,97,110,100,111,109,71,101,110,101,114,97,116,111,114,95,109,116,46,95,95,105,110,100,101,120,10,10,108,111,99,97,108,32,116,121,112,101,44,32,116,111,110,117,109,98,101,114,44,32,101,114,114,111,114,32,61,32,116,121,112,101,44,32,116,111,110,117,109,98,101,114,44,32,101,114,114,111,114,10,108,111,99,97,108,32,102,108,111,111,114,32,61,32,109,97,116,104,46,102,108,111,111,114,10,10,108,111,99,97,108,32,95,114,97,110,100,111,109,32,61,32,82,97,110,100,111,109,71,101,110,101,114,97,116,111,114,46,95,114,97,110,100,111,109,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,103,101,116,114,97,110,100,111,109,40,114,44,32,108,44,32,117,41,10,9,105,102,32,117,32,126,61,32,110,105,108,32,116,104,101,110,10,9,9,105,102,32,116,121,112,101,40,114,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,97,114,103,117,109,101,110,116,32,35,49,32,116,111,32,39,114,97,110,100,111,109,39,32,40,110,117,109,98,101,114,32,101,120,112,101,99,116,101,100,41,34,44,32,50,41,32,101,110,100,10,9,9,105,102,32,116,121,112,101,40,108,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,97,114,103,117,109,101,110,116,32,35,50,32,116,111,32,39,114,97,110,100,111,109,39,32,40,110,117,109,98,101,114,32,101,120,112,101,99,116,101,100,41,34,44,32,50,41,32,101,110,100,10,9,9,114,101,116,117,114,110,32,102,108,111,111,114,40,114,32,42,32,40,117,32,45,32,108,32,43,32,49,41,41,32,43,32,108,10,9,101,108,115,101,105,102,32,108,32,126,61,32,110,105,108,32,116,104,101,110,10,9,9,105,102,32,116,121,112,101,40,108,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,97,114,103,117,109,101,110,116,32,35,49,32,116,111,32,39,114,97,110,100,111,109,39,32,40,110,117,109,98,101,114,32,101,120,112,101,99,116,101,100,41,34,44,32,50,41,32,101,110,100,10,9,9,114,101,116,117,114,110,32,102,108,111,111,114,40,114,32,42,32,108,41,32,43,32,49,10,9,101,108,115,101,10,9,9,114,101,116,117,114,110,32,114,10,9,101,110,100,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,82,97,110,100,111,109,71,101,110,101,114,97,116,111,114,58,114,97,110,100,111,109,40,108,44,32,117,41,10,9,108,111,99,97,108,32,114,32,61,32,95,114,97,110,100,111,109,40,115,101,108,102,41,10,9,114,101,116,117,114,110,32,103,101,116,114,97,110,100,111,109,40,114,44,32,108,44,32,117,41,10,101,110,100,10,10,105,102,32,116,121,112,101,40,106,105,116,41,32,126,61,32,34,116,97,98,108,101,34,32,111,114,32,110,111,116,32,106,105,116,46,115,116,97,116,117,115,40,41,32,116,104,101,110,10,9,45,45,32,76,117,97,74,73,84,39,115,32,70,70,73,32,105,115,32,42,109,117,99,104,42,32,115,108,111,119,101,114,32,116,104,97,110,32,76,79,86,69,39,115,32,114,101,103,117,108,97,114,32,109,101,116,104,111,100,115,32,119,104,101,110,32,116,104,101,32,74,73,84,10,9,45,45,32,99,111,109,112,105,108,101,114,32,105,115,32,100,105,115,97,98,108,101,100,46,10,9,114,101,116,117,114,110,10,101,110,100,10,10,108,111,99,97,108,32,115,116,97,116,117,115,44,32,102,102,105,32,61,32,112,99,97,108,108,40,114,101,113,117,105,114,101,44,32,34,102,102,105,34,41,10,105,102,32,110,111,116,32,115,116,97,116,117,115,32,116,104,101,110,32,114,101,116,117,114,110,32,101,110,100,10,10,112,99,97,108,108,40,102,102,105,46,99,100,101,102,44,32,91,91,10,116,121,112,101,100,101,102,32,115,116,114,117,99,116,32,80,114,111,120,121,32,80,114,111,120,121,59,10,10,116,121,112,101,100,101,102,32,115,116,114,117,99,116,32,70,70,73,95,82,97,110,100,111,109,71,101,110,101,114,97,116,111,114,10,123,10,9,100,111,117,98,108,101,32,40,42,114,97,110,100,111,109,41,40,80,114,111,120,121,32,42,112,41,59,10,125,32,70,70,73,95,82,97,110,100,111,109,71,101,110,101,114,97,116,111,114,59,10,93,93,41,10,10,108,111,99,97,108,32,102,102,105,102,117,110,99,115,32,61,32,102,102,105,46,99,97,115,116,40,34,70,70,73,95,82,97,110,100,111,109,71,101,110,101,114,97,116,111,114,32,42,34,44,32,102,102,105,102,117,110,99,115,112,111,105,110,116,101,114,41,10,10,10,45,45,32,79,118,101,114,119,114,105,116,101,32,115,111,109,101,32,114,101,103,117,108,97,114,32,108,111,118,101,46,109,97,116,104,32,102,117,110,99,116,105,111,110,115,32,119,105,116,104,32,70,70,73,32,105,109,112,108,101,109,101,110,116,97,116,105,111,110,115,46,10,10,102,117,110,99,116,105,111,110,32,82,97,110,100,111,109,71,101,110,101,114,97,116,111,114,58,114,97,110,100,111,109,40,108,44,32,117,41,10,9,108,111,99,97,108,32,114,32,61,32,116,111,110,117,109,98,101,114,40,102,102,105,102,117,110,99,115,46,114,97,110,100,111,109,40,115,101,108,102,41,41,10,9,114,101,116,117,114,110,32,103,101,116,114,97,110,100,111,109,40,114,44,32,108,44,32,117,41,10,101,110,100,10,10,45,45,32,68,79,32,78,79,84,32,82,69,77,79,86,69,32,84,72,69,32,78,69,88,84,32,76,73,78,69,46,32,73,116,32,105,115,32,117,115,101,100,32,116,111,32,108,111,97,100,32,116,104,105,115,32,102,105,108,101,32,97,115,32,97,32,67,43,43,32,115,116,114,105,110,103,46,10,45,45,0,119,114,97,112,95,82,97,110,100,111,109,71,101,110,101,114,97,116,111,114,46,108,117,97,0,73,110,118,97,108,105,100,32,115,121,115,116,101,109,32,99,117,114,115,111,114,32,116,121,112,101,58,32,37,115,0,110,101,119,67,117,114,115,111,114,0,103,101,116,83,121,115,116,101,109,67,117,114,115,111,114,0,115,101,116,67,117,114,115,111,114,0,103,101,116,67,117,114,115,111,114,0,104,97,115,67,117,114,115,111,114,0,115,101,116,86,105,115,105,98,108,101,0,115,101,116,71,114,97,98,98,101,100,0,105,115,71,114,97,98,98,101,100,0,115,101,116,82,101,108,97,116,105,118,101,77,111,100,101,0,103,101,116,82,101,108,97,116,105,118,101,77,111,100,101,0,73,110,118,97,108,105,100,32,66,111,100,121,32,116,121,112,101,58,32,37,115,0,73,110,99,111,114,114,101,99,116,32,110,117,109,98,101,114,32,111,102,32,112,97,114,97,109,101,116,101,114,115,0,112,104,121,115,105,99,115,0,110,101,119,87,111,114,108,100,0,110,101,119,66,111,100,121,0,110,101,119,70,105,120,116,117,114,101,0,110,101,119,67,105,114,99,108,101,83,104,97,112,101,0,110,101,119,82,101,99,116,97,110,103,108,101,83,104,97,112,101,0,110,101,119,80,111,108,121,103,111,110,83,104,97,112,101,0,110,101,119,69,100,103,101,83,104,97,112,101,0,110,101,119,67,104,97,105,110,83,104,97,112,101,0,110,101,119,68,105,115,116,97,110,99,101,74,111,105,110,116,0,110,101,119,77,111,117,115,101,74,111,105,110,116,0,110,101,119,82,101,118,111,108,117,116,101,74,111,105,110,116,0,110,101,119,80,114,105,115,109,97,116,105,99,74,111,105,110,116,0,110,101,119,80,117,108,108,101,121,74,111,105,110,116,0,110,101,119,71,101,97,114,74,111,105,110,116,0,110,101,119,70,114,105,99,116,105,111,110,74,111,105,110,116,0,110,101,119,87,101,108,100,74,111,105,110,116,0,110,101,119,87,104,101,101,108,74,111,105,110,116,0,110,101,119,82,111,112,101,74,111,105,110,116,0,110,101,119,77,111,116,111,114,74,111,105,110,116,0,103,101,116,68,105,115,116,97,110,99,101,0,103,101,116,77,101,116,101,114,0,115,101,116,77,101,116,101,114,0,80,111,108,121,103,111,110,83,104,97,112,101,0,118,97,108,105,100,97,116,101,0,80,114,105,115,109,97,116,105,99,74,111,105,110,116,0,115,101,116,77,97,120,77,111,116,111,114,70,111,114,99,101,0,103,101,116,77,111,116,111,114,70,111,114,99,101,0,103,101,116,77,97,120,77,111,116,111,114,70,111,114,99,101,0,80,117,108,108,101,121,74,111,105,110,116,0,103,101,116,71,114,111,117,110,100,65,110,99,104,111,114,115,0,103,101,116,76,101,110,103,116,104,65,0,103,101,116,76,101,110,103,116,104,66,0,82,101,118,111,108,117,116,101,74,111,105,110,116,0,103,101,116,74,111,105,110,116,65,110,103,108,101,0,115,101,116,76,105,109,105,116,115,69,110,97,98,108,101,100,0,104,97,115,76,105,109,105,116,115,69,110,97,98,108,101,100,0,115,101,116,85,112,112,101,114,76,105,109,105,116,0,115,101,116,76,111,119,101,114,76,105,109,105,116,0,115,101,116,76,105,109,105,116,115,0,103,101,116,76,111,119,101,114,76,105,109,105,116,0,103,101,116,85,112,112,101,114,76,105,109,105,116,0,103,101,116,76,105,109,105,116,115,0,82,111,112,101,74,111,105,110,116,0,103,101,116,77,97,120,76,101,110,103,116,104,0,103,101,116,67,104,105,108,100,67,111,117,110,116,0,99,111,109,112,117,116,101,65,65,66,66,0,99,111,109,112,117,116,101,77,97,115,115,0,83,104,97,112,101,0,87,101,108,100,74,111,105,110,116,0,87,104,101,101,108,74,111,105,110,116,0,103,101,116,74,111,105,110,116,84,114,97,110,115,108,97,116,105,111,110,0,103,101,116,74,111,105,110,116,83,112,101,101,100,0,115,101,116,77,111,116,111,114,69,110,97,98,108,101,100,0,105,115,77,111,116,111,114,69,110,97,98,108,101,100,0,115,101,116,77,111,116,111,114,83,112,101,101,100,0,103,101,116,77,111,116,111,114,83,112,101,101,100,0,115,101,116,77,97,120,77,111,116,111,114,84,111,114,113,117,101,0,103,101,116,77,97,120,77,111,116,111,114,84,111,114,113,117,101,0,103,101,116,77,111,116,111,114,84,111,114,113,117,101,0,115,101,116,83,112,114,105,110,103,70,114,101,113,117,101,110,99,121,0,103,101,116,83,112,114,105,110,103,70,114,101,113,117,101,110,99,121,0,115,101,116,83,112,114,105,110,103,68,97,109,112,105,110,103,82,97,116,105,111,0,103,101,116,83,112,114,105,110,103,68,97,109,112,105,110,103,82,97,116,105,111,0,65,116,116,101,109,112,116,32,116,111,32,117,115,101,32,100,101,115,116,114,111,121,101,100,32,119,111,114,108,100,46,0,87,111,114,108,100,0,115,101,116,67,97,108,108,98,97,99,107,115,0,103,101,116,67,97,108,108,98,97,99,107,115,0,115,101,116,67,111,110,116,97,99,116,70,105,108,116,101,114,0,103,101,116,67,111,110,116,97,99,116,70,105,108,116,101,114,0,115,101,116,71,114,97,118,105,116,121,0,103,101,116,71,114,97,118,105,116,121,0,116,114,97,110,115,108,97,116,101,79,114,105,103,105,110,0,105,115,76,111,99,107,101,100,0,103,101,116,66,111,100,121,67,111,117,110,116,0,103,101,116,74,111,105,110,116,67,111,117,110,116,0,103,101,116,67,111,110,116,97,99,116,67,111,117,110,116,0,103,101,116,66,111,100,121,76,105,115,116,0,113,117,101,114,121,66,111,117,110,100,105,110,103,66,111,120,0,107,105,110,101,109,97,116,105,99,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,52,66,111,100,121,69,0,101,100,103,101,0,99,104,97,105,110,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,83,104,97,112,101,69,0,69,120,116,101,110,115,105,111,110,32,34,37,115,34,32,110,111,116,32,115,117,112,112,111,114,116,101,100,46,0,83,111,117,110,100,68,97,116,97,0,103,101,116,83,97,109,112,108,101,67,111,117,110,116,0,115,101,116,83,97,109,112,108,101,0,103,101,116,83,97,109,112,108,101,0,10,45,45,32,68,79,32,78,79,84,32,82,69,77,79,86,69,32,84,72,69,32,65,66,79,86,69,32,76,73,78,69,46,32,73,116,32,105,115,32,117,115,101,100,32,116,111,32,108,111,97,100,32,116,104,105,115,32,102,105,108,101,32,97,115,32,97,32,67,43,43,32,115,116,114,105,110,103,46,10,45,45,32,84,104,101,114,101,32,105,115,32,97,32,109,97,116,99,104,105,110,103,32,100,101,108,105,109,105,116,101,114,32,97,116,32,116,104,101,32,98,111,116,116,111,109,32,111,102,32,116,104,101,32,102,105,108,101,46,10,10,45,45,91,91,10,67,111,112,121,114,105,103,104,116,32,40,99,41,32,50,48,48,54,45,50,48,49,54,32,76,79,86,69,32,68,101,118,101,108,111,112,109,101,110,116,32,84,101,97,109,10,10,84,104,105,115,32,115,111,102,116,119,97,114,101,32,105,115,32,112,114,111,118,105,100,101,100,32,39,97,115,45,105,115,39,44,32,119,105,116,104,111,117,116,32,97,110,121,32,101,120,112,114,101,115,115,32,111,114,32,105,109,112,108,105,101,100,10,119,97,114,114,97,110,116,121,46,32,32,73,110,32,110,111,32,101,118,101,110,116,32,119,105,108,108,32,116,104,101,32,97,117,116,104,111,114,115,32,98,101,32,104,101,108,100,32,108,105,97,98,108,101,32,102,111,114,32,97,110,121,32,100,97,109,97,103,101,115,10,97,114,105,115,105,110,103,32,102,114,111,109,32,116,104,101,32,117,115,101,32,111,102,32,116,104,105,115,32,115,111,102,116,119,97,114,101,46,10,10,80,101,114,109,105,115,115,105,111,110,32,105,115,32,103,114,97,110,116,101,100,32,116,111,32,97,110,121,111,110,101,32,116,111,32,117,115,101,32,116,104,105,115,32,115,111,102,116,119,97,114,101,32,102,111,114,32,97,110,121,32,112,117,114,112,111,115,101,44,10,105,110,99,108,117,100,105,110,103,32,99,111,109,109,101,114,99,105,97,108,32,97,112,112,108,105,99,97,116,105,111,110,115,44,32,97,110,100,32,116,111,32,97,108,116,101,114,32,105,116,32,97,110,100,32,114,101,100,105,115,116,114,105,98,117,116,101,32,105,116,10,102,114,101,101,108,121,44,32,115,117,98,106,101,99,116,32,116,111,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,114,101,115,116,114,105,99,116,105,111,110,115,58,10,10,49,46,32,84,104,101,32,111,114,105,103,105,110,32,111,102,32,116,104,105,115,32,115,111,102,116,119,97,114,101,32,109,117,115,116,32,110,111,116,32,98,101,32,109,105,115,114,101,112,114,101,115,101,110,116,101,100,59,32,121,111,117,32,109,117,115,116,32,110,111,116,10,99,108,97,105,109,32,116,104,97,116,32,121,111,117,32,119,114,111,116,101,32,116,104,101,32,111,114,105,103,105,110,97,108,32,115,111,102,116,119,97,114,101,46,32,73,102,32,121,111,117,32,117,115,101,32,116,104,105,115,32,115,111,102,116,119,97,114,101,10,105,110,32,97,32,112,114,111,100,117,99,116,44,32,97,110,32,97,99,107,110,111,119,108,101,100,103,109,101,110,116,32,105,110,32,116,104,101,32,112,114,111,100,117,99,116,32,100,111,99,117,109,101,110,116,97,116,105,111,110,32,119,111,117,108,100,32,98,101,10,97,112,112,114,101,99,105,97,116,101,100,32,98,117,116,32,105,115,32,110,111,116,32,114,101,113,117,105,114,101,100,46,10,50,46,32,65,108,116,101,114,101,100,32,115,111,117,114,99,101,32,118,101,114,115,105,111,110,115,32,109,117,115,116,32,98,101,32,112,108,97,105,110,108,121,32,109,97,114,107,101,100,32,97,115,32,115,117,99,104,44,32,97,110,100,32,109,117,115,116,32,110,111,116,32,98,101,10,109,105,115,114,101,112,114,101,115,101,110,116,101,100,32,97,115,32,98,101,105,110,103,32,116,104,101,32,111,114,105,103,105,110,97,108,32,115,111,102,116,119,97,114,101,46,10,51,46,32,84,104,105,115,32,110,111,116,105,99,101,32,109,97,121,32,110,111,116,32,98,101,32,114,101,109,111,118,101,100,32,111,114,32,97,108,116,101,114,101,100,32,102,114,111,109,32,97,110,121,32,115,111,117,114,99,101,32,100,105,115,116,114,105,98,117,116,105,111,110,46,10,45,45,93,93,10,10,108,111,99,97,108,32,83,111,117,110,100,68,97,116,97,95,109,116,32,61,32,46,46,46,10,108,111,99,97,108,32,83,111,117,110,100,68,97,116,97,32,61,32,83,111,117,110,100,68,97,116,97,95,109,116,46,95,95,105,110,100,101,120,10,10,105,102,32,116,121,112,101,40,106,105,116,41,32,126,61,32,34,116,97,98,108,101,34,32,111,114,32,110,111,116,32,106,105,116,46,115,116,97,116,117,115,40,41,32,116,104,101,110,10,9,45,45,32,76,117,97,74,73,84,39,115,32,70,70,73,32,105,115,32,42,109,117,99,104,42,32,115,108,111,119,101,114,32,116,104,97,110,32,76,79,86,69,39,115,32,114,101,103,117,108,97,114,32,109,101,116,104,111,100,115,32,119,104,101,110,32,116,104,101,32,74,73,84,10,9,45,45,32,99,111,109,112,105,108,101,114,32,105,115,32,100,105,115,97,98,108,101,100,46,10,9,114,101,116,117,114,110,10,101,110,100,10,10,108,111,99,97,108,32,115,116,97,116,117,115,44,32,102,102,105,32,61,32,112,99,97,108,108,40,114,101,113,117,105,114,101,44,32,34,102,102,105,34,41,10,105,102,32,110,111,116,32,115,116,97,116,117,115,32,116,104,101,110,32,114,101,116,117,114,110,32,101,110,100,10,10,108,111,99,97,108,32,116,111,110,117,109,98,101,114,44,32,97,115,115,101,114,116,44,32,101,114,114,111,114,32,61,32,116,111,110,117,109,98,101,114,44,32,97,115,115,101,114,116,44,32,101,114,114,111,114,10,10,108,111,99,97,108,32,102,108,111,97,116,32,61,32,102,102,105,46,116,121,112,101,111,102,40,34,102,108,111,97,116,34,41,10,108,111,99,97,108,32,100,97,116,97,116,121,112,101,115,32,61,32,123,102,102,105,46,116,121,112,101,111,102,40,34,117,105,110,116,56,95,116,32,42,34,41,44,32,102,102,105,46,116,121,112,101,111,102,40,34,105,110,116,49,54,95,116,32,42,34,41,125,10,10,108,111,99,97,108,32,116,121,112,101,109,97,120,118,97,108,115,32,61,32,123,48,120,55,70,44,32,48,120,55,70,70,70,125,10,10,108,111,99,97,108,32,95,103,101,116,66,105,116,68,101,112,116,104,32,61,32,83,111,117,110,100,68,97,116,97,46,103,101,116,66,105,116,68,101,112,116,104,10,108,111,99,97,108,32,95,103,101,116,83,97,109,112,108,101,67,111,117,110,116,32,61,32,83,111,117,110,100,68,97,116,97,46,103,101,116,83,97,109,112,108,101,67,111,117,110,116,10,108,111,99,97,108,32,95,103,101,116,83,97,109,112,108,101,82,97,116,101,32,61,32,83,111,117,110,100,68,97,116,97,46,103,101,116,83,97,109,112,108,101,82,97,116,101,10,108,111,99,97,108,32,95,103,101,116,67,104,97,110,110,101,108,115,32,61,32,83,111,117,110,100,68,97,116,97,46,103,101,116,67,104,97,110,110,101,108,115,10,108,111,99,97,108,32,95,103,101,116,68,117,114,97,116,105,111,110,32,61,32,83,111,117,110,100,68,97,116,97,46,103,101,116,68,117,114,97,116,105,111,110,10,10,45,45,32,84,97,98,108,101,32,119,104,105,99,104,32,104,111,108,100,115,32,83,111,117,110,100,68,97,116,97,32,111,98,106,101,99,116,115,32,97,115,32,107,101,121,115,44,32,97,110,100,32,105,110,102,111,114,109,97,116,105,111,110,32,97,98,111,117,116,32,116,104,101,32,111,98,106,101,99,116,115,10,45,45,32,97,115,32,118,97,108,117,101,115,46,32,85,115,101,115,32,119,101,97,107,32,107,101,121,115,32,115,111,32,116,104,101,32,83,111,117,110,100,68,97,116,97,32,111,98,106,101,99,116,115,32,99,97,110,32,115,116,105,108,108,32,98,101,32,71,67,39,100,32,112,114,111,112,101,114,108,121,46,10,108,111,99,97,108,32,111,98,106,101,99,116,99,97,99,104,101,32,61,32,115,101,116,109,101,116,97,116,97,98,108,101,40,123,125,44,32,123,10,9,95,95,109,111,100,101,32,61,32,34,107,34,44,10,9,95,95,105,110,100,101,120,32,61,32,102,117,110,99,116,105,111,110,40,115,101,108,102,44,32,115,111,117,110,100,100,97,116,97,41,10,9,9,108,111,99,97,108,32,98,121,116,101,100,101,112,116,104,32,61,32,95,103,101,116,66,105,116,68,101,112,116,104,40,115,111,117,110,100,100,97,116,97,41,32,47,32,56,10,9,9,108,111,99,97,108,32,112,111,105,110,116,101,114,32,61,32,102,102,105,46,99,97,115,116,40,100,97,116,97,116,121,112,101,115,91,98,121,116,101,100,101,112,116,104,93,44,32,115,111,117,110,100,100,97,116,97,58,103,101,116,80,111,105,110,116,101,114,40,41,41,10,10,9,9,108,111,99,97,108,32,112,32,61,32,123,10,9,9,9,98,121,116,101,100,101,112,116,104,32,61,32,98,121,116,101,100,101,112,116,104,44,10,9,9,9,112,111,105,110,116,101,114,32,61,32,112,111,105,110,116,101,114,44,10,9,9,9,115,105,122,101,32,61,32,115,111,117,110,100,100,97,116,97,58,103,101,116,83,105,122,101,40,41,44,10,9,9,9,109,97,120,118,97,108,117,101,32,61,32,116,121,112,101,109,97,120,118,97,108,115,91,98,121,116,101,100,101,112,116,104,93,44,10,9,9,9,115,97,109,112,108,101,99,111,117,110,116,32,61,32,95,103,101,116,83,97,109,112,108,101,67,111,117,110,116,40,115,111,117,110,100,100,97,116,97,41,44,10,9,9,9,115,97,109,112,108,101,114,97,116,101,32,61,32,95,103,101,116,83,97,109,112,108,101,82,97,116,101,40,115,111,117,110,100,100,97,116,97,41,44,10,9,9,9,99,104,97,110,110,101,108,115,32,61,32,95,103,101,116,67,104,97,110,110,101,108,115,40,115,111,117,110,100,100,97,116,97,41,44,10,9,9,9,100,117,114,97,116,105,111,110,32,61,32,95,103,101,116,68,117,114,97,116,105,111,110,40,115,111,117,110,100,100,97,116,97,41,44,10,9,9,125,10,10,9,9,115,101,108,102,91,115,111,117,110,100,100,97,116,97,93,32,61,32,112,10,9,9,114,101,116,117,114,110,32,112,10,9,101,110,100,44,10,125,41,10,10,10,45,45,32,79,118,101,114,119,114,105,116,101,32,101,120,105,115,116,105,110,103,32,102,117,110,99,116,105,111,110,115,32,119,105,116,104,32,110,101,119,32,70,70,73,32,118,101,114,115,105,111,110,115,46,10,10,102,117,110,99,116,105,111,110,32,83,111,117,110,100,68,97,116,97,58,103,101,116,83,97,109,112,108,101,40,105,41,10,9,105,102,32,116,121,112,101,40,105,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,97,114,103,117,109,101,110,116,32,35,49,32,116,111,32,83,111,117,110,100,68,97,116,97,58,103,101,116,83,97,109,112,108,101,32,40,101,120,112,101,99,116,101,100,32,110,117,109,98,101,114,41,34,44,32,50,41,32,101,110,100,10,10,9,108,111,99,97,108,32,112,32,61,32,111,98,106,101,99,116,99,97,99,104,101,91,115,101,108,102,93,10,10,9,105,102,32,110,111,116,32,40,105,32,62,61,32,48,32,97,110,100,32,105,32,60,32,112,46,115,105,122,101,47,112,46,98,121,116,101,100,101,112,116,104,41,32,116,104,101,110,10,9,9,101,114,114,111,114,40,34,65,116,116,101,109,112,116,32,116,111,32,103,101,116,32,111,117,116,45,111,102,45,114,97,110,103,101,32,115,97,109,112,108,101,33,34,44,32,50,41,10,9,101,110,100,10,10,9,105,102,32,112,46,98,121,116,101,100,101,112,116,104,32,61,61,32,50,32,116,104,101,110,10,9,9,45,45,32,49,54,45,98,105,116,32,100,97,116,97,32,105,115,32,115,116,111,114,101,100,32,97,115,32,115,105,103,110,101,100,32,118,97,108,117,101,115,32,105,110,116,101,114,110,97,108,108,121,46,10,9,9,114,101,116,117,114,110,32,116,111,110,117,109,98,101,114,40,112,46,112,111,105,110,116,101,114,91,105,93,41,32,47,32,112,46,109,97,120,118,97,108,117,101,10,9,101,108,115,101,10,9,9,45,45,32,56,45,98,105,116,32,100,97,116,97,32,105,115,32,115,116,111,114,101,100,32,97,115,32,117,110,115,105,103,110,101,100,32,118,97,108,117,101,115,32,105,110,116,101,114,110,97,108,108,121,46,10,9,9,114,101,116,117,114,110,32,40,116,111,110,117,109,98,101,114,40,112,46,112,111,105,110,116,101,114,91,105,93,41,32,45,32,49,50,56,41,32,47,32,49,50,55,10,9,101,110,100,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,83,111,117,110,100,68,97,116,97,58,115,101,116,83,97,109,112,108,101,40,105,44,32,115,97,109,112,108,101,41,10,9,105,102,32,116,121,112,101,40,105,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,97,114,103,117,109,101,110,116,32,35,49,32,116,111,32,83,111,117,110,100,68,97,116,97,58,115,101,116,83,97,109,112,108,101,32,40,101,120,112,101,99,116,101,100,32,110,117,109,98,101,114,41,34,44,32,50,41,32,101,110,100,10,9,105,102,32,116,121,112,101,40,115,97,109,112,108,101,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,97,114,103,117,109,101,110,116,32,35,50,32,116,111,32,83,111,117,110,100,68,97,116,97,58,115,101,116,83,97,109,112,108,101,32,40,101,120,112,101,99,116,101,100,32,110,117,109,98,101,114,41,34,44,32,50,41,32,101,110,100,10,10,9,108,111,99,97,108,32,112,32,61,32,111,98,106,101,99,116,99,97,99,104,101,91,115,101,108,102,93,10,10,9,105,102,32,110,111,116,32,40,105,32,62,61,32,48,32,97,110,100,32,105,32,60,32,112,46,115,105,122,101,47,112,46,98,121,116,101,100,101,112,116,104,41,32,116,104,101,110,10,9,9,101,114,114,111,114,40,34,65,116,116,101,109,112,116,32,116,111,32,115,101,116,32,111,117,116,45,111,102,45,114,97,110,103,101,32,115,97,109,112,108,101,33,34,44,32,50,41,10,9,101,110,100,10,10,9,105,102,32,112,46,98,121,116,101,100,101,112,116,104,32,61,61,32,50,32,116,104,101,110,10,9,9,45,45,32,49,54,45,98,105,116,32,100,97,116,97,32,105,115,32,115,116,111,114,101,100,32,97,115,32,115,105,103,110,101,100,32,118,97,108,117,101,115,32,105,110,116,101,114,110,97,108,108,121,46,10,9,9,112,46,112,111,105,110,116,101,114,91,105,93,32,61,32,115,97,109,112,108,101,32,42,32,112,46,109,97,120,118,97,108,117,101,10,9,101,108,115,101,10,9,9,45,45,32,56,45,98,105,116,32,100,97,116,97,32,105,115,32,115,116,111,114,101,100,32,97,115,32,117,110,115,105,103,110,101,100,32,118,97,108,117,101,115,32,105,110,116,101,114,110,97,108,108,121,46,10,9,9,45,45,32,84,104,101,32,102,108,111,97,116,32,99,97,115,116,32,105,115,32,110,101,101,100,101,100,32,116,111,32,109,97,107,101,32,118,97,108,117,101,115,32,101,110,100,32,117,112,32,116,104,101,32,115,97,109,101,32,97,115,32,105,110,32,116,104,101,10,9,9,45,45,32,67,43,43,32,118,101,114,115,105,111,110,32,111,102,32,116,104,105,115,32,109,101,116,104,111,100,46,10,9,9,112,46,112,111,105,110,116,101,114,91,105,93,32,61,32,102,102,105,46,99,97,115,116,40,102,108,111,97,116,44,32,40,115,97,109,112,108,101,32,42,32,49,50,55,41,32,43,32,49,50,56,41,10,9,101,110,100,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,83,111,117,110,100,68,97,116,97,58,103,101,116,66,105,116,68,101,112,116,104,40,41,10,9,114,101,116,117,114,110,32,111,98,106,101,99,116,99,97,99,104,101,91,115,101,108,102,93,46,98,121,116,101,100,101,112,116,104,32,42,32,56,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,83,111,117,110,100,68,97,116,97,58,103,101,116,83,97,109,112,108,101,67,111,117,110,116,40,41,10,9,114,101,116,117,114,110,32,111,98,106,101,99,116,99,97,99,104,101,91,115,101,108,102,93,46,115,97,109,112,108,101,99,111,117,110,116,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,83,111,117,110,100,68,97,116,97,58,103,101,116,83,97,109,112,108,101,82,97,116,101,40,41,10,9,114,101,116,117,114,110,32,111,98,106,101,99,116,99,97,99,104,101,91,115,101,108,102,93,46,115,97,109,112,108,101,114,97,116,101,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,83,111,117,110,100,68,97,116,97,58,103,101,116,67,104,97,110,110,101,108,115,40,41,10,9,114,101,116,117,114,110,32,111,98,106,101,99,116,99,97,99,104,101,91,115,101,108,102,93,46,99,104,97,110,110,101,108,115,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,83,111,117,110,100,68,97,116,97,58,103,101,116,68,117,114,97,116,105,111,110,40,41,10,9,114,101,116,117,114,110,32,111,98,106,101,99,116,99,97,99,104,101,91,115,101,108,102,93,46,100,117,114,97,116,105,111,110,10,101,110,100,10,10,45,45,32,68,79,32,78,79,84,32,82,69,77,79,86,69,32,84,72,69,32,78,69,88,84,32,76,73,78,69,46,32,73,116,32,105,115,32,117,115,101,100,32,116,111,32,108,111,97,100,32,116,104,105,115,32,102,105,108,101,32,97,115,32,97,32,67,43,43,32,115,116,114,105,110,103,46,10,45,45,0,83,111,117,110,100,68,97,116,97,46,108,117,97,0,103,101,116,79,83,0,103,101,116,80,114,111,99,101,115,115,111,114,67,111,117,110,116,0,115,101,116,67,108,105,112,98,111,97,114,100,84,101,120,116,0,103,101,116,67,108,105,112,98,111,97,114,100,84,101,120,116,0,103,101,116,80,111,119,101,114,73,110,102,111,0,111,112,101,110,85,82,76,0,118,105,98,114,97,116,101,0,84,104,114,101,97,100,32,99,111,100,101,0,110,101,119,84,104,114,101,97,100,0,110,101,119,67,104,97,110,110,101,108,0,103,101,116,67,104,97,110,110,101,108,0,116,105,109,101,114,0,103,101,116,68,101,108,116,97,0,103,101,116,70,80,83,0,103,101,116,65,118,101,114,97,103,101,68,101,108,116,97,0,103,101,116,84,105,109,101,0,116,111,117,99,104,32,105,100,0,116,111,117,99,104,0,103,101,116,84,111,117,99,104,101,115,0,103,101,116,80,114,101,115,115,117,114,101,0,70,105,108,101,32,105,115,32,110,111,116,32,111,112,101,110,32,97,110,100,32,99,97,110,110,111,116,32,98,101,32,111,112,101,110,101,100,0,83,111,117,114,99,101,32,111,114,32,86,105,100,101,111,83,116,114,101,97,109,32,111,114,32,110,105,108,0,86,105,100,101,111,83,116,114,101,97,109,0,115,101,116,83,121,110,99,0,73,110,118,97,108,105,100,32,119,105,110,100,111,119,32,115,101,116,116,105,110,103,58,32,37,115,0,73,110,118,97,108,105,100,32,102,117,108,108,115,99,114,101,101,110,32,116,121,112,101,58,32,37,115,0,85,110,107,110,111,119,110,32,102,117,108,108,115,99,114,101,101,110,32,116,121,112,101,46,0,77,117,115,116,32,104,97,118,101,32,97,116,32,108,101,97,115,116,32,111,110,101,32,109,101,115,115,97,103,101,98,111,120,32,98,117,116,116,111,110,46,0,101,110,116,101,114,98,117,116,116,111,110,0,101,115,99,97,112,101,98,117,116,116,111,110,0,73,110,118,97,108,105,100,32,109,101,115,115,97,103,101,98,111,120,32,116,121,112,101,58,32,37,115,0,103,101,116,68,105,115,112,108,97,121,67,111,117,110,116,0,103,101,116,68,105,115,112,108,97,121,78,97,109,101,0,115,101,116,77,111,100,101,0,103,101,116,70,117,108,108,115,99,114,101,101,110,77,111,100,101,115,0,115,101,116,70,117,108,108,115,99,114,101,101,110,0,103,101,116,70,117,108,108,115,99,114,101,101,110,0,103,101,116,68,101,115,107,116,111,112,68,105,109,101,110,115,105,111,110,115,0,115,101,116,73,99,111,110,0,103,101,116,73,99,111,110,0,115,101,116,68,105,115,112,108,97,121,83,108,101,101,112,69,110,97,98,108,101,100,0,105,115,68,105,115,112,108,97,121,83,108,101,101,112,69,110,97,98,108,101,100,0,115,101,116,84,105,116,108,101,0,103,101,116,84,105,116,108,101,0,104,97,115,70,111,99,117,115,0,104,97,115,77,111,117,115,101,70,111,99,117,115,0,105,115,86,105,115,105,98,108,101,0,103,101,116,80,105,120,101,108,83,99,97,108,101,0,116,111,80,105,120,101,108,115,0,102,114,111,109,80,105,120,101,108,115,0,109,105,110,105,109,105,122,101,0,109,97,120,105,109,105,122,101,0,115,104,111,119,77,101,115,115,97,103,101,66,111,120,0,114,101,113,117,101,115,116,65,116,116,101,110,116,105,111,110,0,101,110,101,116,95,104,111,115,116,0,115,101,114,118,105,99,101,0,84,114,105,101,100,32,116,111,32,105,110,100,101,120,32,97,32,110,105,108,32,104,111,115,116,33,0,69,114,114,111,114,32,100,117,114,105,110,103,32,115,101,114,118,105,99,101,0,101,110,101,116,95,112,101,101,114,115,0,101,110,101,116,95,112,101,101,114,0,112,101,101,114,0,100,105,115,99,111,110,110,101,99,116,0,99,104,101,99,107,95,101,118,101,110,116,115,0,69,114,114,111,114,32,99,104,101,99,107,105,110,103,32,101,118,101,110,116,0,99,111,109,112,114,101,115,115,95,119,105,116,104,95,114,97,110,103,101,95,99,111,100,101,114,0,72,111,115,116,110,97,109,101,32,116,111,111,32,108,111,110,103,0,70,97,105,108,101,100,32,116,111,32,112,97,114,115,101,32,97,100,100,114,101,115,115,0,77,105,115,115,105,110,103,32,112,111,114,116,32,105,110,32,97,100,100,114,101,115,115,0,70,97,105,108,101,100,32,116,111,32,114,101,115,111,108,118,101,32,104,111,115,116,32,110,97,109,101,0,70,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,112,101,101,114,0,117,110,115,101,113,117,101,110,99,101,100,0,114,101,108,105,97,98,108,101,0,117,110,114,101,108,105,97,98,108,101,0,85,110,107,110,111,119,110,32,112,97,99,107,101,116,32,102,108,97,103,58,32,37,115,0,70,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,112,97,99,107,101,116,0,99,104,97,110,110,101,108,95,108,105,109,105,116,0,98,97,110,100,119,105,100,116,104,95,108,105,109,105,116,0,103,101,116,95,115,111,99,107,101,116,95,97,100,100,114,101,115,115,0,37,100,46,37,100,46,37,100,46,37,100,58,37,100,0,115,111,99,107,101,116,95,103,101,116,95,97,100,100,114,101,115,115,0,116,111,116,97,108,95,115,101,110,116,95,100,97,116,97,0,116,111,116,97,108,95,114,101,99,101,105,118,101,100,95,100,97,116,97,0,115,101,114,118,105,99,101,95,116,105,109,101,0,112,101,101,114,95,99,111,117,110,116,0,103,101,116,95,112,101,101,114,0,73,110,118,97,108,105,100,32,112,101,101,114,32,105,110,100,101,120,0,100,105,115,99,111,110,110,101,99,116,95,110,111,119,0,100,105,115,99,111,110,110,101,99,116,95,108,97,116,101,114,0,112,105,110,103,0,116,104,114,111,116,116,108,101,95,99,111,110,102,105,103,117,114,101,0,112,105,110,103,95,105,110,116,101,114,118,97,108,0,101,110,101,116,58,32,99,111,117,108,100,32,110,111,116,32,102,105,110,100,32,112,101,101,114,32,105,100,33,0,115,116,97,116,101,0,100,105,115,99,111,110,110,101,99,116,101,100,0,99,111,110,110,101,99,116,105,110,103,0,97,99,107,110,111,119,108,101,100,103,105,110,103,95,99,111,110,110,101,99,116,0,99,111,110,110,101,99,116,105,111,110,95,112,101,110,100,105,110,103,0,99,111,110,110,101,99,116,105,111,110,95,115,117,99,99,101,101,100,101,100,0,99,111,110,110,101,99,116,101,100,0,100,105,115,99,111,110,110,101,99,116,105,110,103,0,97,99,107,110,111,119,108,101,100,103,105,110,103,95,100,105,115,99,111,110,110,101,99,116,0,122,111,109,98,105,101,0,99,111,110,110,101,99,116,95,105,100,0,114,111,117,110,100,95,116,114,105,112,95,116,105,109,101,0,108,97,115,116,95,114,111,117,110,100,95,116,114,105,112,95,116,105,109,101,0,101,110,101,116,0,104,111,115,116,95,99,114,101,97,116,101,0,101,110,101,116,58,32,102,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,104,111,115,116,32,40,97,108,114,101,97,100,121,32,108,105,115,116,101,110,105,110,103,63,41,0,108,105,110,107,101,100,95,118,101,114,115,105,111,110,0,37,100,46,37,100,46,37,100,0,115,111,99,107,101,116,46,99,111,114,101,0,109,105,109,101,46,99,111,114,101,0,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,76,117,97,83,111,99,107,101,116,32,104,101,108,112,101,114,32,109,111,100,117,108,101,10,45,45,32,65,117,116,104,111,114,58,32,68,105,101,103,111,32,78,101,104,97,98,10,45,45,32,82,67,83,32,73,68,58,32,36,73,100,58,32,115,111,99,107,101,116,46,108,117,97,44,118,32,49,46,50,50,32,50,48,48,53,47,49,49,47,50,50,32,48,56,58,51,51,58,50,57,32,100,105,101,103,111,32,69,120,112,32,36,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,68,101,99,108,97,114,101,32,109,111,100,117,108,101,32,97,110,100,32,105,109,112,111,114,116,32,100,101,112,101,110,100,101,110,99,105,101,115,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,108,111,99,97,108,32,98,97,115,101,32,61,32,95,71,10,108,111,99,97,108,32,115,116,114,105,110,103,32,61,32,114,101,113,117,105,114,101,40,34,115,116,114,105,110,103,34,41,10,108,111,99,97,108,32,109,97,116,104,32,61,32,114,101,113,117,105,114,101,40,34,109,97,116,104,34,41,10,108,111,99,97,108,32,115,111,99,107,101,116,32,61,32,114,101,113,117,105,114,101,40,34,115,111,99,107,101,116,46,99,111,114,101,34,41,10,109,111,100,117,108,101,40,34,115,111,99,107,101,116,34,41,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,69,120,112,111,114,116,101,100,32,97,117,120,105,108,105,97,114,32,102,117,110,99,116,105,111,110,115,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,102,117,110,99,116,105,111,110,32,99,111,110,110,101,99,116,40,97,100,100,114,101,115,115,44,32,112,111,114,116,44,32,108,97,100,100,114,101,115,115,44,32,108,112,111,114,116,41,10,32,32,32,32,108,111,99,97,108,32,115,111,99,107,44,32,101,114,114,32,61,32,115,111,99,107,101,116,46,116,99,112,40,41,10,32,32,32,32,105,102,32,110,111,116,32,115,111,99,107,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,32,101,110,100,10,32,32,32,32,105,102,32,108,97,100,100,114,101,115,115,32,116,104,101,110,10,32,32,32,32,32,32,32,32,108,111,99,97,108,32,114,101,115,44,32,101,114,114,32,61,32,115,111,99,107,58,98,105,110,100,40,108,97,100,100,114,101,115,115,44,32,108,112,111,114,116,44,32,45,49,41,10,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,114,101,115,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,32,101,110,100,10,32,32,32,32,101,110,100,10,32,32,32,32,108,111,99,97,108,32,114,101,115,44,32,101,114,114,32,61,32,115,111,99,107,58,99,111,110,110,101,99,116,40,97,100,100,114,101,115,115,44,32,112,111,114,116,41,10,32,32,32,32,105,102,32,110,111,116,32,114,101,115,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,32,101,110,100,10,32,32,32,32,114,101,116,117,114,110,32,115,111,99,107,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,98,105,110,100,40,104,111,115,116,44,32,112,111,114,116,44,32,98,97,99,107,108,111,103,41,10,32,32,32,32,108,111,99,97,108,32,115,111,99,107,44,32,101,114,114,32,61,32,115,111,99,107,101,116,46,116,99,112,40,41,10,32,32,32,32,105,102,32,110,111,116,32,115,111,99,107,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,32,101,110,100,10,32,32,32,32,115,111,99,107,58,115,101,116,111,112,116,105,111,110,40,34,114,101,117,115,101,97,100,100,114,34,44,32,116,114,117,101,41,10,32,32,32,32,108,111,99,97,108,32,114,101,115,44,32,101,114,114,32,61,32,115,111,99,107,58,98,105,110,100,40,104,111,115,116,44,32,112,111,114,116,41,10,32,32,32,32,105,102,32,110,111,116,32,114,101,115,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,32,101,110,100,10,32,32,32,32,114,101,115,44,32,101,114,114,32,61,32,115,111,99,107,58,108,105,115,116,101,110,40,98,97,99,107,108,111,103,41,10,32,32,32,32,105,102,32,110,111,116,32,114,101,115,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,32,101,110,100,10,32,32,32,32,114,101,116,117,114,110,32,115,111,99,107,10,101,110,100,10,10,116,114,121,32,61,32,110,101,119,116,114,121,40,41,10,10,102,117,110,99,116,105,111,110,32,99,104,111,111,115,101,40,116,97,98,108,101,41,10,32,32,32,32,114,101,116,117,114,110,32,102,117,110,99,116,105,111,110,40,110,97,109,101,44,32,111,112,116,49,44,32,111,112,116,50,41,10,32,32,32,32,32,32,32,32,105,102,32,98,97,115,101,46,116,121,112,101,40,110,97,109,101,41,32,126,61,32,34,115,116,114,105,110,103,34,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,110,97,109,101,44,32,111,112,116,49,44,32,111,112,116,50,32,61,32,34,100,101,102,97,117,108,116,34,44,32,110,97,109,101,44,32,111,112,116,49,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,108,111,99,97,108,32,102,32,61,32,116,97,98,108,101,91,110,97,109,101,32,111,114,32,34,110,105,108,34,93,10,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,102,32,116,104,101,110,32,98,97,115,101,46,101,114,114,111,114,40,34,117,110,107,110,111,119,110,32,107,101,121,32,40,34,46,46,32,98,97,115,101,46,116,111,115,116,114,105,110,103,40,110,97,109,101,41,32,46,46,34,41,34,44,32,51,41,10,32,32,32,32,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,102,40,111,112,116,49,44,32,111,112,116,50,41,32,101,110,100,10,32,32,32,32,101,110,100,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,83,111,99,107,101,116,32,115,111,117,114,99,101,115,32,97,110,100,32,115,105,110,107,115,44,32,99,111,110,102,111,114,109,105,110,103,32,116,111,32,76,84,78,49,50,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,99,114,101,97,116,101,32,110,97,109,101,115,112,97,99,101,115,32,105,110,115,105,100,101,32,76,117,97,83,111,99,107,101,116,32,110,97,109,101,115,112,97,99,101,10,115,111,117,114,99,101,116,32,61,32,123,125,10,115,105,110,107,116,32,61,32,123,125,10,10,66,76,79,67,75,83,73,90,69,32,61,32,50,48,52,56,10,10,115,105,110,107,116,91,34,99,108,111,115,101,45,119,104,101,110,45,100,111,110,101,34,93,32,61,32,102,117,110,99,116,105,111,110,40,115,111,99,107,41,10,32,32,32,32,114,101,116,117,114,110,32,98,97,115,101,46,115,101,116,109,101,116,97,116,97,98,108,101,40,123,10,32,32,32,32,32,32,32,32,103,101,116,102,100,32,61,32,102,117,110,99,116,105,111,110,40,41,32,114,101,116,117,114,110,32,115,111,99,107,58,103,101,116,102,100,40,41,32,101,110,100,44,10,32,32,32,32,32,32,32,32,100,105,114,116,121,32,61,32,102,117,110,99,116,105,111,110,40,41,32,114,101,116,117,114,110,32,115,111,99,107,58,100,105,114,116,121,40,41,32,101,110,100,10,32,32,32,32,125,44,32,123,10,32,32,32,32,32,32,32,32,95,95,99,97,108,108,32,61,32,102,117,110,99,116,105,111,110,40,115,101,108,102,44,32,99,104,117,110,107,44,32,101,114,114,41,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,99,104,117,110,107,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,115,111,99,107,58,99,108,111,115,101,40,41,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,49,10,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,115,111,99,107,58,115,101,110,100,40,99,104,117,110,107,41,32,101,110,100,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,125,41,10,101,110,100,10,10,115,105,110,107,116,91,34,107,101,101,112,45,111,112,101,110,34,93,32,61,32,102,117,110,99,116,105,111,110,40,115,111,99,107,41,10,32,32,32,32,114,101,116,117,114,110,32,98,97,115,101,46,115,101,116,109,101,116,97,116,97,98,108,101,40,123,10,32,32,32,32,32,32,32,32,103,101,116,102,100,32,61,32,102,117,110,99,116,105,111,110,40,41,32,114,101,116,117,114,110,32,115,111,99,107,58,103,101,116,102,100,40,41,32,101,110,100,44,10,32,32,32,32,32,32,32,32,100,105,114,116,121,32,61,32,102,117,110,99,116,105,111,110,40,41,32,114,101,116,117,114,110,32,115,111,99,107,58,100,105,114,116,121,40,41,32,101,110,100,10,32,32,32,32,125,44,32,123,10,32,32,32,32,32,32,32,32,95,95,99,97,108,108,32,61,32,102,117,110,99,116,105,111,110,40,115,101,108,102,44,32,99,104,117,110,107,44,32,101,114,114,41,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,99,104,117,110,107,32,116,104,101,110,32,114,101,116,117,114,110,32,115,111,99,107,58,115,101,110,100,40,99,104,117,110,107,41,10,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,49,32,101,110,100,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,125,41,10,101,110,100,10,10,115,105,110,107,116,91,34,100,101,102,97,117,108,116,34,93,32,61,32,115,105,110,107,116,91,34,107,101,101,112,45,111,112,101,110,34,93,10,10,115,105,110,107,32,61,32,99,104,111,111,115,101,40,115,105,110,107,116,41,10,10,115,111,117,114,99,101,116,91,34,98,121,45,108,101,110,103,116,104,34,93,32,61,32,102,117,110,99,116,105,111,110,40,115,111,99,107,44,32,108,101,110,103,116,104,41,10,32,32,32,32,114,101,116,117,114,110,32,98,97,115,101,46,115,101,116,109,101,116,97,116,97,98,108,101,40,123,10,32,32,32,32,32,32,32,32,103,101,116,102,100,32,61,32,102,117,110,99,116,105,111,110,40,41,32,114,101,116,117,114,110,32,115,111,99,107,58,103,101,116,102,100,40,41,32,101,110,100,44,10,32,32,32,32,32,32,32,32,100,105,114,116,121,32,61,32,102,117,110,99,116,105,111,110,40,41,32,114,101,116,117,114,110,32,115,111,99,107,58,100,105,114,116,121,40,41,32,101,110,100,10,32,32,32,32,125,44,32,123,10,32,32,32,32,32,32,32,32,95,95,99,97,108,108,32,61,32,102,117,110,99,116,105,111,110,40,41,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,108,101,110,103,116,104,32,60,61,32,48,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,108,111,99,97,108,32,115,105,122,101,32,61,32,109,97,116,104,46,109,105,110,40,115,111,99,107,101,116,46,66,76,79,67,75,83,73,90,69,44,32,108,101,110,103,116,104,41,10,32,32,32,32,32,32,32,32,32,32,32,32,108,111,99,97,108,32,99,104,117,110,107,44,32,101,114,114,32,61,32,115,111,99,107,58,114,101,99,101,105,118,101,40,115,105,122,101,41,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,101,114,114,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,108,101,110,103,116,104,32,61,32,108,101,110,103,116,104,32,45,32,115,116,114,105,110,103,46,108,101,110,40,99,104,117,110,107,41,10,32,32,32,32,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,99,104,117,110,107,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,125,41,10,101,110,100,10,10,115,111,117,114,99,101,116,91,34,117,110,116,105,108,45,99,108,111,115,101,100,34,93,32,61,32,102,117,110,99,116,105,111,110,40,115,111,99,107,41,10,32,32,32,32,108,111,99,97,108,32,100,111,110,101,10,32,32,32,32,114,101,116,117,114,110,32,98,97,115,101,46,115,101,116,109,101,116,97,116,97,98,108,101,40,123,10,32,32,32,32,32,32,32,32,103,101,116,102,100,32,61,32,102,117,110,99,116,105,111,110,40,41,32,114,101,116,117,114,110,32,115,111,99,107,58,103,101,116,102,100,40,41,32,101,110,100,44,10,32,32,32,32,32,32,32,32,100,105,114,116,121,32,61,32,102,117,110,99,116,105,111,110,40,41,32,114,101,116,117,114,110,32,115,111,99,107,58,100,105,114,116,121,40,41,32,101,110,100,10,32,32,32,32,125,44,32,123,10,32,32,32,32,32,32,32,32,95,95,99,97,108,108,32,61,32,102,117,110,99,116,105,111,110,40,41,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,100,111,110,101,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,108,111,99,97,108,32,99,104,117,110,107,44,32,101,114,114,44,32,112,97,114,116,105,97,108,32,61,32,115,111,99,107,58,114,101,99,101,105,118,101,40,115,111,99,107,101,116,46,66,76,79,67,75,83,73,90,69,41,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,101,114,114,32,116,104,101,110,32,114,101,116,117,114,110,32,99,104,117,110,107,10,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,105,102,32,101,114,114,32,61,61,32,34,99,108,111,115,101,100,34,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,115,111,99,107,58,99,108,111,115,101,40,41,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,100,111,110,101,32,61,32,49,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,112,97,114,116,105,97,108,10,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,32,101,110,100,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,125,41,10,101,110,100,10,10,10,115,111,117,114,99,101,116,91,34,100,101,102,97,117,108,116,34,93,32,61,32,115,111,117,114,99,101,116,91,34,117,110,116,105,108,45,99,108,111,115,101,100,34,93,10,10,115,111,117,114,99,101,32,61,32,99,104,111,111,115,101,40,115,111,117,114,99,101,116,41,10,10,115,111,99,107,101,116,46,108,117,97,0,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,70,84,80,32,115,117,112,112,111,114,116,32,102,111,114,32,116,104,101,32,76,117,97,32,108,97,110,103,117,97,103,101,10,45,45,32,76,117,97,83,111,99,107,101,116,32,116,111,111,108,107,105,116,46,10,45,45,32,65,117,116,104,111,114,58,32,68,105,101,103,111,32,78,101,104,97,98,10,45,45,32,82,67,83,32,73,68,58,32,36,73,100,58,32,102,116,112,46,108,117,97,44,118,32,49,46,52,53,32,50,48,48,55,47,48,55,47,49,49,32,49,57,58,50,53,58,52,55,32,100,105,101,103,111,32,69,120,112,32,36,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,68,101,99,108,97,114,101,32,109,111,100,117,108,101,32,97,110,100,32,105,109,112,111,114,116,32,100,101,112,101,110,100,101,110,99,105,101,115,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,108,111,99,97,108,32,98,97,115,101,32,61,32,95,71,10,108,111,99,97,108,32,116,97,98,108,101,32,61,32,114,101,113,117,105,114,101,40,34,116,97,98,108,101,34,41,10,108,111,99,97,108,32,115,116,114,105,110,103,32,61,32,114,101,113,117,105,114,101,40,34,115,116,114,105,110,103,34,41,10,108,111,99,97,108,32,109,97,116,104,32,61,32,114,101,113,117,105,114,101,40,34,109,97,116,104,34,41,10,108,111,99,97,108,32,115,111,99,107,101,116,32,61,32,114,101,113,117,105,114,101,40,34,115,111,99,107,101,116,34,41,10,108,111,99,97,108,32,117,114,108,32,61,32,114,101,113,117,105,114,101,40,34,115,111,99,107,101,116,46,117,114,108,34,41,10,108,111,99,97,108,32,116,112,32,61,32,114,101,113,117,105,114,101,40,34,115,111,99,107,101,116,46,116,112,34,41,10,108,111,99,97,108,32,108,116,110,49,50,32,61,32,114,101,113,117,105,114,101,40,34,108,116,110,49,50,34,41,10,109,111,100,117,108,101,40,34,115,111,99,107,101,116,46,102,116,112,34,41,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,80,114,111,103,114,97,109,32,99,111,110,115,116,97,110,116,115,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,116,105,109,101,111,117,116,32,105,110,32,115,101,99,111,110,100,115,32,98,101,102,111,114,101,32,116,104,101,32,112,114,111,103,114,97,109,32,103,105,118,101,115,32,117,112,32,111,110,32,97,32,99,111,110,110,101,99,116,105,111,110,10,84,73,77,69,79,85,84,32,61,32,54,48,10,45,45,32,100,101,102,97,117,108,116,32,112,111,114,116,32,102,111,114,32,102,116,112,32,115,101,114,118,105,99,101,10,80,79,82,84,32,61,32,50,49,10,45,45,32,116,104,105,115,32,105,115,32,116,104,101,32,100,101,102,97,117,108,116,32,97,110,111,110,121,109,111,117,115,32,112,97,115,115,119,111,114,100,46,32,117,115,101,100,32,119,104,101,110,32,110,111,32,112,97,115,115,119,111,114,100,32,105,115,10,45,45,32,112,114,111,118,105,100,101,100,32,105,110,32,117,114,108,46,32,115,104,111,117,108,100,32,98,101,32,99,104,97,110,103,101,100,32,116,111,32,121,111,117,114,32,101,45,109,97,105,108,46,10,85,83,69,82,32,61,32,34,102,116,112,34,10,80,65,83,83,87,79,82,68,32,61,32,34,97,110,111,110,121,109,111,117,115,64,97,110,111,110,121,109,111,117,115,46,111,114,103,34,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,76,111,119,32,108,101,118,101,108,32,70,84,80,32,65,80,73,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,108,111,99,97,108,32,109,101,116,97,116,32,61,32,123,32,95,95,105,110,100,101,120,32,61,32,123,125,32,125,10,10,102,117,110,99,116,105,111,110,32,111,112,101,110,40,115,101,114,118,101,114,44,32,112,111,114,116,44,32,99,114,101,97,116,101,41,10,32,32,32,32,108,111,99,97,108,32,116,112,32,61,32,115,111,99,107,101,116,46,116,114,121,40,116,112,46,99,111,110,110,101,99,116,40,115,101,114,118,101,114,44,32,112,111,114,116,32,111,114,32,80,79,82,84,44,32,84,73,77,69,79,85,84,44,32,99,114,101,97,116,101,41,41,10,32,32,32,32,108,111,99,97,108,32,102,32,61,32,98,97,115,101,46,115,101,116,109,101,116,97,116,97,98,108,101,40,123,32,116,112,32,61,32,116,112,32,125,44,32,109,101,116,97,116,41,10,32,32,32,32,45,45,32,109,97,107,101,32,115,117,114,101,32,101,118,101,114,121,116,104,105,110,103,32,103,101,116,115,32,99,108,111,115,101,100,32,105,110,32,97,110,32,101,120,99,101,112,116,105,111,110,10,32,32,32,32,102,46,116,114,121,32,61,32,115,111,99,107,101,116,46,110,101,119,116,114,121,40,102,117,110,99,116,105,111,110,40,41,32,102,58,99,108,111,115,101,40,41,32,101,110,100,41,10,32,32,32,32,114,101,116,117,114,110,32,102,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,112,111,114,116,99,111,110,110,101,99,116,40,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,115,101,114,118,101,114,58,115,101,116,116,105,109,101,111,117,116,40,84,73,77,69,79,85,84,41,41,10,32,32,32,32,115,101,108,102,46,100,97,116,97,32,61,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,115,101,114,118,101,114,58,97,99,99,101,112,116,40,41,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,100,97,116,97,58,115,101,116,116,105,109,101,111,117,116,40,84,73,77,69,79,85,84,41,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,112,97,115,118,99,111,110,110,101,99,116,40,41,10,32,32,32,32,115,101,108,102,46,100,97,116,97,32,61,32,115,101,108,102,46,116,114,121,40,115,111,99,107,101,116,46,116,99,112,40,41,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,100,97,116,97,58,115,101,116,116,105,109,101,111,117,116,40,84,73,77,69,79,85,84,41,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,100,97,116,97,58,99,111,110,110,101,99,116,40,115,101,108,102,46,112,97,115,118,116,46,105,112,44,32,115,101,108,102,46,112,97,115,118,116,46,112,111,114,116,41,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,108,111,103,105,110,40,117,115,101,114,44,32,112,97,115,115,119,111,114,100,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,34,117,115,101,114,34,44,32,117,115,101,114,32,111,114,32,85,83,69,82,41,41,10,32,32,32,32,108,111,99,97,108,32,99,111,100,101,44,32,114,101,112,108,121,32,61,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,123,34,50,46,46,34,44,32,51,51,49,125,41,10,32,32,32,32,105,102,32,99,111,100,101,32,61,61,32,51,51,49,32,116,104,101,110,10,32,32,32,32,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,34,112,97,115,115,34,44,32,112,97,115,115,119,111,114,100,32,111,114,32,80,65,83,83,87,79,82,68,41,41,10,32,32,32,32,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,50,46,46,34,41,41,10,32,32,32,32,101,110,100,10,32,32,32,32,114,101,116,117,114,110,32,49,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,112,97,115,118,40,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,34,112,97,115,118,34,41,41,10,32,32,32,32,108,111,99,97,108,32,99,111,100,101,44,32,114,101,112,108,121,32,61,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,50,46,46,34,41,41,10,32,32,32,32,108,111,99,97,108,32,112,97,116,116,101,114,110,32,61,32,34,40,37,100,43,41,37,68,40,37,100,43,41,37,68,40,37,100,43,41,37,68,40,37,100,43,41,37,68,40,37,100,43,41,37,68,40,37,100,43,41,34,10,32,32,32,32,108,111,99,97,108,32,97,44,32,98,44,32,99,44,32,100,44,32,112,49,44,32,112,50,32,61,32,115,111,99,107,101,116,46,115,107,105,112,40,50,44,32,115,116,114,105,110,103,46,102,105,110,100,40,114,101,112,108,121,44,32,112,97,116,116,101,114,110,41,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,97,32,97,110,100,32,98,32,97,110,100,32,99,32,97,110,100,32,100,32,97,110,100,32,112,49,32,97,110,100,32,112,50,44,32,114,101,112,108,121,41,10,32,32,32,32,115,101,108,102,46,112,97,115,118,116,32,61,32,123,10,32,32,32,32,32,32,32,32,105,112,32,61,32,115,116,114,105,110,103,46,102,111,114,109,97,116,40,34,37,100,46,37,100,46,37,100,46,37,100,34,44,32,97,44,32,98,44,32,99,44,32,100,41,44,10,32,32,32,32,32,32,32,32,112,111,114,116,32,61,32,112,49,42,50,53,54,32,43,32,112,50,10,32,32,32,32,125,10,32,32,32,32,105,102,32,115,101,108,102,46,115,101,114,118,101,114,32,116,104,101,110,10,32,32,32,32,32,32,32,32,115,101,108,102,46,115,101,114,118,101,114,58,99,108,111,115,101,40,41,10,32,32,32,32,32,32,32,32,115,101,108,102,46,115,101,114,118,101,114,32,61,32,110,105,108,10,32,32,32,32,101,110,100,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,112,97,115,118,116,46,105,112,44,32,115,101,108,102,46,112,97,115,118,116,46,112,111,114,116,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,112,111,114,116,40,105,112,44,32,112,111,114,116,41,10,32,32,32,32,115,101,108,102,46,112,97,115,118,116,32,61,32,110,105,108,10,32,32,32,32,105,102,32,110,111,116,32,105,112,32,116,104,101,110,10,32,32,32,32,32,32,32,32,105,112,44,32,112,111,114,116,32,61,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,103,101,116,99,111,110,116,114,111,108,40,41,58,103,101,116,115,111,99,107,110,97,109,101,40,41,41,10,32,32,32,32,32,32,32,32,115,101,108,102,46,115,101,114,118,101,114,32,61,32,115,101,108,102,46,116,114,121,40,115,111,99,107,101,116,46,98,105,110,100,40,105,112,44,32,48,41,41,10,32,32,32,32,32,32,32,32,105,112,44,32,112,111,114,116,32,61,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,115,101,114,118,101,114,58,103,101,116,115,111,99,107,110,97,109,101,40,41,41,10,32,32,32,32,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,115,101,114,118,101,114,58,115,101,116,116,105,109,101,111,117,116,40,84,73,77,69,79,85,84,41,41,10,32,32,32,32,101,110,100,10,32,32,32,32,108,111,99,97,108,32,112,108,32,61,32,109,97,116,104,46,109,111,100,40,112,111,114,116,44,32,50,53,54,41,10,32,32,32,32,108,111,99,97,108,32,112,104,32,61,32,40,112,111,114,116,32,45,32,112,108,41,47,50,53,54,10,32,32,32,32,108,111,99,97,108,32,97,114,103,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,115,116,114,105,110,103,46,102,111,114,109,97,116,40,34,37,115,44,37,100,44,37,100,34,44,32,105,112,44,32,112,104,44,32,112,108,41,44,32,34,37,46,34,44,32,34,44,34,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,34,112,111,114,116,34,44,32,97,114,103,41,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,50,46,46,34,41,41,10,32,32,32,32,114,101,116,117,114,110,32,49,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,115,101,110,100,40,115,101,110,100,116,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,112,97,115,118,116,32,111,114,32,115,101,108,102,46,115,101,114,118,101,114,44,32,34,110,101,101,100,32,112,111,114,116,32,111,114,32,112,97,115,118,32,102,105,114,115,116,34,41,10,32,32,32,32,45,45,32,105,102,32,116,104,101,114,101,32,105,115,32,97,32,112,97,115,118,116,32,116,97,98,108,101,44,32,119,101,32,97,108,114,101,97,100,121,32,115,101,110,116,32,97,32,80,65,83,86,32,99,111,109,109,97,110,100,10,32,32,32,32,45,45,32,119,101,32,106,117,115,116,32,103,101,116,32,116,104,101,32,100,97,116,97,32,99,111,110,110,101,99,116,105,111,110,32,105,110,116,111,32,115,101,108,102,46,100,97,116,97,10,32,32,32,32,105,102,32,115,101,108,102,46,112,97,115,118,116,32,116,104,101,110,32,115,101,108,102,58,112,97,115,118,99,111,110,110,101,99,116,40,41,32,101,110,100,10,32,32,32,32,45,45,32,103,101,116,32,116,104,101,32,116,114,97,110,115,102,101,114,32,97,114,103,117,109,101,110,116,32,97,110,100,32,99,111,109,109,97,110,100,10,32,32,32,32,108,111,99,97,108,32,97,114,103,117,109,101,110,116,32,61,32,115,101,110,100,116,46,97,114,103,117,109,101,110,116,32,111,114,10,32,32,32,32,32,32,32,32,117,114,108,46,117,110,101,115,99,97,112,101,40,115,116,114,105,110,103,46,103,115,117,98,40,115,101,110,100,116,46,112,97,116,104,32,111,114,32,34,34,44,32,34,94,91,47,92,92,93,34,44,32,34,34,41,41,10,32,32,32,32,105,102,32,97,114,103,117,109,101,110,116,32,61,61,32,34,34,32,116,104,101,110,32,97,114,103,117,109,101,110,116,32,61,32,110,105,108,32,101,110,100,10,32,32,32,32,108,111,99,97,108,32,99,111,109,109,97,110,100,32,61,32,115,101,110,100,116,46,99,111,109,109,97,110,100,32,111,114,32,34,115,116,111,114,34,10,32,32,32,32,45,45,32,115,101,110,100,32,116,104,101,32,116,114,97,110,115,102,101,114,32,99,111,109,109,97,110,100,32,97,110,100,32,99,104,101,99,107,32,116,104,101,32,114,101,112,108,121,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,99,111,109,109,97,110,100,44,32,97,114,103,117,109,101,110,116,41,41,10,32,32,32,32,108,111,99,97,108,32,99,111,100,101,44,32,114,101,112,108,121,32,61,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,123,34,50,46,46,34,44,32,34,49,46,46,34,125,41,10,32,32,32,32,45,45,32,105,102,32,116,104,101,114,101,32,105,115,32,110,111,116,32,97,32,97,32,112,97,115,118,116,32,116,97,98,108,101,44,32,116,104,101,110,32,116,104,101,114,101,32,105,115,32,97,32,115,101,114,118,101,114,10,32,32,32,32,45,45,32,97,110,100,32,119,101,32,97,108,114,101,97,100,121,32,115,101,110,116,32,97,32,80,79,82,84,32,99,111,109,109,97,110,100,10,32,32,32,32,105,102,32,110,111,116,32,115,101,108,102,46,112,97,115,118,116,32,116,104,101,110,32,115,101,108,102,58,112,111,114,116,99,111,110,110,101,99,116,40,41,32,101,110,100,10,32,32,32,32,45,45,32,103,101,116,32,116,104,101,32,115,105,110,107,44,32,115,111,117,114,99,101,32,97,110,100,32,115,116,101,112,32,102,111,114,32,116,104,101,32,116,114,97,110,115,102,101,114,10,32,32,32,32,108,111,99,97,108,32,115,116,101,112,32,61,32,115,101,110,100,116,46,115,116,101,112,32,111,114,32,108,116,110,49,50,46,112,117,109,112,46,115,116,101,112,10,32,32,32,32,108,111,99,97,108,32,114,101,97,100,116,32,61,32,123,115,101,108,102,46,116,112,46,99,125,10,32,32,32,32,108,111,99,97,108,32,99,104,101,99,107,115,116,101,112,32,61,32,102,117,110,99,116,105,111,110,40,115,114,99,44,32,115,110,107,41,10,32,32,32,32,32,32,32,32,45,45,32,99,104,101,99,107,32,115,116,97,116,117,115,32,105,110,32,99,111,110,116,114,111,108,32,99,111,110,110,101,99,116,105,111,110,32,119,104,105,108,101,32,100,111,119,110,108,111,97,100,105,110,103,10,32,32,32,32,32,32,32,32,108,111,99,97,108,32,114,101,97,100,121,116,32,61,32,115,111,99,107,101,116,46,115,101,108,101,99,116,40,114,101,97,100,116,44,32,110,105,108,44,32,48,41,10,32,32,32,32,32,32,32,32,105,102,32,114,101,97,100,121,116,91,116,112,93,32,116,104,101,110,32,99,111,100,101,32,61,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,50,46,46,34,41,41,32,101,110,100,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,115,116,101,112,40,115,114,99,44,32,115,110,107,41,10,32,32,32,32,101,110,100,10,32,32,32,32,108,111,99,97,108,32,115,105,110,107,32,61,32,115,111,99,107,101,116,46,115,105,110,107,40,34,99,108,111,115,101,45,119,104,101,110,45,100,111,110,101,34,44,32,115,101,108,102,46,100,97,116,97,41,10,32,32,32,32,45,45,32,116,114,97,110,115,102,101,114,32,97,108,108,32,100,97,116,97,32,97,110,100,32,99,104,101,99,107,32,101,114,114,111,114,10,32,32,32,32,115,101,108,102,46,116,114,121,40,108,116,110,49,50,46,112,117,109,112,46,97,108,108,40,115,101,110,100,116,46,115,111,117,114,99,101,44,32,115,105,110,107,44,32,99,104,101,99,107,115,116,101,112,41,41,10,32,32,32,32,105,102,32,115,116,114,105,110,103,46,102,105,110,100,40,99,111,100,101,44,32,34,49,46,46,34,41,32,116,104,101,110,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,50,46,46,34,41,41,32,101,110,100,10,32,32,32,32,45,45,32,100,111,110,101,32,119,105,116,104,32,100,97,116,97,32,99,111,110,110,101,99,116,105,111,110,10,32,32,32,32,115,101,108,102,46,100,97,116,97,58,99,108,111,115,101,40,41,10,32,32,32,32,45,45,32,102,105,110,100,32,111,117,116,32,104,111,119,32,109,97,110,121,32,98,121,116,101,115,32,119,101,114,101,32,115,101,110,116,10,32,32,32,32,108,111,99,97,108,32,115,101,110,116,32,61,32,115,111,99,107,101,116,46,115,107,105,112,40,49,44,32,115,101,108,102,46,100,97,116,97,58,103,101,116,115,116,97,116,115,40,41,41,10,32,32,32,32,115,101,108,102,46,100,97,116,97,32,61,32,110,105,108,10,32,32,32,32,114,101,116,117,114,110,32,115,101,110,116,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,114,101,99,101,105,118,101,40,114,101,99,118,116,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,112,97,115,118,116,32,111,114,32,115,101,108,102,46,115,101,114,118,101,114,44,32,34,110,101,101,100,32,112,111,114,116,32,111,114,32,112,97,115,118,32,102,105,114,115,116,34,41,10,32,32,32,32,105,102,32,115,101,108,102,46,112,97,115,118,116,32,116,104,101,110,32,115,101,108,102,58,112,97,115,118,99,111,110,110,101,99,116,40,41,32,101,110,100,10,32,32,32,32,108,111,99,97,108,32,97,114,103,117,109,101,110,116,32,61,32,114,101,99,118,116,46,97,114,103,117,109,101,110,116,32,111,114,10,32,32,32,32,32,32,32,32,117,114,108,46,117,110,101,115,99,97,112,101,40,115,116,114,105,110,103,46,103,115,117,98,40,114,101,99,118,116,46,112,97,116,104,32,111,114,32,34,34,44,32,34,94,91,47,92,92,93,34,44,32,34,34,41,41,10,32,32,32,32,105,102,32,97,114,103,117,109,101,110,116,32,61,61,32,34,34,32,116,104,101,110,32,97,114,103,117,109,101,110,116,32,61,32,110,105,108,32,101,110,100,10,32,32,32,32,108,111,99,97,108,32,99,111,109,109,97,110,100,32,61,32,114,101,99,118,116,46,99,111,109,109,97,110,100,32,111,114,32,34,114,101,116,114,34,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,99,111,109,109,97,110,100,44,32,97,114,103,117,109,101,110,116,41,41,10,32,32,32,32,108,111,99,97,108,32,99,111,100,101,32,61,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,123,34,49,46,46,34,44,32,34,50,46,46,34,125,41,10,32,32,32,32,105,102,32,110,111,116,32,115,101,108,102,46,112,97,115,118,116,32,116,104,101,110,32,115,101,108,102,58,112,111,114,116,99,111,110,110,101,99,116,40,41,32,101,110,100,10,32,32,32,32,108,111,99,97,108,32,115,111,117,114,99,101,32,61,32,115,111,99,107,101,116,46,115,111,117,114,99,101,40,34,117,110,116,105,108,45,99,108,111,115,101,100,34,44,32,115,101,108,102,46,100,97,116,97,41,10,32,32,32,32,108,111,99,97,108,32,115,116,101,112,32,61,32,114,101,99,118,116,46,115,116,101,112,32,111,114,32,108,116,110,49,50,46,112,117,109,112,46,115,116,101,112,10,32,32,32,32,115,101,108,102,46,116,114,121,40,108,116,110,49,50,46,112,117,109,112,46,97,108,108,40,115,111,117,114,99,101,44,32,114,101,99,118,116,46,115,105,110,107,44,32,115,116,101,112,41,41,10,32,32,32,32,105,102,32,115,116,114,105,110,103,46,102,105,110,100,40,99,111,100,101,44,32,34,49,46,46,34,41,32,116,104,101,110,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,50,46,46,34,41,41,32,101,110,100,10,32,32,32,32,115,101,108,102,46,100,97,116,97,58,99,108,111,115,101,40,41,10,32,32,32,32,115,101,108,102,46,100,97,116,97,32,61,32,110,105,108,10,32,32,32,32,114,101,116,117,114,110,32,49,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,99,119,100,40,100,105,114,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,34,99,119,100,34,44,32,100,105,114,41,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,50,53,48,41,41,10,32,32,32,32,114,101,116,117,114,110,32,49,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,116,121,112,101,40,116,121,112,101,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,34,116,121,112,101,34,44,32,116,121,112,101,41,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,50,48,48,41,41,10,32,32,32,32,114,101,116,117,114,110,32,49,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,103,114,101,101,116,40,41,10,32,32,32,32,108,111,99,97,108,32,99,111,100,101,32,61,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,123,34,49,46,46,34,44,32,34,50,46,46,34,125,41,10,32,32,32,32,105,102,32,115,116,114,105,110,103,46,102,105,110,100,40,99,111,100,101,44,32,34,49,46,46,34,41,32,116,104,101,110,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,50,46,46,34,41,41,32,101,110,100,10,32,32,32,32,114,101,116,117,114,110,32,49,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,113,117,105,116,40,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,34,113,117,105,116,34,41,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,50,46,46,34,41,41,10,32,32,32,32,114,101,116,117,114,110,32,49,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,99,108,111,115,101,40,41,10,32,32,32,32,105,102,32,115,101,108,102,46,100,97,116,97,32,116,104,101,110,32,115,101,108,102,46,100,97,116,97,58,99,108,111,115,101,40,41,32,101,110,100,10,32,32,32,32,105,102,32,115,101,108,102,46,115,101,114,118,101,114,32,116,104,101,110,32,115,101,108,102,46,115,101,114,118,101,114,58,99,108,111,115,101,40,41,32,101,110,100,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,116,112,58,99,108,111,115,101,40,41,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,72,105,103,104,32,108,101,118,101,108,32,70,84,80,32,65,80,73,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,111,118,101,114,114,105,100,101,40,116,41,10,32,32,32,32,105,102,32,116,46,117,114,108,32,116,104,101,110,10,32,32,32,32,32,32,32,32,108,111,99,97,108,32,117,32,61,32,117,114,108,46,112,97,114,115,101,40,116,46,117,114,108,41,10,32,32,32,32,32,32,32,32,102,111,114,32,105,44,118,32,105,110,32,98,97,115,101,46,112,97,105,114,115,40,116,41,32,100,111,10,32,32,32,32,32,32,32,32,32,32,32,32,117,91,105,93,32,61,32,118,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,117,10,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,116,32,101,110,100,10,101,110,100,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,116,112,117,116,40,112,117,116,116,41,10,32,32,32,32,112,117,116,116,32,61,32,111,118,101,114,114,105,100,101,40,112,117,116,116,41,10,32,32,32,32,115,111,99,107,101,116,46,116,114,121,40,112,117,116,116,46,104,111,115,116,44,32,34,109,105,115,115,105,110,103,32,104,111,115,116,110,97,109,101,34,41,10,32,32,32,32,108,111,99,97,108,32,102,32,61,32,111,112,101,110,40,112,117,116,116,46,104,111,115,116,44,32,112,117,116,116,46,112,111,114,116,44,32,112,117,116,116,46,99,114,101,97,116,101,41,10,32,32,32,32,102,58,103,114,101,101,116,40,41,10,32,32,32,32,102,58,108,111,103,105,110,40,112,117,116,116,46,117,115,101,114,44,32,112,117,116,116,46,112,97,115,115,119,111,114,100,41,10,32,32,32,32,105,102,32,112,117,116,116,46,116,121,112,101,32,116,104,101,110,32,102,58,116,121,112,101,40,112,117,116,116,46,116,121,112,101,41,32,101,110,100,10,32,32,32,32,102,58,112,97,115,118,40,41,10,32,32,32,32,108,111,99,97,108,32,115,101,110,116,32,61,32,102,58,115,101,110,100,40,112,117,116,116,41,10,32,32,32,32,102,58,113,117,105,116,40,41,10,32,32,32,32,102,58,99,108,111,115,101,40,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,110,116,10,101,110,100,10,10,108,111,99,97,108,32,100,101,102,97,117,108,116,32,61,32,123,10,9,112,97,116,104,32,61,32,34,47,34,44,10,9,115,99,104,101,109,101,32,61,32,34,102,116,112,34,10,125,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,112,97,114,115,101,40,117,41,10,32,32,32,32,108,111,99,97,108,32,116,32,61,32,115,111,99,107,101,116,46,116,114,121,40,117,114,108,46,112,97,114,115,101,40,117,44,32,100,101,102,97,117,108,116,41,41,10,32,32,32,32,115,111,99,107,101,116,46,116,114,121,40,116,46,115,99,104,101,109,101,32,61,61,32,34,102,116,112,34,44,32,34,119,114,111,110,103,32,115,99,104,101,109,101,32,39,34,32,46,46,32,116,46,115,99,104,101,109,101,32,46,46,32,34,39,34,41,10,32,32,32,32,115,111,99,107,101,116,46,116,114,121,40,116,46,104,111,115,116,44,32,34,109,105,115,115,105,110,103,32,104,111,115,116,110,97,109,101,34,41,10,32,32,32,32,108,111,99,97,108,32,112,97,116,32,61,32,34,94,116,121,112,101,61,40,46,41,36,34,10,32,32,32,32,105,102,32,116,46,112,97,114,97,109,115,32,116,104,101,110,10,32,32,32,32,32,32,32,32,116,46,116,121,112,101,32,61,32,115,111,99,107,101,116,46,115,107,105,112,40,50,44,32,115,116,114,105,110,103,46,102,105,110,100,40,116,46,112,97,114,97,109,115,44,32,112,97,116,41,41,10,32,32,32,32,32,32,32,32,115,111,99,107,101,116,46,116,114,121,40,116,46,116,121,112,101,32,61,61,32,34,97,34,32,111,114,32,116,46,116,121,112,101,32,61,61,32,34,105,34,44,10,32,32,32,32,32,32,32,32,32,32,32,32,34,105,110,118,97,108,105,100,32,116,121,112,101,32,39,34,32,46,46,32,116,46,116,121,112,101,32,46,46,32,34,39,34,41,10,32,32,32,32,101,110,100,10,32,32,32,32,114,101,116,117,114,110,32,116,10,101,110,100,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,115,112,117,116,40,117,44,32,98,111,100,121,41,10,32,32,32,32,108,111,99,97,108,32,112,117,116,116,32,61,32,112,97,114,115,101,40,117,41,10,32,32,32,32,112,117,116,116,46,115,111,117,114,99,101,32,61,32,108,116,110,49,50,46,115,111,117,114,99,101,46,115,116,114,105,110,103,40,98,111,100,121,41,10,32,32,32,32,114,101,116,117,114,110,32,116,112,117,116,40,112,117,116,116,41,10,101,110,100,10,10,112,117,116,32,61,32,115,111,99,107,101,116,46,112,114,111,116,101,99,116,40,102,117,110,99,116,105,111,110,40,112,117,116,116,44,32,98,111,100,121,41,10,32,32,32,32,105,102,32,98,97,115,101,46,116,121,112,101,40,112,117,116,116,41,32,61,61,32,34,115,116,114,105,110,103,34,32,116,104,101,110,32,114,101,116,117,114,110,32,115,112,117,116,40,112,117,116,116,44,32,98,111,100,121,41,10,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,116,112,117,116,40,112,117,116,116,41,32,101,110,100,10,101,110,100,41,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,116,103,101,116,40,103,101,116,116,41,10,32,32,32,32,103,101,116,116,32,61,32,111,118,101,114,114,105,100,101,40,103,101,116,116,41,10,32,32,32,32,115,111,99,107,101,116,46,116,114,121,40,103,101,116,116,46,104,111,115,116,44,32,34,109,105,115,115,105,110,103,32,104,111,115,116,110,97,109,101,34,41,10,32,32,32,32,108,111,99,97,108,32,102,32,61,32,111,112,101,110,40,103,101,116,116,46,104,111,115,116,44,32,103,101,116,116,46,112,111,114,116,44,32,103,101,116,116,46,99,114,101,97,116,101,41,10,32,32,32,32,102,58,103,114,101,101,116,40,41,10,32,32,32,32,102,58,108,111,103,105,110,40,103,101,116,116,46,117,115,101,114,44,32,103,101,116,116,46,112,97,115,115,119,111,114,100,41,10,32,32,32,32,105,102,32,103,101,116,116,46,116,121,112,101,32,116,104,101,110,32,102,58,116,121,112,101,40,103,101,116,116,46,116,121,112,101,41,32,101,110,100,10,32,32,32,32,102,58,112,97,115,118,40,41,10,32,32,32,32,102,58,114,101,99,101,105,118,101,40,103,101,116,116,41,10,32,32,32,32,102,58,113,117,105,116,40,41,10,32,32,32,32,114,101,116,117,114,110,32,102,58,99,108,111,115,101,40,41,10,101,110,100,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,115,103,101,116,40,117,41,10,32,32,32,32,108,111,99,97,108,32,103,101,116,116,32,61,32,112,97,114,115,101,40,117,41,10,32,32,32,32,108,111,99,97,108,32,116,32,61,32,123,125,10,32,32,32,32,103,101,116,116,46,115,105,110,107,32,61,32,108,116,110,49,50,46,115,105,110,107,46,116,97,98,108,101,40,116,41,10,32,32,32,32,116,103,101,116,40,103,101,116,116,41,10,32,32,32,32,114,101,116,117,114,110,32,116,97,98,108,101,46,99,111,110,99,97,116,40,116,41,10,101,110,100,10,10,99,111,109,109,97,110,100,32,61,32,115,111,99,107,101,116,46,112,114,111,116,101,99,116,40,102,117,110,99,116,105,111,110,40,99,109,100,116,41,10,32,32,32,32,99,109,100,116,32,61,32,111,118,101,114,114,105,100,101,40,99,109,100,116,41,10,32,32,32,32,115,111,99,107,101,116,46,116,114,121,40,99,109,100,116,46,104,111,115,116,44,32,34,109,105,115,115,105,110,103,32,104,111,115,116,110,97,109,101,34,41,10,32,32,32,32,115,111,99,107,101,116,46,116,114,121,40,99,109,100,116,46,99,111,109,109,97,110,100,44,32,34,109,105,115,115,105,110,103,32,99,111,109,109,97,110,100,34,41,10,32,32,32,32,108,111,99,97,108,32,102,32,61,32,111,112,101,110,40,99,109,100,116,46,104,111,115,116,44,32,99,109,100,116,46,112,111,114,116,44,32,99,109,100,116,46,99,114,101,97,116,101,41,10,32,32,32,32,102,58,103,114,101,101,116,40,41,10,32,32,32,32,102,58,108,111,103,105,110,40,99,109,100,116,46,117,115,101,114,44,32,99,109,100,116,46,112,97,115,115,119,111,114,100,41,10,32,32,32,32,102,46,116,114,121,40,102,46,116,112,58,99,111,109,109,97,110,100,40,99,109,100,116,46,99,111,109,109,97,110,100,44,32,99,109,100,116,46,97,114,103,117,109,101,110,116,41,41,10,32,32,32,32,105,102,32,99,109,100,116,46,99,104,101,99,107,32,116,104,101,110,32,102,46,116,114,121,40,102,46,116,112,58,99,104,101,99,107,40,99,109,100,116,46,99,104,101,99,107,41,41,32,101,110,100,10,32,32,32,32,102,58,113,117,105,116,40,41,10,32,32,32,32,114,101,116,117,114,110,32,102,58,99,108,111,115,101,40,41,10,101,110,100,41,10,10,103,101,116,32,61,32,115,111,99,107,101,116,46,112,114,111,116,101,99,116,40,102,117,110,99,116,105,111,110,40,103,101,116,116,41,10,32,32,32,32,105,102,32,98,97,115,101,46,116,121,112,101,40,103,101,116,116,41,32,61,61,32,34,115,116,114,105,110,103,34,32,116,104,101,110,32,114,101,116,117,114,110,32,115,103,101,116,40,103,101,116,116,41,10,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,116,103,101,116,40,103,101,116,116,41,32,101,110,100,10,101,110,100,41,10,10,102,116,112,46,108,117,97,0,115,111,99,107,101,116,46,102,116,112,0,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,72,84,84,80,47,49,46,49,32,99,108,105,101,110,116,32,115,117,112,112,111,114,116,32,102,111,114,32,116,104,101,32,76,117,97,32,108,97,110,103,117,97,103,101,46,10,45,45,32,76,117,97,83,111,99,107,101,116,32,116,111,111,108,107,105,116,46,10,45,45,32,65,117,116,104,111,114,58,32,68,105,101,103,111,32,78,101,104,97,98,10,45,45,32,82,67,83,32,73,68,58,32,36,73,100,58,32,104,116,116,112,46,108,117,97,44,118,32,49,46,55,49,32,50,48,48,55,47,49,48,47,49,51,32,50,51,58,53,53,58,50,48,32,100,105,101,103,111,32,69,120,112,32,36,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,68,101,99,108,97,114,101,32,109,111,100,117,108,101,32,97,110,100,32,105,109,112,111,114,116,32,100,101,112,101,110,100,101,110,99,105,101,115,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,108,111,99,97,108,32,115,111,99,107,101,116,32,61,32,114,101,113,117,105,114,101,40,34,115,111,99,107,101,116,34,41,10,108,111,99,97,108,32,117,114,108,32,61,32,114,101,113,117,105,114,101,40,34,115,111,99,107,101,116,46,117,114,108,34,41,10,108,111,99,97,108,32,108,116,110,49,50,32,61,32,114,101,113,117,105,114,101,40,34,108,116,110,49,50,34,41,10,108,111,99,97,108,32,109,105,109,101,32,61,32,114,101,113,117,105,114,101,40,34,109,105,109,101,34,41,10,108,111,99,97,108,32,115,116,114,105,110,103,32,61,32,114,101,113,117,105,114,101,40,34,115,116,114,105,110,103,34,41,10,108,111,99,97,108,32,98,97,115,101,32,61,32,95,71,10,108,111,99,97,108,32,116,97,98,108,101,32,61,32,114,101,113,117,105,114,101,40,34,116,97,98,108,101,34,41,10,109,111,100,117,108,101,40,34,115,111,99,107,101,116,46,104,116,116,112,34,41,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,80,114,111,103,114,97,109,32,99,111,110,115,116,97,110,116,115,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,99,111,110,110,101,99,116,105,111,110,32,116,105,109,101,111,117,116,32,105,110,32,115,101,99,111,110,100,115,10,84,73,77,69,79,85,84,32,61,32,54,48,10,45,45,32,100,101,102,97,117,108,116,32,112,111,114,116,32,102,111,114,32,100,111,99,117,109,101,110,116,32,114,101,116,114,105,101,118,97,108,10,80,79,82,84,32,61,32,56,48,10,45,45,32,117,115,101,114,32,97,103,101,110,116,32,102,105,101,108,100,32,115,101,110,116,32,105,110,32,114,101,113,117,101,115,116,10,85,83,69,82,65,71,69,78,84,32,61,32,115,111,99,107,101,116,46,95,86,69,82,83,73,79,78,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,82,101,97,100,115,32,77,73,77,69,32,104,101,97,100,101,114,115,32,102,114,111,109,32,97,32,99,111,110,110,101,99,116,105,111,110,44,32,117,110,102,111,108,100,105,110,103,32,119,104,101,114,101,32,110,101,101,100,101,100,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,114,101,99,101,105,118,101,104,101,97,100,101,114,115,40,115,111,99,107,44,32,104,101,97,100,101,114,115,41,10,32,32,32,32,108,111,99,97,108,32,108,105,110,101,44,32,110,97,109,101,44,32,118,97,108,117,101,44,32,101,114,114,10,32,32,32,32,104,101,97,100,101,114,115,32,61,32,104,101,97,100,101,114,115,32,111,114,32,123,125,10,32,32,32,32,45,45,32,103,101,116,32,102,105,114,115,116,32,108,105,110,101,10,32,32,32,32,108,105,110,101,44,32,101,114,114,32,61,32,115,111,99,107,58,114,101,99,101,105,118,101,40,41,10,32,32,32,32,105,102,32,101,114,114,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,32,101,110,100,10,32,32,32,32,45,45,32,104,101,97,100,101,114,115,32,103,111,32,117,110,116,105,108,32,97,32,98,108,97,110,107,32,108,105,110,101,32,105,115,32,102,111,117,110,100,10,32,32,32,32,119,104,105,108,101,32,108,105,110,101,32,126,61,32,34,34,32,100,111,10,32,32,32,32,32,32,32,32,45,45,32,103,101,116,32,102,105,101,108,100,45,110,97,109,101,32,97,110,100,32,118,97,108,117,101,10,32,32,32,32,32,32,32,32,110,97,109,101,44,32,118,97,108,117,101,32,61,32,115,111,99,107,101,116,46,115,107,105,112,40,50,44,32,115,116,114,105,110,103,46,102,105,110,100,40,108,105,110,101,44,32,34,94,40,46,45,41,58,37,115,42,40,46,42,41,34,41,41,10,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,40,110,97,109,101,32,97,110,100,32,118,97,108,117,101,41,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,34,109,97,108,102,111,114,109,101,100,32,114,101,112,111,110,115,101,32,104,101,97,100,101,114,115,34,32,101,110,100,10,32,32,32,32,32,32,32,32,110,97,109,101,32,61,32,115,116,114,105,110,103,46,108,111,119,101,114,40,110,97,109,101,41,10,32,32,32,32,32,32,32,32,45,45,32,103,101,116,32,110,101,120,116,32,108,105,110,101,32,40,118,97,108,117,101,32,109,105,103,104,116,32,98,101,32,102,111,108,100,101,100,41,10,32,32,32,32,32,32,32,32,108,105,110,101,44,32,101,114,114,32,32,61,32,115,111,99,107,58,114,101,99,101,105,118,101,40,41,10,32,32,32,32,32,32,32,32,105,102,32,101,114,114,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,32,101,110,100,10,32,32,32,32,32,32,32,32,45,45,32,117,110,102,111,108,100,32,97,110,121,32,102,111,108,100,101,100,32,118,97,108,117,101,115,10,32,32,32,32,32,32,32,32,119,104,105,108,101,32,115,116,114,105,110,103,46,102,105,110,100,40,108,105,110,101,44,32,34,94,37,115,34,41,32,100,111,10,32,32,32,32,32,32,32,32,32,32,32,32,118,97,108,117,101,32,61,32,118,97,108,117,101,32,46,46,32,108,105,110,101,10,32,32,32,32,32,32,32,32,32,32,32,32,108,105,110,101,32,61,32,115,111,99,107,58,114,101,99,101,105,118,101,40,41,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,101,114,114,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,32,101,110,100,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,45,45,32,115,97,118,101,32,112,97,105,114,32,105,110,32,116,97,98,108,101,10,32,32,32,32,32,32,32,32,105,102,32,104,101,97,100,101,114,115,91,110,97,109,101,93,32,116,104,101,110,32,104,101,97,100,101,114,115,91,110,97,109,101,93,32,61,32,104,101,97,100,101,114,115,91,110,97,109,101,93,32,46,46,32,34,44,32,34,32,46,46,32,118,97,108,117,101,10,32,32,32,32,32,32,32,32,101,108,115,101,32,104,101,97,100,101,114,115,91,110,97,109,101,93,32,61,32,118,97,108,117,101,32,101,110,100,10,32,32,32,32,101,110,100,10,32,32,32,32,114,101,116,117,114,110,32,104,101,97,100,101,114,115,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,69,120,116,114,97,32,115,111,117,114,99,101,115,32,97,110,100,32,115,105,110,107,115,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,115,111,99,107,101,116,46,115,111,117,114,99,101,116,91,34,104,116,116,112,45,99,104,117,110,107,101,100,34,93,32,61,32,102,117,110,99,116,105,111,110,40,115,111,99,107,44,32,104,101,97,100,101,114,115,41,10,32,32,32,32,114,101,116,117,114,110,32,98,97,115,101,46,115,101,116,109,101,116,97,116,97,98,108,101,40,123,10,32,32,32,32,32,32,32,32,103,101,116,102,100,32,61,32,102,117,110,99,116,105,111,110,40,41,32,114,101,116,117,114,110,32,115,111,99,107,58,103,101,116,102,100,40,41,32,101,110,100,44,10,32,32,32,32,32,32,32,32,100,105,114,116,121,32,61,32,102,117,110,99,116,105,111,110,40,41,32,114,101,116,117,114,110,32,115,111,99,107,58,100,105,114,116,121,40,41,32,101,110,100,10,32,32,32,32,125,44,32,123,10,32,32,32,32,32,32,32,32,95,95,99,97,108,108,32,61,32,102,117,110,99,116,105,111,110,40,41,10,32,32,32,32,32,32,32,32,32,32,32,32,45,45,32,103,101,116,32,99,104,117,110,107,32,115,105,122,101,44,32,115,107,105,112,32,101,120,116,101,110,116,105,111,110,10,32,32,32,32,32,32,32,32,32,32,32,32,108,111,99,97,108,32,108,105,110,101,44,32,101,114,114,32,61,32,115,111,99,107,58,114,101,99,101,105,118,101,40,41,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,101,114,114,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,108,111,99,97,108,32,115,105,122,101,32,61,32,98,97,115,101,46,116,111,110,117,109,98,101,114,40,115,116,114,105,110,103,46,103,115,117,98,40,108,105,110,101,44,32,34,59,46,42,34,44,32,34,34,41,44,32,49,54,41,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,115,105,122,101,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,34,105,110,118,97,108,105,100,32,99,104,117,110,107,32,115,105,122,101,34,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,45,45,32,119,97,115,32,105,116,32,116,104,101,32,108,97,115,116,32,99,104,117,110,107,63,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,115,105,122,101,32,62,32,48,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,45,45,32,105,102,32,110,111,116,44,32,103,101,116,32,99,104,117,110,107,32,97,110,100,32,115,107,105,112,32,116,101,114,109,105,110,97,116,105,110,103,32,67,82,76,70,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,108,111,99,97,108,32,99,104,117,110,107,44,32,101,114,114,44,32,112,97,114,116,32,61,32,115,111,99,107,58,114,101,99,101,105,118,101,40,115,105,122,101,41,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,99,104,117,110,107,32,116,104,101,110,32,115,111,99,107,58,114,101,99,101,105,118,101,40,41,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,99,104,117,110,107,44,32,101,114,114,10,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,45,45,32,105,102,32,105,116,32,119,97,115,44,32,114,101,97,100,32,116,114,97,105,108,101,114,115,32,105,110,116,111,32,104,101,97,100,101,114,115,32,116,97,98,108,101,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,104,101,97,100,101,114,115,44,32,101,114,114,32,61,32,114,101,99,101,105,118,101,104,101,97,100,101,114,115,40,115,111,99,107,44,32,104,101,97,100,101,114,115,41,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,104,101,97,100,101,114,115,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,125,41,10,101,110,100,10,10,115,111,99,107,101,116,46,115,105,110,107,116,91,34,104,116,116,112,45,99,104,117,110,107,101,100,34,93,32,61,32,102,117,110,99,116,105,111,110,40,115,111,99,107,41,10,32,32,32,32,114,101,116,117,114,110,32,98,97,115,101,46,115,101,116,109,101,116,97,116,97,98,108,101,40,123,10,32,32,32,32,32,32,32,32,103,101,116,102,100,32,61,32,102,117,110,99,116,105,111,110,40,41,32,114,101,116,117,114,110,32,115,111,99,107,58,103,101,116,102,100,40,41,32,101,110,100,44,10,32,32,32,32,32,32,32,32,100,105,114,116,121,32,61,32,102,117,110,99,116,105,111,110,40,41,32,114,101,116,117,114,110,32,115,111,99,107,58,100,105,114,116,121,40,41,32,101,110,100,10,32,32,32,32,125,44,32,123,10,32,32,32,32,32,32,32,32,95,95,99,97,108,108,32,61,32,102,117,110,99,116,105,111,110,40,115,101,108,102,44,32,99,104,117,110,107,44,32,101,114,114,41,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,99,104,117,110,107,32,116,104,101,110,32,114,101,116,117,114,110,32,115,111,99,107,58,115,101,110,100,40,34,48,92,114,92,110,92,114,92,110,34,41,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,108,111,99,97,108,32,115,105,122,101,32,61,32,115,116,114,105,110,103,46,102,111,114,109,97,116,40,34,37,88,92,114,92,110,34,44,32,115,116,114,105,110,103,46,108,101,110,40,99,104,117,110,107,41,41,10,32,32,32,32,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,115,111,99,107,58,115,101,110,100,40,115,105,122,101,32,46,46,32,32,99,104,117,110,107,32,46,46,32,34,92,114,92,110,34,41,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,125,41,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,76,111,119,32,108,101,118,101,108,32,72,84,84,80,32,65,80,73,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,108,111,99,97,108,32,109,101,116,97,116,32,61,32,123,32,95,95,105,110,100,101,120,32,61,32,123,125,32,125,10,10,102,117,110,99,116,105,111,110,32,111,112,101,110,40,104,111,115,116,44,32,112,111,114,116,44,32,99,114,101,97,116,101,41,10,32,32,32,32,45,45,32,99,114,101,97,116,101,32,115,111,99,107,101,116,32,119,105,116,104,32,117,115,101,114,32,99,111,110,110,101,99,116,32,102,117,110,99,116,105,111,110,44,32,111,114,32,119,105,116,104,32,100,101,102,97,117,108,116,10,32,32,32,32,108,111,99,97,108,32,99,32,61,32,115,111,99,107,101,116,46,116,114,121,40,40,99,114,101,97,116,101,32,111,114,32,115,111,99,107,101,116,46,116,99,112,41,40,41,41,10,32,32,32,32,108,111,99,97,108,32,104,32,61,32,98,97,115,101,46,115,101,116,109,101,116,97,116,97,98,108,101,40,123,32,99,32,61,32,99,32,125,44,32,109,101,116,97,116,41,10,32,32,32,32,45,45,32,99,114,101,97,116,101,32,102,105,110,97,108,105,122,101,100,32,116,114,121,10,32,32,32,32,104,46,116,114,121,32,61,32,115,111,99,107,101,116,46,110,101,119,116,114,121,40,102,117,110,99,116,105,111,110,40,41,32,104,58,99,108,111,115,101,40,41,32,101,110,100,41,10,32,32,32,32,45,45,32,115,101,116,32,116,105,109,101,111,117,116,32,98,101,102,111,114,101,32,99,111,110,110,101,99,116,105,110,103,10,32,32,32,32,104,46,116,114,121,40,99,58,115,101,116,116,105,109,101,111,117,116,40,84,73,77,69,79,85,84,41,41,10,32,32,32,32,104,46,116,114,121,40,99,58,99,111,110,110,101,99,116,40,104,111,115,116,44,32,112,111,114,116,32,111,114,32,80,79,82,84,41,41,10,32,32,32,32,45,45,32,104,101,114,101,32,101,118,101,114,121,116,104,105,110,103,32,119,111,114,107,101,100,10,32,32,32,32,114,101,116,117,114,110,32,104,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,115,101,110,100,114,101,113,117,101,115,116,108,105,110,101,40,109,101,116,104,111,100,44,32,117,114,105,41,10,32,32,32,32,108,111,99,97,108,32,114,101,113,108,105,110,101,32,61,32,115,116,114,105,110,103,46,102,111,114,109,97,116,40,34,37,115,32,37,115,32,72,84,84,80,47,49,46,49,92,114,92,110,34,44,32,109,101,116,104,111,100,32,111,114,32,34,71,69,84,34,44,32,117,114,105,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,99,58,115,101,110,100,40,114,101,113,108,105,110,101,41,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,115,101,110,100,104,101,97,100,101,114,115,40,104,101,97,100,101,114,115,41,10,32,32,32,32,108,111,99,97,108,32,104,32,61,32,34,92,114,92,110,34,10,32,32,32,32,102,111,114,32,105,44,32,118,32,105,110,32,98,97,115,101,46,112,97,105,114,115,40,104,101,97,100,101,114,115,41,32,100,111,10,32,32,32,32,32,32,32,32,104,32,61,32,105,32,46,46,32,34,58,32,34,32,46,46,32,118,32,46,46,32,34,92,114,92,110,34,32,46,46,32,104,10,32,32,32,32,101,110,100,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,99,58,115,101,110,100,40,104,41,41,10,32,32,32,32,114,101,116,117,114,110,32,49,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,115,101,110,100,98,111,100,121,40,104,101,97,100,101,114,115,44,32,115,111,117,114,99,101,44,32,115,116,101,112,41,10,32,32,32,32,115,111,117,114,99,101,32,61,32,115,111,117,114,99,101,32,111,114,32,108,116,110,49,50,46,115,111,117,114,99,101,46,101,109,112,116,121,40,41,10,32,32,32,32,115,116,101,112,32,61,32,115,116,101,112,32,111,114,32,108,116,110,49,50,46,112,117,109,112,46,115,116,101,112,10,32,32,32,32,45,45,32,105,102,32,119,101,32,100,111,110,39,116,32,107,110,111,119,32,116,104,101,32,115,105,122,101,32,105,110,32,97,100,118,97,110,99,101,44,32,115,101,110,100,32,99,104,117,110,107,101,100,32,97,110,100,32,104,111,112,101,32,102,111,114,32,116,104,101,32,98,101,115,116,10,32,32,32,32,108,111,99,97,108,32,109,111,100,101,32,61,32,34,104,116,116,112,45,99,104,117,110,107,101,100,34,10,32,32,32,32,105,102,32,104,101,97,100,101,114,115,91,34,99,111,110,116,101,110,116,45,108,101,110,103,116,104,34,93,32,116,104,101,110,32,109,111,100,101,32,61,32,34,107,101,101,112,45,111,112,101,110,34,32,101,110,100,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,116,114,121,40,108,116,110,49,50,46,112,117,109,112,46,97,108,108,40,115,111,117,114,99,101,44,32,115,111,99,107,101,116,46,115,105,110,107,40,109,111,100,101,44,32,115,101,108,102,46,99,41,44,32,115,116,101,112,41,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,114,101,99,101,105,118,101,115,116,97,116,117,115,108,105,110,101,40,41,10,32,32,32,32,108,111,99,97,108,32,115,116,97,116,117,115,32,61,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,99,58,114,101,99,101,105,118,101,40,53,41,41,10,32,32,32,32,45,45,32,105,100,101,110,116,105,102,121,32,72,84,84,80,47,48,46,57,32,114,101,115,112,111,110,115,101,115,44,32,119,104,105,99,104,32,100,111,32,110,111,116,32,99,111,110,116,97,105,110,32,97,32,115,116,97,116,117,115,32,108,105,110,101,10,32,32,32,32,45,45,32,116,104,105,115,32,105,115,32,106,117,115,116,32,97,32,104,101,117,114,105,115,116,105,99,44,32,98,117,116,32,105,115,32,119,104,97,116,32,116,104,101,32,82,70,67,32,114,101,99,111,109,109,101,110,100,115,10,32,32,32,32,105,102,32,115,116,97,116,117,115,32,126,61,32,34,72,84,84,80,47,34,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,115,116,97,116,117,115,32,101,110,100,10,32,32,32,32,45,45,32,111,116,104,101,114,119,105,115,101,32,112,114,111,99,101,101,100,32,114,101,97,100,105,110,103,32,97,32,115,116,97,116,117,115,32,108,105,110,101,10,32,32,32,32,115,116,97,116,117,115,32,61,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,99,58,114,101,99,101,105,118,101,40,34,42,108,34,44,32,115,116,97,116,117,115,41,41,10,32,32,32,32,108,111,99,97,108,32,99,111,100,101,32,61,32,115,111,99,107,101,116,46,115,107,105,112,40,50,44,32,115,116,114,105,110,103,46,102,105,110,100,40,115,116,97,116,117,115,44,32,34,72,84,84,80,47,37,100,42,37,46,37,100,42,32,40,37,100,37,100,37,100,41,34,41,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,116,114,121,40,98,97,115,101,46,116,111,110,117,109,98,101,114,40,99,111,100,101,41,44,32,115,116,97,116,117,115,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,114,101,99,101,105,118,101,104,101,97,100,101,114,115,40,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,116,114,121,40,114,101,99,101,105,118,101,104,101,97,100,101,114,115,40,115,101,108,102,46,99,41,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,114,101,99,101,105,118,101,98,111,100,121,40,104,101,97,100,101,114,115,44,32,115,105,110,107,44,32,115,116,101,112,41,10,32,32,32,32,115,105,110,107,32,61,32,115,105,110,107,32,111,114,32,108,116,110,49,50,46,115,105,110,107,46,110,117,108,108,40,41,10,32,32,32,32,115,116,101,112,32,61,32,115,116,101,112,32,111,114,32,108,116,110,49,50,46,112,117,109,112,46,115,116,101,112,10,32,32,32,32,108,111,99,97,108,32,108,101,110,103,116,104,32,61,32,98,97,115,101,46,116,111,110,117,109,98,101,114,40,104,101,97,100,101,114,115,91,34,99,111,110,116,101,110,116,45,108,101,110,103,116,104,34,93,41,10,32,32,32,32,108,111,99,97,108,32,116,32,61,32,104,101,97,100,101,114,115,91,34,116,114,97,110,115,102,101,114,45,101,110,99,111,100,105,110,103,34,93,32,45,45,32,115,104,111,114,116,99,117,116,10,32,32,32,32,108,111,99,97,108,32,109,111,100,101,32,61,32,34,100,101,102,97,117,108,116,34,32,45,45,32,99,111,110,110,101,99,116,105,111,110,32,99,108,111,115,101,10,32,32,32,32,105,102,32,116,32,97,110,100,32,116,32,126,61,32,34,105,100,101,110,116,105,116,121,34,32,116,104,101,110,32,109,111,100,101,32,61,32,34,104,116,116,112,45,99,104,117,110,107,101,100,34,10,32,32,32,32,101,108,115,101,105,102,32,98,97,115,101,46,116,111,110,117,109,98,101,114,40,104,101,97,100,101,114,115,91,34,99,111,110,116,101,110,116,45,108,101,110,103,116,104,34,93,41,32,116,104,101,110,32,109,111,100,101,32,61,32,34,98,121,45,108,101,110,103,116,104,34,32,101,110,100,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,116,114,121,40,108,116,110,49,50,46,112,117,109,112,46,97,108,108,40,115,111,99,107,101,116,46,115,111,117,114,99,101,40,109,111,100,101,44,32,115,101,108,102,46,99,44,32,108,101,110,103,116,104,41,44,10,32,32,32,32,32,32,32,32,115,105,110,107,44,32,115,116,101,112,41,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,114,101,99,101,105,118,101,48,57,98,111,100,121,40,115,116,97,116,117,115,44,32,115,105,110,107,44,32,115,116,101,112,41,10,32,32,32,32,108,111,99,97,108,32,115,111,117,114,99,101,32,61,32,108,116,110,49,50,46,115,111,117,114,99,101,46,114,101,119,105,110,100,40,115,111,99,107,101,116,46,115,111,117,114,99,101,40,34,117,110,116,105,108,45,99,108,111,115,101,100,34,44,32,115,101,108,102,46,99,41,41,10,32,32,32,32,115,111,117,114,99,101,40,115,116,97,116,117,115,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,116,114,121,40,108,116,110,49,50,46,112,117,109,112,46,97,108,108,40,115,111,117,114,99,101,44,32,115,105,110,107,44,32,115,116,101,112,41,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,99,108,111,115,101,40,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,99,58,99,108,111,115,101,40,41,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,72,105,103,104,32,108,101,118,101,108,32,72,84,84,80,32,65,80,73,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,97,100,106,117,115,116,117,114,105,40,114,101,113,116,41,10,32,32,32,32,108,111,99,97,108,32,117,32,61,32,114,101,113,116,10,32,32,32,32,45,45,32,105,102,32,116,104,101,114,101,32,105,115,32,97,32,112,114,111,120,121,44,32,119,101,32,110,101,101,100,32,116,104,101,32,102,117,108,108,32,117,114,108,46,32,111,116,104,101,114,119,105,115,101,44,32,106,117,115,116,32,97,32,112,97,114,116,46,10,32,32,32,32,105,102,32,110,111,116,32,114,101,113,116,46,112,114,111,120,121,32,97,110,100,32,110,111,116,32,80,82,79,88,89,32,116,104,101,110,10,32,32,32,32,32,32,32,32,117,32,61,32,123,10,32,32,32,32,32,32,32,32,32,32,32,112,97,116,104,32,61,32,115,111,99,107,101,116,46,116,114,121,40,114,101,113,116,46,112,97,116,104,44,32,34,105,110,118,97,108,105,100,32,112,97,116,104,32,39,110,105,108,39,34,41,44,10,32,32,32,32,32,32,32,32,32,32,32,112,97,114,97,109,115,32,61,32,114,101,113,116,46,112,97,114,97,109,115,44,10,32,32,32,32,32,32,32,32,32,32,32,113,117,101,114,121,32,61,32,114,101,113,116,46,113,117,101,114,121,44,10,32,32,32,32,32,32,32,32,32,32,32,102,114,97,103,109,101,110,116,32,61,32,114,101,113,116,46,102,114,97,103,109,101,110,116,10,32,32,32,32,32,32,32,32,125,10,32,32,32,32,101,110,100,10,32,32,32,32,114,101,116,117,114,110,32,117,114,108,46,98,117,105,108,100,40,117,41,10,101,110,100,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,97,100,106,117,115,116,112,114,111,120,121,40,114,101,113,116,41,10,32,32,32,32,108,111,99,97,108,32,112,114,111,120,121,32,61,32,114,101,113,116,46,112,114,111,120,121,32,111,114,32,80,82,79,88,89,10,32,32,32,32,105,102,32,112,114,111,120,121,32,116,104,101,110,10,32,32,32,32,32,32,32,32,112,114,111,120,121,32,61,32,117,114,108,46,112,97,114,115,101,40,112,114,111,120,121,41,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,112,114,111,120,121,46,104,111,115,116,44,32,112,114,111,120,121,46,112,111,114,116,32,111,114,32,51,49,50,56,10,32,32,32,32,101,108,115,101,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,114,101,113,116,46,104,111,115,116,44,32,114,101,113,116,46,112,111,114,116,10,32,32,32,32,101,110,100,10,101,110,100,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,97,100,106,117,115,116,104,101,97,100,101,114,115,40,114,101,113,116,41,10,32,32,32,32,45,45,32,100,101,102,97,117,108,116,32,104,101,97,100,101,114,115,10,32,32,32,32,108,111,99,97,108,32,108,111,119,101,114,32,61,32,123,10,32,32,32,32,32,32,32,32,91,34,117,115,101,114,45,97,103,101,110,116,34,93,32,61,32,85,83,69,82,65,71,69,78,84,44,10,32,32,32,32,32,32,32,32,91,34,104,111,115,116,34,93,32,61,32,114,101,113,116,46,104,111,115,116,44,10,32,32,32,32,32,32,32,32,91,34,99,111,110,110,101,99,116,105,111,110,34,93,32,61,32,34,99,108,111,115,101,44,32,84,69,34,44,10,32,32,32,32,32,32,32,32,91,34,116,101,34,93,32,61,32,34,116,114,97,105,108,101,114,115,34,10,32,32,32,32,125,10,32,32,32,32,45,45,32,105,102,32,119,101,32,104,97,118,101,32,97,117,116,104,101,110,116,105,99,97,116,105,111,110,32,105,110,102,111,114,109,97,116,105,111,110,44,32,112,97,115,115,32,105,116,32,97,108,111,110,103,10,32,32,32,32,105,102,32,114,101,113,116,46,117,115,101,114,32,97,110,100,32,114,101,113,116,46,112,97,115,115,119,111,114,100,32,116,104,101,110,10,32,32,32,32,32,32,32,32,108,111,119,101,114,91,34,97,117,116,104,111,114,105,122,97,116,105,111,110,34,93,32,61,32,10,32,32,32,32,32,32,32,32,32,32,32,32,34,66,97,115,105,99,32,34,32,46,46,32,32,40,109,105,109,101,46,98,54,52,40,114,101,113,116,46,117,115,101,114,32,46,46,32,34,58,34,32,46,46,32,114,101,113,116,46,112,97,115,115,119,111,114,100,41,41,10,32,32,32,32,101,110,100,10,32,32,32,32,45,45,32,111,118,101,114,114,105,100,101,32,119,105,116,104,32,117,115,101,114,32,104,101,97,100,101,114,115,10,32,32,32,32,102,111,114,32,105,44,118,32,105,110,32,98,97,115,101,46,112,97,105,114,115,40,114,101,113,116,46,104,101,97,100,101,114,115,32,111,114,32,108,111,119,101,114,41,32,100,111,10,32,32,32,32,32,32,32,32,108,111,119,101,114,91,115,116,114,105,110,103,46,108,111,119,101,114,40,105,41,93,32,61,32,118,10,32,32,32,32,101,110,100,10,32,32,32,32,114,101,116,117,114,110,32,108,111,119,101,114,10,101,110,100,10,10,45,45,32,100,101,102,97,117,108,116,32,117,114,108,32,112,97,114,116,115,10,108,111,99,97,108,32,100,101,102,97,117,108,116,32,61,32,123,10,32,32,32,32,104,111,115,116,32,61,32,34,34,44,10,32,32,32,32,112,111,114,116,32,61,32,80,79,82,84,44,10,32,32,32,32,112,97,116,104,32,61,34,47,34,44,10,32,32,32,32,115,99,104,101,109,101,32,61,32,34,104,116,116,112,34,10,125,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,97,100,106,117,115,116,114,101,113,117,101,115,116,40,114,101,113,116,41,10,32,32,32,32,45,45,32,112,97,114,115,101,32,117,114,108,32,105,102,32,112,114,111,118,105,100,101,100,10,32,32,32,32,108,111,99,97,108,32,110,114,101,113,116,32,61,32,114,101,113,116,46,117,114,108,32,97,110,100,32,117,114,108,46,112,97,114,115,101,40,114,101,113,116,46,117,114,108,44,32,100,101,102,97,117,108,116,41,32,111,114,32,123,125,10,32,32,32,32,45,45,32,101,120,112,108,105,99,105,116,32,99,111,109,112,111,110,101,110,116,115,32,111,118,101,114,114,105,100,101,32,117,114,108,10,32,32,32,32,102,111,114,32,105,44,118,32,105,110,32,98,97,115,101,46,112,97,105,114,115,40,114,101,113,116,41,32,100,111,32,110,114,101,113,116,91,105,93,32,61,32,118,32,101,110,100,10,32,32,32,32,105,102,32,110,114,101,113,116,46,112,111,114,116,32,61,61,32,34,34,32,116,104,101,110,32,110,114,101,113,116,46,112,111,114,116,32,61,32,56,48,32,101,110,100,10,32,32,32,32,115,111,99,107,101,116,46,116,114,121,40,110,114,101,113,116,46,104,111,115,116,32,97,110,100,32,110,114,101,113,116,46,104,111,115,116,32,126,61,32,34,34,44,32,10,32,32,32,32,32,32,32,32,34,105,110,118,97,108,105,100,32,104,111,115,116,32,39,34,32,46,46,32,98,97,115,101,46,116,111,115,116,114,105,110,103,40,110,114,101,113,116,46,104,111,115,116,41,32,46,46,32,34,39,34,41,10,32,32,32,32,45,45,32,99,111,109,112,117,116,101,32,117,114,105,32,105,102,32,117,115,101,114,32,104,97,115,110,39,116,32,111,118,101,114,114,105,100,101,110,10,32,32,32,32,110,114,101,113,116,46,117,114,105,32,61,32,114,101,113,116,46,117,114,105,32,111,114,32,97,100,106,117,115,116,117,114,105,40,110,114,101,113,116,41,10,32,32,32,32,45,45,32,97,106,117,115,116,32,104,111,115,116,32,97,110,100,32,112,111,114,116,32,105,102,32,116,104,101,114,101,32,105,115,32,97,32,112,114,111,120,121,10,32,32,32,32,110,114,101,113,116,46,104,111,115,116,44,32,110,114,101,113,116,46,112,111,114,116,32,61,32,97,100,106,117,115,116,112,114,111,120,121,40,110,114,101,113,116,41,10,32,32,32,32,45,45,32,97,100,106,117,115,116,32,104,101,97,100,101,114,115,32,105,110,32,114,101,113,117,101,115,116,10,32,32,32,32,110,114,101,113,116,46,104,101,97,100,101,114,115,32,61,32,97,100,106,117,115,116,104,101,97,100,101,114,115,40,110,114,101,113,116,41,10,32,32,32,32,114,101,116,117,114,110,32,110,114,101,113,116,10,101,110,100,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,115,104,111,117,108,100,114,101,100,105,114,101,99,116,40,114,101,113,116,44,32,99,111,100,101,44,32,104,101,97,100,101,114,115,41,10,32,32,32,32,114,101,116,117,114,110,32,104,101,97,100,101,114,115,46,108,111,99,97,116,105,111,110,32,97,110,100,10,32,32,32,32,32,32,32,32,32,32,32,115,116,114,105,110,103,46,103,115,117,98,40,104,101,97,100,101,114,115,46,108,111,99,97,116,105,111,110,44,32,34,37,115,34,44,32,34,34,41,32,126,61,32,34,34,32,97,110,100,10,32,32,32,32,32,32,32,32,32,32,32,40,114,101,113,116,46,114,101,100,105,114,101,99,116,32,126,61,32,102,97,108,115,101,41,32,97,110,100,10,32,32,32,32,32,32,32,32,32,32,32,40,99,111,100,101,32,61,61,32,51,48,49,32,111,114,32,99,111,100,101,32,61,61,32,51,48,50,41,32,97,110,100,10,32,32,32,32,32,32,32,32,32,32,32,40,110,111,116,32,114,101,113,116,46,109,101,116,104,111,100,32,111,114,32,114,101,113,116,46,109,101,116,104,111,100,32,61,61,32,34,71,69,84,34,32,111,114,32,114,101,113,116,46,109,101,116,104,111,100,32,61,61,32,34,72,69,65,68,34,41,10,32,32,32,32,32,32,32,32,32,32,32,97,110,100,32,40,110,111,116,32,114,101,113,116,46,110,114,101,100,105,114,101,99,116,115,32,111,114,32,114,101,113,116,46,110,114,101,100,105,114,101,99,116,115,32,60,32,53,41,10,101,110,100,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,115,104,111,117,108,100,114,101,99,101,105,118,101,98,111,100,121,40,114,101,113,116,44,32,99,111,100,101,41,10,32,32,32,32,105,102,32,114,101,113,116,46,109,101,116,104,111,100,32,61,61,32,34,72,69,65,68,34,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,32,101,110,100,10,32,32,32,32,105,102,32,99,111,100,101,32,61,61,32,50,48,52,32,111,114,32,99,111,100,101,32,61,61,32,51,48,52,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,32,101,110,100,10,32,32,32,32,105,102,32,99,111,100,101,32,62,61,32,49,48,48,32,97,110,100,32,99,111,100,101,32,60,32,50,48,48,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,32,101,110,100,10,32,32,32,32,114,101,116,117,114,110,32,49,10,101,110,100,10,10,45,45,32,102,111,114,119,97,114,100,32,100,101,99,108,97,114,97,116,105,111,110,115,10,108,111,99,97,108,32,116,114,101,113,117,101,115,116,44,32,116,114,101,100,105,114,101,99,116,10,10,102,117,110,99,116,105,111,110,32,116,114,101,100,105,114,101,99,116,40,114,101,113,116,44,32,108,111,99,97,116,105,111,110,41,10,32,32,32,32,108,111,99,97,108,32,114,101,115,117,108,116,44,32,99,111,100,101,44,32,104,101,97,100,101,114,115,44,32,115,116,97,116,117,115,32,61,32,116,114,101,113,117,101,115,116,32,123,10,32,32,32,32,32,32,32,32,45,45,32,116,104,101,32,82,70,67,32,115,97,121,115,32,116,104,101,32,114,101,100,105,114,101,99,116,32,85,82,76,32,104,97,115,32,116,111,32,98,101,32,97,98,115,111,108,117,116,101,44,32,98,117,116,32,115,111,109,101,10,32,32,32,32,32,32,32,32,45,45,32,115,101,114,118,101,114,115,32,100,111,32,110,111,116,32,114,101,115,112,101,99,116,32,116,104,97,116,10,32,32,32,32,32,32,32,32,117,114,108,32,61,32,117,114,108,46,97,98,115,111,108,117,116,101,40,114,101,113,116,46,117,114,108,44,32,108,111,99,97,116,105,111,110,41,44,10,32,32,32,32,32,32,32,32,115,111,117,114,99,101,32,61,32,114,101,113,116,46,115,111,117,114,99,101,44,10,32,32,32,32,32,32,32,32,115,105,110,107,32,61,32,114,101,113,116,46,115,105,110,107,44,10,32,32,32,32,32,32,32,32,104,101,97,100,101,114,115,32,61,32,114,101,113,116,46,104,101,97,100,101,114,115,44,10,32,32,32,32,32,32,32,32,112,114,111,120,121,32,61,32,114,101,113,116,46,112,114,111,120,121,44,32,10,32,32,32,32,32,32,32,32,110,114,101,100,105,114,101,99,116,115,32,61,32,40,114,101,113,116,46,110,114,101,100,105,114,101,99,116,115,32,111,114,32,48,41,32,43,32,49,44,10,32,32,32,32,32,32,32,32,99,114,101,97,116,101,32,61,32,114,101,113,116,46,99,114,101,97,116,101,10,32,32,32,32,125,32,32,32,10,32,32,32,32,45,45,32,112,97,115,115,32,108,111,99,97,116,105,111,110,32,104,101,97,100,101,114,32,98,97,99,107,32,97,115,32,97,32,104,105,110,116,32,119,101,32,114,101,100,105,114,101,99,116,101,100,10,32,32,32,32,104,101,97,100,101,114,115,32,61,32,104,101,97,100,101,114,115,32,111,114,32,123,125,10,32,32,32,32,104,101,97,100,101,114,115,46,108,111,99,97,116,105,111,110,32,61,32,104,101,97,100,101,114,115,46,108,111,99,97,116,105,111,110,32,111,114,32,108,111,99,97,116,105,111,110,10,32,32,32,32,114,101,116,117,114,110,32,114,101,115,117,108,116,44,32,99,111,100,101,44,32,104,101,97,100,101,114,115,44,32,115,116,97,116,117,115,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,116,114,101,113,117,101,115,116,40,114,101,113,116,41,10,32,32,32,32,45,45,32,119,101,32,108,111,111,112,32,117,110,116,105,108,32,119,101,32,103,101,116,32,119,104,97,116,32,119,101,32,119,97,110,116,44,32,111,114,10,32,32,32,32,45,45,32,117,110,116,105,108,32,119,101,32,97,114,101,32,115,117,114,101,32,116,104,101,114,101,32,105,115,32,110,111,32,119,97,121,32,116,111,32,103,101,116,32,105,116,10,32,32,32,32,108,111,99,97,108,32,110,114,101,113,116,32,61,32,97,100,106,117,115,116,114,101,113,117,101,115,116,40,114,101,113,116,41,10,32,32,32,32,108,111,99,97,108,32,104,32,61,32,111,112,101,110,40,110,114,101,113,116,46,104,111,115,116,44,32,110,114,101,113,116,46,112,111,114,116,44,32,110,114,101,113,116,46,99,114,101,97,116,101,41,10,32,32,32,32,45,45,32,115,101,110,100,32,114,101,113,117,101,115,116,32,108,105,110,101,32,97,110,100,32,104,101,97,100,101,114,115,10,32,32,32,32,104,58,115,101,110,100,114,101,113,117,101,115,116,108,105,110,101,40,110,114,101,113,116,46,109,101,116,104,111,100,44,32,110,114,101,113,116,46,117,114,105,41,10,32,32,32,32,104,58,115,101,110,100,104,101,97,100,101,114,115,40,110,114,101,113,116,46,104,101,97,100,101,114,115,41,10,32,32,32,32,45,45,32,105,102,32,116,104,101,114,101,32,105,115,32,97,32,98,111,100,121,44,32,115,101,110,100,32,105,116,10,32,32,32,32,105,102,32,110,114,101,113,116,46,115,111,117,114,99,101,32,116,104,101,110,10,32,32,32,32,32,32,32,32,104,58,115,101,110,100,98,111,100,121,40,110,114,101,113,116,46,104,101,97,100,101,114,115,44,32,110,114,101,113,116,46,115,111,117,114,99,101,44,32,110,114,101,113,116,46,115,116,101,112,41,32,10,32,32,32,32,101,110,100,10,32,32,32,32,108,111,99,97,108,32,99,111,100,101,44,32,115,116,97,116,117,115,32,61,32,104,58,114,101,99,101,105,118,101,115,116,97,116,117,115,108,105,110,101,40,41,10,32,32,32,32,45,45,32,105,102,32,105,116,32,105,115,32,97,110,32,72,84,84,80,47,48,46,57,32,115,101,114,118,101,114,44,32,115,105,109,112,108,121,32,103,101,116,32,116,104,101,32,98,111,100,121,32,97,110,100,32,119,101,32,97,114,101,32,100,111,110,101,10,32,32,32,32,105,102,32,110,111,116,32,99,111,100,101,32,116,104,101,110,10,32,32,32,32,32,32,32,32,104,58,114,101,99,101,105,118,101,48,57,98,111,100,121,40,115,116,97,116,117,115,44,32,110,114,101,113,116,46,115,105,110,107,44,32,110,114,101,113,116,46,115,116,101,112,41,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,49,44,32,50,48,48,10,32,32,32,32,101,110,100,10,32,32,32,32,108,111,99,97,108,32,104,101,97,100,101,114,115,10,32,32,32,32,45,45,32,105,103,110,111,114,101,32,97,110,121,32,49,48,48,45,99,111,110,116,105,110,117,101,32,109,101,115,115,97,103,101,115,10,32,32,32,32,119,104,105,108,101,32,99,111,100,101,32,61,61,32,49,48,48,32,100,111,32,10,32,32,32,32,32,32,32,32,104,101,97,100,101,114,115,32,61,32,104,58,114,101,99,101,105,118,101,104,101,97,100,101,114,115,40,41,10,32,32,32,32,32,32,32,32,99,111,100,101,44,32,115,116,97,116,117,115,32,61,32,104,58,114,101,99,101,105,118,101,115,116,97,116,117,115,108,105,110,101,40,41,10,32,32,32,32,101,110,100,10,32,32,32,32,104,101,97,100,101,114,115,32,61,32,104,58,114,101,99,101,105,118,101,104,101,97,100,101,114,115,40,41,10,32,32,32,32,45,45,32,97,116,32,116,104,105,115,32,112,111,105,110,116,32,119,101,32,115,104,111,117,108,100,32,104,97,118,101,32,97,32,104,111,110,101,115,116,32,114,101,112,108,121,32,102,114,111,109,32,116,104,101,32,115,101,114,118,101,114,10,32,32,32,32,45,45,32,119,101,32,99,97,110,39,116,32,114,101,100,105,114,101,99,116,32,105,102,32,119,101,32,97,108,114,101,97,100,121,32,117,115,101,100,32,116,104,101,32,115,111,117,114,99,101,44,32,115,111,32,119,101,32,114,101,112,111,114,116,32,116,104,101,32,101,114,114,111,114,32,10,32,32,32,32,105,102,32,115,104,111,117,108,100,114,101,100,105,114,101,99,116,40,110,114,101,113,116,44,32,99,111,100,101,44,32,104,101,97,100,101,114,115,41,32,97,110,100,32,110,111,116,32,110,114,101,113,116,46,115,111,117,114,99,101,32,116,104,101,110,10,32,32,32,32,32,32,32,32,104,58,99,108,111,115,101,40,41,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,116,114,101,100,105,114,101,99,116,40,114,101,113,116,44,32,104,101,97,100,101,114,115,46,108,111,99,97,116,105,111,110,41,10,32,32,32,32,101,110,100,10,32,32,32,32,45,45,32,104,101,114,101,32,119,101,32,97,114,101,32,102,105,110,97,108,108,121,32,100,111,110,101,10,32,32,32,32,105,102,32,115,104,111,117,108,100,114,101,99,101,105,118,101,98,111,100,121,40,110,114,101,113,116,44,32,99,111,100,101,41,32,116,104,101,110,10,32,32,32,32,32,32,32,32,104,58,114,101,99,101,105,118,101,98,111,100,121,40,104,101,97,100,101,114,115,44,32,110,114,101,113,116,46,115,105,110,107,44,32,110,114,101,113,116,46,115,116,101,112,41,10,32,32,32,32,101,110,100,10,32,32,32,32,104,58,99,108,111,115,101,40,41,10,32,32,32,32,114,101,116,117,114,110,32,49,44,32,99,111,100,101,44,32,104,101,97,100,101,114,115,44,32,115,116,97,116,117,115,10,101,110,100,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,115,114,101,113,117,101,115,116,40,117,44,32,98,41,10,32,32,32,32,108,111,99,97,108,32,116,32,61,32,123,125,10,32,32,32,32,108,111,99,97,108,32,114,101,113,116,32,61,32,123,10,32,32,32,32,32,32,32,32,117,114,108,32,61,32,117,44,10,32,32,32,32,32,32,32,32,115,105,110,107,32,61,32,108,116,110,49,50,46,115,105,110,107,46,116,97,98,108,101,40,116,41,10,32,32,32,32,125,10,32,32,32,32,105,102,32,98,32,116,104,101,110,10,32,32,32,32,32,32,32,32,114,101,113,116,46,115,111,117,114,99,101,32,61,32,108,116,110,49,50,46,115,111,117,114,99,101,46,115,116,114,105,110,103,40,98,41,10,32,32,32,32,32,32,32,32,114,101,113,116,46,104,101,97,100,101,114,115,32,61,32,123,10,32,32,32,32,32,32,32,32,32,32,32,32,91,34,99,111,110,116,101,110,116,45,108,101,110,103,116,104,34,93,32,61,32,115,116,114,105,110,103,46,108,101,110,40,98,41,44,10,32,32,32,32,32,32,32,32,32,32,32,32,91,34,99,111,110,116,101,110,116,45,116,121,112,101,34,93,32,61,32,34,97,112,112,108,105,99,97,116,105,111,110,47,120,45,119,119,119,45,102,111,114,109,45,117,114,108,101,110,99,111,100,101,100,34,10,32,32,32,32,32,32,32,32,125,10,32,32,32,32,32,32,32,32,114,101,113,116,46,109,101,116,104,111,100,32,61,32,34,80,79,83,84,34,10,32,32,32,32,101,110,100,10,32,32,32,32,108,111,99,97,108,32,99,111,100,101,44,32,104,101,97,100,101,114,115,44,32,115,116,97,116,117,115,32,61,32,115,111,99,107,101,116,46,115,107,105,112,40,49,44,32,116,114,101,113,117,101,115,116,40,114,101,113,116,41,41,10,32,32,32,32,114,101,116,117,114,110,32,116,97,98,108,101,46,99,111,110,99,97,116,40,116,41,44,32,99,111,100,101,44,32,104,101,97,100,101,114,115,44,32,115,116,97,116,117,115,10,101,110,100,10,10,114,101,113,117,101,115,116,32,61,32,115,111,99,107,101,116,46,112,114,111,116,101,99,116,40,102,117,110,99,116,105,111,110,40,114,101,113,116,44,32,98,111,100,121,41,10,32,32,32,32,105,102,32,98,97,115,101,46,116,121,112,101,40,114,101,113,116,41,32,61,61,32,34,115,116,114,105,110,103,34,32,116,104,101,110,32,114,101,116,117,114,110,32,115,114,101,113,117,101,115,116,40,114,101,113,116,44,32,98,111,100,121,41,10,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,116,114,101,113,117,101,115,116,40,114,101,113,116,41,32,101,110,100,10,101,110,100,41,10,104,116,116,112,46,108,117,97,0,115,111,99,107,101,116,46,104,116,116,112,0,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,76,84,78,49,50,32,45,32,70,105,108,116,101,114,115,44,32,115,111,117,114,99,101,115,44,32,115,105,110,107,115,32,97,110,100,32,112,117,109,112,115,46,10,45,45,32,76,117,97,83,111,99,107,101,116,32,116,111,111,108,107,105,116,46,10,45,45,32,65,117,116,104,111,114,58,32,68,105,101,103,111,32,78,101,104,97,98,10,45,45,32,82,67,83,32,73,68,58,32,36,73,100,58,32,108,116,110,49,50,46,108,117,97,44,118,32,49,46,51,49,32,50,48,48,54,47,48,52,47,48,51,32,48,52,58,52,53,58,52,50,32,100,105,101,103,111,32,69,120,112,32,36,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,68,101,99,108,97,114,101,32,109,111,100,117,108,101,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,108,111,99,97,108,32,115,116,114,105,110,103,32,61,32,114,101,113,117,105,114,101,40,34,115,116,114,105,110,103,34,41,10,108,111,99,97,108,32,116,97,98,108,101,32,61,32,114,101,113,117,105,114,101,40,34,116,97,98,108,101,34,41,10,108,111,99,97,108,32,98,97,115,101,32,61,32,95,71,10,109,111,100,117,108,101,40,34,108,116,110,49,50,34,41,10,10,102,105,108,116,101,114,32,61,32,123,125,10,115,111,117,114,99,101,32,61,32,123,125,10,115,105,110,107,32,61,32,123,125,10,112,117,109,112,32,61,32,123,125,10,10,45,45,32,50,48,52,56,32,115,101,101,109,115,32,116,111,32,98,101,32,98,101,116,116,101,114,32,105,110,32,119,105,110,100,111,119,115,46,46,46,10,66,76,79,67,75,83,73,90,69,32,61,32,50,48,52,56,10,95,86,69,82,83,73,79,78,32,61,32,34,76,84,78,49,50,32,49,46,48,46,49,34,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,70,105,108,116,101,114,32,115,116,117,102,102,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,114,101,116,117,114,110,115,32,97,32,104,105,103,104,32,108,101,118,101,108,32,102,105,108,116,101,114,32,116,104,97,116,32,99,121,99,108,101,115,32,97,32,108,111,119,45,108,101,118,101,108,32,102,105,108,116,101,114,10,102,117,110,99,116,105,111,110,32,102,105,108,116,101,114,46,99,121,99,108,101,40,108,111,119,44,32,99,116,120,44,32,101,120,116,114,97,41,10,32,32,32,32,98,97,115,101,46,97,115,115,101,114,116,40,108,111,119,41,10,32,32,32,32,114,101,116,117,114,110,32,102,117,110,99,116,105,111,110,40,99,104,117,110,107,41,10,32,32,32,32,32,32,32,32,108,111,99,97,108,32,114,101,116,10,32,32,32,32,32,32,32,32,114,101,116,44,32,99,116,120,32,61,32,108,111,119,40,99,116,120,44,32,99,104,117,110,107,44,32,101,120,116,114,97,41,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,114,101,116,10,32,32,32,32,101,110,100,10,101,110,100,10,10,45,45,32,99,104,97,105,110,115,32,97,32,98,117,110,99,104,32,111,102,32,102,105,108,116,101,114,115,32,116,111,103,101,116,104,101,114,10,45,45,32,40,116,104,97,110,107,115,32,116,111,32,87,105,109,32,67,111,117,119,101,110,98,101,114,103,41,10,102,117,110,99,116,105,111,110,32,102,105,108,116,101,114,46,99,104,97,105,110,40,46,46,46,41,10,32,32,32,32,108,111,99,97,108,32,110,32,61,32,116,97,98,108,101,46,103,101,116,110,40,97,114,103,41,10,32,32,32,32,108,111,99,97,108,32,116,111,112,44,32,105,110,100,101,120,32,61,32,49,44,32,49,10,32,32,32,32,108,111,99,97,108,32,114,101,116,114,121,32,61,32,34,34,10,32,32,32,32,114,101,116,117,114,110,32,102,117,110,99,116,105,111,110,40,99,104,117,110,107,41,10,32,32,32,32,32,32,32,32,114,101,116,114,121,32,61,32,99,104,117,110,107,32,97,110,100,32,114,101,116,114,121,10,32,32,32,32,32,32,32,32,119,104,105,108,101,32,116,114,117,101,32,100,111,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,105,110,100,101,120,32,61,61,32,116,111,112,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,99,104,117,110,107,32,61,32,97,114,103,91,105,110,100,101,120,93,40,99,104,117,110,107,41,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,99,104,117,110,107,32,61,61,32,34,34,32,111,114,32,116,111,112,32,61,61,32,110,32,116,104,101,110,32,114,101,116,117,114,110,32,99,104,117,110,107,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,105,102,32,99,104,117,110,107,32,116,104,101,110,32,105,110,100,101,120,32,61,32,105,110,100,101,120,32,43,32,49,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,116,111,112,32,61,32,116,111,112,43,49,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,110,100,101,120,32,61,32,116,111,112,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,99,104,117,110,107,32,61,32,97,114,103,91,105,110,100,101,120,93,40,99,104,117,110,107,32,111,114,32,34,34,41,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,99,104,117,110,107,32,61,61,32,34,34,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,110,100,101,120,32,61,32,105,110,100,101,120,32,45,32,49,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,99,104,117,110,107,32,61,32,114,101,116,114,121,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,105,102,32,99,104,117,110,107,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,105,110,100,101,120,32,61,61,32,110,32,116,104,101,110,32,114,101,116,117,114,110,32,99,104,117,110,107,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,32,105,110,100,101,120,32,61,32,105,110,100,101,120,32,43,32,49,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,32,98,97,115,101,46,101,114,114,111,114,40,34,102,105,108,116,101,114,32,114,101,116,117,114,110,101,100,32,105,110,97,112,112,114,111,112,114,105,97,116,101,32,110,105,108,34,41,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,101,110,100,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,83,111,117,114,99,101,32,115,116,117,102,102,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,99,114,101,97,116,101,32,97,110,32,101,109,112,116,121,32,115,111,117,114,99,101,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,101,109,112,116,121,40,41,10,32,32,32,32,114,101,116,117,114,110,32,110,105,108,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,115,111,117,114,99,101,46,101,109,112,116,121,40,41,10,32,32,32,32,114,101,116,117,114,110,32,101,109,112,116,121,10,101,110,100,10,10,45,45,32,114,101,116,117,114,110,115,32,97,32,115,111,117,114,99,101,32,116,104,97,116,32,106,117,115,116,32,111,117,116,112,117,116,115,32,97,110,32,101,114,114,111,114,10,102,117,110,99,116,105,111,110,32,115,111,117,114,99,101,46,101,114,114,111,114,40,101,114,114,41,10,32,32,32,32,114,101,116,117,114,110,32,102,117,110,99,116,105,111,110,40,41,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,10,32,32,32,32,101,110,100,10,101,110,100,10,10,45,45,32,99,114,101,97,116,101,115,32,97,32,102,105,108,101,32,115,111,117,114,99,101,10,102,117,110,99,116,105,111,110,32,115,111,117,114,99,101,46,102,105,108,101,40,104,97,110,100,108,101,44,32,105,111,95,101,114,114,41,10,32,32,32,32,105,102,32,104,97,110,100,108,101,32,116,104,101,110,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,102,117,110,99,116,105,111,110,40,41,10,32,32,32,32,32,32,32,32,32,32,32,32,108,111,99,97,108,32,99,104,117,110,107,32,61,32,104,97,110,100,108,101,58,114,101,97,100,40,66,76,79,67,75,83,73,90,69,41,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,99,104,117,110,107,32,116,104,101,110,32,104,97,110,100,108,101,58,99,108,111,115,101,40,41,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,99,104,117,110,107,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,115,111,117,114,99,101,46,101,114,114,111,114,40,105,111,95,101,114,114,32,111,114,32,34,117,110,97,98,108,101,32,116,111,32,111,112,101,110,32,102,105,108,101,34,41,32,101,110,100,10,101,110,100,10,10,45,45,32,116,117,114,110,115,32,97,32,102,97,110,99,121,32,115,111,117,114,99,101,32,105,110,116,111,32,97,32,115,105,109,112,108,101,32,115,111,117,114,99,101,10,102,117,110,99,116,105,111,110,32,115,111,117,114,99,101,46,115,105,109,112,108,105,102,121,40,115,114,99,41,10,32,32,32,32,98,97,115,101,46,97,115,115,101,114,116,40,115,114,99,41,10,32,32,32,32,114,101,116,117,114,110,32,102,117,110,99,116,105,111,110,40,41,10,32,32,32,32,32,32,32,32,108,111,99,97,108,32,99,104,117,110,107,44,32,101,114,114,95,111,114,95,110,101,119,32,61,32,115,114,99,40,41,10,32,32,32,32,32,32,32,32,115,114,99,32,61,32,101,114,114,95,111,114,95,110,101,119,32,111,114,32,115,114,99,10,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,99,104,117,110,107,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,95,111,114,95,110,101,119,10,32,32,32,32,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,99,104,117,110,107,32,101,110,100,10,32,32,32,32,101,110,100,10,101,110,100,10,10,45,45,32,99,114,101,97,116,101,115,32,115,116,114,105,110,103,32,115,111,117,114,99,101,10,102,117,110,99,116,105,111,110,32,115,111,117,114,99,101,46,115,116,114,105,110,103,40,115,41,10,32,32,32,32,105,102,32,115,32,116,104,101,110,10,32,32,32,32,32,32,32,32,108,111,99,97,108,32,105,32,61,32,49,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,102,117,110,99,116,105,111,110,40,41,10,32,32,32,32,32,32,32,32,32,32,32,32,108,111,99,97,108,32,99,104,117,110,107,32,61,32,115,116,114,105,110,103,46,115,117,98,40,115,44,32,105,44,32,105,43,66,76,79,67,75,83,73,90,69,45,49,41,10,32,32,32,32,32,32,32,32,32,32,32,32,105,32,61,32,105,32,43,32,66,76,79,67,75,83,73,90,69,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,99,104,117,110,107,32,126,61,32,34,34,32,116,104,101,110,32,114,101,116,117,114,110,32,99,104,117,110,107,10,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,110,105,108,32,101,110,100,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,115,111,117,114,99,101,46,101,109,112,116,121,40,41,32,101,110,100,10,101,110,100,10,10,45,45,32,99,114,101,97,116,101,115,32,114,101,119,105,110,100,97,98,108,101,32,115,111,117,114,99,101,10,102,117,110,99,116,105,111,110,32,115,111,117,114,99,101,46,114,101,119,105,110,100,40,115,114,99,41,10,32,32,32,32,98,97,115,101,46,97,115,115,101,114,116,40,115,114,99,41,10,32,32,32,32,108,111,99,97,108,32,116,32,61,32,123,125,10,32,32,32,32,114,101,116,117,114,110,32,102,117,110,99,116,105,111,110,40,99,104,117,110,107,41,10,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,99,104,117,110,107,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,99,104,117,110,107,32,61,32,116,97,98,108,101,46,114,101,109,111,118,101,40,116,41,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,99,104,117,110,107,32,116,104,101,110,32,114,101,116,117,114,110,32,115,114,99,40,41,10,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,99,104,117,110,107,32,101,110,100,10,32,32,32,32,32,32,32,32,101,108,115,101,10,32,32,32,32,32,32,32,32,32,32,32,32,116,97,98,108,101,46,105,110,115,101,114,116,40,116,44,32,99,104,117,110,107,41,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,101,110,100,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,115,111,117,114,99,101,46,99,104,97,105,110,40,115,114,99,44,32,102,41,10,32,32,32,32,98,97,115,101,46,97,115,115,101,114,116,40,115,114,99,32,97,110,100,32,102,41,10,32,32,32,32,108,111,99,97,108,32,108,97,115,116,95,105,110,44,32,108,97,115,116,95,111,117,116,32,61,32,34,34,44,32,34,34,10,32,32,32,32,108,111,99,97,108,32,115,116,97,116,101,32,61,32,34,102,101,101,100,105,110,103,34,10,32,32,32,32,108,111,99,97,108,32,101,114,114,10,32,32,32,32,114,101,116,117,114,110,32,102,117,110,99,116,105,111,110,40,41,10,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,108,97,115,116,95,111,117,116,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,98,97,115,101,46,101,114,114,111,114,40,39,115,111,117,114,99,101,32,105,115,32,101,109,112,116,121,33,39,44,32,50,41,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,119,104,105,108,101,32,116,114,117,101,32,100,111,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,115,116,97,116,101,32,61,61,32,34,102,101,101,100,105,110,103,34,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,108,97,115,116,95,105,110,44,32,101,114,114,32,61,32,115,114,99,40,41,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,101,114,114,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,108,97,115,116,95,111,117,116,32,61,32,102,40,108,97,115,116,95,105,110,41,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,108,97,115,116,95,111,117,116,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,108,97,115,116,95,105,110,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,98,97,115,101,46,101,114,114,111,114,40,39,102,105,108,116,101,114,32,114,101,116,117,114,110,101,100,32,105,110,97,112,112,114,111,112,114,105,97,116,101,32,110,105,108,39,41,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,110,105,108,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,105,102,32,108,97,115,116,95,111,117,116,32,126,61,32,34,34,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,115,116,97,116,101,32,61,32,34,101,97,116,105,110,103,34,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,108,97,115,116,95,105,110,32,116,104,101,110,32,108,97,115,116,95,105,110,32,61,32,34,34,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,108,97,115,116,95,111,117,116,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,108,97,115,116,95,111,117,116,32,61,32,102,40,108,97,115,116,95,105,110,41,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,108,97,115,116,95,111,117,116,32,61,61,32,34,34,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,108,97,115,116,95,105,110,32,61,61,32,34,34,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,115,116,97,116,101,32,61,32,34,102,101,101,100,105,110,103,34,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,98,97,115,101,46,101,114,114,111,114,40,39,102,105,108,116,101,114,32,114,101,116,117,114,110,101,100,32,34,34,39,41,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,105,102,32,110,111,116,32,108,97,115,116,95,111,117,116,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,108,97,115,116,95,105,110,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,98,97,115,101,46,101,114,114,111,114,40,39,102,105,108,116,101,114,32,114,101,116,117,114,110,101,100,32,105,110,97,112,112,114,111,112,114,105,97,116,101,32,110,105,108,39,41,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,110,105,108,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,108,97,115,116,95,111,117,116,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,101,110,100,10,101,110,100,10,10,45,45,32,99,114,101,97,116,101,115,32,97,32,115,111,117,114,99,101,32,116,104,97,116,32,112,114,111,100,117,99,101,115,32,99,111,110,116,101,110,116,115,32,111,102,32,115,101,118,101,114,97,108,32,115,111,117,114,99,101,115,44,32,111,110,101,32,97,102,116,101,114,32,116,104,101,10,45,45,32,111,116,104,101,114,44,32,97,115,32,105,102,32,116,104,101,121,32,119,101,114,101,32,99,111,110,99,97,116,101,110,97,116,101,100,10,45,45,32,40,116,104,97,110,107,115,32,116,111,32,87,105,109,32,67,111,117,119,101,110,98,101,114,103,41,10,102,117,110,99,116,105,111,110,32,115,111,117,114,99,101,46,99,97,116,40,46,46,46,41,10,32,32,32,32,108,111,99,97,108,32,115,114,99,32,61,32,116,97,98,108,101,46,114,101,109,111,118,101,40,97,114,103,44,32,49,41,10,32,32,32,32,114,101,116,117,114,110,32,102,117,110,99,116,105,111,110,40,41,10,32,32,32,32,32,32,32,32,119,104,105,108,101,32,115,114,99,32,100,111,10,32,32,32,32,32,32,32,32,32,32,32,32,108,111,99,97,108,32,99,104,117,110,107,44,32,101,114,114,32,61,32,115,114,99,40,41,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,99,104,117,110,107,32,116,104,101,110,32,114,101,116,117,114,110,32,99,104,117,110,107,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,101,114,114,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,115,114,99,32,61,32,116,97,98,108,101,46,114,101,109,111,118,101,40,97,114,103,44,32,49,41,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,101,110,100,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,83,105,110,107,32,115,116,117,102,102,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,99,114,101,97,116,101,115,32,97,32,115,105,110,107,32,116,104,97,116,32,115,116,111,114,101,115,32,105,110,116,111,32,97,32,116,97,98,108,101,10,102,117,110,99,116,105,111,110,32,115,105,110,107,46,116,97,98,108,101,40,116,41,10,32,32,32,32,116,32,61,32,116,32,111,114,32,123,125,10,32,32,32,32,108,111,99,97,108,32,102,32,61,32,102,117,110,99,116,105,111,110,40,99,104,117,110,107,44,32,101,114,114,41,10,32,32,32,32,32,32,32,32,105,102,32,99,104,117,110,107,32,116,104,101,110,32,116,97,98,108,101,46,105,110,115,101,114,116,40,116,44,32,99,104,117,110,107,41,32,101,110,100,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,49,10,32,32,32,32,101,110,100,10,32,32,32,32,114,101,116,117,114,110,32,102,44,32,116,10,101,110,100,10,10,45,45,32,116,117,114,110,115,32,97,32,102,97,110,99,121,32,115,105,110,107,32,105,110,116,111,32,97,32,115,105,109,112,108,101,32,115,105,110,107,10,102,117,110,99,116,105,111,110,32,115,105,110,107,46,115,105,109,112,108,105,102,121,40,115,110,107,41,10,32,32,32,32,98,97,115,101,46,97,115,115,101,114,116,40,115,110,107,41,10,32,32,32,32,114,101,116,117,114,110,32,102,117,110,99,116,105,111,110,40,99,104,117,110,107,44,32,101,114,114,41,10,32,32,32,32,32,32,32,32,108,111,99,97,108,32,114,101,116,44,32,101,114,114,95,111,114,95,110,101,119,32,61,32,115,110,107,40,99,104,117,110,107,44,32,101,114,114,41,10,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,114,101,116,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,95,111,114,95,110,101,119,32,101,110,100,10,32,32,32,32,32,32,32,32,115,110,107,32,61,32,101,114,114,95,111,114,95,110,101,119,32,111,114,32,115,110,107,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,49,10,32,32,32,32,101,110,100,10,101,110,100,10,10,45,45,32,99,114,101,97,116,101,115,32,97,32,102,105,108,101,32,115,105,110,107,10,102,117,110,99,116,105,111,110,32,115,105,110,107,46,102,105,108,101,40,104,97,110,100,108,101,44,32,105,111,95,101,114,114,41,10,32,32,32,32,105,102,32,104,97,110,100,108,101,32,116,104,101,110,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,102,117,110,99,116,105,111,110,40,99,104,117,110,107,44,32,101,114,114,41,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,99,104,117,110,107,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,104,97,110,100,108,101,58,99,108,111,115,101,40,41,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,49,10,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,104,97,110,100,108,101,58,119,114,105,116,101,40,99,104,117,110,107,41,32,101,110,100,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,115,105,110,107,46,101,114,114,111,114,40,105,111,95,101,114,114,32,111,114,32,34,117,110,97,98,108,101,32,116,111,32,111,112,101,110,32,102,105,108,101,34,41,32,101,110,100,10,101,110,100,10,10,45,45,32,99,114,101,97,116,101,115,32,97,32,115,105,110,107,32,116,104,97,116,32,100,105,115,99,97,114,100,115,32,100,97,116,97,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,110,117,108,108,40,41,10,32,32,32,32,114,101,116,117,114,110,32,49,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,115,105,110,107,46,110,117,108,108,40,41,10,32,32,32,32,114,101,116,117,114,110,32,110,117,108,108,10,101,110,100,10,10,45,45,32,99,114,101,97,116,101,115,32,97,32,115,105,110,107,32,116,104,97,116,32,106,117,115,116,32,114,101,116,117,114,110,115,32,97,110,32,101,114,114,111,114,10,102,117,110,99,116,105,111,110,32,115,105,110,107,46,101,114,114,111,114,40,101,114,114,41,10,32,32,32,32,114,101,116,117,114,110,32,102,117,110,99,116,105,111,110,40,41,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,10,32,32,32,32,101,110,100,10,101,110,100,10,10,45,45,32,99,104,97,105,110,115,32,97,32,115,105,110,107,32,119,105,116,104,32,97,32,102,105,108,116,101,114,10,102,117,110,99,116,105,111,110,32,115,105,110,107,46,99,104,97,105,110,40,102,44,32,115,110,107,41,10,32,32,32,32,98,97,115,101,46,97,115,115,101,114,116,40,102,32,97,110,100,32,115,110,107,41,10,32,32,32,32,114,101,116,117,114,110,32,102,117,110,99,116,105,111,110,40,99,104,117,110,107,44,32,101,114,114,41,10,32,32,32,32,32,32,32,32,105,102,32,99,104,117,110,107,32,126,61,32,34,34,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,108,111,99,97,108,32,102,105,108,116,101,114,101,100,32,61,32,102,40,99,104,117,110,107,41,10,32,32,32,32,32,32,32,32,32,32,32,32,108,111,99,97,108,32,100,111,110,101,32,61,32,99,104,117,110,107,32,97,110,100,32,34,34,10,32,32,32,32,32,32,32,32,32,32,32,32,119,104,105,108,101,32,116,114,117,101,32,100,111,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,108,111,99,97,108,32,114,101,116,44,32,115,110,107,101,114,114,32,61,32,115,110,107,40,102,105,108,116,101,114,101,100,44,32,101,114,114,41,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,114,101,116,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,115,110,107,101,114,114,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,102,105,108,116,101,114,101,100,32,61,61,32,100,111,110,101,32,116,104,101,110,32,114,101,116,117,114,110,32,49,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,102,105,108,116,101,114,101,100,32,61,32,102,40,100,111,110,101,41,10,32,32,32,32,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,49,32,101,110,100,10,32,32,32,32,101,110,100,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,80,117,109,112,32,115,116,117,102,102,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,112,117,109,112,115,32,111,110,101,32,99,104,117,110,107,32,102,114,111,109,32,116,104,101,32,115,111,117,114,99,101,32,116,111,32,116,104,101,32,115,105,110,107,10,102,117,110,99,116,105,111,110,32,112,117,109,112,46,115,116,101,112,40,115,114,99,44,32,115,110,107,41,10,32,32,32,32,108,111,99,97,108,32,99,104,117,110,107,44,32,115,114,99,95,101,114,114,32,61,32,115,114,99,40,41,10,32,32,32,32,108,111,99,97,108,32,114,101,116,44,32,115,110,107,95,101,114,114,32,61,32,115,110,107,40,99,104,117,110,107,44,32,115,114,99,95,101,114,114,41,10,32,32,32,32,105,102,32,99,104,117,110,107,32,97,110,100,32,114,101,116,32,116,104,101,110,32,114,101,116,117,114,110,32,49,10,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,110,105,108,44,32,115,114,99,95,101,114,114,32,111,114,32,115,110,107,95,101,114,114,32,101,110,100,10,101,110,100,10,10,45,45,32,112,117,109,112,115,32,97,108,108,32,100,97,116,97,32,102,114,111,109,32,97,32,115,111,117,114,99,101,32,116,111,32,97,32,115,105,110,107,44,32,117,115,105,110,103,32,97,32,115,116,101,112,32,102,117,110,99,116,105,111,110,10,102,117,110,99,116,105,111,110,32,112,117,109,112,46,97,108,108,40,115,114,99,44,32,115,110,107,44,32,115,116,101,112,41,10,32,32,32,32,98,97,115,101,46,97,115,115,101,114,116,40,115,114,99,32,97,110,100,32,115,110,107,41,10,32,32,32,32,115,116,101,112,32,61,32,115,116,101,112,32,111,114,32,112,117,109,112,46,115,116,101,112,10,32,32,32,32,119,104,105,108,101,32,116,114,117,101,32,100,111,10,32,32,32,32,32,32,32,32,108,111,99,97,108,32,114,101,116,44,32,101,114,114,32,61,32,115,116,101,112,40,115,114,99,44,32,115,110,107,41,10,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,114,101,116,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,101,114,114,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,10,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,49,32,101,110,100,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,101,110,100,10,101,110,100,10,10,108,116,110,49,50,46,108,117,97,0,108,116,110,49,50,0,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,77,73,77,69,32,115,117,112,112,111,114,116,32,102,111,114,32,116,104,101,32,76,117,97,32,108,97,110,103,117,97,103,101,46,10,45,45,32,65,117,116,104,111,114,58,32,68,105,101,103,111,32,78,101,104,97,98,10,45,45,32,67,111,110,102,111,114,109,105,110,103,32,116,111,32,82,70,67,115,32,50,48,52,53,45,50,48,52,57,10,45,45,32,82,67,83,32,73,68,58,32,36,73,100,58,32,109,105,109,101,46,108,117,97,44,118,32,49,46,50,57,32,50,48,48,55,47,48,54,47,49,49,32,50,51,58,52,52,58,53,52,32,100,105,101,103,111,32,69,120,112,32,36,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,68,101,99,108,97,114,101,32,109,111,100,117,108,101,32,97,110,100,32,105,109,112,111,114,116,32,100,101,112,101,110,100,101,110,99,105,101,115,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,108,111,99,97,108,32,98,97,115,101,32,61,32,95,71,10,108,111,99,97,108,32,108,116,110,49,50,32,61,32,114,101,113,117,105,114,101,40,34,108,116,110,49,50,34,41,10,108,111,99,97,108,32,109,105,109,101,32,61,32,114,101,113,117,105,114,101,40,34,109,105,109,101,46,99,111,114,101,34,41,10,108,111,99,97,108,32,105,111,32,61,32,114,101,113,117,105,114,101,40,34,105,111,34,41,10,108,111,99,97,108,32,115,116,114,105,110,103,32,61,32,114,101,113,117,105,114,101,40,34,115,116,114,105,110,103,34,41,10,109,111,100,117,108,101,40,34,109,105,109,101,34,41,10,10,45,45,32,101,110,99,111,100,101,44,32,100,101,99,111,100,101,32,97,110,100,32,119,114,97,112,32,97,108,103,111,114,105,116,104,109,32,116,97,98,108,101,115,10,101,110,99,111,100,101,116,32,61,32,123,125,10,100,101,99,111,100,101,116,32,61,32,123,125,10,119,114,97,112,116,32,61,32,123,125,10,10,45,45,32,99,114,101,97,116,101,115,32,97,32,102,117,110,99,116,105,111,110,32,116,104,97,116,32,99,104,111,111,115,101,115,32,97,32,102,105,108,116,101,114,32,98,121,32,110,97,109,101,32,102,114,111,109,32,97,32,103,105,118,101,110,32,116,97,98,108,101,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,99,104,111,111,115,101,40,116,97,98,108,101,41,10,32,32,32,32,114,101,116,117,114,110,32,102,117,110,99,116,105,111,110,40,110,97,109,101,44,32,111,112,116,49,44,32,111,112,116,50,41,10,32,32,32,32,32,32,32,32,105,102,32,98,97,115,101,46,116,121,112,101,40,110,97,109,101,41,32,126,61,32,34,115,116,114,105,110,103,34,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,110,97,109,101,44,32,111,112,116,49,44,32,111,112,116,50,32,61,32,34,100,101,102,97,117,108,116,34,44,32,110,97,109,101,44,32,111,112,116,49,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,108,111,99,97,108,32,102,32,61,32,116,97,98,108,101,91,110,97,109,101,32,111,114,32,34,110,105,108,34,93,10,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,102,32,116,104,101,110,32,10,32,32,32,32,32,32,32,32,32,32,32,32,98,97,115,101,46,101,114,114,111,114,40,34,117,110,107,110,111,119,110,32,107,101,121,32,40,34,32,46,46,32,98,97,115,101,46,116,111,115,116,114,105,110,103,40,110,97,109,101,41,32,46,46,32,34,41,34,44,32,51,41,10,32,32,32,32,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,102,40,111,112,116,49,44,32,111,112,116,50,41,32,101,110,100,10,32,32,32,32,101,110,100,10,101,110,100,10,10,45,45,32,100,101,102,105,110,101,32,116,104,101,32,101,110,99,111,100,105,110,103,32,102,105,108,116,101,114,115,10,101,110,99,111,100,101,116,91,39,98,97,115,101,54,52,39,93,32,61,32,102,117,110,99,116,105,111,110,40,41,10,32,32,32,32,114,101,116,117,114,110,32,108,116,110,49,50,46,102,105,108,116,101,114,46,99,121,99,108,101,40,98,54,52,44,32,34,34,41,10,101,110,100,10,10,101,110,99,111,100,101,116,91,39,113,117,111,116,101,100,45,112,114,105,110,116,97,98,108,101,39,93,32,61,32,102,117,110,99,116,105,111,110,40,109,111,100,101,41,10,32,32,32,32,114,101,116,117,114,110,32,108,116,110,49,50,46,102,105,108,116,101,114,46,99,121,99,108,101,40,113,112,44,32,34,34,44,10,32,32,32,32,32,32,32,32,40,109,111,100,101,32,61,61,32,34,98,105,110,97,114,121,34,41,32,97,110,100,32,34,61,48,68,61,48,65,34,32,111,114,32,34,92,114,92,110,34,41,10,101,110,100,10,10,45,45,32,100,101,102,105,110,101,32,116,104,101,32,100,101,99,111,100,105,110,103,32,102,105,108,116,101,114,115,10,100,101,99,111,100,101,116,91,39,98,97,115,101,54,52,39,93,32,61,32,102,117,110,99,116,105,111,110,40,41,10,32,32,32,32,114,101,116,117,114,110,32,108,116,110,49,50,46,102,105,108,116,101,114,46,99,121,99,108,101,40,117,110,98,54,52,44,32,34,34,41,10,101,110,100,10,10,100,101,99,111,100,101,116,91,39,113,117,111,116,101,100,45,112,114,105,110,116,97,98,108,101,39,93,32,61,32,102,117,110,99,116,105,111,110,40,41,10,32,32,32,32,114,101,116,117,114,110,32,108,116,110,49,50,46,102,105,108,116,101,114,46,99,121,99,108,101,40,117,110,113,112,44,32,34,34,41,10,101,110,100,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,102,111,114,109,97,116,40,99,104,117,110,107,41,10,32,32,32,32,105,102,32,99,104,117,110,107,32,116,104,101,110,10,32,32,32,32,32,32,32,32,105,102,32,99,104,117,110,107,32,61,61,32,34,34,32,116,104,101,110,32,114,101,116,117,114,110,32,34,39,39,34,10,32,32,32,32,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,115,116,114,105,110,103,46,108,101,110,40,99,104,117,110,107,41,32,101,110,100,10,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,34,110,105,108,34,32,101,110,100,10,101,110,100,10,10,45,45,32,100,101,102,105,110,101,32,116,104,101,32,108,105,110,101,45,119,114,97,112,32,102,105,108,116,101,114,115,10,119,114,97,112,116,91,39,116,101,120,116,39,93,32,61,32,102,117,110,99,116,105,111,110,40,108,101,110,103,116,104,41,10,32,32,32,32,108,101,110,103,116,104,32,61,32,108,101,110,103,116,104,32,111,114,32,55,54,10,32,32,32,32,114,101,116,117,114,110,32,108,116,110,49,50,46,102,105,108,116,101,114,46,99,121,99,108,101,40,119,114,112,44,32,108,101,110,103,116,104,44,32,108,101,110,103,116,104,41,10,101,110,100,10,119,114,97,112,116,91,39,98,97,115,101,54,52,39,93,32,61,32,119,114,97,112,116,91,39,116,101,120,116,39,93,10,119,114,97,112,116,91,39,100,101,102,97,117,108,116,39,93,32,61,32,119,114,97,112,116,91,39,116,101,120,116,39,93,10,10,119,114,97,112,116,91,39,113,117,111,116,101,100,45,112,114,105,110,116,97,98,108,101,39,93,32,61,32,102,117,110,99,116,105,111,110,40,41,10,32,32,32,32,114,101,116,117,114,110,32,108,116,110,49,50,46,102,105,108,116,101,114,46,99,121,99,108,101,40,113,112,119,114,112,44,32,55,54,44,32,55,54,41,10,101,110,100,10,10,45,45,32,102,117,110,99,116,105,111,110,32,116,104,97,116,32,99,104,111,111,115,101,32,116,104,101,32,101,110,99,111,100,105,110,103,44,32,100,101,99,111,100,105,110,103,32,111,114,32,119,114,97,112,32,97,108,103,111,114,105,116,104,109,10,101,110,99,111,100,101,32,61,32,99,104,111,111,115,101,40,101,110,99,111,100,101,116,41,10,100,101,99,111,100,101,32,61,32,99,104,111,111,115,101,40,100,101,99,111,100,101,116,41,10,119,114,97,112,32,61,32,99,104,111,111,115,101,40,119,114,97,112,116,41,10,10,45,45,32,100,101,102,105,110,101,32,116,104,101,32,101,110,100,45,111,102,45,108,105,110,101,32,110,111,114,109,97,108,105,122,97,116,105,111,110,32,102,105,108,116,101,114,10,102,117,110,99,116,105,111,110,32,110,111,114,109,97,108,105,122,101,40,109,97,114,107,101,114,41,10,32,32,32,32,114,101,116,117,114,110,32,108,116,110,49,50,46,102,105,108,116,101,114,46,99,121,99,108,101,40,101,111,108,44,32,48,44,32,109,97,114,107,101,114,41,10,101,110,100,10,10,45,45,32,104,105,103,104,32,108,101,118,101,108,32,115,116,117,102,102,105,110,103,32,102,105,108,116,101,114,10,102,117,110,99,116,105,111,110,32,115,116,117,102,102,40,41,10,32,32,32,32,114,101,116,117,114,110,32,108,116,110,49,50,46,102,105,108,116,101,114,46,99,121,99,108,101,40,100,111,116,44,32,50,41,10,101,110,100,10,109,105,109,101,46,108,117,97,0,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,83,77,84,80,32,99,108,105,101,110,116,32,115,117,112,112,111,114,116,32,102,111,114,32,116,104,101,32,76,117,97,32,108,97,110,103,117,97,103,101,46,10,45,45,32,76,117,97,83,111,99,107,101,116,32,116,111,111,108,107,105,116,46,10,45,45,32,65,117,116,104,111,114,58,32,68,105,101,103,111,32,78,101,104,97,98,10,45,45,32,82,67,83,32,73,68,58,32,36,73,100,58,32,115,109,116,112,46,108,117,97,44,118,32,49,46,52,54,32,50,48,48,55,47,48,51,47,49,50,32,48,52,58,48,56,58,52,48,32,100,105,101,103,111,32,69,120,112,32,36,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,68,101,99,108,97,114,101,32,109,111,100,117,108,101,32,97,110,100,32,105,109,112,111,114,116,32,100,101,112,101,110,100,101,110,99,105,101,115,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,108,111,99,97,108,32,98,97,115,101,32,61,32,95,71,10,108,111,99,97,108,32,99,111,114,111,117,116,105,110,101,32,61,32,114,101,113,117,105,114,101,40,34,99,111,114,111,117,116,105,110,101,34,41,10,108,111,99,97,108,32,115,116,114,105,110,103,32,61,32,114,101,113,117,105,114,101,40,34,115,116,114,105,110,103,34,41,10,108,111,99,97,108,32,109,97,116,104,32,61,32,114,101,113,117,105,114,101,40,34,109,97,116,104,34,41,10,108,111,99,97,108,32,111,115,32,61,32,114,101,113,117,105,114,101,40,34,111,115,34,41,10,108,111,99,97,108,32,115,111,99,107,101,116,32,61,32,114,101,113,117,105,114,101,40,34,115,111,99,107,101,116,34,41,10,108,111,99,97,108,32,116,112,32,61,32,114,101,113,117,105,114,101,40,34,115,111,99,107,101,116,46,116,112,34,41,10,108,111,99,97,108,32,108,116,110,49,50,32,61,32,114,101,113,117,105,114,101,40,34,108,116,110,49,50,34,41,10,108,111,99,97,108,32,109,105,109,101,32,61,32,114,101,113,117,105,114,101,40,34,109,105,109,101,34,41,10,109,111,100,117,108,101,40,34,115,111,99,107,101,116,46,115,109,116,112,34,41,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,80,114,111,103,114,97,109,32,99,111,110,115,116,97,110,116,115,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,116,105,109,101,111,117,116,32,102,111,114,32,99,111,110,110,101,99,116,105,111,110,10,84,73,77,69,79,85,84,32,61,32,54,48,10,45,45,32,100,101,102,97,117,108,116,32,115,101,114,118,101,114,32,117,115,101,100,32,116,111,32,115,101,110,100,32,101,45,109,97,105,108,115,10,83,69,82,86,69,82,32,61,32,34,108,111,99,97,108,104,111,115,116,34,10,45,45,32,100,101,102,97,117,108,116,32,112,111,114,116,10,80,79,82,84,32,61,32,50,53,10,45,45,32,100,111,109,97,105,110,32,117,115,101,100,32,105,110,32,72,69,76,79,32,99,111,109,109,97,110,100,32,97,110,100,32,100,101,102,97,117,108,116,32,115,101,110,100,109,97,105,108,10,45,45,32,73,102,32,119,101,32,97,114,101,32,117,110,100,101,114,32,97,32,67,71,73,44,32,116,114,121,32,116,111,32,103,101,116,32,102,114,111,109,32,101,110,118,105,114,111,110,109,101,110,116,10,68,79,77,65,73,78,32,61,32,111,115,46,103,101,116,101,110,118,40,34,83,69,82,86,69,82,95,78,65,77,69,34,41,32,111,114,32,34,108,111,99,97,108,104,111,115,116,34,10,45,45,32,100,101,102,97,117,108,116,32,116,105,109,101,32,122,111,110,101,32,40,109,101,97,110,115,32,119,101,32,100,111,110,39,116,32,107,110,111,119,41,10,90,79,78,69,32,61,32,34,45,48,48,48,48,34,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,76,111,119,32,108,101,118,101,108,32,83,77,84,80,32,65,80,73,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,108,111,99,97,108,32,109,101,116,97,116,32,61,32,123,32,95,95,105,110,100,101,120,32,61,32,123,125,32,125,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,103,114,101,101,116,40,100,111,109,97,105,110,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,50,46,46,34,41,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,34,69,72,76,79,34,44,32,100,111,109,97,105,110,32,111,114,32,68,79,77,65,73,78,41,41,10,32,32,32,32,114,101,116,117,114,110,32,115,111,99,107,101,116,46,115,107,105,112,40,49,44,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,50,46,46,34,41,41,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,109,97,105,108,40,102,114,111,109,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,34,77,65,73,76,34,44,32,34,70,82,79,77,58,34,32,46,46,32,102,114,111,109,41,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,50,46,46,34,41,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,114,99,112,116,40,116,111,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,34,82,67,80,84,34,44,32,34,84,79,58,34,32,46,46,32,116,111,41,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,50,46,46,34,41,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,100,97,116,97,40,115,114,99,44,32,115,116,101,112,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,34,68,65,84,65,34,41,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,51,46,46,34,41,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,115,111,117,114,99,101,40,115,114,99,44,32,115,116,101,112,41,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,115,101,110,100,40,34,92,114,92,110,46,92,114,92,110,34,41,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,50,46,46,34,41,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,113,117,105,116,40,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,34,81,85,73,84,34,41,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,50,46,46,34,41,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,99,108,111,115,101,40,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,116,112,58,99,108,111,115,101,40,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,108,111,103,105,110,40,117,115,101,114,44,32,112,97,115,115,119,111,114,100,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,34,65,85,84,72,34,44,32,34,76,79,71,73,78,34,41,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,51,46,46,34,41,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,109,105,109,101,46,98,54,52,40,117,115,101,114,41,41,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,51,46,46,34,41,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,109,105,109,101,46,98,54,52,40,112,97,115,115,119,111,114,100,41,41,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,50,46,46,34,41,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,112,108,97,105,110,40,117,115,101,114,44,32,112,97,115,115,119,111,114,100,41,10,32,32,32,32,108,111,99,97,108,32,97,117,116,104,32,61,32,34,80,76,65,73,78,32,34,32,46,46,32,109,105,109,101,46,98,54,52,40,34,92,48,34,32,46,46,32,117,115,101,114,32,46,46,32,34,92,48,34,32,46,46,32,112,97,115,115,119,111,114,100,41,10,32,32,32,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,111,109,109,97,110,100,40,34,65,85,84,72,34,44,32,97,117,116,104,41,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,116,114,121,40,115,101,108,102,46,116,112,58,99,104,101,99,107,40,34,50,46,46,34,41,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,97,117,116,104,40,117,115,101,114,44,32,112,97,115,115,119,111,114,100,44,32,101,120,116,41,10,32,32,32,32,105,102,32,110,111,116,32,117,115,101,114,32,111,114,32,110,111,116,32,112,97,115,115,119,111,114,100,32,116,104,101,110,32,114,101,116,117,114,110,32,49,32,101,110,100,10,32,32,32,32,105,102,32,115,116,114,105,110,103,46,102,105,110,100,40,101,120,116,44,32,34,65,85,84,72,91,94,92,110,93,43,76,79,71,73,78,34,41,32,116,104,101,110,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,58,108,111,103,105,110,40,117,115,101,114,44,32,112,97,115,115,119,111,114,100,41,10,32,32,32,32,101,108,115,101,105,102,32,115,116,114,105,110,103,46,102,105,110,100,40,101,120,116,44,32,34,65,85,84,72,91,94,92,110,93,43,80,76,65,73,78,34,41,32,116,104,101,110,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,58,112,108,97,105,110,40,117,115,101,114,44,32,112,97,115,115,119,111,114,100,41,10,32,32,32,32,101,108,115,101,10,32,32,32,32,32,32,32,32,115,101,108,102,46,116,114,121,40,110,105,108,44,32,34,97,117,116,104,101,110,116,105,99,97,116,105,111,110,32,110,111,116,32,115,117,112,112,111,114,116,101,100,34,41,10,32,32,32,32,101,110,100,10,101,110,100,10,10,45,45,32,115,101,110,100,32,109,101,115,115,97,103,101,32,111,114,32,116,104,114,111,119,32,97,110,32,101,120,99,101,112,116,105,111,110,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,115,101,110,100,40,109,97,105,108,116,41,10,32,32,32,32,115,101,108,102,58,109,97,105,108,40,109,97,105,108,116,46,102,114,111,109,41,10,32,32,32,32,105,102,32,98,97,115,101,46,116,121,112,101,40,109,97,105,108,116,46,114,99,112,116,41,32,61,61,32,34,116,97,98,108,101,34,32,116,104,101,110,10,32,32,32,32,32,32,32,32,102,111,114,32,105,44,118,32,105,110,32,98,97,115,101,46,105,112,97,105,114,115,40,109,97,105,108,116,46,114,99,112,116,41,32,100,111,10,32,32,32,32,32,32,32,32,32,32,32,32,115,101,108,102,58,114,99,112,116,40,118,41,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,101,108,115,101,10,32,32,32,32,32,32,32,32,115,101,108,102,58,114,99,112,116,40,109,97,105,108,116,46,114,99,112,116,41,10,32,32,32,32,101,110,100,10,32,32,32,32,115,101,108,102,58,100,97,116,97,40,108,116,110,49,50,46,115,111,117,114,99,101,46,99,104,97,105,110,40,109,97,105,108,116,46,115,111,117,114,99,101,44,32,109,105,109,101,46,115,116,117,102,102,40,41,41,44,32,109,97,105,108,116,46,115,116,101,112,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,111,112,101,110,40,115,101,114,118,101,114,44,32,112,111,114,116,44,32,99,114,101,97,116,101,41,10,32,32,32,32,108,111,99,97,108,32,116,112,32,61,32,115,111,99,107,101,116,46,116,114,121,40,116,112,46,99,111,110,110,101,99,116,40,115,101,114,118,101,114,32,111,114,32,83,69,82,86,69,82,44,32,112,111,114,116,32,111,114,32,80,79,82,84,44,10,32,32,32,32,32,32,32,32,84,73,77,69,79,85,84,44,32,99,114,101,97,116,101,41,41,10,32,32,32,32,108,111,99,97,108,32,115,32,61,32,98,97,115,101,46,115,101,116,109,101,116,97,116,97,98,108,101,40,123,116,112,32,61,32,116,112,125,44,32,109,101,116,97,116,41,10,32,32,32,32,45,45,32,109,97,107,101,32,115,117,114,101,32,116,112,32,105,115,32,99,108,111,115,101,100,32,105,102,32,119,101,32,103,101,116,32,97,110,32,101,120,99,101,112,116,105,111,110,10,32,32,32,32,115,46,116,114,121,32,61,32,115,111,99,107,101,116,46,110,101,119,116,114,121,40,102,117,110,99,116,105,111,110,40,41,10,32,32,32,32,32,32,32,32,115,58,99,108,111,115,101,40,41,10,32,32,32,32,101,110,100,41,10,32,32,32,32,114,101,116,117,114,110,32,115,10,101,110,100,10,10,45,45,32,99,111,110,118,101,114,116,32,104,101,97,100,101,114,115,32,116,111,32,108,111,119,101,114,99,97,115,101,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,108,111,119,101,114,95,104,101,97,100,101,114,115,40,104,101,97,100,101,114,115,41,10,32,32,32,32,108,111,99,97,108,32,108,111,119,101,114,32,61,32,123,125,10,32,32,32,32,102,111,114,32,105,44,118,32,105,110,32,98,97,115,101,46,112,97,105,114,115,40,104,101,97,100,101,114,115,32,111,114,32,108,111,119,101,114,41,32,100,111,10,32,32,32,32,32,32,32,32,108,111,119,101,114,91,115,116,114,105,110,103,46,108,111,119,101,114,40,105,41,93,32,61,32,118,10,32,32,32,32,101,110,100,10,32,32,32,32,114,101,116,117,114,110,32,108,111,119,101,114,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,77,117,108,116,105,112,97,114,116,32,109,101,115,115,97,103,101,32,115,111,117,114,99,101,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,114,101,116,117,114,110,115,32,97,32,104,111,112,101,102,117,108,108,121,32,117,110,105,113,117,101,32,109,105,109,101,32,98,111,117,110,100,97,114,121,10,108,111,99,97,108,32,115,101,113,110,111,32,61,32,48,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,110,101,119,98,111,117,110,100,97,114,121,40,41,10,32,32,32,32,115,101,113,110,111,32,61,32,115,101,113,110,111,32,43,32,49,10,32,32,32,32,114,101,116,117,114,110,32,115,116,114,105,110,103,46,102,111,114,109,97,116,40,39,37,115,37,48,53,100,61,61,37,48,53,117,39,44,32,111,115,46,100,97,116,101,40,39,37,100,37,109,37,89,37,72,37,77,37,83,39,41,44,10,32,32,32,32,32,32,32,32,109,97,116,104,46,114,97,110,100,111,109,40,48,44,32,57,57,57,57,57,41,44,32,115,101,113,110,111,41,10,101,110,100,10,10,45,45,32,115,101,110,100,95,109,101,115,115,97,103,101,32,102,111,114,119,97,114,100,32,100,101,99,108,97,114,97,116,105,111,110,10,108,111,99,97,108,32,115,101,110,100,95,109,101,115,115,97,103,101,10,10,45,45,32,121,105,101,108,100,32,116,104,101,32,104,101,97,100,101,114,115,32,97,108,108,32,97,116,32,111,110,99,101,44,32,105,116,39,115,32,102,97,115,116,101,114,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,115,101,110,100,95,104,101,97,100,101,114,115,40,104,101,97,100,101,114,115,41,10,32,32,32,32,108,111,99,97,108,32,104,32,61,32,34,92,114,92,110,34,10,32,32,32,32,102,111,114,32,105,44,118,32,105,110,32,98,97,115,101,46,112,97,105,114,115,40,104,101,97,100,101,114,115,41,32,100,111,10,32,32,32,32,32,32,32,32,104,32,61,32,105,32,46,46,32,39,58,32,39,32,46,46,32,118,32,46,46,32,34,92,114,92,110,34,32,46,46,32,104,10,32,32,32,32,101,110,100,10,32,32,32,32,99,111,114,111,117,116,105,110,101,46,121,105,101,108,100,40,104,41,10,101,110,100,10,10,45,45,32,121,105,101,108,100,32,109,117,108,116,105,112,97,114,116,32,109,101,115,115,97,103,101,32,98,111,100,121,32,102,114,111,109,32,97,32,109,117,108,116,105,112,97,114,116,32,109,101,115,115,97,103,101,32,116,97,98,108,101,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,115,101,110,100,95,109,117,108,116,105,112,97,114,116,40,109,101,115,103,116,41,10,32,32,32,32,45,45,32,109,97,107,101,32,115,117,114,101,32,119,101,32,104,97,118,101,32,111,117,114,32,98,111,117,110,100,97,114,121,32,97,110,100,32,115,101,110,100,32,104,101,97,100,101,114,115,10,32,32,32,32,108,111,99,97,108,32,98,100,32,61,32,110,101,119,98,111,117,110,100,97,114,121,40,41,10,32,32,32,32,108,111,99,97,108,32,104,101,97,100,101,114,115,32,61,32,108,111,119,101,114,95,104,101,97,100,101,114,115,40,109,101,115,103,116,46,104,101,97,100,101,114,115,32,111,114,32,123,125,41,10,32,32,32,32,104,101,97,100,101,114,115,91,39,99,111,110,116,101,110,116,45,116,121,112,101,39,93,32,61,32,104,101,97,100,101,114,115,91,39,99,111,110,116,101,110,116,45,116,121,112,101,39,93,32,111,114,32,39,109,117,108,116,105,112,97,114,116,47,109,105,120,101,100,39,10,32,32,32,32,104,101,97,100,101,114,115,91,39,99,111,110,116,101,110,116,45,116,121,112,101,39,93,32,61,32,104,101,97,100,101,114,115,91,39,99,111,110,116,101,110,116,45,116,121,112,101,39,93,32,46,46,10,32,32,32,32,32,32,32,32,39,59,32,98,111,117,110,100,97,114,121,61,34,39,32,46,46,32,32,98,100,32,46,46,32,39,34,39,10,32,32,32,32,115,101,110,100,95,104,101,97,100,101,114,115,40,104,101,97,100,101,114,115,41,10,32,32,32,32,45,45,32,115,101,110,100,32,112,114,101,97,109,98,108,101,10,32,32,32,32,105,102,32,109,101,115,103,116,46,98,111,100,121,46,112,114,101,97,109,98,108,101,32,116,104,101,110,10,32,32,32,32,32,32,32,32,99,111,114,111,117,116,105,110,101,46,121,105,101,108,100,40,109,101,115,103,116,46,98,111,100,121,46,112,114,101,97,109,98,108,101,41,10,32,32,32,32,32,32,32,32,99,111,114,111,117,116,105,110,101,46,121,105,101,108,100,40,34,92,114,92,110,34,41,10,32,32,32,32,101,110,100,10,32,32,32,32,45,45,32,115,101,110,100,32,101,97,99,104,32,112,97,114,116,32,115,101,112,97,114,97,116,101,100,32,98,121,32,97,32,98,111,117,110,100,97,114,121,10,32,32,32,32,102,111,114,32,105,44,32,109,32,105,110,32,98,97,115,101,46,105,112,97,105,114,115,40,109,101,115,103,116,46,98,111,100,121,41,32,100,111,10,32,32,32,32,32,32,32,32,99,111,114,111,117,116,105,110,101,46,121,105,101,108,100,40,34,92,114,92,110,45,45,34,32,46,46,32,98,100,32,46,46,32,34,92,114,92,110,34,41,10,32,32,32,32,32,32,32,32,115,101,110,100,95,109,101,115,115,97,103,101,40,109,41,10,32,32,32,32,101,110,100,10,32,32,32,32,45,45,32,115,101,110,100,32,108,97,115,116,32,98,111,117,110,100,97,114,121,10,32,32,32,32,99,111,114,111,117,116,105,110,101,46,121,105,101,108,100,40,34,92,114,92,110,45,45,34,32,46,46,32,98,100,32,46,46,32,34,45,45,92,114,92,110,92,114,92,110,34,41,10,32,32,32,32,45,45,32,115,101,110,100,32,101,112,105,108,111,103,117,101,10,32,32,32,32,105,102,32,109,101,115,103,116,46,98,111,100,121,46,101,112,105,108,111,103,117,101,32,116,104,101,110,10,32,32,32,32,32,32,32,32,99,111,114,111,117,116,105,110,101,46,121,105,101,108,100,40,109,101,115,103,116,46,98,111,100,121,46,101,112,105,108,111,103,117,101,41,10,32,32,32,32,32,32,32,32,99,111,114,111,117,116,105,110,101,46,121,105,101,108,100,40,34,92,114,92,110,34,41,10,32,32,32,32,101,110,100,10,101,110,100,10,10,45,45,32,121,105,101,108,100,32,109,101,115,115,97,103,101,32,98,111,100,121,32,102,114,111,109,32,97,32,115,111,117,114,99,101,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,115,101,110,100,95,115,111,117,114,99,101,40,109,101,115,103,116,41,10,32,32,32,32,45,45,32,109,97,107,101,32,115,117,114,101,32,119,101,32,104,97,118,101,32,97,32,99,111,110,116,101,110,116,45,116,121,112,101,10,32,32,32,32,108,111,99,97,108,32,104,101,97,100,101,114,115,32,61,32,108,111,119,101,114,95,104,101,97,100,101,114,115,40,109,101,115,103,116,46,104,101,97,100,101,114,115,32,111,114,32,123,125,41,10,32,32,32,32,104,101,97,100,101,114,115,91,39,99,111,110,116,101,110,116,45,116,121,112,101,39,93,32,61,32,104,101,97,100,101,114,115,91,39,99,111,110,116,101,110,116,45,116,121,112,101,39,93,32,111,114,10,32,32,32,32,32,32,32,32,39,116,101,120,116,47,112,108,97,105,110,59,32,99,104,97,114,115,101,116,61,34,105,115,111,45,56,56,53,57,45,49,34,39,10,32,32,32,32,115,101,110,100,95,104,101,97,100,101,114,115,40,104,101,97,100,101,114,115,41,10,32,32,32,32,45,45,32,115,101,110,100,32,98,111,100,121,32,102,114,111,109,32,115,111,117,114,99,101,10,32,32,32,32,119,104,105,108,101,32,116,114,117,101,32,100,111,10,32,32,32,32,32,32,32,32,108,111,99,97,108,32,99,104,117,110,107,44,32,101,114,114,32,61,32,109,101,115,103,116,46,98,111,100,121,40,41,10,32,32,32,32,32,32,32,32,105,102,32,101,114,114,32,116,104,101,110,32,99,111,114,111,117,116,105,110,101,46,121,105,101,108,100,40,110,105,108,44,32,101,114,114,41,10,32,32,32,32,32,32,32,32,101,108,115,101,105,102,32,99,104,117,110,107,32,116,104,101,110,32,99,111,114,111,117,116,105,110,101,46,121,105,101,108,100,40,99,104,117,110,107,41,10,32,32,32,32,32,32,32,32,101,108,115,101,32,98,114,101,97,107,32,101,110,100,10,32,32,32,32,101,110,100,10,101,110,100,10,10,45,45,32,121,105,101,108,100,32,109,101,115,115,97,103,101,32,98,111,100,121,32,102,114,111,109,32,97,32,115,116,114,105,110,103,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,115,101,110,100,95,115,116,114,105,110,103,40,109,101,115,103,116,41,10,32,32,32,32,45,45,32,109,97,107,101,32,115,117,114,101,32,119,101,32,104,97,118,101,32,97,32,99,111,110,116,101,110,116,45,116,121,112,101,10,32,32,32,32,108,111,99,97,108,32,104,101,97,100,101,114,115,32,61,32,108,111,119,101,114,95,104,101,97,100,101,114,115,40,109,101,115,103,116,46,104,101,97,100,101,114,115,32,111,114,32,123,125,41,10,32,32,32,32,104,101,97,100,101,114,115,91,39,99,111,110,116,101,110,116,45,116,121,112,101,39,93,32,61,32,104,101,97,100,101,114,115,91,39,99,111,110,116,101,110,116,45,116,121,112,101,39,93,32,111,114,10,32,32,32,32,32,32,32,32,39,116,101,120,116,47,112,108,97,105,110,59,32,99,104,97,114,115,101,116,61,34,105,115,111,45,56,56,53,57,45,49,34,39,10,32,32,32,32,115,101,110,100,95,104,101,97,100,101,114,115,40,104,101,97,100,101,114,115,41,10,32,32,32,32,45,45,32,115,101,110,100,32,98,111,100,121,32,102,114,111,109,32,115,116,114,105,110,103,10,32,32,32,32,99,111,114,111,117,116,105,110,101,46,121,105,101,108,100,40,109,101,115,103,116,46,98,111,100,121,41,10,101,110,100,10,10,45,45,32,109,101,115,115,97,103,101,32,115,111,117,114,99,101,10,102,117,110,99,116,105,111,110,32,115,101,110,100,95,109,101,115,115,97,103,101,40,109,101,115,103,116,41,10,32,32,32,32,105,102,32,98,97,115,101,46,116,121,112,101,40,109,101,115,103,116,46,98,111,100,121,41,32,61,61,32,34,116,97,98,108,101,34,32,116,104,101,110,32,115,101,110,100,95,109,117,108,116,105,112,97,114,116,40,109,101,115,103,116,41,10,32,32,32,32,101,108,115,101,105,102,32,98,97,115,101,46,116,121,112,101,40,109,101,115,103,116,46,98,111,100,121,41,32,61,61,32,34,102,117,110,99,116,105,111,110,34,32,116,104,101,110,32,115,101,110,100,95,115,111,117,114,99,101,40,109,101,115,103,116,41,10,32,32,32,32,101,108,115,101,32,115,101,110,100,95,115,116,114,105,110,103,40,109,101,115,103,116,41,32,101,110,100,10,101,110,100,10,10,45,45,32,115,101,116,32,100,101,102,97,117,108,32,104,101,97,100,101,114,115,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,97,100,106,117,115,116,95,104,101,97,100,101,114,115,40,109,101,115,103,116,41,10,32,32,32,32,108,111,99,97,108,32,108,111,119,101,114,32,61,32,108,111,119,101,114,95,104,101,97,100,101,114,115,40,109,101,115,103,116,46,104,101,97,100,101,114,115,41,10,32,32,32,32,108,111,119,101,114,91,34,100,97,116,101,34,93,32,61,32,108,111,119,101,114,91,34,100,97,116,101,34,93,32,111,114,10,32,32,32,32,32,32,32,32,111,115,46,100,97,116,101,40,34,33,37,97,44,32,37,100,32,37,98,32,37,89,32,37,72,58,37,77,58,37,83,32,34,41,32,46,46,32,40,109,101,115,103,116,46,122,111,110,101,32,111,114,32,90,79,78,69,41,10,32,32,32,32,108,111,119,101,114,91,34,120,45,109,97,105,108,101,114,34,93,32,61,32,108,111,119,101,114,91,34,120,45,109,97,105,108,101,114,34,93,32,111,114,32,115,111,99,107,101,116,46,95,86,69,82,83,73,79,78,10,32,32,32,32,45,45,32,116,104,105,115,32,99,97,110,39,116,32,98,101,32,111,118,101,114,114,105,100,101,110,10,32,32,32,32,108,111,119,101,114,91,34,109,105,109,101,45,118,101,114,115,105,111,110,34,93,32,61,32,34,49,46,48,34,10,32,32,32,32,114,101,116,117,114,110,32,108,111,119,101,114,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,115,115,97,103,101,40,109,101,115,103,116,41,10,32,32,32,32,109,101,115,103,116,46,104,101,97,100,101,114,115,32,61,32,97,100,106,117,115,116,95,104,101,97,100,101,114,115,40,109,101,115,103,116,41,10,32,32,32,32,45,45,32,99,114,101,97,116,101,32,97,110,100,32,114,101,116,117,114,110,32,109,101,115,115,97,103,101,32,115,111,117,114,99,101,10,32,32,32,32,108,111,99,97,108,32,99,111,32,61,32,99,111,114,111,117,116,105,110,101,46,99,114,101,97,116,101,40,102,117,110,99,116,105,111,110,40,41,32,115,101,110,100,95,109,101,115,115,97,103,101,40,109,101,115,103,116,41,32,101,110,100,41,10,32,32,32,32,114,101,116,117,114,110,32,102,117,110,99,116,105,111,110,40,41,10,32,32,32,32,32,32,32,32,108,111,99,97,108,32,114,101,116,44,32,97,44,32,98,32,61,32,99,111,114,111,117,116,105,110,101,46,114,101,115,117,109,101,40,99,111,41,10,32,32,32,32,32,32,32,32,105,102,32,114,101,116,32,116,104,101,110,32,114,101,116,117,114,110,32,97,44,32,98,10,32,32,32,32,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,110,105,108,44,32,97,32,101,110,100,10,32,32,32,32,101,110,100,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,72,105,103,104,32,108,101,118,101,108,32,83,77,84,80,32,65,80,73,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,115,101,110,100,32,61,32,115,111,99,107,101,116,46,112,114,111,116,101,99,116,40,102,117,110,99,116,105,111,110,40,109,97,105,108,116,41,10,32,32,32,32,108,111,99,97,108,32,115,32,61,32,111,112,101,110,40,109,97,105,108,116,46,115,101,114,118,101,114,44,32,109,97,105,108,116,46,112,111,114,116,44,32,109,97,105,108,116,46,99,114,101,97,116,101,41,10,32,32,32,32,108,111,99,97,108,32,101,120,116,32,61,32,115,58,103,114,101,101,116,40,109,97,105,108,116,46,100,111,109,97,105,110,41,10,32,32,32,32,115,58,97,117,116,104,40,109,97,105,108,116,46,117,115,101,114,44,32,109,97,105,108,116,46,112,97,115,115,119,111,114,100,44,32,101,120,116,41,10,32,32,32,32,115,58,115,101,110,100,40,109,97,105,108,116,41,10,32,32,32,32,115,58,113,117,105,116,40,41,10,32,32,32,32,114,101,116,117,114,110,32,115,58,99,108,111,115,101,40,41,10,101,110,100,41,10,115,109,116,112,46,108,117,97,0,115,111,99,107,101,116,46,115,109,116,112,0,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,85,110,105,102,105,101,100,32,83,77,84,80,47,70,84,80,32,115,117,98,115,121,115,116,101,109,10,45,45,32,76,117,97,83,111,99,107,101,116,32,116,111,111,108,107,105,116,46,10,45,45,32,65,117,116,104,111,114,58,32,68,105,101,103,111,32,78,101,104,97,98,10,45,45,32,82,67,83,32,73,68,58,32,36,73,100,58,32,116,112,46,108,117,97,44,118,32,49,46,50,50,32,50,48,48,54,47,48,51,47,49,52,32,48,57,58,48,52,58,49,53,32,100,105,101,103,111,32,69,120,112,32,36,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,68,101,99,108,97,114,101,32,109,111,100,117,108,101,32,97,110,100,32,105,109,112,111,114,116,32,100,101,112,101,110,100,101,110,99,105,101,115,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,108,111,99,97,108,32,98,97,115,101,32,61,32,95,71,10,108,111,99,97,108,32,115,116,114,105,110,103,32,61,32,114,101,113,117,105,114,101,40,34,115,116,114,105,110,103,34,41,10,108,111,99,97,108,32,115,111,99,107,101,116,32,61,32,114,101,113,117,105,114,101,40,34,115,111,99,107,101,116,34,41,10,108,111,99,97,108,32,108,116,110,49,50,32,61,32,114,101,113,117,105,114,101,40,34,108,116,110,49,50,34,41,10,109,111,100,117,108,101,40,34,115,111,99,107,101,116,46,116,112,34,41,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,80,114,111,103,114,97,109,32,99,111,110,115,116,97,110,116,115,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,84,73,77,69,79,85,84,32,61,32,54,48,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,73,109,112,108,101,109,101,110,116,97,116,105,111,110,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,103,101,116,115,32,115,101,114,118,101,114,32,114,101,112,108,121,32,40,119,111,114,107,115,32,102,111,114,32,83,77,84,80,32,97,110,100,32,70,84,80,41,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,103,101,116,95,114,101,112,108,121,40,99,41,10,32,32,32,32,108,111,99,97,108,32,99,111,100,101,44,32,99,117,114,114,101,110,116,44,32,115,101,112,10,32,32,32,32,108,111,99,97,108,32,108,105,110,101,44,32,101,114,114,32,61,32,99,58,114,101,99,101,105,118,101,40,41,10,32,32,32,32,108,111,99,97,108,32,114,101,112,108,121,32,61,32,108,105,110,101,10,32,32,32,32,105,102,32,101,114,114,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,32,101,110,100,10,32,32,32,32,99,111,100,101,44,32,115,101,112,32,61,32,115,111,99,107,101,116,46,115,107,105,112,40,50,44,32,115,116,114,105,110,103,46,102,105,110,100,40,108,105,110,101,44,32,34,94,40,37,100,37,100,37,100,41,40,46,63,41,34,41,41,10,32,32,32,32,105,102,32,110,111,116,32,99,111,100,101,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,34,105,110,118,97,108,105,100,32,115,101,114,118,101,114,32,114,101,112,108,121,34,32,101,110,100,10,32,32,32,32,105,102,32,115,101,112,32,61,61,32,34,45,34,32,116,104,101,110,32,45,45,32,114,101,112,108,121,32,105,115,32,109,117,108,116,105,108,105,110,101,10,32,32,32,32,32,32,32,32,114,101,112,101,97,116,10,32,32,32,32,32,32,32,32,32,32,32,32,108,105,110,101,44,32,101,114,114,32,61,32,99,58,114,101,99,101,105,118,101,40,41,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,101,114,114,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,114,114,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,99,117,114,114,101,110,116,44,32,115,101,112,32,61,32,115,111,99,107,101,116,46,115,107,105,112,40,50,44,32,115,116,114,105,110,103,46,102,105,110,100,40,108,105,110,101,44,32,34,94,40,37,100,37,100,37,100,41,40,46,63,41,34,41,41,10,32,32,32,32,32,32,32,32,32,32,32,32,114,101,112,108,121,32,61,32,114,101,112,108,121,32,46,46,32,34,92,110,34,32,46,46,32,108,105,110,101,10,32,32,32,32,32,32,32,32,45,45,32,114,101,112,108,121,32,101,110,100,115,32,119,105,116,104,32,115,97,109,101,32,99,111,100,101,10,32,32,32,32,32,32,32,32,117,110,116,105,108,32,99,111,100,101,32,61,61,32,99,117,114,114,101,110,116,32,97,110,100,32,115,101,112,32,61,61,32,34,32,34,10,32,32,32,32,101,110,100,10,32,32,32,32,114,101,116,117,114,110,32,99,111,100,101,44,32,114,101,112,108,121,10,101,110,100,10,10,45,45,32,109,101,116,97,116,97,98,108,101,32,102,111,114,32,115,111,99,107,32,111,98,106,101,99,116,10,108,111,99,97,108,32,109,101,116,97,116,32,61,32,123,32,95,95,105,110,100,101,120,32,61,32,123,125,32,125,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,99,104,101,99,107,40,111,107,41,10,32,32,32,32,108,111,99,97,108,32,99,111,100,101,44,32,114,101,112,108,121,32,61,32,103,101,116,95,114,101,112,108,121,40,115,101,108,102,46,99,41,10,32,32,32,32,105,102,32,110,111,116,32,99,111,100,101,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,114,101,112,108,121,32,101,110,100,10,32,32,32,32,105,102,32,98,97,115,101,46,116,121,112,101,40,111,107,41,32,126,61,32,34,102,117,110,99,116,105,111,110,34,32,116,104,101,110,10,32,32,32,32,32,32,32,32,105,102,32,98,97,115,101,46,116,121,112,101,40,111,107,41,32,61,61,32,34,116,97,98,108,101,34,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,102,111,114,32,105,44,32,118,32,105,110,32,98,97,115,101,46,105,112,97,105,114,115,40,111,107,41,32,100,111,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,115,116,114,105,110,103,46,102,105,110,100,40,99,111,100,101,44,32,118,41,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,98,97,115,101,46,116,111,110,117,109,98,101,114,40,99,111,100,101,41,44,32,114,101,112,108,121,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,110,105,108,44,32,114,101,112,108,121,10,32,32,32,32,32,32,32,32,101,108,115,101,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,115,116,114,105,110,103,46,102,105,110,100,40,99,111,100,101,44,32,111,107,41,32,116,104,101,110,32,114,101,116,117,114,110,32,98,97,115,101,46,116,111,110,117,109,98,101,114,40,99,111,100,101,41,44,32,114,101,112,108,121,10,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,110,105,108,44,32,114,101,112,108,121,32,101,110,100,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,101,108,115,101,32,114,101,116,117,114,110,32,111,107,40,98,97,115,101,46,116,111,110,117,109,98,101,114,40,99,111,100,101,41,44,32,114,101,112,108,121,41,32,101,110,100,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,99,111,109,109,97,110,100,40,99,109,100,44,32,97,114,103,41,10,32,32,32,32,105,102,32,97,114,103,32,116,104,101,110,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,99,58,115,101,110,100,40,99,109,100,32,46,46,32,34,32,34,32,46,46,32,97,114,103,46,46,32,34,92,114,92,110,34,41,10,32,32,32,32,101,108,115,101,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,99,58,115,101,110,100,40,99,109,100,32,46,46,32,34,92,114,92,110,34,41,10,32,32,32,32,101,110,100,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,115,105,110,107,40,115,110,107,44,32,112,97,116,41,10,32,32,32,32,108,111,99,97,108,32,99,104,117,110,107,44,32,101,114,114,32,61,32,99,58,114,101,99,101,105,118,101,40,112,97,116,41,10,32,32,32,32,114,101,116,117,114,110,32,115,110,107,40,99,104,117,110,107,44,32,101,114,114,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,115,101,110,100,40,100,97,116,97,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,99,58,115,101,110,100,40,100,97,116,97,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,114,101,99,101,105,118,101,40,112,97,116,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,99,58,114,101,99,101,105,118,101,40,112,97,116,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,103,101,116,102,100,40,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,99,58,103,101,116,102,100,40,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,100,105,114,116,121,40,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,99,58,100,105,114,116,121,40,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,103,101,116,99,111,110,116,114,111,108,40,41,10,32,32,32,32,114,101,116,117,114,110,32,115,101,108,102,46,99,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,115,111,117,114,99,101,40,115,111,117,114,99,101,44,32,115,116,101,112,41,10,32,32,32,32,108,111,99,97,108,32,115,105,110,107,32,61,32,115,111,99,107,101,116,46,115,105,110,107,40,34,107,101,101,112,45,111,112,101,110,34,44,32,115,101,108,102,46,99,41,10,32,32,32,32,108,111,99,97,108,32,114,101,116,44,32,101,114,114,32,61,32,108,116,110,49,50,46,112,117,109,112,46,97,108,108,40,115,111,117,114,99,101,44,32,115,105,110,107,44,32,115,116,101,112,32,111,114,32,108,116,110,49,50,46,112,117,109,112,46,115,116,101,112,41,10,32,32,32,32,114,101,116,117,114,110,32,114,101,116,44,32,101,114,114,10,101,110,100,10,10,45,45,32,99,108,111,115,101,115,32,116,104,101,32,117,110,100,101,114,108,121,105,110,103,32,99,10,102,117,110,99,116,105,111,110,32,109,101,116,97,116,46,95,95,105,110,100,101,120,58,99,108,111,115,101,40,41,10,32,32,32,32,115,101,108,102,46,99,58,99,108,111,115,101,40,41,10,9,114,101,116,117,114,110,32,49,10,101,110,100,10,10,45,45,32,99,111,110,110,101,99,116,32,119,105,116,104,32,115,101,114,118,101,114,32,97,110,100,32,114,101,116,117,114,110,32,99,32,111,98,106,101,99,116,10,102,117,110,99,116,105,111,110,32,99,111,110,110,101,99,116,40,104,111,115,116,44,32,112,111,114,116,44,32,116,105,109,101,111,117,116,44,32,99,114,101,97,116,101,41,10,32,32,32,32,108,111,99,97,108,32,99,44,32,101,32,61,32,40,99,114,101,97,116,101,32,111,114,32,115,111,99,107,101,116,46,116,99,112,41,40,41,10,32,32,32,32,105,102,32,110,111,116,32,99,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,101,32,101,110,100,10,32,32,32,32,99,58,115,101,116,116,105,109,101,111,117,116,40,116,105,109,101,111,117,116,32,111,114,32,84,73,77,69,79,85,84,41,10,32,32,32,32,108,111,99,97,108,32,114,44,32,101,32,61,32,99,58,99,111,110,110,101,99,116,40,104,111,115,116,44,32,112,111,114,116,41,10,32,32,32,32,105,102,32,110,111,116,32,114,32,116,104,101,110,10,32,32,32,32,32,32,32,32,99,58,99,108,111,115,101,40,41,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,110,105,108,44,32,101,10,32,32,32,32,101,110,100,10,32,32,32,32,114,101,116,117,114,110,32,98,97,115,101,46,115,101,116,109,101,116,97,116,97,98,108,101,40,123,99,32,61,32,99,125,44,32,109,101,116,97,116,41,10,101,110,100,10,10,116,112,46,108,117,97,0,115,111,99,107,101,116,46,116,112,0,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,85,82,73,32,112,97,114,115,105,110,103,44,32,99,111,109,112,111,115,105,116,105,111,110,32,97,110,100,32,114,101,108,97,116,105,118,101,32,85,82,76,32,114,101,115,111,108,117,116,105,111,110,10,45,45,32,76,117,97,83,111,99,107,101,116,32,116,111,111,108,107,105,116,46,10,45,45,32,65,117,116,104,111,114,58,32,68,105,101,103,111,32,78,101,104,97,98,10,45,45,32,82,67,83,32,73,68,58,32,36,73,100,58,32,117,114,108,46,108,117,97,44,118,32,49,46,51,56,32,50,48,48,54,47,48,52,47,48,51,32,48,52,58,52,53,58,52,50,32,100,105,101,103,111,32,69,120,112,32,36,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,68,101,99,108,97,114,101,32,109,111,100,117,108,101,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,108,111,99,97,108,32,115,116,114,105,110,103,32,61,32,114,101,113,117,105,114,101,40,34,115,116,114,105,110,103,34,41,10,108,111,99,97,108,32,98,97,115,101,32,61,32,95,71,10,108,111,99,97,108,32,116,97,98,108,101,32,61,32,114,101,113,117,105,114,101,40,34,116,97,98,108,101,34,41,10,109,111,100,117,108,101,40,34,115,111,99,107,101,116,46,117,114,108,34,41,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,77,111,100,117,108,101,32,118,101,114,115,105,111,110,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,95,86,69,82,83,73,79,78,32,61,32,34,85,82,76,32,49,46,48,46,49,34,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,69,110,99,111,100,101,115,32,97,32,115,116,114,105,110,103,32,105,110,116,111,32,105,116,115,32,101,115,99,97,112,101,100,32,104,101,120,97,100,101,99,105,109,97,108,32,114,101,112,114,101,115,101,110,116,97,116,105,111,110,10,45,45,32,73,110,112,117,116,10,45,45,32,32,32,115,58,32,98,105,110,97,114,121,32,115,116,114,105,110,103,32,116,111,32,98,101,32,101,110,99,111,100,101,100,10,45,45,32,82,101,116,117,114,110,115,10,45,45,32,32,32,101,115,99,97,112,101,100,32,114,101,112,114,101,115,101,110,116,97,116,105,111,110,32,111,102,32,115,116,114,105,110,103,32,98,105,110,97,114,121,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,102,117,110,99,116,105,111,110,32,101,115,99,97,112,101,40,115,41,10,32,32,32,32,114,101,116,117,114,110,32,115,116,114,105,110,103,46,103,115,117,98,40,115,44,32,34,40,91,94,65,45,90,97,45,122,48,45,57,95,93,41,34,44,32,102,117,110,99,116,105,111,110,40,99,41,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,115,116,114,105,110,103,46,102,111,114,109,97,116,40,34,37,37,37,48,50,120,34,44,32,115,116,114,105,110,103,46,98,121,116,101,40,99,41,41,10,32,32,32,32,101,110,100,41,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,80,114,111,116,101,99,116,115,32,97,32,112,97,116,104,32,115,101,103,109,101,110,116,44,32,116,111,32,112,114,101,118,101,110,116,32,105,116,32,102,114,111,109,32,105,110,116,101,114,102,101,114,105,110,103,32,119,105,116,104,32,116,104,101,10,45,45,32,117,114,108,32,112,97,114,115,105,110,103,46,10,45,45,32,73,110,112,117,116,10,45,45,32,32,32,115,58,32,98,105,110,97,114,121,32,115,116,114,105,110,103,32,116,111,32,98,101,32,101,110,99,111,100,101,100,10,45,45,32,82,101,116,117,114,110,115,10,45,45,32,32,32,101,115,99,97,112,101,100,32,114,101,112,114,101,115,101,110,116,97,116,105,111,110,32,111,102,32,115,116,114,105,110,103,32,98,105,110,97,114,121,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,109,97,107,101,95,115,101,116,40,116,41,10,9,108,111,99,97,108,32,115,32,61,32,123,125,10,9,102,111,114,32,105,44,118,32,105,110,32,98,97,115,101,46,105,112,97,105,114,115,40,116,41,32,100,111,10,9,9,115,91,116,91,105,93,93,32,61,32,49,10,9,101,110,100,10,9,114,101,116,117,114,110,32,115,10,101,110,100,10,10,45,45,32,116,104,101,115,101,32,97,114,101,32,97,108,108,111,119,101,100,32,119,105,116,104,105,110,103,32,97,32,112,97,116,104,32,115,101,103,109,101,110,116,44,32,97,108,111,110,103,32,119,105,116,104,32,97,108,112,104,97,110,117,109,10,45,45,32,111,116,104,101,114,32,99,104,97,114,97,99,116,101,114,115,32,109,117,115,116,32,98,101,32,101,115,99,97,112,101,100,10,108,111,99,97,108,32,115,101,103,109,101,110,116,95,115,101,116,32,61,32,109,97,107,101,95,115,101,116,32,123,10,32,32,32,32,34,45,34,44,32,34,95,34,44,32,34,46,34,44,32,34,33,34,44,32,34,126,34,44,32,34,42,34,44,32,34,39,34,44,32,34,40,34,44,10,9,34,41,34,44,32,34,58,34,44,32,34,64,34,44,32,34,38,34,44,32,34,61,34,44,32,34,43,34,44,32,34,36,34,44,32,34,44,34,44,10,125,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,112,114,111,116,101,99,116,95,115,101,103,109,101,110,116,40,115,41,10,9,114,101,116,117,114,110,32,115,116,114,105,110,103,46,103,115,117,98,40,115,44,32,34,40,91,94,65,45,90,97,45,122,48,45,57,95,93,41,34,44,32,102,117,110,99,116,105,111,110,32,40,99,41,10,9,9,105,102,32,115,101,103,109,101,110,116,95,115,101,116,91,99,93,32,116,104,101,110,32,114,101,116,117,114,110,32,99,10,9,9,101,108,115,101,32,114,101,116,117,114,110,32,115,116,114,105,110,103,46,102,111,114,109,97,116,40,34,37,37,37,48,50,120,34,44,32,115,116,114,105,110,103,46,98,121,116,101,40,99,41,41,32,101,110,100,10,9,101,110,100,41,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,69,110,99,111,100,101,115,32,97,32,115,116,114,105,110,103,32,105,110,116,111,32,105,116,115,32,101,115,99,97,112,101,100,32,104,101,120,97,100,101,99,105,109,97,108,32,114,101,112,114,101,115,101,110,116,97,116,105,111,110,10,45,45,32,73,110,112,117,116,10,45,45,32,32,32,115,58,32,98,105,110,97,114,121,32,115,116,114,105,110,103,32,116,111,32,98,101,32,101,110,99,111,100,101,100,10,45,45,32,82,101,116,117,114,110,115,10,45,45,32,32,32,101,115,99,97,112,101,100,32,114,101,112,114,101,115,101,110,116,97,116,105,111,110,32,111,102,32,115,116,114,105,110,103,32,98,105,110,97,114,121,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,102,117,110,99,116,105,111,110,32,117,110,101,115,99,97,112,101,40,115,41,10,32,32,32,32,114,101,116,117,114,110,32,115,116,114,105,110,103,46,103,115,117,98,40,115,44,32,34,37,37,40,37,120,37,120,41,34,44,32,102,117,110,99,116,105,111,110,40,104,101,120,41,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,115,116,114,105,110,103,46,99,104,97,114,40,98,97,115,101,46,116,111,110,117,109,98,101,114,40,104,101,120,44,32,49,54,41,41,10,32,32,32,32,101,110,100,41,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,66,117,105,108,100,115,32,97,32,112,97,116,104,32,102,114,111,109,32,97,32,98,97,115,101,32,112,97,116,104,32,97,110,100,32,97,32,114,101,108,97,116,105,118,101,32,112,97,116,104,10,45,45,32,73,110,112,117,116,10,45,45,32,32,32,98,97,115,101,95,112,97,116,104,10,45,45,32,32,32,114,101,108,97,116,105,118,101,95,112,97,116,104,10,45,45,32,82,101,116,117,114,110,115,10,45,45,32,32,32,99,111,114,114,101,115,112,111,110,100,105,110,103,32,97,98,115,111,108,117,116,101,32,112,97,116,104,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,97,98,115,111,108,117,116,101,95,112,97,116,104,40,98,97,115,101,95,112,97,116,104,44,32,114,101,108,97,116,105,118,101,95,112,97,116,104,41,10,32,32,32,32,105,102,32,115,116,114,105,110,103,46,115,117,98,40,114,101,108,97,116,105,118,101,95,112,97,116,104,44,32,49,44,32,49,41,32,61,61,32,34,47,34,32,116,104,101,110,32,114,101,116,117,114,110,32,114,101,108,97,116,105,118,101,95,112,97,116,104,32,101,110,100,10,32,32,32,32,108,111,99,97,108,32,112,97,116,104,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,98,97,115,101,95,112,97,116,104,44,32,34,91,94,47,93,42,36,34,44,32,34,34,41,10,32,32,32,32,112,97,116,104,32,61,32,112,97,116,104,32,46,46,32,114,101,108,97,116,105,118,101,95,112,97,116,104,10,32,32,32,32,112,97,116,104,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,112,97,116,104,44,32,34,40,91,94,47,93,42,37,46,47,41,34,44,32,102,117,110,99,116,105,111,110,32,40,115,41,10,32,32,32,32,32,32,32,32,105,102,32,115,32,126,61,32,34,46,47,34,32,116,104,101,110,32,114,101,116,117,114,110,32,115,32,101,108,115,101,32,114,101,116,117,114,110,32,34,34,32,101,110,100,10,32,32,32,32,101,110,100,41,10,32,32,32,32,112,97,116,104,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,112,97,116,104,44,32,34,47,37,46,36,34,44,32,34,47,34,41,10,32,32,32,32,108,111,99,97,108,32,114,101,100,117,99,101,100,10,32,32,32,32,119,104,105,108,101,32,114,101,100,117,99,101,100,32,126,61,32,112,97,116,104,32,100,111,10,32,32,32,32,32,32,32,32,114,101,100,117,99,101,100,32,61,32,112,97,116,104,10,32,32,32,32,32,32,32,32,112,97,116,104,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,114,101,100,117,99,101,100,44,32,34,40,91,94,47,93,42,47,37,46,37,46,47,41,34,44,32,102,117,110,99,116,105,111,110,32,40,115,41,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,115,32,126,61,32,34,46,46,47,46,46,47,34,32,116,104,101,110,32,114,101,116,117,114,110,32,34,34,32,101,108,115,101,32,114,101,116,117,114,110,32,115,32,101,110,100,10,32,32,32,32,32,32,32,32,101,110,100,41,10,32,32,32,32,101,110,100,10,32,32,32,32,112,97,116,104,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,114,101,100,117,99,101,100,44,32,34,40,91,94,47,93,42,47,37,46,37,46,41,36,34,44,32,102,117,110,99,116,105,111,110,32,40,115,41,10,32,32,32,32,32,32,32,32,105,102,32,115,32,126,61,32,34,46,46,47,46,46,34,32,116,104,101,110,32,114,101,116,117,114,110,32,34,34,32,101,108,115,101,32,114,101,116,117,114,110,32,115,32,101,110,100,10,32,32,32,32,101,110,100,41,10,32,32,32,32,114,101,116,117,114,110,32,112,97,116,104,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,80,97,114,115,101,115,32,97,32,117,114,108,32,97,110,100,32,114,101,116,117,114,110,115,32,97,32,116,97,98,108,101,32,119,105,116,104,32,97,108,108,32,105,116,115,32,112,97,114,116,115,32,97,99,99,111,114,100,105,110,103,32,116,111,32,82,70,67,32,50,51,57,54,10,45,45,32,84,104,101,32,102,111,108,108,111,119,105,110,103,32,103,114,97,109,109,97,114,32,100,101,115,99,114,105,98,101,115,32,116,104,101,32,110,97,109,101,115,32,103,105,118,101,110,32,116,111,32,116,104,101,32,85,82,76,32,112,97,114,116,115,10,45,45,32,60,117,114,108,62,32,58,58,61,32,60,115,99,104,101,109,101,62,58,47,47,60,97,117,116,104,111,114,105,116,121,62,47,60,112,97,116,104,62,59,60,112,97,114,97,109,115,62,63,60,113,117,101,114,121,62,35,60,102,114,97,103,109,101,110,116,62,10,45,45,32,60,97,117,116,104,111,114,105,116,121,62,32,58,58,61,32,60,117,115,101,114,105,110,102,111,62,64,60,104,111,115,116,62,58,60,112,111,114,116,62,10,45,45,32,60,117,115,101,114,105,110,102,111,62,32,58,58,61,32,60,117,115,101,114,62,91,58,60,112,97,115,115,119,111,114,100,62,93,10,45,45,32,60,112,97,116,104,62,32,58,58,32,61,32,123,60,115,101,103,109,101,110,116,62,47,125,60,115,101,103,109,101,110,116,62,10,45,45,32,73,110,112,117,116,10,45,45,32,32,32,117,114,108,58,32,117,110,105,102,111,114,109,32,114,101,115,111,117,114,99,101,32,108,111,99,97,116,111,114,32,111,102,32,114,101,113,117,101,115,116,10,45,45,32,32,32,100,101,102,97,117,108,116,58,32,116,97,98,108,101,32,119,105,116,104,32,100,101,102,97,117,108,116,32,118,97,108,117,101,115,32,102,111,114,32,101,97,99,104,32,102,105,101,108,100,10,45,45,32,82,101,116,117,114,110,115,10,45,45,32,32,32,116,97,98,108,101,32,119,105,116,104,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,102,105,101,108,100,115,44,32,119,104,101,114,101,32,82,70,67,32,110,97,109,105,110,103,32,99,111,110,118,101,110,116,105,111,110,115,32,104,97,118,101,10,45,45,32,32,32,98,101,101,110,32,112,114,101,115,101,114,118,101,100,58,10,45,45,32,32,32,32,32,115,99,104,101,109,101,44,32,97,117,116,104,111,114,105,116,121,44,32,117,115,101,114,105,110,102,111,44,32,117,115,101,114,44,32,112,97,115,115,119,111,114,100,44,32,104,111,115,116,44,32,112,111,114,116,44,10,45,45,32,32,32,32,32,112,97,116,104,44,32,112,97,114,97,109,115,44,32,113,117,101,114,121,44,32,102,114,97,103,109,101,110,116,10,45,45,32,79,98,115,58,10,45,45,32,32,32,116,104,101,32,108,101,97,100,105,110,103,32,39,47,39,32,105,110,32,123,47,60,112,97,116,104,62,125,32,105,115,32,99,111,110,115,105,100,101,114,101,100,32,112,97,114,116,32,111,102,32,60,112,97,116,104,62,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,102,117,110,99,116,105,111,110,32,112,97,114,115,101,40,117,114,108,44,32,100,101,102,97,117,108,116,41,10,32,32,32,32,45,45,32,105,110,105,116,105,97,108,105,122,101,32,100,101,102,97,117,108,116,32,112,97,114,97,109,101,116,101,114,115,10,32,32,32,32,108,111,99,97,108,32,112,97,114,115,101,100,32,61,32,123,125,10,32,32,32,32,102,111,114,32,105,44,118,32,105,110,32,98,97,115,101,46,112,97,105,114,115,40,100,101,102,97,117,108,116,32,111,114,32,112,97,114,115,101,100,41,32,100,111,32,112,97,114,115,101,100,91,105,93,32,61,32,118,32,101,110,100,10,32,32,32,32,45,45,32,101,109,112,116,121,32,117,114,108,32,105,115,32,112,97,114,115,101,100,32,116,111,32,110,105,108,10,32,32,32,32,105,102,32,110,111,116,32,117,114,108,32,111,114,32,117,114,108,32,61,61,32,34,34,32,116,104,101,110,32,114,101,116,117,114,110,32,110,105,108,44,32,34,105,110,118,97,108,105,100,32,117,114,108,34,32,101,110,100,10,32,32,32,32,45,45,32,114,101,109,111,118,101,32,119,104,105,116,101,115,112,97,99,101,10,32,32,32,32,45,45,32,117,114,108,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,117,114,108,44,32,34,37,115,34,44,32,34,34,41,10,32,32,32,32,45,45,32,103,101,116,32,102,114,97,103,109,101,110,116,10,32,32,32,32,117,114,108,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,117,114,108,44,32,34,35,40,46,42,41,36,34,44,32,102,117,110,99,116,105,111,110,40,102,41,10,32,32,32,32,32,32,32,32,112,97,114,115,101,100,46,102,114,97,103,109,101,110,116,32,61,32,102,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,34,34,10,32,32,32,32,101,110,100,41,10,32,32,32,32,45,45,32,103,101,116,32,115,99,104,101,109,101,10,32,32,32,32,117,114,108,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,117,114,108,44,32,34,94,40,91,37,119,93,91,37,119,37,43,37,45,37,46,93,42,41,37,58,34,44,10,32,32,32,32,32,32,32,32,102,117,110,99,116,105,111,110,40,115,41,32,112,97,114,115,101,100,46,115,99,104,101,109,101,32,61,32,115,59,32,114,101,116,117,114,110,32,34,34,32,101,110,100,41,10,32,32,32,32,45,45,32,103,101,116,32,97,117,116,104,111,114,105,116,121,10,32,32,32,32,117,114,108,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,117,114,108,44,32,34,94,47,47,40,91,94,47,93,42,41,34,44,32,102,117,110,99,116,105,111,110,40,110,41,10,32,32,32,32,32,32,32,32,112,97,114,115,101,100,46,97,117,116,104,111,114,105,116,121,32,61,32,110,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,34,34,10,32,32,32,32,101,110,100,41,10,32,32,32,32,45,45,32,103,101,116,32,113,117,101,114,121,32,115,116,114,105,110,103,105,110,103,10,32,32,32,32,117,114,108,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,117,114,108,44,32,34,37,63,40,46,42,41,34,44,32,102,117,110,99,116,105,111,110,40,113,41,10,32,32,32,32,32,32,32,32,112,97,114,115,101,100,46,113,117,101,114,121,32,61,32,113,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,34,34,10,32,32,32,32,101,110,100,41,10,32,32,32,32,45,45,32,103,101,116,32,112,97,114,97,109,115,10,32,32,32,32,117,114,108,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,117,114,108,44,32,34,37,59,40,46,42,41,34,44,32,102,117,110,99,116,105,111,110,40,112,41,10,32,32,32,32,32,32,32,32,112,97,114,115,101,100,46,112,97,114,97,109,115,32,61,32,112,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,34,34,10,32,32,32,32,101,110,100,41,10,32,32,32,32,45,45,32,112,97,116,104,32,105,115,32,119,104,97,116,101,118,101,114,32,119,97,115,32,108,101,102,116,10,32,32,32,32,105,102,32,117,114,108,32,126,61,32,34,34,32,116,104,101,110,32,112,97,114,115,101,100,46,112,97,116,104,32,61,32,117,114,108,32,101,110,100,10,32,32,32,32,108,111,99,97,108,32,97,117,116,104,111,114,105,116,121,32,61,32,112,97,114,115,101,100,46,97,117,116,104,111,114,105,116,121,10,32,32,32,32,105,102,32,110,111,116,32,97,117,116,104,111,114,105,116,121,32,116,104,101,110,32,114,101,116,117,114,110,32,112,97,114,115,101,100,32,101,110,100,10,32,32,32,32,97,117,116,104,111,114,105,116,121,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,97,117,116,104,111,114,105,116,121,44,34,94,40,91,94,64,93,42,41,64,34,44,10,32,32,32,32,32,32,32,32,102,117,110,99,116,105,111,110,40,117,41,32,112,97,114,115,101,100,46,117,115,101,114,105,110,102,111,32,61,32,117,59,32,114,101,116,117,114,110,32,34,34,32,101,110,100,41,10,32,32,32,32,97,117,116,104,111,114,105,116,121,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,97,117,116,104,111,114,105,116,121,44,32,34,58,40,91,94,58,93,42,41,36,34,44,10,32,32,32,32,32,32,32,32,102,117,110,99,116,105,111,110,40,112,41,32,112,97,114,115,101,100,46,112,111,114,116,32,61,32,112,59,32,114,101,116,117,114,110,32,34,34,32,101,110,100,41,10,32,32,32,32,105,102,32,97,117,116,104,111,114,105,116,121,32,126,61,32,34,34,32,116,104,101,110,32,112,97,114,115,101,100,46,104,111,115,116,32,61,32,97,117,116,104,111,114,105,116,121,32,101,110,100,10,32,32,32,32,108,111,99,97,108,32,117,115,101,114,105,110,102,111,32,61,32,112,97,114,115,101,100,46,117,115,101,114,105,110,102,111,10,32,32,32,32,105,102,32,110,111,116,32,117,115,101,114,105,110,102,111,32,116,104,101,110,32,114,101,116,117,114,110,32,112,97,114,115,101,100,32,101,110,100,10,32,32,32,32,117,115,101,114,105,110,102,111,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,117,115,101,114,105,110,102,111,44,32,34,58,40,91,94,58,93,42,41,36,34,44,10,32,32,32,32,32,32,32,32,102,117,110,99,116,105,111,110,40,112,41,32,112,97,114,115,101,100,46,112,97,115,115,119,111,114,100,32,61,32,112,59,32,114,101,116,117,114,110,32,34,34,32,101,110,100,41,10,32,32,32,32,112,97,114,115,101,100,46,117,115,101,114,32,61,32,117,115,101,114,105,110,102,111,10,32,32,32,32,114,101,116,117,114,110,32,112,97,114,115,101,100,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,82,101,98,117,105,108,100,115,32,97,32,112,97,114,115,101,100,32,85,82,76,32,102,114,111,109,32,105,116,115,32,99,111,109,112,111,110,101,110,116,115,46,10,45,45,32,67,111,109,112,111,110,101,110,116,115,32,97,114,101,32,112,114,111,116,101,99,116,101,100,32,105,102,32,97,110,121,32,114,101,115,101,114,118,101,100,32,111,114,32,117,110,97,108,108,111,119,101,100,32,99,104,97,114,97,99,116,101,114,115,32,97,114,101,32,102,111,117,110,100,10,45,45,32,73,110,112,117,116,10,45,45,32,32,32,112,97,114,115,101,100,58,32,112,97,114,115,101,100,32,85,82,76,44,32,97,115,32,114,101,116,117,114,110,101,100,32,98,121,32,112,97,114,115,101,10,45,45,32,82,101,116,117,114,110,115,10,45,45,32,32,32,97,32,115,116,114,105,110,103,105,110,103,32,119,105,116,104,32,116,104,101,32,99,111,114,114,101,115,112,111,110,100,105,110,103,32,85,82,76,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,102,117,110,99,116,105,111,110,32,98,117,105,108,100,40,112,97,114,115,101,100,41,10,32,32,32,32,108,111,99,97,108,32,112,112,97,116,104,32,61,32,112,97,114,115,101,95,112,97,116,104,40,112,97,114,115,101,100,46,112,97,116,104,32,111,114,32,34,34,41,10,32,32,32,32,108,111,99,97,108,32,117,114,108,32,61,32,98,117,105,108,100,95,112,97,116,104,40,112,112,97,116,104,41,10,32,32,32,32,105,102,32,112,97,114,115,101,100,46,112,97,114,97,109,115,32,116,104,101,110,32,117,114,108,32,61,32,117,114,108,32,46,46,32,34,59,34,32,46,46,32,112,97,114,115,101,100,46,112,97,114,97,109,115,32,101,110,100,10,32,32,32,32,105,102,32,112,97,114,115,101,100,46,113,117,101,114,121,32,116,104,101,110,32,117,114,108,32,61,32,117,114,108,32,46,46,32,34,63,34,32,46,46,32,112,97,114,115,101,100,46,113,117,101,114,121,32,101,110,100,10,9,108,111,99,97,108,32,97,117,116,104,111,114,105,116,121,32,61,32,112,97,114,115,101,100,46,97,117,116,104,111,114,105,116,121,10,9,105,102,32,112,97,114,115,101,100,46,104,111,115,116,32,116,104,101,110,10,9,9,97,117,116,104,111,114,105,116,121,32,61,32,112,97,114,115,101,100,46,104,111,115,116,10,9,9,105,102,32,112,97,114,115,101,100,46,112,111,114,116,32,116,104,101,110,32,97,117,116,104,111,114,105,116,121,32,61,32,97,117,116,104,111,114,105,116,121,32,46,46,32,34,58,34,32,46,46,32,112,97,114,115,101,100,46,112,111,114,116,32,101,110,100,10,9,9,108,111,99,97,108,32,117,115,101,114,105,110,102,111,32,61,32,112,97,114,115,101,100,46,117,115,101,114,105,110,102,111,10,9,9,105,102,32,112,97,114,115,101,100,46,117,115,101,114,32,116,104,101,110,10,9,9,9,117,115,101,114,105,110,102,111,32,61,32,112,97,114,115,101,100,46,117,115,101,114,10,9,9,9,105,102,32,112,97,114,115,101,100,46,112,97,115,115,119,111,114,100,32,116,104,101,110,10,9,9,9,9,117,115,101,114,105,110,102,111,32,61,32,117,115,101,114,105,110,102,111,32,46,46,32,34,58,34,32,46,46,32,112,97,114,115,101,100,46,112,97,115,115,119,111,114,100,10,9,9,9,101,110,100,10,9,9,101,110,100,10,9,9,105,102,32,117,115,101,114,105,110,102,111,32,116,104,101,110,32,97,117,116,104,111,114,105,116,121,32,61,32,117,115,101,114,105,110,102,111,32,46,46,32,34,64,34,32,46,46,32,97,117,116,104,111,114,105,116,121,32,101,110,100,10,9,101,110,100,10,32,32,32,32,105,102,32,97,117,116,104,111,114,105,116,121,32,116,104,101,110,32,117,114,108,32,61,32,34,47,47,34,32,46,46,32,97,117,116,104,111,114,105,116,121,32,46,46,32,117,114,108,32,101,110,100,10,32,32,32,32,105,102,32,112,97,114,115,101,100,46,115,99,104,101,109,101,32,116,104,101,110,32,117,114,108,32,61,32,112,97,114,115,101,100,46,115,99,104,101,109,101,32,46,46,32,34,58,34,32,46,46,32,117,114,108,32,101,110,100,10,32,32,32,32,105,102,32,112,97,114,115,101,100,46,102,114,97,103,109,101,110,116,32,116,104,101,110,32,117,114,108,32,61,32,117,114,108,32,46,46,32,34,35,34,32,46,46,32,112,97,114,115,101,100,46,102,114,97,103,109,101,110,116,32,101,110,100,10,32,32,32,32,45,45,32,117,114,108,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,117,114,108,44,32,34,37,115,34,44,32,34,34,41,10,32,32,32,32,114,101,116,117,114,110,32,117,114,108,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,66,117,105,108,100,115,32,97,32,97,98,115,111,108,117,116,101,32,85,82,76,32,102,114,111,109,32,97,32,98,97,115,101,32,97,110,100,32,97,32,114,101,108,97,116,105,118,101,32,85,82,76,32,97,99,99,111,114,100,105,110,103,32,116,111,32,82,70,67,32,50,51,57,54,10,45,45,32,73,110,112,117,116,10,45,45,32,32,32,98,97,115,101,95,117,114,108,10,45,45,32,32,32,114,101,108,97,116,105,118,101,95,117,114,108,10,45,45,32,82,101,116,117,114,110,115,10,45,45,32,32,32,99,111,114,114,101,115,112,111,110,100,105,110,103,32,97,98,115,111,108,117,116,101,32,117,114,108,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,102,117,110,99,116,105,111,110,32,97,98,115,111,108,117,116,101,40,98,97,115,101,95,117,114,108,44,32,114,101,108,97,116,105,118,101,95,117,114,108,41,10,32,32,32,32,105,102,32,98,97,115,101,46,116,121,112,101,40,98,97,115,101,95,117,114,108,41,32,61,61,32,34,116,97,98,108,101,34,32,116,104,101,110,10,32,32,32,32,32,32,32,32,98,97,115,101,95,112,97,114,115,101,100,32,61,32,98,97,115,101,95,117,114,108,10,32,32,32,32,32,32,32,32,98,97,115,101,95,117,114,108,32,61,32,98,117,105,108,100,40,98,97,115,101,95,112,97,114,115,101,100,41,10,32,32,32,32,101,108,115,101,10,32,32,32,32,32,32,32,32,98,97,115,101,95,112,97,114,115,101,100,32,61,32,112,97,114,115,101,40,98,97,115,101,95,117,114,108,41,10,32,32,32,32,101,110,100,10,32,32,32,32,108,111,99,97,108,32,114,101,108,97,116,105,118,101,95,112,97,114,115,101,100,32,61,32,112,97,114,115,101,40,114,101,108,97,116,105,118,101,95,117,114,108,41,10,32,32,32,32,105,102,32,110,111,116,32,98,97,115,101,95,112,97,114,115,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,114,101,108,97,116,105,118,101,95,117,114,108,10,32,32,32,32,101,108,115,101,105,102,32,110,111,116,32,114,101,108,97,116,105,118,101,95,112,97,114,115,101,100,32,116,104,101,110,32,114,101,116,117,114,110,32,98,97,115,101,95,117,114,108,10,32,32,32,32,101,108,115,101,105,102,32,114,101,108,97,116,105,118,101,95,112,97,114,115,101,100,46,115,99,104,101,109,101,32,116,104,101,110,32,114,101,116,117,114,110,32,114,101,108,97,116,105,118,101,95,117,114,108,10,32,32,32,32,101,108,115,101,10,32,32,32,32,32,32,32,32,114,101,108,97,116,105,118,101,95,112,97,114,115,101,100,46,115,99,104,101,109,101,32,61,32,98,97,115,101,95,112,97,114,115,101,100,46,115,99,104,101,109,101,10,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,114,101,108,97,116,105,118,101,95,112,97,114,115,101,100,46,97,117,116,104,111,114,105,116,121,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,114,101,108,97,116,105,118,101,95,112,97,114,115,101,100,46,97,117,116,104,111,114,105,116,121,32,61,32,98,97,115,101,95,112,97,114,115,101,100,46,97,117,116,104,111,114,105,116,121,10,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,114,101,108,97,116,105,118,101,95,112,97,114,115,101,100,46,112,97,116,104,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,114,101,108,97,116,105,118,101,95,112,97,114,115,101,100,46,112,97,116,104,32,61,32,98,97,115,101,95,112,97,114,115,101,100,46,112,97,116,104,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,114,101,108,97,116,105,118,101,95,112,97,114,115,101,100,46,112,97,114,97,109,115,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,114,101,108,97,116,105,118,101,95,112,97,114,115,101,100,46,112,97,114,97,109,115,32,61,32,98,97,115,101,95,112,97,114,115,101,100,46,112,97,114,97,109,115,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,105,102,32,110,111,116,32,114,101,108,97,116,105,118,101,95,112,97,114,115,101,100,46,113,117,101,114,121,32,116,104,101,110,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,114,101,108,97,116,105,118,101,95,112,97,114,115,101,100,46,113,117,101,114,121,32,61,32,98,97,115,101,95,112,97,114,115,101,100,46,113,117,101,114,121,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,32,32,32,32,101,108,115,101,32,32,32,32,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,114,101,108,97,116,105,118,101,95,112,97,114,115,101,100,46,112,97,116,104,32,61,32,97,98,115,111,108,117,116,101,95,112,97,116,104,40,98,97,115,101,95,112,97,114,115,101,100,46,112,97,116,104,32,111,114,32,34,34,44,10,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,114,101,108,97,116,105,118,101,95,112,97,114,115,101,100,46,112,97,116,104,41,10,32,32,32,32,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,101,110,100,10,32,32,32,32,32,32,32,32,114,101,116,117,114,110,32,98,117,105,108,100,40,114,101,108,97,116,105,118,101,95,112,97,114,115,101,100,41,10,32,32,32,32,101,110,100,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,66,114,101,97,107,115,32,97,32,112,97,116,104,32,105,110,116,111,32,105,116,115,32,115,101,103,109,101,110,116,115,44,32,117,110,101,115,99,97,112,105,110,103,32,116,104,101,32,115,101,103,109,101,110,116,115,10,45,45,32,73,110,112,117,116,10,45,45,32,32,32,112,97,116,104,10,45,45,32,82,101,116,117,114,110,115,10,45,45,32,32,32,115,101,103,109,101,110,116,58,32,97,32,116,97,98,108,101,32,119,105,116,104,32,111,110,101,32,101,110,116,114,121,32,112,101,114,32,115,101,103,109,101,110,116,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,102,117,110,99,116,105,111,110,32,112,97,114,115,101,95,112,97,116,104,40,112,97,116,104,41,10,9,108,111,99,97,108,32,112,97,114,115,101,100,32,61,32,123,125,10,9,112,97,116,104,32,61,32,112,97,116,104,32,111,114,32,34,34,10,9,45,45,112,97,116,104,32,61,32,115,116,114,105,110,103,46,103,115,117,98,40,112,97,116,104,44,32,34,37,115,34,44,32,34,34,41,10,9,115,116,114,105,110,103,46,103,115,117,98,40,112,97,116,104,44,32,34,40,91,94,47,93,43,41,34,44,32,102,117,110,99,116,105,111,110,32,40,115,41,32,116,97,98,108,101,46,105,110,115,101,114,116,40,112,97,114,115,101,100,44,32,115,41,32,101,110,100,41,10,9,102,111,114,32,105,32,61,32,49,44,32,116,97,98,108,101,46,103,101,116,110,40,112,97,114,115,101,100,41,32,100,111,10,9,9,112,97,114,115,101,100,91,105,93,32,61,32,117,110,101,115,99,97,112,101,40,112,97,114,115,101,100,91,105,93,41,10,9,101,110,100,10,9,105,102,32,115,116,114,105,110,103,46,115,117,98,40,112,97,116,104,44,32,49,44,32,49,41,32,61,61,32,34,47,34,32,116,104,101,110,32,112,97,114,115,101,100,46,105,115,95,97,98,115,111,108,117,116,101,32,61,32,49,32,101,110,100,10,9,105,102,32,115,116,114,105,110,103,46,115,117,98,40,112,97,116,104,44,32,45,49,44,32,45,49,41,32,61,61,32,34,47,34,32,116,104,101,110,32,112,97,114,115,101,100,46,105,115,95,100,105,114,101,99,116,111,114,121,32,61,32,49,32,101,110,100,10,9,114,101,116,117,114,110,32,112,97,114,115,101,100,10,101,110,100,10,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,45,45,32,66,117,105,108,100,115,32,97,32,112,97,116,104,32,99,111,109,112,111,110,101,110,116,32,102,114,111,109,32,105,116,115,32,115,101,103,109,101,110,116,115,44,32,101,115,99,97,112,105,110,103,32,112,114,111,116,101,99,116,101,100,32,99,104,97,114,97,99,116,101,114,115,46,10,45,45,32,73,110,112,117,116,10,45,45,32,32,32,112,97,114,115,101,100,58,32,112,97,116,104,32,115,101,103,109,101,110,116,115,10,45,45,32,32,32,117,110,115,97,102,101,58,32,105,102,32,116,114,117,101,44,32,115,101,103,109,101,110,116,115,32,97,114,101,32,110,111,116,32,112,114,111,116,101,99,116,101,100,32,98,101,102,111,114,101,32,112,97,116,104,32,105,115,32,98,117,105,108,116,10,45,45,32,82,101,116,117,114,110,115,10,45,45,32,32,32,112,97,116,104,58,32,99,111,114,114,101,115,112,111,110,100,105,110,103,32,112,97,116,104,32,115,116,114,105,110,103,105,110,103,10,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,10,102,117,110,99,116,105,111,110,32,98,117,105,108,100,95,112,97,116,104,40,112,97,114,115,101,100,44,32,117,110,115,97,102,101,41,10,9,108,111,99,97,108,32,112,97,116,104,32,61,32,34,34,10,9,108,111,99,97,108,32,110,32,61,32,116,97,98,108,101,46,103,101,116,110,40,112,97,114,115,101,100,41,10,9,105,102,32,117,110,115,97,102,101,32,116,104,101,110,10,9,9,102,111,114,32,105,32,61,32,49,44,32,110,45,49,32,100,111,10,9,9,9,112,97,116,104,32,61,32,112,97,116,104,32,46,46,32,112,97,114,115,101,100,91,105,93,10,9,9,9,112,97,116,104,32,61,32,112,97,116,104,32,46,46,32,34,47,34,10,9,9,101,110,100,10,9,9,105,102,32,110,32,62,32,48,32,116,104,101,110,10,9,9,9,112,97,116,104,32,61,32,112,97,116,104,32,46,46,32,112,97,114,115,101,100,91,110,93,10,9,9,9,105,102,32,112,97,114,115,101,100,46,105,115,95,100,105,114,101,99,116,111,114,121,32,116,104,101,110,32,112,97,116,104,32,61,32,112,97,116,104,32,46,46,32,34,47,34,32,101,110,100,10,9,9,101,110,100,10,9,101,108,115,101,10,9,9,102,111,114,32,105,32,61,32,49,44,32,110,45,49,32,100,111,10,9,9,9,112,97,116,104,32,61,32,112,97,116,104,32,46,46,32,112,114,111,116,101,99,116,95,115,101,103,109,101,110,116,40,112,97,114,115,101,100,91,105,93,41,10,9,9,9,112,97,116,104,32,61,32,112,97,116,104,32,46,46,32,34,47,34,10,9,9,101,110,100,10,9,9,105,102,32,110,32,62,32,48,32,116,104,101,110,10,9,9,9,112,97,116,104,32,61,32,112,97,116,104,32,46,46,32,112,114,111,116,101,99,116,95,115,101,103,109,101,110,116,40,112,97,114,115,101,100,91,110,93,41,10,9,9,9,105,102,32,112,97,114,115,101,100,46,105,115,95,100,105,114,101,99,116,111,114,121,32,116,104,101,110,32,112,97,116,104,32,61,32,112,97,116,104,32,46,46,32,34,47,34,32,101,110,100,10,9,9,101,110,100,10,9,101,110,100,10,9,105,102,32,112,97,114,115,101,100,46,105,115,95,97,98,115,111,108,117,116,101,32,116,104,101,110,32,112,97,116,104,32,61,32,34,47,34,32,46,46,32,112,97,116,104,32,101,110,100,10,9,114,101,116,117,114,110,32,112,97,116,104,10,101,110,100,10,117,114,108,46,108,117,97,0,115,111,99,107,101,116,46,117,114,108,0,111,102,102,115,101,116,0,112,111,115,105,116,105,111,110,32,111,117,116,32,111,102,32,114,97,110,103,101,0,105,110,105,116,105,97,108,32,112,111,115,105,116,105,111,110,32,105,115,32,97,32,99,111,110,116,105,110,117,97,116,105,111,110,32,98,121,116,101,0,99,111,100,101,112,111,105,110,116,0,111,117,116,32,111,102,32,114,97,110,103,101,0,105,110,118,97,108,105,100,32,85,84,70,45,56,32,99,111,100,101,0,118,97,108,117,101,32,111,117,116,32,111,102,32,114,97,110,103,101,0,105,110,105,116,105,97,108,32,112,111,115,105,116,105,111,110,32,111,117,116,32,111,102,32,115,116,114,105,110,103,0,102,105,110,97,108,32,112,111,115,105,116,105,111,110,32,111,117,116,32,111,102,32,115,116,114,105,110,103,0,99,111,100,101,115,0,99,104,97,114,112,97,116,116,101,114,110,0,91,37,122,1,45,127,194,45,244,93,91,128,45,191,93,42,0,151,160,137,91,90,15,131,13,201,95,96,53,194,233,7,225,140,36,103,30,69,142,8,99,37,240,21,10,23,190,6,148,247,120,234,75,0,26,197,62,94,252,219,203,117,35,11,32,57,177,33,88,237,149,56,87,174,20,125,136,171,168,68,175,74,165,71,134,139,48,27,166,77,146,158,231,83,111,229,122,60,211,133,230,220,105,92,41,55,46,245,40,244,102,143,54,65,25,63,161,1,216,80,73,209,76,132,187,208,89,18,169,200,196,135,130,116,188,159,86,164,100,109,198,173,186,3,64,52,217,226,250,124,123,5,202,38,147,118,126,255,82,85,212,207,206,59,227,47,16,58,17,182,189,28,42,223,183,170,213,119,248,152,2,44,154,163,70,221,153,101,155,167,43,172,9,129,22,39,253,19,98,108,110,79,113,224,232,178,185,112,104,218,246,97,228,251,34,242,193,238,210,144,12,191,179,162,241,81,51,145,235,249,14,239,107,49,192,214,31,181,199,106,157,184,84,204,176,115,121,50,45,127,4,150,254,138,236,205,93,222,114,67,29,24,72,243,141,128,195,78,66,215,61,156,180,151,160,137,91,90,15,131,13,201,95,96,53,194,233,7,225,140,36,103,30,69,142,8,99,37,240,21,10,23,190,6,148,247,120,234,75,0,26,197,62,94,252,219,203,117,35,11,32,57,177,33,88,237,149,56,87,174,20,125,136,171,168,68,175,74,165,71,134,139,48,27,166,77,146,158,231,83,111,229,122,60,211,133,230,220,105,92,41,55,46,245,40,244,102,143,54,65,25,63,161,1,216,80,73,209,76,132,187,208,89,18,169,200,196,135,130,116,188,159,86,164,100,109,198,173,186,3,64,52,217,226,250,124,123,5,202,38,147,118,126,255,82,85,212,207,206,59,227,47,16,58,17,182,189,28,42,223,183,170,213,119,248,152,2,44,154,163,70,221,153,101,155,167,43,172,9,129,22,39,253,19,98,108,110,79,113,224,232,178,185,112,104,218,246,97,228,251,34,242,193,238,210,144,12,191,179,162,241,81,51,145,235,249,14,239,107,49,192,214,31,181,199,106,157,184,84,204,176,115,121,50,45,127,4,150,254,138,236,205,93,222,114,67,29,24,72,243,141,128,195,78,66,215,61,156,180,151,160,137,91,90,15,131,13,201,95,96,53,194,233,7,225,140,36,103,30,69,142,8,99,37,240,21,10,23,190,6,148,247,120,234,75,0,26,197,62,94,252,219,203,117,35,11,32,57,177,33,88,237,149,56,87,174,20,125,136,171,168,68,175,74,165,71,134,139,48,27,166,77,146,158,231,83,111,229,122,60,211,133,230,220,105,92,41,55,46,245,40,244,102,143,54,65,25,63,161,1,216,80,73,209,76,132,187,208,89,18,169,200,196,135,130,116,188,159,86,164,100,109,198,173,186,3,64,52,217,226,250,124,123,5,202,38,147,118,126,255,82,85,212,207,206,59,227,47,16,58,17,182,189,28,42,223,183,170,213,119,248,152,2,44,154,163,70,221,153,101,155,167,43,172,9,129,22,39,253,19,98,108,110,79,113,224,232,178,185,112,104,218,246,97,228,251,34,242,193,238,210,144,12,191,179,162,241,81,51,145,235,249,14,239,107,49,192,214,31,181,199,106,157,184,84,204,176,115,121,50,45,127,4,150,254,138,236,205,93,222,114,67,29,24,72,243,141,128,195,78,66,215,61,156,180,151,160,137,91,90,15,131,13,201,95,96,53,194,233,7,225,140,36,103,30,69,142,8,99,37,240,21,10,23,190,6,148,247,120,234,75,0,26,197,62,94,252,219,203,117,35,11,32,57,177,33,88,237,149,56,87,174,20,125,136,171,168,68,175,74,165,71,134,139,48,27,166,77,146,158,231,83,111,229,122,60,211,133,230,220,105,92,41,55,46,245,40,244,102,143,54,65,25,63,161,1,216,80,73,209,76,132,187,208,89,18,169,200,196,135,130,116,188,159,86,164,100,109,198,173,186,3,64,52,217,226,250,124,123,5,202,38,147,118,126,255,82,85,212,207,206,59,227,47,16,58,17,182,189,28,42,223,183,170,213,119,248,152,2,44,154,163,70,221,153,101,155,167,43,172,9,129,22,39,253,19,98,108,110,79,113,224,232,178,185,112,104,218,246,97,228,251,34,242,193,238,210,144,12,191,179,162,241,81,51,145,235,249,14,239,107,49,192,214,31,181,199,106,157,184,84,204,176,115,121,50,45,127,4,150,254,138,236,205,93,222,114,67,29,24,72,243,141,128,195,78,66,215,61,156,180,78,52,108,111,118,101,57,69,120,99,101,112,116,105,111,110,69,0,73,110,99,111,114,114,101,99,116,32,110,117,109,98,101,114,32,111,102,32,97,114,103,117,109,101,110,116,115,46,32,71,111,116,32,91,37,100,93,44,32,101,120,112,101,99,116,101,100,32,97,116,32,108,101,97,115,116,32,91,37,100,93,0,73,110,99,111,114,114,101,99,116,32,110,117,109,98,101,114,32,111,102,32,97,114,103,117,109,101,110,116,115,46,32,71,111,116,32,91,37,100,93,44,32,101,120,112,101,99,116,101,100,32,91,37,100,45,37,100,93,0,95,109,111,100,117,108,101,115,0,95,108,111,118,101,111,98,106,101,99,116,115,0,65,116,116,101,109,112,116,101,100,32,116,111,32,117,115,101,32,105,110,118,97,108,105,100,32,114,101,103,105,115,116,114,121,46,0,116,121,112,101,79,102,0,67,97,110,39,116,32,114,101,103,105,115,116,101,114,32,115,101,97,114,99,104,101,114,58,32,112,97,99,107,97,103,101,32,116,97,98,108,101,32,100,111,101,115,32,110,111,116,32,101,120,105,115,116,46,0,115,101,97,114,99,104,101,114,115,0,67,97,110,39,116,32,114,101,103,105,115,116,101,114,32,115,101,97,114,99,104,101,114,58,32,112,97,99,107,97,103,101,46,108,111,97,100,101,114,115,32,116,97,98,108,101,32,100,111,101,115,32,110,111,116,32,101,120,105,115,116,46,0,67,111,117,108,100,32,110,111,116,32,102,105,110,100,32,103,108,111,98,97,108,32,108,111,118,101,33,0,67,111,117,108,100,32,110,111,116,32,102,105,110,100,32,108,111,118,101,46,37,115,33,0,67,111,117,108,100,32,110,111,116,32,102,105,110,100,32,108,111,118,101,46,37,115,46,37,115,33,0,95,108,111,118,101,95,109,97,105,110,116,104,114,101,97,100,0,103,101,116,83,116,114,105,110,103,0,103,101,116,80,111,105,110,116,101,114,0,68,97,116,97,0,115,111,117,110,100,0,110,101,119,68,101,99,111,100,101,114,0,73,110,118,97,108,105,100,32,115,111,117,114,99,101,32,116,121,112,101,58,32,37,115,0,110,101,119,83,111,117,110,100,68,97,116,97,0,68,101,99,111,100,101,114,32,111,114,32,83,111,117,110,100,68,97,116,97,0,73,110,118,97,108,105,100,32,100,105,115,116,97,110,99,101,32,109,111,100,101,108,58,32,37,115,0,67,111,117,108,100,32,110,111,116,32,111,112,101,110,32,97,110,121,32,97,117,100,105,111,32,109,111,100,117,108,101,46,0,97,117,100,105,111,0,103,101,116,83,111,117,114,99,101,67,111,117,110,116,0,110,101,119,83,111,117,114,99,101,0,115,101,116,79,114,105,101,110,116,97,116,105,111,110,0,103,101,116,79,114,105,101,110,116,97,116,105,111,110,0,115,101,116,68,111,112,112,108,101,114,83,99,97,108,101,0,103,101,116,68,111,112,112,108,101,114,83,99,97,108,101,0,115,101,116,68,105,115,116,97,110,99,101,77,111,100,101,108,0,103,101,116,68,105,115,116,97,110,99,101,77,111,100,101,108,0,80,105,116,99,104,32,104,97,115,32,116,111,32,98,101,32,102,105,110,105,116,101,32,97,110,100,32,110,111,116,32,78,97,78,46,0,99,97,110,39,116,32,115,101,101,107,32,116,111,32,97,32,110,101,103,97,116,105,118,101,32,112,111,115,105,116,105,111,110,0,73,110,118,97,108,105,100,32,83,111,117,114,99,101,32,116,105,109,101,32,117,110,105,116,58,32,37,115,0,73,110,118,97,108,105,100,32,118,111,108,117,109,101,32,108,105,109,105,116,115,58,32,91,37,102,58,37,102,93,46,32,77,117,115,116,32,98,101,32,105,110,32,91,48,58,49,93,0,73,110,118,97,108,105,100,32,100,105,115,116,97,110,99,101,115,58,32,37,102,44,32,37,102,46,32,77,117,115,116,32,98,101,32,62,32,48,0,73,110,118,97,108,105,100,32,114,111,108,108,111,102,102,58,32,37,102,46,32,77,117,115,116,32,98,101,32,62,32,48,46,0,85,110,107,110,111,119,110,32,83,111,117,114,99,101,32,116,121,112,101,46,0,83,111,117,114,99,101,0,112,108,97,121,0,114,101,119,105,110,100,0,115,101,116,80,105,116,99,104,0,103,101,116,80,105,116,99,104,0,115,101,116,86,111,108,117,109,101,0,103,101,116,86,111,108,117,109,101,0,115,101,116,86,101,108,111,99,105,116,121,0,103,101,116,86,101,108,111,99,105,116,121,0,115,101,116,67,111,110,101,0,103,101,116,67,111,110,101,0,115,101,116,82,101,108,97,116,105,118,101,0,105,115,82,101,108,97,116,105,118,101,0,115,101,116,76,111,111,112,105,110,103,0,105,115,76,111,111,112,105,110,103,0,105,115,80,108,97,121,105,110,103,0,115,101,116,86,111,108,117,109,101,76,105,109,105,116,115,0,103,101,116,86,111,108,117,109,101,76,105,109,105,116,115,0,115,101,116,65,116,116,101,110,117,97,116,105,111,110,68,105,115,116,97,110,99,101,115,0,103,101,116,65,116,116,101,110,117,97,116,105,111,110,68,105,115,116,97,110,99,101,115,0,115,101,116,82,111,108,108,111,102,102,0,103,101,116,82,111,108,108,111,102,102,0,101,118,101,110,116,0,112,117,109,112,0,112,111,108,108,0,78,52,108,111,118,101,49,48,102,105,108,101,115,121,115,116,101,109,56,70,105,108,101,68,97,116,97,69,0,78,52,108,111,118,101,52,68,97,116,97,69,0,98,97,115,101,54,52,0,67,111,117,108,100,32,110,111,116,32,115,101,116,32,115,111,117,114,99,101,46,0,102,105,108,101,110,97,109,101,44,32,70,105,108,101,44,32,111,114,32,70,105,108,101,68,97,116,97,32,101,120,112,101,99,116,101,100,0,110,101,119,70,105,108,101,0,102,105,108,101,110,97,109,101,32,111,114,32,70,105,108,101,32,101,120,112,101,99,116,101,100,0,73,110,118,97,108,105,100,32,70,105,108,101,68,97,116,97,32,100,101,99,111,100,101,114,58,32,37,115,0,70,105,108,101,32,99,111,117,108,100,32,110,111,116,32,98,101,32,114,101,97,100,46,0,115,116,114,105,110,103,32,111,114,32,68,97,116,97,32,101,120,112,101,99,116,101,100,0,101,120,112,101,99,116,101,100,32,102,105,108,101,110,97,109,101,46,0,77,101,109,111,114,121,32,97,108,108,111,99,97,116,105,111,110,32,101,114,114,111,114,58,32,37,115,10,0,83,121,110,116,97,120,32,101,114,114,111,114,58,32,37,115,10,0,83,105,122,101,32,116,111,111,32,108,97,114,103,101,32,116,111,32,102,105,116,32,105,110,116,111,32,97,32,76,117,97,32,110,117,109,98,101,114,33,0,10,9,110,111,32,39,37,115,39,32,105,110,32,76,79,86,69,32,103,97,109,101,32,100,105,114,101,99,116,111,114,105,101,115,46,0,115,101,116,70,117,115,101,100,0,105,115,70,117,115,101,100,0,95,115,101,116,65,110,100,114,111,105,100,83,97,118,101,69,120,116,101,114,110,97,108,0,115,101,116,73,100,101,110,116,105,116,121,0,103,101,116,73,100,101,110,116,105,116,121,0,115,101,116,83,111,117,114,99,101,0,109,111,117,110,116,0,117,110,109,111,117,110,116,0,103,101,116,87,111,114,107,105,110,103,68,105,114,101,99,116,111,114,121,0,103,101,116,85,115,101,114,68,105,114,101,99,116,111,114,121,0,103,101,116,65,112,112,100,97,116,97,68,105,114,101,99,116,111,114,121,0,103,101,116,83,97,118,101,68,105,114,101,99,116,111,114,121,0,103,101,116,83,111,117,114,99,101,66,97,115,101,68,105,114,101,99,116,111,114,121,0,103,101,116,82,101,97,108,68,105,114,101,99,116,111,114,121,0,103,101,116,69,120,101,99,117,116,97,98,108,101,80,97,116,104,0,101,120,105,115,116,115,0,105,115,68,105,114,101,99,116,111,114,121,0,105,115,83,121,109,108,105,110,107,0,99,114,101,97,116,101,68,105,114,101,99,116,111,114,121,0,97,112,112,101,110,100,0,103,101,116,68,105,114,101,99,116,111,114,121,73,116,101,109,115,0,103,101,116,76,97,115,116,77,111,100,105,102,105,101,100,0,115,101,116,83,121,109,108,105,110,107,115,69,110,97,98,108,101,100,0,97,114,101,83,121,109,108,105,110,107,115,69,110,97,98,108,101,100,0,110,101,119,70,105,108,101,68,97,116,97,0,103,101,116,82,101,113,117,105,114,101,80,97,116,104,0,115,101,116,82,101,113,117,105,114,101,80,97,116,104,0,73,110,118,97,108,105,100,32,84,114,117,101,84,121,112,101,32,102,111,110,116,32,104,105,110,116,105,110,103,32,109,111,100,101,58,32,37,115,0,110,101,119,84,114,117,101,84,121,112,101,82,97,115,116,101,114,105,122,101,114,0,110,101,119,66,77,70,111,110,116,82,97,115,116,101,114,105,122,101,114,0,110,101,119,71,108,121,112,104,68,97,116,97,0,117,110,107,110,111,119,110,32,71,108,121,112,104,68,97,116,97,32,102,111,114,109,97,116,46,0,71,108,121,112,104,68,97,116,97,0,103,101,116,71,108,121,112,104,0,103,101,116,71,108,121,112,104,83,116,114,105,110,103,0,103,101,116,66,101,97,114,105,110,103,0,82,97,115,116,101,114,105,122,101,114,0,103,101,116,65,100,118,97,110,99,101,0,103,101,116,71,108,121,112,104,68,97,116,97,0,103,101,116,71,108,121,112,104,67,111,117,110,116,0,67,97,110,39,116,32,115,101,116,32,115,99,105,115,115,111,114,32,119,105,116,104,32,110,101,103,97,116,105,118,101,32,119,105,100,116,104,32,97,110,100,47,111,114,32,104,101,105,103,104,116,46,0,73,110,118,97,108,105,100,32,115,116,101,110,99,105,108,32,100,114,97,119,32,97,99,116,105,111,110,58,32,37,115,0,73,110,118,97,108,105,100,32,99,111,109,112,97,114,101,32,109,111,100,101,58,32,37,115,0,85,110,107,110,111,119,110,32,99,111,109,112,97,114,101,32,109,111,100,101,46,0,108,111,118,101,46,103,114,97,112,104,105,99,115,32,99,97,110,110,111,116,32,102,117,110,99,116,105,111,110,32,119,105,116,104,111,117,116,32,97,32,119,105,110,100,111,119,33,0,67,97,110,110,111,116,32,108,111,97,100,32,105,109,97,103,101,115,32,119,105,116,104,111,117,116,32,116,104,101,32,108,111,118,101,46,105,109,97,103,101,32,109,111,100,117,108,101,46,0,67,111,117,108,100,32,110,111,116,32,108,111,97,100,32,105,109,97,103,101,46,0,102,111,110,116,0,110,101,119,82,97,115,116,101,114,105,122,101,114,0,73,109,97,103,101,32,109,117,115,116,32,110,111,116,32,98,101,32,99,111,109,112,114,101,115,115,101,100,46,0,110,101,119,73,109,97,103,101,82,97,115,116,101,114,105,122,101,114,0,73,110,118,97,108,105,100,32,83,112,114,105,116,101,66,97,116,99,104,32,117,115,97,103,101,32,104,105,110,116,58,32,37,115,0,73,110,118,97,108,105,100,32,80,97,114,116,105,99,108,101,83,121,115,116,101,109,32,115,105,122,101,0,73,110,118,97,108,105,100,32,67,97,110,118,97,115,32,102,111,114,109,97,116,58,32,37,115,0,67,97,110,118,97,115,32,110,111,116,32,99,114,101,97,116,101,100,44,32,98,117,116,32,110,111,32,101,114,114,111,114,32,116,104,114,111,119,110,46,32,73,32,100,111,110,39,116,32,101,118,101,110,46,46,46,0,103,114,97,112,104,105,99,115,0,95,115,104,97,100,101,114,67,111,100,101,84,111,71,76,83,76,0,67,111,117,108,100,32,110,111,116,32,112,97,114,115,101,32,118,101,114,116,101,120,32,115,104,97,100,101,114,32,99,111,100,101,32,40,109,105,115,115,105,110,103,32,39,112,111,115,105,116,105,111,110,39,32,102,117,110,99,116,105,111,110,63,41,0,67,111,117,108,100,32,110,111,116,32,112,97,114,115,101,32,112,105,120,101,108,32,115,104,97,100,101,114,32,99,111,100,101,32,40,109,105,115,115,105,110,103,32,39,101,102,102,101,99,116,39,32,102,117,110,99,116,105,111,110,63,41,0,109,105,115,115,105,110,103,32,39,112,111,115,105,116,105,111,110,39,32,111,114,32,39,101,102,102,101,99,116,39,32,102,117,110,99,116,105,111,110,63,0,95,116,114,97,110,115,102,111,114,109,71,76,83,76,69,114,114,111,114,77,101,115,115,97,103,101,115,0,116,97,98,108,101,32,111,114,32,110,117,109,98,101,114,32,101,120,112,101,99,116,101,100,0,73,110,118,97,108,105,100,32,109,101,115,104,32,117,115,97,103,101,32,104,105,110,116,58,32,37,115,0,116,97,98,108,101,32,111,102,32,116,97,98,108,101,115,32,101,120,112,101,99,116,101,100,0,73,110,118,97,108,105,100,32,77,101,115,104,32,118,101,114,116,101,120,32,100,97,116,97,32,116,121,112,101,32,110,97,109,101,58,32,37,115,0,78,117,109,98,101,114,32,111,102,32,118,101,114,116,101,120,32,97,116,116,114,105,98,117,116,101,32,99,111,109,112,111,110,101,110,116,115,32,109,117,115,116,32,98,101,32,98,101,116,119,101,101,110,32,49,32,97,110,100,32,52,32,40,103,111,116,32,37,100,41,0,101,120,112,101,99,116,101,100,32,116,97,98,108,101,32,111,102,32,116,97,98,108,101,115,0,116,97,98,108,101,32,111,102,32,116,97,98,108,101,115,0,118,105,100,101,111,0,110,101,119,86,105,100,101,111,83,116,114,101,97,109,0,73,110,118,97,108,105,100,32,98,108,101,110,100,32,109,111,100,101,58,32,37,115,0,73,110,118,97,108,105,100,32,98,108,101,110,100,32,97,108,112,104,97,32,109,111,100,101,58,32,37,115,0,85,110,107,110,111,119,110,32,98,108,101,110,100,32,109,111,100,101,0,85,110,107,110,111,119,110,32,98,108,101,110,100,32,97,108,112,104,97,32,109,111,100,101,0,85,110,107,110,111,119,110,32,109,105,110,105,102,105,99,97,116,105,111,110,32,102,105,108,116,101,114,32,109,111,100,101,0,85,110,107,110,111,119,110,32,109,97,103,110,105,102,105,99,97,116,105,111,110,32,102,105,108,116,101,114,32,109,111,100,101,0,73,110,118,97,108,105,100,32,108,105,110,101,32,115,116,121,108,101,58,32,37,115,0,73,110,118,97,108,105,100,32,108,105,110,101,32,106,111,105,110,32,109,111,100,101,58,32,37,115,0,85,110,107,110,111,119,110,32,108,105,110,101,32,115,116,121,108,101,0,85,110,107,110,111,119,110,32,108,105,110,101,32,106,111,105,110,0,84,114,105,101,100,32,116,111,32,103,101,116,32,110,111,110,101,120,105,115,116,97,110,116,32,109,111,100,117,108,101,32,37,115,46,0,73,110,99,111,114,114,101,99,116,32,109,111,100,117,108,101,32,37,115,0,111,112,101,110,103,108,0,111,112,101,110,103,108,101,115,0,73,110,99,111,114,114,101,99,116,32,97,108,105,103,110,109,101,110,116,58,32,37,115,0,78,117,109,98,101,114,32,111,102,32,118,101,114,116,101,120,32,99,111,109,112,111,110,101,110,116,115,32,109,117,115,116,32,98,101,32,97,32,109,117,108,116,105,112,108,101,32,111,102,32,116,119,111,0,78,101,101,100,32,97,116,32,108,101,97,115,116,32,116,119,111,32,118,101,114,116,105,99,101,115,32,116,111,32,100,114,97,119,32,97,32,108,105,110,101,0,73,110,118,97,108,105,100,32,100,114,97,119,32,109,111,100,101,58,32,37,115,0,73,110,118,97,108,105,100,32,97,114,99,32,109,111,100,101,58,32,37,115,0,78,101,101,100,32,97,116,32,108,101,97,115,116,32,116,104,114,101,101,32,118,101,114,116,105,99,101,115,32,116,111,32,100,114,97,119,32,97,32,112,111,108,121,103,111,110,0,73,110,118,97,108,105,100,32,103,114,97,112,104,105,99,115,32,115,116,97,99,107,32,116,121,112,101,58,32,37,115,0,100,105,115,99,97,114,100,0,112,114,101,115,101,110,116,0,110,101,119,73,109,97,103,101,0,110,101,119,81,117,97,100,0,110,101,119,70,111,110,116,0,110,101,119,73,109,97,103,101,70,111,110,116,0,110,101,119,83,112,114,105,116,101,66,97,116,99,104,0,110,101,119,80,97,114,116,105,99,108,101,83,121,115,116,101,109,0,110,101,119,67,97,110,118,97,115,0,110,101,119,83,104,97,100,101,114,0,110,101,119,77,101,115,104,0,110,101,119,84,101,120,116,0,95,110,101,119,86,105,100,101,111,0,115,101,116,66,97,99,107,103,114,111,117,110,100,67,111,108,111,114,0,103,101,116,66,97,99,107,103,114,111,117,110,100,67,111,108,111,114,0,115,101,116,78,101,119,70,111,110,116,0,115,101,116,67,111,108,111,114,77,97,115,107,0,103,101,116,67,111,108,111,114,77,97,115,107,0,115,101,116,66,108,101,110,100,77,111,100,101,0,103,101,116,66,108,101,110,100,77,111,100,101,0,115,101,116,68,101,102,97,117,108,116,70,105,108,116,101,114,0,103,101,116,68,101,102,97,117,108,116,70,105,108,116,101,114,0,115,101,116,68,101,102,97,117,108,116,77,105,112,109,97,112,70,105,108,116,101,114,0,103,101,116,68,101,102,97,117,108,116,77,105,112,109,97,112,70,105,108,116,101,114,0,115,101,116,76,105,110,101,87,105,100,116,104,0,115,101,116,76,105,110,101,83,116,121,108,101,0,115,101,116,76,105,110,101,74,111,105,110,0,103,101,116,76,105,110,101,87,105,100,116,104,0,103,101,116,76,105,110,101,83,116,121,108,101,0,103,101,116,76,105,110,101,74,111,105,110,0,115,101,116,80,111,105,110,116,83,105,122,101,0,103,101,116,80,111,105,110,116,83,105,122,101,0,115,101,116,87,105,114,101,102,114,97,109,101,0,105,115,87,105,114,101,102,114,97,109,101,0,110,101,119,83,99,114,101,101,110,115,104,111,116,0,115,101,116,67,97,110,118,97,115,0,103,101,116,67,97,110,118,97,115,0,115,101,116,83,104,97,100,101,114,0,103,101,116,83,104,97,100,101,114,0,95,115,101,116,68,101,102,97,117,108,116,83,104,97,100,101,114,67,111,100,101,0,103,101,116,83,117,112,112,111,114,116,101,100,0,103,101,116,67,97,110,118,97,115,70,111,114,109,97,116,115,0,103,101,116,67,111,109,112,114,101,115,115,101,100,73,109,97,103,101,70,111,114,109,97,116,115,0,103,101,116,82,101,110,100,101,114,101,114,73,110,102,111,0,103,101,116,83,121,115,116,101,109,76,105,109,105,116,115,0,103,101,116,83,116,97,116,115,0,100,114,97,119,0,112,114,105,110,116,102,0,105,115,67,114,101,97,116,101,100,0,105,115,71,97,109,109,97,67,111,114,114,101,99,116,0,115,101,116,83,99,105,115,115,111,114,0,105,110,116,101,114,115,101,99,116,83,99,105,115,115,111,114,0,103,101,116,83,99,105,115,115,111,114,0,115,116,101,110,99,105,108,0,115,101,116,83,116,101,110,99,105,108,84,101,115,116,0,103,101,116,83,116,101,110,99,105,108,84,101,115,116,0,114,101,99,116,97,110,103,108,101,0,99,105,114,99,108,101,0,101,108,108,105,112,115,101,0,97,114,99,0,112,111,108,121,103,111,110,0,115,104,101,97,114,0,111,114,105,103,105,110,0,68,114,97,119,97,98,108,101,0,10,45,45,32,68,79,32,78,79,84,32,82,69,77,79,86,69,32,84,72,69,32,65,66,79,86,69,32,76,73,78,69,46,32,73,116,32,105,115,32,117,115,101,100,32,116,111,32,108,111,97,100,32,116,104,105,115,32,102,105,108,101,32,97,115,32,97,32,67,43,43,32,115,116,114,105,110,103,46,10,45,45,32,84,104,101,114,101,32,105,115,32,97,32,109,97,116,99,104,105,110,103,32,100,101,108,105,109,105,116,101,114,32,97,116,32,116,104,101,32,98,111,116,116,111,109,32,111,102,32,116,104,101,32,102,105,108,101,46,10,10,45,45,91,91,10,67,111,112,121,114,105,103,104,116,32,40,99,41,32,50,48,48,54,45,50,48,49,54,32,76,79,86,69,32,68,101,118,101,108,111,112,109,101,110,116,32,84,101,97,109,10,10,84,104,105,115,32,115,111,102,116,119,97,114,101,32,105,115,32,112,114,111,118,105,100,101,100,32,39,97,115,45,105,115,39,44,32,119,105,116,104,111,117,116,32,97,110,121,32,101,120,112,114,101,115,115,32,111,114,32,105,109,112,108,105,101,100,10,119,97,114,114,97,110,116,121,46,32,32,73,110,32,110,111,32,101,118,101,110,116,32,119,105,108,108,32,116,104,101,32,97,117,116,104,111,114,115,32,98,101,32,104,101,108,100,32,108,105,97,98,108,101,32,102,111,114,32,97,110,121,32,100,97,109,97,103,101,115,10,97,114,105,115,105,110,103,32,102,114,111,109,32,116,104,101,32,117,115,101,32,111,102,32,116,104,105,115,32,115,111,102,116,119,97,114,101,46,10,10,80,101,114,109,105,115,115,105,111,110,32,105,115,32,103,114,97,110,116,101,100,32,116,111,32,97,110,121,111,110,101,32,116,111,32,117,115,101,32,116,104,105,115,32,115,111,102,116,119,97,114,101,32,102,111,114,32,97,110,121,32,112,117,114,112,111,115,101,44,10,105,110,99,108,117,100,105,110,103,32,99,111,109,109,101,114,99,105,97,108,32,97,112,112,108,105,99,97,116,105,111,110,115,44,32,97,110,100,32,116,111,32,97,108,116,101,114,32,105,116,32,97,110,100,32,114,101,100,105,115,116,114,105,98,117,116,101,32,105,116,10,102,114,101,101,108,121,44,32,115,117,98,106,101,99,116,32,116,111,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,114,101,115,116,114,105,99,116,105,111,110,115,58,10,10,49,46,32,84,104,101,32,111,114,105,103,105,110,32,111,102,32,116,104,105,115,32,115,111,102,116,119,97,114,101,32,109,117,115,116,32,110,111,116,32,98,101,32,109,105,115,114,101,112,114,101,115,101,110,116,101,100,59,32,121,111,117,32,109,117,115,116,32,110,111,116,10,99,108,97,105,109,32,116,104,97,116,32,121,111,117,32,119,114,111,116,101,32,116,104,101,32,111,114,105,103,105,110,97,108,32,115,111,102,116,119,97,114,101,46,32,73,102,32,121,111,117,32,117,115,101,32,116,104,105,115,32,115,111,102,116,119,97,114,101,10,105,110,32,97,32,112,114,111,100,117,99,116,44,32,97,110,32,97,99,107,110,111,119,108,101,100,103,109,101,110,116,32,105,110,32,116,104,101,32,112,114,111,100,117,99,116,32,100,111,99,117,109,101,110,116,97,116,105,111,110,32,119,111,117,108,100,32,98,101,10,97,112,112,114,101,99,105,97,116,101,100,32,98,117,116,32,105,115,32,110,111,116,32,114,101,113,117,105,114,101,100,46,10,50,46,32,65,108,116,101,114,101,100,32,115,111,117,114,99,101,32,118,101,114,115,105,111,110,115,32,109,117,115,116,32,98,101,32,112,108,97,105,110,108,121,32,109,97,114,107,101,100,32,97,115,32,115,117,99,104,44,32,97,110,100,32,109,117,115,116,32,110,111,116,32,98,101,10,109,105,115,114,101,112,114,101,115,101,110,116,101,100,32,97,115,32,98,101,105,110,103,32,116,104,101,32,111,114,105,103,105,110,97,108,32,115,111,102,116,119,97,114,101,46,10,51,46,32,84,104,105,115,32,110,111,116,105,99,101,32,109,97,121,32,110,111,116,32,98,101,32,114,101,109,111,118,101,100,32,111,114,32,97,108,116,101,114,101,100,32,102,114,111,109,32,97,110,121,32,115,111,117,114,99,101,32,100,105,115,116,114,105,98,117,116,105,111,110,46,10,45,45,93,93,10,10,108,111,99,97,108,32,116,97,98,108,101,95,99,111,110,99,97,116,32,61,32,116,97,98,108,101,46,99,111,110,99,97,116,10,10,45,45,32,83,72,65,68,69,82,83,10,10,108,111,99,97,108,32,71,76,83,76,32,61,32,123,125,10,10,71,76,83,76,46,86,69,82,83,73,79,78,32,61,32,34,35,118,101,114,115,105,111,110,32,49,50,48,34,10,71,76,83,76,46,86,69,82,83,73,79,78,95,69,83,32,61,32,34,35,118,101,114,115,105,111,110,32,49,48,48,34,10,10,71,76,83,76,46,83,89,78,84,65,88,32,61,32,91,91,10,35,105,102,110,100,101,102,32,71,76,95,69,83,10,35,100,101,102,105,110,101,32,108,111,119,112,10,35,100,101,102,105,110,101,32,109,101,100,105,117,109,112,10,35,100,101,102,105,110,101,32,104,105,103,104,112,10,35,101,110,100,105,102,10,35,100,101,102,105,110,101,32,110,117,109,98,101,114,32,102,108,111,97,116,10,35,100,101,102,105,110,101,32,73,109,97,103,101,32,115,97,109,112,108,101,114,50,68,10,35,100,101,102,105,110,101,32,101,120,116,101,114,110,32,117,110,105,102,111,114,109,10,35,100,101,102,105,110,101,32,84,101,120,101,108,32,116,101,120,116,117,114,101,50,68,10,35,112,114,97,103,109,97,32,111,112,116,105,111,110,78,86,40,115,116,114,105,99,116,32,111,110,41,93,93,10,10,45,45,32,85,110,105,102,111,114,109,115,32,115,104,97,114,101,100,32,98,121,32,116,104,101,32,118,101,114,116,101,120,32,97,110,100,32,112,105,120,101,108,32,115,104,97,100,101,114,32,115,116,97,103,101,115,46,10,71,76,83,76,46,85,78,73,70,79,82,77,83,32,61,32,91,91,10,35,105,102,100,101,102,32,71,76,95,69,83,10,47,47,32,65,99,99,111,114,100,105,110,103,32,116,111,32,116,104,101,32,71,76,83,76,32,69,83,32,49,46,48,32,115,112,101,99,44,32,117,110,105,102,111,114,109,32,112,114,101,99,105,115,105,111,110,32,109,117,115,116,32,109,97,116,99,104,32,98,101,116,119,101,101,110,32,115,116,97,103,101,115,44,10,47,47,32,98,117,116,32,119,101,32,99,97,110,39,116,32,103,117,97,114,97,110,116,101,101,32,116,104,97,116,32,104,105,103,104,112,32,105,115,32,97,108,119,97,121,115,32,115,117,112,112,111,114,116,101,100,32,105,110,32,102,114,97,103,109,101,110,116,32,115,104,97,100,101,114,115,46,46,46,10,47,47,32,87,101,32,42,114,101,97,108,108,121,42,32,100,111,110,39,116,32,119,97,110,116,32,116,111,32,117,115,101,32,109,101,100,105,117,109,112,32,102,111,114,32,116,104,101,115,101,32,105,110,32,118,101,114,116,101,120,32,115,104,97,100,101,114,115,32,116,104,111,117,103,104,46,10,35,105,102,32,100,101,102,105,110,101,100,40,86,69,82,84,69,88,41,32,124,124,32,100,101,102,105,110,101,100,40,71,76,95,70,82,65,71,77,69,78,84,95,80,82,69,67,73,83,73,79,78,95,72,73,71,72,41,10,35,100,101,102,105,110,101,32,76,79,86,69,95,85,78,73,70,79,82,77,95,80,82,69,67,73,83,73,79,78,32,104,105,103,104,112,10,35,101,108,115,101,10,35,100,101,102,105,110,101,32,76,79,86,69,95,85,78,73,70,79,82,77,95,80,82,69,67,73,83,73,79,78,32,109,101,100,105,117,109,112,10,35,101,110,100,105,102,10,117,110,105,102,111,114,109,32,76,79,86,69,95,85,78,73,70,79,82,77,95,80,82,69,67,73,83,73,79,78,32,109,97,116,52,32,84,114,97,110,115,102,111,114,109,77,97,116,114,105,120,59,10,117,110,105,102,111,114,109,32,76,79,86,69,95,85,78,73,70,79,82,77,95,80,82,69,67,73,83,73,79,78,32,109,97,116,52,32,80,114,111,106,101,99,116,105,111,110,77,97,116,114,105,120,59,10,117,110,105,102,111,114,109,32,76,79,86,69,95,85,78,73,70,79,82,77,95,80,82,69,67,73,83,73,79,78,32,109,97,116,52,32,84,114,97,110,115,102,111,114,109,80,114,111,106,101,99,116,105,111,110,77,97,116,114,105,120,59,10,117,110,105,102,111,114,109,32,76,79,86,69,95,85,78,73,70,79,82,77,95,80,82,69,67,73,83,73,79,78,32,109,97,116,51,32,78,111,114,109,97,108,77,97,116,114,105,120,59,10,35,101,108,115,101,10,35,100,101,102,105,110,101,32,84,114,97,110,115,102,111,114,109,77,97,116,114,105,120,32,103,108,95,77,111,100,101,108,86,105,101,119,77,97,116,114,105,120,10,35,100,101,102,105,110,101,32,80,114,111,106,101,99,116,105,111,110,77,97,116,114,105,120,32,103,108,95,80,114,111,106,101,99,116,105,111,110,77,97,116,114,105,120,10,35,100,101,102,105,110,101,32,84,114,97,110,115,102,111,114,109,80,114,111,106,101,99,116,105,111,110,77,97,116,114,105,120,32,103,108,95,77,111,100,101,108,86,105,101,119,80,114,111,106,101,99,116,105,111,110,77,97,116,114,105,120,10,35,100,101,102,105,110,101,32,78,111,114,109,97,108,77,97,116,114,105,120,32,103,108,95,78,111,114,109,97,108,77,97,116,114,105,120,10,35,101,110,100,105,102,10,117,110,105,102,111,114,109,32,109,101,100,105,117,109,112,32,118,101,99,52,32,108,111,118,101,95,83,99,114,101,101,110,83,105,122,101,59,93,93,10,10,71,76,83,76,46,70,85,78,67,84,73,79,78,83,32,61,32,91,91,10,102,108,111,97,116,32,103,97,109,109,97,84,111,76,105,110,101,97,114,80,114,101,99,105,115,101,40,102,108,111,97,116,32,99,41,32,123,10,9,114,101,116,117,114,110,32,99,32,60,61,32,48,46,48,52,48,52,53,32,63,32,99,32,42,32,48,46,48,55,55,51,57,57,51,56,48,56,48,52,57,53,52,32,58,32,112,111,119,40,40,99,32,43,32,48,46,48,53,53,41,32,42,32,48,46,57,52,55,56,54,55,50,57,56,53,55,56,50,44,32,50,46,52,41,59,10,125,10,118,101,99,51,32,103,97,109,109,97,84,111,76,105,110,101,97,114,80,114,101,99,105,115,101,40,118,101,99,51,32,99,41,32,123,10,9,98,118,101,99,51,32,108,101,113,32,61,32,108,101,115,115,84,104,97,110,69,113,117,97,108,40,99,44,32,118,101,99,51,40,48,46,48,52,48,52,53,41,41,59,10,9,99,46,114,32,61,32,108,101,113,46,114,32,63,32,99,46,114,32,42,32,48,46,48,55,55,51,57,57,51,56,48,56,48,52,57,53,52,32,58,32,112,111,119,40,40,99,46,114,32,43,32,48,46,48,53,53,41,32,42,32,48,46,57,52,55,56,54,55,50,57,56,53,55,56,50,44,32,50,46,52,41,59,10,9,99,46,103,32,61,32,108,101,113,46,103,32,63,32,99,46,103,32,42,32,48,46,48,55,55,51,57,57,51,56,48,56,48,52,57,53,52,32,58,32,112,111,119,40,40,99,46,103,32,43,32,48,46,48,53,53,41,32,42,32,48,46,57,52,55,56,54,55,50,57,56,53,55,56,50,44,32,50,46,52,41,59,10,9,99,46,98,32,61,32,108,101,113,46,98,32,63,32,99,46,98,32,42,32,48,46,48,55,55,51,57,57,51,56,48,56,48,52,57,53,52,32,58,32,112,111,119,40,40,99,46,98,32,43,32,48,46,48,53,53,41,32,42,32,48,46,57,52,55,56,54,55,50,57,56,53,55,56,50,44,32,50,46,52,41,59,10,9,114,101,116,117,114,110,32,99,59,10,125,10,118,101,99,52,32,103,97,109,109,97,84,111,76,105,110,101,97,114,80,114,101,99,105,115,101,40,118,101,99,52,32,99,41,32,123,32,114,101,116,117,114,110,32,118,101,99,52,40,103,97,109,109,97,84,111,76,105,110,101,97,114,80,114,101,99,105,115,101,40,99,46,114,103,98,41,44,32,99,46,97,41,59,32,125,10,102,108,111,97,116,32,108,105,110,101,97,114,84,111,71,97,109,109,97,80,114,101,99,105,115,101,40,102,108,111,97,116,32,99,41,32,123,10,9,114,101,116,117,114,110,32,99,32,60,32,48,46,48,48,51,49,51,48,56,32,63,32,99,32,42,32,49,50,46,57,50,32,58,32,49,46,48,53,53,32,42,32,112,111,119,40,99,44,32,49,46,48,32,47,32,50,46,52,41,32,45,32,48,46,48,53,53,59,10,125,10,118,101,99,51,32,108,105,110,101,97,114,84,111,71,97,109,109,97,80,114,101,99,105,115,101,40,118,101,99,51,32,99,41,32,123,10,9,98,118,101,99,51,32,108,116,32,61,32,108,101,115,115,84,104,97,110,69,113,117,97,108,40,99,44,32,118,101,99,51,40,48,46,48,48,51,49,51,48,56,41,41,59,10,9,99,46,114,32,61,32,108,116,46,114,32,63,32,99,46,114,32,42,32,49,50,46,57,50,32,58,32,49,46,48,53,53,32,42,32,112,111,119,40,99,46,114,44,32,49,46,48,32,47,32,50,46,52,41,32,45,32,48,46,48,53,53,59,10,9,99,46,103,32,61,32,108,116,46,103,32,63,32,99,46,103,32,42,32,49,50,46,57,50,32,58,32,49,46,48,53,53,32,42,32,112,111,119,40,99,46,103,44,32,49,46,48,32,47,32,50,46,52,41,32,45,32,48,46,48,53,53,59,10,9,99,46,98,32,61,32,108,116,46,98,32,63,32,99,46,98,32,42,32,49,50,46,57,50,32,58,32,49,46,48,53,53,32,42,32,112,111,119,40,99,46,98,44,32,49,46,48,32,47,32,50,46,52,41,32,45,32,48,46,48,53,53,59,10,9,114,101,116,117,114,110,32,99,59,10,125,10,118,101,99,52,32,108,105,110,101,97,114,84,111,71,97,109,109,97,80,114,101,99,105,115,101,40,118,101,99,52,32,99,41,32,123,32,114,101,116,117,114,110,32,118,101,99,52,40,108,105,110,101,97,114,84,111,71,97,109,109,97,80,114,101,99,105,115,101,40,99,46,114,103,98,41,44,32,99,46,97,41,59,32,125,10,10,47,47,32,112,111,119,40,120,44,32,50,46,50,41,32,105,115,110,39,116,32,97,110,32,97,109,97,122,105,110,103,32,97,112,112,114,111,120,105,109,97,116,105,111,110,44,32,98,117,116,32,97,116,32,108,101,97,115,116,32,105,116,39,115,32,101,102,102,105,99,105,101,110,116,46,46,46,10,109,101,100,105,117,109,112,32,102,108,111,97,116,32,103,97,109,109,97,84,111,76,105,110,101,97,114,70,97,115,116,40,109,101,100,105,117,109,112,32,102,108,111,97,116,32,99,41,32,123,32,114,101,116,117,114,110,32,112,111,119,40,109,97,120,40,99,44,32,48,46,48,41,44,32,50,46,50,41,59,32,125,10,109,101,100,105,117,109,112,32,118,101,99,51,32,103,97,109,109,97,84,111,76,105,110,101,97,114,70,97,115,116,40,109,101,100,105,117,109,112,32,118,101,99,51,32,99,41,32,123,32,114,101,116,117,114,110,32,112,111,119,40,109,97,120,40,99,44,32,118,101,99,51,40,48,46,48,41,41,44,32,118,101,99,51,40,50,46,50,41,41,59,32,125,10,109,101,100,105,117,109,112,32,118,101,99,52,32,103,97,109,109,97,84,111,76,105,110,101,97,114,70,97,115,116,40,109,101,100,105,117,109,112,32,118,101,99,52,32,99,41,32,123,32,114,101,116,117,114,110,32,118,101,99,52,40,103,97,109,109,97,84,111,76,105,110,101,97,114,70,97,115,116,40,99,46,114,103,98,41,44,32,99,46,97,41,59,32,125,10,109,101,100,105,117,109,112,32,102,108,111,97,116,32,108,105,110,101,97,114,84,111,71,97,109,109,97,70,97,115,116,40,109,101,100,105,117,109,112,32,102,108,111,97,116,32,99,41,32,123,32,114,101,116,117,114,110,32,112,111,119,40,109,97,120,40,99,44,32,48,46,48,41,44,32,49,46,48,32,47,32,50,46,50,41,59,32,125,10,109,101,100,105,117,109,112,32,118,101,99,51,32,108,105,110,101,97,114,84,111,71,97,109,109,97,70,97,115,116,40,109,101,100,105,117,109,112,32,118,101,99,51,32,99,41,32,123,32,114,101,116,117,114,110,32,112,111,119,40,109,97,120,40,99,44,32,118,101,99,51,40,48,46,48,41,41,44,32,118,101,99,51,40,49,46,48,32,47,32,50,46,50,41,41,59,32,125,10,109,101,100,105,117,109,112,32,118,101,99,52,32,108,105,110,101,97,114,84,111,71,97,109,109,97,70,97,115,116,40,109,101,100,105,117,109,112,32,118,101,99,52,32,99,41,32,123,32,114,101,116,117,114,110,32,118,101,99,52,40,108,105,110,101,97,114,84,111,71,97,109,109,97,70,97,115,116,40,99,46,114,103,98,41,44,32,99,46,97,41,59,32,125,10,10,35,105,102,100,101,102,32,76,79,86,69,95,80,82,69,67,73,83,69,95,71,65,77,77,65,10,35,100,101,102,105,110,101,32,103,97,109,109,97,84,111,76,105,110,101,97,114,32,103,97,109,109,97,84,111,76,105,110,101,97,114,80,114,101,99,105,115,101,10,35,100,101,102,105,110,101,32,108,105,110,101,97,114,84,111,71,97,109,109,97,32,108,105,110,101,97,114,84,111,71,97,109,109,97,80,114,101,99,105,115,101,10,35,101,108,115,101,10,35,100,101,102,105,110,101,32,103,97,109,109,97,84,111,76,105,110,101,97,114,32,103,97,109,109,97,84,111,76,105,110,101,97,114,70,97,115,116,10,35,100,101,102,105,110,101,32,108,105,110,101,97,114,84,111,71,97,109,109,97,32,108,105,110,101,97,114,84,111,71,97,109,109,97,70,97,115,116,10,35,101,110,100,105,102,10,10,35,105,102,100,101,102,32,76,79,86,69,95,71,65,77,77,65,95,67,79,82,82,69,67,84,10,35,100,101,102,105,110,101,32,103,97,109,109,97,67,111,114,114,101,99,116,67,111,108,111,114,32,103,97,109,109,97,84,111,76,105,110,101,97,114,10,35,100,101,102,105,110,101,32,117,110,71,97,109,109,97,67,111,114,114,101,99,116,67,111,108,111,114,32,108,105,110,101,97,114,84,111,71,97,109,109,97,10,35,100,101,102,105,110,101,32,103,97,109,109,97,67,111,114,114,101,99,116,67,111,108,111,114,80,114,101,99,105,115,101,32,103,97,109,109,97,84,111,76,105,110,101,97,114,80,114,101,99,105,115,101,10,35,100,101,102,105,110,101,32,117,110,71,97,109,109,97,67,111,114,114,101,99,116,67,111,108,111,114,80,114,101,99,105,115,101,32,108,105,110,101,97,114,84,111,71,97,109,109,97,80,114,101,99,105,115,101,10,35,100,101,102,105,110,101,32,103,97,109,109,97,67,111,114,114,101,99,116,67,111,108,111,114,70,97,115,116,32,103,97,109,109,97,84,111,76,105,110,101,97,114,70,97,115,116,10,35,100,101,102,105,110,101,32,117,110,71,97,109,109,97,67,111,114,114,101,99,116,67,111,108,111,114,70,97,115,116,32,108,105,110,101,97,114,84,111,71,97,109,109,97,70,97,115,116,10,35,101,108,115,101,10,35,100,101,102,105,110,101,32,103,97,109,109,97,67,111,114,114,101,99,116,67,111,108,111,114,10,35,100,101,102,105,110,101,32,117,110,71,97,109,109,97,67,111,114,114,101,99,116,67,111,108,111,114,10,35,100,101,102,105,110,101,32,103,97,109,109,97,67,111,114,114,101,99,116,67,111,108,111,114,80,114,101,99,105,115,101,10,35,100,101,102,105,110,101,32,117,110,71,97,109,109,97,67,111,114,114,101,99,116,67,111,108,111,114,80,114,101,99,105,115,101,10,35,100,101,102,105,110,101,32,103,97,109,109,97,67,111,114,114,101,99,116,67,111,108,111,114,70,97,115,116,10,35,100,101,102,105,110,101,32,117,110,71,97,109,109,97,67,111,114,114,101,99,116,67,111,108,111,114,70,97,115,116,10,35,101,110,100,105,102,93,93,10,10,71,76,83,76,46,86,69,82,84,69,88,32,61,32,123,10,9,72,69,65,68,69,82,32,61,32,91,91,10,35,100,101,102,105,110,101,32,86,69,82,84,69,88,10,35,100,101,102,105,110,101,32,76,79,86,69,95,80,82,69,67,73,83,69,95,71,65,77,77,65,10,10,97,116,116,114,105,98,117,116,101,32,118,101,99,52,32,86,101,114,116,101,120,80,111,115,105,116,105,111,110,59,10,97,116,116,114,105,98,117,116,101,32,118,101,99,52,32,86,101,114,116,101,120,84,101,120,67,111,111,114,100,59,10,97,116,116,114,105,98,117,116,101,32,118,101,99,52,32,86,101,114,116,101,120,67,111,108,111,114,59,10,97,116,116,114,105,98,117,116,101,32,118,101,99,52,32,67,111,110,115,116,97,110,116,67,111,108,111,114,59,10,10,118,97,114,121,105,110,103,32,118,101,99,52,32,86,97,114,121,105,110,103,84,101,120,67,111,111,114,100,59,10,118,97,114,121,105,110,103,32,118,101,99,52,32,86,97,114,121,105,110,103,67,111,108,111,114,59,10,10,35,105,102,100,101,102,32,71,76,95,69,83,10,117,110,105,102,111,114,109,32,109,101,100,105,117,109,112,32,102,108,111,97,116,32,108,111,118,101,95,80,111,105,110,116,83,105,122,101,59,10,35,101,110,100,105,102,93,93,44,10,10,9,70,79,79,84,69,82,32,61,32,91,91,10,118,111,105,100,32,109,97,105,110,40,41,32,123,10,9,86,97,114,121,105,110,103,84,101,120,67,111,111,114,100,32,61,32,86,101,114,116,101,120,84,101,120,67,111,111,114,100,59,10,9,86,97,114,121,105,110,103,67,111,108,111,114,32,61,32,103,97,109,109,97,67,111,114,114,101,99,116,67,111,108,111,114,40,86,101,114,116,101,120,67,111,108,111,114,41,32,42,32,67,111,110,115,116,97,110,116,67,111,108,111,114,59,10,35,105,102,100,101,102,32,71,76,95,69,83,10,9,103,108,95,80,111,105,110,116,83,105,122,101,32,61,32,108,111,118,101,95,80,111,105,110,116,83,105,122,101,59,10,35,101,110,100,105,102,10,9,103,108,95,80,111,115,105,116,105,111,110,32,61,32,112,111,115,105,116,105,111,110,40,84,114,97,110,115,102,111,114,109,80,114,111,106,101,99,116,105,111,110,77,97,116,114,105,120,44,32,86,101,114,116,101,120,80,111,115,105,116,105,111,110,41,59,10,125,93,93,44,10,125,10,10,71,76,83,76,46,80,73,88,69,76,32,61,32,123,10,9,72,69,65,68,69,82,32,61,32,91,91,10,35,100,101,102,105,110,101,32,80,73,88,69,76,10,10,35,105,102,100,101,102,32,71,76,95,69,83,10,112,114,101,99,105,115,105,111,110,32,109,101,100,105,117,109,112,32,102,108,111,97,116,59,10,35,101,110,100,105,102,10,10,118,97,114,121,105,110,103,32,109,101,100,105,117,109,112,32,118,101,99,52,32,86,97,114,121,105,110,103,84,101,120,67,111,111,114,100,59,10,118,97,114,121,105,110,103,32,109,101,100,105,117,109,112,32,118,101,99,52,32,86,97,114,121,105,110,103,67,111,108,111,114,59,10,10,35,100,101,102,105,110,101,32,108,111,118,101,95,67,97,110,118,97,115,101,115,32,103,108,95,70,114,97,103,68,97,116,97,10,10,117,110,105,102,111,114,109,32,115,97,109,112,108,101,114,50,68,32,95,116,101,120,48,95,59,93,93,44,10,10,9,70,85,78,67,84,73,79,78,83,32,61,32,91,91,10,117,110,105,102,111,114,109,32,115,97,109,112,108,101,114,50,68,32,108,111,118,101,95,86,105,100,101,111,89,67,104,97,110,110,101,108,59,10,117,110,105,102,111,114,109,32,115,97,109,112,108,101,114,50,68,32,108,111,118,101,95,86,105,100,101,111,67,98,67,104,97,110,110,101,108,59,10,117,110,105,102,111,114,109,32,115,97,109,112,108,101,114,50,68,32,108,111,118,101,95,86,105,100,101,111,67,114,67,104,97,110,110,101,108,59,10,10,118,101,99,52,32,86,105,100,101,111,84,101,120,101,108,40,118,101,99,50,32,116,101,120,99,111,111,114,100,115,41,10,123,10,9,118,101,99,51,32,121,117,118,59,10,9,121,117,118,91,48,93,32,61,32,84,101,120,101,108,40,108,111,118,101,95,86,105,100,101,111,89,67,104,97,110,110,101,108,44,32,116,101,120,99,111,111,114,100,115,41,46,114,59,10,9,121,117,118,91,49,93,32,61,32,84,101,120,101,108,40,108,111,118,101,95,86,105,100,101,111,67,98,67,104,97,110,110,101,108,44,32,116,101,120,99,111,111,114,100,115,41,46,114,59,10,9,121,117,118,91,50,93,32,61,32,84,101,120,101,108,40,108,111,118,101,95,86,105,100,101,111,67,114,67,104,97,110,110,101,108,44,32,116,101,120,99,111,111,114,100,115,41,46,114,59,10,9,121,117,118,32,43,61,32,118,101,99,51,40,45,48,46,48,54,50,55,52,53,49,48,49,55,44,32,45,48,46,53,48,49,57,54,48,56,49,52,44,32,45,48,46,53,48,49,57,54,48,56,49,52,41,59,10,10,9,118,101,99,52,32,99,111,108,111,114,59,10,9,99,111,108,111,114,46,114,32,61,32,100,111,116,40,121,117,118,44,32,118,101,99,51,40,49,46,49,54,52,44,32,32,48,46,48,48,48,44,32,32,49,46,53,57,54,41,41,59,10,9,99,111,108,111,114,46,103,32,61,32,100,111,116,40,121,117,118,44,32,118,101,99,51,40,49,46,49,54,52,44,32,45,48,46,51,57,49,44,32,45,48,46,56,49,51,41,41,59,10,9,99,111,108,111,114,46,98,32,61,32,100,111,116,40,121,117,118,44,32,118,101,99,51,40,49,46,49,54,52,44,32,32,50,46,48,49,56,44,32,32,48,46,48,48,48,41,41,59,10,9,99,111,108,111,114,46,97,32,61,32,49,46,48,59,10,10,9,114,101,116,117,114,110,32,103,97,109,109,97,67,111,114,114,101,99,116,67,111,108,111,114,40,99,111,108,111,114,41,59,10,125,93,93,44,10,10,9,70,79,79,84,69,82,32,61,32,91,91,10,118,111,105,100,32,109,97,105,110,40,41,32,123,10,9,47,47,32,102,105,120,32,99,114,97,115,104,105,110,103,32,105,115,115,117,101,32,105,110,32,79,83,88,32,119,104,101,110,32,95,116,101,120,48,95,32,105,115,32,117,110,117,115,101,100,32,119,105,116,104,105,110,32,101,102,102,101,99,116,40,41,10,9,102,108,111,97,116,32,100,117,109,109,121,32,61,32,84,101,120,101,108,40,95,116,101,120,48,95,44,32,118,101,99,50,40,46,53,41,41,46,114,59,10,10,9,47,47,32,83,101,101,32,83,104,97,100,101,114,58,58,99,104,101,99,107,83,101,116,83,99,114,101,101,110,80,97,114,97,109,115,32,105,110,32,83,104,97,100,101,114,46,99,112,112,46,10,9,118,101,99,50,32,112,105,120,101,108,99,111,111,114,100,32,61,32,118,101,99,50,40,103,108,95,70,114,97,103,67,111,111,114,100,46,120,44,32,40,103,108,95,70,114,97,103,67,111,111,114,100,46,121,32,42,32,108,111,118,101,95,83,99,114,101,101,110,83,105,122,101,46,122,41,32,43,32,108,111,118,101,95,83,99,114,101,101,110,83,105,122,101,46,119,41,59,10,10,9,103,108,95,70,114,97,103,67,111,108,111,114,32,61,32,101,102,102,101,99,116,40,86,97,114,121,105,110,103,67,111,108,111,114,44,32,95,116,101,120,48,95,44,32,86,97,114,121,105,110,103,84,101,120,67,111,111,114,100,46,115,116,44,32,112,105,120,101,108,99,111,111,114,100,41,59,10,125,93,93,44,10,10,9,70,79,79,84,69,82,95,77,85,76,84,73,95,67,65,78,86,65,83,32,61,32,91,91,10,118,111,105,100,32,109,97,105,110,40,41,32,123,10,9,47,47,32,102,105,120,32,99,114,97,115,104,105,110,103,32,105,115,115,117,101,32,105,110,32,79,83,88,32,119,104,101,110,32,95,116,101,120,48,95,32,105,115,32,117,110,117,115,101,100,32,119,105,116,104,105,110,32,101,102,102,101,99,116,40,41,10,9,102,108,111,97,116,32,100,117,109,109,121,32,61,32,84,101,120,101,108,40,95,116,101,120,48,95,44,32,118,101,99,50,40,46,53,41,41,46,114,59,10,10,9,47,47,32,83,101,101,32,83,104,97,100,101,114,58,58,99,104,101,99,107,83,101,116,83,99,114,101,101,110,80,97,114,97,109,115,32,105,110,32,83,104,97,100,101,114,46,99,112,112,46,10,9,118,101,99,50,32,112,105,120,101,108,99,111,111,114,100,32,61,32,118,101,99,50,40,103,108,95,70,114,97,103,67,111,111,114,100,46,120,44,32,40,103,108,95,70,114,97,103,67,111,111,114,100,46,121,32,42,32,108,111,118,101,95,83,99,114,101,101,110,83,105,122,101,46,122,41,32,43,32,108,111,118,101,95,83,99,114,101,101,110,83,105,122,101,46,119,41,59,10,10,9,101,102,102,101,99,116,115,40,86,97,114,121,105,110,103,67,111,108,111,114,44,32,95,116,101,120,48,95,44,32,86,97,114,121,105,110,103,84,101,120,67,111,111,114,100,46,115,116,44,32,112,105,120,101,108,99,111,111,114,100,41,59,10,125,93,93,44,10,125,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,99,114,101,97,116,101,86,101,114,116,101,120,67,111,100,101,40,118,101,114,116,101,120,99,111,100,101,44,32,108,97,110,103,41,10,9,108,111,99,97,108,32,118,101,114,116,101,120,99,111,100,101,115,32,61,32,123,10,9,9,108,97,110,103,32,61,61,32,34,103,108,115,108,101,115,34,32,97,110,100,32,71,76,83,76,46,86,69,82,83,73,79,78,95,69,83,32,111,114,32,71,76,83,76,46,86,69,82,83,73,79,78,44,10,9,9,71,76,83,76,46,83,89,78,84,65,88,44,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,105,115,71,97,109,109,97,67,111,114,114,101,99,116,40,41,32,97,110,100,32,34,35,100,101,102,105,110,101,32,76,79,86,69,95,71,65,77,77,65,95,67,79,82,82,69,67,84,32,49,34,32,111,114,32,34,34,44,10,9,9,71,76,83,76,46,86,69,82,84,69,88,46,72,69,65,68,69,82,44,32,71,76,83,76,46,85,78,73,70,79,82,77,83,44,10,9,9,71,76,83,76,46,70,85,78,67,84,73,79,78,83,44,10,9,9,108,97,110,103,32,61,61,32,34,103,108,115,108,101,115,34,32,97,110,100,32,34,35,108,105,110,101,32,49,34,32,111,114,32,34,35,108,105,110,101,32,48,34,44,10,9,9,118,101,114,116,101,120,99,111,100,101,44,10,9,9,71,76,83,76,46,86,69,82,84,69,88,46,70,79,79,84,69,82,44,10,9,125,10,9,114,101,116,117,114,110,32,116,97,98,108,101,95,99,111,110,99,97,116,40,118,101,114,116,101,120,99,111,100,101,115,44,32,34,92,110,34,41,10,101,110,100,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,99,114,101,97,116,101,80,105,120,101,108,67,111,100,101,40,112,105,120,101,108,99,111,100,101,44,32,105,115,95,109,117,108,116,105,99,97,110,118,97,115,44,32,108,97,110,103,41,10,9,108,111,99,97,108,32,112,105,120,101,108,99,111,100,101,115,32,61,32,123,10,9,9,108,97,110,103,32,61,61,32,34,103,108,115,108,101,115,34,32,97,110,100,32,71,76,83,76,46,86,69,82,83,73,79,78,95,69,83,32,111,114,32,71,76,83,76,46,86,69,82,83,73,79,78,44,10,9,9,71,76,83,76,46,83,89,78,84,65,88,44,10,9,9,108,111,118,101,46,103,114,97,112,104,105,99,115,46,105,115,71,97,109,109,97,67,111,114,114,101,99,116,40,41,32,97,110,100,32,34,35,100,101,102,105,110,101,32,76,79,86,69,95,71,65,77,77,65,95,67,79,82,82,69,67,84,32,49,34,32,111,114,32,34,34,44,10,9,9,71,76,83,76,46,80,73,88,69,76,46,72,69,65,68,69,82,44,32,71,76,83,76,46,85,78,73,70,79,82,77,83,44,10,9,9,71,76,83,76,46,70,85,78,67,84,73,79,78,83,44,10,9,9,71,76,83,76,46,80,73,88,69,76,46,70,85,78,67,84,73,79,78,83,44,10,9,9,108,97,110,103,32,61,61,32,34,103,108,115,108,101,115,34,32,97,110,100,32,34,35,108,105,110,101,32,49,34,32,111,114,32,34,35,108,105,110,101,32,48,34,44,10,9,9,112,105,120,101,108,99,111,100,101,44,10,9,9,105,115,95,109,117,108,116,105,99,97,110,118,97,115,32,97,110,100,32,71,76,83,76,46,80,73,88,69,76,46,70,79,79,84,69,82,95,77,85,76,84,73,95,67,65,78,86,65,83,32,111,114,32,71,76,83,76,46,80,73,88,69,76,46,70,79,79,84,69,82,44,10,9,125,10,9,114,101,116,117,114,110,32,116,97,98,108,101,95,99,111,110,99,97,116,40,112,105,120,101,108,99,111,100,101,115,44,32,34,92,110,34,41,10,101,110,100,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,105,115,86,101,114,116,101,120,67,111,100,101,40,99,111,100,101,41,10,9,114,101,116,117,114,110,32,99,111,100,101,58,109,97,116,99,104,40,34,118,101,99,52,37,115,43,112,111,115,105,116,105,111,110,37,115,42,37,40,34,41,32,126,61,32,110,105,108,10,101,110,100,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,105,115,80,105,120,101,108,67,111,100,101,40,99,111,100,101,41,10,9,105,102,32,99,111,100,101,58,109,97,116,99,104,40,34,118,101,99,52,37,115,43,101,102,102,101,99,116,37,115,42,37,40,34,41,32,116,104,101,110,10,9,9,114,101,116,117,114,110,32,116,114,117,101,10,9,101,108,115,101,105,102,32,99,111,100,101,58,109,97,116,99,104,40,34,118,111,105,100,37,115,43,101,102,102,101,99,116,115,37,115,42,37,40,34,41,32,116,104,101,110,10,9,9,45,45,32,102,117,110,99,116,105,111,110,32,102,111,114,32,114,101,110,100,101,114,105,110,103,32,116,111,32,109,117,108,116,105,112,108,101,32,99,97,110,118,97,115,101,115,32,115,105,109,117,108,116,97,110,101,111,117,115,108,121,10,9,9,114,101,116,117,114,110,32,116,114,117,101,44,32,116,114,117,101,10,9,101,108,115,101,10,9,9,114,101,116,117,114,110,32,102,97,108,115,101,10,9,101,110,100,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,95,115,104,97,100,101,114,67,111,100,101,84,111,71,76,83,76,40,97,114,103,49,44,32,97,114,103,50,41,10,9,108,111,99,97,108,32,118,101,114,116,101,120,99,111,100,101,44,32,112,105,120,101,108,99,111,100,101,10,9,108,111,99,97,108,32,105,115,95,109,117,108,116,105,99,97,110,118,97,115,32,61,32,102,97,108,115,101,32,45,45,32,119,104,101,116,104,101,114,32,112,105,120,101,108,32,99,111,100,101,32,104,97,115,32,34,101,102,102,101,99,116,115,34,32,102,117,110,99,116,105,111,110,32,105,110,115,116,101,97,100,32,111,102,32,34,101,102,102,101,99,116,34,10,10,9,108,111,99,97,108,32,108,97,110,103,32,61,32,34,103,108,115,108,34,10,9,105,102,32,40,108,111,118,101,46,103,114,97,112,104,105,99,115,46,103,101,116,82,101,110,100,101,114,101,114,73,110,102,111,40,41,41,32,61,61,32,34,79,112,101,110,71,76,32,69,83,34,32,116,104,101,110,10,9,9,108,97,110,103,32,61,32,34,103,108,115,108,101,115,34,10,9,101,110,100,10,10,9,105,102,32,97,114,103,49,32,116,104,101,110,10,9,9,105,102,32,105,115,86,101,114,116,101,120,67,111,100,101,40,97,114,103,49,41,32,116,104,101,110,10,9,9,9,118,101,114,116,101,120,99,111,100,101,32,61,32,97,114,103,49,32,45,45,32,102,105,114,115,116,32,97,114,103,32,99,111,110,116,97,105,110,115,32,118,101,114,116,101,120,32,115,104,97,100,101,114,32,99,111,100,101,10,9,9,101,110,100,10,10,9,9,108,111,99,97,108,32,105,115,112,105,120,101,108,44,32,105,115,77,117,108,116,105,67,97,110,118,97,115,32,61,32,105,115,80,105,120,101,108,67,111,100,101,40,97,114,103,49,41,10,9,9,105,102,32,105,115,112,105,120,101,108,32,116,104,101,110,10,9,9,9,112,105,120,101,108,99,111,100,101,32,61,32,97,114,103,49,32,45,45,32,102,105,114,115,116,32,97,114,103,32,99,111,110,116,97,105,110,115,32,112,105,120,101,108,32,115,104,97,100,101,114,32,99,111,100,101,10,9,9,9,105,115,95,109,117,108,116,105,99,97,110,118,97,115,32,61,32,105,115,77,117,108,116,105,67,97,110,118,97,115,10,9,9,101,110,100,10,9,101,110,100,10,9,10,9,105,102,32,97,114,103,50,32,116,104,101,110,10,9,9,105,102,32,105,115,86,101,114,116,101,120,67,111,100,101,40,97,114,103,50,41,32,116,104,101,110,10,9,9,9,118,101,114,116,101,120,99,111,100,101,32,61,32,97,114,103,50,32,45,45,32,115,101,99,111,110,100,32,97,114,103,32,99,111,110,116,97,105,110,115,32,118,101,114,116,101,120,32,115,104,97,100,101,114,32,99,111,100,101,10,9,9,101,110,100,10,10,9,9,108,111,99,97,108,32,105,115,112,105,120,101,108,44,32,105,115,77,117,108,116,105,67,97,110,118,97,115,32,61,32,105,115,80,105,120,101,108,67,111,100,101,40,97,114,103,50,41,10,9,9,105,102,32,105,115,112,105,120,101,108,32,116,104,101,110,10,9,9,9,112,105,120,101,108,99,111,100,101,32,61,32,97,114,103,50,32,45,45,32,115,101,99,111,110,100,32,97,114,103,32,99,111,110,116,97,105,110,115,32,112,105,120,101,108,32,115,104,97,100,101,114,32,99,111,100,101,10,9,9,9,105,115,95,109,117,108,116,105,99,97,110,118,97,115,32,61,32,105,115,77,117,108,116,105,67,97,110,118,97,115,10,9,9,101,110,100,10,9,101,110,100,10,10,9,105,102,32,118,101,114,116,101,120,99,111,100,101,32,116,104,101,110,10,9,9,118,101,114,116,101,120,99,111,100,101,32,61,32,99,114,101,97,116,101,86,101,114,116,101,120,67,111,100,101,40,118,101,114,116,101,120,99,111,100,101,44,32,108,97,110,103,41,10,9,101,110,100,10,9,105,102,32,112,105,120,101,108,99,111,100,101,32,116,104,101,110,10,9,9,112,105,120,101,108,99,111,100,101,32,61,32,99,114,101,97,116,101,80,105,120,101,108,67,111,100,101,40,112,105,120,101,108,99,111,100,101,44,32,105,115,95,109,117,108,116,105,99,97,110,118,97,115,44,32,108,97,110,103,41,10,9,101,110,100,10,10,9,114,101,116,117,114,110,32,118,101,114,116,101,120,99,111,100,101,44,32,112,105,120,101,108,99,111,100,101,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,95,116,114,97,110,115,102,111,114,109,71,76,83,76,69,114,114,111,114,77,101,115,115,97,103,101,115,40,109,101,115,115,97,103,101,41,10,9,108,111,99,97,108,32,115,104,97,100,101,114,116,121,112,101,32,61,32,109,101,115,115,97,103,101,58,109,97,116,99,104,40,34,67,97,110,110,111,116,32,99,111,109,112,105,108,101,32,40,37,97,43,41,32,115,104,97,100,101,114,32,99,111,100,101,34,41,10,9,105,102,32,110,111,116,32,115,104,97,100,101,114,116,121,112,101,32,116,104,101,110,32,114,101,116,117,114,110,32,109,101,115,115,97,103,101,32,101,110,100,10,9,108,111,99,97,108,32,108,105,110,101,115,32,61,32,123,34,67,97,110,110,111,116,32,99,111,109,112,105,108,101,32,34,46,46,115,104,97,100,101,114,116,121,112,101,46,46,34,32,115,104,97,100,101,114,32,99,111,100,101,58,34,125,10,9,102,111,114,32,108,32,105,110,32,109,101,115,115,97,103,101,58,103,109,97,116,99,104,40,34,91,94,92,110,93,43,34,41,32,100,111,10,9,9,45,45,32,110,118,105,100,105,97,32,99,111,109,112,105,108,101,114,32,109,101,115,115,97,103,101,58,10,9,9,45,45,32,48,40,60,108,105,110,101,110,117,109,98,101,114,62,41,32,58,32,101,114,114,111,114,47,119,97,114,110,105,110,103,32,91,78,85,77,66,69,82,93,58,32,60,101,114,114,111,114,32,109,101,115,115,97,103,101,62,10,9,9,108,111,99,97,108,32,108,105,110,101,110,117,109,98,101,114,44,32,119,104,97,116,44,32,109,101,115,115,97,103,101,32,61,32,108,58,109,97,116,99,104,40,34,94,48,37,40,40,37,100,43,41,37,41,37,115,42,58,37,115,42,40,37,119,43,41,91,94,58,93,43,58,37,115,42,40,46,43,41,36,34,41,10,9,9,105,102,32,110,111,116,32,108,105,110,101,110,117,109,98,101,114,32,116,104,101,110,10,9,9,9,45,45,32,97,116,105,32,99,111,109,112,105,108,101,114,32,109,101,115,115,97,103,101,58,10,9,9,9,45,45,32,69,82,82,79,82,32,48,58,60,108,105,110,101,110,117,109,98,101,114,62,58,32,101,114,114,111,114,47,119,97,114,110,105,110,103,40,35,91,78,85,77,66,69,82,93,41,32,91,69,82,82,79,82,78,65,77,69,93,58,32,60,101,114,114,111,114,109,101,115,115,97,103,101,62,10,9,9,9,108,105,110,101,110,117,109,98,101,114,44,32,119,104,97,116,44,32,109,101,115,115,97,103,101,32,61,32,108,58,109,97,116,99,104,40,34,94,37,119,43,58,32,48,58,40,37,100,43,41,58,37,115,42,40,37,119,43,41,37,40,91,94,37,41,93,43,37,41,37,115,42,40,46,43,41,36,34,41,10,9,9,9,105,102,32,110,111,116,32,108,105,110,101,110,117,109,98,101,114,32,116,104,101,110,10,9,9,9,9,45,45,32,79,83,88,32,99,111,109,112,105,108,101,114,32,109,101,115,115,97,103,101,32,40,63,41,58,10,9,9,9,9,45,45,32,69,82,82,79,82,58,32,48,58,60,108,105,110,101,110,117,109,98,101,114,62,58,32,60,101,114,114,111,114,109,101,115,115,97,103,101,62,10,9,9,9,9,119,104,97,116,44,32,108,105,110,101,110,117,109,98,101,114,44,32,109,101,115,115,97,103,101,32,61,32,108,58,109,97,116,99,104,40,34,94,40,37,119,43,41,58,32,37,100,43,58,40,37,100,43,41,58,32,40,46,43,41,36,34,41,10,9,9,9,101,110,100,10,9,9,101,110,100,10,9,9,105,102,32,108,105,110,101,110,117,109,98,101,114,32,97,110,100,32,119,104,97,116,32,97,110,100,32,109,101,115,115,97,103,101,32,116,104,101,110,10,9,9,9,108,105,110,101,115,91,35,108,105,110,101,115,43,49,93,32,61,32,40,34,76,105,110,101,32,37,100,58,32,37,115,58,32,37,115,34,41,58,102,111,114,109,97,116,40,108,105,110,101,110,117,109,98,101,114,44,32,119,104,97,116,44,32,109,101,115,115,97,103,101,41,10,9,9,101,110,100,10,9,101,110,100,10,9,45,45,32,100,105,100,32,110,111,116,32,109,97,116,99,104,32,97,110,121,32,107,110,111,119,110,32,101,114,114,111,114,32,109,101,115,115,97,103,101,115,10,9,105,102,32,35,108,105,110,101,115,32,61,61,32,49,32,116,104,101,110,32,114,101,116,117,114,110,32,109,101,115,115,97,103,101,32,101,110,100,10,9,114,101,116,117,114,110,32,116,97,98,108,101,95,99,111,110,99,97,116,40,108,105,110,101,115,44,32,34,92,110,34,41,10,101,110,100,10,10,108,111,99,97,108,32,100,101,102,97,117,108,116,99,111,100,101,32,61,32,123,10,9,118,101,114,116,101,120,32,61,32,91,91,10,118,101,99,52,32,112,111,115,105,116,105,111,110,40,109,97,116,52,32,116,114,97,110,115,102,111,114,109,95,112,114,111,106,44,32,118,101,99,52,32,118,101,114,116,112,111,115,41,32,123,10,9,114,101,116,117,114,110,32,116,114,97,110,115,102,111,114,109,95,112,114,111,106,32,42,32,118,101,114,116,112,111,115,59,10,125,93,93,44,10,9,112,105,120,101,108,32,61,32,91,91,10,118,101,99,52,32,101,102,102,101,99,116,40,109,101,100,105,117,109,112,32,118,101,99,52,32,118,99,111,108,111,114,44,32,73,109,97,103,101,32,116,101,120,44,32,118,101,99,50,32,116,101,120,99,111,111,114,100,44,32,118,101,99,50,32,112,105,120,99,111,111,114,100,41,32,123,10,9,114,101,116,117,114,110,32,84,101,120,101,108,40,116,101,120,44,32,116,101,120,99,111,111,114,100,41,32,42,32,118,99,111,108,111,114,59,10,125,93,93,44,10,9,118,105,100,101,111,112,105,120,101,108,32,61,32,91,91,10,118,101,99,52,32,101,102,102,101,99,116,40,109,101,100,105,117,109,112,32,118,101,99,52,32,118,99,111,108,111,114,44,32,73,109,97,103,101,32,116,101,120,44,32,118,101,99,50,32,116,101,120,99,111,111,114,100,44,32,118,101,99,50,32,112,105,120,99,111,111,114,100,41,32,123,10,9,114,101,116,117,114,110,32,86,105,100,101,111,84,101,120,101,108,40,116,101,120,99,111,111,114,100,41,32,42,32,118,99,111,108,111,114,59,10,125,93,93,44,10,125,10,10,108,111,99,97,108,32,100,101,102,97,117,108,116,115,32,61,32,123,10,9,111,112,101,110,103,108,32,61,32,123,10,9,9,99,114,101,97,116,101,86,101,114,116,101,120,67,111,100,101,40,100,101,102,97,117,108,116,99,111,100,101,46,118,101,114,116,101,120,44,32,34,103,108,115,108,34,41,44,10,9,9,99,114,101,97,116,101,80,105,120,101,108,67,111,100,101,40,100,101,102,97,117,108,116,99,111,100,101,46,112,105,120,101,108,44,32,102,97,108,115,101,44,32,34,103,108,115,108,34,41,44,10,9,9,99,114,101,97,116,101,80,105,120,101,108,67,111,100,101,40,100,101,102,97,117,108,116,99,111,100,101,46,118,105,100,101,111,112,105,120,101,108,44,32,102,97,108,115,101,44,32,34,103,108,115,108,34,41,44,10,9,125,44,10,9,111,112,101,110,103,108,101,115,32,61,32,123,10,9,9,99,114,101,97,116,101,86,101,114,116,101,120,67,111,100,101,40,100,101,102,97,117,108,116,99,111,100,101,46,118,101,114,116,101,120,44,32,34,103,108,115,108,101,115,34,41,44,10,9,9,99,114,101,97,116,101,80,105,120,101,108,67,111,100,101,40,100,101,102,97,117,108,116,99,111,100,101,46,112,105,120,101,108,44,32,102,97,108,115,101,44,32,34,103,108,115,108,101,115,34,41,44,10,9,9,99,114,101,97,116,101,80,105,120,101,108,67,111,100,101,40,100,101,102,97,117,108,116,99,111,100,101,46,118,105,100,101,111,112,105,120,101,108,44,32,102,97,108,115,101,44,32,34,103,108,115,108,101,115,34,41,44,10,9,125,44,10,125,10,10,108,111,118,101,46,103,114,97,112,104,105,99,115,46,95,115,101,116,68,101,102,97,117,108,116,83,104,97,100,101,114,67,111,100,101,40,100,101,102,97,117,108,116,115,41,10,10,102,117,110,99,116,105,111,110,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,110,101,119,86,105,100,101,111,40,102,105,108,101,44,32,108,111,97,100,97,117,100,105,111,41,10,9,108,111,99,97,108,32,118,105,100,101,111,32,61,32,108,111,118,101,46,103,114,97,112,104,105,99,115,46,95,110,101,119,86,105,100,101,111,40,102,105,108,101,41,10,9,108,111,99,97,108,32,115,111,117,114,99,101,44,32,115,117,99,99,101,115,115,10,10,9,105,102,32,108,111,97,100,97,117,100,105,111,32,126,61,32,102,97,108,115,101,32,116,104,101,110,10,9,9,115,117,99,99,101,115,115,44,32,115,111,117,114,99,101,32,61,32,112,99,97,108,108,40,108,111,118,101,46,97,117,100,105,111,46,110,101,119,83,111,117,114,99,101,44,32,118,105,100,101,111,58,103,101,116,83,116,114,101,97,109,40,41,58,103,101,116,70,105,108,101,110,97,109,101,40,41,41,10,9,101,110,100,10,9,105,102,32,115,117,99,99,101,115,115,32,116,104,101,110,10,9,9,118,105,100,101,111,58,115,101,116,83,111,117,114,99,101,40,115,111,117,114,99,101,41,10,9,101,108,115,101,105,102,32,108,111,97,100,97,117,100,105,111,32,61,61,32,116,114,117,101,32,116,104,101,110,10,9,9,101,114,114,111,114,40,34,86,105,100,101,111,32,104,97,100,32,110,111,32,97,117,100,105,111,32,116,114,97,99,107,34,44,32,50,41,10,9,101,108,115,101,10,9,9,118,105,100,101,111,58,103,101,116,83,116,114,101,97,109,40,41,58,115,101,116,83,121,110,99,40,41,10,9,101,110,100,10,10,9,114,101,116,117,114,110,32,118,105,100,101,111,10,101,110,100,10,10,45,45,32,68,79,32,78,79,84,32,82,69,77,79,86,69,32,84,72,69,32,78,69,88,84,32,76,73,78,69,46,32,73,116,32,105,115,32,117,115,101,100,32,116,111,32,108,111,97,100,32,116,104,105,115,32,102,105,108,101,32,97,115,32,97,32,67,43,43,32,115,116,114,105,110,103,46,10,45,45,0,119,114,97,112,95,71,114,97,112,104,105,99,115,46,108,117,97,0,73,109,97,103,101,0,115,101,116,77,105,112,109,97,112,70,105,108,116,101,114,0,103,101,116,77,105,112,109,97,112,70,105,108,116,101,114,0,114,101,102,114,101,115,104,0,103,101,116,68,97,116,97,0,103,101,116,70,108,97,103,115,0,73,110,118,97,108,105,100,32,118,101,114,116,101,120,32,115,116,97,114,116,32,105,110,100,101,120,32,40,109,117,115,116,32,98,101,32,98,101,116,119,101,101,110,32,49,32,97,110,100,32,37,100,41,0,84,111,111,32,109,97,110,121,32,118,101,114,116,105,99,101,115,32,40,101,120,112,101,99,116,101,100,32,97,116,32,109,111,115,116,32,37,100,44,32,103,111,116,32,37,100,41,0,85,110,107,110,111,119,110,32,118,101,114,116,101,120,32,97,116,116,114,105,98,117,116,101,32,100,97,116,97,32,116,121,112,101,46,0,73,110,118,97,108,105,100,32,109,101,115,104,32,100,114,97,119,32,109,111,100,101,58,32,37,115,0,85,110,107,110,111,119,110,32,109,101,115,104,32,100,114,97,119,32,109,111,100,101,46,0,77,101,115,104,0,115,101,116,86,101,114,116,105,99,101,115,0,115,101,116,86,101,114,116,101,120,0,103,101,116,86,101,114,116,101,120,0,115,101,116,86,101,114,116,101,120,65,116,116,114,105,98,117,116,101,0,103,101,116,86,101,114,116,101,120,65,116,116,114,105,98,117,116,101,0,103,101,116,86,101,114,116,101,120,70,111,114,109,97,116,0,115,101,116,65,116,116,114,105,98,117,116,101,69,110,97,98,108,101,100,0,105,115,65,116,116,114,105,98,117,116,101,69,110,97,98,108,101,100,0,115,101,116,86,101,114,116,101,120,77,97,112,0,103,101,116,86,101,114,116,101,120,77,97,112,0,115,101,116,68,114,97,119,77,111,100,101,0,103,101,116,68,114,97,119,77,111,100,101,0,115,101,116,68,114,97,119,82,97,110,103,101,0,103,101,116,68,114,97,119,82,97,110,103,101,0,73,110,118,97,108,105,100,32,105,110,115,101,114,116,32,109,111,100,101,58,32,39,37,115,39,0,85,110,107,110,111,119,110,32,105,110,115,101,114,116,32,109,111,100,101,0,73,110,118,97,108,105,100,32,112,97,114,116,105,99,108,101,32,100,105,115,116,114,105,98,117,116,105,111,110,58,32,37,115,0,73,110,118,97,108,105,100,32,97,114,101,97,32,115,112,114,101,97,100,32,112,97,114,97,109,101,116,101,114,115,32,40,109,117,115,116,32,98,101,32,62,61,32,48,41,0,65,116,32,109,111,115,116,32,101,105,103,104,116,32,40,56,41,32,115,105,122,101,115,32,109,97,121,32,98,101,32,117,115,101,100,46,0,83,105,122,101,32,118,97,114,105,97,116,105,111,110,32,104,97,115,32,116,111,32,98,101,32,98,101,116,119,101,101,110,32,48,32,97,110,100,32,49,44,32,105,110,99,108,117,115,105,118,101,46,0,65,116,32,109,111,115,116,32,101,105,103,104,116,32,40,56,41,32,99,111,108,111,114,115,32,109,97,121,32,98,101,32,117,115,101,100,46,0,101,120,112,101,99,116,101,100,32,52,32,99,111,108,111,114,32,99,111,109,112,111,110,101,110,116,115,0,69,120,112,101,99,116,101,100,32,114,101,100,44,32,103,114,101,101,110,44,32,98,108,117,101,44,32,97,110,100,32,97,108,112,104,97,46,32,79,110,108,121,32,103,111,116,32,37,100,32,111,102,32,52,32,99,111,109,112,111,110,101,110,116,115,46,0,80,97,114,116,105,99,108,101,83,121,115,116,101,109,0,99,108,111,110,101,0,115,101,116,73,110,115,101,114,116,77,111,100,101,0,103,101,116,73,110,115,101,114,116,77,111,100,101,0,115,101,116,69,109,105,115,115,105,111,110,82,97,116,101,0,103,101,116,69,109,105,115,115,105,111,110,82,97,116,101,0,115,101,116,69,109,105,116,116,101,114,76,105,102,101,116,105,109,101,0,103,101,116,69,109,105,116,116,101,114,76,105,102,101,116,105,109,101,0,115,101,116,80,97,114,116,105,99,108,101,76,105,102,101,116,105,109,101,0,103,101,116,80,97,114,116,105,99,108,101,76,105,102,101,116,105,109,101,0,109,111,118,101,84,111,0,115,101,116,65,114,101,97,83,112,114,101,97,100,0,103,101,116,65,114,101,97,83,112,114,101,97,100,0,115,101,116,68,105,114,101,99,116,105,111,110,0,103,101,116,68,105,114,101,99,116,105,111,110,0,115,101,116,83,112,114,101,97,100,0,103,101,116,83,112,114,101,97,100,0,115,101,116,83,112,101,101,100,0,103,101,116,83,112,101,101,100,0,115,101,116,76,105,110,101,97,114,65,99,99,101,108,101,114,97,116,105,111,110,0,103,101,116,76,105,110,101,97,114,65,99,99,101,108,101,114,97,116,105,111,110,0,115,101,116,82,97,100,105,97,108,65,99,99,101,108,101,114,97,116,105,111,110,0,103,101,116,82,97,100,105,97,108,65,99,99,101,108,101,114,97,116,105,111,110,0,115,101,116,84,97,110,103,101,110,116,105,97,108,65,99,99,101,108,101,114,97,116,105,111,110,0,103,101,116,84,97,110,103,101,110,116,105,97,108,65,99,99,101,108,101,114,97,116,105,111,110,0,115,101,116,83,105,122,101,115,0,103,101,116,83,105,122,101,115,0,115,101,116,83,105,122,101,86,97,114,105,97,116,105,111,110,0,103,101,116,83,105,122,101,86,97,114,105,97,116,105,111,110,0,115,101,116,82,111,116,97,116,105,111,110,0,103,101,116,82,111,116,97,116,105,111,110,0,115,101,116,83,112,105,110,0,103,101,116,83,112,105,110,0,115,101,116,83,112,105,110,86,97,114,105,97,116,105,111,110,0,103,101,116,83,112,105,110,86,97,114,105,97,116,105,111,110,0,115,101,116,67,111,108,111,114,115,0,103,101,116,67,111,108,111,114,115,0,115,101,116,81,117,97,100,115,0,103,101,116,81,117,97,100,115,0,115,101,116,79,102,102,115,101,116,0,103,101,116,79,102,102,115,101,116,0,115,101,116,82,101,108,97,116,105,118,101,82,111,116,97,116,105,111,110,0,104,97,115,82,101,108,97,116,105,118,101,82,111,116,97,116,105,111,110,0,114,101,115,101,116,0,101,109,105,116,0,105,115,80,97,117,115,101,100,0,105,115,83,116,111,112,112,101,100,0,117,112,100,97,116,101,0,78,111,32,118,97,114,105,97,98,108,101,32,116,111,32,115,101,110,100,46,0,110,117,109,98,101,114,32,111,114,32,98,111,111,108,101,97,110,0,69,114,114,111,114,32,105,110,32,97,114,103,117,109,101,110,116,32,37,100,58,32,69,120,112,101,99,116,101,100,32,116,97,98,108,101,32,115,105,122,101,32,37,100,44,32,103,111,116,32,37,100,46,0,110,117,109,98,101,114,44,32,98,111,111,108,101,97,110,44,32,111,114,32,116,97,98,108,101,0,69,114,114,111,114,32,105,110,32,97,114,103,117,109,101,110,116,115,46,0,109,97,116,114,105,120,32,116,97,98,108,101,0,100,105,109,101,110,115,105,111,110,0,73,110,118,97,108,105,100,32,109,97,116,114,105,120,32,115,105,122,101,58,32,37,100,120,37,100,32,40,111,110,108,121,32,50,120,50,44,32,51,120,51,32,97,110,100,32,52,120,52,32,109,97,116,114,105,99,101,115,32,97,114,101,32,115,117,112,112,111,114,116,101,100,41,46,0,73,110,118,97,108,105,100,32,109,97,116,114,105,120,32,115,105,122,101,32,97,116,32,97,114,103,117,109,101,110,116,32,37,100,58,32,69,120,112,101,99,116,101,100,32,115,105,122,101,32,37,100,120,37,100,44,32,103,111,116,32,37,100,120,37,100,46,0,110,117,109,98,101,114,44,32,98,111,111,108,101,97,110,44,32,116,97,98,108,101,44,32,105,109,97,103,101,44,32,111,114,32,99,97,110,118,97,115,32,101,120,112,101,99,116,101,100,0,85,110,107,110,111,119,110,32,101,120,116,101,114,110,32,118,97,114,105,97,98,108,101,32,116,121,112,101,32,110,97,109,101,46,0,83,104,97,100,101,114,0,103,101,116,87,97,114,110,105,110,103,115,0,115,101,110,100,73,110,116,0,115,101,110,100,66,111,111,108,101,97,110,0,115,101,110,100,70,108,111,97,116,0,115,101,110,100,67,111,108,111,114,0,115,101,110,100,77,97,116,114,105,120,0,115,101,110,100,84,101,120,116,117,114,101,0,103,101,116,69,120,116,101,114,110,86,97,114,105,97,98,108,101,0,85,110,97,98,108,101,32,116,111,32,100,101,116,101,114,109,105,110,101,32,116,101,120,116,117,114,101,32,116,121,112,101,46,0,83,112,114,105,116,101,66,97,116,99,104,0,115,101,116,84,101,120,116,117,114,101,0,103,101,116,84,101,120,116,117,114,101,0,115,101,116,67,111,108,111,114,0,103,101,116,67,111,108,111,114,0,115,101,116,66,117,102,102,101,114,83,105,122,101,0,103,101,116,66,117,102,102,101,114,83,105,122,101,0,97,116,116,97,99,104,65,116,116,114,105,98,117,116,101,0,73,110,118,97,108,105,100,32,97,108,105,103,110,32,109,111,100,101,58,32,37,115,0,84,101,120,116,0,115,101,116,102,0,97,100,100,102,0,115,101,116,70,111,110,116,0,103,101,116,70,111,110,116,0,86,105,100,101,111,0,103,101,116,83,116,114,101,97,109,0,103,101,116,83,111,117,114,99,101,0,95,115,101,116,83,111,117,114,99,101,0,10,45,45,32,68,79,32,78,79,84,32,82,69,77,79,86,69,32,84,72,69,32,65,66,79,86,69,32,76,73,78,69,46,32,73,116,32,105,115,32,117,115,101,100,32,116,111,32,108,111,97,100,32,116,104,105,115,32,102,105,108,101,32,97,115,32,97,32,67,43,43,32,115,116,114,105,110,103,46,10,45,45,32,84,104,101,114,101,32,105,115,32,97,32,109,97,116,99,104,105,110,103,32,100,101,108,105,109,105,116,101,114,32,97,116,32,116,104,101,32,98,111,116,116,111,109,32,111,102,32,116,104,101,32,102,105,108,101,46,10,10,45,45,91,91,10,67,111,112,121,114,105,103,104,116,32,40,99,41,32,50,48,48,54,45,50,48,49,54,32,76,79,86,69,32,68,101,118,101,108,111,112,109,101,110,116,32,84,101,97,109,10,10,84,104,105,115,32,115,111,102,116,119,97,114,101,32,105,115,32,112,114,111,118,105,100,101,100,32,39,97,115,45,105,115,39,44,32,119,105,116,104,111,117,116,32,97,110,121,32,101,120,112,114,101,115,115,32,111,114,32,105,109,112,108,105,101,100,10,119,97,114,114,97,110,116,121,46,32,32,73,110,32,110,111,32,101,118,101,110,116,32,119,105,108,108,32,116,104,101,32,97,117,116,104,111,114,115,32,98,101,32,104,101,108,100,32,108,105,97,98,108,101,32,102,111,114,32,97,110,121,32,100,97,109,97,103,101,115,10,97,114,105,115,105,110,103,32,102,114,111,109,32,116,104,101,32,117,115,101,32,111,102,32,116,104,105,115,32,115,111,102,116,119,97,114,101,46,10,10,80,101,114,109,105,115,115,105,111,110,32,105,115,32,103,114,97,110,116,101,100,32,116,111,32,97,110,121,111,110,101,32,116,111,32,117,115,101,32,116,104,105,115,32,115,111,102,116,119,97,114,101,32,102,111,114,32,97,110,121,32,112,117,114,112,111,115,101,44,10,105,110,99,108,117,100,105,110,103,32,99,111,109,109,101,114,99,105,97,108,32,97,112,112,108,105,99,97,116,105,111,110,115,44,32,97,110,100,32,116,111,32,97,108,116,101,114,32,105,116,32,97,110,100,32,114,101,100,105,115,116,114,105,98,117,116,101,32,105,116,10,102,114,101,101,108,121,44,32,115,117,98,106,101,99,116,32,116,111,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,114,101,115,116,114,105,99,116,105,111,110,115,58,10,10,49,46,32,84,104,101,32,111,114,105,103,105,110,32,111,102,32,116,104,105,115,32,115,111,102,116,119,97,114,101,32,109,117,115,116,32,110,111,116,32,98,101,32,109,105,115,114,101,112,114,101,115,101,110,116,101,100,59,32,121,111,117,32,109,117,115,116,32,110,111,116,10,99,108,97,105,109,32,116,104,97,116,32,121,111,117,32,119,114,111,116,101,32,116,104,101,32,111,114,105,103,105,110,97,108,32,115,111,102,116,119,97,114,101,46,32,73,102,32,121,111,117,32,117,115,101,32,116,104,105,115,32,115,111,102,116,119,97,114,101,10,105,110,32,97,32,112,114,111,100,117,99,116,44,32,97,110,32,97,99,107,110,111,119,108,101,100,103,109,101,110,116,32,105,110,32,116,104,101,32,112,114,111,100,117,99,116,32,100,111,99,117,109,101,110,116,97,116,105,111,110,32,119,111,117,108,100,32,98,101,10,97,112,112,114,101,99,105,97,116,101,100,32,98,117,116,32,105,115,32,110,111,116,32,114,101,113,117,105,114,101,100,46,10,50,46,32,65,108,116,101,114,101,100,32,115,111,117,114,99,101,32,118,101,114,115,105,111,110,115,32,109,117,115,116,32,98,101,32,112,108,97,105,110,108,121,32,109,97,114,107,101,100,32,97,115,32,115,117,99,104,44,32,97,110,100,32,109,117,115,116,32,110,111,116,32,98,101,10,109,105,115,114,101,112,114,101,115,101,110,116,101,100,32,97,115,32,98,101,105,110,103,32,116,104,101,32,111,114,105,103,105,110,97,108,32,115,111,102,116,119,97,114,101,46,10,51,46,32,84,104,105,115,32,110,111,116,105,99,101,32,109,97,121,32,110,111,116,32,98,101,32,114,101,109,111,118,101,100,32,111,114,32,97,108,116,101,114,101,100,32,102,114,111,109,32,97,110,121,32,115,111,117,114,99,101,32,100,105,115,116,114,105,98,117,116,105,111,110,46,10,45,45,93,93,10,10,108,111,99,97,108,32,86,105,100,101,111,95,109,116,32,61,32,46,46,46,10,108,111,99,97,108,32,86,105,100,101,111,32,61,32,86,105,100,101,111,95,109,116,46,95,95,105,110,100,101,120,10,10,102,117,110,99,116,105,111,110,32,86,105,100,101,111,58,115,101,116,83,111,117,114,99,101,40,115,111,117,114,99,101,41,10,9,115,101,108,102,58,95,115,101,116,83,111,117,114,99,101,40,115,111,117,114,99,101,41,10,9,115,101,108,102,58,103,101,116,83,116,114,101,97,109,40,41,58,115,101,116,83,121,110,99,40,115,111,117,114,99,101,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,86,105,100,101,111,58,112,108,97,121,40,41,10,9,114,101,116,117,114,110,32,115,101,108,102,58,103,101,116,83,116,114,101,97,109,40,41,58,112,108,97,121,40,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,86,105,100,101,111,58,112,97,117,115,101,40,41,10,9,114,101,116,117,114,110,32,115,101,108,102,58,103,101,116,83,116,114,101,97,109,40,41,58,112,97,117,115,101,40,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,86,105,100,101,111,58,115,101,101,107,40,111,102,102,115,101,116,41,10,9,114,101,116,117,114,110,32,115,101,108,102,58,103,101,116,83,116,114,101,97,109,40,41,58,115,101,101,107,40,111,102,102,115,101,116,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,86,105,100,101,111,58,114,101,119,105,110,100,40,41,10,9,114,101,116,117,114,110,32,115,101,108,102,58,103,101,116,83,116,114,101,97,109,40,41,58,114,101,119,105,110,100,40,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,86,105,100,101,111,58,116,101,108,108,40,41,10,9,114,101,116,117,114,110,32,115,101,108,102,58,103,101,116,83,116,114,101,97,109,40,41,58,116,101,108,108,40,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,86,105,100,101,111,58,105,115,80,108,97,121,105,110,103,40,41,10,9,114,101,116,117,114,110,32,115,101,108,102,58,103,101,116,83,116,114,101,97,109,40,41,58,105,115,80,108,97,121,105,110,103,40,41,10,101,110,100,10,10,45,45,32,68,79,32,78,79,84,32,82,69,77,79,86,69,32,84,72,69,32,78,69,88,84,32,76,73,78,69,46,32,73,116,32,105,115,32,117,115,101,100,32,116,111,32,108,111,97,100,32,116,104,105,115,32,102,105,108,101,32,97,115,32,97,32,67,43,43,32,115,116,114,105,110,103,46,10,45,45,0,86,105,100,101,111,46,108,117,97,0,102,105,108,108,0,112,105,101,0,97,108,112,104,97,0,97,100,100,0,115,117,98,116,114,97,99,116,0,109,117,108,116,105,112,108,121,0,108,105,103,104,116,101,110,0,100,97,114,107,101,110,0,115,99,114,101,101,110,0,114,101,112,108,97,99,101,0,97,108,112,104,97,109,117,108,116,105,112,108,121,0,112,114,101,109,117,108,116,105,112,108,105,101,100,0,114,111,117,103,104,0,109,105,116,101,114,0,98,101,118,101,108,0,105,110,99,114,101,109,101,110,116,0,100,101,99,114,101,109,101,110,116,0,105,110,99,114,101,109,101,110,116,119,114,97,112,0,100,101,99,114,101,109,101,110,116,119,114,97,112,0,105,110,118,101,114,116,0,108,101,115,115,0,108,101,113,117,97,108,0,101,113,117,97,108,0,103,101,113,117,97,108,0,103,114,101,97,116,101,114,0,110,111,116,101,113,117,97,108,0,97,108,119,97,121,115,0,109,117,108,116,105,99,97,110,118,97,115,102,111,114,109,97,116,115,0,112,111,105,110,116,115,105,122,101,0,116,101,120,116,117,114,101,115,105,122,101,0,109,117,108,116,105,99,97,110,118,97,115,0,99,97,110,118,97,115,109,115,97,97,0,116,114,97,110,115,102,111,114,109,0,100,114,97,119,99,97,108,108,115,0,99,97,110,118,97,115,115,119,105,116,99,104,101,115,0,99,97,110,118,97,115,101,115,0,105,109,97,103,101,115,0,102,111,110,116,115,0,116,101,120,116,117,114,101,109,101,109,111,114,121,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,56,71,114,97,112,104,105,99,115,69,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,49,52,80,97,114,116,105,99,108,101,83,121,115,116,101,109,69,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,56,68,114,97,119,97,98,108,101,69,0,73,110,118,97,108,105,100,32,98,117,102,102,101,114,32,115,105,122,101,0,117,110,105,102,111,114,109,0,116,111,112,0,98,111,116,116,111,109,0,73,110,118,97,108,105,100,32,80,97,114,116,105,99,108,101,83,121,115,116,101,109,32,115,105,122,101,46,0,73,110,118,97,108,105,100,32,101,109,105,115,115,105,111,110,32,114,97,116,101,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,52,81,117,97,100,69,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,55,84,101,120,116,117,114,101,69,0,99,108,97,109,112,0,99,108,97,109,112,122,101,114,111,0,109,105,114,114,111,114,101,100,114,101,112,101,97,116,0,81,117,97,100,0,115,101,116,86,105,101,119,112,111,114,116,0,103,101,116,86,105,101,119,112,111,114,116,0,85,110,107,110,111,119,110,32,102,105,108,116,101,114,32,109,111,100,101,46,0,115,101,116,87,114,97,112,0,73,110,118,97,108,105,100,32,119,114,97,112,32,109,111,100,101,58,32,37,115,0,73,110,118,97,108,105,100,32,119,114,97,112,32,109,111,100,101,44,32,37,115,0,85,110,107,110,111,119,110,32,119,114,97,112,32,109,111,100,101,46,0,84,101,120,116,117,114,101,0,78,52,108,111,118,101,53,105,109,97,103,101,49,57,67,111,109,112,114,101,115,115,101,100,73,109,97,103,101,68,97,116,97,69,0,68,88,84,49,0,68,88,84,51,0,68,88,84,53,0,66,67,52,0,66,67,52,115,0,66,67,53,0,66,67,53,115,0,66,67,54,104,0,66,67,54,104,115,0,66,67,55,0,80,86,82,49,114,103,98,50,0,80,86,82,49,114,103,98,52,0,80,86,82,49,114,103,98,97,50,0,80,86,82,49,114,103,98,97,52,0,69,84,67,49,0,69,84,67,50,114,103,98,0,69,84,67,50,114,103,98,97,0,69,84,67,50,114,103,98,97,49,0,69,65,67,114,0,69,65,67,114,115,0,69,65,67,114,103,0,69,65,67,114,103,115,0,65,83,84,67,52,120,52,0,65,83,84,67,53,120,52,0,65,83,84,67,53,120,53,0,65,83,84,67,54,120,53,0,65,83,84,67,54,120,54,0,65,83,84,67,56,120,53,0,65,83,84,67,56,120,54,0,65,83,84,67,56,120,56,0,65,83,84,67,49,48,120,53,0,65,83,84,67,49,48,120,54,0,65,83,84,67,49,48,120,56,0,65,83,84,67,49,48,120,49,48,0,65,83,84,67,49,50,120,49,48,0,65,83,84,67,49,50,120,49,50,0,77,105,112,109,97,112,32,108,101,118,101,108,32,37,100,32,100,111,101,115,32,110,111,116,32,101,120,105,115,116,0,67,111,109,112,114,101,115,115,101,100,73,109,97,103,101,68,97,116,97,0,103,101,116,77,105,112,109,97,112,67,111,117,110,116,0,73,110,118,97,108,105,100,32,105,109,97,103,101,32,115,105,122,101,46,0,84,104,101,32,115,105,122,101,32,111,102,32,116,104,101,32,114,97,119,32,98,121,116,101,32,115,116,114,105,110,103,32,109,117,115,116,32,109,97,116,99,104,32,116,104,101,32,73,109,97,103,101,68,97,116,97,39,115,32,97,99,116,117,97,108,32,115,105,122,101,32,105,110,32,98,121,116,101,115,46,0,110,101,119,67,111,109,112,114,101,115,115,101,100,68,97,116,97,0,105,115,67,111,109,112,114,101,115,115,101,100,0,37,115,98,97,100,32,114,101,116,117,114,110,32,118,97,108,117,101,32,35,37,100,32,40,110,117,109,98,101,114,32,101,120,112,101,99,116,101,100,44,32,103,111,116,32,37,115,41,0,73,110,118,97,108,105,100,32,101,110,99,111,100,101,100,32,105,109,97,103,101,32,102,111,114,109,97,116,32,39,37,115,39,46,0,73,109,97,103,101,46,0,73,109,97,103,101,68,97,116,97,0,103,101,116,68,105,109,101,110,115,105,111,110,115,0,103,101,116,80,105,120,101,108,0,115,101,116,80,105,120,101,108,0,101,110,99,111,100,101,0,95,109,97,112,80,105,120,101,108,85,110,115,97,102,101,0,95,112,101,114,102,111,114,109,65,116,111,109,105,99,0,10,45,45,32,68,79,32,78,79,84,32,82,69,77,79,86,69,32,84,72,69,32,65,66,79,86,69,32,76,73,78,69,46,32,73,116,32,105,115,32,117,115,101,100,32,116,111,32,108,111,97,100,32,116,104,105,115,32,102,105,108,101,32,97,115,32,97,32,67,43,43,32,115,116,114,105,110,103,46,10,45,45,32,84,104,101,114,101,32,105,115,32,97,32,109,97,116,99,104,105,110,103,32,100,101,108,105,109,105,116,101,114,32,97,116,32,116,104,101,32,98,111,116,116,111,109,32,111,102,32,116,104,101,32,102,105,108,101,46,10,10,45,45,91,91,10,67,111,112,121,114,105,103,104,116,32,40,99,41,32,50,48,48,54,45,50,48,49,54,32,76,79,86,69,32,68,101,118,101,108,111,112,109,101,110,116,32,84,101,97,109,10,10,84,104,105,115,32,115,111,102,116,119,97,114,101,32,105,115,32,112,114,111,118,105,100,101,100,32,39,97,115,45,105,115,39,44,32,119,105,116,104,111,117,116,32,97,110,121,32,101,120,112,114,101,115,115,32,111,114,32,105,109,112,108,105,101,100,10,119,97,114,114,97,110,116,121,46,32,32,73,110,32,110,111,32,101,118,101,110,116,32,119,105,108,108,32,116,104,101,32,97,117,116,104,111,114,115,32,98,101,32,104,101,108,100,32,108,105,97,98,108,101,32,102,111,114,32,97,110,121,32,100,97,109,97,103,101,115,10,97,114,105,115,105,110,103,32,102,114,111,109,32,116,104,101,32,117,115,101,32,111,102,32,116,104,105,115,32,115,111,102,116,119,97,114,101,46,10,10,80,101,114,109,105,115,115,105,111,110,32,105,115,32,103,114,97,110,116,101,100,32,116,111,32,97,110,121,111,110,101,32,116,111,32,117,115,101,32,116,104,105,115,32,115,111,102,116,119,97,114,101,32,102,111,114,32,97,110,121,32,112,117,114,112,111,115,101,44,10,105,110,99,108,117,100,105,110,103,32,99,111,109,109,101,114,99,105,97,108,32,97,112,112,108,105,99,97,116,105,111,110,115,44,32,97,110,100,32,116,111,32,97,108,116,101,114,32,105,116,32,97,110,100,32,114,101,100,105,115,116,114,105,98,117,116,101,32,105,116,10,102,114,101,101,108,121,44,32,115,117,98,106,101,99,116,32,116,111,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,114,101,115,116,114,105,99,116,105,111,110,115,58,10,10,49,46,32,84,104,101,32,111,114,105,103,105,110,32,111,102,32,116,104,105,115,32,115,111,102,116,119,97,114,101,32,109,117,115,116,32,110,111,116,32,98,101,32,109,105,115,114,101,112,114,101,115,101,110,116,101,100,59,32,121,111,117,32,109,117,115,116,32,110,111,116,10,99,108,97,105,109,32,116,104,97,116,32,121,111,117,32,119,114,111,116,101,32,116,104,101,32,111,114,105,103,105,110,97,108,32,115,111,102,116,119,97,114,101,46,32,73,102,32,121,111,117,32,117,115,101,32,116,104,105,115,32,115,111,102,116,119,97,114,101,10,105,110,32,97,32,112,114,111,100,117,99,116,44,32,97,110,32,97,99,107,110,111,119,108,101,100,103,109,101,110,116,32,105,110,32,116,104,101,32,112,114,111,100,117,99,116,32,100,111,99,117,109,101,110,116,97,116,105,111,110,32,119,111,117,108,100,32,98,101,10,97,112,112,114,101,99,105,97,116,101,100,32,98,117,116,32,105,115,32,110,111,116,32,114,101,113,117,105,114,101,100,46,10,50,46,32,65,108,116,101,114,101,100,32,115,111,117,114,99,101,32,118,101,114,115,105,111,110,115,32,109,117,115,116,32,98,101,32,112,108,97,105,110,108,121,32,109,97,114,107,101,100,32,97,115,32,115,117,99,104,44,32,97,110,100,32,109,117,115,116,32,110,111,116,32,98,101,10,109,105,115,114,101,112,114,101,115,101,110,116,101,100,32,97,115,32,98,101,105,110,103,32,116,104,101,32,111,114,105,103,105,110,97,108,32,115,111,102,116,119,97,114,101,46,10,51,46,32,84,104,105,115,32,110,111,116,105,99,101,32,109,97,121,32,110,111,116,32,98,101,32,114,101,109,111,118,101,100,32,111,114,32,97,108,116,101,114,101,100,32,102,114,111,109,32,97,110,121,32,115,111,117,114,99,101,32,100,105,115,116,114,105,98,117,116,105,111,110,46,10,45,45,93,93,10,10,108,111,99,97,108,32,73,109,97,103,101,68,97,116,97,95,109,116,44,32,102,102,105,102,117,110,99,115,112,111,105,110,116,101,114,32,61,32,46,46,46,10,108,111,99,97,108,32,73,109,97,103,101,68,97,116,97,32,61,32,73,109,97,103,101,68,97,116,97,95,109,116,46,95,95,105,110,100,101,120,10,10,108,111,99,97,108,32,116,111,110,117,109,98,101,114,44,32,97,115,115,101,114,116,44,32,101,114,114,111,114,32,61,32,116,111,110,117,109,98,101,114,44,32,97,115,115,101,114,116,44,32,101,114,114,111,114,10,108,111,99,97,108,32,116,121,112,101,44,32,112,99,97,108,108,32,61,32,116,121,112,101,44,32,112,99,97,108,108,10,10,108,111,99,97,108,32,102,117,110,99,116,105,111,110,32,105,110,115,105,100,101,40,120,44,32,121,44,32,119,44,32,104,41,10,9,114,101,116,117,114,110,32,120,32,62,61,32,48,32,97,110,100,32,120,32,60,32,119,32,97,110,100,32,121,32,62,61,32,48,32,97,110,100,32,121,32,60,32,104,10,101,110,100,10,10,45,45,32,73,109,112,108,101,109,101,110,116,32,116,104,114,101,97,100,45,115,97,102,101,32,73,109,97,103,101,68,97,116,97,58,109,97,112,80,105,120,101,108,32,114,101,103,97,114,100,108,101,115,115,32,111,102,32,119,104,101,116,104,101,114,32,116,104,101,32,70,70,73,32,105,115,10,45,45,32,117,115,101,100,32,111,114,32,110,111,116,46,10,102,117,110,99,116,105,111,110,32,73,109,97,103,101,68,97,116,97,58,109,97,112,80,105,120,101,108,40,102,117,110,99,44,32,105,120,44,32,105,121,44,32,105,119,44,32,105,104,41,10,9,108,111,99,97,108,32,105,100,119,44,32,105,100,104,32,61,32,115,101,108,102,58,103,101,116,68,105,109,101,110,115,105,111,110,115,40,41,10,10,9,105,120,32,61,32,105,120,32,111,114,32,48,10,9,105,121,32,61,32,105,121,32,111,114,32,48,10,9,105,119,32,61,32,105,119,32,111,114,32,105,100,119,10,9,105,104,32,61,32,105,104,32,111,114,32,105,100,104,10,10,9,105,102,32,116,121,112,101,40,105,120,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,97,114,103,117,109,101,110,116,32,35,50,32,116,111,32,73,109,97,103,101,68,97,116,97,58,109,97,112,80,105,120,101,108,32,40,101,120,112,101,99,116,101,100,32,110,117,109,98,101,114,41,34,44,32,50,41,32,101,110,100,10,9,105,102,32,116,121,112,101,40,105,121,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,97,114,103,117,109,101,110,116,32,35,51,32,116,111,32,73,109,97,103,101,68,97,116,97,58,109,97,112,80,105,120,101,108,32,40,101,120,112,101,99,116,101,100,32,110,117,109,98,101,114,41,34,44,32,50,41,32,101,110,100,10,9,105,102,32,116,121,112,101,40,105,119,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,97,114,103,117,109,101,110,116,32,35,52,32,116,111,32,73,109,97,103,101,68,97,116,97,58,109,97,112,80,105,120,101,108,32,40,101,120,112,101,99,116,101,100,32,110,117,109,98,101,114,41,34,44,32,50,41,32,101,110,100,10,9,105,102,32,116,121,112,101,40,105,104,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,97,114,103,117,109,101,110,116,32,35,53,32,116,111,32,73,109,97,103,101,68,97,116,97,58,109,97,112,80,105,120,101,108,32,40,101,120,112,101,99,116,101,100,32,110,117,109,98,101,114,41,34,44,32,50,41,32,101,110,100,10,10,9,105,102,32,116,121,112,101,40,102,117,110,99,41,32,126,61,32,34,102,117,110,99,116,105,111,110,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,97,114,103,117,109,101,110,116,32,35,49,32,116,111,32,73,109,97,103,101,68,97,116,97,58,109,97,112,80,105,120,101,108,32,40,101,120,112,101,99,116,101,100,32,102,117,110,99,116,105,111,110,41,34,44,32,50,41,32,101,110,100,10,9,105,102,32,110,111,116,32,40,105,110,115,105,100,101,40,105,120,44,32,105,121,44,32,105,100,119,44,32,105,100,104,41,32,97,110,100,32,105,110,115,105,100,101,40,105,120,43,105,119,45,49,44,32,105,121,43,105,104,45,49,44,32,105,100,119,44,32,105,100,104,41,41,32,116,104,101,110,32,101,114,114,111,114,40,34,73,110,118,97,108,105,100,32,114,101,99,116,97,110,103,108,101,32,100,105,109,101,110,115,105,111,110,115,34,44,32,50,41,32,101,110,100,10,10,9,45,45,32,112,101,114,102,111,114,109,65,116,111,109,105,99,32,97,110,100,32,109,97,112,80,105,120,101,108,85,110,115,97,102,101,32,104,97,118,101,32,76,117,97,45,67,32,65,80,73,32,97,110,100,32,70,70,73,32,118,101,114,115,105,111,110,115,46,10,9,115,101,108,102,58,95,112,101,114,102,111,114,109,65,116,111,109,105,99,40,115,101,108,102,46,95,109,97,112,80,105,120,101,108,85,110,115,97,102,101,44,32,115,101,108,102,44,32,102,117,110,99,44,32,105,120,44,32,105,121,44,32,105,119,44,32,105,104,41,10,101,110,100,10,10,10,45,45,32,69,118,101,114,121,116,104,105,110,103,32,98,101,108,111,119,32,116,104,105,115,32,112,111,105,110,116,32,105,115,32,101,102,102,105,99,105,101,110,116,32,70,70,73,32,114,101,112,108,97,99,101,109,101,110,116,115,32,102,111,114,32,101,120,105,115,116,105,110,103,10,45,45,32,73,109,97,103,101,68,97,116,97,32,102,117,110,99,116,105,111,110,97,108,105,116,121,46,10,10,105,102,32,116,121,112,101,40,106,105,116,41,32,126,61,32,34,116,97,98,108,101,34,32,111,114,32,110,111,116,32,106,105,116,46,115,116,97,116,117,115,40,41,32,116,104,101,110,10,9,45,45,32,76,117,97,74,73,84,39,115,32,70,70,73,32,105,115,32,42,109,117,99,104,42,32,115,108,111,119,101,114,32,116,104,97,110,32,76,79,86,69,39,115,32,114,101,103,117,108,97,114,32,109,101,116,104,111,100,115,32,119,104,101,110,32,116,104,101,32,74,73,84,10,9,45,45,32,99,111,109,112,105,108,101,114,32,105,115,32,100,105,115,97,98,108,101,100,46,10,9,114,101,116,117,114,110,10,101,110,100,10,10,108,111,99,97,108,32,115,116,97,116,117,115,44,32,102,102,105,32,61,32,112,99,97,108,108,40,114,101,113,117,105,114,101,44,32,34,102,102,105,34,41,10,105,102,32,110,111,116,32,115,116,97,116,117,115,32,116,104,101,110,32,114,101,116,117,114,110,32,101,110,100,10,10,112,99,97,108,108,40,102,102,105,46,99,100,101,102,44,32,91,91,10,116,121,112,101,100,101,102,32,115,116,114,117,99,116,32,80,114,111,120,121,32,80,114,111,120,121,59,10,10,116,121,112,101,100,101,102,32,115,116,114,117,99,116,32,70,70,73,95,73,109,97,103,101,68,97,116,97,10,123,10,9,118,111,105,100,32,40,42,108,111,99,107,77,117,116,101,120,41,40,80,114,111,120,121,32,42,112,41,59,10,9,118,111,105,100,32,40,42,117,110,108,111,99,107,77,117,116,101,120,41,40,80,114,111,120,121,32,42,112,41,59,10,125,32,70,70,73,95,73,109,97,103,101,68,97,116,97,59,10,10,116,121,112,101,100,101,102,32,115,116,114,117,99,116,32,73,109,97,103,101,68,97,116,97,95,80,105,120,101,108,10,123,10,9,117,105,110,116,56,95,116,32,114,44,32,103,44,32,98,44,32,97,59,10,125,32,73,109,97,103,101,68,97,116,97,95,80,105,120,101,108,59,10,93,93,41,10,10,108,111,99,97,108,32,102,102,105,102,117,110,99,115,32,61,32,102,102,105,46,99,97,115,116,40,34,70,70,73,95,73,109,97,103,101,68,97,116,97,32,42,34,44,32,102,102,105,102,117,110,99,115,112,111,105,110,116,101,114,41,10,10,108,111,99,97,108,32,112,105,120,101,108,112,111,105,110,116,101,114,32,61,32,102,102,105,46,116,121,112,101,111,102,40,34,73,109,97,103,101,68,97,116,97,95,80,105,120,101,108,32,42,34,41,10,10,108,111,99,97,108,32,95,103,101,116,87,105,100,116,104,32,61,32,73,109,97,103,101,68,97,116,97,46,103,101,116,87,105,100,116,104,10,108,111,99,97,108,32,95,103,101,116,72,101,105,103,104,116,32,61,32,73,109,97,103,101,68,97,116,97,46,103,101,116,72,101,105,103,104,116,10,108,111,99,97,108,32,95,103,101,116,68,105,109,101,110,115,105,111,110,115,32,61,32,73,109,97,103,101,68,97,116,97,46,103,101,116,68,105,109,101,110,115,105,111,110,115,10,10,45,45,32,84,97,98,108,101,32,119,104,105,99,104,32,104,111,108,100,115,32,73,109,97,103,101,68,97,116,97,32,111,98,106,101,99,116,115,32,97,115,32,107,101,121,115,44,32,97,110,100,32,105,110,102,111,114,109,97,116,105,111,110,32,97,98,111,117,116,32,116,104,101,32,111,98,106,101,99,116,115,10,45,45,32,97,115,32,118,97,108,117,101,115,46,32,85,115,101,115,32,119,101,97,107,32,107,101,121,115,32,115,111,32,116,104,101,32,73,109,97,103,101,68,97,116,97,32,111,98,106,101,99,116,115,32,99,97,110,32,115,116,105,108,108,32,98,101,32,71,67,39,100,32,112,114,111,112,101,114,108,121,46,10,108,111,99,97,108,32,111,98,106,101,99,116,99,97,99,104,101,32,61,32,115,101,116,109,101,116,97,116,97,98,108,101,40,123,125,44,32,123,10,9,95,95,109,111,100,101,32,61,32,34,107,34,44,10,9,95,95,105,110,100,101,120,32,61,32,102,117,110,99,116,105,111,110,40,115,101,108,102,44,32,105,109,97,103,101,100,97,116,97,41,10,9,9,108,111,99,97,108,32,119,105,100,116,104,44,32,104,101,105,103,104,116,32,61,32,95,103,101,116,68,105,109,101,110,115,105,111,110,115,40,105,109,97,103,101,100,97,116,97,41,10,9,9,108,111,99,97,108,32,112,111,105,110,116,101,114,32,61,32,102,102,105,46,99,97,115,116,40,112,105,120,101,108,112,111,105,110,116,101,114,44,32,105,109,97,103,101,100,97,116,97,58,103,101,116,80,111,105,110,116,101,114,40,41,41,10,10,9,9,108,111,99,97,108,32,112,32,61,32,123,10,9,9,9,119,105,100,116,104,32,61,32,119,105,100,116,104,44,10,9,9,9,104,101,105,103,104,116,32,61,32,104,101,105,103,104,116,44,10,9,9,9,112,111,105,110,116,101,114,32,61,32,112,111,105,110,116,101,114,44,10,9,9,125,10,10,9,9,115,101,108,102,91,105,109,97,103,101,100,97,116,97,93,32,61,32,112,10,9,9,114,101,116,117,114,110,32,112,10,9,101,110,100,44,10,125,41,10,10,10,45,45,32,79,118,101,114,119,114,105,116,101,32,101,120,105,115,116,105,110,103,32,102,117,110,99,116,105,111,110,115,32,119,105,116,104,32,110,101,119,32,70,70,73,32,118,101,114,115,105,111,110,115,46,10,10,102,117,110,99,116,105,111,110,32,73,109,97,103,101,68,97,116,97,58,95,112,101,114,102,111,114,109,65,116,111,109,105,99,40,46,46,46,41,10,9,102,102,105,102,117,110,99,115,46,108,111,99,107,77,117,116,101,120,40,115,101,108,102,41,10,9,108,111,99,97,108,32,115,117,99,99,101,115,115,44,32,101,114,114,32,61,32,112,99,97,108,108,40,46,46,46,41,10,9,102,102,105,102,117,110,99,115,46,117,110,108,111,99,107,77,117,116,101,120,40,115,101,108,102,41,10,10,9,105,102,32,110,111,116,32,115,117,99,99,101,115,115,32,116,104,101,110,10,9,9,101,114,114,111,114,40,101,114,114,44,32,51,41,10,9,101,110,100,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,73,109,97,103,101,68,97,116,97,58,95,109,97,112,80,105,120,101,108,85,110,115,97,102,101,40,102,117,110,99,44,32,105,120,44,32,105,121,44,32,105,119,44,32,105,104,41,10,9,108,111,99,97,108,32,112,32,61,32,111,98,106,101,99,116,99,97,99,104,101,91,115,101,108,102,93,10,9,108,111,99,97,108,32,105,100,119,44,32,105,100,104,32,61,32,112,46,119,105,100,116,104,44,32,112,46,104,101,105,103,104,116,10,10,9,108,111,99,97,108,32,112,105,120,101,108,115,32,61,32,112,46,112,111,105,110,116,101,114,10,10,9,102,111,114,32,121,61,105,121,44,32,105,121,43,105,104,45,49,32,100,111,10,9,9,102,111,114,32,120,61,105,120,44,32,105,120,43,105,119,45,49,32,100,111,10,9,9,9,108,111,99,97,108,32,112,32,61,32,112,105,120,101,108,115,91,121,42,105,100,119,43,120,93,10,9,9,9,108,111,99,97,108,32,114,44,32,103,44,32,98,44,32,97,32,61,32,102,117,110,99,40,120,44,32,121,44,32,116,111,110,117,109,98,101,114,40,112,46,114,41,44,32,116,111,110,117,109,98,101,114,40,112,46,103,41,44,32,116,111,110,117,109,98,101,114,40,112,46,98,41,44,32,116,111,110,117,109,98,101,114,40,112,46,97,41,41,10,9,9,9,112,105,120,101,108,115,91,121,42,105,100,119,43,120,93,46,114,32,61,32,114,10,9,9,9,112,105,120,101,108,115,91,121,42,105,100,119,43,120,93,46,103,32,61,32,103,10,9,9,9,112,105,120,101,108,115,91,121,42,105,100,119,43,120,93,46,98,32,61,32,98,10,9,9,9,112,105,120,101,108,115,91,121,42,105,100,119,43,120,93,46,97,32,61,32,97,32,61,61,32,110,105,108,32,97,110,100,32,50,53,53,32,111,114,32,97,10,9,9,101,110,100,10,9,101,110,100,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,73,109,97,103,101,68,97,116,97,58,103,101,116,80,105,120,101,108,40,120,44,32,121,41,10,9,105,102,32,116,121,112,101,40,120,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,97,114,103,117,109,101,110,116,32,35,49,32,116,111,32,73,109,97,103,101,68,97,116,97,58,103,101,116,80,105,120,101,108,32,40,101,120,112,101,99,116,101,100,32,110,117,109,98,101,114,41,34,44,32,50,41,32,101,110,100,10,9,105,102,32,116,121,112,101,40,121,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,97,114,103,117,109,101,110,116,32,35,50,32,116,111,32,73,109,97,103,101,68,97,116,97,58,103,101,116,80,105,120,101,108,32,40,101,120,112,101,99,116,101,100,32,110,117,109,98,101,114,41,34,44,32,50,41,32,101,110,100,10,10,9,108,111,99,97,108,32,112,32,61,32,111,98,106,101,99,116,99,97,99,104,101,91,115,101,108,102,93,10,9,105,102,32,110,111,116,32,105,110,115,105,100,101,40,120,44,32,121,44,32,112,46,119,105,100,116,104,44,32,112,46,104,101,105,103,104,116,41,32,116,104,101,110,32,101,114,114,111,114,40,34,65,116,116,101,109,112,116,32,116,111,32,103,101,116,32,111,117,116,45,111,102,45,114,97,110,103,101,32,112,105,120,101,108,33,34,44,32,50,41,32,101,110,100,10,10,9,102,102,105,102,117,110,99,115,46,108,111,99,107,77,117,116,101,120,40,115,101,108,102,41,10,9,108,111,99,97,108,32,112,105,120,101,108,32,61,32,112,46,112,111,105,110,116,101,114,91,121,32,42,32,112,46,119,105,100,116,104,32,43,32,120,93,10,9,108,111,99,97,108,32,114,44,32,103,44,32,98,44,32,97,32,61,32,116,111,110,117,109,98,101,114,40,112,105,120,101,108,46,114,41,44,32,116,111,110,117,109,98,101,114,40,112,105,120,101,108,46,103,41,44,32,116,111,110,117,109,98,101,114,40,112,105,120,101,108,46,98,41,44,32,116,111,110,117,109,98,101,114,40,112,105,120,101,108,46,97,41,10,9,102,102,105,102,117,110,99,115,46,117,110,108,111,99,107,77,117,116,101,120,40,115,101,108,102,41,10,10,9,114,101,116,117,114,110,32,114,44,32,103,44,32,98,44,32,97,10,101,110,100,10,10,108,111,99,97,108,32,116,101,109,112,112,105,120,101,108,32,61,32,102,102,105,46,110,101,119,40,34,73,109,97,103,101,68,97,116,97,95,80,105,120,101,108,34,41,10,10,102,117,110,99,116,105,111,110,32,73,109,97,103,101,68,97,116,97,58,115,101,116,80,105,120,101,108,40,120,44,32,121,44,32,114,44,32,103,44,32,98,44,32,97,41,10,9,105,102,32,116,121,112,101,40,120,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,97,114,103,117,109,101,110,116,32,35,49,32,116,111,32,73,109,97,103,101,68,97,116,97,58,115,101,116,80,105,120,101,108,32,40,101,120,112,101,99,116,101,100,32,110,117,109,98,101,114,41,34,44,32,50,41,32,101,110,100,10,9,105,102,32,116,121,112,101,40,121,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,97,114,103,117,109,101,110,116,32,35,50,32,116,111,32,73,109,97,103,101,68,97,116,97,58,115,101,116,80,105,120,101,108,32,40,101,120,112,101,99,116,101,100,32,110,117,109,98,101,114,41,34,44,32,50,41,32,101,110,100,10,10,9,105,102,32,116,121,112,101,40,114,41,32,61,61,32,34,116,97,98,108,101,34,32,116,104,101,110,10,9,9,108,111,99,97,108,32,116,32,61,32,114,10,9,9,114,44,32,103,44,32,98,44,32,97,32,61,32,116,91,49,93,44,32,116,91,50,93,44,32,116,91,51,93,44,32,116,91,52,93,10,9,101,110,100,10,10,9,105,102,32,116,121,112,101,40,114,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,114,101,100,32,99,111,108,111,114,32,99,111,109,112,111,110,101,110,116,32,97,114,103,117,109,101,110,116,32,116,111,32,73,109,97,103,101,68,97,116,97,58,115,101,116,80,105,120,101,108,32,40,101,120,112,101,99,116,101,100,32,110,117,109,98,101,114,41,34,44,32,50,41,32,101,110,100,10,9,105,102,32,116,121,112,101,40,103,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,103,114,101,101,110,32,99,111,108,111,114,32,99,111,109,112,111,110,101,110,116,32,97,114,103,117,109,101,110,116,32,116,111,32,73,109,97,103,101,68,97,116,97,58,115,101,116,80,105,120,101,108,32,40,101,120,112,101,99,116,101,100,32,110,117,109,98,101,114,41,34,44,32,50,41,32,101,110,100,10,9,105,102,32,116,121,112,101,40,98,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,98,108,117,101,32,99,111,108,111,114,32,99,111,109,112,111,110,101,110,116,32,97,114,103,117,109,101,110,116,32,116,111,32,73,109,97,103,101,68,97,116,97,58,115,101,116,80,105,120,101,108,32,40,101,120,112,101,99,116,101,100,32,110,117,109,98,101,114,41,34,44,32,50,41,32,101,110,100,10,9,105,102,32,97,32,126,61,32,110,105,108,32,97,110,100,32,116,121,112,101,40,97,41,32,126,61,32,34,110,117,109,98,101,114,34,32,116,104,101,110,32,101,114,114,111,114,40,34,98,97,100,32,97,108,112,104,97,32,99,111,108,111,114,32,99,111,109,112,111,110,101,110,116,32,97,114,103,117,109,101,110,116,32,116,111,32,73,109,97,103,101,68,97,116,97,58,115,101,116,80,105,120,101,108,32,40,101,120,112,101,99,116,101,100,32,110,117,109,98,101,114,41,34,44,32,50,41,32,101,110,100,10,10,9,108,111,99,97,108,32,112,32,61,32,111,98,106,101,99,116,99,97,99,104,101,91,115,101,108,102,93,10,9,105,102,32,110,111,116,32,105,110,115,105,100,101,40,120,44,32,121,44,32,112,46,119,105,100,116,104,44,32,112,46,104,101,105,103,104,116,41,32,116,104,101,110,32,101,114,114,111,114,40,34,65,116,116,101,109,112,116,32,116,111,32,115,101,116,32,111,117,116,45,111,102,45,114,97,110,103,101,32,112,105,120,101,108,33,34,44,32,50,41,32,101,110,100,10,10,9,116,101,109,112,112,105,120,101,108,46,114,32,61,32,114,10,9,116,101,109,112,112,105,120,101,108,46,103,32,61,32,103,10,9,116,101,109,112,112,105,120,101,108,46,98,32,61,32,98,10,9,116,101,109,112,112,105,120,101,108,46,97,32,61,32,97,32,61,61,32,110,105,108,32,97,110,100,32,50,53,53,32,111,114,32,97,10,10,9,102,102,105,102,117,110,99,115,46,108,111,99,107,77,117,116,101,120,40,115,101,108,102,41,10,9,112,46,112,111,105,110,116,101,114,91,121,32,42,32,112,46,119,105,100,116,104,32,43,32,120,93,32,61,32,116,101,109,112,112,105,120,101,108,10,9,102,102,105,102,117,110,99,115,46,117,110,108,111,99,107,77,117,116,101,120,40,115,101,108,102,41,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,73,109,97,103,101,68,97,116,97,58,103,101,116,87,105,100,116,104,40,41,10,9,114,101,116,117,114,110,32,111,98,106,101,99,116,99,97,99,104,101,91,115,101,108,102,93,46,119,105,100,116,104,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,73,109,97,103,101,68,97,116,97,58,103,101,116,72,101,105,103,104,116,40,41,10,9,114,101,116,117,114,110,32,111,98,106,101,99,116,99,97,99,104,101,91,115,101,108,102,93,46,104,101,105,103,104,116,10,101,110,100,10,10,102,117,110,99,116,105,111,110,32,73,109,97,103,101,68,97,116,97,58,103,101,116,68,105,109,101,110,115,105,111,110,115,40,41,10,9,108,111,99,97,108,32,112,32,61,32,111,98,106,101,99,116,99,97,99,104,101,91,115,101,108,102,93,10,9,114,101,116,117,114,110,32,112,46,119,105,100,116,104,44,32,112,46,104,101,105,103,104,116,10,101,110,100,10,10,45,45,32,68,79,32,78,79,84,32,82,69,77,79,86,69,32,84,72,69,32,78,69,88,84,32,76,73,78,69,46,32,73,116,32,105,115,32,117,115,101,100,32,116,111,32,108,111,97,100,32,116,104,105,115,32,102,105,108,101,32,97,115,32,97,32,67,43,43,32,115,116,114,105,110,103,46,10,45,45,0,73,109,97,103,101,68,97,116,97,46,108,117,97,0,73,110,118,97,108,105,100,32,103,97,109,101,112,97,100,32,97,120,105,115,47,98,117,116,116,111,110,58,32,37,115,0,73,110,118,97,108,105,100,32,106,111,121,115,116,105,99,107,32,105,110,112,117,116,32,116,121,112,101,58,32,37,115,0,73,110,118,97,108,105,100,32,106,111,121,115,116,105,99,107,32,104,97,116,58,32,37,115,0,85,110,107,110,111,119,110,32,106,111,121,115,116,105,99,107,32,105,110,112,117,116,32,116,121,112,101,46,0,85,110,107,110,111,119,110,32,106,111,121,115,116,105,99,107,32,104,97,116,46,0,102,105,108,101,115,121,115,116,101,109,0,105,115,70,105,108,101,0,106,111,121,115,116,105,99,107,0,103,101,116,74,111,121,115,116,105,99,107,115,0,103,101,116,74,111,121,115,116,105,99,107,67,111,117,110,116,0,115,101,116,71,97,109,101,112,97,100,77,97,112,112,105,110,103,0,108,111,97,100,71,97,109,101,112,97,100,77,97,112,112,105,110,103,115,0,115,97,118,101,71,97,109,101,112,97,100,77,97,112,112,105,110,103,115,0,73,110,118,97,108,105,100,32,107,101,121,32,99,111,110,115,116,97,110,116,58,32,37,115,0,85,110,107,110,111,119,110,32,115,99,97,110,99,111,100,101,46,0,73,110,118,97,108,105,100,32,115,99,97,110,99,111,100,101,58,32,37,115,0,85,110,107,110,111,119,110,32,107,101,121,32,99,111,110,115,116,97,110,116,0,107,101,121,98,111,97,114,100,0,115,101,116,75,101,121,82,101,112,101,97,116,0,104,97,115,75,101,121,82,101,112,101,97,116,0,115,101,116,84,101,120,116,73,110,112,117,116,0,104,97,115,84,101,120,116,73,110,112,117,116,0,104,97,115,83,99,114,101,101,110,75,101,121,98,111,97,114,100,0,105,115,83,99,97,110,99,111,100,101,68,111,119,110,0,103,101,116,83,99,97,110,99,111,100,101,70,114,111,109,75,101,121,0,103,101,116,75,101,121,70,114,111,109,83,99,97,110,99,111,100,101,0,78,52,108,111,118,101,52,109,97,116,104,49,52,67,111,109,112,114,101,115,115,101,100,68,97,116,97,69,0,108,122,52,0,122,108,105,98,0,103,122,105,112,0,78,52,108,111,118,101,52,109,97,116,104,49,51,76,90,52,67,111,109,112,114,101,115,115,111,114,69,0,78,52,108,111,118,101,52,109,97,116,104,49,48,67,111,109,112,114,101,115,115,111,114,69,0,73,110,118,97,108,105,100,32,102,111,114,109,97,116,32,40,101,120,112,101,99,116,105,110,103,32,76,90,52,41,0,68,97,116,97,32,105,115,32,116,111,111,32,108,97,114,103,101,32,102,111,114,32,76,90,52,32,99,111,109,112,114,101,115,115,111,114,46,0,67,111,117,108,100,32,110,111,116,32,76,90,52,45,99,111,109,112,114,101,115,115,32,100,97,116,97,46,0,73,110,118,97,108,105,100,32,76,90,52,45,99,111,109,112,114,101,115,115,101,100,32,100,97,116,97,32,115,105,122,101,46,0,67,111,117,108,100,32,110,111,116,32,100,101,99,111,109,112,114,101,115,115,32,76,90,52,45,99,111,109,112,114,101,115,115,101,100,32,100,97,116,97,46,0,78,52,108,111,118,101,52,109,97,116,104,49,52,122,108,105,98,67,111,109,112,114,101,115,115,111,114,69,0,73,110,118,97,108,105,100,32,102,111,114,109,97,116,32,40,101,120,112,101,99,116,105,110,103,32,122,108,105,98,32,111,114,32,103,122,105,112,41,0,67,111,117,108,100,32,110,111,116,32,122,108,105,98,47,103,122,105,112,45,99,111,109,112,114,101,115,115,32,100,97,116,97,46,0,67,111,117,108,100,32,110,111,116,32,100,101,99,111,109,112,114,101,115,115,32,122,108,105,98,47,103,122,105,112,45,99,111,109,112,114,101,115,115,101,100,32,100,97,116,97,46,0,78,52,108,111,118,101,52,109,97,116,104,52,77,97,116,104,69,0,108,111,118,101,46,109,97,116,104,0,78,111,116,32,97,32,112,111,108,121,103,111,110,0,67,97,110,110,111,116,32,116,114,105,97,110,103,117,108,97,116,101,32,112,111,108,121,103,111,110,46,0,73,110,118,97,108,105,100,32,99,111,109,112,114,101,115,115,105,111,110,32,102,111,114,109,97,116,46,0,78,52,108,111,118,101,52,109,97,116,104,49,53,82,97,110,100,111,109,71,101,110,101,114,97,116,111,114,69,0,48,120,0,73,110,118,97,108,105,100,32,114,97,110,100,111,109,32,115,116,97,116,101,58,32,37,115,0,66,101,122,105,101,114,67,117,114,118,101,0,103,101,116,68,101,103,114,101,101,0,103,101,116,68,101,114,105,118,97,116,105,118,101,0,103,101,116,67,111,110,116,114,111,108,80,111,105,110,116,0,115,101,116,67,111,110,116,114,111,108,80,111,105,110,116,0,105,110,115,101,114,116,67,111,110,116,114,111,108,80,111,105,110,116,0,114,101,109,111,118,101,67,111,110,116,114,111,108,80,111,105,110,116,0,103,101,116,67,111,110,116,114,111,108,80,111,105,110,116,67,111,117,110,116,0,116,114,97,110,115,108,97,116,101,0,114,111,116,97,116,101,0,115,99,97,108,101,0,101,118,97,108,117,97,116,101,0,103,101,116,83,101,103,109,101,110,116,0,114,101,110,100,101,114,0,114,101,110,100,101,114,83,101,103,109,101,110,116,0,85,110,107,110,111,119,110,32,99,111,109,112,114,101,115,115,101,100,32,100,97,116,97,32,102,111,114,109,97,116,46,0,67,111,109,112,114,101,115,115,101,100,68,97,116,97,0,78,52,108,111,118,101,53,109,111,117,115,101,51,115,100,108,53,77,111,117,115,101,69,0,78,52,108,111,118,101,53,109,111,117,115,101,53,77,111,117,115,101,69,0,108,111,118,101,46,109,111,117,115,101,46,115,100,108,0,97,114,114,111,119,0,105,98,101,97,109,0,119,97,105,116,97,114,114,111,119,0,115,105,122,101,110,119,115,101,0,115,105,122,101,110,101,115,119,0,115,105,122,101,119,101,0,115,105,122,101,110,115,0,115,105,122,101,97,108,108,0,104,97,110,100,0,115,121,115,116,101,109,0,78,52,108,111,118,101,53,109,111,117,115,101,54,67,117,114,115,111,114,69,0,85,110,107,110,111,119,110,32,99,117,114,115,111,114,32,116,121,112,101,46,0,67,117,114,115,111,114,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,53,74,111,105,110,116,69,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,55,80,104,121,115,105,99,115,69,0,108,111,118,101,46,112,104,121,115,105,99,115,46,98,111,120,50,100,0,78,117,109,98,101,114,32,111,102,32,118,101,114,116,101,120,32,99,111,109,112,111,110,101,110,116,115,32,109,117,115,116,32,98,101,32,97,32,109,117,108,116,105,112,108,101,32,111,102,32,116,119,111,46,0,69,120,112,101,99,116,101,100,32,97,32,109,105,110,105,109,117,109,32,111,102,32,51,32,118,101,114,116,105,99,101,115,44,32,103,111,116,32,37,100,46,0,69,120,112,101,99,116,101,100,32,97,32,109,97,120,105,109,117,109,32,111,102,32,37,100,32,118,101,114,116,105,99,101,115,44,32,103,111,116,32,37,100,46,0,80,104,121,115,105,99,115,32,101,114,114,111,114,58,32,105,110,118,97,108,105,100,32,109,101,116,101,114,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,49,50,80,111,108,121,103,111,110,83,104,97,112,101,69,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,49,52,80,114,105,115,109,97,116,105,99,74,111,105,110,116,69,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,49,49,80,117,108,108,101,121,74,111,105,110,116,69,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,49,51,82,101,118,111,108,117,116,101,74,111,105,110,116,69,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,57,82,111,112,101,74,111,105,110,116,69,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,53,83,104,97,112,101,69,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,57,87,101,108,100,74,111,105,110,116,69,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,49,48,87,104,101,101,108,74,111,105,110,116,69,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,53,87,111,114,108,100,49,51,81,117,101,114,121,67,97,108,108,98,97,99,107,69,0,49,53,98,50,81,117,101,114,121,67,97,108,108,98,97,99,107,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,53,87,111,114,108,100,49,53,82,97,121,67,97,115,116,67,97,108,108,98,97,99,107,69,0,49,55,98,50,82,97,121,67,97,115,116,67,97,108,108,98,97,99,107,0,82,97,121,99,97,115,116,32,99,97,108,108,98,97,99,107,32,100,105,100,110,39,116,32,114,101,116,117,114,110,32,97,32,110,117,109,98,101,114,33,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,53,87,111,114,108,100,69,0,49,55,98,50,67,111,110,116,97,99,116,76,105,115,116,101,110,101,114,0,50,49,98,50,68,101,115,116,114,117,99,116,105,111,110,76,105,115,116,101,110,101,114,0,65,32,98,111,100,121,32,104,97,115,32,101,115,99,97,112,101,100,32,77,101,109,111,105,122,101,114,33,0,65,116,116,101,109,112,116,32,116,111,32,117,115,101,32,100,101,115,116,114,111,121,101,100,32,98,111,100,121,46,0,87,114,111,110,103,32,110,117,109,98,101,114,32,111,102,32,112,97,114,97,109,101,116,101,114,115,46,0,66,111,100,121,0,103,101,116,88,0,103,101,116,89,0,103,101,116,65,110,103,108,101,0,103,101,116,80,111,115,105,116,105,111,110,0,103,101,116,76,105,110,101,97,114,86,101,108,111,99,105,116,121,0,103,101,116,87,111,114,108,100,67,101,110,116,101,114,0,103,101,116,76,111,99,97,108,67,101,110,116,101,114,0,103,101,116,65,110,103,117,108,97,114,86,101,108,111,99,105,116,121,0,103,101,116,77,97,115,115,0,103,101,116,73,110,101,114,116,105,97,0,103,101,116,65,110,103,117,108,97,114,68,97,109,112,105,110,103,0,103,101,116,76,105,110,101,97,114,68,97,109,112,105,110,103,0,103,101,116,71,114,97,118,105,116,121,83,99,97,108,101,0,97,112,112,108,121,76,105,110,101,97,114,73,109,112,117,108,115,101,0,97,112,112,108,121,65,110,103,117,108,97,114,73,109,112,117,108,115,101,0,97,112,112,108,121,84,111,114,113,117,101,0,97,112,112,108,121,70,111,114,99,101,0,115,101,116,88,0,115,101,116,89,0,115,101,116,76,105,110,101,97,114,86,101,108,111,99,105,116,121,0,115,101,116,65,110,103,108,101,0,115,101,116,65,110,103,117,108,97,114,86,101,108,111,99,105,116,121,0,115,101,116,80,111,115,105,116,105,111,110,0,114,101,115,101,116,77,97,115,115,68,97,116,97,0,115,101,116,77,97,115,115,68,97,116,97,0,115,101,116,77,97,115,115,0,115,101,116,73,110,101,114,116,105,97,0,115,101,116,65,110,103,117,108,97,114,68,97,109,112,105,110,103,0,115,101,116,76,105,110,101,97,114,68,97,109,112,105,110,103,0,115,101,116,71,114,97,118,105,116,121,83,99,97,108,101,0,115,101,116,84,121,112,101,0,103,101,116,87,111,114,108,100,80,111,105,110,116,0,103,101,116,87,111,114,108,100,86,101,99,116,111,114,0,103,101,116,87,111,114,108,100,80,111,105,110,116,115,0,103,101,116,76,111,99,97,108,80,111,105,110,116,0,103,101,116,76,111,99,97,108,86,101,99,116,111,114,0,103,101,116,76,105,110,101,97,114,86,101,108,111,99,105,116,121,70,114,111,109,87,111,114,108,100,80,111,105,110,116,0,103,101,116,76,105,110,101,97,114,86,101,108,111,99,105,116,121,70,114,111,109,76,111,99,97,108,80,111,105,110,116,0,105,115,66,117,108,108,101,116,0,115,101,116,66,117,108,108,101,116,0,105,115,65,99,116,105,118,101,0,105,115,65,119,97,107,101,0,115,101,116,83,108,101,101,112,105,110,103,65,108,108,111,119,101,100,0,105,115,83,108,101,101,112,105,110,103,65,108,108,111,119,101,100,0,115,101,116,65,99,116,105,118,101,0,115,101,116,65,119,97,107,101,0,115,101,116,70,105,120,101,100,82,111,116,97,116,105,111,110,0,105,115,70,105,120,101,100,82,111,116,97,116,105,111,110,0,103,101,116,87,111,114,108,100,0,103,101,116,70,105,120,116,117,114,101,76,105,115,116,0,103,101,116,74,111,105,110,116,76,105,115,116,0,103,101,116,67,111,110,116,97,99,116,76,105,115,116,0,84,111,111,32,109,97,110,121,32,114,101,116,117,114,110,32,118,97,108,117,101,115,0,67,104,97,105,110,83,104,97,112,101,0,115,101,116,78,101,120,116,86,101,114,116,101,120,0,115,101,116,80,114,101,118,105,111,117,115,86,101,114,116,101,120,0,103,101,116,67,104,105,108,100,69,100,103,101,0,103,101,116,86,101,114,116,101,120,67,111,117,110,116,0,67,105,114,99,108,101,83,104,97,112,101,0,103,101,116,82,97,100,105,117,115,0,115,101,116,82,97,100,105,117,115,0,103,101,116,80,111,105,110,116,0,115,101,116,80,111,105,110,116,0,65,116,116,101,109,112,116,32,116,111,32,117,115,101,32,100,101,115,116,114,111,121,101,100,32,99,111,110,116,97,99,116,46,0,67,111,110,116,97,99,116,0,103,101,116,80,111,115,105,116,105,111,110,115,0,103,101,116,78,111,114,109,97,108,0,105,115,69,110,97,98,108,101,100,0,105,115,84,111,117,99,104,105,110,103,0,115,101,116,69,110,97,98,108,101,100,0,114,101,115,101,116,70,114,105,99,116,105,111,110,0,114,101,115,101,116,82,101,115,116,105,116,117,116,105,111,110,0,115,101,116,84,97,110,103,101,110,116,83,112,101,101,100,0,103,101,116,84,97,110,103,101,110,116,83,112,101,101,100,0,103,101,116,67,104,105,108,100,114,101,110,0,103,101,116,70,105,120,116,117,114,101,115,0,68,105,115,116,97,110,99,101,74,111,105,110,116,0,115,101,116,76,101,110,103,116,104,0,103,101,116,76,101,110,103,116,104,0,69,100,103,101,83,104,97,112,101,0,103,101,116,80,111,105,110,116,115,0,65,116,116,101,109,112,116,32,116,111,32,117,115,101,32,100,101,115,116,114,111,121,101,100,32,102,105,120,116,117,114,101,46,0,70,105,120,116,117,114,101,0,115,101,116,70,114,105,99,116,105,111,110,0,115,101,116,82,101,115,116,105,116,117,116,105,111,110,0,115,101,116,68,101,110,115,105,116,121,0,115,101,116,83,101,110,115,111,114,0,103,101,116,70,114,105,99,116,105,111,110,0,103,101,116,82,101,115,116,105,116,117,116,105,111,110,0,103,101,116,68,101,110,115,105,116,121,0,103,101,116,66,111,100,121,0,103,101,116,83,104,97,112,101,0,105,115,83,101,110,115,111,114,0,116,101,115,116,80,111,105,110,116,0,114,97,121,67,97,115,116,0,115,101,116,70,105,108,116,101,114,68,97,116,97,0,103,101,116,70,105,108,116,101,114,68,97,116,97,0,115,101,116,67,97,116,101,103,111,114,121,0,103,101,116,67,97,116,101,103,111,114,121,0,115,101,116,77,97,115,107,0,103,101,116,77,97,115,107,0,103,101,116,66,111,117,110,100,105,110,103,66,111,120,0,103,101,116,77,97,115,115,68,97,116,97,0,103,101,116,71,114,111,117,112,73,110,100,101,120,0,115,101,116,71,114,111,117,112,73,110,100,101,120,0,70,114,105,99,116,105,111,110,74,111,105,110,116,0,71,101,97,114,74,111,105,110,116,0,115,101,116,82,97,116,105,111,0,103,101,116,82,97,116,105,111,0,103,101,116,74,111,105,110,116,115,0,103,101,116,84,121,112,101,0,103,101,116,66,111,100,105,101,115,0,103,101,116,65,110,99,104,111,114,115,0,103,101,116,82,101,97,99,116,105,111,110,70,111,114,99,101,0,103,101,116,82,101,97,99,116,105,111,110,84,111,114,113,117,101,0,103,101,116,67,111,108,108,105,100,101,67,111,110,110,101,99,116,101,100,0,115,101,116,85,115,101,114,68,97,116,97,0,103,101,116,85,115,101,114,68,97,116,97,0,100,101,115,116,114,111,121,0,105,115,68,101,115,116,114,111,121,101,100,0,74,111,105,110,116,0,77,111,116,111,114,74,111,105,110,116,0,115,101,116,76,105,110,101,97,114,79,102,102,115,101,116,0,103,101,116,76,105,110,101,97,114,79,102,102,115,101,116,0,115,101,116,65,110,103,117,108,97,114,79,102,102,115,101,116,0,103,101,116,65,110,103,117,108,97,114,79,102,102,115,101,116,0,115,101,116,77,97,120,84,111,114,113,117,101,0,103,101,116,77,97,120,84,111,114,113,117,101,0,115,101,116,67,111,114,114,101,99,116,105,111,110,70,97,99,116,111,114,0,103,101,116,67,111,114,114,101,99,116,105,111,110,70,97,99,116,111,114,0,65,116,116,101,109,112,116,32,116,111,32,117,115,101,32,100,101,115,116,114,111,121,101,100,32,106,111,105,110,116,46,0,77,111,117,115,101,74,111,105,110,116,0,115,101,116,84,97,114,103,101,116,0,103,101,116,84,97,114,103,101,116,0,115,101,116,77,97,120,70,111,114,99,101,0,103,101,116,77,97,120,70,111,114,99,101,0,115,101,116,70,114,101,113,117,101,110,99,121,0,103,101,116,70,114,101,113,117,101,110,99,121,0,115,101,116,68,97,109,112,105,110,103,82,97,116,105,111,0,103,101,116,68,97,109,112,105,110,103,82,97,116,105,111,0,100,105,115,116,97,110,99,101,0,114,101,118,111,108,117,116,101,0,112,114,105,115,109,97,116,105,99,0,109,111,117,115,101,0,112,117,108,108,101,121,0,103,101,97,114,0,102,114,105,99,116,105,111,110,0,119,101,108,100,0,119,104,101,101,108,0,114,111,112,101,0,109,111,116,111,114,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,74,111,105,110,116,69,0,78,52,108,111,118,101,53,115,111,117,110,100,55,108,117,108,108,97,98,121,53,83,111,117,110,100,69,0,108,111,118,101,46,115,111,117,110,100,46,108,117,108,108,97,98,121,0,78,52,108,111,118,101,53,115,111,117,110,100,55,108,117,108,108,97,98,121,49,51,86,111,114,98,105,115,68,101,99,111,100,101,114,69,0,67,111,117,108,100,32,110,111,116,32,114,101,97,100,32,79,103,103,32,98,105,116,115,116,114,101,97,109,0,111,103,103,0,111,103,97,0,111,103,118,0,78,52,108,111,118,101,53,115,111,117,110,100,55,108,117,108,108,97,98,121,49,49,87,97,118,101,68,101,99,111,100,101,114,69,0,67,111,117,108,100,32,110,111,116,32,111,112,101,110,32,87,65,86,69,0,67,111,117,108,100,32,110,111,116,32,114,101,116,114,105,101,118,101,32,87,65,86,69,32,115,116,114,101,97,109,32,105,110,102,111,0,77,117,108,116,105,99,104,97,110,110,101,108,32,97,117,100,105,111,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,67,111,117,108,100,32,110,111,116,32,115,101,116,32,111,117,116,112,117,116,32,102,111,114,109,97,116,0,119,97,118,0,78,52,108,111,118,101,53,115,111,117,110,100,53,83,111,117,110,100,69,0,78,52,108,111,118,101,53,115,111,117,110,100,57,83,111,117,110,100,68,97,116,97,69,0,73,110,118,97,108,105,100,32,98,105,116,32,100,101,112,116,104,58,32,37,100,0,78,111,116,32,101,110,111,117,103,104,32,109,101,109,111,114,121,46,0,73,110,118,97,108,105,100,32,115,97,109,112,108,101,32,99,111,117,110,116,58,32,37,100,0,73,110,118,97,108,105,100,32,115,97,109,112,108,101,32,114,97,116,101,58,32,37,100,0,73,110,118,97,108,105,100,32,99,104,97,110,110,101,108,32,99,111,117,110,116,58,32,37,100,0,68,97,116,97,32,105,115,32,116,111,111,32,98,105,103,33,0,65,116,116,101,109,112,116,32,116,111,32,115,101,116,32,111,117,116,45,111,102,45,114,97,110,103,101,32,115,97,109,112,108,101,33,0,65,116,116,101,109,112,116,32,116,111,32,103,101,116,32,111,117,116,45,111,102,45,114,97,110,103,101,32,115,97,109,112,108,101,33,0,68,101,99,111,100,101,114,0,103,101,116,67,104,97,110,110,101,108,115,0,103,101,116,66,105,116,68,101,112,116,104,0,103,101,116,83,97,109,112,108,101,82,97,116,101,0,103,101,116,68,117,114,97,116,105,111,110,0,78,52,108,111,118,101,54,115,121,115,116,101,109,51,115,100,108,54,83,121,115,116,101,109,69,0,108,111,118,101,46,115,121,115,116,101,109,46,115,100,108,0,78,52,108,111,118,101,54,115,121,115,116,101,109,54,83,121,115,116,101,109,69,0,98,97,116,116,101,114,121,0,110,111,98,97,116,116,101,114,121,0,99,104,97,114,103,105,110,103,0,99,104,97,114,103,101,100,0,87,101,98,0,78,52,108,111,118,101,54,116,104,114,101,97,100,49,50,84,104,114,101,97,100,77,111,100,117,108,101,69,0,108,111,118,101,46,116,104,114,101,97,100,46,115,100,108,0,78,52,108,111,118,101,54,116,104,114,101,97,100,49,48,84,104,114,101,97,100,97,98,108,101,69,0,67,104,97,110,110,101,108,0,112,117,115,104,0,115,117,112,112,108,121,0,112,111,112,0,100,101,109,97,110,100,0,112,101,101,107,0,103,101,116,67,111,117,110,116,0,112,101,114,102,111,114,109,65,116,111,109,105,99,0,98,111,111,108,101,97,110,44,32,110,117,109,98,101,114,44,32,115,116,114,105,110,103,44,32,108,111,118,101,32,116,121,112,101,44,32,111,114,32,102,108,97,116,32,116,97,98,108,101,32,101,120,112,101,99,116,101,100,0,84,104,114,101,97,100,0,103,101,116,69,114,114,111,114,0,105,115,82,117,110,110,105,110,103,0,78,52,108,111,118,101,53,116,105,109,101,114,51,115,100,108,53,84,105,109,101,114,69,0,108,111,118,101,46,116,105,109,101,114,46,115,100,108,0,78,52,108,111,118,101,53,116,105,109,101,114,53,84,105,109,101,114,69,0,78,52,108,111,118,101,53,116,111,117,99,104,51,115,100,108,53,84,111,117,99,104,69,0,78,52,108,111,118,101,53,116,111,117,99,104,53,84,111,117,99,104,69,0,73,110,118,97,108,105,100,32,97,99,116,105,118,101,32,116,111,117,99,104,32,73,68,58,32,37,100,0,78,52,108,111,118,101,53,118,105,100,101,111,54,116,104,101,111,114,97,53,86,105,100,101,111,69,0,78,52,108,111,118,101,53,118,105,100,101,111,53,86,105,100,101,111,69,0,108,111,118,101,46,118,105,100,101,111,46,116,104,101,111,114,97,0,78,52,108,111,118,101,53,118,105,100,101,111,54,116,104,101,111,114,97,54,87,111,114,107,101,114,69,0,86,105,100,101,111,87,111,114,107,101,114,0,78,52,108,111,118,101,53,118,105,100,101,111,54,116,104,101,111,114,97,49,49,86,105,100,101,111,83,116,114,101,97,109,69,0,73,110,118,97,108,105,100,32,115,116,114,101,97,109,0,67,111,117,108,100,32,110,111,116,32,102,105,110,100,32,104,101,97,100,101,114,0,78,52,108,111,118,101,53,118,105,100,101,111,49,49,86,105,100,101,111,83,116,114,101,97,109,57,68,101,108,116,97,83,121,110,99,69,0,78,52,108,111,118,101,53,118,105,100,101,111,49,49,86,105,100,101,111,83,116,114,101,97,109,57,70,114,97,109,101,83,121,110,99,69,0,78,52,108,111,118,101,53,118,105,100,101,111,49,49,86,105,100,101,111,83,116,114,101,97,109,49,48,83,111,117,114,99,101,83,121,110,99,69,0,78,52,108,111,118,101,53,118,105,100,101,111,49,49,86,105,100,101,111,83,116,114,101,97,109,69,0,78,52,108,111,118,101,54,83,116,114,101,97,109,69,0,78,52,108,111,118,101,54,119,105,110,100,111,119,51,115,100,108,54,87,105,110,100,111,119,69,0,108,111,118,101,46,119,105,110,100,111,119,46,115,100,108,0,82,80,73,0,65,110,100,114,111,105,100,0,76,79,86,69,95,71,82,65,80,72,73,67,83,95,85,83,69,95,79,80,69,78,71,76,69,83,0,76,79,86,69,95,71,82,65,80,72,73,67,83,95,68,69,66,85,71,0,120,49,49,0,32,45,32,0,32,40,0,85,110,97,98,108,101,32,116,111,32,99,114,101,97,116,101,32,79,112,101,110,71,76,32,119,105,110,100,111,119,0,84,104,105,115,32,112,114,111,103,114,97,109,32,114,101,113,117,105,114,101,115,32,97,32,103,114,97,112,104,105,99,115,32,99,97,114,100,32,97,110,100,32,118,105,100,101,111,32,100,114,105,118,101,114,115,32,119,104,105,99,104,32,115,117,112,112,111,114,116,32,79,112,101,110,71,76,32,50,46,49,32,111,114,32,79,112,101,110,71,76,32,69,83,32,50,46,0,10,10,68,101,116,101,99,116,101,100,32,79,112,101,110,71,76,32,118,101,114,115,105,111,110,58,10,0,10,10,79,112,101,110,71,76,32,99,111,110,116,101,120,116,32,99,114,101,97,116,105,111,110,32,101,114,114,111,114,58,32,0,10,10,83,68,76,32,119,105,110,100,111,119,32,99,114,101,97,116,105,111,110,32,101,114,114,111,114,58,32,0,117,105,107,105,116,0,119,105,110,114,116,0,73,110,118,97,108,105,100,32,100,105,115,112,108,97,121,32,105,110,100,101,120,58,32,37,100,0,67,111,117,108,100,32,110,111,116,32,105,110,105,116,105,97,108,105,122,101,32,83,68,76,32,118,105,100,101,111,32,115,117,98,115,121,115,116,101,109,32,40,37,115,41,0,102,117,108,108,115,99,114,101,101,110,0,102,117,108,108,115,99,114,101,101,110,116,121,112,101,0,118,115,121,110,99,0,109,115,97,97,0,114,101,115,105,122,97,98,108,101,0,109,105,110,119,105,100,116,104,0,109,105,110,104,101,105,103,104,116,0,98,111,114,100,101,114,108,101,115,115,0,99,101,110,116,101,114,101,100,0,100,105,115,112,108,97,121,0,104,105,103,104,100,112,105,0,114,101,102,114,101,115,104,114,97,116,101,0,101,120,99,108,117,115,105,118,101,0,100,101,115,107,116,111,112,0,119,97,114,110,105,110,103,0,78,52,108,111,118,101,54,119,105,110,100,111,119,54,87,105,110,100,111,119,69,0,48,32,60,61,32,105,110,100,101,120,32,38,38,32,105,110,100,101,120,32,60,32,99,104,97,105,110,45,62,109,95,99,111,117,110,116,0,99,97,99,104,101,45,62,99,111,117,110,116,32,60,61,32,51,0,49,50,98,50,67,104,97,105,110,83,104,97,112,101,0,55,98,50,83,104,97,112,101,0,109,95,118,101,114,116,105,99,101,115,32,61,61,32,78,85,76,76,32,38,38,32,109,95,99,111,117,110,116,32,61,61,32,48,0,99,111,117,110,116,32,62,61,32,50,0,98,50,68,105,115,116,97,110,99,101,83,113,117,97,114,101,100,40,118,101,114,116,105,99,101,115,91,105,45,49,93,44,32,118,101,114,116,105,99,101,115,91,105,93,41,32,62,32,98,50,95,108,105,110,101,97,114,83,108,111,112,32,42,32,98,50,95,108,105,110,101,97,114,83,108,111,112,0,99,104,105,108,100,73,110,100,101,120,32,60,32,109,95,99,111,117,110,116,0,98,50,68,105,115,116,97,110,99,101,83,113,117,97,114,101,100,40,118,49,44,32,118,50,41,32,62,32,98,50,95,108,105,110,101,97,114,83,108,111,112,32,42,32,98,50,95,108,105,110,101,97,114,83,108,111,112,0,48,32,60,61,32,105,110,100,101,120,32,38,38,32,105,110,100,101,120,32,60,32,109,95,99,111,117,110,116,32,45,32,49,0,49,51,98,50,67,105,114,99,108,101,83,104,97,112,101,0,49,49,98,50,69,100,103,101,83,104,97,112,101,0,49,52,98,50,80,111,108,121,103,111,110,83,104,97,112,101,0,48,46,48,102,32,60,61,32,108,111,119,101,114,32,38,38,32,108,111,119,101,114,32,60,61,32,105,110,112,117,116,46,109,97,120,70,114,97,99,116,105,111,110,0,109,95,99,111,117,110,116,32,62,61,32,51,0,97,114,101,97,32,62,32,98,50,95,101,112,115,105,108,111,110,0,51,32,60,61,32,99,111,117,110,116,32,38,38,32,99,111,117,110,116,32,60,61,32,98,50,95,109,97,120,80,111,108,121,103,111,110,86,101,114,116,105,99,101,115,0,99,111,117,110,116,32,62,61,32,51,0,101,100,103,101,46,76,101,110,103,116,104,83,113,117,97,114,101,100,40,41,32,62,32,98,50,95,101,112,115,105,108,111,110,32,42,32,98,50,95,101,112,115,105,108,111,110,0,98,50,95,98,108,111,99,107,83,105,122,101,115,32,60,32,85,67,72,65,82,95,77,65,88,0,106,32,60,32,98,50,95,98,108,111,99,107,83,105,122,101,115,0,48,32,60,32,115,105,122,101,0,48,32,60,61,32,105,110,100,101,120,32,38,38,32,105,110,100,101,120,32,60,32,98,50,95,98,108,111,99,107,83,105,122,101,115,0,98,108,111,99,107,67,111,117,110,116,32,42,32,98,108,111,99,107,83,105,122,101,32,60,61,32,98,50,95,99,104,117,110,107,83,105,122,101,0,66,111,120,50,68,32,97,115,115,101,114,116,105,111,110,32,102,97,105,108,101,100,58,32,37,115,0,73,115,76,111,99,107,101,100,40,41,32,61,61,32,102,97,108,115,101,0,109,95,98,111,100,121,67,111,117,110,116,32,62,32,48,0,109,95,106,111,105,110,116,67,111,117,110,116,32,62,32,48,0,98,45,62,73,115,65,99,116,105,118,101,40,41,32,61,61,32,116,114,117,101,0,109,95,98,111,100,121,67,111,117,110,116,32,60,32,109,95,98,111,100,121,67,97,112,97,99,105,116,121,0,109,95,99,111,110,116,97,99,116,67,111,117,110,116,32,60,32,109,95,99,111,110,116,97,99,116,67,97,112,97,99,105,116,121,0,115,116,97,99,107,67,111,117,110,116,32,60,32,115,116,97,99,107,83,105,122,101,0,109,95,106,111,105,110,116,67,111,117,110,116,32,60,32,109,95,106,111,105,110,116,67,97,112,97,99,105,116,121,0,116,121,112,101,65,32,61,61,32,98,50,95,100,121,110,97,109,105,99,66,111,100,121,32,124,124,32,116,121,112,101,66,32,61,61,32,98,50,95,100,121,110,97,109,105,99,66,111,100,121,0,97,108,112,104,97,48,32,60,32,49,46,48,102,0,114,46,76,101,110,103,116,104,83,113,117,97,114,101,100,40,41,32,62,32,48,46,48,102,0,40,109,95,102,108,97,103,115,32,38,32,101,95,108,111,99,107,101,100,41,32,61,61,32,48,0,49,53,98,50,67,111,110,116,97,99,116,70,105,108,116,101,114,0,48,32,60,61,32,116,121,112,101,49,32,38,38,32,116,121,112,101,49,32,60,32,98,50,83,104,97,112,101,58,58,101,95,116,121,112,101,67,111,117,110,116,0,48,32,60,61,32,116,121,112,101,50,32,38,38,32,116,121,112,101,50,32,60,32,98,50,83,104,97,112,101,58,58,101,95,116,121,112,101,67,111,117,110,116,0,115,95,105,110,105,116,105,97,108,105,122,101,100,32,61,61,32,116,114,117,101,0,48,32,60,61,32,116,121,112,101,65,32,38,38,32,116,121,112,101,66,32,60,32,98,50,83,104,97,112,101,58,58,101,95,116,121,112,101,67,111,117,110,116,0,57,98,50,67,111,110,116,97,99,116,0,50,50,98,50,69,100,103,101,65,110,100,67,105,114,99,108,101,67,111,110,116,97,99,116,0,50,51,98,50,69,100,103,101,65,110,100,80,111,108,121,103,111,110,67,111,110,116,97,99,116,0,109,95,102,105,120,116,117,114,101,65,45,62,71,101,116,84,121,112,101,40,41,32,61,61,32,98,50,83,104,97,112,101,58,58,101,95,101,100,103,101,0,50,53,98,50,80,111,108,121,103,111,110,65,110,100,67,105,114,99,108,101,67,111,110,116,97,99,116,0,49,54,98,50,80,111,108,121,103,111,110,67,111,110,116,97,99,116,0,109,95,102,105,120,116,117,114,101,65,45,62,71,101,116,84,121,112,101,40,41,32,61,61,32,98,50,83,104,97,112,101,58,58,101,95,112,111,108,121,103,111,110,0,55,98,50,74,111,105,110,116,0,47,47,32,68,117,109,112,32,105,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,102,111,114,32,116,104,105,115,32,106,111,105,110,116,32,116,121,112,101,46,10,0,100,101,102,45,62,98,111,100,121,65,32,33,61,32,100,101,102,45,62,98,111,100,121,66,0,49,50,98,50,77,111,116,111,114,74,111,105,110,116,0,32,32,98,50,77,111,116,111,114,74,111,105,110,116,68,101,102,32,106,100,59,10,0,32,32,106,100,46,108,105,110,101,97,114,79,102,102,115,101,116,46,83,101,116,40,37,46,49,53,108,101,102,44,32,37,46,49,53,108,101,102,41,59,10,0,32,32,106,100,46,97,110,103,117,108,97,114,79,102,102,115,101,116,32,61,32,37,46,49,53,108,101,102,59,10,0,32,32,106,100,46,99,111,114,114,101,99,116,105,111,110,70,97,99,116,111,114,32,61,32,37,46,49,53,108,101,102,59,10,0,98,50,73,115,86,97,108,105,100,40,102,97,99,116,111,114,41,32,38,38,32,48,46,48,102,32,60,61,32,102,97,99,116,111,114,32,38,38,32,102,97,99,116,111,114,32,60,61,32,49,46,48,102,0,49,50,98,50,77,111,117,115,101,74,111,105,110,116,0,77,111,117,115,101,32,106,111,105,110,116,32,100,117,109,112,105,110,103,32,105,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,46,10,0,100,32,43,32,104,32,42,32,107,32,62,32,98,50,95,101,112,115,105,108,111,110,0,100,101,102,45,62,116,97,114,103,101,116,46,73,115,86,97,108,105,100,40,41,0,98,50,73,115,86,97,108,105,100,40,100,101,102,45,62,109,97,120,70,111,114,99,101,41,32,38,38,32,100,101,102,45,62,109,97,120,70,111,114,99,101,32,62,61,32,48,46,48,102,0,98,50,73,115,86,97,108,105,100,40,100,101,102,45,62,102,114,101,113,117,101,110,99,121,72,122,41,32,38,38,32,100,101,102,45,62,102,114,101,113,117,101,110,99,121,72,122,32,62,61,32,48,46,48,102,0,98,50,73,115,86,97,108,105,100,40,100,101,102,45,62,100,97,109,112,105,110,103,82,97,116,105,111,41,32,38,38,32,100,101,102,45,62,100,97,109,112,105,110,103,82,97,116,105,111,32,62,61,32,48,46,48,102,0,49,54,98,50,80,114,105,115,109,97,116,105,99,74,111,105,110,116,0,32,32,98,50,80,114,105,115,109,97,116,105,99,74,111,105,110,116,68,101,102,32,106,100,59,10,0,32,32,106,100,46,108,111,119,101,114,84,114,97,110,115,108,97,116,105,111,110,32,61,32,37,46,49,53,108,101,102,59,10,0,32,32,106,100,46,117,112,112,101,114,84,114,97,110,115,108,97,116,105,111,110,32,61,32,37,46,49,53,108,101,102,59,10,0,32,32,106,100,46,109,97,120,77,111,116,111,114,70,111,114,99,101,32,61,32,37,46,49,53,108,101,102,59,10,0,49,51,98,50,80,117,108,108,101,121,74,111,105,110,116,0,32,32,98,50,80,117,108,108,101,121,74,111,105,110,116,68,101,102,32,106,100,59,10,0,32,32,106,100,46,103,114,111,117,110,100,65,110,99,104,111,114,65,46,83,101,116,40,37,46,49,53,108,101,102,44,32,37,46,49,53,108,101,102,41,59,10,0,32,32,106,100,46,103,114,111,117,110,100,65,110,99,104,111,114,66,46,83,101,116,40,37,46,49,53,108,101,102,44,32,37,46,49,53,108,101,102,41,59,10,0,32,32,106,100,46,108,101,110,103,116,104,65,32,61,32,37,46,49,53,108,101,102,59,10,0,32,32,106,100,46,108,101,110,103,116,104,66,32,61,32,37,46,49,53,108,101,102,59,10,0,114,97,116,105,111,32,62,32,98,50,95,101,112,115,105,108,111,110,0,100,101,102,45,62,114,97,116,105,111,32,33,61,32,48,46,48,102,0,49,53,98,50,82,101,118,111,108,117,116,101,74,111,105,110,116,0,32,32,98,50,82,101,118,111,108,117,116,101,74,111,105,110,116,68,101,102,32,106,100,59,10,0,32,32,106,100,46,101,110,97,98,108,101,76,105,109,105,116,32,61,32,98,111,111,108,40,37,100,41,59,10,0,32,32,106,100,46,108,111,119,101,114,65,110,103,108,101,32,61,32,37,46,49,53,108,101,102,59,10,0,32,32,106,100,46,117,112,112,101,114,65,110,103,108,101,32,61,32,37,46,49,53,108,101,102,59,10,0,108,111,119,101,114,32,60,61,32,117,112,112,101,114,0,49,49,98,50,82,111,112,101,74,111,105,110,116,0,32,32,98,50,82,111,112,101,74,111,105,110,116,68,101,102,32,106,100,59,10,0,32,32,106,100,46,109,97,120,76,101,110,103,116,104,32,61,32,37,46,49,53,108,101,102,59,10,0,49,49,98,50,87,101,108,100,74,111,105,110,116,0,32,32,98,50,87,101,108,100,74,111,105,110,116,68,101,102,32,106,100,59,10,0,32,32,106,100,46,114,101,102,101,114,101,110,99,101,65,110,103,108,101,32,61,32,37,46,49,53,108,101,102,59,10,0,49,50,98,50,87,104,101,101,108,74,111,105,110,116,0,32,32,98,50,87,104,101,101,108,74,111,105,110,116,68,101,102,32,106,100,59,10,0,32,32,106,100,46,108,111,99,97,108,65,120,105,115,65,46,83,101,116,40,37,46,49,53,108,101,102,44,32,37,46,49,53,108,101,102,41,59,10,0,32,32,106,100,46,101,110,97,98,108,101,77,111,116,111,114,32,61,32,98,111,111,108,40,37,100,41,59,10,0,32,32,106,100,46,109,111,116,111,114,83,112,101,101,100,32,61,32,37,46,49,53,108,101,102,59,10,0,32,32,106,100,46,109,97,120,77,111,116,111,114,84,111,114,113,117,101,32,61,32,37,46,49,53,108,101,102,59,10,0,82,73,70,70,0,87,65,86,69,0,102,109,116,32,0,100,97,116,97,0,115,111,99,107,101,116,0,115,107,105,112,0,95,95,117,110,108,111,97,100,0,76,117,97,83,111,99,107,101,116,32,50,46,48,46,50,0,117,110,97,98,108,101,32,116,111,32,105,110,105,116,105,97,108,105,122,101,32,108,105,98,114,97,114,121,0,97,117,120,105,108,105,97,114,0,101,120,99,101,112,116,0,98,117,102,102,101,114,0,105,110,101,116,0,109,105,109,101,0,100,111,116,0,98,54,52,0,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,114,115,116,117,118,119,120,121,122,48,49,50,51,52,53,54,55,56,57,43,47,0,61,61,61,61,101,111,108,0,13,10,0,113,112,0,48,49,50,51,52,53,54,55,56,57,65,66,67,68,69,70,0,61,13,10,0,113,112,119,114,112,0,117,110,98,54,52,0,117,110,113,112,0,119,114,112,0,77,73,77,69,32,49,46,48,46,50,0,116,99,112,123,109,97,115,116,101,114,125,0,116,99,112,123,97,110,121,125,0,97,99,99,101,112,116,0,116,99,112,123,115,101,114,118,101,114,125,0,116,99,112,123,99,108,105,101,110,116,125,0,98,105,110,100,0,99,111,110,110,101,99,116,0,103,101,116,115,116,97,116,115,0,115,101,116,115,116,97,116,115,0,108,105,115,116,101,110,0,107,101,101,112,97,108,105,118,101,0,116,99,112,45,110,111,100,101,108,97,121,0,108,105,110,103,101,114,0,115,104,117,116,100,111,119,110,0,98,111,116,104,0,105,110,118,97,108,105,100,32,115,104,117,116,100,111,119,110,32,109,101,116,104,111,100,0,116,99,112,0,103,101,116,116,105,109,101,0,105,110,118,97,108,105,100,32,116,105,109,101,111,117,116,32,109,111,100,101,0,117,100,112,123,99,111,110,110,101,99,116,101,100,125,0,117,100,112,123,97,110,121,125,0,100,105,114,116,121,0,103,101,116,102,100,0,103,101,116,112,101,101,114,110,97,109,101,0,103,101,116,115,111,99,107,110,97,109,101,0,114,101,99,101,105,118,101,0,114,101,102,117,115,101,100,0,114,101,99,101,105,118,101,102,114,111,109,0,117,100,112,123,117,110,99,111,110,110,101,99,116,101,100,125,0,115,101,110,100,0,115,101,110,100,116,111,0,105,110,118,97,108,105,100,32,105,112,32,97,100,100,114,101,115,115,0,115,101,116,102,100,0,115,101,116,111,112,116,105,111,110,0,100,111,110,116,114,111,117,116,101,0,98,114,111,97,100,99,97,115,116,0,114,101,117,115,101,97,100,100,114,0,105,112,45,109,117,108,116,105,99,97,115,116,45,116,116,108,0,105,112,45,109,117,108,116,105,99,97,115,116,45,108,111,111,112,0,105,112,45,97,100,100,45,109,101,109,98,101,114,115,104,105,112,0,105,112,45,100,114,111,112,45,109,101,109,98,101,114,115,104,105,112,0,115,101,116,112,101,101,114,110,97,109,101,0,115,101,116,115,111,99,107,110,97,109,101,0,115,101,116,116,105,109,101,111,117,116,0,115,101,108,101,99,116,123,97,98,108,101,125,0,117,100,112,0,104,111,115,116,32,110,111,116,32,102,111,117,110,100,0,97,100,100,114,101,115,115,32,97,108,114,101,97,100,121,32,105,110,32,117,115,101,0,97,108,114,101,97,100,121,32,99,111,110,110,101,99,116,101,100,0,112,101,114,109,105,115,115,105,111,110,32,100,101,110,105,101,100,0,99,111,110,110,101,99,116,105,111,110,32,114,101,102,117,115,101,100,0,78,52,108,111,118,101,53,115,111,117,110,100,55,108,117,108,108,97,98,121,49,51,77,112,103,49,50,51,68,101,99,111,100,101,114,69,0,67,111,117,108,100,32,110,111,116,32,105,110,105,116,105,97,108,105,122,101,32,109,112,103,49,50,51,46,0,67,111,117,108,100,32,110,111,116,32,99,114,101,97,116,101,32,100,101,99,111,100,101,114,46,0,67,111,117,108,100,32,110,111,116,32,115,101,116,32,100,101,99,111,100,101,114,32,99,97,108,108,98,97,99,107,115,46,0,67,111,117,108,100,32,110,111,116,32,111,112,101,110,32,100,101,99,111,100,101,114,46,0,67,111,117,108,100,32,110,111,116,32,103,101,116,32,115,116,114,101,97,109,32,105,110,102,111,114,109,97,116,105,111,110,46,0,109,112,51,0,78,52,108,111,118,101,57,82,101,102,101,114,101,110,99,101,69,0,108,111,118,101,45,114,101,102,101,114,101,110,99,101,115,0,78,52,108,111,118,101,55,86,97,114,105,97,110,116,69,0,78,52,108,111,118,101,53,97,117,100,105,111,52,110,117,108,108,53,65,117,100,105,111,69,0,78,52,108,111,118,101,53,97,117,100,105,111,53,65,117,100,105,111,69,0,108,111,118,101,46,97,117,100,105,111,46,110,117,108,108,0,78,52,108,111,118,101,53,97,117,100,105,111,52,110,117,108,108,54,83,111,117,114,99,101,69,0,78,52,108,111,118,101,53,97,117,100,105,111,54,111,112,101,110,97,108,53,65,117,100,105,111,49,48,80,111,111,108,84,104,114,101,97,100,69,0,78,52,108,111,118,101,53,97,117,100,105,111,54,111,112,101,110,97,108,53,65,117,100,105,111,69,0,108,111,118,101,46,97,117,100,105,111,46,111,112,101,110,97,108,0,65,117,100,105,111,80,111,111,108,0,67,111,117,108,100,32,110,111,116,32,111,112,101,110,32,100,101,118,105,99,101,46,0,67,111,117,108,100,32,110,111,116,32,99,114,101,97,116,101,32,99,111,110,116,101,120,116,46,0,67,111,117,108,100,32,110,111,116,32,109,97,107,101,32,99,111,110,116,101,120,116,32,99,117,114,114,101,110,116,46,0,67,111,117,108,100,32,110,111,116,32,103,101,110,101,114,97,116,101,32,115,111,117,114,99,101,115,46,0,78,52,108,111,118,101,53,97,117,100,105,111,54,111,112,101,110,97,108,49,54,83,116,97,116,105,99,68,97,116,97,66,117,102,102,101,114,69,0,78,52,108,111,118,101,53,97,117,100,105,111,54,111,112,101,110,97,108,54,83,111,117,114,99,101,69,0,84,104,105,115,32,115,112,97,116,105,97,108,32,97,117,100,105,111,32,102,117,110,99,116,105,111,110,97,108,105,116,121,32,105,115,32,111,110,108,121,32,97,118,97,105,108,97,98,108,101,32,102,111,114,32,109,111,110,111,32,83,111,117,114,99,101,115,46,32,69,110,115,117,114,101,32,116,104,101,32,83,111,117,114,99,101,32,105,115,32,110,111,116,32,109,117,108,116,105,45,99,104,97,110,110,101,108,32,98,101,102,111,114,101,32,99,97,108,108,105,110,103,32,116,104,105,115,32,102,117,110,99,116,105,111,110,46,0,78,52,108,111,118,101,53,97,117,100,105,111,54,111,112,101,110,97,108,50,51,83,112,97,116,105,97,108,83,117,112,112,111,114,116,69,120,99,101,112,116,105,111,110,69,0,37,100,45,99,104,97,110,110,101,108,32,83,111,117,114,99,101,115,32,119,105,116,104,32,37,100,32,98,105,116,115,32,112,101,114,32,115,97,109,112,108,101,32,97,114,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,46,0,78,52,108,111,118,101,53,97,117,100,105,111,54,111,112,101,110,97,108,50,50,73,110,118,97,108,105,100,70,111,114,109,97,116,69,120,99,101,112,116,105,111,110,69,0,105,110,118,101,114,115,101,0,105,110,118,101,114,115,101,99,108,97,109,112,101,100,0,108,105,110,101,97,114,99,108,97,109,112,101,100,0,101,120,112,111,110,101,110,116,0,101,120,112,111,110,101,110,116,99,108,97,109,112,101,100,0,78,52,108,111,118,101,53,97,117,100,105,111,54,83,111,117,114,99,101,69,0,115,101,99,111,110,100,115,0,115,97,109,112,108,101,115,0,78,52,108,111,118,101,53,101,118,101,110,116,51,115,100,108,53,69,118,101,110,116,69,0,108,111,118,101,46,101,118,101,110,116,46,115,100,108,0,107,101,121,112,114,101,115,115,101,100,0,107,101,121,114,101,108,101,97,115,101,100,0,116,101,120,116,105,110,112,117,116,0,116,101,120,116,101,100,105,116,101,100,0,109,111,117,115,101,109,111,118,101,100,0,109,111,117,115,101,112,114,101,115,115,101,100,0,109,111,117,115,101,114,101,108,101,97,115,101,100,0,119,104,101,101,108,109,111,118,101,100,0,108,111,118,101,46,116,111,117,99,104,46,115,100,108,0,116,111,117,99,104,112,114,101,115,115,101,100,0,116,111,117,99,104,114,101,108,101,97,115,101,100,0,116,111,117,99,104,109,111,118,101,100,0,106,111,121,115,116,105,99,107,112,114,101,115,115,101,100,0,106,111,121,115,116,105,99,107,114,101,108,101,97,115,101,100,0,106,111,121,115,116,105,99,107,97,120,105,115,0,106,111,121,115,116,105,99,107,104,97,116,0,103,97,109,101,112,97,100,112,114,101,115,115,101,100,0,103,97,109,101,112,97,100,114,101,108,101,97,115,101,100,0,103,97,109,101,112,97,100,97,120,105,115,0,106,111,121,115,116,105,99,107,97,100,100,101,100,0,106,111,121,115,116,105,99,107,114,101,109,111,118,101,100,0,102,111,99,117,115,0,109,111,117,115,101,102,111,99,117,115,0,118,105,115,105,98,108,101,0,114,101,115,105,122,101,0,100,105,114,101,99,116,111,114,121,100,114,111,112,112,101,100,0,102,105,108,101,100,114,111,112,112,101,100,0,113,117,105,116,0,108,111,119,109,101,109,111,114,121,0,67,111,117,108,100,32,110,111,116,32,105,110,105,116,105,97,108,105,122,101,32,83,68,76,32,101,118,101,110,116,115,32,115,117,98,115,121,115,116,101,109,32,40,37,115,41,0,78,52,108,111,118,101,53,101,118,101,110,116,55,77,101,115,115,97,103,101,69,0,78,52,108,111,118,101,53,101,118,101,110,116,53,69,118,101,110,116,69,0,65,114,103,117,109,101,110,116,32,37,100,32,99,97,110,39,116,32,98,101,32,115,116,111,114,101,100,32,115,97,102,101,108,121,10,69,120,112,101,99,116,101,100,32,98,111,111,108,101,97,110,44,32,110,117,109,98,101,114,44,32,115,116,114,105,110,103,32,111,114,32,117,115,101,114,100,97,116,97,46,0,78,52,108,111,118,101,49,48,102,105,108,101,115,121,115,116,101,109,54,112,104,121,115,102,115,49,48,70,105,108,101,115,121,115,116,101,109,69,0,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,46,112,104,121,115,102,115,0,108,111,118,101,47,0,70,105,108,101,32,100,111,101,115,32,110,111,116,32,101,120,105,115,116,46,0,68,97,116,97,32,99,111,117,108,100,32,110,111,116,32,98,101,32,119,114,105,116,116,101,110,46,0,67,111,117,108,100,32,110,111,116,32,100,101,116,101,114,109,105,110,101,32,102,105,108,101,32,109,111,100,105,102,105,99,97,116,105,111,110,32,100,97,116,101,46,0,63,46,108,117,97,0,63,47,105,110,105,116,46,108,117,97,0,78,52,108,111,118,101,49,48,102,105,108,101,115,121,115,116,101,109,49,49,68,114,111,112,112,101,100,70,105,108,101,69,0,97,98,0,78,52,108,111,118,101,49,48,102,105,108,101,115,121,115,116,101,109,52,70,105,108,101,69,0,67,111,117,108,100,32,110,111,116,32,114,101,97,100,32,102,105,108,101,32,37,115,46,0,78,52,108,111,118,101,49,48,102,105,108,101,115,121,115,116,101,109,49,48,70,105,108,101,115,121,115,116,101,109,69,0,68,114,111,112,112,101,100,70,105,108,101,0,103,101,116,83,105,122,101,0,67,111,117,108,100,32,110,111,116,32,100,101,116,101,114,109,105,110,101,32,102,105,108,101,32,115,105,122,101,46,0,83,105,122,101,32,105,115,32,116,111,111,32,108,97,114,103,101,46,0,73,110,99,111,114,114,101,99,116,32,102,105,108,101,32,111,112,101,110,32,109,111,100,101,58,32,37,115,0,105,115,79,112,101,110,0,115,116,114,105,110,103,32,111,114,32,100,97,116,97,32,101,120,112,101,99,116,101,100,0,105,115,69,79,70,0,116,101,108,108,0,73,110,118,97,108,105,100,32,112,111,115,105,116,105,111,110,46,0,78,117,109,98,101,114,32,105,115,32,116,111,111,32,108,97,114,103,101,46,0,67,111,117,108,100,32,110,111,116,32,111,112,101,110,32,102,105,108,101,46,0,70,105,108,101,32,110,101,101,100,115,32,116,111,32,115,116,97,121,32,105,110,32,114,101,97,100,32,109,111,100,101,46,0,67,111,117,108,100,32,110,111,116,32,114,101,97,100,32,102,114,111,109,32,102,105,108,101,46,0,115,101,116,66,117,102,102,101,114,0,73,110,99,111,114,114,101,99,116,32,102,105,108,101,32,98,117,102,102,101,114,32,109,111,100,101,58,32,37,115,0,103,101,116,66,117,102,102,101,114,0,85,110,107,110,111,119,110,32,102,105,108,101,32,98,117,102,102,101,114,32,109,111,100,101,46,0,103,101,116,77,111,100,101,0,85,110,107,110,111,119,110,32,102,105,108,101,32,109,111,100,101,46,0,70,105,108,101,0,70,105,108,101,68,97,116,97,0,103,101,116,70,105,108,101,110,97,109,101,0,103,101,116,69,120,116,101,110,115,105,111,110,0,78,52,108,111,118,101,52,102,111,110,116,56,102,114,101,101,116,121,112,101,52,70,111,110,116,69,0,108,111,118,101,46,102,111,110,116,46,102,114,101,101,116,121,112,101,0,73,110,118,97,108,105,100,32,102,111,110,116,32,102,105,108,101,58,32,37,115,0,84,114,117,101,84,121,112,101,70,111,110,116,32,76,111,97,100,105,110,103,32,101,114,114,111,114,58,32,70,84,95,73,110,105,116,95,70,114,101,101,84,121,112,101,32,102,97,105,108,101,100,0,78,52,108,111,118,101,52,102,111,110,116,56,102,114,101,101,116,121,112,101,49,56,84,114,117,101,84,121,112,101,82,97,115,116,101,114,105,122,101,114,69,0,78,52,108,111,118,101,52,102,111,110,116,49,56,84,114,117,101,84,121,112,101,82,97,115,116,101,114,105,122,101,114,69,0,84,114,117,101,84,121,112,101,32,70,111,110,116,32,103,108,121,112,104,32,101,114,114,111,114,58,32,70,84,95,76,111,97,100,95,71,108,121,112,104,32,102,97,105,108,101,100,32,40,48,120,37,120,41,0,84,114,117,101,84,121,112,101,32,70,111,110,116,32,103,108,121,112,104,32,101,114,114,111,114,58,32,70,84,95,71,101,116,95,71,108,121,112,104,32,102,97,105,108,101,100,32,40,48,120,37,120,41,0,84,114,117,101,84,121,112,101,32,70,111,110,116,32,103,108,121,112,104,32,101,114,114,111,114,58,32,70,84,95,71,108,121,112,104,95,84,111,95,66,105,116,109,97,112,32,102,97,105,108,101,100,32,40,48,120,37,120,41,0,85,110,107,110,111,119,110,32,84,114,117,101,84,121,112,101,32,103,108,121,112,104,32,112,105,120,101,108,32,109,111,100,101,46,0,73,110,118,97,108,105,100,32,84,114,117,101,84,121,112,101,32,102,111,110,116,32,115,105,122,101,58,32,37,100,0,84,114,117,101,84,121,112,101,32,70,111,110,116,32,108,111,97,100,105,110,103,32,101,114,114,111,114,58,32,70,84,95,78,101,119,95,70,97,99,101,32,102,97,105,108,101,100,58,32,48,120,37,120,32,40,112,114,111,98,108,101,109,32,119,105,116,104,32,102,111,110,116,32,102,105,108,101,63,41,0,84,114,117,101,84,121,112,101,32,70,111,110,116,32,108,111,97,100,105,110,103,32,101,114,114,111,114,58,32,70,84,95,83,101,116,95,80,105,120,101,108,95,83,105,122,101,115,32,102,97,105,108,101,100,58,32,48,120,37,120,32,40,105,110,118,97,108,105,100,32,115,105,122,101,63,41,0,78,52,108,111,118,101,52,102,111,110,116,49,54,66,77,70,111,110,116,82,97,115,116,101,114,105,122,101,114,69,0,105,110,102,111,0,115,105,122,101,0,117,110,105,99,111,100,101,0,99,111,109,109,111,110,0,108,105,110,101,72,101,105,103,104,116,0,98,97,115,101,0,112,97,103,101,0,70,105,108,101,115,121,115,116,101,109,32,109,111,100,117,108,101,32,110,111,116,32,108,111,97,100,101,100,33,0,73,109,97,103,101,32,109,111,100,117,108,101,32,110,111,116,32,108,111,97,100,101,100,33,0,120,111,102,102,115,101,116,0,121,111,102,102,115,101,116,0,120,97,100,118,97,110,99,101,0,102,105,114,115,116,0,115,101,99,111,110,100,0,97,109,111,117,110,116,0,73,110,118,97,108,105,100,32,66,77,70,111,110,116,32,102,105,108,101,32,40,110,111,32,99,104,97,114,97,99,116,101,114,32,100,101,102,105,110,105,116,105,111,110,115,63,41,0,73,110,118,97,108,105,100,32,66,77,70,111,110,116,32,99,104,97,114,97,99,116,101,114,32,105,100,32,40,111,110,108,121,32,117,110,105,99,111,100,101,32,97,110,100,32,65,83,67,73,73,32,97,114,101,32,115,117,112,112,111,114,116,101,100,41,0,73,110,118,97,108,105,100,32,66,77,70,111,110,116,32,99,104,97,114,97,99,116,101,114,32,112,97,103,101,32,105,100,58,32,37,100,0,73,110,118,97,108,105,100,32,99,111,111,114,100,105,110,97,116,101,115,32,102,111,114,32,66,77,70,111,110,116,32,99,104,97,114,97,99,116,101,114,32,37,117,46,0,73,110,118,97,108,105,100,32,119,105,100,116,104,32,37,100,32,102,111,114,32,66,77,70,111,110,116,32,99,104,97,114,97,99,116,101,114,32,37,117,46,0,73,110,118,97,108,105,100,32,104,101,105,103,104,116,32,37,100,32,102,111,114,32,66,77,70,111,110,116,32,99,104,97,114,97,99,116,101,114,32,37,117,46,0,0,1,0,0,0,17,1,0,0,4,0,16,79,83,47,50,180,95,244,99,0,0,235,112,0,0,0,86,80,67,76,84,209,138,94,151,0,0,235,200,0,0,0,54,99,109,97,112,164,195,232,160,0,0,177,108,0,0,3,88,99,118,116,32,255,211,29,57,0,0,30,252,0,0,1,252,102,112,103,109,231,180,241,196,0,0,38,96,0,0,0,139,103,97,115,112,0,7,0,7,0,1,1,72,0,0,0,12,103,108,121,102,12,116,65,207,0,0,38,236,0,0,138,126,104,100,109,120,52,240,33,14,0,0,236,0,0,0,21,72,104,101,97,100,221,132,162,208,0,1,1,84,0,0,0,54,104,104,101,97,16,69,8,111,0,0,235,76,0,0,0,36,104,109,116,120,9,198,142,178,0,0,180,196,0,0,4,48,107,101,114,110,220,82,213,153,0,0,189,160,0,0,45,138,108,111,99,97,243,203,210,61,0,0,187,132,0,0,2,26,109,97,120,112,5,71,6,58,0,0,235,44,0,0,0,32,110,97,109,101,217,188,200,181,0,0,1,28,0,0,29,223,112,111,115,116,180,90,47,187,0,0,184,244,0,0,2,142,112,114,101,112,59,7,241,0,0,0,32,248,0,0,5,104,0,0,0,22,1,14,0,1,0,0,0,0,0,0,0,58,0,0,0,1,0,0,0,0,0,1,0,19,0,58,0,1,0,0,0,0,0,2,0,5,0,95,0,1,0,0,0,0,0,3,0,19,0,58,0,1,0,0,0,0,0,4,0,19,0,58,0,1,0,0,0,0,0,5,0,12,0,100,0,1,0,0,0,0,0,6,0,23,0,77,0,1,0,0,0,0,0,7,0,48,0,173,0,1,0,0,0,0,0,8,0,14,8,108,0,1,0,0,0,0,0,11,0,24,9,131,0,1,0,0,0,0,0,13,9,19,0,112,0,3,0,1,4,9,0,0,0,116,9,155,0,3,0,1,4,9,0,1,0,38,10,15,0,3,0,1,4,9,0,2,0,10,10,89,0,3,0,1,4,9,0,3,0,38,10,15,0,3,0,1,4,9,0,4,0,38,10,15,0,3,0,1,4,9,0,5,0,24,10,99,0,3,0,1,4,9,0,6,0,46,10,53,0,3,0,1,4,9,0,7,0,96,10,245,0,3,0,1,4,9,0,8,0,28,26,115,0,3,0,1,4,9,0,11,0,48,28,161,0,3,0,1,4,9,0,13,18,38,10,123,67,111,112,121,114,105,103,104,116,32,40,99,41,32,50,48,48,51,32,98,121,32,66,105,116,115,116,114,101,97,109,44,32,73,110,99,46,32,65,108,108,32,82,105,103,104,116,115,32,82,101,115,101,114,118,101,100,46,66,105,116,115,116,114,101,97,109,32,86,101,114,97,32,83,97,110,115,66,105,116,115,116,114,101,97,109,86,101,114,97,83,97,110,115,45,82,111,109,97,110,82,101,108,101,97,115,101,32,49,46,49,48,67,111,112,121,114,105,103,104,116,32,40,99,41,32,50,48,48,51,32,98,121,32,66,105,116,115,116,114,101,97,109,44,32,73,110,99,46,13,10,65,108,108,32,82,105,103,104,116,115,32,82,101,115,101,114,118,101,100,46,13,10,66,105,116,115,116,114,101,97,109,32,86,101,114,97,32,105,115,32,97,32,116,114,97,100,101,109,97,114,107,32,111,102,32,66,105,116,115,116,114,101,97,109,44,32,73,110,99,46,13,10,13,10,80,101,114,109,105,115,115,105,111,110,32,105,115,32,104,101,114,101,98,121,32,103,114,97,110,116,101,100,44,32,102,114,101,101,32,111,102,32,99,104,97,114,103,101,44,32,116,111,32,97,110,121,32,112,101,114,115,111,110,32,111,98,116,97,105,110,105,110,103,32,97,32,99,111,112,121,32,111,102,32,116,104,101,32,102,111,110,116,115,32,97,99,99,111,109,112,97,110,121,105,110,103,32,116,104,105,115,32,108,105,99,101,110,115,101,32,40,34,70,111,110,116,115,34,41,32,97,110,100,32,97,115,115,111,99,105,97,116,101,100,32,100,111,99,117,109,101,110,116,97,116,105,111,110,32,102,105,108,101,115,32,40,116,104,101,32,34,70,111,110,116,32,83,111,102,116,119,97,114,101,34,41,44,32,116,111,32,114,101,112,114,111,100,117,99,101,32,97,110,100,32,100,105,115,116,114,105,98,117,116,101,32,116,104,101,32,70,111,110,116,32,83,111,102,116,119,97,114,101,44,32,105,110,99,108,117,100,105,110,103,32,119,105,116,104,111,117,116,32,108,105,109,105,116,97,116,105,111,110,32,116,104,101,32,114,105,103,104,116,115,32,116,111,32,117,115,101,44,32,99,111,112,121,44,32,109,101,114,103,101,44,32,112,117,98,108,105,115,104,44,32,100,105,115,116,114,105,98,117,116,101,44,32,97,110,100,47,111,114,32,115,101,108,108,32,99,111,112,105,101,115,32,111,102,32,116,104,101,32,70,111,110,116,32,83,111,102,116,119,97,114,101,44,32,97,110,100,32,116,111,32,112,101,114,109,105,116,32,112,101,114,115,111,110,115,32,116,111,32,119,104,111,109,32,116,104,101,32,70,111,110,116,32,83,111,102,116,119,97,114,101,32,105,115,32,102,117,114,110,105,115,104,101,100,32,116,111,32,100,111,32,115,111,44,32,115,117,98,106,101,99,116,32,116,111,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,99,111,110,100,105,116,105,111,110,115,58,13,10,13,10,84,104,101,32,97,98,111,118,101,32,99,111,112,121,114,105,103,104,116,32,97,110,100,32,116,114,97,100,101,109,97,114,107,32,110,111,116,105,99,101,115,32,97,110,100,32,116,104,105,115,32,112,101,114,109,105,115,115,105,111,110,32,110,111,116,105,99,101,32,115,104,97,108,108,32,98,101,32,105,110,99,108,117,100,101,100,32,105,110,32,97,108,108,32,99,111,112,105,101,115,32,111,102,32,111,110,101,32,111,114,32,109,111,114,101,32,111,102,32,116,104,101,32,70,111,110,116,32,83,111,102,116,119,97,114,101,32,116,121,112,101,102,97,99,101,115,46,13,10,13,10,84,104,101,32,70,111,110,116,32,83,111,102,116,119,97,114,101,32,109,97,121,32,98,101,32,109,111,100,105,102,105,101,100,44,32,97,108,116,101,114,101,100,44,32,111,114,32,97,100,100,101,100,32,116,111,44,32,97,110,100,32,105,110,32,112,97,114,116,105,99,117,108,97,114,32,116,104,101,32,100,101,115,105,103,110,115,32,111,102,32,103,108,121,112,104,115,32,111,114,32,99,104,97,114,97,99,116,101,114,115,32,105,110,32,116,104,101,32,70,111,110,116,115,32,109,97,121,32,98,101,32,109,111,100,105,102,105,101,100,32,97,110,100,32,97,100,100,105,116,105,111,110,97,108,32,103,108,121,112,104,115,32,111,114,32,99,104,97,114,97,99,116,101,114,115,32,109,97,121,32,98,101,32,97,100,100,101,100,32,116,111,32,116,104,101,32,70,111,110,116,115,44,32,111,110,108,121,32,105,102,32,116,104,101,32,102,111,110,116,115,32,97,114,101,32,114,101,110,97,109,101,100,32,116,111,32,110,97,109,101,115,32,110,111,116,32,99,111,110,116,97,105,110,105,110,103,32,101,105,116,104,101,114,32,116,104,101,32,119,111,114,100,115,32,34,66,105,116,115,116,114,101,97,109,34,32,111,114,32,116,104,101,32,119,111,114,100,32,34,86,101,114,97,34,46,13,10,13,10,84,104,105,115,32,76,105,99,101,110,115,101,32,98,101,99,111,109,101,115,32,110,117,108,108,32,97,110,100,32,118,111,105,100,32,116,111,32,116,104,101,32,101,120,116,101,110,116,32,97,112,112,108,105,99,97,98,108,101,32,116,111,32,70,111,110,116,115,32,111,114,32,70,111,110,116,32,83,111,102,116,119,97,114,101,32,116,104,97,116,32,104,97,115,32,98,101,101,110,32,109,111,100,105,102,105,101,100,32,97,110,100,32,105,115,32,100,105,115,116,114,105,98,117,116,101,100,32,117,110,100,101,114,32,116,104,101,32,34,66,105,116,115,116,114,101,97,109,32,86,101,114,97,34,32,110,97,109,101,115,46,13,10,13,10,84,104,101,32,70,111,110,116,32,83,111,102,116,119,97,114,101,32,109,97,121,32,98,101,32,115,111,108,100,32,97,115,32,112,97,114,116,32,111,102,32,97,32,108,97,114,103,101,114,32,115,111,102,116,119,97,114,101,32,112,97,99,107,97,103,101,32,98,117,116,32,110,111,32,99,111,112,121,32,111,102,32,111,110,101,32,111,114,32,109,111,114,101,32,111,102,32,116,104,101,32,70,111,110,116,32,83,111,102,116,119,97,114,101,32,116,121,112,101,102,97,99,101,115,32,109,97,121,32,98,101,32,115,111,108,100,32,98,121,32,105,116,115,101,108,102,46,13,10,13,10,84,72,69,32,70,79,78,84,32,83,79,70,84,87,65,82,69,32,73,83,32,80,82,79,86,73,68,69,68,32,34,65,83,32,73,83,34,44,32,87,73,84,72,79,85,84,32,87,65,82,82,65,78,84,89,32,79,70,32,65,78,89,32,75,73,78,68,44,32,69,88,80,82,69,83,83,32,79,82,32,73,77,80,76,73,69,68,44,32,73,78,67,76,85,68,73,78,71,32,66,85,84,32,78,79,84,32,76,73,77,73,84,69,68,32,84,79,32,65,78,89,32,87,65,82,82,65,78,84,73,69,83,32,79,70,32,77,69,82,67,72,65,78,84,65,66,73,76,73,84,89,44,32,70,73,84,78,69,83,83,32,70,79,82,32,65,32,80,65,82,84,73,67,85,76,65,82,32,80,85,82,80,79,83,69,32,65,78,68,32,78,79,78,73,78,70,82,73,78,71,69,77,69,78,84,32,79,70,32,67,79,80,89,82,73,71,72,84,44,32,80,65,84,69,78,84,44,32,84,82,65,68,69,77,65,82,75,44,32,79,82,32,79,84,72,69,82,32,82,73,71,72,84,46,32,73,78,32,78,79,32,69,86,69,78,84,32,83,72,65,76,76,32,66,73,84,83,84,82,69,65,77,32,79,82,32,84,72,69,32,71,78,79,77,69,32,70,79,85,78,68,65,84,73,79,78,32,66,69,32,76,73,65,66,76,69,32,70,79,82,32,65,78,89,32,67,76,65,73,77,44,32,68,65,77,65,71,69,83,32,79,82,32,79,84,72,69,82,32,76,73,65,66,73,76,73,84,89,44,32,73,78,67,76,85,68,73,78,71,32,65,78,89,32,71,69,78,69,82,65,76,44,32,83,80,69,67,73,65,76,44,32,73,78,68,73,82,69,67,84,44,32,73,78,67,73,68,69,78,84,65,76,44,32,79,82,32,67,79,78,83,69,81,85,69,78,84,73,65,76,32,68,65,77,65,71,69,83,44,32,87,72,69,84,72,69,82,32,73,78,32,65,78,32,65,67,84,73,79,78,32,79,70,32,67,79,78,84,82,65,67,84,44,32,84,79,82,84,32,79,82,32,79,84,72,69,82,87,73,83,69,44,32,65,82,73,83,73,78,71,32,70,82,79,77,44,32,79,85,84,32,79,70,32,84,72,69,32,85,83,69,32,79,82,32,73,78,65,66,73,76,73,84,89,32,84,79,32,85,83,69,32,84,72,69,32,70,79,78,84,32,83,79,70,84,87,65,82,69,32,79,82,32,70,82,79,77,32,79,84,72,69,82,32,68,69,65,76,73,78,71,83,32,73,78,32,84,72,69,32,70,79,78,84,32,83,79,70,84,87,65,82,69,46,13,10,13,10,69,120,99,101,112,116,32,97,115,32,99,111,110,116,97,105,110,101,100,32,105,110,32,116,104,105,115,32,110,111,116,105,99,101,44,32,116,104,101,32,110,97,109,101,115,32,111,102,32,71,110,111,109,101,44,32,116,104,101,32,71,110,111,109,101,32,70,111,117,110,100,97,116,105,111,110,44,32,97,110,100,32,66,105,116,115,116,114,101,97,109,32,73,110,99,46,44,32,115,104,97,108,108,32,110,111,116,32,98,101,32,117,115,101,100,32,105,110,32,97,100,118,101,114,116,105,115,105,110,103,32,111,114,32,111,116,104,101,114,119,105,115,101,32,116,111,32,112,114,111,109,111,116,101,32,116,104,101,32,115,97,108,101,44,32,117,115,101,32,111,114,32,111,116,104,101,114,32,100,101,97,108,105,110,103,115,32,105,110,32,116,104,105,115,32,70,111,110,116,32,83,111,102,116,119,97,114,101,32,119,105,116,104,111,117,116,32,112,114,105,111,114,32,119,114,105,116,116,101,110,32,97,117,116,104,111,114,105,122,97,116,105,111,110,32,102,114,111,109,32,116,104,101,32,71,110,111,109,101,32,70,111,117,110,100,97,116,105,111,110,32,111,114,32,66,105,116,115,116,114,101,97,109,32,73,110,99,46,44,32,114,101,115,112,101,99,116,105,118,101,108,121,46,32,70,111,114,32,102,117,114,116,104,101,114,32,105,110,102,111,114,109,97,116,105,111,110,44,32,99,111,110,116,97,99,116,58,32,102,111,110,116,115,32,97,116,32,103,110,111,109,101,32,100,111,116,32,111,114,103,46,104,116,116,112,58,47,47,119,119,119,46,98,105,116,115,116,114,101,97,109,46,99,111,109,0,67,0,111,0,112,0,121,0,114,0,105,0,103,0,104,0,116,0,32,0,40,0,99,0,41,0,32,0,50,0,48,0,48,0,51,0,32,0,98,0,121,0,32,0,66,0,105,0,116,0,115,0,116,0,114,0,101,0,97,0,109,0,44,0,32,0,73,0,110,0,99,0,46,0,32,0,65,0,108,0,108,0,32,0,82,0,105,0,103,0,104,0,116,0,115,0,32,0,82,0,101,0,115,0,101,0,114,0,118,0,101,0,100,0,46,0,66,0,105,0,116,0,115,0,116,0,114,0,101,0,97,0,109,0,32,0,86,0,101,0,114,0,97,0,32,0,83,0,97,0,110,0,115,0,66,0,105,0,116,0,115,0,116,0,114,0,101,0,97,0,109,0,86,0,101,0,114,0,97,0,83,0,97,0,110,0,115,0,45,0,82,0,111,0,109,0,97,0,110,0,82,0,101,0,108,0,101,0,97,0,115,0,101,0,32,0,49,0,46,0,49,0,48,0,67,0,111,0,112,0,121,0,114,0,105,0,103,0,104,0,116,0,32,0,40,0,99,0,41,0,32,0,50,0,48,0,48,0,51,0,32,0,98,0,121,0,32,0,66,0,105,0,116,0,115,0,116,0,114,0,101,0,97,0,109,0,44,0,32,0,73,0,110,0,99,0,46,0,13,0,10,0,65,0,108,0,108,0,32,0,82,0,105,0,103,0,104,0,116,0,115,0,32,0,82,0,101,0,115,0,101,0,114,0,118,0,101,0,100,0,46,0,13,0,10,0,66,0,105,0,116,0,115,0,116,0,114,0,101,0,97,0,109,0,32,0,86,0,101,0,114,0,97,0,32,0,105,0,115,0,32,0,97,0,32,0,116,0,114,0,97,0,100,0,101,0,109,0,97,0,114,0,107,0,32,0,111,0,102,0,32,0,66,0,105,0,116,0,115,0,116,0,114,0,101,0,97,0,109,0,44,0,32,0,73,0,110,0,99,0,46,0,13,0,10,0,13,0,10,0,80,0,101,0,114,0,109,0,105,0,115,0,115,0,105,0,111,0,110,0,32,0,105,0,115,0,32,0,104,0,101,0,114,0,101,0,98,0,121,0,32,0,103,0,114,0,97,0,110,0,116,0,101,0,100,0,44,0,32,0,102,0,114,0,101,0,101,0,32,0,111,0,102,0,32,0,99,0,104,0,97,0,114,0,103,0,101,0,44,0,32,0,116,0,111,0,32,0,97,0,110,0,121,0,32,0,112,0,101,0,114,0,115,0,111,0,110,0,32,0,111,0,98,0,116,0,97,0,105,0,110,0,105,0,110,0,103,0,32,0,97,0,32,0,99,0,111,0,112,0,121,0,32,0,111,0,102,0,32,0,116,0,104,0,101,0,32,0,102,0,111,0,110,0,116,0,115,0,32,0,97,0,99,0,99,0,111,0,109,0,112,0,97,0,110,0,121,0,105,0,110,0,103,0,32,0,116,0,104,0,105,0,115,0,32,0,108,0,105,0,99,0,101,0,110,0,115,0,101,0,32,0,40,0,34,0,70,0,111,0,110,0,116,0,115,0,34,0,41,0,32,0,97,0,110,0,100,0,32,0,97,0,115,0,115,0,111,0,99,0,105,0,97,0,116,0,101,0,100,0,32,0,100,0,111,0,99,0,117,0,109,0,101,0,110,0,116,0,97,0,116,0,105,0,111,0,110,0,32,0,102,0,105,0,108,0,101,0,115,0,32,0,40,0,116,0,104,0,101,0,32,0,34,0,70,0,111,0,110,0,116,0,32,0,83,0,111,0,102,0,116,0,119,0,97,0,114,0,101,0,34,0,41,0,44,0,32,0,116,0,111,0,32,0,114,0,101,0,112,0,114,0,111,0,100,0,117,0,99,0,101,0,32,0,97,0,110,0,100,0,32,0,100,0,105,0,115,0,116,0,114,0,105,0,98,0,117,0,116,0,101,0,32,0,116,0,104,0,101,0,32,0,70,0,111,0,110,0,116,0,32,0,83,0,111,0,102,0,116,0,119,0,97,0,114,0,101,0,44,0,32,0,105,0,110,0,99,0,108,0,117,0,100,0,105,0,110,0,103,0,32,0,119,0,105,0,116,0,104,0,111,0,117,0,116,0,32,0,108,0,105,0,109,0,105,0,116,0,97,0,116,0,105,0,111,0,110,0,32,0,116,0,104,0,101,0,32,0,114,0,105,0,103,0,104,0,116,0,115,0,32,0,116,0,111,0,32,0,117,0,115,0,101,0,44,0,32,0,99,0,111,0,112,0,121,0,44,0,32,0,109,0,101,0,114,0,103,0,101,0,44,0,32,0,112,0,117,0,98,0,108,0,105,0,115,0,104,0,44,0,32,0,100,0,105,0,115,0,116,0,114,0,105,0,98,0,117,0,116,0,101,0,44,0,32,0,97,0,110,0,100,0,47,0,111,0,114,0,32,0,115,0,101,0,108,0,108,0,32,0,99,0,111,0,112,0,105,0,101,0,115,0,32,0,111,0,102,0,32,0,116,0,104,0,101,0,32,0,70,0,111,0,110,0,116,0,32,0,83,0,111,0,102,0,116,0,119,0,97,0,114,0,101,0,44,0,32,0,97,0,110,0,100,0,32,0,116,0,111,0,32,0,112,0,101,0,114,0,109,0,105,0,116,0,32,0,112,0,101,0,114,0,115,0,111,0,110,0,115,0,32,0,116,0,111,0,32,0,119,0,104,0,111,0,109,0,32,0,116,0,104,0,101,0,32,0,70,0,111,0,110,0,116,0,32,0,83,0,111,0,102,0,116,0,119,0,97,0,114,0,101,0,32,0,105,0,115,0,32,0,102,0,117,0,114,0,110,0,105,0,115,0,104,0,101,0,100,0,32,0,116,0,111,0,32,0,100,0,111,0,32,0,115,0,111,0,44,0,32,0,115,0,117,0,98,0,106,0,101,0,99,0,116,0,32,0,116,0,111,0,32,0,116,0,104,0,101,0,32,0,102,0,111,0,108,0,108,0,111,0,119,0,105,0,110,0,103,0,32,0,99,0,111,0,110,0,100,0,105,0,116,0,105,0,111,0,110,0,115,0,58,0,13,0,10,0,13,0,10,0,84,0,104,0,101,0,32,0,97,0,98,0,111,0,118,0,101,0,32,0,99,0,111,0,112,0,121,0,114,0,105,0,103,0,104,0,116,0,32,0,97,0,110,0,100,0,32,0,116,0,114,0,97,0,100,0,101,0,109,0,97,0,114,0,107,0,32,0,110,0,111,0,116,0,105,0,99,0,101,0,115,0,32,0,97,0,110,0,100,0,32,0,116,0,104,0,105,0,115,0,32,0,112,0,101,0,114,0,109,0,105,0,115,0,115,0,105,0,111,0,110,0,32,0,110,0,111,0,116,0,105,0,99,0,101,0,32,0,115,0,104,0,97,0,108,0,108,0,32,0,98,0,101,0,32,0,105,0,110,0,99,0,108,0,117,0,100,0,101,0,100,0,32,0,105,0,110,0,32,0,97,0,108,0,108,0,32,0,99,0,111,0,112,0,105,0,101,0,115,0,32,0,111,0,102,0,32,0,111,0,110,0,101,0,32,0,111,0,114,0,32,0,109,0,111,0,114,0,101,0,32,0,111,0,102,0,32,0,116,0,104,0,101,0,32,0,70,0,111,0,110,0,116,0,32,0,83,0,111,0,102,0,116,0,119,0,97,0,114,0,101,0,32,0,116,0,121,0,112,0,101,0,102,0,97,0,99,0,101,0,115,0,46,0,13,0,10,0,13,0,10,0,84,0,104,0,101,0,32,0,70,0,111,0,110,0,116,0,32,0,83,0,111,0,102,0,116,0,119,0,97,0,114,0,101,0,32,0,109,0,97,0,121,0,32,0,98,0,101,0,32,0,109,0,111,0,100,0,105,0,102,0,105,0,101,0,100,0,44,0,32,0,97,0,108,0,116,0,101,0,114,0,101,0,100,0,44,0,32,0,111,0,114,0,32,0,97,0,100,0,100,0,101,0,100,0,32,0,116,0,111,0,44,0,32,0,97,0,110,0,100,0,32,0,105,0,110,0,32,0,112,0,97,0,114,0,116,0,105,0,99,0,117,0,108,0,97,0,114,0,32,0,116,0,104,0,101,0,32,0,100,0,101,0,115,0,105,0,103,0,110,0,115,0,32,0,111,0,102,0,32,0,103,0,108,0,121,0,112,0,104,0,115,0,32,0,111,0,114,0,32,0,99,0,104,0,97,0,114,0,97,0,99,0,116,0,101,0,114,0,115,0,32,0,105,0,110,0,32,0,116,0,104,0,101,0,32,0,70,0,111,0,110,0,116,0,115,0,32,0,109,0,97,0,121,0,32,0,98,0,101,0,32,0,109,0,111,0,100,0,105,0,102,0,105,0,101,0,100,0,32,0,97,0,110,0,100,0,32,0,97,0,100,0,100,0,105,0,116,0,105,0,111,0,110,0,97,0,108,0,32,0,103,0,108,0,121,0,112,0,104,0,115,0,32,0,111,0,114,0,32,0,99,0,104,0,97,0,114,0,97,0,99,0,116,0,101,0,114,0,115,0,32,0,109,0,97,0,121,0,32,0,98,0,101,0,32,0,97,0,100,0,100,0,101,0,100,0,32,0,116,0,111,0,32,0,116,0,104,0,101,0,32,0,70,0,111,0,110,0,116,0,115,0,44,0,32,0,111,0,110,0,108,0,121,0,32,0,105,0,102,0,32,0,116,0,104,0,101,0,32,0,102,0,111,0,110,0,116,0,115,0,32,0,97,0,114,0,101,0,32,0,114,0,101,0,110,0,97,0,109,0,101,0,100,0,32,0,116,0,111,0,32,0,110,0,97,0,109,0,101,0,115,0,32,0,110,0,111,0,116,0,32,0,99,0,111,0,110,0,116,0,97,0,105,0,110,0,105,0,110,0,103,0,32,0,101,0,105,0,116,0,104,0,101,0,114,0,32,0,116,0,104,0,101,0,32,0,119,0,111,0,114,0,100,0,115,0,32,0,34,0,66,0,105,0,116,0,115,0,116,0,114,0,101,0,97,0,109,0,34,0,32,0,111,0,114,0,32,0,116,0,104,0,101,0,32,0,119,0,111,0,114,0,100,0,32,0,34,0,86,0,101,0,114,0,97,0,34,0,46,0,13,0,10,0,13,0,10,0,84,0,104,0,105,0,115,0,32,0,76,0,105,0,99,0,101,0,110,0,115,0,101,0,32,0,98,0,101,0,99,0,111,0,109,0,101,0,115,0,32,0,110,0,117,0,108,0,108,0,32,0,97,0,110,0,100,0,32,0,118,0,111,0,105,0,100,0,32,0,116,0,111,0,32,0,116,0,104,0,101,0,32,0,101,0,120,0,116,0,101,0,110,0,116,0,32,0,97,0,112,0,112,0,108,0,105,0,99,0,97,0,98,0,108,0,101,0,32,0,116,0,111,0,32,0,70,0,111,0,110,0,116,0,115,0,32,0,111,0,114,0,32,0,70,0,111,0,110,0,116,0,32,0,83,0,111,0,102,0,116,0,119,0,97,0,114,0,101,0,32,0,116,0,104,0,97,0,116,0,32,0,104,0,97,0,115,0,32,0,98,0,101,0,101,0,110,0,32,0,109,0,111,0,100,0,105,0,102,0,105,0,101,0,100,0,32,0,97,0,110,0,100,0,32,0,105,0,115,0,32,0,100,0,105,0,115,0,116,0,114,0,105,0,98,0,117,0,116,0,101,0,100,0,32,0,117,0,110,0,100,0,101,0,114,0,32,0,116,0,104,0,101,0,32,0,34,0,66,0,105,0,116,0,115,0,116,0,114,0,101,0,97,0,109,0,32,0,86,0,101,0,114,0,97,0,34,0,32,0,110,0,97,0,109,0,101,0,115,0,46,0,13,0,10,0,13,0,10,0,84,0,104,0,101,0,32,0,70,0,111,0,110,0,116,0,32,0,83,0,111,0,102,0,116,0,119,0,97,0,114,0,101,0,32,0,109,0,97,0,121,0,32,0,98,0,101,0,32,0,115,0,111,0,108,0,100,0,32,0,97,0,115,0,32,0,112,0,97,0,114,0,116,0,32,0,111,0,102,0,32,0,97,0,32,0,108,0,97,0,114,0,103,0,101,0,114,0,32,0,115,0,111,0,102,0,116,0,119,0,97,0,114,0,101,0,32,0,112,0,97,0,99,0,107,0,97,0,103,0,101,0,32,0,98,0,117,0,116,0,32,0,110,0,111,0,32,0,99,0,111,0,112,0,121,0,32,0,111,0,102,0,32,0,111,0,110,0,101,0,32,0,111,0,114,0,32,0,109,0,111,0,114,0,101,0,32,0,111,0,102,0,32,0,116,0,104,0,101,0,32,0,70,0,111,0,110,0,116,0,32,0,83,0,111,0,102,0,116,0,119,0,97,0,114,0,101,0,32,0,116,0,121,0,112,0,101,0,102,0,97,0,99,0,101,0,115,0,32,0,109,0,97,0,121,0,32,0,98,0,101,0,32,0,115,0,111,0,108,0,100,0,32,0,98,0,121,0,32,0,105,0,116,0,115,0,101,0,108,0,102,0,46,0,13,0,10,0,13,0,10,0,84,0,72,0,69,0,32,0,70,0,79,0,78,0,84,0,32,0,83,0,79,0,70,0,84,0,87,0,65,0,82,0,69,0,32,0,73,0,83,0,32,0,80,0,82,0,79,0,86,0,73,0,68,0,69,0,68,0,32,0,34,0,65,0,83,0,32,0,73,0,83,0,34,0,44,0,32,0,87,0,73,0,84,0,72,0,79,0,85,0,84,0,32,0,87,0,65,0,82,0,82,0,65,0,78,0,84,0,89,0,32,0,79,0,70,0,32,0,65,0,78,0,89,0,32,0,75,0,73,0,78,0,68,0,44,0,32,0,69,0,88,0,80,0,82,0,69,0,83,0,83,0,32,0,79,0,82,0,32,0,73,0,77,0,80,0,76,0,73,0,69,0,68,0,44,0,32,0,73,0,78,0,67,0,76,0,85,0,68,0,73,0,78,0,71,0,32,0,66,0,85,0,84,0,32,0,78,0,79,0,84,0,32,0,76,0,73,0,77,0,73,0,84,0,69,0,68,0,32,0,84,0,79,0,32,0,65,0,78,0,89,0,32,0,87,0,65,0,82,0,82,0,65,0,78,0,84,0,73,0,69,0,83,0,32,0,79,0,70,0,32,0,77,0,69,0,82,0,67,0,72,0,65,0,78,0,84,0,65,0,66,0,73,0,76,0,73,0,84,0,89,0,44,0,32,0,70,0,73,0,84,0,78,0,69,0,83,0,83,0,32,0,70,0,79,0,82,0,32,0,65,0,32,0,80,0,65,0,82,0,84,0,73,0,67,0,85,0,76,0,65,0,82,0,32,0,80,0,85,0,82,0,80,0,79,0,83,0,69,0,32,0,65,0,78,0,68,0,32,0,78,0,79,0,78,0,73,0,78,0,70,0,82,0,73,0,78,0,71,0,69,0,77,0,69,0,78,0,84,0,32,0,79,0,70,0,32,0,67,0,79,0,80,0,89,0,82,0,73,0,71,0,72,0,84,0,44,0,32,0,80,0,65,0,84,0,69,0,78,0,84,0,44,0,32,0,84,0,82,0,65,0,68,0,69,0,77,0,65,0,82,0,75,0,44,0,32,0,79,0,82,0,32,0,79,0,84,0,72,0,69,0,82,0,32,0,82,0,73,0,71,0,72,0,84,0,46,0,32,0,73,0,78,0,32,0,78,0,79,0,32,0,69,0,86,0,69,0,78,0,84,0,32,0,83,0,72,0,65,0,76,0,76,0,32,0,66,0,73,0,84,0,83,0,84,0,82,0,69,0,65,0,77,0,32,0,79,0,82,0,32,0,84,0,72,0,69,0,32,0,71,0,78,0,79,0,77,0,69,0,32,0,70,0,79,0,85,0,78,0,68,0,65,0,84,0,73,0,79,0,78,0,32,0,66,0,69,0,32,0,76,0,73,0,65,0,66,0,76,0,69,0,32,0,70,0,79,0,82,0,32,0,65,0,78,0,89,0,32,0,67,0,76,0,65,0,73,0,77,0,44,0,32,0,68,0,65,0,77,0,65,0,71,0,69,0,83,0,32,0,79,0,82,0,32,0,79,0,84,0,72,0,69,0,82,0,32,0,76,0,73,0,65,0,66,0,73,0,76,0,73,0,84,0,89,0,44,0,32,0,73,0,78,0,67,0,76,0,85,0,68,0,73,0,78,0,71,0,32,0,65,0,78,0,89,0,32,0,71,0,69,0,78,0,69,0,82,0,65,0,76,0,44,0,32,0,83,0,80,0,69,0,67,0,73,0,65,0,76,0,44,0,32,0,73,0,78,0,68,0,73,0,82,0,69,0,67,0,84,0,44,0,32,0,73,0,78,0,67,0,73,0,68,0,69,0,78,0,84,0,65,0,76,0,44,0,32,0,79,0,82,0,32,0,67,0,79,0,78,0,83,0,69,0,81,0,85,0,69,0,78,0,84,0,73,0,65,0,76,0,32,0,68,0,65,0,77,0,65,0,71,0,69,0,83,0,44,0,32,0,87,0,72,0,69,0,84,0,72,0,69,0,82,0,32,0,73,0,78,0,32,0,65,0,78,0,32,0,65,0,67,0,84,0,73,0,79,0,78,0,32,0,79,0,70,0,32,0,67,0,79,0,78,0,84,0,82,0,65,0,67,0,84,0,44,0,32,0,84,0,79,0,82,0,84,0,32,0,79,0,82,0,32,0,79,0,84,0,72,0,69,0,82,0,87,0,73,0,83,0,69,0,44,0,32,0,65,0,82,0,73,0,83,0,73,0,78,0,71,0,32,0,70,0,82,0,79,0,77,0,44,0,32,0,79,0,85,0,84,0,32,0,79,0,70,0,32,0,84,0,72,0,69,0,32,0,85,0,83,0,69,0,32,0,79,0,82,0,32,0,73,0,78,0,65,0,66,0,73,0,76,0,73,0,84,0,89,0,32,0,84,0,79,0,32,0,85,0,83,0,69,0,32,0,84,0,72,0,69,0,32,0,70,0,79,0,78,0,84,0,32,0,83,0,79,0,70,0,84,0,87,0,65,0,82,0,69,0,32,0,79,0,82,0,32,0,70,0,82,0,79,0,77,0,32,0,79,0,84,0,72,0,69,0,82,0,32,0,68,0,69,0,65,0,76,0,73,0,78,0,71,0,83,0,32,0,73,0,78,0,32,0,84,0,72,0,69,0,32,0,70,0,79,0,78,0,84,0,32,0,83,0,79,0,70,0,84,0,87,0,65,0,82,0,69,0,46,0,13,0,10,0,13,0,10,0,69,0,120,0,99,0,101,0,112,0,116,0,32,0,97,0,115,0,32,0,99,0,111,0,110,0,116,0,97,0,105,0,110,0,101,0,100,0,32,0,105,0,110,0,32,0,116,0,104,0,105,0,115,0,32,0,110,0,111,0,116,0,105,0,99,0,101,0,44,0,32,0,116,0,104,0,101,0,32,0,110,0,97,0,109,0,101,0,115,0,32,0,111,0,102,0,32,0,71,0,110,0,111,0,109,0,101,0,44,0,32,0,116,0,104,0,101,0,32,0,71,0,110,0,111,0,109,0,101,0,32,0,70,0,111,0,117,0,110,0,100,0,97,0,116,0,105,0,111,0,110,0,44,0,32,0,97,0,110,0,100,0,32,0,66,0,105,0,116,0,115,0,116,0,114,0,101,0,97,0,109,0,32,0,73,0,110,0,99,0,46,0,44,0,32,0,115,0,104,0,97,0,108,0,108,0,32,0,110,0,111,0,116,0,32,0,98,0,101,0,32,0,117,0,115,0,101,0,100,0,32,0,105,0,110,0,32,0,97,0,100,0,118,0,101,0,114,0,116,0,105,0,115,0,105,0,110,0,103,0,32,0,111,0,114,0,32,0,111,0,116,0,104,0,101,0,114,0,119,0,105,0,115,0,101,0,32,0,116,0,111,0,32,0,112,0,114,0,111,0,109,0,111,0,116,0,101,0,32,0,116,0,104,0,101,0,32,0,115,0,97,0,108,0,101,0,44,0,32,0,117,0,115,0,101,0,32,0,111,0,114,0,32,0,111,0,116,0,104,0,101,0,114,0,32,0,100,0,101,0,97,0,108,0,105,0,110,0,103,0,115,0,32,0,105,0,110,0,32,0,116,0,104,0,105,0,115,0,32,0,70,0,111,0,110,0,116,0,32,0,83,0,111,0,102,0,116,0,119,0,97,0,114,0,101,0,32,0,119,0,105,0,116,0,104,0,111,0,117,0,116,0,32,0,112,0,114,0,105,0,111,0,114,0,32,0,119,0,114,0,105,0,116,0,116,0,101,0,110,0,32,0,97,0,117,0,116,0,104,0,111,0,114,0,105,0,122,0,97,0,116,0,105,0,111,0,110,0,32,0,102,0,114,0,111,0,109,0,32,0,116,0,104,0,101,0,32,0,71,0,110,0,111,0,109,0,101,0,32,0,70,0,111,0,117,0,110,0,100,0,97,0,116,0,105,0,111,0,110,0,32,0,111,0,114,0,32,0,66,0,105,0,116,0,115,0,116,0,114,0,101,0,97,0,109,0,32,0,73,0,110,0,99,0,46,0,44,0,32,0,114,0,101,0,115,0,112,0,101,0,99,0,116,0,105,0,118,0,101,0,108,0,121,0,46,0,32,0,70,0,111,0,114,0,32,0,102,0,117,0,114,0,116,0,104,0,101,0,114,0,32,0,105,0,110,0,102,0,111,0,114,0,109,0,97,0,116,0,105,0,111,0,110,0,44,0,32,0,99,0,111,0,110,0,116,0,97,0,99,0,116,0,58,0,32,0,102,0,111,0,110,0,116,0,115,0,32,0,97,0,116,0,32,0,103,0,110,0,111,0,109,0,101,0,32,0,100,0,111,0,116,0,32,0,111,0,114,0,103,0,46,0,104,0,116,0,116,0,112,0,58,0,47,0,47,0,119,0,119,0,119,0,46,0,98,0,105,0,116,0,115,0,116,0,114,0,101,0,97,0,109,0,46,0,99,0,111,0,109,0,1,53,0,184,0,203,0,203,0,193,0,170,0,156,1,166,0,184,0,102,0,0,0,113,0,203,0,160,2,178,0,133,0,117,0,184,0,195,1,203,1,137,2,45,0,203,0,166,0,240,0,211,0,170,0,135,0,203,3,170,4,0,1,74,0,51,0,203,0,0,0,217,5,2,0,244,1,84,0,180,0,156,1,57,1,20,1,57,7,6,4,0,4,78,4,180,4,82,4,184,4,231,4,205,0,55,4,115,4,205,4,96,4,115,1,51,3,162,5,86,5,166,5,86,5,57,3,197,2,18,0,201,0,31,0,184,1,223,0,115,0,186,3,233,3,51,3,188,4,68,4,14,0,223,3,205,3,170,0,229,3,170,4,4,0,0,0,203,0,143,0,164,0,123,0,184,0,20,1,111,0,127,2,123,2,82,0,143,0,199,5,205,0,154,0,154,0,111,0,203,0,205,1,158,1,211,0,240,0,186,1,131,0,213,0,152,3,4,2,72,0,158,1,213,0,193,0,203,0,246,0,131,3,84,2,127,0,0,3,51,2,102,0,211,0,199,0,164,0,205,0,143,0,154,0,115,4,0,5,213,1,10,0,254,2,43,0,164,0,180,0,156,0,0,0,98,0,156,0,0,0,29,3,45,5,213,5,213,5,213,5,240,0,127,0,123,0,84,0,164,6,184,6,20,7,35,1,211,0,184,0,203,0,166,1,195,1,236,6,147,0,160,0,211,3,92,3,113,3,219,1,133,4,35,4,168,4,72,0,143,1,57,1,20,1,57,3,96,0,143,5,213,1,154,6,20,7,35,6,102,1,121,4,96,4,96,4,96,4,123,0,156,0,0,2,119,4,96,1,170,0,233,4,96,7,98,0,123,0,197,0,127,2,123,0,0,0,180,2,82,5,205,0,102,0,188,0,102,0,119,6,16,0,205,1,59,1,133,3,137,0,143,0,123,0,0,0,29,0,205,7,74,4,47,0,156,0,156,0,0,7,125,0,111,0,0,0,111,3,53,0,106,0,111,0,123,0,174,0,178,0,45,3,150,0,143,2,123,0,246,0,131,3,84,6,55,5,246,0,143,0,156,4,225,2,102,0,143,1,141,2,246,0,205,3,68,0,41,0,102,4,238,0,115,0,0,20,0,184,2,128,64,255,251,254,3,250,20,3,249,37,3,248,50,3,247,150,3,246,14,3,245,254,3,244,254,3,243,37,3,242,14,3,241,150,3,240,37,3,239,138,65,5,239,254,3,238,150,3,237,150,3,236,250,3,235,250,3,234,254,3,233,58,3,232,66,3,231,254,3,230,50,3,229,228,83,5,229,150,3,228,138,65,5,228,83,3,227,226,47,5,227,250,3,226,47,3,225,254,3,224,254,3,223,50,3,222,20,3,221,150,3,220,254,3,219,18,3,218,125,3,217,187,3,216,254,3,214,138,65,5,214,125,3,213,212,71,5,213,125,3,212,71,3,211,210,27,5,211,254,3,210,27,3,209,254,3,208,254,3,207,254,3,206,254,3,205,150,3,204,203,30,5,204,254,3,203,30,3,202,50,3,201,254,3,198,133,17,5,198,28,3,197,22,3,196,254,3,195,254,3,194,254,3,193,254,3,192,254,3,191,254,3,190,254,3,189,254,3,188,254,3,187,254,3,186,17,3,185,134,37,5,185,254,3,184,183,187,5,184,254,3,183,182,93,5,183,187,3,183,128,4,182,181,37,5,182,93,64,255,3,182,64,4,181,37,3,180,254,3,179,150,3,178,254,3,177,254,3,176,254,3,175,254,3,174,100,3,173,14,3,172,171,37,5,172,100,3,171,170,18,5,171,37,3,170,18,3,169,138,65,5,169,250,3,168,254,3,167,254,3,166,254,3,165,18,3,164,254,3,163,162,14,5,163,50,3,162,14,3,161,100,3,160,138,65,5,160,150,3,159,254,3,158,157,12,5,158,254,3,157,12,3,156,155,25,5,156,100,3,155,154,16,5,155,25,3,154,16,3,153,10,3,152,254,3,151,150,13,5,151,254,3,150,13,3,149,138,65,5,149,150,3,148,147,14,5,148,40,3,147,14,3,146,250,3,145,144,187,5,145,254,3,144,143,93,5,144,187,3,144,128,4,143,142,37,5,143,93,3,143,64,4,142,37,3,141,254,3,140,139,46,5,140,254,3,139,46,3,138,134,37,5,138,65,3,137,136,11,5,137,20,3,136,11,3,135,134,37,5,135,100,3,134,133,17,5,134,37,3,133,17,3,132,254,3,131,130,17,5,131,254,3,130,17,3,129,254,3,128,254,3,127,254,3,64,255,126,125,125,5,126,254,3,125,125,3,124,100,3,123,84,21,5,123,37,3,122,254,3,121,254,3,120,14,3,119,12,3,118,10,3,117,254,3,116,250,3,115,250,3,114,250,3,113,250,3,112,254,3,111,254,3,110,254,3,108,33,3,107,254,3,106,17,66,5,106,83,3,105,254,3,104,125,3,103,17,66,5,102,254,3,101,254,3,100,254,3,99,254,3,98,254,3,97,58,3,96,250,3,94,12,3,93,254,3,91,254,3,90,254,3,89,88,10,5,89,250,3,88,10,3,87,22,25,5,87,50,3,86,254,3,85,84,21,5,85,66,3,84,21,3,83,1,16,5,83,24,3,82,20,3,81,74,19,5,81,254,3,80,11,3,79,254,3,78,77,16,5,78,254,3,77,16,3,76,254,3,75,74,19,5,75,254,3,74,73,16,5,74,19,3,73,29,13,5,73,16,3,72,13,3,71,254,3,70,150,3,69,150,3,68,254,3,67,2,45,5,67,250,3,66,187,3,65,75,3,64,254,3,63,254,3,62,61,18,5,62,20,3,61,60,15,5,61,18,3,60,59,13,5,60,64,255,15,3,59,13,3,58,254,3,57,254,3,56,55,20,5,56,250,3,55,54,16,5,55,20,3,54,53,11,5,54,16,3,53,11,3,52,30,3,51,13,3,50,49,11,5,50,254,3,49,11,3,48,47,11,5,48,13,3,47,11,3,46,45,9,5,46,16,3,45,9,3,44,50,3,43,42,37,5,43,100,3,42,41,18,5,42,37,3,41,18,3,40,39,37,5,40,65,3,39,37,3,38,37,11,5,38,15,3,37,11,3,36,254,3,35,254,3,34,15,3,33,1,16,5,33,18,3,32,100,3,31,250,3,30,29,13,5,30,100,3,29,13,3,28,17,66,5,28,254,3,27,250,3,26,66,3,25,17,66,5,25,254,3,24,100,3,23,22,25,5,23,254,3,22,1,16,5,22,25,3,21,254,3,20,254,3,19,254,3,18,17,66,5,18,254,3,17,2,45,5,17,66,3,16,125,3,15,100,3,14,254,3,13,12,22,5,13,254,3,12,1,16,5,12,22,3,11,254,3,10,16,3,9,254,3,8,2,45,5,8,254,3,7,20,3,6,100,3,4,1,16,5,4,254,3,64,21,3,2,45,5,3,254,3,2,1,16,5,2,45,3,1,16,3,0,254,3,1,184,1,100,133,141,1,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,0,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,43,29,182,6,5,4,3,2,1,0,44,32,16,176,2,37,73,100,176,64,81,88,32,200,89,33,45,44,176,2,37,73,100,176,64,81,88,32,200,89,33,45,44,32,16,7,32,176,0,80,176,13,121,32,184,255,255,80,88,4,27,5,89,176,5,28,176,3,37,8,176,4,37,35,225,32,176,0,80,176,13,121,32,184,255,255,80,88,4,27,5,89,176,5,28,176,3,37,8,225,45,44,75,80,88,32,176,253,69,68,89,33,45,44,176,2,37,69,96,68,45,44,75,83,88,176,2,37,176,2,37,69,68,89,33,33,45,44,69,68,45,0,0,2,0,102,254,150,4,102,5,164,0,3,0,7,0,26,64,12,4,251,0,6,251,1,8,5,127,2,4,0,47,196,212,236,49,0,16,212,236,212,236,48,19,17,33,17,37,33,17,33,102,4,0,252,115,3,27,252,229,254,150,7,14,248,242,114,6,41,0,2,1,53,0,0,2,0,5,213,0,3,0,9,0,64,64,15,7,0,131,4,129,2,8,7,5,1,3,4,0,0,10,16,252,60,236,50,57,57,49,0,47,228,252,204,48,1,75,176,11,84,88,189,0,10,0,64,0,1,0,10,0,10,255,192,56,17,55,56,89,182,0,11,32,11,80,11,3,93,37,51,21,35,17,51,17,3,35,3,1,53,203,203,203,20,162,21,254,254,5,213,253,113,254,155,1,101,0,2,0,197,3,170,2,233,5,213,0,3,0,7,0,77,64,15,5,1,132,4,0,129,8,4,5,6,0,5,2,4,8,16,252,252,220,236,49,0,16,244,60,236,50,48,1,75,176,18,84,75,176,19,84,91,88,189,0,8,0,64,0,1,0,8,0,8,255,192,56,17,55,56,89,64,15,48,9,64,9,80,9,96,9,112,9,160,9,191,9,7,93,1,17,35,17,33,17,35,17,1,111,170,2,36,170,5,213,253,213,2,43,253,213,2,43,0,0,2,0,158,0,0,6,23,5,190,0,3,0,31,0,96,64,49,27,11,0,135,7,4,29,9,5,25,13,2,135,23,19,15,21,17,31,30,28,27,26,23,22,21,20,19,18,17,16,14,13,12,9,8,7,6,5,4,3,2,1,0,26,10,24,6,32,16,252,204,23,57,49,0,47,60,212,60,60,252,60,60,212,60,60,196,50,236,50,50,48,64,17,11,1,11,2,11,12,11,13,20,4,26,17,26,18,20,31,8,1,93,1,33,3,33,11,1,33,19,51,3,33,21,33,3,33,21,33,3,35,19,33,3,35,19,33,53,33,19,33,53,33,19,4,23,254,221,84,1,37,68,104,1,36,105,160,103,1,56,254,161,82,1,62,254,155,104,160,103,254,219,103,161,104,254,197,1,96,84,254,190,1,105,102,3,133,254,178,3,135,254,97,1,159,254,97,154,254,178,153,254,98,1,158,254,98,1,158,153,1,78,154,1,159,0,0,3,0,170,254,211,4,109,6,20,0,33,0,40,0,47,0,213,64,85,34,2,10,11,10,39,1,38,40,2,11,11,10,29,1,30,28,2,47,41,47,27,2,41,41,47,66,19,17,16,34,10,27,41,4,23,6,9,42,33,5,2,23,134,22,6,134,5,17,35,26,138,22,137,16,0,42,138,5,137,2,45,8,22,10,30,7,41,26,18,3,0,9,34,16,9,3,1,7,38,8,13,5,6,48,16,252,60,236,244,23,60,252,23,60,244,228,236,49,0,47,228,236,196,212,228,236,50,196,16,238,16,238,17,18,57,17,57,17,18,23,57,17,18,57,48,75,83,88,7,16,4,237,7,16,14,237,17,23,57,7,16,14,237,17,23,57,7,16,4,237,89,34,1,75,176,9,84,88,189,0,48,0,64,0,1,0,48,0,48,255,192,56,17,55,56,89,1,75,176,12,84,75,176,16,84,91,75,176,15,84,91,88,189,0,48,255,192,0,1,0,48,0,48,0,64,56,17,55,56,89,1,35,3,46,1,39,53,30,1,23,17,46,1,53,52,54,55,53,51,21,30,1,23,21,46,1,39,17,30,1,21,20,6,7,3,17,14,1,21,20,22,23,17,62,1,53,52,38,2,180,100,1,105,210,106,102,209,111,221,201,218,204,100,93,174,83,83,175,92,227,214,227,214,100,116,122,113,225,127,129,123,254,211,1,45,2,45,45,180,64,65,1,1,200,36,172,150,163,188,14,235,232,4,31,27,175,42,46,4,254,85,35,180,156,169,195,15,3,0,1,154,13,106,88,86,96,213,254,79,17,110,90,88,104,0,0,5,0,113,255,227,7,41,5,240,0,11,0,23,0,35,0,39,0,51,0,149,64,54,36,15,37,38,37,38,15,39,36,39,66,0,146,12,30,146,46,141,24,146,36,6,146,12,141,38,18,140,40,36,145,52,39,33,27,37,9,3,13,21,14,9,13,15,33,13,43,14,27,13,15,49,11,52,16,252,196,236,244,236,16,238,246,238,17,57,17,18,57,49,0,16,228,50,244,60,228,236,16,238,246,238,16,238,48,75,83,88,7,16,5,237,7,16,5,237,89,34,1,75,176,9,84,75,176,11,84,91,75,176,12,84,91,75,176,20,84,91,75,176,14,84,91,75,176,13,84,91,88,189,0,52,0,64,0,1,0,52,0,52,255,192,56,17,55,56,89,1,34,6,21,20,22,51,50,54,53,52,38,39,50,22,21,20,6,35,34,38,53,52,54,1,34,6,21,20,22,51,50,54,53,52,38,37,51,1,35,19,50,22,21,20,6,35,34,38,53,52,54,5,209,87,99,99,87,85,99,99,85,158,186,187,157,160,186,187,252,151,86,99,98,87,87,99,100,3,49,160,252,90,160,31,158,188,187,159,159,185,186,2,145,148,132,130,149,149,130,131,149,127,220,187,187,219,219,187,188,219,2,97,149,130,132,148,148,132,129,150,127,249,243,6,13,219,187,189,218,219,188,186,220,0,0,2,0,129,255,227,5,254,5,240,0,9,0,48,1,205,64,150,13,1,14,12,134,17,18,17,11,134,10,11,18,18,17,9,134,0,9,21,22,21,7,1,6,8,134,22,22,21,2,1,3,1,134,29,30,29,0,134,9,0,30,30,29,32,31,2,33,30,17,10,19,10,23,22,21,3,24,20,17,19,10,7,8,2,6,9,17,19,19,10,2,1,2,3,0,17,10,19,10,23,22,2,24,21,17,19,10,20,17,19,19,10,66,18,11,9,3,6,0,10,30,3,40,21,14,6,40,39,6,149,24,43,149,39,148,36,145,24,140,14,19,10,46,11,14,9,0,46,18,21,39,14,30,3,46,18,39,33,14,17,15,19,33,3,18,27,16,49,16,252,236,196,212,212,236,16,198,238,17,57,17,18,57,57,17,57,57,17,57,17,57,49,0,47,198,228,246,230,238,16,238,16,198,17,18,57,17,23,57,17,23,57,48,75,83,88,7,16,5,237,7,5,237,17,23,57,7,16,5,237,17,23,57,7,16,5,237,17,23,57,7,5,237,17,23,57,7,16,5,237,17,23,57,7,16,8,237,7,16,14,237,17,23,57,7,16,14,237,17,23,57,7,16,8,237,7,16,8,237,7,16,14,237,17,23,57,89,34,178,15,50,1,1,93,64,178,7,11,5,34,9,41,28,0,28,1,31,2,23,11,42,0,42,1,38,18,58,0,52,18,68,11,94,0,89,1,90,10,85,18,90,26,90,31,89,48,103,30,123,0,155,0,154,1,153,2,151,8,149,11,147,21,149,22,149,34,153,45,31,9,11,9,12,8,17,12,39,12,40,24,2,27,9,25,11,25,12,25,17,28,20,28,21,22,29,31,50,39,0,39,1,41,9,35,18,42,19,42,20,40,21,47,50,59,9,52,18,57,19,63,50,74,9,76,20,75,21,70,25,79,50,86,1,90,9,89,12,85,18,89,19,92,31,95,50,106,12,105,17,96,50,117,1,121,12,122,17,147,0,147,1,151,2,149,5,156,7,156,8,159,8,154,9,155,11,154,12,144,50,160,50,176,50,57,93,0,93,1,14,1,21,20,22,51,50,54,55,9,1,62,1,55,51,6,2,7,1,35,39,14,1,35,34,0,53,52,54,55,46,1,53,52,54,51,50,22,23,21,46,1,35,34,6,21,20,22,1,242,91,85,212,160,95,166,73,254,123,1,252,59,66,6,186,12,104,93,1,23,252,143,104,228,131,241,254,206,134,134,48,50,222,184,83,165,85,87,158,68,105,131,59,3,35,81,161,88,146,194,63,64,2,143,253,248,89,203,114,132,254,254,126,254,227,147,89,87,1,19,215,128,225,99,63,125,60,162,197,36,36,182,47,49,111,88,51,103,0,1,0,197,3,170,1,111,5,213,0,3,0,66,64,10,1,132,0,129,4,0,5,2,4,4,16,252,236,49,0,16,244,236,48,1,75,176,18,84,75,176,19,84,91,88,189,0,4,0,64,0,1,0,4,0,4,255,192,56,17,55,56,89,64,13,64,5,80,5,96,5,112,5,144,5,160,5,6,93,1,17,35,17,1,111,170,5,213,253,213,2,43,0,0,1,0,176,254,242,2,123,6,18,0,13,0,79,64,15,6,152,0,151,14,13,7,0,3,18,6,0,19,10,14,16,220,228,50,236,17,57,57,49,0,16,252,236,48,1,75,176,19,84,88,189,0,14,0,64,0,1,0,14,0,14,255,192,56,17,55,56,89,1,75,176,15,84,88,189,0,14,255,192,0,1,0,14,0,14,0,64,56,17,55,56,89,1,6,2,21,20,18,23,35,38,2,53,52,18,55,2,123,134,130,131,133,160,150,149,148,151,6,18,230,254,62,231,231,254,59,229,235,1,198,224,223,1,196,236,0,1,0,164,254,242,2,111,6,18,0,13,0,31,64,15,7,152,0,151,14,7,1,0,11,18,4,19,8,0,14,16,220,60,244,236,17,57,57,49,0,16,252,236,48,19,51,22,18,21,20,2,7,35,54,18,53,52,2,164,160,150,149,149,150,160,133,131,131,6,18,236,254,60,223,224,254,58,235,229,1,197,231,231,1,194,0,1,0,61,2,74,3,195,5,240,0,17,0,78,64,44,16,13,11,0,4,12,9,7,4,2,4,8,3,153,5,17,12,153,10,1,14,145,18,8,12,10,3,9,6,17,3,1,3,2,0,20,15,4,11,9,20,13,6,18,16,212,60,228,50,220,60,228,50,23,57,17,18,23,57,49,0,16,244,212,60,236,50,196,236,50,23,57,18,23,57,48,1,5,5,7,37,17,35,17,5,39,37,37,55,5,17,51,17,37,3,195,254,153,1,103,58,254,176,114,254,176,58,1,103,254,153,58,1,80,114,1,80,4,223,194,195,98,203,254,135,1,121,203,98,195,194,99,203,1,121,254,135,203,0,1,0,217,0,0,5,219,5,4,0,11,0,35,64,17,0,9,1,156,7,3,5,2,21,4,0,23,10,6,21,8,12,16,220,252,60,252,60,236,49,0,47,212,60,252,60,196,48,1,17,33,21,33,17,35,17,33,53,33,17,3,174,2,45,253,211,168,253,211,2,45,5,4,253,211,170,253,211,2,45,170,2,45,0,1,0,158,255,18,1,195,0,254,0,5,0,25,64,12,3,158,0,131,6,3,4,1,25,0,24,6,16,252,236,212,204,49,0,16,252,236,48,55,51,21,3,35,19,240,211,164,129,82,254,172,254,192,1,64,0,1,0,100,1,223,2,127,2,131,0,3,0,17,182,0,156,2,4,1,0,4,16,220,204,49,0,16,212,236,48,19,33,21,33,100,2,27,253,229,2,131,164,0,0,1,0,219,0,0,1,174,0,254,0,3,0,17,183,0,131,2,1,25,0,24,4,16,252,236,49,0,47,236,48,55,51,21,35,219,211,211,254,254,0,1,0,0,255,66,2,178,5,213,0,3,0,45,64,20,0,26,1,2,1,2,26,3,0,3,66,2,159,0,129,4,2,0,1,3,47,196,57,57,49,0,16,244,236,48,75,83,88,7,16,5,237,7,16,5,237,89,34,1,51,1,35,2,8,170,253,248,170,5,213,249,109,0,0,2,0,135,255,227,4,143,5,240,0,11,0,23,0,35,64,19,6,160,18,0,160,12,145,18,140,24,9,28,15,30,3,28,21,27,24,16,252,236,244,236,49,0,16,228,244,236,16,238,48,1,34,2,17,16,18,51,50,18,17,16,2,39,50,0,17,16,0,35,34,0,17,16,0,2,139,156,157,157,156,157,157,157,157,251,1,9,254,247,251,251,254,247,1,9,5,80,254,205,254,204,254,205,254,205,1,51,1,51,1,52,1,51,160,254,115,254,134,254,135,254,115,1,141,1,121,1,122,1,141,0,0,1,0,225,0,0,4,90,5,213,0,10,0,75,64,21,66,3,160,4,2,160,5,129,7,0,160,9,8,31,6,28,3,0,31,1,11,16,212,236,196,252,236,49,0,47,236,50,244,236,212,236,48,75,83,88,89,34,1,75,176,15,84,88,189,0,11,255,192,0,1,0,11,0,11,0,64,56,17,55,56,89,180,15,3,15,4,2,93,55,33,17,5,53,37,51,17,33,21,33,254,1,74,254,153,1,101,202,1,74,252,164,170,4,115,72,184,72,250,213,170,0,0,1,0,150,0,0,4,74,5,240,0,28,0,165,64,39,25,26,27,3,24,28,17,5,4,0,17,5,5,4,66,16,161,17,148,13,160,20,145,4,0,160,2,0,16,10,2,1,10,28,23,16,3,6,29,16,252,196,212,236,192,192,17,18,57,49,0,47,236,50,244,236,244,236,48,75,83,88,7,16,5,237,7,5,237,17,23,57,89,34,1,75,176,21,84,75,176,22,84,91,75,176,20,84,91,88,189,0,29,0,64,0,1,0,29,0,29,255,192,56,17,55,56,89,64,50,85,4,86,5,86,7,122,4,122,5,118,27,135,25,7,4,0,4,25,4,26,4,27,5,28,116,0,118,6,117,26,115,27,116,28,130,0,134,25,130,26,130,27,130,28,168,0,168,27,17,93,0,93,37,33,21,33,53,54,0,55,62,1,53,52,38,35,34,6,7,53,62,1,51,50,4,21,20,6,7,6,0,1,137,2,193,252,76,115,1,141,51,97,77,167,134,95,211,120,122,212,88,232,1,20,69,91,25,254,244,170,170,170,119,1,145,58,109,151,73,119,150,66,67,204,49,50,232,194,92,165,112,29,254,235,0,0,1,0,156,255,227,4,115,5,240,0,40,0,123,64,46,0,21,19,10,134,9,31,134,32,19,160,21,13,160,9,147,6,28,160,32,147,35,145,6,140,21,163,41,22,28,19,0,3,20,25,28,38,32,16,28,3,20,31,9,6,41,16,252,196,196,212,236,244,236,17,23,57,57,49,0,16,236,228,244,228,236,16,230,238,16,238,16,238,16,238,17,18,57,48,1,75,176,22,84,75,176,20,84,91,88,189,0,41,0,64,0,1,0,41,0,41,255,192,56,17,55,56,89,64,9,100,30,97,31,97,32,100,33,4,0,93,1,30,1,21,20,4,33,34,38,39,53,30,1,51,50,54,53,52,38,43,1,53,51,50,54,53,52,38,35,34,6,7,53,62,1,51,50,4,21,20,6,3,63,145,163,254,208,254,232,94,199,106,84,200,109,190,199,185,165,174,182,149,158,163,152,83,190,114,115,201,89,230,1,12,142,3,37,31,196,144,221,242,37,37,195,49,50,150,143,132,149,166,119,112,115,123,36,38,180,32,32,209,178,124,171,0,2,0,100,0,0,4,164,5,213,0,2,0,13,0,140,64,29,1,13,3,13,0,3,3,13,66,0,3,11,7,160,5,1,3,129,9,1,12,10,0,28,6,8,4,12,14,16,220,212,60,196,236,50,17,57,49,0,47,228,212,60,236,50,18,57,48,75,83,88,7,16,4,201,7,16,5,201,89,34,1,75,176,11,84,75,176,13,84,91,88,189,0,14,0,64,0,1,0,14,0,14,255,192,56,17,55,56,89,64,42,11,0,42,0,72,0,89,0,105,0,119,0,138,0,7,22,1,43,0,38,1,43,3,54,1,78,1,79,12,79,13,86,1,102,1,117,1,122,3,133,1,13,93,0,93,9,1,33,3,51,17,51,21,35,17,35,17,33,53,3,6,254,2,1,254,53,254,213,213,201,253,94,5,37,252,227,3,205,252,51,168,254,160,1,96,195,0,0,1,0,158,255,227,4,100,5,213,0,29,0,117,64,35,4,26,7,17,134,16,29,26,160,7,20,160,16,137,13,2,160,0,129,13,140,7,164,30,23,28,1,10,3,28,0,10,16,6,30,16,252,196,212,236,16,196,238,49,0,16,228,228,244,236,16,230,238,16,254,196,16,238,17,18,57,48,1,75,176,22,84,75,176,20,84,91,88,189,0,30,0,64,0,1,0,30,0,30,255,192,56,17,55,56,89,1,75,176,15,84,88,189,0,30,255,192,0,1,0,30,0,30,0,64,56,17,55,56,89,19,33,21,33,17,62,1,51,50,0,21,20,0,33,34,38,39,53,30,1,51,50,54,53,52,38,35,34,6,7,221,3,25,253,160,44,88,44,250,1,36,254,212,254,239,94,195,104,90,192,107,173,202,202,173,81,161,84,5,213,170,254,146,15,15,254,238,234,241,254,245,32,32,203,49,48,182,156,156,182,36,38,0,0,2,0,143,255,227,4,150,5,240,0,11,0,36,0,88,64,36,19,6,0,13,134,12,0,160,22,6,160,28,22,165,16,160,12,137,34,145,28,140,37,12,34,9,28,25,30,19,28,3,33,31,27,37,16,252,236,236,244,236,228,49,0,16,228,244,228,252,228,16,238,16,238,16,238,17,18,57,48,64,20,203,0,203,1,205,2,205,3,205,4,203,5,203,6,7,164,30,178,30,2,93,1,93,1,34,6,21,20,22,51,50,54,53,52,38,1,21,46,1,35,34,2,3,62,1,51,50,0,21,20,0,35,32,0,17,16,0,33,50,22,2,164,136,159,159,136,136,159,159,1,9,76,155,76,200,211,15,59,178,107,225,1,5,254,240,226,254,253,254,238,1,80,1,27,76,155,3,59,186,162,161,187,187,161,162,186,2,121,184,36,38,254,242,254,239,87,93,254,239,235,230,254,234,1,141,1,121,1,98,1,165,30,0,0,1,0,168,0,0,4,104,5,213,0,6,0,99,64,24,5,17,2,3,2,3,17,4,5,4,66,5,160,0,129,3,5,3,1,4,1,0,6,7,16,252,204,196,17,57,57,49,0,47,244,236,48,75,83,88,7,16,5,237,7,16,5,237,89,34,1,75,176,22,84,88,189,0,7,0,64,0,1,0,7,0,7,255,192,56,17,55,56,89,64,18,88,2,1,6,3,26,5,57,5,72,5,103,3,176,0,176,6,7,93,0,93,19,33,21,1,35,1,33,168,3,192,253,226,211,1,254,253,51,5,213,86,250,129,5,43,0,0,3,0,139,255,227,4,139,5,240,0,11,0,35,0,47,0,67,64,37,24,12,0,160,39,6,160,30,45,160,18,145,30,140,39,163,48,24,12,36,42,28,21,36,28,15,9,28,21,27,30,3,28,15,33,27,48,16,252,196,236,244,196,236,16,238,16,238,17,57,57,49,0,16,236,228,244,236,16,238,16,238,57,57,48,1,34,6,21,20,22,51,50,54,53,52,38,37,38,38,53,52,54,51,50,22,21,20,6,7,22,22,21,20,4,35,34,36,53,52,54,19,20,22,51,50,54,53,52,38,35,34,6,2,139,144,165,165,144,144,166,165,254,165,130,145,255,222,223,254,145,129,146,163,254,247,247,247,254,247,164,72,145,131,130,147,147,130,131,145,2,197,154,135,135,154,155,134,135,154,86,32,178,128,179,208,208,179,128,178,32,34,198,143,217,232,232,217,143,198,1,97,116,130,130,116,116,130,130,0,0,2,0,129,255,227,4,135,5,240,0,24,0,36,0,88,64,35,7,31,25,1,134,0,25,160,10,165,4,160,0,137,22,31,160,16,145,22,140,37,7,28,28,33,19,30,0,34,34,28,13,27,37,16,252,236,228,244,236,236,49,0,16,228,244,236,16,230,254,245,238,16,238,17,18,57,48,64,22,196,25,194,26,192,27,192,28,192,29,194,30,196,31,7,170,18,188,18,233,18,3,93,1,93,55,53,30,1,51,50,18,19,14,1,35,34,0,53,52,0,51,32,0,17,16,0,33,34,38,1,50,54,53,52,38,35,34,6,21,20,22,225,76,156,75,200,211,15,58,178,108,224,254,251,1,16,226,1,3,1,17,254,177,254,229,76,156,1,62,136,159,159,136,136,159,159,31,184,36,38,1,13,1,18,86,92,1,15,235,230,1,22,254,115,254,134,254,159,254,91,30,2,151,186,162,161,187,187,161,162,186,0,0,2,0,240,0,0,1,195,4,35,0,3,0,7,0,28,64,14,6,131,4,166,0,131,2,5,1,3,4,0,24,8,16,252,60,236,50,49,0,47,236,244,236,48,55,51,21,35,17,51,21,35,240,211,211,211,211,254,254,4,35,254,0,2,0,158,255,18,1,195,4,35,0,3,0,9,0,37,64,19,2,131,0,7,158,4,131,0,166,10,7,8,5,1,25,4,0,24,10,16,252,60,236,50,212,204,49,0,16,228,252,236,16,238,48,19,51,21,35,17,51,21,3,35,19,240,211,211,211,164,129,82,4,35,254,253,217,172,254,192,1,64,0,1,0,217,0,94,5,219,4,166,0,6,0,77,64,42,2,156,3,4,3,1,156,0,1,4,4,3,1,156,2,1,5,6,5,0,156,6,5,66,5,4,2,1,0,5,3,168,6,167,7,1,2,0,36,4,35,7,16,252,236,50,57,49,0,16,244,236,23,57,48,75,83,88,7,4,237,7,16,8,237,7,16,8,237,7,16,4,237,89,34,9,2,21,1,53,1,5,219,251,248,4,8,250,254,5,2,3,240,254,145,254,147,182,1,209,166,1,209,0,0,2,0,217,1,96,5,219,3,162,0,3,0,7,0,28,64,13,0,156,2,6,156,4,8,5,1,4,0,35,8,16,252,60,196,50,49,0,16,212,236,212,236,48,19,33,21,33,21,33,21,33,217,5,2,250,254,5,2,250,254,3,162,168,240,170,0,1,0,217,0,94,5,219,4,166,0,6,0,79,64,43,6,156,0,6,3,4,3,5,156,4,4,3,0,156,1,2,1,6,156,5,6,2,2,1,66,6,5,3,2,0,5,4,168,1,167,7,6,2,36,4,0,35,7,16,252,60,236,57,49,0,16,244,236,23,57,48,75,83,88,7,16,8,237,7,16,4,237,7,16,4,237,7,16,8,237,89,34,19,53,1,21,1,53,1,217,5,2,250,254,4,6,3,240,182,254,47,166,254,47,182,1,109,0,2,0,147,0,0,3,176,5,240,0,3,0,36,0,112,64,43,36,30,9,6,4,10,29,19,4,0,20,134,19,136,16,149,23,145,0,131,2,29,26,13,9,5,4,10,30,1,13,28,26,4,28,5,1,3,0,38,26,19,37,16,220,196,252,236,212,236,16,238,17,57,57,17,18,57,17,18,57,49,0,47,238,246,254,244,238,16,205,17,57,57,23,57,48,1,75,176,12,84,88,189,0,37,0,64,0,1,0,37,0,37,255,192,56,17,55,56,89,182,121,9,122,10,122,32,3,93,37,51,21,35,19,35,53,52,54,63,1,62,1,53,52,38,35,34,6,7,53,62,1,51,50,22,21,20,6,15,1,14,1,7,14,1,21,1,135,203,203,197,191,56,90,90,57,51,131,108,79,179,97,94,193,103,184,223,72,90,88,47,39,8,6,6,254,254,1,145,154,101,130,86,89,53,94,49,89,110,70,67,188,57,56,194,159,76,137,86,86,47,53,25,21,60,52,0,0,2,0,135,254,156,7,113,5,162,0,11,0,76,0,149,64,50,24,12,3,9,169,25,21,27,3,169,76,15,52,51,15,172,48,169,55,21,172,36,169,55,67,77,51,52,30,26,0,40,18,6,24,12,40,26,43,30,40,73,18,43,42,40,73,44,61,77,16,220,236,252,236,16,254,253,254,60,198,16,238,17,18,57,57,49,0,16,212,196,252,236,16,254,237,212,198,16,197,238,50,16,196,238,17,57,57,48,0,75,176,9,84,75,176,12,84,91,75,176,16,84,91,75,176,19,84,91,75,176,20,84,91,88,189,0,77,255,192,0,1,0,77,0,77,0,64,56,17,55,56,89,64,9,15,78,31,78,47,78,63,78,4,1,93,1,20,22,51,50,54,53,52,38,35,34,6,1,14,1,35,34,38,53,52,54,51,50,22,23,53,51,17,62,1,53,52,38,39,38,36,35,34,6,7,6,2,21,20,18,23,22,4,51,50,54,55,23,6,4,35,34,36,39,38,2,53,52,18,55,54,36,51,50,4,23,30,1,21,16,0,5,2,250,142,124,123,141,144,122,121,143,2,33,60,155,103,172,215,216,171,103,156,59,143,146,165,63,64,104,254,213,176,123,226,96,157,177,115,109,105,1,20,157,129,249,104,90,125,254,217,152,185,254,184,128,128,134,136,126,129,1,82,189,212,1,107,123,75,79,254,194,254,232,2,25,143,163,164,142,140,165,164,254,72,77,73,249,200,200,250,75,76,131,253,32,22,223,177,107,188,80,131,139,65,64,102,254,181,193,159,254,234,106,104,109,87,81,111,97,103,131,125,125,1,73,189,182,1,74,125,127,135,174,160,98,230,123,254,249,254,208,6,0,0,2,0,16,0,0,5,104,5,213,0,2,0,10,0,186,64,65,0,17,1,0,4,5,4,2,17,5,5,4,1,17,10,3,10,0,17,2,0,3,3,10,7,17,5,4,6,17,5,5,4,9,17,3,10,8,17,10,3,10,66,0,3,7,149,1,3,129,9,5,9,8,7,6,4,3,2,1,0,9,5,10,11,16,212,196,23,57,49,0,47,60,228,212,236,18,57,48,75,83,88,7,16,5,237,7,5,237,7,16,5,237,7,5,237,7,16,8,237,7,16,5,237,7,16,5,237,7,16,8,237,89,34,178,32,12,1,1,93,64,58,15,0,88,0,118,0,112,0,140,0,5,7,1,8,2,6,3,9,4,22,1,25,2,86,1,88,2,80,12,103,1,104,2,120,1,118,2,124,3,114,4,119,7,120,8,135,1,136,2,128,12,152,2,153,3,150,4,23,93,0,93,9,1,33,1,51,1,35,3,33,3,35,2,188,254,238,2,37,254,123,229,2,57,210,136,253,95,136,213,5,14,253,25,3,174,250,43,1,127,254,129,0,3,0,201,0,0,4,236,5,213,0,8,0,17,0,32,0,67,64,35,25,0,149,10,9,149,18,129,1,149,10,173,31,17,11,8,2,19,25,31,5,0,14,28,22,5,25,28,46,9,0,28,18,4,33,16,252,236,50,252,236,212,236,17,23,57,57,57,49,0,47,236,236,244,236,16,238,57,48,178,15,34,1,1,93,1,17,33,50,54,53,52,38,35,1,17,33,50,54,53,52,38,35,37,33,50,22,21,20,6,7,30,1,21,20,4,35,33,1,147,1,68,163,157,157,163,254,188,1,43,148,145,145,148,254,11,2,4,231,250,128,124,149,165,254,240,251,253,232,2,201,253,221,135,139,140,133,2,102,254,62,111,114,113,112,166,192,177,137,162,20,32,203,152,200,218,0,1,0,115,255,227,5,39,5,240,0,25,0,54,64,26,13,161,14,174,10,149,17,1,161,0,174,4,149,23,145,17,140,26,7,25,13,0,48,20,16,26,16,252,236,50,236,49,0,16,228,244,236,244,236,16,238,246,238,48,180,15,27,31,27,2,1,93,1,21,46,1,35,32,0,17,16,0,33,50,54,55,21,14,1,35,32,0,17,16,0,33,50,22,5,39,102,231,130,255,0,254,240,1,16,1,0,130,231,102,106,237,132,254,173,254,122,1,134,1,83,134,237,5,98,213,95,94,254,199,254,216,254,217,254,199,94,95,211,72,72,1,159,1,103,1,104,1,159,71,0,2,0,201,0,0,5,176,5,213,0,8,0,17,0,46,64,21,0,149,9,129,1,149,16,8,2,16,10,0,5,25,13,50,0,28,9,4,18,16,252,236,244,236,17,57,57,57,57,49,0,47,236,244,236,48,178,96,19,1,1,93,1,17,51,32,0,17,16,0,33,37,33,32,0,17,16,0,41,1,1,147,244,1,53,1,31,254,225,254,203,254,66,1,159,1,178,1,150,254,104,254,80,254,97,5,47,251,119,1,24,1,46,1,44,1,23,166,254,151,254,128,254,126,254,150,0,1,0,201,0,0,4,139,5,213,0,11,0,46,64,21,6,149,4,2,149,0,129,8,149,4,173,10,5,1,9,7,3,28,0,4,12,16,252,236,50,212,196,196,49,0,47,236,236,244,236,16,238,48,178,31,13,1,1,93,19,33,21,33,17,33,21,33,17,33,21,33,201,3,176,253,26,2,199,253,57,2,248,252,62,5,213,170,254,70,170,253,227,170,0,1,0,201,0,0,4,35,5,213,0,9,0,41,64,18,6,149,4,2,149,0,129,4,173,8,5,1,7,3,28,0,4,10,16,252,236,50,212,196,49,0,47,236,244,236,16,238,48,178,15,11,1,1,93,19,33,21,33,17,33,21,33,17,35,201,3,90,253,112,2,80,253,176,202,5,213,170,254,72,170,253,55,0,0,1,0,115,255,227,5,139,5,240,0,29,0,57,64,32,0,5,27,1,149,3,27,149,8,18,161,17,174,21,149,14,145,8,140,30,2,0,28,17,52,4,51,24,25,11,16,30,16,252,236,252,228,252,196,49,0,16,228,244,236,244,236,16,254,212,238,17,57,57,48,37,17,33,53,33,17,6,4,35,32,0,17,16,0,33,50,4,23,21,38,38,35,32,0,17,16,0,33,50,54,4,195,254,182,2,18,117,254,230,160,254,162,254,117,1,139,1,94,146,1,7,111,112,252,139,254,238,254,237,1,19,1,18,107,168,213,1,145,166,253,127,83,85,1,153,1,109,1,110,1,153,72,70,215,95,96,254,206,254,209,254,210,254,206,37,0,1,0,201,0,0,5,59,5,213,0,11,0,44,64,20,8,149,2,173,4,0,129,10,6,7,3,28,5,56,9,1,28,0,4,12,16,252,236,50,252,236,50,49,0,47,60,228,50,252,236,48,178,80,13,1,1,93,19,51,17,33,17,51,17,35,17,33,17,35,201,202,2,222,202,202,253,34,202,5,213,253,156,2,100,250,43,2,199,253,57,0,0,1,0,201,0,0,1,147,5,213,0,3,0,57,183,0,175,2,1,28,0,4,4,16,252,236,49,0,47,236,48,1,75,176,16,84,88,189,0,4,255,192,0,1,0,4,0,4,0,64,56,17,55,56,89,64,13,48,5,64,5,80,5,96,5,143,5,159,5,6,93,19,51,17,35,201,202,202,5,213,250,43,0,1,255,150,254,102,1,147,5,213,0,11,0,77,64,19,11,2,0,7,149,5,176,0,129,12,5,8,6,57,1,28,0,4,12,16,252,236,228,57,57,49,0,16,228,252,236,17,57,57,48,1,75,176,16,84,88,189,0,12,255,192,0,1,0,12,0,12,0,64,56,17,55,56,89,64,13,48,13,64,13,80,13,96,13,143,13,159,13,6,93,19,51,17,16,6,43,1,53,51,50,54,53,201,202,205,227,77,63,134,110,5,213,250,147,254,242,244,170,150,194,0,0,1,0,201,0,0,5,106,5,213,0,10,0,239,64,40,8,17,5,6,5,7,17,6,6,5,3,17,4,5,4,2,17,5,5,4,66,8,5,2,3,3,0,175,9,6,5,1,4,6,8,1,28,0,4,11,16,252,236,50,212,196,17,57,49,0,47,60,236,50,23,57,48,75,83,88,7,16,4,237,7,16,5,237,7,16,5,237,7,16,4,237,89,34,178,8,3,1,1,93,64,146,20,2,1,4,2,9,8,22,2,40,5,40,8,55,2,54,5,52,8,71,2,70,5,67,8,85,2,103,2,118,2,119,5,131,2,136,5,143,8,148,2,155,8,231,2,21,6,3,9,5,9,6,27,3,25,7,5,10,3,10,7,24,3,40,5,43,6,42,7,54,4,54,5,54,6,53,7,48,12,65,3,64,4,69,5,64,6,64,7,64,12,98,3,96,4,104,5,103,7,119,5,112,12,139,3,139,5,142,6,143,7,143,12,154,3,157,6,157,7,182,3,181,7,197,3,197,7,215,3,214,7,232,3,233,4,232,5,234,6,247,3,248,5,249,6,44,93,113,0,93,113,19,51,17,1,33,9,1,33,1,17,35,201,202,2,158,1,4,253,27,3,26,254,246,253,51,202,5,213,253,137,2,119,253,72,252,227,2,207,253,49,0,0,1,0,201,0,0,4,106,5,213,0,5,0,37,64,12,2,149,0,129,4,1,28,3,58,0,4,6,16,252,236,236,49,0,47,228,236,48,64,9,48,7,80,7,128,3,128,4,4,1,93,19,51,17,33,21,33,201,202,2,215,252,95,5,213,250,213,170,0,1,0,201,0,0,6,31,5,213,0,12,0,191,64,52,3,17,7,8,7,2,17,1,2,8,8,7,2,17,3,2,9,10,9,1,17,10,10,9,66,10,7,2,3,8,3,0,175,8,11,5,9,8,3,2,1,5,10,6,28,4,62,10,28,0,4,13,16,252,236,252,236,17,23,57,49,0,47,60,196,236,50,17,23,57,48,75,83,88,7,16,5,237,7,16,8,237,7,16,8,237,7,16,5,237,89,34,178,112,14,1,1,93,64,86,3,7,15,8,15,9,2,10,21,2,20,7,19,10,38,2,38,7,32,7,38,10,32,10,52,7,53,10,105,2,124,2,123,7,121,10,128,2,130,7,130,10,144,2,22,4,1,11,3,19,1,27,3,35,1,44,3,39,8,40,9,52,1,60,3,86,8,89,9,101,8,106,9,118,8,121,9,129,1,141,3,149,1,155,3,20,93,0,93,19,33,9,1,33,17,35,17,1,35,1,17,35,201,1,45,1,125,1,127,1,45,197,254,127,203,254,127,196,5,213,252,8,3,248,250,43,5,31,252,0,4,0,250,225,0,1,0,201,0,0,5,51,5,213,0,9,0,121,64,30,7,17,1,2,1,2,17,6,7,6,66,7,2,3,0,175,8,5,6,1,7,2,28,4,54,7,28,0,4,10,16,252,236,252,236,17,57,57,49,0,47,60,236,50,57,57,48,75,83,88,7,16,4,237,7,16,4,237,89,34,178,31,11,1,1,93,64,48,54,2,56,7,72,2,71,7,105,2,102,7,128,2,7,6,1,9,6,21,1,26,6,70,1,73,6,87,1,88,6,101,1,105,6,121,6,133,1,138,6,149,1,154,6,159,11,16,93,0,93,19,33,1,17,51,17,33,1,17,35,201,1,16,2,150,196,254,240,253,106,196,5,213,251,31,4,225,250,43,4,225,251,31,0,2,0,115,255,227,5,217,5,240,0,11,0,23,0,35,64,19,6,149,18,0,149,12,145,18,140,24,9,25,15,51,3,25,21,16,24,16,252,236,252,236,49,0,16,228,244,236,16,238,48,1,34,0,17,16,0,51,50,0,17,16,0,39,32,0,17,16,0,33,32,0,17,16,0,3,39,220,254,253,1,3,220,220,1,1,254,255,220,1,58,1,120,254,136,254,198,254,197,254,135,1,121,5,76,254,184,254,229,254,230,254,184,1,72,1,26,1,27,1,72,164,254,91,254,158,254,159,254,91,1,164,1,98,1,98,1,165,0,2,0,201,0,0,4,141,5,213,0,8,0,19,0,58,64,24,1,149,16,0,149,9,129,18,16,10,8,2,4,0,5,25,13,63,17,0,28,9,4,20,16,252,236,50,252,236,17,23,57,49,0,47,244,236,212,236,48,64,11,15,21,31,21,63,21,95,21,175,21,5,1,93,1,17,51,50,54,53,52,38,35,37,33,50,4,21,20,4,43,1,17,35,1,147,254,141,154,154,141,254,56,1,200,251,1,1,254,255,251,254,202,5,47,253,207,146,135,134,146,166,227,219,221,226,253,168,0,2,0,115,254,248,5,217,5,240,0,11,0,29,0,82,64,42,17,16,2,15,1,12,13,12,14,1,13,13,12,66,15,30,12,6,149,18,0,149,24,145,18,140,13,30,13,27,15,12,3,9,25,27,51,3,25,21,16,30,16,252,236,252,236,17,57,57,17,57,49,0,16,196,228,244,236,16,238,57,18,57,48,75,83,88,7,16,5,237,7,16,5,237,23,57,89,34,1,34,0,17,16,0,51,50,0,17,16,0,19,1,35,39,6,6,35,32,0,17,16,0,33,32,0,17,16,2,3,39,220,254,253,1,3,220,220,1,1,254,255,63,1,10,244,221,33,35,16,254,197,254,135,1,121,1,59,1,58,1,120,209,5,76,254,184,254,229,254,230,254,184,1,72,1,26,1,27,1,72,250,207,254,221,239,2,2,1,165,1,97,1,98,1,165,254,91,254,158,254,252,254,142,0,0,2,0,201,0,0,5,84,5,213,0,19,0,28,0,177,64,53,9,8,7,3,10,6,17,3,4,3,5,17,4,4,3,66,6,4,0,21,3,4,21,149,9,20,149,13,129,11,4,5,6,3,17,9,0,28,22,14,5,10,25,25,4,17,63,20,10,28,12,4,29,16,252,236,50,252,196,236,17,23,57,17,57,57,57,49,0,47,60,244,236,212,236,18,57,18,57,18,57,48,75,83,88,7,16,5,237,7,16,5,237,17,23,57,89,34,178,64,30,1,1,93,64,66,122,19,1,5,0,5,1,5,2,6,3,7,4,21,0,21,1,20,2,22,3,23,4,37,0,37,1,37,2,38,3,39,6,38,7,38,8,38,9,32,30,54,1,54,2,70,1,70,2,104,5,117,4,117,5,119,19,136,6,136,7,152,6,152,7,31,93,0,93,1,30,1,23,19,35,3,46,1,43,1,17,35,17,33,32,22,21,20,6,1,17,51,50,54,53,52,38,35,3,141,65,123,62,205,217,191,74,139,120,220,202,1,200,1,0,252,131,253,137,254,146,149,149,146,2,188,22,144,126,254,104,1,127,150,98,253,137,5,213,214,216,141,186,2,79,253,238,135,131,131,133,0,0,1,0,135,255,227,4,162,5,240,0,39,0,126,64,60,13,12,2,14,11,2,30,31,30,8,9,2,7,10,2,31,31,30,66,10,11,30,31,4,21,1,0,21,161,20,148,24,149,17,4,149,0,148,37,145,17,140,40,30,10,11,31,27,7,0,34,27,25,14,45,7,25,20,34,40,16,220,196,236,252,236,228,17,18,57,57,57,57,49,0,16,228,244,228,236,16,238,246,238,16,198,17,23,57,48,75,83,88,7,16,14,237,17,23,57,7,16,14,237,17,23,57,89,34,178,15,41,1,1,93,182,31,41,47,41,79,41,3,93,1,21,46,1,35,34,6,21,20,22,31,1,30,1,21,20,4,33,34,38,39,53,30,1,51,50,54,53,52,38,47,1,46,1,53,52,36,51,50,22,4,72,115,204,95,165,179,119,166,122,226,215,254,221,254,231,106,239,128,123,236,114,173,188,135,154,123,226,202,1,23,245,105,218,5,164,197,55,54,128,118,99,101,31,25,43,217,182,217,224,48,47,208,69,70,136,126,110,124,31,24,45,192,171,198,228,38,0,0,1,255,250,0,0,4,233,5,213,0,7,0,74,64,14,6,2,149,0,129,4,1,64,3,28,0,64,5,8,16,212,228,252,228,49,0,47,244,236,50,48,1,75,176,10,84,88,189,0,8,0,64,0,1,0,8,0,8,255,192,56,17,55,56,89,64,19,0,9,31,0,16,1,16,2,31,7,16,9,64,9,112,9,159,9,9,93,3,33,21,33,17,35,17,33,6,4,239,253,238,203,253,238,5,213,170,250,213,5,43,0,0,1,0,178,255,227,5,41,5,213,0,17,0,75,64,22,8,2,17,11,0,5,149,14,140,9,0,129,18,8,28,10,56,1,28,0,65,18,16,252,236,252,236,49,0,16,228,50,244,236,17,57,57,57,57,48,1,75,176,16,84,88,189,0,18,0,64,0,1,0,18,0,18,255,192,56,17,55,56,89,182,31,19,143,19,159,19,3,93,19,51,17,20,22,51,50,54,53,17,51,17,16,0,33,32,0,17,178,203,174,195,194,174,203,254,223,254,230,254,229,254,223,5,213,252,117,240,211,211,240,3,139,252,92,254,220,254,214,1,42,1,36,0,1,0,16,0,0,5,104,5,213,0,6,0,183,64,39,4,17,5,6,5,3,17,2,3,6,6,5,3,17,4,3,0,1,0,2,17,1,1,0,66,3,4,1,175,0,6,4,3,2,0,5,5,1,7,16,212,196,23,57,49,0,47,236,50,57,48,75,83,88,7,16,5,237,7,16,8,237,7,16,8,237,7,16,5,237,89,34,178,80,8,1,1,93,64,98,0,3,42,3,71,4,71,5,90,3,125,3,131,3,7,6,0,7,2,8,4,9,6,21,1,20,2,26,4,26,5,42,0,38,1,38,2,41,4,41,5,37,6,32,8,56,0,51,1,51,2,60,4,60,5,55,6,72,0,69,1,69,2,73,4,73,5,71,6,89,0,86,6,102,2,105,4,105,5,122,0,118,1,118,2,121,4,121,5,117,6,128,8,152,0,151,6,41,93,0,93,33,1,51,9,1,51,1,2,74,253,198,211,1,217,1,218,210,253,199,5,213,251,23,4,233,250,43,0,1,0,68,0,0,7,166,5,213,0,12,1,123,64,73,5,26,6,5,9,10,9,4,26,10,9,3,26,10,11,10,2,26,1,2,11,11,10,6,17,7,8,7,5,17,4,5,8,8,7,2,17,3,2,12,0,12,1,17,0,0,12,66,10,5,2,3,6,3,0,175,11,8,12,11,10,9,8,6,5,4,3,2,1,11,7,0,13,16,212,204,23,57,49,0,47,60,236,50,50,23,57,48,75,83,88,7,16,5,237,7,16,8,237,7,16,8,237,7,16,5,237,7,16,8,237,7,16,5,237,7,5,237,7,16,8,237,89,34,178,0,14,1,1,93,64,242,6,2,6,5,2,10,0,10,0,10,18,10,40,5,36,10,32,10,62,2,62,5,52,10,48,10,76,2,77,5,66,10,64,10,89,2,106,2,107,5,103,10,96,10,123,2,127,2,124,5,127,5,128,10,150,2,149,5,29,7,0,9,2,8,3,0,4,6,5,0,5,0,6,1,7,4,8,0,8,7,9,0,9,4,10,10,12,0,14,26,3,21,4,21,8,25,12,16,14,32,4,33,5,32,6,32,7,32,8,35,9,36,10,37,11,32,14,32,14,60,2,58,3,53,4,51,5,48,8,54,9,57,11,63,12,48,14,70,0,70,1,74,2,64,4,69,5,64,5,66,6,66,7,66,8,64,8,64,9,68,10,77,12,64,14,64,14,88,2,86,8,89,12,80,14,102,2,103,3,97,4,98,5,96,6,96,7,96,8,100,9,100,10,100,11,119,0,118,1,123,2,120,3,119,4,116,5,121,6,121,7,119,8,112,8,120,12,127,12,127,14,134,2,135,3,136,4,137,5,133,9,138,11,143,14,151,4,159,14,175,14,91,93,0,93,19,51,9,1,51,9,1,51,1,35,9,1,35,68,204,1,58,1,57,227,1,58,1,57,205,254,137,254,254,197,254,194,254,5,213,251,18,4,238,251,18,4,238,250,43,5,16,250,240,0,1,0,61,0,0,5,59,5,213,0,11,1,93,64,70,9,17,10,11,10,8,17,7,8,11,11,10,8,17,9,8,5,6,5,7,17,6,6,5,3,17,4,5,4,2,17,1,2,5,5,4,2,17,3,2,11,0,11,1,17,0,0,11,66,11,8,5,2,4,3,0,175,9,6,11,8,5,2,4,0,4,6,0,10,12,16,212,196,220,196,17,23,57,49,0,47,60,236,50,23,57,48,75,83,88,7,16,5,237,7,16,8,237,7,16,8,237,7,16,5,237,7,16,5,237,7,16,8,237,7,16,8,237,7,16,5,237,89,34,1,75,176,12,84,75,176,13,84,91,75,176,14,84,91,88,189,0,12,255,192,0,1,0,12,0,12,0,64,56,17,55,56,89,64,184,7,2,8,8,22,2,25,8,23,11,39,8,39,11,52,2,56,8,54,11,75,8,88,5,91,8,102,2,107,8,119,2,119,11,134,2,128,2,135,5,139,8,133,11,148,2,144,2,151,5,157,8,150,11,27,6,1,9,3,8,7,7,9,22,1,25,3,25,7,23,9,16,13,38,1,40,2,41,3,38,5,40,7,39,9,41,11,32,13,53,0,52,1,60,3,59,4,58,6,59,7,52,9,52,10,56,11,63,13,72,9,79,13,88,11,95,13,101,0,101,1,106,3,106,4,104,5,105,6,105,7,108,9,108,10,120,3,121,6,121,7,120,8,125,9,127,10,120,11,128,0,128,1,131,2,136,3,133,5,132,8,131,11,143,13,144,0,144,1,148,2,151,5,151,6,149,8,147,11,159,13,175,13,64,93,0,93,19,51,9,1,51,9,1,35,9,1,35,1,129,217,1,115,1,117,217,254,32,2,0,217,254,92,254,89,218,2,21,5,213,253,213,2,43,253,51,252,248,2,123,253,133,3,29,0,1,255,252,0,0,4,231,5,213,0,8,0,148,64,40,3,17,4,5,4,2,17,1,2,5,5,4,2,17,3,2,8,0,8,1,17,0,0,8,66,2,3,0,175,6,2,7,4,64,5,28,0,64,7,9,16,212,228,252,228,18,57,49,0,47,236,50,57,48,75,83,88,7,16,5,237,7,16,8,237,7,16,8,237,7,16,5,237,89,34,178,0,10,1,1,93,64,60,5,2,20,2,53,2,48,2,48,5,48,8,70,2,64,2,64,5,64,8,81,2,81,5,81,8,101,2,132,2,147,2,16,22,1,26,3,31,10,38,1,41,3,55,1,56,3,64,10,103,1,104,3,120,3,112,10,159,10,13,93,0,93,3,51,9,1,51,1,17,35,17,4,217,1,158,1,155,217,253,240,203,5,213,253,154,2,102,252,242,253,57,2,199,0,0,1,0,92,0,0,5,31,5,213,0,9,0,155,64,27,3,17,7,8,7,8,17,2,3,2,66,8,149,0,129,3,149,5,8,3,0,1,66,4,0,6,10,16,220,196,212,228,17,57,57,49,0,47,236,244,236,48,75,83,88,7,16,5,237,7,16,5,237,89,34,1,75,176,9,84,75,176,10,84,91,88,189,0,10,0,64,0,1,0,10,0,10,255,192,56,17,55,56,89,64,64,5,2,10,7,24,7,41,2,38,7,56,7,72,2,71,7,72,8,9,5,3,11,8,0,11,22,3,26,8,16,11,47,11,53,3,57,8,63,11,71,3,74,8,79,11,85,3,89,8,102,3,105,8,111,11,119,3,120,8,127,11,159,11,22,93,0,93,19,33,21,1,33,21,33,53,1,33,115,4,149,252,80,3,199,251,61,3,176,252,103,5,213,154,251,111,170,154,4,145,0,0,1,0,176,254,242,2,88,6,20,0,7,0,83,64,15,4,169,6,178,2,169,0,177,8,5,1,3,67,0,8,16,220,252,204,50,49,0,16,252,236,244,236,48,1,75,176,12,84,88,189,0,8,255,192,0,1,0,8,0,8,0,64,56,17,55,56,89,1,75,176,18,84,75,176,19,84,91,88,189,0,8,0,64,0,1,0,8,0,8,255,192,56,17,55,56,89,19,33,21,35,17,51,21,33,176,1,168,240,240,254,88,6,20,143,249,252,143,0,1,0,0,255,66,2,178,5,213,0,3,0,45,64,20,2,26,1,1,0,0,26,3,3,2,66,1,159,0,129,4,2,0,1,3,47,196,57,57,49,0,16,244,236,48,75,83,88,7,16,5,237,7,16,5,237,89,34,19,1,35,1,170,2,8,170,253,248,5,213,249,109,6,147,0,0,1,0,199,254,242,2,111,6,20,0,7,0,60,64,16,3,169,1,178,5,169,0,177,8,0,67,4,6,2,4,8,16,252,60,220,236,49,0,16,252,236,244,236,48,1,75,176,15,84,75,176,16,84,91,88,189,0,8,255,192,0,1,0,8,0,8,0,64,56,17,55,56,89,1,17,33,53,51,17,35,53,2,111,254,88,239,239,6,20,248,222,143,6,4,143,0,1,0,217,3,168,5,219,5,213,0,6,0,24,64,10,3,4,1,0,129,7,3,1,5,7,16,220,204,57,49,0,16,244,204,50,57,48,1,1,35,1,1,35,1,3,188,2,31,201,254,72,254,72,201,2,31,5,213,253,211,1,139,254,117,2,45,0,0,1,255,236,254,29,4,20,254,172,0,3,0,15,181,0,169,1,0,2,4,16,196,196,49,0,212,236,48,1,21,33,53,4,20,251,216,254,172,143,143,0,0,1,0,170,4,240,2,137,6,102,0,3,0,49,64,9,1,180,0,179,4,3,68,1,4,16,220,236,49,0,16,244,236,48,0,75,176,9,84,75,176,14,84,91,88,189,0,4,255,192,0,1,0,4,0,4,0,64,56,17,55,56,89,9,1,35,1,1,111,1,26,153,254,186,6,102,254,138,1,118,0,2,0,123,255,227,4,45,4,123,0,10,0,37,0,188,64,39,25,31,11,23,9,14,0,169,23,6,185,14,17,32,134,31,186,28,185,35,184,17,140,23,12,0,23,3,24,13,9,8,11,31,3,8,20,69,38,16,252,236,204,212,236,50,50,17,57,57,49,0,47,196,228,244,252,244,236,16,198,238,16,238,17,57,17,57,18,57,48,64,110,48,29,48,30,48,31,48,32,48,33,48,34,63,39,64,29,64,30,64,31,64,32,64,33,64,34,80,29,80,30,80,31,80,32,80,33,80,34,80,39,112,39,133,29,135,30,135,31,135,32,135,33,133,34,144,39,160,39,240,39,30,48,30,48,31,48,32,48,33,64,30,64,31,64,32,64,33,80,30,80,31,80,32,80,33,96,30,96,31,96,32,96,33,112,30,112,31,112,32,112,33,128,30,128,31,128,32,128,33,24,93,1,93,1,34,6,21,20,22,51,50,54,61,1,55,17,35,53,14,1,35,34,38,53,52,54,51,33,53,52,38,35,34,6,7,53,62,1,51,50,22,2,190,223,172,129,111,153,185,184,184,63,188,136,172,203,253,251,1,2,167,151,96,182,84,101,190,90,243,240,2,51,102,123,98,115,217,180,41,76,253,129,170,102,97,193,162,189,192,18,127,139,46,46,170,39,39,252,0,0,2,0,186,255,227,4,164,6,20,0,11,0,28,0,56,64,25,3,185,12,15,9,185,24,21,140,15,184,27,151,25,0,18,18,71,24,12,6,8,26,70,29,16,252,236,50,50,244,236,49,0,47,236,228,244,196,236,16,198,238,48,182,96,30,128,30,160,30,3,1,93,1,52,38,35,34,6,21,20,22,51,50,54,1,62,1,51,50,18,17,16,2,35,34,38,39,21,35,17,51,3,229,167,146,146,167,167,146,146,167,253,142,58,177,123,204,255,255,204,123,177,58,185,185,2,47,203,231,231,203,203,231,231,2,82,100,97,254,188,254,248,254,248,254,188,97,100,168,6,20,0,0,1,0,113,255,227,3,231,4,123,0,25,0,63,64,27,0,134,1,136,4,14,134,13,136,10,185,17,4,185,23,184,17,140,26,7,18,13,0,72,20,69,26,16,252,228,50,236,49,0,16,228,244,236,16,254,244,238,16,245,238,48,64,11,15,27,16,27,128,27,144,27,160,27,5,1,93,1,21,46,1,35,34,6,21,20,22,51,50,54,55,21,14,1,35,34,0,17,16,0,33,50,22,3,231,78,157,80,179,198,198,179,80,157,78,77,165,93,253,254,214,1,45,1,6,85,162,4,53,172,43,43,227,205,205,227,43,43,170,36,36,1,62,1,14,1,18,1,58,35,0,2,0,113,255,227,4,90,6,20,0,16,0,28,0,56,64,25,26,185,0,14,20,185,5,8,140,14,184,1,151,3,23,4,0,8,2,71,17,18,11,69,29,16,252,236,244,236,50,50,49,0,47,236,228,244,196,236,16,196,238,48,182,96,30,128,30,160,30,3,1,93,1,17,51,17,35,53,14,1,35,34,2,17,16,18,51,50,22,1,20,22,51,50,54,53,52,38,35,34,6,3,162,184,184,58,177,124,203,255,255,203,124,177,253,199,167,146,146,168,168,146,146,167,3,182,2,94,249,236,168,100,97,1,68,1,8,1,8,1,68,97,254,21,203,231,231,203,203,231,231,0,0,2,0,113,255,227,4,127,4,123,0,20,0,27,0,112,64,36,0,21,1,9,134,8,136,5,21,169,1,5,185,12,1,187,24,185,18,184,12,140,28,27,21,2,8,21,8,0,75,2,18,15,69,28,16,252,236,244,236,196,17,18,57,49,0,16,228,244,236,228,16,238,16,238,16,244,238,17,18,57,48,64,41,63,29,112,29,160,29,208,29,240,29,5,63,0,63,1,63,2,63,21,63,27,5,44,7,47,8,47,9,44,10,111,0,111,1,111,2,111,21,111,27,9,93,113,1,93,1,21,33,30,1,51,50,54,55,21,14,1,35,32,0,17,16,0,51,50,0,7,46,1,35,34,6,7,4,127,252,178,12,205,183,106,199,98,99,208,107,254,244,254,199,1,41,252,226,1,7,184,2,165,136,154,185,14,2,94,90,190,199,52,52,174,42,44,1,56,1,10,1,19,1,67,254,221,196,151,180,174,158,0,0,1,0,47,0,0,2,248,6,20,0,19,0,112,64,28,5,16,1,12,8,169,6,1,135,0,151,14,6,188,10,2,19,7,0,7,9,5,8,13,15,11,76,20,16,252,60,196,252,60,196,196,18,57,57,49,0,47,228,50,252,236,16,238,50,18,57,57,48,1,75,176,10,84,88,189,0,20,255,192,0,1,0,20,0,20,0,64,56,17,55,56,89,1,75,176,14,84,88,189,0,20,0,64,0,1,0,20,0,20,255,192,56,17,55,56,89,182,64,21,80,21,160,21,3,93,1,21,35,34,6,29,1,33,21,33,17,35,17,35,53,51,53,52,54,51,2,248,176,99,77,1,47,254,209,185,176,176,174,189,6,20,153,80,104,99,143,252,47,3,209,143,78,187,171,0,0,2,0,113,254,86,4,90,4,123,0,11,0,40,0,74,64,35,25,12,29,9,18,134,19,22,185,15,3,185,38,35,184,39,188,9,185,15,189,26,29,38,25,0,8,12,71,6,18,18,32,69,41,16,252,196,236,244,236,50,50,49,0,47,196,228,236,228,244,196,236,16,254,213,238,17,18,57,57,48,182,96,42,128,42,160,42,3,1,93,1,52,38,35,34,6,21,20,22,51,50,54,23,16,2,33,34,38,39,53,30,1,51,50,54,61,1,14,1,35,34,2,17,16,18,51,50,22,23,53,51,3,162,165,149,148,165,165,148,149,165,184,254,254,250,97,172,81,81,158,82,181,180,57,178,124,206,252,252,206,124,178,57,184,2,61,200,220,220,200,199,220,220,235,254,226,254,233,29,30,179,44,42,189,191,91,99,98,1,58,1,3,1,4,1,58,98,99,170,0,0,1,0,186,0,0,4,100,6,20,0,19,0,52,64,25,3,9,0,3,14,1,6,135,14,17,184,12,151,10,1,2,8,0,78,13,9,8,11,70,20,16,252,236,50,244,236,49,0,47,60,236,244,196,236,17,18,23,57,48,178,96,21,1,1,93,1,17,35,17,52,38,35,34,6,21,17,35,17,51,17,62,1,51,50,22,4,100,184,124,124,149,172,185,185,66,179,117,193,198,2,164,253,92,2,158,159,158,190,164,253,135,6,20,253,158,101,100,239,0,0,2,0,193,0,0,1,121,6,20,0,3,0,7,0,43,64,14,6,190,4,177,0,188,2,5,1,8,4,0,70,8,16,252,60,236,50,49,0,47,228,252,236,48,64,11,16,9,64,9,80,9,96,9,112,9,5,1,93,19,51,17,35,17,51,21,35,193,184,184,184,184,4,96,251,160,6,20,233,0,0,2,255,219,254,86,1,121,6,20,0,11,0,15,0,68,64,28,11,2,7,0,14,190,12,7,135,5,189,0,188,12,177,16,8,16,5,6,79,13,1,8,12,0,70,16,16,252,60,236,50,228,57,18,57,49,0,16,236,228,244,236,16,238,17,18,57,57,48,64,11,16,17,64,17,80,17,96,17,112,17,5,1,93,19,51,17,20,6,43,1,53,51,50,54,53,17,51,21,35,193,184,163,181,70,49,105,76,184,184,4,96,251,140,214,192,156,97,153,6,40,233,0,1,0,186,0,0,4,156,6,20,0,10,0,188,64,41,8,17,5,6,5,7,17,6,6,5,3,17,4,5,4,2,17,5,5,4,66,8,5,2,3,3,188,0,151,9,6,5,1,4,6,8,1,8,0,70,11,16,252,236,50,212,196,17,57,49,0,47,60,236,228,23,57,48,75,83,88,7,16,4,237,7,16,5,237,7,16,5,237,7,16,4,237,89,34,178,16,12,1,1,93,64,95,4,2,10,8,22,2,39,2,41,5,43,8,86,2,102,2,103,8,115,2,119,5,130,2,137,5,142,8,147,2,150,5,151,8,163,2,18,9,5,9,6,2,11,3,10,7,40,3,39,4,40,5,43,6,43,7,64,12,104,3,96,12,137,3,133,4,137,5,141,6,143,7,154,3,151,7,170,3,167,5,182,7,197,7,214,7,247,3,240,3,247,4,240,4,26,93,113,0,93,19,51,17,1,51,9,1,35,1,17,35,186,185,2,37,235,253,174,2,107,240,253,199,185,6,20,252,105,1,227,253,244,253,172,2,35,253,221,0,1,0,193,0,0,1,121,6,20,0,3,0,34,183,0,151,2,1,8,0,70,4,16,252,236,49,0,47,236,48,64,13,16,5,64,5,80,5,96,5,112,5,240,5,6,1,93,19,51,17,35,193,184,184,6,20,249,236,0,0,1,0,186,0,0,7,29,4,123,0,34,0,90,64,38,6,18,9,24,15,0,6,29,7,21,12,135,29,32,3,184,27,188,25,16,7,0,17,15,8,8,6,80,17,8,15,80,28,24,8,26,70,35,16,252,236,50,252,252,252,236,17,18,57,49,0,47,60,60,228,244,60,196,236,50,17,18,23,57,48,64,19,48,36,80,36,112,36,144,36,160,36,160,36,191,36,223,36,255,36,9,1,93,1,62,1,51,50,22,21,17,35,17,52,38,35,34,6,21,17,35,17,52,38,35,34,6,21,17,35,17,51,21,62,1,51,50,22,4,41,69,192,130,175,190,185,114,117,143,166,185,114,119,141,166,185,185,63,176,121,122,171,3,137,124,118,245,226,253,92,2,158,161,156,190,164,253,135,2,158,162,155,191,163,253,135,4,96,174,103,98,124,0,0,1,0,186,0,0,4,100,4,123,0,19,0,54,64,25,3,9,0,3,14,1,6,135,14,17,184,12,188,10,1,2,8,0,78,13,9,8,11,70,20,16,252,236,50,244,236,49,0,47,60,228,244,196,236,17,18,23,57,48,180,96,21,207,21,2,1,93,1,17,35,17,52,38,35,34,6,21,17,35,17,51,21,62,1,51,50,22,4,100,184,124,124,149,172,185,185,66,179,117,193,198,2,164,253,92,2,158,159,158,190,164,253,135,4,96,174,101,100,239,0,2,0,113,255,227,4,117,4,123,0,11,0,23,0,74,64,19,6,185,18,0,185,12,184,18,140,24,9,18,15,81,3,18,21,69,24,16,252,236,244,236,49,0,16,228,244,236,16,238,48,64,35,63,25,123,0,123,6,127,7,127,8,127,9,127,10,127,11,123,12,127,13,127,14,127,15,127,16,127,17,123,18,160,25,240,25,17,1,93,1,34,6,21,20,22,51,50,54,53,52,38,39,50,0,17,16,0,35,34,0,17,16,0,2,115,148,172,171,149,147,172,172,147,240,1,18,254,238,240,241,254,239,1,17,3,223,231,201,201,231,232,200,199,233,156,254,200,254,236,254,237,254,199,1,57,1,19,1,20,1,56,0,2,0,186,254,86,4,164,4,123,0,16,0,28,0,62,64,27,26,185,0,14,20,185,5,8,184,14,140,1,189,3,188,29,17,18,11,71,23,4,0,8,2,70,29,16,252,236,50,50,244,236,49,0,16,228,228,228,244,196,236,16,196,238,48,64,9,96,30,128,30,160,30,224,30,4,1,93,37,17,35,17,51,21,62,1,51,50,18,17,16,2,35,34,38,1,52,38,35,34,6,21,20,22,51,50,54,1,115,185,185,58,177,123,204,255,255,204,123,177,2,56,167,146,146,167,167,146,146,167,168,253,174,6,10,170,100,97,254,188,254,248,254,248,254,188,97,1,235,203,231,231,203,203,231,231,0,2,0,113,254,86,4,90,4,123,0,11,0,28,0,62,64,27,3,185,12,15,9,185,24,21,184,15,140,27,189,25,188,29,24,12,6,8,26,71,0,18,18,69,29,16,252,236,244,236,50,50,49,0,16,228,228,228,244,196,236,16,198,238,48,64,9,96,30,128,30,160,30,224,30,4,1,93,1,20,22,51,50,54,53,52,38,35,34,6,1,14,1,35,34,2,17,16,18,51,50,22,23,53,51,17,35,1,47,167,146,146,168,168,146,146,167,2,115,58,177,124,203,255,255,203,124,177,58,184,184,2,47,203,231,231,203,203,231,231,253,174,100,97,1,68,1,8,1,8,1,68,97,100,170,249,246,0,0,1,0,186,0,0,3,74,4,123,0,17,0,48,64,20,6,11,7,0,17,11,3,135,14,184,9,188,7,10,6,8,0,8,70,18,16,252,196,236,50,49,0,47,228,244,236,196,212,204,17,18,57,48,180,80,19,159,19,2,1,93,1,46,1,35,34,6,21,17,35,17,51,21,62,1,51,50,22,23,3,74,31,73,44,156,167,185,185,58,186,133,19,46,28,3,180,18,17,203,190,253,178,4,96,174,102,99,5,5,0,1,0,111,255,227,3,199,4,123,0,39,0,231,64,60,13,12,2,14,11,83,31,30,8,9,2,7,10,83,30,31,30,66,10,11,30,31,4,21,0,134,1,137,4,20,134,21,137,24,185,17,4,185,37,184,17,140,40,30,10,11,31,27,7,0,82,27,8,14,7,8,20,34,69,40,16,252,196,236,212,236,228,17,18,57,57,57,57,49,0,16,228,244,236,16,254,245,238,16,245,238,18,23,57,48,75,83,88,7,16,14,237,17,23,57,7,14,237,17,23,57,89,34,178,0,39,1,1,93,64,109,28,10,28,11,28,12,46,9,44,10,44,11,44,12,59,9,59,10,59,11,59,12,11,32,0,32,1,36,2,40,10,40,11,42,19,47,20,47,21,42,22,40,30,40,31,41,32,41,33,36,39,134,10,134,11,134,12,134,13,18,0,0,0,1,2,2,6,10,6,11,3,12,3,13,3,14,3,15,3,16,3,25,3,26,3,27,3,28,4,29,9,39,47,41,63,41,95,41,127,41,128,41,144,41,160,41,240,41,24,93,0,93,113,1,21,46,1,35,34,6,21,20,22,31,1,30,1,21,20,6,35,34,38,39,53,30,1,51,50,54,53,52,38,47,1,46,1,53,52,54,51,50,22,3,139,78,168,90,137,137,98,148,63,196,165,247,216,90,195,108,102,198,97,130,140,101,171,64,171,152,224,206,102,180,4,63,174,40,40,84,84,64,73,33,14,42,153,137,156,182,35,35,190,53,53,89,81,75,80,37,15,36,149,130,158,172,30,0,0,1,0,55,0,0,2,242,5,158,0,19,0,56,64,25,14,5,8,15,3,169,0,17,1,188,8,135,10,11,8,9,2,4,0,8,16,18,14,70,20,16,252,60,196,252,60,196,50,57,57,49,0,47,236,244,60,196,236,50,17,57,57,48,178,175,21,1,1,93,1,17,33,21,33,17,20,22,59,1,21,35,34,38,53,17,35,53,51,17,1,119,1,123,254,133,75,115,189,189,213,162,135,135,5,158,254,194,143,253,160,137,78,154,159,210,2,96,143,1,62,0,0,1,0,174,255,227,4,88,4,96,0,19,0,54,64,25,3,9,0,3,14,1,6,135,14,17,140,10,1,188,12,13,9,8,11,78,2,8,0,70,20,16,252,236,244,236,50,49,0,47,228,50,244,196,236,17,18,23,57,48,180,96,21,207,21,2,1,93,19,17,51,17,20,22,51,50,54,53,17,51,17,35,53,14,1,35,34,38,174,184,124,124,149,173,184,184,67,177,117,193,200,1,186,2,166,253,97,159,159,190,164,2,123,251,160,172,102,99,240,0,0,1,0,61,0,0,4,127,4,96,0,6,1,18,64,39,3,17,4,5,4,2,17,1,2,5,5,4,2,17,3,2,6,0,6,1,17,0,0,6,66,2,3,0,191,5,6,5,3,2,1,5,4,0,7,16,212,196,23,57,49,0,47,236,50,57,48,75,83,88,7,16,5,237,7,16,8,237,7,16,8,237,7,16,5,237,89,34,1,75,176,10,84,88,189,0,7,255,192,0,1,0,7,0,7,0,64,56,17,55,56,89,1,75,176,20,84,75,176,21,84,91,88,189,0,7,0,64,0,1,0,7,0,7,255,192,56,17,55,56,89,64,142,72,2,106,2,123,2,127,2,134,2,128,2,145,2,164,2,8,6,0,6,1,9,3,9,4,21,0,21,1,26,3,26,4,38,0,38,1,41,3,41,4,32,8,53,0,53,1,58,3,58,4,48,8,70,0,70,1,73,3,73,4,70,5,72,6,64,8,86,0,86,1,89,3,89,4,80,8,102,0,102,1,105,3,105,4,103,5,104,6,96,8,117,0,116,1,123,3,123,4,117,5,122,6,133,0,133,1,137,3,137,4,137,5,134,6,150,0,150,1,151,2,154,3,152,4,152,5,151,6,168,5,167,6,176,8,192,8,223,8,255,8,62,93,0,93,19,51,9,1,51,1,35,61,195,1,94,1,94,195,254,92,250,4,96,252,84,3,172,251,160,0,0,1,0,86,0,0,6,53,4,96,0,12,2,1,64,73,5,85,6,5,9,10,9,4,85,10,9,3,85,10,11,10,2,85,1,2,11,11,10,6,17,7,8,7,5,17,4,5,8,8,7,2,17,3,2,12,0,12,1,17,0,0,12,66,10,5,2,3,6,3,0,191,11,8,12,11,10,9,8,6,5,4,3,2,1,11,7,0,13,16,212,204,23,57,49,0,47,60,236,50,50,23,57,48,75,83,88,7,16,5,237,7,16,8,237,7,16,8,237,7,16,5,237,7,16,8,237,7,16,5,237,7,5,237,7,16,8,237,89,34,1,75,176,10,84,75,176,17,84,91,75,176,18,84,91,75,176,19,84,91,75,176,11,84,91,88,189,0,13,255,192,0,1,0,13,0,13,0,64,56,17,55,56,89,1,75,176,12,84,75,176,13,84,91,75,176,16,84,91,88,189,0,13,0,64,0,1,0,13,0,13,255,192,56,17,55,56,89,64,255,5,2,22,2,22,5,34,10,53,10,73,2,73,5,70,10,64,10,91,2,91,5,85,10,80,10,110,2,110,5,102,10,121,2,127,2,121,5,127,5,135,2,153,2,152,5,148,10,188,2,188,5,206,2,199,3,207,5,29,5,2,9,3,6,4,11,5,10,8,11,9,4,11,5,12,21,2,25,3,22,4,26,5,27,8,27,9,20,11,21,12,37,0,37,1,35,2,39,3,33,4,37,5,34,6,34,7,37,8,39,9,36,10,33,11,35,12,57,3,54,4,54,8,57,12,48,14,70,2,72,3,70,4,64,4,66,5,64,6,64,7,64,8,68,9,68,10,68,11,64,14,64,14,86,0,86,1,86,2,80,4,81,5,82,6,82,7,80,8,83,9,84,10,85,11,99,0,100,1,101,2,106,3,101,4,106,5,106,6,106,7,110,9,97,11,103,12,111,14,117,0,117,1,121,2,125,3,120,4,125,5,122,6,127,6,122,7,127,7,120,8,121,9,127,9,123,10,118,11,125,12,135,2,136,5,143,14,151,0,151,1,148,2,147,3,156,4,155,5,152,6,152,7,153,8,64,47,150,12,159,14,166,0,166,1,164,2,164,3,171,4,171,5,169,6,169,7,171,8,164,12,175,14,181,2,177,3,189,4,187,5,184,9,191,14,196,2,195,3,204,4,202,5,121,93,0,93,19,51,27,1,51,27,1,51,1,35,11,1,35,86,184,230,229,217,230,229,184,254,219,217,241,242,217,4,96,252,150,3,106,252,150,3,106,251,160,3,150,252,106,0,1,0,59,0,0,4,121,4,96,0,11,1,90,64,70,5,17,6,7,6,4,17,3,4,7,7,6,4,17,5,4,1,2,1,3,17,2,2,1,11,17,0,1,0,10,17,9,10,1,1,0,10,17,11,10,7,8,7,9,17,8,8,7,66,10,7,4,1,4,8,0,191,5,2,10,7,4,1,4,8,0,2,8,6,12,16,212,196,212,196,17,23,57,49,0,47,60,236,50,23,57,48,75,83,88,7,16,5,237,7,16,8,237,7,16,8,237,7,16,5,237,7,16,5,237,7,16,8,237,7,16,8,237,7,16,5,237,89,34,1,75,176,10,84,75,176,15,84,91,75,176,16,84,91,75,176,17,84,91,88,189,0,12,255,192,0,1,0,12,0,12,0,64,56,17,55,56,89,1,75,176,20,84,88,189,0,12,0,64,0,1,0,12,0,12,255,192,56,17,55,56,89,64,152,10,4,4,10,26,4,21,10,38,10,61,4,49,10,85,4,87,7,88,10,102,10,118,1,122,4,118,7,116,10,141,4,130,10,153,4,159,4,151,7,146,10,144,10,166,1,169,4,175,4,165,7,163,10,160,10,28,10,3,4,5,5,9,10,11,26,3,21,5,21,9,26,11,41,3,38,5,37,9,42,11,32,13,58,1,57,3,55,5,52,7,54,9,57,11,48,13,73,3,70,5,69,9,74,11,64,13,89,0,86,1,89,2,89,3,87,5,86,6,89,7,86,8,86,9,89,11,80,13,111,13,120,1,127,13,155,1,148,7,171,1,164,7,176,13,207,13,223,13,255,13,47,93,0,93,9,2,35,9,1,35,9,1,51,9,1,4,100,254,107,1,170,217,254,186,254,186,217,1,179,254,114,217,1,41,1,41,4,96,253,223,253,193,1,184,254,72,2,74,2,22,254,113,1,143,0,1,0,61,254,86,4,127,4,96,0,15,1,162,64,67,7,8,2,9,17,0,15,10,17,11,10,0,0,15,14,17,15,0,15,13,17,12,13,0,0,15,13,17,14,13,10,11,10,12,17,11,11,10,66,13,11,9,16,0,11,5,135,3,189,14,11,188,16,14,13,12,10,9,6,3,0,8,15,4,15,11,16,16,212,196,196,17,23,57,49,0,16,228,50,244,236,17,57,17,57,18,57,48,75,83,88,7,16,5,237,7,16,8,237,7,16,8,237,7,16,5,237,7,16,8,237,7,5,237,23,50,89,34,1,75,176,10,84,75,176,8,84,91,88,189,0,16,255,192,0,1,0,16,0,16,0,64,56,17,55,56,89,1,75,176,20,84,88,189,0,16,0,64,0,1,0,16,0,16,255,192,56,17,55,56,89,64,240,6,0,5,8,6,9,3,13,22,10,23,13,16,13,35,13,53,13,73,10,79,10,78,13,90,9,90,10,106,10,135,13,128,13,147,13,18,10,0,10,9,6,11,5,12,11,14,11,15,23,1,21,2,16,4,16,5,23,10,20,11,20,12,26,14,26,15,39,0,36,1,36,2,32,4,32,5,41,8,40,9,37,10,36,11,36,12,39,13,42,14,42,15,32,17,55,0,53,1,53,2,48,4,48,5,56,10,54,11,54,12,56,13,57,14,57,15,48,17,65,0,64,1,64,2,64,3,64,4,64,5,64,6,64,7,64,8,66,9,69,10,71,13,73,14,73,15,64,17,84,0,81,1,81,2,85,3,80,4,80,5,86,6,85,7,86,8,87,9,87,10,85,11,85,12,89,14,89,15,80,17,102,1,102,2,104,10,105,14,105,15,96,17,123,8,120,14,120,15,137,0,138,9,133,11,133,12,137,13,137,14,137,15,153,9,149,11,149,12,154,14,154,15,164,11,164,12,171,14,171,15,176,17,207,17,223,17,255,17,101,93,0,93,5,14,1,43,1,53,51,50,54,63,1,1,51,9,1,51,2,147,78,148,124,147,108,76,84,51,33,254,59,195,1,94,1,94,195,104,200,122,154,72,134,84,4,78,252,148,3,108,0,1,0,88,0,0,3,219,4,96,0,9,0,180,64,26,8,17,2,3,2,3,17,7,8,7,66,8,169,0,188,3,169,5,8,3,1,0,4,1,6,10,16,220,196,50,196,17,57,57,49,0,47,236,244,236,48,75,83,88,7,16,5,237,7,16,5,237,89,34,1,75,176,11,84,75,176,12,84,91,88,189,0,10,0,64,0,1,0,10,0,10,255,192,56,17,55,56,89,1,75,176,19,84,88,189,0,10,255,192,0,1,0,10,0,10,0,64,56,17,55,56,89,64,66,5,2,22,2,38,2,71,2,73,7,5,11,8,15,11,24,3,27,8,43,8,32,11,54,3,57,8,48,11,64,1,64,2,69,3,64,4,64,5,67,8,87,3,89,8,95,11,96,1,96,2,102,3,96,4,96,5,98,8,127,11,128,11,175,11,27,93,0,93,19,33,21,1,33,21,33,53,1,33,113,3,106,253,76,2,180,252,125,2,180,253,101,4,96,168,252,219,147,168,3,37,0,1,1,0,254,178,4,23,6,20,0,36,0,130,64,52,25,15,21,11,6,37,9,26,16,21,29,11,5,32,33,3,0,11,169,9,0,169,1,192,9,21,169,19,177,37,12,9,10,5,36,22,25,0,29,10,5,19,2,20,0,32,25,67,10,15,5,37,16,212,60,196,252,60,196,50,57,57,17,18,57,17,18,57,57,17,18,57,57,49,0,16,252,236,196,244,236,16,238,18,23,57,18,57,17,57,57,17,18,57,17,18,57,57,48,1,75,176,12,84,88,189,0,37,255,192,0,1,0,37,0,37,0,64,56,17,55,56,89,178,0,38,1,93,5,21,35,34,38,61,1,52,38,43,1,53,51,50,54,61,1,52,54,59,1,21,35,34,6,29,1,20,6,7,30,1,29,1,20,22,51,4,23,62,249,169,108,142,61,61,143,107,169,249,62,68,141,86,91,110,111,90,86,141,190,144,148,221,239,151,116,143,115,149,240,221,147,143,88,141,248,157,142,25,27,142,156,248,141,88,0,1,1,4,254,29,1,174,6,29,0,3,0,18,183,1,0,177,4,0,5,2,4,16,212,236,49,0,16,252,204,48,1,17,35,17,1,174,170,6,29,248,0,8,0,0,0,1,1,0,254,178,4,23,6,20,0,36,0,158,64,54,31,37,27,22,12,15,8,27,11,21,25,15,4,5,32,3,0,25,169,27,0,169,35,192,27,15,169,17,177,37,28,25,26,21,15,1,4,0,8,26,21,35,18,4,0,26,31,21,67,16,0,11,4,37,16,212,60,196,50,252,60,196,17,18,57,57,17,18,57,17,18,57,57,17,18,57,57,49,0,16,252,236,196,244,236,16,238,18,23,57,17,18,57,57,17,57,17,57,57,17,18,57,48,1,75,176,10,84,88,189,0,37,0,64,0,1,0,37,0,37,255,192,56,17,55,56,89,1,75,176,14,84,88,189,0,37,255,192,0,1,0,37,0,37,0,64,56,17,55,56,89,178,0,38,1,93,5,51,50,54,61,1,52,54,55,46,1,61,1,52,38,43,1,53,51,50,22,29,1,20,22,59,1,21,35,34,6,29,1,20,6,43,1,1,0,70,140,85,90,111,111,90,85,140,70,63,249,167,108,142,62,62,142,108,167,249,63,190,86,143,248,156,142,27,25,142,157,248,142,87,143,147,221,240,149,115,143,116,151,239,221,148,0,1,0,217,1,211,5,219,3,49,0,29,0,35,64,16,1,16,27,12,0,19,4,156,27,19,156,12,30,0,15,30,16,212,196,49,0,16,212,252,212,236,16,192,17,18,57,57,48,1,21,6,6,35,34,39,38,39,38,39,38,35,34,6,7,53,54,54,51,50,23,22,23,22,23,22,51,50,54,5,219,105,179,97,110,146,11,5,7,15,155,94,88,172,98,105,179,97,110,147,10,5,8,14,155,94,86,169,3,49,178,79,68,59,4,2,3,5,62,77,83,178,79,69,60,4,2,3,5,62,76,0,255,255,0,16,0,0,5,104,7,78,2,39,0,36,0,0,0,0,0,7,1,3,0,188,1,117,0,3,0,16,0,0,5,104,7,109,0,11,0,14,0,33,0,203,64,84,12,17,13,12,27,28,27,14,17,28,27,30,17,28,27,29,17,28,28,27,13,17,33,15,33,12,17,14,12,15,15,33,32,17,15,33,31,17,33,15,33,66,12,27,15,13,9,3,193,21,9,30,149,13,9,142,32,28,30,29,28,24,32,31,33,13,18,6,14,24,12,6,27,0,86,24,28,15,6,86,18,28,33,34,16,212,196,212,236,50,16,212,238,50,17,57,17,57,17,18,57,17,57,57,17,18,57,57,49,0,47,60,230,214,238,16,212,238,17,18,57,57,57,48,75,83,88,7,16,5,237,7,5,237,7,16,8,237,7,16,5,237,7,16,5,237,7,5,237,7,5,237,7,16,8,237,89,34,178,32,35,1,1,93,64,32,26,12,115,12,155,12,3,7,15,8,27,80,35,102,13,105,14,117,13,123,14,121,28,121,29,118,32,118,33,128,35,12,93,0,93,1,52,38,35,34,6,21,20,22,51,50,54,3,1,33,1,46,1,53,52,54,51,50,22,21,20,6,7,1,35,3,33,3,35,3,84,89,63,64,87,88,63,63,89,152,254,240,2,33,254,88,61,62,159,115,114,161,63,60,2,20,210,136,253,95,136,213,6,90,63,89,87,65,63,88,88,254,243,253,25,3,78,41,115,73,115,160,161,114,70,118,41,250,139,1,127,254,129,0,255,255,0,115,254,117,5,39,5,240,2,39,0,38,0,0,0,0,0,7,0,221,1,45,0,0,255,255,0,201,0,0,4,139,7,107,2,39,0,40,0,0,0,0,0,7,1,4,0,158,1,117,255,255,0,201,0,0,5,51,7,94,2,39,0,49,0,0,0,0,0,7,1,5,0,254,1,117,255,255,0,115,255,227,5,217,7,78,2,39,0,50,0,0,0,0,0,7,1,3,1,39,1,117,255,255,0,178,255,227,5,41,7,78,2,39,0,56,0,0,0,0,0,7,1,3,0,238,1,117,255,255,0,123,255,227,4,45,6,102,2,39,0,68,0,0,0,0,0,7,0,141,0,82,0,0,255,255,0,123,255,227,4,45,6,102,2,39,0,68,0,0,0,0,0,7,0,67,0,82,0,0,255,255,0,123,255,227,4,45,6,102,2,39,0,68,0,0,0,0,0,7,0,215,0,82,0,0,255,255,0,123,255,227,4,45,6,16,2,39,0,68,0,0,0,0,0,7,0,142,0,82,0,0,255,255,0,123,255,227,4,45,6,55,2,39,0,68,0,0,0,0,0,7,0,216,0,82,0,0,255,255,0,123,255,227,4,45,7,6,2,39,0,68,0,0,0,0,0,7,0,220,0,82,0,0,255,255,0,113,254,117,3,231,4,123,2,39,0,70,0,0,0,0,0,7,0,221,0,143,0,0,255,255,0,113,255,227,4,127,6,102,2,39,0,72,0,0,0,0,0,7,0,141,0,139,0,0,255,255,0,113,255,227,4,127,6,102,2,39,0,72,0,0,0,0,0,7,0,67,0,139,0,0,255,255,0,113,255,227,4,127,6,102,2,39,0,72,0,0,0,0,0,7,0,215,0,139,0,0,255,255,0,113,255,227,4,127,6,16,2,39,0,72,0,0,0,0,0,7,0,142,0,139,0,0,255,255,0,144,0,0,2,111,6,102,2,39,0,214,0,0,0,0,0,7,0,141,255,29,0,0,255,255,255,199,0,0,1,166,6,102,2,39,0,214,0,0,0,0,0,7,0,67,255,29,0,0,255,255,255,222,0,0,2,92,6,102,2,39,0,214,0,0,0,0,0,7,0,215,255,29,0,0,255,255,255,244,0,0,2,70,6,16,2,39,0,214,0,0,0,0,0,7,0,142,255,29,0,0,255,255,0,186,0,0,4,100,6,55,2,39,0,81,0,0,0,0,0,7,0,216,0,152,0,0,255,255,0,113,255,227,4,117,6,102,2,39,0,82,0,0,0,0,0,7,0,141,0,115,0,0,255,255,0,113,255,227,4,117,6,102,2,39,0,82,0,0,0,0,0,7,0,67,0,115,0,0,255,255,0,113,255,227,4,117,6,102,2,39,0,82,0,0,0,0,0,7,0,215,0,115,0,0,255,255,0,113,255,227,4,117,6,16,2,39,0,82,0,0,0,0,0,7,0,142,0,115,0,0,255,255,0,113,255,227,4,117,6,55,2,39,0,82,0,0,0,0,0,7,0,216,0,115,0,0,255,255,0,174,255,227,4,88,6,102,2,39,0,88,0,0,0,0,0,7,0,141,0,123,0,0,255,255,0,174,255,227,4,88,6,102,2,39,0,88,0,0,0,0,0,7,0,67,0,123,0,0,255,255,0,174,255,227,4,88,6,102,2,39,0,88,0,0,0,0,0,7,0,215,0,123,0,0,255,255,0,174,255,227,4,88,6,16,2,39,0,88,0,0,0,0,0,7,0,142,0,123,0,0,0,1,0,57,255,59,3,199,5,213,0,11,0,39,64,20,8,4,185,10,2,0,129,6,194,12,3,89,5,1,87,9,89,7,0,12,16,212,60,236,252,60,236,49,0,16,228,244,212,60,236,50,48,1,51,17,33,21,33,17,35,17,33,53,33,1,168,176,1,111,254,145,176,254,145,1,111,5,213,254,92,153,251,163,4,93,153,0,0,2,0,195,3,117,3,61,5,240,0,11,0,26,0,32,64,17,6,195,21,196,0,195,12,145,27,9,90,18,91,3,90,24,27,16,220,236,252,236,49,0,16,244,236,252,236,48,1,34,6,21,20,22,51,50,54,53,52,38,39,50,22,23,22,22,21,20,6,35,34,38,53,52,54,2,0,80,110,110,80,80,110,111,79,64,118,43,46,46,185,134,135,180,184,5,111,111,80,79,109,109,79,79,112,129,49,46,45,114,66,132,183,180,135,134,186,0,0,2,0,172,254,199,4,35,5,152,0,6,0,33,0,81,64,43,19,22,20,0,15,12,1,11,7,134,8,136,11,16,134,15,136,12,185,20,22,11,185,29,31,28,184,22,140,34,28,21,0,9,30,19,11,15,7,4,18,25,34,16,220,236,212,60,212,60,60,236,50,50,49,0,16,228,244,60,196,236,16,196,254,244,238,16,245,238,18,57,17,18,57,17,18,57,48,37,17,6,6,21,20,22,1,21,38,38,39,3,54,54,55,21,6,6,7,17,35,17,38,0,17,16,0,55,17,51,19,22,22,2,166,147,164,164,2,16,74,136,68,1,70,137,72,65,137,77,102,241,254,247,1,9,241,102,1,73,137,131,3,88,18,226,184,185,226,3,161,172,41,42,3,252,160,5,42,39,170,30,35,7,254,228,1,32,20,1,51,1,1,1,2,1,50,22,1,31,254,225,4,33,0,1,0,129,0,0,4,98,5,240,0,27,0,96,64,33,7,22,8,1,134,0,18,10,169,20,8,12,4,160,0,148,25,145,16,12,160,14,0,13,9,11,7,28,19,15,21,17,28,16,220,60,204,204,252,60,196,212,196,49,0,47,236,50,244,228,236,16,212,60,238,50,16,238,17,57,57,48,1,75,176,12,84,88,189,0,28,255,192,0,1,0,28,0,28,0,64,56,17,55,56,89,180,54,1,54,2,2,0,93,1,21,46,1,35,34,6,29,1,33,21,33,17,33,21,33,53,51,17,35,53,51,53,16,54,51,50,22,4,78,76,136,61,148,116,1,135,254,121,2,45,252,31,236,199,199,214,232,61,151,5,180,182,41,41,155,212,215,143,254,47,170,170,1,209,143,238,1,5,243,31,0,0,2,0,92,255,61,3,162,5,240,0,11,0,62,0,145,64,60,47,48,42,6,0,23,29,48,54,4,13,39,138,38,13,138,12,42,198,38,197,35,16,198,12,197,60,145,35,63,47,6,0,23,48,4,19,29,45,9,54,3,19,87,57,45,87,32,9,87,12,34,26,57,38,34,3,87,51,63,16,220,236,228,196,212,228,236,212,236,16,238,17,57,17,18,57,17,23,57,57,49,0,16,196,244,228,236,16,230,238,16,238,16,238,17,23,57,57,57,17,18,57,48,1,75,176,10,84,75,176,11,84,91,75,176,12,84,91,75,176,14,84,91,88,189,0,63,0,64,0,1,0,63,0,63,255,192,56,17,55,56,89,1,14,1,21,20,22,23,62,1,53,52,38,19,21,46,1,35,34,6,21,20,23,22,23,30,1,21,20,6,7,30,1,21,20,6,35,34,38,39,53,30,1,51,50,54,53,52,47,1,46,1,53,52,54,55,46,1,53,52,54,51,50,22,1,123,63,62,139,250,63,62,143,204,83,143,56,97,108,206,26,14,211,131,92,93,62,57,204,173,73,154,88,87,148,58,102,113,221,25,214,128,93,91,59,59,200,166,73,153,3,168,46,90,46,76,133,135,45,91,46,75,136,2,147,164,39,39,80,71,90,115,15,8,119,154,101,90,140,53,52,109,64,142,168,29,29,164,39,39,84,76,102,123,14,120,153,102,91,143,49,44,112,69,130,159,29,0,0,1,1,51,1,209,3,133,4,33,0,11,0,18,183,9,199,3,12,6,92,0,12,16,212,236,49,0,16,212,236,48,1,52,54,51,50,22,21,20,6,35,34,38,1,51,173,126,124,171,172,125,125,172,2,250,124,171,171,124,125,172,172,0,0,1,0,158,255,59,4,57,5,213,0,13,0,37,64,18,8,2,4,193,0,129,6,2,14,0,7,93,5,3,93,1,11,14,16,212,212,252,220,236,57,49,0,16,196,50,244,236,17,57,48,1,33,17,35,17,35,17,35,17,38,38,53,52,36,2,121,1,192,141,190,142,215,235,1,4,5,213,249,102,6,31,249,225,3,78,17,221,184,190,232,0,0,1,0,186,255,227,4,172,6,20,0,47,0,154,64,48,45,39,33,12,4,6,13,32,0,4,42,22,134,23,26,185,19,42,185,3,151,19,140,46,12,9,13,29,32,33,39,9,8,36,39,8,6,29,8,36,16,22,45,8,16,0,70,48,16,252,196,252,204,16,198,238,212,238,16,238,17,57,57,18,57,18,57,49,0,47,228,254,238,16,254,213,238,18,23,57,23,57,48,64,64,15,5,15,6,15,7,15,39,15,40,138,12,138,13,7,10,6,10,7,10,11,10,12,10,13,10,31,13,32,10,33,12,34,4,38,25,13,25,31,25,32,58,32,58,33,77,31,77,32,73,33,73,34,106,31,106,32,165,6,165,7,166,32,24,93,1,93,19,52,54,51,50,22,23,14,1,21,20,22,31,1,30,1,21,20,6,35,34,38,39,53,30,1,51,50,54,53,52,38,47,1,46,1,53,52,54,55,46,1,35,34,6,21,17,35,186,239,218,208,219,3,151,168,58,65,57,166,96,225,211,64,136,73,80,140,65,116,120,59,101,92,96,87,167,151,8,131,113,130,136,187,4,113,200,219,232,224,8,115,96,47,81,42,37,106,142,100,172,183,25,24,164,30,29,95,91,63,84,62,55,59,135,91,127,172,29,103,112,139,131,251,147,0,0,4,1,27,0,0,6,229,5,205,0,23,0,47,0,56,0,76,0,96,64,54,69,66,67,63,50,201,72,48,201,57,74,67,202,12,57,202,0,201,24,200,12,201,36,72,69,51,48,4,49,66,60,63,57,54,73,49,96,75,54,96,67,60,94,18,9,30,75,94,6,9,30,95,42,77,16,220,228,252,236,16,254,253,196,238,16,238,50,17,57,57,18,57,18,23,57,49,0,47,238,246,254,237,16,237,50,16,238,214,238,57,18,57,57,48,1,34,6,7,6,6,21,20,22,23,22,22,51,50,54,55,54,54,53,52,38,39,38,38,39,50,4,23,22,18,21,20,2,7,6,4,35,34,36,39,38,2,53,52,18,55,54,36,19,35,17,51,50,54,53,52,38,39,50,22,21,20,6,7,22,22,23,23,35,39,38,38,35,35,17,35,17,4,0,131,226,94,94,96,96,94,94,226,131,132,227,94,93,93,94,92,94,227,132,152,1,7,109,109,108,108,109,109,254,249,152,152,254,249,109,109,108,108,109,109,1,7,125,123,123,110,87,88,102,176,174,105,96,24,67,46,137,172,129,59,73,54,66,155,5,102,94,94,94,229,130,129,227,94,94,95,95,94,93,226,131,133,227,93,94,94,103,110,109,109,254,250,154,152,254,251,109,109,110,110,109,109,1,5,152,154,1,6,109,109,110,254,98,254,236,62,75,76,63,103,119,121,86,112,17,8,77,73,223,209,96,51,254,156,3,68,0,3,1,27,0,0,6,229,5,205,0,23,0,47,0,73,0,67,64,38,61,203,62,58,204,65,202,36,49,203,48,52,204,71,202,24,201,0,200,36,201,12,55,97,68,61,48,94,42,9,6,68,94,30,9,6,18,74,16,220,204,252,236,16,254,237,50,16,238,49,0,47,238,246,254,253,238,214,238,16,253,238,214,238,48,1,50,4,23,22,18,21,20,2,7,6,4,35,34,36,39,38,2,53,52,18,55,54,36,23,34,6,7,6,6,21,20,22,23,22,22,51,50,54,55,54,54,53,52,38,39,38,38,23,21,38,38,35,34,6,21,20,22,51,50,54,55,21,6,6,35,34,38,53,52,54,51,50,22,4,0,152,1,7,109,109,108,108,109,109,254,249,152,152,254,249,109,109,108,108,109,109,1,7,152,131,226,94,94,96,96,94,94,226,131,132,227,94,93,93,94,92,94,227,167,66,130,66,149,167,171,155,64,122,66,67,137,70,216,251,251,216,73,136,5,205,110,109,109,254,250,154,152,254,251,109,109,110,110,109,109,1,5,152,154,1,6,109,109,110,103,94,94,94,229,130,129,227,94,94,95,95,94,93,226,131,133,227,93,94,94,245,129,33,32,175,157,159,174,31,34,127,29,28,244,208,209,242,28,0,0,2,1,39,3,147,6,70,5,213,0,12,0,20,0,62,64,33,1,6,7,16,10,4,18,14,9,3,6,201,13,2,0,129,21,1,9,5,98,3,9,98,11,13,99,15,98,19,99,17,21,16,212,228,252,228,212,236,212,236,17,57,49,0,16,244,60,60,236,23,50,212,60,60,196,17,57,48,1,19,19,51,17,35,17,3,35,3,17,35,17,35,21,35,17,35,17,35,53,4,74,174,164,170,113,195,55,203,114,113,203,114,201,5,213,255,0,1,0,253,190,1,228,254,209,1,47,254,28,2,66,94,254,28,1,228,94,0,0,1,1,115,4,238,3,82,6,102,0,3,0,49,64,9,2,180,0,179,4,3,68,1,4,16,212,236,49,0,16,244,236,48,0,75,176,9,84,75,176,14,84,91,88,189,0,4,255,192,0,1,0,4,0,4,0,64,56,17,55,56,89,1,51,1,35,2,139,199,254,186,153,6,102,254,136,0,0,2,0,215,5,70,3,41,6,16,0,3,0,7,0,146,64,14,6,2,206,4,0,205,8,1,100,0,5,100,4,8,16,220,252,212,236,49,0,16,252,60,236,50,48,0,75,176,10,84,75,176,13,84,91,88,189,0,8,0,64,0,1,0,8,0,8,255,192,56,17,55,56,89,1,75,176,12,84,75,176,13,84,91,75,176,14,84,91,75,176,23,84,91,88,189,0,8,255,192,0,1,0,8,0,8,0,64,56,17,55,56,89,1,75,176,15,84,75,176,25,84,91,88,189,0,8,0,64,0,1,0,8,0,8,255,192,56,17,55,56,89,64,17,96,1,96,2,96,5,96,6,112,1,112,2,112,5,112,6,8,93,1,51,21,35,37,51,21,35,2,94,203,203,254,121,203,203,6,16,202,202,202,0,0,1,0,217,0,39,5,219,4,221,0,19,0,62,64,34,13,12,10,3,2,207,4,0,156,6,12,207,14,10,156,18,6,16,8,20,18,14,13,12,8,4,3,2,8,9,5,15,0,20,16,220,60,196,50,23,57,49,0,16,212,60,204,50,252,60,236,16,254,60,236,57,17,18,57,48,19,33,1,23,7,33,21,33,7,33,21,33,1,39,55,33,53,33,55,33,217,3,4,1,0,125,174,1,47,254,72,195,2,123,252,250,254,254,125,174,254,213,1,182,195,253,135,3,162,1,59,102,213,168,240,170,254,199,102,211,170,240,0,2,0,8,0,0,7,72,5,213,0,15,0,19,0,135,64,57,17,17,14,15,14,16,17,15,15,14,13,17,15,14,12,17,14,15,14,66,5,149,3,11,149,17,1,149,16,149,0,129,17,7,149,3,173,13,9,17,16,15,13,12,5,14,10,0,4,8,6,2,28,18,10,14,20,16,212,212,60,236,50,212,196,196,17,18,23,57,49,0,47,60,236,236,196,244,236,236,16,238,16,238,48,75,83,88,7,16,5,237,7,5,237,7,16,5,237,7,16,5,237,89,34,178,128,21,1,1,93,64,19,103,17,119,16,119,17,134,12,133,16,150,17,144,21,160,21,191,21,9,93,1,21,33,17,33,21,33,17,33,21,33,17,33,3,35,1,23,1,33,17,7,53,253,27,2,199,253,57,2,248,252,61,253,240,160,205,2,113,139,254,182,1,203,5,213,170,254,70,170,253,227,170,1,127,254,129,5,213,158,252,240,3,16,0,3,0,102,255,186,5,229,6,23,0,9,0,19,0,43,0,158,64,60,29,31,26,13,43,44,19,10,1,0,4,13,41,38,32,20,13,4,42,38,30,26,4,149,38,13,149,26,145,38,140,44,43,44,42,20,23,16,32,30,35,19,10,1,0,4,29,41,16,7,31,7,25,35,51,16,25,23,16,44,16,252,236,252,236,192,17,18,57,57,23,57,18,57,57,17,18,57,57,17,57,49,0,16,228,244,236,16,238,16,192,16,192,17,18,57,57,18,57,18,23,57,18,57,17,18,57,57,48,64,42,87,0,90,21,87,25,85,33,106,21,101,33,123,21,118,28,117,33,9,70,19,89,0,86,19,106,0,100,19,100,28,106,40,124,0,115,19,118,28,122,40,11,93,1,93,9,1,30,1,51,50,0,17,52,38,39,46,1,35,34,0,17,20,22,23,7,38,2,53,16,0,33,50,22,23,55,23,7,22,18,21,16,0,33,34,38,39,7,39,4,182,253,51,62,161,95,220,1,1,39,121,61,161,95,220,254,253,39,39,134,78,79,1,121,1,59,130,221,87,162,102,170,78,80,254,136,254,198,128,221,91,162,103,4,88,252,178,64,67,1,72,1,26,112,184,184,64,67,254,184,254,229,112,188,68,158,102,1,8,160,1,98,1,165,77,75,191,89,198,103,254,246,158,254,159,254,91,75,75,191,88,0,3,0,221,0,221,5,207,3,238,0,11,0,23,0,47,0,255,64,29,45,27,21,9,33,3,0,36,24,4,21,15,39,33,21,27,15,33,48,12,0,36,24,18,6,42,18,30,48,16,212,196,212,196,17,57,57,57,57,49,0,16,212,196,212,196,16,192,17,18,23,57,18,57,17,18,57,48,64,190,5,2,5,3,5,4,0,5,0,6,0,7,5,8,5,9,5,10,10,16,15,17,15,18,15,19,10,20,21,2,21,3,21,4,16,5,16,6,16,7,21,8,21,9,21,10,26,14,26,15,26,16,31,17,31,18,31,19,26,20,26,21,26,22,36,2,36,3,36,4,32,5,32,6,32,7,36,8,36,9,36,10,42,14,42,15,42,16,47,17,47,18,47,19,42,20,42,21,42,22,53,2,53,3,53,4,48,5,48,6,48,7,53,8,53,9,53,10,58,14,58,15,58,16,63,17,63,18,63,19,58,20,58,21,58,22,69,2,69,3,69,4,64,5,64,6,64,7,69,8,69,9,69,10,74,14,74,15,74,16,79,17,79,18,79,19,74,20,74,21,74,22,86,180,31,176,32,176,33,176,34,176,38,176,39,176,40,180,41,8,93,1,93,1,30,1,51,50,54,53,52,38,35,34,6,7,46,1,35,34,6,21,20,22,51,50,54,23,14,1,35,34,38,53,52,54,51,50,22,23,62,1,51,50,22,21,20,6,35,34,38,3,147,49,134,84,101,128,118,89,82,133,196,49,133,85,102,127,118,89,82,134,144,70,157,94,136,186,167,134,95,153,72,68,158,97,134,188,167,134,94,149,2,47,88,90,135,105,101,134,135,55,88,88,132,106,101,134,136,22,135,127,223,166,175,216,126,138,138,131,225,167,175,214,119,0,0,2,0,217,0,0,5,219,5,4,0,11,0,15,0,46,64,24,5,208,7,3,156,0,208,9,1,12,156,14,13,2,21,4,0,23,12,8,21,10,6,16,16,212,60,236,50,252,60,236,50,49,0,47,236,212,60,236,252,60,236,48,1,17,33,21,33,17,35,17,33,53,33,17,1,33,21,33,3,174,2,45,253,211,168,253,211,2,45,253,211,5,2,250,254,5,4,254,125,170,254,125,1,131,170,1,131,251,166,170,0,2,0,217,0,0,5,219,4,168,0,6,0,10,0,84,64,46,2,156,3,4,3,1,156,0,1,4,4,3,1,156,2,1,5,6,5,0,156,6,5,66,5,4,2,1,0,5,3,209,6,167,7,156,9,1,8,2,0,36,7,4,35,11,16,252,60,236,50,50,57,49,0,47,236,244,236,23,57,48,75,83,88,7,4,237,7,16,8,237,7,16,8,237,7,16,4,237,89,34,9,2,21,1,53,1,1,33,21,33,5,219,252,64,3,192,250,254,5,2,250,254,5,2,250,254,3,248,254,235,254,238,178,1,112,170,1,111,252,2,170,0,0,2,0,217,0,0,5,219,4,168,0,6,0,10,0,86,64,47,6,156,0,6,3,4,3,5,156,4,4,3,0,156,1,2,1,6,156,5,6,2,2,1,66,6,5,3,2,0,5,4,209,1,167,7,156,8,6,7,2,36,9,4,0,35,11,16,252,60,60,236,50,57,49,0,47,236,244,236,23,57,48,75,83,88,7,16,8,237,7,16,4,237,7,16,4,237,7,16,8,237,89,34,19,53,1,21,1,53,1,1,21,33,53,217,5,2,250,254,3,193,1,65,250,254,3,248,176,254,145,170,254,144,178,1,18,253,199,170,170,0,0,1,0,82,0,0,4,195,5,213,0,24,0,198,64,70,16,2,17,22,17,15,2,14,15,22,22,17,15,2,16,15,8,13,8,14,2,13,13,8,66,15,11,9,4,0,211,23,6,18,11,211,20,9,16,13,129,2,12,9,14,3,5,22,15,3,21,18,16,3,0,17,102,19,0,101,1,28,13,102,10,5,101,7,3,25,16,212,60,236,50,236,252,236,50,236,18,23,57,18,57,57,17,23,57,49,0,47,228,50,212,60,236,50,212,60,236,50,17,18,57,48,75,83,88,7,16,5,237,7,16,8,237,7,16,8,237,7,16,5,237,89,34,1,75,176,12,84,88,189,0,25,255,192,0,1,0,25,0,25,0,64,56,17,55,56,89,64,40,134,15,144,15,166,15,160,15,181,15,5,39,12,39,13,39,14,41,16,40,17,40,18,55,14,57,16,135,12,136,18,166,13,165,14,170,16,169,17,14,93,0,93,1,33,17,35,17,33,53,33,53,39,33,53,33,1,51,9,1,51,1,33,21,33,7,21,33,4,141,254,99,201,254,96,1,160,84,254,180,1,8,254,195,190,1,123,1,121,191,254,194,1,8,254,181,84,1,159,1,199,254,57,1,199,123,51,155,123,2,74,253,68,2,188,253,182,123,155,51,0,0,1,0,174,254,86,4,229,4,96,0,32,0,77,64,37,19,25,31,3,22,6,3,9,12,3,1,18,15,6,135,28,22,140,10,1,188,0,189,33,25,9,18,9,8,11,78,31,2,8,0,70,33,16,252,236,50,244,236,196,18,57,49,0,16,228,228,50,244,60,236,220,196,17,23,57,17,18,23,57,48,182,31,34,96,34,207,34,3,1,93,19,17,51,17,20,22,51,50,54,53,17,51,17,20,22,51,50,54,55,21,14,1,35,34,38,39,14,1,35,34,38,39,17,174,184,138,135,148,149,184,35,37,9,32,28,41,73,35,69,82,15,50,145,98,102,143,42,254,86,6,10,253,72,145,148,168,168,2,141,252,162,60,57,11,12,148,23,22,78,80,79,79,78,78,253,215,0,0,2,0,104,255,231,3,193,5,45,0,29,0,41,0,98,64,25,0,39,33,9,27,6,39,21,6,15,33,27,15,21,213,42,12,36,3,0,30,18,36,24,42,16,212,204,220,204,57,57,17,57,49,0,16,228,204,220,204,16,206,16,206,17,18,57,17,18,57,48,1,75,176,12,84,75,176,11,84,91,75,176,14,84,91,75,176,16,84,91,75,176,20,84,91,88,189,0,42,0,64,0,1,0,42,0,42,255,192,56,17,55,56,89,1,62,1,53,52,38,35,34,6,35,34,38,53,52,54,51,50,18,17,16,0,35,34,38,53,52,18,51,50,22,7,52,38,35,34,2,21,20,22,51,50,18,2,244,15,15,73,72,55,144,36,36,48,144,101,180,214,254,223,213,152,203,221,162,101,130,11,87,79,109,141,86,80,109,141,2,109,87,163,75,129,131,116,44,31,62,98,254,202,254,249,254,177,254,70,216,163,198,1,1,91,224,116,125,254,254,207,116,123,1,4,0,0,1,0,25,254,119,5,59,5,193,0,11,0,93,64,20,10,4,12,2,5,7,2,0,7,12,10,5,4,3,1,0,6,6,8,12,16,212,196,23,57,49,0,16,196,212,204,16,206,17,18,57,57,48,64,48,81,3,86,5,80,5,90,10,115,3,112,3,118,4,117,5,112,5,122,10,128,3,128,5,12,90,9,127,2,127,3,112,5,112,6,123,9,116,11,143,2,143,3,128,5,128,6,11,93,1,93,19,33,21,33,9,1,33,21,33,53,9,1,55,4,234,252,65,2,160,253,74,3,239,250,222,2,213,253,73,5,193,193,253,51,253,4,192,149,3,33,2,227,0,0,1,0,156,254,119,5,113,5,193,0,7,0,30,64,15,6,2,215,4,214,0,175,8,3,103,1,5,103,0,8,16,212,236,212,236,49,0,16,252,236,236,50,48,19,33,17,35,17,33,17,35,156,4,213,240,253,10,239,5,193,248,182,6,125,249,131,0,0,1,255,225,255,240,4,170,4,47,0,35,0,203,64,49,11,2,21,31,30,3,0,8,218,15,26,22,0,217,34,216,15,213,24,12,30,27,26,25,24,23,6,36,18,1,0,11,2,4,35,22,21,5,34,31,18,12,35,18,104,5,35,31,36,16,212,212,212,236,18,57,17,18,57,18,57,57,17,23,57,17,18,23,57,49,0,47,60,228,244,236,50,50,16,238,17,23,57,57,57,48,1,75,176,10,84,88,189,0,36,255,192,0,1,0,36,0,36,0,64,56,17,55,56,89,64,86,24,30,24,31,2,9,0,9,1,13,2,13,3,15,4,15,5,15,6,15,7,15,8,15,9,15,10,15,11,15,12,15,13,15,14,15,15,15,16,15,17,15,18,15,19,14,20,13,21,9,22,11,23,8,24,15,24,13,25,8,26,9,35,17,0,17,1,22,2,22,3,23,20,22,21,17,22,23,23,28,24,28,25,17,35,40,93,0,93,1,35,3,14,1,21,20,22,51,50,54,55,7,14,1,35,34,38,53,52,54,55,19,33,3,35,19,35,34,6,7,35,62,1,51,33,4,135,182,105,15,15,47,55,17,46,37,30,30,55,26,118,121,21,34,80,254,186,194,181,195,41,54,60,9,160,28,143,165,3,121,3,145,254,25,74,92,22,58,49,5,5,141,8,8,102,100,46,144,161,1,120,252,111,3,145,64,69,166,125,0,0,1,0,47,254,141,3,250,6,14,0,37,0,38,64,20,32,219,0,26,13,219,19,26,220,7,177,38,10,105,23,106,29,105,4,38,16,220,236,252,236,49,0,16,252,236,220,228,16,222,228,48,1,50,19,54,55,18,18,51,50,22,21,20,6,35,34,38,39,38,38,35,34,3,6,7,2,2,35,34,38,53,52,54,51,50,22,23,22,22,1,55,106,14,2,1,12,190,202,80,100,64,55,42,56,12,6,9,16,107,14,4,4,17,189,196,79,101,68,61,33,48,15,10,10,254,250,2,176,108,57,2,3,1,188,84,65,54,63,38,35,15,72,253,149,193,110,254,33,254,98,83,65,56,63,29,28,18,83,0,0,3,0,115,1,213,3,59,5,240,0,3,0,30,0,41,0,95,64,51,40,7,37,4,31,18,24,16,2,227,0,31,221,16,0,225,37,221,5,10,25,223,24,222,21,221,10,224,28,145,42,0,24,13,31,16,34,6,2,1,40,17,6,107,4,108,24,34,107,13,42,16,220,236,204,252,236,50,50,192,192,17,18,57,57,17,18,57,49,0,16,244,228,252,244,236,16,196,238,237,214,238,16,238,17,18,57,18,57,17,57,57,48,19,33,21,33,1,17,35,53,6,6,35,34,38,53,52,54,51,51,53,52,38,35,34,6,7,53,54,54,51,50,22,5,34,6,21,20,22,51,50,54,53,53,139,2,176,253,80,2,174,149,44,144,93,128,152,191,188,182,117,117,62,136,68,73,145,69,183,179,254,236,161,126,98,82,104,130,2,80,123,2,184,254,64,112,63,68,135,113,135,138,4,91,91,34,34,127,28,28,176,240,67,79,64,77,144,114,29,0,3,0,96,1,213,3,100,5,240,0,3,0,15,0,27,0,46,64,25,2,227,0,225,22,221,10,224,16,221,4,145,28,0,19,13,1,25,107,7,108,19,107,13,28,16,220,236,252,236,57,17,18,57,49,0,16,244,236,244,236,252,236,48,19,33,21,33,1,50,22,21,20,6,35,34,38,53,52,54,23,34,6,21,20,22,51,50,54,53,52,38,139,2,176,253,80,1,88,179,206,206,179,179,208,208,179,105,126,127,104,105,125,124,2,80,123,4,27,221,191,191,219,220,190,191,221,115,161,136,133,160,160,133,137,160,0,1,0,78,0,0,5,207,5,231,0,31,0,64,64,34,9,229,25,145,18,15,3,3,0,229,16,1,17,32,22,19,15,12,31,6,2,1,0,2,109,6,28,28,15,109,12,28,22,32,16,212,236,236,212,236,236,192,192,17,18,57,17,18,57,17,18,57,49,0,47,60,236,23,50,244,236,48,37,21,33,53,54,18,53,52,0,35,34,0,21,20,18,23,21,33,53,33,38,2,53,16,0,33,32,0,17,20,2,7,5,207,253,168,177,198,254,248,216,216,254,247,199,178,253,168,1,63,158,145,1,127,1,49,1,47,1,129,142,161,178,178,178,97,1,76,202,240,1,34,254,221,239,202,254,180,97,178,178,139,1,42,184,1,62,1,138,254,119,254,203,194,254,216,141,0,0,3,0,123,255,227,7,111,4,123,0,6,0,51,0,62,1,3,64,67,39,45,37,61,14,13,0,52,169,37,22,134,21,136,18,0,169,14,58,18,185,28,25,46,134,45,186,42,3,185,14,187,7,49,10,184,31,25,140,37,63,52,55,38,6,15,0,37,55,28,7,38,15,21,0,8,13,61,38,8,15,45,55,8,34,69,63,16,252,236,204,212,252,60,212,236,196,17,18,57,57,17,57,17,18,57,17,18,57,49,0,16,196,228,50,244,60,196,228,252,60,244,236,16,196,238,50,16,238,16,244,238,16,238,17,57,17,57,17,18,57,48,64,129,48,43,48,44,48,45,48,46,48,47,48,48,64,43,64,44,64,45,64,46,64,47,64,48,80,43,80,44,80,45,80,46,80,47,80,48,133,43,133,48,128,64,144,64,160,64,176,64,192,64,208,64,224,64,224,64,240,64,29,63,0,63,6,63,13,63,14,63,15,5,48,44,48,45,48,46,48,47,64,44,64,45,64,46,64,47,80,44,80,45,80,46,80,47,111,0,111,6,111,13,111,14,111,15,96,44,96,45,96,46,96,47,112,44,112,45,112,46,112,47,128,44,128,45,128,46,128,47,29,93,113,1,93,1,46,1,35,34,6,7,3,62,1,51,50,0,29,1,33,30,1,51,50,54,55,21,14,1,35,34,38,39,14,1,35,34,38,53,52,54,51,33,53,52,38,35,34,6,7,53,62,1,51,50,22,3,34,6,21,20,22,51,50,54,61,1,6,182,1,165,137,153,185,14,68,74,212,132,226,1,8,252,178,12,204,183,104,200,100,100,208,106,167,248,77,73,216,143,189,210,253,251,1,2,167,151,96,182,84,101,190,90,142,213,239,223,172,129,111,153,185,2,148,151,180,174,158,1,48,90,94,254,221,250,90,191,200,53,53,174,42,44,121,119,120,120,187,168,189,192,18,127,139,46,46,170,39,39,96,254,24,102,123,98,115,217,180,41,0,3,0,72,255,162,4,156,4,188,0,9,0,19,0,43,0,228,64,60,43,44,38,31,29,26,19,10,1,0,4,13,41,38,32,20,13,4,42,38,30,26,4,185,38,13,185,26,184,38,140,44,43,44,42,20,23,16,32,30,35,19,10,1,0,4,16,7,31,29,7,18,35,81,41,16,18,23,69,44,16,252,236,50,244,236,50,192,17,18,23,57,18,57,57,17,18,57,57,17,57,49,0,16,228,244,236,16,238,16,192,16,192,17,18,57,57,18,57,18,23,57,17,57,57,17,18,57,48,64,112,40,1,63,45,89,20,86,28,85,29,86,32,106,21,102,33,127,0,123,4,127,5,127,6,127,7,127,8,127,9,127,10,127,11,127,12,123,13,122,21,123,26,127,27,127,28,127,29,127,30,127,31,127,32,123,33,127,34,127,35,127,36,127,37,123,38,155,25,149,37,168,25,160,45,240,45,38,89,0,86,19,85,29,90,40,105,0,102,19,101,28,106,40,122,0,116,19,118,28,122,40,137,30,149,24,154,36,162,24,173,36,17,93,1,93,9,1,30,1,51,50,54,53,52,38,39,46,1,35,34,6,21,20,22,23,7,46,1,53,16,0,51,50,22,23,55,23,7,30,1,21,16,0,35,34,38,39,7,39,3,137,254,25,41,103,65,147,172,20,92,42,103,62,151,169,19,20,125,54,54,1,17,241,93,159,67,139,95,146,53,54,254,238,240,96,161,63,139,96,3,33,253,176,42,40,232,200,79,117,154,41,41,235,211,72,110,46,151,77,197,119,1,20,1,56,51,52,168,79,179,77,198,120,254,237,254,199,52,51,168,78,0,2,0,143,255,227,3,172,5,213,0,32,0,36,0,134,64,47,32,26,5,2,4,6,25,0,16,134,15,136,12,0,33,131,35,12,149,19,140,35,129,37,6,34,25,22,9,5,1,0,26,34,9,0,28,1,34,28,33,38,15,9,28,22,37,16,220,236,212,252,236,212,236,17,18,57,17,18,57,17,18,57,18,57,49,0,16,228,244,236,16,254,205,16,244,238,18,57,57,23,57,48,1,75,176,16,84,75,176,18,84,91,75,176,19,84,91,88,189,0,37,255,192,0,1,0,37,0,37,0,64,56,17,55,56,89,64,11,116,4,116,5,116,6,116,7,118,28,5,93,1,51,21,20,6,15,1,14,1,21,20,22,51,50,54,55,21,14,1,35,34,38,53,52,54,63,1,62,1,55,62,1,53,19,35,53,51,1,244,190,55,90,90,58,51,131,109,78,180,96,94,192,103,184,224,73,89,88,48,38,8,7,6,196,202,202,4,68,156,101,130,87,88,53,94,49,89,110,70,67,188,57,56,194,159,76,137,86,86,47,53,25,21,60,54,1,14,254,0,2,1,53,0,0,2,0,5,213,0,3,0,9,0,98,64,15,7,0,131,2,129,4,8,7,4,0,3,5,1,0,10,16,252,60,236,50,57,57,49,0,47,244,252,204,48,1,75,176,11,84,88,189,0,10,0,64,0,1,0,10,0,10,255,192,56,17,55,56,89,1,75,176,15,84,75,176,16,84,91,75,176,19,84,91,88,189,0,10,255,192,0,1,0,10,0,10,0,64,56,17,55,56,89,182,0,11,32,11,80,11,3,93,1,35,53,51,17,35,17,19,51,19,2,0,203,203,203,21,162,20,4,215,254,250,43,2,143,1,101,254,155,0,0,1,0,217,1,31,5,219,3,94,0,5,0,23,64,10,4,156,2,0,6,3,23,1,0,6,16,220,212,236,49,0,16,212,196,236,48,19,33,17,35,17,33,217,5,2,168,251,166,3,94,253,193,1,149,0,0,1,0,61,255,215,5,25,6,125,0,10,0,42,64,24,10,9,8,7,6,5,11,2,4,2,0,11,10,9,7,6,5,4,3,0,8,1,8,11,16,212,204,23,57,49,0,16,212,204,196,17,18,23,57,48,1,51,21,35,1,35,1,7,39,37,1,4,92,189,115,253,174,66,254,193,125,25,1,27,1,0,6,125,96,249,186,3,115,45,80,98,253,59,0,1,0,31,254,86,5,2,6,20,0,35,0,138,64,64,14,13,2,15,12,17,25,30,25,11,10,9,8,4,7,17,30,30,25,66,25,12,19,10,7,30,1,26,10,169,8,19,138,18,230,22,169,15,1,138,0,230,4,169,33,151,15,28,8,36,30,29,27,26,25,12,11,9,8,7,10,0,18,36,16,212,204,23,57,49,0,16,196,50,196,252,236,244,236,16,238,246,238,16,238,50,18,57,57,17,18,57,57,48,75,83,88,7,16,5,237,23,50,7,16,5,237,17,23,57,89,34,1,75,176,12,84,88,189,0,36,255,192,0,1,0,36,0,36,0,64,56,17,55,56,89,1,21,46,1,35,34,6,7,3,33,21,33,3,2,6,35,34,38,39,53,30,1,51,50,54,55,19,35,53,33,19,62,1,51,50,22,5,2,38,80,44,96,114,25,60,1,31,254,195,127,58,188,186,58,100,47,52,97,47,97,109,34,137,248,1,23,63,36,198,151,53,100,5,240,164,29,28,122,132,254,201,143,253,133,254,227,211,21,22,166,33,33,137,166,2,173,143,1,74,183,195,18,0,2,0,217,1,16,5,219,3,244,0,29,0,59,0,63,64,31,46,31,57,42,0,45,34,19,1,16,27,12,30,42,156,49,57,156,34,4,156,27,12,156,19,60,30,0,45,15,60,16,212,60,196,50,49,0,16,212,236,212,236,220,252,212,236,192,17,18,57,57,17,18,57,57,17,18,57,57,48,1,21,6,6,35,34,39,38,39,38,39,38,35,34,6,7,53,54,54,51,50,23,22,23,22,23,22,51,50,54,19,21,6,6,35,34,39,38,39,38,39,38,35,34,6,7,53,54,54,51,50,23,22,23,22,23,22,51,50,54,5,219,105,179,97,110,146,10,7,6,15,155,94,88,172,98,105,179,97,110,147,11,5,6,15,155,94,86,169,103,105,179,97,110,146,10,7,6,15,155,94,88,172,98,105,179,97,110,147,10,5,7,15,155,94,86,169,2,111,179,78,69,59,4,3,2,6,61,76,84,179,78,69,59,5,2,2,6,61,75,1,218,178,79,69,59,4,3,2,6,61,76,83,178,78,69,59,4,2,3,6,61,75,0,2,255,250,0,0,5,96,5,193,0,2,0,6,0,56,64,15,0,3,1,3,5,4,3,2,1,0,5,7,5,6,7,16,212,204,17,23,57,49,0,47,196,204,17,57,48,64,20,99,1,109,2,112,1,120,2,127,2,121,5,118,6,7,110,0,127,0,2,93,1,93,9,1,33,1,51,1,33,2,172,254,94,3,68,253,239,224,2,67,250,154,4,238,251,196,5,15,250,63,0,2,0,158,0,141,4,37,4,35,0,6,0,13,0,134,64,73,3,232,4,5,4,2,232,1,2,5,5,4,2,232,3,2,6,0,6,1,232,0,0,6,10,232,11,12,11,9,232,8,9,12,12,11,9,232,10,9,13,7,13,8,232,7,7,13,66,9,2,11,4,231,7,0,166,14,9,12,5,2,7,3,0,111,5,10,7,111,12,110,14,16,252,252,60,212,236,50,17,57,17,18,57,49,0,16,244,60,236,50,57,57,48,75,83,88,7,16,4,237,7,16,8,237,7,16,8,237,7,16,4,237,7,16,4,237,7,16,8,237,7,16,8,237,7,16,4,237,89,34,1,21,1,1,21,1,53,19,21,1,1,21,1,53,4,37,254,211,1,45,254,43,35,254,211,1,45,254,43,4,35,191,254,244,254,244,191,1,162,82,1,162,191,254,244,254,244,191,1,162,82,0,0,2,0,193,0,141,4,72,4,35,0,6,0,13,0,134,64,73,12,232,13,12,9,10,9,11,232,10,10,9,13,232,7,8,7,12,232,11,12,8,8,7,5,232,6,5,2,3,2,4,232,3,3,2,6,232,0,1,0,5,232,4,5,1,1,0,66,12,5,10,3,231,7,0,166,14,12,8,1,5,0,8,111,10,7,1,111,3,0,112,14,16,252,60,252,212,60,236,18,57,17,18,57,49,0,16,244,60,236,50,57,57,48,75,83,88,7,16,8,237,7,16,4,237,7,16,4,237,7,16,8,237,7,16,8,237,7,16,4,237,7,16,4,237,7,16,8,237,89,34,19,1,21,1,53,1,1,37,1,21,1,53,1,1,193,1,213,254,43,1,45,254,211,1,178,1,213,254,43,1,45,254,211,4,35,254,94,82,254,94,191,1,12,1,12,191,254,94,82,254,94,191,1,12,1,12,0,3,0,236,0,0,7,20,0,254,0,3,0,7,0,11,0,35,64,17,8,4,0,131,10,6,2,4,25,5,0,25,1,9,25,8,12,16,212,252,212,236,212,236,49,0,47,60,60,236,50,50,48,37,51,21,35,37,51,21,35,37,51,21,35,3,150,212,212,2,169,213,213,250,173,213,213,254,254,254,254,254,254,0,255,255,0,16,0,0,5,104,7,107,2,39,0,36,0,0,0,0,0,7,1,6,0,188,1,117,255,255,0,16,0,0,5,104,7,94,2,39,0,36,0,0,0,0,0,7,1,5,0,188,1,117,255,255,0,115,255,227,5,217,7,94,2,39,0,50,0,0,0,0,0,7,1,5,1,39,1,117,0,2,0,115,0,0,8,12,5,213,0,16,0,25,0,59,64,31,5,149,3,17,1,149,0,129,24,7,149,3,173,9,24,18,16,10,21,6,2,28,17,0,4,8,21,25,13,16,26,16,252,236,212,196,196,212,236,50,18,57,57,57,57,49,0,47,236,236,50,244,236,50,16,238,48,1,21,33,17,33,21,33,17,33,21,33,32,0,17,16,0,33,23,35,32,0,17,16,0,33,51,7,250,253,26,2,199,253,57,2,248,251,215,254,79,254,65,1,191,1,177,103,129,254,191,254,192,1,64,1,65,129,5,213,170,254,70,170,253,227,170,1,124,1,112,1,109,1,124,170,254,225,254,224,254,223,254,223,0,3,0,113,255,227,7,195,4,123,0,6,0,39,0,51,0,132,64,49,7,8,0,16,134,15,136,12,0,169,8,46,12,185,22,19,40,3,185,8,187,34,37,31,184,25,19,140,52,6,0,22,34,49,9,15,0,8,7,75,49,18,9,81,43,18,28,69,52,16,252,236,244,252,244,236,196,17,18,57,57,18,57,49,0,16,228,50,244,60,196,228,236,50,16,196,238,50,16,238,16,244,238,17,18,57,48,64,37,63,53,95,53,112,53,159,53,207,53,208,53,240,53,7,63,0,63,6,63,7,63,8,63,9,5,111,0,111,6,111,7,111,8,111,9,5,93,113,1,93,1,46,1,35,34,6,7,5,21,33,30,1,51,50,54,55,21,14,1,35,34,38,39,14,1,35,34,0,17,16,0,51,50,22,23,62,1,51,50,0,37,34,6,21,20,22,51,50,54,53,52,38,7,10,2,164,137,153,185,14,3,72,252,178,12,204,183,106,200,98,100,208,106,160,242,81,71,209,140,241,254,239,1,17,241,140,211,66,78,232,143,226,1,8,250,176,148,172,171,149,147,172,172,2,148,152,179,174,158,53,90,190,199,52,52,174,42,44,110,109,110,109,1,57,1,19,1,20,1,56,111,108,107,112,254,221,135,231,201,201,231,232,200,199,233,0,0,1,0,0,1,233,4,0,2,121,0,3,0,16,182,2,169,0,233,4,1,0,47,198,49,0,16,252,236,48,17,33,21,33,4,0,252,0,2,121,144,0,0,1,0,0,1,233,8,0,2,121,0,3,0,15,181,2,169,0,4,1,0,47,204,49,0,16,212,236,48,17,33,21,33,8,0,248,0,2,121,144,0,2,0,174,3,233,3,109,5,213,0,5,0,11,0,39,64,19,6,0,158,9,3,129,12,9,10,6,25,7,3,4,7,0,25,1,12,16,220,252,204,212,204,16,254,212,206,49,0,16,244,60,236,50,48,1,35,53,19,51,3,5,35,53,19,51,3,1,129,211,164,129,82,1,154,211,164,129,82,3,233,173,1,63,254,193,173,173,1,63,254,193,0,2,0,174,3,233,3,109,5,213,0,5,0,11,0,39,64,19,9,3,158,6,0,129,12,9,10,7,25,6,1,3,4,1,25,0,12,16,220,236,212,204,16,220,238,212,206,49,0,16,244,60,236,50,48,1,51,21,3,35,19,37,51,21,3,35,19,1,0,211,164,129,82,1,154,211,164,129,82,5,213,172,254,192,1,64,172,172,254,192,1,64,0,1,0,174,3,233,1,211,5,213,0,5,0,24,64,11,0,158,3,129,6,3,4,0,25,1,6,16,220,252,212,204,49,0,16,244,236,48,1,35,53,19,51,3,1,129,211,164,129,82,3,233,173,1,63,254,193,0,0,1,0,178,3,254,1,215,5,213,0,5,0,24,64,11,3,158,0,129,6,3,4,1,113,0,6,16,220,236,212,204,49,0,16,244,236,48,1,51,21,3,35,19,1,4,211,164,129,82,5,213,152,254,193,1,63,0,0,3,0,217,0,150,5,219,4,111,0,3,0,7,0,11,0,41,64,20,0,234,2,6,234,4,2,8,156,4,10,12,9,5,1,114,4,0,8,12,16,220,212,60,252,60,196,49,0,16,212,196,252,196,16,238,16,238,48,1,51,21,35,17,51,21,35,1,33,21,33,2,223,246,246,246,246,253,250,5,2,250,254,4,111,246,254,18,245,2,65,170,0,2,0,6,254,35,3,238,6,117,0,3,0,7,0,34,64,17,2,6,0,8,4,6,8,6,4,3,2,1,0,6,5,7,8,16,212,204,23,57,49,0,16,212,204,17,57,18,57,48,9,7,1,250,254,127,1,129,1,129,254,127,1,244,254,12,254,12,5,129,252,207,252,199,3,57,4,37,251,219,251,211,4,45,255,255,0,61,254,86,4,127,6,16,2,39,0,92,0,0,0,0,0,7,0,142,0,94,0,0,255,255,255,252,0,0,4,231,7,78,2,39,0,60,0,0,0,0,0,7,1,3,0,115,1,117,0,1,254,137,255,227,2,205,5,240,0,3,0,43,64,19,0,15,1,2,1,2,15,3,0,3,66,2,140,0,145,4,1,3,4,16,212,204,49,0,16,228,228,48,75,83,88,7,16,5,237,7,16,5,237,89,34,1,51,1,35,2,45,160,252,92,160,5,240,249,243,0,0,2,0,94,0,82,4,188,4,178,0,35,0,47,0,131,64,73,3,9,27,21,4,45,30,0,39,28,2,33,29,12,18,45,20,11,10,3,19,15,1,29,45,185,19,235,15,236,39,185,29,235,33,48,30,12,0,18,4,42,36,20,48,28,21,27,42,29,19,28,24,9,3,36,11,10,1,3,2,36,40,2,115,6,116,42,40,28,115,24,48,16,220,228,236,244,228,236,18,23,57,18,57,57,17,18,57,57,18,57,57,17,18,57,17,18,23,57,49,0,16,212,228,236,244,228,236,16,192,17,18,23,57,18,57,57,17,18,57,57,17,57,57,18,23,57,48,1,55,23,7,22,22,21,20,6,7,23,7,39,6,6,35,34,38,39,7,39,55,38,38,53,52,54,55,39,55,23,54,54,51,50,22,19,52,38,35,34,6,21,20,22,51,50,54,3,123,207,114,206,37,36,38,40,209,114,207,59,116,61,58,120,61,207,113,207,37,37,38,38,207,115,207,55,116,64,60,117,92,155,114,112,158,157,113,113,156,3,225,209,115,206,59,119,62,63,115,57,207,113,207,40,38,37,37,207,115,206,62,118,58,64,116,56,206,115,207,39,37,36,254,124,112,154,154,112,114,156,157,0,0,1,0,158,0,141,2,115,4,35,0,6,0,71,64,37,3,232,4,5,4,2,232,1,2,5,5,4,2,232,3,2,6,0,6,1,232,0,6,66,2,4,231,0,166,7,2,3,0,111,5,110,7,16,252,236,50,57,49,0,16,244,236,57,48,75,83,88,7,4,237,7,16,8,237,7,16,8,237,7,16,4,237,89,34,1,21,1,1,21,1,53,2,115,254,211,1,45,254,43,4,35,191,254,244,254,244,191,1,162,82,0,0,1,0,193,0,141,2,150,4,35,0,6,0,73,64,38,5,232,6,5,2,3,2,4,232,3,3,2,6,232,0,1,0,5,232,4,5,1,1,0,66,5,3,231,0,166,7,5,1,111,3,0,112,7,16,252,60,236,57,49,0,16,244,236,57,48,75,83,88,7,16,8,237,7,16,4,237,7,16,4,237,7,16,8,237,89,34,19,1,21,1,53,1,1,193,1,213,254,43,1,45,254,211,4,35,254,94,82,254,94,191,1,12,1,12,0,0,2,0,47,0,0,4,74,6,20,0,21,0,25,0,155,64,40,11,20,24,7,3,169,0,16,135,14,24,190,22,177,14,151,9,0,188,5,1,17,14,15,4,22,2,8,0,15,20,4,8,8,23,0,10,6,76,26,16,252,60,196,50,196,252,60,196,16,238,50,17,18,57,57,49,0,47,60,230,50,238,254,238,16,238,16,238,50,18,57,57,48,1,75,176,10,84,88,189,0,26,255,192,0,1,0,26,0,26,0,64,56,17,55,56,89,1,75,176,14,84,88,189,0,26,0,64,0,1,0,26,0,26,255,192,56,17,55,56,89,64,26,16,22,16,23,16,24,16,25,4,48,27,80,27,128,15,128,16,128,27,160,27,208,27,239,27,8,93,0,93,1,17,35,17,33,17,35,17,35,53,51,53,52,54,59,1,21,35,34,6,29,1,1,51,21,35,4,74,185,254,7,185,176,176,173,179,185,176,99,77,1,249,185,185,4,96,251,160,3,209,252,47,3,209,143,78,183,175,153,80,104,99,1,178,233,0,1,0,47,0,0,4,74,6,20,0,21,0,132,64,33,8,19,4,15,11,169,9,4,135,0,151,17,9,188,13,2,5,0,10,8,3,8,1,10,12,8,8,16,1,18,14,76,22,16,252,60,196,196,252,60,196,16,238,17,18,57,57,49,0,47,60,230,50,254,238,16,238,50,18,57,57,48,1,75,176,10,84,88,189,0,22,255,192,0,1,0,22,0,22,0,64,56,17,55,56,89,1,75,176,14,84,88,189,0,22,0,64,0,1,0,22,0,22,255,192,56,17,55,56,89,64,17,48,23,80,23,128,10,128,11,128,23,160,23,208,23,239,23,8,93,1,33,17,35,17,33,34,6,29,1,33,21,33,17,35,17,35,53,51,53,52,54,2,74,2,0,185,254,183,99,77,1,47,254,209,185,176,176,174,6,20,249,236,5,123,80,104,99,143,252,47,3,209,143,78,187,171,0,0,1,0,57,255,59,3,199,5,213,0,19,0,62,64,32,18,6,185,0,16,8,185,10,4,0,2,14,10,12,129,2,194,20,15,0,89,17,13,1,87,9,5,89,11,7,3,20,16,212,60,60,236,50,252,60,60,236,50,49,0,16,228,244,196,50,16,196,50,16,238,50,16,238,50,48,37,33,17,35,17,33,53,33,17,33,53,33,17,51,17,33,21,33,17,33,3,199,254,145,176,254,145,1,111,254,145,1,111,176,1,111,254,145,1,111,223,254,92,1,164,154,2,31,153,1,164,254,92,153,253,225,0,1,0,219,2,72,1,174,3,70,0,3,0,18,183,2,131,0,4,1,25,0,4,16,212,236,49,0,16,212,236,48,19,51,21,35,219,211,211,3,70,254,0,1,0,174,255,18,1,211,0,254,0,5,0,24,64,11,3,158,0,131,6,3,4,1,25,0,6,16,212,236,212,204,49,0,16,252,236,48,37,51,21,3,35,19,1,0,211,164,129,82,254,172,254,192,1,64,0,2,0,174,255,18,3,109,0,254,0,5,0,11,0,39,64,19,9,3,158,6,0,131,12,3,4,1,25,0,7,9,10,7,25,6,12,16,220,236,212,204,16,220,238,212,206,49,0,16,252,60,236,50,48,37,51,21,3,35,19,37,51,21,3,35,19,2,154,211,164,129,82,254,102,211,164,129,82,254,172,254,192,1,64,172,172,254,192,1,64,0,0,7,0,113,255,227,10,76,5,240,0,11,0,23,0,35,0,39,0,51,0,63,0,75,0,174,64,68,36,15,37,38,37,38,15,39,36,39,66,64,0,146,12,46,146,30,141,40,146,24,70,6,146,52,12,141,58,38,18,140,36,24,145,76,37,73,67,39,49,43,67,13,61,9,13,15,14,3,13,21,49,13,27,61,14,73,13,21,55,43,13,27,14,33,11,76,16,252,228,236,212,196,236,228,16,238,16,238,246,238,16,238,17,18,57,17,18,57,49,0,16,228,50,244,60,60,228,50,236,50,16,238,246,238,16,238,50,48,75,83,88,7,16,5,237,7,16,5,237,89,34,1,75,176,20,84,75,176,9,84,91,75,176,11,84,91,75,176,12,84,91,75,176,13,84,91,75,176,14,84,91,88,189,0,76,0,64,0,1,0,76,0,76,255,192,56,17,55,56,89,1,34,6,21,20,22,51,50,54,53,52,38,39,50,22,21,20,6,35,34,38,53,52,54,1,50,22,21,20,6,35,34,38,53,52,54,33,51,1,35,19,34,6,21,20,22,51,50,54,53,52,38,1,50,22,21,20,6,35,34,38,53,52,54,23,34,6,21,20,22,51,50,54,53,52,38,8,244,87,100,100,87,85,99,99,85,158,186,187,157,160,186,187,249,116,158,188,187,159,159,185,186,4,37,160,252,90,160,31,86,99,98,87,87,99,100,3,178,158,186,187,157,160,186,187,159,87,99,99,87,85,99,99,2,145,148,132,130,149,149,130,131,149,127,220,187,187,219,219,187,188,219,2,224,219,187,189,218,219,188,186,220,249,243,5,142,149,130,132,148,148,132,129,150,253,159,220,187,187,219,219,187,188,219,127,148,132,130,149,149,130,131,149,0,255,255,0,16,0,0,5,104,7,109,2,39,0,36,0,0,0,0,0,7,1,7,0,188,1,117,255,255,0,201,0,0,4,139,7,109,2,39,0,40,0,0,0,0,0,7,1,7,0,158,1,117,255,255,0,16,0,0,5,104,7,107,2,39,0,36,0,0,0,0,0,7,1,4,0,188,1,117,255,255,0,201,0,0,4,139,7,78,2,39,0,40,0,0,0,0,0,7,1,3,0,158,1,117,255,255,0,201,0,0,4,139,7,107,2,39,0,40,0,0,0,0,0,7,1,6,0,158,1,117,255,255,0,162,0,0,2,31,7,107,2,39,0,44,0,0,0,0,0,7,1,4,255,47,1,117,255,255,255,254,0,0,2,96,7,109,2,39,0,44,0,0,0,0,0,7,1,7,255,47,1,117,255,255,0,6,0,0,2,88,7,78,2,39,0,44,0,0,0,0,0,7,1,3,255,47,1,117,255,255,0,59,0,0,1,186,7,107,2,39,0,44,0,0,0,0,0,7,1,6,255,47,1,117,255,255,0,115,255,227,5,217,7,107,2,39,0,50,0,0,0,0,0,7,1,4,1,39,1,117,255,255,0,115,255,227,5,217,7,109,2,39,0,50,0,0,0,0,0,7,1,7,1,39,1,117,255,255,0,115,255,227,5,217,7,107,2,39,0,50,0,0,0,0,0,7,1,6,1,39,1,117,255,255,0,178,255,227,5,41,7,107,2,39,0,56,0,0,0,0,0,7,1,4,0,238,1,117,255,255,0,178,255,227,5,41,7,109,2,39,0,56,0,0,0,0,0,7,1,7,0,238,1,117,255,255,0,178,255,227,5,41,7,107,2,39,0,56,0,0,0,0,0,7,1,6,0,238,1,117,0,1,0,193,0,0,1,121,4,96,0,3,0,32,183,0,191,2,1,8,0,70,4,16,252,236,49,0,47,236,48,64,11,16,5,64,5,80,5,96,5,112,5,5,1,93,19,51,17,35,193,184,184,4,96,251,160,0,0,1,0,193,4,238,3,63,6,102,0,6,0,55,64,12,4,5,2,180,0,179,7,4,2,117,6,7,16,220,236,57,49,0,16,244,236,50,57,48,0,75,176,9,84,75,176,14,84,91,88,189,0,7,255,192,0,1,0,7,0,7,0,64,56,17,55,56,89,1,51,19,35,39,7,35,1,182,148,245,139,180,180,139,6,102,254,136,245,245,0,1,0,182,5,29,3,74,6,55,0,27,0,99,64,36,0,18,7,14,11,4,1,18,7,15,11,4,18,195,25,7,4,195,21,11,237,28,15,1,14,0,7,21,86,22,119,7,86,8,118,28,16,244,236,252,236,17,57,57,57,57,49,0,16,252,60,252,212,60,236,17,18,57,17,18,57,17,18,57,17,18,57,48,0,75,176,9,84,75,176,12,84,91,88,189,0,28,255,192,0,1,0,28,0,28,0,64,56,17,55,56,89,1,39,46,1,35,34,6,7,35,62,1,51,50,22,31,1,30,1,51,50,54,55,51,14,1,35,34,38,1,252,57,22,33,13,38,36,2,125,2,102,91,38,64,37,57,22,33,13,38,36,2,125,2,102,91,38,64,5,90,55,20,19,73,82,135,147,28,33,55,20,19,73,82,135,147,28,0,0,1,0,213,5,98,3,43,5,246,0,3,0,47,183,2,239,0,238,4,1,0,4,16,212,204,49,0,16,252,236,48,0,75,176,9,84,75,176,14,84,91,88,189,0,4,255,192,0,1,0,4,0,4,0,64,56,17,55,56,89,19,33,21,33,213,2,86,253,170,5,246,148,0,0,1,0,199,5,41,3,57,6,72,0,13,0,87,64,14,11,240,4,7,0,179,14,7,86,8,1,86,0,14,16,220,236,212,236,49,0,16,244,60,212,236,48,0,75,176,9,84,88,189,0,14,255,192,0,1,0,14,0,14,0,64,56,17,55,56,89,0,75,176,15,84,75,176,16,84,91,75,176,17,84,91,88,189,0,14,0,64,0,1,0,14,0,14,255,192,56,17,55,56,89,19,51,30,1,51,50,54,55,51,14,1,35,34,38,199,118,11,97,87,86,96,13,118,10,158,145,145,158,6,72,75,75,74,76,143,144,144,0,1,1,154,5,68,2,102,6,16,0,3,0,44,64,9,2,206,0,205,4,1,100,0,4,16,212,236,49,0,16,252,236,48,0,75,176,9,84,88,189,0,4,255,192,0,1,0,4,0,4,0,64,56,17,55,56,89,1,51,21,35,1,154,204,204,6,16,204,0,0,2,0,238,4,225,3,18,7,6,0,11,0,23,0,95,64,17,3,193,21,242,9,193,15,241,24,0,86,12,120,6,86,18,24,16,212,236,244,236,49,0,16,244,236,244,236,48,0,75,176,9,84,75,176,12,84,91,88,189,0,24,255,192,0,1,0,24,0,24,0,64,56,17,55,56,89,1,75,176,10,84,75,176,11,84,91,75,176,12,84,91,88,189,0,24,255,192,0,1,0,24,0,24,0,64,56,17,55,56,89,1,52,38,35,34,6,21,20,22,51,50,54,55,20,6,35,34,38,53,52,54,51,50,22,2,152,88,64,65,87,87,65,64,88,122,159,115,115,159,159,115,115,159,5,244,63,88,87,64,65,87,88,64,115,160,160,115,115,159,159,0,0,1,1,35,254,117,2,193,0,0,0,19,0,31,64,14,9,6,10,13,243,6,0,19,0,16,39,3,9,20,16,220,212,236,212,204,49,0,47,212,252,196,18,57,48,33,22,22,21,20,6,35,34,38,39,53,22,22,51,50,54,53,52,38,39,2,84,55,54,120,118,46,87,43,34,74,47,59,60,43,45,62,105,48,89,91,12,12,131,17,15,48,46,30,87,61,0,2,0,240,4,238,3,174,6,102,0,3,0,7,0,66,64,17,6,2,180,4,0,179,8,4,7,3,0,5,1,3,5,7,8,16,212,220,212,204,17,57,17,18,57,49,0,16,244,60,236,50,48,0,75,176,9,84,75,176,14,84,91,88,189,0,8,255,192,0,1,0,8,0,8,0,64,56,17,55,56,89,1,51,3,35,3,51,3,35,2,252,178,248,135,129,170,223,137,6,102,254,136,1,120,254,136,0,0,1,1,76,254,117,2,193,0,0,0,19,0,32,64,15,11,14,10,7,243,14,244,0,1,0,10,4,39,17,20,16,212,236,196,212,204,49,0,47,252,252,196,18,57,48,33,51,6,6,21,20,22,51,50,54,55,21,6,6,35,34,38,53,52,54,1,184,119,45,43,55,54,32,62,31,38,68,30,122,115,53,61,88,31,46,46,15,15,133,10,10,87,93,48,105,0,1,0,193,4,238,3,63,6,102,0,6,0,55,64,12,3,0,180,4,1,179,7,3,5,117,1,7,16,220,236,57,49,0,16,244,60,236,57,48,0,75,176,9,84,75,176,14,84,91,88,189,0,7,255,192,0,1,0,7,0,7,0,64,56,17,55,56,89,1,3,51,23,55,51,3,1,182,245,139,180,180,139,245,4,238,1,120,245,245,254,136,0,1,255,242,0,0,4,117,5,213,0,13,0,63,64,30,12,11,10,4,3,2,6,0,6,149,0,129,8,3,4,1,11,14,0,4,5,1,28,12,7,58,9,0,121,14,16,244,60,236,196,252,60,196,17,18,57,17,18,57,49,0,47,228,236,17,23,57,48,180,48,15,80,15,2,1,93,19,51,17,37,23,1,17,33,21,33,17,7,39,55,211,203,1,57,80,254,119,2,215,252,94,148,77,225,5,213,253,152,219,111,254,238,253,227,170,2,59,106,110,158,0,0,1,0,2,0,0,2,72,6,20,0,11,0,94,64,26,10,9,8,4,3,2,6,0,151,6,3,4,1,9,10,0,4,122,5,1,8,10,122,7,0,12,16,212,60,228,252,60,228,17,18,57,17,18,57,49,0,47,236,23,57,48,1,75,176,16,84,88,189,0,12,0,64,0,1,0,12,0,12,255,192,56,17,55,56,89,64,19,16,13,64,13,80,13,96,13,115,4,122,10,112,13,224,13,240,13,9,93,19,51,17,55,23,7,17,35,17,7,39,55,199,184,125,76,201,184,123,74,197,6,20,253,166,90,106,141,252,227,2,154,88,106,141,0,255,255,0,135,255,227,4,162,7,109,2,39,0,54,0,0,0,0,0,7,1,8,0,139,1,117,255,255,0,111,255,227,3,199,6,102,2,39,0,86,0,0,0,0,0,7,0,224,0,23,0,0,255,255,0,92,0,0,5,31,7,109,2,39,0,61,0,0,0,0,0,7,1,8,0,190,1,117,255,255,0,88,0,0,3,219,6,102,2,39,0,93,0,0,0,0,0,7,0,224,0,27,0,0,0,2,1,4,254,162,1,174,5,152,0,3,0,7,0,28,64,13,1,245,0,4,245,5,8,4,0,5,6,2,8,16,220,60,236,50,49,0,16,212,236,212,236,48,1,17,35,17,19,17,35,17,1,174,170,170,170,1,152,253,10,2,246,4,0,253,10,2,246,0,0,2,0,10,0,0,5,186,5,213,0,12,0,25,0,103,64,32,16,9,169,11,13,149,0,129,18,149,14,11,7,7,1,25,19,4,15,13,22,25,4,50,10,17,13,28,8,0,121,26,16,244,60,236,50,196,244,236,16,196,23,57,49,0,47,198,50,238,246,238,16,238,50,48,64,40,32,27,127,27,176,27,3,159,9,159,10,159,11,159,12,159,14,159,15,159,16,159,17,191,9,191,10,191,11,191,12,191,14,191,15,191,16,191,17,16,93,1,93,19,33,32,0,17,16,0,41,1,17,35,53,51,19,17,33,21,33,17,51,32,0,17,16,0,33,211,1,160,1,177,1,150,254,105,254,80,254,96,201,201,203,1,80,254,176,243,1,53,1,31,254,225,254,203,5,213,254,151,254,128,254,126,254,150,2,188,144,1,227,254,29,144,253,234,1,24,1,46,1,44,1,23,0,2,0,113,255,227,4,117,6,20,0,14,0,40,1,39,64,94,37,123,38,37,30,35,30,36,123,35,35,30,15,123,35,30,40,123,39,40,30,35,30,38,39,40,39,37,36,37,40,40,39,34,35,34,31,32,31,33,32,32,31,66,40,39,38,37,34,33,32,31,8,35,30,3,15,35,3,185,27,9,185,21,140,27,35,177,41,38,39,18,12,33,32,24,40,37,35,34,31,5,30,15,6,12,18,18,81,6,18,24,69,41,16,252,236,244,236,17,57,57,23,57,18,57,57,17,18,57,57,49,0,16,236,196,244,236,16,238,18,57,18,57,18,23,57,48,75,83,88,7,16,14,201,7,16,8,201,7,16,8,201,7,16,14,201,7,16,8,237,7,14,237,7,16,5,237,7,16,8,237,89,34,178,63,42,1,1,93,64,118,22,37,43,31,40,34,47,35,47,36,41,37,45,38,45,39,42,40,54,37,70,37,88,32,88,33,96,32,96,33,102,34,117,32,117,33,117,34,19,37,35,37,36,38,38,38,39,39,40,54,36,54,37,70,36,69,37,90,32,90,33,98,32,98,33,127,0,127,1,127,2,122,3,123,9,127,10,127,11,127,12,127,13,127,14,127,15,127,16,127,17,127,18,127,19,127,20,123,21,122,27,122,28,127,29,127,30,118,32,118,33,120,34,160,42,240,42,39,93,0,93,1,46,1,35,34,6,21,20,22,51,50,54,53,52,38,19,22,18,21,20,0,35,34,0,53,52,0,51,50,22,23,39,5,39,37,39,51,23,37,23,5,3,70,50,88,41,167,185,174,146,145,174,54,9,126,114,254,228,230,231,254,229,1,20,221,18,52,42,159,254,193,33,1,25,181,228,127,1,77,33,254,217,3,147,17,16,216,195,188,222,222,188,122,188,1,38,143,254,224,173,255,254,201,1,55,255,250,1,55,5,5,180,107,99,92,204,145,111,97,98,0,255,255,255,252,0,0,4,231,7,107,2,39,0,60,0,0,0,0,0,7,1,4,0,115,1,117,255,255,0,61,254,86,4,127,6,102,2,39,0,92,0,0,0,0,0,7,0,141,0,94,0,0,0,2,0,201,0,0,4,141,5,213,0,12,0,21,0,61,64,27,14,149,9,13,149,2,246,0,129,11,21,15,9,3,4,1,18,25,6,63,13,10,1,28,0,4,22,16,252,236,50,50,252,236,17,23,57,49,0,47,244,252,236,212,236,48,64,9,15,23,31,23,63,23,95,23,4,1,93,19,51,17,51,50,4,21,20,4,43,1,17,35,19,17,51,50,54,53,52,38,35,201,202,254,251,1,1,254,255,251,254,202,202,254,141,154,153,142,5,213,254,248,225,220,220,226,254,174,4,39,253,209,146,134,134,145,0,2,0,186,254,86,4,164,6,20,0,16,0,28,0,62,64,27,20,185,5,8,26,185,0,14,140,8,184,1,189,3,151,29,17,18,11,71,23,4,0,8,2,70,29,16,252,236,50,50,244,236,49,0,16,236,228,228,244,196,236,16,198,238,48,64,9,96,30,128,30,160,30,224,30,4,1,93,37,17,35,17,51,17,62,1,51,50,18,17,16,2,35,34,38,1,52,38,35,34,6,21,20,22,51,50,54,1,115,185,185,58,177,123,204,255,255,204,123,177,2,56,167,146,146,167,167,146,146,167,168,253,174,7,190,253,162,100,97,254,188,254,248,254,248,254,188,97,1,235,203,231,231,203,203,231,231,0,0,1,0,217,2,45,5,219,2,215,0,3,0,17,182,0,156,2,4,1,0,4,16,212,196,49,0,16,212,236,48,19,33,21,33,217,5,2,250,254,2,215,170,0,0,1,1,25,0,63,5,156,4,197,0,11,0,133,64,77,10,156,11,10,7,8,7,9,156,8,8,7,4,156,3,4,7,7,6,5,156,6,7,6,4,156,5,4,1,2,1,3,156,2,2,1,11,156,0,1,0,10,156,9,10,1,1,0,66,10,8,7,6,4,2,1,0,8,5,3,11,9,12,11,10,9,7,5,4,3,1,8,2,0,8,6,12,16,212,60,204,50,23,57,49,0,16,212,60,204,50,23,57,48,75,83,88,7,16,8,237,7,16,5,237,7,16,5,237,7,16,8,237,7,16,5,237,7,16,8,237,7,16,5,237,7,16,8,237,89,34,9,2,7,1,1,39,1,1,55,1,1,5,156,254,55,1,201,119,254,53,254,53,118,1,200,254,56,118,1,203,1,203,4,76,254,53,254,55,121,1,203,254,53,121,1,201,1,203,121,254,53,1,203,0,1,0,137,2,156,2,197,5,223,0,10,0,44,64,24,7,0,221,9,3,221,4,2,221,9,247,5,145,11,8,124,6,93,3,124,1,124,0,11,16,220,244,228,252,228,49,0,16,244,236,236,212,236,16,238,50,48,19,51,17,7,53,55,51,17,51,21,33,156,204,223,230,137,205,253,215,3,10,2,99,41,116,39,253,43,110,0,0,1,0,94,2,156,2,180,5,240,0,24,0,74,64,36,0,125,6,4,0,23,125,6,6,4,66,4,2,0,14,221,15,0,221,2,247,11,221,15,18,145,25,0,14,8,126,1,21,14,3,25,16,220,196,212,196,236,17,57,49,0,16,244,196,236,252,236,16,238,17,18,57,48,75,83,88,7,16,5,237,23,50,7,5,237,89,34,1,33,21,33,53,54,55,0,53,52,38,35,34,6,7,53,54,54,51,50,22,21,20,1,6,1,12,1,168,253,170,34,63,1,88,104,85,52,122,72,77,133,57,145,174,254,181,56,3,14,114,110,31,56,1,49,94,66,81,35,35,123,28,28,132,108,139,254,228,48,0,1,0,98,2,141,2,205,5,240,0,40,0,72,64,39,0,21,19,10,221,9,31,221,32,19,221,21,13,221,9,248,6,247,28,221,32,248,35,145,41,22,19,0,20,25,126,38,16,126,3,20,31,9,41,16,220,196,196,212,236,212,236,17,57,57,57,49,0,16,244,228,236,252,228,236,212,236,16,238,16,238,17,18,57,48,1,22,22,21,20,6,35,34,38,39,53,22,22,51,50,54,53,52,38,35,35,53,51,50,54,53,52,38,35,34,6,7,53,54,54,51,50,22,21,20,6,2,12,92,101,190,177,57,125,70,52,119,67,109,120,111,108,86,94,94,97,100,95,40,102,81,73,128,55,144,169,90,4,96,18,109,82,124,134,21,20,121,27,26,79,70,74,76,108,63,60,58,61,18,23,115,17,18,118,99,69,96,255,255,0,137,255,227,7,127,5,240,0,39,0,240,0,0,0,0,0,39,0,188,3,53,0,0,0,7,1,9,4,139,253,100,255,255,0,137,255,227,7,63,5,240,0,39,0,240,0,0,0,0,0,39,0,188,3,53,0,0,0,7,0,241,4,139,253,100,255,255,0,98,255,227,7,127,5,240,0,39,0,242,0,0,0,0,0,39,0,188,3,53,0,0,0,7,1,9,4,139,253,100,255,255,0,115,255,227,5,139,7,109,2,39,0,42,0,0,0,0,0,7,1,10,1,27,1,117,255,255,0,113,254,86,4,90,6,72,2,39,0,74,0,0,0,0,0,7,0,218,0,139,0,0,255,255,0,201,0,0,1,149,7,80,2,39,0,44,0,0,0,0,0,7,1,11,255,47,1,117,255,255,0,135,254,117,4,162,5,240,2,39,0,54,0,0,0,0,0,7,0,221,0,139,0,0,255,255,0,111,254,117,3,199,4,123,2,39,0,86,0,0,0,0,0,7,0,221,0,23,0,0,255,255,0,115,255,227,5,39,7,107,2,39,0,38,0,0,0,0,0,7,1,4,1,45,1,117,255,255,0,113,255,227,3,231,6,102,2,39,0,70,0,0,0,0,0,7,0,141,0,137,0,0,255,255,0,115,255,227,5,39,7,109,2,39,0,38,0,0,0,0,0,7,1,8,1,45,1,117,255,255,0,113,255,227,3,231,6,102,2,39,0,70,0,0,0,0,0,7,0,224,0,137,0,0,0,2,0,113,255,227,4,244,6,20,0,24,0,36,0,74,64,36,7,3,211,9,1,249,34,185,0,22,28,185,13,16,140,22,184,5,151,11,2,31,12,4,3,0,8,8,10,6,71,25,18,19,69,37,16,252,236,244,60,196,252,23,60,196,49,0,47,236,228,244,196,236,16,196,238,253,60,238,50,48,182,96,38,128,38,160,38,3,1,93,1,17,33,53,33,53,51,21,51,21,35,17,35,53,14,1,35,34,2,17,16,18,51,50,22,1,20,22,51,50,54,53,52,38,35,34,6,3,162,254,186,1,70,184,154,154,184,58,177,124,203,255,255,203,124,177,253,199,167,146,146,168,168,146,146,167,3,182,1,78,125,147,147,125,250,252,168,100,97,1,68,1,8,1,8,1,68,97,254,21,203,231,231,203,203,231,231,0,0,1,0,100,1,223,2,127,2,131,0,3,0,17,182,0,156,2,4,1,0,4,16,220,204,49,0,16,212,236,48,19,33,21,33,100,2,27,253,229,2,131,164,0,0,1,0,219,2,72,1,174,3,70,0,3,0,18,183,2,131,0,4,1,25,0,4,16,212,236,49,0,16,212,236,48,19,51,21,35,219,211,211,3,70,254,0,1,0,0,255,227,4,143,5,240,0,49,1,28,64,58,32,18,211,34,16,43,7,211,9,25,161,26,174,22,149,29,1,161,0,174,4,149,47,145,29,140,41,9,50,43,34,33,41,35,38,18,16,10,3,13,9,17,8,44,32,38,19,7,17,8,17,13,28,25,0,38,42,33,47,60,212,196,50,252,196,196,18,57,57,18,57,57,17,18,57,17,23,57,18,57,57,17,57,57,49,0,16,196,50,228,244,236,244,236,16,238,246,238,16,238,50,221,60,238,50,48,1,75,176,9,84,75,176,12,84,91,75,176,13,84,91,75,176,15,84,91,75,176,23,84,91,75,176,24,84,91,88,189,0,50,255,192,0,1,0,50,0,50,0,64,56,17,55,56,89,64,122,14,0,14,1,11,2,11,49,84,20,105,12,108,14,110,15,111,16,111,17,111,18,111,19,105,20,107,31,111,32,111,33,111,34,111,35,110,36,108,37,105,39,105,45,159,7,159,8,159,9,159,10,159,11,159,12,159,13,159,14,159,15,159,16,159,17,159,18,159,19,150,31,159,32,159,33,159,34,159,35,159,36,159,37,159,38,159,39,159,40,159,41,159,42,159,43,159,44,157,45,50,0,8,0,9,16,8,16,9,32,8,32,9,85,21,83,30,106,21,103,31,10,93,0,93,1,21,46,1,35,34,6,7,33,7,33,14,1,21,20,22,23,33,7,33,30,1,51,50,54,55,21,14,1,35,34,0,3,35,55,51,52,38,53,52,54,53,35,55,51,18,0,51,50,22,4,143,91,169,102,157,202,32,2,65,55,253,230,2,1,1,2,1,190,56,254,138,32,202,157,102,169,91,89,185,96,237,254,203,40,211,55,139,1,1,194,55,156,40,1,54,236,98,185,5,98,213,105,90,200,187,123,24,46,35,32,46,24,123,187,202,90,105,211,72,72,1,34,1,3,123,23,47,32,35,47,23,123,1,1,1,34,71,0,2,0,215,5,14,3,41,5,217,0,3,0,7,0,165,64,13,4,0,206,6,2,8,1,100,0,5,100,4,8,16,212,252,220,236,49,0,16,212,60,236,50,48,0,75,176,14,84,75,176,17,84,91,88,189,0,8,0,64,0,1,0,8,0,8,255,192,56,17,55,56,89,1,75,176,14,84,75,176,13,84,91,75,176,23,84,91,88,189,0,8,255,192,0,1,0,8,0,8,0,64,56,17,55,56,89,1,75,176,17,84,75,176,25,84,91,88,189,0,8,0,64,0,1,0,8,0,8,255,192,56,17,55,56,89,0,75,176,24,84,88,189,0,8,255,192,0,1,0,8,0,8,0,64,56,17,55,56,89,64,17,96,1,96,2,96,5,96,6,112,1,112,2,112,5,112,6,8,1,93,1,51,21,35,37,51,21,35,2,94,203,203,254,121,203,203,5,217,203,203,203,0,1,1,115,4,238,2,240,5,246,0,3,0,127,64,17,2,3,0,3,1,0,0,3,66,0,2,250,4,1,3,3,4,16,196,16,192,49,0,16,244,204,48,75,83,88,7,16,5,201,7,16,5,201,89,34,0,75,176,12,84,88,189,0,4,255,192,0,1,0,4,0,4,0,64,56,17,55,56,89,0,75,176,14,84,88,189,0,4,0,64,0,1,0,4,0,4,255,192,56,17,55,56,89,64,32,6,2,21,2,37,1,37,2,54,2,70,2,86,2,106,1,103,2,9,15,0,15,1,31,0,31,1,47,0,47,1,6,93,1,93,1,51,3,35,2,55,185,228,153,5,246,254,248,0,1,0,182,5,14,3,74,5,233,0,29,0,117,64,33,22,16,15,3,19,12,7,1,0,3,8,23,12,195,4,19,195,27,8,250,30,16,1,15,0,7,22,86,24,7,86,9,30,16,212,236,212,236,17,57,57,57,57,49,0,16,244,60,236,212,236,50,18,23,57,17,18,23,57,48,0,75,176,12,84,88,189,0,30,255,192,0,1,0,30,0,30,0,64,56,17,55,56,89,0,75,176,14,84,88,189,0,30,0,64,0,1,0,30,0,30,255,192,56,17,55,56,89,180,16,11,31,26,2,93,1,39,46,1,35,34,6,29,1,35,52,54,51,50,22,31,1,30,1,51,50,54,61,1,51,14,1,35,34,38,1,252,57,25,31,12,36,40,125,103,86,36,61,48,57,23,34,15,32,40,125,2,103,84,34,59,5,57,33,14,11,50,45,6,101,118,16,27,30,13,12,51,41,6,100,119,16,0,1,1,12,4,238,2,139,5,246,0,3,0,137,64,17,1,2,3,2,0,3,3,2,66,0,1,250,4,1,3,3,4,16,196,16,192,49,0,16,244,204,48,75,83,88,7,16,5,201,7,16,5,201,89,34,0,75,176,12,84,88,189,0,4,255,192,0,1,0,4,0,4,0,64,56,17,55,56,89,0,75,176,14,84,88,189,0,4,0,64,0,1,0,4,0,4,255,192,56,17,55,56,89,64,42,6,0,6,1,22,0,18,1,36,0,36,1,53,1,67,1,85,0,85,1,159,0,159,1,175,0,175,1,14,15,0,15,3,31,0,31,3,47,0,47,3,6,93,1,93,1,19,35,3,1,199,196,153,230,5,246,254,248,1,8,0,1,0,207,4,238,3,49,5,248,0,6,0,119,64,10,4,0,5,2,250,7,4,2,6,7,16,212,196,57,49,0,16,244,60,196,57,48,0,75,176,12,84,88,189,0,7,255,192,0,1,0,7,0,7,0,64,56,17,55,56,89,0,75,176,14,84,88,189,0,7,0,64,0,1,0,7,0,7,255,192,56,17,55,56,89,1,75,176,14,84,88,189,0,7,255,192,0,1,0,7,0,7,0,64,56,17,55,56,89,64,19,15,0,15,1,12,4,31,0,31,1,29,4,47,0,47,1,45,4,9,0,93,1,51,19,35,39,7,35,1,162,188,211,139,166,166,139,5,248,254,246,178,178,0,1,0,207,4,238,3,49,5,248,0,6,0,134,64,10,3,4,1,0,250,7,3,5,1,7,16,212,196,57,49,0,16,244,196,50,57,48,0,75,176,12,84,75,176,9,84,91,75,176,10,84,91,75,176,11,84,91,88,189,0,7,255,192,0,1,0,7,0,7,0,64,56,17,55,56,89,0,75,176,14,84,88,189,0,7,0,64,0,1,0,7,0,7,255,192,56,17,55,56,89,1,75,176,14,84,88,189,0,7,255,192,0,1,0,7,0,7,0,64,56,17,55,56,89,64,19,0,0,3,3,0,6,16,0,18,3,16,6,32,0,34,3,32,6,9,0,93,1,3,51,23,55,51,3,1,162,211,139,166,166,139,211,4,238,1,10,178,178,254,246,0,0,2,0,63,2,156,2,244,5,223,0,2,0,13,0,212,64,22,0,3,11,7,221,5,1,9,247,3,145,14,1,12,10,0,93,6,8,4,12,14,16,220,212,60,196,236,50,17,57,49,0,16,244,252,212,60,236,50,18,57,48,1,75,176,14,84,75,176,15,84,91,75,176,16,84,91,75,176,17,84,91,75,176,11,84,91,75,176,10,84,91,88,189,0,14,0,64,0,1,0,14,0,14,255,192,56,17,55,56,89,0,75,176,17,84,75,176,14,84,91,88,189,0,14,255,192,0,1,0,14,0,14,0,64,56,17,55,56,89,64,84,11,1,29,1,47,1,57,1,73,1,70,3,89,3,105,3,139,3,171,3,187,3,11,1,0,15,1,15,2,15,5,15,6,15,7,15,8,15,11,15,12,15,13,19,0,31,1,31,2,31,5,31,6,31,7,31,8,31,11,31,12,31,13,34,0,53,0,71,0,75,13,83,0,91,13,101,0,132,0,165,0,181,0,30,93,1,93,9,1,33,3,51,17,51,21,35,21,35,53,33,53,1,221,254,203,1,53,22,166,135,135,144,254,98,5,102,254,93,2,28,253,228,109,186,186,121,0,0,1,0,199,5,6,3,57,5,248,0,13,0,106,64,14,7,0,4,195,11,250,14,7,86,8,1,86,0,14,16,212,236,212,236,49,0,16,244,252,196,50,48,0,75,176,12,84,88,189,0,14,255,192,0,1,0,14,0,14,0,64,56,17,55,56,89,0,75,176,14,84,88,189,0,14,0,64,0,1,0,14,0,14,255,192,56,17,55,56,89,1,75,176,14,84,75,176,15,84,91,88,189,0,14,255,192,0,1,0,14,0,14,0,64,56,17,55,56,89,19,51,30,1,51,50,54,55,51,14,1,35,34,38,199,118,13,99,83,82,97,16,118,10,160,143,144,159,5,248,54,57,55,56,119,123,122,0,0,1,1,154,5,14,2,102,5,219,0,3,0,17,182,0,2,250,4,1,0,4,16,212,204,49,0,16,244,204,48,1,51,21,35,1,154,204,204,5,219,205,0,0,0,0,0,2,0,1,0,0,0,0,0,20,0,3,0,1,0,0,1,26,0,0,1,6,0,0,1,0,0,0,0,0,0,0,1,3,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,96,97,0,98,99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,159,160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,208,209,0,210,211,212,213,214,215,216,217,218,219,220,221,222,223,224,0,4,2,62,0,0,0,58,0,32,0,4,0,26,0,126,0,255,1,7,1,17,1,31,1,49,1,66,1,83,1,97,1,120,1,126,1,146,2,199,2,221,3,169,3,192,32,38,32,48,32,58,32,172,33,34,34,6,34,30,34,43,34,72,34,101,37,202,251,2,255,255,0,0,0,32,0,160,1,6,1,12,1,30,1,48,1,65,1,82,1,94,1,120,1,125,1,146,2,198,2,216,3,169,3,192,32,19,32,48,32,57,32,172,33,34,34,2,34,15,34,43,34,72,34,96,37,202,251,1,255,255,255,227,0,0,255,245,0,0,255,216,0,0,255,160,255,94,0,0,255,67,255,104,255,20,0,0,0,0,252,246,252,219,0,0,224,150,224,133,224,86,223,106,0,0,0,0,222,113,222,95,0,0,218,239,5,191,0,1,0,0,0,56,0,0,0,244,0,0,0,252,0,0,0,0,0,250,0,0,0,0,0,0,0,250,0,252,0,0,0,0,1,2,0,0,0,0,0,0,0,0,1,32,1,40,0,0,0,0,1,66,0,0,0,0,0,0,0,172,0,163,0,132,0,133,0,189,0,150,0,231,0,134,0,142,0,139,0,157,0,169,0,164,1,0,0,138,0,217,0,131,0,147,0,241,0,242,0,141,0,151,0,136,0,195,0,221,0,240,0,158,0,170,0,243,0,244,0,245,0,162,0,173,0,201,0,199,0,174,0,98,0,99,0,144,0,100,0,203,0,101,0,200,0,202,0,207,0,204,0,205,0,206,0,232,0,102,0,210,0,208,0,209,0,175,0,103,0,239,0,145,0,213,0,211,0,212,0,104,0,234,0,236,0,137,0,106,0,105,0,107,0,109,0,108,0,110,0,160,0,111,0,113,0,112,0,114,0,115,0,117,0,116,0,118,0,119,0,233,0,120,0,122,0,121,0,123,0,125,0,124,0,184,0,161,0,127,0,126,0,128,0,129,0,235,0,237,0,186,0,253,0,254,0,0,0,0,0,0,0,255,0,248,0,214,0,249,0,250,0,227,0,228,0,215,0,224,0,218,0,219,0,220,0,223,0,216,0,222,0,178,0,179,0,0,0,0,0,0,0,182,0,183,0,196,0,0,0,180,0,181,0,197,0,0,0,130,0,194,0,135,0,0,0,0,0,0,0,171,0,152,0,0,0,0,0,0,0,168,0,154,0,0,0,153,0,238,0,0,0,0,0,188,0,0,0,0,0,0,1,1,0,165,0,0,0,0,0,0,0,146,0,143,0,0,0,0,0,0,0,148,0,149,4,205,0,102,0,0,0,0,2,139,0,0,2,139,0,0,3,53,1,53,3,174,0,197,6,180,0,158,5,23,0,170,7,154,0,113,6,61,0,129,2,51,0,197,3,31,0,176,3,31,0,164,4,0,0,61,6,180,0,217,2,139,0,158,2,227,0,100,2,139,0,219,2,178,0,0,5,23,0,135,5,23,0,225,5,23,0,150,5,23,0,156,5,23,0,100,5,23,0,158,5,23,0,143,5,23,0,168,5,23,0,139,5,23,0,129,2,178,0,240,2,178,0,158,6,180,0,217,6,180,0,217,6,180,0,217,4,63,0,147,8,0,0,135,5,121,0,16,5,125,0,201,5,150,0,115,6,41,0,201,5,14,0,201,4,154,0,201,6,51,0,115,6,4,0,201,2,92,0,201,2,92,255,150,5,63,0,201,4,117,0,201,6,231,0,201,5,252,0,201,6,76,0,115,4,211,0,201,6,76,0,115,5,143,0,201,5,20,0,135,4,227,255,250,5,219,0,178,5,121,0,16,7,233,0,68,5,123,0,61,4,227,255,252,5,123,0,92,3,31,0,176,2,178,0,0,3,31,0,199,6,180,0,217,4,0,255,236,4,0,0,170,4,231,0,123,5,20,0,186,4,102,0,113,5,20,0,113,4,236,0,113,2,209,0,47,5,20,0,113,5,18,0,186,2,57,0,193,2,57,255,219,4,162,0,186,2,57,0,193,7,203,0,186,5,18,0,186,4,229,0,113,5,20,0,186,5,20,0,113,3,74,0,186,4,43,0,111,3,35,0,55,5,18,0,174,4,188,0,61,6,139,0,86,4,188,0,59,4,188,0,61,4,51,0,88,5,23,1,0,2,178,1,4,5,23,1,0,6,180,0,217,5,121,0,16,5,121,0,16,5,150,0,115,5,14,0,201,5,252,0,201,6,76,0,115,5,219,0,178,4,231,0,123,4,231,0,123,4,231,0,123,4,231,0,123,4,231,0,123,4,231,0,123,4,102,0,113,4,236,0,113,4,236,0,113,4,236,0,113,4,236,0,113,2,57,0,144,2,57,255,199,2,57,255,222,2,57,255,244,5,18,0,186,4,229,0,113,4,229,0,113,4,229,0,113,4,229,0,113,4,229,0,113,5,18,0,174,5,18,0,174,5,18,0,174,5,18,0,174,4,0,0,57,4,0,0,195,5,23,0,172,5,23,0,129,4,0,0,92,4,184,1,51,5,23,0,158,5,10,0,186,8,0,1,27,8,0,1,27,8,0,1,39,4,0,1,115,4,0,0,215,6,180,0,217,7,203,0,8,6,76,0,102,6,170,0,221,6,180,0,217,6,180,0,217,6,180,0,217,5,23,0,82,5,23,0,174,4,35,0,104,5,100,0,25,6,14,0,156,4,182,255,225,4,43,0,47,3,197,0,115,3,197,0,96,6,29,0,78,7,219,0,123,4,229,0,72,4,63,0,143,3,53,1,53,6,180,0,217,5,25,0,61,5,23,0,31,6,180,0,217,5,90,255,250,4,229,0,158,4,229,0,193,8,0,0,236,5,23,0,0,5,121,0,16,5,121,0,16,6,76,0,115,8,143,0,115,8,47,0,113,4,0,0,0,8,0,0,0,4,37,0,174,4,37,0,174,2,139,0,174,2,139,0,178,6,180,0,217,3,244,0,6,4,188,0,61,4,227,255,252,1,86,254,137,5,23,0,94,3,51,0,158,3,51,0,193,5,10,0,47,5,10,0,47,4,0,0,57,2,139,0,219,2,139,0,174,4,37,0,174,10,188,0,113,5,121,0,16,5,14,0,201,5,121,0,16,5,14,0,201,5,14,0,201,2,92,0,162,2,92,255,254,2,92,0,6,2,92,0,59,6,76,0,115,6,76,0,115,6,76,0,115,5,219,0,178,5,219,0,178,5,219,0,178,2,57,0,193,4,0,0,193,4,0,0,182,4,0,0,213,4,0,0,199,4,0,1,154,4,0,0,238,4,0,1,35,4,0,0,240,4,0,1,76,4,0,0,193,4,127,255,242,2,70,0,2,5,20,0,135,4,43,0,111,5,123,0,92,4,51,0,88,2,178,1,4,6,51,0,10,4,229,0,113,4,227,255,252,4,188,0,61,4,215,0,201,5,20,0,186,6,180,0,217,6,180,1,25,3,53,0,137,3,53,0,94,3,53,0,98,7,193,0,137,7,193,0,137,7,193,0,98,6,51,0,115,5,20,0,113,2,92,0,201,5,20,0,135,4,43,0,111,5,150,0,115,4,102,0,113,5,150,0,115,4,102,0,113,5,20,0,113,2,227,0,100,2,139,0,219,5,23,0,0,4,0,0,215,4,0,1,115,4,0,0,182,4,0,1,12,4,0,0,207,4,0,0,207,3,53,0,63,4,0,0,199,4,0,1,154,0,2,0,0,0,0,0,0,255,43,0,143,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,12,0,0,0,1,0,2,0,3,0,4,0,5,0,6,0,7,0,8,0,9,0,10,0,11,0,12,0,13,0,14,0,15,0,16,0,17,0,18,0,19,0,20,0,21,0,22,0,23,0,24,0,25,0,26,0,27,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,40,0,41,0,42,0,43,0,44,0,45,0,46,0,47,0,48,0,49,0,50,0,51,0,52,0,53,0,54,0,55,0,56,0,57,0,58,0,59,0,60,0,61,0,62,0,63,0,64,0,65,0,66,0,67,0,68,0,69,0,70,0,71,0,72,0,73,0,74,0,75,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,88,0,89,0,90,0,91,0,92,0,93,0,94,0,95,0,96,0,97,0,98,0,99,0,100,0,101,0,102,0,103,0,104,0,105,0,106,0,107,0,108,0,109,0,110,0,111,0,112,0,113,0,114,0,115,0,116,0,117,0,118,0,119,0,120,0,121,0,122,0,123,0,124,0,125,0,126,0,127,0,128,0,129,0,130,0,131,0,132,0,133,0,134,0,135,0,136,0,137,0,138,0,139,0,140,0,141,0,142,0,143,0,144,0,145,0,146,0,147,0,148,0,149,0,150,0,151,0,152,0,153,0,154,0,155,0,156,0,157,0,158,0,159,0,160,0,161,0,162,0,163,0,164,0,165,0,166,0,167,0,168,0,169,0,170,0,171,0,172,0,173,0,174,0,175,0,176,0,177,0,178,0,179,0,180,0,181,0,182,0,183,0,184,0,185,0,186,0,187,0,188,0,189,0,190,0,191,0,192,0,193,0,194,0,195,0,196,0,197,0,198,0,199,0,200,0,201,0,202,0,203,0,204,0,205,0,206,0,207,0,208,0,209,0,211,0,212,0,213,0,214,0,215,0,216,0,217,0,218,0,219,0,220,0,221,0,222,0,223,0,224,0,225,0,226,0,227,0,228,0,229,0,230,0,231,0,232,0,233,0,234,0,235,0,236,0,237,0,238,0,239,0,240,0,241,0,242,0,243,0,245,0,244,0,246,0,248,0,249,0,250,0,251,0,252,0,253,0,254,0,255,1,0,1,1,1,2,1,3,1,4,1,5,1,6,1,7,1,8,1,9,1,10,1,11,1,12,1,13,9,115,102,116,104,121,112,104,101,110,14,112,101,114,105,111,100,99,101,110,116,101,114,101,100,4,69,117,114,111,5,99,54,52,53,57,5,99,54,52,54,48,5,99,54,52,54,49,5,99,54,52,54,50,5,99,54,52,54,51,5,99,54,52,54,54,5,99,54,52,54,55,5,99,54,52,54,56,5,99,54,52,54,57,0,0,0,0,0,34,0,34,0,34,0,34,0,88,0,147,0,255,1,182,2,79,3,131,3,178,3,246,4,33,4,110,4,152,4,180,4,202,4,222,5,3,5,69,5,130,6,5,6,126,6,225,7,76,7,183,7,252,8,101,8,207,8,238,9,22,9,82,9,115,9,174,10,31,10,223,11,88,11,176,11,251,12,58,12,105,12,147,12,230,13,20,13,61,13,122,14,13,14,47,14,172,15,0,15,69,15,133,15,235,16,117,16,241,17,41,17,112,17,224,18,189,19,138,19,235,20,80,20,139,20,177,20,225,21,3,21,24,21,64,21,214,22,34,22,109,22,185,23,35,23,123,23,223,24,27,24,67,24,128,24,248,25,22,25,119,25,179,26,4,26,82,26,161,26,216,27,135,27,196,28,0,28,157,29,186,30,135,31,119,31,232,32,91,32,114,32,243,33,53,33,66,33,226,33,239,33,252,34,9,34,22,34,35,34,48,34,61,34,74,34,87,34,100,34,113,34,126,34,139,34,152,34,165,34,178,34,191,34,204,34,217,34,230,34,243,35,0,35,13,35,26,35,39,35,52,35,65,35,78,35,91,35,104,35,148,35,207,36,52,36,143,37,51,37,83,37,129,38,20,38,186,39,75,39,144,39,183,40,19,40,88,40,195,41,95,42,37,42,92,42,163,42,233,43,122,43,211,44,68,44,143,44,177,45,80,45,160,46,14,46,82,46,170,47,135,48,65,48,189,49,5,49,33,49,80,49,207,50,72,50,122,50,223,51,70,51,112,51,112,51,125,51,138,51,151,51,230,52,122,52,143,52,163,52,209,52,255,53,28,53,57,53,103,53,145,53,158,53,171,53,207,54,91,54,147,54,205,55,67,55,169,55,235,56,0,56,28,56,74,57,15,57,28,57,41,57,54,57,67,57,80,57,93,57,106,57,119,57,132,57,145,57,158,57,171,57,184,57,197,57,210,57,239,58,28,58,123,58,160,58,229,59,8,59,94,59,142,59,196,59,244,60,34,60,95,60,167,60,180,60,193,60,206,60,219,60,254,61,99,62,59,62,72,62,85,62,152,62,231,62,253,63,97,63,141,63,220,64,58,64,75,64,92,64,109,64,122,64,135,64,148,64,161,64,174,64,187,64,200,64,213,64,226,65,64,65,86,65,107,66,69,66,170,66,247,67,95,67,178,67,255,68,85,68,219,69,42,69,63,0,0,0,0,0,1,0,0,45,134,0,1,7,148,24,0,0,10,21,120,0,16,0,36,255,211,0,16,0,37,255,183,0,16,0,38,0,0,0,16,0,39,0,0,0,16,0,41,0,0,0,16,0,42,0,75,0,16,0,43,0,0,0,16,0,45,0,114,0,16,0,46,0,0,0,16,0,47,0,0,0,16,0,50,0,57,0,16,0,51,0,0,0,16,0,52,0,75,0,16,0,53,0,0,0,16,0,55,255,68,0,16,0,57,255,136,0,16,0,58,255,173,0,16,0,59,255,154,0,16,0,60,255,13,0,16,0,61,0,0,0,16,0,73,0,0,0,16,0,81,0,0,0,16,0,82,0,38,0,16,0,85,0,0,0,16,0,89,255,201,0,16,0,90,0,0,0,16,0,92,255,220,0,16,0,98,255,211,0,16,0,100,0,0,0,16,0,103,0,57,0,16,0,120,0,0,0,16,0,121,0,38,0,16,0,122,0,38,0,16,0,123,0,38,0,16,0,124,0,38,0,16,0,125,0,38,0,16,0,137,0,0,0,16,0,144,0,0,0,16,0,173,255,211,0,16,0,174,255,211,0,16,0,175,0,57,0,16,0,186,255,220,0,16,0,187,255,13,0,16,0,199,255,211,0,16,0,201,255,211,0,16,0,208,0,57,0,16,0,209,0,57,0,16,0,210,0,57,0,16,0,229,0,0,0,16,0,233,0,0,0,16,0,234,255,13,0,16,0,235,255,220,0,16,0,236,0,0,0,16,0,246,0,75,0,16,0,251,0,0,0,16,0,253,0,0,0,36,0,16,255,211,0,36,0,17,255,220,0,36,0,29,255,220,0,36,0,36,0,57,0,36,0,38,255,220,0,36,0,42,255,220,0,36,0,50,255,220,0,36,0,52,255,220,0,36,0,54,0,0,0,36,0,55,255,97,0,36,0,56,0,0,0,36,0,57,255,125,0,36,0,58,255,144,0,36,0,59,0,0,0,36,0,60,255,97,0,36,0,70,255,220,0,36,0,71,255,220,0,36,0,72,255,220,0,36,0,73,255,183,0,36,0,82,255,220,0,36,0,84,255,220,0,36,0,87,255,220,0,36,0,88,0,0,0,36,0,89,255,136,0,36,0,90,255,173,0,36,0,92,255,117,0,36,0,98,0,57,0,36,0,100,255,220,0,36,0,103,255,220,0,36,0,104,0,0,0,36,0,111,255,220,0,36,0,112,255,220,0,36,0,113,255,220,0,36,0,114,255,220,0,36,0,115,255,220,0,36,0,121,255,220,0,36,0,122,255,220,0,36,0,123,255,220,0,36,0,124,255,220,0,36,0,125,255,220,0,36,0,126,0,0,0,36,0,127,0,0,0,36,0,128,0,0,0,36,0,129,0,0,0,36,0,169,255,183,0,36,0,170,0,0,0,36,0,173,0,57,0,36,0,174,0,57,0,36,0,175,255,220,0,36,0,180,254,248,0,36,0,181,255,3,0,36,0,186,255,117,0,36,0,187,255,97,0,36,0,197,0,47,0,36,0,199,0,57,0,36,0,201,0,57,0,36,0,208,255,220,0,36,0,209,255,220,0,36,0,210,255,220,0,36,0,211,0,0,0,36,0,212,0,0,0,36,0,213,0,0,0,36,0,227,0,0,0,36,0,234,255,97,0,36,0,235,255,117,0,36,0,246,255,220,0,36,0,249,0,0,0,36,0,251,255,220,0,36,0,252,255,220,0,36,0,253,255,220,0,36,0,254,255,220,0,37,0,16,0,0,0,37,0,38,255,220,0,37,0,42,255,220,0,37,0,50,255,220,0,37,0,54,255,220,0,37,0,57,255,193,0,37,0,58,255,183,0,37,0,60,255,144,0,37,0,100,255,220,0,37,0,103,255,220,0,37,0,169,255,193,0,37,0,170,255,220,0,37,0,175,255,220,0,37,0,180,255,144,0,37,0,181,255,144,0,37,0,187,255,144,0,37,0,197,255,173,0,37,0,208,255,220,0,37,0,209,255,220,0,37,0,210,255,220,0,37,0,227,255,220,0,37,0,234,255,144,0,37,0,246,255,220,0,37,0,249,255,220,0,37,0,251,255,220,0,37,0,253,255,220,0,38,0,16,0,0,0,38,0,36,0,0,0,38,0,54,0,0,0,38,0,60,255,220,0,38,0,98,0,0,0,38,0,169,255,220,0,38,0,170,255,220,0,38,0,173,0,0,0,38,0,174,0,0,0,38,0,180,0,0,0,38,0,181,0,38,0,38,0,187,255,220,0,38,0,197,0,0,0,38,0,199,0,0,0,38,0,201,0,0,0,38,0,227,0,0,0,38,0,234,255,220,0,38,0,249,0,0,0,39,0,16,0,0,0,39,0,36,255,220,0,39,0,57,255,220,0,39,0,58,0,0,0,39,0,60,255,144,0,39,0,98,255,220,0,39,0,169,255,220,0,39,0,170,255,220,0,39,0,173,255,220,0,39,0,174,255,220,0,39,0,180,255,211,0,39,0,181,255,201,0,39,0,187,255,144,0,39,0,197,255,68,0,39,0,199,255,220,0,39,0,201,255,220,0,39,0,234,255,144,0,41,0,16,0,0,0,41,0,17,254,183,0,41,0,29,255,97,0,41,0,36,255,68,0,41,0,54,255,220,0,41,0,55,255,220,0,41,0,68,255,68,0,41,0,72,255,144,0,41,0,76,255,107,0,41,0,82,255,183,0,41,0,85,255,107,0,41,0,88,255,144,0,41,0,92,255,68,0,41,0,98,255,68,0,41,0,105,255,68,0,41,0,106,255,68,0,41,0,107,255,68,0,41,0,108,255,68,0,41,0,109,255,68,0,41,0,110,255,68,0,41,0,112,255,144,0,41,0,113,255,144,0,41,0,114,255,144,0,41,0,115,255,144,0,41,0,121,255,183,0,41,0,122,255,183,0,41,0,123,255,183,0,41,0,124,255,183,0,41,0,125,255,183,0,41,0,126,255,144,0,41,0,127,255,144,0,41,0,128,255,144,0,41,0,129,255,144,0,41,0,169,0,0,0,41,0,170,0,0,0,41,0,173,255,68,0,41,0,174,255,68,0,41,0,180,255,211,0,41,0,181,0,0,0,41,0,186,255,68,0,41,0,197,254,136,0,41,0,199,255,68,0,41,0,201,255,68,0,41,0,227,255,220,0,41,0,235,255,68,0,41,0,249,255,220,0,42,0,16,0,0,0,42,0,36,0,0,0,42,0,55,255,183,0,42,0,58,0,0,0,42,0,60,255,154,0,42,0,98,0,0,0,42,0,169,255,220,0,42,0,170,255,220,0,42,0,173,0,0,0,42,0,174,0,0,0,42,0,180,255,211,0,42,0,181,255,211,0,42,0,187,255,154,0,42,0,197,255,201,0,42,0,199,0,0,0,42,0,201,0,0,0,42,0,234,255,154,0,43,0,16,0,0,0,43,0,17,255,220,0,43,0,29,0,0,0,43,0,169,0,0,0,43,0,170,0,0,0,43,0,180,255,183,0,43,0,181,255,193,0,43,0,197,255,183,0,45,0,16,255,183,0,45,0,36,255,220,0,45,0,98,255,220,0,45,0,169,255,220,0,45,0,170,255,220,0,45,0,173,255,220,0,45,0,174,255,220,0,45,0,180,255,183,0,45,0,181,255,193,0,45,0,197,255,144,0,45,0,199,255,220,0,45,0,201,255,220,0,46,0,16,255,41,0,46,0,36,255,220,0,46,0,38,255,144,0,46,0,50,255,144,0,46,0,55,255,97,0,46,0,56,255,201,0,46,0,58,255,183,0,46,0,60,255,183,0,46,0,68,255,220,0,46,0,72,255,154,0,46,0,82,255,154,0,46,0,88,255,154,0,46,0,92,255,107,0,46,0,98,255,220,0,46,0,100,255,144,0,46,0,103,255,144,0,46,0,104,255,201,0,46,0,105,255,220,0,46,0,106,255,220,0,46,0,107,255,220,0,46,0,108,255,220,0,46,0,109,255,220,0,46,0,110,255,220,0,46,0,112,255,154,0,46,0,113,255,154,0,46,0,114,255,154,0,46,0,115,255,154,0,46,0,121,255,154,0,46,0,122,255,154,0,46,0,123,255,154,0,46,0,124,255,154,0,46,0,125,255,154,0,46,0,126,255,154,0,46,0,127,255,154,0,46,0,128,255,154,0,46,0,129,255,154,0,46,0,169,255,125,0,46,0,170,0,0,0,46,0,173,255,220,0,46,0,174,255,220,0,46,0,175,255,144,0,46,0,180,255,193,0,46,0,181,255,193,0,46,0,186,255,107,0,46,0,187,255,183,0,46,0,197,0,0,0,46,0,199,255,220,0,46,0,201,255,220,0,46,0,208,255,144,0,46,0,209,255,144,0,46,0,210,255,144,0,46,0,211,255,201,0,46,0,212,255,201,0,46,0,213,255,201,0,46,0,234,255,183,0,46,0,235,255,107,0,46,0,251,255,144,0,46,0,253,255,144,0,47,0,16,255,220,0,47,0,36,0,47,0,47,0,50,255,183,0,47,0,55,254,230,0,47,0,56,255,154,0,47,0,57,255,31,0,47,0,58,255,68,0,47,0,60,254,240,0,47,0,68,0,0,0,47,0,72,255,220,0,47,0,82,255,220,0,47,0,88,255,220,0,47,0,92,255,68,0,47,0,98,0,47,0,47,0,103,255,183,0,47,0,104,255,154,0,47,0,105,0,0,0,47,0,106,0,0,0,47,0,107,0,0,0,47,0,108,0,0,0,47,0,109,0,0,0,47,0,110,0,0,0,47,0,112,255,220,0,47,0,113,255,220,0,47,0,114,255,220,0,47,0,115,255,220,0,47,0,121,255,220,0,47,0,122,255,220,0,47,0,123,255,220,0,47,0,124,255,220,0,47,0,125,255,220,0,47,0,126,255,220,0,47,0,127,255,220,0,47,0,128,255,220,0,47,0,129,255,220,0,47,0,169,0,0,0,47,0,170,0,0,0,47,0,173,0,47,0,47,0,174,0,47,0,47,0,175,255,183,0,47,0,180,254,97,0,47,0,181,253,230,0,47,0,186,255,68,0,47,0,187,254,240,0,47,0,197,0,0,0,47,0,199,0,47,0,47,0,201,0,47,0,47,0,208,255,183,0,47,0,209,255,183,0,47,0,210,255,183,0,47,0,211,255,154,0,47,0,212,255,154,0,47,0,213,255,154,0,47,0,234,254,240,0,47,0,235,255,68,0,50,0,16,0,57,0,50,0,17,255,173,0,50,0,29,255,220,0,50,0,36,255,220,0,50,0,57,255,220,0,50,0,59,255,125,0,50,0,60,255,144,0,50,0,98,255,220,0,50,0,169,255,220,0,50,0,170,0,0,0,50,0,173,255,220,0,50,0,174,255,220,0,50,0,180,255,211,0,50,0,181,255,220,0,50,0,187,255,144,0,50,0,197,255,68,0,50,0,199,255,220,0,50,0,201,255,220,0,50,0,234,255,144,0,51,0,16,255,211,0,51,0,17,254,193,0,51,0,29,0,0,0,51,0,36,255,125,0,51,0,56,0,0,0,51,0,58,0,0,0,51,0,60,255,211,0,51,0,68,255,164,0,51,0,72,255,183,0,51,0,76,255,211,0,51,0,81,255,220,0,51,0,82,255,183,0,51,0,85,255,220,0,51,0,86,255,220,0,51,0,88,255,220,0,51,0,92,0,0,0,51,0,98,255,125,0,51,0,104,0,0,0,51,0,105,255,164,0,51,0,106,255,164,0,51,0,107,255,164,0,51,0,108,255,164,0,51,0,109,255,164,0,51,0,110,255,164,0,51,0,112,255,183,0,51,0,113,255,183,0,51,0,114,255,183,0,51,0,115,255,183,0,51,0,120,255,220,0,51,0,121,255,183,0,51,0,122,255,183,0,51,0,123,255,183,0,51,0,124,255,183,0,51,0,125,255,183,0,51,0,126,255,220,0,51,0,127,255,220,0,51,0,128,255,220,0,51,0,129,255,220,0,51,0,169,255,220,0,51,0,170,0,0,0,51,0,173,255,125,0,51,0,174,255,125,0,51,0,180,0,38,0,51,0,181,0,38,0,51,0,186,0,0,0,51,0,187,255,211,0,51,0,197,254,183,0,51,0,199,255,125,0,51,0,201,255,125,0,51,0,211,0,0,0,51,0,212,0,0,0,51,0,213,0,0,0,51,0,228,255,220,0,51,0,234,255,211,0,51,0,235,0,0,0,51,0,250,255,220,0,52,0,16,0,57,0,52,0,169,0,0,0,52,0,170,0,0,0,52,0,180,255,211,0,52,0,181,255,220,0,52,0,197,255,125,0,53,0,16,255,173,0,53,0,17,255,183,0,53,0,29,255,193,0,53,0,36,255,173,0,53,0,38,255,154,0,53,0,55,255,107,0,53,0,57,255,144,0,53,0,58,255,173,0,53,0,60,255,125,0,53,0,68,255,211,0,53,0,72,255,164,0,53,0,82,255,164,0,53,0,88,255,164,0,53,0,92,255,144,0,53,0,98,255,173,0,53,0,100,255,154,0,53,0,105,255,211,0,53,0,106,255,211,0,53,0,107,255,211,0,53,0,108,255,211,0,53,0,109,255,211,0,53,0,110,255,211,0,53,0,112,255,164,0,53,0,113,255,164,0,53,0,114,255,164,0,53,0,115,255,164,0,53,0,121,255,164,0,53,0,122,255,164,0,53,0,123,255,164,0,53,0,124,255,164,0,53,0,125,255,164,0,53,0,126,255,164,0,53,0,127,255,164,0,53,0,128,255,164,0,53,0,129,255,164,0,53,0,169,255,144,0,53,0,170,255,220,0,53,0,173,255,173,0,53,0,174,255,173,0,53,0,180,255,107,0,53,0,181,255,125,0,53,0,186,255,144,0,53,0,187,255,125,0,53,0,197,255,220,0,53,0,199,255,173,0,53,0,201,255,173,0,53,0,234,255,125,0,53,0,235,255,144,0,53,0,251,255,154,0,53,0,253,255,154,0,54,0,36,0,38,0,54,0,38,0,0,0,54,0,42,0,0,0,54,0,50,0,0,0,54,0,52,0,0,0,54,0,54,0,0,0,54,0,98,0,38,0,54,0,100,0,0,0,54,0,103,0,0,0,54,0,173,0,38,0,54,0,174,0,38,0,54,0,175,0,0,0,54,0,199,0,38,0,54,0,201,0,38,0,54,0,208,0,0,0,54,0,209,0,0,0,54,0,210,0,0,0,54,0,227,0,0,0,54,0,246,0,0,0,54,0,249,0,0,0,54,0,251,0,0,0,54,0,253,0,0,0,55,0,16,255,68,0,55,0,17,255,13,0,55,0,29,255,31,0,55,0,36,255,97,0,55,0,38,255,136,0,55,0,55,255,220,0,55,0,68,254,173,0,55,0,70,254,164,0,55,0,72,254,164,0,55,0,76,255,193,0,55,0,82,254,164,0,55,0,85,254,211,0,55,0,86,254,173,0,55,0,88,254,201,0,55,0,90,254,173,0,55,0,92,254,193,0,55,0,98,255,97,0,55,0,100,255,136,0,55,0,105,254,173,0,55,0,106,254,173,0,55,0,107,254,173,0,55,0,108,254,173,0,55,0,109,254,173,0,55,0,110,254,173,0,55,0,111,254,164,0,55,0,112,254,164,0,55,0,113,254,164,0,55,0,114,254,164,0,55,0,115,254,164,0,55,0,121,254,164,0,55,0,122,254,164,0,55,0,123,254,164,0,55,0,124,254,164,0,55,0,125,254,164,0,55,0,126,254,201,0,55,0,127,254,201,0,55,0,128,254,201,0,55,0,129,254,201,0,55,0,169,255,68,0,55,0,170,255,144,0,55,0,173,255,97,0,55,0,174,255,97,0,55,0,180,0,0,0,55,0,181,255,211,0,55,0,186,254,193,0,55,0,197,254,248,0,55,0,199,255,97,0,55,0,201,255,97,0,55,0,228,254,173,0,55,0,235,254,193,0,55,0,250,254,173,0,55,0,251,255,136,0,55,0,252,254,164,0,55,0,253,255,136,0,55,0,254,254,164,0,56,0,36,0,0,0,56,0,45,0,0,0,56,0,61,255,220,0,56,0,98,0,0,0,56,0,173,0,0,0,56,0,174,0,0,0,56,0,199,0,0,0,56,0,201,0,0,0,56,0,229,255,220,0,57,0,16,255,136,0,57,0,17,254,248,0,57,0,29,255,89,0,57,0,36,255,125,0,57,0,50,255,220,0,57,0,68,255,97,0,57,0,72,255,97,0,57,0,76,255,211,0,57,0,82,255,97,0,57,0,88,255,117,0,57,0,92,255,201,0,57,0,98,255,125,0,57,0,103,255,220,0,57,0,105,255,97,0,57,0,106,255,97,0,57,0,107,255,97,0,57,0,108,255,97,0,57,0,109,255,97,0,57,0,110,255,97,0,57,0,112,255,97,0,57,0,113,255,97,0,57,0,114,255,97,0,57,0,115,255,97,0,57,0,121,255,97,0,57,0,122,255,97,0,57,0,123,255,97,0,57,0,124,255,97,0,57,0,125,255,97,0,57,0,126,255,117,0,57,0,127,255,117,0,57,0,128,255,117,0,57,0,129,255,117,0,57,0,169,255,78,0,57,0,170,255,144,0,57,0,173,255,125,0,57,0,174,255,125,0,57,0,175,255,220,0,57,0,180,0,0,0,57,0,181,0,0,0,57,0,186,255,201,0,57,0,197,254,230,0,57,0,199,255,125,0,57,0,201,255,125,0,57,0,208,255,220,0,57,0,209,255,220,0,57,0,210,255,220,0,57,0,235,255,201,0,58,0,16,255,173,0,58,0,17,255,21,0,58,0,29,255,136,0,58,0,36,255,144,0,58,0,68,255,125,0,58,0,72,255,136,0,58,0,76,255,211,0,58,0,82,255,136,0,58,0,85,255,164,0,58,0,88,255,183,0,58,0,92,255,220,0,58,0,98,255,144,0,58,0,105,255,125,0,58,0,106,255,125,0,58,0,107,255,125,0,58,0,108,255,125,0,58,0,109,255,125,0,58,0,110,255,125,0,58,0,112,255,136,0,58,0,113,255,136,0,58,0,114,255,136,0,58,0,115,255,136,0,58,0,121,255,136,0,58,0,122,255,136,0,58,0,123,255,136,0,58,0,124,255,136,0,58,0,125,255,136,0,58,0,126,255,183,0,58,0,127,255,183,0,58,0,128,255,183,0,58,0,129,255,183,0,58,0,169,255,144,0,58,0,170,255,220,0,58,0,173,255,144,0,58,0,174,255,144,0,58,0,180,255,220,0,58,0,181,0,0,0,58,0,186,255,220,0,58,0,197,254,248,0,58,0,199,255,144,0,58,0,201,255,144,0,58,0,235,255,220,0,59,0,16,255,154,0,59,0,36,0,0,0,59,0,38,255,107,0,59,0,50,255,125,0,59,0,55,255,220,0,59,0,72,255,164,0,59,0,98,0,0,0,59,0,100,255,107,0,59,0,103,255,125,0,59,0,112,255,164,0,59,0,113,255,164,0,59,0,114,255,164,0,59,0,115,255,164,0,59,0,169,255,144,0,59,0,170,0,0,0,59,0,173,0,0,0,59,0,174,0,0,0,59,0,175,255,125,0,59,0,180,255,97,0,59,0,181,255,173,0,59,0,197,255,211,0,59,0,199,0,0,0,59,0,201,0,0,0,59,0,208,255,125,0,59,0,209,255,125,0,59,0,210,255,125,0,59,0,251,255,107,0,59,0,253,255,107,0,60,0,16,255,13,0,60,0,17,254,97,0,60,0,29,254,240,0,60,0,36,255,97,0,60,0,38,255,144,0,60,0,50,255,144,0,60,0,68,254,230,0,60,0,72,254,240,0,60,0,76,255,183,0,60,0,82,254,240,0,60,0,88,255,21,0,60,0,98,255,97,0,60,0,100,255,144,0,60,0,103,255,144,0,60,0,105,254,230,0,60,0,106,254,230,0,60,0,107,254,230,0,60,0,108,254,230,0,60,0,109,254,230,0,60,0,110,254,230,0,60,0,112,254,240,0,60,0,113,254,240,0,60,0,114,254,240,0,60,0,115,254,240,0,60,0,121,254,240,0,60,0,122,254,240,0,60,0,123,254,240,0,60,0,124,254,240,0,60,0,125,254,240,0,60,0,126,255,21,0,60,0,127,255,21,0,60,0,128,255,21,0,60,0,129,255,21,0,60,0,169,255,31,0,60,0,170,255,107,0,60,0,173,255,97,0,60,0,174,255,97,0,60,0,175,255,144,0,60,0,180,255,144,0,60,0,181,255,220,0,60,0,197,254,248,0,60,0,199,255,97,0,60,0,201,255,97,0,60,0,208,255,144,0,60,0,209,255,144,0,60,0,210,255,144,0,60,0,251,255,144,0,60,0,253,255,144,0,61,0,16,255,220,0,61,0,169,0,0,0,61,0,170,0,0,0,61,0,180,255,220,0,61,0,181,255,220,0,61,0,197,255,220,0,72,0,91,255,220,0,73,0,16,255,144,0,73,0,17,255,107,0,73,0,29,255,183,0,73,0,87,255,220,0,73,0,90,255,220,0,73,0,92,255,220,0,73,0,169,255,183,0,73,0,170,255,220,0,73,0,180,0,65,0,73,0,181,0,0,0,73,0,186,255,220,0,73,0,197,255,21,0,73,0,235,255,220,0,78,0,68,255,220,0,78,0,72,255,183,0,78,0,82,255,183,0,78,0,88,255,193,0,78,0,92,255,183,0,78,0,105,255,220,0,78,0,106,255,220,0,78,0,107,255,220,0,78,0,108,255,220,0,78,0,109,255,220,0,78,0,110,255,220,0,78,0,112,255,183,0,78,0,113,255,183,0,78,0,114,255,183,0,78,0,115,255,183,0,78,0,121,255,183,0,78,0,122,255,183,0,78,0,123,255,183,0,78,0,124,255,183,0,78,0,125,255,183,0,78,0,126,255,193,0,78,0,127,255,193,0,78,0,128,255,193,0,78,0,129,255,193,0,78,0,186,255,183,0,78,0,235,255,183,0,81,0,16,0,0,0,81,0,17,0,0,0,81,0,29,0,0,0,81,0,169,0,0,0,81,0,170,0,0,0,81,0,180,255,107,0,81,0,181,255,144,0,81,0,197,255,164,0,82,0,16,0,38,0,82,0,17,255,220,0,82,0,29,0,0,0,82,0,91,255,193,0,82,0,169,0,0,0,82,0,170,0,0,0,82,0,180,255,107,0,82,0,181,255,183,0,82,0,197,255,125,0,85,0,16,255,125,0,85,0,17,255,68,0,85,0,29,255,220,0,85,0,70,255,211,0,85,0,71,255,220,0,85,0,72,255,211,0,85,0,73,0,0,0,85,0,74,255,220,0,85,0,75,255,220,0,85,0,80,255,220,0,85,0,81,255,220,0,85,0,82,255,211,0,85,0,84,255,220,0,85,0,85,255,220,0,85,0,88,0,0,0,85,0,89,0,0,0,85,0,90,0,0,0,85,0,91,255,201,0,85,0,92,0,0,0,85,0,93,0,0,0,85,0,111,255,211,0,85,0,112,255,211,0,85,0,113,255,211,0,85,0,114,255,211,0,85,0,115,255,211,0,85,0,120,255,220,0,85,0,121,255,211,0,85,0,122,255,211,0,85,0,123,255,211,0,85,0,124,255,211,0,85,0,125,255,211,0,85,0,126,0,0,0,85,0,127,0,0,0,85,0,128,0,0,0,85,0,129,0,0,0,85,0,169,255,183,0,85,0,170,0,0,0,85,0,180,0,0,0,85,0,181,0,86,0,85,0,186,0,0,0,85,0,197,254,201,0,85,0,230,0,0,0,85,0,235,0,0,0,85,0,247,255,220,0,85,0,252,255,211,0,85,0,254,255,211,0,89,0,16,255,201,0,89,0,17,255,97,0,89,0,29,255,144,0,89,0,169,255,220,0,89,0,170,255,220,0,89,0,180,0,0,0,89,0,181,255,220,0,89,0,197,254,240,0,90,0,16,0,0,0,90,0,17,255,68,0,90,0,29,255,144,0,90,0,169,255,220,0,90,0,170,255,220,0,90,0,180,0,0,0,90,0,181,0,0,0,90,0,197,255,41,0,91,0,70,255,220,0,91,0,72,255,193,0,91,0,82,255,193,0,91,0,111,255,220,0,91,0,112,255,193,0,91,0,113,255,193,0,91,0,114,255,193,0,91,0,115,255,193,0,91,0,121,255,193,0,91,0,122,255,193,0,91,0,123,255,193,0,91,0,124,255,193,0,91,0,125,255,193,0,91,0,252,255,220,0,91,0,254,255,220,0,92,0,16,255,220,0,92,0,17,254,220,0,92,0,29,255,107,0,92,0,169,255,220,0,92,0,170,255,220,0,92,0,180,0,0,0,92,0,181,0,0,0,92,0,197,254,211,0,98,0,16,255,211,0,98,0,17,255,220,0,98,0,29,255,220,0,98,0,36,0,57,0,98,0,38,255,220,0,98,0,42,255,220,0,98,0,50,255,220,0,98,0,52,255,220,0,98,0,54,0,0,0,98,0,55,255,97,0,98,0,56,0,0,0,98,0,57,255,125,0,98,0,58,255,144,0,98,0,59,0,0,0,98,0,60,255,97,0,98,0,70,255,220,0,98,0,71,255,220,0,98,0,72,255,220,0,98,0,73,255,183,0,98,0,82,255,220,0,98,0,84,255,220,0,98,0,87,255,220,0,98,0,88,0,0,0,98,0,89,255,136,0,98,0,90,255,173,0,98,0,92,255,117,0,98,0,98,0,57,0,98,0,100,255,220,0,98,0,103,255,220,0,98,0,104,0,0,0,98,0,111,255,220,0,98,0,112,255,220,0,98,0,113,255,220,0,98,0,114,255,220,0,98,0,115,255,220,0,98,0,121,255,220,0,98,0,122,255,220,0,98,0,123,255,220,0,98,0,124,255,220,0,98,0,125,255,220,0,98,0,126,0,0,0,98,0,127,0,0,0,98,0,128,0,0,0,98,0,129,0,0,0,98,0,169,255,183,0,98,0,170,0,0,0,98,0,173,0,57,0,98,0,174,0,57,0,98,0,175,255,220,0,98,0,180,254,248,0,98,0,181,255,3,0,98,0,186,255,117,0,98,0,187,255,97,0,98,0,197,0,47,0,98,0,199,0,57,0,98,0,201,0,57,0,98,0,208,255,220,0,98,0,209,255,220,0,98,0,210,255,220,0,98,0,211,0,0,0,98,0,212,0,0,0,98,0,213,0,0,0,98,0,227,0,0,0,98,0,234,255,97,0,98,0,235,255,117,0,98,0,246,255,220,0,98,0,249,0,0,0,98,0,251,255,220,0,98,0,252,255,220,0,98,0,253,255,220,0,98,0,254,255,220,0,100,0,16,0,0,0,100,0,36,0,0,0,100,0,54,0,0,0,100,0,60,255,220,0,100,0,98,0,0,0,100,0,169,255,220,0,100,0,170,255,220,0,100,0,173,0,0,0,100,0,174,0,0,0,100,0,180,0,0,0,100,0,181,0,38,0,100,0,187,255,220,0,100,0,197,0,0,0,100,0,199,0,0,0,100,0,201,0,0,0,100,0,227,0,0,0,100,0,234,255,220,0,100,0,249,0,0,0,103,0,16,0,57,0,103,0,17,255,173,0,103,0,29,255,220,0,103,0,36,255,220,0,103,0,57,255,220,0,103,0,59,255,125,0,103,0,60,255,144,0,103,0,98,255,220,0,103,0,169,255,220,0,103,0,170,0,0,0,103,0,173,255,220,0,103,0,174,255,220,0,103,0,180,255,211,0,103,0,181,255,220,0,103,0,187,255,144,0,103,0,197,255,68,0,103,0,199,255,220,0,103,0,201,255,220,0,103,0,234,255,144,0,104,0,36,0,0,0,104,0,45,0,0,0,104,0,61,255,220,0,104,0,98,0,0,0,104,0,173,0,0,0,104,0,174,0,0,0,104,0,199,0,0,0,104,0,201,0,0,0,104,0,229,255,220,0,112,0,91,255,220,0,113,0,91,255,220,0,114,0,91,255,220,0,115,0,91,255,220,0,120,0,16,0,0,0,120,0,17,0,0,0,120,0,29,0,0,0,120,0,169,0,0,0,120,0,170,0,0,0,120,0,180,255,107,0,120,0,181,255,144,0,120,0,197,255,164,0,121,0,16,0,38,0,121,0,17,255,220,0,121,0,29,0,0,0,121,0,91,255,193,0,121,0,169,0,0,0,121,0,170,0,0,0,121,0,180,255,107,0,121,0,181,255,183,0,121,0,197,255,125,0,122,0,16,0,38,0,122,0,17,255,220,0,122,0,29,0,0,0,122,0,91,255,193,0,122,0,169,0,0,0,122,0,170,0,0,0,122,0,180,255,107,0,122,0,181,255,183,0,122,0,197,255,125,0,123,0,16,0,38,0,123,0,17,255,220,0,123,0,29,0,0,0,123,0,91,255,193,0,123,0,169,0,0,0,123,0,170,0,0,0,123,0,180,255,107,0,123,0,181,255,183,0,123,0,197,255,125,0,124,0,16,0,38,0,124,0,17,255,220,0,124,0,29,0,0,0,124,0,91,255,193,0,124,0,169,0,0,0,124,0,170,0,0,0,124,0,180,255,107,0,124,0,181,255,183,0,124,0,197,255,125,0,125,0,16,0,38,0,125,0,17,255,220,0,125,0,29,0,0,0,125,0,91,255,193,0,125,0,169,0,0,0,125,0,170,0,0,0,125,0,180,255,107,0,125,0,181,255,183,0,125,0,197,255,125,0,137,0,16,0,38,0,137,0,169,0,0,0,137,0,170,0,0,0,137,0,180,255,144,0,137,0,181,255,144,0,137,0,197,255,173,0,144,0,16,0,0,0,144,0,169,0,0,0,144,0,170,0,0,0,144,0,180,255,173,0,144,0,181,255,164,0,144,0,197,255,144,0,169,0,36,0,0,0,169,0,37,255,220,0,169,0,38,255,220,0,169,0,39,255,220,0,169,0,41,0,0,0,169,0,42,255,220,0,169,0,43,0,0,0,169,0,45,255,220,0,169,0,46,0,0,0,169,0,47,0,0,0,169,0,50,0,0,0,169,0,51,0,0,0,169,0,52,0,0,0,169,0,53,0,0,0,169,0,55,255,144,0,169,0,57,255,144,0,169,0,58,255,220,0,169,0,59,0,0,0,169,0,60,255,107,0,169,0,61,0,0,0,169,0,73,0,0,0,169,0,81,0,0,0,169,0,82,0,0,0,169,0,85,0,0,0,169,0,89,255,220,0,169,0,90,255,220,0,169,0,92,255,220,0,169,0,98,0,0,0,169,0,100,255,220,0,169,0,103,0,0,0,169,0,120,0,0,0,169,0,121,0,0,0,169,0,122,0,0,0,169,0,123,0,0,0,169,0,124,0,0,0,169,0,125,0,0,0,169,0,137,0,0,0,169,0,144,0,151,0,169,0,173,0,0,0,169,0,174,0,0,0,169,0,175,0,0,0,169,0,186,255,220,0,169,0,187,255,107,0,169,0,199,0,0,0,169,0,201,0,0,0,169,0,208,0,0,0,169,0,209,0,0,0,169,0,210,0,0,0,169,0,229,0,0,0,169,0,233,0,0,0,169,0,234,255,107,0,169,0,235,255,220,0,169,0,236,0,0,0,169,0,246,255,220,0,169,0,251,255,220,0,169,0,253,255,220,0,170,0,36,255,183,0,170,0,37,255,183,0,170,0,38,255,220,0,170,0,39,255,220,0,170,0,41,0,0,0,170,0,42,0,0,0,170,0,43,0,0,0,170,0,45,255,220,0,170,0,46,0,0,0,170,0,47,0,0,0,170,0,50,255,220,0,170,0,51,0,0,0,170,0,52,0,0,0,170,0,53,0,0,0,170,0,55,255,68,0,170,0,57,255,78,0,170,0,58,255,144,0,170,0,59,255,144,0,170,0,60,255,31,0,170,0,61,0,0,0,170,0,73,0,0,0,170,0,81,0,0,0,170,0,82,0,0,0,170,0,85,0,0,0,170,0,89,255,220,0,170,0,90,255,220,0,170,0,92,255,220,0,170,0,98,255,183,0,170,0,100,255,220,0,170,0,103,255,220,0,170,0,120,0,0,0,170,0,121,0,0,0,170,0,122,0,0,0,170,0,123,0,0,0,170,0,124,0,0,0,170,0,125,0,0,0,170,0,137,0,0,0,170,0,144,0,0,0,170,0,173,255,183,0,170,0,174,255,183,0,170,0,175,255,220,0,170,0,186,255,220,0,170,0,187,255,31,0,170,0,199,255,183,0,170,0,201,255,183,0,170,0,208,255,220,0,170,0,209,255,220,0,170,0,210,255,220,0,170,0,229,0,0,0,170,0,233,0,0,0,170,0,234,255,31,0,170,0,235,255,220,0,170,0,236,0,0,0,170,0,246,0,0,0,170,0,251,255,220,0,170,0,253,255,220,0,173,0,16,255,211,0,173,0,17,255,220,0,173,0,29,255,220,0,173,0,36,0,57,0,173,0,38,255,220,0,173,0,42,255,220,0,173,0,50,255,220,0,173,0,52,255,220,0,173,0,54,0,0,0,173,0,55,255,97,0,173,0,56,0,0,0,173,0,57,255,125,0,173,0,58,255,144,0,173,0,59,0,0,0,173,0,60,255,97,0,173,0,70,255,220,0,173,0,71,255,220,0,173,0,72,255,220,0,173,0,73,255,183,0,173,0,82,255,220,0,173,0,84,255,220,0,173,0,87,255,220,0,173,0,88,0,0,0,173,0,89,255,136,0,173,0,90,255,173,0,173,0,92,255,117,0,173,0,98,0,57,0,173,0,100,255,220,0,173,0,103,255,220,0,173,0,104,0,0,0,173,0,111,255,220,0,173,0,112,255,220,0,173,0,113,255,220,0,173,0,114,255,220,0,173,0,115,255,220,0,173,0,121,255,220,0,173,0,122,255,220,0,173,0,123,255,220,0,173,0,124,255,220,0,173,0,125,255,220,0,173,0,126,0,0,0,173,0,127,0,0,0,173,0,128,0,0,0,173,0,129,0,0,0,173,0,169,255,183,0,173,0,170,0,0,0,173,0,173,0,57,0,173,0,174,0,57,0,173,0,175,255,220,0,173,0,180,254,248,0,173,0,181,255,3,0,173,0,186,255,117,0,173,0,187,255,97,0,173,0,197,0,47,0,173,0,199,0,57,0,173,0,201,0,57,0,173,0,208,255,220,0,173,0,209,255,220,0,173,0,210,255,220,0,173,0,211,0,0,0,173,0,212,0,0,0,173,0,213,0,0,0,173,0,227,0,0,0,173,0,234,255,97,0,173,0,235,255,117,0,173,0,246,255,220,0,173,0,249,0,0,0,173,0,251,255,220,0,173,0,252,255,220,0,173,0,253,255,220,0,173,0,254,255,220,0,174,0,16,255,211,0,174,0,17,255,220,0,174,0,29,255,220,0,174,0,36,0,57,0,174,0,38,255,220,0,174,0,42,255,220,0,174,0,50,255,220,0,174,0,52,255,220,0,174,0,54,0,0,0,174,0,55,255,97,0,174,0,56,0,0,0,174,0,57,255,125,0,174,0,58,255,144,0,174,0,59,0,0,0,174,0,60,255,97,0,174,0,70,255,220,0,174,0,71,255,220,0,174,0,72,255,220,0,174,0,73,255,183,0,174,0,82,255,220,0,174,0,84,255,220,0,174,0,87,255,220,0,174,0,88,0,0,0,174,0,89,255,136,0,174,0,90,255,173,0,174,0,92,255,117,0,174,0,98,0,57,0,174,0,100,255,220,0,174,0,103,255,220,0,174,0,104,0,0,0,174,0,111,255,220,0,174,0,112,255,220,0,174,0,113,255,220,0,174,0,114,255,220,0,174,0,115,255,220,0,174,0,121,255,220,0,174,0,122,255,220,0,174,0,123,255,220,0,174,0,124,255,220,0,174,0,125,255,220,0,174,0,126,0,0,0,174,0,127,0,0,0,174,0,128,0,0,0,174,0,129,0,0,0,174,0,169,255,183,0,174,0,170,0,0,0,174,0,173,0,57,0,174,0,174,0,57,0,174,0,175,255,220,0,174,0,180,254,248,0,174,0,181,255,3,0,174,0,186,255,117,0,174,0,187,255,97,0,174,0,197,0,47,0,174,0,199,0,57,0,174,0,201,0,57,0,174,0,208,255,220,0,174,0,209,255,220,0,174,0,210,255,220,0,174,0,211,0,0,0,174,0,212,0,0,0,174,0,213,0,0,0,174,0,227,0,0,0,174,0,234,255,97,0,174,0,235,255,117,0,174,0,246,255,220,0,174,0,249,0,0,0,174,0,251,255,220,0,174,0,252,255,220,0,174,0,253,255,220,0,174,0,254,255,220,0,175,0,16,0,57,0,175,0,17,255,173,0,175,0,29,255,220,0,175,0,36,255,220,0,175,0,57,255,220,0,175,0,59,255,125,0,175,0,60,255,144,0,175,0,98,255,220,0,175,0,169,255,220,0,175,0,170,0,0,0,175,0,173,255,220,0,175,0,174,255,220,0,175,0,180,255,211,0,175,0,181,255,220,0,175,0,187,255,144,0,175,0,197,255,68,0,175,0,199,255,220,0,175,0,201,255,220,0,175,0,234,255,144,0,180,0,36,254,248,0,180,0,37,255,193,0,180,0,38,255,183,0,180,0,39,255,193,0,180,0,41,255,193,0,180,0,42,255,183,0,180,0,43,255,193,0,180,0,45,255,193,0,180,0,46,255,193,0,180,0,47,255,193,0,180,0,50,255,183,0,180,0,51,255,193,0,180,0,52,255,183,0,180,0,53,255,193,0,180,0,55,0,0,0,180,0,57,0,0,0,180,0,58,0,0,0,180,0,59,255,136,0,180,0,60,0,0,0,180,0,61,255,220,0,180,0,73,255,183,0,180,0,81,255,144,0,180,0,82,255,107,0,180,0,85,255,144,0,180,0,89,255,183,0,180,0,90,255,183,0,180,0,92,255,183,0,180,0,98,254,248,0,180,0,100,255,183,0,180,0,103,255,183,0,180,0,120,255,144,0,180,0,121,255,107,0,180,0,122,255,107,0,180,0,123,255,107,0,180,0,124,255,107,0,180,0,125,255,107,0,180,0,137,255,193,0,180,0,144,254,125,0,180,0,173,254,248,0,180,0,174,254,248,0,180,0,175,255,183,0,180,0,186,255,183,0,180,0,187,0,0,0,180,0,199,254,248,0,180,0,201,254,248,0,180,0,208,255,183,0,180,0,209,255,183,0,180,0,210,255,183,0,180,0,229,255,220,0,180,0,233,255,183,0,180,0,234,0,0,0,180,0,235,255,183,0,180,0,236,255,193,0,180,0,246,255,183,0,180,0,251,255,183,0,180,0,253,255,183,0,186,0,16,255,220,0,186,0,17,254,220,0,186,0,29,255,107,0,186,0,169,255,220,0,186,0,170,255,220,0,186,0,180,0,0,0,186,0,181,0,0,0,186,0,197,254,211,0,187,0,16,255,13,0,187,0,17,254,97,0,187,0,29,254,240,0,187,0,36,255,97,0,187,0,38,255,144,0,187,0,50,255,144,0,187,0,68,254,230,0,187,0,72,254,240,0,187,0,76,255,183,0,187,0,82,254,240,0,187,0,88,255,21,0,187,0,98,255,97,0,187,0,100,255,144,0,187,0,103,255,144,0,187,0,105,254,230,0,187,0,106,254,230,0,187,0,107,254,230,0,187,0,108,254,230,0,187,0,109,254,230,0,187,0,110,254,230,0,187,0,112,254,240,0,187,0,113,254,240,0,187,0,114,254,240,0,187,0,115,254,240,0,187,0,121,254,240,0,187,0,122,254,240,0,187,0,123,254,240,0,187,0,124,254,240,0,187,0,125,254,240,0,187,0,126,255,21,0,187,0,127,255,21,0,187,0,128,255,21,0,187,0,129,255,21,0,187,0,169,255,31,0,187,0,170,255,107,0,187,0,173,255,97,0,187,0,174,255,97,0,187,0,175,255,144,0,187,0,180,255,144,0,187,0,181,255,220,0,187,0,197,254,248,0,187,0,199,255,97,0,187,0,201,255,97,0,187,0,208,255,144,0,187,0,209,255,144,0,187,0,210,255,144,0,187,0,251,255,144,0,187,0,253,255,144,0,197,0,36,0,38,0,197,0,37,255,183,0,197,0,38,255,144,0,197,0,39,255,183,0,197,0,41,255,183,0,197,0,42,255,183,0,197,0,43,255,183,0,197,0,45,0,47,0,197,0,46,255,183,0,197,0,47,255,183,0,197,0,50,255,144,0,197,0,51,255,183,0,197,0,52,255,144,0,197,0,53,255,183,0,197,0,55,254,230,0,197,0,57,254,136,0,197,0,58,255,3,0,197,0,59,255,183,0,197,0,60,254,136,0,197,0,61,0,0,0,197,0,73,255,220,0,197,0,81,255,183,0,197,0,82,255,183,0,197,0,85,255,183,0,197,0,89,255,21,0,197,0,90,255,60,0,197,0,92,255,144,0,197,0,98,0,38,0,197,0,100,255,144,0,197,0,103,255,144,0,197,0,120,255,183,0,197,0,121,255,183,0,197,0,122,255,183,0,197,0,123,255,183,0,197,0,124,255,183,0,197,0,125,255,183,0,197,0,137,255,183,0,197,0,144,0,38,0,197,0,173,0,38,0,197,0,174,0,38,0,197,0,175,255,144,0,197,0,186,255,144,0,197,0,187,254,136,0,197,0,199,0,38,0,197,0,201,0,38,0,197,0,208,255,144,0,197,0,209,255,144,0,197,0,210,255,144,0,197,0,229,0,0,0,197,0,233,255,183,0,197,0,234,254,136,0,197,0,235,255,144,0,197,0,236,255,183,0,197,0,246,255,183,0,197,0,251,255,144,0,197,0,253,255,144,0,199,0,16,255,211,0,199,0,17,255,220,0,199,0,29,255,220,0,199,0,36,0,57,0,199,0,38,255,220,0,199,0,42,255,220,0,199,0,50,255,220,0,199,0,52,255,220,0,199,0,54,0,0,0,199,0,55,255,97,0,199,0,56,0,0,0,199,0,57,255,125,0,199,0,58,255,144,0,199,0,59,0,0,0,199,0,60,255,97,0,199,0,70,255,220,0,199,0,71,255,220,0,199,0,72,255,220,0,199,0,73,255,183,0,199,0,82,255,220,0,199,0,84,255,220,0,199,0,87,255,220,0,199,0,88,0,0,0,199,0,89,255,136,0,199,0,90,255,173,0,199,0,92,255,117,0,199,0,98,0,57,0,199,0,100,255,220,0,199,0,103,255,220,0,199,0,104,0,0,0,199,0,111,255,220,0,199,0,112,255,220,0,199,0,113,255,220,0,199,0,114,255,220,0,199,0,115,255,220,0,199,0,121,255,220,0,199,0,122,255,220,0,199,0,123,255,220,0,199,0,124,255,220,0,199,0,125,255,220,0,199,0,126,0,0,0,199,0,127,0,0,0,199,0,128,0,0,0,199,0,129,0,0,0,199,0,169,255,183,0,199,0,170,0,0,0,199,0,173,0,57,0,199,0,174,0,57,0,199,0,175,255,220,0,199,0,180,254,248,0,199,0,181,255,3,0,199,0,186,255,117,0,199,0,187,255,97,0,199,0,197,0,47,0,199,0,199,0,57,0,199,0,201,0,57,0,199,0,208,255,220,0,199,0,209,255,220,0,199,0,210,255,220,0,199,0,211,0,0,0,199,0,212,0,0,0,199,0,213,0,0,0,199,0,227,0,0,0,199,0,234,255,97,0,199,0,235,255,117,0,199,0,246,255,220,0,199,0,249,0,0,0,199,0,251,255,220,0,199,0,252,255,220,0,199,0,253,255,220,0,199,0,254,255,220,0,201,0,16,255,211,0,201,0,17,255,220,0,201,0,29,255,220,0,201,0,36,0,57,0,201,0,38,255,220,0,201,0,42,255,220,0,201,0,50,255,220,0,201,0,52,255,220,0,201,0,54,0,0,0,201,0,55,255,97,0,201,0,56,0,0,0,201,0,57,255,125,0,201,0,58,255,144,0,201,0,59,0,0,0,201,0,60,255,97,0,201,0,70,255,220,0,201,0,71,255,220,0,201,0,72,255,220,0,201,0,73,255,183,0,201,0,82,255,220,0,201,0,84,255,220,0,201,0,87,255,220,0,201,0,88,0,0,0,201,0,89,255,136,0,201,0,90,255,173,0,201,0,92,255,117,0,201,0,98,0,57,0,201,0,100,255,220,0,201,0,103,255,220,0,201,0,104,0,0,0,201,0,111,255,220,0,201,0,112,255,220,0,201,0,113,255,220,0,201,0,114,255,220,0,201,0,115,255,220,0,201,0,121,255,220,0,201,0,122,255,220,0,201,0,123,255,220,0,201,0,124,255,220,0,201,0,125,255,220,0,201,0,126,0,0,0,201,0,127,0,0,0,201,0,128,0,0,0,201,0,129,0,0,0,201,0,169,255,183,0,201,0,170,0,0,0,201,0,173,0,57,0,201,0,174,0,57,0,201,0,175,255,220,0,201,0,180,254,248,0,201,0,181,255,3,0,201,0,186,255,117,0,201,0,187,255,97,0,201,0,197,0,47,0,201,0,199,0,57,0,201,0,201,0,57,0,201,0,208,255,220,0,201,0,209,255,220,0,201,0,210,255,220,0,201,0,211,0,0,0,201,0,212,0,0,0,201,0,213,0,0,0,201,0,227,0,0,0,201,0,234,255,97,0,201,0,235,255,117,0,201,0,246,255,220,0,201,0,249,0,0,0,201,0,251,255,220,0,201,0,252,255,220,0,201,0,253,255,220,0,201,0,254,255,220,0,208,0,16,0,57,0,208,0,17,255,173,0,208,0,29,255,220,0,208,0,36,255,220,0,208,0,57,255,220,0,208,0,59,255,125,0,208,0,60,255,144,0,208,0,98,255,220,0,208,0,169,255,220,0,208,0,170,0,0,0,208,0,173,255,220,0,208,0,174,255,220,0,208,0,180,255,211,0,208,0,181,255,220,0,208,0,187,255,144,0,208,0,197,255,68,0,208,0,199,255,220,0,208,0,201,255,220,0,208,0,234,255,144,0,209,0,16,0,57,0,209,0,17,255,173,0,209,0,29,255,220,0,209,0,36,255,220,0,209,0,57,255,220,0,209,0,59,255,125,0,209,0,60,255,144,0,209,0,98,255,220,0,209,0,169,255,220,0,209,0,170,0,0,0,209,0,173,255,220,0,209,0,174,255,220,0,209,0,180,255,211,0,209,0,181,255,220,0,209,0,187,255,144,0,209,0,197,255,68,0,209,0,199,255,220,0,209,0,201,255,220,0,209,0,234,255,144,0,210,0,16,0,57,0,210,0,17,255,173,0,210,0,29,255,220,0,210,0,36,255,220,0,210,0,57,255,220,0,210,0,59,255,125,0,210,0,60,255,144,0,210,0,98,255,220,0,210,0,169,255,220,0,210,0,170,0,0,0,210,0,173,255,220,0,210,0,174,255,220,0,210,0,180,255,211,0,210,0,181,255,220,0,210,0,187,255,144,0,210,0,197,255,68,0,210,0,199,255,220,0,210,0,201,255,220,0,210,0,234,255,144,0,211,0,36,0,0,0,211,0,45,0,0,0,211,0,61,255,220,0,211,0,98,0,0,0,211,0,173,0,0,0,211,0,174,0,0,0,211,0,199,0,0,0,211,0,201,0,0,0,211,0,229,255,220,0,212,0,36,0,0,0,212,0,45,0,0,0,212,0,61,255,220,0,212,0,98,0,0,0,212,0,173,0,0,0,212,0,174,0,0,0,212,0,199,0,0,0,212,0,201,0,0,0,212,0,229,255,220,0,213,0,36,0,0,0,213,0,45,0,0,0,213,0,61,255,220,0,213,0,98,0,0,0,213,0,173,0,0,0,213,0,174,0,0,0,213,0,199,0,0,0,213,0,201,0,0,0,213,0,229,255,220,0,227,0,36,0,38,0,227,0,38,0,0,0,227,0,42,0,0,0,227,0,50,0,0,0,227,0,52,0,0,0,227,0,54,0,0,0,227,0,98,0,38,0,227,0,100,0,0,0,227,0,103,0,0,0,227,0,173,0,38,0,227,0,174,0,38,0,227,0,175,0,0,0,227,0,199,0,38,0,227,0,201,0,38,0,227,0,208,0,0,0,227,0,209,0,0,0,227,0,210,0,0,0,227,0,227,0,0,0,227,0,246,0,0,0,227,0,249,0,0,0,227,0,251,0,0,0,227,0,253,0,0,0,229,0,16,255,220,0,229,0,169,0,0,0,229,0,170,0,0,0,229,0,180,255,220,0,229,0,181,255,220,0,229,0,197,255,220,0,233,0,16,0,0,0,233,0,169,0,0,0,233,0,170,0,0,0,233,0,180,255,164,0,233,0,181,255,144,0,233,0,197,255,183,0,234,0,16,255,13,0,234,0,17,254,97,0,234,0,29,254,240,0,234,0,36,255,97,0,234,0,38,255,144,0,234,0,50,255,144,0,234,0,68,254,230,0,234,0,72,254,240,0,234,0,76,255,183,0,234,0,82,254,240,0,234,0,88,255,21,0,234,0,98,255,97,0,234,0,100,255,144,0,234,0,103,255,144,0,234,0,105,254,230,0,234,0,106,254,230,0,234,0,107,254,230,0,234,0,108,254,230,0,234,0,109,254,230,0,234,0,110,254,230,0,234,0,112,254,240,0,234,0,113,254,240,0,234,0,114,254,240,0,234,0,115,254,240,0,234,0,121,254,240,0,234,0,122,254,240,0,234,0,123,254,240,0,234,0,124,254,240,0,234,0,125,254,240,0,234,0,126,255,21,0,234,0,127,255,21,0,234,0,128,255,21,0,234,0,129,255,21,0,234,0,169,255,31,0,234,0,170,255,107,0,234,0,173,255,97,0,234,0,174,255,97,0,234,0,175,255,144,0,234,0,180,255,144,0,234,0,181,255,220,0,234,0,197,254,248,0,234,0,199,255,97,0,234,0,201,255,97,0,234,0,208,255,144,0,234,0,209,255,144,0,234,0,210,255,144,0,234,0,251,255,144,0,234,0,253,255,144,0,235,0,16,255,220,0,235,0,17,254,220,0,235,0,29,255,107,0,235,0,169,255,220,0,235,0,170,255,220,0,235,0,180,0,0,0,235,0,181,0,0,0,235,0,197,254,211,0,236,0,16,0,0,0,236,0,17,255,107,0,236,0,29,255,183,0,236,0,169,0,0,0,236,0,170,0,0,0,236,0,180,255,220,0,236,0,181,0,0,0,236,0,197,255,68,0,246,0,16,0,0,0,246,0,36,0,0,0,246,0,55,255,183,0,246,0,58,0,0,0,246,0,60,255,154,0,246,0,98,0,0,0,246,0,169,255,220,0,246,0,170,255,220,0,246,0,173,0,0,0,246,0,174,0,0,0,246,0,180,255,211,0,246,0,181,255,211,0,246,0,187,255,154,0,246,0,197,255,201,0,246,0,199,0,0,0,246,0,201,0,0,0,246,0,234,255,154,0,249,0,36,0,38,0,249,0,38,0,0,0,249,0,42,0,0,0,249,0,50,0,0,0,249,0,52,0,0,0,249,0,54,0,0,0,249,0,98,0,38,0,249,0,100,0,0,0,249,0,103,0,0,0,249,0,173,0,38,0,249,0,174,0,38,0,249,0,175,0,0,0,249,0,199,0,38,0,249,0,201,0,38,0,249,0,208,0,0,0,249,0,209,0,0,0,249,0,210,0,0,0,249,0,227,0,0,0,249,0,246,0,0,0,249,0,249,0,0,0,249,0,251,0,0,0,249,0,253,0,0,0,251,0,16,0,0,0,251,0,36,0,0,0,251,0,54,0,0,0,251,0,60,255,220,0,251,0,98,0,0,0,251,0,169,255,220,0,251,0,170,255,220,0,251,0,173,0,0,0,251,0,174,0,0,0,251,0,180,0,0,0,251,0,181,0,38,0,251,0,187,255,220,0,251,0,197,0,0,0,251,0,199,0,0,0,251,0,201,0,0,0,251,0,227,0,0,0,251,0,234,255,220,0,251,0,249,0,0,0,253,0,16,0,0,0,253,0,36,0,0,0,253,0,54,0,0,0,253,0,60,255,220,0,253,0,98,0,0,0,253,0,169,255,220,0,253,0,170,255,220,0,253,0,173,0,0,0,253,0,174,0,0,0,253,0,180,0,0,0,253,0,181,0,38,0,253,0,187,255,220,0,253,0,197,0,0,0,253,0,199,0,0,0,253,0,201,0,0,0,253,0,227,0,0,0,253,0,234,255,220,0,253,0,249,0,0,0,0,0,1,0,0,1,12,0,77,0,7,0,66,0,4,0,2,0,16,0,64,0,7,0,0,4,21,5,104,0,3,0,1,0,1,0,0,7,109,254,29,0,0,10,188,254,137,254,137,10,76,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,12,0,1,4,14,1,144,0,5,0,4,5,71,4,204,0,0,254,66,5,71,4,204,0,0,2,83,0,143,2,102,8,2,2,11,6,3,3,8,4,2,2,4,128,0,0,175,16,0,32,74,0,0,0,0,0,0,0,0,66,105,116,115,0,64,0,32,251,2,6,20,254,20,1,154,7,109,1,227,0,0,0,1,0,0,0,0,0,0,0,1,0,0,66,0,29,177,2,139,4,96,0,0,35,99,5,213,0,0,86,101,114,97,83,97,110,115,0,0,0,0,0,0,0,0,255,255,255,255,54,255,255,254,54,50,56,82,48,48,0,0,64,0,0,0,0,0,0,20,0,0,1,16,9,12,5,0,3,3,3,4,8,6,9,8,2,4,4,5,8,3,3,3,3,6,6,6,6,6,6,6,6,6,6,3,3,8,8,8,5,10,6,7,7,7,6,6,7,7,3,3,6,5,8,7,7,6,7,6,7,5,7,6,7,6,5,5,4,3,4,8,5,5,6,6,6,6,6,3,6,6,3,3,5,3,9,6,6,6,6,4,5,4,6,5,7,5,5,6,5,3,5,8,6,6,7,6,7,7,7,6,6,6,6,6,6,6,6,6,6,6,3,3,3,3,6,6,6,6,6,6,6,6,6,6,5,5,6,6,5,5,6,6,9,9,9,5,5,8,9,7,8,8,8,8,6,6,5,6,7,5,5,4,4,7,9,6,5,3,8,6,6,8,6,6,6,9,6,6,6,7,10,9,5,9,5,5,3,3,8,4,5,5,2,6,4,4,6,6,5,3,3,5,12,6,6,6,6,6,3,3,3,3,7,7,7,7,7,7,3,5,5,5,5,5,5,5,5,5,5,5,3,7,5,5,6,3,7,6,5,5,6,6,8,8,4,4,4,9,9,9,7,6,3,7,5,7,6,7,6,6,3,3,6,5,5,5,5,5,5,4,5,5,0,0,10,13,6,0,3,3,4,5,8,6,10,9,3,4,4,5,8,3,4,3,3,6,6,6,6,6,6,6,6,6,6,3,3,8,8,8,5,11,7,7,8,8,7,6,8,8,3,3,7,6,9,8,8,7,8,7,7,5,8,7,9,6,7,6,4,3,4,8,5,5,6,6,5,6,6,4,6,6,2,2,5,2,10,6,6,6,6,4,5,4,6,6,8,6,6,5,6,3,6,8,7,7,8,7,8,8,8,6,6,6,6,6,6,5,6,6,6,6,2,2,2,2,6,6,6,6,6,6,6,6,6,6,5,5,6,6,5,6,6,6,10,10,10,5,5,8,10,8,8,8,8,8,6,7,5,7,8,6,5,5,5,8,10,6,5,4,8,6,6,8,7,6,6,10,6,7,7,8,11,10,5,10,5,5,3,3,8,5,6,7,2,6,4,4,6,6,5,3,3,5,13,7,7,7,7,7,3,3,3,3,8,8,8,8,8,8,2,5,5,5,5,5,5,5,5,5,5,6,2,7,5,6,5,3,8,6,7,6,7,6,8,8,4,4,4,10,10,10,8,6,3,7,5,8,5,8,5,6,4,3,6,5,5,5,5,5,5,4,5,5,0,0,11,15,7,0,3,3,3,5,9,7,10,10,3,4,4,6,9,4,4,4,4,7,7,7,7,7,7,7,7,7,7,4,4,9,9,9,6,12,7,8,8,8,7,6,9,8,3,3,7,6,9,8,9,7,9,7,8,7,8,7,9,7,7,9,4,4,4,9,6,6,7,7,6,7,7,4,7,7,3,3,6,3,11,7,7,7,7,5,7,4,7,6,9,6,6,5,7,4,7,9,7,7,8,7,8,9,8,7,7,7,7,7,7,6,7,7,7,7,3,3,3,3,7,7,7,7,7,7,7,7,7,7,6,6,7,7,6,6,7,7,11,11,11,6,6,9,11,9,9,9,9,9,7,7,6,7,8,6,6,5,5,8,11,7,6,3,9,7,7,9,7,7,7,11,7,7,7,9,12,11,6,11,6,6,4,4,9,5,6,7,2,7,4,4,7,7,6,4,4,6,15,7,7,7,7,7,3,3,3,3,9,9,9,8,8,8,3,6,6,6,6,6,6,6,6,6,6,6,3,8,7,9,5,4,8,7,7,6,7,7,9,9,4,4,4,11,11,11,9,7,3,8,7,8,6,8,6,7,4,4,7,6,6,6,6,6,6,4,6,6,0,0,12,16,7,0,4,4,5,5,10,8,11,10,3,5,5,6,10,4,4,4,4,8,8,8,8,8,8,8,8,8,8,4,4,10,10,10,6,13,8,8,8,9,8,7,9,9,3,3,7,6,10,9,9,8,9,8,8,7,9,8,11,9,7,9,5,4,5,10,6,6,8,8,7,8,8,4,8,8,3,3,7,3,11,8,8,8,8,5,7,5,8,6,9,6,6,5,8,4,8,10,8,8,8,8,9,9,9,8,8,8,8,8,8,7,8,8,8,8,3,3,3,3,8,8,8,8,8,8,8,8,8,8,6,6,8,8,6,7,8,8,12,12,12,6,6,10,12,9,10,10,10,10,8,8,6,8,9,7,6,6,6,9,12,8,6,5,10,8,8,10,8,7,7,12,8,8,8,9,13,13,6,12,6,6,4,4,10,6,6,7,2,8,5,5,7,7,6,4,4,6,16,8,8,8,8,8,3,3,3,3,9,9,9,9,9,9,3,6,6,6,6,6,6,6,6,6,6,6,3,8,7,9,5,4,9,8,7,6,8,8,10,10,5,5,5,12,12,12,9,8,3,8,7,8,7,8,7,8,4,4,8,6,6,6,6,6,6,5,6,6,0,0,13,17,8,0,4,4,5,5,11,8,12,11,3,5,5,7,11,4,5,4,4,8,8,8,8,8,8,8,8,8,8,4,4,11,11,11,7,13,9,9,9,10,8,7,10,10,3,3,8,7,11,10,10,8,10,8,9,7,10,9,11,10,7,10,5,4,5,11,7,7,8,8,7,8,8,4,8,8,3,3,7,3,13,8,8,8,8,5,7,5,8,7,9,7,7,7,8,4,8,11,9,9,9,8,10,10,10,8,8,8,8,8,8,7,8,8,8,8,3,3,3,3,8,8,8,8,8,8,8,8,8,8,7,7,8,8,7,8,8,8,13,13,13,7,7,11,13,10,11,11,11,11,8,8,7,9,10,8,7,6,6,10,13,8,7,5,11,8,8,11,9,8,8,13,8,9,9,10,14,13,7,13,7,7,4,4,11,6,7,7,2,8,5,5,8,8,7,4,4,7,17,9,8,9,8,8,3,3,3,3,10,10,10,10,10,10,3,7,7,7,7,7,7,7,7,7,7,7,3,9,7,10,7,4,10,8,7,7,8,8,11,11,5,5,5,13,13,13,10,8,3,9,7,9,7,9,7,8,5,4,8,7,7,7,7,7,7,5,7,7,0,0,14,19,8,0,4,4,5,5,12,9,13,12,3,5,5,7,12,4,5,4,5,9,9,9,9,9,9,9,9,9,9,5,5,12,12,12,7,14,9,10,10,11,9,8,11,10,3,3,9,7,12,10,11,9,11,10,9,9,10,9,13,11,9,10,5,5,5,12,7,7,8,9,8,9,9,4,9,9,3,3,8,3,13,9,9,9,9,5,8,5,9,7,11,7,7,8,9,5,9,12,9,9,10,9,10,11,10,8,8,8,8,8,8,8,9,9,9,9,3,3,3,3,9,9,9,9,9,9,9,9,9,9,7,7,9,9,7,8,9,9,14,14,14,7,7,12,14,11,12,12,12,12,9,9,7,9,11,8,7,7,7,11,14,9,7,5,12,9,9,12,9,9,9,14,9,9,9,11,15,15,7,14,7,7,4,4,12,7,7,9,2,9,6,6,8,8,7,4,4,7,19,9,9,9,9,9,3,3,3,3,11,11,11,10,10,10,3,7,7,7,7,7,7,7,7,7,7,7,3,9,8,10,8,5,11,9,9,7,9,9,12,12,6,6,6,14,14,14,11,9,3,9,8,10,8,10,8,9,5,4,9,7,7,7,7,7,7,6,7,7,0,0,15,20,9,0,5,5,6,6,13,10,14,11,3,6,6,8,13,5,5,5,5,10,10,10,10,10,10,10,10,10,10,5,5,13,13,13,8,15,10,10,10,11,9,9,12,11,3,3,10,8,13,11,12,9,12,10,10,9,11,10,15,10,9,11,6,5,6,13,8,8,9,9,8,9,9,5,9,9,3,3,8,3,15,9,9,9,9,6,8,6,9,8,13,10,8,8,10,5,10,13,10,10,10,9,11,12,11,9,9,9,9,9,9,8,9,9,9,9,3,3,3,3,9,9,9,9,9,9,9,9,9,9,8,8,10,10,8,9,10,9,15,15,15,8,8,13,15,12,13,13,13,13,10,9,8,10,11,9,8,7,7,11,15,9,8,6,13,10,10,13,10,9,9,15,10,10,10,12,16,15,8,15,8,8,5,5,13,7,8,9,3,10,6,6,9,9,8,5,5,8,20,10,9,10,9,9,3,3,3,3,12,12,12,11,11,11,3,8,8,8,8,8,8,8,8,8,8,8,3,10,8,11,8,5,12,9,9,8,9,9,13,13,6,6,6,15,15,15,12,9,3,10,8,10,8,10,8,9,5,5,10,8,8,8,8,8,8,6,8,8,0,0,16,21,10,0,5,5,6,6,13,10,15,12,3,6,6,8,13,5,6,5,5,10,10,10,10,10,10,10,10,10,10,5,5,13,13,13,9,16,11,11,11,12,10,9,12,12,5,5,10,9,13,12,13,10,13,11,10,9,12,11,17,11,9,12,6,5,6,13,8,8,9,10,9,10,9,6,10,10,3,3,9,3,15,10,10,10,10,7,9,6,10,9,13,10,9,9,10,5,10,13,11,11,11,10,12,13,12,9,9,9,9,9,9,9,9,9,9,9,3,3,3,3,10,10,10,10,10,10,10,10,10,10,8,8,10,10,8,9,10,10,16,16,16,8,8,13,16,13,13,13,13,13,10,10,8,11,12,9,8,8,8,12,16,10,9,6,13,10,10,13,11,10,10,16,10,11,11,13,17,15,8,16,8,8,5,5,13,8,9,9,3,10,6,6,10,10,8,5,5,8,21,11,10,11,10,10,5,5,5,5,13,13,13,12,12,12,3,8,8,8,8,8,8,8,8,8,8,9,4,10,9,12,9,5,13,10,9,9,10,10,13,13,6,6,6,16,16,16,12,10,5,10,9,11,9,11,9,10,6,5,10,8,8,8,8,8,8,6,8,8,0,0,17,23,10,0,5,5,7,8,14,11,16,13,5,7,7,9,14,5,6,5,6,11,11,11,11,11,11,11,11,11,11,6,6,14,14,14,9,17,11,12,12,13,11,10,13,13,6,6,12,9,15,13,13,10,13,12,11,10,13,11,18,13,10,12,7,6,7,14,9,9,10,10,8,10,10,6,10,11,5,5,10,5,17,11,10,10,10,7,8,7,11,10,15,10,10,8,11,6,11,14,11,11,12,11,13,13,13,10,10,10,10,10,10,8,10,10,10,10,5,5,5,5,11,10,10,10,10,10,11,11,11,11,9,9,11,11,9,10,11,11,17,17,17,9,9,14,16,13,14,14,14,14,11,11,9,11,13,10,9,8,8,13,16,10,9,7,14,11,11,14,11,10,10,17,11,11,11,13,18,17,9,17,9,9,5,5,14,8,10,10,3,11,7,7,10,10,9,5,5,9,23,11,11,11,11,11,6,6,6,6,13,13,13,13,13,13,5,9,9,9,9,9,9,9,9,9,9,10,5,11,8,12,8,6,13,10,10,10,10,10,14,14,7,7,7,16,16,16,13,10,6,11,8,12,8,12,8,10,6,5,11,9,9,9,9,9,9,7,9,9,0,0,18,24,11,0,6,6,7,8,15,11,17,13,4,7,7,9,15,6,7,6,6,11,11,11,11,11,11,11,11,11,11,6,6,15,15,15,10,18,12,12,13,14,11,10,14,14,6,6,12,10,16,14,14,11,14,13,11,12,14,12,19,13,12,13,7,6,7,15,9,9,10,11,9,11,11,6,11,11,5,5,10,5,17,11,11,11,11,8,8,7,11,11,16,11,11,9,11,6,11,15,12,12,13,11,14,14,14,10,10,10,10,10,10,9,11,11,11,11,5,5,5,5,11,11,11,11,11,11,11,11,11,11,9,9,11,11,9,11,11,11,18,18,18,9,9,15,17,14,15,15,15,15,11,11,9,12,14,11,9,8,8,14,17,11,10,7,15,11,11,15,12,11,11,18,11,12,12,14,19,19,9,18,9,9,6,6,15,9,11,12,3,11,7,7,11,11,9,6,6,9,24,12,11,12,11,11,6,6,6,6,14,14,14,14,14,14,5,9,9,9,9,9,9,9,9,9,9,10,5,11,8,13,9,6,14,11,12,11,11,11,15,15,7,7,7,17,17,17,14,11,6,11,8,13,9,13,9,11,7,6,11,9,9,9,9,9,9,7,9,9,0,0,19,26,11,0,6,6,8,8,16,12,18,14,4,7,7,10,16,6,7,6,6,12,12,12,12,12,12,12,12,12,12,6,6,16,16,16,10,19,13,13,13,15,12,11,15,14,6,6,12,11,16,14,15,12,15,13,12,12,14,13,20,14,12,13,7,6,7,16,10,10,11,11,9,11,11,6,11,12,5,5,11,5,17,12,11,11,11,8,9,8,12,11,17,11,11,11,12,6,12,16,13,13,13,12,14,15,14,11,11,11,11,11,11,9,11,11,11,11,5,5,5,5,12,11,11,11,11,11,12,12,12,12,10,10,12,12,10,11,12,12,19,19,19,10,10,16,18,15,16,16,16,16,12,12,10,13,14,11,10,9,9,15,18,11,10,8,16,12,12,16,13,12,12,19,12,13,13,15,20,19,10,19,10,10,6,6,16,9,11,12,3,12,8,8,11,11,10,6,6,10,26,13,12,13,12,12,6,6,6,6,15,15,15,14,14,14,5,10,10,10,10,10,10,10,10,10,10,11,5,12,9,13,11,6,15,11,12,11,12,11,16,16,8,8,8,18,18,18,15,11,6,12,9,13,9,13,9,11,7,6,12,10,10,10,10,10,10,8,10,10,0,0,20,27,12,0,6,6,8,10,17,13,19,15,6,8,8,10,17,6,7,6,7,13,13,13,13,13,13,13,13,13,13,7,7,17,17,17,11,20,14,14,14,15,13,12,16,15,6,6,13,11,17,15,16,12,16,14,13,12,15,14,20,14,12,14,8,7,8,17,10,10,12,13,11,13,12,7,13,13,6,6,12,6,20,13,12,13,13,8,10,8,13,11,17,11,11,11,13,7,13,17,14,14,14,13,15,16,15,12,12,12,12,12,12,11,12,12,12,12,6,6,6,6,13,12,12,12,12,12,13,13,13,13,10,10,13,13,10,12,13,13,20,20,20,10,10,17,20,16,17,17,17,17,13,13,10,13,15,12,10,9,9,15,19,12,11,8,17,13,13,17,13,12,12,20,13,14,14,16,21,20,10,20,10,10,6,6,17,10,11,12,3,13,8,8,13,13,10,6,6,10,27,14,13,14,13,13,6,6,6,6,16,16,16,15,15,15,6,10,10,10,10,10,10,10,10,10,10,11,6,13,10,14,11,7,15,12,12,11,12,13,17,17,8,8,8,19,19,19,16,13,6,13,10,14,11,14,11,13,7,6,13,10,10,10,10,10,10,8,10,10,0,0,21,28,13,0,7,7,8,10,18,13,20,16,6,8,8,11,18,7,8,7,7,13,13,13,13,13,13,13,13,13,13,7,7,18,18,18,11,21,14,14,15,16,13,12,16,16,6,6,14,12,18,16,17,13,17,15,13,13,15,14,21,14,13,14,8,7,8,18,11,11,13,13,12,13,13,7,13,13,6,6,12,6,20,14,13,13,13,9,11,8,14,11,17,13,13,11,13,7,13,18,14,14,15,13,16,17,15,13,13,13,13,13,13,12,13,13,13,13,6,6,6,6,14,13,13,13,13,13,14,14,14,14,11,11,13,13,11,12,13,13,21,21,21,11,11,18,20,17,18,18,18,18,13,14,11,14,16,12,11,10,10,16,20,13,11,8,18,13,13,18,14,13,13,21,13,14,14,17,22,22,11,21,11,11,7,7,18,10,13,13,4,13,8,8,13,13,11,7,7,11,28,14,13,14,13,13,6,6,6,6,17,17,17,15,15,15,6,11,11,11,11,11,11,11,11,11,11,12,6,13,11,14,11,7,16,13,13,13,13,13,18,18,8,8,8,20,20,20,16,13,6,13,11,15,12,15,12,13,8,7,13,11,11,11,11,11,11,8,11,11,0,0,22,30,13,0,7,7,9,10,18,14,21,17,6,9,9,11,18,7,8,7,7,14,14,14,14,14,14,14,14,14,14,7,7,18,18,18,12,22,15,15,15,17,14,13,17,17,6,6,14,12,19,16,17,13,17,15,14,13,16,15,22,15,13,15,9,7,9,18,11,11,13,14,12,14,13,8,14,14,6,6,13,6,22,14,13,14,14,9,11,9,14,14,18,14,14,12,14,7,14,18,15,15,15,14,16,17,16,13,13,13,13,13,13,12,13,13,13,13,6,6,6,6,14,13,13,13,13,13,14,14,14,14,11,11,14,14,11,13,14,14,22,22,22,11,11,18,21,17,18,18,18,18,14,14,11,15,17,13,11,10,10,17,21,13,12,9,18,14,14,18,15,13,13,22,14,15,15,17,24,22,11,22,11,11,7,7,18,11,14,13,4,14,9,9,13,13,11,7,7,11,30,15,14,15,14,14,6,6,6,6,17,17,17,16,16,16,6,11,11,11,11,11,11,11,11,11,11,12,6,14,11,15,12,7,17,13,13,14,13,14,18,18,9,9,9,21,21,21,17,14,6,14,11,15,12,15,12,14,8,7,14,11,11,11,11,11,11,9,11,11,0,0,23,31,14,0,7,7,9,11,19,15,22,18,6,9,9,12,19,7,8,7,8,15,15,15,15,15,15,15,15,15,15,8,8,19,19,19,12,23,16,16,16,18,15,13,18,17,7,7,15,13,20,17,18,14,18,16,15,14,17,16,23,16,14,16,9,8,9,19,12,12,14,15,13,15,14,8,15,15,6,6,13,6,22,15,14,14,14,9,12,9,15,14,19,14,14,12,15,8,15,19,16,16,16,15,17,18,17,14,14,14,14,14,14,13,14,14,14,14,6,6,6,6,15,14,14,14,14,14,15,15,15,15,12,12,15,15,12,14,15,14,23,23,23,12,12,19,22,18,19,19,19,19,15,15,12,16,17,14,12,11,11,18,21,14,12,9,19,15,15,19,15,14,14,23,15,16,16,18,25,24,12,23,12,12,7,7,19,11,14,14,4,15,9,9,15,15,12,7,7,12,31,16,15,16,15,15,7,7,7,7,18,18,18,17,17,17,6,12,12,12,12,12,12,12,12,12,12,13,6,15,12,16,12,8,18,14,14,14,14,14,19,19,9,9,9,22,22,22,18,15,7,15,12,16,13,16,13,15,8,7,15,12,12,12,12,12,12,9,12,12,0,0,24,32,14,0,8,8,10,11,20,15,23,19,7,9,9,12,20,8,9,8,8,15,15,15,15,15,15,15,15,15,15,8,8,20,20,20,13,24,16,16,17,18,15,14,19,18,7,7,16,13,21,18,19,14,19,17,15,15,18,16,24,16,15,16,9,8,9,20,12,12,14,15,13,15,14,8,15,15,7,7,14,6,24,15,14,15,15,10,12,9,15,15,20,15,15,13,15,8,15,20,16,16,17,15,18,19,18,14,14,14,14,14,14,13,14,14,14,14,7,7,7,7,15,14,14,14,14,14,15,15,15,15,12,12,15,15,12,14,15,15,24,24,24,12,12,20,23,19,20,20,20,20,15,15,12,16,18,14,13,11,11,18,23,14,13,10,20,15,15,20,16,15,15,24,15,16,16,19,26,24,12,24,12,12,8,8,20,12,15,15,4,15,10,10,15,15,12,8,8,12,32,16,15,16,15,15,7,7,7,7,19,19,19,18,18,18,7,12,12,12,12,12,12,12,12,12,12,13,6,15,12,16,13,8,19,14,15,15,15,15,20,20,10,10,10,23,23,23,19,15,7,15,12,17,13,17,13,15,9,8,15,12,12,12,12,12,12,10,12,12,0,0,25,34,15,0,8,8,10,12,21,16,24,20,7,10,10,13,21,8,9,8,8,16,16,16,16,16,16,16,16,16,16,8,8,21,21,21,13,25,17,17,17,19,16,14,19,19,7,7,16,14,22,19,20,15,20,17,16,15,18,17,25,17,15,17,10,8,10,21,13,13,15,16,14,16,15,9,16,16,7,7,14,7,24,16,15,16,16,10,13,10,16,15,20,15,15,13,16,8,16,21,17,17,17,16,19,20,18,15,15,15,15,15,15,14,15,15,15,15,7,7,7,7,16,15,15,15,15,15,16,16,16,16,13,13,16,16,13,15,16,16,25,25,25,13,13,21,24,20,21,21,21,21,16,16,13,17,19,15,13,12,12,19,25,15,13,10,21,16,16,21,17,15,15,25,16,17,17,20,27,26,13,25,13,13,8,8,21,12,15,15,4,16,10,10,16,16,13,8,8,13,34,17,16,17,16,16,7,7,7,7,20,20,20,18,18,18,7,13,13,13,13,13,13,13,13,13,13,14,7,16,13,17,13,8,19,15,15,15,15,16,21,21,10,10,10,24,24,24,19,16,7,16,13,17,14,17,14,16,9,8,16,13,13,13,13,13,13,10,13,13,0,0,26,35,16,0,8,8,10,12,22,17,25,20,7,10,10,13,22,8,9,8,9,17,17,17,17,17,17,17,17,17,17,9,9,22,22,22,14,26,18,18,18,20,16,15,20,20,8,8,17,14,22,19,20,16,20,18,17,16,19,18,26,18,16,18,10,9,10,22,13,13,16,17,14,17,16,9,17,16,7,7,15,7,25,16,16,17,17,11,14,10,16,15,21,15,15,14,17,9,17,22,18,18,18,16,19,20,19,16,16,16,16,16,16,14,16,16,16,16,7,7,7,7,16,16,16,16,16,16,16,16,16,16,13,13,17,17,13,15,17,16,26,26,26,13,13,22,25,20,22,22,22,22,17,17,13,18,20,15,14,12,12,20,26,16,14,10,22,17,17,22,17,16,16,26,17,18,18,20,28,27,13,26,13,13,8,8,22,13,15,16,4,17,10,10,16,16,13,8,8,13,35,18,16,18,16,16,8,8,8,8,20,20,20,19,19,19,7,13,13,13,13,13,13,13,13,13,13,15,7,17,14,18,14,9,20,16,16,15,16,17,22,22,10,10,10,25,25,25,20,17,8,17,14,18,14,18,14,17,9,8,17,13,13,13,13,13,13,10,13,13,0,0,27,36,16,0,9,9,11,12,23,17,26,21,7,11,11,14,23,9,10,9,9,17,17,17,17,17,17,17,17,17,17,9,9,23,23,23,14,27,18,19,19,21,17,16,21,20,8,8,18,15,23,20,21,16,21,19,17,17,20,18,27,19,17,19,11,9,11,23,14,14,17,17,15,17,17,10,17,17,8,8,16,8,26,17,17,17,17,11,14,11,17,16,22,16,16,14,17,9,17,23,18,18,19,17,20,21,20,17,17,17,17,17,17,15,17,17,17,17,8,8,8,8,17,17,17,17,17,17,17,17,17,17,14,14,17,17,14,16,17,17,27,27,27,14,14,23,26,21,22,23,23,23,17,17,14,18,20,16,14,13,13,21,27,17,14,11,23,17,17,23,18,17,17,27,17,18,18,21,29,28,14,27,14,14,9,9,23,13,16,17,5,17,11,11,17,17,14,9,9,14,36,18,17,18,17,17,8,8,8,8,21,21,21,20,20,20,8,14,14,14,14,14,14,14,14,14,14,15,8,17,14,19,14,9,21,17,17,16,16,17,23,23,11,11,11,26,26,26,21,17,8,17,14,19,15,19,15,17,10,9,17,14,14,14,14,14,14,11,14,14,0,0,28,38,17,0,9,9,11,13,23,18,27,22,8,11,11,14,23,9,10,9,9,18,18,18,18,18,18,18,18,18,18,9,9,23,23,23,15,28,19,19,20,22,18,16,22,21,8,8,18,16,24,21,22,17,22,19,18,17,21,19,28,19,17,19,11,9,11,23,14,14,17,18,15,18,17,10,18,18,8,8,16,8,27,18,17,18,18,12,15,11,18,17,23,17,17,15,18,9,18,23,19,19,20,18,21,22,21,17,17,17,17,17,17,15,17,17,17,17,8,8,8,8,18,17,17,17,17,17,18,18,18,18,14,14,18,18,14,17,18,18,28,28,28,14,14,23,27,22,23,23,23,23,18,18,14,19,21,16,15,13,13,21,28,17,15,11,23,18,18,23,19,17,17,28,18,19,19,22,30,29,14,28,15,15,9,9,23,14,17,17,5,18,11,11,18,18,14,9,9,15,38,19,18,19,18,18,8,8,8,8,22,22,22,21,21,21,8,14,14,14,14,14,14,14,14,14,14,16,8,18,15,19,15,9,22,17,17,17,17,18,23,23,11,11,11,27,27,27,22,18,8,18,15,20,15,20,15,18,10,9,18,14,14,14,14,14,14,11,14,14,0,0,0,0,0,2,0,8,0,2,255,255,0,3,0,1,0,0,0,2,0,0,12,80,10,236,95,15,60,245,0,31,8,0,0,0,0,0,186,185,240,184,0,0,0,0,186,194,103,145,254,137,254,29,10,76,7,109,0,0,0,8,0,1,0,0,0,0,0,0,78,52,108,111,118,101,52,102,111,110,116,52,70,111,110,116,69,0,78,52,108,111,118,101,52,102,111,110,116,49,53,68,101,102,97,117,108,116,70,111,110,116,68,97,116,97,69,0,78,52,117,116,102,56,57,101,120,99,101,112,116,105,111,110,69,0,78,52,117,116,102,56,49,53,110,111,116,95,101,110,111,117,103,104,95,114,111,111,109,69,0,78,111,116,32,101,110,111,117,103,104,32,115,112,97,99,101,0,78,52,117,116,102,56,49,50,105,110,118,97,108,105,100,95,117,116,102,56,69,0,73,110,118,97,108,105,100,32,85,84,70,45,56,0,78,52,117,116,102,56,49,56,105,110,118,97,108,105,100,95,99,111,100,101,95,112,111,105,110,116,69,0,73,110,118,97,108,105,100,32,99,111,100,101,32,112,111,105,110,116,0,78,52,108,111,118,101,52,102,111,110,116,57,71,108,121,112,104,68,97,116,97,69,0,108,117,109,105,110,97,110,99,101,97,108,112,104,97,0,114,103,98,97,0,78,52,108,111,118,101,52,102,111,110,116,49,53,73,109,97,103,101,82,97,115,116,101,114,105,122,101,114,69,0,78,52,108,111,118,101,52,102,111,110,116,49,48,82,97,115,116,101,114,105,122,101,114,69,0,108,105,103,104,116,0,109,111,110,111,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,54,111,112,101,110,103,108,54,67,97,110,118,97,115,69,0,67,97,110,110,111,116,32,100,114,97,119,32,97,32,67,97,110,118,97,115,32,116,111,32,105,116,115,101,108,102,46,0,104,100,114,0,114,103,98,97,52,0,114,103,98,53,97,49,0,114,103,98,53,54,53,0,114,56,0,114,103,56,0,114,103,98,49,48,97,50,0,114,103,49,49,98,49,48,102,0,114,49,54,102,0,114,103,49,54,102,0,114,103,98,97,49,54,102,0,114,51,50,102,0,114,103,51,50,102,0,114,103,98,97,51,50,102,0,115,114,103,98,0,84,104,105,115,32,115,121,115,116,101,109,32,99,97,110,39,116,32,115,105,109,117,108,116,97,110,101,111,117,115,108,121,32,114,101,110,100,101,114,32,116,111,32,37,100,32,99,97,110,118,97,115,101,115,46,0,77,117,108,116,105,45,99,97,110,118,97,115,32,114,101,110,100,101,114,105,110,103,32,105,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,119,105,116,104,32,77,83,65,65,46,0,65,108,108,32,99,97,110,118,97,115,101,115,32,109,117,115,116,32,104,97,118,101,32,116,104,101,32,115,97,109,101,32,100,105,109,101,110,115,105,111,110,115,46,0,84,104,105,115,32,115,121,115,116,101,109,32,100,111,101,115,110,39,116,32,115,117,112,112,111,114,116,32,109,117,108,116,105,45,99,97,110,118,97,115,32,114,101,110,100,101,114,105,110,103,32,119,105,116,104,32,100,105,102,102,101,114,101,110,116,32,99,97,110,118,97,115,32,102,111,114,109,97,116,115,46,0,73,110,118,97,108,105,100,32,73,109,97,103,101,68,97,116,97,32,114,101,99,116,97,110,103,108,101,32,100,105,109,101,110,115,105,111,110,115,46,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,54,111,112,101,110,103,108,52,70,111,110,116,69,0,67,111,117,108,100,32,110,111,116,32,99,114,101,97,116,101,32,102,111,110,116,32,116,101,120,116,117,114,101,33,0,99,101,110,116,101,114,0,106,117,115,116,105,102,121,0,73,110,118,97,108,105,100,32,117,116,102,45,56,32,105,116,101,114,97,116,111,114,32,112,111,115,105,116,105,111,110,0,67,111,109,112,97,114,105,110,103,32,117,116,102,45,56,32,105,116,101,114,97,116,111,114,115,32,100,101,102,105,110,101,100,32,119,105,116,104,32,100,105,102,102,101,114,101,110,116,32,114,97,110,103,101,115,0,85,84,70,45,56,32,100,101,99,111,100,105,110,103,32,101,114,114,111,114,58,32,37,115,0,78,83,116,51,95,95,49,49,57,98,97,115,105,99,95,105,115,116,114,105,110,103,115,116,114,101,97,109,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,78,83,95,57,97,108,108,111,99,97,116,111,114,73,99,69,69,69,69,0,70,111,110,116,32,102,97,108,108,98,97,99,107,115,32,109,117,115,116,32,98,101,32,111,102,32,116,104,101,32,115,97,109,101,32,102,111,110,116,32,116,121,112,101,46,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,54,111,112,101,110,103,108,56,71,76,66,117,102,102,101,114,69,0,73,110,118,97,108,105,100,32,110,117,109,98,101,114,32,111,102,32,113,117,97,100,115,46,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,54,111,112,101,110,103,108,56,71,114,97,112,104,105,99,115,69,0,108,111,118,101,46,103,114,97,112,104,105,99,115,46,111,112,101,110,103,108,0,79,112,101,110,71,76,58,32,37,115,32,91,115,111,117,114,99,101,61,37,115,44,32,116,121,112,101,61,37,115,44,32,115,101,118,101,114,105,116,121,61,37,115,44,32,105,100,61,37,100,93,10,0,79,112,101,110,71,76,32,100,101,98,117,103,32,111,117,116,112,117,116,32,101,110,97,98,108,101,100,32,40,76,79,86,69,95,71,82,65,80,72,73,67,83,95,68,69,66,85,71,61,49,41,0,67,111,117,108,100,32,110,111,116,32,114,101,108,111,97,100,32,97,108,108,32,118,111,108,97,116,105,108,101,32,111,98,106,101,99,116,115,46,0,84,104,101,32,39,108,105,103,104,116,101,110,39,32,97,110,100,32,39,100,97,114,107,101,110,39,32,98,108,101,110,100,32,109,111,100,101,115,32,97,114,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,111,110,32,116,104,105,115,32,115,121,115,116,101,109,46,0,84,104,101,32,39,37,115,39,32,98,108,101,110,100,32,109,111,100,101,32,109,117,115,116,32,98,101,32,117,115,101,100,32,119,105,116,104,32,112,114,101,109,117,108,116,105,112,108,105,101,100,32,97,108,112,104,97,46,0,70,111,110,116,32,109,111,100,117,108,101,32,104,97,115,32,110,111,116,32,98,101,101,110,32,108,111,97,100,101,100,46,0,78,117,109,98,101,114,32,111,102,32,99,108,101,97,114,32,99,111,108,111,114,115,32,109,117,115,116,32,109,97,116,99,104,32,116,104,101,32,110,117,109,98,101,114,32,111,102,32,97,99,116,105,118,101,32,99,97,110,118,97,115,101,115,32,40,37,108,100,41,0,67,97,110,118,97,115,101,115,32,97,114,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,98,121,32,121,111,117,114,32,79,112,101,110,71,76,32,100,114,105,118,101,114,115,33,0,114,103,98,97,56,0,84,104,101,32,37,115,32,99,97,110,118,97,115,32,102,111,114,109,97,116,32,105,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,98,121,32,121,111,117,114,32,79,112,101,110,71,76,32,100,114,105,118,101,114,115,46,0,67,97,110,110,111,116,32,99,114,101,97,116,101,32,99,97,110,118,97,115,58,32,119,105,100,116,104,32,111,102,32,37,100,32,112,105,120,101,108,115,32,105,115,32,116,111,111,32,108,97,114,103,101,32,102,111,114,32,116,104,105,115,32,115,121,115,116,101,109,46,0,67,97,110,110,111,116,32,99,114,101,97,116,101,32,99,97,110,118,97,115,58,32,104,101,105,103,104,116,32,111,102,32,37,100,32,112,105,120,101,108,115,32,105,115,32,116,111,111,32,108,97,114,103,101,32,102,111,114,32,116,104,105,115,32,115,121,115,116,101,109,46,0,67,97,110,110,111,116,32,99,114,101,97,116,101,32,99,97,110,118,97,115,58,32,0,78,111,116,32,115,117,112,112,111,114,116,101,100,32,98,121,32,121,111,117,114,32,79,112,101,110,71,76,32,100,114,105,118,101,114,115,46,0,84,101,120,116,117,114,101,32,102,111,114,109,97,116,32,99,97,110,110,111,116,32,98,101,32,114,101,110,100,101,114,101,100,32,116,111,32,111,110,32,116,104,105,115,32,115,121,115,116,101,109,46,0,69,114,114,111,114,32,105,110,32,103,114,97,112,104,105,99,115,32,100,114,105,118,101,114,46,0,77,97,121,32,110,111,116,32,98,101,32,115,117,112,112,111,114,116,101,100,32,98,121,32,121,111,117,114,32,79,112,101,110,71,76,32,100,114,105,118,101,114,115,46,0,67,97,110,110,111,116,32,99,114,101,97,116,101,32,99,97,110,118,97,115,58,32,65,108,105,101,110,115,32,100,105,100,32,105,116,32,40,79,112,101,110,71,76,32,101,114,114,111,114,32,99,111,100,101,58,32,0,79,112,101,110,71,76,32,69,83,0,79,112,101,110,71,76,0,67,97,110,110,111,116,32,114,101,116,114,105,101,118,101,32,114,101,110,100,101,114,101,114,32,118,101,114,115,105,111,110,32,105,110,102,111,114,109,97,116,105,111,110,46,0,67,97,110,110,111,116,32,114,101,116,114,105,101,118,101,32,114,101,110,100,101,114,101,114,32,118,101,110,100,111,114,32,105,110,102,111,114,109,97,116,105,111,110,46,0,67,97,110,110,111,116,32,114,101,116,114,105,101,118,101,32,114,101,110,100,101,114,101,114,32,100,101,118,105,99,101,32,105,110,102,111,114,109,97,116,105,111,110,46,0,77,97,120,105,109,117,109,32,115,116,97,99,107,32,100,101,112,116,104,32,114,101,97,99,104,101,100,32,40,109,111,114,101,32,112,117,115,104,101,115,32,116,104,97,110,32,112,111,112,115,63,41,0,77,105,110,105,109,117,109,32,115,116,97,99,107,32,100,101,112,116,104,32,114,101,97,99,104,101,100,32,40,109,111,114,101,32,112,111,112,115,32,116,104,97,110,32,112,117,115,104,101,115,63,41,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,54,111,112,101,110,103,108,53,73,109,97,103,101,69,0,78,111,110,45,109,105,112,109,97,112,112,101,100,32,105,109,97,103,101,32,99,97,110,110,111,116,32,104,97,118,101,32,109,105,112,109,97,112,32,102,105,108,116,101,114,105,110,103,46,0,115,82,71,66,32,0,67,97,110,110,111,116,32,99,114,101,97,116,101,32,105,109,97,103,101,58,32,37,115,37,115,32,99,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,115,32,97,114,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,111,110,32,116,104,105,115,32,115,121,115,116,101,109,46,0,99,97,110,110,111,116,32,99,114,101,97,116,101,32,105,109,97,103,101,58,32,102,111,114,109,97,116,32,105,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,111,110,32,116,104,105,115,32,115,121,115,116,101,109,46,0,115,82,71,66,32,105,109,97,103,101,115,32,97,114,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,111,110,32,116,104,105,115,32,115,121,115,116,101,109,46,0,255,255,255,255,255,192,192,255,255,192,192,255,255,255,255,255,67,97,110,110,111,116,32,99,114,101,97,116,101,32,105,109,97,103,101,32,40,79,112,101,110,71,76,32,101,114,114,111,114,58,32,37,115,41,0,109,105,112,109,97,112,115,0,108,105,110,101,97,114,0,73,109,97,103,101,32,100,111,101,115,32,110,111,116,32,104,97,118,101,32,97,108,108,32,114,101,113,117,105,114,101,100,32,109,105,112,109,97,112,32,108,101,118,101,108,115,32,40,101,120,112,101,99,116,101,100,32,37,100,44,32,103,111,116,32,37,100,41,0,87,105,100,116,104,32,111,102,32,105,109,97,103,101,32,109,105,112,109,97,112,32,108,101,118,101,108,32,37,100,32,105,115,32,105,110,99,111,114,114,101,99,116,32,40,101,120,112,101,99,116,101,100,32,37,100,44,32,103,111,116,32,37,100,41,0,72,101,105,103,104,116,32,111,102,32,105,109,97,103,101,32,109,105,112,109,97,112,32,108,101,118,101,108,32,37,100,32,105,115,32,105,110,99,111,114,114,101,99,116,32,40,101,120,112,101,99,116,101,100,32,37,100,44,32,103,111,116,32,37,100,41,0,73,109,97,103,101,32,99,97,110,110,111,116,32,104,97,118,101,32,109,105,112,109,97,112,115,58,32,99,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,32,100,97,116,97,32,100,111,101,115,32,110,111,116,32,104,97,118,101,32,97,108,108,32,114,101,113,117,105,114,101,100,32,109,105,112,109,97,112,32,108,101,118,101,108,115,46,0,65,108,108,32,105,109,97,103,101,32,109,105,112,109,97,112,32,108,101,118,101,108,115,32,109,117,115,116,32,104,97,118,101,32,116,104,101,32,115,97,109,101,32,102,111,114,109,97,116,46,0,73,110,118,97,108,105,100,32,114,101,99,116,97,110,103,108,101,32,100,105,109,101,110,115,105,111,110,115,46,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,54,111,112,101,110,103,108,52,77,101,115,104,69,0,77,101,115,104,32,109,117,115,116,32,104,97,118,101,32,97,110,32,101,110,97,98,108,101,100,32,86,101,114,116,101,120,80,111,115,105,116,105,111,110,32,97,116,116,114,105,98,117,116,101,32,116,111,32,98,101,32,100,114,97,119,110,46,0,115,116,114,101,97,109,0,100,121,110,97,109,105,99,0,115,116,97,116,105,99,0,102,97,110,0,115,116,114,105,112,0,116,114,105,97,110,103,108,101,115,0,68,117,112,108,105,99,97,116,101,32,118,101,114,116,101,120,32,97,116,116,114,105,98,117,116,101,32,110,97,109,101,58,32,37,115,0,98,121,116,101,32,118,101,114,116,101,120,32,97,116,116,114,105,98,117,116,101,115,32,109,117,115,116,32,104,97,118,101,32,52,32,99,111,109,112,111,110,101,110,116,115,46,0,86,101,114,116,101,120,32,97,116,116,114,105,98,117,116,101,115,32,109,117,115,116,32,104,97,118,101,32,98,101,116,119,101,101,110,32,49,32,97,110,100,32,52,32,99,111,109,112,111,110,101,110,116,115,46,0,68,97,116,97,32,115,105,122,101,32,105,115,32,116,111,111,32,115,109,97,108,108,32,102,111,114,32,115,112,101,99,105,102,105,101,100,32,118,101,114,116,101,120,32,97,116,116,114,105,98,117,116,101,32,102,111,114,109,97,116,115,46,0,73,110,118,97,108,105,100,32,110,117,109,98,101,114,32,111,102,32,118,101,114,116,105,99,101,115,32,40,37,100,41,46,0,73,110,118,97,108,105,100,32,118,101,114,116,101,120,32,105,110,100,101,120,58,32,37,108,100,0,73,110,118,97,108,105,100,32,118,101,114,116,101,120,32,97,116,116,114,105,98,117,116,101,32,105,110,100,101,120,58,32,37,100,0,77,101,115,104,32,100,111,101,115,32,110,111,116,32,104,97,118,101,32,97,110,32,97,116,116,97,99,104,101,100,32,118,101,114,116,101,120,32,97,116,116,114,105,98,117,116,101,32,110,97,109,101,100,32,39,37,115,39,0,67,97,110,110,111,116,32,97,116,116,97,99,104,32,97,32,77,101,115,104,32,119,104,105,99,104,32,104,97,115,32,97,116,116,97,99,104,101,100,32,77,101,115,104,101,115,32,111,102,32,105,116,115,32,111,119,110,46,0,73,110,118,97,108,105,100,32,118,101,114,116,101,120,32,109,97,112,32,118,97,108,117,101,58,32,37,100,0,73,110,118,97,108,105,100,32,100,114,97,119,32,114,97,110,103,101,46,0,65,84,73,32,84,101,99,104,110,111,108,111,103,105,101,115,0,78,86,73,68,73,65,0,73,110,116,101,108,0,77,101,115,97,0,65,112,112,108,101,32,67,111,109,112,117,116,101,114,0,65,112,112,108,101,32,73,110,99,46,0,77,105,99,114,111,115,111,102,116,0,73,109,97,103,105,110,97,116,105,111,110,0,81,117,97,108,99,111,109,109,0,66,114,111,97,100,99,111,109,0,86,105,118,97,110,116,101,0,73,110,118,97,108,105,100,32,116,101,120,116,117,114,101,32,117,110,105,116,32,105,110,100,101,120,32,40,37,100,41,46,0,73,110,118,97,108,105,100,32,116,101,120,116,117,114,101,32,117,110,105,116,32,105,110,100,101,120,46,0,48,120,37,120,0,79,112,101,110,71,76,32,99,111,110,116,101,120,116,32,104,97,115,32,98,101,101,110,32,108,111,115,116,0,105,110,118,97,108,105,100,32,102,114,97,109,101,98,117,102,102,101,114,32,111,112,101,114,97,116,105,111,110,0,111,117,116,32,111,102,32,109,101,109,111,114,121,0,105,110,118,97,108,105,100,32,111,112,101,114,97,116,105,111,110,0,105,110,118,97,108,105,100,32,101,110,117,109,0,110,111,32,101,114,114,111,114,0,108,111,119,0,109,101,100,105,117,109,0,104,105,103,104,0,111,116,104,101,114,0,76,79,86,69,0,101,120,116,101,114,110,97,108,0,115,104,97,100,101,114,0,119,105,110,100,111,119,0,65,80,73,0,112,111,114,116,97,98,105,108,105,116,121,0,112,101,114,102,111,114,109,97,110,99,101,0,117,110,100,101,102,105,110,101,100,32,98,101,104,97,118,105,111,114,0,100,101,112,114,101,99,97,116,101,100,32,98,101,104,97,118,105,111,114,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,54,111,112,101,110,103,108,49,52,80,97,114,116,105,99,108,101,83,121,115,116,101,109,69,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,54,111,112,101,110,103,108,56,80,111,108,121,108,105,110,101,69,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,54,111,112,101,110,103,108,49,54,78,111,110,101,74,111,105,110,80,111,108,121,108,105,110,101,69,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,54,111,112,101,110,103,108,49,55,77,105,116,101,114,74,111,105,110,80,111,108,121,108,105,110,101,69,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,54,111,112,101,110,103,108,49,55,66,101,118,101,108,74,111,105,110,80,111,108,121,108,105,110,101,69,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,54,111,112,101,110,103,108,54,83,104,97,100,101,114,69,0,67,97,110,110,111,116,32,99,114,101,97,116,101,32,115,104,97,100,101,114,32,111,98,106,101,99,116,58,32,117,110,107,110,111,119,110,32,115,104,97,100,101,114,32,116,121,112,101,46,0,67,97,110,110,111,116,32,99,114,101,97,116,101,32,37,115,32,115,104,97,100,101,114,32,111,98,106,101,99,116,58,32,37,115,32,115,104,97,100,101,114,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,46,0,67,97,110,110,111,116,32,99,114,101,97,116,101,32,37,115,32,115,104,97,100,101,114,32,111,98,106,101,99,116,46,0,67,97,110,110,111,116,32,99,111,109,112,105,108,101,32,37,115,32,115,104,97,100,101,114,32,99,111,100,101,58,10,37,115,0,67,97,110,110,111,116,32,99,114,101,97,116,101,32,115,104,97,100,101,114,32,112,114,111,103,114,97,109,32,111,98,106,101,99,116,46,0,67,97,110,110,111,116,32,108,105,110,107,32,115,104,97,100,101,114,32,112,114,111,103,114,97,109,32,111,98,106,101,99,116,58,10,37,115,0,91,48,93,0,118,101,114,116,101,120,0,112,105,120,101,108,0,102,108,111,97,116,0,105,110,116,0,98,111,111,108,0,105,109,97,103,101,0,86,101,114,116,101,120,80,111,115,105,116,105,111,110,0,86,101,114,116,101,120,84,101,120,67,111,111,114,100,0,86,101,114,116,101,120,67,111,108,111,114,0,67,111,110,115,116,97,110,116,67,111,108,111,114,0,84,114,97,110,115,102,111,114,109,77,97,116,114,105,120,0,80,114,111,106,101,99,116,105,111,110,77,97,116,114,105,120,0,84,114,97,110,115,102,111,114,109,80,114,111,106,101,99,116,105,111,110,77,97,116,114,105,120,0,78,111,114,109,97,108,77,97,116,114,105,120,0,108,111,118,101,95,80,111,105,110,116,83,105,122,101,0,108,111,118,101,95,83,99,114,101,101,110,83,105,122,101,0,108,111,118,101,95,86,105,100,101,111,89,67,104,97,110,110,101,108,0,108,111,118,101,95,86,105,100,101,111,67,98,67,104,97,110,110,101,108,0,108,111,118,101,95,86,105,100,101,111,67,114,67,104,97,110,110,101,108,0,67,97,110,110,111,116,32,99,114,101,97,116,101,32,115,104,97,100,101,114,58,32,110,111,32,115,111,117,114,99,101,32,99,111,100,101,33,0,32,115,104,97,100,101,114,58,10,0,86,97,114,105,97,98,108,101,32,39,37,115,39,32,100,111,101,115,32,110,111,116,32,101,120,105,115,116,46,10,65,32,99,111,109,109,111,110,32,101,114,114,111,114,32,105,115,32,116,111,32,100,101,102,105,110,101,32,98,117,116,32,110,111,116,32,117,115,101,32,116,104,101,32,118,97,114,105,97,98,108,101,46,0,78,111,32,97,99,116,105,118,101,32,115,104,97,100,101,114,32,112,114,111,103,114,97,109,46,0,86,97,108,117,101,32,115,105,122,101,32,111,102,32,37,100,32,100,111,101,115,32,110,111,116,32,109,97,116,99,104,32,118,97,114,105,97,98,108,101,32,115,105,122,101,32,111,102,32,37,100,46,0,73,110,118,97,108,105,100,32,110,117,109,98,101,114,32,111,102,32,118,97,108,117,101,115,32,40,101,120,112,101,99,116,101,100,32,37,100,44,32,103,111,116,32,37,100,41,46,0,67,97,110,110,111,116,32,115,101,110,100,32,97,32,118,97,108,117,101,32,111,102,32,116,104,105,115,32,116,121,112,101,32,116,111,32,97,110,32,73,109,97,103,101,32,118,97,114,105,97,98,108,101,46,0,67,97,110,110,111,116,32,99,111,110,118,101,114,116,32,98,101,116,119,101,101,110,32,102,108,111,97,116,32,97,110,100,32,105,110,116,46,0,73,110,118,97,108,105,100,32,109,97,116,114,105,120,32,115,105,122,101,58,32,37,100,120,37,100,32,40,99,97,110,32,111,110,108,121,32,115,101,116,32,50,120,50,44,32,51,120,51,32,111,114,32,52,120,52,32,109,97,116,114,105,99,101,115,46,41,0,78,111,32,109,111,114,101,32,116,101,120,116,117,114,101,32,117,110,105,116,115,32,97,118,97,105,108,97,98,108,101,32,102,111,114,32,115,104,97,100,101,114,46,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,54,111,112,101,110,103,108,49,49,83,112,114,105,116,101,66,97,116,99,104,69,0,77,101,115,104,32,119,105,116,104,32,97,116,116,114,105,98,117,116,101,32,39,37,115,39,32,97,116,116,97,99,104,101,100,32,116,111,32,116,104,105,115,32,83,112,114,105,116,101,66,97,116,99,104,32,104,97,115,32,116,111,111,32,102,101,119,32,118,101,114,116,105,99,101,115,0,73,110,118,97,108,105,100,32,83,112,114,105,116,101,66,97,116,99,104,32,115,105,122,101,46,0,77,101,115,104,32,104,97,115,32,116,111,111,32,102,101,119,32,118,101,114,116,105,99,101,115,32,116,111,32,98,101,32,97,116,116,97,99,104,101,100,32,116,111,32,116,104,105,115,32,83,112,114,105,116,101,66,97,116,99,104,32,40,97,116,32,108,101,97,115,116,32,37,100,32,118,101,114,116,105,99,101,115,32,97,114,101,32,114,101,113,117,105,114,101,100,41,0,84,104,101,32,115,112,101,99,105,102,105,101,100,32,109,101,115,104,32,100,111,101,115,32,110,111,116,32,104,97,118,101,32,97,32,118,101,114,116,101,120,32,97,116,116,114,105,98,117,116,101,32,110,97,109,101,100,32,39,37,115,39,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,54,111,112,101,110,103,108,52,84,101,120,116,69,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,54,111,112,101,110,103,108,53,86,105,100,101,111,69,0,73,110,118,97,108,105,100,32,116,101,120,116,117,114,101,32,102,105,108,116,101,114,46,0,85,110,107,110,111,119,110,32,67,97,110,118,97,115,32,102,111,114,109,97,116,46,0,67,97,110,118,97,115,0,114,101,110,100,101,114,84,111,0,110,101,119,73,109,97,103,101,68,97,116,97,0,103,101,116,70,111,114,109,97,116,0,103,101,116,77,83,65,65,0,73,110,118,97,108,105,100,32,102,105,108,116,101,114,32,109,111,100,101,58,32,37,115,0,70,111,110,116,0,103,101,116,72,101,105,103,104,116,0,103,101,116,87,105,100,116,104,0,103,101,116,87,114,97,112,0,115,101,116,76,105,110,101,72,101,105,103,104,116,0,103,101,116,76,105,110,101,72,101,105,103,104,116,0,115,101,116,70,105,108,116,101,114,0,103,101,116,70,105,108,116,101,114,0,103,101,116,65,115,99,101,110,116,0,103,101,116,68,101,115,99,101,110,116,0,103,101,116,66,97,115,101,108,105,110,101,0,104,97,115,71,108,121,112,104,115,0,115,101,116,70,97,108,108,98,97,99,107,115,0,78,52,108,111,118,101,56,103,114,97,112,104,105,99,115,56,86,111,108,97,116,105,108,101,69,0,78,52,108,111,118,101,53,105,109,97,103,101,54,109,97,103,112,105,101,53,73,109,97,103,101,69,0,78,52,108,111,118,101,53,105,109,97,103,101,53,73,109,97,103,101,69,0,108,111,118,101,46,105,109,97,103,101,46,109,97,103,112,105,101,0,78,52,108,111,118,101,53,105,109,97,103,101,54,109,97,103,112,105,101,57,73,109,97,103,101,68,97,116,97,69,0,78,111,32,115,117,105,116,97,98,108,101,32,105,109,97,103,101,32,101,110,99,111,100,101,114,32,102,111,114,32,37,115,32,102,111,114,109,97,116,46,0,67,111,117,108,100,32,110,111,116,32,100,101,99,111,100,101,32,102,105,108,101,32,39,37,115,39,32,116,111,32,73,109,97,103,101,68,97,116,97,58,32,117,110,115,117,112,112,111,114,116,101,100,32,102,105,108,101,32,102,111,114,109,97,116,0,67,111,117,108,100,32,110,111,116,32,99,111,110,118,101,114,116,32,105,109,97,103,101,33,0,78,52,108,111,118,101,53,105,109,97,103,101,54,109,97,103,112,105,101,49,48,75,84,88,72,97,110,100,108,101,114,69,0,78,52,108,111,118,101,53,105,109,97,103,101,54,109,97,103,112,105,101,50,51,67,111,109,112,114,101,115,115,101,100,70,111,114,109,97,116,72,97,110,100,108,101,114,69,0,171,75,84,88,32,49,49,187,13,10,26,10,67,111,117,108,100,32,110,111,116,32,100,101,99,111,100,101,32,99,111,109,112,114,101,115,115,101,100,32,100,97,116,97,32,40,110,111,116,32,97,32,75,84,88,32,102,105,108,101,63,41,0,85,110,115,117,112,112,111,114,116,101,100,32,105,109,97,103,101,32,102,111,114,109,97,116,32,105,110,32,75,84,88,32,102,105,108,101,46,0,84,101,120,116,117,114,101,32,97,114,114,97,121,115,32,105,110,32,75,84,88,32,102,105,108,101,115,32,97,114,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,46,0,51,68,32,116,101,120,116,117,114,101,115,32,105,110,32,75,84,88,32,102,105,108,101,115,32,97,114,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,46,0,67,117,98,101,109,97,112,32,116,101,120,116,117,114,101,115,32,105,110,32,75,84,88,32,102,105,108,101,115,32,97,114,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,46,0,67,111,117,108,100,32,110,111,116,32,112,97,114,115,101,32,75,84,88,32,102,105,108,101,58,32,117,110,101,120,112,101,99,116,101,100,32,69,79,70,46,0,78,52,108,111,118,101,53,105,109,97,103,101,54,109,97,103,112,105,101,49,48,80,75,77,72,97,110,100,108,101,114,69,0,80,75,77,32,67,111,117,108,100,32,110,111,116,32,100,101,99,111,100,101,32,99,111,109,112,114,101,115,115,101,100,32,100,97,116,97,32,40,110,111,116,32,97,32,80,75,77,32,102,105,108,101,63,41,0,67,111,117,108,100,32,110,111,116,32,112,97,114,115,101,32,80,75,77,32,102,105,108,101,58,32,117,110,115,117,112,112,111,114,116,101,100,32,116,101,120,116,117,114,101,32,102,111,114,109,97,116,46,0,78,52,108,111,118,101,53,105,109,97,103,101,54,109,97,103,112,105,101,49,48,80,78,71,72,97,110,100,108,101,114,69,0,67,111,117,108,100,32,110,111,116,32,100,101,99,111,100,101,32,80,78,71,32,105,109,97,103,101,32,40,37,115,41,0,80,78,71,32,101,110,99,111,100,101,114,32,99,97,110,110,111,116,32,101,110,99,111,100,101,32,116,111,32,110,111,110,45,80,78,71,32,102,111,114,109,97,116,46,0,67,111,117,108,100,32,110,111,116,32,101,110,99,111,100,101,32,80,78,71,32,105,109,97,103,101,32,40,37,115,41,0,78,52,108,111,118,101,53,105,109,97,103,101,54,109,97,103,112,105,101,49,48,80,86,82,72,97,110,100,108,101,114,69,0,67,111,117,108,100,32,110,111,116,32,100,101,99,111,100,101,32,99,111,109,112,114,101,115,115,101,100,32,100,97,116,97,32,40,110,111,116,32,97,32,80,86,82,32,102,105,108,101,63,41,0,73,109,97,103,101,32,100,101,112,116,104,115,32,103,114,101,97,116,101,114,32,116,104,97,110,32,49,32,105,110,32,80,86,82,32,102,105,108,101,115,32,97,114,101,32,117,110,115,117,112,112,111,114,116,101,100,46,0,67,111,117,108,100,32,110,111,116,32,112,97,114,115,101,32,80,86,82,32,102,105,108,101,58,32,117,110,115,117,112,112,111,114,116,101,100,32,105,109,97,103,101,32,102,111,114,109,97,116,46,0,67,111,117,108,100,32,110,111,116,32,112,97,114,115,101,32,80,86,82,32,102,105,108,101,58,32,105,110,118,97,108,105,100,32,115,105,122,101,32,99,97,108,99,117,108,97,116,105,111,110,46,0,78,52,108,111,118,101,53,105,109,97,103,101,54,109,97,103,112,105,101,49,48,83,84,66,72,97,110,100,108,101,114,69,0,110,111,32,83,79,73,0,101,120,112,101,99,116,101,100,32,109,97,114,107,101,114,0,98,97,100,32,68,82,73,32,108,101,110,0,98,97,100,32,68,81,84,32,116,121,112,101,0,98,97,100,32,68,81,84,32,116,97,98,108,101,0,0,1,8,16,9,2,3,10,17,24,32,25,18,11,4,5,12,19,26,33,40,48,41,34,27,20,13,6,7,14,21,28,35,42,49,56,57,50,43,36,29,22,15,23,30,37,44,51,58,59,52,45,38,31,39,46,53,60,61,54,47,55,62,63,63,63,63,63,63,63,63,63,63,63,63,63,63,63,63,98,97,100,32,68,72,84,32,104,101,97,100,101,114,0,98,97,100,32,99,111,100,101,32,108,101,110,103,116,104,115,0,110,111,32,83,79,70,0,98,97,100,32,83,79,70,32,108,101,110,0,111,110,108,121,32,56,45,98,105,116,0,110,111,32,104,101,97,100,101,114,32,104,101,105,103,104,116,0,48,32,119,105,100,116,104,0,98,97,100,32,99,111,109,112,111,110,101,110,116,32,99,111,117,110,116,0,98,97,100,32,99,111,109,112,111,110,101,110,116,32,73,68,0,98,97,100,32,72,0,98,97,100,32,86,0,98,97,100,32,84,81,0,116,111,111,32,108,97,114,103,101,0,111,117,116,111,102,109,101,109,0,117,110,107,110,111,119,110,32,105,109,97,103,101,32,116,121,112,101,0,98,97,100,32,114,101,113,95,99,111,109,112,0,98,97,100,32,83,79,83,32,99,111,109,112,111,110,101,110,116,32,99,111,117,110,116,0,98,97,100,32,83,79,83,32,108,101,110,0,98,97,100,32,68,67,32,104,117,102,102,0,98,97,100,32,65,67,32,104,117,102,102,0,98,97,100,32,83,79,83,0,40,40,40,106,45,62,99,111,100,101,95,98,117,102,102,101,114,41,32,62,62,32,40,51,50,32,45,32,104,45,62,115,105,122,101,91,99,93,41,41,32,38,32,115,116,98,105,95,95,98,109,97,115,107,91,104,45,62,115,105,122,101,91,99,93,93,41,32,61,61,32,104,45,62,99,111,100,101,91,99,93,0,67,111,117,108,100,32,110,111,116,32,100,101,99,111,100,101,32,105,109,97,103,101,32,40,115,116,98,95,105,109,97,103,101,32,97,115,115,101,114,116,105,111,110,32,39,37,115,39,32,102,97,105,108,101,100,41,0,98,97,100,32,104,117,102,102,109,97,110,32,99,111,100,101,0,110,32,62,61,32,48,32,38,38,32,110,32,60,32,40,105,110,116,41,32,40,115,105,122,101,111,102,40,115,116,98,105,95,95,98,109,97,115,107,41,47,115,105,122,101,111,102,40,42,115,116,98,105,95,95,98,109,97,115,107,41,41,0,99,97,110,39,116,32,109,101,114,103,101,32,100,99,32,97,110,100,32,97,99,0,106,117,110,107,32,98,101,102,111,114,101,32,109,97,114,107,101,114,0,110,111,116,32,66,77,80,0,117,110,107,110,111,119,110,32,66,77,80,0,98,97,100,32,66,77,80,0,109,111,110,111,99,104,114,111,109,101,0,66,77,80,32,82,76,69,0,104,115,122,32,61,61,32,49,48,56,32,124,124,32,104,115,122,32,61,61,32,49,50,52,0,105,110,118,97,108,105,100,0,98,97,100,32,98,112,112,0,98,97,100,32,109,97,115,107,115,0,114,101,113,95,99,111,109,112,32,62,61,32,49,32,38,38,32,114,101,113,95,99,111,109,112,32,60,61,32,52,0,98,97,100,32,112,97,108,101,116,116,101,0,67,111,117,108,100,32,110,111,116,32,100,101,99,111,100,101,32,105,109,97,103,101,32,119,105,116,104,32,115,116,98,95,105,109,97,103,101,46,0,73,110,118,97,108,105,100,32,102,111,114,109,97,116,46,0,78,52,108,111,118,101,53,105,109,97,103,101,57,73,109,97,103,101,68,97,116,97,69,0,116,103,97,0,112,110,103,0,65,116,116,101,109,112,116,32,116,111,32,115,101,116,32,111,117,116,45,111,102,45,114,97,110,103,101,32,112,105,120,101,108,33,0,65,116,116,101,109,112,116,32,116,111,32,103,101,116,32,111,117,116,45,111,102,45,114,97,110,103,101,32,112,105,120,101,108,33,0,78,52,108,111,118,101,56,106,111,121,115,116,105,99,107,51,115,100,108,56,74,111,121,115,116,105,99,107,69,0,78,52,108,111,118,101,56,106,111,121,115,116,105,99,107,56,74,111,121,115,116,105,99,107,69,0,78,52,108,111,118,101,56,106,111,121,115,116,105,99,107,51,115,100,108,49,52,74,111,121,115,116,105,99,107,77,111,100,117,108,101,69,0,78,52,108,111,118,101,56,106,111,121,115,116,105,99,107,49,52,74,111,121,115,116,105,99,107,77,111,100,117,108,101,69,0,108,111,118,101,46,106,111,121,115,116,105,99,107,46,115,100,108,0,73,110,118,97,108,105,100,32,106,111,121,115,116,105,99,107,32,71,85,73,68,58,32,37,115,0,44,67,111,110,116,114,111,108,108,101,114,44,0,73,110,118,97,108,105,100,32,106,111,121,115,116,105,99,107,32,105,110,112,117,116,32,118,97,108,117,101,46,0,73,110,118,97,108,105,100,32,103,97,109,101,112,97,100,32,97,120,105,115,47,98,117,116,116,111,110,46,0,44,32,0,112,108,97,116,102,111,114,109,58,0,73,110,118,97,108,105,100,32,103,97,109,101,112,97,100,32,109,97,112,112,105,110,103,115,46,0,44,10,0,67,111,117,108,100,32,110,111,116,32,105,110,105,116,105,97,108,105,122,101,32,83,68,76,32,106,111,121,115,116,105,99,107,32,115,117,98,115,121,115,116,101,109,32,40,37,115,41,0,114,117,0,114,100,0,108,117,0,108,100,0,116,114,105,103,103,101,114,108,101,102,116,0,116,114,105,103,103,101,114,114,105,103,104,116,0,97,120,105,115,0,98,117,116,116,111,110,0,104,97,116,0,73,110,118,97,108,105,100,0,73,110,118,97,108,105,100,32,103,97,109,101,112,97,100,32,97,120,105,115,58,32,37,115,0,73,110,118,97,108,105,100,32,103,97,109,101,112,97,100,32,98,117,116,116,111,110,58,32,37,115,0,74,111,121,115,116,105,99,107,0,105,115,67,111,110,110,101,99,116,101,100,0,103,101,116,78,97,109,101,0,103,101,116,73,68,0,103,101,116,71,85,73,68,0,103,101,116,65,120,105,115,67,111,117,110,116,0,103,101,116,66,117,116,116,111,110,67,111,117,110,116,0,103,101,116,72,97,116,67,111,117,110,116,0,103,101,116,65,120,105,115,0,103,101,116,65,120,101,115,0,103,101,116,72,97,116,0,105,115,68,111,119,110,0,105,115,71,97,109,101,112,97,100,0,103,101,116,71,97,109,101,112,97,100,65,120,105,115,0,105,115,71,97,109,101,112,97,100,68,111,119,110,0,105,115,86,105,98,114,97,116,105,111,110,83,117,112,112,111,114,116,101,100,0,115,101,116,86,105,98,114,97,116,105,111,110,0,103,101,116,86,105,98,114,97,116,105,111,110,0,103,101,116,67,111,110,110,101,99,116,101,100,73,110,100,101,120,0,103,101,116,71,97,109,101,112,97,100,77,97,112,112,105,110,103,0,78,52,108,111,118,101,56,107,101,121,98,111,97,114,100,51,115,100,108,56,75,101,121,98,111,97,114,100,69,0,78,52,108,111,118,101,56,107,101,121,98,111,97,114,100,56,75,101,121,98,111,97,114,100,69,0,108,111,118,101,46,107,101,121,98,111,97,114,100,46,115,100,108,0,117,110,107,110,111,119,110,0,101,115,99,97,112,101,0,98,97,99,107,115,112,97,99,101,0,116,97,98,0,115,112,97,99,101,0,33,0,34,0,35,0,36,0,38,0,39,0,40,0,41,0,44,0,50,0,51,0,52,0,56,0,58,0,59,0,60,0,61,0,62,0,64,0,91,0,92,0,93,0,94,0,96,0,99,0,101,0,103,0,105,0,106,0,107,0,111,0,112,0,113,0,115,0,116,0,117,0,118,0,122,0,99,97,112,115,108,111,99,107,0,102,49,0,102,50,0,102,51,0,102,52,0,102,53,0,102,54,0,102,55,0,102,56,0,102,57,0,102,49,48,0,102,49,49,0,102,49,50,0,112,114,105,110,116,115,99,114,101,101,110,0,115,99,114,111,108,108,108,111,99,107,0,112,97,117,115,101,0,104,111,109,101,0,112,97,103,101,117,112,0,100,101,108,101,116,101,0,112,97,103,101,100,111,119,110,0,114,105,103,104,116,0,108,101,102,116,0,100,111,119,110,0,117,112,0,110,117,109,108,111,99,107,0,107,112,47,0,107,112,42,0,107,112,45,0,107,112,43,0,107,112,101,110,116,101,114,0,107,112,48,0,107,112,49,0,107,112,50,0,107,112,51,0,107,112,52,0,107,112,53,0,107,112,54,0,107,112,55,0,107,112,56,0,107,112,57,0,107,112,46,0,107,112,44,0,107,112,61,0,97,112,112,108,105,99,97,116,105,111,110,0,112,111,119,101,114,0,102,49,51,0,102,49,52,0,102,49,53,0,102,49,54,0,102,49,55,0,102,49,56,0,102,49,57,0,102,50,48,0,102,50,49,0,102,50,50,0,102,50,51,0,102,50,52,0,104,101,108,112,0,109,101,110,117,0,97,103,97,105,110,0,117,110,100,111,0,99,117,116,0,99,111,112,121,0,112,97,115,116,101,0,118,111,108,117,109,101,117,112,0,118,111,108,117,109,101,100,111,119,110,0,97,108,116,101,114,97,115,101,0,115,121,115,114,101,113,0,99,97,110,99,101,108,0,99,108,101,97,114,0,112,114,105,111,114,0,114,101,116,117,114,110,50,0,115,101,112,97,114,97,116,111,114,0,111,117,116,0,111,112,101,114,0,99,108,101,97,114,97,103,97,105,110,0,116,104,115,111,117,115,97,110,100,115,115,101,112,97,114,97,116,111,114,0,100,101,99,105,109,97,108,115,101,112,97,114,97,116,111,114,0,99,117,114,114,101,110,99,121,117,110,105,116,0,99,117,114,114,101,110,99,121,115,117,98,117,110,105,116,0,108,99,116,114,108,0,108,115,104,105,102,116,0,108,97,108,116,0,108,103,117,105,0,114,99,116,114,108,0,114,115,104,105,102,116,0,114,97,108,116,0,114,103,117,105,0,97,117,100,105,111,110,101,120,116,0,97,117,100,105,111,112,114,101,118,0,97,117,100,105,111,115,116,111,112,0,97,117,100,105,111,112,108,97,121,0,97,117,100,105,111,109,117,116,101,0,109,101,100,105,97,115,101,108,101,99,116,0,119,119,119,0,109,97,105,108,0,99,97,108,99,117,108,97,116,111,114,0,99,111,109,112,117,116,101,114,0,97,112,112,115,101,97,114,99,104,0,97,112,112,104,111,109,101,0,97,112,112,98,97,99,107,0,97,112,112,102,111,114,119,97,114,100,0,97,112,112,115,116,111,112,0,97,112,112,114,101,102,114,101,115,104,0,97,112,112,98,111,111,107,109,97,114,107,115,0,98,114,105,103,104,116,110,101,115,115,100,111,119,110,0,98,114,105,103,104,116,110,101,115,115,117,112,0,100,105,115,112,108,97,121,115,119,105,116,99,104,0,107,98,100,105,108,108,117,109,116,111,103,103,108,101,0,107,98,100,105,108,108,117,109,100,111,119,110,0,107,98,100,105,108,108,117,109,117,112,0,101,106,101,99,116,0,115,108,101,101,112,0,110,111,110,117,115,35,0,110,111,110,117,115,98,97,99,107,115,108,97,115,104,0,107,112,61,52,48,48,0,105,110,116,101,114,110,97,116,105,111,110,97,108,49,0,105,110,116,101,114,110,97,116,105,111,110,97,108,50,0,105,110,116,101,114,110,97,116,105,111,110,97,108,51,0,105,110,116,101,114,110,97,116,105,111,110,97,108,52,0,105,110,116,101,114,110,97,116,105,111,110,97,108,53,0,105,110,116,101,114,110,97,116,105,111,110,97,108,54,0,105,110,116,101,114,110,97,116,105,111,110,97,108,55,0,105,110,116,101,114,110,97,116,105,111,110,97,108,56,0,105,110,116,101,114,110,97,116,105,111,110,97,108,57,0,108,97,110,103,49,0,108,97,110,103,50,0,108,97,110,103,51,0,108,97,110,103,52,0,108,97,110,103,53,0,108,97,110,103,54,0,108,97,110,103,55,0,108,97,110,103,56,0,108,97,110,103,57,0,99,114,115,101,108,0,101,120,115,101,108,0,107,112,48,48,0,107,112,48,48,48,0,107,112,40,0,107,112,41,0,107,112,123,0,107,112,125,0,107,112,116,97,98,0,107,112,98,97,99,107,115,112,97,99,101,0,107,112,97,0,107,112,98,0,107,112,99,0,107,112,100,0,107,112,101,0,107,112,102,0,107,112,120,111,114,0,107,112,111,119,101,114,0,107,112,37,0,107,112,60,0,107,112,62,0,107,112,38,0,107,112,38,38,0,107,112,124,0,107,112,124,124,0,107,112,58,0,107,112,35,0,107,112,32,0,107,112,64,0,107,112,33,0,107,112,109,101,109,115,116,111,114,101,0,107,112,109,101,109,114,101,99,97,108,108,0,107,112,109,101,109,99,108,101,97,114,0,107,112,109,101,109,43,0,107,112,109,101,109,45,0,107,112,109,101,109,42,0,107,112,109,101,109,47,0,107,112,43,45,0,107,112,99,108,101,97,114,0,107,112,99,108,101,97,114,101,110,116,114,121,0,107,112,98,105,110,97,114,121,0,107,112,111,99,116,97,108,0,107,112,100,101,99,105,109,97,108,0,107,112,104,101,120,0,97,99,115,101,97,114,99,104,0,97,99,104,111,109,101,0,97,99,98,97,99,107,0,97,99,102,111,114,119,97,114,100,0,97,99,115,116,111,112,0,97,99,114,101,102,114,101,115,104,0,97,99,98,111,111,107,109,97,114,107,115,0,97,112,112,49,0,97,112,112,50,0,67,111,110,115,116,97,110,116,32,37,115,32,111,117,116,32,111,102,32,98,111,117,110,100,115,32,119,105,116,104,32,37,117,33,10,0,78,52,108,111,118,101,52,109,97,116,104,49,49,66,101,122,105,101,114,67,117,114,118,101,69,0,67,97,110,110,111,116,32,100,101,114,105,118,101,32,97,32,99,117,114,118,101,32,111,102,32,100,101,103,114,101,101,32,60,32,49,46,0,73,110,118,97,108,105,100,32,101,118,97,108,117,97,116,105,111,110,32,112,97,114,97,109,101,116,101,114,58,32,109,117,115,116,32,98,101,32,98,101,116,119,101,101,110,32,48,32,97,110,100,32,49,0,73,110,118,97,108,105,100,32,66,101,122,105,101,114,32,99,117,114,118,101,58,32,78,111,116,32,101,110,111,117,103,104,32,99,111,110,116,114,111,108,32,112,111,105,110,116,115,46,0,73,110,118,97,108,105,100,32,115,101,103,109,101,110,116,32,112,97,114,97,109,101,116,101,114,115,58,32,109,117,115,116,32,98,101,32,98,101,116,119,101,101,110,32,48,32,97,110,100,32,49,0,73,110,118,97,108,105,100,32,115,101,103,109,101,110,116,32,112,97,114,97,109,101,116,101,114,115,58,32,116,49,32,109,117,115,116,32,98,101,32,115,109,97,108,108,101,114,32,116,104,97,110,32,116,50,0,78,52,108,111,118,101,53,109,111,117,115,101,51,115,100,108,54,67,117,114,115,111,114,69,0,67,97,110,110,111,116,32,99,114,101,97,116,101,32,99,117,114,115,111,114,58,32,111,117,116,32,111,102,32,109,101,109,111,114,121,33,0,67,97,110,110,111,116,32,99,114,101,97,116,101,32,99,117,114,115,111,114,58,32,37,115,0,67,97,110,110,111,116,32,99,114,101,97,116,101,32,115,121,115,116,101,109,32,99,117,114,115,111,114,58,32,105,110,118,97,108,105,100,32,116,121,112,101,46,0,67,97,110,110,111,116,32,99,114,101,97,116,101,32,115,121,115,116,101,109,32,99,117,114,115,111,114,58,32,37,115,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,52,66,111,100,121,69,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,49,48,67,104,97,105,110,83,104,97,112,101,69,0,80,104,121,115,105,99,115,32,101,114,114,111,114,58,32,67,97,110,39,116,32,99,97,108,108,32,115,101,116,78,101,120,116,86,101,114,116,101,120,32,111,110,32,97,32,108,111,111,112,32,67,104,97,105,110,83,104,97,112,101,0,80,104,121,115,105,99,115,32,101,114,114,111,114,58,32,67,97,110,39,116,32,99,97,108,108,32,115,101,116,80,114,101,118,105,111,117,115,86,101,114,116,101,120,32,111,110,32,97,32,108,111,111,112,32,67,104,97,105,110,83,104,97,112,101,0,80,104,121,115,105,99,115,32,101,114,114,111,114,58,32,105,110,100,101,120,32,111,117,116,32,111,102,32,98,111,117,110,100,115,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,49,49,67,105,114,99,108,101,83,104,97,112,101,69,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,55,67,111,110,116,97,99,116,69,0,65,32,102,105,120,116,117,114,101,32,104,97,115,32,101,115,99,97,112,101,100,32,77,101,109,111,105,122,101,114,33,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,49,51,68,105,115,116,97,110,99,101,74,111,105,110,116,69,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,57,69,100,103,101,83,104,97,112,101,69,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,55,70,105,120,116,117,114,101,69,0,98,50,73,115,86,97,108,105,100,40,100,101,110,115,105,116,121,41,32,38,38,32,100,101,110,115,105,116,121,32,62,61,32,48,46,48,102,0,86,97,108,117,101,115,32,109,117,115,116,32,98,101,32,105,110,32,114,97,110,103,101,32,49,45,49,54,46,0,98,105,116,115,101,116,32,115,101,116,32,97,114,103,117,109,101,110,116,32,111,117,116,32,111,102,32,114,97,110,103,101,0,48,32,60,61,32,99,104,105,108,100,73,110,100,101,120,32,38,38,32,99,104,105,108,100,73,110,100,101,120,32,60,32,109,95,112,114,111,120,121,67,111,117,110,116,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,49,51,70,114,105,99,116,105,111,110,74,111,105,110,116,69,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,57,71,101,97,114,74,111,105,110,116,69,0,65,32,106,111,105,110,116,32,104,97,115,32,101,115,99,97,112,101,100,32,77,101,109,111,105,122,101,114,33,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,49,48,77,111,116,111,114,74,111,105,110,116,69,0,78,52,108,111,118,101,55,112,104,121,115,105,99,115,53,98,111,120,50,100,49,48,77,111,117,115,101,74,111,105,110,116,69,0,78,52,108,111,118,101,53,115,111,117,110,100,55,108,117,108,108,97,98,121,55,68,101,99,111,100,101,114,69,0,78,52,108,111,118,101,53,115,111,117,110,100,55,68,101,99,111,100,101,114,69,0,78,52,108,111,118,101,53,115,111,117,110,100,55,108,117,108,108,97,98,121,49,52,77,111,100,80,108,117,103,68,101,99,111,100,101,114,69,0,67,111,117,108,100,32,110,111,116,32,108,111,97,100,32,102,105,108,101,32,119,105,116,104,32,77,111,100,80,108,117,103,46,0,54,57,57,0,97,98,99,0,97,109,102,0,97,109,115,0,100,98,109,0,100,109,102,0,100,115,109,0,102,97,114,0,105,116,0,106,50,98,0,109,100,108,0,109,101,100,0,109,105,100,0,109,116,50,0,109,116,109,0,111,107,116,0,112,97,116,0,112,115,109,0,115,51,109,0,115,116,109,0,117,108,116,0,117,109,120,0,120,109,0,78,52,108,111,118,101,54,116,104,114,101,97,100,51,115,100,108,53,77,117,116,101,120,69,0,78,52,108,111,118,101,54,116,104,114,101,97,100,53,77,117,116,101,120,69,0,78,52,108,111,118,101,54,116,104,114,101,97,100,51,115,100,108,49,49,67,111,110,100,105,116,105,111,110,97,108,69,0,78,52,108,111,118,101,54,116,104,114,101,97,100,49,49,67,111,110,100,105,116,105,111,110,97,108,69,0,78,52,108,111,118,101,54,116,104,114,101,97,100,55,67,104,97,110,110,101,108,69,0,78,52,108,111,118,101,54,116,104,114,101,97,100,57,76,117,97,84,104,114,101,97,100,69,0,108,111,118,101,0,108,111,118,101,46,116,104,114,101,97,100,0,108,111,118,101,46,102,105,108,101,115,121,115,116,101,109,0,116,104,114,101,97,100,101,114,114,111,114,0,100,101,110,32,62,32,48,46,48,102,0,48,32,60,61,32,101,100,103,101,49,32,38,38,32,101,100,103,101,49,32,60,32,112,111,108,121,49,45,62,109,95,99,111,117,110,116,0,109,95,110,111,100,101,67,111,117,110,116,32,61,61,32,109,95,110,111,100,101,67,97,112,97,99,105,116,121,0,48,32,60,61,32,110,111,100,101,73,100,32,38,38,32,110,111,100,101,73,100,32,60,32,109,95,110,111,100,101,67,97,112,97,99,105,116,121,0,48,32,60,32,109,95,110,111,100,101,67,111,117,110,116,0,105,65,32,33,61,32,98,50,95,110,117,108,108,78,111,100,101,0,48,32,60,61,32,105,66,32,38,38,32,105,66,32,60,32,109,95,110,111,100,101,67,97,112,97,99,105,116,121,0,48,32,60,61,32,105,67,32,38,38,32,105,67,32,60,32,109,95,110,111,100,101,67,97,112,97,99,105,116,121,0,48,32,60,61,32,105,70,32,38,38,32,105,70,32,60,32,109,95,110,111,100,101,67,97,112,97,99,105,116,121,0,48,32,60,61,32,105,71,32,38,38,32,105,71,32,60,32,109,95,110,111,100,101,67,97,112,97,99,105,116,121,0,109,95,110,111,100,101,115,91,67,45,62,112,97,114,101,110,116,93,46,99,104,105,108,100,50,32,61,61,32,105,65,0,48,32,60,61,32,105,68,32,38,38,32,105,68,32,60,32,109,95,110,111,100,101,67,97,112,97,99,105,116,121,0,48,32,60,61,32,105,69,32,38,38,32,105,69,32,60,32,109,95,110,111,100,101,67,97,112,97,99,105,116,121,0,109,95,110,111,100,101,115,91,66,45,62,112,97,114,101,110,116,93,46,99,104,105,108,100,50,32,61,61,32,105,65,0,99,104,105,108,100,49,32,33,61,32,98,50,95,110,117,108,108,78,111,100,101,0,99,104,105,108,100,50,32,33,61,32,98,50,95,110,117,108,108,78,111,100,101,0,109,95,110,111,100,101,115,91,112,114,111,120,121,73,100,93,46,73,115,76,101,97,102,40,41,0,116,97,114,103,101,116,32,62,32,116,111,108,101,114,97,110,99,101,0,48,32,60,32,99,111,117,110,116,32,38,38,32,99,111,117,110,116,32,60,32,51,0,48,32,60,61,32,105,110,100,101,120,32,38,38,32,105,110,100,101,120,32,60,32,109,95,99,111,117,110,116,0,109,95,105,110,100,101,120,32,61,61,32,48,0,109,95,101,110,116,114,121,67,111,117,110,116,32,61,61,32,48,0,109,95,101,110,116,114,121,67,111,117,110,116,32,60,32,98,50,95,109,97,120,83,116,97,99,107,69,110,116,114,105,101,115,0,109,95,101,110,116,114,121,67,111,117,110,116,32,62,32,48,0,112,32,61,61,32,101,110,116,114,121,45,62,100,97,116,97,0,98,100,45,62,112,111,115,105,116,105,111,110,46,73,115,86,97,108,105,100,40,41,0,98,100,45,62,108,105,110,101,97,114,86,101,108,111,99,105,116,121,46,73,115,86,97,108,105,100,40,41,0,98,50,73,115,86,97,108,105,100,40,98,100,45,62,97,110,103,108,101,41,0,98,50,73,115,86,97,108,105,100,40,98,100,45,62,97,110,103,117,108,97,114,86,101,108,111,99,105,116,121,41,0,98,50,73,115,86,97,108,105,100,40,98,100,45,62,97,110,103,117,108,97,114,68,97,109,112,105,110,103,41,32,38,38,32,98,100,45,62,97,110,103,117,108,97,114,68,97,109,112,105,110,103,32,62,61,32,48,46,48,102,0,98,50,73,115,86,97,108,105,100,40,98,100,45,62,108,105,110,101,97,114,68,97,109,112,105,110,103,41,32,38,38,32,98,100,45,62,108,105,110,101,97,114,68,97,109,112,105,110,103,32,62,61,32,48,46,48,102,0,109,95,119,111,114,108,100,45,62,73,115,76,111,99,107,101,100,40,41,32,61,61,32,102,97,108,115,101,0,109,95,116,121,112,101,32,61,61,32,98,50,95,100,121,110,97,109,105,99,66,111,100,121,0,109,95,73,32,62,32,48,46,48,102,0,102,105,120,116,117,114,101,45,62,109,95,98,111,100,121,32,61,61,32,116,104,105,115,0,109,95,102,105,120,116,117,114,101,67,111,117,110,116,32,62,32,48,0,102,111,117,110,100,0,48,32,60,61,32,112,114,111,120,121,73,100,32,38,38,32,112,114,111,120,121,73,100,32,60,32,109,95,110,111,100,101,67,97,112,97,99,105,116,121,0,109,95,99,111,117,110,116,32,62,32,48,0,109,95,112,114,111,120,121,67,111,117,110,116,32,61,61,32,48,0,116,111,105,73,110,100,101,120,65,32,60,32,109,95,98,111,100,121,67,111,117,110,116,0,116,111,105,73,110,100,101,120,66,32,60,32,109,95,98,111,100,121,67,111,117,110,116,0,50,51,98,50,67,104,97,105,110,65,110,100,67,105,114,99,108,101,67,111,110,116,97,99,116,0,50,52,98,50,67,104,97,105,110,65,110,100,80,111,108,121,103,111,110,67,111,110,116,97,99,116,0,109,95,102,105,120,116,117,114,101,65,45,62,71,101,116,84,121,112,101,40,41,32,61,61,32,98,50,83,104,97,112,101,58,58,101,95,99,104,97,105,110,0,109,95,102,105,120,116,117,114,101,66,45,62,71,101,116,84,121,112,101,40,41,32,61,61,32,98,50,83,104,97,112,101,58,58,101,95,112,111,108,121,103,111,110,0,49,53,98,50,67,105,114,99,108,101,67,111,110,116,97,99,116,0,109,95,102,105,120,116,117,114,101,65,45,62,71,101,116,84,121,112,101,40,41,32,61,61,32,98,50,83,104,97,112,101,58,58,101,95,99,105,114,99,108,101,0,109,95,102,105,120,116,117,114,101,66,45,62,71,101,116,84,121,112,101,40,41,32,61,61,32,98,50,83,104,97,112,101,58,58,101,95,99,105,114,99,108,101,0,1,112,111,105,110,116,67,111,117,110,116,32,62,32,48,0,109,97,110,105,102,111,108,100,45,62,112,111,105,110,116,67,111,117,110,116,32,62,32,48,0,112,111,105,110,116,67,111,117,110,116,32,61,61,32,49,32,124,124,32,112,111,105,110,116,67,111,117,110,116,32,61,61,32,50,0,97,46,120,32,62,61,32,48,46,48,102,32,38,38,32,97,46,121,32,62,61,32,48,46,48,102,0,112,99,45,62,112,111,105,110,116,67,111,117,110,116,32,62,32,48,0,49,53,98,50,68,105,115,116,97,110,99,101,74,111,105,110,116,0,32,32,98,50,68,105,115,116,97,110,99,101,74,111,105,110,116,68,101,102,32,106,100,59,10,0,32,32,106,100,46,108,101,110,103,116,104,32,61,32,37,46,49,53,108,101,102,59,10,0,32,32,106,100,46,102,114,101,113,117,101,110,99,121,72,122,32,61,32,37,46,49,53,108,101,102,59,10,0,32,32,106,100,46,100,97,109,112,105,110,103,82,97,116,105,111,32,61,32,37,46,49,53,108,101,102,59,10,0,49,53,98,50,70,114,105,99,116,105,111,110,74,111,105,110,116,0,32,32,98,50,70,114,105,99,116,105,111,110,74,111,105,110,116,68,101,102,32,106,100,59,10,0,32,32,106,100,46,108,111,99,97,108,65,110,99,104,111,114,65,46,83,101,116,40,37,46,49,53,108,101,102,44,32,37,46,49,53,108,101,102,41,59,10,0,32,32,106,100,46,108,111,99,97,108,65,110,99,104,111,114,66,46,83,101,116,40,37,46,49,53,108,101,102,44,32,37,46,49,53,108,101,102,41,59,10,0,32,32,106,100,46,109,97,120,70,111,114,99,101,32,61,32,37,46,49,53,108,101,102,59,10,0,32,32,106,100,46,109,97,120,84,111,114,113,117,101,32,61,32,37,46,49,53,108,101,102,59,10,0,98,50,73,115,86,97,108,105,100,40,102,111,114,99,101,41,32,38,38,32,102,111,114,99,101,32,62,61,32,48,46,48,102,0,98,50,73,115,86,97,108,105,100,40,116,111,114,113,117,101,41,32,38,38,32,116,111,114,113,117,101,32,62,61,32,48,46,48,102,0,49,49,98,50,71,101,97,114,74,111,105,110,116,0,32,32,98,50,71,101,97,114,74,111,105,110,116,68,101,102,32,106,100,59,10,0,32,32,106,100,46,98,111,100,121,65,32,61,32,98,111,100,105,101,115,91,37,100,93,59,10,0,32,32,106,100,46,98,111,100,121,66,32,61,32,98,111,100,105,101,115,91,37,100,93,59,10,0,32,32,106,100,46,99,111,108,108,105,100,101,67,111,110,110,101,99,116,101,100,32,61,32,98,111,111,108,40,37,100,41,59,10,0,32,32,106,100,46,106,111,105,110,116,49,32,61,32,106,111,105,110,116,115,91,37,100,93,59,10,0,32,32,106,100,46,106,111,105,110,116,50,32,61,32,106,111,105,110,116,115,91,37,100,93,59,10,0,32,32,106,100,46,114,97,116,105,111,32,61,32,37,46,49,53,108,101,102,59,10,0,32,32,106,111,105,110,116,115,91,37,100,93,32,61,32,109,95,119,111,114,108,100,45,62,67,114,101,97,116,101,74,111,105,110,116,40,38,106,100,41,59,10,0,109,95,116,121,112,101,65,32,61,61,32,101,95,114,101,118,111,108,117,116,101,74,111,105,110,116,32,124,124,32,109,95,116,121,112,101,65,32,61,61,32,101,95,112,114,105,115,109,97,116,105,99,74,111,105,110,116,0,109,95,116,121,112,101,66,32,61,61,32,101,95,114,101,118,111,108,117,116,101,74,111,105,110,116,32,124,124,32,109,95,116,121,112,101,66,32,61,61,32,101,95,112,114,105,115,109,97,116,105,99,74,111,105,110,116,0,98,50,73,115,86,97,108,105,100,40,114,97,116,105,111,41,0,79,112,101,110,71,76,32,69,83,32,0,87,101,98,71,76,32,0,103,108,80,111,105,110,116,83,105,122,101,0,103,108,80,111,108,121,103,111,110,77,111,100,101,0,103,108,84,101,120,73,109,97,103,101,49,68,0,103,108,68,114,97,119,66,117,102,102,101,114,0,103,108,76,111,103,105,99,79,112,0,103,108,80,105,120,101,108,83,116,111,114,101,102,0,103,108,82,101,97,100,66,117,102,102,101,114,0,103,108,71,101,116,68,111,117,98,108,101,118,0,103,108,71,101,116,84,101,120,73,109,97,103,101,0,103,108,71,101,116,84,101,120,76,101,118,101,108,80,97,114,97,109,101,116,101,114,102,118,0,103,108,71,101,116,84,101,120,76,101,118,101,108,80,97,114,97,109,101,116,101,114,105,118,0,103,108,78,101,119,76,105,115,116,0,103,108,69,110,100,76,105,115,116,0,103,108,67,97,108,108,76,105,115,116,0,103,108,67,97,108,108,76,105,115,116,115,0,103,108,68,101,108,101,116,101,76,105,115,116,115,0,103,108,71,101,110,76,105,115,116,115,0,103,108,76,105,115,116,66,97,115,101,0,103,108,66,101,103,105,110,0,103,108,66,105,116,109,97,112,0,103,108,67,111,108,111,114,51,98,0,103,108,67,111,108,111,114,51,98,118,0,103,108,67,111,108,111,114,51,100,0,103,108,67,111,108,111,114,51,100,118,0,103,108,67,111,108,111,114,51,102,0,103,108,67,111,108,111,114,51,102,118,0,103,108,67,111,108,111,114,51,105,0,103,108,67,111,108,111,114,51,105,118,0,103,108,67,111,108,111,114,51,115,0,103,108,67,111,108,111,114,51,115,118,0,103,108,67,111,108,111,114,51,117,98,0,103,108,67,111,108,111,114,51,117,98,118,0,103,108,67,111,108,111,114,51,117,105,0,103,108,67,111,108,111,114,51,117,105,118,0,103,108,67,111,108,111,114,51,117,115,0,103,108,67,111,108,111,114,51,117,115,118,0,103,108,67,111,108,111,114,52,98,0,103,108,67,111,108,111,114,52,98,118,0,103,108,67,111,108,111,114,52,100,0,103,108,67,111,108,111,114,52,100,118,0,103,108,67,111,108,111,114,52,102,0,103,108,67,111,108,111,114,52,102,118,0,103,108,67,111,108,111,114,52,105,0,103,108,67,111,108,111,114,52,105,118,0,103,108,67,111,108,111,114,52,115,0,103,108,67,111,108,111,114,52,115,118,0,103,108,67,111,108,111,114,52,117,98,0,103,108,67,111,108,111,114,52,117,98,118,0,103,108,67,111,108,111,114,52,117,105,0,103,108,67,111,108,111,114,52,117,105,118,0,103,108,67,111,108,111,114,52,117,115,0,103,108,67,111,108,111,114,52,117,115,118,0,103,108,69,100,103,101,70,108,97,103,0,103,108,69,100,103,101,70,108,97,103,118,0,103,108,69,110,100,0,103,108,73,110,100,101,120,100,0,103,108,73,110,100,101,120,100,118,0,103,108,73,110,100,101,120,102,0,103,108,73,110,100,101,120,102,118,0,103,108,73,110,100,101,120,105,0,103,108,73,110,100,101,120,105,118,0,103,108,73,110,100,101,120,115,0,103,108,73,110,100,101,120,115,118,0,103,108,78,111,114,109,97,108,51,98,0,103,108,78,111,114,109,97,108,51,98,118,0,103,108,78,111,114,109,97,108,51,100,0,103,108,78,111,114,109,97,108,51,100,118,0,103,108,78,111,114,109,97,108,51,102,0,103,108,78,111,114,109,97,108,51,102,118,0,103,108,78,111,114,109,97,108,51,105,0,103,108,78,111,114,109,97,108,51,105,118,0,103,108,78,111,114,109,97,108,51,115,0,103,108,78,111,114,109,97,108,51,115,118,0,103,108,82,97,115,116,101,114,80,111,115,50,100,0,103,108,82,97,115,116,101,114,80,111,115,50,100,118,0,103,108,82,97,115,116,101,114,80,111,115,50,102,0,103,108,82,97,115,116,101,114,80,111,115,50,102,118,0,103,108,82,97,115,116,101,114,80,111,115,50,105,0,103,108,82,97,115,116,101,114,80,111,115,50,105,118,0,103,108,82,97,115,116,101,114,80,111,115,50,115,0,103,108,82,97,115,116,101,114,80,111,115,50,115,118,0,103,108,82,97,115,116,101,114,80,111,115,51,100,0,103,108,82,97,115,116,101,114,80,111,115,51,100,118,0,103,108,82,97,115,116,101,114,80,111,115,51,102,0,103,108,82,97,115,116,101,114,80,111,115,51,102,118,0,103,108,82,97,115,116,101,114,80,111,115,51,105,0,103,108,82,97,115,116,101,114,80,111,115,51,105,118,0,103,108,82,97,115,116,101,114,80,111,115,51,115,0,103,108,82,97,115,116,101,114,80,111,115,51,115,118,0,103,108,82,97,115,116,101,114,80,111,115,52,100,0,103,108,82,97,115,116,101,114,80,111,115,52,100,118,0,103,108,82,97,115,116,101,114,80,111,115,52,102,0,103,108,82,97,115,116,101,114,80,111,115,52,102,118,0,103,108,82,97,115,116,101,114,80,111,115,52,105,0,103,108,82,97,115,116,101,114,80,111,115,52,105,118,0,103,108,82,97,115,116,101,114,80,111,115,52,115,0,103,108,82,97,115,116,101,114,80,111,115,52,115,118,0,103,108,82,101,99,116,100,0,103,108,82,101,99,116,100,118,0,103,108,82,101,99,116,102,0,103,108,82,101,99,116,102,118,0,103,108,82,101,99,116,105,0,103,108,82,101,99,116,105,118,0,103,108,82,101,99,116,115,0,103,108,82,101,99,116,115,118,0,103,108,84,101,120,67,111,111,114,100,49,100,0,103,108,84,101,120,67,111,111,114,100,49,100,118,0,103,108,84,101,120,67,111,111,114,100,49,102,0,103,108,84,101,120,67,111,111,114,100,49,102,118,0,103,108,84,101,120,67,111,111,114,100,49,105,0,103,108,84,101,120,67,111,111,114,100,49,105,118,0,103,108,84,101,120,67,111,111,114,100,49,115,0,103,108,84,101,120,67,111,111,114,100,49,115,118,0,103,108,84,101,120,67,111,111,114,100,50,100,0,103,108,84,101,120,67,111,111,114,100,50,100,118,0,103,108,84,101,120,67,111,111,114,100,50,102,0,103,108,84,101,120,67,111,111,114,100,50,102,118,0,103,108,84,101,120,67,111,111,114,100,50,105,0,103,108,84,101,120,67,111,111,114,100,50,105,118,0,103,108,84,101,120,67,111,111,114,100,50,115,0,103,108,84,101,120,67,111,111,114,100,50,115,118,0,103,108,84,101,120,67,111,111,114,100,51,100,0,103,108,84,101,120,67,111,111,114,100,51,100,118,0,103,108,84,101,120,67,111,111,114,100,51,102,0,103,108,84,101,120,67,111,111,114,100,51,102,118,0,103,108,84,101,120,67,111,111,114,100,51,105,0,103,108,84,101,120,67,111,111,114,100,51,105,118,0,103,108,84,101,120,67,111,111,114,100,51,115,0,103,108,84,101,120,67,111,111,114,100,51,115,118,0,103,108,84,101,120,67,111,111,114,100,52,100,0,103,108,84,101,120,67,111,111,114,100,52,100,118,0,103,108,84,101,120,67,111,111,114,100,52,102,0,103,108,84,101,120,67,111,111,114,100,52,102,118,0,103,108,84,101,120,67,111,111,114,100,52,105,0,103,108,84,101,120,67,111,111,114,100,52,105,118,0,103,108,84,101,120,67,111,111,114,100,52,115,0,103,108,84,101,120,67,111,111,114,100,52,115,118,0,103,108,86,101,114,116,101,120,50,100,0,103,108,86,101,114,116,101,120,50,100,118,0,103,108,86,101,114,116,101,120,50,102,0,103,108,86,101,114,116,101,120,50,102,118,0,103,108,86,101,114,116,101,120,50,105,0,103,108,86,101,114,116,101,120,50,105,118,0,103,108,86,101,114,116,101,120,50,115,0,103,108,86,101,114,116,101,120,50,115,118,0,103,108,86,101,114,116,101,120,51,100,0,103,108,86,101,114,116,101,120,51,100,118,0,103,108,86,101,114,116,101,120,51,102,0,103,108,86,101,114,116,101,120,51,102,118,0,103,108,86,101,114,116,101,120,51,105,0,103,108,86,101,114,116,101,120,51,105,118,0,103,108,86,101,114,116,101,120,51,115,0,103,108,86,101,114,116,101,120,51,115,118,0,103,108,86,101,114,116,101,120,52,100,0,103,108,86,101,114,116,101,120,52,100,118,0,103,108,86,101,114,116,101,120,52,102,0,103,108,86,101,114,116,101,120,52,102,118,0,103,108,86,101,114,116,101,120,52,105,0,103,108,86,101,114,116,101,120,52,105,118,0,103,108,86,101,114,116,101,120,52,115,0,103,108,86,101,114,116,101,120,52,115,118,0,103,108,67,108,105,112,80,108,97,110,101,0,103,108,67,111,108,111,114,77,97,116,101,114,105,97,108,0,103,108,70,111,103,102,0,103,108,70,111,103,102,118,0,103,108,70,111,103,105,0,103,108,70,111,103,105,118,0,103,108,76,105,103,104,116,102,0,103,108,76,105,103,104,116,102,118,0,103,108,76,105,103,104,116,105,0,103,108,76,105,103,104,116,105,118,0,103,108,76,105,103,104,116,77,111,100,101,108,102,0,103,108,76,105,103,104,116,77,111,100,101,108,102,118,0,103,108,76,105,103,104,116,77,111,100,101,108,105,0,103,108,76,105,103,104,116,77,111,100,101,108,105,118,0,103,108,76,105,110,101,83,116,105,112,112,108,101,0,103,108,77,97,116,101,114,105,97,108,102,0,103,108,77,97,116,101,114,105,97,108,102,118,0,103,108,77,97,116,101,114,105,97,108,105,0,103,108,77,97,116,101,114,105,97,108,105,118,0,103,108,80,111,108,121,103,111,110,83,116,105,112,112,108,101,0,103,108,83,104,97,100,101,77,111,100,101,108,0,103,108,84,101,120,69,110,118,102,0,103,108,84,101,120,69,110,118,102,118,0,103,108,84,101,120,69,110,118,105,0,103,108,84,101,120,69,110,118,105,118,0,103,108,84,101,120,71,101,110,100,0,103,108,84,101,120,71,101,110,100,118,0,103,108,84,101,120,71,101,110,102,0,103,108,84,101,120,71,101,110,102,118,0,103,108,84,101,120,71,101,110,105,0,103,108,84,101,120,71,101,110,105,118,0,103,108,70,101,101,100,98,97,99,107,66,117,102,102,101,114,0,103,108,83,101,108,101,99,116,66,117,102,102,101,114,0,103,108,82,101,110,100,101,114,77,111,100,101,0,103,108,73,110,105,116,78,97,109,101,115,0,103,108,76,111,97,100,78,97,109,101,0,103,108,80,97,115,115,84,104,114,111,117,103,104,0,103,108,80,111,112,78,97,109,101,0,103,108,80,117,115,104,78,97,109,101,0,103,108,67,108,101,97,114,65,99,99,117,109,0,103,108,67,108,101,97,114,73,110,100,101,120,0,103,108,73,110,100,101,120,77,97,115,107,0,103,108,65,99,99,117,109,0,103,108,80,111,112,65,116,116,114,105,98,0,103,108,80,117,115,104,65,116,116,114,105,98,0,103,108,77,97,112,49,100,0,103,108,77,97,112,49,102,0,103,108,77,97,112,50,100,0,103,108,77,97,112,50,102,0,103,108,77,97,112,71,114,105,100,49,100,0,103,108,77,97,112,71,114,105,100,49,102,0,103,108,77,97,112,71,114,105,100,50,100,0,103,108,77,97,112,71,114,105,100,50,102,0,103,108,69,118,97,108,67,111,111,114,100,49,100,0,103,108,69,118,97,108,67,111,111,114,100,49,100,118,0,103,108,69,118,97,108,67,111,111,114,100,49,102,0,103,108,69,118,97,108,67,111,111,114,100,49,102,118,0,103,108,69,118,97,108,67,111,111,114,100,50,100,0,103,108,69,118,97,108,67,111,111,114,100,50,100,118,0,103,108,69,118,97,108,67,111,111,114,100,50,102,0,103,108,69,118,97,108,67,111,111,114,100,50,102,118,0,103,108,69,118,97,108,77,101,115,104,49,0,103,108,69,118,97,108,80,111,105,110,116,49,0,103,108,69,118,97,108,77,101,115,104,50,0,103,108,69,118,97,108,80,111,105,110,116,50,0,103,108,65,108,112,104,97,70,117,110,99,0,103,108,80,105,120,101,108,90,111,111,109,0,103,108,80,105,120,101,108,84,114,97,110,115,102,101,114,102,0,103,108,80,105,120,101,108,84,114,97,110,115,102,101,114,105,0,103,108,80,105,120,101,108,77,97,112,102,118,0,103,108,80,105,120,101,108,77,97,112,117,105,118,0,103,108,80,105,120,101,108,77,97,112,117,115,118,0,103,108,67,111,112,121,80,105,120,101,108,115,0,103,108,68,114,97,119,80,105,120,101,108,115,0,103,108,71,101,116,67,108,105,112,80,108,97,110,101,0,103,108,71,101,116,76,105,103,104,116,102,118,0,103,108,71,101,116,76,105,103,104,116,105,118,0,103,108,71,101,116,77,97,112,100,118,0,103,108,71,101,116,77,97,112,102,118,0,103,108,71,101,116,77,97,112,105,118,0,103,108,71,101,116,77,97,116,101,114,105,97,108,102,118,0,103,108,71,101,116,77,97,116,101,114,105,97,108,105,118,0,103,108,71,101,116,80,105,120,101,108,77,97,112,102,118,0,103,108,71,101,116,80,105,120,101,108,77,97,112,117,105,118,0,103,108,71,101,116,80,105,120,101,108,77,97,112,117,115,118,0,103,108,71,101,116,80,111,108,121,103,111,110,83,116,105,112,112,108,101,0,103,108,71,101,116,84,101,120,69,110,118,102,118,0,103,108,71,101,116,84,101,120,69,110,118,105,118,0,103,108,71,101,116,84,101,120,71,101,110,100,118,0,103,108,71,101,116,84,101,120,71,101,110,102,118,0,103,108,71,101,116,84,101,120,71,101,110,105,118,0,103,108,73,115,76,105,115,116,0,103,108,76,111,97,100,77,97,116,114,105,120,100,0,103,108,77,117,108,116,77,97,116,114,105,120,102,0,103,108,77,117,108,116,77,97,116,114,105,120,100,0,103,108,79,114,116,104,111,0,103,108,80,111,112,77,97,116,114,105,120,0,103,108,80,117,115,104,77,97,116,114,105,120,0,103,108,82,111,116,97,116,101,100,0,103,108,83,99,97,108,101,100,0,103,108,83,99,97,108,101,102,0,103,108,84,114,97,110,115,108,97,116,101,100,0,103,108,84,114,97,110,115,108,97,116,101,102,0,103,108,67,111,112,121,84,101,120,73,109,97,103,101,49,68,0,103,108,67,111,112,121,84,101,120,83,117,98,73,109,97,103,101,49,68,0,103,108,84,101,120,83,117,98,73,109,97,103,101,49,68,0,103,108,65,114,114,97,121,69,108,101,109,101,110,116,0,103,108,68,105,115,97,98,108,101,67,108,105,101,110,116,83,116,97,116,101,0,103,108,69,100,103,101,70,108,97,103,80,111,105,110,116,101,114,0,103,108,73,110,100,101,120,80,111,105,110,116,101,114,0,103,108,73,110,116,101,114,108,101,97,118,101,100,65,114,114,97,121,115,0,103,108,65,114,101,84,101,120,116,117,114,101,115,82,101,115,105,100,101,110,116,0,103,108,80,114,105,111,114,105,116,105,122,101,84,101,120,116,117,114,101,115,0,103,108,73,110,100,101,120,117,98,0,103,108,73,110,100,101,120,117,98,118,0,103,108,80,111,112,67,108,105,101,110,116,65,116,116,114,105,98,0,103,108,80,117,115,104,67,108,105,101,110,116,65,116,116,114,105,98,0,103,108,84,101,120,73,109,97,103,101,51,68,0,103,108,84,101,120,83,117,98,73,109,97,103,101,51,68,0,103,108,67,111,112,121,84,101,120,83,117,98,73,109,97,103,101,51,68,0,103,108,67,111,109,112,114,101,115,115,101,100,84,101,120,73,109,97,103,101,51,68,0,103,108,67,111,109,112,114,101,115,115,101,100,84,101,120,73,109,97,103,101,49,68,0,103,108,67,111,109,112,114,101,115,115,101,100,84,101,120,83,117,98,73,109,97,103,101,51,68,0,103,108,67,111,109,112,114,101,115,115,101,100,84,101,120,83,117,98,73,109,97,103,101,49,68,0,103,108,71,101,116,67,111,109,112,114,101,115,115,101,100,84,101,120,73,109,97,103,101,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,49,100,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,49,100,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,49,102,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,49,102,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,49,105,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,49,105,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,49,115,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,49,115,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,50,100,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,50,100,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,50,102,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,50,102,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,50,105,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,50,105,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,50,115,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,50,115,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,51,100,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,51,100,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,51,102,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,51,102,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,51,105,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,51,105,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,51,115,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,51,115,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,52,100,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,52,100,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,52,102,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,52,102,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,52,105,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,52,105,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,52,115,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,52,115,118,0,103,108,76,111,97,100,84,114,97,110,115,112,111,115,101,77,97,116,114,105,120,102,0,103,108,76,111,97,100,84,114,97,110,115,112,111,115,101,77,97,116,114,105,120,100,0,103,108,77,117,108,116,84,114,97,110,115,112,111,115,101,77,97,116,114,105,120,102,0,103,108,77,117,108,116,84,114,97,110,115,112,111,115,101,77,97,116,114,105,120,100,0,103,108,77,117,108,116,105,68,114,97,119,65,114,114,97,121,115,0,103,108,77,117,108,116,105,68,114,97,119,69,108,101,109,101,110,116,115,0,103,108,80,111,105,110,116,80,97,114,97,109,101,116,101,114,102,0,103,108,80,111,105,110,116,80,97,114,97,109,101,116,101,114,102,118,0,103,108,80,111,105,110,116,80,97,114,97,109,101,116,101,114,105,0,103,108,80,111,105,110,116,80,97,114,97,109,101,116,101,114,105,118,0,103,108,70,111,103,67,111,111,114,100,102,0,103,108,70,111,103,67,111,111,114,100,102,118,0,103,108,70,111,103,67,111,111,114,100,100,0,103,108,70,111,103,67,111,111,114,100,100,118,0,103,108,70,111,103,67,111,111,114,100,80,111,105,110,116,101,114,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,98,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,98,118,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,100,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,100,118,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,102,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,102,118,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,105,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,105,118,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,115,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,115,118,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,117,98,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,117,98,118,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,117,105,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,117,105,118,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,117,115,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,117,115,118,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,80,111,105,110,116,101,114,0,103,108,87,105,110,100,111,119,80,111,115,50,100,0,103,108,87,105,110,100,111,119,80,111,115,50,100,118,0,103,108,87,105,110,100,111,119,80,111,115,50,102,0,103,108,87,105,110,100,111,119,80,111,115,50,102,118,0,103,108,87,105,110,100,111,119,80,111,115,50,105,0,103,108,87,105,110,100,111,119,80,111,115,50,105,118,0,103,108,87,105,110,100,111,119,80,111,115,50,115,0,103,108,87,105,110,100,111,119,80,111,115,50,115,118,0,103,108,87,105,110,100,111,119,80,111,115,51,100,0,103,108,87,105,110,100,111,119,80,111,115,51,100,118,0,103,108,87,105,110,100,111,119,80,111,115,51,102,0,103,108,87,105,110,100,111,119,80,111,115,51,102,118,0,103,108,87,105,110,100,111,119,80,111,115,51,105,0,103,108,87,105,110,100,111,119,80,111,115,51,105,118,0,103,108,87,105,110,100,111,119,80,111,115,51,115,0,103,108,87,105,110,100,111,119,80,111,115,51,115,118,0,103,108,71,101,110,81,117,101,114,105,101,115,0,103,108,68,101,108,101,116,101,81,117,101,114,105,101,115,0,103,108,73,115,81,117,101,114,121,0,103,108,66,101,103,105,110,81,117,101,114,121,0,103,108,69,110,100,81,117,101,114,121,0,103,108,71,101,116,81,117,101,114,121,105,118,0,103,108,71,101,116,81,117,101,114,121,79,98,106,101,99,116,105,118,0,103,108,71,101,116,81,117,101,114,121,79,98,106,101,99,116,117,105,118,0,103,108,71,101,116,66,117,102,102,101,114,83,117,98,68,97,116,97,0,103,108,77,97,112,66,117,102,102,101,114,0,103,108,85,110,109,97,112,66,117,102,102,101,114,0,103,108,71,101,116,66,117,102,102,101,114,80,111,105,110,116,101,114,118,0,103,108,71,101,116,86,101,114,116,101,120,65,116,116,114,105,98,100,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,49,100,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,49,100,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,49,115,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,49,115,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,50,100,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,50,100,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,50,115,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,50,115,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,51,100,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,51,100,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,51,115,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,51,115,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,78,98,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,78,105,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,78,115,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,78,117,98,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,78,117,98,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,78,117,105,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,78,117,115,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,98,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,100,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,100,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,105,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,115,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,115,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,117,98,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,117,105,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,117,115,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,50,120,51,102,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,51,120,50,102,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,50,120,52,102,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,52,120,50,102,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,51,120,52,102,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,52,120,51,102,118,0,103,108,67,111,108,111,114,77,97,115,107,105,0,103,108,71,101,116,66,111,111,108,101,97,110,105,95,118,0,103,108,71,101,116,73,110,116,101,103,101,114,105,95,118,0,103,108,69,110,97,98,108,101,105,0,103,108,68,105,115,97,98,108,101,105,0,103,108,73,115,69,110,97,98,108,101,100,105,0,103,108,66,101,103,105,110,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,0,103,108,69,110,100,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,0,103,108,66,105,110,100,66,117,102,102,101,114,82,97,110,103,101,0,103,108,66,105,110,100,66,117,102,102,101,114,66,97,115,101,0,103,108,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,86,97,114,121,105,110,103,115,0,103,108,71,101,116,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,86,97,114,121,105,110,103,0,103,108,67,108,97,109,112,67,111,108,111,114,0,103,108,66,101,103,105,110,67,111,110,100,105,116,105,111,110,97,108,82,101,110,100,101,114,0,103,108,69,110,100,67,111,110,100,105,116,105,111,110,97,108,82,101,110,100,101,114,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,80,111,105,110,116,101,114,0,103,108,71,101,116,86,101,114,116,101,120,65,116,116,114,105,98,73,105,118,0,103,108,71,101,116,86,101,114,116,101,120,65,116,116,114,105,98,73,117,105,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,49,105,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,50,105,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,51,105,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,52,105,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,49,117,105,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,50,117,105,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,51,117,105,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,52,117,105,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,49,105,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,50,105,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,51,105,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,52,105,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,49,117,105,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,50,117,105,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,51,117,105,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,52,117,105,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,52,98,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,52,115,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,52,117,98,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,52,117,115,118,0,103,108,71,101,116,85,110,105,102,111,114,109,117,105,118,0,103,108,66,105,110,100,70,114,97,103,68,97,116,97,76,111,99,97,116,105,111,110,0,103,108,71,101,116,70,114,97,103,68,97,116,97,76,111,99,97,116,105,111,110,0,103,108,85,110,105,102,111,114,109,49,117,105,0,103,108,85,110,105,102,111,114,109,50,117,105,0,103,108,85,110,105,102,111,114,109,51,117,105,0,103,108,85,110,105,102,111,114,109,52,117,105,0,103,108,85,110,105,102,111,114,109,49,117,105,118,0,103,108,85,110,105,102,111,114,109,50,117,105,118,0,103,108,85,110,105,102,111,114,109,51,117,105,118,0,103,108,85,110,105,102,111,114,109,52,117,105,118,0,103,108,84,101,120,80,97,114,97,109,101,116,101,114,73,105,118,0,103,108,84,101,120,80,97,114,97,109,101,116,101,114,73,117,105,118,0,103,108,71,101,116,84,101,120,80,97,114,97,109,101,116,101,114,73,105,118,0,103,108,71,101,116,84,101,120,80,97,114,97,109,101,116,101,114,73,117,105,118,0,103,108,67,108,101,97,114,66,117,102,102,101,114,105,118,0,103,108,67,108,101,97,114,66,117,102,102,101,114,117,105,118,0,103,108,67,108,101,97,114,66,117,102,102,101,114,102,118,0,103,108,67,108,101,97,114,66,117,102,102,101,114,102,105,0,103,108,71,101,116,83,116,114,105,110,103,105,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,49,68,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,51,68,0,103,108,66,108,105,116,70,114,97,109,101,98,117,102,102,101,114,0,103,108,82,101,110,100,101,114,98,117,102,102,101,114,83,116,111,114,97,103,101,77,117,108,116,105,115,97,109,112,108,101,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,76,97,121,101,114,0,103,108,77,97,112,66,117,102,102,101,114,82,97,110,103,101,0,103,108,70,108,117,115,104,77,97,112,112,101,100,66,117,102,102,101,114,82,97,110,103,101,0,103,108,73,115,86,101,114,116,101,120,65,114,114,97,121,0,103,108,84,101,120,66,117,102,102,101,114,0,103,108,80,114,105,109,105,116,105,118,101,82,101,115,116,97,114,116,73,110,100,101,120,0,103,108,67,111,112,121,66,117,102,102,101,114,83,117,98,68,97,116,97,0,103,108,71,101,116,85,110,105,102,111,114,109,73,110,100,105,99,101,115,0,103,108,71,101,116,65,99,116,105,118,101,85,110,105,102,111,114,109,115,105,118,0,103,108,71,101,116,65,99,116,105,118,101,85,110,105,102,111,114,109,78,97,109,101,0,103,108,71,101,116,85,110,105,102,111,114,109,66,108,111,99,107,73,110,100,101,120,0,103,108,71,101,116,65,99,116,105,118,101,85,110,105,102,111,114,109,66,108,111,99,107,105,118,0,103,108,71,101,116,65,99,116,105,118,101,85,110,105,102,111,114,109,66,108,111,99,107,78,97,109,101,0,103,108,85,110,105,102,111,114,109,66,108,111,99,107,66,105,110,100,105,110,103,0,103,108,68,114,97,119,69,108,101,109,101,110,116,115,66,97,115,101,86,101,114,116,101,120,0,103,108,68,114,97,119,82,97,110,103,101,69,108,101,109,101,110,116,115,66,97,115,101,86,101,114,116,101,120,0,103,108,68,114,97,119,69,108,101,109,101,110,116,115,73,110,115,116,97,110,99,101,100,66,97,115,101,86,101,114,116,101,120,0,103,108,77,117,108,116,105,68,114,97,119,69,108,101,109,101,110,116,115,66,97,115,101,86,101,114,116,101,120,0,103,108,80,114,111,118,111,107,105,110,103,86,101,114,116,101,120,0,103,108,70,101,110,99,101,83,121,110,99,0,103,108,73,115,83,121,110,99,0,103,108,68,101,108,101,116,101,83,121,110,99,0,103,108,67,108,105,101,110,116,87,97,105,116,83,121,110,99,0,103,108,87,97,105,116,83,121,110,99,0,103,108,71,101,116,73,110,116,101,103,101,114,54,52,118,0,103,108,71,101,116,83,121,110,99,105,118,0,103,108,71,101,116,73,110,116,101,103,101,114,54,52,105,95,118,0,103,108,71,101,116,66,117,102,102,101,114,80,97,114,97,109,101,116,101,114,105,54,52,118,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,0,103,108,84,101,120,73,109,97,103,101,50,68,77,117,108,116,105,115,97,109,112,108,101,0,103,108,84,101,120,73,109,97,103,101,51,68,77,117,108,116,105,115,97,109,112,108,101,0,103,108,71,101,116,77,117,108,116,105,115,97,109,112,108,101,102,118,0,103,108,83,97,109,112,108,101,77,97,115,107,105,0,103,108,66,105,110,100,70,114,97,103,68,97,116,97,76,111,99,97,116,105,111,110,73,110,100,101,120,101,100,0,103,108,71,101,116,70,114,97,103,68,97,116,97,73,110,100,101,120,0,103,108,71,101,110,83,97,109,112,108,101,114,115,0,103,108,68,101,108,101,116,101,83,97,109,112,108,101,114,115,0,103,108,73,115,83,97,109,112,108,101,114,0,103,108,66,105,110,100,83,97,109,112,108,101,114,0,103,108,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,105,0,103,108,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,105,118,0,103,108,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,102,0,103,108,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,102,118,0,103,108,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,73,105,118,0,103,108,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,73,117,105,118,0,103,108,71,101,116,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,105,118,0,103,108,71,101,116,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,73,105,118,0,103,108,71,101,116,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,102,118,0,103,108,71,101,116,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,73,117,105,118,0,103,108,81,117,101,114,121,67,111,117,110,116,101,114,0,103,108,71,101,116,81,117,101,114,121,79,98,106,101,99,116,105,54,52,118,0,103,108,71,101,116,81,117,101,114,121,79,98,106,101,99,116,117,105,54,52,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,80,49,117,105,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,80,49,117,105,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,80,50,117,105,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,80,50,117,105,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,80,51,117,105,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,80,51,117,105,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,80,52,117,105,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,80,52,117,105,118,0,103,108,86,101,114,116,101,120,80,50,117,105,0,103,108,86,101,114,116,101,120,80,50,117,105,118,0,103,108,86,101,114,116,101,120,80,51,117,105,0,103,108,86,101,114,116,101,120,80,51,117,105,118,0,103,108,86,101,114,116,101,120,80,52,117,105,0,103,108,86,101,114,116,101,120,80,52,117,105,118,0,103,108,84,101,120,67,111,111,114,100,80,49,117,105,0,103,108,84,101,120,67,111,111,114,100,80,49,117,105,118,0,103,108,84,101,120,67,111,111,114,100,80,50,117,105,0,103,108,84,101,120,67,111,111,114,100,80,50,117,105,118,0,103,108,84,101,120,67,111,111,114,100,80,51,117,105,0,103,108,84,101,120,67,111,111,114,100,80,51,117,105,118,0,103,108,84,101,120,67,111,111,114,100,80,52,117,105,0,103,108,84,101,120,67,111,111,114,100,80,52,117,105,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,80,49,117,105,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,80,49,117,105,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,80,50,117,105,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,80,50,117,105,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,80,51,117,105,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,80,51,117,105,118,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,80,52,117,105,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,80,52,117,105,118,0,103,108,78,111,114,109,97,108,80,51,117,105,0,103,108,78,111,114,109,97,108,80,51,117,105,118,0,103,108,67,111,108,111,114,80,51,117,105,0,103,108,67,111,108,111,114,80,51,117,105,118,0,103,108,67,111,108,111,114,80,52,117,105,0,103,108,67,111,108,111,114,80,52,117,105,118,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,80,51,117,105,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,80,51,117,105,118,0,103,108,77,105,110,83,97,109,112,108,101,83,104,97,100,105,110,103,0,103,108,66,108,101,110,100,69,113,117,97,116,105,111,110,105,0,103,108,66,108,101,110,100,69,113,117,97,116,105,111,110,83,101,112,97,114,97,116,101,105,0,103,108,66,108,101,110,100,70,117,110,99,105,0,103,108,66,108,101,110,100,70,117,110,99,83,101,112,97,114,97,116,101,105,0,103,108,68,114,97,119,65,114,114,97,121,115,73,110,100,105,114,101,99,116,0,103,108,68,114,97,119,69,108,101,109,101,110,116,115,73,110,100,105,114,101,99,116,0,103,108,85,110,105,102,111,114,109,49,100,0,103,108,85,110,105,102,111,114,109,50,100,0,103,108,85,110,105,102,111,114,109,51,100,0,103,108,85,110,105,102,111,114,109,52,100,0,103,108,85,110,105,102,111,114,109,49,100,118,0,103,108,85,110,105,102,111,114,109,50,100,118,0,103,108,85,110,105,102,111,114,109,51,100,118,0,103,108,85,110,105,102,111,114,109,52,100,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,50,100,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,51,100,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,52,100,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,50,120,51,100,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,50,120,52,100,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,51,120,50,100,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,51,120,52,100,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,52,120,50,100,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,52,120,51,100,118,0,103,108,71,101,116,85,110,105,102,111,114,109,100,118,0,103,108,71,101,116,83,117,98,114,111,117,116,105,110,101,85,110,105,102,111,114,109,76,111,99,97,116,105,111,110,0,103,108,71,101,116,83,117,98,114,111,117,116,105,110,101,73,110,100,101,120,0,103,108,71,101,116,65,99,116,105,118,101,83,117,98,114,111,117,116,105,110,101,85,110,105,102,111,114,109,105,118,0,103,108,71,101,116,65,99,116,105,118,101,83,117,98,114,111,117,116,105,110,101,85,110,105,102,111,114,109,78,97,109,101,0,103,108,71,101,116,65,99,116,105,118,101,83,117,98,114,111,117,116,105,110,101,78,97,109,101,0,103,108,85,110,105,102,111,114,109,83,117,98,114,111,117,116,105,110,101,115,117,105,118,0,103,108,71,101,116,85,110,105,102,111,114,109,83,117,98,114,111,117,116,105,110,101,117,105,118,0,103,108,71,101,116,80,114,111,103,114,97,109,83,116,97,103,101,105,118,0,103,108,80,97,116,99,104,80,97,114,97,109,101,116,101,114,105,0,103,108,80,97,116,99,104,80,97,114,97,109,101,116,101,114,102,118,0,103,108,66,105,110,100,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,0,103,108,68,101,108,101,116,101,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,115,0,103,108,71,101,110,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,115,0,103,108,73,115,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,0,103,108,80,97,117,115,101,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,0,103,108,82,101,115,117,109,101,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,0,103,108,68,114,97,119,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,0,103,108,68,114,97,119,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,83,116,114,101,97,109,0,103,108,66,101,103,105,110,81,117,101,114,121,73,110,100,101,120,101,100,0,103,108,69,110,100,81,117,101,114,121,73,110,100,101,120,101,100,0,103,108,71,101,116,81,117,101,114,121,73,110,100,101,120,101,100,105,118,0,103,108,71,101,116,80,114,111,103,114,97,109,66,105,110,97,114,121,0,103,108,80,114,111,103,114,97,109,66,105,110,97,114,121,0,103,108,80,114,111,103,114,97,109,80,97,114,97,109,101,116,101,114,105,0,103,108,85,115,101,80,114,111,103,114,97,109,83,116,97,103,101,115,0,103,108,65,99,116,105,118,101,83,104,97,100,101,114,80,114,111,103,114,97,109,0,103,108,67,114,101,97,116,101,83,104,97,100,101,114,80,114,111,103,114,97,109,118,0,103,108,66,105,110,100,80,114,111,103,114,97,109,80,105,112,101,108,105,110,101,0,103,108,68,101,108,101,116,101,80,114,111,103,114,97,109,80,105,112,101,108,105,110,101,115,0,103,108,71,101,110,80,114,111,103,114,97,109,80,105,112,101,108,105,110,101,115,0,103,108,73,115,80,114,111,103,114,97,109,80,105,112,101,108,105,110,101,0,103,108,71,101,116,80,114,111,103,114,97,109,80,105,112,101,108,105,110,101,105,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,105,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,105,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,102,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,102,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,100,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,100,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,117,105,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,117,105,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,105,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,105,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,102,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,102,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,100,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,100,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,117,105,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,117,105,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,105,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,105,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,102,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,102,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,100,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,100,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,117,105,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,117,105,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,105,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,105,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,102,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,102,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,100,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,100,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,117,105,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,117,105,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,50,102,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,51,102,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,52,102,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,50,100,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,51,100,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,52,100,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,50,120,51,102,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,51,120,50,102,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,50,120,52,102,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,52,120,50,102,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,51,120,52,102,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,52,120,51,102,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,50,120,51,100,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,51,120,50,100,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,50,120,52,100,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,52,120,50,100,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,51,120,52,100,118,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,52,120,51,100,118,0,103,108,86,97,108,105,100,97,116,101,80,114,111,103,114,97,109,80,105,112,101,108,105,110,101,0,103,108,71,101,116,80,114,111,103,114,97,109,80,105,112,101,108,105,110,101,73,110,102,111,76,111,103,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,49,100,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,50,100,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,51,100,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,52,100,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,49,100,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,50,100,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,51,100,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,52,100,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,80,111,105,110,116,101,114,0,103,108,71,101,116,86,101,114,116,101,120,65,116,116,114,105,98,76,100,118,0,103,108,86,105,101,119,112,111,114,116,65,114,114,97,121,118,0,103,108,86,105,101,119,112,111,114,116,73,110,100,101,120,101,100,102,0,103,108,86,105,101,119,112,111,114,116,73,110,100,101,120,101,100,102,118,0,103,108,83,99,105,115,115,111,114,65,114,114,97,121,118,0,103,108,83,99,105,115,115,111,114,73,110,100,101,120,101,100,0,103,108,83,99,105,115,115,111,114,73,110,100,101,120,101,100,118,0,103,108,68,101,112,116,104,82,97,110,103,101,65,114,114,97,121,118,0,103,108,68,101,112,116,104,82,97,110,103,101,73,110,100,101,120,101,100,0,103,108,71,101,116,70,108,111,97,116,105,95,118,0,103,108,71,101,116,68,111,117,98,108,101,105,95,118,0,103,108,68,114,97,119,65,114,114,97,121,115,73,110,115,116,97,110,99,101,100,66,97,115,101,73,110,115,116,97,110,99,101,0,103,108,68,114,97,119,69,108,101,109,101,110,116,115,73,110,115,116,97,110,99,101,100,66,97,115,101,73,110,115,116,97,110,99,101,0,103,108,68,114,97,119,69,108,101,109,101,110,116,115,73,110,115,116,97,110,99,101,100,66,97,115,101,86,101,114,116,101,120,66,97,115,101,73,110,115,116,97,110,99,101,0,103,108,71,101,116,73,110,116,101,114,110,97,108,102,111,114,109,97,116,105,118,0,103,108,71,101,116,65,99,116,105,118,101,65,116,111,109,105,99,67,111,117,110,116,101,114,66,117,102,102,101,114,105,118,0,103,108,66,105,110,100,73,109,97,103,101,84,101,120,116,117,114,101,0,103,108,77,101,109,111,114,121,66,97,114,114,105,101,114,0,103,108,84,101,120,83,116,111,114,97,103,101,49,68,0,103,108,84,101,120,83,116,111,114,97,103,101,50,68,0,103,108,84,101,120,83,116,111,114,97,103,101,51,68,0,103,108,68,114,97,119,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,73,110,115,116,97,110,99,101,100,0,103,108,68,114,97,119,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,83,116,114,101,97,109,73,110,115,116,97,110,99,101,100,0,103,108,67,108,101,97,114,66,117,102,102,101,114,68,97,116,97,0,103,108,67,108,101,97,114,66,117,102,102,101,114,83,117,98,68,97,116,97,0,103,108,68,105,115,112,97,116,99,104,67,111,109,112,117,116,101,0,103,108,68,105,115,112,97,116,99,104,67,111,109,112,117,116,101,73,110,100,105,114,101,99,116,0,103,108,67,111,112,121,73,109,97,103,101,83,117,98,68,97,116,97,0,103,108,70,114,97,109,101,98,117,102,102,101,114,80,97,114,97,109,101,116,101,114,105,0,103,108,71,101,116,70,114,97,109,101,98,117,102,102,101,114,80,97,114,97,109,101,116,101,114,105,118,0,103,108,71,101,116,73,110,116,101,114,110,97,108,102,111,114,109,97,116,105,54,52,118,0,103,108,73,110,118,97,108,105,100,97,116,101,84,101,120,83,117,98,73,109,97,103,101,0,103,108,73,110,118,97,108,105,100,97,116,101,84,101,120,73,109,97,103,101,0,103,108,73,110,118,97,108,105,100,97,116,101,66,117,102,102,101,114,83,117,98,68,97,116,97,0,103,108,73,110,118,97,108,105,100,97,116,101,66,117,102,102,101,114,68,97,116,97,0,103,108,73,110,118,97,108,105,100,97,116,101,70,114,97,109,101,98,117,102,102,101,114,0,103,108,73,110,118,97,108,105,100,97,116,101,83,117,98,70,114,97,109,101,98,117,102,102,101,114,0,103,108,77,117,108,116,105,68,114,97,119,65,114,114,97,121,115,73,110,100,105,114,101,99,116,0,103,108,77,117,108,116,105,68,114,97,119,69,108,101,109,101,110,116,115,73,110,100,105,114,101,99,116,0,103,108,71,101,116,80,114,111,103,114,97,109,73,110,116,101,114,102,97,99,101,105,118,0,103,108,71,101,116,80,114,111,103,114,97,109,82,101,115,111,117,114,99,101,73,110,100,101,120,0,103,108,71,101,116,80,114,111,103,114,97,109,82,101,115,111,117,114,99,101,78,97,109,101,0,103,108,71,101,116,80,114,111,103,114,97,109,82,101,115,111,117,114,99,101,105,118,0,103,108,71,101,116,80,114,111,103,114,97,109,82,101,115,111,117,114,99,101,76,111,99,97,116,105,111,110,0,103,108,71,101,116,80,114,111,103,114,97,109,82,101,115,111,117,114,99,101,76,111,99,97,116,105,111,110,73,110,100,101,120,0,103,108,83,104,97,100,101,114,83,116,111,114,97,103,101,66,108,111,99,107,66,105,110,100,105,110,103,0,103,108,84,101,120,66,117,102,102,101,114,82,97,110,103,101,0,103,108,84,101,120,83,116,111,114,97,103,101,50,68,77,117,108,116,105,115,97,109,112,108,101,0,103,108,84,101,120,83,116,111,114,97,103,101,51,68,77,117,108,116,105,115,97,109,112,108,101,0,103,108,84,101,120,116,117,114,101,86,105,101,119,0,103,108,66,105,110,100,86,101,114,116,101,120,66,117,102,102,101,114,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,70,111,114,109,97,116,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,70,111,114,109,97,116,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,70,111,114,109,97,116,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,66,105,110,100,105,110,103,0,103,108,86,101,114,116,101,120,66,105,110,100,105,110,103,68,105,118,105,115,111,114,0,103,108,68,101,98,117,103,77,101,115,115,97,103,101,67,111,110,116,114,111,108,0,103,108,68,101,98,117,103,77,101,115,115,97,103,101,73,110,115,101,114,116,0,103,108,68,101,98,117,103,77,101,115,115,97,103,101,67,97,108,108,98,97,99,107,0,103,108,71,101,116,68,101,98,117,103,77,101,115,115,97,103,101,76,111,103,0,103,108,80,117,115,104,68,101,98,117,103,71,114,111,117,112,0,103,108,80,111,112,68,101,98,117,103,71,114,111,117,112,0,103,108,79,98,106,101,99,116,76,97,98,101,108,0,103,108,71,101,116,79,98,106,101,99,116,76,97,98,101,108,0,103,108,79,98,106,101,99,116,80,116,114,76,97,98,101,108,0,103,108,71,101,116,79,98,106,101,99,116,80,116,114,76,97,98,101,108,0,103,108,66,117,102,102,101,114,83,116,111,114,97,103,101,0,103,108,67,108,101,97,114,84,101,120,73,109,97,103,101,0,103,108,67,108,101,97,114,84,101,120,83,117,98,73,109,97,103,101,0,103,108,66,105,110,100,66,117,102,102,101,114,115,66,97,115,101,0,103,108,66,105,110,100,66,117,102,102,101,114,115,82,97,110,103,101,0,103,108,66,105,110,100,84,101,120,116,117,114,101,115,0,103,108,66,105,110,100,83,97,109,112,108,101,114,115,0,103,108,66,105,110,100,73,109,97,103,101,84,101,120,116,117,114,101,115,0,103,108,66,105,110,100,86,101,114,116,101,120,66,117,102,102,101,114,115,0,103,108,67,108,105,112,67,111,110,116,114,111,108,0,103,108,67,114,101,97,116,101,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,115,0,103,108,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,66,117,102,102,101,114,66,97,115,101,0,103,108,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,66,117,102,102,101,114,82,97,110,103,101,0,103,108,71,101,116,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,105,118,0,103,108,71,101,116,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,105,95,118,0,103,108,71,101,116,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,105,54,52,95,118,0,103,108,67,114,101,97,116,101,66,117,102,102,101,114,115,0,103,108,78,97,109,101,100,66,117,102,102,101,114,83,116,111,114,97,103,101,0,103,108,78,97,109,101,100,66,117,102,102,101,114,68,97,116,97,0,103,108,78,97,109,101,100,66,117,102,102,101,114,83,117,98,68,97,116,97,0,103,108,67,111,112,121,78,97,109,101,100,66,117,102,102,101,114,83,117,98,68,97,116,97,0,103,108,67,108,101,97,114,78,97,109,101,100,66,117,102,102,101,114,68,97,116,97,0,103,108,67,108,101,97,114,78,97,109,101,100,66,117,102,102,101,114,83,117,98,68,97,116,97,0,103,108,77,97,112,78,97,109,101,100,66,117,102,102,101,114,0,103,108,77,97,112,78,97,109,101,100,66,117,102,102,101,114,82,97,110,103,101,0,103,108,85,110,109,97,112,78,97,109,101,100,66,117,102,102,101,114,0,103,108,70,108,117,115,104,77,97,112,112,101,100,78,97,109,101,100,66,117,102,102,101,114,82,97,110,103,101,0,103,108,71,101,116,78,97,109,101,100,66,117,102,102,101,114,80,97,114,97,109,101,116,101,114,105,118,0,103,108,71,101,116,78,97,109,101,100,66,117,102,102,101,114,80,97,114,97,109,101,116,101,114,105,54,52,118,0,103,108,71,101,116,78,97,109,101,100,66,117,102,102,101,114,80,111,105,110,116,101,114,118,0,103,108,71,101,116,78,97,109,101,100,66,117,102,102,101,114,83,117,98,68,97,116,97,0,103,108,67,114,101,97,116,101,70,114,97,109,101,98,117,102,102,101,114,115,0,103,108,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,82,101,110,100,101,114,98,117,102,102,101,114,0,103,108,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,80,97,114,97,109,101,116,101,114,105,0,103,108,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,0,103,108,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,76,97,121,101,114,0,103,108,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,68,114,97,119,66,117,102,102,101,114,0,103,108,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,68,114,97,119,66,117,102,102,101,114,115,0,103,108,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,82,101,97,100,66,117,102,102,101,114,0,103,108,73,110,118,97,108,105,100,97,116,101,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,68,97,116,97,0,103,108,73,110,118,97,108,105,100,97,116,101,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,83,117,98,68,97,116,97,0,103,108,67,108,101,97,114,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,105,118,0,103,108,67,108,101,97,114,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,117,105,118,0,103,108,67,108,101,97,114,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,102,118,0,103,108,67,108,101,97,114,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,102,105,0,103,108,66,108,105,116,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,0,103,108,67,104,101,99,107,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,83,116,97,116,117,115,0,103,108,71,101,116,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,80,97,114,97,109,101,116,101,114,105,118,0,103,108,71,101,116,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,65,116,116,97,99,104,109,101,110,116,80,97,114,97,109,101,116,101,114,105,118,0,103,108,67,114,101,97,116,101,82,101,110,100,101,114,98,117,102,102,101,114,115,0,103,108,78,97,109,101,100,82,101,110,100,101,114,98,117,102,102,101,114,83,116,111,114,97,103,101,0,103,108,78,97,109,101,100,82,101,110,100,101,114,98,117,102,102,101,114,83,116,111,114,97,103,101,77,117,108,116,105,115,97,109,112,108,101,0,103,108,71,101,116,78,97,109,101,100,82,101,110,100,101,114,98,117,102,102,101,114,80,97,114,97,109,101,116,101,114,105,118,0,103,108,67,114,101,97,116,101,84,101,120,116,117,114,101,115,0,103,108,84,101,120,116,117,114,101,66,117,102,102,101,114,0,103,108,84,101,120,116,117,114,101,66,117,102,102,101,114,82,97,110,103,101,0,103,108,84,101,120,116,117,114,101,83,116,111,114,97,103,101,49,68,0,103,108,84,101,120,116,117,114,101,83,116,111,114,97,103,101,50,68,0,103,108,84,101,120,116,117,114,101,83,116,111,114,97,103,101,51,68,0,103,108,84,101,120,116,117,114,101,83,116,111,114,97,103,101,50,68,77,117,108,116,105,115,97,109,112,108,101,0,103,108,84,101,120,116,117,114,101,83,116,111,114,97,103,101,51,68,77,117,108,116,105,115,97,109,112,108,101,0,103,108,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,49,68,0,103,108,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,50,68,0,103,108,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,51,68,0,103,108,67,111,109,112,114,101,115,115,101,100,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,49,68,0,103,108,67,111,109,112,114,101,115,115,101,100,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,50,68,0,103,108,67,111,109,112,114,101,115,115,101,100,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,51,68,0,103,108,67,111,112,121,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,49,68,0,103,108,67,111,112,121,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,50,68,0,103,108,67,111,112,121,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,51,68,0,103,108,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,102,0,103,108,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,102,118,0,103,108,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,105,0,103,108,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,73,105,118,0,103,108,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,73,117,105,118,0,103,108,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,105,118,0,103,108,71,101,110,101,114,97,116,101,84,101,120,116,117,114,101,77,105,112,109,97,112,0,103,108,66,105,110,100,84,101,120,116,117,114,101,85,110,105,116,0,103,108,71,101,116,84,101,120,116,117,114,101,73,109,97,103,101,0,103,108,71,101,116,67,111,109,112,114,101,115,115,101,100,84,101,120,116,117,114,101,73,109,97,103,101,0,103,108,71,101,116,84,101,120,116,117,114,101,76,101,118,101,108,80,97,114,97,109,101,116,101,114,102,118,0,103,108,71,101,116,84,101,120,116,117,114,101,76,101,118,101,108,80,97,114,97,109,101,116,101,114,105,118,0,103,108,71,101,116,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,102,118,0,103,108,71,101,116,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,73,105,118,0,103,108,71,101,116,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,73,117,105,118,0,103,108,71,101,116,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,105,118,0,103,108,67,114,101,97,116,101,86,101,114,116,101,120,65,114,114,97,121,115,0,103,108,68,105,115,97,98,108,101,86,101,114,116,101,120,65,114,114,97,121,65,116,116,114,105,98,0,103,108,69,110,97,98,108,101,86,101,114,116,101,120,65,114,114,97,121,65,116,116,114,105,98,0,103,108,86,101,114,116,101,120,65,114,114,97,121,69,108,101,109,101,110,116,66,117,102,102,101,114,0,103,108,86,101,114,116,101,120,65,114,114,97,121,86,101,114,116,101,120,66,117,102,102,101,114,0,103,108,86,101,114,116,101,120,65,114,114,97,121,86,101,114,116,101,120,66,117,102,102,101,114,115,0,103,108,86,101,114,116,101,120,65,114,114,97,121,65,116,116,114,105,98,66,105,110,100,105,110,103,0,103,108,86,101,114,116,101,120,65,114,114,97,121,65,116,116,114,105,98,70,111,114,109,97,116,0,103,108,86,101,114,116,101,120,65,114,114,97,121,65,116,116,114,105,98,73,70,111,114,109,97,116,0,103,108,86,101,114,116,101,120,65,114,114,97,121,65,116,116,114,105,98,76,70,111,114,109,97,116,0,103,108,86,101,114,116,101,120,65,114,114,97,121,66,105,110,100,105,110,103,68,105,118,105,115,111,114,0,103,108,71,101,116,86,101,114,116,101,120,65,114,114,97,121,105,118,0,103,108,71,101,116,86,101,114,116,101,120,65,114,114,97,121,73,110,100,101,120,101,100,105,118,0,103,108,71,101,116,86,101,114,116,101,120,65,114,114,97,121,73,110,100,101,120,101,100,54,52,105,118,0,103,108,67,114,101,97,116,101,83,97,109,112,108,101,114,115,0,103,108,67,114,101,97,116,101,80,114,111,103,114,97,109,80,105,112,101,108,105,110,101,115,0,103,108,67,114,101,97,116,101,81,117,101,114,105,101,115,0,103,108,71,101,116,81,117,101,114,121,66,117,102,102,101,114,79,98,106,101,99,116,105,54,52,118,0,103,108,71,101,116,81,117,101,114,121,66,117,102,102,101,114,79,98,106,101,99,116,105,118,0,103,108,71,101,116,81,117,101,114,121,66,117,102,102,101,114,79,98,106,101,99,116,117,105,54,52,118,0,103,108,71,101,116,81,117,101,114,121,66,117,102,102,101,114,79,98,106,101,99,116,117,105,118,0,103,108,77,101,109,111,114,121,66,97,114,114,105,101,114,66,121,82,101,103,105,111,110,0,103,108,71,101,116,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,0,103,108,71,101,116,67,111,109,112,114,101,115,115,101,100,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,0,103,108,71,101,116,71,114,97,112,104,105,99,115,82,101,115,101,116,83,116,97,116,117,115,0,103,108,71,101,116,110,67,111,109,112,114,101,115,115,101,100,84,101,120,73,109,97,103,101,0,103,108,71,101,116,110,84,101,120,73,109,97,103,101,0,103,108,71,101,116,110,85,110,105,102,111,114,109,100,118,0,103,108,71,101,116,110,85,110,105,102,111,114,109,102,118,0,103,108,71,101,116,110,85,110,105,102,111,114,109,105,118,0,103,108,71,101,116,110,85,110,105,102,111,114,109,117,105,118,0,103,108,82,101,97,100,110,80,105,120,101,108,115,0,103,108,71,101,116,110,77,97,112,100,118,0,103,108,71,101,116,110,77,97,112,102,118,0,103,108,71,101,116,110,77,97,112,105,118,0,103,108,71,101,116,110,80,105,120,101,108,77,97,112,102,118,0,103,108,71,101,116,110,80,105,120,101,108,77,97,112,117,105,118,0,103,108,71,101,116,110,80,105,120,101,108,77,97,112,117,115,118,0,103,108,71,101,116,110,80,111,108,121,103,111,110,83,116,105,112,112,108,101,0,103,108,71,101,116,110,67,111,108,111,114,84,97,98,108,101,0,103,108,71,101,116,110,67,111,110,118,111,108,117,116,105,111,110,70,105,108,116,101,114,0,103,108,71,101,116,110,83,101,112,97,114,97,98,108,101,70,105,108,116,101,114,0,103,108,71,101,116,110,72,105,115,116,111,103,114,97,109,0,103,108,71,101,116,110,77,105,110,109,97,120,0,103,108,84,101,120,116,117,114,101,66,97,114,114,105,101,114,0,103,108,66,108,101,110,100,66,97,114,114,105,101,114,0,103,108,80,114,105,109,105,116,105,118,101,66,111,117,110,100,105,110,103,66,111,120,0,71,76,95,75,72,82,95,98,108,101,110,100,95,101,113,117,97,116,105,111,110,95,97,100,118,97,110,99,101,100,0,71,76,95,75,72,82,95,98,108,101,110,100,95,101,113,117,97,116,105,111,110,95,97,100,118,97,110,99,101,100,95,99,111,104,101,114,101,110,116,0,71,76,95,75,72,82,95,99,111,110,116,101,120,116,95,102,108,117,115,104,95,99,111,110,116,114,111,108,0,71,76,95,75,72,82,95,100,101,98,117,103,0,71,76,95,75,72,82,95,110,111,95,101,114,114,111,114,0,71,76,95,75,72,82,95,114,111,98,117,115,116,95,98,117,102,102,101,114,95,97,99,99,101,115,115,95,98,101,104,97,118,105,111,114,0,71,76,95,75,72,82,95,114,111,98,117,115,116,110,101,115,115,0,71,76,95,75,72,82,95,116,101,120,116,117,114,101,95,99,111,109,112,114,101,115,115,105,111,110,95,97,115,116,99,95,104,100,114,0,71,76,95,75,72,82,95,116,101,120,116,117,114,101,95,99,111,109,112,114,101,115,115,105,111,110,95,97,115,116,99,95,108,100,114,0,71,76,95,75,72,82,95,116,101,120,116,117,114,101,95,99,111,109,112,114,101,115,115,105,111,110,95,97,115,116,99,95,115,108,105,99,101,100,95,51,100,0,71,76,95,65,82,66,95,97,114,114,97,121,115,95,111,102,95,97,114,114,97,121,115,0,71,76,95,65,82,66,95,98,97,115,101,95,105,110,115,116,97,110,99,101,0,71,76,95,65,82,66,95,98,105,110,100,108,101,115,115,95,116,101,120,116,117,114,101,0,71,76,95,65,82,66,95,98,108,101,110,100,95,102,117,110,99,95,101,120,116,101,110,100,101,100,0,71,76,95,65,82,66,95,98,117,102,102,101,114,95,115,116,111,114,97,103,101,0,71,76,95,65,82,66,95,99,108,95,101,118,101,110,116,0,71,76,95,65,82,66,95,99,108,101,97,114,95,98,117,102,102,101,114,95,111,98,106,101,99,116,0,71,76,95,65,82,66,95,99,108,101,97,114,95,116,101,120,116,117,114,101,0,71,76,95,65,82,66,95,99,108,105,112,95,99,111,110,116,114,111,108,0,71,76,95,65,82,66,95,99,111,108,111,114,95,98,117,102,102,101,114,95,102,108,111,97,116,0,71,76,95,65,82,66,95,99,111,109,112,97,116,105,98,105,108,105,116,121,0,71,76,95,65,82,66,95,99,111,109,112,114,101,115,115,101,100,95,116,101,120,116,117,114,101,95,112,105,120,101,108,95,115,116,111,114,97,103,101,0,71,76,95,65,82,66,95,99,111,109,112,117,116,101,95,115,104,97,100,101,114,0,71,76,95,65,82,66,95,99,111,109,112,117,116,101,95,118,97,114,105,97,98,108,101,95,103,114,111,117,112,95,115,105,122,101,0,71,76,95,65,82,66,95,99,111,110,100,105,116,105,111,110,97,108,95,114,101,110,100,101,114,95,105,110,118,101,114,116,101,100,0,71,76,95,65,82,66,95,99,111,110,115,101,114,118,97,116,105,118,101,95,100,101,112,116,104,0,71,76,95,65,82,66,95,99,111,112,121,95,98,117,102,102,101,114,0,71,76,95,65,82,66,95,99,111,112,121,95,105,109,97,103,101,0,71,76,95,65,82,66,95,99,117,108,108,95,100,105,115,116,97,110,99,101,0,71,76,95,65,82,66,95,100,101,98,117,103,95,111,117,116,112,117,116,0,71,76,95,65,82,66,95,100,101,112,116,104,95,98,117,102,102,101,114,95,102,108,111,97,116,0,71,76,95,65,82,66,95,100,101,112,116,104,95,99,108,97,109,112,0,71,76,95,65,82,66,95,100,101,114,105,118,97,116,105,118,101,95,99,111,110,116,114,111,108,0,71,76,95,65,82,66,95,100,105,114,101,99,116,95,115,116,97,116,101,95,97,99,99,101,115,115,0,71,76,95,65,82,66,95,100,114,97,119,95,98,117,102,102,101,114,115,95,98,108,101,110,100,0,71,76,95,65,82,66,95,100,114,97,119,95,101,108,101,109,101,110,116,115,95,98,97,115,101,95,118,101,114,116,101,120,0,71,76,95,65,82,66,95,100,114,97,119,95,105,110,100,105,114,101,99,116,0,71,76,95,65,82,66,95,100,114,97,119,95,105,110,115,116,97,110,99,101,100,0,71,76,95,65,82,66,95,101,110,104,97,110,99,101,100,95,108,97,121,111,117,116,115,0,71,76,95,65,82,66,95,69,83,50,95,99,111,109,112,97,116,105,98,105,108,105,116,121,0,71,76,95,65,82,66,95,69,83,51,95,49,95,99,111,109,112,97,116,105,98,105,108,105,116,121,0,71,76,95,65,82,66,95,69,83,51,95,50,95,99,111,109,112,97,116,105,98,105,108,105,116,121,0,71,76,95,65,82,66,95,69,83,51,95,99,111,109,112,97,116,105,98,105,108,105,116,121,0,71,76,95,65,82,66,95,101,120,112,108,105,99,105,116,95,97,116,116,114,105,98,95,108,111,99,97,116,105,111,110,0,71,76,95,65,82,66,95,101,120,112,108,105,99,105,116,95,117,110,105,102,111,114,109,95,108,111,99,97,116,105,111,110,0,71,76,95,65,82,66,95,102,114,97,103,109,101,110,116,95,99,111,111,114,100,95,99,111,110,118,101,110,116,105,111,110,115,0,71,76,95,65,82,66,95,102,114,97,103,109,101,110,116,95,108,97,121,101,114,95,118,105,101,119,112,111,114,116,0,71,76,95,65,82,66,95,102,114,97,103,109,101,110,116,95,115,104,97,100,101,114,95,105,110,116,101,114,108,111,99,107,0,71,76,95,65,82,66,95,102,114,97,109,101,98,117,102,102,101,114,95,110,111,95,97,116,116,97,99,104,109,101,110,116,115,0,71,76,95,65,82,66,95,102,114,97,109,101,98,117,102,102,101,114,95,111,98,106,101,99,116,0,71,76,95,65,82,66,95,102,114,97,109,101,98,117,102,102,101,114,95,115,82,71,66,0,71,76,95,65,82,66,95,103,101,111,109,101,116,114,121,95,115,104,97,100,101,114,52,0,71,76,95,65,82,66,95,103,101,116,95,112,114,111,103,114,97,109,95,98,105,110,97,114,121,0,71,76,95,65,82,66,95,103,101,116,95,116,101,120,116,117,114,101,95,115,117,98,95,105,109,97,103,101,0,71,76,95,65,82,66,95,103,112,117,95,115,104,97,100,101,114,53,0,71,76,95,65,82,66,95,103,112,117,95,115,104,97,100,101,114,95,102,112,54,52,0,71,76,95,65,82,66,95,103,112,117,95,115,104,97,100,101,114,95,105,110,116,54,52,0,71,76,95,65,82,66,95,104,97,108,102,95,102,108,111,97,116,95,112,105,120,101,108,0,71,76,95,65,82,66,95,104,97,108,102,95,102,108,111,97,116,95,118,101,114,116,101,120,0,71,76,95,65,82,66,95,105,110,100,105,114,101,99,116,95,112,97,114,97,109,101,116,101,114,115,0,71,76,95,65,82,66,95,105,110,115,116,97,110,99,101,100,95,97,114,114,97,121,115,0,71,76,95,65,82,66,95,105,110,116,101,114,110,97,108,102,111,114,109,97,116,95,113,117,101,114,121,0,71,76,95,65,82,66,95,105,110,116,101,114,110,97,108,102,111,114,109,97,116,95,113,117,101,114,121,50,0,71,76,95,65,82,66,95,105,110,118,97,108,105,100,97,116,101,95,115,117,98,100,97,116,97,0,71,76,95,65,82,66,95,109,97,112,95,98,117,102,102,101,114,95,97,108,105,103,110,109,101,110,116,0,71,76,95,65,82,66,95,109,97,112,95,98,117,102,102,101,114,95,114,97,110,103,101,0,71,76,95,65,82,66,95,109,117,108,116,105,95,98,105,110,100,0,71,76,95,65,82,66,95,109,117,108,116,105,95,100,114,97,119,95,105,110,100,105,114,101,99,116,0,71,76,95,65,82,66,95,111,99,99,108,117,115,105,111,110,95,113,117,101,114,121,50,0,71,76,95,65,82,66,95,112,97,114,97,108,108,101,108,95,115,104,97,100,101,114,95,99,111,109,112,105,108,101,0,71,76,95,65,82,66,95,112,105,112,101,108,105,110,101,95,115,116,97,116,105,115,116,105,99,115,95,113,117,101,114,121,0,71,76,95,65,82,66,95,112,105,120,101,108,95,98,117,102,102,101,114,95,111,98,106,101,99,116,0,71,76,95,65,82,66,95,112,111,115,116,95,100,101,112,116,104,95,99,111,118,101,114,97,103,101,0,71,76,95,65,82,66,95,112,114,111,103,114,97,109,95,105,110,116,101,114,102,97,99,101,95,113,117,101,114,121,0,71,76,95,65,82,66,95,112,114,111,118,111,107,105,110,103,95,118,101,114,116,101,120,0,71,76,95,65,82,66,95,113,117,101,114,121,95,98,117,102,102,101,114,95,111,98,106,101,99,116,0,71,76,95,65,82,66,95,114,111,98,117,115,116,95,98,117,102,102,101,114,95,97,99,99,101,115,115,95,98,101,104,97,118,105,111,114,0,71,76,95,65,82,66,95,114,111,98,117,115,116,110,101,115,115,0,71,76,95,65,82,66,95,114,111,98,117,115,116,110,101,115,115,95,105,115,111,108,97,116,105,111,110,0,71,76,95,65,82,66,95,115,97,109,112,108,101,95,108,111,99,97,116,105,111,110,115,0,71,76,95,65,82,66,95,115,97,109,112,108,101,95,115,104,97,100,105,110,103,0,71,76,95,65,82,66,95,115,97,109,112,108,101,114,95,111,98,106,101,99,116,115,0,71,76,95,65,82,66,95,115,101,97,109,108,101,115,115,95,99,117,98,101,95,109,97,112,0,71,76,95,65,82,66,95,115,101,97,109,108,101,115,115,95,99,117,98,101,109,97,112,95,112,101,114,95,116,101,120,116,117,114,101,0,71,76,95,65,82,66,95,115,101,112,97,114,97,116,101,95,115,104,97,100,101,114,95,111,98,106,101,99,116,115,0,71,76,95,65,82,66,95,115,104,97,100,101,114,95,97,116,111,109,105,99,95,99,111,117,110,116,101,114,95,111,112,115,0,71,76,95,65,82,66,95,115,104,97,100,101,114,95,97,116,111,109,105,99,95,99,111,117,110,116,101,114,115,0,71,76,95,65,82,66,95,115,104,97,100,101,114,95,98,97,108,108,111,116,0,71,76,95,65,82,66,95,115,104,97,100,101,114,95,98,105,116,95,101,110,99,111,100,105,110,103,0,71,76,95,65,82,66,95,115,104,97,100,101,114,95,99,108,111,99,107,0,71,76,95,65,82,66,95,115,104,97,100,101,114,95,100,114,97,119,95,112,97,114,97,109,101,116,101,114,115,0,71,76,95,65,82,66,95,115,104,97,100,101,114,95,103,114,111,117,112,95,118,111,116,101,0,71,76,95,65,82,66,95,115,104,97,100,101,114,95,105,109,97,103,101,95,108,111,97,100,95,115,116,111,114,101,0,71,76,95,65,82,66,95,115,104,97,100,101,114,95,105,109,97,103,101,95,115,105,122,101,0,71,76,95,65,82,66,95,115,104,97,100,101,114,95,112,114,101,99,105,115,105,111,110,0,71,76,95,65,82,66,95,115,104,97,100,101,114,95,115,116,101,110,99,105,108,95,101,120,112,111,114,116,0,71,76,95,65,82,66,95,115,104,97,100,101,114,95,115,116,111,114,97,103,101,95,98,117,102,102,101,114,95,111,98,106,101,99,116,0,71,76,95,65,82,66,95,115,104,97,100,101,114,95,115,117,98,114,111,117,116,105,110,101,0,71,76,95,65,82,66,95,115,104,97,100,101,114,95,116,101,120,116,117,114,101,95,105,109,97,103,101,95,115,97,109,112,108,101,115,0,71,76,95,65,82,66,95,115,104,97,100,101,114,95,116,101,120,116,117,114,101,95,108,111,100,0,71,76,95,65,82,66,95,115,104,97,100,101,114,95,118,105,101,119,112,111,114,116,95,108,97,121,101,114,95,97,114,114,97,121,0,71,76,95,65,82,66,95,115,104,97,100,105,110,103,95,108,97,110,103,117,97,103,101,95,52,50,48,112,97,99,107,0,71,76,95,65,82,66,95,115,104,97,100,105,110,103,95,108,97,110,103,117,97,103,101,95,105,110,99,108,117,100,101,0,71,76,95,65,82,66,95,115,104,97,100,105,110,103,95,108,97,110,103,117,97,103,101,95,112,97,99,107,105,110,103,0,71,76,95,65,82,66,95,115,112,97,114,115,101,95,98,117,102,102,101,114,0,71,76,95,65,82,66,95,115,112,97,114,115,101,95,116,101,120,116,117,114,101,0,71,76,95,65,82,66,95,115,112,97,114,115,101,95,116,101,120,116,117,114,101,50,0,71,76,95,65,82,66,95,115,112,97,114,115,101,95,116,101,120,116,117,114,101,95,99,108,97,109,112,0,71,76,95,65,82,66,95,115,116,101,110,99,105,108,95,116,101,120,116,117,114,105,110,103,0,71,76,95,65,82,66,95,115,121,110,99,0,71,76,95,65,82,66,95,116,101,115,115,101,108,108,97,116,105,111,110,95,115,104,97,100,101,114,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,98,97,114,114,105,101,114,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,98,117,102,102,101,114,95,111,98,106,101,99,116,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,98,117,102,102,101,114,95,111,98,106,101,99,116,95,114,103,98,51,50,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,98,117,102,102,101,114,95,114,97,110,103,101,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,99,111,109,112,114,101,115,115,105,111,110,95,98,112,116,99,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,99,111,109,112,114,101,115,115,105,111,110,95,114,103,116,99,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,99,117,98,101,95,109,97,112,95,97,114,114,97,121,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,102,105,108,116,101,114,95,109,105,110,109,97,120,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,102,108,111,97,116,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,103,97,116,104,101,114,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,109,105,114,114,111,114,95,99,108,97,109,112,95,116,111,95,101,100,103,101,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,109,117,108,116,105,115,97,109,112,108,101,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,113,117,101,114,121,95,108,101,118,101,108,115,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,113,117,101,114,121,95,108,111,100,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,114,101,99,116,97,110,103,108,101,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,114,103,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,114,103,98,49,48,95,97,50,117,105,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,115,116,101,110,99,105,108,56,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,115,116,111,114,97,103,101,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,115,116,111,114,97,103,101,95,109,117,108,116,105,115,97,109,112,108,101,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,115,119,105,122,122,108,101,0,71,76,95,65,82,66,95,116,101,120,116,117,114,101,95,118,105,101,119,0,71,76,95,65,82,66,95,116,105,109,101,114,95,113,117,101,114,121,0,71,76,95,65,82,66,95,116,114,97,110,115,102,111,114,109,95,102,101,101,100,98,97,99,107,50,0,71,76,95,65,82,66,95,116,114,97,110,115,102,111,114,109,95,102,101,101,100,98,97,99,107,51,0,71,76,95,65,82,66,95,116,114,97,110,115,102,111,114,109,95,102,101,101,100,98,97,99,107,95,105,110,115,116,97,110,99,101,100,0,71,76,95,65,82,66,95,116,114,97,110,115,102,111,114,109,95,102,101,101,100,98,97,99,107,95,111,118,101,114,102,108,111,119,95,113,117,101,114,121,0,71,76,95,65,82,66,95,117,110,105,102,111,114,109,95,98,117,102,102,101,114,95,111,98,106,101,99,116,0,71,76,95,65,82,66,95,118,101,114,116,101,120,95,97,114,114,97,121,95,98,103,114,97,0,71,76,95,65,82,66,95,118,101,114,116,101,120,95,97,114,114,97,121,95,111,98,106,101,99,116,0,71,76,95,65,82,66,95,118,101,114,116,101,120,95,97,116,116,114,105,98,95,54,52,98,105,116,0,71,76,95,65,82,66,95,118,101,114,116,101,120,95,97,116,116,114,105,98,95,98,105,110,100,105,110,103,0,71,76,95,65,82,66,95,118,101,114,116,101,120,95,116,121,112,101,95,49,48,102,95,49,49,102,95,49,49,102,95,114,101,118,0,71,76,95,65,82,66,95,118,101,114,116,101,120,95,116,121,112,101,95,50,95,49,48,95,49,48,95,49,48,95,114,101,118,0,71,76,95,65,82,66,95,118,105,101,119,112,111,114,116,95,97,114,114,97,121,0,71,76,95,69,88,84,95,98,97,115,101,95,105,110,115,116,97,110,99,101,0,71,76,95,69,88,84,95,98,103,114,97,0,71,76,95,69,88,84,95,98,105,110,100,97,98,108,101,95,117,110,105,102,111,114,109,0,71,76,95,69,88,84,95,98,108,101,110,100,95,102,117,110,99,95,101,120,116,101,110,100,101,100,0,71,76,95,69,88,84,95,98,108,101,110,100,95,109,105,110,109,97,120,0,71,76,95,69,88,84,95,98,117,102,102,101,114,95,115,116,111,114,97,103,101,0,71,76,95,69,88,84,95,99,111,108,111,114,95,98,117,102,102,101,114,95,102,108,111,97,116,0,71,76,95,69,88,84,95,99,111,108,111,114,95,98,117,102,102,101,114,95,104,97,108,102,95,102,108,111,97,116,0,71,76,95,69,88,84,95,99,111,112,121,95,105,109,97,103,101,0,71,76,95,69,88,84,95,100,101,98,117,103,95,108,97,98,101,108,0,71,76,95,69,88,84,95,100,101,98,117,103,95,109,97,114,107,101,114,0,71,76,95,69,88,84,95,100,105,114,101,99,116,95,115,116,97,116,101,95,97,99,99,101,115,115,0,71,76,95,69,88,84,95,100,105,115,99,97,114,100,95,102,114,97,109,101,98,117,102,102,101,114,0,71,76,95,69,88,84,95,100,105,115,106,111,105,110,116,95,116,105,109,101,114,95,113,117,101,114,121,0,71,76,95,69,88,84,95,100,114,97,119,95,98,117,102,102,101,114,115,0,71,76,95,69,88,84,95,100,114,97,119,95,98,117,102,102,101,114,115,50,0,71,76,95,69,88,84,95,100,114,97,119,95,98,117,102,102,101,114,115,95,105,110,100,101,120,101,100,0,71,76,95,69,88,84,95,100,114,97,119,95,101,108,101,109,101,110,116,115,95,98,97,115,101,95,118,101,114,116,101,120,0,71,76,95,69,88,84,95,100,114,97,119,95,105,110,115,116,97,110,99,101,100,0,71,76,95,69,88,84,95,102,108,111,97,116,95,98,108,101,110,100,0,71,76,95,69,88,84,95,102,114,97,109,101,98,117,102,102,101,114,95,98,108,105,116,0,71,76,95,69,88,84,95,102,114,97,109,101,98,117,102,102,101,114,95,109,117,108,116,105,115,97,109,112,108,101,0,71,76,95,69,88,84,95,102,114,97,109,101,98,117,102,102,101,114,95,109,117,108,116,105,115,97,109,112,108,101,95,98,108,105,116,95,115,99,97,108,101,100,0,71,76,95,69,88,84,95,102,114,97,109,101,98,117,102,102,101,114,95,111,98,106,101,99,116,0,71,76,95,69,88,84,95,102,114,97,109,101,98,117,102,102,101,114,95,115,82,71,66,0,71,76,95,69,88,84,95,103,101,111,109,101,116,114,121,95,112,111,105,110,116,95,115,105,122,101,0,71,76,95,69,88,84,95,103,101,111,109,101,116,114,121,95,115,104,97,100,101,114,0,71,76,95,69,88,84,95,103,101,111,109,101,116,114,121,95,115,104,97,100,101,114,52,0,71,76,95,69,88,84,95,103,112,117,95,112,114,111,103,114,97,109,95,112,97,114,97,109,101,116,101,114,115,0,71,76,95,69,88,84,95,103,112,117,95,115,104,97,100,101,114,52,0,71,76,95,69,88,84,95,103,112,117,95,115,104,97,100,101,114,53,0,71,76,95,69,88,84,95,105,110,115,116,97,110,99,101,100,95,97,114,114,97,121,115,0,71,76,95,69,88,84,95,109,97,112,95,98,117,102,102,101,114,95,114,97,110,103,101,0,71,76,95,69,88,84,95,109,117,108,116,105,95,100,114,97,119,95,105,110,100,105,114,101,99,116,0,71,76,95,69,88,84,95,109,117,108,116,105,115,97,109,112,108,101,100,95,99,111,109,112,97,116,105,98,105,108,105,116,121,0,71,76,95,69,88,84,95,109,117,108,116,105,115,97,109,112,108,101,100,95,114,101,110,100,101,114,95,116,111,95,116,101,120,116,117,114,101,0,71,76,95,69,88,84,95,109,117,108,116,105,118,105,101,119,95,100,114,97,119,95,98,117,102,102,101,114,115,0,71,76,95,69,88,84,95,111,99,99,108,117,115,105,111,110,95,113,117,101,114,121,95,98,111,111,108,101,97,110,0,71,76,95,69,88,84,95,112,97,99,107,101,100,95,100,101,112,116,104,95,115,116,101,110,99,105,108,0,71,76,95,69,88,84,95,112,97,99,107,101,100,95,102,108,111,97,116,0,71,76,95,69,88,84,95,112,105,120,101,108,95,98,117,102,102,101,114,95,111,98,106,101,99,116,0,71,76,95,69,88,84,95,112,111,108,121,103,111,110,95,111,102,102,115,101,116,95,99,108,97,109,112,0,71,76,95,69,88,84,95,112,111,115,116,95,100,101,112,116,104,95,99,111,118,101,114,97,103,101,0,71,76,95,69,88,84,95,112,114,105,109,105,116,105,118,101,95,98,111,117,110,100,105,110,103,95,98,111,120,0,71,76,95,69,88,84,95,112,114,111,118,111,107,105,110,103,95,118,101,114,116,101,120,0,71,76,95,69,88,84,95,112,118,114,116,99,95,115,82,71,66,0,71,76,95,69,88,84,95,114,97,115,116,101,114,95,109,117,108,116,105,115,97,109,112,108,101,0,71,76,95,69,88,84,95,114,101,97,100,95,102,111,114,109,97,116,95,98,103,114,97,0,71,76,95,69,88,84,95,114,101,110,100,101,114,95,115,110,111,114,109,0,71,76,95,69,88,84,95,114,111,98,117,115,116,110,101,115,115,0,71,76,95,69,88,84,95,115,101,112,97,114,97,116,101,95,115,104,97,100,101,114,95,111,98,106,101,99,116,115,0,71,76,95,69,88,84,95,115,104,97,100,101,114,95,102,114,97,109,101,98,117,102,102,101,114,95,102,101,116,99,104,0,71,76,95,69,88,84,95,115,104,97,100,101,114,95,105,109,97,103,101,95,108,111,97,100,95,102,111,114,109,97,116,116,101,100,0,71,76,95,69,88,84,95,115,104,97,100,101,114,95,105,109,97,103,101,95,108,111,97,100,95,115,116,111,114,101,0,71,76,95,69,88,84,95,115,104,97,100,101,114,95,105,109,112,108,105,99,105,116,95,99,111,110,118,101,114,115,105,111,110,115,0,71,76,95,69,88,84,95,115,104,97,100,101,114,95,105,110,116,101,103,101,114,95,109,105,120,0,71,76,95,69,88,84,95,115,104,97,100,101,114,95,105,111,95,98,108,111,99,107,115,0,71,76,95,69,88,84,95,115,104,97,100,101,114,95,112,105,120,101,108,95,108,111,99,97,108,95,115,116,111,114,97,103,101,0,71,76,95,69,88,84,95,115,104,97,100,101,114,95,116,101,120,116,117,114,101,95,108,111,100,0,71,76,95,69,88,84,95,115,104,97,100,111,119,95,115,97,109,112,108,101,114,115,0,71,76,95,69,88,84,95,115,112,97,114,115,101,95,116,101,120,116,117,114,101,0,71,76,95,69,88,84,95,115,112,97,114,115,101,95,116,101,120,116,117,114,101,50,0,71,76,95,69,88,84,95,115,82,71,66,0,71,76,95,69,88,84,95,115,82,71,66,95,119,114,105,116,101,95,99,111,110,116,114,111,108,0,71,76,95,69,88,84,95,115,116,101,110,99,105,108,95,99,108,101,97,114,95,116,97,103,0,71,76,95,69,88,84,95,116,101,115,115,101,108,108,97,116,105,111,110,95,112,111,105,110,116,95,115,105,122,101,0,71,76,95,69,88,84,95,116,101,115,115,101,108,108,97,116,105,111,110,95,115,104,97,100,101,114,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,97,114,114,97,121,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,98,111,114,100,101,114,95,99,108,97,109,112,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,98,117,102,102,101,114,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,98,117,102,102,101,114,95,111,98,106,101,99,116,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,99,111,109,112,114,101,115,115,105,111,110,95,100,120,116,49,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,99,111,109,112,114,101,115,115,105,111,110,95,108,97,116,99,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,99,111,109,112,114,101,115,115,105,111,110,95,114,103,116,99,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,99,111,109,112,114,101,115,115,105,111,110,95,115,51,116,99,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,99,117,98,101,95,109,97,112,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,99,117,98,101,95,109,97,112,95,97,114,114,97,121,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,102,105,108,116,101,114,95,97,110,105,115,111,116,114,111,112,105,99,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,102,105,108,116,101,114,95,109,105,110,109,97,120,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,102,111,114,109,97,116,95,66,71,82,65,56,56,56,56,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,105,110,116,101,103,101,114,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,110,111,114,109,49,54,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,114,103,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,115,104,97,114,101,100,95,101,120,112,111,110,101,110,116,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,115,110,111,114,109,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,115,82,71,66,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,115,82,71,66,95,100,101,99,111,100,101,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,115,82,71,66,95,82,56,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,115,82,71,66,95,82,71,56,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,115,116,111,114,97,103,101,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,115,119,105,122,122,108,101,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,116,121,112,101,95,50,95,49,48,95,49,48,95,49,48,95,82,69,86,0,71,76,95,69,88,84,95,116,101,120,116,117,114,101,95,118,105,101,119,0,71,76,95,69,88,84,95,116,105,109,101,114,95,113,117,101,114,121,0,71,76,95,69,88,84,95,116,114,97,110,115,102,111,114,109,95,102,101,101,100,98,97,99,107,0,71,76,95,69,88,84,95,117,110,112,97,99,107,95,115,117,98,105,109,97,103,101,0,71,76,95,69,88,84,95,118,101,114,116,101,120,95,97,114,114,97,121,95,98,103,114,97,0,71,76,95,69,88,84,95,118,101,114,116,101,120,95,97,116,116,114,105,98,95,54,52,98,105,116,0,71,76,95,69,88,84,95,120,49,49,95,115,121,110,99,95,111,98,106,101,99,116,0,71,76,95,69,88,84,95,89,85,86,95,116,97,114,103,101,116,0,71,76,95,79,69,83,95,99,111,109,112,114,101,115,115,101,100,95,69,84,67,49,95,82,71,66,56,95,115,117,98,95,116,101,120,116,117,114,101,0,71,76,95,79,69,83,95,99,111,109,112,114,101,115,115,101,100,95,69,84,67,49,95,82,71,66,56,95,116,101,120,116,117,114,101,0,71,76,95,79,69,83,95,99,111,112,121,95,105,109,97,103,101,0,71,76,95,79,69,83,95,100,101,112,116,104,50,52,0,71,76,95,79,69,83,95,100,101,112,116,104,51,50,0,71,76,95,79,69,83,95,100,101,112,116,104,95,116,101,120,116,117,114,101,0,71,76,95,79,69,83,95,100,114,97,119,95,98,117,102,102,101,114,115,95,105,110,100,101,120,101,100,0,71,76,95,79,69,83,95,100,114,97,119,95,101,108,101,109,101,110,116,115,95,98,97,115,101,95,118,101,114,116,101,120,0,71,76,95,79,69,83,95,69,71,76,95,105,109,97,103,101,0,71,76,95,79,69,83,95,69,71,76,95,105,109,97,103,101,95,101,120,116,101,114,110,97,108,0,71,76,95,79,69,83,95,69,71,76,95,105,109,97,103,101,95,101,120,116,101,114,110,97,108,95,101,115,115,108,51,0,71,76,95,79,69,83,95,101,108,101,109,101,110,116,95,105,110,100,101,120,95,117,105,110,116,0,71,76,95,79,69,83,95,102,98,111,95,114,101,110,100,101,114,95,109,105,112,109,97,112,0,71,76,95,79,69,83,95,102,114,97,103,109,101,110,116,95,112,114,101,99,105,115,105,111,110,95,104,105,103,104,0,71,76,95,79,69,83,95,103,101,111,109,101,116,114,121,95,112,111,105,110,116,95,115,105,122,101,0,71,76,95,79,69,83,95,103,101,111,109,101,116,114,121,95,115,104,97,100,101,114,0,71,76,95,79,69,83,95,103,101,116,95,112,114,111,103,114,97,109,95,98,105,110,97,114,121,0,71,76,95,79,69,83,95,103,112,117,95,115,104,97,100,101,114,53,0,71,76,95,79,69,83,95,109,97,112,98,117,102,102,101,114,0,71,76,95,79,69,83,95,112,97,99,107,101,100,95,100,101,112,116,104,95,115,116,101,110,99,105,108,0,71,76,95,79,69,83,95,112,114,105,109,105,116,105,118,101,95,98,111,117,110,100,105,110,103,95,98,111,120,0,71,76,95,79,69,83,95,114,101,113,117,105,114,101,100,95,105,110,116,101,114,110,97,108,102,111,114,109,97,116,0,71,76,95,79,69,83,95,114,103,98,56,95,114,103,98,97,56,0,71,76,95,79,69,83,95,115,97,109,112,108,101,95,115,104,97,100,105,110,103,0,71,76,95,79,69,83,95,115,97,109,112,108,101,95,118,97,114,105,97,98,108,101,115,0,71,76,95,79,69,83,95,115,104,97,100,101,114,95,105,109,97,103,101,95,97,116,111,109,105,99,0,71,76,95,79,69,83,95,115,104,97,100,101,114,95,105,111,95,98,108,111,99,107,115,0,71,76,95,79,69,83,95,115,104,97,100,101,114,95,109,117,108,116,105,115,97,109,112,108,101,95,105,110,116,101,114,112,111,108,97,116,105,111,110,0,71,76,95,79,69,83,95,115,116,97,110,100,97,114,100,95,100,101,114,105,118,97,116,105,118,101,115,0,71,76,95,79,69,83,95,115,116,101,110,99,105,108,49,0,71,76,95,79,69,83,95,115,116,101,110,99,105,108,52,0,71,76,95,79,69,83,95,115,117,114,102,97,99,101,108,101,115,115,95,99,111,110,116,101,120,116,0,71,76,95,79,69,83,95,116,101,115,115,101,108,108,97,116,105,111,110,95,112,111,105,110,116,95,115,105,122,101,0,71,76,95,79,69,83,95,116,101,115,115,101,108,108,97,116,105,111,110,95,115,104,97,100,101,114,0,71,76,95,79,69,83,95,116,101,120,116,117,114,101,95,51,68,0,71,76,95,79,69,83,95,116,101,120,116,117,114,101,95,98,111,114,100,101,114,95,99,108,97,109,112,0,71,76,95,79,69,83,95,116,101,120,116,117,114,101,95,98,117,102,102,101,114,0,71,76,95,79,69,83,95,116,101,120,116,117,114,101,95,99,111,109,112,114,101,115,115,105,111,110,95,97,115,116,99,0,71,76,95,79,69,83,95,116,101,120,116,117,114,101,95,99,117,98,101,95,109,97,112,95,97,114,114,97,121,0,71,76,95,79,69,83,95,116,101,120,116,117,114,101,95,102,108,111,97,116,0,71,76,95,79,69,83,95,116,101,120,116,117,114,101,95,102,108,111,97,116,95,108,105,110,101,97,114,0,71,76,95,79,69,83,95,116,101,120,116,117,114,101,95,104,97,108,102,95,102,108,111,97,116,0,71,76,95,79,69,83,95,116,101,120,116,117,114,101,95,104,97,108,102,95,102,108,111,97,116,95,108,105,110,101,97,114,0,71,76,95,79,69,83,95,116,101,120,116,117,114,101,95,110,112,111,116,0,71,76,95,79,69,83,95,116,101,120,116,117,114,101,95,115,116,101,110,99,105,108,56,0,71,76,95,79,69,83,95,116,101,120,116,117,114,101,95,115,116,111,114,97,103,101,95,109,117,108,116,105,115,97,109,112,108,101,95,50,100,95,97,114,114,97,121,0,71,76,95,79,69,83,95,116,101,120,116,117,114,101,95,118,105,101,119,0,71,76,95,79,69,83,95,118,101,114,116,101,120,95,97,114,114,97,121,95,111,98,106,101,99,116,0,71,76,95,79,69,83,95,118,101,114,116,101,120,95,104,97,108,102,95,102,108,111,97,116,0,71,76,95,79,69,83,95,118,101,114,116,101,120,95,116,121,112,101,95,49,48,95,49,48,95,49,48,95,50,0,71,76,95,65,77,68,95,98,108,101,110,100,95,109,105,110,109,97,120,95,102,97,99,116,111,114,0,71,76,95,65,77,68,95,99,111,109,112,114,101,115,115,101,100,95,51,68,67,95,116,101,120,116,117,114,101,0,71,76,95,65,77,68,95,99,111,109,112,114,101,115,115,101,100,95,65,84,67,95,116,101,120,116,117,114,101,0,71,76,95,65,77,68,95,99,111,110,115,101,114,118,97,116,105,118,101,95,100,101,112,116,104,0,71,76,95,65,77,68,95,100,101,98,117,103,95,111,117,116,112,117,116,0,71,76,95,65,77,68,95,100,101,112,116,104,95,99,108,97,109,112,95,115,101,112,97,114,97,116,101,0,71,76,95,65,77,68,95,100,114,97,119,95,98,117,102,102,101,114,115,95,98,108,101,110,100,0,71,76,95,65,77,68,95,103,99,110,95,115,104,97,100,101,114,0,71,76,95,65,77,68,95,103,112,117,95,115,104,97,100,101,114,95,105,110,116,54,52,0,71,76,95,65,77,68,95,105,110,116,101,114,108,101,97,118,101,100,95,101,108,101,109,101,110,116,115,0,71,76,95,65,77,68,95,109,117,108,116,105,95,100,114,97,119,95,105,110,100,105,114,101,99,116,0,71,76,95,65,77,68,95,110,97,109,101,95,103,101,110,95,100,101,108,101,116,101,0,71,76,95,65,77,68,95,111,99,99,108,117,115,105,111,110,95,113,117,101,114,121,95,101,118,101,110,116,0,71,76,95,65,77,68,95,112,101,114,102,111,114,109,97,110,99,101,95,109,111,110,105,116,111,114,0,71,76,95,65,77,68,95,112,105,110,110,101,100,95,109,101,109,111,114,121,0,71,76,95,65,77,68,95,112,114,111,103,114,97,109,95,98,105,110,97,114,121,95,90,52,48,48,0,71,76,95,65,77,68,95,113,117,101,114,121,95,98,117,102,102,101,114,95,111,98,106,101,99,116,0,71,76,95,65,77,68,95,115,97,109,112,108,101,95,112,111,115,105,116,105,111,110,115,0,71,76,95,65,77,68,95,115,101,97,109,108,101,115,115,95,99,117,98,101,109,97,112,95,112,101,114,95,116,101,120,116,117,114,101,0,71,76,95,65,77,68,95,115,104,97,100,101,114,95,97,116,111,109,105,99,95,99,111,117,110,116,101,114,95,111,112,115,0,71,76,95,65,77,68,95,115,104,97,100,101,114,95,115,116,101,110,99,105,108,95,101,120,112,111,114,116,0,71,76,95,65,77,68,95,115,104,97,100,101,114,95,116,114,105,110,97,114,121,95,109,105,110,109,97,120,0,71,76,95,65,77,68,95,115,112,97,114,115,101,95,116,101,120,116,117,114,101,0,71,76,95,65,77,68,95,115,116,101,110,99,105,108,95,111,112,101,114,97,116,105,111,110,95,101,120,116,101,110,100,101,100,0,71,76,95,65,77,68,95,116,101,120,116,117,114,101,95,116,101,120,116,117,114,101,52,0,71,76,95,65,77,68,95,116,114,97,110,115,102,111,114,109,95,102,101,101,100,98,97,99,107,51,95,108,105,110,101,115,95,116,114,105,97,110,103,108,101,115,0,71,76,95,65,77,68,95,116,114,97,110,115,102,111,114,109,95,102,101,101,100,98,97,99,107,52,0,71,76,95,65,77,68,95,118,101,114,116,101,120,95,115,104,97,100,101,114,95,108,97,121,101,114,0,71,76,95,65,77,68,95,118,101,114,116,101,120,95,115,104,97,100,101,114,95,116,101,115,115,101,108,108,97,116,111,114,0,71,76,95,65,77,68,95,118,101,114,116,101,120,95,115,104,97,100,101,114,95,118,105,101,119,112,111,114,116,95,105,110,100,101,120,0,71,76,95,65,78,68,82,79,73,68,95,101,120,116,101,110,115,105,111,110,95,112,97,99,107,95,101,115,51,49,97,0,71,76,95,65,78,71,76,69,95,100,101,112,116,104,95,116,101,120,116,117,114,101,0,71,76,95,65,78,71,76,69,95,102,114,97,109,101,98,117,102,102,101,114,95,98,108,105,116,0,71,76,95,65,78,71,76,69,95,102,114,97,109,101,98,117,102,102,101,114,95,109,117,108,116,105,115,97,109,112,108,101,0,71,76,95,65,78,71,76,69,95,105,110,115,116,97,110,99,101,100,95,97,114,114,97,121,115,0,71,76,95,65,78,71,76,69,95,112,97,99,107,95,114,101,118,101,114,115,101,95,114,111,119,95,111,114,100,101,114,0,71,76,95,65,78,71,76,69,95,112,114,111,103,114,97,109,95,98,105,110,97,114,121,0,71,76,95,65,78,71,76,69,95,116,101,120,116,117,114,101,95,99,111,109,112,114,101,115,115,105,111,110,95,100,120,116,51,0,71,76,95,65,78,71,76,69,95,116,101,120,116,117,114,101,95,99,111,109,112,114,101,115,115,105,111,110,95,100,120,116,53,0,71,76,95,65,78,71,76,69,95,116,101,120,116,117,114,101,95,117,115,97,103,101,0,71,76,95,65,78,71,76,69,95,116,114,97,110,115,108,97,116,101,100,95,115,104,97,100,101,114,95,115,111,117,114,99,101,0,71,76,95,65,80,80,76,69,95,97,117,120,95,100,101,112,116,104,95,115,116,101,110,99,105,108,0,71,76,95,65,80,80,76,69,95,99,108,105,101,110,116,95,115,116,111,114,97,103,101,0,71,76,95,65,80,80,76,69,95,99,108,105,112,95,100,105,115,116,97,110,99,101,0,71,76,95,65,80,80,76,69,95,99,111,108,111,114,95,98,117,102,102,101,114,95,112,97,99,107,101,100,95,102,108,111,97,116,0,71,76,95,65,80,80,76,69,95,99,111,112,121,95,116,101,120,116,117,114,101,95,108,101,118,101,108,115,0,71,76,95,65,80,80,76,69,95,101,108,101,109,101,110,116,95,97,114,114,97,121,0,71,76,95,65,80,80,76,69,95,102,101,110,99,101,0,71,76,95,65,80,80,76,69,95,102,108,111,97,116,95,112,105,120,101,108,115,0,71,76,95,65,80,80,76,69,95,102,108,117,115,104,95,98,117,102,102,101,114,95,114,97,110,103,101,0,71,76,95,65,80,80,76,69,95,102,114,97,109,101,98,117,102,102,101,114,95,109,117,108,116,105,115,97,109,112,108,101,0,71,76,95,65,80,80,76,69,95,111,98,106,101,99,116,95,112,117,114,103,101,97,98,108,101,0,71,76,95,65,80,80,76,69,95,114,103,98,95,52,50,50,0,71,76,95,65,80,80,76,69,95,114,111,119,95,98,121,116,101,115,0,71,76,95,65,80,80,76,69,95,115,121,110,99,0,71,76,95,65,80,80,76,69,95,116,101,120,116,117,114,101,95,102,111,114,109,97,116,95,66,71,82,65,56,56,56,56,0,71,76,95,65,80,80,76,69,95,116,101,120,116,117,114,101,95,109,97,120,95,108,101,118,101,108,0,71,76,95,65,80,80,76,69,95,116,101,120,116,117,114,101,95,112,97,99,107,101,100,95,102,108,111,97,116,0,71,76,95,65,80,80,76,69,95,116,101,120,116,117,114,101,95,114,97,110,103,101,0,71,76,95,65,80,80,76,69,95,118,101,114,116,101,120,95,97,114,114,97,121,95,111,98,106,101,99,116,0,71,76,95,65,80,80,76,69,95,118,101,114,116,101,120,95,97,114,114,97,121,95,114,97,110,103,101,0,71,76,95,65,80,80,76,69,95,118,101,114,116,101,120,95,112,114,111,103,114,97,109,95,101,118,97,108,117,97,116,111,114,115,0,71,76,95,65,80,80,76,69,95,121,99,98,99,114,95,52,50,50,0,71,76,95,65,82,77,95,109,97,108,105,95,112,114,111,103,114,97,109,95,98,105,110,97,114,121,0,71,76,95,65,82,77,95,109,97,108,105,95,115,104,97,100,101,114,95,98,105,110,97,114,121,0,71,76,95,65,82,77,95,114,103,98,97,56,0,71,76,95,65,82,77,95,115,104,97,100,101,114,95,102,114,97,109,101,98,117,102,102,101,114,95,102,101,116,99,104,0,71,76,95,65,82,77,95,115,104,97,100,101,114,95,102,114,97,109,101,98,117,102,102,101,114,95,102,101,116,99,104,95,100,101,112,116,104,95,115,116,101,110,99,105,108,0,71,76,95,65,84,73,95,109,101,109,105,110,102,111,0,71,76,95,65,84,73,95,112,105,120,101,108,95,102,111,114,109,97,116,95,102,108,111,97,116,0,71,76,95,68,77,80,95,112,114,111,103,114,97,109,95,98,105,110,97,114,121,0,71,76,95,68,77,80,95,115,104,97,100,101,114,95,98,105,110,97,114,121,0,71,76,95,70,74,95,115,104,97,100,101,114,95,98,105,110,97,114,121,95,71,67,67,83,79,0,71,76,95,71,82,69,77,69,68,89,95,102,114,97,109,101,95,116,101,114,109,105,110,97,116,111,114,0,71,76,95,71,82,69,77,69,68,89,95,115,116,114,105,110,103,95,109,97,114,107,101,114,0,71,76,95,73,77,71,95,109,117,108,116,105,115,97,109,112,108,101,100,95,114,101,110,100,101,114,95,116,111,95,116,101,120,116,117,114,101,0,71,76,95,73,77,71,95,112,114,111,103,114,97,109,95,98,105,110,97,114,121,0,71,76,95,73,77,71,95,114,101,97,100,95,102,111,114,109,97,116,0,71,76,95,73,77,71,95,115,104,97,100,101,114,95,98,105,110,97,114,121,0,71,76,95,73,77,71,95,116,101,120,116,117,114,101,95,99,111,109,112,114,101,115,115,105,111,110,95,112,118,114,116,99,0,71,76,95,73,77,71,95,116,101,120,116,117,114,101,95,99,111,109,112,114,101,115,115,105,111,110,95,112,118,114,116,99,50,0,71,76,95,73,77,71,95,116,101,120,116,117,114,101,95,102,105,108,116,101,114,95,99,117,98,105,99,0,71,76,95,73,78,71,82,95,98,108,101,110,100,95,102,117,110,99,95,115,101,112,97,114,97,116,101,0,71,76,95,73,78,84,69,76,95,102,114,97,103,109,101,110,116,95,115,104,97,100,101,114,95,111,114,100,101,114,105,110,103,0,71,76,95,73,78,84,69,76,95,102,114,97,109,101,98,117,102,102,101,114,95,67,77,65,65,0,71,76,95,73,78,84,69,76,95,109,97,112,95,116,101,120,116,117,114,101,0,71,76,95,73,78,84,69,76,95,112,101,114,102,111,114,109,97,110,99,101,95,113,117,101,114,121,0,71,76,95,77,69,83,65,95,112,97,99,107,95,105,110,118,101,114,116,0,71,76,95,77,69,83,65,95,121,99,98,99,114,95,116,101,120,116,117,114,101,0,71,76,95,77,69,83,65,88,95,116,101,120,116,117,114,101,95,115,116,97,99,107,0,71,76,95,78,86,95,98,105,110,100,108,101,115,115,95,109,117,108,116,105,95,100,114,97,119,95,105,110,100,105,114,101,99,116,0,71,76,95,78,86,95,98,105,110,100,108,101,115,115,95,109,117,108,116,105,95,100,114,97,119,95,105,110,100,105,114,101,99,116,95,99,111,117,110,116,0,71,76,95,78,86,95,98,105,110,100,108,101,115,115,95,116,101,120,116,117,114,101,0,71,76,95,78,86,95,98,108,101,110,100,95,101,113,117,97,116,105,111,110,95,97,100,118,97,110,99,101,100,0,71,76,95,78,86,95,98,108,101,110,100,95,101,113,117,97,116,105,111,110,95,97,100,118,97,110,99,101,100,95,99,111,104,101,114,101,110,116,0,71,76,95,78,86,95,99,111,109,109,97,110,100,95,108,105,115,116,0,71,76,95,78,86,95,99,111,109,112,117,116,101,95,112,114,111,103,114,97,109,53,0,71,76,95,78,86,95,99,111,110,100,105,116,105,111,110,97,108,95,114,101,110,100,101,114,0,71,76,95,78,86,95,99,111,110,115,101,114,118,97,116,105,118,101,95,114,97,115,116,101,114,0,71,76,95,78,86,95,99,111,110,115,101,114,118,97,116,105,118,101,95,114,97,115,116,101,114,95,100,105,108,97,116,101,0,71,76,95,78,86,95,99,111,112,121,95,98,117,102,102,101,114,0,71,76,95,78,86,95,99,111,112,121,95,105,109,97,103,101,0,71,76,95,78,86,95,99,111,118,101,114,97,103,101,95,115,97,109,112,108,101,0,71,76,95,78,86,95,100,101,101,112,95,116,101,120,116,117,114,101,51,68,0,71,76,95,78,86,95,100,101,112,116,104,95,98,117,102,102,101,114,95,102,108,111,97,116,0,71,76,95,78,86,95,100,101,112,116,104,95,99,108,97,109,112,0,71,76,95,78,86,95,100,101,112,116,104,95,110,111,110,108,105,110,101,97,114,0,71,76,95,78,86,95,100,114,97,119,95,98,117,102,102,101,114,115,0,71,76,95,78,86,95,100,114,97,119,95,105,110,115,116,97,110,99,101,100,0,71,76,95,78,86,95,100,114,97,119,95,116,101,120,116,117,114,101,0,71,76,95,78,86,95,101,120,112,108,105,99,105,116,95,97,116,116,114,105,98,95,108,111,99,97,116,105,111,110,0,71,76,95,78,86,95,101,120,112,108,105,99,105,116,95,109,117,108,116,105,115,97,109,112,108,101,0,71,76,95,78,86,95,102,98,111,95,99,111,108,111,114,95,97,116,116,97,99,104,109,101,110,116,115,0,71,76,95,78,86,95,102,101,110,99,101,0,71,76,95,78,86,95,102,105,108,108,95,114,101,99,116,97,110,103,108,101,0,71,76,95,78,86,95,102,108,111,97,116,95,98,117,102,102,101,114,0,71,76,95,78,86,95,102,114,97,103,109,101,110,116,95,99,111,118,101,114,97,103,101,95,116,111,95,99,111,108,111,114,0,71,76,95,78,86,95,102,114,97,103,109,101,110,116,95,112,114,111,103,114,97,109,52,0,71,76,95,78,86,95,102,114,97,103,109,101,110,116,95,115,104,97,100,101,114,95,105,110,116,101,114,108,111,99,107,0,71,76,95,78,86,95,102,114,97,109,101,98,117,102,102,101,114,95,98,108,105,116,0,71,76,95,78,86,95,102,114,97,109,101,98,117,102,102,101,114,95,109,105,120,101,100,95,115,97,109,112,108,101,115,0,71,76,95,78,86,95,102,114,97,109,101,98,117,102,102,101,114,95,109,117,108,116,105,115,97,109,112,108,101,0,71,76,95,78,86,95,102,114,97,109,101,98,117,102,102,101,114,95,109,117,108,116,105,115,97,109,112,108,101,95,99,111,118,101,114,97,103,101,0,71,76,95,78,86,95,103,101,110,101,114,97,116,101,95,109,105,112,109,97,112,95,115,82,71,66,0,71,76,95,78,86,95,103,101,111,109,101,116,114,121,95,112,114,111,103,114,97,109,52,0,71,76,95,78,86,95,103,101,111,109,101,116,114,121,95,115,104,97,100,101,114,52,0,71,76,95,78,86,95,103,101,111,109,101,116,114,121,95,115,104,97,100,101,114,95,112,97,115,115,116,104,114,111,117,103,104,0,71,76,95,78,86,95,103,112,117,95,112,114,111,103,114,97,109,52,0,71,76,95,78,86,95,103,112,117,95,112,114,111,103,114,97,109,53,0,71,76,95,78,86,95,103,112,117,95,112,114,111,103,114,97,109,53,95,109,101,109,95,101,120,116,101,110,100,101,100,0,71,76,95,78,86,95,103,112,117,95,115,104,97,100,101,114,53,0,71,76,95,78,86,95,104,97,108,102,95,102,108,111,97,116,0,71,76,95,78,86,95,105,109,97,103,101,95,102,111,114,109,97,116,115,0,71,76,95,78,86,95,105,110,115,116,97,110,99,101,100,95,97,114,114,97,121,115,0,71,76,95,78,86,95,105,110,116,101,114,110,97,108,102,111,114,109,97,116,95,115,97,109,112,108,101,95,113,117,101,114,121,0,71,76,95,78,86,95,109,117,108,116,105,115,97,109,112,108,101,95,99,111,118,101,114,97,103,101,0,71,76,95,78,86,95,110,111,110,95,115,113,117,97,114,101,95,109,97,116,114,105,99,101,115,0,71,76,95,78,86,95,111,99,99,108,117,115,105,111,110,95,113,117,101,114,121,0,71,76,95,78,86,95,112,97,114,97,109,101,116,101,114,95,98,117,102,102,101,114,95,111,98,106,101,99,116,0,71,76,95,78,86,95,112,97,114,97,109,101,116,101,114,95,98,117,102,102,101,114,95,111,98,106,101,99,116,50,0,71,76,95,78,86,95,112,97,116,104,95,114,101,110,100,101,114,105,110,103,0,71,76,95,78,86,95,112,97,116,104,95,114,101,110,100,101,114,105,110,103,95,115,104,97,114,101,100,95,101,100,103,101,0,71,76,95,78,86,95,112,111,108,121,103,111,110,95,109,111,100,101,0,71,76,95,78,86,95,112,114,101,115,101,110,116,95,118,105,100,101,111,0,71,76,95,78,86,95,112,114,105,109,105,116,105,118,101,95,114,101,115,116,97,114,116,0,71,76,95,78,86,95,114,101,97,100,95,98,117,102,102,101,114,0,71,76,95,78,86,95,114,101,97,100,95,98,117,102,102,101,114,95,102,114,111,110,116,0,71,76,95,78,86,95,114,101,97,100,95,100,101,112,116,104,0,71,76,95,78,86,95,114,101,97,100,95,100,101,112,116,104,95,115,116,101,110,99,105,108,0,71,76,95,78,86,95,114,101,97,100,95,115,116,101,110,99,105,108,0,71,76,95,78,86,95,115,97,109,112,108,101,95,108,111,99,97,116,105,111,110,115,0,71,76,95,78,86,95,115,97,109,112,108,101,95,109,97,115,107,95,111,118,101,114,114,105,100,101,95,99,111,118,101,114,97,103,101,0,71,76,95,78,86,95,115,104,97,100,101,114,95,97,116,111,109,105,99,95,99,111,117,110,116,101,114,115,0,71,76,95,78,86,95,115,104,97,100,101,114,95,97,116,111,109,105,99,95,102,108,111,97,116,0,71,76,95,78,86,95,115,104,97,100,101,114,95,97,116,111,109,105,99,95,102,112,49,54,95,118,101,99,116,111,114,0,71,76,95,78,86,95,115,104,97,100,101,114,95,97,116,111,109,105,99,95,105,110,116,54,52,0,71,76,95,78,86,95,115,104,97,100,101,114,95,98,117,102,102,101,114,95,108,111,97,100,0,71,76,95,78,86,95,115,104,97,100,101,114,95,98,117,102,102,101,114,95,115,116,111,114,101,0,71,76,95,78,86,95,115,104,97,100,101,114,95,110,111,112,101,114,115,112,101,99,116,105,118,101,95,105,110,116,101,114,112,111,108,97,116,105,111,110,0,71,76,95,78,86,95,115,104,97,100,101,114,95,115,116,111,114,97,103,101,95,98,117,102,102,101,114,95,111,98,106,101,99,116,0,71,76,95,78,86,95,115,104,97,100,101,114,95,116,104,114,101,97,100,95,103,114,111,117,112,0,71,76,95,78,86,95,115,104,97,100,101,114,95,116,104,114,101,97,100,95,115,104,117,102,102,108,101,0,71,76,95,78,86,95,115,104,97,100,111,119,95,115,97,109,112,108,101,114,115,95,97,114,114,97,121,0,71,76,95,78,86,95,115,104,97,100,111,119,95,115,97,109,112,108,101,114,115,95,99,117,98,101,0,71,76,95,78,86,95,115,82,71,66,95,102,111,114,109,97,116,115,0,71,76,95,78,86,95,116,101,115,115,101,108,108,97,116,105,111,110,95,112,114,111,103,114,97,109,53,0,71,76,95,78,86,95,116,101,120,116,117,114,101,95,98,97,114,114,105,101,114,0,71,76,95,78,86,95,116,101,120,116,117,114,101,95,98,111,114,100,101,114,95,99,108,97,109,112,0,71,76,95,78,86,95,116,101,120,116,117,114,101,95,99,111,109,112,114,101,115,115,105,111,110,95,115,51,116,99,95,117,112,100,97,116,101,0,71,76,95,78,86,95,116,101,120,116,117,114,101,95,109,117,108,116,105,115,97,109,112,108,101,0,71,76,95,78,86,95,116,101,120,116,117,114,101,95,110,112,111,116,95,50,68,95,109,105,112,109,97,112,0,71,76,95,78,86,95,116,114,97,110,115,102,111,114,109,95,102,101,101,100,98,97,99,107,0,71,76,95,78,86,95,116,114,97,110,115,102,111,114,109,95,102,101,101,100,98,97,99,107,50,0,71,76,95,78,86,95,117,110,105,102,111,114,109,95,98,117,102,102,101,114,95,117,110,105,102,105,101,100,95,109,101,109,111,114,121,0,71,76,95,78,86,95,118,100,112,97,117,95,105,110,116,101,114,111,112,0,71,76,95,78,86,95,118,101,114,116,101,120,95,97,116,116,114,105,98,95,105,110,116,101,103,101,114,95,54,52,98,105,116,0,71,76,95,78,86,95,118,101,114,116,101,120,95,98,117,102,102,101,114,95,117,110,105,102,105,101,100,95,109,101,109,111,114,121,0,71,76,95,78,86,95,118,101,114,116,101,120,95,112,114,111,103,114,97,109,52,0,71,76,95,78,86,95,118,105,100,101,111,95,99,97,112,116,117,114,101,0,71,76,95,78,86,95,118,105,101,119,112,111,114,116,95,97,114,114,97,121,0,71,76,95,78,86,95,118,105,101,119,112,111,114,116,95,97,114,114,97,121,50,0,71,76,95,78,86,88,95,99,111,110,100,105,116,105,111,110,97,108,95,114,101,110,100,101,114,0,71,76,95,78,86,88,95,103,112,117,95,109,101,109,111,114,121,95,105,110,102,111,0,71,76,95,79,86,82,95,109,117,108,116,105,118,105,101,119,0,71,76,95,79,86,82,95,109,117,108,116,105,118,105,101,119,50,0,71,76,95,79,86,82,95,109,117,108,116,105,118,105,101,119,95,109,117,108,116,105,115,97,109,112,108,101,100,95,114,101,110,100,101,114,95,116,111,95,116,101,120,116,117,114,101,0,71,76,95,81,67,79,77,95,97,108,112,104,97,95,116,101,115,116,0,71,76,95,81,67,79,77,95,98,105,110,110,105,110,103,95,99,111,110,116,114,111,108,0,71,76,95,81,67,79,77,95,100,114,105,118,101,114,95,99,111,110,116,114,111,108,0,71,76,95,81,67,79,77,95,101,120,116,101,110,100,101,100,95,103,101,116,0,71,76,95,81,67,79,77,95,101,120,116,101,110,100,101,100,95,103,101,116,50,0,71,76,95,81,67,79,77,95,112,101,114,102,109,111,110,95,103,108,111,98,97,108,95,109,111,100,101,0,71,76,95,81,67,79,77,95,116,105,108,101,100,95,114,101,110,100,101,114,105,110,103,0,71,76,95,81,67,79,77,95,119,114,105,116,101,111,110,108,121,95,114,101,110,100,101,114,105,110,103,0,71,76,95,86,73,86,95,115,104,97,100,101,114,95,98,105,110,97,114,121,0,103,108,66,108,101,110,100,66,97,114,114,105,101,114,75,72,82,0,103,108,68,101,98,117,103,77,101,115,115,97,103,101,67,111,110,116,114,111,108,75,72,82,0,103,108,68,101,98,117,103,77,101,115,115,97,103,101,73,110,115,101,114,116,75,72,82,0,103,108,68,101,98,117,103,77,101,115,115,97,103,101,67,97,108,108,98,97,99,107,75,72,82,0,103,108,71,101,116,68,101,98,117,103,77,101,115,115,97,103,101,76,111,103,75,72,82,0,103,108,80,117,115,104,68,101,98,117,103,71,114,111,117,112,75,72,82,0,103,108,80,111,112,68,101,98,117,103,71,114,111,117,112,75,72,82,0,103,108,79,98,106,101,99,116,76,97,98,101,108,75,72,82,0,103,108,71,101,116,79,98,106,101,99,116,76,97,98,101,108,75,72,82,0,103,108,79,98,106,101,99,116,80,116,114,76,97,98,101,108,75,72,82,0,103,108,71,101,116,79,98,106,101,99,116,80,116,114,76,97,98,101,108,75,72,82,0,103,108,71,101,116,80,111,105,110,116,101,114,118,75,72,82,0,103,108,71,101,116,71,114,97,112,104,105,99,115,82,101,115,101,116,83,116,97,116,117,115,75,72,82,0,103,108,82,101,97,100,110,80,105,120,101,108,115,75,72,82,0,103,108,71,101,116,110,85,110,105,102,111,114,109,102,118,75,72,82,0,103,108,71,101,116,110,85,110,105,102,111,114,109,105,118,75,72,82,0,103,108,71,101,116,110,85,110,105,102,111,114,109,117,105,118,75,72,82,0,103,108,71,101,116,84,101,120,116,117,114,101,72,97,110,100,108,101,65,82,66,0,103,108,71,101,116,84,101,120,116,117,114,101,83,97,109,112,108,101,114,72,97,110,100,108,101,65,82,66,0,103,108,77,97,107,101,84,101,120,116,117,114,101,72,97,110,100,108,101,82,101,115,105,100,101,110,116,65,82,66,0,103,108,77,97,107,101,84,101,120,116,117,114,101,72,97,110,100,108,101,78,111,110,82,101,115,105,100,101,110,116,65,82,66,0,103,108,71,101,116,73,109,97,103,101,72,97,110,100,108,101,65,82,66,0,103,108,77,97,107,101,73,109,97,103,101,72,97,110,100,108,101,82,101,115,105,100,101,110,116,65,82,66,0,103,108,77,97,107,101,73,109,97,103,101,72,97,110,100,108,101,78,111,110,82,101,115,105,100,101,110,116,65,82,66,0,103,108,85,110,105,102,111,114,109,72,97,110,100,108,101,117,105,54,52,65,82,66,0,103,108,85,110,105,102,111,114,109,72,97,110,100,108,101,117,105,54,52,118,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,72,97,110,100,108,101,117,105,54,52,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,72,97,110,100,108,101,117,105,54,52,118,65,82,66,0,103,108,73,115,84,101,120,116,117,114,101,72,97,110,100,108,101,82,101,115,105,100,101,110,116,65,82,66,0,103,108,73,115,73,109,97,103,101,72,97,110,100,108,101,82,101,115,105,100,101,110,116,65,82,66,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,49,117,105,54,52,65,82,66,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,49,117,105,54,52,118,65,82,66,0,103,108,71,101,116,86,101,114,116,101,120,65,116,116,114,105,98,76,117,105,54,52,118,65,82,66,0,103,108,67,114,101,97,116,101,83,121,110,99,70,114,111,109,67,76,101,118,101,110,116,65,82,66,0,103,108,67,108,97,109,112,67,111,108,111,114,65,82,66,0,103,108,68,105,115,112,97,116,99,104,67,111,109,112,117,116,101,71,114,111,117,112,83,105,122,101,65,82,66,0,103,108,68,101,98,117,103,77,101,115,115,97,103,101,67,111,110,116,114,111,108,65,82,66,0,103,108,68,101,98,117,103,77,101,115,115,97,103,101,73,110,115,101,114,116,65,82,66,0,103,108,68,101,98,117,103,77,101,115,115,97,103,101,67,97,108,108,98,97,99,107,65,82,66,0,103,108,71,101,116,68,101,98,117,103,77,101,115,115,97,103,101,76,111,103,65,82,66,0,103,108,66,108,101,110,100,69,113,117,97,116,105,111,110,105,65,82,66,0,103,108,66,108,101,110,100,69,113,117,97,116,105,111,110,83,101,112,97,114,97,116,101,105,65,82,66,0,103,108,66,108,101,110,100,70,117,110,99,105,65,82,66,0,103,108,66,108,101,110,100,70,117,110,99,83,101,112,97,114,97,116,101,105,65,82,66,0,103,108,68,114,97,119,65,114,114,97,121,115,73,110,115,116,97,110,99,101,100,65,82,66,0,103,108,68,114,97,119,69,108,101,109,101,110,116,115,73,110,115,116,97,110,99,101,100,65,82,66,0,103,108,80,114,105,109,105,116,105,118,101,66,111,117,110,100,105,110,103,66,111,120,65,82,66,0,103,108,80,114,111,103,114,97,109,80,97,114,97,109,101,116,101,114,105,65,82,66,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,65,82,66,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,76,97,121,101,114,65,82,66,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,70,97,99,101,65,82,66,0,103,108,85,110,105,102,111,114,109,49,105,54,52,65,82,66,0,103,108,85,110,105,102,111,114,109,50,105,54,52,65,82,66,0,103,108,85,110,105,102,111,114,109,51,105,54,52,65,82,66,0,103,108,85,110,105,102,111,114,109,52,105,54,52,65,82,66,0,103,108,85,110,105,102,111,114,109,49,105,54,52,118,65,82,66,0,103,108,85,110,105,102,111,114,109,50,105,54,52,118,65,82,66,0,103,108,85,110,105,102,111,114,109,51,105,54,52,118,65,82,66,0,103,108,85,110,105,102,111,114,109,52,105,54,52,118,65,82,66,0,103,108,85,110,105,102,111,114,109,49,117,105,54,52,65,82,66,0,103,108,85,110,105,102,111,114,109,50,117,105,54,52,65,82,66,0,103,108,85,110,105,102,111,114,109,51,117,105,54,52,65,82,66,0,103,108,85,110,105,102,111,114,109,52,117,105,54,52,65,82,66,0,103,108,85,110,105,102,111,114,109,49,117,105,54,52,118,65,82,66,0,103,108,85,110,105,102,111,114,109,50,117,105,54,52,118,65,82,66,0,103,108,85,110,105,102,111,114,109,51,117,105,54,52,118,65,82,66,0,103,108,85,110,105,102,111,114,109,52,117,105,54,52,118,65,82,66,0,103,108,71,101,116,85,110,105,102,111,114,109,105,54,52,118,65,82,66,0,103,108,71,101,116,85,110,105,102,111,114,109,117,105,54,52,118,65,82,66,0,103,108,71,101,116,110,85,110,105,102,111,114,109,105,54,52,118,65,82,66,0,103,108,71,101,116,110,85,110,105,102,111,114,109,117,105,54,52,118,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,105,54,52,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,105,54,52,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,105,54,52,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,105,54,52,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,105,54,52,118,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,105,54,52,118,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,105,54,52,118,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,105,54,52,118,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,117,105,54,52,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,117,105,54,52,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,117,105,54,52,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,117,105,54,52,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,117,105,54,52,118,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,117,105,54,52,118,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,117,105,54,52,118,65,82,66,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,117,105,54,52,118,65,82,66,0,103,108,77,117,108,116,105,68,114,97,119,65,114,114,97,121,115,73,110,100,105,114,101,99,116,67,111,117,110,116,65,82,66,0,103,108,77,117,108,116,105,68,114,97,119,69,108,101,109,101,110,116,115,73,110,100,105,114,101,99,116,67,111,117,110,116,65,82,66,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,68,105,118,105,115,111,114,65,82,66,0,103,108,77,97,120,83,104,97,100,101,114,67,111,109,112,105,108,101,114,84,104,114,101,97,100,115,65,82,66,0,103,108,71,101,116,71,114,97,112,104,105,99,115,82,101,115,101,116,83,116,97,116,117,115,65,82,66,0,103,108,71,101,116,110,84,101,120,73,109,97,103,101,65,82,66,0,103,108,82,101,97,100,110,80,105,120,101,108,115,65,82,66,0,103,108,71,101,116,110,67,111,109,112,114,101,115,115,101,100,84,101,120,73,109,97,103,101,65,82,66,0,103,108,71,101,116,110,85,110,105,102,111,114,109,102,118,65,82,66,0,103,108,71,101,116,110,85,110,105,102,111,114,109,105,118,65,82,66,0,103,108,71,101,116,110,85,110,105,102,111,114,109,117,105,118,65,82,66,0,103,108,71,101,116,110,85,110,105,102,111,114,109,100,118,65,82,66,0,103,108,71,101,116,110,77,97,112,100,118,65,82,66,0,103,108,71,101,116,110,77,97,112,102,118,65,82,66,0,103,108,71,101,116,110,77,97,112,105,118,65,82,66,0,103,108,71,101,116,110,80,105,120,101,108,77,97,112,102,118,65,82,66,0,103,108,71,101,116,110,80,105,120,101,108,77,97,112,117,105,118,65,82,66,0,103,108,71,101,116,110,80,105,120,101,108,77,97,112,117,115,118,65,82,66,0,103,108,71,101,116,110,80,111,108,121,103,111,110,83,116,105,112,112,108,101,65,82,66,0,103,108,71,101,116,110,67,111,108,111,114,84,97,98,108,101,65,82,66,0,103,108,71,101,116,110,67,111,110,118,111,108,117,116,105,111,110,70,105,108,116,101,114,65,82,66,0,103,108,71,101,116,110,83,101,112,97,114,97,98,108,101,70,105,108,116,101,114,65,82,66,0,103,108,71,101,116,110,72,105,115,116,111,103,114,97,109,65,82,66,0,103,108,71,101,116,110,77,105,110,109,97,120,65,82,66,0,103,108,70,114,97,109,101,98,117,102,102,101,114,83,97,109,112,108,101,76,111,99,97,116,105,111,110,115,102,118,65,82,66,0,103,108,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,83,97,109,112,108,101,76,111,99,97,116,105,111,110,115,102,118,65,82,66,0,103,108,69,118,97,108,117,97,116,101,68,101,112,116,104,86,97,108,117,101,115,65,82,66,0,103,108,77,105,110,83,97,109,112,108,101,83,104,97,100,105,110,103,65,82,66,0,103,108,78,97,109,101,100,83,116,114,105,110,103,65,82,66,0,103,108,68,101,108,101,116,101,78,97,109,101,100,83,116,114,105,110,103,65,82,66,0,103,108,67,111,109,112,105,108,101,83,104,97,100,101,114,73,110,99,108,117,100,101,65,82,66,0,103,108,73,115,78,97,109,101,100,83,116,114,105,110,103,65,82,66,0,103,108,71,101,116,78,97,109,101,100,83,116,114,105,110,103,65,82,66,0,103,108,71,101,116,78,97,109,101,100,83,116,114,105,110,103,105,118,65,82,66,0,103,108,66,117,102,102,101,114,80,97,103,101,67,111,109,109,105,116,109,101,110,116,65,82,66,0,103,108,78,97,109,101,100,66,117,102,102,101,114,80,97,103,101,67,111,109,109,105,116,109,101,110,116,69,88,84,0,103,108,78,97,109,101,100,66,117,102,102,101,114,80,97,103,101,67,111,109,109,105,116,109,101,110,116,65,82,66,0,103,108,84,101,120,80,97,103,101,67,111,109,109,105,116,109,101,110,116,65,82,66,0,103,108,84,101,120,66,117,102,102,101,114,65,82,66,0,103,108,68,114,97,119,65,114,114,97,121,115,73,110,115,116,97,110,99,101,100,66,97,115,101,73,110,115,116,97,110,99,101,69,88,84,0,103,108,68,114,97,119,69,108,101,109,101,110,116,115,73,110,115,116,97,110,99,101,100,66,97,115,101,73,110,115,116,97,110,99,101,69,88,84,0,103,108,68,114,97,119,69,108,101,109,101,110,116,115,73,110,115,116,97,110,99,101,100,66,97,115,101,86,101,114,116,101,120,66,97,115,101,73,110,115,116,97,110,99,101,69,88,84,0,103,108,85,110,105,102,111,114,109,66,117,102,102,101,114,69,88,84,0,103,108,71,101,116,85,110,105,102,111,114,109,66,117,102,102,101,114,83,105,122,101,69,88,84,0,103,108,71,101,116,85,110,105,102,111,114,109,79,102,102,115,101,116,69,88,84,0,103,108,66,105,110,100,70,114,97,103,68,97,116,97,76,111,99,97,116,105,111,110,73,110,100,101,120,101,100,69,88,84,0,103,108,66,105,110,100,70,114,97,103,68,97,116,97,76,111,99,97,116,105,111,110,69,88,84,0,103,108,71,101,116,80,114,111,103,114,97,109,82,101,115,111,117,114,99,101,76,111,99,97,116,105,111,110,73,110,100,101,120,69,88,84,0,103,108,71,101,116,70,114,97,103,68,97,116,97,73,110,100,101,120,69,88,84,0,103,108,66,108,101,110,100,69,113,117,97,116,105,111,110,69,88,84,0,103,108,66,117,102,102,101,114,83,116,111,114,97,103,101,69,88,84,0,103,108,67,111,112,121,73,109,97,103,101,83,117,98,68,97,116,97,69,88,84,0,103,108,76,97,98,101,108,79,98,106,101,99,116,69,88,84,0,103,108,71,101,116,79,98,106,101,99,116,76,97,98,101,108,69,88,84,0,103,108,73,110,115,101,114,116,69,118,101,110,116,77,97,114,107,101,114,69,88,84,0,103,108,80,117,115,104,71,114,111,117,112,77,97,114,107,101,114,69,88,84,0,103,108,80,111,112,71,114,111,117,112,77,97,114,107,101,114,69,88,84,0,103,108,77,97,116,114,105,120,76,111,97,100,102,69,88,84,0,103,108,77,97,116,114,105,120,76,111,97,100,100,69,88,84,0,103,108,77,97,116,114,105,120,77,117,108,116,102,69,88,84,0,103,108,77,97,116,114,105,120,77,117,108,116,100,69,88,84,0,103,108,77,97,116,114,105,120,76,111,97,100,73,100,101,110,116,105,116,121,69,88,84,0,103,108,77,97,116,114,105,120,82,111,116,97,116,101,102,69,88,84,0,103,108,77,97,116,114,105,120,82,111,116,97,116,101,100,69,88,84,0,103,108,77,97,116,114,105,120,83,99,97,108,101,102,69,88,84,0,103,108,77,97,116,114,105,120,83,99,97,108,101,100,69,88,84,0,103,108,77,97,116,114,105,120,84,114,97,110,115,108,97,116,101,102,69,88,84,0,103,108,77,97,116,114,105,120,84,114,97,110,115,108,97,116,101,100,69,88,84,0,103,108,77,97,116,114,105,120,70,114,117,115,116,117,109,69,88,84,0,103,108,77,97,116,114,105,120,79,114,116,104,111,69,88,84,0,103,108,77,97,116,114,105,120,80,111,112,69,88,84,0,103,108,77,97,116,114,105,120,80,117,115,104,69,88,84,0,103,108,67,108,105,101,110,116,65,116,116,114,105,98,68,101,102,97,117,108,116,69,88,84,0,103,108,80,117,115,104,67,108,105,101,110,116,65,116,116,114,105,98,68,101,102,97,117,108,116,69,88,84,0,103,108,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,102,69,88,84,0,103,108,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,102,118,69,88,84,0,103,108,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,105,69,88,84,0,103,108,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,105,118,69,88,84,0,103,108,84,101,120,116,117,114,101,73,109,97,103,101,49,68,69,88,84,0,103,108,84,101,120,116,117,114,101,73,109,97,103,101,50,68,69,88,84,0,103,108,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,49,68,69,88,84,0,103,108,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,50,68,69,88,84,0,103,108,67,111,112,121,84,101,120,116,117,114,101,73,109,97,103,101,49,68,69,88,84,0,103,108,67,111,112,121,84,101,120,116,117,114,101,73,109,97,103,101,50,68,69,88,84,0,103,108,67,111,112,121,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,49,68,69,88,84,0,103,108,67,111,112,121,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,50,68,69,88,84,0,103,108,71,101,116,84,101,120,116,117,114,101,73,109,97,103,101,69,88,84,0,103,108,71,101,116,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,102,118,69,88,84,0,103,108,71,101,116,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,105,118,69,88,84,0,103,108,71,101,116,84,101,120,116,117,114,101,76,101,118,101,108,80,97,114,97,109,101,116,101,114,102,118,69,88,84,0,103,108,71,101,116,84,101,120,116,117,114,101,76,101,118,101,108,80,97,114,97,109,101,116,101,114,105,118,69,88,84,0,103,108,84,101,120,116,117,114,101,73,109,97,103,101,51,68,69,88,84,0,103,108,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,51,68,69,88,84,0,103,108,67,111,112,121,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,51,68,69,88,84,0,103,108,66,105,110,100,77,117,108,116,105,84,101,120,116,117,114,101,69,88,84,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,80,111,105,110,116,101,114,69,88,84,0,103,108,77,117,108,116,105,84,101,120,69,110,118,102,69,88,84,0,103,108,77,117,108,116,105,84,101,120,69,110,118,102,118,69,88,84,0,103,108,77,117,108,116,105,84,101,120,69,110,118,105,69,88,84,0,103,108,77,117,108,116,105,84,101,120,69,110,118,105,118,69,88,84,0,103,108,77,117,108,116,105,84,101,120,71,101,110,100,69,88,84,0,103,108,77,117,108,116,105,84,101,120,71,101,110,100,118,69,88,84,0,103,108,77,117,108,116,105,84,101,120,71,101,110,102,69,88,84,0,103,108,77,117,108,116,105,84,101,120,71,101,110,102,118,69,88,84,0,103,108,77,117,108,116,105,84,101,120,71,101,110,105,69,88,84,0,103,108,77,117,108,116,105,84,101,120,71,101,110,105,118,69,88,84,0,103,108,71,101,116,77,117,108,116,105,84,101,120,69,110,118,102,118,69,88,84,0,103,108,71,101,116,77,117,108,116,105,84,101,120,69,110,118,105,118,69,88,84,0,103,108,71,101,116,77,117,108,116,105,84,101,120,71,101,110,100,118,69,88,84,0,103,108,71,101,116,77,117,108,116,105,84,101,120,71,101,110,102,118,69,88,84,0,103,108,71,101,116,77,117,108,116,105,84,101,120,71,101,110,105,118,69,88,84,0,103,108,77,117,108,116,105,84,101,120,80,97,114,97,109,101,116,101,114,105,69,88,84,0,103,108,77,117,108,116,105,84,101,120,80,97,114,97,109,101,116,101,114,105,118,69,88,84,0,103,108,77,117,108,116,105,84,101,120,80,97,114,97,109,101,116,101,114,102,69,88,84,0,103,108,77,117,108,116,105,84,101,120,80,97,114,97,109,101,116,101,114,102,118,69,88,84,0,103,108,77,117,108,116,105,84,101,120,73,109,97,103,101,49,68,69,88,84,0,103,108,77,117,108,116,105,84,101,120,73,109,97,103,101,50,68,69,88,84,0,103,108,77,117,108,116,105,84,101,120,83,117,98,73,109,97,103,101,49,68,69,88,84,0,103,108,77,117,108,116,105,84,101,120,83,117,98,73,109,97,103,101,50,68,69,88,84,0,103,108,67,111,112,121,77,117,108,116,105,84,101,120,73,109,97,103,101,49,68,69,88,84,0,103,108,67,111,112,121,77,117,108,116,105,84,101,120,73,109,97,103,101,50,68,69,88,84,0,103,108,67,111,112,121,77,117,108,116,105,84,101,120,83,117,98,73,109,97,103,101,49,68,69,88,84,0,103,108,67,111,112,121,77,117,108,116,105,84,101,120,83,117,98,73,109,97,103,101,50,68,69,88,84,0,103,108,71,101,116,77,117,108,116,105,84,101,120,73,109,97,103,101,69,88,84,0,103,108,71,101,116,77,117,108,116,105,84,101,120,80,97,114,97,109,101,116,101,114,102,118,69,88,84,0,103,108,71,101,116,77,117,108,116,105,84,101,120,80,97,114,97,109,101,116,101,114,105,118,69,88,84,0,103,108,71,101,116,77,117,108,116,105,84,101,120,76,101,118,101,108,80,97,114,97,109,101,116,101,114,102,118,69,88,84,0,103,108,71,101,116,77,117,108,116,105,84,101,120,76,101,118,101,108,80,97,114,97,109,101,116,101,114,105,118,69,88,84,0,103,108,77,117,108,116,105,84,101,120,73,109,97,103,101,51,68,69,88,84,0,103,108,77,117,108,116,105,84,101,120,83,117,98,73,109,97,103,101,51,68,69,88,84,0,103,108,67,111,112,121,77,117,108,116,105,84,101,120,83,117,98,73,109,97,103,101,51,68,69,88,84,0,103,108,69,110,97,98,108,101,67,108,105,101,110,116,83,116,97,116,101,73,110,100,101,120,101,100,69,88,84,0,103,108,68,105,115,97,98,108,101,67,108,105,101,110,116,83,116,97,116,101,73,110,100,101,120,101,100,69,88,84,0,103,108,71,101,116,70,108,111,97,116,73,110,100,101,120,101,100,118,69,88,84,0,103,108,71,101,116,68,111,117,98,108,101,73,110,100,101,120,101,100,118,69,88,84,0,103,108,71,101,116,80,111,105,110,116,101,114,73,110,100,101,120,101,100,118,69,88,84,0,103,108,69,110,97,98,108,101,73,110,100,101,120,101,100,69,88,84,0,103,108,68,105,115,97,98,108,101,73,110,100,101,120,101,100,69,88,84,0,103,108,73,115,69,110,97,98,108,101,100,73,110,100,101,120,101,100,69,88,84,0,103,108,71,101,116,73,110,116,101,103,101,114,73,110,100,101,120,101,100,118,69,88,84,0,103,108,71,101,116,66,111,111,108,101,97,110,73,110,100,101,120,101,100,118,69,88,84,0,103,108,67,111,109,112,114,101,115,115,101,100,84,101,120,116,117,114,101,73,109,97,103,101,51,68,69,88,84,0,103,108,67,111,109,112,114,101,115,115,101,100,84,101,120,116,117,114,101,73,109,97,103,101,50,68,69,88,84,0,103,108,67,111,109,112,114,101,115,115,101,100,84,101,120,116,117,114,101,73,109,97,103,101,49,68,69,88,84,0,103,108,67,111,109,112,114,101,115,115,101,100,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,51,68,69,88,84,0,103,108,67,111,109,112,114,101,115,115,101,100,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,50,68,69,88,84,0,103,108,67,111,109,112,114,101,115,115,101,100,84,101,120,116,117,114,101,83,117,98,73,109,97,103,101,49,68,69,88,84,0,103,108,71,101,116,67,111,109,112,114,101,115,115,101,100,84,101,120,116,117,114,101,73,109,97,103,101,69,88,84,0,103,108,67,111,109,112,114,101,115,115,101,100,77,117,108,116,105,84,101,120,73,109,97,103,101,51,68,69,88,84,0,103,108,67,111,109,112,114,101,115,115,101,100,77,117,108,116,105,84,101,120,73,109,97,103,101,50,68,69,88,84,0,103,108,67,111,109,112,114,101,115,115,101,100,77,117,108,116,105,84,101,120,73,109,97,103,101,49,68,69,88,84,0,103,108,67,111,109,112,114,101,115,115,101,100,77,117,108,116,105,84,101,120,83,117,98,73,109,97,103,101,51,68,69,88,84,0,103,108,67,111,109,112,114,101,115,115,101,100,77,117,108,116,105,84,101,120,83,117,98,73,109,97,103,101,50,68,69,88,84,0,103,108,67,111,109,112,114,101,115,115,101,100,77,117,108,116,105,84,101,120,83,117,98,73,109,97,103,101,49,68,69,88,84,0,103,108,71,101,116,67,111,109,112,114,101,115,115,101,100,77,117,108,116,105,84,101,120,73,109,97,103,101,69,88,84,0,103,108,77,97,116,114,105,120,76,111,97,100,84,114,97,110,115,112,111,115,101,102,69,88,84,0,103,108,77,97,116,114,105,120,76,111,97,100,84,114,97,110,115,112,111,115,101,100,69,88,84,0,103,108,77,97,116,114,105,120,77,117,108,116,84,114,97,110,115,112,111,115,101,102,69,88,84,0,103,108,77,97,116,114,105,120,77,117,108,116,84,114,97,110,115,112,111,115,101,100,69,88,84,0,103,108,78,97,109,101,100,66,117,102,102,101,114,68,97,116,97,69,88,84,0,103,108,78,97,109,101,100,66,117,102,102,101,114,83,117,98,68,97,116,97,69,88,84,0,103,108,77,97,112,78,97,109,101,100,66,117,102,102,101,114,69,88,84,0,103,108,85,110,109,97,112,78,97,109,101,100,66,117,102,102,101,114,69,88,84,0,103,108,71,101,116,78,97,109,101,100,66,117,102,102,101,114,80,97,114,97,109,101,116,101,114,105,118,69,88,84,0,103,108,71,101,116,78,97,109,101,100,66,117,102,102,101,114,80,111,105,110,116,101,114,118,69,88,84,0,103,108,71,101,116,78,97,109,101,100,66,117,102,102,101,114,83,117,98,68,97,116,97,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,102,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,102,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,102,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,102,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,105,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,105,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,105,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,105,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,102,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,102,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,102,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,102,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,105,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,105,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,105,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,105,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,50,102,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,51,102,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,52,102,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,50,120,51,102,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,51,120,50,102,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,50,120,52,102,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,52,120,50,102,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,51,120,52,102,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,52,120,51,102,118,69,88,84,0,103,108,84,101,120,116,117,114,101,66,117,102,102,101,114,69,88,84,0,103,108,77,117,108,116,105,84,101,120,66,117,102,102,101,114,69,88,84,0,103,108,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,73,105,118,69,88,84,0,103,108,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,73,117,105,118,69,88,84,0,103,108,71,101,116,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,73,105,118,69,88,84,0,103,108,71,101,116,84,101,120,116,117,114,101,80,97,114,97,109,101,116,101,114,73,117,105,118,69,88,84,0,103,108,77,117,108,116,105,84,101,120,80,97,114,97,109,101,116,101,114,73,105,118,69,88,84,0,103,108,77,117,108,116,105,84,101,120,80,97,114,97,109,101,116,101,114,73,117,105,118,69,88,84,0,103,108,71,101,116,77,117,108,116,105,84,101,120,80,97,114,97,109,101,116,101,114,73,105,118,69,88,84,0,103,108,71,101,116,77,117,108,116,105,84,101,120,80,97,114,97,109,101,116,101,114,73,117,105,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,117,105,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,117,105,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,117,105,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,117,105,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,117,105,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,117,105,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,117,105,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,117,105,118,69,88,84,0,103,108,78,97,109,101,100,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,115,52,102,118,69,88,84,0,103,108,78,97,109,101,100,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,73,52,105,69,88,84,0,103,108,78,97,109,101,100,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,73,52,105,118,69,88,84,0,103,108,78,97,109,101,100,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,115,73,52,105,118,69,88,84,0,103,108,78,97,109,101,100,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,73,52,117,105,69,88,84,0,103,108,78,97,109,101,100,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,73,52,117,105,118,69,88,84,0,103,108,78,97,109,101,100,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,115,73,52,117,105,118,69,88,84,0,103,108,71,101,116,78,97,109,101,100,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,73,105,118,69,88,84,0,103,108,71,101,116,78,97,109,101,100,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,73,117,105,118,69,88,84,0,103,108,69,110,97,98,108,101,67,108,105,101,110,116,83,116,97,116,101,105,69,88,84,0,103,108,68,105,115,97,98,108,101,67,108,105,101,110,116,83,116,97,116,101,105,69,88,84,0,103,108,71,101,116,70,108,111,97,116,105,95,118,69,88,84,0,103,108,71,101,116,68,111,117,98,108,101,105,95,118,69,88,84,0,103,108,71,101,116,80,111,105,110,116,101,114,105,95,118,69,88,84,0,103,108,78,97,109,101,100,80,114,111,103,114,97,109,83,116,114,105,110,103,69,88,84,0,103,108,78,97,109,101,100,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,52,100,69,88,84,0,103,108,78,97,109,101,100,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,52,100,118,69,88,84,0,103,108,78,97,109,101,100,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,52,102,69,88,84,0,103,108,78,97,109,101,100,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,52,102,118,69,88,84,0,103,108,71,101,116,78,97,109,101,100,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,100,118,69,88,84,0,103,108,71,101,116,78,97,109,101,100,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,102,118,69,88,84,0,103,108,71,101,116,78,97,109,101,100,80,114,111,103,114,97,109,105,118,69,88,84,0,103,108,71,101,116,78,97,109,101,100,80,114,111,103,114,97,109,83,116,114,105,110,103,69,88,84,0,103,108,78,97,109,101,100,82,101,110,100,101,114,98,117,102,102,101,114,83,116,111,114,97,103,101,69,88,84,0,103,108,71,101,116,78,97,109,101,100,82,101,110,100,101,114,98,117,102,102,101,114,80,97,114,97,109,101,116,101,114,105,118,69,88,84,0,103,108,78,97,109,101,100,82,101,110,100,101,114,98,117,102,102,101,114,83,116,111,114,97,103,101,77,117,108,116,105,115,97,109,112,108,101,69,88,84,0,103,108,78,97,109,101,100,82,101,110,100,101,114,98,117,102,102,101,114,83,116,111,114,97,103,101,77,117,108,116,105,115,97,109,112,108,101,67,111,118,101,114,97,103,101,69,88,84,0,103,108,67,104,101,99,107,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,83,116,97,116,117,115,69,88,84,0,103,108,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,49,68,69,88,84,0,103,108,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,50,68,69,88,84,0,103,108,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,51,68,69,88,84,0,103,108,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,82,101,110,100,101,114,98,117,102,102,101,114,69,88,84,0,103,108,71,101,116,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,65,116,116,97,99,104,109,101,110,116,80,97,114,97,109,101,116,101,114,105,118,69,88,84,0,103,108,71,101,110,101,114,97,116,101,84,101,120,116,117,114,101,77,105,112,109,97,112,69,88,84,0,103,108,71,101,110,101,114,97,116,101,77,117,108,116,105,84,101,120,77,105,112,109,97,112,69,88,84,0,103,108,70,114,97,109,101,98,117,102,102,101,114,68,114,97,119,66,117,102,102,101,114,69,88,84,0,103,108,70,114,97,109,101,98,117,102,102,101,114,68,114,97,119,66,117,102,102,101,114,115,69,88,84,0,103,108,70,114,97,109,101,98,117,102,102,101,114,82,101,97,100,66,117,102,102,101,114,69,88,84,0,103,108,71,101,116,70,114,97,109,101,98,117,102,102,101,114,80,97,114,97,109,101,116,101,114,105,118,69,88,84,0,103,108,78,97,109,101,100,67,111,112,121,66,117,102,102,101,114,83,117,98,68,97,116,97,69,88,84,0,103,108,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,69,88,84,0,103,108,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,76,97,121,101,114,69,88,84,0,103,108,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,70,97,99,101,69,88,84,0,103,108,84,101,120,116,117,114,101,82,101,110,100,101,114,98,117,102,102,101,114,69,88,84,0,103,108,77,117,108,116,105,84,101,120,82,101,110,100,101,114,98,117,102,102,101,114,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,86,101,114,116,101,120,79,102,102,115,101,116,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,67,111,108,111,114,79,102,102,115,101,116,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,69,100,103,101,70,108,97,103,79,102,102,115,101,116,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,73,110,100,101,120,79,102,102,115,101,116,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,78,111,114,109,97,108,79,102,102,115,101,116,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,84,101,120,67,111,111,114,100,79,102,102,115,101,116,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,77,117,108,116,105,84,101,120,67,111,111,114,100,79,102,102,115,101,116,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,70,111,103,67,111,111,114,100,79,102,102,115,101,116,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,83,101,99,111,110,100,97,114,121,67,111,108,111,114,79,102,102,115,101,116,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,86,101,114,116,101,120,65,116,116,114,105,98,79,102,102,115,101,116,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,86,101,114,116,101,120,65,116,116,114,105,98,73,79,102,102,115,101,116,69,88,84,0,103,108,69,110,97,98,108,101,86,101,114,116,101,120,65,114,114,97,121,69,88,84,0,103,108,68,105,115,97,98,108,101,86,101,114,116,101,120,65,114,114,97,121,69,88,84,0,103,108,69,110,97,98,108,101,86,101,114,116,101,120,65,114,114,97,121,65,116,116,114,105,98,69,88,84,0,103,108,68,105,115,97,98,108,101,86,101,114,116,101,120,65,114,114,97,121,65,116,116,114,105,98,69,88,84,0,103,108,71,101,116,86,101,114,116,101,120,65,114,114,97,121,73,110,116,101,103,101,114,118,69,88,84,0,103,108,71,101,116,86,101,114,116,101,120,65,114,114,97,121,80,111,105,110,116,101,114,118,69,88,84,0,103,108,71,101,116,86,101,114,116,101,120,65,114,114,97,121,73,110,116,101,103,101,114,105,95,118,69,88,84,0,103,108,71,101,116,86,101,114,116,101,120,65,114,114,97,121,80,111,105,110,116,101,114,105,95,118,69,88,84,0,103,108,77,97,112,78,97,109,101,100,66,117,102,102,101,114,82,97,110,103,101,69,88,84,0,103,108,70,108,117,115,104,77,97,112,112,101,100,78,97,109,101,100,66,117,102,102,101,114,82,97,110,103,101,69,88,84,0,103,108,78,97,109,101,100,66,117,102,102,101,114,83,116,111,114,97,103,101,69,88,84,0,103,108,67,108,101,97,114,78,97,109,101,100,66,117,102,102,101,114,68,97,116,97,69,88,84,0,103,108,67,108,101,97,114,78,97,109,101,100,66,117,102,102,101,114,83,117,98,68,97,116,97,69,88,84,0,103,108,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,80,97,114,97,109,101,116,101,114,105,69,88,84,0,103,108,71,101,116,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,80,97,114,97,109,101,116,101,114,105,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,100,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,100,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,100,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,100,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,100,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,100,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,100,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,100,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,50,100,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,51,100,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,52,100,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,50,120,51,100,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,50,120,52,100,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,51,120,50,100,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,51,120,52,100,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,52,120,50,100,118,69,88,84,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,77,97,116,114,105,120,52,120,51,100,118,69,88,84,0,103,108,84,101,120,116,117,114,101,66,117,102,102,101,114,82,97,110,103,101,69,88,84,0,103,108,84,101,120,116,117,114,101,83,116,111,114,97,103,101,49,68,69,88,84,0,103,108,84,101,120,116,117,114,101,83,116,111,114,97,103,101,50,68,69,88,84,0,103,108,84,101,120,116,117,114,101,83,116,111,114,97,103,101,51,68,69,88,84,0,103,108,84,101,120,116,117,114,101,83,116,111,114,97,103,101,50,68,77,117,108,116,105,115,97,109,112,108,101,69,88,84,0,103,108,84,101,120,116,117,114,101,83,116,111,114,97,103,101,51,68,77,117,108,116,105,115,97,109,112,108,101,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,66,105,110,100,86,101,114,116,101,120,66,117,102,102,101,114,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,86,101,114,116,101,120,65,116,116,114,105,98,70,111,114,109,97,116,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,86,101,114,116,101,120,65,116,116,114,105,98,73,70,111,114,109,97,116,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,86,101,114,116,101,120,65,116,116,114,105,98,76,70,111,114,109,97,116,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,86,101,114,116,101,120,65,116,116,114,105,98,66,105,110,100,105,110,103,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,86,101,114,116,101,120,66,105,110,100,105,110,103,68,105,118,105,115,111,114,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,86,101,114,116,101,120,65,116,116,114,105,98,76,79,102,102,115,101,116,69,88,84,0,103,108,84,101,120,116,117,114,101,80,97,103,101,67,111,109,109,105,116,109,101,110,116,69,88,84,0,103,108,86,101,114,116,101,120,65,114,114,97,121,86,101,114,116,101,120,65,116,116,114,105,98,68,105,118,105,115,111,114,69,88,84,0,103,108,68,105,115,99,97,114,100,70,114,97,109,101,98,117,102,102,101,114,69,88,84,0,103,108,71,101,110,81,117,101,114,105,101,115,69,88,84,0,103,108,68,101,108,101,116,101,81,117,101,114,105,101,115,69,88,84,0,103,108,73,115,81,117,101,114,121,69,88,84,0,103,108,66,101,103,105,110,81,117,101,114,121,69,88,84,0,103,108,69,110,100,81,117,101,114,121,69,88,84,0,103,108,81,117,101,114,121,67,111,117,110,116,101,114,69,88,84,0,103,108,71,101,116,81,117,101,114,121,105,118,69,88,84,0,103,108,71,101,116,81,117,101,114,121,79,98,106,101,99,116,105,118,69,88,84,0,103,108,71,101,116,81,117,101,114,121,79,98,106,101,99,116,117,105,118,69,88,84,0,103,108,71,101,116,81,117,101,114,121,79,98,106,101,99,116,105,54,52,118,69,88,84,0,103,108,71,101,116,81,117,101,114,121,79,98,106,101,99,116,117,105,54,52,118,69,88,84,0,103,108,68,114,97,119,66,117,102,102,101,114,115,69,88,84,0,103,108,67,111,108,111,114,77,97,115,107,73,110,100,101,120,101,100,69,88,84,0,103,108,69,110,97,98,108,101,105,69,88,84,0,103,108,68,105,115,97,98,108,101,105,69,88,84,0,103,108,66,108,101,110,100,69,113,117,97,116,105,111,110,105,69,88,84,0,103,108,66,108,101,110,100,69,113,117,97,116,105,111,110,83,101,112,97,114,97,116,101,105,69,88,84,0,103,108,66,108,101,110,100,70,117,110,99,105,69,88,84,0,103,108,66,108,101,110,100,70,117,110,99,83,101,112,97,114,97,116,101,105,69,88,84,0,103,108,67,111,108,111,114,77,97,115,107,105,69,88,84,0,103,108,73,115,69,110,97,98,108,101,100,105,69,88,84,0,103,108,68,114,97,119,69,108,101,109,101,110,116,115,66,97,115,101,86,101,114,116,101,120,69,88,84,0,103,108,68,114,97,119,82,97,110,103,101,69,108,101,109,101,110,116,115,66,97,115,101,86,101,114,116,101,120,69,88,84,0,103,108,68,114,97,119,69,108,101,109,101,110,116,115,73,110,115,116,97,110,99,101,100,66,97,115,101,86,101,114,116,101,120,69,88,84,0,103,108,77,117,108,116,105,68,114,97,119,69,108,101,109,101,110,116,115,66,97,115,101,86,101,114,116,101,120,69,88,84,0,103,108,68,114,97,119,65,114,114,97,121,115,73,110,115,116,97,110,99,101,100,69,88,84,0,103,108,68,114,97,119,69,108,101,109,101,110,116,115,73,110,115,116,97,110,99,101,100,69,88,84,0,103,108,66,108,105,116,70,114,97,109,101,98,117,102,102,101,114,69,88,84,0,103,108,82,101,110,100,101,114,98,117,102,102,101,114,83,116,111,114,97,103,101,77,117,108,116,105,115,97,109,112,108,101,69,88,84,0,103,108,73,115,82,101,110,100,101,114,98,117,102,102,101,114,69,88,84,0,103,108,66,105,110,100,82,101,110,100,101,114,98,117,102,102,101,114,69,88,84,0,103,108,68,101,108,101,116,101,82,101,110,100,101,114,98,117,102,102,101,114,115,69,88,84,0,103,108,71,101,110,82,101,110,100,101,114,98,117,102,102,101,114,115,69,88,84,0,103,108,82,101,110,100,101,114,98,117,102,102,101,114,83,116,111,114,97,103,101,69,88,84,0,103,108,71,101,116,82,101,110,100,101,114,98,117,102,102,101,114,80,97,114,97,109,101,116,101,114,105,118,69,88,84,0,103,108,73,115,70,114,97,109,101,98,117,102,102,101,114,69,88,84,0,103,108,66,105,110,100,70,114,97,109,101,98,117,102,102,101,114,69,88,84,0,103,108,68,101,108,101,116,101,70,114,97,109,101,98,117,102,102,101,114,115,69,88,84,0,103,108,71,101,110,70,114,97,109,101,98,117,102,102,101,114,115,69,88,84,0,103,108,67,104,101,99,107,70,114,97,109,101,98,117,102,102,101,114,83,116,97,116,117,115,69,88,84,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,49,68,69,88,84,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,50,68,69,88,84,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,51,68,69,88,84,0,103,108,70,114,97,109,101,98,117,102,102,101,114,82,101,110,100,101,114,98,117,102,102,101,114,69,88,84,0,103,108,71,101,116,70,114,97,109,101,98,117,102,102,101,114,65,116,116,97,99,104,109,101,110,116,80,97,114,97,109,101,116,101,114,105,118,69,88,84,0,103,108,71,101,110,101,114,97,116,101,77,105,112,109,97,112,69,88,84,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,69,88,84,0,103,108,80,114,111,103,114,97,109,80,97,114,97,109,101,116,101,114,105,69,88,84,0,103,108,80,114,111,103,114,97,109,69,110,118,80,97,114,97,109,101,116,101,114,115,52,102,118,69,88,84,0,103,108,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,115,52,102,118,69,88,84,0,103,108,71,101,116,85,110,105,102,111,114,109,117,105,118,69,88,84,0,103,108,71,101,116,70,114,97,103,68,97,116,97,76,111,99,97,116,105,111,110,69,88,84,0,103,108,85,110,105,102,111,114,109,49,117,105,69,88,84,0,103,108,85,110,105,102,111,114,109,50,117,105,69,88,84,0,103,108,85,110,105,102,111,114,109,51,117,105,69,88,84,0,103,108,85,110,105,102,111,114,109,52,117,105,69,88,84,0,103,108,85,110,105,102,111,114,109,49,117,105,118,69,88,84,0,103,108,85,110,105,102,111,114,109,50,117,105,118,69,88,84,0,103,108,85,110,105,102,111,114,109,51,117,105,118,69,88,84,0,103,108,85,110,105,102,111,114,109,52,117,105,118,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,68,105,118,105,115,111,114,69,88,84,0,103,108,77,97,112,66,117,102,102,101,114,82,97,110,103,101,69,88,84,0,103,108,70,108,117,115,104,77,97,112,112,101,100,66,117,102,102,101,114,82,97,110,103,101,69,88,84,0,103,108,77,117,108,116,105,68,114,97,119,65,114,114,97,121,115,73,110,100,105,114,101,99,116,69,88,84,0,103,108,77,117,108,116,105,68,114,97,119,69,108,101,109,101,110,116,115,73,110,100,105,114,101,99,116,69,88,84,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,50,68,77,117,108,116,105,115,97,109,112,108,101,69,88,84,0,103,108,82,101,97,100,66,117,102,102,101,114,73,110,100,101,120,101,100,69,88,84,0,103,108,68,114,97,119,66,117,102,102,101,114,115,73,110,100,101,120,101,100,69,88,84,0,103,108,71,101,116,73,110,116,101,103,101,114,105,95,118,69,88,84,0,103,108,80,111,108,121,103,111,110,79,102,102,115,101,116,67,108,97,109,112,69,88,84,0,103,108,80,114,105,109,105,116,105,118,101,66,111,117,110,100,105,110,103,66,111,120,69,88,84,0,103,108,80,114,111,118,111,107,105,110,103,86,101,114,116,101,120,69,88,84,0,103,108,82,97,115,116,101,114,83,97,109,112,108,101,115,69,88,84,0,103,108,71,101,116,71,114,97,112,104,105,99,115,82,101,115,101,116,83,116,97,116,117,115,69,88,84,0,103,108,82,101,97,100,110,80,105,120,101,108,115,69,88,84,0,103,108,71,101,116,110,85,110,105,102,111,114,109,102,118,69,88,84,0,103,108,71,101,116,110,85,110,105,102,111,114,109,105,118,69,88,84,0,103,108,85,115,101,83,104,97,100,101,114,80,114,111,103,114,97,109,69,88,84,0,103,108,65,99,116,105,118,101,80,114,111,103,114,97,109,69,88,84,0,103,108,67,114,101,97,116,101,83,104,97,100,101,114,80,114,111,103,114,97,109,69,88,84,0,103,108,65,99,116,105,118,101,83,104,97,100,101,114,80,114,111,103,114,97,109,69,88,84,0,103,108,66,105,110,100,80,114,111,103,114,97,109,80,105,112,101,108,105,110,101,69,88,84,0,103,108,67,114,101,97,116,101,83,104,97,100,101,114,80,114,111,103,114,97,109,118,69,88,84,0,103,108,68,101,108,101,116,101,80,114,111,103,114,97,109,80,105,112,101,108,105,110,101,115,69,88,84,0,103,108,71,101,110,80,114,111,103,114,97,109,80,105,112,101,108,105,110,101,115,69,88,84,0,103,108,71,101,116,80,114,111,103,114,97,109,80,105,112,101,108,105,110,101,73,110,102,111,76,111,103,69,88,84,0,103,108,71,101,116,80,114,111,103,114,97,109,80,105,112,101,108,105,110,101,105,118,69,88,84,0,103,108,73,115,80,114,111,103,114,97,109,80,105,112,101,108,105,110,101,69,88,84,0,103,108,85,115,101,80,114,111,103,114,97,109,83,116,97,103,101,115,69,88,84,0,103,108,86,97,108,105,100,97,116,101,80,114,111,103,114,97,109,80,105,112,101,108,105,110,101,69,88,84,0,103,108,66,105,110,100,73,109,97,103,101,84,101,120,116,117,114,101,69,88,84,0,103,108,77,101,109,111,114,121,66,97,114,114,105,101,114,69,88,84,0,103,108,84,101,120,80,97,103,101,67,111,109,109,105,116,109,101,110,116,69,88,84,0,103,108,83,116,101,110,99,105,108,67,108,101,97,114,84,97,103,69,88,84,0,103,108,80,97,116,99,104,80,97,114,97,109,101,116,101,114,105,69,88,84,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,76,97,121,101,114,69,88,84,0,103,108,84,101,120,80,97,114,97,109,101,116,101,114,73,105,118,69,88,84,0,103,108,84,101,120,80,97,114,97,109,101,116,101,114,73,117,105,118,69,88,84,0,103,108,71,101,116,84,101,120,80,97,114,97,109,101,116,101,114,73,105,118,69,88,84,0,103,108,71,101,116,84,101,120,80,97,114,97,109,101,116,101,114,73,117,105,118,69,88,84,0,103,108,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,73,105,118,69,88,84,0,103,108,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,73,117,105,118,69,88,84,0,103,108,71,101,116,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,73,105,118,69,88,84,0,103,108,71,101,116,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,73,117,105,118,69,88,84,0,103,108,84,101,120,66,117,102,102,101,114,69,88,84,0,103,108,84,101,120,66,117,102,102,101,114,82,97,110,103,101,69,88,84,0,103,108,67,108,101,97,114,67,111,108,111,114,73,105,69,88,84,0,103,108,67,108,101,97,114,67,111,108,111,114,73,117,105,69,88,84,0,103,108,84,101,120,83,116,111,114,97,103,101,49,68,69,88,84,0,103,108,84,101,120,83,116,111,114,97,103,101,50,68,69,88,84,0,103,108,84,101,120,83,116,111,114,97,103,101,51,68,69,88,84,0,103,108,84,101,120,116,117,114,101,86,105,101,119,69,88,84,0,103,108,66,101,103,105,110,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,69,88,84,0,103,108,69,110,100,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,69,88,84,0,103,108,66,105,110,100,66,117,102,102,101,114,82,97,110,103,101,69,88,84,0,103,108,66,105,110,100,66,117,102,102,101,114,79,102,102,115,101,116,69,88,84,0,103,108,66,105,110,100,66,117,102,102,101,114,66,97,115,101,69,88,84,0,103,108,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,86,97,114,121,105,110,103,115,69,88,84,0,103,108,71,101,116,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,86,97,114,121,105,110,103,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,49,100,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,50,100,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,51,100,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,52,100,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,49,100,118,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,50,100,118,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,51,100,118,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,52,100,118,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,80,111,105,110,116,101,114,69,88,84,0,103,108,71,101,116,86,101,114,116,101,120,65,116,116,114,105,98,76,100,118,69,88,84,0,103,108,73,109,112,111,114,116,83,121,110,99,69,88,84,0,103,108,67,111,112,121,73,109,97,103,101,83,117,98,68,97,116,97,79,69,83,0,103,108,69,110,97,98,108,101,105,79,69,83,0,103,108,68,105,115,97,98,108,101,105,79,69,83,0,103,108,66,108,101,110,100,69,113,117,97,116,105,111,110,105,79,69,83,0,103,108,66,108,101,110,100,69,113,117,97,116,105,111,110,83,101,112,97,114,97,116,101,105,79,69,83,0,103,108,66,108,101,110,100,70,117,110,99,105,79,69,83,0,103,108,66,108,101,110,100,70,117,110,99,83,101,112,97,114,97,116,101,105,79,69,83,0,103,108,67,111,108,111,114,77,97,115,107,105,79,69,83,0,103,108,73,115,69,110,97,98,108,101,100,105,79,69,83,0,103,108,68,114,97,119,69,108,101,109,101,110,116,115,66,97,115,101,86,101,114,116,101,120,79,69,83,0,103,108,68,114,97,119,82,97,110,103,101,69,108,101,109,101,110,116,115,66,97,115,101,86,101,114,116,101,120,79,69,83,0,103,108,68,114,97,119,69,108,101,109,101,110,116,115,73,110,115,116,97,110,99,101,100,66,97,115,101,86,101,114,116,101,120,79,69,83,0,103,108,77,117,108,116,105,68,114,97,119,69,108,101,109,101,110,116,115,66,97,115,101,86,101,114,116,101,120,79,69,83,0,103,108,69,71,76,73,109,97,103,101,84,97,114,103,101,116,84,101,120,116,117,114,101,50,68,79,69,83,0,103,108,69,71,76,73,109,97,103,101,84,97,114,103,101,116,82,101,110,100,101,114,98,117,102,102,101,114,83,116,111,114,97,103,101,79,69,83,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,79,69,83,0,103,108,71,101,116,80,114,111,103,114,97,109,66,105,110,97,114,121,79,69,83,0,103,108,80,114,111,103,114,97,109,66,105,110,97,114,121,79,69,83,0,103,108,77,97,112,66,117,102,102,101,114,79,69,83,0,103,108,85,110,109,97,112,66,117,102,102,101,114,79,69,83,0,103,108,71,101,116,66,117,102,102,101,114,80,111,105,110,116,101,114,118,79,69,83,0,103,108,80,114,105,109,105,116,105,118,101,66,111,117,110,100,105,110,103,66,111,120,79,69,83,0,103,108,77,105,110,83,97,109,112,108,101,83,104,97,100,105,110,103,79,69,83,0,103,108,80,97,116,99,104,80,97,114,97,109,101,116,101,114,105,79,69,83,0,103,108,84,101,120,73,109,97,103,101,51,68,79,69,83,0,103,108,84,101,120,83,117,98,73,109,97,103,101,51,68,79,69,83,0,103,108,67,111,112,121,84,101,120,83,117,98,73,109,97,103,101,51,68,79,69,83,0,103,108,67,111,109,112,114,101,115,115,101,100,84,101,120,73,109,97,103,101,51,68,79,69,83,0,103,108,67,111,109,112,114,101,115,115,101,100,84,101,120,83,117,98,73,109,97,103,101,51,68,79,69,83,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,51,68,79,69,83,0,103,108,84,101,120,80,97,114,97,109,101,116,101,114,73,105,118,79,69,83,0,103,108,84,101,120,80,97,114,97,109,101,116,101,114,73,117,105,118,79,69,83,0,103,108,71,101,116,84,101,120,80,97,114,97,109,101,116,101,114,73,105,118,79,69,83,0,103,108,71,101,116,84,101,120,80,97,114,97,109,101,116,101,114,73,117,105,118,79,69,83,0,103,108,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,73,105,118,79,69,83,0,103,108,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,73,117,105,118,79,69,83,0,103,108,71,101,116,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,73,105,118,79,69,83,0,103,108,71,101,116,83,97,109,112,108,101,114,80,97,114,97,109,101,116,101,114,73,117,105,118,79,69,83,0,103,108,84,101,120,66,117,102,102,101,114,79,69,83,0,103,108,84,101,120,66,117,102,102,101,114,82,97,110,103,101,79,69,83,0,103,108,84,101,120,83,116,111,114,97,103,101,51,68,77,117,108,116,105,115,97,109,112,108,101,79,69,83,0,103,108,84,101,120,116,117,114,101,86,105,101,119,79,69,83,0,103,108,66,105,110,100,86,101,114,116,101,120,65,114,114,97,121,79,69,83,0,103,108,68,101,108,101,116,101,86,101,114,116,101,120,65,114,114,97,121,115,79,69,83,0,103,108,71,101,110,86,101,114,116,101,120,65,114,114,97,121,115,79,69,83,0,103,108,73,115,86,101,114,116,101,120,65,114,114,97,121,79,69,83,0,103,108,68,101,98,117,103,77,101,115,115,97,103,101,69,110,97,98,108,101,65,77,68,0,103,108,68,101,98,117,103,77,101,115,115,97,103,101,73,110,115,101,114,116,65,77,68,0,103,108,68,101,98,117,103,77,101,115,115,97,103,101,67,97,108,108,98,97,99,107,65,77,68,0,103,108,71,101,116,68,101,98,117,103,77,101,115,115,97,103,101,76,111,103,65,77,68,0,103,108,66,108,101,110,100,70,117,110,99,73,110,100,101,120,101,100,65,77,68,0,103,108,66,108,101,110,100,70,117,110,99,83,101,112,97,114,97,116,101,73,110,100,101,120,101,100,65,77,68,0,103,108,66,108,101,110,100,69,113,117,97,116,105,111,110,73,110,100,101,120,101,100,65,77,68,0,103,108,66,108,101,110,100,69,113,117,97,116,105,111,110,83,101,112,97,114,97,116,101,73,110,100,101,120,101,100,65,77,68,0,103,108,85,110,105,102,111,114,109,49,105,54,52,78,86,0,103,108,85,110,105,102,111,114,109,50,105,54,52,78,86,0,103,108,85,110,105,102,111,114,109,51,105,54,52,78,86,0,103,108,85,110,105,102,111,114,109,52,105,54,52,78,86,0,103,108,85,110,105,102,111,114,109,49,105,54,52,118,78,86,0,103,108,85,110,105,102,111,114,109,50,105,54,52,118,78,86,0,103,108,85,110,105,102,111,114,109,51,105,54,52,118,78,86,0,103,108,85,110,105,102,111,114,109,52,105,54,52,118,78,86,0,103,108,85,110,105,102,111,114,109,49,117,105,54,52,78,86,0,103,108,85,110,105,102,111,114,109,50,117,105,54,52,78,86,0,103,108,85,110,105,102,111,114,109,51,117,105,54,52,78,86,0,103,108,85,110,105,102,111,114,109,52,117,105,54,52,78,86,0,103,108,85,110,105,102,111,114,109,49,117,105,54,52,118,78,86,0,103,108,85,110,105,102,111,114,109,50,117,105,54,52,118,78,86,0,103,108,85,110,105,102,111,114,109,51,117,105,54,52,118,78,86,0,103,108,85,110,105,102,111,114,109,52,117,105,54,52,118,78,86,0,103,108,71,101,116,85,110,105,102,111,114,109,105,54,52,118,78,86,0,103,108,71,101,116,85,110,105,102,111,114,109,117,105,54,52,118,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,105,54,52,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,105,54,52,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,105,54,52,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,105,54,52,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,105,54,52,118,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,105,54,52,118,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,105,54,52,118,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,105,54,52,118,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,117,105,54,52,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,117,105,54,52,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,117,105,54,52,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,117,105,54,52,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,49,117,105,54,52,118,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,50,117,105,54,52,118,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,51,117,105,54,52,118,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,52,117,105,54,52,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,80,97,114,97,109,101,116,101,114,105,65,77,68,0,103,108,77,117,108,116,105,68,114,97,119,65,114,114,97,121,115,73,110,100,105,114,101,99,116,65,77,68,0,103,108,77,117,108,116,105,68,114,97,119,69,108,101,109,101,110,116,115,73,110,100,105,114,101,99,116,65,77,68,0,103,108,71,101,110,78,97,109,101,115,65,77,68,0,103,108,68,101,108,101,116,101,78,97,109,101,115,65,77,68,0,103,108,73,115,78,97,109,101,65,77,68,0,103,108,81,117,101,114,121,79,98,106,101,99,116,80,97,114,97,109,101,116,101,114,117,105,65,77,68,0,103,108,71,101,116,80,101,114,102,77,111,110,105,116,111,114,71,114,111,117,112,115,65,77,68,0,103,108,71,101,116,80,101,114,102,77,111,110,105,116,111,114,67,111,117,110,116,101,114,115,65,77,68,0,103,108,71,101,116,80,101,114,102,77,111,110,105,116,111,114,71,114,111,117,112,83,116,114,105,110,103,65,77,68,0,103,108,71,101,116,80,101,114,102,77,111,110,105,116,111,114,67,111,117,110,116,101,114,83,116,114,105,110,103,65,77,68,0,103,108,71,101,116,80,101,114,102,77,111,110,105,116,111,114,67,111,117,110,116,101,114,73,110,102,111,65,77,68,0,103,108,71,101,110,80,101,114,102,77,111,110,105,116,111,114,115,65,77,68,0,103,108,68,101,108,101,116,101,80,101,114,102,77,111,110,105,116,111,114,115,65,77,68,0,103,108,83,101,108,101,99,116,80,101,114,102,77,111,110,105,116,111,114,67,111,117,110,116,101,114,115,65,77,68,0,103,108,66,101,103,105,110,80,101,114,102,77,111,110,105,116,111,114,65,77,68,0,103,108,69,110,100,80,101,114,102,77,111,110,105,116,111,114,65,77,68,0,103,108,71,101,116,80,101,114,102,77,111,110,105,116,111,114,67,111,117,110,116,101,114,68,97,116,97,65,77,68,0,103,108,83,101,116,77,117,108,116,105,115,97,109,112,108,101,102,118,65,77,68,0,103,108,84,101,120,83,116,111,114,97,103,101,83,112,97,114,115,101,65,77,68,0,103,108,84,101,120,116,117,114,101,83,116,111,114,97,103,101,83,112,97,114,115,101,65,77,68,0,103,108,83,116,101,110,99,105,108,79,112,86,97,108,117,101,65,77,68,0,103,108,84,101,115,115,101,108,108,97,116,105,111,110,70,97,99,116,111,114,65,77,68,0,103,108,84,101,115,115,101,108,108,97,116,105,111,110,77,111,100,101,65,77,68,0,103,108,66,108,105,116,70,114,97,109,101,98,117,102,102,101,114,65,78,71,76,69,0,103,108,82,101,110,100,101,114,98,117,102,102,101,114,83,116,111,114,97,103,101,77,117,108,116,105,115,97,109,112,108,101,65,78,71,76,69,0,103,108,68,114,97,119,65,114,114,97,121,115,73,110,115,116,97,110,99,101,100,65,78,71,76,69,0,103,108,68,114,97,119,69,108,101,109,101,110,116,115,73,110,115,116,97,110,99,101,100,65,78,71,76,69,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,68,105,118,105,115,111,114,65,78,71,76,69,0,103,108,71,101,116,84,114,97,110,115,108,97,116,101,100,83,104,97,100,101,114,83,111,117,114,99,101,65,78,71,76,69,0,103,108,67,111,112,121,84,101,120,116,117,114,101,76,101,118,101,108,115,65,80,80,76,69,0,103,108,69,108,101,109,101,110,116,80,111,105,110,116,101,114,65,80,80,76,69,0,103,108,68,114,97,119,69,108,101,109,101,110,116,65,114,114,97,121,65,80,80,76,69,0,103,108,68,114,97,119,82,97,110,103,101,69,108,101,109,101,110,116,65,114,114,97,121,65,80,80,76,69,0,103,108,77,117,108,116,105,68,114,97,119,69,108,101,109,101,110,116,65,114,114,97,121,65,80,80,76,69,0,103,108,77,117,108,116,105,68,114,97,119,82,97,110,103,101,69,108,101,109,101,110,116,65,114,114,97,121,65,80,80,76,69,0,103,108,71,101,110,70,101,110,99,101,115,65,80,80,76,69,0,103,108,68,101,108,101,116,101,70,101,110,99,101,115,65,80,80,76,69,0,103,108,83,101,116,70,101,110,99,101,65,80,80,76,69,0,103,108,73,115,70,101,110,99,101,65,80,80,76,69,0,103,108,84,101,115,116,70,101,110,99,101,65,80,80,76,69,0,103,108,70,105,110,105,115,104,70,101,110,99,101,65,80,80,76,69,0,103,108,84,101,115,116,79,98,106,101,99,116,65,80,80,76,69,0,103,108,70,105,110,105,115,104,79,98,106,101,99,116,65,80,80,76,69,0,103,108,66,117,102,102,101,114,80,97,114,97,109,101,116,101,114,105,65,80,80,76,69,0,103,108,70,108,117,115,104,77,97,112,112,101,100,66,117,102,102,101,114,82,97,110,103,101,65,80,80,76,69,0,103,108,82,101,110,100,101,114,98,117,102,102,101,114,83,116,111,114,97,103,101,77,117,108,116,105,115,97,109,112,108,101,65,80,80,76,69,0,103,108,82,101,115,111,108,118,101,77,117,108,116,105,115,97,109,112,108,101,70,114,97,109,101,98,117,102,102,101,114,65,80,80,76,69,0,103,108,79,98,106,101,99,116,80,117,114,103,101,97,98,108,101,65,80,80,76,69,0,103,108,79,98,106,101,99,116,85,110,112,117,114,103,101,97,98,108,101,65,80,80,76,69,0,103,108,71,101,116,79,98,106,101,99,116,80,97,114,97,109,101,116,101,114,105,118,65,80,80,76,69,0,103,108,70,101,110,99,101,83,121,110,99,65,80,80,76,69,0,103,108,73,115,83,121,110,99,65,80,80,76,69,0,103,108,68,101,108,101,116,101,83,121,110,99,65,80,80,76,69,0,103,108,67,108,105,101,110,116,87,97,105,116,83,121,110,99,65,80,80,76,69,0,103,108,87,97,105,116,83,121,110,99,65,80,80,76,69,0,103,108,71,101,116,73,110,116,101,103,101,114,54,52,118,65,80,80,76,69,0,103,108,71,101,116,83,121,110,99,105,118,65,80,80,76,69,0,103,108,84,101,120,116,117,114,101,82,97,110,103,101,65,80,80,76,69,0,103,108,71,101,116,84,101,120,80,97,114,97,109,101,116,101,114,80,111,105,110,116,101,114,118,65,80,80,76,69,0,103,108,66,105,110,100,86,101,114,116,101,120,65,114,114,97,121,65,80,80,76,69,0,103,108,68,101,108,101,116,101,86,101,114,116,101,120,65,114,114,97,121,115,65,80,80,76,69,0,103,108,71,101,110,86,101,114,116,101,120,65,114,114,97,121,115,65,80,80,76,69,0,103,108,73,115,86,101,114,116,101,120,65,114,114,97,121,65,80,80,76,69,0,103,108,86,101,114,116,101,120,65,114,114,97,121,82,97,110,103,101,65,80,80,76,69,0,103,108,70,108,117,115,104,86,101,114,116,101,120,65,114,114,97,121,82,97,110,103,101,65,80,80,76,69,0,103,108,86,101,114,116,101,120,65,114,114,97,121,80,97,114,97,109,101,116,101,114,105,65,80,80,76,69,0,103,108,69,110,97,98,108,101,86,101,114,116,101,120,65,116,116,114,105,98,65,80,80,76,69,0,103,108,68,105,115,97,98,108,101,86,101,114,116,101,120,65,116,116,114,105,98,65,80,80,76,69,0,103,108,73,115,86,101,114,116,101,120,65,116,116,114,105,98,69,110,97,98,108,101,100,65,80,80,76,69,0,103,108,77,97,112,86,101,114,116,101,120,65,116,116,114,105,98,49,100,65,80,80,76,69,0,103,108,77,97,112,86,101,114,116,101,120,65,116,116,114,105,98,49,102,65,80,80,76,69,0,103,108,77,97,112,86,101,114,116,101,120,65,116,116,114,105,98,50,100,65,80,80,76,69,0,103,108,77,97,112,86,101,114,116,101,120,65,116,116,114,105,98,50,102,65,80,80,76,69,0,103,108,70,114,97,109,101,84,101,114,109,105,110,97,116,111,114,71,82,69,77,69,68,89,0,103,108,83,116,114,105,110,103,77,97,114,107,101,114,71,82,69,77,69,68,89,0,103,108,82,101,110,100,101,114,98,117,102,102,101,114,83,116,111,114,97,103,101,77,117,108,116,105,115,97,109,112,108,101,73,77,71,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,50,68,77,117,108,116,105,115,97,109,112,108,101,73,77,71,0,103,108,66,108,101,110,100,70,117,110,99,83,101,112,97,114,97,116,101,73,78,71,82,0,103,108,65,112,112,108,121,70,114,97,109,101,98,117,102,102,101,114,65,116,116,97,99,104,109,101,110,116,67,77,65,65,73,78,84,69,76,0,103,108,83,121,110,99,84,101,120,116,117,114,101,73,78,84,69,76,0,103,108,85,110,109,97,112,84,101,120,116,117,114,101,50,68,73,78,84,69,76,0,103,108,77,97,112,84,101,120,116,117,114,101,50,68,73,78,84,69,76,0,103,108,66,101,103,105,110,80,101,114,102,81,117,101,114,121,73,78,84,69,76,0,103,108,67,114,101,97,116,101,80,101,114,102,81,117,101,114,121,73,78,84,69,76,0,103,108,68,101,108,101,116,101,80,101,114,102,81,117,101,114,121,73,78,84,69,76,0,103,108,69,110,100,80,101,114,102,81,117,101,114,121,73,78,84,69,76,0,103,108,71,101,116,70,105,114,115,116,80,101,114,102,81,117,101,114,121,73,100,73,78,84,69,76,0,103,108,71,101,116,78,101,120,116,80,101,114,102,81,117,101,114,121,73,100,73,78,84,69,76,0,103,108,71,101,116,80,101,114,102,67,111,117,110,116,101,114,73,110,102,111,73,78,84,69,76,0,103,108,71,101,116,80,101,114,102,81,117,101,114,121,68,97,116,97,73,78,84,69,76,0,103,108,71,101,116,80,101,114,102,81,117,101,114,121,73,100,66,121,78,97,109,101,73,78,84,69,76,0,103,108,71,101,116,80,101,114,102,81,117,101,114,121,73,110,102,111,73,78,84,69,76,0,103,108,77,117,108,116,105,68,114,97,119,65,114,114,97,121,115,73,110,100,105,114,101,99,116,66,105,110,100,108,101,115,115,78,86,0,103,108,77,117,108,116,105,68,114,97,119,69,108,101,109,101,110,116,115,73,110,100,105,114,101,99,116,66,105,110,100,108,101,115,115,78,86,0,103,108,77,117,108,116,105,68,114,97,119,65,114,114,97,121,115,73,110,100,105,114,101,99,116,66,105,110,100,108,101,115,115,67,111,117,110,116,78,86,0,103,108,77,117,108,116,105,68,114,97,119,69,108,101,109,101,110,116,115,73,110,100,105,114,101,99,116,66,105,110,100,108,101,115,115,67,111,117,110,116,78,86,0,103,108,71,101,116,84,101,120,116,117,114,101,72,97,110,100,108,101,78,86,0,103,108,71,101,116,84,101,120,116,117,114,101,83,97,109,112,108,101,114,72,97,110,100,108,101,78,86,0,103,108,77,97,107,101,84,101,120,116,117,114,101,72,97,110,100,108,101,82,101,115,105,100,101,110,116,78,86,0,103,108,77,97,107,101,84,101,120,116,117,114,101,72,97,110,100,108,101,78,111,110,82,101,115,105,100,101,110,116,78,86,0,103,108,71,101,116,73,109,97,103,101,72,97,110,100,108,101,78,86,0,103,108,77,97,107,101,73,109,97,103,101,72,97,110,100,108,101,82,101,115,105,100,101,110,116,78,86,0,103,108,77,97,107,101,73,109,97,103,101,72,97,110,100,108,101,78,111,110,82,101,115,105,100,101,110,116,78,86,0,103,108,85,110,105,102,111,114,109,72,97,110,100,108,101,117,105,54,52,78,86,0,103,108,85,110,105,102,111,114,109,72,97,110,100,108,101,117,105,54,52,118,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,72,97,110,100,108,101,117,105,54,52,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,72,97,110,100,108,101,117,105,54,52,118,78,86,0,103,108,73,115,84,101,120,116,117,114,101,72,97,110,100,108,101,82,101,115,105,100,101,110,116,78,86,0,103,108,73,115,73,109,97,103,101,72,97,110,100,108,101,82,101,115,105,100,101,110,116,78,86,0,103,108,66,108,101,110,100,80,97,114,97,109,101,116,101,114,105,78,86,0,103,108,66,108,101,110,100,66,97,114,114,105,101,114,78,86,0,103,108,67,114,101,97,116,101,83,116,97,116,101,115,78,86,0,103,108,68,101,108,101,116,101,83,116,97,116,101,115,78,86,0,103,108,73,115,83,116,97,116,101,78,86,0,103,108,83,116,97,116,101,67,97,112,116,117,114,101,78,86,0,103,108,71,101,116,67,111,109,109,97,110,100,72,101,97,100,101,114,78,86,0,103,108,71,101,116,83,116,97,103,101,73,110,100,101,120,78,86,0,103,108,68,114,97,119,67,111,109,109,97,110,100,115,78,86,0,103,108,68,114,97,119,67,111,109,109,97,110,100,115,65,100,100,114,101,115,115,78,86,0,103,108,68,114,97,119,67,111,109,109,97,110,100,115,83,116,97,116,101,115,78,86,0,103,108,68,114,97,119,67,111,109,109,97,110,100,115,83,116,97,116,101,115,65,100,100,114,101,115,115,78,86,0,103,108,67,114,101,97,116,101,67,111,109,109,97,110,100,76,105,115,116,115,78,86,0,103,108,68,101,108,101,116,101,67,111,109,109,97,110,100,76,105,115,116,115,78,86,0,103,108,73,115,67,111,109,109,97,110,100,76,105,115,116,78,86,0,103,108,76,105,115,116,68,114,97,119,67,111,109,109,97,110,100,115,83,116,97,116,101,115,67,108,105,101,110,116,78,86,0,103,108,67,111,109,109,97,110,100,76,105,115,116,83,101,103,109,101,110,116,115,78,86,0,103,108,67,111,109,112,105,108,101,67,111,109,109,97,110,100,76,105,115,116,78,86,0,103,108,67,97,108,108,67,111,109,109,97,110,100,76,105,115,116,78,86,0,103,108,66,101,103,105,110,67,111,110,100,105,116,105,111,110,97,108,82,101,110,100,101,114,78,86,0,103,108,69,110,100,67,111,110,100,105,116,105,111,110,97,108,82,101,110,100,101,114,78,86,0,103,108,83,117,98,112,105,120,101,108,80,114,101,99,105,115,105,111,110,66,105,97,115,78,86,0,103,108,67,111,110,115,101,114,118,97,116,105,118,101,82,97,115,116,101,114,80,97,114,97,109,101,116,101,114,102,78,86,0,103,108,67,111,112,121,66,117,102,102,101,114,83,117,98,68,97,116,97,78,86,0,103,108,67,111,112,121,73,109,97,103,101,83,117,98,68,97,116,97,78,86,0,103,108,67,111,118,101,114,97,103,101,77,97,115,107,78,86,0,103,108,67,111,118,101,114,97,103,101,79,112,101,114,97,116,105,111,110,78,86,0,103,108,68,101,112,116,104,82,97,110,103,101,100,78,86,0,103,108,67,108,101,97,114,68,101,112,116,104,100,78,86,0,103,108,68,101,112,116,104,66,111,117,110,100,115,100,78,86,0,103,108,68,114,97,119,66,117,102,102,101,114,115,78,86,0,103,108,68,114,97,119,65,114,114,97,121,115,73,110,115,116,97,110,99,101,100,78,86,0,103,108,68,114,97,119,69,108,101,109,101,110,116,115,73,110,115,116,97,110,99,101,100,78,86,0,103,108,68,114,97,119,84,101,120,116,117,114,101,78,86,0,103,108,71,101,116,77,117,108,116,105,115,97,109,112,108,101,102,118,78,86,0,103,108,83,97,109,112,108,101,77,97,115,107,73,110,100,101,120,101,100,78,86,0,103,108,84,101,120,82,101,110,100,101,114,98,117,102,102,101,114,78,86,0,103,108,68,101,108,101,116,101,70,101,110,99,101,115,78,86,0,103,108,71,101,110,70,101,110,99,101,115,78,86,0,103,108,73,115,70,101,110,99,101,78,86,0,103,108,84,101,115,116,70,101,110,99,101,78,86,0,103,108,71,101,116,70,101,110,99,101,105,118,78,86,0,103,108,70,105,110,105,115,104,70,101,110,99,101,78,86,0,103,108,83,101,116,70,101,110,99,101,78,86,0,103,108,70,114,97,103,109,101,110,116,67,111,118,101,114,97,103,101,67,111,108,111,114,78,86,0,103,108,66,108,105,116,70,114,97,109,101,98,117,102,102,101,114,78,86,0,103,108,67,111,118,101,114,97,103,101,77,111,100,117,108,97,116,105,111,110,84,97,98,108,101,78,86,0,103,108,71,101,116,67,111,118,101,114,97,103,101,77,111,100,117,108,97,116,105,111,110,84,97,98,108,101,78,86,0,103,108,67,111,118,101,114,97,103,101,77,111,100,117,108,97,116,105,111,110,78,86,0,103,108,82,101,110,100,101,114,98,117,102,102,101,114,83,116,111,114,97,103,101,77,117,108,116,105,115,97,109,112,108,101,78,86,0,103,108,82,101,110,100,101,114,98,117,102,102,101,114,83,116,111,114,97,103,101,77,117,108,116,105,115,97,109,112,108,101,67,111,118,101,114,97,103,101,78,86,0,103,108,80,114,111,103,114,97,109,86,101,114,116,101,120,76,105,109,105,116,78,86,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,70,97,99,101,69,88,84,0,103,108,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,73,52,105,78,86,0,103,108,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,73,52,105,118,78,86,0,103,108,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,115,73,52,105,118,78,86,0,103,108,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,73,52,117,105,78,86,0,103,108,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,73,52,117,105,118,78,86,0,103,108,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,115,73,52,117,105,118,78,86,0,103,108,80,114,111,103,114,97,109,69,110,118,80,97,114,97,109,101,116,101,114,73,52,105,78,86,0,103,108,80,114,111,103,114,97,109,69,110,118,80,97,114,97,109,101,116,101,114,73,52,105,118,78,86,0,103,108,80,114,111,103,114,97,109,69,110,118,80,97,114,97,109,101,116,101,114,115,73,52,105,118,78,86,0,103,108,80,114,111,103,114,97,109,69,110,118,80,97,114,97,109,101,116,101,114,73,52,117,105,78,86,0,103,108,80,114,111,103,114,97,109,69,110,118,80,97,114,97,109,101,116,101,114,73,52,117,105,118,78,86,0,103,108,80,114,111,103,114,97,109,69,110,118,80,97,114,97,109,101,116,101,114,115,73,52,117,105,118,78,86,0,103,108,71,101,116,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,73,105,118,78,86,0,103,108,71,101,116,80,114,111,103,114,97,109,76,111,99,97,108,80,97,114,97,109,101,116,101,114,73,117,105,118,78,86,0,103,108,71,101,116,80,114,111,103,114,97,109,69,110,118,80,97,114,97,109,101,116,101,114,73,105,118,78,86,0,103,108,71,101,116,80,114,111,103,114,97,109,69,110,118,80,97,114,97,109,101,116,101,114,73,117,105,118,78,86,0,103,108,80,114,111,103,114,97,109,83,117,98,114,111,117,116,105,110,101,80,97,114,97,109,101,116,101,114,115,117,105,118,78,86,0,103,108,71,101,116,80,114,111,103,114,97,109,83,117,98,114,111,117,116,105,110,101,80,97,114,97,109,101,116,101,114,117,105,118,78,86,0,103,108,86,101,114,116,101,120,50,104,78,86,0,103,108,86,101,114,116,101,120,50,104,118,78,86,0,103,108,86,101,114,116,101,120,51,104,78,86,0,103,108,86,101,114,116,101,120,51,104,118,78,86,0,103,108,86,101,114,116,101,120,52,104,78,86,0,103,108,86,101,114,116,101,120,52,104,118,78,86,0,103,108,78,111,114,109,97,108,51,104,78,86,0,103,108,78,111,114,109,97,108,51,104,118,78,86,0,103,108,67,111,108,111,114,51,104,78,86,0,103,108,67,111,108,111,114,51,104,118,78,86,0,103,108,67,111,108,111,114,52,104,78,86,0,103,108,67,111,108,111,114,52,104,118,78,86,0,103,108,84,101,120,67,111,111,114,100,49,104,78,86,0,103,108,84,101,120,67,111,111,114,100,49,104,118,78,86,0,103,108,84,101,120,67,111,111,114,100,50,104,78,86,0,103,108,84,101,120,67,111,111,114,100,50,104,118,78,86,0,103,108,84,101,120,67,111,111,114,100,51,104,78,86,0,103,108,84,101,120,67,111,111,114,100,51,104,118,78,86,0,103,108,84,101,120,67,111,111,114,100,52,104,78,86,0,103,108,84,101,120,67,111,111,114,100,52,104,118,78,86,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,49,104,78,86,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,49,104,118,78,86,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,50,104,78,86,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,50,104,118,78,86,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,51,104,78,86,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,51,104,118,78,86,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,52,104,78,86,0,103,108,77,117,108,116,105,84,101,120,67,111,111,114,100,52,104,118,78,86,0,103,108,70,111,103,67,111,111,114,100,104,78,86,0,103,108,70,111,103,67,111,111,114,100,104,118,78,86,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,104,78,86,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,51,104,118,78,86,0,103,108,86,101,114,116,101,120,87,101,105,103,104,116,104,78,86,0,103,108,86,101,114,116,101,120,87,101,105,103,104,116,104,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,49,104,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,49,104,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,50,104,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,50,104,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,51,104,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,51,104,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,104,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,104,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,115,49,104,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,115,50,104,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,115,51,104,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,115,52,104,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,68,105,118,105,115,111,114,78,86,0,103,108,71,101,116,73,110,116,101,114,110,97,108,102,111,114,109,97,116,83,97,109,112,108,101,105,118,78,86,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,50,120,51,102,118,78,86,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,51,120,50,102,118,78,86,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,50,120,52,102,118,78,86,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,52,120,50,102,118,78,86,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,51,120,52,102,118,78,86,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,52,120,51,102,118,78,86,0,103,108,71,101,110,79,99,99,108,117,115,105,111,110,81,117,101,114,105,101,115,78,86,0,103,108,68,101,108,101,116,101,79,99,99,108,117,115,105,111,110,81,117,101,114,105,101,115,78,86,0,103,108,73,115,79,99,99,108,117,115,105,111,110,81,117,101,114,121,78,86,0,103,108,66,101,103,105,110,79,99,99,108,117,115,105,111,110,81,117,101,114,121,78,86,0,103,108,69,110,100,79,99,99,108,117,115,105,111,110,81,117,101,114,121,78,86,0,103,108,71,101,116,79,99,99,108,117,115,105,111,110,81,117,101,114,121,105,118,78,86,0,103,108,71,101,116,79,99,99,108,117,115,105,111,110,81,117,101,114,121,117,105,118,78,86,0,103,108,80,114,111,103,114,97,109,66,117,102,102,101,114,80,97,114,97,109,101,116,101,114,115,102,118,78,86,0,103,108,80,114,111,103,114,97,109,66,117,102,102,101,114,80,97,114,97,109,101,116,101,114,115,73,105,118,78,86,0,103,108,80,114,111,103,114,97,109,66,117,102,102,101,114,80,97,114,97,109,101,116,101,114,115,73,117,105,118,78,86,0,103,108,71,101,110,80,97,116,104,115,78,86,0,103,108,68,101,108,101,116,101,80,97,116,104,115,78,86,0,103,108,73,115,80,97,116,104,78,86,0,103,108,80,97,116,104,67,111,109,109,97,110,100,115,78,86,0,103,108,80,97,116,104,67,111,111,114,100,115,78,86,0,103,108,80,97,116,104,83,117,98,67,111,109,109,97,110,100,115,78,86,0,103,108,80,97,116,104,83,117,98,67,111,111,114,100,115,78,86,0,103,108,80,97,116,104,83,116,114,105,110,103,78,86,0,103,108,80,97,116,104,71,108,121,112,104,115,78,86,0,103,108,80,97,116,104,71,108,121,112,104,82,97,110,103,101,78,86,0,103,108,87,101,105,103,104,116,80,97,116,104,115,78,86,0,103,108,67,111,112,121,80,97,116,104,78,86,0,103,108,73,110,116,101,114,112,111,108,97,116,101,80,97,116,104,115,78,86,0,103,108,84,114,97,110,115,102,111,114,109,80,97,116,104,78,86,0,103,108,80,97,116,104,80,97,114,97,109,101,116,101,114,105,118,78,86,0,103,108,80,97,116,104,80,97,114,97,109,101,116,101,114,105,78,86,0,103,108,80,97,116,104,80,97,114,97,109,101,116,101,114,102,118,78,86,0,103,108,80,97,116,104,80,97,114,97,109,101,116,101,114,102,78,86,0,103,108,80,97,116,104,68,97,115,104,65,114,114,97,121,78,86,0,103,108,80,97,116,104,83,116,101,110,99,105,108,70,117,110,99,78,86,0,103,108,80,97,116,104,83,116,101,110,99,105,108,68,101,112,116,104,79,102,102,115,101,116,78,86,0,103,108,83,116,101,110,99,105,108,70,105,108,108,80,97,116,104,78,86,0,103,108,83,116,101,110,99,105,108,83,116,114,111,107,101,80,97,116,104,78,86,0,103,108,83,116,101,110,99,105,108,70,105,108,108,80,97,116,104,73,110,115,116,97,110,99,101,100,78,86,0,103,108,83,116,101,110,99,105,108,83,116,114,111,107,101,80,97,116,104,73,110,115,116,97,110,99,101,100,78,86,0,103,108,80,97,116,104,67,111,118,101,114,68,101,112,116,104,70,117,110,99,78,86,0,103,108,67,111,118,101,114,70,105,108,108,80,97,116,104,78,86,0,103,108,67,111,118,101,114,83,116,114,111,107,101,80,97,116,104,78,86,0,103,108,67,111,118,101,114,70,105,108,108,80,97,116,104,73,110,115,116,97,110,99,101,100,78,86,0,103,108,67,111,118,101,114,83,116,114,111,107,101,80,97,116,104,73,110,115,116,97,110,99,101,100,78,86,0,103,108,71,101,116,80,97,116,104,80,97,114,97,109,101,116,101,114,105,118,78,86,0,103,108,71,101,116,80,97,116,104,80,97,114,97,109,101,116,101,114,102,118,78,86,0,103,108,71,101,116,80,97,116,104,67,111,109,109,97,110,100,115,78,86,0,103,108,71,101,116,80,97,116,104,67,111,111,114,100,115,78,86,0,103,108,71,101,116,80,97,116,104,68,97,115,104,65,114,114,97,121,78,86,0,103,108,71,101,116,80,97,116,104,77,101,116,114,105,99,115,78,86,0,103,108,71,101,116,80,97,116,104,77,101,116,114,105,99,82,97,110,103,101,78,86,0,103,108,71,101,116,80,97,116,104,83,112,97,99,105,110,103,78,86,0,103,108,73,115,80,111,105,110,116,73,110,70,105,108,108,80,97,116,104,78,86,0,103,108,73,115,80,111,105,110,116,73,110,83,116,114,111,107,101,80,97,116,104,78,86,0,103,108,71,101,116,80,97,116,104,76,101,110,103,116,104,78,86,0,103,108,80,111,105,110,116,65,108,111,110,103,80,97,116,104,78,86,0,103,108,77,97,116,114,105,120,76,111,97,100,51,120,50,102,78,86,0,103,108,77,97,116,114,105,120,76,111,97,100,51,120,51,102,78,86,0,103,108,77,97,116,114,105,120,76,111,97,100,84,114,97,110,115,112,111,115,101,51,120,51,102,78,86,0,103,108,77,97,116,114,105,120,77,117,108,116,51,120,50,102,78,86,0,103,108,77,97,116,114,105,120,77,117,108,116,51,120,51,102,78,86,0,103,108,77,97,116,114,105,120,77,117,108,116,84,114,97,110,115,112,111,115,101,51,120,51,102,78,86,0,103,108,83,116,101,110,99,105,108,84,104,101,110,67,111,118,101,114,70,105,108,108,80,97,116,104,78,86,0,103,108,83,116,101,110,99,105,108,84,104,101,110,67,111,118,101,114,83,116,114,111,107,101,80,97,116,104,78,86,0,103,108,83,116,101,110,99,105,108,84,104,101,110,67,111,118,101,114,70,105,108,108,80,97,116,104,73,110,115,116,97,110,99,101,100,78,86,0,103,108,83,116,101,110,99,105,108,84,104,101,110,67,111,118,101,114,83,116,114,111,107,101,80,97,116,104,73,110,115,116,97,110,99,101,100,78,86,0,103,108,80,97,116,104,71,108,121,112,104,73,110,100,101,120,82,97,110,103,101,78,86,0,103,108,80,97,116,104,71,108,121,112,104,73,110,100,101,120,65,114,114,97,121,78,86,0,103,108,80,97,116,104,77,101,109,111,114,121,71,108,121,112,104,73,110,100,101,120,65,114,114,97,121,78,86,0,103,108,80,114,111,103,114,97,109,80,97,116,104,70,114,97,103,109,101,110,116,73,110,112,117,116,71,101,110,78,86,0,103,108,71,101,116,80,114,111,103,114,97,109,82,101,115,111,117,114,99,101,102,118,78,86,0,103,108,80,97,116,104,67,111,108,111,114,71,101,110,78,86,0,103,108,80,97,116,104,84,101,120,71,101,110,78,86,0,103,108,80,97,116,104,70,111,103,71,101,110,78,86,0,103,108,71,101,116,80,97,116,104,67,111,108,111,114,71,101,110,105,118,78,86,0,103,108,71,101,116,80,97,116,104,67,111,108,111,114,71,101,110,102,118,78,86,0,103,108,71,101,116,80,97,116,104,84,101,120,71,101,110,105,118,78,86,0,103,108,71,101,116,80,97,116,104,84,101,120,71,101,110,102,118,78,86,0,103,108,80,111,108,121,103,111,110,77,111,100,101,78,86,0,103,108,80,114,101,115,101,110,116,70,114,97,109,101,75,101,121,101,100,78,86,0,103,108,80,114,101,115,101,110,116,70,114,97,109,101,68,117,97,108,70,105,108,108,78,86,0,103,108,71,101,116,86,105,100,101,111,105,118,78,86,0,103,108,71,101,116,86,105,100,101,111,117,105,118,78,86,0,103,108,71,101,116,86,105,100,101,111,105,54,52,118,78,86,0,103,108,71,101,116,86,105,100,101,111,117,105,54,52,118,78,86,0,103,108,80,114,105,109,105,116,105,118,101,82,101,115,116,97,114,116,78,86,0,103,108,80,114,105,109,105,116,105,118,101,82,101,115,116,97,114,116,73,110,100,101,120,78,86,0,103,108,82,101,97,100,66,117,102,102,101,114,78,86,0,103,108,70,114,97,109,101,98,117,102,102,101,114,83,97,109,112,108,101,76,111,99,97,116,105,111,110,115,102,118,78,86,0,103,108,78,97,109,101,100,70,114,97,109,101,98,117,102,102,101,114,83,97,109,112,108,101,76,111,99,97,116,105,111,110,115,102,118,78,86,0,103,108,82,101,115,111,108,118,101,68,101,112,116,104,86,97,108,117,101,115,78,86,0,103,108,77,97,107,101,66,117,102,102,101,114,82,101,115,105,100,101,110,116,78,86,0,103,108,77,97,107,101,66,117,102,102,101,114,78,111,110,82,101,115,105,100,101,110,116,78,86,0,103,108,73,115,66,117,102,102,101,114,82,101,115,105,100,101,110,116,78,86,0,103,108,77,97,107,101,78,97,109,101,100,66,117,102,102,101,114,82,101,115,105,100,101,110,116,78,86,0,103,108,77,97,107,101,78,97,109,101,100,66,117,102,102,101,114,78,111,110,82,101,115,105,100,101,110,116,78,86,0,103,108,73,115,78,97,109,101,100,66,117,102,102,101,114,82,101,115,105,100,101,110,116,78,86,0,103,108,71,101,116,66,117,102,102,101,114,80,97,114,97,109,101,116,101,114,117,105,54,52,118,78,86,0,103,108,71,101,116,78,97,109,101,100,66,117,102,102,101,114,80,97,114,97,109,101,116,101,114,117,105,54,52,118,78,86,0,103,108,71,101,116,73,110,116,101,103,101,114,117,105,54,52,118,78,86,0,103,108,85,110,105,102,111,114,109,117,105,54,52,78,86,0,103,108,85,110,105,102,111,114,109,117,105,54,52,118,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,117,105,54,52,78,86,0,103,108,80,114,111,103,114,97,109,85,110,105,102,111,114,109,117,105,54,52,118,78,86,0,103,108,84,101,120,116,117,114,101,66,97,114,114,105,101,114,78,86,0,103,108,84,101,120,73,109,97,103,101,50,68,77,117,108,116,105,115,97,109,112,108,101,67,111,118,101,114,97,103,101,78,86,0,103,108,84,101,120,73,109,97,103,101,51,68,77,117,108,116,105,115,97,109,112,108,101,67,111,118,101,114,97,103,101,78,86,0,103,108,84,101,120,116,117,114,101,73,109,97,103,101,50,68,77,117,108,116,105,115,97,109,112,108,101,78,86,0,103,108,84,101,120,116,117,114,101,73,109,97,103,101,51,68,77,117,108,116,105,115,97,109,112,108,101,78,86,0,103,108,84,101,120,116,117,114,101,73,109,97,103,101,50,68,77,117,108,116,105,115,97,109,112,108,101,67,111,118,101,114,97,103,101,78,86,0,103,108,84,101,120,116,117,114,101,73,109,97,103,101,51,68,77,117,108,116,105,115,97,109,112,108,101,67,111,118,101,114,97,103,101,78,86,0,103,108,66,101,103,105,110,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,78,86,0,103,108,69,110,100,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,78,86,0,103,108,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,65,116,116,114,105,98,115,78,86,0,103,108,66,105,110,100,66,117,102,102,101,114,82,97,110,103,101,78,86,0,103,108,66,105,110,100,66,117,102,102,101,114,79,102,102,115,101,116,78,86,0,103,108,66,105,110,100,66,117,102,102,101,114,66,97,115,101,78,86,0,103,108,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,86,97,114,121,105,110,103,115,78,86,0,103,108,65,99,116,105,118,101,86,97,114,121,105,110,103,78,86,0,103,108,71,101,116,86,97,114,121,105,110,103,76,111,99,97,116,105,111,110,78,86,0,103,108,71,101,116,65,99,116,105,118,101,86,97,114,121,105,110,103,78,86,0,103,108,71,101,116,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,86,97,114,121,105,110,103,78,86,0,103,108,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,83,116,114,101,97,109,65,116,116,114,105,98,115,78,86,0,103,108,66,105,110,100,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,78,86,0,103,108,68,101,108,101,116,101,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,115,78,86,0,103,108,71,101,110,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,115,78,86,0,103,108,73,115,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,78,86,0,103,108,80,97,117,115,101,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,78,86,0,103,108,82,101,115,117,109,101,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,78,86,0,103,108,68,114,97,119,84,114,97,110,115,102,111,114,109,70,101,101,100,98,97,99,107,78,86,0,103,108,86,68,80,65,85,73,110,105,116,78,86,0,103,108,86,68,80,65,85,70,105,110,105,78,86,0,103,108,86,68,80,65,85,82,101,103,105,115,116,101,114,86,105,100,101,111,83,117,114,102,97,99,101,78,86,0,103,108,86,68,80,65,85,82,101,103,105,115,116,101,114,79,117,116,112,117,116,83,117,114,102,97,99,101,78,86,0,103,108,86,68,80,65,85,73,115,83,117,114,102,97,99,101,78,86,0,103,108,86,68,80,65,85,85,110,114,101,103,105,115,116,101,114,83,117,114,102,97,99,101,78,86,0,103,108,86,68,80,65,85,71,101,116,83,117,114,102,97,99,101,105,118,78,86,0,103,108,86,68,80,65,85,83,117,114,102,97,99,101,65,99,99,101,115,115,78,86,0,103,108,86,68,80,65,85,77,97,112,83,117,114,102,97,99,101,115,78,86,0,103,108,86,68,80,65,85,85,110,109,97,112,83,117,114,102,97,99,101,115,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,49,105,54,52,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,50,105,54,52,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,51,105,54,52,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,52,105,54,52,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,49,105,54,52,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,50,105,54,52,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,51,105,54,52,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,52,105,54,52,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,49,117,105,54,52,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,50,117,105,54,52,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,51,117,105,54,52,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,52,117,105,54,52,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,49,117,105,54,52,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,50,117,105,54,52,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,51,117,105,54,52,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,52,117,105,54,52,118,78,86,0,103,108,71,101,116,86,101,114,116,101,120,65,116,116,114,105,98,76,105,54,52,118,78,86,0,103,108,71,101,116,86,101,114,116,101,120,65,116,116,114,105,98,76,117,105,54,52,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,76,70,111,114,109,97,116,78,86,0,103,108,66,117,102,102,101,114,65,100,100,114,101,115,115,82,97,110,103,101,78,86,0,103,108,86,101,114,116,101,120,70,111,114,109,97,116,78,86,0,103,108,78,111,114,109,97,108,70,111,114,109,97,116,78,86,0,103,108,67,111,108,111,114,70,111,114,109,97,116,78,86,0,103,108,73,110,100,101,120,70,111,114,109,97,116,78,86,0,103,108,84,101,120,67,111,111,114,100,70,111,114,109,97,116,78,86,0,103,108,69,100,103,101,70,108,97,103,70,111,114,109,97,116,78,86,0,103,108,83,101,99,111,110,100,97,114,121,67,111,108,111,114,70,111,114,109,97,116,78,86,0,103,108,70,111,103,67,111,111,114,100,70,111,114,109,97,116,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,70,111,114,109,97,116,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,70,111,114,109,97,116,78,86,0,103,108,71,101,116,73,110,116,101,103,101,114,117,105,54,52,105,95,118,78,86,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,49,105,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,50,105,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,51,105,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,52,105,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,49,117,105,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,50,117,105,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,51,117,105,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,52,117,105,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,49,105,118,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,50,105,118,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,51,105,118,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,52,105,118,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,49,117,105,118,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,50,117,105,118,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,51,117,105,118,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,52,117,105,118,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,52,98,118,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,52,115,118,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,52,117,98,118,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,52,117,115,118,69,88,84,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,73,80,111,105,110,116,101,114,69,88,84,0,103,108,71,101,116,86,101,114,116,101,120,65,116,116,114,105,98,73,105,118,69,88,84,0,103,108,71,101,116,86,101,114,116,101,120,65,116,116,114,105,98,73,117,105,118,69,88,84,0,103,108,66,101,103,105,110,86,105,100,101,111,67,97,112,116,117,114,101,78,86,0,103,108,66,105,110,100,86,105,100,101,111,67,97,112,116,117,114,101,83,116,114,101,97,109,66,117,102,102,101,114,78,86,0,103,108,66,105,110,100,86,105,100,101,111,67,97,112,116,117,114,101,83,116,114,101,97,109,84,101,120,116,117,114,101,78,86,0,103,108,69,110,100,86,105,100,101,111,67,97,112,116,117,114,101,78,86,0,103,108,71,101,116,86,105,100,101,111,67,97,112,116,117,114,101,105,118,78,86,0,103,108,71,101,116,86,105,100,101,111,67,97,112,116,117,114,101,83,116,114,101,97,109,105,118,78,86,0,103,108,71,101,116,86,105,100,101,111,67,97,112,116,117,114,101,83,116,114,101,97,109,102,118,78,86,0,103,108,71,101,116,86,105,100,101,111,67,97,112,116,117,114,101,83,116,114,101,97,109,100,118,78,86,0,103,108,86,105,100,101,111,67,97,112,116,117,114,101,78,86,0,103,108,86,105,100,101,111,67,97,112,116,117,114,101,83,116,114,101,97,109,80,97,114,97,109,101,116,101,114,105,118,78,86,0,103,108,86,105,100,101,111,67,97,112,116,117,114,101,83,116,114,101,97,109,80,97,114,97,109,101,116,101,114,102,118,78,86,0,103,108,86,105,100,101,111,67,97,112,116,117,114,101,83,116,114,101,97,109,80,97,114,97,109,101,116,101,114,100,118,78,86,0,103,108,86,105,101,119,112,111,114,116,65,114,114,97,121,118,78,86,0,103,108,86,105,101,119,112,111,114,116,73,110,100,101,120,101,100,102,78,86,0,103,108,86,105,101,119,112,111,114,116,73,110,100,101,120,101,100,102,118,78,86,0,103,108,83,99,105,115,115,111,114,65,114,114,97,121,118,78,86,0,103,108,83,99,105,115,115,111,114,73,110,100,101,120,101,100,78,86,0,103,108,83,99,105,115,115,111,114,73,110,100,101,120,101,100,118,78,86,0,103,108,68,101,112,116,104,82,97,110,103,101,65,114,114,97,121,102,118,78,86,0,103,108,68,101,112,116,104,82,97,110,103,101,73,110,100,101,120,101,100,102,78,86,0,103,108,71,101,116,70,108,111,97,116,105,95,118,78,86,0,103,108,69,110,97,98,108,101,105,78,86,0,103,108,68,105,115,97,98,108,101,105,78,86,0,103,108,73,115,69,110,97,98,108,101,100,105,78,86,0,103,108,66,101,103,105,110,67,111,110,100,105,116,105,111,110,97,108,82,101,110,100,101,114,78,86,88,0,103,108,69,110,100,67,111,110,100,105,116,105,111,110,97,108,82,101,110,100,101,114,78,86,88,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,77,117,108,116,105,118,105,101,119,79,86,82,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,77,117,108,116,105,115,97,109,112,108,101,77,117,108,116,105,118,105,101,119,79,86,82,0,103,108,65,108,112,104,97,70,117,110,99,81,67,79,77,0,103,108,71,101,116,68,114,105,118,101,114,67,111,110,116,114,111,108,115,81,67,79,77,0,103,108,71,101,116,68,114,105,118,101,114,67,111,110,116,114,111,108,83,116,114,105,110,103,81,67,79,77,0,103,108,69,110,97,98,108,101,68,114,105,118,101,114,67,111,110,116,114,111,108,81,67,79,77,0,103,108,68,105,115,97,98,108,101,68,114,105,118,101,114,67,111,110,116,114,111,108,81,67,79,77,0,103,108,69,120,116,71,101,116,84,101,120,116,117,114,101,115,81,67,79,77,0,103,108,69,120,116,71,101,116,66,117,102,102,101,114,115,81,67,79,77,0,103,108,69,120,116,71,101,116,82,101,110,100,101,114,98,117,102,102,101,114,115,81,67,79,77,0,103,108,69,120,116,71,101,116,70,114,97,109,101,98,117,102,102,101,114,115,81,67,79,77,0,103,108,69,120,116,71,101,116,84,101,120,76,101,118,101,108,80,97,114,97,109,101,116,101,114,105,118,81,67,79,77,0,103,108,69,120,116,84,101,120,79,98,106,101,99,116,83,116,97,116,101,79,118,101,114,114,105,100,101,105,81,67,79,77,0,103,108,69,120,116,71,101,116,84,101,120,83,117,98,73,109,97,103,101,81,67,79,77,0,103,108,69,120,116,71,101,116,66,117,102,102,101,114,80,111,105,110,116,101,114,118,81,67,79,77,0,103,108,69,120,116,71,101,116,83,104,97,100,101,114,115,81,67,79,77,0,103,108,69,120,116,71,101,116,80,114,111,103,114,97,109,115,81,67,79,77,0,103,108,69,120,116,73,115,80,114,111,103,114,97,109,66,105,110,97,114,121,81,67,79,77,0,103,108,69,120,116,71,101,116,80,114,111,103,114,97,109,66,105,110,97,114,121,83,111,117,114,99,101,81,67,79,77,0,103,108,83,116,97,114,116,84,105,108,105,110,103,81,67,79,77,0,103,108,69,110,100,84,105,108,105,110,103,81,67,79,77,0,78,55,108,111,100,101,112,110,103,53,83,116,97,116,101,69,0,49,50,76,111,100,101,80,78,71,83,116,97,116,101,0,73,68,65,84,0,73,72,68,82,0,80,76,84,69,0,116,82,78,83,0,98,75,71,68,0,112,72,89,115,0,116,73,77,69,0,122,84,88,116,0,116,69,88,116,0,76,111,100,101,80,78,71,0,50,48,49,52,49,49,51,48,0,105,84,88,116,0,73,69,78,68,0,117,110,107,110,111,119,110,32,101,114,114,111,114,32,99,111,100,101,0,122,101,114,111,32,119,105,100,116,104,32,111,114,32,104,101,105,103,104,116,32,105,115,32,105,110,118,97,108,105,100,0,116,111,111,32,109,97,110,121,32,112,105,120,101,108,115,44,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,105,110,118,97,108,105,100,32,100,101,99,111,109,112,114,101,115,115,101,100,32,105,100,97,116,32,115,105,122,101,0,119,105,110,100,111,119,115,105,122,101,32,109,117,115,116,32,98,101,32,97,32,112,111,119,101,114,32,111,102,32,116,119,111,0,116,101,120,116,32,99,104,117,110,107,32,107,101,121,119,111,114,100,32,116,111,111,32,115,104,111,114,116,32,111,114,32,108,111,110,103,58,32,109,117,115,116,32,104,97,118,101,32,115,105,122,101,32,49,45,55,57,0,105,110,118,97,108,105,100,32,102,105,108,116,101,114,32,115,116,114,97,116,101,103,121,32,103,105,118,101,110,32,102,111,114,32,76,111,100,101,80,78,71,69,110,99,111,100,101,114,83,101,116,116,105,110,103,115,46,102,105,108,116,101,114,95,115,116,114,97,116,101,103,121,0,109,117,115,116,32,112,114,111,118,105,100,101,32,99,117,115,116,111,109,32,122,108,105,98,32,102,117,110,99,116,105,111,110,32,112,111,105,110,116,101,114,32,105,102,32,76,79,68,69,80,78,71,95,67,79,77,80,73,76,69,95,90,76,73,66,32,105,115,32,110,111,116,32,100,101,102,105,110,101,100,0,105,109,112,111,115,115,105,98,108,101,32,111,102,102,115,101,116,32,105,110,32,108,122,55,55,32,101,110,99,111,100,105,110,103,32,40,105,110,116,101,114,110,97,108,32,98,117,103,41,0,103,105,118,101,110,32,105,109,97,103,101,32,116,111,111,32,115,109,97,108,108,32,116,111,32,99,111,110,116,97,105,110,32,97,108,108,32,112,105,120,101,108,115,32,116,111,32,98,101,32,101,110,99,111,100,101,100,0,109,101,109,111,114,121,32,97,108,108,111,99,97,116,105,111,110,32,102,97,105,108,101,100,0,99,111,108,111,114,32,99,111,110,118,101,114,115,105,111,110,32,116,111,32,112,97,108,101,116,116,101,32,114,101,113,117,101,115,116,101,100,32,119,104,105,108,101,32,97,32,99,111,108,111,114,32,105,115,110,39,116,32,105,110,32,112,97,108,101,116,116,101,0,108,97,122,121,32,109,97,116,99,104,105,110,103,32,97,116,32,112,111,115,32,48,32,105,115,32,105,109,112,111,115,115,105,98,108,101,0,116,114,105,101,100,32,99,114,101,97,116,105,110,103,32,97,32,116,114,101,101,32,111,102,32,48,32,115,121,109,98,111,108,115,0,102,97,105,108,101,100,32,116,111,32,111,112,101,110,32,102,105,108,101,32,102,111,114,32,119,114,105,116,105,110,103,0,102,97,105,108,101,100,32,116,111,32,111,112,101,110,32,102,105,108,101,32,102,111,114,32,114,101,97,100,105,110,103,0,105,110,116,101,103,101,114,32,111,118,101,114,102,108,111,119,32,105,110,32,98,117,102,102,101,114,32,115,105,122,101,0,105,84,88,116,32,99,104,117,110,107,32,116,111,111,32,115,104,111,114,116,32,116,111,32,99,111,110,116,97,105,110,32,114,101,113,117,105,114,101,100,32,98,121,116,101,115,0,110,111,32,110,117,108,108,32,116,101,114,109,105,110,97,116,105,111,110,32,99,104,97,114,32,102,111,117,110,100,32,119,104,105,108,101,32,100,101,99,111,100,105,110,103,32,116,101,120,116,32,99,104,117,110,107,0,105,110,118,97,108,105,100,32,112,72,89,115,32,99,104,117,110,107,32,115,105,122,101,0,105,110,118,97,108,105,100,32,116,73,77,69,32,99,104,117,110,107,32,115,105,122,101,0,119,104,105,108,101,32,100,101,99,111,100,105,110,103,44,32,117,110,101,120,105,115,116,105,110,103,32,99,111,109,112,114,101,115,115,105,111,110,32,109,101,116,104,111,100,32,101,110,99,111,117,110,116,101,114,105,110,103,32,105,110,32,122,84,88,116,32,111,114,32,105,84,88,116,32,99,104,117,110,107,32,40,105,116,32,109,117,115,116,32,98,101,32,48,41,0,117,110,101,120,105,115,116,105,110,103,32,105,110,116,101,114,108,97,99,101,32,109,111,100,101,32,103,105,118,101,110,32,116,111,32,101,110,99,111,100,101,114,32,40,109,117,115,116,32,98,101,32,48,32,111,114,32,49,41,0,117,110,107,110,111,119,110,32,99,104,117,110,107,32,116,121,112,101,32,119,105,116,104,32,39,99,114,105,116,105,99,97,108,39,32,102,108,97,103,32,101,110,99,111,117,110,116,101,114,101,100,32,98,121,32,116,104,101,32,100,101,99,111,100,101,114,0,116,114,105,101,100,32,116,111,32,101,110,99,111,100,101,32,97,32,80,76,84,69,32,99,104,117,110,107,32,119,105,116,104,32,97,32,112,97,108,101,116,116,101,32,116,104,97,116,32,104,97,115,32,108,101,115,115,32,116,104,97,110,32,49,32,111,114,32,109,111,114,101,32,116,104,97,110,32,50,53,54,32,99,111,108,111,114,115,0,116,104,101,32,108,101,110,103,116,104,32,111,102,32,97,32,116,101,120,116,32,99,104,117,110,107,32,107,101,121,119,111,114,100,32,103,105,118,101,110,32,116,111,32,116,104,101,32,101,110,99,111,100,101,114,32,105,115,32,115,109,97,108,108,101,114,32,116,104,97,110,32,116,104,101,32,109,105,110,105,109,117,109,32,111,102,32,49,32,98,121,116,101,0,116,104,101,32,108,101,110,103,116,104,32,111,102,32,97,32,116,101,120,116,32,99,104,117,110,107,32,107,101,121,119,111,114,100,32,103,105,118,101,110,32,116,111,32,116,104,101,32,101,110,99,111,100,101,114,32,105,115,32,108,111,110,103,101,114,32,116,104,97,110,32,116,104,101,32,109,97,120,105,109,117,109,32,111,102,32,55,57,32,98,121,116,101,115,0,116,104,101,32,108,101,110,103,116,104,32,111,102,32,116,104,101,32,69,78,68,32,115,121,109,98,111,108,32,50,53,54,32,105,110,32,116,104,101,32,72,117,102,102,109,97,110,32,116,114,101,101,32,105,115,32,48,0,108,101,110,103,116,104,32,111,102,32,97,32,99,104,117,110,107,32,116,111,111,32,108,111,110,103,44,32,109,97,120,32,97,108,108,111,119,101,100,32,102,111,114,32,80,78,71,32,105,115,32,50,49,52,55,52,56,51,54,52,55,32,98,121,116,101,115,32,112,101,114,32,99,104,117,110,107,0,99,111,110,118,101,114,115,105,111,110,32,102,114,111,109,32,99,111,108,111,114,32,116,111,32,103,114,101,121,115,99,97,108,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,105,110,118,97,108,105,100,32,66,84,89,80,69,32,103,105,118,101,110,32,105,110,32,116,104,101,32,115,101,116,116,105,110,103,115,32,111,102,32,116,104,101,32,101,110,99,111,100,101,114,32,40,111,110,108,121,32,48,44,32,49,32,97,110,100,32,50,32,97,114,101,32,97,108,108,111,119,101,100,41,0,105,110,118,97,108,105,100,32,119,105,110,100,111,119,32,115,105,122,101,32,103,105,118,101,110,32,105,110,32,116,104,101,32,115,101,116,116,105,110,103,115,32,111,102,32,116,104,101,32,101,110,99,111,100,101,114,32,40,109,117,115,116,32,98,101,32,48,45,51,50,55,54,56,41,0,114,101,113,117,101,115,116,101,100,32,99,111,108,111,114,32,99,111,110,118,101,114,115,105,111,110,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,105,110,118,97,108,105,100,32,65,68,76,69,82,51,50,32,101,110,99,111,117,110,116,101,114,101,100,32,40,99,104,101,99,107,105,110,103,32,65,68,76,69,82,51,50,32,99,97,110,32,98,101,32,100,105,115,97,98,108,101,100,41,0,105,110,118,97,108,105,100,32,67,82,67,32,101,110,99,111,117,110,116,101,114,101,100,32,40,99,104,101,99,107,105,110,103,32,67,82,67,32,99,97,110,32,98,101,32,100,105,115,97,98,108,101,100,41,0,103,105,118,101,110,32,111,117,116,112,117,116,32,105,109,97,103,101,32,99,111,108,111,114,116,121,112,101,32,111,114,32,98,105,116,100,101,112,116,104,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,102,111,114,32,99,111,108,111,114,32,99,111,110,118,101,114,115,105,111,110,0,106,117,109,112,101,100,32,112,97,115,116,32,116,114,101,101,32,119,104,105,108,101,32,103,101,110,101,114,97,116,105,110,103,32,104,117,102,102,109,97,110,32,116,114,101,101,0,114,101,112,101,97,116,32,115,121,109,98,111,108,32,105,110,32,116,114,101,101,32,119,104,105,108,101,32,116,104,101,114,101,32,119,97,115,32,110,111,32,118,97,108,117,101,32,115,121,109,98,111,108,32,121,101,116,0,115,105,122,101,32,111,102,32,122,108,105,98,32,100,97,116,97,32,116,111,111,32,115,109,97,108,108,0,106,117,109,112,101,100,32,112,97,115,116,32,109,101,109,111,114,121,32,119,104,105,108,101,32,105,110,102,108,97,116,105,110,103,0,106,117,109,112,101,100,32,112,97,115,116,32,109,101,109,111,114,121,32,119,104,105,108,101,32,105,110,102,108,97,116,105,110,103,32,104,117,102,102,109,97,110,32,98,108,111,99,107,0,106,117,109,112,101,100,32,112,97,115,116,32,109,101,109,111,114,121,32,119,104,105,108,101,32,103,101,110,101,114,97,116,105,110,103,32,100,121,110,97,109,105,99,32,104,117,102,102,109,97,110,32,116,114,101,101,0,101,109,112,116,121,32,105,110,112,117,116,32,111,114,32,102,105,108,101,32,100,111,101,115,110,39,116,32,101,120,105,115,116,0,98,75,71,68,32,99,104,117,110,107,32,104,97,115,32,119,114,111,110,103,32,115,105,122,101,32,102,111,114,32,82,71,66,32,105,109,97,103,101,0,98,75,71,68,32,99,104,117,110,107,32,104,97,115,32,119,114,111,110,103,32,115,105,122,101,32,102,111,114,32,103,114,101,121,115,99,97,108,101,32,105,109,97,103,101,0,98,75,71,68,32,99,104,117,110,107,32,104,97,115,32,119,114,111,110,103,32,115,105,122,101,32,102,111,114,32,112,97,108,101,116,116,101,32,105,109,97,103,101,0,116,82,78,83,32,99,104,117,110,107,32,97,112,112,101,97,114,101,100,32,119,104,105,108,101,32,105,116,32,119,97,115,32,110,111,116,32,97,108,108,111,119,101,100,32,102,111,114,32,116,104,105,115,32,99,111,108,111,114,32,116,121,112,101,0,116,82,78,83,32,99,104,117,110,107,32,104,97,115,32,119,114,111,110,103,32,115,105,122,101,32,102,111,114,32,82,71,66,32,105,109,97,103,101,0,116,82,78,83,32,99,104,117,110,107,32,104,97,115,32,119,114,111,110,103,32,115,105,122,101,32,102,111,114,32,103,114,101,121,115,99,97,108,101,32,105,109,97,103,101,0,109,111,114,101,32,112,97,108,101,116,116,101,32,97,108,112,104,97,32,118,97,108,117,101,115,32,103,105,118,101,110,32,105,110,32,116,82,78,83,32,99,104,117,110,107,32,116,104,97,110,32,116,104,101,114,101,32,97,114,101,32,99,111,108,111,114,115,32,105,110,32,116,104,101,32,112,97,108,101,116,116,101,0,116,104,101,32,112,97,108,101,116,116,101,32,105,115,32,116,111,111,32,98,105,103,0,105,108,108,101,103,97,108,32,98,105,116,32,100,101,112,116,104,32,102,111,114,32,116,104,105,115,32,99,111,108,111,114,32,116,121,112,101,32,103,105,118,101,110,0,105,108,108,101,103,97,108,32,80,78,71,32,102,105,108,116,101,114,32,116,121,112,101,32,101,110,99,111,117,110,116,101,114,101,100,0,99,104,117,110,107,32,108,101,110,103,116,104,32,111,102,32,97,32,99,104,117,110,107,32,105,115,32,116,111,111,32,108,97,114,103,101,32,111,114,32,116,104,101,32,99,104,117,110,107,32,116,111,111,32,115,109,97,108,108,0,105,108,108,101,103,97,108,32,80,78,71,32,105,110,116,101,114,108,97,99,101,32,109,101,116,104,111,100,0,105,108,108,101,103,97,108,32,80,78,71,32,102,105,108,116,101,114,32,109,101,116,104,111,100,0,105,108,108,101,103,97,108,32,80,78,71,32,99,111,109,112,114,101,115,115,105,111,110,32,109,101,116,104,111,100,0,105,108,108,101,103,97,108,32,80,78,71,32,99,111,108,111,114,32,116,121,112,101,32,111,114,32,98,112,112,0,99,104,117,110,107,32,108,101,110,103,116,104,32,116,111,111,32,108,97,114,103,101,44,32,99,104,117,110,107,32,98,114,111,107,101,110,32,111,102,102,32,97,116,32,101,110,100,32,111,102,32,102,105,108,101,0,102,105,114,115,116,32,99,104,117,110,107,32,105,115,32,110,111,116,32,116,104,101,32,104,101,97,100,101,114,32,99,104,117,110,107,0,105,110,99,111,114,114,101,99,116,32,80,78,71,32,115,105,103,110,97,116,117,114,101,44,32,105,116,39,115,32,110,111,32,80,78,71,32,111,114,32,99,111,114,114,117,112,116,101,100,0,80,78,71,32,102,105,108,101,32,105,115,32,115,109,97,108,108,101,114,32,116,104,97,110,32,97,32,80,78,71,32,104,101,97,100,101,114,0,70,68,73,67,84,32,101,110,99,111,117,110,116,101,114,101,100,32,105,110,32,122,108,105,98,32,104,101,97,100,101,114,32,119,104,105,108,101,32,105,116,39,115,32,110,111,116,32,117,115,101,100,32,102,111,114,32,80,78,71,0,105,110,118,97,108,105,100,32,99,111,109,112,114,101,115,115,105,111,110,32,109,101,116,104,111,100,32,105,110,32,122,108,105,98,32,104,101,97,100,101,114,0,105,110,118,97,108,105,100,32,70,67,72,69,67,75,32,105,110,32,122,108,105,98,32,104,101,97,100,101,114,0,101,110,100,32,111,102,32,105,110,32,98,117,102,102,101,114,32,109,101,109,111,114,121,32,114,101,97,99,104,101,100,32,119,104,105,108,101,32,105,110,102,108,97,116,105,110,103,0,101,110,100,32,111,102,32,111,117,116,32,98,117,102,102,101,114,32,109,101,109,111,114,121,32,114,101,97,99,104,101,100,32,119,104,105,108,101,32,105,110,102,108,97,116,105,110,103,0,78,76,69,78,32,105,115,32,110,111,116,32,111,110,101,115,32,99,111,109,112,108,101,109,101,110,116,32,111,102,32,76,69,78,32,105,110,32,97,32,100,101,102,108,97,116,101,32,98,108,111,99,107,0,105,110,118,97,108,105,100,32,100,101,102,108,97,116,101,32,98,108,111,99,107,32,66,84,89,80,69,32,101,110,99,111,117,110,116,101,114,101,100,32,119,104,105,108,101,32,100,101,99,111,100,105,110,103,0,105,110,118,97,108,105,100,32,100,105,115,116,97,110,99,101,32,99,111,100,101,32,119,104,105,108,101,32,105,110,102,108,97,116,105,110,103,0,117,110,101,120,105,115,116,105,110,103,32,99,111,100,101,32,119,104,105,108,101,32,112,114,111,99,101,115,115,105,110,103,32,100,121,110,97,109,105,99,32,100,101,102,108,97,116,101,32,98,108,111,99,107,0,112,114,111,98,108,101,109,32,119,104,105,108,101,32,112,114,111,99,101,115,115,105,110,103,32,100,121,110,97,109,105,99,32,100,101,102,108,97,116,101,32,98,108,111,99,107,0,101,114,114,111,114,32,105,110,32,99,111,100,101,32,116,114,101,101,32,109,97,100,101,32,105,116,32,106,117,109,112,32,111,117,116,115,105,100,101,32,111,102,32,104,117,102,102,109,97,110,32,116,114,101,101,0,101,110,100,32,111,102,32,105,110,112,117,116,32,109,101,109,111,114,121,32,114,101,97,99,104,101,100,32,119,105,116,104,111,117,116,32,104,117,102,102,109,97,110,32,101,110,100,32,99,111,100,101,0,110,111,116,104,105,110,103,32,100,111,110,101,32,121,101,116,0,110,111,32,101,114,114,111,114,44,32,101,118,101,114,121,116,104,105,110,103,32,119,101,110,116,32,111,107,0,99,108,97,115,115,0,105,110,118,97,108,105,100,32,111,98,106,101,99,116,32,112,97,115,115,101,100,32,116,111,32,39,97,117,120,105,108,105,97,114,46,99,58,95,95,116,111,115,116,114,105,110,103,39,0,37,46,51,53,115,32,101,120,112,101,99,116,101,100,0,42,108,0,105,110,118,97,108,105,100,32,114,101,99,101,105,118,101,32,112,97,116,116,101,114,110,0,110,101,119,116,114,121,0,112,114,111,116,101,99,116,0,100,110,115,0,116,111,105,112,0,105,112,0,97,108,105,97,115,0,116,111,104,111,115,116,110,97,109,101,0,103,101,116,104,111,115,116,110,97,109,101,0,103,101,116,104,111,115,116,110,97,109,101,32,102,97,105,108,101,100,0,103,101,116,112,101,101,114,110,97,109,101,32,102,97,105,108,101,100,0,103,101,116,115,111,99,107,110,97,109,101,32,102,97,105,108,101,100,0,99,108,111,115,101,100,0,117,110,115,117,112,112,111,114,116,101,100,32,111,112,116,105,111,110,32,96,37,46,51,53,115,39,0,115,101,116,115,111,99,107,111,112,116,32,102,97,105,108,101,100,0,98,111,111,108,101,97,110,32,39,111,110,39,32,102,105,101,108,100,32,101,120,112,101,99,116,101,100,0,116,105,109,101,111,117,116,0,110,117,109,98,101,114,32,39,116,105,109,101,111,117,116,39,32,102,105,101,108,100,32,101,120,112,101,99,116,101,100,0,109,117,108,116,105,97,100,100,114,0,115,116,114,105,110,103,32,39,109,117,108,116,105,97,100,100,114,39,32,102,105,101,108,100,32,101,120,112,101,99,116,101,100,0,105,110,118,97,108,105,100,32,39,109,117,108,116,105,97,100,100,114,39,32,105,112,32,97,100,100,114,101,115,115,0,105,110,116,101,114,102,97,99,101,0,115,116,114,105,110,103,32,39,105,110,116,101,114,102,97,99,101,39,32,102,105,101,108,100,32,101,120,112,101,99,116,101,100,0,42,0,105,110,118,97,108,105,100,32,39,105,110,116,101,114,102,97,99,101,39,32,105,112,32,97,100,100,114,101,115,115,0,124,36,36,36,125,114,115,116,117,118,119,120,121,122,123,36,36,36,36,36,36,36,62,63,64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,36,36,36,36,36,36,88,89,90,91,92,93,94,95,96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,0,78,52,108,111,118,101,49,48,102,105,108,101,115,121,115,116,101,109,54,112,104,121,115,102,115,52,70,105,108,101,69,0,80,104,121,115,70,83,32,105,115,32,110,111,116,32,105,110,105,116,105,97,108,105,122,101,100,46,0,67,111,117,108,100,32,110,111,116,32,111,112,101,110,32,102,105,108,101,32,37,115,46,32,68,111,101,115,32,110,111,116,32,101,120,105,115,116,46,0,67,111,117,108,100,32,110,111,116,32,115,101,116,32,119,114,105,116,101,32,100,105,114,101,99,116,111,114,121,46,0,67,111,117,108,100,32,110,111,116,32,111,112,101,110,32,102,105,108,101,32,37,115,32,40,37,115,41,0,70,105,108,101,32,105,115,32,110,111,116,32,111,112,101,110,101,100,32,102,111,114,32,114,101,97,100,105,110,103,46,0,73,110,118,97,108,105,100,32,114,101,97,100,32,115,105,122,101,46,0,70,105,108,101,32,105,115,32,110,111,116,32,111,112,101,110,101,100,32,102,111,114,32,119,114,105,116,105,110,103,46,0,73,110,118,97,108,105,100,32,119,114,105,116,101,32,115,105,122,101,46,0,70,83,46,115,121,110,99,102,115,40,102,97,108,115,101,44,32,102,117,110,99,116,105,111,110,32,40,101,114,114,41,32,123,32,105,102,32,40,101,114,114,41,32,123,32,99,111,110,115,111,108,101,46,108,111,103,40,101,114,114,41,59,32,125,32,125,41,59,0,78,52,108,111,118,101,53,105,109,97,103,101,54,109,97,103,112,105,101,49,49,65,83,84,67,72,97,110,100,108,101,114,69,0,67,111,117,108,100,32,110,111,116,32,100,101,99,111,100,101,32,99,111,109,112,114,101,115,115,101,100,32,100,97,116,97,32,40,110,111,116,32,97,110,32,46,97,115,116,99,32,102,105,108,101,63,41,0,67,111,117,108,100,32,110,111,116,32,112,97,114,115,101,32,46,97,115,116,99,32,102,105,108,101,58,32,117,110,115,117,112,112,111,114,116,101,100,32,65,83,84,67,32,102,111,114,109,97,116,32,37,100,120,37,100,120,37,100,46,0,67,111,117,108,100,32,110,111,116,32,112,97,114,115,101,32,46,97,115,116,99,32,102,105,108,101,58,32,102,105,108,101,32,105,115,32,116,111,111,32,115,109,97,108,108,46,0,79,117,116,32,111,102,32,109,101,109,111,114,121,46,0,78,52,108,111,118,101,53,105,109,97,103,101,54,109,97,103,112,105,101,49,57,67,111,109,112,114,101,115,115,101,100,73,109,97,103,101,68,97,116,97,69,0,67,111,117,108,100,32,110,111,116,32,112,97,114,115,101,32,99,111,109,112,114,101,115,115,101,100,32,100,97,116,97,58,32,85,110,107,110,111,119,110,32,102,111,114,109,97,116,46,0,67,111,117,108,100,32,110,111,116,32,112,97,114,115,101,32,99,111,109,112,114,101,115,115,101,100,32,100,97,116,97,46,0,67,111,117,108,100,32,110,111,116,32,112,97,114,115,101,32,99,111,109,112,114,101,115,115,101,100,32,100,97,116,97,58,32,78,111,32,118,97,108,105,100,32,100,97,116,97,63,0,78,52,108,111,118,101,53,105,109,97,103,101,54,109,97,103,112,105,101,49,48,68,68,83,72,97,110,100,108,101,114,69,0,67,111,117,108,100,32,110,111,116,32,100,101,99,111,100,101,32,99,111,109,112,114,101,115,115,101,100,32,100,97,116,97,32,40,110,111,116,32,97,32,68,68,83,32,102,105,108,101,63,41,0,67,111,117,108,100,32,110,111,116,32,112,97,114,115,101,32,99,111,109,112,114,101,115,115,101,100,32,100,97,116,97,58,32,85,110,115,117,112,112,111,114,116,101,100,32,102,111,114,109,97,116,46,0,67,111,117,108,100,32,110,111,116,32,112,97,114,115,101,32,99,111,109,112,114,101,115,115,101,100,32,100,97,116,97,58,32,78,111,32,114,101,97,100,97,98,108,101,32,116,101,120,116,117,114,101,32,100,97,116,97,46,0,78,52,108,111,118,101,53,105,109,97,103,101,54,109,97,103,112,105,101,49,51,70,111,114,109,97,116,72,97,110,100,108,101,114,69,0,73,109,97,103,101,32,100,101,99,111,100,105,110,103,32,105,115,32,110,111,116,32,105,109,112,108,101,109,101,110,116,101,100,32,102,111,114,32,116,104,105,115,32,102,111,114,109,97,116,32,98,97,99,107,101,110,100,46,0,73,109,97,103,101,32,101,110,99,111,100,105,110,103,32,105,115,32,110,111,116,32,105,109,112,108,101,109,101,110,116,101,100,32,102,111,114,32,116,104,105,115,32,102,111,114,109,97,116,32,98,97,99,107,101,110,100,46,0,78,52,108,111,118,101,54,116,104,114,101,97,100,51,115,100,108,54,84,104,114,101,97,100,69,0,78,52,108,111,118,101,54,116,104,114,101,97,100,54,84,104,114,101,97,100,69,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,66,97,100,32,116,97,103,32,108,101,110,103,116,104,32,40,110,111,116,32,115,121,110,99,104,115,97,102,101,41,58,32,48,120,37,48,50,120,37,48,50,120,37,48,50,120,37,48,50,120,59,32,89,111,117,32,103,111,116,32,97,32,98,97,100,32,73,68,51,32,116,97,103,32,104,101,114,101,46,10,0,78,111,116,101,58,32,73,68,51,118,50,46,37,105,32,114,101,118,32,37,105,32,116,97,103,32,111,102,32,37,108,117,32,98,121,116,101,115,10,0,78,111,116,101,58,32,83,107,105,112,112,105,110,103,32,73,68,51,118,50,32,116,97,103,32,112,101,114,32,117,115,101,114,32,114,101,113,117,101,115,116,46,10,0,91,105,100,51,46,99,58,37,105,93,32,119,97,114,110,105,110,103,58,32,73,68,51,118,50,58,32,87,111,110,39,116,32,112,97,114,115,101,32,116,104,101,32,73,68,51,118,50,32,116,97,103,32,119,105,116,104,32,109,97,106,111,114,32,118,101,114,115,105,111,110,32,37,117,32,97,110,100,32,102,108,97,103,115,32,48,120,37,120,117,32,45,32,115,111,109,101,32,101,120,116,114,97,32,99,111,100,101,32,109,97,121,32,98,101,32,110,101,101,100,101,100,10,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,66,97,100,32,40,110,111,110,45,115,121,110,99,104,115,97,102,101,41,32,116,97,103,32,111,102,102,115,101,116,58,32,48,120,37,48,50,120,37,48,50,120,37,48,50,120,37,48,50,120,10,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,73,68,51,118,50,58,32,110,111,110,45,115,121,110,99,115,97,102,101,32,115,105,122,101,32,111,102,32,37,115,32,102,114,97,109,101,44,32,115,107,105,112,112,105,110,103,32,116,104,101,32,114,101,109,97,105,110,100,101,114,32,111,102,32,116,97,103,10,0,78,111,116,101,58,32,73,68,51,118,50,32,37,115,32,102,114,97,109,101,32,111,102,32,115,105,122,101,32,37,108,117,10,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,87,104,111,97,33,32,73,68,51,118,50,32,102,114,97,109,101,32,99,108,97,105,109,115,32,116,111,32,98,101,32,108,97,114,103,101,114,32,116,104,97,110,32,116,104,101,32,119,104,111,108,101,32,114,101,115,116,32,111,102,32,116,104,101,32,116,97,103,46,10,0,67,79,77,0,84,65,76,0,84,66,80,0,84,67,77,0,84,67,79,0,84,67,82,0,84,68,65,0,84,68,89,0,84,69,78,0,84,70,84,0,84,73,77,0,84,75,69,0,84,76,65,0,84,76,69,0,84,77,84,0,84,79,65,0,84,79,70,0,84,79,76,0,84,79,82,0,84,79,84,0,84,80,49,0,84,80,50,0,84,80,51,0,84,80,52,0,84,80,65,0,84,80,66,0,84,82,67,0,84,82,75,0,84,83,73,0,84,83,83,0,84,84,49,0,84,84,50,0,84,84,51,0,84,88,84,0,84,88,88,0,84,89,69,0,67,79,77,77,0,84,65,76,66,0,84,66,80,77,0,84,67,79,77,0,84,67,79,78,0,84,67,79,80,0,84,68,65,84,0,84,68,76,89,0,84,69,78,67,0,84,70,76,84,0,84,73,77,69,0,84,75,69,89,0,84,76,65,78,0,84,76,69,78,0,84,77,69,68,0,84,79,80,69,0,84,79,70,78,0,84,79,76,89,0,84,79,82,89,0,84,79,65,76,0,84,80,69,49,0,84,80,69,50,0,84,80,69,51,0,84,80,69,52,0,84,80,79,83,0,84,80,85,66,0,84,83,82,67,0,84,82,68,65,0,84,82,67,75,0,84,83,73,90,0,84,83,83,69,0,84,73,84,49,0,84,73,84,50,0,84,73,84,51,0,84,69,88,84,0,84,88,88,88,0,84,89,69,82,0,84,114,97,110,115,108,97,116,101,100,32,73,68,51,118,50,46,50,32,102,114,97,109,101,32,37,115,32,116,111,32,37,115,10,0,73,103,110,111,114,105,110,103,32,117,110,116,114,97,110,115,108,97,116,101,100,32,73,68,51,118,50,46,50,32,102,114,97,109,101,32,37,99,37,99,37,99,10,0,67,79,77,77,0,84,88,88,88,0,82,86,65,50,0,85,83,76,84,0,91,105,100,51,46,99,58,37,105,93,32,119,97,114,110,105,110,103,58,32,73,68,51,118,50,58,32,115,107,105,112,112,105,110,103,32,105,110,118,97,108,105,100,47,117,110,115,117,112,112,111,114,116,101,100,32,102,114,97,109,101,10,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,73,68,51,118,50,58,32,117,110,97,98,108,101,32,116,111,32,97,108,108,111,99,97,116,101,32,119,111,114,107,105,110,103,32,98,117,102,102,101,114,32,102,111,114,32,100,101,45,117,110,115,121,110,99,10,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,73,110,118,97,108,105,100,32,102,114,97,109,101,32,115,105,122,101,32,111,102,32,37,108,117,32,40,116,111,111,32,115,109,97,108,108,32,102,111,114,32,97,110,121,116,104,105,110,103,41,46,10,0,117,110,107,110,111,119,110,33,0,85,84,70,45,56,0,85,84,70,45,49,54,32,66,69,0,85,84,70,45,49,54,32,66,79,77,0,76,97,116,105,110,32,49,0,78,111,116,101,58,32,83,116,111,114,105,110,103,32,99,111,109,109,101,110,116,32,102,114,111,109,32,37,115,32,101,110,99,111,100,105,110,103,10,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,85,110,97,98,108,101,32,116,111,32,97,116,116,97,99,104,32,110,101,119,32,99,111,109,109,101,110,116,33,10,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,78,111,32,99,111,109,109,101,110,116,32,116,101,120,116,32,47,32,118,97,108,105,100,32,100,101,115,99,114,105,112,116,105,111,110,63,10,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,67,97,110,110,111,116,32,114,101,115,105,122,101,32,116,97,114,103,101,116,32,115,116,114,105,110,103,44,32,111,117,116,32,111,102,32,109,101,109,111,114,121,63,10,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,85,110,107,110,111,119,110,32,116,101,120,116,32,101,110,99,111,100,105,110,103,32,37,117,44,32,73,32,116,97,107,101,32,110,111,32,99,104,97,110,99,101,115,44,32,115,111,114,114,121,33,10,0,91,105,100,51,46,99,58,37,105,93,32,119,97,114,110,105,110,103,58,32,87,101,105,114,100,32,116,97,103,32,115,105,122,101,32,37,100,32,102,111,114,32,101,110,99,111,100,105,110,103,32,37,117,32,45,32,73,32,119,105,108,108,32,112,114,111,98,97,98,108,121,32,116,114,105,109,32,116,111,111,32,101,97,114,108,121,32,111,114,32,115,111,109,101,116,104,105,110,103,32,98,117,116,32,73,32,116,104,105,110,107,32,116,104,101,32,77,80,51,32,105,115,32,98,114,111,107,101,110,46,10,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,73,110,118,97,108,105,100,32,85,84,70,49,54,32,115,117,114,114,111,103,97,116,101,32,112,97,105,114,32,97,116,32,37,108,105,32,40,48,120,37,48,52,108,120,41,46,10,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,117,110,97,98,108,101,32,116,111,32,99,111,110,118,101,114,116,32,115,116,114,105,110,103,32,116,111,32,85,84,70,45,56,32,40,111,117,116,32,111,102,32,109,101,109,111,114,121,44,32,106,117,110,107,32,105,110,112,117,116,63,41,33,10,0,78,111,116,101,58,32,73,68,51,32,99,111,109,109,47,117,115,108,116,32,100,101,115,99,32,111,102,32,108,101,110,103,116,104,32,37,108,117,46,10,0,78,111,116,101,58,32,73,68,51,32,99,111,109,109,47,117,115,108,116,32,116,101,120,116,32,111,102,32,108,101,110,103,116,104,32,37,108,117,46,10,0,114,118,97,0,114,118,97,95,109,105,120,0,114,118,97,95,116,114,97,99,107,0,114,118,97,95,114,97,100,105,111,0,114,118,97,95,97,108,98,117,109,0,114,118,97,95,97,117,100,105,111,112,104,105,108,101,0,114,118,97,95,117,115,101,114,0,78,111,116,101,58,32,82,86,65,32,118,97,108,117,101,32,37,102,100,66,10,0,78,111,116,101,58,32,83,116,111,114,105,110,103,32,101,120,116,114,97,32,102,114,111,109,32,37,115,32,101,110,99,111,100,105,110,103,10,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,78,111,32,101,120,116,114,97,32,102,114,97,109,101,32,116,101,120,116,32,47,32,118,97,108,105,100,32,100,101,115,99,114,105,112,116,105,111,110,63,10,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,85,110,97,98,108,101,32,116,111,32,97,116,116,97,99,104,32,110,101,119,32,101,120,116,114,97,32,116,101,120,116,33,10,0,114,101,112,108,97,121,103,97,105,110,95,116,114,97,99,107,95,0,78,111,116,101,58,32,82,86,65,32,82,101,112,108,97,121,71,97,105,110,32,116,114,97,99,107,32,103,97,105,110,47,112,101,97,107,10,0,114,101,112,108,97,121,103,97,105,110,95,116,114,97,99,107,95,112,101,97,107,0,114,101,112,108,97,121,103,97,105,110,95,116,114,97,99,107,95,103,97,105,110,0,114,101,112,108,97,121,103,97,105,110,95,97,108,98,117,109,95,0,78,111,116,101,58,32,82,86,65,32,82,101,112,108,97,121,71,97,105,110,32,97,108,98,117,109,32,103,97,105,110,47,112,101,97,107,10,0,114,101,112,108,97,121,103,97,105,110,95,97,108,98,117,109,95,112,101,97,107,0,114,101,112,108,97,121,103,97,105,110,95,97,108,98,117,109,95,103,97,105,110,0,78,111,116,101,58,32,82,86,65,32,112,101,97,107,32,37,102,10,0,78,111,116,101,58,32,82,86,65,32,103,97,105,110,32,37,102,100,66,10,0,78,111,116,101,58,32,82,86,65,50,32,105,100,101,110,116,105,102,105,99,97,116,105,111,110,32,34,37,115,34,10,0,97,108,98,117,109,0,97,117,100,105,111,112,104,105,108,101,0,117,115,101,114,0,78,111,116,101,58,32,83,116,111,114,105,110,103,32,116,101,120,116,32,102,114,111,109,32,37,115,32,101,110,99,111,100,105,110,103,10,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,85,110,97,98,108,101,32,116,111,32,97,116,116,97,99,104,32,110,101,119,32,116,101,120,116,33,10,0,78,111,116,101,58,32,73,68,51,118,50,32,37,99,37,99,37,99,37,99,32,116,101,120,116,32,102,114,97,109,101,58,32,37,115,10,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,73,68,51,118,50,58,32,117,110,107,110,111,119,110,32,102,114,97,109,101,32,116,121,112,101,32,37,105,10,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,73,68,51,118,50,58,32,68,117,104,44,32,110,111,116,32,97,98,108,101,32,116,111,32,114,101,97,100,32,73,68,51,118,50,32,116,97,103,32,100,97,116,97,46,10,0,91,105,100,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,73,68,51,118,50,58,32,65,114,114,103,33,32,85,110,97,98,108,101,32,116,111,32,97,108,108,111,99,97,116,101,32,37,108,117,32,98,121,116,101,115,32,102,111,114,32,105,110,116,101,114,112,114,101,116,105,110,103,32,73,68,51,118,50,32,100,97,116,97,32,45,32,116,114,121,105,110,103,32,116,111,32,115,107,105,112,32,105,110,115,116,101,97,100,46,10,0,78,111,116,101,58,32,74,117,110,107,32,97,116,32,116,104,101,32,98,101,103,105,110,110,105,110,103,32,40,48,120,37,48,56,108,120,41,10,0,78,111,116,101,58,32,76,111,111,107,115,32,108,105,107,101,32,97,32,82,73,70,70,32,104,101,97,100,101,114,46,10,0,78,111,116,101,58,32,83,107,105,112,112,101,100,32,82,73,70,70,32,104,101,97,100,101,114,33,10,0,91,112,97,114,115,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,89,111,117,32,102,111,111,108,101,100,32,109,101,32,116,111,111,32,111,102,116,101,110,46,32,82,101,102,117,115,105,110,103,32,116,111,32,103,117,101,115,115,32,102,114,101,101,32,102,111,114,109,97,116,32,102,114,97,109,101,32,115,105,122,101,32,95,97,103,97,105,110,95,46,10,0,91,112,97,114,115,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,67,97,110,110,111,116,32,108,111,111,107,32,102,111,114,32,102,114,101,101,102,111,114,109,97,116,32,102,114,97,109,101,32,115,105,122,101,32,119,105,116,104,32,110,111,110,45,115,101,101,107,97,98,108,101,32,97,110,100,32,110,111,110,45,98,117,102,102,101,114,101,100,32,115,116,114,101,97,109,33,10,0,78,111,116,101,58,32,102,114,101,101,32,102,111,114,109,97,116,32,102,114,97,109,101,32,115,105,122,101,32,37,108,105,10,0,91,112,97,114,115,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,69,110,99,111,117,110,116,101,114,101,100,32,102,114,101,101,32,102,111,114,109,97,116,32,104,101,97,100,101,114,44,32,98,117,116,32,102,97,105,108,101,100,32,116,111,32,103,117,101,115,115,32,102,114,97,109,101,32,115,105,122,101,46,10,0,84,114,117,110,99,97,116,105,110,103,32,115,116,101,114,101,111,32,98,111,117,110,100,97,114,121,32,116,111,32,115,105,100,101,98,97,110,100,32,108,105,109,105,116,46,10,0,78,111,116,101,58,32,109,105,115,115,105,110,103,32,37,100,32,98,121,116,101,115,32,105,110,32,98,105,116,32,114,101,115,101,114,118,111,105,114,32,102,111,114,32,102,114,97,109,101,32,37,108,105,10,0,91,108,97,121,101,114,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,98,105,103,95,118,97,108,117,101,115,32,116,111,111,32,108,97,114,103,101,33,10,0,91,108,97,121,101,114,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,66,108,111,99,107,116,121,112,101,32,61,61,32,48,32,97,110,100,32,119,105,110,100,111,119,45,115,119,105,116,99,104,105,110,103,32,61,61,32,49,32,110,111,116,32,97,108,108,111,119,101,100,46,10,0,91,108,97,121,101,114,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,98,97,100,32,102,114,97,109,101,32,45,32,117,110,97,98,108,101,32,116,111,32,103,101,116,32,118,97,108,105,100,32,115,105,100,101,105,110,102,111,10,0,6,5,5,5,6,5,7,3,11,10,0,0,7,7,7,0,6,6,6,3,8,8,5,0,9,9,9,9,9,9,12,6,18,18,0,0,12,12,12,0,12,9,9,6,15,12,9,0,6,9,9,9,6,9,12,6,15,18,0,0,6,15,12,0,6,12,9,6,6,18,9,0,0,0,0,0,3,1,1,1,2,2,2,3,3,3,4,4,0,1,2,3,0,1,2,3,1,2,3,1,2,3,2,3,91,108,97,121,101,114,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,97,116,116,101,109,112,116,101,100,32,120,114,112,110,116,32,111,118,101,114,102,108,111,119,32,40,37,112,32,33,60,32,37,112,41,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,2,2,3,3,3,2,0,91,108,97,121,101,114,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,100,101,113,117,97,110,116,105,122,97,116,105,111,110,32,102,97,105,108,101,100,33,10,0,91,112,97,114,115,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,76,97,121,101,114,32,116,121,112,101,32,37,105,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,105,110,32,116,104,105,115,32,98,117,105,108,100,33,10,0,91,112,97,114,115,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,70,114,97,109,101,32,115,105,122,101,32,116,111,111,32,98,105,103,58,32,37,100,10,0,91,112,97,114,115,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,71,105,118,105,110,103,32,117,112,32,115,101,97,114,99,104,105,110,103,32,118,97,108,105,100,32,77,80,69,71,32,104,101,97,100,101,114,32,97,102,116,101,114,32,37,108,105,32,98,121,116,101,115,32,111,102,32,106,117,110,107,46,10,0,78,111,116,101,58,32,83,107,105,112,112,101,100,32,73,68,51,118,49,32,116,97,103,46,10,0,78,111,116,101,58,32,73,108,108,101,103,97,108,32,65,117,100,105,111,45,77,80,69,71,45,72,101,97,100,101,114,32,48,120,37,48,56,108,120,32,97,116,32,111,102,102,115,101,116,32,37,108,105,46,10,0,78,111,116,101,58,32,67,111,117,108,100,32,98,101,32,97,32,66,77,80,32,97,108,98,117,109,32,97,114,116,46,10,0,78,111,116,101,58,32,84,114,121,105,110,103,32,116,111,32,114,101,115,121,110,99,46,46,46,10,0,78,111,116,101,58,32,72,105,116,32,101,110,100,32,111,102,32,40,97,118,97,105,108,97,98,108,101,41,32,100,97,116,97,32,100,117,114,105,110,103,32,114,101,115,121,110,99,46,10,0,78,111,116,101,58,32,83,107,105,112,112,101,100,32,37,108,105,32,98,121,116,101,115,32,105,110,32,105,110,112,117,116,46,10,0,91,112,97,114,115,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,71,105,118,105,110,103,32,117,112,32,114,101,115,121,110,99,32,97,102,116,101,114,32,37,108,105,32,98,121,116,101,115,32,45,32,121,111,117,114,32,115,116,114,101,97,109,32,105,115,32,110,111,116,32,110,105,99,101,46,46,46,32,40,109,97,121,98,101,32,105,110,99,114,101,97,115,105,110,103,32,114,101,115,121,110,99,32,108,105,109,105,116,32,99,111,117,108,100,32,104,101,108,112,41,46,10,0,91,112,97,114,115,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,110,111,116,32,97,116,116,101,109,112,116,105,110,103,32,116,111,32,114,101,115,121,110,99,46,46,46,10,0,91,112,97,114,115,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,99,97,110,110,111,116,32,115,101,101,107,33,10,0,91,112,97,114,115,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,67,97,110,110,111,116,32,115,101,101,107,32,98,97,99,107,33,10,0,91,112,97,114,115,101,46,99,58,37,105,93,32,119,97,114,110,105,110,103,58,32,67,97,110,110,111,116,32,114,101,97,100,32,110,101,120,116,32,104,101,97,100,101,114,44,32,97,32,111,110,101,45,102,114,97,109,101,32,115,116,114,101,97,109,63,32,68,117,104,46,46,46,10,0,91,112,97,114,115,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,67,97,110,110,111,116,32,115,101,101,107,32,51,32,98,121,116,101,115,32,98,97,99,107,33,10,0,78,111,116,101,58,32,88,105,110,103,47,76,97,109,101,47,73,110,102,111,32,104,101,97,100,101,114,32,100,101,116,101,99,116,101,100,10,0,78,111,116,101,58,32,73,103,110,111,114,105,110,103,32,88,105,110,103,32,102,114,97,109,101,115,32,98,101,99,97,117,115,101,32,111,102,32,77,80,71,49,50,51,95,73,71,78,79,82,69,95,83,84,82,69,65,77,76,69,78,71,84,72,10,0,78,111,116,101,58,32,88,105,110,103,58,32,37,108,117,32,102,114,97,109,101,115,10,0,78,111,116,101,58,32,73,103,110,111,114,105,110,103,32,88,105,110,103,32,98,121,116,101,115,32,98,101,99,97,117,115,101,32,111,102,32,77,80,71,49,50,51,95,73,71,78,79,82,69,95,83,84,82,69,65,77,76,69,78,71,84,72,10,0,78,111,116,101,58,32,88,105,110,103,32,115,116,114,101,97,109,32,115,105,122,101,32,37,108,117,32,100,105,102,102,101,114,115,32,98,121,32,37,102,37,37,32,102,114,111,109,32,100,101,116,101,114,109,105,110,101,100,47,103,105,118,101,110,32,102,105,108,101,32,115,105,122,101,33,10,0,87,97,114,110,105,110,103,58,32,88,105,110,103,32,115,116,114,101,97,109,32,115,105,122,101,32,111,102,102,32,98,121,32,109,111,114,101,32,116,104,97,110,32,49,37,37,44,32,102,117,122,122,121,32,115,101,101,107,105,110,103,32,109,97,121,32,98,101,32,101,118,101,110,32,109,111,114,101,32,102,117,122,122,121,32,116,104,97,110,32,98,121,32,100,101,115,105,103,110,33,10,0,78,111,116,101,58,32,88,105,110,103,58,32,37,108,117,32,98,121,116,101,115,10,0,78,111,116,101,58,32,88,105,110,103,58,32,113,117,97,108,105,116,121,32,61,32,37,108,117,10,0,78,111,116,101,58,32,73,110,102,111,58,32,69,110,99,111,100,101,114,58,32,37,115,10,0,76,65,77,69,0,37,117,46,37,117,37,115,0,78,111,116,101,58,32,73,110,102,111,58,32,79,108,100,32,76,65,77,69,32,100,101,116,101,99,116,101,100,44,32,117,115,105,110,103,32,82,101,112,108,97,121,71,97,105,110,32,112,114,101,97,109,112,32,111,102,32,37,102,32,100,66,46,10,0,78,111,116,101,58,32,73,110,102,111,58,32,67,97,110,110,111,116,32,100,101,116,101,114,109,105,110,101,32,76,65,77,69,32,118,101,114,115,105,111,110,46,10,0,78,111,116,101,58,32,73,110,102,111,58,32,114,101,118,32,37,117,10,0,78,111,116,101,58,32,73,110,102,111,58,32,118,98,114,32,109,111,100,101,32,37,117,10,0,78,111,116,101,58,32,73,110,102,111,58,32,112,101,97,107,32,61,32,37,102,32,40,73,32,119,111,110,39,116,32,117,115,101,32,116,104,105,115,41,10,0,78,111,116,101,58,32,73,110,102,111,58,32,82,97,100,105,111,32,71,97,105,110,32,61,32,37,48,51,46,49,102,100,66,10,0,78,111,116,101,58,32,73,110,102,111,58,32,65,117,100,105,111,112,104,105,108,101,32,71,97,105,110,32,61,32,37,48,51,46,49,102,100,66,10,0,78,111,116,101,58,32,73,110,102,111,58,32,65,66,82,32,114,97,116,101,32,61,32,37,117,10,0,78,111,116,101,58,32,69,110,99,111,100,101,114,32,100,101,108,97,121,32,61,32,37,105,59,32,112,97,100,100,105,110,103,32,61,32,37,105,10,0,78,111,116,101,58,32,100,111,105,110,103,32,82,86,65,32,119,105,116,104,32,103,97,105,110,32,37,102,10,0,91,102,114,97,109,101,46,99,58,37,105,93,32,119,97,114,110,105,110,103,58,32,108,105,109,105,116,105,110,103,32,115,99,97,108,101,32,118,97,108,117,101,32,116,111,32,37,102,32,116,111,32,112,114,101,118,101,110,116,32,99,108,105,112,112,105,110,103,32,119,105,116,104,32,105,110,100,105,99,97,116,101,100,32,112,101,97,107,32,102,97,99,116,111,114,32,111,102,32,37,102,10,0,10,87,97,114,110,105,110,103,58,32,69,110,99,111,117,110,116,101,114,101,100,32,109,111,114,101,32,100,97,116,97,32,97,102,116,101,114,32,97,110,110,111,117,110,99,101,100,32,101,110,100,32,111,102,32,116,114,97,99,107,32,40,102,114,97,109,101,32,37,108,105,47,37,108,105,41,46,32,70,114,97,110,107,101,110,115,116,101,105,110,33,10,0,91,105,110,100,101,120,46,99,58,37,105,93,32,101,114,114,111,114,58,32,102,97,105,108,101,100,32,116,111,32,114,101,115,105,122,101,32,105,110,100,101,120,33,10,0,78,111,116,101,58,32,69,110,97,98,108,105,110,103,32,97,108,108,32,102,111,114,109,97,116,115,46,10,0,91,102,114,97,109,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,104,97,118,101,32,101,120,116,101,114,110,97,108,32,98,117,102,102,101,114,32,111,102,32,115,105,122,101,32,37,108,117,44,32,110,101,101,100,32,37,108,117,10,0,91,102,114,97,109,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,66,97,100,32,100,111,119,110,95,115,97,109,112,108,101,32,40,37,105,41,32,46,46,46,32,115,104,111,117,108,100,32,110,111,116,32,98,101,32,112,111,115,115,105,98,108,101,33,33,10,0,91,102,114,97,109,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,66,97,100,32,100,111,119,110,95,115,97,109,112,108,101,32,46,46,46,32,115,104,111,117,108,100,32,110,111,116,32,98,101,32,112,111,115,115,105,98,108,101,33,33,10,0,10,87,97,114,110,105,110,103,58,32,82,101,97,108,32,115,97,109,112,108,101,32,99,111,117,110,116,32,100,105,102,102,101,114,115,32,102,114,111,109,32,103,105,118,101,110,32,103,97,112,108,101,115,115,32,115,97,109,112,108,101,32,99,111,117,110,116,46,32,70,114,97,110,107,101,110,115,116,101,105,110,32,115,116,114,101,97,109,63,10,0,91,102,114,97,109,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,69,110,100,32,115,97,109,112,108,101,32,99,111,117,110,116,32,115,109,97,108,108,101,114,32,116,104,97,110,32,103,97,112,108,101,115,115,32,101,110,100,33,32,40,37,108,105,32,60,32,37,108,105,41,46,32,68,105,115,97,98,108,105,110,103,32,103,97,112,108,101,115,115,32,109,111,100,101,32,102,114,111,109,32,110,111,119,32,111,110,46,10,0,109,111,110,111,44,32,0,115,116,101,114,101,111,44,32,0,56,98,105,116,44,32,0,91,102,111,114,109,97,116,46,99,58,37,105,93,32,101,114,114,111,114,58,32,85,110,97,98,108,101,32,116,111,32,115,101,116,32,117,112,32,111,117,116,112,117,116,32,102,111,114,109,97,116,33,32,67,111,110,115,116,114,97,105,110,116,115,58,32,37,115,37,115,37,108,105,72,122,46,10,0,91,102,111,114,109,97,116,46,99,58,37,105,93,32,101,114,114,111,114,58,32,85,110,97,98,108,101,32,116,111,32,115,101,116,32,117,112,32,111,117,116,112,117,116,32,102,111,114,109,97,116,33,32,67,111,110,115,116,114,97,105,110,116,115,58,32,37,115,37,115,37,108,105,44,32,37,108,105,32,111,114,32,37,108,105,72,122,46,10,0,91,102,111,114,109,97,116,46,99,58,37,105,93,32,101,114,114,111,114,58,32,83,111,109,101,32,117,110,107,110,111,119,110,32,101,110,99,111,100,105,110,103,63,63,63,32,40,37,105,41,10,0,78,111,116,101,58,32,68,105,115,97,98,108,105,110,103,32,97,108,108,32,102,111,114,109,97,116,115,46,10,0,78,111,116,101,58,32,87,97,110,116,32,116,111,32,101,110,97,98,108,101,32,102,111,114,109,97,116,32,37,108,105,47,37,105,32,102,111,114,32,101,110,99,111,100,105,110,103,115,32,48,120,37,120,46,10,0,91,111,112,116,105,109,105,122,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,115,101,116,95,115,121,110,116,104,95,102,117,110,99,116,105,111,110,115,58,32,84,104,105,115,32,111,117,116,112,117,116,32,102,111,114,109,97,116,32,105,115,32,100,105,115,97,98,108,101,100,32,105,110,32,116,104,105,115,32,98,117,105,108,100,33,10,0,91,111,112,116,105,109,105,122,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,115,101,116,95,115,121,110,116,104,95,102,117,110,99,116,105,111,110,115,58,32,84,104,105,115,32,114,101,115,97,109,112,108,105,110,103,32,109,111,100,101,32,105,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,105,110,32,116,104,105,115,32,98,117,105,108,100,33,10,0,91,111,112,116,105,109,105,122,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,85,110,97,98,108,101,32,116,111,32,100,101,116,101,114,109,105,110,101,32,97,99,116,105,118,101,32,100,101,99,111,100,101,114,32,116,121,112,101,32,45,45,32,116,104,105,115,32,105,115,32,83,69,82,73,79,85,83,32,98,48,114,107,97,103,101,33,10,0,91,111,112,116,105,109,105,122,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,70,97,105,108,101,100,32,116,111,32,115,101,116,32,117,112,32,100,101,99,111,100,101,114,32,98,117,102,102,101,114,115,33,10,0,91,116,97,98,105,110,105,116,46,99,58,37,105,93,32,101,114,114,111,114,58,32,67,97,110,39,116,32,97,108,108,111,99,97,116,101,32,49,54,32,116,111,32,56,32,99,111,110,118,101,114,116,101,114,32,116,97,98,108,101,33,10,0,91,116,97,98,105,110,105,116,46,99,58,37,105,93,32,101,114,114,111,114,58,32,67,111,110,118,101,114,114,111,114,32,37,100,32,37,100,10,0,91,111,112,116,105,109,105,122,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,70,97,105,108,101,100,32,116,111,32,115,101,116,32,117,112,32,99,111,110,118,49,54,116,111,56,32,116,97,98,108,101,33,10,0,97,117,116,111,0,103,101,110,101,114,105,99,95,100,105,116,104,101,114,0,105,51,56,54,0,105,52,56,54,0,105,53,56,54,0,105,53,56,54,95,100,105,116,104,101,114,0,77,77,88,0,51,68,78,111,119,0,51,68,78,111,119,69,120,116,0,65,108,116,105,86,101,99,0,83,83,69,0,120,56,54,45,54,52,0,65,82,77,0,78,69,79,78,0,91,111,112,116,105,109,105,122,101,46,99,58,37,105,93,32,101,114,114,111,114,58,32,121,111,117,32,119,97,110,116,101,100,32,100,101,99,111,100,101,114,32,116,121,112,101,32,37,105,44,32,73,32,111,110,108,121,32,104,97,118,101,32,37,105,10,0,68,101,99,111,100,101,114,58,32,37,115,10,0,91,114,101,97,100,101,114,115,46,99,58,37,105,93,32,101,114,114,111,114,58,32,70,97,105,108,101,100,32,116,111,32,97,100,100,32,98,117,102,102,101,114,44,32,114,101,116,117,114,110,58,32,37,105,10,0,91,114,101,97,100,101,114,115,46,99,58,37,105,93,32,101,114,114,111,114,58,32,84,105,109,101,111,117,116,32,114,101,97,100,105,110,103,32,100,111,101,115,32,110,111,116,32,119,111,114,107,32,119,105,116,104,32,117,115,101,114,45,112,114,111,118,105,100,101,100,32,114,101,97,100,32,102,117,110,99,116,105,111,110,46,32,73,109,112,108,101,109,101,110,116,32,105,116,32,121,111,117,114,115,101,108,102,33,10,0,91,114,101,97,100,101,114,115,46,99,58,37,105,93,32,101,114,114,111,114,58,32,115,116,114,101,97,109,32,116,105,109,101,100,32,111,117,116,10,0,84,65,71,0,91,114,101,97,100,101,114,115,46,99,58,37,105,93,32,101,114,114,111,114,58,32,109,112,103,49,50,51,32,112,114,111,103,114,97,109,109,101,114,32,101,114,114,111,114,58,32,73,32,100,111,110,39,116,32,100,111,32,73,67,89,32,111,110,32,115,101,101,107,97,98,108,101,32,115,116,114,101,97,109,115,46,10,0,91,114,101,97,100,101,114,115,46,99,58,37,105,93,32,101,114,114,111,114,58,32,105,99,121,32,98,111,117,110,100,97,114,121,32,114,101,97,100,10,0,91,114,101,97,100,101,114,115,46,99,58,37,105,93,32,101,114,114,111,114,58,32,114,101,97,100,105,110,103,32,105,99,121,32,115,105,122,101,10,0,91,114,101,97,100,101,114,115,46,99,58,37,105,93,32,101,114,114,111,114,58,32,114,101,97,100,105,110,103,32,105,99,121,45,109,101,116,97,10,0,91,114,101,97,100,101,114,115,46,99,58,37,105,93,32,101,114,114,111,114,58,32,99,97,110,110,111,116,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,109,101,116,97,95,98,117,102,102,32,40,37,108,117,32,98,121,116,101,115,41,32,46,46,46,32,116,114,121,105,110,103,32,116,111,32,115,107,105,112,32,116,104,101,32,109,101,116,97,100,97,116,97,33,10,0,91,114,101,97,100,101,114,115,46,99,58,37,105,93,32,101,114,114,111,114,58,32,114,101,97,100,105,110,103,32,116,104,101,32,114,101,115,116,32,111,102,32,37,108,105,10,0,91,114,101,97,100,101,114,115,46,99,58,37,105,93,32,101,114,114,111,114,58,32,109,112,103,49,50,51,32,80,114,111,103,114,97,109,109,101,114,39,115,32,102,97,117,108,116,58,32,105,110,118,97,108,105,100,32,114,101,97,100,101,114,10,0,91,114,101,97,100,101,114,115,46,99,58,37,105,93,32,101,114,114,111,114,58,32,98,117,102,102,101,114,32,114,101,97,100,105,110,103,10,0,91,114,101,97,100,101,114,115,46,99,58,37,105,93,32,101,114,114,111,114,58,32,117,110,97,98,108,101,32,116,111,32,97,100,100,32,116,111,32,99,104,97,105,110,44,32,114,101,116,117,114,110,58,32,37,105,10,0,78,111,116,101,58,32,73,110,112,117,116,32,100,97,116,97,32,101,110,100,46,10,0,91,114,101,97,100,101,114,115,46,99,58,37,105,93,32,101,114,114,111,114,58,32,103,111,116,99,111,117,110,116,32,33,61,32,99,111,117,110,116,10,0,91,108,105,98,109,112,103,49,50,51,46,99,58,37,105,93,32,101,114,114,111,114,58,32,100,101,99,111,100,101,95,117,112,100,97,116,101,40,41,32,104,97,115,32,98,101,101,110,32,99,97,108,108,101,100,32,98,101,102,111,114,101,32,114,101,97,100,105,110,103,32,116,104,101,32,102,105,114,115,116,32,77,80,69,71,32,102,114,97,109,101,33,32,73,110,116,101,114,110,97,108,32,112,114,111,103,114,97,109,109,105,110,103,32,101,114,114,111,114,46,10,0,73,110,105,116,32,114,97,116,101,32,99,111,110,118,101,114,116,101,114,58,32,37,108,100,45,62,37,108,100,10,0,91,110,116,111,109,46,99,58,37,105,93,32,101,114,114,111,114,58,32,78,116,111,77,32,99,111,110,118,101,114,116,101,114,58,32,105,108,108,101,103,97,108,32,114,97,116,101,115,10,0,91,110,116,111,109,46,99,58,37,105,93,32,101,114,114,111,114,58,32,109,97,120,46,32,49,58,37,105,32,99,111,110,118,101,114,115,105,111,110,32,97,108,108,111,119,101,100,32,40,37,108,117,32,118,115,32,37,108,117,41,33,10,0,78,111,116,101,58,32,98,114,111,107,101,110,32,102,114,97,109,101,32,37,108,105,44,32,102,105,108,108,105,110,103,32,117,112,32,119,105,116,104,32,37,108,117,32,122,101,114,111,101,115,44,32,102,114,111,109,32,37,108,117,10,0,10,78,111,116,101,58,32,67,117,116,32,102,114,97,109,101,32,37,108,105,32,98,117,102,102,101,114,32,111,110,32,101,110,100,32,111,102,32,115,116,114,101,97,109,32,116,111,32,37,108,105,32,115,97,109,112,108,101,115,44,32,102,105,108,108,32,110,111,119,32,37,108,117,32,98,121,116,101,115,46,10,0,10,78,111,116,101,58,32,67,117,116,32,102,114,97,109,101,32,37,108,105,32,98,117,102,102,101,114,32,111,110,32,98,101,103,105,110,110,105,110,103,32,111,102,32,115,116,114,101,97,109,32,98,121,32,37,108,105,32,115,97,109,112,108,101,115,44,32,102,105,108,108,32,110,111,119,32,37,108,117,32,98,121,116,101,115,46,10,0,65,108,114,101,97,100,121,32,105,110,105,116,105,97,108,105,122,101,100,0,33,101,120,116,101,114,110,97,108,65,108,108,111,99,97,116,111,114,0,47,85,115,101,114,115,47,116,97,110,110,101,114,47,99,111,100,101,47,99,47,108,111,118,101,47,109,101,103,97,115,111,117,114,99,101,47,108,105,98,115,47,112,104,121,115,102,115,45,50,46,48,46,51,47,112,104,121,115,102,115,46,99,0,115,101,116,68,101,102,97,117,108,116,65,108,108,111,99,97,116,111,114,0,97,114,103,118,48,32,105,115,32,78,85,76,76,0,37,115,117,115,101,114,115,37,115,37,115,0,78,111,116,32,105,110,105,116,105,97,108,105,122,101,100,0,70,105,108,101,115,32,115,116,105,108,108,32,111,112,101,110,0,73,110,115,101,99,117,114,101,32,102,105,108,101,110,97,109,101,0,78,111,32,115,117,99,104,32,101,110,116,114,121,32,105,110,32,115,101,97,114,99,104,32,112,97,116,104,0,87,114,105,116,101,32,100,105,114,101,99,116,111,114,121,32,105,115,32,110,111,116,32,115,101,116,0,109,110,116,112,110,116,108,101,110,32,62,32,49,0,118,101,114,105,102,121,80,97,116,104,0,80,97,116,104,32,110,111,116,32,102,111,117,110,100,0,83,121,109,98,111,108,105,99,32,108,105,110,107,115,32,97,114,101,32,100,105,115,97,98,108,101,100,0,101,110,117,109,101,114,97,116,101,70,114,111,109,77,111,117,110,116,80,111,105,110,116,0,78,111,116,32,97,32,102,105,108,101,32,104,97,110,100,108,101,0,70,105,108,101,32,97,108,114,101,97,100,121,32,111,112,101,110,32,102,111,114,32,119,114,105,116,105,110,103,0,70,105,108,101,32,97,108,114,101,97,100,121,32,111,112,101,110,32,102,111,114,32,114,101,97,100,105,110,103,0,98,117,102,102,101,114,32,109,117,115,116,32,102,105,116,32,105,110,32,51,50,45,98,105,116,115,0,85,83,69,82,0,72,79,77,69,0,114,99,32,60,61,32,109,97,120,0,47,85,115,101,114,115,47,116,97,110,110,101,114,47,99,111,100,101,47,99,47,108,111,118,101,47,109,101,103,97,115,111,117,114,99,101,47,108,105,98,115,47,112,104,121,115,102,115,45,50,46,48,46,51,47,112,108,97,116,102,111,114,109,47,112,111,115,105,120,46,99,0,95,95,80,72,89,83,70,83,95,112,108,97,116,102,111,114,109,82,101,97,100,0,95,95,80,72,89,83,70,83,95,112,108,97,116,102,111,114,109,87,114,105,116,101,0,47,112,114,111,99,47,115,101,108,102,47,101,120,101,0,47,112,114,111,99,47,37,108,108,117,47,101,120,101,0,80,65,84,72,0,73,110,118,97,108,105,100,32,97,114,103,117,109,101,110,116,0,78,111,110,45,97,114,99,104,105,118,101,44,32,100,105,114,101,99,116,32,102,105,108,101,115,121,115,116,101,109,32,73,47,79,0,90,73,80,0,80,107,90,105,112,47,87,105,110,90,105,112,47,73,110,102,111,45,90,105,112,32,99,111,109,112,97,116,105,98,108,101,0,82,121,97,110,32,67,46,32,71,111,114,100,111,110,32,60,105,99,99,117,108,117,115,64,105,99,99,117,108,117,115,46,111,114,103,62,0,104,116,116,112,58,47,47,105,99,99,117,108,117,115,46,111,114,103,47,112,104,121,115,102,115,47,0,90,73,80,32,98,105,103,103,101,114,32,116,104,97,110,32,50,32,103,105,103,115,63,33,0,78,111,116,32,97,110,32,97,114,99,104,105,118,101,0,65,114,99,104,105,118,101,32,105,115,32,114,101,97,100,45,111,110,108,121,0,65,114,99,104,105,118,101,32,116,121,112,101,32,117,110,115,117,112,112,111,114,116,101,100,0,67,111,114,114,117,112,116,101,100,32,97,114,99,104,105,118,101,0,70,105,108,101,32,110,111,116,32,102,111,117,110,100,0,73,110,102,105,110,105,116,101,32,115,121,109,98,111,108,105,99,32,108,105,110,107,32,108,111,111,112,0,109,101,109,111,114,121,32,101,114,114,111,114,0,118,101,114,115,105,111,110,32,101,114,114,111,114,0,117,110,107,110,111,119,110,32,101,114,114,111,114,0,78,111,116,32,97,32,100,105,114,101,99,116,111,114,121,0,79,112,101,114,97,116,105,111,110,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,80,97,115,116,32,101,110,100,32,111,102,32,102,105,108,101,0,110,111,32,99,97,108,108,105,110,103,32,101,110,118,105,114,111,110,109,101,110,116,0,40,42,116,101,109,112,111,114,97,114,121,41,0,116,97,105,108,0,61,40,116,97,105,108,32,99,97,108,108,41,0,61,91,67,93,0,109,97,105,110,0,76,117,97,0,103,108,111,98,97,108,0,102,105,101,108,100,0,117,112,118,97,108,117,101,0,97,116,116,101,109,112,116,32,116,111,32,37,115,32,37,115,32,39,37,115,39,32,40,97,32,37,115,32,118,97,108,117,101,41,0,97,116,116,101,109,112,116,32,116,111,32,37,115,32,97,32,37,115,32,118,97,108,117,101,0,99,111,110,99,97,116,101,110,97,116,101,0,112,101,114,102,111,114,109,32,97,114,105,116,104,109,101,116,105,99,32,111,110,0,97,116,116,101,109,112,116,32,116,111,32,99,111,109,112,97,114,101,32,116,119,111,32,37,115,32,118,97,108,117,101,115,0,97,116,116,101,109,112,116,32,116,111,32,99,111,109,112,97,114,101,32,37,115,32,119,105,116,104,32,37,115,0,101,114,114,111,114,32,105,110,32,101,114,114,111,114,32,104,97,110,100,108,105,110,103,0,115,116,97,99,107,32,111,118,101,114,102,108,111,119,0,67,32,115,116,97,99,107,32,111,118,101,114,102,108,111,119,0,99,97,110,110,111,116,32,114,101,115,117,109,101,32,110,111,110,45,115,117,115,112,101,110,100,101,100,32,99,111,114,111,117,116,105,110,101,0,97,116,116,101,109,112,116,32,116,111,32,121,105,101,108,100,32,97,99,114,111,115,115,32,109,101,116,97,109,101,116,104,111,100,47,67,45,99,97,108,108,32,98,111,117,110,100,97,114,121,0,109,101,109,111,114,121,32,97,108,108,111,99,97,116,105,111,110,32,101,114,114,111,114,58,32,98,108,111,99,107,32,116,111,111,32,98,105,103,0,0,1,2,2,3,3,3,3,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,10,13,0,91,115,116,114,105,110,103,32,34,0,34,93,0,96,113,84,96,80,113,108,49,16,60,84,108,124,124,124,124,124,124,96,96,96,104,34,188,188,188,228,228,84,84,16,98,98,132,20,0,81,80,99,104,117,110,107,32,104,97,115,32,116,111,111,32,109,97,110,121,32,115,121,110,116,97,120,32,108,101,118,101,108,115,0,99,97,110,110,111,116,32,117,115,101,32,39,46,46,46,39,32,111,117,116,115,105,100,101,32,97,32,118,97,114,97,114,103,32,102,117,110,99,116,105,111,110,0,39,37,115,39,32,101,120,112,101,99,116,101,100,0,109,97,105,110,32,102,117,110,99,116,105,111,110,32,104,97,115,32,109,111,114,101,32,116,104,97,110,32,37,100,32,37,115,0,105,116,101,109,115,32,105,110,32,97,32,99,111,110,115,116,114,117,99,116,111,114,0,102,117,110,99,116,105,111,110,32,97,116,32,108,105,110,101,32,37,100,32,104,97,115,32,109,111,114,101,32,116,104,97,110,32,37,100,32,37,115,0,39,37,115,39,32,101,120,112,101,99,116,101,100,32,40,116,111,32,99,108,111,115,101,32,39,37,115,39,32,97,116,32,108,105,110,101,32,37,100,41,0,115,101,108,102,0,108,111,99,97,108,32,118,97,114,105,97,98,108,101,115,0,116,111,111,32,109,97,110,121,32,108,111,99,97,108,32,118,97,114,105,97,98,108,101,115,0,97,114,103,0,60,110,97,109,101,62,32,111,114,32,39,46,46,46,39,32,101,120,112,101,99,116,101,100,0,117,112,118,97,108,117,101,115,0,117,110,101,120,112,101,99,116,101,100,32,115,121,109,98,111,108,0,97,109,98,105,103,117,111,117,115,32,115,121,110,116,97,120,32,40,102,117,110,99,116,105,111,110,32,99,97,108,108,32,120,32,110,101,119,32,115,116,97,116,101,109,101,110,116,41,0,102,117,110,99,116,105,111,110,32,97,114,103,117,109,101,110,116,115,32,101,120,112,101,99,116,101,100,0,6,6,6,6,7,7,7,7,7,7,10,9,5,4,3,3,3,3,3,3,3,3,3,3,3,3,2,2,1,1,40,102,111,114,32,105,110,100,101,120,41,0,40,102,111,114,32,108,105,109,105,116,41,0,40,102,111,114,32,115,116,101,112,41,0,40,102,111,114,32,103,101,110,101,114,97,116,111,114,41,0,40,102,111,114,32,115,116,97,116,101,41,0,40,102,111,114,32,99,111,110,116,114,111,108,41,0,39,61,39,32,111,114,32,39,105,110,39,32,101,120,112,101,99,116,101,100,0,110,111,32,108,111,111,112,32,116,111,32,98,114,101,97,107,0,115,121,110,116,97,120,32,101,114,114,111,114,0,118,97,114,105,97,98,108,101,115,32,105,110,32,97,115,115,105,103,110,109,101,110,116,0,110,111,116,32,101,110,111,117,103,104,32,109,101,109,111,114,121,0,105,110,118,97,108,105,100,32,107,101,121,32,116,111,32,39,110,101,120,116,39,0,116,97,98,108,101,32,111,118,101,114,102,108,111,119,0,116,97,98,108,101,32,105,110,100,101,120,32,105,115,32,110,105,108,0,116,97,98,108,101,32,105,110,100,101,120,32,105,115,32,78,97,78,0,98,111,111,108,101,97,110,0,117,115,101,114,100,97,116,97,0,110,117,109,98,101,114,0,116,104,114,101,97,100,0,112,114,111,116,111,0,117,112,118,97,108,0,95,95,110,101,119,105,110,100,101,120,0,95,95,101,113,0,95,95,97,100,100,0,95,95,115,117,98,0,95,95,109,117,108,0,95,95,100,105,118,0,95,95,109,111,100,0,95,95,112,111,119,0,95,95,117,110,109,0,95,95,108,101,110,0,95,95,108,116,0,95,95,108,101,0,95,95,99,111,110,99,97,116,0,95,95,99,97,108,108,0,98,105,110,97,114,121,32,115,116,114,105,110,103,0,37,115,58,32,37,115,32,105,110,32,112,114,101,99,111,109,112,105,108,101,100,32,99,104,117,110,107,0,117,110,101,120,112,101,99,116,101,100,32,101,110,100,0,98,97,100,32,104,101,97,100,101,114,0,61,63,0,99,111,100,101,32,116,111,111,32,100,101,101,112,0,98,97,100,32,105,110,116,101,103,101,114,0,98,97,100,32,99,111,110,115,116,97,110,116,0,98,97,100,32,99,111,100,101,0,105,110,100,101,120,0,108,111,111,112,32,105,110,32,103,101,116,116,97,98,108,101,0,108,111,111,112,32,105,110,32,115,101,116,116,97,98,108,101,0,115,116,114,105,110,103,32,108,101,110,103,116,104,32,111,118,101,114,102,108,111,119,0,103,101,116,32,108,101,110,103,116,104,32,111,102,0,39,102,111,114,39,32,105,110,105,116,105,97,108,32,118,97,108,117,101,32,109,117,115,116,32,98,101,32,97,32,110,117,109,98,101,114,0,39,102,111,114,39,32,108,105,109,105,116,32,109,117,115,116,32,98,101,32,97,32,110,117,109,98,101,114,0,39,102,111,114,39,32,115,116,101,112,32,109,117,115,116,32,98,101,32,97,32,110,117,109,98,101,114,0,98,97,100,32,97,114,103,117,109,101,110,116,32,35,37,100,32,40,37,115,41,0,83,108,0,37,115,58,37,100,58,32,0,110,0,109,101,116,104,111,100,0,99,97,108,108,105,110,103,32,39,37,115,39,32,111,110,32,98,97,100,32,115,101,108,102,32,40,37,115,41,0,98,97,100,32,97,114,103,117,109,101,110,116,32,35,37,100,32,116,111,32,39,37,115,39,32,40,37,115,41,0,37,115,32,101,120,112,101,99,116,101,100,44,32,103,111,116,32,37,115,0,105,110,118,97,108,105,100,32,111,112,116,105,111,110,32,39,37,115,39,0,115,116,97,99,107,32,111,118,101,114,102,108,111,119,32,40,37,115,41,0,118,97,108,117,101,32,101,120,112,101,99,116,101,100,0,61,115,116,100,105,110,0,64,37,115,0,99,97,110,110,111,116,32,37,115,32,37,115,58,32,37,115,0,114,101,111,112,101,110,0,80,65,78,73,67,58,32,117,110,112,114,111,116,101,99,116,101,100,32,101,114,114,111,114,32,105,110,32,99,97,108,108,32,116,111,32,76,117,97,32,65,80,73,32,40,37,115,41,10,0,99,111,110,116,114,111,108,32,115,116,114,117,99,116,117,114,101,32,116,111,111,32,108,111,110,103,0,99,111,100,101,32,115,105,122,101,32,111,118,101,114,102,108,111,119,0,102,117,110,99,116,105,111,110,32,111,114,32,101,120,112,114,101,115,115,105,111,110,32,116,111,111,32,99,111,109,112,108,101,120,0,99,111,110,115,116,97,110,116,32,116,97,98,108,101,32,111,118,101,114,102,108,111,119,0,97,110,100,0,98,114,101,97,107,0,100,111,0,101,108,115,101,0,101,108,115,101,105,102,0,102,111,114,0,102,117,110,99,116,105,111,110,0,105,102,0,105,110,0,108,111,99,97,108,0,110,111,116,0,111,114,0,114,101,112,101,97,116,0,116,104,101,110,0,117,110,116,105,108,0,119,104,105,108,101,0,46,46,0,46,46,46,0,61,61,0,62,61,0,60,61,0,126,61,0,60,110,117,109,98,101,114,62,0,60,110,97,109,101,62,0,60,115,116,114,105,110,103,62,0,60,101,111,102,62,0,99,104,97,114,40,37,100,41,0,37,115,58,37,100,58,32,37,115,0,108,101,120,105,99,97,108,32,101,108,101,109,101,110,116,32,116,111,111,32,108,111,110,103,0,37,115,32,110,101,97,114,32,39,37,115,39,0,99,104,117,110,107,32,104,97,115,32,116,111,111,32,109,97,110,121,32,108,105,110,101,115,0,117,110,102,105,110,105,115,104,101,100,32,108,111,110,103,32,115,116,114,105,110,103,0,117,110,102,105,110,105,115,104,101,100,32,108,111,110,103,32,99,111,109,109,101,110,116,0,110,101,115,116,105,110,103,32,111,102,32,91,91,46,46,46,93,93,32,105,115,32,100,101,112,114,101,99,97,116,101,100,0,105,110,118,97,108,105,100,32,108,111,110,103,32,115,116,114,105,110,103,32,100,101,108,105,109,105,116,101,114,0,117,110,102,105,110,105,115,104,101,100,32,115,116,114,105,110,103,0,101,115,99,97,112,101,32,115,101,113,117,101,110,99,101,32,116,111,111,32,108,97,114,103,101,0,69,101,0,109,97,108,102,111,114,109,101,100,32,110,117,109,98,101,114,0,95,71,0,97,115,115,101,114,116,0,97,115,115,101,114,116,105,111,110,32,102,97,105,108,101,100,33,0,99,111,108,108,101,99,116,103,97,114,98,97,103,101,0,99,111,108,108,101,99,116,0,115,116,111,112,0,114,101,115,116,97,114,116,0,115,116,101,112,0,115,101,116,112,97,117,115,101,0,115,101,116,115,116,101,112,109,117,108,0,100,111,102,105,108,101,0,101,114,114,111,114,0,103,99,105,110,102,111,0,108,101,118,101,108,32,109,117,115,116,32,98,101,32,110,111,110,45,110,101,103,97,116,105,118,101,0,105,110,118,97,108,105,100,32,108,101,118,101,108,0,110,111,32,102,117,110,99,116,105,111,110,32,101,110,118,105,114,111,110,109,101,110,116,32,102,111,114,32,116,97,105,108,32,99,97,108,108,32,97,116,32,108,101,118,101,108,32,37,100,0,95,95,109,101,116,97,116,97,98,108,101,0,108,111,97,100,102,105,108,101,0,108,111,97,100,0,61,40,108,111,97,100,41,0,116,111,111,32,109,97,110,121,32,110,101,115,116,101,100,32,102,117,110,99,116,105,111,110,115,0,114,101,97,100,101,114,32,102,117,110,99,116,105,111,110,32,109,117,115,116,32,114,101,116,117,114,110,32,97,32,115,116,114,105,110,103,0,108,111,97,100,115,116,114,105,110,103,0,110,101,120,116,0,112,99,97,108,108,0,112,114,105,110,116,0,116,111,115,116,114,105,110,103,0,39,116,111,115,116,114,105,110,103,39,32,109,117,115,116,32,114,101,116,117,114,110,32,97,32,115,116,114,105,110,103,32,116,111,32,39,112,114,105,110,116,39,0,114,97,119,101,113,117,97,108,0,114,97,119,103,101,116,0,114,97,119,115,101,116,0,115,101,108,101,99,116,0,105,110,100,101,120,32,111,117,116,32,111,102,32,114,97,110,103,101,0,99,97,110,110,111,116,32,99,104,97,110,103,101,32,97,32,112,114,111,116,101,99,116,101,100,32,109,101,116,97,116,97,98,108,101,0,116,111,110,117,109,98,101,114,0,98,97,115,101,32,111,117,116,32,111,102,32,114,97,110,103,101,0,110,105,108,0,37,115,58,32,37,112,0,117,110,112,97,99,107,0,116,111,111,32,109,97,110,121,32,114,101,115,117,108,116,115,32,116,111,32,117,110,112,97,99,107,0,120,112,99,97,108,108,0,76,117,97,32,53,46,49,0,95,86,69,82,83,73,79,78,0,105,112,97,105,114,115,0,112,97,105,114,115,0,107,118,0,95,95,109,111,100,101,0,98,111,111,108,101,97,110,32,111,114,32,112,114,111,120,121,32,101,120,112,101,99,116,101,100,0,110,101,119,112,114,111,120,121,0,99,111,114,111,117,116,105,110,101,0,99,114,101,97,116,101,0,76,117,97,32,102,117,110,99,116,105,111,110,32,101,120,112,101,99,116,101,100,0,114,101,115,117,109,101,0,99,111,114,111,117,116,105,110,101,32,101,120,112,101,99,116,101,100,0,116,111,111,32,109,97,110,121,32,97,114,103,117,109,101,110,116,115,32,116,111,32,114,101,115,117,109,101,0,114,117,110,110,105,110,103,0,115,117,115,112,101,110,100,101,100,0,110,111,114,109,97,108,0,100,101,97,100,0,99,97,110,110,111,116,32,114,101,115,117,109,101,32,37,115,32,99,111,114,111,117,116,105,110,101,0,116,111,111,32,109,97,110,121,32,114,101,115,117,108,116,115,32,116,111,32,114,101,115,117,109,101,0,115,116,97,116,117,115,0,119,114,97,112,0,121,105,101,108,100,0,100,101,98,117,103,0,108,117,97,95,100,101,98,117,103,62,32,0,99,111,110,116,10,0,61,40,100,101,98,117,103,32,99,111,109,109,97,110,100,41,0,103,101,116,102,101,110,118,0,103,101,116,104,111,111,107,0,104,99,97,108,108,0,114,101,116,117,114,110,0,116,97,105,108,32,114,101,116,117,114,110,0,101,120,116,101,114,110,97,108,32,104,111,111,107,0,103,101,116,105,110,102,111,0,102,108,110,83,117,0,62,37,115,0,102,117,110,99,116,105,111,110,32,111,114,32,108,101,118,101,108,32,101,120,112,101,99,116,101,100,0,115,104,111,114,116,95,115,114,99,0,108,105,110,101,100,101,102,105,110,101,100,0,108,97,115,116,108,105,110,101,100,101,102,105,110,101,100,0,119,104,97,116,0,99,117,114,114,101,110,116,108,105,110,101,0,110,117,112,115,0,110,97,109,101,119,104,97,116,0,97,99,116,105,118,101,108,105,110,101,115,0,102,117,110,99,0,103,101,116,108,111,99,97,108,0,108,101,118,101,108,32,111,117,116,32,111,102,32,114,97,110,103,101,0,103,101,116,114,101,103,105,115,116,114,121,0,103,101,116,109,101,116,97,116,97,98,108,101,0,103,101,116,117,112,118,97,108,117,101,0,115,101,116,102,101,110,118,0,39,115,101,116,102,101,110,118,39,32,99,97,110,110,111,116,32,99,104,97,110,103,101,32,101,110,118,105,114,111,110,109,101,110,116,32,111,102,32,103,105,118,101,110,32,111,98,106,101,99,116,0,115,101,116,104,111,111,107,0,115,101,116,108,111,99,97,108,0,115,101,116,109,101,116,97,116,97,98,108,101,0,110,105,108,32,111,114,32,116,97,98,108,101,32,101,120,112,101,99,116,101,100,0,115,101,116,117,112,118,97,108,117,101,0,116,114,97,99,101,98,97,99,107,0,10,0,115,116,97,99,107,32,116,114,97,99,101,98,97,99,107,58,0,10,9,46,46,46,0,10,9,0,83,110,108,0,37,115,58,0,37,100,58,0,32,105,110,32,102,117,110,99,116,105,111,110,32,39,37,115,39,0,32,105,110,32,109,97,105,110,32,99,104,117,110,107,0,32,63,0,32,105,110,32,102,117,110,99,116,105,111,110,32,60,37,115,58,37,100,62,0,70,73,76,69,42,0,99,108,111,115,101,0,97,116,116,101,109,112,116,32,116,111,32,117,115,101,32,97,32,99,108,111,115,101,100,32,102,105,108,101,0,95,95,99,108,111,115,101,0,102,108,117,115,104,0,108,105,110,101,115,0,102,105,108,101,32,105,115,32,97,108,114,101,97,100,121,32,99,108,111,115,101,100,0,114,101,97,100,0,116,111,111,32,109,97,110,121,32,97,114,103,117,109,101,110,116,115,0,105,110,118,97,108,105,100,32,111,112,116,105,111,110,0,37,108,102,0,105,110,118,97,108,105,100,32,102,111,114,109,97,116,0,115,101,101,107,0,99,117,114,0,115,101,116,0,101,110,100,0,115,101,116,118,98,117,102,0,102,117,108,108,0,108,105,110,101,0,119,114,105,116,101,0,37,46,49,52,103,0,95,95,116,111,115,116,114,105,110,103,0,102,105,108,101,32,40,99,108,111,115,101,100,41,0,102,105,108,101,32,40,37,112,41,0,105,111,0,115,116,97,110,100,97,114,100,32,37,115,32,102,105,108,101,32,105,115,32,99,108,111,115,101,100,0,111,117,116,112,117,116,0,105,110,112,117,116,0,111,112,101,110,0,112,111,112,101,110,0,39,112,111,112,101,110,39,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,116,109,112,102,105,108,101,0,116,121,112,101,0,99,108,111,115,101,100,32,102,105,108,101,0,102,105,108,101,0,99,97,110,110,111,116,32,99,108,111,115,101,32,115,116,97,110,100,97,114,100,32,102,105,108,101,0,115,116,100,105,110,0,115,116,100,111,117,116,0,115,116,100,101,114,114,0,109,97,116,104,0,97,98,115,0,97,99,111,115,0,97,115,105,110,0,97,116,97,110,50,0,97,116,97,110,0,99,101,105,108,0,99,111,115,104,0,99,111,115,0,100,101,103,0,101,120,112,0,102,108,111,111,114,0,102,109,111,100,0,102,114,101,120,112,0,108,100,101,120,112,0,108,111,103,49,48,0,108,111,103,0,109,97,120,0,109,111,100,102,0,112,111,119,0,114,97,100,0,114,97,110,100,111,109,0,105,110,116,101,114,118,97,108,32,105,115,32,101,109,112,116,121,0,119,114,111,110,103,32,110,117,109,98,101,114,32,111,102,32,97,114,103,117,109,101,110,116,115,0,114,97,110,100,111,109,115,101,101,100,0,115,105,110,104,0,115,105,110,0,115,113,114,116,0,116,97,110,104,0,116,97,110,0,112,105,0,104,117,103,101,0,109,111,100,0,111,115,0,99,108,111,99,107,0,100,97,116,101,0,37,99,0,42,116,0,115,101,99,0,104,111,117,114,0,100,97,121,0,109,111,110,116,104,0,121,101,97,114,0,119,100,97,121,0,121,100,97,121,0,105,115,100,115,116,0,100,105,102,102,116,105,109,101,0,101,120,101,99,117,116,101,0,101,120,105,116,0,103,101,116,101,110,118,0,37,115,58,32,37,115,0,114,101,110,97,109,101,0,115,101,116,108,111,99,97,108,101,0,99,111,108,108,97,116,101,0,99,116,121,112,101,0,109,111,110,101,116,97,114,121,0,110,117,109,101,114,105,99,0,116,105,109,101,0,102,105,101,108,100,32,39,37,115,39,32,109,105,115,115,105,110,103,32,105,110,32,100,97,116,101,32,116,97,98,108,101,0,116,109,112,110,97,109,101,0,117,110,97,98,108,101,32,116,111,32,103,101,110,101,114,97,116,101,32,97,32,117,110,105,113,117,101,32,102,105,108,101,110,97,109,101,0,116,97,98,108,101,0,99,111,110,99,97,116,0,105,110,118,97,108,105,100,32,118,97,108,117,101,32,40,37,115,41,32,97,116,32,105,110,100,101,120,32,37,100,32,105,110,32,116,97,98,108,101,32,102,111,114,32,39,99,111,110,99,97,116,39,0,102,111,114,101,97,99,104,0,102,111,114,101,97,99,104,105,0,103,101,116,110,0,109,97,120,110,0,105,110,115,101,114,116,0,119,114,111,110,103,32,110,117,109,98,101,114,32,111,102,32,97,114,103,117,109,101,110,116,115,32,116,111,32,39,105,110,115,101,114,116,39,0,114,101,109,111,118,101,0,115,101,116,110,0,39,115,101,116,110,39,32,105,115,32,111,98,115,111,108,101,116,101,0,115,111,114,116,0,105,110,118,97,108,105,100,32,111,114,100,101,114,32,102,117,110,99,116,105,111,110,32,102,111,114,32,115,111,114,116,105,110,103,0,115,116,114,105,110,103,0,98,121,116,101,0,115,116,114,105,110,103,32,115,108,105,99,101,32,116,111,111,32,108,111,110,103,0,99,104,97,114,0,105,110,118,97,108,105,100,32,118,97,108,117,101,0,100,117,109,112,0,117,110,97,98,108,101,32,116,111,32,100,117,109,112,32,103,105,118,101,110,32,102,117,110,99,116,105,111,110,0,102,105,110,100,0,94,36,42,43,63,46,40,91,37,45,0,116,111,111,32,109,97,110,121,32,99,97,112,116,117,114,101,115,0,105,110,118,97,108,105,100,32,112,97,116,116,101,114,110,32,99,97,112,116,117,114,101,0,117,110,98,97,108,97,110,99,101,100,32,112,97,116,116,101,114,110,0,109,105,115,115,105,110,103,32,39,91,39,32,97,102,116,101,114,32,39,37,37,102,39,32,105,110,32,112,97,116,116,101,114,110,0,109,97,108,102,111,114,109,101,100,32,112,97,116,116,101,114,110,32,40,101,110,100,115,32,119,105,116,104,32,39,37,37,39,41,0,109,97,108,102,111,114,109,101,100,32,112,97,116,116,101,114,110,32,40,109,105,115,115,105,110,103,32,39,93,39,41,0,105,110,118,97,108,105,100,32,99,97,112,116,117,114,101,32,105,110,100,101,120,0,117,110,102,105,110,105,115,104,101,100,32,99,97,112,116,117,114,101,0,110,111,32,118,97,108,117,101,0,45,43,32,35,48,0,105,110,118,97,108,105,100,32,102,111,114,109,97,116,32,40,114,101,112,101,97,116,101,100,32,102,108,97,103,115,41,0,105,110,118,97,108,105,100,32,102,111,114,109,97,116,32,40,119,105,100,116,104,32,111,114,32,112,114,101,99,105,115,105,111,110,32,116,111,111,32,108,111,110,103,41,0,92,114,0,92,48,48,48,0,105,110,118,97,108,105,100,32,111,112,116,105,111,110,32,39,37,37,37,99,39,32,116,111,32,39,102,111,114,109,97,116,39,0,103,102,105,110,100,0,39,115,116,114,105,110,103,46,103,102,105,110,100,39,32,119,97,115,32,114,101,110,97,109,101,100,32,116,111,32,39,115,116,114,105,110,103,46,103,109,97,116,99,104,39,0,103,109,97,116,99,104,0,103,115,117,98,0,115,116,114,105,110,103,47,102,117,110,99,116,105,111,110,47,116,97,98,108,101,32,101,120,112,101,99,116,101,100,0,105,110,118,97,108,105,100,32,114,101,112,108,97,99,101,109,101,110,116,32,118,97,108,117,101,32,40,97,32,37,115,41,0,108,101,110,0,108,111,119,101,114,0,109,97,116,99,104,0,114,101,112,0,114,101,118,101,114,115,101,0,115,117,98,0,117,112,112,101,114,0,95,76,79,65,68,76,73,66,0,95,95,103,99,0,112,97,99,107,97,103,101,0,108,111,97,100,108,105,98,0,76,79,65,68,76,73,66,58,32,0,100,121,110,97,109,105,99,32,108,105,98,114,97,114,105,101,115,32,110,111,116,32,101,110,97,98,108,101,100,59,32,99,104,101,99,107,32,121,111,117,114,32,76,117,97,32,105,110,115,116,97,108,108,97,116,105,111,110,0,97,98,115,101,110,116,0,105,110,105,116,0,115,101,101,97,108,108,0,95,95,105,110,100,101,120,0,112,114,101,108,111,97,100,0,39,112,97,99,107,97,103,101,46,112,114,101,108,111,97,100,39,32,109,117,115,116,32,98,101,32,97,32,116,97,98,108,101,0,10,9,110,111,32,102,105,101,108,100,32,112,97,99,107,97,103,101,46,112,114,101,108,111,97,100,91,39,37,115,39,93,0,112,97,116,104,0,47,0,39,112,97,99,107,97,103,101,46,37,115,39,32,109,117,115,116,32,98,101,32,97,32,115,116,114,105,110,103,0,63,0,10,9,110,111,32,102,105,108,101,32,39,37,115,39,0,101,114,114,111,114,32,108,111,97,100,105,110,103,32,109,111,100,117,108,101,32,39,37,115,39,32,102,114,111,109,32,102,105,108,101,32,39,37,115,39,58,10,9,37,115,0,99,112,97,116,104,0,95,0,108,117,97,111,112,101,110,95,37,115,0,10,9,110,111,32,109,111,100,117,108,101,32,39,37,115,39,32,105,110,32,102,105,108,101,32,39,37,115,39,0,108,111,97,100,101,114,115,0,76,85,65,95,80,65,84,72,0,46,47,63,46,108,117,97,59,47,117,115,114,47,108,111,99,97,108,47,115,104,97,114,101,47,108,117,97,47,53,46,49,47,63,46,108,117,97,59,47,117,115,114,47,108,111,99,97,108,47,115,104,97,114,101,47,108,117,97,47,53,46,49,47,63,47,105,110,105,116,46,108,117,97,59,47,117,115,114,47,108,111,99,97,108,47,108,105,98,47,108,117,97,47,53,46,49,47,63,46,108,117,97,59,47,117,115,114,47,108,111,99,97,108,47,108,105,98,47,108,117,97,47,53,46,49,47,63,47,105,110,105,116,46,108,117,97,0,59,59,0,59,1,59,0,76,85,65,95,67,80,65,84,72,0,46,47,63,46,115,111,59,47,117,115,114,47,108,111,99,97,108,47,108,105,98,47,108,117,97,47,53,46,49,47,63,46,115,111,59,47,117,115,114,47,108,111,99,97,108,47,108,105,98,47,108,117,97,47,53,46,49,47,108,111,97,100,97,108,108,46,115,111,0,47,10,59,10,63,10,33,10,45,0,99,111,110,102,105,103,0,95,76,79,65,68,69,68,0,108,111,97,100,101,100,0,109,111,100,117,108,101,0,110,97,109,101,32,99,111,110,102,108,105,99,116,32,102,111,114,32,109,111,100,117,108,101,32,39,37,115,39,0,95,78,65,77,69,0,95,77,0,95,80,65,67,75,65,71,69,0,102,0,39,109,111,100,117,108,101,39,32,110,111,116,32,99,97,108,108,101,100,32,102,114,111,109,32,97,32,76,117,97,32,102,117,110,99,116,105,111,110,0,114,101,113,117,105,114,101,0,108,111,111,112,32,111,114,32,112,114,101,118,105,111,117,115,32,101,114,114,111,114,32,108,111,97,100,105,110,103,32,109,111,100,117,108,101,32,39,37,115,39,0,39,112,97,99,107,97,103,101,46,108,111,97,100,101,114,115,39,32,109,117,115,116,32,98,101,32,97,32,116,97,98,108,101,0,109,111,100,117,108,101,32,39,37,115,39,32,110,111,116,32,102,111,117,110,100,58,37,115,0,0,3,1,4,2,0,0,0,0,1,4,8,16,32,64,96,128,255,255,255,255,255,255,255,0,0,0,0,0,0,10,8,0,0,0,0,0,0,24,32,0,255,254,252,248,240,0,0,0,1,2,4,8,16,0,0,0,254,253,251,250,248,247,245,244,242,240,239,237,236,234,233,232,230,229,227,226,224,223,222,220,219,218,217,215,214,213,212,211,210,209,208,207,206,205,204,203,202,201,200,200,199,198,197,197,196,196,195,195,194,194,194,193,193,193,192,192,192,192,192,192,192,192,192,192,192,193,193,193,194,194,194,195,195,196,196,197,197,198,199,200,200,201,202,203,204,205,206,207,208,209,210,211,212,213,214,215,217,218,219,220,222,223,224,226,227,229,230,232,233,234,236,237,239,240,242,244,245,247,248,250,251,253,254,0,2,3,5,6,8,9,11,12,14,16,17,19,20,22,23,24,26,27,29,30,32,33,34,36,37,38,39,41,42,43,44,45,46,47,48,49,50,51,52,53,54,55,56,56,57,58,59,59,60,60,61,61,62,62,62,63,63,63,64,64,64,64,64,64,64,64,64,64,64,63,63,63,62,62,62,61,61,60,60,59,59,58,57,56,56,55,54,53,52,51,50,49,48,47,46,45,44,43,42,41,39,38,37,36,34,33,32,30,29,27,26,24,23,22,20,19,17,16,14,12,11,9,8,6,5,3,2,70,70,0,70,67,0,57,99,32,110,32,118,0,57,99,32,110,32,48,0,67,99,32,112,0,70,48,70,48,48,49,37,48,50,88,0,77,77,65,66,67,95,78,79,95,82,65,78,68,79,77,95,80,73,67,75,0,45,37,108,100,0,47,117,115,114,47,108,111,99,97,108,47,115,104,97,114,101,47,116,105,109,105,100,105,116,121,47,105,110,115,116,114,117,109,101,110,116,115,0,47,117,115,114,47,108,111,99,97,108,47,115,104,97,114,101,47,116,105,109,105,100,105,116,121,47,116,105,109,105,100,105,116,121,46,99,102,103,0,77,77,80,65,84,95,80,65,84,72,95,84,79,95,67,70,71,0,47,116,105,109,105,100,105,116,121,46,99,102,103,0,47,105,110,115,116,114,117,109,101,110,116,115,0,99,97,110,32,110,111,116,32,111,112,101,110,32,37,115,44,32,117,115,101,32,101,110,118,105,114,111,110,109,101,110,116,32,118,97,114,105,97,98,108,101,32,77,77,80,65,84,95,80,65,84,72,95,84,79,95,67,70,71,32,102,111,114,32,116,104,101,32,100,105,114,101,99,116,111,114,121,0,108,111,97,100,95,112,97,116,32,62,32,37,115,10,0,100,114,117,109,115,101,116,0,115,111,117,114,99,101,0,118,32,61,32,43,100,111,119,110,98,111,119,43,0,126,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,0,43,110,105,108,43,0,117,32,61,32,43,117,112,98,111,119,43,0,79,32,61,32,43,99,111,100,97,43,0,83,32,61,32,43,115,101,103,110,111,43,0,80,32,61,32,43,117,112,112,101,114,109,111,114,100,101,110,116,43,0,77,32,61,32,43,108,111,119,101,114,109,111,114,100,101,110,116,43,0,76,32,61,32,43,101,109,112,104,97,115,105,115,43,0,72,32,61,32,43,102,101,114,109,97,116,97,43,0,84,32,61,32,43,116,114,105,108,108,43,0,126,32,61,32,43,114,111,108,108,43,0,70,97,105,108,117,114,101,58,32,67,104,111,114,100,32,110,97,109,101,32,99,97,110,110,111,116,32,101,120,99,101,101,100,32,55,32,99,104,97,114,97,99,116,101,114,115,44,32,37,115,0,108,111,97,100,95,97,98,99,32,62,32,37,115,10,0,70,97,105,108,117,114,101,58,32,78,97,109,101,100,32,99,104,111,114,100,32,99,97,110,110,111,116,32,104,97,118,101,32,109,111,114,101,32,116,104,97,110,32,54,32,110,111,116,101,115,44,32,37,115,0,70,97,105,108,117,114,101,58,32,84,111,111,32,109,97,110,121,32,71,117,105,116,97,114,32,67,104,111,114,100,32,78,97,109,101,115,32,117,115,101,100,44,32,37,115,0,109,0,55,0,109,55,0,109,97,106,55,0,77,55,0,54,0,109,54,0,97,117,103,0,43,0,97,117,103,55,0,55,43,0,100,105,109,0,100,105,109,55,0,57,0,109,57,0,109,97,106,57,0,77,57,0,49,49,0,100,105,109,57,0,115,117,115,0,115,117,115,57,0,55,115,117,115,0,55,115,117,115,52,0,55,115,117,115,57,0,57,115,117,115,52,0,53,0,49,51,0,67,68,69,70,71,65,66,99,100,101,102,103,97,98,0,67,68,69,70,71,65,66,67,68,69,70,71,65,66,99,100,101,102,103,97,98,99,100,101,102,103,97,98,0,88,58,0,116,61,0,111,99,116,97,118,101,61,0,116,114,97,110,115,112,111,115,101,61,0,109,61,0,109,105,100,100,108,101,61,0,99,108,101,102,61,0,116,114,101,98,108,101,0,100,0,71,0,69,0,98,97,115,115,0,68,44,0,65,44,0,70,44,0,66,44,44,0,116,101,110,111,114,0,97,108,116,111,0,43,45,0,118,97,0,32,67,32,68,32,69,70,32,71,32,65,32,66,99,32,100,32,101,102,32,103,32,97,32,98,0,76,58,0,32,37,100,32,47,32,37,100,0,77,58,0,110,111,110,101,0,67,124,0,32,47,32,37,100,0,80,58,0,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,40,41,46,48,49,50,51,52,53,54,55,56,57,32,0,105,110,118,97,108,105,100,32,99,104,97,114,97,99,116,101,114,115,32,105,110,32,112,97,114,116,32,115,116,114,105,110,103,32,115,99,97,110,110,105,110,103,32,80,58,37,115,0,87,97,114,110,105,110,103,58,32,85,110,98,97,108,97,110,99,101,100,32,114,105,103,104,116,32,112,97,114,101,110,115,32,105,110,32,80,58,32,100,101,102,105,110,105,116,105,111,110,32,37,115,0,87,97,114,110,105,110,103,58,32,85,110,98,97,108,97,110,99,101,100,32,108,101,102,116,32,112,97,114,101,110,115,32,105,110,32,80,58,32,100,101,102,105,110,105,116,105,111,110,32,37,115,0,81,58,0,70,97,105,108,117,114,101,58,32,105,110,99,111,110,118,101,110,105,101,110,116,32,116,101,109,112,111,32,99,104,97,110,103,101,32,105,110,32,109,105,100,100,108,101,32,111,102,32,118,111,105,99,101,32,40,37,115,41,0,84,58,0,82,58,0,104,111,114,110,112,105,112,101,0,86,58,0,75,58,0,109,105,110,0,109,97,106,0,72,112,0,72,80,0,66,109,0,109,105,110,111,114,0,109,97,106,111,114,0,55,32,115,104,97,114,112,115,58,32,67,35,32,32,32,32,65,35,109,32,32,32,71,35,77,105,120,32,68,35,68,111,114,32,69,35,80,104,114,32,70,35,76,121,100,32,66,35,76,111,99,32,0,54,32,115,104,97,114,112,115,58,32,70,35,32,32,32,32,68,35,109,32,32,32,67,35,77,105,120,32,71,35,68,111,114,32,65,35,80,104,114,32,66,76,121,100,32,32,69,35,76,111,99,32,0,53,32,115,104,97,114,112,115,58,32,66,32,32,32,32,32,71,35,109,32,32,32,70,35,77,105,120,32,67,35,68,111,114,32,68,35,80,104,114,32,69,76,121,100,32,32,65,35,76,111,99,32,0,52,32,115,104,97,114,112,115,58,32,69,32,32,32,32,32,67,35,109,32,32,32,66,77,105,120,32,32,70,35,68,111,114,32,71,35,80,104,114,32,65,76,121,100,32,32,68,35,76,111,99,32,0,51,32,115,104,97,114,112,115,58,32,65,32,32,32,32,32,70,35,109,32,32,32,69,77,105,120,32,32,66,68,111,114,32,32,67,35,80,104,114,32,68,76,121,100,32,32,71,35,76,111,99,32,0,50,32,115,104,97,114,112,115,58,32,68,32,32,32,32,32,66,109,32,32,32,32,65,77,105,120,32,32,69,68,111,114,32,32,70,35,80,104,114,32,71,76,121,100,32,32,67,35,76,111,99,32,0,49,32,115,104,97,114,112,32,58,32,71,32,32,32,32,32,69,109,32,32,32,32,68,77,105,120,32,32,65,68,111,114,32,32,66,80,104,114,32,32,67,76,121,100,32,32,70,35,76,111,99,32,0,48,32,115,104,97,114,112,115,58,32,67,32,32,32,32,32,65,109,32,32,32,32,71,77,105,120,32,32,68,68,111,114,32,32,69,80,104,114,32,32,70,76,121,100,32,32,66,76,111,99,32,32,0,49,32,102,108,97,116,32,32,58,32,70,32,32,32,32,32,68,109,32,32,32,32,67,77,105,120,32,32,71,68,111,114,32,32,65,80,104,114,32,32,66,98,76,121,100,32,69,76,111,99,32,32,0,50,32,102,108,97,116,115,32,58,32,66,98,32,32,32,32,71,109,32,32,32,32,70,77,105,120,32,32,67,68,111,114,32,32,68,80,104,114,32,32,69,98,76,121,100,32,65,76,111,99,32,32,0,51,32,102,108,97,116,115,32,58,32,69,98,32,32,32,32,67,109,32,32,32,32,66,98,77,105,120,32,70,68,111,114,32,32,71,80,104,114,32,32,65,98,76,121,100,32,68,76,111,99,32,32,0,52,32,102,108,97,116,115,32,58,32,65,98,32,32,32,32,70,109,32,32,32,32,69,98,77,105,120,32,66,98,68,111,114,32,67,80,104,114,32,32,68,98,76,121,100,32,71,76,111,99,32,32,0,53,32,102,108,97,116,115,32,58,32,68,98,32,32,32,32,66,98,109,32,32,32,65,98,77,105,120,32,69,98,68,111,114,32,70,80,104,114,32,32,71,98,76,121,100,32,67,76,111,99,32,32,0,54,32,102,108,97,116,115,32,58,32,71,98,32,32,32,32,69,98,109,32,32,32,68,98,77,105,120,32,65,98,68,111,114,32,66,98,80,104,114,32,67,98,76,121,100,32,70,76,111,99,32,32,0,55,32,102,108,97,116,115,32,58,32,67,98,32,32,32,32,65,98,109,32,32,32,71,98,77,105,120,32,68,98,68,111,114,32,69,98,80,104,114,32,70,98,76,121,100,32,66,98,76,111,99,32,0,70,97,105,108,117,114,101,58,32,85,110,114,101,99,111,103,110,105,115,101,100,32,75,58,32,102,105,101,108,100,32,37,115,0,32,67,32,68,32,69,70,32,71,32,65,66,32,99,32,100,32,101,102,32,103,32,97,98,32,0,32,67,32,68,69,32,70,32,71,32,65,66,32,99,32,100,101,32,102,32,103,32,97,98,32,0,32,67,32,68,69,32,70,32,71,65,32,66,32,99,32,100,101,32,102,32,103,97,32,98,32,0,32,67,68,32,69,32,70,32,71,65,32,66,32,99,100,32,101,32,102,32,103,97,32,98,32,0,32,67,68,32,69,32,70,71,32,65,32,66,32,99,100,32,101,32,102,103,32,97,32,98,32,0,32,67,32,68,32,69,32,70,71,32,65,32,66,99,32,100,32,101,32,102,103,32,97,32,98,0,67,32,68,32,69,32,70,71,32,65,32,66,99,32,100,32,101,32,102,103,32,97,32,98,32,0,67,32,68,32,69,70,32,71,32,65,32,66,99,32,100,32,101,102,32,103,32,97,32,98,32,0,37,115,37,115,0,102,122,99,122,102,122,99,122,0,111,110,0,111,102,102,0,102,98,99,122,48,49,50,51,52,53,54,55,56,57,103,104,105,106,71,72,73,74,0,102,122,99,122,99,122,102,122,99,122,99,122,0,102,122,99,122,99,122,0,102,122,99,102,122,99,0,102,122,99,102,122,99,102,122,99,0,102,122,99,102,122,99,102,122,99,102,122,99,0,102,122,99,102,122,99,102,122,99,102,122,99,102,122,99,102,122,99,102,122,99,102,122,99,102,122,99,0,102,122,99,122,102,122,99,122,102,122,99,122,102,122,99,122,102,122,99,122,102,122,99,122,102,122,99,122,0,112,97,114,116,115,32,40,37,115,41,32,115,101,116,32,98,117,116,32,110,111,116,32,117,115,101,100,0,109,58,0,108,105,110,101,32,110,111,116,32,112,114,111,112,101,114,108,121,32,99,111,110,116,105,110,117,101,100,10,37,115,0,37,37,0,77,73,68,73,0,98,97,115,115,112,114,111,103,0,98,97,115,115,118,111,108,0,98,101,97,116,0,98,101,97,116,115,116,114,105,110,103,0,99,104,111,114,100,110,97,109,101,0,70,97,105,108,117,114,101,58,32,66,97,100,32,102,111,114,109,97,116,32,102,111,114,32,99,104,111,114,100,110,97,109,101,32,99,111,109,109,97,110,100,44,32,37,115,0,99,104,111,114,100,112,114,111,103,0,99,104,111,114,100,118,111,108,0,100,114,111,110,101,0,100,114,111,110,101,111,102,102,0,100,114,111,110,101,111,110,0,100,114,117,109,0,100,122,48,49,50,51,52,53,54,55,56,57,0,100,114,117,109,111,102,102,0,100,114,117,109,111,110,0,103,99,104,111,114,100,0,103,99,104,111,114,100,111,102,102,0,103,99,104,111,114,100,111,110,0,99,104,97,110,110,101,108,0,112,114,111,103,114,97,109,0,118,111,105,99,101,0,105,110,115,116,114,117,109,101,110,116,61,0,109,117,116,101,0,98,97,110,107,61,0,108,101,103,97,116,111,0,109,97,99,114,111,32,108,105,110,101,32,116,111,111,32,108,111,110,103,10,37,115,0,85,58,0,119,58,0,87,58,0,100,58,0,119,97,114,110,105,110,103,58,32,111,108,100,32,115,116,121,108,101,32,100,101,99,111,114,97,116,105,111,110,115,32,110,111,116,32,104,97,110,100,108,101,100,10,37,115,0,115,58,0,119,97,114,110,105,110,103,58,32,110,101,119,32,115,116,121,108,101,32,100,101,99,111,114,97,116,105,111,110,115,32,110,111,116,32,104,97,110,100,108,101,100,10,37,115,0,73,58,0,97,98,99,45,105,110,99,108,117,100,101,0,102,97,105,108,117,114,101,58,32,116,111,111,32,109,97,110,121,32,97,98,99,45,105,110,99,108,117,100,101,39,115,44,32,37,115,0,114,0,102,97,105,108,117,114,101,58,32,97,98,99,45,105,110,99,108,117,100,101,32,102,105,108,101,32,37,115,32,110,111,116,32,102,111,117,110,100,0,102,97,105,108,117,114,101,58,32,97,98,99,45,105,110,99,108,117,100,101,32,109,105,115,115,105,110,103,32,102,105,108,101,32,110,97,109,101,44,32,37,115,0,97,98,99,100,101,102,103,65,66,67,68,69,70,71,94,95,61,0,44,39,0,115,116,114,97,110,103,101,32,110,111,116,101,32,101,110,99,111,117,110,116,101,114,101,100,32,115,99,97,110,110,105,110,103,32,37,115,0,115,104,111,114,116,32,110,111,116,101,115,32,105,110,32,99,104,111,114,100,32,99,97,110,32,110,111,116,32,98,101,32,116,105,101,100,58,10,37,115,0,110,111,116,101,0,95,94,60,62,64,0,70,97,105,108,117,114,101,58,32,117,110,114,101,99,111,103,110,105,122,101,100,32,99,104,111,114,100,110,97,109,101,32,37,115,0,102,97,100,101,0,116,111,32,99,111,100,97,0,116,111,0,99,111,100,97,0,97,114,112,101,103,103,105,111,0,99,114,101,115,99,101,110,100,111,40,0,99,114,101,115,99,101,110,100,111,41,0,60,40,0,60,41,0,100,105,109,105,109,117,101,110,100,111,40,0,100,105,109,105,110,117,101,110,100,111,41,0,62,40,0,62,41,0,117,112,98,111,119,0,100,111,119,110,98,111,119,0,116,114,105,108,108,0,116,114,105,108,108,40,0,116,114,105,108,108,41,0,97,99,99,101,110,116,0,101,109,112,104,97,115,105,115,0,102,101,114,109,97,116,97,0,102,105,110,101,0,115,101,103,110,111,0,116,111,99,111,100,97,0,68,46,67,46,0,68,46,83,46,0,100,97,99,97,112,111,0,100,97,99,111,100,97,0,124,91,58,93,33,0,109,112,0,109,102,0,115,102,122,0,97,98,99,100,101,102,103,65,66,67,68,69,70,71,94,95,61,88,0,122,120,0,90,0,97,98,99,32,88,58,37,115,32,104,97,115,32,110,111,32,98,111,100,121,0,77,77,65,66,67,95,68,85,77,80,84,82,65,67,75,83,0,80,58,37,115,10,0,32,32,37,99,32,0,65,108,108,0,37,51,100,32,0,116,114,97,99,107,32,37,100,46,37,100,32,99,104,97,110,61,37,100,32,37,115,10,0,110,111,110,111,116,101,115,0,32,32,37,54,100,46,37,48,50,100,46,37,100,37,99,37,99,32,37,100,46,37,100,32,37,115,32,0,67,67,67,68,68,69,70,70,71,71,65,65,66,99,99,100,100,101,102,102,103,103,97,97,98,98,0,98,32,35,32,35,32,32,35,32,35,32,35,32,32,35,32,35,32,32,35,32,35,32,35,32,35,0,67,77,68,32,37,99,58,32,103,99,104,111,114,100,32,37,115,37,115,0,47,37,115,0,67,77,68,32,37,99,32,64,37,112,32,48,120,37,48,56,108,88,10,0,45,45,0,105,110,118,97,108,105,100,32,103,109,32,37,100,0,78,79,84,69,32,37,115,32,111,99,116,97,118,101,32,37,100,32,105,110,115,116,32,37,115,32,118,111,108,32,37,48,51,100,10,0,97,108,108,0,37,115,37,99,37,115,0,46,112,97,116,0,37,115,37,99,37,115,46,112,97,116,0,63,63,63,63,63,63,0,37,100,58,37,115,0,97,109,112,0,84,105,109,105,100,105,116,121,32,71,77,32,112,97,116,99,104,101,115,0,77,84,104,100,0,77,77,77,73,68,95,68,69,66,85,71,0,77,77,77,73,68,95,86,69,82,66,79,83,69,0,77,77,77,73,68,95,83,80,69,69,68,0,83,99,97,110,110,105,110,103,32,77,73,68,73,32,119,105,116,104,32,102,111,114,109,97,116,58,32,37,100,32,114,101,115,111,108,117,116,105,111,110,58,32,37,100,32,116,114,97,99,107,115,58,32,37,100,10,0,77,117,108,116,105,112,108,121,105,110,103,32,114,101,115,111,108,117,116,105,111,110,32,97,110,100,32,100,101,108,116,97,116,105,109,101,115,32,98,121,32,37,100,32,116,111,32,103,101,116,32,37,100,32,109,105,100,105,116,105,99,107,115,32,112,101,114,32,112,97,116,116,101,114,110,114,111,119,10,0,80,97,114,115,105,110,103,32,116,114,97,99,107,32,37,100,10,0,77,84,114,107,0,105,110,118,97,108,105,100,32,116,114,97,99,107,45,99,104,117,110,107,32,39,37,115,39,32,105,115,32,110,111,116,32,39,77,84,114,107,39,0,108,111,97,100,95,109,105,100,32,62,32,37,115,10,0,37,50,100,32,37,48,56,108,100,32,32,32,32,32,32,32,78,111,116,101,32,111,102,102,58,32,99,104,32,37,100,32,48,120,37,48,50,120,32,48,120,37,48,50,120,10,0,37,50,100,32,37,48,56,108,100,32,78,111,116,101,32,32,111,110,58,32,99,104,32,37,100,32,48,120,37,48,50,120,32,48,120,37,48,50,120,10,0,37,50,100,32,37,48,56,108,100,32,110,111,116,101,32,111,102,102,58,32,99,104,32,37,100,32,48,120,37,48,50,120,10,0,37,50,100,32,37,48,56,108,100,32,112,111,108,121,112,104,111,110,105,99,32,107,101,121,32,112,114,101,115,115,117,114,101,58,32,99,104,32,37,100,32,48,120,37,48,50,120,32,48,120,37,48,50,120,10,0,37,108,100,32,37,100,32,97,108,108,32,110,111,116,101,115,32,111,102,102,10,0,37,50,100,32,37,48,56,108,100,32,99,111,110,116,114,111,108,32,99,104,97,110,103,101,58,32,99,104,32,37,100,32,48,120,37,48,50,120,32,48,120,37,48,50,120,10,0,37,50,100,32,37,48,56,108,100,32,112,114,111,103,114,97,109,32,99,104,97,110,103,101,58,32,99,104,32,37,100,32,37,100,10,0,37,50,100,32,37,48,56,108,100,32,99,104,97,110,110,101,108,32,112,114,101,115,115,117,114,101,58,32,99,104,32,37,100,32,48,120,37,48,50,120,10,0,37,50,100,32,37,48,56,108,100,32,112,105,116,99,104,32,119,104,101,101,108,32,99,104,97,110,103,101,58,32,99,104,32,37,100,32,37,100,10,0,37,50,100,32,37,48,56,108,100,32,115,121,115,101,120,58,32,48,120,37,48,50,120,0,32,37,48,50,88,0,37,50,100,32,37,48,56,108,100,32,115,111,110,103,32,112,111,115,105,116,105,111,110,32,112,111,105,110,116,101,114,58,32,37,100,0,37,50,100,32,37,48,56,108,100,32,115,121,115,101,120,32,99,111,110,116,105,110,117,101,100,58,32,37,108,100,0,37,50,100,32,37,48,56,108,100,32,77,69,84,65,32,116,114,97,99,107,110,97,109,101,58,37,115,10,0,37,50,100,32,37,48,56,108,100,32,77,69,84,65,32,116,101,109,112,111,58,37,100,10,0,37,50,100,32,37,48,56,108,100,32,77,69,84,65,32,101,110,100,32,111,102,32,116,114,97,99,107,10,0,37,108,100,0,77,101,116,97,32,101,118,101,110,116,32,110,111,116,32,97,116,32,101,110,100,32,111,102,32,116,114,97,99,107,44,32,37,115,32,98,121,116,101,115,32,108,101,102,116,32,105,110,32,116,114,97,99,107,0,37,50,100,32,37,48,56,108,100,32,77,69,84,65,32,116,121,112,101,32,48,120,37,48,50,120,10,0,37,50,100,32,37,48,56,108,100,32,83,121,115,116,101,109,32,116,121,112,101,32,48,120,37,48,50,120,10,0,37,50,100,32,37,48,56,108,100,32,117,110,107,110,111,119,110,32,114,117,110,110,105,110,103,115,116,97,116,117,115,58,32,48,120,37,48,50,120,32,115,107,105,112,112,101,100,58,48,120,37,48,50,120,10,0,115,117,112,101,114,102,108,117,111,117,115,0,110,111,0,68,101,116,101,114,109,105,110,105,110,103,32,112,101,114,99,117,115,115,105,111,110,32,99,104,97,110,110,101,108,0,80,101,114,99,117,115,115,105,111,110,32,99,104,97,110,110,101,108,32,105,115,32,37,100,10,83,116,114,105,112,112,105,110,103,32,111,102,102,32,115,105,108,101,110,99,101,115,32,97,110,100,32,111,116,104,101,114,32,111,112,116,105,109,97,108,105,115,97,116,105,111,110,115,10,0,37,108,100,32,62,32,37,108,100,32,110,111,116,101,32,37,100,0,100,114,117,109,32,116,114,97,99,107,32,101,110,100,115,32,119,105,116,104,32,110,111,116,101,32,111,110,32,40,37,115,41,0,99,104,97,110,110,101,108,32,37,100,44,32,37,108,100,32,62,32,37,108,100,32,110,111,116,101,32,37,100,0,109,101,108,111,100,121,32,116,114,97,99,107,32,101,110,100,115,32,119,105,116,104,32,110,111,116,101,32,111,110,32,40,37,115,41,0,116,114,97,99,107,116,105,109,101,32,32,61,32,37,108,100,10,0,115,112,101,101,100,32,32,32,32,32,32,61,32,37,100,10,0,109,105,100,105,115,112,101,101,100,32,32,61,32,37,100,10,0,109,105,100,105,102,111,114,109,97,116,32,61,32,37,100,10,0,114,101,115,111,108,117,116,105,111,110,32,61,32,37,100,10,0,109,105,100,105,116,114,97,99,107,115,32,61,32,37,100,10,0,100,105,118,105,100,101,114,32,32,32,32,61,32,37,100,10,0,116,101,109,112,111,32,32,32,32,32,32,61,32,37,100,10,0,112,101,114,99,117,115,115,105,111,110,32,61,32,37,100,10,0,100,101,108,116,97,116,105,109,101,32,32,61,32,37,108,100,10,0,84,82,65,67,75,32,37,50,100,32,99,104,97,110,61,37,100,32,110,111,116,101,61,48,120,37,48,50,120,32,118,111,108,61,37,100,32,112,97,110,61,48,120,37,48,50,120,32,105,110,115,116,114,61,37,100,10,0,78,79,84,69,0,67,84,82,76,0,37,50,100,32,37,54,108,100,32,37,115,32,37,51,100,32,37,51,100,32,37,51,100,32,0,102,120,98,114,107,0,102,120,115,121,110,99,0,112,114,111,103,32,37,100,10,0,109,97,105,110,118,111,108,32,37,100,10,0,109,111,100,119,104,101,101,108,32,37,100,10,0,119,104,101,101,108,100,111,119,110,32,37,100,10,0,119,104,101,101,108,117,112,32,37,100,10,0,116,109,112,111,32,37,100,10,0,65,100,106,117,115,116,105,110,103,32,116,101,109,112,111,32,37,100,32,116,111,32,50,53,53,10,0,65,100,106,117,115,116,101,100,32,109,97,120,105,109,117,109,32,116,101,109,112,111,32,102,114,111,109,32,37,100,32,116,111,32,37,100,32,116,111,32,103,101,116,32,37,100,32,109,105,100,105,116,105,99,107,115,32,112,101,114,32,112,97,116,116,101,114,110,114,111,119,10,0,71,101,110,101,114,97,116,105,110,103,32,37,100,32,112,97,116,116,101,114,110,115,32,119,105,116,104,32,115,112,101,101,100,32,37,100,10,0,68,111,110,101,0,71,70,49,80,65,84,67,72,49,49,48,0,73,68,35,48,48,48,48,48,50,0,37,115,32,99,97,110,111,110,32,37,100,45,118,32,40,70,114,46,32,74,97,99,113,117,101,115,41,0,37,100,45,118,111,105,99,101,32,99,97,110,111,110,32,40,70,114,46,32,74,97,99,113,117,101,115,41,0,37,100,58,85,110,116,105,116,108,101,100,32,71,77,32,112,97,116,99,104,0,84,105,109,105,100,105,116,121,32,71,77,32,112,97,116,99,104,0,99,32,100,32,101,32,99,124,99,32,100,32,101,32,99,124,101,32,102,32,103,46,46,124,101,32,102,32,103,46,46,124,103,97,103,102,101,32,99,124,103,97,103,102,101,32,99,124,99,32,71,32,99,46,46,124,99,32,71,32,99,46,46,124,0,33,83,67,82,69,65,77,33,0,66,77,79,68,50,83,84,77,0,179,164,152,141,131,123,116,110,104,99,77,84,77,0,69,120,116,114,101,109,101,0,77,65,83,95,85,84,114,97,99,107,95,86,48,48,0,69,120,116,101,110,100,101,100,32,77,111,100,117,108,101,0,65,68,80,67,77,0,65,83,89,76,85,77,32,77,117,115,105,99,32,70,111,114,109,97,116,32,86,49,46,48,0,116,104,101,111,114,97,0,3,4,2,0,1,5,6,7,3,4,0,2,1,5,6,7,3,2,4,0,1,5,6,7,3,2,0,4,1,5,6,7,0,3,4,2,1,5,6,7,0,5,3,4,2,1,6,7,0,1,2,3,4,5,6,7,12,4,3,3,4,4,5,5,8,8,8,8,3,3,6,1,6,15,28,64,24,32,1,0,1,1,1,4,8,1,1,8,1,1,1,1,1,2,2,2,2,4,8,2,2,2,4,2,2,2,2,2,8,2,4,8,15,16,17,88,80,1,0,48,14,56,57,58,59,60,62,64,66,68,72,2,4,6,8,18,20,22,24,26,32,12,28,40,0,1,8,16,9,2,3,10,17,24,32,25,18,11,4,5,12,19,26,33,40,48,41,34,27,20,13,6,7,14,21,28,35,42,49,56,57,50,43,36,29,22,15,23,30,37,44,51,58,59,52,45,38,31,39,46,53,60,61,54,47,55,62,63,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,0,3,1,2,0,1,2,3,4,8,0,0,0,0,0,0,0,1,2,3,4,5,8,9,0,0,0,0,0,1,2,3,4,6,8,10,0,0,0,0,0,1,2,3,4,5,6,7,8,9,10,11,6,8,8,12,241,241,242,242,243,243,244,244,245,245,246,246,247,247,248,248,249,249,250,250,251,251,252,252,253,253,254,254,255,255,0,0,0,1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,11,12,12,13,13,14,14,15,15,0,249,249,249,249,250,250,250,250,251,251,251,251,252,252,252,252,253,253,253,253,254,254,254,254,255,255,255,255,0,0,0,0,0,0,0,1,1,1,1,2,2,2,2,3,3,3,3,4,4,4,4,5,5,5,5,6,6,6,6,7,7,7,7,0,255,0,255,0,255,0,255,0,255,0,255,0,255,0,255,0,255,0,255,0,255,0,255,0,255,0,255,0,255,0,255,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,255,255,255,0,255,255,255,0,255,255,255,0,255,255,255,0,255,255,255,0,255,255,255,0,255,255,255,0,255,255,255,0,1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,0,99,105,100,0,79,84,84,79,0,46,95,0,47,46,46,110,97,109,101,100,102,111,114,107,47,114,115,114,99,0,47,114,115,114,99,0,114,101,115,111,117,114,99,101,46,102,114,107,47,0,46,114,101,115,111,117,114,99,101,47,0,37,0,46,65,112,112,108,101,68,111,117,98,108,101,47,0,114,98,0,66,68,70,0,67,79,77,77,69,78,84,0,83,84,65,82,84,70,79,78,84,0,65,86,71,95,67,65,80,73,84,65,76,95,87,73,68,84,72,0,65,86,71,95,76,79,87,69,82,67,65,83,69,95,87,73,68,84,72,0,67,65,80,95,72,69,73,71,72,84,0,67,72,65,82,83,69,84,95,67,79,76,76,69,67,84,73,79,78,83,0,67,79,80,89,82,73,71,72,84,0,68,69,70,65,85,76,84,95,67,72,65,82,0,68,69,83,84,73,78,65,84,73,79,78,0,68,69,86,73,67,69,95,70,79,78,84,95,78,65,77,69,0,69,78,68,95,83,80,65,67,69,0,70,65,67,69,95,78,65,77,69,0,70,73,71,85,82,69,95,87,73,68,84,72,0,70,79,78,84,0,70,79,78,84,78,65,77,69,95,82,69,71,73,83,84,82,89,0,70,79,78,84,95,65,83,67,69,78,84,0,70,79,78,84,95,68,69,83,67,69,78,84,0,70,79,85,78,68,82,89,0,70,85,76,76,95,78,65,77,69,0,73,84,65,76,73,67,95,65,78,71,76,69,0,77,65,88,95,83,80,65,67,69,0,77,73,78,95,83,80,65,67,69,0,78,79,82,77,95,83,80,65,67,69,0,78,79,84,73,67,69,0,81,85,65,68,95,87,73,68,84,72,0,82,65,87,95,65,83,67,69,78,84,0,82,65,87,95,65,86,69,82,65,71,69,95,87,73,68,84,72,0,82,65,87,95,65,86,71,95,67,65,80,73,84,65,76,95,87,73,68,84,72,0,82,65,87,95,65,86,71,95,76,79,87,69,82,67,65,83,69,95,87,73,68,84,72,0,82,65,87,95,67,65,80,95,72,69,73,71,72,84,0,82,65,87,95,68,69,83,67,69,78,84,0,82,65,87,95,69,78,68,95,83,80,65,67,69,0,82,65,87,95,70,73,71,85,82,69,95,87,73,68,84,72,0,82,65,87,95,77,65,88,95,83,80,65,67,69,0,82,65,87,95,77,73,78,95,83,80,65,67,69,0,82,65,87,95,78,79,82,77,95,83,80,65,67,69,0,82,65,87,95,80,73,88,69,76,95,83,73,90,69,0,82,65,87,95,80,79,73,78,84,95,83,73,90,69,0,82,65,87,95,80,73,88,69,76,83,73,90,69,0,82,65,87,95,80,79,73,78,84,83,73,90,69,0,82,65,87,95,81,85,65,68,95,87,73,68,84,72,0,82,65,87,95,83,77,65,76,76,95,67,65,80,95,83,73,90,69,0,82,65,87,95,83,84,82,73,75,69,79,85,84,95,65,83,67,69,78,84,0,82,65,87,95,83,84,82,73,75,69,79,85,84,95,68,69,83,67,69,78,84,0,82,65,87,95,83,85,66,83,67,82,73,80,84,95,83,73,90,69,0,82,65,87,95,83,85,66,83,67,82,73,80,84,95,88,0,82,65,87,95,83,85,66,83,67,82,73,80,84,95,89,0,82,65,87,95,83,85,80,69,82,83,67,82,73,80,84,95,83,73,90,69,0,82,65,87,95,83,85,80,69,82,83,67,82,73,80,84,95,88,0,82,65,87,95,83,85,80,69,82,83,67,82,73,80,84,95,89,0,82,65,87,95,85,78,68,69,82,76,73,78,69,95,80,79,83,73,84,73,79,78,0,82,65,87,95,85,78,68,69,82,76,73,78,69,95,84,72,73,67,75,78,69,83,83,0,82,65,87,95,88,95,72,69,73,71,72,84,0,82,69,76,65,84,73,86,69,95,83,69,84,87,73,68,84,72,0,82,69,76,65,84,73,86,69,95,87,69,73,71,72,84,0,82,69,83,79,76,85,84,73,79,78,0,83,77,65,76,76,95,67,65,80,95,83,73,90,69,0,83,80,65,67,73,78,71,0,83,84,82,73,75,69,79,85,84,95,65,83,67,69,78,84,0,83,84,82,73,75,69,79,85,84,95,68,69,83,67,69,78,84,0,83,85,66,83,67,82,73,80,84,95,83,73,90,69,0,83,85,66,83,67,82,73,80,84,95,88,0,83,85,66,83,67,82,73,80,84,95,89,0,83,85,80,69,82,83,67,82,73,80,84,95,83,73,90,69,0,83,85,80,69,82,83,67,82,73,80,84,95,88,0,83,85,80,69,82,83,67,82,73,80,84,95,89,0,85,78,68,69,82,76,73,78,69,95,80,79,83,73,84,73,79,78,0,85,78,68,69,82,76,73,78,69,95,84,72,73,67,75,78,69,83,83,0,87,69,73,71,72,84,0,88,95,72,69,73,71,72,84,0,95,77,85,76,69,95,66,65,83,69,76,73,78,69,95,79,70,70,83,69,84,0,95,77,85,76,69,95,82,69,76,65,84,73,86,69,95,67,79,77,80,79,83,69,0,83,84,65,82,84,80,82,79,80,69,82,84,73,69,83,0,32,43,0,0,0,0,0,0,0,255,3,126,0,0,0,126,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,255,3,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,2,3,4,5,6,7,8,9,0,0,0,0,0,0,0,10,11,12,13,14,15,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,10,11,12,13,14,15,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,69,78,68,80,82,79,80,69,82,84,73,69,83,0,37,104,100,0,67,72,65,82,83,0,69,78,68,70,79,78,84,0,69,78,68,67,72,65,82,0,83,84,65,82,84,67,72,65,82,0,69,78,67,79,68,73,78,71,0,255,128,192,224,240,248,252,254,83,87,73,68,84,72,0,68,87,73,68,84,72,0,66,66,88,0,66,73,84,77,65,80,0,95,88,70,82,69,69,56,54,95,71,76,89,80,72,95,82,65,78,71,69,83,0,70,79,78,84,66,79,85,78,68,73,78,71,66,79,88,0,45,0,83,73,90,69,0,99,102,102,0,67,70,70,0,100,97,114,107,101,110,105,110,103,45,112,97,114,97,109,101,116,101,114,115,0,104,105,110,116,105,110,103,45,101,110,103,105,110,101,0,110,111,45,115,116,101,109,45,100,97,114,107,101,110,105,110,103,0,1,0,1,1,1,0,1,0,1,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,0,1,1,1,0,1,1,1,1,1,1,1,1,1,1,0,0,116,49,99,105,100,0,67,73,68,32,84,121,112,101,32,49,0,67,73,68,0,37,33,80,83,45,65,100,111,98,101,45,51,46,48,32,82,101,115,111,117,114,99,101,45,67,73,68,70,111,110,116,0,83,116,97,114,116,68,97,116,97,0,47,115,102,110,116,115,0,40,72,101,120,41,0,37,65,68,79,66,101,103,105,110,70,111,110,116,68,105,99,116,0,67,73,68,70,111,110,116,78,97,109,101,0,67,73,68,70,111,110,116,86,101,114,115,105,111,110,0,67,73,68,70,111,110,116,84,121,112,101,0,82,101,103,105,115,116,114,121,0,79,114,100,101,114,105,110,103,0,83,117,112,112,108,101,109,101,110,116,0,85,73,68,66,97,115,101,0,67,73,68,77,97,112,79,102,102,115,101,116,0,70,68,66,121,116,101,115,0,71,68,66,121,116,101,115,0,67,73,68,67,111,117,110,116,0,83,117,98,114,77,97,112,79,102,102,115,101,116,0,83,68,66,121,116,101,115,0,83,117,98,114,67,111,117,110,116,0,108,101,110,66,117,105,108,100,67,104,97,114,65,114,114,97,121,0,70,111,114,99,101,66,111,108,100,84,104,114,101,115,104,111,108,100,0,70,68,65,114,114,97,121,0,112,99,102,0,80,67,70,0,83,76,65,78,84,0,79,98,108,105,113,117,101,0,87,69,73,71,72,84,95,78,65,77,69,0,83,69,84,87,73,68,84,72,95,78,65,77,69,0,65,68,68,95,83,84,89,76,69,95,78,65,77,69,0,70,65,77,73,76,89,95,78,65,77,69,0,65,86,69,82,65,71,69,95,87,73,68,84,72,0,80,79,73,78,84,95,83,73,90,69,0,80,73,88,69,76,95,83,73,90,69,0,82,69,83,79,76,85,84,73,79,78,95,88,0,82,69,83,79,76,85,84,73,79,78,95,89,0,49,48,54,52,54,0,56,56,53,57,0,49,0,112,102,114,0,112,102,114,45,109,101,116,114,105,99,115,0,80,70,82,0,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,0,1,2,3,4,5,6,7,8,9,255,255,255,255,255,255,255,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,255,255,255,255,255,255,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,255,255,255,255,255,83,116,97,114,116,70,111,110,116,77,101,116,114,105,99,115,0,65,115,99,101,110,100,101,114,0,65,120,105,115,76,97,98,101,108,0,65,120,105,115,84,121,112,101,0,67,67,0,67,72,0,67,97,112,72,101,105,103,104,116,0,67,104,97,114,87,105,100,116,104,0,67,104,97,114,97,99,116,101,114,83,101,116,0,67,104,97,114,97,99,116,101,114,115,0,68,101,115,99,101,110,100,101,114,0,69,110,99,111,100,105,110,103,83,99,104,101,109,101,0,69,110,100,65,120,105,115,0,69,110,100,67,104,97,114,77,101,116,114,105,99,115,0,69,110,100,67,111,109,112,111,115,105,116,101,115,0,69,110,100,68,105,114,101,99,116,105,111,110,0,69,110,100,70,111,110,116,77,101,116,114,105,99,115,0,69,110,100,75,101,114,110,68,97,116,97,0,69,110,100,75,101,114,110,80,97,105,114,115,0,69,110,100,84,114,97,99,107,75,101,114,110,0,69,115,99,67,104,97,114,0,73,115,66,97,115,101,70,111,110,116,0,73,115,67,73,68,70,111,110,116,0,73,115,70,105,120,101,100,80,105,116,99,104,0,73,115,70,105,120,101,100,86,0,75,80,0,75,80,72,0,75,80,88,0,75,80,89,0,77,97,112,112,105,110,103,83,99,104,101,109,101,0,77,101,116,114,105,99,115,83,101,116,115,0,78,0,80,67,67,0,83,116,97,114,116,65,120,105,115,0,83,116,97,114,116,67,104,97,114,77,101,116,114,105,99,115,0,83,116,97,114,116,67,111,109,112,111,115,105,116,101,115,0,83,116,97,114,116,68,105,114,101,99,116,105,111,110,0,83,116,97,114,116,75,101,114,110,68,97,116,97,0,83,116,97,114,116,75,101,114,110,80,97,105,114,115,0,83,116,97,114,116,75,101,114,110,80,97,105,114,115,48,0,83,116,97,114,116,75,101,114,110,80,97,105,114,115,49,0,83,116,97,114,116,84,114,97,99,107,75,101,114,110,0,84,114,97,99,107,75,101,114,110,0,86,86,0,86,86,101,99,116,111,114,0,86,101,114,115,105,111,110,0,87,0,87,48,0,87,48,88,0,87,48,89,0,87,49,0,87,49,88,0,87,49,89,0,87,88,0,87,89,0,88,72,101,105,103,104,116,0,112,115,110,97,109,101,115,0,0,52,0,106,2,167,3,63,4,220,6,125,9,143,10,23,11,137,12,199,14,246,15,87,16,233,17,219,18,104,19,88,22,110,23,32,23,71,24,77,27,156,29,73,31,247,32,107,32,222,33,55,34,154,35,218,58,10,64,122,72,188,80,109,88,104,93,61,98,168,106,91,114,111,115,237,122,180,127,255,135,164,143,132,149,213,158,108,161,115,168,175,183,147,197,199,202,25,204,166,208,209,209,81,215,26,65,143,0,65,0,140,0,175,0,193,1,15,1,147,1,233,1,251,2,7,2,40,2,57,2,82,2,91,2,128,2,136,2,154,69,131,0,198,0,150,0,158,0,167,225,227,245,244,101,128,1,252,237,225,227,242,239,110,128,1,226,243,237,225,236,108,128,247,230,225,227,245,244,101,129,0,193,0,185,243,237,225,236,108,128,247,225,226,242,229,246,101,134,1,2,0,213,0,221,0,232,0,243,0,251,1,7,225,227,245,244,101,128,30,174,227,249,242,233,236,236,233,99,128,4,208,228,239,244,226,229,236,239,119,128,30,182,231,242,225,246,101,128,30,176,232,239,239,235,225,226,239,246,101,128,30,178,244,233,236,228,101,128,30,180,99,4,1,25,1,32,1,121,1,137,225,242,239,110,128,1,205,233,242,99,2,1,40,1,45,236,101,128,36,182,245,237,230,236,229,120,134,0,194,1,66,1,74,1,85,1,93,1,105,1,113,225,227,245,244,101,128,30,164,228,239,244,226,229,236,239,119,128,30,172,231,242,225,246,101,128,30,166,232,239,239,235,225,226,239,246,101,128,30,168,243,237,225,236,108,128,247,226,244,233,236,228,101,128,30,170,245,244,101,129,246,201,1,129,243,237,225,236,108,128,247,180,249,242,233,236,236,233,99,128,4,16,100,3,1,155,1,165,1,209,226,236,231,242,225,246,101,128,2,0,233,229,242,229,243,233,115,131,0,196,1,181,1,192,1,201,227,249,242,233,236,236,233,99,128,4,210,237,225,227,242,239,110,128,1,222,243,237,225,236,108,128,247,228,239,116,2,1,216,1,224,226,229,236,239,119,128,30,160,237,225,227,242,239,110,128,1,224,231,242,225,246,101,129,0,192,1,243,243,237,225,236,108,128,247,224,232,239,239,235,225,226,239,246,101,128,30,162,105,2,2,13,2,25,229,227,249,242,233,236,236,233,99,128,4,212,238,246,229,242,244,229,228,226,242,229,246,101,128,2,2,236,240,232,97,129,3,145,2,49,244,239,238,239,115,128,3,134,109,2,2,63,2,71,225,227,242,239,110,128,1,0,239,238,239,243,240,225,227,101,128,255,33,239,231,239,238,229,107,128,1,4,242,233,238,103,131,0,197,2,104,2,112,2,120,225,227,245,244,101,128,1,250,226,229,236,239,119,128,30,0,243,237,225,236,108,128,247,229,243,237,225,236,108,128,247,97,244,233,236,228,101,129,0,195,2,146,243,237,225,236,108,128,247,227,249,226,225,242,237,229,238,233,225,110,128,5,49,66,137,0,66,2,189,2,198,2,223,3,3,3,10,3,22,3,34,3,46,3,54,227,233,242,227,236,101,128,36,183,228,239,116,2,2,206,2,215,225,227,227,229,238,116,128,30,2,226,229,236,239,119,128,30,4,101,3,2,231,2,242,2,254,227,249,242,233,236,236,233,99,128,4,17,238,225,242,237,229,238,233,225,110,128,5,50,244,97,128,3,146,232,239,239,107,128,1,129,236,233,238,229,226,229,236,239,119,128,30,6,237,239,238,239,243,240,225,227,101,128,255,34,242,229,246,229,243,237,225,236,108,128,246,244,243,237,225,236,108,128,247,98,244,239,240,226,225,114,128,1,130,67,137,0,67,3,85,3,127,3,193,3,210,3,224,4,171,4,188,4,200,4,212,97,3,3,93,3,104,3,111,225,242,237,229,238,233,225,110,128,5,62,227,245,244,101,128,1,6,242,239,110,129,246,202,3,119,243,237,225,236,108,128,246,245,99,3,3,135,3,142,3,171,225,242,239,110,128,1,12,229,228,233,236,236,97,130,0,199,3,155,3,163,225,227,245,244,101,128,30,8,243,237,225,236,108,128,247,231,233,242,99,2,3,179,3,184,236,101,128,36,184,245,237,230,236,229,120,128,1,8,228,239,116,129,1,10,3,201,225,227,227,229,238,116,128,1,10,229,228,233,236,236,225,243,237,225,236,108,128,247,184,104,4,3,234,3,246,4,161,4,165,225,225,242,237,229,238,233,225,110,128,5,73,101,6,4,4,4,24,4,35,4,103,4,115,4,136,225,226,235,232,225,243,233,225,238,227,249,242,233,236,236,233,99,128,4,188,227,249,242,233,236,236,233,99,128,4,39,100,2,4,41,4,85,229,243,227,229,238,228,229,114,2,4,54,4,74,225,226,235,232,225,243,233,225,238,227,249,242,233,236,236,233,99,128,4,190,227,249,242,233,236,236,233,99,128,4,182,233,229,242,229,243,233,243,227,249,242,233,236,236,233,99,128,4,244,232,225,242,237,229,238,233,225,110,128,5,67,235,232,225,235,225,243,243,233,225,238,227,249,242,233,236,236,233,99,128,4,203,246,229,242,244,233,227,225,236,243,244,242,239,235,229,227,249,242,233,236,236,233,99,128,4,184,105,128,3,167,239,239,107,128,1,135,233,242,227,245,237,230,236,229,248,243,237,225,236,108,128,246,246,237,239,238,239,243,240,225,227,101,128,255,35,239,225,242,237,229,238,233,225,110,128,5,81,243,237,225,236,108,128,247,99,68,142,0,68,4,252,5,10,5,36,5,96,5,121,5,166,5,173,5,231,5,244,6,0,6,12,6,28,6,48,6,57,90,129,1,241,5,2,227,225,242,239,110,128,1,196,97,2,5,16,5,27,225,242,237,229,238,233,225,110,128,5,52,230,242,233,227,225,110,128,1,137,99,4,5,46,5,53,5,62,5,89,225,242,239,110,128,1,14,229,228,233,236,236,97,128,30,16,233,242,99,2,5,70,5,75,236,101,128,36,185,245,237,230,236,229,248,226,229,236,239,119,128,30,18,242,239,225,116,128,1,16,228,239,116,2,5,104,5,113,225,227,227,229,238,116,128,30,10,226,229,236,239,119,128,30,12,101,3,5,129,5,140,5,150,227,249,242,233,236,236,233,99,128,4,20,233,227,239,240,244,233,99,128,3,238,236,244,97,129,34,6,5,158,231,242,229,229,107,128,3,148,232,239,239,107,128,1,138,105,2,5,179,5,218,229,242,229,243,233,115,131,246,203,5,194,5,202,5,210,193,227,245,244,101,128,246,204,199,242,225,246,101,128,246,205,243,237,225,236,108,128,247,168,231,225,237,237,225,231,242,229,229,107,128,3,220,234,229,227,249,242,233,236,236,233,99,128,4,2,236,233,238,229,226,229,236,239,119,128,30,14,237,239,238,239,243,240,225,227,101,128,255,36,239,244,225,227,227,229,238,244,243,237,225,236,108,128,246,247,115,2,6,34,6,41,236,225,243,104,128,1,16,237,225,236,108,128,247,100,244,239,240,226,225,114,128,1,139,122,131,1,242,6,67,6,75,6,112,227,225,242,239,110,128,1,197,101,2,6,81,6,101,225,226,235,232,225,243,233,225,238,227,249,242,233,236,236,233,99,128,4,224,227,249,242,233,236,236,233,99,128,4,5,232,229,227,249,242,233,236,236,233,99,128,4,15,69,146,0,69,6,165,6,183,6,191,7,89,7,153,7,165,7,183,7,211,8,7,8,36,8,94,8,169,8,189,8,208,8,248,9,44,9,109,9,115,225,227,245,244,101,129,0,201,6,175,243,237,225,236,108,128,247,233,226,242,229,246,101,128,1,20,99,5,6,203,6,210,6,224,6,236,7,79,225,242,239,110,128,1,26,229,228,233,236,236,225,226,242,229,246,101,128,30,28,232,225,242,237,229,238,233,225,110,128,5,53,233,242,99,2,6,244,6,249,236,101,128,36,186,245,237,230,236,229,120,135,0,202,7,16,7,24,7,32,7,43,7,51,7,63,7,71,225,227,245,244,101,128,30,190,226,229,236,239,119,128,30,24,228,239,244,226,229,236,239,119,128,30,198,231,242,225,246,101,128,30,192,232,239,239,235,225,226,239,246,101,128,30,194,243,237,225,236,108,128,247,234,244,233,236,228,101,128,30,196,249,242,233,236,236,233,99,128,4,4,100,3,7,97,7,107,7,127,226,236,231,242,225,246,101,128,2,4,233,229,242,229,243,233,115,129,0,203,7,119,243,237,225,236,108,128,247,235,239,116,130,1,22,7,136,7,145,225,227,227,229,238,116,128,1,22,226,229,236,239,119,128,30,184,230,227,249,242,233,236,236,233,99,128,4,36,231,242,225,246,101,129,0,200,7,175,243,237,225,236,108,128,247,232,104,2,7,189,7,200,225,242,237,229,238,233,225,110,128,5,55,239,239,235,225,226,239,246,101,128,30,186,105,3,7,219,7,230,7,245,231,232,244,242,239,237,225,110,128,33,103,238,246,229,242,244,229,228,226,242,229,246,101,128,2,6,239,244,233,230,233,229,228,227,249,242,233,236,236,233,99,128,4,100,108,2,8,13,8,24,227,249,242,233,236,236,233,99,128,4,27,229,246,229,238,242,239,237,225,110,128,33,106,109,3,8,44,8,72,8,83,225,227,242,239,110,130,1,18,8,56,8,64,225,227,245,244,101,128,30,22,231,242,225,246,101,128,30,20,227,249,242,233,236,236,233,99,128,4,28,239,238,239,243,240,225,227,101,128,255,37,110,4,8,104,8,115,8,135,8,154,227,249,242,233,236,236,233,99,128,4,29,228,229,243,227,229,238,228,229,242,227,249,242,233,236,236,233,99,128,4,162,103,129,1,74,8,141,232,229,227,249,242,233,236,236,233,99,128,4,164,232,239,239,235,227,249,242,233,236,236,233,99,128,4,199,111,2,8,175,8,183,231,239,238,229,107,128,1,24,240,229,110,128,1,144,240,243,233,236,239,110,129,3,149,8,200,244,239,238,239,115,128,3,136,114,2,8,214,8,225,227,249,242,233,236,236,233,99,128,4,32,229,246,229,242,243,229,100,129,1,142,8,237,227,249,242,233,236,236,233,99,128,4,45,115,4,9,2,9,13,9,33,9,37,227,249,242,233,236,236,233,99,128,4,33,228,229,243,227,229,238,228,229,242,227,249,242,233,236,236,233,99,128,4,170,104,128,1,169,237,225,236,108,128,247,101,116,3,9,52,9,78,9,92,97,130,3,151,9,60,9,70,242,237,229,238,233,225,110,128,5,56,244,239,238,239,115,128,3,137,104,129,0,208,9,84,243,237,225,236,108,128,247,240,233,236,228,101,129,30,188,9,101,226,229,236,239,119,128,30,26,245,242,111,128,32,172,250,104,130,1,183,9,124,9,132,227,225,242,239,110,128,1,238,242,229,246,229,242,243,229,100,128,1,184,70,136,0,70,9,163,9,172,9,184,9,212,9,219,9,248,10,4,10,15,227,233,242,227,236,101,128,36,187,228,239,244,225,227,227,229,238,116,128,30,30,101,2,9,190,9,202,232,225,242,237,229,238,233,225,110,128,5,86,233,227,239,240,244,233,99,128,3,228,232,239,239,107,128,1,145,105,2,9,225,9,238,244,225,227,249,242,233,236,236,233,99,128,4,114,246,229,242,239,237,225,110,128,33,100,237,239,238,239,243,240,225,227,101,128,255,38,239,245,242,242,239,237,225,110,128,33,99,243,237,225,236,108,128,247,102,71,140,0,71,10,51,10,61,10,107,10,115,10,176,10,193,10,205,11,39,11,52,11,65,11,90,11,107,194,243,241,245,225,242,101,128,51,135,97,3,10,69,10,76,10,94,227,245,244,101,128,1,244,237,237,97,129,3,147,10,84,225,230,242,233,227,225,110,128,1,148,238,231,233,225,227,239,240,244,233,99,128,3,234,226,242,229,246,101,128,1,30,99,4,10,125,10,132,10,141,10,163,225,242,239,110,128,1,230,229,228,233,236,236,97,128,1,34,233,242,99,2,10,149,10,154,236,101,128,36,188,245,237,230,236,229,120,128,1,28,239,237,237,225,225,227,227,229,238,116,128,1,34,228,239,116,129,1,32,10,184,225,227,227,229,238,116,128,1,32,229,227,249,242,233,236,236,233,99,128,4,19,104,3,10,213,10,226,11,33,225,228,225,242,237,229,238,233,225,110,128,5,66,101,3,10,234,10,255,11,16,237,233,228,228,236,229,232,239,239,235,227,249,242,233,236,236,233,99,128,4,148,243,244,242,239,235,229,227,249,242,233,236,236,233,99,128,4,146,245,240,244,245,242,238,227,249,242,233,236,236,233,99,128,4,144,239,239,107,128,1,147,233,237,225,242,237,229,238,233,225,110,128,5,51,234,229,227,249,242,233,236,236,233,99,128,4,3,109,2,11,71,11,79,225,227,242,239,110,128,30,32,239,238,239,243,240,225,227,101,128,255,39,242,225,246,101,129,246,206,11,99,243,237,225,236,108,128,247,96,115,2,11,113,11,129,237,225,236,108,129,247,103,11,122,232,239,239,107,128,2,155,244,242,239,235,101,128,1,228,72,140,0,72,11,165,11,190,11,198,11,208,12,17,12,40,12,77,12,117,12,129,12,157,12,165,12,189,177,184,53,3,11,175,11,180,11,185,179,51,128,37,207,180,51,128,37,170,181,49,128,37,171,178,178,176,183,51,128,37,161,208,243,241,245,225,242,101,128,51,203,97,3,11,216,11,236,12,0,225,226,235,232,225,243,233,225,238,227,249,242,233,236,236,233,99,128,4,168,228,229,243,227,229,238,228,229,242,227,249,242,233,236,236,233,99,128,4,178,242,228,243,233,231,238,227,249,242,233,236,236,233,99,128,4,42,98,2,12,23,12,28,225,114,128,1,38,242,229,246,229,226,229,236,239,119,128,30,42,99,2,12,46,12,55,229,228,233,236,236,97,128,30,40,233,242,99,2,12,63,12,68,236,101,128,36,189,245,237,230,236,229,120,128,1,36,100,2,12,83,12,93,233,229,242,229,243,233,115,128,30,38,239,116,2,12,100,12,109,225,227,227,229,238,116,128,30,34,226,229,236,239,119,128,30,36,237,239,238,239,243,240,225,227,101,128,255,40,111,2,12,135,12,146,225,242,237,229,238,233,225,110,128,5,64,242,233,227,239,240,244,233,99,128,3,232,243,237,225,236,108,128,247,104,245,238,231,225,242,245,237,236,225,245,116,129,246,207,12,181,243,237,225,236,108,128,246,248,250,243,241,245,225,242,101,128,51,144,73,146,0,73,12,239,12,251,12,255,13,11,13,29,13,37,13,94,13,181,13,214,13,224,13,242,13,254,14,48,14,86,14,99,14,166,14,187,14,205,193,227,249,242,233,236,236,233,99,128,4,47,74,128,1,50,213,227,249,242,233,236,236,233,99,128,4,46,225,227,245,244,101,129,0,205,13,21,243,237,225,236,108,128,247,237,226,242,229,246,101,128,1,44,99,3,13,45,13,52,13,84,225,242,239,110,128,1,207,233,242,99,2,13,60,13,65,236,101,128,36,190,245,237,230,236,229,120,129,0,206,13,76,243,237,225,236,108,128,247,238,249,242,233,236,236,233,99,128,4,6,100,3,13,102,13,112,13,155,226,236,231,242,225,246,101,128,2,8,233,229,242,229,243,233,115,131,0,207,13,128,13,136,13,147,225,227,245,244,101,128,30,46,227,249,242,233,236,236,233,99,128,4,228,243,237,225,236,108,128,247,239,239,116,130,1,48,13,164,13,173,225,227,227,229,238,116,128,1,48,226,229,236,239,119,128,30,202,101,2,13,187,13,203,226,242,229,246,229,227,249,242,233,236,236,233,99,128,4,214,227,249,242,233,236,236,233,99,128,4,21,230,242,225,235,244,245,114,128,33,17,231,242,225,246,101,129,0,204,13,234,243,237,225,236,108,128,247,236,232,239,239,235,225,226,239,246,101,128,30,200,105,3,14,6,14,17,14,32,227,249,242,233,236,236,233,99,128,4,24,238,246,229,242,244,229,228,226,242,229,246,101,128,2,10,243,232,239,242,244,227,249,242,233,236,236,233,99,128,4,25,109,2,14,54,14,75,225,227,242,239,110,129,1,42,14,64,227,249,242,233,236,236,233,99,128,4,226,239,238,239,243,240,225,227,101,128,255,41,238,233,225,242,237,229,238,233,225,110,128,5,59,111,3,14,107,14,118,14,126,227,249,242,233,236,236,233,99,128,4,1,231,239,238,229,107,128,1,46,244,97,131,3,153,14,137,14,147,14,158,225,230,242,233,227,225,110,128,1,150,228,233,229,242,229,243,233,115,128,3,170,244,239,238,239,115,128,3,138,115,2,14,172,14,179,237,225,236,108,128,247,105,244,242,239,235,101,128,1,151,244,233,236,228,101,129,1,40,14,197,226,229,236,239,119,128,30,44,250,232,233,244,243,97,2,14,216,14,227,227,249,242,233,236,236,233,99,128,4,116,228,226,236,231,242,225,246,229,227,249,242,233,236,236,233,99,128,4,118,74,134,0,74,15,6,15,18,15,41,15,53,15,67,15,79,225,225,242,237,229,238,233,225,110,128,5,65,227,233,242,99,2,15,27,15,32,236,101,128,36,191,245,237,230,236,229,120,128,1,52,229,227,249,242,233,236,236,233,99,128,4,8,232,229,232,225,242,237,229,238,233,225,110,128,5,75,237,239,238,239,243,240,225,227,101,128,255,42,243,237,225,236,108,128,247,106,75,140,0,75,15,115,15,125,15,135,16,18,16,65,16,76,16,106,16,143,16,156,16,168,16,180,16,208,194,243,241,245,225,242,101,128,51,133,203,243,241,245,225,242,101,128,51,205,97,7,15,151,15,169,15,191,15,211,15,226,15,232,15,249,226,225,243,232,235,233,242,227,249,242,233,236,236,233,99,128,4,160,99,2,15,175,15,181,245,244,101,128,30,48,249,242,233,236,236,233,99,128,4,26,228,229,243,227,229,238,228,229,242,227,249,242,233,236,236,233,99,128,4,154,232,239,239,235,227,249,242,233,236,236,233,99,128,4,195,240,240,97,128,3,154,243,244,242,239,235,229,227,249,242,233,236,236,233,99,128,4,158,246,229,242,244,233,227,225,236,243,244,242,239,235,229,227,249,242,233,236,236,233,99,128,4,156,99,4,16,28,16,35,16,44,16,52,225,242,239,110,128,1,232,229,228,233,236,236,97,128,1,54,233,242,227,236,101,128,36,192,239,237,237,225,225,227,227,229,238,116,128,1,54,228,239,244,226,229,236,239,119,128,30,50,101,2,16,82,16,94,232,225,242,237,229,238,233,225,110,128,5,84,238,225,242,237,229,238,233,225,110,128,5,63,104,3,16,114,16,126,16,137,225,227,249,242,233,236,236,233,99,128,4,37,229,233,227,239,240,244,233,99,128,3,230,239,239,107,128,1,152,234,229,227,249,242,233,236,236,233,99,128,4,12,236,233,238,229,226,229,236,239,119,128,30,52,237,239,238,239,243,240,225,227,101,128,255,43,239,240,240,97,2,16,189,16,200,227,249,242,233,236,236,233,99,128,4,128,231,242,229,229,107,128,3,222,115,2,16,214,16,226,233,227,249,242,233,236,236,233,99,128,4,110,237,225,236,108,128,247,107,76,138,0,76,17,1,17,5,17,9,17,29,17,95,17,133,17,147,17,165,17,177,17,189,74,128,1,199,76,128,246,191,97,2,17,15,17,22,227,245,244,101,128,1,57,237,226,228,97,128,3,155,99,4,17,39,17,46,17,55,17,82,225,242,239,110,128,1,61,229,228,233,236,236,97,128,1,59,233,242,99,2,17,63,17,68,236,101,128,36,193,245,237,230,236,229,248,226,229,236,239,119,128,30,60,239,237,237,225,225,227,227,229,238,116,128,1,59,228,239,116,130,1,63,17,105,17,114,225,227,227,229,238,116,128,1,63,226,229,236,239,119,129,30,54,17,124,237,225,227,242,239,110,128,30,56,233,247,238,225,242,237,229,238,233,225,110,128,5,60,106,129,1,200,17,153,229,227,249,242,233,236,236,233,99,128,4,9,236,233,238,229,226,229,236,239,119,128,30,58,237,239,238,239,243,240,225,227,101,128,255,44,115,2,17,195,17,212,236,225,243,104,129,1,65,17,204,243,237,225,236,108,128,246,249,237,225,236,108,128,247,108,77,137,0,77,17,241,17,251,18,24,18,33,18,58,18,71,18,83,18,91,18,100,194,243,241,245,225,242,101,128,51,134,225,99,2,18,2,18,18,242,239,110,129,246,208,18,10,243,237,225,236,108,128,247,175,245,244,101,128,30,62,227,233,242,227,236,101,128,36,194,228,239,116,2,18,41,18,50,225,227,227,229,238,116,128,30,64,226,229,236,239,119,128,30,66,229,238,225,242,237,229,238,233,225,110,128,5,68,237,239,238,239,243,240,225,227,101,128,255,45,243,237,225,236,108,128,247,109,244,245,242,238,229,100,128,1,156,117,128,3,156,78,141,0,78,18,134,18,138,18,146,18,212,18,237,18,248,19,3,19,21,19,33,19,45,19,58,19,66,19,84,74,128,1,202,225,227,245,244,101,128,1,67,99,4,18,156,18,163,18,172,18,199,225,242,239,110,128,1,71,229,228,233,236,236,97,128,1,69,233,242,99,2,18,180,18,185,236,101,128,36,195,245,237,230,236,229,248,226,229,236,239,119,128,30,74,239,237,237,225,225,227,227,229,238,116,128,1,69,228,239,116,2,18,220,18,229,225,227,227,229,238,116,128,30,68,226,229,236,239,119,128,30,70,232,239,239,235,236,229,230,116,128,1,157,233,238,229,242,239,237,225,110,128,33,104,106,129,1,203,19,9,229,227,249,242,233,236,236,233,99,128,4,10,236,233,238,229,226,229,236,239,119,128,30,72,237,239,238,239,243,240,225,227,101,128,255,46,239,247,225,242,237,229,238,233,225,110,128,5,70,243,237,225,236,108,128,247,110,244,233,236,228,101,129,0,209,19,76,243,237,225,236,108,128,247,241,117,128,3,157,79,141,0,79,19,118,19,132,19,150,19,203,20,78,20,152,20,187,21,48,21,69,21,213,21,223,21,254,22,53,69,129,1,82,19,124,243,237,225,236,108,128,246,250,225,227,245,244,101,129,0,211,19,142,243,237,225,236,108,128,247,243,98,2,19,156,19,196,225,242,242,229,100,2,19,166,19,177,227,249,242,233,236,236,233,99,128,4,232,228,233,229,242,229,243,233,243,227,249,242,233,236,236,233,99,128,4,234,242,229,246,101,128,1,78,99,4,19,213,19,220,19,235,20,68,225,242,239,110,128,1,209,229,238,244,229,242,229,228,244,233,236,228,101,128,1,159,233,242,99,2,19,243,19,248,236,101,128,36,196,245,237,230,236,229,120,134,0,212,20,13,20,21,20,32,20,40,20,52,20,60,225,227,245,244,101,128,30,208,228,239,244,226,229,236,239,119,128,30,216,231,242,225,246,101,128,30,210,232,239,239,235,225,226,239,246,101,128,30,212,243,237,225,236,108,128,247,244,244,233,236,228,101,128,30,214,249,242,233,236,236,233,99,128,4,30,100,3,20,86,20,109,20,142,226,108,2,20,93,20,101,225,227,245,244,101,128,1,80,231,242,225,246,101,128,2,12,233,229,242,229,243,233,115,130,0,214,20,123,20,134,227,249,242,233,236,236,233,99,128,4,230,243,237,225,236,108,128,247,246,239,244,226,229,236,239,119,128,30,204,103,2,20,158,20,170,239,238,229,235,243,237,225,236,108,128,246,251,242,225,246,101,129,0,210,20,179,243,237,225,236,108,128,247,242,104,4,20,197,20,208,20,212,21,34,225,242,237,229,238,233,225,110,128,5,85,109,128,33,38,111,2,20,218,20,228,239,235,225,226,239,246,101,128,30,206,242,110,133,1,160,20,243,20,251,21,6,21,14,21,26,225,227,245,244,101,128,30,218,228,239,244,226,229,236,239,119,128,30,226,231,242,225,246,101,128,30,220,232,239,239,235,225,226,239,246,101,128,30,222,244,233,236,228,101,128,30,224,245,238,231,225,242,245,237,236,225,245,116,128,1,80,105,129,1,162,21,54,238,246,229,242,244,229,228,226,242,229,246,101,128,2,14,109,4,21,79,21,107,21,184,21,202,225,227,242,239,110,130,1,76,21,91,21,99,225,227,245,244,101,128,30,82,231,242,225,246,101,128,30,80,229,231,97,132,33,38,21,121,21,132,21,140,21,156,227,249,242,233,236,236,233,99,128,4,96,231,242,229,229,107,128,3,169,242,239,245,238,228,227,249,242,233,236,236,233,99,128,4,122,116,2,21,162,21,177,233,244,236,239,227,249,242,233,236,236,233,99,128,4,124,239,238,239,115,128,3,143,233,227,242,239,110,129,3,159,21,194,244,239,238,239,115,128,3,140,239,238,239,243,240,225,227,101,128,255,47,238,229,242,239,237,225,110,128,33,96,111,2,21,229,21,248,231,239,238,229,107,129,1,234,21,239,237,225,227,242,239,110,128,1,236,240,229,110,128,1,134,115,3,22,6,22,33,22,40,236,225,243,104,130,0,216,22,17,22,25,225,227,245,244,101,128,1,254,243,237,225,236,108,128,247,248,237,225,236,108,128,247,111,244,242,239,235,229,225,227,245,244,101,128,1,254,116,2,22,59,22,70,227,249,242,233,236,236,233,99,128,4,126,233,236,228,101,131,0,213,22,83,22,91,22,102,225,227,245,244,101,128,30,76,228,233,229,242,229,243,233,115,128,30,78,243,237,225,236,108,128,247,245,80,136,0,80,22,130,22,138,22,147,22,159,22,211,22,227,22,246,23,2,225,227,245,244,101,128,30,84,227,233,242,227,236,101,128,36,197,228,239,244,225,227,227,229,238,116,128,30,86,101,3,22,167,22,178,22,190,227,249,242,233,236,236,233,99,128,4,31,232,225,242,237,229,238,233,225,110,128,5,74,237,233,228,228,236,229,232,239,239,235,227,249,242,233,236,236,233,99,128,4,166,104,2,22,217,22,221,105,128,3,166,239,239,107,128,1,164,105,129,3,160,22,233,247,242,225,242,237,229,238,233,225,110,128,5,83,237,239,238,239,243,240,225,227,101,128,255,48,115,2,23,8,23,25,105,129,3,168,23,14,227,249,242,233,236,236,233,99,128,4,112,237,225,236,108,128,247,112,81,131,0,81,23,42,23,51,23,63,227,233,242,227,236,101,128,36,198,237,239,238,239,243,240,225,227,101,128,255,49,243,237,225,236,108,128,247,113,82,138,0,82,23,95,23,119,23,166,23,217,23,230,23,240,23,245,24,19,24,31,24,43,97,2,23,101,23,112,225,242,237,229,238,233,225,110,128,5,76,227,245,244,101,128,1,84,99,4,23,129,23,136,23,145,23,153,225,242,239,110,128,1,88,229,228,233,236,236,97,128,1,86,233,242,227,236,101,128,36,199,239,237,237,225,225,227,227,229,238,116,128,1,86,100,2,23,172,23,182,226,236,231,242,225,246,101,128,2,16,239,116,2,23,189,23,198,225,227,227,229,238,116,128,30,88,226,229,236,239,119,129,30,90,23,208,237,225,227,242,239,110,128,30,92,229,232,225,242,237,229,238,233,225,110,128,5,80,230,242,225,235,244,245,114,128,33,28,232,111,128,3,161,233,110,2,23,252,24,5,231,243,237,225,236,108,128,246,252,246,229,242,244,229,228,226,242,229,246,101,128,2,18,236,233,238,229,226,229,236,239,119,128,30,94,237,239,238,239,243,240,225,227,101,128,255,50,243,237,225,236,108,129,247,114,24,53,233,238,246,229,242,244,229,100,129,2,129,24,66,243,245,240,229,242,233,239,114,128,2,182,83,139,0,83,24,103,26,17,26,55,26,182,26,221,26,250,27,84,27,105,27,117,27,135,27,143,70,6,24,117,24,209,24,241,25,77,25,119,25,221,48,9,24,137,24,145,24,153,24,161,24,169,24,177,24,185,24,193,24,201,177,176,176,176,48,128,37,12,178,176,176,176,48,128,37,20,179,176,176,176,48,128,37,16,180,176,176,176,48,128,37,24,181,176,176,176,48,128,37,60,182,176,176,176,48,128,37,44,183,176,176,176,48,128,37,52,184,176,176,176,48,128,37,28,185,176,176,176,48,128,37,36,49,3,24,217,24,225,24,233,176,176,176,176,48,128,37,0,177,176,176,176,48,128,37,2,185,176,176,176,48,128,37,97,50,9,25,5,25,13,25,21,25,29,25,37,25,45,25,53,25,61,25,69,176,176,176,176,48,128,37,98,177,176,176,176,48,128,37,86,178,176,176,176,48,128,37,85,179,176,176,176,48,128,37,99,180,176,176,176,48,128,37,81,181,176,176,176,48,128,37,87,182,176,176,176,48,128,37,93,183,176,176,176,48,128,37,92,184,176,176,176,48,128,37,91,51,4,25,87,25,95,25,103,25,111,182,176,176,176,48,128,37,94,183,176,176,176,48,128,37,95,184,176,176,176,48,128,37,90,185,176,176,176,48,128,37,84,52,10,25,141,25,149,25,157,25,165,25,173,25,181,25,189,25,197,25,205,25,213,176,176,176,176,48,128,37,105,177,176,176,176,48,128,37,102,178,176,176,176,48,128,37,96,179,176,176,176,48,128,37,80,180,176,176,176,48,128,37,108,181,176,176,176,48,128,37,103,182,176,176,176,48,128,37,104,183,176,176,176,48,128,37,100,184,176,176,176,48,128,37,101,185,176,176,176,48,128,37,89,53,5,25,233,25,241,25,249,26,1,26,9,176,176,176,176,48,128,37,88,177,176,176,176,48,128,37,82,178,176,176,176,48,128,37,83,179,176,176,176,48,128,37,107,180,176,176,176,48,128,37,106,97,2,26,23,26,44,227,245,244,101,129,1,90,26,32,228,239,244,225,227,227,229,238,116,128,30,100,237,240,233,231,242,229,229,107,128,3,224,99,5,26,67,26,98,26,107,26,147,26,169,225,242,239,110,130,1,96,26,78,26,90,228,239,244,225,227,227,229,238,116,128,30,102,243,237,225,236,108,128,246,253,229,228,233,236,236,97,128,1,94,232,247,97,130,1,143,26,117,26,128,227,249,242,233,236,236,233,99,128,4,216,228,233,229,242,229,243,233,243,227,249,242,233,236,236,233,99,128,4,218,233,242,99,2,26,155,26,160,236,101,128,36,200,245,237,230,236,229,120,128,1,92,239,237,237,225,225,227,227,229,238,116,128,2,24,228,239,116,2,26,190,26,199,225,227,227,229,238,116,128,30,96,226,229,236,239,119,129,30,98,26,209,228,239,244,225,227,227,229,238,116,128,30,104,101,2,26,227,26,239,232,225,242,237,229,238,233,225,110,128,5,77,246,229,238,242,239,237,225,110,128,33,102,104,5,27,6,27,34,27,48,27,59,27,72,97,2,27,12,27,23,225,242,237,229,238,233,225,110,128,5,71,227,249,242,233,236,236,233,99,128,4,40,227,232,225,227,249,242,233,236,236,233,99,128,4,41,229,233,227,239,240,244,233,99,128,3,226,232,225,227,249,242,233,236,236,233,99,128,4,186,233,237,225,227,239,240,244,233,99,128,3,236,105,2,27,90,27,96,231,237,97,128,3,163,248,242,239,237,225,110,128,33,101,237,239,238,239,243,240,225,227,101,128,255,51,239,230,244,243,233,231,238,227,249,242,233,236,236,233,99,128,4,44,243,237,225,236,108,128,247,115,244,233,231,237,225,231,242,229,229,107,128,3,218,84,141,0,84,27,186,27,191,27,197,28,7,28,32,28,96,28,147,28,177,28,189,28,201,28,246,29,6,29,46,225,117,128,3,164,226,225,114,128,1,102,99,4,27,207,27,214,27,223,27,250,225,242,239,110,128,1,100,229,228,233,236,236,97,128,1,98,233,242,99,2,27,231,27,236,236,101,128,36,201,245,237,230,236,229,248,226,229,236,239,119,128,30,112,239,237,237,225,225,227,227,229,238,116,128,1,98,228,239,116,2,28,15,28,24,225,227,227,229,238,116,128,30,106,226,229,236,239,119,128,30,108,101,4,28,42,28,53,28,73,28,82,227,249,242,233,236,236,233,99,128,4,34,228,229,243,227,229,238,228,229,242,227,249,242,233,236,236,233,99,128,4,172,238,242,239,237,225,110,128,33,105,244,243,229,227,249,242,233,236,236,233,99,128,4,180,104,3,28,104,28,110,28,136,229,244,97,128,3,152,111,2,28,116,28,121,239,107,128,1,172,242,110,129,0,222,28,128,243,237,225,236,108,128,247,254,242,229,229,242,239,237,225,110,128,33,98,105,2,28,153,28,164,236,228,229,243,237,225,236,108,128,246,254,247,238,225,242,237,229,238,233,225,110,128,5,79,236,233,238,229,226,229,236,239,119,128,30,110,237,239,238,239,243,240,225,227,101,128,255,52,111,2,28,207,28,218,225,242,237,229,238,233,225,110,128,5,57,238,101,3,28,227,28,234,28,240,230,233,246,101,128,1,188,243,233,120,128,1,132,244,247,111,128,1,167,242,229,244,242,239,230,236,229,248,232,239,239,107,128,1,174,115,3,29,14,29,26,29,39,229,227,249,242,233,236,236,233,99,128,4,38,232,229,227,249,242,233,236,236,233,99,128,4,11,237,225,236,108,128,247,116,119,2,29,52,29,64,229,236,246,229,242,239,237,225,110,128,33,107,239,242,239,237,225,110,128,33,97,85,142,0,85,29,105,29,123,29,131,29,198,30,69,30,87,30,198,30,214,30,226,31,21,31,30,31,142,31,149,31,219,225,227,245,244,101,129,0,218,29,115,243,237,225,236,108,128,247,250,226,242,229,246,101,128,1,108,99,3,29,139,29,146,29,188,225,242,239,110,128,1,211,233,242,99,2,29,154,29,159,236,101,128,36,202,245,237,230,236,229,120,130,0,219,29,172,29,180,226,229,236,239,119,128,30,118,243,237,225,236,108,128,247,251,249,242,233,236,236,233,99,128,4,35,100,3,29,206,29,229,30,59,226,108,2,29,213,29,221,225,227,245,244,101,128,1,112,231,242,225,246,101,128,2,20,233,229,242,229,243,233,115,134,0,220,29,251,30,3,30,11,30,34,30,42,30,51,225,227,245,244,101,128,1,215,226,229,236,239,119,128,30,114,99,2,30,17,30,24,225,242,239,110,128,1,217,249,242,233,236,236,233,99,128,4,240,231,242,225,246,101,128,1,219,237,225,227,242,239,110,128,1,213,243,237,225,236,108,128,247,252,239,244,226,229,236,239,119,128,30,228,231,242,225,246,101,129,0,217,30,79,243,237,225,236,108,128,247,249,104,2,30,93,30,171,111,2,30,99,30,109,239,235,225,226,239,246,101,128,30,230,242,110,133,1,175,30,124,30,132,30,143,30,151,30,163,225,227,245,244,101,128,30,232,228,239,244,226,229,236,239,119,128,30,240,231,242,225,246,101,128,30,234,232,239,239,235,225,226,239,246,101,128,30,236,244,233,236,228,101,128,30,238,245,238,231,225,242,245,237,236,225,245,116,129,1,112,30,187,227,249,242,233,236,236,233,99,128,4,242,233,238,246,229,242,244,229,228,226,242,229,246,101,128,2,22,235,227,249,242,233,236,236,233,99,128,4,120,109,2,30,232,31,10,225,227,242,239,110,130,1,106,30,244,30,255,227,249,242,233,236,236,233,99,128,4,238,228,233,229,242,229,243,233,115,128,30,122,239,238,239,243,240,225,227,101,128,255,53,239,231,239,238,229,107,128,1,114,240,243,233,236,239,110,133,3,165,31,49,31,53,31,90,31,121,31,134,49,128,3,210,97,2,31,59,31,81,227,245,244,229,232,239,239,235,243,249,237,226,239,236,231,242,229,229,107,128,3,211,230,242,233,227,225,110,128,1,177,228,233,229,242,229,243,233,115,129,3,171,31,103,232,239,239,235,243,249,237,226,239,236,231,242,229,229,107,128,3,212,232,239,239,235,243,249,237,226,239,108,128,3,210,244,239,238,239,115,128,3,142,242,233,238,103,128,1,110,115,3,31,157,31,172,31,179,232,239,242,244,227,249,242,233,236,236,233,99,128,4,14,237,225,236,108,128,247,117,244,242,225,233,231,232,116,2,31,191,31,202,227,249,242,233,236,236,233,99,128,4,174,243,244,242,239,235,229,227,249,242,233,236,236,233,99,128,4,176,244,233,236,228,101,130,1,104,31,231,31,239,225,227,245,244,101,128,30,120,226,229,236,239,119,128,30,116,86,136,0,86,32,11,32,20,32,31,32,60,32,67,32,79,32,91,32,99,227,233,242,227,236,101,128,36,203,228,239,244,226,229,236,239,119,128,30,126,101,2,32,37,32,48,227,249,242,233,236,236,233,99,128,4,18,247,225,242,237,229,238,233,225,110,128,5,78,232,239,239,107,128,1,178,237,239,238,239,243,240,225,227,101,128,255,54,239,225,242,237,229,238,233,225,110,128,5,72,243,237,225,236,108,128,247,118,244,233,236,228,101,128,30,124,87,134,0,87,32,123,32,131,32,154,32,194,32,202,32,214,225,227,245,244,101,128,30,130,227,233,242,99,2,32,140,32,145,236,101,128,36,204,245,237,230,236,229,120,128,1,116,100,2,32,160,32,170,233,229,242,229,243,233,115,128,30,132,239,116,2,32,177,32,186,225,227,227,229,238,116,128,30,134,226,229,236,239,119,128,30,136,231,242,225,246,101,128,30,128,237,239,238,239,243,240,225,227,101,128,255,55,243,237,225,236,108,128,247,119,88,134,0,88,32,238,32,247,33,18,33,31,33,35,33,47,227,233,242,227,236,101,128,36,205,100,2,32,253,33,7,233,229,242,229,243,233,115,128,30,140,239,244,225,227,227,229,238,116,128,30,138,229,232,225,242,237,229,238,233,225,110,128,5,61,105,128,3,158,237,239,238,239,243,240,225,227,101,128,255,56,243,237,225,236,108,128,247,120,89,139,0,89,33,81,33,116,33,139,33,189,33,228,33,236,33,253,34,40,34,52,34,60,34,68,97,2,33,87,33,104,227,245,244,101,129,0,221,33,96,243,237,225,236,108,128,247,253,244,227,249,242,233,236,236,233,99,128,4,98,227,233,242,99,2,33,125,33,130,236,101,128,36,206,245,237,230,236,229,120,128,1,118,100,2,33,145,33,165,233,229,242,229,243,233,115,129,1,120,33,157,243,237,225,236,108,128,247,255,239,116,2,33,172,33,181,225,227,227,229,238,116,128,30,142,226,229,236,239,119,128,30,244,229,114,2,33,196,33,208,233,227,249,242,233,236,236,233,99,128,4,43,245,228,233,229,242,229,243,233,243,227,249,242,233,236,236,233,99,128,4,248,231,242,225,246,101,128,30,242,232,239,239,107,129,1,179,33,245,225,226,239,246,101,128,30,246,105,3,34,5,34,16,34,27,225,242,237,229,238,233,225,110,128,5,69,227,249,242,233,236,236,233,99,128,4,7,247,238,225,242,237,229,238,233,225,110,128,5,82,237,239,238,239,243,240,225,227,101,128,255,57,243,237,225,236,108,128,247,121,244,233,236,228,101,128,30,248,245,115,2,34,75,34,113,226,233,103,2,34,83,34,94,227,249,242,233,236,236,233,99,128,4,106,233,239,244,233,230,233,229,228,227,249,242,233,236,236,233,99,128,4,108,236,233,244,244,236,101,2,34,124,34,135,227,249,242,233,236,236,233,99,128,4,102,233,239,244,233,230,233,229,228,227,249,242,233,236,236,233,99,128,4,104,90,136,0,90,34,174,34,198,34,243,35,14,35,81,35,173,35,185,35,197,97,2,34,180,34,191,225,242,237,229,238,233,225,110,128,5,54,227,245,244,101,128,1,121,99,2,34,204,34,221,225,242,239,110,129,1,125,34,213,243,237,225,236,108,128,246,255,233,242,99,2,34,229,34,234,236,101,128,36,207,245,237,230,236,229,120,128,30,144,228,239,116,130,1,123,34,253,35,6,225,227,227,229,238,116,128,1,123,226,229,236,239,119,128,30,146,101,3,35,22,35,33,35,76,227,249,242,233,236,236,233,99,128,4,23,100,2,35,39,35,58,229,243,227,229,238,228,229,242,227,249,242,233,236,236,233,99,128,4,152,233,229,242,229,243,233,243,227,249,242,233,236,236,233,99,128,4,222,244,97,128,3,150,232,101,4,35,92,35,103,35,119,35,130,225,242,237,229,238,233,225,110,128,5,58,226,242,229,246,229,227,249,242,233,236,236,233,99,128,4,193,227,249,242,233,236,236,233,99,128,4,22,100,2,35,136,35,155,229,243,227,229,238,228,229,242,227,249,242,233,236,236,233,99,128,4,150,233,229,242,229,243,233,243,227,249,242,233,236,236,233,99,128,4,220,236,233,238,229,226,229,236,239,119,128,30,148,237,239,238,239,243,240,225,227,101,128,255,58,115,2,35,203,35,210,237,225,236,108,128,247,122,244,242,239,235,101,128,1,181,97,158,0,97,36,26,38,154,39,4,39,68,39,132,39,196,40,4,40,68,40,126,40,190,41,70,41,217,42,137,42,237,43,17,49,192,49,229,50,0,50,225,51,7,52,96,52,168,53,123,53,132,54,5,56,13,57,3,57,50,57,201,57,215,49,138,39,1,36,50,36,114,36,154,36,218,37,26,37,90,37,154,37,218,38,26,38,90,48,138,39,33,36,74,36,78,36,82,36,86,36,90,36,94,36,98,36,102,36,106,36,110,48,128,39,94,49,128,39,97,50,128,39,98,51,128,39,99,52,128,39,100,53,128,39,16,54,128,39,101,55,128,39,102,56,128,39,103,57,128,38,96,49,134,38,27,36,130,36,134,36,138,36,142,36,146,36,150,48,128,38,101,49,128,38,102,50,128,38,99,55,128,39,9,56,128,39,8,57,128,39,7,50,138,38,30,36,178,36,182,36,186,36,190,36,194,36,198,36,202,36,206,36,210,36,214,48,128,36,96,49,128,36,97,50,128,36,98,51,128,36,99,52,128,36,100,53,128,36,101,54,128,36,102,55,128,36,103,56,128,36,104,57,128,36,105,51,138,39,12,36,242,36,246,36,250,36,254,37,2,37,6,37,10,37,14,37,18,37,22,48,128,39,118,49,128,39,119,50,128,39,120,51,128,39,121,52,128,39,122,53,128,39,123,54,128,39,124,55,128,39,125,56,128,39,126,57,128,39,127,52,138,39,13,37,50,37,54,37,58,37,62,37,66,37,70,37,74,37,78,37,82,37,86,48,128,39,128,49,128,39,129,50,128,39,130,51,128,39,131,52,128,39,132,53,128,39,133,54,128,39,134,55,128,39,135,56,128,39,136,57,128,39,137,53,138,39,14,37,114,37,118,37,122,37,126,37,130,37,134,37,138,37,142,37,146,37,150,48,128,39,138,49,128,39,139,50,128,39,140,51,128,39,141,52,128,39,142,53,128,39,143,54,128,39,144,55,128,39,145,56,128,39,146,57,128,39,147,54,138,39,15,37,178,37,182,37,186,37,190,37,194,37,198,37,202,37,206,37,210,37,214,48,128,39,148,49,128,33,146,50,128,39,163,51,128,33,148,52,128,33,149,53,128,39,153,54,128,39,155,55,128,39,156,56,128,39,157,57,128,39,158,55,138,39,17,37,242,37,246,37,250,37,254,38,2,38,6,38,10,38,14,38,18,38,22,48,128,39,159,49,128,39,160,50,128,39,161,51,128,39,162,52,128,39,164,53,128,39,165,54,128,39,166,55,128,39,167,56,128,39,168,57,128,39,169,56,138,39,18,38,50,38,54,38,58,38,62,38,66,38,70,38,74,38,78,38,82,38,86,48,128,39,171,49,128,39,173,50,128,39,175,51,128,39,178,52,128,39,179,53,128,39,181,54,128,39,184,55,128,39,186,56,128,39,187,57,128,39,188,57,138,39,19,38,114,38,118,38,122,38,126,38,130,38,134,38,138,38,142,38,146,38,150,48,128,39,189,49,128,39,190,50,128,39,154,51,128,39,170,52,128,39,182,53,128,39,185,54,128,39,152,55,128,39,180,56,128,39,183,57,128,39,172,50,138,39,2,38,178,38,224,38,228,38,232,38,236,38,240,38,244,38,248,38,252,39,0,48,135,39,20,38,196,38,200,38,204,38,208,38,212,38,216,38,220,48,128,39,174,49,128,39,177,50,128,39,3,51,128,39,80,52,128,39,82,53,128,39,110,54,128,39,112,49,128,39,21,50,128,39,22,51,128,39,23,52,128,39,24,53,128,39,25,54,128,39,26,55,128,39,27,56,128,39,28,57,128,39,34,51,138,39,4,39,28,39,32,39,36,39,40,39,44,39,48,39,52,39,56,39,60,39,64,48,128,39,35,49,128,39,36,50,128,39,37,51,128,39,38,52,128,39,39,53,128,38,5,54,128,39,41,55,128,39,42,56,128,39,43,57,128,39,44,52,138,38,14,39,92,39,96,39,100,39,104,39,108,39,112,39,116,39,120,39,124,39,128,48,128,39,45,49,128,39,46,50,128,39,47,51,128,39,48,52,128,39,49,53,128,39,50,54,128,39,51,55,128,39,52,56,128,39,53,57,128,39,54,53,138,39,6,39,156,39,160,39,164,39,168,39,172,39,176,39,180,39,184,39,188,39,192,48,128,39,55,49,128,39,56,50,128,39,57,51,128,39,58,52,128,39,59,53,128,39,60,54,128,39,61,55,128,39,62,56,128,39,63,57,128,39,64,54,138,39,29,39,220,39,224,39,228,39,232,39,236,39,240,39,244,39,248,39,252,40,0,48,128,39,65,49,128,39,66,50,128,39,67,51,128,39,68,52,128,39,69,53,128,39,70,54,128,39,71,55,128,39,72,56,128,39,73,57,128,39,74,55,138,39,30,40,28,40,32,40,36,40,40,40,44,40,48,40,52,40,56,40,60,40,64,48,128,39,75,49,128,37,207,50,128,39,77,51,128,37,160,52,128,39,79,53,128,39,81,54,128,37,178,55,128,37,188,56,128,37,198,57,128,39,86,56,137,39,31,40,90,40,94,40,98,40,102,40,106,40,110,40,114,40,118,40,122,49,128,37,215,50,128,39,88,51,128,39,89,52,128,39,90,53,128,39,111,54,128,39,113,55,128,39,114,56,128,39,115,57,128,39,104,57,138,39,32,40,150,40,154,40,158,40,162,40,166,40,170,40,174,40,178,40,182,40,186,48,128,39,105,49,128,39,108,50,128,39,109,51,128,39,106,52,128,39,107,53,128,39,116,54,128,39,117,55,128,39,91,56,128,39,92,57,128,39,93,97,7,40,206,40,216,40,223,40,230,40,255,41,15,41,26,226,229,238,231,225,236,105,128,9,134,227,245,244,101,128,0,225,228,229,246,97,128,9,6,231,117,2,40,237,40,246,234,225,242,225,244,105,128,10,134,242,237,245,235,232,105,128,10,6,237,225,244,242,225,231,245,242,237,245,235,232,105,128,10,62,242,245,243,241,245,225,242,101,128,51,3,246,239,247,229,236,243,233,231,110,3,41,42,41,52,41,59,226,229,238,231,225,236,105,128,9,190,228,229,246,97,128,9,62,231,245,234,225,242,225,244,105,128,10,190,98,4,41,80,41,121,41,130,41,140,226,242,229,246,233,225,244,233,239,110,2,41,95,41,110,237,225,242,235,225,242,237,229,238,233,225,110,128,5,95,243,233,231,238,228,229,246,97,128,9,112,229,238,231,225,236,105,128,9,133,239,240,239,237,239,230,111,128,49,26,242,229,246,101,134,1,3,41,159,41,167,41,178,41,189,41,197,41,209,225,227,245,244,101,128,30,175,227,249,242,233,236,236,233,99,128,4,209,228,239,244,226,229,236,239,119,128,30,183,231,242,225,246,101,128,30,177,232,239,239,235,225,226,239,246,101,128,30,179,244,233,236,228,101,128,30,181,99,4,41,227,41,234,42,57,42,127,225,242,239,110,128,1,206,233,242,99,2,41,242,41,247,236,101,128,36,208,245,237,230,236,229,120,133,0,226,42,10,42,18,42,29,42,37,42,49,225,227,245,244,101,128,30,165,228,239,244,226,229,236,239,119,128,30,173,231,242,225,246,101,128,30,167,232,239,239,235,225,226,239,246,101,128,30,169,244,233,236,228,101,128,30,171,245,244,101,133,0,180,42,73,42,84,42,101,42,108,42,117,226,229,236,239,247,227,237,98,128,3,23,99,2,42,90,42,95,237,98,128,3,1,239,237,98,128,3,1,228,229,246,97,128,9,84,236,239,247,237,239,100,128,2,207,244,239,238,229,227,237,98,128,3,65,249,242,233,236,236,233,99,128,4,48,100,5,42,149,42,159,42,173,42,179,42,213,226,236,231,242,225,246,101,128,2,1,228,225,235,231,245,242,237,245,235,232,105,128,10,113,229,246,97,128,9,5,233,229,242,229,243,233,115,130,0,228,42,193,42,204,227,249,242,233,236,236,233,99,128,4,211,237,225,227,242,239,110,128,1,223,239,116,2,42,220,42,228,226,229,236,239,119,128,30,161,237,225,227,242,239,110,128,1,225,101,131,0,230,42,247,42,255,43,8,225,227,245,244,101,128,1,253,235,239,242,229,225,110,128,49,80,237,225,227,242,239,110,128,1,227,230,233,105,6,43,33,43,53,45,246,45,252,46,11,49,111,48,2,43,39,43,46,176,178,176,56,128,32,21,184,185,180,49,128,32,164,177,48,3,43,62,45,86,45,221,48,9,43,82,43,102,43,164,43,226,44,32,44,94,44,156,44,218,45,24,49,3,43,90,43,94,43,98,55,128,4,16,56,128,4,17,57,128,4,18,50,10,43,124,43,128,43,132,43,136,43,140,43,144,43,148,43,152,43,156,43,160,48,128,4,19,49,128,4,20,50,128,4,21,51,128,4,1,52,128,4,22,53,128,4,23,54,128,4,24,55,128,4,25,56,128,4,26,57,128,4,27,51,10,43,186,43,190,43,194,43,198,43,202,43,206,43,210,43,214,43,218,43,222,48,128,4,28,49,128,4,29,50,128,4,30,51,128,4,31,52,128,4,32,53,128,4,33,54,128,4,34,55,128,4,35,56,128,4,36,57,128,4,37,52,10,43,248,43,252,44,0,44,4,44,8,44,12,44,16,44,20,44,24,44,28,48,128,4,38,49,128,4,39,50,128,4,40,51,128,4,41,52,128,4,42,53,128,4,43,54,128,4,44,55,128,4,45,56,128,4,46,57,128,4,47,53,10,44,54,44,58,44,62,44,66,44,70,44,74,44,78,44,82,44,86,44,90,48,128,4,144,49,128,4,2,50,128,4,3,51,128,4,4,52,128,4,5,53,128,4,6,54,128,4,7,55,128,4,8,56,128,4,9,57,128,4,10,54,10,44,116,44,120,44,124,44,128,44,132,44,136,44,140,44,144,44,148,44,152,48,128,4,11,49,128,4,12,50,128,4,14,51,128,246,196,52,128,246,197,53,128,4,48,54,128,4,49,55,128,4,50,56,128,4,51,57,128,4,52,55,10,44,178,44,182,44,186,44,190,44,194,44,198,44,202,44,206,44,210,44,214,48,128,4,53,49,128,4,81,50,128,4,54,51,128,4,55,52,128,4,56,53,128,4,57,54,128,4,58,55,128,4,59,56,128,4,60,57,128,4,61,56,10,44,240,44,244,44,248,44,252,45,0,45,4,45,8,45,12,45,16,45,20,48,128,4,62,49,128,4,63,50,128,4,64,51,128,4,65,52,128,4,66,53,128,4,67,54,128,4,68,55,128,4,69,56,128,4,70,57,128,4,71,57,10,45,46,45,50,45,54,45,58,45,62,45,66,45,70,45,74,45,78,45,82,48,128,4,72,49,128,4,73,50,128,4,74,51,128,4,75,52,128,4,76,53,128,4,77,54,128,4,78,55,128,4,79,56,128,4,145,57,128,4,82,49,4,45,96,45,158,45,163,45,189,48,10,45,118,45,122,45,126,45,130,45,134,45,138,45,142,45,146,45,150,45,154,48,128,4,83,49,128,4,84,50,128,4,85,51,128,4,86,52,128,4,87,53,128,4,88,54,128,4,89,55,128,4,90,56,128,4,91,57,128,4,92,177,48,128,4,94,52,4,45,173,45,177,45,181,45,185,53,128,4,15,54,128,4,98,55,128,4,114,56,128,4,116,57,5,45,201,45,205,45,209,45,213,45,217,50,128,246,198,51,128,4,95,52,128,4,99,53,128,4,115,54,128,4,117,56,2,45,227,45,241,51,2,45,233,45,237,49,128,246,199,50,128,246,200,180,54,128,4,217,178,185,57,128,32,14,179,48,2,46,3,46,7,48,128,32,15,49,128,32,13,181,55,7,46,28,46,98,47,163,47,240,48,197,49,34,49,105,51,2,46,34,46,48,56,2,46,40,46,44,49,128,6,106,56,128,6,12,57,8,46,66,46,70,46,74,46,78,46,82,46,86,46,90,46,94,50,128,6,96,51,128,6,97,52,128,6,98,53,128,6,99,54,128,6,100,55,128,6,101,56,128,6,102,57,128,6,103,52,7,46,114,46,146,46,208,47,14,47,46,47,102,47,158,48,5,46,126,46,130,46,134,46,138,46,142,48,128,6,104,49,128,6,105,51,128,6,27,55,128,6,31,57,128,6,33,49,10,46,168,46,172,46,176,46,180,46,184,46,188,46,192,46,196,46,200,46,204,48,128,6,34,49,128,6,35,50,128,6,36,51,128,6,37,52,128,6,38,53,128,6,39,54,128,6,40,55,128,6,41,56,128,6,42,57,128,6,43,50,10,46,230,46,234,46,238,46,242,46,246,46,250,46,254,47,2,47,6,47,10,48,128,6,44,49,128,6,45,50,128,6,46,51,128,6,47,52,128,6,48,53,128,6,49,54,128,6,50,55,128,6,51,56,128,6,52,57,128,6,53,51,5,47,26,47,30,47,34,47,38,47,42,48,128,6,54,49,128,6,55,50,128,6,56,51,128,6,57,52,128,6,58,52,9,47,66,47,70,47,74,47,78,47,82,47,86,47,90,47,94,47,98,48,128,6,64,49,128,6,65,50,128,6,66,51,128,6,67,52,128,6,68,53,128,6,69,54,128,6,70,56,128,6,72,57,128,6,73,53,9,47,122,47,126,47,130,47,134,47,138,47,142,47,146,47,150,47,154,48,128,6,74,49,128,6,75,50,128,6,76,51,128,6,77,52,128,6,78,53,128,6,79,54,128,6,80,55,128,6,81,56,128,6,82,183,48,128,6,71,53,3,47,171,47,203,47,235,48,5,47,183,47,187,47,191,47,195,47,199,53,128,6,164,54,128,6,126,55,128,6,134,56,128,6,152,57,128,6,175,49,5,47,215,47,219,47,223,47,227,47,231,49,128,6,121,50,128,6,136,51,128,6,145,52,128,6,186,57,128,6,210,179,52,128,6,213,54,7,48,0,48,5,48,10,48,15,48,53,48,115,48,177,179,54,128,32,170,180,53,128,5,190,181,56,128,5,195,54,6,48,29,48,33,48,37,48,41,48,45,48,49,52,128,5,208,53,128,5,209,54,128,5,210,55,128,5,211,56,128,5,212,57,128,5,213,55,10,48,75,48,79,48,83,48,87,48,91,48,95,48,99,48,103,48,107,48,111,48,128,5,214,49,128,5,215,50,128,5,216,51,128,5,217,52,128,5,218,53,128,5,219,54,128,5,220,55,128,5,221,56,128,5,222,57,128,5,223,56,10,48,137,48,141,48,145,48,149,48,153,48,157,48,161,48,165,48,169,48,173,48,128,5,224,49,128,5,225,50,128,5,226,51,128,5,227,52,128,5,228,53,128,5,229,54,128,5,230,55,128,5,231,56,128,5,232,57,128,5,233,57,3,48,185,48,189,48,193,48,128,5,234,52,128,251,42,53,128,251,43,55,4,48,207,48,221,48,241,48,246,48,2,48,213,48,217,48,128,251,75,53,128,251,31,49,3,48,229,48,233,48,237,54,128,5,240,55,128,5,241,56,128,5,242,178,51,128,251,53,57,7,49,6,49,10,49,14,49,18,49,22,49,26,49,30,51,128,5,180,52,128,5,181,53,128,5,182,54,128,5,187,55,128,5,184,56,128,5,183,57,128,5,176,56,3,49,42,49,86,49,91,48,7,49,58,49,62,49,66,49,70,49,74,49,78,49,82,48,128,5,178,49,128,5,177,50,128,5,179,51,128,5,194,52,128,5,193,54,128,5,185,55,128,5,188,179,57,128,5,189,52,2,49,97,49,101,49,128,5,191,50,128,5,192,185,178,57,128,2,188,54,3,49,119,49,178,49,185,49,4,49,129,49,145,49,151,49,172,50,2,49,135,49,140,180,56,128,33,5,184,57,128,33,19,179,181,50,128,33,22,181,55,3,49,160,49,164,49,168,51,128,32,44,52,128,32,45,53,128,32,46,182,182,52,128,32,12,179,177,182,55,128,6,109,180,185,179,55,128,2,189,103,2,49,198,49,205,242,225,246,101,128,0,224,117,2,49,211,49,220,234,225,242,225,244,105,128,10,133,242,237,245,235,232,105,128,10,5,104,2,49,235,49,245,233,242,225,231,225,238,97,128,48,66,239,239,235,225,226,239,246,101,128,30,163,105,7,50,16,50,41,50,48,50,60,50,85,50,101,50,181,98,2,50,22,50,31,229,238,231,225,236,105,128,9,144,239,240,239,237,239,230,111,128,49,30,228,229,246,97,128,9,16,229,227,249,242,233,236,236,233,99,128,4,213,231,117,2,50,67,50,76,234,225,242,225,244,105,128,10,144,242,237,245,235,232,105,128,10,16,237,225,244,242,225,231,245,242,237,245,235,232,105,128,10,72,110,5,50,113,50,122,50,136,50,152,50,167,225,242,225,226,233,99,128,6,57,230,233,238,225,236,225,242,225,226,233,99,128,254,202,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,203,237,229,228,233,225,236,225,242,225,226,233,99,128,254,204,246,229,242,244,229,228,226,242,229,246,101,128,2,3,246,239,247,229,236,243,233,231,110,3,50,197,50,207,50,214,226,229,238,231,225,236,105,128,9,200,228,229,246,97,128,9,72,231,245,234,225,242,225,244,105,128,10,200,107,2,50,231,50,255,225,244,225,235,225,238,97,129,48,162,50,243,232,225,236,230,247,233,228,244,104,128,255,113,239,242,229,225,110,128,49,79,108,3,51,15,52,71,52,80,101,2,51,21,52,66,102,136,5,208,51,41,51,50,51,65,51,79,51,168,51,182,52,37,52,51,225,242,225,226,233,99,128,6,39,228,225,231,229,243,232,232,229,226,242,229,119,128,251,48,230,233,238,225,236,225,242,225,226,233,99,128,254,142,104,2,51,85,51,160,225,237,250,97,2,51,94,51,127,225,226,239,246,101,2,51,104,51,113,225,242,225,226,233,99,128,6,35,230,233,238,225,236,225,242,225,226,233,99,128,254,132,226,229,236,239,119,2,51,137,51,146,225,242,225,226,233,99,128,6,37,230,233,238,225,236,225,242,225,226,233,99,128,254,136,229,226,242,229,119,128,5,208,236,225,237,229,228,232,229,226,242,229,119,128,251,79,237,97,2,51,189,51,225,228,228,225,225,226,239,246,101,2,51,202,51,211,225,242,225,226,233,99,128,6,34,230,233,238,225,236,225,242,225,226,233,99,128,254,130,235,243,245,242,97,4,51,239,51,248,52,6,52,22,225,242,225,226,233,99,128,6,73,230,233,238,225,236,225,242,225,226,233,99,128,254,240,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,243,237,229,228,233,225,236,225,242,225,226,233,99,128,254,244,240,225,244,225,232,232,229,226,242,229,119,128,251,46,241,225,237,225,244,243,232,229,226,242,229,119,128,251,47,240,104,128,33,53,236,229,241,245,225,108,128,34,76,240,232,97,129,3,177,52,88,244,239,238,239,115,128,3,172,109,4,52,106,52,114,52,125,52,159,225,227,242,239,110,128,1,1,239,238,239,243,240,225,227,101,128,255,65,240,229,242,243,225,238,100,130,0,38,52,139,52,151,237,239,238,239,243,240,225,227,101,128,255,6,243,237,225,236,108,128,247,38,243,241,245,225,242,101,128,51,194,110,4,52,178,52,189,53,55,53,65,226,239,240,239,237,239,230,111,128,49,34,103,4,52,199,52,210,52,224,53,47,226,239,240,239,237,239,230,111,128,49,36,235,232,225,238,235,232,245,244,232,225,105,128,14,90,236,101,131,34,32,52,235,53,32,53,39,226,242,225,227,235,229,116,2,52,247,53,11,236,229,230,116,129,48,8,53,0,246,229,242,244,233,227,225,108,128,254,63,242,233,231,232,116,129,48,9,53,21,246,229,242,244,233,227,225,108,128,254,64,236,229,230,116,128,35,41,242,233,231,232,116,128,35,42,243,244,242,239,109,128,33,43,239,244,229,236,229,233,97,128,3,135,117,2,53,71,53,83,228,225,244,244,225,228,229,246,97,128,9,82,243,246,225,242,97,3,53,95,53,105,53,112,226,229,238,231,225,236,105,128,9,130,228,229,246,97,128,9,2,231,245,234,225,242,225,244,105,128,10,130,239,231,239,238,229,107,128,1,5,112,3,53,140,53,164,53,194,97,2,53,146,53,158,225,244,239,243,241,245,225,242,101,128,51,0,242,229,110,128,36,156,239,243,244,242,239,240,232,101,2,53,177,53,188,225,242,237,229,238,233,225,110,128,5,90,237,239,100,128,2,188,112,2,53,200,53,205,236,101,128,248,255,242,111,2,53,212,53,220,225,227,232,229,115,128,34,80,120,2,53,226,53,246,229,241,245,225,108,129,34,72,53,236,239,242,233,237,225,231,101,128,34,82,233,237,225,244,229,236,249,229,241,245,225,108,128,34,69,114,4,54,15,54,42,54,46,54,91,225,229,97,2,54,23,54,33,229,235,239,242,229,225,110,128,49,142,235,239,242,229,225,110,128,49,141,99,128,35,18,105,2,54,52,54,66,231,232,244,232,225,236,230,242,233,238,103,128,30,154,238,103,130,0,229,54,75,54,83,225,227,245,244,101,128,1,251,226,229,236,239,119,128,30,1,242,239,119,8,54,111,54,118,54,247,55,57,55,107,55,162,55,185,56,4,226,239,244,104,128,33,148,100,3,54,126,54,165,54,212,225,243,104,4,54,138,54,145,54,152,54,160,228,239,247,110,128,33,227,236,229,230,116,128,33,224,242,233,231,232,116,128,33,226,245,112,128,33,225,226,108,5,54,178,54,185,54,192,54,199,54,207,226,239,244,104,128,33,212,228,239,247,110,128,33,211,236,229,230,116,128,33,208,242,233,231,232,116,128,33,210,245,112,128,33,209,239,247,110,131,33,147,54,224,54,231,54,239,236,229,230,116,128,33,153,242,233,231,232,116,128,33,152,247,232,233,244,101,128,33,233,104,2,54,253,55,48,229,225,100,4,55,9,55,19,55,29,55,40,228,239,247,238,237,239,100,128,2,197,236,229,230,244,237,239,100,128,2,194,242,233,231,232,244,237,239,100,128,2,195,245,240,237,239,100,128,2,196,239,242,233,250,229,120,128,248,231,236,229,230,116,131,33,144,55,70,55,87,55,99,228,226,108,129,33,208,55,78,243,244,242,239,235,101,128,33,205,239,246,229,242,242,233,231,232,116,128,33,198,247,232,233,244,101,128,33,230,242,233,231,232,116,132,33,146,55,123,55,135,55,143,55,154,228,226,236,243,244,242,239,235,101,128,33,207,232,229,225,246,121,128,39,158,239,246,229,242,236,229,230,116,128,33,196,247,232,233,244,101,128,33,232,244,225,98,2,55,170,55,177,236,229,230,116,128,33,228,242,233,231,232,116,128,33,229,245,112,132,33,145,55,198,55,226,55,244,55,252,100,2,55,204,55,216,110,129,33,149,55,210,226,243,101,128,33,168,239,247,238,226,225,243,101,128,33,168,236,229,230,116,129,33,150,55,235,239,230,228,239,247,110,128,33,197,242,233,231,232,116,128,33,151,247,232,233,244,101,128,33,231,246,229,242,244,229,120,128,248,230,115,5,56,25,56,101,56,146,56,229,56,239,99,2,56,31,56,83,233,105,2,56,38,56,61,227,233,242,227,245,109,129,0,94,56,49,237,239,238,239,243,240,225,227,101,128,255,62,244,233,236,228,101,129,0,126,56,71,237,239,238,239,243,240,225,227,101,128,255,94,242,233,240,116,129,2,81,56,92,244,245,242,238,229,100,128,2,82,237,225,236,108,2,56,110,56,121,232,233,242,225,231,225,238,97,128,48,65,235,225,244,225,235,225,238,97,129,48,161,56,134,232,225,236,230,247,233,228,244,104,128,255,103,244,229,242,233,115,2,56,156,56,225,107,131,0,42,56,166,56,194,56,217,97,2,56,172,56,186,236,244,239,238,229,225,242,225,226,233,99,128,6,109,242,225,226,233,99,128,6,109,109,2,56,200,56,206,225,244,104,128,34,23,239,238,239,243,240,225,227,101,128,255,10,243,237,225,236,108,128,254,97,109,128,32,66,245,240,229,242,233,239,114,128,246,233,249,237,240,244,239,244,233,227,225,236,236,249,229,241,245,225,108,128,34,67,116,132,0,64,57,15,57,22,57,34,57,42,233,236,228,101,128,0,227,237,239,238,239,243,240,225,227,101,128,255,32,243,237,225,236,108,128,254,107,245,242,238,229,100,128,2,80,117,6,57,64,57,89,57,96,57,121,57,141,57,157,98,2,57,70,57,79,229,238,231,225,236,105,128,9,148,239,240,239,237,239,230,111,128,49,32,228,229,246,97,128,9,20,231,117,2,57,103,57,112,234,225,242,225,244,105,128,10,148,242,237,245,235,232,105,128,10,20,236,229,238,231,244,232,237,225,242,235,226,229,238,231,225,236,105,128,9,215,237,225,244,242,225,231,245,242,237,245,235,232,105,128,10,76,246,239,247,229,236,243,233,231,110,3,57,173,57,183,57,190,226,229,238,231,225,236,105,128,9,204,228,229,246,97,128,9,76,231,245,234,225,242,225,244,105,128,10,204,246,225,231,242,225,232,225,228,229,246,97,128,9,61,121,2,57,221,57,233,226,225,242,237,229,238,233,225,110,128,5,97,233,110,130,5,226,57,242,58,1,225,236,244,239,238,229,232,229,226,242,229,119,128,251,32,232,229,226,242,229,119,128,5,226,98,144,0,98,58,46,58,181,58,192,58,201,58,226,60,11,60,73,60,146,62,72,62,84,62,127,62,135,62,145,64,15,64,39,64,48,97,7,58,62,58,72,58,96,58,103,58,128,58,152,58,163,226,229,238,231,225,236,105,128,9,172,227,235,243,236,225,243,104,129,0,92,58,84,237,239,238,239,243,240,225,227,101,128,255,60,228,229,246,97,128,9,44,231,117,2,58,110,58,119,234,225,242,225,244,105,128,10,172,242,237,245,235,232,105,128,10,44,104,2,58,134,58,144,233,242,225,231,225,238,97,128,48,112,244,244,232,225,105,128,14,63,235,225,244,225,235,225,238,97,128,48,208,114,129,0,124,58,169,237,239,238,239,243,240,225,227,101,128,255,92,226,239,240,239,237,239,230,111,128,49,5,227,233,242,227,236,101,128,36,209,228,239,116,2,58,209,58,218,225,227,227,229,238,116,128,30,3,226,229,236,239,119,128,30,5,101,6,58,240,59,5,59,28,59,170,59,181,59,193,225,237,229,228,243,233,248,244,229,229,238,244,232,238,239,244,229,115,128,38,108,99,2,59,11,59,18,225,245,243,101,128,34,53,249,242,233,236,236,233,99,128,4,49,104,5,59,40,59,49,59,63,59,93,59,152,225,242,225,226,233,99,128,6,40,230,233,238,225,236,225,242,225,226,233,99,128,254,144,105,2,59,69,59,84,238,233,244,233,225,236,225,242,225,226,233,99,128,254,145,242,225,231,225,238,97,128,48,121,237,101,2,59,100,59,113,228,233,225,236,225,242,225,226,233,99,128,254,146,229,237,105,2,59,121,59,136,238,233,244,233,225,236,225,242,225,226,233,99,128,252,159,243,239,236,225,244,229,228,225,242,225,226,233,99,128,252,8,238,239,239,238,230,233,238,225,236,225,242,225,226,233,99,128,252,109,235,225,244,225,235,225,238,97,128,48,217,238,225,242,237,229,238,233,225,110,128,5,98,116,132,5,209,59,205,59,225,59,245,59,254,97,129,3,178,59,211,243,249,237,226,239,236,231,242,229,229,107,128,3,208,228,225,231,229,243,104,129,251,49,59,236,232,229,226,242,229,119,128,251,49,232,229,226,242,229,119,128,5,209,242,225,230,229,232,229,226,242,229,119,128,251,76,104,2,60,17,60,67,97,3,60,25,60,35,60,42,226,229,238,231,225,236,105,128,9,173,228,229,246,97,128,9,45,231,117,2,60,49,60,58,234,225,242,225,244,105,128,10,173,242,237,245,235,232,105,128,10,45,239,239,107,128,2,83,105,5,60,85,60,96,60,107,60,121,60,135,232,233,242,225,231,225,238,97,128,48,115,235,225,244,225,235,225,238,97,128,48,211,236,225,226,233,225,236,227,236,233,227,107,128,2,152,238,228,233,231,245,242,237,245,235,232,105,128,10,2,242,245,243,241,245,225,242,101,128,51,49,108,3,60,154,62,55,62,66,97,2,60,160,62,50,227,107,6,60,175,60,184,60,221,61,114,61,169,61,221,227,233,242,227,236,101,128,37,207,100,2,60,190,60,199,233,225,237,239,238,100,128,37,198,239,247,238,240,239,233,238,244,233,238,231,244,242,233,225,238,231,236,101,128,37,188,108,2,60,227,61,74,101,2,60,233,61,13,230,244,240,239,233,238,244,233,238,103,2,60,248,61,2,240,239,233,238,244,229,114,128,37,196,244,242,233,225,238,231,236,101,128,37,192,238,244,233,227,245,236,225,242,226,242,225,227,235,229,116,2,61,33,61,53,236,229,230,116,129,48,16,61,42,246,229,242,244,233,227,225,108,128,254,59,242,233,231,232,116,129,48,17,61,63,246,229,242,244,233,227,225,108,128,254,60,239,247,229,114,2,61,83,61,98,236,229,230,244,244,242,233,225,238,231,236,101,128,37,227,242,233,231,232,244,244,242,233,225,238,231,236,101,128,37,226,114,2,61,120,61,131,229,227,244,225,238,231,236,101,128,37,172,233,231,232,244,240,239,233,238,244,233,238,103,2,61,148,61,158,240,239,233,238,244,229,114,128,37,186,244,242,233,225,238,231,236,101,128,37,182,115,3,61,177,61,207,61,215,109,2,61,183,61,195,225,236,236,243,241,245,225,242,101,128,37,170,233,236,233,238,231,230,225,227,101,128,38,59,241,245,225,242,101,128,37,160,244,225,114,128,38,5,245,240,112,2,61,229,62,11,229,114,2,61,236,61,251,236,229,230,244,244,242,233,225,238,231,236,101,128,37,228,242,233,231,232,244,244,242,233,225,238,231,236,101,128,37,229,239,233,238,244,233,238,103,2,62,23,62,39,243,237,225,236,236,244,242,233,225,238,231,236,101,128,37,180,244,242,233,225,238,231,236,101,128,37,178,238,107,128,36,35,233,238,229,226,229,236,239,119,128,30,7,239,227,107,128,37,136,237,239,238,239,243,240,225,227,101,128,255,66,111,3,62,92,62,105,62,116,226,225,233,237,225,233,244,232,225,105,128,14,26,232,233,242,225,231,225,238,97,128,48,124,235,225,244,225,235,225,238,97,128,48,220,240,225,242,229,110,128,36,157,241,243,241,245,225,242,101,128,51,195,114,4,62,155,63,149,63,222,64,5,225,99,2,62,162,63,56,101,3,62,170,62,175,62,243,229,120,128,248,244,236,229,230,116,133,0,123,62,192,62,197,62,219,62,227,62,232,226,116,128,248,243,109,2,62,203,62,208,233,100,128,248,242,239,238,239,243,240,225,227,101,128,255,91,243,237,225,236,108,128,254,91,244,112,128,248,241,246,229,242,244,233,227,225,108,128,254,55,242,233,231,232,116,133,0,125,63,5,63,10,63,32,63,40,63,45,226,116,128,248,254,109,2,63,16,63,21,233,100,128,248,253,239,238,239,243,240,225,227,101,128,255,93,243,237,225,236,108,128,254,92,244,112,128,248,252,246,229,242,244,233,227,225,108,128,254,56,235,229,116,2,63,64,63,106,236,229,230,116,132,0,91,63,79,63,84,63,89,63,101,226,116,128,248,240,229,120,128,248,239,237,239,238,239,243,240,225,227,101,128,255,59,244,112,128,248,238,242,233,231,232,116,132,0,93,63,122,63,127,63,132,63,144,226,116,128,248,251,229,120,128,248,250,237,239,238,239,243,240,225,227,101,128,255,61,244,112,128,248,249,229,246,101,131,2,216,63,161,63,172,63,178,226,229,236,239,247,227,237,98,128,3,46,227,237,98,128,3,6,233,238,246,229,242,244,229,100,3,63,193,63,204,63,210,226,229,236,239,247,227,237,98,128,3,47,227,237,98,128,3,17,228,239,245,226,236,229,227,237,98,128,3,97,233,228,231,101,2,63,231,63,242,226,229,236,239,247,227,237,98,128,3,42,233,238,246,229,242,244,229,228,226,229,236,239,247,227,237,98,128,3,58,239,235,229,238,226,225,114,128,0,166,115,2,64,21,64,29,244,242,239,235,101,128,1,128,245,240,229,242,233,239,114,128,246,234,244,239,240,226,225,114,128,1,131,117,3,64,56,64,67,64,78,232,233,242,225,231,225,238,97,128,48,118,235,225,244,225,235,225,238,97,128,48,214,236,108,2,64,85,64,115,229,116,130,32,34,64,94,64,104,233,238,246,229,242,243,101,128,37,216,239,240,229,242,225,244,239,114,128,34,25,243,229,249,101,128,37,206,99,143,0,99,64,156,65,105,65,116,65,180,65,211,66,48,67,215,68,199,69,43,69,92,72,84,72,92,72,102,72,114,72,147,97,9,64,176,64,187,64,197,64,204,64,211,64,236,64,246,65,42,65,51,225,242,237,229,238,233,225,110,128,5,110,226,229,238,231,225,236,105,128,9,154,227,245,244,101,128,1,7,228,229,246,97,128,9,26,231,117,2,64,218,64,227,234,225,242,225,244,105,128,10,154,242,237,245,235,232,105,128,10,26,236,243,241,245,225,242,101,128,51,136,238,228,242,225,226,233,238,228,117,4,65,8,65,18,65,24,65,31,226,229,238,231,225,236,105,128,9,129,227,237,98,128,3,16,228,229,246,97,128,9,1,231,245,234,225,242,225,244,105,128,10,129,240,243,236,239,227,107,128,33,234,114,3,65,59,65,65,65,91,229,239,102,128,33,5,239,110,130,2,199,65,74,65,85,226,229,236,239,247,227,237,98,128,3,44,227,237,98,128,3,12,242,233,225,231,229,242,229,244,245,242,110,128,33,181,226,239,240,239,237,239,230,111,128,49,24,99,4,65,126,65,133,65,152,65,174,225,242,239,110,128,1,13,229,228,233,236,236,97,129,0,231,65,144,225,227,245,244,101,128,30,9,233,242,99,2,65,160,65,165,236,101,128,36,210,245,237,230,236,229,120,128,1,9,245,242,108,128,2,85,100,2,65,186,65,202,239,116,129,1,11,65,193,225,227,227,229,238,116,128,1,11,243,241,245,225,242,101,128,51,197,101,2,65,217,65,233,228,233,236,236,97,129,0,184,65,227,227,237,98,128,3,39,238,116,132,0,162,65,246,66,14,66,26,66,37,105,2,65,252,66,4,231,242,225,228,101,128,33,3,238,230,229,242,233,239,114,128,246,223,237,239,238,239,243,240,225,227,101,128,255,224,239,236,228,243,244,249,236,101,128,247,162,243,245,240,229,242,233,239,114,128,246,224,104,5,66,60,66,123,66,134,67,62,67,154,97,4,66,70,66,81,66,91,66,98,225,242,237,229,238,233,225,110,128,5,121,226,229,238,231,225,236,105,128,9,155,228,229,246,97,128,9,27,231,117,2,66,105,66,114,234,225,242,225,244,105,128,10,155,242,237,245,235,232,105,128,10,27,226,239,240,239,237,239,230,111,128,49,20,101,6,66,148,66,168,66,192,67,4,67,16,67,37,225,226,235,232,225,243,233,225,238,227,249,242,233,236,236,233,99,128,4,189,99,2,66,174,66,182,235,237,225,242,107,128,39,19,249,242,233,236,236,233,99,128,4,71,100,2,66,198,66,242,229,243,227,229,238,228,229,114,2,66,211,66,231,225,226,235,232,225,243,233,225,238,227,249,242,233,236,236,233,99,128,4,191,227,249,242,233,236,236,233,99,128,4,183,233,229,242,229,243,233,243,227,249,242,233,236,236,233,99,128,4,245,232,225,242,237,229,238,233,225,110,128,5,115,235,232,225,235,225,243,243,233,225,238,227,249,242,233,236,236,233,99,128,4,204,246,229,242,244,233,227,225,236,243,244,242,239,235,229,227,249,242,233,236,236,233,99,128,4,185,105,129,3,199,67,68,229,245,227,104,4,67,81,67,116,67,131,67,140,97,2,67,87,67,102,227,233,242,227,236,229,235,239,242,229,225,110,128,50,119,240,225,242,229,238,235,239,242,229,225,110,128,50,23,227,233,242,227,236,229,235,239,242,229,225,110,128,50,105,235,239,242,229,225,110,128,49,74,240,225,242,229,238,235,239,242,229,225,110,128,50,9,111,2,67,160,67,210,227,104,3,67,169,67,191,67,201,225,110,2,67,176,67,184,231,244,232,225,105,128,14,10,244,232,225,105,128,14,8,233,238,231,244,232,225,105,128,14,9,239,229,244,232,225,105,128,14,12,239,107,128,1,136,105,2,67,221,68,67,229,245,99,5,67,235,68,14,68,29,68,38,68,52,97,2,67,241,68,0,227,233,242,227,236,229,235,239,242,229,225,110,128,50,118,240,225,242,229,238,235,239,242,229,225,110,128,50,22,227,233,242,227,236,229,235,239,242,229,225,110,128,50,104,235,239,242,229,225,110,128,49,72,240,225,242,229,238,235,239,242,229,225,110,128,50,8,245,240,225,242,229,238,235,239,242,229,225,110,128,50,28,242,99,2,68,74,68,169,236,101,132,37,203,68,87,68,98,68,103,68,127,237,245,236,244,233,240,236,121,128,34,151,239,116,128,34,153,112,2,68,109,68,115,236,245,115,128,34,149,239,243,244,225,236,237,225,242,107,128,48,54,247,233,244,104,2,68,136,68,152,236,229,230,244,232,225,236,230,226,236,225,227,107,128,37,208,242,233,231,232,244,232,225,236,230,226,236,225,227,107,128,37,209,245,237,230,236,229,120,130,2,198,68,182,68,193,226,229,236,239,247,227,237,98,128,3,45,227,237,98,128,3,2,108,3,68,207,68,213,69,11,229,225,114,128,35,39,233,227,107,4,68,225,68,236,68,245,68,255,225,236,246,229,239,236,225,114,128,1,194,228,229,238,244,225,108,128,1,192,236,225,244,229,242,225,108,128,1,193,242,229,244,242,239,230,236,229,120,128,1,195,245,98,129,38,99,69,18,243,245,233,116,2,69,27,69,35,226,236,225,227,107,128,38,99,247,232,233,244,101,128,38,103,109,3,69,51,69,65,69,76,227,245,226,229,228,243,241,245,225,242,101,128,51,164,239,238,239,243,240,225,227,101,128,255,67,243,241,245,225,242,229,228,243,241,245,225,242,101,128,51,160,111,8,69,110,69,121,69,208,70,150,71,179,71,210,72,61,72,70,225,242,237,229,238,233,225,110,128,5,129,236,239,110,131,0,58,69,133,69,158,69,177,237,239,110,2,69,141,69,149,229,244,225,242,121,128,32,161,239,243,240,225,227,101,128,255,26,115,2,69,164,69,170,233,231,110,128,32,161,237,225,236,108,128,254,85,244,242,233,225,238,231,245,236,225,114,2,69,192,69,202,232,225,236,230,237,239,100,128,2,209,237,239,100,128,2,208,109,2,69,214,70,143,237,97,134,0,44,69,231,70,39,70,50,70,62,70,92,70,115,97,3,69,239,70,9,70,17,226,239,246,101,2,69,248,69,254,227,237,98,128,3,19,242,233,231,232,244,227,237,98,128,3,21,227,227,229,238,116,128,246,195,114,2,70,23,70,30,225,226,233,99,128,6,12,237,229,238,233,225,110,128,5,93,233,238,230,229,242,233,239,114,128,246,225,237,239,238,239,243,240,225,227,101,128,255,12,242,229,246,229,242,243,229,100,2,70,75,70,86,225,226,239,246,229,227,237,98,128,3,20,237,239,100,128,2,189,115,2,70,98,70,105,237,225,236,108,128,254,80,245,240,229,242,233,239,114,128,246,226,244,245,242,238,229,100,2,70,126,70,137,225,226,239,246,229,227,237,98,128,3,18,237,239,100,128,2,187,240,225,243,115,128,38,60,110,2,70,156,70,165,231,242,245,229,238,116,128,34,69,116,2,70,171,70,185,239,245,242,233,238,244,229,231,242,225,108,128,34,46,242,239,108,142,35,3,70,219,70,225,70,240,70,255,71,43,71,88,71,102,71,107,71,112,71,117,71,123,71,128,71,169,71,174,193,195,75,128,0,6,66,2,70,231,70,236,197,76,128,0,7,83,128,0,8,67,2,70,246,70,251,193,78,128,0,24,82,128,0,13,68,3,71,7,71,33,71,38,67,4,71,17,71,21,71,25,71,29,49,128,0,17,50,128,0,18,51,128,0,19,52,128,0,20,197,76,128,0,127,204,69,128,0,16,69,5,71,55,71,59,71,64,71,69,71,74,77,128,0,25,206,81,128,0,5,207,84,128,0,4,211,67,128,0,27,84,2,71,80,71,84,66,128,0,23,88,128,0,3,70,2,71,94,71,98,70,128,0,12,83,128,0,28,199,83,128,0,29,200,84,128,0,9,204,70,128,0,10,206,193,75,128,0,21,210,83,128,0,30,83,5,71,140,71,144,71,154,71,159,71,164,73,128,0,15,79,129,0,14,71,150,84,128,0,2,212,88,128,0,1,213,66,128,0,26,217,78,128,0,22,213,83,128,0,31,214,84,128,0,11,240,249,242,233,231,232,116,129,0,169,71,191,115,2,71,197,71,203,225,238,115,128,248,233,229,242,233,102,128,246,217,114,2,71,216,72,44,238,229,242,226,242,225,227,235,229,116,2,71,231,72,9,236,229,230,116,130,48,12,71,242,71,254,232,225,236,230,247,233,228,244,104,128,255,98,246,229,242,244,233,227,225,108,128,254,65,242,233,231,232,116,130,48,13,72,21,72,33,232,225,236,230,247,233,228,244,104,128,255,99,246,229,242,244,233,227,225,108,128,254,66,240,239,242,225,244,233,239,238,243,241,245,225,242,101,128,51,127,243,241,245,225,242,101,128,51,199,246,229,242,235,231,243,241,245,225,242,101,128,51,198,240,225,242,229,110,128,36,158,242,245,250,229,233,242,111,128,32,162,243,244,242,229,244,227,232,229,100,128,2,151,245,114,2,72,121,72,139,236,121,2,72,128,72,134,225,238,100,128,34,207,239,114,128,34,206,242,229,238,227,121,128,0,164,249,114,4,72,158,72,166,72,173,72,181,194,242,229,246,101,128,246,209,198,236,229,120,128,246,210,226,242,229,246,101,128,246,212,230,236,229,120,128,246,213,100,146,0,100,72,228,74,110,75,134,75,194,76,114,77,68,77,130,78,59,78,72,78,81,78,107,78,132,78,141,79,208,79,216,79,227,79,247,80,19,97,11,72,252,73,7,73,17,73,89,73,152,73,163,73,174,73,243,74,49,74,55,74,85,225,242,237,229,238,233,225,110,128,5,100,226,229,238,231,225,236,105,128,9,166,100,5,73,29,73,38,73,44,73,58,73,74,225,242,225,226,233,99,128,6,54,229,246,97,128,9,38,230,233,238,225,236,225,242,225,226,233,99,128,254,190,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,191,237,229,228,233,225,236,225,242,225,226,233,99,128,254,192,103,3,73,97,73,114,73,128,229,243,104,129,5,188,73,105,232,229,226,242,229,119,128,5,188,231,229,114,129,32,32,73,122,228,226,108,128,32,33,117,2,73,134,73,143,234,225,242,225,244,105,128,10,166,242,237,245,235,232,105,128,10,38,232,233,242,225,231,225,238,97,128,48,96,235,225,244,225,235,225,238,97,128,48,192,108,3,73,182,73,191,73,229,225,242,225,226,233,99,128,6,47,229,116,130,5,211,73,200,73,220,228,225,231,229,243,104,129,251,51,73,211,232,229,226,242,229,119,128,251,51,232,229,226,242,229,119,128,5,211,230,233,238,225,236,225,242,225,226,233,99,128,254,170,237,237,97,3,73,253,74,6,74,18,225,242,225,226,233,99,128,6,79,236,239,247,225,242,225,226,233,99,128,6,79,244,225,238,97,2,74,27,74,41,236,244,239,238,229,225,242,225,226,233,99,128,6,76,242,225,226,233,99,128,6,76,238,228,97,128,9,100,242,231,97,2,74,63,74,72,232,229,226,242,229,119,128,5,167,236,229,230,244,232,229,226,242,229,119,128,5,167,243,233,225,240,238,229,245,237,225,244,225,227,249,242,233,236,236,233,227,227,237,98,128,4,133,98,3,74,118,75,115,75,125,108,9,74,138,74,146,75,3,75,11,75,27,75,38,75,56,75,70,75,81,199,242,225,246,101,128,246,211,97,2,74,152,74,209,238,231,236,229,226,242,225,227,235,229,116,2,74,168,74,188,236,229,230,116,129,48,10,74,177,246,229,242,244,233,227,225,108,128,254,61,242,233,231,232,116,129,48,11,74,198,246,229,242,244,233,227,225,108,128,254,62,114,2,74,215,74,236,227,232,233,238,246,229,242,244,229,228,226,229,236,239,247,227,237,98,128,3,43,242,239,119,2,74,244,74,251,236,229,230,116,128,33,212,242,233,231,232,116,128,33,210,228,225,238,228,97,128,9,101,231,242,225,246,101,129,246,214,75,21,227,237,98,128,3,15,233,238,244,229,231,242,225,108,128,34,44,236,239,247,236,233,238,101,129,32,23,75,50,227,237,98,128,3,51,239,246,229,242,236,233,238,229,227,237,98,128,3,63,240,242,233,237,229,237,239,100,128,2,186,246,229,242,244,233,227,225,108,2,75,94,75,100,226,225,114,128,32,22,236,233,238,229,225,226,239,246,229,227,237,98,128,3,14,239,240,239,237,239,230,111,128,49,9,243,241,245,225,242,101,128,51,200,99,4,75,144,75,151,75,160,75,187,225,242,239,110,128,1,15,229,228,233,236,236,97,128,30,17,233,242,99,2,75,168,75,173,236,101,128,36,211,245,237,230,236,229,248,226,229,236,239,119,128,30,19,242,239,225,116,128,1,17,100,4,75,204,76,29,76,39,76,90,97,4,75,214,75,224,75,231,76,0,226,229,238,231,225,236,105,128,9,161,228,229,246,97,128,9,33,231,117,2,75,238,75,247,234,225,242,225,244,105,128,10,161,242,237,245,235,232,105,128,10,33,108,2,76,6,76,15,225,242,225,226,233,99,128,6,136,230,233,238,225,236,225,242,225,226,233,99,128,251,137,228,232,225,228,229,246,97,128,9,92,232,97,3,76,48,76,58,76,65,226,229,238,231,225,236,105,128,9,162,228,229,246,97,128,9,34,231,117,2,76,72,76,81,234,225,242,225,244,105,128,10,162,242,237,245,235,232,105,128,10,34,239,116,2,76,97,76,106,225,227,227,229,238,116,128,30,11,226,229,236,239,119,128,30,13,101,8,76,132,76,185,76,192,76,217,76,227,76,238,77,27,77,63,99,2,76,138,76,175,233,237,225,236,243,229,240,225,242,225,244,239,114,2,76,156,76,165,225,242,225,226,233,99,128,6,107,240,229,242,243,233,225,110,128,6,107,249,242,233,236,236,233,99,128,4,52,231,242,229,101,128,0,176,232,105,2,76,199,76,208,232,229,226,242,229,119,128,5,173,242,225,231,225,238,97,128,48,103,233,227,239,240,244,233,99,128,3,239,235,225,244,225,235,225,238,97,128,48,199,108,2,76,244,77,11,229,244,101,2,76,252,77,3,236,229,230,116,128,35,43,242,233,231,232,116,128,35,38,244,97,129,3,180,77,18,244,245,242,238,229,100,128,1,141,238,239,237,233,238,225,244,239,242,237,233,238,245,243,239,238,229,238,245,237,229,242,225,244,239,242,226,229,238,231,225,236,105,128,9,248,250,104,128,2,164,104,2,77,74,77,124,97,3,77,82,77,92,77,99,226,229,238,231,225,236,105,128,9,167,228,229,246,97,128,9,39,231,117,2,77,106,77,115,234,225,242,225,244,105,128,10,167,242,237,245,235,232,105,128,10,39,239,239,107,128,2,87,105,6,77,144,77,193,77,253,78,8,78,19,78,29,97,2,77,150,77,172,236,249,244,233,235,225,244,239,238,239,115,129,3,133,77,166,227,237,98,128,3,68,237,239,238,100,129,38,102,77,181,243,245,233,244,247,232,233,244,101,128,38,98,229,242,229,243,233,115,133,0,168,77,212,77,220,77,231,77,237,77,245,225,227,245,244,101,128,246,215,226,229,236,239,247,227,237,98,128,3,36,227,237,98,128,3,8,231,242,225,246,101,128,246,216,244,239,238,239,115,128,3,133,232,233,242,225,231,225,238,97,128,48,98,235,225,244,225,235,225,238,97,128,48,194,244,244,239,237,225,242,107,128,48,3,246,105,2,78,36,78,47,228,101,129,0,247,78,43,115,128,34,35,243,233,239,238,243,236,225,243,104,128,34,21,234,229,227,249,242,233,236,236,233,99,128,4,82,235,243,232,225,228,101,128,37,147,108,2,78,87,78,98,233,238,229,226,229,236,239,119,128,30,15,243,241,245,225,242,101,128,51,151,109,2,78,113,78,121,225,227,242,239,110,128,1,17,239,238,239,243,240,225,227,101,128,255,68,238,226,236,239,227,107,128,37,132,111,10,78,163,78,175,78,185,78,196,78,207,79,23,79,28,79,39,79,154,79,180,227,232,225,228,225,244,232,225,105,128,14,14,228,229,235,244,232,225,105,128,14,20,232,233,242,225,231,225,238,97,128,48,105,235,225,244,225,235,225,238,97,128,48,201,236,236,225,114,132,0,36,78,222,78,233,78,245,79,0,233,238,230,229,242,233,239,114,128,246,227,237,239,238,239,243,240,225,227,101,128,255,4,239,236,228,243,244,249,236,101,128,247,36,115,2,79,6,79,13,237,225,236,108,128,254,105,245,240,229,242,233,239,114,128,246,228,238,103,128,32,171,242,245,243,241,245,225,242,101,128,51,38,116,6,79,53,79,70,79,92,79,103,79,135,79,142,225,227,227,229,238,116,129,2,217,79,64,227,237,98,128,3,7,226,229,236,239,247,99,2,79,81,79,86,237,98,128,3,35,239,237,98,128,3,35,235,225,244,225,235,225,238,97,128,48,251,236,229,243,115,2,79,112,79,116,105,128,1,49,106,129,246,190,79,122,243,244,242,239,235,229,232,239,239,107,128,2,132,237,225,244,104,128,34,197,244,229,228,227,233,242,227,236,101,128,37,204,245,226,236,229,249,239,228,240,225,244,225,104,129,251,31,79,171,232,229,226,242,229,119,128,251,31,247,238,244,225,227,107,2,79,191,79,202,226,229,236,239,247,227,237,98,128,3,30,237,239,100,128,2,213,240,225,242,229,110,128,36,159,243,245,240,229,242,233,239,114,128,246,235,116,2,79,233,79,239,225,233,108,128,2,86,239,240,226,225,114,128,1,140,117,2,79,253,80,8,232,233,242,225,231,225,238,97,128,48,101,235,225,244,225,235,225,238,97,128,48,197,122,132,1,243,80,31,80,40,80,59,80,96,225,236,244,239,238,101,128,2,163,99,2,80,46,80,53,225,242,239,110,128,1,198,245,242,108,128,2,165,101,2,80,65,80,85,225,226,235,232,225,243,233,225,238,227,249,242,233,236,236,233,99,128,4,225,227,249,242,233,236,236,233,99,128,4,85,232,229,227,249,242,233,236,236,233,99,128,4,95,101,151,0,101,80,159,80,178,80,212,81,186,81,248,82,25,82,37,82,60,82,113,83,225,84,27,84,129,84,245,85,124,85,199,85,230,86,36,86,89,87,24,87,157,87,177,87,221,88,56,97,2,80,165,80,172,227,245,244,101,128,0,233,242,244,104,128,38,65,98,3,80,186,80,195,80,205,229,238,231,225,236,105,128,9,143,239,240,239,237,239,230,111,128,49,28,242,229,246,101,128,1,21,99,5,80,224,81,41,81,55,81,87,81,176,97,2,80,230,81,35,238,228,242,97,3,80,241,80,248,81,3,228,229,246,97,128,9,13,231,245,234,225,242,225,244,105,128,10,141,246,239,247,229,236,243,233,231,110,2,81,17,81,24,228,229,246,97,128,9,69,231,245,234,225,242,225,244,105,128,10,197,242,239,110,128,1,27,229,228,233,236,236,225,226,242,229,246,101,128,30,29,104,2,81,61,81,72,225,242,237,229,238,233,225,110,128,5,101,249,233,247,238,225,242,237,229,238,233,225,110,128,5,135,233,242,99,2,81,95,81,100,236,101,128,36,212,245,237,230,236,229,120,134,0,234,81,121,81,129,81,137,81,148,81,156,81,168,225,227,245,244,101,128,30,191,226,229,236,239,119,128,30,25,228,239,244,226,229,236,239,119,128,30,199,231,242,225,246,101,128,30,193,232,239,239,235,225,226,239,246,101,128,30,195,244,233,236,228,101,128,30,197,249,242,233,236,236,233,99,128,4,84,100,4,81,196,81,206,81,212,81,222,226,236,231,242,225,246,101,128,2,5,229,246,97,128,9,15,233,229,242,229,243,233,115,128,0,235,239,116,130,1,23,81,231,81,240,225,227,227,229,238,116,128,1,23,226,229,236,239,119,128,30,185,101,2,81,254,82,9,231,245,242,237,245,235,232,105,128,10,15,237,225,244,242,225,231,245,242,237,245,235,232,105,128,10,71,230,227,249,242,233,236,236,233,99,128,4,68,103,2,82,43,82,50,242,225,246,101,128,0,232,245,234,225,242,225,244,105,128,10,143,104,4,82,70,82,81,82,92,82,102,225,242,237,229,238,233,225,110,128,5,103,226,239,240,239,237,239,230,111,128,49,29,233,242,225,231,225,238,97,128,48,72,239,239,235,225,226,239,246,101,128,30,187,105,4,82,123,82,134,83,192,83,207,226,239,240,239,237,239,230,111,128,49,31,231,232,116,142,0,56,82,168,82,177,82,187,82,217,82,224,83,6,83,31,83,76,83,110,83,122,83,133,83,166,83,174,83,185,225,242,225,226,233,99,128,6,104,226,229,238,231,225,236,105,128,9,238,227,233,242,227,236,101,129,36,103,82,198,233,238,246,229,242,243,229,243,225,238,243,243,229,242,233,102,128,39,145,228,229,246,97,128,9,110,229,229,110,2,82,232,82,241,227,233,242,227,236,101,128,36,113,112,2,82,247,82,254,225,242,229,110,128,36,133,229,242,233,239,100,128,36,153,231,117,2,83,13,83,22,234,225,242,225,244,105,128,10,238,242,237,245,235,232,105,128,10,110,104,2,83,37,83,63,97,2,83,43,83,54,227,235,225,242,225,226,233,99,128,6,104,238,231,250,232,239,117,128,48,40,238,239,244,229,226,229,225,237,229,100,128,38,107,105,2,83,82,83,100,228,229,239,231,242,225,240,232,233,227,240,225,242,229,110,128,50,39,238,230,229,242,233,239,114,128,32,136,237,239,238,239,243,240,225,227,101,128,255,24,239,236,228,243,244,249,236,101,128,247,56,112,2,83,139,83,146,225,242,229,110,128,36,123,229,114,2,83,153,83,159,233,239,100,128,36,143,243,233,225,110,128,6,248,242,239,237,225,110,128,33,119,243,245,240,229,242,233,239,114,128,32,120,244,232,225,105,128,14,88,238,246,229,242,244,229,228,226,242,229,246,101,128,2,7,239,244,233,230,233,229,228,227,249,242,233,236,236,233,99,128,4,101,107,2,83,231,83,255,225,244,225,235,225,238,97,129,48,168,83,243,232,225,236,230,247,233,228,244,104,128,255,116,111,2,84,5,84,20,238,235,225,242,231,245,242,237,245,235,232,105,128,10,116,242,229,225,110,128,49,84,108,3,84,35,84,46,84,107,227,249,242,233,236,236,233,99,128,4,59,101,2,84,52,84,59,237,229,238,116,128,34,8,246,229,110,3,84,69,84,78,84,99,227,233,242,227,236,101,128,36,106,112,2,84,84,84,91,225,242,229,110,128,36,126,229,242,233,239,100,128,36,146,242,239,237,225,110,128,33,122,236,233,240,243,233,115,129,32,38,84,118,246,229,242,244,233,227,225,108,128,34,238,109,5,84,141,84,169,84,180,84,200,84,211,225,227,242,239,110,130,1,19,84,153,84,161,225,227,245,244,101,128,30,23,231,242,225,246,101,128,30,21,227,249,242,233,236,236,233,99,128,4,60,228,225,243,104,129,32,20,84,189,246,229,242,244,233,227,225,108,128,254,49,239,238,239,243,240,225,227,101,128,255,69,112,2,84,217,84,237,232,225,243,233,243,237,225,242,235,225,242,237,229,238,233,225,110,128,5,91,244,249,243,229,116,128,34,5,110,6,85,3,85,14,85,25,85,69,85,101,85,116,226,239,240,239,237,239,230,111,128,49,35,227,249,242,233,236,236,233,99,128,4,61,100,2,85,31,85,50,225,243,104,129,32,19,85,39,246,229,242,244,233,227,225,108,128,254,50,229,243,227,229,238,228,229,242,227,249,242,233,236,236,233,99,128,4,163,103,130,1,75,85,77,85,88,226,239,240,239,237,239,230,111,128,49,37,232,229,227,249,242,233,236,236,233,99,128,4,165,232,239,239,235,227,249,242,233,236,236,233,99,128,4,200,243,240,225,227,101,128,32,2,111,3,85,132,85,140,85,149,231,239,238,229,107,128,1,25,235,239,242,229,225,110,128,49,83,240,229,110,130,2,91,85,159,85,168,227,236,239,243,229,100,128,2,154,242,229,246,229,242,243,229,100,130,2,92,85,183,85,192,227,236,239,243,229,100,128,2,94,232,239,239,107,128,2,93,112,2,85,205,85,212,225,242,229,110,128,36,160,243,233,236,239,110,129,3,181,85,222,244,239,238,239,115,128,3,173,241,117,2,85,237,86,25,225,108,130,0,61,85,246,86,2,237,239,238,239,243,240,225,227,101,128,255,29,115,2,86,8,86,15,237,225,236,108,128,254,102,245,240,229,242,233,239,114,128,32,124,233,246,225,236,229,238,227,101,128,34,97,114,3,86,44,86,55,86,66,226,239,240,239,237,239,230,111,128,49,38,227,249,242,233,236,236,233,99,128,4,64,229,246,229,242,243,229,100,129,2,88,86,78,227,249,242,233,236,236,233,99,128,4,77,115,6,86,103,86,114,86,134,86,215,87,4,87,14,227,249,242,233,236,236,233,99,128,4,65,228,229,243,227,229,238,228,229,242,227,249,242,233,236,236,233,99,128,4,171,104,132,2,131,86,146,86,153,86,184,86,199,227,245,242,108,128,2,134,239,242,116,2,86,161,86,168,228,229,246,97,128,9,14,246,239,247,229,236,243,233,231,238,228,229,246,97,128,9,70,242,229,246,229,242,243,229,228,236,239,239,112,128,1,170,243,241,245,225,244,242,229,246,229,242,243,229,100,128,2,133,237,225,236,108,2,86,224,86,235,232,233,242,225,231,225,238,97,128,48,71,235,225,244,225,235,225,238,97,129,48,167,86,248,232,225,236,230,247,233,228,244,104,128,255,106,244,233,237,225,244,229,100,128,33,46,245,240,229,242,233,239,114,128,246,236,116,5,87,36,87,62,87,66,87,83,87,149,97,130,3,183,87,44,87,54,242,237,229,238,233,225,110,128,5,104,244,239,238,239,115,128,3,174,104,128,0,240,233,236,228,101,129,30,189,87,75,226,229,236,239,119,128,30,27,238,225,232,244,97,3,87,95,87,127,87,136,230,239,245,235,104,2,87,105,87,114,232,229,226,242,229,119,128,5,145,236,229,230,244,232,229,226,242,229,119,128,5,145,232,229,226,242,229,119,128,5,145,236,229,230,244,232,229,226,242,229,119,128,5,145,245,242,238,229,100,128,1,221,117,2,87,163,87,172,235,239,242,229,225,110,128,49,97,242,111,128,32,172,246,239,247,229,236,243,233,231,110,3,87,193,87,203,87,210,226,229,238,231,225,236,105,128,9,199,228,229,246,97,128,9,71,231,245,234,225,242,225,244,105,128,10,199,120,2,87,227,88,44,227,236,225,109,132,0,33,87,242,87,253,88,24,88,36,225,242,237,229,238,233,225,110,128,5,92,100,2,88,3,88,8,226,108,128,32,60,239,247,110,129,0,161,88,16,243,237,225,236,108,128,247,161,237,239,238,239,243,240,225,227,101,128,255,1,243,237,225,236,108,128,247,33,233,243,244,229,238,244,233,225,108,128,34,3,250,104,131,2,146,88,67,88,86,88,97,99,2,88,73,88,80,225,242,239,110,128,1,239,245,242,108,128,2,147,242,229,246,229,242,243,229,100,128,1,185,244,225,233,108,128,1,186,102,140,0,102,88,132,88,214,88,225,88,234,88,246,89,93,89,109,91,117,91,130,91,156,93,33,93,41,97,4,88,142,88,149,88,160,88,171,228,229,246,97,128,9,94,231,245,242,237,245,235,232,105,128,10,94,232,242,229,238,232,229,233,116,128,33,9,244,232,97,3,88,181,88,190,88,202,225,242,225,226,233,99,128,6,78,236,239,247,225,242,225,226,233,99,128,6,78,244,225,238,225,242,225,226,233,99,128,6,75,226,239,240,239,237,239,230,111,128,49,8,227,233,242,227,236,101,128,36,213,228,239,244,225,227,227,229,238,116,128,30,31,101,3,88,254,89,76,89,86,104,4,89,8,89,31,89,45,89,61,225,114,2,89,15,89,22,225,226,233,99,128,6,65,237,229,238,233,225,110,128,5,134,230,233,238,225,236,225,242,225,226,233,99,128,254,210,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,211,237,229,228,233,225,236,225,242,225,226,233,99,128,254,212,233,227,239,240,244,233,99,128,3,229,237,225,236,101,128,38,64,102,130,251,0,89,101,89,105,105,128,251,3,108,128,251,4,105,136,251,1,89,129,89,169,89,180,89,202,90,68,90,85,90,93,90,106,230,244,229,229,110,2,89,139,89,148,227,233,242,227,236,101,128,36,110,112,2,89,154,89,161,225,242,229,110,128,36,130,229,242,233,239,100,128,36,150,231,245,242,229,228,225,243,104,128,32,18,236,236,229,100,2,89,189,89,195,226,239,120,128,37,160,242,229,227,116,128,37,172,238,225,108,5,89,216,89,255,90,16,90,33,90,49,235,225,102,130,5,218,89,226,89,246,228,225,231,229,243,104,129,251,58,89,237,232,229,226,242,229,119,128,251,58,232,229,226,242,229,119,128,5,218,237,229,109,129,5,221,90,7,232,229,226,242,229,119,128,5,221,238,245,110,129,5,223,90,24,232,229,226,242,229,119,128,5,223,240,101,129,5,227,90,40,232,229,226,242,229,119,128,5,227,244,243,225,228,105,129,5,229,90,59,232,229,226,242,229,119,128,5,229,242,243,244,244,239,238,229,227,232,233,238,229,243,101,128,2,201,243,232,229,249,101,128,37,201,244,225,227,249,242,233,236,236,233,99,128,4,115,246,101,142,0,53,90,139,90,148,90,158,90,188,90,195,90,205,90,230,91,1,91,35,91,47,91,58,91,91,91,99,91,110,225,242,225,226,233,99,128,6,101,226,229,238,231,225,236,105,128,9,235,227,233,242,227,236,101,129,36,100,90,169,233,238,246,229,242,243,229,243,225,238,243,243,229,242,233,102,128,39,142,228,229,246,97,128,9,107,229,233,231,232,244,232,115,128,33,93,231,117,2,90,212,90,221,234,225,242,225,244,105,128,10,235,242,237,245,235,232,105,128,10,107,232,97,2,90,237,90,248,227,235,225,242,225,226,233,99,128,6,101,238,231,250,232,239,117,128,48,37,105,2,91,7,91,25,228,229,239,231,242,225,240,232,233,227,240,225,242,229,110,128,50,36,238,230,229,242,233,239,114,128,32,133,237,239,238,239,243,240,225,227,101,128,255,21,239,236,228,243,244,249,236,101,128,247,53,112,2,91,64,91,71,225,242,229,110,128,36,120,229,114,2,91,78,91,84,233,239,100,128,36,140,243,233,225,110,128,6,245,242,239,237,225,110,128,33,116,243,245,240,229,242,233,239,114,128,32,117,244,232,225,105,128,14,85,108,129,251,2,91,123,239,242,233,110,128,1,146,109,2,91,136,91,147,239,238,239,243,240,225,227,101,128,255,70,243,241,245,225,242,101,128,51,153,111,4,91,166,91,188,91,200,91,207,230,97,2,91,173,91,181,238,244,232,225,105,128,14,31,244,232,225,105,128,14,29,238,231,237,225,238,244,232,225,105,128,14,79,242,225,236,108,128,34,0,245,114,142,0,52,91,240,91,249,92,3,92,33,92,40,92,65,92,92,92,126,92,138,92,157,92,168,92,201,92,209,92,220,225,242,225,226,233,99,128,6,100,226,229,238,231,225,236,105,128,9,234,227,233,242,227,236,101,129,36,99,92,14,233,238,246,229,242,243,229,243,225,238,243,243,229,242,233,102,128,39,141,228,229,246,97,128,9,106,231,117,2,92,47,92,56,234,225,242,225,244,105,128,10,234,242,237,245,235,232,105,128,10,106,232,97,2,92,72,92,83,227,235,225,242,225,226,233,99,128,6,100,238,231,250,232,239,117,128,48,36,105,2,92,98,92,116,228,229,239,231,242,225,240,232,233,227,240,225,242,229,110,128,50,35,238,230,229,242,233,239,114,128,32,132,237,239,238,239,243,240,225,227,101,128,255,20,238,245,237,229,242,225,244,239,242,226,229,238,231,225,236,105,128,9,247,239,236,228,243,244,249,236,101,128,247,52,112,2,92,174,92,181,225,242,229,110,128,36,119,229,114,2,92,188,92,194,233,239,100,128,36,139,243,233,225,110,128,6,244,242,239,237,225,110,128,33,115,243,245,240,229,242,233,239,114,128,32,116,116,2,92,226,93,8,229,229,110,2,92,234,92,243,227,233,242,227,236,101,128,36,109,112,2,92,249,93,0,225,242,229,110,128,36,129,229,242,233,239,100,128,36,149,104,2,93,14,93,19,225,105,128,14,84,244,239,238,229,227,232,233,238,229,243,101,128,2,203,240,225,242,229,110,128,36,161,242,97,2,93,48,93,56,227,244,233,239,110,128,32,68,238,99,128,32,163,103,144,0,103,93,97,94,43,94,66,94,127,94,144,95,65,96,58,96,143,96,156,97,14,97,39,97,67,97,89,98,34,98,56,98,158,97,9,93,117,93,127,93,134,93,141,93,205,93,230,93,241,93,252,94,30,226,229,238,231,225,236,105,128,9,151,227,245,244,101,128,1,245,228,229,246,97,128,9,23,102,4,93,151,93,160,93,174,93,190,225,242,225,226,233,99,128,6,175,230,233,238,225,236,225,242,225,226,233,99,128,251,147,233,238,233,244,233,225,236,225,242,225,226,233,99,128,251,148,237,229,228,233,225,236,225,242,225,226,233,99,128,251,149,231,117,2,93,212,93,221,234,225,242,225,244,105,128,10,151,242,237,245,235,232,105,128,10,23,232,233,242,225,231,225,238,97,128,48,76,235,225,244,225,235,225,238,97,128,48,172,237,237,97,130,3,179,94,6,94,19,236,225,244,233,238,243,237,225,236,108,128,2,99,243,245,240,229,242,233,239,114,128,2,224,238,231,233,225,227,239,240,244,233,99,128,3,235,98,2,94,49,94,59,239,240,239,237,239,230,111,128,49,13,242,229,246,101,128,1,31,99,4,94,76,94,83,94,92,94,114,225,242,239,110,128,1,231,229,228,233,236,236,97,128,1,35,233,242,99,2,94,100,94,105,236,101,128,36,214,245,237,230,236,229,120,128,1,29,239,237,237,225,225,227,227,229,238,116,128,1,35,228,239,116,129,1,33,94,135,225,227,227,229,238,116,128,1,33,101,6,94,158,94,169,94,180,94,191,94,210,95,56,227,249,242,233,236,236,233,99,128,4,51,232,233,242,225,231,225,238,97,128,48,82,235,225,244,225,235,225,238,97,128,48,178,239,237,229,244,242,233,227,225,236,236,249,229,241,245,225,108,128,34,81,114,3,94,218,95,11,95,21,229,243,104,3,94,228,94,243,94,252,225,227,227,229,238,244,232,229,226,242,229,119,128,5,156,232,229,226,242,229,119,128,5,243,237,245,241,228,225,237,232,229,226,242,229,119,128,5,157,237,225,238,228,226,236,115,128,0,223,243,232,225,249,233,109,2,95,32,95,47,225,227,227,229,238,244,232,229,226,242,229,119,128,5,158,232,229,226,242,229,119,128,5,244,244,225,237,225,242,107,128,48,19,104,5,95,77,95,210,96,17,96,42,96,48,97,4,95,87,95,97,95,120,95,145,226,229,238,231,225,236,105,128,9,152,100,2,95,103,95,114,225,242,237,229,238,233,225,110,128,5,114,229,246,97,128,9,24,231,117,2,95,127,95,136,234,225,242,225,244,105,128,10,152,242,237,245,235,232,105,128,10,24,233,110,4,95,156,95,165,95,179,95,195,225,242,225,226,233,99,128,6,58,230,233,238,225,236,225,242,225,226,233,99,128,254,206,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,207,237,229,228,233,225,236,225,242,225,226,233,99,128,254,208,101,3,95,218,95,239,96,0,237,233,228,228,236,229,232,239,239,235,227,249,242,233,236,236,233,99,128,4,149,243,244,242,239,235,229,227,249,242,233,236,236,233,99,128,4,147,245,240,244,245,242,238,227,249,242,233,236,236,233,99,128,4,145,232,97,2,96,24,96,31,228,229,246,97,128,9,90,231,245,242,237,245,235,232,105,128,10,90,239,239,107,128,2,96,250,243,241,245,225,242,101,128,51,147,105,3,96,66,96,77,96,88,232,233,242,225,231,225,238,97,128,48,78,235,225,244,225,235,225,238,97,128,48,174,109,2,96,94,96,105,225,242,237,229,238,233,225,110,128,5,99,229,108,130,5,210,96,114,96,134,228,225,231,229,243,104,129,251,50,96,125,232,229,226,242,229,119,128,251,50,232,229,226,242,229,119,128,5,210,234,229,227,249,242,233,236,236,233,99,128,4,83,236,239,244,244,225,108,2,96,167,96,184,233,238,246,229,242,244,229,228,243,244,242,239,235,101,128,1,190,243,244,239,112,132,2,148,96,199,96,210,96,216,96,248,233,238,246,229,242,244,229,100,128,2,150,237,239,100,128,2,192,242,229,246,229,242,243,229,100,130,2,149,96,231,96,237,237,239,100,128,2,193,243,245,240,229,242,233,239,114,128,2,228,243,244,242,239,235,101,129,2,161,97,3,242,229,246,229,242,243,229,100,128,2,162,109,2,97,20,97,28,225,227,242,239,110,128,30,33,239,238,239,243,240,225,227,101,128,255,71,111,2,97,45,97,56,232,233,242,225,231,225,238,97,128,48,84,235,225,244,225,235,225,238,97,128,48,180,240,97,2,97,74,97,80,242,229,110,128,36,162,243,241,245,225,242,101,128,51,172,114,2,97,95,97,192,97,2,97,101,97,109,228,233,229,238,116,128,34,7,246,101,134,0,96,97,126,97,137,97,154,97,161,97,170,97,182,226,229,236,239,247,227,237,98,128,3,22,99,2,97,143,97,148,237,98,128,3,0,239,237,98,128,3,0,228,229,246,97,128,9,83,236,239,247,237,239,100,128,2,206,237,239,238,239,243,240,225,227,101,128,255,64,244,239,238,229,227,237,98,128,3,64,229,225,244,229,114,132,0,62,97,208,97,227,97,239,98,26,229,241,245,225,108,129,34,101,97,218,239,242,236,229,243,115,128,34,219,237,239,238,239,243,240,225,227,101,128,255,30,111,2,97,245,98,15,114,2,97,251,98,8,229,241,245,233,246,225,236,229,238,116,128,34,115,236,229,243,115,128,34,119,246,229,242,229,241,245,225,108,128,34,103,243,237,225,236,108,128,254,101,115,2,98,40,98,48,227,242,233,240,116,128,2,97,244,242,239,235,101,128,1,229,117,4,98,66,98,77,98,134,98,145,232,233,242,225,231,225,238,97,128,48,80,233,108,2,98,84,98,109,236,229,237,239,116,2,98,94,98,101,236,229,230,116,128,0,171,242,233,231,232,116,128,0,187,243,233,238,231,108,2,98,119,98,126,236,229,230,116,128,32,57,242,233,231,232,116,128,32,58,235,225,244,225,235,225,238,97,128,48,176,242,225,237,245,243,241,245,225,242,101,128,51,24,249,243,241,245,225,242,101,128,51,201,104,144,0,104,98,204,101,90,101,125,101,162,101,202,103,90,103,110,104,75,104,87,104,99,105,167,105,175,105,186,105,195,106,19,106,23,97,13,98,232,99,15,99,25,99,55,99,80,99,158,99,170,99,195,99,210,99,239,99,252,100,54,100,63,97,2,98,238,99,1,226,235,232,225,243,233,225,238,227,249,242,233,236,236,233,99,128,4,169,236,244,239,238,229,225,242,225,226,233,99,128,6,193,226,229,238,231,225,236,105,128,9,185,228,101,2,99,32,99,50,243,227,229,238,228,229,242,227,249,242,233,236,236,233,99,128,4,179,246,97,128,9,57,231,117,2,99,62,99,71,234,225,242,225,244,105,128,10,185,242,237,245,235,232,105,128,10,57,104,4,99,90,99,99,99,113,99,143,225,242,225,226,233,99,128,6,45,230,233,238,225,236,225,242,225,226,233,99,128,254,162,105,2,99,119,99,134,238,233,244,233,225,236,225,242,225,226,233,99,128,254,163,242,225,231,225,238,97,128,48,111,237,229,228,233,225,236,225,242,225,226,233,99,128,254,164,233,244,245,243,241,245,225,242,101,128,51,42,235,225,244,225,235,225,238,97,129,48,207,99,183,232,225,236,230,247,233,228,244,104,128,255,138,236,225,238,244,231,245,242,237,245,235,232,105,128,10,77,237,250,97,2,99,218,99,227,225,242,225,226,233,99,128,6,33,236,239,247,225,242,225,226,233,99,128,6,33,238,231,245,236,230,233,236,236,229,114,128,49,100,114,2,100,2,100,18,228,243,233,231,238,227,249,242,233,236,236,233,99,128,4,74,240,239,239,110,2,100,27,100,40,236,229,230,244,226,225,242,226,245,112,128,33,188,242,233,231,232,244,226,225,242,226,245,112,128,33,192,243,241,245,225,242,101,128,51,202,244,225,102,3,100,73,100,165,101,0,240,225,244,225,104,134,5,178,100,93,100,98,100,112,100,121,100,136,100,152,177,54,128,5,178,50,2,100,104,100,108,51,128,5,178,102,128,5,178,232,229,226,242,229,119,128,5,178,238,225,242,242,239,247,232,229,226,242,229,119,128,5,178,241,245,225,242,244,229,242,232,229,226,242,229,119,128,5,178,247,233,228,229,232,229,226,242,229,119,128,5,178,241,225,237,225,244,115,135,5,179,100,188,100,193,100,198,100,203,100,212,100,227,100,243,177,98,128,5,179,178,56,128,5,179,179,52,128,5,179,232,229,226,242,229,119,128,5,179,238,225,242,242,239,247,232,229,226,242,229,119,128,5,179,241,245,225,242,244,229,242,232,229,226,242,229,119,128,5,179,247,233,228,229,232,229,226,242,229,119,128,5,179,243,229,231,239,108,135,5,177,101,22,101,27,101,32,101,37,101,46,101,61,101,77,177,55,128,5,177,178,52,128,5,177,179,48,128,5,177,232,229,226,242,229,119,128,5,177,238,225,242,242,239,247,232,229,226,242,229,119,128,5,177,241,245,225,242,244,229,242,232,229,226,242,229,119,128,5,177,247,233,228,229,232,229,226,242,229,119,128,5,177,98,3,101,98,101,103,101,113,225,114,128,1,39,239,240,239,237,239,230,111,128,49,15,242,229,246,229,226,229,236,239,119,128,30,43,99,2,101,131,101,140,229,228,233,236,236,97,128,30,41,233,242,99,2,101,148,101,153,236,101,128,36,215,245,237,230,236,229,120,128,1,37,100,2,101,168,101,178,233,229,242,229,243,233,115,128,30,39,239,116,2,101,185,101,194,225,227,227,229,238,116,128,30,35,226,229,236,239,119,128,30,37,101,136,5,212,101,222,101,255,102,19,102,248,103,8,103,53,103,62,103,75,225,242,116,129,38,101,101,230,243,245,233,116,2,101,239,101,247,226,236,225,227,107,128,38,101,247,232,233,244,101,128,38,97,228,225,231,229,243,104,129,251,52,102,10,232,229,226,242,229,119,128,251,52,104,6,102,33,102,61,102,69,102,119,102,165,102,214,97,2,102,39,102,53,236,244,239,238,229,225,242,225,226,233,99,128,6,193,242,225,226,233,99,128,6,71,229,226,242,229,119,128,5,212,230,233,238,225,236,97,2,102,80,102,111,236,116,2,102,87,102,99,239,238,229,225,242,225,226,233,99,128,251,167,244,247,239,225,242,225,226,233,99,128,254,234,242,225,226,233,99,128,254,234,232,225,237,250,225,225,226,239,246,101,2,102,134,102,148,230,233,238,225,236,225,242,225,226,233,99,128,251,165,233,243,239,236,225,244,229,228,225,242,225,226,233,99,128,251,164,105,2,102,171,102,205,238,233,244,233,225,236,97,2,102,183,102,197,236,244,239,238,229,225,242,225,226,233,99,128,251,168,242,225,226,233,99,128,254,235,242,225,231,225,238,97,128,48,120,237,229,228,233,225,236,97,2,102,226,102,240,236,244,239,238,229,225,242,225,226,233,99,128,251,169,242,225,226,233,99,128,254,236,233,243,229,233,229,242,225,243,241,245,225,242,101,128,51,123,107,2,103,14,103,38,225,244,225,235,225,238,97,129,48,216,103,26,232,225,236,230,247,233,228,244,104,128,255,141,245,244,225,225,242,245,243,241,245,225,242,101,128,51,54,238,231,232,239,239,107,128,2,103,242,245,244,245,243,241,245,225,242,101,128,51,57,116,129,5,215,103,81,232,229,226,242,229,119,128,5,215,232,239,239,107,129,2,102,103,99,243,245,240,229,242,233,239,114,128,2,177,105,4,103,120,103,205,103,216,103,241,229,245,104,4,103,132,103,167,103,182,103,191,97,2,103,138,103,153,227,233,242,227,236,229,235,239,242,229,225,110,128,50,123,240,225,242,229,238,235,239,242,229,225,110,128,50,27,227,233,242,227,236,229,235,239,242,229,225,110,128,50,109,235,239,242,229,225,110,128,49,78,240,225,242,229,238,235,239,242,229,225,110,128,50,13,232,233,242,225,231,225,238,97,128,48,114,235,225,244,225,235,225,238,97,129,48,210,103,229,232,225,236,230,247,233,228,244,104,128,255,139,242,233,113,134,5,180,104,3,104,8,104,22,104,31,104,46,104,62,177,52,128,5,180,50,2,104,14,104,18,49,128,5,180,100,128,5,180,232,229,226,242,229,119,128,5,180,238,225,242,242,239,247,232,229,226,242,229,119,128,5,180,241,245,225,242,244,229,242,232,229,226,242,229,119,128,5,180,247,233,228,229,232,229,226,242,229,119,128,5,180,236,233,238,229,226,229,236,239,119,128,30,150,237,239,238,239,243,240,225,227,101,128,255,72,111,9,104,119,104,130,104,154,104,179,105,11,105,24,105,110,105,150,105,161,225,242,237,229,238,233,225,110,128,5,112,232,105,2,104,137,104,145,240,244,232,225,105,128,14,43,242,225,231,225,238,97,128,48,123,235,225,244,225,235,225,238,97,129,48,219,104,167,232,225,236,230,247,233,228,244,104,128,255,142,236,225,109,135,5,185,104,199,104,204,104,209,104,214,104,223,104,238,104,254,177,57,128,5,185,178,54,128,5,185,179,50,128,5,185,232,229,226,242,229,119,128,5,185,238,225,242,242,239,247,232,229,226,242,229,119,128,5,185,241,245,225,242,244,229,242,232,229,226,242,229,119,128,5,185,247,233,228,229,232,229,226,242,229,119,128,5,185,238,239,235,232,245,235,244,232,225,105,128,14,46,111,2,105,30,105,100,107,4,105,40,105,52,105,58,105,80,225,226,239,246,229,227,239,237,98,128,3,9,227,237,98,128,3,9,240,225,236,225,244,225,236,233,250,229,228,226,229,236,239,247,227,237,98,128,3,33,242,229,244,242,239,230,236,229,248,226,229,236,239,247,227,237,98,128,3,34,238,243,241,245,225,242,101,128,51,66,114,2,105,116,105,143,105,2,105,122,105,131,227,239,240,244,233,99,128,3,233,250,239,238,244,225,236,226,225,114,128,32,21,238,227,237,98,128,3,27,244,243,240,242,233,238,231,115,128,38,104,245,243,101,128,35,2,240,225,242,229,110,128,36,163,243,245,240,229,242,233,239,114,128,2,176,244,245,242,238,229,100,128,2,101,117,4,105,205,105,216,105,229,105,254,232,233,242,225,231,225,238,97,128,48,117,233,233,244,239,243,241,245,225,242,101,128,51,51,235,225,244,225,235,225,238,97,129,48,213,105,242,232,225,236,230,247,233,228,244,104,128,255,140,238,231,225,242,245,237,236,225,245,116,129,2,221,106,13,227,237,98,128,3,11,118,128,1,149,249,240,232,229,110,132,0,45,106,39,106,50,106,62,106,85,233,238,230,229,242,233,239,114,128,246,229,237,239,238,239,243,240,225,227,101,128,255,13,115,2,106,68,106,75,237,225,236,108,128,254,99,245,240,229,242,233,239,114,128,246,230,244,247,111,128,32,16,105,149,0,105,106,137,106,160,106,194,106,241,110,123,110,243,111,24,111,51,111,213,111,217,111,255,112,21,112,105,113,14,113,89,113,97,113,110,113,197,113,254,114,26,114,70,225,99,2,106,144,106,150,245,244,101,128,0,237,249,242,233,236,236,233,99,128,4,79,98,3,106,168,106,177,106,187,229,238,231,225,236,105,128,9,135,239,240,239,237,239,230,111,128,49,39,242,229,246,101,128,1,45,99,3,106,202,106,209,106,231,225,242,239,110,128,1,208,233,242,99,2,106,217,106,222,236,101,128,36,216,245,237,230,236,229,120,128,0,238,249,242,233,236,236,233,99,128,4,86,100,4,106,251,107,5,110,80,110,113,226,236,231,242,225,246,101,128,2,9,101,2,107,11,110,75,239,231,242,225,240,104,7,107,32,107,46,107,59,109,244,110,19,110,32,110,44,229,225,242,244,232,227,233,242,227,236,101,128,50,143,230,233,242,229,227,233,242,227,236,101,128,50,139,233,99,14,107,90,107,106,107,205,108,3,108,69,108,98,108,114,108,171,108,220,108,232,109,3,109,70,109,208,109,237,225,236,236,233,225,238,227,229,240,225,242,229,110,128,50,63,99,4,107,116,107,127,107,141,107,148,225,236,236,240,225,242,229,110,128,50,58,229,238,244,242,229,227,233,242,227,236,101,128,50,165,236,239,243,101,128,48,6,111,3,107,156,107,171,107,191,237,237,97,129,48,1,107,164,236,229,230,116,128,255,100,238,231,242,225,244,245,236,225,244,233,239,238,240,225,242,229,110,128,50,55,242,242,229,227,244,227,233,242,227,236,101,128,50,163,101,3,107,213,107,225,107,242,225,242,244,232,240,225,242,229,110,128,50,47,238,244,229,242,240,242,233,243,229,240,225,242,229,110,128,50,61,248,227,229,236,236,229,238,244,227,233,242,227,236,101,128,50,157,102,2,108,9,108,24,229,243,244,233,246,225,236,240,225,242,229,110,128,50,64,105,2,108,30,108,59,238,225,238,227,233,225,108,2,108,42,108,51,227,233,242,227,236,101,128,50,150,240,225,242,229,110,128,50,54,242,229,240,225,242,229,110,128,50,43,104,2,108,75,108,86,225,246,229,240,225,242,229,110,128,50,50,233,231,232,227,233,242,227,236,101,128,50,164,233,244,229,242,225,244,233,239,238,237,225,242,107,128,48,5,108,3,108,122,108,148,108,160,225,226,239,114,2,108,131,108,140,227,233,242,227,236,101,128,50,152,240,225,242,229,110,128,50,56,229,230,244,227,233,242,227,236,101,128,50,167,239,247,227,233,242,227,236,101,128,50,166,109,2,108,177,108,209,101,2,108,183,108,198,228,233,227,233,238,229,227,233,242,227,236,101,128,50,169,244,225,236,240,225,242,229,110,128,50,46,239,239,238,240,225,242,229,110,128,50,42,238,225,237,229,240,225,242,229,110,128,50,52,112,2,108,238,108,246,229,242,233,239,100,128,48,2,242,233,238,244,227,233,242,227,236,101,128,50,158,114,2,109,9,109,57,101,3,109,17,109,28,109,43,225,227,232,240,225,242,229,110,128,50,67,240,242,229,243,229,238,244,240,225,242,229,110,128,50,57,243,239,245,242,227,229,240,225,242,229,110,128,50,62,233,231,232,244,227,233,242,227,236,101,128,50,168,115,5,109,82,109,111,109,125,109,150,109,178,101,2,109,88,109,101,227,242,229,244,227,233,242,227,236,101,128,50,153,236,230,240,225,242,229,110,128,50,66,239,227,233,229,244,249,240,225,242,229,110,128,50,51,112,2,109,131,109,137,225,227,101,128,48,0,229,227,233,225,236,240,225,242,229,110,128,50,53,116,2,109,156,109,167,239,227,235,240,225,242,229,110,128,50,49,245,228,249,240,225,242,229,110,128,50,59,117,2,109,184,109,193,238,240,225,242,229,110,128,50,48,240,229,242,246,233,243,229,240,225,242,229,110,128,50,60,119,2,109,214,109,226,225,244,229,242,240,225,242,229,110,128,50,44,239,239,228,240,225,242,229,110,128,50,45,250,229,242,111,128,48,7,109,2,109,250,110,7,229,244,225,236,227,233,242,227,236,101,128,50,142,239,239,238,227,233,242,227,236,101,128,50,138,238,225,237,229,227,233,242,227,236,101,128,50,148,243,245,238,227,233,242,227,236,101,128,50,144,119,2,110,50,110,63,225,244,229,242,227,233,242,227,236,101,128,50,140,239,239,228,227,233,242,227,236,101,128,50,141,246,97,128,9,7,233,229,242,229,243,233,115,130,0,239,110,94,110,102,225,227,245,244,101,128,30,47,227,249,242,233,236,236,233,99,128,4,229,239,244,226,229,236,239,119,128,30,203,101,3,110,131,110,147,110,158,226,242,229,246,229,227,249,242,233,236,236,233,99,128,4,215,227,249,242,233,236,236,233,99,128,4,53,245,238,103,4,110,170,110,205,110,220,110,229,97,2,110,176,110,191,227,233,242,227,236,229,235,239,242,229,225,110,128,50,117,240,225,242,229,238,235,239,242,229,225,110,128,50,21,227,233,242,227,236,229,235,239,242,229,225,110,128,50,103,235,239,242,229,225,110,128,49,71,240,225,242,229,238,235,239,242,229,225,110,128,50,7,103,2,110,249,111,0,242,225,246,101,128,0,236,117,2,111,6,111,15,234,225,242,225,244,105,128,10,135,242,237,245,235,232,105,128,10,7,104,2,111,30,111,40,233,242,225,231,225,238,97,128,48,68,239,239,235,225,226,239,246,101,128,30,201,105,8,111,69,111,79,111,90,111,97,111,122,111,138,111,153,111,169,226,229,238,231,225,236,105,128,9,136,227,249,242,233,236,236,233,99,128,4,56,228,229,246,97,128,9,8,231,117,2,111,104,111,113,234,225,242,225,244,105,128,10,136,242,237,245,235,232,105,128,10,8,237,225,244,242,225,231,245,242,237,245,235,232,105,128,10,64,238,246,229,242,244,229,228,226,242,229,246,101,128,2,11,243,232,239,242,244,227,249,242,233,236,236,233,99,128,4,57,246,239,247,229,236,243,233,231,110,3,111,185,111,195,111,202,226,229,238,231,225,236,105,128,9,192,228,229,246,97,128,9,64,231,245,234,225,242,225,244,105,128,10,192,106,128,1,51,107,2,111,223,111,247,225,244,225,235,225,238,97,129,48,164,111,235,232,225,236,230,247,233,228,244,104,128,255,114,239,242,229,225,110,128,49,99,108,2,112,5,112,10,228,101,128,2,220,245,249,232,229,226,242,229,119,128,5,172,109,2,112,27,112,94,97,3,112,35,112,55,112,80,227,242,239,110,129,1,43,112,44,227,249,242,233,236,236,233,99,128,4,227,231,229,239,242,225,240,240,242,239,248,233,237,225,244,229,236,249,229,241,245,225,108,128,34,83,244,242,225,231,245,242,237,245,235,232,105,128,10,63,239,238,239,243,240,225,227,101,128,255,73,110,5,112,117,112,127,112,136,112,148,112,232,227,242,229,237,229,238,116,128,34,6,230,233,238,233,244,121,128,34,30,233,225,242,237,229,238,233,225,110,128,5,107,116,2,112,154,112,222,101,2,112,160,112,211,231,242,225,108,131,34,43,112,173,112,191,112,196,98,2,112,179,112,187,239,244,244,239,109,128,35,33,116,128,35,33,229,120,128,248,245,116,2,112,202,112,207,239,112,128,35,32,112,128,35,32,242,243,229,227,244,233,239,110,128,34,41,233,243,241,245,225,242,101,128,51,5,118,3,112,240,112,249,113,2,226,245,236,236,229,116,128,37,216,227,233,242,227,236,101,128,37,217,243,237,233,236,229,230,225,227,101,128,38,59,111,3,113,22,113,33,113,41,227,249,242,233,236,236,233,99,128,4,81,231,239,238,229,107,128,1,47,244,97,131,3,185,113,52,113,73,113,81,228,233,229,242,229,243,233,115,129,3,202,113,65,244,239,238,239,115,128,3,144,236,225,244,233,110,128,2,105,244,239,238,239,115,128,3,175,240,225,242,229,110,128,36,164,242,233,231,245,242,237,245,235,232,105,128,10,114,115,4,113,120,113,165,113,179,113,187,237,225,236,108,2,113,129,113,140,232,233,242,225,231,225,238,97,128,48,67,235,225,244,225,235,225,238,97,129,48,163,113,153,232,225,236,230,247,233,228,244,104,128,255,104,243,232,225,242,226,229,238,231,225,236,105,128,9,250,244,242,239,235,101,128,2,104,245,240,229,242,233,239,114,128,246,237,116,2,113,203,113,237,229,242,225,244,233,239,110,2,113,215,113,226,232,233,242,225,231,225,238,97,128,48,157,235,225,244,225,235,225,238,97,128,48,253,233,236,228,101,129,1,41,113,246,226,229,236,239,119,128,30,45,117,2,114,4,114,15,226,239,240,239,237,239,230,111,128,49,41,227,249,242,233,236,236,233,99,128,4,78,246,239,247,229,236,243,233,231,110,3,114,42,114,52,114,59,226,229,238,231,225,236,105,128,9,191,228,229,246,97,128,9,63,231,245,234,225,242,225,244,105,128,10,191,250,232,233,244,243,97,2,114,81,114,92,227,249,242,233,236,236,233,99,128,4,117,228,226,236,231,242,225,246,229,227,249,242,233,236,236,233,99,128,4,119,106,138,0,106,114,135,114,198,114,209,115,3,115,19,115,132,115,201,115,206,115,218,115,226,97,4,114,145,114,156,114,166,114,173,225,242,237,229,238,233,225,110,128,5,113,226,229,238,231,225,236,105,128,9,156,228,229,246,97,128,9,28,231,117,2,114,180,114,189,234,225,242,225,244,105,128,10,156,242,237,245,235,232,105,128,10,28,226,239,240,239,237,239,230,111,128,49,16,99,3,114,217,114,224,114,246,225,242,239,110,128,1,240,233,242,99,2,114,232,114,237,236,101,128,36,217,245,237,230,236,229,120,128,1,53,242,239,243,243,229,228,244,225,233,108,128,2,157,228,239,244,236,229,243,243,243,244,242,239,235,101,128,2,95,101,3,115,27,115,38,115,103,227,249,242,233,236,236,233,99,128,4,88,229,109,4,115,49,115,58,115,72,115,88,225,242,225,226,233,99,128,6,44,230,233,238,225,236,225,242,225,226,233,99,128,254,158,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,159,237,229,228,233,225,236,225,242,225,226,233,99,128,254,160,104,2,115,109,115,118,225,242,225,226,233,99,128,6,152,230,233,238,225,236,225,242,225,226,233,99,128,251,139,104,2,115,138,115,188,97,3,115,146,115,156,115,163,226,229,238,231,225,236,105,128,9,157,228,229,246,97,128,9,29,231,117,2,115,170,115,179,234,225,242,225,244,105,128,10,157,242,237,245,235,232,105,128,10,29,229,232,225,242,237,229,238,233,225,110,128,5,123,233,115,128,48,4,237,239,238,239,243,240,225,227,101,128,255,74,240,225,242,229,110,128,36,165,243,245,240,229,242,233,239,114,128,2,178,107,146,0,107,116,21,118,110,118,121,118,183,118,194,119,28,119,42,120,150,121,90,121,103,121,129,121,178,122,60,122,82,122,95,122,118,122,160,122,170,97,12,116,47,116,79,116,101,116,131,116,245,117,14,117,44,117,69,117,175,117,189,118,56,118,85,98,2,116,53,116,70,225,243,232,235,233,242,227,249,242,233,236,236,233,99,128,4,161,229,238,231,225,236,105,128,9,149,99,2,116,85,116,91,245,244,101,128,30,49,249,242,233,236,236,233,99,128,4,58,228,101,2,116,108,116,126,243,227,229,238,228,229,242,227,249,242,233,236,236,233,99,128,4,155,246,97,128,9,21,102,135,5,219,116,149,116,158,116,178,116,192,116,201,116,217,116,232,225,242,225,226,233,99,128,6,67,228,225,231,229,243,104,129,251,59,116,169,232,229,226,242,229,119,128,251,59,230,233,238,225,236,225,242,225,226,233,99,128,254,218,232,229,226,242,229,119,128,5,219,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,219,237,229,228,233,225,236,225,242,225,226,233,99,128,254,220,242,225,230,229,232,229,226,242,229,119,128,251,77,231,117,2,116,252,117,5,234,225,242,225,244,105,128,10,149,242,237,245,235,232,105,128,10,21,104,2,117,20,117,30,233,242,225,231,225,238,97,128,48,75,239,239,235,227,249,242,233,236,236,233,99,128,4,196,235,225,244,225,235,225,238,97,129,48,171,117,57,232,225,236,230,247,233,228,244,104,128,255,118,112,2,117,75,117,96,240,97,129,3,186,117,82,243,249,237,226,239,236,231,242,229,229,107,128,3,240,249,229,239,245,110,3,117,108,117,122,117,156,237,233,229,245,237,235,239,242,229,225,110,128,49,113,112,2,117,128,117,143,232,233,229,245,240,232,235,239,242,229,225,110,128,49,132,233,229,245,240,235,239,242,229,225,110,128,49,120,243,243,225,238,231,240,233,229,245,240,235,239,242,229,225,110,128,49,121,242,239,242,233,233,243,241,245,225,242,101,128,51,13,115,5,117,201,117,245,118,4,118,12,118,40,232,233,228,225,225,245,244,111,2,117,214,117,223,225,242,225,226,233,99,128,6,64,238,239,243,233,228,229,226,229,225,242,233,238,231,225,242,225,226,233,99,128,6,64,237,225,236,236,235,225,244,225,235,225,238,97,128,48,245,241,245,225,242,101,128,51,132,242,97,2,118,19,118,28,225,242,225,226,233,99,128,6,80,244,225,238,225,242,225,226,233,99,128,6,77,244,242,239,235,229,227,249,242,233,236,236,233,99,128,4,159,244,225,232,233,242,225,240,242,239,236,239,238,231,237,225,242,235,232,225,236,230,247,233,228,244,104,128,255,112,246,229,242,244,233,227,225,236,243,244,242,239,235,229,227,249,242,233,236,236,233,99,128,4,157,226,239,240,239,237,239,230,111,128,49,14,99,4,118,131,118,153,118,162,118,170,97,2,118,137,118,147,236,243,241,245,225,242,101,128,51,137,242,239,110,128,1,233,229,228,233,236,236,97,128,1,55,233,242,227,236,101,128,36,218,239,237,237,225,225,227,227,229,238,116,128,1,55,228,239,244,226,229,236,239,119,128,30,51,101,4,118,204,118,231,119,0,119,12,104,2,118,210,118,221,225,242,237,229,238,233,225,110,128,5,132,233,242,225,231,225,238,97,128,48,81,235,225,244,225,235,225,238,97,129,48,177,118,244,232,225,236,230,247,233,228,244,104,128,255,121,238,225,242,237,229,238,233,225,110,128,5,111,243,237,225,236,236,235,225,244,225,235,225,238,97,128,48,246,231,242,229,229,238,236,225,238,228,233,99,128,1,56,104,6,119,56,119,185,119,196,119,221,120,52,120,140,97,5,119,68,119,78,119,89,119,96,119,121,226,229,238,231,225,236,105,128,9,150,227,249,242,233,236,236,233,99,128,4,69,228,229,246,97,128,9,22,231,117,2,119,103,119,112,234,225,242,225,244,105,128,10,150,242,237,245,235,232,105,128,10,22,104,4,119,131,119,140,119,154,119,170,225,242,225,226,233,99,128,6,46,230,233,238,225,236,225,242,225,226,233,99,128,254,166,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,167,237,229,228,233,225,236,225,242,225,226,233,99,128,254,168,229,233,227,239,240,244,233,99,128,3,231,232,97,2,119,203,119,210,228,229,246,97,128,9,89,231,245,242,237,245,235,232,105,128,10,89,233,229,245,235,104,4,119,235,120,14,120,29,120,38,97,2,119,241,120,0,227,233,242,227,236,229,235,239,242,229,225,110,128,50,120,240,225,242,229,238,235,239,242,229,225,110,128,50,24,227,233,242,227,236,229,235,239,242,229,225,110,128,50,106,235,239,242,229,225,110,128,49,75,240,225,242,229,238,235,239,242,229,225,110,128,50,10,111,4,120,62,120,111,120,121,120,126,235,104,4,120,73,120,82,120,91,120,101,225,233,244,232,225,105,128,14,2,239,238,244,232,225,105,128,14,5,245,225,244,244,232,225,105,128,14,3,247,225,233,244,232,225,105,128,14,4,237,245,244,244,232,225,105,128,14,91,239,107,128,1,153,242,225,235,232,225,238,231,244,232,225,105,128,14,6,250,243,241,245,225,242,101,128,51,145,105,4,120,160,120,171,120,196,120,245,232,233,242,225,231,225,238,97,128,48,77,235,225,244,225,235,225,238,97,129,48,173,120,184,232,225,236,230,247,233,228,244,104,128,255,119,242,111,3,120,205,120,220,120,236,231,245,242,225,237,245,243,241,245,225,242,101,128,51,21,237,229,229,244,239,242,245,243,241,245,225,242,101,128,51,22,243,241,245,225,242,101,128,51,20,249,229,239,107,5,121,4,121,39,121,54,121,63,121,77,97,2,121,10,121,25,227,233,242,227,236,229,235,239,242,229,225,110,128,50,110,240,225,242,229,238,235,239,242,229,225,110,128,50,14,227,233,242,227,236,229,235,239,242,229,225,110,128,50,96,235,239,242,229,225,110,128,49,49,240,225,242,229,238,235,239,242,229,225,110,128,50,0,243,233,239,243,235,239,242,229,225,110,128,49,51,234,229,227,249,242,233,236,236,233,99,128,4,92,108,2,121,109,121,120,233,238,229,226,229,236,239,119,128,30,53,243,241,245,225,242,101,128,51,152,109,3,121,137,121,151,121,162,227,245,226,229,228,243,241,245,225,242,101,128,51,166,239,238,239,243,240,225,227,101,128,255,75,243,241,245,225,242,229,228,243,241,245,225,242,101,128,51,162,111,5,121,190,121,216,121,254,122,10,122,24,104,2,121,196,121,206,233,242,225,231,225,238,97,128,48,83,237,243,241,245,225,242,101,128,51,192,235,97,2,121,223,121,231,233,244,232,225,105,128,14,1,244,225,235,225,238,97,129,48,179,121,242,232,225,236,230,247,233,228,244,104,128,255,122,239,240,239,243,241,245,225,242,101,128,51,30,240,240,225,227,249,242,233,236,236,233,99,128,4,129,114,2,122,30,122,50,229,225,238,243,244,225,238,228,225,242,228,243,249,237,226,239,108,128,50,127,239,238,233,243,227,237,98,128,3,67,240,97,2,122,67,122,73,242,229,110,128,36,166,243,241,245,225,242,101,128,51,170,243,233,227,249,242,233,236,236,233,99,128,4,111,116,2,122,101,122,110,243,241,245,225,242,101,128,51,207,245,242,238,229,100,128,2,158,117,2,122,124,122,135,232,233,242,225,231,225,238,97,128,48,79,235,225,244,225,235,225,238,97,129,48,175,122,148,232,225,236,230,247,233,228,244,104,128,255,120,246,243,241,245,225,242,101,128,51,184,247,243,241,245,225,242,101,128,51,190,108,146,0,108,122,220,124,247,125,20,125,86,125,124,126,20,126,29,126,45,126,69,126,87,126,205,126,246,127,125,127,133,127,166,127,175,127,183,127,245,97,7,122,236,122,246,122,253,123,4,123,29,123,45,124,235,226,229,238,231,225,236,105,128,9,178,227,245,244,101,128,1,58,228,229,246,97,128,9,50,231,117,2,123,11,123,20,234,225,242,225,244,105,128,10,178,242,237,245,235,232,105,128,10,50,235,235,232,225,238,231,249,225,239,244,232,225,105,128,14,69,109,10,123,67,124,6,124,23,124,61,124,75,124,94,124,110,124,130,124,150,124,173,97,2,123,73,123,254,236,229,102,4,123,85,123,99,123,191,123,208,230,233,238,225,236,225,242,225,226,233,99,128,254,252,232,225,237,250,97,2,123,109,123,150,225,226,239,246,101,2,123,119,123,133,230,233,238,225,236,225,242,225,226,233,99,128,254,248,233,243,239,236,225,244,229,228,225,242,225,226,233,99,128,254,247,226,229,236,239,119,2,123,160,123,174,230,233,238,225,236,225,242,225,226,233,99,128,254,250,233,243,239,236,225,244,229,228,225,242,225,226,233,99,128,254,249,233,243,239,236,225,244,229,228,225,242,225,226,233,99,128,254,251,237,225,228,228,225,225,226,239,246,101,2,123,223,123,237,230,233,238,225,236,225,242,225,226,233,99,128,254,246,233,243,239,236,225,244,229,228,225,242,225,226,233,99,128,254,245,242,225,226,233,99,128,6,68,226,228,97,129,3,187,124,14,243,244,242,239,235,101,128,1,155,229,100,130,5,220,124,32,124,52,228,225,231,229,243,104,129,251,60,124,43,232,229,226,242,229,119,128,251,60,232,229,226,242,229,119,128,5,220,230,233,238,225,236,225,242,225,226,233,99,128,254,222,232,225,232,233,238,233,244,233,225,236,225,242,225,226,233,99,128,252,202,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,223,234,229,229,237,233,238,233,244,233,225,236,225,242,225,226,233,99,128,252,201,235,232,225,232,233,238,233,244,233,225,236,225,242,225,226,233,99,128,252,203,236,225,237,232,229,232,233,243,239,236,225,244,229,228,225,242,225,226,233,99,128,253,242,237,101,2,124,180,124,193,228,233,225,236,225,242,225,226,233,99,128,254,224,229,109,2,124,200,124,219,232,225,232,233,238,233,244,233,225,236,225,242,225,226,233,99,128,253,136,233,238,233,244,233,225,236,225,242,225,226,233,99,128,252,204,242,231,229,227,233,242,227,236,101,128,37,239,98,3,124,255,125,4,125,10,225,114,128,1,154,229,236,116,128,2,108,239,240,239,237,239,230,111,128,49,12,99,4,125,30,125,37,125,46,125,73,225,242,239,110,128,1,62,229,228,233,236,236,97,128,1,60,233,242,99,2,125,54,125,59,236,101,128,36,219,245,237,230,236,229,248,226,229,236,239,119,128,30,61,239,237,237,225,225,227,227,229,238,116,128,1,60,228,239,116,130,1,64,125,96,125,105,225,227,227,229,238,116,128,1,64,226,229,236,239,119,129,30,55,125,115,237,225,227,242,239,110,128,30,57,101,3,125,132,125,170,126,15,230,116,2,125,139,125,155,225,238,231,236,229,225,226,239,246,229,227,237,98,128,3,26,244,225,227,235,226,229,236,239,247,227,237,98,128,3,24,243,115,132,0,60,125,183,125,205,125,217,126,7,229,241,245,225,108,129,34,100,125,193,239,242,231,242,229,225,244,229,114,128,34,218,237,239,238,239,243,240,225,227,101,128,255,28,111,2,125,223,125,252,114,2,125,229,125,242,229,241,245,233,246,225,236,229,238,116,128,34,114,231,242,229,225,244,229,114,128,34,118,246,229,242,229,241,245,225,108,128,34,102,243,237,225,236,108,128,254,100,250,104,128,2,110,230,226,236,239,227,107,128,37,140,232,239,239,235,242,229,244,242,239,230,236,229,120,128,2,109,105,2,126,51,126,56,242,97,128,32,164,247,238,225,242,237,229,238,233,225,110,128,5,108,106,129,1,201,126,75,229,227,249,242,233,236,236,233,99,128,4,89,108,132,246,192,126,99,126,123,126,134,126,143,97,2,126,105,126,112,228,229,246,97,128,9,51,231,245,234,225,242,225,244,105,128,10,179,233,238,229,226,229,236,239,119,128,30,59,236,225,228,229,246,97,128,9,52,246,239,227,225,236,233,99,3,126,157,126,167,126,174,226,229,238,231,225,236,105,128,9,225,228,229,246,97,128,9,97,246,239,247,229,236,243,233,231,110,2,126,188,126,198,226,229,238,231,225,236,105,128,9,227,228,229,246,97,128,9,99,109,3,126,213,126,226,126,237,233,228,228,236,229,244,233,236,228,101,128,2,107,239,238,239,243,240,225,227,101,128,255,76,243,241,245,225,242,101,128,51,208,111,6,127,4,127,16,127,58,127,69,127,75,127,117,227,232,245,236,225,244,232,225,105,128,14,44,231,233,227,225,108,3,127,28,127,34,127,53,225,238,100,128,34,39,238,239,116,129,0,172,127,42,242,229,246,229,242,243,229,100,128,35,16,239,114,128,34,40,236,233,238,231,244,232,225,105,128,14,37,238,231,115,128,1,127,247,236,233,238,101,2,127,85,127,108,99,2,127,91,127,103,229,238,244,229,242,236,233,238,101,128,254,78,237,98,128,3,50,228,225,243,232,229,100,128,254,77,250,229,238,231,101,128,37,202,240,225,242,229,110,128,36,167,115,3,127,141,127,148,127,156,236,225,243,104,128,1,66,241,245,225,242,101,128,33,19,245,240,229,242,233,239,114,128,246,238,244,243,232,225,228,101,128,37,145,245,244,232,225,105,128,14,38,246,239,227,225,236,233,99,3,127,197,127,207,127,214,226,229,238,231,225,236,105,128,9,140,228,229,246,97,128,9,12,246,239,247,229,236,243,233,231,110,2,127,228,127,238,226,229,238,231,225,236,105,128,9,226,228,229,246,97,128,9,98,248,243,241,245,225,242,101,128,51,211,109,144,0,109,128,35,130,144,130,169,130,196,130,221,132,18,132,40,133,95,133,125,133,174,134,25,134,47,134,72,134,81,135,108,135,136,97,12,128,61,128,71,128,135,128,142,128,167,128,215,130,51,130,76,130,81,130,95,130,107,130,112,226,229,238,231,225,236,105,128,9,174,99,2,128,77,128,129,242,239,110,132,0,175,128,91,128,102,128,108,128,117,226,229,236,239,247,227,237,98,128,3,49,227,237,98,128,3,4,236,239,247,237,239,100,128,2,205,237,239,238,239,243,240,225,227,101,128,255,227,245,244,101,128,30,63,228,229,246,97,128,9,46,231,117,2,128,149,128,158,234,225,242,225,244,105,128,10,174,242,237,245,235,232,105,128,10,46,104,2,128,173,128,205,225,240,225,235,104,2,128,183,128,192,232,229,226,242,229,119,128,5,164,236,229,230,244,232,229,226,242,229,119,128,5,164,233,242,225,231,225,238,97,128,48,126,105,5,128,227,129,40,129,103,129,133,130,39,227,232,225,244,244,225,247,97,3,128,242,129,17,129,24,236,239,119,2,128,250,129,5,236,229,230,244,244,232,225,105,128,248,149,242,233,231,232,244,244,232,225,105,128,248,148,244,232,225,105,128,14,75,245,240,240,229,242,236,229,230,244,244,232,225,105,128,248,147,229,107,3,129,49,129,80,129,87,236,239,119,2,129,57,129,68,236,229,230,244,244,232,225,105,128,248,140,242,233,231,232,244,244,232,225,105,128,248,139,244,232,225,105,128,14,72,245,240,240,229,242,236,229,230,244,244,232,225,105,128,248,138,232,225,238,225,235,225,116,2,129,115,129,126,236,229,230,244,244,232,225,105,128,248,132,244,232,225,105,128,14,49,116,3,129,141,129,169,129,232,225,233,235,232,117,2,129,151,129,162,236,229,230,244,244,232,225,105,128,248,137,244,232,225,105,128,14,71,232,111,3,129,178,129,209,129,216,236,239,119,2,129,186,129,197,236,229,230,244,244,232,225,105,128,248,143,242,233,231,232,244,244,232,225,105,128,248,142,244,232,225,105,128,14,73,245,240,240,229,242,236,229,230,244,244,232,225,105,128,248,141,242,105,3,129,241,130,16,130,23,236,239,119,2,129,249,130,4,236,229,230,244,244,232,225,105,128,248,146,242,233,231,232,244,244,232,225,105,128,248,145,244,232,225,105,128,14,74,245,240,240,229,242,236,229,230,244,244,232,225,105,128,248,144,249,225,237,239,235,244,232,225,105,128,14,70,235,225,244,225,235,225,238,97,129,48,222,130,64,232,225,236,230,247,233,228,244,104,128,255,143,236,101,128,38,66,238,243,249,239,238,243,241,245,225,242,101,128,51,71,241,225,230,232,229,226,242,229,119,128,5,190,242,115,128,38,66,115,2,130,118,130,136,239,242,225,227,233,242,227,236,229,232,229,226,242,229,119,128,5,175,241,245,225,242,101,128,51,131,98,2,130,150,130,160,239,240,239,237,239,230,111,128,49,7,243,241,245,225,242,101,128,51,212,99,2,130,175,130,183,233,242,227,236,101,128,36,220,245,226,229,228,243,241,245,225,242,101,128,51,165,228,239,116,2,130,204,130,213,225,227,227,229,238,116,128,30,65,226,229,236,239,119,128,30,67,101,7,130,237,131,108,131,119,131,134,131,159,131,196,131,208,101,2,130,243,131,95,109,4,130,253,131,6,131,20,131,36,225,242,225,226,233,99,128,6,69,230,233,238,225,236,225,242,225,226,233,99,128,254,226,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,227,237,101,2,131,43,131,56,228,233,225,236,225,242,225,226,233,99,128,254,228,229,237,105,2,131,64,131,79,238,233,244,233,225,236,225,242,225,226,233,99,128,252,209,243,239,236,225,244,229,228,225,242,225,226,233,99,128,252,72,244,239,242,245,243,241,245,225,242,101,128,51,77,232,233,242,225,231,225,238,97,128,48,129,233,250,233,229,242,225,243,241,245,225,242,101,128,51,126,235,225,244,225,235,225,238,97,129,48,225,131,147,232,225,236,230,247,233,228,244,104,128,255,146,109,130,5,222,131,167,131,187,228,225,231,229,243,104,129,251,62,131,178,232,229,226,242,229,119,128,251,62,232,229,226,242,229,119,128,5,222,238,225,242,237,229,238,233,225,110,128,5,116,242,235,232,97,3,131,219,131,228,132,5,232,229,226,242,229,119,128,5,165,235,229,230,245,236,97,2,131,239,131,248,232,229,226,242,229,119,128,5,166,236,229,230,244,232,229,226,242,229,119,128,5,166,236,229,230,244,232,229,226,242,229,119,128,5,165,104,2,132,24,132,30,239,239,107,128,2,113,250,243,241,245,225,242,101,128,51,146,105,6,132,54,132,91,132,228,132,239,133,8,133,65,228,100,2,132,61,132,86,236,229,228,239,244,235,225,244,225,235,225,238,225,232,225,236,230,247,233,228,244,104,128,255,101,239,116,128,0,183,229,245,109,5,132,105,132,140,132,155,132,164,132,215,97,2,132,111,132,126,227,233,242,227,236,229,235,239,242,229,225,110,128,50,114,240,225,242,229,238,235,239,242,229,225,110,128,50,18,227,233,242,227,236,229,235,239,242,229,225,110,128,50,100,235,239,242,229,225,110,128,49,65,112,2,132,170,132,202,97,2,132,176,132,190,238,243,233,239,243,235,239,242,229,225,110,128,49,112,242,229,238,235,239,242,229,225,110,128,50,4,233,229,245,240,235,239,242,229,225,110,128,49,110,243,233,239,243,235,239,242,229,225,110,128,49,111,232,233,242,225,231,225,238,97,128,48,127,235,225,244,225,235,225,238,97,129,48,223,132,252,232,225,236,230,247,233,228,244,104,128,255,144,238,117,2,133,15,133,60,115,132,34,18,133,27,133,38,133,47,133,53,226,229,236,239,247,227,237,98,128,3,32,227,233,242,227,236,101,128,34,150,237,239,100,128,2,215,240,236,245,115,128,34,19,244,101,128,32,50,242,105,2,133,72,133,86,226,225,225,242,245,243,241,245,225,242,101,128,51,74,243,241,245,225,242,101,128,51,73,108,2,133,101,133,116,239,238,231,236,229,231,244,245,242,238,229,100,128,2,112,243,241,245,225,242,101,128,51,150,109,3,133,133,133,147,133,158,227,245,226,229,228,243,241,245,225,242,101,128,51,163,239,238,239,243,240,225,227,101,128,255,77,243,241,245,225,242,229,228,243,241,245,225,242,101,128,51,159,111,5,133,186,133,212,133,237,133,247,134,0,104,2,133,192,133,202,233,242,225,231,225,238,97,128,48,130,237,243,241,245,225,242,101,128,51,193,235,225,244,225,235,225,238,97,129,48,226,133,225,232,225,236,230,247,233,228,244,104,128,255,147,236,243,241,245,225,242,101,128,51,214,237,225,244,232,225,105,128,14,33,246,229,242,243,243,241,245,225,242,101,129,51,167,134,15,228,243,241,245,225,242,101,128,51,168,240,97,2,134,32,134,38,242,229,110,128,36,168,243,241,245,225,242,101,128,51,171,115,2,134,53,134,62,243,241,245,225,242,101,128,51,179,245,240,229,242,233,239,114,128,246,239,244,245,242,238,229,100,128,2,111,117,141,0,181,134,111,134,115,134,125,134,149,134,159,134,181,134,192,134,217,134,240,134,250,135,24,135,88,135,98,49,128,0,181,225,243,241,245,225,242,101,128,51,130,227,104,2,134,132,134,142,231,242,229,225,244,229,114,128,34,107,236,229,243,115,128,34,106,230,243,241,245,225,242,101,128,51,140,103,2,134,165,134,172,242,229,229,107,128,3,188,243,241,245,225,242,101,128,51,141,232,233,242,225,231,225,238,97,128,48,128,235,225,244,225,235,225,238,97,129,48,224,134,205,232,225,236,230,247,233,228,244,104,128,255,145,108,2,134,223,134,232,243,241,245,225,242,101,128,51,149,244,233,240,236,121,128,0,215,237,243,241,245,225,242,101,128,51,155,238,225,104,2,135,2,135,11,232,229,226,242,229,119,128,5,163,236,229,230,244,232,229,226,242,229,119,128,5,163,115,2,135,30,135,79,233,99,3,135,39,135,56,135,67,225,236,238,239,244,101,129,38,106,135,50,228,226,108,128,38,107,230,236,225,244,243,233,231,110,128,38,109,243,232,225,242,240,243,233,231,110,128,38,111,243,241,245,225,242,101,128,51,178,246,243,241,245,225,242,101,128,51,182,247,243,241,245,225,242,101,128,51,188,118,2,135,114,135,127,237,229,231,225,243,241,245,225,242,101,128,51,185,243,241,245,225,242,101,128,51,183,119,2,135,142,135,155,237,229,231,225,243,241,245,225,242,101,128,51,191,243,241,245,225,242,101,128,51,189,110,150,0,110,135,212,136,90,136,114,136,180,136,205,137,7,137,17,137,84,137,127,139,161,139,179,139,204,139,235,140,5,140,70,142,52,142,60,142,85,142,93,143,61,143,71,143,81,97,8,135,230,135,250,136,1,136,8,136,33,136,44,136,69,136,81,98,2,135,236,135,245,229,238,231,225,236,105,128,9,168,236,97,128,34,7,227,245,244,101,128,1,68,228,229,246,97,128,9,40,231,117,2,136,15,136,24,234,225,242,225,244,105,128,10,168,242,237,245,235,232,105,128,10,40,232,233,242,225,231,225,238,97,128,48,106,235,225,244,225,235,225,238,97,129,48,202,136,57,232,225,236,230,247,233,228,244,104,128,255,133,240,239,243,244,242,239,240,232,101,128,1,73,243,241,245,225,242,101,128,51,129,98,2,136,96,136,106,239,240,239,237,239,230,111,128,49,11,243,240,225,227,101,128,0,160,99,4,136,124,136,131,136,140,136,167,225,242,239,110,128,1,72,229,228,233,236,236,97,128,1,70,233,242,99,2,136,148,136,153,236,101,128,36,221,245,237,230,236,229,248,226,229,236,239,119,128,30,75,239,237,237,225,225,227,227,229,238,116,128,1,70,228,239,116,2,136,188,136,197,225,227,227,229,238,116,128,30,69,226,229,236,239,119,128,30,71,101,3,136,213,136,224,136,249,232,233,242,225,231,225,238,97,128,48,109,235,225,244,225,235,225,238,97,129,48,205,136,237,232,225,236,230,247,233,228,244,104,128,255,136,247,243,232,229,241,229,236,243,233,231,110,128,32,170,230,243,241,245,225,242,101,128,51,139,103,2,137,23,137,73,97,3,137,31,137,41,137,48,226,229,238,231,225,236,105,128,9,153,228,229,246,97,128,9,25,231,117,2,137,55,137,64,234,225,242,225,244,105,128,10,153,242,237,245,235,232,105,128,10,25,239,238,231,245,244,232,225,105,128,14,7,104,2,137,90,137,100,233,242,225,231,225,238,97,128,48,147,239,239,107,2,137,108,137,115,236,229,230,116,128,2,114,242,229,244,242,239,230,236,229,120,128,2,115,105,4,137,137,138,50,138,61,138,119,229,245,110,7,137,155,137,190,137,222,137,236,137,245,138,22,138,35,97,2,137,161,137,176,227,233,242,227,236,229,235,239,242,229,225,110,128,50,111,240,225,242,229,238,235,239,242,229,225,110,128,50,15,227,105,2,137,197,137,209,229,245,227,235,239,242,229,225,110,128,49,53,242,227,236,229,235,239,242,229,225,110,128,50,97,232,233,229,245,232,235,239,242,229,225,110,128,49,54,235,239,242,229,225,110,128,49,52,240,97,2,137,252,138,10,238,243,233,239,243,235,239,242,229,225,110,128,49,104,242,229,238,235,239,242,229,225,110,128,50,1,243,233,239,243,235,239,242,229,225,110,128,49,103,244,233,235,229,245,244,235,239,242,229,225,110,128,49,102,232,233,242,225,231,225,238,97,128,48,107,107,2,138,67,138,91,225,244,225,235,225,238,97,129,48,203,138,79,232,225,236,230,247,233,228,244,104,128,255,134,232,225,232,233,116,2,138,101,138,112,236,229,230,244,244,232,225,105,128,248,153,244,232,225,105,128,14,77,238,101,141,0,57,138,150,138,159,138,169,138,199,138,206,138,231,139,2,139,36,139,48,139,59,139,92,139,100,139,111,225,242,225,226,233,99,128,6,105,226,229,238,231,225,236,105,128,9,239,227,233,242,227,236,101,129,36,104,138,180,233,238,246,229,242,243,229,243,225,238,243,243,229,242,233,102,128,39,146,228,229,246,97,128,9,111,231,117,2,138,213,138,222,234,225,242,225,244,105,128,10,239,242,237,245,235,232,105,128,10,111,232,97,2,138,238,138,249,227,235,225,242,225,226,233,99,128,6,105,238,231,250,232,239,117,128,48,41,105,2,139,8,139,26,228,229,239,231,242,225,240,232,233,227,240,225,242,229,110,128,50,40,238,230,229,242,233,239,114,128,32,137,237,239,238,239,243,240,225,227,101,128,255,25,239,236,228,243,244,249,236,101,128,247,57,112,2,139,65,139,72,225,242,229,110,128,36,124,229,114,2,139,79,139,85,233,239,100,128,36,144,243,233,225,110,128,6,249,242,239,237,225,110,128,33,120,243,245,240,229,242,233,239,114,128,32,121,116,2,139,117,139,155,229,229,110,2,139,125,139,134,227,233,242,227,236,101,128,36,114,112,2,139,140,139,147,225,242,229,110,128,36,134,229,242,233,239,100,128,36,154,232,225,105,128,14,89,106,129,1,204,139,167,229,227,249,242,233,236,236,233,99,128,4,90,235,225,244,225,235,225,238,97,129,48,243,139,192,232,225,236,230,247,233,228,244,104,128,255,157,108,2,139,210,139,224,229,231,242,233,231,232,244,236,239,238,103,128,1,158,233,238,229,226,229,236,239,119,128,30,73,109,2,139,241,139,252,239,238,239,243,240,225,227,101,128,255,78,243,241,245,225,242,101,128,51,154,110,2,140,11,140,61,97,3,140,19,140,29,140,36,226,229,238,231,225,236,105,128,9,163,228,229,246,97,128,9,35,231,117,2,140,43,140,52,234,225,242,225,244,105,128,10,163,242,237,245,235,232,105,128,10,35,238,225,228,229,246,97,128,9,41,111,6,140,84,140,95,140,120,140,161,141,113,142,40,232,233,242,225,231,225,238,97,128,48,110,235,225,244,225,235,225,238,97,129,48,206,140,108,232,225,236,230,247,233,228,244,104,128,255,137,110,3,140,128,140,144,140,153,226,242,229,225,235,233,238,231,243,240,225,227,101,128,0,160,229,238,244,232,225,105,128,14,19,245,244,232,225,105,128,14,25,239,110,7,140,178,140,187,140,201,140,235,140,251,141,36,141,95,225,242,225,226,233,99,128,6,70,230,233,238,225,236,225,242,225,226,233,99,128,254,230,231,232,245,238,238,97,2,140,212,140,221,225,242,225,226,233,99,128,6,186,230,233,238,225,236,225,242,225,226,233,99,128,251,159,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,231,234,229,229,237,105,2,141,5,141,20,238,233,244,233,225,236,225,242,225,226,233,99,128,252,210,243,239,236,225,244,229,228,225,242,225,226,233,99,128,252,75,237,101,2,141,43,141,56,228,233,225,236,225,242,225,226,233,99,128,254,232,229,237,105,2,141,64,141,79,238,233,244,233,225,236,225,242,225,226,233,99,128,252,213,243,239,236,225,244,229,228,225,242,225,226,233,99,128,252,78,238,239,239,238,230,233,238,225,236,225,242,225,226,233,99,128,252,141,116,7,141,129,141,140,141,169,141,204,141,216,141,236,142,6,227,239,238,244,225,233,238,115,128,34,12,101,2,141,146,141,162,236,229,237,229,238,116,129,34,9,141,157,239,102,128,34,9,241,245,225,108,128,34,96,231,242,229,225,244,229,114,129,34,111,141,181,238,239,114,2,141,189,141,197,229,241,245,225,108,128,34,113,236,229,243,115,128,34,121,233,228,229,238,244,233,227,225,108,128,34,98,236,229,243,115,129,34,110,141,225,238,239,242,229,241,245,225,108,128,34,112,112,2,141,242,141,252,225,242,225,236,236,229,108,128,34,38,242,229,227,229,228,229,115,128,34,128,243,117,3,142,15,142,22,142,31,226,243,229,116,128,34,132,227,227,229,229,228,115,128,34,129,240,229,242,243,229,116,128,34,133,247,225,242,237,229,238,233,225,110,128,5,118,240,225,242,229,110,128,36,169,115,2,142,66,142,75,243,241,245,225,242,101,128,51,177,245,240,229,242,233,239,114,128,32,127,244,233,236,228,101,128,0,241,117,132,3,189,142,105,142,116,142,197,143,24,232,233,242,225,231,225,238,97,128,48,108,107,2,142,122,142,146,225,244,225,235,225,238,97,129,48,204,142,134,232,225,236,230,247,233,228,244,104,128,255,135,244,97,3,142,155,142,165,142,172,226,229,238,231,225,236,105,128,9,188,228,229,246,97,128,9,60,231,117,2,142,179,142,188,234,225,242,225,244,105,128,10,188,242,237,245,235,232,105,128,10,60,109,2,142,203,142,237,226,229,242,243,233,231,110,130,0,35,142,217,142,229,237,239,238,239,243,240,225,227,101,128,255,3,243,237,225,236,108,128,254,95,229,114,2,142,244,143,20,225,236,243,233,231,110,2,142,255,143,7,231,242,229,229,107,128,3,116,236,239,247,229,242,231,242,229,229,107,128,3,117,111,128,33,22,110,130,5,224,143,32,143,52,228,225,231,229,243,104,129,251,64,143,43,232,229,226,242,229,119,128,251,64,232,229,226,242,229,119,128,5,224,246,243,241,245,225,242,101,128,51,181,247,243,241,245,225,242,101,128,51,187,249,97,3,143,90,143,100,143,107,226,229,238,231,225,236,105,128,9,158,228,229,246,97,128,9,30,231,117,2,143,114,143,123,234,225,242,225,244,105,128,10,158,242,237,245,235,232,105,128,10,30,111,147,0,111,143,174,143,196,144,18,144,188,145,4,145,19,145,59,145,182,145,203,145,241,145,252,146,174,148,8,148,72,148,105,148,151,149,24,149,71,149,83,97,2,143,180,143,187,227,245,244,101,128,0,243,238,231,244,232,225,105,128,14,45,98,4,143,206,143,248,144,1,144,11,225,242,242,229,100,130,2,117,143,218,143,229,227,249,242,233,236,236,233,99,128,4,233,228,233,229,242,229,243,233,243,227,249,242,233,236,236,233,99,128,4,235,229,238,231,225,236,105,128,9,147,239,240,239,237,239,230,111,128,49,27,242,229,246,101,128,1,79,99,3,144,26,144,99,144,178,97,2,144,32,144,93,238,228,242,97,3,144,43,144,50,144,61,228,229,246,97,128,9,17,231,245,234,225,242,225,244,105,128,10,145,246,239,247,229,236,243,233,231,110,2,144,75,144,82,228,229,246,97,128,9,73,231,245,234,225,242,225,244,105,128,10,201,242,239,110,128,1,210,233,242,99,2,144,107,144,112,236,101,128,36,222,245,237,230,236,229,120,133,0,244,144,131,144,139,144,150,144,158,144,170,225,227,245,244,101,128,30,209,228,239,244,226,229,236,239,119,128,30,217,231,242,225,246,101,128,30,211,232,239,239,235,225,226,239,246,101,128,30,213,244,233,236,228,101,128,30,215,249,242,233,236,236,233,99,128,4,62,100,4,144,198,144,221,144,227,144,250,226,108,2,144,205,144,213,225,227,245,244,101,128,1,81,231,242,225,246,101,128,2,13,229,246,97,128,9,19,233,229,242,229,243,233,115,129,0,246,144,239,227,249,242,233,236,236,233,99,128,4,231,239,244,226,229,236,239,119,128,30,205,101,129,1,83,145,10,235,239,242,229,225,110,128,49,90,103,3,145,27,145,42,145,49,239,238,229,107,129,2,219,145,36,227,237,98,128,3,40,242,225,246,101,128,0,242,245,234,225,242,225,244,105,128,10,147,104,4,145,69,145,80,145,90,145,168,225,242,237,229,238,233,225,110,128,5,133,233,242,225,231,225,238,97,128,48,74,111,2,145,96,145,106,239,235,225,226,239,246,101,128,30,207,242,110,133,1,161,145,121,145,129,145,140,145,148,145,160,225,227,245,244,101,128,30,219,228,239,244,226,229,236,239,119,128,30,227,231,242,225,246,101,128,30,221,232,239,239,235,225,226,239,246,101,128,30,223,244,233,236,228,101,128,30,225,245,238,231,225,242,245,237,236,225,245,116,128,1,81,105,129,1,163,145,188,238,246,229,242,244,229,228,226,242,229,246,101,128,2,15,107,2,145,209,145,233,225,244,225,235,225,238,97,129,48,170,145,221,232,225,236,230,247,233,228,244,104,128,255,117,239,242,229,225,110,128,49,87,236,229,232,229,226,242,229,119,128,5,171,109,6,146,10,146,38,146,45,146,134,146,145,146,163,225,227,242,239,110,130,1,77,146,22,146,30,225,227,245,244,101,128,30,83,231,242,225,246,101,128,30,81,228,229,246,97,128,9,80,229,231,97,133,3,201,146,61,146,65,146,76,146,90,146,106,49,128,3,214,227,249,242,233,236,236,233,99,128,4,97,236,225,244,233,238,227,236,239,243,229,100,128,2,119,242,239,245,238,228,227,249,242,233,236,236,233,99,128,4,123,116,2,146,112,146,127,233,244,236,239,227,249,242,233,236,236,233,99,128,4,125,239,238,239,115,128,3,206,231,245,234,225,242,225,244,105,128,10,208,233,227,242,239,110,129,3,191,146,155,244,239,238,239,115,128,3,204,239,238,239,243,240,225,227,101,128,255,79,238,101,145,0,49,146,213,146,222,146,232,147,6,147,31,147,40,147,49,147,74,147,108,147,142,147,154,147,173,147,184,147,217,147,227,147,235,147,246,225,242,225,226,233,99,128,6,97,226,229,238,231,225,236,105,128,9,231,227,233,242,227,236,101,129,36,96,146,243,233,238,246,229,242,243,229,243,225,238,243,243,229,242,233,102,128,39,138,100,2,147,12,147,18,229,246,97,128,9,103,239,244,229,238,236,229,225,228,229,114,128,32,36,229,233,231,232,244,104,128,33,91,230,233,244,244,229,100,128,246,220,231,117,2,147,56,147,65,234,225,242,225,244,105,128,10,231,242,237,245,235,232,105,128,10,103,232,97,3,147,83,147,94,147,99,227,235,225,242,225,226,233,99,128,6,97,236,102,128,0,189,238,231,250,232,239,117,128,48,33,105,2,147,114,147,132,228,229,239,231,242,225,240,232,233,227,240,225,242,229,110,128,50,32,238,230,229,242,233,239,114,128,32,129,237,239,238,239,243,240,225,227,101,128,255,17,238,245,237,229,242,225,244,239,242,226,229,238,231,225,236,105,128,9,244,239,236,228,243,244,249,236,101,128,247,49,112,2,147,190,147,197,225,242,229,110,128,36,116,229,114,2,147,204,147,210,233,239,100,128,36,136,243,233,225,110,128,6,241,241,245,225,242,244,229,114,128,0,188,242,239,237,225,110,128,33,112,243,245,240,229,242,233,239,114,128,0,185,244,104,2,147,253,148,2,225,105,128,14,81,233,242,100,128,33,83,111,3,148,16,148,50,148,66,103,2,148,22,148,40,239,238,229,107,129,1,235,148,31,237,225,227,242,239,110,128,1,237,245,242,237,245,235,232,105,128,10,19,237,225,244,242,225,231,245,242,237,245,235,232,105,128,10,75,240,229,110,128,2,84,112,3,148,80,148,87,148,98,225,242,229,110,128,36,170,229,238,226,245,236,236,229,116,128,37,230,244,233,239,110,128,35,37,114,2,148,111,148,140,100,2,148,117,148,128,230,229,237,233,238,233,238,101,128,0,170,237,225,243,227,245,236,233,238,101,128,0,186,244,232,239,231,239,238,225,108,128,34,31,115,5,148,163,148,195,148,212,149,1,149,14,232,239,242,116,2,148,172,148,179,228,229,246,97,128,9,18,246,239,247,229,236,243,233,231,238,228,229,246,97,128,9,74,236,225,243,104,129,0,248,148,204,225,227,245,244,101,128,1,255,237,225,236,108,2,148,221,148,232,232,233,242,225,231,225,238,97,128,48,73,235,225,244,225,235,225,238,97,129,48,169,148,245,232,225,236,230,247,233,228,244,104,128,255,107,244,242,239,235,229,225,227,245,244,101,128,1,255,245,240,229,242,233,239,114,128,246,240,116,2,149,30,149,41,227,249,242,233,236,236,233,99,128,4,127,233,236,228,101,130,0,245,149,52,149,60,225,227,245,244,101,128,30,77,228,233,229,242,229,243,233,115,128,30,79,245,226,239,240,239,237,239,230,111,128,49,33,118,2,149,89,149,170,229,114,2,149,96,149,162,236,233,238,101,131,32,62,149,109,149,132,149,155,99,2,149,115,149,127,229,238,244,229,242,236,233,238,101,128,254,74,237,98,128,3,5,100,2,149,138,149,146,225,243,232,229,100,128,254,73,226,236,247,225,246,121,128,254,76,247,225,246,121,128,254,75,243,227,239,242,101,128,0,175,239,247,229,236,243,233,231,110,3,149,185,149,195,149,202,226,229,238,231,225,236,105,128,9,203,228,229,246,97,128,9,75,231,245,234,225,242,225,244,105,128,10,203,112,145,0,112,149,251,152,123,152,134,152,143,152,155,154,80,154,90,155,82,156,101,156,191,156,217,157,92,157,100,158,2,158,60,158,88,158,98,97,14,150,25,150,57,150,67,150,74,150,81,150,129,150,140,150,154,150,165,150,212,150,226,151,238,152,21,152,111,97,2,150,31,150,43,237,240,243,243,241,245,225,242,101,128,51,128,243,229,238,244,239,243,241,245,225,242,101,128,51,43,226,229,238,231,225,236,105,128,9,170,227,245,244,101,128,30,85,228,229,246,97,128,9,42,103,2,150,87,150,105,101,2,150,93,150,100,228,239,247,110,128,33,223,245,112,128,33,222,117,2,150,111,150,120,234,225,242,225,244,105,128,10,170,242,237,245,235,232,105,128,10,42,232,233,242,225,231,225,238,97,128,48,113,233,249,225,238,238,239,233,244,232,225,105,128,14,47,235,225,244,225,235,225,238,97,128,48,209,108,2,150,171,150,196,225,244,225,236,233,250,225,244,233,239,238,227,249,242,233,236,236,233,227,227,237,98,128,4,132,239,227,232,235,225,227,249,242,233,236,236,233,99,128,4,192,238,243,233,239,243,235,239,242,229,225,110,128,49,127,114,3,150,234,150,255,151,227,97,2,150,240,150,248,231,242,225,240,104,128,0,182,236,236,229,108,128,34,37,229,110,2,151,6,151,116,236,229,230,116,136,0,40,151,29,151,44,151,49,151,54,151,65,151,77,151,100,151,105,225,236,244,239,238,229,225,242,225,226,233,99,128,253,62,226,116,128,248,237,229,120,128,248,236,233,238,230,229,242,233,239,114,128,32,141,237,239,238,239,243,240,225,227,101,128,255,8,115,2,151,83,151,90,237,225,236,108,128,254,89,245,240,229,242,233,239,114,128,32,125,244,112,128,248,235,246,229,242,244,233,227,225,108,128,254,53,242,233,231,232,116,136,0,41,151,140,151,155,151,160,151,165,151,176,151,188,151,211,151,216,225,236,244,239,238,229,225,242,225,226,233,99,128,253,63,226,116,128,248,248,229,120,128,248,247,233,238,230,229,242,233,239,114,128,32,142,237,239,238,239,243,240,225,227,101,128,255,9,115,2,151,194,151,201,237,225,236,108,128,254,90,245,240,229,242,233,239,114,128,32,126,244,112,128,248,246,246,229,242,244,233,227,225,108,128,254,54,244,233,225,236,228,233,230,102,128,34,2,115,3,151,246,152,1,152,13,229,241,232,229,226,242,229,119,128,5,192,232,244,225,232,229,226,242,229,119,128,5,153,241,245,225,242,101,128,51,169,244,225,104,134,5,183,152,39,152,53,152,58,152,67,152,82,152,98,49,2,152,45,152,49,49,128,5,183,100,128,5,183,178,97,128,5,183,232,229,226,242,229,119,128,5,183,238,225,242,242,239,247,232,229,226,242,229,119,128,5,183,241,245,225,242,244,229,242,232,229,226,242,229,119,128,5,183,247,233,228,229,232,229,226,242,229,119,128,5,183,250,229,242,232,229,226,242,229,119,128,5,161,226,239,240,239,237,239,230,111,128,49,6,227,233,242,227,236,101,128,36,223,228,239,244,225,227,227,229,238,116,128,30,87,101,137,5,228,152,177,152,188,152,208,152,220,152,240,153,86,153,97,153,118,154,73,227,249,242,233,236,236,233,99,128,4,63,228,225,231,229,243,104,129,251,68,152,199,232,229,226,242,229,119,128,251,68,229,250,233,243,241,245,225,242,101,128,51,59,230,233,238,225,236,228,225,231,229,243,232,232,229,226,242,229,119,128,251,67,104,5,152,252,153,19,153,27,153,41,153,71,225,114,2,153,3,153,10,225,226,233,99,128,6,126,237,229,238,233,225,110,128,5,122,229,226,242,229,119,128,5,228,230,233,238,225,236,225,242,225,226,233,99,128,251,87,105,2,153,47,153,62,238,233,244,233,225,236,225,242,225,226,233,99,128,251,88,242,225,231,225,238,97,128,48,122,237,229,228,233,225,236,225,242,225,226,233,99,128,251,89,235,225,244,225,235,225,238,97,128,48,218,237,233,228,228,236,229,232,239,239,235,227,249,242,233,236,236,233,99,128,4,167,114,5,153,130,153,142,153,184,154,49,154,62,225,230,229,232,229,226,242,229,119,128,251,78,227,229,238,116,131,0,37,153,155,153,164,153,176,225,242,225,226,233,99,128,6,106,237,239,238,239,243,240,225,227,101,128,255,5,243,237,225,236,108,128,254,106,105,2,153,190,154,31,239,100,134,0,46,153,207,153,218,153,229,153,241,153,252,154,8,225,242,237,229,238,233,225,110,128,5,137,227,229,238,244,229,242,229,100,128,0,183,232,225,236,230,247,233,228,244,104,128,255,97,233,238,230,229,242,233,239,114,128,246,231,237,239,238,239,243,240,225,227,101,128,255,14,115,2,154,14,154,21,237,225,236,108,128,254,82,245,240,229,242,233,239,114,128,246,232,243,240,239,237,229,238,233,231,242,229,229,235,227,237,98,128,3,66,240,229,238,228,233,227,245,236,225,114,128,34,165,244,232,239,245,243,225,238,100,128,32,48,243,229,244,97,128,32,167,230,243,241,245,225,242,101,128,51,138,104,3,154,98,154,148,155,29,97,3,154,106,154,116,154,123,226,229,238,231,225,236,105,128,9,171,228,229,246,97,128,9,43,231,117,2,154,130,154,139,234,225,242,225,244,105,128,10,171,242,237,245,235,232,105,128,10,43,105,133,3,198,154,162,154,166,154,252,155,4,155,15,49,128,3,213,229,245,240,104,4,154,179,154,214,154,229,154,238,97,2,154,185,154,200,227,233,242,227,236,229,235,239,242,229,225,110,128,50,122,240,225,242,229,238,235,239,242,229,225,110,128,50,26,227,233,242,227,236,229,235,239,242,229,225,110,128,50,108,235,239,242,229,225,110,128,49,77,240,225,242,229,238,235,239,242,229,225,110,128,50,12,236,225,244,233,110,128,2,120,238,244,232,245,244,232,225,105,128,14,58,243,249,237,226,239,236,231,242,229,229,107,128,3,213,111,3,155,37,155,42,155,68,239,107,128,1,165,240,104,2,155,49,155,58,225,238,244,232,225,105,128,14,30,245,238,231,244,232,225,105,128,14,28,243,225,237,240,232,225,239,244,232,225,105,128,14,32,105,133,3,192,155,96,156,52,156,63,156,74,156,88,229,245,112,6,155,112,155,147,155,179,155,207,155,221,156,17,97,2,155,118,155,133,227,233,242,227,236,229,235,239,242,229,225,110,128,50,115,240,225,242,229,238,235,239,242,229,225,110,128,50,19,227,105,2,155,154,155,166,229,245,227,235,239,242,229,225,110,128,49,118,242,227,236,229,235,239,242,229,225,110,128,50,101,107,2,155,185,155,199,233,249,229,239,235,235,239,242,229,225,110,128,49,114,239,242,229,225,110,128,49,66,240,225,242,229,238,235,239,242,229,225,110,128,50,5,243,233,239,115,2,155,230,156,2,107,2,155,236,155,250,233,249,229,239,235,235,239,242,229,225,110,128,49,116,239,242,229,225,110,128,49,68,244,233,235,229,245,244,235,239,242,229,225,110,128,49,117,116,2,156,23,156,38,232,233,229,245,244,232,235,239,242,229,225,110,128,49,119,233,235,229,245,244,235,239,242,229,225,110,128,49,115,232,233,242,225,231,225,238,97,128,48,116,235,225,244,225,235,225,238,97,128,48,212,243,249,237,226,239,236,231,242,229,229,107,128,3,214,247,242,225,242,237,229,238,233,225,110,128,5,131,236,245,115,132,0,43,156,115,156,126,156,135,156,168,226,229,236,239,247,227,237,98,128,3,31,227,233,242,227,236,101,128,34,149,109,2,156,141,156,148,233,238,245,115,128,0,177,111,2,156,154,156,158,100,128,2,214,238,239,243,240,225,227,101,128,255,11,115,2,156,174,156,181,237,225,236,108,128,254,98,245,240,229,242,233,239,114,128,32,122,109,2,156,197,156,208,239,238,239,243,240,225,227,101,128,255,80,243,241,245,225,242,101,128,51,216,111,5,156,229,156,240,157,51,157,62,157,72,232,233,242,225,231,225,238,97,128,48,125,233,238,244,233,238,231,233,238,228,229,120,4,157,4,157,16,157,28,157,41,228,239,247,238,247,232,233,244,101,128,38,31,236,229,230,244,247,232,233,244,101,128,38,28,242,233,231,232,244,247,232,233,244,101,128,38,30,245,240,247,232,233,244,101,128,38,29,235,225,244,225,235,225,238,97,128,48,221,240,236,225,244,232,225,105,128,14,27,243,244,225,236,237,225,242,107,129,48,18,157,85,230,225,227,101,128,48,32,240,225,242,229,110,128,36,171,114,3,157,108,157,134,157,159,101,2,157,114,157,122,227,229,228,229,115,128,34,122,243,227,242,233,240,244,233,239,110,128,33,30,233,237,101,2,157,142,157,148,237,239,100,128,2,185,242,229,246,229,242,243,229,100,128,32,53,111,4,157,169,157,176,157,186,157,199,228,245,227,116,128,34,15,234,229,227,244,233,246,101,128,35,5,236,239,238,231,229,228,235,225,238,97,128,48,252,112,2,157,205,157,242,101,2,157,211,157,218,236,236,239,114,128,35,24,242,243,117,2,157,226,157,233,226,243,229,116,128,34,130,240,229,242,243,229,116,128,34,131,239,242,244,233,239,110,129,34,55,157,253,225,108,128,34,29,115,2,158,8,158,51,105,130,3,200,158,16,158,27,227,249,242,233,236,236,233,99,128,4,113,236,233,240,238,229,245,237,225,244,225,227,249,242,233,236,236,233,227,227,237,98,128,4,134,243,241,245,225,242,101,128,51,176,117,2,158,66,158,77,232,233,242,225,231,225,238,97,128,48,119,235,225,244,225,235,225,238,97,128,48,215,246,243,241,245,225,242,101,128,51,180,247,243,241,245,225,242,101,128,51,186,113,136,0,113,158,128,159,177,159,188,159,197,159,204,159,216,159,254,160,6,97,4,158,138,158,161,158,225,159,160,100,2,158,144,158,150,229,246,97,128,9,88,237,225,232,229,226,242,229,119,128,5,168,102,4,158,171,158,180,158,194,158,210,225,242,225,226,233,99,128,6,66,230,233,238,225,236,225,242,225,226,233,99,128,254,214,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,215,237,229,228,233,225,236,225,242,225,226,233,99,128,254,216,237,225,244,115,136,5,184,158,248,159,12,159,26,159,31,159,36,159,45,159,60,159,147,49,3,159,0,159,4,159,8,48,128,5,184,97,128,5,184,99,128,5,184,50,2,159,18,159,22,55,128,5,184,57,128,5,184,179,51,128,5,184,228,101,128,5,184,232,229,226,242,229,119,128,5,184,238,225,242,242,239,247,232,229,226,242,229,119,128,5,184,113,2,159,66,159,132,225,244,225,110,4,159,79,159,88,159,103,159,119,232,229,226,242,229,119,128,5,184,238,225,242,242,239,247,232,229,226,242,229,119,128,5,184,241,245,225,242,244,229,242,232,229,226,242,229,119,128,5,184,247,233,228,229,232,229,226,242,229,119,128,5,184,245,225,242,244,229,242,232,229,226,242,229,119,128,5,184,247,233,228,229,232,229,226,242,229,119,128,5,184,242,238,229,249,240,225,242,225,232,229,226,242,229,119,128,5,159,226,239,240,239,237,239,230,111,128,49,17,227,233,242,227,236,101,128,36,224,232,239,239,107,128,2,160,237,239,238,239,243,240,225,227,101,128,255,81,239,102,130,5,231,159,225,159,245,228,225,231,229,243,104,129,251,71,159,236,232,229,226,242,229,119,128,251,71,232,229,226,242,229,119,128,5,231,240,225,242,229,110,128,36,172,117,4,160,16,160,28,160,117,160,204,225,242,244,229,242,238,239,244,101,128,38,105,226,245,244,115,135,5,187,160,49,160,54,160,59,160,64,160,73,160,88,160,104,177,56,128,5,187,178,53,128,5,187,179,49,128,5,187,232,229,226,242,229,119,128,5,187,238,225,242,242,239,247,232,229,226,242,229,119,128,5,187,241,245,225,242,244,229,242,232,229,226,242,229,119,128,5,187,247,233,228,229,232,229,226,242,229,119,128,5,187,229,243,244,233,239,110,133,0,63,160,136,160,159,160,176,160,184,160,196,225,114,2,160,143,160,150,225,226,233,99,128,6,31,237,229,238,233,225,110,128,5,94,228,239,247,110,129,0,191,160,168,243,237,225,236,108,128,247,191,231,242,229,229,107,128,3,126,237,239,238,239,243,240,225,227,101,128,255,31,243,237,225,236,108,128,247,63,239,244,101,4,160,216,161,31,161,51,161,80,228,226,108,133,0,34,160,232,160,239,160,246,161,2,161,23,226,225,243,101,128,32,30,236,229,230,116,128,32,28,237,239,238,239,243,240,225,227,101,128,255,2,240,242,233,237,101,129,48,30,161,12,242,229,246,229,242,243,229,100,128,48,29,242,233,231,232,116,128,32,29,236,229,230,116,129,32,24,161,40,242,229,246,229,242,243,229,100,128,32,27,114,2,161,57,161,67,229,246,229,242,243,229,100,128,32,27,233,231,232,116,129,32,25,161,76,110,128,1,73,243,233,238,231,108,2,161,90,161,97,226,225,243,101,128,32,26,101,129,0,39,161,103,237,239,238,239,243,240,225,227,101,128,255,7,114,145,0,114,161,153,162,157,162,168,162,215,163,10,164,27,164,51,164,146,166,180,166,217,166,229,167,27,167,35,167,197,167,208,167,243,168,87,97,11,161,177,161,188,161,198,161,205,162,14,162,30,162,55,162,66,162,91,162,114,162,151,225,242,237,229,238,233,225,110,128,5,124,226,229,238,231,225,236,105,128,9,176,227,245,244,101,128,1,85,100,4,161,215,161,221,161,235,162,5,229,246,97,128,9,48,233,227,225,108,129,34,26,161,230,229,120,128,248,229,239,246,229,242,243,243,241,245,225,242,101,129,51,174,161,251,228,243,241,245,225,242,101,128,51,175,243,241,245,225,242,101,128,51,173,230,101,129,5,191,162,21,232,229,226,242,229,119,128,5,191,231,117,2,162,37,162,46,234,225,242,225,244,105,128,10,176,242,237,245,235,232,105,128,10,48,232,233,242,225,231,225,238,97,128,48,137,235,225,244,225,235,225,238,97,129,48,233,162,79,232,225,236,230,247,233,228,244,104,128,255,151,236,239,247,229,242,228,233,225,231,239,238,225,236,226,229,238,231,225,236,105,128,9,241,109,2,162,120,162,143,233,228,228,236,229,228,233,225,231,239,238,225,236,226,229,238,231,225,236,105,128,9,240,243,232,239,242,110,128,2,100,244,233,111,128,34,54,226,239,240,239,237,239,230,111,128,49,22,99,4,162,178,162,185,162,194,162,202,225,242,239,110,128,1,89,229,228,233,236,236,97,128,1,87,233,242,227,236,101,128,36,225,239,237,237,225,225,227,227,229,238,116,128,1,87,100,2,162,221,162,231,226,236,231,242,225,246,101,128,2,17,239,116,2,162,238,162,247,225,227,227,229,238,116,128,30,89,226,229,236,239,119,129,30,91,163,1,237,225,227,242,239,110,128,30,93,101,6,163,24,163,69,163,104,163,159,163,184,163,217,102,2,163,30,163,43,229,242,229,238,227,229,237,225,242,107,128,32,59,236,229,248,243,117,2,163,53,163,60,226,243,229,116,128,34,134,240,229,242,243,229,116,128,34,135,231,233,243,244,229,114,2,163,80,163,85,229,100,128,0,174,115,2,163,91,163,97,225,238,115,128,248,232,229,242,233,102,128,246,218,104,3,163,112,163,135,163,149,225,114,2,163,119,163,126,225,226,233,99,128,6,49,237,229,238,233,225,110,128,5,128,230,233,238,225,236,225,242,225,226,233,99,128,254,174,233,242,225,231,225,238,97,128,48,140,235,225,244,225,235,225,238,97,129,48,236,163,172,232,225,236,230,247,233,228,244,104,128,255,154,243,104,130,5,232,163,193,163,208,228,225,231,229,243,232,232,229,226,242,229,119,128,251,72,232,229,226,242,229,119,128,5,232,118,3,163,225,163,238,164,14,229,242,243,229,228,244,233,236,228,101,128,34,61,233,97,2,163,245,163,254,232,229,226,242,229,119,128,5,151,237,245,231,242,225,243,232,232,229,226,242,229,119,128,5,151,236,239,231,233,227,225,236,238,239,116,128,35,16,230,233,243,232,232,239,239,107,129,2,126,164,40,242,229,246,229,242,243,229,100,128,2,127,104,2,164,57,164,80,97,2,164,63,164,73,226,229,238,231,225,236,105,128,9,221,228,229,246,97,128,9,93,111,131,3,193,164,90,164,119,164,133,239,107,129,2,125,164,97,244,245,242,238,229,100,129,2,123,164,108,243,245,240,229,242,233,239,114,128,2,181,243,249,237,226,239,236,231,242,229,229,107,128,3,241,244,233,227,232,239,239,235,237,239,100,128,2,222,105,6,164,160,165,204,165,250,166,5,166,30,166,166,229,245,108,9,164,182,164,217,164,232,164,246,165,36,165,50,165,136,165,149,165,184,97,2,164,188,164,203,227,233,242,227,236,229,235,239,242,229,225,110,128,50,113,240,225,242,229,238,235,239,242,229,225,110,128,50,17,227,233,242,227,236,229,235,239,242,229,225,110,128,50,99,232,233,229,245,232,235,239,242,229,225,110,128,49,64,107,2,164,252,165,28,233,249,229,239,107,2,165,6,165,15,235,239,242,229,225,110,128,49,58,243,233,239,243,235,239,242,229,225,110,128,49,105,239,242,229,225,110,128,49,57,237,233,229,245,237,235,239,242,229,225,110,128,49,59,112,3,165,58,165,90,165,105,97,2,165,64,165,78,238,243,233,239,243,235,239,242,229,225,110,128,49,108,242,229,238,235,239,242,229,225,110,128,50,3,232,233,229,245,240,232,235,239,242,229,225,110,128,49,63,233,229,245,112,2,165,114,165,123,235,239,242,229,225,110,128,49,60,243,233,239,243,235,239,242,229,225,110,128,49,107,243,233,239,243,235,239,242,229,225,110,128,49,61,116,2,165,155,165,170,232,233,229,245,244,232,235,239,242,229,225,110,128,49,62,233,235,229,245,244,235,239,242,229,225,110,128,49,106,249,229,239,242,233,238,232,233,229,245,232,235,239,242,229,225,110,128,49,109,231,232,116,2,165,212,165,220,225,238,231,236,101,128,34,31,116,2,165,226,165,240,225,227,235,226,229,236,239,247,227,237,98,128,3,25,242,233,225,238,231,236,101,128,34,191,232,233,242,225,231,225,238,97,128,48,138,235,225,244,225,235,225,238,97,129,48,234,166,18,232,225,236,230,247,233,228,244,104,128,255,152,110,2,166,36,166,152,103,131,2,218,166,46,166,57,166,63,226,229,236,239,247,227,237,98,128,3,37,227,237,98,128,3,10,232,225,236,102,2,166,72,166,118,236,229,230,116,131,2,191,166,85,166,96,166,107,225,242,237,229,238,233,225,110,128,5,89,226,229,236,239,247,227,237,98,128,3,28,227,229,238,244,229,242,229,100,128,2,211,242,233,231,232,116,130,2,190,166,130,166,141,226,229,236,239,247,227,237,98,128,3,57,227,229,238,244,229,242,229,100,128,2,210,246,229,242,244,229,228,226,242,229,246,101,128,2,19,244,244,239,242,245,243,241,245,225,242,101,128,51,81,108,2,166,186,166,197,233,238,229,226,229,236,239,119,128,30,95,239,238,231,236,229,103,129,2,124,166,208,244,245,242,238,229,100,128,2,122,237,239,238,239,243,240,225,227,101,128,255,82,111,3,166,237,166,248,167,17,232,233,242,225,231,225,238,97,128,48,141,235,225,244,225,235,225,238,97,129,48,237,167,5,232,225,236,230,247,233,228,244,104,128,255,155,242,245,225,244,232,225,105,128,14,35,240,225,242,229,110,128,36,173,114,3,167,43,167,79,167,109,97,3,167,51,167,61,167,68,226,229,238,231,225,236,105,128,9,220,228,229,246,97,128,9,49,231,245,242,237,245,235,232,105,128,10,92,229,104,2,167,86,167,95,225,242,225,226,233,99,128,6,145,230,233,238,225,236,225,242,225,226,233,99,128,251,141,246,239,227,225,236,233,99,4,167,125,167,135,167,142,167,153,226,229,238,231,225,236,105,128,9,224,228,229,246,97,128,9,96,231,245,234,225,242,225,244,105,128,10,224,246,239,247,229,236,243,233,231,110,3,167,169,167,179,167,186,226,229,238,231,225,236,105,128,9,196,228,229,246,97,128,9,68,231,245,234,225,242,225,244,105,128,10,196,243,245,240,229,242,233,239,114,128,246,241,116,2,167,214,167,222,226,236,239,227,107,128,37,144,245,242,238,229,100,129,2,121,167,232,243,245,240,229,242,233,239,114,128,2,180,117,4,167,253,168,8,168,33,168,80,232,233,242,225,231,225,238,97,128,48,139,235,225,244,225,235,225,238,97,129,48,235,168,21,232,225,236,230,247,233,228,244,104,128,255,153,112,2,168,39,168,74,229,101,2,168,46,168,60,237,225,242,235,226,229,238,231,225,236,105,128,9,242,243,233,231,238,226,229,238,231,225,236,105,128,9,243,233,225,104,128,246,221,244,232,225,105,128,14,36,246,239,227,225,236,233,99,4,168,103,168,113,168,120,168,131,226,229,238,231,225,236,105,128,9,139,228,229,246,97,128,9,11,231,245,234,225,242,225,244,105,128,10,139,246,239,247,229,236,243,233,231,110,3,168,147,168,157,168,164,226,229,238,231,225,236,105,128,9,195,228,229,246,97,128,9,67,231,245,234,225,242,225,244,105,128,10,195,115,147,0,115,168,217,170,187,170,198,171,68,171,107,174,49,174,60,176,203,179,85,179,131,179,158,180,93,180,160,181,193,181,203,182,133,182,206,183,120,183,130,97,9,168,237,168,247,169,12,169,84,169,109,169,120,169,145,169,177,169,217,226,229,238,231,225,236,105,128,9,184,227,245,244,101,129,1,91,169,0,228,239,244,225,227,227,229,238,116,128,30,101,100,5,169,24,169,33,169,39,169,53,169,69,225,242,225,226,233,99,128,6,53,229,246,97,128,9,56,230,233,238,225,236,225,242,225,226,233,99,128,254,186,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,187,237,229,228,233,225,236,225,242,225,226,233,99,128,254,188,231,117,2,169,91,169,100,234,225,242,225,244,105,128,10,184,242,237,245,235,232,105,128,10,56,232,233,242,225,231,225,238,97,128,48,85,235,225,244,225,235,225,238,97,129,48,181,169,133,232,225,236,230,247,233,228,244,104,128,255,123,236,236,225,236,236,225,232,239,245,225,236,225,249,232,229,247,225,243,225,236,236,225,237,225,242,225,226,233,99,128,253,250,237,229,235,104,130,5,225,169,188,169,208,228,225,231,229,243,104,129,251,65,169,199,232,229,226,242,229,119,128,251,65,232,229,226,242,229,119,128,5,225,242,97,5,169,230,170,48,170,56,170,106,170,114,97,5,169,242,169,250,170,2,170,33,170,41,225,244,232,225,105,128,14,50,229,244,232,225,105,128,14,65,233,237,225,233,109,2,170,12,170,23,225,236,225,233,244,232,225,105,128,14,68,245,225,238,244,232,225,105,128,14,67,237,244,232,225,105,128,14,51,244,232,225,105,128,14,48,229,244,232,225,105,128,14,64,105,3,170,64,170,88,170,99,105,2,170,70,170,81,236,229,230,244,244,232,225,105,128,248,134,244,232,225,105,128,14,53,236,229,230,244,244,232,225,105,128,248,133,244,232,225,105,128,14,52,239,244,232,225,105,128,14,66,117,3,170,122,170,172,170,179,101,3,170,130,170,154,170,165,101,2,170,136,170,147,236,229,230,244,244,232,225,105,128,248,136,244,232,225,105,128,14,55,236,229,230,244,244,232,225,105,128,248,135,244,232,225,105,128,14,54,244,232,225,105,128,14,56,245,244,232,225,105,128,14,57,226,239,240,239,237,239,230,111,128,49,25,99,5,170,210,170,231,170,240,171,33,171,55,225,242,239,110,129,1,97,170,219,228,239,244,225,227,227,229,238,116,128,30,103,229,228,233,236,236,97,128,1,95,232,247,97,131,2,89,170,252,171,7,171,26,227,249,242,233,236,236,233,99,128,4,217,228,233,229,242,229,243,233,243,227,249,242,233,236,236,233,99,128,4,219,232,239,239,107,128,2,90,233,242,99,2,171,41,171,46,236,101,128,36,226,245,237,230,236,229,120,128,1,93,239,237,237,225,225,227,227,229,238,116,128,2,25,228,239,116,2,171,76,171,85,225,227,227,229,238,116,128,30,97,226,229,236,239,119,129,30,99,171,95,228,239,244,225,227,227,229,238,116,128,30,105,101,9,171,127,171,143,171,178,171,243,172,90,172,117,172,142,172,223,172,250,225,231,245,236,236,226,229,236,239,247,227,237,98,128,3,60,99,2,171,149,171,171,239,238,100,129,32,51,171,157,244,239,238,229,227,232,233,238,229,243,101,128,2,202,244,233,239,110,128,0,167,229,110,4,171,189,171,198,171,212,171,228,225,242,225,226,233,99,128,6,51,230,233,238,225,236,225,242,225,226,233,99,128,254,178,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,179,237,229,228,233,225,236,225,242,225,226,233,99,128,254,180,231,239,108,135,5,182,172,7,172,21,172,26,172,35,172,50,172,66,172,77,49,2,172,13,172,17,51,128,5,182,102,128,5,182,178,99,128,5,182,232,229,226,242,229,119,128,5,182,238,225,242,242,239,247,232,229,226,242,229,119,128,5,182,241,245,225,242,244,229,242,232,229,226,242,229,119,128,5,182,244,225,232,229,226,242,229,119,128,5,146,247,233,228,229,232,229,226,242,229,119,128,5,182,104,2,172,96,172,107,225,242,237,229,238,233,225,110,128,5,125,233,242,225,231,225,238,97,128,48,91,235,225,244,225,235,225,238,97,129,48,187,172,130,232,225,236,230,247,233,228,244,104,128,255,126,237,105,2,172,149,172,192,227,239,236,239,110,131,0,59,172,163,172,172,172,184,225,242,225,226,233,99,128,6,27,237,239,238,239,243,240,225,227,101,128,255,27,243,237,225,236,108,128,254,84,246,239,233,227,229,228,237,225,242,235,235,225,238,97,129,48,156,172,211,232,225,236,230,247,233,228,244,104,128,255,159,238,116,2,172,230,172,240,233,243,241,245,225,242,101,128,51,34,239,243,241,245,225,242,101,128,51,35,246,229,110,142,0,55,173,28,173,37,173,47,173,77,173,84,173,94,173,119,173,146,173,180,173,192,173,203,173,236,173,244,173,255,225,242,225,226,233,99,128,6,103,226,229,238,231,225,236,105,128,9,237,227,233,242,227,236,101,129,36,102,173,58,233,238,246,229,242,243,229,243,225,238,243,243,229,242,233,102,128,39,144,228,229,246,97,128,9,109,229,233,231,232,244,232,115,128,33,94,231,117,2,173,101,173,110,234,225,242,225,244,105,128,10,237,242,237,245,235,232,105,128,10,109,232,97,2,173,126,173,137,227,235,225,242,225,226,233,99,128,6,103,238,231,250,232,239,117,128,48,39,105,2,173,152,173,170,228,229,239,231,242,225,240,232,233,227,240,225,242,229,110,128,50,38,238,230,229,242,233,239,114,128,32,135,237,239,238,239,243,240,225,227,101,128,255,23,239,236,228,243,244,249,236,101,128,247,55,112,2,173,209,173,216,225,242,229,110,128,36,122,229,114,2,173,223,173,229,233,239,100,128,36,142,243,233,225,110,128,6,247,242,239,237,225,110,128,33,118,243,245,240,229,242,233,239,114,128,32,119,116,2,174,5,174,43,229,229,110,2,174,13,174,22,227,233,242,227,236,101,128,36,112,112,2,174,28,174,35,225,242,229,110,128,36,132,229,242,233,239,100,128,36,152,232,225,105,128,14,87,230,244,232,249,240,232,229,110,128,0,173,104,7,174,76,175,50,175,61,175,75,176,20,176,33,176,197,97,6,174,90,174,101,174,111,174,122,175,9,175,34,225,242,237,229,238,233,225,110,128,5,119,226,229,238,231,225,236,105,128,9,182,227,249,242,233,236,236,233,99,128,4,72,100,2,174,128,174,224,228,97,4,174,139,174,148,174,179,174,193,225,242,225,226,233,99,128,6,81,228,225,237,237,97,2,174,158,174,167,225,242,225,226,233,99,128,252,97,244,225,238,225,242,225,226,233,99,128,252,94,230,225,244,232,225,225,242,225,226,233,99,128,252,96,235,225,243,242,97,2,174,203,174,212,225,242,225,226,233,99,128,252,98,244,225,238,225,242,225,226,233,99,128,252,95,101,132,37,146,174,236,174,243,174,251,175,4,228,225,242,107,128,37,147,236,233,231,232,116,128,37,145,237,229,228,233,245,109,128,37,146,246,97,128,9,54,231,117,2,175,16,175,25,234,225,242,225,244,105,128,10,182,242,237,245,235,232,105,128,10,54,236,243,232,229,236,229,244,232,229,226,242,229,119,128,5,147,226,239,240,239,237,239,230,111,128,49,21,227,232,225,227,249,242,233,236,236,233,99,128,4,73,101,4,175,85,175,150,175,160,175,177,229,110,4,175,96,175,105,175,119,175,135,225,242,225,226,233,99,128,6,52,230,233,238,225,236,225,242,225,226,233,99,128,254,182,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,183,237,229,228,233,225,236,225,242,225,226,233,99,128,254,184,233,227,239,240,244,233,99,128,3,227,241,229,108,129,32,170,175,168,232,229,226,242,229,119,128,32,170,246,97,134,5,176,175,194,175,209,175,223,175,232,175,247,176,7,49,2,175,200,175,205,177,53,128,5,176,53,128,5,176,50,2,175,215,175,219,50,128,5,176,101,128,5,176,232,229,226,242,229,119,128,5,176,238,225,242,242,239,247,232,229,226,242,229,119,128,5,176,241,245,225,242,244,229,242,232,229,226,242,229,119,128,5,176,247,233,228,229,232,229,226,242,229,119,128,5,176,232,225,227,249,242,233,236,236,233,99,128,4,187,105,2,176,39,176,50,237,225,227,239,240,244,233,99,128,3,237,110,131,5,233,176,60,176,143,176,152,100,2,176,66,176,132,225,231,229,243,104,130,251,73,176,78,176,87,232,229,226,242,229,119,128,251,73,115,2,176,93,176,113,232,233,238,228,239,116,129,251,44,176,104,232,229,226,242,229,119,128,251,44,233,238,228,239,116,129,251,45,176,123,232,229,226,242,229,119,128,251,45,239,244,232,229,226,242,229,119,128,5,193,232,229,226,242,229,119,128,5,233,115,2,176,158,176,178,232,233,238,228,239,116,129,251,42,176,169,232,229,226,242,229,119,128,251,42,233,238,228,239,116,129,251,43,176,188,232,229,226,242,229,119,128,251,43,239,239,107,128,2,130,105,8,176,221,177,9,177,20,177,45,177,75,177,83,177,96,178,11,231,237,97,131,3,195,176,233,176,237,176,245,49,128,3,194,230,233,238,225,108,128,3,194,236,245,238,225,244,229,243,249,237,226,239,236,231,242,229,229,107,128,3,242,232,233,242,225,231,225,238,97,128,48,87,235,225,244,225,235,225,238,97,129,48,183,177,33,232,225,236,230,247,233,228,244,104,128,255,124,236,245,113,2,177,53,177,62,232,229,226,242,229,119,128,5,189,236,229,230,244,232,229,226,242,229,119,128,5,189,237,233,236,225,114,128,34,60,238,228,239,244,232,229,226,242,229,119,128,5,194,239,115,6,177,111,177,146,177,178,177,206,177,220,177,252,97,2,177,117,177,132,227,233,242,227,236,229,235,239,242,229,225,110,128,50,116,240,225,242,229,238,235,239,242,229,225,110,128,50,20,227,105,2,177,153,177,165,229,245,227,235,239,242,229,225,110,128,49,126,242,227,236,229,235,239,242,229,225,110,128,50,102,107,2,177,184,177,198,233,249,229,239,235,235,239,242,229,225,110,128,49,122,239,242,229,225,110,128,49,69,238,233,229,245,238,235,239,242,229,225,110,128,49,123,112,2,177,226,177,239,225,242,229,238,235,239,242,229,225,110,128,50,6,233,229,245,240,235,239,242,229,225,110,128,49,125,244,233,235,229,245,244,235,239,242,229,225,110,128,49,124,120,141,0,54,178,41,178,50,178,60,178,90,178,97,178,122,178,149,178,183,178,195,178,206,178,239,178,247,179,2,225,242,225,226,233,99,128,6,102,226,229,238,231,225,236,105,128,9,236,227,233,242,227,236,101,129,36,101,178,71,233,238,246,229,242,243,229,243,225,238,243,243,229,242,233,102,128,39,143,228,229,246,97,128,9,108,231,117,2,178,104,178,113,234,225,242,225,244,105,128,10,236,242,237,245,235,232,105,128,10,108,232,97,2,178,129,178,140,227,235,225,242,225,226,233,99,128,6,102,238,231,250,232,239,117,128,48,38,105,2,178,155,178,173,228,229,239,231,242,225,240,232,233,227,240,225,242,229,110,128,50,37,238,230,229,242,233,239,114,128,32,134,237,239,238,239,243,240,225,227,101,128,255,22,239,236,228,243,244,249,236,101,128,247,54,112,2,178,212,178,219,225,242,229,110,128,36,121,229,114,2,178,226,178,232,233,239,100,128,36,141,243,233,225,110,128,6,246,242,239,237,225,110,128,33,117,243,245,240,229,242,233,239,114,128,32,118,116,2,179,8,179,79,229,229,110,2,179,16,179,58,99,2,179,22,179,30,233,242,227,236,101,128,36,111,245,242,242,229,238,227,249,228,229,238,239,237,233,238,225,244,239,242,226,229,238,231,225,236,105,128,9,249,112,2,179,64,179,71,225,242,229,110,128,36,131,229,242,233,239,100,128,36,151,232,225,105,128,14,86,108,2,179,91,179,111,225,243,104,129,0,47,179,99,237,239,238,239,243,240,225,227,101,128,255,15,239,238,103,129,1,127,179,119,228,239,244,225,227,227,229,238,116,128,30,155,109,2,179,137,179,147,233,236,229,230,225,227,101,128,38,58,239,238,239,243,240,225,227,101,128,255,83,111,6,179,172,179,222,179,233,180,2,180,47,180,58,102,2,179,178,179,192,240,225,243,245,241,232,229,226,242,229,119,128,5,195,116,2,179,198,179,207,232,249,240,232,229,110,128,0,173,243,233,231,238,227,249,242,233,236,236,233,99,128,4,76,232,233,242,225,231,225,238,97,128,48,93,235,225,244,225,235,225,238,97,129,48,189,179,246,232,225,236,230,247,233,228,244,104,128,255,127,236,233,228,245,115,2,180,12,180,29,236,239,238,231,239,246,229,242,236,225,249,227,237,98,128,3,56,243,232,239,242,244,239,246,229,242,236,225,249,227,237,98,128,3,55,242,245,243,233,244,232,225,105,128,14,41,115,3,180,66,180,76,180,84,225,236,225,244,232,225,105,128,14,40,239,244,232,225,105,128,14,11,245,225,244,232,225,105,128,14,42,240,97,3,180,102,180,122,180,154,227,101,129,0,32,180,109,232,225,227,235,225,242,225,226,233,99,128,0,32,228,101,129,38,96,180,129,243,245,233,116,2,180,138,180,146,226,236,225,227,107,128,38,96,247,232,233,244,101,128,38,100,242,229,110,128,36,174,241,245,225,242,101,11,180,188,180,199,180,213,180,238,180,255,181,25,181,40,181,73,181,100,181,156,181,171,226,229,236,239,247,227,237,98,128,3,59,99,2,180,205,180,209,99,128,51,196,109,128,51,157,228,233,225,231,239,238,225,236,227,242,239,243,243,232,225,244,227,232,230,233,236,108,128,37,169,232,239,242,233,250,239,238,244,225,236,230,233,236,108,128,37,164,107,2,181,5,181,9,103,128,51,143,109,129,51,158,181,15,227,225,240,233,244,225,108,128,51,206,108,2,181,31,181,35,110,128,51,209,239,103,128,51,210,109,4,181,50,181,54,181,59,181,63,103,128,51,142,233,108,128,51,213,109,128,51,156,243,241,245,225,242,229,100,128,51,161,239,242,244,232,239,231,239,238,225,236,227,242,239,243,243,232,225,244,227,232,230,233,236,108,128,37,166,245,240,240,229,114,2,181,110,181,133,236,229,230,244,244,239,236,239,247,229,242,242,233,231,232,244,230,233,236,108,128,37,167,242,233,231,232,244,244,239,236,239,247,229,242,236,229,230,244,230,233,236,108,128,37,168,246,229,242,244,233,227,225,236,230,233,236,108,128,37,165,247,232,233,244,229,247,233,244,232,243,237,225,236,236,226,236,225,227,107,128,37,163,242,243,241,245,225,242,101,128,51,219,115,2,181,209,182,123,97,4,181,219,181,229,181,236,181,247,226,229,238,231,225,236,105,128,9,183,228,229,246,97,128,9,55,231,245,234,225,242,225,244,105,128,10,183,238,103,8,182,10,182,24,182,38,182,52,182,67,182,81,182,95,182,108,227,233,229,245,227,235,239,242,229,225,110,128,49,73,232,233,229,245,232,235,239,242,229,225,110,128,49,133,233,229,245,238,231,235,239,242,229,225,110,128,49,128,235,233,249,229,239,235,235,239,242,229,225,110,128,49,50,238,233,229,245,238,235,239,242,229,225,110,128,49,101,240,233,229,245,240,235,239,242,229,225,110,128,49,67,243,233,239,243,235,239,242,229,225,110,128,49,70,244,233,235,229,245,244,235,239,242,229,225,110,128,49,56,245,240,229,242,233,239,114,128,246,242,116,2,182,139,182,162,229,242,236,233,238,103,129,0,163,182,150,237,239,238,239,243,240,225,227,101,128,255,225,242,239,235,101,2,182,171,182,188,236,239,238,231,239,246,229,242,236,225,249,227,237,98,128,3,54,243,232,239,242,244,239,246,229,242,236,225,249,227,237,98,128,3,53,117,7,182,222,182,254,183,20,183,31,183,72,183,82,183,86,226,243,229,116,130,34,130,182,233,182,244,238,239,244,229,241,245,225,108,128,34,138,239,242,229,241,245,225,108,128,34,134,99,2,183,4,183,12,227,229,229,228,115,128,34,123,232,244,232,225,116,128,34,11,232,233,242,225,231,225,238,97,128,48,89,107,2,183,37,183,61,225,244,225,235,225,238,97,129,48,185,183,49,232,225,236,230,247,233,228,244,104,128,255,125,245,238,225,242,225,226,233,99,128,6,82,237,237,225,244,233,239,110,128,34,17,110,128,38,60,240,229,242,243,229,116,130,34,131,183,99,183,110,238,239,244,229,241,245,225,108,128,34,139,239,242,229,241,245,225,108,128,34,135,246,243,241,245,225,242,101,128,51,220,249,239,245,247,225,229,242,225,243,241,245,225,242,101,128,51,124,116,144,0,116,183,183,184,192,184,213,185,100,185,140,187,188,191,70,192,145,192,157,192,169,193,202,193,227,194,57,194,237,195,165,195,255,97,10,183,205,183,215,183,236,183,243,184,12,184,90,184,107,184,132,184,146,184,150,226,229,238,231,225,236,105,128,9,164,227,107,2,183,222,183,229,228,239,247,110,128,34,164,236,229,230,116,128,34,163,228,229,246,97,128,9,36,231,117,2,183,250,184,3,234,225,242,225,244,105,128,10,164,242,237,245,235,232,105,128,10,36,104,4,184,22,184,31,184,45,184,75,225,242,225,226,233,99,128,6,55,230,233,238,225,236,225,242,225,226,233,99,128,254,194,105,2,184,51,184,66,238,233,244,233,225,236,225,242,225,226,233,99,128,254,195,242,225,231,225,238,97,128,48,95,237,229,228,233,225,236,225,242,225,226,233,99,128,254,196,233,243,249,239,245,229,242,225,243,241,245,225,242,101,128,51,125,235,225,244,225,235,225,238,97,129,48,191,184,120,232,225,236,230,247,233,228,244,104,128,255,128,244,247,229,229,236,225,242,225,226,233,99,128,6,64,117,128,3,196,118,130,5,234,184,158,184,183,228,225,231,229,115,129,251,74,184,168,104,129,251,74,184,174,232,229,226,242,229,119,128,251,74,232,229,226,242,229,119,128,5,234,98,2,184,198,184,203,225,114,128,1,103,239,240,239,237,239,230,111,128,49,10,99,6,184,227,184,234,184,241,184,250,185,60,185,87,225,242,239,110,128,1,101,227,245,242,108,128,2,168,229,228,233,236,236,97,128,1,99,232,229,104,4,185,6,185,15,185,29,185,45,225,242,225,226,233,99,128,6,134,230,233,238,225,236,225,242,225,226,233,99,128,251,123,233,238,233,244,233,225,236,225,242,225,226,233,99,128,251,124,237,229,228,233,225,236,225,242,225,226,233,99,128,251,125,233,242,99,2,185,68,185,73,236,101,128,36,227,245,237,230,236,229,248,226,229,236,239,119,128,30,113,239,237,237,225,225,227,227,229,238,116,128,1,99,100,2,185,106,185,116,233,229,242,229,243,233,115,128,30,151,239,116,2,185,123,185,132,225,227,227,229,238,116,128,30,107,226,229,236,239,119,128,30,109,101,9,185,160,185,171,185,191,186,201,186,226,187,34,187,101,187,106,187,158,227,249,242,233,236,236,233,99,128,4,66,228,229,243,227,229,238,228,229,242,227,249,242,233,236,236,233,99,128,4,173,104,7,185,207,185,216,185,230,186,14,186,44,186,85,186,183,225,242,225,226,233,99,128,6,42,230,233,238,225,236,225,242,225,226,233,99,128,254,150,232,225,232,105,2,185,239,185,254,238,233,244,233,225,236,225,242,225,226,233,99,128,252,162,243,239,236,225,244,229,228,225,242,225,226,233,99,128,252,12,105,2,186,20,186,35,238,233,244,233,225,236,225,242,225,226,233,99,128,254,151,242,225,231,225,238,97,128,48,102,234,229,229,237,105,2,186,54,186,69,238,233,244,233,225,236,225,242,225,226,233,99,128,252,161,243,239,236,225,244,229,228,225,242,225,226,233,99,128,252,11,109,2,186,91,186,125,225,242,226,245,244,97,2,186,102,186,111,225,242,225,226,233,99,128,6,41,230,233,238,225,236,225,242,225,226,233,99,128,254,148,101,2,186,131,186,144,228,233,225,236,225,242,225,226,233,99,128,254,152,229,237,105,2,186,152,186,167,238,233,244,233,225,236,225,242,225,226,233,99,128,252,164,243,239,236,225,244,229,228,225,242,225,226,233,99,128,252,14,238,239,239,238,230,233,238,225,236,225,242,225,226,233,99,128,252,115,235,225,244,225,235,225,238,97,129,48,198,186,214,232,225,236,230,247,233,228,244,104,128,255,131,108,2,186,232,186,251,229,240,232,239,238,101,129,33,33,186,243,226,236,225,227,107,128,38,14,233,243,232,97,2,187,4,187,19,231,229,228,239,236,225,232,229,226,242,229,119,128,5,160,241,229,244,225,238,225,232,229,226,242,229,119,128,5,169,110,4,187,44,187,53,187,72,187,93,227,233,242,227,236,101,128,36,105,233,228,229,239,231,242,225,240,232,233,227,240,225,242,229,110,128,50,41,112,2,187,78,187,85,225,242,229,110,128,36,125,229,242,233,239,100,128,36,145,242,239,237,225,110,128,33,121,243,104,128,2,167,116,131,5,216,187,116,187,136,187,145,228,225,231,229,243,104,129,251,56,187,127,232,229,226,242,229,119,128,251,56,232,229,226,242,229,119,128,5,216,243,229,227,249,242,233,236,236,233,99,128,4,181,246,233,114,2,187,166,187,175,232,229,226,242,229,119,128,5,155,236,229,230,244,232,229,226,242,229,119,128,5,155,104,6,187,202,188,98,188,220,189,96,190,3,191,60,97,5,187,214,187,224,187,231,188,0,188,29,226,229,238,231,225,236,105,128,9,165,228,229,246,97,128,9,37,231,117,2,187,238,187,247,234,225,242,225,244,105,128,10,165,242,237,245,235,232,105,128,10,37,108,2,188,6,188,15,225,242,225,226,233,99,128,6,48,230,233,238,225,236,225,242,225,226,233,99,128,254,172,238,244,232,225,235,232,225,116,3,188,44,188,75,188,82,236,239,119,2,188,52,188,63,236,229,230,244,244,232,225,105,128,248,152,242,233,231,232,244,244,232,225,105,128,248,151,244,232,225,105,128,14,76,245,240,240,229,242,236,229,230,244,244,232,225,105,128,248,150,101,3,188,106,188,170,188,193,104,4,188,116,188,125,188,139,188,155,225,242,225,226,233,99,128,6,43,230,233,238,225,236,225,242,225,226,233,99,128,254,154,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,155,237,229,228,233,225,236,225,242,225,226,233,99,128,254,156,242,101,2,188,177,188,186,229,248,233,243,244,115,128,34,3,230,239,242,101,128,34,52,244,97,130,3,184,188,202,188,206,49,128,3,209,243,249,237,226,239,236,231,242,229,229,107,128,3,209,105,2,188,226,189,56,229,245,244,104,4,188,239,189,18,189,33,189,42,97,2,188,245,189,4,227,233,242,227,236,229,235,239,242,229,225,110,128,50,121,240,225,242,229,238,235,239,242,229,225,110,128,50,25,227,233,242,227,236,229,235,239,242,229,225,110,128,50,107,235,239,242,229,225,110,128,49,76,240,225,242,229,238,235,239,242,229,225,110,128,50,11,242,244,229,229,110,2,189,66,189,75,227,233,242,227,236,101,128,36,108,112,2,189,81,189,88,225,242,229,110,128,36,128,229,242,233,239,100,128,36,148,111,6,189,110,189,127,189,132,189,146,189,151,189,204,238,225,238,231,237,239,238,244,232,239,244,232,225,105,128,14,17,239,107,128,1,173,240,232,245,244,232,225,239,244,232,225,105,128,14,18,242,110,128,0,254,244,104,3,189,160,189,184,189,194,97,2,189,166,189,176,232,225,238,244,232,225,105,128,14,23,238,244,232,225,105,128,14,16,239,238,231,244,232,225,105,128,14,24,245,238,231,244,232,225,105,128,14,22,245,243,225,238,100,2,189,214,189,225,227,249,242,233,236,236,233,99,128,4,130,243,243,229,240,225,242,225,244,239,114,2,189,240,189,249,225,242,225,226,233,99,128,6,108,240,229,242,243,233,225,110,128,6,108,242,229,101,144,0,51,190,41,190,50,190,60,190,90,190,97,190,107,190,132,190,159,190,193,190,205,190,224,190,235,191,12,191,34,191,42,191,53,225,242,225,226,233,99,128,6,99,226,229,238,231,225,236,105,128,9,233,227,233,242,227,236,101,129,36,98,190,71,233,238,246,229,242,243,229,243,225,238,243,243,229,242,233,102,128,39,140,228,229,246,97,128,9,105,229,233,231,232,244,232,115,128,33,92,231,117,2,190,114,190,123,234,225,242,225,244,105,128,10,233,242,237,245,235,232,105,128,10,105,232,97,2,190,139,190,150,227,235,225,242,225,226,233,99,128,6,99,238,231,250,232,239,117,128,48,35,105,2,190,165,190,183,228,229,239,231,242,225,240,232,233,227,240,225,242,229,110,128,50,34,238,230,229,242,233,239,114,128,32,131,237,239,238,239,243,240,225,227,101,128,255,19,238,245,237,229,242,225,244,239,242,226,229,238,231,225,236,105,128,9,246,239,236,228,243,244,249,236,101,128,247,51,112,2,190,241,190,248,225,242,229,110,128,36,118,229,114,2,190,255,191,5,233,239,100,128,36,138,243,233,225,110,128,6,243,241,245,225,242,244,229,242,115,129,0,190,191,25,229,237,228,225,243,104,128,246,222,242,239,237,225,110,128,33,114,243,245,240,229,242,233,239,114,128,0,179,244,232,225,105,128,14,83,250,243,241,245,225,242,101,128,51,148,105,7,191,86,191,97,191,212,192,54,192,66,192,115,192,132,232,233,242,225,231,225,238,97,128,48,97,107,2,191,103,191,127,225,244,225,235,225,238,97,129,48,193,191,115,232,225,236,230,247,233,228,244,104,128,255,129,229,245,116,4,191,139,191,174,191,189,191,198,97,2,191,145,191,160,227,233,242,227,236,229,235,239,242,229,225,110,128,50,112,240,225,242,229,238,235,239,242,229,225,110,128,50,16,227,233,242,227,236,229,235,239,242,229,225,110,128,50,98,235,239,242,229,225,110,128,49,55,240,225,242,229,238,235,239,242,229,225,110,128,50,2,236,228,101,133,2,220,191,228,191,239,192,0,192,12,192,40,226,229,236,239,247,227,237,98,128,3,48,99,2,191,245,191,250,237,98,128,3,3,239,237,98,128,3,3,228,239,245,226,236,229,227,237,98,128,3,96,111,2,192,18,192,28,240,229,242,225,244,239,114,128,34,60,246,229,242,236,225,249,227,237,98,128,3,52,246,229,242,244,233,227,225,236,227,237,98,128,3,62,237,229,243,227,233,242,227,236,101,128,34,151,112,2,192,72,192,102,229,232,97,2,192,80,192,89,232,229,226,242,229,119,128,5,150,236,229,230,244,232,229,226,242,229,119,128,5,150,240,233,231,245,242,237,245,235,232,105,128,10,112,244,236,239,227,249,242,233,236,236,233,227,227,237,98,128,4,131,247,238,225,242,237,229,238,233,225,110,128,5,127,236,233,238,229,226,229,236,239,119,128,30,111,237,239,238,239,243,240,225,227,101,128,255,84,111,7,192,185,192,196,192,207,192,232,193,96,193,108,193,192,225,242,237,229,238,233,225,110,128,5,105,232,233,242,225,231,225,238,97,128,48,104,235,225,244,225,235,225,238,97,129,48,200,192,220,232,225,236,230,247,233,228,244,104,128,255,132,110,3,192,240,193,82,193,87,101,4,192,250,193,63,193,70,193,76,226,225,114,4,193,6,193,35,193,45,193,54,229,248,244,242,97,2,193,16,193,26,232,233,231,232,237,239,100,128,2,229,236,239,247,237,239,100,128,2,233,232,233,231,232,237,239,100,128,2,230,236,239,247,237,239,100,128,2,232,237,233,228,237,239,100,128,2,231,230,233,246,101,128,1,189,243,233,120,128,1,133,244,247,111,128,1,168,239,115,128,3,132,243,241,245,225,242,101,128,51,39,240,225,244,225,235,244,232,225,105,128,14,15,242,244,239,233,243,229,243,232,229,236,236,226,242,225,227,235,229,116,2,193,131,193,161,236,229,230,116,130,48,20,193,142,193,150,243,237,225,236,108,128,254,93,246,229,242,244,233,227,225,108,128,254,57,242,233,231,232,116,130,48,21,193,173,193,181,243,237,225,236,108,128,254,94,246,229,242,244,233,227,225,108,128,254,58,244,225,239,244,232,225,105,128,14,21,240,97,2,193,209,193,221,236,225,244,225,236,232,239,239,107,128,1,171,242,229,110,128,36,175,114,3,193,235,194,10,194,25,225,228,229,237,225,242,107,129,33,34,193,247,115,2,193,253,194,3,225,238,115,128,248,234,229,242,233,102,128,246,219,229,244,242,239,230,236,229,248,232,239,239,107,128,2,136,233,225,103,4,194,37,194,42,194,47,194,52,228,110,128,37,188,236,102,128,37,196,242,116,128,37,186,245,112,128,37,178,115,132,2,166,194,69,194,108,194,214,194,227,225,228,105,130,5,230,194,79,194,99,228,225,231,229,243,104,129,251,70,194,90,232,229,226,242,229,119,128,251,70,232,229,226,242,229,119,128,5,230,101,2,194,114,194,125,227,249,242,233,236,236,233,99,128,4,70,242,101,134,5,181,194,142,194,156,194,161,194,170,194,185,194,201,49,2,194,148,194,152,50,128,5,181,101,128,5,181,178,98,128,5,181,232,229,226,242,229,119,128,5,181,238,225,242,242,239,247,232,229,226,242,229,119,128,5,181,241,245,225,242,244,229,242,232,229,226,242,229,119,128,5,181,247,233,228,229,232,229,226,242,229,119,128,5,181,232,229,227,249,242,233,236,236,233,99,128,4,91,245,240,229,242,233,239,114,128,246,243,116,4,194,247,195,41,195,106,195,157,97,3,194,255,195,9,195,16,226,229,238,231,225,236,105,128,9,159,228,229,246,97,128,9,31,231,117,2,195,23,195,32,234,225,242,225,244,105,128,10,159,242,237,245,235,232,105,128,10,31,229,104,4,195,52,195,61,195,75,195,91,225,242,225,226,233,99,128,6,121,230,233,238,225,236,225,242,225,226,233,99,128,251,103,233,238,233,244,233,225,236,225,242,225,226,233,99,128,251,104,237,229,228,233,225,236,225,242,225,226,233,99,128,251,105,232,97,3,195,115,195,125,195,132,226,229,238,231,225,236,105,128,9,160,228,229,246,97,128,9,32,231,117,2,195,139,195,148,234,225,242,225,244,105,128,10,160,242,237,245,235,232,105,128,10,32,245,242,238,229,100,128,2,135,117,3,195,173,195,184,195,209,232,233,242,225,231,225,238,97,128,48,100,235,225,244,225,235,225,238,97,129,48,196,195,197,232,225,236,230,247,233,228,244,104,128,255,130,243,237,225,236,108,2,195,219,195,230,232,233,242,225,231,225,238,97,128,48,99,235,225,244,225,235,225,238,97,129,48,195,195,243,232,225,236,230,247,233,228,244,104,128,255,111,119,2,196,5,196,110,101,2,196,11,196,59,236,246,101,3,196,21,196,30,196,51,227,233,242,227,236,101,128,36,107,112,2,196,36,196,43,225,242,229,110,128,36,127,229,242,233,239,100,128,36,147,242,239,237,225,110,128,33,123,238,244,121,3,196,69,196,78,196,89,227,233,242,227,236,101,128,36,115,232,225,238,231,250,232,239,117,128,83,68,112,2,196,95,196,102,225,242,229,110,128,36,135,229,242,233,239,100,128,36,155,111,142,0,50,196,142,196,151,196,161,196,191,196,243,197,12,197,39,197,73,197,85,197,104,197,115,197,148,197,156,197,180,225,242,225,226,233,99,128,6,98,226,229,238,231,225,236,105,128,9,232,227,233,242,227,236,101,129,36,97,196,172,233,238,246,229,242,243,229,243,225,238,243,243,229,242,233,102,128,39,139,100,2,196,197,196,203,229,246,97,128,9,104,239,116,2,196,210,196,221,229,238,236,229,225,228,229,114,128,32,37,236,229,225,228,229,114,129,32,37,196,232,246,229,242,244,233,227,225,108,128,254,48,231,117,2,196,250,197,3,234,225,242,225,244,105,128,10,232,242,237,245,235,232,105,128,10,104,232,97,2,197,19,197,30,227,235,225,242,225,226,233,99,128,6,98,238,231,250,232,239,117,128,48,34,105,2,197,45,197,63,228,229,239,231,242,225,240,232,233,227,240,225,242,229,110,128,50,33,238,230,229,242,233,239,114,128,32,130,237,239,238,239,243,240,225,227,101,128,255,18,238,245,237,229,242,225,244,239,242,226,229,238,231,225,236,105,128,9,245,239,236,228,243,244,249,236,101,128,247,50,112,2,197,121,197,128,225,242,229,110,128,36,117,229,114,2,197,135,197,141,233,239,100,128,36,137,243,233,225,110,128,6,242,242,239,237,225,110,128,33,113,115,2,197,162,197,170,244,242,239,235,101,128,1,187,245,240,229,242,233,239,114,128,0,178,244,104,2,197,187,197,192,225,105,128,14,82,233,242,228,115,128,33,84,117,145,0,117,197,237,197,245,198,30,198,87,198,225,199,6,199,129,199,145,199,196,200,10,200,91,200,100,200,219,200,243,201,95,201,123,201,237,225,227,245,244,101,128,0,250,98,4,197,255,198,4,198,13,198,23,225,114,128,2,137,229,238,231,225,236,105,128,9,137,239,240,239,237,239,230,111,128,49,40,242,229,246,101,128,1,109,99,3,198,38,198,45,198,77,225,242,239,110,128,1,212,233,242,99,2,198,53,198,58,236,101,128,36,228,245,237,230,236,229,120,129,0,251,198,69,226,229,236,239,119,128,30,119,249,242,233,236,236,233,99,128,4,67,100,5,198,99,198,110,198,133,198,139,198,215,225,244,244,225,228,229,246,97,128,9,81,226,108,2,198,117,198,125,225,227,245,244,101,128,1,113,231,242,225,246,101,128,2,21,229,246,97,128,9,9,233,229,242,229,243,233,115,133,0,252,198,159,198,167,198,175,198,198,198,206,225,227,245,244,101,128,1,216,226,229,236,239,119,128,30,115,99,2,198,181,198,188,225,242,239,110,128,1,218,249,242,233,236,236,233,99,128,4,241,231,242,225,246,101,128,1,220,237,225,227,242,239,110,128,1,214,239,244,226,229,236,239,119,128,30,229,103,2,198,231,198,238,242,225,246,101,128,0,249,117,2,198,244,198,253,234,225,242,225,244,105,128,10,137,242,237,245,235,232,105,128,10,9,104,3,199,14,199,24,199,102,233,242,225,231,225,238,97,128,48,70,111,2,199,30,199,40,239,235,225,226,239,246,101,128,30,231,242,110,133,1,176,199,55,199,63,199,74,199,82,199,94,225,227,245,244,101,128,30,233,228,239,244,226,229,236,239,119,128,30,241,231,242,225,246,101,128,30,235,232,239,239,235,225,226,239,246,101,128,30,237,244,233,236,228,101,128,30,239,245,238,231,225,242,245,237,236,225,245,116,129,1,113,199,118,227,249,242,233,236,236,233,99,128,4,243,233,238,246,229,242,244,229,228,226,242,229,246,101,128,2,23,107,3,199,153,199,177,199,188,225,244,225,235,225,238,97,129,48,166,199,165,232,225,236,230,247,233,228,244,104,128,255,115,227,249,242,233,236,236,233,99,128,4,121,239,242,229,225,110,128,49,92,109,2,199,202,199,255,97,2,199,208,199,241,227,242,239,110,130,1,107,199,219,199,230,227,249,242,233,236,236,233,99,128,4,239,228,233,229,242,229,243,233,115,128,30,123,244,242,225,231,245,242,237,245,235,232,105,128,10,65,239,238,239,243,240,225,227,101,128,255,85,110,2,200,16,200,71,228,229,242,243,227,239,242,101,132,0,95,200,35,200,41,200,53,200,64,228,226,108,128,32,23,237,239,238,239,243,240,225,227,101,128,255,63,246,229,242,244,233,227,225,108,128,254,51,247,225,246,121,128,254,79,105,2,200,77,200,82,239,110,128,34,42,246,229,242,243,225,108,128,34,0,239,231,239,238,229,107,128,1,115,112,5,200,112,200,119,200,127,200,142,200,193,225,242,229,110,128,36,176,226,236,239,227,107,128,37,128,240,229,242,228,239,244,232,229,226,242,229,119,128,5,196,243,233,236,239,110,131,3,197,200,156,200,177,200,185,228,233,229,242,229,243,233,115,129,3,203,200,169,244,239,238,239,115,128,3,176,236,225,244,233,110,128,2,138,244,239,238,239,115,128,3,205,244,225,227,107,2,200,202,200,213,226,229,236,239,247,227,237,98,128,3,29,237,239,100,128,2,212,114,2,200,225,200,237,225,231,245,242,237,245,235,232,105,128,10,115,233,238,103,128,1,111,115,3,200,251,201,10,201,55,232,239,242,244,227,249,242,233,236,236,233,99,128,4,94,237,225,236,108,2,201,19,201,30,232,233,242,225,231,225,238,97,128,48,69,235,225,244,225,235,225,238,97,129,48,165,201,43,232,225,236,230,247,233,228,244,104,128,255,105,244,242,225,233,231,232,116,2,201,67,201,78,227,249,242,233,236,236,233,99,128,4,175,243,244,242,239,235,229,227,249,242,233,236,236,233,99,128,4,177,244,233,236,228,101,130,1,105,201,107,201,115,225,227,245,244,101,128,30,121,226,229,236,239,119,128,30,117,117,5,201,135,201,145,201,152,201,177,201,193,226,229,238,231,225,236,105,128,9,138,228,229,246,97,128,9,10,231,117,2,201,159,201,168,234,225,242,225,244,105,128,10,138,242,237,245,235,232,105,128,10,10,237,225,244,242,225,231,245,242,237,245,235,232,105,128,10,66,246,239,247,229,236,243,233,231,110,3,201,209,201,219,201,226,226,229,238,231,225,236,105,128,9,194,228,229,246,97,128,9,66,231,245,234,225,242,225,244,105,128,10,194,246,239,247,229,236,243,233,231,110,3,201,253,202,7,202,14,226,229,238,231,225,236,105,128,9,193,228,229,246,97,128,9,65,231,245,234,225,242,225,244,105,128,10,193,118,139,0,118,202,51,202,199,202,208,202,219,203,148,203,155,203,253,204,9,204,109,204,117,204,138,97,4,202,61,202,68,202,93,202,104,228,229,246,97,128,9,53,231,117,2,202,75,202,84,234,225,242,225,244,105,128,10,181,242,237,245,235,232,105,128,10,53,235,225,244,225,235,225,238,97,128,48,247,118,132,5,213,202,116,202,143,202,175,202,187,228,225,231,229,243,104,130,251,53,202,129,202,134,182,53,128,251,53,232,229,226,242,229,119,128,251,53,104,2,202,149,202,157,229,226,242,229,119,128,5,213,239,236,225,109,129,251,75,202,166,232,229,226,242,229,119,128,251,75,246,225,246,232,229,226,242,229,119,128,5,240,249,239,228,232,229,226,242,229,119,128,5,241,227,233,242,227,236,101,128,36,229,228,239,244,226,229,236,239,119,128,30,127,101,6,202,233,202,244,203,52,203,63,203,69,203,136,227,249,242,233,236,236,233,99,128,4,50,104,4,202,254,203,7,203,21,203,37,225,242,225,226,233,99,128,6,164,230,233,238,225,236,225,242,225,226,233,99,128,251,107,233,238,233,244,233,225,236,225,242,225,226,233,99,128,251,108,237,229,228,233,225,236,225,242,225,226,233,99,128,251,109,235,225,244,225,235,225,238,97,128,48,249,238,245,115,128,38,64,242,244,233,227,225,108,2,203,80,203,86,226,225,114,128,0,124,236,233,238,101,4,203,99,203,110,203,121,203,130,225,226,239,246,229,227,237,98,128,3,13,226,229,236,239,247,227,237,98,128,3,41,236,239,247,237,239,100,128,2,204,237,239,100,128,2,200,247,225,242,237,229,238,233,225,110,128,5,126,232,239,239,107,128,2,139,105,3,203,163,203,174,203,213,235,225,244,225,235,225,238,97,128,48,248,242,225,237,97,3,203,185,203,195,203,202,226,229,238,231,225,236,105,128,9,205,228,229,246,97,128,9,77,231,245,234,225,242,225,244,105,128,10,205,243,225,242,231,97,3,203,225,203,235,203,242,226,229,238,231,225,236,105,128,9,131,228,229,246,97,128,9,3,231,245,234,225,242,225,244,105,128,10,131,237,239,238,239,243,240,225,227,101,128,255,86,111,3,204,17,204,28,204,98,225,242,237,229,238,233,225,110,128,5,120,233,227,229,100,2,204,37,204,73,233,244,229,242,225,244,233,239,110,2,204,51,204,62,232,233,242,225,231,225,238,97,128,48,158,235,225,244,225,235,225,238,97,128,48,254,237,225,242,235,235,225,238,97,129,48,155,204,86,232,225,236,230,247,233,228,244,104,128,255,158,235,225,244,225,235,225,238,97,128,48,250,240,225,242,229,110,128,36,177,116,2,204,123,204,130,233,236,228,101,128,30,125,245,242,238,229,100,128,2,140,117,2,204,144,204,155,232,233,242,225,231,225,238,97,128,48,148,235,225,244,225,235,225,238,97,128,48,244,119,143,0,119,204,200,205,177,205,187,205,210,205,250,206,61,206,69,208,40,208,81,208,93,208,168,208,176,208,183,208,194,208,203,97,8,204,218,204,225,204,235,204,246,205,28,205,60,205,72,205,108,227,245,244,101,128,30,131,229,235,239,242,229,225,110,128,49,89,232,233,242,225,231,225,238,97,128,48,143,107,2,204,252,205,20,225,244,225,235,225,238,97,129,48,239,205,8,232,225,236,230,247,233,228,244,104,128,255,156,239,242,229,225,110,128,49,88,243,237,225,236,108,2,205,38,205,49,232,233,242,225,231,225,238,97,128,48,142,235,225,244,225,235,225,238,97,128,48,238,244,244,239,243,241,245,225,242,101,128,51,87,118,2,205,78,205,86,229,228,225,243,104,128,48,28,249,245,238,228,229,242,243,227,239,242,229,246,229,242,244,233,227,225,108,128,254,52,119,3,205,116,205,125,205,139,225,242,225,226,233,99,128,6,72,230,233,238,225,236,225,242,225,226,233,99,128,254,238,232,225,237,250,225,225,226,239,246,101,2,205,154,205,163,225,242,225,226,233,99,128,6,36,230,233,238,225,236,225,242,225,226,233,99,128,254,134,226,243,241,245,225,242,101,128,51,221,227,233,242,99,2,205,196,205,201,236,101,128,36,230,245,237,230,236,229,120,128,1,117,100,2,205,216,205,226,233,229,242,229,243,233,115,128,30,133,239,116,2,205,233,205,242,225,227,227,229,238,116,128,30,135,226,229,236,239,119,128,30,137,101,4,206,4,206,15,206,27,206,51,232,233,242,225,231,225,238,97,128,48,145,233,229,242,243,244,242,225,243,115,128,33,24,107,2,206,33,206,43,225,244,225,235,225,238,97,128,48,241,239,242,229,225,110,128,49,94,239,235,239,242,229,225,110,128,49,93,231,242,225,246,101,128,30,129,232,233,244,101,8,206,90,206,99,206,183,207,17,207,101,207,146,207,198,207,254,226,245,236,236,229,116,128,37,230,99,2,206,105,206,125,233,242,227,236,101,129,37,203,206,115,233,238,246,229,242,243,101,128,37,217,239,242,238,229,242,226,242,225,227,235,229,116,2,206,142,206,162,236,229,230,116,129,48,14,206,151,246,229,242,244,233,227,225,108,128,254,67,242,233,231,232,116,129,48,15,206,172,246,229,242,244,233,227,225,108,128,254,68,100,2,206,189,206,230,233,225,237,239,238,100,129,37,199,206,200,227,239,238,244,225,233,238,233,238,231,226,236,225,227,235,243,237,225,236,236,228,233,225,237,239,238,100,128,37,200,239,247,238,240,239,233,238,244,233,238,103,2,206,246,207,6,243,237,225,236,236,244,242,233,225,238,231,236,101,128,37,191,244,242,233,225,238,231,236,101,128,37,189,236,101,2,207,24,207,66,230,244,240,239,233,238,244,233,238,103,2,207,39,207,55,243,237,225,236,236,244,242,233,225,238,231,236,101,128,37,195,244,242,233,225,238,231,236,101,128,37,193,238,244,233,227,245,236,225,242,226,242,225,227,235,229,116,2,207,86,207,93,236,229,230,116,128,48,22,242,233,231,232,116,128,48,23,242,233,231,232,244,240,239,233,238,244,233,238,103,2,207,119,207,135,243,237,225,236,236,244,242,233,225,238,231,236,101,128,37,185,244,242,233,225,238,231,236,101,128,37,183,115,3,207,154,207,184,207,192,109,2,207,160,207,172,225,236,236,243,241,245,225,242,101,128,37,171,233,236,233,238,231,230,225,227,101,128,38,58,241,245,225,242,101,128,37,161,244,225,114,128,38,6,116,2,207,204,207,215,229,236,229,240,232,239,238,101,128,38,15,239,242,244,239,233,243,229,243,232,229,236,236,226,242,225,227,235,229,116,2,207,239,207,246,236,229,230,116,128,48,24,242,233,231,232,116,128,48,25,245,240,240,239,233,238,244,233,238,103,2,208,13,208,29,243,237,225,236,236,244,242,233,225,238,231,236,101,128,37,181,244,242,233,225,238,231,236,101,128,37,179,105,2,208,46,208,57,232,233,242,225,231,225,238,97,128,48,144,107,2,208,63,208,73,225,244,225,235,225,238,97,128,48,240,239,242,229,225,110,128,49,95,237,239,238,239,243,240,225,227,101,128,255,87,111,4,208,103,208,114,208,139,208,157,232,233,242,225,231,225,238,97,128,48,146,235,225,244,225,235,225,238,97,129,48,242,208,127,232,225,236,230,247,233,228,244,104,128,255,102,110,129,32,169,208,145,237,239,238,239,243,240,225,227,101,128,255,230,247,225,229,238,244,232,225,105,128,14,39,240,225,242,229,110,128,36,178,242,233,238,103,128,30,152,243,245,240,229,242,233,239,114,128,2,183,244,245,242,238,229,100,128,2,141,249,238,110,128,1,191,120,137,0,120,208,231,208,242,208,253,209,6,209,33,209,46,209,50,209,62,209,70,225,226,239,246,229,227,237,98,128,3,61,226,239,240,239,237,239,230,111,128,49,18,227,233,242,227,236,101,128,36,231,100,2,209,12,209,22,233,229,242,229,243,233,115,128,30,141,239,244,225,227,227,229,238,116,128,30,139,229,232,225,242,237,229,238,233,225,110,128,5,109,105,128,3,190,237,239,238,239,243,240,225,227,101,128,255,88,240,225,242,229,110,128,36,179,243,245,240,229,242,233,239,114,128,2,227,121,143,0,121,209,115,210,74,210,97,210,137,212,103,212,111,212,128,212,192,212,204,213,201,213,241,213,253,214,8,214,29,215,2,97,11,209,139,209,151,209,161,209,168,209,175,209,185,209,210,209,221,210,3,210,16,210,62,225,228,239,243,241,245,225,242,101,128,51,78,226,229,238,231,225,236,105,128,9,175,227,245,244,101,128,0,253,228,229,246,97,128,9,47,229,235,239,242,229,225,110,128,49,82,231,117,2,209,192,209,201,234,225,242,225,244,105,128,10,175,242,237,245,235,232,105,128,10,47,232,233,242,225,231,225,238,97,128,48,132,107,2,209,227,209,251,225,244,225,235,225,238,97,129,48,228,209,239,232,225,236,230,247,233,228,244,104,128,255,148,239,242,229,225,110,128,49,81,237,225,235,235,225,238,244,232,225,105,128,14,78,243,237,225,236,108,2,210,26,210,37,232,233,242,225,231,225,238,97,128,48,131,235,225,244,225,235,225,238,97,129,48,227,210,50,232,225,236,230,247,233,228,244,104,128,255,108,244,227,249,242,233,236,236,233,99,128,4,99,227,233,242,99,2,210,83,210,88,236,101,128,36,232,245,237,230,236,229,120,128,1,119,100,2,210,103,210,113,233,229,242,229,243,233,115,128,0,255,239,116,2,210,120,210,129,225,227,227,229,238,116,128,30,143,226,229,236,239,119,128,30,245,101,7,210,153,211,161,211,170,211,188,211,220,212,40,212,91,104,8,210,171,210,180,210,214,210,228,211,45,211,61,211,120,211,138,225,242,225,226,233,99,128,6,74,226,225,242,242,229,101,2,210,191,210,200,225,242,225,226,233,99,128,6,210,230,233,238,225,236,225,242,225,226,233,99,128,251,175,230,233,238,225,236,225,242,225,226,233,99,128,254,242,232,225,237,250,225,225,226,239,246,101,4,210,247,211,0,211,14,211,30,225,242,225,226,233,99,128,6,38,230,233,238,225,236,225,242,225,226,233,99,128,254,138,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,139,237,229,228,233,225,236,225,242,225,226,233,99,128,254,140,233,238,233,244,233,225,236,225,242,225,226,233,99,128,254,243,237,101,2,211,68,211,81,228,233,225,236,225,242,225,226,233,99,128,254,244,229,237,105,2,211,89,211,104,238,233,244,233,225,236,225,242,225,226,233,99,128,252,221,243,239,236,225,244,229,228,225,242,225,226,233,99,128,252,88,238,239,239,238,230,233,238,225,236,225,242,225,226,233,99,128,252,148,244,232,242,229,229,228,239,244,243,226,229,236,239,247,225,242,225,226,233,99,128,6,209,235,239,242,229,225,110,128,49,86,110,129,0,165,211,176,237,239,238,239,243,240,225,227,101,128,255,229,111,2,211,194,211,203,235,239,242,229,225,110,128,49,85,242,233,238,232,233,229,245,232,235,239,242,229,225,110,128,49,134,114,3,211,228,212,8,212,20,225,232,226,229,238,249,239,237,111,2,211,242,211,251,232,229,226,242,229,119,128,5,170,236,229,230,244,232,229,226,242,229,119,128,5,170,233,227,249,242,233,236,236,233,99,128,4,75,245,228,233,229,242,229,243,233,243,227,249,242,233,236,236,233,99,128,4,249,243,233,229,245,238,103,3,212,53,212,62,212,78,235,239,242,229,225,110,128,49,129,240,225,238,243,233,239,243,235,239,242,229,225,110,128,49,131,243,233,239,243,235,239,242,229,225,110,128,49,130,244,233,246,232,229,226,242,229,119,128,5,154,231,242,225,246,101,128,30,243,232,239,239,107,129,1,180,212,120,225,226,239,246,101,128,30,247,105,5,212,140,212,151,212,162,212,171,212,179,225,242,237,229,238,233,225,110,128,5,117,227,249,242,233,236,236,233,99,128,4,87,235,239,242,229,225,110,128,49,98,238,249,225,238,103,128,38,47,247,238,225,242,237,229,238,233,225,110,128,5,130,237,239,238,239,243,240,225,227,101,128,255,89,111,7,212,220,213,34,213,45,213,55,213,93,213,139,213,148,100,131,5,217,212,230,212,250,213,3,228,225,231,229,243,104,129,251,57,212,241,232,229,226,242,229,119,128,251,57,232,229,226,242,229,119,128,5,217,249,239,100,2,213,11,213,20,232,229,226,242,229,119,128,5,242,240,225,244,225,232,232,229,226,242,229,119,128,251,31,232,233,242,225,231,225,238,97,128,48,136,233,235,239,242,229,225,110,128,49,137,107,2,213,61,213,85,225,244,225,235,225,238,97,129,48,232,213,73,232,225,236,230,247,233,228,244,104,128,255,150,239,242,229,225,110,128,49,91,243,237,225,236,108,2,213,103,213,114,232,233,242,225,231,225,238,97,128,48,135,235,225,244,225,235,225,238,97,129,48,231,213,127,232,225,236,230,247,233,228,244,104,128,255,110,244,231,242,229,229,107,128,3,243,121,2,213,154,213,191,97,2,213,160,213,170,229,235,239,242,229,225,110,128,49,136,107,2,213,176,213,184,239,242,229,225,110,128,49,135,244,232,225,105,128,14,34,233,238,231,244,232,225,105,128,14,13,112,2,213,207,213,214,225,242,229,110,128,36,180,239,231,229,231,242,225,237,237,229,238,105,129,3,122,213,230,231,242,229,229,235,227,237,98,128,3,69,114,129,1,166,213,247,233,238,103,128,30,153,243,245,240,229,242,233,239,114,128,2,184,116,2,214,14,214,21,233,236,228,101,128,30,249,245,242,238,229,100,128,2,142,117,5,214,41,214,52,214,62,214,100,214,232,232,233,242,225,231,225,238,97,128,48,134,233,235,239,242,229,225,110,128,49,140,107,2,214,68,214,92,225,244,225,235,225,238,97,129,48,230,214,80,232,225,236,230,247,233,228,244,104,128,255,149,239,242,229,225,110,128,49,96,115,3,214,108,214,146,214,187,226,233,103,2,214,116,214,127,227,249,242,233,236,236,233,99,128,4,107,233,239,244,233,230,233,229,228,227,249,242,233,236,236,233,99,128,4,109,236,233,244,244,236,101,2,214,157,214,168,227,249,242,233,236,236,233,99,128,4,103,233,239,244,233,230,233,229,228,227,249,242,233,236,236,233,99,128,4,105,237,225,236,108,2,214,196,214,207,232,233,242,225,231,225,238,97,128,48,133,235,225,244,225,235,225,238,97,129,48,229,214,220,232,225,236,230,247,233,228,244,104,128,255,109,249,101,2,214,239,214,248,235,239,242,229,225,110,128,49,139,239,235,239,242,229,225,110,128,49,138,249,97,2,215,9,215,19,226,229,238,231,225,236,105,128,9,223,228,229,246,97,128,9,95,122,142,0,122,215,58,216,66,216,77,216,120,216,147,217,182,218,34,218,76,218,88,218,100,218,128,218,136,218,152,218,161,97,10,215,80,215,91,215,98,215,105,215,116,215,194,215,224,215,235,216,15,216,27,225,242,237,229,238,233,225,110,128,5,102,227,245,244,101,128,1,122,228,229,246,97,128,9,91,231,245,242,237,245,235,232,105,128,10,91,104,4,215,126,215,135,215,149,215,179,225,242,225,226,233,99,128,6,56,230,233,238,225,236,225,242,225,226,233,99,128,254,198,105,2,215,155,215,170,238,233,244,233,225,236,225,242,225,226,233,99,128,254,199,242,225,231,225,238,97,128,48,86,237,229,228,233,225,236,225,242,225,226,233,99,128,254,200,233,110,2,215,201,215,210,225,242,225,226,233,99,128,6,50,230,233,238,225,236,225,242,225,226,233,99,128,254,176,235,225,244,225,235,225,238,97,128,48,182,241,229,102,2,215,243,216,1,231,225,228,239,236,232,229,226,242,229,119,128,5,149,241,225,244,225,238,232,229,226,242,229,119,128,5,148,242,241,225,232,229,226,242,229,119,128,5,152,249,233,110,130,5,214,216,37,216,57,228,225,231,229,243,104,129,251,54,216,48,232,229,226,242,229,119,128,251,54,232,229,226,242,229,119,128,5,214,226,239,240,239,237,239,230,111,128,49,23,99,3,216,85,216,92,216,114,225,242,239,110,128,1,126,233,242,99,2,216,100,216,105,236,101,128,36,233,245,237,230,236,229,120,128,30,145,245,242,108,128,2,145,228,239,116,130,1,124,216,130,216,139,225,227,227,229,238,116,128,1,124,226,229,236,239,119,128,30,147,101,6,216,161,216,172,216,215,216,226,216,237,217,177,227,249,242,233,236,236,233,99,128,4,55,100,2,216,178,216,197,229,243,227,229,238,228,229,242,227,249,242,233,236,236,233,99,128,4,153,233,229,242,229,243,233,243,227,249,242,233,236,236,233,99,128,4,223,232,233,242,225,231,225,238,97,128,48,92,235,225,244,225,235,225,238,97,128,48,188,242,111,140,0,48,217,10,217,19,217,29,217,36,217,61,217,74,217,85,217,97,217,108,217,118,217,129,217,136,225,242,225,226,233,99,128,6,96,226,229,238,231,225,236,105,128,9,230,228,229,246,97,128,9,102,231,117,2,217,43,217,52,234,225,242,225,244,105,128,10,230,242,237,245,235,232,105,128,10,102,232,225,227,235,225,242,225,226,233,99,128,6,96,233,238,230,229,242,233,239,114,128,32,128,237,239,238,239,243,240,225,227,101,128,255,16,239,236,228,243,244,249,236,101,128,247,48,240,229,242,243,233,225,110,128,6,240,243,245,240,229,242,233,239,114,128,32,112,244,232,225,105,128,14,80,247,233,228,244,104,3,217,148,217,157,217,169,234,239,233,238,229,114,128,254,255,238,239,238,234,239,233,238,229,114,128,32,12,243,240,225,227,101,128,32,11,244,97,128,3,182,104,2,217,188,217,199,226,239,240,239,237,239,230,111,128,49,19,101,4,217,209,217,220,217,236,217,247,225,242,237,229,238,233,225,110,128,5,106,226,242,229,246,229,227,249,242,233,236,236,233,99,128,4,194,227,249,242,233,236,236,233,99,128,4,54,100,2,217,253,218,16,229,243,227,229,238,228,229,242,227,249,242,233,236,236,233,99,128,4,151,233,229,242,229,243,233,243,227,249,242,233,236,236,233,99,128,4,221,105,3,218,42,218,53,218,64,232,233,242,225,231,225,238,97,128,48,88,235,225,244,225,235,225,238,97,128,48,184,238,239,242,232,229,226,242,229,119,128,5,174,236,233,238,229,226,229,236,239,119,128,30,149,237,239,238,239,243,240,225,227,101,128,255,90,111,2,218,106,218,117,232,233,242,225,231,225,238,97,128,48,94,235,225,244,225,235,225,238,97,128,48,190,240,225,242,229,110,128,36,181,242,229,244,242,239,230,236,229,248,232,239,239,107,128,2,144,243,244,242,239,235,101,128,1,182,117,2,218,167,218,178,232,233,242,225,231,225,238,97,128,48,90,235,225,244,225,235,225,238,97,128,48,186,68,101,108,116,97,0,79,109,101,103,97,0,102,114,97,99,116,105,111,110,0,104,121,112,104,101,110,0,109,97,99,114,111,110,0,109,117,0,112,101,114,105,111,100,99,101,110,116,101,114,101,100,0,115,112,97,99,101,0,84,99,111,109,109,97,97,99,99,101,110,116,0,116,99,111,109,109,97,97,99,99,101,110,116,0,46,110,117,108,108,0,110,111,110,109,97,114,107,105,110,103,114,101,116,117,114,110,0,110,111,116,101,113,117,97,108,0,105,110,102,105,110,105,116,121,0,108,101,115,115,101,113,117,97,108,0,103,114,101,97,116,101,114,101,113,117,97,108,0,112,97,114,116,105,97,108,100,105,102,102,0,115,117,109,109,97,116,105,111,110,0,112,114,111,100,117,99,116,0,112,105,0,105,110,116,101,103,114,97,108,0,79,109,101,103,97,0,114,97,100,105,99,97,108,0,97,112,112,114,111,120,101,113,117,97,108,0,68,101,108,116,97,0,110,111,110,98,114,101,97,107,105,110,103,115,112,97,99,101,0,108,111,122,101,110,103,101,0,97,112,112,108,101,0,102,114,97,110,99,0,71,98,114,101,118,101,0,103,98,114,101,118,101,0,73,100,111,116,97,99,99,101,110,116,0,83,99,101,100,105,108,108,97,0,115,99,101,100,105,108,108,97,0,67,97,99,117,116,101,0,99,97,99,117,116,101,0,67,99,97,114,111,110,0,99,99,97,114,111,110,0,100,99,114,111,97,116,0,46,110,111,116,100,101,102,0,115,112,97,99,101,0,101,120,99,108,97,109,0,113,117,111,116,101,100,98,108,0,110,117,109,98,101,114,115,105,103,110,0,100,111,108,108,97,114,0,112,101,114,99,101,110,116,0,97,109,112,101,114,115,97,110,100,0,113,117,111,116,101,114,105,103,104,116,0,112,97,114,101,110,108,101,102,116,0,112,97,114,101,110,114,105,103,104,116,0,97,115,116,101,114,105,115,107,0,112,108,117,115,0,99,111,109,109,97,0,104,121,112,104,101,110,0,112,101,114,105,111,100,0,115,108,97,115,104,0,122,101,114,111,0,111,110,101,0,116,119,111,0,116,104,114,101,101,0,102,111,117,114,0,102,105,118,101,0,115,105,120,0,115,101,118,101,110,0,101,105,103,104,116,0,110,105,110,101,0,99,111,108,111,110,0,115,101,109,105,99,111,108,111,110,0,108,101,115,115,0,101,113,117,97,108,0,103,114,101,97,116,101,114,0,113,117,101,115,116,105,111,110,0,97,116,0,65,0,66,0,67,0,68,0,69,0,70,0,71,0,72,0,73,0,74,0,75,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,88,0,89,0,90,0,98,114,97,99,107,101,116,108,101,102,116,0,98,97,99,107,115,108,97,115,104,0,98,114,97,99,107,101,116,114,105,103,104,116,0,97,115,99,105,105,99,105,114,99,117,109,0,117,110,100,101,114,115,99,111,114,101,0,113,117,111,116,101,108,101,102,116,0,97,0,98,0,99,0,100,0,101,0,102,0,103,0,104,0,105,0,106,0,107,0,108,0,109,0,110,0,111,0,112,0,113,0,114,0,115,0,116,0,117,0,118,0,119,0,120,0,121,0,122,0,98,114,97,99,101,108,101,102,116,0,98,97,114,0,98,114,97,99,101,114,105,103,104,116,0,97,115,99,105,105,116,105,108,100,101,0,101,120,99,108,97,109,100,111,119,110,0,99,101,110,116,0,115,116,101,114,108,105,110,103,0,102,114,97,99,116,105,111,110,0,121,101,110,0,102,108,111,114,105,110,0,115,101,99,116,105,111,110,0,99,117,114,114,101,110,99,121,0,113,117,111,116,101,115,105,110,103,108,101,0,113,117,111,116,101,100,98,108,108,101,102,116,0,103,117,105,108,108,101,109,111,116,108,101,102,116,0,103,117,105,108,115,105,110,103,108,108,101,102,116,0,103,117,105,108,115,105,110,103,108,114,105,103,104,116,0,102,105,0,102,108,0,101,110,100,97,115,104,0,100,97,103,103,101,114,0,100,97,103,103,101,114,100,98,108,0,112,101,114,105,111,100,99,101,110,116,101,114,101,100,0,112,97,114,97,103,114,97,112,104,0,98,117,108,108,101,116,0,113,117,111,116,101,115,105,110,103,108,98,97,115,101,0,113,117,111,116,101,100,98,108,98,97,115,101,0,113,117,111,116,101,100,98,108,114,105,103,104,116,0,103,117,105,108,108,101,109,111,116,114,105,103,104,116,0,101,108,108,105,112,115,105,115,0,112,101,114,116,104,111,117,115,97,110,100,0,113,117,101,115,116,105,111,110,100,111,119,110,0,103,114,97,118,101,0,97,99,117,116,101,0,99,105,114,99,117,109,102,108,101,120,0,116,105,108,100,101,0,109,97,99,114,111,110,0,98,114,101,118,101,0,100,111,116,97,99,99,101,110,116,0,100,105,101,114,101,115,105,115,0,114,105,110,103,0,99,101,100,105,108,108,97,0,104,117,110,103,97,114,117,109,108,97,117,116,0,111,103,111,110,101,107,0,99,97,114,111,110,0,101,109,100,97,115,104,0,65,69,0,111,114,100,102,101,109,105,110,105,110,101,0,76,115,108,97,115,104,0,79,115,108,97,115,104,0,79,69,0,111,114,100,109,97,115,99,117,108,105,110,101,0,97,101,0,100,111,116,108,101,115,115,105,0,108,115,108,97,115,104,0,111,115,108,97,115,104,0,111,101,0,103,101,114,109,97,110,100,98,108,115,0,111,110,101,115,117,112,101,114,105,111,114,0,108,111,103,105,99,97,108,110,111,116,0,109,117,0,116,114,97,100,101,109,97,114,107,0,69,116,104,0,111,110,101,104,97,108,102,0,112,108,117,115,109,105,110,117,115,0,84,104,111,114,110,0,111,110,101,113,117,97,114,116,101,114,0,100,105,118,105,100,101,0,98,114,111,107,101,110,98,97,114,0,100,101,103,114,101,101,0,116,104,111,114,110,0,116,104,114,101,101,113,117,97,114,116,101,114,115,0,116,119,111,115,117,112,101,114,105,111,114,0,114,101,103,105,115,116,101,114,101,100,0,109,105,110,117,115,0,101,116,104,0,109,117,108,116,105,112,108,121,0,116,104,114,101,101,115,117,112,101,114,105,111,114,0,99,111,112,121,114,105,103,104,116,0,65,97,99,117,116,101,0,65,99,105,114,99,117,109,102,108,101,120,0,65,100,105,101,114,101,115,105,115,0,65,103,114,97,118,101,0,65,114,105,110,103,0,65,116,105,108,100,101,0,67,99,101,100,105,108,108,97,0,69,97,99,117,116,101,0,69,99,105,114,99,117,109,102,108,101,120,0,69,100,105,101,114,101,115,105,115,0,69,103,114,97,118,101,0,73,97,99,117,116,101,0,73,99,105,114,99,117,109,102,108,101,120,0,73,100,105,101,114,101,115,105,115,0,73,103,114,97,118,101,0,78,116,105,108,100,101,0,79,97,99,117,116,101,0,79,99,105,114,99,117,109,102,108,101,120,0,79,100,105,101,114,101,115,105,115,0,79,103,114,97,118,101,0,79,116,105,108,100,101,0,83,99,97,114,111,110,0,85,97,99,117,116,101,0,85,99,105,114,99,117,109,102,108,101,120,0,85,100,105,101,114,101,115,105,115,0,85,103,114,97,118,101,0,89,97,99,117,116,101,0,89,100,105,101,114,101,115,105,115,0,90,99,97,114,111,110,0,97,97,99,117,116,101,0,97,99,105,114,99,117,109,102,108,101,120,0,97,100,105,101,114,101,115,105,115,0,97,103,114,97,118,101,0,97,114,105,110,103,0,97,116,105,108,100,101,0,99,99,101,100,105,108,108,97,0,101,97,99,117,116,101,0,101,99,105,114,99,117,109,102,108,101,120,0,101,100,105,101,114,101,115,105,115,0,101,103,114,97,118,101,0,105,97,99,117,116,101,0,105,99,105,114,99,117,109,102,108,101,120,0,105,100,105,101,114,101,115,105,115,0,105,103,114,97,118,101,0,110,116,105,108,100,101,0,111,97,99,117,116,101,0,111,99,105,114,99,117,109,102,108,101,120,0,111,100,105,101,114,101,115,105,115,0,111,103,114,97,118,101,0,111,116,105,108,100,101,0,115,99,97,114,111,110,0,117,97,99,117,116,101,0,117,99,105,114,99,117,109,102,108,101,120,0,117,100,105,101,114,101,115,105,115,0,117,103,114,97,118,101,0,121,97,99,117,116,101,0,121,100,105,101,114,101,115,105,115,0,122,99,97,114,111,110,0,101,120,99,108,97,109,115,109,97,108,108,0,72,117,110,103,97,114,117,109,108,97,117,116,115,109,97,108,108,0,100,111,108,108,97,114,111,108,100,115,116,121,108,101,0,100,111,108,108,97,114,115,117,112,101,114,105,111,114,0,97,109,112,101,114,115,97,110,100,115,109,97,108,108,0,65,99,117,116,101,115,109,97,108,108,0,112,97,114,101,110,108,101,102,116,115,117,112,101,114,105,111,114,0,112,97,114,101,110,114,105,103,104,116,115,117,112,101,114,105,111,114,0,116,119,111,100,111,116,101,110,108,101,97,100,101,114,0,111,110,101,100,111,116,101,110,108,101,97,100,101,114,0,122,101,114,111,111,108,100,115,116,121,108,101,0,111,110,101,111,108,100,115,116,121,108,101,0,116,119,111,111,108,100,115,116,121,108,101,0,116,104,114,101,101,111,108,100,115,116,121,108,101,0,102,111,117,114,111,108,100,115,116,121,108,101,0,102,105,118,101,111,108,100,115,116,121,108,101,0,115,105,120,111,108,100,115,116,121,108,101,0,115,101,118,101,110,111,108,100,115,116,121,108,101,0,101,105,103,104,116,111,108,100,115,116,121,108,101,0,110,105,110,101,111,108,100,115,116,121,108,101,0,99,111,109,109,97,115,117,112,101,114,105,111,114,0,116,104,114,101,101,113,117,97,114,116,101,114,115,101,109,100,97,115,104,0,112,101,114,105,111,100,115,117,112,101,114,105,111,114,0,113,117,101,115,116,105,111,110,115,109,97,108,108,0,97,115,117,112,101,114,105,111,114,0,98,115,117,112,101,114,105,111,114,0,99,101,110,116,115,117,112,101,114,105,111,114,0,100,115,117,112,101,114,105,111,114,0,101,115,117,112,101,114,105,111,114,0,105,115,117,112,101,114,105,111,114,0,108,115,117,112,101,114,105,111,114,0,109,115,117,112,101,114,105,111,114,0,110,115,117,112,101,114,105,111,114,0,111,115,117,112,101,114,105,111,114,0,114,115,117,112,101,114,105,111,114,0,115,115,117,112,101,114,105,111,114,0,116,115,117,112,101,114,105,111,114,0,102,102,0,102,102,105,0,102,102,108,0,112,97,114,101,110,108,101,102,116,105,110,102,101,114,105,111,114,0,112,97,114,101,110,114,105,103,104,116,105,110,102,101,114,105,111,114,0,67,105,114,99,117,109,102,108,101,120,115,109,97,108,108,0,104,121,112,104,101,110,115,117,112,101,114,105,111,114,0,71,114,97,118,101,115,109,97,108,108,0,65,115,109,97,108,108,0,66,115,109,97,108,108,0,67,115,109,97,108,108,0,68,115,109,97,108,108,0,69,115,109,97,108,108,0,70,115,109,97,108,108,0,71,115,109,97,108,108,0,72,115,109,97,108,108,0,73,115,109,97,108,108,0,74,115,109,97,108,108,0,75,115,109,97,108,108,0,76,115,109,97,108,108,0,77,115,109,97,108,108,0,78,115,109,97,108,108,0,79,115,109,97,108,108,0,80,115,109,97,108,108,0,81,115,109,97,108,108,0,82,115,109,97,108,108,0,83,115,109,97,108,108,0,84,115,109,97,108,108,0,85,115,109,97,108,108,0,86,115,109,97,108,108,0,87,115,109,97,108,108,0,88,115,109,97,108,108,0,89,115,109,97,108,108,0,90,115,109,97,108,108,0,99,111,108,111,110,109,111,110,101,116,97,114,121,0,111,110,101,102,105,116,116,101,100,0,114,117,112,105,97,104,0,84,105,108,100,101,115,109,97,108,108,0,101,120,99,108,97,109,100,111,119,110,115,109,97,108,108,0,99,101,110,116,111,108,100,115,116,121,108,101,0,76,115,108,97,115,104,115,109,97,108,108,0,83,99,97,114,111,110,115,109,97,108,108,0,90,99,97,114,111,110,115,109,97,108,108,0,68,105,101,114,101,115,105,115,115,109,97,108,108,0,66,114,101,118,101,115,109,97,108,108,0,67,97,114,111,110,115,109,97,108,108,0,68,111,116,97,99,99,101,110,116,115,109,97,108,108,0,77,97,99,114,111,110,115,109,97,108,108,0,102,105,103,117,114,101,100,97,115,104,0,104,121,112,104,101,110,105,110,102,101,114,105,111,114,0,79,103,111,110,101,107,115,109,97,108,108,0,82,105,110,103,115,109,97,108,108,0,67,101,100,105,108,108,97,115,109,97,108,108,0,113,117,101,115,116,105,111,110,100,111,119,110,115,109,97,108,108,0,111,110,101,101,105,103,104,116,104,0,116,104,114,101,101,101,105,103,104,116,104,115,0,102,105,118,101,101,105,103,104,116,104,115,0,115,101,118,101,110,101,105,103,104,116,104,115,0,111,110,101,116,104,105,114,100,0,116,119,111,116,104,105,114,100,115,0,122,101,114,111,115,117,112,101,114,105,111,114,0,102,111,117,114,115,117,112,101,114,105,111,114,0,102,105,118,101,115,117,112,101,114,105,111,114,0,115,105,120,115,117,112,101,114,105,111,114,0,115,101,118,101,110,115,117,112,101,114,105,111,114,0,101,105,103,104,116,115,117,112,101,114,105,111,114,0,110,105,110,101,115,117,112,101,114,105,111,114,0,122,101,114,111,105,110,102,101,114,105,111,114,0,111,110,101,105,110,102,101,114,105,111,114,0,116,119,111,105,110,102,101,114,105,111,114,0,116,104,114,101,101,105,110,102,101,114,105,111,114,0,102,111,117,114,105,110,102,101,114,105,111,114,0,102,105,118,101,105,110,102,101,114,105,111,114,0,115,105,120,105,110,102,101,114,105,111,114,0,115,101,118,101,110,105,110,102,101,114,105,111,114,0,101,105,103,104,116,105,110,102,101,114,105,111,114,0,110,105,110,101,105,110,102,101,114,105,111,114,0,99,101,110,116,105,110,102,101,114,105,111,114,0,100,111,108,108,97,114,105,110,102,101,114,105,111,114,0,112,101,114,105,111,100,105,110,102,101,114,105,111,114,0,99,111,109,109,97,105,110,102,101,114,105,111,114,0,65,103,114,97,118,101,115,109,97,108,108,0,65,97,99,117,116,101,115,109,97,108,108,0,65,99,105,114,99,117,109,102,108,101,120,115,109,97,108,108,0,65,116,105,108,100,101,115,109,97,108,108,0,65,100,105,101,114,101,115,105,115,115,109,97,108,108,0,65,114,105,110,103,115,109,97,108,108,0,65,69,115,109,97,108,108,0,67,99,101,100,105,108,108,97,115,109,97,108,108,0,69,103,114,97,118,101,115,109,97,108,108,0,69,97,99,117,116,101,115,109,97,108,108,0,69,99,105,114,99,117,109,102,108,101,120,115,109,97,108,108,0,69,100,105,101,114,101,115,105,115,115,109,97,108,108,0,73,103,114,97,118,101,115,109,97,108,108,0,73,97,99,117,116,101,115,109,97,108,108,0,73,99,105,114,99,117,109,102,108,101,120,115,109,97,108,108,0,73,100,105,101,114,101,115,105,115,115,109,97,108,108,0,69,116,104,115,109,97,108,108,0,78,116,105,108,100,101,115,109,97,108,108,0,79,103,114,97,118,101,115,109,97,108,108,0,79,97,99,117,116,101,115,109,97,108,108,0,79,99,105,114,99,117,109,102,108,101,120,115,109,97,108,108,0,79,116,105,108,100,101,115,109,97,108,108,0,79,100,105,101,114,101,115,105,115,115,109,97,108,108,0,79,69,115,109,97,108,108,0,79,115,108,97,115,104,115,109,97,108,108,0,85,103,114,97,118,101,115,109,97,108,108,0,85,97,99,117,116,101,115,109,97,108,108,0,85,99,105,114,99,117,109,102,108,101,120,115,109,97,108,108,0,85,100,105,101,114,101,115,105,115,115,109,97,108,108,0,89,97,99,117,116,101,115,109,97,108,108,0,84,104,111,114,110,115,109,97,108,108,0,89,100,105,101,114,101,115,105,115,115,109,97,108,108,0,48,48,49,46,48,48,48,0,48,48,49,46,48,48,49,0,48,48,49,46,48,48,50,0,48,48,49,46,48,48,51,0,66,108,97,99,107,0,66,111,108,100,0,66,111,111,107,0,76,105,103,104,116,0,77,101,100,105,117,109,0,82,101,103,117,108,97,114,0,82,111,109,97,110,0,83,101,109,105,98,111,108,100,0,114,97,115,116,101,114,49,0,115,102,110,116,45,116,97,98,108,101,0,98,100,102,0,67,72,65,82,83,69,84,95,82,69,71,73,83,84,82,89,0,67,72,65,82,83,69,84,95,69,78,67,79,68,73,78,71,0,116,116,45,99,109,97,112,115,0,115,109,111,111,116,104,0,115,109,111,111,116,104,45,108,99,100,0,115,109,111,111,116,104,45,108,99,100,118,0,84,114,117,101,84,121,112,101,0,83,108,97,110,116,0,116,114,117,101,116,121,112,101,45,101,110,103,105,110,101,0,116,116,45,103,108,121,102,0,105,110,116,101,114,112,114,101,116,101,114,45,118,101,114,115,105,111,110,0,115,102,110,116,0,68,70,75,97,105,83,104,111,45,83,66,0,0,0,0,0,0,0,0,0,68,70,75,97,105,83,104,117,0,0,0,0,0,0,0,0,0,0,0,0,68,70,75,97,105,45,83,66,0,0,0,0,0,0,0,0,0,0,0,0,72,117,97,84,105,97,110,75,97,105,84,105,63,0,0,0,0,0,0,0,72,117,97,84,105,97,110,83,111,110,103,84,105,63,0,0,0,0,0,0,77,105,110,103,40,102,111,114,32,73,83,79,49,48,54,52,54,41,0,0,77,105,110,103,76,105,85,0,0,0,0,0,0,0,0,0,0,0,0,0,80,77,105,110,103,76,105,85,0,0,0,0,0,0,0,0,0,0,0,0,77,105,110,103,76,105,52,51,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,255,254,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,2,3,4,5,6,7,8,9,3,5,7,9,11,13,15,17,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,32,32,32,32,32,32,2,2,0,80,16,16,16,16,16,16,16,16,0,0,16,0,16,16,16,16,18,16,0,34,1,17,16,32,0,16,32,16,16,0,16,16,0,0,0,0,16,16,16,16,16,0,32,32,0,0,32,32,0,0,32,17,32,17,17,17,32,33,33,1,1,0,0,16,33,33,33,33,33,33,17,17,16,0,33,33,17,16,16,16,33,33,33,33,17,17,17,17,17,17,17,17,17,17,17,17,32,16,16,16,16,16,16,16,32,32,0,0,0,0,16,16,0,32,32,0,0,16,32,32,17,16,51,33,33,16,32,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,2,3,4,5,6,7,8,1,2,3,4,5,6,7,8,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,116,121,112,101,49,0,84,121,112,101,32,49,0,107,101,114,110,105,110,103,0,109,117,108,116,105,45,109,97,115,116,101,114,115,0,87,105,100,116,104,0,79,112,116,105,99,97,108,83,105,122,101,0,112,115,104,105,110,116,101,114,0,37,33,80,83,45,65,100,111,98,101,70,111,110,116,0,37,33,70,111,110,116,84,121,112,101,0,101,101,120,101,99,0,99,108,111,115,101,102,105,108,101,0,85,110,105,113,117,101,73,68,0,108,101,110,73,86,0,76,97,110,103,117,97,103,101,71,114,111,117,112,0,112,97,115,115,119,111,114,100,0,66,108,117,101,83,99,97,108,101,0,66,108,117,101,83,104,105,102,116,0,66,108,117,101,70,117,122,122,0,66,108,117,101,86,97,108,117,101,115,0,79,116,104,101,114,66,108,117,101,115,0,70,97,109,105,108,121,66,108,117,101,115,0,70,97,109,105,108,121,79,116,104,101,114,66,108,117,101,115,0,83,116,100,72,87,0,83,116,100,86,87,0,77,105,110,70,101,97,116,117,114,101,0,83,116,101,109,83,110,97,112,72,0,83,116,101,109,83,110,97,112,86,0,69,120,112,97,110,115,105,111,110,70,97,99,116,111,114,0,70,111,114,99,101,66,111,108,100,0,78,68,86,0,67,68,86,0,68,101,115,105,103,110,86,101,99,116,111,114,0,83,117,98,114,115,0,100,117,112,0,112,117,116,0,139,247,225,13,14,80,114,105,118,97,116,101,0,66,108,101,110,100,68,101,115,105,103,110,80,111,115,105,116,105,111,110,115,0,66,108,101,110,100,68,101,115,105,103,110,77,97,112,0,66,108,101,110,100,65,120,105,115,84,121,112,101,115,0,87,101,105,103,104,116,86,101,99,116,111,114,0,66,117,105,108,100,67,104,97,114,65,114,114,97,121,0,66,108,97,99,107,0,116,121,112,101,52,50,0,116,114,117,101,116,121,112,101,0,103,108,121,112,104,45,100,105,99,116,0,112,111,115,116,115,99,114,105,112,116,45,102,111,110,116,45,110,97,109,101,0,112,111,115,116,115,99,114,105,112,116,45,105,110,102,111,0,84,121,112,101,32,52,50,0,112,111,115,116,115,99,114,105,112,116,45,99,109,97,112,115,0,112,115,97,117,120,0,37,33,80,83,45,84,114,117,101,84,121,112,101,70,111,110,116,0,70,111,110,116,68,105,114,101,99,116,111,114,121,0,107,110,111,119,110,0,118,101,114,115,105,111,110,0,78,111,116,105,99,101,0,70,117,108,108,78,97,109,101,0,70,97,109,105,108,121,78,97,109,101,0,87,101,105,103,104,116,0,73,116,97,108,105,99,65,110,103,108,101,0,105,115,70,105,120,101,100,80,105,116,99,104,0,85,110,100,101,114,108,105,110,101,80,111,115,105,116,105,111,110,0,85,110,100,101,114,108,105,110,101,84,104,105,99,107,110,101,115,115,0,70,83,84,121,112,101,0,70,111,110,116,78,97,109,101,0,80,97,105,110,116,84,121,112,101,0,70,111,110,116,84,121,112,101,0,83,116,114,111,107,101,87,105,100,116,104,0,70,111,110,116,66,66,111,120,0,70,111,110,116,77,97,116,114,105,120,0,69,110,99,111,100,105,110,103,0,46,110,111,116,100,101,102,0,83,116,97,110,100,97,114,100,69,110,99,111,100,105,110,103,0,69,120,112,101,114,116,69,110,99,111,100,105,110,103,0,73,83,79,76,97,116,105,110,49,69,110,99,111,100,105,110,103,0,67,104,97,114,83,116,114,105,110,103,115,0,115,102,110,116,115,0,119,105,110,102,111,110,116,115,0,102,111,110,116,45,102,111,114,109,97,116,0,87,105,110,100,111,119,115,32,70,78,84,0,82,101,103,117,108,97,114,0,66,111,108,100,32,73,116,97,108,105,99,0,66,111,108,100,0,73,116,97,108,105,99,0,216,167,216,165,217,132,217,131,216,183,216,184,0,216,170,216,171,216,183,216,184,217,131,0,208,145,208,146,208,149,208,159,208,151,208,158,208,161,208,173,0,208,145,208,146,208,149,208,168,208,151,208,158,208,161,208,173,0,209,133,208,191,208,189,209,136,208,181,208,183,208,190,209,129,0,209,128,209,131,209,132,0,224,164,149,224,164,174,224,164,133,224,164,134,224,164,165,224,164,167,224,164,173,224,164,182,0,224,164,136,224,164,144,224,164,147,224,164,148,224,164,191,224,165,128,224,165,139,224,165,140,0,224,164,149,224,164,174,224,164,133,224,164,134,224,164,165,224,164,167,224,164,173,224,164,182,0,224,165,129,224,165,131,0,206,147,206,146,206,149,206,150,206,152,206,159,206,169,0,206,146,206,148,206,150,206,158,206,152,206,159,0,206,178,206,184,206,180,206,182,206,187,206,190,0,206,177,206,181,206,185,206,191,207,128,207,131,207,132,207,137,0,206,178,206,179,206,183,206,188,207,129,207,134,207,135,207,136,0,215,145,215,147,215,148,215,151,215,154,215,155,215,157,215,161,0,215,145,215,152,215,155,215,157,215,161,215,166,0,215,167,215,154,215,159,215,163,215,165,0,84,72,69,90,79,67,81,83,0,72,69,90,76,79,67,85,83,0,102,105,106,107,100,98,104,0,120,122,114,111,101,115,99,0,112,113,103,106,121,0,224,176,135,224,176,140,224,176,153,224,176,158,224,176,163,224,176,177,224,177,175,0,224,176,133,224,176,149,224,176,154,224,176,176,224,176,189,224,177,168,224,177,172,0,224,184,154,224,185,128,224,185,129,224,184,173,224,184,129,224,184,178,0,224,184,154,224,184,155,224,184,169,224,184,175,224,184,173,224,184,162,224,184,174,0,224,184,155,224,184,157,224,184,159,0,224,185,130,224,185,131,224,185,132,0,224,184,142,224,184,143,224,184,164,224,184,166,0,224,184,141,224,184,144,0,224,185,144,224,185,145,224,185,147,0,228,187,150,228,187,172,228,189,160,228,190,134,229,128,145,229,136,176,229,146,140,229,156,176,229,175,185,229,176,141,229,176,177,229,184,173,230,136,145,230,151,182,230,153,130,230,156,131,230,157,165,231,130,186,232,131,189,232,136,176,232,170,170,232,175,180,232,191,153,233,128,153,233,189,138,124,229,134,155,229,144,140,229,183,178,230,132,191,230,151,162,230,152,159,230,152,175,230,153,175,230,176,145,231,133,167,231,142,176,231,143,190,231,144,134,231,148,168,231,189,174,232,166,129,232,187,141,233,130,163,233,133,141,233,135,140,233,150,139,233,155,183,233,156,178,233,157,162,233,161,190,0,228,184,170,228,184,186,228,186,186,228,187,150,228,187,165,228,187,172,228,189,160,228,190,134,229,128,139,229,128,145,229,136,176,229,146,140,229,164,167,229,175,185,229,176,141,229,176,177,230,136,145,230,151,182,230,153,130,230,156,137,230,157,165,231,130,186,232,166,129,232,170,170,232,175,180,124,228,184,187,228,186,155,229,155,160,229,174,131,230,131,179,230,132,143,231,144,134,231,148,159,231,149,182,231,156,139,231,157,128,231,189,174,232,128,133,232,135,170,232,145,151,232,163,161,232,191,135,232,191,152,232,191,155,233,128,178,233,129,142,233,129,147,233,130,132,233,135,140,233,157,162,0,97,117,116,111,102,105,116,116,101,114,0,112,114,111,112,101,114,116,105,101,115,0,102,97,108,108,98,97,99,107,45,115,99,114,105,112,116,0,100,101,102,97,117,108,116,45,115,99,114,105,112,116,0,105,110,99,114,101,97,115,101,45,120,45,104,101,105,103,104,116,0,119,97,114,112,105,110,103,0,103,108,121,112,104,45,116,111,45,115,99,114,105,112,116,45,109,97,112,0,111,118,101,114,115,117,98,115,99,114,105,98,101,100,32,100,121,110,97,109,105,99,32,98,105,116,32,108,101,110,103,116,104,115,32,116,114,101,101,0,105,110,99,111,109,112,108,101,116,101,32,100,121,110,97,109,105,99,32,98,105,116,32,108,101,110,103,116,104,115,32,116,114,101,101,0,111,118,101,114,115,117,98,115,99,114,105,98,101,100,32,108,105,116,101,114,97,108,47,108,101,110,103,116,104,32,116,114,101,101,0,105,110,99,111,109,112,108,101,116,101,32,108,105,116,101,114,97,108,47,108,101,110,103,116,104,32,116,114,101,101,0,111,118,101,114,115,117,98,115,99,114,105,98,101,100,32,100,105,115,116,97,110,99,101,32,116,114,101,101,0,105,110,99,111,109,112,108,101,116,101,32,100,105,115,116,97,110,99,101,32,116,114,101,101,0,101,109,112,116,121,32,100,105,115,116,97,110,99,101,32,116,114,101,101,32,119,105,116,104,32,108,101,110,103,116,104,115,0,118,111,114,98,105,115,0,79,103,103,83,0,85,110,107,110,111,119,110,32,79,112,101,110,71,76,32,97,116,116,114,105,98,117,116,101,0,37,115,0,86,69,82,66,79,83,69,0,68,69,66,85,71,0,73,78,70,79,0,87,65,82,78,0,69,82,82,79,82,0,67,82,73,84,73,67,65,76,0,37,115,58,32,37,115,10,0,86,105,100,101,111,32,115,117,98,115,121,115,116,101,109,32,104,97,115,32,110,111,116,32,98,101,101,110,32,105,110,105,116,105,97,108,105,122,101,100,0,78,111,32,71,76,32,100,114,105,118,101,114,32,104,97,115,32,98,101,101,110,32,108,111,97,100,101,100,0,78,111,32,100,121,110,97,109,105,99,32,71,76,32,115,117,112,112,111,114,116,32,105,110,32,118,105,100,101,111,32,100,114,105,118,101,114,0,70,97,105,108,101,100,32,103,101,116,116,105,110,103,32,79,112,101,110,71,76,32,103,108,71,101,116,73,110,116,101,103,101,114,118,32,101,110,116,114,121,32,112,111,105,110,116,0,70,97,105,108,101,100,32,103,101,116,116,105,110,103,32,79,112,101,110,71,76,32,103,108,71,101,116,69,114,114,111,114,32,101,110,116,114,121,32,112,111,105,110,116,0,79,112,101,110,71,76,32,101,114,114,111,114,58,32,71,76,95,73,78,86,65,76,73,68,95,69,78,85,77,0,79,112,101,110,71,76,32,101,114,114,111,114,58,32,71,76,95,73,78,86,65,76,73,68,95,86,65,76,85,69,0,79,112,101,110,71,76,32,101,114,114,111,114,58,32,37,48,56,88,0,85,110,107,110,111,119,110,32,79,112,101,110,71,76,32,99,111,110,116,101,120,116,32,102,108,97,103,32,37,100,0,85,110,107,110,111,119,110,32,79,112,101,110,71,76,32,99,111,110,116,101,120,116,32,112,114,111,102,105,108,101,32,37,100,0,78,111,32,79,112,101,110,71,76,32,115,117,112,112,111,114,116,32,105,110,32,118,105,100,101,111,32,100,114,105,118,101,114,0,100,105,115,112,108,97,121,73,110,100,101,120,32,109,117,115,116,32,98,101,32,105,110,32,116,104,101,32,114,97,110,103,101,32,48,32,45,32,37,100,0,112,111,105,110,116,115,0,99,111,117,110,116,0,67,111,117,108,100,110,39,116,32,102,105,110,100,32,97,110,121,32,100,105,115,112,108,97,121,115,0,109,111,100,101,0,77,105,115,115,105,110,103,32,100,101,115,105,114,101,100,32,109,111,100,101,32,111,114,32,99,108,111,115,101,115,116,32,109,111,100,101,32,112,97,114,97,109,101,116,101,114,0,67,111,117,108,100,110,39,116,32,102,105,110,100,32,100,105,115,112,108,97,121,32,109,111,100,101,32,109,97,116,99,104,0,78,111,32,118,105,100,101,111,32,109,111,100,101,32,108,97,114,103,101,32,101,110,111,117,103,104,32,102,111,114,32,37,100,120,37,100,0,86,105,100,101,111,32,100,114,105,118,101,114,32,100,111,101,115,110,39,116,32,115,117,112,112,111,114,116,32,99,104,97,110,103,105,110,103,32,100,105,115,112,108,97,121,32,109,111,100,101,0,84,104,101,32,101,118,101,110,116,32,115,121,115,116,101,109,32,104,97,115,32,98,101,101,110,32,115,104,117,116,32,100,111,119,110,0,69,118,101,110,116,32,113,117,101,117,101,32,105,115,32,102,117,108,108,32,40,37,100,32,101,118,101,110,116,115,41,0,67,111,117,108,100,110,39,116,32,108,111,99,107,32,101,118,101,110,116,32,113,117,101,117,101,0,69,82,82,79,82,58,32,78,117,109,80,111,105,110,116,115,32,61,32,37,105,10,0,110,111,32,103,101,115,116,117,114,101,32,116,111,117,99,104,32,100,101,118,105,99,101,115,32,114,101,103,105,115,116,101,114,101,100,0,79,117,116,32,111,102,32,109,101,109,111,114,121,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,102,114,111,109,32,100,97,116,97,115,116,114,101,97,109,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,116,111,32,100,97,116,97,115,116,114,101,97,109,0,69,114,114,111,114,32,115,101,101,107,105,110,103,32,105,110,32,100,97,116,97,115,116,114,101,97,109,0,84,104,97,116,32,111,112,101,114,97,116,105,111,110,32,105,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,85,110,107,110,111,119,110,32,83,68,76,32,101,114,114,111,114,0,83,68,76,95,70,105,108,108,82,101,99,116,40,41,58,32,85,110,115,117,112,112,111,114,116,101,100,32,115,117,114,102,97,99,101,32,102,111,114,109,97,116,0,65,0,66,0,114,101,115,117,108,116,0,83,68,76,95,70,105,108,108,82,101,99,116,40,41,58,32,89,111,117,32,109,117,115,116,32,108,111,99,107,32,116,104,101,32,115,117,114,102,97,99,101,0,0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,159,160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255,0,2,4,6,8,10,12,14,16,18,20,22,24,26,28,30,32,34,36,38,40,42,44,46,48,50,52,54,56,58,60,62,64,66,68,70,72,74,76,78,80,82,84,86,88,90,92,94,96,98,100,102,104,106,108,110,112,114,116,118,120,122,124,126,128,130,132,134,136,138,140,142,144,146,148,150,152,154,156,158,160,162,164,166,168,170,172,174,176,178,180,182,184,186,188,190,192,194,196,198,200,202,204,206,208,210,212,214,216,218,220,222,224,226,228,230,232,234,236,238,240,242,244,246,248,250,252,255,0,4,8,12,16,20,24,28,32,36,40,44,48,52,56,60,64,68,72,76,80,85,89,93,97,101,105,109,113,117,121,125,129,133,137,141,145,149,153,157,161,165,170,174,178,182,186,190,194,198,202,206,210,214,218,222,226,230,234,238,242,246,250,255,0,8,16,24,32,41,49,57,65,74,82,90,98,106,115,123,131,139,148,156,164,172,180,189,197,205,213,222,230,238,246,255,0,17,34,51,68,85,102,119,136,153,170,187,204,221,238,255,0,36,72,109,145,182,218,255,0,85,170,255,0,255,255,83,68,76,95,83,101,116,80,105,120,101,108,70,111,114,109,97,116,80,97,108,101,116,116,101,40,41,32,112,97,115,115,101,100,32,78,85,76,76,32,102,111,114,109,97,116,0,83,68,76,95,83,101,116,80,105,120,101,108,70,111,114,109,97,116,80,97,108,101,116,116,101,40,41,32,112,97,115,115,101,100,32,97,32,112,97,108,101,116,116,101,32,116,104,97,116,32,100,111,101,115,110,39,116,32,109,97,116,99,104,32,116,104,101,32,102,111,114,109,97,116,0,102,111,114,109,97,116,0,84,104,101,32,115,112,101,99,105,102,105,101,100,32,119,105,110,100,111,119,32,105,115,110,39,116,32,97,110,32,79,112,101,110,71,76,32,119,105,110,100,111,119,0,105,100,0,73,110,118,97,108,105,100,32,114,101,110,100,101,114,101,114,0,73,110,118,97,108,105,100,32,116,101,120,116,117,114,101,0,82,101,110,100,101,114,101,114,32,100,111,101,115,110,39,116,32,115,117,112,112,111,114,116,32,113,117,101,114,121,105,110,103,32,111,117,116,112,117,116,32,115,105,122,101,0,103,101,110,101,114,105,99,0,85,78,75,78,79,87,78,0,71,76,95,79,85,84,95,79,70,95,77,69,77,79,82,89,0,71,76,95,73,78,86,65,76,73,68,95,79,80,69,82,65,84,73,79,78,0,71,76,95,73,78,86,65,76,73,68,95,86,65,76,85,69,0,71,76,95,73,78,86,65,76,73,68,95,69,78,85,77,0,37,115,58,32,37,115,32,40,37,100,41,58,32,37,115,32,37,115,32,40,48,120,37,88,41,0,47,85,115,101,114,115,47,116,97,110,110,101,114,47,46,101,109,115,99,114,105,112,116,101,110,95,112,111,114,116,115,47,115,100,108,50,47,83,68,76,50,45,118,101,114,115,105,111,110,95,49,48,47,115,114,99,47,114,101,110,100,101,114,47,111,112,101,110,103,108,101,115,50,47,83,68,76,95,114,101,110,100,101,114,95,103,108,101,115,50,46,99,0,105,110,116,32,71,76,69,83,50,95,85,112,100,97,116,101,86,105,101,119,112,111,114,116,40,83,68,76,95,82,101,110,100,101,114,101,114,32,42,41,0,118,111,105,100,32,71,76,69,83,50,95,68,101,115,116,114,111,121,82,101,110,100,101,114,101,114,40,83,68,76,95,82,101,110,100,101,114,101,114,32,42,41,0,67,111,117,108,100,110,39,116,32,108,111,97,100,32,71,76,69,83,50,32,102,117,110,99,116,105,111,110,32,37,115,58,32,37,115,10,0,78,111,32,79,112,101,110,71,76,32,99,111,110,116,101,120,116,32,104,97,115,32,98,101,101,110,32,109,97,100,101,32,99,117,114,114,101,110,116,0,83,101,116,116,105,110,103,32,116,104,101,32,115,119,97,112,32,105,110,116,101,114,118,97,108,32,105,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,84,101,120,116,117,114,101,32,102,111,114,109,97,116,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,105,110,116,32,71,76,69,83,50,95,67,114,101,97,116,101,84,101,120,116,117,114,101,40,83,68,76,95,82,101,110,100,101,114,101,114,32,42,44,32,83,68,76,95,84,101,120,116,117,114,101,32,42,41,0,103,108,71,101,110,84,101,120,117,114,101,115,40,41,0,103,108,84,101,120,73,109,97,103,101,50,68,40,41,0,103,108,84,101,120,83,117,98,73,109,97,103,101,50,68,40,41,0,105,110,116,32,71,76,69,83,50,95,85,112,100,97,116,101,84,101,120,116,117,114,101,40,83,68,76,95,82,101,110,100,101,114,101,114,32,42,44,32,83,68,76,95,84,101,120,116,117,114,101,32,42,44,32,99,111,110,115,116,32,83,68,76,95,82,101,99,116,32,42,44,32,99,111,110,115,116,32,118,111,105,100,32,42,44,32,105,110,116,41,0,105,110,116,32,71,76,69,83,50,95,85,112,100,97,116,101,84,101,120,116,117,114,101,89,85,86,40,83,68,76,95,82,101,110,100,101,114,101,114,32,42,44,32,83,68,76,95,84,101,120,116,117,114,101,32,42,44,32,99,111,110,115,116,32,83,68,76,95,82,101,99,116,32,42,44,32,99,111,110,115,116,32,85,105,110,116,56,32,42,44,32,105,110,116,44,32,99,111,110,115,116,32,85,105,110,116,56,32,42,44,32,105,110,116,44,32,99,111,110,115,116,32,85,105,110,116,56,32,42,44,32,105,110,116,41,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,50,68,40,41,32,102,97,105,108,101,100,0,32,32,32,32,32,112,114,101,99,105,115,105,111,110,32,109,101,100,105,117,109,112,32,102,108,111,97,116,59,32,32,32,32,32,117,110,105,102,111,114,109,32,115,97,109,112,108,101,114,50,68,32,117,95,116,101,120,116,117,114,101,59,32,32,32,32,32,117,110,105,102,111,114,109,32,115,97,109,112,108,101,114,50,68,32,117,95,116,101,120,116,117,114,101,95,117,59,32,32,32,32,32,117,110,105,102,111,114,109,32,118,101,99,52,32,117,95,109,111,100,117,108,97,116,105,111,110,59,32,32,32,32,32,118,97,114,121,105,110,103,32,118,101,99,50,32,118,95,116,101,120,67,111,111,114,100,59,32,32,32,32,32,32,32,32,32,118,111,105,100,32,109,97,105,110,40,41,32,32,32,32,32,123,32,32,32,32,32,32,32,32,32,109,101,100,105,117,109,112,32,118,101,99,51,32,121,117,118,59,32,32,32,32,32,32,32,32,32,108,111,119,112,32,118,101,99,51,32,114,103,98,59,32,32,32,32,32,32,32,32,32,121,117,118,46,120,32,61,32,116,101,120,116,117,114,101,50,68,40,117,95,116,101,120,116,117,114,101,44,32,32,32,118,95,116,101,120,67,111,111,114,100,41,46,114,59,32,32,32,32,32,32,32,32,32,121,117,118,46,121,122,32,61,32,116,101,120,116,117,114,101,50,68,40,117,95,116,101,120,116,117,114,101,95,117,44,32,118,95,116,101,120,67,111,111,114,100,41,46,97,114,32,45,32,48,46,53,59,32,32,32,32,32,32,32,32,32,114,103,98,32,61,32,109,97,116,51,40,32,49,44,32,32,32,32,32,32,32,32,49,44,32,32,32,32,32,32,32,49,44,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,48,44,32,32,32,32,32,32,32,45,48,46,51,57,52,54,53,44,32,50,46,48,51,50,49,49,44,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,49,46,49,51,57,56,51,44,32,45,48,46,53,56,48,54,48,44,32,48,41,32,42,32,121,117,118,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,32,61,32,118,101,99,52,40,114,103,98,44,32,49,41,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,32,42,61,32,117,95,109,111,100,117,108,97,116,105,111,110,59,32,32,32,32,32,125,32,0,32,32,32,32,32,112,114,101,99,105,115,105,111,110,32,109,101,100,105,117,109,112,32,102,108,111,97,116,59,32,32,32,32,32,117,110,105,102,111,114,109,32,115,97,109,112,108,101,114,50,68,32,117,95,116,101,120,116,117,114,101,59,32,32,32,32,32,117,110,105,102,111,114,109,32,115,97,109,112,108,101,114,50,68,32,117,95,116,101,120,116,117,114,101,95,117,59,32,32,32,32,32,117,110,105,102,111,114,109,32,118,101,99,52,32,117,95,109,111,100,117,108,97,116,105,111,110,59,32,32,32,32,32,118,97,114,121,105,110,103,32,118,101,99,50,32,118,95,116,101,120,67,111,111,114,100,59,32,32,32,32,32,32,32,32,32,118,111,105,100,32,109,97,105,110,40,41,32,32,32,32,32,123,32,32,32,32,32,32,32,32,32,109,101,100,105,117,109,112,32,118,101,99,51,32,121,117,118,59,32,32,32,32,32,32,32,32,32,108,111,119,112,32,118,101,99,51,32,114,103,98,59,32,32,32,32,32,32,32,32,32,121,117,118,46,120,32,61,32,116,101,120,116,117,114,101,50,68,40,117,95,116,101,120,116,117,114,101,44,32,32,32,118,95,116,101,120,67,111,111,114,100,41,46,114,59,32,32,32,32,32,32,32,32,32,121,117,118,46,121,122,32,61,32,116,101,120,116,117,114,101,50,68,40,117,95,116,101,120,116,117,114,101,95,117,44,32,118,95,116,101,120,67,111,111,114,100,41,46,114,97,32,45,32,48,46,53,59,32,32,32,32,32,32,32,32,32,114,103,98,32,61,32,109,97,116,51,40,32,49,44,32,32,32,32,32,32,32,32,49,44,32,32,32,32,32,32,32,49,44,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,48,44,32,32,32,32,32,32,32,45,48,46,51,57,52,54,53,44,32,50,46,48,51,50,49,49,44,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,49,46,49,51,57,56,51,44,32,45,48,46,53,56,48,54,48,44,32,48,41,32,42,32,121,117,118,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,32,61,32,118,101,99,52,40,114,103,98,44,32,49,41,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,32,42,61,32,117,95,109,111,100,117,108,97,116,105,111,110,59,32,32,32,32,32,125,32,0,32,32,32,32,32,112,114,101,99,105,115,105,111,110,32,109,101,100,105,117,109,112,32,102,108,111,97,116,59,32,32,32,32,32,117,110,105,102,111,114,109,32,115,97,109,112,108,101,114,50,68,32,117,95,116,101,120,116,117,114,101,59,32,32,32,32,32,117,110,105,102,111,114,109,32,115,97,109,112,108,101,114,50,68,32,117,95,116,101,120,116,117,114,101,95,117,59,32,32,32,32,32,117,110,105,102,111,114,109,32,115,97,109,112,108,101,114,50,68,32,117,95,116,101,120,116,117,114,101,95,118,59,32,32,32,32,32,117,110,105,102,111,114,109,32,118,101,99,52,32,117,95,109,111,100,117,108,97,116,105,111,110,59,32,32,32,32,32,118,97,114,121,105,110,103,32,118,101,99,50,32,118,95,116,101,120,67,111,111,114,100,59,32,32,32,32,32,32,32,32,32,118,111,105,100,32,109,97,105,110,40,41,32,32,32,32,32,123,32,32,32,32,32,32,32,32,32,109,101,100,105,117,109,112,32,118,101,99,51,32,121,117,118,59,32,32,32,32,32,32,32,32,32,108,111,119,112,32,118,101,99,51,32,114,103,98,59,32,32,32,32,32,32,32,32,32,121,117,118,46,120,32,61,32,116,101,120,116,117,114,101,50,68,40,117,95,116,101,120,116,117,114,101,44,32,32,32,118,95,116,101,120,67,111,111,114,100,41,46,114,59,32,32,32,32,32,32,32,32,32,121,117,118,46,121,32,61,32,116,101,120,116,117,114,101,50,68,40,117,95,116,101,120,116,117,114,101,95,117,44,32,118,95,116,101,120,67,111,111,114,100,41,46,114,32,45,32,48,46,53,59,32,32,32,32,32,32,32,32,32,121,117,118,46,122,32,61,32,116,101,120,116,117,114,101,50,68,40,117,95,116,101,120,116,117,114,101,95,118,44,32,118,95,116,101,120,67,111,111,114,100,41,46,114,32,45,32,48,46,53,59,32,32,32,32,32,32,32,32,32,114,103,98,32,61,32,109,97,116,51,40,32,49,44,32,32,32,32,32,32,32,32,49,44,32,32,32,32,32,32,32,49,44,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,48,44,32,32,32,32,32,32,32,45,48,46,51,57,52,54,53,44,32,50,46,48,51,50,49,49,44,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,49,46,49,51,57,56,51,44,32,45,48,46,53,56,48,54,48,44,32,48,41,32,42,32,121,117,118,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,32,61,32,118,101,99,52,40,114,103,98,44,32,49,41,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,32,42,61,32,117,95,109,111,100,117,108,97,116,105,111,110,59,32,32,32,32,32,125,32,0,32,32,32,32,32,112,114,101,99,105,115,105,111,110,32,109,101,100,105,117,109,112,32,102,108,111,97,116,59,32,32,32,32,32,117,110,105,102,111,114,109,32,115,97,109,112,108,101,114,50,68,32,117,95,116,101,120,116,117,114,101,59,32,32,32,32,32,117,110,105,102,111,114,109,32,118,101,99,52,32,117,95,109,111,100,117,108,97,116,105,111,110,59,32,32,32,32,32,118,97,114,121,105,110,103,32,118,101,99,50,32,118,95,116,101,120,67,111,111,114,100,59,32,32,32,32,32,32,32,32,32,118,111,105,100,32,109,97,105,110,40,41,32,32,32,32,32,123,32,32,32,32,32,32,32,32,32,118,101,99,52,32,97,98,103,114,32,61,32,116,101,120,116,117,114,101,50,68,40,117,95,116,101,120,116,117,114,101,44,32,118,95,116,101,120,67,111,111,114,100,41,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,32,61,32,97,98,103,114,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,46,97,32,61,32,49,46,48,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,32,42,61,32,117,95,109,111,100,117,108,97,116,105,111,110,59,32,32,32,32,32,125,32,0,32,32,32,32,32,112,114,101,99,105,115,105,111,110,32,109,101,100,105,117,109,112,32,102,108,111,97,116,59,32,32,32,32,32,117,110,105,102,111,114,109,32,115,97,109,112,108,101,114,50,68,32,117,95,116,101,120,116,117,114,101,59,32,32,32,32,32,117,110,105,102,111,114,109,32,118,101,99,52,32,117,95,109,111,100,117,108,97,116,105,111,110,59,32,32,32,32,32,118,97,114,121,105,110,103,32,118,101,99,50,32,118,95,116,101,120,67,111,111,114,100,59,32,32,32,32,32,32,32,32,32,118,111,105,100,32,109,97,105,110,40,41,32,32,32,32,32,123,32,32,32,32,32,32,32,32,32,118,101,99,52,32,97,98,103,114,32,61,32,116,101,120,116,117,114,101,50,68,40,117,95,116,101,120,116,117,114,101,44,32,118,95,116,101,120,67,111,111,114,100,41,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,32,61,32,97,98,103,114,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,46,114,32,61,32,97,98,103,114,46,98,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,46,98,32,61,32,97,98,103,114,46,114,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,46,97,32,61,32,49,46,48,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,32,42,61,32,117,95,109,111,100,117,108,97,116,105,111,110,59,32,32,32,32,32,125,32,0,32,32,32,32,32,112,114,101,99,105,115,105,111,110,32,109,101,100,105,117,109,112,32,102,108,111,97,116,59,32,32,32,32,32,117,110,105,102,111,114,109,32,115,97,109,112,108,101,114,50,68,32,117,95,116,101,120,116,117,114,101,59,32,32,32,32,32,117,110,105,102,111,114,109,32,118,101,99,52,32,117,95,109,111,100,117,108,97,116,105,111,110,59,32,32,32,32,32,118,97,114,121,105,110,103,32,118,101,99,50,32,118,95,116,101,120,67,111,111,114,100,59,32,32,32,32,32,32,32,32,32,118,111,105,100,32,109,97,105,110,40,41,32,32,32,32,32,123,32,32,32,32,32,32,32,32,32,118,101,99,52,32,97,98,103,114,32,61,32,116,101,120,116,117,114,101,50,68,40,117,95,116,101,120,116,117,114,101,44,32,118,95,116,101,120,67,111,111,114,100,41,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,32,61,32,97,98,103,114,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,46,114,32,61,32,97,98,103,114,46,98,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,46,98,32,61,32,97,98,103,114,46,114,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,32,42,61,32,117,95,109,111,100,117,108,97,116,105,111,110,59,32,32,32,32,32,125,32,0,32,32,32,32,32,112,114,101,99,105,115,105,111,110,32,109,101,100,105,117,109,112,32,102,108,111,97,116,59,32,32,32,32,32,117,110,105,102,111,114,109,32,115,97,109,112,108,101,114,50,68,32,117,95,116,101,120,116,117,114,101,59,32,32,32,32,32,117,110,105,102,111,114,109,32,118,101,99,52,32,117,95,109,111,100,117,108,97,116,105,111,110,59,32,32,32,32,32,118,97,114,121,105,110,103,32,118,101,99,50,32,118,95,116,101,120,67,111,111,114,100,59,32,32,32,32,32,32,32,32,32,118,111,105,100,32,109,97,105,110,40,41,32,32,32,32,32,123,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,32,61,32,116,101,120,116,117,114,101,50,68,40,117,95,116,101,120,116,117,114,101,44,32,118,95,116,101,120,67,111,111,114,100,41,59,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,32,42,61,32,117,95,109,111,100,117,108,97,116,105,111,110,59,32,32,32,32,32,125,32,0,32,32,32,32,32,112,114,101,99,105,115,105,111,110,32,109,101,100,105,117,109,112,32,102,108,111,97,116,59,32,32,32,32,32,117,110,105,102,111,114,109,32,118,101,99,52,32,117,95,99,111,108,111,114,59,32,32,32,32,32,32,32,32,32,118,111,105,100,32,109,97,105,110,40,41,32,32,32,32,32,123,32,32,32,32,32,32,32,32,32,103,108,95,70,114,97,103,67,111,108,111,114,32,61,32,117,95,99,111,108,111,114,59,32,32,32,32,32,125,32,0,32,32,32,32,32,117,110,105,102,111,114,109,32,109,97,116,52,32,117,95,112,114,111,106,101,99,116,105,111,110,59,32,32,32,32,32,97,116,116,114,105,98,117,116,101,32,118,101,99,50,32,97,95,112,111,115,105,116,105,111,110,59,32,32,32,32,32,97,116,116,114,105,98,117,116,101,32,118,101,99,50,32,97,95,116,101,120,67,111,111,114,100,59,32,32,32,32,32,97,116,116,114,105,98,117,116,101,32,102,108,111,97,116,32,97,95,97,110,103,108,101,59,32,32,32,32,32,97,116,116,114,105,98,117,116,101,32,118,101,99,50,32,97,95,99,101,110,116,101,114,59,32,32,32,32,32,118,97,114,121,105,110,103,32,118,101,99,50,32,118,95,116,101,120,67,111,111,114,100,59,32,32,32,32,32,32,32,32,32,118,111,105,100,32,109,97,105,110,40,41,32,32,32,32,32,123,32,32,32,32,32,32,32,32,32,102,108,111,97,116,32,97,110,103,108,101,32,61,32,114,97,100,105,97,110,115,40,97,95,97,110,103,108,101,41,59,32,32,32,32,32,32,32,32,32,102,108,111,97,116,32,99,32,61,32,99,111,115,40,97,110,103,108,101,41,59,32,32,32,32,32,32,32,32,32,102,108,111,97,116,32,115,32,61,32,115,105,110,40,97,110,103,108,101,41,59,32,32,32,32,32,32,32,32,32,109,97,116,50,32,114,111,116,97,116,105,111,110,77,97,116,114,105,120,32,61,32,109,97,116,50,40,99,44,32,45,115,44,32,115,44,32,99,41,59,32,32,32,32,32,32,32,32,32,118,101,99,50,32,112,111,115,105,116,105,111,110,32,61,32,114,111,116,97,116,105,111,110,77,97,116,114,105,120,32,42,32,40,97,95,112,111,115,105,116,105,111,110,32,45,32,97,95,99,101,110,116,101,114,41,32,43,32,97,95,99,101,110,116,101,114,59,32,32,32,32,32,32,32,32,32,118,95,116,101,120,67,111,111,114,100,32,61,32,97,95,116,101,120,67,111,111,114,100,59,32,32,32,32,32,32,32,32,32,103,108,95,80,111,115,105,116,105,111,110,32,61,32,117,95,112,114,111,106,101,99,116,105,111,110,32,42,32,118,101,99,52,40,112,111,115,105,116,105,111,110,44,32,48,46,48,44,32,49,46,48,41,59,32,32,32,32,32,32,32,32,103,108,95,80,111,105,110,116,83,105,122,101,32,61,32,49,46,48,59,32,32,32,32,32,125,32,0,78,111,32,115,104,97,100,101,114,32,109,97,116,99,104,105,110,103,32,116,104,101,32,114,101,113,117,101,115,116,101,100,32,99,104,97,114,97,99,116,101,114,105,115,116,105,99,115,32,119,97,115,32,102,111,117,110,100,0,84,104,101,32,115,112,101,99,105,102,105,101,100,32,115,104,97,100,101,114,32,99,97,110,110,111,116,32,98,101,32,108,111,97,100,101,100,32,111,110,32,116,104,101,32,99,117,114,114,101,110,116,32,112,108,97,116,102,111,114,109,0,70,97,105,108,101,100,32,116,111,32,108,111,97,100,32,116,104,101,32,115,104,97,100,101,114,58,32,37,115,0,70,97,105,108,101,100,32,116,111,32,108,111,97,100,32,116,104,101,32,115,104,97,100,101,114,0,97,95,112,111,115,105,116,105,111,110,0,97,95,116,101,120,67,111,111,114,100,0,97,95,97,110,103,108,101,0,97,95,99,101,110,116,101,114,0,70,97,105,108,101,100,32,116,111,32,108,105,110,107,32,115,104,97,100,101,114,32,112,114,111,103,114,97,109,0,117,95,112,114,111,106,101,99,116,105,111,110,0,117,95,116,101,120,116,117,114,101,95,118,0,117,95,116,101,120,116,117,114,101,95,117,0,117,95,116,101,120,116,117,114,101,0,117,95,109,111,100,117,108,97,116,105,111,110,0,117,95,99,111,108,111,114,0,105,110,116,32,71,76,69,83,50,95,82,101,110,100,101,114,68,114,97,119,76,105,110,101,115,40,83,68,76,95,82,101,110,100,101,114,101,114,32,42,44,32,99,111,110,115,116,32,83,68,76,95,70,80,111,105,110,116,32,42,44,32,105,110,116,41,0,105,110,116,32,71,76,69,83,50,95,82,101,110,100,101,114,70,105,108,108,82,101,99,116,115,40,83,68,76,95,82,101,110,100,101,114,101,114,32,42,44,32,99,111,110,115,116,32,83,68,76,95,70,82,101,99,116,32,42,44,32,105,110,116,41,0,85,110,115,117,112,112,111,114,116,101,100,32,116,101,120,116,117,114,101,32,102,111,114,109,97,116,0,105,110,116,32,71,76,69,83,50,95,82,101,110,100,101,114,67,111,112,121,40,83,68,76,95,82,101,110,100,101,114,101,114,32,42,44,32,83,68,76,95,84,101,120,116,117,114,101,32,42,44,32,99,111,110,115,116,32,83,68,76,95,82,101,99,116,32,42,44,32,99,111,110,115,116,32,83,68,76,95,70,82,101,99,116,32,42,41,0,105,110,116,32,71,76,69,83,50,95,82,101,110,100,101,114,67,111,112,121,69,120,40,83,68,76,95,82,101,110,100,101,114,101,114,32,42,44,32,83,68,76,95,84,101,120,116,117,114,101,32,42,44,32,99,111,110,115,116,32,83,68,76,95,82,101,99,116,32,42,44,32,99,111,110,115,116,32,83,68,76,95,70,82,101,99,116,32,42,44,32,99,111,110,115,116,32,100,111,117,98,108,101,44,32,99,111,110,115,116,32,83,68,76,95,70,80,111,105,110,116,32,42,44,32,99,111,110,115,116,32,83,68,76,95,82,101,110,100,101,114,101,114,70,108,105,112,41,0,103,108,82,101,97,100,80,105,120,101,108,115,40,41,0,105,110,116,32,71,76,69,83,50,95,82,101,110,100,101,114,82,101,97,100,80,105,120,101,108,115,40,83,68,76,95,82,101,110,100,101,114,101,114,32,42,44,32,99,111,110,115,116,32,83,68,76,95,82,101,99,116,32,42,44,32,85,105,110,116,51,50,44,32,118,111,105,100,32,42,44,32,105,110,116,41,0,100,115,116,0,100,115,116,95,112,105,116,99,104,0,85,110,107,110,111,119,110,32,70,79,85,82,67,67,32,112,105,120,101,108,32,102,111,114,109,97,116,0,73,110,100,101,120,101,100,32,112,105,120,101,108,32,102,111,114,109,97,116,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,70,79,85,82,67,67,32,112,105,120,101,108,32,102,111,114,109,97,116,115,32,97,114,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,83,68,76,95,66,76,73,84,95,67,80,85,95,70,69,65,84,85,82,69,83,0,37,117,0,66,108,105,116,32,99,111,109,98,105,110,97,116,105,111,110,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,84,104,101,32,115,112,101,99,105,102,105,101,100,32,119,105,110,100,111,119,32,104,97,115,32,110,111,116,32,98,101,101,110,32,109,97,100,101,32,99,117,114,114,101,110,116,0,118,111,105,100,32,71,76,69,83,50,95,82,101,115,101,116,83,116,97,116,101,40,83,68,76,95,82,101,110,100,101,114,101,114,32,42,41,0,111,112,101,110,103,108,101,115,50,0,85,110,107,110,111,119,110,32,112,105,120,101,108,32,102,111,114,109,97,116,0,110,99,111,108,111,114,115,0,112,97,108,101,116,116,101,0,67,97,110,39,116,32,99,114,101,97,116,101,32,114,101,110,100,101,114,101,114,32,102,111,114,32,78,85,76,76,32,115,117,114,102,97,99,101,0,83,111,102,116,119,97,114,101,32,114,101,110,100,101,114,101,114,32,100,111,101,115,110,39,116,32,104,97,118,101,32,97,110,32,111,117,116,112,117,116,32,115,117,114,102,97,99,101,0,85,110,107,110,111,119,110,32,116,101,120,116,117,114,101,32,102,111,114,109,97,116,0,83,68,76,95,68,114,97,119,80,111,105,110,116,115,40,41,58,32,85,110,115,117,112,112,111,114,116,101,100,32,115,117,114,102,97,99,101,32,102,111,114,109,97,116,0,83,68,76,95,66,108,101,110,100,80,111,105,110,116,115,40,41,58,32,85,110,115,117,112,112,111,114,116,101,100,32,115,117,114,102,97,99,101,32,102,111,114,109,97,116,0,83,68,76,95,68,114,97,119,76,105,110,101,115,40,41,58,32,80,97,115,115,101,100,32,78,85,76,76,32,100,101,115,116,105,110,97,116,105,111,110,32,115,117,114,102,97,99,101,0,83,68,76,95,68,114,97,119,76,105,110,101,115,40,41,58,32,85,110,115,117,112,112,111,114,116,101,100,32,115,117,114,102,97,99,101,32,102,111,114,109,97,116,0,114,101,99,116,0,88,49,0,89,49,0,88,50,0,89,50,0,83,68,76,95,68,114,97,119,80,111,105,110,116,40,41,58,32,85,110,115,117,112,112,111,114,116,101,100,32,115,117,114,102,97,99,101,32,102,111,114,109,97,116,0,83,68,76,95,66,108,101,110,100,76,105,110,101,115,40,41,58,32,80,97,115,115,101,100,32,78,85,76,76,32,100,101,115,116,105,110,97,116,105,111,110,32,115,117,114,102,97,99,101,0,83,68,76,95,66,108,101,110,100,76,105,110,101,115,40,41,58,32,85,110,115,117,112,112,111,114,116,101,100,32,115,117,114,102,97,99,101,32,102,111,114,109,97,116,0,83,68,76,95,66,108,101,110,100,80,111,105,110,116,40,41,58,32,85,110,115,117,112,112,111,114,116,101,100,32,115,117,114,102,97,99,101,32,102,111,114,109,97,116,0,83,68,76,95,70,105,108,108,82,101,99,116,115,40,41,32,112,97,115,115,101,100,32,78,85,76,76,32,114,101,99,116,115,0,80,97,115,115,101,100,32,78,85,76,76,32,100,101,115,116,105,110,97,116,105,111,110,32,115,117,114,102,97,99,101,0,83,68,76,95,66,108,101,110,100,70,105,108,108,82,101,99,116,115,40,41,58,32,85,110,115,117,112,112,111,114,116,101,100,32,115,117,114,102,97,99,101,32,102,111,114,109,97,116,0,83,68,76,95,85,112,112,101,114,66,108,105,116,58,32,112,97,115,115,101,100,32,97,32,78,85,76,76,32,115,117,114,102,97,99,101,0,83,117,114,102,97,99,101,115,32,109,117,115,116,32,110,111,116,32,98,101,32,108,111,99,107,101,100,32,100,117,114,105,110,103,32,98,108,105,116,0,83,68,76,95,85,112,112,101,114,66,108,105,116,83,99,97,108,101,100,58,32,112,97,115,115,101,100,32,97,32,78,85,76,76,32,115,117,114,102,97,99,101,0,79,110,108,121,32,119,111,114,107,115,32,119,105,116,104,32,115,97,109,101,32,102,111,114,109,97,116,32,115,117,114,102,97,99,101,115,0,73,110,118,97,108,105,100,32,115,111,117,114,99,101,32,98,108,105,116,32,114,101,99,116,97,110,103,108,101,0,73,110,118,97,108,105,100,32,100,101,115,116,105,110,97,116,105,111,110,32,98,108,105,116,32,114,101,99,116,97,110,103,108,101,0,85,110,97,98,108,101,32,116,111,32,108,111,99,107,32,100,101,115,116,105,110,97,116,105,111,110,32,115,117,114,102,97,99,101,0,85,110,97,98,108,101,32,116,111,32,108,111,99,107,32,115,111,117,114,99,101,32,115,117,114,102,97,99,101,0,115,117,114,102,97,99,101,0,107,101,121,0,69,109,112,116,121,32,100,101,115,116,105,110,97,116,105,111,110,32,112,97,108,101,116,116,101,0,83,68,76,95,82,69,78,68,69,82,95,83,67,65,76,69,95,81,85,65,76,73,84,89,0,110,101,97,114,101,115,116,0,84,114,105,101,100,32,116,111,32,114,101,97,100,32,111,117,116,115,105,100,101,32,111,102,32,115,117,114,102,97,99,101,32,98,111,117,110,100,115,0,87,105,110,100,111,119,32,115,117,114,102,97,99,101,32,105,115,32,105,110,118,97,108,105,100,44,32,112,108,101,97,115,101,32,99,97,108,108,32,83,68,76,95,71,101,116,87,105,110,100,111,119,83,117,114,102,97,99,101,40,41,32,116,111,32,103,101,116,32,97,32,110,101,119,32,115,117,114,102,97,99,101,0,83,68,76,32,101,109,115,99,114,105,112,116,101,110,32,118,105,100,101,111,32,100,114,105,118,101,114,0,83,68,76,95,86,73,68,69,79,95,77,73,78,73,77,73,90,69,95,79,78,95,70,79,67,85,83,95,76,79,83,83,0,48,0,123,32,114,101,116,117,114,110,32,115,99,114,101,101,110,46,119,105,100,116,104,59,32,125,0,123,32,114,101,116,117,114,110,32,115,99,114,101,101,110,46,104,101,105,103,104,116,59,32,125,0,48,49,50,51,52,53,54,55,56,57,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,100,101,102,97,117,108,116,0,123,32,105,102,32,40,77,111,100,117,108,101,91,39,99,97,110,118,97,115,39,93,41,32,123,32,77,111,100,117,108,101,91,39,99,97,110,118,97,115,39,93,46,115,116,121,108,101,91,39,99,117,114,115,111,114,39,93,32,61,32,77,111,100,117,108,101,91,39,80,111,105,110,116,101,114,95,115,116,114,105,110,103,105,102,121,39,93,40,36,48,41,59,32,125,32,114,101,116,117,114,110,32,48,59,32,125,0,105,102,32,40,77,111,100,117,108,101,91,39,99,97,110,118,97,115,39,93,41,32,123,32,77,111,100,117,108,101,91,39,99,97,110,118,97,115,39,93,46,115,116,121,108,101,91,39,99,117,114,115,111,114,39,93,32,61,32,39,110,111,110,101,39,59,32,125,0,112,111,105,110,116,101,114,0,110,111,116,45,97,108,108,111,119,101,100,0,110,115,45,114,101,115,105,122,101,0,101,119,45,114,101,115,105,122,101,0,110,101,115,119,45,114,101,115,105,122,101,0,110,119,115,101,45,114,101,115,105,122,101,0,112,114,111,103,114,101,115,115,0,99,114,111,115,115,104,97,105,114,0,119,97,105,116,0,116,101,120,116,0,79,112,101,110,71,76,32,108,105,98,114,97,114,121,32,97,108,114,101,97,100,121,32,108,111,97,100,101,100,0,69,71,76,32,105,109,112,108,101,109,101,110,116,97,116,105,111,110,32,100,111,101,115,32,110,111,116,32,115,117,112,112,111,114,116,32,115,82,71,66,32,115,121,115,116,101,109,32,102,114,97,109,101,98,117,102,102,101,114,115,0,67,111,117,108,100,110,39,116,32,102,105,110,100,32,109,97,116,99,104,105,110,103,32,69,71,76,32,99,111,110,102,105,103,0,67,111,117,108,100,32,110,111,116,32,99,114,101,97,116,101,32,71,76,69,83,32,119,105,110,100,111,119,32,115,117,114,102,97,99,101,0,78,111,32,119,105,110,100,111,119,32,104,97,115,32,102,111,99,117,115,0,35,99,97,110,118,97,115,0,35,100,111,99,117,109,101,110,116,0,35,119,105,110,100,111,119,0,85,110,107,110,111,119,110,32,116,111,117,99,104,32,100,101,118,105,99,101,0,83,68,76,95,69,77,83,67,82,73,80,84,69,78,95,75,69,89,66,79,65,82,68,95,69,76,69,77,69,78,84,0,123,32,105,102,32,40,116,121,112,101,111,102,32,77,111,100,117,108,101,91,39,115,101,116,87,105,110,100,111,119,84,105,116,108,101,39,93,32,33,61,61,32,39,117,110,100,101,102,105,110,101,100,39,41,32,123,32,77,111,100,117,108,101,91,39,115,101,116,87,105,110,100,111,119,84,105,116,108,101,39,93,40,77,111,100,117,108,101,91,39,80,111,105,110,116,101,114,95,115,116,114,105,110,103,105,102,121,39,93,40,36,48,41,41,59,32,125,32,114,101,116,117,114,110,32,48,59,32,125,0,67,111,117,108,100,110,39,116,32,102,105,110,100,32,102,114,97,109,101,98,117,102,102,101,114,32,115,117,114,102,97,99,101,32,102,111,114,32,119,105,110,100,111,119,0,123,32,118,97,114,32,119,32,61,32,36,48,59,32,118,97,114,32,104,32,61,32,36,49,59,32,118,97,114,32,112,105,120,101,108,115,32,61,32,36,50,59,32,105,102,32,40,33,77,111,100,117,108,101,91,39,83,68,76,50,39,93,41,32,77,111,100,117,108,101,91,39,83,68,76,50,39,93,32,61,32,123,125,59,32,118,97,114,32,83,68,76,50,32,61,32,77,111,100,117,108,101,91,39,83,68,76,50,39,93,59,32,105,102,32,40,83,68,76,50,46,99,116,120,67,97,110,118,97,115,32,33,61,61,32,77,111,100,117,108,101,91,39,99,97,110,118,97,115,39,93,41,32,123,32,83,68,76,50,46,99,116,120,32,61,32,77,111,100,117,108,101,91,39,99,114,101,97,116,101,67,111,110,116,101,120,116,39,93,40,77,111,100,117,108,101,91,39,99,97,110,118,97,115,39,93,44,32,102,97,108,115,101,44,32,116,114,117,101,41,59,32,83,68,76,50,46,99,116,120,67,97,110,118,97,115,32,61,32,77,111,100,117,108,101,91,39,99,97,110,118,97,115,39,93,59,32,125,32,105,102,32,40,83,68,76,50,46,119,32,33,61,61,32,119,32,124,124,32,83,68,76,50,46,104,32,33,61,61,32,104,32,124,124,32,83,68,76,50,46,105,109,97,103,101,67,116,120,32,33,61,61,32,83,68,76,50,46,99,116,120,41,32,123,32,83,68,76,50,46,105,109,97,103,101,32,61,32,83,68,76,50,46,99,116,120,46,99,114,101,97,116,101,73,109,97,103,101,68,97,116,97,40,119,44,32,104,41,59,32,83,68,76,50,46,119,32,61,32,119,59,32,83,68,76,50,46,104,32,61,32,104,59,32,83,68,76,50,46,105,109,97,103,101,67,116,120,32,61,32,83,68,76,50,46,99,116,120,59,32,125,32,118,97,114,32,100,97,116,97,32,61,32,83,68,76,50,46,105,109,97,103,101,46,100,97,116,97,59,32,118,97,114,32,115,114,99,32,61,32,112,105,120,101,108,115,32,62,62,32,50,59,32,118,97,114,32,100,115,116,32,61,32,48,59,32,118,97,114,32,110,117,109,59,32,105,102,32,40,116,121,112,101,111,102,32,67,97,110,118,97,115,80,105,120,101,108,65,114,114,97,121,32,33,61,61,32,39,117,110,100,101,102,105,110,101,100,39,32,38,38,32,100,97,116,97,32,105,110,115,116,97,110,99,101,111,102,32,67,97,110,118,97,115,80,105,120,101,108,65,114,114,97,121,41,32,123,32,110,117,109,32,61,32,100,97,116,97,46,108,101,110,103,116,104,59,32,119,104,105,108,101,32,40,100,115,116,32,60,32,110,117,109,41,32,123,32,118,97,114,32,118,97,108,32,61,32,72,69,65,80,51,50,91,115,114,99,93,59,32,100,97,116,97,91,100,115,116,32,93,32,61,32,118,97,108,32,38,32,48,120,102,102,59,32,100,97,116,97,91,100,115,116,43,49,93,32,61,32,40,118,97,108,32,62,62,32,56,41,32,38,32,48,120,102,102,59,32,100,97,116,97,91,100,115,116,43,50,93,32,61,32,40,118,97,108,32,62,62,32,49,54,41,32,38,32,48,120,102,102,59,32,100,97,116,97,91,100,115,116,43,51,93,32,61,32,48,120,102,102,59,32,115,114,99,43,43,59,32,100,115,116,32,43,61,32,52,59,32,125,32,125,32,101,108,115,101,32,123,32,105,102,32,40,83,68,76,50,46,100,97,116,97,51,50,68,97,116,97,32,33,61,61,32,100,97,116,97,41,32,123,32,83,68,76,50,46,100,97,116,97,51,50,32,61,32,110,101,119,32,73,110,116,51,50,65,114,114,97,121,40,100,97,116,97,46,98,117,102,102,101,114,41,59,32,83,68,76,50,46,100,97,116,97,56,32,61,32,110,101,119,32,85,105,110,116,56,65,114,114,97,121,40,100,97,116,97,46,98,117,102,102,101,114,41,59,32,125,32,118,97,114,32,100,97,116,97,51,50,32,61,32,83,68,76,50,46,100,97,116,97,51,50,59,32,110,117,109,32,61,32,100,97,116,97,51,50,46,108,101,110,103,116,104,59,32,100,97,116,97,51,50,46,115,101,116,40,72,69,65,80,51,50,46,115,117,98,97,114,114,97,121,40,115,114,99,44,32,115,114,99,32,43,32,110,117,109,41,41,59,32,118,97,114,32,100,97,116,97,56,32,61,32,83,68,76,50,46,100,97,116,97,56,59,32,118,97,114,32,105,32,61,32,51,59,32,118,97,114,32,106,32,61,32,105,32,43,32,52,42,110,117,109,59,32,105,102,32,40,110,117,109,32,37,32,56,32,61,61,32,48,41,32,123,32,119,104,105,108,101,32,40,105,32,60,32,106,41,32,123,32,100,97,116,97,56,91,105,93,32,61,32,48,120,102,102,59,32,105,32,61,32,105,32,43,32,52,32,124,32,48,59,32,100,97,116,97,56,91,105,93,32,61,32,48,120,102,102,59,32,105,32,61,32,105,32,43,32,52,32,124,32,48,59,32,100,97,116,97,56,91,105,93,32,61,32,48,120,102,102,59,32,105,32,61,32,105,32,43,32,52,32,124,32,48,59,32,100,97,116,97,56,91,105,93,32,61,32,48,120,102,102,59,32,105,32,61,32,105,32,43,32,52,32,124,32,48,59,32,100,97,116,97,56,91,105,93,32,61,32,48,120,102,102,59,32,105,32,61,32,105,32,43,32,52,32,124,32,48,59,32,100,97,116,97,56,91,105,93,32,61,32,48,120,102,102,59,32,105,32,61,32,105,32,43,32,52,32,124,32,48,59,32,100,97,116,97,56,91,105,93,32,61,32,48,120,102,102,59,32,105,32,61,32,105,32,43,32,52,32,124,32,48,59,32,100,97,116,97,56,91,105,93,32,61,32,48,120,102,102,59,32,105,32,61,32,105,32,43,32,52,32,124,32,48,59,32,125,32,125,32,101,108,115,101,32,123,32,119,104,105,108,101,32,40,105,32,60,32,106,41,32,123,32,100,97,116,97,56,91,105,93,32,61,32,48,120,102,102,59,32,105,32,61,32,105,32,43,32,52,32,124,32,48,59,32,125,32,125,32,125,32,83,68,76,50,46,99,116,120,46,112,117,116,73,109,97,103,101,68,97,116,97,40,83,68,76,50,46,105,109,97,103,101,44,32,48,44,32,48,41,59,32,114,101,116,117,114,110,32,48,59,32,125,0,67,111,117,108,100,32,110,111,116,32,103,101,116,32,69,71,76,32,100,105,115,112,108,97,121,0,67,111,117,108,100,32,110,111,116,32,105,110,105,116,105,97,108,105,122,101,32,69,71,76,0,70,97,105,108,101,100,32,108,111,97,100,105,110,103,32,37,115,58,32,37,115,0,67,111,117,108,100,32,110,111,116,32,99,114,101,97,116,101,32,69,71,76,32,99,111,110,116,101,120,116,32,40,99,111,110,116,101,120,116,32,97,116,116,114,105,98,117,116,101,115,32,97,114,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,41,0,67,111,117,108,100,32,110,111,116,32,99,114,101,97,116,101,32,69,71,76,32,99,111,110,116,101,120,116,0,79,112,101,110,71,76,32,110,111,116,32,105,110,105,116,105,97,108,105,122,101,100,0,85,110,97,98,108,101,32,116,111,32,109,97,107,101,32,69,71,76,32,99,111,110,116,101,120,116,32,99,117,114,114,101,110,116,0,67,111,117,108,100,32,110,111,116,32,109,97,107,101,32,69,71,76,32,99,111,110,116,101,120,116,32,99,117,114,114,101,110,116,0,69,71,76,32,110,111,116,32,105,110,105,116,105,97,108,105,122,101,100,0,85,110,97,98,108,101,32,116,111,32,115,101,116,32,116,104,101,32,69,71,76,32,115,119,97,112,32,105,110,116,101,114,118,97,108,0,83,68,76,32,101,109,115,99,114,105,112,116,101,110,32,97,117,100,105,111,32,100,114,105,118,101,114,0,78,111,32,99,111,109,112,97,116,105,98,108,101,32,97,117,100,105,111,32,102,111,114,109,97,116,33,0,123,32,105,102,40,116,121,112,101,111,102,40,83,68,76,50,41,32,61,61,61,32,39,117,110,100,101,102,105,110,101,100,39,41,32,83,68,76,50,32,61,32,123,125,59,32,105,102,40,116,121,112,101,111,102,40,83,68,76,50,46,97,117,100,105,111,41,32,61,61,61,32,39,117,110,100,101,102,105,110,101,100,39,41,32,83,68,76,50,46,97,117,100,105,111,32,61,32,123,125,59,32,105,102,32,40,33,83,68,76,50,46,97,117,100,105,111,67,111,110,116,101,120,116,41,32,123,32,105,102,32,40,116,121,112,101,111,102,40,65,117,100,105,111,67,111,110,116,101,120,116,41,32,33,61,61,32,39,117,110,100,101,102,105,110,101,100,39,41,32,123,32,83,68,76,50,46,97,117,100,105,111,67,111,110,116,101,120,116,32,61,32,110,101,119,32,65,117,100,105,111,67,111,110,116,101,120,116,40,41,59,32,125,32,101,108,115,101,32,105,102,32,40,116,121,112,101,111,102,40,119,101,98,107,105,116,65,117,100,105,111,67,111,110,116,101,120,116,41,32,33,61,61,32,39,117,110,100,101,102,105,110,101,100,39,41,32,123,32,83,68,76,50,46,97,117,100,105,111,67,111,110,116,101,120,116,32,61,32,110,101,119,32,119,101,98,107,105,116,65,117,100,105,111,67,111,110,116,101,120,116,40,41,59,32,125,32,101,108,115,101,32,123,32,114,101,116,117,114,110,32,45,49,59,32,125,32,125,32,114,101,116,117,114,110,32,48,59,32,125,0,87,101,98,32,65,117,100,105,111,32,65,80,73,32,105,115,32,110,111,116,32,97,118,97,105,108,97,98,108,101,33,0,123,32,114,101,116,117,114,110,32,83,68,76,50,46,97,117,100,105,111,67,111,110,116,101,120,116,91,39,115,97,109,112,108,101,82,97,116,101,39,93,59,32,125,0,123,32,83,68,76,50,46,97,117,100,105,111,46,115,99,114,105,112,116,80,114,111,99,101,115,115,111,114,78,111,100,101,32,61,32,83,68,76,50,46,97,117,100,105,111,67,111,110,116,101,120,116,91,39,99,114,101,97,116,101,83,99,114,105,112,116,80,114,111,99,101,115,115,111,114,39,93,40,36,49,44,32,48,44,32,36,48,41,59,32,83,68,76,50,46,97,117,100,105,111,46,115,99,114,105,112,116,80,114,111,99,101,115,115,111,114,78,111,100,101,91,39,111,110,97,117,100,105,111,112,114,111,99,101,115,115,39,93,32,61,32,102,117,110,99,116,105,111,110,32,40,101,41,32,123,32,83,68,76,50,46,97,117,100,105,111,46,99,117,114,114,101,110,116,79,117,116,112,117,116,66,117,102,102,101,114,32,61,32,101,91,39,111,117,116,112,117,116,66,117,102,102,101,114,39,93,59,32,82,117,110,116,105,109,101,46,100,121,110,67,97,108,108,40,39,118,105,39,44,32,36,50,44,32,91,36,51,93,41,59,32,125,59,32,83,68,76,50,46,97,117,100,105,111,46,115,99,114,105,112,116,80,114,111,99,101,115,115,111,114,78,111,100,101,91,39,99,111,110,110,101,99,116,39,93,40,83,68,76,50,46,97,117,100,105,111,67,111,110,116,101,120,116,91,39,100,101,115,116,105,110,97,116,105,111,110,39,93,41,59,32,125,0,78,111,32,98,117,102,102,101,114,32,97,108,108,111,99,97,116,101,100,32,102,111,114,32,99,111,110,118,101,114,115,105,111,110,0,123,32,118,97,114,32,110,117,109,67,104,97,110,110,101,108,115,32,61,32,83,68,76,50,46,97,117,100,105,111,46,99,117,114,114,101,110,116,79,117,116,112,117,116,66,117,102,102,101,114,91,39,110,117,109,98,101,114,79,102,67,104,97,110,110,101,108,115,39,93,59,32,102,111,114,32,40,118,97,114,32,99,32,61,32,48,59,32,99,32,60,32,110,117,109,67,104,97,110,110,101,108,115,59,32,43,43,99,41,32,123,32,118,97,114,32,99,104,97,110,110,101,108,68,97,116,97,32,61,32,83,68,76,50,46,97,117,100,105,111,46,99,117,114,114,101,110,116,79,117,116,112,117,116,66,117,102,102,101,114,91,39,103,101,116,67,104,97,110,110,101,108,68,97,116,97,39,93,40,99,41,59,32,105,102,32,40,99,104,97,110,110,101,108,68,97,116,97,46,108,101,110,103,116,104,32,33,61,32,36,49,41,32,123,32,116,104,114,111,119,32,39,87,101,98,32,65,117,100,105,111,32,111,117,116,112,117,116,32,98,117,102,102,101,114,32,108,101,110,103,116,104,32,109,105,115,109,97,116,99,104,33,32,68,101,115,116,105,110,97,116,105,111,110,32,115,105,122,101,58,32,39,32,43,32,99,104,97,110,110,101,108,68,97,116,97,46,108,101,110,103,116,104,32,43,32,39,32,115,97,109,112,108,101,115,32,118,115,32,101,120,112,101,99,116,101,100,32,39,32,43,32,36,49,32,43,32,39,32,115,97,109,112,108,101,115,33,39,59,32,125,32,102,111,114,32,40,118,97,114,32,106,32,61,32,48,59,32,106,32,60,32,36,49,59,32,43,43,106,41,32,123,32,99,104,97,110,110,101,108,68,97,116,97,91,106,93,32,61,32,72,69,65,80,70,51,50,91,36,48,32,43,32,40,40,106,42,110,117,109,67,104,97,110,110,101,108,115,32,43,32,99,41,32,60,60,32,50,41,32,62,62,32,50,93,59,32,125,32,125,32,125,0,123,32,105,102,32,40,116,121,112,101,111,102,40,65,117,100,105,111,67,111,110,116,101,120,116,41,32,33,61,61,32,39,117,110,100,101,102,105,110,101,100,39,41,32,123,32,114,101,116,117,114,110,32,49,59,32,125,32,101,108,115,101,32,105,102,32,40,116,121,112,101,111,102,40,119,101,98,107,105,116,65,117,100,105,111,67,111,110,116,101,120,116,41,32,33,61,61,32,39,117,110,100,101,102,105,110,101,100,39,41,32,123,32,114,101,116,117,114,110,32,49,59,32,125,32,114,101,116,117,114,110,32,48,59,32,125,0,78,111,32,97,117,100,105,111,32,99,111,110,116,101,120,116,32,97,118,97,105,108,97,98,108,101,0,83,68,76,32,100,117,109,109,121,32,118,105,100,101,111,32,100,114,105,118,101,114,0,95,83,68,76,95,68,117,109,109,121,83,117,114,102,97,99,101,0,67,111,117,108,100,110,39,116,32,102,105,110,100,32,100,117,109,109,121,32,115,117,114,102,97,99,101,32,102,111,114,32,119,105,110,100,111,119,0,83,68,76,95,86,73,68,69,79,95,68,85,77,77,89,95,83,65,86,69,95,70,82,65,77,69,83,0,83,68,76,95,119,105,110,100,111,119,37,100,45,37,56,46,56,100,46,98,109,112,0,83,68,76,95,82,87,70,114,111,109,70,105,108,101,40,41,58,32,78,111,32,102,105,108,101,32,111,114,32,110,111,32,109,111,100,101,32,115,112,101,99,105,102,105,101,100,0,67,111,117,108,100,110,39,116,32,111,112,101,110,32,37,115,0,37,100,32,98,112,112,32,66,77,80,32,102,105,108,101,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,67,111,117,108,100,110,39,116,32,99,111,110,118,101,114,116,32,105,109,97,103,101,32,116,111,32,37,100,32,98,112,112,0,100,105,115,107,0,100,105,114,101,99,116,45,116,111,45,100,105,115,107,32,97,117,100,105,111,0,83,68,76,95,68,73,83,75,65,85,68,73,79,70,73,76,69,0,115,100,108,97,117,100,105,111,46,114,97,119,0,83,68,76,95,68,73,83,75,65,85,68,73,79,68,69,76,65,89,0,119,98,0,87,65,82,78,73,78,71,58,32,89,111,117,32,97,114,101,32,117,115,105,110,103,32,116,104,101,32,83,68,76,32,100,105,115,107,32,119,114,105,116,101,114,32,97,117,100,105,111,32,100,114,105,118,101,114,33,10,32,87,114,105,116,105,110,103,32,116,111,32,102,105,108,101,32,91,37,115,93,46,10,0,100,117,109,109,121,0,83,68,76,32,100,117,109,109,121,32,97,117,100,105,111,32,100,114,105,118,101,114,0,83,68,76,95,78,79,95,83,73,71,78,65,76,95,72,65,78,68,76,69,82,83,0,83,68,76,32,110,111,116,32,98,117,105,108,116,32,119,105,116,104,32,116,104,114,101,97,100,32,115,117,112,112,111,114,116,0,83,68,76,84,105,109,101,114,0,84,104,114,101,97,100,115,32,97,114,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,111,110,32,116,104,105,115,32,112,108,97,116,102,111,114,109,0,83,68,76,95,77,79,85,83,69,95,82,69,76,65,84,73,86,69,95,77,79,68,69,95,87,65,82,80,0,88,98,111,120,0,88,45,66,111,120,0,83,68,76,95,69,86,69,78,84,95,81,85,69,85,69,95,83,84,65,84,73,83,84,73,67,83,0,83,68,76,32,69,86,69,78,84,32,81,85,69,85,69,58,32,77,97,120,105,109,117,109,32,101,118,101,110,116,115,32,105,110,45,102,108,105,103,104,116,58,32,37,100,10,0,83,68,76,95,86,73,68,69,79,68,82,73,86,69,82,0,37,115,32,110,111,116,32,97,118,97,105,108,97,98,108,101,0,78,111,32,97,118,97,105,108,97,98,108,101,32,118,105,100,101,111,32,100,101,118,105,99,101,0,84,104,101,32,118,105,100,101,111,32,100,114,105,118,101,114,32,100,105,100,32,110,111,116,32,97,100,100,32,97,110,121,32,100,105,115,112,108,97,121,115,0,115,111,102,116,119,97,114,101,0,83,68,76,95,70,82,65,77,69,66,85,70,70,69,82,95,65,67,67,69,76,69,82,65,84,73,79,78,0,95,83,68,76,95,87,105,110,100,111,119,84,101,120,116,117,114,101,68,97,116,97,0,73,110,118,97,108,105,100,32,119,105,110,100,111,119,0,95,83,68,76,95,87,105,110,100,111,119,82,101,110,100,101,114,68,97,116,97,0,82,101,110,100,101,114,101,114,32,97,108,114,101,97,100,121,32,97,115,115,111,99,105,97,116,101,100,32,119,105,116,104,32,119,105,110,100,111,119,0,83,68,76,95,82,69,78,68,69,82,95,86,83,89,78,67,0,83,68,76,95,82,69,78,68,69,82,95,68,82,73,86,69,82,0,67,111,117,108,100,110,39,116,32,102,105,110,100,32,109,97,116,99,104,105,110,103,32,114,101,110,100,101,114,32,100,114,105,118,101,114,0,105,110,100,101,120,32,109,117,115,116,32,98,101,32,45,49,32,111,114,32,105,110,32,116,104,101,32,114,97,110,103,101,32,111,102,32,48,32,45,32,37,100,0,84,101,120,116,117,114,101,32,119,97,115,32,110,111,116,32,99,114,101,97,116,101,100,32,119,105,116,104,32,116,104,105,115,32,114,101,110,100,101,114,101,114,0,84,101,120,116,117,114,101,32,110,111,116,32,99,114,101,97,116,101,100,32,119,105,116,104,32,83,68,76,95,84,69,88,84,85,82,69,65,67,67,69,83,83,95,84,65,82,71,69,84,0,67,114,101,97,116,101,100,32,114,101,110,100,101,114,101,114,58,32,37,115,0,78,111,32,104,97,114,100,119,97,114,101,32,97,99,99,101,108,101,114,97,116,101,100,32,114,101,110,100,101,114,101,114,115,32,97,118,97,105,108,97,98,108,101,0,73,110,118,97,108,105,100,32,116,101,120,116,117,114,101,32,102,111,114,109,97,116,0,80,97,108,101,116,116,105,122,101,100,32,116,101,120,116,117,114,101,115,32,97,114,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,84,101,120,116,117,114,101,32,100,105,109,101,110,115,105,111,110,115,32,99,97,110,39,116,32,98,101,32,48,0,84,101,120,116,117,114,101,32,100,105,109,101,110,115,105,111,110,115,32,97,114,101,32,108,105,109,105,116,101,100,32,116,111,32,37,100,120,37,100,0,85,110,115,117,112,112,111,114,116,101,100,32,89,85,86,32,102,111,114,109,97,116,0,78,111,32,119,105,110,100,111,119,32,116,101,120,116,117,114,101,32,100,97,116,97,0,119,105,100,116,104,0,104,101,105,103,104,116,0,114,101,99,116,115,0,115,112,97,110,0,110,117,109,114,101,99,116,115,0,112,105,120,101,108,115,0,112,105,116,99,104,0,83,68,76,95,76,111,99,107,84,101,120,116,117,114,101,40,41,58,32,116,101,120,116,117,114,101,32,109,117,115,116,32,98,101,32,115,116,114,101,97,109,105,110,103,0,89,86,49,50,32,97,110,100,32,73,89,85,86,32,116,101,120,116,117,114,101,115,32,111,110,108,121,32,115,117,112,112,111,114,116,32,102,117,108,108,32,115,117,114,102,97,99,101,32,108,111,99,107,115,0,73,110,118,97,108,105,100,32,116,97,114,103,101,116,32,112,105,120,101,108,32,102,111,114,109,97,116,0,85,110,115,117,112,112,111,114,116,101,100,32,89,85,86,32,100,101,115,116,105,110,97,116,105,111,110,32,102,111,114,109,97,116,0,85,110,115,117,112,112,111,114,116,101,100,32,89,85,86,32,102,111,114,109,97,116,32,105,110,32,99,111,112,121,0,83,68,76,95,86,73,68,69,79,95,65,76,76,79,87,95,83,67,82,69,69,78,83,65,86,69,82,0,83,68,76,95,65,85,68,73,79,68,82,73,86,69,82,0,65,117,100,105,111,32,116,97,114,103,101,116,32,39,37,115,39,32,110,111,116,32,97,118,97,105,108,97,98,108,101,0,78,111,32,97,118,97,105,108,97,98,108,101,32,97,117,100,105,111,32,100,101,118,105,99,101,0,83,121,115,116,101,109,32,97,117,100,105,111,32,111,117,116,112,117,116,32,100,101,118,105,99,101,0,83,121,115,116,101,109,32,97,117,100,105,111,32,99,97,112,116,117,114,101,32,100,101,118,105,99,101,0,83,68,76,95,74,79,89,83,84,73,67,75,95,65,76,76,79,87,95,66,65,67,75,71,82,79,85,78,68,95,69,86,69,78,84,83,0,110,97,109,101,0,99,97,108,108,98,97,99,107,0,71,97,109,101,112,97,100,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,67,111,117,108,100,32,110,111,116,32,115,101,116,32,103,97,109,101,112,97,100,32,99,111,110,110,101,99,116,32,99,97,108,108,98,97,99,107,0,67,111,117,108,100,32,110,111,116,32,115,101,116,32,103,97,109,101,112,97,100,32,100,105,115,99,111,110,110,101,99,116,32,99,97,108,108,98,97,99,107,0,101,109,115,99,114,105,112,116,101,110,44,83,116,97,110,100,97,114,100,32,71,97,109,101,112,97,100,44,97,58,98,48,44,98,58,98,49,44,98,97,99,107,58,98,56,44,100,112,100,111,119,110,58,98,49,51,44,100,112,108,101,102,116,58,98,49,52,44,100,112,114,105,103,104,116,58,98,49,53,44,100,112,117,112,58,98,49,50,44,103,117,105,100,101,58,98,49,54,44,108,101,102,116,115,104,111,117,108,100,101,114,58,98,52,44,108,101,102,116,115,116,105,99,107,58,98,49,48,44,108,101,102,116,116,114,105,103,103,101,114,58,98,54,44,108,101,102,116,120,58,97,48,44,108,101,102,116,121,58,97,49,44,114,105,103,104,116,115,104,111,117,108,100,101,114,58,98,53,44,114,105,103,104,116,115,116,105,99,107,58,98,49,49,44,114,105,103,104,116,116,114,105,103,103,101,114,58,98,55,44,114,105,103,104,116,120,58,97,50,44,114,105,103,104,116,121,58,97,51,44,115,116,97,114,116,58,98,57,44,120,58,98,50,44,121,58,98,51,44,0,109,97,112,112,105,110,103,83,116,114,105,110,103,0,67,111,117,108,100,110,39,116,32,112,97,114,115,101,32,71,85,73,68,32,102,114,111,109,32,37,115,0,120,105,110,112,117,116,0,101,109,115,99,114,105,112,116,101,110,0,67,111,117,108,100,110,39,116,32,112,97,114,115,101,32,110,97,109,101,32,102,114,111,109,32,37,115,0,67,111,117,108,100,110,39,116,32,112,97,114,115,101,32,37,115,0,97,0,98,0,120,0,121,0,98,97,99,107,0,103,117,105,100,101,0,115,116,97,114,116,0,108,101,102,116,115,116,105,99,107,0,114,105,103,104,116,115,116,105,99,107,0,108,101,102,116,115,104,111,117,108,100,101,114,0,114,105,103,104,116,115,104,111,117,108,100,101,114,0,100,112,117,112,0,100,112,100,111,119,110,0,100,112,108,101,102,116,0,100,112,114,105,103,104,116,0,108,101,102,116,120,0,108,101,102,116,121,0,114,105,103,104,116,120,0,114,105,103,104,116,121,0,108,101,102,116,116,114,105,103,103,101,114,0,114,105,103,104,116,116,114,105,103,103,101,114,0,65,120,105,115,32,105,110,100,101,120,32,116,111,111,32,108,97,114,103,101,58,32,37,100,0,66,117,116,116,111,110,32,105,110,100,101,120,32,116,111,111,32,108,97,114,103,101,58,32,37,100,0,72,97,116,32,105,110,100,101,120,32,116,111,111,32,108,97,114,103,101,58,32,37,100,0,66,117,116,116,111,110,32,110,97,109,101,32,116,111,111,32,108,97,114,103,101,58,32,37,115,0,74,111,121,115,116,105,99,107,32,98,117,116,116,111,110,32,110,97,109,101,32,116,111,111,32,108,97,114,103,101,58,32,37,115,0,83,68,76,95,71,65,77,69,67,79,78,84,82,79,76,76,69,82,67,79,78,70,73,71,0,83,68,76,32,110,111,116,32,98,117,105,108,116,32,119,105,116,104,32,104,97,112,116,105,99,32,40,102,111,114,99,101,32,102,101,101,100,98,97,99,107,41,32,115,117,112,112,111,114,116,0,109,101,115,115,97,103,101,98,111,120,100,97,116,97,0,78,111,32,109,101,115,115,97,103,101,32,115,121,115,116,101,109,32,97,118,97,105,108,97,98,108,101,0,69,109,115,99,114,105,112,116,101,110,0,80,97,114,97,109,101,116,101,114,32,39,37,115,39,32,105,115,32,105,110,118,97,108,105,100,0,115,99,97,110,99,111,100,101,0,67,117,114,115,111,114,32,110,111,116,32,97,115,115,111,99,105,97,116,101,100,32,119,105,116,104,32,116,104,101,32,99,117,114,114,101,110,116,32,109,111,117,115,101,0,67,117,114,115,111,114,115,32,97,114,101,32,110,111,116,32,99,117,114,114,101,110,116,108,121,32,115,117,112,112,111,114,116,101,100,0,67,117,114,115,111,114,32,104,111,116,32,115,112,111,116,32,100,111,101,115,110,39,116,32,108,105,101,32,119,105,116,104,105,110,32,99,117,114,115,111,114,0,80,97,115,115,101,100,32,78,85,76,76,32,99,117,114,115,111,114,32,115,117,114,102,97,99,101,0,67,114,101,97,116,101,83,121,115,116,101,109,67,117,114,115,111,114,32,105,115,32,110,111,116,32,99,117,114,114,101,110,116,108,121,32,115,117,112,112,111,114,116,101,100,0,72,97,112,116,105,99,58,32,84,104,101,114,101,32,97,114,101,32,37,100,32,104,97,112,116,105,99,32,100,101,118,105,99,101,115,32,97,118,97,105,108,97,98,108,101,0,76,111,103,105,99,32,101,114,114,111,114,58,32,78,111,32,104,97,112,116,105,99,32,100,101,118,105,99,101,115,32,97,118,97,105,108,97,98,108,101,46,0,72,97,112,116,105,99,58,32,73,110,118,97,108,105,100,32,104,97,112,116,105,99,32,100,101,118,105,99,101,32,105,100,101,110,116,105,102,105,101,114,0,74,111,121,115,116,105,99,107,32,104,97,115,110,39,116,32,98,101,101,110,32,111,112,101,110,101,100,32,121,101,116,0,72,97,112,116,105,99,58,32,74,111,121,115,116,105,99,107,32,105,115,110,39,116,32,118,97,108,105,100,46,0,72,97,112,116,105,99,58,32,74,111,121,115,116,105,99,107,32,105,115,110,39,116,32,97,32,104,97,112,116,105,99,32,100,101,118,105,99,101,46,0,72,97,112,116,105,99,58,32,73,110,118,97,108,105,100,32,101,102,102,101,99,116,32,105,100,101,110,116,105,102,105,101,114,46,0,72,97,112,116,105,99,58,32,69,102,102,101,99,116,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,98,121,32,104,97,112,116,105,99,32,100,101,118,105,99,101,46,0,72,97,112,116,105,99,58,32,68,101,118,105,99,101,32,104,97,115,32,110,111,32,102,114,101,101,32,115,112,97,99,101,32,108,101,102,116,46,0,72,97,112,116,105,99,58,32,85,112,100,97,116,105,110,103,32,101,102,102,101,99,116,32,116,121,112,101,32,105,115,32,105,108,108,101,103,97,108,46,0,72,97,112,116,105,99,58,32,68,101,118,105,99,101,32,100,111,101,115,32,110,111,116,32,115,117,112,112,111,114,116,32,115,116,97,116,117,115,32,113,117,101,114,105,101,115,46,0,48,49,50,51,52,53,54,55,56,57,97,98,99,100,101,102,0,37,115,44,37,115,44,37,115,0,84,104,101,114,101,32,97,114,101,32,37,100,32,106,111,121,115,116,105,99,107,115,32,97,118,97,105,108,97,98,108,101,0,67,111,117,108,100,110,39,116,32,102,105,110,100,32,109,97,112,112,105,110,103,32,102,111,114,32,100,101,118,105,99,101,32,40,37,100,41,0,78,111,32,115,117,99,104,32,100,101,118,105,99,101,0,74,111,121,115,116,105,99,107,32,97,108,114,101,97,100,121,32,111,112,101,110,101,100,0,74,111,121,115,116,105,99,107,32,111,110,108,121,32,104,97,115,32,37,100,32,97,120,101,115,0,74,111,121,115,116,105,99,107,32,111,110,108,121,32,104,97,115,32,37,100,32,98,117,116,116,111,110,115,0,74,111,121,115,116,105,99,107,32,111,110,108,121,32,104,97,115,32,37,100,32,104,97,116,115,0,87,105,110,100,111,119,32,105,115,32,116,111,111,32,108,97,114,103,101,46,0,83,68,76,95,86,73,68,69,79,95,72,73,71,72,68,80,73,95,68,73,83,65,66,76,69,68,0,86,105,100,101,111,32,115,117,98,115,121,115,116,101,109,32,109,117,115,116,32,98,101,32,105,110,105,116,105,97,108,105,122,101,100,32,116,111,32,115,101,116,32,99,108,105,112,98,111,97,114,100,32,116,101,120,116,0,86,105,100,101,111,32,115,117,98,115,121,115,116,101,109,32,109,117,115,116,32,98,101,32,105,110,105,116,105,97,108,105,122,101,100,32,116,111,32,103,101,116,32,99,108,105,112,98,111,97,114,100,32,116,101,120,116,0,105,110,100,101,120,32,109,117,115,116,32,98,101,32,105,110,32,116,104,101,32,114,97,110,103,101,32,111,102,32,48,32,45,32,37,100,0,119,0,104,0,109,105,110,95,119,0,109,105,110,95,104,0,79,75,0,80,97,115,115,101,100,32,97,32,78,85,76,76,32,99,111,110,100,105,116,105,111,110,32,118,97,114,105,97,98,108,101,0,105,110,99,111,114,114,101,99,116,32,104,101,97,100,101,114,32,99,104,101,99,107,0,117,110,107,110,111,119,110,32,99,111,109,112,114,101,115,115,105,111,110,32,109,101,116,104,111,100,0,105,110,118,97,108,105,100,32,119,105,110,100,111,119,32,115,105,122,101,0,117,110,107,110,111,119,110,32,104,101,97,100,101,114,32,102,108,97,103,115,32,115,101,116,0,104,101,97,100,101,114,32,99,114,99,32,109,105,115,109,97,116,99,104,0,105,110,118,97,108,105,100,32,98,108,111,99,107,32,116,121,112,101,0,105,110,118,97,108,105,100,32,115,116,111,114,101,100,32,98,108,111,99,107,32,108,101,110,103,116,104,115,0,116,111,111,32,109,97,110,121,32,108,101,110,103,116,104,32,111,114,32,100,105,115,116,97,110,99,101,32,115,121,109,98,111,108,115,0,105,110,118,97,108,105,100,32,99,111,100,101,32,108,101,110,103,116,104,115,32,115,101,116,0,105,110,118,97,108,105,100,32,98,105,116,32,108,101,110,103,116,104,32,114,101,112,101,97,116,0,105,110,118,97,108,105,100,32,99,111,100,101,32,45,45,32,109,105,115,115,105,110,103,32,101,110,100,45,111,102,45,98,108,111,99,107,0,105,110,118,97,108,105,100,32,108,105,116,101,114,97,108,47,108,101,110,103,116,104,115,32,115,101,116,0,105,110,118,97,108,105,100,32,100,105,115,116,97,110,99,101,115,32,115,101,116,0,105,110,99,111,114,114,101,99,116,32,100,97,116,97,32,99,104,101,99,107,0,105,110,99,111,114,114,101,99,116,32,108,101,110,103,116,104,32,99,104,101,99,107,0,0,1,2,3,4,4,5,5,6,6,6,6,7,7,7,7,8,8,8,8,8,8,8,8,9,9,9,9,9,9,9,9,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,0,0,16,17,18,18,19,19,20,20,20,20,21,21,21,21,22,22,22,22,22,22,22,22,23,23,23,23,23,23,23,23,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,0,1,2,3,4,5,6,7,8,8,9,9,10,10,11,11,12,12,12,12,13,13,13,13,14,14,14,14,15,15,15,15,16,16,16,16,16,16,16,16,17,17,17,17,17,17,17,17,18,18,18,18,18,18,18,18,19,19,19,19,19,19,19,19,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,28,16,17,18,0,8,7,9,6,10,5,11,4,12,3,13,2,14,1,15,49,46,50,46,56,0,110,101,101,100,32,100,105,99,116,105,111,110,97,114,121,0,115,116,114,101,97,109,32,101,110,100,0,102,105,108,101,32,101,114,114,111,114,0,115,116,114,101,97,109,32,101,114,114,111,114,0,100,97,116,97,32,101,114,114,111,114,0,105,110,115,117,102,102,105,99,105,101,110,116,32,109,101,109,111,114,121,0,98,117,102,102,101,114,32,101,114,114,111,114,0,105,110,99,111,109,112,97,116,105,98,108,101,32,118,101,114,115,105,111,110,0,105,110,118,97,108,105,100,32,100,105,115,116,97,110,99,101,32,116,111,111,32,102,97,114,32,98,97,99,107,0,105,110,118,97,108,105,100,32,100,105,115,116,97,110,99,101,32,99,111,100,101,0,105,110,118,97,108,105,100,32,108,105,116,101,114,97,108,47,108,101,110,103,116,104,32,99,111,100,101,0,69,88,84,0,65,82,66,0,79,69,83,0,65,78,71,76,69,0,103,108,67,114,101,97,116,101,80,114,111,103,114,97,109,79,98,106,101,99,116,0,103,108,67,114,101,97,116,101,80,114,111,103,114,97,109,0,103,108,85,115,101,80,114,111,103,114,97,109,79,98,106,101,99,116,0,103,108,85,115,101,80,114,111,103,114,97,109,0,103,108,67,114,101,97,116,101,83,104,97,100,101,114,79,98,106,101,99,116,0,103,108,67,114,101,97,116,101,83,104,97,100,101,114,0,103,108,65,116,116,97,99,104,79,98,106,101,99,116,0,103,108,65,116,116,97,99,104,83,104,97,100,101,114,0,103,108,68,101,116,97,99,104,79,98,106,101,99,116,0,103,108,68,101,116,97,99,104,83,104,97,100,101,114,0,103,108,80,105,120,101,108,83,116,111,114,101,105,0,103,108,71,101,116,83,116,114,105,110,103,0,103,108,71,101,116,73,110,116,101,103,101,114,118,0,103,108,71,101,116,70,108,111,97,116,118,0,103,108,71,101,116,66,111,111,108,101,97,110,118,0,103,108,71,101,110,84,101,120,116,117,114,101,115,0,103,108,68,101,108,101,116,101,84,101,120,116,117,114,101,115,0,103,108,67,111,109,112,114,101,115,115,101,100,84,101,120,73,109,97,103,101,50,68,0,103,108,67,111,109,112,114,101,115,115,101,100,84,101,120,83,117,98,73,109,97,103,101,50,68,0,103,108,84,101,120,73,109,97,103,101,50,68,0,103,108,84,101,120,83,117,98,73,109,97,103,101,50,68,0,103,108,82,101,97,100,80,105,120,101,108,115,0,103,108,66,105,110,100,84,101,120,116,117,114,101,0,103,108,71,101,116,84,101,120,80,97,114,97,109,101,116,101,114,102,118,0,103,108,71,101,116,84,101,120,80,97,114,97,109,101,116,101,114,105,118,0,103,108,84,101,120,80,97,114,97,109,101,116,101,114,102,118,0,103,108,84,101,120,80,97,114,97,109,101,116,101,114,105,118,0,103,108,73,115,84,101,120,116,117,114,101,0,103,108,71,101,110,66,117,102,102,101,114,115,0,103,108,68,101,108,101,116,101,66,117,102,102,101,114,115,0,103,108,71,101,116,66,117,102,102,101,114,80,97,114,97,109,101,116,101,114,105,118,0,103,108,66,117,102,102,101,114,68,97,116,97,0,103,108,66,117,102,102,101,114,83,117,98,68,97,116,97,0,103,108,73,115,66,117,102,102,101,114,0,103,108,71,101,110,82,101,110,100,101,114,98,117,102,102,101,114,115,0,103,108,68,101,108,101,116,101,82,101,110,100,101,114,98,117,102,102,101,114,115,0,103,108,66,105,110,100,82,101,110,100,101,114,98,117,102,102,101,114,0,103,108,71,101,116,82,101,110,100,101,114,98,117,102,102,101,114,80,97,114,97,109,101,116,101,114,105,118,0,103,108,73,115,82,101,110,100,101,114,98,117,102,102,101,114,0,103,108,71,101,116,85,110,105,102,111,114,109,102,118,0,103,108,71,101,116,85,110,105,102,111,114,109,105,118,0,103,108,71,101,116,85,110,105,102,111,114,109,76,111,99,97,116,105,111,110,0,103,108,71,101,116,86,101,114,116,101,120,65,116,116,114,105,98,102,118,0,103,108,71,101,116,86,101,114,116,101,120,65,116,116,114,105,98,105,118,0,103,108,71,101,116,86,101,114,116,101,120,65,116,116,114,105,98,80,111,105,110,116,101,114,118,0,103,108,71,101,116,65,99,116,105,118,101,85,110,105,102,111,114,109,0,103,108,85,110,105,102,111,114,109,49,102,0,103,108,85,110,105,102,111,114,109,50,102,0,103,108,85,110,105,102,111,114,109,51,102,0,103,108,85,110,105,102,111,114,109,52,102,0,103,108,85,110,105,102,111,114,109,49,105,0,103,108,85,110,105,102,111,114,109,50,105,0,103,108,85,110,105,102,111,114,109,51,105,0,103,108,85,110,105,102,111,114,109,52,105,0,103,108,85,110,105,102,111,114,109,49,105,118,0,103,108,85,110,105,102,111,114,109,50,105,118,0,103,108,85,110,105,102,111,114,109,51,105,118,0,103,108,85,110,105,102,111,114,109,52,105,118,0,103,108,85,110,105,102,111,114,109,49,102,118,0,103,108,85,110,105,102,111,114,109,50,102,118,0,103,108,85,110,105,102,111,114,109,51,102,118,0,103,108,85,110,105,102,111,114,109,52,102,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,50,102,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,51,102,118,0,103,108,85,110,105,102,111,114,109,77,97,116,114,105,120,52,102,118,0,103,108,66,105,110,100,66,117,102,102,101,114,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,49,102,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,50,102,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,51,102,118,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,102,118,0,103,108,71,101,116,65,116,116,114,105,98,76,111,99,97,116,105,111,110,0,103,108,71,101,116,65,99,116,105,118,101,65,116,116,114,105,98,0,103,108,68,101,108,101,116,101,83,104,97,100,101,114,0,103,108,71,101,116,65,116,116,97,99,104,101,100,83,104,97,100,101,114,115,0,103,108,83,104,97,100,101,114,83,111,117,114,99,101,0,103,108,71,101,116,83,104,97,100,101,114,83,111,117,114,99,101,0,103,108,67,111,109,112,105,108,101,83,104,97,100,101,114,0,103,108,71,101,116,83,104,97,100,101,114,73,110,102,111,76,111,103,0,103,108,71,101,116,83,104,97,100,101,114,105,118,0,103,108,71,101,116,80,114,111,103,114,97,109,105,118,0,103,108,73,115,83,104,97,100,101,114,0,103,108,68,101,108,101,116,101,80,114,111,103,114,97,109,0,103,108,71,101,116,83,104,97,100,101,114,80,114,101,99,105,115,105,111,110,70,111,114,109,97,116,0,103,108,76,105,110,107,80,114,111,103,114,97,109,0,103,108,71,101,116,80,114,111,103,114,97,109,73,110,102,111,76,111,103,0,103,108,86,97,108,105,100,97,116,101,80,114,111,103,114,97,109,0,103,108,73,115,80,114,111,103,114,97,109,0,103,108,66,105,110,100,65,116,116,114,105,98,76,111,99,97,116,105,111,110,0,103,108,66,105,110,100,70,114,97,109,101,98,117,102,102,101,114,0,103,108,71,101,110,70,114,97,109,101,98,117,102,102,101,114,115,0,103,108,68,101,108,101,116,101,70,114,97,109,101,98,117,102,102,101,114,115,0,103,108,70,114,97,109,101,98,117,102,102,101,114,82,101,110,100,101,114,98,117,102,102,101,114,0,103,108,70,114,97,109,101,98,117,102,102,101,114,84,101,120,116,117,114,101,50,68,0,103,108,71,101,116,70,114,97,109,101,98,117,102,102,101,114,65,116,116,97,99,104,109,101,110,116,80,97,114,97,109,101,116,101,114,105,118,0,103,108,73,115,70,114,97,109,101,98,117,102,102,101,114,0,103,108,68,101,108,101,116,101,79,98,106,101,99,116,0,103,108,71,101,116,79,98,106,101,99,116,80,97,114,97,109,101,116,101,114,105,118,0,103,108,71,101,116,73,110,102,111,76,111,103,0,103,108,66,105,110,100,80,114,111,103,114,97,109,0,103,108,71,101,116,80,111,105,110,116,101,114,118,0,103,108,68,114,97,119,82,97,110,103,101,69,108,101,109,101,110,116,115,0,103,108,69,110,97,98,108,101,67,108,105,101,110,116,83,116,97,116,101,0,103,108,86,101,114,116,101,120,80,111,105,110,116,101,114,0,103,108,84,101,120,67,111,111,114,100,80,111,105,110,116,101,114,0,103,108,78,111,114,109,97,108,80,111,105,110,116,101,114,0,103,108,67,111,108,111,114,80,111,105,110,116,101,114,0,103,108,67,108,105,101,110,116,65,99,116,105,118,101,84,101,120,116,117,114,101,0,103,108,71,101,110,86,101,114,116,101,120,65,114,114,97,121,115,0,103,108,68,101,108,101,116,101,86,101,114,116,101,120,65,114,114,97,121,115,0,103,108,66,105,110,100,86,101,114,116,101,120,65,114,114,97,121,0,103,108,77,97,116,114,105,120,77,111,100,101,0,103,108,76,111,97,100,73,100,101,110,116,105,116,121,0,103,108,76,111,97,100,77,97,116,114,105,120,102,0,103,108,70,114,117,115,116,117,109,0,103,108,82,111,116,97,116,101,102,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,80,111,105,110,116,101,114,0,103,108,69,110,97,98,108,101,86,101,114,116,101,120,65,116,116,114,105,98,65,114,114,97,121,0,103,108,68,105,115,97,98,108,101,86,101,114,116,101,120,65,116,116,114,105,98,65,114,114,97,121,0,103,108,68,114,97,119,65,114,114,97,121,115,0,103,108,68,114,97,119,69,108,101,109,101,110,116,115,0,103,108,83,104,97,100,101,114,66,105,110,97,114,121,0,103,108,82,101,108,101,97,115,101,83,104,97,100,101,114,67,111,109,112,105,108,101,114,0,103,108,71,101,116,69,114,114,111,114,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,68,105,118,105,115,111,114,0,103,108,68,114,97,119,65,114,114,97,121,115,73,110,115,116,97,110,99,101,100,0,103,108,68,114,97,119,69,108,101,109,101,110,116,115,73,110,115,116,97,110,99,101,100,0,103,108,70,105,110,105,115,104,0,103,108,70,108,117,115,104,0,103,108,67,108,101,97,114,68,101,112,116,104,0,103,108,67,108,101,97,114,68,101,112,116,104,102,0,103,108,68,101,112,116,104,70,117,110,99,0,103,108,69,110,97,98,108,101,0,103,108,68,105,115,97,98,108,101,0,103,108,70,114,111,110,116,70,97,99,101,0,103,108,67,117,108,108,70,97,99,101,0,103,108,67,108,101,97,114,0,103,108,76,105,110,101,87,105,100,116,104,0,103,108,67,108,101,97,114,83,116,101,110,99,105,108,0,103,108,68,101,112,116,104,77,97,115,107,0,103,108,83,116,101,110,99,105,108,77,97,115,107,0,103,108,67,104,101,99,107,70,114,97,109,101,98,117,102,102,101,114,83,116,97,116,117,115,0,103,108,71,101,110,101,114,97,116,101,77,105,112,109,97,112,0,103,108,65,99,116,105,118,101,84,101,120,116,117,114,101,0,103,108,66,108,101,110,100,69,113,117,97,116,105,111,110,0,103,108,73,115,69,110,97,98,108,101,100,0,103,108,66,108,101,110,100,70,117,110,99,0,103,108,66,108,101,110,100,69,113,117,97,116,105,111,110,83,101,112,97,114,97,116,101,0,103,108,68,101,112,116,104,82,97,110,103,101,0,103,108,68,101,112,116,104,82,97,110,103,101,102,0,103,108,83,116,101,110,99,105,108,77,97,115,107,83,101,112,97,114,97,116,101,0,103,108,72,105,110,116,0,103,108,80,111,108,121,103,111,110,79,102,102,115,101,116,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,49,102,0,103,108,83,97,109,112,108,101,67,111,118,101,114,97,103,101,0,103,108,84,101,120,80,97,114,97,109,101,116,101,114,105,0,103,108,84,101,120,80,97,114,97,109,101,116,101,114,102,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,50,102,0,103,108,83,116,101,110,99,105,108,70,117,110,99,0,103,108,83,116,101,110,99,105,108,79,112,0,103,108,86,105,101,119,112,111,114,116,0,103,108,67,108,101,97,114,67,111,108,111,114,0,103,108,83,99,105,115,115,111,114,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,51,102,0,103,108,67,111,108,111,114,77,97,115,107,0,103,108,82,101,110,100,101,114,98,117,102,102,101,114,83,116,111,114,97,103,101,0,103,108,66,108,101,110,100,70,117,110,99,83,101,112,97,114,97,116,101,0,103,108,66,108,101,110,100,67,111,108,111,114,0,103,108,83,116,101,110,99,105,108,70,117,110,99,83,101,112,97,114,97,116,101,0,103,108,83,116,101,110,99,105,108,79,112,83,101,112,97,114,97,116,101,0,103,108,86,101,114,116,101,120,65,116,116,114,105,98,52,102,0,103,108,67,111,112,121,84,101,120,73,109,97,103,101,50,68,0,103,108,67,111,112,121,84,101,120,83,117,98,73,109,97,103,101,50,68,0,103,108,68,114,97,119,66,117,102,102,101,114,115,0,123,32,77,111,100,117,108,101,46,112,114,105,110,116,69,114,114,40,39,98,97,100,32,110,97,109,101,32,105,110,32,103,101,116,80,114,111,99,65,100,100,114,101,115,115,58,32,39,32,43,32,91,80,111,105,110,116,101,114,95,115,116,114,105,110,103,105,102,121,40,36,48,41,44,32,80,111,105,110,116,101,114,95,115,116,114,105,110,103,105,102,121,40,36,49,41,93,41,59,32,125,0,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,0,1,2,3,4,5,6,7,8,9,255,255,255,255,255,255,255,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,255,255,255,255,255,255,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,0,1,2,4,7,3,6,5,0,17,0,10,0,17,17,17,0,0,0,0,5,0,0,0,0,0,0,9,0,0,0,0,11,0,0,0,0,0,0,0,0,17,0,15,10,17,17,17,3,10,7,0,1,19,9,11,11,0,0,9,6,11,0,0,11,0,6,17,0,0,0,17,17,17,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,11,0,0,0,0,0,0,0,0,17,0,10,10,17,17,17,0,10,0,0,2,0,9,11,0,0,0,9,0,11,0,0,11,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,12,0,0,0,0,0,0,0,0,0,0,0,12,0,0,0,0,12,0,0,0,0,9,12,0,0,0,0,0,12,0,0,12,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,14,0,0,0,0,0,0,0,0,0,0,0,13,0,0,0,4,13,0,0,0,0,9,14,0,0,0,0,0,14,0,0,14,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16,0,0,0,0,0,0,0,0,0,0,0,15,0,0,0,0,15,0,0,0,0,9,16,0,0,0,0,0,16,0,0,16,0,0,18,0,0,0,18,18,18,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,18,0,0,0,18,18,18,0,0,0,0,0,0,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,11,0,0,0,0,0,0,0,0,0,0,0,10,0,0,0,0,10,0,0,0,0,9,11,0,0,0,0,0,11,0,0,11,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,12,0,0,0,0,0,0,0,0,0,0,0,12,0,0,0,0,12,0,0,0,0,9,12,0,0,0,0,0,12,0,0,12,0,0,48,49,50,51,52,53,54,55,56,57,65,66,67,68,69,70,45,43,32,32,32,48,88,48,120,0,84,33,34,25,13,1,2,3,17,75,28,12,16,4,11,29,18,30,39,104,110,111,112,113,98,32,5,6,15,19,20,21,26,8,22,7,40,36,23,24,9,10,14,27,31,37,35,131,130,125,38,42,43,60,61,62,63,67,71,74,77,88,89,90,91,92,93,94,95,96,97,99,100,101,102,103,105,106,107,108,114,115,116,121,122,123,124,0,73,108,108,101,103,97,108,32,98,121,116,101,32,115,101,113,117,101,110,99,101,0,68,111,109,97,105,110,32,101,114,114,111,114,0,82,101,115,117,108,116,32,110,111,116,32,114,101,112,114,101,115,101,110,116,97,98,108,101,0,78,111,116,32,97,32,116,116,121,0,80,101,114,109,105,115,115,105,111,110,32,100,101,110,105,101,100,0,79,112,101,114,97,116,105,111,110,32,110,111,116,32,112,101,114,109,105,116,116,101,100,0,78,111,32,115,117,99,104,32,102,105,108,101,32,111,114,32,100,105,114,101,99,116,111,114,121,0,78,111,32,115,117,99,104,32,112,114,111,99,101,115,115,0,70,105,108,101,32,101,120,105,115,116,115,0,86,97,108,117,101,32,116,111,111,32,108,97,114,103,101,32,102,111,114,32,100,97,116,97,32,116,121,112,101,0,78,111,32,115,112,97,99,101,32,108,101,102,116,32,111,110,32,100,101,118,105,99,101,0,79,117,116,32,111,102,32,109,101,109,111,114,121,0,82,101,115,111,117,114,99,101,32,98,117,115,121,0,73,110,116,101,114,114,117,112,116,101,100,32,115,121,115,116,101,109,32,99,97,108,108,0,82,101,115,111,117,114,99,101,32,116,101,109,112,111,114,97,114,105,108,121,32,117,110,97,118,97,105,108,97,98,108,101,0,73,110,118,97,108,105,100,32,115,101,101,107,0,67,114,111,115,115,45,100,101,118,105,99,101,32,108,105,110,107,0,82,101,97,100,45,111,110,108,121,32,102,105,108,101,32,115,121,115,116,101,109,0,68,105,114,101,99,116,111,114,121,32,110,111,116,32,101,109,112,116,121,0,67,111,110,110,101,99,116,105,111,110,32,114,101,115,101,116,32,98,121,32,112,101,101,114,0,79,112,101,114,97,116,105,111,110,32,116,105,109,101,100,32,111,117,116,0,67,111,110,110,101,99,116,105,111,110,32,114,101,102,117,115,101,100,0,72,111,115,116,32,105,115,32,100,111,119,110,0,72,111,115,116,32,105,115,32,117,110,114,101,97,99,104,97,98,108,101,0,65,100,100,114,101,115,115,32,105,110,32,117,115,101,0,66,114,111,107,101,110,32,112,105,112,101,0,73,47,79,32,101,114,114,111,114,0,78,111,32,115,117,99,104,32,100,101,118,105,99,101,32,111,114,32,97,100,100,114,101,115,115,0,66,108,111,99,107,32,100,101,118,105,99,101,32,114,101,113,117,105,114,101,100,0,78,111,32,115,117,99,104,32,100,101,118,105,99,101,0,78,111,116,32,97,32,100,105,114,101,99,116,111,114,121,0,73,115,32,97,32,100,105,114,101,99,116,111,114,121,0,84,101,120,116,32,102,105,108,101,32,98,117,115,121,0,69,120,101,99,32,102,111,114,109,97,116,32,101,114,114,111,114,0,73,110,118,97,108,105,100,32,97,114,103,117,109,101,110,116,0,65,114,103,117,109,101,110,116,32,108,105,115,116,32,116,111,111,32,108,111,110,103,0,83,121,109,98,111,108,105,99,32,108,105,110,107,32,108,111,111,112,0,70,105,108,101,110,97,109,101,32,116,111,111,32,108,111,110,103,0,84,111,111,32,109,97,110,121,32,111,112,101,110,32,102,105,108,101,115,32,105,110,32,115,121,115,116,101,109,0,78,111,32,102,105,108,101,32,100,101,115,99,114,105,112,116,111,114,115,32,97,118,97,105,108,97,98,108,101,0,66,97,100,32,102,105,108,101,32,100,101,115,99,114,105,112,116,111,114,0,78,111,32,99,104,105,108,100,32,112,114,111,99,101,115,115,0,66,97,100,32,97,100,100,114,101,115,115,0,70,105,108,101,32,116,111,111,32,108,97,114,103,101,0,84,111,111,32,109,97,110,121,32,108,105,110,107,115,0,78,111,32,108,111,99,107,115,32,97,118,97,105,108,97,98,108,101,0,82,101,115,111,117,114,99,101,32,100,101,97,100,108,111,99,107,32,119,111,117,108,100,32,111,99,99,117,114,0,83,116,97,116,101,32,110,111,116,32,114,101,99,111,118,101,114,97,98,108,101,0,80,114,101,118,105,111,117,115,32,111,119,110,101,114,32,100,105,101,100,0,79,112,101,114,97,116,105,111,110,32,99,97,110,99,101,108,101,100,0,70,117,110,99,116,105,111,110,32,110,111,116,32,105,109,112,108,101,109,101,110,116,101,100,0,78,111,32,109,101,115,115,97,103,101,32,111,102,32,100,101,115,105,114,101,100,32,116,121,112,101,0,73,100,101,110,116,105,102,105,101,114,32,114,101,109,111,118,101,100,0,68,101,118,105,99,101,32,110,111,116,32,97,32,115,116,114,101,97,109,0,78,111,32,100,97,116,97,32,97,118,97,105,108,97,98,108,101,0,68,101,118,105,99,101,32,116,105,109,101,111,117,116,0,79,117,116,32,111,102,32,115,116,114,101,97,109,115,32,114,101,115,111,117,114,99,101,115,0,76,105,110,107,32,104,97,115,32,98,101,101,110,32,115,101,118,101,114,101,100,0,80,114,111,116,111,99,111,108,32,101,114,114,111,114,0,66,97,100,32,109,101,115,115,97,103,101,0,70,105,108,101,32,100,101,115,99,114,105,112,116,111,114,32,105,110,32,98,97,100,32,115,116,97,116,101,0,78,111,116,32,97,32,115,111,99,107,101,116,0,68,101,115,116,105,110,97,116,105,111,110,32,97,100,100,114,101,115,115,32,114,101,113,117,105,114,101,100,0,77,101,115,115,97,103,101,32,116,111,111,32,108,97,114,103,101,0,80,114,111,116,111,99,111,108,32,119,114,111,110,103,32,116,121,112,101,32,102,111,114,32,115,111,99,107,101,116,0,80,114,111,116,111,99,111,108,32,110,111,116,32,97,118,97,105,108,97,98,108,101,0,80,114,111,116,111,99,111,108,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,83,111,99,107,101,116,32,116,121,112,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,78,111,116,32,115,117,112,112,111,114,116,101,100,0,80,114,111,116,111,99,111,108,32,102,97,109,105,108,121,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,65,100,100,114,101,115,115,32,102,97,109,105,108,121,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,98,121,32,112,114,111,116,111,99,111,108,0,65,100,100,114,101,115,115,32,110,111,116,32,97,118,97,105,108,97,98,108,101,0,78,101,116,119,111,114,107,32,105,115,32,100,111,119,110,0,78,101,116,119,111,114,107,32,117,110,114,101,97,99,104,97,98,108,101,0,67,111,110,110,101,99,116,105,111,110,32,114,101,115,101,116,32,98,121,32,110,101,116,119,111,114,107,0,67,111,110,110,101,99,116,105,111,110,32,97,98,111,114,116,101,100,0,78,111,32,98,117,102,102,101,114,32,115,112,97,99,101,32,97,118,97,105,108,97,98,108,101,0,83,111,99,107,101,116,32,105,115,32,99,111,110,110,101,99,116,101,100,0,83,111,99,107,101,116,32,110,111,116,32,99,111,110,110,101,99,116,101,100,0,67,97,110,110,111,116,32,115,101,110,100,32,97,102,116,101,114,32,115,111,99,107,101,116,32,115,104,117,116,100,111,119,110,0,79,112,101,114,97,116,105,111,110,32,97,108,114,101,97,100,121,32,105,110,32,112,114,111,103,114,101,115,115,0,79,112,101,114,97,116,105,111,110,32,105,110,32,112,114,111,103,114,101,115,115,0,83,116,97,108,101,32,102,105,108,101,32,104,97,110,100,108,101,0,82,101,109,111,116,101,32,73,47,79,32,101,114,114,111,114,0,81,117,111,116,97,32,101,120,99,101,101,100,101,100,0,78,111,32,109,101,100,105,117,109,32,102,111,117,110,100,0,87,114,111,110,103,32,109,101,100,105,117,109,32,116,121,112,101,0,78,111,32,101,114,114,111,114,32,105,110,102,111,114,109,97,116,105,111,110,0,0,40,110,117,108,108,41,0,45,48,88,43,48,88,32,48,88,45,48,120,43,48,120,32,48,120,0,105,110,102,0,73,78,70,0,78,65,78,0,47,112,114,111,99,47,115,101,108,102,47,102,100,47,0,105,110,102,105,110,105,116,121,0,110,97,110,0,46,0,80,79,83,73,88,0,67,46,85,84,70,45,56,0,114,119,97,0,72,111,115,116,32,110,111,116,32,102,111,117,110,100,0,84,114,121,32,97,103,97,105,110,0,78,111,110,45,114,101,99,111,118,101,114,97,98,108,101,32,101,114,114,111,114,0,65,100,100,114,101,115,115,32,110,111,116,32,97,118,97,105,108,97,98,108,101,0,0,85,110,107,110,111,119,110,32,101,114,114,111,114,0,37,100,46,37,100,46,37,100,46,37,100,0,47,116,109,112,0,47,116,109,112,47,116,37,120,45,37,120,0,119,43,0,78,83,116,51,95,95,49,49,48,95,95,115,116,100,105,110,98,117,102,73,99,69,69,0,117,110,115,117,112,112,111,114,116,101,100,32,108,111,99,97,108,101,32,102,111,114,32,115,116,97,110,100,97,114,100,32,105,110,112,117,116,0,78,83,116,51,95,95,49,49,49,95,95,115,116,100,111,117,116,98,117,102,73,99,69,69,0,78,83,116,51,95,95,49,49,48,95,95,115,116,100,105,110,98,117,102,73,119,69,69,0,78,83,116,51,95,95,49,49,49,95,95,115,116,100,111,117,116,98,117,102,73,119,69,69,0,95,95,110,101,120,116,95,112,114,105,109,101,32,111,118,101,114,102,108,111,119,0,98,97,115,105,99,95,115,116,114,105,110,103,0,78,83,116,51,95,95,49,57,98,97,115,105,99,95,105,111,115,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,0,78,83,116,51,95,95,49,56,105,111,115,95,98,97,115,101,69,0,78,83,116,51,95,95,49,57,98,97,115,105,99,95,105,111,115,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,0,78,83,116,51,95,95,49,49,53,98,97,115,105,99,95,115,116,114,101,97,109,98,117,102,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,0,78,83,116,51,95,95,49,49,53,98,97,115,105,99,95,115,116,114,101,97,109,98,117,102,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,0,78,83,116,51,95,95,49,49,51,98,97,115,105,99,95,105,115,116,114,101,97,109,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,0,78,83,116,51,95,95,49,49,51,98,97,115,105,99,95,105,115,116,114,101,97,109,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,0,78,83,116,51,95,95,49,49,51,98,97,115,105,99,95,111,115,116,114,101,97,109,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,0,78,83,116,51,95,95,49,49,51,98,97,115,105,99,95,111,115,116,114,101,97,109,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,0,78,83,116,51,95,95,49,49,52,98,97,115,105,99,95,105,111,115,116,114,101,97,109,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,0,78,83,116,51,95,95,49,56,105,111,115,95,98,97,115,101,55,102,97,105,108,117,114,101,69,0,78,83,116,51,95,95,49,49,57,95,95,105,111,115,116,114,101,97,109,95,99,97,116,101,103,111,114,121,69,0,105,111,115,116,114,101,97,109,0,117,110,115,112,101,99,105,102,105,101,100,32,105,111,115,116,114,101,97,109,95,99,97,116,101,103,111,114,121,32,101,114,114,111,114,0,105,111,115,95,98,97,115,101,58,58,99,108,101,97,114,0,78,83,116,51,95,95,49,55,99,111,108,108,97,116,101,73,99,69,69,0,78,83,116,51,95,95,49,54,108,111,99,97,108,101,53,102,97,99,101,116,69,0,78,83,116,51,95,95,49,55,99,111,108,108,97,116,101,73,119,69,69,0,78,83,116,51,95,95,49,55,110,117,109,95,103,101,116,73,99,78,83,95,49,57,105,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,69,69,0,78,83,116,51,95,95,49,57,95,95,110,117,109,95,103,101,116,73,99,69,69,0,78,83,116,51,95,95,49,49,52,95,95,110,117,109,95,103,101,116,95,98,97,115,101,69,0,48,49,50,51,52,53,54,55,56,57,97,98,99,100,101,102,65,66,67,68,69,70,120,88,43,45,112,80,105,73,110,78,0,67,0,37,112,0,78,83,116,51,95,95,49,55,110,117,109,95,103,101,116,73,119,78,83,95,49,57,105,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,69,69,0,78,83,116,51,95,95,49,57,95,95,110,117,109,95,103,101,116,73,119,69,69,0,78,83,116,51,95,95,49,55,110,117,109,95,112,117,116,73,99,78,83,95,49,57,111,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,69,69,0,78,83,116,51,95,95,49,57,95,95,110,117,109,95,112,117,116,73,99,69,69,0,78,83,116,51,95,95,49,49,52,95,95,110,117,109,95,112,117,116,95,98,97,115,101,69,0,37,0,0,0,0,0,108,0,108,108,0,76,0,37,112,0,0,0,0,78,83,116,51,95,95,49,55,110,117,109,95,112,117,116,73,119,78,83,95,49,57,111,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,69,69,0,78,83,116,51,95,95,49,57,95,95,110,117,109,95,112,117,116,73,119,69,69,0,78,83,116,51,95,95,49,56,116,105,109,101,95,103,101,116,73,99,78,83,95,49,57,105,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,69,69,0,78,83,116,51,95,95,49,57,116,105,109,101,95,98,97,115,101,69,0,78,83,116,51,95,95,49,50,48,95,95,116,105,109,101,95,103,101,116,95,99,95,115,116,111,114,97,103,101,73,99,69,69,0,37,72,58,37,77,58,37,83,37,109,47,37,100,47,37,121,37,89,45,37,109,45,37,100,37,73,58,37,77,58,37,83,32,37,112,37,72,58,37,77,37,72,58,37,77,58,37,83,83,117,110,100,97,121,0,77,111,110,100,97,121,0,84,117,101,115,100,97,121,0,87,101,100,110,101,115,100,97,121,0,84,104,117,114,115,100,97,121,0,70,114,105,100,97,121,0,83,97,116,117,114,100,97,121,0,83,117,110,0,77,111,110,0,84,117,101,0,87,101,100,0,84,104,117,0,70,114,105,0,83,97,116,0,74,97,110,117,97,114,121,0,70,101,98,114,117,97,114,121,0,77,97,114,99,104,0,65,112,114,105,108,0,77,97,121,0,74,117,110,101,0,74,117,108,121,0,65,117,103,117,115,116,0,83,101,112,116,101,109,98,101,114,0,79,99,116,111,98,101,114,0,78,111,118,101,109,98,101,114,0,68,101,99,101,109,98,101,114,0,74,97,110,0,70,101,98,0,77,97,114,0,65,112,114,0,74,117,110,0,74,117,108,0,65,117,103,0,83,101,112,0,79,99,116,0,78,111,118,0,68,101,99,0,65,77,0,80,77,0,37,97,32,37,98,32,37,100,32,37,72,58,37,77,58,37,83,32,37,89,0,37,73,58,37,77,58,37,83,32,37,112,0,37,109,47,37,100,47,37,121,0,37,72,58,37,77,58,37,83,0,78,83,116,51,95,95,49,56,116,105,109,101,95,103,101,116,73,119,78,83,95,49,57,105,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,69,69,0,78,83,116,51,95,95,49,50,48,95,95,116,105,109,101,95,103,101,116,95,99,95,115,116,111,114,97,103,101,73,119,69,69,0,78,83,116,51,95,95,49,56,116,105,109,101,95,112,117,116,73,99,78,83,95,49,57,111,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,69,69,0,78,83,116,51,95,95,49,49,48,95,95,116,105,109,101,95,112,117,116,69,0,78,83,116,51,95,95,49,56,116,105,109,101,95,112,117,116,73,119,78,83,95,49,57,111,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,69,69,0,108,111,99,97,108,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,78,83,116,51,95,95,49,49,48,109,111,110,101,121,112,117,110,99,116,73,99,76,98,48,69,69,69,0,78,83,116,51,95,95,49,49,48,109,111,110,101,121,95,98,97,115,101,69,0,78,83,116,51,95,95,49,49,48,109,111,110,101,121,112,117,110,99,116,73,99,76,98,49,69,69,69,0,78,83,116,51,95,95,49,49,48,109,111,110,101,121,112,117,110,99,116,73,119,76,98,48,69,69,69,0,78,83,116,51,95,95,49,49,48,109,111,110,101,121,112,117,110,99,116,73,119,76,98,49,69,69,69,0,78,83,116,51,95,95,49,57,109,111,110,101,121,95,103,101,116,73,99,78,83,95,49,57,105,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,69,69,0,78,83,116,51,95,95,49,49,49,95,95,109,111,110,101,121,95,103,101,116,73,99,69,69,0,48,49,50,51,52,53,54,55,56,57,0,37,76,102,0,109,111,110,101,121,95,103,101,116,32,101,114,114,111,114,0,78,83,116,51,95,95,49,57,109,111,110,101,121,95,103,101,116,73,119,78,83,95,49,57,105,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,69,69,0,78,83,116,51,95,95,49,49,49,95,95,109,111,110,101,121,95,103,101,116,73,119,69,69,0,48,49,50,51,52,53,54,55,56,57,0,78,83,116,51,95,95,49,57,109,111,110,101,121,95,112,117,116,73,99,78,83,95,49,57,111,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,69,69,0,78,83,116,51,95,95,49,49,49,95,95,109,111,110,101,121,95,112,117,116,73,99,69,69,0,37,46,48,76,102,0,78,83,116,51,95,95,49,57,109,111,110,101,121,95,112,117,116,73,119,78,83,95,49,57,111,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,69,69,0,78,83,116,51,95,95,49,49,49,95,95,109,111,110,101,121,95,112,117,116,73,119,69,69,0,78,83,116,51,95,95,49,56,109,101,115,115,97,103,101,115,73,99,69,69,0,78,83,116,51,95,95,49,49,51,109,101,115,115,97,103,101,115,95,98,97,115,101,69,0,78,83,116,51,95,95,49,56,109,101,115,115,97,103,101,115,73,119,69,69,0,78,83,116,51,95,95,49,49,54,95,95,110,97,114,114,111,119,95,116,111,95,117,116,102,56,73,76,106,51,50,69,69,69,0,78,83,116,51,95,95,49,55,99,111,100,101,99,118,116,73,68,105,99,49,49,95,95,109,98,115,116,97,116,101,95,116,69,69,0,78,83,116,51,95,95,49,49,50,99,111,100,101,99,118,116,95,98,97,115,101,69,0,78,83,116,51,95,95,49,49,55,95,95,119,105,100,101,110,95,102,114,111,109,95,117,116,102,56,73,76,106,51,50,69,69,69,0,78,83,116,51,95,95,49,55,99,111,100,101,99,118,116,73,99,99,49,49,95,95,109,98,115,116,97,116,101,95,116,69,69,0,78,83,116,51,95,95,49,55,99,111,100,101,99,118,116,73,119,99,49,49,95,95,109,98,115,116,97,116,101,95,116,69,69,0,78,83,116,51,95,95,49,55,99,111,100,101,99,118,116,73,68,115,99,49,49,95,95,109,98,115,116,97,116,101,95,116,69,69,0,78,83,116,51,95,95,49,54,108,111,99,97,108,101,53,95,95,105,109,112,69,0,78,83,116,51,95,95,49,53,99,116,121,112,101,73,99,69,69,0,78,83,116,51,95,95,49,49,48,99,116,121,112,101,95,98,97,115,101,69,0,78,83,116,51,95,95,49,53,99,116,121,112,101,73,119,69,69,0,78,83,116,51,95,95,49,56,110,117,109,112,117,110,99,116,73,99,69,69,0,116,114,117,101,0,102,97,108,115,101,0,78,83,116,51,95,95,49,56,110,117,109,112,117,110,99,116,73,119,69,69,0,118,101,99,116,111,114,0,78,83,116,51,95,95,49,49,52,95,95,115,104,97,114,101,100,95,99,111,117,110,116,69,0,78,83,116,51,95,95,49,49,52,101,114,114,111,114,95,99,97,116,101,103,111,114,121,69,0,78,83,116,51,95,95,49,49,50,115,121,115,116,101,109,95,101,114,114,111,114,69,0,78,83,116,51,95,95,49,49,50,95,95,100,111,95,109,101,115,115,97,103,101,69,0,58,32,0,99,97,110,110,111,116,32,122,101,114,111,32,111,117,116,32,116,104,114,101,97,100,32,118,97,108,117,101,32,102,111,114,32,95,95,99,120,97,95,103,101,116,95,103,108,111,98,97,108,115,40,41,0,99,97,110,110,111,116,32,99,114,101,97,116,101,32,112,116,104,114,101,97,100,32,107,101,121,32,102,111,114,32,95,95,99,120,97,95,103,101,116,95,103,108,111,98,97,108,115,40,41,0,112,116,104,114,101,97,100,95,111,110,99,101,32,102,97,105,108,117,114,101,32,105,110,32,95,95,99,120,97,95,103,101,116,95,103,108,111,98,97,108,115,95,102,97,115,116,40,41,0,78,49,48,95,95,99,120,120,97,98,105,118,49,50,48,95,95,115,105,95,99,108,97,115,115,95,116,121,112,101,95,105,110,102,111,69,0,78,49,48,95,95,99,120,120,97,98,105,118,49,49,54,95,95,115,104,105,109,95,116,121,112,101,95,105,110,102,111,69,0,83,116,57,116,121,112,101,95,105,110,102,111,0,78,49,48,95,95,99,120,120,97,98,105,118,49,49,55,95,95,99,108,97,115,115,95,116,121,112,101,95,105,110,102,111,69,0,83,116,57,101,120,99,101,112,116,105,111,110,0,117,110,99,97,117,103,104,116,0,116,101,114,109,105,110,97,116,105,110,103,32,119,105,116,104,32,37,115,32,101,120,99,101,112,116,105,111,110,32,111,102,32,116,121,112,101,32,37,115,58,32,37,115,0,116,101,114,109,105,110,97,116,105,110,103,32,119,105,116,104,32,37,115,32,101,120,99,101,112,116,105,111,110,32,111,102,32,116,121,112,101,32,37,115,0,116,101,114,109,105,110,97,116,105,110,103,32,119,105,116,104,32,37,115,32,102,111,114,101,105,103,110,32,101,120,99,101,112,116,105,111,110,0,116,101,114,109,105,110,97,116,105,110,103,0,116,101,114,109,105,110,97,116,101,95,104,97,110,100,108,101,114,32,117,110,101,120,112,101,99,116,101,100,108,121,32,114,101,116,117,114,110,101,100,0,116,101,114,109,105,110,97,116,101,95,104,97,110,100,108,101,114,32,117,110,101,120,112,101,99,116,101,100,108,121,32,116,104,114,101,119,32,97,110,32,101,120,99,101,112,116,105,111,110,0,83,116,57,98,97,100,95,97,108,108,111,99,0,115,116,100,58,58,98,97,100,95,97,108,108,111,99,0,83,116,49,49,108,111,103,105,99,95,101,114,114,111,114,0,83,116,49,51,114,117,110,116,105,109,101,95,101,114,114,111,114,0,83,116,49,50,108,101,110,103,116,104,95,101,114,114,111,114,0,83,116,49,50,111,117,116,95,111,102,95,114,97,110,103,101,0,83,116,49,52,111,118,101,114,102,108,111,119,95,101,114,114,111,114,0,83,116,56,98,97,100,95,99,97,115,116,0,115,116,100,58,58,98,97,100,95,99,97,115,116,0,83,116,49,48,98,97,100,95,116,121,112,101,105,100,0,115,116,100,58,58,98,97,100,95,116,121,112,101,105,100,0,78,49,48,95,95,99,120,120,97,98,105,118,49,49,57,95,95,112,111,105,110,116,101,114,95,116,121,112,101,95,105,110,102,111,69,0,78,49,48,95,95,99,120,120,97,98,105,118,49,49,55,95,95,112,98,97,115,101,95,116,121,112,101,95,105,110,102,111,69,0,78,49,48,95,95,99,120,120,97,98,105,118,49,50,49,95,95,118,109,105,95,99,108,97,115,115,95,116,121,112,101,95,105,110,102,111,69,0], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE);





/* no memory initializer */
var tempDoublePtr = STATICTOP; STATICTOP += 16;

assert(tempDoublePtr % 8 == 0);

function copyTempFloat(ptr) { // functions, because inlining this code increases code size too much

  HEAP8[tempDoublePtr] = HEAP8[ptr];

  HEAP8[tempDoublePtr+1] = HEAP8[ptr+1];

  HEAP8[tempDoublePtr+2] = HEAP8[ptr+2];

  HEAP8[tempDoublePtr+3] = HEAP8[ptr+3];

}

function copyTempDouble(ptr) {

  HEAP8[tempDoublePtr] = HEAP8[ptr];

  HEAP8[tempDoublePtr+1] = HEAP8[ptr+1];

  HEAP8[tempDoublePtr+2] = HEAP8[ptr+2];

  HEAP8[tempDoublePtr+3] = HEAP8[ptr+3];

  HEAP8[tempDoublePtr+4] = HEAP8[ptr+4];

  HEAP8[tempDoublePtr+5] = HEAP8[ptr+5];

  HEAP8[tempDoublePtr+6] = HEAP8[ptr+6];

  HEAP8[tempDoublePtr+7] = HEAP8[ptr+7];

}

// {{PRE_LIBRARY}}


  
  var JSEvents={keyEvent:0,mouseEvent:0,wheelEvent:0,uiEvent:0,focusEvent:0,deviceOrientationEvent:0,deviceMotionEvent:0,fullscreenChangeEvent:0,pointerlockChangeEvent:0,visibilityChangeEvent:0,touchEvent:0,previousFullscreenElement:null,previousScreenX:null,previousScreenY:null,removeEventListenersRegistered:false,registerRemoveEventListeners:function () {
        if (!JSEvents.removeEventListenersRegistered) {
        __ATEXIT__.push(function() {
            for(var i = JSEvents.eventHandlers.length-1; i >= 0; --i) {
              JSEvents._removeHandler(i);
            }
           });
          JSEvents.removeEventListenersRegistered = true;
        }
      },findEventTarget:function (target) {
        if (target) {
          if (typeof target == "number") {
            target = Pointer_stringify(target);
          }
          if (target == '#window') return window;
          else if (target == '#document') return document;
          else if (target == '#screen') return window.screen;
          else if (target == '#canvas') return Module['canvas'];
  
          if (typeof target == 'string') return document.getElementById(target);
          else return target;
        } else {
          // The sensible target varies between events, but use window as the default
          // since DOM events mostly can default to that. Specific callback registrations
          // override their own defaults.
          return window;
        }
      },deferredCalls:[],deferCall:function (targetFunction, precedence, argsList) {
        function arraysHaveEqualContent(arrA, arrB) {
          if (arrA.length != arrB.length) return false;
  
          for(var i in arrA) {
            if (arrA[i] != arrB[i]) return false;
          }
          return true;
        }
        // Test if the given call was already queued, and if so, don't add it again.
        for(var i in JSEvents.deferredCalls) {
          var call = JSEvents.deferredCalls[i];
          if (call.targetFunction == targetFunction && arraysHaveEqualContent(call.argsList, argsList)) {
            return;
          }
        }
        JSEvents.deferredCalls.push({
          targetFunction: targetFunction,
          precedence: precedence,
          argsList: argsList
        });
  
        JSEvents.deferredCalls.sort(function(x,y) { return x.precedence < y.precedence; });
      },removeDeferredCalls:function (targetFunction) {
        for(var i = 0; i < JSEvents.deferredCalls.length; ++i) {
          if (JSEvents.deferredCalls[i].targetFunction == targetFunction) {
            JSEvents.deferredCalls.splice(i, 1);
            --i;
          }
        }
      },canPerformEventHandlerRequests:function () {
        return JSEvents.inEventHandler && JSEvents.currentEventHandler.allowsDeferredCalls;
      },runDeferredCalls:function () {
        if (!JSEvents.canPerformEventHandlerRequests()) {
          return;
        }
        for(var i = 0; i < JSEvents.deferredCalls.length; ++i) {
          var call = JSEvents.deferredCalls[i];
          JSEvents.deferredCalls.splice(i, 1);
          --i;
          call.targetFunction.apply(this, call.argsList);
        }
      },inEventHandler:0,currentEventHandler:null,eventHandlers:[],isInternetExplorer:function () { return navigator.userAgent.indexOf('MSIE') !== -1 || navigator.appVersion.indexOf('Trident/') > 0; },removeAllHandlersOnTarget:function (target, eventTypeString) {
        for(var i = 0; i < JSEvents.eventHandlers.length; ++i) {
          if (JSEvents.eventHandlers[i].target == target && 
            (!eventTypeString || eventTypeString == JSEvents.eventHandlers[i].eventTypeString)) {
             JSEvents._removeHandler(i--);
           }
        }
      },_removeHandler:function (i) {
        var h = JSEvents.eventHandlers[i];
        h.target.removeEventListener(h.eventTypeString, h.eventListenerFunc, h.useCapture);
        JSEvents.eventHandlers.splice(i, 1);
      },registerOrRemoveHandler:function (eventHandler) {
        var jsEventHandler = function jsEventHandler(event) {
          // Increment nesting count for the event handler.
          ++JSEvents.inEventHandler;
          JSEvents.currentEventHandler = eventHandler;
          // Process any old deferred calls the user has placed.
          JSEvents.runDeferredCalls();
          // Process the actual event, calls back to user C code handler.
          eventHandler.handlerFunc(event);
          // Process any new deferred calls that were placed right now from this event handler.
          JSEvents.runDeferredCalls();
          // Out of event handler - restore nesting count.
          --JSEvents.inEventHandler;
        }
        
        if (eventHandler.callbackfunc) {
          eventHandler.eventListenerFunc = jsEventHandler;
          eventHandler.target.addEventListener(eventHandler.eventTypeString, jsEventHandler, eventHandler.useCapture);
          JSEvents.eventHandlers.push(eventHandler);
          JSEvents.registerRemoveEventListeners();
        } else {
          for(var i = 0; i < JSEvents.eventHandlers.length; ++i) {
            if (JSEvents.eventHandlers[i].target == eventHandler.target
             && JSEvents.eventHandlers[i].eventTypeString == eventHandler.eventTypeString) {
               JSEvents._removeHandler(i--);
             }
          }
        }
      },registerKeyEventCallback:function (target, userData, useCapture, callbackfunc, eventTypeId, eventTypeString) {
        if (!JSEvents.keyEvent) {
          JSEvents.keyEvent = _malloc( 164 );
        }
        var handlerFunc = function(event) {
          var e = event || window.event;
          writeStringToMemory(e.key ? e.key : "", JSEvents.keyEvent + 0 );
          writeStringToMemory(e.code ? e.code : "", JSEvents.keyEvent + 32 );
          HEAP32[(((JSEvents.keyEvent)+(64))>>2)]=e.location;
          HEAP32[(((JSEvents.keyEvent)+(68))>>2)]=e.ctrlKey;
          HEAP32[(((JSEvents.keyEvent)+(72))>>2)]=e.shiftKey;
          HEAP32[(((JSEvents.keyEvent)+(76))>>2)]=e.altKey;
          HEAP32[(((JSEvents.keyEvent)+(80))>>2)]=e.metaKey;
          HEAP32[(((JSEvents.keyEvent)+(84))>>2)]=e.repeat;
          writeStringToMemory(e.locale ? e.locale : "", JSEvents.keyEvent + 88 );
          writeStringToMemory(e.char ? e.char : "", JSEvents.keyEvent + 120 );
          HEAP32[(((JSEvents.keyEvent)+(152))>>2)]=e.charCode;
          HEAP32[(((JSEvents.keyEvent)+(156))>>2)]=e.keyCode;
          HEAP32[(((JSEvents.keyEvent)+(160))>>2)]=e.which;
          var shouldCancel = Runtime.dynCall('iiii', callbackfunc, [eventTypeId, JSEvents.keyEvent, userData]);
          if (shouldCancel) {
            e.preventDefault();
          }
        };
  
        var eventHandler = {
          target: JSEvents.findEventTarget(target),
          allowsDeferredCalls: JSEvents.isInternetExplorer() ? false : true, // MSIE doesn't allow fullscreen and pointerlock requests from key handlers, others do.
          eventTypeString: eventTypeString,
          callbackfunc: callbackfunc,
          handlerFunc: handlerFunc,
          useCapture: useCapture
        };
        JSEvents.registerOrRemoveHandler(eventHandler);
      },getBoundingClientRectOrZeros:function (target) {
        return target.getBoundingClientRect ? target.getBoundingClientRect() : { left: 0, top: 0 };
      },fillMouseEventData:function (eventStruct, e, target) {
        HEAPF64[((eventStruct)>>3)]=JSEvents.tick();
        HEAP32[(((eventStruct)+(8))>>2)]=e.screenX;
        HEAP32[(((eventStruct)+(12))>>2)]=e.screenY;
        HEAP32[(((eventStruct)+(16))>>2)]=e.clientX;
        HEAP32[(((eventStruct)+(20))>>2)]=e.clientY;
        HEAP32[(((eventStruct)+(24))>>2)]=e.ctrlKey;
        HEAP32[(((eventStruct)+(28))>>2)]=e.shiftKey;
        HEAP32[(((eventStruct)+(32))>>2)]=e.altKey;
        HEAP32[(((eventStruct)+(36))>>2)]=e.metaKey;
        HEAP16[(((eventStruct)+(40))>>1)]=e.button;
        HEAP16[(((eventStruct)+(42))>>1)]=e.buttons;
        HEAP32[(((eventStruct)+(44))>>2)]=e["movementX"] || e["mozMovementX"] || e["webkitMovementX"] || (e.screenX-JSEvents.previousScreenX);
        HEAP32[(((eventStruct)+(48))>>2)]=e["movementY"] || e["mozMovementY"] || e["webkitMovementY"] || (e.screenY-JSEvents.previousScreenY);
  
        if (Module['canvas']) {
          var rect = Module['canvas'].getBoundingClientRect();
          HEAP32[(((eventStruct)+(60))>>2)]=e.clientX - rect.left;
          HEAP32[(((eventStruct)+(64))>>2)]=e.clientY - rect.top;
        } else { // Canvas is not initialized, return 0.
          HEAP32[(((eventStruct)+(60))>>2)]=0;
          HEAP32[(((eventStruct)+(64))>>2)]=0;
        }
        if (target) {
          var rect = JSEvents.getBoundingClientRectOrZeros(target);
          HEAP32[(((eventStruct)+(52))>>2)]=e.clientX - rect.left;
          HEAP32[(((eventStruct)+(56))>>2)]=e.clientY - rect.top;        
        } else { // No specific target passed, return 0.
          HEAP32[(((eventStruct)+(52))>>2)]=0;
          HEAP32[(((eventStruct)+(56))>>2)]=0;
        }
        JSEvents.previousScreenX = e.screenX;
        JSEvents.previousScreenY = e.screenY;
      },registerMouseEventCallback:function (target, userData, useCapture, callbackfunc, eventTypeId, eventTypeString) {
        if (!JSEvents.mouseEvent) {
          JSEvents.mouseEvent = _malloc( 72 );
        }
        target = JSEvents.findEventTarget(target);
        var handlerFunc = function(event) {
          var e = event || window.event;
          JSEvents.fillMouseEventData(JSEvents.mouseEvent, e, target);
          var shouldCancel = Runtime.dynCall('iiii', callbackfunc, [eventTypeId, JSEvents.mouseEvent, userData]);
          if (shouldCancel) {
            e.preventDefault();
          }
        };
  
        var eventHandler = {
          target: target,
          allowsDeferredCalls: eventTypeString != 'mousemove' && eventTypeString != 'mouseenter' && eventTypeString != 'mouseleave', // Mouse move events do not allow fullscreen/pointer lock requests to be handled in them!
          eventTypeString: eventTypeString,
          callbackfunc: callbackfunc,
          handlerFunc: handlerFunc,
          useCapture: useCapture
        };
        // In IE, mousedown events don't either allow deferred calls to be run!
        if (JSEvents.isInternetExplorer() && eventTypeString == 'mousedown') eventHandler.allowsDeferredCalls = false;
        JSEvents.registerOrRemoveHandler(eventHandler);
      },registerWheelEventCallback:function (target, userData, useCapture, callbackfunc, eventTypeId, eventTypeString) {
        if (!JSEvents.wheelEvent) {
          JSEvents.wheelEvent = _malloc( 104 );
        }
        target = JSEvents.findEventTarget(target);
        // The DOM Level 3 events spec event 'wheel'
        var wheelHandlerFunc = function(event) {
          var e = event || window.event;
          JSEvents.fillMouseEventData(JSEvents.wheelEvent, e, target);
          HEAPF64[(((JSEvents.wheelEvent)+(72))>>3)]=e["deltaX"];
          HEAPF64[(((JSEvents.wheelEvent)+(80))>>3)]=e["deltaY"];
          HEAPF64[(((JSEvents.wheelEvent)+(88))>>3)]=e["deltaZ"];
          HEAP32[(((JSEvents.wheelEvent)+(96))>>2)]=e["deltaMode"];
          var shouldCancel = Runtime.dynCall('iiii', callbackfunc, [eventTypeId, JSEvents.wheelEvent, userData]);
          if (shouldCancel) {
            e.preventDefault();
          }
        };
        // The 'mousewheel' event as implemented in Safari 6.0.5
        var mouseWheelHandlerFunc = function(event) {
          var e = event || window.event;
          JSEvents.fillMouseEventData(JSEvents.wheelEvent, e, target);
          HEAPF64[(((JSEvents.wheelEvent)+(72))>>3)]=e["wheelDeltaX"];
          HEAPF64[(((JSEvents.wheelEvent)+(80))>>3)]=-e["wheelDeltaY"] /* Invert to unify direction with the DOM Level 3 wheel event. */;
          HEAPF64[(((JSEvents.wheelEvent)+(88))>>3)]=0 /* Not available */;
          HEAP32[(((JSEvents.wheelEvent)+(96))>>2)]=0 /* DOM_DELTA_PIXEL */;
          var shouldCancel = Runtime.dynCall('iiii', callbackfunc, [eventTypeId, JSEvents.wheelEvent, userData]);
          if (shouldCancel) {
            e.preventDefault();
          }
        };
  
        var eventHandler = {
          target: target,
          allowsDeferredCalls: true,
          eventTypeString: eventTypeString,
          callbackfunc: callbackfunc,
          handlerFunc: (eventTypeString == 'wheel') ? wheelHandlerFunc : mouseWheelHandlerFunc,
          useCapture: useCapture
        };
        JSEvents.registerOrRemoveHandler(eventHandler);
      },pageScrollPos:function () {
        if (window.pageXOffset > 0 || window.pageYOffset > 0) {
          return [window.pageXOffset, window.pageYOffset];
        }
        if (typeof document.documentElement.scrollLeft !== 'undefined' || typeof document.documentElement.scrollTop !== 'undefined') {
          return [document.documentElement.scrollLeft, document.documentElement.scrollTop];
        }
        return [document.body.scrollLeft|0, document.body.scrollTop|0];
      },registerUiEventCallback:function (target, userData, useCapture, callbackfunc, eventTypeId, eventTypeString) {
        if (!JSEvents.uiEvent) {
          JSEvents.uiEvent = _malloc( 36 );
        }
  
        if (eventTypeString == "scroll" && !target) {
          target = document; // By default read scroll events on document rather than window.
        } else {
          target = JSEvents.findEventTarget(target);
        }
  
        var handlerFunc = function(event) {
          var e = event || window.event;
          if (e.target != target) {
            // Never take ui events such as scroll via a 'bubbled' route, but always from the direct element that
            // was targeted. Otherwise e.g. if app logs a message in response to a page scroll, the Emscripten log
            // message box could cause to scroll, generating a new (bubbled) scroll message, causing a new log print,
            // causing a new scroll, etc..
            return;
          }
          var scrollPos = JSEvents.pageScrollPos();
          HEAP32[((JSEvents.uiEvent)>>2)]=e.detail;
          HEAP32[(((JSEvents.uiEvent)+(4))>>2)]=document.body.clientWidth;
          HEAP32[(((JSEvents.uiEvent)+(8))>>2)]=document.body.clientHeight;
          HEAP32[(((JSEvents.uiEvent)+(12))>>2)]=window.innerWidth;
          HEAP32[(((JSEvents.uiEvent)+(16))>>2)]=window.innerHeight;
          HEAP32[(((JSEvents.uiEvent)+(20))>>2)]=window.outerWidth;
          HEAP32[(((JSEvents.uiEvent)+(24))>>2)]=window.outerHeight;
          HEAP32[(((JSEvents.uiEvent)+(28))>>2)]=scrollPos[0];
          HEAP32[(((JSEvents.uiEvent)+(32))>>2)]=scrollPos[1];
          var shouldCancel = Runtime.dynCall('iiii', callbackfunc, [eventTypeId, JSEvents.uiEvent, userData]);
          if (shouldCancel) {
            e.preventDefault();
          }
        };
  
        var eventHandler = {
          target: target,
          allowsDeferredCalls: false, // Neither scroll or resize events allow running requests inside them.
          eventTypeString: eventTypeString,
          callbackfunc: callbackfunc,
          handlerFunc: handlerFunc,
          useCapture: useCapture
        };
        JSEvents.registerOrRemoveHandler(eventHandler);
      },getNodeNameForTarget:function (target) {
        if (!target) return '';
        if (target == window) return '#window';
        if (target == window.screen) return '#screen';
        return (target && target.nodeName) ? target.nodeName : '';
      },registerFocusEventCallback:function (target, userData, useCapture, callbackfunc, eventTypeId, eventTypeString) {
        if (!JSEvents.focusEvent) {
          JSEvents.focusEvent = _malloc( 256 );
        }
        var handlerFunc = function(event) {
          var e = event || window.event;
  
          var nodeName = JSEvents.getNodeNameForTarget(e.target);
          var id = e.target.id ? e.target.id : '';
          writeStringToMemory(nodeName, JSEvents.focusEvent + 0 );
          writeStringToMemory(id, JSEvents.focusEvent + 128 );
          var shouldCancel = Runtime.dynCall('iiii', callbackfunc, [eventTypeId, JSEvents.focusEvent, userData]);
          if (shouldCancel) {
            e.preventDefault();
          }
        };
  
        var eventHandler = {
          target: JSEvents.findEventTarget(target),
          allowsDeferredCalls: false,
          eventTypeString: eventTypeString,
          callbackfunc: callbackfunc,
          handlerFunc: handlerFunc,
          useCapture: useCapture
        };
        JSEvents.registerOrRemoveHandler(eventHandler);
      },tick:function () {
        if (window['performance'] && window['performance']['now']) return window['performance']['now']();
        else return Date.now();
      },registerDeviceOrientationEventCallback:function (target, userData, useCapture, callbackfunc, eventTypeId, eventTypeString) {
        if (!JSEvents.deviceOrientationEvent) {
          JSEvents.deviceOrientationEvent = _malloc( 40 );
        }
        var handlerFunc = function(event) {
          var e = event || window.event;
  
          HEAPF64[((JSEvents.deviceOrientationEvent)>>3)]=JSEvents.tick();
          HEAPF64[(((JSEvents.deviceOrientationEvent)+(8))>>3)]=e.alpha;
          HEAPF64[(((JSEvents.deviceOrientationEvent)+(16))>>3)]=e.beta;
          HEAPF64[(((JSEvents.deviceOrientationEvent)+(24))>>3)]=e.gamma;
          HEAP32[(((JSEvents.deviceOrientationEvent)+(32))>>2)]=e.absolute;
  
          var shouldCancel = Runtime.dynCall('iiii', callbackfunc, [eventTypeId, JSEvents.deviceOrientationEvent, userData]);
          if (shouldCancel) {
            e.preventDefault();
          }
        };
  
        var eventHandler = {
          target: JSEvents.findEventTarget(target),
          allowsDeferredCalls: false,
          eventTypeString: eventTypeString,
          callbackfunc: callbackfunc,
          handlerFunc: handlerFunc,
          useCapture: useCapture
        };
        JSEvents.registerOrRemoveHandler(eventHandler);
      },registerDeviceMotionEventCallback:function (target, userData, useCapture, callbackfunc, eventTypeId, eventTypeString) {
        if (!JSEvents.deviceMotionEvent) {
          JSEvents.deviceMotionEvent = _malloc( 80 );
        }
        var handlerFunc = function(event) {
          var e = event || window.event;
  
          HEAPF64[((JSEvents.deviceOrientationEvent)>>3)]=JSEvents.tick();
          HEAPF64[(((JSEvents.deviceMotionEvent)+(8))>>3)]=e.acceleration.x;
          HEAPF64[(((JSEvents.deviceMotionEvent)+(16))>>3)]=e.acceleration.y;
          HEAPF64[(((JSEvents.deviceMotionEvent)+(24))>>3)]=e.acceleration.z;
          HEAPF64[(((JSEvents.deviceMotionEvent)+(32))>>3)]=e.accelerationIncludingGravity.x;
          HEAPF64[(((JSEvents.deviceMotionEvent)+(40))>>3)]=e.accelerationIncludingGravity.y;
          HEAPF64[(((JSEvents.deviceMotionEvent)+(48))>>3)]=e.accelerationIncludingGravity.z;
          HEAPF64[(((JSEvents.deviceMotionEvent)+(56))>>3)]=e.rotationRate.alpha;
          HEAPF64[(((JSEvents.deviceMotionEvent)+(64))>>3)]=e.rotationRate.beta;
          HEAPF64[(((JSEvents.deviceMotionEvent)+(72))>>3)]=e.rotationRate.gamma;
  
          var shouldCancel = Runtime.dynCall('iiii', callbackfunc, [eventTypeId, JSEvents.deviceMotionEvent, userData]);
          if (shouldCancel) {
            e.preventDefault();
          }
        };
  
        var eventHandler = {
          target: JSEvents.findEventTarget(target),
          allowsDeferredCalls: false,
          eventTypeString: eventTypeString,
          callbackfunc: callbackfunc,
          handlerFunc: handlerFunc,
          useCapture: useCapture
        };
        JSEvents.registerOrRemoveHandler(eventHandler);
      },screenOrientation:function () {
        if (!window.screen) return undefined;
        return window.screen.orientation || window.screen.mozOrientation || window.screen.webkitOrientation || window.screen.msOrientation;
      },fillOrientationChangeEventData:function (eventStruct, e) {
        var orientations  = ["portrait-primary", "portrait-secondary", "landscape-primary", "landscape-secondary"];
        var orientations2 = ["portrait",         "portrait",           "landscape",         "landscape"];
  
        var orientationString = JSEvents.screenOrientation();
        var orientation = orientations.indexOf(orientationString);
        if (orientation == -1) {
          orientation = orientations2.indexOf(orientationString);
        }
  
        HEAP32[((eventStruct)>>2)]=1 << orientation;
        HEAP32[(((eventStruct)+(4))>>2)]=window.orientation;
      },registerOrientationChangeEventCallback:function (target, userData, useCapture, callbackfunc, eventTypeId, eventTypeString) {
        if (!JSEvents.orientationChangeEvent) {
          JSEvents.orientationChangeEvent = _malloc( 8 );
        }
  
        if (!target) {
          target = window.screen; // Orientation events need to be captured from 'window.screen' instead of 'window'
        } else {
          target = JSEvents.findEventTarget(target);
        }
  
        var handlerFunc = function(event) {
          var e = event || window.event;
  
          JSEvents.fillOrientationChangeEventData(JSEvents.orientationChangeEvent, e);
  
          var shouldCancel = Runtime.dynCall('iiii', callbackfunc, [eventTypeId, JSEvents.orientationChangeEvent, userData]);
          if (shouldCancel) {
            e.preventDefault();
          }
        };
  
        if (eventTypeString == "orientationchange" && window.screen.mozOrientation !== undefined) {
          eventTypeString = "mozorientationchange";
        }
  
        var eventHandler = {
          target: target,
          allowsDeferredCalls: false,
          eventTypeString: eventTypeString,
          callbackfunc: callbackfunc,
          handlerFunc: handlerFunc,
          useCapture: useCapture
        };
        JSEvents.registerOrRemoveHandler(eventHandler);
      },fullscreenEnabled:function () {
        return document.fullscreenEnabled || document.mozFullscreenEnabled || document.mozFullScreenEnabled || document.webkitFullscreenEnabled || document.msFullscreenEnabled;
      },fillFullscreenChangeEventData:function (eventStruct, e) {
        var fullscreenElement = document.fullscreenElement || document.mozFullScreenElement || document.webkitFullscreenElement || document.msFullscreenElement;
        var isFullscreen = !!fullscreenElement;
        HEAP32[((eventStruct)>>2)]=isFullscreen;
        HEAP32[(((eventStruct)+(4))>>2)]=JSEvents.fullscreenEnabled();
        // If transitioning to fullscreen, report info about the element that is now fullscreen.
        // If transitioning to windowed mode, report info about the element that just was fullscreen.
        var reportedElement = isFullscreen ? fullscreenElement : JSEvents.previousFullscreenElement;
        var nodeName = JSEvents.getNodeNameForTarget(reportedElement);
        var id = (reportedElement && reportedElement.id) ? reportedElement.id : '';
        writeStringToMemory(nodeName, eventStruct + 8 );
        writeStringToMemory(id, eventStruct + 136 );
        HEAP32[(((eventStruct)+(264))>>2)]=reportedElement ? reportedElement.clientWidth : 0;
        HEAP32[(((eventStruct)+(268))>>2)]=reportedElement ? reportedElement.clientHeight : 0;
        HEAP32[(((eventStruct)+(272))>>2)]=screen.width;
        HEAP32[(((eventStruct)+(276))>>2)]=screen.height;
        if (isFullscreen) {
          JSEvents.previousFullscreenElement = fullscreenElement;
        }
      },registerFullscreenChangeEventCallback:function (target, userData, useCapture, callbackfunc, eventTypeId, eventTypeString) {
        if (!JSEvents.fullscreenChangeEvent) {
          JSEvents.fullscreenChangeEvent = _malloc( 280 );
        }
  
        if (!target) {
          target = document; // Fullscreen change events need to be captured from 'document' by default instead of 'window'
        } else {
          target = JSEvents.findEventTarget(target);
        }
  
        var handlerFunc = function(event) {
          var e = event || window.event;
  
          JSEvents.fillFullscreenChangeEventData(JSEvents.fullscreenChangeEvent, e);
  
          var shouldCancel = Runtime.dynCall('iiii', callbackfunc, [eventTypeId, JSEvents.fullscreenChangeEvent, userData]);
          if (shouldCancel) {
            e.preventDefault();
          }
        };
  
        var eventHandler = {
          target: target,
          allowsDeferredCalls: false,
          eventTypeString: eventTypeString,
          callbackfunc: callbackfunc,
          handlerFunc: handlerFunc,
          useCapture: useCapture
        };
        JSEvents.registerOrRemoveHandler(eventHandler);
      },resizeCanvasForFullscreen:function (target, strategy) {
        var restoreOldStyle = __registerRestoreOldStyle(target);
        var cssWidth = strategy.softFullscreen ? window.innerWidth : screen.width;
        var cssHeight = strategy.softFullscreen ? window.innerHeight : screen.height;
        var rect = target.getBoundingClientRect();
        var windowedCssWidth = rect.right - rect.left;
        var windowedCssHeight = rect.bottom - rect.top;
        var windowedRttWidth = target.width;
        var windowedRttHeight = target.height;
  
        if (strategy.scaleMode == 3) {
          __setLetterbox(target, (cssHeight - windowedCssHeight) / 2, (cssWidth - windowedCssWidth) / 2);
          cssWidth = windowedCssWidth;
          cssHeight = windowedCssHeight;
        } else if (strategy.scaleMode == 2) {
          if (cssWidth*windowedRttHeight < windowedRttWidth*cssHeight) {
            var desiredCssHeight = windowedRttHeight * cssWidth / windowedRttWidth;
            __setLetterbox(target, (cssHeight - desiredCssHeight) / 2, 0);
            cssHeight = desiredCssHeight;
          } else {
            var desiredCssWidth = windowedRttWidth * cssHeight / windowedRttHeight;
            __setLetterbox(target, 0, (cssWidth - desiredCssWidth) / 2);
            cssWidth = desiredCssWidth;
          }
        }
  
        // If we are adding padding, must choose a background color or otherwise Chrome will give the
        // padding a default white color. Do it only if user has not customized their own background color.
        if (!target.style.backgroundColor) target.style.backgroundColor = 'black';
        // IE11 does the same, but requires the color to be set in the document body.
        if (!document.body.style.backgroundColor) document.body.style.backgroundColor = 'black'; // IE11
        // Firefox always shows black letterboxes independent of style color.
  
        target.style.width = cssWidth + 'px';
        target.style.height = cssHeight + 'px';
  
        if (strategy.filteringMode == 1) {
          target.style.imageRendering = 'optimizeSpeed';
          target.style.imageRendering = '-moz-crisp-edges';
          target.style.imageRendering = '-o-crisp-edges';
          target.style.imageRendering = '-webkit-optimize-contrast';
          target.style.imageRendering = 'optimize-contrast';
          target.style.imageRendering = 'crisp-edges';
          target.style.imageRendering = 'pixelated';
        }
  
        var dpiScale = (strategy.canvasResolutionScaleMode == 2) ? window.devicePixelRatio : 1;
        if (strategy.canvasResolutionScaleMode != 0) {
          target.width = cssWidth * dpiScale;
          target.height = cssHeight * dpiScale;
          if (target.GLctxObject) target.GLctxObject.GLctx.viewport(0, 0, target.width, target.height);
        }
        return restoreOldStyle;
      },requestFullscreen:function (target, strategy) {
        // EMSCRIPTEN_FULLSCREEN_SCALE_DEFAULT + EMSCRIPTEN_FULLSCREEN_CANVAS_SCALE_NONE is a mode where no extra logic is performed to the DOM elements.
        if (strategy.scaleMode != 0 || strategy.canvasResolutionScaleMode != 0) {
          JSEvents.resizeCanvasForFullscreen(target, strategy);
        }
  
        if (target.requestFullscreen) {
          target.requestFullscreen();
        } else if (target.msRequestFullscreen) {
          target.msRequestFullscreen();
        } else if (target.mozRequestFullScreen) {
          target.mozRequestFullScreen();
        } else if (target.mozRequestFullscreen) {
          target.mozRequestFullscreen();
        } else if (target.webkitRequestFullscreen) {
          target.webkitRequestFullscreen(Element.ALLOW_KEYBOARD_INPUT);
        } else {
          if (typeof JSEvents.fullscreenEnabled() === 'undefined') {
            return -1;
          } else {
            return -3;
          }
        }
  
        if (strategy.canvasResizedCallback) {
          Runtime.dynCall('iiii', strategy.canvasResizedCallback, [37, 0, strategy.canvasResizedCallbackUserData]);
        }
  
        return 0;
      },fillPointerlockChangeEventData:function (eventStruct, e) {
        var pointerLockElement = document.pointerLockElement || document.mozPointerLockElement || document.webkitPointerLockElement || document.msPointerLockElement;
        var isPointerlocked = !!pointerLockElement;
        HEAP32[((eventStruct)>>2)]=isPointerlocked;
        var nodeName = JSEvents.getNodeNameForTarget(pointerLockElement);
        var id = (pointerLockElement && pointerLockElement.id) ? pointerLockElement.id : '';
        writeStringToMemory(nodeName, eventStruct + 4 );
        writeStringToMemory(id, eventStruct + 132);
      },registerPointerlockChangeEventCallback:function (target, userData, useCapture, callbackfunc, eventTypeId, eventTypeString) {
        if (!JSEvents.pointerlockChangeEvent) {
          JSEvents.pointerlockChangeEvent = _malloc( 260 );
        }
  
        if (!target) {
          target = document; // Pointer lock change events need to be captured from 'document' by default instead of 'window'
        } else {
          target = JSEvents.findEventTarget(target);
        }
  
        var handlerFunc = function(event) {
          var e = event || window.event;
  
          JSEvents.fillPointerlockChangeEventData(JSEvents.pointerlockChangeEvent, e);
  
          var shouldCancel = Runtime.dynCall('iiii', callbackfunc, [eventTypeId, JSEvents.pointerlockChangeEvent, userData]);
          if (shouldCancel) {
            e.preventDefault();
          }
        };
  
        var eventHandler = {
          target: target,
          allowsDeferredCalls: false,
          eventTypeString: eventTypeString,
          callbackfunc: callbackfunc,
          handlerFunc: handlerFunc,
          useCapture: useCapture
        };
        JSEvents.registerOrRemoveHandler(eventHandler);
      },requestPointerLock:function (target) {
        if (target.requestPointerLock) {
          target.requestPointerLock();
        } else if (target.mozRequestPointerLock) {
          target.mozRequestPointerLock();
        } else if (target.webkitRequestPointerLock) {
          target.webkitRequestPointerLock();
        } else if (target.msRequestPointerLock) {
          target.msRequestPointerLock();
        } else {
          // document.body is known to accept pointer lock, so use that to differentiate if the user passed a bad element,
          // or if the whole browser just doesn't support the feature.
          if (document.body.requestPointerLock || document.body.mozRequestPointerLock || document.body.webkitRequestPointerLock || document.body.msRequestPointerLock) {
            return -3;
          } else {
            return -1;
          }
        }
        return 0;
      },fillVisibilityChangeEventData:function (eventStruct, e) {
        var visibilityStates = [ "hidden", "visible", "prerender", "unloaded" ];
        var visibilityState = visibilityStates.indexOf(document.visibilityState);
  
        HEAP32[((eventStruct)>>2)]=document.hidden;
        HEAP32[(((eventStruct)+(4))>>2)]=visibilityState;
      },registerVisibilityChangeEventCallback:function (target, userData, useCapture, callbackfunc, eventTypeId, eventTypeString) {
        if (!JSEvents.visibilityChangeEvent) {
          JSEvents.visibilityChangeEvent = _malloc( 8 );
        }
  
        if (!target) {
          target = document; // Visibility change events need to be captured from 'document' by default instead of 'window'
        } else {
          target = JSEvents.findEventTarget(target);
        }
  
        var handlerFunc = function(event) {
          var e = event || window.event;
  
          JSEvents.fillVisibilityChangeEventData(JSEvents.visibilityChangeEvent, e);
  
          var shouldCancel = Runtime.dynCall('iiii', callbackfunc, [eventTypeId, JSEvents.visibilityChangeEvent, userData]);
          if (shouldCancel) {
            e.preventDefault();
          }
        };
  
        var eventHandler = {
          target: target,
          allowsDeferredCalls: false,
          eventTypeString: eventTypeString,
          callbackfunc: callbackfunc,
          handlerFunc: handlerFunc,
          useCapture: useCapture
        };
        JSEvents.registerOrRemoveHandler(eventHandler);
      },registerTouchEventCallback:function (target, userData, useCapture, callbackfunc, eventTypeId, eventTypeString) {
        if (!JSEvents.touchEvent) {
          JSEvents.touchEvent = _malloc( 1684 );
        }
  
        target = JSEvents.findEventTarget(target);
  
        var handlerFunc = function(event) {
          var e = event || window.event;
  
          var touches = {};
          for(var i = 0; i < e.touches.length; ++i) {
            var touch = e.touches[i];
            touches[touch.identifier] = touch;
          }
          for(var i = 0; i < e.changedTouches.length; ++i) {
            var touch = e.changedTouches[i];
            touches[touch.identifier] = touch;
            touch.changed = true;
          }
          for(var i = 0; i < e.targetTouches.length; ++i) {
            var touch = e.targetTouches[i];
            touches[touch.identifier].onTarget = true;
          }
          
          var ptr = JSEvents.touchEvent;
          HEAP32[(((ptr)+(4))>>2)]=e.ctrlKey;
          HEAP32[(((ptr)+(8))>>2)]=e.shiftKey;
          HEAP32[(((ptr)+(12))>>2)]=e.altKey;
          HEAP32[(((ptr)+(16))>>2)]=e.metaKey;
          ptr += 20; // Advance to the start of the touch array.
          var canvasRect = Module['canvas'] ? Module['canvas'].getBoundingClientRect() : undefined;
          var targetRect = JSEvents.getBoundingClientRectOrZeros(target);
          var numTouches = 0;
          for(var i in touches) {
            var t = touches[i];
            HEAP32[((ptr)>>2)]=t.identifier;
            HEAP32[(((ptr)+(4))>>2)]=t.screenX;
            HEAP32[(((ptr)+(8))>>2)]=t.screenY;
            HEAP32[(((ptr)+(12))>>2)]=t.clientX;
            HEAP32[(((ptr)+(16))>>2)]=t.clientY;
            HEAP32[(((ptr)+(20))>>2)]=t.pageX;
            HEAP32[(((ptr)+(24))>>2)]=t.pageY;
            HEAP32[(((ptr)+(28))>>2)]=t.changed;
            HEAP32[(((ptr)+(32))>>2)]=t.onTarget;
            if (canvasRect) {
              HEAP32[(((ptr)+(44))>>2)]=t.clientX - canvasRect.left;
              HEAP32[(((ptr)+(48))>>2)]=t.clientY - canvasRect.top;
            } else {
              HEAP32[(((ptr)+(44))>>2)]=0;
              HEAP32[(((ptr)+(48))>>2)]=0;            
            }
            HEAP32[(((ptr)+(36))>>2)]=t.clientX - targetRect.left;
            HEAP32[(((ptr)+(40))>>2)]=t.clientY - targetRect.top;
            
            ptr += 52;
  
            if (++numTouches >= 32) {
              break;
            }
          }
          HEAP32[((JSEvents.touchEvent)>>2)]=numTouches;
  
          var shouldCancel = Runtime.dynCall('iiii', callbackfunc, [eventTypeId, JSEvents.touchEvent, userData]);
          if (shouldCancel) {
            e.preventDefault();
          }
        };
  
        var eventHandler = {
          target: target,
          allowsDeferredCalls: false, // XXX Currently disabled, see bug https://bugzilla.mozilla.org/show_bug.cgi?id=966493
          // Once the above bug is resolved, enable the following condition if possible:
          // allowsDeferredCalls: eventTypeString == 'touchstart',
          eventTypeString: eventTypeString,
          callbackfunc: callbackfunc,
          handlerFunc: handlerFunc,
          useCapture: useCapture
        };
        JSEvents.registerOrRemoveHandler(eventHandler);
      },fillGamepadEventData:function (eventStruct, e) {
        HEAPF64[((eventStruct)>>3)]=e.timestamp;
        for(var i = 0; i < e.axes.length; ++i) {
          HEAPF64[(((eventStruct+i*8)+(16))>>3)]=e.axes[i];
        }
        for(var i = 0; i < e.buttons.length; ++i) {
          if (typeof(e.buttons[i]) === 'object') {
            HEAPF64[(((eventStruct+i*8)+(528))>>3)]=e.buttons[i].value;
          } else {
            HEAPF64[(((eventStruct+i*8)+(528))>>3)]=e.buttons[i];
          }
        }
        for(var i = 0; i < e.buttons.length; ++i) {
          if (typeof(e.buttons[i]) === 'object') {
            HEAP32[(((eventStruct+i*4)+(1040))>>2)]=e.buttons[i].pressed;
          } else {
            HEAP32[(((eventStruct+i*4)+(1040))>>2)]=e.buttons[i] == 1.0;
          }
        }
        HEAP32[(((eventStruct)+(1296))>>2)]=e.connected;
        HEAP32[(((eventStruct)+(1300))>>2)]=e.index;
        HEAP32[(((eventStruct)+(8))>>2)]=e.axes.length;
        HEAP32[(((eventStruct)+(12))>>2)]=e.buttons.length;
        writeStringToMemory(e.id, eventStruct + 1304 );
        writeStringToMemory(e.mapping, eventStruct + 1368 );
      },registerGamepadEventCallback:function (target, userData, useCapture, callbackfunc, eventTypeId, eventTypeString) {
        if (!JSEvents.gamepadEvent) {
          JSEvents.gamepadEvent = _malloc( 1432 );
        }
  
        var handlerFunc = function(event) {
          var e = event || window.event;
  
          JSEvents.fillGamepadEventData(JSEvents.gamepadEvent, e.gamepad);
  
          var shouldCancel = Runtime.dynCall('iiii', callbackfunc, [eventTypeId, JSEvents.gamepadEvent, userData]);
          if (shouldCancel) {
            e.preventDefault();
          }
        };
  
        var eventHandler = {
          target: JSEvents.findEventTarget(target),
          allowsDeferredCalls: true,
          eventTypeString: eventTypeString,
          callbackfunc: callbackfunc,
          handlerFunc: handlerFunc,
          useCapture: useCapture
        };
        JSEvents.registerOrRemoveHandler(eventHandler);
      },registerBeforeUnloadEventCallback:function (target, userData, useCapture, callbackfunc, eventTypeId, eventTypeString) {
        var handlerFunc = function(event) {
          var e = event || window.event;
  
          var confirmationMessage = Runtime.dynCall('iiii', callbackfunc, [eventTypeId, 0, userData]);
          
          if (confirmationMessage) {
            confirmationMessage = Pointer_stringify(confirmationMessage);
          }
          if (confirmationMessage) {
            e.preventDefault();
            e.returnValue = confirmationMessage;
            return confirmationMessage;
          }
        };
  
        var eventHandler = {
          target: JSEvents.findEventTarget(target),
          allowsDeferredCalls: false,
          eventTypeString: eventTypeString,
          callbackfunc: callbackfunc,
          handlerFunc: handlerFunc,
          useCapture: useCapture
        };
        JSEvents.registerOrRemoveHandler(eventHandler);
      },battery:function () { return navigator.battery || navigator.mozBattery || navigator.webkitBattery; },fillBatteryEventData:function (eventStruct, e) {
        HEAPF64[((eventStruct)>>3)]=e.chargingTime;
        HEAPF64[(((eventStruct)+(8))>>3)]=e.dischargingTime;
        HEAPF64[(((eventStruct)+(16))>>3)]=e.level;
        HEAP32[(((eventStruct)+(24))>>2)]=e.charging;
      },registerBatteryEventCallback:function (target, userData, useCapture, callbackfunc, eventTypeId, eventTypeString) {
        if (!JSEvents.batteryEvent) {
          JSEvents.batteryEvent = _malloc( 32 );
        }
  
        var handlerFunc = function(event) {
          var e = event || window.event;
  
          JSEvents.fillBatteryEventData(JSEvents.batteryEvent, JSEvents.battery());
  
          var shouldCancel = Runtime.dynCall('iiii', callbackfunc, [eventTypeId, JSEvents.batteryEvent, userData]);
          if (shouldCancel) {
            e.preventDefault();
          }
        };
  
        var eventHandler = {
          target: JSEvents.findEventTarget(target),
          allowsDeferredCalls: false,
          eventTypeString: eventTypeString,
          callbackfunc: callbackfunc,
          handlerFunc: handlerFunc,
          useCapture: useCapture
        };
        JSEvents.registerOrRemoveHandler(eventHandler);
      },registerWebGlEventCallback:function (target, userData, useCapture, callbackfunc, eventTypeId, eventTypeString) {
        if (!target) {
          target = Module['canvas'];
        }
        var handlerFunc = function(event) {
          var e = event || window.event;
  
          var shouldCancel = Runtime.dynCall('iiii', callbackfunc, [eventTypeId, 0, userData]);
          if (shouldCancel) {
            e.preventDefault();
          }
        };
  
        var eventHandler = {
          target: JSEvents.findEventTarget(target),
          allowsDeferredCalls: false,
          eventTypeString: eventTypeString,
          callbackfunc: callbackfunc,
          handlerFunc: handlerFunc,
          useCapture: useCapture
        };
        JSEvents.registerOrRemoveHandler(eventHandler);
      }};function _emscripten_set_visibilitychange_callback(userData, useCapture, callbackfunc) {
      JSEvents.registerVisibilityChangeEventCallback(document, userData, useCapture, callbackfunc, 21, "visibilitychange");
      return 0;
    }

  
  var GL={counter:1,lastError:0,buffers:[],mappedBuffers:{},programs:[],framebuffers:[],renderbuffers:[],textures:[],uniforms:[],shaders:[],vaos:[],contexts:[],currentContext:null,currArrayBuffer:0,currElementArrayBuffer:0,byteSizeByTypeRoot:5120,byteSizeByType:[1,1,2,2,4,4,4,2,3,4,8],programInfos:{},stringCache:{},packAlignment:4,unpackAlignment:4,init:function () {
        GL.createLog2ceilLookup(GL.MAX_TEMP_BUFFER_SIZE);
        GL.miniTempBuffer = new Float32Array(GL.MINI_TEMP_BUFFER_SIZE);
        for (var i = 0; i < GL.MINI_TEMP_BUFFER_SIZE; i++) {
          GL.miniTempBufferViews[i] = GL.miniTempBuffer.subarray(0, i+1);
        }
      },recordError:function recordError(errorCode) {
        if (!GL.lastError) {
          GL.lastError = errorCode;
        }
      },getNewId:function (table) {
        var ret = GL.counter++;
        for (var i = table.length; i < ret; i++) {
          table[i] = null;
        }
        return ret;
      },MINI_TEMP_BUFFER_SIZE:16,miniTempBuffer:null,miniTempBufferViews:[0],MAX_TEMP_BUFFER_SIZE:2097152,numTempVertexBuffersPerSize:64,log2ceilLookup:null,createLog2ceilLookup:function (maxValue) {
        GL.log2ceilLookup = new Uint8Array(maxValue+1);
        var log2 = 0;
        var pow2 = 1;
        GL.log2ceilLookup[0] = 0;
        for(var i = 1; i <= maxValue; ++i) {
          if (i > pow2) {
            pow2 <<= 1;
            ++log2;
          }
          GL.log2ceilLookup[i] = log2;
        }
      },generateTempBuffers:function (quads, context) {
        var largestIndex = GL.log2ceilLookup[GL.MAX_TEMP_BUFFER_SIZE];
        context.tempVertexBufferCounters1 = [];
        context.tempVertexBufferCounters2 = [];
        context.tempVertexBufferCounters1.length = context.tempVertexBufferCounters2.length = largestIndex+1;
        context.tempVertexBuffers1 = [];
        context.tempVertexBuffers2 = [];
        context.tempVertexBuffers1.length = context.tempVertexBuffers2.length = largestIndex+1;
        context.tempIndexBuffers = [];
        context.tempIndexBuffers.length = largestIndex+1;
        for(var i = 0; i <= largestIndex; ++i) {
          context.tempIndexBuffers[i] = null; // Created on-demand
          context.tempVertexBufferCounters1[i] = context.tempVertexBufferCounters2[i] = 0;
          var ringbufferLength = GL.numTempVertexBuffersPerSize;
          context.tempVertexBuffers1[i] = [];
          context.tempVertexBuffers2[i] = [];
          var ringbuffer1 = context.tempVertexBuffers1[i];
          var ringbuffer2 = context.tempVertexBuffers2[i];
          ringbuffer1.length = ringbuffer2.length = ringbufferLength;
          for(var j = 0; j < ringbufferLength; ++j) {
            ringbuffer1[j] = ringbuffer2[j] = null; // Created on-demand
          }
        }
  
        if (quads) {
          // GL_QUAD indexes can be precalculated
          context.tempQuadIndexBuffer = GLctx.createBuffer();
          context.GLctx.bindBuffer(context.GLctx.ELEMENT_ARRAY_BUFFER, context.tempQuadIndexBuffer);
          var numIndexes = GL.MAX_TEMP_BUFFER_SIZE >> 1;
          var quadIndexes = new Uint16Array(numIndexes);
          var i = 0, v = 0;
          while (1) {
            quadIndexes[i++] = v;
            if (i >= numIndexes) break;
            quadIndexes[i++] = v+1;
            if (i >= numIndexes) break;
            quadIndexes[i++] = v+2;
            if (i >= numIndexes) break;
            quadIndexes[i++] = v;
            if (i >= numIndexes) break;
            quadIndexes[i++] = v+2;
            if (i >= numIndexes) break;
            quadIndexes[i++] = v+3;
            if (i >= numIndexes) break;
            v += 4;
          }
          context.GLctx.bufferData(context.GLctx.ELEMENT_ARRAY_BUFFER, quadIndexes, context.GLctx.STATIC_DRAW);
          context.GLctx.bindBuffer(context.GLctx.ELEMENT_ARRAY_BUFFER, null);
        }
      },getTempVertexBuffer:function getTempVertexBuffer(sizeBytes) {
        var idx = GL.log2ceilLookup[sizeBytes];
        var ringbuffer = GL.currentContext.tempVertexBuffers1[idx];
        var nextFreeBufferIndex = GL.currentContext.tempVertexBufferCounters1[idx];
        GL.currentContext.tempVertexBufferCounters1[idx] = (GL.currentContext.tempVertexBufferCounters1[idx]+1) & (GL.numTempVertexBuffersPerSize-1);
        var vbo = ringbuffer[nextFreeBufferIndex];
        if (vbo) {
          return vbo;
        }
        var prevVBO = GLctx.getParameter(GLctx.ARRAY_BUFFER_BINDING);
        ringbuffer[nextFreeBufferIndex] = GLctx.createBuffer();
        GLctx.bindBuffer(GLctx.ARRAY_BUFFER, ringbuffer[nextFreeBufferIndex]);
        GLctx.bufferData(GLctx.ARRAY_BUFFER, 1 << idx, GLctx.DYNAMIC_DRAW);
        GLctx.bindBuffer(GLctx.ARRAY_BUFFER, prevVBO);
        return ringbuffer[nextFreeBufferIndex];
      },getTempIndexBuffer:function getTempIndexBuffer(sizeBytes) {
        var idx = GL.log2ceilLookup[sizeBytes];
        var ibo = GL.currentContext.tempIndexBuffers[idx];
        if (ibo) {
          return ibo;
        }
        var prevIBO = GLctx.getParameter(GLctx.ELEMENT_ARRAY_BUFFER_BINDING);
        GL.currentContext.tempIndexBuffers[idx] = GLctx.createBuffer();
        GLctx.bindBuffer(GLctx.ELEMENT_ARRAY_BUFFER, GL.currentContext.tempIndexBuffers[idx]);
        GLctx.bufferData(GLctx.ELEMENT_ARRAY_BUFFER, 1 << idx, GLctx.DYNAMIC_DRAW);
        GLctx.bindBuffer(GLctx.ELEMENT_ARRAY_BUFFER, prevIBO);
        return GL.currentContext.tempIndexBuffers[idx];
      },newRenderingFrameStarted:function newRenderingFrameStarted() {
        if (!GL.currentContext) {
          return;
        }
        var vb = GL.currentContext.tempVertexBuffers1;
        GL.currentContext.tempVertexBuffers1 = GL.currentContext.tempVertexBuffers2;
        GL.currentContext.tempVertexBuffers2 = vb;
        vb = GL.currentContext.tempVertexBufferCounters1;
        GL.currentContext.tempVertexBufferCounters1 = GL.currentContext.tempVertexBufferCounters2;
        GL.currentContext.tempVertexBufferCounters2 = vb;
        var largestIndex = GL.log2ceilLookup[GL.MAX_TEMP_BUFFER_SIZE];
        for(var i = 0; i <= largestIndex; ++i) {
          GL.currentContext.tempVertexBufferCounters1[i] = 0;
        }
      },getSource:function (shader, count, string, length) {
        var source = '';
        for (var i = 0; i < count; ++i) {
          var frag;
          if (length) {
            var len = HEAP32[(((length)+(i*4))>>2)];
            if (len < 0) {
              frag = Pointer_stringify(HEAP32[(((string)+(i*4))>>2)]);
            } else {
              frag = Pointer_stringify(HEAP32[(((string)+(i*4))>>2)], len);
            }
          } else {
            frag = Pointer_stringify(HEAP32[(((string)+(i*4))>>2)]);
          }
          source += frag;
        }
        return source;
      },calcBufLength:function calcBufLength(size, type, stride, count) {
        if (stride > 0) {
          return count * stride;  // XXXvlad this is not exactly correct I don't think
        }
        var typeSize = GL.byteSizeByType[type - GL.byteSizeByTypeRoot];
        return size * typeSize * count;
      },usedTempBuffers:[],preDrawHandleClientVertexAttribBindings:function preDrawHandleClientVertexAttribBindings(count) {
        GL.resetBufferBinding = false;
  
        // TODO: initial pass to detect ranges we need to upload, might not need an upload per attrib
        for (var i = 0; i < GL.currentContext.maxVertexAttribs; ++i) {
          var cb = GL.currentContext.clientBuffers[i];
          if (!cb.clientside || !cb.enabled) continue;
  
          GL.resetBufferBinding = true;
  
          var size = GL.calcBufLength(cb.size, cb.type, cb.stride, count);
          var buf = GL.getTempVertexBuffer(size);
          GLctx.bindBuffer(GLctx.ARRAY_BUFFER, buf);
          GLctx.bufferSubData(GLctx.ARRAY_BUFFER,
                                   0,
                                   HEAPU8.subarray(cb.ptr, cb.ptr + size));
          GLctx.vertexAttribPointer(i, cb.size, cb.type, cb.normalized, cb.stride, 0);
        }
      },postDrawHandleClientVertexAttribBindings:function postDrawHandleClientVertexAttribBindings() {
        if (GL.resetBufferBinding) {
          GLctx.bindBuffer(GLctx.ARRAY_BUFFER, GL.buffers[GL.currArrayBuffer]);
        }
      },createContext:function (canvas, webGLContextAttributes) {
        if (typeof webGLContextAttributes.majorVersion === 'undefined' && typeof webGLContextAttributes.minorVersion === 'undefined') {
          webGLContextAttributes.majorVersion = 1;
          webGLContextAttributes.minorVersion = 0;
        }
        var ctx;
        var errorInfo = '?';
        function onContextCreationError(event) {
          errorInfo = event.statusMessage || errorInfo;
        }
        try {
          canvas.addEventListener('webglcontextcreationerror', onContextCreationError, false);
          try {
            if (webGLContextAttributes.majorVersion == 1 && webGLContextAttributes.minorVersion == 0) {
              ctx = canvas.getContext("webgl", webGLContextAttributes) || canvas.getContext("experimental-webgl", webGLContextAttributes);
            } else if (webGLContextAttributes.majorVersion == 2 && webGLContextAttributes.minorVersion == 0) {
              ctx = canvas.getContext("webgl2", webGLContextAttributes) || canvas.getContext("experimental-webgl2", webGLContextAttributes);
            } else {
              throw 'Unsupported WebGL context version ' + majorVersion + '.' + minorVersion + '!'
            }
          } finally {
            canvas.removeEventListener('webglcontextcreationerror', onContextCreationError, false);
          }
          if (!ctx) throw ':(';
        } catch (e) {
          Module.print('Could not create canvas: ' + [errorInfo, e, JSON.stringify(webGLContextAttributes)]);
          return 0;
        }
        // possible GL_DEBUG entry point: ctx = wrapDebugGL(ctx);
  
        if (!ctx) return 0;
        return GL.registerContext(ctx, webGLContextAttributes);
      },registerContext:function (ctx, webGLContextAttributes) {
        var handle = GL.getNewId(GL.contexts);
        var context = {
          handle: handle,
          version: webGLContextAttributes.majorVersion,
          GLctx: ctx
        };
        // Store the created context object so that we can access the context given a canvas without having to pass the parameters again.
        if (ctx.canvas) ctx.canvas.GLctxObject = context;
        GL.contexts[handle] = context;
        if (typeof webGLContextAttributes['enableExtensionsByDefault'] === 'undefined' || webGLContextAttributes.enableExtensionsByDefault) {
          GL.initExtensions(context);
        }
        return handle;
      },makeContextCurrent:function (contextHandle) {
        var context = GL.contexts[contextHandle];
        if (!context) return false;
        GLctx = Module.ctx = context.GLctx; // Active WebGL context object.
        GL.currentContext = context; // Active Emscripten GL layer context object.
        return true;
      },getContext:function (contextHandle) {
        return GL.contexts[contextHandle];
      },deleteContext:function (contextHandle) {
        if (GL.currentContext === GL.contexts[contextHandle]) GL.currentContext = null;
        if (typeof JSEvents === 'object') JSEvents.removeAllHandlersOnTarget(GL.contexts[contextHandle].GLctx.canvas); // Release all JS event handlers on the DOM element that the GL context is associated with since the context is now deleted.
        if (GL.contexts[contextHandle] && GL.contexts[contextHandle].GLctx.canvas) GL.contexts[contextHandle].GLctx.canvas.GLctxObject = undefined; // Make sure the canvas object no longer refers to the context object so there are no GC surprises.
        GL.contexts[contextHandle] = null;
      },initExtensions:function (context) {
        // If this function is called without a specific context object, init the extensions of the currently active context.
        if (!context) context = GL.currentContext;
  
        if (context.initExtensionsDone) return;
        context.initExtensionsDone = true;
  
        var GLctx = context.GLctx;
  
        context.maxVertexAttribs = GLctx.getParameter(GLctx.MAX_VERTEX_ATTRIBS);
        context.clientBuffers = [];
        for (var i = 0; i < context.maxVertexAttribs; i++) {
          context.clientBuffers[i] = { enabled: false, clientside: false, size: 0, type: 0, normalized: 0, stride: 0, ptr: 0 };
        }
  
        GL.generateTempBuffers(false, context);
  
        // Detect the presence of a few extensions manually, this GL interop layer itself will need to know if they exist. 
  
        if (context.version < 2) {
          // Extension available from Firefox 26 and Google Chrome 30
          var instancedArraysExt = GLctx.getExtension('ANGLE_instanced_arrays');
          if (instancedArraysExt) {
            GLctx['vertexAttribDivisor'] = function(index, divisor) { instancedArraysExt['vertexAttribDivisorANGLE'](index, divisor); };
            GLctx['drawArraysInstanced'] = function(mode, first, count, primcount) { instancedArraysExt['drawArraysInstancedANGLE'](mode, first, count, primcount); };
            GLctx['drawElementsInstanced'] = function(mode, count, type, indices, primcount) { instancedArraysExt['drawElementsInstancedANGLE'](mode, count, type, indices, primcount); };
          }
  
          // Extension available from Firefox 25 and WebKit
          var vaoExt = GLctx.getExtension('OES_vertex_array_object');
          if (vaoExt) {
            GLctx['createVertexArray'] = function() { return vaoExt['createVertexArrayOES'](); };
            GLctx['deleteVertexArray'] = function(vao) { vaoExt['deleteVertexArrayOES'](vao); };
            GLctx['bindVertexArray'] = function(vao) { vaoExt['bindVertexArrayOES'](vao); };
            GLctx['isVertexArray'] = function(vao) { return vaoExt['isVertexArrayOES'](vao); };
          }
  
          var drawBuffersExt = GLctx.getExtension('WEBGL_draw_buffers');
          if (drawBuffersExt) {
            GLctx['drawBuffers'] = function(n, bufs) { drawBuffersExt['drawBuffersWEBGL'](n, bufs); };
          }
        }
  
        // These are the 'safe' feature-enabling extensions that don't add any performance impact related to e.g. debugging, and
        // should be enabled by default so that client GLES2/GL code will not need to go through extra hoops to get its stuff working.
        // As new extensions are ratified at http://www.khronos.org/registry/webgl/extensions/ , feel free to add your new extensions
        // here, as long as they don't produce a performance impact for users that might not be using those extensions.
        // E.g. debugging-related extensions should probably be off by default.
        var automaticallyEnabledExtensions = [ "OES_texture_float", "OES_texture_half_float", "OES_standard_derivatives",
                                               "OES_vertex_array_object", "WEBGL_compressed_texture_s3tc", "WEBGL_depth_texture",
                                               "OES_element_index_uint", "EXT_texture_filter_anisotropic", "ANGLE_instanced_arrays",
                                               "OES_texture_float_linear", "OES_texture_half_float_linear", "WEBGL_compressed_texture_atc",
                                               "WEBGL_compressed_texture_pvrtc", "EXT_color_buffer_half_float", "WEBGL_color_buffer_float",
                                               "EXT_frag_depth", "EXT_sRGB", "WEBGL_draw_buffers", "WEBGL_shared_resources",
                                               "EXT_shader_texture_lod" ];
  
        function shouldEnableAutomatically(extension) {
          var ret = false;
          automaticallyEnabledExtensions.forEach(function(include) {
            if (ext.indexOf(include) != -1) {
              ret = true;
            }
          });
          return ret;
        }
  
        var exts = GLctx.getSupportedExtensions();
        if (exts && exts.length > 0) {
          GLctx.getSupportedExtensions().forEach(function(ext) {
            if (automaticallyEnabledExtensions.indexOf(ext) != -1) {
              GLctx.getExtension(ext); // Calling .getExtension enables that extension permanently, no need to store the return value to be enabled.
            }
          });
        }
      },populateUniformTable:function (program) {
        var p = GL.programs[program];
        GL.programInfos[program] = {
          uniforms: {},
          maxUniformLength: 0, // This is eagerly computed below, since we already enumerate all uniforms anyway.
          maxAttributeLength: -1 // This is lazily computed and cached, computed when/if first asked, "-1" meaning not computed yet.
        };
  
        var ptable = GL.programInfos[program];
        var utable = ptable.uniforms;
        // A program's uniform table maps the string name of an uniform to an integer location of that uniform.
        // The global GL.uniforms map maps integer locations to WebGLUniformLocations.
        var numUniforms = GLctx.getProgramParameter(p, GLctx.ACTIVE_UNIFORMS);
        for (var i = 0; i < numUniforms; ++i) {
          var u = GLctx.getActiveUniform(p, i);
  
          var name = u.name;
          ptable.maxUniformLength = Math.max(ptable.maxUniformLength, name.length+1);
  
          // Strip off any trailing array specifier we might have got, e.g. "[0]".
          if (name.indexOf(']', name.length-1) !== -1) {
            var ls = name.lastIndexOf('[');
            name = name.slice(0, ls);
          }
  
          // Optimize memory usage slightly: If we have an array of uniforms, e.g. 'vec3 colors[3];', then 
          // only store the string 'colors' in utable, and 'colors[0]', 'colors[1]' and 'colors[2]' will be parsed as 'colors'+i.
          // Note that for the GL.uniforms table, we still need to fetch the all WebGLUniformLocations for all the indices.
          var loc = GLctx.getUniformLocation(p, name);
          var id = GL.getNewId(GL.uniforms);
          utable[name] = [u.size, id];
          GL.uniforms[id] = loc;
  
          for (var j = 1; j < u.size; ++j) {
            var n = name + '['+j+']';
            loc = GLctx.getUniformLocation(p, n);
            id = GL.getNewId(GL.uniforms);
  
            GL.uniforms[id] = loc;
          }
        }
      }};function _emscripten_glIsRenderbuffer(renderbuffer) {
      var rb = GL.renderbuffers[renderbuffer];
      if (!rb) return 0;
      return GLctx.isRenderbuffer(rb);
    }

  function _emscripten_glStencilMaskSeparate(x0, x1) { GLctx.stencilMaskSeparate(x0, x1) }

  
  
  
  
  function _emscripten_set_main_loop_timing(mode, value) {
      Browser.mainLoop.timingMode = mode;
      Browser.mainLoop.timingValue = value;
  
      if (!Browser.mainLoop.func) {
        console.error('emscripten_set_main_loop_timing: Cannot set timing mode for main loop since a main loop does not exist! Call emscripten_set_main_loop first to set one up.');
        return 1; // Return non-zero on failure, can't set timing mode when there is no main loop.
      }
  
      if (mode == 0 /*EM_TIMING_SETTIMEOUT*/) {
        Browser.mainLoop.scheduler = function Browser_mainLoop_scheduler_setTimeout() {
          setTimeout(Browser.mainLoop.runner, value); // doing this each time means that on exception, we stop
        };
        Browser.mainLoop.method = 'timeout';
      } else if (mode == 1 /*EM_TIMING_RAF*/) {
        Browser.mainLoop.scheduler = function Browser_mainLoop_scheduler_rAF() {
          Browser.requestAnimationFrame(Browser.mainLoop.runner);
        };
        Browser.mainLoop.method = 'rAF';
      } else if (mode == 2 /*EM_TIMING_SETIMMEDIATE*/) {
        if (!window['setImmediate']) {
          // Emulate setImmediate. (note: not a complete polyfill, we don't emulate clearImmediate() to keep code size to minimum, since not needed)
          var setImmediates = [];
          var emscriptenMainLoopMessageId = '__emcc';
          function Browser_setImmediate_messageHandler(event) {
            if (event.source === window && event.data === emscriptenMainLoopMessageId) {
              event.stopPropagation();
              setImmediates.shift()();
            }
          }
          window.addEventListener("message", Browser_setImmediate_messageHandler, true);
          window['setImmediate'] = function Browser_emulated_setImmediate(func) {
            setImmediates.push(func);
            window.postMessage(emscriptenMainLoopMessageId, "*");
          }
        }
        Browser.mainLoop.scheduler = function Browser_mainLoop_scheduler_setImmediate() {
          window['setImmediate'](Browser.mainLoop.runner);
        };
        Browser.mainLoop.method = 'immediate';
      }
      return 0;
    }function _emscripten_set_main_loop(func, fps, simulateInfiniteLoop, arg, noSetTiming) {
      Module['noExitRuntime'] = true;
  
      assert(!Browser.mainLoop.func, 'emscripten_set_main_loop: there can only be one main loop function at once: call emscripten_cancel_main_loop to cancel the previous one before setting a new one with different parameters.');
  
      Browser.mainLoop.func = func;
      Browser.mainLoop.arg = arg;
  
      var thisMainLoopId = Browser.mainLoop.currentlyRunningMainloop;
  
      Browser.mainLoop.runner = function Browser_mainLoop_runner() {
        if (ABORT) return;
        if (Browser.mainLoop.queue.length > 0) {
          var start = Date.now();
          var blocker = Browser.mainLoop.queue.shift();
          blocker.func(blocker.arg);
          if (Browser.mainLoop.remainingBlockers) {
            var remaining = Browser.mainLoop.remainingBlockers;
            var next = remaining%1 == 0 ? remaining-1 : Math.floor(remaining);
            if (blocker.counted) {
              Browser.mainLoop.remainingBlockers = next;
            } else {
              // not counted, but move the progress along a tiny bit
              next = next + 0.5; // do not steal all the next one's progress
              Browser.mainLoop.remainingBlockers = (8*remaining + next)/9;
            }
          }
          console.log('main loop blocker "' + blocker.name + '" took ' + (Date.now() - start) + ' ms'); //, left: ' + Browser.mainLoop.remainingBlockers);
          Browser.mainLoop.updateStatus();
  
          // catches pause/resume main loop from blocker execution
          if (thisMainLoopId < Browser.mainLoop.currentlyRunningMainloop) return;
  
          setTimeout(Browser.mainLoop.runner, 0);
          return;
        }
  
        // catch pauses from non-main loop sources
        if (thisMainLoopId < Browser.mainLoop.currentlyRunningMainloop) return;
  
        // Implement very basic swap interval control
        Browser.mainLoop.currentFrameNumber = Browser.mainLoop.currentFrameNumber + 1 | 0;
        if (Browser.mainLoop.timingMode == 1/*EM_TIMING_RAF*/ && Browser.mainLoop.timingValue > 1 && Browser.mainLoop.currentFrameNumber % Browser.mainLoop.timingValue != 0) {
          // Not the scheduled time to render this frame - skip.
          Browser.mainLoop.scheduler();
          return;
        }
  
        // Signal GL rendering layer that processing of a new frame is about to start. This helps it optimize
        // VBO double-buffering and reduce GPU stalls.
        GL.newRenderingFrameStarted();
  
        if (Browser.mainLoop.method === 'timeout' && Module.ctx) {
          Module.printErr('Looks like you are rendering without using requestAnimationFrame for the main loop. You should use 0 for the frame rate in emscripten_set_main_loop in order to use requestAnimationFrame, as that can greatly improve your frame rates!');
          Browser.mainLoop.method = ''; // just warn once per call to set main loop
        }
  
        Browser.mainLoop.runIter(function() {
          if (typeof arg !== 'undefined') {
            Runtime.dynCall('vi', func, [arg]);
          } else {
            Runtime.dynCall('v', func);
          }
        });
  
        // catch pauses from the main loop itself
        if (thisMainLoopId < Browser.mainLoop.currentlyRunningMainloop) return;
  
        // Queue new audio data. This is important to be right after the main loop invocation, so that we will immediately be able
        // to queue the newest produced audio samples.
        // TODO: Consider adding pre- and post- rAF callbacks so that GL.newRenderingFrameStarted() and SDL.audio.queueNewAudioData()
        //       do not need to be hardcoded into this function, but can be more generic.
        if (typeof SDL === 'object' && SDL.audio && SDL.audio.queueNewAudioData) SDL.audio.queueNewAudioData();
  
        Browser.mainLoop.scheduler();
      }
  
      if (!noSetTiming) {
        if (fps && fps > 0) _emscripten_set_main_loop_timing(0/*EM_TIMING_SETTIMEOUT*/, 1000.0 / fps);
        else _emscripten_set_main_loop_timing(1/*EM_TIMING_RAF*/, 1); // Do rAF by rendering each frame (no decimating)
  
        Browser.mainLoop.scheduler();
      }
  
      if (simulateInfiniteLoop) {
        throw 'SimulateInfiniteLoop';
      }
    }var Browser={mainLoop:{scheduler:null,method:"",currentlyRunningMainloop:0,func:null,arg:0,timingMode:0,timingValue:0,currentFrameNumber:0,queue:[],pause:function () {
          Browser.mainLoop.scheduler = null;
          Browser.mainLoop.currentlyRunningMainloop++; // Incrementing this signals the previous main loop that it's now become old, and it must return.
        },resume:function () {
          Browser.mainLoop.currentlyRunningMainloop++;
          var timingMode = Browser.mainLoop.timingMode;
          var timingValue = Browser.mainLoop.timingValue;
          var func = Browser.mainLoop.func;
          Browser.mainLoop.func = null;
          _emscripten_set_main_loop(func, 0, false, Browser.mainLoop.arg, true /* do not set timing and call scheduler, we will do it on the next lines */);
          _emscripten_set_main_loop_timing(timingMode, timingValue);
          Browser.mainLoop.scheduler();
        },updateStatus:function () {
          if (Module['setStatus']) {
            var message = Module['statusMessage'] || 'Please wait...';
            var remaining = Browser.mainLoop.remainingBlockers;
            var expected = Browser.mainLoop.expectedBlockers;
            if (remaining) {
              if (remaining < expected) {
                Module['setStatus'](message + ' (' + (expected - remaining) + '/' + expected + ')');
              } else {
                Module['setStatus'](message);
              }
            } else {
              Module['setStatus']('');
            }
          }
        },runIter:function (func) {
          if (ABORT) return;
          if (Module['preMainLoop']) {
            var preRet = Module['preMainLoop']();
            if (preRet === false) {
              return; // |return false| skips a frame
            }
          }
          try {
            func();
          } catch (e) {
            if (e instanceof ExitStatus) {
              return;
            } else {
              if (e && typeof e === 'object' && e.stack) Module.printErr('exception thrown: ' + [e, e.stack]);
              throw e;
            }
          }
          if (Module['postMainLoop']) Module['postMainLoop']();
        }},isFullScreen:false,pointerLock:false,moduleContextCreatedCallbacks:[],workers:[],init:function () {
        if (!Module["preloadPlugins"]) Module["preloadPlugins"] = []; // needs to exist even in workers
  
        if (Browser.initted) return;
        Browser.initted = true;
  
        try {
          new Blob();
          Browser.hasBlobConstructor = true;
        } catch(e) {
          Browser.hasBlobConstructor = false;
          console.log("warning: no blob constructor, cannot create blobs with mimetypes");
        }
        Browser.BlobBuilder = typeof MozBlobBuilder != "undefined" ? MozBlobBuilder : (typeof WebKitBlobBuilder != "undefined" ? WebKitBlobBuilder : (!Browser.hasBlobConstructor ? console.log("warning: no BlobBuilder") : null));
        Browser.URLObject = typeof window != "undefined" ? (window.URL ? window.URL : window.webkitURL) : undefined;
        if (!Module.noImageDecoding && typeof Browser.URLObject === 'undefined') {
          console.log("warning: Browser does not support creating object URLs. Built-in browser image decoding will not be available.");
          Module.noImageDecoding = true;
        }
  
        // Support for plugins that can process preloaded files. You can add more of these to
        // your app by creating and appending to Module.preloadPlugins.
        //
        // Each plugin is asked if it can handle a file based on the file's name. If it can,
        // it is given the file's raw data. When it is done, it calls a callback with the file's
        // (possibly modified) data. For example, a plugin might decompress a file, or it
        // might create some side data structure for use later (like an Image element, etc.).
  
        var imagePlugin = {};
        imagePlugin['canHandle'] = function imagePlugin_canHandle(name) {
          return !Module.noImageDecoding && /\.(jpg|jpeg|png|bmp)$/i.test(name);
        };
        imagePlugin['handle'] = function imagePlugin_handle(byteArray, name, onload, onerror) {
          var b = null;
          if (Browser.hasBlobConstructor) {
            try {
              b = new Blob([byteArray], { type: Browser.getMimetype(name) });
              if (b.size !== byteArray.length) { // Safari bug #118630
                // Safari's Blob can only take an ArrayBuffer
                b = new Blob([(new Uint8Array(byteArray)).buffer], { type: Browser.getMimetype(name) });
              }
            } catch(e) {
              Runtime.warnOnce('Blob constructor present but fails: ' + e + '; falling back to blob builder');
            }
          }
          if (!b) {
            var bb = new Browser.BlobBuilder();
            bb.append((new Uint8Array(byteArray)).buffer); // we need to pass a buffer, and must copy the array to get the right data range
            b = bb.getBlob();
          }
          var url = Browser.URLObject.createObjectURL(b);
          assert(typeof url == 'string', 'createObjectURL must return a url as a string');
          var img = new Image();
          img.onload = function img_onload() {
            assert(img.complete, 'Image ' + name + ' could not be decoded');
            var canvas = document.createElement('canvas');
            canvas.width = img.width;
            canvas.height = img.height;
            var ctx = canvas.getContext('2d');
            ctx.drawImage(img, 0, 0);
            Module["preloadedImages"][name] = canvas;
            Browser.URLObject.revokeObjectURL(url);
            if (onload) onload(byteArray);
          };
          img.onerror = function img_onerror(event) {
            console.log('Image ' + url + ' could not be decoded');
            if (onerror) onerror();
          };
          img.src = url;
        };
        Module['preloadPlugins'].push(imagePlugin);
  
        var audioPlugin = {};
        audioPlugin['canHandle'] = function audioPlugin_canHandle(name) {
          return !Module.noAudioDecoding && name.substr(-4) in { '.ogg': 1, '.wav': 1, '.mp3': 1 };
        };
        audioPlugin['handle'] = function audioPlugin_handle(byteArray, name, onload, onerror) {
          var done = false;
          function finish(audio) {
            if (done) return;
            done = true;
            Module["preloadedAudios"][name] = audio;
            if (onload) onload(byteArray);
          }
          function fail() {
            if (done) return;
            done = true;
            Module["preloadedAudios"][name] = new Audio(); // empty shim
            if (onerror) onerror();
          }
          if (Browser.hasBlobConstructor) {
            try {
              var b = new Blob([byteArray], { type: Browser.getMimetype(name) });
            } catch(e) {
              return fail();
            }
            var url = Browser.URLObject.createObjectURL(b); // XXX we never revoke this!
            assert(typeof url == 'string', 'createObjectURL must return a url as a string');
            var audio = new Audio();
            audio.addEventListener('canplaythrough', function() { finish(audio) }, false); // use addEventListener due to chromium bug 124926
            audio.onerror = function audio_onerror(event) {
              if (done) return;
              console.log('warning: browser could not fully decode audio ' + name + ', trying slower base64 approach');
              function encode64(data) {
                var BASE = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';
                var PAD = '=';
                var ret = '';
                var leftchar = 0;
                var leftbits = 0;
                for (var i = 0; i < data.length; i++) {
                  leftchar = (leftchar << 8) | data[i];
                  leftbits += 8;
                  while (leftbits >= 6) {
                    var curr = (leftchar >> (leftbits-6)) & 0x3f;
                    leftbits -= 6;
                    ret += BASE[curr];
                  }
                }
                if (leftbits == 2) {
                  ret += BASE[(leftchar&3) << 4];
                  ret += PAD + PAD;
                } else if (leftbits == 4) {
                  ret += BASE[(leftchar&0xf) << 2];
                  ret += PAD;
                }
                return ret;
              }
              audio.src = 'data:audio/x-' + name.substr(-3) + ';base64,' + encode64(byteArray);
              finish(audio); // we don't wait for confirmation this worked - but it's worth trying
            };
            audio.src = url;
            // workaround for chrome bug 124926 - we do not always get oncanplaythrough or onerror
            Browser.safeSetTimeout(function() {
              finish(audio); // try to use it even though it is not necessarily ready to play
            }, 10000);
          } else {
            return fail();
          }
        };
        Module['preloadPlugins'].push(audioPlugin);
  
        // Canvas event setup
  
        var canvas = Module['canvas'];
        function pointerLockChange() {
          Browser.pointerLock = document['pointerLockElement'] === canvas ||
                                document['mozPointerLockElement'] === canvas ||
                                document['webkitPointerLockElement'] === canvas ||
                                document['msPointerLockElement'] === canvas;
        }
        if (canvas) {
          // forced aspect ratio can be enabled by defining 'forcedAspectRatio' on Module
          // Module['forcedAspectRatio'] = 4 / 3;
  
          canvas.requestPointerLock = canvas['requestPointerLock'] ||
                                      canvas['mozRequestPointerLock'] ||
                                      canvas['webkitRequestPointerLock'] ||
                                      canvas['msRequestPointerLock'] ||
                                      function(){};
          canvas.exitPointerLock = document['exitPointerLock'] ||
                                   document['mozExitPointerLock'] ||
                                   document['webkitExitPointerLock'] ||
                                   document['msExitPointerLock'] ||
                                   function(){}; // no-op if function does not exist
          canvas.exitPointerLock = canvas.exitPointerLock.bind(document);
  
  
          document.addEventListener('pointerlockchange', pointerLockChange, false);
          document.addEventListener('mozpointerlockchange', pointerLockChange, false);
          document.addEventListener('webkitpointerlockchange', pointerLockChange, false);
          document.addEventListener('mspointerlockchange', pointerLockChange, false);
  
          if (Module['elementPointerLock']) {
            canvas.addEventListener("click", function(ev) {
              if (!Browser.pointerLock && canvas.requestPointerLock) {
                canvas.requestPointerLock();
                ev.preventDefault();
              }
            }, false);
          }
        }
      },createContext:function (canvas, useWebGL, setInModule, webGLContextAttributes) {
        if (useWebGL && Module.ctx && canvas == Module.canvas) return Module.ctx; // no need to recreate GL context if it's already been created for this canvas.
  
        var ctx;
        var contextHandle;
        if (useWebGL) {
          // For GLES2/desktop GL compatibility, adjust a few defaults to be different to WebGL defaults, so that they align better with the desktop defaults.
          var contextAttributes = {
            antialias: false,
            alpha: false
          };
  
          if (webGLContextAttributes) {
            for (var attribute in webGLContextAttributes) {
              contextAttributes[attribute] = webGLContextAttributes[attribute];
            }
          }
  
          contextHandle = GL.createContext(canvas, contextAttributes);
          if (contextHandle) {
            ctx = GL.getContext(contextHandle).GLctx;
          }
          // Set the background of the WebGL canvas to black
          canvas.style.backgroundColor = "black";
        } else {
          ctx = canvas.getContext('2d');
        }
  
        if (!ctx) return null;
  
        if (setInModule) {
          if (!useWebGL) assert(typeof GLctx === 'undefined', 'cannot set in module if GLctx is used, but we are a non-GL context that would replace it');
  
          Module.ctx = ctx;
          if (useWebGL) GL.makeContextCurrent(contextHandle);
          Module.useWebGL = useWebGL;
          Browser.moduleContextCreatedCallbacks.forEach(function(callback) { callback() });
          Browser.init();
        }
        return ctx;
      },destroyContext:function (canvas, useWebGL, setInModule) {},fullScreenHandlersInstalled:false,lockPointer:undefined,resizeCanvas:undefined,requestFullScreen:function (lockPointer, resizeCanvas, vrDevice) {
        Browser.lockPointer = lockPointer;
        Browser.resizeCanvas = resizeCanvas;
        Browser.vrDevice = vrDevice;
        if (typeof Browser.lockPointer === 'undefined') Browser.lockPointer = true;
        if (typeof Browser.resizeCanvas === 'undefined') Browser.resizeCanvas = false;
        if (typeof Browser.vrDevice === 'undefined') Browser.vrDevice = null;
  
        var canvas = Module['canvas'];
        function fullScreenChange() {
          Browser.isFullScreen = false;
          var canvasContainer = canvas.parentNode;
          if ((document['webkitFullScreenElement'] || document['webkitFullscreenElement'] ||
               document['mozFullScreenElement'] || document['mozFullscreenElement'] ||
               document['fullScreenElement'] || document['fullscreenElement'] ||
               document['msFullScreenElement'] || document['msFullscreenElement'] ||
               document['webkitCurrentFullScreenElement']) === canvasContainer) {
            canvas.cancelFullScreen = document['cancelFullScreen'] ||
                                      document['mozCancelFullScreen'] ||
                                      document['webkitCancelFullScreen'] ||
                                      document['msExitFullscreen'] ||
                                      document['exitFullscreen'] ||
                                      function() {};
            canvas.cancelFullScreen = canvas.cancelFullScreen.bind(document);
            if (Browser.lockPointer) canvas.requestPointerLock();
            Browser.isFullScreen = true;
            if (Browser.resizeCanvas) Browser.setFullScreenCanvasSize();
          } else {
  
            // remove the full screen specific parent of the canvas again to restore the HTML structure from before going full screen
            canvasContainer.parentNode.insertBefore(canvas, canvasContainer);
            canvasContainer.parentNode.removeChild(canvasContainer);
  
            if (Browser.resizeCanvas) Browser.setWindowedCanvasSize();
          }
          if (Module['onFullScreen']) Module['onFullScreen'](Browser.isFullScreen);
          Browser.updateCanvasDimensions(canvas);
        }
  
        if (!Browser.fullScreenHandlersInstalled) {
          Browser.fullScreenHandlersInstalled = true;
          document.addEventListener('fullscreenchange', fullScreenChange, false);
          document.addEventListener('mozfullscreenchange', fullScreenChange, false);
          document.addEventListener('webkitfullscreenchange', fullScreenChange, false);
          document.addEventListener('MSFullscreenChange', fullScreenChange, false);
        }
  
        // create a new parent to ensure the canvas has no siblings. this allows browsers to optimize full screen performance when its parent is the full screen root
        var canvasContainer = document.createElement("div");
        canvas.parentNode.insertBefore(canvasContainer, canvas);
        canvasContainer.appendChild(canvas);
  
        // use parent of canvas as full screen root to allow aspect ratio correction (Firefox stretches the root to screen size)
        canvasContainer.requestFullScreen = canvasContainer['requestFullScreen'] ||
                                            canvasContainer['mozRequestFullScreen'] ||
                                            canvasContainer['msRequestFullscreen'] ||
                                           (canvasContainer['webkitRequestFullScreen'] ? function() { canvasContainer['webkitRequestFullScreen'](Element['ALLOW_KEYBOARD_INPUT']) } : null);
  
        if (vrDevice) {
          canvasContainer.requestFullScreen({ vrDisplay: vrDevice });
        } else {
          canvasContainer.requestFullScreen();
        }
      },nextRAF:0,fakeRequestAnimationFrame:function (func) {
        // try to keep 60fps between calls to here
        var now = Date.now();
        if (Browser.nextRAF === 0) {
          Browser.nextRAF = now + 1000/60;
        } else {
          while (now + 2 >= Browser.nextRAF) { // fudge a little, to avoid timer jitter causing us to do lots of delay:0
            Browser.nextRAF += 1000/60;
          }
        }
        var delay = Math.max(Browser.nextRAF - now, 0);
        setTimeout(func, delay);
      },requestAnimationFrame:function requestAnimationFrame(func) {
        if (typeof window === 'undefined') { // Provide fallback to setTimeout if window is undefined (e.g. in Node.js)
          Browser.fakeRequestAnimationFrame(func);
        } else {
          if (!window.requestAnimationFrame) {
            window.requestAnimationFrame = window['requestAnimationFrame'] ||
                                           window['mozRequestAnimationFrame'] ||
                                           window['webkitRequestAnimationFrame'] ||
                                           window['msRequestAnimationFrame'] ||
                                           window['oRequestAnimationFrame'] ||
                                           Browser.fakeRequestAnimationFrame;
          }
          window.requestAnimationFrame(func);
        }
      },safeCallback:function (func) {
        return function() {
          if (!ABORT) return func.apply(null, arguments);
        };
      },allowAsyncCallbacks:true,queuedAsyncCallbacks:[],pauseAsyncCallbacks:function () {
        Browser.allowAsyncCallbacks = false;
      },resumeAsyncCallbacks:function () { // marks future callbacks as ok to execute, and synchronously runs any remaining ones right now
        Browser.allowAsyncCallbacks = true;
        if (Browser.queuedAsyncCallbacks.length > 0) {
          var callbacks = Browser.queuedAsyncCallbacks;
          Browser.queuedAsyncCallbacks = [];
          callbacks.forEach(function(func) {
            func();
          });
        }
      },safeRequestAnimationFrame:function (func) {
        return Browser.requestAnimationFrame(function() {
          if (ABORT) return;
          if (Browser.allowAsyncCallbacks) {
            func();
          } else {
            Browser.queuedAsyncCallbacks.push(func);
          }
        });
      },safeSetTimeout:function (func, timeout) {
        Module['noExitRuntime'] = true;
        return setTimeout(function() {
          if (ABORT) return;
          if (Browser.allowAsyncCallbacks) {
            func();
          } else {
            Browser.queuedAsyncCallbacks.push(func);
          }
        }, timeout);
      },safeSetInterval:function (func, timeout) {
        Module['noExitRuntime'] = true;
        return setInterval(function() {
          if (ABORT) return;
          if (Browser.allowAsyncCallbacks) {
            func();
          } // drop it on the floor otherwise, next interval will kick in
        }, timeout);
      },getMimetype:function (name) {
        return {
          'jpg': 'image/jpeg',
          'jpeg': 'image/jpeg',
          'png': 'image/png',
          'bmp': 'image/bmp',
          'ogg': 'audio/ogg',
          'wav': 'audio/wav',
          'mp3': 'audio/mpeg'
        }[name.substr(name.lastIndexOf('.')+1)];
      },getUserMedia:function (func) {
        if(!window.getUserMedia) {
          window.getUserMedia = navigator['getUserMedia'] ||
                                navigator['mozGetUserMedia'];
        }
        window.getUserMedia(func);
      },getMovementX:function (event) {
        return event['movementX'] ||
               event['mozMovementX'] ||
               event['webkitMovementX'] ||
               0;
      },getMovementY:function (event) {
        return event['movementY'] ||
               event['mozMovementY'] ||
               event['webkitMovementY'] ||
               0;
      },getMouseWheelDelta:function (event) {
        var delta = 0;
        switch (event.type) {
          case 'DOMMouseScroll':
            delta = event.detail;
            break;
          case 'mousewheel':
            delta = event.wheelDelta;
            break;
          case 'wheel':
            delta = event['deltaY'];
            break;
          default:
            throw 'unrecognized mouse wheel event: ' + event.type;
        }
        return delta;
      },mouseX:0,mouseY:0,mouseMovementX:0,mouseMovementY:0,touches:{},lastTouches:{},calculateMouseEvent:function (event) { // event should be mousemove, mousedown or mouseup
        if (Browser.pointerLock) {
          // When the pointer is locked, calculate the coordinates
          // based on the movement of the mouse.
          // Workaround for Firefox bug 764498
          if (event.type != 'mousemove' &&
              ('mozMovementX' in event)) {
            Browser.mouseMovementX = Browser.mouseMovementY = 0;
          } else {
            Browser.mouseMovementX = Browser.getMovementX(event);
            Browser.mouseMovementY = Browser.getMovementY(event);
          }
  
          // check if SDL is available
          if (typeof SDL != "undefined") {
          	Browser.mouseX = SDL.mouseX + Browser.mouseMovementX;
          	Browser.mouseY = SDL.mouseY + Browser.mouseMovementY;
          } else {
          	// just add the mouse delta to the current absolut mouse position
          	// FIXME: ideally this should be clamped against the canvas size and zero
          	Browser.mouseX += Browser.mouseMovementX;
          	Browser.mouseY += Browser.mouseMovementY;
          }
        } else {
          // Otherwise, calculate the movement based on the changes
          // in the coordinates.
          var rect = Module["canvas"].getBoundingClientRect();
          var cw = Module["canvas"].width;
          var ch = Module["canvas"].height;
  
          // Neither .scrollX or .pageXOffset are defined in a spec, but
          // we prefer .scrollX because it is currently in a spec draft.
          // (see: http://www.w3.org/TR/2013/WD-cssom-view-20131217/)
          var scrollX = ((typeof window.scrollX !== 'undefined') ? window.scrollX : window.pageXOffset);
          var scrollY = ((typeof window.scrollY !== 'undefined') ? window.scrollY : window.pageYOffset);
          // If this assert lands, it's likely because the browser doesn't support scrollX or pageXOffset
          // and we have no viable fallback.
          assert((typeof scrollX !== 'undefined') && (typeof scrollY !== 'undefined'), 'Unable to retrieve scroll position, mouse positions likely broken.');
  
          if (event.type === 'touchstart' || event.type === 'touchend' || event.type === 'touchmove') {
            var touch = event.touch;
            if (touch === undefined) {
              return; // the "touch" property is only defined in SDL
  
            }
            var adjustedX = touch.pageX - (scrollX + rect.left);
            var adjustedY = touch.pageY - (scrollY + rect.top);
  
            adjustedX = adjustedX * (cw / rect.width);
            adjustedY = adjustedY * (ch / rect.height);
  
            var coords = { x: adjustedX, y: adjustedY };
  
            if (event.type === 'touchstart') {
              Browser.lastTouches[touch.identifier] = coords;
              Browser.touches[touch.identifier] = coords;
            } else if (event.type === 'touchend' || event.type === 'touchmove') {
              var last = Browser.touches[touch.identifier];
              if (!last) last = coords;
              Browser.lastTouches[touch.identifier] = last;
              Browser.touches[touch.identifier] = coords;
            }
            return;
          }
  
          var x = event.pageX - (scrollX + rect.left);
          var y = event.pageY - (scrollY + rect.top);
  
          // the canvas might be CSS-scaled compared to its backbuffer;
          // SDL-using content will want mouse coordinates in terms
          // of backbuffer units.
          x = x * (cw / rect.width);
          y = y * (ch / rect.height);
  
          Browser.mouseMovementX = x - Browser.mouseX;
          Browser.mouseMovementY = y - Browser.mouseY;
          Browser.mouseX = x;
          Browser.mouseY = y;
        }
      },asyncLoad:function (url, onload, onerror, noRunDep) {
        Module['readAsync'](url, function(arrayBuffer) {
          assert(arrayBuffer, 'Loading data file "' + url + '" failed (no arrayBuffer).');
          onload(new Uint8Array(arrayBuffer));
          if (!noRunDep) removeRunDependency('al ' + url);
        }, function(event) {
          if (onerror) {
            onerror();
          } else {
            throw 'Loading data file "' + url + '" failed.';
          }
        });
        if (!noRunDep) addRunDependency('al ' + url);
      },resizeListeners:[],updateResizeListeners:function () {
        var canvas = Module['canvas'];
        Browser.resizeListeners.forEach(function(listener) {
          listener(canvas.width, canvas.height);
        });
      },setCanvasSize:function (width, height, noUpdates) {
        var canvas = Module['canvas'];
        Browser.updateCanvasDimensions(canvas, width, height);
        if (!noUpdates) Browser.updateResizeListeners();
      },windowedWidth:0,windowedHeight:0,setFullScreenCanvasSize:function () {
        // check if SDL is available
        if (typeof SDL != "undefined") {
        	var flags = HEAPU32[((SDL.screen+Runtime.QUANTUM_SIZE*0)>>2)];
        	flags = flags | 0x00800000; // set SDL_FULLSCREEN flag
        	HEAP32[((SDL.screen+Runtime.QUANTUM_SIZE*0)>>2)]=flags
        }
        Browser.updateResizeListeners();
      },setWindowedCanvasSize:function () {
        // check if SDL is available
        if (typeof SDL != "undefined") {
        	var flags = HEAPU32[((SDL.screen+Runtime.QUANTUM_SIZE*0)>>2)];
        	flags = flags & ~0x00800000; // clear SDL_FULLSCREEN flag
        	HEAP32[((SDL.screen+Runtime.QUANTUM_SIZE*0)>>2)]=flags
        }
        Browser.updateResizeListeners();
      },updateCanvasDimensions:function (canvas, wNative, hNative) {
        if (wNative && hNative) {
          canvas.widthNative = wNative;
          canvas.heightNative = hNative;
        } else {
          wNative = canvas.widthNative;
          hNative = canvas.heightNative;
        }
        var w = wNative;
        var h = hNative;
        if (Module['forcedAspectRatio'] && Module['forcedAspectRatio'] > 0) {
          if (w/h < Module['forcedAspectRatio']) {
            w = Math.round(h * Module['forcedAspectRatio']);
          } else {
            h = Math.round(w / Module['forcedAspectRatio']);
          }
        }
        if (((document['webkitFullScreenElement'] || document['webkitFullscreenElement'] ||
             document['mozFullScreenElement'] || document['mozFullscreenElement'] ||
             document['fullScreenElement'] || document['fullscreenElement'] ||
             document['msFullScreenElement'] || document['msFullscreenElement'] ||
             document['webkitCurrentFullScreenElement']) === canvas.parentNode) && (typeof screen != 'undefined')) {
           var factor = Math.min(screen.width / w, screen.height / h);
           w = Math.round(w * factor);
           h = Math.round(h * factor);
        }
        if (Browser.resizeCanvas) {
          if (canvas.width  != w) canvas.width  = w;
          if (canvas.height != h) canvas.height = h;
          if (typeof canvas.style != 'undefined') {
            canvas.style.removeProperty( "width");
            canvas.style.removeProperty("height");
          }
        } else {
          if (canvas.width  != wNative) canvas.width  = wNative;
          if (canvas.height != hNative) canvas.height = hNative;
          if (typeof canvas.style != 'undefined') {
            if (w != wNative || h != hNative) {
              canvas.style.setProperty( "width", w + "px", "important");
              canvas.style.setProperty("height", h + "px", "important");
            } else {
              canvas.style.removeProperty( "width");
              canvas.style.removeProperty("height");
            }
          }
        }
      },wgetRequests:{},nextWgetRequestHandle:0,getNextWgetRequestHandle:function () {
        var handle = Browser.nextWgetRequestHandle;
        Browser.nextWgetRequestHandle++;
        return handle;
      }};var EGL={errorCode:12288,defaultDisplayInitialized:false,currentContext:0,currentReadSurface:0,currentDrawSurface:0,stringCache:{},setErrorCode:function (code) {
        EGL.errorCode = code;
      },chooseConfig:function (display, attribList, config, config_size, numConfigs) { 
        if (display != 62000 /* Magic ID for Emscripten 'default display' */) {
          EGL.setErrorCode(0x3008 /* EGL_BAD_DISPLAY */);
          return 0;
        }
        // TODO: read attribList.
        if ((!config || !config_size) && !numConfigs) {
          EGL.setErrorCode(0x300C /* EGL_BAD_PARAMETER */);
          return 0;
        }
        if (numConfigs) {
          HEAP32[((numConfigs)>>2)]=1; // Total number of supported configs: 1.
        }
        if (config && config_size > 0) {
          HEAP32[((config)>>2)]=62002; 
        }
        
        EGL.setErrorCode(0x3000 /* EGL_SUCCESS */);
        return 1;
      }};
  
  function _eglWaitClient() {
      EGL.setErrorCode(0x3000 /* EGL_SUCCESS */);
      return 1;
    }function _eglTerminate(display) {
      if (display != 62000 /* Magic ID for Emscripten 'default display' */) {
        EGL.setErrorCode(0x3008 /* EGL_BAD_DISPLAY */);
        return 0;
      }
      EGL.currentContext = 0;
      EGL.currentReadSurface = 0;
      EGL.currentDrawSurface = 0;
      EGL.defaultDisplayInitialized = false;
      EGL.setErrorCode(0x3000 /* EGL_SUCCESS */);
      return 1;
    }

  function _emscripten_glStencilMask(x0) { GLctx.stencilMask(x0) }

  
   
  Module["_llvm_bswap_i32"] = _llvm_bswap_i32;function _llvm_bswap_i64(l, h) {
      var retl = _llvm_bswap_i32(h)>>>0;
      var reth = _llvm_bswap_i32(l)>>>0;
      return ((asm["setTempRet0"](reth),retl)|0);
    }

  function _pthread_mutex_lock() {}

  
  
  function _free() {
  }
  Module["_free"] = _free;function ___cxa_free_exception(ptr) {
      try {
        return _free(ptr);
      } catch(e) { // XXX FIXME
        Module.printErr('exception during cxa_free_exception: ' + e);
      }
    }
  
  var EXCEPTIONS={last:0,caught:[],infos:{},deAdjust:function (adjusted) {
        if (!adjusted || EXCEPTIONS.infos[adjusted]) return adjusted;
        for (var ptr in EXCEPTIONS.infos) {
          var info = EXCEPTIONS.infos[ptr];
          if (info.adjusted === adjusted) {
            return ptr;
          }
        }
        return adjusted;
      },addRef:function (ptr) {
        if (!ptr) return;
        var info = EXCEPTIONS.infos[ptr];
        info.refcount++;
      },decRef:function (ptr) {
        if (!ptr) return;
        var info = EXCEPTIONS.infos[ptr];
        assert(info.refcount > 0);
        info.refcount--;
        if (info.refcount === 0) {
          if (info.destructor) {
            Runtime.dynCall('vi', info.destructor, [ptr]);
          }
          delete EXCEPTIONS.infos[ptr];
          ___cxa_free_exception(ptr);
        }
      },clearRef:function (ptr) {
        if (!ptr) return;
        var info = EXCEPTIONS.infos[ptr];
        info.refcount = 0;
      }};function ___cxa_end_catch() {
      if (___cxa_end_catch.rethrown) {
        ___cxa_end_catch.rethrown = false;
        return;
      }
      // Clear state flag.
      asm['setThrew'](0);
      // Call destructor if one is registered then clear it.
      var ptr = EXCEPTIONS.caught.pop();
      if (ptr) {
        EXCEPTIONS.decRef(EXCEPTIONS.deAdjust(ptr));
        EXCEPTIONS.last = 0; // XXX in decRef?
      }
    }

  function _emscripten_set_mouseleave_callback(target, userData, useCapture, callbackfunc) {
      JSEvents.registerMouseEventCallback(target, userData, useCapture, callbackfunc, 34, "mouseleave");
      return 0;
    }

  function _emscripten_glStencilFunc(x0, x1, x2) { GLctx.stencilFunc(x0, x1, x2) }

  
  var __currentFullscreenStrategy={};function _emscripten_exit_fullscreen() {
      if (typeof JSEvents.fullscreenEnabled() === 'undefined') return -1;
      // Make sure no queued up calls will fire after this.
      JSEvents.removeDeferredCalls(JSEvents.requestFullscreen);
  
      if (document.exitFullscreen) {
        document.exitFullscreen();
      } else if (document.msExitFullscreen) {
        document.msExitFullscreen();
      } else if (document.mozCancelFullScreen) {
        document.mozCancelFullScreen();
      } else if (document.webkitExitFullscreen) {
        document.webkitExitFullscreen();
      } else {
        return -1;
      }
  
      if (__currentFullscreenStrategy.canvasResizedCallback) {
        Runtime.dynCall('iiii', __currentFullscreenStrategy.canvasResizedCallback, [37, 0, __currentFullscreenStrategy.canvasResizedCallbackUserData]);
      }
  
      return 0;
    }

  
  
  
  var ERRNO_CODES={EPERM:1,ENOENT:2,ESRCH:3,EINTR:4,EIO:5,ENXIO:6,E2BIG:7,ENOEXEC:8,EBADF:9,ECHILD:10,EAGAIN:11,EWOULDBLOCK:11,ENOMEM:12,EACCES:13,EFAULT:14,ENOTBLK:15,EBUSY:16,EEXIST:17,EXDEV:18,ENODEV:19,ENOTDIR:20,EISDIR:21,EINVAL:22,ENFILE:23,EMFILE:24,ENOTTY:25,ETXTBSY:26,EFBIG:27,ENOSPC:28,ESPIPE:29,EROFS:30,EMLINK:31,EPIPE:32,EDOM:33,ERANGE:34,ENOMSG:42,EIDRM:43,ECHRNG:44,EL2NSYNC:45,EL3HLT:46,EL3RST:47,ELNRNG:48,EUNATCH:49,ENOCSI:50,EL2HLT:51,EDEADLK:35,ENOLCK:37,EBADE:52,EBADR:53,EXFULL:54,ENOANO:55,EBADRQC:56,EBADSLT:57,EDEADLOCK:35,EBFONT:59,ENOSTR:60,ENODATA:61,ETIME:62,ENOSR:63,ENONET:64,ENOPKG:65,EREMOTE:66,ENOLINK:67,EADV:68,ESRMNT:69,ECOMM:70,EPROTO:71,EMULTIHOP:72,EDOTDOT:73,EBADMSG:74,ENOTUNIQ:76,EBADFD:77,EREMCHG:78,ELIBACC:79,ELIBBAD:80,ELIBSCN:81,ELIBMAX:82,ELIBEXEC:83,ENOSYS:38,ENOTEMPTY:39,ENAMETOOLONG:36,ELOOP:40,EOPNOTSUPP:95,EPFNOSUPPORT:96,ECONNRESET:104,ENOBUFS:105,EAFNOSUPPORT:97,EPROTOTYPE:91,ENOTSOCK:88,ENOPROTOOPT:92,ESHUTDOWN:108,ECONNREFUSED:111,EADDRINUSE:98,ECONNABORTED:103,ENETUNREACH:101,ENETDOWN:100,ETIMEDOUT:110,EHOSTDOWN:112,EHOSTUNREACH:113,EINPROGRESS:115,EALREADY:114,EDESTADDRREQ:89,EMSGSIZE:90,EPROTONOSUPPORT:93,ESOCKTNOSUPPORT:94,EADDRNOTAVAIL:99,ENETRESET:102,EISCONN:106,ENOTCONN:107,ETOOMANYREFS:109,EUSERS:87,EDQUOT:122,ESTALE:116,ENOTSUP:95,ENOMEDIUM:123,EILSEQ:84,EOVERFLOW:75,ECANCELED:125,ENOTRECOVERABLE:131,EOWNERDEAD:130,ESTRPIPE:86};
  
  var ERRNO_MESSAGES={0:"Success",1:"Not super-user",2:"No such file or directory",3:"No such process",4:"Interrupted system call",5:"I/O error",6:"No such device or address",7:"Arg list too long",8:"Exec format error",9:"Bad file number",10:"No children",11:"No more processes",12:"Not enough core",13:"Permission denied",14:"Bad address",15:"Block device required",16:"Mount device busy",17:"File exists",18:"Cross-device link",19:"No such device",20:"Not a directory",21:"Is a directory",22:"Invalid argument",23:"Too many open files in system",24:"Too many open files",25:"Not a typewriter",26:"Text file busy",27:"File too large",28:"No space left on device",29:"Illegal seek",30:"Read only file system",31:"Too many links",32:"Broken pipe",33:"Math arg out of domain of func",34:"Math result not representable",35:"File locking deadlock error",36:"File or path name too long",37:"No record locks available",38:"Function not implemented",39:"Directory not empty",40:"Too many symbolic links",42:"No message of desired type",43:"Identifier removed",44:"Channel number out of range",45:"Level 2 not synchronized",46:"Level 3 halted",47:"Level 3 reset",48:"Link number out of range",49:"Protocol driver not attached",50:"No CSI structure available",51:"Level 2 halted",52:"Invalid exchange",53:"Invalid request descriptor",54:"Exchange full",55:"No anode",56:"Invalid request code",57:"Invalid slot",59:"Bad font file fmt",60:"Device not a stream",61:"No data (for no delay io)",62:"Timer expired",63:"Out of streams resources",64:"Machine is not on the network",65:"Package not installed",66:"The object is remote",67:"The link has been severed",68:"Advertise error",69:"Srmount error",70:"Communication error on send",71:"Protocol error",72:"Multihop attempted",73:"Cross mount point (not really error)",74:"Trying to read unreadable message",75:"Value too large for defined data type",76:"Given log. name not unique",77:"f.d. invalid for this operation",78:"Remote address changed",79:"Can   access a needed shared lib",80:"Accessing a corrupted shared lib",81:".lib section in a.out corrupted",82:"Attempting to link in too many libs",83:"Attempting to exec a shared library",84:"Illegal byte sequence",86:"Streams pipe error",87:"Too many users",88:"Socket operation on non-socket",89:"Destination address required",90:"Message too long",91:"Protocol wrong type for socket",92:"Protocol not available",93:"Unknown protocol",94:"Socket type not supported",95:"Not supported",96:"Protocol family not supported",97:"Address family not supported by protocol family",98:"Address already in use",99:"Address not available",100:"Network interface is not configured",101:"Network is unreachable",102:"Connection reset by network",103:"Connection aborted",104:"Connection reset by peer",105:"No buffer space available",106:"Socket is already connected",107:"Socket is not connected",108:"Can't send after socket shutdown",109:"Too many references",110:"Connection timed out",111:"Connection refused",112:"Host is down",113:"Host is unreachable",114:"Socket already connected",115:"Connection already in progress",116:"Stale file handle",122:"Quota exceeded",123:"No medium (in tape drive)",125:"Operation canceled",130:"Previous owner died",131:"State not recoverable"};
  
  function ___setErrNo(value) {
      if (Module['___errno_location']) HEAP32[((Module['___errno_location']())>>2)]=value;
      else Module.printErr('failed to set errno from JS');
      return value;
    }
  
  var PATH={splitPath:function (filename) {
        var splitPathRe = /^(\/?|)([\s\S]*?)((?:\.{1,2}|[^\/]+?|)(\.[^.\/]*|))(?:[\/]*)$/;
        return splitPathRe.exec(filename).slice(1);
      },normalizeArray:function (parts, allowAboveRoot) {
        // if the path tries to go above the root, `up` ends up > 0
        var up = 0;
        for (var i = parts.length - 1; i >= 0; i--) {
          var last = parts[i];
          if (last === '.') {
            parts.splice(i, 1);
          } else if (last === '..') {
            parts.splice(i, 1);
            up++;
          } else if (up) {
            parts.splice(i, 1);
            up--;
          }
        }
        // if the path is allowed to go above the root, restore leading ..s
        if (allowAboveRoot) {
          for (; up--; up) {
            parts.unshift('..');
          }
        }
        return parts;
      },normalize:function (path) {
        var isAbsolute = path.charAt(0) === '/',
            trailingSlash = path.substr(-1) === '/';
        // Normalize the path
        path = PATH.normalizeArray(path.split('/').filter(function(p) {
          return !!p;
        }), !isAbsolute).join('/');
        if (!path && !isAbsolute) {
          path = '.';
        }
        if (path && trailingSlash) {
          path += '/';
        }
        return (isAbsolute ? '/' : '') + path;
      },dirname:function (path) {
        var result = PATH.splitPath(path),
            root = result[0],
            dir = result[1];
        if (!root && !dir) {
          // No dirname whatsoever
          return '.';
        }
        if (dir) {
          // It has a dirname, strip trailing slash
          dir = dir.substr(0, dir.length - 1);
        }
        return root + dir;
      },basename:function (path) {
        // EMSCRIPTEN return '/'' for '/', not an empty string
        if (path === '/') return '/';
        var lastSlash = path.lastIndexOf('/');
        if (lastSlash === -1) return path;
        return path.substr(lastSlash+1);
      },extname:function (path) {
        return PATH.splitPath(path)[3];
      },join:function () {
        var paths = Array.prototype.slice.call(arguments, 0);
        return PATH.normalize(paths.join('/'));
      },join2:function (l, r) {
        return PATH.normalize(l + '/' + r);
      },resolve:function () {
        var resolvedPath = '',
          resolvedAbsolute = false;
        for (var i = arguments.length - 1; i >= -1 && !resolvedAbsolute; i--) {
          var path = (i >= 0) ? arguments[i] : FS.cwd();
          // Skip empty and invalid entries
          if (typeof path !== 'string') {
            throw new TypeError('Arguments to path.resolve must be strings');
          } else if (!path) {
            return ''; // an invalid portion invalidates the whole thing
          }
          resolvedPath = path + '/' + resolvedPath;
          resolvedAbsolute = path.charAt(0) === '/';
        }
        // At this point the path should be resolved to a full absolute path, but
        // handle relative paths to be safe (might happen when process.cwd() fails)
        resolvedPath = PATH.normalizeArray(resolvedPath.split('/').filter(function(p) {
          return !!p;
        }), !resolvedAbsolute).join('/');
        return ((resolvedAbsolute ? '/' : '') + resolvedPath) || '.';
      },relative:function (from, to) {
        from = PATH.resolve(from).substr(1);
        to = PATH.resolve(to).substr(1);
        function trim(arr) {
          var start = 0;
          for (; start < arr.length; start++) {
            if (arr[start] !== '') break;
          }
          var end = arr.length - 1;
          for (; end >= 0; end--) {
            if (arr[end] !== '') break;
          }
          if (start > end) return [];
          return arr.slice(start, end - start + 1);
        }
        var fromParts = trim(from.split('/'));
        var toParts = trim(to.split('/'));
        var length = Math.min(fromParts.length, toParts.length);
        var samePartsLength = length;
        for (var i = 0; i < length; i++) {
          if (fromParts[i] !== toParts[i]) {
            samePartsLength = i;
            break;
          }
        }
        var outputParts = [];
        for (var i = samePartsLength; i < fromParts.length; i++) {
          outputParts.push('..');
        }
        outputParts = outputParts.concat(toParts.slice(samePartsLength));
        return outputParts.join('/');
      }};
  
  var TTY={ttys:[],init:function () {
        // https://github.com/kripken/emscripten/pull/1555
        // if (ENVIRONMENT_IS_NODE) {
        //   // currently, FS.init does not distinguish if process.stdin is a file or TTY
        //   // device, it always assumes it's a TTY device. because of this, we're forcing
        //   // process.stdin to UTF8 encoding to at least make stdin reading compatible
        //   // with text files until FS.init can be refactored.
        //   process['stdin']['setEncoding']('utf8');
        // }
      },shutdown:function () {
        // https://github.com/kripken/emscripten/pull/1555
        // if (ENVIRONMENT_IS_NODE) {
        //   // inolen: any idea as to why node -e 'process.stdin.read()' wouldn't exit immediately (with process.stdin being a tty)?
        //   // isaacs: because now it's reading from the stream, you've expressed interest in it, so that read() kicks off a _read() which creates a ReadReq operation
        //   // inolen: I thought read() in that case was a synchronous operation that just grabbed some amount of buffered data if it exists?
        //   // isaacs: it is. but it also triggers a _read() call, which calls readStart() on the handle
        //   // isaacs: do process.stdin.pause() and i'd think it'd probably close the pending call
        //   process['stdin']['pause']();
        // }
      },register:function (dev, ops) {
        TTY.ttys[dev] = { input: [], output: [], ops: ops };
        FS.registerDevice(dev, TTY.stream_ops);
      },stream_ops:{open:function (stream) {
          var tty = TTY.ttys[stream.node.rdev];
          if (!tty) {
            throw new FS.ErrnoError(ERRNO_CODES.ENODEV);
          }
          stream.tty = tty;
          stream.seekable = false;
        },close:function (stream) {
          // flush any pending line data
          stream.tty.ops.flush(stream.tty);
        },flush:function (stream) {
          stream.tty.ops.flush(stream.tty);
        },read:function (stream, buffer, offset, length, pos /* ignored */) {
          if (!stream.tty || !stream.tty.ops.get_char) {
            throw new FS.ErrnoError(ERRNO_CODES.ENXIO);
          }
          var bytesRead = 0;
          for (var i = 0; i < length; i++) {
            var result;
            try {
              result = stream.tty.ops.get_char(stream.tty);
            } catch (e) {
              throw new FS.ErrnoError(ERRNO_CODES.EIO);
            }
            if (result === undefined && bytesRead === 0) {
              throw new FS.ErrnoError(ERRNO_CODES.EAGAIN);
            }
            if (result === null || result === undefined) break;
            bytesRead++;
            buffer[offset+i] = result;
          }
          if (bytesRead) {
            stream.node.timestamp = Date.now();
          }
          return bytesRead;
        },write:function (stream, buffer, offset, length, pos) {
          if (!stream.tty || !stream.tty.ops.put_char) {
            throw new FS.ErrnoError(ERRNO_CODES.ENXIO);
          }
          for (var i = 0; i < length; i++) {
            try {
              stream.tty.ops.put_char(stream.tty, buffer[offset+i]);
            } catch (e) {
              throw new FS.ErrnoError(ERRNO_CODES.EIO);
            }
          }
          if (length) {
            stream.node.timestamp = Date.now();
          }
          return i;
        }},default_tty_ops:{get_char:function (tty) {
          if (!tty.input.length) {
            var result = null;
            if (ENVIRONMENT_IS_NODE) {
              // we will read data by chunks of BUFSIZE
              var BUFSIZE = 256;
              var buf = new Buffer(BUFSIZE);
              var bytesRead = 0;
  
              var fd = process.stdin.fd;
              // Linux and Mac cannot use process.stdin.fd (which isn't set up as sync)
              var usingDevice = false;
              try {
                fd = fs.openSync('/dev/stdin', 'r');
                usingDevice = true;
              } catch (e) {}
  
              bytesRead = fs.readSync(fd, buf, 0, BUFSIZE, null);
  
              if (usingDevice) { fs.closeSync(fd); }
              if (bytesRead > 0) {
                result = buf.slice(0, bytesRead).toString('utf-8');
              } else {
                result = null;
              }
  
            } else if (typeof window != 'undefined' &&
              typeof window.prompt == 'function') {
              // Browser.
              result = window.prompt('Input: ');  // returns null on cancel
              if (result !== null) {
                result += '\n';
              }
            } else if (typeof readline == 'function') {
              // Command line.
              result = readline();
              if (result !== null) {
                result += '\n';
              }
            }
            if (!result) {
              return null;
            }
            tty.input = intArrayFromString(result, true);
          }
          return tty.input.shift();
        },put_char:function (tty, val) {
          if (val === null || val === 10) {
            Module['print'](UTF8ArrayToString(tty.output, 0));
            tty.output = [];
          } else {
            if (val != 0) tty.output.push(val); // val == 0 would cut text output off in the middle.
          }
        },flush:function (tty) {
          if (tty.output && tty.output.length > 0) {
            Module['print'](UTF8ArrayToString(tty.output, 0));
            tty.output = [];
          }
        }},default_tty1_ops:{put_char:function (tty, val) {
          if (val === null || val === 10) {
            Module['printErr'](UTF8ArrayToString(tty.output, 0));
            tty.output = [];
          } else {
            if (val != 0) tty.output.push(val);
          }
        },flush:function (tty) {
          if (tty.output && tty.output.length > 0) {
            Module['printErr'](UTF8ArrayToString(tty.output, 0));
            tty.output = [];
          }
        }}};
  
  var MEMFS={ops_table:null,mount:function (mount) {
        return MEMFS.createNode(null, '/', 16384 | 511 /* 0777 */, 0);
      },createNode:function (parent, name, mode, dev) {
        if (FS.isBlkdev(mode) || FS.isFIFO(mode)) {
          // no supported
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        if (!MEMFS.ops_table) {
          MEMFS.ops_table = {
            dir: {
              node: {
                getattr: MEMFS.node_ops.getattr,
                setattr: MEMFS.node_ops.setattr,
                lookup: MEMFS.node_ops.lookup,
                mknod: MEMFS.node_ops.mknod,
                rename: MEMFS.node_ops.rename,
                unlink: MEMFS.node_ops.unlink,
                rmdir: MEMFS.node_ops.rmdir,
                readdir: MEMFS.node_ops.readdir,
                symlink: MEMFS.node_ops.symlink
              },
              stream: {
                llseek: MEMFS.stream_ops.llseek
              }
            },
            file: {
              node: {
                getattr: MEMFS.node_ops.getattr,
                setattr: MEMFS.node_ops.setattr
              },
              stream: {
                llseek: MEMFS.stream_ops.llseek,
                read: MEMFS.stream_ops.read,
                write: MEMFS.stream_ops.write,
                allocate: MEMFS.stream_ops.allocate,
                mmap: MEMFS.stream_ops.mmap,
                msync: MEMFS.stream_ops.msync
              }
            },
            link: {
              node: {
                getattr: MEMFS.node_ops.getattr,
                setattr: MEMFS.node_ops.setattr,
                readlink: MEMFS.node_ops.readlink
              },
              stream: {}
            },
            chrdev: {
              node: {
                getattr: MEMFS.node_ops.getattr,
                setattr: MEMFS.node_ops.setattr
              },
              stream: FS.chrdev_stream_ops
            }
          };
        }
        var node = FS.createNode(parent, name, mode, dev);
        if (FS.isDir(node.mode)) {
          node.node_ops = MEMFS.ops_table.dir.node;
          node.stream_ops = MEMFS.ops_table.dir.stream;
          node.contents = {};
        } else if (FS.isFile(node.mode)) {
          node.node_ops = MEMFS.ops_table.file.node;
          node.stream_ops = MEMFS.ops_table.file.stream;
          node.usedBytes = 0; // The actual number of bytes used in the typed array, as opposed to contents.buffer.byteLength which gives the whole capacity.
          // When the byte data of the file is populated, this will point to either a typed array, or a normal JS array. Typed arrays are preferred
          // for performance, and used by default. However, typed arrays are not resizable like normal JS arrays are, so there is a small disk size
          // penalty involved for appending file writes that continuously grow a file similar to std::vector capacity vs used -scheme.
          node.contents = null; 
        } else if (FS.isLink(node.mode)) {
          node.node_ops = MEMFS.ops_table.link.node;
          node.stream_ops = MEMFS.ops_table.link.stream;
        } else if (FS.isChrdev(node.mode)) {
          node.node_ops = MEMFS.ops_table.chrdev.node;
          node.stream_ops = MEMFS.ops_table.chrdev.stream;
        }
        node.timestamp = Date.now();
        // add the new node to the parent
        if (parent) {
          parent.contents[name] = node;
        }
        return node;
      },getFileDataAsRegularArray:function (node) {
        if (node.contents && node.contents.subarray) {
          var arr = [];
          for (var i = 0; i < node.usedBytes; ++i) arr.push(node.contents[i]);
          return arr; // Returns a copy of the original data.
        }
        return node.contents; // No-op, the file contents are already in a JS array. Return as-is.
      },getFileDataAsTypedArray:function (node) {
        if (!node.contents) return new Uint8Array;
        if (node.contents.subarray) return node.contents.subarray(0, node.usedBytes); // Make sure to not return excess unused bytes.
        return new Uint8Array(node.contents);
      },expandFileStorage:function (node, newCapacity) {
        // If we are asked to expand the size of a file that already exists, revert to using a standard JS array to store the file
        // instead of a typed array. This makes resizing the array more flexible because we can just .push() elements at the back to
        // increase the size.
        if (node.contents && node.contents.subarray && newCapacity > node.contents.length) {
          node.contents = MEMFS.getFileDataAsRegularArray(node);
          node.usedBytes = node.contents.length; // We might be writing to a lazy-loaded file which had overridden this property, so force-reset it.
        }
  
        if (!node.contents || node.contents.subarray) { // Keep using a typed array if creating a new storage, or if old one was a typed array as well.
          var prevCapacity = node.contents ? node.contents.buffer.byteLength : 0;
          if (prevCapacity >= newCapacity) return; // No need to expand, the storage was already large enough.
          // Don't expand strictly to the given requested limit if it's only a very small increase, but instead geometrically grow capacity.
          // For small filesizes (<1MB), perform size*2 geometric increase, but for large sizes, do a much more conservative size*1.125 increase to
          // avoid overshooting the allocation cap by a very large margin.
          var CAPACITY_DOUBLING_MAX = 1024 * 1024;
          newCapacity = Math.max(newCapacity, (prevCapacity * (prevCapacity < CAPACITY_DOUBLING_MAX ? 2.0 : 1.125)) | 0);
          if (prevCapacity != 0) newCapacity = Math.max(newCapacity, 256); // At minimum allocate 256b for each file when expanding.
          var oldContents = node.contents;
          node.contents = new Uint8Array(newCapacity); // Allocate new storage.
          if (node.usedBytes > 0) node.contents.set(oldContents.subarray(0, node.usedBytes), 0); // Copy old data over to the new storage.
          return;
        }
        // Not using a typed array to back the file storage. Use a standard JS array instead.
        if (!node.contents && newCapacity > 0) node.contents = [];
        while (node.contents.length < newCapacity) node.contents.push(0);
      },resizeFileStorage:function (node, newSize) {
        if (node.usedBytes == newSize) return;
        if (newSize == 0) {
          node.contents = null; // Fully decommit when requesting a resize to zero.
          node.usedBytes = 0;
          return;
        }
        if (!node.contents || node.contents.subarray) { // Resize a typed array if that is being used as the backing store.
          var oldContents = node.contents;
          node.contents = new Uint8Array(new ArrayBuffer(newSize)); // Allocate new storage.
          if (oldContents) {
            node.contents.set(oldContents.subarray(0, Math.min(newSize, node.usedBytes))); // Copy old data over to the new storage.
          }
          node.usedBytes = newSize;
          return;
        }
        // Backing with a JS array.
        if (!node.contents) node.contents = [];
        if (node.contents.length > newSize) node.contents.length = newSize;
        else while (node.contents.length < newSize) node.contents.push(0);
        node.usedBytes = newSize;
      },node_ops:{getattr:function (node) {
          var attr = {};
          // device numbers reuse inode numbers.
          attr.dev = FS.isChrdev(node.mode) ? node.id : 1;
          attr.ino = node.id;
          attr.mode = node.mode;
          attr.nlink = 1;
          attr.uid = 0;
          attr.gid = 0;
          attr.rdev = node.rdev;
          if (FS.isDir(node.mode)) {
            attr.size = 4096;
          } else if (FS.isFile(node.mode)) {
            attr.size = node.usedBytes;
          } else if (FS.isLink(node.mode)) {
            attr.size = node.link.length;
          } else {
            attr.size = 0;
          }
          attr.atime = new Date(node.timestamp);
          attr.mtime = new Date(node.timestamp);
          attr.ctime = new Date(node.timestamp);
          // NOTE: In our implementation, st_blocks = Math.ceil(st_size/st_blksize),
          //       but this is not required by the standard.
          attr.blksize = 4096;
          attr.blocks = Math.ceil(attr.size / attr.blksize);
          return attr;
        },setattr:function (node, attr) {
          if (attr.mode !== undefined) {
            node.mode = attr.mode;
          }
          if (attr.timestamp !== undefined) {
            node.timestamp = attr.timestamp;
          }
          if (attr.size !== undefined) {
            MEMFS.resizeFileStorage(node, attr.size);
          }
        },lookup:function (parent, name) {
          throw FS.genericErrors[ERRNO_CODES.ENOENT];
        },mknod:function (parent, name, mode, dev) {
          return MEMFS.createNode(parent, name, mode, dev);
        },rename:function (old_node, new_dir, new_name) {
          // if we're overwriting a directory at new_name, make sure it's empty.
          if (FS.isDir(old_node.mode)) {
            var new_node;
            try {
              new_node = FS.lookupNode(new_dir, new_name);
            } catch (e) {
            }
            if (new_node) {
              for (var i in new_node.contents) {
                throw new FS.ErrnoError(ERRNO_CODES.ENOTEMPTY);
              }
            }
          }
          // do the internal rewiring
          delete old_node.parent.contents[old_node.name];
          old_node.name = new_name;
          new_dir.contents[new_name] = old_node;
          old_node.parent = new_dir;
        },unlink:function (parent, name) {
          delete parent.contents[name];
        },rmdir:function (parent, name) {
          var node = FS.lookupNode(parent, name);
          for (var i in node.contents) {
            throw new FS.ErrnoError(ERRNO_CODES.ENOTEMPTY);
          }
          delete parent.contents[name];
        },readdir:function (node) {
          var entries = ['.', '..']
          for (var key in node.contents) {
            if (!node.contents.hasOwnProperty(key)) {
              continue;
            }
            entries.push(key);
          }
          return entries;
        },symlink:function (parent, newname, oldpath) {
          var node = MEMFS.createNode(parent, newname, 511 /* 0777 */ | 40960, 0);
          node.link = oldpath;
          return node;
        },readlink:function (node) {
          if (!FS.isLink(node.mode)) {
            throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
          }
          return node.link;
        }},stream_ops:{read:function (stream, buffer, offset, length, position) {
          var contents = stream.node.contents;
          if (position >= stream.node.usedBytes) return 0;
          var size = Math.min(stream.node.usedBytes - position, length);
          assert(size >= 0);
          if (size > 8 && contents.subarray) { // non-trivial, and typed array
            buffer.set(contents.subarray(position, position + size), offset);
          } else {
            for (var i = 0; i < size; i++) buffer[offset + i] = contents[position + i];
          }
          return size;
        },write:function (stream, buffer, offset, length, position, canOwn) {
          if (!length) return 0;
          var node = stream.node;
          node.timestamp = Date.now();
  
          if (buffer.subarray && (!node.contents || node.contents.subarray)) { // This write is from a typed array to a typed array?
            if (canOwn) { // Can we just reuse the buffer we are given?
              assert(position === 0, 'canOwn must imply no weird position inside the file');
              node.contents = buffer.subarray(offset, offset + length);
              node.usedBytes = length;
              return length;
            } else if (node.usedBytes === 0 && position === 0) { // If this is a simple first write to an empty file, do a fast set since we don't need to care about old data.
              node.contents = new Uint8Array(buffer.subarray(offset, offset + length));
              node.usedBytes = length;
              return length;
            } else if (position + length <= node.usedBytes) { // Writing to an already allocated and used subrange of the file?
              node.contents.set(buffer.subarray(offset, offset + length), position);
              return length;
            }
          }
  
          // Appending to an existing file and we need to reallocate, or source data did not come as a typed array.
          MEMFS.expandFileStorage(node, position+length);
          if (node.contents.subarray && buffer.subarray) node.contents.set(buffer.subarray(offset, offset + length), position); // Use typed array write if available.
          else {
            for (var i = 0; i < length; i++) {
             node.contents[position + i] = buffer[offset + i]; // Or fall back to manual write if not.
            }
          }
          node.usedBytes = Math.max(node.usedBytes, position+length);
          return length;
        },llseek:function (stream, offset, whence) {
          var position = offset;
          if (whence === 1) {  // SEEK_CUR.
            position += stream.position;
          } else if (whence === 2) {  // SEEK_END.
            if (FS.isFile(stream.node.mode)) {
              position += stream.node.usedBytes;
            }
          }
          if (position < 0) {
            throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
          }
          return position;
        },allocate:function (stream, offset, length) {
          MEMFS.expandFileStorage(stream.node, offset + length);
          stream.node.usedBytes = Math.max(stream.node.usedBytes, offset + length);
        },mmap:function (stream, buffer, offset, length, position, prot, flags) {
          if (!FS.isFile(stream.node.mode)) {
            throw new FS.ErrnoError(ERRNO_CODES.ENODEV);
          }
          var ptr;
          var allocated;
          var contents = stream.node.contents;
          // Only make a new copy when MAP_PRIVATE is specified.
          if ( !(flags & 2) &&
                (contents.buffer === buffer || contents.buffer === buffer.buffer) ) {
            // We can't emulate MAP_SHARED when the file is not backed by the buffer
            // we're mapping to (e.g. the HEAP buffer).
            allocated = false;
            ptr = contents.byteOffset;
          } else {
            // Try to avoid unnecessary slices.
            if (position > 0 || position + length < stream.node.usedBytes) {
              if (contents.subarray) {
                contents = contents.subarray(position, position + length);
              } else {
                contents = Array.prototype.slice.call(contents, position, position + length);
              }
            }
            allocated = true;
            ptr = _malloc(length);
            if (!ptr) {
              throw new FS.ErrnoError(ERRNO_CODES.ENOMEM);
            }
            buffer.set(contents, ptr);
          }
          return { ptr: ptr, allocated: allocated };
        },msync:function (stream, buffer, offset, length, mmapFlags) {
          if (!FS.isFile(stream.node.mode)) {
            throw new FS.ErrnoError(ERRNO_CODES.ENODEV);
          }
          if (mmapFlags & 2) {
            // MAP_PRIVATE calls need not to be synced back to underlying fs
            return 0;
          }
  
          var bytesWritten = MEMFS.stream_ops.write(stream, buffer, 0, length, offset, false);
          // should we check if bytesWritten and length are the same?
          return 0;
        }}};
  
  var IDBFS={dbs:{},indexedDB:function () {
        if (typeof indexedDB !== 'undefined') return indexedDB;
        var ret = null;
        if (typeof window === 'object') ret = window.indexedDB || window.mozIndexedDB || window.webkitIndexedDB || window.msIndexedDB;
        assert(ret, 'IDBFS used, but indexedDB not supported');
        return ret;
      },DB_VERSION:21,DB_STORE_NAME:"FILE_DATA",mount:function (mount) {
        // reuse all of the core MEMFS functionality
        return MEMFS.mount.apply(null, arguments);
      },syncfs:function (mount, populate, callback) {
        IDBFS.getLocalSet(mount, function(err, local) {
          if (err) return callback(err);
  
          IDBFS.getRemoteSet(mount, function(err, remote) {
            if (err) return callback(err);
  
            var src = populate ? remote : local;
            var dst = populate ? local : remote;
  
            IDBFS.reconcile(src, dst, callback);
          });
        });
      },getDB:function (name, callback) {
        // check the cache first
        var db = IDBFS.dbs[name];
        if (db) {
          return callback(null, db);
        }
  
        var req;
        try {
          req = IDBFS.indexedDB().open(name, IDBFS.DB_VERSION);
        } catch (e) {
          return callback(e);
        }
        req.onupgradeneeded = function(e) {
          var db = e.target.result;
          var transaction = e.target.transaction;
  
          var fileStore;
  
          if (db.objectStoreNames.contains(IDBFS.DB_STORE_NAME)) {
            fileStore = transaction.objectStore(IDBFS.DB_STORE_NAME);
          } else {
            fileStore = db.createObjectStore(IDBFS.DB_STORE_NAME);
          }
  
          if (!fileStore.indexNames.contains('timestamp')) {
            fileStore.createIndex('timestamp', 'timestamp', { unique: false });
          }
        };
        req.onsuccess = function() {
          db = req.result;
  
          // add to the cache
          IDBFS.dbs[name] = db;
          callback(null, db);
        };
        req.onerror = function(e) {
          callback(this.error);
          e.preventDefault();
        };
      },getLocalSet:function (mount, callback) {
        var entries = {};
  
        function isRealDir(p) {
          return p !== '.' && p !== '..';
        };
        function toAbsolute(root) {
          return function(p) {
            return PATH.join2(root, p);
          }
        };
  
        var check = FS.readdir(mount.mountpoint).filter(isRealDir).map(toAbsolute(mount.mountpoint));
  
        while (check.length) {
          var path = check.pop();
          var stat;
  
          try {
            stat = FS.stat(path);
          } catch (e) {
            return callback(e);
          }
  
          if (FS.isDir(stat.mode)) {
            check.push.apply(check, FS.readdir(path).filter(isRealDir).map(toAbsolute(path)));
          }
  
          entries[path] = { timestamp: stat.mtime };
        }
  
        return callback(null, { type: 'local', entries: entries });
      },getRemoteSet:function (mount, callback) {
        var entries = {};
  
        IDBFS.getDB(mount.mountpoint, function(err, db) {
          if (err) return callback(err);
  
          var transaction = db.transaction([IDBFS.DB_STORE_NAME], 'readonly');
          transaction.onerror = function(e) {
            callback(this.error);
            e.preventDefault();
          };
  
          var store = transaction.objectStore(IDBFS.DB_STORE_NAME);
          var index = store.index('timestamp');
  
          index.openKeyCursor().onsuccess = function(event) {
            var cursor = event.target.result;
  
            if (!cursor) {
              return callback(null, { type: 'remote', db: db, entries: entries });
            }
  
            entries[cursor.primaryKey] = { timestamp: cursor.key };
  
            cursor.continue();
          };
        });
      },loadLocalEntry:function (path, callback) {
        var stat, node;
  
        try {
          var lookup = FS.lookupPath(path);
          node = lookup.node;
          stat = FS.stat(path);
        } catch (e) {
          return callback(e);
        }
  
        if (FS.isDir(stat.mode)) {
          return callback(null, { timestamp: stat.mtime, mode: stat.mode });
        } else if (FS.isFile(stat.mode)) {
          // Performance consideration: storing a normal JavaScript array to a IndexedDB is much slower than storing a typed array.
          // Therefore always convert the file contents to a typed array first before writing the data to IndexedDB.
          node.contents = MEMFS.getFileDataAsTypedArray(node);
          return callback(null, { timestamp: stat.mtime, mode: stat.mode, contents: node.contents });
        } else {
          return callback(new Error('node type not supported'));
        }
      },storeLocalEntry:function (path, entry, callback) {
        try {
          if (FS.isDir(entry.mode)) {
            FS.mkdir(path, entry.mode);
          } else if (FS.isFile(entry.mode)) {
            FS.writeFile(path, entry.contents, { encoding: 'binary', canOwn: true });
          } else {
            return callback(new Error('node type not supported'));
          }
  
          FS.chmod(path, entry.mode);
          FS.utime(path, entry.timestamp, entry.timestamp);
        } catch (e) {
          return callback(e);
        }
  
        callback(null);
      },removeLocalEntry:function (path, callback) {
        try {
          var lookup = FS.lookupPath(path);
          var stat = FS.stat(path);
  
          if (FS.isDir(stat.mode)) {
            FS.rmdir(path);
          } else if (FS.isFile(stat.mode)) {
            FS.unlink(path);
          }
        } catch (e) {
          return callback(e);
        }
  
        callback(null);
      },loadRemoteEntry:function (store, path, callback) {
        var req = store.get(path);
        req.onsuccess = function(event) { callback(null, event.target.result); };
        req.onerror = function(e) {
          callback(this.error);
          e.preventDefault();
        };
      },storeRemoteEntry:function (store, path, entry, callback) {
        var req = store.put(entry, path);
        req.onsuccess = function() { callback(null); };
        req.onerror = function(e) {
          callback(this.error);
          e.preventDefault();
        };
      },removeRemoteEntry:function (store, path, callback) {
        var req = store.delete(path);
        req.onsuccess = function() { callback(null); };
        req.onerror = function(e) {
          callback(this.error);
          e.preventDefault();
        };
      },reconcile:function (src, dst, callback) {
        var total = 0;
  
        var create = [];
        Object.keys(src.entries).forEach(function (key) {
          var e = src.entries[key];
          var e2 = dst.entries[key];
          if (!e2 || e.timestamp > e2.timestamp) {
            create.push(key);
            total++;
          }
        });
  
        var remove = [];
        Object.keys(dst.entries).forEach(function (key) {
          var e = dst.entries[key];
          var e2 = src.entries[key];
          if (!e2) {
            remove.push(key);
            total++;
          }
        });
  
        if (!total) {
          return callback(null);
        }
  
        var errored = false;
        var completed = 0;
        var db = src.type === 'remote' ? src.db : dst.db;
        var transaction = db.transaction([IDBFS.DB_STORE_NAME], 'readwrite');
        var store = transaction.objectStore(IDBFS.DB_STORE_NAME);
  
        function done(err) {
          if (err) {
            if (!done.errored) {
              done.errored = true;
              return callback(err);
            }
            return;
          }
          if (++completed >= total) {
            return callback(null);
          }
        };
  
        transaction.onerror = function(e) {
          done(this.error);
          e.preventDefault();
        };
  
        // sort paths in ascending order so directory entries are created
        // before the files inside them
        create.sort().forEach(function (path) {
          if (dst.type === 'local') {
            IDBFS.loadRemoteEntry(store, path, function (err, entry) {
              if (err) return done(err);
              IDBFS.storeLocalEntry(path, entry, done);
            });
          } else {
            IDBFS.loadLocalEntry(path, function (err, entry) {
              if (err) return done(err);
              IDBFS.storeRemoteEntry(store, path, entry, done);
            });
          }
        });
  
        // sort paths in descending order so files are deleted before their
        // parent directories
        remove.sort().reverse().forEach(function(path) {
          if (dst.type === 'local') {
            IDBFS.removeLocalEntry(path, done);
          } else {
            IDBFS.removeRemoteEntry(store, path, done);
          }
        });
      }};
  
  var NODEFS={isWindows:false,staticInit:function () {
        NODEFS.isWindows = !!process.platform.match(/^win/);
      },mount:function (mount) {
        assert(ENVIRONMENT_IS_NODE);
        return NODEFS.createNode(null, '/', NODEFS.getMode(mount.opts.root), 0);
      },createNode:function (parent, name, mode, dev) {
        if (!FS.isDir(mode) && !FS.isFile(mode) && !FS.isLink(mode)) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        var node = FS.createNode(parent, name, mode);
        node.node_ops = NODEFS.node_ops;
        node.stream_ops = NODEFS.stream_ops;
        return node;
      },getMode:function (path) {
        var stat;
        try {
          stat = fs.lstatSync(path);
          if (NODEFS.isWindows) {
            // On Windows, directories return permission bits 'rw-rw-rw-', even though they have 'rwxrwxrwx', so
            // propagate write bits to execute bits.
            stat.mode = stat.mode | ((stat.mode & 146) >> 1);
          }
        } catch (e) {
          if (!e.code) throw e;
          throw new FS.ErrnoError(ERRNO_CODES[e.code]);
        }
        return stat.mode;
      },realPath:function (node) {
        var parts = [];
        while (node.parent !== node) {
          parts.push(node.name);
          node = node.parent;
        }
        parts.push(node.mount.opts.root);
        parts.reverse();
        return PATH.join.apply(null, parts);
      },flagsToPermissionStringMap:{0:"r",1:"r+",2:"r+",64:"r",65:"r+",66:"r+",129:"rx+",193:"rx+",514:"w+",577:"w",578:"w+",705:"wx",706:"wx+",1024:"a",1025:"a",1026:"a+",1089:"a",1090:"a+",1153:"ax",1154:"ax+",1217:"ax",1218:"ax+",4096:"rs",4098:"rs+"},flagsToPermissionString:function (flags) {
        flags &= ~0100000 /*O_LARGEFILE*/; // Ignore this flag from musl, otherwise node.js fails to open the file.
        flags &= ~02000000 /*O_CLOEXEC*/; // Some applications may pass it; it makes no sense for a single process.
        if (flags in NODEFS.flagsToPermissionStringMap) {
          return NODEFS.flagsToPermissionStringMap[flags];
        } else {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
      },node_ops:{getattr:function (node) {
          var path = NODEFS.realPath(node);
          var stat;
          try {
            stat = fs.lstatSync(path);
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
          // node.js v0.10.20 doesn't report blksize and blocks on Windows. Fake them with default blksize of 4096.
          // See http://support.microsoft.com/kb/140365
          if (NODEFS.isWindows && !stat.blksize) {
            stat.blksize = 4096;
          }
          if (NODEFS.isWindows && !stat.blocks) {
            stat.blocks = (stat.size+stat.blksize-1)/stat.blksize|0;
          }
          return {
            dev: stat.dev,
            ino: stat.ino,
            mode: stat.mode,
            nlink: stat.nlink,
            uid: stat.uid,
            gid: stat.gid,
            rdev: stat.rdev,
            size: stat.size,
            atime: stat.atime,
            mtime: stat.mtime,
            ctime: stat.ctime,
            blksize: stat.blksize,
            blocks: stat.blocks
          };
        },setattr:function (node, attr) {
          var path = NODEFS.realPath(node);
          try {
            if (attr.mode !== undefined) {
              fs.chmodSync(path, attr.mode);
              // update the common node structure mode as well
              node.mode = attr.mode;
            }
            if (attr.timestamp !== undefined) {
              var date = new Date(attr.timestamp);
              fs.utimesSync(path, date, date);
            }
            if (attr.size !== undefined) {
              fs.truncateSync(path, attr.size);
            }
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        },lookup:function (parent, name) {
          var path = PATH.join2(NODEFS.realPath(parent), name);
          var mode = NODEFS.getMode(path);
          return NODEFS.createNode(parent, name, mode);
        },mknod:function (parent, name, mode, dev) {
          var node = NODEFS.createNode(parent, name, mode, dev);
          // create the backing node for this in the fs root as well
          var path = NODEFS.realPath(node);
          try {
            if (FS.isDir(node.mode)) {
              fs.mkdirSync(path, node.mode);
            } else {
              fs.writeFileSync(path, '', { mode: node.mode });
            }
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
          return node;
        },rename:function (oldNode, newDir, newName) {
          var oldPath = NODEFS.realPath(oldNode);
          var newPath = PATH.join2(NODEFS.realPath(newDir), newName);
          try {
            fs.renameSync(oldPath, newPath);
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        },unlink:function (parent, name) {
          var path = PATH.join2(NODEFS.realPath(parent), name);
          try {
            fs.unlinkSync(path);
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        },rmdir:function (parent, name) {
          var path = PATH.join2(NODEFS.realPath(parent), name);
          try {
            fs.rmdirSync(path);
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        },readdir:function (node) {
          var path = NODEFS.realPath(node);
          try {
            return fs.readdirSync(path);
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        },symlink:function (parent, newName, oldPath) {
          var newPath = PATH.join2(NODEFS.realPath(parent), newName);
          try {
            fs.symlinkSync(oldPath, newPath);
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        },readlink:function (node) {
          var path = NODEFS.realPath(node);
          try {
            path = fs.readlinkSync(path);
            path = NODEJS_PATH.relative(NODEJS_PATH.resolve(node.mount.opts.root), path);
            return path;
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        }},stream_ops:{open:function (stream) {
          var path = NODEFS.realPath(stream.node);
          try {
            if (FS.isFile(stream.node.mode)) {
              stream.nfd = fs.openSync(path, NODEFS.flagsToPermissionString(stream.flags));
            }
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        },close:function (stream) {
          try {
            if (FS.isFile(stream.node.mode) && stream.nfd) {
              fs.closeSync(stream.nfd);
            }
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        },read:function (stream, buffer, offset, length, position) {
          if (length === 0) return 0; // node errors on 0 length reads
          // FIXME this is terrible.
          var nbuffer = new Buffer(length);
          var res;
          try {
            res = fs.readSync(stream.nfd, nbuffer, 0, length, position);
          } catch (e) {
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
          if (res > 0) {
            for (var i = 0; i < res; i++) {
              buffer[offset + i] = nbuffer[i];
            }
          }
          return res;
        },write:function (stream, buffer, offset, length, position) {
          // FIXME this is terrible.
          var nbuffer = new Buffer(buffer.subarray(offset, offset + length));
          var res;
          try {
            res = fs.writeSync(stream.nfd, nbuffer, 0, length, position);
          } catch (e) {
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
          return res;
        },llseek:function (stream, offset, whence) {
          var position = offset;
          if (whence === 1) {  // SEEK_CUR.
            position += stream.position;
          } else if (whence === 2) {  // SEEK_END.
            if (FS.isFile(stream.node.mode)) {
              try {
                var stat = fs.fstatSync(stream.nfd);
                position += stat.size;
              } catch (e) {
                throw new FS.ErrnoError(ERRNO_CODES[e.code]);
              }
            }
          }
  
          if (position < 0) {
            throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
          }
  
          return position;
        }}};
  
  var WORKERFS={DIR_MODE:16895,FILE_MODE:33279,reader:null,mount:function (mount) {
        assert(ENVIRONMENT_IS_WORKER);
        if (!WORKERFS.reader) WORKERFS.reader = new FileReaderSync();
        var root = WORKERFS.createNode(null, '/', WORKERFS.DIR_MODE, 0);
        var createdParents = {};
        function ensureParent(path) {
          // return the parent node, creating subdirs as necessary
          var parts = path.split('/');
          var parent = root;
          for (var i = 0; i < parts.length-1; i++) {
            var curr = parts.slice(0, i+1).join('/');
            if (!createdParents[curr]) {
              createdParents[curr] = WORKERFS.createNode(parent, curr, WORKERFS.DIR_MODE, 0);
            }
            parent = createdParents[curr];
          }
          return parent;
        }
        function base(path) {
          var parts = path.split('/');
          return parts[parts.length-1];
        }
        // We also accept FileList here, by using Array.prototype
        Array.prototype.forEach.call(mount.opts["files"] || [], function(file) {
          WORKERFS.createNode(ensureParent(file.name), base(file.name), WORKERFS.FILE_MODE, 0, file, file.lastModifiedDate);
        });
        (mount.opts["blobs"] || []).forEach(function(obj) {
          WORKERFS.createNode(ensureParent(obj["name"]), base(obj["name"]), WORKERFS.FILE_MODE, 0, obj["data"]);
        });
        (mount.opts["packages"] || []).forEach(function(pack) {
          pack['metadata'].files.forEach(function(file) {
            var name = file.filename.substr(1); // remove initial slash
            WORKERFS.createNode(ensureParent(name), base(name), WORKERFS.FILE_MODE, 0, pack['blob'].slice(file.start, file.end));
          });
        });
        return root;
      },createNode:function (parent, name, mode, dev, contents, mtime) {
        var node = FS.createNode(parent, name, mode);
        node.mode = mode;
        node.node_ops = WORKERFS.node_ops;
        node.stream_ops = WORKERFS.stream_ops;
        node.timestamp = (mtime || new Date).getTime();
        assert(WORKERFS.FILE_MODE !== WORKERFS.DIR_MODE);
        if (mode === WORKERFS.FILE_MODE) {
          node.size = contents.size;
          node.contents = contents;
        } else {
          node.size = 4096;
          node.contents = {};
        }
        if (parent) {
          parent.contents[name] = node;
        }
        return node;
      },node_ops:{getattr:function (node) {
          return {
            dev: 1,
            ino: undefined,
            mode: node.mode,
            nlink: 1,
            uid: 0,
            gid: 0,
            rdev: undefined,
            size: node.size,
            atime: new Date(node.timestamp),
            mtime: new Date(node.timestamp),
            ctime: new Date(node.timestamp),
            blksize: 4096,
            blocks: Math.ceil(node.size / 4096),
          };
        },setattr:function (node, attr) {
          if (attr.mode !== undefined) {
            node.mode = attr.mode;
          }
          if (attr.timestamp !== undefined) {
            node.timestamp = attr.timestamp;
          }
        },lookup:function (parent, name) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOENT);
        },mknod:function (parent, name, mode, dev) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        },rename:function (oldNode, newDir, newName) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        },unlink:function (parent, name) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        },rmdir:function (parent, name) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        },readdir:function (node) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        },symlink:function (parent, newName, oldPath) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        },readlink:function (node) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }},stream_ops:{read:function (stream, buffer, offset, length, position) {
          if (position >= stream.node.size) return 0;
          var chunk = stream.node.contents.slice(position, position + length);
          var ab = WORKERFS.reader.readAsArrayBuffer(chunk);
          buffer.set(new Uint8Array(ab), offset);
          return chunk.size;
        },write:function (stream, buffer, offset, length, position) {
          throw new FS.ErrnoError(ERRNO_CODES.EIO);
        },llseek:function (stream, offset, whence) {
          var position = offset;
          if (whence === 1) {  // SEEK_CUR.
            position += stream.position;
          } else if (whence === 2) {  // SEEK_END.
            if (FS.isFile(stream.node.mode)) {
              position += stream.node.size;
            }
          }
          if (position < 0) {
            throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
          }
          return position;
        }}};
  
  var _stdin=STATICTOP; STATICTOP += 16;;
  
  var _stdout=STATICTOP; STATICTOP += 16;;
  
  var _stderr=STATICTOP; STATICTOP += 16;;var FS={root:null,mounts:[],devices:[null],streams:[],nextInode:1,nameTable:null,currentPath:"/",initialized:false,ignorePermissions:true,trackingDelegate:{},tracking:{openFlags:{READ:1,WRITE:2}},ErrnoError:null,genericErrors:{},filesystems:null,syncFSRequests:0,handleFSError:function (e) {
        if (!(e instanceof FS.ErrnoError)) throw e + ' : ' + stackTrace();
        return ___setErrNo(e.errno);
      },lookupPath:function (path, opts) {
        path = PATH.resolve(FS.cwd(), path);
        opts = opts || {};
  
        if (!path) return { path: '', node: null };
  
        var defaults = {
          follow_mount: true,
          recurse_count: 0
        };
        for (var key in defaults) {
          if (opts[key] === undefined) {
            opts[key] = defaults[key];
          }
        }
  
        if (opts.recurse_count > 8) {  // max recursive lookup of 8
          throw new FS.ErrnoError(ERRNO_CODES.ELOOP);
        }
  
        // split the path
        var parts = PATH.normalizeArray(path.split('/').filter(function(p) {
          return !!p;
        }), false);
  
        // start at the root
        var current = FS.root;
        var current_path = '/';
  
        for (var i = 0; i < parts.length; i++) {
          var islast = (i === parts.length-1);
          if (islast && opts.parent) {
            // stop resolving
            break;
          }
  
          current = FS.lookupNode(current, parts[i]);
          current_path = PATH.join2(current_path, parts[i]);
  
          // jump to the mount's root node if this is a mountpoint
          if (FS.isMountpoint(current)) {
            if (!islast || (islast && opts.follow_mount)) {
              current = current.mounted.root;
            }
          }
  
          // by default, lookupPath will not follow a symlink if it is the final path component.
          // setting opts.follow = true will override this behavior.
          if (!islast || opts.follow) {
            var count = 0;
            while (FS.isLink(current.mode)) {
              var link = FS.readlink(current_path);
              current_path = PATH.resolve(PATH.dirname(current_path), link);
  
              var lookup = FS.lookupPath(current_path, { recurse_count: opts.recurse_count });
              current = lookup.node;
  
              if (count++ > 40) {  // limit max consecutive symlinks to 40 (SYMLOOP_MAX).
                throw new FS.ErrnoError(ERRNO_CODES.ELOOP);
              }
            }
          }
        }
  
        return { path: current_path, node: current };
      },getPath:function (node) {
        var path;
        while (true) {
          if (FS.isRoot(node)) {
            var mount = node.mount.mountpoint;
            if (!path) return mount;
            return mount[mount.length-1] !== '/' ? mount + '/' + path : mount + path;
          }
          path = path ? node.name + '/' + path : node.name;
          node = node.parent;
        }
      },hashName:function (parentid, name) {
        var hash = 0;
  
  
        for (var i = 0; i < name.length; i++) {
          hash = ((hash << 5) - hash + name.charCodeAt(i)) | 0;
        }
        return ((parentid + hash) >>> 0) % FS.nameTable.length;
      },hashAddNode:function (node) {
        var hash = FS.hashName(node.parent.id, node.name);
        node.name_next = FS.nameTable[hash];
        FS.nameTable[hash] = node;
      },hashRemoveNode:function (node) {
        var hash = FS.hashName(node.parent.id, node.name);
        if (FS.nameTable[hash] === node) {
          FS.nameTable[hash] = node.name_next;
        } else {
          var current = FS.nameTable[hash];
          while (current) {
            if (current.name_next === node) {
              current.name_next = node.name_next;
              break;
            }
            current = current.name_next;
          }
        }
      },lookupNode:function (parent, name) {
        var err = FS.mayLookup(parent);
        if (err) {
          throw new FS.ErrnoError(err, parent);
        }
        var hash = FS.hashName(parent.id, name);
        for (var node = FS.nameTable[hash]; node; node = node.name_next) {
          var nodeName = node.name;
          if (node.parent.id === parent.id && nodeName === name) {
            return node;
          }
        }
        // if we failed to find it in the cache, call into the VFS
        return FS.lookup(parent, name);
      },createNode:function (parent, name, mode, rdev) {
        if (!FS.FSNode) {
          FS.FSNode = function(parent, name, mode, rdev) {
            if (!parent) {
              parent = this;  // root node sets parent to itself
            }
            this.parent = parent;
            this.mount = parent.mount;
            this.mounted = null;
            this.id = FS.nextInode++;
            this.name = name;
            this.mode = mode;
            this.node_ops = {};
            this.stream_ops = {};
            this.rdev = rdev;
          };
  
          FS.FSNode.prototype = {};
  
          // compatibility
          var readMode = 292 | 73;
          var writeMode = 146;
  
          // NOTE we must use Object.defineProperties instead of individual calls to
          // Object.defineProperty in order to make closure compiler happy
          Object.defineProperties(FS.FSNode.prototype, {
            read: {
              get: function() { return (this.mode & readMode) === readMode; },
              set: function(val) { val ? this.mode |= readMode : this.mode &= ~readMode; }
            },
            write: {
              get: function() { return (this.mode & writeMode) === writeMode; },
              set: function(val) { val ? this.mode |= writeMode : this.mode &= ~writeMode; }
            },
            isFolder: {
              get: function() { return FS.isDir(this.mode); }
            },
            isDevice: {
              get: function() { return FS.isChrdev(this.mode); }
            }
          });
        }
  
        var node = new FS.FSNode(parent, name, mode, rdev);
  
        FS.hashAddNode(node);
  
        return node;
      },destroyNode:function (node) {
        FS.hashRemoveNode(node);
      },isRoot:function (node) {
        return node === node.parent;
      },isMountpoint:function (node) {
        return !!node.mounted;
      },isFile:function (mode) {
        return (mode & 61440) === 32768;
      },isDir:function (mode) {
        return (mode & 61440) === 16384;
      },isLink:function (mode) {
        return (mode & 61440) === 40960;
      },isChrdev:function (mode) {
        return (mode & 61440) === 8192;
      },isBlkdev:function (mode) {
        return (mode & 61440) === 24576;
      },isFIFO:function (mode) {
        return (mode & 61440) === 4096;
      },isSocket:function (mode) {
        return (mode & 49152) === 49152;
      },flagModes:{"r":0,"rs":1052672,"r+":2,"w":577,"wx":705,"xw":705,"w+":578,"wx+":706,"xw+":706,"a":1089,"ax":1217,"xa":1217,"a+":1090,"ax+":1218,"xa+":1218},modeStringToFlags:function (str) {
        var flags = FS.flagModes[str];
        if (typeof flags === 'undefined') {
          throw new Error('Unknown file open mode: ' + str);
        }
        return flags;
      },flagsToPermissionString:function (flag) {
        var perms = ['r', 'w', 'rw'][flag & 3];
        if ((flag & 512)) {
          perms += 'w';
        }
        return perms;
      },nodePermissions:function (node, perms) {
        if (FS.ignorePermissions) {
          return 0;
        }
        // return 0 if any user, group or owner bits are set.
        if (perms.indexOf('r') !== -1 && !(node.mode & 292)) {
          return ERRNO_CODES.EACCES;
        } else if (perms.indexOf('w') !== -1 && !(node.mode & 146)) {
          return ERRNO_CODES.EACCES;
        } else if (perms.indexOf('x') !== -1 && !(node.mode & 73)) {
          return ERRNO_CODES.EACCES;
        }
        return 0;
      },mayLookup:function (dir) {
        var err = FS.nodePermissions(dir, 'x');
        if (err) return err;
        if (!dir.node_ops.lookup) return ERRNO_CODES.EACCES;
        return 0;
      },mayCreate:function (dir, name) {
        try {
          var node = FS.lookupNode(dir, name);
          return ERRNO_CODES.EEXIST;
        } catch (e) {
        }
        return FS.nodePermissions(dir, 'wx');
      },mayDelete:function (dir, name, isdir) {
        var node;
        try {
          node = FS.lookupNode(dir, name);
        } catch (e) {
          return e.errno;
        }
        var err = FS.nodePermissions(dir, 'wx');
        if (err) {
          return err;
        }
        if (isdir) {
          if (!FS.isDir(node.mode)) {
            return ERRNO_CODES.ENOTDIR;
          }
          if (FS.isRoot(node) || FS.getPath(node) === FS.cwd()) {
            return ERRNO_CODES.EBUSY;
          }
        } else {
          if (FS.isDir(node.mode)) {
            return ERRNO_CODES.EISDIR;
          }
        }
        return 0;
      },mayOpen:function (node, flags) {
        if (!node) {
          return ERRNO_CODES.ENOENT;
        }
        if (FS.isLink(node.mode)) {
          return ERRNO_CODES.ELOOP;
        } else if (FS.isDir(node.mode)) {
          if (FS.flagsToPermissionString(flags) !== 'r' || // opening for write
              (flags & 512)) { // TODO: check for O_SEARCH? (== search for dir only)
            return ERRNO_CODES.EISDIR;
          }
        }
        return FS.nodePermissions(node, FS.flagsToPermissionString(flags));
      },MAX_OPEN_FDS:4096,nextfd:function (fd_start, fd_end) {
        fd_start = fd_start || 0;
        fd_end = fd_end || FS.MAX_OPEN_FDS;
        for (var fd = fd_start; fd <= fd_end; fd++) {
          if (!FS.streams[fd]) {
            return fd;
          }
        }
        throw new FS.ErrnoError(ERRNO_CODES.EMFILE);
      },getStream:function (fd) {
        return FS.streams[fd];
      },createStream:function (stream, fd_start, fd_end) {
        if (!FS.FSStream) {
          FS.FSStream = function(){};
          FS.FSStream.prototype = {};
          // compatibility
          Object.defineProperties(FS.FSStream.prototype, {
            object: {
              get: function() { return this.node; },
              set: function(val) { this.node = val; }
            },
            isRead: {
              get: function() { return (this.flags & 2097155) !== 1; }
            },
            isWrite: {
              get: function() { return (this.flags & 2097155) !== 0; }
            },
            isAppend: {
              get: function() { return (this.flags & 1024); }
            }
          });
        }
        // clone it, so we can return an instance of FSStream
        var newStream = new FS.FSStream();
        for (var p in stream) {
          newStream[p] = stream[p];
        }
        stream = newStream;
        var fd = FS.nextfd(fd_start, fd_end);
        stream.fd = fd;
        FS.streams[fd] = stream;
        return stream;
      },closeStream:function (fd) {
        FS.streams[fd] = null;
      },chrdev_stream_ops:{open:function (stream) {
          var device = FS.getDevice(stream.node.rdev);
          // override node's stream ops with the device's
          stream.stream_ops = device.stream_ops;
          // forward the open call
          if (stream.stream_ops.open) {
            stream.stream_ops.open(stream);
          }
        },llseek:function () {
          throw new FS.ErrnoError(ERRNO_CODES.ESPIPE);
        }},major:function (dev) {
        return ((dev) >> 8);
      },minor:function (dev) {
        return ((dev) & 0xff);
      },makedev:function (ma, mi) {
        return ((ma) << 8 | (mi));
      },registerDevice:function (dev, ops) {
        FS.devices[dev] = { stream_ops: ops };
      },getDevice:function (dev) {
        return FS.devices[dev];
      },getMounts:function (mount) {
        var mounts = [];
        var check = [mount];
  
        while (check.length) {
          var m = check.pop();
  
          mounts.push(m);
  
          check.push.apply(check, m.mounts);
        }
  
        return mounts;
      },syncfs:function (populate, callback) {
        if (typeof(populate) === 'function') {
          callback = populate;
          populate = false;
        }
  
        FS.syncFSRequests++;
  
        if (FS.syncFSRequests > 1) {
          console.log('warning: ' + FS.syncFSRequests + ' FS.syncfs operations in flight at once, probably just doing extra work');
        }
  
        var mounts = FS.getMounts(FS.root.mount);
        var completed = 0;
  
        function doCallback(err) {
          assert(FS.syncFSRequests > 0);
          FS.syncFSRequests--;
          return callback(err);
        }
  
        function done(err) {
          if (err) {
            if (!done.errored) {
              done.errored = true;
              return doCallback(err);
            }
            return;
          }
          if (++completed >= mounts.length) {
            doCallback(null);
          }
        };
  
        // sync all mounts
        mounts.forEach(function (mount) {
          if (!mount.type.syncfs) {
            return done(null);
          }
          mount.type.syncfs(mount, populate, done);
        });
      },mount:function (type, opts, mountpoint) {
        var root = mountpoint === '/';
        var pseudo = !mountpoint;
        var node;
  
        if (root && FS.root) {
          throw new FS.ErrnoError(ERRNO_CODES.EBUSY);
        } else if (!root && !pseudo) {
          var lookup = FS.lookupPath(mountpoint, { follow_mount: false });
  
          mountpoint = lookup.path;  // use the absolute path
          node = lookup.node;
  
          if (FS.isMountpoint(node)) {
            throw new FS.ErrnoError(ERRNO_CODES.EBUSY);
          }
  
          if (!FS.isDir(node.mode)) {
            throw new FS.ErrnoError(ERRNO_CODES.ENOTDIR);
          }
        }
  
        var mount = {
          type: type,
          opts: opts,
          mountpoint: mountpoint,
          mounts: []
        };
  
        // create a root node for the fs
        var mountRoot = type.mount(mount);
        mountRoot.mount = mount;
        mount.root = mountRoot;
  
        if (root) {
          FS.root = mountRoot;
        } else if (node) {
          // set as a mountpoint
          node.mounted = mount;
  
          // add the new mount to the current mount's children
          if (node.mount) {
            node.mount.mounts.push(mount);
          }
        }
  
        return mountRoot;
      },unmount:function (mountpoint) {
        var lookup = FS.lookupPath(mountpoint, { follow_mount: false });
  
        if (!FS.isMountpoint(lookup.node)) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
  
        // destroy the nodes for this mount, and all its child mounts
        var node = lookup.node;
        var mount = node.mounted;
        var mounts = FS.getMounts(mount);
  
        Object.keys(FS.nameTable).forEach(function (hash) {
          var current = FS.nameTable[hash];
  
          while (current) {
            var next = current.name_next;
  
            if (mounts.indexOf(current.mount) !== -1) {
              FS.destroyNode(current);
            }
  
            current = next;
          }
        });
  
        // no longer a mountpoint
        node.mounted = null;
  
        // remove this mount from the child mounts
        var idx = node.mount.mounts.indexOf(mount);
        assert(idx !== -1);
        node.mount.mounts.splice(idx, 1);
      },lookup:function (parent, name) {
        return parent.node_ops.lookup(parent, name);
      },mknod:function (path, mode, dev) {
        var lookup = FS.lookupPath(path, { parent: true });
        var parent = lookup.node;
        var name = PATH.basename(path);
        if (!name || name === '.' || name === '..') {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        var err = FS.mayCreate(parent, name);
        if (err) {
          throw new FS.ErrnoError(err);
        }
        if (!parent.node_ops.mknod) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        return parent.node_ops.mknod(parent, name, mode, dev);
      },create:function (path, mode) {
        mode = mode !== undefined ? mode : 438 /* 0666 */;
        mode &= 4095;
        mode |= 32768;
        return FS.mknod(path, mode, 0);
      },mkdir:function (path, mode) {
        mode = mode !== undefined ? mode : 511 /* 0777 */;
        mode &= 511 | 512;
        mode |= 16384;
        return FS.mknod(path, mode, 0);
      },mkdev:function (path, mode, dev) {
        if (typeof(dev) === 'undefined') {
          dev = mode;
          mode = 438 /* 0666 */;
        }
        mode |= 8192;
        return FS.mknod(path, mode, dev);
      },symlink:function (oldpath, newpath) {
        if (!PATH.resolve(oldpath)) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOENT);
        }
        var lookup = FS.lookupPath(newpath, { parent: true });
        var parent = lookup.node;
        if (!parent) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOENT);
        }
        var newname = PATH.basename(newpath);
        var err = FS.mayCreate(parent, newname);
        if (err) {
          throw new FS.ErrnoError(err);
        }
        if (!parent.node_ops.symlink) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        return parent.node_ops.symlink(parent, newname, oldpath);
      },rename:function (old_path, new_path) {
        var old_dirname = PATH.dirname(old_path);
        var new_dirname = PATH.dirname(new_path);
        var old_name = PATH.basename(old_path);
        var new_name = PATH.basename(new_path);
        // parents must exist
        var lookup, old_dir, new_dir;
        try {
          lookup = FS.lookupPath(old_path, { parent: true });
          old_dir = lookup.node;
          lookup = FS.lookupPath(new_path, { parent: true });
          new_dir = lookup.node;
        } catch (e) {
          throw new FS.ErrnoError(ERRNO_CODES.EBUSY);
        }
        if (!old_dir || !new_dir) throw new FS.ErrnoError(ERRNO_CODES.ENOENT);
        // need to be part of the same mount
        if (old_dir.mount !== new_dir.mount) {
          throw new FS.ErrnoError(ERRNO_CODES.EXDEV);
        }
        // source must exist
        var old_node = FS.lookupNode(old_dir, old_name);
        // old path should not be an ancestor of the new path
        var relative = PATH.relative(old_path, new_dirname);
        if (relative.charAt(0) !== '.') {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        // new path should not be an ancestor of the old path
        relative = PATH.relative(new_path, old_dirname);
        if (relative.charAt(0) !== '.') {
          throw new FS.ErrnoError(ERRNO_CODES.ENOTEMPTY);
        }
        // see if the new path already exists
        var new_node;
        try {
          new_node = FS.lookupNode(new_dir, new_name);
        } catch (e) {
          // not fatal
        }
        // early out if nothing needs to change
        if (old_node === new_node) {
          return;
        }
        // we'll need to delete the old entry
        var isdir = FS.isDir(old_node.mode);
        var err = FS.mayDelete(old_dir, old_name, isdir);
        if (err) {
          throw new FS.ErrnoError(err);
        }
        // need delete permissions if we'll be overwriting.
        // need create permissions if new doesn't already exist.
        err = new_node ?
          FS.mayDelete(new_dir, new_name, isdir) :
          FS.mayCreate(new_dir, new_name);
        if (err) {
          throw new FS.ErrnoError(err);
        }
        if (!old_dir.node_ops.rename) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        if (FS.isMountpoint(old_node) || (new_node && FS.isMountpoint(new_node))) {
          throw new FS.ErrnoError(ERRNO_CODES.EBUSY);
        }
        // if we are going to change the parent, check write permissions
        if (new_dir !== old_dir) {
          err = FS.nodePermissions(old_dir, 'w');
          if (err) {
            throw new FS.ErrnoError(err);
          }
        }
        try {
          if (FS.trackingDelegate['willMovePath']) {
            FS.trackingDelegate['willMovePath'](old_path, new_path);
          }
        } catch(e) {
          console.log("FS.trackingDelegate['willMovePath']('"+old_path+"', '"+new_path+"') threw an exception: " + e.message);
        }
        // remove the node from the lookup hash
        FS.hashRemoveNode(old_node);
        // do the underlying fs rename
        try {
          old_dir.node_ops.rename(old_node, new_dir, new_name);
        } catch (e) {
          throw e;
        } finally {
          // add the node back to the hash (in case node_ops.rename
          // changed its name)
          FS.hashAddNode(old_node);
        }
        try {
          if (FS.trackingDelegate['onMovePath']) FS.trackingDelegate['onMovePath'](old_path, new_path);
        } catch(e) {
          console.log("FS.trackingDelegate['onMovePath']('"+old_path+"', '"+new_path+"') threw an exception: " + e.message);
        }
      },rmdir:function (path) {
        var lookup = FS.lookupPath(path, { parent: true });
        var parent = lookup.node;
        var name = PATH.basename(path);
        var node = FS.lookupNode(parent, name);
        var err = FS.mayDelete(parent, name, true);
        if (err) {
          throw new FS.ErrnoError(err);
        }
        if (!parent.node_ops.rmdir) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        if (FS.isMountpoint(node)) {
          throw new FS.ErrnoError(ERRNO_CODES.EBUSY);
        }
        try {
          if (FS.trackingDelegate['willDeletePath']) {
            FS.trackingDelegate['willDeletePath'](path);
          }
        } catch(e) {
          console.log("FS.trackingDelegate['willDeletePath']('"+path+"') threw an exception: " + e.message);
        }
        parent.node_ops.rmdir(parent, name);
        FS.destroyNode(node);
        try {
          if (FS.trackingDelegate['onDeletePath']) FS.trackingDelegate['onDeletePath'](path);
        } catch(e) {
          console.log("FS.trackingDelegate['onDeletePath']('"+path+"') threw an exception: " + e.message);
        }
      },readdir:function (path) {
        var lookup = FS.lookupPath(path, { follow: true });
        var node = lookup.node;
        if (!node.node_ops.readdir) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOTDIR);
        }
        return node.node_ops.readdir(node);
      },unlink:function (path) {
        var lookup = FS.lookupPath(path, { parent: true });
        var parent = lookup.node;
        var name = PATH.basename(path);
        var node = FS.lookupNode(parent, name);
        var err = FS.mayDelete(parent, name, false);
        if (err) {
          // POSIX says unlink should set EPERM, not EISDIR
          if (err === ERRNO_CODES.EISDIR) err = ERRNO_CODES.EPERM;
          throw new FS.ErrnoError(err);
        }
        if (!parent.node_ops.unlink) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        if (FS.isMountpoint(node)) {
          throw new FS.ErrnoError(ERRNO_CODES.EBUSY);
        }
        try {
          if (FS.trackingDelegate['willDeletePath']) {
            FS.trackingDelegate['willDeletePath'](path);
          }
        } catch(e) {
          console.log("FS.trackingDelegate['willDeletePath']('"+path+"') threw an exception: " + e.message);
        }
        parent.node_ops.unlink(parent, name);
        FS.destroyNode(node);
        try {
          if (FS.trackingDelegate['onDeletePath']) FS.trackingDelegate['onDeletePath'](path);
        } catch(e) {
          console.log("FS.trackingDelegate['onDeletePath']('"+path+"') threw an exception: " + e.message);
        }
      },readlink:function (path) {
        var lookup = FS.lookupPath(path);
        var link = lookup.node;
        if (!link) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOENT);
        }
        if (!link.node_ops.readlink) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        return PATH.resolve(FS.getPath(link.parent), link.node_ops.readlink(link));
      },stat:function (path, dontFollow) {
        var lookup = FS.lookupPath(path, { follow: !dontFollow });
        var node = lookup.node;
        if (!node) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOENT);
        }
        if (!node.node_ops.getattr) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        return node.node_ops.getattr(node);
      },lstat:function (path) {
        return FS.stat(path, true);
      },chmod:function (path, mode, dontFollow) {
        var node;
        if (typeof path === 'string') {
          var lookup = FS.lookupPath(path, { follow: !dontFollow });
          node = lookup.node;
        } else {
          node = path;
        }
        if (!node.node_ops.setattr) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        node.node_ops.setattr(node, {
          mode: (mode & 4095) | (node.mode & ~4095),
          timestamp: Date.now()
        });
      },lchmod:function (path, mode) {
        FS.chmod(path, mode, true);
      },fchmod:function (fd, mode) {
        var stream = FS.getStream(fd);
        if (!stream) {
          throw new FS.ErrnoError(ERRNO_CODES.EBADF);
        }
        FS.chmod(stream.node, mode);
      },chown:function (path, uid, gid, dontFollow) {
        var node;
        if (typeof path === 'string') {
          var lookup = FS.lookupPath(path, { follow: !dontFollow });
          node = lookup.node;
        } else {
          node = path;
        }
        if (!node.node_ops.setattr) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        node.node_ops.setattr(node, {
          timestamp: Date.now()
          // we ignore the uid / gid for now
        });
      },lchown:function (path, uid, gid) {
        FS.chown(path, uid, gid, true);
      },fchown:function (fd, uid, gid) {
        var stream = FS.getStream(fd);
        if (!stream) {
          throw new FS.ErrnoError(ERRNO_CODES.EBADF);
        }
        FS.chown(stream.node, uid, gid);
      },truncate:function (path, len) {
        if (len < 0) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        var node;
        if (typeof path === 'string') {
          var lookup = FS.lookupPath(path, { follow: true });
          node = lookup.node;
        } else {
          node = path;
        }
        if (!node.node_ops.setattr) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        if (FS.isDir(node.mode)) {
          throw new FS.ErrnoError(ERRNO_CODES.EISDIR);
        }
        if (!FS.isFile(node.mode)) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        var err = FS.nodePermissions(node, 'w');
        if (err) {
          throw new FS.ErrnoError(err);
        }
        node.node_ops.setattr(node, {
          size: len,
          timestamp: Date.now()
        });
      },ftruncate:function (fd, len) {
        var stream = FS.getStream(fd);
        if (!stream) {
          throw new FS.ErrnoError(ERRNO_CODES.EBADF);
        }
        if ((stream.flags & 2097155) === 0) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        FS.truncate(stream.node, len);
      },utime:function (path, atime, mtime) {
        var lookup = FS.lookupPath(path, { follow: true });
        var node = lookup.node;
        node.node_ops.setattr(node, {
          timestamp: Math.max(atime, mtime)
        });
      },open:function (path, flags, mode, fd_start, fd_end) {
        if (path === "") {
          throw new FS.ErrnoError(ERRNO_CODES.ENOENT);
        }
        flags = typeof flags === 'string' ? FS.modeStringToFlags(flags) : flags;
        mode = typeof mode === 'undefined' ? 438 /* 0666 */ : mode;
        if ((flags & 64)) {
          mode = (mode & 4095) | 32768;
        } else {
          mode = 0;
        }
        var node;
        if (typeof path === 'object') {
          node = path;
        } else {
          path = PATH.normalize(path);
          try {
            var lookup = FS.lookupPath(path, {
              follow: !(flags & 131072)
            });
            node = lookup.node;
          } catch (e) {
            // ignore
          }
        }
        // perhaps we need to create the node
        var created = false;
        if ((flags & 64)) {
          if (node) {
            // if O_CREAT and O_EXCL are set, error out if the node already exists
            if ((flags & 128)) {
              throw new FS.ErrnoError(ERRNO_CODES.EEXIST);
            }
          } else {
            // node doesn't exist, try to create it
            node = FS.mknod(path, mode, 0);
            created = true;
          }
        }
        if (!node) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOENT);
        }
        // can't truncate a device
        if (FS.isChrdev(node.mode)) {
          flags &= ~512;
        }
        // if asked only for a directory, then this must be one
        if ((flags & 65536) && !FS.isDir(node.mode)) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOTDIR);
        }
        // check permissions, if this is not a file we just created now (it is ok to
        // create and write to a file with read-only permissions; it is read-only
        // for later use)
        if (!created) {
          var err = FS.mayOpen(node, flags);
          if (err) {
            throw new FS.ErrnoError(err);
          }
        }
        // do truncation if necessary
        if ((flags & 512)) {
          FS.truncate(node, 0);
        }
        // we've already handled these, don't pass down to the underlying vfs
        flags &= ~(128 | 512);
  
        // register the stream with the filesystem
        var stream = FS.createStream({
          node: node,
          path: FS.getPath(node),  // we want the absolute path to the node
          flags: flags,
          seekable: true,
          position: 0,
          stream_ops: node.stream_ops,
          // used by the file family libc calls (fopen, fwrite, ferror, etc.)
          ungotten: [],
          error: false
        }, fd_start, fd_end);
        // call the new stream's open function
        if (stream.stream_ops.open) {
          stream.stream_ops.open(stream);
        }
        if (Module['logReadFiles'] && !(flags & 1)) {
          if (!FS.readFiles) FS.readFiles = {};
          if (!(path in FS.readFiles)) {
            FS.readFiles[path] = 1;
            Module['printErr']('read file: ' + path);
          }
        }
        try {
          if (FS.trackingDelegate['onOpenFile']) {
            var trackingFlags = 0;
            if ((flags & 2097155) !== 1) {
              trackingFlags |= FS.tracking.openFlags.READ;
            }
            if ((flags & 2097155) !== 0) {
              trackingFlags |= FS.tracking.openFlags.WRITE;
            }
            FS.trackingDelegate['onOpenFile'](path, trackingFlags);
          }
        } catch(e) {
          console.log("FS.trackingDelegate['onOpenFile']('"+path+"', flags) threw an exception: " + e.message);
        }
        return stream;
      },close:function (stream) {
        if (stream.getdents) stream.getdents = null; // free readdir state
        try {
          if (stream.stream_ops.close) {
            stream.stream_ops.close(stream);
          }
        } catch (e) {
          throw e;
        } finally {
          FS.closeStream(stream.fd);
        }
      },llseek:function (stream, offset, whence) {
        if (!stream.seekable || !stream.stream_ops.llseek) {
          throw new FS.ErrnoError(ERRNO_CODES.ESPIPE);
        }
        stream.position = stream.stream_ops.llseek(stream, offset, whence);
        stream.ungotten = [];
        return stream.position;
      },read:function (stream, buffer, offset, length, position) {
        if (length < 0 || position < 0) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        if ((stream.flags & 2097155) === 1) {
          throw new FS.ErrnoError(ERRNO_CODES.EBADF);
        }
        if (FS.isDir(stream.node.mode)) {
          throw new FS.ErrnoError(ERRNO_CODES.EISDIR);
        }
        if (!stream.stream_ops.read) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        var seeking = true;
        if (typeof position === 'undefined') {
          position = stream.position;
          seeking = false;
        } else if (!stream.seekable) {
          throw new FS.ErrnoError(ERRNO_CODES.ESPIPE);
        }
        var bytesRead = stream.stream_ops.read(stream, buffer, offset, length, position);
        if (!seeking) stream.position += bytesRead;
        return bytesRead;
      },write:function (stream, buffer, offset, length, position, canOwn) {
        if (length < 0 || position < 0) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        if ((stream.flags & 2097155) === 0) {
          throw new FS.ErrnoError(ERRNO_CODES.EBADF);
        }
        if (FS.isDir(stream.node.mode)) {
          throw new FS.ErrnoError(ERRNO_CODES.EISDIR);
        }
        if (!stream.stream_ops.write) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        if (stream.flags & 1024) {
          // seek to the end before writing in append mode
          FS.llseek(stream, 0, 2);
        }
        var seeking = true;
        if (typeof position === 'undefined') {
          position = stream.position;
          seeking = false;
        } else if (!stream.seekable) {
          throw new FS.ErrnoError(ERRNO_CODES.ESPIPE);
        }
        var bytesWritten = stream.stream_ops.write(stream, buffer, offset, length, position, canOwn);
        if (!seeking) stream.position += bytesWritten;
        try {
          if (stream.path && FS.trackingDelegate['onWriteToFile']) FS.trackingDelegate['onWriteToFile'](stream.path);
        } catch(e) {
          console.log("FS.trackingDelegate['onWriteToFile']('"+path+"') threw an exception: " + e.message);
        }
        return bytesWritten;
      },allocate:function (stream, offset, length) {
        if (offset < 0 || length <= 0) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        if ((stream.flags & 2097155) === 0) {
          throw new FS.ErrnoError(ERRNO_CODES.EBADF);
        }
        if (!FS.isFile(stream.node.mode) && !FS.isDir(node.mode)) {
          throw new FS.ErrnoError(ERRNO_CODES.ENODEV);
        }
        if (!stream.stream_ops.allocate) {
          throw new FS.ErrnoError(ERRNO_CODES.EOPNOTSUPP);
        }
        stream.stream_ops.allocate(stream, offset, length);
      },mmap:function (stream, buffer, offset, length, position, prot, flags) {
        // TODO if PROT is PROT_WRITE, make sure we have write access
        if ((stream.flags & 2097155) === 1) {
          throw new FS.ErrnoError(ERRNO_CODES.EACCES);
        }
        if (!stream.stream_ops.mmap) {
          throw new FS.ErrnoError(ERRNO_CODES.ENODEV);
        }
        return stream.stream_ops.mmap(stream, buffer, offset, length, position, prot, flags);
      },msync:function (stream, buffer, offset, length, mmapFlags) {
        if (!stream || !stream.stream_ops.msync) {
          return 0;
        }
        return stream.stream_ops.msync(stream, buffer, offset, length, mmapFlags);
      },munmap:function (stream) {
        return 0;
      },ioctl:function (stream, cmd, arg) {
        if (!stream.stream_ops.ioctl) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOTTY);
        }
        return stream.stream_ops.ioctl(stream, cmd, arg);
      },readFile:function (path, opts) {
        opts = opts || {};
        opts.flags = opts.flags || 'r';
        opts.encoding = opts.encoding || 'binary';
        if (opts.encoding !== 'utf8' && opts.encoding !== 'binary') {
          throw new Error('Invalid encoding type "' + opts.encoding + '"');
        }
        var ret;
        var stream = FS.open(path, opts.flags);
        var stat = FS.stat(path);
        var length = stat.size;
        var buf = new Uint8Array(length);
        FS.read(stream, buf, 0, length, 0);
        if (opts.encoding === 'utf8') {
          ret = UTF8ArrayToString(buf, 0);
        } else if (opts.encoding === 'binary') {
          ret = buf;
        }
        FS.close(stream);
        return ret;
      },writeFile:function (path, data, opts) {
        opts = opts || {};
        opts.flags = opts.flags || 'w';
        opts.encoding = opts.encoding || 'utf8';
        if (opts.encoding !== 'utf8' && opts.encoding !== 'binary') {
          throw new Error('Invalid encoding type "' + opts.encoding + '"');
        }
        var stream = FS.open(path, opts.flags, opts.mode);
        if (opts.encoding === 'utf8') {
          var buf = new Uint8Array(lengthBytesUTF8(data)+1);
          var actualNumBytes = stringToUTF8Array(data, buf, 0, buf.length);
          FS.write(stream, buf, 0, actualNumBytes, 0, opts.canOwn);
        } else if (opts.encoding === 'binary') {
          FS.write(stream, data, 0, data.length, 0, opts.canOwn);
        }
        FS.close(stream);
      },cwd:function () {
        return FS.currentPath;
      },chdir:function (path) {
        var lookup = FS.lookupPath(path, { follow: true });
        if (!FS.isDir(lookup.node.mode)) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOTDIR);
        }
        var err = FS.nodePermissions(lookup.node, 'x');
        if (err) {
          throw new FS.ErrnoError(err);
        }
        FS.currentPath = lookup.path;
      },createDefaultDirectories:function () {
        FS.mkdir('/tmp');
        FS.mkdir('/home');
        FS.mkdir('/home/web_user');
      },createDefaultDevices:function () {
        // create /dev
        FS.mkdir('/dev');
        // setup /dev/null
        FS.registerDevice(FS.makedev(1, 3), {
          read: function() { return 0; },
          write: function(stream, buffer, offset, length, pos) { return length; }
        });
        FS.mkdev('/dev/null', FS.makedev(1, 3));
        // setup /dev/tty and /dev/tty1
        // stderr needs to print output using Module['printErr']
        // so we register a second tty just for it.
        TTY.register(FS.makedev(5, 0), TTY.default_tty_ops);
        TTY.register(FS.makedev(6, 0), TTY.default_tty1_ops);
        FS.mkdev('/dev/tty', FS.makedev(5, 0));
        FS.mkdev('/dev/tty1', FS.makedev(6, 0));
        // setup /dev/[u]random
        var random_device;
        if (typeof crypto !== 'undefined') {
          // for modern web browsers
          var randomBuffer = new Uint8Array(1);
          random_device = function() { crypto.getRandomValues(randomBuffer); return randomBuffer[0]; };
        } else if (ENVIRONMENT_IS_NODE) {
          // for nodejs
          random_device = function() { return require('crypto').randomBytes(1)[0]; };
        } else {
          // default for ES5 platforms
          random_device = function() { return (Math.random()*256)|0; };
        }
        FS.createDevice('/dev', 'random', random_device);
        FS.createDevice('/dev', 'urandom', random_device);
        // we're not going to emulate the actual shm device,
        // just create the tmp dirs that reside in it commonly
        FS.mkdir('/dev/shm');
        FS.mkdir('/dev/shm/tmp');
      },createSpecialDirectories:function () {
        // create /proc/self/fd which allows /proc/self/fd/6 => readlink gives the name of the stream for fd 6 (see test_unistd_ttyname)
        FS.mkdir('/proc');
        FS.mkdir('/proc/self');
        FS.mkdir('/proc/self/fd');
        FS.mount({
          mount: function() {
            var node = FS.createNode('/proc/self', 'fd', 16384 | 0777, 73);
            node.node_ops = {
              lookup: function(parent, name) {
                var fd = +name;
                var stream = FS.getStream(fd);
                if (!stream) throw new FS.ErrnoError(ERRNO_CODES.EBADF);
                var ret = {
                  parent: null,
                  mount: { mountpoint: 'fake' },
                  node_ops: { readlink: function() { return stream.path } }
                };
                ret.parent = ret; // make it look like a simple root node
                return ret;
              }
            };
            return node;
          }
        }, {}, '/proc/self/fd');
      },createStandardStreams:function () {
        // TODO deprecate the old functionality of a single
        // input / output callback and that utilizes FS.createDevice
        // and instead require a unique set of stream ops
  
        // by default, we symlink the standard streams to the
        // default tty devices. however, if the standard streams
        // have been overwritten we create a unique device for
        // them instead.
        if (Module['stdin']) {
          FS.createDevice('/dev', 'stdin', Module['stdin']);
        } else {
          FS.symlink('/dev/tty', '/dev/stdin');
        }
        if (Module['stdout']) {
          FS.createDevice('/dev', 'stdout', null, Module['stdout']);
        } else {
          FS.symlink('/dev/tty', '/dev/stdout');
        }
        if (Module['stderr']) {
          FS.createDevice('/dev', 'stderr', null, Module['stderr']);
        } else {
          FS.symlink('/dev/tty1', '/dev/stderr');
        }
  
        // open default streams for the stdin, stdout and stderr devices
        var stdin = FS.open('/dev/stdin', 'r');
        assert(stdin.fd === 0, 'invalid handle for stdin (' + stdin.fd + ')');
  
        var stdout = FS.open('/dev/stdout', 'w');
        assert(stdout.fd === 1, 'invalid handle for stdout (' + stdout.fd + ')');
  
        var stderr = FS.open('/dev/stderr', 'w');
        assert(stderr.fd === 2, 'invalid handle for stderr (' + stderr.fd + ')');
      },ensureErrnoError:function () {
        if (FS.ErrnoError) return;
        FS.ErrnoError = function ErrnoError(errno, node) {
          //Module.printErr(stackTrace()); // useful for debugging
          this.node = node;
          this.setErrno = function(errno) {
            this.errno = errno;
            for (var key in ERRNO_CODES) {
              if (ERRNO_CODES[key] === errno) {
                this.code = key;
                break;
              }
            }
          };
          this.setErrno(errno);
          this.message = ERRNO_MESSAGES[errno];
          if (this.stack) this.stack = demangleAll(this.stack);
        };
        FS.ErrnoError.prototype = new Error();
        FS.ErrnoError.prototype.constructor = FS.ErrnoError;
        // Some errors may happen quite a bit, to avoid overhead we reuse them (and suffer a lack of stack info)
        [ERRNO_CODES.ENOENT].forEach(function(code) {
          FS.genericErrors[code] = new FS.ErrnoError(code);
          FS.genericErrors[code].stack = '<generic error, no stack>';
        });
      },staticInit:function () {
        FS.ensureErrnoError();
  
        FS.nameTable = new Array(4096);
  
        FS.mount(MEMFS, {}, '/');
  
        FS.createDefaultDirectories();
        FS.createDefaultDevices();
        FS.createSpecialDirectories();
  
        FS.filesystems = {
          'MEMFS': MEMFS,
          'IDBFS': IDBFS,
          'NODEFS': NODEFS,
          'WORKERFS': WORKERFS,
        };
      },init:function (input, output, error) {
        assert(!FS.init.initialized, 'FS.init was previously called. If you want to initialize later with custom parameters, remove any earlier calls (note that one is automatically added to the generated code)');
        FS.init.initialized = true;
  
        FS.ensureErrnoError();
  
        // Allow Module.stdin etc. to provide defaults, if none explicitly passed to us here
        Module['stdin'] = input || Module['stdin'];
        Module['stdout'] = output || Module['stdout'];
        Module['stderr'] = error || Module['stderr'];
  
        FS.createStandardStreams();
      },quit:function () {
        FS.init.initialized = false;
        // force-flush all streams, so we get musl std streams printed out
        var fflush = Module['_fflush'];
        if (fflush) fflush(0);
        // close all of our streams
        for (var i = 0; i < FS.streams.length; i++) {
          var stream = FS.streams[i];
          if (!stream) {
            continue;
          }
          FS.close(stream);
        }
      },getMode:function (canRead, canWrite) {
        var mode = 0;
        if (canRead) mode |= 292 | 73;
        if (canWrite) mode |= 146;
        return mode;
      },joinPath:function (parts, forceRelative) {
        var path = PATH.join.apply(null, parts);
        if (forceRelative && path[0] == '/') path = path.substr(1);
        return path;
      },absolutePath:function (relative, base) {
        return PATH.resolve(base, relative);
      },standardizePath:function (path) {
        return PATH.normalize(path);
      },findObject:function (path, dontResolveLastLink) {
        var ret = FS.analyzePath(path, dontResolveLastLink);
        if (ret.exists) {
          return ret.object;
        } else {
          ___setErrNo(ret.error);
          return null;
        }
      },analyzePath:function (path, dontResolveLastLink) {
        // operate from within the context of the symlink's target
        try {
          var lookup = FS.lookupPath(path, { follow: !dontResolveLastLink });
          path = lookup.path;
        } catch (e) {
        }
        var ret = {
          isRoot: false, exists: false, error: 0, name: null, path: null, object: null,
          parentExists: false, parentPath: null, parentObject: null
        };
        try {
          var lookup = FS.lookupPath(path, { parent: true });
          ret.parentExists = true;
          ret.parentPath = lookup.path;
          ret.parentObject = lookup.node;
          ret.name = PATH.basename(path);
          lookup = FS.lookupPath(path, { follow: !dontResolveLastLink });
          ret.exists = true;
          ret.path = lookup.path;
          ret.object = lookup.node;
          ret.name = lookup.node.name;
          ret.isRoot = lookup.path === '/';
        } catch (e) {
          ret.error = e.errno;
        };
        return ret;
      },createFolder:function (parent, name, canRead, canWrite) {
        var path = PATH.join2(typeof parent === 'string' ? parent : FS.getPath(parent), name);
        var mode = FS.getMode(canRead, canWrite);
        return FS.mkdir(path, mode);
      },createPath:function (parent, path, canRead, canWrite) {
        parent = typeof parent === 'string' ? parent : FS.getPath(parent);
        var parts = path.split('/').reverse();
        while (parts.length) {
          var part = parts.pop();
          if (!part) continue;
          var current = PATH.join2(parent, part);
          try {
            FS.mkdir(current);
          } catch (e) {
            // ignore EEXIST
          }
          parent = current;
        }
        return current;
      },createFile:function (parent, name, properties, canRead, canWrite) {
        var path = PATH.join2(typeof parent === 'string' ? parent : FS.getPath(parent), name);
        var mode = FS.getMode(canRead, canWrite);
        return FS.create(path, mode);
      },createDataFile:function (parent, name, data, canRead, canWrite, canOwn) {
        var path = name ? PATH.join2(typeof parent === 'string' ? parent : FS.getPath(parent), name) : parent;
        var mode = FS.getMode(canRead, canWrite);
        var node = FS.create(path, mode);
        if (data) {
          if (typeof data === 'string') {
            var arr = new Array(data.length);
            for (var i = 0, len = data.length; i < len; ++i) arr[i] = data.charCodeAt(i);
            data = arr;
          }
          // make sure we can write to the file
          FS.chmod(node, mode | 146);
          var stream = FS.open(node, 'w');
          FS.write(stream, data, 0, data.length, 0, canOwn);
          FS.close(stream);
          FS.chmod(node, mode);
        }
        return node;
      },createDevice:function (parent, name, input, output) {
        var path = PATH.join2(typeof parent === 'string' ? parent : FS.getPath(parent), name);
        var mode = FS.getMode(!!input, !!output);
        if (!FS.createDevice.major) FS.createDevice.major = 64;
        var dev = FS.makedev(FS.createDevice.major++, 0);
        // Create a fake device that a set of stream ops to emulate
        // the old behavior.
        FS.registerDevice(dev, {
          open: function(stream) {
            stream.seekable = false;
          },
          close: function(stream) {
            // flush any pending line data
            if (output && output.buffer && output.buffer.length) {
              output(10);
            }
          },
          read: function(stream, buffer, offset, length, pos /* ignored */) {
            var bytesRead = 0;
            for (var i = 0; i < length; i++) {
              var result;
              try {
                result = input();
              } catch (e) {
                throw new FS.ErrnoError(ERRNO_CODES.EIO);
              }
              if (result === undefined && bytesRead === 0) {
                throw new FS.ErrnoError(ERRNO_CODES.EAGAIN);
              }
              if (result === null || result === undefined) break;
              bytesRead++;
              buffer[offset+i] = result;
            }
            if (bytesRead) {
              stream.node.timestamp = Date.now();
            }
            return bytesRead;
          },
          write: function(stream, buffer, offset, length, pos) {
            for (var i = 0; i < length; i++) {
              try {
                output(buffer[offset+i]);
              } catch (e) {
                throw new FS.ErrnoError(ERRNO_CODES.EIO);
              }
            }
            if (length) {
              stream.node.timestamp = Date.now();
            }
            return i;
          }
        });
        return FS.mkdev(path, mode, dev);
      },createLink:function (parent, name, target, canRead, canWrite) {
        var path = PATH.join2(typeof parent === 'string' ? parent : FS.getPath(parent), name);
        return FS.symlink(target, path);
      },forceLoadFile:function (obj) {
        if (obj.isDevice || obj.isFolder || obj.link || obj.contents) return true;
        var success = true;
        if (typeof XMLHttpRequest !== 'undefined') {
          throw new Error("Lazy loading should have been performed (contents set) in createLazyFile, but it was not. Lazy loading only works in web workers. Use --embed-file or --preload-file in emcc on the main thread.");
        } else if (Module['read']) {
          // Command-line.
          try {
            // WARNING: Can't read binary files in V8's d8 or tracemonkey's js, as
            //          read() will try to parse UTF8.
            obj.contents = intArrayFromString(Module['read'](obj.url), true);
            obj.usedBytes = obj.contents.length;
          } catch (e) {
            success = false;
          }
        } else {
          throw new Error('Cannot load without read() or XMLHttpRequest.');
        }
        if (!success) ___setErrNo(ERRNO_CODES.EIO);
        return success;
      },createLazyFile:function (parent, name, url, canRead, canWrite) {
        // Lazy chunked Uint8Array (implements get and length from Uint8Array). Actual getting is abstracted away for eventual reuse.
        function LazyUint8Array() {
          this.lengthKnown = false;
          this.chunks = []; // Loaded chunks. Index is the chunk number
        }
        LazyUint8Array.prototype.get = function LazyUint8Array_get(idx) {
          if (idx > this.length-1 || idx < 0) {
            return undefined;
          }
          var chunkOffset = idx % this.chunkSize;
          var chunkNum = (idx / this.chunkSize)|0;
          return this.getter(chunkNum)[chunkOffset];
        }
        LazyUint8Array.prototype.setDataGetter = function LazyUint8Array_setDataGetter(getter) {
          this.getter = getter;
        }
        LazyUint8Array.prototype.cacheLength = function LazyUint8Array_cacheLength() {
          // Find length
          var xhr = new XMLHttpRequest();
          xhr.open('HEAD', url, false);
          xhr.send(null);
          if (!(xhr.status >= 200 && xhr.status < 300 || xhr.status === 304)) throw new Error("Couldn't load " + url + ". Status: " + xhr.status);
          var datalength = Number(xhr.getResponseHeader("Content-length"));
          var header;
          var hasByteServing = (header = xhr.getResponseHeader("Accept-Ranges")) && header === "bytes";
          var usesGzip = (header = xhr.getResponseHeader("Content-Encoding")) && header === "gzip";
  
          var chunkSize = 1024*1024; // Chunk size in bytes
  
          if (!hasByteServing) chunkSize = datalength;
  
          // Function to get a range from the remote URL.
          var doXHR = (function(from, to) {
            if (from > to) throw new Error("invalid range (" + from + ", " + to + ") or no bytes requested!");
            if (to > datalength-1) throw new Error("only " + datalength + " bytes available! programmer error!");
  
            // TODO: Use mozResponseArrayBuffer, responseStream, etc. if available.
            var xhr = new XMLHttpRequest();
            xhr.open('GET', url, false);
            if (datalength !== chunkSize) xhr.setRequestHeader("Range", "bytes=" + from + "-" + to);
  
            // Some hints to the browser that we want binary data.
            if (typeof Uint8Array != 'undefined') xhr.responseType = 'arraybuffer';
            if (xhr.overrideMimeType) {
              xhr.overrideMimeType('text/plain; charset=x-user-defined');
            }
  
            xhr.send(null);
            if (!(xhr.status >= 200 && xhr.status < 300 || xhr.status === 304)) throw new Error("Couldn't load " + url + ". Status: " + xhr.status);
            if (xhr.response !== undefined) {
              return new Uint8Array(xhr.response || []);
            } else {
              return intArrayFromString(xhr.responseText || '', true);
            }
          });
          var lazyArray = this;
          lazyArray.setDataGetter(function(chunkNum) {
            var start = chunkNum * chunkSize;
            var end = (chunkNum+1) * chunkSize - 1; // including this byte
            end = Math.min(end, datalength-1); // if datalength-1 is selected, this is the last block
            if (typeof(lazyArray.chunks[chunkNum]) === "undefined") {
              lazyArray.chunks[chunkNum] = doXHR(start, end);
            }
            if (typeof(lazyArray.chunks[chunkNum]) === "undefined") throw new Error("doXHR failed!");
            return lazyArray.chunks[chunkNum];
          });
  
          if (usesGzip || !datalength) {
            // if the server uses gzip or doesn't supply the length, we have to download the whole file to get the (uncompressed) length
            chunkSize = datalength = 1; // this will force getter(0)/doXHR do download the whole file
            datalength = this.getter(0).length;
            chunkSize = datalength;
            console.log("LazyFiles on gzip forces download of the whole file when length is accessed");
          }
  
          this._length = datalength;
          this._chunkSize = chunkSize;
          this.lengthKnown = true;
        }
        if (typeof XMLHttpRequest !== 'undefined') {
          if (!ENVIRONMENT_IS_WORKER) throw 'Cannot do synchronous binary XHRs outside webworkers in modern browsers. Use --embed-file or --preload-file in emcc';
          var lazyArray = new LazyUint8Array();
          Object.defineProperties(lazyArray, {
            length: {
              get: function() {
                if(!this.lengthKnown) {
                  this.cacheLength();
                }
                return this._length;
              }
            },
            chunkSize: {
              get: function() {
                if(!this.lengthKnown) {
                  this.cacheLength();
                }
                return this._chunkSize;
              }
            }
          });
  
          var properties = { isDevice: false, contents: lazyArray };
        } else {
          var properties = { isDevice: false, url: url };
        }
  
        var node = FS.createFile(parent, name, properties, canRead, canWrite);
        // This is a total hack, but I want to get this lazy file code out of the
        // core of MEMFS. If we want to keep this lazy file concept I feel it should
        // be its own thin LAZYFS proxying calls to MEMFS.
        if (properties.contents) {
          node.contents = properties.contents;
        } else if (properties.url) {
          node.contents = null;
          node.url = properties.url;
        }
        // Add a function that defers querying the file size until it is asked the first time.
        Object.defineProperties(node, {
          usedBytes: {
            get: function() { return this.contents.length; }
          }
        });
        // override each stream op with one that tries to force load the lazy file first
        var stream_ops = {};
        var keys = Object.keys(node.stream_ops);
        keys.forEach(function(key) {
          var fn = node.stream_ops[key];
          stream_ops[key] = function forceLoadLazyFile() {
            if (!FS.forceLoadFile(node)) {
              throw new FS.ErrnoError(ERRNO_CODES.EIO);
            }
            return fn.apply(null, arguments);
          };
        });
        // use a custom read function
        stream_ops.read = function stream_ops_read(stream, buffer, offset, length, position) {
          if (!FS.forceLoadFile(node)) {
            throw new FS.ErrnoError(ERRNO_CODES.EIO);
          }
          var contents = stream.node.contents;
          if (position >= contents.length)
            return 0;
          var size = Math.min(contents.length - position, length);
          assert(size >= 0);
          if (contents.slice) { // normal array
            for (var i = 0; i < size; i++) {
              buffer[offset + i] = contents[position + i];
            }
          } else {
            for (var i = 0; i < size; i++) { // LazyUint8Array from sync binary XHR
              buffer[offset + i] = contents.get(position + i);
            }
          }
          return size;
        };
        node.stream_ops = stream_ops;
        return node;
      },createPreloadedFile:function (parent, name, url, canRead, canWrite, onload, onerror, dontCreateFile, canOwn, preFinish) {
        Browser.init(); // XXX perhaps this method should move onto Browser?
        // TODO we should allow people to just pass in a complete filename instead
        // of parent and name being that we just join them anyways
        var fullname = name ? PATH.resolve(PATH.join2(parent, name)) : parent;
        var dep = getUniqueRunDependency('cp ' + fullname); // might have several active requests for the same fullname
        function processData(byteArray) {
          function finish(byteArray) {
            if (preFinish) preFinish();
            if (!dontCreateFile) {
              FS.createDataFile(parent, name, byteArray, canRead, canWrite, canOwn);
            }
            if (onload) onload();
            removeRunDependency(dep);
          }
          var handled = false;
          Module['preloadPlugins'].forEach(function(plugin) {
            if (handled) return;
            if (plugin['canHandle'](fullname)) {
              plugin['handle'](byteArray, fullname, finish, function() {
                if (onerror) onerror();
                removeRunDependency(dep);
              });
              handled = true;
            }
          });
          if (!handled) finish(byteArray);
        }
        addRunDependency(dep);
        if (typeof url == 'string') {
          Browser.asyncLoad(url, function(byteArray) {
            processData(byteArray);
          }, onerror);
        } else {
          processData(url);
        }
      },indexedDB:function () {
        return window.indexedDB || window.mozIndexedDB || window.webkitIndexedDB || window.msIndexedDB;
      },DB_NAME:function () {
        return 'EM_FS_' + window.location.pathname;
      },DB_VERSION:20,DB_STORE_NAME:"FILE_DATA",saveFilesToDB:function (paths, onload, onerror) {
        onload = onload || function(){};
        onerror = onerror || function(){};
        var indexedDB = FS.indexedDB();
        try {
          var openRequest = indexedDB.open(FS.DB_NAME(), FS.DB_VERSION);
        } catch (e) {
          return onerror(e);
        }
        openRequest.onupgradeneeded = function openRequest_onupgradeneeded() {
          console.log('creating db');
          var db = openRequest.result;
          db.createObjectStore(FS.DB_STORE_NAME);
        };
        openRequest.onsuccess = function openRequest_onsuccess() {
          var db = openRequest.result;
          var transaction = db.transaction([FS.DB_STORE_NAME], 'readwrite');
          var files = transaction.objectStore(FS.DB_STORE_NAME);
          var ok = 0, fail = 0, total = paths.length;
          function finish() {
            if (fail == 0) onload(); else onerror();
          }
          paths.forEach(function(path) {
            var putRequest = files.put(FS.analyzePath(path).object.contents, path);
            putRequest.onsuccess = function putRequest_onsuccess() { ok++; if (ok + fail == total) finish() };
            putRequest.onerror = function putRequest_onerror() { fail++; if (ok + fail == total) finish() };
          });
          transaction.onerror = onerror;
        };
        openRequest.onerror = onerror;
      },loadFilesFromDB:function (paths, onload, onerror) {
        onload = onload || function(){};
        onerror = onerror || function(){};
        var indexedDB = FS.indexedDB();
        try {
          var openRequest = indexedDB.open(FS.DB_NAME(), FS.DB_VERSION);
        } catch (e) {
          return onerror(e);
        }
        openRequest.onupgradeneeded = onerror; // no database to load from
        openRequest.onsuccess = function openRequest_onsuccess() {
          var db = openRequest.result;
          try {
            var transaction = db.transaction([FS.DB_STORE_NAME], 'readonly');
          } catch(e) {
            onerror(e);
            return;
          }
          var files = transaction.objectStore(FS.DB_STORE_NAME);
          var ok = 0, fail = 0, total = paths.length;
          function finish() {
            if (fail == 0) onload(); else onerror();
          }
          paths.forEach(function(path) {
            var getRequest = files.get(path);
            getRequest.onsuccess = function getRequest_onsuccess() {
              if (FS.analyzePath(path).exists) {
                FS.unlink(path);
              }
              FS.createDataFile(PATH.dirname(path), PATH.basename(path), getRequest.result, true, true, true);
              ok++;
              if (ok + fail == total) finish();
            };
            getRequest.onerror = function getRequest_onerror() { fail++; if (ok + fail == total) finish() };
          });
          transaction.onerror = onerror;
        };
        openRequest.onerror = onerror;
      }};var SYSCALLS={DEFAULT_POLLMASK:5,mappings:{},umask:511,calculateAt:function (dirfd, path) {
        if (path[0] !== '/') {
          // relative path
          var dir;
          if (dirfd === -100) {
            dir = FS.cwd();
          } else {
            var dirstream = FS.getStream(dirfd);
            if (!dirstream) throw new FS.ErrnoError(ERRNO_CODES.EBADF);
            dir = dirstream.path;
          }
          path = PATH.join2(dir, path);
        }
        return path;
      },doStat:function (func, path, buf) {
        try {
          var stat = func(path);
        } catch (e) {
          if (e && e.node && PATH.normalize(path) !== PATH.normalize(FS.getPath(e.node))) {
            // an error occurred while trying to look up the path; we should just report ENOTDIR
            return -ERRNO_CODES.ENOTDIR;
          }
          throw e;
        }
        HEAP32[((buf)>>2)]=stat.dev;
        HEAP32[(((buf)+(4))>>2)]=0;
        HEAP32[(((buf)+(8))>>2)]=stat.ino;
        HEAP32[(((buf)+(12))>>2)]=stat.mode;
        HEAP32[(((buf)+(16))>>2)]=stat.nlink;
        HEAP32[(((buf)+(20))>>2)]=stat.uid;
        HEAP32[(((buf)+(24))>>2)]=stat.gid;
        HEAP32[(((buf)+(28))>>2)]=stat.rdev;
        HEAP32[(((buf)+(32))>>2)]=0;
        HEAP32[(((buf)+(36))>>2)]=stat.size;
        HEAP32[(((buf)+(40))>>2)]=4096;
        HEAP32[(((buf)+(44))>>2)]=stat.blocks;
        HEAP32[(((buf)+(48))>>2)]=(stat.atime.getTime() / 1000)|0;
        HEAP32[(((buf)+(52))>>2)]=0;
        HEAP32[(((buf)+(56))>>2)]=(stat.mtime.getTime() / 1000)|0;
        HEAP32[(((buf)+(60))>>2)]=0;
        HEAP32[(((buf)+(64))>>2)]=(stat.ctime.getTime() / 1000)|0;
        HEAP32[(((buf)+(68))>>2)]=0;
        HEAP32[(((buf)+(72))>>2)]=stat.ino;
        return 0;
      },doMsync:function (addr, stream, len, flags) {
        var buffer = new Uint8Array(HEAPU8.subarray(addr, addr + len));
        FS.msync(stream, buffer, 0, len, flags);
      },doMkdir:function (path, mode) {
        // remove a trailing slash, if one - /a/b/ has basename of '', but
        // we want to create b in the context of this function
        path = PATH.normalize(path);
        if (path[path.length-1] === '/') path = path.substr(0, path.length-1);
        FS.mkdir(path, mode, 0);
        return 0;
      },doMknod:function (path, mode, dev) {
        // we don't want this in the JS API as it uses mknod to create all nodes.
        switch (mode & 61440) {
          case 32768:
          case 8192:
          case 24576:
          case 4096:
          case 49152:
            break;
          default: return -ERRNO_CODES.EINVAL;
        }
        FS.mknod(path, mode, dev);
        return 0;
      },doReadlink:function (path, buf, bufsize) {
        if (bufsize <= 0) return -ERRNO_CODES.EINVAL;
        var ret = FS.readlink(path);
        ret = ret.slice(0, Math.max(0, bufsize));
        writeStringToMemory(ret, buf, true);
        return ret.length;
      },doAccess:function (path, amode) {
        if (amode & ~7) {
          // need a valid mode
          return -ERRNO_CODES.EINVAL;
        }
        var node;
        var lookup = FS.lookupPath(path, { follow: true });
        node = lookup.node;
        var perms = '';
        if (amode & 4) perms += 'r';
        if (amode & 2) perms += 'w';
        if (amode & 1) perms += 'x';
        if (perms /* otherwise, they've just passed F_OK */ && FS.nodePermissions(node, perms)) {
          return -ERRNO_CODES.EACCES;
        }
        return 0;
      },doDup:function (path, flags, suggestFD) {
        var suggest = FS.getStream(suggestFD);
        if (suggest) FS.close(suggest);
        return FS.open(path, flags, 0, suggestFD, suggestFD).fd;
      },doReadv:function (stream, iov, iovcnt, offset) {
        var ret = 0;
        for (var i = 0; i < iovcnt; i++) {
          var ptr = HEAP32[(((iov)+(i*8))>>2)];
          var len = HEAP32[(((iov)+(i*8 + 4))>>2)];
          var curr = FS.read(stream, HEAP8,ptr, len, offset);
          if (curr < 0) return -1;
          ret += curr;
          if (curr < len) break; // nothing more to read
        }
        return ret;
      },doWritev:function (stream, iov, iovcnt, offset) {
        var ret = 0;
        for (var i = 0; i < iovcnt; i++) {
          var ptr = HEAP32[(((iov)+(i*8))>>2)];
          var len = HEAP32[(((iov)+(i*8 + 4))>>2)];
          var curr = FS.write(stream, HEAP8,ptr, len, offset);
          if (curr < 0) return -1;
          ret += curr;
        }
        return ret;
      },varargs:0,get:function (varargs) {
        SYSCALLS.varargs += 4;
        var ret = HEAP32[(((SYSCALLS.varargs)-(4))>>2)];
        return ret;
      },getStr:function () {
        var ret = Pointer_stringify(SYSCALLS.get());
        return ret;
      },getStreamFromFD:function () {
        var stream = FS.getStream(SYSCALLS.get());
        if (!stream) throw new FS.ErrnoError(ERRNO_CODES.EBADF);
        return stream;
      },getSocketFromFD:function () {
        var socket = SOCKFS.getSocket(SYSCALLS.get());
        if (!socket) throw new FS.ErrnoError(ERRNO_CODES.EBADF);
        return socket;
      },getSocketAddress:function (allowNull) {
        var addrp = SYSCALLS.get(), addrlen = SYSCALLS.get();
        if (allowNull && addrp === 0) return null;
        var info = __read_sockaddr(addrp, addrlen);
        if (info.errno) throw new FS.ErrnoError(info.errno);
        info.addr = DNS.lookup_addr(info.addr) || info.addr;
        return info;
      },get64:function () {
        var low = SYSCALLS.get(), high = SYSCALLS.get();
        if (low >= 0) assert(high === 0);
        else assert(high === -1);
        return low;
      },getZero:function () {
        assert(SYSCALLS.get() === 0);
      }};function ___syscall118(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // fsync
      var stream = SYSCALLS.getStreamFromFD();
      return 0; // we can't do anything synchronously; the in-memory FS is already synced to
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function _emscripten_glVertexPointer(){ throw 'Legacy GL function (glVertexPointer) called. If you want legacy GL emulation, you need to compile with -s LEGACY_GL_EMULATION=1 to enable legacy GL emulation.'; }

  function _emscripten_glUniform3iv(location, count, value) {
      location = GL.uniforms[location];
      count *= 3;
      value = HEAP32.subarray((value)>>2,(value+count*4)>>2);
      GLctx.uniform3iv(location, value);
    }

  
  var _emscripten_GetProcAddress=undefined;
  Module["_emscripten_GetProcAddress"] = _emscripten_GetProcAddress;function _eglGetProcAddress(name_) {
      return _emscripten_GetProcAddress(name_);
    }

  function _emscripten_glShaderSource(shader, count, string, length) {
      var source = GL.getSource(shader, count, string, length);
      GLctx.shaderSource(GL.shaders[shader], source);
    }

  function _pthread_mutex_init() {}

  var _llvm_pow_f32=Math_pow;

  
  
  var _tzname=STATICTOP; STATICTOP += 16;;
  
  var _daylight=STATICTOP; STATICTOP += 16;;
  
  var _timezone=STATICTOP; STATICTOP += 16;;function _tzset() {
      // TODO: Use (malleable) environment variables instead of system settings.
      if (_tzset.called) return;
      _tzset.called = true;
  
      HEAP32[((_timezone)>>2)]=-(new Date()).getTimezoneOffset() * 60;
  
      var winter = new Date(2000, 0, 1);
      var summer = new Date(2000, 6, 1);
      HEAP32[((_daylight)>>2)]=Number(winter.getTimezoneOffset() != summer.getTimezoneOffset());
  
      function extractZone(date) {
        var match = date.toTimeString().match(/\(([A-Za-z ]+)\)$/);
        return match ? match[1] : "GMT";
      };
      var winterName = extractZone(winter);
      var summerName = extractZone(summer);
      var winterNamePtr = allocate(intArrayFromString(winterName), 'i8', ALLOC_NORMAL);
      var summerNamePtr = allocate(intArrayFromString(summerName), 'i8', ALLOC_NORMAL);
      if (summer.getTimezoneOffset() < winter.getTimezoneOffset()) {
        // Northern hemisphere
        HEAP32[((_tzname)>>2)]=winterNamePtr;
        HEAP32[(((_tzname)+(4))>>2)]=summerNamePtr;
      } else {
        HEAP32[((_tzname)>>2)]=summerNamePtr;
        HEAP32[(((_tzname)+(4))>>2)]=winterNamePtr;
      }
    }function _mktime(tmPtr) {
      _tzset();
      var date = new Date(HEAP32[(((tmPtr)+(20))>>2)] + 1900,
                          HEAP32[(((tmPtr)+(16))>>2)],
                          HEAP32[(((tmPtr)+(12))>>2)],
                          HEAP32[(((tmPtr)+(8))>>2)],
                          HEAP32[(((tmPtr)+(4))>>2)],
                          HEAP32[((tmPtr)>>2)],
                          0);
  
      // There's an ambiguous hour when the time goes back; the tm_isdst field is
      // used to disambiguate it.  Date() basically guesses, so we fix it up if it
      // guessed wrong, or fill in tm_isdst with the guess if it's -1.
      var dst = HEAP32[(((tmPtr)+(32))>>2)];
      var guessedOffset = date.getTimezoneOffset();
      var start = new Date(date.getFullYear(), 0, 1);
      var summerOffset = new Date(2000, 6, 1).getTimezoneOffset();
      var winterOffset = start.getTimezoneOffset();
      var dstOffset = Math.min(winterOffset, summerOffset); // DST is in December in South
      if (dst < 0) {
        HEAP32[(((tmPtr)+(32))>>2)]=Number(dstOffset == guessedOffset);
      } else if ((dst > 0) != (dstOffset == guessedOffset)) {
        var nonDstOffset = Math.max(winterOffset, summerOffset);
        var trueOffset = dst > 0 ? dstOffset : nonDstOffset;
        // Don't try setMinutes(date.getMinutes() + ...) -- it's messed up.
        date.setTime(date.getTime() + (trueOffset - guessedOffset)*60000);
      }
  
      HEAP32[(((tmPtr)+(24))>>2)]=date.getDay();
      var yday = ((date.getTime() - start.getTime()) / (1000 * 60 * 60 * 24))|0;
      HEAP32[(((tmPtr)+(28))>>2)]=yday;
  
      return (date.getTime() / 1000)|0;
    }

  function _emscripten_glIsTexture(texture) {
      var texture = GL.textures[texture];
      if (!texture) return 0;
      return GLctx.isTexture(texture);
    }

  function _emscripten_glTexParameterf(x0, x1, x2) { GLctx.texParameterf(x0, x1, x2) }

  
  var DLFCN={error:null,errorMsg:null,loadedLibs:{},loadedLibNames:{}};function _dlerror() {
      // char *dlerror(void);
      // http://pubs.opengroup.org/onlinepubs/009695399/functions/dlerror.html
      if (DLFCN.errorMsg === null) {
        return 0;
      } else {
        if (DLFCN.error) _free(DLFCN.error);
        var msgArr = intArrayFromString(DLFCN.errorMsg);
        DLFCN.error = allocate(msgArr, 'i8', ALLOC_NORMAL);
        DLFCN.errorMsg = null;
        return DLFCN.error;
      }
    }

  var _llvm_sqrt_f64=Math_sqrt;

  function _eglWaitGL() {
  return _eglWaitClient.apply(null, arguments)
  }

  function ___syscall54(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // ioctl
      var stream = SYSCALLS.getStreamFromFD(), op = SYSCALLS.get();
      switch (op) {
        case 21505: {
          if (!stream.tty) return -ERRNO_CODES.ENOTTY;
          return 0;
        }
        case 21506: {
          if (!stream.tty) return -ERRNO_CODES.ENOTTY;
          return 0; // no-op, not actually adjusting terminal settings
        }
        case 21519: {
          if (!stream.tty) return -ERRNO_CODES.ENOTTY;
          var argp = SYSCALLS.get();
          HEAP32[((argp)>>2)]=0;
          return 0;
        }
        case 21520: {
          if (!stream.tty) return -ERRNO_CODES.ENOTTY;
          return -ERRNO_CODES.EINVAL; // not supported
        }
        case 21531: {
          var argp = SYSCALLS.get();
          return FS.ioctl(stream, op, argp);
        }
        default: abort('bad ioctl syscall ' + op);
      }
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  
  var ___tm_current=STATICTOP; STATICTOP += 48;;
  
  
  var ___tm_timezone=allocate(intArrayFromString("GMT"), "i8", ALLOC_STATIC);function _localtime_r(time, tmPtr) {
      _tzset();
      var date = new Date(HEAP32[((time)>>2)]*1000);
      HEAP32[((tmPtr)>>2)]=date.getSeconds();
      HEAP32[(((tmPtr)+(4))>>2)]=date.getMinutes();
      HEAP32[(((tmPtr)+(8))>>2)]=date.getHours();
      HEAP32[(((tmPtr)+(12))>>2)]=date.getDate();
      HEAP32[(((tmPtr)+(16))>>2)]=date.getMonth();
      HEAP32[(((tmPtr)+(20))>>2)]=date.getFullYear()-1900;
      HEAP32[(((tmPtr)+(24))>>2)]=date.getDay();
  
      var start = new Date(date.getFullYear(), 0, 1);
      var yday = ((date.getTime() - start.getTime()) / (1000 * 60 * 60 * 24))|0;
      HEAP32[(((tmPtr)+(28))>>2)]=yday;
      HEAP32[(((tmPtr)+(36))>>2)]=-(date.getTimezoneOffset() * 60);
  
      // DST is in December in South
      var summerOffset = new Date(2000, 6, 1).getTimezoneOffset();
      var winterOffset = start.getTimezoneOffset();
      var dst = (date.getTimezoneOffset() == Math.min(winterOffset, summerOffset))|0;
      HEAP32[(((tmPtr)+(32))>>2)]=dst;
  
      var zonePtr = HEAP32[(((_tzname)+(dst ? Runtime.QUANTUM_SIZE : 0))>>2)];
      HEAP32[(((tmPtr)+(40))>>2)]=zonePtr;
  
      return tmPtr;
    }function _localtime(time) {
      return _localtime_r(time, ___tm_current);
    }

  function _eglSwapBuffers() {
  
      if (!EGL.defaultDisplayInitialized) {
        EGL.setErrorCode(0x3001 /* EGL_NOT_INITIALIZED */);
      } else if (!Module.ctx) {
        EGL.setErrorCode(0x3002 /* EGL_BAD_ACCESS */);
      } else if (Module.ctx.isContextLost()) {
        EGL.setErrorCode(0x300E /* EGL_CONTEXT_LOST */);
      } else {
        // According to documentation this does an implicit flush.
        // Due to discussion at https://github.com/kripken/emscripten/pull/1871
        // the flush was removed since this _may_ result in slowing code down.
        //_glFlush();
        EGL.setErrorCode(0x3000 /* EGL_SUCCESS */);
        return 1 /* EGL_TRUE */;
      }
      return 0 /* EGL_FALSE */;
    }

  function _emscripten_glFrustum() {
  Module['printErr']('missing function: emscripten_glFrustum'); abort(-1);
  }

  function _emscripten_glGetTexParameterfv(target, pname, params) {
      if (!params) {
        // GLES2 specification does not specify how to behave if params is a null pointer. Since calling this function does not make sense
        // if p == null, issue a GL error to notify user about it. 
        GL.recordError(0x0501 /* GL_INVALID_VALUE */);
        return;
      }
      HEAPF32[((params)>>2)]=GLctx.getTexParameter(target, pname);
    }

  function _emscripten_glUniform4i(location, v0, v1, v2, v3) {
      location = GL.uniforms[location];
      GLctx.uniform4i(location, v0, v1, v2, v3);
    }

  function _emscripten_glBindRenderbuffer(target, renderbuffer) {
      GLctx.bindRenderbuffer(target, renderbuffer ? GL.renderbuffers[renderbuffer] : null);
    }

  function _emscripten_glViewport(x0, x1, x2, x3) { GLctx.viewport(x0, x1, x2, x3) }

  
  var AL={contexts:[],currentContext:null,alcErr:0,stringCache:{},alcStringCache:{},QUEUE_INTERVAL:25,QUEUE_LOOKAHEAD:100,newSrcId:1,updateSources:function updateSources(context) {
        // If we are animating using the requestAnimationFrame method, then the main loop does not run when in the background.
        // To give a perfect glitch-free audio stop when switching from foreground to background, we need to avoid updating
        // audio altogether when in the background, so detect that case and kill audio buffer streaming if so.
        if (Browser.mainLoop.timingMode == 1/*EM_TIMING_RAF*/ && document['visibilityState'] != 'visible') return;
  
        for (var srcId in context.src) {
          AL.updateSource(context.src[srcId]);
        }
      },updateSource:function updateSource(src) {
        if (src.state !== 0x1012 /* AL_PLAYING */) {
          return;
        }
  
        var currentTime = AL.currentContext.ctx.currentTime;
        var startTime = src.bufferPosition;
  
        for (var i = src.buffersPlayed; i < src.queue.length; i++) {
          var entry = src.queue[i];
  
          var startOffset = startTime - currentTime;
          var endTime = startTime + entry.buffer.duration;
  
          // Clean up old buffers.
          if (currentTime >= endTime) {
            // Update our location in the queue.
            src.bufferPosition = endTime;
            src.buffersPlayed = i + 1;
  
            // Stop / restart the source when we hit the end.
            if (src.buffersPlayed >= src.queue.length) {
              if (src.loop) {
                AL.setSourceState(src, 0x1012 /* AL_PLAYING */);
              } else {
                AL.setSourceState(src, 0x1014 /* AL_STOPPED */);
              }
            }
          }
          // Process all buffers that'll be played before the next tick.
          else if (startOffset < (AL.QUEUE_LOOKAHEAD / 1000) && !entry.src) {
            // If the start offset is negative, we need to offset the actual buffer.
            var offset = Math.abs(Math.min(startOffset, 0));
  
            entry.src = AL.currentContext.ctx.createBufferSource();
            entry.src.buffer = entry.buffer;
            entry.src.connect(src.gain);
            if (typeof(entry.src.start) !== 'undefined') {
              entry.src.start(startTime, offset);
            } else if (typeof(entry.src.noteOn) !== 'undefined') {
              entry.src.noteOn(startTime);
            }
          }
  
          startTime = endTime;
        }
      },setSourceState:function setSourceState(src, state) {
        if (state === 0x1012 /* AL_PLAYING */) {
          if (src.state !== 0x1013 /* AL_PAUSED */) {
            src.state = 0x1012 /* AL_PLAYING */;
            // Reset our position.
            src.bufferPosition = AL.currentContext.ctx.currentTime;
            src.buffersPlayed = 0;
          } else {
            src.state = 0x1012 /* AL_PLAYING */;
            // Use the current offset from src.bufferPosition to resume at the correct point.
            src.bufferPosition = AL.currentContext.ctx.currentTime - src.bufferPosition;
          }
          AL.stopSourceQueue(src);
          AL.updateSource(src);
        } else if (state === 0x1013 /* AL_PAUSED */) {
          if (src.state === 0x1012 /* AL_PLAYING */) {
            src.state = 0x1013 /* AL_PAUSED */;
            // Store off the current offset to restore with on resume.
            src.bufferPosition = AL.currentContext.ctx.currentTime - src.bufferPosition;
            AL.stopSourceQueue(src);
          }
        } else if (state === 0x1014 /* AL_STOPPED */) {
          if (src.state !== 0x1011 /* AL_INITIAL */) {
            src.state = 0x1014 /* AL_STOPPED */;
            src.buffersPlayed = src.queue.length;
            AL.stopSourceQueue(src);
          }
        } else if (state == 0x1011 /* AL_INITIAL */) {
          if (src.state !== 0x1011 /* AL_INITIAL */) {
            src.state = 0x1011 /* AL_INITIAL */;
            src.bufferPosition = 0;
            src.buffersPlayed = 0;
          }
        }
      },stopSourceQueue:function stopSourceQueue(src) {
        for (var i = 0; i < src.queue.length; i++) {
          var entry = src.queue[i];
          if (entry.src) {
            entry.src.stop(0);
            entry.src = null;
          }
        }
      }};function _alcGetIntegerv(device, param, size, data) {
      if (size == 0 || !data) {
        AL.currentContext.err = 0xA003 /* AL_INVALID_VALUE */;
        return;
      }
  
      switch(param) {
      case 0x1000 /* ALC_MAJOR_VERSION */:
        HEAP32[((data)>>2)]=1;
        break;
      case 0x1001 /* ALC_MINOR_VERSION */:
        HEAP32[((data)>>2)]=1;
        break;
      case 0x1002 /* ALC_ATTRIBUTES_SIZE */:
        if (!device) {
          AL.alcErr = 0xA001 /* ALC_INVALID_DEVICE */;
          return 0;
        }
        HEAP32[((data)>>2)]=1;
        break;
      case 0x1003 /* ALC_ALL_ATTRIBUTES */:
        if (!device) {
          AL.alcErr = 0xA001 /* ALC_INVALID_DEVICE */;
          return 0;
        }
        HEAP32[((data)>>2)]=0;
        break;
      case 0x20003 /* ALC_MAX_AUXILIARY_SENDS */:
        if (!device) {
          AL.currentContext.err = 0xA001 /* ALC_INVALID_DEVICE */;
          return 0;
        }
        HEAP32[((data)>>2)]=1;
      default:
        AL.alcErr = 0xA003 /* ALC_INVALID_ENUM */;
        break;
      }
    }

  function _dlclose(handle) {
      // int dlclose(void *handle);
      // http://pubs.opengroup.org/onlinepubs/009695399/functions/dlclose.html
      if (!DLFCN.loadedLibs[handle]) {
        DLFCN.errorMsg = 'Tried to dlclose() unopened handle: ' + handle;
        return 1;
      } else {
        var lib_record = DLFCN.loadedLibs[handle];
        if (--lib_record.refcount == 0) {
          if (lib_record.module.cleanups) {
            lib_record.module.cleanups.forEach(function(cleanup) { cleanup() });
          }
          delete DLFCN.loadedLibNames[lib_record.name];
          delete DLFCN.loadedLibs[handle];
        }
        return 0;
      }
    }

  function _emscripten_get_gamepad_status(index, gamepadState) {
      if (!navigator.getGamepads && !navigator.webkitGetGamepads) return -1;
      var gamepads;
      if (navigator.getGamepads) {
        gamepads = navigator.getGamepads();
      } else if (navigator.webkitGetGamepads) {
        gamepads = navigator.webkitGetGamepads();
      }
      if (index < 0 || index >= gamepads.length) {
        return -5;
      }
      // For previously disconnected gamepads there should be a null at the index.
      // This is because gamepads must keep their original position in the array.
      // For example, removing the first of two gamepads produces [null, gamepad].
      // Older implementations of the Gamepad API used undefined instead of null.
      // The following check works because null and undefined evaluate to false.
      if (!gamepads[index]) {
        // There is a "false" but no gamepad at index because it was disconnected.
        return -7;
      }
      // There should be a gamepad at index which can be queried.
      JSEvents.fillGamepadEventData(gamepadState, gamepads[index]);
      return 0;
    }

  var _llvm_pow_f64=Math_pow;

  function _emscripten_glCopyTexImage2D(x0, x1, x2, x3, x4, x5, x6, x7) { GLctx.copyTexImage2D(x0, x1, x2, x3, x4, x5, x6, x7) }

  function _emscripten_glTexParameterfv(target, pname, params) {
      var param = HEAPF32[((params)>>2)];
      GLctx.texParameterf(target, pname, param);
    }

  function _emscripten_glDepthRangef(x0, x1) { GLctx.depthRange(x0, x1) }

  var _llvm_fabs_f32=Math_abs;

  function _emscripten_glUniform3f(location, v0, v1, v2) {
      location = GL.uniforms[location];
      GLctx.uniform3f(location, v0, v1, v2);
    }

  function ___resumeException(ptr) {
      if (!EXCEPTIONS.last) { EXCEPTIONS.last = ptr; }
      EXCEPTIONS.clearRef(EXCEPTIONS.deAdjust(ptr)); // exception refcount should be cleared, but don't free it
      throw ptr;
    }

  function _emscripten_glGetObjectParameterivARB() {
  Module['printErr']('missing function: emscripten_glGetObjectParameterivARB'); abort(-1);
  }

  function _emscripten_glBlendFunc(x0, x1) { GLctx.blendFunc(x0, x1) }

  function _emscripten_glUniform3i(location, v0, v1, v2) {
      location = GL.uniforms[location];
      GLctx.uniform3i(location, v0, v1, v2);
    }

  function _emscripten_glStencilOp(x0, x1, x2) { GLctx.stencilOp(x0, x1, x2) }

  function _alcDestroyContext(context) {
      // Stop playback, etc
      clearInterval(AL.contexts[context - 1].interval);
    }

  function _emscripten_glBindAttribLocation(program, index, name) {
      name = Pointer_stringify(name);
      GLctx.bindAttribLocation(GL.programs[program], index, name);
    }

  
  var PROCINFO={ppid:1,pid:42,sid:42,pgid:42};function ___syscall20(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // getpid
      return PROCINFO.pid;
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function _emscripten_glEnableVertexAttribArray(index) {
      var cb = GL.currentContext.clientBuffers[index];
      assert(cb, index);
      cb.enabled = true;
      GLctx.enableVertexAttribArray(index);
    }

   
  Module["_memset"] = _memset;

  function _emscripten_glUniform1i(location, v0) {
      location = GL.uniforms[location];
      GLctx.uniform1i(location, v0);
    }

  function _alDeleteBuffers(count, buffers)
    {
      if (!AL.currentContext) {
        return;
      }
      if (count > AL.currentContext.buf.length) {
        AL.currentContext.err = 0xA003 /* AL_INVALID_VALUE */;
        return;
      }
  
      for (var i = 0; i < count; ++i) {
        var bufferIdx = HEAP32[(((buffers)+(i*4))>>2)] - 1;
  
        // Make sure the buffer index is valid.
        if (bufferIdx >= AL.currentContext.buf.length || !AL.currentContext.buf[bufferIdx]) {
          AL.currentContext.err = 0xA001 /* AL_INVALID_NAME */;
          return;
        }
  
        // Make sure the buffer is no longer in use.
        var buffer = AL.currentContext.buf[bufferIdx];
        for (var srcId in AL.currentContext.src) {
          var src = AL.currentContext.src[srcId];
          if (!src) {
            continue;
          }
          for (var k = 0; k < src.queue.length; k++) {
            if (buffer === src.queue[k].buffer) {
              AL.currentContext.err = 0xA004 /* AL_INVALID_OPERATION */;
              return;
            }
          }
        }
      }
  
      for (var i = 0; i < count; ++i) {
        var bufferIdx = HEAP32[(((buffers)+(i*4))>>2)] - 1;
        delete AL.currentContext.buf[bufferIdx];
      }
    }

  function _atexit(func, arg) {
      __ATEXIT__.unshift({ func: func, arg: arg });
    }


  function _emscripten_set_touchcancel_callback(target, userData, useCapture, callbackfunc) {
      JSEvents.registerTouchEventCallback(target, userData, useCapture, callbackfunc, 25, "touchcancel");
      return 0;
    }

  
  
  var GLUT={initTime:null,idleFunc:null,displayFunc:null,keyboardFunc:null,keyboardUpFunc:null,specialFunc:null,specialUpFunc:null,reshapeFunc:null,motionFunc:null,passiveMotionFunc:null,mouseFunc:null,buttons:0,modifiers:0,initWindowWidth:256,initWindowHeight:256,initDisplayMode:18,windowX:0,windowY:0,windowWidth:0,windowHeight:0,requestedAnimationFrame:false,saveModifiers:function (event) {
        GLUT.modifiers = 0;
        if (event['shiftKey'])
          GLUT.modifiers += 1; /* GLUT_ACTIVE_SHIFT */
        if (event['ctrlKey'])
          GLUT.modifiers += 2; /* GLUT_ACTIVE_CTRL */
        if (event['altKey'])
          GLUT.modifiers += 4; /* GLUT_ACTIVE_ALT */
      },onMousemove:function (event) {
        /* Send motion event only if the motion changed, prevents
         * spamming our app with uncessary callback call. It does happen in
         * Chrome on Windows.
         */
        var lastX = Browser.mouseX;
        var lastY = Browser.mouseY;
        Browser.calculateMouseEvent(event);
        var newX = Browser.mouseX;
        var newY = Browser.mouseY;
        if (newX == lastX && newY == lastY) return;
  
        if (GLUT.buttons == 0 && event.target == Module["canvas"] && GLUT.passiveMotionFunc) {
          event.preventDefault();
          GLUT.saveModifiers(event);
          Runtime.dynCall('vii', GLUT.passiveMotionFunc, [lastX, lastY]);
        } else if (GLUT.buttons != 0 && GLUT.motionFunc) {
          event.preventDefault();
          GLUT.saveModifiers(event);
          Runtime.dynCall('vii', GLUT.motionFunc, [lastX, lastY]);
        }
      },getSpecialKey:function (keycode) {
          var key = null;
          switch (keycode) {
            case 8:  key = 120 /* backspace */; break;
            case 46: key = 111 /* delete */; break;
  
            case 0x70 /*DOM_VK_F1*/: key = 1 /* GLUT_KEY_F1 */; break;
            case 0x71 /*DOM_VK_F2*/: key = 2 /* GLUT_KEY_F2 */; break;
            case 0x72 /*DOM_VK_F3*/: key = 3 /* GLUT_KEY_F3 */; break;
            case 0x73 /*DOM_VK_F4*/: key = 4 /* GLUT_KEY_F4 */; break;
            case 0x74 /*DOM_VK_F5*/: key = 5 /* GLUT_KEY_F5 */; break;
            case 0x75 /*DOM_VK_F6*/: key = 6 /* GLUT_KEY_F6 */; break;
            case 0x76 /*DOM_VK_F7*/: key = 7 /* GLUT_KEY_F7 */; break;
            case 0x77 /*DOM_VK_F8*/: key = 8 /* GLUT_KEY_F8 */; break;
            case 0x78 /*DOM_VK_F9*/: key = 9 /* GLUT_KEY_F9 */; break;
            case 0x79 /*DOM_VK_F10*/: key = 10 /* GLUT_KEY_F10 */; break;
            case 0x7a /*DOM_VK_F11*/: key = 11 /* GLUT_KEY_F11 */; break;
            case 0x7b /*DOM_VK_F12*/: key = 12 /* GLUT_KEY_F12 */; break;
            case 0x25 /*DOM_VK_LEFT*/: key = 100 /* GLUT_KEY_LEFT */; break;
            case 0x26 /*DOM_VK_UP*/: key = 101 /* GLUT_KEY_UP */; break;
            case 0x27 /*DOM_VK_RIGHT*/: key = 102 /* GLUT_KEY_RIGHT */; break;
            case 0x28 /*DOM_VK_DOWN*/: key = 103 /* GLUT_KEY_DOWN */; break;
            case 0x21 /*DOM_VK_PAGE_UP*/: key = 104 /* GLUT_KEY_PAGE_UP */; break;
            case 0x22 /*DOM_VK_PAGE_DOWN*/: key = 105 /* GLUT_KEY_PAGE_DOWN */; break;
            case 0x24 /*DOM_VK_HOME*/: key = 106 /* GLUT_KEY_HOME */; break;
            case 0x23 /*DOM_VK_END*/: key = 107 /* GLUT_KEY_END */; break;
            case 0x2d /*DOM_VK_INSERT*/: key = 108 /* GLUT_KEY_INSERT */; break;
  
            case 16   /*DOM_VK_SHIFT*/:
            case 0x05 /*DOM_VK_LEFT_SHIFT*/:
              key = 112 /* GLUT_KEY_SHIFT_L */;
              break;
            case 0x06 /*DOM_VK_RIGHT_SHIFT*/:
              key = 113 /* GLUT_KEY_SHIFT_R */;
              break;
  
            case 17   /*DOM_VK_CONTROL*/:
            case 0x03 /*DOM_VK_LEFT_CONTROL*/:
              key = 114 /* GLUT_KEY_CONTROL_L */;
              break;
            case 0x04 /*DOM_VK_RIGHT_CONTROL*/:
              key = 115 /* GLUT_KEY_CONTROL_R */;
              break;
  
            case 18   /*DOM_VK_ALT*/:
            case 0x02 /*DOM_VK_LEFT_ALT*/:
              key = 116 /* GLUT_KEY_ALT_L */;
              break;
            case 0x01 /*DOM_VK_RIGHT_ALT*/:
              key = 117 /* GLUT_KEY_ALT_R */;
              break;
          };
          return key;
      },getASCIIKey:function (event) {
        if (event['ctrlKey'] || event['altKey'] || event['metaKey']) return null;
  
        var keycode = event['keyCode'];
  
        /* The exact list is soooo hard to find in a canonical place! */
  
        if (48 <= keycode && keycode <= 57)
          return keycode; // numeric  TODO handle shift?
        if (65 <= keycode && keycode <= 90)
          return event['shiftKey'] ? keycode : keycode + 32;
        if (96 <= keycode && keycode <= 105)
          return keycode - 48; // numpad numbers    
        if (106 <= keycode && keycode <= 111)
          return keycode - 106 + 42; // *,+-./  TODO handle shift?
  
        switch (keycode) {
          case 9:  // tab key
          case 13: // return key
          case 27: // escape
          case 32: // space
          case 61: // equal
            return keycode;
        }
  
        var s = event['shiftKey'];
        switch (keycode) {
          case 186: return s ? 58 : 59; // colon / semi-colon
          case 187: return s ? 43 : 61; // add / equal (these two may be wrong)
          case 188: return s ? 60 : 44; // less-than / comma
          case 189: return s ? 95 : 45; // dash
          case 190: return s ? 62 : 46; // greater-than / period
          case 191: return s ? 63 : 47; // forward slash
          case 219: return s ? 123 : 91; // open bracket
          case 220: return s ? 124 : 47; // back slash
          case 221: return s ? 125 : 93; // close braket
          case 222: return s ? 34 : 39; // single quote
        }
  
        return null;
      },onKeydown:function (event) {
        if (GLUT.specialFunc || GLUT.keyboardFunc) {
          var key = GLUT.getSpecialKey(event['keyCode']);
          if (key !== null) {
            if( GLUT.specialFunc ) {
              event.preventDefault();
              GLUT.saveModifiers(event);
              Runtime.dynCall('viii', GLUT.specialFunc, [key, Browser.mouseX, Browser.mouseY]);
            }
          }
          else
          {
            key = GLUT.getASCIIKey(event);
            if( key !== null && GLUT.keyboardFunc ) {
              event.preventDefault();
              GLUT.saveModifiers(event);
              Runtime.dynCall('viii', GLUT.keyboardFunc, [key, Browser.mouseX, Browser.mouseY]);
            }
          }
        }
      },onKeyup:function (event) {
        if (GLUT.specialUpFunc || GLUT.keyboardUpFunc) {
          var key = GLUT.getSpecialKey(event['keyCode']);
          if (key !== null) {
            if(GLUT.specialUpFunc) {
              event.preventDefault ();
              GLUT.saveModifiers(event);
              Runtime.dynCall('viii', GLUT.specialUpFunc, [key, Browser.mouseX, Browser.mouseY]);
            }
          }
          else
          {
            key = GLUT.getASCIIKey(event);
            if( key !== null && GLUT.keyboardUpFunc ) {
              event.preventDefault ();
              GLUT.saveModifiers(event);
              Runtime.dynCall('viii', GLUT.keyboardUpFunc, [key, Browser.mouseX, Browser.mouseY]);
            }
          }
        }
      },onMouseButtonDown:function (event) {
        Browser.calculateMouseEvent(event);
  
        GLUT.buttons |= (1 << event['button']);
  
        if (event.target == Module["canvas"] && GLUT.mouseFunc) {
          try {
            event.target.setCapture();
          } catch (e) {}
          event.preventDefault();
          GLUT.saveModifiers(event);
          Runtime.dynCall('viiii', GLUT.mouseFunc, [event['button'], 0/*GLUT_DOWN*/, Browser.mouseX, Browser.mouseY]);
        }
      },onMouseButtonUp:function (event) {
        Browser.calculateMouseEvent(event);
  
        GLUT.buttons &= ~(1 << event['button']);
  
        if (GLUT.mouseFunc) {
          event.preventDefault();
          GLUT.saveModifiers(event);
          Runtime.dynCall('viiii', GLUT.mouseFunc, [event['button'], 1/*GLUT_UP*/, Browser.mouseX, Browser.mouseY]);
        }
      },onMouseWheel:function (event) {
        Browser.calculateMouseEvent(event);
  
        // cross-browser wheel delta
        var e = window.event || event; // old IE support
        // Note the minus sign that flips browser wheel direction (positive direction scrolls page down) to native wheel direction (positive direction is mouse wheel up)
        var delta = -Browser.getMouseWheelDelta(event);
        delta = (delta == 0) ? 0 : (delta > 0 ? Math.max(delta, 1) : Math.min(delta, -1)); // Quantize to integer so that minimum scroll is at least +/- 1.
  
        var button = 3; // wheel up
        if (delta < 0) {
          button = 4; // wheel down
        }
  
        if (GLUT.mouseFunc) {
          event.preventDefault();
          GLUT.saveModifiers(event);
          Runtime.dynCall('viiii', GLUT.mouseFunc, [button, 0/*GLUT_DOWN*/, Browser.mouseX, Browser.mouseY]);
        }
      },onFullScreenEventChange:function (event) {
        var width;
        var height;
        if (document["fullScreen"] || document["mozFullScreen"] || document["webkitIsFullScreen"]) {
          width = screen["width"];
          height = screen["height"];
        } else {
          width = GLUT.windowWidth;
          height = GLUT.windowHeight;
          // TODO set position
          document.removeEventListener('fullscreenchange', GLUT.onFullScreenEventChange, true);
          document.removeEventListener('mozfullscreenchange', GLUT.onFullScreenEventChange, true);
          document.removeEventListener('webkitfullscreenchange', GLUT.onFullScreenEventChange, true);
        }
        Browser.setCanvasSize(width, height);
        /* Can't call _glutReshapeWindow as that requests cancelling fullscreen. */
        if (GLUT.reshapeFunc) {
          // console.log("GLUT.reshapeFunc (from FS): " + width + ", " + height);
          Runtime.dynCall('vii', GLUT.reshapeFunc, [width, height]);
        }
        _glutPostRedisplay();
      },requestFullScreen:function () {
        var RFS = Module["canvas"]['requestFullscreen'] ||
                  Module["canvas"]['requestFullScreen'] ||
                  Module["canvas"]['mozRequestFullScreen'] ||
                  Module["canvas"]['webkitRequestFullScreen'] ||
                  (function() {});
        RFS.apply(Module["canvas"], []);
      },cancelFullScreen:function () {
        var CFS = document['exitFullscreen'] ||
                  document['cancelFullScreen'] ||
                  document['mozCancelFullScreen'] ||
                  document['webkitCancelFullScreen'] ||
                  (function() {});
        CFS.apply(document, []);
      }};function _glutDestroyWindow(name) {
      Module.ctx = Browser.destroyContext(Module['canvas'], true, true);
      return 1;
    }function _eglDestroyContext(display, context) {
      if (display != 62000 /* Magic ID for Emscripten 'default display' */) {
        EGL.setErrorCode(0x3008 /* EGL_BAD_DISPLAY */);
        return 0;
      }
  
      if (context != 62004 /* Magic ID for Emscripten EGLContext */) {
        EGL.setErrorCode(0x3006 /* EGL_BAD_CONTEXT */);
        return 0;
      }
  
      EGL.setErrorCode(0x3000 /* EGL_SUCCESS */);
      return 1;
    }

  function ___lock() {}

  function _emscripten_glBlendFuncSeparate(x0, x1, x2, x3) { GLctx.blendFuncSeparate(x0, x1, x2, x3) }

  function _emscripten_glGetVertexAttribPointerv(index, pname, pointer) {
      if (!pointer) {
        // GLES2 specification does not specify how to behave if pointer is a null pointer. Since calling this function does not make sense
        // if pointer == null, issue a GL error to notify user about it. 
        GL.recordError(0x0501 /* GL_INVALID_VALUE */);
        return;
      }
      if (GL.currentContext.clientBuffers[index].enabled) {
        Module.printErr("glGetVertexAttribPointer on client-side array: not supported, bad data returned");
      }
      HEAP32[((pointer)>>2)]=GLctx.getVertexAttribOffset(index, pname);
    }

  function _emscripten_glVertexAttrib3f(x0, x1, x2, x3) { GLctx.vertexAttrib3f(x0, x1, x2, x3) }

  function _clock() {
      if (_clock.start === undefined) _clock.start = Date.now();
      return ((Date.now() - _clock.start) * (1000000 / 1000))|0;
    }

  function ___cxa_guard_abort() {}

  function _emscripten_glCopyTexSubImage2D(x0, x1, x2, x3, x4, x5, x6, x7) { GLctx.copyTexSubImage2D(x0, x1, x2, x3, x4, x5, x6, x7) }

  function _emscripten_glNormalPointer() {
  Module['printErr']('missing function: emscripten_glNormalPointer'); abort(-1);
  }

  function _emscripten_glTexParameteri(x0, x1, x2) { GLctx.texParameteri(x0, x1, x2) }

  function _emscripten_get_pointerlock_status(pointerlockStatus) {
      if (pointerlockStatus) JSEvents.fillPointerlockChangeEventData(pointerlockStatus);
      if (!document.body.requestPointerLock && !document.body.mozRequestPointerLock && !document.body.webkitRequestPointerLock && !document.body.msRequestPointerLock) {
        return -1;
      }
      return 0;
    }

  function _alSourcePlay(source) {
      if (!AL.currentContext) {
        return;
      }
      var src = AL.currentContext.src[source];
      if (!src) {
        AL.currentContext.err = 0xA001 /* AL_INVALID_NAME */;
        return;
      }
      AL.setSourceState(src, 0x1012 /* AL_PLAYING */);
    }

  function _eglSwapInterval(display, interval) {
      if (display != 62000 /* Magic ID for Emscripten 'default display' */) {
        EGL.setErrorCode(0x3008 /* EGL_BAD_DISPLAY */);
        return 0;
      }
      if (interval == 0) _emscripten_set_main_loop_timing(0/*EM_TIMING_SETTIMEOUT*/, 0);
      else _emscripten_set_main_loop_timing(1/*EM_TIMING_RAF*/, interval);
  
      EGL.setErrorCode(0x3000 /* EGL_SUCCESS */);
      return 1;
    }

  
  function emscriptenWebGLGetVertexAttrib(index, pname, params, type) {
      if (!params) {
        // GLES2 specification does not specify how to behave if params is a null pointer. Since calling this function does not make sense
        // if params == null, issue a GL error to notify user about it. 
        GL.recordError(0x0501 /* GL_INVALID_VALUE */);
        return;
      }
      if (GL.currentContext.clientBuffers[index].enabled) {
        Module.printErr("glGetVertexAttrib*v on client-side array: not supported, bad data returned");
      }
      var data = GLctx.getVertexAttrib(index, pname);
      if (typeof data == 'number' || typeof data == 'boolean') {
        switch (type) {
          case 'Integer': HEAP32[((params)>>2)]=data; break;
          case 'Float': HEAPF32[((params)>>2)]=data; break;
          case 'FloatToInteger': HEAP32[((params)>>2)]=Math.fround(data); break;
          default: throw 'internal emscriptenWebGLGetVertexAttrib() error, bad type: ' + type;
        }
      } else {
        for (var i = 0; i < data.length; i++) {
          switch (type) {
            case 'Integer': HEAP32[(((params)+(i))>>2)]=data[i]; break;
            case 'Float': HEAPF32[(((params)+(i))>>2)]=data[i]; break;
            case 'FloatToInteger': HEAP32[(((params)+(i))>>2)]=Math.fround(data[i]); break;
            default: throw 'internal emscriptenWebGLGetVertexAttrib() error, bad type: ' + type;
          }
        }
      }
    }function _emscripten_glGetVertexAttribfv(index, pname, params) {
      // N.B. This function may only be called if the vertex attribute was specified using the function glVertexAttrib*f(),
      // otherwise the results are undefined. (GLES3 spec 6.1.12)
      emscriptenWebGLGetVertexAttrib(index, pname, params, 'Float');
    }

  function _emscripten_set_touchstart_callback(target, userData, useCapture, callbackfunc) {
      JSEvents.registerTouchEventCallback(target, userData, useCapture, callbackfunc, 22, "touchstart");
      return 0;
    }

  function _emscripten_get_battery_status(batteryState) {
      if (!JSEvents.battery()) return -1; 
      JSEvents.fillBatteryEventData(batteryState, JSEvents.battery());
      return 0;
    }

  function ___cxa_guard_acquire(variable) {
      if (!HEAP8[((variable)>>0)]) { // ignore SAFE_HEAP stuff because llvm mixes i64 and i8 here
        HEAP8[((variable)>>0)]=1;
        return 1;
      }
      return 0;
    }

  function _emscripten_glDrawArraysInstanced(mode, first, count, primcount) {
      GLctx['drawArraysInstanced'](mode, first, count, primcount);
    }

  function _emscripten_glDeleteBuffers(n, buffers) {
      for (var i = 0; i < n; i++) {
        var id = HEAP32[(((buffers)+(i*4))>>2)];
        var buffer = GL.buffers[id];
  
        // From spec: "glDeleteBuffers silently ignores 0's and names that do not
        // correspond to existing buffer objects."
        if (!buffer) continue;
  
        GLctx.deleteBuffer(buffer);
        buffer.name = 0;
        GL.buffers[id] = null;
  
        if (id == GL.currArrayBuffer) GL.currArrayBuffer = 0;
        if (id == GL.currElementArrayBuffer) GL.currElementArrayBuffer = 0;
      }
    }

  function _emscripten_glUniformMatrix2fv(location, count, transpose, value) {
      location = GL.uniforms[location];
      var view;
      if (count === 1) {
        // avoid allocation for the common case of uploading one uniform matrix
        view = GL.miniTempBufferViews[3];
        for (var i = 0; i < 4; i++) {
          view[i] = HEAPF32[(((value)+(i*4))>>2)];
        }
      } else {
        view = HEAPF32.subarray((value)>>2,(value+count*16)>>2);
      }
      GLctx.uniformMatrix2fv(location, transpose, view);
    }

  function _emscripten_glDeleteShader(id) {
      if (!id) return;
      var shader = GL.shaders[id];
      if (!shader) { // glDeleteShader actually signals an error when deleting a nonexisting object, unlike some other GL delete functions.
        GL.recordError(0x0501 /* GL_INVALID_VALUE */);
        return;
      }
      GLctx.deleteShader(shader);
      GL.shaders[id] = null;
    }

  function ___syscall5(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // open
      var pathname = SYSCALLS.getStr(), flags = SYSCALLS.get(), mode = SYSCALLS.get() // optional TODO
      var stream = FS.open(pathname, flags, mode);
      return stream.fd;
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function ___syscall4(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // write
      var stream = SYSCALLS.getStreamFromFD(), buf = SYSCALLS.get(), count = SYSCALLS.get();
      return FS.write(stream, HEAP8,buf, count);
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function ___syscall6(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // close
      var stream = SYSCALLS.getStreamFromFD();
      FS.close(stream);
      return 0;
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function _getpwuid(uid) {
      return 0; // NULL
    }

  
  var cttz_i8 = allocate([8,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,7,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0], "i8", ALLOC_STATIC); 
  Module["_llvm_cttz_i32"] = _llvm_cttz_i32;function _llvm_cttz_i64(l, h) {
      var ret = _llvm_cttz_i32(l);
      if (ret == 32) ret += _llvm_cttz_i32(h);
      return ((asm["setTempRet0"](0),ret)|0);
    }

  function _alGetListenerf(pname, value) {
      if (!AL.currentContext) {
        return;
      }
      switch (pname) {
      case 0x100A /* AL_GAIN */:
        HEAPF32[((value)>>2)]=AL.currentContext.gain.gain.value
        break;
      default:
        AL.currentContext.err = 0xA002 /* AL_INVALID_ENUM */;
        break;
      }
  
    }

  function _emscripten_glGetVertexAttribiv(index, pname, params) {
      // N.B. This function may only be called if the vertex attribute was specified using the function glVertexAttrib*f(),
      // otherwise the results are undefined. (GLES3 spec 6.1.12)
      emscriptenWebGLGetVertexAttrib(index, pname, params, 'FloatToInteger');
    }

  function _emscripten_glUniformMatrix4fv(location, count, transpose, value) {
      location = GL.uniforms[location];
      var view;
      if (count === 1) {
        // avoid allocation for the common case of uploading one uniform matrix
        view = GL.miniTempBufferViews[15];
        for (var i = 0; i < 16; i++) {
          view[i] = HEAPF32[(((value)+(i*4))>>2)];
        }
      } else {
        view = HEAPF32.subarray((value)>>2,(value+count*64)>>2);
      }
      GLctx.uniformMatrix4fv(location, transpose, view);
    }

  function _emscripten_glEnableClientState() {
  Module['printErr']('missing function: emscripten_glEnableClientState'); abort(-1);
  }

  function ___syscall265(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // clock_nanosleep
      return 0;
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function _emscripten_glGetPointerv() {
  Module['printErr']('missing function: emscripten_glGetPointerv'); abort(-1);
  }

  function _alcCaptureSamples() {
  Module['printErr']('missing function: alcCaptureSamples'); abort(-1);
  }

  function _eglChooseConfig(display, attrib_list, configs, config_size, numConfigs) { 
      return EGL.chooseConfig(display, attrib_list, configs, config_size, numConfigs);
    }

  function ___syscall146(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // writev
      var stream = SYSCALLS.getStreamFromFD(), iov = SYSCALLS.get(), iovcnt = SYSCALLS.get();
      return SYSCALLS.doWritev(stream, iov, iovcnt);
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  
  function _alSource3f(source, param, v1, v2, v3) {
      if (!AL.currentContext) {
        return;
      }
      var src = AL.currentContext.src[source];
      if (!src) {
        AL.currentContext.err = 0xA001 /* AL_INVALID_NAME */;
        return;
      }
      switch (param) {
      case 0x1004 /* AL_POSITION */:
        src.position = [v1, v2, v3];
        break;
      case 0x1005 /* AL_DIRECTION */:
        src.direction = [v1, v2, v3];
        break;
      case 0x1006 /* AL_VELOCITY */:
        src.velocity = [v1, v2, v3];
        break;
      default:
        AL.currentContext.err = 0xA002 /* AL_INVALID_ENUM */;
        break;
      }
    }function _alSourcefv(source, param, value) {
      _alSource3f(source, param,
        HEAPF32[((value)>>2)],
        HEAPF32[(((value)+(4))>>2)],
        HEAPF32[(((value)+(8))>>2)]);
    }

  function ___syscall145(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // readv
      var stream = SYSCALLS.getStreamFromFD(), iov = SYSCALLS.get(), iovcnt = SYSCALLS.get();
      return SYSCALLS.doReadv(stream, iov, iovcnt);
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function ___cxa_atexit() {
  return _atexit.apply(null, arguments)
  }

  function _emscripten_glStencilFuncSeparate(x0, x1, x2, x3) { GLctx.stencilFuncSeparate(x0, x1, x2, x3) }

  function _eglGetConfigAttrib(display, config, attribute, value) {
      if (display != 62000 /* Magic ID for Emscripten 'default display' */) {
        EGL.setErrorCode(0x3008 /* EGL_BAD_DISPLAY */);
        return 0;
      }
      if (config != 62002 /* Magic ID for the only EGLConfig supported by Emscripten */) {
        EGL.setErrorCode(0x3005 /* EGL_BAD_CONFIG */);
        return 0;
      }
      if (!value) {
        EGL.setErrorCode(0x300C /* EGL_BAD_PARAMETER */);
        return 0;
      }
      EGL.setErrorCode(0x3000 /* EGL_SUCCESS */);
      switch(attribute) {
      case 0x3020: // EGL_BUFFER_SIZE
        HEAP32[((value)>>2)]=32;
        return 1;
      case 0x3021: // EGL_ALPHA_SIZE
        HEAP32[((value)>>2)]=8;
        return 1;
      case 0x3022: // EGL_BLUE_SIZE
        HEAP32[((value)>>2)]=8;
        return 1;
      case 0x3023: // EGL_GREEN_SIZE
        HEAP32[((value)>>2)]=8;
        return 1;
      case 0x3024: // EGL_RED_SIZE
        HEAP32[((value)>>2)]=8;
        return 1;
      case 0x3025: // EGL_DEPTH_SIZE
        HEAP32[((value)>>2)]=24;
        return 1;
      case 0x3026: // EGL_STENCIL_SIZE
        HEAP32[((value)>>2)]=8;
        return 1;
      case 0x3027: // EGL_CONFIG_CAVEAT
        // We can return here one of EGL_NONE (0x3038), EGL_SLOW_CONFIG (0x3050) or EGL_NON_CONFORMANT_CONFIG (0x3051).
        HEAP32[((value)>>2)]=0x3038;
        return 1;
      case 0x3028: // EGL_CONFIG_ID
        HEAP32[((value)>>2)]=62002;
        return 1;
      case 0x3029: // EGL_LEVEL
        HEAP32[((value)>>2)]=0;
        return 1;
      case 0x302A: // EGL_MAX_PBUFFER_HEIGHT
        HEAP32[((value)>>2)]=4096;
        return 1;
      case 0x302B: // EGL_MAX_PBUFFER_PIXELS
        HEAP32[((value)>>2)]=16777216;
        return 1;
      case 0x302C: // EGL_MAX_PBUFFER_WIDTH
        HEAP32[((value)>>2)]=4096;
        return 1;
      case 0x302D: // EGL_NATIVE_RENDERABLE
        HEAP32[((value)>>2)]=0;
        return 1;
      case 0x302E: // EGL_NATIVE_VISUAL_ID
        HEAP32[((value)>>2)]=0;
        return 1;
      case 0x302F: // EGL_NATIVE_VISUAL_TYPE
        HEAP32[((value)>>2)]=0x3038;
        return 1;
      case 0x3031: // EGL_SAMPLES
        HEAP32[((value)>>2)]=4;
        return 1;
      case 0x3032: // EGL_SAMPLE_BUFFERS
        HEAP32[((value)>>2)]=1;
        return 1;
      case 0x3033: // EGL_SURFACE_TYPE
        HEAP32[((value)>>2)]=0x0004;
        return 1;
      case 0x3034: // EGL_TRANSPARENT_TYPE
        // If this returns EGL_TRANSPARENT_RGB (0x3052), transparency is used through color-keying. No such thing applies to Emscripten canvas.
        HEAP32[((value)>>2)]=0x3038;
        return 1;
      case 0x3035: // EGL_TRANSPARENT_BLUE_VALUE
      case 0x3036: // EGL_TRANSPARENT_GREEN_VALUE
      case 0x3037: // EGL_TRANSPARENT_RED_VALUE
        // "If EGL_TRANSPARENT_TYPE is EGL_NONE, then the values for EGL_TRANSPARENT_RED_VALUE, EGL_TRANSPARENT_GREEN_VALUE, and EGL_TRANSPARENT_BLUE_VALUE are undefined."
        HEAP32[((value)>>2)]=-1;
        return 1;
      case 0x3039: // EGL_BIND_TO_TEXTURE_RGB
      case 0x303A: // EGL_BIND_TO_TEXTURE_RGBA
        HEAP32[((value)>>2)]=0;
        return 1;
      case 0x303B: // EGL_MIN_SWAP_INTERVAL
      case 0x303C: // EGL_MAX_SWAP_INTERVAL
        HEAP32[((value)>>2)]=1;
        return 1;
      case 0x303D: // EGL_LUMINANCE_SIZE
      case 0x303E: // EGL_ALPHA_MASK_SIZE
        HEAP32[((value)>>2)]=0;
        return 1;
      case 0x303F: // EGL_COLOR_BUFFER_TYPE
        // EGL has two types of buffers: EGL_RGB_BUFFER and EGL_LUMINANCE_BUFFER.
        HEAP32[((value)>>2)]=0x308E;
        return 1;
      case 0x3040: // EGL_RENDERABLE_TYPE
        // A bit combination of EGL_OPENGL_ES_BIT,EGL_OPENVG_BIT,EGL_OPENGL_ES2_BIT and EGL_OPENGL_BIT.
        HEAP32[((value)>>2)]=0x0004;
        return 1;
      case 0x3042: // EGL_CONFORMANT
        // "EGL_CONFORMANT is a mask indicating if a client API context created with respect to the corresponding EGLConfig will pass the required conformance tests for that API."
        HEAP32[((value)>>2)]=0;
        return 1;
      default:
        EGL.setErrorCode(0x3004 /* EGL_BAD_ATTRIBUTE */);
        return 0;
      }
    }

   
  Module["_i64Add"] = _i64Add;

  
  function __ZSt18uncaught_exceptionv() { // std::uncaught_exception()
      return !!__ZSt18uncaught_exceptionv.uncaught_exception;
    }
  
  function ___cxa_find_matching_catch() {
      var thrown = EXCEPTIONS.last;
      if (!thrown) {
        // just pass through the null ptr
        return ((asm["setTempRet0"](0),0)|0);
      }
      var info = EXCEPTIONS.infos[thrown];
      var throwntype = info.type;
      if (!throwntype) {
        // just pass through the thrown ptr
        return ((asm["setTempRet0"](0),thrown)|0);
      }
      var typeArray = Array.prototype.slice.call(arguments);
  
      var pointer = Module['___cxa_is_pointer_type'](throwntype);
      // can_catch receives a **, add indirection
      if (!___cxa_find_matching_catch.buffer) ___cxa_find_matching_catch.buffer = _malloc(4);
      HEAP32[((___cxa_find_matching_catch.buffer)>>2)]=thrown;
      thrown = ___cxa_find_matching_catch.buffer;
      // The different catch blocks are denoted by different types.
      // Due to inheritance, those types may not precisely match the
      // type of the thrown object. Find one which matches, and
      // return the type of the catch block which should be called.
      for (var i = 0; i < typeArray.length; i++) {
        if (typeArray[i] && Module['___cxa_can_catch'](typeArray[i], throwntype, thrown)) {
          thrown = HEAP32[((thrown)>>2)]; // undo indirection
          info.adjusted = thrown;
          return ((asm["setTempRet0"](typeArray[i]),thrown)|0);
        }
      }
      // Shouldn't happen unless we have bogus data in typeArray
      // or encounter a type for which emscripten doesn't have suitable
      // typeinfo defined. Best-efforts match just in case.
      thrown = HEAP32[((thrown)>>2)]; // undo indirection
      return ((asm["setTempRet0"](throwntype),thrown)|0);
    }function ___cxa_throw(ptr, type, destructor) {
      EXCEPTIONS.infos[ptr] = {
        ptr: ptr,
        adjusted: ptr,
        type: type,
        destructor: destructor,
        refcount: 0
      };
      EXCEPTIONS.last = ptr;
      if (!("uncaught_exception" in __ZSt18uncaught_exceptionv)) {
        __ZSt18uncaught_exceptionv.uncaught_exception = 1;
      } else {
        __ZSt18uncaught_exceptionv.uncaught_exception++;
      }
      throw ptr;
    }

  function _emscripten_set_touchend_callback(target, userData, useCapture, callbackfunc) {
      JSEvents.registerTouchEventCallback(target, userData, useCapture, callbackfunc, 23, "touchend");
      return 0;
    }

  
  
  function __setLetterbox(element, topBottom, leftRight) {
      if (JSEvents.isInternetExplorer()) {
        // Cannot use padding on IE11, because IE11 computes padding in addition to the size, unlike
        // other browsers, which treat padding to be part of the size.
        // e.g.
        // FF, Chrome: If CSS size = 1920x1080, padding-leftright = 460, padding-topbottomx40, then content size = (1920 - 2*460) x (1080-2*40) = 1000x1000px, and total element size = 1920x1080px.
        //       IE11: If CSS size = 1920x1080, padding-leftright = 460, padding-topbottomx40, then content size = 1920x1080px and total element size = (1920+2*460) x (1080+2*40)px.
        // IE11  treats margin like Chrome and FF treat padding.
        element.style.marginLeft = element.style.marginRight = leftRight + 'px';
        element.style.marginTop = element.style.marginBottom = topBottom + 'px';
      } else {
        // Cannot use margin to specify letterboxes in FF or Chrome, since those ignore margins in fullscreen mode.
        element.style.paddingLeft = element.style.paddingRight = leftRight + 'px';
        element.style.paddingTop = element.style.paddingBottom = topBottom + 'px';
      }
    }function _emscripten_do_request_fullscreen(target, strategy) {
      if (typeof JSEvents.fullscreenEnabled() === 'undefined') return -1;
      if (!JSEvents.fullscreenEnabled()) return -3;
      if (!target) target = '#canvas';
      target = JSEvents.findEventTarget(target);
      if (!target) return -4;
  
      if (!target.requestFullscreen && !target.msRequestFullscreen && !target.mozRequestFullScreen && !target.mozRequestFullscreen && !target.webkitRequestFullscreen) {
        return -3;
      }
  
      var canPerformRequests = JSEvents.canPerformEventHandlerRequests();
  
      // Queue this function call if we're not currently in an event handler and the user saw it appropriate to do so.
      if (!canPerformRequests) {
        if (strategy.deferUntilInEventHandler) {
          JSEvents.deferCall(JSEvents.requestFullscreen, 1 /* priority over pointer lock */, [target, strategy]);
          return 1;
        } else {
          return -2;
        }
      }
  
      return JSEvents.requestFullscreen(target, strategy);
    }
  
  function __registerRestoreOldStyle(canvas) {
      var oldWidth = canvas.width;
      var oldHeight = canvas.height;
      var oldCssWidth = canvas.style.width;
      var oldCssHeight = canvas.style.height;
      var oldBackgroundColor = canvas.style.backgroundColor; // Chrome reads color from here.
      var oldDocumentBackgroundColor = document.body.style.backgroundColor; // IE11 reads color from here.
      // Firefox always has black background color.
      var oldPaddingLeft = canvas.style.paddingLeft; // Chrome, FF, Safari
      var oldPaddingRight = canvas.style.paddingRight;
      var oldPaddingTop = canvas.style.paddingTop;
      var oldPaddingBottom = canvas.style.paddingBottom;
      var oldMarginLeft = canvas.style.marginLeft; // IE11
      var oldMarginRight = canvas.style.marginRight;
      var oldMarginTop = canvas.style.marginTop;
      var oldMarginBottom = canvas.style.marginBottom;
      var oldDocumentBodyMargin = document.body.style.margin;
      var oldDocumentOverflow = document.documentElement.style.overflow; // Chrome, Firefox
      var oldDocumentScroll = document.body.scroll; // IE
      var oldImageRendering = canvas.style.imageRendering;
  
      function restoreOldStyle() {
        var fullscreenElement = document.fullscreenElement || document.mozFullScreenElement || document.webkitFullscreenElement || document.msFullscreenElement;
        if (!fullscreenElement) {
          document.removeEventListener('fullscreenchange', restoreOldStyle);
          document.removeEventListener('mozfullscreenchange', restoreOldStyle);
          document.removeEventListener('webkitfullscreenchange', restoreOldStyle);
          document.removeEventListener('MSFullscreenChange', restoreOldStyle);
  
          canvas.width = oldWidth;
          canvas.height = oldHeight;
          canvas.style.width = oldCssWidth;
          canvas.style.height = oldCssHeight;
          canvas.style.backgroundColor = oldBackgroundColor; // Chrome
          // IE11 hack: assigning 'undefined' or an empty string to document.body.style.backgroundColor has no effect, so first assign back the default color
          // before setting the undefined value. Setting undefined value is also important, or otherwise we would later treat that as something that the user
          // had explicitly set so subsequent fullscreen transitions would not set background color properly.
          if (!oldDocumentBackgroundColor) document.body.style.backgroundColor = 'white';
          document.body.style.backgroundColor = oldDocumentBackgroundColor; // IE11
          canvas.style.paddingLeft = oldPaddingLeft; // Chrome, FF, Safari
          canvas.style.paddingRight = oldPaddingRight;
          canvas.style.paddingTop = oldPaddingTop;
          canvas.style.paddingBottom = oldPaddingBottom;
          canvas.style.marginLeft = oldMarginLeft; // IE11
          canvas.style.marginRight = oldMarginRight;
          canvas.style.marginTop = oldMarginTop;
          canvas.style.marginBottom = oldMarginBottom;
          document.body.style.margin = oldDocumentBodyMargin;
          document.documentElement.style.overflow = oldDocumentOverflow; // Chrome, Firefox
          document.body.scroll = oldDocumentScroll; // IE
          canvas.style.imageRendering = oldImageRendering;
          if (canvas.GLctxObject) canvas.GLctxObject.GLctx.viewport(0, 0, oldWidth, oldHeight);
  
          if (__currentFullscreenStrategy.canvasResizedCallback) {
            Runtime.dynCall('iiii', __currentFullscreenStrategy.canvasResizedCallback, [37, 0, __currentFullscreenStrategy.canvasResizedCallbackUserData]);
          }
        }
      }
      document.addEventListener('fullscreenchange', restoreOldStyle);
      document.addEventListener('mozfullscreenchange', restoreOldStyle);
      document.addEventListener('webkitfullscreenchange', restoreOldStyle);
      document.addEventListener('MSFullscreenChange', restoreOldStyle);
      return restoreOldStyle;
    }function _emscripten_request_fullscreen_strategy(target, deferUntilInEventHandler, fullscreenStrategy) {
      var strategy = {};
      strategy.scaleMode = HEAP32[((fullscreenStrategy)>>2)];
      strategy.canvasResolutionScaleMode = HEAP32[(((fullscreenStrategy)+(4))>>2)];
      strategy.filteringMode = HEAP32[(((fullscreenStrategy)+(8))>>2)];
      strategy.deferUntilInEventHandler = deferUntilInEventHandler;
      strategy.canvasResizedCallback = HEAP32[(((fullscreenStrategy)+(12))>>2)];
      strategy.canvasResizedCallbackUserData = HEAP32[(((fullscreenStrategy)+(16))>>2)];
      __currentFullscreenStrategy = strategy;
  
      return _emscripten_do_request_fullscreen(target, strategy);
    }

  function _alSourceStop(source) {
      if (!AL.currentContext) {
        return;
      }
      var src = AL.currentContext.src[source];
      if (!src) {
        AL.currentContext.err = 0xA001 /* AL_INVALID_NAME */;
        return;
      }
      AL.setSourceState(src, 0x1014 /* AL_STOPPED */);
    }

  function _emscripten_glGetProgramiv(program, pname, p) {
      if (!p) {
        // GLES2 specification does not specify how to behave if p is a null pointer. Since calling this function does not make sense
        // if p == null, issue a GL error to notify user about it. 
        GL.recordError(0x0501 /* GL_INVALID_VALUE */);
        return;
      }
      if (pname == 0x8B84) { // GL_INFO_LOG_LENGTH
        var log = GLctx.getProgramInfoLog(GL.programs[program]);
        if (log === null) log = '(unknown error)';
        HEAP32[((p)>>2)]=log.length + 1;
      } else if (pname == 0x8B87 /* GL_ACTIVE_UNIFORM_MAX_LENGTH */) {
        var ptable = GL.programInfos[program];
        if (ptable) {
          HEAP32[((p)>>2)]=ptable.maxUniformLength;
          return;
        } else if (program < GL.counter) {
          GL.recordError(0x0502 /* GL_INVALID_OPERATION */);
        } else {
          GL.recordError(0x0501 /* GL_INVALID_VALUE */);
        }
      } else if (pname == 0x8B8A /* GL_ACTIVE_ATTRIBUTE_MAX_LENGTH */) {
        var ptable = GL.programInfos[program];
        if (ptable) {
          if (ptable.maxAttributeLength == -1) {
            var program = GL.programs[program];
            var numAttribs = GLctx.getProgramParameter(program, GLctx.ACTIVE_ATTRIBUTES);
            ptable.maxAttributeLength = 0; // Spec says if there are no active attribs, 0 must be returned.
            for(var i = 0; i < numAttribs; ++i) {
              var activeAttrib = GLctx.getActiveAttrib(program, i);
              ptable.maxAttributeLength = Math.max(ptable.maxAttributeLength, activeAttrib.name.length+1);
            }
          }
          HEAP32[((p)>>2)]=ptable.maxAttributeLength;
          return;
        } else if (program < GL.counter) {
          GL.recordError(0x0502 /* GL_INVALID_OPERATION */);
        } else {
          GL.recordError(0x0501 /* GL_INVALID_VALUE */);
        }
      } else {
        HEAP32[((p)>>2)]=GLctx.getProgramParameter(GL.programs[program], pname);
      }
    }

  function _emscripten_glFinish() { GLctx.finish() }

  function ___syscall3(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // read
      var stream = SYSCALLS.getStreamFromFD(), buf = SYSCALLS.get(), count = SYSCALLS.get();
      return FS.read(stream, HEAP8,buf, count);
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function _emscripten_glDepthFunc(x0) { GLctx.depthFunc(x0) }

  
  function _realloc() { throw 'bad' }
  Module["_realloc"] = _realloc; 
  Module["_saveSetjmp"] = _saveSetjmp;

  function _emscripten_get_num_gamepads() {
      if (!navigator.getGamepads && !navigator.webkitGetGamepads) return -1;
      if (navigator.getGamepads) {
        return navigator.getGamepads().length;
      } else if (navigator.webkitGetGamepads) {
        return navigator.webkitGetGamepads().length;
      }
    }

  function _emscripten_set_blur_callback(target, userData, useCapture, callbackfunc) {
      JSEvents.registerFocusEventCallback(target, userData, useCapture, callbackfunc, 12, "blur");
      return 0;
    }

  function _emscripten_glReleaseShaderCompiler() {
      // NOP (as allowed by GLES 2.0 spec)
    }

  function _sigaction(signum, act, oldact) {
      //int sigaction(int signum, const struct sigaction *act, struct sigaction *oldact);
      Module.printErr('Calling stub instead of sigaction()');
      return 0;
    }

  function _emscripten_glUniform4iv(location, count, value) {
      location = GL.uniforms[location];
      count *= 4;
      value = HEAP32.subarray((value)>>2,(value+count*4)>>2);
      GLctx.uniform4iv(location, value);
    }

  function _glClear(x0) { GLctx.clear(x0) }

  
  function emscriptenWebGLGetUniform(program, location, params, type) {
      if (!params) {
        // GLES2 specification does not specify how to behave if params is a null pointer. Since calling this function does not make sense
        // if params == null, issue a GL error to notify user about it. 
        GL.recordError(0x0501 /* GL_INVALID_VALUE */);
        return;
      }
      var data = GLctx.getUniform(GL.programs[program], GL.uniforms[location]);
      if (typeof data == 'number' || typeof data == 'boolean') {
        switch (type) {
          case 'Integer': HEAP32[((params)>>2)]=data; break;
          case 'Float': HEAPF32[((params)>>2)]=data; break;
          default: throw 'internal emscriptenWebGLGetUniform() error, bad type: ' + type;
        }
      } else {
        for (var i = 0; i < data.length; i++) {
          switch (type) {
            case 'Integer': HEAP32[(((params)+(i))>>2)]=data[i]; break;
            case 'Float': HEAPF32[(((params)+(i))>>2)]=data[i]; break;
            default: throw 'internal emscriptenWebGLGetUniform() error, bad type: ' + type;
          }
        }
      }
    }function _emscripten_glGetUniformiv(program, location, params) {
      emscriptenWebGLGetUniform(program, location, params, 'Integer');
    }

  function _emscripten_set_resize_callback(target, userData, useCapture, callbackfunc) {
      JSEvents.registerUiEventCallback(target, userData, useCapture, callbackfunc, 10, "resize");
      return 0;
    }

  function _emscripten_glLoadIdentity(){ throw 'Legacy GL function (glLoadIdentity) called. If you want legacy GL emulation, you need to compile with -s LEGACY_GL_EMULATION=1 to enable legacy GL emulation.'; }

  
  
  
  
  var _environ=STATICTOP; STATICTOP += 16;;var ___environ=_environ;function ___buildEnvironment(env) {
      // WARNING: Arbitrary limit!
      var MAX_ENV_VALUES = 64;
      var TOTAL_ENV_SIZE = 1024;
  
      // Statically allocate memory for the environment.
      var poolPtr;
      var envPtr;
      if (!___buildEnvironment.called) {
        ___buildEnvironment.called = true;
        // Set default values. Use string keys for Closure Compiler compatibility.
        ENV['USER'] = ENV['LOGNAME'] = 'web_user';
        ENV['PATH'] = '/';
        ENV['PWD'] = '/';
        ENV['HOME'] = '/home/web_user';
        ENV['LANG'] = 'C';
        ENV['_'] = Module['thisProgram'];
        // Allocate memory.
        poolPtr = allocate(TOTAL_ENV_SIZE, 'i8', ALLOC_STATIC);
        envPtr = allocate(MAX_ENV_VALUES * 4,
                          'i8*', ALLOC_STATIC);
        HEAP32[((envPtr)>>2)]=poolPtr;
        HEAP32[((_environ)>>2)]=envPtr;
      } else {
        envPtr = HEAP32[((_environ)>>2)];
        poolPtr = HEAP32[((envPtr)>>2)];
      }
  
      // Collect key=value lines.
      var strings = [];
      var totalSize = 0;
      for (var key in env) {
        if (typeof env[key] === 'string') {
          var line = key + '=' + env[key];
          strings.push(line);
          totalSize += line.length;
        }
      }
      if (totalSize > TOTAL_ENV_SIZE) {
        throw new Error('Environment size exceeded TOTAL_ENV_SIZE!');
      }
  
      // Make new.
      var ptrSize = 4;
      for (var i = 0; i < strings.length; i++) {
        var line = strings[i];
        writeAsciiToMemory(line, poolPtr);
        HEAP32[(((envPtr)+(i * ptrSize))>>2)]=poolPtr;
        poolPtr += line.length + 1;
      }
      HEAP32[(((envPtr)+(strings.length * ptrSize))>>2)]=0;
    }var ENV={};function _getenv(name) {
      // char *getenv(const char *name);
      // http://pubs.opengroup.org/onlinepubs/009695399/functions/getenv.html
      if (name === 0) return 0;
      name = Pointer_stringify(name);
      if (!ENV.hasOwnProperty(name)) return 0;
  
      if (_getenv.ret) _free(_getenv.ret);
      _getenv.ret = allocate(intArrayFromString(ENV[name]), 'i8', ALLOC_NORMAL);
      return _getenv.ret;
    }

  function _emscripten_set_element_css_size(target, width, height) {
      if (!target) {
        target = Module['canvas'];
      } else {
        target = JSEvents.findEventTarget(target);
      }
  
      if (!target) return -4;
  
      target.style.setProperty("width", width + "px");
      target.style.setProperty("height", height + "px");
  
      return 0;
    }

  function _alDistanceModel(model) {
      if (model !== 0 /* AL_NONE */) {
      }
    }

  function _emscripten_glAttachShader(program, shader) {
      GLctx.attachShader(GL.programs[program],
                              GL.shaders[shader]);
    }

  function _alSourceUnqueueBuffers(source, count, buffers) {
      if (!AL.currentContext) {
        return;
      }
      var src = AL.currentContext.src[source];
      if (!src) {
        AL.currentContext.err = 0xA001 /* AL_INVALID_NAME */;
        return;
      }
  
      if (count > src.buffersPlayed) {
        AL.currentContext.err = 0xA003 /* AL_INVALID_VALUE */;
        return;
      }
  
      for (var i = 0; i < count; i++) {
        var entry = src.queue.shift();
        // Write the buffers index out to the return list.
        for (var j = 0; j < AL.currentContext.buf.length; j++) {
          var b = AL.currentContext.buf[j];
          if (b && b == entry.buffer) {
            HEAP32[(((buffers)+(i*4))>>2)]=j+1;
            break;
          }
        }
        src.buffersPlayed--;
      }
  
      AL.updateSource(src);
    }

  function _emscripten_glGetRenderbufferParameteriv(target, pname, params) {
      if (!params) {
        // GLES2 specification does not specify how to behave if params is a null pointer. Since calling this function does not make sense
        // if params == null, issue a GL error to notify user about it. 
        GL.recordError(0x0501 /* GL_INVALID_VALUE */);
        return;
      }
      HEAP32[((params)>>2)]=GLctx.getRenderbufferParameter(target, pname);
    }

  function _alGetSourcef(source, param, value) {
      if (!AL.currentContext) {
        return;
      }
      var src = AL.currentContext.src[source];
      if (!src) {
        AL.currentContext.err = 0xA001 /* AL_INVALID_NAME */;
        return;
      }
      switch (param) {
      // case 0x1003 /* AL_PITCH */:
      //   break;
      case 0x100A /* AL_GAIN */:
        HEAPF32[((value)>>2)]=src.gain.gain.value
        break;
      // case 0x100D /* AL_MIN_GAIN */:
      //   break;
      // case 0x100E /* AL_MAX_GAIN */:
      //   break;
      case 0x1023 /* AL_MAX_DISTANCE */:
        HEAPF32[((value)>>2)]=src.maxDistance
        break;
      case 0x1021 /* AL_ROLLOFF_FACTOR */:
        HEAPF32[((value)>>2)]=src.rolloffFactor
        break;
      case 0x1022 /* AL_CONE_OUTER_GAIN */:
        HEAPF32[((value)>>2)]=src.coneOuterGain
        break;
      case 0x1001 /* AL_CONE_INNER_ANGLE */:
        HEAPF32[((value)>>2)]=src.coneInnerAngle
        break;
      case 0x1002 /* AL_CONE_OUTER_ANGLE */:
        HEAPF32[((value)>>2)]=src.coneOuterAngle
        break;
      case 0x1020 /* AL_REFERENCE_DISTANCE */:
        HEAPF32[((value)>>2)]=src.refDistance
        break;
      // case 0x1024 /* AL_SEC_OFFSET */:
      //   break;
      // case 0x1025 /* AL_SAMPLE_OFFSET */:
      //   break;
      // case 0x1026 /* AL_BYTE_OFFSET */:
      //   break;
      default:
        AL.currentContext.err = 0xA002 /* AL_INVALID_ENUM */;
        break;
      }
    }

  function _emscripten_request_pointerlock(target, deferUntilInEventHandler) {
      if (!target) target = '#canvas';
      target = JSEvents.findEventTarget(target);
      if (!target) return -4;
      if (!target.requestPointerLock && !target.mozRequestPointerLock && !target.webkitRequestPointerLock && !target.msRequestPointerLock) {
        return -1;
      }
  
      var canPerformRequests = JSEvents.canPerformEventHandlerRequests();
  
      // Queue this function call if we're not currently in an event handler and the user saw it appropriate to do so.
      if (!canPerformRequests) {
        if (deferUntilInEventHandler) {
          JSEvents.deferCall(JSEvents.requestPointerLock, 2 /* priority below fullscreen */, [target]);
          return 1;
        } else {
          return -2;
        }
      }
  
      return JSEvents.requestPointerLock(target);
    }

  
  function _gmtime_r(time, tmPtr) {
      var date = new Date(HEAP32[((time)>>2)]*1000);
      HEAP32[((tmPtr)>>2)]=date.getUTCSeconds();
      HEAP32[(((tmPtr)+(4))>>2)]=date.getUTCMinutes();
      HEAP32[(((tmPtr)+(8))>>2)]=date.getUTCHours();
      HEAP32[(((tmPtr)+(12))>>2)]=date.getUTCDate();
      HEAP32[(((tmPtr)+(16))>>2)]=date.getUTCMonth();
      HEAP32[(((tmPtr)+(20))>>2)]=date.getUTCFullYear()-1900;
      HEAP32[(((tmPtr)+(24))>>2)]=date.getUTCDay();
      HEAP32[(((tmPtr)+(36))>>2)]=0;
      HEAP32[(((tmPtr)+(32))>>2)]=0;
      var start = Date.UTC(date.getUTCFullYear(), 0, 1, 0, 0, 0, 0);
      var yday = ((date.getTime() - start) / (1000 * 60 * 60 * 24))|0;
      HEAP32[(((tmPtr)+(28))>>2)]=yday;
      HEAP32[(((tmPtr)+(40))>>2)]=___tm_timezone;
  
      return tmPtr;
    }function _gmtime(time) {
      return _gmtime_r(time, ___tm_current);
    }

  function _eglCreateWindowSurface(display, config, win, attrib_list) { 
      if (display != 62000 /* Magic ID for Emscripten 'default display' */) {
        EGL.setErrorCode(0x3008 /* EGL_BAD_DISPLAY */);
        return 0;
      }
      if (config != 62002 /* Magic ID for the only EGLConfig supported by Emscripten */) {
        EGL.setErrorCode(0x3005 /* EGL_BAD_CONFIG */);
        return 0;
      }
      // TODO: Examine attrib_list! Parameters that can be present there are:
      // - EGL_RENDER_BUFFER (must be EGL_BACK_BUFFER)
      // - EGL_VG_COLORSPACE (can't be set)
      // - EGL_VG_ALPHA_FORMAT (can't be set)
      EGL.setErrorCode(0x3000 /* EGL_SUCCESS */);
      return 62006; /* Magic ID for Emscripten 'default surface' */
    }

  function _alListenerfv(param, values) {
      if (!AL.currentContext) {
        return;
      }
      switch (param) {
      case 0x1004 /* AL_POSITION */:
        var x = HEAPF32[((values)>>2)];
        var y = HEAPF32[(((values)+(4))>>2)];
        var z = HEAPF32[(((values)+(8))>>2)];
        AL.currentContext.ctx.listener._position = [x, y, z];
        AL.currentContext.ctx.listener.setPosition(x, y, z);
        break;
      case 0x1006 /* AL_VELOCITY */:
        var x = HEAPF32[((values)>>2)];
        var y = HEAPF32[(((values)+(4))>>2)];
        var z = HEAPF32[(((values)+(8))>>2)];
        AL.currentContext.ctx.listener._velocity = [x, y, z];
        AL.currentContext.ctx.listener.setVelocity(x, y, z);
        break;
      case 0x100F /* AL_ORIENTATION */:
        var x = HEAPF32[((values)>>2)];
        var y = HEAPF32[(((values)+(4))>>2)];
        var z = HEAPF32[(((values)+(8))>>2)];
        var x2 = HEAPF32[(((values)+(12))>>2)];
        var y2 = HEAPF32[(((values)+(16))>>2)];
        var z2 = HEAPF32[(((values)+(20))>>2)];
        AL.currentContext.ctx.listener._orientation = [x, y, z, x2, y2, z2];
        AL.currentContext.ctx.listener.setOrientation(x, y, z, x2, y2, z2);
        break;
      default:
        AL.currentContext.err = 0xA002 /* AL_INVALID_ENUM */;
        break;
      }
    }

  function _pthread_cond_broadcast() {
      return 0;
    }

  function _alGetSourcei(source, param, value) {
      if (!AL.currentContext) {
        return;
      }
      var src = AL.currentContext.src[source];
      if (!src) {
        AL.currentContext.err = 0xA001 /* AL_INVALID_NAME */;
        return;
      }
  
      // Being that we have no way to receive end events from buffer nodes,
      // we currently proccess and update a source's buffer queue every
      // ~QUEUE_INTERVAL milliseconds. However, this interval is not precise,
      // so we also forcefully update the source when alGetSourcei is queried
      // to aid in the common scenario of application calling alGetSourcei(AL_BUFFERS_PROCESSED)
      // to recycle buffers.
      AL.updateSource(src);
  
      switch (param) {
      case 0x202 /* AL_SOURCE_RELATIVE */:
        HEAP32[((value)>>2)]=src.panner ? 1 : 0;
        break;
      case 0x1001 /* AL_CONE_INNER_ANGLE */:
        HEAP32[((value)>>2)]=src.coneInnerAngle;
        break;
      case 0x1002 /* AL_CONE_OUTER_ANGLE */:
        HEAP32[((value)>>2)]=src.coneOuterAngle;
        break;
      case 0x1007 /* AL_LOOPING */:
        HEAP32[((value)>>2)]=src.loop;
        break;
      case 0x1009 /* AL_BUFFER */:
        if (!src.queue.length) {
          HEAP32[((value)>>2)]=0;
        } else {
          // Find the first unprocessed buffer.
          var buffer = src.queue[src.buffersPlayed].buffer;
          // Return its index.
          for (var i = 0; i < AL.currentContext.buf.length; ++i) {
            if (buffer == AL.currentContext.buf[i]) {
              HEAP32[((value)>>2)]=i+1;
              return;
            }
          }
          HEAP32[((value)>>2)]=0;
        }
        break;
      case 0x1010 /* AL_SOURCE_STATE */:
        HEAP32[((value)>>2)]=src.state;
        break;
      case 0x1015 /* AL_BUFFERS_QUEUED */:
        HEAP32[((value)>>2)]=src.queue.length
        break;
      case 0x1016 /* AL_BUFFERS_PROCESSED */:
        if (src.loop) {
          HEAP32[((value)>>2)]=0
        } else {
          HEAP32[((value)>>2)]=src.buffersPlayed
        }
        break;
      default:
        AL.currentContext.err = 0xA002 /* AL_INVALID_ENUM */;
        break;
      }
    }

  function _gettimeofday(ptr) {
      var now = Date.now();
      HEAP32[((ptr)>>2)]=(now/1000)|0; // seconds
      HEAP32[(((ptr)+(4))>>2)]=((now % 1000)*1000)|0; // microseconds
      return 0;
    }

  function _emscripten_glClearStencil(x0) { GLctx.clearStencil(x0) }

  function _emscripten_glDetachShader(program, shader) {
      GLctx.detachShader(GL.programs[program],
                              GL.shaders[shader]);
    }

  function _emscripten_get_device_pixel_ratio() {
      return window.devicePixelRatio || 1.0;
    }

  function _emscripten_glDeleteVertexArrays(n, vaos) {
      for(var i = 0; i < n; i++) {
        var id = HEAP32[(((vaos)+(i*4))>>2)];
        GLctx['deleteVertexArray'](GL.vaos[id]);
        GL.vaos[id] = null;
      }
    }

  function _alGenSources(count, sources) {
      if (!AL.currentContext) {
        return;
      }
      for (var i = 0; i < count; ++i) {
        var gain = AL.currentContext.ctx.createGain();
        gain.connect(AL.currentContext.gain);
        AL.currentContext.src[AL.newSrcId] = {
          state: 0x1011 /* AL_INITIAL */,
          queue: [],
          loop: false,
          get refDistance() {
            return this._refDistance || 1;
          },
          set refDistance(val) {
            this._refDistance = val;
            if (this.panner) this.panner.refDistance = val;
          },
          get maxDistance() {
            return this._maxDistance || 10000;
          },
          set maxDistance(val) {
            this._maxDistance = val;
            if (this.panner) this.panner.maxDistance = val;
          },
          get rolloffFactor() {
            return this._rolloffFactor || 1;
          },
          set rolloffFactor(val) {
            this._rolloffFactor = val;
            if (this.panner) this.panner.rolloffFactor = val;
          },
          get position() {
            return this._position || [0, 0, 0];
          },
          set position(val) {
            this._position = val;
            if (this.panner) this.panner.setPosition(val[0], val[1], val[2]);
          },
          get velocity() {
            return this._velocity || [0, 0, 0];
          },
          set velocity(val) {
            this._velocity = val;
            if (this.panner) this.panner.setVelocity(val[0], val[1], val[2]);
          },
          get direction() {
            return this._direction || [0, 0, 0];
          },
          set direction(val) {
            this._direction = val;
            if (this.panner) this.panner.setOrientation(val[0], val[1], val[2]);
          },
          get coneOuterGain() {
            return this._coneOuterGain || 0.0;
          },
          set coneOuterGain(val) {
            this._coneOuterGain = val;
            if (this.panner) this.panner.coneOuterGain = val;
          },
          get coneInnerAngle() {
            return this._coneInnerAngle || 360.0;
          },
          set coneInnerAngle(val) {
            this._coneInnerAngle = val;
            if (this.panner) this.panner.coneInnerAngle = val;
          },
          get coneOuterAngle() {
            return this._coneOuterAngle || 360.0;
          },
          set coneOuterAngle(val) {
            this._coneOuterAngle = val;
            if (this.panner) this.panner.coneOuterAngle = val;
          },
          gain: gain,
          panner: null,
          buffersPlayed: 0,
          bufferPosition: 0
        };
        HEAP32[(((sources)+(i*4))>>2)]=AL.newSrcId;
        AL.newSrcId++;
      }
    }

  function _alGetFloat() {
  Module['printErr']('missing function: alGetFloat'); abort(-1);
  }

  function _emscripten_get_element_css_size(target, width, height) {
      if (!target) {
        target = Module['canvas'];
      } else {
        target = JSEvents.findEventTarget(target);
      }
  
      if (!target) return -4;
  
      if (target.getBoundingClientRect) {
        var rect = target.getBoundingClientRect();
        HEAPF64[((width)>>3)]=rect.right - rect.left;
        HEAPF64[((height)>>3)]=rect.bottom - rect.top;
      } else {
        HEAPF64[((width)>>3)]=target.clientWidth;
        HEAPF64[((height)>>3)]=target.clientHeight;
      }
  
      return 0;
    }

  function _emscripten_glMatrixMode(){ throw 'Legacy GL function (glMatrixMode) called. If you want legacy GL emulation, you need to compile with -s LEGACY_GL_EMULATION=1 to enable legacy GL emulation.'; }

  function _pthread_mutex_destroy() {}

  function _emscripten_glGetTexParameteriv(target, pname, params) {
      if (!params) {
        // GLES2 specification does not specify how to behave if params is a null pointer. Since calling this function does not make sense
        // if p == null, issue a GL error to notify user about it. 
        GL.recordError(0x0501 /* GL_INVALID_VALUE */);
        return;
      }
      HEAP32[((params)>>2)]=GLctx.getTexParameter(target, pname);
    }

  function ___syscall122(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // uname
      var buf = SYSCALLS.get();
      if (!buf) return -ERRNO_CODES.EFAULT
      var layout = {"sysname":0,"nodename":65,"domainname":325,"machine":260,"version":195,"release":130,"__size__":390};
      function copyString(element, value) {
        var offset = layout[element];
        writeAsciiToMemory(value, buf + offset);
      }
      copyString('sysname', 'Emscripten');
      copyString('nodename', 'emscripten');
      copyString('release', '1.0');
      copyString('version', '#1');
      copyString('machine', 'x86-JS');
      return 0;
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function _alDeleteSources(count, sources) {
      if (!AL.currentContext) {
        return;
      }
      for (var i = 0; i < count; ++i) {
        var sourceIdx = HEAP32[(((sources)+(i*4))>>2)];
        delete AL.currentContext.src[sourceIdx];
      }
    }

  var _llvm_ctlz_i32=true;

  function _emscripten_glGenerateMipmap(x0) { GLctx.generateMipmap(x0) }

  function _emscripten_glSampleCoverage(x0, x1) { GLctx.sampleCoverage(x0, x1) }

  function _emscripten_glCullFace(x0) { GLctx.cullFace(x0) }

  function _emscripten_glUniform4f(location, v0, v1, v2, v3) {
      location = GL.uniforms[location];
      GLctx.uniform4f(location, v0, v1, v2, v3);
    }

  function _emscripten_glUseProgram(program) {
      GLctx.useProgram(program ? GL.programs[program] : null);
    }

  function _emscripten_glHint(x0, x1) { GLctx.hint(x0, x1) }

  function ___syscall63(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // dup2
      var old = SYSCALLS.getStreamFromFD(), suggestFD = SYSCALLS.get();
      if (old.fd === suggestFD) return suggestFD;
      return SYSCALLS.doDup(old.path, old.flags, suggestFD);
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function _dlsym(handle, symbol) {
      // void *dlsym(void *restrict handle, const char *restrict name);
      // http://pubs.opengroup.org/onlinepubs/009695399/functions/dlsym.html
      symbol = Pointer_stringify(symbol);
  
      if (!DLFCN.loadedLibs[handle]) {
        DLFCN.errorMsg = 'Tried to dlsym() from an unopened handle: ' + handle;
        return 0;
      } else {
        var lib = DLFCN.loadedLibs[handle];
        symbol = '_' + symbol;
        if (lib.cached_functions.hasOwnProperty(symbol)) {
          return lib.cached_functions[symbol];
        }
        if (!lib.module.hasOwnProperty(symbol)) {
          DLFCN.errorMsg = ('Tried to lookup unknown symbol "' + symbol +
                                 '" in dynamic lib: ' + lib.name);
          return 0;
        } else {
          var result = lib.module[symbol];
          if (typeof result == 'function') {
            result = Runtime.addFunction(result);
            lib.cached_functions = result;
          }
          return result;
        }
      }
    }

  function _emscripten_glUniform2fv(location, count, value) {
      location = GL.uniforms[location];
      var view;
      if (count === 1) {
        // avoid allocation for the common case of uploading one uniform
        view = GL.miniTempBufferViews[1];
        view[0] = HEAPF32[((value)>>2)];
        view[1] = HEAPF32[(((value)+(4))>>2)];
      } else {
        view = HEAPF32.subarray((value)>>2,(value+count*8)>>2);
      }
      GLctx.uniform2fv(location, view);
    }

  
  
  function __isLeapYear(year) {
        return year%4 === 0 && (year%100 !== 0 || year%400 === 0);
    }
  
  function __arraySum(array, index) {
      var sum = 0;
      for (var i = 0; i <= index; sum += array[i++]);
      return sum;
    }
  
  
  var __MONTH_DAYS_LEAP=[31,29,31,30,31,30,31,31,30,31,30,31];
  
  var __MONTH_DAYS_REGULAR=[31,28,31,30,31,30,31,31,30,31,30,31];function __addDays(date, days) {
      var newDate = new Date(date.getTime());
      while(days > 0) {
        var leap = __isLeapYear(newDate.getFullYear());
        var currentMonth = newDate.getMonth();
        var daysInCurrentMonth = (leap ? __MONTH_DAYS_LEAP : __MONTH_DAYS_REGULAR)[currentMonth];
  
        if (days > daysInCurrentMonth-newDate.getDate()) {
          // we spill over to next month
          days -= (daysInCurrentMonth-newDate.getDate()+1);
          newDate.setDate(1);
          if (currentMonth < 11) {
            newDate.setMonth(currentMonth+1)
          } else {
            newDate.setMonth(0);
            newDate.setFullYear(newDate.getFullYear()+1);
          }
        } else {
          // we stay in current month 
          newDate.setDate(newDate.getDate()+days);
          return newDate;
        }
      }
  
      return newDate;
    }function _strftime(s, maxsize, format, tm) {
      // size_t strftime(char *restrict s, size_t maxsize, const char *restrict format, const struct tm *restrict timeptr);
      // http://pubs.opengroup.org/onlinepubs/009695399/functions/strftime.html
  
      var tm_zone = HEAP32[(((tm)+(40))>>2)];
  
      var date = {
        tm_sec: HEAP32[((tm)>>2)],
        tm_min: HEAP32[(((tm)+(4))>>2)],
        tm_hour: HEAP32[(((tm)+(8))>>2)],
        tm_mday: HEAP32[(((tm)+(12))>>2)],
        tm_mon: HEAP32[(((tm)+(16))>>2)],
        tm_year: HEAP32[(((tm)+(20))>>2)],
        tm_wday: HEAP32[(((tm)+(24))>>2)],
        tm_yday: HEAP32[(((tm)+(28))>>2)],
        tm_isdst: HEAP32[(((tm)+(32))>>2)],
        tm_gmtoff: HEAP32[(((tm)+(36))>>2)],
        tm_zone: tm_zone ? Pointer_stringify(tm_zone) : ''
      };
  
      var pattern = Pointer_stringify(format);
  
      // expand format
      var EXPANSION_RULES_1 = {
        '%c': '%a %b %d %H:%M:%S %Y',     // Replaced by the locale's appropriate date and time representation - e.g., Mon Aug  3 14:02:01 2013
        '%D': '%m/%d/%y',                 // Equivalent to %m / %d / %y
        '%F': '%Y-%m-%d',                 // Equivalent to %Y - %m - %d
        '%h': '%b',                       // Equivalent to %b
        '%r': '%I:%M:%S %p',              // Replaced by the time in a.m. and p.m. notation
        '%R': '%H:%M',                    // Replaced by the time in 24-hour notation
        '%T': '%H:%M:%S',                 // Replaced by the time
        '%x': '%m/%d/%y',                 // Replaced by the locale's appropriate date representation
        '%X': '%H:%M:%S'                  // Replaced by the locale's appropriate date representation
      };
      for (var rule in EXPANSION_RULES_1) {
        pattern = pattern.replace(new RegExp(rule, 'g'), EXPANSION_RULES_1[rule]);
      }
  
      var WEEKDAYS = ['Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday'];
      var MONTHS = ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December'];
  
      function leadingSomething(value, digits, character) {
        var str = typeof value === 'number' ? value.toString() : (value || '');
        while (str.length < digits) {
          str = character[0]+str;
        }
        return str;
      };
  
      function leadingNulls(value, digits) {
        return leadingSomething(value, digits, '0');
      };
  
      function compareByDay(date1, date2) {
        function sgn(value) {
          return value < 0 ? -1 : (value > 0 ? 1 : 0);
        };
  
        var compare;
        if ((compare = sgn(date1.getFullYear()-date2.getFullYear())) === 0) {
          if ((compare = sgn(date1.getMonth()-date2.getMonth())) === 0) {
            compare = sgn(date1.getDate()-date2.getDate());
          }
        }
        return compare;
      };
  
      function getFirstWeekStartDate(janFourth) {
          switch (janFourth.getDay()) {
            case 0: // Sunday
              return new Date(janFourth.getFullYear()-1, 11, 29);
            case 1: // Monday
              return janFourth;
            case 2: // Tuesday
              return new Date(janFourth.getFullYear(), 0, 3);
            case 3: // Wednesday
              return new Date(janFourth.getFullYear(), 0, 2);
            case 4: // Thursday
              return new Date(janFourth.getFullYear(), 0, 1);
            case 5: // Friday
              return new Date(janFourth.getFullYear()-1, 11, 31);
            case 6: // Saturday
              return new Date(janFourth.getFullYear()-1, 11, 30);
          }
      };
  
      function getWeekBasedYear(date) {
          var thisDate = __addDays(new Date(date.tm_year+1900, 0, 1), date.tm_yday);
  
          var janFourthThisYear = new Date(thisDate.getFullYear(), 0, 4);
          var janFourthNextYear = new Date(thisDate.getFullYear()+1, 0, 4);
  
          var firstWeekStartThisYear = getFirstWeekStartDate(janFourthThisYear);
          var firstWeekStartNextYear = getFirstWeekStartDate(janFourthNextYear);
  
          if (compareByDay(firstWeekStartThisYear, thisDate) <= 0) {
            // this date is after the start of the first week of this year
            if (compareByDay(firstWeekStartNextYear, thisDate) <= 0) {
              return thisDate.getFullYear()+1;
            } else {
              return thisDate.getFullYear();
            }
          } else { 
            return thisDate.getFullYear()-1;
          }
      };
  
      var EXPANSION_RULES_2 = {
        '%a': function(date) {
          return WEEKDAYS[date.tm_wday].substring(0,3);
        },
        '%A': function(date) {
          return WEEKDAYS[date.tm_wday];
        },
        '%b': function(date) {
          return MONTHS[date.tm_mon].substring(0,3);
        },
        '%B': function(date) {
          return MONTHS[date.tm_mon];
        },
        '%C': function(date) {
          var year = date.tm_year+1900;
          return leadingNulls((year/100)|0,2);
        },
        '%d': function(date) {
          return leadingNulls(date.tm_mday, 2);
        },
        '%e': function(date) {
          return leadingSomething(date.tm_mday, 2, ' ');
        },
        '%g': function(date) {
          // %g, %G, and %V give values according to the ISO 8601:2000 standard week-based year. 
          // In this system, weeks begin on a Monday and week 1 of the year is the week that includes 
          // January 4th, which is also the week that includes the first Thursday of the year, and 
          // is also the first week that contains at least four days in the year. 
          // If the first Monday of January is the 2nd, 3rd, or 4th, the preceding days are part of 
          // the last week of the preceding year; thus, for Saturday 2nd January 1999, 
          // %G is replaced by 1998 and %V is replaced by 53. If December 29th, 30th, 
          // or 31st is a Monday, it and any following days are part of week 1 of the following year. 
          // Thus, for Tuesday 30th December 1997, %G is replaced by 1998 and %V is replaced by 01.
          
          return getWeekBasedYear(date).toString().substring(2);
        },
        '%G': function(date) {
          return getWeekBasedYear(date);
        },
        '%H': function(date) {
          return leadingNulls(date.tm_hour, 2);
        },
        '%I': function(date) {
          var twelveHour = date.tm_hour;
          if (twelveHour == 0) twelveHour = 12;
          else if (twelveHour > 12) twelveHour -= 12;
          return leadingNulls(twelveHour, 2);
        },
        '%j': function(date) {
          // Day of the year (001-366)
          return leadingNulls(date.tm_mday+__arraySum(__isLeapYear(date.tm_year+1900) ? __MONTH_DAYS_LEAP : __MONTH_DAYS_REGULAR, date.tm_mon-1), 3);
        },
        '%m': function(date) {
          return leadingNulls(date.tm_mon+1, 2);
        },
        '%M': function(date) {
          return leadingNulls(date.tm_min, 2);
        },
        '%n': function() {
          return '\n';
        },
        '%p': function(date) {
          if (date.tm_hour >= 0 && date.tm_hour < 12) {
            return 'AM';
          } else {
            return 'PM';
          }
        },
        '%S': function(date) {
          return leadingNulls(date.tm_sec, 2);
        },
        '%t': function() {
          return '\t';
        },
        '%u': function(date) {
          var day = new Date(date.tm_year+1900, date.tm_mon+1, date.tm_mday, 0, 0, 0, 0);
          return day.getDay() || 7;
        },
        '%U': function(date) {
          // Replaced by the week number of the year as a decimal number [00,53]. 
          // The first Sunday of January is the first day of week 1; 
          // days in the new year before this are in week 0. [ tm_year, tm_wday, tm_yday]
          var janFirst = new Date(date.tm_year+1900, 0, 1);
          var firstSunday = janFirst.getDay() === 0 ? janFirst : __addDays(janFirst, 7-janFirst.getDay());
          var endDate = new Date(date.tm_year+1900, date.tm_mon, date.tm_mday);
          
          // is target date after the first Sunday?
          if (compareByDay(firstSunday, endDate) < 0) {
            // calculate difference in days between first Sunday and endDate
            var februaryFirstUntilEndMonth = __arraySum(__isLeapYear(endDate.getFullYear()) ? __MONTH_DAYS_LEAP : __MONTH_DAYS_REGULAR, endDate.getMonth()-1)-31;
            var firstSundayUntilEndJanuary = 31-firstSunday.getDate();
            var days = firstSundayUntilEndJanuary+februaryFirstUntilEndMonth+endDate.getDate();
            return leadingNulls(Math.ceil(days/7), 2);
          }
  
          return compareByDay(firstSunday, janFirst) === 0 ? '01': '00';
        },
        '%V': function(date) {
          // Replaced by the week number of the year (Monday as the first day of the week) 
          // as a decimal number [01,53]. If the week containing 1 January has four 
          // or more days in the new year, then it is considered week 1. 
          // Otherwise, it is the last week of the previous year, and the next week is week 1. 
          // Both January 4th and the first Thursday of January are always in week 1. [ tm_year, tm_wday, tm_yday]
          var janFourthThisYear = new Date(date.tm_year+1900, 0, 4);
          var janFourthNextYear = new Date(date.tm_year+1901, 0, 4);
  
          var firstWeekStartThisYear = getFirstWeekStartDate(janFourthThisYear);
          var firstWeekStartNextYear = getFirstWeekStartDate(janFourthNextYear);
  
          var endDate = __addDays(new Date(date.tm_year+1900, 0, 1), date.tm_yday);
  
          if (compareByDay(endDate, firstWeekStartThisYear) < 0) {
            // if given date is before this years first week, then it belongs to the 53rd week of last year
            return '53';
          } 
  
          if (compareByDay(firstWeekStartNextYear, endDate) <= 0) {
            // if given date is after next years first week, then it belongs to the 01th week of next year
            return '01';
          }
  
          // given date is in between CW 01..53 of this calendar year
          var daysDifference;
          if (firstWeekStartThisYear.getFullYear() < date.tm_year+1900) {
            // first CW of this year starts last year
            daysDifference = date.tm_yday+32-firstWeekStartThisYear.getDate()
          } else {
            // first CW of this year starts this year
            daysDifference = date.tm_yday+1-firstWeekStartThisYear.getDate();
          }
          return leadingNulls(Math.ceil(daysDifference/7), 2);
        },
        '%w': function(date) {
          var day = new Date(date.tm_year+1900, date.tm_mon+1, date.tm_mday, 0, 0, 0, 0);
          return day.getDay();
        },
        '%W': function(date) {
          // Replaced by the week number of the year as a decimal number [00,53]. 
          // The first Monday of January is the first day of week 1; 
          // days in the new year before this are in week 0. [ tm_year, tm_wday, tm_yday]
          var janFirst = new Date(date.tm_year, 0, 1);
          var firstMonday = janFirst.getDay() === 1 ? janFirst : __addDays(janFirst, janFirst.getDay() === 0 ? 1 : 7-janFirst.getDay()+1);
          var endDate = new Date(date.tm_year+1900, date.tm_mon, date.tm_mday);
  
          // is target date after the first Monday?
          if (compareByDay(firstMonday, endDate) < 0) {
            var februaryFirstUntilEndMonth = __arraySum(__isLeapYear(endDate.getFullYear()) ? __MONTH_DAYS_LEAP : __MONTH_DAYS_REGULAR, endDate.getMonth()-1)-31;
            var firstMondayUntilEndJanuary = 31-firstMonday.getDate();
            var days = firstMondayUntilEndJanuary+februaryFirstUntilEndMonth+endDate.getDate();
            return leadingNulls(Math.ceil(days/7), 2);
          }
          return compareByDay(firstMonday, janFirst) === 0 ? '01': '00';
        },
        '%y': function(date) {
          // Replaced by the last two digits of the year as a decimal number [00,99]. [ tm_year]
          return (date.tm_year+1900).toString().substring(2);
        },
        '%Y': function(date) {
          // Replaced by the year as a decimal number (for example, 1997). [ tm_year]
          return date.tm_year+1900;
        },
        '%z': function(date) {
          // Replaced by the offset from UTC in the ISO 8601:2000 standard format ( +hhmm or -hhmm ).
          // For example, "-0430" means 4 hours 30 minutes behind UTC (west of Greenwich).
          var off = date.tm_gmtoff;
          var ahead = off >= 0;
          off = Math.abs(off) / 60;
          // convert from minutes into hhmm format (which means 60 minutes = 100 units)
          off = (off / 60)*100 + (off % 60);
          return (ahead ? '+' : '-') + String("0000" + off).slice(-4);
        },
        '%Z': function(date) {
          return date.tm_zone;
        },
        '%%': function() {
          return '%';
        }
      };
      for (var rule in EXPANSION_RULES_2) {
        if (pattern.indexOf(rule) >= 0) {
          pattern = pattern.replace(new RegExp(rule, 'g'), EXPANSION_RULES_2[rule](date));
        }
      }
  
      var bytes = intArrayFromString(pattern, false);
      if (bytes.length > maxsize) {
        return 0;
      } 
  
      writeArrayToMemory(bytes, s);
      return bytes.length-1;
    }function _strftime_l(s, maxsize, format, tm) {
      return _strftime(s, maxsize, format, tm); // no locale support yet
    }

  function _abort() {
      Module['abort']();
    }

  function _emscripten_glVertexAttribDivisor(index, divisor) {
      GLctx['vertexAttribDivisor'](index, divisor);
    }

  function _emscripten_glFramebufferRenderbuffer(target, attachment, renderbuffertarget, renderbuffer) {
      GLctx.framebufferRenderbuffer(target, attachment, renderbuffertarget,
                                         GL.renderbuffers[renderbuffer]);
    }

  function _alSourcePause(source) {
      if (!AL.currentContext) {
        return;
      }
      var src = AL.currentContext.src[source];
      if (!src) {
        AL.currentContext.err = 0xA001 /* AL_INVALID_NAME */;
        return;
      }
      AL.setSourceState(src, 0x1013 /* AL_PAUSED */);
    }

  function _alGenBuffers(count, buffers) {
      if (!AL.currentContext) {
        return;
      }
      for (var i = 0; i < count; ++i) {
        AL.currentContext.buf.push(null);
        HEAP32[(((buffers)+(i*4))>>2)]=AL.currentContext.buf.length;
      }
    }

  function _emscripten_glPolygonOffset(x0, x1) { GLctx.polygonOffset(x0, x1) }

  function _emscripten_glIsBuffer(buffer) {
      var b = GL.buffers[buffer];
      if (!b) return 0;
      return GLctx.isBuffer(b);
    }

  function _emscripten_glUniform2iv(location, count, value) {
      location = GL.uniforms[location];
      count *= 2;
      value = HEAP32.subarray((value)>>2,(value+count*4)>>2);
      GLctx.uniform2iv(location, value);
    }

  
  var PTHREAD_SPECIFIC={};function _pthread_getspecific(key) {
      return PTHREAD_SPECIFIC[key] || 0;
    }

  function _emscripten_glVertexAttrib1fv(index, v) {
      v = HEAPF32.subarray((v)>>2,(v+4)>>2);
      GLctx.vertexAttrib1fv(index, v);
    }

  function _alBufferData(buffer, format, data, size, freq) {
      if (!AL.currentContext) {
        return;
      }
      if (buffer > AL.currentContext.buf.length) {
        return;
      }
      var channels, bytes;
      switch (format) {
      case 0x1100 /* AL_FORMAT_MONO8 */:
        bytes = 1;
        channels = 1;
        break;
      case 0x1101 /* AL_FORMAT_MONO16 */:
        bytes = 2;
        channels = 1;
        break;
      case 0x1102 /* AL_FORMAT_STEREO8 */:
        bytes = 1;
        channels = 2;
        break;
      case 0x1103 /* AL_FORMAT_STEREO16 */:
        bytes = 2;
        channels = 2;
        break;
      case 0x10010 /* AL_FORMAT_MONO_FLOAT32 */:
        bytes = 4;
        channels = 1;
        break;
      case 0x10011 /* AL_FORMAT_STEREO_FLOAT32 */:
        bytes = 4;
        channels = 2;
        break;
      default:
        return;
      }
      try {
        AL.currentContext.buf[buffer - 1] = AL.currentContext.ctx.createBuffer(channels, size / (bytes * channels), freq);
        AL.currentContext.buf[buffer - 1].bytesPerSample =  bytes;
      } catch (e) {
        AL.currentContext.err = 0xA003 /* AL_INVALID_VALUE */;
        return;
      }
      var buf = new Array(channels);
      for (var i = 0; i < channels; ++i) {
        buf[i] = AL.currentContext.buf[buffer - 1].getChannelData(i);
      }
      for (var i = 0; i < size / (bytes * channels); ++i) {
        for (var j = 0; j < channels; ++j) {
          switch (bytes) {
          case 1:
            var val = HEAP8[(((data)+(i*channels+j))>>0)] & 0xff;  // unsigned
            buf[j][i] = -1.0 + val * (2/256);
            break;
          case 2:
            var val = HEAP16[(((data)+(2*(i*channels+j)))>>1)];
            buf[j][i] = val/32768;
            break;
          case 4:
            buf[j][i] = HEAPF32[(((data)+(4*(i*channels+j)))>>2)];
            break;
          }
        }
      }
    }

  function _emscripten_glGetActiveAttrib(program, index, bufSize, length, size, type, name) {
      program = GL.programs[program];
      var info = GLctx.getActiveAttrib(program, index);
      if (!info) return; // If an error occurs, nothing will be written to length, size and type and name.
  
      var infoname = info.name.slice(0, Math.max(0, bufSize - 1));
      if (bufSize > 0 && name) {
        writeStringToMemory(infoname, name);
        if (length) HEAP32[((length)>>2)]=infoname.length;
      } else {
        if (length) HEAP32[((length)>>2)]=0;
      }
  
      if (size) HEAP32[((size)>>2)]=info.size;
      if (type) HEAP32[((type)>>2)]=info.type;
    }

  function _eglDestroySurface(display, surface) { 
      if (display != 62000 /* Magic ID for Emscripten 'default display' */) {
        EGL.setErrorCode(0x3008 /* EGL_BAD_DISPLAY */);
        return 0; 
      }
      if (surface != 62006 /* Magic ID for the only EGLSurface supported by Emscripten */) {
        EGL.setErrorCode(0x300D /* EGL_BAD_SURFACE */);
        return 1;
      }
      if (EGL.currentReadSurface == surface) {
        EGL.currentReadSurface = 0;
      }
      if (EGL.currentDrawSurface == surface) {
        EGL.currentDrawSurface = 0;
      }
      EGL.setErrorCode(0x3000 /* EGL_SUCCESS */);
      return 1; /* Magic ID for Emscripten 'default surface' */
    }


  var _emscripten_asm_const_int=true;

  function _emscripten_glUniform2f(location, v0, v1) {
      location = GL.uniforms[location];
      GLctx.uniform2f(location, v0, v1);
    }

  function ___syscall330(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // dup3
      var old = SYSCALLS.getStreamFromFD(), suggestFD = SYSCALLS.get(), flags = SYSCALLS.get();
      assert(!flags);
      if (old.fd === suggestFD) return -ERRNO_CODES.EINVAL;
      return SYSCALLS.doDup(old.path, old.flags, suggestFD);
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function _emscripten_glUniform2i(location, v0, v1) {
      location = GL.uniforms[location];
      GLctx.uniform2i(location, v0, v1);
    }

  function _emscripten_glDeleteRenderbuffers(n, renderbuffers) {
      for (var i = 0; i < n; i++) {
        var id = HEAP32[(((renderbuffers)+(i*4))>>2)];
        var renderbuffer = GL.renderbuffers[id];
        if (!renderbuffer) continue; // GL spec: "glDeleteRenderbuffers silently ignores 0s and names that do not correspond to existing renderbuffer objects".
        GLctx.deleteRenderbuffer(renderbuffer);
        renderbuffer.name = 0;
        GL.renderbuffers[id] = null;
      }
    }

  function ___cxa_pure_virtual() {
      ABORT = true;
      throw 'Pure virtual function called!';
    }

  
   
  Module["_testSetjmp"] = _testSetjmp;function _longjmp(env, value) {
      asm['setThrew'](env, value || 1);
      throw 'longjmp';
    }

  function _emscripten_glDepthMask(x0) { GLctx.depthMask(x0) }

  function _emscripten_set_mousedown_callback(target, userData, useCapture, callbackfunc) {
      JSEvents.registerMouseEventCallback(target, userData, useCapture, callbackfunc, 5, "mousedown");
      return 0;
    }

  function _emscripten_glDepthRange(x0, x1) { GLctx.depthRange(x0, x1) }

  function _emscripten_set_fullscreenchange_callback(target, userData, useCapture, callbackfunc) {
      if (typeof JSEvents.fullscreenEnabled() === 'undefined') return -1;
      if (!target) target = document;
      else {
        target = JSEvents.findEventTarget(target);
        if (!target) return -4;
      }
      JSEvents.registerFullscreenChangeEventCallback(target, userData, useCapture, callbackfunc, 19, "fullscreenchange");
      JSEvents.registerFullscreenChangeEventCallback(target, userData, useCapture, callbackfunc, 19, "mozfullscreenchange");
      JSEvents.registerFullscreenChangeEventCallback(target, userData, useCapture, callbackfunc, 19, "webkitfullscreenchange");
      JSEvents.registerFullscreenChangeEventCallback(target, userData, useCapture, callbackfunc, 19, "msfullscreenchange");
      return 0;
    }

  function ___syscall140(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // llseek
      var stream = SYSCALLS.getStreamFromFD(), offset_high = SYSCALLS.get(), offset_low = SYSCALLS.get(), result = SYSCALLS.get(), whence = SYSCALLS.get();
      var offset = offset_low;
      assert(offset_high === 0);
      FS.llseek(stream, offset, whence);
      HEAP32[((result)>>2)]=stream.position;
      if (stream.getdents && offset === 0 && whence === 0) stream.getdents = null; // reset readdir state
      return 0;
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function _emscripten_glVertexAttrib1f(x0, x1) { GLctx.vertexAttrib1f(x0, x1) }

  function _emscripten_glGetShaderPrecisionFormat(shaderType, precisionType, range, precision) {
      var result = GLctx.getShaderPrecisionFormat(shaderType, precisionType);
      HEAP32[((range)>>2)]=result.rangeMin;
      HEAP32[(((range)+(4))>>2)]=result.rangeMax;
      HEAP32[((precision)>>2)]=result.precision;
    }

  function _emscripten_glUniform1fv(location, count, value) {
      location = GL.uniforms[location];
      var view;
      if (count === 1) {
        // avoid allocation for the common case of uploading one uniform
        view = GL.miniTempBufferViews[0];
        view[0] = HEAPF32[((value)>>2)];
      } else {
        view = HEAPF32.subarray((value)>>2,(value+count*4)>>2);
      }
      GLctx.uniform1fv(location, view);
    }

  
  var SOCKFS={mount:function (mount) {
        // If Module['websocket'] has already been defined (e.g. for configuring
        // the subprotocol/url) use that, if not initialise it to a new object.
        Module['websocket'] = (Module['websocket'] && 
                               ('object' === typeof Module['websocket'])) ? Module['websocket'] : {};
  
        // Add the Event registration mechanism to the exported websocket configuration
        // object so we can register network callbacks from native JavaScript too.
        // For more documentation see system/include/emscripten/emscripten.h
        Module['websocket']._callbacks = {};
        Module['websocket']['on'] = function(event, callback) {
  	    if ('function' === typeof callback) {
  		  this._callbacks[event] = callback;
          }
  	    return this;
        };
  
        Module['websocket'].emit = function(event, param) {
  	    if ('function' === typeof this._callbacks[event]) {
  		  this._callbacks[event].call(this, param);
          }
        };
  
        // If debug is enabled register simple default logging callbacks for each Event.
  
        return FS.createNode(null, '/', 16384 | 511 /* 0777 */, 0);
      },createSocket:function (family, type, protocol) {
        var streaming = type == 1;
        if (protocol) {
          assert(streaming == (protocol == 6)); // if SOCK_STREAM, must be tcp
        }
  
        // create our internal socket structure
        var sock = {
          family: family,
          type: type,
          protocol: protocol,
          server: null,
          error: null, // Used in getsockopt for SOL_SOCKET/SO_ERROR test
          peers: {},
          pending: [],
          recv_queue: [],
          sock_ops: SOCKFS.websocket_sock_ops
        };
  
        // create the filesystem node to store the socket structure
        var name = SOCKFS.nextname();
        var node = FS.createNode(SOCKFS.root, name, 49152, 0);
        node.sock = sock;
  
        // and the wrapping stream that enables library functions such
        // as read and write to indirectly interact with the socket
        var stream = FS.createStream({
          path: name,
          node: node,
          flags: FS.modeStringToFlags('r+'),
          seekable: false,
          stream_ops: SOCKFS.stream_ops
        });
  
        // map the new stream to the socket structure (sockets have a 1:1
        // relationship with a stream)
        sock.stream = stream;
  
        return sock;
      },getSocket:function (fd) {
        var stream = FS.getStream(fd);
        if (!stream || !FS.isSocket(stream.node.mode)) {
          return null;
        }
        return stream.node.sock;
      },stream_ops:{poll:function (stream) {
          var sock = stream.node.sock;
          return sock.sock_ops.poll(sock);
        },ioctl:function (stream, request, varargs) {
          var sock = stream.node.sock;
          return sock.sock_ops.ioctl(sock, request, varargs);
        },read:function (stream, buffer, offset, length, position /* ignored */) {
          var sock = stream.node.sock;
          var msg = sock.sock_ops.recvmsg(sock, length);
          if (!msg) {
            // socket is closed
            return 0;
          }
          buffer.set(msg.buffer, offset);
          return msg.buffer.length;
        },write:function (stream, buffer, offset, length, position /* ignored */) {
          var sock = stream.node.sock;
          return sock.sock_ops.sendmsg(sock, buffer, offset, length);
        },close:function (stream) {
          var sock = stream.node.sock;
          sock.sock_ops.close(sock);
        }},nextname:function () {
        if (!SOCKFS.nextname.current) {
          SOCKFS.nextname.current = 0;
        }
        return 'socket[' + (SOCKFS.nextname.current++) + ']';
      },websocket_sock_ops:{createPeer:function (sock, addr, port) {
          var ws;
  
          if (typeof addr === 'object') {
            ws = addr;
            addr = null;
            port = null;
          }
  
          if (ws) {
            // for sockets that've already connected (e.g. we're the server)
            // we can inspect the _socket property for the address
            if (ws._socket) {
              addr = ws._socket.remoteAddress;
              port = ws._socket.remotePort;
            }
            // if we're just now initializing a connection to the remote,
            // inspect the url property
            else {
              var result = /ws[s]?:\/\/([^:]+):(\d+)/.exec(ws.url);
              if (!result) {
                throw new Error('WebSocket URL must be in the format ws(s)://address:port');
              }
              addr = result[1];
              port = parseInt(result[2], 10);
            }
          } else {
            // create the actual websocket object and connect
            try {
              // runtimeConfig gets set to true if WebSocket runtime configuration is available.
              var runtimeConfig = (Module['websocket'] && ('object' === typeof Module['websocket']));
  
              // The default value is 'ws://' the replace is needed because the compiler replaces '//' comments with '#'
              // comments without checking context, so we'd end up with ws:#, the replace swaps the '#' for '//' again.
              var url = 'ws:#'.replace('#', '//');
  
              if (runtimeConfig) {
                if ('string' === typeof Module['websocket']['url']) {
                  url = Module['websocket']['url']; // Fetch runtime WebSocket URL config.
                }
              }
  
              if (url === 'ws://' || url === 'wss://') { // Is the supplied URL config just a prefix, if so complete it.
                var parts = addr.split('/');
                url = url + parts[0] + ":" + port + "/" + parts.slice(1).join('/');
              }
  
              // Make the WebSocket subprotocol (Sec-WebSocket-Protocol) default to binary if no configuration is set.
              var subProtocols = 'binary'; // The default value is 'binary'
  
              if (runtimeConfig) {
                if ('string' === typeof Module['websocket']['subprotocol']) {
                  subProtocols = Module['websocket']['subprotocol']; // Fetch runtime WebSocket subprotocol config.
                }
              }
  
              // The regex trims the string (removes spaces at the beginning and end, then splits the string by
              // <any space>,<any space> into an Array. Whitespace removal is important for Websockify and ws.
              subProtocols = subProtocols.replace(/^ +| +$/g,"").split(/ *, */);
  
              // The node ws library API for specifying optional subprotocol is slightly different than the browser's.
              var opts = ENVIRONMENT_IS_NODE ? {'protocol': subProtocols.toString()} : subProtocols;
  
              // If node we use the ws library.
              var WebSocketConstructor;
              if (ENVIRONMENT_IS_NODE) {
                WebSocketConstructor = require('ws');
              } else if (ENVIRONMENT_IS_WEB) {
                WebSocketConstructor = window['WebSocket'];
              } else {
                WebSocketConstructor = WebSocket;
              }
              ws = new WebSocketConstructor(url, opts);
              ws.binaryType = 'arraybuffer';
            } catch (e) {
              throw new FS.ErrnoError(ERRNO_CODES.EHOSTUNREACH);
            }
          }
  
  
          var peer = {
            addr: addr,
            port: port,
            socket: ws,
            dgram_send_queue: []
          };
  
          SOCKFS.websocket_sock_ops.addPeer(sock, peer);
          SOCKFS.websocket_sock_ops.handlePeerEvents(sock, peer);
  
          // if this is a bound dgram socket, send the port number first to allow
          // us to override the ephemeral port reported to us by remotePort on the
          // remote end.
          if (sock.type === 2 && typeof sock.sport !== 'undefined') {
            peer.dgram_send_queue.push(new Uint8Array([
                255, 255, 255, 255,
                'p'.charCodeAt(0), 'o'.charCodeAt(0), 'r'.charCodeAt(0), 't'.charCodeAt(0),
                ((sock.sport & 0xff00) >> 8) , (sock.sport & 0xff)
            ]));
          }
  
          return peer;
        },getPeer:function (sock, addr, port) {
          return sock.peers[addr + ':' + port];
        },addPeer:function (sock, peer) {
          sock.peers[peer.addr + ':' + peer.port] = peer;
        },removePeer:function (sock, peer) {
          delete sock.peers[peer.addr + ':' + peer.port];
        },handlePeerEvents:function (sock, peer) {
          var first = true;
  
          var handleOpen = function () {
  
            Module['websocket'].emit('open', sock.stream.fd);
  
            try {
              var queued = peer.dgram_send_queue.shift();
              while (queued) {
                peer.socket.send(queued);
                queued = peer.dgram_send_queue.shift();
              }
            } catch (e) {
              // not much we can do here in the way of proper error handling as we've already
              // lied and said this data was sent. shut it down.
              peer.socket.close();
            }
          };
  
          function handleMessage(data) {
            assert(typeof data !== 'string' && data.byteLength !== undefined);  // must receive an ArrayBuffer
            data = new Uint8Array(data);  // make a typed array view on the array buffer
  
  
            // if this is the port message, override the peer's port with it
            var wasfirst = first;
            first = false;
            if (wasfirst &&
                data.length === 10 &&
                data[0] === 255 && data[1] === 255 && data[2] === 255 && data[3] === 255 &&
                data[4] === 'p'.charCodeAt(0) && data[5] === 'o'.charCodeAt(0) && data[6] === 'r'.charCodeAt(0) && data[7] === 't'.charCodeAt(0)) {
              // update the peer's port and it's key in the peer map
              var newport = ((data[8] << 8) | data[9]);
              SOCKFS.websocket_sock_ops.removePeer(sock, peer);
              peer.port = newport;
              SOCKFS.websocket_sock_ops.addPeer(sock, peer);
              return;
            }
  
            sock.recv_queue.push({ addr: peer.addr, port: peer.port, data: data });
            Module['websocket'].emit('message', sock.stream.fd);
          };
  
          if (ENVIRONMENT_IS_NODE) {
            peer.socket.on('open', handleOpen);
            peer.socket.on('message', function(data, flags) {
              if (!flags.binary) {
                return;
              }
              handleMessage((new Uint8Array(data)).buffer);  // copy from node Buffer -> ArrayBuffer
            });
            peer.socket.on('close', function() {
              Module['websocket'].emit('close', sock.stream.fd);
            });
            peer.socket.on('error', function(error) {
              // Although the ws library may pass errors that may be more descriptive than
              // ECONNREFUSED they are not necessarily the expected error code e.g. 
              // ENOTFOUND on getaddrinfo seems to be node.js specific, so using ECONNREFUSED
              // is still probably the most useful thing to do.
              sock.error = ERRNO_CODES.ECONNREFUSED; // Used in getsockopt for SOL_SOCKET/SO_ERROR test.
              Module['websocket'].emit('error', [sock.stream.fd, sock.error, 'ECONNREFUSED: Connection refused']);
              // don't throw
            });
          } else {
            peer.socket.onopen = handleOpen;
            peer.socket.onclose = function() {
              Module['websocket'].emit('close', sock.stream.fd);
            };
            peer.socket.onmessage = function peer_socket_onmessage(event) {
              handleMessage(event.data);
            };
            peer.socket.onerror = function(error) {
              // The WebSocket spec only allows a 'simple event' to be thrown on error,
              // so we only really know as much as ECONNREFUSED.
              sock.error = ERRNO_CODES.ECONNREFUSED; // Used in getsockopt for SOL_SOCKET/SO_ERROR test.
              Module['websocket'].emit('error', [sock.stream.fd, sock.error, 'ECONNREFUSED: Connection refused']);
            };
          }
        },poll:function (sock) {
          if (sock.type === 1 && sock.server) {
            // listen sockets should only say they're available for reading
            // if there are pending clients.
            return sock.pending.length ? (64 | 1) : 0;
          }
  
          var mask = 0;
          var dest = sock.type === 1 ?  // we only care about the socket state for connection-based sockets
            SOCKFS.websocket_sock_ops.getPeer(sock, sock.daddr, sock.dport) :
            null;
  
          if (sock.recv_queue.length ||
              !dest ||  // connection-less sockets are always ready to read
              (dest && dest.socket.readyState === dest.socket.CLOSING) ||
              (dest && dest.socket.readyState === dest.socket.CLOSED)) {  // let recv return 0 once closed
            mask |= (64 | 1);
          }
  
          if (!dest ||  // connection-less sockets are always ready to write
              (dest && dest.socket.readyState === dest.socket.OPEN)) {
            mask |= 4;
          }
  
          if ((dest && dest.socket.readyState === dest.socket.CLOSING) ||
              (dest && dest.socket.readyState === dest.socket.CLOSED)) {
            mask |= 16;
          }
  
          return mask;
        },ioctl:function (sock, request, arg) {
          switch (request) {
            case 21531:
              var bytes = 0;
              if (sock.recv_queue.length) {
                bytes = sock.recv_queue[0].data.length;
              }
              HEAP32[((arg)>>2)]=bytes;
              return 0;
            default:
              return ERRNO_CODES.EINVAL;
          }
        },close:function (sock) {
          // if we've spawned a listen server, close it
          if (sock.server) {
            try {
              sock.server.close();
            } catch (e) {
            }
            sock.server = null;
          }
          // close any peer connections
          var peers = Object.keys(sock.peers);
          for (var i = 0; i < peers.length; i++) {
            var peer = sock.peers[peers[i]];
            try {
              peer.socket.close();
            } catch (e) {
            }
            SOCKFS.websocket_sock_ops.removePeer(sock, peer);
          }
          return 0;
        },bind:function (sock, addr, port) {
          if (typeof sock.saddr !== 'undefined' || typeof sock.sport !== 'undefined') {
            throw new FS.ErrnoError(ERRNO_CODES.EINVAL);  // already bound
          }
          sock.saddr = addr;
          sock.sport = port;
          // in order to emulate dgram sockets, we need to launch a listen server when
          // binding on a connection-less socket
          // note: this is only required on the server side
          if (sock.type === 2) {
            // close the existing server if it exists
            if (sock.server) {
              sock.server.close();
              sock.server = null;
            }
            // swallow error operation not supported error that occurs when binding in the
            // browser where this isn't supported
            try {
              sock.sock_ops.listen(sock, 0);
            } catch (e) {
              if (!(e instanceof FS.ErrnoError)) throw e;
              if (e.errno !== ERRNO_CODES.EOPNOTSUPP) throw e;
            }
          }
        },connect:function (sock, addr, port) {
          if (sock.server) {
            throw new FS.ErrnoError(ERRNO_CODES.EOPNOTSUPP);
          }
  
          // TODO autobind
          // if (!sock.addr && sock.type == 2) {
          // }
  
          // early out if we're already connected / in the middle of connecting
          if (typeof sock.daddr !== 'undefined' && typeof sock.dport !== 'undefined') {
            var dest = SOCKFS.websocket_sock_ops.getPeer(sock, sock.daddr, sock.dport);
            if (dest) {
              if (dest.socket.readyState === dest.socket.CONNECTING) {
                throw new FS.ErrnoError(ERRNO_CODES.EALREADY);
              } else {
                throw new FS.ErrnoError(ERRNO_CODES.EISCONN);
              }
            }
          }
  
          // add the socket to our peer list and set our
          // destination address / port to match
          var peer = SOCKFS.websocket_sock_ops.createPeer(sock, addr, port);
          sock.daddr = peer.addr;
          sock.dport = peer.port;
  
          // always "fail" in non-blocking mode
          throw new FS.ErrnoError(ERRNO_CODES.EINPROGRESS);
        },listen:function (sock, backlog) {
          if (!ENVIRONMENT_IS_NODE) {
            throw new FS.ErrnoError(ERRNO_CODES.EOPNOTSUPP);
          }
          if (sock.server) {
             throw new FS.ErrnoError(ERRNO_CODES.EINVAL);  // already listening
          }
          var WebSocketServer = require('ws').Server;
          var host = sock.saddr;
          sock.server = new WebSocketServer({
            host: host,
            port: sock.sport
            // TODO support backlog
          });
          Module['websocket'].emit('listen', sock.stream.fd); // Send Event with listen fd.
  
          sock.server.on('connection', function(ws) {
            if (sock.type === 1) {
              var newsock = SOCKFS.createSocket(sock.family, sock.type, sock.protocol);
  
              // create a peer on the new socket
              var peer = SOCKFS.websocket_sock_ops.createPeer(newsock, ws);
              newsock.daddr = peer.addr;
              newsock.dport = peer.port;
  
              // push to queue for accept to pick up
              sock.pending.push(newsock);
              Module['websocket'].emit('connection', newsock.stream.fd);
            } else {
              // create a peer on the listen socket so calling sendto
              // with the listen socket and an address will resolve
              // to the correct client
              SOCKFS.websocket_sock_ops.createPeer(sock, ws);
              Module['websocket'].emit('connection', sock.stream.fd);
            }
          });
          sock.server.on('closed', function() {
            Module['websocket'].emit('close', sock.stream.fd);
            sock.server = null;
          });
          sock.server.on('error', function(error) {
            // Although the ws library may pass errors that may be more descriptive than
            // ECONNREFUSED they are not necessarily the expected error code e.g. 
            // ENOTFOUND on getaddrinfo seems to be node.js specific, so using EHOSTUNREACH
            // is still probably the most useful thing to do. This error shouldn't
            // occur in a well written app as errors should get trapped in the compiled
            // app's own getaddrinfo call.
            sock.error = ERRNO_CODES.EHOSTUNREACH; // Used in getsockopt for SOL_SOCKET/SO_ERROR test.
            Module['websocket'].emit('error', [sock.stream.fd, sock.error, 'EHOSTUNREACH: Host is unreachable']);
            // don't throw
          });
        },accept:function (listensock) {
          if (!listensock.server) {
            throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
          }
          var newsock = listensock.pending.shift();
          newsock.stream.flags = listensock.stream.flags;
          return newsock;
        },getname:function (sock, peer) {
          var addr, port;
          if (peer) {
            if (sock.daddr === undefined || sock.dport === undefined) {
              throw new FS.ErrnoError(ERRNO_CODES.ENOTCONN);
            }
            addr = sock.daddr;
            port = sock.dport;
          } else {
            // TODO saddr and sport will be set for bind()'d UDP sockets, but what
            // should we be returning for TCP sockets that've been connect()'d?
            addr = sock.saddr || 0;
            port = sock.sport || 0;
          }
          return { addr: addr, port: port };
        },sendmsg:function (sock, buffer, offset, length, addr, port) {
          if (sock.type === 2) {
            // connection-less sockets will honor the message address,
            // and otherwise fall back to the bound destination address
            if (addr === undefined || port === undefined) {
              addr = sock.daddr;
              port = sock.dport;
            }
            // if there was no address to fall back to, error out
            if (addr === undefined || port === undefined) {
              throw new FS.ErrnoError(ERRNO_CODES.EDESTADDRREQ);
            }
          } else {
            // connection-based sockets will only use the bound
            addr = sock.daddr;
            port = sock.dport;
          }
  
          // find the peer for the destination address
          var dest = SOCKFS.websocket_sock_ops.getPeer(sock, addr, port);
  
          // early out if not connected with a connection-based socket
          if (sock.type === 1) {
            if (!dest || dest.socket.readyState === dest.socket.CLOSING || dest.socket.readyState === dest.socket.CLOSED) {
              throw new FS.ErrnoError(ERRNO_CODES.ENOTCONN);
            } else if (dest.socket.readyState === dest.socket.CONNECTING) {
              throw new FS.ErrnoError(ERRNO_CODES.EAGAIN);
            }
          }
  
          // create a copy of the incoming data to send, as the WebSocket API
          // doesn't work entirely with an ArrayBufferView, it'll just send
          // the entire underlying buffer
          var data;
          if (buffer instanceof Array || buffer instanceof ArrayBuffer) {
            data = buffer.slice(offset, offset + length);
          } else {  // ArrayBufferView
            data = buffer.buffer.slice(buffer.byteOffset + offset, buffer.byteOffset + offset + length);
          }
  
          // if we're emulating a connection-less dgram socket and don't have
          // a cached connection, queue the buffer to send upon connect and
          // lie, saying the data was sent now.
          if (sock.type === 2) {
            if (!dest || dest.socket.readyState !== dest.socket.OPEN) {
              // if we're not connected, open a new connection
              if (!dest || dest.socket.readyState === dest.socket.CLOSING || dest.socket.readyState === dest.socket.CLOSED) {
                dest = SOCKFS.websocket_sock_ops.createPeer(sock, addr, port);
              }
              dest.dgram_send_queue.push(data);
              return length;
            }
          }
  
          try {
            // send the actual data
            dest.socket.send(data);
            return length;
          } catch (e) {
            throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
          }
        },recvmsg:function (sock, length) {
          // http://pubs.opengroup.org/onlinepubs/7908799/xns/recvmsg.html
          if (sock.type === 1 && sock.server) {
            // tcp servers should not be recv()'ing on the listen socket
            throw new FS.ErrnoError(ERRNO_CODES.ENOTCONN);
          }
  
          var queued = sock.recv_queue.shift();
          if (!queued) {
            if (sock.type === 1) {
              var dest = SOCKFS.websocket_sock_ops.getPeer(sock, sock.daddr, sock.dport);
  
              if (!dest) {
                // if we have a destination address but are not connected, error out
                throw new FS.ErrnoError(ERRNO_CODES.ENOTCONN);
              }
              else if (dest.socket.readyState === dest.socket.CLOSING || dest.socket.readyState === dest.socket.CLOSED) {
                // return null if the socket has closed
                return null;
              }
              else {
                // else, our socket is in a valid state but truly has nothing available
                throw new FS.ErrnoError(ERRNO_CODES.EAGAIN);
              }
            } else {
              throw new FS.ErrnoError(ERRNO_CODES.EAGAIN);
            }
          }
  
          // queued.data will be an ArrayBuffer if it's unadulterated, but if it's
          // requeued TCP data it'll be an ArrayBufferView
          var queuedLength = queued.data.byteLength || queued.data.length;
          var queuedOffset = queued.data.byteOffset || 0;
          var queuedBuffer = queued.data.buffer || queued.data;
          var bytesRead = Math.min(length, queuedLength);
          var res = {
            buffer: new Uint8Array(queuedBuffer, queuedOffset, bytesRead),
            addr: queued.addr,
            port: queued.port
          };
  
  
          // push back any unread data for TCP connections
          if (sock.type === 1 && bytesRead < queuedLength) {
            var bytesRemaining = queuedLength - bytesRead;
            queued.data = new Uint8Array(queuedBuffer, queuedOffset + bytesRead, bytesRemaining);
            sock.recv_queue.unshift(queued);
          }
  
          return res;
        }}};
  
  
  function __inet_pton4_raw(str) {
      var b = str.split('.');
      for (var i = 0; i < 4; i++) {
        var tmp = Number(b[i]);
        if (isNaN(tmp)) return null;
        b[i] = tmp;
      }
      return (b[0] | (b[1] << 8) | (b[2] << 16) | (b[3] << 24)) >>> 0;
    }
  
  
  var _htons=undefined;
  Module["_htons"] = _htons;function __inet_pton6_raw(str) {
      var words;
      var w, offset, z, i;
      /* http://home.deds.nl/~aeron/regex/ */
      var valid6regx = /^((?=.*::)(?!.*::.+::)(::)?([\dA-F]{1,4}:(:|\b)|){5}|([\dA-F]{1,4}:){6})((([\dA-F]{1,4}((?!\3)::|:\b|$))|(?!\2\3)){2}|(((2[0-4]|1\d|[1-9])?\d|25[0-5])\.?\b){4})$/i
      var parts = [];
      if (!valid6regx.test(str)) {
        return null;
      }
      if (str === "::") {
        return [0, 0, 0, 0, 0, 0, 0, 0];
      }
      // Z placeholder to keep track of zeros when splitting the string on ":"
      if (str.indexOf("::") === 0) {
        str = str.replace("::", "Z:"); // leading zeros case
      } else {
        str = str.replace("::", ":Z:");
      }
  
      if (str.indexOf(".") > 0) {
        // parse IPv4 embedded stress
        str = str.replace(new RegExp('[.]', 'g'), ":");
        words = str.split(":");
        words[words.length-4] = parseInt(words[words.length-4]) + parseInt(words[words.length-3])*256;
        words[words.length-3] = parseInt(words[words.length-2]) + parseInt(words[words.length-1])*256;
        words = words.slice(0, words.length-2);
      } else {
        words = str.split(":");
      }
  
      offset = 0; z = 0;
      for (w=0; w < words.length; w++) {
        if (typeof words[w] === 'string') {
          if (words[w] === 'Z') {
            // compressed zeros - write appropriate number of zero words
            for (z = 0; z < (8 - words.length+1); z++) {
              parts[w+z] = 0;
            }
            offset = z-1;
          } else {
            // parse hex to field to 16-bit value and write it in network byte-order
            parts[w+offset] = _htons(parseInt(words[w],16));
          }
        } else {
          // parsed IPv4 words
          parts[w+offset] = words[w];
        }
      }
      return [
        (parts[1] << 16) | parts[0],
        (parts[3] << 16) | parts[2],
        (parts[5] << 16) | parts[4],
        (parts[7] << 16) | parts[6]
      ];
    }var DNS={address_map:{id:1,addrs:{},names:{}},lookup_name:function (name) {
        // If the name is already a valid ipv4 / ipv6 address, don't generate a fake one.
        var res = __inet_pton4_raw(name);
        if (res !== null) {
          return name;
        }
        res = __inet_pton6_raw(name);
        if (res !== null) {
          return name;
        }
  
        // See if this name is already mapped.
        var addr;
  
        if (DNS.address_map.addrs[name]) {
          addr = DNS.address_map.addrs[name];
        } else {
          var id = DNS.address_map.id++;
          assert(id < 65535, 'exceeded max address mappings of 65535');
  
          addr = '172.29.' + (id & 0xff) + '.' + (id & 0xff00);
  
          DNS.address_map.names[addr] = name;
          DNS.address_map.addrs[name] = addr;
        }
  
        return addr;
      },lookup_addr:function (addr) {
        if (DNS.address_map.names[addr]) {
          return DNS.address_map.names[addr];
        }
  
        return null;
      }};
  
  
  var Sockets={BUFFER_SIZE:10240,MAX_BUFFER_SIZE:10485760,nextFd:1,fds:{},nextport:1,maxport:65535,peer:null,connections:{},portmap:{},localAddr:4261412874,addrPool:[33554442,50331658,67108874,83886090,100663306,117440522,134217738,150994954,167772170,184549386,201326602,218103818,234881034]};
  
  function __inet_ntop4_raw(addr) {
      return (addr & 0xff) + '.' + ((addr >> 8) & 0xff) + '.' + ((addr >> 16) & 0xff) + '.' + ((addr >> 24) & 0xff)
    }
  
  
  var _ntohs=undefined;
  Module["_ntohs"] = _ntohs;function __inet_ntop6_raw(ints) {
      //  ref:  http://www.ietf.org/rfc/rfc2373.txt - section 2.5.4
      //  Format for IPv4 compatible and mapped  128-bit IPv6 Addresses
      //  128-bits are split into eight 16-bit words
      //  stored in network byte order (big-endian)
      //  |                80 bits               | 16 |      32 bits        |
      //  +-----------------------------------------------------------------+
      //  |               10 bytes               |  2 |      4 bytes        |
      //  +--------------------------------------+--------------------------+
      //  +               5 words                |  1 |      2 words        |
      //  +--------------------------------------+--------------------------+
      //  |0000..............................0000|0000|    IPv4 ADDRESS     | (compatible)
      //  +--------------------------------------+----+---------------------+
      //  |0000..............................0000|FFFF|    IPv4 ADDRESS     | (mapped)
      //  +--------------------------------------+----+---------------------+
      var str = "";
      var word = 0;
      var longest = 0;
      var lastzero = 0;
      var zstart = 0;
      var len = 0;
      var i = 0;
      var parts = [
        ints[0] & 0xffff,
        (ints[0] >> 16),
        ints[1] & 0xffff,
        (ints[1] >> 16),
        ints[2] & 0xffff,
        (ints[2] >> 16),
        ints[3] & 0xffff,
        (ints[3] >> 16)
      ];
  
      // Handle IPv4-compatible, IPv4-mapped, loopback and any/unspecified addresses
  
      var hasipv4 = true;
      var v4part = "";
      // check if the 10 high-order bytes are all zeros (first 5 words)
      for (i = 0; i < 5; i++) {
        if (parts[i] !== 0) { hasipv4 = false; break; }
      }
  
      if (hasipv4) {
        // low-order 32-bits store an IPv4 address (bytes 13 to 16) (last 2 words)
        v4part = __inet_ntop4_raw(parts[6] | (parts[7] << 16));
        // IPv4-mapped IPv6 address if 16-bit value (bytes 11 and 12) == 0xFFFF (6th word)
        if (parts[5] === -1) {
          str = "::ffff:";
          str += v4part;
          return str;
        }
        // IPv4-compatible IPv6 address if 16-bit value (bytes 11 and 12) == 0x0000 (6th word)
        if (parts[5] === 0) {
          str = "::";
          //special case IPv6 addresses
          if(v4part === "0.0.0.0") v4part = ""; // any/unspecified address
          if(v4part === "0.0.0.1") v4part = "1";// loopback address
          str += v4part;
          return str;
        }
      }
  
      // Handle all other IPv6 addresses
  
      // first run to find the longest contiguous zero words
      for (word = 0; word < 8; word++) {
        if (parts[word] === 0) {
          if (word - lastzero > 1) {
            len = 0;
          }
          lastzero = word;
          len++;
        }
        if (len > longest) {
          longest = len;
          zstart = word - longest + 1;
        }
      }
  
      for (word = 0; word < 8; word++) {
        if (longest > 1) {
          // compress contiguous zeros - to produce "::"
          if (parts[word] === 0 && word >= zstart && word < (zstart + longest) ) {
            if (word === zstart) {
              str += ":";
              if (zstart === 0) str += ":"; //leading zeros case
            }
            continue;
          }
        }
        // converts 16-bit words from big-endian to little-endian before converting to hex string
        str += Number(_ntohs(parts[word] & 0xffff)).toString(16);
        str += word < 7 ? ":" : "";
      }
      return str;
    }function __read_sockaddr(sa, salen) {
      // family / port offsets are common to both sockaddr_in and sockaddr_in6
      var family = HEAP16[((sa)>>1)];
      var port = _ntohs(HEAP16[(((sa)+(2))>>1)]);
      var addr;
  
      switch (family) {
        case 2:
          if (salen !== 16) {
            return { errno: ERRNO_CODES.EINVAL };
          }
          addr = HEAP32[(((sa)+(4))>>2)];
          addr = __inet_ntop4_raw(addr);
          break;
        case 10:
          if (salen !== 28) {
            return { errno: ERRNO_CODES.EINVAL };
          }
          addr = [
            HEAP32[(((sa)+(8))>>2)],
            HEAP32[(((sa)+(12))>>2)],
            HEAP32[(((sa)+(16))>>2)],
            HEAP32[(((sa)+(20))>>2)]
          ];
          addr = __inet_ntop6_raw(addr);
          break;
        default:
          return { errno: ERRNO_CODES.EAFNOSUPPORT };
      }
  
      return { family: family, addr: addr, port: port };
    }
  
  function __write_sockaddr(sa, family, addr, port) {
      switch (family) {
        case 2:
          addr = __inet_pton4_raw(addr);
          HEAP16[((sa)>>1)]=family;
          HEAP32[(((sa)+(4))>>2)]=addr;
          HEAP16[(((sa)+(2))>>1)]=_htons(port);
          break;
        case 10:
          addr = __inet_pton6_raw(addr);
          HEAP32[((sa)>>2)]=family;
          HEAP32[(((sa)+(8))>>2)]=addr[0];
          HEAP32[(((sa)+(12))>>2)]=addr[1];
          HEAP32[(((sa)+(16))>>2)]=addr[2];
          HEAP32[(((sa)+(20))>>2)]=addr[3];
          HEAP16[(((sa)+(2))>>1)]=_htons(port);
          HEAP32[(((sa)+(4))>>2)]=0;
          HEAP32[(((sa)+(24))>>2)]=0;
          break;
        default:
          return { errno: ERRNO_CODES.EAFNOSUPPORT };
      }
      // kind of lame, but let's match _read_sockaddr's interface
      return {};
    }function ___syscall102(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // socketcall
      var call = SYSCALLS.get(), socketvararg = SYSCALLS.get();
      // socketcalls pass the rest of the arguments in a struct
      SYSCALLS.varargs = socketvararg;
      switch (call) {
        case 1: { // socket
          var domain = SYSCALLS.get(), type = SYSCALLS.get(), protocol = SYSCALLS.get();
          var sock = SOCKFS.createSocket(domain, type, protocol);
          assert(sock.stream.fd < 64); // XXX ? select() assumes socket fd values are in 0..63
          return sock.stream.fd;
        }
        case 2: { // bind
          var sock = SYSCALLS.getSocketFromFD(), info = SYSCALLS.getSocketAddress();
          sock.sock_ops.bind(sock, info.addr, info.port);
          return 0;
        }
        case 3: { // connect
          var sock = SYSCALLS.getSocketFromFD(), info = SYSCALLS.getSocketAddress();
          sock.sock_ops.connect(sock, info.addr, info.port);
          return 0;
        }
        case 4: { // listen
          var sock = SYSCALLS.getSocketFromFD(), backlog = SYSCALLS.get();
          sock.sock_ops.listen(sock, backlog);
          return 0;
        }
        case 5: { // accept
          var sock = SYSCALLS.getSocketFromFD(), addr = SYSCALLS.get(), addrlen = SYSCALLS.get();
          var newsock = sock.sock_ops.accept(sock);
          if (addr) {
            var res = __write_sockaddr(addr, newsock.family, DNS.lookup_name(newsock.daddr), newsock.dport);
            assert(!res.errno);
          }
          return newsock.stream.fd;
        }
        case 6: { // getsockname
          var sock = SYSCALLS.getSocketFromFD(), addr = SYSCALLS.get(), addrlen = SYSCALLS.get();
          // TODO: sock.saddr should never be undefined, see TODO in websocket_sock_ops.getname
          var res = __write_sockaddr(addr, sock.family, DNS.lookup_name(sock.saddr || '0.0.0.0'), sock.sport);
          assert(!res.errno);
          return 0;
        }
        case 7: { // getpeername
          var sock = SYSCALLS.getSocketFromFD(), addr = SYSCALLS.get(), addrlen = SYSCALLS.get();
          if (!sock.daddr) {
            return -ERRNO_CODES.ENOTCONN; // The socket is not connected.
          }
          var res = __write_sockaddr(addr, sock.family, DNS.lookup_name(sock.daddr), sock.dport);
          assert(!res.errno);
          return 0;
        }
        case 11: { // sendto
          var sock = SYSCALLS.getSocketFromFD(), message = SYSCALLS.get(), length = SYSCALLS.get(), flags = SYSCALLS.get(), dest = SYSCALLS.getSocketAddress(true);
          if (!dest) {
            // send, no address provided
            return FS.write(sock.stream, HEAP8,message, length);
          } else {
            // sendto an address
            return sock.sock_ops.sendmsg(sock, HEAP8,message, length, dest.addr, dest.port);
          }
        }
        case 12: { // recvfrom
          var sock = SYSCALLS.getSocketFromFD(), buf = SYSCALLS.get(), len = SYSCALLS.get(), flags = SYSCALLS.get(), addr = SYSCALLS.get(), addrlen = SYSCALLS.get();
          var msg = sock.sock_ops.recvmsg(sock, len);
          if (!msg) return 0; // socket is closed
          if (addr) {
            var res = __write_sockaddr(addr, sock.family, DNS.lookup_name(msg.addr), msg.port);
            assert(!res.errno);
          }
          HEAPU8.set(msg.buffer, buf);
          return msg.buffer.byteLength;
        }
        case 14: { // setsockopt
          return -ERRNO_CODES.ENOPROTOOPT; // The option is unknown at the level indicated.
        }
        case 15: { // getsockopt
          var sock = SYSCALLS.getSocketFromFD(), level = SYSCALLS.get(), optname = SYSCALLS.get(), optval = SYSCALLS.get(), optlen = SYSCALLS.get();
          // Minimal getsockopt aimed at resolving https://github.com/kripken/emscripten/issues/2211
          // so only supports SOL_SOCKET with SO_ERROR.
          if (level === 1) {
            if (optname === 4) {
              HEAP32[((optval)>>2)]=sock.error;
              HEAP32[((optlen)>>2)]=4;
              sock.error = null; // Clear the error (The SO_ERROR option obtains and then clears this field).
              return 0;
            }
          }
          return -ERRNO_CODES.ENOPROTOOPT; // The option is unknown at the level indicated.
        }
        case 16: { // sendmsg
          var sock = SYSCALLS.getSocketFromFD(), message = SYSCALLS.get(), flags = SYSCALLS.get();
          var iov = HEAP32[(((message)+(8))>>2)];
          var num = HEAP32[(((message)+(12))>>2)];
          // read the address and port to send to
          var addr, port;
          var name = HEAP32[((message)>>2)];
          var namelen = HEAP32[(((message)+(4))>>2)];
          if (name) {
            var info = __read_sockaddr(name, namelen);
            if (info.errno) return -info.errno;
            port = info.port;
            addr = DNS.lookup_addr(info.addr) || info.addr;
          }
          // concatenate scatter-gather arrays into one message buffer
          var total = 0;
          for (var i = 0; i < num; i++) {
            total += HEAP32[(((iov)+((8 * i) + 4))>>2)];
          }
          var view = new Uint8Array(total);
          var offset = 0;
          for (var i = 0; i < num; i++) {
            var iovbase = HEAP32[(((iov)+((8 * i) + 0))>>2)];
            var iovlen = HEAP32[(((iov)+((8 * i) + 4))>>2)];
            for (var j = 0; j < iovlen; j++) {  
              view[offset++] = HEAP8[(((iovbase)+(j))>>0)];
            }
          }
          // write the buffer
          return sock.sock_ops.sendmsg(sock, view, 0, total, addr, port);
        }
        case 17: { // recvmsg
          var sock = SYSCALLS.getSocketFromFD(), message = SYSCALLS.get(), flags = SYSCALLS.get();
          var iov = HEAP32[(((message)+(8))>>2)];
          var num = HEAP32[(((message)+(12))>>2)];
          // get the total amount of data we can read across all arrays
          var total = 0;
          for (var i = 0; i < num; i++) {
            total += HEAP32[(((iov)+((8 * i) + 4))>>2)];
          }
          // try to read total data
          var msg = sock.sock_ops.recvmsg(sock, total);
          if (!msg) return 0; // socket is closed
  
          // TODO honor flags:
          // MSG_OOB
          // Requests out-of-band data. The significance and semantics of out-of-band data are protocol-specific.
          // MSG_PEEK
          // Peeks at the incoming message.
          // MSG_WAITALL
          // Requests that the function block until the full amount of data requested can be returned. The function may return a smaller amount of data if a signal is caught, if the connection is terminated, if MSG_PEEK was specified, or if an error is pending for the socket.
  
          // write the source address out
          var name = HEAP32[((message)>>2)];
          if (name) {
            var res = __write_sockaddr(name, sock.family, DNS.lookup_name(msg.addr), msg.port);
            assert(!res.errno);
          }
          // write the buffer out to the scatter-gather arrays
          var bytesRead = 0;
          var bytesRemaining = msg.buffer.byteLength;
          for (var i = 0; bytesRemaining > 0 && i < num; i++) {
            var iovbase = HEAP32[(((iov)+((8 * i) + 0))>>2)];
            var iovlen = HEAP32[(((iov)+((8 * i) + 4))>>2)];
            if (!iovlen) {
              continue;
            }
            var length = Math.min(iovlen, bytesRemaining);
            var buf = msg.buffer.subarray(bytesRead, bytesRead + length);
            HEAPU8.set(buf, iovbase + bytesRead);
            bytesRead += length;
            bytesRemaining -= length;
          }
  
          // TODO set msghdr.msg_flags
          // MSG_EOR
          // End of record was received (if supported by the protocol).
          // MSG_OOB
          // Out-of-band data was received.
          // MSG_TRUNC
          // Normal data was truncated.
          // MSG_CTRUNC
  
          return bytesRead;
        }
        default: abort('unsupported socketcall syscall ' + call);
      }
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function _emscripten_set_wheel_callback(target, userData, useCapture, callbackfunc) {
      target = JSEvents.findEventTarget(target);
      if (typeof target.onwheel !== 'undefined') {
        JSEvents.registerWheelEventCallback(target, userData, useCapture, callbackfunc, 9, "wheel");
        return 0;
      } else if (typeof target.onmousewheel !== 'undefined') {
        JSEvents.registerWheelEventCallback(target, userData, useCapture, callbackfunc, 9, "mousewheel");
        return 0;
      } else {
        return -1;
      }
    }

  function _emscripten_set_gamepaddisconnected_callback(userData, useCapture, callbackfunc) {
      if (!navigator.getGamepads && !navigator.webkitGetGamepads) return -1;
      JSEvents.registerGamepadEventCallback(window, userData, useCapture, callbackfunc, 27, "gamepaddisconnected");
      return 0;
   }

  function _emscripten_set_mouseenter_callback(target, userData, useCapture, callbackfunc) {
      JSEvents.registerMouseEventCallback(target, userData, useCapture, callbackfunc, 33, "mouseenter");
      return 0;
    }

  function _emscripten_glBindProgramARB() {
  Module['printErr']('missing function: emscripten_glBindProgramARB'); abort(-1);
  }


  function _emscripten_glCheckFramebufferStatus(x0) { return GLctx.checkFramebufferStatus(x0) }

  function _emscripten_glDeleteProgram(id) {
      if (!id) return;
      var program = GL.programs[id];
      if (!program) { // glDeleteProgram actually signals an error when deleting a nonexisting object, unlike some other GL delete functions.
        GL.recordError(0x0501 /* GL_INVALID_VALUE */);
        return;
      }
      GLctx.deleteProgram(program);
      program.name = 0;
      GL.programs[id] = null;
      GL.programInfos[id] = null;
    }

  function _emscripten_glDisable(x0) { GLctx.disable(x0) }

  function _emscripten_glVertexAttrib3fv(index, v) {
      v = HEAPF32.subarray((v)>>2,(v+12)>>2);
      GLctx.vertexAttrib3fv(index, v);
    }

  function _sysconf(name) {
      // long sysconf(int name);
      // http://pubs.opengroup.org/onlinepubs/009695399/functions/sysconf.html
      switch(name) {
        case 30: return PAGE_SIZE;
        case 85: return totalMemory / PAGE_SIZE;
        case 132:
        case 133:
        case 12:
        case 137:
        case 138:
        case 15:
        case 235:
        case 16:
        case 17:
        case 18:
        case 19:
        case 20:
        case 149:
        case 13:
        case 10:
        case 236:
        case 153:
        case 9:
        case 21:
        case 22:
        case 159:
        case 154:
        case 14:
        case 77:
        case 78:
        case 139:
        case 80:
        case 81:
        case 82:
        case 68:
        case 67:
        case 164:
        case 11:
        case 29:
        case 47:
        case 48:
        case 95:
        case 52:
        case 51:
        case 46:
          return 200809;
        case 79:
          return 0;
        case 27:
        case 246:
        case 127:
        case 128:
        case 23:
        case 24:
        case 160:
        case 161:
        case 181:
        case 182:
        case 242:
        case 183:
        case 184:
        case 243:
        case 244:
        case 245:
        case 165:
        case 178:
        case 179:
        case 49:
        case 50:
        case 168:
        case 169:
        case 175:
        case 170:
        case 171:
        case 172:
        case 97:
        case 76:
        case 32:
        case 173:
        case 35:
          return -1;
        case 176:
        case 177:
        case 7:
        case 155:
        case 8:
        case 157:
        case 125:
        case 126:
        case 92:
        case 93:
        case 129:
        case 130:
        case 131:
        case 94:
        case 91:
          return 1;
        case 74:
        case 60:
        case 69:
        case 70:
        case 4:
          return 1024;
        case 31:
        case 42:
        case 72:
          return 32;
        case 87:
        case 26:
        case 33:
          return 2147483647;
        case 34:
        case 1:
          return 47839;
        case 38:
        case 36:
          return 99;
        case 43:
        case 37:
          return 2048;
        case 0: return 2097152;
        case 3: return 65536;
        case 28: return 32768;
        case 44: return 32767;
        case 75: return 16384;
        case 39: return 1000;
        case 89: return 700;
        case 71: return 256;
        case 40: return 255;
        case 2: return 100;
        case 180: return 64;
        case 25: return 20;
        case 5: return 16;
        case 6: return 6;
        case 73: return 4;
        case 84: {
          if (typeof navigator === 'object') return navigator['hardwareConcurrency'] || 1;
          return 1;
        }
      }
      ___setErrNo(ERRNO_CODES.EINVAL);
      return -1;
    }

  function _emscripten_glIsFramebuffer(framebuffer) {
      var fb = GL.framebuffers[framebuffer];
      if (!fb) return 0;
      return GLctx.isFramebuffer(fb);
    }

  function _emscripten_glLineWidth(x0) { GLctx.lineWidth(x0) }

  function _emscripten_glGetString(name_) {
      if (GL.stringCache[name_]) return GL.stringCache[name_];
      var ret; 
      switch(name_) {
        case 0x1F00 /* GL_VENDOR */:
        case 0x1F01 /* GL_RENDERER */:
        case 0x1F02 /* GL_VERSION */:
          ret = allocate(intArrayFromString(GLctx.getParameter(name_)), 'i8', ALLOC_NORMAL);
          break;
        case 0x1F03 /* GL_EXTENSIONS */:
          var exts = GLctx.getSupportedExtensions();
          var gl_exts = [];
          for (var i in exts) {
            gl_exts.push(exts[i]);
            gl_exts.push("GL_" + exts[i]);
          }
          ret = allocate(intArrayFromString(gl_exts.join(' ')), 'i8', ALLOC_NORMAL);
          break;
        case 0x8B8C /* GL_SHADING_LANGUAGE_VERSION */:
          var glslVersion = GLctx.getParameter(GLctx.SHADING_LANGUAGE_VERSION);
          // Map WebGL GL_SHADING_LANGUAGE_VERSION string format to GLES format.
          if (glslVersion.indexOf('WebGL GLSL ES 1.0') != -1) glslVersion = 'OpenGL ES GLSL ES 1.00 (WebGL)';
          ret = allocate(intArrayFromString(glslVersion), 'i8', ALLOC_NORMAL);
          break;
        default:
          GL.recordError(0x0500/*GL_INVALID_ENUM*/);
          return 0;
      }
      GL.stringCache[name_] = ret;
      return ret;
    }

  function ___syscall195(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // SYS_stat64
      var path = SYSCALLS.getStr(), buf = SYSCALLS.get();
      return SYSCALLS.doStat(FS.stat, path, buf);
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  
  
  function emscriptenWebGLComputeImageSize(width, height, sizePerPixel, alignment) {
      function roundedToNextMultipleOf(x, y) {
        return Math.floor((x + y - 1) / y) * y
      }
      var plainRowSize = width * sizePerPixel;
      var alignedRowSize = roundedToNextMultipleOf(plainRowSize, alignment);
      return (height <= 0) ? 0 :
               ((height - 1) * alignedRowSize + plainRowSize);
    }function emscriptenWebGLGetTexPixelData(type, format, width, height, pixels, internalFormat) {
      var sizePerPixel;
      var numChannels;
      switch(format) {
        case 0x1906 /* GL_ALPHA */:
        case 0x1909 /* GL_LUMINANCE */:
        case 0x1902 /* GL_DEPTH_COMPONENT */:
        case 0x1903 /* GL_RED */:
          numChannels = 1;
          break;
        case 0x190A /* GL_LUMINANCE_ALPHA */:
        case 0x8227 /* GL_RG */:
          numChannels = 2;
          break;
        case 0x1907 /* GL_RGB */:
        case 0x8C40 /* GL_SRGB_EXT */:
          numChannels = 3;
          break;
        case 0x1908 /* GL_RGBA */:
        case 0x8C42 /* GL_SRGB_ALPHA_EXT */:
          numChannels = 4;
          break;
        default:
          GL.recordError(0x0500); // GL_INVALID_ENUM
          return {
            pixels: null,
            internalFormat: 0x0
          };
      }
      switch (type) {
        case 0x1401 /* GL_UNSIGNED_BYTE */:
          sizePerPixel = numChannels*1;
          break;
        case 0x1403 /* GL_UNSIGNED_SHORT */:
        case 0x8D61 /* GL_HALF_FLOAT_OES */:
          sizePerPixel = numChannels*2;
          break;
        case 0x1405 /* GL_UNSIGNED_INT */:
        case 0x1406 /* GL_FLOAT */:
          sizePerPixel = numChannels*4;
          break;
        case 0x84FA /* UNSIGNED_INT_24_8_WEBGL/UNSIGNED_INT_24_8 */:
          sizePerPixel = 4;
          break;
        case 0x8363 /* GL_UNSIGNED_SHORT_5_6_5 */:
        case 0x8033 /* GL_UNSIGNED_SHORT_4_4_4_4 */:
        case 0x8034 /* GL_UNSIGNED_SHORT_5_5_5_1 */:
          sizePerPixel = 2;
          break;
        default:
          GL.recordError(0x0500); // GL_INVALID_ENUM
          return {
            pixels: null,
            internalFormat: 0x0
          };
      }
      var bytes = emscriptenWebGLComputeImageSize(width, height, sizePerPixel, GL.unpackAlignment);
      if (type == 0x1401 /* GL_UNSIGNED_BYTE */) {
        pixels = HEAPU8.subarray((pixels),(pixels+bytes));
      } else if (type == 0x1406 /* GL_FLOAT */) {
        pixels = HEAPF32.subarray((pixels)>>2,(pixels+bytes)>>2);
      } else if (type == 0x1405 /* GL_UNSIGNED_INT */ || type == 0x84FA /* UNSIGNED_INT_24_8_WEBGL */) {
        pixels = HEAPU32.subarray((pixels)>>2,(pixels+bytes)>>2);
      } else {
        pixels = HEAPU16.subarray((pixels)>>1,(pixels+bytes)>>1);
      }
      return {
        pixels: pixels,
        internalFormat: internalFormat
      };
    }function _emscripten_glTexSubImage2D(target, level, xoffset, yoffset, width, height, format, type, pixels) {
      var pixelData;
      if (pixels) {
        pixelData = emscriptenWebGLGetTexPixelData(type, format, width, height, pixels, -1).pixels;
      } else {
        pixelData = null;
      }
      GLctx.texSubImage2D(target, level, xoffset, yoffset, width, height, format, type, pixelData);
    }

  function ___syscall197(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // SYS_fstat64
      var stream = SYSCALLS.getStreamFromFD(), buf = SYSCALLS.get();
      return SYSCALLS.doStat(FS.stat, stream.path, buf);
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function ___syscall196(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // SYS_lstat64
      var path = SYSCALLS.getStr(), buf = SYSCALLS.get();
      return SYSCALLS.doStat(FS.lstat, path, buf);
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  
  function ___syscall202(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // getgid32
      return 0;
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }function ___syscall199() {
  return ___syscall202.apply(null, arguments)
  }

  function _emscripten_glRotatef() {
  Module['printErr']('missing function: emscripten_glRotatef'); abort(-1);
  }

  function _emscripten_glVertexAttrib2f(x0, x1, x2) { GLctx.vertexAttrib2f(x0, x1, x2) }

  
  function emscriptenWebGLGet(name_, p, type) {
      // Guard against user passing a null pointer.
      // Note that GLES2 spec does not say anything about how passing a null pointer should be treated.
      // Testing on desktop core GL 3, the application crashes on glGetIntegerv to a null pointer, but
      // better to report an error instead of doing anything random.
      if (!p) {
        GL.recordError(0x0501 /* GL_INVALID_VALUE */);
        return;
      }
      var ret = undefined;
      switch(name_) { // Handle a few trivial GLES values
        case 0x8DFA: // GL_SHADER_COMPILER
          ret = 1;
          break;
        case 0x8DF8: // GL_SHADER_BINARY_FORMATS
          if (type !== 'Integer' && type !== 'Integer64') {
            GL.recordError(0x0500); // GL_INVALID_ENUM
          }
          return; // Do not write anything to the out pointer, since no binary formats are supported.
        case 0x8DF9: // GL_NUM_SHADER_BINARY_FORMATS
          ret = 0;
          break;
        case 0x86A2: // GL_NUM_COMPRESSED_TEXTURE_FORMATS
          // WebGL doesn't have GL_NUM_COMPRESSED_TEXTURE_FORMATS (it's obsolete since GL_COMPRESSED_TEXTURE_FORMATS returns a JS array that can be queried for length),
          // so implement it ourselves to allow C++ GLES2 code get the length.
          var formats = GLctx.getParameter(0x86A3 /*GL_COMPRESSED_TEXTURE_FORMATS*/);
          ret = formats.length;
          break;
        case 0x8B9A: // GL_IMPLEMENTATION_COLOR_READ_TYPE
          ret = 0x1401; // GL_UNSIGNED_BYTE
          break;
        case 0x8B9B: // GL_IMPLEMENTATION_COLOR_READ_FORMAT
          ret = 0x1908; // GL_RGBA
          break;
      }
  
      if (ret === undefined) {
        var result = GLctx.getParameter(name_);
        switch (typeof(result)) {
          case "number":
            ret = result;
            break;
          case "boolean":
            ret = result ? 1 : 0;
            break;
          case "string":
            GL.recordError(0x0500); // GL_INVALID_ENUM
            return;
          case "object":
            if (result === null) {
              // null is a valid result for some (e.g., which buffer is bound - perhaps nothing is bound), but otherwise
              // can mean an invalid name_, which we need to report as an error
              switch(name_) {
                case 0x8894: // ARRAY_BUFFER_BINDING
                case 0x8B8D: // CURRENT_PROGRAM
                case 0x8895: // ELEMENT_ARRAY_BUFFER_BINDING
                case 0x8CA6: // FRAMEBUFFER_BINDING
                case 0x8CA7: // RENDERBUFFER_BINDING
                case 0x8069: // TEXTURE_BINDING_2D
                case 0x8514: { // TEXTURE_BINDING_CUBE_MAP
                  ret = 0;
                  break;
                }
                default: {
                  GL.recordError(0x0500); // GL_INVALID_ENUM
                  return;
                }
              }
            } else if (result instanceof Float32Array ||
                       result instanceof Uint32Array ||
                       result instanceof Int32Array ||
                       result instanceof Array) {
              for (var i = 0; i < result.length; ++i) {
                switch (type) {
                  case 'Integer': HEAP32[(((p)+(i*4))>>2)]=result[i];   break;
                  case 'Float':   HEAPF32[(((p)+(i*4))>>2)]=result[i]; break;
                  case 'Boolean': HEAP8[(((p)+(i))>>0)]=result[i] ? 1 : 0;    break;
                  default: throw 'internal glGet error, bad type: ' + type;
                }
              }
              return;
            } else if (result instanceof WebGLBuffer ||
                       result instanceof WebGLProgram ||
                       result instanceof WebGLFramebuffer ||
                       result instanceof WebGLRenderbuffer ||
                       result instanceof WebGLTexture) {
              ret = result.name | 0;
            } else {
              GL.recordError(0x0500); // GL_INVALID_ENUM
              return;
            }
            break;
          default:
            GL.recordError(0x0500); // GL_INVALID_ENUM
            return;
        }
      }
  
      switch (type) {
        case 'Integer64': (tempI64 = [ret>>>0,(tempDouble=ret,(+(Math_abs(tempDouble))) >= 1.0 ? (tempDouble > 0.0 ? ((Math_min((+(Math_floor((tempDouble)/4294967296.0))), 4294967295.0))|0)>>>0 : (~~((+(Math_ceil((tempDouble - +(((~~(tempDouble)))>>>0))/4294967296.0)))))>>>0) : 0)],HEAP32[((p)>>2)]=tempI64[0],HEAP32[(((p)+(4))>>2)]=tempI64[1]);    break;
        case 'Integer': HEAP32[((p)>>2)]=ret;    break;
        case 'Float':   HEAPF32[((p)>>2)]=ret;  break;
        case 'Boolean': HEAP8[((p)>>0)]=ret ? 1 : 0; break;
        default: throw 'internal glGet error, bad type: ' + type;
      }
    }function _emscripten_glGetIntegerv(name_, p) {
      emscriptenWebGLGet(name_, p, 'Integer');
    }

  function _llvm_ctpop_i32(x) {
      var ret = 0;
      while (x) {
        if (x&1) ret++;
        x >>>= 1;
      }
      return ret;
    }

  function _emscripten_glGetFramebufferAttachmentParameteriv(target, attachment, pname, params) {
      var result = GLctx.getFramebufferAttachmentParameter(target, attachment, pname);
      HEAP32[((params)>>2)]=result;
    }

  function _emscripten_glClientActiveTexture() {
  Module['printErr']('missing function: emscripten_glClientActiveTexture'); abort(-1);
  }

  function _emscripten_set_focus_callback(target, userData, useCapture, callbackfunc) {
      JSEvents.registerFocusEventCallback(target, userData, useCapture, callbackfunc, 13, "focus");
      return 0;
    }

  
  function _emscripten_memcpy_big(dest, src, num) {
      HEAPU8.set(HEAPU8.subarray(src, src+num), dest);
      return dest;
    } 
  Module["_memcpy"] = _memcpy;

  function _emscripten_glGetShaderInfoLog(shader, maxLength, length, infoLog) {
      var log = GLctx.getShaderInfoLog(GL.shaders[shader]);
      if (log === null) log = '(unknown error)';
      log = log.substr(0, maxLength - 1);
      if (maxLength > 0 && infoLog) {
        writeStringToMemory(log, infoLog);
        if (length) HEAP32[((length)>>2)]=log.length;
      } else {
        if (length) HEAP32[((length)>>2)]=0;
      }
    }

  function _emscripten_set_mouseup_callback(target, userData, useCapture, callbackfunc) {
      JSEvents.registerMouseEventCallback(target, userData, useCapture, callbackfunc, 6, "mouseup");
      return 0;
    }

  function _emscripten_glStencilOpSeparate(x0, x1, x2, x3) { GLctx.stencilOpSeparate(x0, x1, x2, x3) }

  
  function _glutInitDisplayMode(mode) {
      GLUT.initDisplayMode = mode;
    }
  
  function _glutCreateWindow(name) {
      var contextAttributes = {
        antialias: ((GLUT.initDisplayMode & 0x0080 /*GLUT_MULTISAMPLE*/) != 0),
        depth: ((GLUT.initDisplayMode & 0x0010 /*GLUT_DEPTH*/) != 0),
        stencil: ((GLUT.initDisplayMode & 0x0020 /*GLUT_STENCIL*/) != 0)
      };
      Module.ctx = Browser.createContext(Module['canvas'], true, true, contextAttributes);
      return Module.ctx ? 1 /* a new GLUT window ID for the created context */ : 0 /* failure */;
    }function _eglCreateContext(display, config, hmm, contextAttribs) {
      if (display != 62000 /* Magic ID for Emscripten 'default display' */) {
        EGL.setErrorCode(0x3008 /* EGL_BAD_DISPLAY */);
        return 0;
      }
  
      // EGL 1.4 spec says default EGL_CONTEXT_CLIENT_VERSION is GLES1, but this is not supported by Emscripten.
      // So user must pass EGL_CONTEXT_CLIENT_VERSION == 2 to initialize EGL.
      var glesContextVersion = 1;
      for(;;) {
        var param = HEAP32[((contextAttribs)>>2)];
        if (param == 0x3098 /*EGL_CONTEXT_CLIENT_VERSION*/) {
          glesContextVersion = HEAP32[(((contextAttribs)+(4))>>2)];
        } else if (param == 0x3038 /*EGL_NONE*/) {
          break;
        } else {
          /* EGL1.4 specifies only EGL_CONTEXT_CLIENT_VERSION as supported attribute */
          EGL.setErrorCode(0x3004 /*EGL_BAD_ATTRIBUTE*/);
          return 0;
        }
        contextAttribs += 8;
      }
      if (glesContextVersion != 2) {
        EGL.setErrorCode(0x3005 /* EGL_BAD_CONFIG */);
        return 0; /* EGL_NO_CONTEXT */
      }
  
      _glutInitDisplayMode(0xB2 /* GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_MULTISAMPLE | GLUT_STENCIL */);
      EGL.windowID = _glutCreateWindow();
      if (EGL.windowID != 0) {
        EGL.setErrorCode(0x3000 /* EGL_SUCCESS */);
        // Note: This function only creates a context, but it shall not make it active.
        return 62004; // Magic ID for Emscripten EGLContext
      } else {
        EGL.setErrorCode(0x3009 /* EGL_BAD_MATCH */); // By the EGL 1.4 spec, an implementation that does not support GLES2 (WebGL in this case), this error code is set.
        return 0; /* EGL_NO_CONTEXT */
      }
    }

  function _emscripten_glReadPixels(x, y, width, height, format, type, pixels) {
      var data = emscriptenWebGLGetTexPixelData(type, format, width, height, pixels, format);
      if (!data.pixels) {
        GL.recordError(0x0500/*GL_INVALID_ENUM*/);
        return;
      }
      GLctx.readPixels(x, y, width, height, format, type, data.pixels);
    }

  function _emscripten_glCompressedTexSubImage2D(target, level, xoffset, yoffset, width, height, format, imageSize, data) {
      var heapView;
      if (data) {
        heapView = HEAPU8.subarray((data),(data+imageSize));
      } else {
        heapView = null;
      }
      GLctx['compressedTexSubImage2D'](target, level, xoffset, yoffset, width, height, format, heapView);
    }

  function _emscripten_glGetError() {
      // First return any GL error generated by the emscripten library_gl.js interop layer.
      if (GL.lastError) {
        var error = GL.lastError;
        GL.lastError = 0/*GL_NO_ERROR*/;
        return error;
      } else { // If there were none, return the GL error from the browser GL context.
        return GLctx.getError();
      }
    }

  function _eglBindAPI(api) {
      if (api == 0x30A0 /* EGL_OPENGL_ES_API */) {
        EGL.setErrorCode(0x3000 /* EGL_SUCCESS */);
        return 1;
      } else { // if (api == 0x30A1 /* EGL_OPENVG_API */ || api == 0x30A2 /* EGL_OPENGL_API */) {
        EGL.setErrorCode(0x300C /* EGL_BAD_PARAMETER */);
        return 0;
      }
    }

  function ___syscall39(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // mkdir
      var path = SYSCALLS.getStr(), mode = SYSCALLS.get();
      return SYSCALLS.doMkdir(path, mode);
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function ___syscall38(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // rename
      var old_path = SYSCALLS.getStr(), new_path = SYSCALLS.getStr();
      FS.rename(old_path, new_path);
      return 0;
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function ___syscall33(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // access
      var path = SYSCALLS.getStr(), amode = SYSCALLS.get();
      return SYSCALLS.doAccess(path, amode);
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function _pthread_cleanup_push(routine, arg) {
      __ATEXIT__.push(function() { Runtime.dynCall('vi', routine, [arg]) })
      _pthread_cleanup_push.level = __ATEXIT__.length;
    }

  function _emscripten_glFramebufferTexture2D(target, attachment, textarget, texture, level) {
      GLctx.framebufferTexture2D(target, attachment, textarget,
                                      GL.textures[texture], level);
    }

  function _alSourcei(source, param, value) {
      if (!AL.currentContext) {
        return;
      }
      var src = AL.currentContext.src[source];
      if (!src) {
        AL.currentContext.err = 0xA001 /* AL_INVALID_NAME */;
        return;
      }
      switch (param) {
      case 0x1001 /* AL_CONE_INNER_ANGLE */:
        src.coneInnerAngle = value;
        break;
      case 0x1002 /* AL_CONE_OUTER_ANGLE */:
        src.coneOuterAngle = value;
        break;
      case 0x1007 /* AL_LOOPING */:
        src.loop = (value === 1 /* AL_TRUE */);
        break;
      case 0x1009 /* AL_BUFFER */:
        var buffer = AL.currentContext.buf[value - 1];
        if (value == 0) {
          src.queue = [];
        } else {
          src.queue = [{ buffer: buffer }];
        }
        AL.updateSource(src);
        break;
      case 0x202 /* AL_SOURCE_RELATIVE */:
        if (value === 0 /* AL_FALSE */) {
          if (src.panner) {
            src.panner = null;
  
            // Disconnect from the panner.
            src.gain.disconnect();
  
            src.gain.connect(AL.currentContext.gain);
          }
        } else if (value === 1 /* AL_TRUE */) {
          if (!src.panner) {
            var panner = src.panner = AL.currentContext.ctx.createPanner();
            panner.panningModel = "equalpower";
            panner.distanceModel = "linear";
            panner.refDistance = src.refDistance;
            panner.maxDistance = src.maxDistance;
            panner.rolloffFactor = src.rolloffFactor;
            panner.setPosition(src.position[0], src.position[1], src.position[2]);
            panner.setVelocity(src.velocity[0], src.velocity[1], src.velocity[2]);
            panner.connect(AL.currentContext.gain);
  
            // Disconnect from the default source.
            src.gain.disconnect();
  
            src.gain.connect(panner);
          }
        } else {
          AL.currentContext.err = 0xA003 /* AL_INVALID_VALUE */;
        }
        break;
      default:
        AL.currentContext.err = 0xA002 /* AL_INVALID_ENUM */;
        break;
      }
    }

  function _emscripten_glIsEnabled(x0) { return GLctx.isEnabled(x0) }

  function _alSourceQueueBuffers(source, count, buffers) {
      if (!AL.currentContext) {
        return;
      }
      var src = AL.currentContext.src[source];
      if (!src) {
        AL.currentContext.err = 0xA001 /* AL_INVALID_NAME */;
        return;
      }
      for (var i = 0; i < count; ++i) {
        var bufferIdx = HEAP32[(((buffers)+(i*4))>>2)];
        if (bufferIdx > AL.currentContext.buf.length) {
          AL.currentContext.err = 0xA001 /* AL_INVALID_NAME */;
          return;
        }
      }
  
      for (var i = 0; i < count; ++i) {
        var bufferIdx = HEAP32[(((buffers)+(i*4))>>2)];
        var buffer = AL.currentContext.buf[bufferIdx - 1];
        src.queue.push({ buffer: buffer, src: null });
      }
  
      AL.updateSource(src);
    }

  function _alSourcef(source, param, value) {
      if (!AL.currentContext) {
        return;
      }
      var src = AL.currentContext.src[source];
      if (!src) {
        AL.currentContext.err = 0xA001 /* AL_INVALID_NAME */;
        return;
      }
      switch (param) {
      case 0x1003 /* AL_PITCH */:
        break;
      case 0x100A /* AL_GAIN */:
        src.gain.gain.value = value;
        break;
      // case 0x100D /* AL_MIN_GAIN */:
      //   break;
      // case 0x100E /* AL_MAX_GAIN */:
      //   break;
      case 0x1023 /* AL_MAX_DISTANCE */:
        src.maxDistance = value;
        break;
      case 0x1021 /* AL_ROLLOFF_FACTOR */:
        src.rolloffFactor = value;
        break;
      case 0x1022 /* AL_CONE_OUTER_GAIN */:
        src.coneOuterGain = value;
        break;
      case 0x1001 /* AL_CONE_INNER_ANGLE */:
        src.coneInnerAngle = value;
        break;
      case 0x1002 /* AL_CONE_OUTER_ANGLE */:
        src.coneOuterAngle = value;
        break;
      case 0x1020 /* AL_REFERENCE_DISTANCE */:
        src.refDistance = value;
        break;
      default:
        AL.currentContext.err = 0xA002 /* AL_INVALID_ENUM */;
        break;
      }
    }

   
  Module["_memmove"] = _memmove;

  function _emscripten_glClearDepthf(x0) { GLctx.clearDepth(x0) }

  
  var __sigalrm_handler=0;function _signal(sig, func) {
      if (sig == 14 /*SIGALRM*/) {
        __sigalrm_handler = func;
      } else {
        Module.printErr('Calling stub instead of signal()');
      }
      return 0;
    }

  function _emscripten_glVertexAttrib4f(x0, x1, x2, x3, x4) { GLctx.vertexAttrib4f(x0, x1, x2, x3, x4) }

  function ___gxx_personality_v0() {
    }

  var _emscripten_asm_const=true;

  function ___cxa_rethrow() {
      ___cxa_end_catch.rethrown = true;
      var ptr = EXCEPTIONS.caught.pop();
      EXCEPTIONS.last = ptr;
      throw ptr;
    }

  function _pthread_cond_wait() {
      return 0;
    }

  function _alListenerf(param, value) {
      if (!AL.currentContext) {
        return;
      }
      switch (param) {
      case 0x100A /* AL_GAIN */:
        AL.currentContext.gain.gain.value = value;
        break;
      default:
        AL.currentContext.err = 0xA002 /* AL_INVALID_ENUM */;
        break;
      }
    }

  function _emscripten_glClear(x0) { GLctx.clear(x0) }

   
  Module["_llvm_ctlz_i64"] = _llvm_ctlz_i64;

  function _alGetError() {
      if (!AL.currentContext) {
        return 0xA004 /* AL_INVALID_OPERATION */;
      } else {
        // Reset error on get.
        var err = AL.currentContext.err;
        AL.currentContext.err = 0 /* AL_NO_ERROR */;
        return err;
      }
    }

  
  function _emscripten_get_now() {
      if (!_emscripten_get_now.actual) {
        if (ENVIRONMENT_IS_NODE) {
          _emscripten_get_now.actual = function _emscripten_get_now_actual() {
            var t = process['hrtime']();
            return t[0] * 1e3 + t[1] / 1e6;
          }
        } else if (typeof dateNow !== 'undefined') {
          _emscripten_get_now.actual = dateNow;
        } else if (typeof self === 'object' && self['performance'] && typeof self['performance']['now'] === 'function') {
          _emscripten_get_now.actual = function _emscripten_get_now_actual() { return self['performance']['now'](); };
        } else if (typeof performance === 'object' && typeof performance['now'] === 'function') {
          _emscripten_get_now.actual = function _emscripten_get_now_actual() { return performance['now'](); };
        } else {
          _emscripten_get_now.actual = Date.now;
        }
      }
      return _emscripten_get_now.actual();
    }
  
  function _emscripten_get_now_is_monotonic() {
      // return whether emscripten_get_now is guaranteed monotonic; the Date.now
      // implementation is not :(
      return ENVIRONMENT_IS_NODE || (typeof dateNow !== 'undefined') ||
          ((ENVIRONMENT_IS_WEB || ENVIRONMENT_IS_WORKER) && self['performance'] && self['performance']['now']);
    }function _clock_gettime(clk_id, tp) {
      // int clock_gettime(clockid_t clk_id, struct timespec *tp);
      var now;
      if (clk_id === 0) {
        now = Date.now();
      } else if (clk_id === 1 && _emscripten_get_now_is_monotonic()) {
        now = _emscripten_get_now();
      } else {
        ___setErrNo(ERRNO_CODES.EINVAL);
        return -1;
      }
      HEAP32[((tp)>>2)]=(now/1000)|0; // seconds
      HEAP32[(((tp)+(4))>>2)]=((now % 1000)*1000*1000)|0; // nanoseconds
      return 0;
    }

  function _alGetListenerfv(pname, values) {
      if (!AL.currentContext) {
        return;
      }
      switch (pname) {
      case 0x1004 /* AL_POSITION */:
        var position = AL.currentContext.ctx.listener._position || [0,0,0];
        HEAPF32[((values)>>2)]=position[0]
        HEAPF32[(((values)+(4))>>2)]=position[1]
        HEAPF32[(((values)+(8))>>2)]=position[2]
        break;
      case 0x1006 /* AL_VELOCITY */:
        var velocity = AL.currentContext.ctx.listener._velocity || [0,0,0];
        HEAPF32[((values)>>2)]=velocity[0]
        HEAPF32[(((values)+(4))>>2)]=velocity[1]
        HEAPF32[(((values)+(8))>>2)]=velocity[2]
        break;
      case 0x100F /* AL_ORIENTATION */:
        var orientation = AL.currentContext.ctx.listener._orientation || [0,0,-1,0,1,0];
        HEAPF32[((values)>>2)]=orientation[0]
        HEAPF32[(((values)+(4))>>2)]=orientation[1]
        HEAPF32[(((values)+(8))>>2)]=orientation[2]
        HEAPF32[(((values)+(12))>>2)]=orientation[3]
        HEAPF32[(((values)+(16))>>2)]=orientation[4]
        HEAPF32[(((values)+(20))>>2)]=orientation[5]
        break;
      default:
        AL.currentContext.err = 0xA002 /* AL_INVALID_ENUM */;
        break;
      }
    }

  function _emscripten_glBindBuffer(target, buffer) {
      var bufferObj = buffer ? GL.buffers[buffer] : null;
  
      if (target == GLctx.ARRAY_BUFFER) {
        GL.currArrayBuffer = buffer;
      } else if (target == GLctx.ELEMENT_ARRAY_BUFFER) {
        GL.currElementArrayBuffer = buffer;
      }
  
      GLctx.bindBuffer(target, bufferObj);
    }

  function _emscripten_glDisableVertexAttribArray(index) {
      var cb = GL.currentContext.clientBuffers[index];
      assert(cb, index);
      cb.enabled = false;
      GLctx.disableVertexAttribArray(index);
    }

  function _alcOpenDevice(deviceName) {
      if (typeof(AudioContext) !== "undefined" ||
          typeof(webkitAudioContext) !== "undefined") {
        return 1; // non-null pointer -- we just simulate one device
      } else {
        return 0;
      }
    }

   
  Module["_bitshift64Shl"] = _bitshift64Shl;

  function _emscripten_glGetBufferParameteriv(target, value, data) {
      if (!data) {
        // GLES2 specification does not specify how to behave if data is a null pointer. Since calling this function does not make sense
        // if data == null, issue a GL error to notify user about it. 
        GL.recordError(0x0501 /* GL_INVALID_VALUE */);
        return;
      }
      HEAP32[((data)>>2)]=GLctx.getBufferParameter(target, value);
    }

  function _emscripten_glGetAttachedShaders(program, maxCount, count, shaders) {
      var result = GLctx.getAttachedShaders(GL.programs[program]);
      var len = result.length;
      if (len > maxCount) {
        len = maxCount;
      }
      HEAP32[((count)>>2)]=len;
      for (var i = 0; i < len; ++i) {
        var id = GL.shaders.indexOf(result[i]);
        assert(id !== -1, 'shader not bound to local id');
        HEAP32[(((shaders)+(i*4))>>2)]=id;
      }
    }

  function _alSourceRewind(source) {
      if (!AL.currentContext) {
        return;
      }
      var src = AL.currentContext.src[source];
      if (!src) {
        AL.currentContext.err = 0xA001 /* AL_INVALID_NAME */;
        return;
      }
      // Stop the source first to clear the source queue
      AL.setSourceState(src, 0x1014 /* AL_STOPPED */);
      // Now set the state of AL_INITIAL according to the specification
      AL.setSourceState(src, 0x1011 /* AL_INITIAL */);
    }

  function _emscripten_glGenRenderbuffers(n, renderbuffers) {
      for (var i = 0; i < n; i++) {
        var renderbuffer = GLctx.createRenderbuffer();
        if (!renderbuffer) {
          GL.recordError(0x0502 /* GL_INVALID_OPERATION */);
          while(i < n) HEAP32[(((renderbuffers)+(i++*4))>>2)]=0;
          return;
        }
        var id = GL.getNewId(GL.renderbuffers);
        renderbuffer.name = id;
        GL.renderbuffers[id] = renderbuffer;
        HEAP32[(((renderbuffers)+(i*4))>>2)]=id;
      }
    }

  function _alGetSourcefv(source, param, values) {
      if (!AL.currentContext) {
        return;
      }
      var src = AL.currentContext.src[source];
      if (!src) {
        AL.currentContext.err = 0xA001 /* AL_INVALID_NAME */;
        return;
      }
      switch (param) {
      case 0x1003 /* AL_PITCH */:
      case 0x100A /* AL_GAIN */:
      case 0x100D /* AL_MIN_GAIN */:
      case 0x100E /* AL_MAX_GAIN */:
      case 0x1023 /* AL_MAX_DISTANCE */:
      case 0x1021 /* AL_ROLLOFF_FACTOR */:
      case 0x1022 /* AL_CONE_OUTER_GAIN */:
      case 0x1001 /* AL_CONE_INNER_ANGLE */:
      case 0x1002 /* AL_CONE_OUTER_ANGLE */:
      case 0x1020 /* AL_REFERENCE_DISTANCE */:
      case 0x1024 /* AL_SEC_OFFSET */:
      case 0x1025 /* AL_SAMPLE_OFFSET */:
      case 0x1026 /* AL_BYTE_OFFSET */:
        _alGetSourcef(source, param, values);
        break;
      case 0x1004 /* AL_POSITION */:
        var position = src.position;
        HEAPF32[((values)>>2)]=position[0]
        HEAPF32[(((values)+(4))>>2)]=position[1]
        HEAPF32[(((values)+(8))>>2)]=position[2]
        break;
      case 0x1005 /* AL_DIRECTION */:
        var direction = src.direction;
        HEAPF32[((values)>>2)]=direction[0]
        HEAPF32[(((values)+(4))>>2)]=direction[1]
        HEAPF32[(((values)+(8))>>2)]=direction[2]
        break;
      case 0x1006 /* AL_VELOCITY */:
        var velocity = src.velocity;
        HEAPF32[((values)>>2)]=velocity[0]
        HEAPF32[(((values)+(4))>>2)]=velocity[1]
        HEAPF32[(((values)+(8))>>2)]=velocity[2]
        break;
      default:
        AL.currentContext.err = 0xA002 /* AL_INVALID_ENUM */;
        break;
      }
    }

  function _emscripten_glFrontFace(x0) { GLctx.frontFace(x0) }

  function _emscripten_glActiveTexture(x0) { GLctx.activeTexture(x0) }

  function _emscripten_glUniform1iv(location, count, value) {
      location = GL.uniforms[location];
      value = HEAP32.subarray((value)>>2,(value+count*4)>>2);
      GLctx.uniform1iv(location, value);
    }

  function _emscripten_glGetAttribLocation(program, name) {
      program = GL.programs[program];
      name = Pointer_stringify(name);
      return GLctx.getAttribLocation(program, name);
    }

  function _emscripten_glTexCoordPointer() {
  Module['printErr']('missing function: emscripten_glTexCoordPointer'); abort(-1);
  }

  function _emscripten_glGetInfoLogARB() {
  Module['printErr']('missing function: emscripten_glGetInfoLogARB'); abort(-1);
  }

  
  function __exit(status) {
      // void _exit(int status);
      // http://pubs.opengroup.org/onlinepubs/000095399/functions/exit.html
      Module['exit'](status);
    }function _exit(status) {
      __exit(status);
    }

  function _pthread_setspecific(key, value) {
      if (!(key in PTHREAD_SPECIFIC)) {
        return ERRNO_CODES.EINVAL;
      }
      PTHREAD_SPECIFIC[key] = value;
      return 0;
    }

  
  function _gethostbyname(name) {
      name = Pointer_stringify(name);
  
      // generate hostent
      var ret = _malloc(20); // XXX possibly leaked, as are others here
      var nameBuf = _malloc(name.length+1);
      writeStringToMemory(name, nameBuf);
      HEAP32[((ret)>>2)]=nameBuf;
      var aliasesBuf = _malloc(4);
      HEAP32[((aliasesBuf)>>2)]=0;
      HEAP32[(((ret)+(4))>>2)]=aliasesBuf;
      var afinet = 2;
      HEAP32[(((ret)+(8))>>2)]=afinet;
      HEAP32[(((ret)+(12))>>2)]=4;
      var addrListBuf = _malloc(12);
      HEAP32[((addrListBuf)>>2)]=addrListBuf+8;
      HEAP32[(((addrListBuf)+(4))>>2)]=0;
      HEAP32[(((addrListBuf)+(8))>>2)]=__inet_pton4_raw(DNS.lookup_name(name));
      HEAP32[(((ret)+(16))>>2)]=addrListBuf;
      return ret;
    }function _gethostbyaddr(addr, addrlen, type) {
      if (type !== 2) {
        ___setErrNo(ERRNO_CODES.EAFNOSUPPORT);
        // TODO: set h_errno
        return null;
      }
      addr = HEAP32[((addr)>>2)]; // addr is in_addr
      var host = __inet_ntop4_raw(addr);
      var lookup = DNS.lookup_addr(host);
      if (lookup) {
        host = lookup;
      }
      var hostp = allocate(intArrayFromString(host), 'i8', ALLOC_STACK);
      return _gethostbyname(hostp);
    }

  function _emscripten_set_keyup_callback(target, userData, useCapture, callbackfunc) {
      JSEvents.registerKeyEventCallback(target, userData, useCapture, callbackfunc, 3, "keyup");
      return 0;
    }

  function _emscripten_glRenderbufferStorage(x0, x1, x2, x3) { GLctx.renderbufferStorage(x0, x1, x2, x3) }

  function _emscripten_set_mousemove_callback(target, userData, useCapture, callbackfunc) {
      JSEvents.registerMouseEventCallback(target, userData, useCapture, callbackfunc, 8, "mousemove");
      return 0;
    }

  function ___syscall10(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // unlink
      var path = SYSCALLS.getStr();
      FS.unlink(path);
      return 0;
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function _emscripten_glTexParameteriv(target, pname, params) {
      var param = HEAP32[((params)>>2)];
      GLctx.texParameteri(target, pname, param);
    }

  function _emscripten_glShaderBinary() {
      GL.recordError(0x0500/*GL_INVALID_ENUM*/);
    }

  function _emscripten_glIsProgram(program) {
      var program = GL.programs[program];
      if (!program) return 0;
      return GLctx.isProgram(program);
    }

  function ___cxa_begin_catch(ptr) {
      __ZSt18uncaught_exceptionv.uncaught_exception--;
      EXCEPTIONS.caught.push(ptr);
      EXCEPTIONS.addRef(EXCEPTIONS.deAdjust(ptr));
      return ptr;
    }

  function _eglInitialize(display, majorVersion, minorVersion) {
      if (display == 62000 /* Magic ID for Emscripten 'default display' */) {
        if (majorVersion) {
          HEAP32[((majorVersion)>>2)]=1; // Advertise EGL Major version: '1'
        }
        if (minorVersion) {
          HEAP32[((minorVersion)>>2)]=4; // Advertise EGL Minor version: '4'
        }
        EGL.defaultDisplayInitialized = true;
        EGL.setErrorCode(0x3000 /* EGL_SUCCESS */);
        return 1;
      } 
      else {
        EGL.setErrorCode(0x3008 /* EGL_BAD_DISPLAY */);
        return 0;
      }
    }

  function _emscripten_glBlendColor(x0, x1, x2, x3) { GLctx.blendColor(x0, x1, x2, x3) }

  function _emscripten_glGetShaderiv(shader, pname, p) {
      if (!p) {
        // GLES2 specification does not specify how to behave if p is a null pointer. Since calling this function does not make sense
        // if p == null, issue a GL error to notify user about it. 
        GL.recordError(0x0501 /* GL_INVALID_VALUE */);
        return;
      }
      if (pname == 0x8B84) { // GL_INFO_LOG_LENGTH
        var log = GLctx.getShaderInfoLog(GL.shaders[shader]);
        if (log === null) log = '(unknown error)';
        HEAP32[((p)>>2)]=log.length + 1;
      } else {
        HEAP32[((p)>>2)]=GLctx.getShaderParameter(GL.shaders[shader], pname);
      }
    }

  function _emscripten_glUniformMatrix3fv(location, count, transpose, value) {
      location = GL.uniforms[location];
      var view;
      if (count === 1) {
        // avoid allocation for the common case of uploading one uniform matrix
        view = GL.miniTempBufferViews[8];
        for (var i = 0; i < 9; i++) {
          view[i] = HEAPF32[(((value)+(i*4))>>2)];
        }
      } else {
        view = HEAPF32.subarray((value)>>2,(value+count*36)>>2);
      }
      GLctx.uniformMatrix3fv(location, transpose, view);
    }


  function _emscripten_glUniform4fv(location, count, value) {
      location = GL.uniforms[location];
      var view;
      if (count === 1) {
        // avoid allocation for the common case of uploading one uniform
        view = GL.miniTempBufferViews[3];
        view[0] = HEAPF32[((value)>>2)];
        view[1] = HEAPF32[(((value)+(4))>>2)];
        view[2] = HEAPF32[(((value)+(8))>>2)];
        view[3] = HEAPF32[(((value)+(12))>>2)];
      } else {
        view = HEAPF32.subarray((value)>>2,(value+count*16)>>2);
      }
      GLctx.uniform4fv(location, view);
    }

  function _emscripten_set_keypress_callback(target, userData, useCapture, callbackfunc) {
      JSEvents.registerKeyEventCallback(target, userData, useCapture, callbackfunc, 1, "keypress");
      return 0;
    }


  function _emscripten_glGenFramebuffers(n, ids) {
      for (var i = 0; i < n; ++i) {
        var framebuffer = GLctx.createFramebuffer();
        if (!framebuffer) {
          GL.recordError(0x0502 /* GL_INVALID_OPERATION */);
          while(i < n) HEAP32[(((ids)+(i++*4))>>2)]=0;
          return;
        }
        var id = GL.getNewId(GL.framebuffers);
        framebuffer.name = id;
        GL.framebuffers[id] = framebuffer;
        HEAP32[(((ids)+(i*4))>>2)]=id;
      }
    }

  function ___syscall40(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // rmdir
      var path = SYSCALLS.getStr();
      FS.rmdir(path);
      return 0;
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function _emscripten_exit_pointerlock() {
      // Make sure no queued up calls will fire after this.
      JSEvents.removeDeferredCalls(JSEvents.requestPointerLock);
  
      if (document.exitPointerLock) {
        document.exitPointerLock();
      } else if (document.msExitPointerLock) {
        document.msExitPointerLock();
      } else if (document.mozExitPointerLock) {
        document.mozExitPointerLock();
      } else if (document.webkitExitPointerLock) {
        document.webkitExitPointerLock();
      } else {
        return -1;
      }
      return 0;
    }

  function _emscripten_glBlendEquationSeparate(x0, x1) { GLctx.blendEquationSeparate(x0, x1) }

  function _eglWaitNative(nativeEngineId) {
      EGL.setErrorCode(0x3000 /* EGL_SUCCESS */);
      return 1;
    }

  
  function _usleep(useconds) {
      // int usleep(useconds_t useconds);
      // http://pubs.opengroup.org/onlinepubs/000095399/functions/usleep.html
      // We're single-threaded, so use a busy loop. Super-ugly.
      var msec = useconds / 1000;
      if ((ENVIRONMENT_IS_WEB || ENVIRONMENT_IS_WORKER) && self['performance'] && self['performance']['now']) {
        var start = self['performance']['now']();
        while (self['performance']['now']() - start < msec) {
          // Do nothing.
        }
      } else {
        var start = Date.now();
        while (Date.now() - start < msec) {
          // Do nothing.
        }
      }
      return 0;
    }
  Module["_usleep"] = _usleep;function _nanosleep(rqtp, rmtp) {
      // int nanosleep(const struct timespec  *rqtp, struct timespec *rmtp);
      var seconds = HEAP32[((rqtp)>>2)];
      var nanoseconds = HEAP32[(((rqtp)+(4))>>2)];
      if (rmtp !== 0) {
        HEAP32[((rmtp)>>2)]=0;
        HEAP32[(((rmtp)+(4))>>2)]=0;
      }
      return _usleep((seconds * 1e6) + (nanoseconds / 1000));
    }

  function _emscripten_glBindTexture(target, texture) {
      GLctx.bindTexture(target, texture ? GL.textures[texture] : null);
    }

  function _emscripten_glDrawRangeElements() {
  Module['printErr']('missing function: emscripten_glDrawRangeElements'); abort(-1);
  }

  function _emscripten_glGenTextures(n, textures) {
      for (var i = 0; i < n; i++) {
        var texture = GLctx.createTexture();
        if (!texture) {
          GL.recordError(0x0502 /* GL_INVALID_OPERATION */); // GLES + EGL specs don't specify what should happen here, so best to issue an error and create IDs with 0.
          while(i < n) HEAP32[(((textures)+(i++*4))>>2)]=0;
          return;
        }
        var id = GL.getNewId(GL.textures);
        texture.name = id;
        GL.textures[id] = texture;
        HEAP32[(((textures)+(i*4))>>2)]=id;
      }
    }

  function _emscripten_glVertexAttrib2fv(index, v) {
      v = HEAPF32.subarray((v)>>2,(v+8)>>2);
      GLctx.vertexAttrib2fv(index, v);
    }

  function _emscripten_glGetActiveUniform(program, index, bufSize, length, size, type, name) {
      program = GL.programs[program];
      var info = GLctx.getActiveUniform(program, index);
      if (!info) return; // If an error occurs, nothing will be written to length, size, type and name.
  
      var infoname = info.name.slice(0, Math.max(0, bufSize - 1));
      if (bufSize > 0 && name) {
        writeStringToMemory(infoname, name);
        if (length) HEAP32[((length)>>2)]=infoname.length;
      } else {
        if (length) HEAP32[((length)>>2)]=0;
      }
  
      if (size) HEAP32[((size)>>2)]=info.size;
      if (type) HEAP32[((type)>>2)]=info.type;
    }

  function _emscripten_glDeleteObjectARB() {
  Module['printErr']('missing function: emscripten_glDeleteObjectARB'); abort(-1);
  }

  function _emscripten_set_touchmove_callback(target, userData, useCapture, callbackfunc) {
      JSEvents.registerTouchEventCallback(target, userData, useCapture, callbackfunc, 24, "touchmove");
      return 0;
    }

  function _emscripten_glUniform1f(location, v0) {
      location = GL.uniforms[location];
      GLctx.uniform1f(location, v0);
    }

  function _alcCreateContext(device, attrList) {
      if (device != 1) {
        return 0;
      }
  
      if (attrList) {
        return 0;
      }
  
      var ctx;
      try {
        ctx = new AudioContext();
      } catch (e) {
        try {
          ctx = new webkitAudioContext();
        } catch (e) {}
      }
  
      if (ctx) {
        // Old Web Audio API (e.g. Safari 6.0.5) had an inconsistently named createGainNode function.
        if (typeof(ctx.createGain) === 'undefined') ctx.createGain = ctx.createGainNode;
  
        var gain = ctx.createGain();
        gain.connect(ctx.destination);
        var context = {
          ctx: ctx,
          err: 0,
          src: {},
          buf: [],
          interval: setInterval(function() { AL.updateSources(context); }, AL.QUEUE_INTERVAL),
          gain: gain
        };
        AL.contexts.push(context);
        return AL.contexts.length;
      } else {
        return 0;
      }
    }

  function _emscripten_glVertexAttribPointer(index, size, type, normalized, stride, ptr) {
      var cb = GL.currentContext.clientBuffers[index];
      assert(cb, index);
      if (!GL.currArrayBuffer) {
        cb.size = size;
        cb.type = type;
        cb.normalized = normalized;
        cb.stride = stride;
        cb.ptr = ptr;
        cb.clientside = true;
        return;
      }
      cb.clientside = false;
      GLctx.vertexAttribPointer(index, size, type, normalized, stride, ptr);
    }

  function _alcCloseDevice(device) {
      // Stop playback, etc
    }

  function _emscripten_glDrawArrays(mode, first, count) {
      // bind any client-side buffers
      GL.preDrawHandleClientVertexAttribBindings(first + count);
  
      GLctx.drawArrays(mode, first, count);
  
      GL.postDrawHandleClientVertexAttribBindings();
    }

  function _emscripten_glGenBuffers(n, buffers) {
      for (var i = 0; i < n; i++) {
        var buffer = GLctx.createBuffer();
        if (!buffer) {
          GL.recordError(0x0502 /* GL_INVALID_OPERATION */);
          while(i < n) HEAP32[(((buffers)+(i++*4))>>2)]=0;
          return;
        }
        var id = GL.getNewId(GL.buffers);
        buffer.name = id;
        GL.buffers[id] = buffer;
        HEAP32[(((buffers)+(i*4))>>2)]=id;
      }
    }

  function _emscripten_glClearDepth(x0) { GLctx.clearDepth(x0) }

   
  Module["_i64Subtract"] = _i64Subtract;

  
  var PTHREAD_SPECIFIC_NEXT_KEY=1;function _pthread_key_create(key, destructor) {
      if (key == 0) {
        return ERRNO_CODES.EINVAL;
      }
      HEAP32[((key)>>2)]=PTHREAD_SPECIFIC_NEXT_KEY;
      // values start at 0
      PTHREAD_SPECIFIC[PTHREAD_SPECIFIC_NEXT_KEY] = 0;
      PTHREAD_SPECIFIC_NEXT_KEY++;
      return 0;
    }

  function _system(command) {
      // int system(const char *command);
      // http://pubs.opengroup.org/onlinepubs/000095399/functions/system.html
      // Can't call external programs.
      ___setErrNo(ERRNO_CODES.EAGAIN);
      return -1;
    }

  function _emscripten_glGetUniformLocation(program, name) {
      name = Pointer_stringify(name);
  
      var arrayOffset = 0;
      // If user passed an array accessor "[index]", parse the array index off the accessor.
      if (name.indexOf(']', name.length-1) !== -1) {
        var ls = name.lastIndexOf('[');
        var arrayIndex = name.slice(ls+1, -1);
        if (arrayIndex.length > 0) {
          arrayOffset = parseInt(arrayIndex);
          if (arrayOffset < 0) {
            return -1;
          }
        }
        name = name.slice(0, ls);
      }
  
      var ptable = GL.programInfos[program];
      if (!ptable) {
        return -1;
      }
      var utable = ptable.uniforms;
      var uniformInfo = utable[name]; // returns pair [ dimension_of_uniform_array, uniform_location ]
      if (uniformInfo && arrayOffset < uniformInfo[0]) { // Check if user asked for an out-of-bounds element, i.e. for 'vec4 colors[3];' user could ask for 'colors[10]' which should return -1.
        return uniformInfo[1]+arrayOffset;
      } else {
        return -1;
      }
    }

  function _emscripten_glEnable(x0) { GLctx.enable(x0) }

  function _emscripten_glScissor(x0, x1, x2, x3) { GLctx.scissor(x0, x1, x2, x3) }

   
  Module["_bitshift64Lshr"] = _bitshift64Lshr;

  function _emscripten_glLinkProgram(program) {
      GLctx.linkProgram(GL.programs[program]);
      GL.programInfos[program] = null; // uniforms no longer keep the same names after linking
      GL.populateUniformTable(program);
    }

  function _emscripten_glValidateProgram(program) {
      GLctx.validateProgram(GL.programs[program]);
    }

  function _emscripten_longjmp(env, value) {
      _longjmp(env, value);
    }

  function _emscripten_glGetUniformfv(program, location, params) {
      emscriptenWebGLGetUniform(program, location, params, 'Float');
    }

  function _emscripten_glDrawBuffers(n, bufs) {
      var bufArray = [];
      for (var i = 0; i < n; i++)
        bufArray.push(HEAP32[(((bufs)+(i*4))>>2)]);
  
      // Work around Firefox WebGL 2 bug, see https://github.com/kripken/emscripten/issues/3890.
      if (n == 0) {
        // If a FBO is bound, glDrawBuffers(0, *) means glDrawBuffers(1, GL_NONE).
        // If an FBO is not bound, glDrawBuffers(0, *) is an error.
        if (GLctx.getParameter(GLctx['DRAW_FRAMEBUFFER_BINDING']) != 0) bufArray.push(GLctx['NONE']);
        else {
          GL.recordError(0x0502 /* GL_INVALID_OPERATION */);
          return;
        }
      }
  
      GLctx['drawBuffers'](bufArray);
    }

  function ___cxa_find_matching_catch_2() {
          return ___cxa_find_matching_catch.apply(null, arguments);
        }

  function ___cxa_find_matching_catch_3() {
          return ___cxa_find_matching_catch.apply(null, arguments);
        }

  function _pthread_mutex_unlock() {}

  function _emscripten_glBindFramebuffer(target, framebuffer) {
      GLctx.bindFramebuffer(target, framebuffer ? GL.framebuffers[framebuffer] : null);
    }

  function _emscripten_glBlendEquation(x0) { GLctx.blendEquation(x0) }

  function _emscripten_glBufferSubData(target, offset, size, data) {
      GLctx.bufferSubData(target, offset, HEAPU8.subarray(data, data+size));
    }

  function _alcCaptureStart() {
  Module['printErr']('missing function: alcCaptureStart'); abort(-1);
  }

  function _emscripten_set_keydown_callback(target, userData, useCapture, callbackfunc) {
      JSEvents.registerKeyEventCallback(target, userData, useCapture, callbackfunc, 2, "keydown");
      return 0;
    }

  function _emscripten_glBufferData(target, size, data, usage) {
      switch (usage) { // fix usages, WebGL only has *_DRAW
        case 0x88E1: // GL_STREAM_READ
        case 0x88E2: // GL_STREAM_COPY
          usage = 0x88E0; // GL_STREAM_DRAW
          break;
        case 0x88E5: // GL_STATIC_READ
        case 0x88E6: // GL_STATIC_COPY
          usage = 0x88E4; // GL_STATIC_DRAW
          break;
        case 0x88E9: // GL_DYNAMIC_READ
        case 0x88EA: // GL_DYNAMIC_COPY
          usage = 0x88E8; // GL_DYNAMIC_DRAW
          break;
      }
      if (!data) {
        GLctx.bufferData(target, size, usage);
      } else {
        GLctx.bufferData(target, HEAPU8.subarray(data, data+size), usage);
      }
    }

  function _sbrk(bytes) {
      // Implement a Linux-like 'memory area' for our 'process'.
      // Changes the size of the memory area by |bytes|; returns the
      // address of the previous top ('break') of the memory area
      // We control the "dynamic" memory - DYNAMIC_BASE to DYNAMICTOP
      var self = _sbrk;
      if (!self.called) {
        DYNAMICTOP = alignMemoryPage(DYNAMICTOP); // make sure we start out aligned
        self.called = true;
        assert(Runtime.dynamicAlloc);
        self.alloc = Runtime.dynamicAlloc;
        Runtime.dynamicAlloc = function() { abort('cannot dynamically allocate, sbrk now has control') };
      }
      var ret = DYNAMICTOP;
      if (bytes != 0) {
        var success = self.alloc(bytes);
        if (!success) return -1 >>> 0; // sbrk failure code
      }
      return ret;  // Previous break location.
    }

  function _alcMakeContextCurrent(context) {
      if (context == 0) {
        AL.currentContext = null;
        return 0;
      } else {
        AL.currentContext = AL.contexts[context - 1];
        return 1;
      }
    }

  function _emscripten_glVertexAttrib4fv(index, v) {
      v = HEAPF32.subarray((v)>>2,(v+16)>>2);
      GLctx.vertexAttrib4fv(index, v);
    }

  function _emscripten_glDeleteFramebuffers(n, framebuffers) {
      for (var i = 0; i < n; ++i) {
        var id = HEAP32[(((framebuffers)+(i*4))>>2)];
        var framebuffer = GL.framebuffers[id];
        if (!framebuffer) continue; // GL spec: "glDeleteFramebuffers silently ignores 0s and names that do not correspond to existing framebuffer objects".
        GLctx.deleteFramebuffer(framebuffer);
        framebuffer.name = 0;
        GL.framebuffers[id] = null;
      }
    }

  function ___syscall85(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // readlink
      var path = SYSCALLS.getStr(), buf = SYSCALLS.get(), bufsize = SYSCALLS.get();
      return SYSCALLS.doReadlink(path, buf, bufsize);
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function _emscripten_glGetShaderSource(shader, bufSize, length, source) {
      var result = GLctx.getShaderSource(GL.shaders[shader]);
      if (!result) return; // If an error occurs, nothing will be written to length or source.
      result = result.slice(0, Math.max(0, bufSize - 1));
      if (bufSize > 0 && source) {
        writeStringToMemory(result, source);
        if (length) HEAP32[((length)>>2)]=result.length;
      } else {
        if (length) HEAP32[((length)>>2)]=0;
      }
    }


  function _emscripten_glBindVertexArray(vao) {
      GLctx['bindVertexArray'](GL.vaos[vao]);
    }

  function ___cxa_guard_release() {}

  function _emscripten_set_gamepadconnected_callback(userData, useCapture, callbackfunc) {
      if (!navigator.getGamepads && !navigator.webkitGetGamepads) return -1;
      JSEvents.registerGamepadEventCallback(window, userData, useCapture, callbackfunc, 26, "gamepadconnected");
      return 0;
    }


  function _emscripten_glGetFloatv(name_, p) {
      emscriptenWebGLGet(name_, p, 'Float');
    }

  function ___assert_fail(condition, filename, line, func) {
      ABORT = true;
      throw 'Assertion failed: ' + Pointer_stringify(condition) + ', at: ' + [filename ? Pointer_stringify(filename) : 'unknown filename', line, func ? Pointer_stringify(func) : 'unknown function'] + ' at ' + stackTrace();
    }

  function _alDopplerFactor(value) {
      Runtime.warnOnce('alDopplerFactor() is not yet implemented! Ignoring all calls to it.');
    }

  function _emscripten_glUniform3fv(location, count, value) {
      location = GL.uniforms[location];
      var view;
      if (count === 1) {
        // avoid allocation for the common case of uploading one uniform
        view = GL.miniTempBufferViews[2];
        view[0] = HEAPF32[((value)>>2)];
        view[1] = HEAPF32[(((value)+(4))>>2)];
        view[2] = HEAPF32[(((value)+(8))>>2)];
      } else {
        view = HEAPF32.subarray((value)>>2,(value+count*12)>>2);
      }
      GLctx.uniform3fv(location, view);
    }

  function _emscripten_glDrawElementsInstanced(mode, count, type, indices, primcount) {
      GLctx['drawElementsInstanced'](mode, count, type, indices, primcount);
    }

  function _eglMakeCurrent(display, draw, read, context) { 
      if (display != 62000 /* Magic ID for Emscripten 'default display' */) {
        EGL.setErrorCode(0x3008 /* EGL_BAD_DISPLAY */);
        return 0 /* EGL_FALSE */;
      }
      //\todo An EGL_NOT_INITIALIZED error is generated if EGL is not initialized for dpy. 
      if (context != 0 && context != 62004 /* Magic ID for Emscripten EGLContext */) {
        EGL.setErrorCode(0x3006 /* EGL_BAD_CONTEXT */);
        return 0;
      }
      if ((read != 0 && read != 62006) || (draw != 0 && draw != 62006 /* Magic ID for Emscripten 'default surface' */)) {
        EGL.setErrorCode(0x300D /* EGL_BAD_SURFACE */);
        return 0;
      }
      EGL.currentContext = context;
      EGL.currentDrawSurface = draw;
      EGL.currentReadSurface = read;
      EGL.setErrorCode(0x3000 /* EGL_SUCCESS */);
      return 1 /* EGL_TRUE */;
    }

  function _emscripten_glDrawElements(mode, count, type, indices) {
      var buf;
      if (!GL.currElementArrayBuffer) {
        var size = GL.calcBufLength(1, type, 0, count);
        buf = GL.getTempIndexBuffer(size);
        GLctx.bindBuffer(GLctx.ELEMENT_ARRAY_BUFFER, buf);
        GLctx.bufferSubData(GLctx.ELEMENT_ARRAY_BUFFER,
                                 0,
                                 HEAPU8.subarray(indices, indices + size));
        // the index is now 0
        indices = 0;
      }
  
      // bind any client-side buffers
      GL.preDrawHandleClientVertexAttribBindings(count);
  
      GLctx.drawElements(mode, count, type, indices);
  
      GL.postDrawHandleClientVertexAttribBindings(count);
  
      if (!GL.currElementArrayBuffer) {
        GLctx.bindBuffer(GLctx.ELEMENT_ARRAY_BUFFER, null);
      }
    }

  function _emscripten_set_main_loop_arg(func, arg, fps, simulateInfiniteLoop) {
      _emscripten_set_main_loop(func, fps, simulateInfiniteLoop, arg);
    }

  function _emscripten_glCreateProgram() {
      var id = GL.getNewId(GL.programs);
      var program = GLctx.createProgram();
      program.name = id;
      GL.programs[id] = program;
      return id;
    }

  function _pthread_once(ptr, func) {
      if (!_pthread_once.seen) _pthread_once.seen = {};
      if (ptr in _pthread_once.seen) return;
      Runtime.dynCall('v', func);
      _pthread_once.seen[ptr] = 1;
    }

  function _emscripten_glCompressedTexImage2D(target, level, internalFormat, width, height, border, imageSize, data) {
      var heapView;
      if (data) {
        heapView = HEAPU8.subarray((data),(data+imageSize));
      } else {
        heapView = null;
      }
      GLctx['compressedTexImage2D'](target, level, internalFormat, width, height, border, heapView);
    }

  function _emscripten_glClearColor(x0, x1, x2, x3) { GLctx.clearColor(x0, x1, x2, x3) }

  function ___unlock() {}

  var _llvm_fabs_f64=Math_abs;

  function _alcCaptureStop() {
  Module['printErr']('missing function: alcCaptureStop'); abort(-1);
  }

  function _setenv(envname, envval, overwrite) {
      // int setenv(const char *envname, const char *envval, int overwrite);
      // http://pubs.opengroup.org/onlinepubs/009695399/functions/setenv.html
      if (envname === 0) {
        ___setErrNo(ERRNO_CODES.EINVAL);
        return -1;
      }
      var name = Pointer_stringify(envname);
      var val = Pointer_stringify(envval);
      if (name === '' || name.indexOf('=') !== -1) {
        ___setErrNo(ERRNO_CODES.EINVAL);
        return -1;
      }
      if (ENV.hasOwnProperty(name) && !overwrite) return 0;
      ENV[name] = val;
      ___buildEnvironment(ENV);
      return 0;
    }

   
  Module["_bitshift64Ashr"] = _bitshift64Ashr;

  function _emscripten_glLoadMatrixf() {
  Module['printErr']('missing function: emscripten_glLoadMatrixf'); abort(-1);
  }

  
  function _malloc(bytes) {
      /* Over-allocate to make sure it is byte-aligned by 8.
       * This will leak memory, but this is only the dummy
       * implementation (replaced by dlmalloc normally) so
       * not an issue.
       */
      var ptr = Runtime.dynamicAlloc(bytes + 8);
      return (ptr+8) & 0xFFFFFFF8;
    }
  Module["_malloc"] = _malloc;function ___cxa_allocate_exception(size) {
      return _malloc(size);
    }

  function ___syscall142(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // newselect
      // readfds are supported,
      // writefds checks socket open status
      // exceptfds not supported
      // timeout is always 0 - fully async
      var nfds = SYSCALLS.get(), readfds = SYSCALLS.get(), writefds = SYSCALLS.get(), exceptfds = SYSCALLS.get(), timeout = SYSCALLS.get();
  
      assert(nfds <= 64, 'nfds must be less than or equal to 64');  // fd sets have 64 bits // TODO: this could be 1024 based on current musl headers
      assert(!exceptfds, 'exceptfds not supported');
  
      var total = 0;
      
      var srcReadLow = (readfds ? HEAP32[((readfds)>>2)] : 0),
          srcReadHigh = (readfds ? HEAP32[(((readfds)+(4))>>2)] : 0);
      var srcWriteLow = (writefds ? HEAP32[((writefds)>>2)] : 0),
          srcWriteHigh = (writefds ? HEAP32[(((writefds)+(4))>>2)] : 0);
      var srcExceptLow = (exceptfds ? HEAP32[((exceptfds)>>2)] : 0),
          srcExceptHigh = (exceptfds ? HEAP32[(((exceptfds)+(4))>>2)] : 0);
  
      var dstReadLow = 0,
          dstReadHigh = 0;
      var dstWriteLow = 0,
          dstWriteHigh = 0;
      var dstExceptLow = 0,
          dstExceptHigh = 0;
  
      var allLow = (readfds ? HEAP32[((readfds)>>2)] : 0) |
                   (writefds ? HEAP32[((writefds)>>2)] : 0) |
                   (exceptfds ? HEAP32[((exceptfds)>>2)] : 0);
      var allHigh = (readfds ? HEAP32[(((readfds)+(4))>>2)] : 0) |
                    (writefds ? HEAP32[(((writefds)+(4))>>2)] : 0) |
                    (exceptfds ? HEAP32[(((exceptfds)+(4))>>2)] : 0);
  
      function check(fd, low, high, val) {
        return (fd < 32 ? (low & val) : (high & val));
      }
  
      for (var fd = 0; fd < nfds; fd++) {
        var mask = 1 << (fd % 32);
        if (!(check(fd, allLow, allHigh, mask))) {
          continue;  // index isn't in the set
        }
  
        var stream = FS.getStream(fd);
        if (!stream) throw new FS.ErrnoError(ERRNO_CODES.EBADF);
  
        var flags = SYSCALLS.DEFAULT_POLLMASK;
  
        if (stream.stream_ops.poll) {
          flags = stream.stream_ops.poll(stream);
        }
  
        if ((flags & 1) && check(fd, srcReadLow, srcReadHigh, mask)) {
          fd < 32 ? (dstReadLow = dstReadLow | mask) : (dstReadHigh = dstReadHigh | mask);
          total++;
        }
        if ((flags & 4) && check(fd, srcWriteLow, srcWriteHigh, mask)) {
          fd < 32 ? (dstWriteLow = dstWriteLow | mask) : (dstWriteHigh = dstWriteHigh | mask);
          total++;
        }
        if ((flags & 2) && check(fd, srcExceptLow, srcExceptHigh, mask)) {
          fd < 32 ? (dstExceptLow = dstExceptLow | mask) : (dstExceptHigh = dstExceptHigh | mask);
          total++;
        }
      }
  
      if (readfds) {
        HEAP32[((readfds)>>2)]=dstReadLow;
        HEAP32[(((readfds)+(4))>>2)]=dstReadHigh;
      }
      if (writefds) {
        HEAP32[((writefds)>>2)]=dstWriteLow;
        HEAP32[(((writefds)+(4))>>2)]=dstWriteHigh;
      }
      if (exceptfds) {
        HEAP32[((exceptfds)>>2)]=dstExceptLow;
        HEAP32[(((exceptfds)+(4))>>2)]=dstExceptHigh;
      }
      
      return total;
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function _emscripten_glGetProgramInfoLog(program, maxLength, length, infoLog) {
      var log = GLctx.getProgramInfoLog(GL.programs[program]);
      if (log === null) log = '(unknown error)';
  
      log = log.substr(0, maxLength - 1);
      if (maxLength > 0 && infoLog) {
        writeStringToMemory(log, infoLog);
        if (length) HEAP32[((length)>>2)]=log.length;
      } else {
        if (length) HEAP32[((length)>>2)]=0;
      }
    }

  function _emscripten_glTexImage2D(target, level, internalFormat, width, height, border, format, type, pixels) {
      var pixelData;
      if (pixels) {
        var data = emscriptenWebGLGetTexPixelData(type, format, width, height, pixels, internalFormat);
        pixelData = data.pixels;
        internalFormat = data.internalFormat;
      } else {
        pixelData = null;
      }
      GLctx.texImage2D(target, level, internalFormat, width, height, border, format, type, pixelData);
    }

  function _emscripten_glColorPointer() {
  Module['printErr']('missing function: emscripten_glColorPointer'); abort(-1);
  }

  function _emscripten_glFlush() { GLctx.flush() }

  function ___syscall183(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // getcwd
      var buf = SYSCALLS.get(), size = SYSCALLS.get();
      if (size === 0) return -ERRNO_CODES.EINVAL;
      var cwd = FS.cwd();
      if (size < cwd.length + 1) return -ERRNO_CODES.ERANGE;
      writeAsciiToMemory(cwd, buf);
      return buf;
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function _llvm_eh_typeid_for(type) {
      return type;
    }

  function _emscripten_glCreateShader(shaderType) {
      var id = GL.getNewId(GL.shaders);
      GL.shaders[id] = GLctx.createShader(shaderType);
      return id;
    }

  function _alcGetError(device) {
      var err = AL.alcErr;
      AL.alcErr = 0;
      return err;
    }

  function _emscripten_glIsShader(shader) {
      var s = GL.shaders[shader];
      if (!s) return 0;
      return GLctx.isShader(s);
    }

  function _difftime(time1, time0) {
      return time1 - time0;
    }

  function _emscripten_glColorMask(x0, x1, x2, x3) { GLctx.colorMask(x0, x1, x2, x3) }

  function _emscripten_glPixelStorei(pname, param) {
      if (pname == 0x0D05 /* GL_PACK_ALIGNMENT */) {
        GL.packAlignment = param;
      } else if (pname == 0x0cf5 /* GL_UNPACK_ALIGNMENT */) {
        GL.unpackAlignment = param;
      }
      GLctx.pixelStorei(pname, param);
    }

  function _pthread_cleanup_pop() {
      assert(_pthread_cleanup_push.level == __ATEXIT__.length, 'cannot pop if something else added meanwhile!');
      __ATEXIT__.pop();
      _pthread_cleanup_push.level = __ATEXIT__.length;
    }

  function _emscripten_glDeleteTextures(n, textures) {
      for (var i = 0; i < n; i++) {
        var id = HEAP32[(((textures)+(i*4))>>2)];
        var texture = GL.textures[id];
        if (!texture) continue; // GL spec: "glDeleteTextures silently ignores 0s and names that do not correspond to existing textures".
        GLctx.deleteTexture(texture);
        texture.name = 0;
        GL.textures[id] = null;
      }
    }

  function _eglGetDisplay(nativeDisplayType) {
      EGL.setErrorCode(0x3000 /* EGL_SUCCESS */);
      // Note: As a 'conformant' implementation of EGL, we would prefer to init here only if the user
      //       calls this function with EGL_DEFAULT_DISPLAY. Other display IDs would be preferred to be unsupported
      //       and EGL_NO_DISPLAY returned. Uncomment the following code lines to do this.
      // Instead, an alternative route has been preferred, namely that the Emscripten EGL implementation
      // "emulates" X11, and eglGetDisplay is expected to accept/receive a pointer to an X11 Display object.
      // Therefore, be lax and allow anything to be passed in, and return the magic handle to our default EGLDisplay object.
  
  //    if (nativeDisplayType == 0 /* EGL_DEFAULT_DISPLAY */) {
          return 62000; // Magic ID for Emscripten 'default display'
  //    }
  //    else
  //      return 0; // EGL_NO_DISPLAY
    }

  function _emscripten_set_canvas_size(width, height) {
      Browser.setCanvasSize(width, height);
    }

  function _emscripten_glCompileShader(shader) {
      GLctx.compileShader(GL.shaders[shader]);
    }

  function _emscripten_glGenVertexArrays(n, arrays) {
  
      for(var i = 0; i < n; i++) {
        var vao = GLctx['createVertexArray']();
        if (!vao) {
          GL.recordError(0x0502 /* GL_INVALID_OPERATION */);
          while(i < n) HEAP32[(((arrays)+(i++*4))>>2)]=0;
          return;
        }
        var id = GL.getNewId(GL.vaos);
        vao.name = id;
        GL.vaos[id] = vao;
        HEAP32[(((arrays)+(i*4))>>2)]=id;
      }
    }

  function _time(ptr) {
      var ret = (Date.now()/1000)|0;
      if (ptr) {
        HEAP32[((ptr)>>2)]=ret;
      }
      return ret;
    }

  function _pthread_self() {
      //FIXME: assumes only a single thread
      return 0;
    }

  function _emscripten_glGetBooleanv(name_, p) {
      emscriptenWebGLGet(name_, p, 'Boolean');
    }

  function ___syscall221(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // fcntl64
      var stream = SYSCALLS.getStreamFromFD(), cmd = SYSCALLS.get();
      switch (cmd) {
        case 0: {
          var arg = SYSCALLS.get();
          if (arg < 0) {
            return -ERRNO_CODES.EINVAL;
          }
          var newStream;
          newStream = FS.open(stream.path, stream.flags, 0, arg);
          return newStream.fd;
        }
        case 1:
        case 2:
          return 0;  // FD_CLOEXEC makes no sense for a single process.
        case 3:
          return stream.flags;
        case 4: {
          var arg = SYSCALLS.get();
          stream.flags |= arg;
          return 0;
        }
        case 12:
        case 12: {
          var arg = SYSCALLS.get();
          var offset = 0;
          // We're always unlocked.
          HEAP16[(((arg)+(offset))>>1)]=2;
          return 0;
        }
        case 13:
        case 14:
        case 13:
        case 14:
          return 0; // Pretend that the locking is successful.
        case 16:
        case 8:
          return -ERRNO_CODES.EINVAL; // These are for sockets. We don't have them fully implemented yet.
        case 9:
          // musl trusts getown return values, due to a bug where they must be, as they overlap with errors. just return -1 here, so fnctl() returns that, and we set errno ourselves.
          ___setErrNo(ERRNO_CODES.EINVAL);
          return -1;
        default: {
          return -ERRNO_CODES.EINVAL;
        }
      }
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function ___syscall220(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // SYS_getdents64
      var stream = SYSCALLS.getStreamFromFD(), dirp = SYSCALLS.get(), count = SYSCALLS.get();
      if (!stream.getdents) {
        stream.getdents = FS.readdir(stream.path);
      }
      var pos = 0;
      while (stream.getdents.length > 0 && pos + 268 < count) {
        var id;
        var type;
        var name = stream.getdents.pop();
        assert(name.length < 256); // limit of dirent struct
        if (name[0] === '.') {
          id = 1;
          type = 4; // DT_DIR
        } else {
          var child = FS.lookupNode(stream.node, name);
          id = child.id;
          type = FS.isChrdev(child.mode) ? 2 :  // DT_CHR, character device.
                 FS.isDir(child.mode) ? 4 :     // DT_DIR, directory.
                 FS.isLink(child.mode) ? 10 :   // DT_LNK, symbolic link.
                 8;                             // DT_REG, regular file.
        }
        HEAP32[((dirp + pos)>>2)]=id;
        HEAP32[(((dirp + pos)+(4))>>2)]=stream.position;
        HEAP16[(((dirp + pos)+(8))>>1)]=268;
        HEAP8[(((dirp + pos)+(10))>>0)]=type;
        for (var i = 0; i < name.length; i++) {
          HEAP8[(((dirp + pos)+(11 + i))>>0)]=name.charCodeAt(i);
        }
        HEAP8[(((dirp + pos)+(11 + i))>>0)]=0;
        pos += 268;
      }
      return pos;
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  var ___dso_handle=STATICTOP; STATICTOP += 16;;
var GLctx; GL.init();
Module["requestFullScreen"] = function Module_requestFullScreen(lockPointer, resizeCanvas, vrDevice) { Browser.requestFullScreen(lockPointer, resizeCanvas, vrDevice) };
  Module["requestAnimationFrame"] = function Module_requestAnimationFrame(func) { Browser.requestAnimationFrame(func) };
  Module["setCanvasSize"] = function Module_setCanvasSize(width, height, noUpdates) { Browser.setCanvasSize(width, height, noUpdates) };
  Module["pauseMainLoop"] = function Module_pauseMainLoop() { Browser.mainLoop.pause() };
  Module["resumeMainLoop"] = function Module_resumeMainLoop() { Browser.mainLoop.resume() };
  Module["getUserMedia"] = function Module_getUserMedia() { Browser.getUserMedia() }
  Module["createContext"] = function Module_createContext(canvas, useWebGL, setInModule, webGLContextAttributes) { return Browser.createContext(canvas, useWebGL, setInModule, webGLContextAttributes) };
FS.staticInit();__ATINIT__.unshift(function() { if (!Module["noFSInit"] && !FS.init.initialized) FS.init() });__ATMAIN__.push(function() { FS.ignorePermissions = false });__ATEXIT__.push(function() { FS.quit() });Module["FS_createFolder"] = FS.createFolder;Module["FS_createPath"] = FS.createPath;Module["FS_createDataFile"] = FS.createDataFile;Module["FS_createPreloadedFile"] = FS.createPreloadedFile;Module["FS_createLazyFile"] = FS.createLazyFile;Module["FS_createLink"] = FS.createLink;Module["FS_createDevice"] = FS.createDevice;Module["FS_unlink"] = FS.unlink;;
__ATINIT__.unshift(function() { TTY.init() });__ATEXIT__.push(function() { TTY.shutdown() });;
if (ENVIRONMENT_IS_NODE) { var fs = require("fs"); var NODEJS_PATH = require("path"); NODEFS.staticInit(); };
___buildEnvironment(ENV);;
__ATINIT__.push(function() { SOCKFS.root = FS.mount(SOCKFS, {}, null); });;
STACK_BASE = STACKTOP = Runtime.alignMemory(STATICTOP);

staticSealed = true; // seal the static portion of memory

STACK_MAX = STACK_BASE + TOTAL_STACK;

DYNAMIC_BASE = DYNAMICTOP = Runtime.alignMemory(STACK_MAX);

assert(DYNAMIC_BASE < TOTAL_MEMORY, "TOTAL_MEMORY not big enough for stack");



function nullFunc_iiiiiid(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiiid'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiidiii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiidiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiddddddi(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiddddddi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiiddi(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiiddi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iidddddddddi(x) { Module["printErr"]("Invalid function pointer called with signature 'iidddddddddi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiiiiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiiiiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iidd(x) { Module["printErr"]("Invalid function pointer called with signature 'iidd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_vidi(x) { Module["printErr"]("Invalid function pointer called with signature 'vidi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_vidd(x) { Module["printErr"]("Invalid function pointer called with signature 'vidd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiiiiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiiiiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viidddiddddddd(x) { Module["printErr"]("Invalid function pointer called with signature 'viidddiddddddd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiiiiidi(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiiiiidi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_di(x) { Module["printErr"]("Invalid function pointer called with signature 'di'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiddi(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiddi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiddddd(x) { Module["printErr"]("Invalid function pointer called with signature 'iiddddd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_d(x) { Module["printErr"]("Invalid function pointer called with signature 'd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_dii(x) { Module["printErr"]("Invalid function pointer called with signature 'dii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiidddddi(x) { Module["printErr"]("Invalid function pointer called with signature 'viiidddddi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_did(x) { Module["printErr"]("Invalid function pointer called with signature 'did'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_vdddddd(x) { Module["printErr"]("Invalid function pointer called with signature 'vdddddd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiddddi(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiddddi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_vdd(x) { Module["printErr"]("Invalid function pointer called with signature 'vdd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_vdi(x) { Module["printErr"]("Invalid function pointer called with signature 'vdi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiid(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiid'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiddi(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiddi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiddddi(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiddddi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iidddd(x) { Module["printErr"]("Invalid function pointer called with signature 'iidddd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiddddddi(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiddddddi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiiiiidi(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiiiiidi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viidddd(x) { Module["printErr"]("Invalid function pointer called with signature 'viidddd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_ddd(x) { Module["printErr"]("Invalid function pointer called with signature 'ddd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiidi(x) { Module["printErr"]("Invalid function pointer called with signature 'viiidi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_vidddd(x) { Module["printErr"]("Invalid function pointer called with signature 'vidddd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_diid(x) { Module["printErr"]("Invalid function pointer called with signature 'diid'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiddd(x) { Module["printErr"]("Invalid function pointer called with signature 'iiddd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiddi(x) { Module["printErr"]("Invalid function pointer called with signature 'iiddi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiiiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiiiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_diii(x) { Module["printErr"]("Invalid function pointer called with signature 'diii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiididdddddddd(x) { Module["printErr"]("Invalid function pointer called with signature 'iiididdddddddd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiddi(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiddi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiid(x) { Module["printErr"]("Invalid function pointer called with signature 'viiid'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viidii(x) { Module["printErr"]("Invalid function pointer called with signature 'viidii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiid(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiid'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_vid(x) { Module["printErr"]("Invalid function pointer called with signature 'vid'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiidi(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiidi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiidd(x) { Module["printErr"]("Invalid function pointer called with signature 'iiidd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_vii(x) { Module["printErr"]("Invalid function pointer called with signature 'vii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viidd(x) { Module["printErr"]("Invalid function pointer called with signature 'viidd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viddii(x) { Module["printErr"]("Invalid function pointer called with signature 'viddii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viidi(x) { Module["printErr"]("Invalid function pointer called with signature 'viidi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viddd(x) { Module["printErr"]("Invalid function pointer called with signature 'viddd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiiiiid(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiiiiid'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiiiiiiiiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiiiiiiiiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iid(x) { Module["printErr"]("Invalid function pointer called with signature 'iid'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_dddd(x) { Module["printErr"]("Invalid function pointer called with signature 'dddd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iii(x) { Module["printErr"]("Invalid function pointer called with signature 'iii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiddi(x) { Module["printErr"]("Invalid function pointer called with signature 'viiddi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiddddddddd(x) { Module["printErr"]("Invalid function pointer called with signature 'viiddddddddd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_vdddd(x) { Module["printErr"]("Invalid function pointer called with signature 'vdddd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viii(x) { Module["printErr"]("Invalid function pointer called with signature 'viii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_v(x) { Module["printErr"]("Invalid function pointer called with signature 'v'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viid(x) { Module["printErr"]("Invalid function pointer called with signature 'viid'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_vd(x) { Module["printErr"]("Invalid function pointer called with signature 'vd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_vi(x) { Module["printErr"]("Invalid function pointer called with signature 'vi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viddddddddd(x) { Module["printErr"]("Invalid function pointer called with signature 'viddddddddd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_ii(x) { Module["printErr"]("Invalid function pointer called with signature 'ii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_ddddd(x) { Module["printErr"]("Invalid function pointer called with signature 'ddddd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiddddddddd(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiddddddddd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_dd(x) { Module["printErr"]("Invalid function pointer called with signature 'dd'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiidii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiidii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_diiiid(x) { Module["printErr"]("Invalid function pointer called with signature 'diiiid'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiidddddddddi(x) { Module["printErr"]("Invalid function pointer called with signature 'iiidddddddddi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_i(x) { Module["printErr"]("Invalid function pointer called with signature 'i'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiidii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiidii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiiiiiid(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiiiiiid'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiidddddi(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiidddddi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function invoke_iiiiiid(index,a1,a2,a3,a4,a5,a6) {
  try {
    return Module["dynCall_iiiiiid"](index,a1,a2,a3,a4,a5,a6);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiidiii(index,a1,a2,a3,a4,a5,a6,a7) {
  try {
    Module["dynCall_viiidiii"](index,a1,a2,a3,a4,a5,a6,a7);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiiddddddi(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  try {
    Module["dynCall_viiiddddddi"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiiiii(index,a1,a2,a3,a4,a5,a6) {
  try {
    return Module["dynCall_iiiiiii"](index,a1,a2,a3,a4,a5,a6);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiiiddi(index,a1,a2,a3,a4,a5,a6,a7) {
  try {
    Module["dynCall_viiiiddi"](index,a1,a2,a3,a4,a5,a6,a7);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iidddddddddi(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  try {
    return Module["dynCall_iidddddddddi"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  try {
    Module["dynCall_viiiiiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iidd(index,a1,a2,a3) {
  try {
    return Module["dynCall_iidd"](index,a1,a2,a3);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_vidi(index,a1,a2,a3) {
  try {
    Module["dynCall_vidi"](index,a1,a2,a3);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_vidd(index,a1,a2,a3) {
  try {
    Module["dynCall_vidd"](index,a1,a2,a3);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  try {
    return Module["dynCall_iiiiiiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viidddiddddddd(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13) {
  try {
    Module["dynCall_viidddiddddddd"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiiiiiidi(index,a1,a2,a3,a4,a5,a6,a7,a8,a9) {
  try {
    Module["dynCall_viiiiiiidi"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_di(index,a1) {
  try {
    return Module["dynCall_di"](index,a1);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiiddi(index,a1,a2,a3,a4,a5,a6) {
  try {
    return Module["dynCall_iiiiddi"](index,a1,a2,a3,a4,a5,a6);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiddddd(index,a1,a2,a3,a4,a5,a6) {
  try {
    return Module["dynCall_iiddddd"](index,a1,a2,a3,a4,a5,a6);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_d(index) {
  try {
    return Module["dynCall_d"](index);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_dii(index,a1,a2) {
  try {
    return Module["dynCall_dii"](index,a1,a2);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiidddddi(index,a1,a2,a3,a4,a5,a6,a7,a8,a9) {
  try {
    Module["dynCall_viiidddddi"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_did(index,a1,a2) {
  try {
    return Module["dynCall_did"](index,a1,a2);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9) {
  try {
    return Module["dynCall_iiiiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_vdddddd(index,a1,a2,a3,a4,a5,a6) {
  try {
    Module["dynCall_vdddddd"](index,a1,a2,a3,a4,a5,a6);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiiddddi(index,a1,a2,a3,a4,a5,a6,a7,a8) {
  try {
    Module["dynCall_viiiddddi"](index,a1,a2,a3,a4,a5,a6,a7,a8);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_vdd(index,a1,a2) {
  try {
    Module["dynCall_vdd"](index,a1,a2);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_vdi(index,a1,a2) {
  try {
    Module["dynCall_vdi"](index,a1,a2);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiid(index,a1,a2,a3,a4) {
  try {
    return Module["dynCall_iiiid"](index,a1,a2,a3,a4);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiiiiii(index,a1,a2,a3,a4,a5,a6,a7) {
  try {
    return Module["dynCall_iiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiddi(index,a1,a2,a3,a4,a5) {
  try {
    return Module["dynCall_iiiddi"](index,a1,a2,a3,a4,a5);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiiddddi(index,a1,a2,a3,a4,a5,a6,a7,a8) {
  try {
    return Module["dynCall_iiiiddddi"](index,a1,a2,a3,a4,a5,a6,a7,a8);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iidddd(index,a1,a2,a3,a4,a5) {
  try {
    return Module["dynCall_iidddd"](index,a1,a2,a3,a4,a5);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiiddddddi(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  try {
    return Module["dynCall_iiiiddddddi"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiiiiiidi(index,a1,a2,a3,a4,a5,a6,a7,a8,a9) {
  try {
    return Module["dynCall_iiiiiiiidi"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viidddd(index,a1,a2,a3,a4,a5,a6) {
  try {
    Module["dynCall_viidddd"](index,a1,a2,a3,a4,a5,a6);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiii(index,a1,a2,a3,a4) {
  try {
    return Module["dynCall_iiiii"](index,a1,a2,a3,a4);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiiiii(index,a1,a2,a3,a4,a5,a6) {
  try {
    Module["dynCall_viiiiii"](index,a1,a2,a3,a4,a5,a6);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_ddd(index,a1,a2) {
  try {
    return Module["dynCall_ddd"](index,a1,a2);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiidi(index,a1,a2,a3,a4,a5) {
  try {
    Module["dynCall_viiidi"](index,a1,a2,a3,a4,a5);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_vidddd(index,a1,a2,a3,a4,a5) {
  try {
    Module["dynCall_vidddd"](index,a1,a2,a3,a4,a5);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_diid(index,a1,a2,a3) {
  try {
    return Module["dynCall_diid"](index,a1,a2,a3);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiddd(index,a1,a2,a3,a4) {
  try {
    return Module["dynCall_iiddd"](index,a1,a2,a3,a4);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiddi(index,a1,a2,a3,a4) {
  try {
    return Module["dynCall_iiddi"](index,a1,a2,a3,a4);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  try {
    Module["dynCall_viiiiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_diii(index,a1,a2,a3) {
  try {
    return Module["dynCall_diii"](index,a1,a2,a3);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiididdddddddd(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13) {
  try {
    return Module["dynCall_iiididdddddddd"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiiddi(index,a1,a2,a3,a4,a5,a6) {
  try {
    Module["dynCall_viiiddi"](index,a1,a2,a3,a4,a5,a6);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiiii(index,a1,a2,a3,a4,a5) {
  try {
    return Module["dynCall_iiiiii"](index,a1,a2,a3,a4,a5);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiid(index,a1,a2,a3,a4) {
  try {
    Module["dynCall_viiid"](index,a1,a2,a3,a4);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viidii(index,a1,a2,a3,a4,a5) {
  try {
    Module["dynCall_viidii"](index,a1,a2,a3,a4,a5);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiiid(index,a1,a2,a3,a4,a5) {
  try {
    return Module["dynCall_iiiiid"](index,a1,a2,a3,a4,a5);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiii(index,a1,a2,a3,a4) {
  try {
    Module["dynCall_viiii"](index,a1,a2,a3,a4);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiiii(index,a1,a2,a3,a4,a5) {
  try {
    Module["dynCall_viiiii"](index,a1,a2,a3,a4,a5);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_vid(index,a1,a2) {
  try {
    Module["dynCall_vid"](index,a1,a2);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiidi(index,a1,a2,a3,a4,a5) {
  try {
    return Module["dynCall_iiiidi"](index,a1,a2,a3,a4,a5);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiidd(index,a1,a2,a3,a4) {
  try {
    return Module["dynCall_iiidd"](index,a1,a2,a3,a4);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_vii(index,a1,a2) {
  try {
    Module["dynCall_vii"](index,a1,a2);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viidd(index,a1,a2,a3,a4) {
  try {
    Module["dynCall_viidd"](index,a1,a2,a3,a4);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viddii(index,a1,a2,a3,a4,a5) {
  try {
    Module["dynCall_viddii"](index,a1,a2,a3,a4,a5);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viidi(index,a1,a2,a3,a4) {
  try {
    Module["dynCall_viidi"](index,a1,a2,a3,a4);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viddd(index,a1,a2,a3,a4) {
  try {
    Module["dynCall_viddd"](index,a1,a2,a3,a4);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8) {
  try {
    Module["dynCall_viiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7,a8);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiiiiiid(index,a1,a2,a3,a4,a5,a6,a7,a8) {
  try {
    Module["dynCall_viiiiiiid"](index,a1,a2,a3,a4,a5,a6,a7,a8);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiiiiiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  try {
    Module["dynCall_viiiiiiiiiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iid(index,a1,a2) {
  try {
    return Module["dynCall_iid"](index,a1,a2);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_dddd(index,a1,a2,a3) {
  try {
    return Module["dynCall_dddd"](index,a1,a2,a3);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiiiiii(index,a1,a2,a3,a4,a5,a6,a7) {
  try {
    Module["dynCall_viiiiiii"](index,a1,a2,a3,a4,a5,a6,a7);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9) {
  try {
    Module["dynCall_viiiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iii(index,a1,a2) {
  try {
    return Module["dynCall_iii"](index,a1,a2);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiddi(index,a1,a2,a3,a4,a5) {
  try {
    Module["dynCall_viiddi"](index,a1,a2,a3,a4,a5);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiddddddddd(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  try {
    Module["dynCall_viiddddddddd"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_vdddd(index,a1,a2,a3,a4) {
  try {
    Module["dynCall_vdddd"](index,a1,a2,a3,a4);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viii(index,a1,a2,a3) {
  try {
    Module["dynCall_viii"](index,a1,a2,a3);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_v(index) {
  try {
    Module["dynCall_v"](index);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viid(index,a1,a2,a3) {
  try {
    Module["dynCall_viid"](index,a1,a2,a3);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_vd(index,a1) {
  try {
    Module["dynCall_vd"](index,a1);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_vi(index,a1) {
  try {
    Module["dynCall_vi"](index,a1);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viddddddddd(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  try {
    Module["dynCall_viddddddddd"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_ii(index,a1) {
  try {
    return Module["dynCall_ii"](index,a1);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_ddddd(index,a1,a2,a3,a4) {
  try {
    return Module["dynCall_ddddd"](index,a1,a2,a3,a4);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiddddddddd(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  try {
    return Module["dynCall_iiiddddddddd"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiii(index,a1,a2,a3) {
  try {
    return Module["dynCall_iiii"](index,a1,a2,a3);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_dd(index,a1) {
  try {
    return Module["dynCall_dd"](index,a1);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiiidii(index,a1,a2,a3,a4,a5,a6,a7) {
  try {
    Module["dynCall_viiiidii"](index,a1,a2,a3,a4,a5,a6,a7);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_diiiid(index,a1,a2,a3,a4,a5) {
  try {
    return Module["dynCall_diiiid"](index,a1,a2,a3,a4,a5);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiidddddddddi(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  try {
    return Module["dynCall_iiidddddddddi"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_i(index) {
  try {
    return Module["dynCall_i"](index);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiiidii(index,a1,a2,a3,a4,a5,a6,a7) {
  try {
    return Module["dynCall_iiiiidii"](index,a1,a2,a3,a4,a5,a6,a7);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8) {
  try {
    return Module["dynCall_iiiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7,a8);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_viiiiiiiid(index,a1,a2,a3,a4,a5,a6,a7,a8,a9) {
  try {
    Module["dynCall_viiiiiiiid"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

function invoke_iiiidddddi(index,a1,a2,a3,a4,a5,a6,a7,a8,a9) {
  try {
    return Module["dynCall_iiiidddddi"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    asm["setThrew"](1, 0);
  }
}

Module.asmGlobalArg = { "Math": Math, "Int8Array": Int8Array, "Int16Array": Int16Array, "Int32Array": Int32Array, "Uint8Array": Uint8Array, "Uint16Array": Uint16Array, "Uint32Array": Uint32Array, "Float32Array": Float32Array, "Float64Array": Float64Array, "NaN": NaN, "Infinity": Infinity, "byteLength": byteLength };

Module.asmLibraryArg = { "abort": abort, "assert": assert, "nullFunc_iiiiiid": nullFunc_iiiiiid, "nullFunc_viiidiii": nullFunc_viiidiii, "nullFunc_viiiddddddi": nullFunc_viiiddddddi, "nullFunc_iiiiiii": nullFunc_iiiiiii, "nullFunc_viiiiddi": nullFunc_viiiiddi, "nullFunc_iidddddddddi": nullFunc_iidddddddddi, "nullFunc_viiiiiiiiiii": nullFunc_viiiiiiiiiii, "nullFunc_iidd": nullFunc_iidd, "nullFunc_vidi": nullFunc_vidi, "nullFunc_vidd": nullFunc_vidd, "nullFunc_iiiiiiiiiiii": nullFunc_iiiiiiiiiiii, "nullFunc_viidddiddddddd": nullFunc_viidddiddddddd, "nullFunc_viiiiiiidi": nullFunc_viiiiiiidi, "nullFunc_di": nullFunc_di, "nullFunc_iiiiddi": nullFunc_iiiiddi, "nullFunc_iiddddd": nullFunc_iiddddd, "nullFunc_d": nullFunc_d, "nullFunc_dii": nullFunc_dii, "nullFunc_viiidddddi": nullFunc_viiidddddi, "nullFunc_did": nullFunc_did, "nullFunc_iiiiiiiiii": nullFunc_iiiiiiiiii, "nullFunc_vdddddd": nullFunc_vdddddd, "nullFunc_viiiddddi": nullFunc_viiiddddi, "nullFunc_vdd": nullFunc_vdd, "nullFunc_vdi": nullFunc_vdi, "nullFunc_iiiid": nullFunc_iiiid, "nullFunc_iiiiiiii": nullFunc_iiiiiiii, "nullFunc_iiiddi": nullFunc_iiiddi, "nullFunc_iiiiddddi": nullFunc_iiiiddddi, "nullFunc_iidddd": nullFunc_iidddd, "nullFunc_iiiiddddddi": nullFunc_iiiiddddddi, "nullFunc_iiiiiiiidi": nullFunc_iiiiiiiidi, "nullFunc_viidddd": nullFunc_viidddd, "nullFunc_iiiii": nullFunc_iiiii, "nullFunc_viiiiii": nullFunc_viiiiii, "nullFunc_ddd": nullFunc_ddd, "nullFunc_viiidi": nullFunc_viiidi, "nullFunc_vidddd": nullFunc_vidddd, "nullFunc_diid": nullFunc_diid, "nullFunc_iiddd": nullFunc_iiddd, "nullFunc_iiddi": nullFunc_iiddi, "nullFunc_viiiiiiiiii": nullFunc_viiiiiiiiii, "nullFunc_diii": nullFunc_diii, "nullFunc_iiididdddddddd": nullFunc_iiididdddddddd, "nullFunc_viiiddi": nullFunc_viiiddi, "nullFunc_iiiiii": nullFunc_iiiiii, "nullFunc_viiid": nullFunc_viiid, "nullFunc_viidii": nullFunc_viidii, "nullFunc_iiiiid": nullFunc_iiiiid, "nullFunc_viiii": nullFunc_viiii, "nullFunc_viiiii": nullFunc_viiiii, "nullFunc_vid": nullFunc_vid, "nullFunc_iiiidi": nullFunc_iiiidi, "nullFunc_iiidd": nullFunc_iiidd, "nullFunc_vii": nullFunc_vii, "nullFunc_viidd": nullFunc_viidd, "nullFunc_viddii": nullFunc_viddii, "nullFunc_viidi": nullFunc_viidi, "nullFunc_viddd": nullFunc_viddd, "nullFunc_viiiiiiii": nullFunc_viiiiiiii, "nullFunc_viiiiiiid": nullFunc_viiiiiiid, "nullFunc_viiiiiiiiiiiiiii": nullFunc_viiiiiiiiiiiiiii, "nullFunc_iid": nullFunc_iid, "nullFunc_dddd": nullFunc_dddd, "nullFunc_viiiiiii": nullFunc_viiiiiii, "nullFunc_viiiiiiiii": nullFunc_viiiiiiiii, "nullFunc_iii": nullFunc_iii, "nullFunc_viiddi": nullFunc_viiddi, "nullFunc_viiddddddddd": nullFunc_viiddddddddd, "nullFunc_vdddd": nullFunc_vdddd, "nullFunc_viii": nullFunc_viii, "nullFunc_v": nullFunc_v, "nullFunc_viid": nullFunc_viid, "nullFunc_vd": nullFunc_vd, "nullFunc_vi": nullFunc_vi, "nullFunc_viddddddddd": nullFunc_viddddddddd, "nullFunc_ii": nullFunc_ii, "nullFunc_ddddd": nullFunc_ddddd, "nullFunc_iiiddddddddd": nullFunc_iiiddddddddd, "nullFunc_iiii": nullFunc_iiii, "nullFunc_dd": nullFunc_dd, "nullFunc_viiiidii": nullFunc_viiiidii, "nullFunc_diiiid": nullFunc_diiiid, "nullFunc_iiidddddddddi": nullFunc_iiidddddddddi, "nullFunc_i": nullFunc_i, "nullFunc_iiiiidii": nullFunc_iiiiidii, "nullFunc_iiiiiiiii": nullFunc_iiiiiiiii, "nullFunc_viiiiiiiid": nullFunc_viiiiiiiid, "nullFunc_iiiidddddi": nullFunc_iiiidddddi, "invoke_iiiiiid": invoke_iiiiiid, "invoke_viiidiii": invoke_viiidiii, "invoke_viiiddddddi": invoke_viiiddddddi, "invoke_iiiiiii": invoke_iiiiiii, "invoke_viiiiddi": invoke_viiiiddi, "invoke_iidddddddddi": invoke_iidddddddddi, "invoke_viiiiiiiiiii": invoke_viiiiiiiiiii, "invoke_iidd": invoke_iidd, "invoke_vidi": invoke_vidi, "invoke_vidd": invoke_vidd, "invoke_iiiiiiiiiiii": invoke_iiiiiiiiiiii, "invoke_viidddiddddddd": invoke_viidddiddddddd, "invoke_viiiiiiidi": invoke_viiiiiiidi, "invoke_di": invoke_di, "invoke_iiiiddi": invoke_iiiiddi, "invoke_iiddddd": invoke_iiddddd, "invoke_d": invoke_d, "invoke_dii": invoke_dii, "invoke_viiidddddi": invoke_viiidddddi, "invoke_did": invoke_did, "invoke_iiiiiiiiii": invoke_iiiiiiiiii, "invoke_vdddddd": invoke_vdddddd, "invoke_viiiddddi": invoke_viiiddddi, "invoke_vdd": invoke_vdd, "invoke_vdi": invoke_vdi, "invoke_iiiid": invoke_iiiid, "invoke_iiiiiiii": invoke_iiiiiiii, "invoke_iiiddi": invoke_iiiddi, "invoke_iiiiddddi": invoke_iiiiddddi, "invoke_iidddd": invoke_iidddd, "invoke_iiiiddddddi": invoke_iiiiddddddi, "invoke_iiiiiiiidi": invoke_iiiiiiiidi, "invoke_viidddd": invoke_viidddd, "invoke_iiiii": invoke_iiiii, "invoke_viiiiii": invoke_viiiiii, "invoke_ddd": invoke_ddd, "invoke_viiidi": invoke_viiidi, "invoke_vidddd": invoke_vidddd, "invoke_diid": invoke_diid, "invoke_iiddd": invoke_iiddd, "invoke_iiddi": invoke_iiddi, "invoke_viiiiiiiiii": invoke_viiiiiiiiii, "invoke_diii": invoke_diii, "invoke_iiididdddddddd": invoke_iiididdddddddd, "invoke_viiiddi": invoke_viiiddi, "invoke_iiiiii": invoke_iiiiii, "invoke_viiid": invoke_viiid, "invoke_viidii": invoke_viidii, "invoke_iiiiid": invoke_iiiiid, "invoke_viiii": invoke_viiii, "invoke_viiiii": invoke_viiiii, "invoke_vid": invoke_vid, "invoke_iiiidi": invoke_iiiidi, "invoke_iiidd": invoke_iiidd, "invoke_vii": invoke_vii, "invoke_viidd": invoke_viidd, "invoke_viddii": invoke_viddii, "invoke_viidi": invoke_viidi, "invoke_viddd": invoke_viddd, "invoke_viiiiiiii": invoke_viiiiiiii, "invoke_viiiiiiid": invoke_viiiiiiid, "invoke_viiiiiiiiiiiiiii": invoke_viiiiiiiiiiiiiii, "invoke_iid": invoke_iid, "invoke_dddd": invoke_dddd, "invoke_viiiiiii": invoke_viiiiiii, "invoke_viiiiiiiii": invoke_viiiiiiiii, "invoke_iii": invoke_iii, "invoke_viiddi": invoke_viiddi, "invoke_viiddddddddd": invoke_viiddddddddd, "invoke_vdddd": invoke_vdddd, "invoke_viii": invoke_viii, "invoke_v": invoke_v, "invoke_viid": invoke_viid, "invoke_vd": invoke_vd, "invoke_vi": invoke_vi, "invoke_viddddddddd": invoke_viddddddddd, "invoke_ii": invoke_ii, "invoke_ddddd": invoke_ddddd, "invoke_iiiddddddddd": invoke_iiiddddddddd, "invoke_iiii": invoke_iiii, "invoke_dd": invoke_dd, "invoke_viiiidii": invoke_viiiidii, "invoke_diiiid": invoke_diiiid, "invoke_iiidddddddddi": invoke_iiidddddddddi, "invoke_i": invoke_i, "invoke_iiiiidii": invoke_iiiiidii, "invoke_iiiiiiiii": invoke_iiiiiiiii, "invoke_viiiiiiiid": invoke_viiiiiiiid, "invoke_iiiidddddi": invoke_iiiidddddi, "_emscripten_glGetTexParameterfv": _emscripten_glGetTexParameterfv, "___syscall221": ___syscall221, "__inet_ntop6_raw": __inet_ntop6_raw, "___syscall220": ___syscall220, "_emscripten_glBlendFuncSeparate": _emscripten_glBlendFuncSeparate, "_emscripten_glGetIntegerv": _emscripten_glGetIntegerv, "___assert_fail": ___assert_fail, "_emscripten_glDepthFunc": _emscripten_glDepthFunc, "_emscripten_memcpy_big": _emscripten_memcpy_big, "_emscripten_glUniform1f": _emscripten_glUniform1f, "emscriptenWebGLComputeImageSize": emscriptenWebGLComputeImageSize, "_alGetFloat": _alGetFloat, "_emscripten_glUniform1i": _emscripten_glUniform1i, "_emscripten_glIsProgram": _emscripten_glIsProgram, "_gmtime_r": _gmtime_r, "___cxa_rethrow": ___cxa_rethrow, "_emscripten_glTexParameteriv": _emscripten_glTexParameteriv, "___syscall140": ___syscall140, "_alSourcePause": _alSourcePause, "___syscall142": ___syscall142, "___syscall145": ___syscall145, "___syscall146": ___syscall146, "_emscripten_glAttachShader": _emscripten_glAttachShader, "_emscripten_get_now_is_monotonic": _emscripten_get_now_is_monotonic, "_alcCreateContext": _alcCreateContext, "_emscripten_glTexParameterfv": _emscripten_glTexParameterfv, "__inet_ntop4_raw": __inet_ntop4_raw, "_emscripten_glUniformMatrix2fv": _emscripten_glUniformMatrix2fv, "_emscripten_glDrawArraysInstanced": _emscripten_glDrawArraysInstanced, "_alcMakeContextCurrent": _alcMakeContextCurrent, "_emscripten_glVertexAttrib2fv": _emscripten_glVertexAttrib2fv, "_emscripten_glBufferData": _emscripten_glBufferData, "_alSourcef": _alSourcef, "_emscripten_glFlush": _emscripten_glFlush, "_alSourcei": _alSourcei, "_alGenBuffers": _alGenBuffers, "_nanosleep": _nanosleep, "_pthread_once": _pthread_once, "_eglWaitClient": _eglWaitClient, "_emscripten_glTexCoordPointer": _emscripten_glTexCoordPointer, "_emscripten_set_blur_callback": _emscripten_set_blur_callback, "_emscripten_glLoadMatrixf": _emscripten_glLoadMatrixf, "_emscripten_glStencilFuncSeparate": _emscripten_glStencilFuncSeparate, "_emscripten_glVertexAttrib3f": _emscripten_glVertexAttrib3f, "_pthread_mutex_lock": _pthread_mutex_lock, "___cxa_guard_abort": ___cxa_guard_abort, "_dlerror": _dlerror, "_emscripten_get_gamepad_status": _emscripten_get_gamepad_status, "___syscall20": ___syscall20, "_emscripten_glUniform1iv": _emscripten_glUniform1iv, "_alListenerfv": _alListenerfv, "emscriptenWebGLGetUniform": emscriptenWebGLGetUniform, "_emscripten_glGetBufferParameteriv": _emscripten_glGetBufferParameteriv, "_emscripten_glVertexAttrib4fv": _emscripten_glVertexAttrib4fv, "_pthread_getspecific": _pthread_getspecific, "_alListenerf": _alListenerf, "_emscripten_glDepthRange": _emscripten_glDepthRange, "_emscripten_asm_const_iii": _emscripten_asm_const_iii, "_eglMakeCurrent": _eglMakeCurrent, "_emscripten_glCopyTexImage2D": _emscripten_glCopyTexImage2D, "_emscripten_glFramebufferTexture2D": _emscripten_glFramebufferTexture2D, "_emscripten_glStencilFunc": _emscripten_glStencilFunc, "___syscall183": ___syscall183, "_alDistanceModel": _alDistanceModel, "_localtime": _localtime, "_emscripten_glRenderbufferStorage": _emscripten_glRenderbufferStorage, "_emscripten_set_keydown_callback": _emscripten_set_keydown_callback, "_emscripten_glVertexPointer": _emscripten_glVertexPointer, "__read_sockaddr": __read_sockaddr, "_eglInitialize": _eglInitialize, "_emscripten_glBufferSubData": _emscripten_glBufferSubData, "_emscripten_glGetUniformfv": _emscripten_glGetUniformfv, "_emscripten_glStencilOp": _emscripten_glStencilOp, "_emscripten_glBlendEquation": _emscripten_glBlendEquation, "_pthread_self": _pthread_self, "_emscripten_glVertexAttrib1fv": _emscripten_glVertexAttrib1fv, "_dlclose": _dlclose, "_emscripten_glGetProgramInfoLog": _emscripten_glGetProgramInfoLog, "_getpwuid": _getpwuid, "___cxa_throw": ___cxa_throw, "_emscripten_glUniform2fv": _emscripten_glUniform2fv, "_emscripten_glBindBuffer": _emscripten_glBindBuffer, "_emscripten_glGetFloatv": _emscripten_glGetFloatv, "_pthread_mutex_init": _pthread_mutex_init, "_emscripten_glGenRenderbuffers": _emscripten_glGenRenderbuffers, "_eglGetDisplay": _eglGetDisplay, "___syscall63": ___syscall63, "_emscripten_glCullFace": _emscripten_glCullFace, "_llvm_fabs_f64": _llvm_fabs_f64, "_emscripten_glStencilMaskSeparate": _emscripten_glStencilMaskSeparate, "_emscripten_glUniform3fv": _emscripten_glUniform3fv, "_emscripten_asm_const_ii": _emscripten_asm_const_ii, "_alSource3f": _alSource3f, "_emscripten_set_keypress_callback": _emscripten_set_keypress_callback, "_emscripten_glDisableVertexAttribArray": _emscripten_glDisableVertexAttribArray, "_eglBindAPI": _eglBindAPI, "_eglCreateContext": _eglCreateContext, "_clock": _clock, "_emscripten_glGetBooleanv": _emscripten_glGetBooleanv, "_emscripten_glVertexAttribDivisor": _emscripten_glVertexAttribDivisor, "_emscripten_glGenBuffers": _emscripten_glGenBuffers, "_emscripten_glDeleteObjectARB": _emscripten_glDeleteObjectARB, "_emscripten_glGetShaderPrecisionFormat": _emscripten_glGetShaderPrecisionFormat, "__write_sockaddr": __write_sockaddr, "_emscripten_request_fullscreen_strategy": _emscripten_request_fullscreen_strategy, "_emscripten_glIsEnabled": _emscripten_glIsEnabled, "_alGetListenerf": _alGetListenerf, "_emscripten_glStencilOpSeparate": _emscripten_glStencilOpSeparate, "_pthread_cleanup_pop": _pthread_cleanup_pop, "_emscripten_asm_const_iiiii": _emscripten_asm_const_iiiii, "___syscall122": ___syscall122, "___cxa_free_exception": ___cxa_free_exception, "___cxa_find_matching_catch": ___cxa_find_matching_catch, "_emscripten_glShaderBinary": _emscripten_glShaderBinary, "___cxa_guard_release": ___cxa_guard_release, "_emscripten_glValidateProgram": _emscripten_glValidateProgram, "_emscripten_glUniform4iv": _emscripten_glUniform4iv, "___setErrNo": ___setErrNo, "_eglSwapBuffers": _eglSwapBuffers, "_emscripten_glVertexAttrib2f": _emscripten_glVertexAttrib2f, "___resumeException": ___resumeException, "_mktime": _mktime, "_emscripten_glGetError": _emscripten_glGetError, "_gethostbyaddr": _gethostbyaddr, "_emscripten_glReadPixels": _emscripten_glReadPixels, "_eglCreateWindowSurface": _eglCreateWindowSurface, "_emscripten_glClearStencil": _emscripten_glClearStencil, "emscriptenWebGLGet": emscriptenWebGLGet, "_emscripten_get_device_pixel_ratio": _emscripten_get_device_pixel_ratio, "_emscripten_set_mouseup_callback": _emscripten_set_mouseup_callback, "_emscripten_glFinish": _emscripten_glFinish, "_emscripten_glClearDepth": _emscripten_glClearDepth, "_emscripten_glUniform1fv": _emscripten_glUniform1fv, "_emscripten_set_resize_callback": _emscripten_set_resize_callback, "_emscripten_glUniform4i": _emscripten_glUniform4i, "_alDopplerFactor": _alDopplerFactor, "_llvm_pow_f64": _llvm_pow_f64, "_emscripten_glUniform4f": _emscripten_glUniform4f, "_emscripten_glBlendFunc": _emscripten_glBlendFunc, "_emscripten_glStencilMask": _emscripten_glStencilMask, "_alcGetIntegerv": _alcGetIntegerv, "_localtime_r": _localtime_r, "_clock_gettime": _clock_gettime, "_strftime": _strftime, "_alcDestroyContext": _alcDestroyContext, "_emscripten_glGetVertexAttribiv": _emscripten_glGetVertexAttribiv, "_emscripten_glUniformMatrix3fv": _emscripten_glUniformMatrix3fv, "___syscall33": ___syscall33, "_pthread_key_create": _pthread_key_create, "_emscripten_glDeleteFramebuffers": _emscripten_glDeleteFramebuffers, "__setLetterbox": __setLetterbox, "___syscall39": ___syscall39, "___syscall38": ___syscall38, "_emscripten_glUniform4fv": _emscripten_glUniform4fv, "_emscripten_glGetObjectParameterivARB": _emscripten_glGetObjectParameterivARB, "_emscripten_glGetUniformiv": _emscripten_glGetUniformiv, "_eglDestroySurface": _eglDestroySurface, "_sigaction": _sigaction, "_difftime": _difftime, "_emscripten_set_mousemove_callback": _emscripten_set_mousemove_callback, "___cxa_find_matching_catch_2": ___cxa_find_matching_catch_2, "___cxa_find_matching_catch_3": ___cxa_find_matching_catch_3, "_emscripten_glDeleteTextures": _emscripten_glDeleteTextures, "_eglDestroyContext": _eglDestroyContext, "_emscripten_exit_fullscreen": _emscripten_exit_fullscreen, "_strftime_l": _strftime_l, "_llvm_eh_typeid_for": _llvm_eh_typeid_for, "__exit": __exit, "_alGetListenerfv": _alGetListenerfv, "_emscripten_glColorMask": _emscripten_glColorMask, "_pthread_mutex_unlock": _pthread_mutex_unlock, "_emscripten_glBindTexture": _emscripten_glBindTexture, "_emscripten_set_main_loop": _emscripten_set_main_loop, "_alGenSources": _alGenSources, "_emscripten_glIsShader": _emscripten_glIsShader, "_emscripten_asm_const_iiii": _emscripten_asm_const_iiii, "_alcOpenDevice": _alcOpenDevice, "_emscripten_glCompressedTexImage2D": _emscripten_glCompressedTexImage2D, "_emscripten_glGetInfoLogARB": _emscripten_glGetInfoLogARB, "_emscripten_longjmp": _emscripten_longjmp, "_atexit": _atexit, "_emscripten_glDeleteVertexArrays": _emscripten_glDeleteVertexArrays, "_emscripten_glReleaseShaderCompiler": _emscripten_glReleaseShaderCompiler, "_dlsym": _dlsym, "___cxa_guard_acquire": ___cxa_guard_acquire, "_emscripten_glFrontFace": _emscripten_glFrontFace, "__ZSt18uncaught_exceptionv": __ZSt18uncaught_exceptionv, "_emscripten_glUseProgram": _emscripten_glUseProgram, "__addDays": __addDays, "_emscripten_set_touchmove_callback": _emscripten_set_touchmove_callback, "_emscripten_glUniform3iv": _emscripten_glUniform3iv, "_emscripten_set_main_loop_arg": _emscripten_set_main_loop_arg, "_sysconf": _sysconf, "_emscripten_glLineWidth": _emscripten_glLineWidth, "___cxa_atexit": ___cxa_atexit, "_emscripten_glScissor": _emscripten_glScissor, "_pthread_cleanup_push": _pthread_cleanup_push, "_emscripten_set_element_css_size": _emscripten_set_element_css_size, "_alSourcePlay": _alSourcePlay, "_emscripten_glIsBuffer": _emscripten_glIsBuffer, "_emscripten_glVertexAttrib1f": _emscripten_glVertexAttrib1f, "_emscripten_set_keyup_callback": _emscripten_set_keyup_callback, "_emscripten_glCompressedTexSubImage2D": _emscripten_glCompressedTexSubImage2D, "_emscripten_glGetAttachedShaders": _emscripten_glGetAttachedShaders, "_emscripten_glGenTextures": _emscripten_glGenTextures, "_alcCaptureSamples": _alcCaptureSamples, "_eglGetConfigAttrib": _eglGetConfigAttrib, "_alGetSourcei": _alGetSourcei, "_alGetSourcef": _alGetSourcef, "_gmtime": _gmtime, "_alSourceQueueBuffers": _alSourceQueueBuffers, "_emscripten_glGetTexParameteriv": _emscripten_glGetTexParameteriv, "___syscall330": ___syscall330, "_emscripten_set_mousedown_callback": _emscripten_set_mousedown_callback, "_emscripten_glClientActiveTexture": _emscripten_glClientActiveTexture, "_emscripten_glCheckFramebufferStatus": _emscripten_glCheckFramebufferStatus, "_eglWaitGL": _eglWaitGL, "___syscall118": ___syscall118, "_emscripten_glUniform3f": _emscripten_glUniform3f, "_emscripten_glUniform3i": _emscripten_glUniform3i, "_emscripten_glDeleteShader": _emscripten_glDeleteShader, "_alDeleteSources": _alDeleteSources, "_emscripten_glGetUniformLocation": _emscripten_glGetUniformLocation, "_emscripten_glEnableVertexAttribArray": _emscripten_glEnableVertexAttribArray, "_alGetError": _alGetError, "_emscripten_get_now": _emscripten_get_now, "__registerRestoreOldStyle": __registerRestoreOldStyle, "emscriptenWebGLGetTexPixelData": emscriptenWebGLGetTexPixelData, "_gettimeofday": _gettimeofday, "_eglWaitNative": _eglWaitNative, "___syscall202": ___syscall202, "_emscripten_glEnableClientState": _emscripten_glEnableClientState, "_eglChooseConfig": _eglChooseConfig, "___cxa_allocate_exception": ___cxa_allocate_exception, "_emscripten_get_num_gamepads": _emscripten_get_num_gamepads, "___buildEnvironment": ___buildEnvironment, "_tzset": _tzset, "_emscripten_glGetAttribLocation": _emscripten_glGetAttribLocation, "_emscripten_glDisable": _emscripten_glDisable, "___cxa_end_catch": ___cxa_end_catch, "_emscripten_glDeleteRenderbuffers": _emscripten_glDeleteRenderbuffers, "_emscripten_glDrawElementsInstanced": _emscripten_glDrawElementsInstanced, "_emscripten_glVertexAttrib4f": _emscripten_glVertexAttrib4f, "_emscripten_glPixelStorei": _emscripten_glPixelStorei, "_llvm_fabs_f32": _llvm_fabs_f32, "_getenv": _getenv, "_emscripten_set_gamepaddisconnected_callback": _emscripten_set_gamepaddisconnected_callback, "_llvm_cttz_i64": _llvm_cttz_i64, "_gethostbyname": _gethostbyname, "_alcCloseDevice": _alcCloseDevice, "_emscripten_glFramebufferRenderbuffer": _emscripten_glFramebufferRenderbuffer, "_alcCaptureStart": _alcCaptureStart, "_llvm_ctpop_i32": _llvm_ctpop_i32, "_emscripten_glRotatef": _emscripten_glRotatef, "_emscripten_glGetShaderiv": _emscripten_glGetShaderiv, "_emscripten_get_battery_status": _emscripten_get_battery_status, "___cxa_pure_virtual": ___cxa_pure_virtual, "_emscripten_glUniformMatrix4fv": _emscripten_glUniformMatrix4fv, "_emscripten_glGetPointerv": _emscripten_glGetPointerv, "_pthread_cond_wait": _pthread_cond_wait, "_emscripten_set_touchstart_callback": _emscripten_set_touchstart_callback, "_emscripten_glIsRenderbuffer": _emscripten_glIsRenderbuffer, "_emscripten_request_pointerlock": _emscripten_request_pointerlock, "___syscall40": ___syscall40, "_llvm_bswap_i64": _llvm_bswap_i64, "_emscripten_set_touchcancel_callback": _emscripten_set_touchcancel_callback, "__inet_pton6_raw": __inet_pton6_raw, "_emscripten_set_focus_callback": _emscripten_set_focus_callback, "_emscripten_glGetVertexAttribfv": _emscripten_glGetVertexAttribfv, "_emscripten_glVertexAttrib3fv": _emscripten_glVertexAttrib3fv, "_emscripten_glCompileShader": _emscripten_glCompileShader, "_glClear": _glClear, "__arraySum": __arraySum, "_emscripten_glLinkProgram": _emscripten_glLinkProgram, "_alDeleteBuffers": _alDeleteBuffers, "_alGetSourcefv": _alGetSourcefv, "_emscripten_get_pointerlock_status": _emscripten_get_pointerlock_status, "_emscripten_glDrawRangeElements": _emscripten_glDrawRangeElements, "___unlock": ___unlock, "_pthread_setspecific": _pthread_setspecific, "_emscripten_glClearColor": _emscripten_glClearColor, "_emscripten_glCreateProgram": _emscripten_glCreateProgram, "_emscripten_glDetachShader": _emscripten_glDetachShader, "_setenv": _setenv, "_emscripten_do_request_fullscreen": _emscripten_do_request_fullscreen, "_emscripten_set_mouseleave_callback": _emscripten_set_mouseleave_callback, "_emscripten_get_element_css_size": _emscripten_get_element_css_size, "_emscripten_set_fullscreenchange_callback": _emscripten_set_fullscreenchange_callback, "_emscripten_glVertexAttribPointer": _emscripten_glVertexAttribPointer, "_alBufferData": _alBufferData, "_emscripten_glDrawArrays": _emscripten_glDrawArrays, "_emscripten_glPolygonOffset": _emscripten_glPolygonOffset, "_longjmp": _longjmp, "_emscripten_glBlendColor": _emscripten_glBlendColor, "_signal": _signal, "_emscripten_set_main_loop_timing": _emscripten_set_main_loop_timing, "_sbrk": _sbrk, "___cxa_begin_catch": ___cxa_begin_catch, "_emscripten_glGetProgramiv": _emscripten_glGetProgramiv, "_emscripten_glGetShaderSource": _emscripten_glGetShaderSource, "_emscripten_glTexImage2D": _emscripten_glTexImage2D, "__isLeapYear": __isLeapYear, "_emscripten_glBlendEquationSeparate": _emscripten_glBlendEquationSeparate, "_emscripten_glGetString": _emscripten_glGetString, "_emscripten_glIsFramebuffer": _emscripten_glIsFramebuffer, "_emscripten_glBindProgramARB": _emscripten_glBindProgramARB, "_glutCreateWindow": _glutCreateWindow, "___syscall85": ___syscall85, "_alcGetError": _alcGetError, "_emscripten_glUniform2i": _emscripten_glUniform2i, "_emscripten_glUniform2f": _emscripten_glUniform2f, "_alSourcefv": _alSourcefv, "_emscripten_glTexParameterf": _emscripten_glTexParameterf, "_emscripten_glTexParameteri": _emscripten_glTexParameteri, "_glutInitDisplayMode": _glutInitDisplayMode, "_emscripten_glGenVertexArrays": _emscripten_glGenVertexArrays, "_emscripten_set_visibilitychange_callback": _emscripten_set_visibilitychange_callback, "_eglGetProcAddress": _eglGetProcAddress, "_emscripten_glBindAttribLocation": _emscripten_glBindAttribLocation, "_llvm_pow_f32": _llvm_pow_f32, "_emscripten_glDrawElements": _emscripten_glDrawElements, "_emscripten_set_canvas_size": _emscripten_set_canvas_size, "_emscripten_glTexSubImage2D": _emscripten_glTexSubImage2D, "_emscripten_asm_const_v": _emscripten_asm_const_v, "_emscripten_glClearDepthf": _emscripten_glClearDepthf, "_emscripten_set_mouseenter_callback": _emscripten_set_mouseenter_callback, "_emscripten_glMatrixMode": _emscripten_glMatrixMode, "___syscall10": ___syscall10, "_emscripten_glNormalPointer": _emscripten_glNormalPointer, "_emscripten_glHint": _emscripten_glHint, "_emscripten_glEnable": _emscripten_glEnable, "___syscall3": ___syscall3, "___lock": ___lock, "_emscripten_glBindFramebuffer": _emscripten_glBindFramebuffer, "___syscall6": ___syscall6, "___syscall5": ___syscall5, "___syscall4": ___syscall4, "_emscripten_glBindRenderbuffer": _emscripten_glBindRenderbuffer, "_time": _time, "_emscripten_glGetFramebufferAttachmentParameteriv": _emscripten_glGetFramebufferAttachmentParameteriv, "_exit": _exit, "_emscripten_glGetActiveAttrib": _emscripten_glGetActiveAttrib, "__inet_pton4_raw": __inet_pton4_raw, "___syscall102": ___syscall102, "___syscall265": ___syscall265, "_emscripten_glClear": _emscripten_glClear, "_emscripten_glGetShaderInfoLog": _emscripten_glGetShaderInfoLog, "_emscripten_glGetVertexAttribPointerv": _emscripten_glGetVertexAttribPointerv, "_llvm_sqrt_f64": _llvm_sqrt_f64, "_alSourceRewind": _alSourceRewind, "_emscripten_glGetActiveUniform": _emscripten_glGetActiveUniform, "emscriptenWebGLGetVertexAttrib": emscriptenWebGLGetVertexAttrib, "___syscall197": ___syscall197, "___syscall196": ___syscall196, "___syscall195": ___syscall195, "_eglSwapInterval": _eglSwapInterval, "___syscall199": ___syscall199, "_emscripten_glDeleteProgram": _emscripten_glDeleteProgram, "_glutDestroyWindow": _glutDestroyWindow, "_emscripten_glCreateShader": _emscripten_glCreateShader, "_pthread_mutex_destroy": _pthread_mutex_destroy, "_emscripten_glColorPointer": _emscripten_glColorPointer, "_emscripten_glViewport": _emscripten_glViewport, "_pthread_cond_broadcast": _pthread_cond_broadcast, "_emscripten_glDepthMask": _emscripten_glDepthMask, "_alcCaptureStop": _alcCaptureStop, "_emscripten_glDrawBuffers": _emscripten_glDrawBuffers, "_alSourceStop": _alSourceStop, "_emscripten_exit_pointerlock": _emscripten_exit_pointerlock, "_emscripten_set_gamepadconnected_callback": _emscripten_set_gamepadconnected_callback, "_abort": _abort, "_emscripten_glGenFramebuffers": _emscripten_glGenFramebuffers, "_emscripten_glLoadIdentity": _emscripten_glLoadIdentity, "_emscripten_glShaderSource": _emscripten_glShaderSource, "___gxx_personality_v0": ___gxx_personality_v0, "_system": _system, "_usleep": _usleep, "_emscripten_set_touchend_callback": _emscripten_set_touchend_callback, "_emscripten_glGetRenderbufferParameteriv": _emscripten_glGetRenderbufferParameteriv, "_eglTerminate": _eglTerminate, "_emscripten_glSampleCoverage": _emscripten_glSampleCoverage, "_emscripten_glFrustum": _emscripten_glFrustum, "_emscripten_glDepthRangef": _emscripten_glDepthRangef, "_emscripten_glGenerateMipmap": _emscripten_glGenerateMipmap, "_emscripten_glIsTexture": _emscripten_glIsTexture, "_emscripten_glBindVertexArray": _emscripten_glBindVertexArray, "_emscripten_glActiveTexture": _emscripten_glActiveTexture, "_emscripten_set_wheel_callback": _emscripten_set_wheel_callback, "_emscripten_glDeleteBuffers": _emscripten_glDeleteBuffers, "___syscall54": ___syscall54, "_emscripten_glUniform2iv": _emscripten_glUniform2iv, "_emscripten_asm_const_i": _emscripten_asm_const_i, "_emscripten_glCopyTexSubImage2D": _emscripten_glCopyTexSubImage2D, "_alSourceUnqueueBuffers": _alSourceUnqueueBuffers, "STACKTOP": STACKTOP, "STACK_MAX": STACK_MAX, "tempDoublePtr": tempDoublePtr, "ABORT": ABORT, "cttz_i8": cttz_i8, "___dso_handle": ___dso_handle };
// EMSCRIPTEN_START_ASM
var asm = (function(global, env, buffer) {
  'almost asm';
  
  
  var Int8View = global.Int8Array;
  var Int16View = global.Int16Array;
  var Int32View = global.Int32Array;
  var Uint8View = global.Uint8Array;
  var Uint16View = global.Uint16Array;
  var Uint32View = global.Uint32Array;
  var Float32View = global.Float32Array;
  var Float64View = global.Float64Array;
  var HEAP8 = new Int8View(buffer);
  var HEAP16 = new Int16View(buffer);
  var HEAP32 = new Int32View(buffer);
  var HEAPU8 = new Uint8View(buffer);
  var HEAPU16 = new Uint16View(buffer);
  var HEAPU32 = new Uint32View(buffer);
  var HEAPF32 = new Float32View(buffer);
  var HEAPF64 = new Float64View(buffer);
  var byteLength = global.byteLength;


  var STACKTOP=env.STACKTOP|0;
  var STACK_MAX=env.STACK_MAX|0;
  var tempDoublePtr=env.tempDoublePtr|0;
  var ABORT=env.ABORT|0;
  var cttz_i8=env.cttz_i8|0;
  var ___dso_handle=env.___dso_handle|0;

  var __THREW__ = 0;
  var threwValue = 0;
  var setjmpId = 0;
  var undef = 0;
  var nan = global.NaN, inf = global.Infinity;
  var tempInt = 0, tempBigInt = 0, tempBigIntP = 0, tempBigIntS = 0, tempBigIntR = 0.0, tempBigIntI = 0, tempBigIntD = 0, tempValue = 0, tempDouble = 0.0;

  var tempRet0 = 0;
  var tempRet1 = 0;
  var tempRet2 = 0;
  var tempRet3 = 0;
  var tempRet4 = 0;
  var tempRet5 = 0;
  var tempRet6 = 0;
  var tempRet7 = 0;
  var tempRet8 = 0;
  var tempRet9 = 0;
  var Math_floor=global.Math.floor;
  var Math_abs=global.Math.abs;
  var Math_sqrt=global.Math.sqrt;
  var Math_pow=global.Math.pow;
  var Math_cos=global.Math.cos;
  var Math_sin=global.Math.sin;
  var Math_tan=global.Math.tan;
  var Math_acos=global.Math.acos;
  var Math_asin=global.Math.asin;
  var Math_atan=global.Math.atan;
  var Math_atan2=global.Math.atan2;
  var Math_exp=global.Math.exp;
  var Math_log=global.Math.log;
  var Math_ceil=global.Math.ceil;
  var Math_imul=global.Math.imul;
  var Math_min=global.Math.min;
  var Math_clz32=global.Math.clz32;
  var abort=env.abort;
  var assert=env.assert;
  var nullFunc_iiiiiid=env.nullFunc_iiiiiid;
  var nullFunc_viiidiii=env.nullFunc_viiidiii;
  var nullFunc_viiiddddddi=env.nullFunc_viiiddddddi;
  var nullFunc_iiiiiii=env.nullFunc_iiiiiii;
  var nullFunc_viiiiddi=env.nullFunc_viiiiddi;
  var nullFunc_iidddddddddi=env.nullFunc_iidddddddddi;
  var nullFunc_viiiiiiiiiii=env.nullFunc_viiiiiiiiiii;
  var nullFunc_iidd=env.nullFunc_iidd;
  var nullFunc_vidi=env.nullFunc_vidi;
  var nullFunc_vidd=env.nullFunc_vidd;
  var nullFunc_iiiiiiiiiiii=env.nullFunc_iiiiiiiiiiii;
  var nullFunc_viidddiddddddd=env.nullFunc_viidddiddddddd;
  var nullFunc_viiiiiiidi=env.nullFunc_viiiiiiidi;
  var nullFunc_di=env.nullFunc_di;
  var nullFunc_iiiiddi=env.nullFunc_iiiiddi;
  var nullFunc_iiddddd=env.nullFunc_iiddddd;
  var nullFunc_d=env.nullFunc_d;
  var nullFunc_dii=env.nullFunc_dii;
  var nullFunc_viiidddddi=env.nullFunc_viiidddddi;
  var nullFunc_did=env.nullFunc_did;
  var nullFunc_iiiiiiiiii=env.nullFunc_iiiiiiiiii;
  var nullFunc_vdddddd=env.nullFunc_vdddddd;
  var nullFunc_viiiddddi=env.nullFunc_viiiddddi;
  var nullFunc_vdd=env.nullFunc_vdd;
  var nullFunc_vdi=env.nullFunc_vdi;
  var nullFunc_iiiid=env.nullFunc_iiiid;
  var nullFunc_iiiiiiii=env.nullFunc_iiiiiiii;
  var nullFunc_iiiddi=env.nullFunc_iiiddi;
  var nullFunc_iiiiddddi=env.nullFunc_iiiiddddi;
  var nullFunc_iidddd=env.nullFunc_iidddd;
  var nullFunc_iiiiddddddi=env.nullFunc_iiiiddddddi;
  var nullFunc_iiiiiiiidi=env.nullFunc_iiiiiiiidi;
  var nullFunc_viidddd=env.nullFunc_viidddd;
  var nullFunc_iiiii=env.nullFunc_iiiii;
  var nullFunc_viiiiii=env.nullFunc_viiiiii;
  var nullFunc_ddd=env.nullFunc_ddd;
  var nullFunc_viiidi=env.nullFunc_viiidi;
  var nullFunc_vidddd=env.nullFunc_vidddd;
  var nullFunc_diid=env.nullFunc_diid;
  var nullFunc_iiddd=env.nullFunc_iiddd;
  var nullFunc_iiddi=env.nullFunc_iiddi;
  var nullFunc_viiiiiiiiii=env.nullFunc_viiiiiiiiii;
  var nullFunc_diii=env.nullFunc_diii;
  var nullFunc_iiididdddddddd=env.nullFunc_iiididdddddddd;
  var nullFunc_viiiddi=env.nullFunc_viiiddi;
  var nullFunc_iiiiii=env.nullFunc_iiiiii;
  var nullFunc_viiid=env.nullFunc_viiid;
  var nullFunc_viidii=env.nullFunc_viidii;
  var nullFunc_iiiiid=env.nullFunc_iiiiid;
  var nullFunc_viiii=env.nullFunc_viiii;
  var nullFunc_viiiii=env.nullFunc_viiiii;
  var nullFunc_vid=env.nullFunc_vid;
  var nullFunc_iiiidi=env.nullFunc_iiiidi;
  var nullFunc_iiidd=env.nullFunc_iiidd;
  var nullFunc_vii=env.nullFunc_vii;
  var nullFunc_viidd=env.nullFunc_viidd;
  var nullFunc_viddii=env.nullFunc_viddii;
  var nullFunc_viidi=env.nullFunc_viidi;
  var nullFunc_viddd=env.nullFunc_viddd;
  var nullFunc_viiiiiiii=env.nullFunc_viiiiiiii;
  var nullFunc_viiiiiiid=env.nullFunc_viiiiiiid;
  var nullFunc_viiiiiiiiiiiiiii=env.nullFunc_viiiiiiiiiiiiiii;
  var nullFunc_iid=env.nullFunc_iid;
  var nullFunc_dddd=env.nullFunc_dddd;
  var nullFunc_viiiiiii=env.nullFunc_viiiiiii;
  var nullFunc_viiiiiiiii=env.nullFunc_viiiiiiiii;
  var nullFunc_iii=env.nullFunc_iii;
  var nullFunc_viiddi=env.nullFunc_viiddi;
  var nullFunc_viiddddddddd=env.nullFunc_viiddddddddd;
  var nullFunc_vdddd=env.nullFunc_vdddd;
  var nullFunc_viii=env.nullFunc_viii;
  var nullFunc_v=env.nullFunc_v;
  var nullFunc_viid=env.nullFunc_viid;
  var nullFunc_vd=env.nullFunc_vd;
  var nullFunc_vi=env.nullFunc_vi;
  var nullFunc_viddddddddd=env.nullFunc_viddddddddd;
  var nullFunc_ii=env.nullFunc_ii;
  var nullFunc_ddddd=env.nullFunc_ddddd;
  var nullFunc_iiiddddddddd=env.nullFunc_iiiddddddddd;
  var nullFunc_iiii=env.nullFunc_iiii;
  var nullFunc_dd=env.nullFunc_dd;
  var nullFunc_viiiidii=env.nullFunc_viiiidii;
  var nullFunc_diiiid=env.nullFunc_diiiid;
  var nullFunc_iiidddddddddi=env.nullFunc_iiidddddddddi;
  var nullFunc_i=env.nullFunc_i;
  var nullFunc_iiiiidii=env.nullFunc_iiiiidii;
  var nullFunc_iiiiiiiii=env.nullFunc_iiiiiiiii;
  var nullFunc_viiiiiiiid=env.nullFunc_viiiiiiiid;
  var nullFunc_iiiidddddi=env.nullFunc_iiiidddddi;
  var invoke_iiiiiid=env.invoke_iiiiiid;
  var invoke_viiidiii=env.invoke_viiidiii;
  var invoke_viiiddddddi=env.invoke_viiiddddddi;
  var invoke_iiiiiii=env.invoke_iiiiiii;
  var invoke_viiiiddi=env.invoke_viiiiddi;
  var invoke_iidddddddddi=env.invoke_iidddddddddi;
  var invoke_viiiiiiiiiii=env.invoke_viiiiiiiiiii;
  var invoke_iidd=env.invoke_iidd;
  var invoke_vidi=env.invoke_vidi;
  var invoke_vidd=env.invoke_vidd;
  var invoke_iiiiiiiiiiii=env.invoke_iiiiiiiiiiii;
  var invoke_viidddiddddddd=env.invoke_viidddiddddddd;
  var invoke_viiiiiiidi=env.invoke_viiiiiiidi;
  var invoke_di=env.invoke_di;
  var invoke_iiiiddi=env.invoke_iiiiddi;
  var invoke_iiddddd=env.invoke_iiddddd;
  var invoke_d=env.invoke_d;
  var invoke_dii=env.invoke_dii;
  var invoke_viiidddddi=env.invoke_viiidddddi;
  var invoke_did=env.invoke_did;
  var invoke_iiiiiiiiii=env.invoke_iiiiiiiiii;
  var invoke_vdddddd=env.invoke_vdddddd;
  var invoke_viiiddddi=env.invoke_viiiddddi;
  var invoke_vdd=env.invoke_vdd;
  var invoke_vdi=env.invoke_vdi;
  var invoke_iiiid=env.invoke_iiiid;
  var invoke_iiiiiiii=env.invoke_iiiiiiii;
  var invoke_iiiddi=env.invoke_iiiddi;
  var invoke_iiiiddddi=env.invoke_iiiiddddi;
  var invoke_iidddd=env.invoke_iidddd;
  var invoke_iiiiddddddi=env.invoke_iiiiddddddi;
  var invoke_iiiiiiiidi=env.invoke_iiiiiiiidi;
  var invoke_viidddd=env.invoke_viidddd;
  var invoke_iiiii=env.invoke_iiiii;
  var invoke_viiiiii=env.invoke_viiiiii;
  var invoke_ddd=env.invoke_ddd;
  var invoke_viiidi=env.invoke_viiidi;
  var invoke_vidddd=env.invoke_vidddd;
  var invoke_diid=env.invoke_diid;
  var invoke_iiddd=env.invoke_iiddd;
  var invoke_iiddi=env.invoke_iiddi;
  var invoke_viiiiiiiiii=env.invoke_viiiiiiiiii;
  var invoke_diii=env.invoke_diii;
  var invoke_iiididdddddddd=env.invoke_iiididdddddddd;
  var invoke_viiiddi=env.invoke_viiiddi;
  var invoke_iiiiii=env.invoke_iiiiii;
  var invoke_viiid=env.invoke_viiid;
  var invoke_viidii=env.invoke_viidii;
  var invoke_iiiiid=env.invoke_iiiiid;
  var invoke_viiii=env.invoke_viiii;
  var invoke_viiiii=env.invoke_viiiii;
  var invoke_vid=env.invoke_vid;
  var invoke_iiiidi=env.invoke_iiiidi;
  var invoke_iiidd=env.invoke_iiidd;
  var invoke_vii=env.invoke_vii;
  var invoke_viidd=env.invoke_viidd;
  var invoke_viddii=env.invoke_viddii;
  var invoke_viidi=env.invoke_viidi;
  var invoke_viddd=env.invoke_viddd;
  var invoke_viiiiiiii=env.invoke_viiiiiiii;
  var invoke_viiiiiiid=env.invoke_viiiiiiid;
  var invoke_viiiiiiiiiiiiiii=env.invoke_viiiiiiiiiiiiiii;
  var invoke_iid=env.invoke_iid;
  var invoke_dddd=env.invoke_dddd;
  var invoke_viiiiiii=env.invoke_viiiiiii;
  var invoke_viiiiiiiii=env.invoke_viiiiiiiii;
  var invoke_iii=env.invoke_iii;
  var invoke_viiddi=env.invoke_viiddi;
  var invoke_viiddddddddd=env.invoke_viiddddddddd;
  var invoke_vdddd=env.invoke_vdddd;
  var invoke_viii=env.invoke_viii;
  var invoke_v=env.invoke_v;
  var invoke_viid=env.invoke_viid;
  var invoke_vd=env.invoke_vd;
  var invoke_vi=env.invoke_vi;
  var invoke_viddddddddd=env.invoke_viddddddddd;
  var invoke_ii=env.invoke_ii;
  var invoke_ddddd=env.invoke_ddddd;
  var invoke_iiiddddddddd=env.invoke_iiiddddddddd;
  var invoke_iiii=env.invoke_iiii;
  var invoke_dd=env.invoke_dd;
  var invoke_viiiidii=env.invoke_viiiidii;
  var invoke_diiiid=env.invoke_diiiid;
  var invoke_iiidddddddddi=env.invoke_iiidddddddddi;
  var invoke_i=env.invoke_i;
  var invoke_iiiiidii=env.invoke_iiiiidii;
  var invoke_iiiiiiiii=env.invoke_iiiiiiiii;
  var invoke_viiiiiiiid=env.invoke_viiiiiiiid;
  var invoke_iiiidddddi=env.invoke_iiiidddddi;
  var _emscripten_glGetTexParameterfv=env._emscripten_glGetTexParameterfv;
  var ___syscall221=env.___syscall221;
  var __inet_ntop6_raw=env.__inet_ntop6_raw;
  var ___syscall220=env.___syscall220;
  var _emscripten_glBlendFuncSeparate=env._emscripten_glBlendFuncSeparate;
  var _emscripten_glGetIntegerv=env._emscripten_glGetIntegerv;
  var ___assert_fail=env.___assert_fail;
  var _emscripten_glDepthFunc=env._emscripten_glDepthFunc;
  var _emscripten_memcpy_big=env._emscripten_memcpy_big;
  var _emscripten_glUniform1f=env._emscripten_glUniform1f;
  var emscriptenWebGLComputeImageSize=env.emscriptenWebGLComputeImageSize;
  var _alGetFloat=env._alGetFloat;
  var _emscripten_glUniform1i=env._emscripten_glUniform1i;
  var _emscripten_glIsProgram=env._emscripten_glIsProgram;
  var _gmtime_r=env._gmtime_r;
  var ___cxa_rethrow=env.___cxa_rethrow;
  var _emscripten_glTexParameteriv=env._emscripten_glTexParameteriv;
  var ___syscall140=env.___syscall140;
  var _alSourcePause=env._alSourcePause;
  var ___syscall142=env.___syscall142;
  var ___syscall145=env.___syscall145;
  var ___syscall146=env.___syscall146;
  var _emscripten_glAttachShader=env._emscripten_glAttachShader;
  var _emscripten_get_now_is_monotonic=env._emscripten_get_now_is_monotonic;
  var _alcCreateContext=env._alcCreateContext;
  var _emscripten_glTexParameterfv=env._emscripten_glTexParameterfv;
  var __inet_ntop4_raw=env.__inet_ntop4_raw;
  var _emscripten_glUniformMatrix2fv=env._emscripten_glUniformMatrix2fv;
  var _emscripten_glDrawArraysInstanced=env._emscripten_glDrawArraysInstanced;
  var _alcMakeContextCurrent=env._alcMakeContextCurrent;
  var _emscripten_glVertexAttrib2fv=env._emscripten_glVertexAttrib2fv;
  var _emscripten_glBufferData=env._emscripten_glBufferData;
  var _alSourcef=env._alSourcef;
  var _emscripten_glFlush=env._emscripten_glFlush;
  var _alSourcei=env._alSourcei;
  var _alGenBuffers=env._alGenBuffers;
  var _nanosleep=env._nanosleep;
  var _pthread_once=env._pthread_once;
  var _eglWaitClient=env._eglWaitClient;
  var _emscripten_glTexCoordPointer=env._emscripten_glTexCoordPointer;
  var _emscripten_set_blur_callback=env._emscripten_set_blur_callback;
  var _emscripten_glLoadMatrixf=env._emscripten_glLoadMatrixf;
  var _emscripten_glStencilFuncSeparate=env._emscripten_glStencilFuncSeparate;
  var _emscripten_glVertexAttrib3f=env._emscripten_glVertexAttrib3f;
  var _pthread_mutex_lock=env._pthread_mutex_lock;
  var ___cxa_guard_abort=env.___cxa_guard_abort;
  var _dlerror=env._dlerror;
  var _emscripten_get_gamepad_status=env._emscripten_get_gamepad_status;
  var ___syscall20=env.___syscall20;
  var _emscripten_glUniform1iv=env._emscripten_glUniform1iv;
  var _alListenerfv=env._alListenerfv;
  var emscriptenWebGLGetUniform=env.emscriptenWebGLGetUniform;
  var _emscripten_glGetBufferParameteriv=env._emscripten_glGetBufferParameteriv;
  var _emscripten_glVertexAttrib4fv=env._emscripten_glVertexAttrib4fv;
  var _pthread_getspecific=env._pthread_getspecific;
  var _alListenerf=env._alListenerf;
  var _emscripten_glDepthRange=env._emscripten_glDepthRange;
  var _emscripten_asm_const_iii=env._emscripten_asm_const_iii;
  var _eglMakeCurrent=env._eglMakeCurrent;
  var _emscripten_glCopyTexImage2D=env._emscripten_glCopyTexImage2D;
  var _emscripten_glFramebufferTexture2D=env._emscripten_glFramebufferTexture2D;
  var _emscripten_glStencilFunc=env._emscripten_glStencilFunc;
  var ___syscall183=env.___syscall183;
  var _alDistanceModel=env._alDistanceModel;
  var _localtime=env._localtime;
  var _emscripten_glRenderbufferStorage=env._emscripten_glRenderbufferStorage;
  var _emscripten_set_keydown_callback=env._emscripten_set_keydown_callback;
  var _emscripten_glVertexPointer=env._emscripten_glVertexPointer;
  var __read_sockaddr=env.__read_sockaddr;
  var _eglInitialize=env._eglInitialize;
  var _emscripten_glBufferSubData=env._emscripten_glBufferSubData;
  var _emscripten_glGetUniformfv=env._emscripten_glGetUniformfv;
  var _emscripten_glStencilOp=env._emscripten_glStencilOp;
  var _emscripten_glBlendEquation=env._emscripten_glBlendEquation;
  var _pthread_self=env._pthread_self;
  var _emscripten_glVertexAttrib1fv=env._emscripten_glVertexAttrib1fv;
  var _dlclose=env._dlclose;
  var _emscripten_glGetProgramInfoLog=env._emscripten_glGetProgramInfoLog;
  var _getpwuid=env._getpwuid;
  var ___cxa_throw=env.___cxa_throw;
  var _emscripten_glUniform2fv=env._emscripten_glUniform2fv;
  var _emscripten_glBindBuffer=env._emscripten_glBindBuffer;
  var _emscripten_glGetFloatv=env._emscripten_glGetFloatv;
  var _pthread_mutex_init=env._pthread_mutex_init;
  var _emscripten_glGenRenderbuffers=env._emscripten_glGenRenderbuffers;
  var _eglGetDisplay=env._eglGetDisplay;
  var ___syscall63=env.___syscall63;
  var _emscripten_glCullFace=env._emscripten_glCullFace;
  var _llvm_fabs_f64=env._llvm_fabs_f64;
  var _emscripten_glStencilMaskSeparate=env._emscripten_glStencilMaskSeparate;
  var _emscripten_glUniform3fv=env._emscripten_glUniform3fv;
  var _emscripten_asm_const_ii=env._emscripten_asm_const_ii;
  var _alSource3f=env._alSource3f;
  var _emscripten_set_keypress_callback=env._emscripten_set_keypress_callback;
  var _emscripten_glDisableVertexAttribArray=env._emscripten_glDisableVertexAttribArray;
  var _eglBindAPI=env._eglBindAPI;
  var _eglCreateContext=env._eglCreateContext;
  var _clock=env._clock;
  var _emscripten_glGetBooleanv=env._emscripten_glGetBooleanv;
  var _emscripten_glVertexAttribDivisor=env._emscripten_glVertexAttribDivisor;
  var _emscripten_glGenBuffers=env._emscripten_glGenBuffers;
  var _emscripten_glDeleteObjectARB=env._emscripten_glDeleteObjectARB;
  var _emscripten_glGetShaderPrecisionFormat=env._emscripten_glGetShaderPrecisionFormat;
  var __write_sockaddr=env.__write_sockaddr;
  var _emscripten_request_fullscreen_strategy=env._emscripten_request_fullscreen_strategy;
  var _emscripten_glIsEnabled=env._emscripten_glIsEnabled;
  var _alGetListenerf=env._alGetListenerf;
  var _emscripten_glStencilOpSeparate=env._emscripten_glStencilOpSeparate;
  var _pthread_cleanup_pop=env._pthread_cleanup_pop;
  var _emscripten_asm_const_iiiii=env._emscripten_asm_const_iiiii;
  var ___syscall122=env.___syscall122;
  var ___cxa_free_exception=env.___cxa_free_exception;
  var ___cxa_find_matching_catch=env.___cxa_find_matching_catch;
  var _emscripten_glShaderBinary=env._emscripten_glShaderBinary;
  var ___cxa_guard_release=env.___cxa_guard_release;
  var _emscripten_glValidateProgram=env._emscripten_glValidateProgram;
  var _emscripten_glUniform4iv=env._emscripten_glUniform4iv;
  var ___setErrNo=env.___setErrNo;
  var _eglSwapBuffers=env._eglSwapBuffers;
  var _emscripten_glVertexAttrib2f=env._emscripten_glVertexAttrib2f;
  var ___resumeException=env.___resumeException;
  var _mktime=env._mktime;
  var _emscripten_glGetError=env._emscripten_glGetError;
  var _gethostbyaddr=env._gethostbyaddr;
  var _emscripten_glReadPixels=env._emscripten_glReadPixels;
  var _eglCreateWindowSurface=env._eglCreateWindowSurface;
  var _emscripten_glClearStencil=env._emscripten_glClearStencil;
  var emscriptenWebGLGet=env.emscriptenWebGLGet;
  var _emscripten_get_device_pixel_ratio=env._emscripten_get_device_pixel_ratio;
  var _emscripten_set_mouseup_callback=env._emscripten_set_mouseup_callback;
  var _emscripten_glFinish=env._emscripten_glFinish;
  var _emscripten_glClearDepth=env._emscripten_glClearDepth;
  var _emscripten_glUniform1fv=env._emscripten_glUniform1fv;
  var _emscripten_set_resize_callback=env._emscripten_set_resize_callback;
  var _emscripten_glUniform4i=env._emscripten_glUniform4i;
  var _alDopplerFactor=env._alDopplerFactor;
  var _llvm_pow_f64=env._llvm_pow_f64;
  var _emscripten_glUniform4f=env._emscripten_glUniform4f;
  var _emscripten_glBlendFunc=env._emscripten_glBlendFunc;
  var _emscripten_glStencilMask=env._emscripten_glStencilMask;
  var _alcGetIntegerv=env._alcGetIntegerv;
  var _localtime_r=env._localtime_r;
  var _clock_gettime=env._clock_gettime;
  var _strftime=env._strftime;
  var _alcDestroyContext=env._alcDestroyContext;
  var _emscripten_glGetVertexAttribiv=env._emscripten_glGetVertexAttribiv;
  var _emscripten_glUniformMatrix3fv=env._emscripten_glUniformMatrix3fv;
  var ___syscall33=env.___syscall33;
  var _pthread_key_create=env._pthread_key_create;
  var _emscripten_glDeleteFramebuffers=env._emscripten_glDeleteFramebuffers;
  var __setLetterbox=env.__setLetterbox;
  var ___syscall39=env.___syscall39;
  var ___syscall38=env.___syscall38;
  var _emscripten_glUniform4fv=env._emscripten_glUniform4fv;
  var _emscripten_glGetObjectParameterivARB=env._emscripten_glGetObjectParameterivARB;
  var _emscripten_glGetUniformiv=env._emscripten_glGetUniformiv;
  var _eglDestroySurface=env._eglDestroySurface;
  var _sigaction=env._sigaction;
  var _difftime=env._difftime;
  var _emscripten_set_mousemove_callback=env._emscripten_set_mousemove_callback;
  var ___cxa_find_matching_catch_2=env.___cxa_find_matching_catch_2;
  var ___cxa_find_matching_catch_3=env.___cxa_find_matching_catch_3;
  var _emscripten_glDeleteTextures=env._emscripten_glDeleteTextures;
  var _eglDestroyContext=env._eglDestroyContext;
  var _emscripten_exit_fullscreen=env._emscripten_exit_fullscreen;
  var _strftime_l=env._strftime_l;
  var _llvm_eh_typeid_for=env._llvm_eh_typeid_for;
  var __exit=env.__exit;
  var _alGetListenerfv=env._alGetListenerfv;
  var _emscripten_glColorMask=env._emscripten_glColorMask;
  var _pthread_mutex_unlock=env._pthread_mutex_unlock;
  var _emscripten_glBindTexture=env._emscripten_glBindTexture;
  var _emscripten_set_main_loop=env._emscripten_set_main_loop;
  var _alGenSources=env._alGenSources;
  var _emscripten_glIsShader=env._emscripten_glIsShader;
  var _emscripten_asm_const_iiii=env._emscripten_asm_const_iiii;
  var _alcOpenDevice=env._alcOpenDevice;
  var _emscripten_glCompressedTexImage2D=env._emscripten_glCompressedTexImage2D;
  var _emscripten_glGetInfoLogARB=env._emscripten_glGetInfoLogARB;
  var _emscripten_longjmp=env._emscripten_longjmp;
  var _atexit=env._atexit;
  var _emscripten_glDeleteVertexArrays=env._emscripten_glDeleteVertexArrays;
  var _emscripten_glReleaseShaderCompiler=env._emscripten_glReleaseShaderCompiler;
  var _dlsym=env._dlsym;
  var ___cxa_guard_acquire=env.___cxa_guard_acquire;
  var _emscripten_glFrontFace=env._emscripten_glFrontFace;
  var __ZSt18uncaught_exceptionv=env.__ZSt18uncaught_exceptionv;
  var _emscripten_glUseProgram=env._emscripten_glUseProgram;
  var __addDays=env.__addDays;
  var _emscripten_set_touchmove_callback=env._emscripten_set_touchmove_callback;
  var _emscripten_glUniform3iv=env._emscripten_glUniform3iv;
  var _emscripten_set_main_loop_arg=env._emscripten_set_main_loop_arg;
  var _sysconf=env._sysconf;
  var _emscripten_glLineWidth=env._emscripten_glLineWidth;
  var ___cxa_atexit=env.___cxa_atexit;
  var _emscripten_glScissor=env._emscripten_glScissor;
  var _pthread_cleanup_push=env._pthread_cleanup_push;
  var _emscripten_set_element_css_size=env._emscripten_set_element_css_size;
  var _alSourcePlay=env._alSourcePlay;
  var _emscripten_glIsBuffer=env._emscripten_glIsBuffer;
  var _emscripten_glVertexAttrib1f=env._emscripten_glVertexAttrib1f;
  var _emscripten_set_keyup_callback=env._emscripten_set_keyup_callback;
  var _emscripten_glCompressedTexSubImage2D=env._emscripten_glCompressedTexSubImage2D;
  var _emscripten_glGetAttachedShaders=env._emscripten_glGetAttachedShaders;
  var _emscripten_glGenTextures=env._emscripten_glGenTextures;
  var _alcCaptureSamples=env._alcCaptureSamples;
  var _eglGetConfigAttrib=env._eglGetConfigAttrib;
  var _alGetSourcei=env._alGetSourcei;
  var _alGetSourcef=env._alGetSourcef;
  var _gmtime=env._gmtime;
  var _alSourceQueueBuffers=env._alSourceQueueBuffers;
  var _emscripten_glGetTexParameteriv=env._emscripten_glGetTexParameteriv;
  var ___syscall330=env.___syscall330;
  var _emscripten_set_mousedown_callback=env._emscripten_set_mousedown_callback;
  var _emscripten_glClientActiveTexture=env._emscripten_glClientActiveTexture;
  var _emscripten_glCheckFramebufferStatus=env._emscripten_glCheckFramebufferStatus;
  var _eglWaitGL=env._eglWaitGL;
  var ___syscall118=env.___syscall118;
  var _emscripten_glUniform3f=env._emscripten_glUniform3f;
  var _emscripten_glUniform3i=env._emscripten_glUniform3i;
  var _emscripten_glDeleteShader=env._emscripten_glDeleteShader;
  var _alDeleteSources=env._alDeleteSources;
  var _emscripten_glGetUniformLocation=env._emscripten_glGetUniformLocation;
  var _emscripten_glEnableVertexAttribArray=env._emscripten_glEnableVertexAttribArray;
  var _alGetError=env._alGetError;
  var _emscripten_get_now=env._emscripten_get_now;
  var __registerRestoreOldStyle=env.__registerRestoreOldStyle;
  var emscriptenWebGLGetTexPixelData=env.emscriptenWebGLGetTexPixelData;
  var _gettimeofday=env._gettimeofday;
  var _eglWaitNative=env._eglWaitNative;
  var ___syscall202=env.___syscall202;
  var _emscripten_glEnableClientState=env._emscripten_glEnableClientState;
  var _eglChooseConfig=env._eglChooseConfig;
  var ___cxa_allocate_exception=env.___cxa_allocate_exception;
  var _emscripten_get_num_gamepads=env._emscripten_get_num_gamepads;
  var ___buildEnvironment=env.___buildEnvironment;
  var _tzset=env._tzset;
  var _emscripten_glGetAttribLocation=env._emscripten_glGetAttribLocation;
  var _emscripten_glDisable=env._emscripten_glDisable;
  var ___cxa_end_catch=env.___cxa_end_catch;
  var _emscripten_glDeleteRenderbuffers=env._emscripten_glDeleteRenderbuffers;
  var _emscripten_glDrawElementsInstanced=env._emscripten_glDrawElementsInstanced;
  var _emscripten_glVertexAttrib4f=env._emscripten_glVertexAttrib4f;
  var _emscripten_glPixelStorei=env._emscripten_glPixelStorei;
  var _llvm_fabs_f32=env._llvm_fabs_f32;
  var _getenv=env._getenv;
  var _emscripten_set_gamepaddisconnected_callback=env._emscripten_set_gamepaddisconnected_callback;
  var _llvm_cttz_i64=env._llvm_cttz_i64;
  var _gethostbyname=env._gethostbyname;
  var _alcCloseDevice=env._alcCloseDevice;
  var _emscripten_glFramebufferRenderbuffer=env._emscripten_glFramebufferRenderbuffer;
  var _alcCaptureStart=env._alcCaptureStart;
  var _llvm_ctpop_i32=env._llvm_ctpop_i32;
  var _emscripten_glRotatef=env._emscripten_glRotatef;
  var _emscripten_glGetShaderiv=env._emscripten_glGetShaderiv;
  var _emscripten_get_battery_status=env._emscripten_get_battery_status;
  var ___cxa_pure_virtual=env.___cxa_pure_virtual;
  var _emscripten_glUniformMatrix4fv=env._emscripten_glUniformMatrix4fv;
  var _emscripten_glGetPointerv=env._emscripten_glGetPointerv;
  var _pthread_cond_wait=env._pthread_cond_wait;
  var _emscripten_set_touchstart_callback=env._emscripten_set_touchstart_callback;
  var _emscripten_glIsRenderbuffer=env._emscripten_glIsRenderbuffer;
  var _emscripten_request_pointerlock=env._emscripten_request_pointerlock;
  var ___syscall40=env.___syscall40;
  var _llvm_bswap_i64=env._llvm_bswap_i64;
  var _emscripten_set_touchcancel_callback=env._emscripten_set_touchcancel_callback;
  var __inet_pton6_raw=env.__inet_pton6_raw;
  var _emscripten_set_focus_callback=env._emscripten_set_focus_callback;
  var _emscripten_glGetVertexAttribfv=env._emscripten_glGetVertexAttribfv;
  var _emscripten_glVertexAttrib3fv=env._emscripten_glVertexAttrib3fv;
  var _emscripten_glCompileShader=env._emscripten_glCompileShader;
  var _glClear=env._glClear;
  var __arraySum=env.__arraySum;
  var _emscripten_glLinkProgram=env._emscripten_glLinkProgram;
  var _alDeleteBuffers=env._alDeleteBuffers;
  var _alGetSourcefv=env._alGetSourcefv;
  var _emscripten_get_pointerlock_status=env._emscripten_get_pointerlock_status;
  var _emscripten_glDrawRangeElements=env._emscripten_glDrawRangeElements;
  var ___unlock=env.___unlock;
  var _pthread_setspecific=env._pthread_setspecific;
  var _emscripten_glClearColor=env._emscripten_glClearColor;
  var _emscripten_glCreateProgram=env._emscripten_glCreateProgram;
  var _emscripten_glDetachShader=env._emscripten_glDetachShader;
  var _setenv=env._setenv;
  var _emscripten_do_request_fullscreen=env._emscripten_do_request_fullscreen;
  var _emscripten_set_mouseleave_callback=env._emscripten_set_mouseleave_callback;
  var _emscripten_get_element_css_size=env._emscripten_get_element_css_size;
  var _emscripten_set_fullscreenchange_callback=env._emscripten_set_fullscreenchange_callback;
  var _emscripten_glVertexAttribPointer=env._emscripten_glVertexAttribPointer;
  var _alBufferData=env._alBufferData;
  var _emscripten_glDrawArrays=env._emscripten_glDrawArrays;
  var _emscripten_glPolygonOffset=env._emscripten_glPolygonOffset;
  var _longjmp=env._longjmp;
  var _emscripten_glBlendColor=env._emscripten_glBlendColor;
  var _signal=env._signal;
  var _emscripten_set_main_loop_timing=env._emscripten_set_main_loop_timing;
  var _sbrk=env._sbrk;
  var ___cxa_begin_catch=env.___cxa_begin_catch;
  var _emscripten_glGetProgramiv=env._emscripten_glGetProgramiv;
  var _emscripten_glGetShaderSource=env._emscripten_glGetShaderSource;
  var _emscripten_glTexImage2D=env._emscripten_glTexImage2D;
  var __isLeapYear=env.__isLeapYear;
  var _emscripten_glBlendEquationSeparate=env._emscripten_glBlendEquationSeparate;
  var _emscripten_glGetString=env._emscripten_glGetString;
  var _emscripten_glIsFramebuffer=env._emscripten_glIsFramebuffer;
  var _emscripten_glBindProgramARB=env._emscripten_glBindProgramARB;
  var _glutCreateWindow=env._glutCreateWindow;
  var ___syscall85=env.___syscall85;
  var _alcGetError=env._alcGetError;
  var _emscripten_glUniform2i=env._emscripten_glUniform2i;
  var _emscripten_glUniform2f=env._emscripten_glUniform2f;
  var _alSourcefv=env._alSourcefv;
  var _emscripten_glTexParameterf=env._emscripten_glTexParameterf;
  var _emscripten_glTexParameteri=env._emscripten_glTexParameteri;
  var _glutInitDisplayMode=env._glutInitDisplayMode;
  var _emscripten_glGenVertexArrays=env._emscripten_glGenVertexArrays;
  var _emscripten_set_visibilitychange_callback=env._emscripten_set_visibilitychange_callback;
  var _eglGetProcAddress=env._eglGetProcAddress;
  var _emscripten_glBindAttribLocation=env._emscripten_glBindAttribLocation;
  var _llvm_pow_f32=env._llvm_pow_f32;
  var _emscripten_glDrawElements=env._emscripten_glDrawElements;
  var _emscripten_set_canvas_size=env._emscripten_set_canvas_size;
  var _emscripten_glTexSubImage2D=env._emscripten_glTexSubImage2D;
  var _emscripten_asm_const_v=env._emscripten_asm_const_v;
  var _emscripten_glClearDepthf=env._emscripten_glClearDepthf;
  var _emscripten_set_mouseenter_callback=env._emscripten_set_mouseenter_callback;
  var _emscripten_glMatrixMode=env._emscripten_glMatrixMode;
  var ___syscall10=env.___syscall10;
  var _emscripten_glNormalPointer=env._emscripten_glNormalPointer;
  var _emscripten_glHint=env._emscripten_glHint;
  var _emscripten_glEnable=env._emscripten_glEnable;
  var ___syscall3=env.___syscall3;
  var ___lock=env.___lock;
  var _emscripten_glBindFramebuffer=env._emscripten_glBindFramebuffer;
  var ___syscall6=env.___syscall6;
  var ___syscall5=env.___syscall5;
  var ___syscall4=env.___syscall4;
  var _emscripten_glBindRenderbuffer=env._emscripten_glBindRenderbuffer;
  var _time=env._time;
  var _emscripten_glGetFramebufferAttachmentParameteriv=env._emscripten_glGetFramebufferAttachmentParameteriv;
  var _exit=env._exit;
  var _emscripten_glGetActiveAttrib=env._emscripten_glGetActiveAttrib;
  var __inet_pton4_raw=env.__inet_pton4_raw;
  var ___syscall102=env.___syscall102;
  var ___syscall265=env.___syscall265;
  var _emscripten_glClear=env._emscripten_glClear;
  var _emscripten_glGetShaderInfoLog=env._emscripten_glGetShaderInfoLog;
  var _emscripten_glGetVertexAttribPointerv=env._emscripten_glGetVertexAttribPointerv;
  var _llvm_sqrt_f64=env._llvm_sqrt_f64;
  var _alSourceRewind=env._alSourceRewind;
  var _emscripten_glGetActiveUniform=env._emscripten_glGetActiveUniform;
  var emscriptenWebGLGetVertexAttrib=env.emscriptenWebGLGetVertexAttrib;
  var ___syscall197=env.___syscall197;
  var ___syscall196=env.___syscall196;
  var ___syscall195=env.___syscall195;
  var _eglSwapInterval=env._eglSwapInterval;
  var ___syscall199=env.___syscall199;
  var _emscripten_glDeleteProgram=env._emscripten_glDeleteProgram;
  var _glutDestroyWindow=env._glutDestroyWindow;
  var _emscripten_glCreateShader=env._emscripten_glCreateShader;
  var _pthread_mutex_destroy=env._pthread_mutex_destroy;
  var _emscripten_glColorPointer=env._emscripten_glColorPointer;
  var _emscripten_glViewport=env._emscripten_glViewport;
  var _pthread_cond_broadcast=env._pthread_cond_broadcast;
  var _emscripten_glDepthMask=env._emscripten_glDepthMask;
  var _alcCaptureStop=env._alcCaptureStop;
  var _emscripten_glDrawBuffers=env._emscripten_glDrawBuffers;
  var _alSourceStop=env._alSourceStop;
  var _emscripten_exit_pointerlock=env._emscripten_exit_pointerlock;
  var _emscripten_set_gamepadconnected_callback=env._emscripten_set_gamepadconnected_callback;
  var _abort=env._abort;
  var _emscripten_glGenFramebuffers=env._emscripten_glGenFramebuffers;
  var _emscripten_glLoadIdentity=env._emscripten_glLoadIdentity;
  var _emscripten_glShaderSource=env._emscripten_glShaderSource;
  var ___gxx_personality_v0=env.___gxx_personality_v0;
  var _system=env._system;
  var _usleep=env._usleep;
  var _emscripten_set_touchend_callback=env._emscripten_set_touchend_callback;
  var _emscripten_glGetRenderbufferParameteriv=env._emscripten_glGetRenderbufferParameteriv;
  var _eglTerminate=env._eglTerminate;
  var _emscripten_glSampleCoverage=env._emscripten_glSampleCoverage;
  var _emscripten_glFrustum=env._emscripten_glFrustum;
  var _emscripten_glDepthRangef=env._emscripten_glDepthRangef;
  var _emscripten_glGenerateMipmap=env._emscripten_glGenerateMipmap;
  var _emscripten_glIsTexture=env._emscripten_glIsTexture;
  var _emscripten_glBindVertexArray=env._emscripten_glBindVertexArray;
  var _emscripten_glActiveTexture=env._emscripten_glActiveTexture;
  var _emscripten_set_wheel_callback=env._emscripten_set_wheel_callback;
  var _emscripten_glDeleteBuffers=env._emscripten_glDeleteBuffers;
  var ___syscall54=env.___syscall54;
  var _emscripten_glUniform2iv=env._emscripten_glUniform2iv;
  var _emscripten_asm_const_i=env._emscripten_asm_const_i;
  var _emscripten_glCopyTexSubImage2D=env._emscripten_glCopyTexSubImage2D;
  var _alSourceUnqueueBuffers=env._alSourceUnqueueBuffers;
  var tempFloat = 0.0;

function _emscripten_replace_memory(newBuffer) {
  if ((byteLength(newBuffer) & 0xffffff || byteLength(newBuffer) <= 0xffffff) || byteLength(newBuffer) > 0x80000000) return false;
  HEAP8 = new Int8View(newBuffer);
  HEAP16 = new Int16View(newBuffer);
  HEAP32 = new Int32View(newBuffer);
  HEAPU8 = new Uint8View(newBuffer);
  HEAPU16 = new Uint16View(newBuffer);
  HEAPU32 = new Uint32View(newBuffer);
  HEAPF32 = new Float32View(newBuffer);
  HEAPF64 = new Float64View(newBuffer);
  buffer = newBuffer;
  return true;
}

// EMSCRIPTEN_START_FUNCS

function stackAlloc(size) {
  size = size|0;
  var ret = 0;
  ret = STACKTOP;
  STACKTOP = (STACKTOP + size)|0;
  STACKTOP = (STACKTOP + 15)&-16;
if ((STACKTOP|0) >= (STACK_MAX|0)) abort();

  return ret|0;
}
function stackSave() {
  return STACKTOP|0;
}
function stackRestore(top) {
  top = top|0;
  STACKTOP = top;
}
function establishStackSpace(stackBase, stackMax) {
  stackBase = stackBase|0;
  stackMax = stackMax|0;
  STACKTOP = stackBase;
  STACK_MAX = stackMax;
}

function setThrew(threw, value) {
  threw = threw|0;
  value = value|0;
  if ((__THREW__|0) == 0) {
    __THREW__ = threw;
    threwValue = value;
  }
}
function copyTempFloat(ptr) {
  ptr = ptr|0;
  HEAP8[tempDoublePtr>>0] = HEAP8[ptr>>0];
  HEAP8[tempDoublePtr+1>>0] = HEAP8[ptr+1>>0];
  HEAP8[tempDoublePtr+2>>0] = HEAP8[ptr+2>>0];
  HEAP8[tempDoublePtr+3>>0] = HEAP8[ptr+3>>0];
}
function copyTempDouble(ptr) {
  ptr = ptr|0;
  HEAP8[tempDoublePtr>>0] = HEAP8[ptr>>0];
  HEAP8[tempDoublePtr+1>>0] = HEAP8[ptr+1>>0];
  HEAP8[tempDoublePtr+2>>0] = HEAP8[ptr+2>>0];
  HEAP8[tempDoublePtr+3>>0] = HEAP8[ptr+3>>0];
  HEAP8[tempDoublePtr+4>>0] = HEAP8[ptr+4>>0];
  HEAP8[tempDoublePtr+5>>0] = HEAP8[ptr+5>>0];
  HEAP8[tempDoublePtr+6>>0] = HEAP8[ptr+6>>0];
  HEAP8[tempDoublePtr+7>>0] = HEAP8[ptr+7>>0];
}

function setTempRet0(value) {
  value = value|0;
  tempRet0 = value;
}
function getTempRet0() {
  return tempRet0|0;
}

function _main($argc,$argv) {
 $argc = $argc|0;
 $argv = $argv|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, $loopArg = 0, $retval$0 = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $loopArg = sp + 16|0;
 $0 = (_love_version()|0); //@line 209 "love.cpp"
 $1 = (_strcmp(180253,$0)|0); //@line 209 "love.cpp"
 $2 = ($1|0)==(0); //@line 209 "love.cpp"
 if (!($2)) {
  $3 = (_love_version()|0); //@line 212 "love.cpp"
  HEAP32[$vararg_buffer>>2] = 180253; //@line 211 "love.cpp"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 211 "love.cpp"
  HEAP32[$vararg_ptr1>>2] = $3; //@line 211 "love.cpp"
  (_printf(179968,$vararg_buffer)|0); //@line 211 "love.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 326 "love.cpp"
 }
 $4 = ($argc|0)>(1); //@line 217 "love.cpp"
 if ($4) {
  $5 = ((($argv)) + 4|0); //@line 217 "love.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 217 "love.cpp"
  $7 = (_strcmp($6,180049)|0); //@line 217 "love.cpp"
  $8 = ($7|0)==(0); //@line 217 "love.cpp"
  if ($8) {
   $9 = (_love_version()|0); //@line 219 "love.cpp"
   $10 = (_love_codename()|0); //@line 219 "love.cpp"
   HEAP32[$vararg_buffer2>>2] = $9; //@line 219 "love.cpp"
   $vararg_ptr5 = ((($vararg_buffer2)) + 4|0); //@line 219 "love.cpp"
   HEAP32[$vararg_ptr5>>2] = $10; //@line 219 "love.cpp"
   (_printf(180059,$vararg_buffer2)|0); //@line 219 "love.cpp"
   $$0 = 0;
   STACKTOP = sp;return ($$0|0); //@line 326 "love.cpp"
  }
 }
 $11 = (_luaL_newstate()|0); //@line 224 "love.cpp"
 _luaL_openlibs($11); //@line 225 "love.cpp"
 _lua_getfield($11,-10002,523111); //@line 143 "love.cpp"
 _lua_getfield($11,-1,523223); //@line 144 "love.cpp"
 _lua_pushcclosure($11,3346,0); //@line 145 "love.cpp"
 _lua_setfield($11,-2,445428); //@line 146 "love.cpp"
 _lua_settop($11,-3); //@line 147 "love.cpp"
 _lua_createtable($11,0,0); //@line 237 "love.cpp"
 $12 = ($argc|0)>(0); //@line 239 "love.cpp"
 if ($12) {
  $13 = HEAP32[$argv>>2]|0; //@line 241 "love.cpp"
  _lua_pushstring($11,$13); //@line 241 "love.cpp"
  _lua_rawseti($11,-2,-2); //@line 242 "love.cpp"
 }
 _lua_pushstring($11,180073); //@line 245 "love.cpp"
 _lua_rawseti($11,-2,-1); //@line 246 "love.cpp"
 if ($4) {
  $i$01 = 1;
  while(1) {
   $18 = (($argv) + ($i$01<<2)|0); //@line 250 "love.cpp"
   $19 = HEAP32[$18>>2]|0; //@line 250 "love.cpp"
   _lua_pushstring($11,$19); //@line 250 "love.cpp"
   _lua_rawseti($11,-2,$i$01); //@line 251 "love.cpp"
   $20 = (($i$01) + 1)|0; //@line 248 "love.cpp"
   $exitcond = ($20|0)==($argc|0); //@line 248 "love.cpp"
   if ($exitcond) {
    break;
   } else {
    $i$01 = $20;
   }
  }
 }
 _lua_setfield($11,-10002,518559); //@line 254 "love.cpp"
 _lua_getfield($11,-10002,523817); //@line 258 "love.cpp"
 _lua_pushstring($11,445428); //@line 259 "love.cpp"
 _lua_call($11,1,1); //@line 260 "love.cpp"
 _lua_pushboolean($11,1); //@line 266 "love.cpp"
 _lua_setfield($11,-2,180091); //@line 267 "love.cpp"
 _lua_settop($11,-2); //@line 271 "love.cpp"
 _lua_getfield($11,-10002,523817); //@line 274 "love.cpp"
 _lua_pushstring($11,236662); //@line 275 "love.cpp"
 _lua_call($11,1,1); //@line 276 "love.cpp"
 _lua_pushboolean($11,0); //@line 280 "love.cpp"
 _lua_call($11,1,1); //@line 281 "love.cpp"
 $14 = (_luaL_ref($11,-10000)|0); //@line 283 "love.cpp"
 HEAP32[$loopArg>>2] = $11; //@line 284 "love.cpp"
 $15 = ((($loopArg)) + 4|0); //@line 284 "love.cpp"
 HEAP32[$15>>2] = $14; //@line 284 "love.cpp"
 _emscripten_set_main_loop_arg((3347|0),($loopArg|0),0,1); //@line 286 "love.cpp"
 $16 = (_lua_isnumber($11,-1)|0); //@line 303 "love.cpp"
 $17 = ($16|0)==(0); //@line 303 "love.cpp"
 if ($17) {
  $retval$0 = 0;
 } else {
  $21 = (+_lua_tonumber($11,-1)); //@line 304 "love.cpp"
  $22 = (~~(($21))); //@line 304 "love.cpp"
  $retval$0 = $22;
 }
 _lua_close($11); //@line 306 "love.cpp"
 $$0 = $retval$0;
 STACKTOP = sp;return ($$0|0); //@line 326 "love.cpp"
}
function __ZZ4mainEN3__08__invokeEPv($arg) {
 $arg = $arg|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
 $1 = ($0|0)==(0|0); //@line 288 "love.cpp"
 if (!($1)) {
  $2 = HEAP32[$0>>2]|0; //@line 289 "love.cpp"
  $3 = ((($2)) + 40|0); //@line 289 "love.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 289 "love.cpp"
  FUNCTION_TABLE_vi[$4 & 8191]($0); //@line 289 "love.cpp"
 }
 $5 = HEAP32[$arg>>2]|0; //@line 292 "love.cpp"
 $6 = ((($arg)) + 4|0); //@line 293 "love.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 293 "love.cpp"
 _lua_rawgeti($5,-10000,$7); //@line 293 "love.cpp"
 $8 = (_lua_pcall($5,0,0,0)|0); //@line 294 "love.cpp"
 $9 = ($8|0)==(0); //@line 294 "love.cpp"
 if ($9) {
  STACKTOP = sp;return; //@line 286 "love.cpp"
 }
 $10 = (_lua_tolstring($5,-1,0)|0); //@line 295 "love.cpp"
 HEAP32[$vararg_buffer>>2] = $10; //@line 295 "love.cpp"
 (_printf(180096,$vararg_buffer)|0); //@line 295 "love.cpp"
 STACKTOP = sp;return; //@line 286 "love.cpp"
}
function __ZN4love6ModuleD2Ev($this) {
 $this = $this|0;
 var $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i$i = 0, $$0$i$i$i$i1 = 0, $$01$i$i$i = 0, $$01$i$i$i$i = 0, $$lcssa = 0, $$lcssa55 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $it$sroa$0$011 = 0;
 var $it$sroa$0$011$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (9240); //@line 64 "common/Module.cpp"
 $0 = HEAP32[156420]|0; //@line 41 "common/Module.cpp"
 $1 = ($0|0)==(0|0); //@line 41 "common/Module.cpp"
 do {
  if ($1) {
   __THREW__ = 0;
   $2 = (invoke_ii(3348,12)|0); //@line 42 "common/Module.cpp"
   $3 = __THREW__; __THREW__ = 0;
   $4 = $3&1;
   if ($4) {
    $12 = ___cxa_find_matching_catch_3(0|0)|0;
    $13 = tempRet0;
    ___clang_call_terminate($12); //@line 85 "common/Module.cpp"
    // unreachable; //@line 85 "common/Module.cpp"
   } else {
    $5 = ((($2)) + 4|0); //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP32[$5>>2] = 0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $6 = ((($2)) + 8|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$6>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$2>>2] = $5; //@line 1130 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP32[156420] = $2; //@line 42 "common/Module.cpp"
    $8 = $2;
    break;
   }
  } else {
   $8 = $0;
  }
 } while(0);
 $7 = HEAP32[$8>>2]|0; //@line 911 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $9 = ((($8)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $10 = $7; //@line 672 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $11 = ($10|0)==($9|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 L7: do {
  if (!($11)) {
   $15 = $10;$it$sroa$0$011 = $7;
   while(1) {
    $14 = ((($15)) + 28|0); //@line 70 "common/Module.cpp"
    $16 = HEAP32[$14>>2]|0; //@line 70 "common/Module.cpp"
    $17 = ($16|0)==($this|0); //@line 70 "common/Module.cpp"
    $18 = $it$sroa$0$011; //@line 656 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($17) {
     $$lcssa = $15;$$lcssa55 = $18;$it$sroa$0$011$lcssa = $it$sroa$0$011;
     break;
    }
    $39 = ((($18)) + 4|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $40 = HEAP32[$39>>2]|0; //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $41 = ($40|0)==(0|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($41) {
     $$0$i$i$i = $18;
     while(1) {
      $44 = ((($$0$i$i$i)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $45 = HEAP32[$44>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $46 = HEAP32[$45>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $47 = ($46|0)==($$0$i$i$i|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      if ($47) {
       $$01$i$i$i = $45;
       break;
      } else {
       $$0$i$i$i = $45;
      }
     }
    } else {
     $$0$i$i$i$i = $40;
     while(1) {
      $42 = HEAP32[$$0$i$i$i$i>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $43 = ($42|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      if ($43) {
       $$01$i$i$i = $$0$i$i$i$i;
       break;
      } else {
       $$0$i$i$i$i = $42;
      }
     }
    }
    $48 = $$01$i$i$i; //@line 656 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $49 = ($$01$i$i$i|0)==($9|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($49) {
     break L7;
    } else {
     $15 = $$01$i$i$i;$it$sroa$0$011 = $48;
    }
   }
   $19 = ((($$lcssa)) + 4|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $20 = HEAP32[$19>>2]|0; //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $21 = ($20|0)==(0|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($21) {
    $$0$i$i$i$i1 = $$lcssa55;
    while(1) {
     $24 = ((($$0$i$i$i$i1)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $25 = HEAP32[$24>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $26 = HEAP32[$25>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $27 = ($26|0)==($$0$i$i$i$i1|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     if ($27) {
      $$01$i$i$i$i = $25;
      break;
     } else {
      $$0$i$i$i$i1 = $25;
     }
    }
   } else {
    $$0$i$i$i$i$i = $20;
    while(1) {
     $22 = HEAP32[$$0$i$i$i$i$i>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $23 = ($22|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     if ($23) {
      $$01$i$i$i$i = $$0$i$i$i$i$i;
      break;
     } else {
      $$0$i$i$i$i$i = $22;
     }
    }
   }
   $28 = ($10|0)==($$lcssa|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($28) {
    $29 = $$01$i$i$i$i; //@line 656 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP32[$8>>2] = $29; //@line 1980 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   }
   $30 = ((($8)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $31 = HEAP32[$30>>2]|0; //@line 1981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $32 = (($31) + -1)|0; //@line 1981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$30>>2] = $32; //@line 1981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $33 = ((($8)) + 4|0); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $34 = HEAP32[$33>>2]|0; //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__113__tree_removeIPNS_16__tree_node_baseIPvEEEEvT_S5_($34,$$lcssa55); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $35 = ((($$lcssa)) + 16|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($35); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
   $36 = $it$sroa$0$011$lcssa; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __ZdlPv($36); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
 } while(0);
 $37 = HEAP32[156402]|0; //@line 80 "common/Module.cpp"
 $38 = ($37|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($38) {
  HEAP32[156402] = 0; //@line 81 "common/Module.cpp"
 }
 $56 = HEAP32[(625612)>>2]|0; //@line 80 "common/Module.cpp"
 $57 = ($56|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($57) {
  HEAP32[(625612)>>2] = 0; //@line 81 "common/Module.cpp"
 }
 $58 = HEAP32[(625616)>>2]|0; //@line 80 "common/Module.cpp"
 $59 = ($58|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($59) {
  HEAP32[(625616)>>2] = 0; //@line 81 "common/Module.cpp"
 }
 $60 = HEAP32[(625620)>>2]|0; //@line 80 "common/Module.cpp"
 $61 = ($60|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($61) {
  HEAP32[(625620)>>2] = 0; //@line 81 "common/Module.cpp"
 }
 $62 = HEAP32[(625624)>>2]|0; //@line 80 "common/Module.cpp"
 $63 = ($62|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($63) {
  HEAP32[(625624)>>2] = 0; //@line 81 "common/Module.cpp"
 }
 $64 = HEAP32[(625628)>>2]|0; //@line 80 "common/Module.cpp"
 $65 = ($64|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($65) {
  HEAP32[(625628)>>2] = 0; //@line 81 "common/Module.cpp"
 }
 $66 = HEAP32[(625632)>>2]|0; //@line 80 "common/Module.cpp"
 $67 = ($66|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($67) {
  HEAP32[(625632)>>2] = 0; //@line 81 "common/Module.cpp"
 }
 $68 = HEAP32[(625636)>>2]|0; //@line 80 "common/Module.cpp"
 $69 = ($68|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($69) {
  HEAP32[(625636)>>2] = 0; //@line 81 "common/Module.cpp"
 }
 $70 = HEAP32[(625640)>>2]|0; //@line 80 "common/Module.cpp"
 $71 = ($70|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($71) {
  HEAP32[(625640)>>2] = 0; //@line 81 "common/Module.cpp"
 }
 $72 = HEAP32[(625644)>>2]|0; //@line 80 "common/Module.cpp"
 $73 = ($72|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($73) {
  HEAP32[(625644)>>2] = 0; //@line 81 "common/Module.cpp"
 }
 $74 = HEAP32[(625648)>>2]|0; //@line 80 "common/Module.cpp"
 $75 = ($74|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($75) {
  HEAP32[(625648)>>2] = 0; //@line 81 "common/Module.cpp"
 }
 $76 = HEAP32[(625652)>>2]|0; //@line 80 "common/Module.cpp"
 $77 = ($76|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($77) {
  HEAP32[(625652)>>2] = 0; //@line 81 "common/Module.cpp"
 }
 $78 = HEAP32[(625656)>>2]|0; //@line 80 "common/Module.cpp"
 $79 = ($78|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($79) {
  HEAP32[(625656)>>2] = 0; //@line 81 "common/Module.cpp"
 }
 $80 = HEAP32[(625660)>>2]|0; //@line 80 "common/Module.cpp"
 $81 = ($80|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($81) {
  HEAP32[(625660)>>2] = 0; //@line 81 "common/Module.cpp"
 }
 $82 = HEAP32[(625664)>>2]|0; //@line 80 "common/Module.cpp"
 $83 = ($82|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($83) {
  HEAP32[(625664)>>2] = 0; //@line 81 "common/Module.cpp"
 }
 $84 = HEAP32[(625668)>>2]|0; //@line 80 "common/Module.cpp"
 $85 = ($84|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($85) {
  HEAP32[(625668)>>2] = 0; //@line 81 "common/Module.cpp"
 }
 $86 = HEAP32[(625672)>>2]|0; //@line 80 "common/Module.cpp"
 $87 = ($86|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($87) {
  HEAP32[(625672)>>2] = 0; //@line 81 "common/Module.cpp"
 }
 $88 = HEAP32[(625676)>>2]|0; //@line 80 "common/Module.cpp"
 $89 = ($88|0)==($this|0); //@line 80 "common/Module.cpp"
 if ($89) {
  HEAP32[(625676)>>2] = 0; //@line 81 "common/Module.cpp"
 }
 $51 = HEAP32[156420]|0; //@line 49 "common/Module.cpp"
 $90 = ($51|0)==(0|0); //@line 49 "common/Module.cpp"
 if ($90) {
  return; //@line 85 "common/Module.cpp"
 }
 $50 = ((($51)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $52 = HEAP32[$50>>2]|0; //@line 1000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $53 = ($52|0)==(0); //@line 1000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 if (!($53)) {
  return; //@line 85 "common/Module.cpp"
 }
 $54 = ((($51)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $55 = HEAP32[$54>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ModuleEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE7destroyEPNS_11__tree_nodeISB_PvEE($51,$55); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZdlPv($51); //@line 51 "common/Module.cpp"
 HEAP32[156420] = 0; //@line 52 "common/Module.cpp"
 return; //@line 85 "common/Module.cpp"
}
function ___clang_call_terminate($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 (___cxa_begin_catch(($0|0))|0);
 __ZSt9terminatev();
 // unreachable;
}
function __ZNSt3__113__tree_removeIPNS_16__tree_node_baseIPvEEEEvT_S5_($__root,$__z) {
 $__root = $__root|0;
 $__z = $__z|0;
 var $$$0 = 0, $$0 = 0, $$0$i$i = 0, $$1 = 0, $$2 = 0, $$2$be = 0, $$4 = 0, $$6 = 0, $$cast9 = 0, $$in = 0, $$in2 = 0, $$lcssa = 0, $$lcssa104 = 0, $$lcssa119 = 0, $$lcssa22 = 0, $$ph = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phi29Z2D = 0, $$pre28 = 0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__w$0 = 0, $__w$1 = 0, $__w$1$$2 = 0, $__w$1$$23 = 0, $__w$1$be = 0;
 var $__w$1$be$in = 0, $__w$2 = 0, $__w$2$lcssa = 0, $__w$2100 = 0, $__w$2101 = 0, $__w$3 = 0, $__w$4 = 0, $__w$4111 = 0, $__w$4112 = 0, $__w$5 = 0, $or$cond = 0, $or$cond4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$__z>>2]|0; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $1 = ($0|0)==(0|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 do {
  if ($1) {
   $$ph = $__z;
   label = 5;
  } else {
   $2 = ((($__z)) + 4|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3 = HEAP32[$2>>2]|0; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4 = ($3|0)==(0|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($4) {
    $11 = $__z;$14 = $0;$211 = $__z;
    label = 7;
    break;
   } else {
    $$0$i$i = $3;
    while(1) {
     $5 = HEAP32[$$0$i$i>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $6 = ($5|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     if ($6) {
      $$ph = $$0$i$i;
      label = 5;
      break;
     } else {
      $$0$i$i = $5;
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 5) {
  $7 = ((($$ph)) + 4|0); //@line 329 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $8 = HEAP32[$7>>2]|0; //@line 329 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $9 = ($8|0)==(0|0); //@line 333 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($9) {
   $$pre28 = ((($$ph)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$phi29Z2D = $$pre28;$18 = $$ph;$19 = 0;$212 = 0;$35 = $$ph;
  } else {
   $11 = $$ph;$14 = $8;$211 = $$ph;
   label = 7;
  }
 }
 if ((label|0) == 7) {
  $10 = ((($11)) + 8|0); //@line 334 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $12 = HEAP32[$10>>2]|0; //@line 334 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $13 = ((($14)) + 8|0); //@line 334 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$13>>2] = $12; //@line 334 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre$phi29Z2D = $10;$18 = $11;$19 = $14;$212 = 1;$35 = $211;
 }
 $15 = HEAP32[$$pre$phi29Z2D>>2]|0; //@line 347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $16 = HEAP32[$15>>2]|0; //@line 347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $17 = ($16|0)==($18|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($17) {
  HEAP32[$15>>2] = $19; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $20 = ($18|0)==($__root|0); //@line 338 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($20) {
   $$0 = $19;$__w$0 = 0;
  } else {
   $21 = ((($15)) + 4|0); //@line 339 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $22 = HEAP32[$21>>2]|0; //@line 339 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$0 = $__root;$__w$0 = $22;
  }
 } else {
  $23 = ((($15)) + 4|0); //@line 345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$23>>2] = $19; //@line 345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$0 = $__root;$__w$0 = $16;
 }
 $24 = ((($18)) + 12|0); //@line 349 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $25 = HEAP8[$24>>0]|0; //@line 349 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $26 = ($25<<24>>24)!=(0); //@line 349 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $27 = ($18|0)==($__z|0); //@line 352 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($27) {
  $$1 = $$0;
 } else {
  $28 = ((($__z)) + 8|0); //@line 355 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $29 = HEAP32[$28>>2]|0; //@line 355 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$pre$phi29Z2D>>2] = $29; //@line 355 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $30 = HEAP32[$28>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $31 = HEAP32[$30>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $32 = ($31|0)==($__z|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$cast9 = $29; //@line 357 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($32) {
   HEAP32[$$cast9>>2] = $18; //@line 357 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  } else {
   $33 = ((($$cast9)) + 4|0); //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$33>>2] = $18; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  }
  $34 = HEAP32[$__z>>2]|0; //@line 360 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$35>>2] = $34; //@line 360 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $36 = ((($34)) + 8|0); //@line 361 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$36>>2] = $18; //@line 361 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $37 = ((($__z)) + 4|0); //@line 362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $38 = HEAP32[$37>>2]|0; //@line 362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $39 = ((($18)) + 4|0); //@line 362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$39>>2] = $38; //@line 362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $40 = ($38|0)==(0|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if (!($40)) {
   $41 = ((($38)) + 8|0); //@line 364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$41>>2] = $18; //@line 364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  }
  $42 = ((($__z)) + 12|0); //@line 365 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $43 = HEAP8[$42>>0]|0; //@line 365 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP8[$24>>0] = $43; //@line 365 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $44 = ($$0|0)==($__z|0); //@line 366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$$0 = $44 ? $18 : $$0; //@line 366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$1 = $$$0;
 }
 $45 = ($$1|0)!=(0|0); //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $or$cond = $26 & $45; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if (!($or$cond)) {
  return; //@line 503 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
 if ($212) {
  $46 = ((($19)) + 12|0); //@line 386 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP8[$46>>0] = 1; //@line 386 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  return; //@line 503 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 } else {
  $$2 = $$1;$__w$1 = $__w$0;
 }
 while(1) {
  $47 = ((($__w$1)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $48 = HEAP32[$47>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $49 = HEAP32[$48>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $50 = ($49|0)==($__w$1|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $51 = ((($__w$1)) + 12|0); //@line 450 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $52 = HEAP8[$51>>0]|0; //@line 450 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $53 = ($52<<24>>24)!=(0); //@line 450 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($50) {
   if ($53) {
    $$6 = $$2;$__w$4 = $__w$1;
   } else {
    HEAP8[$51>>0] = 1; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $134 = ((($48)) + 12|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP8[$134>>0] = 0; //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $135 = HEAP32[$48>>2]|0; //@line 227 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $136 = ((($135)) + 4|0); //@line 228 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $137 = HEAP32[$136>>2]|0; //@line 228 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP32[$48>>2] = $137; //@line 228 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $138 = ($137|0)==(0|0); //@line 229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if (!($138)) {
     $139 = ((($137)) + 8|0); //@line 230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     HEAP32[$139>>2] = $48; //@line 230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    }
    $140 = ((($48)) + 8|0); //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $141 = HEAP32[$140>>2]|0; //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $142 = ((($135)) + 8|0); //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP32[$142>>2] = $141; //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $143 = HEAP32[$140>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $144 = HEAP32[$143>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $145 = ($144|0)==($48|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($145) {
     HEAP32[$143>>2] = $135; //@line 233 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    } else {
     $146 = ((($143)) + 4|0); //@line 235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     HEAP32[$146>>2] = $135; //@line 235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    }
    HEAP32[$136>>2] = $48; //@line 236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP32[$140>>2] = $135; //@line 237 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $147 = ((($__w$1)) + 4|0); //@line 457 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $148 = HEAP32[$147>>2]|0; //@line 457 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $149 = ($$2|0)==($148|0); //@line 457 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $__w$1$$23 = $149 ? $__w$1 : $$2; //@line 457 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $150 = HEAP32[$148>>2]|0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $$6 = $__w$1$$23;$__w$4 = $150;
   }
   $151 = HEAP32[$__w$4>>2]|0; //@line 482 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $152 = ($151|0)==(0|0); //@line 463 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if (!($152)) {
    $153 = ((($151)) + 12|0); //@line 463 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $154 = HEAP8[$153>>0]|0; //@line 463 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $155 = ($154<<24>>24)==(0); //@line 463 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($155) {
     $175 = $151;$__w$4111 = $__w$4;
     label = 68;
     break;
    }
   }
   $156 = ((($__w$4)) + 4|0); //@line 464 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $157 = HEAP32[$156>>2]|0; //@line 464 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $158 = ($157|0)==(0|0); //@line 464 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if (!($158)) {
    $159 = ((($157)) + 12|0); //@line 464 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $160 = HEAP8[$159>>0]|0; //@line 464 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $161 = ($160<<24>>24)==(0); //@line 464 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($161) {
     label = 67;
     break;
    }
   }
   $162 = ((($__w$4)) + 12|0); //@line 466 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP8[$162>>0] = 0; //@line 466 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $163 = ((($__w$4)) + 8|0); //@line 467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $164 = HEAP32[$163>>2]|0; //@line 467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $165 = ((($164)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $166 = HEAP8[$165>>0]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $167 = ($166<<24>>24)==(0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $168 = ($164|0)==($$6|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $or$cond4 = $168 | $167; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($or$cond4) {
    $$lcssa119 = $165;
    label = 64;
    break;
   }
   $169 = ((($164)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $170 = HEAP32[$169>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $171 = HEAP32[$170>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $172 = ($171|0)==($164|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $173 = ((($170)) + 4|0); //@line 476 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$in2 = $172 ? $173 : $170; //@line 475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$2$be = $$6;$__w$1$be$in = $$in2;
  } else {
   if ($53) {
    $$4 = $$2;$__w$2 = $__w$1;
   } else {
    HEAP8[$51>>0] = 1; //@line 400 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $54 = ((($48)) + 12|0); //@line 401 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP8[$54>>0] = 0; //@line 401 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $55 = ((($48)) + 4|0); //@line 207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $56 = HEAP32[$55>>2]|0; //@line 207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $57 = HEAP32[$56>>2]|0; //@line 208 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP32[$55>>2] = $57; //@line 208 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $58 = ($57|0)==(0|0); //@line 209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if (!($58)) {
     $59 = ((($57)) + 8|0); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     HEAP32[$59>>2] = $48; //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    }
    $60 = ((($48)) + 8|0); //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $61 = HEAP32[$60>>2]|0; //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $62 = ((($56)) + 8|0); //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP32[$62>>2] = $61; //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $63 = HEAP32[$60>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $64 = HEAP32[$63>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $65 = ($64|0)==($48|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($65) {
     HEAP32[$63>>2] = $56; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    } else {
     $66 = ((($63)) + 4|0); //@line 215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     HEAP32[$66>>2] = $56; //@line 215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    }
    HEAP32[$56>>2] = $48; //@line 216 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP32[$60>>2] = $56; //@line 217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $67 = HEAP32[$__w$1>>2]|0; //@line 405 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $68 = ($$2|0)==($67|0); //@line 405 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $__w$1$$2 = $68 ? $__w$1 : $$2; //@line 405 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $69 = ((($67)) + 4|0); //@line 408 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $70 = HEAP32[$69>>2]|0; //@line 408 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $$4 = $__w$1$$2;$__w$2 = $70;
   }
   $71 = HEAP32[$__w$2>>2]|0; //@line 227 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $72 = ($71|0)==(0|0); //@line 411 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if (!($72)) {
    $73 = ((($71)) + 12|0); //@line 411 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $74 = HEAP8[$73>>0]|0; //@line 411 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $75 = ($74<<24>>24)==(0); //@line 411 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($75) {
     $$lcssa = $__w$2;$$lcssa104 = $71;$__w$2$lcssa = $__w$2;
     label = 38;
     break;
    }
   }
   $76 = ((($__w$2)) + 4|0); //@line 412 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $77 = HEAP32[$76>>2]|0; //@line 412 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $78 = ($77|0)==(0|0); //@line 412 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if (!($78)) {
    $79 = ((($77)) + 12|0); //@line 412 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $80 = HEAP8[$79>>0]|0; //@line 412 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $81 = ($80<<24>>24)==(0); //@line 412 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($81) {
     $213 = $71;$214 = $__w$2;$97 = $77;$__w$2100 = $__w$2;
     label = 39;
     break;
    }
   }
   $82 = ((($__w$2)) + 12|0); //@line 414 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP8[$82>>0] = 0; //@line 414 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $83 = ((($__w$2)) + 8|0); //@line 415 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $84 = HEAP32[$83>>2]|0; //@line 415 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $85 = ($84|0)==($$4|0); //@line 417 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($85) {
    $$lcssa22 = $$4;
    label = 36;
    break;
   }
   $86 = ((($84)) + 12|0); //@line 417 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $87 = HEAP8[$86>>0]|0; //@line 417 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $88 = ($87<<24>>24)==(0); //@line 417 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($88) {
    $$lcssa22 = $84;
    label = 36;
    break;
   }
   $90 = ((($84)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $91 = HEAP32[$90>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $92 = HEAP32[$91>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $93 = ($92|0)==($84|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $94 = ((($91)) + 4|0); //@line 424 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$in = $93 ? $94 : $91; //@line 423 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$2$be = $$4;$__w$1$be$in = $$in;
  }
  $__w$1$be = HEAP32[$__w$1$be$in>>2]|0; //@line 475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$2 = $$2$be;$__w$1 = $__w$1$be;
 }
 if ((label|0) == 36) {
  $89 = ((($$lcssa22)) + 12|0); //@line 419 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP8[$89>>0] = 1; //@line 419 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  return; //@line 503 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
 else if ((label|0) == 38) {
  $$phi$trans$insert = ((($__w$2$lcssa)) + 4|0); //@line 430 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 430 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $95 = ($$pre|0)==(0|0); //@line 430 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($95) {
   $101 = $$lcssa104;$105 = $$lcssa;$__w$2101 = $__w$2$lcssa;
   label = 40;
  } else {
   $213 = $$lcssa104;$214 = $$lcssa;$97 = $$pre;$__w$2100 = $__w$2$lcssa;
   label = 39;
  }
 }
 else if ((label|0) == 64) {
  HEAP8[$$lcssa119>>0] = 1; //@line 471 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  return; //@line 503 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
 else if ((label|0) == 67) {
  if ($152) {
   $__w$4112 = $__w$4;
   label = 69;
  } else {
   $175 = $151;$__w$4111 = $__w$4;
   label = 68;
  }
 }
 if ((label|0) == 39) {
  $96 = ((($97)) + 12|0); //@line 430 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $98 = HEAP8[$96>>0]|0; //@line 430 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $99 = ($98<<24>>24)==(0); //@line 430 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($99) {
   $121 = $97;$__w$3 = $__w$2100;
   label = 46;
  } else {
   $101 = $213;$105 = $214;$__w$2101 = $__w$2100;
   label = 40;
  }
 }
 else if ((label|0) == 68) {
  $174 = ((($175)) + 12|0); //@line 482 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $176 = HEAP8[$174>>0]|0; //@line 482 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $177 = ($176<<24>>24)==(0); //@line 482 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($177) {
   $198 = $175;$__w$5 = $__w$4111;
   label = 75;
  } else {
   $__w$4112 = $__w$4111;
   label = 69;
  }
 }
 if ((label|0) == 40) {
  $100 = ((($101)) + 12|0); //@line 433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP8[$100>>0] = 1; //@line 433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $102 = ((($__w$2101)) + 12|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP8[$102>>0] = 0; //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $103 = ((($101)) + 4|0); //@line 228 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $104 = HEAP32[$103>>2]|0; //@line 228 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$105>>2] = $104; //@line 228 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $106 = ($104|0)==(0|0); //@line 229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if (!($106)) {
   $107 = ((($104)) + 8|0); //@line 230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$107>>2] = $__w$2101; //@line 230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  }
  $108 = ((($__w$2101)) + 8|0); //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $109 = HEAP32[$108>>2]|0; //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $110 = ((($101)) + 8|0); //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$110>>2] = $109; //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $111 = HEAP32[$108>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $112 = HEAP32[$111>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $113 = ($112|0)==($__w$2101|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($113) {
   HEAP32[$111>>2] = $101; //@line 233 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  } else {
   $114 = ((($111)) + 4|0); //@line 235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$114>>2] = $101; //@line 235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  }
  HEAP32[$103>>2] = $__w$2101; //@line 236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$108>>2] = $101; //@line 237 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $121 = $__w$2101;$__w$3 = $101;
  label = 46;
 }
 else if ((label|0) == 69) {
  $178 = ((($__w$4112)) + 4|0); //@line 485 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $179 = HEAP32[$178>>2]|0; //@line 485 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $180 = ((($179)) + 12|0); //@line 485 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP8[$180>>0] = 1; //@line 485 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $181 = ((($__w$4112)) + 12|0); //@line 486 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP8[$181>>0] = 0; //@line 486 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $182 = HEAP32[$179>>2]|0; //@line 208 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$178>>2] = $182; //@line 208 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $183 = ($182|0)==(0|0); //@line 209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if (!($183)) {
   $184 = ((($182)) + 8|0); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$184>>2] = $__w$4112; //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  }
  $185 = ((($__w$4112)) + 8|0); //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $186 = HEAP32[$185>>2]|0; //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $187 = ((($179)) + 8|0); //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$187>>2] = $186; //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $188 = HEAP32[$185>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $189 = HEAP32[$188>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $190 = ($189|0)==($__w$4112|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($190) {
   HEAP32[$188>>2] = $179; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  } else {
   $191 = ((($188)) + 4|0); //@line 215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$191>>2] = $179; //@line 215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  }
  HEAP32[$179>>2] = $__w$4112; //@line 216 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$185>>2] = $179; //@line 217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $198 = $__w$4112;$__w$5 = $179;
  label = 75;
 }
 if ((label|0) == 46) {
  $115 = ((($__w$3)) + 8|0); //@line 441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $116 = HEAP32[$115>>2]|0; //@line 441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $117 = ((($116)) + 12|0); //@line 441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $118 = HEAP8[$117>>0]|0; //@line 441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $119 = ((($__w$3)) + 12|0); //@line 441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP8[$119>>0] = $118; //@line 441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP8[$117>>0] = 1; //@line 442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $120 = ((($121)) + 12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP8[$120>>0] = 1; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $122 = ((($116)) + 4|0); //@line 207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $123 = HEAP32[$122>>2]|0; //@line 207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $124 = HEAP32[$123>>2]|0; //@line 208 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$122>>2] = $124; //@line 208 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $125 = ($124|0)==(0|0); //@line 209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if (!($125)) {
   $126 = ((($124)) + 8|0); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$126>>2] = $116; //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  }
  $127 = ((($116)) + 8|0); //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $128 = HEAP32[$127>>2]|0; //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $129 = ((($123)) + 8|0); //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$129>>2] = $128; //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $130 = HEAP32[$127>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $131 = HEAP32[$130>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $132 = ($131|0)==($116|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($132) {
   HEAP32[$130>>2] = $123; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  } else {
   $133 = ((($130)) + 4|0); //@line 215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$133>>2] = $123; //@line 215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  }
  HEAP32[$123>>2] = $116; //@line 216 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$127>>2] = $123; //@line 217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  return; //@line 503 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
 else if ((label|0) == 75) {
  $192 = ((($__w$5)) + 8|0); //@line 493 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $193 = HEAP32[$192>>2]|0; //@line 493 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $194 = ((($193)) + 12|0); //@line 493 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $195 = HEAP8[$194>>0]|0; //@line 493 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $196 = ((($__w$5)) + 12|0); //@line 493 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP8[$196>>0] = $195; //@line 493 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP8[$194>>0] = 1; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $197 = ((($198)) + 12|0); //@line 495 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP8[$197>>0] = 1; //@line 495 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $199 = HEAP32[$193>>2]|0; //@line 227 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $200 = ((($199)) + 4|0); //@line 228 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $201 = HEAP32[$200>>2]|0; //@line 228 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$193>>2] = $201; //@line 228 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $202 = ($201|0)==(0|0); //@line 229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if (!($202)) {
   $203 = ((($201)) + 8|0); //@line 230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$203>>2] = $193; //@line 230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  }
  $204 = ((($193)) + 8|0); //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $205 = HEAP32[$204>>2]|0; //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $206 = ((($199)) + 8|0); //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$206>>2] = $205; //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $207 = HEAP32[$204>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $208 = HEAP32[$207>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $209 = ($208|0)==($193|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($209) {
   HEAP32[$207>>2] = $199; //@line 233 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  } else {
   $210 = ((($207)) + 4|0); //@line 235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$210>>2] = $199; //@line 235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  }
  HEAP32[$200>>2] = $193; //@line 236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$204>>2] = $199; //@line 237 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  return; //@line 503 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
}
function __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ModuleEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE7destroyEPNS_11__tree_nodeISB_PvEE($this,$__nd) {
 $this = $this|0;
 $__nd = $__nd|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__nd|0)==(0|0); //@line 1440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($0) {
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 } else {
  $1 = HEAP32[$__nd>>2]|0; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ModuleEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE7destroyEPNS_11__tree_nodeISB_PvEE($this,$1); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2 = ((($__nd)) + 4|0); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3 = HEAP32[$2>>2]|0; //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ModuleEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE7destroyEPNS_11__tree_nodeISB_PvEE($this,$3); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4 = ((($__nd)) + 16|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
  __ZdlPv($__nd); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
}
function __ZN4love6ModuleD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ModuleD2Ev($this); //@line 64 "common/Module.cpp"
 __ZdlPv($this); //@line 64 "common/Module.cpp"
 return; //@line 85 "common/Module.cpp"
}
function __ZN4love6Module16registerInstanceEPS0_($instance) {
 $instance = $instance|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_ptr6 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 56|0;
 $name = sp + 40|0;
 $1 = sp + 24|0;
 $2 = $instance;
 $3 = ($instance|0)==(0|0); //@line 89 "common/Module.cpp"
 if ($3) {
  $4 = (___cxa_allocate_exception(16)|0); //@line 90 "common/Module.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($4|0),(180122|0),($vararg_buffer|0)); //@line 90 "common/Module.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   ___cxa_throw(($4|0),(96|0),(344|0)); //@line 90 "common/Module.cpp"
   // unreachable; //@line 90 "common/Module.cpp"
  }
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  ___cxa_free_exception(($4|0)); //@line 90 "common/Module.cpp"
  $$1 = $7;$$12 = $8;
  ___resumeException($$1|0);
  // unreachable;
 }
 $9 = HEAP32[$instance>>2]|0; //@line 92 "common/Module.cpp"
 $10 = ((($9)) + 20|0); //@line 92 "common/Module.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 92 "common/Module.cpp"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($instance)|0); //@line 92 "common/Module.cpp"
 $13 = (_strlen($12)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($name,$12,$13); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = HEAP32[156420]|0; //@line 41 "common/Module.cpp"
 $15 = ($14|0)==(0|0); //@line 41 "common/Module.cpp"
 if ($15) {
  __THREW__ = 0;
  $16 = (invoke_ii(3348,12)|0); //@line 42 "common/Module.cpp"
  $17 = __THREW__; __THREW__ = 0;
  $18 = $17&1;
  if ($18) {
   label = 10;
  } else {
   $19 = ((($16)) + 4|0); //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$19>>2] = 0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $20 = ((($16)) + 8|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$20>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$16>>2] = $19; //@line 1130 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[156420] = $16; //@line 42 "common/Module.cpp"
   $21 = $16;
   label = 8;
  }
 } else {
  $21 = $14;
  label = 8;
 }
 do {
  if ((label|0) == 8) {
   $22 = (__ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ModuleEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE4findIS7_EENS_15__tree_iteratorISB_PNS_11__tree_nodeISB_PvEEiEERKT_($21,$name)|0); //@line 1090 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $23 = ((($21)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $24 = ($22|0)==($23|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if (!($24)) {
    $25 = ((($22)) + 28|0); //@line 100 "common/Module.cpp"
    $26 = HEAP32[$25>>2]|0; //@line 100 "common/Module.cpp"
    $27 = ($26|0)==($instance|0); //@line 100 "common/Module.cpp"
    if ($27) {
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($name); //@line 116 "common/Module.cpp"
     STACKTOP = sp;return; //@line 116 "common/Module.cpp"
    }
    $30 = (___cxa_allocate_exception(16)|0); //@line 102 "common/Module.cpp"
    $31 = HEAP32[$instance>>2]|0; //@line 102 "common/Module.cpp"
    $32 = ((($31)) + 20|0); //@line 102 "common/Module.cpp"
    $33 = HEAP32[$32>>2]|0; //@line 102 "common/Module.cpp"
    __THREW__ = 0;
    $34 = (invoke_ii($33|0,($instance|0))|0); //@line 102 "common/Module.cpp"
    $35 = __THREW__; __THREW__ = 0;
    $36 = $35&1;
    if (!($36)) {
     __THREW__ = 0;
     HEAP32[$vararg_buffer1>>2] = $34; //@line 102 "common/Module.cpp"
     invoke_viii(3349,($30|0),(180146|0),($vararg_buffer1|0)); //@line 102 "common/Module.cpp"
     $37 = __THREW__; __THREW__ = 0;
     $38 = $37&1;
     if (!($38)) {
      __THREW__ = 0;
      invoke_viii(3350,($30|0),(96|0),(344|0)); //@line 102 "common/Module.cpp"
      $39 = __THREW__; __THREW__ = 0;
      label = 10;
      break;
     }
    }
    $40 = ___cxa_find_matching_catch_2()|0;
    $41 = tempRet0;
    ___cxa_free_exception(($30|0)); //@line 102 "common/Module.cpp"
    $$0 = $40;$$01 = $41;
    break;
   }
   __THREW__ = 0;
   invoke_vii(3351,($1|0),($name|0)); //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
   $42 = __THREW__; __THREW__ = 0;
   $43 = $42&1;
   if ($43) {
    label = 10;
   } else {
    $44 = ((($1)) + 12|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
    HEAP32[$44>>2] = $2; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
    __THREW__ = 0;
    invoke_viii(3352,($0|0),($21|0),($1|0)); //@line 1038 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $45 = __THREW__; __THREW__ = 0;
    $46 = $45&1;
    if ($46) {
     $68 = ___cxa_find_matching_catch_2()|0;
     $69 = tempRet0;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     $$0 = $68;$$01 = $69;
     break;
    }
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
    $47 = HEAP32[$instance>>2]|0; //@line 107 "common/Module.cpp"
    $48 = ((($47)) + 16|0); //@line 107 "common/Module.cpp"
    $49 = HEAP32[$48>>2]|0; //@line 107 "common/Module.cpp"
    __THREW__ = 0;
    $50 = (invoke_ii($49|0,($instance|0))|0); //@line 107 "common/Module.cpp"
    $51 = __THREW__; __THREW__ = 0;
    $52 = $51&1;
    if ($52) {
     label = 10;
    } else {
     $53 = (625608 + ($50<<2)|0); //@line 109 "common/Module.cpp"
     $54 = HEAP32[$53>>2]|0; //@line 109 "common/Module.cpp"
     $55 = ($54|0)==(0|0); //@line 109 "common/Module.cpp"
     if (!($55)) {
      $56 = HEAP32[$54>>2]|0; //@line 112 "common/Module.cpp"
      $57 = ((($56)) + 20|0); //@line 112 "common/Module.cpp"
      $58 = HEAP32[$57>>2]|0; //@line 112 "common/Module.cpp"
      __THREW__ = 0;
      $59 = (invoke_ii($58|0,($54|0))|0); //@line 112 "common/Module.cpp"
      $60 = __THREW__; __THREW__ = 0;
      $61 = $60&1;
      if ($61) {
       label = 10;
       break;
      }
      $62 = HEAP32[$instance>>2]|0; //@line 112 "common/Module.cpp"
      $63 = ((($62)) + 20|0); //@line 112 "common/Module.cpp"
      $64 = HEAP32[$63>>2]|0; //@line 112 "common/Module.cpp"
      __THREW__ = 0;
      $65 = (invoke_ii($64|0,($instance|0))|0); //@line 112 "common/Module.cpp"
      $66 = __THREW__; __THREW__ = 0;
      $67 = $66&1;
      if ($67) {
       label = 10;
       break;
      }
      HEAP32[$vararg_buffer3>>2] = $59; //@line 111 "common/Module.cpp"
      $vararg_ptr6 = ((($vararg_buffer3)) + 4|0); //@line 111 "common/Module.cpp"
      HEAP32[$vararg_ptr6>>2] = $65; //@line 111 "common/Module.cpp"
      (_printf(180176,$vararg_buffer3)|0); //@line 111 "common/Module.cpp"
     }
     HEAP32[$53>>2] = $2; //@line 115 "common/Module.cpp"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($name); //@line 116 "common/Module.cpp"
     STACKTOP = sp;return; //@line 116 "common/Module.cpp"
    }
   }
  }
 } while(0);
 if ((label|0) == 10) {
  $28 = ___cxa_find_matching_catch_2()|0;
  $29 = tempRet0;
  $$0 = $28;$$01 = $29;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($name); //@line 116 "common/Module.cpp"
 $$1 = $$0;$$12 = $$01;
 ___resumeException($$1|0);
 // unreachable;
}
function __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ModuleEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE4findIS7_EENS_15__tree_iteratorISB_PNS_11__tree_nodeISB_PvEEiEERKT_($this,$__v) {
 $this = $this|0;
 $__v = $__v|0;
 var $$$$i$i$i$i = 0, $$$$i$i$i$i$i = 0, $$0$ph$lcssa$i = 0, $$0$ph6$i = 0, $$01$ph5$i = 0, $$013$i = 0, $$013$i$lcssa = 0, $$sroa$0$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $2 = ($1|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if (!($2)) {
  $3 = HEAP8[$__v>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $4 = $3 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $5 = ($4<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $6 = ((($__v)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = HEAP32[$6>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $8 = $3&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = $8 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $10 = $5 ? $9 : $7; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $11 = ((($__v)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $12 = HEAP32[$11>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $13 = ((($__v)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $14 = $5 ? $13 : $12; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $$0$ph6$i = $0;$$01$ph5$i = $1;
  L3: while(1) {
   $$013$i = $$01$ph5$i;
   while(1) {
    $15 = ((($$013$i)) + 16|0); //@line 2102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $16 = HEAP8[$15>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $17 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $18 = ($17<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $19 = ((($$013$i)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $20 = HEAP32[$19>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $21 = $16&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $22 = $21 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $23 = $18 ? $22 : $20; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $24 = ((($$013$i)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $25 = HEAP32[$24>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $26 = ((($15)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $27 = $18 ? $26 : $25; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $28 = ($10>>>0)<($23>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $29 = $28 ? $10 : $23; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $30 = (_memcmp($27,$14,$29)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $31 = ($30|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $32 = ($23>>>0)<($10>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $$$$i$i$i$i$i = $32 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $33 = $31 ? $$$$i$i$i$i$i : $30; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $34 = ($33|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if (!($34)) {
     $$013$i$lcssa = $$013$i;
     break;
    }
    $37 = ((($$013$i)) + 4|0); //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $38 = HEAP32[$37>>2]|0; //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $39 = ($38|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($39) {
     $$0$ph$lcssa$i = $$0$ph6$i;
     break L3;
    } else {
     $$013$i = $38;
    }
   }
   $35 = HEAP32[$$013$i$lcssa>>2]|0; //@line 2105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $36 = ($35|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($36) {
    $$0$ph$lcssa$i = $$013$i$lcssa;
    break;
   } else {
    $$0$ph6$i = $$013$i$lcssa;$$01$ph5$i = $35;
   }
  }
  $40 = ($$0$ph$lcssa$i|0)==($0|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if (!($40)) {
   $41 = ((($$0$ph$lcssa$i)) + 16|0); //@line 650 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $42 = HEAP8[$41>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $43 = $42 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $44 = ($43<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $45 = ((($$0$ph$lcssa$i)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $46 = HEAP32[$45>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $47 = $42&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $48 = $47 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $49 = $44 ? $48 : $46; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $50 = ((($$0$ph$lcssa$i)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $51 = HEAP32[$50>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $52 = ((($41)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $53 = $44 ? $52 : $51; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $54 = ($49>>>0)<($10>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $55 = $54 ? $49 : $10; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $56 = (_memcmp($14,$53,$55)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $57 = ($56|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $58 = ($10>>>0)<($49>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$$$i$i$i$i = $58 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $59 = $57 ? $$$$i$i$i$i : $56; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $60 = ($59|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if (!($60)) {
    $$sroa$0$0 = $$0$ph$lcssa$i;
    return ($$sroa$0$0|0); //@line 2032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   }
  }
 }
 $$sroa$0$0 = $0;
 return ($$sroa$0$0|0); //@line 2032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
}
function __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ModuleEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE15__insert_uniqueINS_4pairIS7_SA_EEEENSJ_INS_15__tree_iteratorISB_PNS_11__tree_nodeISB_PvEEiEEbEEOT_($agg$result,$this,$__v) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $__v = $__v|0;
 var $$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__inserted$0$i = 0, $__parent$i = 0, $__r$0$i = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__parent$i = sp;
 $0 = (__Znwj(32)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $1 = ((($0)) + 16|0); //@line 1731 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 ;HEAP32[$1>>2]=HEAP32[$__v>>2]|0;HEAP32[$1+4>>2]=HEAP32[$__v+4>>2]|0;HEAP32[$1+8>>2]=HEAP32[$__v+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 ;HEAP32[$__v>>2]=0|0;HEAP32[$__v+4>>2]=0|0;HEAP32[$__v+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = ((($0)) + 28|0); //@line 316 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $3 = ((($__v)) + 12|0); //@line 316 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $4 = HEAP32[$3>>2]|0; //@line 316 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 HEAP32[$2>>2] = $4; //@line 316 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $5 = (__ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ModuleEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE12__find_equalISB_EERPNS_16__tree_node_baseIPvEESN_RKT_($this,$__parent$i,$1)|0); //@line 1923 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $6 = HEAP32[$5>>2]|0; //@line 1924 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $7 = ($6|0)==(0|0); //@line 1926 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($7) {
  $8 = HEAP32[$__parent$i>>2]|0; //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$0>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $9 = ((($0)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$9>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $10 = ((($0)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$10>>2] = $8; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5>>2] = $0; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $11 = HEAP32[$this>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $12 = HEAP32[$11>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $13 = ($12|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($13) {
   $17 = $0;
  } else {
   $14 = $12; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$this>>2] = $14; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i = HEAP32[$5>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $17 = $$pre$i$i;
  }
  $15 = ((($this)) + 4|0); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $16 = HEAP32[$15>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($16,$17); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $18 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $19 = HEAP32[$18>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $20 = (($19) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$18>>2] = $20; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $__inserted$0$i = 1;$__r$0$i = $0;
 } else {
  $__inserted$0$i = 0;$__r$0$i = $6;
 }
 $21 = $__r$0$i; //@line 679 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$agg$result>>2] = $21; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $22 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 HEAP8[$22>>0] = $__inserted$0$i; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $not$ = ($__inserted$0$i<<24>>24)==(0); //@line 2629 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($not$)) {
  STACKTOP = sp;return; //@line 1809 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 __ZdlPv($0); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 STACKTOP = sp;return; //@line 1809 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
}
function __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ModuleEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE12__find_equalISB_EERPNS_16__tree_node_baseIPvEESN_RKT_($this,$__parent,$__v) {
 $this = $this|0;
 $__parent = $__parent|0;
 $__v = $__v|0;
 var $$$$i$i$i$i = 0, $$$$i$i$i$i1 = 0, $$0 = 0, $$lcssa = 0, $$lcssa19 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__nd$0 = 0, $__nd$0$lcssa = 0, $__nd$0$lcssa17 = 0, $__nd$0$lcssa18 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $2 = ($1|0)==(0|0); //@line 1609 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($2) {
  HEAP32[$__parent>>2] = $0; //@line 1640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$0 = $0;
  return ($$0|0); //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
 $3 = HEAP8[$__v>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = $3 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ($4<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = ((($__v)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = HEAP32[$6>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $3&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $8 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $5 ? $9 : $7; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = ((($__v)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = HEAP32[$11>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ((($__v)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = $5 ? $13 : $12; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $__nd$0 = $1;
 while(1) {
  $15 = ((($__nd$0)) + 16|0); //@line 1613 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $16 = HEAP8[$15>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $17 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $18 = ($17<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $19 = ((($__nd$0)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $20 = HEAP32[$19>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $21 = $16&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $22 = $21 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $23 = $18 ? $22 : $20; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $24 = ((($__nd$0)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $25 = HEAP32[$24>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $26 = ((($15)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $27 = $18 ? $26 : $25; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $28 = ($23>>>0)<($10>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $29 = $28 ? $23 : $10; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $30 = (_memcmp($14,$27,$29)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $31 = ($30|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $32 = ($10>>>0)<($23>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $$$$i$i$i$i1 = $32 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $33 = $31 ? $$$$i$i$i$i1 : $30; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $34 = ($33|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($34) {
   $35 = HEAP32[$__nd$0>>2]|0; //@line 1615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $36 = ($35|0)==(0|0); //@line 1615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($36) {
    $$lcssa19 = $__nd$0;$__nd$0$lcssa18 = $__nd$0;
    label = 5;
    break;
   } else {
    $$sink = $35;
   }
  } else {
   $37 = $32 ? $10 : $23; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $38 = (_memcmp($27,$14,$37)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $39 = ($38|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$$$i$i$i$i = $28 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $40 = $39 ? $$$$i$i$i$i : $38; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $41 = ($40|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if (!($41)) {
    $__nd$0$lcssa = $__nd$0;
    label = 10;
    break;
   }
   $42 = ((($__nd$0)) + 4|0); //@line 1625 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $43 = HEAP32[$42>>2]|0; //@line 1625 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $44 = ($43|0)==(0|0); //@line 1625 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($44) {
    $$lcssa = $42;$__nd$0$lcssa17 = $__nd$0;
    label = 9;
    break;
   } else {
    $$sink = $43;
   }
  }
  $__nd$0 = $$sink;
 }
 if ((label|0) == 5) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa18; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$0 = $$lcssa19;
  return ($$0|0); //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
 else if ((label|0) == 9) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa17; //@line 1629 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$0 = $$lcssa;
  return ($$0|0); //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
 else if ((label|0) == 10) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa; //@line 1635 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$0 = $__parent;
  return ($$0|0); //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
 return (0)|0;
}
function __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($__root,$__x) {
 $__root = $__root|0;
 $__x = $__x|0;
 var $$010 = 0, $$010$lcssa59 = 0, $$010$lcssa60 = 0, $$lcssa = 0, $$lcssa61 = 0, $$lcssa62 = 0, $$lcssa63 = 0, $$lcssa64 = 0, $$lcssa65 = 0, $$lcssa66 = 0, $$lcssa67 = 0, $$lcssa68 = 0, $$lcssa69 = 0, $$pre = 0, $$pre16 = 0, $$pre18 = 0, $$pre19 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__x|0)==($__root|0); //@line 252 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $1 = ((($__x)) + 12|0); //@line 252 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $2 = $0&1; //@line 252 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP8[$1>>0] = $2; //@line 252 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($0) {
  return; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 } else {
  $$010 = $__x;
 }
 while(1) {
  $3 = ((($$010)) + 8|0); //@line 253 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4 = HEAP32[$3>>2]|0; //@line 271 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5 = ((($4)) + 12|0); //@line 253 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6 = HEAP8[$5>>0]|0; //@line 253 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $7 = ($6<<24>>24)==(0); //@line 253 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if (!($7)) {
   label = 37;
   break;
  }
  $8 = ((($4)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $9 = HEAP32[$8>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $10 = HEAP32[$9>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $11 = ($10|0)==($4|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($11) {
   $12 = ((($9)) + 4|0); //@line 258 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $13 = HEAP32[$12>>2]|0; //@line 258 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $14 = ($13|0)==(0|0); //@line 259 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($14) {
    $$010$lcssa60 = $$010;$$lcssa62 = $4;$$lcssa63 = $8;$$lcssa65 = $8;$$lcssa67 = $9;$$lcssa69 = $9;
    label = 8;
    break;
   }
   $15 = ((($13)) + 12|0); //@line 259 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $16 = HEAP8[$15>>0]|0; //@line 259 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $17 = ($16<<24>>24)==(0); //@line 259 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if (!($17)) {
    $$010$lcssa60 = $$010;$$lcssa62 = $4;$$lcssa63 = $8;$$lcssa65 = $8;$$lcssa67 = $9;$$lcssa69 = $9;
    label = 8;
    break;
   }
   HEAP8[$5>>0] = 1; //@line 262 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $18 = ($9|0)==($__root|0); //@line 264 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $19 = ((($9)) + 12|0); //@line 264 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $20 = $18&1; //@line 264 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP8[$19>>0] = $20; //@line 264 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP8[$15>>0] = 1; //@line 265 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  } else {
   $52 = ($10|0)==(0|0); //@line 285 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($52) {
    $$010$lcssa59 = $$010;$$lcssa = $8;$$lcssa61 = $4;$$lcssa64 = $8;$$lcssa66 = $9;$$lcssa68 = $9;
    label = 24;
    break;
   }
   $53 = ((($10)) + 12|0); //@line 285 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $54 = HEAP8[$53>>0]|0; //@line 285 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $55 = ($54<<24>>24)==(0); //@line 285 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if (!($55)) {
    $$010$lcssa59 = $$010;$$lcssa = $8;$$lcssa61 = $4;$$lcssa64 = $8;$$lcssa66 = $9;$$lcssa68 = $9;
    label = 24;
    break;
   }
   HEAP8[$5>>0] = 1; //@line 288 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $56 = ($9|0)==($__root|0); //@line 290 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $57 = ((($9)) + 12|0); //@line 290 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $58 = $56&1; //@line 290 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP8[$57>>0] = $58; //@line 290 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP8[$53>>0] = 1; //@line 291 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  }
  $21 = ($9|0)==($__root|0); //@line 253 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($21) {
   label = 37;
   break;
  } else {
   $$010 = $9;
  }
 }
 if ((label|0) == 8) {
  $22 = HEAP32[$$lcssa62>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $23 = ($22|0)==($$010$lcssa60|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($23) {
   $37 = $$lcssa62;$39 = $$lcssa69;
  } else {
   $24 = $$lcssa67; //@line 256 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $25 = ((($$lcssa62)) + 4|0); //@line 207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $26 = HEAP32[$25>>2]|0; //@line 207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $27 = HEAP32[$26>>2]|0; //@line 208 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$25>>2] = $27; //@line 208 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $28 = ($27|0)==(0|0); //@line 209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($28) {
    $31 = $24;
   } else {
    $29 = ((($27)) + 8|0); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP32[$29>>2] = $$lcssa62; //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $$pre18 = HEAP32[$$lcssa63>>2]|0; //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $31 = $$pre18;
   }
   $30 = ((($26)) + 8|0); //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$30>>2] = $31; //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $32 = HEAP32[$$lcssa65>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $33 = HEAP32[$32>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $34 = ($33|0)==($$lcssa62|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($34) {
    HEAP32[$32>>2] = $26; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   } else {
    $35 = ((($32)) + 4|0); //@line 215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP32[$35>>2] = $26; //@line 215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   }
   HEAP32[$26>>2] = $$lcssa62; //@line 216 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$$lcssa63>>2] = $26; //@line 217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre19 = HEAP32[$30>>2]|0; //@line 276 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $37 = $26;$39 = $$pre19;
  }
  $36 = ((($37)) + 12|0); //@line 275 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP8[$36>>0] = 1; //@line 275 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $38 = ((($39)) + 12|0); //@line 277 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP8[$38>>0] = 0; //@line 277 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $40 = HEAP32[$39>>2]|0; //@line 227 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $41 = ((($40)) + 4|0); //@line 228 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $42 = HEAP32[$41>>2]|0; //@line 228 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$39>>2] = $42; //@line 228 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $43 = ($42|0)==(0|0); //@line 229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if (!($43)) {
   $44 = ((($42)) + 8|0); //@line 230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$44>>2] = $39; //@line 230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  }
  $45 = ((($39)) + 8|0); //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $46 = HEAP32[$45>>2]|0; //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $47 = ((($40)) + 8|0); //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$47>>2] = $46; //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $48 = HEAP32[$45>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $49 = HEAP32[$48>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $50 = ($49|0)==($39|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($50) {
   HEAP32[$48>>2] = $40; //@line 233 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  } else {
   $51 = ((($48)) + 4|0); //@line 235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$51>>2] = $40; //@line 235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  }
  HEAP32[$41>>2] = $39; //@line 236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$45>>2] = $40; //@line 237 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  return; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
 else if ((label|0) == 24) {
  $59 = HEAP32[$$lcssa61>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $60 = ($59|0)==($$010$lcssa59|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($60) {
   $61 = $$lcssa66; //@line 256 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $62 = HEAP32[$$lcssa61>>2]|0; //@line 227 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $63 = ((($62)) + 4|0); //@line 228 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $64 = HEAP32[$63>>2]|0; //@line 228 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$$lcssa61>>2] = $64; //@line 228 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $65 = ($64|0)==(0|0); //@line 229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($65) {
    $68 = $61;
   } else {
    $66 = ((($64)) + 8|0); //@line 230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP32[$66>>2] = $$lcssa61; //@line 230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $$pre = HEAP32[$$lcssa>>2]|0; //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $68 = $$pre;
   }
   $67 = ((($62)) + 8|0); //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$67>>2] = $68; //@line 231 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $69 = HEAP32[$$lcssa64>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $70 = HEAP32[$69>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $71 = ($70|0)==($$lcssa61|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($71) {
    HEAP32[$69>>2] = $62; //@line 233 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   } else {
    $72 = ((($69)) + 4|0); //@line 235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP32[$72>>2] = $62; //@line 235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   }
   HEAP32[$63>>2] = $$lcssa61; //@line 236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$$lcssa>>2] = $62; //@line 237 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre16 = HEAP32[$67>>2]|0; //@line 302 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $74 = $62;$76 = $$pre16;
  } else {
   $74 = $$lcssa61;$76 = $$lcssa68;
  }
  $73 = ((($74)) + 12|0); //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP8[$73>>0] = 1; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $75 = ((($76)) + 12|0); //@line 303 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP8[$75>>0] = 0; //@line 303 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $77 = ((($76)) + 4|0); //@line 207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $78 = HEAP32[$77>>2]|0; //@line 207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $79 = HEAP32[$78>>2]|0; //@line 208 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$77>>2] = $79; //@line 208 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $80 = ($79|0)==(0|0); //@line 209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if (!($80)) {
   $81 = ((($79)) + 8|0); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$81>>2] = $76; //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  }
  $82 = ((($76)) + 8|0); //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $83 = HEAP32[$82>>2]|0; //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $84 = ((($78)) + 8|0); //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$84>>2] = $83; //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $85 = HEAP32[$82>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $86 = HEAP32[$85>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $87 = ($86|0)==($76|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($87) {
   HEAP32[$85>>2] = $78; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  } else {
   $88 = ((($85)) + 4|0); //@line 215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$88>>2] = $78; //@line 215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  }
  HEAP32[$78>>2] = $76; //@line 216 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$82>>2] = $78; //@line 217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  return; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
 else if ((label|0) == 37) {
  return; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
}
function __ZN4love6Module11getInstanceERKNSt3__112basic_stringIcNS1_11char_traitsIcEENS1_9allocatorIcEEEE($name) {
 $name = $name|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[156420]|0; //@line 41 "common/Module.cpp"
 $1 = ($0|0)==(0|0); //@line 41 "common/Module.cpp"
 if ($1) {
  $2 = (__Znwj(12)|0); //@line 42 "common/Module.cpp"
  $3 = ((($2)) + 4|0); //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3>>2] = 0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4 = ((($2)) + 8|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2>>2] = $3; //@line 1130 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[156420] = $2; //@line 42 "common/Module.cpp"
  $5 = $2;
 } else {
  $5 = $0;
 }
 $6 = (__ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ModuleEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE4findIS7_EENS_15__tree_iteratorISB_PNS_11__tree_nodeISB_PvEEiEERKT_($5,$name)|0); //@line 1090 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $7 = ((($5)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = ($7|0)==($6|0); //@line 672 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($8) {
  $$0 = 0;
  return ($$0|0); //@line 128 "common/Module.cpp"
 }
 $9 = ((($6)) + 28|0); //@line 127 "common/Module.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 127 "common/Module.cpp"
 $$0 = $10;
 return ($$0|0); //@line 128 "common/Module.cpp"
}
function __ZN4love6Object6retainEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 917 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/atomic"
 $1 = HEAP32[$0>>2]|0;HEAP32[$0>>2] = (($1+1)|0); //@line 917 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/atomic"
 return; //@line 49 "common/Object.cpp"
}
function __ZN4love6Object7releaseEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 923 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/atomic"
 $1 = HEAP32[$0>>2]|0;HEAP32[$0>>2] = (($1-1)|0); //@line 923 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/atomic"
 $2 = ($1|0)==(1); //@line 54 "common/Object.cpp"
 if (!($2)) {
  return; //@line 59 "common/Object.cpp"
 }
 /* fence */; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/atomic"
 $3 = ($this|0)==(0|0); //@line 57 "common/Object.cpp"
 if ($3) {
  return; //@line 59 "common/Object.cpp"
 }
 $4 = HEAP32[$this>>2]|0; //@line 57 "common/Object.cpp"
 $5 = ((($4)) + 4|0); //@line 57 "common/Object.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 57 "common/Object.cpp"
 FUNCTION_TABLE_vi[$6 & 8191]($this); //@line 57 "common/Object.cpp"
 return; //@line 59 "common/Object.cpp"
}
function __ZN4love6ObjectC2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (9272); //@line 29 "common/Object.cpp"
 $0 = ((($this)) + 4|0); //@line 887 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/atomic"
 HEAP32[$0>>2] = 1; //@line 887 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/atomic"
 return; //@line 30 "common/Object.cpp"
}
function __ZN4love6ObjectD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 39 "common/Object.cpp"
}
function __ZNK4love6Object17getReferenceCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 835 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/atomic"
 $1 = HEAP32[$0>>2]|0; //@line 835 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/atomic"
 return ($1|0); //@line 43 "common/Object.cpp"
}
function _love_version() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (180253|0); //@line 203 "modules/love/love.cpp"
}
function _love_codename() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (180260|0); //@line 208 "modules/love/love.cpp"
}
function _luaopen_love($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $i1$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 (__ZN4love23luax_insistpinnedthreadEP9lua_State($L)|0); //@line 263 "modules/love/love.cpp"
 (__ZN4love17luax_insistglobalEP9lua_StatePKc($L,445428)|0); //@line 265 "modules/love/love.cpp"
 _lua_pushstring($L,180253); //@line 268 "modules/love/love.cpp"
 _lua_setfield($L,-2,180272); //@line 269 "modules/love/love.cpp"
 _lua_pushnumber($L,0.0); //@line 271 "modules/love/love.cpp"
 _lua_setfield($L,-2,180281); //@line 272 "modules/love/love.cpp"
 _lua_pushnumber($L,10.0); //@line 273 "modules/love/love.cpp"
 _lua_setfield($L,-2,180296); //@line 274 "modules/love/love.cpp"
 _lua_pushnumber($L,1.0); //@line 275 "modules/love/love.cpp"
 _lua_setfield($L,-2,180311); //@line 276 "modules/love/love.cpp"
 _lua_pushstring($L,180260); //@line 278 "modules/love/love.cpp"
 _lua_setfield($L,-2,180329); //@line 279 "modules/love/love.cpp"
 _lua_pushcclosure($L,3353,0); //@line 291 "modules/love/love.cpp"
 _lua_setfield($L,-2,180347); //@line 292 "modules/love/love.cpp"
 _lua_createtable($L,0,0); //@line 294 "modules/love/love.cpp"
 _lua_pushstring($L,180253); //@line 298 "modules/love/love.cpp"
 _lua_rawseti($L,-2,1); //@line 299 "modules/love/love.cpp"
 _lua_pushstring($L,180364); //@line 298 "modules/love/love.cpp"
 _lua_rawseti($L,-2,2); //@line 299 "modules/love/love.cpp"
 _lua_setfield($L,-2,180371); //@line 302 "modules/love/love.cpp"
 _lua_pushcclosure($L,3354,0); //@line 304 "modules/love/love.cpp"
 _lua_setfield($L,-2,180387); //@line 305 "modules/love/love.cpp"
 _lua_pushcclosure($L,3355,0); //@line 307 "modules/love/love.cpp"
 _lua_setfield($L,-2,180533); //@line 308 "modules/love/love.cpp"
 _lua_pushstring($L,359672); //@line 311 "modules/love/love.cpp"
 _lua_setfield($L,-2,180553); //@line 325 "modules/love/love.cpp"
 $i1$01 = 0;
 while(1) {
  $0 = (9592 + ($i1$01<<3)|0); //@line 328 "modules/love/love.cpp"
  $1 = HEAP32[$0>>2]|0; //@line 328 "modules/love/love.cpp"
  $2 = (((9592 + ($i1$01<<3)|0)) + 4|0); //@line 329 "modules/love/love.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 329 "modules/love/love.cpp"
  (__ZN4love12luax_preloadEP9lua_StatePFiS1_EPKc($L,$3,$1)|0); //@line 329 "modules/love/love.cpp"
  $4 = (($i1$01) + 1)|0; //@line 328 "modules/love/love.cpp"
  $5 = ($4|0)==(20); //@line 328 "modules/love/love.cpp"
  if ($5) {
   break;
  } else {
   $i1$01 = $4;
  }
 }
 (__ZN4love11w_Data_openEP9lua_State($L)|0); //@line 332 "modules/love/love.cpp"
 (__ZN4love9luasocket6__openEP9lua_State($L)|0); //@line 335 "modules/love/love.cpp"
 (__ZN4love12luax_preloadEP9lua_StatePFiS1_EPKc($L,3356,267475)|0); //@line 338 "modules/love/love.cpp"
 (__ZN4love12luax_preloadEP9lua_StatePFiS1_EPKc($L,3357,180557)|0); //@line 341 "modules/love/love.cpp"
 return 1; //@line 344 "modules/love/love.cpp"
}
function __ZL18w__setGammaCorrectP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_toboolean($L,1)|0); //@line 256 "modules/love/love.cpp"
 $1 = ($0|0)!=(0); //@line 256 "modules/love/love.cpp"
 __ZN4love8graphics15setGammaCorrectEb($1); //@line 256 "modules/love/love.cpp"
 return 0; //@line 258 "modules/love/love.cpp"
}
function __ZL17w_love_getVersionP9lua_State($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _lua_pushinteger($L,0); //@line 213 "modules/love/love.cpp"
 _lua_pushinteger($L,10); //@line 214 "modules/love/love.cpp"
 _lua_pushinteger($L,1); //@line 215 "modules/love/love.cpp"
 _lua_pushstring($L,180260); //@line 216 "modules/love/love.cpp"
 return 4; //@line 217 "modules/love/love.cpp"
}
function __ZL26w_love_isVersionCompatibleP9lua_State($L) {
 $L = $L|0;
 var $$0$i = 0, $$01$i = 0, $$02 = 0, $$03 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0.0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $9 = 0, $eh$lpad$body$i$index7Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $ss = 0, $version = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 172|0;
 $version = sp + 160|0;
 $ss = sp + 16|0;
 $1 = sp;
 ;HEAP32[$version>>2]=0|0;HEAP32[$version+4>>2]=0|0;HEAP32[$version+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 $2 = (invoke_iii(3358,($L|0),1)|0); //@line 224 "modules/love/love.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 L1: do {
  if ($4) {
   label = 6;
  } else {
   $5 = ($2|0)==(4); //@line 224 "modules/love/love.cpp"
   L3: do {
    if ($5) {
     __THREW__ = 0;
     $6 = (invoke_iiii(3359,($L|0),1,(0|0))|0); //@line 225 "modules/love/love.cpp"
     $7 = __THREW__; __THREW__ = 0;
     $8 = $7&1;
     if ($8) {
      label = 6;
      break L1;
     }
     __THREW__ = 0;
     (invoke_iii(3360,($version|0),($6|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $9 = __THREW__; __THREW__ = 0;
     $10 = $9&1;
     if ($10) {
      label = 6;
      break L1;
     }
    } else {
     __THREW__ = 0;
     $15 = (+invoke_dii(3361,($L|0),1)); //@line 228 "modules/love/love.cpp"
     $16 = __THREW__; __THREW__ = 0;
     $17 = $16&1;
     if ($17) {
      label = 6;
      break L1;
     }
     $18 = (~~(($15))); //@line 228 "modules/love/love.cpp"
     __THREW__ = 0;
     $19 = (+invoke_dii(3361,($L|0),2)); //@line 229 "modules/love/love.cpp"
     $20 = __THREW__; __THREW__ = 0;
     $21 = $20&1;
     if ($21) {
      label = 6;
      break L1;
     }
     $22 = (~~(($19))); //@line 229 "modules/love/love.cpp"
     __THREW__ = 0;
     $23 = (+invoke_dii(3361,($L|0),3)); //@line 230 "modules/love/love.cpp"
     $24 = __THREW__; __THREW__ = 0;
     $25 = $24&1;
     if ($25) {
      label = 6;
      break L1;
     }
     $26 = (~~(($23))); //@line 230 "modules/love/love.cpp"
     $27 = ((($ss)) + 64|0); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     $28 = ((($ss)) + 8|0); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     HEAP32[$28>>2] = (9320); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     $29 = ((($ss)) + 12|0); //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     HEAP32[$ss>>2] = ((9524)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
     HEAP32[$27>>2] = ((9544)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
     $30 = ((($ss)) + 4|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
     HEAP32[$30>>2] = 0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
     $31 = ((($ss)) + 64|0); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     __THREW__ = 0;
     invoke_vii(3362,($31|0),($29|0)); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     $32 = __THREW__; __THREW__ = 0;
     $33 = $32&1;
     if ($33) {
      $44 = ___cxa_find_matching_catch_2()|0;
      $45 = tempRet0;
      $$0$i = $45;$$01$i = $44;
     } else {
      $34 = ((($ss)) + 136|0); //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
      HEAP32[$34>>2] = 0; //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
      $35 = ((($ss)) + 140|0); //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
      HEAP32[$35>>2] = -1; //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
      HEAP32[$ss>>2] = (9300); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
      HEAP32[$27>>2] = (9340); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
      HEAP32[$28>>2] = (9320); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
      __THREW__ = 0;
      invoke_vi(3363,($29|0)); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
      $36 = __THREW__; __THREW__ = 0;
      $37 = $36&1;
      do {
       if ($37) {
        $46 = ___cxa_find_matching_catch_2()|0;
        $47 = tempRet0;
        $eh$lpad$body$i$index7Z2D = $47;$eh$lpad$body$i$indexZ2D = $46;
       } else {
        HEAP32[$29>>2] = (9356); //@line 243 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
        $38 = ((($ss)) + 44|0); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
        $39 = ((($ss)) + 60|0); //@line 242 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
        ;HEAP32[$38>>2]=0|0;HEAP32[$38+4>>2]=0|0;HEAP32[$38+8>>2]=0|0;HEAP32[$38+12>>2]=0|0; //@line 241 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
        HEAP32[$39>>2] = 24; //@line 242 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
        ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        __THREW__ = 0;
        invoke_vii(3364,($29|0),($0|0)); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
        $40 = __THREW__; __THREW__ = 0;
        $41 = $40&1;
        if ($41) {
         $42 = ___cxa_find_matching_catch_2()|0;
         $43 = tempRet0;
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($38); //@line 245 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
         __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($29); //@line 245 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
         $eh$lpad$body$i$index7Z2D = $43;$eh$lpad$body$i$indexZ2D = $42;
         break;
        }
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
        $48 = ((($ss)) + 8|0); //@line 235 "modules/love/love.cpp"
        __THREW__ = 0;
        $49 = (invoke_iii(3365,($48|0),($18|0))|0); //@line 235 "modules/love/love.cpp"
        $50 = __THREW__; __THREW__ = 0;
        $51 = $50&1;
        if (!($51)) {
         __THREW__ = 0;
         $52 = (invoke_iiii(3366,($49|0),(619449|0),1)|0); //@line 894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
         $53 = __THREW__; __THREW__ = 0;
         $54 = $53&1;
         if (!($54)) {
          __THREW__ = 0;
          $55 = (invoke_iii(3365,($52|0),($22|0))|0); //@line 235 "modules/love/love.cpp"
          $56 = __THREW__; __THREW__ = 0;
          $57 = $56&1;
          if (!($57)) {
           __THREW__ = 0;
           $58 = (invoke_iiii(3366,($55|0),(619449|0),1)|0); //@line 894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
           $59 = __THREW__; __THREW__ = 0;
           $60 = $59&1;
           if (!($60)) {
            __THREW__ = 0;
            (invoke_iii(3365,($58|0),($26|0))|0); //@line 235 "modules/love/love.cpp"
            $61 = __THREW__; __THREW__ = 0;
            $62 = $61&1;
            if (!($62)) {
             __THREW__ = 0;
             invoke_vii(3367,($1|0),($29|0)); //@line 964 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
             $63 = __THREW__; __THREW__ = 0;
             $64 = $63&1;
             if (!($64)) {
              $65 = HEAP8[$version>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              $66 = $65 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              $67 = ($66<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              if ($67) {
               $71 = ((($version)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               HEAP8[$71>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               HEAP8[$version>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              } else {
               $68 = ((($version)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $69 = HEAP32[$68>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               HEAP8[$69>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $70 = ((($version)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               HEAP32[$70>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              }
              __THREW__ = 0;
              invoke_vii(3368,($version|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              $72 = __THREW__; __THREW__ = 0;
              $73 = $72&1;
              if ($73) {
               $74 = ___cxa_find_matching_catch_3(0|0)|0;
               $75 = tempRet0;
               ___clang_call_terminate($74); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              } else {
               ;HEAP32[$version>>2]=HEAP32[$1>>2]|0;HEAP32[$version+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$version+8>>2]=HEAP32[$1+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 237 "modules/love/love.cpp"
               HEAP32[$ss>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
               HEAP32[$27>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
               HEAP32[$48>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
               HEAP32[$29>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
               __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($38); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
               __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($29); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
               __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($27); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
               break L3;
              }
             }
            }
           }
          }
         }
        }
        $76 = ___cxa_find_matching_catch_2()|0;
        $77 = tempRet0;
        HEAP32[$ss>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$27>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$48>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$29>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($38); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($29); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($27); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        $$02 = $76;$$03 = $77;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($version); //@line 251 "modules/love/love.cpp"
        ___resumeException($$02|0);
        // unreachable;
       }
      } while(0);
      $$0$i = $eh$lpad$body$i$index7Z2D;$$01$i = $eh$lpad$body$i$indexZ2D;
     }
     __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($27); //@line 900 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     $eh$lpad$body$index2Z2D = $$0$i;$eh$lpad$body$indexZ2D = $$01$i;
     break L1;
    }
   } while(0);
   $11 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7compareEPKc($version,180253)|0); //@line 242 "modules/love/love.cpp"
   $12 = ($11|0)==(0); //@line 242 "modules/love/love.cpp"
   if (!($12)) {
    $78 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7compareEPKc($version,180364)|0); //@line 242 "modules/love/love.cpp"
    $79 = ($78|0)==(0); //@line 242 "modules/love/love.cpp"
    if (!($79)) {
     __THREW__ = 0;
     invoke_vii(3369,($L|0),0); //@line 249 "modules/love/love.cpp"
     $82 = __THREW__; __THREW__ = 0;
     $83 = $82&1;
     if ($83) {
      label = 6;
      break;
     }
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($version); //@line 251 "modules/love/love.cpp"
     STACKTOP = sp;return 1; //@line 251 "modules/love/love.cpp"
    }
   }
   __THREW__ = 0;
   invoke_vii(3369,($L|0),1); //@line 245 "modules/love/love.cpp"
   $80 = __THREW__; __THREW__ = 0;
   $81 = $80&1;
   if ($81) {
    label = 6;
   } else {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($version); //@line 251 "modules/love/love.cpp"
    STACKTOP = sp;return 1; //@line 251 "modules/love/love.cpp"
   }
  }
 } while(0);
 if ((label|0) == 6) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  $eh$lpad$body$index2Z2D = $14;$eh$lpad$body$indexZ2D = $13;
 }
 $$02 = $eh$lpad$body$indexZ2D;$$03 = $eh$lpad$body$index2Z2D;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($version); //@line 251 "modules/love/love.cpp"
 ___resumeException($$02|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__118basic_stringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED1Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $0 = ((($this)) + 64|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$0>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $1 = ((($this)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$1>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $2 = ((($this)) + 12|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$2>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $3 = ((($this)) + 44|0); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($2); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 return; //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
}
function __ZNSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $0 = ((($this)) + 32|0); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($this); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 return; //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
}
function __ZNSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $0 = ((($this)) + 32|0); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($this); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZdlPv($this); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 return; //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
}
function __ZNSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEE7seekoffExNS_8ios_base7seekdirEj($agg$result,$this,$0,$1,$__way,$__wch) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $0 = $0|0;
 $1 = $1|0;
 $__way = $__way|0;
 $__wch = $__wch|0;
 var $$pre = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, $or$cond3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($this)) + 44|0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $3 = HEAP32[$2>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $4 = ((($this)) + 24|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $5 = HEAP32[$4>>2]|0; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $6 = ($3>>>0)<($5>>>0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $7 = $3; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $8 = $5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 if ($6) {
  HEAP32[$2>>2] = $5; //@line 565 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $55 = $8;
 } else {
  $55 = $7;
 }
 $9 = $__wch & 24; //@line 566 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $10 = ($9|0)==(0); //@line 566 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 if ($10) {
  $11 = $agg$result; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $12 = $11; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$12>>2] = 0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $13 = (($11) + 4)|0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $14 = $13; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$14>>2] = 0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $15 = ((($agg$result)) + 8|0); //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $16 = $15; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $17 = $16; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$17>>2] = -1; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $18 = (($16) + 4)|0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $19 = $18; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$19>>2] = -1; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  return; //@line 607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 }
 $20 = ($9|0)==(24); //@line 568 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $21 = ($__way|0)==(1); //@line 569 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $or$cond = $21 & $20; //@line 569 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 if ($or$cond) {
  $22 = $agg$result; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $23 = $22; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$23>>2] = 0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $24 = (($22) + 4)|0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $25 = $24; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$25>>2] = 0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $26 = ((($agg$result)) + 8|0); //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $27 = $26; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $28 = $27; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$28>>2] = -1; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $29 = (($27) + 4)|0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $30 = $29; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$30>>2] = -1; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  return; //@line 607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 }
 L12: do {
  switch ($__way|0) {
  case 0:  {
   $67 = 0;$68 = 0;
   break;
  }
  case 1:  {
   $31 = $__wch & 8; //@line 578 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $32 = ($31|0)==(0); //@line 578 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   if ($32) {
    $40 = ((($this)) + 20|0); //@line 176 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
    $41 = HEAP32[$40>>2]|0; //@line 176 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
    $42 = (($8) - ($41))|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $43 = ($42|0)<(0); //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $44 = $43 << 31 >> 31; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $67 = $42;$68 = $44;
    break L12;
   } else {
    $33 = ((($this)) + 12|0); //@line 170 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
    $34 = HEAP32[$33>>2]|0; //@line 170 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
    $35 = ((($this)) + 8|0); //@line 169 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
    $36 = HEAP32[$35>>2]|0; //@line 169 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
    $37 = (($34) - ($36))|0; //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $38 = ($37|0)<(0); //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $39 = $38 << 31 >> 31; //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $67 = $37;$68 = $39;
    break L12;
   }
   break;
  }
  case 2:  {
   $45 = ((($this)) + 32|0); //@line 584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $46 = HEAP8[$45>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $47 = $46 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $48 = ($47<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if ($48) {
    $51 = ((($45)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $53 = $51;
   } else {
    $49 = ((($this)) + 40|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $50 = HEAP32[$49>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $53 = $50;
   }
   $52 = $53; //@line 584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $54 = (($55) - ($52))|0; //@line 584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $56 = ($54|0)<(0); //@line 584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $57 = $56 << 31 >> 31; //@line 584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $67 = $54;$68 = $57;
   break;
  }
  default: {
   $58 = $agg$result; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $59 = $58; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP32[$59>>2] = 0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $60 = (($58) + 4)|0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $61 = $60; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP32[$61>>2] = 0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $62 = ((($agg$result)) + 8|0); //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $63 = $62; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $64 = $63; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP32[$64>>2] = -1; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $65 = (($63) + 4)|0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $66 = $65; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP32[$66>>2] = -1; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   return; //@line 607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  }
  }
 } while(0);
 $69 = (_i64Add(($67|0),($68|0),($0|0),($1|0))|0); //@line 589 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $70 = tempRet0; //@line 589 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $71 = ($70|0)<(0); //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 if (!($71)) {
  $72 = ((($this)) + 32|0); //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $73 = HEAP8[$72>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $74 = $73 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $75 = ($74<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($75) {
   $78 = ((($72)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $80 = $78;
  } else {
   $76 = ((($this)) + 40|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $77 = HEAP32[$76>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $80 = $77;
  }
  $79 = $80; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $81 = (($55) - ($79))|0; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $82 = ($81|0)<(0); //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $83 = $82 << 31 >> 31; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $84 = ($83|0)<($70|0); //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $85 = ($81>>>0)<($69>>>0); //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $86 = ($83|0)==($70|0); //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $87 = $86 & $85; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $88 = $84 | $87; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  if (!($88)) {
   $98 = ($69|0)==(0); //@line 592 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $99 = ($70|0)==(0); //@line 592 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $100 = $98 & $99; //@line 592 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $$pre = $__wch & 8; //@line 599 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   if (!($100)) {
    $101 = ($$pre|0)==(0); //@line 594 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    if (!($101)) {
     $102 = ((($this)) + 12|0); //@line 170 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
     $103 = HEAP32[$102>>2]|0; //@line 170 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
     $104 = ($103|0)==(0|0); //@line 594 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     if ($104) {
      $105 = $agg$result; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $106 = $105; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP32[$106>>2] = 0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $107 = (($105) + 4)|0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $108 = $107; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP32[$108>>2] = 0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $109 = ((($agg$result)) + 8|0); //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $110 = $109; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $111 = $110; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP32[$111>>2] = -1; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $112 = (($110) + 4)|0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $113 = $112; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP32[$113>>2] = -1; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      return; //@line 607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     }
    }
    $114 = $__wch & 16; //@line 596 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $115 = ($114|0)!=(0); //@line 596 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $116 = ($5|0)==(0|0); //@line 596 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $or$cond3 = $115 & $116; //@line 596 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    if ($or$cond3) {
     $117 = $agg$result; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $118 = $117; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP32[$118>>2] = 0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $119 = (($117) + 4)|0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $120 = $119; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP32[$120>>2] = 0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $121 = ((($agg$result)) + 8|0); //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $122 = $121; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $123 = $122; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP32[$123>>2] = -1; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $124 = (($122) + 4)|0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $125 = $124; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP32[$125>>2] = -1; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     return; //@line 607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    }
   }
   $126 = ($$pre|0)==(0); //@line 599 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   if (!($126)) {
    $127 = ((($this)) + 8|0); //@line 169 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
    $128 = HEAP32[$127>>2]|0; //@line 169 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
    $129 = (($128) + ($69)|0); //@line 600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $130 = ((($this)) + 12|0); //@line 428 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
    HEAP32[$130>>2] = $129; //@line 428 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
    $131 = ((($this)) + 16|0); //@line 429 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
    HEAP32[$131>>2] = $55; //@line 429 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
   }
   $132 = $__wch & 16; //@line 601 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $133 = ($132|0)==(0); //@line 601 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   if (!($133)) {
    $134 = ((($this)) + 20|0); //@line 176 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
    $135 = HEAP32[$134>>2]|0; //@line 176 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
    $136 = (($135) + ($69)|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
    HEAP32[$4>>2] = $136; //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
   }
   $137 = $agg$result; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $138 = $137; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP32[$138>>2] = 0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $139 = (($137) + 4)|0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $140 = $139; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP32[$140>>2] = 0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $141 = ((($agg$result)) + 8|0); //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $142 = $141; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $143 = $142; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP32[$143>>2] = $69; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $144 = (($142) + 4)|0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $145 = $144; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP32[$145>>2] = $70; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   return; //@line 607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  }
 }
 $89 = $agg$result; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $90 = $89; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 HEAP32[$90>>2] = 0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $91 = (($89) + 4)|0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $92 = $91; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 HEAP32[$92>>2] = 0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $93 = ((($agg$result)) + 8|0); //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $94 = $93; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $95 = $94; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 HEAP32[$95>>2] = -1; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $96 = (($94) + 4)|0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $97 = $96; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 HEAP32[$97>>2] = -1; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 return; //@line 607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
}
function __ZNSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEE7seekposENS_4fposI11__mbstate_tEEj($agg$result,$this,$__sp,$__wch) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $__sp = $__sp|0;
 $__wch = $__wch|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $1 = ((($0)) + 16|0); //@line 615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $2 = HEAP32[$1>>2]|0; //@line 615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $3 = ((($__sp)) + 8|0); //@line 475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = $3; //@line 475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = $4; //@line 475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = HEAP32[$5>>2]|0; //@line 475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = (($4) + 4)|0; //@line 475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $7; //@line 475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = HEAP32[$8>>2]|0; //@line 475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 FUNCTION_TABLE_viiiiii[$2 & 8191]($agg$result,$this,$6,$9,0,$__wch); //@line 615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 return; //@line 615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
}
function __ZNSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEE9underflowEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 44|0); //@line 480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $1 = HEAP32[$0>>2]|0; //@line 480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $2 = ((($this)) + 24|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $3 = HEAP32[$2>>2]|0; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $4 = ($1>>>0)<($3>>>0); //@line 480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 if ($4) {
  HEAP32[$0>>2] = $3; //@line 481 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $12 = $3;
 } else {
  $12 = $1;
 }
 $5 = ((($this)) + 48|0); //@line 482 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $6 = HEAP32[$5>>2]|0; //@line 482 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $7 = $6 & 8; //@line 482 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $8 = ($7|0)==(0); //@line 482 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 if ($8) {
  $$0 = -1;
  return ($$0|0); //@line 490 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 }
 $9 = ((($this)) + 16|0); //@line 171 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $10 = HEAP32[$9>>2]|0; //@line 171 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $11 = ($10>>>0)<($12>>>0); //@line 484 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 if ($11) {
  HEAP32[$9>>2] = $12; //@line 429 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $16 = $12;
 } else {
  $16 = $10;
 }
 $13 = ((($this)) + 12|0); //@line 170 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $14 = HEAP32[$13>>2]|0; //@line 170 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $15 = ($14>>>0)<($16>>>0); //@line 486 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 if (!($15)) {
  $$0 = -1;
  return ($$0|0); //@line 490 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 }
 $17 = HEAP8[$14>>0]|0; //@line 487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $18 = $17&255; //@line 658 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $$0 = $18;
 return ($$0|0); //@line 490 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
}
function __ZNSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEE9pbackfailEi($this,$__c) {
 $this = $this|0;
 $__c = $__c|0;
 var $$0 = 0, $$in = 0, $$pre = 0, $$pre$phi2Z2D = 0, $$pre$phiZ2D = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 44|0); //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $1 = HEAP32[$0>>2]|0; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $2 = ((($this)) + 24|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $3 = HEAP32[$2>>2]|0; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $4 = ($1>>>0)<($3>>>0); //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 if ($4) {
  HEAP32[$0>>2] = $3; //@line 497 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $$in = $3;
 } else {
  $$in = $1;
 }
 $5 = $$in; //@line 502 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $6 = ((($this)) + 8|0); //@line 169 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $7 = HEAP32[$6>>2]|0; //@line 169 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $8 = ((($this)) + 12|0); //@line 170 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $9 = HEAP32[$8>>2]|0; //@line 170 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $10 = ($7>>>0)<($9>>>0); //@line 498 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 if (!($10)) {
  $$0 = -1;
  return ($$0|0); //@line 514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 }
 $11 = ($__c|0)==(-1); //@line 660 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($11) {
  $12 = ((($9)) + -1|0); //@line 502 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  HEAP32[$8>>2] = $12; //@line 428 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $13 = ((($this)) + 16|0); //@line 429 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  HEAP32[$13>>2] = $5; //@line 429 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $$0 = 0;
  return ($$0|0); //@line 514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 }
 $14 = ((($this)) + 48|0); //@line 505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $15 = HEAP32[$14>>2]|0; //@line 505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $16 = $15 & 16; //@line 505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $17 = ($16|0)==(0); //@line 505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 if ($17) {
  $18 = $__c&255; //@line 656 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $19 = ((($9)) + -1|0); //@line 506 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $20 = HEAP8[$19>>0]|0; //@line 506 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $21 = ($18<<24>>24)==($20<<24>>24); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($21) {
   $$pre$phi2Z2D = $18;$$pre$phiZ2D = $19;
  } else {
   $$0 = -1;
   return ($$0|0); //@line 514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  }
 } else {
  $$pre = ((($9)) + -1|0); //@line 508 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $$pre1 = $__c&255; //@line 656 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $$pre$phi2Z2D = $$pre1;$$pre$phiZ2D = $$pre;
 }
 HEAP32[$8>>2] = $$pre$phiZ2D; //@line 428 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $22 = ((($this)) + 16|0); //@line 429 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 HEAP32[$22>>2] = $5; //@line 429 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 HEAP8[$$pre$phiZ2D>>0] = $$pre$phi2Z2D; //@line 509 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $$0 = $__c;
 return ($$0|0); //@line 514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
}
function __ZNSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEE8overflowEi($this,$__c) {
 $this = $this|0;
 $__c = $__c|0;
 var $$1 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phi3Z2D = 0, $$pre$phi5Z2D = 0, $$pre$phiZ2D = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = ($__c|0)==(-1); //@line 660 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($1) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 556 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 }
 $2 = ((($this)) + 12|0); //@line 170 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $3 = HEAP32[$2>>2]|0; //@line 170 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $4 = ((($this)) + 8|0); //@line 169 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $5 = HEAP32[$4>>2]|0; //@line 169 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $6 = (($3) - ($5))|0; //@line 522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $7 = ((($this)) + 24|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $8 = HEAP32[$7>>2]|0; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $9 = ((($this)) + 28|0); //@line 178 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $10 = HEAP32[$9>>2]|0; //@line 178 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $11 = ($8|0)==($10|0); //@line 523 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 do {
  if ($11) {
   $12 = ((($this)) + 48|0); //@line 525 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $13 = HEAP32[$12>>2]|0; //@line 525 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $14 = $13 & 16; //@line 525 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $15 = ($14|0)==(0); //@line 525 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   if ($15) {
    $$1 = -1;
    STACKTOP = sp;return ($$1|0); //@line 556 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   }
   $16 = ((($this)) + 20|0); //@line 176 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
   $17 = HEAP32[$16>>2]|0; //@line 176 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
   $18 = $8; //@line 531 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $19 = (($18) - ($17))|0; //@line 531 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $20 = ((($this)) + 44|0); //@line 532 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $21 = HEAP32[$20>>2]|0; //@line 532 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $22 = (($21) - ($17))|0; //@line 532 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $23 = ((($this)) + 32|0); //@line 533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   __THREW__ = 0;
   invoke_vii(3370,($23|0),0); //@line 533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $24 = __THREW__; __THREW__ = 0;
   $25 = $24&1;
   if (!($25)) {
    $26 = HEAP8[$23>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $27 = $26 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $28 = ($27<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if ($28) {
     $31 = 10;
    } else {
     $29 = HEAP32[$23>>2]|0; //@line 1736 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $30 = $29 & -2; //@line 1736 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $phitmp$i = (($30) + -1)|0; //@line 1435 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $31 = $phitmp$i;
    }
    __THREW__ = 0;
    invoke_viii(3371,($23|0),($31|0),0); //@line 1438 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $32 = __THREW__; __THREW__ = 0;
    $33 = $32&1;
    if (!($33)) {
     $34 = HEAP8[$23>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $35 = $34 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $36 = ($35<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($36) {
      $41 = ((($23)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $42 = $34&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $43 = $42 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $45 = $41;$46 = $43;
     } else {
      $37 = ((($this)) + 40|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $38 = HEAP32[$37>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $39 = ((($this)) + 36|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $40 = HEAP32[$39>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $45 = $38;$46 = $40;
     }
     $44 = (($45) + ($46)|0); //@line 536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     HEAP32[$16>>2] = $45; //@line 445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
     HEAP32[$9>>2] = $44; //@line 446 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
     $47 = (($45) + ($19)|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
     HEAP32[$7>>2] = $47; //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
     $48 = (($45) + ($22)|0); //@line 538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     HEAP32[$20>>2] = $48; //@line 538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     $$pre$phi3Z2D = $20;$$pre$phi5Z2D = $12;$$pre$phiZ2D = $20;$52 = $47;$54 = $48;$71 = $44;
     break;
    }
   }
   $49 = ___cxa_find_matching_catch_3(0|0)|0;
   $50 = tempRet0;
   (___cxa_begin_catch(($49|0))|0); //@line 540 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   ___cxa_end_catch(); //@line 544 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $$1 = -1;
   STACKTOP = sp;return ($$1|0); //@line 556 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  } else {
   $$phi$trans$insert = ((($this)) + 44|0); //@line 547 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $$pre4 = ((($this)) + 48|0); //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $$pre$phi3Z2D = $$phi$trans$insert;$$pre$phi5Z2D = $$pre4;$$pre$phiZ2D = $$phi$trans$insert;$52 = $8;$54 = $$pre;$71 = $10;
  }
 } while(0);
 $51 = ((($52)) + 1|0); //@line 547 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$0>>2] = $51; //@line 547 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $53 = ($51>>>0)<($54>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $55 = $53 ? $$pre$phiZ2D : $0; //@line 2648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $56 = HEAP32[$55>>2]|0; //@line 551 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$$pre$phi3Z2D>>2] = $56; //@line 547 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $57 = HEAP32[$$pre$phi5Z2D>>2]|0; //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $58 = $57 & 8; //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $59 = ($58|0)==(0); //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 if (!($59)) {
  $60 = ((($this)) + 32|0); //@line 550 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $61 = HEAP8[$60>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $62 = $61 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $63 = ($62<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($63) {
   $66 = ((($60)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $68 = $66;
  } else {
   $64 = ((($this)) + 40|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $65 = HEAP32[$64>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $68 = $65;
  }
  $67 = (($68) + ($6)|0); //@line 551 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  HEAP32[$4>>2] = $68; //@line 427 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  HEAP32[$2>>2] = $67; //@line 428 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $69 = ((($this)) + 16|0); //@line 429 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  HEAP32[$69>>2] = $56; //@line 429 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 }
 $70 = ($52|0)==($71|0); //@line 349 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 if ($70) {
  $72 = HEAP32[$this>>2]|0; //@line 350 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $73 = ((($72)) + 52|0); //@line 350 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $74 = HEAP32[$73>>2]|0; //@line 350 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $75 = $__c & 255; //@line 658 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $76 = (FUNCTION_TABLE_iii[$74 & 8191]($this,$75)|0); //@line 350 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $$1 = $76;
  STACKTOP = sp;return ($$1|0); //@line 556 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 } else {
  $77 = $__c&255; //@line 553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  HEAP32[$7>>2] = $51; //@line 351 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  HEAP8[$52>>0] = $77; //@line 351 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $78 = $__c & 255; //@line 658 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $$1 = $78;
  STACKTOP = sp;return ($$1|0); //@line 556 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 }
 return (0)|0;
}
function __ZNSt3__118basic_stringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $0 = ((($this)) + 64|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$0>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $1 = ((($this)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$1>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $2 = ((($this)) + 12|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$2>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $3 = ((($this)) + 44|0); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($2); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZdlPv($this); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 return; //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
}
function __ZThn8_NSt3__118basic_stringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED1Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$0>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $1 = ((($0)) + 64|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$1>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $2 = ((($0)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$2>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $3 = ((($0)) + 12|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$3>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $4 = ((($0)) + 44|0); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($3); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($1); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 return; //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
}
function __ZThn8_NSt3__118basic_stringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$0>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $1 = ((($0)) + 64|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$1>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $2 = ((($0)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$2>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $3 = ((($0)) + 12|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$3>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $4 = ((($0)) + 44|0); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($3); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($1); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZdlPv($0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 return; //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
}
function __ZTv0_n12_NSt3__118basic_stringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED1Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $1 = ((($0)) + -12|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $2 = HEAP32[$1>>2]|0; //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $3 = (($this) + ($2)|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$3>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $4 = ((($3)) + 64|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$4>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $5 = ((($3)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$5>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $6 = ((($3)) + 12|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$6>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $7 = ((($3)) + 44|0); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($6); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($4); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 return; //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
}
function __ZTv0_n12_NSt3__118basic_stringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $1 = ((($0)) + -12|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $2 = HEAP32[$1>>2]|0; //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $3 = (($this) + ($2)|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$3>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $4 = ((($3)) + 64|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$4>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $5 = ((($3)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$5>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $6 = ((($3)) + 12|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$6>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $7 = ((($3)) + 44|0); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($6); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($4); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZdlPv($3); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 return; //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
}
function __ZNSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEE3strERKNS_12basic_stringIcS2_S4_EE($this,$__s) {
 $this = $this|0;
 $__s = $__s|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 (__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEaSERKS5_($0,$__s)|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $1 = ((($this)) + 44|0); //@line 456 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$1>>2] = 0; //@line 456 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $2 = ((($this)) + 48|0); //@line 457 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $3 = HEAP32[$2>>2]|0; //@line 464 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $4 = $3 & 8; //@line 457 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $5 = ($4|0)==(0); //@line 457 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 if (!($5)) {
  $6 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = $6 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $8 = ($7<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($8) {
   $14 = $6&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $15 = $14 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $16 = (((($0)) + 1|0) + ($15)|0); //@line 459 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   HEAP32[$1>>2] = $16; //@line 459 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $17 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $18 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$in = $16;$21 = $17;$23 = $18;
  } else {
   $9 = ((($this)) + 40|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $10 = HEAP32[$9>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $11 = ((($this)) + 36|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $12 = HEAP32[$11>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $13 = (($10) + ($12)|0); //@line 459 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   HEAP32[$1>>2] = $13; //@line 459 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $$in = $13;$21 = $10;$23 = $10;
  }
  $19 = $$in; //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $20 = ((($this)) + 8|0); //@line 427 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  HEAP32[$20>>2] = $21; //@line 427 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $22 = ((($this)) + 12|0); //@line 428 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  HEAP32[$22>>2] = $23; //@line 428 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $24 = ((($this)) + 16|0); //@line 429 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  HEAP32[$24>>2] = $19; //@line 429 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 }
 $25 = $3 & 16; //@line 464 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $26 = ($25|0)==(0); //@line 464 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 if ($26) {
  return; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 }
 $27 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $28 = $27 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $29 = ($28<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($29) {
  $30 = $27&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $31 = $30 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $32 = (((($0)) + 1|0) + ($31)|0); //@line 467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  HEAP32[$1>>2] = $32; //@line 467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $40 = 10;$63 = $31;
 } else {
  $33 = ((($this)) + 36|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $34 = HEAP32[$33>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $35 = ((($this)) + 40|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $36 = HEAP32[$35>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $37 = (($36) + ($34)|0); //@line 467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  HEAP32[$1>>2] = $37; //@line 467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $38 = HEAP32[$0>>2]|0; //@line 1736 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $39 = $38 & -2; //@line 1736 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $phitmp$i = (($39) + -1)|0; //@line 1435 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $40 = $phitmp$i;$63 = $34;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6resizeEjc($0,$40,0); //@line 1438 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $41 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $42 = $41 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $43 = ($42<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($43) {
  $48 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $49 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $50 = $41&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $51 = $50 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $53 = $49;$54 = $51;$56 = $48;
 } else {
  $44 = ((($this)) + 40|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $45 = HEAP32[$44>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $46 = ((($this)) + 36|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $47 = HEAP32[$46>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $53 = $45;$54 = $47;$56 = $45;
 }
 $52 = (($53) + ($54)|0); //@line 470 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $55 = ((($this)) + 24|0); //@line 445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 HEAP32[$55>>2] = $56; //@line 445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $57 = ((($this)) + 20|0); //@line 445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 HEAP32[$57>>2] = $56; //@line 445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $58 = ((($this)) + 28|0); //@line 446 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 HEAP32[$58>>2] = $52; //@line 446 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $59 = HEAP32[$2>>2]|0; //@line 471 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $60 = $59 & 3; //@line 471 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $61 = ($60|0)==(0); //@line 471 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 if ($61) {
  return; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 }
 $62 = (($56) + ($63)|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 HEAP32[$55>>2] = $62; //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 return; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
}
function __ZNSt3__124__put_character_sequenceIcNS_11char_traitsIcEEEERNS_13basic_ostreamIT_T0_EES7_PKS4_j($__os,$__str,$__len) {
 $__os = $__os|0;
 $__str = $__str|0;
 $__len = $__len|0;
 var $$0 = 0, $$byval_copy = 0, $$pre$phi2Z2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, $__s = 0, $eh$lpad$body$indexZ2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 20|0;
 $0 = sp + 16|0;
 $__s = sp + 8|0;
 $1 = sp;
 __THREW__ = 0;
 invoke_vii(3372,($__s|0),($__os|0)); //@line 751 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 L1: do {
  if ($3) {
   $51 = ___cxa_find_matching_catch_3(0|0)|0;
   $52 = tempRet0;
   $$0 = $51;$$pre$phi2Z2D = $__os;$$pre$phiZ2D = $__os;
  } else {
   $4 = HEAP8[$__s>>0]|0; //@line 240 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
   $5 = ($4<<24>>24)==(0); //@line 240 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
   do {
    if (!($5)) {
     $6 = HEAP32[$__os>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $7 = ((($6)) + -12|0); //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $8 = HEAP32[$7>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $9 = (($__os) + ($8)|0); //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $10 = ((($9)) + 24|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     $11 = HEAP32[$10>>2]|0; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     HEAP32[$1>>2] = $11; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $12 = ((($9)) + 4|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     $13 = HEAP32[$12>>2]|0; //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     $14 = (($__str) + ($__len)|0); //@line 758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
     $15 = ((($9)) + 76|0); //@line 742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     $16 = HEAP32[$15>>2]|0; //@line 742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     $17 = ($16|0)==(-1); //@line 660 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     do {
      if ($17) {
       __THREW__ = 0;
       $18 = (invoke_ii(3373,($9|0))|0); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
       $19 = __THREW__; __THREW__ = 0;
       $20 = $19&1;
       if ($20) {
        label = 13;
       } else {
        HEAP32[$0>>2] = $18; //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
        __THREW__ = 0;
        $21 = (invoke_iii(3374,($0|0),(1207576|0))|0); //@line 181 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
        $22 = __THREW__; __THREW__ = 0;
        $23 = $22&1;
        if (!($23)) {
         $24 = HEAP32[$21>>2]|0; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
         $25 = ((($24)) + 28|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
         $26 = HEAP32[$25>>2]|0; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
         __THREW__ = 0;
         $27 = (invoke_iii($26|0,($21|0),32)|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
         $28 = __THREW__; __THREW__ = 0;
         $29 = $28&1;
         if (!($29)) {
          __ZNSt3__16localeD2Ev($0); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
          $32 = $27 << 24 >> 24; //@line 743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
          HEAP32[$15>>2] = $32; //@line 743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
          $34 = $32;
          label = 9;
          break;
         }
        }
        $30 = ___cxa_find_matching_catch_3(0|0)|0;
        $31 = tempRet0;
        __ZNSt3__16localeD2Ev($0); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
        $eh$lpad$body$indexZ2D = $30;
       }
      } else {
       $34 = $16;
       label = 9;
      }
     } while(0);
     if ((label|0) == 9) {
      $33 = $34&255; //@line 744 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
      $35 = $13 & 176; //@line 757 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
      $36 = ($35|0)==(32); //@line 757 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
      $37 = $36 ? $14 : $__str; //@line 757 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
      __THREW__ = 0;
      ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0; //@line 755 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
      $38 = (invoke_iiiiiii(3375,($$byval_copy|0),($__str|0),($37|0),($14|0),($9|0),($33|0))|0); //@line 755 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
      $39 = __THREW__; __THREW__ = 0;
      $40 = $39&1;
      if ($40) {
       label = 13;
      } else {
       $41 = ($38|0)==(0|0); //@line 909 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       if (!($41)) {
        break;
       }
       $42 = HEAP32[$__os>>2]|0; //@line 763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
       $43 = ((($42)) + -12|0); //@line 763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
       $44 = HEAP32[$43>>2]|0; //@line 763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
       $45 = (($__os) + ($44)|0); //@line 763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
       $46 = ((($45)) + 16|0); //@line 524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
       $47 = HEAP32[$46>>2]|0; //@line 524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
       $48 = $47 | 5; //@line 524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
       __THREW__ = 0;
       invoke_vii(3376,($45|0),($48|0)); //@line 524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
       $49 = __THREW__; __THREW__ = 0;
       $50 = $49&1;
       if ($50) {
        label = 13;
       } else {
        break;
       }
      }
     }
     if ((label|0) == 13) {
      $53 = ___cxa_find_matching_catch_3(0|0)|0;
      $54 = tempRet0;
      $eh$lpad$body$indexZ2D = $53;
     }
     __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE6sentryD2Ev($__s); //@line 766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
     $$0 = $eh$lpad$body$indexZ2D;$$pre$phi2Z2D = $__os;$$pre$phiZ2D = $__os;
     break L1;
    }
   } while(0);
   __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE6sentryD2Ev($__s); //@line 766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
   STACKTOP = sp;return ($__os|0); //@line 772 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
  }
 } while(0);
 (___cxa_begin_catch(($$0|0))|0); //@line 766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
 $55 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 769 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
 $56 = ((($55)) + -12|0); //@line 769 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
 $57 = HEAP32[$56>>2]|0; //@line 769 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
 $58 = (($$pre$phi2Z2D) + ($57)|0); //@line 769 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
 __THREW__ = 0;
 invoke_vi(3377,($58|0)); //@line 769 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
 $59 = __THREW__; __THREW__ = 0;
 $60 = $59&1;
 if (!($60)) {
  ___cxa_end_catch(); //@line 770 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
  STACKTOP = sp;return ($__os|0); //@line 772 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
 }
 $61 = ___cxa_find_matching_catch_2()|0;
 $62 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 770 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
 $63 = __THREW__; __THREW__ = 0;
 $64 = $63&1;
 if ($64) {
  $65 = ___cxa_find_matching_catch_3(0|0)|0;
  $66 = tempRet0;
  ___clang_call_terminate($65); //@line 770 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
  // unreachable; //@line 770 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
 } else {
  ___resumeException($61|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZNSt3__116__pad_and_outputIcNS_11char_traitsIcEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($__s,$__ob,$__op,$__oe,$__iob,$__fl) {
 $__s = $__s|0;
 $__ob = $__ob|0;
 $__op = $__op|0;
 $__oe = $__oe|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 var $$sroa$0$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $__ns$0 = 0, $__sp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__sp = sp;
 $0 = HEAP32[$__s>>2]|0; //@line 1522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
 $1 = ($0|0)==(0|0); //@line 1486 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
 if ($1) {
  $$sroa$0$2 = 0;
  STACKTOP = sp;return ($$sroa$0$2|0); //@line 1523 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
 }
 $2 = $__oe; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
 $3 = $__ob; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
 $4 = (($2) - ($3))|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
 $5 = ((($__iob)) + 12|0); //@line 499 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 $6 = HEAP32[$5>>2]|0; //@line 499 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 $7 = ($6|0)>($4|0); //@line 1490 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
 $8 = (($6) - ($4))|0; //@line 1491 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
 $__ns$0 = $7 ? $8 : 0; //@line 1490 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
 $9 = $__op; //@line 1494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
 $10 = (($9) - ($3))|0; //@line 1494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
 $11 = ($10|0)>(0); //@line 1495 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
 if ($11) {
  $12 = HEAP32[$0>>2]|0; //@line 360 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $13 = ((($12)) + 48|0); //@line 360 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $14 = HEAP32[$13>>2]|0; //@line 360 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $15 = (FUNCTION_TABLE_iiii[$14 & 8191]($0,$__ob,$10)|0); //@line 360 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $16 = ($15|0)==($10|0); //@line 1497 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
  if (!($16)) {
   HEAP32[$__s>>2] = 0; //@line 1499 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
   $$sroa$0$2 = 0;
   STACKTOP = sp;return ($$sroa$0$2|0); //@line 1523 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
  }
 }
 $17 = ($__ns$0|0)>(0); //@line 1503 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
 do {
  if ($17) {
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEjc($__sp,$__ns$0,$__fl); //@line 2133 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $18 = HEAP8[$__sp>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $19 = $18 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $20 = ($19<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $21 = ((($__sp)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $22 = HEAP32[$21>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $23 = ((($__sp)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $24 = $20 ? $23 : $22; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $25 = HEAP32[$0>>2]|0; //@line 360 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
   $26 = ((($25)) + 48|0); //@line 360 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
   $27 = HEAP32[$26>>2]|0; //@line 360 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
   __THREW__ = 0;
   $28 = (invoke_iiii($27|0,($0|0),($24|0),($__ns$0|0))|0); //@line 360 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
   $29 = __THREW__; __THREW__ = 0;
   $30 = $29&1;
   if ($30) {
    $32 = ___cxa_find_matching_catch_2()|0;
    $33 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__sp); //@line 1511 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
    ___resumeException($32|0);
    // unreachable;
   }
   $31 = ($28|0)==($__ns$0|0); //@line 1506 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
   if ($31) {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__sp); //@line 1511 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
    break;
   }
   HEAP32[$__s>>2] = 0; //@line 1508 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__sp); //@line 1511 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
   $$sroa$0$2 = 0;
   STACKTOP = sp;return ($$sroa$0$2|0); //@line 1523 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
  }
 } while(0);
 $34 = (($2) - ($9))|0; //@line 1512 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
 $35 = ($34|0)>(0); //@line 1513 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
 if ($35) {
  $36 = HEAP32[$0>>2]|0; //@line 360 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $37 = ((($36)) + 48|0); //@line 360 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $38 = HEAP32[$37>>2]|0; //@line 360 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $39 = (FUNCTION_TABLE_iiii[$38 & 8191]($0,$__op,$34)|0); //@line 360 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $40 = ($39|0)==($34|0); //@line 1515 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
  if (!($40)) {
   HEAP32[$__s>>2] = 0; //@line 1517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
   $$sroa$0$2 = 0;
   STACKTOP = sp;return ($$sroa$0$2|0); //@line 1523 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
  }
 }
 HEAP32[$5>>2] = 0; //@line 507 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 $$sroa$0$2 = $0;
 STACKTOP = sp;return ($$sroa$0$2|0); //@line 1523 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/locale"
}
function __ZNKSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEE3strEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$02$i$i = 0, $$02$i$i5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__p$0$lcssa$i$i = 0, $__p$0$lcssa$i$i10 = 0;
 var $__p$0$ph$i$i = 0, $__p$0$ph$i$i3 = 0, $__p$01$i$i = 0, $__p$01$i$i6 = 0, $scevgep4$i$i = 0, $scevgep4$i$i8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $1 = HEAP32[$0>>2]|0; //@line 440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $2 = $1 & 16; //@line 440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $3 = ($2|0)==(0); //@line 440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 if (!($3)) {
  $4 = ((($this)) + 44|0); //@line 442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $5 = HEAP32[$4>>2]|0; //@line 442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $6 = ((($this)) + 24|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $7 = HEAP32[$6>>2]|0; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $8 = ($5>>>0)<($7>>>0); //@line 442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  if ($8) {
   HEAP32[$4>>2] = $7; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $13 = $7;
  } else {
   $13 = $5;
  }
  $9 = ((($this)) + 20|0); //@line 176 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $10 = HEAP32[$9>>2]|0; //@line 176 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
  $11 = $10; //@line 2241 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $12 = $13; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $14 = (($12) - ($11))|0; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $15 = ($14>>>0)>(4294967279); //@line 2202 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($15) {
   __ZNKSt3__121__basic_string_commonILb1EE20__throw_length_errorEv($agg$result); //@line 2203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   // unreachable;
  }
  $16 = ($14>>>0)<(11); //@line 2205 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($16) {
   $17 = $14 << 1; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $18 = $17&255; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP8[$agg$result>>0] = $18; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $19 = ((($agg$result)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $__p$0$ph$i$i3 = $19;
  } else {
   $20 = (($14) + 16)|0; //@line 1771 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $21 = $20 & -16; //@line 1771 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $22 = (__Znwj($21)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $23 = ((($agg$result)) + 8|0); //@line 1740 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP32[$23>>2] = $22; //@line 1740 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $24 = $21 | 1; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP32[$agg$result>>2] = $24; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $25 = ((($agg$result)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP32[$25>>2] = $14; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $__p$0$ph$i$i3 = $22;
  }
  $26 = ($10|0)==($13|0); //@line 2218 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($26) {
   $__p$0$lcssa$i$i10 = $__p$0$ph$i$i3;
  } else {
   $$02$i$i5 = $10;$__p$01$i$i6 = $__p$0$ph$i$i3;
   while(1) {
    $27 = HEAP8[$$02$i$i5>>0]|0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$__p$01$i$i6>>0] = $27; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $28 = ((($$02$i$i5)) + 1|0); //@line 2218 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $29 = ((($__p$01$i$i6)) + 1|0); //@line 2218 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $30 = ($28|0)==($13|0); //@line 2218 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if ($30) {
     break;
    } else {
     $$02$i$i5 = $28;$__p$01$i$i6 = $29;
    }
   }
   $scevgep4$i$i8 = (($__p$0$ph$i$i3) + ($14)|0); //@line 631 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $__p$0$lcssa$i$i10 = $scevgep4$i$i8;
  }
  HEAP8[$__p$0$lcssa$i$i10>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  return; //@line 449 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 }
 $31 = $1 & 8; //@line 446 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $32 = ($31|0)==(0); //@line 446 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 if ($32) {
  ;HEAP32[$agg$result>>2]=0|0;HEAP32[$agg$result+4>>2]=0|0;HEAP32[$agg$result+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  return; //@line 449 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 }
 $33 = ((($this)) + 8|0); //@line 169 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $34 = HEAP32[$33>>2]|0; //@line 169 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $35 = ((($this)) + 16|0); //@line 171 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $36 = HEAP32[$35>>2]|0; //@line 171 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
 $37 = $34; //@line 2241 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $38 = $36; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $39 = (($38) - ($37))|0; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $40 = ($39>>>0)>(4294967279); //@line 2202 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($40) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_length_errorEv($agg$result); //@line 2203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  // unreachable;
 }
 $41 = ($39>>>0)<(11); //@line 2205 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($41) {
  $42 = $39 << 1; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $43 = $42&255; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$agg$result>>0] = $43; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $44 = ((($agg$result)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $__p$0$ph$i$i = $44;
 } else {
  $45 = (($39) + 16)|0; //@line 1771 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $46 = $45 & -16; //@line 1771 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $47 = (__Znwj($46)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $48 = ((($agg$result)) + 8|0); //@line 1740 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$48>>2] = $47; //@line 1740 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $49 = $46 | 1; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$agg$result>>2] = $49; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $50 = ((($agg$result)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$50>>2] = $39; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $__p$0$ph$i$i = $47;
 }
 $51 = ($34|0)==($36|0); //@line 2218 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($51) {
  $__p$0$lcssa$i$i = $__p$0$ph$i$i;
 } else {
  $$02$i$i = $34;$__p$01$i$i = $__p$0$ph$i$i;
  while(1) {
   $52 = HEAP8[$$02$i$i>>0]|0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP8[$__p$01$i$i>>0] = $52; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $53 = ((($$02$i$i)) + 1|0); //@line 2218 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $54 = ((($__p$01$i$i)) + 1|0); //@line 2218 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $55 = ($53|0)==($36|0); //@line 2218 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if ($55) {
    break;
   } else {
    $$02$i$i = $53;$__p$01$i$i = $54;
   }
  }
  $scevgep4$i$i = (($__p$0$ph$i$i) + ($39)|0); //@line 631 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $__p$0$lcssa$i$i = $scevgep4$i$i;
 }
 HEAP8[$__p$0$lcssa$i$i>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 return; //@line 449 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
}
function _luaopen_love_nogame($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_loadbuffer($L,180751,55900,236651)|0); //@line 435 "modules/love/love.cpp"
 $1 = ($0|0)==(0); //@line 435 "modules/love/love.cpp"
 if (!($1)) {
  return 1; //@line 438 "modules/love/love.cpp"
 }
 _lua_call($L,0,1); //@line 436 "modules/love/love.cpp"
 return 1; //@line 438 "modules/love/love.cpp"
}
function _luaopen_love_boot($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_loadbuffer($L,236672,17263,253935)|0); //@line 443 "modules/love/love.cpp"
 $1 = ($0|0)==(0); //@line 443 "modules/love/love.cpp"
 if (!($1)) {
  return 1; //@line 446 "modules/love/love.cpp"
 }
 _lua_call($L,0,1); //@line 444 "modules/love/love.cpp"
 return 1; //@line 446 "modules/love/love.cpp"
}
function __ZN4love4math21w__getRandomGeneratorEP9lua_State($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,25,(623992)); //@line 45 "modules/math/wrap_Math.cpp"
 return 1; //@line 46 "modules/math/wrap_Math.cpp"
}
function __ZN4love4math20w_newRandomGeneratorEP9lua_State($L) {
 $L = $L|0;
 var $$02 = 0, $$03 = 0, $$1 = 0, $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 24|0;
 $vararg_buffer = sp + 16|0;
 $0 = sp + 8|0;
 $1 = sp;
 $2 = (_lua_gettop($L)|0); //@line 52 "modules/math/wrap_Math.cpp"
 $3 = ($2|0)>(0); //@line 52 "modules/math/wrap_Math.cpp"
 if ($3) {
  __ZN4love4math20luax_checkrandomseedEP9lua_Statei($0,$L,1); //@line 53 "modules/math/wrap_Math.cpp"
  $4 = $0; //@line 53 "modules/math/wrap_Math.cpp"
  $5 = $4; //@line 53 "modules/math/wrap_Math.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 53 "modules/math/wrap_Math.cpp"
  $7 = (($4) + 4)|0; //@line 53 "modules/math/wrap_Math.cpp"
  $8 = $7; //@line 53 "modules/math/wrap_Math.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 53 "modules/math/wrap_Math.cpp"
  $15 = $6;$18 = $9;
 } else {
  $15 = 0;$18 = 0;
 }
 $10 = (__ZN4love4math4Math18newRandomGeneratorEv(623984)|0); //@line 55 "modules/math/wrap_Math.cpp"
 $11 = (_lua_gettop($L)|0); //@line 57 "modules/math/wrap_Math.cpp"
 $12 = ($11|0)>(0); //@line 57 "modules/math/wrap_Math.cpp"
 if ($12) {
  $13 = $1; //@line 63 "modules/math/wrap_Math.cpp"
  $14 = $13; //@line 63 "modules/math/wrap_Math.cpp"
  HEAP32[$14>>2] = $15; //@line 63 "modules/math/wrap_Math.cpp"
  $16 = (($13) + 4)|0; //@line 63 "modules/math/wrap_Math.cpp"
  $17 = $16; //@line 63 "modules/math/wrap_Math.cpp"
  HEAP32[$17>>2] = $18; //@line 63 "modules/math/wrap_Math.cpp"
  __THREW__ = 0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;HEAP32[$$byval_copy+4>>2]=HEAP32[$1+4>>2]|0; //@line 63 "modules/math/wrap_Math.cpp"
  invoke_vii(3379,($10|0),($$byval_copy|0)); //@line 63 "modules/math/wrap_Math.cpp"
  $19 = __THREW__; __THREW__ = 0;
  $20 = $19&1;
  if ($20) {
   $21 = ___cxa_find_matching_catch_3(96|0)|0;
   $22 = tempRet0;
   $23 = (_llvm_eh_typeid_for((96|0))|0); //@line 64 "modules/math/wrap_Math.cpp"
   $24 = ($22|0)==($23|0); //@line 64 "modules/math/wrap_Math.cpp"
   if (!($24)) {
    $$02 = $22;$$03 = $21;
    ___resumeException($$03|0);
    // unreachable;
   }
   $25 = (___cxa_begin_catch(($21|0))|0); //@line 64 "modules/math/wrap_Math.cpp"
   $26 = HEAP32[$10>>2]|0; //@line 67 "modules/math/wrap_Math.cpp"
   $27 = ((($26)) + 12|0); //@line 67 "modules/math/wrap_Math.cpp"
   $28 = HEAP32[$27>>2]|0; //@line 67 "modules/math/wrap_Math.cpp"
   __THREW__ = 0;
   invoke_vi($28|0,($10|0)); //@line 67 "modules/math/wrap_Math.cpp"
   $29 = __THREW__; __THREW__ = 0;
   $30 = $29&1;
   if (!($30)) {
    $31 = HEAP32[$25>>2]|0; //@line 69 "modules/math/wrap_Math.cpp"
    $32 = ((($31)) + 8|0); //@line 69 "modules/math/wrap_Math.cpp"
    $33 = HEAP32[$32>>2]|0; //@line 69 "modules/math/wrap_Math.cpp"
    $34 = (FUNCTION_TABLE_ii[$33 & 8191]($25)|0); //@line 69 "modules/math/wrap_Math.cpp"
    __THREW__ = 0;
    invoke_vii(3380,($L|0),($34|0)); //@line 69 "modules/math/wrap_Math.cpp"
    $35 = __THREW__; __THREW__ = 0;
    $36 = $35&1;
    if (!($36)) {
     ___cxa_end_catch(); //@line 70 "modules/math/wrap_Math.cpp"
     $41 = (_lua_tolstring($L,-1,0)|0); //@line 73 "modules/math/wrap_Math.cpp"
     HEAP32[$vararg_buffer>>2] = $41; //@line 73 "modules/math/wrap_Math.cpp"
     $42 = (_luaL_error($L,595867,$vararg_buffer)|0); //@line 73 "modules/math/wrap_Math.cpp"
     $$1 = $42;
     STACKTOP = sp;return ($$1|0); //@line 79 "modules/math/wrap_Math.cpp"
    }
   }
   $37 = ___cxa_find_matching_catch_2()|0;
   $38 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 70 "modules/math/wrap_Math.cpp"
   $39 = __THREW__; __THREW__ = 0;
   $40 = $39&1;
   if ($40) {
    $46 = ___cxa_find_matching_catch_3(0|0)|0;
    $47 = tempRet0;
    ___clang_call_terminate($46); //@line 70 "modules/math/wrap_Math.cpp"
    // unreachable; //@line 70 "modules/math/wrap_Math.cpp"
   } else {
    $$02 = $38;$$03 = $37;
    ___resumeException($$03|0);
    // unreachable;
   }
  }
 }
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,25,$10); //@line 76 "modules/math/wrap_Math.cpp"
 $43 = HEAP32[$10>>2]|0; //@line 77 "modules/math/wrap_Math.cpp"
 $44 = ((($43)) + 12|0); //@line 77 "modules/math/wrap_Math.cpp"
 $45 = HEAP32[$44>>2]|0; //@line 77 "modules/math/wrap_Math.cpp"
 FUNCTION_TABLE_vi[$45 & 8191]($10); //@line 77 "modules/math/wrap_Math.cpp"
 $$1 = 1;
 STACKTOP = sp;return ($$1|0); //@line 79 "modules/math/wrap_Math.cpp"
}
function __ZN4love4math16w_newBezierCurveEP9lua_State($L) {
 $L = $L|0;
 var $$off = 0, $$off13 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0.0, $116 = 0, $117 = 0, $118 = 0.0, $119 = 0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0, $123 = 0.0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0.0, $57 = 0, $58 = 0, $59 = 0.0;
 var $6 = 0, $60 = 0.0, $61 = 0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$019 = 0, $i2$020 = 0, $lpad$phi$index = 0, $lpad$phi$index3 = 0, $points = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i1 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i2 = 0, $v = 0, $v3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $points = sp + 16|0;
 $v = sp + 8|0;
 $v3 = sp;
 HEAP32[$points>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $0 = ((($points)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($points)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 $2 = (invoke_iii(3358,($L|0),1)|0); //@line 84 "modules/math/wrap_Math.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 L1: do {
  if ($4) {
   label = 14;
  } else {
   $5 = ($2|0)==(5); //@line 84 "modules/math/wrap_Math.cpp"
   L3: do {
    if ($5) {
     __THREW__ = 0;
     $6 = (invoke_iii(3381,($L|0),1)|0); //@line 86 "modules/math/wrap_Math.cpp"
     $7 = __THREW__; __THREW__ = 0;
     $8 = $7&1;
     if ($8) {
      label = 14;
      break L1;
     }
     $9 = (($6|0) / 2)&-1; //@line 87 "modules/math/wrap_Math.cpp"
     $10 = ((($points)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $11 = HEAP32[$10>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $12 = HEAP32[$points>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $13 = (($11) - ($12))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $14 = $13 >> 3; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $15 = ($14>>>0)<($9>>>0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($15) {
      $16 = HEAP32[$0>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $17 = (($16) - ($12))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $18 = $17 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$off13 = (($6) + 1)|0; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $19 = ($$off13>>>0)<(3); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if ($19) {
       $25 = 0;
      } else {
       $20 = $9 << 3; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       __THREW__ = 0;
       $21 = (invoke_ii(3348,($20|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       $22 = __THREW__; __THREW__ = 0;
       $23 = $22&1;
       if ($23) {
        label = 14;
        break L1;
       }
       $25 = $21;
      }
      $24 = (($25) + ($18<<3)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $26 = $24; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $27 = (($25) + ($9<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $29 = (0 - ($18))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $30 = (($24) + ($29<<3)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $31 = $30; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $32 = $12; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      _memcpy(($30|0),($32|0),($17|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$points>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$0>>2] = $26; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$10>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $33 = ($12|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if (!($33)) {
       __ZdlPv($32); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      }
     }
     $34 = ($6|0)<(1); //@line 88 "modules/math/wrap_Math.cpp"
     if (!($34)) {
      $35 = ((($v)) + 4|0); //@line 220 "./common/Vector.h"
      $i$019 = 1;
      while(1) {
       __THREW__ = 0;
       invoke_viii(3382,($L|0),1,($i$019|0)); //@line 90 "modules/math/wrap_Math.cpp"
       $51 = __THREW__; __THREW__ = 0;
       $52 = $51&1;
       if ($52) {
        break;
       }
       $53 = (($i$019) + 1)|0; //@line 91 "modules/math/wrap_Math.cpp"
       __THREW__ = 0;
       invoke_viii(3382,($L|0),1,($53|0)); //@line 91 "modules/math/wrap_Math.cpp"
       $54 = __THREW__; __THREW__ = 0;
       $55 = $54&1;
       if ($55) {
        break;
       }
       HEAPF32[$v>>2] = 0.0; //@line 219 "./common/Vector.h"
       HEAPF32[$35>>2] = 0.0; //@line 220 "./common/Vector.h"
       __THREW__ = 0;
       $56 = (+invoke_dii(3361,($L|0),-2)); //@line 94 "modules/math/wrap_Math.cpp"
       $57 = __THREW__; __THREW__ = 0;
       $58 = $57&1;
       if ($58) {
        break;
       }
       $59 = $56; //@line 94 "modules/math/wrap_Math.cpp"
       HEAPF32[$v>>2] = $59; //@line 94 "modules/math/wrap_Math.cpp"
       __THREW__ = 0;
       $60 = (+invoke_dii(3361,($L|0),-1)); //@line 95 "modules/math/wrap_Math.cpp"
       $61 = __THREW__; __THREW__ = 0;
       $62 = $61&1;
       if ($62) {
        break;
       }
       $63 = $60; //@line 95 "modules/math/wrap_Math.cpp"
       HEAPF32[$35>>2] = $63; //@line 95 "modules/math/wrap_Math.cpp"
       $64 = HEAP32[$0>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $65 = HEAP32[$10>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $66 = ($64|0)==($65|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($66) {
        __THREW__ = 0;
        invoke_vii(3383,($points|0),($v|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $79 = __THREW__; __THREW__ = 0;
        $80 = $79&1;
        if ($80) {
         break;
        }
       } else {
        $67 = $v; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $68 = $67; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $69 = HEAP32[$68>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $70 = (($67) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $71 = $70; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $72 = HEAP32[$71>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $73 = $64; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $74 = $73; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        HEAP32[$74>>2] = $69; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $75 = (($73) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $76 = $75; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        HEAP32[$76>>2] = $72; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $77 = HEAP32[$0>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $78 = ((($77)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$0>>2] = $78; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       __THREW__ = 0;
       invoke_vii(3384,($L|0),-3); //@line 98 "modules/math/wrap_Math.cpp"
       $81 = __THREW__; __THREW__ = 0;
       $82 = $81&1;
       if ($82) {
        break;
       }
       $83 = (($i$019) + 2)|0; //@line 88 "modules/math/wrap_Math.cpp"
       $84 = ($83|0)>($6|0); //@line 88 "modules/math/wrap_Math.cpp"
       if ($84) {
        break L3;
       } else {
        $i$019 = $83;
       }
      }
      $36 = ___cxa_find_matching_catch_2()|0;
      $37 = tempRet0;
      $lpad$phi$index = $36;$lpad$phi$index3 = $37;
      break L1;
     }
    } else {
     __THREW__ = 0;
     $85 = (invoke_ii(3385,($L|0))|0); //@line 103 "modules/math/wrap_Math.cpp"
     $86 = __THREW__; __THREW__ = 0;
     $87 = $86&1;
     if ($87) {
      label = 14;
      break L1;
     }
     $88 = (($85|0) / 2)&-1; //@line 104 "modules/math/wrap_Math.cpp"
     $89 = ((($points)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $90 = HEAP32[$89>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $91 = HEAP32[$points>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $92 = (($90) - ($91))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $93 = $92 >> 3; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $94 = ($93>>>0)<($88>>>0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($94) {
      $95 = HEAP32[$0>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $96 = (($95) - ($91))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $97 = $96 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$off = (($85) + 1)|0; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $98 = ($$off>>>0)<(3); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if ($98) {
       $104 = 0;
      } else {
       $99 = $88 << 3; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       __THREW__ = 0;
       $100 = (invoke_ii(3348,($99|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       $101 = __THREW__; __THREW__ = 0;
       $102 = $101&1;
       if ($102) {
        label = 14;
        break L1;
       }
       $104 = $100;
      }
      $103 = (($104) + ($97<<3)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $105 = $103; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $106 = (($104) + ($88<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $107 = $106; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $108 = (0 - ($97))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $109 = (($103) + ($108<<3)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $110 = $109; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $111 = $91; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      _memcpy(($109|0),($111|0),($96|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$points>>2] = $110; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$0>>2] = $105; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$89>>2] = $107; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $112 = ($91|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if (!($112)) {
       __ZdlPv($111); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      }
     }
     $113 = ($85|0)<(1); //@line 105 "modules/math/wrap_Math.cpp"
     if (!($113)) {
      $114 = ((($v3)) + 4|0); //@line 220 "./common/Vector.h"
      $i2$020 = 1;
      while(1) {
       HEAPF32[$v3>>2] = 0.0; //@line 219 "./common/Vector.h"
       HEAPF32[$114>>2] = 0.0; //@line 220 "./common/Vector.h"
       __THREW__ = 0;
       $115 = (+invoke_dii(3361,($L|0),($i2$020|0))); //@line 108 "modules/math/wrap_Math.cpp"
       $116 = __THREW__; __THREW__ = 0;
       $117 = $116&1;
       if ($117) {
        break;
       }
       $118 = $115; //@line 108 "modules/math/wrap_Math.cpp"
       HEAPF32[$v3>>2] = $118; //@line 108 "modules/math/wrap_Math.cpp"
       $119 = (($i2$020) + 1)|0; //@line 109 "modules/math/wrap_Math.cpp"
       __THREW__ = 0;
       $120 = (+invoke_dii(3361,($L|0),($119|0))); //@line 109 "modules/math/wrap_Math.cpp"
       $121 = __THREW__; __THREW__ = 0;
       $122 = $121&1;
       if ($122) {
        break;
       }
       $123 = $120; //@line 109 "modules/math/wrap_Math.cpp"
       HEAPF32[$114>>2] = $123; //@line 109 "modules/math/wrap_Math.cpp"
       $124 = HEAP32[$0>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $125 = HEAP32[$89>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $126 = ($124|0)==($125|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($126) {
        __THREW__ = 0;
        invoke_vii(3383,($points|0),($v3|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $139 = __THREW__; __THREW__ = 0;
        $140 = $139&1;
        if ($140) {
         break;
        }
       } else {
        $127 = $v3; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $128 = $127; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $129 = HEAP32[$128>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $130 = (($127) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $131 = $130; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $132 = HEAP32[$131>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $133 = $124; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $134 = $133; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        HEAP32[$134>>2] = $129; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $135 = (($133) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $136 = $135; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        HEAP32[$136>>2] = $132; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $137 = HEAP32[$0>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $138 = ((($137)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$0>>2] = $138; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       $141 = (($i2$020) + 2)|0; //@line 105 "modules/math/wrap_Math.cpp"
       $142 = ($141|0)>($85|0); //@line 105 "modules/math/wrap_Math.cpp"
       if ($142) {
        break L3;
       } else {
        $i2$020 = $141;
       }
      }
      $38 = ___cxa_find_matching_catch_2()|0;
      $39 = tempRet0;
      $lpad$phi$index = $38;$lpad$phi$index3 = $39;
      break L1;
     }
    }
   } while(0);
   __THREW__ = 0;
   $143 = (invoke_iii(3386,(623984|0),($points|0))|0); //@line 114 "modules/math/wrap_Math.cpp"
   $144 = __THREW__; __THREW__ = 0;
   $145 = $144&1;
   if ($145) {
    label = 14;
   } else {
    __THREW__ = 0;
    invoke_viii(3387,($L|0),26,($143|0)); //@line 115 "modules/math/wrap_Math.cpp"
    $146 = __THREW__; __THREW__ = 0;
    $147 = $146&1;
    if ($147) {
     label = 14;
    } else {
     $148 = HEAP32[$143>>2]|0; //@line 116 "modules/math/wrap_Math.cpp"
     $149 = ((($148)) + 12|0); //@line 116 "modules/math/wrap_Math.cpp"
     $150 = HEAP32[$149>>2]|0; //@line 116 "modules/math/wrap_Math.cpp"
     __THREW__ = 0;
     invoke_vi($150|0,($143|0)); //@line 116 "modules/math/wrap_Math.cpp"
     $151 = __THREW__; __THREW__ = 0;
     $152 = $151&1;
     if ($152) {
      label = 14;
     } else {
      $153 = HEAP32[$points>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $154 = ($153|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $155 = $153; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($154) {
       STACKTOP = sp;return 1; //@line 118 "modules/math/wrap_Math.cpp"
      }
      $156 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $157 = ($156|0)==($153|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($157)) {
       $scevgep$i$i$i = ((($156)) + -8|0);
       $158 = $scevgep$i$i$i;
       $159 = (($158) - ($155))|0;
       $160 = $159 >>> 3;
       $161 = $160 ^ -1;
       $scevgep4$i$i$i = (($156) + ($161<<3)|0);
       HEAP32[$0>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($153); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      STACKTOP = sp;return 1; //@line 118 "modules/math/wrap_Math.cpp"
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 14) {
  $40 = ___cxa_find_matching_catch_2()|0;
  $41 = tempRet0;
  $lpad$phi$index = $40;$lpad$phi$index3 = $41;
 }
 $42 = HEAP32[$points>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $43 = ($42|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $44 = $42; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($43) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $45 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $46 = ($45|0)==($42|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($46)) {
  $scevgep$i$i$i1 = ((($45)) + -8|0);
  $47 = $scevgep$i$i$i1;
  $48 = (($47) - ($44))|0;
  $49 = $48 >>> 3;
  $50 = $49 ^ -1;
  $scevgep4$i$i$i2 = (($45) + ($50<<3)|0);
  HEAP32[$0>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($42); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(536870911); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 3; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(268435455); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 2; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$42 = $18;
  } else {
   $$0$i1 = $17;$50 = $18;$51 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 536870911;$50 = $12;$51 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 3; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $51;$42 = $50;
 }
 $24 = (($25) + ($26<<3)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = $__x; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $31 = HEAP32[$30>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $32 = (($29) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = $32; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = HEAP32[$33>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = $24; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = $35; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$36>>2] = $31; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = (($35) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $37; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$38>>2] = $34; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $39 = ((($24)) + 8|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $40 = $39; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $41 = (($42) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $43 = $41 >> 3; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $44 = (0 - ($43))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $45 = (($24) + ($44<<3)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $46 = $45; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($45|0),($47|0),($41|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $46; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $40; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $48 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($48) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $49 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($49); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love4math13w_triangulateEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $$01$i = 0, $$02 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa61 = 0, $$lcssa63 = 0, $$off = 0, $$off26 = 0, $$pre$phi2$i$i$i$iZ2D = 0, $$pre$phiZ2D = 0, $$pre1$i$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0, $102 = 0.0, $103 = 0;
 var $104 = 0, $105 = 0.0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0;
 var $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0;
 var $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0;
 var $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0.0, $211 = 0.0;
 var $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0.0, $218 = 0.0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0.0, $225 = 0.0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0;
 var $230 = 0, $231 = 0.0, $232 = 0.0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0.0, $239 = 0.0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0.0, $246 = 0.0, $247 = 0, $248 = 0;
 var $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0;
 var $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0, $55 = 0, $56 = 0.0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0.0, $98 = 0, $99 = 0, $eh$lpad$body$index12Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i$037 = 0, $i2$038 = 0, $i4$0 = 0, $lpad$phi32$index = 0, $lpad$phi32$index5 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i$i$i$i$i$i = 0, $scevgep$i$i$i13 = 0, $scevgep$i$i$i16 = 0, $scevgep$i$i$i21 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i$i$i$i$i$i = 0, $scevgep4$i$i$i14 = 0;
 var $scevgep4$i$i$i17 = 0, $scevgep4$i$i$i22 = 0, $triangles = 0, $v = 0, $v3 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vertices = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 160|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 88|0;
 $1 = sp + 76|0;
 $vertices = sp + 64|0;
 $v = sp + 44|0;
 $v3 = sp + 24|0;
 $triangles = sp + 12|0;
 HEAP32[$vertices>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($vertices)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($vertices)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$3>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 $4 = (invoke_iii(3358,($L|0),1)|0); //@line 123 "modules/math/wrap_Math.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 L1: do {
  if ($6) {
   label = 14;
  } else {
   $7 = ($4|0)==(5); //@line 123 "modules/math/wrap_Math.cpp"
   L3: do {
    if ($7) {
     __THREW__ = 0;
     $8 = (invoke_iii(3381,($L|0),1)|0); //@line 125 "modules/math/wrap_Math.cpp"
     $9 = __THREW__; __THREW__ = 0;
     $10 = $9&1;
     if ($10) {
      label = 14;
      break L1;
     }
     $11 = (($8|0) / 2)&-1; //@line 126 "modules/math/wrap_Math.cpp"
     $12 = ((($vertices)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $13 = HEAP32[$12>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $14 = HEAP32[$vertices>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $15 = (($13) - ($14))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $16 = (($15|0) / 20)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $17 = ($16>>>0)<($11>>>0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($17) {
      $18 = HEAP32[$2>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $19 = (($18) - ($14))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $20 = (($19|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$off26 = (($8) + 1)|0; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $21 = ($$off26>>>0)<(3); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if ($21) {
       $27 = 0;
      } else {
       $22 = ($11*20)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       __THREW__ = 0;
       $23 = (invoke_ii(3348,($22|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       $24 = __THREW__; __THREW__ = 0;
       $25 = $24&1;
       if ($25) {
        label = 14;
        break L1;
       }
       $27 = $23;
      }
      $26 = (($27) + (($20*20)|0)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $28 = $26; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $29 = (($27) + (($11*20)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $30 = $29; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $31 = (($19|0) / -20)&-1; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $32 = (($26) + (($31*20)|0)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $33 = $32; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $34 = $14; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      _memcpy(($32|0),($34|0),($19|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$vertices>>2] = $33; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$2>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$12>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $35 = ($14|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if (!($35)) {
       __ZdlPv($34); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      }
     }
     $36 = ($8|0)<(1); //@line 127 "modules/math/wrap_Math.cpp"
     if ($36) {
      $$pre$phiZ2D = $vertices;
     } else {
      $37 = ((($v)) + 4|0); //@line 134 "modules/math/wrap_Math.cpp"
      $i$037 = 1;
      while(1) {
       __THREW__ = 0;
       invoke_viii(3382,($L|0),1,($i$037|0)); //@line 129 "modules/math/wrap_Math.cpp"
       $44 = __THREW__; __THREW__ = 0;
       $45 = $44&1;
       if ($45) {
        break;
       }
       $46 = (($i$037) + 1)|0; //@line 130 "modules/math/wrap_Math.cpp"
       __THREW__ = 0;
       invoke_viii(3382,($L|0),1,($46|0)); //@line 130 "modules/math/wrap_Math.cpp"
       $47 = __THREW__; __THREW__ = 0;
       $48 = $47&1;
       if ($48) {
        break;
       }
       __THREW__ = 0;
       $49 = (+invoke_dii(3361,($L|0),-2)); //@line 133 "modules/math/wrap_Math.cpp"
       $50 = __THREW__; __THREW__ = 0;
       $51 = $50&1;
       if ($51) {
        break;
       }
       $52 = $49; //@line 133 "modules/math/wrap_Math.cpp"
       HEAPF32[$v>>2] = $52; //@line 133 "modules/math/wrap_Math.cpp"
       __THREW__ = 0;
       $53 = (+invoke_dii(3361,($L|0),-1)); //@line 134 "modules/math/wrap_Math.cpp"
       $54 = __THREW__; __THREW__ = 0;
       $55 = $54&1;
       if ($55) {
        break;
       }
       $56 = $53; //@line 134 "modules/math/wrap_Math.cpp"
       HEAPF32[$37>>2] = $56; //@line 134 "modules/math/wrap_Math.cpp"
       $57 = HEAP32[$2>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $58 = HEAP32[$12>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $59 = ($57|0)==($58|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($59) {
        __THREW__ = 0;
        invoke_vii(3388,($vertices|0),($v|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $61 = __THREW__; __THREW__ = 0;
        $62 = $61&1;
        if ($62) {
         break;
        }
       } else {
        ;HEAP32[$57>>2]=HEAP32[$v>>2]|0;HEAP32[$57+4>>2]=HEAP32[$v+4>>2]|0;HEAP32[$57+8>>2]=HEAP32[$v+8>>2]|0;HEAP32[$57+12>>2]=HEAP32[$v+12>>2]|0;HEAP32[$57+16>>2]=HEAP32[$v+16>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $60 = ((($57)) + 20|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$2>>2] = $60; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       __THREW__ = 0;
       invoke_vii(3384,($L|0),-3); //@line 137 "modules/math/wrap_Math.cpp"
       $63 = __THREW__; __THREW__ = 0;
       $64 = $63&1;
       if ($64) {
        break;
       }
       $65 = (($i$037) + 2)|0; //@line 127 "modules/math/wrap_Math.cpp"
       $66 = ($65|0)>($8|0); //@line 127 "modules/math/wrap_Math.cpp"
       if ($66) {
        $$pre$phiZ2D = $vertices;
        break L3;
       } else {
        $i$037 = $65;
       }
      }
      $38 = ___cxa_find_matching_catch_2()|0;
      $39 = tempRet0;
      $lpad$phi32$index = $38;$lpad$phi32$index5 = $39;
      label = 15;
      break L1;
     }
    } else {
     __THREW__ = 0;
     $67 = (invoke_ii(3385,($L|0))|0); //@line 142 "modules/math/wrap_Math.cpp"
     $68 = __THREW__; __THREW__ = 0;
     $69 = $68&1;
     if ($69) {
      label = 14;
      break L1;
     }
     $70 = (($67|0) / 2)&-1; //@line 143 "modules/math/wrap_Math.cpp"
     $71 = ((($vertices)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $72 = HEAP32[$71>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $73 = HEAP32[$vertices>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $74 = (($72) - ($73))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $75 = (($74|0) / 20)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $76 = ($75>>>0)<($70>>>0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($76) {
      $77 = HEAP32[$2>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $78 = (($77) - ($73))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $79 = (($78|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$off = (($67) + 1)|0; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $80 = ($$off>>>0)<(3); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if ($80) {
       $86 = 0;
      } else {
       $81 = ($70*20)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       __THREW__ = 0;
       $82 = (invoke_ii(3348,($81|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       $83 = __THREW__; __THREW__ = 0;
       $84 = $83&1;
       if ($84) {
        label = 14;
        break L1;
       }
       $86 = $82;
      }
      $85 = (($86) + (($79*20)|0)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $87 = $85; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $88 = (($86) + (($70*20)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $89 = $88; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $90 = (($78|0) / -20)&-1; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $91 = (($85) + (($90*20)|0)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $92 = $91; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $93 = $73; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      _memcpy(($91|0),($93|0),($78|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$vertices>>2] = $92; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$2>>2] = $87; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$71>>2] = $89; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $94 = ($73|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if (!($94)) {
       __ZdlPv($93); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      }
     }
     $95 = ($67|0)<(1); //@line 144 "modules/math/wrap_Math.cpp"
     if ($95) {
      $$pre$phiZ2D = $vertices;
     } else {
      $96 = ((($v3)) + 4|0); //@line 148 "modules/math/wrap_Math.cpp"
      $i2$038 = 1;
      while(1) {
       __THREW__ = 0;
       $97 = (+invoke_dii(3361,($L|0),($i2$038|0))); //@line 147 "modules/math/wrap_Math.cpp"
       $98 = __THREW__; __THREW__ = 0;
       $99 = $98&1;
       if ($99) {
        break;
       }
       $100 = $97; //@line 147 "modules/math/wrap_Math.cpp"
       HEAPF32[$v3>>2] = $100; //@line 147 "modules/math/wrap_Math.cpp"
       $101 = (($i2$038) + 1)|0; //@line 148 "modules/math/wrap_Math.cpp"
       __THREW__ = 0;
       $102 = (+invoke_dii(3361,($L|0),($101|0))); //@line 148 "modules/math/wrap_Math.cpp"
       $103 = __THREW__; __THREW__ = 0;
       $104 = $103&1;
       if ($104) {
        break;
       }
       $105 = $102; //@line 148 "modules/math/wrap_Math.cpp"
       HEAPF32[$96>>2] = $105; //@line 148 "modules/math/wrap_Math.cpp"
       $106 = HEAP32[$2>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $107 = HEAP32[$71>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $108 = ($106|0)==($107|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($108) {
        __THREW__ = 0;
        invoke_vii(3388,($vertices|0),($v3|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $110 = __THREW__; __THREW__ = 0;
        $111 = $110&1;
        if ($111) {
         break;
        }
       } else {
        ;HEAP32[$106>>2]=HEAP32[$v3>>2]|0;HEAP32[$106+4>>2]=HEAP32[$v3+4>>2]|0;HEAP32[$106+8>>2]=HEAP32[$v3+8>>2]|0;HEAP32[$106+12>>2]=HEAP32[$v3+12>>2]|0;HEAP32[$106+16>>2]=HEAP32[$v3+16>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $109 = ((($106)) + 20|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$2>>2] = $109; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       $112 = (($i2$038) + 2)|0; //@line 144 "modules/math/wrap_Math.cpp"
       $113 = ($112|0)>($67|0); //@line 144 "modules/math/wrap_Math.cpp"
       if ($113) {
        $$pre$phiZ2D = $vertices;
        break L3;
       } else {
        $i2$038 = $112;
       }
      }
      $40 = ___cxa_find_matching_catch_2()|0;
      $41 = tempRet0;
      $lpad$phi32$index = $40;$lpad$phi32$index5 = $41;
      label = 15;
      break L1;
     }
    }
   } while(0);
   $114 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $115 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $116 = (($114) - ($115))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $117 = (($116|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $118 = ($117>>>0)<(3); //@line 153 "modules/math/wrap_Math.cpp"
   L48: do {
    if ($118) {
     __THREW__ = 0;
     $119 = (invoke_iiii(3389,($L|0),(253944|0),($vararg_buffer|0))|0); //@line 154 "modules/math/wrap_Math.cpp"
     $120 = __THREW__; __THREW__ = 0;
     $121 = $120&1;
     if ($121) {
      label = 14;
      break L1;
     } else {
      $$0 = $119;
     }
    } else {
     HEAP32[$triangles>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $122 = ((($triangles)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$122>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $123 = ((($triangles)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$123>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $124 = ($116|0)==(60); //@line 159 "modules/math/wrap_Math.cpp"
     $125 = $115; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($124) {
      $126 = ((($125)) + 20|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $127 = ((($125)) + 40|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $128 = $115; //@line 74 "./common/math.h"
      ;HEAP32[$0>>2]=HEAP32[$128>>2]|0;HEAP32[$0+4>>2]=HEAP32[$128+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$128+8>>2]|0;HEAP32[$0+12>>2]=HEAP32[$128+12>>2]|0;HEAP32[$0+16>>2]=HEAP32[$128+16>>2]|0; //@line 74 "./common/math.h"
      $129 = ((($0)) + 20|0); //@line 74 "./common/math.h"
      ;HEAP32[$129>>2]=HEAP32[$126>>2]|0;HEAP32[$129+4>>2]=HEAP32[$126+4>>2]|0;HEAP32[$129+8>>2]=HEAP32[$126+8>>2]|0;HEAP32[$129+12>>2]=HEAP32[$126+12>>2]|0;HEAP32[$129+16>>2]=HEAP32[$126+16>>2]|0; //@line 74 "./common/math.h"
      $130 = ((($0)) + 40|0); //@line 74 "./common/math.h"
      ;HEAP32[$130>>2]=HEAP32[$127>>2]|0;HEAP32[$130+4>>2]=HEAP32[$127+4>>2]|0;HEAP32[$130+8>>2]=HEAP32[$127+8>>2]|0;HEAP32[$130+12>>2]=HEAP32[$127+12>>2]|0;HEAP32[$130+16>>2]=HEAP32[$127+16>>2]|0; //@line 74 "./common/math.h"
      __THREW__ = 0;
      invoke_vii(3390,($triangles|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $131 = __THREW__; __THREW__ = 0;
      $132 = $131&1;
      if ($132) {
       label = 52;
      } else {
       label = 51;
      }
     } else {
      __THREW__ = 0;
      invoke_viii(3391,($1|0),(623984|0),($vertices|0)); //@line 162 "modules/math/wrap_Math.cpp"
      $133 = __THREW__; __THREW__ = 0;
      $134 = $133&1;
      if ($134) {
       label = 52;
      } else {
       $135 = HEAP32[$triangles>>2]|0; //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $136 = ($135|0)==(0|0); //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $137 = $135; //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($136) {
        $$pre1$i$i$i$i = ((($triangles)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $$pre$phi2$i$i$i$iZ2D = $$pre1$i$i$i$i;
       } else {
        $138 = HEAP32[$122>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $139 = ($138|0)==($135|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($139)) {
         $scevgep$i$i$i$i$i$i$i$i = ((($138)) + -60|0);
         $140 = $scevgep$i$i$i$i$i$i$i$i;
         $141 = (($140) - ($137))|0;
         $142 = (($141>>>0) / 60)&-1;
         $143 = $142 ^ -1;
         $scevgep4$i$i$i$i$i$i$i$i = (($138) + (($143*60)|0)|0);
         HEAP32[$122>>2] = $scevgep4$i$i$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
        __ZdlPv($135); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        $144 = ((($triangles)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        HEAP32[$144>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$122>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$triangles>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $$pre$phi2$i$i$i$iZ2D = $144;
       }
       $145 = HEAP32[$1>>2]|0; //@line 1343 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$triangles>>2] = $145; //@line 1343 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $146 = ((($1)) + 4|0); //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $147 = HEAP32[$146>>2]|0; //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$122>>2] = $147; //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $148 = ((($1)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $149 = HEAP32[$148>>2]|0; //@line 1345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$$pre$phi2$i$i$i$iZ2D>>2] = $149; //@line 1345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$148>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$146>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$1>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       label = 51;
      }
     }
     L62: do {
      if ((label|0) == 51) {
       label = 60;
      }
      else if ((label|0) == 52) {
       $150 = ___cxa_find_matching_catch_3(8992|0)|0;
       $151 = tempRet0;
       $152 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
       $153 = ($151|0)==($152|0); //@line 535 "./common/runtime.h"
       do {
        if ($153) {
         $154 = (___cxa_begin_catch(($150|0))|0); //@line 535 "./common/runtime.h"
         $155 = HEAP32[$154>>2]|0; //@line 539 "./common/runtime.h"
         $156 = ((($155)) + 8|0); //@line 539 "./common/runtime.h"
         $157 = HEAP32[$156>>2]|0; //@line 539 "./common/runtime.h"
         $158 = (FUNCTION_TABLE_ii[$157 & 8191]($154)|0); //@line 539 "./common/runtime.h"
         __THREW__ = 0;
         invoke_vii(3380,($L|0),($158|0)); //@line 539 "./common/runtime.h"
         $159 = __THREW__; __THREW__ = 0;
         $160 = $159&1;
         if ($160) {
          $168 = ___cxa_find_matching_catch_2()|0;
          $169 = tempRet0;
          __THREW__ = 0;
          invoke_v(3378); //@line 540 "./common/runtime.h"
          $170 = __THREW__; __THREW__ = 0;
          $171 = $170&1;
          if (!($171)) {
           $$01$i = $169;$$02$i = $168;
           break;
          }
          $172 = ___cxa_find_matching_catch_3(0|0)|0;
          $173 = tempRet0;
          ___clang_call_terminate($172); //@line 540 "./common/runtime.h"
          // unreachable; //@line 540 "./common/runtime.h"
         }
         __THREW__ = 0;
         invoke_v(3378); //@line 540 "./common/runtime.h"
         $161 = __THREW__; __THREW__ = 0;
         $162 = $161&1;
         if ($162) {
          label = 68;
          break L62;
         }
         __THREW__ = 0;
         $163 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
         $164 = __THREW__; __THREW__ = 0;
         $165 = $164&1;
         if ($165) {
          label = 68;
          break L62;
         }
         __THREW__ = 0;
         HEAP32[$vararg_buffer1>>2] = $163; //@line 543 "./common/runtime.h"
         (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer1|0))|0); //@line 543 "./common/runtime.h"
         $166 = __THREW__; __THREW__ = 0;
         $167 = $166&1;
         if ($167) {
          label = 68;
          break L62;
         } else {
          label = 60;
          break L62;
         }
        } else {
         $$01$i = $151;$$02$i = $150;
        }
       } while(0);
       $eh$lpad$body$index12Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
      }
     } while(0);
     do {
      if ((label|0) == 60) {
       $174 = HEAP32[$122>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $175 = HEAP32[$triangles>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $176 = (($174) - ($175))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $177 = (($176|0) / 60)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       __THREW__ = 0;
       invoke_viii(3393,($L|0),($177|0),0); //@line 165 "modules/math/wrap_Math.cpp"
       $178 = __THREW__; __THREW__ = 0;
       $179 = $178&1;
       if ($179) {
        label = 68;
       } else {
        $i4$0 = 0;
        while(1) {
         $180 = HEAP32[$122>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $181 = HEAP32[$triangles>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $182 = (($180) - ($181))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $183 = (($182|0) / 60)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $184 = ($i4$0|0)<($183|0); //@line 166 "modules/math/wrap_Math.cpp"
         $185 = $181; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if (!($184)) {
          $$lcssa = $180;$$lcssa61 = $181;$$lcssa63 = $185;
          break;
         }
         __THREW__ = 0;
         invoke_viii(3393,($L|0),6,0); //@line 170 "modules/math/wrap_Math.cpp"
         $207 = __THREW__; __THREW__ = 0;
         $208 = $207&1;
         if ($208) {
          label = 67;
          break;
         }
         $209 = (($185) + (($i4$0*60)|0)|0); //@line 171 "modules/math/wrap_Math.cpp"
         $210 = +HEAPF32[$209>>2]; //@line 171 "modules/math/wrap_Math.cpp"
         $211 = $210; //@line 171 "modules/math/wrap_Math.cpp"
         __THREW__ = 0;
         invoke_vid(3394,($L|0),(+$211)); //@line 171 "modules/math/wrap_Math.cpp"
         $212 = __THREW__; __THREW__ = 0;
         $213 = $212&1;
         if ($213) {
          label = 67;
          break;
         }
         __THREW__ = 0;
         invoke_viii(3395,($L|0),-2,1); //@line 172 "modules/math/wrap_Math.cpp"
         $214 = __THREW__; __THREW__ = 0;
         $215 = $214&1;
         if ($215) {
          label = 67;
          break;
         }
         $216 = (((($185) + (($i4$0*60)|0)|0)) + 4|0); //@line 173 "modules/math/wrap_Math.cpp"
         $217 = +HEAPF32[$216>>2]; //@line 173 "modules/math/wrap_Math.cpp"
         $218 = $217; //@line 173 "modules/math/wrap_Math.cpp"
         __THREW__ = 0;
         invoke_vid(3394,($L|0),(+$218)); //@line 173 "modules/math/wrap_Math.cpp"
         $219 = __THREW__; __THREW__ = 0;
         $220 = $219&1;
         if ($220) {
          label = 67;
          break;
         }
         __THREW__ = 0;
         invoke_viii(3395,($L|0),-2,2); //@line 174 "modules/math/wrap_Math.cpp"
         $221 = __THREW__; __THREW__ = 0;
         $222 = $221&1;
         if ($222) {
          label = 67;
          break;
         }
         $223 = (((($185) + (($i4$0*60)|0)|0)) + 20|0); //@line 175 "modules/math/wrap_Math.cpp"
         $224 = +HEAPF32[$223>>2]; //@line 175 "modules/math/wrap_Math.cpp"
         $225 = $224; //@line 175 "modules/math/wrap_Math.cpp"
         __THREW__ = 0;
         invoke_vid(3394,($L|0),(+$225)); //@line 175 "modules/math/wrap_Math.cpp"
         $226 = __THREW__; __THREW__ = 0;
         $227 = $226&1;
         if ($227) {
          label = 67;
          break;
         }
         __THREW__ = 0;
         invoke_viii(3395,($L|0),-2,3); //@line 176 "modules/math/wrap_Math.cpp"
         $228 = __THREW__; __THREW__ = 0;
         $229 = $228&1;
         if ($229) {
          label = 67;
          break;
         }
         $230 = (((($185) + (($i4$0*60)|0)|0)) + 24|0); //@line 177 "modules/math/wrap_Math.cpp"
         $231 = +HEAPF32[$230>>2]; //@line 177 "modules/math/wrap_Math.cpp"
         $232 = $231; //@line 177 "modules/math/wrap_Math.cpp"
         __THREW__ = 0;
         invoke_vid(3394,($L|0),(+$232)); //@line 177 "modules/math/wrap_Math.cpp"
         $233 = __THREW__; __THREW__ = 0;
         $234 = $233&1;
         if ($234) {
          label = 67;
          break;
         }
         __THREW__ = 0;
         invoke_viii(3395,($L|0),-2,4); //@line 178 "modules/math/wrap_Math.cpp"
         $235 = __THREW__; __THREW__ = 0;
         $236 = $235&1;
         if ($236) {
          label = 67;
          break;
         }
         $237 = (((($185) + (($i4$0*60)|0)|0)) + 40|0); //@line 179 "modules/math/wrap_Math.cpp"
         $238 = +HEAPF32[$237>>2]; //@line 179 "modules/math/wrap_Math.cpp"
         $239 = $238; //@line 179 "modules/math/wrap_Math.cpp"
         __THREW__ = 0;
         invoke_vid(3394,($L|0),(+$239)); //@line 179 "modules/math/wrap_Math.cpp"
         $240 = __THREW__; __THREW__ = 0;
         $241 = $240&1;
         if ($241) {
          label = 67;
          break;
         }
         __THREW__ = 0;
         invoke_viii(3395,($L|0),-2,5); //@line 180 "modules/math/wrap_Math.cpp"
         $242 = __THREW__; __THREW__ = 0;
         $243 = $242&1;
         if ($243) {
          label = 67;
          break;
         }
         $244 = (((($185) + (($i4$0*60)|0)|0)) + 44|0); //@line 181 "modules/math/wrap_Math.cpp"
         $245 = +HEAPF32[$244>>2]; //@line 181 "modules/math/wrap_Math.cpp"
         $246 = $245; //@line 181 "modules/math/wrap_Math.cpp"
         __THREW__ = 0;
         invoke_vid(3394,($L|0),(+$246)); //@line 181 "modules/math/wrap_Math.cpp"
         $247 = __THREW__; __THREW__ = 0;
         $248 = $247&1;
         if ($248) {
          label = 67;
          break;
         }
         __THREW__ = 0;
         invoke_viii(3395,($L|0),-2,6); //@line 182 "modules/math/wrap_Math.cpp"
         $249 = __THREW__; __THREW__ = 0;
         $250 = $249&1;
         if ($250) {
          label = 67;
          break;
         }
         $251 = (($i4$0) + 1)|0; //@line 184 "modules/math/wrap_Math.cpp"
         __THREW__ = 0;
         invoke_viii(3395,($L|0),-2,($251|0)); //@line 184 "modules/math/wrap_Math.cpp"
         $252 = __THREW__; __THREW__ = 0;
         $253 = $252&1;
         if ($253) {
          label = 67;
          break;
         } else {
          $i4$0 = $251;
         }
        }
        if ((label|0) == 67) {
         $194 = ___cxa_find_matching_catch_2()|0;
         $195 = tempRet0;
         $eh$lpad$body$index12Z2D = $195;$eh$lpad$body$indexZ2D = $194;
         break;
        }
        $186 = $$lcssa; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $187 = ($$lcssa61|0)==(0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $188 = $$lcssa61; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($187)) {
         $189 = ($186|0)==($$lcssa63|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if (!($189)) {
          $scevgep$i$i$i13 = ((($186)) + -60|0);
          $190 = $scevgep$i$i$i13;
          $191 = (($190) - ($$lcssa61))|0;
          $192 = (($191>>>0) / 60)&-1;
          $193 = $192 ^ -1;
          $scevgep4$i$i$i14 = (($186) + (($193*60)|0)|0);
          HEAP32[$122>>2] = $scevgep4$i$i$i14; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         }
         __ZdlPv($188); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        }
        $$0 = 1;
        break L48;
       }
      }
     } while(0);
     if ((label|0) == 68) {
      $196 = ___cxa_find_matching_catch_2()|0;
      $197 = tempRet0;
      $eh$lpad$body$index12Z2D = $197;$eh$lpad$body$indexZ2D = $196;
     }
     $198 = HEAP32[$triangles>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $199 = ($198|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $200 = $198; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($199) {
      $$01 = $eh$lpad$body$indexZ2D;$$02 = $eh$lpad$body$index12Z2D;
      break L1;
     }
     $201 = HEAP32[$122>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $202 = ($201|0)==($198|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($202)) {
      $scevgep$i$i$i16 = ((($201)) + -60|0);
      $203 = $scevgep$i$i$i16;
      $204 = (($203) - ($200))|0;
      $205 = (($204>>>0) / 60)&-1;
      $206 = $205 ^ -1;
      $scevgep4$i$i$i17 = (($201) + (($206*60)|0)|0);
      HEAP32[$122>>2] = $scevgep4$i$i$i17; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __ZdlPv($198); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $$01 = $eh$lpad$body$indexZ2D;$$02 = $eh$lpad$body$index12Z2D;
     break L1;
    }
   } while(0);
   $254 = HEAP32[$vertices>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $255 = ($254|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $256 = $254; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($255) {
    STACKTOP = sp;return ($$0|0); //@line 188 "modules/math/wrap_Math.cpp"
   }
   $257 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $258 = ($257|0)==($254|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($258)) {
    $scevgep$i$i$i21 = ((($257)) + -20|0);
    $259 = $scevgep$i$i$i21;
    $260 = (($259) - ($256))|0;
    $261 = (($260>>>0) / 20)&-1;
    $262 = $261 ^ -1;
    $scevgep4$i$i$i22 = (($257) + (($262*20)|0)|0);
    HEAP32[$2>>2] = $scevgep4$i$i$i22; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($254); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   STACKTOP = sp;return ($$0|0); //@line 188 "modules/math/wrap_Math.cpp"
  }
 } while(0);
 if ((label|0) == 14) {
  $42 = ___cxa_find_matching_catch_2()|0;
  $43 = tempRet0;
  $lpad$phi32$index = $42;$lpad$phi32$index5 = $43;
  label = 15;
 }
 if ((label|0) == 15) {
  $$01 = $lpad$phi32$index;$$02 = $lpad$phi32$index5;
 }
 $263 = HEAP32[$vertices>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $264 = ($263|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $265 = $263; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($264) {
  ___resumeException($$01|0);
  // unreachable;
 }
 $266 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $267 = ($266|0)==($263|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($267)) {
  $scevgep$i$i$i = ((($266)) + -20|0);
  $268 = $scevgep$i$i$i;
  $269 = (($268) - ($265))|0;
  $270 = (($269>>>0) / 20)&-1;
  $271 = $270 ^ -1;
  $scevgep4$i$i$i = (($266) + (($271*20)|0)|0);
  HEAP32[$2>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($263); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$01|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__16vectorIN4love6VertexENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($3|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(214748364); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9|0) / 20)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(107374182); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $10 << 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = (($19|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$32 = $18;
  } else {
   $$0$i1 = $17;$39 = $18;$40 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 214748364;$39 = $12;$40 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = ($$0$i1*20)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $40;$32 = $39;
 }
 $24 = (($25) + (($26*20)|0)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + (($$0$i2*20)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 ;HEAP32[$24>>2]=HEAP32[$__x>>2]|0;HEAP32[$24+4>>2]=HEAP32[$__x+4>>2]|0;HEAP32[$24+8>>2]=HEAP32[$__x+8>>2]|0;HEAP32[$24+12>>2]=HEAP32[$__x+12>>2]|0;HEAP32[$24+16>>2]=HEAP32[$__x+16>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $29 = ((($24)) + 20|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = $29; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = (($32) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = (($31|0) / -20)&-1; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = (($24) + (($33*20)|0)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = $34; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($34|0),($36|0),($31|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $35; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $37 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($37) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $38 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($38); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNSt3__16vectorIN4love8TriangleENS_9allocatorIS2_EEE21__push_back_slow_pathIS2_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($3|0) / 60)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(71582788); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9|0) / 60)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(35791394); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $10 << 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = (($19|0) / 60)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$32 = $18;
  } else {
   $$0$i1 = $17;$39 = $18;$40 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13|0) / 60)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 71582788;$39 = $12;$40 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = ($$0$i1*60)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $40;$32 = $39;
 }
 $24 = (($25) + (($26*60)|0)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + (($$0$i2*60)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 dest=$24; src=$__x; stop=dest+60|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $29 = ((($24)) + 60|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = $29; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = (($32) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = (($31|0) / -60)&-1; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = (($24) + (($33*60)|0)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = $34; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($34|0),($36|0),($31|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $35; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $37 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($37) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $38 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($38); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love4math10w_isConvexEP9lua_State($L) {
 $L = $L|0;
 var $$off = 0, $$off12 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0.0, $105 = 0, $106 = 0, $107 = 0.0, $108 = 0, $109 = 0.0, $11 = 0, $110 = 0, $111 = 0, $112 = 0.0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0.0, $57 = 0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$018 = 0, $i2$019 = 0, $lpad$phi$index = 0;
 var $lpad$phi$index3 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i1 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i2 = 0, $v = 0, $v3 = 0, $vertices = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vertices = sp + 40|0;
 $v = sp + 20|0;
 $v3 = sp;
 HEAP32[$vertices>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $0 = ((($vertices)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($vertices)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 $2 = (invoke_iii(3358,($L|0),1)|0); //@line 193 "modules/math/wrap_Math.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 L1: do {
  if ($4) {
   label = 14;
  } else {
   $5 = ($2|0)==(5); //@line 193 "modules/math/wrap_Math.cpp"
   L3: do {
    if ($5) {
     __THREW__ = 0;
     $6 = (invoke_iii(3381,($L|0),1)|0); //@line 195 "modules/math/wrap_Math.cpp"
     $7 = __THREW__; __THREW__ = 0;
     $8 = $7&1;
     if ($8) {
      label = 14;
      break L1;
     }
     $9 = (($6|0) / 2)&-1; //@line 196 "modules/math/wrap_Math.cpp"
     $10 = ((($vertices)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $11 = HEAP32[$10>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $12 = HEAP32[$vertices>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $13 = (($11) - ($12))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $14 = (($13|0) / 20)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $15 = ($14>>>0)<($9>>>0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($15) {
      $16 = HEAP32[$0>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $17 = (($16) - ($12))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $18 = (($17|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$off12 = (($6) + 1)|0; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $19 = ($$off12>>>0)<(3); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if ($19) {
       $25 = 0;
      } else {
       $20 = ($9*20)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       __THREW__ = 0;
       $21 = (invoke_ii(3348,($20|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       $22 = __THREW__; __THREW__ = 0;
       $23 = $22&1;
       if ($23) {
        label = 14;
        break L1;
       }
       $25 = $21;
      }
      $24 = (($25) + (($18*20)|0)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $26 = $24; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $27 = (($25) + (($9*20)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $29 = (($17|0) / -20)&-1; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $30 = (($24) + (($29*20)|0)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $31 = $30; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $32 = $12; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      _memcpy(($30|0),($32|0),($17|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$vertices>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$0>>2] = $26; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$10>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $33 = ($12|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if (!($33)) {
       __ZdlPv($32); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      }
     }
     $34 = ($6|0)<(1); //@line 197 "modules/math/wrap_Math.cpp"
     if (!($34)) {
      $35 = ((($v)) + 4|0); //@line 204 "modules/math/wrap_Math.cpp"
      $i$018 = 1;
      while(1) {
       __THREW__ = 0;
       invoke_viii(3382,($L|0),1,($i$018|0)); //@line 199 "modules/math/wrap_Math.cpp"
       $51 = __THREW__; __THREW__ = 0;
       $52 = $51&1;
       if ($52) {
        break;
       }
       $53 = (($i$018) + 1)|0; //@line 200 "modules/math/wrap_Math.cpp"
       __THREW__ = 0;
       invoke_viii(3382,($L|0),1,($53|0)); //@line 200 "modules/math/wrap_Math.cpp"
       $54 = __THREW__; __THREW__ = 0;
       $55 = $54&1;
       if ($55) {
        break;
       }
       __THREW__ = 0;
       $56 = (+invoke_dii(3361,($L|0),-2)); //@line 203 "modules/math/wrap_Math.cpp"
       $57 = __THREW__; __THREW__ = 0;
       $58 = $57&1;
       if ($58) {
        break;
       }
       $59 = $56; //@line 203 "modules/math/wrap_Math.cpp"
       HEAPF32[$v>>2] = $59; //@line 203 "modules/math/wrap_Math.cpp"
       __THREW__ = 0;
       $60 = (+invoke_dii(3361,($L|0),-1)); //@line 204 "modules/math/wrap_Math.cpp"
       $61 = __THREW__; __THREW__ = 0;
       $62 = $61&1;
       if ($62) {
        break;
       }
       $63 = $60; //@line 204 "modules/math/wrap_Math.cpp"
       HEAPF32[$35>>2] = $63; //@line 204 "modules/math/wrap_Math.cpp"
       $64 = HEAP32[$0>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $65 = HEAP32[$10>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $66 = ($64|0)==($65|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($66) {
        __THREW__ = 0;
        invoke_vii(3388,($vertices|0),($v|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $68 = __THREW__; __THREW__ = 0;
        $69 = $68&1;
        if ($69) {
         break;
        }
       } else {
        ;HEAP32[$64>>2]=HEAP32[$v>>2]|0;HEAP32[$64+4>>2]=HEAP32[$v+4>>2]|0;HEAP32[$64+8>>2]=HEAP32[$v+8>>2]|0;HEAP32[$64+12>>2]=HEAP32[$v+12>>2]|0;HEAP32[$64+16>>2]=HEAP32[$v+16>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $67 = ((($64)) + 20|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$0>>2] = $67; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       __THREW__ = 0;
       invoke_vii(3384,($L|0),-3); //@line 207 "modules/math/wrap_Math.cpp"
       $70 = __THREW__; __THREW__ = 0;
       $71 = $70&1;
       if ($71) {
        break;
       }
       $72 = (($i$018) + 2)|0; //@line 197 "modules/math/wrap_Math.cpp"
       $73 = ($72|0)>($6|0); //@line 197 "modules/math/wrap_Math.cpp"
       if ($73) {
        break L3;
       } else {
        $i$018 = $72;
       }
      }
      $36 = ___cxa_find_matching_catch_2()|0;
      $37 = tempRet0;
      $lpad$phi$index = $36;$lpad$phi$index3 = $37;
      break L1;
     }
    } else {
     __THREW__ = 0;
     $74 = (invoke_ii(3385,($L|0))|0); //@line 212 "modules/math/wrap_Math.cpp"
     $75 = __THREW__; __THREW__ = 0;
     $76 = $75&1;
     if ($76) {
      label = 14;
      break L1;
     }
     $77 = (($74|0) / 2)&-1; //@line 213 "modules/math/wrap_Math.cpp"
     $78 = ((($vertices)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $79 = HEAP32[$78>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $80 = HEAP32[$vertices>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $81 = (($79) - ($80))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $82 = (($81|0) / 20)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $83 = ($82>>>0)<($77>>>0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($83) {
      $84 = HEAP32[$0>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $85 = (($84) - ($80))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $86 = (($85|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$off = (($74) + 1)|0; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $87 = ($$off>>>0)<(3); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if ($87) {
       $93 = 0;
      } else {
       $88 = ($77*20)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       __THREW__ = 0;
       $89 = (invoke_ii(3348,($88|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       $90 = __THREW__; __THREW__ = 0;
       $91 = $90&1;
       if ($91) {
        label = 14;
        break L1;
       }
       $93 = $89;
      }
      $92 = (($93) + (($86*20)|0)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $94 = $92; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $95 = (($93) + (($77*20)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $96 = $95; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $97 = (($85|0) / -20)&-1; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $98 = (($92) + (($97*20)|0)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $99 = $98; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $100 = $80; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      _memcpy(($98|0),($100|0),($85|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$vertices>>2] = $99; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$0>>2] = $94; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$78>>2] = $96; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $101 = ($80|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if (!($101)) {
       __ZdlPv($100); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      }
     }
     $102 = ($74|0)<(1); //@line 214 "modules/math/wrap_Math.cpp"
     if (!($102)) {
      $103 = ((($v3)) + 4|0); //@line 218 "modules/math/wrap_Math.cpp"
      $i2$019 = 1;
      while(1) {
       __THREW__ = 0;
       $104 = (+invoke_dii(3361,($L|0),($i2$019|0))); //@line 217 "modules/math/wrap_Math.cpp"
       $105 = __THREW__; __THREW__ = 0;
       $106 = $105&1;
       if ($106) {
        break;
       }
       $107 = $104; //@line 217 "modules/math/wrap_Math.cpp"
       HEAPF32[$v3>>2] = $107; //@line 217 "modules/math/wrap_Math.cpp"
       $108 = (($i2$019) + 1)|0; //@line 218 "modules/math/wrap_Math.cpp"
       __THREW__ = 0;
       $109 = (+invoke_dii(3361,($L|0),($108|0))); //@line 218 "modules/math/wrap_Math.cpp"
       $110 = __THREW__; __THREW__ = 0;
       $111 = $110&1;
       if ($111) {
        break;
       }
       $112 = $109; //@line 218 "modules/math/wrap_Math.cpp"
       HEAPF32[$103>>2] = $112; //@line 218 "modules/math/wrap_Math.cpp"
       $113 = HEAP32[$0>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $114 = HEAP32[$78>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $115 = ($113|0)==($114|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($115) {
        __THREW__ = 0;
        invoke_vii(3388,($vertices|0),($v3|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $117 = __THREW__; __THREW__ = 0;
        $118 = $117&1;
        if ($118) {
         break;
        }
       } else {
        ;HEAP32[$113>>2]=HEAP32[$v3>>2]|0;HEAP32[$113+4>>2]=HEAP32[$v3+4>>2]|0;HEAP32[$113+8>>2]=HEAP32[$v3+8>>2]|0;HEAP32[$113+12>>2]=HEAP32[$v3+12>>2]|0;HEAP32[$113+16>>2]=HEAP32[$v3+16>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $116 = ((($113)) + 20|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$0>>2] = $116; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       $119 = (($i2$019) + 2)|0; //@line 214 "modules/math/wrap_Math.cpp"
       $120 = ($119|0)>($74|0); //@line 214 "modules/math/wrap_Math.cpp"
       if ($120) {
        break L3;
       } else {
        $i2$019 = $119;
       }
      }
      $38 = ___cxa_find_matching_catch_2()|0;
      $39 = tempRet0;
      $lpad$phi$index = $38;$lpad$phi$index3 = $39;
      break L1;
     }
    }
   } while(0);
   __THREW__ = 0;
   $121 = (invoke_iii(3396,(623984|0),($vertices|0))|0); //@line 223 "modules/math/wrap_Math.cpp"
   $122 = __THREW__; __THREW__ = 0;
   $123 = $122&1;
   if ($123) {
    label = 14;
   } else {
    __THREW__ = 0;
    invoke_vii(3397,($L|0),($121|0)); //@line 223 "modules/math/wrap_Math.cpp"
    $124 = __THREW__; __THREW__ = 0;
    $125 = $124&1;
    if ($125) {
     label = 14;
    } else {
     $126 = HEAP32[$vertices>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $127 = ($126|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $128 = $126; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($127) {
      STACKTOP = sp;return 1; //@line 225 "modules/math/wrap_Math.cpp"
     }
     $129 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $130 = ($129|0)==($126|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($130)) {
      $scevgep$i$i$i = ((($129)) + -20|0);
      $131 = $scevgep$i$i$i;
      $132 = (($131) - ($128))|0;
      $133 = (($132>>>0) / 20)&-1;
      $134 = $133 ^ -1;
      $scevgep4$i$i$i = (($129) + (($134*20)|0)|0);
      HEAP32[$0>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __ZdlPv($126); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     STACKTOP = sp;return 1; //@line 225 "modules/math/wrap_Math.cpp"
    }
   }
  }
 } while(0);
 if ((label|0) == 14) {
  $40 = ___cxa_find_matching_catch_2()|0;
  $41 = tempRet0;
  $lpad$phi$index = $40;$lpad$phi$index3 = $41;
 }
 $42 = HEAP32[$vertices>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $43 = ($42|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $44 = $42; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($43) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $45 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $46 = ($45|0)==($42|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($46)) {
  $scevgep$i$i$i1 = ((($45)) + -20|0);
  $47 = $scevgep$i$i$i1;
  $48 = (($47) - ($44))|0;
  $49 = (($48>>>0) / 20)&-1;
  $50 = $49 ^ -1;
  $scevgep4$i$i$i2 = (($45) + (($50*20)|0)|0);
  HEAP32[$0>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($42); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love4math15w_gammaToLinearEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0, $color = 0, $exitcond = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $color = sp;
 $0 = (__ZN4love4mathL12getGammaArgsEP9lua_StatePf($L,$color)|0); //@line 262 "modules/math/wrap_Math.cpp"
 $1 = ($0|0)>(0); //@line 264 "modules/math/wrap_Math.cpp"
 if ($1) {
  $i$01 = 0;
 } else {
  STACKTOP = sp;return ($0|0); //@line 272 "modules/math/wrap_Math.cpp"
 }
 while(1) {
  $2 = ($i$01|0)<(3); //@line 267 "modules/math/wrap_Math.cpp"
  $3 = (($color) + ($i$01<<2)|0); //@line 268 "modules/math/wrap_Math.cpp"
  $4 = +HEAPF32[$3>>2]; //@line 268 "modules/math/wrap_Math.cpp"
  if ($2) {
   $5 = (+__ZNK4love4math4Math13gammaToLinearEf(623984,$4)); //@line 268 "modules/math/wrap_Math.cpp"
   HEAPF32[$3>>2] = $5; //@line 268 "modules/math/wrap_Math.cpp"
   $7 = $5;
  } else {
   $7 = $4;
  }
  $6 = $7 * 255.0; //@line 269 "modules/math/wrap_Math.cpp"
  $8 = $6; //@line 269 "modules/math/wrap_Math.cpp"
  _lua_pushnumber($L,$8); //@line 269 "modules/math/wrap_Math.cpp"
  $9 = (($i$01) + 1)|0; //@line 264 "modules/math/wrap_Math.cpp"
  $exitcond = ($9|0)==($0|0); //@line 264 "modules/math/wrap_Math.cpp"
  if ($exitcond) {
   break;
  } else {
   $i$01 = $9;
  }
 }
 STACKTOP = sp;return ($0|0); //@line 272 "modules/math/wrap_Math.cpp"
}
function __ZN4love4mathL12getGammaArgsEP9lua_StatePf($L,$color) {
 $L = $L|0;
 $color = $color|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $i2$04 = 0, $numcomponents$02 = 0, $numcomponents$02$lcssa = 0, $numcomponents$13 = 0, $numcomponents$29 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_type($L,1)|0); //@line 231 "modules/math/wrap_Math.cpp"
 $1 = ($0|0)==(5); //@line 231 "modules/math/wrap_Math.cpp"
 do {
  if ($1) {
   $2 = (__ZN4love11luax_objlenEP9lua_Statei($L,1)|0); //@line 233 "modules/math/wrap_Math.cpp"
   $3 = ($2|0)>(0); //@line 234 "modules/math/wrap_Math.cpp"
   if ($3) {
    $i$01 = 1;$numcomponents$02 = 0;
   } else {
    _lua_settop($L,-1); //@line 241 "modules/math/wrap_Math.cpp"
    break;
   }
   while(1) {
    _lua_rawgeti($L,1,$i$01); //@line 236 "modules/math/wrap_Math.cpp"
    $4 = (+_luaL_checknumber($L,-1)); //@line 237 "modules/math/wrap_Math.cpp"
    $5 = $4; //@line 237 "modules/math/wrap_Math.cpp"
    $6 = $5 / 255.0; //@line 237 "modules/math/wrap_Math.cpp"
    $7 = (($i$01) + -1)|0; //@line 237 "modules/math/wrap_Math.cpp"
    $8 = (($color) + ($7<<2)|0); //@line 237 "modules/math/wrap_Math.cpp"
    HEAPF32[$8>>2] = $6; //@line 237 "modules/math/wrap_Math.cpp"
    $9 = (($numcomponents$02) + 1)|0; //@line 238 "modules/math/wrap_Math.cpp"
    $10 = (($i$01) + 1)|0; //@line 234 "modules/math/wrap_Math.cpp"
    $11 = ($i$01|0)<($2|0); //@line 234 "modules/math/wrap_Math.cpp"
    $12 = ($10|0)<(5); //@line 234 "modules/math/wrap_Math.cpp"
    $13 = $11 & $12; //@line 234 "modules/math/wrap_Math.cpp"
    if ($13) {
     $i$01 = $10;$numcomponents$02 = $9;
    } else {
     $$lcssa = $9;$numcomponents$02$lcssa = $numcomponents$02;
     break;
    }
   }
   $14 = (-2 - ($numcomponents$02$lcssa))|0; //@line 241 "modules/math/wrap_Math.cpp"
   _lua_settop($L,$14); //@line 241 "modules/math/wrap_Math.cpp"
   $numcomponents$29 = $$lcssa;
   return ($numcomponents$29|0); //@line 256 "modules/math/wrap_Math.cpp"
  } else {
   $15 = (_lua_gettop($L)|0); //@line 245 "modules/math/wrap_Math.cpp"
   $16 = ($15|0)>(0); //@line 246 "modules/math/wrap_Math.cpp"
   if ($16) {
    $i2$04 = 1;$numcomponents$13 = 0;
    while(1) {
     $17 = (+_luaL_checknumber($L,$i2$04)); //@line 248 "modules/math/wrap_Math.cpp"
     $18 = $17; //@line 248 "modules/math/wrap_Math.cpp"
     $19 = $18 / 255.0; //@line 248 "modules/math/wrap_Math.cpp"
     $20 = (($i2$04) + -1)|0; //@line 248 "modules/math/wrap_Math.cpp"
     $21 = (($color) + ($20<<2)|0); //@line 248 "modules/math/wrap_Math.cpp"
     HEAPF32[$21>>2] = $19; //@line 248 "modules/math/wrap_Math.cpp"
     $22 = (($numcomponents$13) + 1)|0; //@line 249 "modules/math/wrap_Math.cpp"
     $23 = (($i2$04) + 1)|0; //@line 246 "modules/math/wrap_Math.cpp"
     $24 = ($i2$04|0)<($15|0); //@line 246 "modules/math/wrap_Math.cpp"
     $25 = ($23|0)<(5); //@line 246 "modules/math/wrap_Math.cpp"
     $26 = $24 & $25; //@line 246 "modules/math/wrap_Math.cpp"
     if ($26) {
      $i2$04 = $23;$numcomponents$13 = $22;
     } else {
      $numcomponents$29 = $22;
      break;
     }
    }
    return ($numcomponents$29|0); //@line 256 "modules/math/wrap_Math.cpp"
   }
  }
 } while(0);
 (+_luaL_checknumber($L,1)); //@line 254 "modules/math/wrap_Math.cpp"
 $numcomponents$29 = 0;
 return ($numcomponents$29|0); //@line 256 "modules/math/wrap_Math.cpp"
}
function __ZN4love4math15w_linearToGammaEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0, $color = 0, $exitcond = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $color = sp;
 $0 = (__ZN4love4mathL12getGammaArgsEP9lua_StatePf($L,$color)|0); //@line 278 "modules/math/wrap_Math.cpp"
 $1 = ($0|0)>(0); //@line 280 "modules/math/wrap_Math.cpp"
 if ($1) {
  $i$01 = 0;
 } else {
  STACKTOP = sp;return ($0|0); //@line 288 "modules/math/wrap_Math.cpp"
 }
 while(1) {
  $2 = ($i$01|0)<(3); //@line 283 "modules/math/wrap_Math.cpp"
  $3 = (($color) + ($i$01<<2)|0); //@line 284 "modules/math/wrap_Math.cpp"
  $4 = +HEAPF32[$3>>2]; //@line 284 "modules/math/wrap_Math.cpp"
  if ($2) {
   $5 = (+__ZNK4love4math4Math13linearToGammaEf(623984,$4)); //@line 284 "modules/math/wrap_Math.cpp"
   HEAPF32[$3>>2] = $5; //@line 284 "modules/math/wrap_Math.cpp"
   $7 = $5;
  } else {
   $7 = $4;
  }
  $6 = $7 * 255.0; //@line 285 "modules/math/wrap_Math.cpp"
  $8 = $6; //@line 285 "modules/math/wrap_Math.cpp"
  _lua_pushnumber($L,$8); //@line 285 "modules/math/wrap_Math.cpp"
  $9 = (($i$01) + 1)|0; //@line 280 "modules/math/wrap_Math.cpp"
  $exitcond = ($9|0)==($0|0); //@line 280 "modules/math/wrap_Math.cpp"
  if ($exitcond) {
   break;
  } else {
   $i$01 = $9;
  }
 }
 STACKTOP = sp;return ($0|0); //@line 288 "modules/math/wrap_Math.cpp"
}
function __ZN4love4math7w_noiseEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0;
 var $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0;
 var $args = 0, $i$01 = 0, $val$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $args = sp;
 $0 = (_lua_gettop($L)|0); //@line 293 "modules/math/wrap_Math.cpp"
 $1 = ($0|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $2 = $1 ? 1 : $0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $3 = ($2|0)>(4); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $4 = $3 ? 4 : $2; //@line 293 "modules/math/wrap_Math.cpp"
 $5 = ($4|0)>(0); //@line 296 "modules/math/wrap_Math.cpp"
 if ($5) {
  $i$01 = 0;
  while(1) {
   $6 = (($i$01) + 1)|0; //@line 297 "modules/math/wrap_Math.cpp"
   $7 = (+_luaL_checknumber($L,$6)); //@line 297 "modules/math/wrap_Math.cpp"
   $8 = $7; //@line 297 "modules/math/wrap_Math.cpp"
   $9 = (($args) + ($i$01<<2)|0); //@line 297 "modules/math/wrap_Math.cpp"
   HEAPF32[$9>>2] = $8; //@line 297 "modules/math/wrap_Math.cpp"
   $10 = ($6|0)<($4|0); //@line 296 "modules/math/wrap_Math.cpp"
   if ($10) {
    $i$01 = $6;
   } else {
    break;
   }
  }
 }
 switch ($4|0) {
 case 1:  {
  $11 = +HEAPF32[$args>>2]; //@line 304 "modules/math/wrap_Math.cpp"
  $12 = (+__ZN16SimplexNoise12345noiseEf($11)); //@line 165 "modules/math/MathModule.h"
  $13 = $12 * 0.5; //@line 165 "modules/math/MathModule.h"
  $14 = $13 + 0.5; //@line 165 "modules/math/MathModule.h"
  $val$0 = $14;
  break;
 }
 case 2:  {
  $15 = +HEAPF32[$args>>2]; //@line 307 "modules/math/wrap_Math.cpp"
  $16 = ((($args)) + 4|0); //@line 307 "modules/math/wrap_Math.cpp"
  $17 = +HEAPF32[$16>>2]; //@line 307 "modules/math/wrap_Math.cpp"
  $18 = (+__ZN16SimplexNoise12345noiseEff($15,$17)); //@line 170 "modules/math/MathModule.h"
  $19 = $18 * 0.5; //@line 170 "modules/math/MathModule.h"
  $20 = $19 + 0.5; //@line 170 "modules/math/MathModule.h"
  $val$0 = $20;
  break;
 }
 case 3:  {
  $21 = +HEAPF32[$args>>2]; //@line 310 "modules/math/wrap_Math.cpp"
  $22 = ((($args)) + 4|0); //@line 310 "modules/math/wrap_Math.cpp"
  $23 = +HEAPF32[$22>>2]; //@line 310 "modules/math/wrap_Math.cpp"
  $24 = ((($args)) + 8|0); //@line 310 "modules/math/wrap_Math.cpp"
  $25 = +HEAPF32[$24>>2]; //@line 310 "modules/math/wrap_Math.cpp"
  $26 = (+__ZN9Noise12345noiseEfff($21,$23,$25)); //@line 178 "modules/math/MathModule.h"
  $27 = $26 * 0.5; //@line 178 "modules/math/MathModule.h"
  $28 = $27 + 0.5; //@line 178 "modules/math/MathModule.h"
  $val$0 = $28;
  break;
 }
 case 4:  {
  $29 = +HEAPF32[$args>>2]; //@line 313 "modules/math/wrap_Math.cpp"
  $30 = ((($args)) + 4|0); //@line 313 "modules/math/wrap_Math.cpp"
  $31 = +HEAPF32[$30>>2]; //@line 313 "modules/math/wrap_Math.cpp"
  $32 = ((($args)) + 8|0); //@line 313 "modules/math/wrap_Math.cpp"
  $33 = +HEAPF32[$32>>2]; //@line 313 "modules/math/wrap_Math.cpp"
  $34 = ((($args)) + 12|0); //@line 313 "modules/math/wrap_Math.cpp"
  $35 = +HEAPF32[$34>>2]; //@line 313 "modules/math/wrap_Math.cpp"
  $36 = (+__ZN9Noise12345noiseEffff($29,$31,$33,$35)); //@line 183 "modules/math/MathModule.h"
  $37 = $36 * 0.5; //@line 183 "modules/math/MathModule.h"
  $38 = $37 + 0.5; //@line 183 "modules/math/MathModule.h"
  $val$0 = $38;
  break;
 }
 default: {
  $val$0 = 0.0;
 }
 }
 $39 = $val$0; //@line 317 "modules/math/wrap_Math.cpp"
 _lua_pushnumber($L,$39); //@line 317 "modules/math/wrap_Math.cpp"
 STACKTOP = sp;return 1; //@line 318 "modules/math/wrap_Math.cpp"
}
function __ZN4love4math10w_compressEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$01$i2 = 0, $$02$i = 0, $$02$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, $cdata$0 = 0, $cdata$2 = 0, $format = 0, $rawsize = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $format = sp + 24|0;
 $rawsize = sp + 20|0;
 $0 = (_lua_type($L,2)|0); //@line 323 "modules/math/wrap_Math.cpp"
 $1 = ($0|0)<(1); //@line 323 "modules/math/wrap_Math.cpp"
 if ($1) {
  HEAP32[$format>>2] = 0; //@line 324 "modules/math/wrap_Math.cpp"
  $57 = $format;
 } else {
  $2 = (_luaL_checklstring($L,2,0)|0); //@line 323 "modules/math/wrap_Math.cpp"
  HEAP32[$format>>2] = 0; //@line 324 "modules/math/wrap_Math.cpp"
  $3 = ($2|0)==(0|0); //@line 326 "modules/math/wrap_Math.cpp"
  if ($3) {
   $57 = $format;
  } else {
   $4 = (__ZN4love4math10Compressor11getConstantEPKcRNS1_6FormatE($2,$format)|0); //@line 326 "modules/math/wrap_Math.cpp"
   if ($4) {
    $57 = $format;
   } else {
    HEAP32[$vararg_buffer>>2] = $2; //@line 327 "modules/math/wrap_Math.cpp"
    $5 = (_luaL_error($L,253984,$vararg_buffer)|0); //@line 327 "modules/math/wrap_Math.cpp"
    $$0 = $5;$58 = $format;
    STACKTOP = sp;return ($$0|0); //@line 346 "modules/math/wrap_Math.cpp"
   }
  }
 }
 $6 = (+_luaL_optnumber($L,3,-1.0)); //@line 329 "modules/math/wrap_Math.cpp"
 $7 = (~~(($6))); //@line 329 "modules/math/wrap_Math.cpp"
 $8 = (_lua_isstring($L,1)|0); //@line 332 "modules/math/wrap_Math.cpp"
 $9 = ($8|0)==(0); //@line 332 "modules/math/wrap_Math.cpp"
 do {
  if ($9) {
   $34 = (__ZN4love14luax_checktypeINS_4DataEEEPT_P9lua_StateiNS_4TypeE($L,1,2)|0); //@line 340 "modules/math/wrap_Math.cpp"
   $35 = HEAP32[$format>>2]|0; //@line 341 "modules/math/wrap_Math.cpp"
   __THREW__ = 0;
   $36 = (invoke_iiiii(3399,(623984|0),($35|0),($34|0),($7|0))|0); //@line 341 "modules/math/wrap_Math.cpp"
   $37 = __THREW__; __THREW__ = 0;
   $38 = $37&1;
   if ($38) {
    $39 = ___cxa_find_matching_catch_3(8992|0)|0;
    $40 = tempRet0;
    $41 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
    $42 = ($40|0)==($41|0); //@line 535 "./common/runtime.h"
    if (!($42)) {
     $$01$i2 = $40;$$02$i1 = $39;
     ___resumeException($$02$i1|0);
     // unreachable;
    }
    $43 = (___cxa_begin_catch(($39|0))|0); //@line 535 "./common/runtime.h"
    $44 = HEAP32[$43>>2]|0; //@line 539 "./common/runtime.h"
    $45 = ((($44)) + 8|0); //@line 539 "./common/runtime.h"
    $46 = HEAP32[$45>>2]|0; //@line 539 "./common/runtime.h"
    $47 = (FUNCTION_TABLE_ii[$46 & 8191]($43)|0); //@line 539 "./common/runtime.h"
    __THREW__ = 0;
    invoke_vii(3380,($L|0),($47|0)); //@line 539 "./common/runtime.h"
    $48 = __THREW__; __THREW__ = 0;
    $49 = $48&1;
    if (!($49)) {
     ___cxa_end_catch(); //@line 540 "./common/runtime.h"
     $50 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
     HEAP32[$vararg_buffer4>>2] = $50; //@line 543 "./common/runtime.h"
     (_luaL_error($L,595867,$vararg_buffer4)|0); //@line 543 "./common/runtime.h"
     $cdata$2 = 0;
     break;
    }
    $51 = ___cxa_find_matching_catch_2()|0;
    $52 = tempRet0;
    __THREW__ = 0;
    invoke_v(3378); //@line 540 "./common/runtime.h"
    $53 = __THREW__; __THREW__ = 0;
    $54 = $53&1;
    if ($54) {
     $55 = ___cxa_find_matching_catch_3(0|0)|0;
     $56 = tempRet0;
     ___clang_call_terminate($55); //@line 540 "./common/runtime.h"
     // unreachable; //@line 540 "./common/runtime.h"
    } else {
     $$01$i2 = $52;$$02$i1 = $51;
     ___resumeException($$02$i1|0);
     // unreachable;
    }
   } else {
    $cdata$2 = $36;
   }
  } else {
   HEAP32[$rawsize>>2] = 0; //@line 334 "modules/math/wrap_Math.cpp"
   $10 = (_luaL_checklstring($L,1,$rawsize)|0); //@line 335 "modules/math/wrap_Math.cpp"
   $11 = HEAP32[$format>>2]|0; //@line 336 "modules/math/wrap_Math.cpp"
   $12 = HEAP32[$rawsize>>2]|0; //@line 336 "modules/math/wrap_Math.cpp"
   __THREW__ = 0;
   $13 = (invoke_iiiiii(3398,(623984|0),($11|0),($10|0),($12|0),($7|0))|0); //@line 336 "modules/math/wrap_Math.cpp"
   $14 = __THREW__; __THREW__ = 0;
   $15 = $14&1;
   do {
    if ($15) {
     $16 = ___cxa_find_matching_catch_3(8992|0)|0;
     $17 = tempRet0;
     $18 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
     $19 = ($17|0)==($18|0); //@line 535 "./common/runtime.h"
     if (!($19)) {
      $$01$i = $17;$$02$i = $16;
      ___resumeException($$02$i|0);
      // unreachable;
     }
     $20 = (___cxa_begin_catch(($16|0))|0); //@line 535 "./common/runtime.h"
     $21 = HEAP32[$20>>2]|0; //@line 539 "./common/runtime.h"
     $22 = ((($21)) + 8|0); //@line 539 "./common/runtime.h"
     $23 = HEAP32[$22>>2]|0; //@line 539 "./common/runtime.h"
     $24 = (FUNCTION_TABLE_ii[$23 & 8191]($20)|0); //@line 539 "./common/runtime.h"
     __THREW__ = 0;
     invoke_vii(3380,($L|0),($24|0)); //@line 539 "./common/runtime.h"
     $25 = __THREW__; __THREW__ = 0;
     $26 = $25&1;
     if (!($26)) {
      ___cxa_end_catch(); //@line 540 "./common/runtime.h"
      $27 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
      HEAP32[$vararg_buffer1>>2] = $27; //@line 543 "./common/runtime.h"
      (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
      $cdata$0 = 0;
      break;
     }
     $28 = ___cxa_find_matching_catch_2()|0;
     $29 = tempRet0;
     __THREW__ = 0;
     invoke_v(3378); //@line 540 "./common/runtime.h"
     $30 = __THREW__; __THREW__ = 0;
     $31 = $30&1;
     if ($31) {
      $32 = ___cxa_find_matching_catch_3(0|0)|0;
      $33 = tempRet0;
      ___clang_call_terminate($32); //@line 540 "./common/runtime.h"
      // unreachable; //@line 540 "./common/runtime.h"
     } else {
      $$01$i = $29;$$02$i = $28;
      ___resumeException($$02$i|0);
      // unreachable;
     }
    } else {
     $cdata$0 = $13;
    }
   } while(0);
   $cdata$2 = $cdata$0;
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,27,$cdata$2); //@line 344 "modules/math/wrap_Math.cpp"
 $$0 = 1;$58 = $57;
 STACKTOP = sp;return ($$0|0); //@line 346 "modules/math/wrap_Math.cpp"
}
function __ZN4love14luax_checktypeINS_4DataEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love4math12w_decompressEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$01$i3 = 0, $$02$i = 0, $$02$i2 = 0, $$1 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, $cbytes$0 = 0, $compressedsize = 0, $format = 0, $rawbytes$1 = 0, $rawbytes$3 = 0, $rawsize = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $rawsize = sp + 28|0;
 $format = sp + 24|0;
 $compressedsize = sp + 20|0;
 HEAP32[$rawsize>>2] = 0; //@line 351 "modules/math/wrap_Math.cpp"
 $0 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,1,27)|0); //@line 353 "modules/math/wrap_Math.cpp"
 if ($0) {
  $1 = (__ZN4love4math24luax_checkcompresseddataEP9lua_Statei($L,1)|0); //@line 355 "modules/math/wrap_Math.cpp"
  $2 = (__ZNK4love4math14CompressedData19getDecompressedSizeEv($1)|0); //@line 356 "modules/math/wrap_Math.cpp"
  HEAP32[$rawsize>>2] = $2; //@line 356 "modules/math/wrap_Math.cpp"
  __THREW__ = 0;
  $3 = (invoke_iiii(3400,(623984|0),($1|0),($rawsize|0))|0); //@line 357 "modules/math/wrap_Math.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $6 = ___cxa_find_matching_catch_3(8992|0)|0;
   $7 = tempRet0;
   $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
   if (!($9)) {
    $$01$i = $7;$$02$i = $6;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
   $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
   $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
   $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
   $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if (!($16)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $17; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $18 = HEAP32[$rawsize>>2]|0; //@line 382 "modules/math/wrap_Math.cpp"
    _lua_pushlstring($L,0,$18); //@line 382 "modules/math/wrap_Math.cpp"
    $$1 = 1;
    STACKTOP = sp;return ($$1|0); //@line 386 "modules/math/wrap_Math.cpp"
   }
   $19 = ___cxa_find_matching_catch_2()|0;
   $20 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $21 = __THREW__; __THREW__ = 0;
   $22 = $21&1;
   if ($22) {
    $23 = ___cxa_find_matching_catch_3(0|0)|0;
    $24 = tempRet0;
    ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $20;$$02$i = $19;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $rawbytes$3 = $3;
  }
 } else {
  HEAP32[$format>>2] = 0; //@line 361 "modules/math/wrap_Math.cpp"
  $25 = (_luaL_checklstring($L,2,0)|0); //@line 362 "modules/math/wrap_Math.cpp"
  $26 = (__ZN4love4math10Compressor11getConstantEPKcRNS1_6FormatE($25,$format)|0); //@line 364 "modules/math/wrap_Math.cpp"
  if (!($26)) {
   HEAP32[$vararg_buffer4>>2] = $25; //@line 365 "modules/math/wrap_Math.cpp"
   $61 = (_luaL_error($L,253984,$vararg_buffer4)|0); //@line 365 "modules/math/wrap_Math.cpp"
   $$1 = $61;
   STACKTOP = sp;return ($$1|0); //@line 386 "modules/math/wrap_Math.cpp"
  }
  HEAP32[$compressedsize>>2] = 0; //@line 367 "modules/math/wrap_Math.cpp"
  $27 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,1,2)|0); //@line 370 "modules/math/wrap_Math.cpp"
  if ($27) {
   $28 = (__ZN4love14luax_checktypeINS_4DataEEEPT_P9lua_StateiNS_4TypeE($L,1,2)|0); //@line 372 "modules/math/wrap_Math.cpp"
   $29 = HEAP32[$28>>2]|0; //@line 373 "modules/math/wrap_Math.cpp"
   $30 = ((($29)) + 16|0); //@line 373 "modules/math/wrap_Math.cpp"
   $31 = HEAP32[$30>>2]|0; //@line 373 "modules/math/wrap_Math.cpp"
   $32 = (FUNCTION_TABLE_ii[$31 & 8191]($28)|0); //@line 373 "modules/math/wrap_Math.cpp"
   $33 = HEAP32[$28>>2]|0; //@line 374 "modules/math/wrap_Math.cpp"
   $34 = ((($33)) + 20|0); //@line 374 "modules/math/wrap_Math.cpp"
   $35 = HEAP32[$34>>2]|0; //@line 374 "modules/math/wrap_Math.cpp"
   $36 = (FUNCTION_TABLE_ii[$35 & 8191]($28)|0); //@line 374 "modules/math/wrap_Math.cpp"
   HEAP32[$compressedsize>>2] = $36; //@line 374 "modules/math/wrap_Math.cpp"
   $39 = $36;$cbytes$0 = $32;
  } else {
   $37 = (_luaL_checklstring($L,1,$compressedsize)|0); //@line 377 "modules/math/wrap_Math.cpp"
   $$pre = HEAP32[$compressedsize>>2]|0; //@line 379 "modules/math/wrap_Math.cpp"
   $39 = $$pre;$cbytes$0 = $37;
  }
  $38 = HEAP32[$format>>2]|0; //@line 379 "modules/math/wrap_Math.cpp"
  __THREW__ = 0;
  $40 = (invoke_iiiiii(3401,(623984|0),($38|0),($cbytes$0|0),($39|0),($rawsize|0))|0); //@line 379 "modules/math/wrap_Math.cpp"
  $41 = __THREW__; __THREW__ = 0;
  $42 = $41&1;
  do {
   if ($42) {
    $43 = ___cxa_find_matching_catch_3(8992|0)|0;
    $44 = tempRet0;
    $45 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
    $46 = ($44|0)==($45|0); //@line 535 "./common/runtime.h"
    if (!($46)) {
     $$01$i3 = $44;$$02$i2 = $43;
     ___resumeException($$02$i2|0);
     // unreachable;
    }
    $47 = (___cxa_begin_catch(($43|0))|0); //@line 535 "./common/runtime.h"
    $48 = HEAP32[$47>>2]|0; //@line 539 "./common/runtime.h"
    $49 = ((($48)) + 8|0); //@line 539 "./common/runtime.h"
    $50 = HEAP32[$49>>2]|0; //@line 539 "./common/runtime.h"
    $51 = (FUNCTION_TABLE_ii[$50 & 8191]($47)|0); //@line 539 "./common/runtime.h"
    __THREW__ = 0;
    invoke_vii(3380,($L|0),($51|0)); //@line 539 "./common/runtime.h"
    $52 = __THREW__; __THREW__ = 0;
    $53 = $52&1;
    if (!($53)) {
     ___cxa_end_catch(); //@line 540 "./common/runtime.h"
     $54 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
     HEAP32[$vararg_buffer1>>2] = $54; //@line 543 "./common/runtime.h"
     (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
     $rawbytes$1 = 0;
     break;
    }
    $55 = ___cxa_find_matching_catch_2()|0;
    $56 = tempRet0;
    __THREW__ = 0;
    invoke_v(3378); //@line 540 "./common/runtime.h"
    $57 = __THREW__; __THREW__ = 0;
    $58 = $57&1;
    if ($58) {
     $59 = ___cxa_find_matching_catch_3(0|0)|0;
     $60 = tempRet0;
     ___clang_call_terminate($59); //@line 540 "./common/runtime.h"
     // unreachable; //@line 540 "./common/runtime.h"
    } else {
     $$01$i3 = $56;$$02$i2 = $55;
     ___resumeException($$02$i2|0);
     // unreachable;
    }
   } else {
    $rawbytes$1 = $40;
   }
  } while(0);
  $rawbytes$3 = $rawbytes$1;
 }
 $62 = HEAP32[$rawsize>>2]|0; //@line 382 "modules/math/wrap_Math.cpp"
 _lua_pushlstring($L,$rawbytes$3,$62); //@line 382 "modules/math/wrap_Math.cpp"
 $63 = ($rawbytes$3|0)==(0|0); //@line 383 "modules/math/wrap_Math.cpp"
 if ($63) {
  $$1 = 1;
  STACKTOP = sp;return ($$1|0); //@line 386 "modules/math/wrap_Math.cpp"
 }
 __ZdaPv($rawbytes$3); //@line 383 "modules/math/wrap_Math.cpp"
 $$1 = 1;
 STACKTOP = sp;return ($$1|0); //@line 386 "modules/math/wrap_Math.cpp"
}
function _luaopen_love_math($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $w = sp;
 $0 = HEAP32[155996]|0; //@line 457 "modules/math/wrap_Math.cpp"
 $1 = ((($0)) + 8|0); //@line 457 "modules/math/wrap_Math.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 457 "modules/math/wrap_Math.cpp"
 FUNCTION_TABLE_vi[$2 & 8191](623984); //@line 457 "modules/math/wrap_Math.cpp"
 HEAP32[$w>>2] = 623984; //@line 460 "modules/math/wrap_Math.cpp"
 $3 = ((($w)) + 4|0); //@line 461 "modules/math/wrap_Math.cpp"
 HEAP32[$3>>2] = 521854; //@line 461 "modules/math/wrap_Math.cpp"
 $4 = ((($w)) + 8|0); //@line 462 "modules/math/wrap_Math.cpp"
 HEAP32[$4>>2] = 3; //@line 462 "modules/math/wrap_Math.cpp"
 $5 = ((($w)) + 12|0); //@line 463 "modules/math/wrap_Math.cpp"
 HEAP32[$5>>2] = 9760; //@line 463 "modules/math/wrap_Math.cpp"
 $6 = ((($w)) + 16|0); //@line 464 "modules/math/wrap_Math.cpp"
 HEAP32[$6>>2] = 9848; //@line 464 "modules/math/wrap_Math.cpp"
 $7 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 466 "modules/math/wrap_Math.cpp"
 (_luaL_loadbuffer($L,254148,3426,257574)|0); //@line 469 "modules/math/wrap_Math.cpp"
 _lua_pushvalue($L,-2); //@line 470 "modules/math/wrap_Math.cpp"
 _lua_pushlightuserdata($L,9864); //@line 471 "modules/math/wrap_Math.cpp"
 _lua_call($L,2,0); //@line 472 "modules/math/wrap_Math.cpp"
 STACKTOP = sp;return ($7|0); //@line 474 "modules/math/wrap_Math.cpp"
}
function __ZN4love4math3__58__invokeEf($x) {
 $x = +$x;
 var $0 = 0.0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+__ZN16SimplexNoise12345noiseEf($x)); //@line 165 "modules/math/MathModule.h"
 $1 = $0 * 0.5; //@line 165 "modules/math/MathModule.h"
 $2 = $1 + 0.5; //@line 165 "modules/math/MathModule.h"
 return (+$2); //@line 402 "modules/math/wrap_Math.cpp"
}
function __ZN4love4math3__68__invokeEff($x,$y) {
 $x = +$x;
 $y = +$y;
 var $0 = 0.0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+__ZN16SimplexNoise12345noiseEff($x,$y)); //@line 170 "modules/math/MathModule.h"
 $1 = $0 * 0.5; //@line 170 "modules/math/MathModule.h"
 $2 = $1 + 0.5; //@line 170 "modules/math/MathModule.h"
 return (+$2); //@line 406 "modules/math/wrap_Math.cpp"
}
function __ZN4love4math3__78__invokeEfff($x,$y,$z) {
 $x = +$x;
 $y = +$y;
 $z = +$z;
 var $0 = 0.0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+__ZN9Noise12345noiseEfff($x,$y,$z)); //@line 178 "modules/math/MathModule.h"
 $1 = $0 * 0.5; //@line 178 "modules/math/MathModule.h"
 $2 = $1 + 0.5; //@line 178 "modules/math/MathModule.h"
 return (+$2); //@line 410 "modules/math/wrap_Math.cpp"
}
function __ZN4love4math3__88__invokeEffff($x,$y,$z,$w) {
 $x = +$x;
 $y = +$y;
 $z = +$z;
 $w = +$w;
 var $0 = 0.0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+__ZN9Noise12345noiseEffff($x,$y,$z,$w)); //@line 183 "modules/math/MathModule.h"
 $1 = $0 * 0.5; //@line 183 "modules/math/MathModule.h"
 $2 = $1 + 0.5; //@line 183 "modules/math/MathModule.h"
 return (+$2); //@line 414 "modules/math/wrap_Math.cpp"
}
function __ZN4love4math3__98__invokeEf($c) {
 $c = +$c;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+__ZNK4love4math4Math13gammaToLinearEf(623984,$c)); //@line 421 "modules/math/wrap_Math.cpp"
 return (+$0); //@line 419 "modules/math/wrap_Math.cpp"
}
function __ZN4love4math4__108__invokeEf($c) {
 $c = +$c;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+__ZNK4love4math4Math13linearToGammaEf(623984,$c)); //@line 425 "modules/math/wrap_Math.cpp"
 return (+$0); //@line 423 "modules/math/wrap_Math.cpp"
}
function __ZN4love4math20luax_checkrandomseedEP9lua_Statei($agg$result,$L,$idx) {
 $agg$result = $agg$result|0;
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0;
 var $or$cond$i1 = 0, $or$cond$i4 = 0, $or$cond1$i2 = 0, $or$cond1$i5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($idx) + 1)|0; //@line 53 "modules/math/wrap_RandomGenerator.cpp"
 $1 = (_lua_type($L,$0)|0); //@line 53 "modules/math/wrap_RandomGenerator.cpp"
 $2 = ($1|0)<(1); //@line 53 "modules/math/wrap_RandomGenerator.cpp"
 $3 = (+_luaL_checknumber($L,$idx)); //@line 39 "modules/math/wrap_RandomGenerator.cpp"
 $4 = $3 == inf; //@line 43 "modules/math/wrap_RandomGenerator.cpp"
 $5 = $3 == -inf; //@line 43 "modules/math/wrap_RandomGenerator.cpp"
 $or$cond$i4 = $4 | $5; //@line 43 "modules/math/wrap_RandomGenerator.cpp"
 $6 = ($3 != $3) | (0.0 != 0.0); //@line 43 "modules/math/wrap_RandomGenerator.cpp"
 $or$cond1$i5 = $6 | $or$cond$i4; //@line 43 "modules/math/wrap_RandomGenerator.cpp"
 if ($2) {
  if ($or$cond1$i5) {
   (_luaL_argerror($L,$idx,257588)|0); //@line 44 "modules/math/wrap_RandomGenerator.cpp"
  }
  $14 = (~~$3)>>>0; //@line 46 "modules/math/wrap_RandomGenerator.cpp"
  $15 = +Math_abs($3) >= 1.0 ? $3 > 0.0 ? (~~+Math_min(+Math_floor($3 / 4294967296.0), 4294967295.0)) >>> 0 : ~~+Math_ceil(($3 - +(~~$3 >>> 0)) / 4294967296.0) >>> 0 : 0; //@line 46 "modules/math/wrap_RandomGenerator.cpp"
  $16 = $agg$result; //@line 59 "modules/math/wrap_RandomGenerator.cpp"
  $17 = $16; //@line 59 "modules/math/wrap_RandomGenerator.cpp"
  HEAP32[$17>>2] = $14; //@line 59 "modules/math/wrap_RandomGenerator.cpp"
  $18 = (($16) + 4)|0; //@line 59 "modules/math/wrap_RandomGenerator.cpp"
  $19 = $18; //@line 59 "modules/math/wrap_RandomGenerator.cpp"
  HEAP32[$19>>2] = $15; //@line 59 "modules/math/wrap_RandomGenerator.cpp"
  return; //@line 61 "modules/math/wrap_RandomGenerator.cpp"
 }
 if ($or$cond1$i5) {
  (_luaL_argerror($L,$idx,257588)|0); //@line 44 "modules/math/wrap_RandomGenerator.cpp"
 }
 $7 = (~~(($3))>>>0); //@line 46 "modules/math/wrap_RandomGenerator.cpp"
 HEAP32[$agg$result>>2] = $7; //@line 55 "modules/math/wrap_RandomGenerator.cpp"
 $8 = (+_luaL_checknumber($L,$0)); //@line 39 "modules/math/wrap_RandomGenerator.cpp"
 $9 = $8 == inf; //@line 43 "modules/math/wrap_RandomGenerator.cpp"
 $10 = $8 == -inf; //@line 43 "modules/math/wrap_RandomGenerator.cpp"
 $or$cond$i1 = $9 | $10; //@line 43 "modules/math/wrap_RandomGenerator.cpp"
 $11 = ($8 != $8) | (0.0 != 0.0); //@line 43 "modules/math/wrap_RandomGenerator.cpp"
 $or$cond1$i2 = $11 | $or$cond$i1; //@line 43 "modules/math/wrap_RandomGenerator.cpp"
 if ($or$cond1$i2) {
  (_luaL_argerror($L,$0,257588)|0); //@line 44 "modules/math/wrap_RandomGenerator.cpp"
 }
 $12 = (~~(($8))>>>0); //@line 46 "modules/math/wrap_RandomGenerator.cpp"
 $13 = ((($agg$result)) + 4|0); //@line 56 "modules/math/wrap_RandomGenerator.cpp"
 HEAP32[$13>>2] = $12; //@line 56 "modules/math/wrap_RandomGenerator.cpp"
 return; //@line 61 "modules/math/wrap_RandomGenerator.cpp"
}
function __ZN4love14luax_checktypeINS_4math15RandomGeneratorEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love4math25w_RandomGenerator__randomEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4math15RandomGeneratorEEEPT_P9lua_StateiNS_4TypeE($L,1,25)|0); //@line 66 "modules/math/wrap_RandomGenerator.cpp"
 $1 = (__ZN4love4math15RandomGenerator4randEv($0)|0); //@line 76 "modules/math/RandomGenerator.h"
 $2 = tempRet0; //@line 76 "modules/math/RandomGenerator.h"
 $3 = (+($1>>>0)) + (4294967296.0*(+($2>>>0))); //@line 76 "modules/math/RandomGenerator.h"
 $4 = $3 * 5.4210108624275222E-20; //@line 76 "modules/math/RandomGenerator.h"
 _lua_pushnumber($L,$4); //@line 72 "modules/math/wrap_RandomGenerator.cpp"
 return 1; //@line 73 "modules/math/wrap_RandomGenerator.cpp"
}
function __ZN4love4math30w_RandomGenerator_randomNormalEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4math15RandomGeneratorEEEPT_P9lua_StateiNS_4TypeE($L,1,25)|0); //@line 66 "modules/math/wrap_RandomGenerator.cpp"
 $1 = (+_luaL_optnumber($L,2,1.0)); //@line 80 "modules/math/wrap_RandomGenerator.cpp"
 $2 = (+_luaL_optnumber($L,3,0.0)); //@line 81 "modules/math/wrap_RandomGenerator.cpp"
 $3 = (+__ZN4love4math15RandomGenerator12randomNormalEd($0,$1)); //@line 82 "modules/math/wrap_RandomGenerator.cpp"
 $4 = $2 + $3; //@line 84 "modules/math/wrap_RandomGenerator.cpp"
 _lua_pushnumber($L,$4); //@line 84 "modules/math/wrap_RandomGenerator.cpp"
 return 1; //@line 85 "modules/math/wrap_RandomGenerator.cpp"
}
function __ZN4love4math25w_RandomGenerator_setSeedEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 16|0;
 $vararg_buffer = sp + 8|0;
 $0 = sp;
 $1 = (__ZN4love14luax_checktypeINS_4math15RandomGeneratorEEEPT_P9lua_StateiNS_4TypeE($L,1,25)|0); //@line 66 "modules/math/wrap_RandomGenerator.cpp"
 __THREW__ = 0;
 invoke_viii(3402,($0|0),($L|0),2); //@line 91 "modules/math/wrap_RandomGenerator.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  __THREW__ = 0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;HEAP32[$$byval_copy+4>>2]=HEAP32[$0+4>>2]|0; //@line 91 "modules/math/wrap_RandomGenerator.cpp"
  invoke_vii(3379,($1|0),($$byval_copy|0)); //@line 91 "modules/math/wrap_RandomGenerator.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if (!($5)) {
   STACKTOP = sp;return 0; //@line 92 "modules/math/wrap_RandomGenerator.cpp"
  }
 }
 $6 = ___cxa_find_matching_catch_3(8992|0)|0;
 $7 = tempRet0;
 $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
 if (!($9)) {
  $$01$i = $7;$$02$i = $6;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $17; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 92 "modules/math/wrap_RandomGenerator.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love4math25w_RandomGenerator_getSeedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0, $4 = 0, $5 = 0.0, $s = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $s = sp;
 $0 = (__ZN4love14luax_checktypeINS_4math15RandomGeneratorEEEPT_P9lua_StateiNS_4TypeE($L,1,25)|0); //@line 66 "modules/math/wrap_RandomGenerator.cpp"
 __ZNK4love4math15RandomGenerator7getSeedEv($s,$0); //@line 98 "modules/math/wrap_RandomGenerator.cpp"
 $1 = HEAP32[$s>>2]|0; //@line 99 "modules/math/wrap_RandomGenerator.cpp"
 $2 = (+($1>>>0)); //@line 99 "modules/math/wrap_RandomGenerator.cpp"
 _lua_pushnumber($L,$2); //@line 99 "modules/math/wrap_RandomGenerator.cpp"
 $3 = ((($s)) + 4|0); //@line 100 "modules/math/wrap_RandomGenerator.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 100 "modules/math/wrap_RandomGenerator.cpp"
 $5 = (+($4>>>0)); //@line 100 "modules/math/wrap_RandomGenerator.cpp"
 _lua_pushnumber($L,$5); //@line 100 "modules/math/wrap_RandomGenerator.cpp"
 STACKTOP = sp;return 2; //@line 101 "modules/math/wrap_RandomGenerator.cpp"
}
function __ZN4love4math26w_RandomGenerator_setStateEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 $1 = (__ZN4love14luax_checktypeINS_4math15RandomGeneratorEEEPT_P9lua_StateiNS_4TypeE($L,1,25)|0); //@line 66 "modules/math/wrap_RandomGenerator.cpp"
 __THREW__ = 0;
 invoke_viii(3403,($0|0),($L|0),2); //@line 107 "modules/math/wrap_RandomGenerator.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 do {
  if ($3) {
   $8 = ___cxa_find_matching_catch_3(8992|0)|0;
   $9 = tempRet0;
   $eh$lpad$body$i$index2Z2D = $9;$eh$lpad$body$i$indexZ2D = $8;
  } else {
   __THREW__ = 0;
   invoke_vii(3404,($1|0),($0|0)); //@line 107 "modules/math/wrap_RandomGenerator.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   if ($5) {
    $6 = ___cxa_find_matching_catch_3(8992|0)|0;
    $7 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 107 "modules/math/wrap_RandomGenerator.cpp"
    $eh$lpad$body$i$index2Z2D = $7;$eh$lpad$body$i$indexZ2D = $6;
    break;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 107 "modules/math/wrap_RandomGenerator.cpp"
   STACKTOP = sp;return 0; //@line 108 "modules/math/wrap_RandomGenerator.cpp"
  }
 } while(0);
 $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $11 = ($eh$lpad$body$i$index2Z2D|0)==($10|0); //@line 535 "./common/runtime.h"
 if (!($11)) {
  $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $12 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = ((($13)) + 8|0); //@line 539 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($16|0)); //@line 539 "./common/runtime.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if (!($18)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $19 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $19; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 108 "modules/math/wrap_RandomGenerator.cpp"
 }
 $20 = ___cxa_find_matching_catch_2()|0;
 $21 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if ($23) {
  $24 = ___cxa_find_matching_catch_3(0|0)|0;
  $25 = tempRet0;
  ___clang_call_terminate($24); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $21;$$02$i = $20;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love4math26w_RandomGenerator_getStateEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__ZN4love14luax_checktypeINS_4math15RandomGeneratorEEEPT_P9lua_StateiNS_4TypeE($L,1,25)|0); //@line 66 "modules/math/wrap_RandomGenerator.cpp"
 __ZNK4love4math15RandomGenerator8getStateEv($0,$1); //@line 114 "modules/math/wrap_RandomGenerator.cpp"
 __THREW__ = 0;
 invoke_vii(3405,($L|0),($0|0)); //@line 114 "modules/math/wrap_RandomGenerator.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 114 "modules/math/wrap_RandomGenerator.cpp"
  ___resumeException($4|0);
  // unreachable;
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 114 "modules/math/wrap_RandomGenerator.cpp"
  STACKTOP = sp;return 1; //@line 115 "modules/math/wrap_RandomGenerator.cpp"
 }
 return (0)|0;
}
function _luaopen_randomgenerator($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 148 "modules/math/wrap_RandomGenerator.cpp"
 HEAP32[$vararg_buffer>>2] = 9888; //@line 150 "modules/math/wrap_RandomGenerator.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 150 "modules/math/wrap_RandomGenerator.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 150 "modules/math/wrap_RandomGenerator.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,25,257608,$vararg_buffer)|0); //@line 150 "modules/math/wrap_RandomGenerator.cpp"
 __ZN4love21luax_gettypemetatableEP9lua_StateNS_4TypeE($L,25); //@line 152 "modules/math/wrap_RandomGenerator.cpp"
 $1 = (_lua_type($L,-1)|0); //@line 156 "modules/math/wrap_RandomGenerator.cpp"
 $2 = ($1|0)==(5); //@line 156 "modules/math/wrap_RandomGenerator.cpp"
 if (!($2)) {
  _lua_settop($L,-2); //@line 165 "modules/math/wrap_RandomGenerator.cpp"
  STACKTOP = sp;return ($0|0); //@line 167 "modules/math/wrap_RandomGenerator.cpp"
 }
 (_luaL_loadbuffer($L,257679,2474,260153)|0); //@line 158 "modules/math/wrap_RandomGenerator.cpp"
 _lua_pushvalue($L,-2); //@line 159 "modules/math/wrap_RandomGenerator.cpp"
 _lua_pushlightuserdata($L,9944); //@line 160 "modules/math/wrap_RandomGenerator.cpp"
 _lua_call($L,2,0); //@line 161 "modules/math/wrap_RandomGenerator.cpp"
 _lua_settop($L,-2); //@line 165 "modules/math/wrap_RandomGenerator.cpp"
 STACKTOP = sp;return ($0|0); //@line 167 "modules/math/wrap_RandomGenerator.cpp"
}
function __ZN4love4math3__28__invokeEPNS_5ProxyE($p) {
 $p = $p|0;
 var $$0$i = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($p|0)==(0|0); //@line 129 "modules/math/wrap_RandomGenerator.cpp"
 if ($0) {
  $$0$i = 0.0;
  return (+$$0$i); //@line 126 "modules/math/wrap_RandomGenerator.cpp"
 }
 $1 = HEAP32[$p>>2]|0; //@line 129 "modules/math/wrap_RandomGenerator.cpp"
 $2 = HEAP32[156504]|0; //@line 129 "modules/math/wrap_RandomGenerator.cpp"
 $3 = (($2) + ($1<<3)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $4 = HEAP32[$3>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $5 = $4 & 33554432; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $6 = ($5|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if ($6) {
  $$0$i = 0.0;
  return (+$$0$i); //@line 126 "modules/math/wrap_RandomGenerator.cpp"
 }
 $7 = ((($p)) + 4|0); //@line 132 "modules/math/wrap_RandomGenerator.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 132 "modules/math/wrap_RandomGenerator.cpp"
 $9 = (__ZN4love4math15RandomGenerator4randEv($8)|0); //@line 76 "modules/math/RandomGenerator.h"
 $10 = tempRet0; //@line 76 "modules/math/RandomGenerator.h"
 $11 = (+($9>>>0)) + (4294967296.0*(+($10>>>0))); //@line 76 "modules/math/RandomGenerator.h"
 $12 = $11 * 5.4210108624275222E-20; //@line 76 "modules/math/RandomGenerator.h"
 $$0$i = $12;
 return (+$$0$i); //@line 126 "modules/math/wrap_RandomGenerator.cpp"
}
function __ZN4love5mouse11w_newCursorEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0, $cursor$0 = 0;
 var $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 35 "modules/mouse/wrap_Mouse.cpp"
 $0 = (_lua_isstring($L,1)|0); //@line 39 "modules/mouse/wrap_Mouse.cpp"
 $1 = ($0|0)==(0); //@line 39 "modules/mouse/wrap_Mouse.cpp"
 if ($1) {
  $2 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,1,5)|0); //@line 39 "modules/mouse/wrap_Mouse.cpp"
  if ($2) {
   label = 4;
  } else {
   $3 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,1,7)|0); //@line 39 "modules/mouse/wrap_Mouse.cpp"
   if ($3) {
    label = 4;
   }
  }
 } else {
  label = 4;
 }
 if ((label|0) == 4) {
  (__ZN4love12luax_convobjEP9lua_StateiPKcS3_($L,1,438231,439339)|0); //@line 40 "modules/mouse/wrap_Mouse.cpp"
 }
 $4 = (__ZN4love14luax_checktypeINS_5image9ImageDataEEEPT_P9lua_StateiNS_4TypeE($L,1,22)|0); //@line 42 "modules/mouse/wrap_Mouse.cpp"
 $5 = (+_luaL_optnumber($L,2,0.0)); //@line 43 "modules/mouse/wrap_Mouse.cpp"
 $6 = (~~(($5))); //@line 43 "modules/mouse/wrap_Mouse.cpp"
 $7 = (+_luaL_optnumber($L,3,0.0)); //@line 44 "modules/mouse/wrap_Mouse.cpp"
 $8 = (~~(($7))); //@line 44 "modules/mouse/wrap_Mouse.cpp"
 $9 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
 $10 = HEAP32[$9>>2]|0; //@line 46 "modules/mouse/wrap_Mouse.cpp"
 $11 = ((($10)) + 24|0); //@line 46 "modules/mouse/wrap_Mouse.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 46 "modules/mouse/wrap_Mouse.cpp"
 __THREW__ = 0;
 $13 = (invoke_iiiii($12|0,($9|0),($4|0),($6|0),($8|0))|0); //@line 46 "modules/mouse/wrap_Mouse.cpp"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if (!($15)) {
  $cursor$0 = $13;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,31,$cursor$0); //@line 48 "modules/mouse/wrap_Mouse.cpp"
  $34 = HEAP32[$cursor$0>>2]|0; //@line 49 "modules/mouse/wrap_Mouse.cpp"
  $35 = ((($34)) + 12|0); //@line 49 "modules/mouse/wrap_Mouse.cpp"
  $36 = HEAP32[$35>>2]|0; //@line 49 "modules/mouse/wrap_Mouse.cpp"
  FUNCTION_TABLE_vi[$36 & 8191]($cursor$0); //@line 49 "modules/mouse/wrap_Mouse.cpp"
  STACKTOP = sp;return 1; //@line 50 "modules/mouse/wrap_Mouse.cpp"
 }
 $16 = ___cxa_find_matching_catch_3(8992|0)|0;
 $17 = tempRet0;
 $18 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $19 = ($17|0)==($18|0); //@line 535 "./common/runtime.h"
 if (!($19)) {
  $$01$i = $17;$$02$i = $16;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $20 = (___cxa_begin_catch(($16|0))|0); //@line 535 "./common/runtime.h"
 $21 = HEAP32[$20>>2]|0; //@line 539 "./common/runtime.h"
 $22 = ((($21)) + 8|0); //@line 539 "./common/runtime.h"
 $23 = HEAP32[$22>>2]|0; //@line 539 "./common/runtime.h"
 $24 = (FUNCTION_TABLE_ii[$23 & 8191]($20)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($24|0)); //@line 539 "./common/runtime.h"
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if (!($26)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $27 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $27; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $cursor$0 = 0;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,31,$cursor$0); //@line 48 "modules/mouse/wrap_Mouse.cpp"
  $34 = HEAP32[$cursor$0>>2]|0; //@line 49 "modules/mouse/wrap_Mouse.cpp"
  $35 = ((($34)) + 12|0); //@line 49 "modules/mouse/wrap_Mouse.cpp"
  $36 = HEAP32[$35>>2]|0; //@line 49 "modules/mouse/wrap_Mouse.cpp"
  FUNCTION_TABLE_vi[$36 & 8191]($cursor$0); //@line 49 "modules/mouse/wrap_Mouse.cpp"
  STACKTOP = sp;return 1; //@line 50 "modules/mouse/wrap_Mouse.cpp"
 }
 $28 = ___cxa_find_matching_catch_2()|0;
 $29 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $30 = __THREW__; __THREW__ = 0;
 $31 = $30&1;
 if ($31) {
  $32 = ___cxa_find_matching_catch_3(0|0)|0;
  $33 = tempRet0;
  ___clang_call_terminate($32); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $29;$$02$i = $28;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love14luax_checktypeINS_5image9ImageDataEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love5mouse17w_getSystemCursorEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cursor$0 = 0, $systemCursor = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $systemCursor = sp + 12|0;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 55 "modules/mouse/wrap_Mouse.cpp"
 $1 = (__ZN4love5mouse6Cursor11getConstantEPKcRNS1_12SystemCursorE($0,$systemCursor)|0); //@line 58 "modules/mouse/wrap_Mouse.cpp"
 if (!($1)) {
  HEAP32[$vararg_buffer>>2] = $0; //@line 59 "modules/mouse/wrap_Mouse.cpp"
  $2 = (_luaL_error($L,260178,$vararg_buffer)|0); //@line 59 "modules/mouse/wrap_Mouse.cpp"
  $$0 = $2;
  STACKTOP = sp;return ($$0|0); //@line 66 "modules/mouse/wrap_Mouse.cpp"
 }
 $3 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
 $4 = HEAP32[$3>>2]|0; //@line 62 "modules/mouse/wrap_Mouse.cpp"
 $5 = ((($4)) + 28|0); //@line 62 "modules/mouse/wrap_Mouse.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 62 "modules/mouse/wrap_Mouse.cpp"
 $7 = HEAP32[$systemCursor>>2]|0; //@line 62 "modules/mouse/wrap_Mouse.cpp"
 __THREW__ = 0;
 $8 = (invoke_iii($6|0,($3|0),($7|0))|0); //@line 62 "modules/mouse/wrap_Mouse.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 do {
  if ($10) {
   $11 = ___cxa_find_matching_catch_3(8992|0)|0;
   $12 = tempRet0;
   $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $14 = ($12|0)==($13|0); //@line 535 "./common/runtime.h"
   if (!($14)) {
    $$01$i = $12;$$02$i = $11;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $15 = (___cxa_begin_catch(($11|0))|0); //@line 535 "./common/runtime.h"
   $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
   $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
   $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
   $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer1>>2] = $22; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
    $cursor$0 = 0;
    break;
   }
   $23 = ___cxa_find_matching_catch_2()|0;
   $24 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    $27 = ___cxa_find_matching_catch_3(0|0)|0;
    $28 = tempRet0;
    ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $24;$$02$i = $23;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $cursor$0 = $8;
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,31,$cursor$0); //@line 64 "modules/mouse/wrap_Mouse.cpp"
 $$0 = 1;
 STACKTOP = sp;return ($$0|0); //@line 66 "modules/mouse/wrap_Mouse.cpp"
}
function __ZN4love5mouse11w_setCursorEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_type($L,1)|0); //@line 71 "modules/mouse/wrap_Mouse.cpp"
 $1 = ($0|0)<(1); //@line 71 "modules/mouse/wrap_Mouse.cpp"
 if ($1) {
  $2 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
  $3 = HEAP32[$2>>2]|0; //@line 73 "modules/mouse/wrap_Mouse.cpp"
  $4 = ((($3)) + 36|0); //@line 73 "modules/mouse/wrap_Mouse.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 73 "modules/mouse/wrap_Mouse.cpp"
  FUNCTION_TABLE_vi[$5 & 8191]($2); //@line 73 "modules/mouse/wrap_Mouse.cpp"
  return 0; //@line 80 "modules/mouse/wrap_Mouse.cpp"
 } else {
  $6 = (__ZN4love5mouse16luax_checkcursorEP9lua_Statei($L,1)|0); //@line 77 "modules/mouse/wrap_Mouse.cpp"
  $7 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
  $8 = HEAP32[$7>>2]|0; //@line 78 "modules/mouse/wrap_Mouse.cpp"
  $9 = ((($8)) + 32|0); //@line 78 "modules/mouse/wrap_Mouse.cpp"
  $10 = HEAP32[$9>>2]|0; //@line 78 "modules/mouse/wrap_Mouse.cpp"
  FUNCTION_TABLE_vii[$10 & 8191]($7,$6); //@line 78 "modules/mouse/wrap_Mouse.cpp"
  return 0; //@line 80 "modules/mouse/wrap_Mouse.cpp"
 }
 return (0)|0;
}
function __ZN4love5mouse11w_getCursorEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 84 "modules/mouse/wrap_Mouse.cpp"
 $2 = ((($1)) + 40|0); //@line 84 "modules/mouse/wrap_Mouse.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 84 "modules/mouse/wrap_Mouse.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 84 "modules/mouse/wrap_Mouse.cpp"
 $5 = ($4|0)==(0|0); //@line 86 "modules/mouse/wrap_Mouse.cpp"
 if ($5) {
  _lua_pushnil($L); //@line 89 "modules/mouse/wrap_Mouse.cpp"
  return 1; //@line 91 "modules/mouse/wrap_Mouse.cpp"
 } else {
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,31,$4); //@line 87 "modules/mouse/wrap_Mouse.cpp"
  return 1; //@line 91 "modules/mouse/wrap_Mouse.cpp"
 }
 return (0)|0;
}
function __ZN4love5mouse11w_hasCursorEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 96 "modules/mouse/wrap_Mouse.cpp"
 $2 = ((($1)) + 44|0); //@line 96 "modules/mouse/wrap_Mouse.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 96 "modules/mouse/wrap_Mouse.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 96 "modules/mouse/wrap_Mouse.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 96 "modules/mouse/wrap_Mouse.cpp"
 return 1; //@line 97 "modules/mouse/wrap_Mouse.cpp"
}
function __ZN4love5mouse6w_getXEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 102 "modules/mouse/wrap_Mouse.cpp"
 $2 = ((($1)) + 48|0); //@line 102 "modules/mouse/wrap_Mouse.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 102 "modules/mouse/wrap_Mouse.cpp"
 $4 = (+FUNCTION_TABLE_di[$3 & 4095]($0)); //@line 102 "modules/mouse/wrap_Mouse.cpp"
 _lua_pushnumber($L,$4); //@line 102 "modules/mouse/wrap_Mouse.cpp"
 return 1; //@line 103 "modules/mouse/wrap_Mouse.cpp"
}
function __ZN4love5mouse6w_getYEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 108 "modules/mouse/wrap_Mouse.cpp"
 $2 = ((($1)) + 52|0); //@line 108 "modules/mouse/wrap_Mouse.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 108 "modules/mouse/wrap_Mouse.cpp"
 $4 = (+FUNCTION_TABLE_di[$3 & 4095]($0)); //@line 108 "modules/mouse/wrap_Mouse.cpp"
 _lua_pushnumber($L,$4); //@line 108 "modules/mouse/wrap_Mouse.cpp"
 return 1; //@line 109 "modules/mouse/wrap_Mouse.cpp"
}
function __ZN4love5mouse13w_getPositionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $x = 0, $y = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $x = sp + 8|0;
 $y = sp;
 $0 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 115 "modules/mouse/wrap_Mouse.cpp"
 $2 = ((($1)) + 56|0); //@line 115 "modules/mouse/wrap_Mouse.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 115 "modules/mouse/wrap_Mouse.cpp"
 FUNCTION_TABLE_viii[$3 & 8191]($0,$x,$y); //@line 115 "modules/mouse/wrap_Mouse.cpp"
 $4 = +HEAPF64[$x>>3]; //@line 116 "modules/mouse/wrap_Mouse.cpp"
 _lua_pushnumber($L,$4); //@line 116 "modules/mouse/wrap_Mouse.cpp"
 $5 = +HEAPF64[$y>>3]; //@line 117 "modules/mouse/wrap_Mouse.cpp"
 _lua_pushnumber($L,$5); //@line 117 "modules/mouse/wrap_Mouse.cpp"
 STACKTOP = sp;return 2; //@line 118 "modules/mouse/wrap_Mouse.cpp"
}
function __ZN4love5mouse6w_setXEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1)); //@line 123 "modules/mouse/wrap_Mouse.cpp"
 $1 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 124 "modules/mouse/wrap_Mouse.cpp"
 $3 = ((($2)) + 60|0); //@line 124 "modules/mouse/wrap_Mouse.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 124 "modules/mouse/wrap_Mouse.cpp"
 FUNCTION_TABLE_vid[$4 & 8191]($1,$0); //@line 124 "modules/mouse/wrap_Mouse.cpp"
 return 0; //@line 125 "modules/mouse/wrap_Mouse.cpp"
}
function __ZN4love5mouse6w_setYEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1)); //@line 130 "modules/mouse/wrap_Mouse.cpp"
 $1 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 131 "modules/mouse/wrap_Mouse.cpp"
 $3 = ((($2)) + 64|0); //@line 131 "modules/mouse/wrap_Mouse.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 131 "modules/mouse/wrap_Mouse.cpp"
 FUNCTION_TABLE_vid[$4 & 8191]($1,$0); //@line 131 "modules/mouse/wrap_Mouse.cpp"
 return 0; //@line 132 "modules/mouse/wrap_Mouse.cpp"
}
function __ZN4love5mouse13w_setPositionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1)); //@line 137 "modules/mouse/wrap_Mouse.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 138 "modules/mouse/wrap_Mouse.cpp"
 $2 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
 $3 = HEAP32[$2>>2]|0; //@line 139 "modules/mouse/wrap_Mouse.cpp"
 $4 = ((($3)) + 68|0); //@line 139 "modules/mouse/wrap_Mouse.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 139 "modules/mouse/wrap_Mouse.cpp"
 FUNCTION_TABLE_vidd[$5 & 8191]($2,$0,$1); //@line 139 "modules/mouse/wrap_Mouse.cpp"
 return 0; //@line 140 "modules/mouse/wrap_Mouse.cpp"
}
function __ZN4love5mouse8w_isDownEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buttons = 0, $i$07 = 0, $lpad$phi$index = 0;
 var $lpad$phi$index2 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i1 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $buttons = sp + 4|0;
 $0 = sp;
 $1 = (_lua_gettop($L)|0); //@line 145 "modules/mouse/wrap_Mouse.cpp"
 HEAP32[$buttons>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($buttons)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($buttons)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$3>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $4 = ((($buttons)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ($1|0)==(0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 L1: do {
  if ($5) {
   label = 4;
  } else {
   $6 = $1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __THREW__ = 0;
   $7 = (invoke_ii(3348,($6|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $8 = __THREW__; __THREW__ = 0;
   $9 = $8&1;
   if ($9) {
    label = 6;
   } else {
    $10 = $7; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $11 = (($7) + ($1<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $12 = $11; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    HEAP32[$buttons>>2] = $10; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$2>>2] = $10; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$4>>2] = $12; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $13 = ($1|0)>(0); //@line 149 "modules/mouse/wrap_Mouse.cpp"
    if ($13) {
     $i$07 = 0;
     while(1) {
      $34 = (($i$07) + 1)|0; //@line 150 "modules/mouse/wrap_Mouse.cpp"
      __THREW__ = 0;
      $35 = (+invoke_dii(3361,($L|0),($34|0))); //@line 150 "modules/mouse/wrap_Mouse.cpp"
      $36 = __THREW__; __THREW__ = 0;
      $37 = $36&1;
      if ($37) {
       break;
      }
      $38 = (~~(($35))); //@line 150 "modules/mouse/wrap_Mouse.cpp"
      HEAP32[$0>>2] = $38; //@line 150 "modules/mouse/wrap_Mouse.cpp"
      $39 = HEAP32[$2>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $40 = HEAP32[$4>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $41 = ($39>>>0)<($40>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($41) {
       HEAP32[$39>>2] = $38; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $42 = ((($39)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$2>>2] = $42; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      } else {
       __THREW__ = 0;
       invoke_vii(3406,($buttons|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $43 = __THREW__; __THREW__ = 0;
       $44 = $43&1;
       if ($44) {
        break;
       }
      }
      $45 = ($34|0)<($1|0); //@line 149 "modules/mouse/wrap_Mouse.cpp"
      if ($45) {
       $i$07 = $34;
      } else {
       label = 4;
       break L1;
      }
     }
     $21 = ___cxa_find_matching_catch_2()|0;
     $22 = tempRet0;
     $lpad$phi$index = $21;$lpad$phi$index2 = $22;
    } else {
     label = 4;
    }
   }
  }
 } while(0);
 if ((label|0) == 4) {
  $14 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
  $15 = HEAP32[$14>>2]|0; //@line 152 "modules/mouse/wrap_Mouse.cpp"
  $16 = ((($15)) + 76|0); //@line 152 "modules/mouse/wrap_Mouse.cpp"
  $17 = HEAP32[$16>>2]|0; //@line 152 "modules/mouse/wrap_Mouse.cpp"
  __THREW__ = 0;
  $18 = (invoke_iii($17|0,($14|0),($buttons|0))|0); //@line 152 "modules/mouse/wrap_Mouse.cpp"
  $19 = __THREW__; __THREW__ = 0;
  $20 = $19&1;
  if ($20) {
   label = 6;
  } else {
   __THREW__ = 0;
   invoke_vii(3397,($L|0),($18|0)); //@line 152 "modules/mouse/wrap_Mouse.cpp"
   $46 = __THREW__; __THREW__ = 0;
   $47 = $46&1;
   if ($47) {
    label = 6;
   } else {
    $48 = HEAP32[$buttons>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $49 = ($48|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $50 = $48; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($49) {
     STACKTOP = sp;return 1; //@line 154 "modules/mouse/wrap_Mouse.cpp"
    }
    $51 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $52 = ($51|0)==($48|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($52)) {
     $scevgep$i$i$i = ((($51)) + -4|0);
     $53 = $scevgep$i$i$i;
     $54 = (($53) - ($50))|0;
     $55 = $54 >>> 2;
     $56 = $55 ^ -1;
     $scevgep4$i$i$i = (($51) + ($56<<2)|0);
     HEAP32[$2>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($48); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    STACKTOP = sp;return 1; //@line 154 "modules/mouse/wrap_Mouse.cpp"
   }
  }
 }
 if ((label|0) == 6) {
  $23 = ___cxa_find_matching_catch_2()|0;
  $24 = tempRet0;
  $lpad$phi$index = $23;$lpad$phi$index2 = $24;
 }
 $25 = HEAP32[$buttons>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $26 = ($25|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = $25; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($26) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $28 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $29 = ($28|0)==($25|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($29)) {
  $scevgep$i$i$i1 = ((($28)) + -4|0);
  $30 = $scevgep$i$i$i1;
  $31 = (($30) - ($27))|0;
  $32 = $31 >>> 2;
  $33 = $32 ^ -1;
  $scevgep4$i$i$i2 = (($28) + ($33<<2)|0);
  HEAP32[$2>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($25); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__16vectorIiNS_9allocatorIiEEE21__push_back_slow_pathIiEEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love5mouse12w_setVisibleEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,1)|0); //@line 158 "modules/mouse/wrap_Mouse.cpp"
 $1 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 159 "modules/mouse/wrap_Mouse.cpp"
 $3 = ((($2)) + 72|0); //@line 159 "modules/mouse/wrap_Mouse.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 159 "modules/mouse/wrap_Mouse.cpp"
 FUNCTION_TABLE_vii[$4 & 8191]($1,$0); //@line 159 "modules/mouse/wrap_Mouse.cpp"
 return 0; //@line 160 "modules/mouse/wrap_Mouse.cpp"
}
function __ZN4love5mouse11w_isVisibleEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 165 "modules/mouse/wrap_Mouse.cpp"
 $2 = ((($1)) + 80|0); //@line 165 "modules/mouse/wrap_Mouse.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 165 "modules/mouse/wrap_Mouse.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 165 "modules/mouse/wrap_Mouse.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 165 "modules/mouse/wrap_Mouse.cpp"
 return 1; //@line 166 "modules/mouse/wrap_Mouse.cpp"
}
function __ZN4love5mouse12w_setGrabbedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,1)|0); //@line 171 "modules/mouse/wrap_Mouse.cpp"
 $1 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 172 "modules/mouse/wrap_Mouse.cpp"
 $3 = ((($2)) + 84|0); //@line 172 "modules/mouse/wrap_Mouse.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 172 "modules/mouse/wrap_Mouse.cpp"
 FUNCTION_TABLE_vii[$4 & 8191]($1,$0); //@line 172 "modules/mouse/wrap_Mouse.cpp"
 return 0; //@line 173 "modules/mouse/wrap_Mouse.cpp"
}
function __ZN4love5mouse11w_isGrabbedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 178 "modules/mouse/wrap_Mouse.cpp"
 $2 = ((($1)) + 88|0); //@line 178 "modules/mouse/wrap_Mouse.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 178 "modules/mouse/wrap_Mouse.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 178 "modules/mouse/wrap_Mouse.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 178 "modules/mouse/wrap_Mouse.cpp"
 return 1; //@line 179 "modules/mouse/wrap_Mouse.cpp"
}
function __ZN4love5mouse17w_setRelativeModeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,1)|0); //@line 184 "modules/mouse/wrap_Mouse.cpp"
 $1 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 185 "modules/mouse/wrap_Mouse.cpp"
 $3 = ((($2)) + 92|0); //@line 185 "modules/mouse/wrap_Mouse.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 185 "modules/mouse/wrap_Mouse.cpp"
 $5 = (FUNCTION_TABLE_iii[$4 & 8191]($1,$0)|0); //@line 185 "modules/mouse/wrap_Mouse.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$5); //@line 185 "modules/mouse/wrap_Mouse.cpp"
 return 1; //@line 186 "modules/mouse/wrap_Mouse.cpp"
}
function __ZN4love5mouse17w_getRelativeModeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 191 "modules/mouse/wrap_Mouse.cpp"
 $2 = ((($1)) + 96|0); //@line 191 "modules/mouse/wrap_Mouse.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 191 "modules/mouse/wrap_Mouse.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 191 "modules/mouse/wrap_Mouse.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 191 "modules/mouse/wrap_Mouse.cpp"
 return 1; //@line 192 "modules/mouse/wrap_Mouse.cpp"
}
function _luaopen_love_mouse($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0;
 var $eh$lpad$body$i$indexZ2D = 0, $instance$sroa$0$1 = 0, $vararg_buffer = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $w = sp + 4|0;
 $0 = HEAP32[(625644)>>2]|0; //@line 98 "./common/Module.h"
 $1 = $0; //@line 228 "modules/mouse/wrap_Mouse.cpp"
 $2 = ($0|0)==(0|0); //@line 229 "modules/mouse/wrap_Mouse.cpp"
 L1: do {
  if ($2) {
   __THREW__ = 0;
   $3 = (invoke_ii(3348,24)|0); //@line 231 "modules/mouse/wrap_Mouse.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   do {
    if ($5) {
     $11 = ___cxa_find_matching_catch_3(8992|0)|0;
     $12 = tempRet0;
     $eh$lpad$body$i$index2Z2D = $12;$eh$lpad$body$i$indexZ2D = $11;
    } else {
     __THREW__ = 0;
     invoke_vi(3407,($3|0)); //@line 231 "modules/mouse/wrap_Mouse.cpp"
     $6 = __THREW__; __THREW__ = 0;
     $7 = $6&1;
     if ($7) {
      $8 = ___cxa_find_matching_catch_3(8992|0)|0;
      $9 = tempRet0;
      __ZdlPv($3); //@line 231 "modules/mouse/wrap_Mouse.cpp"
      $eh$lpad$body$i$index2Z2D = $9;$eh$lpad$body$i$indexZ2D = $8;
      break;
     } else {
      $10 = $3; //@line 231 "modules/mouse/wrap_Mouse.cpp"
      $instance$sroa$0$1 = $10;
      break L1;
     }
    }
   } while(0);
   $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $14 = ($eh$lpad$body$i$index2Z2D|0)==($13|0); //@line 535 "./common/runtime.h"
   if (!($14)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $15 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
   $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
   $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
   $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $instance$sroa$0$1 = $1;
    break;
   }
   $23 = ___cxa_find_matching_catch_2()|0;
   $24 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    $27 = ___cxa_find_matching_catch_3(0|0)|0;
    $28 = tempRet0;
    ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $24;$$02$i = $23;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $29 = HEAP32[$0>>2]|0; //@line 234 "modules/mouse/wrap_Mouse.cpp"
   $30 = ((($29)) + 8|0); //@line 234 "modules/mouse/wrap_Mouse.cpp"
   $31 = HEAP32[$30>>2]|0; //@line 234 "modules/mouse/wrap_Mouse.cpp"
   FUNCTION_TABLE_vi[$31 & 8191]($0); //@line 234 "modules/mouse/wrap_Mouse.cpp"
   $instance$sroa$0$1 = $1;
  }
 } while(0);
 HEAP32[$w>>2] = $instance$sroa$0$1; //@line 237 "modules/mouse/wrap_Mouse.cpp"
 $32 = ((($w)) + 4|0); //@line 238 "modules/mouse/wrap_Mouse.cpp"
 HEAP32[$32>>2] = 358914; //@line 238 "modules/mouse/wrap_Mouse.cpp"
 $33 = ((($w)) + 8|0); //@line 239 "modules/mouse/wrap_Mouse.cpp"
 HEAP32[$33>>2] = 3; //@line 239 "modules/mouse/wrap_Mouse.cpp"
 $34 = ((($w)) + 12|0); //@line 240 "modules/mouse/wrap_Mouse.cpp"
 HEAP32[$34>>2] = 9948; //@line 240 "modules/mouse/wrap_Mouse.cpp"
 $35 = ((($w)) + 16|0); //@line 241 "modules/mouse/wrap_Mouse.cpp"
 HEAP32[$35>>2] = 10100; //@line 241 "modules/mouse/wrap_Mouse.cpp"
 $36 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 243 "modules/mouse/wrap_Mouse.cpp"
 STACKTOP = sp;return ($36|0); //@line 243 "modules/mouse/wrap_Mouse.cpp"
}
function __ZN4love7physics5box2d10w_newWorldEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0.0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $w$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 54 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = (+_luaL_optnumber($L,1,0.0)); //@line 56 "modules/physics/box2d/wrap_Physics.cpp"
 $1 = $0; //@line 56 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = (+_luaL_optnumber($L,2,0.0)); //@line 57 "modules/physics/box2d/wrap_Physics.cpp"
 $3 = $2; //@line 57 "modules/physics/box2d/wrap_Physics.cpp"
 $4 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,3,1)|0); //@line 58 "modules/physics/box2d/wrap_Physics.cpp"
 $5 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $6 = (invoke_iiddi(3408,($5|0),(+$1),(+$3),($4|0))|0); //@line 61 "modules/physics/box2d/wrap_Physics.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 do {
  if ($8) {
   $9 = ___cxa_find_matching_catch_3(8992|0)|0;
   $10 = tempRet0;
   $11 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $12 = ($10|0)==($11|0); //@line 535 "./common/runtime.h"
   if (!($12)) {
    $$01$i = $10;$$02$i = $9;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $13 = (___cxa_begin_catch(($9|0))|0); //@line 535 "./common/runtime.h"
   $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
   $15 = ((($14)) + 8|0); //@line 539 "./common/runtime.h"
   $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
   $17 = (FUNCTION_TABLE_ii[$16 & 8191]($13)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($17|0)); //@line 539 "./common/runtime.h"
   $18 = __THREW__; __THREW__ = 0;
   $19 = $18&1;
   if (!($19)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $20 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $20; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $w$0 = 0;
    break;
   }
   $21 = ___cxa_find_matching_catch_2()|0;
   $22 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $23 = __THREW__; __THREW__ = 0;
   $24 = $23&1;
   if ($24) {
    $25 = ___cxa_find_matching_catch_3(0|0)|0;
    $26 = tempRet0;
    ___clang_call_terminate($25); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $22;$$02$i = $21;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $w$0 = $6;
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,32,$w$0); //@line 62 "modules/physics/box2d/wrap_Physics.cpp"
 $27 = HEAP32[$w$0>>2]|0; //@line 63 "modules/physics/box2d/wrap_Physics.cpp"
 $28 = ((($27)) + 12|0); //@line 63 "modules/physics/box2d/wrap_Physics.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 63 "modules/physics/box2d/wrap_Physics.cpp"
 FUNCTION_TABLE_vi[$29 & 8191]($w$0); //@line 63 "modules/physics/box2d/wrap_Physics.cpp"
 STACKTOP = sp;return 1; //@line 65 "modules/physics/box2d/wrap_Physics.cpp"
}
function __ZN4love7physics5box2d9w_newBodyEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$02$i = 0, $$pre = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $body$0 = 0, $btype = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $btype = sp + 12|0;
 $0 = (__ZN4love7physics5box2d15luax_checkworldEP9lua_Statei($L,1)|0); //@line 70 "modules/physics/box2d/wrap_Physics.cpp"
 $1 = (+_luaL_optnumber($L,2,0.0)); //@line 71 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = $1; //@line 71 "modules/physics/box2d/wrap_Physics.cpp"
 $3 = (+_luaL_optnumber($L,3,0.0)); //@line 72 "modules/physics/box2d/wrap_Physics.cpp"
 $4 = $3; //@line 72 "modules/physics/box2d/wrap_Physics.cpp"
 HEAP32[$btype>>2] = 1; //@line 74 "modules/physics/box2d/wrap_Physics.cpp"
 $5 = (_lua_type($L,4)|0); //@line 75 "modules/physics/box2d/wrap_Physics.cpp"
 $6 = ($5|0)<(1); //@line 75 "modules/physics/box2d/wrap_Physics.cpp"
 do {
  if ($6) {
   $12 = 1;
  } else {
   $7 = (_lua_tolstring($L,4,0)|0); //@line 75 "modules/physics/box2d/wrap_Physics.cpp"
   $8 = ($7|0)==(0|0); //@line 76 "modules/physics/box2d/wrap_Physics.cpp"
   if ($8) {
    $12 = 1;
   } else {
    $9 = (__ZN4love7physics4Body11getConstantEPKcRNS1_4TypeE($7,$btype)|0); //@line 76 "modules/physics/box2d/wrap_Physics.cpp"
    if ($9) {
     $$pre = HEAP32[$btype>>2]|0; //@line 80 "modules/physics/box2d/wrap_Physics.cpp"
     $12 = $$pre;
     break;
    }
    HEAP32[$vararg_buffer>>2] = $7; //@line 77 "modules/physics/box2d/wrap_Physics.cpp"
    $10 = (_luaL_error($L,260329,$vararg_buffer)|0); //@line 77 "modules/physics/box2d/wrap_Physics.cpp"
    $$0 = $10;
    STACKTOP = sp;return ($$0|0); //@line 84 "modules/physics/box2d/wrap_Physics.cpp"
   }
  }
 } while(0);
 $11 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $13 = (invoke_iiiddi(3409,($11|0),($0|0),(+$2),(+$4),($12|0))|0); //@line 80 "modules/physics/box2d/wrap_Physics.cpp"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 do {
  if ($15) {
   $16 = ___cxa_find_matching_catch_3(8992|0)|0;
   $17 = tempRet0;
   $18 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $19 = ($17|0)==($18|0); //@line 535 "./common/runtime.h"
   if (!($19)) {
    $$01$i = $17;$$02$i = $16;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $20 = (___cxa_begin_catch(($16|0))|0); //@line 535 "./common/runtime.h"
   $21 = HEAP32[$20>>2]|0; //@line 539 "./common/runtime.h"
   $22 = ((($21)) + 8|0); //@line 539 "./common/runtime.h"
   $23 = HEAP32[$22>>2]|0; //@line 539 "./common/runtime.h"
   $24 = (FUNCTION_TABLE_ii[$23 & 8191]($20)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($24|0)); //@line 539 "./common/runtime.h"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if (!($26)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $27 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer1>>2] = $27; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
    $body$0 = 0;
    break;
   }
   $28 = ___cxa_find_matching_catch_2()|0;
   $29 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $30 = __THREW__; __THREW__ = 0;
   $31 = $30&1;
   if ($31) {
    $32 = ___cxa_find_matching_catch_3(0|0)|0;
    $33 = tempRet0;
    ___clang_call_terminate($32); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $29;$$02$i = $28;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $body$0 = $13;
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,34,$body$0); //@line 81 "modules/physics/box2d/wrap_Physics.cpp"
 $34 = HEAP32[$body$0>>2]|0; //@line 82 "modules/physics/box2d/wrap_Physics.cpp"
 $35 = ((($34)) + 12|0); //@line 82 "modules/physics/box2d/wrap_Physics.cpp"
 $36 = HEAP32[$35>>2]|0; //@line 82 "modules/physics/box2d/wrap_Physics.cpp"
 FUNCTION_TABLE_vi[$36 & 8191]($body$0); //@line 82 "modules/physics/box2d/wrap_Physics.cpp"
 $$0 = 1;
 STACKTOP = sp;return ($$0|0); //@line 84 "modules/physics/box2d/wrap_Physics.cpp"
}
function __ZN4love7physics5box2d12w_newFixtureEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fixture$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 86 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,1)|0); //@line 88 "modules/physics/box2d/wrap_Physics.cpp"
 $1 = (__ZN4love7physics5box2d15luax_checkshapeEP9lua_Statei($L,2)|0); //@line 89 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = (+_luaL_optnumber($L,3,1.0)); //@line 90 "modules/physics/box2d/wrap_Physics.cpp"
 $3 = $2; //@line 90 "modules/physics/box2d/wrap_Physics.cpp"
 $4 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $5 = (invoke_iiiid(3410,($4|0),($0|0),($1|0),(+$3))|0); //@line 92 "modules/physics/box2d/wrap_Physics.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 do {
  if ($7) {
   $8 = ___cxa_find_matching_catch_3(8992|0)|0;
   $9 = tempRet0;
   $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $11 = ($9|0)==($10|0); //@line 535 "./common/runtime.h"
   if (!($11)) {
    $$01$i = $9;$$02$i = $8;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $12 = (___cxa_begin_catch(($8|0))|0); //@line 535 "./common/runtime.h"
   $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
   $14 = ((($13)) + 8|0); //@line 539 "./common/runtime.h"
   $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
   $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($16|0)); //@line 539 "./common/runtime.h"
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if (!($18)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $19 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $19; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $fixture$0 = 0;
    break;
   }
   $20 = ___cxa_find_matching_catch_2()|0;
   $21 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $22 = __THREW__; __THREW__ = 0;
   $23 = $22&1;
   if ($23) {
    $24 = ___cxa_find_matching_catch_3(0|0)|0;
    $25 = tempRet0;
    ___clang_call_terminate($24); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $21;$$02$i = $20;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $fixture$0 = $5;
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,35,$fixture$0); //@line 93 "modules/physics/box2d/wrap_Physics.cpp"
 $26 = HEAP32[$fixture$0>>2]|0; //@line 94 "modules/physics/box2d/wrap_Physics.cpp"
 $27 = ((($26)) + 12|0); //@line 94 "modules/physics/box2d/wrap_Physics.cpp"
 $28 = HEAP32[$27>>2]|0; //@line 94 "modules/physics/box2d/wrap_Physics.cpp"
 FUNCTION_TABLE_vi[$28 & 8191]($fixture$0); //@line 94 "modules/physics/box2d/wrap_Physics.cpp"
 STACKTOP = sp;return 1; //@line 95 "modules/physics/box2d/wrap_Physics.cpp"
}
function __ZN4love7physics5box2d16w_newCircleShapeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$01$i2 = 0, $$02$i = 0, $$02$i1 = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $shape$0 = 0, $shape2$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 16|0; //@line 98 "modules/physics/box2d/wrap_Physics.cpp"
 $vararg_buffer1 = sp + 8|0; //@line 98 "modules/physics/box2d/wrap_Physics.cpp"
 $vararg_buffer = sp; //@line 98 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = (_lua_gettop($L)|0); //@line 100 "modules/physics/box2d/wrap_Physics.cpp"
 switch ($0|0) {
 case 1:  {
  $1 = (+_luaL_checknumber($L,1)); //@line 104 "modules/physics/box2d/wrap_Physics.cpp"
  $2 = $1; //@line 104 "modules/physics/box2d/wrap_Physics.cpp"
  $3 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
  __THREW__ = 0;
  $4 = (invoke_iid(3411,($3|0),(+$2))|0); //@line 106 "modules/physics/box2d/wrap_Physics.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  do {
   if ($6) {
    $7 = ___cxa_find_matching_catch_3(8992|0)|0;
    $8 = tempRet0;
    $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
    $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
    if (!($10)) {
     $$01$i = $8;$$02$i = $7;
     ___resumeException($$02$i|0);
     // unreachable;
    }
    $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
    $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
    $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
    $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
    $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
    __THREW__ = 0;
    invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
    $16 = __THREW__; __THREW__ = 0;
    $17 = $16&1;
    if (!($17)) {
     ___cxa_end_catch(); //@line 540 "./common/runtime.h"
     $18 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
     HEAP32[$vararg_buffer>>2] = $18; //@line 543 "./common/runtime.h"
     (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
     $shape$0 = 0;
     break;
    }
    $19 = ___cxa_find_matching_catch_2()|0;
    $20 = tempRet0;
    __THREW__ = 0;
    invoke_v(3378); //@line 540 "./common/runtime.h"
    $21 = __THREW__; __THREW__ = 0;
    $22 = $21&1;
    if ($22) {
     $23 = ___cxa_find_matching_catch_3(0|0)|0;
     $24 = tempRet0;
     ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
     // unreachable; //@line 540 "./common/runtime.h"
    } else {
     $$01$i = $20;$$02$i = $19;
     ___resumeException($$02$i|0);
     // unreachable;
    }
   } else {
    $shape$0 = $4;
   }
  } while(0);
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,37,$shape$0); //@line 107 "modules/physics/box2d/wrap_Physics.cpp"
  $25 = HEAP32[$shape$0>>2]|0; //@line 108 "modules/physics/box2d/wrap_Physics.cpp"
  $26 = ((($25)) + 12|0); //@line 108 "modules/physics/box2d/wrap_Physics.cpp"
  $27 = HEAP32[$26>>2]|0; //@line 108 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$27 & 8191]($shape$0); //@line 108 "modules/physics/box2d/wrap_Physics.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 124 "modules/physics/box2d/wrap_Physics.cpp"
  break;
 }
 case 3:  {
  $28 = (+_luaL_checknumber($L,1)); //@line 113 "modules/physics/box2d/wrap_Physics.cpp"
  $29 = $28; //@line 113 "modules/physics/box2d/wrap_Physics.cpp"
  $30 = (+_luaL_checknumber($L,2)); //@line 114 "modules/physics/box2d/wrap_Physics.cpp"
  $31 = $30; //@line 114 "modules/physics/box2d/wrap_Physics.cpp"
  $32 = (+_luaL_checknumber($L,3)); //@line 115 "modules/physics/box2d/wrap_Physics.cpp"
  $33 = $32; //@line 115 "modules/physics/box2d/wrap_Physics.cpp"
  $34 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
  __THREW__ = 0;
  $35 = (invoke_iiddd(3412,($34|0),(+$29),(+$31),(+$33))|0); //@line 117 "modules/physics/box2d/wrap_Physics.cpp"
  $36 = __THREW__; __THREW__ = 0;
  $37 = $36&1;
  do {
   if ($37) {
    $38 = ___cxa_find_matching_catch_3(8992|0)|0;
    $39 = tempRet0;
    $40 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
    $41 = ($39|0)==($40|0); //@line 535 "./common/runtime.h"
    if (!($41)) {
     $$01$i2 = $39;$$02$i1 = $38;
     ___resumeException($$02$i1|0);
     // unreachable;
    }
    $42 = (___cxa_begin_catch(($38|0))|0); //@line 535 "./common/runtime.h"
    $43 = HEAP32[$42>>2]|0; //@line 539 "./common/runtime.h"
    $44 = ((($43)) + 8|0); //@line 539 "./common/runtime.h"
    $45 = HEAP32[$44>>2]|0; //@line 539 "./common/runtime.h"
    $46 = (FUNCTION_TABLE_ii[$45 & 8191]($42)|0); //@line 539 "./common/runtime.h"
    __THREW__ = 0;
    invoke_vii(3380,($L|0),($46|0)); //@line 539 "./common/runtime.h"
    $47 = __THREW__; __THREW__ = 0;
    $48 = $47&1;
    if (!($48)) {
     ___cxa_end_catch(); //@line 540 "./common/runtime.h"
     $49 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
     HEAP32[$vararg_buffer1>>2] = $49; //@line 543 "./common/runtime.h"
     (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
     $shape2$0 = 0;
     break;
    }
    $50 = ___cxa_find_matching_catch_2()|0;
    $51 = tempRet0;
    __THREW__ = 0;
    invoke_v(3378); //@line 540 "./common/runtime.h"
    $52 = __THREW__; __THREW__ = 0;
    $53 = $52&1;
    if ($53) {
     $54 = ___cxa_find_matching_catch_3(0|0)|0;
     $55 = tempRet0;
     ___clang_call_terminate($54); //@line 540 "./common/runtime.h"
     // unreachable; //@line 540 "./common/runtime.h"
    } else {
     $$01$i2 = $51;$$02$i1 = $50;
     ___resumeException($$02$i1|0);
     // unreachable;
    }
   } else {
    $shape2$0 = $35;
   }
  } while(0);
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,37,$shape2$0); //@line 118 "modules/physics/box2d/wrap_Physics.cpp"
  $56 = HEAP32[$shape2$0>>2]|0; //@line 119 "modules/physics/box2d/wrap_Physics.cpp"
  $57 = ((($56)) + 12|0); //@line 119 "modules/physics/box2d/wrap_Physics.cpp"
  $58 = HEAP32[$57>>2]|0; //@line 119 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$58 & 8191]($shape2$0); //@line 119 "modules/physics/box2d/wrap_Physics.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 124 "modules/physics/box2d/wrap_Physics.cpp"
  break;
 }
 default: {
  $59 = (_luaL_error($L,260351,$vararg_buffer4)|0); //@line 123 "modules/physics/box2d/wrap_Physics.cpp"
  $$0 = $59;
  STACKTOP = sp;return ($$0|0); //@line 124 "modules/physics/box2d/wrap_Physics.cpp"
 }
 }
 return (0)|0;
}
function __ZN4love7physics5box2d19w_newRectangleShapeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$01$i3 = 0, $$02$i = 0, $$02$i2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0.0;
 var $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0.0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $shape$0 = 0, $shape3$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 16|0; //@line 126 "modules/physics/box2d/wrap_Physics.cpp"
 $vararg_buffer1 = sp + 8|0; //@line 126 "modules/physics/box2d/wrap_Physics.cpp"
 $vararg_buffer = sp; //@line 126 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = (_lua_gettop($L)|0); //@line 128 "modules/physics/box2d/wrap_Physics.cpp"
 $1 = ($0|0)==(2); //@line 130 "modules/physics/box2d/wrap_Physics.cpp"
 if ($1) {
  $2 = (+_luaL_checknumber($L,1)); //@line 132 "modules/physics/box2d/wrap_Physics.cpp"
  $3 = $2; //@line 132 "modules/physics/box2d/wrap_Physics.cpp"
  $4 = (+_luaL_checknumber($L,2)); //@line 133 "modules/physics/box2d/wrap_Physics.cpp"
  $5 = $4; //@line 133 "modules/physics/box2d/wrap_Physics.cpp"
  $6 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
  __THREW__ = 0;
  $7 = (invoke_iidd(3413,($6|0),(+$3),(+$5))|0); //@line 135 "modules/physics/box2d/wrap_Physics.cpp"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  do {
   if ($9) {
    $10 = ___cxa_find_matching_catch_3(8992|0)|0;
    $11 = tempRet0;
    $12 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
    $13 = ($11|0)==($12|0); //@line 535 "./common/runtime.h"
    if (!($13)) {
     $$01$i = $11;$$02$i = $10;
     ___resumeException($$02$i|0);
     // unreachable;
    }
    $14 = (___cxa_begin_catch(($10|0))|0); //@line 535 "./common/runtime.h"
    $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
    $16 = ((($15)) + 8|0); //@line 539 "./common/runtime.h"
    $17 = HEAP32[$16>>2]|0; //@line 539 "./common/runtime.h"
    $18 = (FUNCTION_TABLE_ii[$17 & 8191]($14)|0); //@line 539 "./common/runtime.h"
    __THREW__ = 0;
    invoke_vii(3380,($L|0),($18|0)); //@line 539 "./common/runtime.h"
    $19 = __THREW__; __THREW__ = 0;
    $20 = $19&1;
    if (!($20)) {
     ___cxa_end_catch(); //@line 540 "./common/runtime.h"
     $21 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
     HEAP32[$vararg_buffer>>2] = $21; //@line 543 "./common/runtime.h"
     (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
     $shape$0 = 0;
     break;
    }
    $22 = ___cxa_find_matching_catch_2()|0;
    $23 = tempRet0;
    __THREW__ = 0;
    invoke_v(3378); //@line 540 "./common/runtime.h"
    $24 = __THREW__; __THREW__ = 0;
    $25 = $24&1;
    if ($25) {
     $26 = ___cxa_find_matching_catch_3(0|0)|0;
     $27 = tempRet0;
     ___clang_call_terminate($26); //@line 540 "./common/runtime.h"
     // unreachable; //@line 540 "./common/runtime.h"
    } else {
     $$01$i = $23;$$02$i = $22;
     ___resumeException($$02$i|0);
     // unreachable;
    }
   } else {
    $shape$0 = $7;
   }
  } while(0);
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,38,$shape$0); //@line 136 "modules/physics/box2d/wrap_Physics.cpp"
  $28 = HEAP32[$shape$0>>2]|0; //@line 137 "modules/physics/box2d/wrap_Physics.cpp"
  $29 = ((($28)) + 12|0); //@line 137 "modules/physics/box2d/wrap_Physics.cpp"
  $30 = HEAP32[$29>>2]|0; //@line 137 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$30 & 8191]($shape$0); //@line 137 "modules/physics/box2d/wrap_Physics.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 155 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $31 = $0 | 1; //@line 140 "modules/physics/box2d/wrap_Physics.cpp"
 $32 = ($31|0)==(5); //@line 140 "modules/physics/box2d/wrap_Physics.cpp"
 if (!($32)) {
  $68 = (_luaL_error($L,260351,$vararg_buffer4)|0); //@line 154 "modules/physics/box2d/wrap_Physics.cpp"
  $$0 = $68;
  STACKTOP = sp;return ($$0|0); //@line 155 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $33 = (+_luaL_checknumber($L,1)); //@line 142 "modules/physics/box2d/wrap_Physics.cpp"
 $34 = $33; //@line 142 "modules/physics/box2d/wrap_Physics.cpp"
 $35 = (+_luaL_checknumber($L,2)); //@line 143 "modules/physics/box2d/wrap_Physics.cpp"
 $36 = $35; //@line 143 "modules/physics/box2d/wrap_Physics.cpp"
 $37 = (+_luaL_checknumber($L,3)); //@line 144 "modules/physics/box2d/wrap_Physics.cpp"
 $38 = $37; //@line 144 "modules/physics/box2d/wrap_Physics.cpp"
 $39 = (+_luaL_checknumber($L,4)); //@line 145 "modules/physics/box2d/wrap_Physics.cpp"
 $40 = $39; //@line 145 "modules/physics/box2d/wrap_Physics.cpp"
 $41 = (+_luaL_optnumber($L,5,0.0)); //@line 146 "modules/physics/box2d/wrap_Physics.cpp"
 $42 = $41; //@line 146 "modules/physics/box2d/wrap_Physics.cpp"
 $43 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $44 = (invoke_iiddddd(3414,($43|0),(+$34),(+$36),(+$38),(+$40),(+$42))|0); //@line 148 "modules/physics/box2d/wrap_Physics.cpp"
 $45 = __THREW__; __THREW__ = 0;
 $46 = $45&1;
 do {
  if ($46) {
   $47 = ___cxa_find_matching_catch_3(8992|0)|0;
   $48 = tempRet0;
   $49 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $50 = ($48|0)==($49|0); //@line 535 "./common/runtime.h"
   if (!($50)) {
    $$01$i3 = $48;$$02$i2 = $47;
    ___resumeException($$02$i2|0);
    // unreachable;
   }
   $51 = (___cxa_begin_catch(($47|0))|0); //@line 535 "./common/runtime.h"
   $52 = HEAP32[$51>>2]|0; //@line 539 "./common/runtime.h"
   $53 = ((($52)) + 8|0); //@line 539 "./common/runtime.h"
   $54 = HEAP32[$53>>2]|0; //@line 539 "./common/runtime.h"
   $55 = (FUNCTION_TABLE_ii[$54 & 8191]($51)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($55|0)); //@line 539 "./common/runtime.h"
   $56 = __THREW__; __THREW__ = 0;
   $57 = $56&1;
   if (!($57)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $58 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer1>>2] = $58; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
    $shape3$0 = 0;
    break;
   }
   $59 = ___cxa_find_matching_catch_2()|0;
   $60 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $61 = __THREW__; __THREW__ = 0;
   $62 = $61&1;
   if ($62) {
    $63 = ___cxa_find_matching_catch_3(0|0)|0;
    $64 = tempRet0;
    ___clang_call_terminate($63); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i3 = $60;$$02$i2 = $59;
    ___resumeException($$02$i2|0);
    // unreachable;
   }
  } else {
   $shape3$0 = $44;
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,38,$shape3$0); //@line 149 "modules/physics/box2d/wrap_Physics.cpp"
 $65 = HEAP32[$shape3$0>>2]|0; //@line 150 "modules/physics/box2d/wrap_Physics.cpp"
 $66 = ((($65)) + 12|0); //@line 150 "modules/physics/box2d/wrap_Physics.cpp"
 $67 = HEAP32[$66>>2]|0; //@line 150 "modules/physics/box2d/wrap_Physics.cpp"
 FUNCTION_TABLE_vi[$67 & 8191]($shape3$0); //@line 150 "modules/physics/box2d/wrap_Physics.cpp"
 $$0 = 1;
 STACKTOP = sp;return ($$0|0); //@line 155 "modules/physics/box2d/wrap_Physics.cpp"
}
function __ZN4love7physics5box2d14w_newEdgeShapeEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0.0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0, $9 = 0, $shape$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 157 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = (+_luaL_checknumber($L,1)); //@line 159 "modules/physics/box2d/wrap_Physics.cpp"
 $1 = $0; //@line 159 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = (+_luaL_checknumber($L,2)); //@line 160 "modules/physics/box2d/wrap_Physics.cpp"
 $3 = $2; //@line 160 "modules/physics/box2d/wrap_Physics.cpp"
 $4 = (+_luaL_checknumber($L,3)); //@line 161 "modules/physics/box2d/wrap_Physics.cpp"
 $5 = $4; //@line 161 "modules/physics/box2d/wrap_Physics.cpp"
 $6 = (+_luaL_checknumber($L,4)); //@line 162 "modules/physics/box2d/wrap_Physics.cpp"
 $7 = $6; //@line 162 "modules/physics/box2d/wrap_Physics.cpp"
 $8 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $9 = (invoke_iidddd(3415,($8|0),(+$1),(+$3),(+$5),(+$7))|0); //@line 164 "modules/physics/box2d/wrap_Physics.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if (!($11)) {
  $shape$0 = $9;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,39,$shape$0); //@line 165 "modules/physics/box2d/wrap_Physics.cpp"
  $30 = HEAP32[$shape$0>>2]|0; //@line 166 "modules/physics/box2d/wrap_Physics.cpp"
  $31 = ((($30)) + 12|0); //@line 166 "modules/physics/box2d/wrap_Physics.cpp"
  $32 = HEAP32[$31>>2]|0; //@line 166 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$32 & 8191]($shape$0); //@line 166 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 167 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $12 = ___cxa_find_matching_catch_3(8992|0)|0;
 $13 = tempRet0;
 $14 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $15 = ($13|0)==($14|0); //@line 535 "./common/runtime.h"
 if (!($15)) {
  $$01$i = $13;$$02$i = $12;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $16 = (___cxa_begin_catch(($12|0))|0); //@line 535 "./common/runtime.h"
 $17 = HEAP32[$16>>2]|0; //@line 539 "./common/runtime.h"
 $18 = ((($17)) + 8|0); //@line 539 "./common/runtime.h"
 $19 = HEAP32[$18>>2]|0; //@line 539 "./common/runtime.h"
 $20 = (FUNCTION_TABLE_ii[$19 & 8191]($16)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($20|0)); //@line 539 "./common/runtime.h"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if (!($22)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $23 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $23; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $shape$0 = 0;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,39,$shape$0); //@line 165 "modules/physics/box2d/wrap_Physics.cpp"
  $30 = HEAP32[$shape$0>>2]|0; //@line 166 "modules/physics/box2d/wrap_Physics.cpp"
  $31 = ((($30)) + 12|0); //@line 166 "modules/physics/box2d/wrap_Physics.cpp"
  $32 = HEAP32[$31>>2]|0; //@line 166 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$32 & 8191]($shape$0); //@line 166 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 167 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $24 = ___cxa_find_matching_catch_2()|0;
 $25 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  $28 = ___cxa_find_matching_catch_3(0|0)|0;
  $29 = tempRet0;
  ___clang_call_terminate($28); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $25;$$02$i = $24;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d17w_newPolygonShapeEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ret$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 170 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $1 = (invoke_iii(3416,($0|0),($L|0))|0); //@line 173 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  $ret$0 = $1;
  STACKTOP = sp;return ($ret$0|0); //@line 174 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $4 = ___cxa_find_matching_catch_3(8992|0)|0;
 $5 = tempRet0;
 $6 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $7 = ($5|0)==($6|0); //@line 535 "./common/runtime.h"
 if (!($7)) {
  $$01$i = $5;$$02$i = $4;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $8 = (___cxa_begin_catch(($4|0))|0); //@line 535 "./common/runtime.h"
 $9 = HEAP32[$8>>2]|0; //@line 539 "./common/runtime.h"
 $10 = ((($9)) + 8|0); //@line 539 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($8)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($12|0)); //@line 539 "./common/runtime.h"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if (!($14)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $15 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $15; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $ret$0 = 0;
  STACKTOP = sp;return ($ret$0|0); //@line 174 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $16 = ___cxa_find_matching_catch_2()|0;
 $17 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $20 = ___cxa_find_matching_catch_3(0|0)|0;
  $21 = tempRet0;
  ___clang_call_terminate($20); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $17;$$02$i = $16;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d15w_newChainShapeEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ret$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 177 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $1 = (invoke_iii(3417,($0|0),($L|0))|0); //@line 180 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  $ret$0 = $1;
  STACKTOP = sp;return ($ret$0|0); //@line 181 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $4 = ___cxa_find_matching_catch_3(8992|0)|0;
 $5 = tempRet0;
 $6 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $7 = ($5|0)==($6|0); //@line 535 "./common/runtime.h"
 if (!($7)) {
  $$01$i = $5;$$02$i = $4;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $8 = (___cxa_begin_catch(($4|0))|0); //@line 535 "./common/runtime.h"
 $9 = HEAP32[$8>>2]|0; //@line 539 "./common/runtime.h"
 $10 = ((($9)) + 8|0); //@line 539 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($8)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($12|0)); //@line 539 "./common/runtime.h"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if (!($14)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $15 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $15; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $ret$0 = 0;
  STACKTOP = sp;return ($ret$0|0); //@line 181 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $16 = ___cxa_find_matching_catch_2()|0;
 $17 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $20 = ___cxa_find_matching_catch_3(0|0)|0;
  $21 = tempRet0;
  ___clang_call_terminate($20); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $17;$$02$i = $16;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d18w_newDistanceJointEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $j$0 = 0, $vararg_buffer = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 184 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,1)|0); //@line 186 "modules/physics/box2d/wrap_Physics.cpp"
 $1 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,2)|0); //@line 187 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = (+_luaL_checknumber($L,3)); //@line 188 "modules/physics/box2d/wrap_Physics.cpp"
 $3 = $2; //@line 188 "modules/physics/box2d/wrap_Physics.cpp"
 $4 = (+_luaL_checknumber($L,4)); //@line 189 "modules/physics/box2d/wrap_Physics.cpp"
 $5 = $4; //@line 189 "modules/physics/box2d/wrap_Physics.cpp"
 $6 = (+_luaL_checknumber($L,5)); //@line 190 "modules/physics/box2d/wrap_Physics.cpp"
 $7 = $6; //@line 190 "modules/physics/box2d/wrap_Physics.cpp"
 $8 = (+_luaL_checknumber($L,6)); //@line 191 "modules/physics/box2d/wrap_Physics.cpp"
 $9 = $8; //@line 191 "modules/physics/box2d/wrap_Physics.cpp"
 $10 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,7,0)|0); //@line 192 "modules/physics/box2d/wrap_Physics.cpp"
 $11 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $12 = (invoke_iiiiddddi(3418,($11|0),($0|0),($1|0),(+$3),(+$5),(+$7),(+$9),($10|0))|0); //@line 195 "modules/physics/box2d/wrap_Physics.cpp"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if (!($14)) {
  $j$0 = $12;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,43,$j$0); //@line 197 "modules/physics/box2d/wrap_Physics.cpp"
  $33 = HEAP32[$j$0>>2]|0; //@line 198 "modules/physics/box2d/wrap_Physics.cpp"
  $34 = ((($33)) + 12|0); //@line 198 "modules/physics/box2d/wrap_Physics.cpp"
  $35 = HEAP32[$34>>2]|0; //@line 198 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$35 & 8191]($j$0); //@line 198 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 199 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $15 = ___cxa_find_matching_catch_3(8992|0)|0;
 $16 = tempRet0;
 $17 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $18 = ($16|0)==($17|0); //@line 535 "./common/runtime.h"
 if (!($18)) {
  $$01$i = $16;$$02$i = $15;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $19 = (___cxa_begin_catch(($15|0))|0); //@line 535 "./common/runtime.h"
 $20 = HEAP32[$19>>2]|0; //@line 539 "./common/runtime.h"
 $21 = ((($20)) + 8|0); //@line 539 "./common/runtime.h"
 $22 = HEAP32[$21>>2]|0; //@line 539 "./common/runtime.h"
 $23 = (FUNCTION_TABLE_ii[$22 & 8191]($19)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($23|0)); //@line 539 "./common/runtime.h"
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if (!($25)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $26 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $26; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $j$0 = 0;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,43,$j$0); //@line 197 "modules/physics/box2d/wrap_Physics.cpp"
  $33 = HEAP32[$j$0>>2]|0; //@line 198 "modules/physics/box2d/wrap_Physics.cpp"
  $34 = ((($33)) + 12|0); //@line 198 "modules/physics/box2d/wrap_Physics.cpp"
  $35 = HEAP32[$34>>2]|0; //@line 198 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$35 & 8191]($j$0); //@line 198 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 199 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $27 = ___cxa_find_matching_catch_2()|0;
 $28 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $29 = __THREW__; __THREW__ = 0;
 $30 = $29&1;
 if ($30) {
  $31 = ___cxa_find_matching_catch_3(0|0)|0;
  $32 = tempRet0;
  ___clang_call_terminate($31); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $28;$$02$i = $27;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d15w_newMouseJointEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $j$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 202 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,1)|0); //@line 204 "modules/physics/box2d/wrap_Physics.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 205 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = $1; //@line 205 "modules/physics/box2d/wrap_Physics.cpp"
 $3 = (+_luaL_checknumber($L,3)); //@line 206 "modules/physics/box2d/wrap_Physics.cpp"
 $4 = $3; //@line 206 "modules/physics/box2d/wrap_Physics.cpp"
 $5 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $6 = (invoke_iiidd(3419,($5|0),($0|0),(+$2),(+$4))|0); //@line 208 "modules/physics/box2d/wrap_Physics.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 do {
  if ($8) {
   $9 = ___cxa_find_matching_catch_3(8992|0)|0;
   $10 = tempRet0;
   $11 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $12 = ($10|0)==($11|0); //@line 535 "./common/runtime.h"
   if (!($12)) {
    $$01$i = $10;$$02$i = $9;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $13 = (___cxa_begin_catch(($9|0))|0); //@line 535 "./common/runtime.h"
   $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
   $15 = ((($14)) + 8|0); //@line 539 "./common/runtime.h"
   $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
   $17 = (FUNCTION_TABLE_ii[$16 & 8191]($13)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($17|0)); //@line 539 "./common/runtime.h"
   $18 = __THREW__; __THREW__ = 0;
   $19 = $18&1;
   if (!($19)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $20 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $20; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $j$0 = 0;
    break;
   }
   $21 = ___cxa_find_matching_catch_2()|0;
   $22 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $23 = __THREW__; __THREW__ = 0;
   $24 = $23&1;
   if ($24) {
    $25 = ___cxa_find_matching_catch_3(0|0)|0;
    $26 = tempRet0;
    ___clang_call_terminate($25); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $22;$$02$i = $21;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $j$0 = $6;
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,42,$j$0); //@line 209 "modules/physics/box2d/wrap_Physics.cpp"
 $27 = HEAP32[$j$0>>2]|0; //@line 210 "modules/physics/box2d/wrap_Physics.cpp"
 $28 = ((($27)) + 12|0); //@line 210 "modules/physics/box2d/wrap_Physics.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 210 "modules/physics/box2d/wrap_Physics.cpp"
 FUNCTION_TABLE_vi[$29 & 8191]($j$0); //@line 210 "modules/physics/box2d/wrap_Physics.cpp"
 STACKTOP = sp;return 1; //@line 211 "modules/physics/box2d/wrap_Physics.cpp"
}
function __ZN4love7physics5box2d18w_newRevoluteJointEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $j$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 214 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,1)|0); //@line 216 "modules/physics/box2d/wrap_Physics.cpp"
 $1 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,2)|0); //@line 217 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = (+_luaL_checknumber($L,3)); //@line 218 "modules/physics/box2d/wrap_Physics.cpp"
 $3 = $2; //@line 218 "modules/physics/box2d/wrap_Physics.cpp"
 $4 = (+_luaL_checknumber($L,4)); //@line 219 "modules/physics/box2d/wrap_Physics.cpp"
 $5 = $4; //@line 219 "modules/physics/box2d/wrap_Physics.cpp"
 $6 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,5,0)|0); //@line 220 "modules/physics/box2d/wrap_Physics.cpp"
 $7 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $8 = (invoke_iiiiddi(3420,($7|0),($0|0),($1|0),(+$3),(+$5),($6|0))|0); //@line 223 "modules/physics/box2d/wrap_Physics.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if (!($10)) {
  $j$0 = $8;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,45,$j$0); //@line 225 "modules/physics/box2d/wrap_Physics.cpp"
  $29 = HEAP32[$j$0>>2]|0; //@line 226 "modules/physics/box2d/wrap_Physics.cpp"
  $30 = ((($29)) + 12|0); //@line 226 "modules/physics/box2d/wrap_Physics.cpp"
  $31 = HEAP32[$30>>2]|0; //@line 226 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$31 & 8191]($j$0); //@line 226 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 227 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $11 = ___cxa_find_matching_catch_3(8992|0)|0;
 $12 = tempRet0;
 $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $14 = ($12|0)==($13|0); //@line 535 "./common/runtime.h"
 if (!($14)) {
  $$01$i = $12;$$02$i = $11;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $15 = (___cxa_begin_catch(($11|0))|0); //@line 535 "./common/runtime.h"
 $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
 $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
 $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
 $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if (!($21)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $j$0 = 0;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,45,$j$0); //@line 225 "modules/physics/box2d/wrap_Physics.cpp"
  $29 = HEAP32[$j$0>>2]|0; //@line 226 "modules/physics/box2d/wrap_Physics.cpp"
  $30 = ((($29)) + 12|0); //@line 226 "modules/physics/box2d/wrap_Physics.cpp"
  $31 = HEAP32[$30>>2]|0; //@line 226 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$31 & 8191]($j$0); //@line 226 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 227 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $23 = ___cxa_find_matching_catch_2()|0;
 $24 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if ($26) {
  $27 = ___cxa_find_matching_catch_3(0|0)|0;
  $28 = tempRet0;
  ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $24;$$02$i = $23;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d19w_newPrismaticJointEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $5 = 0.0;
 var $6 = 0, $7 = 0, $8 = 0.0, $9 = 0.0, $ax$0 = 0.0, $ay$0 = 0.0, $ay$0$in = 0.0, $j$0 = 0, $storemerge$in = 0, $vararg_buffer = 0, $xB$sroa$0$0$in = 0.0, $yB$sroa$0$0$in = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 230 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,1)|0); //@line 232 "modules/physics/box2d/wrap_Physics.cpp"
 $1 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,2)|0); //@line 233 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = (+_luaL_checknumber($L,3)); //@line 234 "modules/physics/box2d/wrap_Physics.cpp"
 $3 = $2; //@line 234 "modules/physics/box2d/wrap_Physics.cpp"
 $4 = (+_luaL_checknumber($L,4)); //@line 235 "modules/physics/box2d/wrap_Physics.cpp"
 $5 = $4; //@line 235 "modules/physics/box2d/wrap_Physics.cpp"
 $6 = (_lua_gettop($L)|0); //@line 238 "modules/physics/box2d/wrap_Physics.cpp"
 $7 = ($6|0)>(7); //@line 238 "modules/physics/box2d/wrap_Physics.cpp"
 $8 = (+_luaL_checknumber($L,5)); //@line 240 "modules/physics/box2d/wrap_Physics.cpp"
 $9 = $8; //@line 240 "modules/physics/box2d/wrap_Physics.cpp"
 $10 = (+_luaL_checknumber($L,6)); //@line 241 "modules/physics/box2d/wrap_Physics.cpp"
 if ($7) {
  $11 = $10; //@line 241 "modules/physics/box2d/wrap_Physics.cpp"
  $12 = (+_luaL_checknumber($L,7)); //@line 242 "modules/physics/box2d/wrap_Physics.cpp"
  $13 = $12; //@line 242 "modules/physics/box2d/wrap_Physics.cpp"
  $14 = (+_luaL_checknumber($L,8)); //@line 243 "modules/physics/box2d/wrap_Physics.cpp"
  $15 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,9,0)|0); //@line 244 "modules/physics/box2d/wrap_Physics.cpp"
  $ax$0 = $13;$ay$0$in = $14;$storemerge$in = $15;$xB$sroa$0$0$in = $9;$yB$sroa$0$0$in = $11;
 } else {
  $16 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,7,0)|0); //@line 252 "modules/physics/box2d/wrap_Physics.cpp"
  $ax$0 = $9;$ay$0$in = $10;$storemerge$in = $16;$xB$sroa$0$0$in = $3;$yB$sroa$0$0$in = $5;
 }
 $ay$0 = $ay$0$in; //@line 243 "modules/physics/box2d/wrap_Physics.cpp"
 $17 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $18 = (invoke_iiiiddddddi(3421,($17|0),($0|0),($1|0),(+$3),(+$5),(+$xB$sroa$0$0$in),(+$yB$sroa$0$0$in),(+$ax$0),(+$ay$0),($storemerge$in|0))|0); //@line 256 "modules/physics/box2d/wrap_Physics.cpp"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if (!($20)) {
  $j$0 = $18;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,44,$j$0); //@line 258 "modules/physics/box2d/wrap_Physics.cpp"
  $39 = HEAP32[$j$0>>2]|0; //@line 259 "modules/physics/box2d/wrap_Physics.cpp"
  $40 = ((($39)) + 12|0); //@line 259 "modules/physics/box2d/wrap_Physics.cpp"
  $41 = HEAP32[$40>>2]|0; //@line 259 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$41 & 8191]($j$0); //@line 259 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 260 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $21 = ___cxa_find_matching_catch_3(8992|0)|0;
 $22 = tempRet0;
 $23 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $24 = ($22|0)==($23|0); //@line 535 "./common/runtime.h"
 if (!($24)) {
  $$01$i = $22;$$02$i = $21;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $25 = (___cxa_begin_catch(($21|0))|0); //@line 535 "./common/runtime.h"
 $26 = HEAP32[$25>>2]|0; //@line 539 "./common/runtime.h"
 $27 = ((($26)) + 8|0); //@line 539 "./common/runtime.h"
 $28 = HEAP32[$27>>2]|0; //@line 539 "./common/runtime.h"
 $29 = (FUNCTION_TABLE_ii[$28 & 8191]($25)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($29|0)); //@line 539 "./common/runtime.h"
 $30 = __THREW__; __THREW__ = 0;
 $31 = $30&1;
 if (!($31)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $32 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $32; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $j$0 = 0;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,44,$j$0); //@line 258 "modules/physics/box2d/wrap_Physics.cpp"
  $39 = HEAP32[$j$0>>2]|0; //@line 259 "modules/physics/box2d/wrap_Physics.cpp"
  $40 = ((($39)) + 12|0); //@line 259 "modules/physics/box2d/wrap_Physics.cpp"
  $41 = HEAP32[$40>>2]|0; //@line 259 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$41 & 8191]($j$0); //@line 259 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 260 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $33 = ___cxa_find_matching_catch_2()|0;
 $34 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $35 = __THREW__; __THREW__ = 0;
 $36 = $35&1;
 if ($36) {
  $37 = ___cxa_find_matching_catch_3(0|0)|0;
  $38 = tempRet0;
  ___clang_call_terminate($37); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $34;$$02$i = $33;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d16w_newPulleyJointEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $$byval_copy = 0, $$byval_copy1 = 0, $$byval_copy2 = 0, $$byval_copy3 = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0;
 var $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0;
 var $j$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy3 = sp + 64|0;
 $$byval_copy2 = sp + 56|0;
 $$byval_copy1 = sp + 48|0;
 $$byval_copy = sp + 40|0;
 $vararg_buffer = sp;
 $0 = sp + 32|0;
 $1 = sp + 24|0;
 $2 = sp + 16|0;
 $3 = sp + 8|0;
 $4 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,1)|0); //@line 265 "modules/physics/box2d/wrap_Physics.cpp"
 $5 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,2)|0); //@line 266 "modules/physics/box2d/wrap_Physics.cpp"
 $6 = (+_luaL_checknumber($L,3)); //@line 267 "modules/physics/box2d/wrap_Physics.cpp"
 $7 = $6; //@line 267 "modules/physics/box2d/wrap_Physics.cpp"
 $8 = (+_luaL_checknumber($L,4)); //@line 268 "modules/physics/box2d/wrap_Physics.cpp"
 $9 = $8; //@line 268 "modules/physics/box2d/wrap_Physics.cpp"
 $10 = (+_luaL_checknumber($L,5)); //@line 269 "modules/physics/box2d/wrap_Physics.cpp"
 $11 = $10; //@line 269 "modules/physics/box2d/wrap_Physics.cpp"
 $12 = (+_luaL_checknumber($L,6)); //@line 270 "modules/physics/box2d/wrap_Physics.cpp"
 $13 = $12; //@line 270 "modules/physics/box2d/wrap_Physics.cpp"
 $14 = (+_luaL_checknumber($L,7)); //@line 271 "modules/physics/box2d/wrap_Physics.cpp"
 $15 = $14; //@line 271 "modules/physics/box2d/wrap_Physics.cpp"
 $16 = (+_luaL_checknumber($L,8)); //@line 272 "modules/physics/box2d/wrap_Physics.cpp"
 $17 = $16; //@line 272 "modules/physics/box2d/wrap_Physics.cpp"
 $18 = (+_luaL_checknumber($L,9)); //@line 273 "modules/physics/box2d/wrap_Physics.cpp"
 $19 = $18; //@line 273 "modules/physics/box2d/wrap_Physics.cpp"
 $20 = (+_luaL_checknumber($L,10)); //@line 274 "modules/physics/box2d/wrap_Physics.cpp"
 $21 = $20; //@line 274 "modules/physics/box2d/wrap_Physics.cpp"
 $22 = (+_luaL_optnumber($L,11,1.0)); //@line 275 "modules/physics/box2d/wrap_Physics.cpp"
 $23 = $22; //@line 275 "modules/physics/box2d/wrap_Physics.cpp"
 $24 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,12,1)|0); //@line 276 "modules/physics/box2d/wrap_Physics.cpp"
 $25 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 HEAPF32[$0>>2] = $7; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $26 = ((($0)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$26>>2] = $9; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$1>>2] = $11; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$27>>2] = $13; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$2>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $28 = ((($2)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$28>>2] = $17; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$3>>2] = $19; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $29 = ((($3)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$29>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;HEAP32[$$byval_copy+4>>2]=HEAP32[$0+4>>2]|0; //@line 280 "modules/physics/box2d/wrap_Physics.cpp"
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;HEAP32[$$byval_copy1+4>>2]=HEAP32[$1+4>>2]|0; //@line 280 "modules/physics/box2d/wrap_Physics.cpp"
 ;HEAP32[$$byval_copy2>>2]=HEAP32[$2>>2]|0;HEAP32[$$byval_copy2+4>>2]=HEAP32[$2+4>>2]|0; //@line 280 "modules/physics/box2d/wrap_Physics.cpp"
 ;HEAP32[$$byval_copy3>>2]=HEAP32[$3>>2]|0;HEAP32[$$byval_copy3+4>>2]=HEAP32[$3+4>>2]|0; //@line 280 "modules/physics/box2d/wrap_Physics.cpp"
 $30 = (invoke_iiiiiiiidi(3422,($25|0),($4|0),($5|0),($$byval_copy|0),($$byval_copy1|0),($$byval_copy2|0),($$byval_copy3|0),(+$23),($24|0))|0); //@line 280 "modules/physics/box2d/wrap_Physics.cpp"
 $31 = __THREW__; __THREW__ = 0;
 $32 = $31&1;
 if (!($32)) {
  $j$0 = $30;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,46,$j$0); //@line 282 "modules/physics/box2d/wrap_Physics.cpp"
  $51 = HEAP32[$j$0>>2]|0; //@line 283 "modules/physics/box2d/wrap_Physics.cpp"
  $52 = ((($51)) + 12|0); //@line 283 "modules/physics/box2d/wrap_Physics.cpp"
  $53 = HEAP32[$52>>2]|0; //@line 283 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$53 & 8191]($j$0); //@line 283 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 284 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $33 = ___cxa_find_matching_catch_3(8992|0)|0;
 $34 = tempRet0;
 $35 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $36 = ($34|0)==($35|0); //@line 535 "./common/runtime.h"
 if (!($36)) {
  $$01$i = $34;$$02$i = $33;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $37 = (___cxa_begin_catch(($33|0))|0); //@line 535 "./common/runtime.h"
 $38 = HEAP32[$37>>2]|0; //@line 539 "./common/runtime.h"
 $39 = ((($38)) + 8|0); //@line 539 "./common/runtime.h"
 $40 = HEAP32[$39>>2]|0; //@line 539 "./common/runtime.h"
 $41 = (FUNCTION_TABLE_ii[$40 & 8191]($37)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($41|0)); //@line 539 "./common/runtime.h"
 $42 = __THREW__; __THREW__ = 0;
 $43 = $42&1;
 if (!($43)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $44 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $44; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $j$0 = 0;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,46,$j$0); //@line 282 "modules/physics/box2d/wrap_Physics.cpp"
  $51 = HEAP32[$j$0>>2]|0; //@line 283 "modules/physics/box2d/wrap_Physics.cpp"
  $52 = ((($51)) + 12|0); //@line 283 "modules/physics/box2d/wrap_Physics.cpp"
  $53 = HEAP32[$52>>2]|0; //@line 283 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$53 & 8191]($j$0); //@line 283 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 284 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $45 = ___cxa_find_matching_catch_2()|0;
 $46 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $47 = __THREW__; __THREW__ = 0;
 $48 = $47&1;
 if ($48) {
  $49 = ___cxa_find_matching_catch_3(0|0)|0;
  $50 = tempRet0;
  ___clang_call_terminate($49); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $46;$$02$i = $45;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d14w_newGearJointEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $j$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 287 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = (__ZN4love7physics5box2d15luax_checkjointEP9lua_Statei($L,1)|0); //@line 289 "modules/physics/box2d/wrap_Physics.cpp"
 $1 = (__ZN4love7physics5box2d15luax_checkjointEP9lua_Statei($L,2)|0); //@line 290 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = (+_luaL_optnumber($L,3,1.0)); //@line 291 "modules/physics/box2d/wrap_Physics.cpp"
 $3 = $2; //@line 291 "modules/physics/box2d/wrap_Physics.cpp"
 $4 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,4,0)|0); //@line 292 "modules/physics/box2d/wrap_Physics.cpp"
 $5 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $6 = (invoke_iiiidi(3423,($5|0),($0|0),($1|0),(+$3),($4|0))|0); //@line 296 "modules/physics/box2d/wrap_Physics.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 do {
  if ($8) {
   $9 = ___cxa_find_matching_catch_3(8992|0)|0;
   $10 = tempRet0;
   $11 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $12 = ($10|0)==($11|0); //@line 535 "./common/runtime.h"
   if (!($12)) {
    $$01$i = $10;$$02$i = $9;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $13 = (___cxa_begin_catch(($9|0))|0); //@line 535 "./common/runtime.h"
   $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
   $15 = ((($14)) + 8|0); //@line 539 "./common/runtime.h"
   $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
   $17 = (FUNCTION_TABLE_ii[$16 & 8191]($13)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($17|0)); //@line 539 "./common/runtime.h"
   $18 = __THREW__; __THREW__ = 0;
   $19 = $18&1;
   if (!($19)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $20 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $20; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $j$0 = 0;
    break;
   }
   $21 = ___cxa_find_matching_catch_2()|0;
   $22 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $23 = __THREW__; __THREW__ = 0;
   $24 = $23&1;
   if ($24) {
    $25 = ___cxa_find_matching_catch_3(0|0)|0;
    $26 = tempRet0;
    ___clang_call_terminate($25); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $22;$$02$i = $21;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $j$0 = $6;
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,47,$j$0); //@line 298 "modules/physics/box2d/wrap_Physics.cpp"
 $27 = HEAP32[$j$0>>2]|0; //@line 299 "modules/physics/box2d/wrap_Physics.cpp"
 $28 = ((($27)) + 12|0); //@line 299 "modules/physics/box2d/wrap_Physics.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 299 "modules/physics/box2d/wrap_Physics.cpp"
 FUNCTION_TABLE_vi[$29 & 8191]($j$0); //@line 299 "modules/physics/box2d/wrap_Physics.cpp"
 STACKTOP = sp;return 1; //@line 300 "modules/physics/box2d/wrap_Physics.cpp"
}
function __ZN4love7physics5box2d18w_newFrictionJointEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0.0;
 var $9 = 0.0, $j$0 = 0, $storemerge$in = 0, $vararg_buffer = 0, $xB$sroa$0$0$in = 0.0, $yB$sroa$0$0$in = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 303 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,1)|0); //@line 305 "modules/physics/box2d/wrap_Physics.cpp"
 $1 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,2)|0); //@line 306 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = (+_luaL_checknumber($L,3)); //@line 307 "modules/physics/box2d/wrap_Physics.cpp"
 $3 = $2; //@line 307 "modules/physics/box2d/wrap_Physics.cpp"
 $4 = (+_luaL_checknumber($L,4)); //@line 308 "modules/physics/box2d/wrap_Physics.cpp"
 $5 = $4; //@line 308 "modules/physics/box2d/wrap_Physics.cpp"
 $6 = (_lua_gettop($L)|0); //@line 311 "modules/physics/box2d/wrap_Physics.cpp"
 $7 = ($6|0)>(5); //@line 311 "modules/physics/box2d/wrap_Physics.cpp"
 if ($7) {
  $8 = (+_luaL_checknumber($L,5)); //@line 313 "modules/physics/box2d/wrap_Physics.cpp"
  $9 = $8; //@line 313 "modules/physics/box2d/wrap_Physics.cpp"
  $10 = (+_luaL_checknumber($L,6)); //@line 314 "modules/physics/box2d/wrap_Physics.cpp"
  $11 = $10; //@line 314 "modules/physics/box2d/wrap_Physics.cpp"
  $12 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,7,0)|0); //@line 315 "modules/physics/box2d/wrap_Physics.cpp"
  $storemerge$in = $12;$xB$sroa$0$0$in = $9;$yB$sroa$0$0$in = $11;
 } else {
  $13 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,5,0)|0); //@line 321 "modules/physics/box2d/wrap_Physics.cpp"
  $storemerge$in = $13;$xB$sroa$0$0$in = $3;$yB$sroa$0$0$in = $5;
 }
 $14 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $15 = (invoke_iiiiddddi(3424,($14|0),($0|0),($1|0),(+$3),(+$5),(+$xB$sroa$0$0$in),(+$yB$sroa$0$0$in),($storemerge$in|0))|0); //@line 325 "modules/physics/box2d/wrap_Physics.cpp"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if (!($17)) {
  $j$0 = $15;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,48,$j$0); //@line 327 "modules/physics/box2d/wrap_Physics.cpp"
  $36 = HEAP32[$j$0>>2]|0; //@line 328 "modules/physics/box2d/wrap_Physics.cpp"
  $37 = ((($36)) + 12|0); //@line 328 "modules/physics/box2d/wrap_Physics.cpp"
  $38 = HEAP32[$37>>2]|0; //@line 328 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$38 & 8191]($j$0); //@line 328 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 329 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $18 = ___cxa_find_matching_catch_3(8992|0)|0;
 $19 = tempRet0;
 $20 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $21 = ($19|0)==($20|0); //@line 535 "./common/runtime.h"
 if (!($21)) {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $22 = (___cxa_begin_catch(($18|0))|0); //@line 535 "./common/runtime.h"
 $23 = HEAP32[$22>>2]|0; //@line 539 "./common/runtime.h"
 $24 = ((($23)) + 8|0); //@line 539 "./common/runtime.h"
 $25 = HEAP32[$24>>2]|0; //@line 539 "./common/runtime.h"
 $26 = (FUNCTION_TABLE_ii[$25 & 8191]($22)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($26|0)); //@line 539 "./common/runtime.h"
 $27 = __THREW__; __THREW__ = 0;
 $28 = $27&1;
 if (!($28)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $29 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $29; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $j$0 = 0;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,48,$j$0); //@line 327 "modules/physics/box2d/wrap_Physics.cpp"
  $36 = HEAP32[$j$0>>2]|0; //@line 328 "modules/physics/box2d/wrap_Physics.cpp"
  $37 = ((($36)) + 12|0); //@line 328 "modules/physics/box2d/wrap_Physics.cpp"
  $38 = HEAP32[$37>>2]|0; //@line 328 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$38 & 8191]($j$0); //@line 328 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 329 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $30 = ___cxa_find_matching_catch_2()|0;
 $31 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $32 = __THREW__; __THREW__ = 0;
 $33 = $32&1;
 if ($33) {
  $34 = ___cxa_find_matching_catch_3(0|0)|0;
  $35 = tempRet0;
  ___clang_call_terminate($34); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $31;$$02$i = $30;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d14w_newWeldJointEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0.0;
 var $9 = 0.0, $j$0 = 0, $storemerge$in = 0, $vararg_buffer = 0, $xB$sroa$0$0$in = 0.0, $yB$sroa$0$0$in = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 332 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,1)|0); //@line 334 "modules/physics/box2d/wrap_Physics.cpp"
 $1 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,2)|0); //@line 335 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = (+_luaL_checknumber($L,3)); //@line 336 "modules/physics/box2d/wrap_Physics.cpp"
 $3 = $2; //@line 336 "modules/physics/box2d/wrap_Physics.cpp"
 $4 = (+_luaL_checknumber($L,4)); //@line 337 "modules/physics/box2d/wrap_Physics.cpp"
 $5 = $4; //@line 337 "modules/physics/box2d/wrap_Physics.cpp"
 $6 = (_lua_gettop($L)|0); //@line 340 "modules/physics/box2d/wrap_Physics.cpp"
 $7 = ($6|0)>(5); //@line 340 "modules/physics/box2d/wrap_Physics.cpp"
 if ($7) {
  $8 = (+_luaL_checknumber($L,5)); //@line 342 "modules/physics/box2d/wrap_Physics.cpp"
  $9 = $8; //@line 342 "modules/physics/box2d/wrap_Physics.cpp"
  $10 = (+_luaL_checknumber($L,6)); //@line 343 "modules/physics/box2d/wrap_Physics.cpp"
  $11 = $10; //@line 343 "modules/physics/box2d/wrap_Physics.cpp"
  $12 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,7,0)|0); //@line 344 "modules/physics/box2d/wrap_Physics.cpp"
  $storemerge$in = $12;$xB$sroa$0$0$in = $9;$yB$sroa$0$0$in = $11;
 } else {
  $13 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,5,0)|0); //@line 350 "modules/physics/box2d/wrap_Physics.cpp"
  $storemerge$in = $13;$xB$sroa$0$0$in = $3;$yB$sroa$0$0$in = $5;
 }
 $14 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $15 = (invoke_iiiiddddi(3425,($14|0),($0|0),($1|0),(+$3),(+$5),(+$xB$sroa$0$0$in),(+$yB$sroa$0$0$in),($storemerge$in|0))|0); //@line 354 "modules/physics/box2d/wrap_Physics.cpp"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if (!($17)) {
  $j$0 = $15;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,49,$j$0); //@line 356 "modules/physics/box2d/wrap_Physics.cpp"
  $36 = HEAP32[$j$0>>2]|0; //@line 357 "modules/physics/box2d/wrap_Physics.cpp"
  $37 = ((($36)) + 12|0); //@line 357 "modules/physics/box2d/wrap_Physics.cpp"
  $38 = HEAP32[$37>>2]|0; //@line 357 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$38 & 8191]($j$0); //@line 357 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 358 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $18 = ___cxa_find_matching_catch_3(8992|0)|0;
 $19 = tempRet0;
 $20 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $21 = ($19|0)==($20|0); //@line 535 "./common/runtime.h"
 if (!($21)) {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $22 = (___cxa_begin_catch(($18|0))|0); //@line 535 "./common/runtime.h"
 $23 = HEAP32[$22>>2]|0; //@line 539 "./common/runtime.h"
 $24 = ((($23)) + 8|0); //@line 539 "./common/runtime.h"
 $25 = HEAP32[$24>>2]|0; //@line 539 "./common/runtime.h"
 $26 = (FUNCTION_TABLE_ii[$25 & 8191]($22)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($26|0)); //@line 539 "./common/runtime.h"
 $27 = __THREW__; __THREW__ = 0;
 $28 = $27&1;
 if (!($28)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $29 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $29; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $j$0 = 0;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,49,$j$0); //@line 356 "modules/physics/box2d/wrap_Physics.cpp"
  $36 = HEAP32[$j$0>>2]|0; //@line 357 "modules/physics/box2d/wrap_Physics.cpp"
  $37 = ((($36)) + 12|0); //@line 357 "modules/physics/box2d/wrap_Physics.cpp"
  $38 = HEAP32[$37>>2]|0; //@line 357 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$38 & 8191]($j$0); //@line 357 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 358 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $30 = ___cxa_find_matching_catch_2()|0;
 $31 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $32 = __THREW__; __THREW__ = 0;
 $33 = $32&1;
 if ($33) {
  $34 = ___cxa_find_matching_catch_3(0|0)|0;
  $35 = tempRet0;
  ___clang_call_terminate($34); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $31;$$02$i = $30;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d15w_newWheelJointEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $5 = 0.0;
 var $6 = 0, $7 = 0, $8 = 0.0, $9 = 0.0, $ax$0 = 0.0, $ay$0 = 0.0, $ay$0$in = 0.0, $j$0 = 0, $storemerge$in = 0, $vararg_buffer = 0, $xB$sroa$0$0$in = 0.0, $yB$sroa$0$0$in = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 361 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,1)|0); //@line 363 "modules/physics/box2d/wrap_Physics.cpp"
 $1 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,2)|0); //@line 364 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = (+_luaL_checknumber($L,3)); //@line 365 "modules/physics/box2d/wrap_Physics.cpp"
 $3 = $2; //@line 365 "modules/physics/box2d/wrap_Physics.cpp"
 $4 = (+_luaL_checknumber($L,4)); //@line 366 "modules/physics/box2d/wrap_Physics.cpp"
 $5 = $4; //@line 366 "modules/physics/box2d/wrap_Physics.cpp"
 $6 = (_lua_gettop($L)|0); //@line 369 "modules/physics/box2d/wrap_Physics.cpp"
 $7 = ($6|0)>(7); //@line 369 "modules/physics/box2d/wrap_Physics.cpp"
 $8 = (+_luaL_checknumber($L,5)); //@line 371 "modules/physics/box2d/wrap_Physics.cpp"
 $9 = $8; //@line 371 "modules/physics/box2d/wrap_Physics.cpp"
 $10 = (+_luaL_checknumber($L,6)); //@line 372 "modules/physics/box2d/wrap_Physics.cpp"
 if ($7) {
  $11 = $10; //@line 372 "modules/physics/box2d/wrap_Physics.cpp"
  $12 = (+_luaL_checknumber($L,7)); //@line 373 "modules/physics/box2d/wrap_Physics.cpp"
  $13 = $12; //@line 373 "modules/physics/box2d/wrap_Physics.cpp"
  $14 = (+_luaL_checknumber($L,8)); //@line 374 "modules/physics/box2d/wrap_Physics.cpp"
  $15 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,9,0)|0); //@line 375 "modules/physics/box2d/wrap_Physics.cpp"
  $ax$0 = $13;$ay$0$in = $14;$storemerge$in = $15;$xB$sroa$0$0$in = $9;$yB$sroa$0$0$in = $11;
 } else {
  $16 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,7,0)|0); //@line 383 "modules/physics/box2d/wrap_Physics.cpp"
  $ax$0 = $9;$ay$0$in = $10;$storemerge$in = $16;$xB$sroa$0$0$in = $3;$yB$sroa$0$0$in = $5;
 }
 $ay$0 = $ay$0$in; //@line 374 "modules/physics/box2d/wrap_Physics.cpp"
 $17 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $18 = (invoke_iiiiddddddi(3426,($17|0),($0|0),($1|0),(+$3),(+$5),(+$xB$sroa$0$0$in),(+$yB$sroa$0$0$in),(+$ax$0),(+$ay$0),($storemerge$in|0))|0); //@line 388 "modules/physics/box2d/wrap_Physics.cpp"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if (!($20)) {
  $j$0 = $18;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,51,$j$0); //@line 390 "modules/physics/box2d/wrap_Physics.cpp"
  $39 = HEAP32[$j$0>>2]|0; //@line 391 "modules/physics/box2d/wrap_Physics.cpp"
  $40 = ((($39)) + 12|0); //@line 391 "modules/physics/box2d/wrap_Physics.cpp"
  $41 = HEAP32[$40>>2]|0; //@line 391 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$41 & 8191]($j$0); //@line 391 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 392 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $21 = ___cxa_find_matching_catch_3(8992|0)|0;
 $22 = tempRet0;
 $23 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $24 = ($22|0)==($23|0); //@line 535 "./common/runtime.h"
 if (!($24)) {
  $$01$i = $22;$$02$i = $21;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $25 = (___cxa_begin_catch(($21|0))|0); //@line 535 "./common/runtime.h"
 $26 = HEAP32[$25>>2]|0; //@line 539 "./common/runtime.h"
 $27 = ((($26)) + 8|0); //@line 539 "./common/runtime.h"
 $28 = HEAP32[$27>>2]|0; //@line 539 "./common/runtime.h"
 $29 = (FUNCTION_TABLE_ii[$28 & 8191]($25)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($29|0)); //@line 539 "./common/runtime.h"
 $30 = __THREW__; __THREW__ = 0;
 $31 = $30&1;
 if (!($31)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $32 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $32; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $j$0 = 0;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,51,$j$0); //@line 390 "modules/physics/box2d/wrap_Physics.cpp"
  $39 = HEAP32[$j$0>>2]|0; //@line 391 "modules/physics/box2d/wrap_Physics.cpp"
  $40 = ((($39)) + 12|0); //@line 391 "modules/physics/box2d/wrap_Physics.cpp"
  $41 = HEAP32[$40>>2]|0; //@line 391 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$41 & 8191]($j$0); //@line 391 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 392 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $33 = ___cxa_find_matching_catch_2()|0;
 $34 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $35 = __THREW__; __THREW__ = 0;
 $36 = $35&1;
 if ($36) {
  $37 = ___cxa_find_matching_catch_3(0|0)|0;
  $38 = tempRet0;
  ___clang_call_terminate($37); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $34;$$02$i = $33;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d14w_newRopeJointEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0;
 var $j$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 395 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,1)|0); //@line 397 "modules/physics/box2d/wrap_Physics.cpp"
 $1 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,2)|0); //@line 398 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = (+_luaL_checknumber($L,3)); //@line 399 "modules/physics/box2d/wrap_Physics.cpp"
 $3 = $2; //@line 399 "modules/physics/box2d/wrap_Physics.cpp"
 $4 = (+_luaL_checknumber($L,4)); //@line 400 "modules/physics/box2d/wrap_Physics.cpp"
 $5 = $4; //@line 400 "modules/physics/box2d/wrap_Physics.cpp"
 $6 = (+_luaL_checknumber($L,5)); //@line 401 "modules/physics/box2d/wrap_Physics.cpp"
 $7 = $6; //@line 401 "modules/physics/box2d/wrap_Physics.cpp"
 $8 = (+_luaL_checknumber($L,6)); //@line 402 "modules/physics/box2d/wrap_Physics.cpp"
 $9 = $8; //@line 402 "modules/physics/box2d/wrap_Physics.cpp"
 $10 = (+_luaL_checknumber($L,7)); //@line 403 "modules/physics/box2d/wrap_Physics.cpp"
 $11 = $10; //@line 403 "modules/physics/box2d/wrap_Physics.cpp"
 $12 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,8,0)|0); //@line 404 "modules/physics/box2d/wrap_Physics.cpp"
 $13 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $14 = (invoke_iiiidddddi(3427,($13|0),($0|0),($1|0),(+$3),(+$5),(+$7),(+$9),(+$11),($12|0))|0); //@line 407 "modules/physics/box2d/wrap_Physics.cpp"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  $j$0 = $14;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,50,$j$0); //@line 409 "modules/physics/box2d/wrap_Physics.cpp"
  $35 = HEAP32[$j$0>>2]|0; //@line 410 "modules/physics/box2d/wrap_Physics.cpp"
  $36 = ((($35)) + 12|0); //@line 410 "modules/physics/box2d/wrap_Physics.cpp"
  $37 = HEAP32[$36>>2]|0; //@line 410 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$37 & 8191]($j$0); //@line 410 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 411 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $17 = ___cxa_find_matching_catch_3(8992|0)|0;
 $18 = tempRet0;
 $19 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $20 = ($18|0)==($19|0); //@line 535 "./common/runtime.h"
 if (!($20)) {
  $$01$i = $18;$$02$i = $17;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $21 = (___cxa_begin_catch(($17|0))|0); //@line 535 "./common/runtime.h"
 $22 = HEAP32[$21>>2]|0; //@line 539 "./common/runtime.h"
 $23 = ((($22)) + 8|0); //@line 539 "./common/runtime.h"
 $24 = HEAP32[$23>>2]|0; //@line 539 "./common/runtime.h"
 $25 = (FUNCTION_TABLE_ii[$24 & 8191]($21)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($25|0)); //@line 539 "./common/runtime.h"
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if (!($27)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $28 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $28; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $j$0 = 0;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,50,$j$0); //@line 409 "modules/physics/box2d/wrap_Physics.cpp"
  $35 = HEAP32[$j$0>>2]|0; //@line 410 "modules/physics/box2d/wrap_Physics.cpp"
  $36 = ((($35)) + 12|0); //@line 410 "modules/physics/box2d/wrap_Physics.cpp"
  $37 = HEAP32[$36>>2]|0; //@line 410 "modules/physics/box2d/wrap_Physics.cpp"
  FUNCTION_TABLE_vi[$37 & 8191]($j$0); //@line 410 "modules/physics/box2d/wrap_Physics.cpp"
  STACKTOP = sp;return 1; //@line 411 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $29 = ___cxa_find_matching_catch_2()|0;
 $30 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $31 = __THREW__; __THREW__ = 0;
 $32 = $31&1;
 if ($32) {
  $33 = ___cxa_find_matching_catch_3(0|0)|0;
  $34 = tempRet0;
  ___clang_call_terminate($33); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $30;$$02$i = $29;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d15w_newMotorJointEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$01$i2 = 0, $$02$i = 0, $$02$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0.0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $j$2 = 0, $vararg_buffer = 0;
 var $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 414 "modules/physics/box2d/wrap_Physics.cpp"
 $vararg_buffer = sp; //@line 414 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,1)|0); //@line 416 "modules/physics/box2d/wrap_Physics.cpp"
 $1 = (__ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,2)|0); //@line 417 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = (_lua_type($L,3)|0); //@line 419 "modules/physics/box2d/wrap_Physics.cpp"
 $3 = ($2|0)<(1); //@line 419 "modules/physics/box2d/wrap_Physics.cpp"
 do {
  if ($3) {
   $29 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
   __THREW__ = 0;
   $30 = (invoke_iiii(3429,($29|0),($0|0),($1|0))|0); //@line 429 "modules/physics/box2d/wrap_Physics.cpp"
   $31 = __THREW__; __THREW__ = 0;
   $32 = $31&1;
   if ($32) {
    $33 = ___cxa_find_matching_catch_3(8992|0)|0;
    $34 = tempRet0;
    $35 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
    $36 = ($34|0)==($35|0); //@line 535 "./common/runtime.h"
    if (!($36)) {
     $$01$i2 = $34;$$02$i1 = $33;
     ___resumeException($$02$i1|0);
     // unreachable;
    }
    $37 = (___cxa_begin_catch(($33|0))|0); //@line 535 "./common/runtime.h"
    $38 = HEAP32[$37>>2]|0; //@line 539 "./common/runtime.h"
    $39 = ((($38)) + 8|0); //@line 539 "./common/runtime.h"
    $40 = HEAP32[$39>>2]|0; //@line 539 "./common/runtime.h"
    $41 = (FUNCTION_TABLE_ii[$40 & 8191]($37)|0); //@line 539 "./common/runtime.h"
    __THREW__ = 0;
    invoke_vii(3380,($L|0),($41|0)); //@line 539 "./common/runtime.h"
    $42 = __THREW__; __THREW__ = 0;
    $43 = $42&1;
    if (!($43)) {
     ___cxa_end_catch(); //@line 540 "./common/runtime.h"
     $44 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
     HEAP32[$vararg_buffer1>>2] = $44; //@line 543 "./common/runtime.h"
     (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
     $j$2 = 0;
     break;
    }
    $45 = ___cxa_find_matching_catch_2()|0;
    $46 = tempRet0;
    __THREW__ = 0;
    invoke_v(3378); //@line 540 "./common/runtime.h"
    $47 = __THREW__; __THREW__ = 0;
    $48 = $47&1;
    if ($48) {
     $49 = ___cxa_find_matching_catch_3(0|0)|0;
     $50 = tempRet0;
     ___clang_call_terminate($49); //@line 540 "./common/runtime.h"
     // unreachable; //@line 540 "./common/runtime.h"
    } else {
     $$01$i2 = $46;$$02$i1 = $45;
     ___resumeException($$02$i1|0);
     // unreachable;
    }
   } else {
    $j$2 = $30;
   }
  } else {
   $4 = (+_luaL_checknumber($L,3)); //@line 421 "modules/physics/box2d/wrap_Physics.cpp"
   $5 = $4; //@line 421 "modules/physics/box2d/wrap_Physics.cpp"
   $6 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,4,0)|0); //@line 422 "modules/physics/box2d/wrap_Physics.cpp"
   $7 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
   __THREW__ = 0;
   $8 = (invoke_iiiidi(3428,($7|0),($0|0),($1|0),(+$5),($6|0))|0); //@line 424 "modules/physics/box2d/wrap_Physics.cpp"
   $9 = __THREW__; __THREW__ = 0;
   $10 = $9&1;
   if ($10) {
    $11 = ___cxa_find_matching_catch_3(8992|0)|0;
    $12 = tempRet0;
    $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
    $14 = ($12|0)==($13|0); //@line 535 "./common/runtime.h"
    if (!($14)) {
     $$01$i = $12;$$02$i = $11;
     ___resumeException($$02$i|0);
     // unreachable;
    }
    $15 = (___cxa_begin_catch(($11|0))|0); //@line 535 "./common/runtime.h"
    $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
    $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
    $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
    $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
    __THREW__ = 0;
    invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
    $20 = __THREW__; __THREW__ = 0;
    $21 = $20&1;
    if (!($21)) {
     ___cxa_end_catch(); //@line 540 "./common/runtime.h"
     $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
     HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
     (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
     $j$2 = 0;
     break;
    }
    $23 = ___cxa_find_matching_catch_2()|0;
    $24 = tempRet0;
    __THREW__ = 0;
    invoke_v(3378); //@line 540 "./common/runtime.h"
    $25 = __THREW__; __THREW__ = 0;
    $26 = $25&1;
    if ($26) {
     $27 = ___cxa_find_matching_catch_3(0|0)|0;
     $28 = tempRet0;
     ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
     // unreachable; //@line 540 "./common/runtime.h"
    } else {
     $$01$i = $24;$$02$i = $23;
     ___resumeException($$02$i|0);
     // unreachable;
    }
   } else {
    $j$2 = $8;
   }
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,52,$j$2); //@line 431 "modules/physics/box2d/wrap_Physics.cpp"
 $51 = HEAP32[$j$2>>2]|0; //@line 432 "modules/physics/box2d/wrap_Physics.cpp"
 $52 = ((($51)) + 12|0); //@line 432 "modules/physics/box2d/wrap_Physics.cpp"
 $53 = HEAP32[$52>>2]|0; //@line 432 "modules/physics/box2d/wrap_Physics.cpp"
 FUNCTION_TABLE_vi[$53 & 8191]($j$2); //@line 432 "modules/physics/box2d/wrap_Physics.cpp"
 STACKTOP = sp;return 1; //@line 433 "modules/physics/box2d/wrap_Physics.cpp"
}
function __ZN4love7physics5box2d13w_getDistanceEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZN4love7physics5box2d7Physics11getDistanceEP9lua_State($0,$L)|0); //@line 438 "modules/physics/box2d/wrap_Physics.cpp"
 return ($1|0); //@line 438 "modules/physics/box2d/wrap_Physics.cpp"
}
function __ZN4love7physics5box2d10w_setMeterEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0.0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 441 "modules/physics/box2d/wrap_Physics.cpp"
 $0 = (+_luaL_checknumber($L,1)); //@line 443 "modules/physics/box2d/wrap_Physics.cpp"
 $1 = (~~(($0))); //@line 443 "modules/physics/box2d/wrap_Physics.cpp"
 __THREW__ = 0;
 invoke_vi(3430,($1|0)); //@line 444 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  STACKTOP = sp;return 0; //@line 445 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $4 = ___cxa_find_matching_catch_3(8992|0)|0;
 $5 = tempRet0;
 $6 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $7 = ($5|0)==($6|0); //@line 535 "./common/runtime.h"
 if (!($7)) {
  $$01$i = $5;$$02$i = $4;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $8 = (___cxa_begin_catch(($4|0))|0); //@line 535 "./common/runtime.h"
 $9 = HEAP32[$8>>2]|0; //@line 539 "./common/runtime.h"
 $10 = ((($9)) + 8|0); //@line 539 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($8)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($12|0)); //@line 539 "./common/runtime.h"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if (!($14)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $15 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $15; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 445 "modules/physics/box2d/wrap_Physics.cpp"
 }
 $16 = ___cxa_find_matching_catch_2()|0;
 $17 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $20 = ___cxa_find_matching_catch_3(0|0)|0;
  $21 = tempRet0;
  ___clang_call_terminate($20); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $17;$$02$i = $16;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d10w_getMeterEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love7physics5box2d7Physics8getMeterEv()|0); //@line 450 "modules/physics/box2d/wrap_Physics.cpp"
 _lua_pushinteger($L,$0); //@line 450 "modules/physics/box2d/wrap_Physics.cpp"
 return 1; //@line 451 "modules/physics/box2d/wrap_Physics.cpp"
}
function _luaopen_love_physics($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $instance$sroa$0$1 = 0, $vararg_buffer = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $w = sp + 4|0;
 $0 = HEAP32[(625648)>>2]|0; //@line 98 "./common/Module.h"
 $1 = $0; //@line 510 "modules/physics/box2d/wrap_Physics.cpp"
 $2 = ($0|0)==(0|0); //@line 511 "modules/physics/box2d/wrap_Physics.cpp"
 L1: do {
  if ($2) {
   __THREW__ = 0;
   $3 = (invoke_ii(3348,8)|0); //@line 513 "modules/physics/box2d/wrap_Physics.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   do {
    if ($5) {
     $15 = ___cxa_find_matching_catch_3(8992|0)|0;
     $16 = tempRet0;
     $eh$lpad$body$i$index2Z2D = $16;$eh$lpad$body$i$indexZ2D = $15;
    } else {
     $6 = $3; //@line 513 "modules/physics/box2d/wrap_Physics.cpp"
     $7 = $6; //@line 513 "modules/physics/box2d/wrap_Physics.cpp"
     HEAP32[$7>>2] = 0; //@line 513 "modules/physics/box2d/wrap_Physics.cpp"
     $8 = (($6) + 4)|0; //@line 513 "modules/physics/box2d/wrap_Physics.cpp"
     $9 = $8; //@line 513 "modules/physics/box2d/wrap_Physics.cpp"
     HEAP32[$9>>2] = 0; //@line 513 "modules/physics/box2d/wrap_Physics.cpp"
     __THREW__ = 0;
     invoke_vi(3431,($3|0)); //@line 33 "./common/Module.h"
     $10 = __THREW__; __THREW__ = 0;
     $11 = $10&1;
     if ($11) {
      $12 = ___cxa_find_matching_catch_3(8992|0)|0;
      $13 = tempRet0;
      __ZdlPv($3); //@line 513 "modules/physics/box2d/wrap_Physics.cpp"
      $eh$lpad$body$i$index2Z2D = $13;$eh$lpad$body$i$indexZ2D = $12;
      break;
     } else {
      HEAP32[$3>>2] = (17104); //@line 55 "modules/physics/box2d/Physics.h"
      $14 = $3; //@line 513 "modules/physics/box2d/wrap_Physics.cpp"
      $instance$sroa$0$1 = $14;
      break L1;
     }
    }
   } while(0);
   $17 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $18 = ($eh$lpad$body$i$index2Z2D|0)==($17|0); //@line 535 "./common/runtime.h"
   if (!($18)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $19 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $20 = HEAP32[$19>>2]|0; //@line 539 "./common/runtime.h"
   $21 = ((($20)) + 8|0); //@line 539 "./common/runtime.h"
   $22 = HEAP32[$21>>2]|0; //@line 539 "./common/runtime.h"
   $23 = (FUNCTION_TABLE_ii[$22 & 8191]($19)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($23|0)); //@line 539 "./common/runtime.h"
   $24 = __THREW__; __THREW__ = 0;
   $25 = $24&1;
   if (!($25)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $26 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $26; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $instance$sroa$0$1 = $1;
    break;
   }
   $27 = ___cxa_find_matching_catch_2()|0;
   $28 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $29 = __THREW__; __THREW__ = 0;
   $30 = $29&1;
   if ($30) {
    $31 = ___cxa_find_matching_catch_3(0|0)|0;
    $32 = tempRet0;
    ___clang_call_terminate($31); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $28;$$02$i = $27;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $33 = HEAP32[$0>>2]|0; //@line 516 "modules/physics/box2d/wrap_Physics.cpp"
   $34 = ((($33)) + 8|0); //@line 516 "modules/physics/box2d/wrap_Physics.cpp"
   $35 = HEAP32[$34>>2]|0; //@line 516 "modules/physics/box2d/wrap_Physics.cpp"
   FUNCTION_TABLE_vi[$35 & 8191]($0); //@line 516 "modules/physics/box2d/wrap_Physics.cpp"
   $instance$sroa$0$1 = $1;
  }
 } while(0);
 HEAP32[$w>>2] = $instance$sroa$0$1; //@line 519 "modules/physics/box2d/wrap_Physics.cpp"
 $36 = ((($w)) + 4|0); //@line 520 "modules/physics/box2d/wrap_Physics.cpp"
 HEAP32[$36>>2] = 260382; //@line 520 "modules/physics/box2d/wrap_Physics.cpp"
 $37 = ((($w)) + 8|0); //@line 521 "modules/physics/box2d/wrap_Physics.cpp"
 HEAP32[$37>>2] = 3; //@line 521 "modules/physics/box2d/wrap_Physics.cpp"
 $38 = ((($w)) + 12|0); //@line 522 "modules/physics/box2d/wrap_Physics.cpp"
 HEAP32[$38>>2] = 10108; //@line 522 "modules/physics/box2d/wrap_Physics.cpp"
 $39 = ((($w)) + 16|0); //@line 523 "modules/physics/box2d/wrap_Physics.cpp"
 HEAP32[$39>>2] = 10292; //@line 523 "modules/physics/box2d/wrap_Physics.cpp"
 $40 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 525 "modules/physics/box2d/wrap_Physics.cpp"
 STACKTOP = sp;return ($40|0); //@line 525 "modules/physics/box2d/wrap_Physics.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d12PolygonShapeEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d24w_PolygonShape_getPointsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d12PolygonShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,38)|0); //@line 32 "modules/physics/box2d/wrap_PolygonShape.cpp"
 _lua_remove($L,1); //@line 38 "modules/physics/box2d/wrap_PolygonShape.cpp"
 $1 = (__ZN4love7physics5box2d12PolygonShape9getPointsEP9lua_State($0,$L)|0); //@line 39 "modules/physics/box2d/wrap_PolygonShape.cpp"
 return ($1|0); //@line 39 "modules/physics/box2d/wrap_PolygonShape.cpp"
}
function __ZN4love7physics5box2d23w_PolygonShape_validateEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d12PolygonShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,38)|0); //@line 32 "modules/physics/box2d/wrap_PolygonShape.cpp"
 $1 = (__ZNK4love7physics5box2d12PolygonShape8validateEv($0)|0); //@line 45 "modules/physics/box2d/wrap_PolygonShape.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$1); //@line 45 "modules/physics/box2d/wrap_PolygonShape.cpp"
 return 1; //@line 46 "modules/physics/box2d/wrap_PolygonShape.cpp"
}
function _luaopen_polygonshape($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 56 "modules/physics/box2d/wrap_PolygonShape.cpp"
 HEAP32[$vararg_buffer>>2] = 10748; //@line 58 "modules/physics/box2d/wrap_PolygonShape.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 58 "modules/physics/box2d/wrap_PolygonShape.cpp"
 HEAP32[$vararg_ptr1>>2] = 10380; //@line 58 "modules/physics/box2d/wrap_PolygonShape.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 58 "modules/physics/box2d/wrap_PolygonShape.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 58 "modules/physics/box2d/wrap_PolygonShape.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,38,260689,$vararg_buffer)|0); //@line 58 "modules/physics/box2d/wrap_PolygonShape.cpp"
 STACKTOP = sp;return ($0|0); //@line 58 "modules/physics/box2d/wrap_PolygonShape.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d36w_PrismaticJoint_getJointTranslationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 39 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,1,44)|0); //@line 33 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d14PrismaticJoint19getJointTranslationEv($0)); //@line 42 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $3 = $2; //@line 42 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 _lua_pushnumber($L,$3); //@line 42 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 STACKTOP = sp;return 1; //@line 43 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d30w_PrismaticJoint_getJointSpeedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 46 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,1,44)|0); //@line 33 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d14PrismaticJoint13getJointSpeedEv($0)); //@line 49 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $3 = $2; //@line 49 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 _lua_pushnumber($L,$3); //@line 49 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 STACKTOP = sp;return 1; //@line 50 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d32w_PrismaticJoint_setMotorEnabledEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 53 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,1,44)|0); //@line 33 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $2 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,2)|0); //@line 56 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 __ZN4love7physics5box2d14PrismaticJoint15setMotorEnabledEb($0,$2); //@line 57 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 STACKTOP = sp;return 0; //@line 58 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d31w_PrismaticJoint_isMotorEnabledEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 61 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,1,44)|0); //@line 33 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $2 = (__ZNK4love7physics5box2d14PrismaticJoint14isMotorEnabledEv($0)|0); //@line 64 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$2); //@line 64 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 STACKTOP = sp;return 1; //@line 65 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d33w_PrismaticJoint_setMaxMotorForceEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 68 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,1,44)|0); //@line 33 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 71 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $3 = $2; //@line 71 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 __ZN4love7physics5box2d14PrismaticJoint16setMaxMotorForceEf($0,$3); //@line 72 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 STACKTOP = sp;return 0; //@line 73 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d30w_PrismaticJoint_setMotorSpeedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 76 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,1,44)|0); //@line 33 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 79 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $3 = $2; //@line 79 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 __ZN4love7physics5box2d14PrismaticJoint13setMotorSpeedEf($0,$3); //@line 80 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 STACKTOP = sp;return 0; //@line 81 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d30w_PrismaticJoint_getMotorSpeedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 84 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,1,44)|0); //@line 33 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d14PrismaticJoint13getMotorSpeedEv($0)); //@line 87 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $3 = $2; //@line 87 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 _lua_pushnumber($L,$3); //@line 87 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 STACKTOP = sp;return 1; //@line 88 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d30w_PrismaticJoint_getMotorForceEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 91 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,1,44)|0); //@line 33 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 94 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $3 = $2; //@line 94 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $4 = (+__ZNK4love7physics5box2d14PrismaticJoint13getMotorForceEf($0,$3)); //@line 95 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $5 = $4; //@line 95 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 _lua_pushnumber($L,$5); //@line 95 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 STACKTOP = sp;return 1; //@line 96 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d33w_PrismaticJoint_getMaxMotorForceEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 99 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,1,44)|0); //@line 33 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d14PrismaticJoint16getMaxMotorForceEv($0)); //@line 102 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $3 = $2; //@line 102 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 _lua_pushnumber($L,$3); //@line 102 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 STACKTOP = sp;return 1; //@line 103 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d33w_PrismaticJoint_setLimitsEnabledEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 106 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,1,44)|0); //@line 33 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $2 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,2)|0); //@line 109 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 __ZN4love7physics5box2d14PrismaticJoint16setLimitsEnabledEb($0,$2); //@line 110 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 STACKTOP = sp;return 0; //@line 111 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d33w_PrismaticJoint_hasLimitsEnabledEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 114 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,1,44)|0); //@line 33 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $2 = (__ZNK4love7physics5box2d14PrismaticJoint16hasLimitsEnabledEv($0)|0); //@line 117 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$2); //@line 117 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 STACKTOP = sp;return 1; //@line 118 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d30w_PrismaticJoint_setUpperLimitEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0.0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 121 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $vararg_buffer = sp; //@line 121 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,1,44)|0); //@line 33 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 124 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $3 = $2; //@line 124 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 __THREW__ = 0;
 invoke_vid(3432,($0|0),(+$3)); //@line 125 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  STACKTOP = sp;return 0; //@line 126 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $6 = ___cxa_find_matching_catch_3(8992|0)|0;
 $7 = tempRet0;
 $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
 if (!($9)) {
  $$01$i = $7;$$02$i = $6;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $17; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 126 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d30w_PrismaticJoint_setLowerLimitEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0.0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 129 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $vararg_buffer = sp; //@line 129 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,1,44)|0); //@line 33 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 132 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $3 = $2; //@line 132 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 __THREW__ = 0;
 invoke_vid(3433,($0|0),(+$3)); //@line 133 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  STACKTOP = sp;return 0; //@line 134 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $6 = ___cxa_find_matching_catch_3(8992|0)|0;
 $7 = tempRet0;
 $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
 if (!($9)) {
  $$01$i = $7;$$02$i = $6;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $17; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 134 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d26w_PrismaticJoint_setLimitsEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 137 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $vararg_buffer = sp; //@line 137 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,1,44)|0); //@line 33 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 140 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $3 = $2; //@line 140 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $4 = (+_luaL_checknumber($L,3)); //@line 141 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $5 = $4; //@line 141 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 __THREW__ = 0;
 invoke_vidd(3434,($0|0),(+$3),(+$5)); //@line 142 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  STACKTOP = sp;return 0; //@line 143 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $8 = ___cxa_find_matching_catch_3(8992|0)|0;
 $9 = tempRet0;
 $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $11 = ($9|0)==($10|0); //@line 535 "./common/runtime.h"
 if (!($11)) {
  $$01$i = $9;$$02$i = $8;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $12 = (___cxa_begin_catch(($8|0))|0); //@line 535 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = ((($13)) + 8|0); //@line 539 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($16|0)); //@line 539 "./common/runtime.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if (!($18)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $19 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $19; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 143 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $20 = ___cxa_find_matching_catch_2()|0;
 $21 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if ($23) {
  $24 = ___cxa_find_matching_catch_3(0|0)|0;
  $25 = tempRet0;
  ___clang_call_terminate($24); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $21;$$02$i = $20;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d30w_PrismaticJoint_getLowerLimitEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 146 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,1,44)|0); //@line 33 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d14PrismaticJoint13getLowerLimitEv($0)); //@line 149 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $3 = $2; //@line 149 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 _lua_pushnumber($L,$3); //@line 149 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 STACKTOP = sp;return 1; //@line 150 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d30w_PrismaticJoint_getUpperLimitEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 153 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,1,44)|0); //@line 33 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d14PrismaticJoint13getUpperLimitEv($0)); //@line 156 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $3 = $2; //@line 156 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 _lua_pushnumber($L,$3); //@line 156 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 STACKTOP = sp;return 1; //@line 157 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d26w_PrismaticJoint_getLimitsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 160 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d14PrismaticJointEEEPT_P9lua_StateiNS_4TypeE($L,1,44)|0); //@line 33 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 }
 _lua_remove($L,1); //@line 163 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $2 = (__ZN4love7physics5box2d14PrismaticJoint9getLimitsEP9lua_State($0,$L)|0); //@line 164 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 STACKTOP = sp;return ($2|0); //@line 164 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
}
function _luaopen_prismaticjoint($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 189 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 HEAP32[$vararg_buffer>>2] = 18560; //@line 191 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 191 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 HEAP32[$vararg_ptr1>>2] = 10404; //@line 191 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 191 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 191 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,44,260711,$vararg_buffer)|0); //@line 191 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
 STACKTOP = sp;return ($0|0); //@line 191 "modules/physics/box2d/wrap_PrismaticJoint.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d11PulleyJointEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d30w_PulleyJoint_getGroundAnchorsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 38 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d11PulleyJointEEEPT_P9lua_StateiNS_4TypeE($L,1,46)|0); //@line 32 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 }
 _lua_remove($L,1); //@line 41 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 $2 = (__ZN4love7physics5box2d11PulleyJoint16getGroundAnchorsEP9lua_State($0,$L)|0); //@line 42 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 STACKTOP = sp;return ($2|0); //@line 42 "modules/physics/box2d/wrap_PulleyJoint.cpp"
}
function __ZN4love7physics5box2d24w_PulleyJoint_getLengthAEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 45 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d11PulleyJointEEEPT_P9lua_StateiNS_4TypeE($L,1,46)|0); //@line 32 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d11PulleyJoint10getLengthAEv($0)); //@line 48 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 $3 = $2; //@line 48 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 _lua_pushnumber($L,$3); //@line 48 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 STACKTOP = sp;return 1; //@line 49 "modules/physics/box2d/wrap_PulleyJoint.cpp"
}
function __ZN4love7physics5box2d24w_PulleyJoint_getLengthBEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 52 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d11PulleyJointEEEPT_P9lua_StateiNS_4TypeE($L,1,46)|0); //@line 32 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d11PulleyJoint10getLengthBEv($0)); //@line 55 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 $3 = $2; //@line 55 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 _lua_pushnumber($L,$3); //@line 55 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 STACKTOP = sp;return 1; //@line 56 "modules/physics/box2d/wrap_PulleyJoint.cpp"
}
function __ZN4love7physics5box2d22w_PulleyJoint_getRatioEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 59 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d11PulleyJointEEEPT_P9lua_StateiNS_4TypeE($L,1,46)|0); //@line 32 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d11PulleyJoint8getRatioEv($0)); //@line 62 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 $3 = $2; //@line 62 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 _lua_pushnumber($L,$3); //@line 62 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 STACKTOP = sp;return 1; //@line 63 "modules/physics/box2d/wrap_PulleyJoint.cpp"
}
function _luaopen_pulleyjoint($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 75 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 HEAP32[$vararg_buffer>>2] = 18560; //@line 77 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 77 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 HEAP32[$vararg_ptr1>>2] = 10548; //@line 77 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 77 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 77 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,46,260774,$vararg_buffer)|0); //@line 77 "modules/physics/box2d/wrap_PulleyJoint.cpp"
 STACKTOP = sp;return ($0|0); //@line 77 "modules/physics/box2d/wrap_PulleyJoint.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d29w_RevoluteJoint_getJointAngleEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 39 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,1,45)|0); //@line 33 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d13RevoluteJoint13getJointAngleEv($0)); //@line 42 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $3 = $2; //@line 42 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 _lua_pushnumber($L,$3); //@line 42 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 STACKTOP = sp;return 1; //@line 43 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d29w_RevoluteJoint_getJointSpeedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 46 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,1,45)|0); //@line 33 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d13RevoluteJoint13getJointSpeedEv($0)); //@line 49 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $3 = $2; //@line 49 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 _lua_pushnumber($L,$3); //@line 49 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 STACKTOP = sp;return 1; //@line 50 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d31w_RevoluteJoint_setMotorEnabledEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 53 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,1,45)|0); //@line 33 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $2 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,2)|0); //@line 56 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 __ZN4love7physics5box2d13RevoluteJoint15setMotorEnabledEb($0,$2); //@line 57 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 STACKTOP = sp;return 0; //@line 58 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d30w_RevoluteJoint_isMotorEnabledEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 61 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,1,45)|0); //@line 33 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $2 = (__ZNK4love7physics5box2d13RevoluteJoint14isMotorEnabledEv($0)|0); //@line 64 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$2); //@line 64 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 STACKTOP = sp;return 1; //@line 65 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d33w_RevoluteJoint_setMaxMotorTorqueEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 68 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,1,45)|0); //@line 33 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 71 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $3 = $2; //@line 71 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 __ZN4love7physics5box2d13RevoluteJoint17setMaxMotorTorqueEf($0,$3); //@line 72 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 STACKTOP = sp;return 0; //@line 73 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d29w_RevoluteJoint_setMotorSpeedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 76 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,1,45)|0); //@line 33 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 79 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $3 = $2; //@line 79 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 __ZN4love7physics5box2d13RevoluteJoint13setMotorSpeedEf($0,$3); //@line 80 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 STACKTOP = sp;return 0; //@line 81 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d29w_RevoluteJoint_getMotorSpeedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 84 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,1,45)|0); //@line 33 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d13RevoluteJoint13getMotorSpeedEv($0)); //@line 87 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $3 = $2; //@line 87 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 _lua_pushnumber($L,$3); //@line 87 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 STACKTOP = sp;return 1; //@line 88 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d30w_RevoluteJoint_getMotorTorqueEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 91 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,1,45)|0); //@line 33 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 94 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $3 = $2; //@line 94 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $4 = (+__ZNK4love7physics5box2d13RevoluteJoint14getMotorTorqueEf($0,$3)); //@line 95 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $5 = $4; //@line 95 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 _lua_pushnumber($L,$5); //@line 95 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 STACKTOP = sp;return 1; //@line 96 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d33w_RevoluteJoint_getMaxMotorTorqueEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 99 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,1,45)|0); //@line 33 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d13RevoluteJoint17getMaxMotorTorqueEv($0)); //@line 102 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $3 = $2; //@line 102 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 _lua_pushnumber($L,$3); //@line 102 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 STACKTOP = sp;return 1; //@line 103 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d32w_RevoluteJoint_setLimitsEnabledEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 106 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,1,45)|0); //@line 33 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $2 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,2)|0); //@line 109 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 __ZN4love7physics5box2d13RevoluteJoint16setLimitsEnabledEb($0,$2); //@line 110 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 STACKTOP = sp;return 0; //@line 111 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d32w_RevoluteJoint_hasLimitsEnabledEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 114 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,1,45)|0); //@line 33 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $2 = (__ZNK4love7physics5box2d13RevoluteJoint16hasLimitsEnabledEv($0)|0); //@line 117 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$2); //@line 117 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 STACKTOP = sp;return 1; //@line 118 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d29w_RevoluteJoint_setUpperLimitEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0.0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 121 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $vararg_buffer = sp; //@line 121 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,1,45)|0); //@line 33 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 124 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $3 = $2; //@line 124 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 __THREW__ = 0;
 invoke_vid(3435,($0|0),(+$3)); //@line 125 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  STACKTOP = sp;return 0; //@line 126 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $6 = ___cxa_find_matching_catch_3(8992|0)|0;
 $7 = tempRet0;
 $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
 if (!($9)) {
  $$01$i = $7;$$02$i = $6;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $17; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 126 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d29w_RevoluteJoint_setLowerLimitEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0.0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 129 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $vararg_buffer = sp; //@line 129 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,1,45)|0); //@line 33 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 132 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $3 = $2; //@line 132 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 __THREW__ = 0;
 invoke_vid(3436,($0|0),(+$3)); //@line 133 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  STACKTOP = sp;return 0; //@line 134 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $6 = ___cxa_find_matching_catch_3(8992|0)|0;
 $7 = tempRet0;
 $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
 if (!($9)) {
  $$01$i = $7;$$02$i = $6;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $17; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 134 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d25w_RevoluteJoint_setLimitsEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 137 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $vararg_buffer = sp; //@line 137 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,1,45)|0); //@line 33 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 140 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $3 = $2; //@line 140 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $4 = (+_luaL_checknumber($L,3)); //@line 141 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $5 = $4; //@line 141 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 __THREW__ = 0;
 invoke_vidd(3437,($0|0),(+$3),(+$5)); //@line 142 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  STACKTOP = sp;return 0; //@line 143 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $8 = ___cxa_find_matching_catch_3(8992|0)|0;
 $9 = tempRet0;
 $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $11 = ($9|0)==($10|0); //@line 535 "./common/runtime.h"
 if (!($11)) {
  $$01$i = $9;$$02$i = $8;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $12 = (___cxa_begin_catch(($8|0))|0); //@line 535 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = ((($13)) + 8|0); //@line 539 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($16|0)); //@line 539 "./common/runtime.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if (!($18)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $19 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $19; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 143 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $20 = ___cxa_find_matching_catch_2()|0;
 $21 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if ($23) {
  $24 = ___cxa_find_matching_catch_3(0|0)|0;
  $25 = tempRet0;
  ___clang_call_terminate($24); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $21;$$02$i = $20;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d29w_RevoluteJoint_getLowerLimitEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 146 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,1,45)|0); //@line 33 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d13RevoluteJoint13getLowerLimitEv($0)); //@line 149 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $3 = $2; //@line 149 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 _lua_pushnumber($L,$3); //@line 149 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 STACKTOP = sp;return 1; //@line 150 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d29w_RevoluteJoint_getUpperLimitEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 153 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,1,45)|0); //@line 33 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d13RevoluteJoint13getUpperLimitEv($0)); //@line 156 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $3 = $2; //@line 156 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 _lua_pushnumber($L,$3); //@line 156 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 STACKTOP = sp;return 1; //@line 157 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d25w_RevoluteJoint_getLimitsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 160 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13RevoluteJointEEEPT_P9lua_StateiNS_4TypeE($L,1,45)|0); //@line 33 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 }
 _lua_remove($L,1); //@line 163 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $2 = (__ZN4love7physics5box2d13RevoluteJoint9getLimitsEP9lua_State($0,$L)|0); //@line 164 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 STACKTOP = sp;return ($2|0); //@line 164 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
}
function _luaopen_revolutejoint($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 189 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 HEAP32[$vararg_buffer>>2] = 18560; //@line 191 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 191 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 HEAP32[$vararg_ptr1>>2] = 10588; //@line 191 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 191 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 191 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,45,260825,$vararg_buffer)|0); //@line 191 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
 STACKTOP = sp;return ($0|0); //@line 191 "modules/physics/box2d/wrap_RevoluteJoint.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d9RopeJointEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d24w_RopeJoint_getMaxLengthEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 38 "modules/physics/box2d/wrap_RopeJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d9RopeJointEEEPT_P9lua_StateiNS_4TypeE($L,1,50)|0); //@line 32 "modules/physics/box2d/wrap_RopeJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_RopeJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_RopeJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d9RopeJoint12getMaxLengthEv($0)); //@line 41 "modules/physics/box2d/wrap_RopeJoint.cpp"
 $3 = $2; //@line 41 "modules/physics/box2d/wrap_RopeJoint.cpp"
 _lua_pushnumber($L,$3); //@line 41 "modules/physics/box2d/wrap_RopeJoint.cpp"
 STACKTOP = sp;return 1; //@line 42 "modules/physics/box2d/wrap_RopeJoint.cpp"
}
function _luaopen_ropejoint($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 51 "modules/physics/box2d/wrap_RopeJoint.cpp"
 HEAP32[$vararg_buffer>>2] = 18560; //@line 53 "modules/physics/box2d/wrap_RopeJoint.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 53 "modules/physics/box2d/wrap_RopeJoint.cpp"
 HEAP32[$vararg_ptr1>>2] = 10732; //@line 53 "modules/physics/box2d/wrap_RopeJoint.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 53 "modules/physics/box2d/wrap_RopeJoint.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 53 "modules/physics/box2d/wrap_RopeJoint.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,50,260963,$vararg_buffer)|0); //@line 53 "modules/physics/box2d/wrap_RopeJoint.cpp"
 STACKTOP = sp;return ($0|0); //@line 53 "modules/physics/box2d/wrap_RopeJoint.cpp"
}
function __ZN4love7physics5box2d15w_Shape_getTypeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $type = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $type = sp;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5ShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,36)|0); //@line 33 "modules/physics/box2d/wrap_Shape.cpp"
 HEAP32[$type>>2] = 1282466; //@line 39 "modules/physics/box2d/wrap_Shape.cpp"
 $1 = (__ZNK4love7physics5box2d5Shape7getTypeEv($0)|0); //@line 40 "modules/physics/box2d/wrap_Shape.cpp"
 (__ZN4love7physics5Shape11getConstantENS1_4TypeERPKc($1,$type)|0); //@line 40 "modules/physics/box2d/wrap_Shape.cpp"
 $2 = HEAP32[$type>>2]|0; //@line 41 "modules/physics/box2d/wrap_Shape.cpp"
 _lua_pushstring($L,$2); //@line 41 "modules/physics/box2d/wrap_Shape.cpp"
 STACKTOP = sp;return 1; //@line 42 "modules/physics/box2d/wrap_Shape.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d5ShapeEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d17w_Shape_getRadiusEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5ShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,36)|0); //@line 33 "modules/physics/box2d/wrap_Shape.cpp"
 $1 = (+__ZNK4love7physics5box2d5Shape9getRadiusEv($0)); //@line 48 "modules/physics/box2d/wrap_Shape.cpp"
 $2 = $1; //@line 49 "modules/physics/box2d/wrap_Shape.cpp"
 _lua_pushnumber($L,$2); //@line 49 "modules/physics/box2d/wrap_Shape.cpp"
 return 1; //@line 50 "modules/physics/box2d/wrap_Shape.cpp"
}
function __ZN4love7physics5box2d21w_Shape_getChildCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5ShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,36)|0); //@line 33 "modules/physics/box2d/wrap_Shape.cpp"
 $1 = (__ZNK4love7physics5box2d5Shape13getChildCountEv($0)|0); //@line 56 "modules/physics/box2d/wrap_Shape.cpp"
 _lua_pushinteger($L,$1); //@line 57 "modules/physics/box2d/wrap_Shape.cpp"
 return 1; //@line 58 "modules/physics/box2d/wrap_Shape.cpp"
}
function __ZN4love7physics5box2d17w_Shape_testPointEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0, $12 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5ShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,36)|0); //@line 33 "modules/physics/box2d/wrap_Shape.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 64 "modules/physics/box2d/wrap_Shape.cpp"
 $2 = $1; //@line 64 "modules/physics/box2d/wrap_Shape.cpp"
 $3 = (+_luaL_checknumber($L,3)); //@line 65 "modules/physics/box2d/wrap_Shape.cpp"
 $4 = $3; //@line 65 "modules/physics/box2d/wrap_Shape.cpp"
 $5 = (+_luaL_checknumber($L,4)); //@line 66 "modules/physics/box2d/wrap_Shape.cpp"
 $6 = $5; //@line 66 "modules/physics/box2d/wrap_Shape.cpp"
 $7 = (+_luaL_checknumber($L,5)); //@line 67 "modules/physics/box2d/wrap_Shape.cpp"
 $8 = $7; //@line 67 "modules/physics/box2d/wrap_Shape.cpp"
 $9 = (+_luaL_checknumber($L,6)); //@line 68 "modules/physics/box2d/wrap_Shape.cpp"
 $10 = $9; //@line 68 "modules/physics/box2d/wrap_Shape.cpp"
 $11 = (__ZNK4love7physics5box2d5Shape9testPointEfffff($0,$2,$4,$6,$8,$10)|0); //@line 69 "modules/physics/box2d/wrap_Shape.cpp"
 $12 = $11&1; //@line 70 "modules/physics/box2d/wrap_Shape.cpp"
 _lua_pushboolean($L,$12); //@line 70 "modules/physics/box2d/wrap_Shape.cpp"
 return 1; //@line 71 "modules/physics/box2d/wrap_Shape.cpp"
}
function __ZN4love7physics5box2d15w_Shape_rayCastEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ret$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 74 "modules/physics/box2d/wrap_Shape.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5ShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,36)|0); //@line 33 "modules/physics/box2d/wrap_Shape.cpp"
 _lua_remove($L,1); //@line 77 "modules/physics/box2d/wrap_Shape.cpp"
 __THREW__ = 0;
 $1 = (invoke_iii(3438,($0|0),($L|0))|0); //@line 79 "modules/physics/box2d/wrap_Shape.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  $ret$0 = $1;
  STACKTOP = sp;return ($ret$0|0); //@line 80 "modules/physics/box2d/wrap_Shape.cpp"
 }
 $4 = ___cxa_find_matching_catch_3(8992|0)|0;
 $5 = tempRet0;
 $6 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $7 = ($5|0)==($6|0); //@line 535 "./common/runtime.h"
 if (!($7)) {
  $$01$i = $5;$$02$i = $4;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $8 = (___cxa_begin_catch(($4|0))|0); //@line 535 "./common/runtime.h"
 $9 = HEAP32[$8>>2]|0; //@line 539 "./common/runtime.h"
 $10 = ((($9)) + 8|0); //@line 539 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($8)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($12|0)); //@line 539 "./common/runtime.h"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if (!($14)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $15 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $15; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $ret$0 = 0;
  STACKTOP = sp;return ($ret$0|0); //@line 80 "modules/physics/box2d/wrap_Shape.cpp"
 }
 $16 = ___cxa_find_matching_catch_2()|0;
 $17 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $20 = ___cxa_find_matching_catch_3(0|0)|0;
  $21 = tempRet0;
  ___clang_call_terminate($20); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $17;$$02$i = $16;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d19w_Shape_computeAABBEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5ShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,36)|0); //@line 33 "modules/physics/box2d/wrap_Shape.cpp"
 _lua_remove($L,1); //@line 86 "modules/physics/box2d/wrap_Shape.cpp"
 $1 = (__ZNK4love7physics5box2d5Shape11computeAABBEP9lua_State($0,$L)|0); //@line 87 "modules/physics/box2d/wrap_Shape.cpp"
 return ($1|0); //@line 87 "modules/physics/box2d/wrap_Shape.cpp"
}
function __ZN4love7physics5box2d19w_Shape_computeMassEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5ShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,36)|0); //@line 33 "modules/physics/box2d/wrap_Shape.cpp"
 _lua_remove($L,1); //@line 93 "modules/physics/box2d/wrap_Shape.cpp"
 $1 = (__ZNK4love7physics5box2d5Shape11computeMassEP9lua_State($0,$L)|0); //@line 94 "modules/physics/box2d/wrap_Shape.cpp"
 return ($1|0); //@line 94 "modules/physics/box2d/wrap_Shape.cpp"
}
function __ZN4love7physics5box2d15luax_checkshapeEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5ShapeEEEPT_P9lua_StateiNS_4TypeE($L,$idx,36)|0); //@line 33 "modules/physics/box2d/wrap_Shape.cpp"
 return ($0|0); //@line 33 "modules/physics/box2d/wrap_Shape.cpp"
}
function _luaopen_shape($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 109 "modules/physics/box2d/wrap_Shape.cpp"
 HEAP32[$vararg_buffer>>2] = 10748; //@line 111 "modules/physics/box2d/wrap_Shape.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 111 "modules/physics/box2d/wrap_Shape.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 111 "modules/physics/box2d/wrap_Shape.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,36,261024,$vararg_buffer)|0); //@line 111 "modules/physics/box2d/wrap_Shape.cpp"
 STACKTOP = sp;return ($0|0); //@line 111 "modules/physics/box2d/wrap_Shape.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d9WeldJointEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d24w_WeldJoint_setFrequencyEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 38 "modules/physics/box2d/wrap_WeldJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d9WeldJointEEEPT_P9lua_StateiNS_4TypeE($L,1,49)|0); //@line 32 "modules/physics/box2d/wrap_WeldJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_WeldJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_WeldJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 41 "modules/physics/box2d/wrap_WeldJoint.cpp"
 $3 = $2; //@line 41 "modules/physics/box2d/wrap_WeldJoint.cpp"
 __ZN4love7physics5box2d9WeldJoint12setFrequencyEf($0,$3); //@line 42 "modules/physics/box2d/wrap_WeldJoint.cpp"
 STACKTOP = sp;return 0; //@line 43 "modules/physics/box2d/wrap_WeldJoint.cpp"
}
function __ZN4love7physics5box2d24w_WeldJoint_getFrequencyEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 46 "modules/physics/box2d/wrap_WeldJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d9WeldJointEEEPT_P9lua_StateiNS_4TypeE($L,1,49)|0); //@line 32 "modules/physics/box2d/wrap_WeldJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_WeldJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_WeldJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d9WeldJoint12getFrequencyEv($0)); //@line 49 "modules/physics/box2d/wrap_WeldJoint.cpp"
 $3 = $2; //@line 49 "modules/physics/box2d/wrap_WeldJoint.cpp"
 _lua_pushnumber($L,$3); //@line 49 "modules/physics/box2d/wrap_WeldJoint.cpp"
 STACKTOP = sp;return 1; //@line 50 "modules/physics/box2d/wrap_WeldJoint.cpp"
}
function __ZN4love7physics5box2d27w_WeldJoint_setDampingRatioEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 53 "modules/physics/box2d/wrap_WeldJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d9WeldJointEEEPT_P9lua_StateiNS_4TypeE($L,1,49)|0); //@line 32 "modules/physics/box2d/wrap_WeldJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_WeldJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_WeldJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 56 "modules/physics/box2d/wrap_WeldJoint.cpp"
 $3 = $2; //@line 56 "modules/physics/box2d/wrap_WeldJoint.cpp"
 __ZN4love7physics5box2d9WeldJoint15setDampingRatioEf($0,$3); //@line 57 "modules/physics/box2d/wrap_WeldJoint.cpp"
 STACKTOP = sp;return 0; //@line 58 "modules/physics/box2d/wrap_WeldJoint.cpp"
}
function __ZN4love7physics5box2d27w_WeldJoint_getDampingRatioEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 61 "modules/physics/box2d/wrap_WeldJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d9WeldJointEEEPT_P9lua_StateiNS_4TypeE($L,1,49)|0); //@line 32 "modules/physics/box2d/wrap_WeldJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_WeldJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_WeldJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d9WeldJoint15getDampingRatioEv($0)); //@line 64 "modules/physics/box2d/wrap_WeldJoint.cpp"
 $3 = $2; //@line 64 "modules/physics/box2d/wrap_WeldJoint.cpp"
 _lua_pushnumber($L,$3); //@line 64 "modules/physics/box2d/wrap_WeldJoint.cpp"
 STACKTOP = sp;return 1; //@line 65 "modules/physics/box2d/wrap_WeldJoint.cpp"
}
function _luaopen_weldjoint($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 77 "modules/physics/box2d/wrap_WeldJoint.cpp"
 HEAP32[$vararg_buffer>>2] = 18560; //@line 79 "modules/physics/box2d/wrap_WeldJoint.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 79 "modules/physics/box2d/wrap_WeldJoint.cpp"
 HEAP32[$vararg_ptr1>>2] = 10812; //@line 79 "modules/physics/box2d/wrap_WeldJoint.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 79 "modules/physics/box2d/wrap_WeldJoint.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 79 "modules/physics/box2d/wrap_WeldJoint.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,49,261030,$vararg_buffer)|0); //@line 79 "modules/physics/box2d/wrap_WeldJoint.cpp"
 STACKTOP = sp;return ($0|0); //@line 79 "modules/physics/box2d/wrap_WeldJoint.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d10WheelJointEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d32w_WheelJoint_getJointTranslationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 38 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10WheelJointEEEPT_P9lua_StateiNS_4TypeE($L,1,51)|0); //@line 32 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_WheelJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_WheelJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d10WheelJoint19getJointTranslationEv($0)); //@line 41 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $3 = $2; //@line 41 "modules/physics/box2d/wrap_WheelJoint.cpp"
 _lua_pushnumber($L,$3); //@line 41 "modules/physics/box2d/wrap_WheelJoint.cpp"
 STACKTOP = sp;return 1; //@line 42 "modules/physics/box2d/wrap_WheelJoint.cpp"
}
function __ZN4love7physics5box2d26w_WheelJoint_getJointSpeedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 45 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10WheelJointEEEPT_P9lua_StateiNS_4TypeE($L,1,51)|0); //@line 32 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_WheelJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_WheelJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d10WheelJoint13getJointSpeedEv($0)); //@line 48 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $3 = $2; //@line 48 "modules/physics/box2d/wrap_WheelJoint.cpp"
 _lua_pushnumber($L,$3); //@line 48 "modules/physics/box2d/wrap_WheelJoint.cpp"
 STACKTOP = sp;return 1; //@line 49 "modules/physics/box2d/wrap_WheelJoint.cpp"
}
function __ZN4love7physics5box2d28w_WheelJoint_setMotorEnabledEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 52 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10WheelJointEEEPT_P9lua_StateiNS_4TypeE($L,1,51)|0); //@line 32 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_WheelJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_WheelJoint.cpp"
 }
 $2 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,2)|0); //@line 55 "modules/physics/box2d/wrap_WheelJoint.cpp"
 __ZN4love7physics5box2d10WheelJoint15setMotorEnabledEb($0,$2); //@line 56 "modules/physics/box2d/wrap_WheelJoint.cpp"
 STACKTOP = sp;return 0; //@line 57 "modules/physics/box2d/wrap_WheelJoint.cpp"
}
function __ZN4love7physics5box2d27w_WheelJoint_isMotorEnabledEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 60 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10WheelJointEEEPT_P9lua_StateiNS_4TypeE($L,1,51)|0); //@line 32 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_WheelJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_WheelJoint.cpp"
 }
 $2 = (__ZNK4love7physics5box2d10WheelJoint14isMotorEnabledEv($0)|0); //@line 63 "modules/physics/box2d/wrap_WheelJoint.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$2); //@line 63 "modules/physics/box2d/wrap_WheelJoint.cpp"
 STACKTOP = sp;return 1; //@line 64 "modules/physics/box2d/wrap_WheelJoint.cpp"
}
function __ZN4love7physics5box2d26w_WheelJoint_setMotorSpeedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 67 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10WheelJointEEEPT_P9lua_StateiNS_4TypeE($L,1,51)|0); //@line 32 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_WheelJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_WheelJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 70 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $3 = $2; //@line 70 "modules/physics/box2d/wrap_WheelJoint.cpp"
 __ZN4love7physics5box2d10WheelJoint13setMotorSpeedEf($0,$3); //@line 71 "modules/physics/box2d/wrap_WheelJoint.cpp"
 STACKTOP = sp;return 0; //@line 72 "modules/physics/box2d/wrap_WheelJoint.cpp"
}
function __ZN4love7physics5box2d26w_WheelJoint_getMotorSpeedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 75 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10WheelJointEEEPT_P9lua_StateiNS_4TypeE($L,1,51)|0); //@line 32 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_WheelJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_WheelJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d10WheelJoint13getMotorSpeedEv($0)); //@line 78 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $3 = $2; //@line 78 "modules/physics/box2d/wrap_WheelJoint.cpp"
 _lua_pushnumber($L,$3); //@line 78 "modules/physics/box2d/wrap_WheelJoint.cpp"
 STACKTOP = sp;return 1; //@line 79 "modules/physics/box2d/wrap_WheelJoint.cpp"
}
function __ZN4love7physics5box2d30w_WheelJoint_setMaxMotorTorqueEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 82 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10WheelJointEEEPT_P9lua_StateiNS_4TypeE($L,1,51)|0); //@line 32 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_WheelJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_WheelJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 85 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $3 = $2; //@line 85 "modules/physics/box2d/wrap_WheelJoint.cpp"
 __ZN4love7physics5box2d10WheelJoint17setMaxMotorTorqueEf($0,$3); //@line 86 "modules/physics/box2d/wrap_WheelJoint.cpp"
 STACKTOP = sp;return 0; //@line 87 "modules/physics/box2d/wrap_WheelJoint.cpp"
}
function __ZN4love7physics5box2d30w_WheelJoint_getMaxMotorTorqueEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 90 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10WheelJointEEEPT_P9lua_StateiNS_4TypeE($L,1,51)|0); //@line 32 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_WheelJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_WheelJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d10WheelJoint17getMaxMotorTorqueEv($0)); //@line 93 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $3 = $2; //@line 93 "modules/physics/box2d/wrap_WheelJoint.cpp"
 _lua_pushnumber($L,$3); //@line 93 "modules/physics/box2d/wrap_WheelJoint.cpp"
 STACKTOP = sp;return 1; //@line 94 "modules/physics/box2d/wrap_WheelJoint.cpp"
}
function __ZN4love7physics5box2d27w_WheelJoint_getMotorTorqueEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 97 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10WheelJointEEEPT_P9lua_StateiNS_4TypeE($L,1,51)|0); //@line 32 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_WheelJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_WheelJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 100 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $3 = $2; //@line 100 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $4 = (+__ZNK4love7physics5box2d10WheelJoint14getMotorTorqueEf($0,$3)); //@line 101 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $5 = $4; //@line 101 "modules/physics/box2d/wrap_WheelJoint.cpp"
 _lua_pushnumber($L,$5); //@line 101 "modules/physics/box2d/wrap_WheelJoint.cpp"
 STACKTOP = sp;return 1; //@line 102 "modules/physics/box2d/wrap_WheelJoint.cpp"
}
function __ZN4love7physics5box2d31w_WheelJoint_setSpringFrequencyEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 105 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10WheelJointEEEPT_P9lua_StateiNS_4TypeE($L,1,51)|0); //@line 32 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_WheelJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_WheelJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 108 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $3 = $2; //@line 108 "modules/physics/box2d/wrap_WheelJoint.cpp"
 __ZN4love7physics5box2d10WheelJoint18setSpringFrequencyEf($0,$3); //@line 109 "modules/physics/box2d/wrap_WheelJoint.cpp"
 STACKTOP = sp;return 0; //@line 110 "modules/physics/box2d/wrap_WheelJoint.cpp"
}
function __ZN4love7physics5box2d31w_WheelJoint_getSpringFrequencyEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 113 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10WheelJointEEEPT_P9lua_StateiNS_4TypeE($L,1,51)|0); //@line 32 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_WheelJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_WheelJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d10WheelJoint18getSpringFrequencyEv($0)); //@line 116 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $3 = $2; //@line 116 "modules/physics/box2d/wrap_WheelJoint.cpp"
 _lua_pushnumber($L,$3); //@line 116 "modules/physics/box2d/wrap_WheelJoint.cpp"
 STACKTOP = sp;return 1; //@line 117 "modules/physics/box2d/wrap_WheelJoint.cpp"
}
function __ZN4love7physics5box2d34w_WheelJoint_setSpringDampingRatioEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 120 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10WheelJointEEEPT_P9lua_StateiNS_4TypeE($L,1,51)|0); //@line 32 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_WheelJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_WheelJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 123 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $3 = $2; //@line 123 "modules/physics/box2d/wrap_WheelJoint.cpp"
 __ZN4love7physics5box2d10WheelJoint21setSpringDampingRatioEf($0,$3); //@line 124 "modules/physics/box2d/wrap_WheelJoint.cpp"
 STACKTOP = sp;return 0; //@line 125 "modules/physics/box2d/wrap_WheelJoint.cpp"
}
function __ZN4love7physics5box2d34w_WheelJoint_getSpringDampingRatioEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 128 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10WheelJointEEEPT_P9lua_StateiNS_4TypeE($L,1,51)|0); //@line 32 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_WheelJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_WheelJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d10WheelJoint21getSpringDampingRatioEv($0)); //@line 131 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $3 = $2; //@line 131 "modules/physics/box2d/wrap_WheelJoint.cpp"
 _lua_pushnumber($L,$3); //@line 131 "modules/physics/box2d/wrap_WheelJoint.cpp"
 STACKTOP = sp;return 1; //@line 132 "modules/physics/box2d/wrap_WheelJoint.cpp"
}
function _luaopen_wheeljoint($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 153 "modules/physics/box2d/wrap_WheelJoint.cpp"
 HEAP32[$vararg_buffer>>2] = 18560; //@line 155 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 155 "modules/physics/box2d/wrap_WheelJoint.cpp"
 HEAP32[$vararg_ptr1>>2] = 10852; //@line 155 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 155 "modules/physics/box2d/wrap_WheelJoint.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 155 "modules/physics/box2d/wrap_WheelJoint.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,51,261040,$vararg_buffer)|0); //@line 155 "modules/physics/box2d/wrap_WheelJoint.cpp"
 STACKTOP = sp;return ($0|0); //@line 155 "modules/physics/box2d/wrap_WheelJoint.cpp"
}
function __ZN4love7physics5box2d15luax_checkworldEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 30 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,$idx,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if ($1) {
  STACKTOP = sp;return ($0|0); //@line 35 "modules/physics/box2d/wrap_World.cpp"
 }
 (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 STACKTOP = sp;return ($0|0); //@line 35 "modules/physics/box2d/wrap_World.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d14w_World_updateEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0.0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 38 "modules/physics/box2d/wrap_World.cpp"
 $vararg_buffer = sp; //@line 38 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 41 "modules/physics/box2d/wrap_World.cpp"
 $3 = $2; //@line 41 "modules/physics/box2d/wrap_World.cpp"
 __ZN4love7physics5box2d5World13setCallbacksLEP9lua_State($0,$L); //@line 43 "modules/physics/box2d/wrap_World.cpp"
 __THREW__ = 0;
 invoke_vid(3439,($0|0),(+$3)); //@line 44 "modules/physics/box2d/wrap_World.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  STACKTOP = sp;return 0; //@line 45 "modules/physics/box2d/wrap_World.cpp"
 }
 $6 = ___cxa_find_matching_catch_3(8992|0)|0;
 $7 = tempRet0;
 $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
 if (!($9)) {
  $$01$i = $7;$$02$i = $6;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $17; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 45 "modules/physics/box2d/wrap_World.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d20w_World_setCallbacksEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 48 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 _lua_remove($L,1); //@line 51 "modules/physics/box2d/wrap_World.cpp"
 $2 = (__ZN4love7physics5box2d5World12setCallbacksEP9lua_State($0,$L)|0); //@line 52 "modules/physics/box2d/wrap_World.cpp"
 STACKTOP = sp;return ($2|0); //@line 52 "modules/physics/box2d/wrap_World.cpp"
}
function __ZN4love7physics5box2d20w_World_getCallbacksEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 55 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 _lua_remove($L,1); //@line 58 "modules/physics/box2d/wrap_World.cpp"
 $2 = (__ZN4love7physics5box2d5World12getCallbacksEP9lua_State($0,$L)|0); //@line 59 "modules/physics/box2d/wrap_World.cpp"
 STACKTOP = sp;return ($2|0); //@line 59 "modules/physics/box2d/wrap_World.cpp"
}
function __ZN4love7physics5box2d24w_World_setContactFilterEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 62 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 _lua_remove($L,1); //@line 65 "modules/physics/box2d/wrap_World.cpp"
 $2 = (__ZN4love7physics5box2d5World16setContactFilterEP9lua_State($0,$L)|0); //@line 66 "modules/physics/box2d/wrap_World.cpp"
 STACKTOP = sp;return ($2|0); //@line 66 "modules/physics/box2d/wrap_World.cpp"
}
function __ZN4love7physics5box2d24w_World_getContactFilterEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 69 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 _lua_remove($L,1); //@line 72 "modules/physics/box2d/wrap_World.cpp"
 $2 = (__ZN4love7physics5box2d5World16getContactFilterEP9lua_State($0,$L)|0); //@line 73 "modules/physics/box2d/wrap_World.cpp"
 STACKTOP = sp;return ($2|0); //@line 73 "modules/physics/box2d/wrap_World.cpp"
}
function __ZN4love7physics5box2d18w_World_setGravityEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 76 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 79 "modules/physics/box2d/wrap_World.cpp"
 $3 = $2; //@line 79 "modules/physics/box2d/wrap_World.cpp"
 $4 = (+_luaL_checknumber($L,3)); //@line 80 "modules/physics/box2d/wrap_World.cpp"
 $5 = $4; //@line 80 "modules/physics/box2d/wrap_World.cpp"
 __ZN4love7physics5box2d5World10setGravityEff($0,$3,$5); //@line 81 "modules/physics/box2d/wrap_World.cpp"
 STACKTOP = sp;return 0; //@line 82 "modules/physics/box2d/wrap_World.cpp"
}
function __ZN4love7physics5box2d18w_World_getGravityEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 85 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 _lua_remove($L,1); //@line 88 "modules/physics/box2d/wrap_World.cpp"
 $2 = (__ZN4love7physics5box2d5World10getGravityEP9lua_State($0,$L)|0); //@line 89 "modules/physics/box2d/wrap_World.cpp"
 STACKTOP = sp;return ($2|0); //@line 89 "modules/physics/box2d/wrap_World.cpp"
}
function __ZN4love7physics5box2d23w_World_translateOriginEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 92 "modules/physics/box2d/wrap_World.cpp"
 $vararg_buffer = sp; //@line 92 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 95 "modules/physics/box2d/wrap_World.cpp"
 $3 = $2; //@line 95 "modules/physics/box2d/wrap_World.cpp"
 $4 = (+_luaL_checknumber($L,3)); //@line 96 "modules/physics/box2d/wrap_World.cpp"
 $5 = $4; //@line 96 "modules/physics/box2d/wrap_World.cpp"
 __THREW__ = 0;
 invoke_vidd(3440,($0|0),(+$3),(+$5)); //@line 97 "modules/physics/box2d/wrap_World.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  STACKTOP = sp;return 0; //@line 98 "modules/physics/box2d/wrap_World.cpp"
 }
 $8 = ___cxa_find_matching_catch_3(8992|0)|0;
 $9 = tempRet0;
 $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $11 = ($9|0)==($10|0); //@line 535 "./common/runtime.h"
 if (!($11)) {
  $$01$i = $9;$$02$i = $8;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $12 = (___cxa_begin_catch(($8|0))|0); //@line 535 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = ((($13)) + 8|0); //@line 539 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($16|0)); //@line 539 "./common/runtime.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if (!($18)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $19 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $19; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 98 "modules/physics/box2d/wrap_World.cpp"
 }
 $20 = ___cxa_find_matching_catch_2()|0;
 $21 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if ($23) {
  $24 = ___cxa_find_matching_catch_3(0|0)|0;
  $25 = tempRet0;
  ___clang_call_terminate($24); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $21;$$02$i = $20;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d26w_World_setSleepingAllowedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 101 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 $2 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,2)|0); //@line 104 "modules/physics/box2d/wrap_World.cpp"
 __ZN4love7physics5box2d5World18setSleepingAllowedEb($0,$2); //@line 105 "modules/physics/box2d/wrap_World.cpp"
 STACKTOP = sp;return 0; //@line 106 "modules/physics/box2d/wrap_World.cpp"
}
function __ZN4love7physics5box2d25w_World_isSleepingAllowedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 109 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 $2 = (__ZNK4love7physics5box2d5World17isSleepingAllowedEv($0)|0); //@line 112 "modules/physics/box2d/wrap_World.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$2); //@line 112 "modules/physics/box2d/wrap_World.cpp"
 STACKTOP = sp;return 1; //@line 113 "modules/physics/box2d/wrap_World.cpp"
}
function __ZN4love7physics5box2d16w_World_isLockedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 116 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 $2 = (__ZNK4love7physics5box2d5World8isLockedEv($0)|0); //@line 119 "modules/physics/box2d/wrap_World.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$2); //@line 119 "modules/physics/box2d/wrap_World.cpp"
 STACKTOP = sp;return 1; //@line 120 "modules/physics/box2d/wrap_World.cpp"
}
function __ZN4love7physics5box2d20w_World_getBodyCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 123 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 $2 = (__ZNK4love7physics5box2d5World12getBodyCountEv($0)|0); //@line 126 "modules/physics/box2d/wrap_World.cpp"
 _lua_pushinteger($L,$2); //@line 126 "modules/physics/box2d/wrap_World.cpp"
 STACKTOP = sp;return 1; //@line 127 "modules/physics/box2d/wrap_World.cpp"
}
function __ZN4love7physics5box2d21w_World_getJointCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 130 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 $2 = (__ZNK4love7physics5box2d5World13getJointCountEv($0)|0); //@line 133 "modules/physics/box2d/wrap_World.cpp"
 _lua_pushinteger($L,$2); //@line 133 "modules/physics/box2d/wrap_World.cpp"
 STACKTOP = sp;return 1; //@line 134 "modules/physics/box2d/wrap_World.cpp"
}
function __ZN4love7physics5box2d23w_World_getContactCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 137 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 $2 = (__ZNK4love7physics5box2d5World15getContactCountEv($0)|0); //@line 140 "modules/physics/box2d/wrap_World.cpp"
 _lua_pushinteger($L,$2); //@line 140 "modules/physics/box2d/wrap_World.cpp"
 STACKTOP = sp;return 1; //@line 141 "modules/physics/box2d/wrap_World.cpp"
}
function __ZN4love7physics5box2d19w_World_getBodyListEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ret$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 144 "modules/physics/box2d/wrap_World.cpp"
 $vararg_buffer = sp; //@line 144 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 _lua_remove($L,1); //@line 147 "modules/physics/box2d/wrap_World.cpp"
 __THREW__ = 0;
 $2 = (invoke_iii(3441,($0|0),($L|0))|0); //@line 149 "modules/physics/box2d/wrap_World.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  $ret$0 = $2;
  STACKTOP = sp;return ($ret$0|0); //@line 150 "modules/physics/box2d/wrap_World.cpp"
 }
 $5 = ___cxa_find_matching_catch_3(8992|0)|0;
 $6 = tempRet0;
 $7 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $8 = ($6|0)==($7|0); //@line 535 "./common/runtime.h"
 if (!($8)) {
  $$01$i = $6;$$02$i = $5;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $9 = (___cxa_begin_catch(($5|0))|0); //@line 535 "./common/runtime.h"
 $10 = HEAP32[$9>>2]|0; //@line 539 "./common/runtime.h"
 $11 = ((($10)) + 8|0); //@line 539 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = (FUNCTION_TABLE_ii[$12 & 8191]($9)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($13|0)); //@line 539 "./common/runtime.h"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if (!($15)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $16 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $16; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  $ret$0 = 0;
  STACKTOP = sp;return ($ret$0|0); //@line 150 "modules/physics/box2d/wrap_World.cpp"
 }
 $17 = ___cxa_find_matching_catch_2()|0;
 $18 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $21 = ___cxa_find_matching_catch_3(0|0)|0;
  $22 = tempRet0;
  ___clang_call_terminate($21); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $18;$$02$i = $17;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d20w_World_getJointListEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ret$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 153 "modules/physics/box2d/wrap_World.cpp"
 $vararg_buffer = sp; //@line 153 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 _lua_remove($L,1); //@line 156 "modules/physics/box2d/wrap_World.cpp"
 __THREW__ = 0;
 $2 = (invoke_iii(3442,($0|0),($L|0))|0); //@line 158 "modules/physics/box2d/wrap_World.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  $ret$0 = $2;
  STACKTOP = sp;return ($ret$0|0); //@line 159 "modules/physics/box2d/wrap_World.cpp"
 }
 $5 = ___cxa_find_matching_catch_3(8992|0)|0;
 $6 = tempRet0;
 $7 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $8 = ($6|0)==($7|0); //@line 535 "./common/runtime.h"
 if (!($8)) {
  $$01$i = $6;$$02$i = $5;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $9 = (___cxa_begin_catch(($5|0))|0); //@line 535 "./common/runtime.h"
 $10 = HEAP32[$9>>2]|0; //@line 539 "./common/runtime.h"
 $11 = ((($10)) + 8|0); //@line 539 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = (FUNCTION_TABLE_ii[$12 & 8191]($9)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($13|0)); //@line 539 "./common/runtime.h"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if (!($15)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $16 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $16; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  $ret$0 = 0;
  STACKTOP = sp;return ($ret$0|0); //@line 159 "modules/physics/box2d/wrap_World.cpp"
 }
 $17 = ___cxa_find_matching_catch_2()|0;
 $18 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $21 = ___cxa_find_matching_catch_3(0|0)|0;
  $22 = tempRet0;
  ___clang_call_terminate($21); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $18;$$02$i = $17;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d22w_World_getContactListEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ret$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 162 "modules/physics/box2d/wrap_World.cpp"
 $vararg_buffer = sp; //@line 162 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 _lua_remove($L,1); //@line 165 "modules/physics/box2d/wrap_World.cpp"
 __THREW__ = 0;
 $2 = (invoke_iii(3443,($0|0),($L|0))|0); //@line 167 "modules/physics/box2d/wrap_World.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  $ret$0 = $2;
  STACKTOP = sp;return ($ret$0|0); //@line 168 "modules/physics/box2d/wrap_World.cpp"
 }
 $5 = ___cxa_find_matching_catch_3(8992|0)|0;
 $6 = tempRet0;
 $7 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $8 = ($6|0)==($7|0); //@line 535 "./common/runtime.h"
 if (!($8)) {
  $$01$i = $6;$$02$i = $5;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $9 = (___cxa_begin_catch(($5|0))|0); //@line 535 "./common/runtime.h"
 $10 = HEAP32[$9>>2]|0; //@line 539 "./common/runtime.h"
 $11 = ((($10)) + 8|0); //@line 539 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = (FUNCTION_TABLE_ii[$12 & 8191]($9)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($13|0)); //@line 539 "./common/runtime.h"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if (!($15)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $16 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $16; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  $ret$0 = 0;
  STACKTOP = sp;return ($ret$0|0); //@line 168 "modules/physics/box2d/wrap_World.cpp"
 }
 $17 = ___cxa_find_matching_catch_2()|0;
 $18 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $21 = ___cxa_find_matching_catch_3(0|0)|0;
  $22 = tempRet0;
  ___clang_call_terminate($21); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $18;$$02$i = $17;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d24w_World_queryBoundingBoxEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 171 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 _lua_remove($L,1); //@line 174 "modules/physics/box2d/wrap_World.cpp"
 $2 = (__ZN4love7physics5box2d5World16queryBoundingBoxEP9lua_State($0,$L)|0); //@line 175 "modules/physics/box2d/wrap_World.cpp"
 STACKTOP = sp;return ($2|0); //@line 175 "modules/physics/box2d/wrap_World.cpp"
}
function __ZN4love7physics5box2d15w_World_rayCastEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ret$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 178 "modules/physics/box2d/wrap_World.cpp"
 $vararg_buffer = sp; //@line 178 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 _lua_remove($L,1); //@line 181 "modules/physics/box2d/wrap_World.cpp"
 __THREW__ = 0;
 $2 = (invoke_iii(3444,($0|0),($L|0))|0); //@line 183 "modules/physics/box2d/wrap_World.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  $ret$0 = $2;
  STACKTOP = sp;return ($ret$0|0); //@line 184 "modules/physics/box2d/wrap_World.cpp"
 }
 $5 = ___cxa_find_matching_catch_3(8992|0)|0;
 $6 = tempRet0;
 $7 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $8 = ($6|0)==($7|0); //@line 535 "./common/runtime.h"
 if (!($8)) {
  $$01$i = $6;$$02$i = $5;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $9 = (___cxa_begin_catch(($5|0))|0); //@line 535 "./common/runtime.h"
 $10 = HEAP32[$9>>2]|0; //@line 539 "./common/runtime.h"
 $11 = ((($10)) + 8|0); //@line 539 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = (FUNCTION_TABLE_ii[$12 & 8191]($9)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($13|0)); //@line 539 "./common/runtime.h"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if (!($15)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $16 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $16; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  $ret$0 = 0;
  STACKTOP = sp;return ($ret$0|0); //@line 184 "modules/physics/box2d/wrap_World.cpp"
 }
 $17 = ___cxa_find_matching_catch_2()|0;
 $18 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $21 = ___cxa_find_matching_catch_3(0|0)|0;
  $22 = tempRet0;
  ___clang_call_terminate($21); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $18;$$02$i = $17;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d15w_World_destroyEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 187 "modules/physics/box2d/wrap_World.cpp"
 $vararg_buffer = sp; //@line 187 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 32 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_World.cpp"
 if (!($1)) {
  (_luaL_error($L,261277,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_World.cpp"
 }
 __THREW__ = 0;
 invoke_vi(3445,($0|0)); //@line 190 "modules/physics/box2d/wrap_World.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  STACKTOP = sp;return 0; //@line 191 "modules/physics/box2d/wrap_World.cpp"
 }
 $4 = ___cxa_find_matching_catch_3(8992|0)|0;
 $5 = tempRet0;
 $6 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $7 = ($5|0)==($6|0); //@line 535 "./common/runtime.h"
 if (!($7)) {
  $$01$i = $5;$$02$i = $4;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $8 = (___cxa_begin_catch(($4|0))|0); //@line 535 "./common/runtime.h"
 $9 = HEAP32[$8>>2]|0; //@line 539 "./common/runtime.h"
 $10 = ((($9)) + 8|0); //@line 539 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($8)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($12|0)); //@line 539 "./common/runtime.h"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if (!($14)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $15 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $15; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 191 "modules/physics/box2d/wrap_World.cpp"
 }
 $16 = ___cxa_find_matching_catch_2()|0;
 $17 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $20 = ___cxa_find_matching_catch_3(0|0)|0;
  $21 = tempRet0;
  ___clang_call_terminate($20); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $17;$$02$i = $16;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d19w_World_isDestroyedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5WorldEEEPT_P9lua_StateiNS_4TypeE($L,1,32)|0); //@line 196 "modules/physics/box2d/wrap_World.cpp"
 $1 = (__ZNK4love7physics5box2d5World7isValidEv($0)|0); //@line 197 "modules/physics/box2d/wrap_World.cpp"
 $2 = $1 ^ 1; //@line 197 "modules/physics/box2d/wrap_World.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$2); //@line 197 "modules/physics/box2d/wrap_World.cpp"
 return 1; //@line 198 "modules/physics/box2d/wrap_World.cpp"
}
function _luaopen_world($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 228 "modules/physics/box2d/wrap_World.cpp"
 HEAP32[$vararg_buffer>>2] = 10964; //@line 230 "modules/physics/box2d/wrap_World.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 230 "modules/physics/box2d/wrap_World.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 230 "modules/physics/box2d/wrap_World.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,32,261309,$vararg_buffer)|0); //@line 230 "modules/physics/box2d/wrap_World.cpp"
 STACKTOP = sp;return ($0|0); //@line 230 "modules/physics/box2d/wrap_World.cpp"
}
function __ZN4love7physics4BodyD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 30 "modules/physics/Body.cpp"
}
function __GLOBAL__sub_I_Body_cpp() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[(625692)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(625704)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(625716)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(625728)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(625740)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(625752)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(625764)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(625776)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(625780)>>2]=0|0;HEAP32[(625780)+4>>2]=0|0;HEAP32[(625780)+8>>2]=0|0;HEAP32[(625780)+12>>2]=0|0; //@line 44 "./common/StringMap.h"
 $0 = HEAP32[2785]|0; //@line 49 "./common/StringMap.h"
 $1 = HEAP32[(11144)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_7physics4Body4TypeELj4EE3addEPKcS3_(625684,$0,$1)|0); //@line 49 "./common/StringMap.h"
 $2 = HEAP32[(11148)>>2]|0; //@line 49 "./common/StringMap.h"
 $3 = HEAP32[(11152)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_7physics4Body4TypeELj4EE3addEPKcS3_(625684,$2,$3)|0); //@line 49 "./common/StringMap.h"
 $4 = HEAP32[(11156)>>2]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(11160)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_7physics4Body4TypeELj4EE3addEPKcS3_(625684,$4,$5)|0); //@line 49 "./common/StringMap.h"
 return;
}
function __ZN4love9StringMapINS_7physics4Body4TypeELj4EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0;
 var $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 7; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = $13 & 7; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = $22 & 7; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = $27 & 7; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = $32 & 7; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = $37 & 7; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $42 = (($hash$0$lcssa$i) + 6)|0; //@line 112 "./common/StringMap.h"
       $43 = $42 & 7; //@line 112 "./common/StringMap.h"
       $44 = (((($this) + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
       $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
       $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
       if ($46) {
        $$lcssa = $43;$$lcssa10 = $44;
        label = 5;
       } else {
        $47 = (($hash$0$lcssa$i) + 7)|0; //@line 112 "./common/StringMap.h"
        $48 = $47 & 7; //@line 112 "./common/StringMap.h"
        $49 = (((($this) + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
        $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
        $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
        if ($51) {
         $$lcssa = $48;$$lcssa10 = $49;
         label = 5;
        } else {
         $inserted$2$off0 = 0;
        }
       }
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(3); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 96|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love7physics4Body11getConstantEPKcRNS1_4TypeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = $9 & 7; //@line 72 "./common/StringMap.h"
  $11 = (((625684 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (625684 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(8); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((625684 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 34 "modules/physics/Body.cpp"
  return ($$$2$i|0); //@line 34 "modules/physics/Body.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 34 "modules/physics/Body.cpp"
  return ($$$2$i|0); //@line 34 "modules/physics/Body.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics4Body11getConstantENS1_4TypeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(3); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (625780 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 39 "modules/physics/Body.cpp"
}
function __ZN4love7physics5ShapeD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 30 "modules/physics/Shape.cpp"
}
function __GLOBAL__sub_I_Shape_cpp() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[(625804)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(625816)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(625828)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(625840)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(625852)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(625864)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(625876)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(625888)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(625900)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(625912)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(625916)>>2]=0|0;HEAP32[(625916)+4>>2]=0|0;HEAP32[(625916)+8>>2]=0|0;HEAP32[(625916)+12>>2]=0|0;HEAP32[(625916)+16>>2]=0|0; //@line 44 "./common/StringMap.h"
 $0 = HEAP32[2791]|0; //@line 49 "./common/StringMap.h"
 $1 = HEAP32[(11168)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_7physics5Shape4TypeELj5EE3addEPKcS3_(625796,$0,$1)|0); //@line 49 "./common/StringMap.h"
 $2 = HEAP32[(11172)>>2]|0; //@line 49 "./common/StringMap.h"
 $3 = HEAP32[(11176)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_7physics5Shape4TypeELj5EE3addEPKcS3_(625796,$2,$3)|0); //@line 49 "./common/StringMap.h"
 $4 = HEAP32[(11180)>>2]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(11184)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_7physics5Shape4TypeELj5EE3addEPKcS3_(625796,$4,$5)|0); //@line 49 "./common/StringMap.h"
 $6 = HEAP32[(11188)>>2]|0; //@line 49 "./common/StringMap.h"
 $7 = HEAP32[(11192)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_7physics5Shape4TypeELj5EE3addEPKcS3_(625796,$6,$7)|0); //@line 49 "./common/StringMap.h"
 return;
}
function __ZN4love9StringMapINS_7physics5Shape4TypeELj5EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $42 = (($hash$0$lcssa$i) + 6)|0; //@line 112 "./common/StringMap.h"
       $43 = (($42>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
       $44 = (((($this) + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
       $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
       $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
       if ($46) {
        $$lcssa = $43;$$lcssa10 = $44;
        label = 5;
       } else {
        $47 = (($hash$0$lcssa$i) + 7)|0; //@line 112 "./common/StringMap.h"
        $48 = (($47>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
        $49 = (((($this) + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
        $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
        $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
        if ($51) {
         $$lcssa = $48;$$lcssa10 = $49;
         label = 5;
        } else {
         $52 = (($hash$0$lcssa$i) + 8)|0; //@line 112 "./common/StringMap.h"
         $53 = (($52>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
         $54 = (((($this) + (($53*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
         $55 = HEAP8[$54>>0]|0; //@line 114 "./common/StringMap.h"
         $56 = ($55<<24>>24)==(0); //@line 114 "./common/StringMap.h"
         if ($56) {
          $$lcssa = $53;$$lcssa10 = $54;
          label = 5;
         } else {
          $57 = (($hash$0$lcssa$i) + 9)|0; //@line 112 "./common/StringMap.h"
          $58 = (($57>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
          $59 = (((($this) + (($58*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
          $60 = HEAP8[$59>>0]|0; //@line 114 "./common/StringMap.h"
          $61 = ($60<<24>>24)==(0); //@line 114 "./common/StringMap.h"
          if ($61) {
           $$lcssa = $58;$$lcssa10 = $59;
           label = 5;
          } else {
           $inserted$2$off0 = 0;
          }
         }
        }
       }
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(4); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 120|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love7physics5Shape11getConstantENS1_4TypeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(4); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (625916 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 39 "modules/physics/Shape.cpp"
}
function __ZN4love5sound12w_newDecoderEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 35 "modules/sound/wrap_Sound.cpp"
 $vararg_buffer = sp; //@line 35 "modules/sound/wrap_Sound.cpp"
 $0 = (__ZN4love10filesystem16luax_getfiledataEP9lua_Statei($L,1)|0); //@line 37 "modules/sound/wrap_Sound.cpp"
 $1 = (+_luaL_optnumber($L,2,16384.0)); //@line 38 "modules/sound/wrap_Sound.cpp"
 $2 = (~~(($1))); //@line 38 "modules/sound/wrap_Sound.cpp"
 $3 = HEAP32[(625652)>>2]|0; //@line 98 "./common/Module.h"
 $4 = HEAP32[$3>>2]|0; //@line 42 "modules/sound/wrap_Sound.cpp"
 $5 = ((($4)) + 24|0); //@line 42 "modules/sound/wrap_Sound.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 42 "modules/sound/wrap_Sound.cpp"
 __THREW__ = 0;
 $7 = (invoke_iiii($6|0,($3|0),($0|0),($2|0))|0); //@line 42 "modules/sound/wrap_Sound.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 do {
  if ($9) {
   $10 = ___cxa_find_matching_catch_3(8992|0)|0;
   $11 = tempRet0;
   $12 = (_llvm_eh_typeid_for((8992|0))|0); //@line 556 "./common/runtime.h"
   $13 = ($11|0)==($12|0); //@line 556 "./common/runtime.h"
   if (!($13)) {
    $$01$i = $11;$$02$i = $10;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $14 = (___cxa_begin_catch(($10|0))|0); //@line 556 "./common/runtime.h"
   $15 = HEAP32[$14>>2]|0; //@line 560 "./common/runtime.h"
   $16 = ((($15)) + 8|0); //@line 560 "./common/runtime.h"
   $17 = HEAP32[$16>>2]|0; //@line 560 "./common/runtime.h"
   $18 = (FUNCTION_TABLE_ii[$17 & 8191]($14)|0); //@line 560 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($18|0)); //@line 560 "./common/runtime.h"
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   if (!($20)) {
    ___cxa_end_catch(); //@line 561 "./common/runtime.h"
    $21 = HEAP32[$0>>2]|0; //@line 43 "modules/sound/wrap_Sound.cpp"
    $22 = ((($21)) + 12|0); //@line 43 "modules/sound/wrap_Sound.cpp"
    $23 = HEAP32[$22>>2]|0; //@line 43 "modules/sound/wrap_Sound.cpp"
    FUNCTION_TABLE_vi[$23 & 8191]($0); //@line 43 "modules/sound/wrap_Sound.cpp"
    $24 = (_lua_tolstring($L,-1,0)|0); //@line 566 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $24; //@line 566 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 566 "./common/runtime.h"
    break;
   }
   $25 = ___cxa_find_matching_catch_2()|0;
   $26 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 561 "./common/runtime.h"
   $27 = __THREW__; __THREW__ = 0;
   $28 = $27&1;
   if ($28) {
    $29 = ___cxa_find_matching_catch_3(0|0)|0;
    $30 = tempRet0;
    ___clang_call_terminate($29); //@line 561 "./common/runtime.h"
    // unreachable; //@line 561 "./common/runtime.h"
   } else {
    $$01$i = $26;$$02$i = $25;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $31 = HEAP32[$0>>2]|0; //@line 43 "modules/sound/wrap_Sound.cpp"
   $32 = ((($31)) + 12|0); //@line 43 "modules/sound/wrap_Sound.cpp"
   $33 = HEAP32[$32>>2]|0; //@line 43 "modules/sound/wrap_Sound.cpp"
   FUNCTION_TABLE_vi[$33 & 8191]($0); //@line 43 "modules/sound/wrap_Sound.cpp"
   $34 = ($7|0)==(0|0); //@line 46 "modules/sound/wrap_Sound.cpp"
   if (!($34)) {
    __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,30,$7); //@line 49 "modules/sound/wrap_Sound.cpp"
    $44 = HEAP32[$7>>2]|0; //@line 50 "modules/sound/wrap_Sound.cpp"
    $45 = ((($44)) + 12|0); //@line 50 "modules/sound/wrap_Sound.cpp"
    $46 = HEAP32[$45>>2]|0; //@line 50 "modules/sound/wrap_Sound.cpp"
    FUNCTION_TABLE_vi[$46 & 8191]($7); //@line 50 "modules/sound/wrap_Sound.cpp"
    $$0 = 1;
    STACKTOP = sp;return ($$0|0); //@line 52 "modules/sound/wrap_Sound.cpp"
   }
  }
 } while(0);
 $35 = (__ZNK4love10filesystem8FileData12getExtensionEv($0)|0); //@line 47 "modules/sound/wrap_Sound.cpp"
 $36 = HEAP8[$35>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $37 = $36 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $38 = ($37<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $39 = ((($35)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $40 = HEAP32[$39>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $41 = ((($35)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $42 = $38 ? $41 : $40; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 HEAP32[$vararg_buffer1>>2] = $42; //@line 47 "modules/sound/wrap_Sound.cpp"
 $43 = (_luaL_error($L,261558,$vararg_buffer1)|0); //@line 47 "modules/sound/wrap_Sound.cpp"
 $$0 = $43;
 STACKTOP = sp;return ($$0|0); //@line 52 "modules/sound/wrap_Sound.cpp"
}
function __ZN4love5sound14w_newSoundDataEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$01$i2 = 0, $$02$i = 0, $$02$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0.0, $60 = 0, $7 = 0, $8 = 0.0, $9 = 0, $t$2 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 54 "modules/sound/wrap_Sound.cpp"
 $vararg_buffer = sp; //@line 54 "modules/sound/wrap_Sound.cpp"
 $0 = (_lua_isnumber($L,1)|0); //@line 58 "modules/sound/wrap_Sound.cpp"
 $1 = ($0|0)==(0); //@line 58 "modules/sound/wrap_Sound.cpp"
 do {
  if ($1) {
   $32 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,1,30)|0); //@line 71 "modules/sound/wrap_Sound.cpp"
   if (!($32)) {
    (__ZN4love5sound12w_newDecoderEP9lua_State($L)|0); //@line 73 "modules/sound/wrap_Sound.cpp"
    _lua_replace($L,1); //@line 74 "modules/sound/wrap_Sound.cpp"
   }
   $33 = HEAP32[(625652)>>2]|0; //@line 98 "./common/Module.h"
   __THREW__ = 0;
   $34 = (invoke_iii(3447,($L|0),1)|0); //@line 77 "modules/sound/wrap_Sound.cpp"
   $35 = __THREW__; __THREW__ = 0;
   $36 = $35&1;
   if (!($36)) {
    __THREW__ = 0;
    $37 = (invoke_iii(3448,($33|0),($34|0))|0); //@line 77 "modules/sound/wrap_Sound.cpp"
    $38 = __THREW__; __THREW__ = 0;
    $39 = $38&1;
    if (!($39)) {
     $t$2 = $37;
     break;
    }
   }
   $40 = ___cxa_find_matching_catch_3(8992|0)|0;
   $41 = tempRet0;
   $42 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $43 = ($41|0)==($42|0); //@line 535 "./common/runtime.h"
   if (!($43)) {
    $$01$i2 = $41;$$02$i1 = $40;
    ___resumeException($$02$i1|0);
    // unreachable;
   }
   $44 = (___cxa_begin_catch(($40|0))|0); //@line 535 "./common/runtime.h"
   $45 = HEAP32[$44>>2]|0; //@line 539 "./common/runtime.h"
   $46 = ((($45)) + 8|0); //@line 539 "./common/runtime.h"
   $47 = HEAP32[$46>>2]|0; //@line 539 "./common/runtime.h"
   $48 = (FUNCTION_TABLE_ii[$47 & 8191]($44)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($48|0)); //@line 539 "./common/runtime.h"
   $49 = __THREW__; __THREW__ = 0;
   $50 = $49&1;
   if (!($50)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $51 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer1>>2] = $51; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
    $t$2 = 0;
    break;
   }
   $52 = ___cxa_find_matching_catch_2()|0;
   $53 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $54 = __THREW__; __THREW__ = 0;
   $55 = $54&1;
   if ($55) {
    $56 = ___cxa_find_matching_catch_3(0|0)|0;
    $57 = tempRet0;
    ___clang_call_terminate($56); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i2 = $53;$$02$i1 = $52;
    ___resumeException($$02$i1|0);
    // unreachable;
   }
  } else {
   $2 = (+_luaL_checknumber($L,1)); //@line 60 "modules/sound/wrap_Sound.cpp"
   $3 = (~~(($2))); //@line 60 "modules/sound/wrap_Sound.cpp"
   $4 = (+_luaL_optnumber($L,2,44100.0)); //@line 61 "modules/sound/wrap_Sound.cpp"
   $5 = (~~(($4))); //@line 61 "modules/sound/wrap_Sound.cpp"
   $6 = (+_luaL_optnumber($L,3,16.0)); //@line 62 "modules/sound/wrap_Sound.cpp"
   $7 = (~~(($6))); //@line 62 "modules/sound/wrap_Sound.cpp"
   $8 = (+_luaL_optnumber($L,4,2.0)); //@line 63 "modules/sound/wrap_Sound.cpp"
   $9 = (~~(($8))); //@line 63 "modules/sound/wrap_Sound.cpp"
   $10 = HEAP32[(625652)>>2]|0; //@line 98 "./common/Module.h"
   __THREW__ = 0;
   $11 = (invoke_iiiiii(3446,($10|0),($3|0),($5|0),($7|0),($9|0))|0); //@line 65 "modules/sound/wrap_Sound.cpp"
   $12 = __THREW__; __THREW__ = 0;
   $13 = $12&1;
   if ($13) {
    $14 = ___cxa_find_matching_catch_3(8992|0)|0;
    $15 = tempRet0;
    $16 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
    $17 = ($15|0)==($16|0); //@line 535 "./common/runtime.h"
    if (!($17)) {
     $$01$i = $15;$$02$i = $14;
     ___resumeException($$02$i|0);
     // unreachable;
    }
    $18 = (___cxa_begin_catch(($14|0))|0); //@line 535 "./common/runtime.h"
    $19 = HEAP32[$18>>2]|0; //@line 539 "./common/runtime.h"
    $20 = ((($19)) + 8|0); //@line 539 "./common/runtime.h"
    $21 = HEAP32[$20>>2]|0; //@line 539 "./common/runtime.h"
    $22 = (FUNCTION_TABLE_ii[$21 & 8191]($18)|0); //@line 539 "./common/runtime.h"
    __THREW__ = 0;
    invoke_vii(3380,($L|0),($22|0)); //@line 539 "./common/runtime.h"
    $23 = __THREW__; __THREW__ = 0;
    $24 = $23&1;
    if (!($24)) {
     ___cxa_end_catch(); //@line 540 "./common/runtime.h"
     $25 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
     HEAP32[$vararg_buffer>>2] = $25; //@line 543 "./common/runtime.h"
     (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
     $t$2 = 0;
     break;
    }
    $26 = ___cxa_find_matching_catch_2()|0;
    $27 = tempRet0;
    __THREW__ = 0;
    invoke_v(3378); //@line 540 "./common/runtime.h"
    $28 = __THREW__; __THREW__ = 0;
    $29 = $28&1;
    if ($29) {
     $30 = ___cxa_find_matching_catch_3(0|0)|0;
     $31 = tempRet0;
     ___clang_call_terminate($30); //@line 540 "./common/runtime.h"
     // unreachable; //@line 540 "./common/runtime.h"
    } else {
     $$01$i = $27;$$02$i = $26;
     ___resumeException($$02$i|0);
     // unreachable;
    }
   } else {
    $t$2 = $11;
   }
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,29,$t$2); //@line 80 "modules/sound/wrap_Sound.cpp"
 $58 = HEAP32[$t$2>>2]|0; //@line 81 "modules/sound/wrap_Sound.cpp"
 $59 = ((($58)) + 12|0); //@line 81 "modules/sound/wrap_Sound.cpp"
 $60 = HEAP32[$59>>2]|0; //@line 81 "modules/sound/wrap_Sound.cpp"
 FUNCTION_TABLE_vi[$60 & 8191]($t$2); //@line 81 "modules/sound/wrap_Sound.cpp"
 STACKTOP = sp;return 1; //@line 82 "modules/sound/wrap_Sound.cpp"
}
function _luaopen_love_sound($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0;
 var $eh$lpad$body$i$indexZ2D = 0, $instance$sroa$0$1 = 0, $vararg_buffer = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $w = sp + 4|0;
 $0 = HEAP32[(625652)>>2]|0; //@line 98 "./common/Module.h"
 $1 = $0; //@line 102 "modules/sound/wrap_Sound.cpp"
 $2 = ($0|0)==(0|0); //@line 103 "modules/sound/wrap_Sound.cpp"
 L1: do {
  if ($2) {
   __THREW__ = 0;
   $3 = (invoke_ii(3348,8)|0); //@line 105 "modules/sound/wrap_Sound.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   do {
    if ($5) {
     $11 = ___cxa_find_matching_catch_3(8992|0)|0;
     $12 = tempRet0;
     $eh$lpad$body$i$index2Z2D = $12;$eh$lpad$body$i$indexZ2D = $11;
    } else {
     __THREW__ = 0;
     invoke_vi(3449,($3|0)); //@line 105 "modules/sound/wrap_Sound.cpp"
     $6 = __THREW__; __THREW__ = 0;
     $7 = $6&1;
     if ($7) {
      $8 = ___cxa_find_matching_catch_3(8992|0)|0;
      $9 = tempRet0;
      __ZdlPv($3); //@line 105 "modules/sound/wrap_Sound.cpp"
      $eh$lpad$body$i$index2Z2D = $9;$eh$lpad$body$i$indexZ2D = $8;
      break;
     } else {
      $10 = $3; //@line 105 "modules/sound/wrap_Sound.cpp"
      $instance$sroa$0$1 = $10;
      break L1;
     }
    }
   } while(0);
   $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $14 = ($eh$lpad$body$i$index2Z2D|0)==($13|0); //@line 535 "./common/runtime.h"
   if (!($14)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $15 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
   $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
   $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
   $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $instance$sroa$0$1 = $1;
    break;
   }
   $23 = ___cxa_find_matching_catch_2()|0;
   $24 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    $27 = ___cxa_find_matching_catch_3(0|0)|0;
    $28 = tempRet0;
    ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $24;$$02$i = $23;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $29 = HEAP32[$0>>2]|0; //@line 108 "modules/sound/wrap_Sound.cpp"
   $30 = ((($29)) + 8|0); //@line 108 "modules/sound/wrap_Sound.cpp"
   $31 = HEAP32[$30>>2]|0; //@line 108 "modules/sound/wrap_Sound.cpp"
   FUNCTION_TABLE_vi[$31 & 8191]($0); //@line 108 "modules/sound/wrap_Sound.cpp"
   $instance$sroa$0$1 = $1;
  }
 } while(0);
 HEAP32[$w>>2] = $instance$sroa$0$1; //@line 112 "modules/sound/wrap_Sound.cpp"
 $32 = ((($w)) + 4|0); //@line 113 "modules/sound/wrap_Sound.cpp"
 HEAP32[$32>>2] = 327505; //@line 113 "modules/sound/wrap_Sound.cpp"
 $33 = ((($w)) + 8|0); //@line 114 "modules/sound/wrap_Sound.cpp"
 HEAP32[$33>>2] = 59; //@line 114 "modules/sound/wrap_Sound.cpp"
 $34 = ((($w)) + 12|0); //@line 115 "modules/sound/wrap_Sound.cpp"
 HEAP32[$34>>2] = 11196; //@line 115 "modules/sound/wrap_Sound.cpp"
 $35 = ((($w)) + 16|0); //@line 116 "modules/sound/wrap_Sound.cpp"
 HEAP32[$35>>2] = 11220; //@line 116 "modules/sound/wrap_Sound.cpp"
 $36 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 118 "modules/sound/wrap_Sound.cpp"
 STACKTOP = sp;return ($36|0); //@line 118 "modules/sound/wrap_Sound.cpp"
}
function __ZN4love14luax_checktypeINS_5sound9SoundDataEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love5sound23w_SoundData_getChannelsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5sound9SoundDataEEEPT_P9lua_StateiNS_4TypeE($L,1,29)|0); //@line 42 "modules/sound/wrap_SoundData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 48 "modules/sound/wrap_SoundData.cpp"
 $2 = ((($1)) + 24|0); //@line 48 "modules/sound/wrap_SoundData.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 48 "modules/sound/wrap_SoundData.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 48 "modules/sound/wrap_SoundData.cpp"
 _lua_pushinteger($L,$4); //@line 48 "modules/sound/wrap_SoundData.cpp"
 return 1; //@line 49 "modules/sound/wrap_SoundData.cpp"
}
function __ZN4love5sound23w_SoundData_getBitDepthEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5sound9SoundDataEEEPT_P9lua_StateiNS_4TypeE($L,1,29)|0); //@line 42 "modules/sound/wrap_SoundData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 55 "modules/sound/wrap_SoundData.cpp"
 $2 = ((($1)) + 28|0); //@line 55 "modules/sound/wrap_SoundData.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 55 "modules/sound/wrap_SoundData.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 55 "modules/sound/wrap_SoundData.cpp"
 _lua_pushinteger($L,$4); //@line 55 "modules/sound/wrap_SoundData.cpp"
 return 1; //@line 56 "modules/sound/wrap_SoundData.cpp"
}
function __ZN4love5sound25w_SoundData_getSampleRateEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5sound9SoundDataEEEPT_P9lua_StateiNS_4TypeE($L,1,29)|0); //@line 42 "modules/sound/wrap_SoundData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 62 "modules/sound/wrap_SoundData.cpp"
 $2 = ((($1)) + 32|0); //@line 62 "modules/sound/wrap_SoundData.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 62 "modules/sound/wrap_SoundData.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 62 "modules/sound/wrap_SoundData.cpp"
 _lua_pushinteger($L,$4); //@line 62 "modules/sound/wrap_SoundData.cpp"
 return 1; //@line 63 "modules/sound/wrap_SoundData.cpp"
}
function __ZN4love5sound26w_SoundData_getSampleCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5sound9SoundDataEEEPT_P9lua_StateiNS_4TypeE($L,1,29)|0); //@line 42 "modules/sound/wrap_SoundData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 69 "modules/sound/wrap_SoundData.cpp"
 $2 = ((($1)) + 36|0); //@line 69 "modules/sound/wrap_SoundData.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 69 "modules/sound/wrap_SoundData.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 69 "modules/sound/wrap_SoundData.cpp"
 _lua_pushinteger($L,$4); //@line 69 "modules/sound/wrap_SoundData.cpp"
 return 1; //@line 70 "modules/sound/wrap_SoundData.cpp"
}
function __ZN4love5sound23w_SoundData_getDurationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5sound9SoundDataEEEPT_P9lua_StateiNS_4TypeE($L,1,29)|0); //@line 42 "modules/sound/wrap_SoundData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 76 "modules/sound/wrap_SoundData.cpp"
 $2 = ((($1)) + 40|0); //@line 76 "modules/sound/wrap_SoundData.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 76 "modules/sound/wrap_SoundData.cpp"
 $4 = (+FUNCTION_TABLE_di[$3 & 4095]($0)); //@line 76 "modules/sound/wrap_SoundData.cpp"
 $5 = $4; //@line 76 "modules/sound/wrap_SoundData.cpp"
 _lua_pushnumber($L,$5); //@line 76 "modules/sound/wrap_SoundData.cpp"
 return 1; //@line 77 "modules/sound/wrap_SoundData.cpp"
}
function __ZN4love5sound21w_SoundData_setSampleEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0.0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 80 "modules/sound/wrap_SoundData.cpp"
 $0 = (__ZN4love14luax_checktypeINS_5sound9SoundDataEEEPT_P9lua_StateiNS_4TypeE($L,1,29)|0); //@line 42 "modules/sound/wrap_SoundData.cpp"
 $1 = (_luaL_checkinteger($L,2)|0); //@line 83 "modules/sound/wrap_SoundData.cpp"
 $2 = (+_luaL_checknumber($L,3)); //@line 84 "modules/sound/wrap_SoundData.cpp"
 $3 = $2; //@line 84 "modules/sound/wrap_SoundData.cpp"
 __THREW__ = 0;
 invoke_viid(3450,($0|0),($1|0),(+$3)); //@line 86 "modules/sound/wrap_SoundData.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  STACKTOP = sp;return 0; //@line 87 "modules/sound/wrap_SoundData.cpp"
 }
 $6 = ___cxa_find_matching_catch_3(8992|0)|0;
 $7 = tempRet0;
 $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
 if (!($9)) {
  $$01$i = $7;$$02$i = $6;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $17; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 87 "modules/sound/wrap_SoundData.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5sound21w_SoundData_getSampleEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 90 "modules/sound/wrap_SoundData.cpp"
 $0 = (__ZN4love14luax_checktypeINS_5sound9SoundDataEEEPT_P9lua_StateiNS_4TypeE($L,1,29)|0); //@line 42 "modules/sound/wrap_SoundData.cpp"
 $1 = (_luaL_checkinteger($L,2)|0); //@line 93 "modules/sound/wrap_SoundData.cpp"
 __THREW__ = 0;
 $2 = (+invoke_dii(3451,($0|0),($1|0))); //@line 95 "modules/sound/wrap_SoundData.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  $5 = $2; //@line 95 "modules/sound/wrap_SoundData.cpp"
  __THREW__ = 0;
  invoke_vid(3394,($L|0),(+$5)); //@line 95 "modules/sound/wrap_SoundData.cpp"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if (!($7)) {
   STACKTOP = sp;return 1; //@line 96 "modules/sound/wrap_SoundData.cpp"
  }
 }
 $8 = ___cxa_find_matching_catch_3(8992|0)|0;
 $9 = tempRet0;
 $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $11 = ($9|0)==($10|0); //@line 535 "./common/runtime.h"
 if (!($11)) {
  $$01$i = $9;$$02$i = $8;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $12 = (___cxa_begin_catch(($8|0))|0); //@line 535 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = ((($13)) + 8|0); //@line 539 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($16|0)); //@line 539 "./common/runtime.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if (!($18)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $19 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $19; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 1; //@line 96 "modules/sound/wrap_SoundData.cpp"
 }
 $20 = ___cxa_find_matching_catch_2()|0;
 $21 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if ($23) {
  $24 = ___cxa_find_matching_catch_3(0|0)|0;
  $25 = tempRet0;
  ___clang_call_terminate($24); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $21;$$02$i = $20;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function _luaopen_sounddata($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 111 "modules/sound/wrap_SoundData.cpp"
 HEAP32[$vararg_buffer>>2] = 12272; //@line 113 "modules/sound/wrap_SoundData.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 113 "modules/sound/wrap_SoundData.cpp"
 HEAP32[$vararg_ptr1>>2] = 11232; //@line 113 "modules/sound/wrap_SoundData.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 113 "modules/sound/wrap_SoundData.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 113 "modules/sound/wrap_SoundData.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,29,261588,$vararg_buffer)|0); //@line 113 "modules/sound/wrap_SoundData.cpp"
 __ZN4love21luax_gettypemetatableEP9lua_StateNS_4TypeE($L,29); //@line 115 "modules/sound/wrap_SoundData.cpp"
 $1 = (_lua_type($L,-1)|0); //@line 118 "modules/sound/wrap_SoundData.cpp"
 $2 = ($1|0)==(5); //@line 118 "modules/sound/wrap_SoundData.cpp"
 if (!($2)) {
  _lua_settop($L,-2); //@line 126 "modules/sound/wrap_SoundData.cpp"
  STACKTOP = sp;return ($0|0); //@line 128 "modules/sound/wrap_SoundData.cpp"
 }
 (_luaL_loadbuffer($L,261633,4222,265855)|0); //@line 120 "modules/sound/wrap_SoundData.cpp"
 _lua_pushvalue($L,-2); //@line 121 "modules/sound/wrap_SoundData.cpp"
 _lua_call($L,1,0); //@line 122 "modules/sound/wrap_SoundData.cpp"
 _lua_settop($L,-2); //@line 126 "modules/sound/wrap_SoundData.cpp"
 STACKTOP = sp;return ($0|0); //@line 128 "modules/sound/wrap_SoundData.cpp"
}
function __ZN4love6system7w_getOSEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = HEAP32[(625656)>>2]|0; //@line 98 "./common/Module.h"
 __ZNK4love6system6System5getOSEv($0,$1); //@line 34 "modules/system/wrap_System.cpp"
 __THREW__ = 0;
 invoke_vii(3405,($L|0),($0|0)); //@line 34 "modules/system/wrap_System.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 34 "modules/system/wrap_System.cpp"
  ___resumeException($4|0);
  // unreachable;
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 34 "modules/system/wrap_System.cpp"
  STACKTOP = sp;return 1; //@line 35 "modules/system/wrap_System.cpp"
 }
 return (0)|0;
}
function __ZN4love6system19w_getProcessorCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625656)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 40 "modules/system/wrap_System.cpp"
 $2 = ((($1)) + 24|0); //@line 40 "modules/system/wrap_System.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 40 "modules/system/wrap_System.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 40 "modules/system/wrap_System.cpp"
 _lua_pushinteger($L,$4); //@line 40 "modules/system/wrap_System.cpp"
 return 1; //@line 41 "modules/system/wrap_System.cpp"
}
function __ZN4love6system18w_setClipboardTextEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (_luaL_checklstring($L,1,0)|0); //@line 46 "modules/system/wrap_System.cpp"
 $2 = HEAP32[(625656)>>2]|0; //@line 98 "./common/Module.h"
 $3 = HEAP32[$2>>2]|0; //@line 47 "modules/system/wrap_System.cpp"
 $4 = ((($3)) + 28|0); //@line 47 "modules/system/wrap_System.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 47 "modules/system/wrap_System.cpp"
 $6 = (_strlen($1)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($0,$1,$6); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_vii($5|0,($2|0),($0|0)); //@line 47 "modules/system/wrap_System.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 47 "modules/system/wrap_System.cpp"
  ___resumeException($9|0);
  // unreachable;
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 47 "modules/system/wrap_System.cpp"
  STACKTOP = sp;return 0; //@line 48 "modules/system/wrap_System.cpp"
 }
 return (0)|0;
}
function __ZN4love6system18w_getClipboardTextEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = HEAP32[(625656)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 53 "modules/system/wrap_System.cpp"
 $3 = ((($2)) + 32|0); //@line 53 "modules/system/wrap_System.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 53 "modules/system/wrap_System.cpp"
 FUNCTION_TABLE_vii[$4 & 8191]($0,$1); //@line 53 "modules/system/wrap_System.cpp"
 __THREW__ = 0;
 invoke_vii(3405,($L|0),($0|0)); //@line 53 "modules/system/wrap_System.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 53 "modules/system/wrap_System.cpp"
  ___resumeException($7|0);
  // unreachable;
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 53 "modules/system/wrap_System.cpp"
  STACKTOP = sp;return 1; //@line 54 "modules/system/wrap_System.cpp"
 }
 return (0)|0;
}
function __ZN4love6system14w_getPowerInfoEP9lua_State($L) {
 $L = $L|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $percent = 0, $seconds = 0, $str = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $seconds = sp + 8|0;
 $percent = sp + 4|0;
 $str = sp;
 HEAP32[$seconds>>2] = -1; //@line 59 "modules/system/wrap_System.cpp"
 HEAP32[$percent>>2] = -1; //@line 59 "modules/system/wrap_System.cpp"
 $0 = HEAP32[(625656)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 62 "modules/system/wrap_System.cpp"
 $2 = ((($1)) + 36|0); //@line 62 "modules/system/wrap_System.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 62 "modules/system/wrap_System.cpp"
 $4 = (FUNCTION_TABLE_iiii[$3 & 8191]($0,$seconds,$percent)|0); //@line 62 "modules/system/wrap_System.cpp"
 $5 = (__ZN4love6system6System11getConstantENS1_10PowerStateERPKc($4,$str)|0); //@line 64 "modules/system/wrap_System.cpp"
 if ($5) {
  $$pre = HEAP32[$str>>2]|0; //@line 67 "modules/system/wrap_System.cpp"
  $6 = $$pre;
 } else {
  HEAP32[$str>>2] = 442424; //@line 65 "modules/system/wrap_System.cpp"
  $6 = 442424;
 }
 _lua_pushstring($L,$6); //@line 67 "modules/system/wrap_System.cpp"
 $7 = HEAP32[$percent>>2]|0; //@line 69 "modules/system/wrap_System.cpp"
 $8 = ($7|0)>(-1); //@line 69 "modules/system/wrap_System.cpp"
 if ($8) {
  _lua_pushinteger($L,$7); //@line 70 "modules/system/wrap_System.cpp"
 } else {
  _lua_pushnil($L); //@line 72 "modules/system/wrap_System.cpp"
 }
 $9 = HEAP32[$seconds>>2]|0; //@line 74 "modules/system/wrap_System.cpp"
 $10 = ($9|0)>(-1); //@line 74 "modules/system/wrap_System.cpp"
 if ($10) {
  _lua_pushinteger($L,$9); //@line 75 "modules/system/wrap_System.cpp"
  STACKTOP = sp;return 3; //@line 79 "modules/system/wrap_System.cpp"
 } else {
  _lua_pushnil($L); //@line 77 "modules/system/wrap_System.cpp"
  STACKTOP = sp;return 3; //@line 79 "modules/system/wrap_System.cpp"
 }
 return (0)|0;
}
function __ZN4love6system9w_openURLEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $url = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $url = sp;
 __ZN4love16luax_checkstringEP9lua_Statei($url,$L,1); //@line 84 "modules/system/wrap_System.cpp"
 $0 = HEAP32[(625656)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 85 "modules/system/wrap_System.cpp"
 $2 = ((($1)) + 40|0); //@line 85 "modules/system/wrap_System.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 85 "modules/system/wrap_System.cpp"
 __THREW__ = 0;
 $4 = (invoke_iii($3|0,($0|0),($url|0))|0); //@line 85 "modules/system/wrap_System.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  __THREW__ = 0;
  invoke_vii(3397,($L|0),($4|0)); //@line 85 "modules/system/wrap_System.cpp"
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if (!($8)) {
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($url); //@line 87 "modules/system/wrap_System.cpp"
   STACKTOP = sp;return 1; //@line 87 "modules/system/wrap_System.cpp"
  }
 }
 $9 = ___cxa_find_matching_catch_2()|0;
 $10 = tempRet0;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($url); //@line 87 "modules/system/wrap_System.cpp"
 ___resumeException($9|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love6system9w_vibrateEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_optnumber($L,1,0.5)); //@line 91 "modules/system/wrap_System.cpp"
 $1 = HEAP32[(625656)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 92 "modules/system/wrap_System.cpp"
 $3 = ((($2)) + 44|0); //@line 92 "modules/system/wrap_System.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 92 "modules/system/wrap_System.cpp"
 FUNCTION_TABLE_vid[$4 & 8191]($1,$0); //@line 92 "modules/system/wrap_System.cpp"
 return 0; //@line 93 "modules/system/wrap_System.cpp"
}
function _luaopen_love_system($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $instance$0 = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $w = sp;
 $0 = HEAP32[(625656)>>2]|0; //@line 98 "./common/Module.h"
 $1 = ($0|0)==(0|0); //@line 111 "modules/system/wrap_System.cpp"
 do {
  if ($1) {
   $2 = (__Znwj(8)|0); //@line 113 "modules/system/wrap_System.cpp"
   __THREW__ = 0;
   invoke_vi(3452,($2|0)); //@line 113 "modules/system/wrap_System.cpp"
   $3 = __THREW__; __THREW__ = 0;
   $4 = $3&1;
   if ($4) {
    $5 = ___cxa_find_matching_catch_2()|0;
    $6 = tempRet0;
    __ZdlPv($2); //@line 113 "modules/system/wrap_System.cpp"
    ___resumeException($5|0);
    // unreachable;
   } else {
    $instance$0 = $2;
    break;
   }
  } else {
   $7 = HEAP32[$0>>2]|0; //@line 116 "modules/system/wrap_System.cpp"
   $8 = ((($7)) + 8|0); //@line 116 "modules/system/wrap_System.cpp"
   $9 = HEAP32[$8>>2]|0; //@line 116 "modules/system/wrap_System.cpp"
   FUNCTION_TABLE_vi[$9 & 8191]($0); //@line 116 "modules/system/wrap_System.cpp"
   $instance$0 = $0;
  }
 } while(0);
 HEAP32[$w>>2] = $instance$0; //@line 119 "modules/system/wrap_System.cpp"
 $10 = ((($w)) + 4|0); //@line 120 "modules/system/wrap_System.cpp"
 HEAP32[$10>>2] = 356113; //@line 120 "modules/system/wrap_System.cpp"
 $11 = ((($w)) + 8|0); //@line 121 "modules/system/wrap_System.cpp"
 HEAP32[$11>>2] = 3; //@line 121 "modules/system/wrap_System.cpp"
 $12 = ((($w)) + 12|0); //@line 122 "modules/system/wrap_System.cpp"
 HEAP32[$12>>2] = 11296; //@line 122 "modules/system/wrap_System.cpp"
 $13 = ((($w)) + 16|0); //@line 123 "modules/system/wrap_System.cpp"
 HEAP32[$13>>2] = 0; //@line 123 "modules/system/wrap_System.cpp"
 $14 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 125 "modules/system/wrap_System.cpp"
 STACKTOP = sp;return ($14|0); //@line 125 "modules/system/wrap_System.cpp"
}
function __ZN4love6thread11w_newThreadEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $data$0 = 0, $idxs = 0;
 var $name = 0, $slen = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 36|0;
 $slen = sp + 32|0;
 $idxs = sp + 24|0;
 $0 = sp + 12|0;
 $1 = sp;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($name,265956,11); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 $2 = (invoke_iii(3453,($L|0),1)|0); //@line 45 "modules/thread/wrap_ThreadModule.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 L1: do {
  if (!($4)) {
   $5 = ($2|0)==(0); //@line 45 "modules/thread/wrap_ThreadModule.cpp"
   if ($5) {
    __THREW__ = 0;
    $35 = (invoke_iiii(3458,($L|0),1,5)|0); //@line 64 "modules/thread/wrap_ThreadModule.cpp"
    $36 = __THREW__; __THREW__ = 0;
    $37 = $36&1;
    if ($37) {
     break;
    }
    if ($35) {
     __THREW__ = 0;
     (invoke_iiiii(3457,($L|0),1,(354839|0),(328985|0))|0); //@line 65 "modules/thread/wrap_ThreadModule.cpp"
     $38 = __THREW__; __THREW__ = 0;
     $39 = $38&1;
     if ($39) {
      break;
     }
    }
   } else {
    HEAP32[$slen>>2] = 0; //@line 47 "modules/thread/wrap_ThreadModule.cpp"
    __THREW__ = 0;
    $6 = (invoke_iiii(3392,($L|0),1,($slen|0))|0); //@line 48 "modules/thread/wrap_ThreadModule.cpp"
    $7 = __THREW__; __THREW__ = 0;
    $8 = $7&1;
    if ($8) {
     break;
    }
    $9 = HEAP32[$slen>>2]|0; //@line 51 "modules/thread/wrap_ThreadModule.cpp"
    $10 = ($9>>>0)>(1023); //@line 51 "modules/thread/wrap_ThreadModule.cpp"
    if ($10) {
     label = 6;
    } else {
     $11 = (_memchr($6,10,$9)|0); //@line 51 "modules/thread/wrap_ThreadModule.cpp"
     $12 = ($11|0)==(0|0); //@line 51 "modules/thread/wrap_ThreadModule.cpp"
     if ($12) {
      __THREW__ = 0;
      (invoke_iiiii(3457,($L|0),1,(354839|0),(328985|0))|0); //@line 62 "modules/thread/wrap_ThreadModule.cpp"
      $33 = __THREW__; __THREW__ = 0;
      $34 = $33&1;
      if ($34) {
       break;
      }
     } else {
      label = 6;
     }
    }
    if ((label|0) == 6) {
     __THREW__ = 0;
     invoke_vii(3454,($L|0),1); //@line 54 "modules/thread/wrap_ThreadModule.cpp"
     $13 = __THREW__; __THREW__ = 0;
     $14 = $13&1;
     if ($14) {
      break;
     }
     __THREW__ = 0;
     invoke_vii(3380,($L|0),(522489|0)); //@line 55 "modules/thread/wrap_ThreadModule.cpp"
     $15 = __THREW__; __THREW__ = 0;
     $16 = $15&1;
     if ($16) {
      break;
     }
     __THREW__ = 0;
     $17 = (invoke_ii(3385,($L|0))|0); //@line 56 "modules/thread/wrap_ThreadModule.cpp"
     $18 = __THREW__; __THREW__ = 0;
     $19 = $18&1;
     if ($19) {
      break;
     }
     $20 = (($17) + -1)|0; //@line 56 "modules/thread/wrap_ThreadModule.cpp"
     HEAP32[$idxs>>2] = $20; //@line 56 "modules/thread/wrap_ThreadModule.cpp"
     __THREW__ = 0;
     $21 = (invoke_ii(3385,($L|0))|0); //@line 56 "modules/thread/wrap_ThreadModule.cpp"
     $22 = __THREW__; __THREW__ = 0;
     $23 = $22&1;
     if ($23) {
      break;
     }
     $24 = ((($idxs)) + 4|0); //@line 56 "modules/thread/wrap_ThreadModule.cpp"
     HEAP32[$24>>2] = $21; //@line 56 "modules/thread/wrap_ThreadModule.cpp"
     __THREW__ = 0;
     (invoke_iiiiii(3455,($L|0),($idxs|0),2,(354839|0),(328985|0))|0); //@line 57 "modules/thread/wrap_ThreadModule.cpp"
     $25 = __THREW__; __THREW__ = 0;
     $26 = $25&1;
     if ($26) {
      break;
     }
     __THREW__ = 0;
     invoke_vii(3384,($L|0),-2); //@line 58 "modules/thread/wrap_ThreadModule.cpp"
     $27 = __THREW__; __THREW__ = 0;
     $28 = $27&1;
     if ($28) {
      break;
     }
     __THREW__ = 0;
     invoke_vii(3456,($L|0),1); //@line 59 "modules/thread/wrap_ThreadModule.cpp"
     $29 = __THREW__; __THREW__ = 0;
     $30 = $29&1;
     if ($30) {
      break;
     }
    }
   }
   __THREW__ = 0;
   $40 = (invoke_iiii(3458,($L|0),1,7)|0); //@line 67 "modules/thread/wrap_ThreadModule.cpp"
   $41 = __THREW__; __THREW__ = 0;
   $42 = $41&1;
   if (!($42)) {
    do {
     if ($40) {
      __THREW__ = 0;
      $43 = (invoke_iiii(3459,($L|0),1,7)|0); //@line 69 "modules/thread/wrap_ThreadModule.cpp"
      $44 = __THREW__; __THREW__ = 0;
      $45 = $44&1;
      if ($45) {
       break L1;
      }
      __THREW__ = 0;
      invoke_viii(3460,($1|0),(442495|0),1); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $46 = __THREW__; __THREW__ = 0;
      $47 = $46&1;
      if ($47) {
       break L1;
      }
      __THREW__ = 0;
      $48 = (invoke_ii(3461,($43|0))|0); //@line 70 "modules/thread/wrap_ThreadModule.cpp"
      $49 = __THREW__; __THREW__ = 0;
      $50 = $49&1;
      if (!($50)) {
       $51 = HEAP8[$48>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $52 = $51 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $53 = ($52<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $54 = ((($48)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $55 = HEAP32[$54>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $56 = ((($48)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $57 = $53 ? $56 : $55; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $58 = ((($48)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $59 = HEAP32[$58>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $60 = $51&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $61 = $60 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $62 = $53 ? $61 : $59; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       __THREW__ = 0;
       $63 = (invoke_iiii(3462,($1|0),($57|0),($62|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $64 = __THREW__; __THREW__ = 0;
       $65 = $64&1;
       if (!($65)) {
        ;HEAP32[$0>>2]=HEAP32[$63>>2]|0;HEAP32[$0+4>>2]=HEAP32[$63+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$63+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        ;HEAP32[$63>>2]=0|0;HEAP32[$63+4>>2]=0|0;HEAP32[$63+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $66 = HEAP8[$name>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $67 = $66 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $68 = ($67<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if ($68) {
         $72 = ((($name)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         HEAP8[$72>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         HEAP8[$name>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        } else {
         $69 = ((($name)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $70 = HEAP32[$69>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         HEAP8[$70>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $71 = ((($name)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         HEAP32[$71>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        }
        __THREW__ = 0;
        invoke_vii(3368,($name|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $73 = __THREW__; __THREW__ = 0;
        $74 = $73&1;
        if ($74) {
         $75 = ___cxa_find_matching_catch_3(0|0)|0;
         $76 = tempRet0;
         ___clang_call_terminate($75); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        } else {
         ;HEAP32[$name>>2]=HEAP32[$0>>2]|0;HEAP32[$name+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$name+8>>2]=HEAP32[$0+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 70 "modules/thread/wrap_ThreadModule.cpp"
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 70 "modules/thread/wrap_ThreadModule.cpp"
         $data$0 = $43;
         break;
        }
       }
      }
      $77 = ___cxa_find_matching_catch_2()|0;
      $78 = tempRet0;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 70 "modules/thread/wrap_ThreadModule.cpp"
      $$0 = $77;$$01 = $78;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($name); //@line 82 "modules/thread/wrap_ThreadModule.cpp"
      ___resumeException($$0|0);
      // unreachable;
     } else {
      __THREW__ = 0;
      $79 = (invoke_iiii(3463,($L|0),1,2)|0); //@line 75 "modules/thread/wrap_ThreadModule.cpp"
      $80 = __THREW__; __THREW__ = 0;
      $81 = $80&1;
      if ($81) {
       break L1;
      } else {
       $data$0 = $79;
      }
     }
    } while(0);
    $82 = HEAP32[(625660)>>2]|0; //@line 98 "./common/Module.h"
    $83 = HEAP32[$82>>2]|0; //@line 78 "modules/thread/wrap_ThreadModule.cpp"
    $84 = ((($83)) + 24|0); //@line 78 "modules/thread/wrap_ThreadModule.cpp"
    $85 = HEAP32[$84>>2]|0; //@line 78 "modules/thread/wrap_ThreadModule.cpp"
    __THREW__ = 0;
    $86 = (invoke_iiii($85|0,($82|0),($name|0),($data$0|0))|0); //@line 78 "modules/thread/wrap_ThreadModule.cpp"
    $87 = __THREW__; __THREW__ = 0;
    $88 = $87&1;
    if (!($88)) {
     __THREW__ = 0;
     invoke_viii(3387,($L|0),53,($86|0)); //@line 79 "modules/thread/wrap_ThreadModule.cpp"
     $89 = __THREW__; __THREW__ = 0;
     $90 = $89&1;
     if (!($90)) {
      $91 = HEAP32[$86>>2]|0; //@line 80 "modules/thread/wrap_ThreadModule.cpp"
      $92 = ((($91)) + 12|0); //@line 80 "modules/thread/wrap_ThreadModule.cpp"
      $93 = HEAP32[$92>>2]|0; //@line 80 "modules/thread/wrap_ThreadModule.cpp"
      __THREW__ = 0;
      invoke_vi($93|0,($86|0)); //@line 80 "modules/thread/wrap_ThreadModule.cpp"
      $94 = __THREW__; __THREW__ = 0;
      $95 = $94&1;
      if (!($95)) {
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($name); //@line 82 "modules/thread/wrap_ThreadModule.cpp"
       STACKTOP = sp;return 1; //@line 82 "modules/thread/wrap_ThreadModule.cpp"
      }
     }
    }
   }
  }
 } while(0);
 $31 = ___cxa_find_matching_catch_2()|0;
 $32 = tempRet0;
 $$0 = $31;$$01 = $32;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($name); //@line 82 "modules/thread/wrap_ThreadModule.cpp"
 ___resumeException($$0|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love14luax_checktypeINS_10filesystem8FileDataEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love6thread12w_newChannelEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625660)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 86 "modules/thread/wrap_ThreadModule.cpp"
 $2 = ((($1)) + 28|0); //@line 86 "modules/thread/wrap_ThreadModule.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 86 "modules/thread/wrap_ThreadModule.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 86 "modules/thread/wrap_ThreadModule.cpp"
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,54,$4); //@line 87 "modules/thread/wrap_ThreadModule.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 88 "modules/thread/wrap_ThreadModule.cpp"
 $6 = ((($5)) + 12|0); //@line 88 "modules/thread/wrap_ThreadModule.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 88 "modules/thread/wrap_ThreadModule.cpp"
 FUNCTION_TABLE_vi[$7 & 8191]($4); //@line 88 "modules/thread/wrap_ThreadModule.cpp"
 return 1; //@line 89 "modules/thread/wrap_ThreadModule.cpp"
}
function __ZN4love6thread12w_getChannelEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp;
 __ZN4love16luax_checkstringEP9lua_Statei($name,$L,1); //@line 94 "modules/thread/wrap_ThreadModule.cpp"
 $0 = HEAP32[(625660)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 95 "modules/thread/wrap_ThreadModule.cpp"
 $2 = ((($1)) + 32|0); //@line 95 "modules/thread/wrap_ThreadModule.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 95 "modules/thread/wrap_ThreadModule.cpp"
 __THREW__ = 0;
 $4 = (invoke_iii($3|0,($0|0),($name|0))|0); //@line 95 "modules/thread/wrap_ThreadModule.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  __THREW__ = 0;
  invoke_viii(3387,($L|0),54,($4|0)); //@line 96 "modules/thread/wrap_ThreadModule.cpp"
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if (!($8)) {
   $9 = HEAP32[$4>>2]|0; //@line 97 "modules/thread/wrap_ThreadModule.cpp"
   $10 = ((($9)) + 12|0); //@line 97 "modules/thread/wrap_ThreadModule.cpp"
   $11 = HEAP32[$10>>2]|0; //@line 97 "modules/thread/wrap_ThreadModule.cpp"
   __THREW__ = 0;
   invoke_vi($11|0,($4|0)); //@line 97 "modules/thread/wrap_ThreadModule.cpp"
   $12 = __THREW__; __THREW__ = 0;
   $13 = $12&1;
   if (!($13)) {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($name); //@line 99 "modules/thread/wrap_ThreadModule.cpp"
    STACKTOP = sp;return 1; //@line 99 "modules/thread/wrap_ThreadModule.cpp"
   }
  }
 }
 $14 = ___cxa_find_matching_catch_2()|0;
 $15 = tempRet0;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($name); //@line 99 "modules/thread/wrap_ThreadModule.cpp"
 ___resumeException($14|0);
 // unreachable;
 return (0)|0;
}
function _luaopen_love_thread($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $instance$sroa$0$1 = 0, $vararg_buffer = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $w = sp + 4|0;
 $0 = HEAP32[(625660)>>2]|0; //@line 98 "./common/Module.h"
 $1 = $0; //@line 118 "modules/thread/wrap_ThreadModule.cpp"
 $2 = ($0|0)==(0|0); //@line 119 "modules/thread/wrap_ThreadModule.cpp"
 L1: do {
  if ($2) {
   __THREW__ = 0;
   $3 = (invoke_ii(3348,8)|0); //@line 121 "modules/thread/wrap_ThreadModule.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   do {
    if ($5) {
     $15 = ___cxa_find_matching_catch_3(8992|0)|0;
     $16 = tempRet0;
     $eh$lpad$body$i$index2Z2D = $16;$eh$lpad$body$i$indexZ2D = $15;
    } else {
     $6 = $3; //@line 121 "modules/thread/wrap_ThreadModule.cpp"
     $7 = $6; //@line 121 "modules/thread/wrap_ThreadModule.cpp"
     HEAP32[$7>>2] = 0; //@line 121 "modules/thread/wrap_ThreadModule.cpp"
     $8 = (($6) + 4)|0; //@line 121 "modules/thread/wrap_ThreadModule.cpp"
     $9 = $8; //@line 121 "modules/thread/wrap_ThreadModule.cpp"
     HEAP32[$9>>2] = 0; //@line 121 "modules/thread/wrap_ThreadModule.cpp"
     __THREW__ = 0;
     invoke_vi(3431,($3|0)); //@line 33 "./common/Module.h"
     $10 = __THREW__; __THREW__ = 0;
     $11 = $10&1;
     if ($11) {
      $12 = ___cxa_find_matching_catch_3(8992|0)|0;
      $13 = tempRet0;
      __ZdlPv($3); //@line 121 "modules/thread/wrap_ThreadModule.cpp"
      $eh$lpad$body$i$index2Z2D = $13;$eh$lpad$body$i$indexZ2D = $12;
      break;
     } else {
      HEAP32[$3>>2] = (19380); //@line 41 "modules/thread/ThreadModule.h"
      $14 = $3; //@line 121 "modules/thread/wrap_ThreadModule.cpp"
      $instance$sroa$0$1 = $14;
      break L1;
     }
    }
   } while(0);
   $17 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $18 = ($eh$lpad$body$i$index2Z2D|0)==($17|0); //@line 535 "./common/runtime.h"
   if (!($18)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $19 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $20 = HEAP32[$19>>2]|0; //@line 539 "./common/runtime.h"
   $21 = ((($20)) + 8|0); //@line 539 "./common/runtime.h"
   $22 = HEAP32[$21>>2]|0; //@line 539 "./common/runtime.h"
   $23 = (FUNCTION_TABLE_ii[$22 & 8191]($19)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($23|0)); //@line 539 "./common/runtime.h"
   $24 = __THREW__; __THREW__ = 0;
   $25 = $24&1;
   if (!($25)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $26 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $26; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $instance$sroa$0$1 = $1;
    break;
   }
   $27 = ___cxa_find_matching_catch_2()|0;
   $28 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $29 = __THREW__; __THREW__ = 0;
   $30 = $29&1;
   if ($30) {
    $31 = ___cxa_find_matching_catch_3(0|0)|0;
    $32 = tempRet0;
    ___clang_call_terminate($31); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $28;$$02$i = $27;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $33 = HEAP32[$0>>2]|0; //@line 124 "modules/thread/wrap_ThreadModule.cpp"
   $34 = ((($33)) + 8|0); //@line 124 "modules/thread/wrap_ThreadModule.cpp"
   $35 = HEAP32[$34>>2]|0; //@line 124 "modules/thread/wrap_ThreadModule.cpp"
   FUNCTION_TABLE_vi[$35 & 8191]($0); //@line 124 "modules/thread/wrap_ThreadModule.cpp"
   $instance$sroa$0$1 = $1;
  }
 } while(0);
 HEAP32[$w>>2] = $instance$sroa$0$1; //@line 127 "modules/thread/wrap_ThreadModule.cpp"
 $36 = ((($w)) + 4|0); //@line 128 "modules/thread/wrap_ThreadModule.cpp"
 HEAP32[$36>>2] = 518991; //@line 128 "modules/thread/wrap_ThreadModule.cpp"
 $37 = ((($w)) + 8|0); //@line 129 "modules/thread/wrap_ThreadModule.cpp"
 HEAP32[$37>>2] = 3; //@line 129 "modules/thread/wrap_ThreadModule.cpp"
 $38 = ((($w)) + 12|0); //@line 130 "modules/thread/wrap_ThreadModule.cpp"
 HEAP32[$38>>2] = 11360; //@line 130 "modules/thread/wrap_ThreadModule.cpp"
 $39 = ((($w)) + 16|0); //@line 131 "modules/thread/wrap_ThreadModule.cpp"
 HEAP32[$39>>2] = 11392; //@line 131 "modules/thread/wrap_ThreadModule.cpp"
 $40 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 133 "modules/thread/wrap_ThreadModule.cpp"
 STACKTOP = sp;return ($40|0); //@line 133 "modules/thread/wrap_ThreadModule.cpp"
}
function __ZN4love5timer6w_stepEP9lua_State($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[(625664)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 37 "modules/timer/wrap_Timer.cpp"
 $3 = ((($2)) + 24|0); //@line 37 "modules/timer/wrap_Timer.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 37 "modules/timer/wrap_Timer.cpp"
 FUNCTION_TABLE_vi[$4 & 8191]($1); //@line 37 "modules/timer/wrap_Timer.cpp"
 return 0; //@line 38 "modules/timer/wrap_Timer.cpp"
}
function __ZN4love5timer10w_getDeltaEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625664)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 43 "modules/timer/wrap_Timer.cpp"
 $2 = ((($1)) + 32|0); //@line 43 "modules/timer/wrap_Timer.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 43 "modules/timer/wrap_Timer.cpp"
 $4 = (+FUNCTION_TABLE_di[$3 & 4095]($0)); //@line 43 "modules/timer/wrap_Timer.cpp"
 _lua_pushnumber($L,$4); //@line 43 "modules/timer/wrap_Timer.cpp"
 return 1; //@line 44 "modules/timer/wrap_Timer.cpp"
}
function __ZN4love5timer8w_getFPSEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625664)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 49 "modules/timer/wrap_Timer.cpp"
 $2 = ((($1)) + 36|0); //@line 49 "modules/timer/wrap_Timer.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 49 "modules/timer/wrap_Timer.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 49 "modules/timer/wrap_Timer.cpp"
 _lua_pushinteger($L,$4); //@line 49 "modules/timer/wrap_Timer.cpp"
 return 1; //@line 50 "modules/timer/wrap_Timer.cpp"
}
function __ZN4love5timer17w_getAverageDeltaEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625664)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 55 "modules/timer/wrap_Timer.cpp"
 $2 = ((($1)) + 40|0); //@line 55 "modules/timer/wrap_Timer.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 55 "modules/timer/wrap_Timer.cpp"
 $4 = (+FUNCTION_TABLE_di[$3 & 4095]($0)); //@line 55 "modules/timer/wrap_Timer.cpp"
 _lua_pushnumber($L,$4); //@line 55 "modules/timer/wrap_Timer.cpp"
 return 1; //@line 56 "modules/timer/wrap_Timer.cpp"
}
function __ZN4love5timer7w_sleepEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625664)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 61 "modules/timer/wrap_Timer.cpp"
 $2 = ((($1)) + 28|0); //@line 61 "modules/timer/wrap_Timer.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 61 "modules/timer/wrap_Timer.cpp"
 $4 = (+_luaL_checknumber($L,1)); //@line 61 "modules/timer/wrap_Timer.cpp"
 FUNCTION_TABLE_vid[$3 & 8191]($0,$4); //@line 61 "modules/timer/wrap_Timer.cpp"
 return 0; //@line 62 "modules/timer/wrap_Timer.cpp"
}
function __ZN4love5timer9w_getTimeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625664)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 67 "modules/timer/wrap_Timer.cpp"
 $2 = ((($1)) + 44|0); //@line 67 "modules/timer/wrap_Timer.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 67 "modules/timer/wrap_Timer.cpp"
 $4 = (+FUNCTION_TABLE_di[$3 & 4095]($0)); //@line 67 "modules/timer/wrap_Timer.cpp"
 _lua_pushnumber($L,$4); //@line 67 "modules/timer/wrap_Timer.cpp"
 return 1; //@line 68 "modules/timer/wrap_Timer.cpp"
}
function _luaopen_love_timer($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0;
 var $eh$lpad$body$i$indexZ2D = 0, $instance$sroa$0$1 = 0, $vararg_buffer = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $w = sp + 4|0;
 $0 = HEAP32[(625664)>>2]|0; //@line 98 "./common/Module.h"
 $1 = $0; //@line 86 "modules/timer/wrap_Timer.cpp"
 $2 = ($0|0)==(0|0); //@line 87 "modules/timer/wrap_Timer.cpp"
 L1: do {
  if ($2) {
   __THREW__ = 0;
   $3 = (invoke_ii(3348,72)|0); //@line 89 "modules/timer/wrap_Timer.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   do {
    if ($5) {
     $11 = ___cxa_find_matching_catch_3(8992|0)|0;
     $12 = tempRet0;
     $eh$lpad$body$i$index2Z2D = $12;$eh$lpad$body$i$indexZ2D = $11;
    } else {
     __THREW__ = 0;
     invoke_vi(3464,($3|0)); //@line 89 "modules/timer/wrap_Timer.cpp"
     $6 = __THREW__; __THREW__ = 0;
     $7 = $6&1;
     if ($7) {
      $8 = ___cxa_find_matching_catch_3(8992|0)|0;
      $9 = tempRet0;
      __ZdlPv($3); //@line 89 "modules/timer/wrap_Timer.cpp"
      $eh$lpad$body$i$index2Z2D = $9;$eh$lpad$body$i$indexZ2D = $8;
      break;
     } else {
      $10 = $3; //@line 89 "modules/timer/wrap_Timer.cpp"
      $instance$sroa$0$1 = $10;
      break L1;
     }
    }
   } while(0);
   $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $14 = ($eh$lpad$body$i$index2Z2D|0)==($13|0); //@line 535 "./common/runtime.h"
   if (!($14)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $15 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
   $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
   $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
   $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $instance$sroa$0$1 = $1;
    break;
   }
   $23 = ___cxa_find_matching_catch_2()|0;
   $24 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    $27 = ___cxa_find_matching_catch_3(0|0)|0;
    $28 = tempRet0;
    ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $24;$$02$i = $23;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $29 = HEAP32[$0>>2]|0; //@line 92 "modules/timer/wrap_Timer.cpp"
   $30 = ((($29)) + 8|0); //@line 92 "modules/timer/wrap_Timer.cpp"
   $31 = HEAP32[$30>>2]|0; //@line 92 "modules/timer/wrap_Timer.cpp"
   FUNCTION_TABLE_vi[$31 & 8191]($0); //@line 92 "modules/timer/wrap_Timer.cpp"
   $instance$sroa$0$1 = $1;
  }
 } while(0);
 HEAP32[$w>>2] = $instance$sroa$0$1; //@line 95 "modules/timer/wrap_Timer.cpp"
 $32 = ((($w)) + 4|0); //@line 96 "modules/timer/wrap_Timer.cpp"
 HEAP32[$32>>2] = 266000; //@line 96 "modules/timer/wrap_Timer.cpp"
 $33 = ((($w)) + 8|0); //@line 97 "modules/timer/wrap_Timer.cpp"
 HEAP32[$33>>2] = 3; //@line 97 "modules/timer/wrap_Timer.cpp"
 $34 = ((($w)) + 12|0); //@line 98 "modules/timer/wrap_Timer.cpp"
 HEAP32[$34>>2] = 11404; //@line 98 "modules/timer/wrap_Timer.cpp"
 $35 = ((($w)) + 16|0); //@line 99 "modules/timer/wrap_Timer.cpp"
 HEAP32[$35>>2] = 0; //@line 99 "modules/timer/wrap_Timer.cpp"
 $36 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 101 "modules/timer/wrap_Timer.cpp"
 STACKTOP = sp;return ($36|0); //@line 101 "modules/timer/wrap_Timer.cpp"
}
function __ZN4love5touch12w_getTouchesEP9lua_State($L) {
 $L = $L|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625668)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 45 "modules/touch/wrap_Touch.cpp"
 $2 = ((($1)) + 24|0); //@line 45 "modules/touch/wrap_Touch.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 45 "modules/touch/wrap_Touch.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 45 "modules/touch/wrap_Touch.cpp"
 $5 = ((($4)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$5>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = HEAP32[$4>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = (($6) - ($7))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8|0) / 48)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 _lua_createtable($L,$9,0); //@line 47 "modules/touch/wrap_Touch.cpp"
 $10 = HEAP32[$5>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = HEAP32[$4>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($10|0)==($11|0); //@line 49 "modules/touch/wrap_Touch.cpp"
 if ($12) {
  return 1; //@line 60 "modules/touch/wrap_Touch.cpp"
 } else {
  $$in = $11;$i$01 = 0;
 }
 while(1) {
  $13 = $$in; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) + (($i$01*48)|0)|0); //@line 56 "modules/touch/wrap_Touch.cpp"
  $15 = $14; //@line 56 "modules/touch/wrap_Touch.cpp"
  $16 = $15; //@line 56 "modules/touch/wrap_Touch.cpp"
  $17 = HEAP32[$16>>2]|0; //@line 56 "modules/touch/wrap_Touch.cpp"
  $18 = (($15) + 4)|0; //@line 56 "modules/touch/wrap_Touch.cpp"
  $19 = $18; //@line 56 "modules/touch/wrap_Touch.cpp"
  $20 = HEAP32[$19>>2]|0; //@line 56 "modules/touch/wrap_Touch.cpp"
  $21 = $17; //@line 56 "modules/touch/wrap_Touch.cpp"
  _lua_pushlightuserdata($L,$21); //@line 56 "modules/touch/wrap_Touch.cpp"
  $22 = (($i$01) + 1)|0; //@line 57 "modules/touch/wrap_Touch.cpp"
  _lua_rawseti($L,-2,$22); //@line 57 "modules/touch/wrap_Touch.cpp"
  $23 = HEAP32[$5>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $24 = HEAP32[$4>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $25 = (($23) - ($24))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $26 = (($25|0) / 48)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $27 = ($22>>>0)<($26>>>0); //@line 49 "modules/touch/wrap_Touch.cpp"
  if ($27) {
   $$in = $24;$i$01 = $22;
  } else {
   break;
  }
 }
 return 1; //@line 60 "modules/touch/wrap_Touch.cpp"
}
function __ZN4love5touch13w_getPositionEP9lua_State($L) {
 $L = $L|0;
 var $$0$in$i = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $touch$sroa$4$0 = 0.0;
 var $touch$sroa$5$0 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 63 "modules/touch/wrap_Touch.cpp"
 $0 = (_lua_type($L,1)|0); //@line 37 "modules/touch/wrap_Touch.cpp"
 $1 = ($0|0)==(2); //@line 37 "modules/touch/wrap_Touch.cpp"
 if ($1) {
  $3 = (_lua_touserdata($L,1)|0); //@line 40 "modules/touch/wrap_Touch.cpp"
  $4 = $3; //@line 40 "modules/touch/wrap_Touch.cpp"
  $$0$in$i = $4;
 } else {
  $2 = (_luax_typerror($L,1,266046)|0); //@line 38 "modules/touch/wrap_Touch.cpp"
  $$0$in$i = $2;
 }
 $5 = ($$0$in$i|0)<(0); //@line 40 "modules/touch/wrap_Touch.cpp"
 $6 = $5 << 31 >> 31; //@line 40 "modules/touch/wrap_Touch.cpp"
 $7 = HEAP32[(625668)>>2]|0; //@line 98 "./common/Module.h"
 $8 = HEAP32[$7>>2]|0; //@line 68 "modules/touch/wrap_Touch.cpp"
 $9 = ((($8)) + 28|0); //@line 68 "modules/touch/wrap_Touch.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 68 "modules/touch/wrap_Touch.cpp"
 __THREW__ = 0;
 $11 = (invoke_iiii($10|0,($7|0),($$0$in$i|0),($6|0))|0); //@line 68 "modules/touch/wrap_Touch.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if (!($13)) {
  $14 = ((($11)) + 8|0); //@line 68 "modules/touch/wrap_Touch.cpp"
  $15 = +HEAPF64[$14>>3]; //@line 68 "modules/touch/wrap_Touch.cpp"
  $16 = ((($11)) + 16|0); //@line 68 "modules/touch/wrap_Touch.cpp"
  $17 = +HEAPF64[$16>>3]; //@line 68 "modules/touch/wrap_Touch.cpp"
  $touch$sroa$4$0 = $15;$touch$sroa$5$0 = $17;
  _lua_pushnumber($L,$touch$sroa$4$0); //@line 70 "modules/touch/wrap_Touch.cpp"
  _lua_pushnumber($L,$touch$sroa$5$0); //@line 71 "modules/touch/wrap_Touch.cpp"
  STACKTOP = sp;return 2; //@line 73 "modules/touch/wrap_Touch.cpp"
 }
 $18 = ___cxa_find_matching_catch_3(8992|0)|0;
 $19 = tempRet0;
 $20 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $21 = ($19|0)==($20|0); //@line 535 "./common/runtime.h"
 if (!($21)) {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $22 = (___cxa_begin_catch(($18|0))|0); //@line 535 "./common/runtime.h"
 $23 = HEAP32[$22>>2]|0; //@line 539 "./common/runtime.h"
 $24 = ((($23)) + 8|0); //@line 539 "./common/runtime.h"
 $25 = HEAP32[$24>>2]|0; //@line 539 "./common/runtime.h"
 $26 = (FUNCTION_TABLE_ii[$25 & 8191]($22)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($26|0)); //@line 539 "./common/runtime.h"
 $27 = __THREW__; __THREW__ = 0;
 $28 = $27&1;
 if (!($28)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $29 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $29; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $touch$sroa$4$0 = 0.0;$touch$sroa$5$0 = 0.0;
  _lua_pushnumber($L,$touch$sroa$4$0); //@line 70 "modules/touch/wrap_Touch.cpp"
  _lua_pushnumber($L,$touch$sroa$5$0); //@line 71 "modules/touch/wrap_Touch.cpp"
  STACKTOP = sp;return 2; //@line 73 "modules/touch/wrap_Touch.cpp"
 }
 $30 = ___cxa_find_matching_catch_2()|0;
 $31 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $32 = __THREW__; __THREW__ = 0;
 $33 = $32&1;
 if ($33) {
  $34 = ___cxa_find_matching_catch_3(0|0)|0;
  $35 = tempRet0;
  ___clang_call_terminate($34); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $31;$$02$i = $30;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5touch13w_getPressureEP9lua_State($L) {
 $L = $L|0;
 var $$0$in$i = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $touch$sroa$4$0 = 0.0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 76 "modules/touch/wrap_Touch.cpp"
 $0 = (_lua_type($L,1)|0); //@line 37 "modules/touch/wrap_Touch.cpp"
 $1 = ($0|0)==(2); //@line 37 "modules/touch/wrap_Touch.cpp"
 if ($1) {
  $3 = (_lua_touserdata($L,1)|0); //@line 40 "modules/touch/wrap_Touch.cpp"
  $4 = $3; //@line 40 "modules/touch/wrap_Touch.cpp"
  $$0$in$i = $4;
 } else {
  $2 = (_luax_typerror($L,1,266046)|0); //@line 38 "modules/touch/wrap_Touch.cpp"
  $$0$in$i = $2;
 }
 $5 = ($$0$in$i|0)<(0); //@line 40 "modules/touch/wrap_Touch.cpp"
 $6 = $5 << 31 >> 31; //@line 40 "modules/touch/wrap_Touch.cpp"
 $7 = HEAP32[(625668)>>2]|0; //@line 98 "./common/Module.h"
 $8 = HEAP32[$7>>2]|0; //@line 81 "modules/touch/wrap_Touch.cpp"
 $9 = ((($8)) + 28|0); //@line 81 "modules/touch/wrap_Touch.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 81 "modules/touch/wrap_Touch.cpp"
 __THREW__ = 0;
 $11 = (invoke_iiii($10|0,($7|0),($$0$in$i|0),($6|0))|0); //@line 81 "modules/touch/wrap_Touch.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if (!($13)) {
  $14 = ((($11)) + 40|0); //@line 81 "modules/touch/wrap_Touch.cpp"
  $15 = +HEAPF64[$14>>3]; //@line 81 "modules/touch/wrap_Touch.cpp"
  $touch$sroa$4$0 = $15;
  _lua_pushnumber($L,$touch$sroa$4$0); //@line 83 "modules/touch/wrap_Touch.cpp"
  STACKTOP = sp;return 1; //@line 84 "modules/touch/wrap_Touch.cpp"
 }
 $16 = ___cxa_find_matching_catch_3(8992|0)|0;
 $17 = tempRet0;
 $18 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $19 = ($17|0)==($18|0); //@line 535 "./common/runtime.h"
 if (!($19)) {
  $$01$i = $17;$$02$i = $16;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $20 = (___cxa_begin_catch(($16|0))|0); //@line 535 "./common/runtime.h"
 $21 = HEAP32[$20>>2]|0; //@line 539 "./common/runtime.h"
 $22 = ((($21)) + 8|0); //@line 539 "./common/runtime.h"
 $23 = HEAP32[$22>>2]|0; //@line 539 "./common/runtime.h"
 $24 = (FUNCTION_TABLE_ii[$23 & 8191]($20)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($24|0)); //@line 539 "./common/runtime.h"
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if (!($26)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $27 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $27; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $touch$sroa$4$0 = 0.0;
  _lua_pushnumber($L,$touch$sroa$4$0); //@line 83 "modules/touch/wrap_Touch.cpp"
  STACKTOP = sp;return 1; //@line 84 "modules/touch/wrap_Touch.cpp"
 }
 $28 = ___cxa_find_matching_catch_2()|0;
 $29 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $30 = __THREW__; __THREW__ = 0;
 $31 = $30&1;
 if ($31) {
  $32 = ___cxa_find_matching_catch_3(0|0)|0;
  $33 = tempRet0;
  ___clang_call_terminate($32); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $29;$$02$i = $28;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function _luaopen_love_touch($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $instance$sroa$0$1 = 0, $vararg_buffer = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $w = sp + 4|0;
 $0 = HEAP32[(625668)>>2]|0; //@line 98 "./common/Module.h"
 $1 = $0; //@line 97 "modules/touch/wrap_Touch.cpp"
 $2 = ($0|0)==(0|0); //@line 98 "modules/touch/wrap_Touch.cpp"
 L1: do {
  if ($2) {
   __THREW__ = 0;
   $3 = (invoke_ii(3348,20)|0); //@line 100 "modules/touch/wrap_Touch.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   do {
    if ($5) {
     $14 = ___cxa_find_matching_catch_3(8992|0)|0;
     $15 = tempRet0;
     $eh$lpad$body$i$index2Z2D = $15;$eh$lpad$body$i$indexZ2D = $14;
    } else {
     ;HEAP32[$3>>2]=0|0;HEAP32[$3+4>>2]=0|0;HEAP32[$3+8>>2]=0|0;HEAP32[$3+12>>2]=0|0;HEAP32[$3+16>>2]=0|0; //@line 100 "modules/touch/wrap_Touch.cpp"
     __THREW__ = 0;
     invoke_vi(3431,($3|0)); //@line 33 "./common/Module.h"
     $6 = __THREW__; __THREW__ = 0;
     $7 = $6&1;
     if ($7) {
      $8 = ___cxa_find_matching_catch_3(8992|0)|0;
      $9 = tempRet0;
      __ZdlPv($3); //@line 100 "modules/touch/wrap_Touch.cpp"
      $eh$lpad$body$i$index2Z2D = $9;$eh$lpad$body$i$indexZ2D = $8;
      break;
     } else {
      HEAP32[$3>>2] = (19668); //@line 37 "modules/touch/sdl/Touch.h"
      $10 = ((($3)) + 8|0); //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$10>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $11 = ((($3)) + 12|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$11>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $12 = ((($3)) + 16|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$12>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $13 = $3; //@line 100 "modules/touch/wrap_Touch.cpp"
      $instance$sroa$0$1 = $13;
      break L1;
     }
    }
   } while(0);
   $16 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $17 = ($eh$lpad$body$i$index2Z2D|0)==($16|0); //@line 535 "./common/runtime.h"
   if (!($17)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $18 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $19 = HEAP32[$18>>2]|0; //@line 539 "./common/runtime.h"
   $20 = ((($19)) + 8|0); //@line 539 "./common/runtime.h"
   $21 = HEAP32[$20>>2]|0; //@line 539 "./common/runtime.h"
   $22 = (FUNCTION_TABLE_ii[$21 & 8191]($18)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($22|0)); //@line 539 "./common/runtime.h"
   $23 = __THREW__; __THREW__ = 0;
   $24 = $23&1;
   if (!($24)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $25 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $25; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $instance$sroa$0$1 = $1;
    break;
   }
   $26 = ___cxa_find_matching_catch_2()|0;
   $27 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $28 = __THREW__; __THREW__ = 0;
   $29 = $28&1;
   if ($29) {
    $30 = ___cxa_find_matching_catch_3(0|0)|0;
    $31 = tempRet0;
    ___clang_call_terminate($30); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $27;$$02$i = $26;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $32 = HEAP32[$0>>2]|0; //@line 103 "modules/touch/wrap_Touch.cpp"
   $33 = ((($32)) + 8|0); //@line 103 "modules/touch/wrap_Touch.cpp"
   $34 = HEAP32[$33>>2]|0; //@line 103 "modules/touch/wrap_Touch.cpp"
   FUNCTION_TABLE_vi[$34 & 8191]($0); //@line 103 "modules/touch/wrap_Touch.cpp"
   $instance$sroa$0$1 = $1;
  }
 } while(0);
 HEAP32[$w>>2] = $instance$sroa$0$1; //@line 106 "modules/touch/wrap_Touch.cpp"
 $35 = ((($w)) + 4|0); //@line 107 "modules/touch/wrap_Touch.cpp"
 HEAP32[$35>>2] = 266055; //@line 107 "modules/touch/wrap_Touch.cpp"
 $36 = ((($w)) + 8|0); //@line 108 "modules/touch/wrap_Touch.cpp"
 HEAP32[$36>>2] = 3; //@line 108 "modules/touch/wrap_Touch.cpp"
 $37 = ((($w)) + 12|0); //@line 109 "modules/touch/wrap_Touch.cpp"
 HEAP32[$37>>2] = 11460; //@line 109 "modules/touch/wrap_Touch.cpp"
 $38 = ((($w)) + 16|0); //@line 110 "modules/touch/wrap_Touch.cpp"
 HEAP32[$38>>2] = 0; //@line 110 "modules/touch/wrap_Touch.cpp"
 $39 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 112 "modules/touch/wrap_Touch.cpp"
 STACKTOP = sp;return ($39|0); //@line 112 "modules/touch/wrap_Touch.cpp"
}
function __ZN4love5video16w_newVideoStreamEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $stream$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 35 "modules/video/wrap_Video.cpp"
 $vararg_buffer = sp; //@line 35 "modules/video/wrap_Video.cpp"
 $0 = (__ZN4love10filesystem12luax_getfileEP9lua_Statei($L,1)|0); //@line 37 "modules/video/wrap_Video.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 42 "modules/video/wrap_Video.cpp"
 $2 = ((($1)) + 24|0); //@line 42 "modules/video/wrap_Video.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 42 "modules/video/wrap_Video.cpp"
 __THREW__ = 0;
 $4 = (invoke_ii($3|0,($0|0))|0); //@line 42 "modules/video/wrap_Video.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 do {
  if (!($6)) {
   if (!($4)) {
    $7 = HEAP32[$0>>2]|0; //@line 42 "modules/video/wrap_Video.cpp"
    $8 = ((($7)) + 16|0); //@line 42 "modules/video/wrap_Video.cpp"
    $9 = HEAP32[$8>>2]|0; //@line 42 "modules/video/wrap_Video.cpp"
    __THREW__ = 0;
    $10 = (invoke_iii($9|0,($0|0),1)|0); //@line 42 "modules/video/wrap_Video.cpp"
    $11 = __THREW__; __THREW__ = 0;
    $12 = $11&1;
    if ($12) {
     break;
    }
    if (!($10)) {
     __THREW__ = 0;
     (invoke_iiii(3389,($L|0),(266084|0),($vararg_buffer|0))|0); //@line 43 "modules/video/wrap_Video.cpp"
     $13 = __THREW__; __THREW__ = 0;
     $14 = $13&1;
     if ($14) {
      break;
     }
    }
   }
   $15 = HEAP32[(625676)>>2]|0; //@line 98 "./common/Module.h"
   $16 = HEAP32[$15>>2]|0; //@line 45 "modules/video/wrap_Video.cpp"
   $17 = ((($16)) + 24|0); //@line 45 "modules/video/wrap_Video.cpp"
   $18 = HEAP32[$17>>2]|0; //@line 45 "modules/video/wrap_Video.cpp"
   __THREW__ = 0;
   $19 = (invoke_iii($18|0,($15|0),($0|0))|0); //@line 45 "modules/video/wrap_Video.cpp"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    $stream$0 = $19;
    __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,55,$stream$0); //@line 48 "modules/video/wrap_Video.cpp"
    $40 = HEAP32[$stream$0>>2]|0; //@line 49 "modules/video/wrap_Video.cpp"
    $41 = ((($40)) + 12|0); //@line 49 "modules/video/wrap_Video.cpp"
    $42 = HEAP32[$41>>2]|0; //@line 49 "modules/video/wrap_Video.cpp"
    FUNCTION_TABLE_vi[$42 & 8191]($stream$0); //@line 49 "modules/video/wrap_Video.cpp"
    STACKTOP = sp;return 1; //@line 50 "modules/video/wrap_Video.cpp"
   }
  }
 } while(0);
 $22 = ___cxa_find_matching_catch_3(8992|0)|0;
 $23 = tempRet0;
 $24 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $25 = ($23|0)==($24|0); //@line 535 "./common/runtime.h"
 if (!($25)) {
  $$01$i = $23;$$02$i = $22;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $26 = (___cxa_begin_catch(($22|0))|0); //@line 535 "./common/runtime.h"
 $27 = HEAP32[$26>>2]|0; //@line 539 "./common/runtime.h"
 $28 = ((($27)) + 8|0); //@line 539 "./common/runtime.h"
 $29 = HEAP32[$28>>2]|0; //@line 539 "./common/runtime.h"
 $30 = (FUNCTION_TABLE_ii[$29 & 8191]($26)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($30|0)); //@line 539 "./common/runtime.h"
 $31 = __THREW__; __THREW__ = 0;
 $32 = $31&1;
 if (!($32)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $33 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $33; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  $stream$0 = 0;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,55,$stream$0); //@line 48 "modules/video/wrap_Video.cpp"
  $40 = HEAP32[$stream$0>>2]|0; //@line 49 "modules/video/wrap_Video.cpp"
  $41 = ((($40)) + 12|0); //@line 49 "modules/video/wrap_Video.cpp"
  $42 = HEAP32[$41>>2]|0; //@line 49 "modules/video/wrap_Video.cpp"
  FUNCTION_TABLE_vi[$42 & 8191]($stream$0); //@line 49 "modules/video/wrap_Video.cpp"
  STACKTOP = sp;return 1; //@line 50 "modules/video/wrap_Video.cpp"
 }
 $34 = ___cxa_find_matching_catch_2()|0;
 $35 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $36 = __THREW__; __THREW__ = 0;
 $37 = $36&1;
 if ($37) {
  $38 = ___cxa_find_matching_catch_3(0|0)|0;
  $39 = tempRet0;
  ___clang_call_terminate($38); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $35;$$02$i = $34;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function _luaopen_love_video($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0;
 var $eh$lpad$body$i$indexZ2D = 0, $instance$sroa$0$1 = 0, $vararg_buffer = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $w = sp + 4|0;
 $0 = HEAP32[(625676)>>2]|0; //@line 98 "./common/Module.h"
 $1 = $0; //@line 67 "modules/video/wrap_Video.cpp"
 $2 = ($0|0)==(0|0); //@line 68 "modules/video/wrap_Video.cpp"
 L1: do {
  if ($2) {
   __THREW__ = 0;
   $3 = (invoke_ii(3348,12)|0); //@line 70 "modules/video/wrap_Video.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   do {
    if ($5) {
     $11 = ___cxa_find_matching_catch_3(8992|0)|0;
     $12 = tempRet0;
     $eh$lpad$body$i$index2Z2D = $12;$eh$lpad$body$i$indexZ2D = $11;
    } else {
     __THREW__ = 0;
     invoke_vi(3465,($3|0)); //@line 70 "modules/video/wrap_Video.cpp"
     $6 = __THREW__; __THREW__ = 0;
     $7 = $6&1;
     if ($7) {
      $8 = ___cxa_find_matching_catch_3(8992|0)|0;
      $9 = tempRet0;
      __ZdlPv($3); //@line 70 "modules/video/wrap_Video.cpp"
      $eh$lpad$body$i$index2Z2D = $9;$eh$lpad$body$i$indexZ2D = $8;
      break;
     } else {
      $10 = $3; //@line 70 "modules/video/wrap_Video.cpp"
      $instance$sroa$0$1 = $10;
      break L1;
     }
    }
   } while(0);
   $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $14 = ($eh$lpad$body$i$index2Z2D|0)==($13|0); //@line 535 "./common/runtime.h"
   if (!($14)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $15 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
   $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
   $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
   $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $instance$sroa$0$1 = $1;
    break;
   }
   $23 = ___cxa_find_matching_catch_2()|0;
   $24 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    $27 = ___cxa_find_matching_catch_3(0|0)|0;
    $28 = tempRet0;
    ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $24;$$02$i = $23;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $29 = HEAP32[$0>>2]|0; //@line 73 "modules/video/wrap_Video.cpp"
   $30 = ((($29)) + 8|0); //@line 73 "modules/video/wrap_Video.cpp"
   $31 = HEAP32[$30>>2]|0; //@line 73 "modules/video/wrap_Video.cpp"
   FUNCTION_TABLE_vi[$31 & 8191]($0); //@line 73 "modules/video/wrap_Video.cpp"
   $instance$sroa$0$1 = $1;
  }
 } while(0);
 HEAP32[$w>>2] = $instance$sroa$0$1; //@line 76 "modules/video/wrap_Video.cpp"
 $32 = ((($w)) + 4|0); //@line 77 "modules/video/wrap_Video.cpp"
 HEAP32[$32>>2] = 330161; //@line 77 "modules/video/wrap_Video.cpp"
 $33 = ((($w)) + 8|0); //@line 78 "modules/video/wrap_Video.cpp"
 HEAP32[$33>>2] = 3; //@line 78 "modules/video/wrap_Video.cpp"
 $34 = ((($w)) + 12|0); //@line 79 "modules/video/wrap_Video.cpp"
 HEAP32[$34>>2] = 11492; //@line 79 "modules/video/wrap_Video.cpp"
 $35 = ((($w)) + 16|0); //@line 80 "modules/video/wrap_Video.cpp"
 HEAP32[$35>>2] = 11508; //@line 80 "modules/video/wrap_Video.cpp"
 $36 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 82 "modules/video/wrap_Video.cpp"
 STACKTOP = sp;return ($36|0); //@line 82 "modules/video/wrap_Video.cpp"
}
function __ZN4love14luax_checktypeINS_5video11VideoStreamEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love5video21w_VideoStream_setSyncEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5video11VideoStreamEEEPT_P9lua_StateiNS_4TypeE($L,1,55)|0); //@line 30 "modules/video/wrap_VideoStream.cpp"
 $1 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,2,28)|0); //@line 37 "modules/video/wrap_VideoStream.cpp"
 if ($1) {
  $2 = (_lua_touserdata($L,2)|0); //@line 516 "./common/runtime.h"
  $3 = ((($2)) + 4|0); //@line 516 "./common/runtime.h"
  $4 = HEAP32[$3>>2]|0; //@line 516 "./common/runtime.h"
  $5 = (__Znwj(12)|0); //@line 40 "modules/video/wrap_VideoStream.cpp"
  __THREW__ = 0;
  invoke_vii(3466,($5|0),($4|0)); //@line 40 "modules/video/wrap_VideoStream.cpp"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if ($7) {
   $14 = ___cxa_find_matching_catch_2()|0;
   $15 = tempRet0;
   __ZdlPv($5); //@line 40 "modules/video/wrap_VideoStream.cpp"
   $$01 = $14;$$02 = $15;
   ___resumeException($$01|0);
   // unreachable;
  } else {
   $8 = HEAP32[$0>>2]|0; //@line 41 "modules/video/wrap_VideoStream.cpp"
   $9 = ((($8)) + 64|0); //@line 41 "modules/video/wrap_VideoStream.cpp"
   $10 = HEAP32[$9>>2]|0; //@line 41 "modules/video/wrap_VideoStream.cpp"
   FUNCTION_TABLE_vii[$10 & 8191]($0,$5); //@line 41 "modules/video/wrap_VideoStream.cpp"
   $11 = HEAP32[$5>>2]|0; //@line 42 "modules/video/wrap_VideoStream.cpp"
   $12 = ((($11)) + 12|0); //@line 42 "modules/video/wrap_VideoStream.cpp"
   $13 = HEAP32[$12>>2]|0; //@line 42 "modules/video/wrap_VideoStream.cpp"
   FUNCTION_TABLE_vi[$13 & 8191]($5); //@line 42 "modules/video/wrap_VideoStream.cpp"
   $$0 = 0;
   return ($$0|0); //@line 60 "modules/video/wrap_VideoStream.cpp"
  }
 }
 $16 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,2,55)|0); //@line 44 "modules/video/wrap_VideoStream.cpp"
 if ($16) {
  $17 = (_lua_touserdata($L,2)|0); //@line 516 "./common/runtime.h"
  $18 = ((($17)) + 4|0); //@line 516 "./common/runtime.h"
  $19 = HEAP32[$18>>2]|0; //@line 516 "./common/runtime.h"
  $20 = HEAP32[$0>>2]|0; //@line 47 "modules/video/wrap_VideoStream.cpp"
  $21 = ((($20)) + 64|0); //@line 47 "modules/video/wrap_VideoStream.cpp"
  $22 = HEAP32[$21>>2]|0; //@line 47 "modules/video/wrap_VideoStream.cpp"
  $23 = HEAP32[$19>>2]|0; //@line 47 "modules/video/wrap_VideoStream.cpp"
  $24 = ((($23)) + 68|0); //@line 47 "modules/video/wrap_VideoStream.cpp"
  $25 = HEAP32[$24>>2]|0; //@line 47 "modules/video/wrap_VideoStream.cpp"
  $26 = (FUNCTION_TABLE_ii[$25 & 8191]($19)|0); //@line 47 "modules/video/wrap_VideoStream.cpp"
  FUNCTION_TABLE_vii[$22 & 8191]($0,$26); //@line 47 "modules/video/wrap_VideoStream.cpp"
  $$0 = 0;
  return ($$0|0); //@line 60 "modules/video/wrap_VideoStream.cpp"
 }
 $27 = (_lua_type($L,2)|0); //@line 49 "modules/video/wrap_VideoStream.cpp"
 $28 = ($27|0)<(1); //@line 49 "modules/video/wrap_VideoStream.cpp"
 if (!($28)) {
  $44 = (_luax_typerror($L,2,266122)|0); //@line 57 "modules/video/wrap_VideoStream.cpp"
  $$0 = $44;
  return ($$0|0); //@line 60 "modules/video/wrap_VideoStream.cpp"
 }
 $29 = (__Znwj(40)|0); //@line 51 "modules/video/wrap_VideoStream.cpp"
 __THREW__ = 0;
 invoke_vi(3467,($29|0)); //@line 51 "modules/video/wrap_VideoStream.cpp"
 $30 = __THREW__; __THREW__ = 0;
 $31 = $30&1;
 if ($31) {
  $42 = ___cxa_find_matching_catch_2()|0;
  $43 = tempRet0;
  __ZdlPv($29); //@line 51 "modules/video/wrap_VideoStream.cpp"
  $$01 = $42;$$02 = $43;
  ___resumeException($$01|0);
  // unreachable;
 } else {
  $32 = HEAP32[$0>>2]|0; //@line 52 "modules/video/wrap_VideoStream.cpp"
  $33 = ((($32)) + 68|0); //@line 52 "modules/video/wrap_VideoStream.cpp"
  $34 = HEAP32[$33>>2]|0; //@line 52 "modules/video/wrap_VideoStream.cpp"
  $35 = (FUNCTION_TABLE_ii[$34 & 8191]($0)|0); //@line 52 "modules/video/wrap_VideoStream.cpp"
  __ZN4love5video11VideoStream9FrameSync9copyStateEPKS2_($29,$35); //@line 52 "modules/video/wrap_VideoStream.cpp"
  $36 = HEAP32[$0>>2]|0; //@line 53 "modules/video/wrap_VideoStream.cpp"
  $37 = ((($36)) + 64|0); //@line 53 "modules/video/wrap_VideoStream.cpp"
  $38 = HEAP32[$37>>2]|0; //@line 53 "modules/video/wrap_VideoStream.cpp"
  FUNCTION_TABLE_vii[$38 & 8191]($0,$29); //@line 53 "modules/video/wrap_VideoStream.cpp"
  $39 = HEAP32[$29>>2]|0; //@line 54 "modules/video/wrap_VideoStream.cpp"
  $40 = ((($39)) + 12|0); //@line 54 "modules/video/wrap_VideoStream.cpp"
  $41 = HEAP32[$40>>2]|0; //@line 54 "modules/video/wrap_VideoStream.cpp"
  FUNCTION_TABLE_vi[$41 & 8191]($29); //@line 54 "modules/video/wrap_VideoStream.cpp"
  $$0 = 0;
  return ($$0|0); //@line 60 "modules/video/wrap_VideoStream.cpp"
 }
 return (0)|0;
}
function __ZN4love5video25w_VideoStream_getFilenameEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5video11VideoStreamEEEPT_P9lua_StateiNS_4TypeE($L,1,55)|0); //@line 30 "modules/video/wrap_VideoStream.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 65 "modules/video/wrap_VideoStream.cpp"
 $2 = ((($1)) + 40|0); //@line 65 "modules/video/wrap_VideoStream.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 65 "modules/video/wrap_VideoStream.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 65 "modules/video/wrap_VideoStream.cpp"
 __ZN4love15luax_pushstringEP9lua_StateRKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE($L,$4); //@line 65 "modules/video/wrap_VideoStream.cpp"
 return 1; //@line 66 "modules/video/wrap_VideoStream.cpp"
}
function __ZN4love5video18w_VideoStream_playEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5video11VideoStreamEEEPT_P9lua_StateiNS_4TypeE($L,1,55)|0); //@line 30 "modules/video/wrap_VideoStream.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 72 "modules/video/wrap_VideoStream.cpp"
 $2 = ((($1)) + 44|0); //@line 72 "modules/video/wrap_VideoStream.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 72 "modules/video/wrap_VideoStream.cpp"
 FUNCTION_TABLE_vi[$3 & 8191]($0); //@line 72 "modules/video/wrap_VideoStream.cpp"
 return 0; //@line 73 "modules/video/wrap_VideoStream.cpp"
}
function __ZN4love5video19w_VideoStream_pauseEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5video11VideoStreamEEEPT_P9lua_StateiNS_4TypeE($L,1,55)|0); //@line 30 "modules/video/wrap_VideoStream.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 79 "modules/video/wrap_VideoStream.cpp"
 $2 = ((($1)) + 48|0); //@line 79 "modules/video/wrap_VideoStream.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 79 "modules/video/wrap_VideoStream.cpp"
 FUNCTION_TABLE_vi[$3 & 8191]($0); //@line 79 "modules/video/wrap_VideoStream.cpp"
 return 0; //@line 80 "modules/video/wrap_VideoStream.cpp"
}
function __ZN4love5video18w_VideoStream_seekEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5video11VideoStreamEEEPT_P9lua_StateiNS_4TypeE($L,1,55)|0); //@line 30 "modules/video/wrap_VideoStream.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 86 "modules/video/wrap_VideoStream.cpp"
 $2 = HEAP32[$0>>2]|0; //@line 87 "modules/video/wrap_VideoStream.cpp"
 $3 = ((($2)) + 52|0); //@line 87 "modules/video/wrap_VideoStream.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 87 "modules/video/wrap_VideoStream.cpp"
 FUNCTION_TABLE_vid[$4 & 8191]($0,$1); //@line 87 "modules/video/wrap_VideoStream.cpp"
 return 0; //@line 88 "modules/video/wrap_VideoStream.cpp"
}
function __ZN4love5video20w_VideoStream_rewindEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5video11VideoStreamEEEPT_P9lua_StateiNS_4TypeE($L,1,55)|0); //@line 30 "modules/video/wrap_VideoStream.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 94 "modules/video/wrap_VideoStream.cpp"
 $2 = ((($1)) + 52|0); //@line 94 "modules/video/wrap_VideoStream.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 94 "modules/video/wrap_VideoStream.cpp"
 FUNCTION_TABLE_vid[$3 & 8191]($0,0.0); //@line 94 "modules/video/wrap_VideoStream.cpp"
 return 0; //@line 95 "modules/video/wrap_VideoStream.cpp"
}
function __ZN4love5video18w_VideoStream_tellEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5video11VideoStreamEEEPT_P9lua_StateiNS_4TypeE($L,1,55)|0); //@line 30 "modules/video/wrap_VideoStream.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 101 "modules/video/wrap_VideoStream.cpp"
 $2 = ((($1)) + 56|0); //@line 101 "modules/video/wrap_VideoStream.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 101 "modules/video/wrap_VideoStream.cpp"
 $4 = (+FUNCTION_TABLE_di[$3 & 4095]($0)); //@line 101 "modules/video/wrap_VideoStream.cpp"
 _lua_pushnumber($L,$4); //@line 101 "modules/video/wrap_VideoStream.cpp"
 return 1; //@line 102 "modules/video/wrap_VideoStream.cpp"
}
function __ZN4love5video23w_VideoStream_isPlayingEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5video11VideoStreamEEEPT_P9lua_StateiNS_4TypeE($L,1,55)|0); //@line 30 "modules/video/wrap_VideoStream.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 108 "modules/video/wrap_VideoStream.cpp"
 $2 = ((($1)) + 60|0); //@line 108 "modules/video/wrap_VideoStream.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 108 "modules/video/wrap_VideoStream.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 108 "modules/video/wrap_VideoStream.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 108 "modules/video/wrap_VideoStream.cpp"
 return 1; //@line 109 "modules/video/wrap_VideoStream.cpp"
}
function __ZN4love5video19luaopen_videostreamEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 125 "modules/video/wrap_VideoStream.cpp"
 HEAP32[$vararg_buffer>>2] = 11516; //@line 127 "modules/video/wrap_VideoStream.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 127 "modules/video/wrap_VideoStream.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 127 "modules/video/wrap_VideoStream.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,55,266151,$vararg_buffer)|0); //@line 127 "modules/video/wrap_VideoStream.cpp"
 STACKTOP = sp;return ($0|0); //@line 127 "modules/video/wrap_VideoStream.cpp"
}
function __ZN4love6window17w_getDisplayCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 33 "modules/window/wrap_Window.cpp"
 $2 = ((($1)) + 48|0); //@line 33 "modules/window/wrap_Window.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 33 "modules/window/wrap_Window.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 33 "modules/window/wrap_Window.cpp"
 _lua_pushinteger($L,$4); //@line 33 "modules/window/wrap_Window.cpp"
 return 1; //@line 34 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window16w_getDisplayNameEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0.0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 37 "modules/window/wrap_Window.cpp"
 $0 = (+_luaL_checknumber($L,1)); //@line 39 "modules/window/wrap_Window.cpp"
 $1 = (~~(($0))); //@line 39 "modules/window/wrap_Window.cpp"
 $2 = (($1) + -1)|0; //@line 39 "modules/window/wrap_Window.cpp"
 $3 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $4 = HEAP32[$3>>2]|0; //@line 42 "modules/window/wrap_Window.cpp"
 $5 = ((($4)) + 52|0); //@line 42 "modules/window/wrap_Window.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 42 "modules/window/wrap_Window.cpp"
 __THREW__ = 0;
 $7 = (invoke_iii($6|0,($3|0),($2|0))|0); //@line 42 "modules/window/wrap_Window.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if (!($9)) {
  $name$0 = $7;
  _lua_pushstring($L,$name$0); //@line 44 "modules/window/wrap_Window.cpp"
  STACKTOP = sp;return 1; //@line 45 "modules/window/wrap_Window.cpp"
 }
 $10 = ___cxa_find_matching_catch_3(8992|0)|0;
 $11 = tempRet0;
 $12 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $13 = ($11|0)==($12|0); //@line 535 "./common/runtime.h"
 if (!($13)) {
  $$01$i = $11;$$02$i = $10;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $14 = (___cxa_begin_catch(($10|0))|0); //@line 535 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = ((($15)) + 8|0); //@line 539 "./common/runtime.h"
 $17 = HEAP32[$16>>2]|0; //@line 539 "./common/runtime.h"
 $18 = (FUNCTION_TABLE_ii[$17 & 8191]($14)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($18|0)); //@line 539 "./common/runtime.h"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if (!($20)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $21 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $21; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $name$0 = 0;
  _lua_pushstring($L,$name$0); //@line 44 "modules/window/wrap_Window.cpp"
  STACKTOP = sp;return 1; //@line 45 "modules/window/wrap_Window.cpp"
 }
 $22 = ___cxa_find_matching_catch_2()|0;
 $23 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $26 = ___cxa_find_matching_catch_3(0|0)|0;
  $27 = tempRet0;
  ___clang_call_terminate($26); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $23;$$02$i = $22;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love6window9w_setModeEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $$3 = 0, $$4 = 0, $$lcssa = 0, $0 = 0.0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0.0, $74 = 0, $75 = 0.0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $name$i = 0, $name$i10 = 0, $name$i11 = 0, $name$i12 = 0, $name$i13 = 0, $name$i14 = 0, $name$i15 = 0, $name$i16 = 0, $name$i5 = 0, $name$i6 = 0, $name$i7 = 0;
 var $name$i8 = 0, $name$i9 = 0, $phitmp = 0, $phitmp4 = 0, $setting = 0, $settings = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 80|0;
 $vararg_buffer1 = sp + 72|0;
 $vararg_buffer = sp + 64|0;
 $name$i16 = sp + 136|0;
 $name$i15 = sp + 132|0;
 $name$i14 = sp + 128|0;
 $name$i13 = sp + 124|0;
 $name$i12 = sp + 120|0;
 $name$i11 = sp + 116|0;
 $name$i10 = sp + 112|0;
 $name$i9 = sp + 108|0;
 $name$i8 = sp + 104|0;
 $name$i7 = sp + 100|0;
 $name$i6 = sp + 96|0;
 $name$i5 = sp + 92|0;
 $name$i = sp + 88|0;
 $setting = sp + 84|0;
 $settings = sp;
 $0 = (+_luaL_checknumber($L,1)); //@line 57 "modules/window/wrap_Window.cpp"
 $1 = (~~(($0))); //@line 57 "modules/window/wrap_Window.cpp"
 $2 = (+_luaL_checknumber($L,2)); //@line 58 "modules/window/wrap_Window.cpp"
 $3 = (~~(($2))); //@line 58 "modules/window/wrap_Window.cpp"
 $4 = (_lua_type($L,3)|0); //@line 60 "modules/window/wrap_Window.cpp"
 $5 = ($4|0)<(1); //@line 60 "modules/window/wrap_Window.cpp"
 if ($5) {
  $6 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
  $7 = HEAP32[$6>>2]|0; //@line 62 "modules/window/wrap_Window.cpp"
  $8 = ((($7)) + 24|0); //@line 62 "modules/window/wrap_Window.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 62 "modules/window/wrap_Window.cpp"
  $10 = (FUNCTION_TABLE_iiiii[$9 & 8191]($6,$1,$3,0)|0); //@line 62 "modules/window/wrap_Window.cpp"
  __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$10); //@line 62 "modules/window/wrap_Window.cpp"
  $$4 = 1;
  STACKTOP = sp;return ($$4|0); //@line 128 "modules/window/wrap_Window.cpp"
 }
 _luaL_checktype($L,3,5); //@line 66 "modules/window/wrap_Window.cpp"
 _lua_pushnil($L); //@line 69 "modules/window/wrap_Window.cpp"
 $11 = (_lua_next($L,3)|0); //@line 70 "modules/window/wrap_Window.cpp"
 $12 = ($11|0)==(0); //@line 70 "modules/window/wrap_Window.cpp"
 L5: do {
  if (!($12)) {
   while(1) {
    $13 = (_lua_type($L,-2)|0); //@line 72 "modules/window/wrap_Window.cpp"
    $14 = ($13|0)==(4); //@line 72 "modules/window/wrap_Window.cpp"
    if (!($14)) {
     label = 6;
     break;
    }
    $16 = (_luaL_checklstring($L,-2,0)|0); //@line 75 "modules/window/wrap_Window.cpp"
    $17 = (__ZN4love6window6Window11getConstantEPKcRNS1_7SettingE($16,$setting)|0); //@line 78 "modules/window/wrap_Window.cpp"
    if (!($17)) {
     $$lcssa = $16;
     label = 9;
     break;
    }
    _lua_settop($L,-2); //@line 81 "modules/window/wrap_Window.cpp"
    $18 = (_lua_next($L,3)|0); //@line 70 "modules/window/wrap_Window.cpp"
    $19 = ($18|0)==(0); //@line 70 "modules/window/wrap_Window.cpp"
    if ($19) {
     break L5;
    }
   }
   if ((label|0) == 6) {
    $15 = (_luax_typerror($L,-2,522489)|0); //@line 73 "modules/window/wrap_Window.cpp"
    $$4 = $15;
    STACKTOP = sp;return ($$4|0); //@line 128 "modules/window/wrap_Window.cpp"
   }
   else if ((label|0) == 9) {
    HEAP32[$vararg_buffer>>2] = $$lcssa; //@line 79 "modules/window/wrap_Window.cpp"
    $20 = (_luaL_error($L,266171,$vararg_buffer)|0); //@line 79 "modules/window/wrap_Window.cpp"
    $$4 = $20;
    STACKTOP = sp;return ($$4|0); //@line 128 "modules/window/wrap_Window.cpp"
   }
  }
 } while(0);
 HEAP8[$settings>>0] = 0; //@line 204 "./modules/window/Window.h"
 $21 = ((($settings)) + 4|0); //@line 205 "./modules/window/Window.h"
 HEAP32[$21>>2] = 1; //@line 205 "./modules/window/Window.h"
 $22 = ((($settings)) + 8|0); //@line 206 "./modules/window/Window.h"
 HEAP8[$22>>0] = 1; //@line 206 "./modules/window/Window.h"
 $23 = ((($settings)) + 12|0); //@line 207 "./modules/window/Window.h"
 HEAP32[$23>>2] = 0; //@line 207 "./modules/window/Window.h"
 $24 = ((($settings)) + 16|0); //@line 208 "./modules/window/Window.h"
 HEAP8[$24>>0] = 0; //@line 208 "./modules/window/Window.h"
 $25 = ((($settings)) + 20|0); //@line 209 "./modules/window/Window.h"
 HEAP32[$25>>2] = 1; //@line 209 "./modules/window/Window.h"
 $26 = ((($settings)) + 24|0); //@line 210 "./modules/window/Window.h"
 HEAP32[$26>>2] = 1; //@line 210 "./modules/window/Window.h"
 $27 = ((($settings)) + 28|0); //@line 211 "./modules/window/Window.h"
 HEAP8[$27>>0] = 0; //@line 211 "./modules/window/Window.h"
 $28 = ((($settings)) + 29|0); //@line 212 "./modules/window/Window.h"
 HEAP8[$28>>0] = 1; //@line 212 "./modules/window/Window.h"
 $29 = ((($settings)) + 32|0); //@line 213 "./modules/window/Window.h"
 HEAP32[$29>>2] = 0; //@line 213 "./modules/window/Window.h"
 $30 = ((($settings)) + 36|0); //@line 214 "./modules/window/Window.h"
 HEAP8[$30>>0] = 0; //@line 214 "./modules/window/Window.h"
 $31 = ((($settings)) + 40|0); //@line 215 "./modules/window/Window.h"
 HEAPF64[$31>>3] = 0.0; //@line 215 "./modules/window/Window.h"
 $32 = ((($settings)) + 48|0); //@line 216 "./modules/window/Window.h"
 HEAP8[$32>>0] = 0; //@line 216 "./modules/window/Window.h"
 $33 = ((($settings)) + 52|0); //@line 217 "./modules/window/Window.h"
 HEAP32[$33>>2] = 0; //@line 217 "./modules/window/Window.h"
 $34 = ((($settings)) + 56|0); //@line 218 "./modules/window/Window.h"
 HEAP32[$34>>2] = 0; //@line 218 "./modules/window/Window.h"
 HEAP32[$name$i>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
 (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(1,$name$i)|0); //@line 51 "modules/window/wrap_Window.cpp"
 $35 = HEAP32[$name$i>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
 _lua_getfield($L,3,$35); //@line 86 "modules/window/wrap_Window.cpp"
 $36 = (_lua_type($L,-1)|0); //@line 87 "modules/window/wrap_Window.cpp"
 $37 = ($36|0)<(1); //@line 87 "modules/window/wrap_Window.cpp"
 if ($37) {
  HEAP32[$21>>2] = 1; //@line 96 "modules/window/wrap_Window.cpp"
  label = 14;
 } else {
  $38 = (_luaL_checklstring($L,-1,0)|0); //@line 89 "modules/window/wrap_Window.cpp"
  $39 = (__ZN4love6window6Window11getConstantEPKcRNS1_14FullscreenTypeE($38,$21)|0); //@line 90 "modules/window/wrap_Window.cpp"
  if ($39) {
   label = 14;
  } else {
   HEAP32[$vararg_buffer1>>2] = $38; //@line 91 "modules/window/wrap_Window.cpp"
   $40 = (_luaL_error($L,266198,$vararg_buffer1)|0); //@line 91 "modules/window/wrap_Window.cpp"
   $$3 = $40;
  }
 }
 do {
  if ((label|0) == 14) {
   _lua_settop($L,-2); //@line 98 "modules/window/wrap_Window.cpp"
   HEAP32[$name$i5>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
   (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(0,$name$i5)|0); //@line 51 "modules/window/wrap_Window.cpp"
   $41 = HEAP32[$name$i5>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
   $42 = (__ZN4love13luax_boolflagEP9lua_StateiPKcb($L,3,$41,0)|0); //@line 100 "modules/window/wrap_Window.cpp"
   $43 = $42&1; //@line 100 "modules/window/wrap_Window.cpp"
   HEAP8[$settings>>0] = $43; //@line 100 "modules/window/wrap_Window.cpp"
   HEAP32[$name$i6>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
   (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(2,$name$i6)|0); //@line 51 "modules/window/wrap_Window.cpp"
   $44 = HEAP32[$name$i6>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
   $45 = (__ZN4love13luax_boolflagEP9lua_StateiPKcb($L,3,$44,1)|0); //@line 101 "modules/window/wrap_Window.cpp"
   $46 = $45&1; //@line 101 "modules/window/wrap_Window.cpp"
   HEAP8[$22>>0] = $46; //@line 101 "modules/window/wrap_Window.cpp"
   HEAP32[$name$i7>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
   (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(3,$name$i7)|0); //@line 51 "modules/window/wrap_Window.cpp"
   $47 = HEAP32[$name$i7>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
   $48 = (__ZN4love12luax_intflagEP9lua_StateiPKci($L,3,$47,0)|0); //@line 102 "modules/window/wrap_Window.cpp"
   HEAP32[$23>>2] = $48; //@line 102 "modules/window/wrap_Window.cpp"
   HEAP32[$name$i8>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
   (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(4,$name$i8)|0); //@line 51 "modules/window/wrap_Window.cpp"
   $49 = HEAP32[$name$i8>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
   $50 = (__ZN4love13luax_boolflagEP9lua_StateiPKcb($L,3,$49,0)|0); //@line 103 "modules/window/wrap_Window.cpp"
   $51 = $50&1; //@line 103 "modules/window/wrap_Window.cpp"
   HEAP8[$24>>0] = $51; //@line 103 "modules/window/wrap_Window.cpp"
   HEAP32[$name$i9>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
   (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(5,$name$i9)|0); //@line 51 "modules/window/wrap_Window.cpp"
   $52 = HEAP32[$name$i9>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
   $53 = (__ZN4love12luax_intflagEP9lua_StateiPKci($L,3,$52,1)|0); //@line 104 "modules/window/wrap_Window.cpp"
   HEAP32[$25>>2] = $53; //@line 104 "modules/window/wrap_Window.cpp"
   HEAP32[$name$i10>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
   (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(6,$name$i10)|0); //@line 51 "modules/window/wrap_Window.cpp"
   $54 = HEAP32[$name$i10>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
   $55 = (__ZN4love12luax_intflagEP9lua_StateiPKci($L,3,$54,1)|0); //@line 105 "modules/window/wrap_Window.cpp"
   HEAP32[$26>>2] = $55; //@line 105 "modules/window/wrap_Window.cpp"
   HEAP32[$name$i11>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
   (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(7,$name$i11)|0); //@line 51 "modules/window/wrap_Window.cpp"
   $56 = HEAP32[$name$i11>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
   $57 = (__ZN4love13luax_boolflagEP9lua_StateiPKcb($L,3,$56,0)|0); //@line 106 "modules/window/wrap_Window.cpp"
   $58 = $57&1; //@line 106 "modules/window/wrap_Window.cpp"
   HEAP8[$27>>0] = $58; //@line 106 "modules/window/wrap_Window.cpp"
   HEAP32[$name$i12>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
   (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(8,$name$i12)|0); //@line 51 "modules/window/wrap_Window.cpp"
   $59 = HEAP32[$name$i12>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
   $60 = (__ZN4love13luax_boolflagEP9lua_StateiPKcb($L,3,$59,1)|0); //@line 107 "modules/window/wrap_Window.cpp"
   $61 = $60&1; //@line 107 "modules/window/wrap_Window.cpp"
   HEAP8[$28>>0] = $61; //@line 107 "modules/window/wrap_Window.cpp"
   HEAP32[$name$i13>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
   (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(9,$name$i13)|0); //@line 51 "modules/window/wrap_Window.cpp"
   $62 = HEAP32[$name$i13>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
   $63 = (__ZN4love12luax_intflagEP9lua_StateiPKci($L,3,$62,1)|0); //@line 108 "modules/window/wrap_Window.cpp"
   $64 = (($63) + -1)|0; //@line 108 "modules/window/wrap_Window.cpp"
   HEAP32[$29>>2] = $64; //@line 108 "modules/window/wrap_Window.cpp"
   HEAP32[$name$i14>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
   (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(10,$name$i14)|0); //@line 51 "modules/window/wrap_Window.cpp"
   $65 = HEAP32[$name$i14>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
   $66 = (__ZN4love13luax_boolflagEP9lua_StateiPKcb($L,3,$65,0)|0); //@line 109 "modules/window/wrap_Window.cpp"
   $67 = $66&1; //@line 109 "modules/window/wrap_Window.cpp"
   HEAP8[$30>>0] = $67; //@line 109 "modules/window/wrap_Window.cpp"
   HEAP32[$name$i15>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
   (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(12,$name$i15)|0); //@line 51 "modules/window/wrap_Window.cpp"
   $68 = HEAP32[$name$i15>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
   _lua_getfield($L,3,$68); //@line 111 "modules/window/wrap_Window.cpp"
   HEAP32[$name$i16>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
   (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(13,$name$i16)|0); //@line 51 "modules/window/wrap_Window.cpp"
   $69 = HEAP32[$name$i16>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
   _lua_getfield($L,3,$69); //@line 112 "modules/window/wrap_Window.cpp"
   $70 = (_lua_type($L,-2)|0); //@line 113 "modules/window/wrap_Window.cpp"
   $71 = ($70|0)<(1); //@line 113 "modules/window/wrap_Window.cpp"
   if ($71) {
    $72 = (_lua_type($L,-1)|0); //@line 113 "modules/window/wrap_Window.cpp"
    $phitmp = ($72|0)>(0);
    $phitmp4 = $phitmp&1;
    HEAP8[$32>>0] = $phitmp4; //@line 113 "modules/window/wrap_Window.cpp"
    if ($phitmp) {
     label = 17;
    }
   } else {
    HEAP8[$32>>0] = 1; //@line 113 "modules/window/wrap_Window.cpp"
    label = 17;
   }
   if ((label|0) == 17) {
    $73 = (+_luaL_optnumber($L,-2,0.0)); //@line 116 "modules/window/wrap_Window.cpp"
    $74 = (~~(($73))); //@line 116 "modules/window/wrap_Window.cpp"
    HEAP32[$33>>2] = $74; //@line 116 "modules/window/wrap_Window.cpp"
    $75 = (+_luaL_optnumber($L,-1,0.0)); //@line 117 "modules/window/wrap_Window.cpp"
    $76 = (~~(($75))); //@line 117 "modules/window/wrap_Window.cpp"
    HEAP32[$34>>2] = $76; //@line 117 "modules/window/wrap_Window.cpp"
   }
   _lua_settop($L,-3); //@line 119 "modules/window/wrap_Window.cpp"
   $77 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
   $78 = HEAP32[$77>>2]|0; //@line 124 "modules/window/wrap_Window.cpp"
   $79 = ((($78)) + 24|0); //@line 124 "modules/window/wrap_Window.cpp"
   $80 = HEAP32[$79>>2]|0; //@line 124 "modules/window/wrap_Window.cpp"
   __THREW__ = 0;
   $81 = (invoke_iiiii($80|0,($77|0),($1|0),($3|0),($settings|0))|0); //@line 124 "modules/window/wrap_Window.cpp"
   $82 = __THREW__; __THREW__ = 0;
   $83 = $82&1;
   if (!($83)) {
    __THREW__ = 0;
    invoke_vii(3397,($L|0),($81|0)); //@line 124 "modules/window/wrap_Window.cpp"
    $84 = __THREW__; __THREW__ = 0;
    $85 = $84&1;
    if (!($85)) {
     $$3 = 1;
     break;
    }
   }
   $86 = ___cxa_find_matching_catch_3(8992|0)|0;
   $87 = tempRet0;
   $88 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $89 = ($87|0)==($88|0); //@line 535 "./common/runtime.h"
   if (!($89)) {
    $$01$i = $87;$$02$i = $86;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $90 = (___cxa_begin_catch(($86|0))|0); //@line 535 "./common/runtime.h"
   $91 = HEAP32[$90>>2]|0; //@line 539 "./common/runtime.h"
   $92 = ((($91)) + 8|0); //@line 539 "./common/runtime.h"
   $93 = HEAP32[$92>>2]|0; //@line 539 "./common/runtime.h"
   $94 = (FUNCTION_TABLE_ii[$93 & 8191]($90)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($94|0)); //@line 539 "./common/runtime.h"
   $95 = __THREW__; __THREW__ = 0;
   $96 = $95&1;
   if (!($96)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $97 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer4>>2] = $97; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer4)|0); //@line 543 "./common/runtime.h"
    $$3 = 1;
    break;
   }
   $98 = ___cxa_find_matching_catch_2()|0;
   $99 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $100 = __THREW__; __THREW__ = 0;
   $101 = $100&1;
   if ($101) {
    $102 = ___cxa_find_matching_catch_3(0|0)|0;
    $103 = tempRet0;
    ___clang_call_terminate($102); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $99;$$02$i = $98;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  }
 } while(0);
 $$4 = $$3;
 STACKTOP = sp;return ($$4|0); //@line 128 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window9w_getModeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0.0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fstypestr = 0, $h = 0;
 var $name$i = 0, $name$i1 = 0, $name$i10 = 0, $name$i11 = 0, $name$i12 = 0, $name$i13 = 0, $name$i2 = 0, $name$i3 = 0, $name$i4 = 0, $name$i5 = 0, $name$i6 = 0, $name$i7 = 0, $name$i8 = 0, $name$i9 = 0, $settings = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name$i13 = sp + 128|0;
 $name$i12 = sp + 124|0;
 $name$i11 = sp + 120|0;
 $name$i10 = sp + 116|0;
 $name$i9 = sp + 112|0;
 $name$i8 = sp + 108|0;
 $name$i7 = sp + 104|0;
 $name$i6 = sp + 100|0;
 $name$i5 = sp + 96|0;
 $name$i4 = sp + 92|0;
 $name$i3 = sp + 88|0;
 $name$i2 = sp + 84|0;
 $name$i1 = sp + 80|0;
 $name$i = sp + 76|0;
 $w = sp + 72|0;
 $h = sp + 68|0;
 $settings = sp;
 $fstypestr = sp + 64|0;
 HEAP8[$settings>>0] = 0; //@line 204 "./modules/window/Window.h"
 $0 = ((($settings)) + 4|0); //@line 205 "./modules/window/Window.h"
 HEAP32[$0>>2] = 1; //@line 205 "./modules/window/Window.h"
 $1 = ((($settings)) + 8|0); //@line 206 "./modules/window/Window.h"
 HEAP8[$1>>0] = 1; //@line 206 "./modules/window/Window.h"
 $2 = ((($settings)) + 12|0); //@line 207 "./modules/window/Window.h"
 HEAP32[$2>>2] = 0; //@line 207 "./modules/window/Window.h"
 $3 = ((($settings)) + 16|0); //@line 208 "./modules/window/Window.h"
 HEAP8[$3>>0] = 0; //@line 208 "./modules/window/Window.h"
 $4 = ((($settings)) + 20|0); //@line 209 "./modules/window/Window.h"
 HEAP32[$4>>2] = 1; //@line 209 "./modules/window/Window.h"
 $5 = ((($settings)) + 24|0); //@line 210 "./modules/window/Window.h"
 HEAP32[$5>>2] = 1; //@line 210 "./modules/window/Window.h"
 $6 = ((($settings)) + 28|0); //@line 211 "./modules/window/Window.h"
 HEAP8[$6>>0] = 0; //@line 211 "./modules/window/Window.h"
 $7 = ((($settings)) + 29|0); //@line 212 "./modules/window/Window.h"
 HEAP8[$7>>0] = 1; //@line 212 "./modules/window/Window.h"
 $8 = ((($settings)) + 32|0); //@line 213 "./modules/window/Window.h"
 HEAP32[$8>>2] = 0; //@line 213 "./modules/window/Window.h"
 $9 = ((($settings)) + 36|0); //@line 214 "./modules/window/Window.h"
 HEAP8[$9>>0] = 0; //@line 214 "./modules/window/Window.h"
 $10 = ((($settings)) + 40|0); //@line 215 "./modules/window/Window.h"
 HEAPF64[$10>>3] = 0.0; //@line 215 "./modules/window/Window.h"
 $11 = ((($settings)) + 48|0); //@line 216 "./modules/window/Window.h"
 HEAP8[$11>>0] = 0; //@line 216 "./modules/window/Window.h"
 $12 = ((($settings)) + 52|0); //@line 217 "./modules/window/Window.h"
 HEAP32[$12>>2] = 0; //@line 217 "./modules/window/Window.h"
 $13 = ((($settings)) + 56|0); //@line 218 "./modules/window/Window.h"
 HEAP32[$13>>2] = 0; //@line 218 "./modules/window/Window.h"
 $14 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $15 = HEAP32[$14>>2]|0; //@line 134 "modules/window/wrap_Window.cpp"
 $16 = ((($15)) + 28|0); //@line 134 "modules/window/wrap_Window.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 134 "modules/window/wrap_Window.cpp"
 FUNCTION_TABLE_viiii[$17 & 8191]($14,$w,$h,$settings); //@line 134 "modules/window/wrap_Window.cpp"
 $18 = HEAP32[$w>>2]|0; //@line 135 "modules/window/wrap_Window.cpp"
 $19 = (+($18|0)); //@line 135 "modules/window/wrap_Window.cpp"
 _lua_pushnumber($L,$19); //@line 135 "modules/window/wrap_Window.cpp"
 $20 = HEAP32[$h>>2]|0; //@line 136 "modules/window/wrap_Window.cpp"
 $21 = (+($20|0)); //@line 136 "modules/window/wrap_Window.cpp"
 _lua_pushnumber($L,$21); //@line 136 "modules/window/wrap_Window.cpp"
 _lua_createtable($L,0,0); //@line 138 "modules/window/wrap_Window.cpp"
 HEAP32[$fstypestr>>2] = 360894; //@line 140 "modules/window/wrap_Window.cpp"
 $22 = HEAP32[$0>>2]|0; //@line 141 "modules/window/wrap_Window.cpp"
 (__ZN4love6window6Window11getConstantENS1_14FullscreenTypeERPKc($22,$fstypestr)|0); //@line 141 "modules/window/wrap_Window.cpp"
 $23 = HEAP32[$fstypestr>>2]|0; //@line 143 "modules/window/wrap_Window.cpp"
 _lua_pushstring($L,$23); //@line 143 "modules/window/wrap_Window.cpp"
 HEAP32[$name$i2>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
 (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(1,$name$i2)|0); //@line 51 "modules/window/wrap_Window.cpp"
 $24 = HEAP32[$name$i2>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
 _lua_setfield($L,-2,$24); //@line 144 "modules/window/wrap_Window.cpp"
 $25 = HEAP8[$settings>>0]|0; //@line 146 "modules/window/wrap_Window.cpp"
 $26 = ($25<<24>>24)!=(0); //@line 146 "modules/window/wrap_Window.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$26); //@line 146 "modules/window/wrap_Window.cpp"
 HEAP32[$name$i3>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
 (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(0,$name$i3)|0); //@line 51 "modules/window/wrap_Window.cpp"
 $27 = HEAP32[$name$i3>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
 _lua_setfield($L,-2,$27); //@line 147 "modules/window/wrap_Window.cpp"
 $28 = HEAP8[$1>>0]|0; //@line 149 "modules/window/wrap_Window.cpp"
 $29 = ($28<<24>>24)!=(0); //@line 149 "modules/window/wrap_Window.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$29); //@line 149 "modules/window/wrap_Window.cpp"
 HEAP32[$name$i4>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
 (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(2,$name$i4)|0); //@line 51 "modules/window/wrap_Window.cpp"
 $30 = HEAP32[$name$i4>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
 _lua_setfield($L,-2,$30); //@line 150 "modules/window/wrap_Window.cpp"
 $31 = HEAP32[$2>>2]|0; //@line 152 "modules/window/wrap_Window.cpp"
 _lua_pushinteger($L,$31); //@line 152 "modules/window/wrap_Window.cpp"
 HEAP32[$name$i5>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
 (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(3,$name$i5)|0); //@line 51 "modules/window/wrap_Window.cpp"
 $32 = HEAP32[$name$i5>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
 _lua_setfield($L,-2,$32); //@line 153 "modules/window/wrap_Window.cpp"
 $33 = HEAP8[$3>>0]|0; //@line 155 "modules/window/wrap_Window.cpp"
 $34 = ($33<<24>>24)!=(0); //@line 155 "modules/window/wrap_Window.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$34); //@line 155 "modules/window/wrap_Window.cpp"
 HEAP32[$name$i6>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
 (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(4,$name$i6)|0); //@line 51 "modules/window/wrap_Window.cpp"
 $35 = HEAP32[$name$i6>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
 _lua_setfield($L,-2,$35); //@line 156 "modules/window/wrap_Window.cpp"
 $36 = HEAP32[$4>>2]|0; //@line 158 "modules/window/wrap_Window.cpp"
 _lua_pushinteger($L,$36); //@line 158 "modules/window/wrap_Window.cpp"
 HEAP32[$name$i7>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
 (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(5,$name$i7)|0); //@line 51 "modules/window/wrap_Window.cpp"
 $37 = HEAP32[$name$i7>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
 _lua_setfield($L,-2,$37); //@line 159 "modules/window/wrap_Window.cpp"
 $38 = HEAP32[$5>>2]|0; //@line 161 "modules/window/wrap_Window.cpp"
 _lua_pushinteger($L,$38); //@line 161 "modules/window/wrap_Window.cpp"
 HEAP32[$name$i8>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
 (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(6,$name$i8)|0); //@line 51 "modules/window/wrap_Window.cpp"
 $39 = HEAP32[$name$i8>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
 _lua_setfield($L,-2,$39); //@line 162 "modules/window/wrap_Window.cpp"
 $40 = HEAP8[$6>>0]|0; //@line 164 "modules/window/wrap_Window.cpp"
 $41 = ($40<<24>>24)!=(0); //@line 164 "modules/window/wrap_Window.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$41); //@line 164 "modules/window/wrap_Window.cpp"
 HEAP32[$name$i9>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
 (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(7,$name$i9)|0); //@line 51 "modules/window/wrap_Window.cpp"
 $42 = HEAP32[$name$i9>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
 _lua_setfield($L,-2,$42); //@line 165 "modules/window/wrap_Window.cpp"
 $43 = HEAP8[$7>>0]|0; //@line 167 "modules/window/wrap_Window.cpp"
 $44 = ($43<<24>>24)!=(0); //@line 167 "modules/window/wrap_Window.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$44); //@line 167 "modules/window/wrap_Window.cpp"
 HEAP32[$name$i10>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
 (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(8,$name$i10)|0); //@line 51 "modules/window/wrap_Window.cpp"
 $45 = HEAP32[$name$i10>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
 _lua_setfield($L,-2,$45); //@line 168 "modules/window/wrap_Window.cpp"
 $46 = HEAP32[$8>>2]|0; //@line 171 "modules/window/wrap_Window.cpp"
 $47 = (($46) + 1)|0; //@line 171 "modules/window/wrap_Window.cpp"
 _lua_pushinteger($L,$47); //@line 171 "modules/window/wrap_Window.cpp"
 HEAP32[$name$i11>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
 (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(9,$name$i11)|0); //@line 51 "modules/window/wrap_Window.cpp"
 $48 = HEAP32[$name$i11>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
 _lua_setfield($L,-2,$48); //@line 172 "modules/window/wrap_Window.cpp"
 $49 = HEAP8[$9>>0]|0; //@line 174 "modules/window/wrap_Window.cpp"
 $50 = ($49<<24>>24)!=(0); //@line 174 "modules/window/wrap_Window.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$50); //@line 174 "modules/window/wrap_Window.cpp"
 HEAP32[$name$i12>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
 (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(10,$name$i12)|0); //@line 51 "modules/window/wrap_Window.cpp"
 $51 = HEAP32[$name$i12>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
 _lua_setfield($L,-2,$51); //@line 175 "modules/window/wrap_Window.cpp"
 $52 = +HEAPF64[$10>>3]; //@line 177 "modules/window/wrap_Window.cpp"
 _lua_pushnumber($L,$52); //@line 177 "modules/window/wrap_Window.cpp"
 HEAP32[$name$i13>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
 (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(11,$name$i13)|0); //@line 51 "modules/window/wrap_Window.cpp"
 $53 = HEAP32[$name$i13>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
 _lua_setfield($L,-2,$53); //@line 178 "modules/window/wrap_Window.cpp"
 $54 = HEAP32[$12>>2]|0; //@line 180 "modules/window/wrap_Window.cpp"
 _lua_pushinteger($L,$54); //@line 180 "modules/window/wrap_Window.cpp"
 HEAP32[$name$i1>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
 (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(12,$name$i1)|0); //@line 51 "modules/window/wrap_Window.cpp"
 $55 = HEAP32[$name$i1>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
 _lua_setfield($L,-2,$55); //@line 181 "modules/window/wrap_Window.cpp"
 $56 = HEAP32[$13>>2]|0; //@line 183 "modules/window/wrap_Window.cpp"
 _lua_pushinteger($L,$56); //@line 183 "modules/window/wrap_Window.cpp"
 HEAP32[$name$i>>2] = 0; //@line 50 "modules/window/wrap_Window.cpp"
 (__ZN4love6window6Window11getConstantENS1_7SettingERPKc(13,$name$i)|0); //@line 51 "modules/window/wrap_Window.cpp"
 $57 = HEAP32[$name$i>>2]|0; //@line 52 "modules/window/wrap_Window.cpp"
 _lua_setfield($L,-2,$57); //@line 184 "modules/window/wrap_Window.cpp"
 STACKTOP = sp;return 3; //@line 186 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window20w_getFullscreenModesEP9lua_State($L) {
 $L = $L|0;
 var $$lcssa = 0, $$lcssa21 = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $displayindex = 0, $i$0 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $modes = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i1 = 0;
 var $scevgep4$i$i$i = 0, $scevgep4$i$i$i2 = 0, $x = 0, $y = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $displayindex = sp + 20|0;
 $x = sp + 16|0;
 $y = sp + 12|0;
 $modes = sp;
 HEAP32[$displayindex>>2] = 0; //@line 191 "modules/window/wrap_Window.cpp"
 $0 = (_lua_type($L,1)|0); //@line 192 "modules/window/wrap_Window.cpp"
 $1 = ($0|0)<(1); //@line 192 "modules/window/wrap_Window.cpp"
 if ($1) {
  $5 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
  $6 = HEAP32[$5>>2]|0; //@line 197 "modules/window/wrap_Window.cpp"
  $7 = ((($6)) + 68|0); //@line 197 "modules/window/wrap_Window.cpp"
  $8 = HEAP32[$7>>2]|0; //@line 197 "modules/window/wrap_Window.cpp"
  FUNCTION_TABLE_viiii[$8 & 8191]($5,$x,$y,$displayindex); //@line 197 "modules/window/wrap_Window.cpp"
  $$pre = HEAP32[$displayindex>>2]|0; //@line 200 "modules/window/wrap_Window.cpp"
  $13 = $$pre;
 } else {
  $2 = (+_luaL_checknumber($L,1)); //@line 193 "modules/window/wrap_Window.cpp"
  $3 = (~~(($2))); //@line 193 "modules/window/wrap_Window.cpp"
  $4 = (($3) + -1)|0; //@line 193 "modules/window/wrap_Window.cpp"
  HEAP32[$displayindex>>2] = $4; //@line 193 "modules/window/wrap_Window.cpp"
  $13 = $4;
 }
 $9 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $10 = HEAP32[$9>>2]|0; //@line 200 "modules/window/wrap_Window.cpp"
 $11 = ((($10)) + 56|0); //@line 200 "modules/window/wrap_Window.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 200 "modules/window/wrap_Window.cpp"
 FUNCTION_TABLE_viii[$12 & 8191]($modes,$9,$13); //@line 200 "modules/window/wrap_Window.cpp"
 $14 = ((($modes)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = HEAP32[$14>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = HEAP32[$modes>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = (($15) - ($16))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $18 = $17 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 __THREW__ = 0;
 invoke_viii(3393,($L|0),($18|0),0); //@line 202 "modules/window/wrap_Window.cpp"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 do {
  if ($20) {
   $37 = ___cxa_find_matching_catch_2()|0;
   $38 = tempRet0;
   $$pre$phiZ2D = $modes;$lpad$phi$index = $37;$lpad$phi$index2 = $38;
  } else {
   $i$0 = 0;
   while(1) {
    $21 = HEAP32[$14>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $22 = HEAP32[$modes>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $23 = (($21) - ($22))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $24 = $23 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $25 = ($i$0>>>0)<($24>>>0); //@line 204 "modules/window/wrap_Window.cpp"
    if (!($25)) {
     $$lcssa = $21;$$lcssa21 = $22;
     break;
    }
    $48 = (($i$0) + 1)|0; //@line 206 "modules/window/wrap_Window.cpp"
    __THREW__ = 0;
    invoke_vii(3468,($L|0),($48|0)); //@line 206 "modules/window/wrap_Window.cpp"
    $49 = __THREW__; __THREW__ = 0;
    $50 = $49&1;
    if ($50) {
     label = 12;
     break;
    }
    __THREW__ = 0;
    invoke_viii(3393,($L|0),0,2); //@line 207 "modules/window/wrap_Window.cpp"
    $51 = __THREW__; __THREW__ = 0;
    $52 = $51&1;
    if ($52) {
     label = 12;
     break;
    }
    $53 = HEAP32[$modes>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $54 = (($53) + ($i$0<<3)|0); //@line 211 "modules/window/wrap_Window.cpp"
    $55 = HEAP32[$54>>2]|0; //@line 211 "modules/window/wrap_Window.cpp"
    __THREW__ = 0;
    invoke_vii(3468,($L|0),($55|0)); //@line 211 "modules/window/wrap_Window.cpp"
    $56 = __THREW__; __THREW__ = 0;
    $57 = $56&1;
    if ($57) {
     label = 12;
     break;
    }
    __THREW__ = 0;
    invoke_viii(3469,($L|0),-2,(609941|0)); //@line 212 "modules/window/wrap_Window.cpp"
    $58 = __THREW__; __THREW__ = 0;
    $59 = $58&1;
    if ($59) {
     label = 12;
     break;
    }
    $60 = HEAP32[$modes>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $61 = (((($60) + ($i$0<<3)|0)) + 4|0); //@line 214 "modules/window/wrap_Window.cpp"
    $62 = HEAP32[$61>>2]|0; //@line 214 "modules/window/wrap_Window.cpp"
    __THREW__ = 0;
    invoke_vii(3468,($L|0),($62|0)); //@line 214 "modules/window/wrap_Window.cpp"
    $63 = __THREW__; __THREW__ = 0;
    $64 = $63&1;
    if ($64) {
     label = 12;
     break;
    }
    __THREW__ = 0;
    invoke_viii(3469,($L|0),-2,(609947|0)); //@line 215 "modules/window/wrap_Window.cpp"
    $65 = __THREW__; __THREW__ = 0;
    $66 = $65&1;
    if ($66) {
     label = 12;
     break;
    }
    __THREW__ = 0;
    invoke_vii(3470,($L|0),-3); //@line 219 "modules/window/wrap_Window.cpp"
    $67 = __THREW__; __THREW__ = 0;
    $68 = $67&1;
    if ($68) {
     label = 12;
     break;
    } else {
     $i$0 = $48;
    }
   }
   if ((label|0) == 12) {
    $35 = ___cxa_find_matching_catch_2()|0;
    $36 = tempRet0;
    $$pre$phiZ2D = $modes;$lpad$phi$index = $35;$lpad$phi$index2 = $36;
    break;
   }
   $26 = $$lcssa; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $27 = ($$lcssa21|0)==(0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $28 = $$lcssa21; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($27) {
    STACKTOP = sp;return 1; //@line 223 "modules/window/wrap_Window.cpp"
   }
   $29 = $$lcssa21; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $30 = ($26|0)==($29|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($30)) {
    $scevgep$i$i$i1 = ((($26)) + -8|0);
    $31 = $scevgep$i$i$i1;
    $32 = (($31) - ($$lcssa21))|0;
    $33 = $32 >>> 3;
    $34 = $33 ^ -1;
    $scevgep4$i$i$i2 = (($26) + ($34<<3)|0);
    HEAP32[$14>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($28); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   STACKTOP = sp;return 1; //@line 223 "modules/window/wrap_Window.cpp"
  }
 } while(0);
 $39 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $40 = ($39|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $41 = $39; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($40) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $42 = HEAP32[$14>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $43 = ($42|0)==($39|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($43)) {
  $scevgep$i$i$i = ((($42)) + -8|0);
  $44 = $scevgep$i$i$i;
  $45 = (($44) - ($41))|0;
  $46 = $45 >>> 3;
  $47 = $46 ^ -1;
  $scevgep4$i$i$i = (($42) + ($47<<3)|0);
  HEAP32[$14>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($39); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love6window15w_setFullscreenEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $fstype = 0, $success$0$in = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $fstype = sp + 4|0;
 $0 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,1)|0); //@line 227 "modules/window/wrap_Window.cpp"
 HEAP32[$fstype>>2] = 2; //@line 228 "modules/window/wrap_Window.cpp"
 $1 = (_lua_type($L,2)|0); //@line 230 "modules/window/wrap_Window.cpp"
 $2 = ($1|0)<(1); //@line 230 "modules/window/wrap_Window.cpp"
 do {
  if ($2) {
   label = 5;
  } else {
   $3 = (_luaL_checklstring($L,2,0)|0); //@line 230 "modules/window/wrap_Window.cpp"
   $4 = ($3|0)==(0|0); //@line 231 "modules/window/wrap_Window.cpp"
   if ($4) {
    label = 5;
   } else {
    $5 = (__ZN4love6window6Window11getConstantEPKcRNS1_14FullscreenTypeE($3,$fstype)|0); //@line 231 "modules/window/wrap_Window.cpp"
    if ($5) {
     $$pre = HEAP32[$fstype>>2]|0; //@line 235 "modules/window/wrap_Window.cpp"
     $8 = ($$pre|0)==(2); //@line 235 "modules/window/wrap_Window.cpp"
     $9 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
     if ($8) {
      $11 = $9;
      label = 7;
      break;
     }
     $15 = HEAP32[$9>>2]|0; //@line 238 "modules/window/wrap_Window.cpp"
     $16 = ((($15)) + 36|0); //@line 238 "modules/window/wrap_Window.cpp"
     $17 = HEAP32[$16>>2]|0; //@line 238 "modules/window/wrap_Window.cpp"
     $18 = (FUNCTION_TABLE_iiii[$17 & 8191]($9,$0,$$pre)|0); //@line 238 "modules/window/wrap_Window.cpp"
     $success$0$in = $18;
     break;
    } else {
     HEAP32[$vararg_buffer>>2] = $3; //@line 232 "modules/window/wrap_Window.cpp"
     $6 = (_luaL_error($L,266198,$vararg_buffer)|0); //@line 232 "modules/window/wrap_Window.cpp"
     $$0 = $6;
     STACKTOP = sp;return ($$0|0); //@line 242 "modules/window/wrap_Window.cpp"
    }
   }
  }
 } while(0);
 if ((label|0) == 5) {
  $7 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
  $11 = $7;
  label = 7;
 }
 if ((label|0) == 7) {
  $10 = HEAP32[$11>>2]|0; //@line 236 "modules/window/wrap_Window.cpp"
  $12 = ((($10)) + 40|0); //@line 236 "modules/window/wrap_Window.cpp"
  $13 = HEAP32[$12>>2]|0; //@line 236 "modules/window/wrap_Window.cpp"
  $14 = (FUNCTION_TABLE_iii[$13 & 8191]($11,$0)|0); //@line 236 "modules/window/wrap_Window.cpp"
  $success$0$in = $14;
 }
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$success$0$in); //@line 240 "modules/window/wrap_Window.cpp"
 $$0 = 1;
 STACKTOP = sp;return ($$0|0); //@line 242 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window15w_getFullscreenEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $h = 0, $settings = 0, $typestr = 0, $vararg_buffer = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp + 64|0;
 $w = sp + 76|0;
 $h = sp + 72|0;
 $settings = sp;
 $typestr = sp + 68|0;
 HEAP8[$settings>>0] = 0; //@line 204 "./modules/window/Window.h"
 $0 = ((($settings)) + 4|0); //@line 205 "./modules/window/Window.h"
 HEAP32[$0>>2] = 1; //@line 205 "./modules/window/Window.h"
 $1 = ((($settings)) + 8|0); //@line 206 "./modules/window/Window.h"
 HEAP8[$1>>0] = 1; //@line 206 "./modules/window/Window.h"
 $2 = ((($settings)) + 12|0); //@line 207 "./modules/window/Window.h"
 HEAP32[$2>>2] = 0; //@line 207 "./modules/window/Window.h"
 $3 = ((($settings)) + 16|0); //@line 208 "./modules/window/Window.h"
 HEAP8[$3>>0] = 0; //@line 208 "./modules/window/Window.h"
 $4 = ((($settings)) + 20|0); //@line 209 "./modules/window/Window.h"
 HEAP32[$4>>2] = 1; //@line 209 "./modules/window/Window.h"
 $5 = ((($settings)) + 24|0); //@line 210 "./modules/window/Window.h"
 HEAP32[$5>>2] = 1; //@line 210 "./modules/window/Window.h"
 $6 = ((($settings)) + 28|0); //@line 211 "./modules/window/Window.h"
 HEAP8[$6>>0] = 0; //@line 211 "./modules/window/Window.h"
 $7 = ((($settings)) + 29|0); //@line 212 "./modules/window/Window.h"
 HEAP8[$7>>0] = 1; //@line 212 "./modules/window/Window.h"
 $8 = ((($settings)) + 32|0); //@line 213 "./modules/window/Window.h"
 HEAP32[$8>>2] = 0; //@line 213 "./modules/window/Window.h"
 $9 = ((($settings)) + 36|0); //@line 214 "./modules/window/Window.h"
 HEAP8[$9>>0] = 0; //@line 214 "./modules/window/Window.h"
 $10 = ((($settings)) + 40|0); //@line 215 "./modules/window/Window.h"
 HEAPF64[$10>>3] = 0.0; //@line 215 "./modules/window/Window.h"
 $11 = ((($settings)) + 48|0); //@line 216 "./modules/window/Window.h"
 HEAP8[$11>>0] = 0; //@line 216 "./modules/window/Window.h"
 $12 = ((($settings)) + 52|0); //@line 217 "./modules/window/Window.h"
 HEAP32[$12>>2] = 0; //@line 217 "./modules/window/Window.h"
 $13 = ((($settings)) + 56|0); //@line 218 "./modules/window/Window.h"
 HEAP32[$13>>2] = 0; //@line 218 "./modules/window/Window.h"
 $14 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $15 = HEAP32[$14>>2]|0; //@line 248 "modules/window/wrap_Window.cpp"
 $16 = ((($15)) + 28|0); //@line 248 "modules/window/wrap_Window.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 248 "modules/window/wrap_Window.cpp"
 FUNCTION_TABLE_viiii[$17 & 8191]($14,$w,$h,$settings); //@line 248 "modules/window/wrap_Window.cpp"
 $18 = HEAP32[$0>>2]|0; //@line 251 "modules/window/wrap_Window.cpp"
 $19 = (__ZN4love6window6Window11getConstantENS1_14FullscreenTypeERPKc($18,$typestr)|0); //@line 251 "modules/window/wrap_Window.cpp"
 if ($19) {
  $20 = HEAP8[$settings>>0]|0; //@line 254 "modules/window/wrap_Window.cpp"
  $21 = ($20<<24>>24)!=(0); //@line 254 "modules/window/wrap_Window.cpp"
  __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$21); //@line 254 "modules/window/wrap_Window.cpp"
  $22 = HEAP32[$typestr>>2]|0; //@line 255 "modules/window/wrap_Window.cpp"
  _lua_pushstring($L,$22); //@line 255 "modules/window/wrap_Window.cpp"
  STACKTOP = sp;return 2; //@line 256 "modules/window/wrap_Window.cpp"
 }
 (_luaL_error($L,266226,$vararg_buffer)|0); //@line 252 "modules/window/wrap_Window.cpp"
 $20 = HEAP8[$settings>>0]|0; //@line 254 "modules/window/wrap_Window.cpp"
 $21 = ($20<<24>>24)!=(0); //@line 254 "modules/window/wrap_Window.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$21); //@line 254 "modules/window/wrap_Window.cpp"
 $22 = HEAP32[$typestr>>2]|0; //@line 255 "modules/window/wrap_Window.cpp"
 _lua_pushstring($L,$22); //@line 255 "modules/window/wrap_Window.cpp"
 STACKTOP = sp;return 2; //@line 256 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window8w_isOpenEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 261 "modules/window/wrap_Window.cpp"
 $2 = ((($1)) + 72|0); //@line 261 "modules/window/wrap_Window.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 261 "modules/window/wrap_Window.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 261 "modules/window/wrap_Window.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 261 "modules/window/wrap_Window.cpp"
 return 1; //@line 262 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window7w_closeEP9lua_State($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 267 "modules/window/wrap_Window.cpp"
 $3 = ((($2)) + 32|0); //@line 267 "modules/window/wrap_Window.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 267 "modules/window/wrap_Window.cpp"
 FUNCTION_TABLE_vi[$4 & 8191]($1); //@line 267 "modules/window/wrap_Window.cpp"
 return 0; //@line 268 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window22w_getDesktopDimensionsEP9lua_State($L) {
 $L = $L|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0.0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $displayindex = 0, $height = 0, $width = 0;
 var $x = 0, $y = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $width = sp + 16|0;
 $height = sp + 12|0;
 $displayindex = sp + 8|0;
 $x = sp + 4|0;
 $y = sp;
 HEAP32[$width>>2] = 0; //@line 273 "modules/window/wrap_Window.cpp"
 HEAP32[$height>>2] = 0; //@line 273 "modules/window/wrap_Window.cpp"
 HEAP32[$displayindex>>2] = 0; //@line 274 "modules/window/wrap_Window.cpp"
 $0 = (_lua_type($L,1)|0); //@line 275 "modules/window/wrap_Window.cpp"
 $1 = ($0|0)<(1); //@line 275 "modules/window/wrap_Window.cpp"
 if ($1) {
  $5 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
  $6 = HEAP32[$5>>2]|0; //@line 280 "modules/window/wrap_Window.cpp"
  $7 = ((($6)) + 68|0); //@line 280 "modules/window/wrap_Window.cpp"
  $8 = HEAP32[$7>>2]|0; //@line 280 "modules/window/wrap_Window.cpp"
  FUNCTION_TABLE_viiii[$8 & 8191]($5,$x,$y,$displayindex); //@line 280 "modules/window/wrap_Window.cpp"
  $$pre = HEAP32[$displayindex>>2]|0; //@line 282 "modules/window/wrap_Window.cpp"
  $13 = $$pre;
 } else {
  $2 = (+_luaL_checknumber($L,1)); //@line 276 "modules/window/wrap_Window.cpp"
  $3 = (~~(($2))); //@line 276 "modules/window/wrap_Window.cpp"
  $4 = (($3) + -1)|0; //@line 276 "modules/window/wrap_Window.cpp"
  HEAP32[$displayindex>>2] = $4; //@line 276 "modules/window/wrap_Window.cpp"
  $13 = $4;
 }
 $9 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $10 = HEAP32[$9>>2]|0; //@line 282 "modules/window/wrap_Window.cpp"
 $11 = ((($10)) + 60|0); //@line 282 "modules/window/wrap_Window.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 282 "modules/window/wrap_Window.cpp"
 FUNCTION_TABLE_viiii[$12 & 8191]($9,$13,$width,$height); //@line 282 "modules/window/wrap_Window.cpp"
 $14 = HEAP32[$width>>2]|0; //@line 283 "modules/window/wrap_Window.cpp"
 _lua_pushinteger($L,$14); //@line 283 "modules/window/wrap_Window.cpp"
 $15 = HEAP32[$height>>2]|0; //@line 284 "modules/window/wrap_Window.cpp"
 _lua_pushinteger($L,$15); //@line 284 "modules/window/wrap_Window.cpp"
 STACKTOP = sp;return 2; //@line 285 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window13w_setPositionEP9lua_State($L) {
 $L = $L|0;
 var $$pre = 0, $0 = 0.0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0.0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, $displayindex = 0;
 var $x_unused = 0, $y_unused = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $displayindex = sp + 8|0;
 $x_unused = sp + 4|0;
 $y_unused = sp;
 $0 = (+_luaL_checknumber($L,1)); //@line 290 "modules/window/wrap_Window.cpp"
 $1 = (~~(($0))); //@line 290 "modules/window/wrap_Window.cpp"
 $2 = (+_luaL_checknumber($L,2)); //@line 291 "modules/window/wrap_Window.cpp"
 $3 = (~~(($2))); //@line 291 "modules/window/wrap_Window.cpp"
 HEAP32[$displayindex>>2] = 0; //@line 293 "modules/window/wrap_Window.cpp"
 $4 = (_lua_type($L,3)|0); //@line 294 "modules/window/wrap_Window.cpp"
 $5 = ($4|0)<(1); //@line 294 "modules/window/wrap_Window.cpp"
 if ($5) {
  $9 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
  $10 = HEAP32[$9>>2]|0; //@line 299 "modules/window/wrap_Window.cpp"
  $11 = ((($10)) + 68|0); //@line 299 "modules/window/wrap_Window.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 299 "modules/window/wrap_Window.cpp"
  FUNCTION_TABLE_viiii[$12 & 8191]($9,$x_unused,$y_unused,$displayindex); //@line 299 "modules/window/wrap_Window.cpp"
  $$pre = HEAP32[$displayindex>>2]|0; //@line 302 "modules/window/wrap_Window.cpp"
  $17 = $$pre;
 } else {
  $6 = (+_luaL_checknumber($L,3)); //@line 295 "modules/window/wrap_Window.cpp"
  $7 = (~~(($6))); //@line 295 "modules/window/wrap_Window.cpp"
  $8 = (($7) + -1)|0; //@line 295 "modules/window/wrap_Window.cpp"
  HEAP32[$displayindex>>2] = $8; //@line 295 "modules/window/wrap_Window.cpp"
  $17 = $8;
 }
 $13 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $14 = HEAP32[$13>>2]|0; //@line 302 "modules/window/wrap_Window.cpp"
 $15 = ((($14)) + 64|0); //@line 302 "modules/window/wrap_Window.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 302 "modules/window/wrap_Window.cpp"
 FUNCTION_TABLE_viiii[$16 & 8191]($13,$1,$3,$17); //@line 302 "modules/window/wrap_Window.cpp"
 STACKTOP = sp;return 0; //@line 303 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window13w_getPositionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $displayindex = 0, $x = 0, $y = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $x = sp + 8|0;
 $y = sp + 4|0;
 $displayindex = sp;
 HEAP32[$x>>2] = 0; //@line 308 "modules/window/wrap_Window.cpp"
 HEAP32[$y>>2] = 0; //@line 309 "modules/window/wrap_Window.cpp"
 HEAP32[$displayindex>>2] = 0; //@line 310 "modules/window/wrap_Window.cpp"
 $0 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 311 "modules/window/wrap_Window.cpp"
 $2 = ((($1)) + 68|0); //@line 311 "modules/window/wrap_Window.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 311 "modules/window/wrap_Window.cpp"
 FUNCTION_TABLE_viiii[$3 & 8191]($0,$x,$y,$displayindex); //@line 311 "modules/window/wrap_Window.cpp"
 $4 = HEAP32[$x>>2]|0; //@line 312 "modules/window/wrap_Window.cpp"
 _lua_pushinteger($L,$4); //@line 312 "modules/window/wrap_Window.cpp"
 $5 = HEAP32[$y>>2]|0; //@line 313 "modules/window/wrap_Window.cpp"
 _lua_pushinteger($L,$5); //@line 313 "modules/window/wrap_Window.cpp"
 $6 = HEAP32[$displayindex>>2]|0; //@line 314 "modules/window/wrap_Window.cpp"
 $7 = (($6) + 1)|0; //@line 314 "modules/window/wrap_Window.cpp"
 _lua_pushinteger($L,$7); //@line 314 "modules/window/wrap_Window.cpp"
 STACKTOP = sp;return 3; //@line 315 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window9w_setIconEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5image9ImageDataEEEPT_P9lua_StateiNS_4TypeE($L,1,22)|0); //@line 320 "modules/window/wrap_Window.cpp"
 $1 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 321 "modules/window/wrap_Window.cpp"
 $3 = ((($2)) + 84|0); //@line 321 "modules/window/wrap_Window.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 321 "modules/window/wrap_Window.cpp"
 $5 = (FUNCTION_TABLE_iii[$4 & 8191]($1,$0)|0); //@line 321 "modules/window/wrap_Window.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$5); //@line 321 "modules/window/wrap_Window.cpp"
 return 1; //@line 322 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window9w_getIconEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 327 "modules/window/wrap_Window.cpp"
 $2 = ((($1)) + 88|0); //@line 327 "modules/window/wrap_Window.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 327 "modules/window/wrap_Window.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 327 "modules/window/wrap_Window.cpp"
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,22,$4); //@line 328 "modules/window/wrap_Window.cpp"
 return 1; //@line 329 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window24w_setDisplaySleepEnabledEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 334 "modules/window/wrap_Window.cpp"
 $2 = ((($1)) + 92|0); //@line 334 "modules/window/wrap_Window.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 334 "modules/window/wrap_Window.cpp"
 $4 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,1)|0); //@line 334 "modules/window/wrap_Window.cpp"
 FUNCTION_TABLE_vii[$3 & 8191]($0,$4); //@line 334 "modules/window/wrap_Window.cpp"
 return 0; //@line 335 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window23w_isDisplaySleepEnabledEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 340 "modules/window/wrap_Window.cpp"
 $2 = ((($1)) + 96|0); //@line 340 "modules/window/wrap_Window.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 340 "modules/window/wrap_Window.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 340 "modules/window/wrap_Window.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 340 "modules/window/wrap_Window.cpp"
 return 1; //@line 341 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window10w_setTitleEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $title = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $title = sp;
 __ZN4love16luax_checkstringEP9lua_Statei($title,$L,1); //@line 346 "modules/window/wrap_Window.cpp"
 $0 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 347 "modules/window/wrap_Window.cpp"
 $2 = ((($1)) + 76|0); //@line 347 "modules/window/wrap_Window.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 347 "modules/window/wrap_Window.cpp"
 __THREW__ = 0;
 invoke_vii($3|0,($0|0),($title|0)); //@line 347 "modules/window/wrap_Window.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($title); //@line 349 "modules/window/wrap_Window.cpp"
  ___resumeException($6|0);
  // unreachable;
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($title); //@line 349 "modules/window/wrap_Window.cpp"
  STACKTOP = sp;return 0; //@line 349 "modules/window/wrap_Window.cpp"
 }
 return (0)|0;
}
function __ZN4love6window10w_getTitleEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 353 "modules/window/wrap_Window.cpp"
 $2 = ((($1)) + 80|0); //@line 353 "modules/window/wrap_Window.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 353 "modules/window/wrap_Window.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 353 "modules/window/wrap_Window.cpp"
 __ZN4love15luax_pushstringEP9lua_StateRKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE($L,$4); //@line 353 "modules/window/wrap_Window.cpp"
 return 1; //@line 354 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window10w_hasFocusEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 359 "modules/window/wrap_Window.cpp"
 $2 = ((($1)) + 112|0); //@line 359 "modules/window/wrap_Window.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 359 "modules/window/wrap_Window.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 359 "modules/window/wrap_Window.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 359 "modules/window/wrap_Window.cpp"
 return 1; //@line 360 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window15w_hasMouseFocusEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 365 "modules/window/wrap_Window.cpp"
 $2 = ((($1)) + 116|0); //@line 365 "modules/window/wrap_Window.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 365 "modules/window/wrap_Window.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 365 "modules/window/wrap_Window.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 365 "modules/window/wrap_Window.cpp"
 return 1; //@line 366 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window11w_isVisibleEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 371 "modules/window/wrap_Window.cpp"
 $2 = ((($1)) + 120|0); //@line 371 "modules/window/wrap_Window.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 371 "modules/window/wrap_Window.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 371 "modules/window/wrap_Window.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 371 "modules/window/wrap_Window.cpp"
 return 1; //@line 372 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window15w_getPixelScaleEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 377 "modules/window/wrap_Window.cpp"
 $2 = ((($1)) + 152|0); //@line 377 "modules/window/wrap_Window.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 377 "modules/window/wrap_Window.cpp"
 $4 = (+FUNCTION_TABLE_di[$3 & 4095]($0)); //@line 377 "modules/window/wrap_Window.cpp"
 _lua_pushnumber($L,$4); //@line 377 "modules/window/wrap_Window.cpp"
 return 1; //@line 378 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window10w_toPixelsEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0.0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0, $px = 0, $py = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $px = sp + 8|0;
 $py = sp;
 $0 = (+_luaL_checknumber($L,1)); //@line 383 "modules/window/wrap_Window.cpp"
 $1 = (_lua_type($L,2)|0); //@line 385 "modules/window/wrap_Window.cpp"
 $2 = ($1|0)<(1); //@line 385 "modules/window/wrap_Window.cpp"
 if ($2) {
  $3 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
  $4 = HEAP32[$3>>2]|0; //@line 387 "modules/window/wrap_Window.cpp"
  $5 = ((($4)) + 156|0); //@line 387 "modules/window/wrap_Window.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 387 "modules/window/wrap_Window.cpp"
  $7 = (+FUNCTION_TABLE_did[$6 & 2047]($3,$0)); //@line 387 "modules/window/wrap_Window.cpp"
  _lua_pushnumber($L,$7); //@line 387 "modules/window/wrap_Window.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 400 "modules/window/wrap_Window.cpp"
 } else {
  $8 = (+_luaL_checknumber($L,2)); //@line 391 "modules/window/wrap_Window.cpp"
  HEAPF64[$px>>3] = 0.0; //@line 392 "modules/window/wrap_Window.cpp"
  HEAPF64[$py>>3] = 0.0; //@line 392 "modules/window/wrap_Window.cpp"
  $9 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
  $10 = HEAP32[$9>>2]|0; //@line 394 "modules/window/wrap_Window.cpp"
  $11 = ((($10)) + 160|0); //@line 394 "modules/window/wrap_Window.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 394 "modules/window/wrap_Window.cpp"
  FUNCTION_TABLE_viddii[$12 & 4095]($9,$0,$8,$px,$py); //@line 394 "modules/window/wrap_Window.cpp"
  $13 = +HEAPF64[$px>>3]; //@line 396 "modules/window/wrap_Window.cpp"
  _lua_pushnumber($L,$13); //@line 396 "modules/window/wrap_Window.cpp"
  $14 = +HEAPF64[$py>>3]; //@line 397 "modules/window/wrap_Window.cpp"
  _lua_pushnumber($L,$14); //@line 397 "modules/window/wrap_Window.cpp"
  $$0 = 2;
  STACKTOP = sp;return ($$0|0); //@line 400 "modules/window/wrap_Window.cpp"
 }
 return (0)|0;
}
function __ZN4love6window12w_fromPixelsEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0.0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0, $wx = 0, $wy = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $wx = sp + 8|0;
 $wy = sp;
 $0 = (+_luaL_checknumber($L,1)); //@line 404 "modules/window/wrap_Window.cpp"
 $1 = (_lua_type($L,2)|0); //@line 406 "modules/window/wrap_Window.cpp"
 $2 = ($1|0)<(1); //@line 406 "modules/window/wrap_Window.cpp"
 if ($2) {
  $3 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
  $4 = HEAP32[$3>>2]|0; //@line 408 "modules/window/wrap_Window.cpp"
  $5 = ((($4)) + 164|0); //@line 408 "modules/window/wrap_Window.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 408 "modules/window/wrap_Window.cpp"
  $7 = (+FUNCTION_TABLE_did[$6 & 2047]($3,$0)); //@line 408 "modules/window/wrap_Window.cpp"
  _lua_pushnumber($L,$7); //@line 408 "modules/window/wrap_Window.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 421 "modules/window/wrap_Window.cpp"
 } else {
  $8 = (+_luaL_checknumber($L,2)); //@line 412 "modules/window/wrap_Window.cpp"
  HEAPF64[$wx>>3] = 0.0; //@line 413 "modules/window/wrap_Window.cpp"
  HEAPF64[$wy>>3] = 0.0; //@line 413 "modules/window/wrap_Window.cpp"
  $9 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
  $10 = HEAP32[$9>>2]|0; //@line 415 "modules/window/wrap_Window.cpp"
  $11 = ((($10)) + 168|0); //@line 415 "modules/window/wrap_Window.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 415 "modules/window/wrap_Window.cpp"
  FUNCTION_TABLE_viddii[$12 & 4095]($9,$0,$8,$wx,$wy); //@line 415 "modules/window/wrap_Window.cpp"
  $13 = +HEAPF64[$wx>>3]; //@line 417 "modules/window/wrap_Window.cpp"
  _lua_pushnumber($L,$13); //@line 417 "modules/window/wrap_Window.cpp"
  $14 = +HEAPF64[$wy>>3]; //@line 418 "modules/window/wrap_Window.cpp"
  _lua_pushnumber($L,$14); //@line 418 "modules/window/wrap_Window.cpp"
  $$0 = 2;
  STACKTOP = sp;return ($$0|0); //@line 421 "modules/window/wrap_Window.cpp"
 }
 return (0)|0;
}
function __ZN4love6window10w_minimizeEP9lua_State($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 425 "modules/window/wrap_Window.cpp"
 $3 = ((($2)) + 100|0); //@line 425 "modules/window/wrap_Window.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 425 "modules/window/wrap_Window.cpp"
 FUNCTION_TABLE_vi[$4 & 8191]($1); //@line 425 "modules/window/wrap_Window.cpp"
 return 0; //@line 426 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window10w_maximizeEP9lua_State($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 431 "modules/window/wrap_Window.cpp"
 $3 = ((($2)) + 104|0); //@line 431 "modules/window/wrap_Window.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 431 "modules/window/wrap_Window.cpp"
 FUNCTION_TABLE_vi[$4 & 8191]($1); //@line 431 "modules/window/wrap_Window.cpp"
 return 0; //@line 432 "modules/window/wrap_Window.cpp"
}
function __ZN4love6window16w_showMessageBoxEP9lua_State($L) {
 $L = $L|0;
 var $$02 = 0, $$03 = 0, $$3 = 0, $$pre$i$i = 0, $$pre$i$i12 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0.0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $data = 0, $i$020 = 0, $lpad$phi$index = 0, $lpad$phi$index7 = 0, $storemerge = 0, $storemerge7 = 0, $storemerge7$in = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $data = sp + 32|0;
 $0 = sp + 20|0;
 $1 = ((($data)) + 4|0); //@line 437 "modules/window/wrap_Window.cpp"
 $2 = ((($data)) + 16|0); //@line 437 "modules/window/wrap_Window.cpp"
 $3 = ((($data)) + 28|0); //@line 437 "modules/window/wrap_Window.cpp"
 $4 = ((($data)) + 32|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ((($data)) + 40|0); //@line 437 "modules/window/wrap_Window.cpp"
 $6 = ((($data)) + 44|0); //@line 437 "modules/window/wrap_Window.cpp"
 $7 = ((($data)) + 48|0); //@line 437 "modules/window/wrap_Window.cpp"
 dest=$data; stop=dest+48|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));HEAP8[$data+48>>0]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 HEAP32[$data>>2] = 2; //@line 438 "modules/window/wrap_Window.cpp"
 __THREW__ = 0;
 $8 = (invoke_iiii(3359,($L|0),1,(0|0))|0); //@line 440 "modules/window/wrap_Window.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 L1: do {
  if ($10) {
   label = 12;
  } else {
   __THREW__ = 0;
   (invoke_iii(3360,($1|0),($8|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $11 = __THREW__; __THREW__ = 0;
   $12 = $11&1;
   if ($12) {
    label = 12;
   } else {
    __THREW__ = 0;
    $13 = (invoke_iiii(3359,($L|0),2,(0|0))|0); //@line 441 "modules/window/wrap_Window.cpp"
    $14 = __THREW__; __THREW__ = 0;
    $15 = $14&1;
    if ($15) {
     label = 12;
    } else {
     __THREW__ = 0;
     (invoke_iii(3360,($2|0),($13|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $16 = __THREW__; __THREW__ = 0;
     $17 = $16&1;
     if ($17) {
      label = 12;
     } else {
      __THREW__ = 0;
      $18 = (invoke_iii(3358,($L|0),3)|0); //@line 445 "modules/window/wrap_Window.cpp"
      $19 = __THREW__; __THREW__ = 0;
      $20 = $19&1;
      if ($20) {
       label = 12;
      } else {
       $21 = ($18|0)==(5); //@line 445 "modules/window/wrap_Window.cpp"
       L7: do {
        if ($21) {
         __THREW__ = 0;
         $22 = (invoke_iii(3381,($L|0),3)|0); //@line 447 "modules/window/wrap_Window.cpp"
         $23 = __THREW__; __THREW__ = 0;
         $24 = $23&1;
         if ($24) {
          label = 12;
          break L1;
         }
         $25 = ($22|0)==(0); //@line 448 "modules/window/wrap_Window.cpp"
         if ($25) {
          __THREW__ = 0;
          $27 = (invoke_iiii(3389,($L|0),(266251|0),($vararg_buffer|0))|0); //@line 449 "modules/window/wrap_Window.cpp"
          $28 = __THREW__; __THREW__ = 0;
          $29 = $28&1;
          if ($29) {
           label = 12;
           break L1;
          } else {
           $$3 = $27;
           break;
          }
         }
         $26 = ((($data)) + 36|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $i$020 = 0;
         while(1) {
          $36 = (($i$020) + 1)|0; //@line 454 "modules/window/wrap_Window.cpp"
          __THREW__ = 0;
          invoke_viii(3382,($L|0),3,($36|0)); //@line 454 "modules/window/wrap_Window.cpp"
          $37 = __THREW__; __THREW__ = 0;
          $38 = $37&1;
          if ($38) {
           label = 11;
           break;
          }
          __THREW__ = 0;
          invoke_viii(3403,($0|0),($L|0),-1); //@line 455 "modules/window/wrap_Window.cpp"
          $39 = __THREW__; __THREW__ = 0;
          $40 = $39&1;
          if ($40) {
           label = 11;
           break;
          }
          $41 = HEAP32[$4>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $42 = HEAP32[$26>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $43 = ($41>>>0)<($42>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          if ($43) {
           ;HEAP32[$41>>2]=HEAP32[$0>>2]|0;HEAP32[$41+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$41+8>>2]=HEAP32[$0+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $44 = HEAP32[$4>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $45 = ((($44)) + 12|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           HEAP32[$4>>2] = $45; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          } else {
           __THREW__ = 0;
           invoke_vii(3472,($3|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $46 = __THREW__; __THREW__ = 0;
           $47 = $46&1;
           if ($47) {
            label = 22;
            break;
           }
          }
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 455 "modules/window/wrap_Window.cpp"
          __THREW__ = 0;
          invoke_vii(3384,($L|0),-2); //@line 456 "modules/window/wrap_Window.cpp"
          $48 = __THREW__; __THREW__ = 0;
          $49 = $48&1;
          if ($49) {
           label = 11;
           break;
          }
          $50 = ($36>>>0)<($22>>>0); //@line 452 "modules/window/wrap_Window.cpp"
          if ($50) {
           $i$020 = $36;
          } else {
           label = 14;
           break;
          }
         }
         if ((label|0) == 11) {
          $30 = ___cxa_find_matching_catch_2()|0;
          $31 = tempRet0;
          $lpad$phi$index = $30;$lpad$phi$index7 = $31;
          label = 13;
          break L1;
         }
         else if ((label|0) == 14) {
          __THREW__ = 0;
          invoke_viii(3471,($L|0),3,(266293|0)); //@line 460 "modules/window/wrap_Window.cpp"
          $34 = __THREW__; __THREW__ = 0;
          $35 = $34&1;
          if ($35) {
           label = 12;
           break L1;
          }
          __THREW__ = 0;
          $53 = (invoke_iii(3358,($L|0),-1)|0); //@line 461 "modules/window/wrap_Window.cpp"
          $54 = __THREW__; __THREW__ = 0;
          $55 = $54&1;
          if ($55) {
           label = 12;
           break L1;
          }
          $56 = ($53|0)<(1); //@line 461 "modules/window/wrap_Window.cpp"
          if ($56) {
           $storemerge = 0;
          } else {
           __THREW__ = 0;
           $57 = (+invoke_dii(3361,($L|0),-1)); //@line 462 "modules/window/wrap_Window.cpp"
           $58 = __THREW__; __THREW__ = 0;
           $59 = $58&1;
           if ($59) {
            label = 12;
            break L1;
           }
           $60 = (~~(($57))); //@line 462 "modules/window/wrap_Window.cpp"
           $61 = (($60) + -1)|0; //@line 462 "modules/window/wrap_Window.cpp"
           $storemerge = $61;
          }
          HEAP32[$5>>2] = $storemerge; //@line 464 "modules/window/wrap_Window.cpp"
          __THREW__ = 0;
          invoke_vii(3384,($L|0),-2); //@line 465 "modules/window/wrap_Window.cpp"
          $62 = __THREW__; __THREW__ = 0;
          $63 = $62&1;
          if ($63) {
           label = 12;
           break L1;
          }
          __THREW__ = 0;
          invoke_viii(3471,($L|0),3,(266305|0)); //@line 468 "modules/window/wrap_Window.cpp"
          $64 = __THREW__; __THREW__ = 0;
          $65 = $64&1;
          if ($65) {
           label = 12;
           break L1;
          }
          __THREW__ = 0;
          $66 = (invoke_iii(3358,($L|0),-1)|0); //@line 469 "modules/window/wrap_Window.cpp"
          $67 = __THREW__; __THREW__ = 0;
          $68 = $67&1;
          if ($68) {
           label = 12;
           break L1;
          }
          $69 = ($66|0)<(1); //@line 469 "modules/window/wrap_Window.cpp"
          if ($69) {
           $74 = HEAP32[$4>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $75 = HEAP32[$3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $76 = (($74) - ($75))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $77 = (($76|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $storemerge7$in = $77;
          } else {
           __THREW__ = 0;
           $70 = (+invoke_dii(3361,($L|0),-1)); //@line 470 "modules/window/wrap_Window.cpp"
           $71 = __THREW__; __THREW__ = 0;
           $72 = $71&1;
           if ($72) {
            label = 12;
            break L1;
           }
           $73 = (~~(($70))); //@line 470 "modules/window/wrap_Window.cpp"
           $storemerge7$in = $73;
          }
          $storemerge7 = (($storemerge7$in) + -1)|0; //@line 470 "modules/window/wrap_Window.cpp"
          HEAP32[$6>>2] = $storemerge7; //@line 472 "modules/window/wrap_Window.cpp"
          __THREW__ = 0;
          invoke_vii(3384,($L|0),-2); //@line 473 "modules/window/wrap_Window.cpp"
          $78 = __THREW__; __THREW__ = 0;
          $79 = $78&1;
          if ($79) {
           label = 12;
           break L1;
          }
          __THREW__ = 0;
          $80 = (invoke_iii(3358,($L|0),4)|0); //@line 475 "modules/window/wrap_Window.cpp"
          $81 = __THREW__; __THREW__ = 0;
          $82 = $81&1;
          if ($82) {
           label = 12;
           break L1;
          }
          $83 = ($80|0)<(1); //@line 475 "modules/window/wrap_Window.cpp"
          do {
           if (!($83)) {
            __THREW__ = 0;
            $84 = (invoke_iiii(3359,($L|0),4,(0|0))|0); //@line 475 "modules/window/wrap_Window.cpp"
            $85 = __THREW__; __THREW__ = 0;
            $86 = $85&1;
            if ($86) {
             label = 12;
             break L1;
            }
            $87 = ($84|0)==(0|0); //@line 476 "modules/window/wrap_Window.cpp"
            if ($87) {
             break;
            }
            __THREW__ = 0;
            $88 = (invoke_iii(3473,($84|0),($data|0))|0); //@line 476 "modules/window/wrap_Window.cpp"
            $89 = __THREW__; __THREW__ = 0;
            $90 = $89&1;
            if ($90) {
             label = 12;
             break L1;
            }
            if ($88) {
             break;
            }
            __THREW__ = 0;
            HEAP32[$vararg_buffer1>>2] = $84; //@line 477 "modules/window/wrap_Window.cpp"
            $91 = (invoke_iiii(3389,($L|0),(266318|0),($vararg_buffer1|0))|0); //@line 477 "modules/window/wrap_Window.cpp"
            $92 = __THREW__; __THREW__ = 0;
            $93 = $92&1;
            if ($93) {
             label = 12;
             break L1;
            } else {
             $$3 = $91;
             break L7;
            }
           }
          } while(0);
          __THREW__ = 0;
          $94 = (invoke_iiii(3474,($L|0),5,1)|0); //@line 479 "modules/window/wrap_Window.cpp"
          $95 = __THREW__; __THREW__ = 0;
          $96 = $95&1;
          if ($96) {
           label = 12;
           break L1;
          }
          $97 = $94&1; //@line 479 "modules/window/wrap_Window.cpp"
          HEAP8[$7>>0] = $97; //@line 479 "modules/window/wrap_Window.cpp"
          $98 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
          $99 = HEAP32[$98>>2]|0; //@line 481 "modules/window/wrap_Window.cpp"
          $100 = ((($99)) + 180|0); //@line 481 "modules/window/wrap_Window.cpp"
          $101 = HEAP32[$100>>2]|0; //@line 481 "modules/window/wrap_Window.cpp"
          __THREW__ = 0;
          $102 = (invoke_iii($101|0,($98|0),($data|0))|0); //@line 481 "modules/window/wrap_Window.cpp"
          $103 = __THREW__; __THREW__ = 0;
          $104 = $103&1;
          if ($104) {
           label = 12;
           break L1;
          }
          $105 = (($102) + 1)|0; //@line 482 "modules/window/wrap_Window.cpp"
          __THREW__ = 0;
          invoke_vii(3468,($L|0),($105|0)); //@line 482 "modules/window/wrap_Window.cpp"
          $106 = __THREW__; __THREW__ = 0;
          $107 = $106&1;
          if ($107) {
           label = 12;
           break L1;
          } else {
           $$3 = 1;
           break;
          }
         }
         else if ((label|0) == 22) {
          $51 = ___cxa_find_matching_catch_2()|0;
          $52 = tempRet0;
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 455 "modules/window/wrap_Window.cpp"
          $$02 = $51;$$03 = $52;
          break L1;
         }
        } else {
         __THREW__ = 0;
         $108 = (invoke_iii(3358,($L|0),3)|0); //@line 486 "modules/window/wrap_Window.cpp"
         $109 = __THREW__; __THREW__ = 0;
         $110 = $109&1;
         if ($110) {
          label = 12;
          break L1;
         }
         $111 = ($108|0)<(1); //@line 486 "modules/window/wrap_Window.cpp"
         if (!($111)) {
          __THREW__ = 0;
          $112 = (invoke_iiii(3359,($L|0),3,(0|0))|0); //@line 486 "modules/window/wrap_Window.cpp"
          $113 = __THREW__; __THREW__ = 0;
          $114 = $113&1;
          if ($114) {
           label = 12;
           break L1;
          }
          $115 = ($112|0)==(0|0); //@line 487 "modules/window/wrap_Window.cpp"
          if (!($115)) {
           __THREW__ = 0;
           $116 = (invoke_iii(3473,($112|0),($data|0))|0); //@line 487 "modules/window/wrap_Window.cpp"
           $117 = __THREW__; __THREW__ = 0;
           $118 = $117&1;
           if ($118) {
            label = 12;
            break L1;
           }
           if (!($116)) {
            __THREW__ = 0;
            HEAP32[$vararg_buffer3>>2] = $112; //@line 488 "modules/window/wrap_Window.cpp"
            $119 = (invoke_iiii(3389,($L|0),(266318|0),($vararg_buffer3|0))|0); //@line 488 "modules/window/wrap_Window.cpp"
            $120 = __THREW__; __THREW__ = 0;
            $121 = $120&1;
            if ($121) {
             label = 12;
             break L1;
            } else {
             $$3 = $119;
             break;
            }
           }
          }
         }
         __THREW__ = 0;
         $122 = (invoke_iiii(3474,($L|0),4,1)|0); //@line 490 "modules/window/wrap_Window.cpp"
         $123 = __THREW__; __THREW__ = 0;
         $124 = $123&1;
         if ($124) {
          label = 12;
          break L1;
         }
         $125 = $122&1; //@line 490 "modules/window/wrap_Window.cpp"
         HEAP8[$7>>0] = $125; //@line 490 "modules/window/wrap_Window.cpp"
         $126 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
         $127 = HEAP32[$126>>2]|0; //@line 493 "modules/window/wrap_Window.cpp"
         $128 = ((($127)) + 176|0); //@line 493 "modules/window/wrap_Window.cpp"
         $129 = HEAP32[$128>>2]|0; //@line 493 "modules/window/wrap_Window.cpp"
         $130 = HEAP32[$data>>2]|0; //@line 493 "modules/window/wrap_Window.cpp"
         __THREW__ = 0;
         $131 = (invoke_iiiiii($129|0,($126|0),($1|0),($2|0),($130|0),($122|0))|0); //@line 493 "modules/window/wrap_Window.cpp"
         $132 = __THREW__; __THREW__ = 0;
         $133 = $132&1;
         if ($133) {
          label = 12;
          break L1;
         }
         __THREW__ = 0;
         invoke_vii(3397,($L|0),($131|0)); //@line 494 "modules/window/wrap_Window.cpp"
         $134 = __THREW__; __THREW__ = 0;
         $135 = $134&1;
         if ($135) {
          label = 12;
          break L1;
         } else {
          $$3 = 1;
         }
        }
       } while(0);
       $136 = ((($data)) + 28|0); //@line 92 "./modules/window/Window.h"
       $137 = HEAP32[$136>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $138 = ($137|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($138) {
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 92 "./modules/window/Window.h"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 92 "./modules/window/Window.h"
        STACKTOP = sp;return ($$3|0); //@line 498 "modules/window/wrap_Window.cpp"
       }
       $139 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $140 = ($139|0)==($137|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($140) {
        $145 = $137;
       } else {
        $142 = $139;
        while(1) {
         $141 = ((($142)) + -12|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$4>>2] = $141; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($141); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $143 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $144 = ($143|0)==($137|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($144) {
          break;
         } else {
          $142 = $143;
         }
        }
        $$pre$i$i12 = HEAP32[$136>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $145 = $$pre$i$i12;
       }
       __ZdlPv($145); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 92 "./modules/window/Window.h"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 92 "./modules/window/Window.h"
       STACKTOP = sp;return ($$3|0); //@line 498 "modules/window/wrap_Window.cpp"
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 12) {
  $32 = ___cxa_find_matching_catch_2()|0;
  $33 = tempRet0;
  $lpad$phi$index = $32;$lpad$phi$index7 = $33;
  label = 13;
 }
 if ((label|0) == 13) {
  $$02 = $lpad$phi$index;$$03 = $lpad$phi$index7;
 }
 $146 = ((($data)) + 28|0); //@line 92 "./modules/window/Window.h"
 $147 = HEAP32[$146>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $148 = ($147|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($148) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 92 "./modules/window/Window.h"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 92 "./modules/window/Window.h"
  ___resumeException($$02|0);
  // unreachable;
 }
 $149 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $150 = ($149|0)==($147|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($150) {
  $155 = $147;
 } else {
  $152 = $149;
  while(1) {
   $151 = ((($152)) + -12|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$4>>2] = $151; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($151); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $153 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $154 = ($153|0)==($147|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($154) {
    break;
   } else {
    $152 = $153;
   }
  }
  $$pre$i$i = HEAP32[$146>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $155 = $$pre$i$i;
 }
 __ZdlPv($155); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 92 "./modules/window/Window.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 92 "./modules/window/Window.h"
 ___resumeException($$02|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__16vectorINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEENS4_IS6_EEE21__push_back_slow_pathIS6_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$01$i$i = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $__v$sroa$5$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($3|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(357913941); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $10 = $$pre;
 } else {
  $10 = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($10))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = (($9|0) / 12)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($11>>>0)<(178956970); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $16 = $11 << 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = ($16>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $18 = $17 ? $5 : $16; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = (($19) - ($10))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = (($20|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = ($18|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($22) {
   $$0$i2 = 0;$26 = 0;$27 = $21;
  } else {
   $$0$i1 = $18;$54 = $21;
   label = 6;
  }
 } else {
  $13 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) - ($10))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = (($14|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 357913941;$54 = $15;
  label = 6;
 }
 if ((label|0) == 6) {
  $23 = ($$0$i1*12)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = (__Znwj($23)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$26 = $24;$27 = $54;
 }
 $25 = (($26) + (($27*12)|0)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $25; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = (($26) + (($$0$i2*12)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = $29; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 ;HEAP32[$25>>2]=HEAP32[$__x>>2]|0;HEAP32[$25+4>>2]=HEAP32[$__x+4>>2]|0;HEAP32[$25+8>>2]=HEAP32[$__x+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 ;HEAP32[$__x>>2]=0|0;HEAP32[$__x+4>>2]=0|0;HEAP32[$__x+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $31 = ((($25)) + 12|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = $31; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $33 = HEAP32[$this>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $34 = HEAP32[$0>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $35 = ($34|0)==($33|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($35) {
  $36 = $33; //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $44 = $28;$46 = $36;$48 = $33;
 } else {
  $$01$i$i = $34;$38 = $25;$__v$sroa$5$0 = $28;
  while(1) {
   $37 = ((($38)) + -12|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $39 = ((($$01$i$i)) + -12|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   ;HEAP32[$37>>2]=HEAP32[$39>>2]|0;HEAP32[$37+4>>2]=HEAP32[$39+4>>2]|0;HEAP32[$37+8>>2]=HEAP32[$39+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   ;HEAP32[$39>>2]=0|0;HEAP32[$39+4>>2]=0|0;HEAP32[$39+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $40 = $__v$sroa$5$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $41 = ((($40)) + -12|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $42 = $41; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $43 = ($39|0)==($33|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if ($43) {
    $$lcssa = $42;
    break;
   } else {
    $$01$i$i = $39;$38 = $41;$__v$sroa$5$0 = $42;
   }
  }
  $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $$pre3 = HEAP32[$0>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $44 = $$lcssa;$46 = $$pre$i;$48 = $$pre3;
 }
 HEAP32[$this>>2] = $44; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $32; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $45 = $46; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $47 = ($48|0)==($45|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if (!($47)) {
  $50 = $48;
  while(1) {
   $49 = ((($50)) + -12|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($49); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $51 = ($49|0)==($45|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($51) {
    break;
   } else {
    $50 = $49;
   }
  }
 }
 $52 = ($46|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($52) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $53 = $46; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($53); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love6window18w_requestAttentionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,1,0)|0); //@line 502 "modules/window/wrap_Window.cpp"
 $1 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 503 "modules/window/wrap_Window.cpp"
 $3 = ((($2)) + 184|0); //@line 503 "modules/window/wrap_Window.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 503 "modules/window/wrap_Window.cpp"
 FUNCTION_TABLE_vii[$4 & 8191]($1,$0); //@line 503 "modules/window/wrap_Window.cpp"
 return 0; //@line 504 "modules/window/wrap_Window.cpp"
}
function _luaopen_love_window($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0;
 var $eh$lpad$body$i$indexZ2D = 0, $instance$sroa$0$1 = 0, $vararg_buffer = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $w = sp + 4|0;
 $0 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $1 = $0; //@line 543 "modules/window/wrap_Window.cpp"
 $2 = ($0|0)==(0|0); //@line 544 "modules/window/wrap_Window.cpp"
 L1: do {
  if ($2) {
   __THREW__ = 0;
   $3 = (invoke_ii(3348,128)|0); //@line 545 "modules/window/wrap_Window.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   do {
    if ($5) {
     $11 = ___cxa_find_matching_catch_3(8992|0)|0;
     $12 = tempRet0;
     $eh$lpad$body$i$index2Z2D = $12;$eh$lpad$body$i$indexZ2D = $11;
    } else {
     __THREW__ = 0;
     invoke_vi(3475,($3|0)); //@line 545 "modules/window/wrap_Window.cpp"
     $6 = __THREW__; __THREW__ = 0;
     $7 = $6&1;
     if ($7) {
      $8 = ___cxa_find_matching_catch_3(8992|0)|0;
      $9 = tempRet0;
      __ZdlPv($3); //@line 545 "modules/window/wrap_Window.cpp"
      $eh$lpad$body$i$index2Z2D = $9;$eh$lpad$body$i$indexZ2D = $8;
      break;
     } else {
      $10 = $3; //@line 545 "modules/window/wrap_Window.cpp"
      $instance$sroa$0$1 = $10;
      break L1;
     }
    }
   } while(0);
   $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $14 = ($eh$lpad$body$i$index2Z2D|0)==($13|0); //@line 535 "./common/runtime.h"
   if (!($14)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $15 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
   $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
   $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
   $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $instance$sroa$0$1 = $1;
    break;
   }
   $23 = ___cxa_find_matching_catch_2()|0;
   $24 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    $27 = ___cxa_find_matching_catch_3(0|0)|0;
    $28 = tempRet0;
    ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $24;$$02$i = $23;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $29 = HEAP32[$0>>2]|0; //@line 547 "modules/window/wrap_Window.cpp"
   $30 = ((($29)) + 8|0); //@line 547 "modules/window/wrap_Window.cpp"
   $31 = HEAP32[$30>>2]|0; //@line 547 "modules/window/wrap_Window.cpp"
   FUNCTION_TABLE_vi[$31 & 8191]($0); //@line 547 "modules/window/wrap_Window.cpp"
   $instance$sroa$0$1 = $1;
  }
 } while(0);
 HEAP32[$w>>2] = $instance$sroa$0$1; //@line 550 "modules/window/wrap_Window.cpp"
 $32 = ((($w)) + 4|0); //@line 551 "modules/window/wrap_Window.cpp"
 HEAP32[$32>>2] = 437644; //@line 551 "modules/window/wrap_Window.cpp"
 $33 = ((($w)) + 8|0); //@line 552 "modules/window/wrap_Window.cpp"
 HEAP32[$33>>2] = 3; //@line 552 "modules/window/wrap_Window.cpp"
 $34 = ((($w)) + 12|0); //@line 553 "modules/window/wrap_Window.cpp"
 HEAP32[$34>>2] = 11588; //@line 553 "modules/window/wrap_Window.cpp"
 $35 = ((($w)) + 16|0); //@line 554 "modules/window/wrap_Window.cpp"
 HEAP32[$35>>2] = 0; //@line 554 "modules/window/wrap_Window.cpp"
 $36 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 556 "modules/window/wrap_Window.cpp"
 STACKTOP = sp;return ($36|0); //@line 556 "modules/window/wrap_Window.cpp"
}
function __Z12luaopen_enetP9lua_State($l) {
 $l = $l|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 (_enet_initialize()|0); //@line 778 "libraries/enet/enet.cpp"
 (_atexit((3476|0))|0); //@line 779 "libraries/enet/enet.cpp"
 (_luaL_newmetatable($l,266649)|0); //@line 782 "libraries/enet/enet.cpp"
 _lua_createtable($l,0,0); //@line 783 "libraries/enet/enet.cpp"
 _luax_register($l,0,11828); //@line 784 "libraries/enet/enet.cpp"
 _lua_setfield($l,-2,523215); //@line 785 "libraries/enet/enet.cpp"
 _lua_pushcclosure($l,313,0); //@line 786 "libraries/enet/enet.cpp"
 _lua_setfield($l,-2,523106); //@line 787 "libraries/enet/enet.cpp"
 (_luaL_newmetatable($l,266726)|0); //@line 789 "libraries/enet/enet.cpp"
 _lua_createtable($l,0,0); //@line 790 "libraries/enet/enet.cpp"
 _luax_register($l,0,11964); //@line 791 "libraries/enet/enet.cpp"
 _lua_setfield($l,-2,523215); //@line 792 "libraries/enet/enet.cpp"
 _lua_pushcclosure($l,3477,0); //@line 793 "libraries/enet/enet.cpp"
 _lua_setfield($l,-2,521666); //@line 794 "libraries/enet/enet.cpp"
 _lua_createtable($l,0,0); //@line 797 "libraries/enet/enet.cpp"
 _lua_createtable($l,0,0); //@line 799 "libraries/enet/enet.cpp"
 _lua_pushstring($l,442531); //@line 800 "libraries/enet/enet.cpp"
 _lua_setfield($l,-2,520718); //@line 801 "libraries/enet/enet.cpp"
 (_lua_setmetatable($l,-2)|0); //@line 802 "libraries/enet/enet.cpp"
 _lua_setfield($l,-10000,266715); //@line 804 "libraries/enet/enet.cpp"
 _luax_register($l,267475,12092); //@line 806 "libraries/enet/enet.cpp"
 return 1; //@line 809 "libraries/enet/enet.cpp"
}
function __ZL12host_serviceP9lua_State($l) {
 $l = $l|0;
 var $$0 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, $event = 0, $timeout$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $event = sp + 12|0;
 $0 = (_luaL_checkudata($l,1,266649)|0); //@line 274 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 274 "libraries/enet/enet.cpp"
 $2 = ($1|0)==(0|0); //@line 275 "libraries/enet/enet.cpp"
 if ($2) {
  $3 = (_luaL_error($l,266667,$vararg_buffer)|0); //@line 276 "libraries/enet/enet.cpp"
  $$1 = $3;
  STACKTOP = sp;return ($$1|0); //@line 290 "libraries/enet/enet.cpp"
 }
 $4 = (_lua_gettop($l)|0); //@line 281 "libraries/enet/enet.cpp"
 $5 = ($4|0)>(1); //@line 281 "libraries/enet/enet.cpp"
 if ($5) {
  $6 = (+_luaL_checknumber($l,2)); //@line 282 "libraries/enet/enet.cpp"
  $7 = (~~(($6))); //@line 282 "libraries/enet/enet.cpp"
  $timeout$0 = $7;
 } else {
  $timeout$0 = 0;
 }
 $8 = (_enet_host_service($1,$event,$timeout$0)|0); //@line 284 "libraries/enet/enet.cpp"
 $9 = ($8|0)==(0); //@line 285 "libraries/enet/enet.cpp"
 do {
  if ($9) {
   $$0 = 0;
  } else {
   $10 = ($8|0)<(0); //@line 286 "libraries/enet/enet.cpp"
   if ($10) {
    $11 = (_luaL_error($l,266694,$vararg_buffer1)|0); //@line 286 "libraries/enet/enet.cpp"
    $$0 = $11;
    break;
   } else {
    __ZL10push_eventP9lua_StateP10_ENetEvent($l,$event); //@line 288 "libraries/enet/enet.cpp"
    $$0 = 1;
    break;
   }
  }
 } while(0);
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0); //@line 290 "libraries/enet/enet.cpp"
}
function __ZL10push_eventP9lua_StateP10_ENetEvent($l,$event) {
 $l = $l|0;
 $event = $event|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _lua_createtable($l,0,0); //@line 127 "libraries/enet/enet.cpp"
 $0 = ((($event)) + 4|0); //@line 129 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 129 "libraries/enet/enet.cpp"
 $2 = ($1|0)==(0|0); //@line 129 "libraries/enet/enet.cpp"
 if (!($2)) {
  __ZL9push_peerP9lua_StateP9_ENetPeer($l,$1); //@line 130 "libraries/enet/enet.cpp"
  _lua_setfield($l,-2,266736); //@line 131 "libraries/enet/enet.cpp"
 }
 $3 = HEAP32[$event>>2]|0; //@line 134 "libraries/enet/enet.cpp"
 switch ($3|0) {
 case 1:  {
  $4 = ((($event)) + 12|0); //@line 136 "libraries/enet/enet.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 136 "libraries/enet/enet.cpp"
  _lua_pushinteger($l,$5); //@line 136 "libraries/enet/enet.cpp"
  _lua_setfield($l,-2,363661); //@line 137 "libraries/enet/enet.cpp"
  _lua_pushstring($l,363962); //@line 139 "libraries/enet/enet.cpp"
  _lua_setfield($l,-2,521785); //@line 163 "libraries/enet/enet.cpp"
  return; //@line 164 "libraries/enet/enet.cpp"
  break;
 }
 case 2:  {
  $6 = ((($event)) + 12|0); //@line 142 "libraries/enet/enet.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 142 "libraries/enet/enet.cpp"
  _lua_pushinteger($l,$7); //@line 142 "libraries/enet/enet.cpp"
  _lua_setfield($l,-2,363661); //@line 143 "libraries/enet/enet.cpp"
  _lua_pushstring($l,266741); //@line 145 "libraries/enet/enet.cpp"
  _lua_setfield($l,-2,521785); //@line 163 "libraries/enet/enet.cpp"
  return; //@line 164 "libraries/enet/enet.cpp"
  break;
 }
 case 3:  {
  $8 = ((($event)) + 16|0); //@line 148 "libraries/enet/enet.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 148 "libraries/enet/enet.cpp"
  $10 = ((($9)) + 8|0); //@line 148 "libraries/enet/enet.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 148 "libraries/enet/enet.cpp"
  $12 = ((($9)) + 12|0); //@line 148 "libraries/enet/enet.cpp"
  $13 = HEAP32[$12>>2]|0; //@line 148 "libraries/enet/enet.cpp"
  _lua_pushlstring($l,$11,$13); //@line 148 "libraries/enet/enet.cpp"
  _lua_setfield($l,-2,363661); //@line 149 "libraries/enet/enet.cpp"
  $14 = ((($event)) + 8|0); //@line 151 "libraries/enet/enet.cpp"
  $15 = HEAP8[$14>>0]|0; //@line 151 "libraries/enet/enet.cpp"
  $16 = $15&255; //@line 151 "libraries/enet/enet.cpp"
  _lua_pushinteger($l,$16); //@line 151 "libraries/enet/enet.cpp"
  _lua_setfield($l,-2,526910); //@line 152 "libraries/enet/enet.cpp"
  _lua_pushstring($l,364155); //@line 154 "libraries/enet/enet.cpp"
  $17 = HEAP32[$8>>2]|0; //@line 156 "libraries/enet/enet.cpp"
  _enet_packet_destroy($17); //@line 156 "libraries/enet/enet.cpp"
  _lua_setfield($l,-2,521785); //@line 163 "libraries/enet/enet.cpp"
  return; //@line 164 "libraries/enet/enet.cpp"
  break;
 }
 case 0:  {
  _lua_pushstring($l,525139); //@line 159 "libraries/enet/enet.cpp"
  _lua_setfield($l,-2,521785); //@line 163 "libraries/enet/enet.cpp"
  return; //@line 164 "libraries/enet/enet.cpp"
  break;
 }
 default: {
  _lua_setfield($l,-2,521785); //@line 163 "libraries/enet/enet.cpp"
  return; //@line 164 "libraries/enet/enet.cpp"
 }
 }
}
function __ZL9push_peerP9lua_StateP9_ENetPeer($l,$peer) {
 $l = $l|0;
 $peer = $peer|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _lua_getfield($l,-10000,266715); //@line 106 "libraries/enet/enet.cpp"
 _lua_pushlightuserdata($l,$peer); //@line 107 "libraries/enet/enet.cpp"
 _lua_gettable($l,-2); //@line 108 "libraries/enet/enet.cpp"
 $0 = (_lua_type($l,-1)|0); //@line 110 "libraries/enet/enet.cpp"
 $1 = ($0|0)==(0); //@line 110 "libraries/enet/enet.cpp"
 if (!($1)) {
  _lua_remove($l,-2); //@line 123 "libraries/enet/enet.cpp"
  return; //@line 124 "libraries/enet/enet.cpp"
 }
 _lua_settop($l,-2); //@line 112 "libraries/enet/enet.cpp"
 $2 = (_lua_newuserdata($l,4)|0); //@line 114 "libraries/enet/enet.cpp"
 HEAP32[$2>>2] = $peer; //@line 114 "libraries/enet/enet.cpp"
 _lua_getfield($l,-10000,266726); //@line 115 "libraries/enet/enet.cpp"
 (_lua_setmetatable($l,-2)|0); //@line 116 "libraries/enet/enet.cpp"
 _lua_pushlightuserdata($l,$peer); //@line 118 "libraries/enet/enet.cpp"
 _lua_pushvalue($l,-2); //@line 119 "libraries/enet/enet.cpp"
 _lua_settable($l,-4); //@line 121 "libraries/enet/enet.cpp"
 _lua_remove($l,-2); //@line 123 "libraries/enet/enet.cpp"
 return; //@line 124 "libraries/enet/enet.cpp"
}
function __ZL17host_check_eventsP9lua_State($l) {
 $l = $l|0;
 var $$0 = 0, $$1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $event = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $event = sp + 12|0;
 $0 = (_luaL_checkudata($l,1,266649)|0); //@line 296 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 296 "libraries/enet/enet.cpp"
 $2 = ($1|0)==(0|0); //@line 297 "libraries/enet/enet.cpp"
 if ($2) {
  $3 = (_luaL_error($l,266667,$vararg_buffer)|0); //@line 298 "libraries/enet/enet.cpp"
  $$1 = $3;
  STACKTOP = sp;return ($$1|0); //@line 307 "libraries/enet/enet.cpp"
 }
 $4 = (_enet_host_check_events($1,$event)|0); //@line 301 "libraries/enet/enet.cpp"
 $5 = ($4|0)==(0); //@line 302 "libraries/enet/enet.cpp"
 do {
  if ($5) {
   $$0 = 0;
  } else {
   $6 = ($4|0)<(0); //@line 303 "libraries/enet/enet.cpp"
   if ($6) {
    $7 = (_luaL_error($l,266765,$vararg_buffer1)|0); //@line 303 "libraries/enet/enet.cpp"
    $$0 = $7;
    break;
   } else {
    __ZL10push_eventP9lua_StateP10_ENetEvent($l,$event); //@line 305 "libraries/enet/enet.cpp"
    $$0 = 1;
    break;
   }
  }
 } while(0);
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0); //@line 307 "libraries/enet/enet.cpp"
}
function __ZL30host_compress_with_range_coderP9lua_State($l) {
 $l = $l|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 313 "libraries/enet/enet.cpp"
 $0 = (_luaL_checkudata($l,1,266649)|0); //@line 314 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 314 "libraries/enet/enet.cpp"
 $2 = ($1|0)==(0|0); //@line 315 "libraries/enet/enet.cpp"
 do {
  if ($2) {
   $3 = (_luaL_error($l,266667,$vararg_buffer)|0); //@line 316 "libraries/enet/enet.cpp"
   $$0 = $3;
  } else {
   $4 = (_enet_host_compress_with_range_coder($1)|0); //@line 319 "libraries/enet/enet.cpp"
   $5 = ($4|0)==(0); //@line 320 "libraries/enet/enet.cpp"
   if ($5) {
    _lua_pushboolean($l,1); //@line 321 "libraries/enet/enet.cpp"
    $$0 = 1;
    break;
   } else {
    _lua_pushboolean($l,0); //@line 323 "libraries/enet/enet.cpp"
    $$0 = 1;
    break;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0); //@line 327 "libraries/enet/enet.cpp"
}
function __ZL12host_connectP9lua_State($l) {
 $l = $l|0;
 var $$0 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0, $address = 0;
 var $channel_count$0 = 0, $data$0 = 0, $data$1 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $address = sp + 16|0;
 $0 = (_luaL_checkudata($l,1,266649)|0); //@line 337 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 337 "libraries/enet/enet.cpp"
 $2 = ($1|0)==(0|0); //@line 338 "libraries/enet/enet.cpp"
 if ($2) {
  $3 = (_luaL_error($l,266667,$vararg_buffer)|0); //@line 339 "libraries/enet/enet.cpp"
  $$1 = $3;
  STACKTOP = sp;return ($$1|0); //@line 366 "libraries/enet/enet.cpp"
 }
 $4 = (_luaL_checklstring($l,2,0)|0); //@line 347 "libraries/enet/enet.cpp"
 __ZL13parse_addressP9lua_StatePKcP12_ENetAddress($l,$4,$address); //@line 347 "libraries/enet/enet.cpp"
 $5 = (_lua_gettop($l)|0); //@line 349 "libraries/enet/enet.cpp"
 switch ($5|0) {
 case 4:  {
  $6 = (_lua_type($l,4)|0); //@line 351 "libraries/enet/enet.cpp"
  $7 = ($6|0)==(0); //@line 351 "libraries/enet/enet.cpp"
  if ($7) {
   $data$0 = 0;
   label = 6;
  } else {
   $8 = (+_luaL_checknumber($l,4)); //@line 351 "libraries/enet/enet.cpp"
   $9 = (~~(($8))); //@line 351 "libraries/enet/enet.cpp"
   $data$0 = $9;
   label = 6;
  }
  break;
 }
 case 3:  {
  $data$0 = 0;
  label = 6;
  break;
 }
 default: {
  $channel_count$0 = 1;$data$1 = 0;
 }
 }
 if ((label|0) == 6) {
  $10 = (_lua_type($l,3)|0); //@line 353 "libraries/enet/enet.cpp"
  $11 = ($10|0)==(0); //@line 353 "libraries/enet/enet.cpp"
  if ($11) {
   $channel_count$0 = 1;$data$1 = $data$0;
  } else {
   $12 = (+_luaL_checknumber($l,3)); //@line 353 "libraries/enet/enet.cpp"
   $13 = (~~(($12))); //@line 353 "libraries/enet/enet.cpp"
   $channel_count$0 = $13;$data$1 = $data$0;
  }
 }
 $14 = (_enet_host_connect($1,$address,$channel_count$0,$data$1)|0); //@line 357 "libraries/enet/enet.cpp"
 $15 = ($14|0)==(0|0); //@line 359 "libraries/enet/enet.cpp"
 if ($15) {
  $16 = (_luaL_error($l,266906,$vararg_buffer1)|0); //@line 360 "libraries/enet/enet.cpp"
  $$0 = $16;
 } else {
  __ZL9push_peerP9lua_StateP9_ENetPeer($l,$14); //@line 363 "libraries/enet/enet.cpp"
  $$0 = 1;
 }
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0); //@line 366 "libraries/enet/enet.cpp"
}
function __ZL13parse_addressP9lua_StatePKcP12_ENetAddress($l,$addr_str,$address) {
 $l = $l|0;
 $addr_str = $addr_str|0;
 $address = $address|0;
 var $$pre = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $c$05 = 0, $host_i$02 = 0, $host_i$1 = 0, $host_i$1$lcssa = 0, $host_str = 0, $or$cond = 0, $port_i$0$lcssa7 = 0, $port_i$0$lcssa8 = 0, $port_i$03 = 0, $port_i$1 = 0;
 var $port_i$1$lcssa = 0, $port_str = 0, $scanning_port$04 = 0, $scanning_port$1 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $host_str = sp + 64|0;
 $port_str = sp + 32|0;
 dest=$host_str; stop=dest+128|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 49 "libraries/enet/enet.cpp"
 dest=$port_str; stop=dest+32|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 50 "libraries/enet/enet.cpp"
 $0 = HEAP8[$addr_str>>0]|0; //@line 55 "libraries/enet/enet.cpp"
 $1 = ($0<<24>>24)==(0); //@line 55 "libraries/enet/enet.cpp"
 if ($1) {
  HEAP8[$host_str>>0] = 0; //@line 68 "libraries/enet/enet.cpp"
  HEAP8[$port_str>>0] = 0; //@line 69 "libraries/enet/enet.cpp"
  $port_i$0$lcssa8 = 0;
  label = 11;
 } else {
  $27 = $0;$c$05 = $addr_str;$host_i$02 = 0;$port_i$03 = 0;$scanning_port$04 = 0;
  while(1) {
   $2 = ($host_i$02|0)>(127); //@line 56 "libraries/enet/enet.cpp"
   $3 = ($port_i$03|0)>(31); //@line 56 "libraries/enet/enet.cpp"
   $or$cond = $2 | $3; //@line 56 "libraries/enet/enet.cpp"
   if ($or$cond) {
    (_luaL_error($l,266812,$vararg_buffer)|0); //@line 56 "libraries/enet/enet.cpp"
    $$pre = HEAP8[$c$05>>0]|0; //@line 58 "libraries/enet/enet.cpp"
    $7 = $$pre;
   } else {
    $7 = $27;
   }
   $4 = ($scanning_port$04|0)==(0); //@line 57 "libraries/enet/enet.cpp"
   if ($4) {
    $8 = ($7<<24>>24)==(58); //@line 60 "libraries/enet/enet.cpp"
    if ($8) {
     $host_i$1 = $host_i$02;$port_i$1 = $port_i$03;$scanning_port$1 = 1;
    } else {
     $9 = (($host_i$02) + 1)|0; //@line 63 "libraries/enet/enet.cpp"
     $10 = (($host_str) + ($host_i$02)|0); //@line 63 "libraries/enet/enet.cpp"
     HEAP8[$10>>0] = $7; //@line 63 "libraries/enet/enet.cpp"
     $host_i$1 = $9;$port_i$1 = $port_i$03;$scanning_port$1 = 0;
    }
   } else {
    $5 = (($port_i$03) + 1)|0; //@line 58 "libraries/enet/enet.cpp"
    $6 = (($port_str) + ($port_i$03)|0); //@line 58 "libraries/enet/enet.cpp"
    HEAP8[$6>>0] = $7; //@line 58 "libraries/enet/enet.cpp"
    $host_i$1 = $host_i$02;$port_i$1 = $5;$scanning_port$1 = $scanning_port$04;
   }
   $11 = ((($c$05)) + 1|0); //@line 66 "libraries/enet/enet.cpp"
   $12 = HEAP8[$11>>0]|0; //@line 55 "libraries/enet/enet.cpp"
   $13 = ($12<<24>>24)==(0); //@line 55 "libraries/enet/enet.cpp"
   if ($13) {
    $host_i$1$lcssa = $host_i$1;$port_i$1$lcssa = $port_i$1;
    break;
   } else {
    $27 = $12;$c$05 = $11;$host_i$02 = $host_i$1;$port_i$03 = $port_i$1;$scanning_port$04 = $scanning_port$1;
   }
  }
  $14 = (($host_str) + ($host_i$1$lcssa)|0); //@line 68 "libraries/enet/enet.cpp"
  HEAP8[$14>>0] = 0; //@line 68 "libraries/enet/enet.cpp"
  $15 = (($port_str) + ($port_i$1$lcssa)|0); //@line 69 "libraries/enet/enet.cpp"
  HEAP8[$15>>0] = 0; //@line 69 "libraries/enet/enet.cpp"
  $16 = ($host_i$1$lcssa|0)==(0); //@line 71 "libraries/enet/enet.cpp"
  if ($16) {
   $port_i$0$lcssa8 = $port_i$1$lcssa;
   label = 11;
  } else {
   $port_i$0$lcssa7 = $port_i$1$lcssa;
  }
 }
 if ((label|0) == 11) {
  (_luaL_error($l,266830,$vararg_buffer1)|0); //@line 71 "libraries/enet/enet.cpp"
  $port_i$0$lcssa7 = $port_i$0$lcssa8;
 }
 $17 = ($port_i$0$lcssa7|0)==(0); //@line 72 "libraries/enet/enet.cpp"
 if ($17) {
  (_luaL_error($l,266854,$vararg_buffer3)|0); //@line 72 "libraries/enet/enet.cpp"
 }
 $18 = (_strcmp(506969,$host_str)|0); //@line 74 "libraries/enet/enet.cpp"
 $19 = ($18|0)==(0); //@line 74 "libraries/enet/enet.cpp"
 if ($19) {
  HEAP32[$address>>2] = 0; //@line 75 "libraries/enet/enet.cpp"
 } else {
  $20 = (_enet_address_set_host($address,$host_str)|0); //@line 77 "libraries/enet/enet.cpp"
  $21 = ($20|0)==(0); //@line 77 "libraries/enet/enet.cpp"
  if (!($21)) {
   (_luaL_error($l,266878,$vararg_buffer5)|0); //@line 78 "libraries/enet/enet.cpp"
  }
 }
 $22 = (_strcmp(506969,$port_str)|0); //@line 82 "libraries/enet/enet.cpp"
 $23 = ($22|0)==(0); //@line 82 "libraries/enet/enet.cpp"
 if ($23) {
  $$sink = 0;
  $26 = ((($address)) + 4|0); //@line 85 "libraries/enet/enet.cpp"
  HEAP16[$26>>1] = $$sink; //@line 85 "libraries/enet/enet.cpp"
  STACKTOP = sp;return; //@line 87 "libraries/enet/enet.cpp"
 }
 $24 = (_atoi($port_str)|0); //@line 85 "libraries/enet/enet.cpp"
 $25 = $24&65535; //@line 85 "libraries/enet/enet.cpp"
 $$sink = $25;
 $26 = ((($address)) + 4|0); //@line 85 "libraries/enet/enet.cpp"
 HEAP16[$26>>1] = $$sink; //@line 85 "libraries/enet/enet.cpp"
 STACKTOP = sp;return; //@line 87 "libraries/enet/enet.cpp"
}
function __ZL10host_flushP9lua_State($l) {
 $l = $l|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 368 "libraries/enet/enet.cpp"
 $0 = (_luaL_checkudata($l,1,266649)|0); //@line 369 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 369 "libraries/enet/enet.cpp"
 $2 = ($1|0)==(0|0); //@line 370 "libraries/enet/enet.cpp"
 if ($2) {
  $3 = (_luaL_error($l,266667,$vararg_buffer)|0); //@line 371 "libraries/enet/enet.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 375 "libraries/enet/enet.cpp"
 } else {
  _enet_host_flush($1); //@line 373 "libraries/enet/enet.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 375 "libraries/enet/enet.cpp"
 }
 return (0)|0;
}
function __ZL14host_broadcastP9lua_State($l) {
 $l = $l|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $channel_id = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $channel_id = sp + 4|0;
 $0 = (_luaL_checkudata($l,1,266649)|0); //@line 378 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 378 "libraries/enet/enet.cpp"
 $2 = ($1|0)==(0|0); //@line 379 "libraries/enet/enet.cpp"
 if ($2) {
  $3 = (_luaL_error($l,266667,$vararg_buffer)|0); //@line 380 "libraries/enet/enet.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 387 "libraries/enet/enet.cpp"
 } else {
  $4 = (__ZL11read_packetP9lua_StateiPh($l,$channel_id)|0); //@line 384 "libraries/enet/enet.cpp"
  $5 = HEAP8[$channel_id>>0]|0; //@line 385 "libraries/enet/enet.cpp"
  _enet_host_broadcast($1,$5,$4); //@line 385 "libraries/enet/enet.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 387 "libraries/enet/enet.cpp"
 }
 return (0)|0;
}
function __ZL11read_packetP9lua_StateiPh($l,$channel_id) {
 $l = $l|0;
 $channel_id = $channel_id|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $flags$11 = 0, $flags$12 = 0, $size = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $size = sp + 12|0;
 $0 = (_lua_gettop($l)|0); //@line 172 "libraries/enet/enet.cpp"
 $1 = (_luaL_checklstring($l,2,$size)|0); //@line 173 "libraries/enet/enet.cpp"
 HEAP8[$channel_id>>0] = 0; //@line 177 "libraries/enet/enet.cpp"
 $2 = ($0|0)>(3); //@line 179 "libraries/enet/enet.cpp"
 if ($2) {
  $3 = (_lua_type($l,4)|0); //@line 179 "libraries/enet/enet.cpp"
  $4 = ($3|0)==(0); //@line 179 "libraries/enet/enet.cpp"
  if ($4) {
   $flags$12 = 1;
   label = 8;
  } else {
   $5 = (_luaL_checklstring($l,4,0)|0); //@line 180 "libraries/enet/enet.cpp"
   $6 = (_strcmp(266928,$5)|0); //@line 181 "libraries/enet/enet.cpp"
   $7 = ($6|0)==(0); //@line 181 "libraries/enet/enet.cpp"
   if ($7) {
    $flags$12 = 2;
    label = 8;
   } else {
    $8 = (_strcmp(266940,$5)|0); //@line 183 "libraries/enet/enet.cpp"
    $9 = ($8|0)==(0); //@line 183 "libraries/enet/enet.cpp"
    if ($9) {
     $flags$12 = 1;
     label = 8;
    } else {
     $10 = (_strcmp(266949,$5)|0); //@line 185 "libraries/enet/enet.cpp"
     $11 = ($10|0)==(0); //@line 185 "libraries/enet/enet.cpp"
     if ($11) {
      $flags$12 = 0;
      label = 8;
     } else {
      HEAP32[$vararg_buffer>>2] = $5; //@line 188 "libraries/enet/enet.cpp"
      (_luaL_error($l,266960,$vararg_buffer)|0); //@line 188 "libraries/enet/enet.cpp"
      $flags$12 = 1;
      label = 8;
     }
    }
   }
  }
 } else {
  $12 = ($0|0)>(2); //@line 192 "libraries/enet/enet.cpp"
  if ($12) {
   $flags$12 = 1;
   label = 8;
  } else {
   $flags$11 = 1;
  }
 }
 if ((label|0) == 8) {
  $13 = (_lua_type($l,3)|0); //@line 192 "libraries/enet/enet.cpp"
  $14 = ($13|0)==(0); //@line 192 "libraries/enet/enet.cpp"
  if ($14) {
   $flags$11 = $flags$12;
  } else {
   $15 = (+_luaL_checknumber($l,3)); //@line 193 "libraries/enet/enet.cpp"
   $16 = (~~(($15))); //@line 193 "libraries/enet/enet.cpp"
   $17 = $16&255; //@line 193 "libraries/enet/enet.cpp"
   HEAP8[$channel_id>>0] = $17; //@line 193 "libraries/enet/enet.cpp"
   $flags$11 = $flags$12;
  }
 }
 $18 = HEAP32[$size>>2]|0; //@line 196 "libraries/enet/enet.cpp"
 $19 = (_enet_packet_create($1,$18,$flags$11)|0); //@line 196 "libraries/enet/enet.cpp"
 $20 = ($19|0)==(0|0); //@line 197 "libraries/enet/enet.cpp"
 if (!($20)) {
  STACKTOP = sp;return ($19|0); //@line 201 "libraries/enet/enet.cpp"
 }
 (_luaL_error($l,266984,$vararg_buffer1)|0); //@line 198 "libraries/enet/enet.cpp"
 STACKTOP = sp;return ($19|0); //@line 201 "libraries/enet/enet.cpp"
}
function __ZL18host_channel_limitP9lua_State($l) {
 $l = $l|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 390 "libraries/enet/enet.cpp"
 $0 = (_luaL_checkudata($l,1,266649)|0); //@line 391 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 391 "libraries/enet/enet.cpp"
 $2 = ($1|0)==(0|0); //@line 392 "libraries/enet/enet.cpp"
 if ($2) {
  $3 = (_luaL_error($l,266667,$vararg_buffer)|0); //@line 393 "libraries/enet/enet.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 398 "libraries/enet/enet.cpp"
 } else {
  $4 = (+_luaL_checknumber($l,2)); //@line 395 "libraries/enet/enet.cpp"
  $5 = (~~(($4))); //@line 395 "libraries/enet/enet.cpp"
  _enet_host_channel_limit($1,$5); //@line 396 "libraries/enet/enet.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 398 "libraries/enet/enet.cpp"
 }
 return (0)|0;
}
function __ZL20host_bandwidth_limitP9lua_State($l) {
 $l = $l|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 400 "libraries/enet/enet.cpp"
 $0 = (_luaL_checkudata($l,1,266649)|0); //@line 401 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 401 "libraries/enet/enet.cpp"
 $2 = ($1|0)==(0|0); //@line 402 "libraries/enet/enet.cpp"
 if ($2) {
  $3 = (_luaL_error($l,266667,$vararg_buffer)|0); //@line 403 "libraries/enet/enet.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 409 "libraries/enet/enet.cpp"
 } else {
  $4 = (+_luaL_checknumber($l,2)); //@line 405 "libraries/enet/enet.cpp"
  $5 = (~~(($4))); //@line 405 "libraries/enet/enet.cpp"
  $6 = (+_luaL_checknumber($l,2)); //@line 406 "libraries/enet/enet.cpp"
  $7 = (~~(($6))); //@line 406 "libraries/enet/enet.cpp"
  _enet_host_bandwidth_limit($1,$5,$7); //@line 407 "libraries/enet/enet.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 409 "libraries/enet/enet.cpp"
 }
 return (0)|0;
}
function __ZL23host_get_socket_addressP9lua_State($l) {
 $l = $l|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $address = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_ptr3 = 0;
 var $vararg_ptr4 = 0, $vararg_ptr5 = 0, $vararg_ptr6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $address = sp + 32|0;
 $0 = (_luaL_checkudata($l,1,266649)|0); //@line 412 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 412 "libraries/enet/enet.cpp"
 $2 = ($1|0)==(0|0); //@line 413 "libraries/enet/enet.cpp"
 if ($2) {
  $3 = (_luaL_error($l,266667,$vararg_buffer)|0); //@line 414 "libraries/enet/enet.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 426 "libraries/enet/enet.cpp"
 } else {
  $4 = HEAP32[$1>>2]|0; //@line 417 "libraries/enet/enet.cpp"
  (_enet_socket_get_address($4,$address)|0); //@line 417 "libraries/enet/enet.cpp"
  $5 = HEAP32[$address>>2]|0; //@line 420 "libraries/enet/enet.cpp"
  $6 = $5 & 255; //@line 420 "libraries/enet/enet.cpp"
  $7 = $5 >>> 8; //@line 421 "libraries/enet/enet.cpp"
  $8 = $7 & 255; //@line 421 "libraries/enet/enet.cpp"
  $9 = $5 >>> 16; //@line 422 "libraries/enet/enet.cpp"
  $10 = $9 & 255; //@line 422 "libraries/enet/enet.cpp"
  $11 = $5 >>> 24; //@line 423 "libraries/enet/enet.cpp"
  $12 = ((($address)) + 4|0); //@line 424 "libraries/enet/enet.cpp"
  $13 = HEAP16[$12>>1]|0; //@line 424 "libraries/enet/enet.cpp"
  $14 = $13&65535; //@line 424 "libraries/enet/enet.cpp"
  HEAP32[$vararg_buffer1>>2] = $6; //@line 419 "libraries/enet/enet.cpp"
  $vararg_ptr3 = ((($vararg_buffer1)) + 4|0); //@line 419 "libraries/enet/enet.cpp"
  HEAP32[$vararg_ptr3>>2] = $8; //@line 419 "libraries/enet/enet.cpp"
  $vararg_ptr4 = ((($vararg_buffer1)) + 8|0); //@line 419 "libraries/enet/enet.cpp"
  HEAP32[$vararg_ptr4>>2] = $10; //@line 419 "libraries/enet/enet.cpp"
  $vararg_ptr5 = ((($vararg_buffer1)) + 12|0); //@line 419 "libraries/enet/enet.cpp"
  HEAP32[$vararg_ptr5>>2] = $11; //@line 419 "libraries/enet/enet.cpp"
  $vararg_ptr6 = ((($vararg_buffer1)) + 16|0); //@line 419 "libraries/enet/enet.cpp"
  HEAP32[$vararg_ptr6>>2] = $14; //@line 419 "libraries/enet/enet.cpp"
  (_lua_pushfstring($l,267057,$vararg_buffer1)|0); //@line 419 "libraries/enet/enet.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 426 "libraries/enet/enet.cpp"
 }
 return (0)|0;
}
function __ZL7host_gcP9lua_State($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkudata($l,1,266649)|0); //@line 490 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 492 "libraries/enet/enet.cpp"
 $2 = ($1|0)==(0|0); //@line 492 "libraries/enet/enet.cpp"
 if (!($2)) {
  _enet_host_destroy($1); //@line 493 "libraries/enet/enet.cpp"
 }
 HEAP32[$0>>2] = 0; //@line 495 "libraries/enet/enet.cpp"
 return 0; //@line 496 "libraries/enet/enet.cpp"
}
function __ZL20host_total_sent_dataP9lua_State($l) {
 $l = $l|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 427 "libraries/enet/enet.cpp"
 $0 = (_luaL_checkudata($l,1,266649)|0); //@line 428 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 428 "libraries/enet/enet.cpp"
 $2 = ($1|0)==(0|0); //@line 429 "libraries/enet/enet.cpp"
 if ($2) {
  $3 = (_luaL_error($l,266667,$vararg_buffer)|0); //@line 430 "libraries/enet/enet.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 436 "libraries/enet/enet.cpp"
 } else {
  $4 = ((($1)) + 10364|0); //@line 433 "libraries/enet/enet.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 433 "libraries/enet/enet.cpp"
  _lua_pushinteger($l,$5); //@line 433 "libraries/enet/enet.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 436 "libraries/enet/enet.cpp"
 }
 return (0)|0;
}
function __ZL24host_total_received_dataP9lua_State($l) {
 $l = $l|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 438 "libraries/enet/enet.cpp"
 $0 = (_luaL_checkudata($l,1,266649)|0); //@line 439 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 439 "libraries/enet/enet.cpp"
 $2 = ($1|0)==(0|0); //@line 440 "libraries/enet/enet.cpp"
 if ($2) {
  $3 = (_luaL_error($l,266667,$vararg_buffer)|0); //@line 441 "libraries/enet/enet.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 447 "libraries/enet/enet.cpp"
 } else {
  $4 = ((($1)) + 10372|0); //@line 444 "libraries/enet/enet.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 444 "libraries/enet/enet.cpp"
  _lua_pushinteger($l,$5); //@line 444 "libraries/enet/enet.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 447 "libraries/enet/enet.cpp"
 }
 return (0)|0;
}
function __ZL17host_service_timeP9lua_State($l) {
 $l = $l|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 448 "libraries/enet/enet.cpp"
 $0 = (_luaL_checkudata($l,1,266649)|0); //@line 449 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 449 "libraries/enet/enet.cpp"
 $2 = ($1|0)==(0|0); //@line 450 "libraries/enet/enet.cpp"
 if ($2) {
  $3 = (_luaL_error($l,266667,$vararg_buffer)|0); //@line 451 "libraries/enet/enet.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 457 "libraries/enet/enet.cpp"
 } else {
  $4 = ((($1)) + 48|0); //@line 454 "libraries/enet/enet.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 454 "libraries/enet/enet.cpp"
  _lua_pushinteger($l,$5); //@line 454 "libraries/enet/enet.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 457 "libraries/enet/enet.cpp"
 }
 return (0)|0;
}
function __ZL15host_peer_countP9lua_State($l) {
 $l = $l|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 459 "libraries/enet/enet.cpp"
 $0 = (_luaL_checkudata($l,1,266649)|0); //@line 460 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 460 "libraries/enet/enet.cpp"
 $2 = ($1|0)==(0|0); //@line 461 "libraries/enet/enet.cpp"
 if ($2) {
  $3 = (_luaL_error($l,266667,$vararg_buffer)|0); //@line 462 "libraries/enet/enet.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 468 "libraries/enet/enet.cpp"
 } else {
  $4 = ((($1)) + 40|0); //@line 465 "libraries/enet/enet.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 465 "libraries/enet/enet.cpp"
  _lua_pushinteger($l,$5); //@line 465 "libraries/enet/enet.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 468 "libraries/enet/enet.cpp"
 }
 return (0)|0;
}
function __ZL13host_get_peerP9lua_State($l) {
 $l = $l|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 470 "libraries/enet/enet.cpp"
 $0 = (_luaL_checkudata($l,1,266649)|0); //@line 471 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 471 "libraries/enet/enet.cpp"
 $2 = ($1|0)==(0|0); //@line 472 "libraries/enet/enet.cpp"
 if ($2) {
  $3 = (_luaL_error($l,266667,$vararg_buffer)|0); //@line 473 "libraries/enet/enet.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 486 "libraries/enet/enet.cpp"
 }
 $4 = (+_luaL_checknumber($l,2)); //@line 476 "libraries/enet/enet.cpp"
 $5 = (~~(($4))); //@line 476 "libraries/enet/enet.cpp"
 $6 = (($5) + -1)|0; //@line 476 "libraries/enet/enet.cpp"
 $7 = ($5|0)<(1); //@line 478 "libraries/enet/enet.cpp"
 if ($7) {
  label = 5;
 } else {
  $8 = ((($1)) + 40|0); //@line 478 "libraries/enet/enet.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 478 "libraries/enet/enet.cpp"
  $10 = ($6>>>0)<($9>>>0); //@line 478 "libraries/enet/enet.cpp"
  if (!($10)) {
   label = 5;
  }
 }
 if ((label|0) == 5) {
  (_luaL_argerror($l,2,267160)|0); //@line 479 "libraries/enet/enet.cpp"
 }
 $11 = ((($1)) + 36|0); //@line 482 "libraries/enet/enet.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 482 "libraries/enet/enet.cpp"
 $13 = (($12) + (($6*380)|0)|0); //@line 482 "libraries/enet/enet.cpp"
 __ZL9push_peerP9lua_StateP9_ENetPeer($l,$13); //@line 484 "libraries/enet/enet.cpp"
 $$0 = 1;
 STACKTOP = sp;return ($$0|0); //@line 486 "libraries/enet/enet.cpp"
}
function __ZL15peer_disconnectP9lua_State($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkudata($l,1,266726)|0); //@line 592 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 592 "libraries/enet/enet.cpp"
 $2 = (_lua_gettop($l)|0); //@line 594 "libraries/enet/enet.cpp"
 $3 = ($2|0)>(1); //@line 594 "libraries/enet/enet.cpp"
 if ($3) {
  $4 = (+_luaL_checknumber($l,2)); //@line 594 "libraries/enet/enet.cpp"
  $5 = (~~(($4))); //@line 594 "libraries/enet/enet.cpp"
  $6 = $5;
 } else {
  $6 = 0;
 }
 _enet_peer_disconnect($1,$6); //@line 595 "libraries/enet/enet.cpp"
 return 0; //@line 596 "libraries/enet/enet.cpp"
}
function __ZL19peer_disconnect_nowP9lua_State($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkudata($l,1,266726)|0); //@line 600 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 600 "libraries/enet/enet.cpp"
 $2 = (_lua_gettop($l)|0); //@line 602 "libraries/enet/enet.cpp"
 $3 = ($2|0)>(1); //@line 602 "libraries/enet/enet.cpp"
 if ($3) {
  $4 = (+_luaL_checknumber($l,2)); //@line 602 "libraries/enet/enet.cpp"
  $5 = (~~(($4))); //@line 602 "libraries/enet/enet.cpp"
  $6 = $5;
 } else {
  $6 = 0;
 }
 _enet_peer_disconnect_now($1,$6); //@line 603 "libraries/enet/enet.cpp"
 return 0; //@line 604 "libraries/enet/enet.cpp"
}
function __ZL21peer_disconnect_laterP9lua_State($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkudata($l,1,266726)|0); //@line 608 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 608 "libraries/enet/enet.cpp"
 $2 = (_lua_gettop($l)|0); //@line 610 "libraries/enet/enet.cpp"
 $3 = ($2|0)>(1); //@line 610 "libraries/enet/enet.cpp"
 if ($3) {
  $4 = (+_luaL_checknumber($l,2)); //@line 610 "libraries/enet/enet.cpp"
  $5 = (~~(($4))); //@line 610 "libraries/enet/enet.cpp"
  $6 = $5;
 } else {
  $6 = 0;
 }
 _enet_peer_disconnect_later($1,$6); //@line 611 "libraries/enet/enet.cpp"
 return 0; //@line 612 "libraries/enet/enet.cpp"
}
function __ZL10peer_resetP9lua_State($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkudata($l,1,266726)|0); //@line 675 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 675 "libraries/enet/enet.cpp"
 _enet_peer_reset($1); //@line 676 "libraries/enet/enet.cpp"
 return 0; //@line 677 "libraries/enet/enet.cpp"
}
function __ZL9peer_pingP9lua_State($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkudata($l,1,266726)|0); //@line 512 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 512 "libraries/enet/enet.cpp"
 _enet_peer_ping($1); //@line 513 "libraries/enet/enet.cpp"
 return 0; //@line 514 "libraries/enet/enet.cpp"
}
function __ZL12peer_receiveP9lua_State($l) {
 $l = $l|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $channel_id = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $channel_id = sp;
 $0 = (_luaL_checkudata($l,1,266726)|0); //@line 681 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 681 "libraries/enet/enet.cpp"
 HEAP8[$channel_id>>0] = 0; //@line 683 "libraries/enet/enet.cpp"
 $2 = (_lua_gettop($l)|0); //@line 685 "libraries/enet/enet.cpp"
 $3 = ($2|0)>(1); //@line 685 "libraries/enet/enet.cpp"
 if ($3) {
  $4 = (+_luaL_checknumber($l,2)); //@line 686 "libraries/enet/enet.cpp"
  $5 = (~~(($4))); //@line 686 "libraries/enet/enet.cpp"
  $6 = $5&255; //@line 686 "libraries/enet/enet.cpp"
  HEAP8[$channel_id>>0] = $6; //@line 686 "libraries/enet/enet.cpp"
 }
 $7 = (_enet_peer_receive($1,$channel_id)|0); //@line 689 "libraries/enet/enet.cpp"
 $8 = ($7|0)==(0|0); //@line 690 "libraries/enet/enet.cpp"
 if ($8) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 697 "libraries/enet/enet.cpp"
 }
 $9 = ((($7)) + 8|0); //@line 692 "libraries/enet/enet.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 692 "libraries/enet/enet.cpp"
 $11 = ((($7)) + 12|0); //@line 692 "libraries/enet/enet.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 692 "libraries/enet/enet.cpp"
 _lua_pushlstring($l,$10,$12); //@line 692 "libraries/enet/enet.cpp"
 $13 = HEAP8[$channel_id>>0]|0; //@line 693 "libraries/enet/enet.cpp"
 $14 = $13&255; //@line 693 "libraries/enet/enet.cpp"
 _lua_pushinteger($l,$14); //@line 693 "libraries/enet/enet.cpp"
 _enet_packet_destroy($7); //@line 695 "libraries/enet/enet.cpp"
 $$0 = 2;
 STACKTOP = sp;return ($$0|0); //@line 697 "libraries/enet/enet.cpp"
}
function __ZL9peer_sendP9lua_State($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $channel_id = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $channel_id = sp;
 $0 = (_luaL_checkudata($l,1,266726)|0); //@line 709 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 709 "libraries/enet/enet.cpp"
 $2 = (__ZL11read_packetP9lua_StateiPh($l,$channel_id)|0); //@line 712 "libraries/enet/enet.cpp"
 $3 = HEAP8[$channel_id>>0]|0; //@line 715 "libraries/enet/enet.cpp"
 (_enet_peer_send($1,$3,$2)|0); //@line 715 "libraries/enet/enet.cpp"
 STACKTOP = sp;return 0; //@line 716 "libraries/enet/enet.cpp"
}
function __ZL23peer_throttle_configureP9lua_State($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkudata($l,1,266726)|0); //@line 518 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 518 "libraries/enet/enet.cpp"
 $2 = (+_luaL_checknumber($l,2)); //@line 520 "libraries/enet/enet.cpp"
 $3 = (~~(($2))); //@line 520 "libraries/enet/enet.cpp"
 $4 = (+_luaL_checknumber($l,3)); //@line 521 "libraries/enet/enet.cpp"
 $5 = (~~(($4))); //@line 521 "libraries/enet/enet.cpp"
 $6 = (+_luaL_checknumber($l,4)); //@line 522 "libraries/enet/enet.cpp"
 $7 = (~~(($6))); //@line 522 "libraries/enet/enet.cpp"
 _enet_peer_throttle_configure($1,$3,$5,$7); //@line 524 "libraries/enet/enet.cpp"
 return 0; //@line 525 "libraries/enet/enet.cpp"
}
function __ZL18peer_ping_intervalP9lua_State($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkudata($l,1,266726)|0); //@line 554 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 554 "libraries/enet/enet.cpp"
 $2 = (_lua_gettop($l)|0); //@line 556 "libraries/enet/enet.cpp"
 $3 = ($2|0)>(1); //@line 556 "libraries/enet/enet.cpp"
 if ($3) {
  $4 = (+_luaL_checknumber($l,2)); //@line 557 "libraries/enet/enet.cpp"
  $5 = (~~(($4))); //@line 557 "libraries/enet/enet.cpp"
  _enet_peer_ping_interval($1,$5); //@line 558 "libraries/enet/enet.cpp"
 }
 $6 = ((($1)) + 136|0); //@line 561 "libraries/enet/enet.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 561 "libraries/enet/enet.cpp"
 _lua_pushinteger($l,$7); //@line 561 "libraries/enet/enet.cpp"
 return 1; //@line 563 "libraries/enet/enet.cpp"
}
function __ZL12peer_timeoutP9lua_State($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0.0, $timeout_limit$0 = 0, $timeout_maximum$0 = 0, $timeout_maximum$1 = 0, $timeout_maximum$2 = 0, $timeout_minimum$0 = 0, $timeout_minimum$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkudata($l,1,266726)|0); //@line 567 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 567 "libraries/enet/enet.cpp"
 $2 = (_lua_gettop($l)|0); //@line 573 "libraries/enet/enet.cpp"
 switch ($2|0) {
 case 4:  {
  $3 = (_lua_type($l,4)|0); //@line 575 "libraries/enet/enet.cpp"
  $4 = ($3|0)==(0); //@line 575 "libraries/enet/enet.cpp"
  if ($4) {
   $timeout_maximum$0 = 0;
   label = 4;
  } else {
   $5 = (+_luaL_checknumber($l,4)); //@line 575 "libraries/enet/enet.cpp"
   $6 = (~~(($5))); //@line 575 "libraries/enet/enet.cpp"
   $timeout_maximum$0 = $6;
   label = 4;
  }
  break;
 }
 case 3:  {
  $timeout_maximum$0 = 0;
  label = 4;
  break;
 }
 case 2:  {
  $timeout_maximum$1 = 0;$timeout_minimum$0 = 0;
  label = 6;
  break;
 }
 default: {
  $timeout_limit$0 = 0;$timeout_maximum$2 = 0;$timeout_minimum$1 = 0;
 }
 }
 if ((label|0) == 4) {
  $7 = (_lua_type($l,3)|0); //@line 577 "libraries/enet/enet.cpp"
  $8 = ($7|0)==(0); //@line 577 "libraries/enet/enet.cpp"
  if ($8) {
   $timeout_maximum$1 = $timeout_maximum$0;$timeout_minimum$0 = 0;
   label = 6;
  } else {
   $9 = (+_luaL_checknumber($l,3)); //@line 577 "libraries/enet/enet.cpp"
   $10 = (~~(($9))); //@line 577 "libraries/enet/enet.cpp"
   $timeout_maximum$1 = $timeout_maximum$0;$timeout_minimum$0 = $10;
   label = 6;
  }
 }
 if ((label|0) == 6) {
  $11 = (_lua_type($l,2)|0); //@line 579 "libraries/enet/enet.cpp"
  $12 = ($11|0)==(0); //@line 579 "libraries/enet/enet.cpp"
  if ($12) {
   $timeout_limit$0 = 0;$timeout_maximum$2 = $timeout_maximum$1;$timeout_minimum$1 = $timeout_minimum$0;
  } else {
   $13 = (+_luaL_checknumber($l,2)); //@line 579 "libraries/enet/enet.cpp"
   $14 = (~~(($13))); //@line 579 "libraries/enet/enet.cpp"
   $timeout_limit$0 = $14;$timeout_maximum$2 = $timeout_maximum$1;$timeout_minimum$1 = $timeout_minimum$0;
  }
 }
 _enet_peer_timeout($1,$timeout_limit$0,$timeout_minimum$1,$timeout_maximum$2); //@line 582 "libraries/enet/enet.cpp"
 $15 = ((($1)) + 140|0); //@line 584 "libraries/enet/enet.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 584 "libraries/enet/enet.cpp"
 _lua_pushinteger($l,$16); //@line 584 "libraries/enet/enet.cpp"
 $17 = ((($1)) + 144|0); //@line 585 "libraries/enet/enet.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 585 "libraries/enet/enet.cpp"
 _lua_pushinteger($l,$18); //@line 585 "libraries/enet/enet.cpp"
 $19 = ((($1)) + 148|0); //@line 586 "libraries/enet/enet.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 586 "libraries/enet/enet.cpp"
 _lua_pushinteger($l,$20); //@line 586 "libraries/enet/enet.cpp"
 return 3; //@line 588 "libraries/enet/enet.cpp"
}
function __ZL10peer_indexP9lua_State($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $peer_index$0$lcssa$i = 0, $peer_index$02$i = 0, $peer_index$03$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 615 "libraries/enet/enet.cpp"
 $0 = (_luaL_checkudata($l,1,266726)|0); //@line 616 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 616 "libraries/enet/enet.cpp"
 $2 = ((($1)) + 8|0); //@line 618 "libraries/enet/enet.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 618 "libraries/enet/enet.cpp"
 $4 = ((($3)) + 40|0); //@line 94 "libraries/enet/enet.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 94 "libraries/enet/enet.cpp"
 $6 = ($5|0)==(0); //@line 94 "libraries/enet/enet.cpp"
 L1: do {
  if ($6) {
   $peer_index$0$lcssa$i = 0;
  } else {
   $7 = ((($3)) + 36|0); //@line 95 "libraries/enet/enet.cpp"
   $8 = HEAP32[$7>>2]|0; //@line 95 "libraries/enet/enet.cpp"
   $peer_index$03$i = 0;
   while(1) {
    $9 = (($8) + (($peer_index$03$i*380)|0)|0); //@line 95 "libraries/enet/enet.cpp"
    $10 = ($9|0)==($1|0); //@line 95 "libraries/enet/enet.cpp"
    if ($10) {
     $peer_index$02$i = $peer_index$03$i;
     break;
    }
    $11 = (($peer_index$03$i) + 1)|0; //@line 94 "libraries/enet/enet.cpp"
    $12 = ($11>>>0)<($5>>>0); //@line 94 "libraries/enet/enet.cpp"
    if ($12) {
     $peer_index$03$i = $11;
    } else {
     $peer_index$0$lcssa$i = $11;
     break L1;
    }
   }
   $13 = (($peer_index$02$i) + 1)|0; //@line 619 "libraries/enet/enet.cpp"
   _lua_pushinteger($l,$13); //@line 619 "libraries/enet/enet.cpp"
   STACKTOP = sp;return 1; //@line 621 "libraries/enet/enet.cpp"
  }
 } while(0);
 (_luaL_error($l,267249,$vararg_buffer)|0); //@line 99 "libraries/enet/enet.cpp"
 $peer_index$02$i = $peer_index$0$lcssa$i;
 $13 = (($peer_index$02$i) + 1)|0; //@line 619 "libraries/enet/enet.cpp"
 _lua_pushinteger($l,$13); //@line 619 "libraries/enet/enet.cpp"
 STACKTOP = sp;return 1; //@line 621 "libraries/enet/enet.cpp"
}
function __ZL10peer_stateP9lua_State($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkudata($l,1,266726)|0); //@line 625 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 625 "libraries/enet/enet.cpp"
 $2 = ((($1)) + 36|0); //@line 627 "libraries/enet/enet.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 627 "libraries/enet/enet.cpp"
 do {
  switch ($3|0) {
  case 0:  {
   _lua_pushstring($l,267285); //@line 629 "libraries/enet/enet.cpp"
   break;
  }
  case 1:  {
   _lua_pushstring($l,267298); //@line 632 "libraries/enet/enet.cpp"
   break;
  }
  case 2:  {
   _lua_pushstring($l,267309); //@line 635 "libraries/enet/enet.cpp"
   break;
  }
  case 3:  {
   _lua_pushstring($l,267331); //@line 638 "libraries/enet/enet.cpp"
   break;
  }
  case 4:  {
   _lua_pushstring($l,267350); //@line 641 "libraries/enet/enet.cpp"
   break;
  }
  case 5:  {
   _lua_pushstring($l,267371); //@line 644 "libraries/enet/enet.cpp"
   break;
  }
  case 6:  {
   _lua_pushstring($l,267194); //@line 647 "libraries/enet/enet.cpp"
   break;
  }
  case 7:  {
   _lua_pushstring($l,267381); //@line 650 "libraries/enet/enet.cpp"
   break;
  }
  case 8:  {
   _lua_pushstring($l,267395); //@line 653 "libraries/enet/enet.cpp"
   break;
  }
  case 9:  {
   _lua_pushstring($l,267420); //@line 656 "libraries/enet/enet.cpp"
   break;
  }
  default: {
   _lua_pushstring($l,442424); //@line 659 "libraries/enet/enet.cpp"
  }
  }
 } while(0);
 return 1; //@line 662 "libraries/enet/enet.cpp"
}
function __ZL15peer_connect_idP9lua_State($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkudata($l,1,266726)|0); //@line 666 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 666 "libraries/enet/enet.cpp"
 $2 = ((($1)) + 16|0); //@line 668 "libraries/enet/enet.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 668 "libraries/enet/enet.cpp"
 _lua_pushinteger($l,$3); //@line 668 "libraries/enet/enet.cpp"
 return 1; //@line 670 "libraries/enet/enet.cpp"
}
function __ZL20peer_round_trip_timeP9lua_State($l) {
 $l = $l|0;
 var $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkudata($l,1,266726)|0); //@line 529 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 529 "libraries/enet/enet.cpp"
 $2 = (_lua_gettop($l)|0); //@line 531 "libraries/enet/enet.cpp"
 $3 = ($2|0)>(1); //@line 531 "libraries/enet/enet.cpp"
 if ($3) {
  $4 = (+_luaL_checknumber($l,2)); //@line 532 "libraries/enet/enet.cpp"
  $5 = (~~(($4))); //@line 532 "libraries/enet/enet.cpp"
  $6 = ((($1)) + 168|0); //@line 533 "libraries/enet/enet.cpp"
  HEAP32[$6>>2] = $5; //@line 533 "libraries/enet/enet.cpp"
  $7 = $5;
  _lua_pushinteger($l,$7); //@line 536 "libraries/enet/enet.cpp"
  return 1; //@line 538 "libraries/enet/enet.cpp"
 } else {
  $$phi$trans$insert = ((($1)) + 168|0); //@line 536 "libraries/enet/enet.cpp"
  $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 536 "libraries/enet/enet.cpp"
  $7 = $$pre;
  _lua_pushinteger($l,$7); //@line 536 "libraries/enet/enet.cpp"
  return 1; //@line 538 "libraries/enet/enet.cpp"
 }
 return (0)|0;
}
function __ZL25peer_last_round_trip_timeP9lua_State($l) {
 $l = $l|0;
 var $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkudata($l,1,266726)|0); //@line 542 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 542 "libraries/enet/enet.cpp"
 $2 = (_lua_gettop($l)|0); //@line 544 "libraries/enet/enet.cpp"
 $3 = ($2|0)>(1); //@line 544 "libraries/enet/enet.cpp"
 if ($3) {
  $4 = (+_luaL_checknumber($l,2)); //@line 545 "libraries/enet/enet.cpp"
  $5 = (~~(($4))); //@line 545 "libraries/enet/enet.cpp"
  $6 = ((($1)) + 152|0); //@line 546 "libraries/enet/enet.cpp"
  HEAP32[$6>>2] = $5; //@line 546 "libraries/enet/enet.cpp"
  $7 = $5;
  _lua_pushinteger($l,$7); //@line 548 "libraries/enet/enet.cpp"
  return 1; //@line 550 "libraries/enet/enet.cpp"
 } else {
  $$phi$trans$insert = ((($1)) + 152|0); //@line 548 "libraries/enet/enet.cpp"
  $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 548 "libraries/enet/enet.cpp"
  $7 = $$pre;
  _lua_pushinteger($l,$7); //@line 548 "libraries/enet/enet.cpp"
  return 1; //@line 550 "libraries/enet/enet.cpp"
 }
 return (0)|0;
}
function __ZL13peer_tostringP9lua_State($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $host_str = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $host_str = sp;
 $0 = (_luaL_checkudata($l,1,266726)|0); //@line 500 "libraries/enet/enet.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 500 "libraries/enet/enet.cpp"
 $2 = ((($1)) + 24|0); //@line 502 "libraries/enet/enet.cpp"
 (_enet_address_get_host_ip($2,$host_str,128)|0); //@line 502 "libraries/enet/enet.cpp"
 _lua_pushstring($l,$host_str); //@line 504 "libraries/enet/enet.cpp"
 _lua_pushstring($l,442485); //@line 505 "libraries/enet/enet.cpp"
 $3 = ((($1)) + 28|0); //@line 506 "libraries/enet/enet.cpp"
 $4 = HEAP16[$3>>1]|0; //@line 506 "libraries/enet/enet.cpp"
 $5 = $4&65535; //@line 506 "libraries/enet/enet.cpp"
 _lua_pushinteger($l,$5); //@line 506 "libraries/enet/enet.cpp"
 _lua_concat($l,3); //@line 507 "libraries/enet/enet.cpp"
 STACKTOP = sp;return 1; //@line 508 "libraries/enet/enet.cpp"
}
function __ZL11host_createP9lua_State($l) {
 $l = $l|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0, $address = 0, $address$ = 0, $channel_count$0 = 0, $channel_count$1 = 0, $have_address$0 = 0, $in_bandwidth$0 = 0, $in_bandwidth$1 = 0, $in_bandwidth$2 = 0, $out_bandwidth$0 = 0, $out_bandwidth$1 = 0, $out_bandwidth$2 = 0, $out_bandwidth$3 = 0, $peer_count$0 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $address = sp;
 $0 = (_lua_gettop($l)|0); //@line 221 "libraries/enet/enet.cpp"
 $1 = ($0|0)==(0); //@line 221 "libraries/enet/enet.cpp"
 if ($1) {
  $have_address$0 = 0;
 } else {
  $2 = (_lua_type($l,1)|0); //@line 221 "libraries/enet/enet.cpp"
  $3 = ($2|0)==(0); //@line 221 "libraries/enet/enet.cpp"
  if ($3) {
   $have_address$0 = 0;
  } else {
   $4 = (_luaL_checklstring($l,1,0)|0); //@line 224 "libraries/enet/enet.cpp"
   __ZL13parse_addressP9lua_StatePKcP12_ENetAddress($l,$4,$address); //@line 224 "libraries/enet/enet.cpp"
   $have_address$0 = 1;
  }
 }
 $5 = (_lua_gettop($l)|0); //@line 227 "libraries/enet/enet.cpp"
 switch ($5|0) {
 case 5:  {
  $6 = (_lua_type($l,5)|0); //@line 229 "libraries/enet/enet.cpp"
  $7 = ($6|0)==(0); //@line 229 "libraries/enet/enet.cpp"
  if ($7) {
   $out_bandwidth$0 = 0;
   label = 7;
  } else {
   $8 = (+_luaL_checknumber($l,5)); //@line 229 "libraries/enet/enet.cpp"
   $9 = (~~(($8))); //@line 229 "libraries/enet/enet.cpp"
   $out_bandwidth$0 = $9;
   label = 7;
  }
  break;
 }
 case 4:  {
  $out_bandwidth$0 = 0;
  label = 7;
  break;
 }
 case 3:  {
  $in_bandwidth$0 = 0;$out_bandwidth$1 = 0;
  label = 9;
  break;
 }
 case 2:  {
  $channel_count$0 = 1;$in_bandwidth$1 = 0;$out_bandwidth$2 = 0;
  label = 11;
  break;
 }
 default: {
  $channel_count$1 = 1;$in_bandwidth$2 = 0;$out_bandwidth$3 = 0;$peer_count$0 = 64;
 }
 }
 if ((label|0) == 7) {
  $10 = (_lua_type($l,4)|0); //@line 231 "libraries/enet/enet.cpp"
  $11 = ($10|0)==(0); //@line 231 "libraries/enet/enet.cpp"
  if ($11) {
   $in_bandwidth$0 = 0;$out_bandwidth$1 = $out_bandwidth$0;
   label = 9;
  } else {
   $12 = (+_luaL_checknumber($l,4)); //@line 231 "libraries/enet/enet.cpp"
   $13 = (~~(($12))); //@line 231 "libraries/enet/enet.cpp"
   $in_bandwidth$0 = $13;$out_bandwidth$1 = $out_bandwidth$0;
   label = 9;
  }
 }
 if ((label|0) == 9) {
  $14 = (_lua_type($l,3)|0); //@line 233 "libraries/enet/enet.cpp"
  $15 = ($14|0)==(0); //@line 233 "libraries/enet/enet.cpp"
  if ($15) {
   $channel_count$0 = 1;$in_bandwidth$1 = $in_bandwidth$0;$out_bandwidth$2 = $out_bandwidth$1;
   label = 11;
  } else {
   $16 = (+_luaL_checknumber($l,3)); //@line 233 "libraries/enet/enet.cpp"
   $17 = (~~(($16))); //@line 233 "libraries/enet/enet.cpp"
   $channel_count$0 = $17;$in_bandwidth$1 = $in_bandwidth$0;$out_bandwidth$2 = $out_bandwidth$1;
   label = 11;
  }
 }
 if ((label|0) == 11) {
  $18 = (_lua_type($l,2)|0); //@line 235 "libraries/enet/enet.cpp"
  $19 = ($18|0)==(0); //@line 235 "libraries/enet/enet.cpp"
  if ($19) {
   $channel_count$1 = $channel_count$0;$in_bandwidth$2 = $in_bandwidth$1;$out_bandwidth$3 = $out_bandwidth$2;$peer_count$0 = 64;
  } else {
   $20 = (+_luaL_checknumber($l,2)); //@line 235 "libraries/enet/enet.cpp"
   $21 = (~~(($20))); //@line 235 "libraries/enet/enet.cpp"
   $channel_count$1 = $channel_count$0;$in_bandwidth$2 = $in_bandwidth$1;$out_bandwidth$3 = $out_bandwidth$2;$peer_count$0 = $21;
  }
 }
 $22 = ($have_address$0|0)!=(0); //@line 240 "libraries/enet/enet.cpp"
 $address$ = $22 ? $address : 0; //@line 240 "libraries/enet/enet.cpp"
 $23 = (_enet_host_create($address$,$peer_count$0,$channel_count$1,$in_bandwidth$2,$out_bandwidth$3)|0); //@line 240 "libraries/enet/enet.cpp"
 $24 = ($23|0)==(0|0); //@line 243 "libraries/enet/enet.cpp"
 if ($24) {
  _lua_pushnil($l); //@line 244 "libraries/enet/enet.cpp"
  _lua_pushstring($l,267492); //@line 245 "libraries/enet/enet.cpp"
  $$0 = 2;
  STACKTOP = sp;return ($$0|0); //@line 254 "libraries/enet/enet.cpp"
 } else {
  $25 = (_lua_newuserdata($l,4)|0); //@line 249 "libraries/enet/enet.cpp"
  HEAP32[$25>>2] = $23; //@line 249 "libraries/enet/enet.cpp"
  _lua_getfield($l,-10000,266649); //@line 250 "libraries/enet/enet.cpp"
  (_lua_setmetatable($l,-2)|0); //@line 251 "libraries/enet/enet.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 254 "libraries/enet/enet.cpp"
 }
 return (0)|0;
}
function __ZL14linked_versionP9lua_State($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 256 "libraries/enet/enet.cpp"
 $0 = (_enet_linked_version()|0); //@line 258 "libraries/enet/enet.cpp"
 $1 = $0 >>> 16; //@line 258 "libraries/enet/enet.cpp"
 $2 = $1 & 255; //@line 258 "libraries/enet/enet.cpp"
 $3 = (_enet_linked_version()|0); //@line 259 "libraries/enet/enet.cpp"
 $4 = $3 >>> 8; //@line 259 "libraries/enet/enet.cpp"
 $5 = $4 & 255; //@line 259 "libraries/enet/enet.cpp"
 $6 = (_enet_linked_version()|0); //@line 260 "libraries/enet/enet.cpp"
 $7 = $6 & 255; //@line 260 "libraries/enet/enet.cpp"
 HEAP32[$vararg_buffer>>2] = $2; //@line 257 "libraries/enet/enet.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 257 "libraries/enet/enet.cpp"
 HEAP32[$vararg_ptr1>>2] = $5; //@line 257 "libraries/enet/enet.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 257 "libraries/enet/enet.cpp"
 HEAP32[$vararg_ptr2>>2] = $7; //@line 257 "libraries/enet/enet.cpp"
 (_lua_pushfstring($l,267556,$vararg_buffer)|0); //@line 257 "libraries/enet/enet.cpp"
 STACKTOP = sp;return 1; //@line 261 "libraries/enet/enet.cpp"
}
function _enet_linked_version() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 66315;
}
function _enet_malloc($size) {
 $size = $size|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $memory = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $size;
 $1 = HEAP32[3029]|0;
 $2 = $0;
 $3 = (FUNCTION_TABLE_ii[$1 & 8191]($2)|0);
 $memory = $3;
 $4 = $memory;
 $5 = ($4|0)==(0|0);
 if (!($5)) {
  $7 = $memory;
  STACKTOP = sp;return ($7|0);
 }
 $6 = HEAP32[(12124)>>2]|0;
 FUNCTION_TABLE_v[$6 & 8191]();
 $7 = $memory;
 STACKTOP = sp;return ($7|0);
}
function _enet_free($memory) {
 $memory = $memory|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $memory;
 $1 = HEAP32[(12120)>>2]|0;
 $2 = $0;
 FUNCTION_TABLE_vi[$1 & 8191]($2);
 STACKTOP = sp;return;
}
function _enet_range_coder_create() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $rangeCoder = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = (_enet_malloc(65536)|0);
 $rangeCoder = $1;
 $2 = $rangeCoder;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $0 = 0;
 } else {
  $4 = $rangeCoder;
  $0 = $4;
 }
 $5 = $0;
 STACKTOP = sp;return ($5|0);
}
function _enet_range_coder_destroy($context) {
 $context = $context|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $rangeCoder = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $context;
 $1 = $0;
 $rangeCoder = $1;
 $2 = $rangeCoder;
 $3 = ($2|0)==(0|0);
 if ($3) {
  STACKTOP = sp;return;
 }
 $4 = $rangeCoder;
 _enet_free($4);
 STACKTOP = sp;return;
}
function _enet_range_coder_compress($context,$inBuffers,$inBufferCount,$inLimit,$outData,$outLimit) {
 $context = $context|0;
 $inBuffers = $inBuffers|0;
 $inBufferCount = $inBufferCount|0;
 $inLimit = $inLimit|0;
 $outData = $outData|0;
 $outLimit = $outLimit|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0;
 var $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0;
 var $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0;
 var $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0;
 var $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0;
 var $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0;
 var $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0;
 var $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0;
 var $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0;
 var $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0;
 var $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0;
 var $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0;
 var $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0;
 var $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0;
 var $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0;
 var $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0;
 var $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0;
 var $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0;
 var $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0;
 var $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0;
 var $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0;
 var $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0;
 var $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0;
 var $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0;
 var $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0;
 var $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0;
 var $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $count = 0, $encodeLow = 0, $encodeRange = 0;
 var $inData = 0, $inEnd = 0, $nextSymbol = 0, $node = 0, $node1 = 0, $or$cond = 0, $or$cond3 = 0, $order = 0, $outEnd = 0, $outStart = 0, $parent = 0, $predicted = 0, $rangeCoder = 0, $root = 0, $subcontext = 0, $symbol = 0, $total = 0, $under = 0, $value = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $predicted = sp + 94|0;
 $1 = $context;
 $2 = $inBuffers;
 $3 = $inBufferCount;
 $4 = $inLimit;
 $5 = $outData;
 $6 = $outLimit;
 $7 = $1;
 $rangeCoder = $7;
 $8 = $5;
 $outStart = $8;
 $9 = $6;
 $10 = $5;
 $11 = (($10) + ($9)|0);
 $outEnd = $11;
 $encodeLow = 0;
 $encodeRange = -1;
 HEAP16[$predicted>>1] = 0;
 $order = 0;
 $nextSymbol = 0;
 $12 = $rangeCoder;
 $13 = ($12|0)==(0|0);
 $14 = $3;
 $15 = ($14>>>0)<=(0);
 $or$cond = $13 | $15;
 $16 = $4;
 $17 = ($16>>>0)<=(0);
 $or$cond3 = $or$cond | $17;
 if ($or$cond3) {
  $0 = 0;
  $796 = $0;
  STACKTOP = sp;return ($796|0);
 }
 $18 = $2;
 $19 = HEAP32[$18>>2]|0;
 $inData = $19;
 $20 = $2;
 $21 = ((($20)) + 4|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = $inData;
 $24 = (($23) + ($22)|0);
 $inEnd = $24;
 $25 = $2;
 $26 = ((($25)) + 8|0);
 $2 = $26;
 $27 = $3;
 $28 = (($27) + -1)|0;
 $3 = $28;
 $29 = $nextSymbol;
 $30 = (($29) + 1)|0;
 $nextSymbol = $30;
 $31 = $rangeCoder;
 $32 = (($31) + ($29<<4)|0);
 $root = $32;
 $33 = $root;
 HEAP8[$33>>0] = 0;
 $34 = $root;
 $35 = ((($34)) + 1|0);
 HEAP8[$35>>0] = 0;
 $36 = $root;
 $37 = ((($36)) + 2|0);
 HEAP16[$37>>1] = 0;
 $38 = $root;
 $39 = ((($38)) + 4|0);
 HEAP16[$39>>1] = 0;
 $40 = $root;
 $41 = ((($40)) + 6|0);
 HEAP16[$41>>1] = 0;
 $42 = $root;
 $43 = ((($42)) + 8|0);
 HEAP16[$43>>1] = 0;
 $44 = $root;
 $45 = ((($44)) + 10|0);
 HEAP16[$45>>1] = 0;
 $46 = $root;
 $47 = ((($46)) + 12|0);
 HEAP16[$47>>1] = 0;
 $48 = $root;
 $49 = ((($48)) + 14|0);
 HEAP16[$49>>1] = 0;
 $50 = $root;
 $51 = ((($50)) + 10|0);
 HEAP16[$51>>1] = 1;
 $52 = $root;
 $53 = ((($52)) + 12|0);
 HEAP16[$53>>1] = 257;
 $54 = $root;
 $55 = ((($54)) + 8|0);
 HEAP16[$55>>1] = 0;
 L5: while(1) {
  $parent = $predicted;
  $56 = $inData;
  $57 = $inEnd;
  $58 = ($56>>>0)>=($57>>>0);
  if ($58) {
   $59 = $3;
   $60 = ($59>>>0)<=(0);
   if ($60) {
    label = 75;
    break;
   }
   $61 = $2;
   $62 = HEAP32[$61>>2]|0;
   $inData = $62;
   $63 = $2;
   $64 = ((($63)) + 4|0);
   $65 = HEAP32[$64>>2]|0;
   $66 = $inData;
   $67 = (($66) + ($65)|0);
   $inEnd = $67;
   $68 = $2;
   $69 = ((($68)) + 8|0);
   $2 = $69;
   $70 = $3;
   $71 = (($70) + -1)|0;
   $3 = $71;
  }
  $72 = $inData;
  $73 = ((($72)) + 1|0);
  $inData = $73;
  $74 = HEAP8[$72>>0]|0;
  $value = $74;
  $75 = HEAP16[$predicted>>1]|0;
  $76 = $75&65535;
  $77 = $rangeCoder;
  $78 = (($77) + ($76<<4)|0);
  $subcontext = $78;
  while(1) {
   $79 = $subcontext;
   $80 = $root;
   $81 = ($79|0)!=($80|0);
   $82 = $value;
   $83 = $82&255;
   if (!($81)) {
    label = 46;
    break;
   }
   $84 = 0;
   $85 = $84&65535;
   $under = $85;
   $count = 0;
   $86 = $subcontext;
   $87 = ((($86)) + 8|0);
   $88 = HEAP16[$87>>1]|0;
   $89 = ($88<<16>>16)!=(0);
   do {
    if ($89) {
     $121 = $subcontext;
     $122 = $subcontext;
     $123 = ((($122)) + 8|0);
     $124 = HEAP16[$123>>1]|0;
     $125 = $124&65535;
     $126 = (($121) + ($125<<4)|0);
     $node = $126;
     while(1) {
      $127 = $value;
      $128 = $127&255;
      $129 = $node;
      $130 = HEAP8[$129>>0]|0;
      $131 = $130&255;
      $132 = ($128|0)<($131|0);
      if ($132) {
       $133 = $node;
       $134 = ((($133)) + 2|0);
       $135 = HEAP16[$134>>1]|0;
       $136 = $135&65535;
       $137 = (($136) + 2)|0;
       $138 = $137&65535;
       HEAP16[$134>>1] = $138;
       $139 = $node;
       $140 = ((($139)) + 4|0);
       $141 = HEAP16[$140>>1]|0;
       $142 = ($141<<16>>16)!=(0);
       if (!($142)) {
        label = 15;
        break;
       }
       $143 = $node;
       $144 = ((($143)) + 4|0);
       $145 = HEAP16[$144>>1]|0;
       $146 = $145&65535;
       $147 = $node;
       $148 = (($147) + ($146<<4)|0);
       $node = $148;
       continue;
      }
      $180 = $value;
      $181 = $180&255;
      $182 = $node;
      $183 = HEAP8[$182>>0]|0;
      $184 = $183&255;
      $185 = ($181|0)>($184|0);
      $186 = $node;
      if (!($185)) {
       label = 20;
       break;
      }
      $187 = ((($186)) + 2|0);
      $188 = HEAP16[$187>>1]|0;
      $189 = $188&65535;
      $190 = $under;
      $191 = $190&65535;
      $192 = (($191) + ($189))|0;
      $193 = $192&65535;
      $under = $193;
      $194 = $node;
      $195 = ((($194)) + 6|0);
      $196 = HEAP16[$195>>1]|0;
      $197 = ($196<<16>>16)!=(0);
      if (!($197)) {
       label = 19;
       break;
      }
      $198 = $node;
      $199 = ((($198)) + 6|0);
      $200 = HEAP16[$199>>1]|0;
      $201 = $200&65535;
      $202 = $node;
      $203 = (($202) + ($201<<4)|0);
      $node = $203;
     }
     if ((label|0) == 15) {
      label = 0;
      $149 = $nextSymbol;
      $150 = (($149) + 1)|0;
      $nextSymbol = $150;
      $151 = $rangeCoder;
      $152 = (($151) + ($149<<4)|0);
      $symbol = $152;
      $153 = $value;
      $154 = $symbol;
      HEAP8[$154>>0] = $153;
      $155 = $symbol;
      $156 = ((($155)) + 1|0);
      HEAP8[$156>>0] = 2;
      $157 = $symbol;
      $158 = ((($157)) + 2|0);
      HEAP16[$158>>1] = 2;
      $159 = $symbol;
      $160 = ((($159)) + 4|0);
      HEAP16[$160>>1] = 0;
      $161 = $symbol;
      $162 = ((($161)) + 6|0);
      HEAP16[$162>>1] = 0;
      $163 = $symbol;
      $164 = ((($163)) + 8|0);
      HEAP16[$164>>1] = 0;
      $165 = $symbol;
      $166 = ((($165)) + 10|0);
      HEAP16[$166>>1] = 0;
      $167 = $symbol;
      $168 = ((($167)) + 12|0);
      HEAP16[$168>>1] = 0;
      $169 = $symbol;
      $170 = ((($169)) + 14|0);
      HEAP16[$170>>1] = 0;
      $171 = $symbol;
      $172 = $node;
      $173 = $171;
      $174 = $172;
      $175 = (($173) - ($174))|0;
      $176 = (($175|0) / 16)&-1;
      $177 = $176&65535;
      $178 = $node;
      $179 = ((($178)) + 4|0);
      HEAP16[$179>>1] = $177;
      break;
     }
     else if ((label|0) == 19) {
      label = 0;
      $204 = $nextSymbol;
      $205 = (($204) + 1)|0;
      $nextSymbol = $205;
      $206 = $rangeCoder;
      $207 = (($206) + ($204<<4)|0);
      $symbol = $207;
      $208 = $value;
      $209 = $symbol;
      HEAP8[$209>>0] = $208;
      $210 = $symbol;
      $211 = ((($210)) + 1|0);
      HEAP8[$211>>0] = 2;
      $212 = $symbol;
      $213 = ((($212)) + 2|0);
      HEAP16[$213>>1] = 2;
      $214 = $symbol;
      $215 = ((($214)) + 4|0);
      HEAP16[$215>>1] = 0;
      $216 = $symbol;
      $217 = ((($216)) + 6|0);
      HEAP16[$217>>1] = 0;
      $218 = $symbol;
      $219 = ((($218)) + 8|0);
      HEAP16[$219>>1] = 0;
      $220 = $symbol;
      $221 = ((($220)) + 10|0);
      HEAP16[$221>>1] = 0;
      $222 = $symbol;
      $223 = ((($222)) + 12|0);
      HEAP16[$223>>1] = 0;
      $224 = $symbol;
      $225 = ((($224)) + 14|0);
      HEAP16[$225>>1] = 0;
      $226 = $symbol;
      $227 = $node;
      $228 = $226;
      $229 = $227;
      $230 = (($228) - ($229))|0;
      $231 = (($230|0) / 16)&-1;
      $232 = $231&65535;
      $233 = $node;
      $234 = ((($233)) + 6|0);
      HEAP16[$234>>1] = $232;
      break;
     }
     else if ((label|0) == 20) {
      label = 0;
      $235 = ((($186)) + 1|0);
      $236 = HEAP8[$235>>0]|0;
      $237 = $236&255;
      $238 = $count;
      $239 = $238&65535;
      $240 = (($239) + ($237))|0;
      $241 = $240&65535;
      $count = $241;
      $242 = $node;
      $243 = ((($242)) + 2|0);
      $244 = HEAP16[$243>>1]|0;
      $245 = $244&65535;
      $246 = $node;
      $247 = ((($246)) + 1|0);
      $248 = HEAP8[$247>>0]|0;
      $249 = $248&255;
      $250 = (($245) - ($249))|0;
      $251 = $under;
      $252 = $251&65535;
      $253 = (($252) + ($250))|0;
      $254 = $253&65535;
      $under = $254;
      $255 = $node;
      $256 = ((($255)) + 2|0);
      $257 = HEAP16[$256>>1]|0;
      $258 = $257&65535;
      $259 = (($258) + 2)|0;
      $260 = $259&65535;
      HEAP16[$256>>1] = $260;
      $261 = $node;
      $262 = ((($261)) + 1|0);
      $263 = HEAP8[$262>>0]|0;
      $264 = $263&255;
      $265 = (($264) + 2)|0;
      $266 = $265&255;
      HEAP8[$262>>0] = $266;
      $267 = $node;
      $symbol = $267;
      break;
     }
    } else {
     $90 = $nextSymbol;
     $91 = (($90) + 1)|0;
     $nextSymbol = $91;
     $92 = $rangeCoder;
     $93 = (($92) + ($90<<4)|0);
     $symbol = $93;
     $94 = $value;
     $95 = $symbol;
     HEAP8[$95>>0] = $94;
     $96 = $symbol;
     $97 = ((($96)) + 1|0);
     HEAP8[$97>>0] = 2;
     $98 = $symbol;
     $99 = ((($98)) + 2|0);
     HEAP16[$99>>1] = 2;
     $100 = $symbol;
     $101 = ((($100)) + 4|0);
     HEAP16[$101>>1] = 0;
     $102 = $symbol;
     $103 = ((($102)) + 6|0);
     HEAP16[$103>>1] = 0;
     $104 = $symbol;
     $105 = ((($104)) + 8|0);
     HEAP16[$105>>1] = 0;
     $106 = $symbol;
     $107 = ((($106)) + 10|0);
     HEAP16[$107>>1] = 0;
     $108 = $symbol;
     $109 = ((($108)) + 12|0);
     HEAP16[$109>>1] = 0;
     $110 = $symbol;
     $111 = ((($110)) + 14|0);
     HEAP16[$111>>1] = 0;
     $112 = $symbol;
     $113 = $subcontext;
     $114 = $112;
     $115 = $113;
     $116 = (($114) - ($115))|0;
     $117 = (($116|0) / 16)&-1;
     $118 = $117&65535;
     $119 = $subcontext;
     $120 = ((($119)) + 8|0);
     HEAP16[$120>>1] = $118;
    }
   } while(0);
   $268 = $symbol;
   $269 = $rangeCoder;
   $270 = $268;
   $271 = $269;
   $272 = (($270) - ($271))|0;
   $273 = (($272|0) / 16)&-1;
   $274 = $273&65535;
   $275 = $parent;
   HEAP16[$275>>1] = $274;
   $276 = $symbol;
   $277 = ((($276)) + 14|0);
   $parent = $277;
   $278 = $subcontext;
   $279 = ((($278)) + 12|0);
   $280 = HEAP16[$279>>1]|0;
   $total = $280;
   $281 = $count;
   $282 = $281&65535;
   $283 = ($282|0)>(0);
   L30: do {
    if ($283) {
     $284 = $total;
     $285 = $284&65535;
     $286 = $encodeRange;
     $287 = (($286>>>0) / ($285>>>0))&-1;
     $encodeRange = $287;
     $288 = $subcontext;
     $289 = ((($288)) + 10|0);
     $290 = HEAP16[$289>>1]|0;
     $291 = $290&65535;
     $292 = $under;
     $293 = $292&65535;
     $294 = (($291) + ($293))|0;
     $295 = $encodeRange;
     $296 = Math_imul($294, $295)|0;
     $297 = $encodeLow;
     $298 = (($297) + ($296))|0;
     $encodeLow = $298;
     $299 = $count;
     $300 = $299&65535;
     $301 = $encodeRange;
     $302 = Math_imul($301, $300)|0;
     $encodeRange = $302;
     while(1) {
      $303 = $encodeLow;
      $304 = $encodeLow;
      $305 = $encodeRange;
      $306 = (($304) + ($305))|0;
      $307 = $303 ^ $306;
      $308 = ($307>>>0)>=(16777216);
      if ($308) {
       $309 = $encodeRange;
       $310 = ($309>>>0)>=(65536);
       if ($310) {
        break L30;
       }
       $311 = $encodeLow;
       $312 = (0 - ($311))|0;
       $313 = $312 & 65535;
       $encodeRange = $313;
      }
      $314 = $5;
      $315 = $outEnd;
      $316 = ($314>>>0)>=($315>>>0);
      if ($316) {
       label = 27;
       break L5;
      }
      $317 = $encodeLow;
      $318 = $317 >>> 24;
      $319 = $318&255;
      $320 = $5;
      $321 = ((($320)) + 1|0);
      $5 = $321;
      HEAP8[$320>>0] = $319;
      $322 = $encodeRange;
      $323 = $322 << 8;
      $encodeRange = $323;
      $324 = $encodeLow;
      $325 = $324 << 8;
      $encodeLow = $325;
     }
    } else {
     $326 = $subcontext;
     $327 = ((($326)) + 10|0);
     $328 = HEAP16[$327>>1]|0;
     $329 = $328&65535;
     $330 = ($329|0)>(0);
     L40: do {
      if ($330) {
       $331 = $subcontext;
       $332 = ((($331)) + 10|0);
       $333 = HEAP16[$332>>1]|0;
       $334 = $333&65535;
       $335 = $total;
       $336 = $335&65535;
       $337 = ($334|0)<($336|0);
       if ($337) {
        $338 = $total;
        $339 = $338&65535;
        $340 = $encodeRange;
        $341 = (($340>>>0) / ($339>>>0))&-1;
        $encodeRange = $341;
        $342 = $encodeRange;
        $343 = 0;
        $344 = $encodeLow;
        $345 = (($344) + ($343))|0;
        $encodeLow = $345;
        $346 = $subcontext;
        $347 = ((($346)) + 10|0);
        $348 = HEAP16[$347>>1]|0;
        $349 = $348&65535;
        $350 = $encodeRange;
        $351 = Math_imul($350, $349)|0;
        $encodeRange = $351;
        while(1) {
         $352 = $encodeLow;
         $353 = $encodeLow;
         $354 = $encodeRange;
         $355 = (($353) + ($354))|0;
         $356 = $352 ^ $355;
         $357 = ($356>>>0)>=(16777216);
         if ($357) {
          $358 = $encodeRange;
          $359 = ($358>>>0)>=(65536);
          if ($359) {
           break L40;
          }
          $360 = $encodeLow;
          $361 = (0 - ($360))|0;
          $362 = $361 & 65535;
          $encodeRange = $362;
         }
         $363 = $5;
         $364 = $outEnd;
         $365 = ($363>>>0)>=($364>>>0);
         if ($365) {
          label = 36;
          break L5;
         }
         $366 = $encodeLow;
         $367 = $366 >>> 24;
         $368 = $367&255;
         $369 = $5;
         $370 = ((($369)) + 1|0);
         $5 = $370;
         HEAP8[$369>>0] = $368;
         $371 = $encodeRange;
         $372 = $371 << 8;
         $encodeRange = $372;
         $373 = $encodeLow;
         $374 = $373 << 8;
         $encodeLow = $374;
        }
       }
      }
     } while(0);
     $375 = $subcontext;
     $376 = ((($375)) + 10|0);
     $377 = HEAP16[$376>>1]|0;
     $378 = $377&65535;
     $379 = (($378) + 5)|0;
     $380 = $379&65535;
     HEAP16[$376>>1] = $380;
     $381 = $subcontext;
     $382 = ((($381)) + 12|0);
     $383 = HEAP16[$382>>1]|0;
     $384 = $383&65535;
     $385 = (($384) + 5)|0;
     $386 = $385&65535;
     HEAP16[$382>>1] = $386;
    }
   } while(0);
   $387 = $subcontext;
   $388 = ((($387)) + 12|0);
   $389 = HEAP16[$388>>1]|0;
   $390 = $389&65535;
   $391 = (($390) + 2)|0;
   $392 = $391&65535;
   HEAP16[$388>>1] = $392;
   $393 = $count;
   $394 = $393&65535;
   $395 = ($394|0)>(251);
   if ($395) {
    label = 41;
   } else {
    $396 = $subcontext;
    $397 = ((($396)) + 12|0);
    $398 = HEAP16[$397>>1]|0;
    $399 = $398&65535;
    $400 = ($399|0)>(65280);
    if ($400) {
     label = 41;
    }
   }
   if ((label|0) == 41) {
    label = 0;
    $401 = $subcontext;
    $402 = ((($401)) + 8|0);
    $403 = HEAP16[$402>>1]|0;
    $404 = $403&65535;
    $405 = ($404|0)!=(0);
    if ($405) {
     $406 = $subcontext;
     $407 = $subcontext;
     $408 = ((($407)) + 8|0);
     $409 = HEAP16[$408>>1]|0;
     $410 = $409&65535;
     $411 = (($406) + ($410<<4)|0);
     $412 = (_enet_symbol_rescale($411)|0);
     $413 = $412&65535;
     $415 = $413;
    } else {
     $415 = 0;
    }
    $414 = $415&65535;
    $416 = $subcontext;
    $417 = ((($416)) + 12|0);
    HEAP16[$417>>1] = $414;
    $418 = $subcontext;
    $419 = ((($418)) + 10|0);
    $420 = HEAP16[$419>>1]|0;
    $421 = $420&65535;
    $422 = $421 >> 1;
    $423 = $subcontext;
    $424 = ((($423)) + 10|0);
    $425 = HEAP16[$424>>1]|0;
    $426 = $425&65535;
    $427 = (($426) - ($422))|0;
    $428 = $427&65535;
    HEAP16[$424>>1] = $428;
    $429 = $subcontext;
    $430 = ((($429)) + 10|0);
    $431 = HEAP16[$430>>1]|0;
    $432 = $431&65535;
    $433 = (($432) + 0)|0;
    $434 = $subcontext;
    $435 = ((($434)) + 12|0);
    $436 = HEAP16[$435>>1]|0;
    $437 = $436&65535;
    $438 = (($437) + ($433))|0;
    $439 = $438&65535;
    HEAP16[$435>>1] = $439;
   }
   $440 = $count;
   $441 = $440&65535;
   $442 = ($441|0)>(0);
   if ($442) {
    break;
   }
   $443 = $subcontext;
   $444 = ((($443)) + 14|0);
   $445 = HEAP16[$444>>1]|0;
   $446 = $445&65535;
   $447 = $rangeCoder;
   $448 = (($447) + ($446<<4)|0);
   $subcontext = $448;
  }
  do {
   if ((label|0) == 46) {
    label = 0;
    $449 = $83;
    $450 = $449&65535;
    $under = $450;
    $count = 1;
    $451 = $root;
    $452 = ((($451)) + 8|0);
    $453 = HEAP16[$452>>1]|0;
    $454 = ($453<<16>>16)!=(0);
    do {
     if ($454) {
      $486 = $root;
      $487 = $root;
      $488 = ((($487)) + 8|0);
      $489 = HEAP16[$488>>1]|0;
      $490 = $489&65535;
      $491 = (($486) + ($490<<4)|0);
      $node1 = $491;
      while(1) {
       $492 = $value;
       $493 = $492&255;
       $494 = $node1;
       $495 = HEAP8[$494>>0]|0;
       $496 = $495&255;
       $497 = ($493|0)<($496|0);
       if ($497) {
        $498 = $node1;
        $499 = ((($498)) + 2|0);
        $500 = HEAP16[$499>>1]|0;
        $501 = $500&65535;
        $502 = (($501) + 3)|0;
        $503 = $502&65535;
        HEAP16[$499>>1] = $503;
        $504 = $node1;
        $505 = ((($504)) + 4|0);
        $506 = HEAP16[$505>>1]|0;
        $507 = ($506<<16>>16)!=(0);
        if (!($507)) {
         label = 52;
         break;
        }
        $508 = $node1;
        $509 = ((($508)) + 4|0);
        $510 = HEAP16[$509>>1]|0;
        $511 = $510&65535;
        $512 = $node1;
        $513 = (($512) + ($511<<4)|0);
        $node1 = $513;
        continue;
       }
       $545 = $value;
       $546 = $545&255;
       $547 = $node1;
       $548 = HEAP8[$547>>0]|0;
       $549 = $548&255;
       $550 = ($546|0)>($549|0);
       $551 = $node1;
       if (!($550)) {
        label = 57;
        break;
       }
       $552 = ((($551)) + 2|0);
       $553 = HEAP16[$552>>1]|0;
       $554 = $553&65535;
       $555 = $under;
       $556 = $555&65535;
       $557 = (($556) + ($554))|0;
       $558 = $557&65535;
       $under = $558;
       $559 = $node1;
       $560 = ((($559)) + 6|0);
       $561 = HEAP16[$560>>1]|0;
       $562 = ($561<<16>>16)!=(0);
       if (!($562)) {
        label = 56;
        break;
       }
       $563 = $node1;
       $564 = ((($563)) + 6|0);
       $565 = HEAP16[$564>>1]|0;
       $566 = $565&65535;
       $567 = $node1;
       $568 = (($567) + ($566<<4)|0);
       $node1 = $568;
      }
      if ((label|0) == 52) {
       label = 0;
       $514 = $nextSymbol;
       $515 = (($514) + 1)|0;
       $nextSymbol = $515;
       $516 = $rangeCoder;
       $517 = (($516) + ($514<<4)|0);
       $symbol = $517;
       $518 = $value;
       $519 = $symbol;
       HEAP8[$519>>0] = $518;
       $520 = $symbol;
       $521 = ((($520)) + 1|0);
       HEAP8[$521>>0] = 3;
       $522 = $symbol;
       $523 = ((($522)) + 2|0);
       HEAP16[$523>>1] = 3;
       $524 = $symbol;
       $525 = ((($524)) + 4|0);
       HEAP16[$525>>1] = 0;
       $526 = $symbol;
       $527 = ((($526)) + 6|0);
       HEAP16[$527>>1] = 0;
       $528 = $symbol;
       $529 = ((($528)) + 8|0);
       HEAP16[$529>>1] = 0;
       $530 = $symbol;
       $531 = ((($530)) + 10|0);
       HEAP16[$531>>1] = 0;
       $532 = $symbol;
       $533 = ((($532)) + 12|0);
       HEAP16[$533>>1] = 0;
       $534 = $symbol;
       $535 = ((($534)) + 14|0);
       HEAP16[$535>>1] = 0;
       $536 = $symbol;
       $537 = $node1;
       $538 = $536;
       $539 = $537;
       $540 = (($538) - ($539))|0;
       $541 = (($540|0) / 16)&-1;
       $542 = $541&65535;
       $543 = $node1;
       $544 = ((($543)) + 4|0);
       HEAP16[$544>>1] = $542;
       break;
      }
      else if ((label|0) == 56) {
       label = 0;
       $569 = $nextSymbol;
       $570 = (($569) + 1)|0;
       $nextSymbol = $570;
       $571 = $rangeCoder;
       $572 = (($571) + ($569<<4)|0);
       $symbol = $572;
       $573 = $value;
       $574 = $symbol;
       HEAP8[$574>>0] = $573;
       $575 = $symbol;
       $576 = ((($575)) + 1|0);
       HEAP8[$576>>0] = 3;
       $577 = $symbol;
       $578 = ((($577)) + 2|0);
       HEAP16[$578>>1] = 3;
       $579 = $symbol;
       $580 = ((($579)) + 4|0);
       HEAP16[$580>>1] = 0;
       $581 = $symbol;
       $582 = ((($581)) + 6|0);
       HEAP16[$582>>1] = 0;
       $583 = $symbol;
       $584 = ((($583)) + 8|0);
       HEAP16[$584>>1] = 0;
       $585 = $symbol;
       $586 = ((($585)) + 10|0);
       HEAP16[$586>>1] = 0;
       $587 = $symbol;
       $588 = ((($587)) + 12|0);
       HEAP16[$588>>1] = 0;
       $589 = $symbol;
       $590 = ((($589)) + 14|0);
       HEAP16[$590>>1] = 0;
       $591 = $symbol;
       $592 = $node1;
       $593 = $591;
       $594 = $592;
       $595 = (($593) - ($594))|0;
       $596 = (($595|0) / 16)&-1;
       $597 = $596&65535;
       $598 = $node1;
       $599 = ((($598)) + 6|0);
       HEAP16[$599>>1] = $597;
       break;
      }
      else if ((label|0) == 57) {
       label = 0;
       $600 = ((($551)) + 1|0);
       $601 = HEAP8[$600>>0]|0;
       $602 = $601&255;
       $603 = $count;
       $604 = $603&65535;
       $605 = (($604) + ($602))|0;
       $606 = $605&65535;
       $count = $606;
       $607 = $node1;
       $608 = ((($607)) + 2|0);
       $609 = HEAP16[$608>>1]|0;
       $610 = $609&65535;
       $611 = $node1;
       $612 = ((($611)) + 1|0);
       $613 = HEAP8[$612>>0]|0;
       $614 = $613&255;
       $615 = (($610) - ($614))|0;
       $616 = $under;
       $617 = $616&65535;
       $618 = (($617) + ($615))|0;
       $619 = $618&65535;
       $under = $619;
       $620 = $node1;
       $621 = ((($620)) + 2|0);
       $622 = HEAP16[$621>>1]|0;
       $623 = $622&65535;
       $624 = (($623) + 3)|0;
       $625 = $624&65535;
       HEAP16[$621>>1] = $625;
       $626 = $node1;
       $627 = ((($626)) + 1|0);
       $628 = HEAP8[$627>>0]|0;
       $629 = $628&255;
       $630 = (($629) + 3)|0;
       $631 = $630&255;
       HEAP8[$627>>0] = $631;
       $632 = $node1;
       $symbol = $632;
       break;
      }
     } else {
      $455 = $nextSymbol;
      $456 = (($455) + 1)|0;
      $nextSymbol = $456;
      $457 = $rangeCoder;
      $458 = (($457) + ($455<<4)|0);
      $symbol = $458;
      $459 = $value;
      $460 = $symbol;
      HEAP8[$460>>0] = $459;
      $461 = $symbol;
      $462 = ((($461)) + 1|0);
      HEAP8[$462>>0] = 3;
      $463 = $symbol;
      $464 = ((($463)) + 2|0);
      HEAP16[$464>>1] = 3;
      $465 = $symbol;
      $466 = ((($465)) + 4|0);
      HEAP16[$466>>1] = 0;
      $467 = $symbol;
      $468 = ((($467)) + 6|0);
      HEAP16[$468>>1] = 0;
      $469 = $symbol;
      $470 = ((($469)) + 8|0);
      HEAP16[$470>>1] = 0;
      $471 = $symbol;
      $472 = ((($471)) + 10|0);
      HEAP16[$472>>1] = 0;
      $473 = $symbol;
      $474 = ((($473)) + 12|0);
      HEAP16[$474>>1] = 0;
      $475 = $symbol;
      $476 = ((($475)) + 14|0);
      HEAP16[$476>>1] = 0;
      $477 = $symbol;
      $478 = $root;
      $479 = $477;
      $480 = $478;
      $481 = (($479) - ($480))|0;
      $482 = (($481|0) / 16)&-1;
      $483 = $482&65535;
      $484 = $root;
      $485 = ((($484)) + 8|0);
      HEAP16[$485>>1] = $483;
     }
    } while(0);
    $633 = $symbol;
    $634 = $rangeCoder;
    $635 = $633;
    $636 = $634;
    $637 = (($635) - ($636))|0;
    $638 = (($637|0) / 16)&-1;
    $639 = $638&65535;
    $640 = $parent;
    HEAP16[$640>>1] = $639;
    $641 = $symbol;
    $642 = ((($641)) + 14|0);
    $parent = $642;
    $643 = $root;
    $644 = ((($643)) + 12|0);
    $645 = HEAP16[$644>>1]|0;
    $total = $645;
    $646 = $total;
    $647 = $646&65535;
    $648 = $encodeRange;
    $649 = (($648>>>0) / ($647>>>0))&-1;
    $encodeRange = $649;
    $650 = $root;
    $651 = ((($650)) + 10|0);
    $652 = HEAP16[$651>>1]|0;
    $653 = $652&65535;
    $654 = $under;
    $655 = $654&65535;
    $656 = (($653) + ($655))|0;
    $657 = $encodeRange;
    $658 = Math_imul($656, $657)|0;
    $659 = $encodeLow;
    $660 = (($659) + ($658))|0;
    $encodeLow = $660;
    $661 = $count;
    $662 = $661&65535;
    $663 = $encodeRange;
    $664 = Math_imul($663, $662)|0;
    $encodeRange = $664;
    while(1) {
     $665 = $encodeLow;
     $666 = $encodeLow;
     $667 = $encodeRange;
     $668 = (($666) + ($667))|0;
     $669 = $665 ^ $668;
     $670 = ($669>>>0)>=(16777216);
     if ($670) {
      $671 = $encodeRange;
      $672 = ($671>>>0)>=(65536);
      if ($672) {
       break;
      }
      $673 = $encodeLow;
      $674 = (0 - ($673))|0;
      $675 = $674 & 65535;
      $encodeRange = $675;
     }
     $676 = $5;
     $677 = $outEnd;
     $678 = ($676>>>0)>=($677>>>0);
     if ($678) {
      label = 63;
      break L5;
     }
     $679 = $encodeLow;
     $680 = $679 >>> 24;
     $681 = $680&255;
     $682 = $5;
     $683 = ((($682)) + 1|0);
     $5 = $683;
     HEAP8[$682>>0] = $681;
     $684 = $encodeRange;
     $685 = $684 << 8;
     $encodeRange = $685;
     $686 = $encodeLow;
     $687 = $686 << 8;
     $encodeLow = $687;
    }
    $688 = $root;
    $689 = ((($688)) + 12|0);
    $690 = HEAP16[$689>>1]|0;
    $691 = $690&65535;
    $692 = (($691) + 3)|0;
    $693 = $692&65535;
    HEAP16[$689>>1] = $693;
    $694 = $count;
    $695 = $694&65535;
    $696 = ($695|0)>(250);
    if (!($696)) {
     $697 = $root;
     $698 = ((($697)) + 12|0);
     $699 = HEAP16[$698>>1]|0;
     $700 = $699&65535;
     $701 = ($700|0)>(65280);
     if (!($701)) {
      break;
     }
    }
    $702 = $root;
    $703 = ((($702)) + 8|0);
    $704 = HEAP16[$703>>1]|0;
    $705 = $704&65535;
    $706 = ($705|0)!=(0);
    if ($706) {
     $707 = $root;
     $708 = $root;
     $709 = ((($708)) + 8|0);
     $710 = HEAP16[$709>>1]|0;
     $711 = $710&65535;
     $712 = (($707) + ($711<<4)|0);
     $713 = (_enet_symbol_rescale($712)|0);
     $714 = $713&65535;
     $716 = $714;
    } else {
     $716 = 0;
    }
    $715 = $716&65535;
    $717 = $root;
    $718 = ((($717)) + 12|0);
    HEAP16[$718>>1] = $715;
    $719 = $root;
    $720 = ((($719)) + 10|0);
    $721 = HEAP16[$720>>1]|0;
    $722 = $721&65535;
    $723 = $722 >> 1;
    $724 = $root;
    $725 = ((($724)) + 10|0);
    $726 = HEAP16[$725>>1]|0;
    $727 = $726&65535;
    $728 = (($727) - ($723))|0;
    $729 = $728&65535;
    HEAP16[$725>>1] = $729;
    $730 = $root;
    $731 = ((($730)) + 10|0);
    $732 = HEAP16[$731>>1]|0;
    $733 = $732&65535;
    $734 = (($733) + 256)|0;
    $735 = $root;
    $736 = ((($735)) + 12|0);
    $737 = HEAP16[$736>>1]|0;
    $738 = $737&65535;
    $739 = (($738) + ($734))|0;
    $740 = $739&65535;
    HEAP16[$736>>1] = $740;
   }
  } while(0);
  $741 = $order;
  $742 = ($741>>>0)>=(2);
  if ($742) {
   $743 = HEAP16[$predicted>>1]|0;
   $744 = $743&65535;
   $745 = $rangeCoder;
   $746 = (($745) + ($744<<4)|0);
   $747 = ((($746)) + 14|0);
   $748 = HEAP16[$747>>1]|0;
   HEAP16[$predicted>>1] = $748;
  } else {
   $749 = $order;
   $750 = (($749) + 1)|0;
   $order = $750;
  }
  $751 = $nextSymbol;
  $752 = ($751>>>0)>=(4094);
  if (!($752)) {
   continue;
  }
  $nextSymbol = 0;
  $753 = $nextSymbol;
  $754 = (($753) + 1)|0;
  $nextSymbol = $754;
  $755 = $rangeCoder;
  $756 = (($755) + ($753<<4)|0);
  $root = $756;
  $757 = $root;
  HEAP8[$757>>0] = 0;
  $758 = $root;
  $759 = ((($758)) + 1|0);
  HEAP8[$759>>0] = 0;
  $760 = $root;
  $761 = ((($760)) + 2|0);
  HEAP16[$761>>1] = 0;
  $762 = $root;
  $763 = ((($762)) + 4|0);
  HEAP16[$763>>1] = 0;
  $764 = $root;
  $765 = ((($764)) + 6|0);
  HEAP16[$765>>1] = 0;
  $766 = $root;
  $767 = ((($766)) + 8|0);
  HEAP16[$767>>1] = 0;
  $768 = $root;
  $769 = ((($768)) + 10|0);
  HEAP16[$769>>1] = 0;
  $770 = $root;
  $771 = ((($770)) + 12|0);
  HEAP16[$771>>1] = 0;
  $772 = $root;
  $773 = ((($772)) + 14|0);
  HEAP16[$773>>1] = 0;
  $774 = $root;
  $775 = ((($774)) + 10|0);
  HEAP16[$775>>1] = 1;
  $776 = $root;
  $777 = ((($776)) + 12|0);
  HEAP16[$777>>1] = 257;
  $778 = $root;
  $779 = ((($778)) + 8|0);
  HEAP16[$779>>1] = 0;
  HEAP16[$predicted>>1] = 0;
  $order = 0;
 }
 if ((label|0) == 27) {
  $0 = 0;
  $796 = $0;
  STACKTOP = sp;return ($796|0);
 }
 else if ((label|0) == 36) {
  $0 = 0;
  $796 = $0;
  STACKTOP = sp;return ($796|0);
 }
 else if ((label|0) == 63) {
  $0 = 0;
  $796 = $0;
  STACKTOP = sp;return ($796|0);
 }
 else if ((label|0) == 75) {
  while(1) {
   label = 0;
   $780 = $encodeLow;
   $781 = ($780|0)!=(0);
   $782 = $5;
   if (!($781)) {
    label = 79;
    break;
   }
   $783 = $outEnd;
   $784 = ($782>>>0)>=($783>>>0);
   if ($784) {
    label = 77;
    break;
   }
   $785 = $encodeLow;
   $786 = $785 >>> 24;
   $787 = $786&255;
   $788 = $5;
   $789 = ((($788)) + 1|0);
   $5 = $789;
   HEAP8[$788>>0] = $787;
   $790 = $encodeLow;
   $791 = $790 << 8;
   $encodeLow = $791;
   label = 75;
  }
  if ((label|0) == 77) {
   $0 = 0;
   $796 = $0;
   STACKTOP = sp;return ($796|0);
  }
  else if ((label|0) == 79) {
   $792 = $outStart;
   $793 = $782;
   $794 = $792;
   $795 = (($793) - ($794))|0;
   $0 = $795;
   $796 = $0;
   STACKTOP = sp;return ($796|0);
  }
 }
 return (0)|0;
}
function _enet_symbol_rescale($symbol) {
 $symbol = $symbol|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $total = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $symbol;
 $total = 0;
 while(1) {
  $1 = $0;
  $2 = ((($1)) + 1|0);
  $3 = HEAP8[$2>>0]|0;
  $4 = $3&255;
  $5 = $4 >> 1;
  $6 = $0;
  $7 = ((($6)) + 1|0);
  $8 = HEAP8[$7>>0]|0;
  $9 = $8&255;
  $10 = (($9) - ($5))|0;
  $11 = $10&255;
  HEAP8[$7>>0] = $11;
  $12 = $0;
  $13 = ((($12)) + 1|0);
  $14 = HEAP8[$13>>0]|0;
  $15 = $14&255;
  $16 = $0;
  $17 = ((($16)) + 2|0);
  HEAP16[$17>>1] = $15;
  $18 = $0;
  $19 = ((($18)) + 4|0);
  $20 = HEAP16[$19>>1]|0;
  $21 = ($20<<16>>16)!=(0);
  if ($21) {
   $22 = $0;
   $23 = $0;
   $24 = ((($23)) + 4|0);
   $25 = HEAP16[$24>>1]|0;
   $26 = $25&65535;
   $27 = (($22) + ($26<<4)|0);
   $28 = (_enet_symbol_rescale($27)|0);
   $29 = $28&65535;
   $30 = $0;
   $31 = ((($30)) + 2|0);
   $32 = HEAP16[$31>>1]|0;
   $33 = $32&65535;
   $34 = (($33) + ($29))|0;
   $35 = $34&65535;
   HEAP16[$31>>1] = $35;
  }
  $36 = $0;
  $37 = ((($36)) + 2|0);
  $38 = HEAP16[$37>>1]|0;
  $39 = $38&65535;
  $40 = $total;
  $41 = $40&65535;
  $42 = (($41) + ($39))|0;
  $43 = $42&65535;
  $total = $43;
  $44 = $0;
  $45 = ((($44)) + 6|0);
  $46 = HEAP16[$45>>1]|0;
  $47 = ($46<<16>>16)!=(0);
  if (!($47)) {
   break;
  }
  $48 = $0;
  $49 = ((($48)) + 6|0);
  $50 = HEAP16[$49>>1]|0;
  $51 = $50&65535;
  $52 = $0;
  $53 = (($52) + ($51<<4)|0);
  $0 = $53;
 }
 $54 = $total;
 STACKTOP = sp;return ($54|0);
}
function _enet_range_coder_decompress($context,$inData,$inLimit,$outData,$outLimit) {
 $context = $context|0;
 $inData = $inData|0;
 $inLimit = $inLimit|0;
 $outData = $outData|0;
 $outLimit = $outLimit|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0;
 var $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0;
 var $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0;
 var $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0;
 var $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0;
 var $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0;
 var $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0;
 var $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0;
 var $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0;
 var $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0;
 var $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0;
 var $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0;
 var $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0;
 var $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0;
 var $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0;
 var $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0;
 var $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0;
 var $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0;
 var $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0;
 var $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0;
 var $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0;
 var $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0;
 var $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0;
 var $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0;
 var $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0;
 var $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0;
 var $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0;
 var $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0;
 var $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0;
 var $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0;
 var $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0;
 var $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0;
 var $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0;
 var $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0;
 var $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0;
 var $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0;
 var $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0;
 var $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0;
 var $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0;
 var $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0;
 var $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0;
 var $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0;
 var $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0;
 var $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0;
 var $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0;
 var $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0;
 var $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0;
 var $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0;
 var $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0;
 var $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0;
 var $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0;
 var $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $after = 0, $after2 = 0, $before = 0, $before3 = 0, $bottom = 0, $code = 0, $count = 0, $decodeCode = 0, $decodeLow = 0;
 var $decodeRange = 0, $inEnd = 0, $nextSymbol = 0, $node = 0, $node1 = 0, $node4 = 0, $or$cond = 0, $order = 0, $outEnd = 0, $outStart = 0, $parent = 0, $patch = 0, $predicted = 0, $rangeCoder = 0, $root = 0, $subcontext = 0, $symbol = 0, $total = 0, $under = 0, $value = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $predicted = sp + 110|0;
 $1 = $context;
 $2 = $inData;
 $3 = $inLimit;
 $4 = $outData;
 $5 = $outLimit;
 $6 = $1;
 $rangeCoder = $6;
 $7 = $4;
 $outStart = $7;
 $8 = $5;
 $9 = $4;
 $10 = (($9) + ($8)|0);
 $outEnd = $10;
 $11 = $3;
 $12 = $2;
 $13 = (($12) + ($11)|0);
 $inEnd = $13;
 $decodeLow = 0;
 $decodeCode = 0;
 $decodeRange = -1;
 HEAP16[$predicted>>1] = 0;
 $order = 0;
 $nextSymbol = 0;
 $14 = $rangeCoder;
 $15 = ($14|0)==(0|0);
 $16 = $3;
 $17 = ($16>>>0)<=(0);
 $or$cond = $15 | $17;
 if ($or$cond) {
  $0 = 0;
  $1130 = $0;
  STACKTOP = sp;return ($1130|0);
 }
 $18 = $nextSymbol;
 $19 = (($18) + 1)|0;
 $nextSymbol = $19;
 $20 = $rangeCoder;
 $21 = (($20) + ($18<<4)|0);
 $root = $21;
 $22 = $root;
 HEAP8[$22>>0] = 0;
 $23 = $root;
 $24 = ((($23)) + 1|0);
 HEAP8[$24>>0] = 0;
 $25 = $root;
 $26 = ((($25)) + 2|0);
 HEAP16[$26>>1] = 0;
 $27 = $root;
 $28 = ((($27)) + 4|0);
 HEAP16[$28>>1] = 0;
 $29 = $root;
 $30 = ((($29)) + 6|0);
 HEAP16[$30>>1] = 0;
 $31 = $root;
 $32 = ((($31)) + 8|0);
 HEAP16[$32>>1] = 0;
 $33 = $root;
 $34 = ((($33)) + 10|0);
 HEAP16[$34>>1] = 0;
 $35 = $root;
 $36 = ((($35)) + 12|0);
 HEAP16[$36>>1] = 0;
 $37 = $root;
 $38 = ((($37)) + 14|0);
 HEAP16[$38>>1] = 0;
 $39 = $root;
 $40 = ((($39)) + 10|0);
 HEAP16[$40>>1] = 1;
 $41 = $root;
 $42 = ((($41)) + 12|0);
 HEAP16[$42>>1] = 257;
 $43 = $root;
 $44 = ((($43)) + 8|0);
 HEAP16[$44>>1] = 0;
 $45 = $2;
 $46 = $inEnd;
 $47 = ($45>>>0)<($46>>>0);
 if ($47) {
  $48 = $2;
  $49 = ((($48)) + 1|0);
  $2 = $49;
  $50 = HEAP8[$48>>0]|0;
  $51 = $50&255;
  $52 = $51 << 24;
  $53 = $decodeCode;
  $54 = $53 | $52;
  $decodeCode = $54;
 }
 $55 = $2;
 $56 = $inEnd;
 $57 = ($55>>>0)<($56>>>0);
 if ($57) {
  $58 = $2;
  $59 = ((($58)) + 1|0);
  $2 = $59;
  $60 = HEAP8[$58>>0]|0;
  $61 = $60&255;
  $62 = $61 << 16;
  $63 = $decodeCode;
  $64 = $63 | $62;
  $decodeCode = $64;
 }
 $65 = $2;
 $66 = $inEnd;
 $67 = ($65>>>0)<($66>>>0);
 if ($67) {
  $68 = $2;
  $69 = ((($68)) + 1|0);
  $2 = $69;
  $70 = HEAP8[$68>>0]|0;
  $71 = $70&255;
  $72 = $71 << 8;
  $73 = $decodeCode;
  $74 = $73 | $72;
  $decodeCode = $74;
 }
 $75 = $2;
 $76 = $inEnd;
 $77 = ($75>>>0)<($76>>>0);
 if ($77) {
  $78 = $2;
  $79 = ((($78)) + 1|0);
  $2 = $79;
  $80 = HEAP8[$78>>0]|0;
  $81 = $80&255;
  $82 = $decodeCode;
  $83 = $82 | $81;
  $decodeCode = $83;
 }
 L16: while(1) {
  $value = 0;
  $parent = $predicted;
  $84 = HEAP16[$predicted>>1]|0;
  $85 = $84&65535;
  $86 = $rangeCoder;
  $87 = (($86) + ($85<<4)|0);
  $subcontext = $87;
  L18: while(1) {
   $88 = $subcontext;
   $89 = $root;
   $90 = ($88|0)!=($89|0);
   if (!($90)) {
    label = 47;
    break;
   }
   $91 = $subcontext;
   $92 = ((($91)) + 10|0);
   $93 = HEAP16[$92>>1]|0;
   $94 = $93&65535;
   $95 = ($94|0)<=(0);
   L21: do {
    if (!($95)) {
     $96 = $subcontext;
     $97 = ((($96)) + 12|0);
     $98 = HEAP16[$97>>1]|0;
     $total = $98;
     $99 = $subcontext;
     $100 = ((($99)) + 10|0);
     $101 = HEAP16[$100>>1]|0;
     $102 = $101&65535;
     $103 = $total;
     $104 = $103&65535;
     $105 = ($102|0)>=($104|0);
     if (!($105)) {
      $106 = $decodeCode;
      $107 = $decodeLow;
      $108 = (($106) - ($107))|0;
      $109 = $total;
      $110 = $109&65535;
      $111 = $decodeRange;
      $112 = (($111>>>0) / ($110>>>0))&-1;
      $decodeRange = $112;
      $113 = (($108>>>0) / ($112>>>0))&-1;
      $114 = $113&65535;
      $code = $114;
      $115 = $code;
      $116 = $115&65535;
      $117 = $subcontext;
      $118 = ((($117)) + 10|0);
      $119 = HEAP16[$118>>1]|0;
      $120 = $119&65535;
      $121 = ($116|0)<($120|0);
      if (!($121)) {
       label = 23;
       break L18;
      }
      $122 = $decodeRange;
      $123 = 0;
      $124 = $decodeLow;
      $125 = (($124) + ($123))|0;
      $decodeLow = $125;
      $126 = $subcontext;
      $127 = ((($126)) + 10|0);
      $128 = HEAP16[$127>>1]|0;
      $129 = $128&65535;
      $130 = $decodeRange;
      $131 = Math_imul($130, $129)|0;
      $decodeRange = $131;
      while(1) {
       $132 = $decodeLow;
       $133 = $decodeLow;
       $134 = $decodeRange;
       $135 = (($133) + ($134))|0;
       $136 = $132 ^ $135;
       $137 = ($136>>>0)>=(16777216);
       if ($137) {
        $138 = $decodeRange;
        $139 = ($138>>>0)>=(65536);
        if ($139) {
         break L21;
        }
        $140 = $decodeLow;
        $141 = (0 - ($140))|0;
        $142 = $141 & 65535;
        $decodeRange = $142;
       }
       $143 = $decodeCode;
       $144 = $143 << 8;
       $decodeCode = $144;
       $145 = $2;
       $146 = $inEnd;
       $147 = ($145>>>0)<($146>>>0);
       if ($147) {
        $148 = $2;
        $149 = ((($148)) + 1|0);
        $2 = $149;
        $150 = HEAP8[$148>>0]|0;
        $151 = $150&255;
        $152 = $decodeCode;
        $153 = $152 | $151;
        $decodeCode = $153;
       }
       $154 = $decodeRange;
       $155 = $154 << 8;
       $decodeRange = $155;
       $156 = $decodeLow;
       $157 = $156 << 8;
       $decodeLow = $157;
      }
     }
    }
   } while(0);
   $372 = $subcontext;
   $373 = ((($372)) + 14|0);
   $374 = HEAP16[$373>>1]|0;
   $375 = $374&65535;
   $376 = $rangeCoder;
   $377 = (($376) + ($375<<4)|0);
   $subcontext = $377;
  }
  do {
   if ((label|0) == 23) {
    label = 0;
    $158 = $subcontext;
    $159 = ((($158)) + 10|0);
    $160 = HEAP16[$159>>1]|0;
    $161 = $160&65535;
    $162 = $code;
    $163 = $162&65535;
    $164 = (($163) - ($161))|0;
    $165 = $164&65535;
    $code = $165;
    $under = 0;
    $count = 0;
    $166 = $subcontext;
    $167 = ((($166)) + 8|0);
    $168 = HEAP16[$167>>1]|0;
    $169 = ($168<<16>>16)!=(0);
    if (!($169)) {
     label = 24;
     break L16;
    }
    $170 = $subcontext;
    $171 = $subcontext;
    $172 = ((($171)) + 8|0);
    $173 = HEAP16[$172>>1]|0;
    $174 = $173&65535;
    $175 = (($170) + ($174<<4)|0);
    $node = $175;
    while(1) {
     $176 = $under;
     $177 = $176&65535;
     $178 = $node;
     $179 = ((($178)) + 2|0);
     $180 = HEAP16[$179>>1]|0;
     $181 = $180&65535;
     $182 = (($177) + ($181))|0;
     $183 = $node;
     $184 = HEAP8[$183>>0]|0;
     $185 = $184&255;
     $186 = (($185) + 1)|0;
     $187 = 0;
     $188 = (($182) + ($187))|0;
     $189 = $188&65535;
     $after = $189;
     $190 = $node;
     $191 = ((($190)) + 1|0);
     $192 = HEAP8[$191>>0]|0;
     $193 = $192&255;
     $194 = (($193) + 0)|0;
     $195 = $194&65535;
     $before = $195;
     $196 = $code;
     $197 = $196&65535;
     $198 = $after;
     $199 = $198&65535;
     $200 = ($197|0)>=($199|0);
     if ($200) {
      $201 = $node;
      $202 = ((($201)) + 2|0);
      $203 = HEAP16[$202>>1]|0;
      $204 = $203&65535;
      $205 = $under;
      $206 = $205&65535;
      $207 = (($206) + ($204))|0;
      $208 = $207&65535;
      $under = $208;
      $209 = $node;
      $210 = ((($209)) + 6|0);
      $211 = HEAP16[$210>>1]|0;
      $212 = ($211<<16>>16)!=(0);
      if (!($212)) {
       label = 29;
       break L16;
      }
      $213 = $node;
      $214 = ((($213)) + 6|0);
      $215 = HEAP16[$214>>1]|0;
      $216 = $215&65535;
      $217 = $node;
      $218 = (($217) + ($216<<4)|0);
      $node = $218;
      continue;
     }
     $219 = $code;
     $220 = $219&65535;
     $221 = $after;
     $222 = $221&65535;
     $223 = $before;
     $224 = $223&65535;
     $225 = (($222) - ($224))|0;
     $226 = ($220|0)<($225|0);
     $227 = $node;
     if (!($226)) {
      break;
     }
     $228 = ((($227)) + 2|0);
     $229 = HEAP16[$228>>1]|0;
     $230 = $229&65535;
     $231 = (($230) + 2)|0;
     $232 = $231&65535;
     HEAP16[$228>>1] = $232;
     $233 = $node;
     $234 = ((($233)) + 4|0);
     $235 = HEAP16[$234>>1]|0;
     $236 = ($235<<16>>16)!=(0);
     if (!($236)) {
      label = 33;
      break L16;
     }
     $237 = $node;
     $238 = ((($237)) + 4|0);
     $239 = HEAP16[$238>>1]|0;
     $240 = $239&65535;
     $241 = $node;
     $242 = (($241) + ($240<<4)|0);
     $node = $242;
    }
    $243 = HEAP8[$227>>0]|0;
    $value = $243;
    $244 = $node;
    $245 = ((($244)) + 1|0);
    $246 = HEAP8[$245>>0]|0;
    $247 = $246&255;
    $248 = $count;
    $249 = $248&65535;
    $250 = (($249) + ($247))|0;
    $251 = $250&65535;
    $count = $251;
    $252 = $after;
    $253 = $252&65535;
    $254 = $before;
    $255 = $254&65535;
    $256 = (($253) - ($255))|0;
    $257 = $256&65535;
    $under = $257;
    $258 = $node;
    $259 = ((($258)) + 2|0);
    $260 = HEAP16[$259>>1]|0;
    $261 = $260&65535;
    $262 = (($261) + 2)|0;
    $263 = $262&65535;
    HEAP16[$259>>1] = $263;
    $264 = $node;
    $265 = ((($264)) + 1|0);
    $266 = HEAP8[$265>>0]|0;
    $267 = $266&255;
    $268 = (($267) + 2)|0;
    $269 = $268&255;
    HEAP8[$265>>0] = $269;
    $270 = $node;
    $symbol = $270;
    $271 = $symbol;
    $272 = $rangeCoder;
    $273 = $271;
    $274 = $272;
    $275 = (($273) - ($274))|0;
    $276 = (($275|0) / 16)&-1;
    $277 = $276&65535;
    $bottom = $277;
    $278 = $subcontext;
    $279 = ((($278)) + 10|0);
    $280 = HEAP16[$279>>1]|0;
    $281 = $280&65535;
    $282 = $under;
    $283 = $282&65535;
    $284 = (($281) + ($283))|0;
    $285 = $decodeRange;
    $286 = Math_imul($284, $285)|0;
    $287 = $decodeLow;
    $288 = (($287) + ($286))|0;
    $decodeLow = $288;
    $289 = $count;
    $290 = $289&65535;
    $291 = $decodeRange;
    $292 = Math_imul($291, $290)|0;
    $decodeRange = $292;
    while(1) {
     $293 = $decodeLow;
     $294 = $decodeLow;
     $295 = $decodeRange;
     $296 = (($294) + ($295))|0;
     $297 = $293 ^ $296;
     $298 = ($297>>>0)>=(16777216);
     if ($298) {
      $299 = $decodeRange;
      $300 = ($299>>>0)>=(65536);
      if ($300) {
       break;
      }
      $301 = $decodeLow;
      $302 = (0 - ($301))|0;
      $303 = $302 & 65535;
      $decodeRange = $303;
     }
     $304 = $decodeCode;
     $305 = $304 << 8;
     $decodeCode = $305;
     $306 = $2;
     $307 = $inEnd;
     $308 = ($306>>>0)<($307>>>0);
     if ($308) {
      $309 = $2;
      $310 = ((($309)) + 1|0);
      $2 = $310;
      $311 = HEAP8[$309>>0]|0;
      $312 = $311&255;
      $313 = $decodeCode;
      $314 = $313 | $312;
      $decodeCode = $314;
     }
     $315 = $decodeRange;
     $316 = $315 << 8;
     $decodeRange = $316;
     $317 = $decodeLow;
     $318 = $317 << 8;
     $decodeLow = $318;
    }
    $319 = $subcontext;
    $320 = ((($319)) + 12|0);
    $321 = HEAP16[$320>>1]|0;
    $322 = $321&65535;
    $323 = (($322) + 2)|0;
    $324 = $323&65535;
    HEAP16[$320>>1] = $324;
    $325 = $count;
    $326 = $325&65535;
    $327 = ($326|0)>(251);
    if (!($327)) {
     $328 = $subcontext;
     $329 = ((($328)) + 12|0);
     $330 = HEAP16[$329>>1]|0;
     $331 = $330&65535;
     $332 = ($331|0)>(65280);
     if (!($332)) {
      break;
     }
    }
    $333 = $subcontext;
    $334 = ((($333)) + 8|0);
    $335 = HEAP16[$334>>1]|0;
    $336 = $335&65535;
    $337 = ($336|0)!=(0);
    if ($337) {
     $338 = $subcontext;
     $339 = $subcontext;
     $340 = ((($339)) + 8|0);
     $341 = HEAP16[$340>>1]|0;
     $342 = $341&65535;
     $343 = (($338) + ($342<<4)|0);
     $344 = (_enet_symbol_rescale($343)|0);
     $345 = $344&65535;
     $347 = $345;
    } else {
     $347 = 0;
    }
    $346 = $347&65535;
    $348 = $subcontext;
    $349 = ((($348)) + 12|0);
    HEAP16[$349>>1] = $346;
    $350 = $subcontext;
    $351 = ((($350)) + 10|0);
    $352 = HEAP16[$351>>1]|0;
    $353 = $352&65535;
    $354 = $353 >> 1;
    $355 = $subcontext;
    $356 = ((($355)) + 10|0);
    $357 = HEAP16[$356>>1]|0;
    $358 = $357&65535;
    $359 = (($358) - ($354))|0;
    $360 = $359&65535;
    HEAP16[$356>>1] = $360;
    $361 = $subcontext;
    $362 = ((($361)) + 10|0);
    $363 = HEAP16[$362>>1]|0;
    $364 = $363&65535;
    $365 = (($364) + 0)|0;
    $366 = $subcontext;
    $367 = ((($366)) + 12|0);
    $368 = HEAP16[$367>>1]|0;
    $369 = $368&65535;
    $370 = (($369) + ($365))|0;
    $371 = $370&65535;
    HEAP16[$367>>1] = $371;
   }
   else if ((label|0) == 47) {
    label = 0;
    $378 = $root;
    $379 = ((($378)) + 12|0);
    $380 = HEAP16[$379>>1]|0;
    $total = $380;
    $381 = $decodeCode;
    $382 = $decodeLow;
    $383 = (($381) - ($382))|0;
    $384 = $total;
    $385 = $384&65535;
    $386 = $decodeRange;
    $387 = (($386>>>0) / ($385>>>0))&-1;
    $decodeRange = $387;
    $388 = (($383>>>0) / ($387>>>0))&-1;
    $389 = $388&65535;
    $code = $389;
    $390 = $code;
    $391 = $390&65535;
    $392 = $root;
    $393 = ((($392)) + 10|0);
    $394 = HEAP16[$393>>1]|0;
    $395 = $394&65535;
    $396 = ($391|0)<($395|0);
    if ($396) {
     label = 48;
     break L16;
    }
    $433 = $root;
    $434 = ((($433)) + 10|0);
    $435 = HEAP16[$434>>1]|0;
    $436 = $435&65535;
    $437 = $code;
    $438 = $437&65535;
    $439 = (($438) - ($436))|0;
    $440 = $439&65535;
    $code = $440;
    $under = 0;
    $count = 1;
    $441 = $root;
    $442 = ((($441)) + 8|0);
    $443 = HEAP16[$442>>1]|0;
    $444 = ($443<<16>>16)!=(0);
    do {
     if ($444) {
      $487 = $root;
      $488 = $root;
      $489 = ((($488)) + 8|0);
      $490 = HEAP16[$489>>1]|0;
      $491 = $490&65535;
      $492 = (($487) + ($491<<4)|0);
      $node1 = $492;
      while(1) {
       $493 = $under;
       $494 = $493&65535;
       $495 = $node1;
       $496 = ((($495)) + 2|0);
       $497 = HEAP16[$496>>1]|0;
       $498 = $497&65535;
       $499 = (($494) + ($498))|0;
       $500 = $node1;
       $501 = HEAP8[$500>>0]|0;
       $502 = $501&255;
       $503 = (($502) + 1)|0;
       $504 = $503;
       $505 = (($499) + ($504))|0;
       $506 = $505&65535;
       $after2 = $506;
       $507 = $node1;
       $508 = ((($507)) + 1|0);
       $509 = HEAP8[$508>>0]|0;
       $510 = $509&255;
       $511 = (($510) + 1)|0;
       $512 = $511&65535;
       $before3 = $512;
       $513 = $code;
       $514 = $513&65535;
       $515 = $after2;
       $516 = $515&65535;
       $517 = ($514|0)>=($516|0);
       if ($517) {
        $518 = $node1;
        $519 = ((($518)) + 2|0);
        $520 = HEAP16[$519>>1]|0;
        $521 = $520&65535;
        $522 = $under;
        $523 = $522&65535;
        $524 = (($523) + ($521))|0;
        $525 = $524&65535;
        $under = $525;
        $526 = $node1;
        $527 = ((($526)) + 6|0);
        $528 = HEAP16[$527>>1]|0;
        $529 = ($528<<16>>16)!=(0);
        $530 = $node1;
        if (!($529)) {
         label = 61;
         break;
        }
        $531 = ((($530)) + 6|0);
        $532 = HEAP16[$531>>1]|0;
        $533 = $532&65535;
        $534 = $node1;
        $535 = (($534) + ($533<<4)|0);
        $node1 = $535;
        continue;
       }
       $588 = $code;
       $589 = $588&65535;
       $590 = $after2;
       $591 = $590&65535;
       $592 = $before3;
       $593 = $592&65535;
       $594 = (($591) - ($593))|0;
       $595 = ($589|0)<($594|0);
       $596 = $node1;
       if (!($595)) {
        label = 66;
        break;
       }
       $597 = ((($596)) + 2|0);
       $598 = HEAP16[$597>>1]|0;
       $599 = $598&65535;
       $600 = (($599) + 3)|0;
       $601 = $600&65535;
       HEAP16[$597>>1] = $601;
       $602 = $node1;
       $603 = ((($602)) + 4|0);
       $604 = HEAP16[$603>>1]|0;
       $605 = ($604<<16>>16)!=(0);
       $606 = $node1;
       if (!($605)) {
        label = 65;
        break;
       }
       $607 = ((($606)) + 4|0);
       $608 = HEAP16[$607>>1]|0;
       $609 = $608&65535;
       $610 = $node1;
       $611 = (($610) + ($609<<4)|0);
       $node1 = $611;
      }
      if ((label|0) == 61) {
       label = 0;
       $536 = HEAP8[$530>>0]|0;
       $537 = $536&255;
       $538 = (($537) + 1)|0;
       $539 = $code;
       $540 = $539&65535;
       $541 = $after2;
       $542 = $541&65535;
       $543 = (($540) - ($542))|0;
       $544 = (($543|0) / 1)&-1;
       $545 = (($538) + ($544))|0;
       $546 = $545&255;
       $value = $546;
       $547 = $code;
       $548 = $547&65535;
       $549 = $code;
       $550 = $549&65535;
       $551 = $after2;
       $552 = $551&65535;
       $553 = (($550) - ($552))|0;
       $554 = (($553|0) % 1)&-1;
       $555 = (($548) - ($554))|0;
       $556 = $555&65535;
       $under = $556;
       $557 = $nextSymbol;
       $558 = (($557) + 1)|0;
       $nextSymbol = $558;
       $559 = $rangeCoder;
       $560 = (($559) + ($557<<4)|0);
       $symbol = $560;
       $561 = $value;
       $562 = $symbol;
       HEAP8[$562>>0] = $561;
       $563 = $symbol;
       $564 = ((($563)) + 1|0);
       HEAP8[$564>>0] = 3;
       $565 = $symbol;
       $566 = ((($565)) + 2|0);
       HEAP16[$566>>1] = 3;
       $567 = $symbol;
       $568 = ((($567)) + 4|0);
       HEAP16[$568>>1] = 0;
       $569 = $symbol;
       $570 = ((($569)) + 6|0);
       HEAP16[$570>>1] = 0;
       $571 = $symbol;
       $572 = ((($571)) + 8|0);
       HEAP16[$572>>1] = 0;
       $573 = $symbol;
       $574 = ((($573)) + 10|0);
       HEAP16[$574>>1] = 0;
       $575 = $symbol;
       $576 = ((($575)) + 12|0);
       HEAP16[$576>>1] = 0;
       $577 = $symbol;
       $578 = ((($577)) + 14|0);
       HEAP16[$578>>1] = 0;
       $579 = $symbol;
       $580 = $node1;
       $581 = $579;
       $582 = $580;
       $583 = (($581) - ($582))|0;
       $584 = (($583|0) / 16)&-1;
       $585 = $584&65535;
       $586 = $node1;
       $587 = ((($586)) + 6|0);
       HEAP16[$587>>1] = $585;
       break;
      }
      else if ((label|0) == 65) {
       label = 0;
       $612 = HEAP8[$606>>0]|0;
       $613 = $612&255;
       $614 = (($613) - 1)|0;
       $615 = $after2;
       $616 = $615&65535;
       $617 = $before3;
       $618 = $617&65535;
       $619 = (($616) - ($618))|0;
       $620 = $code;
       $621 = $620&65535;
       $622 = (($619) - ($621))|0;
       $623 = (($622) - 1)|0;
       $624 = (($623|0) / 1)&-1;
       $625 = (($614) - ($624))|0;
       $626 = $625&255;
       $value = $626;
       $627 = $code;
       $628 = $627&65535;
       $629 = $after2;
       $630 = $629&65535;
       $631 = $before3;
       $632 = $631&65535;
       $633 = (($630) - ($632))|0;
       $634 = $code;
       $635 = $634&65535;
       $636 = (($633) - ($635))|0;
       $637 = (($636) - 1)|0;
       $638 = (($637|0) % 1)&-1;
       $639 = (($628) - ($638))|0;
       $640 = $639&65535;
       $under = $640;
       $641 = $nextSymbol;
       $642 = (($641) + 1)|0;
       $nextSymbol = $642;
       $643 = $rangeCoder;
       $644 = (($643) + ($641<<4)|0);
       $symbol = $644;
       $645 = $value;
       $646 = $symbol;
       HEAP8[$646>>0] = $645;
       $647 = $symbol;
       $648 = ((($647)) + 1|0);
       HEAP8[$648>>0] = 3;
       $649 = $symbol;
       $650 = ((($649)) + 2|0);
       HEAP16[$650>>1] = 3;
       $651 = $symbol;
       $652 = ((($651)) + 4|0);
       HEAP16[$652>>1] = 0;
       $653 = $symbol;
       $654 = ((($653)) + 6|0);
       HEAP16[$654>>1] = 0;
       $655 = $symbol;
       $656 = ((($655)) + 8|0);
       HEAP16[$656>>1] = 0;
       $657 = $symbol;
       $658 = ((($657)) + 10|0);
       HEAP16[$658>>1] = 0;
       $659 = $symbol;
       $660 = ((($659)) + 12|0);
       HEAP16[$660>>1] = 0;
       $661 = $symbol;
       $662 = ((($661)) + 14|0);
       HEAP16[$662>>1] = 0;
       $663 = $symbol;
       $664 = $node1;
       $665 = $663;
       $666 = $664;
       $667 = (($665) - ($666))|0;
       $668 = (($667|0) / 16)&-1;
       $669 = $668&65535;
       $670 = $node1;
       $671 = ((($670)) + 4|0);
       HEAP16[$671>>1] = $669;
       break;
      }
      else if ((label|0) == 66) {
       label = 0;
       $672 = HEAP8[$596>>0]|0;
       $value = $672;
       $673 = $node1;
       $674 = ((($673)) + 1|0);
       $675 = HEAP8[$674>>0]|0;
       $676 = $675&255;
       $677 = $count;
       $678 = $677&65535;
       $679 = (($678) + ($676))|0;
       $680 = $679&65535;
       $count = $680;
       $681 = $after2;
       $682 = $681&65535;
       $683 = $before3;
       $684 = $683&65535;
       $685 = (($682) - ($684))|0;
       $686 = $685&65535;
       $under = $686;
       $687 = $node1;
       $688 = ((($687)) + 2|0);
       $689 = HEAP16[$688>>1]|0;
       $690 = $689&65535;
       $691 = (($690) + 3)|0;
       $692 = $691&65535;
       HEAP16[$688>>1] = $692;
       $693 = $node1;
       $694 = ((($693)) + 1|0);
       $695 = HEAP8[$694>>0]|0;
       $696 = $695&255;
       $697 = (($696) + 3)|0;
       $698 = $697&255;
       HEAP8[$694>>0] = $698;
       $699 = $node1;
       $symbol = $699;
       break;
      }
     } else {
      $445 = $code;
      $446 = $445&65535;
      $447 = (($446|0) / 1)&-1;
      $448 = $447&255;
      $value = $448;
      $449 = $code;
      $450 = $449&65535;
      $451 = $code;
      $452 = $451&65535;
      $453 = (($452|0) % 1)&-1;
      $454 = (($450) - ($453))|0;
      $455 = $454&65535;
      $under = $455;
      $456 = $nextSymbol;
      $457 = (($456) + 1)|0;
      $nextSymbol = $457;
      $458 = $rangeCoder;
      $459 = (($458) + ($456<<4)|0);
      $symbol = $459;
      $460 = $value;
      $461 = $symbol;
      HEAP8[$461>>0] = $460;
      $462 = $symbol;
      $463 = ((($462)) + 1|0);
      HEAP8[$463>>0] = 3;
      $464 = $symbol;
      $465 = ((($464)) + 2|0);
      HEAP16[$465>>1] = 3;
      $466 = $symbol;
      $467 = ((($466)) + 4|0);
      HEAP16[$467>>1] = 0;
      $468 = $symbol;
      $469 = ((($468)) + 6|0);
      HEAP16[$469>>1] = 0;
      $470 = $symbol;
      $471 = ((($470)) + 8|0);
      HEAP16[$471>>1] = 0;
      $472 = $symbol;
      $473 = ((($472)) + 10|0);
      HEAP16[$473>>1] = 0;
      $474 = $symbol;
      $475 = ((($474)) + 12|0);
      HEAP16[$475>>1] = 0;
      $476 = $symbol;
      $477 = ((($476)) + 14|0);
      HEAP16[$477>>1] = 0;
      $478 = $symbol;
      $479 = $root;
      $480 = $478;
      $481 = $479;
      $482 = (($480) - ($481))|0;
      $483 = (($482|0) / 16)&-1;
      $484 = $483&65535;
      $485 = $root;
      $486 = ((($485)) + 8|0);
      HEAP16[$486>>1] = $484;
     }
    } while(0);
    $700 = $symbol;
    $701 = $rangeCoder;
    $702 = $700;
    $703 = $701;
    $704 = (($702) - ($703))|0;
    $705 = (($704|0) / 16)&-1;
    $706 = $705&65535;
    $bottom = $706;
    $707 = $root;
    $708 = ((($707)) + 10|0);
    $709 = HEAP16[$708>>1]|0;
    $710 = $709&65535;
    $711 = $under;
    $712 = $711&65535;
    $713 = (($710) + ($712))|0;
    $714 = $decodeRange;
    $715 = Math_imul($713, $714)|0;
    $716 = $decodeLow;
    $717 = (($716) + ($715))|0;
    $decodeLow = $717;
    $718 = $count;
    $719 = $718&65535;
    $720 = $decodeRange;
    $721 = Math_imul($720, $719)|0;
    $decodeRange = $721;
    while(1) {
     $722 = $decodeLow;
     $723 = $decodeLow;
     $724 = $decodeRange;
     $725 = (($723) + ($724))|0;
     $726 = $722 ^ $725;
     $727 = ($726>>>0)>=(16777216);
     if ($727) {
      $728 = $decodeRange;
      $729 = ($728>>>0)>=(65536);
      if ($729) {
       break;
      }
      $730 = $decodeLow;
      $731 = (0 - ($730))|0;
      $732 = $731 & 65535;
      $decodeRange = $732;
     }
     $733 = $decodeCode;
     $734 = $733 << 8;
     $decodeCode = $734;
     $735 = $2;
     $736 = $inEnd;
     $737 = ($735>>>0)<($736>>>0);
     if ($737) {
      $738 = $2;
      $739 = ((($738)) + 1|0);
      $2 = $739;
      $740 = HEAP8[$738>>0]|0;
      $741 = $740&255;
      $742 = $decodeCode;
      $743 = $742 | $741;
      $decodeCode = $743;
     }
     $744 = $decodeRange;
     $745 = $744 << 8;
     $decodeRange = $745;
     $746 = $decodeLow;
     $747 = $746 << 8;
     $decodeLow = $747;
    }
    $748 = $root;
    $749 = ((($748)) + 12|0);
    $750 = HEAP16[$749>>1]|0;
    $751 = $750&65535;
    $752 = (($751) + 3)|0;
    $753 = $752&65535;
    HEAP16[$749>>1] = $753;
    $754 = $count;
    $755 = $754&65535;
    $756 = ($755|0)>(250);
    if (!($756)) {
     $757 = $root;
     $758 = ((($757)) + 12|0);
     $759 = HEAP16[$758>>1]|0;
     $760 = $759&65535;
     $761 = ($760|0)>(65280);
     if (!($761)) {
      break;
     }
    }
    $762 = $root;
    $763 = ((($762)) + 8|0);
    $764 = HEAP16[$763>>1]|0;
    $765 = $764&65535;
    $766 = ($765|0)!=(0);
    if ($766) {
     $767 = $root;
     $768 = $root;
     $769 = ((($768)) + 8|0);
     $770 = HEAP16[$769>>1]|0;
     $771 = $770&65535;
     $772 = (($767) + ($771<<4)|0);
     $773 = (_enet_symbol_rescale($772)|0);
     $774 = $773&65535;
     $776 = $774;
    } else {
     $776 = 0;
    }
    $775 = $776&65535;
    $777 = $root;
    $778 = ((($777)) + 12|0);
    HEAP16[$778>>1] = $775;
    $779 = $root;
    $780 = ((($779)) + 10|0);
    $781 = HEAP16[$780>>1]|0;
    $782 = $781&65535;
    $783 = $782 >> 1;
    $784 = $root;
    $785 = ((($784)) + 10|0);
    $786 = HEAP16[$785>>1]|0;
    $787 = $786&65535;
    $788 = (($787) - ($783))|0;
    $789 = $788&65535;
    HEAP16[$785>>1] = $789;
    $790 = $root;
    $791 = ((($790)) + 10|0);
    $792 = HEAP16[$791>>1]|0;
    $793 = $792&65535;
    $794 = (($793) + 256)|0;
    $795 = $root;
    $796 = ((($795)) + 12|0);
    $797 = HEAP16[$796>>1]|0;
    $798 = $797&65535;
    $799 = (($798) + ($794))|0;
    $800 = $799&65535;
    HEAP16[$796>>1] = $800;
   }
  } while(0);
  $801 = HEAP16[$predicted>>1]|0;
  $802 = $801&65535;
  $803 = $rangeCoder;
  $804 = (($803) + ($802<<4)|0);
  $patch = $804;
  while(1) {
   $805 = $patch;
   $806 = $subcontext;
   $807 = ($805|0)!=($806|0);
   if (!($807)) {
    break;
   }
   $808 = $value;
   $809 = $808&255;
   $810 = 0;
   $811 = $810&65535;
   $under = $811;
   $count = 0;
   $812 = $patch;
   $813 = ((($812)) + 8|0);
   $814 = HEAP16[$813>>1]|0;
   $815 = ($814<<16>>16)!=(0);
   do {
    if ($815) {
     $847 = $patch;
     $848 = $patch;
     $849 = ((($848)) + 8|0);
     $850 = HEAP16[$849>>1]|0;
     $851 = $850&65535;
     $852 = (($847) + ($851<<4)|0);
     $node4 = $852;
     while(1) {
      $853 = $value;
      $854 = $853&255;
      $855 = $node4;
      $856 = HEAP8[$855>>0]|0;
      $857 = $856&255;
      $858 = ($854|0)<($857|0);
      if ($858) {
       $859 = $node4;
       $860 = ((($859)) + 2|0);
       $861 = HEAP16[$860>>1]|0;
       $862 = $861&65535;
       $863 = (($862) + 2)|0;
       $864 = $863&65535;
       HEAP16[$860>>1] = $864;
       $865 = $node4;
       $866 = ((($865)) + 4|0);
       $867 = HEAP16[$866>>1]|0;
       $868 = ($867<<16>>16)!=(0);
       if (!($868)) {
        label = 87;
        break;
       }
       $869 = $node4;
       $870 = ((($869)) + 4|0);
       $871 = HEAP16[$870>>1]|0;
       $872 = $871&65535;
       $873 = $node4;
       $874 = (($873) + ($872<<4)|0);
       $node4 = $874;
       continue;
      }
      $906 = $value;
      $907 = $906&255;
      $908 = $node4;
      $909 = HEAP8[$908>>0]|0;
      $910 = $909&255;
      $911 = ($907|0)>($910|0);
      $912 = $node4;
      if (!($911)) {
       label = 92;
       break;
      }
      $913 = ((($912)) + 2|0);
      $914 = HEAP16[$913>>1]|0;
      $915 = $914&65535;
      $916 = $under;
      $917 = $916&65535;
      $918 = (($917) + ($915))|0;
      $919 = $918&65535;
      $under = $919;
      $920 = $node4;
      $921 = ((($920)) + 6|0);
      $922 = HEAP16[$921>>1]|0;
      $923 = ($922<<16>>16)!=(0);
      if (!($923)) {
       label = 91;
       break;
      }
      $924 = $node4;
      $925 = ((($924)) + 6|0);
      $926 = HEAP16[$925>>1]|0;
      $927 = $926&65535;
      $928 = $node4;
      $929 = (($928) + ($927<<4)|0);
      $node4 = $929;
     }
     if ((label|0) == 87) {
      label = 0;
      $875 = $nextSymbol;
      $876 = (($875) + 1)|0;
      $nextSymbol = $876;
      $877 = $rangeCoder;
      $878 = (($877) + ($875<<4)|0);
      $symbol = $878;
      $879 = $value;
      $880 = $symbol;
      HEAP8[$880>>0] = $879;
      $881 = $symbol;
      $882 = ((($881)) + 1|0);
      HEAP8[$882>>0] = 2;
      $883 = $symbol;
      $884 = ((($883)) + 2|0);
      HEAP16[$884>>1] = 2;
      $885 = $symbol;
      $886 = ((($885)) + 4|0);
      HEAP16[$886>>1] = 0;
      $887 = $symbol;
      $888 = ((($887)) + 6|0);
      HEAP16[$888>>1] = 0;
      $889 = $symbol;
      $890 = ((($889)) + 8|0);
      HEAP16[$890>>1] = 0;
      $891 = $symbol;
      $892 = ((($891)) + 10|0);
      HEAP16[$892>>1] = 0;
      $893 = $symbol;
      $894 = ((($893)) + 12|0);
      HEAP16[$894>>1] = 0;
      $895 = $symbol;
      $896 = ((($895)) + 14|0);
      HEAP16[$896>>1] = 0;
      $897 = $symbol;
      $898 = $node4;
      $899 = $897;
      $900 = $898;
      $901 = (($899) - ($900))|0;
      $902 = (($901|0) / 16)&-1;
      $903 = $902&65535;
      $904 = $node4;
      $905 = ((($904)) + 4|0);
      HEAP16[$905>>1] = $903;
      break;
     }
     else if ((label|0) == 91) {
      label = 0;
      $930 = $nextSymbol;
      $931 = (($930) + 1)|0;
      $nextSymbol = $931;
      $932 = $rangeCoder;
      $933 = (($932) + ($930<<4)|0);
      $symbol = $933;
      $934 = $value;
      $935 = $symbol;
      HEAP8[$935>>0] = $934;
      $936 = $symbol;
      $937 = ((($936)) + 1|0);
      HEAP8[$937>>0] = 2;
      $938 = $symbol;
      $939 = ((($938)) + 2|0);
      HEAP16[$939>>1] = 2;
      $940 = $symbol;
      $941 = ((($940)) + 4|0);
      HEAP16[$941>>1] = 0;
      $942 = $symbol;
      $943 = ((($942)) + 6|0);
      HEAP16[$943>>1] = 0;
      $944 = $symbol;
      $945 = ((($944)) + 8|0);
      HEAP16[$945>>1] = 0;
      $946 = $symbol;
      $947 = ((($946)) + 10|0);
      HEAP16[$947>>1] = 0;
      $948 = $symbol;
      $949 = ((($948)) + 12|0);
      HEAP16[$949>>1] = 0;
      $950 = $symbol;
      $951 = ((($950)) + 14|0);
      HEAP16[$951>>1] = 0;
      $952 = $symbol;
      $953 = $node4;
      $954 = $952;
      $955 = $953;
      $956 = (($954) - ($955))|0;
      $957 = (($956|0) / 16)&-1;
      $958 = $957&65535;
      $959 = $node4;
      $960 = ((($959)) + 6|0);
      HEAP16[$960>>1] = $958;
      break;
     }
     else if ((label|0) == 92) {
      label = 0;
      $961 = ((($912)) + 1|0);
      $962 = HEAP8[$961>>0]|0;
      $963 = $962&255;
      $964 = $count;
      $965 = $964&65535;
      $966 = (($965) + ($963))|0;
      $967 = $966&65535;
      $count = $967;
      $968 = $node4;
      $969 = ((($968)) + 2|0);
      $970 = HEAP16[$969>>1]|0;
      $971 = $970&65535;
      $972 = $node4;
      $973 = ((($972)) + 1|0);
      $974 = HEAP8[$973>>0]|0;
      $975 = $974&255;
      $976 = (($971) - ($975))|0;
      $977 = $under;
      $978 = $977&65535;
      $979 = (($978) + ($976))|0;
      $980 = $979&65535;
      $under = $980;
      $981 = $node4;
      $982 = ((($981)) + 2|0);
      $983 = HEAP16[$982>>1]|0;
      $984 = $983&65535;
      $985 = (($984) + 2)|0;
      $986 = $985&65535;
      HEAP16[$982>>1] = $986;
      $987 = $node4;
      $988 = ((($987)) + 1|0);
      $989 = HEAP8[$988>>0]|0;
      $990 = $989&255;
      $991 = (($990) + 2)|0;
      $992 = $991&255;
      HEAP8[$988>>0] = $992;
      $993 = $node4;
      $symbol = $993;
      break;
     }
    } else {
     $816 = $nextSymbol;
     $817 = (($816) + 1)|0;
     $nextSymbol = $817;
     $818 = $rangeCoder;
     $819 = (($818) + ($816<<4)|0);
     $symbol = $819;
     $820 = $value;
     $821 = $symbol;
     HEAP8[$821>>0] = $820;
     $822 = $symbol;
     $823 = ((($822)) + 1|0);
     HEAP8[$823>>0] = 2;
     $824 = $symbol;
     $825 = ((($824)) + 2|0);
     HEAP16[$825>>1] = 2;
     $826 = $symbol;
     $827 = ((($826)) + 4|0);
     HEAP16[$827>>1] = 0;
     $828 = $symbol;
     $829 = ((($828)) + 6|0);
     HEAP16[$829>>1] = 0;
     $830 = $symbol;
     $831 = ((($830)) + 8|0);
     HEAP16[$831>>1] = 0;
     $832 = $symbol;
     $833 = ((($832)) + 10|0);
     HEAP16[$833>>1] = 0;
     $834 = $symbol;
     $835 = ((($834)) + 12|0);
     HEAP16[$835>>1] = 0;
     $836 = $symbol;
     $837 = ((($836)) + 14|0);
     HEAP16[$837>>1] = 0;
     $838 = $symbol;
     $839 = $patch;
     $840 = $838;
     $841 = $839;
     $842 = (($840) - ($841))|0;
     $843 = (($842|0) / 16)&-1;
     $844 = $843&65535;
     $845 = $patch;
     $846 = ((($845)) + 8|0);
     HEAP16[$846>>1] = $844;
    }
   } while(0);
   $994 = $symbol;
   $995 = $rangeCoder;
   $996 = $994;
   $997 = $995;
   $998 = (($996) - ($997))|0;
   $999 = (($998|0) / 16)&-1;
   $1000 = $999&65535;
   $1001 = $parent;
   HEAP16[$1001>>1] = $1000;
   $1002 = $symbol;
   $1003 = ((($1002)) + 14|0);
   $parent = $1003;
   $1004 = $count;
   $1005 = $1004&65535;
   $1006 = ($1005|0)<=(0);
   if ($1006) {
    $1007 = $patch;
    $1008 = ((($1007)) + 10|0);
    $1009 = HEAP16[$1008>>1]|0;
    $1010 = $1009&65535;
    $1011 = (($1010) + 5)|0;
    $1012 = $1011&65535;
    HEAP16[$1008>>1] = $1012;
    $1013 = $patch;
    $1014 = ((($1013)) + 12|0);
    $1015 = HEAP16[$1014>>1]|0;
    $1016 = $1015&65535;
    $1017 = (($1016) + 5)|0;
    $1018 = $1017&65535;
    HEAP16[$1014>>1] = $1018;
   }
   $1019 = $patch;
   $1020 = ((($1019)) + 12|0);
   $1021 = HEAP16[$1020>>1]|0;
   $1022 = $1021&65535;
   $1023 = (($1022) + 2)|0;
   $1024 = $1023&65535;
   HEAP16[$1020>>1] = $1024;
   $1025 = $count;
   $1026 = $1025&65535;
   $1027 = ($1026|0)>(251);
   if ($1027) {
    label = 97;
   } else {
    $1028 = $patch;
    $1029 = ((($1028)) + 12|0);
    $1030 = HEAP16[$1029>>1]|0;
    $1031 = $1030&65535;
    $1032 = ($1031|0)>(65280);
    if ($1032) {
     label = 97;
    }
   }
   if ((label|0) == 97) {
    label = 0;
    $1033 = $patch;
    $1034 = ((($1033)) + 8|0);
    $1035 = HEAP16[$1034>>1]|0;
    $1036 = $1035&65535;
    $1037 = ($1036|0)!=(0);
    if ($1037) {
     $1038 = $patch;
     $1039 = $patch;
     $1040 = ((($1039)) + 8|0);
     $1041 = HEAP16[$1040>>1]|0;
     $1042 = $1041&65535;
     $1043 = (($1038) + ($1042<<4)|0);
     $1044 = (_enet_symbol_rescale($1043)|0);
     $1045 = $1044&65535;
     $1047 = $1045;
    } else {
     $1047 = 0;
    }
    $1046 = $1047&65535;
    $1048 = $patch;
    $1049 = ((($1048)) + 12|0);
    HEAP16[$1049>>1] = $1046;
    $1050 = $patch;
    $1051 = ((($1050)) + 10|0);
    $1052 = HEAP16[$1051>>1]|0;
    $1053 = $1052&65535;
    $1054 = $1053 >> 1;
    $1055 = $patch;
    $1056 = ((($1055)) + 10|0);
    $1057 = HEAP16[$1056>>1]|0;
    $1058 = $1057&65535;
    $1059 = (($1058) - ($1054))|0;
    $1060 = $1059&65535;
    HEAP16[$1056>>1] = $1060;
    $1061 = $patch;
    $1062 = ((($1061)) + 10|0);
    $1063 = HEAP16[$1062>>1]|0;
    $1064 = $1063&65535;
    $1065 = (($1064) + 0)|0;
    $1066 = $patch;
    $1067 = ((($1066)) + 12|0);
    $1068 = HEAP16[$1067>>1]|0;
    $1069 = $1068&65535;
    $1070 = (($1069) + ($1065))|0;
    $1071 = $1070&65535;
    HEAP16[$1067>>1] = $1071;
   }
   $1072 = $patch;
   $1073 = ((($1072)) + 14|0);
   $1074 = HEAP16[$1073>>1]|0;
   $1075 = $1074&65535;
   $1076 = $rangeCoder;
   $1077 = (($1076) + ($1075<<4)|0);
   $patch = $1077;
  }
  $1078 = $bottom;
  $1079 = $parent;
  HEAP16[$1079>>1] = $1078;
  $1080 = $4;
  $1081 = $outEnd;
  $1082 = ($1080>>>0)>=($1081>>>0);
  if ($1082) {
   label = 102;
   break;
  }
  $1083 = $value;
  $1084 = $4;
  $1085 = ((($1084)) + 1|0);
  $4 = $1085;
  HEAP8[$1084>>0] = $1083;
  $1086 = $order;
  $1087 = ($1086>>>0)>=(2);
  if ($1087) {
   $1088 = HEAP16[$predicted>>1]|0;
   $1089 = $1088&65535;
   $1090 = $rangeCoder;
   $1091 = (($1090) + ($1089<<4)|0);
   $1092 = ((($1091)) + 14|0);
   $1093 = HEAP16[$1092>>1]|0;
   HEAP16[$predicted>>1] = $1093;
  } else {
   $1094 = $order;
   $1095 = (($1094) + 1)|0;
   $order = $1095;
  }
  $1096 = $nextSymbol;
  $1097 = ($1096>>>0)>=(4094);
  if (!($1097)) {
   continue;
  }
  $nextSymbol = 0;
  $1098 = $nextSymbol;
  $1099 = (($1098) + 1)|0;
  $nextSymbol = $1099;
  $1100 = $rangeCoder;
  $1101 = (($1100) + ($1098<<4)|0);
  $root = $1101;
  $1102 = $root;
  HEAP8[$1102>>0] = 0;
  $1103 = $root;
  $1104 = ((($1103)) + 1|0);
  HEAP8[$1104>>0] = 0;
  $1105 = $root;
  $1106 = ((($1105)) + 2|0);
  HEAP16[$1106>>1] = 0;
  $1107 = $root;
  $1108 = ((($1107)) + 4|0);
  HEAP16[$1108>>1] = 0;
  $1109 = $root;
  $1110 = ((($1109)) + 6|0);
  HEAP16[$1110>>1] = 0;
  $1111 = $root;
  $1112 = ((($1111)) + 8|0);
  HEAP16[$1112>>1] = 0;
  $1113 = $root;
  $1114 = ((($1113)) + 10|0);
  HEAP16[$1114>>1] = 0;
  $1115 = $root;
  $1116 = ((($1115)) + 12|0);
  HEAP16[$1116>>1] = 0;
  $1117 = $root;
  $1118 = ((($1117)) + 14|0);
  HEAP16[$1118>>1] = 0;
  $1119 = $root;
  $1120 = ((($1119)) + 10|0);
  HEAP16[$1120>>1] = 1;
  $1121 = $root;
  $1122 = ((($1121)) + 12|0);
  HEAP16[$1122>>1] = 257;
  $1123 = $root;
  $1124 = ((($1123)) + 8|0);
  HEAP16[$1124>>1] = 0;
  HEAP16[$predicted>>1] = 0;
  $order = 0;
 }
 if ((label|0) == 24) {
  $0 = 0;
  $1130 = $0;
  STACKTOP = sp;return ($1130|0);
 }
 else if ((label|0) == 29) {
  $0 = 0;
  $1130 = $0;
  STACKTOP = sp;return ($1130|0);
 }
 else if ((label|0) == 33) {
  $0 = 0;
  $1130 = $0;
  STACKTOP = sp;return ($1130|0);
 }
 else if ((label|0) == 48) {
  $397 = $decodeRange;
  $398 = 0;
  $399 = $decodeLow;
  $400 = (($399) + ($398))|0;
  $decodeLow = $400;
  $401 = $root;
  $402 = ((($401)) + 10|0);
  $403 = HEAP16[$402>>1]|0;
  $404 = $403&65535;
  $405 = $decodeRange;
  $406 = Math_imul($405, $404)|0;
  $decodeRange = $406;
  while(1) {
   $407 = $decodeLow;
   $408 = $decodeLow;
   $409 = $decodeRange;
   $410 = (($408) + ($409))|0;
   $411 = $407 ^ $410;
   $412 = ($411>>>0)>=(16777216);
   if ($412) {
    $413 = $decodeRange;
    $414 = ($413>>>0)>=(65536);
    if ($414) {
     break;
    }
    $415 = $decodeLow;
    $416 = (0 - ($415))|0;
    $417 = $416 & 65535;
    $decodeRange = $417;
   }
   $418 = $decodeCode;
   $419 = $418 << 8;
   $decodeCode = $419;
   $420 = $2;
   $421 = $inEnd;
   $422 = ($420>>>0)<($421>>>0);
   if ($422) {
    $423 = $2;
    $424 = ((($423)) + 1|0);
    $2 = $424;
    $425 = HEAP8[$423>>0]|0;
    $426 = $425&255;
    $427 = $decodeCode;
    $428 = $427 | $426;
    $decodeCode = $428;
   }
   $429 = $decodeRange;
   $430 = $429 << 8;
   $decodeRange = $430;
   $431 = $decodeLow;
   $432 = $431 << 8;
   $decodeLow = $432;
  }
  $1125 = $4;
  $1126 = $outStart;
  $1127 = $1125;
  $1128 = $1126;
  $1129 = (($1127) - ($1128))|0;
  $0 = $1129;
  $1130 = $0;
  STACKTOP = sp;return ($1130|0);
 }
 else if ((label|0) == 102) {
  $0 = 0;
  $1130 = $0;
  STACKTOP = sp;return ($1130|0);
 }
 return (0)|0;
}
function _enet_host_compress_with_range_coder($host) {
 $host = $host|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $compressor = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $compressor = sp;
 $1 = $host;
 ;HEAP32[$compressor>>2]=0|0;HEAP32[$compressor+4>>2]=0|0;HEAP32[$compressor+8>>2]=0|0;HEAP32[$compressor+12>>2]=0|0;
 $2 = (_enet_range_coder_create()|0);
 HEAP32[$compressor>>2] = $2;
 $3 = HEAP32[$compressor>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  $0 = -1;
  $9 = $0;
  STACKTOP = sp;return ($9|0);
 } else {
  $5 = ((($compressor)) + 4|0);
  HEAP32[$5>>2] = 3478;
  $6 = ((($compressor)) + 8|0);
  HEAP32[$6>>2] = 3479;
  $7 = ((($compressor)) + 12|0);
  HEAP32[$7>>2] = 3480;
  $8 = $1;
  _enet_host_compress($8,$compressor);
  $0 = 0;
  $9 = $0;
  STACKTOP = sp;return ($9|0);
 }
 return (0)|0;
}
function _enet_host_create($address,$peerCount,$channelLimit,$incomingBandwidth,$outgoingBandwidth) {
 $address = $address|0;
 $peerCount = $peerCount|0;
 $channelLimit = $channelLimit|0;
 $incomingBandwidth = $incomingBandwidth|0;
 $outgoingBandwidth = $outgoingBandwidth|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $currentPeer = 0, $host = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $address;
 $2 = $peerCount;
 $3 = $channelLimit;
 $4 = $incomingBandwidth;
 $5 = $outgoingBandwidth;
 $6 = $2;
 $7 = ($6>>>0)>(4095);
 if ($7) {
  $0 = 0;
  $201 = $0;
  STACKTOP = sp;return ($201|0);
 }
 $8 = (_enet_malloc(10396)|0);
 $host = $8;
 $9 = $host;
 $10 = ($9|0)==(0|0);
 if ($10) {
  $0 = 0;
  $201 = $0;
  STACKTOP = sp;return ($201|0);
 }
 $11 = $host;
 _memset(($11|0),0,10396)|0;
 $12 = $2;
 $13 = ($12*380)|0;
 $14 = (_enet_malloc($13)|0);
 $15 = $host;
 $16 = ((($15)) + 36|0);
 HEAP32[$16>>2] = $14;
 $17 = $host;
 $18 = ((($17)) + 36|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($19|0)==(0|0);
 $21 = $host;
 if ($20) {
  _enet_free($21);
  $0 = 0;
  $201 = $0;
  STACKTOP = sp;return ($201|0);
 }
 $22 = ((($21)) + 36|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = $2;
 $25 = ($24*380)|0;
 _memset(($23|0),0,($25|0))|0;
 $26 = (_enet_socket_create(2)|0);
 $27 = $host;
 HEAP32[$27>>2] = $26;
 $28 = $host;
 $29 = HEAP32[$28>>2]|0;
 $30 = ($29|0)==(-1);
 do {
  if (!($30)) {
   $31 = $1;
   $32 = ($31|0)!=(0|0);
   if ($32) {
    $33 = $host;
    $34 = HEAP32[$33>>2]|0;
    $35 = $1;
    $36 = (_enet_socket_bind($34,$35)|0);
    $37 = ($36|0)<(0);
    if ($37) {
     break;
    }
   }
   $47 = $host;
   $48 = HEAP32[$47>>2]|0;
   (_enet_socket_set_option($48,1,1)|0);
   $49 = $host;
   $50 = HEAP32[$49>>2]|0;
   (_enet_socket_set_option($50,2,1)|0);
   $51 = $host;
   $52 = HEAP32[$51>>2]|0;
   (_enet_socket_set_option($52,3,262144)|0);
   $53 = $host;
   $54 = HEAP32[$53>>2]|0;
   (_enet_socket_set_option($54,4,262144)|0);
   $55 = $1;
   $56 = ($55|0)!=(0|0);
   if ($56) {
    $57 = $host;
    $58 = HEAP32[$57>>2]|0;
    $59 = $host;
    $60 = ((($59)) + 4|0);
    $61 = (_enet_socket_get_address($58,$60)|0);
    $62 = ($61|0)<(0);
    if ($62) {
     $63 = $host;
     $64 = ((($63)) + 4|0);
     $65 = $1;
     ;HEAP32[$64>>2]=HEAP32[$65>>2]|0;HEAP32[$64+4>>2]=HEAP32[$65+4>>2]|0;
    }
   }
   $66 = $3;
   $67 = ($66|0)==(0);
   $68 = $3;
   $69 = ($68>>>0)>(255);
   $or$cond = $67 | $69;
   if ($or$cond) {
    $3 = 255;
   } else {
    $70 = $3;
    $71 = ($70>>>0)<(1);
    if ($71) {
     $3 = 1;
    }
   }
   $72 = $host;
   $73 = $72;
   $74 = $host;
   $75 = ((($74)) + 28|0);
   HEAP32[$75>>2] = $73;
   $76 = (_enet_host_random_seed()|0);
   $77 = $host;
   $78 = ((($77)) + 28|0);
   $79 = HEAP32[$78>>2]|0;
   $80 = (($79) + ($76))|0;
   HEAP32[$78>>2] = $80;
   $81 = $host;
   $82 = ((($81)) + 28|0);
   $83 = HEAP32[$82>>2]|0;
   $84 = $83 << 16;
   $85 = $host;
   $86 = ((($85)) + 28|0);
   $87 = HEAP32[$86>>2]|0;
   $88 = $87 >>> 16;
   $89 = $84 | $88;
   $90 = $host;
   $91 = ((($90)) + 28|0);
   HEAP32[$91>>2] = $89;
   $92 = $3;
   $93 = $host;
   $94 = ((($93)) + 44|0);
   HEAP32[$94>>2] = $92;
   $95 = $4;
   $96 = $host;
   $97 = ((($96)) + 12|0);
   HEAP32[$97>>2] = $95;
   $98 = $5;
   $99 = $host;
   $100 = ((($99)) + 16|0);
   HEAP32[$100>>2] = $98;
   $101 = $host;
   $102 = ((($101)) + 20|0);
   HEAP32[$102>>2] = 0;
   $103 = $host;
   $104 = ((($103)) + 32|0);
   HEAP32[$104>>2] = 0;
   $105 = $host;
   $106 = ((($105)) + 24|0);
   HEAP32[$106>>2] = 1400;
   $107 = $2;
   $108 = $host;
   $109 = ((($108)) + 40|0);
   HEAP32[$109>>2] = $107;
   $110 = $host;
   $111 = ((($110)) + 1608|0);
   HEAP32[$111>>2] = 0;
   $112 = $host;
   $113 = ((($112)) + 2132|0);
   HEAP32[$113>>2] = 0;
   $114 = $host;
   $115 = ((($114)) + 2136|0);
   HEAP32[$115>>2] = 0;
   $116 = $host;
   $117 = ((($116)) + 10348|0);
   HEAP32[$117>>2] = 0;
   $118 = $host;
   $119 = ((($118)) + 10348|0);
   $120 = ((($119)) + 4|0);
   HEAP16[$120>>1] = 0;
   $121 = $host;
   $122 = ((($121)) + 10356|0);
   HEAP32[$122>>2] = 0;
   $123 = $host;
   $124 = ((($123)) + 10360|0);
   HEAP32[$124>>2] = 0;
   $125 = $host;
   $126 = ((($125)) + 10364|0);
   HEAP32[$126>>2] = 0;
   $127 = $host;
   $128 = ((($127)) + 10368|0);
   HEAP32[$128>>2] = 0;
   $129 = $host;
   $130 = ((($129)) + 10372|0);
   HEAP32[$130>>2] = 0;
   $131 = $host;
   $132 = ((($131)) + 10376|0);
   HEAP32[$132>>2] = 0;
   $133 = $host;
   $134 = ((($133)) + 10384|0);
   HEAP32[$134>>2] = 0;
   $135 = $host;
   $136 = ((($135)) + 10388|0);
   HEAP32[$136>>2] = 0;
   $137 = $host;
   $138 = ((($137)) + 10392|0);
   HEAP32[$138>>2] = 4095;
   $139 = $host;
   $140 = ((($139)) + 2140|0);
   HEAP32[$140>>2] = 0;
   $141 = $host;
   $142 = ((($141)) + 2140|0);
   $143 = ((($142)) + 4|0);
   HEAP32[$143>>2] = 0;
   $144 = $host;
   $145 = ((($144)) + 2140|0);
   $146 = ((($145)) + 8|0);
   HEAP32[$146>>2] = 0;
   $147 = $host;
   $148 = ((($147)) + 2140|0);
   $149 = ((($148)) + 12|0);
   HEAP32[$149>>2] = 0;
   $150 = $host;
   $151 = ((($150)) + 10380|0);
   HEAP32[$151>>2] = 0;
   $152 = $host;
   $153 = ((($152)) + 52|0);
   _enet_list_clear($153);
   $154 = $host;
   $155 = ((($154)) + 36|0);
   $156 = HEAP32[$155>>2]|0;
   $currentPeer = $156;
   while(1) {
    $157 = $currentPeer;
    $158 = $host;
    $159 = ((($158)) + 40|0);
    $160 = HEAP32[$159>>2]|0;
    $161 = $host;
    $162 = ((($161)) + 36|0);
    $163 = HEAP32[$162>>2]|0;
    $164 = (($163) + (($160*380)|0)|0);
    $165 = ($157>>>0)<($164>>>0);
    $166 = $host;
    if (!($165)) {
     break;
    }
    $167 = $currentPeer;
    $168 = ((($167)) + 8|0);
    HEAP32[$168>>2] = $166;
    $169 = $currentPeer;
    $170 = $host;
    $171 = ((($170)) + 36|0);
    $172 = HEAP32[$171>>2]|0;
    $173 = $169;
    $174 = $172;
    $175 = (($173) - ($174))|0;
    $176 = (($175|0) / 380)&-1;
    $177 = $176&65535;
    $178 = $currentPeer;
    $179 = ((($178)) + 14|0);
    HEAP16[$179>>1] = $177;
    $180 = $currentPeer;
    $181 = ((($180)) + 21|0);
    HEAP8[$181>>0] = -1;
    $182 = $currentPeer;
    $183 = ((($182)) + 20|0);
    HEAP8[$183>>0] = -1;
    $184 = $currentPeer;
    $185 = ((($184)) + 32|0);
    HEAP32[$185>>2] = 0;
    $186 = $currentPeer;
    $187 = ((($186)) + 192|0);
    _enet_list_clear($187);
    $188 = $currentPeer;
    $189 = ((($188)) + 200|0);
    _enet_list_clear($189);
    $190 = $currentPeer;
    $191 = ((($190)) + 208|0);
    _enet_list_clear($191);
    $192 = $currentPeer;
    $193 = ((($192)) + 216|0);
    _enet_list_clear($193);
    $194 = $currentPeer;
    $195 = ((($194)) + 224|0);
    _enet_list_clear($195);
    $196 = $currentPeer;
    $197 = ((($196)) + 232|0);
    _enet_list_clear($197);
    $198 = $currentPeer;
    _enet_peer_reset($198);
    $199 = $currentPeer;
    $200 = ((($199)) + 380|0);
    $currentPeer = $200;
   }
   $0 = $166;
   $201 = $0;
   STACKTOP = sp;return ($201|0);
  }
 } while(0);
 $38 = $host;
 $39 = HEAP32[$38>>2]|0;
 $40 = ($39|0)!=(-1);
 if ($40) {
  $41 = $host;
  $42 = HEAP32[$41>>2]|0;
  _enet_socket_destroy($42);
 }
 $43 = $host;
 $44 = ((($43)) + 36|0);
 $45 = HEAP32[$44>>2]|0;
 _enet_free($45);
 $46 = $host;
 _enet_free($46);
 $0 = 0;
 $201 = $0;
 STACKTOP = sp;return ($201|0);
}
function _enet_host_destroy($host) {
 $host = $host|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $currentPeer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $host;
 $1 = $0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  STACKTOP = sp;return;
 }
 $3 = $0;
 $4 = HEAP32[$3>>2]|0;
 _enet_socket_destroy($4);
 $5 = $0;
 $6 = ((($5)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $currentPeer = $7;
 while(1) {
  $8 = $currentPeer;
  $9 = $0;
  $10 = ((($9)) + 40|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = $0;
  $13 = ((($12)) + 36|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = (($14) + (($11*380)|0)|0);
  $16 = ($8>>>0)<($15>>>0);
  if (!($16)) {
   break;
  }
  $17 = $currentPeer;
  _enet_peer_reset($17);
  $18 = $currentPeer;
  $19 = ((($18)) + 380|0);
  $currentPeer = $19;
 }
 $20 = $0;
 $21 = ((($20)) + 2140|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ($22|0)!=(0|0);
 if ($23) {
  $24 = $0;
  $25 = ((($24)) + 2140|0);
  $26 = ((($25)) + 12|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = ($27|0)!=(0|0);
  if ($28) {
   $29 = $0;
   $30 = ((($29)) + 2140|0);
   $31 = ((($30)) + 12|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = $0;
   $34 = ((($33)) + 2140|0);
   $35 = HEAP32[$34>>2]|0;
   FUNCTION_TABLE_vi[$32 & 8191]($35);
  }
 }
 $36 = $0;
 $37 = ((($36)) + 36|0);
 $38 = HEAP32[$37>>2]|0;
 _enet_free($38);
 $39 = $0;
 _enet_free($39);
 STACKTOP = sp;return;
}
function _enet_host_connect($host,$address,$channelCount,$data) {
 $host = $host|0;
 $address = $address|0;
 $channelCount = $channelCount|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $channel = 0, $command = 0, $currentPeer = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $command = sp + 32|0;
 $1 = $host;
 $2 = $address;
 $3 = $channelCount;
 $4 = $data;
 $5 = $3;
 $6 = ($5>>>0)<(1);
 if ($6) {
  $3 = 1;
 } else {
  $7 = $3;
  $8 = ($7>>>0)>(255);
  if ($8) {
   $3 = 255;
  }
 }
 $9 = $1;
 $10 = ((($9)) + 36|0);
 $11 = HEAP32[$10>>2]|0;
 $currentPeer = $11;
 while(1) {
  $12 = $currentPeer;
  $13 = $1;
  $14 = ((($13)) + 40|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = $1;
  $17 = ((($16)) + 36|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = (($18) + (($15*380)|0)|0);
  $20 = ($12>>>0)<($19>>>0);
  if (!($20)) {
   break;
  }
  $21 = $currentPeer;
  $22 = ((($21)) + 36|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = ($23|0)==(0);
  if ($24) {
   break;
  }
  $25 = $currentPeer;
  $26 = ((($25)) + 380|0);
  $currentPeer = $26;
 }
 $27 = $currentPeer;
 $28 = $1;
 $29 = ((($28)) + 40|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = $1;
 $32 = ((($31)) + 36|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = (($33) + (($30*380)|0)|0);
 $35 = ($27>>>0)>=($34>>>0);
 if ($35) {
  $0 = 0;
  $170 = $0;
  STACKTOP = sp;return ($170|0);
 }
 $36 = $3;
 $37 = ($36*60)|0;
 $38 = (_enet_malloc($37)|0);
 $39 = $currentPeer;
 $40 = ((($39)) + 40|0);
 HEAP32[$40>>2] = $38;
 $41 = $currentPeer;
 $42 = ((($41)) + 40|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = ($43|0)==(0|0);
 if ($44) {
  $0 = 0;
  $170 = $0;
  STACKTOP = sp;return ($170|0);
 }
 $45 = $3;
 $46 = $currentPeer;
 $47 = ((($46)) + 44|0);
 HEAP32[$47>>2] = $45;
 $48 = $currentPeer;
 $49 = ((($48)) + 36|0);
 HEAP32[$49>>2] = 1;
 $50 = $currentPeer;
 $51 = ((($50)) + 24|0);
 $52 = $2;
 ;HEAP32[$51>>2]=HEAP32[$52>>2]|0;HEAP32[$51+4>>2]=HEAP32[$52+4>>2]|0;
 $53 = $1;
 $54 = ((($53)) + 28|0);
 $55 = HEAP32[$54>>2]|0;
 $56 = (($55) + 1)|0;
 HEAP32[$54>>2] = $56;
 $57 = $currentPeer;
 $58 = ((($57)) + 16|0);
 HEAP32[$58>>2] = $56;
 $59 = $1;
 $60 = ((($59)) + 16|0);
 $61 = HEAP32[$60>>2]|0;
 $62 = ($61|0)==(0);
 if ($62) {
  $63 = $currentPeer;
  $64 = ((($63)) + 180|0);
  HEAP32[$64>>2] = 65536;
 } else {
  $65 = $1;
  $66 = ((($65)) + 16|0);
  $67 = HEAP32[$66>>2]|0;
  $68 = (($67>>>0) / 65536)&-1;
  $69 = $68<<12;
  $70 = $currentPeer;
  $71 = ((($70)) + 180|0);
  HEAP32[$71>>2] = $69;
 }
 $72 = $currentPeer;
 $73 = ((($72)) + 180|0);
 $74 = HEAP32[$73>>2]|0;
 $75 = ($74>>>0)<(4096);
 $76 = $currentPeer;
 $77 = ((($76)) + 180|0);
 if ($75) {
  HEAP32[$77>>2] = 4096;
 } else {
  $78 = HEAP32[$77>>2]|0;
  $79 = ($78>>>0)>(65536);
  if ($79) {
   $80 = $currentPeer;
   $81 = ((($80)) + 180|0);
   HEAP32[$81>>2] = 65536;
  }
 }
 $82 = $currentPeer;
 $83 = ((($82)) + 40|0);
 $84 = HEAP32[$83>>2]|0;
 $channel = $84;
 while(1) {
  $85 = $channel;
  $86 = $3;
  $87 = $currentPeer;
  $88 = ((($87)) + 40|0);
  $89 = HEAP32[$88>>2]|0;
  $90 = (($89) + (($86*60)|0)|0);
  $91 = ($85>>>0)<($90>>>0);
  if (!($91)) {
   break;
  }
  $92 = $channel;
  HEAP16[$92>>1] = 0;
  $93 = $channel;
  $94 = ((($93)) + 2|0);
  HEAP16[$94>>1] = 0;
  $95 = $channel;
  $96 = ((($95)) + 38|0);
  HEAP16[$96>>1] = 0;
  $97 = $channel;
  $98 = ((($97)) + 40|0);
  HEAP16[$98>>1] = 0;
  $99 = $channel;
  $100 = ((($99)) + 44|0);
  _enet_list_clear($100);
  $101 = $channel;
  $102 = ((($101)) + 52|0);
  _enet_list_clear($102);
  $103 = $channel;
  $104 = ((($103)) + 4|0);
  HEAP16[$104>>1] = 0;
  $105 = $channel;
  $106 = ((($105)) + 6|0);
  dest=$106; stop=dest+32|0; do { HEAP16[dest>>1]=0|0; dest=dest+2|0; } while ((dest|0) < (stop|0));
  $107 = $channel;
  $108 = ((($107)) + 60|0);
  $channel = $108;
 }
 HEAP8[$command>>0] = -126;
 $109 = ((($command)) + 1|0);
 HEAP8[$109>>0] = -1;
 $110 = $currentPeer;
 $111 = ((($110)) + 14|0);
 $112 = HEAP16[$111>>1]|0;
 $113 = (_htons($112)|0);
 $114 = ((($command)) + 4|0);
 HEAP8[$114>>0]=$113&255;HEAP8[$114+1>>0]=$113>>8;
 $115 = $currentPeer;
 $116 = ((($115)) + 21|0);
 $117 = HEAP8[$116>>0]|0;
 $118 = ((($command)) + 6|0);
 HEAP8[$118>>0] = $117;
 $119 = $currentPeer;
 $120 = ((($119)) + 20|0);
 $121 = HEAP8[$120>>0]|0;
 $122 = ((($command)) + 7|0);
 HEAP8[$122>>0] = $121;
 $123 = $currentPeer;
 $124 = ((($123)) + 176|0);
 $125 = HEAP32[$124>>2]|0;
 $126 = (_htonl($125)|0);
 $127 = ((($command)) + 8|0);
 HEAP8[$127>>0]=$126&255;HEAP8[$127+1>>0]=($126>>8)&255;HEAP8[$127+2>>0]=($126>>16)&255;HEAP8[$127+3>>0]=$126>>24;
 $128 = $currentPeer;
 $129 = ((($128)) + 180|0);
 $130 = HEAP32[$129>>2]|0;
 $131 = (_htonl($130)|0);
 $132 = ((($command)) + 12|0);
 HEAP8[$132>>0]=$131&255;HEAP8[$132+1>>0]=($131>>8)&255;HEAP8[$132+2>>0]=($131>>16)&255;HEAP8[$132+3>>0]=$131>>24;
 $133 = $3;
 $134 = (_htonl($133)|0);
 $135 = ((($command)) + 16|0);
 HEAP8[$135>>0]=$134&255;HEAP8[$135+1>>0]=($134>>8)&255;HEAP8[$135+2>>0]=($134>>16)&255;HEAP8[$135+3>>0]=$134>>24;
 $136 = $1;
 $137 = ((($136)) + 12|0);
 $138 = HEAP32[$137>>2]|0;
 $139 = (_htonl($138)|0);
 $140 = ((($command)) + 20|0);
 HEAP8[$140>>0]=$139&255;HEAP8[$140+1>>0]=($139>>8)&255;HEAP8[$140+2>>0]=($139>>16)&255;HEAP8[$140+3>>0]=$139>>24;
 $141 = $1;
 $142 = ((($141)) + 16|0);
 $143 = HEAP32[$142>>2]|0;
 $144 = (_htonl($143)|0);
 $145 = ((($command)) + 24|0);
 HEAP8[$145>>0]=$144&255;HEAP8[$145+1>>0]=($144>>8)&255;HEAP8[$145+2>>0]=($144>>16)&255;HEAP8[$145+3>>0]=$144>>24;
 $146 = $currentPeer;
 $147 = ((($146)) + 132|0);
 $148 = HEAP32[$147>>2]|0;
 $149 = (_htonl($148)|0);
 $150 = ((($command)) + 28|0);
 HEAP8[$150>>0]=$149&255;HEAP8[$150+1>>0]=($149>>8)&255;HEAP8[$150+2>>0]=($149>>16)&255;HEAP8[$150+3>>0]=$149>>24;
 $151 = $currentPeer;
 $152 = ((($151)) + 124|0);
 $153 = HEAP32[$152>>2]|0;
 $154 = (_htonl($153)|0);
 $155 = ((($command)) + 32|0);
 HEAP8[$155>>0]=$154&255;HEAP8[$155+1>>0]=($154>>8)&255;HEAP8[$155+2>>0]=($154>>16)&255;HEAP8[$155+3>>0]=$154>>24;
 $156 = $currentPeer;
 $157 = ((($156)) + 128|0);
 $158 = HEAP32[$157>>2]|0;
 $159 = (_htonl($158)|0);
 $160 = ((($command)) + 36|0);
 HEAP8[$160>>0]=$159&255;HEAP8[$160+1>>0]=($159>>8)&255;HEAP8[$160+2>>0]=($159>>16)&255;HEAP8[$160+3>>0]=$159>>24;
 $161 = $currentPeer;
 $162 = ((($161)) + 16|0);
 $163 = HEAP32[$162>>2]|0;
 $164 = ((($command)) + 40|0);
 HEAP8[$164>>0]=$163&255;HEAP8[$164+1>>0]=($163>>8)&255;HEAP8[$164+2>>0]=($163>>16)&255;HEAP8[$164+3>>0]=$163>>24;
 $165 = $4;
 $166 = (_htonl($165)|0);
 $167 = ((($command)) + 44|0);
 HEAP8[$167>>0]=$166&255;HEAP8[$167+1>>0]=($166>>8)&255;HEAP8[$167+2>>0]=($166>>16)&255;HEAP8[$167+3>>0]=$166>>24;
 $168 = $currentPeer;
 (_enet_peer_queue_outgoing_command($168,$command,0,0,0)|0);
 $169 = $currentPeer;
 $0 = $169;
 $170 = $0;
 STACKTOP = sp;return ($170|0);
}
function _enet_host_broadcast($host,$channelID,$packet) {
 $host = $host|0;
 $channelID = $channelID|0;
 $packet = $packet|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $currentPeer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $host;
 $1 = $channelID;
 $2 = $packet;
 $3 = $0;
 $4 = ((($3)) + 36|0);
 $5 = HEAP32[$4>>2]|0;
 $currentPeer = $5;
 while(1) {
  $6 = $currentPeer;
  $7 = $0;
  $8 = ((($7)) + 40|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = $0;
  $11 = ((($10)) + 36|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = (($12) + (($9*380)|0)|0);
  $14 = ($6>>>0)<($13>>>0);
  if (!($14)) {
   break;
  }
  $15 = $currentPeer;
  $16 = ((($15)) + 36|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = ($17|0)!=(5);
  if (!($18)) {
   $19 = $currentPeer;
   $20 = $1;
   $21 = $2;
   (_enet_peer_send($19,$20,$21)|0);
  }
  $22 = $currentPeer;
  $23 = ((($22)) + 380|0);
  $currentPeer = $23;
 }
 $24 = $2;
 $25 = HEAP32[$24>>2]|0;
 $26 = ($25|0)==(0);
 if (!($26)) {
  STACKTOP = sp;return;
 }
 $27 = $2;
 _enet_packet_destroy($27);
 STACKTOP = sp;return;
}
function _enet_host_compress($host,$compressor) {
 $host = $host|0;
 $compressor = $compressor|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $host;
 $1 = $compressor;
 $2 = $0;
 $3 = ((($2)) + 2140|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)!=(0|0);
 if ($5) {
  $6 = $0;
  $7 = ((($6)) + 2140|0);
  $8 = ((($7)) + 12|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)!=(0|0);
  if ($10) {
   $11 = $0;
   $12 = ((($11)) + 2140|0);
   $13 = ((($12)) + 12|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = $0;
   $16 = ((($15)) + 2140|0);
   $17 = HEAP32[$16>>2]|0;
   FUNCTION_TABLE_vi[$14 & 8191]($17);
  }
 }
 $18 = $1;
 $19 = ($18|0)!=(0|0);
 $20 = $0;
 $21 = ((($20)) + 2140|0);
 if ($19) {
  $22 = $1;
  ;HEAP32[$21>>2]=HEAP32[$22>>2]|0;HEAP32[$21+4>>2]=HEAP32[$22+4>>2]|0;HEAP32[$21+8>>2]=HEAP32[$22+8>>2]|0;HEAP32[$21+12>>2]=HEAP32[$22+12>>2]|0;
  STACKTOP = sp;return;
 } else {
  HEAP32[$21>>2] = 0;
  STACKTOP = sp;return;
 }
}
function _enet_host_channel_limit($host,$channelLimit) {
 $host = $host|0;
 $channelLimit = $channelLimit|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $host;
 $1 = $channelLimit;
 $2 = $1;
 $3 = ($2|0)==(0);
 $4 = $1;
 $5 = ($4>>>0)>(255);
 $or$cond = $3 | $5;
 if ($or$cond) {
  $1 = 255;
 } else {
  $6 = $1;
  $7 = ($6>>>0)<(1);
  if ($7) {
   $1 = 1;
  }
 }
 $8 = $1;
 $9 = $0;
 $10 = ((($9)) + 44|0);
 HEAP32[$10>>2] = $8;
 STACKTOP = sp;return;
}
function _enet_host_bandwidth_limit($host,$incomingBandwidth,$outgoingBandwidth) {
 $host = $host|0;
 $incomingBandwidth = $incomingBandwidth|0;
 $outgoingBandwidth = $outgoingBandwidth|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $host;
 $1 = $incomingBandwidth;
 $2 = $outgoingBandwidth;
 $3 = $1;
 $4 = $0;
 $5 = ((($4)) + 12|0);
 HEAP32[$5>>2] = $3;
 $6 = $2;
 $7 = $0;
 $8 = ((($7)) + 16|0);
 HEAP32[$8>>2] = $6;
 $9 = $0;
 $10 = ((($9)) + 32|0);
 HEAP32[$10>>2] = 1;
 STACKTOP = sp;return;
}
function _enet_host_bandwidth_throttle($host) {
 $host = $host|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $bandwidth = 0, $bandwidthLimit = 0, $command = 0, $dataTotal = 0, $elapsedTime = 0, $needsAdjustment = 0, $peer = 0, $peerBandwidth = 0, $peersRemaining = 0, $throttle = 0, $timeCurrent = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $command = sp + 48|0;
 $0 = $host;
 $1 = (_enet_time_get()|0);
 $timeCurrent = $1;
 $2 = $timeCurrent;
 $3 = $0;
 $4 = ((($3)) + 20|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($2) - ($5))|0;
 $elapsedTime = $6;
 $7 = $0;
 $8 = ((($7)) + 10384|0);
 $9 = HEAP32[$8>>2]|0;
 $peersRemaining = $9;
 $dataTotal = -1;
 $bandwidth = -1;
 $throttle = 0;
 $bandwidthLimit = 0;
 $10 = $0;
 $11 = ((($10)) + 10388|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12>>>0)>(0);
 $14 = $13 ? 1 : 0;
 $needsAdjustment = $14;
 $15 = $elapsedTime;
 $16 = ($15>>>0)<(1000);
 if ($16) {
  STACKTOP = sp;return;
 }
 $17 = $timeCurrent;
 $18 = $0;
 $19 = ((($18)) + 20|0);
 HEAP32[$19>>2] = $17;
 $20 = $peersRemaining;
 $21 = ($20|0)==(0);
 if ($21) {
  STACKTOP = sp;return;
 }
 $22 = $0;
 $23 = ((($22)) + 16|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ($24|0)!=(0);
 L7: do {
  if ($25) {
   $dataTotal = 0;
   $26 = $0;
   $27 = ((($26)) + 16|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = $elapsedTime;
   $30 = Math_imul($28, $29)|0;
   $31 = (($30>>>0) / 1000)&-1;
   $bandwidth = $31;
   $32 = $0;
   $33 = ((($32)) + 36|0);
   $34 = HEAP32[$33>>2]|0;
   $peer = $34;
   while(1) {
    $35 = $peer;
    $36 = $0;
    $37 = ((($36)) + 40|0);
    $38 = HEAP32[$37>>2]|0;
    $39 = $0;
    $40 = ((($39)) + 36|0);
    $41 = HEAP32[$40>>2]|0;
    $42 = (($41) + (($38*380)|0)|0);
    $43 = ($35>>>0)<($42>>>0);
    if (!($43)) {
     break L7;
    }
    $44 = $peer;
    $45 = ((($44)) + 36|0);
    $46 = HEAP32[$45>>2]|0;
    $47 = ($46|0)!=(5);
    if ($47) {
     $48 = $peer;
     $49 = ((($48)) + 36|0);
     $50 = HEAP32[$49>>2]|0;
     $51 = ($50|0)!=(6);
     if (!($51)) {
      label = 8;
     }
    } else {
     label = 8;
    }
    if ((label|0) == 8) {
     label = 0;
     $52 = $peer;
     $53 = ((($52)) + 68|0);
     $54 = HEAP32[$53>>2]|0;
     $55 = $dataTotal;
     $56 = (($55) + ($54))|0;
     $dataTotal = $56;
    }
    $57 = $peer;
    $58 = ((($57)) + 380|0);
    $peer = $58;
   }
  }
 } while(0);
 L17: while(1) {
  $59 = $peersRemaining;
  $60 = ($59>>>0)>(0);
  $61 = $needsAdjustment;
  $62 = ($61|0)!=(0);
  $63 = $60 ? $62 : 0;
  if (!($63)) {
   break;
  }
  $needsAdjustment = 0;
  $64 = $dataTotal;
  $65 = $bandwidth;
  $66 = ($64>>>0)<=($65>>>0);
  if ($66) {
   $throttle = 32;
  } else {
   $67 = $bandwidth;
   $68 = $67<<5;
   $69 = $dataTotal;
   $70 = (($68>>>0) / ($69>>>0))&-1;
   $throttle = $70;
  }
  $71 = $0;
  $72 = ((($71)) + 36|0);
  $73 = HEAP32[$72>>2]|0;
  $peer = $73;
  while(1) {
   $74 = $peer;
   $75 = $0;
   $76 = ((($75)) + 40|0);
   $77 = HEAP32[$76>>2]|0;
   $78 = $0;
   $79 = ((($78)) + 36|0);
   $80 = HEAP32[$79>>2]|0;
   $81 = (($80) + (($77*380)|0)|0);
   $82 = ($74>>>0)<($81>>>0);
   if (!($82)) {
    continue L17;
   }
   $83 = $peer;
   $84 = ((($83)) + 36|0);
   $85 = HEAP32[$84>>2]|0;
   $86 = ($85|0)!=(5);
   if ($86) {
    $87 = $peer;
    $88 = ((($87)) + 36|0);
    $89 = HEAP32[$88>>2]|0;
    $90 = ($89|0)!=(6);
    if (!($90)) {
     label = 18;
    }
   } else {
    label = 18;
   }
   if ((label|0) == 18) {
    label = 0;
    $91 = $peer;
    $92 = ((($91)) + 48|0);
    $93 = HEAP32[$92>>2]|0;
    $94 = ($93|0)==(0);
    if (!($94)) {
     $95 = $peer;
     $96 = ((($95)) + 60|0);
     $97 = HEAP32[$96>>2]|0;
     $98 = $timeCurrent;
     $99 = ($97|0)==($98|0);
     if (!($99)) {
      $100 = $peer;
      $101 = ((($100)) + 48|0);
      $102 = HEAP32[$101>>2]|0;
      $103 = $elapsedTime;
      $104 = Math_imul($102, $103)|0;
      $105 = (($104>>>0) / 1000)&-1;
      $peerBandwidth = $105;
      $106 = $throttle;
      $107 = $peer;
      $108 = ((($107)) + 68|0);
      $109 = HEAP32[$108>>2]|0;
      $110 = Math_imul($106, $109)|0;
      $111 = (($110>>>0) / 32)&-1;
      $112 = $peerBandwidth;
      $113 = ($111>>>0)<=($112>>>0);
      if (!($113)) {
       $114 = $peerBandwidth;
       $115 = $114<<5;
       $116 = $peer;
       $117 = ((($116)) + 68|0);
       $118 = HEAP32[$117>>2]|0;
       $119 = (($115>>>0) / ($118>>>0))&-1;
       $120 = $peer;
       $121 = ((($120)) + 112|0);
       HEAP32[$121>>2] = $119;
       $122 = $peer;
       $123 = ((($122)) + 112|0);
       $124 = HEAP32[$123>>2]|0;
       $125 = ($124|0)==(0);
       if ($125) {
        $126 = $peer;
        $127 = ((($126)) + 112|0);
        HEAP32[$127>>2] = 1;
       }
       $128 = $peer;
       $129 = ((($128)) + 108|0);
       $130 = HEAP32[$129>>2]|0;
       $131 = $peer;
       $132 = ((($131)) + 112|0);
       $133 = HEAP32[$132>>2]|0;
       $134 = ($130>>>0)>($133>>>0);
       if ($134) {
        $135 = $peer;
        $136 = ((($135)) + 112|0);
        $137 = HEAP32[$136>>2]|0;
        $138 = $peer;
        $139 = ((($138)) + 108|0);
        HEAP32[$139>>2] = $137;
       }
       $140 = $timeCurrent;
       $141 = $peer;
       $142 = ((($141)) + 60|0);
       HEAP32[$142>>2] = $140;
       $143 = $peer;
       $144 = ((($143)) + 64|0);
       HEAP32[$144>>2] = 0;
       $145 = $peer;
       $146 = ((($145)) + 68|0);
       HEAP32[$146>>2] = 0;
       $needsAdjustment = 1;
       $147 = $peersRemaining;
       $148 = (($147) + -1)|0;
       $peersRemaining = $148;
       $149 = $peerBandwidth;
       $150 = $bandwidth;
       $151 = (($150) - ($149))|0;
       $bandwidth = $151;
       $152 = $peerBandwidth;
       $153 = $dataTotal;
       $154 = (($153) - ($152))|0;
       $dataTotal = $154;
      }
     }
    }
   }
   $155 = $peer;
   $156 = ((($155)) + 380|0);
   $peer = $156;
  }
 }
 $157 = $peersRemaining;
 $158 = ($157>>>0)>(0);
 L42: do {
  if ($158) {
   $159 = $dataTotal;
   $160 = $bandwidth;
   $161 = ($159>>>0)<=($160>>>0);
   if ($161) {
    $throttle = 32;
   } else {
    $162 = $bandwidth;
    $163 = $162<<5;
    $164 = $dataTotal;
    $165 = (($163>>>0) / ($164>>>0))&-1;
    $throttle = $165;
   }
   $166 = $0;
   $167 = ((($166)) + 36|0);
   $168 = HEAP32[$167>>2]|0;
   $peer = $168;
   while(1) {
    $169 = $peer;
    $170 = $0;
    $171 = ((($170)) + 40|0);
    $172 = HEAP32[$171>>2]|0;
    $173 = $0;
    $174 = ((($173)) + 36|0);
    $175 = HEAP32[$174>>2]|0;
    $176 = (($175) + (($172*380)|0)|0);
    $177 = ($169>>>0)<($176>>>0);
    if (!($177)) {
     break L42;
    }
    $178 = $peer;
    $179 = ((($178)) + 36|0);
    $180 = HEAP32[$179>>2]|0;
    $181 = ($180|0)!=(5);
    if ($181) {
     $182 = $peer;
     $183 = ((($182)) + 36|0);
     $184 = HEAP32[$183>>2]|0;
     $185 = ($184|0)!=(6);
     if (!($185)) {
      label = 35;
     }
    } else {
     label = 35;
    }
    if ((label|0) == 35) {
     label = 0;
     $186 = $peer;
     $187 = ((($186)) + 60|0);
     $188 = HEAP32[$187>>2]|0;
     $189 = $timeCurrent;
     $190 = ($188|0)==($189|0);
     if (!($190)) {
      $191 = $throttle;
      $192 = $peer;
      $193 = ((($192)) + 112|0);
      HEAP32[$193>>2] = $191;
      $194 = $peer;
      $195 = ((($194)) + 108|0);
      $196 = HEAP32[$195>>2]|0;
      $197 = $peer;
      $198 = ((($197)) + 112|0);
      $199 = HEAP32[$198>>2]|0;
      $200 = ($196>>>0)>($199>>>0);
      if ($200) {
       $201 = $peer;
       $202 = ((($201)) + 112|0);
       $203 = HEAP32[$202>>2]|0;
       $204 = $peer;
       $205 = ((($204)) + 108|0);
       HEAP32[$205>>2] = $203;
      }
      $206 = $peer;
      $207 = ((($206)) + 64|0);
      HEAP32[$207>>2] = 0;
      $208 = $peer;
      $209 = ((($208)) + 68|0);
      HEAP32[$209>>2] = 0;
     }
    }
    $210 = $peer;
    $211 = ((($210)) + 380|0);
    $peer = $211;
   }
  }
 } while(0);
 $212 = $0;
 $213 = ((($212)) + 32|0);
 $214 = HEAP32[$213>>2]|0;
 $215 = ($214|0)!=(0);
 if (!($215)) {
  STACKTOP = sp;return;
 }
 $216 = $0;
 $217 = ((($216)) + 32|0);
 HEAP32[$217>>2] = 0;
 $218 = $0;
 $219 = ((($218)) + 10384|0);
 $220 = HEAP32[$219>>2]|0;
 $peersRemaining = $220;
 $221 = $0;
 $222 = ((($221)) + 12|0);
 $223 = HEAP32[$222>>2]|0;
 $bandwidth = $223;
 $needsAdjustment = 1;
 $224 = $bandwidth;
 $225 = ($224|0)==(0);
 L64: do {
  if ($225) {
   $bandwidthLimit = 0;
  } else {
   L66: while(1) {
    $226 = $peersRemaining;
    $227 = ($226>>>0)>(0);
    $228 = $needsAdjustment;
    $229 = ($228|0)!=(0);
    $230 = $227 ? $229 : 0;
    if (!($230)) {
     break L64;
    }
    $needsAdjustment = 0;
    $231 = $bandwidth;
    $232 = $peersRemaining;
    $233 = (($231>>>0) / ($232>>>0))&-1;
    $bandwidthLimit = $233;
    $234 = $0;
    $235 = ((($234)) + 36|0);
    $236 = HEAP32[$235>>2]|0;
    $peer = $236;
    while(1) {
     $237 = $peer;
     $238 = $0;
     $239 = ((($238)) + 40|0);
     $240 = HEAP32[$239>>2]|0;
     $241 = $0;
     $242 = ((($241)) + 36|0);
     $243 = HEAP32[$242>>2]|0;
     $244 = (($243) + (($240*380)|0)|0);
     $245 = ($237>>>0)<($244>>>0);
     if (!($245)) {
      continue L66;
     }
     $246 = $peer;
     $247 = ((($246)) + 36|0);
     $248 = HEAP32[$247>>2]|0;
     $249 = ($248|0)!=(5);
     if ($249) {
      $250 = $peer;
      $251 = ((($250)) + 36|0);
      $252 = HEAP32[$251>>2]|0;
      $253 = ($252|0)!=(6);
      if (!($253)) {
       label = 48;
      }
     } else {
      label = 48;
     }
     do {
      if ((label|0) == 48) {
       label = 0;
       $254 = $peer;
       $255 = ((($254)) + 56|0);
       $256 = HEAP32[$255>>2]|0;
       $257 = $timeCurrent;
       $258 = ($256|0)==($257|0);
       if (!($258)) {
        $259 = $peer;
        $260 = ((($259)) + 52|0);
        $261 = HEAP32[$260>>2]|0;
        $262 = ($261>>>0)>(0);
        if ($262) {
         $263 = $peer;
         $264 = ((($263)) + 52|0);
         $265 = HEAP32[$264>>2]|0;
         $266 = $bandwidthLimit;
         $267 = ($265>>>0)>=($266>>>0);
         if ($267) {
          break;
         }
        }
        $268 = $timeCurrent;
        $269 = $peer;
        $270 = ((($269)) + 56|0);
        HEAP32[$270>>2] = $268;
        $needsAdjustment = 1;
        $271 = $peersRemaining;
        $272 = (($271) + -1)|0;
        $peersRemaining = $272;
        $273 = $peer;
        $274 = ((($273)) + 52|0);
        $275 = HEAP32[$274>>2]|0;
        $276 = $bandwidth;
        $277 = (($276) - ($275))|0;
        $bandwidth = $277;
       }
      }
     } while(0);
     $278 = $peer;
     $279 = ((($278)) + 380|0);
     $peer = $279;
    }
   }
  }
 } while(0);
 $280 = $0;
 $281 = ((($280)) + 36|0);
 $282 = HEAP32[$281>>2]|0;
 $peer = $282;
 while(1) {
  $283 = $peer;
  $284 = $0;
  $285 = ((($284)) + 40|0);
  $286 = HEAP32[$285>>2]|0;
  $287 = $0;
  $288 = ((($287)) + 36|0);
  $289 = HEAP32[$288>>2]|0;
  $290 = (($289) + (($286*380)|0)|0);
  $291 = ($283>>>0)<($290>>>0);
  if (!($291)) {
   break;
  }
  $292 = $peer;
  $293 = ((($292)) + 36|0);
  $294 = HEAP32[$293>>2]|0;
  $295 = ($294|0)!=(5);
  if ($295) {
   $296 = $peer;
   $297 = ((($296)) + 36|0);
   $298 = HEAP32[$297>>2]|0;
   $299 = ($298|0)!=(6);
   if (!($299)) {
    label = 57;
   }
  } else {
   label = 57;
  }
  if ((label|0) == 57) {
   label = 0;
   HEAP8[$command>>0] = -118;
   $300 = ((($command)) + 1|0);
   HEAP8[$300>>0] = -1;
   $301 = $0;
   $302 = ((($301)) + 16|0);
   $303 = HEAP32[$302>>2]|0;
   $304 = (_htonl($303)|0);
   $305 = ((($command)) + 8|0);
   HEAP8[$305>>0]=$304&255;HEAP8[$305+1>>0]=($304>>8)&255;HEAP8[$305+2>>0]=($304>>16)&255;HEAP8[$305+3>>0]=$304>>24;
   $306 = $peer;
   $307 = ((($306)) + 56|0);
   $308 = HEAP32[$307>>2]|0;
   $309 = $timeCurrent;
   $310 = ($308|0)==($309|0);
   if ($310) {
    $311 = $peer;
    $312 = ((($311)) + 52|0);
    $313 = HEAP32[$312>>2]|0;
    $314 = (_htonl($313)|0);
    $315 = ((($command)) + 4|0);
    HEAP8[$315>>0]=$314&255;HEAP8[$315+1>>0]=($314>>8)&255;HEAP8[$315+2>>0]=($314>>16)&255;HEAP8[$315+3>>0]=$314>>24;
   } else {
    $316 = $bandwidthLimit;
    $317 = (_htonl($316)|0);
    $318 = ((($command)) + 4|0);
    HEAP8[$318>>0]=$317&255;HEAP8[$318+1>>0]=($317>>8)&255;HEAP8[$318+2>>0]=($317>>16)&255;HEAP8[$318+3>>0]=$317>>24;
   }
   $319 = $peer;
   (_enet_peer_queue_outgoing_command($319,$command,0,0,0)|0);
  }
  $320 = $peer;
  $321 = ((($320)) + 380|0);
  $peer = $321;
 }
 STACKTOP = sp;return;
}
function _enet_list_clear($list) {
 $list = $list|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $list;
 $1 = $0;
 $2 = $0;
 HEAP32[$2>>2] = $1;
 $3 = $0;
 $4 = $0;
 $5 = ((($4)) + 4|0);
 HEAP32[$5>>2] = $3;
 STACKTOP = sp;return;
}
function _enet_list_insert($position,$data) {
 $position = $position|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $result = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $position;
 $1 = $data;
 $2 = $1;
 $result = $2;
 $3 = $0;
 $4 = ((($3)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $result;
 $7 = ((($6)) + 4|0);
 HEAP32[$7>>2] = $5;
 $8 = $0;
 $9 = $result;
 HEAP32[$9>>2] = $8;
 $10 = $result;
 $11 = $result;
 $12 = ((($11)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 HEAP32[$13>>2] = $10;
 $14 = $result;
 $15 = $0;
 $16 = ((($15)) + 4|0);
 HEAP32[$16>>2] = $14;
 $17 = $result;
 STACKTOP = sp;return ($17|0);
}
function _enet_list_remove($position) {
 $position = $position|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $position;
 $1 = $0;
 $2 = HEAP32[$1>>2]|0;
 $3 = $0;
 $4 = ((($3)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 HEAP32[$5>>2] = $2;
 $6 = $0;
 $7 = ((($6)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $0;
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + 4|0);
 HEAP32[$11>>2] = $8;
 $12 = $0;
 STACKTOP = sp;return ($12|0);
}
function _enet_list_move($position,$dataFirst,$dataLast) {
 $position = $position|0;
 $dataFirst = $dataFirst|0;
 $dataLast = $dataLast|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $first = 0, $last = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $position;
 $1 = $dataFirst;
 $2 = $dataLast;
 $3 = $1;
 $first = $3;
 $4 = $2;
 $last = $4;
 $5 = $last;
 $6 = HEAP32[$5>>2]|0;
 $7 = $first;
 $8 = ((($7)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 HEAP32[$9>>2] = $6;
 $10 = $first;
 $11 = ((($10)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = $last;
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($14)) + 4|0);
 HEAP32[$15>>2] = $12;
 $16 = $0;
 $17 = ((($16)) + 4|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = $first;
 $20 = ((($19)) + 4|0);
 HEAP32[$20>>2] = $18;
 $21 = $0;
 $22 = $last;
 HEAP32[$22>>2] = $21;
 $23 = $first;
 $24 = $first;
 $25 = ((($24)) + 4|0);
 $26 = HEAP32[$25>>2]|0;
 HEAP32[$26>>2] = $23;
 $27 = $last;
 $28 = $0;
 $29 = ((($28)) + 4|0);
 HEAP32[$29>>2] = $27;
 $30 = $first;
 STACKTOP = sp;return ($30|0);
}
function _enet_packet_create($data,$dataLength,$flags) {
 $data = $data|0;
 $dataLength = $dataLength|0;
 $flags = $flags|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $packet = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $data;
 $2 = $dataLength;
 $3 = $flags;
 $4 = (_enet_malloc(24)|0);
 $packet = $4;
 $5 = $packet;
 $6 = ($5|0)==(0|0);
 if ($6) {
  $0 = 0;
  $45 = $0;
  STACKTOP = sp;return ($45|0);
 }
 $7 = $3;
 $8 = $7 & 4;
 $9 = ($8|0)!=(0);
 do {
  if ($9) {
   $10 = $1;
   $11 = $packet;
   $12 = ((($11)) + 8|0);
   HEAP32[$12>>2] = $10;
  } else {
   $13 = $2;
   $14 = ($13>>>0)<=(0);
   if ($14) {
    $15 = $packet;
    $16 = ((($15)) + 8|0);
    HEAP32[$16>>2] = 0;
    break;
   }
   $17 = $2;
   $18 = (_enet_malloc($17)|0);
   $19 = $packet;
   $20 = ((($19)) + 8|0);
   HEAP32[$20>>2] = $18;
   $21 = $packet;
   $22 = ((($21)) + 8|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = ($23|0)==(0|0);
   if ($24) {
    $25 = $packet;
    _enet_free($25);
    $0 = 0;
    $45 = $0;
    STACKTOP = sp;return ($45|0);
   } else {
    $26 = $1;
    $27 = ($26|0)!=(0|0);
    if (!($27)) {
     break;
    }
    $28 = $packet;
    $29 = ((($28)) + 8|0);
    $30 = HEAP32[$29>>2]|0;
    $31 = $1;
    $32 = $2;
    _memcpy(($30|0),($31|0),($32|0))|0;
    break;
   }
  }
 } while(0);
 $33 = $packet;
 HEAP32[$33>>2] = 0;
 $34 = $3;
 $35 = $packet;
 $36 = ((($35)) + 4|0);
 HEAP32[$36>>2] = $34;
 $37 = $2;
 $38 = $packet;
 $39 = ((($38)) + 12|0);
 HEAP32[$39>>2] = $37;
 $40 = $packet;
 $41 = ((($40)) + 16|0);
 HEAP32[$41>>2] = 0;
 $42 = $packet;
 $43 = ((($42)) + 20|0);
 HEAP32[$43>>2] = 0;
 $44 = $packet;
 $0 = $44;
 $45 = $0;
 STACKTOP = sp;return ($45|0);
}
function _enet_packet_destroy($packet) {
 $packet = $packet|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $packet;
 $1 = $0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  STACKTOP = sp;return;
 }
 $3 = $0;
 $4 = ((($3)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)!=(0|0);
 if ($6) {
  $7 = $0;
  $8 = ((($7)) + 16|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = $0;
  FUNCTION_TABLE_vi[$9 & 8191]($10);
 }
 $11 = $0;
 $12 = ((($11)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = $13 & 4;
 $15 = ($14|0)!=(0);
 if (!($15)) {
  $16 = $0;
  $17 = ((($16)) + 8|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = ($18|0)!=(0|0);
  if ($19) {
   $20 = $0;
   $21 = ((($20)) + 8|0);
   $22 = HEAP32[$21>>2]|0;
   _enet_free($22);
  }
 }
 $23 = $0;
 _enet_free($23);
 STACKTOP = sp;return;
}
function _enet_peer_throttle_configure($peer,$interval,$acceleration,$deceleration) {
 $peer = $peer|0;
 $interval = $interval|0;
 $acceleration = $acceleration|0;
 $deceleration = $deceleration|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $command = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $command = sp + 16|0;
 $0 = $peer;
 $1 = $interval;
 $2 = $acceleration;
 $3 = $deceleration;
 $4 = $1;
 $5 = $0;
 $6 = ((($5)) + 132|0);
 HEAP32[$6>>2] = $4;
 $7 = $2;
 $8 = $0;
 $9 = ((($8)) + 124|0);
 HEAP32[$9>>2] = $7;
 $10 = $3;
 $11 = $0;
 $12 = ((($11)) + 128|0);
 HEAP32[$12>>2] = $10;
 HEAP8[$command>>0] = -117;
 $13 = ((($command)) + 1|0);
 HEAP8[$13>>0] = -1;
 $14 = $1;
 $15 = (_htonl($14)|0);
 $16 = ((($command)) + 4|0);
 HEAP8[$16>>0]=$15&255;HEAP8[$16+1>>0]=($15>>8)&255;HEAP8[$16+2>>0]=($15>>16)&255;HEAP8[$16+3>>0]=$15>>24;
 $17 = $2;
 $18 = (_htonl($17)|0);
 $19 = ((($command)) + 8|0);
 HEAP8[$19>>0]=$18&255;HEAP8[$19+1>>0]=($18>>8)&255;HEAP8[$19+2>>0]=($18>>16)&255;HEAP8[$19+3>>0]=$18>>24;
 $20 = $3;
 $21 = (_htonl($20)|0);
 $22 = ((($command)) + 12|0);
 HEAP8[$22>>0]=$21&255;HEAP8[$22+1>>0]=($21>>8)&255;HEAP8[$22+2>>0]=($21>>16)&255;HEAP8[$22+3>>0]=$21>>24;
 $23 = $0;
 (_enet_peer_queue_outgoing_command($23,$command,0,0,0)|0);
 STACKTOP = sp;return;
}
function _enet_peer_queue_outgoing_command($peer,$command,$packet,$offset,$length) {
 $peer = $peer|0;
 $command = $command|0;
 $packet = $packet|0;
 $offset = $offset|0;
 $length = $length|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $outgoingCommand = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $peer;
 $2 = $command;
 $3 = $packet;
 $4 = $offset;
 $5 = $length;
 $6 = (_enet_malloc(84)|0);
 $outgoingCommand = $6;
 $7 = $outgoingCommand;
 $8 = ($7|0)==(0|0);
 if ($8) {
  $0 = 0;
  $29 = $0;
  STACKTOP = sp;return ($29|0);
 }
 $9 = $outgoingCommand;
 $10 = ((($9)) + 32|0);
 $11 = $2;
 dest=$10; src=$11; stop=dest+48|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
 $12 = $4;
 $13 = $outgoingCommand;
 $14 = ((($13)) + 24|0);
 HEAP32[$14>>2] = $12;
 $15 = $5;
 $16 = $outgoingCommand;
 $17 = ((($16)) + 28|0);
 HEAP16[$17>>1] = $15;
 $18 = $3;
 $19 = $outgoingCommand;
 $20 = ((($19)) + 80|0);
 HEAP32[$20>>2] = $18;
 $21 = $3;
 $22 = ($21|0)!=(0|0);
 if ($22) {
  $23 = $3;
  $24 = HEAP32[$23>>2]|0;
  $25 = (($24) + 1)|0;
  HEAP32[$23>>2] = $25;
 }
 $26 = $1;
 $27 = $outgoingCommand;
 _enet_peer_setup_outgoing_command($26,$27);
 $28 = $outgoingCommand;
 $0 = $28;
 $29 = $0;
 STACKTOP = sp;return ($29|0);
}
function _enet_peer_setup_outgoing_command($peer,$outgoingCommand) {
 $peer = $peer|0;
 $outgoingCommand = $outgoingCommand|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $channel = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $peer;
 $1 = $outgoingCommand;
 $2 = $1;
 $3 = ((($2)) + 32|0);
 $4 = ((($3)) + 1|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = $0;
 $8 = ((($7)) + 40|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (($9) + (($6*60)|0)|0);
 $channel = $10;
 $11 = $1;
 $12 = ((($11)) + 32|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = (_enet_protocol_command_size($13)|0);
 $15 = $1;
 $16 = ((($15)) + 28|0);
 $17 = HEAP16[$16>>1]|0;
 $18 = $17&65535;
 $19 = (($14) + ($18))|0;
 $20 = $0;
 $21 = ((($20)) + 68|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = (($22) + ($19))|0;
 HEAP32[$21>>2] = $23;
 $24 = $1;
 $25 = ((($24)) + 32|0);
 $26 = ((($25)) + 1|0);
 $27 = HEAP8[$26>>0]|0;
 $28 = $27&255;
 $29 = ($28|0)==(255);
 do {
  if ($29) {
   $30 = $0;
   $31 = ((($30)) + 188|0);
   $32 = HEAP16[$31>>1]|0;
   $33 = (($32) + 1)<<16>>16;
   HEAP16[$31>>1] = $33;
   $34 = $0;
   $35 = ((($34)) + 188|0);
   $36 = HEAP16[$35>>1]|0;
   $37 = $1;
   $38 = ((($37)) + 8|0);
   HEAP16[$38>>1] = $36;
   $39 = $1;
   $40 = ((($39)) + 10|0);
   HEAP16[$40>>1] = 0;
  } else {
   $41 = $1;
   $42 = ((($41)) + 32|0);
   $43 = HEAP8[$42>>0]|0;
   $44 = $43&255;
   $45 = $44 & 128;
   $46 = ($45|0)!=(0);
   if ($46) {
    $47 = $channel;
    $48 = HEAP16[$47>>1]|0;
    $49 = (($48) + 1)<<16>>16;
    HEAP16[$47>>1] = $49;
    $50 = $channel;
    $51 = ((($50)) + 2|0);
    HEAP16[$51>>1] = 0;
    $52 = $channel;
    $53 = HEAP16[$52>>1]|0;
    $54 = $1;
    $55 = ((($54)) + 8|0);
    HEAP16[$55>>1] = $53;
    $56 = $1;
    $57 = ((($56)) + 10|0);
    HEAP16[$57>>1] = 0;
    break;
   }
   $58 = $1;
   $59 = ((($58)) + 32|0);
   $60 = HEAP8[$59>>0]|0;
   $61 = $60&255;
   $62 = $61 & 64;
   $63 = ($62|0)!=(0);
   if ($63) {
    $64 = $0;
    $65 = ((($64)) + 246|0);
    $66 = HEAP16[$65>>1]|0;
    $67 = (($66) + 1)<<16>>16;
    HEAP16[$65>>1] = $67;
    $68 = $1;
    $69 = ((($68)) + 8|0);
    HEAP16[$69>>1] = 0;
    $70 = $1;
    $71 = ((($70)) + 10|0);
    HEAP16[$71>>1] = 0;
    break;
   }
   $72 = $1;
   $73 = ((($72)) + 24|0);
   $74 = HEAP32[$73>>2]|0;
   $75 = ($74|0)==(0);
   if ($75) {
    $76 = $channel;
    $77 = ((($76)) + 2|0);
    $78 = HEAP16[$77>>1]|0;
    $79 = (($78) + 1)<<16>>16;
    HEAP16[$77>>1] = $79;
   }
   $80 = $channel;
   $81 = HEAP16[$80>>1]|0;
   $82 = $1;
   $83 = ((($82)) + 8|0);
   HEAP16[$83>>1] = $81;
   $84 = $channel;
   $85 = ((($84)) + 2|0);
   $86 = HEAP16[$85>>1]|0;
   $87 = $1;
   $88 = ((($87)) + 10|0);
   HEAP16[$88>>1] = $86;
  }
 } while(0);
 $89 = $1;
 $90 = ((($89)) + 30|0);
 HEAP16[$90>>1] = 0;
 $91 = $1;
 $92 = ((($91)) + 12|0);
 HEAP32[$92>>2] = 0;
 $93 = $1;
 $94 = ((($93)) + 16|0);
 HEAP32[$94>>2] = 0;
 $95 = $1;
 $96 = ((($95)) + 20|0);
 HEAP32[$96>>2] = 0;
 $97 = $1;
 $98 = ((($97)) + 8|0);
 $99 = HEAP16[$98>>1]|0;
 $100 = (_htons($99)|0);
 $101 = $1;
 $102 = ((($101)) + 32|0);
 $103 = ((($102)) + 2|0);
 HEAP16[$103>>1] = $100;
 $104 = $1;
 $105 = ((($104)) + 32|0);
 $106 = HEAP8[$105>>0]|0;
 $107 = $106&255;
 $108 = $107 & 15;
 switch ($108|0) {
 case 7:  {
  $109 = $1;
  $110 = ((($109)) + 10|0);
  $111 = HEAP16[$110>>1]|0;
  $112 = (_htons($111)|0);
  $113 = $1;
  $114 = ((($113)) + 32|0);
  $115 = ((($114)) + 4|0);
  HEAP16[$115>>1] = $112;
  break;
 }
 case 9:  {
  $116 = $0;
  $117 = ((($116)) + 246|0);
  $118 = HEAP16[$117>>1]|0;
  $119 = (_htons($118)|0);
  $120 = $1;
  $121 = ((($120)) + 32|0);
  $122 = ((($121)) + 4|0);
  HEAP16[$122>>1] = $119;
  break;
 }
 default: {
 }
 }
 $123 = $1;
 $124 = ((($123)) + 32|0);
 $125 = HEAP8[$124>>0]|0;
 $126 = $125&255;
 $127 = $126 & 128;
 $128 = ($127|0)!=(0);
 $129 = $0;
 if ($128) {
  $130 = ((($129)) + 216|0);
  $131 = $1;
  (_enet_list_insert($130,$131)|0);
  STACKTOP = sp;return;
 } else {
  $132 = ((($129)) + 224|0);
  $133 = $1;
  (_enet_list_insert($132,$133)|0);
  STACKTOP = sp;return;
 }
}
function _enet_peer_throttle($peer,$rtt) {
 $peer = $peer|0;
 $rtt = $rtt|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $peer;
 $2 = $rtt;
 $3 = $1;
 $4 = ((($3)) + 152|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $1;
 $7 = ((($6)) + 160|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($5>>>0)<=($8>>>0);
 if ($9) {
  $10 = $1;
  $11 = ((($10)) + 112|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = $1;
  $14 = ((($13)) + 108|0);
  HEAP32[$14>>2] = $12;
 } else {
  $15 = $2;
  $16 = $1;
  $17 = ((($16)) + 152|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = ($15>>>0)<($18>>>0);
  if ($19) {
   $20 = $1;
   $21 = ((($20)) + 124|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = $1;
   $24 = ((($23)) + 108|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = (($25) + ($22))|0;
   HEAP32[$24>>2] = $26;
   $27 = $1;
   $28 = ((($27)) + 108|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = $1;
   $31 = ((($30)) + 112|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = ($29>>>0)>($32>>>0);
   if ($33) {
    $34 = $1;
    $35 = ((($34)) + 112|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = $1;
    $38 = ((($37)) + 108|0);
    HEAP32[$38>>2] = $36;
   }
   $0 = 1;
   $64 = $0;
   STACKTOP = sp;return ($64|0);
  }
  $39 = $2;
  $40 = $1;
  $41 = ((($40)) + 152|0);
  $42 = HEAP32[$41>>2]|0;
  $43 = $1;
  $44 = ((($43)) + 160|0);
  $45 = HEAP32[$44>>2]|0;
  $46 = $45<<1;
  $47 = (($42) + ($46))|0;
  $48 = ($39>>>0)>($47>>>0);
  if ($48) {
   $49 = $1;
   $50 = ((($49)) + 108|0);
   $51 = HEAP32[$50>>2]|0;
   $52 = $1;
   $53 = ((($52)) + 128|0);
   $54 = HEAP32[$53>>2]|0;
   $55 = ($51>>>0)>($54>>>0);
   $56 = $1;
   if ($55) {
    $57 = ((($56)) + 128|0);
    $58 = HEAP32[$57>>2]|0;
    $59 = $1;
    $60 = ((($59)) + 108|0);
    $61 = HEAP32[$60>>2]|0;
    $62 = (($61) - ($58))|0;
    HEAP32[$60>>2] = $62;
   } else {
    $63 = ((($56)) + 108|0);
    HEAP32[$63>>2] = 0;
   }
   $0 = -1;
   $64 = $0;
   STACKTOP = sp;return ($64|0);
  }
 }
 $0 = 0;
 $64 = $0;
 STACKTOP = sp;return ($64|0);
}
function _enet_peer_send($peer,$channelID,$packet) {
 $peer = $peer|0;
 $channelID = $channelID|0;
 $packet = $packet|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $channel = 0, $command = 0, $commandNumber = 0, $fragment = 0, $fragmentCount = 0, $fragmentLength = 0, $fragmentNumber = 0, $fragmentOffset = 0, $fragments = 0;
 var $startSequenceNumber = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $command = sp + 56|0;
 $fragments = sp + 8|0;
 $1 = $peer;
 $2 = $channelID;
 $3 = $packet;
 $4 = $2;
 $5 = $4&255;
 $6 = $1;
 $7 = ((($6)) + 40|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (($8) + (($5*60)|0)|0);
 $channel = $9;
 $10 = $1;
 $11 = ((($10)) + 36|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)!=(5);
 if (!($13)) {
  $14 = $2;
  $15 = $14&255;
  $16 = $1;
  $17 = ((($16)) + 44|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = ($15>>>0)>=($18>>>0);
  if (!($19)) {
   $20 = $3;
   $21 = ((($20)) + 12|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = ($22>>>0)>(1073741824);
   if (!($23)) {
    $24 = $1;
    $25 = ((($24)) + 176|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = (($26) - 4)|0;
    $28 = (($27) - 24)|0;
    $fragmentLength = $28;
    $29 = $1;
    $30 = ((($29)) + 8|0);
    $31 = HEAP32[$30>>2]|0;
    $32 = ((($31)) + 2136|0);
    $33 = HEAP32[$32>>2]|0;
    $34 = ($33|0)!=(0|0);
    if ($34) {
     $35 = $fragmentLength;
     $36 = (($35) - 4)|0;
     $fragmentLength = $36;
    }
    $37 = $3;
    $38 = ((($37)) + 12|0);
    $39 = HEAP32[$38>>2]|0;
    $40 = $fragmentLength;
    $41 = ($39>>>0)>($40>>>0);
    if (!($41)) {
     $167 = $2;
     $168 = ((($command)) + 1|0);
     HEAP8[$168>>0] = $167;
     $169 = $3;
     $170 = ((($169)) + 4|0);
     $171 = HEAP32[$170>>2]|0;
     $172 = $171 & 3;
     $173 = ($172|0)==(2);
     do {
      if ($173) {
       HEAP8[$command>>0] = 73;
       $174 = $3;
       $175 = ((($174)) + 12|0);
       $176 = HEAP32[$175>>2]|0;
       $177 = $176&65535;
       $178 = (_htons($177)|0);
       $179 = ((($command)) + 6|0);
       HEAP8[$179>>0]=$178&255;HEAP8[$179+1>>0]=$178>>8;
      } else {
       $180 = $3;
       $181 = ((($180)) + 4|0);
       $182 = HEAP32[$181>>2]|0;
       $183 = $182 & 1;
       $184 = ($183|0)!=(0);
       if (!($184)) {
        $185 = $channel;
        $186 = ((($185)) + 2|0);
        $187 = HEAP16[$186>>1]|0;
        $188 = $187&65535;
        $189 = ($188|0)>=(65535);
        if (!($189)) {
         HEAP8[$command>>0] = 7;
         $196 = $3;
         $197 = ((($196)) + 12|0);
         $198 = HEAP32[$197>>2]|0;
         $199 = $198&65535;
         $200 = (_htons($199)|0);
         $201 = ((($command)) + 6|0);
         HEAP8[$201>>0]=$200&255;HEAP8[$201+1>>0]=$200>>8;
         break;
        }
       }
       HEAP8[$command>>0] = -122;
       $190 = $3;
       $191 = ((($190)) + 12|0);
       $192 = HEAP32[$191>>2]|0;
       $193 = $192&65535;
       $194 = (_htons($193)|0);
       $195 = ((($command)) + 4|0);
       HEAP8[$195>>0]=$194&255;HEAP8[$195+1>>0]=$194>>8;
      }
     } while(0);
     $202 = $1;
     $203 = $3;
     $204 = $3;
     $205 = ((($204)) + 12|0);
     $206 = HEAP32[$205>>2]|0;
     $207 = $206&65535;
     $208 = (_enet_peer_queue_outgoing_command($202,$command,$203,0,$207)|0);
     $209 = ($208|0)==(0|0);
     if ($209) {
      $0 = -1;
      $210 = $0;
      STACKTOP = sp;return ($210|0);
     } else {
      $0 = 0;
      $210 = $0;
      STACKTOP = sp;return ($210|0);
     }
    }
    $42 = $3;
    $43 = ((($42)) + 12|0);
    $44 = HEAP32[$43>>2]|0;
    $45 = $fragmentLength;
    $46 = (($44) + ($45))|0;
    $47 = (($46) - 1)|0;
    $48 = $fragmentLength;
    $49 = (($47>>>0) / ($48>>>0))&-1;
    $fragmentCount = $49;
    $50 = $fragmentCount;
    $51 = ($50>>>0)>(1048576);
    if ($51) {
     $0 = -1;
     $210 = $0;
     STACKTOP = sp;return ($210|0);
    }
    $52 = $3;
    $53 = ((($52)) + 4|0);
    $54 = HEAP32[$53>>2]|0;
    $55 = $54 & 9;
    $56 = ($55|0)==(8);
    if ($56) {
     $57 = $channel;
     $58 = ((($57)) + 2|0);
     $59 = HEAP16[$58>>1]|0;
     $60 = $59&65535;
     $61 = ($60|0)<(65535);
     if ($61) {
      $commandNumber = 12;
      $62 = $channel;
      $63 = ((($62)) + 2|0);
      $64 = HEAP16[$63>>1]|0;
      $65 = $64&65535;
      $66 = (($65) + 1)|0;
      $67 = $66&65535;
      $68 = (_htons($67)|0);
      $startSequenceNumber = $68;
     } else {
      label = 13;
     }
    } else {
     label = 13;
    }
    if ((label|0) == 13) {
     $commandNumber = -120;
     $69 = $channel;
     $70 = HEAP16[$69>>1]|0;
     $71 = $70&65535;
     $72 = (($71) + 1)|0;
     $73 = $72&65535;
     $74 = (_htons($73)|0);
     $startSequenceNumber = $74;
    }
    _enet_list_clear($fragments);
    $fragmentNumber = 0;
    $fragmentOffset = 0;
    while(1) {
     $75 = $fragmentOffset;
     $76 = $3;
     $77 = ((($76)) + 12|0);
     $78 = HEAP32[$77>>2]|0;
     $79 = ($75>>>0)<($78>>>0);
     if (!($79)) {
      break;
     }
     $80 = $3;
     $81 = ((($80)) + 12|0);
     $82 = HEAP32[$81>>2]|0;
     $83 = $fragmentOffset;
     $84 = (($82) - ($83))|0;
     $85 = $fragmentLength;
     $86 = ($84>>>0)<($85>>>0);
     if ($86) {
      $87 = $3;
      $88 = ((($87)) + 12|0);
      $89 = HEAP32[$88>>2]|0;
      $90 = $fragmentOffset;
      $91 = (($89) - ($90))|0;
      $fragmentLength = $91;
     }
     $92 = (_enet_malloc(84)|0);
     $fragment = $92;
     $93 = $fragment;
     $94 = ($93|0)==(0|0);
     if ($94) {
      label = 19;
      break;
     }
     $101 = $fragmentOffset;
     $102 = $fragment;
     $103 = ((($102)) + 24|0);
     HEAP32[$103>>2] = $101;
     $104 = $fragmentLength;
     $105 = $104&65535;
     $106 = $fragment;
     $107 = ((($106)) + 28|0);
     HEAP16[$107>>1] = $105;
     $108 = $3;
     $109 = $fragment;
     $110 = ((($109)) + 80|0);
     HEAP32[$110>>2] = $108;
     $111 = $commandNumber;
     $112 = $fragment;
     $113 = ((($112)) + 32|0);
     HEAP8[$113>>0] = $111;
     $114 = $2;
     $115 = $fragment;
     $116 = ((($115)) + 32|0);
     $117 = ((($116)) + 1|0);
     HEAP8[$117>>0] = $114;
     $118 = $startSequenceNumber;
     $119 = $fragment;
     $120 = ((($119)) + 32|0);
     $121 = ((($120)) + 4|0);
     HEAP16[$121>>1] = $118;
     $122 = $fragmentLength;
     $123 = $122&65535;
     $124 = (_htons($123)|0);
     $125 = $fragment;
     $126 = ((($125)) + 32|0);
     $127 = ((($126)) + 6|0);
     HEAP16[$127>>1] = $124;
     $128 = $fragmentCount;
     $129 = (_htonl($128)|0);
     $130 = $fragment;
     $131 = ((($130)) + 32|0);
     $132 = ((($131)) + 8|0);
     HEAP32[$132>>2] = $129;
     $133 = $fragmentNumber;
     $134 = (_htonl($133)|0);
     $135 = $fragment;
     $136 = ((($135)) + 32|0);
     $137 = ((($136)) + 12|0);
     HEAP32[$137>>2] = $134;
     $138 = $3;
     $139 = ((($138)) + 12|0);
     $140 = HEAP32[$139>>2]|0;
     $141 = (_htonl($140)|0);
     $142 = $fragment;
     $143 = ((($142)) + 32|0);
     $144 = ((($143)) + 16|0);
     HEAP32[$144>>2] = $141;
     $145 = $fragmentOffset;
     $146 = (_ntohl($145)|0);
     $147 = $fragment;
     $148 = ((($147)) + 32|0);
     $149 = ((($148)) + 20|0);
     HEAP32[$149>>2] = $146;
     $150 = $fragment;
     (_enet_list_insert($fragments,$150)|0);
     $151 = $fragmentNumber;
     $152 = (($151) + 1)|0;
     $fragmentNumber = $152;
     $153 = $fragmentLength;
     $154 = $fragmentOffset;
     $155 = (($154) + ($153))|0;
     $fragmentOffset = $155;
    }
    if ((label|0) == 19) {
     while(1) {
      label = 0;
      $95 = HEAP32[$fragments>>2]|0;
      $96 = ($95|0)==($fragments|0);
      $97 = $96 ^ 1;
      if (!($97)) {
       break;
      }
      $98 = HEAP32[$fragments>>2]|0;
      $99 = (_enet_list_remove($98)|0);
      $fragment = $99;
      $100 = $fragment;
      _enet_free($100);
      label = 19;
     }
     $0 = -1;
     $210 = $0;
     STACKTOP = sp;return ($210|0);
    }
    $156 = $fragmentNumber;
    $157 = $3;
    $158 = HEAP32[$157>>2]|0;
    $159 = (($158) + ($156))|0;
    HEAP32[$157>>2] = $159;
    while(1) {
     $160 = HEAP32[$fragments>>2]|0;
     $161 = ($160|0)==($fragments|0);
     $162 = $161 ^ 1;
     if (!($162)) {
      break;
     }
     $163 = HEAP32[$fragments>>2]|0;
     $164 = (_enet_list_remove($163)|0);
     $fragment = $164;
     $165 = $1;
     $166 = $fragment;
     _enet_peer_setup_outgoing_command($165,$166);
    }
    $0 = 0;
    $210 = $0;
    STACKTOP = sp;return ($210|0);
   }
  }
 }
 $0 = -1;
 $210 = $0;
 STACKTOP = sp;return ($210|0);
}
function _enet_peer_receive($peer,$channelID) {
 $peer = $peer|0;
 $channelID = $channelID|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $incomingCommand = 0, $packet = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $peer;
 $2 = $channelID;
 $3 = $1;
 $4 = ((($3)) + 232|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $1;
 $7 = ((($6)) + 232|0);
 $8 = ($5|0)==($7|0);
 if ($8) {
  $0 = 0;
  $35 = $0;
  STACKTOP = sp;return ($35|0);
 }
 $9 = $1;
 $10 = ((($9)) + 232|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (_enet_list_remove($11)|0);
 $incomingCommand = $12;
 $13 = $2;
 $14 = ($13|0)!=(0|0);
 if ($14) {
  $15 = $incomingCommand;
  $16 = ((($15)) + 12|0);
  $17 = ((($16)) + 1|0);
  $18 = HEAP8[$17>>0]|0;
  $19 = $2;
  HEAP8[$19>>0] = $18;
 }
 $20 = $incomingCommand;
 $21 = ((($20)) + 72|0);
 $22 = HEAP32[$21>>2]|0;
 $packet = $22;
 $23 = $packet;
 $24 = HEAP32[$23>>2]|0;
 $25 = (($24) + -1)|0;
 HEAP32[$23>>2] = $25;
 $26 = $incomingCommand;
 $27 = ((($26)) + 68|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ($28|0)!=(0|0);
 if ($29) {
  $30 = $incomingCommand;
  $31 = ((($30)) + 68|0);
  $32 = HEAP32[$31>>2]|0;
  _enet_free($32);
 }
 $33 = $incomingCommand;
 _enet_free($33);
 $34 = $packet;
 $0 = $34;
 $35 = $0;
 STACKTOP = sp;return ($35|0);
}
function _enet_peer_reset_queues($peer) {
 $peer = $peer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $channel = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $peer;
 $1 = $0;
 $2 = ((($1)) + 240|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)!=(0);
 if ($4) {
  $5 = $0;
  (_enet_list_remove($5)|0);
  $6 = $0;
  $7 = ((($6)) + 240|0);
  HEAP32[$7>>2] = 0;
 }
 while(1) {
  $8 = $0;
  $9 = ((($8)) + 192|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = $0;
  $12 = ((($11)) + 192|0);
  $13 = ($10|0)==($12|0);
  $14 = $13 ^ 1;
  $15 = $0;
  if (!($14)) {
   break;
  }
  $16 = ((($15)) + 192|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = (_enet_list_remove($17)|0);
  _enet_free($18);
 }
 $19 = ((($15)) + 200|0);
 _enet_peer_reset_outgoing_commands($19);
 $20 = $0;
 $21 = ((($20)) + 208|0);
 _enet_peer_reset_outgoing_commands($21);
 $22 = $0;
 $23 = ((($22)) + 216|0);
 _enet_peer_reset_outgoing_commands($23);
 $24 = $0;
 $25 = ((($24)) + 224|0);
 _enet_peer_reset_outgoing_commands($25);
 $26 = $0;
 $27 = ((($26)) + 232|0);
 _enet_peer_reset_incoming_commands($27);
 $28 = $0;
 $29 = ((($28)) + 40|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ($30|0)!=(0|0);
 if (!($31)) {
  $57 = $0;
  $58 = ((($57)) + 40|0);
  HEAP32[$58>>2] = 0;
  $59 = $0;
  $60 = ((($59)) + 44|0);
  HEAP32[$60>>2] = 0;
  STACKTOP = sp;return;
 }
 $32 = $0;
 $33 = ((($32)) + 44|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = ($34>>>0)>(0);
 if (!($35)) {
  $57 = $0;
  $58 = ((($57)) + 40|0);
  HEAP32[$58>>2] = 0;
  $59 = $0;
  $60 = ((($59)) + 44|0);
  HEAP32[$60>>2] = 0;
  STACKTOP = sp;return;
 }
 $36 = $0;
 $37 = ((($36)) + 40|0);
 $38 = HEAP32[$37>>2]|0;
 $channel = $38;
 while(1) {
  $39 = $channel;
  $40 = $0;
  $41 = ((($40)) + 44|0);
  $42 = HEAP32[$41>>2]|0;
  $43 = $0;
  $44 = ((($43)) + 40|0);
  $45 = HEAP32[$44>>2]|0;
  $46 = (($45) + (($42*60)|0)|0);
  $47 = ($39>>>0)<($46>>>0);
  if (!($47)) {
   break;
  }
  $48 = $channel;
  $49 = ((($48)) + 44|0);
  _enet_peer_reset_incoming_commands($49);
  $50 = $channel;
  $51 = ((($50)) + 52|0);
  _enet_peer_reset_incoming_commands($51);
  $52 = $channel;
  $53 = ((($52)) + 60|0);
  $channel = $53;
 }
 $54 = $0;
 $55 = ((($54)) + 40|0);
 $56 = HEAP32[$55>>2]|0;
 _enet_free($56);
 $57 = $0;
 $58 = ((($57)) + 40|0);
 HEAP32[$58>>2] = 0;
 $59 = $0;
 $60 = ((($59)) + 44|0);
 HEAP32[$60>>2] = 0;
 STACKTOP = sp;return;
}
function _enet_peer_reset_outgoing_commands($queue) {
 $queue = $queue|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $outgoingCommand = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $queue;
 while(1) {
  $1 = $0;
  $2 = HEAP32[$1>>2]|0;
  $3 = $0;
  $4 = ($2|0)==($3|0);
  $5 = $4 ^ 1;
  if (!($5)) {
   break;
  }
  $6 = $0;
  $7 = HEAP32[$6>>2]|0;
  $8 = (_enet_list_remove($7)|0);
  $outgoingCommand = $8;
  $9 = $outgoingCommand;
  $10 = ((($9)) + 80|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($11|0)!=(0|0);
  if ($12) {
   $13 = $outgoingCommand;
   $14 = ((($13)) + 80|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = HEAP32[$15>>2]|0;
   $17 = (($16) + -1)|0;
   HEAP32[$15>>2] = $17;
   $18 = $outgoingCommand;
   $19 = ((($18)) + 80|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = HEAP32[$20>>2]|0;
   $22 = ($21|0)==(0);
   if ($22) {
    $23 = $outgoingCommand;
    $24 = ((($23)) + 80|0);
    $25 = HEAP32[$24>>2]|0;
    _enet_packet_destroy($25);
   }
  }
  $26 = $outgoingCommand;
  _enet_free($26);
 }
 STACKTOP = sp;return;
}
function _enet_peer_reset_incoming_commands($queue) {
 $queue = $queue|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $queue;
 $1 = $0;
 $2 = $0;
 $3 = HEAP32[$2>>2]|0;
 $4 = $0;
 _enet_peer_remove_incoming_commands($1,$3,$4);
 STACKTOP = sp;return;
}
function _enet_peer_remove_incoming_commands($queue,$startCommand,$endCommand) {
 $queue = $queue|0;
 $startCommand = $startCommand|0;
 $endCommand = $endCommand|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $currentCommand = 0, $incomingCommand = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $queue;
 $1 = $startCommand;
 $2 = $endCommand;
 $3 = $1;
 $currentCommand = $3;
 while(1) {
  $4 = $currentCommand;
  $5 = $2;
  $6 = ($4|0)!=($5|0);
  if (!($6)) {
   break;
  }
  $7 = $currentCommand;
  $incomingCommand = $7;
  $8 = $currentCommand;
  $9 = HEAP32[$8>>2]|0;
  $currentCommand = $9;
  $10 = $incomingCommand;
  (_enet_list_remove($10)|0);
  $11 = $incomingCommand;
  $12 = ((($11)) + 72|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ($13|0)!=(0|0);
  if ($14) {
   $15 = $incomingCommand;
   $16 = ((($15)) + 72|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = HEAP32[$17>>2]|0;
   $19 = (($18) + -1)|0;
   HEAP32[$17>>2] = $19;
   $20 = $incomingCommand;
   $21 = ((($20)) + 72|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = HEAP32[$22>>2]|0;
   $24 = ($23|0)==(0);
   if ($24) {
    $25 = $incomingCommand;
    $26 = ((($25)) + 72|0);
    $27 = HEAP32[$26>>2]|0;
    _enet_packet_destroy($27);
   }
  }
  $28 = $incomingCommand;
  $29 = ((($28)) + 68|0);
  $30 = HEAP32[$29>>2]|0;
  $31 = ($30|0)!=(0|0);
  if ($31) {
   $32 = $incomingCommand;
   $33 = ((($32)) + 68|0);
   $34 = HEAP32[$33>>2]|0;
   _enet_free($34);
  }
  $35 = $incomingCommand;
  _enet_free($35);
 }
 STACKTOP = sp;return;
}
function _enet_peer_on_connect($peer) {
 $peer = $peer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $peer;
 $1 = $0;
 $2 = ((($1)) + 36|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)!=(5);
 if (!($4)) {
  STACKTOP = sp;return;
 }
 $5 = $0;
 $6 = ((($5)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)!=(6);
 if (!($8)) {
  STACKTOP = sp;return;
 }
 $9 = $0;
 $10 = ((($9)) + 48|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)!=(0);
 if ($12) {
  $13 = $0;
  $14 = ((($13)) + 8|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ((($15)) + 10388|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = (($17) + 1)|0;
  HEAP32[$16>>2] = $18;
 }
 $19 = $0;
 $20 = ((($19)) + 8|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ((($21)) + 10384|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = (($23) + 1)|0;
 HEAP32[$22>>2] = $24;
 STACKTOP = sp;return;
}
function _enet_peer_on_disconnect($peer) {
 $peer = $peer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $peer;
 $1 = $0;
 $2 = ((($1)) + 36|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(5);
 if (!($4)) {
  $5 = $0;
  $6 = ((($5)) + 36|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)==(6);
  if (!($8)) {
   STACKTOP = sp;return;
  }
 }
 $9 = $0;
 $10 = ((($9)) + 48|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)!=(0);
 if ($12) {
  $13 = $0;
  $14 = ((($13)) + 8|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ((($15)) + 10388|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = (($17) + -1)|0;
  HEAP32[$16>>2] = $18;
 }
 $19 = $0;
 $20 = ((($19)) + 8|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ((($21)) + 10384|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = (($23) + -1)|0;
 HEAP32[$22>>2] = $24;
 STACKTOP = sp;return;
}
function _enet_peer_reset($peer) {
 $peer = $peer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $peer;
 $1 = $0;
 _enet_peer_on_disconnect($1);
 $2 = $0;
 $3 = ((($2)) + 12|0);
 HEAP16[$3>>1] = 4095;
 $4 = $0;
 $5 = ((($4)) + 16|0);
 HEAP32[$5>>2] = 0;
 $6 = $0;
 $7 = ((($6)) + 36|0);
 HEAP32[$7>>2] = 0;
 $8 = $0;
 $9 = ((($8)) + 48|0);
 HEAP32[$9>>2] = 0;
 $10 = $0;
 $11 = ((($10)) + 52|0);
 HEAP32[$11>>2] = 0;
 $12 = $0;
 $13 = ((($12)) + 56|0);
 HEAP32[$13>>2] = 0;
 $14 = $0;
 $15 = ((($14)) + 60|0);
 HEAP32[$15>>2] = 0;
 $16 = $0;
 $17 = ((($16)) + 64|0);
 HEAP32[$17>>2] = 0;
 $18 = $0;
 $19 = ((($18)) + 68|0);
 HEAP32[$19>>2] = 0;
 $20 = $0;
 $21 = ((($20)) + 72|0);
 HEAP32[$21>>2] = 0;
 $22 = $0;
 $23 = ((($22)) + 76|0);
 HEAP32[$23>>2] = 0;
 $24 = $0;
 $25 = ((($24)) + 80|0);
 HEAP32[$25>>2] = 0;
 $26 = $0;
 $27 = ((($26)) + 84|0);
 HEAP32[$27>>2] = 0;
 $28 = $0;
 $29 = ((($28)) + 88|0);
 HEAP32[$29>>2] = 0;
 $30 = $0;
 $31 = ((($30)) + 92|0);
 HEAP32[$31>>2] = 0;
 $32 = $0;
 $33 = ((($32)) + 96|0);
 HEAP32[$33>>2] = 0;
 $34 = $0;
 $35 = ((($34)) + 100|0);
 HEAP32[$35>>2] = 0;
 $36 = $0;
 $37 = ((($36)) + 104|0);
 HEAP32[$37>>2] = 0;
 $38 = $0;
 $39 = ((($38)) + 108|0);
 HEAP32[$39>>2] = 32;
 $40 = $0;
 $41 = ((($40)) + 112|0);
 HEAP32[$41>>2] = 32;
 $42 = $0;
 $43 = ((($42)) + 116|0);
 HEAP32[$43>>2] = 0;
 $44 = $0;
 $45 = ((($44)) + 120|0);
 HEAP32[$45>>2] = 0;
 $46 = $0;
 $47 = ((($46)) + 124|0);
 HEAP32[$47>>2] = 2;
 $48 = $0;
 $49 = ((($48)) + 128|0);
 HEAP32[$49>>2] = 2;
 $50 = $0;
 $51 = ((($50)) + 132|0);
 HEAP32[$51>>2] = 5000;
 $52 = $0;
 $53 = ((($52)) + 136|0);
 HEAP32[$53>>2] = 500;
 $54 = $0;
 $55 = ((($54)) + 140|0);
 HEAP32[$55>>2] = 32;
 $56 = $0;
 $57 = ((($56)) + 144|0);
 HEAP32[$57>>2] = 5000;
 $58 = $0;
 $59 = ((($58)) + 148|0);
 HEAP32[$59>>2] = 30000;
 $60 = $0;
 $61 = ((($60)) + 152|0);
 HEAP32[$61>>2] = 500;
 $62 = $0;
 $63 = ((($62)) + 156|0);
 HEAP32[$63>>2] = 500;
 $64 = $0;
 $65 = ((($64)) + 160|0);
 HEAP32[$65>>2] = 0;
 $66 = $0;
 $67 = ((($66)) + 164|0);
 HEAP32[$67>>2] = 0;
 $68 = $0;
 $69 = ((($68)) + 168|0);
 HEAP32[$69>>2] = 500;
 $70 = $0;
 $71 = ((($70)) + 172|0);
 HEAP32[$71>>2] = 0;
 $72 = $0;
 $73 = ((($72)) + 8|0);
 $74 = HEAP32[$73>>2]|0;
 $75 = ((($74)) + 24|0);
 $76 = HEAP32[$75>>2]|0;
 $77 = $0;
 $78 = ((($77)) + 176|0);
 HEAP32[$78>>2] = $76;
 $79 = $0;
 $80 = ((($79)) + 184|0);
 HEAP32[$80>>2] = 0;
 $81 = $0;
 $82 = ((($81)) + 188|0);
 HEAP16[$82>>1] = 0;
 $83 = $0;
 $84 = ((($83)) + 180|0);
 HEAP32[$84>>2] = 65536;
 $85 = $0;
 $86 = ((($85)) + 244|0);
 HEAP16[$86>>1] = 0;
 $87 = $0;
 $88 = ((($87)) + 246|0);
 HEAP16[$88>>1] = 0;
 $89 = $0;
 $90 = ((($89)) + 376|0);
 HEAP32[$90>>2] = 0;
 $91 = $0;
 $92 = ((($91)) + 248|0);
 dest=$92; stop=dest+128|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $93 = $0;
 _enet_peer_reset_queues($93);
 STACKTOP = sp;return;
}
function _enet_peer_ping($peer) {
 $peer = $peer|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $command = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $command = sp + 8|0;
 $0 = $peer;
 $1 = $0;
 $2 = ((($1)) + 36|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)!=(5);
 if ($4) {
  STACKTOP = sp;return;
 }
 HEAP8[$command>>0] = -123;
 $5 = ((($command)) + 1|0);
 HEAP8[$5>>0] = -1;
 $6 = $0;
 (_enet_peer_queue_outgoing_command($6,$command,0,0,0)|0);
 STACKTOP = sp;return;
}
function _enet_peer_ping_interval($peer,$pingInterval) {
 $peer = $peer|0;
 $pingInterval = $pingInterval|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $peer;
 $1 = $pingInterval;
 $2 = $1;
 $3 = ($2|0)!=(0);
 $4 = $1;
 $5 = $3 ? $4 : 500;
 $6 = $0;
 $7 = ((($6)) + 136|0);
 HEAP32[$7>>2] = $5;
 STACKTOP = sp;return;
}
function _enet_peer_timeout($peer,$timeoutLimit,$timeoutMinimum,$timeoutMaximum) {
 $peer = $peer|0;
 $timeoutLimit = $timeoutLimit|0;
 $timeoutMinimum = $timeoutMinimum|0;
 $timeoutMaximum = $timeoutMaximum|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $peer;
 $1 = $timeoutLimit;
 $2 = $timeoutMinimum;
 $3 = $timeoutMaximum;
 $4 = $1;
 $5 = ($4|0)!=(0);
 $6 = $1;
 $7 = $5 ? $6 : 32;
 $8 = $0;
 $9 = ((($8)) + 140|0);
 HEAP32[$9>>2] = $7;
 $10 = $2;
 $11 = ($10|0)!=(0);
 $12 = $2;
 $13 = $11 ? $12 : 5000;
 $14 = $0;
 $15 = ((($14)) + 144|0);
 HEAP32[$15>>2] = $13;
 $16 = $3;
 $17 = ($16|0)!=(0);
 $18 = $3;
 $19 = $17 ? $18 : 30000;
 $20 = $0;
 $21 = ((($20)) + 148|0);
 HEAP32[$21>>2] = $19;
 STACKTOP = sp;return;
}
function _enet_peer_disconnect_now($peer,$data) {
 $peer = $peer|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $command = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $command = sp + 8|0;
 $0 = $peer;
 $1 = $data;
 $2 = $0;
 $3 = ((($2)) + 36|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0);
 if ($5) {
  STACKTOP = sp;return;
 }
 $6 = $0;
 $7 = ((($6)) + 36|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)!=(9);
 if ($9) {
  $10 = $0;
  $11 = ((($10)) + 36|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ($12|0)!=(7);
  if ($13) {
   $14 = $0;
   _enet_peer_reset_queues($14);
   HEAP8[$command>>0] = 68;
   $15 = ((($command)) + 1|0);
   HEAP8[$15>>0] = -1;
   $16 = $1;
   $17 = (_htonl($16)|0);
   $18 = ((($command)) + 4|0);
   HEAP8[$18>>0]=$17&255;HEAP8[$18+1>>0]=($17>>8)&255;HEAP8[$18+2>>0]=($17>>16)&255;HEAP8[$18+3>>0]=$17>>24;
   $19 = $0;
   (_enet_peer_queue_outgoing_command($19,$command,0,0,0)|0);
   $20 = $0;
   $21 = ((($20)) + 8|0);
   $22 = HEAP32[$21>>2]|0;
   _enet_host_flush($22);
  }
 }
 $23 = $0;
 _enet_peer_reset($23);
 STACKTOP = sp;return;
}
function _enet_peer_disconnect($peer,$data) {
 $peer = $peer|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $command = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $command = sp + 8|0;
 $0 = $peer;
 $1 = $data;
 $2 = $0;
 $3 = ((($2)) + 36|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(7);
 if ($5) {
  STACKTOP = sp;return;
 }
 $6 = $0;
 $7 = ((($6)) + 36|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0);
 if ($9) {
  STACKTOP = sp;return;
 }
 $10 = $0;
 $11 = ((($10)) + 36|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)==(8);
 if ($13) {
  STACKTOP = sp;return;
 }
 $14 = $0;
 $15 = ((($14)) + 36|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ($16|0)==(9);
 if ($17) {
  STACKTOP = sp;return;
 }
 $18 = $0;
 _enet_peer_reset_queues($18);
 HEAP8[$command>>0] = 4;
 $19 = ((($command)) + 1|0);
 HEAP8[$19>>0] = -1;
 $20 = $1;
 $21 = (_htonl($20)|0);
 $22 = ((($command)) + 4|0);
 HEAP8[$22>>0]=$21&255;HEAP8[$22+1>>0]=($21>>8)&255;HEAP8[$22+2>>0]=($21>>16)&255;HEAP8[$22+3>>0]=$21>>24;
 $23 = $0;
 $24 = ((($23)) + 36|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = ($25|0)==(5);
 if ($26) {
  label = 7;
 } else {
  $27 = $0;
  $28 = ((($27)) + 36|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = ($29|0)==(6);
  if ($30) {
   label = 7;
  } else {
   $35 = HEAP8[$command>>0]|0;
   $36 = $35&255;
   $37 = $36 | 64;
   $38 = $37&255;
   HEAP8[$command>>0] = $38;
  }
 }
 if ((label|0) == 7) {
  $31 = HEAP8[$command>>0]|0;
  $32 = $31&255;
  $33 = $32 | 128;
  $34 = $33&255;
  HEAP8[$command>>0] = $34;
 }
 $39 = $0;
 (_enet_peer_queue_outgoing_command($39,$command,0,0,0)|0);
 $40 = $0;
 $41 = ((($40)) + 36|0);
 $42 = HEAP32[$41>>2]|0;
 $43 = ($42|0)==(5);
 if (!($43)) {
  $44 = $0;
  $45 = ((($44)) + 36|0);
  $46 = HEAP32[$45>>2]|0;
  $47 = ($46|0)==(6);
  if (!($47)) {
   $51 = $0;
   $52 = ((($51)) + 8|0);
   $53 = HEAP32[$52>>2]|0;
   _enet_host_flush($53);
   $54 = $0;
   _enet_peer_reset($54);
   STACKTOP = sp;return;
  }
 }
 $48 = $0;
 _enet_peer_on_disconnect($48);
 $49 = $0;
 $50 = ((($49)) + 36|0);
 HEAP32[$50>>2] = 7;
 STACKTOP = sp;return;
}
function _enet_peer_disconnect_later($peer,$data) {
 $peer = $peer|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $peer;
 $1 = $data;
 $2 = $0;
 $3 = ((($2)) + 36|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(5);
 if ($5) {
  label = 3;
 } else {
  $6 = $0;
  $7 = ((($6)) + 36|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8|0)==(6);
  if ($9) {
   label = 3;
  }
 }
 do {
  if ((label|0) == 3) {
   $10 = $0;
   $11 = ((($10)) + 216|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = $0;
   $14 = ((($13)) + 216|0);
   $15 = ($12|0)==($14|0);
   if ($15) {
    $16 = $0;
    $17 = ((($16)) + 224|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = $0;
    $20 = ((($19)) + 224|0);
    $21 = ($18|0)==($20|0);
    if ($21) {
     $22 = $0;
     $23 = ((($22)) + 200|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = $0;
     $26 = ((($25)) + 200|0);
     $27 = ($24|0)==($26|0);
     if ($27) {
      break;
     }
    }
   }
   $28 = $0;
   $29 = ((($28)) + 36|0);
   HEAP32[$29>>2] = 6;
   $30 = $1;
   $31 = $0;
   $32 = ((($31)) + 376|0);
   HEAP32[$32>>2] = $30;
   STACKTOP = sp;return;
  }
 } while(0);
 $33 = $0;
 $34 = $1;
 _enet_peer_disconnect($33,$34);
 STACKTOP = sp;return;
}
function _enet_peer_queue_acknowledgement($peer,$command,$sentTime) {
 $peer = $peer|0;
 $command = $command|0;
 $sentTime = $sentTime|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $8 = 0, $9 = 0, $acknowledgement = 0, $channel = 0, $currentWindow = 0;
 var $reliableWindow = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $peer;
 $2 = $command;
 $3 = $sentTime;
 $4 = $2;
 $5 = ((($4)) + 1|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = $6&255;
 $8 = $1;
 $9 = ((($8)) + 44|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($7>>>0)<($10>>>0);
 if ($11) {
  $12 = $2;
  $13 = ((($12)) + 1|0);
  $14 = HEAP8[$13>>0]|0;
  $15 = $14&255;
  $16 = $1;
  $17 = ((($16)) + 40|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = (($18) + (($15*60)|0)|0);
  $channel = $19;
  $20 = $2;
  $21 = ((($20)) + 2|0);
  $22 = HEAPU8[$21>>0]|(HEAPU8[$21+1>>0]<<8);
  $23 = $22&65535;
  $24 = (($23|0) / 4096)&-1;
  $25 = $24&65535;
  $reliableWindow = $25;
  $26 = $channel;
  $27 = ((($26)) + 38|0);
  $28 = HEAP16[$27>>1]|0;
  $29 = $28&65535;
  $30 = (($29|0) / 4096)&-1;
  $31 = $30&65535;
  $currentWindow = $31;
  $32 = $2;
  $33 = ((($32)) + 2|0);
  $34 = HEAPU8[$33>>0]|(HEAPU8[$33+1>>0]<<8);
  $35 = $34&65535;
  $36 = $channel;
  $37 = ((($36)) + 38|0);
  $38 = HEAP16[$37>>1]|0;
  $39 = $38&65535;
  $40 = ($35|0)<($39|0);
  if ($40) {
   $41 = $reliableWindow;
   $42 = $41&65535;
   $43 = (($42) + 16)|0;
   $44 = $43&65535;
   $reliableWindow = $44;
  }
  $45 = $reliableWindow;
  $46 = $45&65535;
  $47 = $currentWindow;
  $48 = $47&65535;
  $49 = (($48) + 8)|0;
  $50 = (($49) - 1)|0;
  $51 = ($46|0)>=($50|0);
  if ($51) {
   $52 = $reliableWindow;
   $53 = $52&65535;
   $54 = $currentWindow;
   $55 = $54&65535;
   $56 = (($55) + 8)|0;
   $57 = ($53|0)<=($56|0);
   if ($57) {
    $0 = 0;
    $76 = $0;
    STACKTOP = sp;return ($76|0);
   }
  }
 }
 $58 = (_enet_malloc(60)|0);
 $acknowledgement = $58;
 $59 = $acknowledgement;
 $60 = ($59|0)==(0|0);
 if ($60) {
  $0 = 0;
  $76 = $0;
  STACKTOP = sp;return ($76|0);
 } else {
  $61 = $1;
  $62 = ((($61)) + 68|0);
  $63 = HEAP32[$62>>2]|0;
  $64 = (($63) + 8)|0;
  HEAP32[$62>>2] = $64;
  $65 = $3;
  $66 = $65&65535;
  $67 = $acknowledgement;
  $68 = ((($67)) + 8|0);
  HEAP32[$68>>2] = $66;
  $69 = $acknowledgement;
  $70 = ((($69)) + 12|0);
  $71 = $2;
  dest=$70; src=$71; stop=dest+48|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
  $72 = $1;
  $73 = ((($72)) + 192|0);
  $74 = $acknowledgement;
  (_enet_list_insert($73,$74)|0);
  $75 = $acknowledgement;
  $0 = $75;
  $76 = $0;
  STACKTOP = sp;return ($76|0);
 }
 return (0)|0;
}
function _enet_peer_dispatch_incoming_unreliable_commands($peer,$channel) {
 $peer = $peer|0;
 $channel = $channel|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $currentCommand = 0, $currentWindow = 0, $droppedCommand = 0, $incomingCommand = 0, $reliableWindow = 0, $startCommand = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $peer;
 $1 = $channel;
 $2 = $1;
 $3 = ((($2)) + 52|0);
 $4 = HEAP32[$3>>2]|0;
 $currentCommand = $4;
 $startCommand = $4;
 $droppedCommand = $4;
 L1: while(1) {
  $5 = $currentCommand;
  $6 = $1;
  $7 = ((($6)) + 52|0);
  $8 = ($5|0)!=($7|0);
  if (!($8)) {
   break;
  }
  $9 = $currentCommand;
  $incomingCommand = $9;
  $10 = $incomingCommand;
  $11 = ((($10)) + 12|0);
  $12 = HEAP8[$11>>0]|0;
  $13 = $12&255;
  $14 = $13 & 15;
  $15 = ($14|0)==(9);
  L4: do {
   if (!($15)) {
    $16 = $incomingCommand;
    $17 = ((($16)) + 8|0);
    $18 = HEAP16[$17>>1]|0;
    $19 = $18&65535;
    $20 = $1;
    $21 = ((($20)) + 38|0);
    $22 = HEAP16[$21>>1]|0;
    $23 = $22&65535;
    $24 = ($19|0)==($23|0);
    $25 = $incomingCommand;
    do {
     if ($24) {
      $26 = ((($25)) + 64|0);
      $27 = HEAP32[$26>>2]|0;
      $28 = ($27>>>0)<=(0);
      if ($28) {
       $29 = $incomingCommand;
       $30 = ((($29)) + 10|0);
       $31 = HEAP16[$30>>1]|0;
       $32 = $1;
       $33 = ((($32)) + 40|0);
       HEAP16[$33>>1] = $31;
       break L4;
      }
      $34 = $startCommand;
      $35 = $currentCommand;
      $36 = ($34|0)!=($35|0);
      if (!($36)) {
       $55 = $droppedCommand;
       $56 = $currentCommand;
       $57 = ($55|0)!=($56|0);
       if (!($57)) {
        break;
       }
       $58 = $currentCommand;
       $59 = ((($58)) + 4|0);
       $60 = HEAP32[$59>>2]|0;
       $droppedCommand = $60;
       break;
      }
      $37 = $0;
      $38 = ((($37)) + 232|0);
      $39 = $startCommand;
      $40 = $currentCommand;
      $41 = ((($40)) + 4|0);
      $42 = HEAP32[$41>>2]|0;
      (_enet_list_move($38,$39,$42)|0);
      $43 = $0;
      $44 = ((($43)) + 240|0);
      $45 = HEAP32[$44>>2]|0;
      $46 = ($45|0)!=(0);
      if (!($46)) {
       $47 = $0;
       $48 = ((($47)) + 8|0);
       $49 = HEAP32[$48>>2]|0;
       $50 = ((($49)) + 52|0);
       $51 = $0;
       (_enet_list_insert($50,$51)|0);
       $52 = $0;
       $53 = ((($52)) + 240|0);
       HEAP32[$53>>2] = 1;
      }
      $54 = $currentCommand;
      $droppedCommand = $54;
     } else {
      $61 = ((($25)) + 8|0);
      $62 = HEAP16[$61>>1]|0;
      $63 = $62&65535;
      $64 = (($63|0) / 4096)&-1;
      $65 = $64&65535;
      $reliableWindow = $65;
      $66 = $1;
      $67 = ((($66)) + 38|0);
      $68 = HEAP16[$67>>1]|0;
      $69 = $68&65535;
      $70 = (($69|0) / 4096)&-1;
      $71 = $70&65535;
      $currentWindow = $71;
      $72 = $incomingCommand;
      $73 = ((($72)) + 8|0);
      $74 = HEAP16[$73>>1]|0;
      $75 = $74&65535;
      $76 = $1;
      $77 = ((($76)) + 38|0);
      $78 = HEAP16[$77>>1]|0;
      $79 = $78&65535;
      $80 = ($75|0)<($79|0);
      if ($80) {
       $81 = $reliableWindow;
       $82 = $81&65535;
       $83 = (($82) + 16)|0;
       $84 = $83&65535;
       $reliableWindow = $84;
      }
      $85 = $reliableWindow;
      $86 = $85&65535;
      $87 = $currentWindow;
      $88 = $87&65535;
      $89 = ($86|0)>=($88|0);
      if ($89) {
       $90 = $reliableWindow;
       $91 = $90&65535;
       $92 = $currentWindow;
       $93 = $92&65535;
       $94 = (($93) + 8)|0;
       $95 = (($94) - 1)|0;
       $96 = ($91|0)<($95|0);
       if ($96) {
        break L1;
       }
      }
      $97 = $currentCommand;
      $98 = HEAP32[$97>>2]|0;
      $droppedCommand = $98;
      $99 = $startCommand;
      $100 = $currentCommand;
      $101 = ($99|0)!=($100|0);
      if ($101) {
       $102 = $0;
       $103 = ((($102)) + 232|0);
       $104 = $startCommand;
       $105 = $currentCommand;
       $106 = ((($105)) + 4|0);
       $107 = HEAP32[$106>>2]|0;
       (_enet_list_move($103,$104,$107)|0);
       $108 = $0;
       $109 = ((($108)) + 240|0);
       $110 = HEAP32[$109>>2]|0;
       $111 = ($110|0)!=(0);
       if (!($111)) {
        $112 = $0;
        $113 = ((($112)) + 8|0);
        $114 = HEAP32[$113>>2]|0;
        $115 = ((($114)) + 52|0);
        $116 = $0;
        (_enet_list_insert($115,$116)|0);
        $117 = $0;
        $118 = ((($117)) + 240|0);
        HEAP32[$118>>2] = 1;
       }
      }
     }
    } while(0);
    $119 = $currentCommand;
    $120 = HEAP32[$119>>2]|0;
    $startCommand = $120;
   }
  } while(0);
  $121 = $currentCommand;
  $122 = HEAP32[$121>>2]|0;
  $currentCommand = $122;
 }
 $123 = $startCommand;
 $124 = $currentCommand;
 $125 = ($123|0)!=($124|0);
 if (!($125)) {
  $144 = $1;
  $145 = ((($144)) + 52|0);
  $146 = $1;
  $147 = ((($146)) + 52|0);
  $148 = HEAP32[$147>>2]|0;
  $149 = $droppedCommand;
  _enet_peer_remove_incoming_commands($145,$148,$149);
  STACKTOP = sp;return;
 }
 $126 = $0;
 $127 = ((($126)) + 232|0);
 $128 = $startCommand;
 $129 = $currentCommand;
 $130 = ((($129)) + 4|0);
 $131 = HEAP32[$130>>2]|0;
 (_enet_list_move($127,$128,$131)|0);
 $132 = $0;
 $133 = ((($132)) + 240|0);
 $134 = HEAP32[$133>>2]|0;
 $135 = ($134|0)!=(0);
 if (!($135)) {
  $136 = $0;
  $137 = ((($136)) + 8|0);
  $138 = HEAP32[$137>>2]|0;
  $139 = ((($138)) + 52|0);
  $140 = $0;
  (_enet_list_insert($139,$140)|0);
  $141 = $0;
  $142 = ((($141)) + 240|0);
  HEAP32[$142>>2] = 1;
 }
 $143 = $currentCommand;
 $droppedCommand = $143;
 $144 = $1;
 $145 = ((($144)) + 52|0);
 $146 = $1;
 $147 = ((($146)) + 52|0);
 $148 = HEAP32[$147>>2]|0;
 $149 = $droppedCommand;
 _enet_peer_remove_incoming_commands($145,$148,$149);
 STACKTOP = sp;return;
}
function _enet_peer_dispatch_incoming_reliable_commands($peer,$channel) {
 $peer = $peer|0;
 $channel = $channel|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $9 = 0, $currentCommand = 0, $incomingCommand = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $peer;
 $1 = $channel;
 $2 = $1;
 $3 = ((($2)) + 44|0);
 $4 = HEAP32[$3>>2]|0;
 $currentCommand = $4;
 while(1) {
  $5 = $currentCommand;
  $6 = $1;
  $7 = ((($6)) + 44|0);
  $8 = ($5|0)!=($7|0);
  if (!($8)) {
   break;
  }
  $9 = $currentCommand;
  $incomingCommand = $9;
  $10 = $incomingCommand;
  $11 = ((($10)) + 64|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ($12>>>0)>(0);
  if ($13) {
   break;
  }
  $14 = $incomingCommand;
  $15 = ((($14)) + 8|0);
  $16 = HEAP16[$15>>1]|0;
  $17 = $16&65535;
  $18 = $1;
  $19 = ((($18)) + 38|0);
  $20 = HEAP16[$19>>1]|0;
  $21 = $20&65535;
  $22 = (($21) + 1)|0;
  $23 = $22&65535;
  $24 = $23&65535;
  $25 = ($17|0)!=($24|0);
  if ($25) {
   break;
  }
  $26 = $incomingCommand;
  $27 = ((($26)) + 8|0);
  $28 = HEAP16[$27>>1]|0;
  $29 = $1;
  $30 = ((($29)) + 38|0);
  HEAP16[$30>>1] = $28;
  $31 = $incomingCommand;
  $32 = ((($31)) + 60|0);
  $33 = HEAP32[$32>>2]|0;
  $34 = ($33>>>0)>(0);
  if ($34) {
   $35 = $incomingCommand;
   $36 = ((($35)) + 60|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = (($37) - 1)|0;
   $39 = $1;
   $40 = ((($39)) + 38|0);
   $41 = HEAP16[$40>>1]|0;
   $42 = $41&65535;
   $43 = (($42) + ($38))|0;
   $44 = $43&65535;
   HEAP16[$40>>1] = $44;
  }
  $45 = $currentCommand;
  $46 = HEAP32[$45>>2]|0;
  $currentCommand = $46;
 }
 $47 = $currentCommand;
 $48 = $1;
 $49 = ((($48)) + 44|0);
 $50 = HEAP32[$49>>2]|0;
 $51 = ($47|0)==($50|0);
 if ($51) {
  STACKTOP = sp;return;
 }
 $52 = $1;
 $53 = ((($52)) + 40|0);
 HEAP16[$53>>1] = 0;
 $54 = $0;
 $55 = ((($54)) + 232|0);
 $56 = $1;
 $57 = ((($56)) + 44|0);
 $58 = HEAP32[$57>>2]|0;
 $59 = $currentCommand;
 $60 = ((($59)) + 4|0);
 $61 = HEAP32[$60>>2]|0;
 (_enet_list_move($55,$58,$61)|0);
 $62 = $0;
 $63 = ((($62)) + 240|0);
 $64 = HEAP32[$63>>2]|0;
 $65 = ($64|0)!=(0);
 if (!($65)) {
  $66 = $0;
  $67 = ((($66)) + 8|0);
  $68 = HEAP32[$67>>2]|0;
  $69 = ((($68)) + 52|0);
  $70 = $0;
  (_enet_list_insert($69,$70)|0);
  $71 = $0;
  $72 = ((($71)) + 240|0);
  HEAP32[$72>>2] = 1;
 }
 $73 = $1;
 $74 = ((($73)) + 52|0);
 $75 = HEAP32[$74>>2]|0;
 $76 = $1;
 $77 = ((($76)) + 52|0);
 $78 = ($75|0)==($77|0);
 if ($78) {
  STACKTOP = sp;return;
 }
 $79 = $0;
 $80 = $1;
 _enet_peer_dispatch_incoming_unreliable_commands($79,$80);
 STACKTOP = sp;return;
}
function _enet_peer_queue_incoming_command($peer,$command,$packet,$fragmentCount) {
 $peer = $peer|0;
 $command = $command|0;
 $packet = $packet|0;
 $fragmentCount = $fragmentCount|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $channel = 0, $currentCommand = 0, $currentWindow = 0, $incomingCommand = 0, $reliableSequenceNumber = 0, $reliableWindow = 0, $unreliableSequenceNumber = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $peer;
 $2 = $command;
 $3 = $packet;
 $4 = $fragmentCount;
 $5 = $2;
 $6 = ((($5)) + 1|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $9 = $1;
 $10 = ((($9)) + 40|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + (($8*60)|0)|0);
 $channel = $12;
 $unreliableSequenceNumber = 0;
 $reliableSequenceNumber = 0;
 $13 = $1;
 $14 = ((($13)) + 36|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)==(6);
 L1: do {
  if ($16) {
   label = 44;
  } else {
   $17 = $2;
   $18 = HEAP8[$17>>0]|0;
   $19 = $18&255;
   $20 = $19 & 15;
   $21 = ($20|0)!=(9);
   if ($21) {
    $22 = $2;
    $23 = ((($22)) + 2|0);
    $24 = HEAPU8[$23>>0]|(HEAPU8[$23+1>>0]<<8);
    $25 = $24&65535;
    $reliableSequenceNumber = $25;
    $26 = $reliableSequenceNumber;
    $27 = (($26>>>0) / 4096)&-1;
    $28 = $27&65535;
    $reliableWindow = $28;
    $29 = $channel;
    $30 = ((($29)) + 38|0);
    $31 = HEAP16[$30>>1]|0;
    $32 = $31&65535;
    $33 = (($32|0) / 4096)&-1;
    $34 = $33&65535;
    $currentWindow = $34;
    $35 = $reliableSequenceNumber;
    $36 = $channel;
    $37 = ((($36)) + 38|0);
    $38 = HEAP16[$37>>1]|0;
    $39 = $38&65535;
    $40 = ($35>>>0)<($39>>>0);
    if ($40) {
     $41 = $reliableWindow;
     $42 = $41&65535;
     $43 = (($42) + 16)|0;
     $44 = $43&65535;
     $reliableWindow = $44;
    }
    $45 = $reliableWindow;
    $46 = $45&65535;
    $47 = $currentWindow;
    $48 = $47&65535;
    $49 = ($46|0)<($48|0);
    if ($49) {
     label = 44;
     break;
    }
    $50 = $reliableWindow;
    $51 = $50&65535;
    $52 = $currentWindow;
    $53 = $52&65535;
    $54 = (($53) + 8)|0;
    $55 = (($54) - 1)|0;
    $56 = ($51|0)>=($55|0);
    if ($56) {
     label = 44;
     break;
    }
   }
   $57 = $2;
   $58 = HEAP8[$57>>0]|0;
   $59 = $58&255;
   $60 = $59 & 15;
   L10: do {
    switch ($60|0) {
    case 6: case 8:  {
     $61 = $reliableSequenceNumber;
     $62 = $channel;
     $63 = ((($62)) + 38|0);
     $64 = HEAP16[$63>>1]|0;
     $65 = $64&65535;
     $66 = ($61|0)==($65|0);
     if ($66) {
      label = 44;
      break L1;
     }
     $67 = $channel;
     $68 = ((($67)) + 44|0);
     $69 = ((($68)) + 4|0);
     $70 = HEAP32[$69>>2]|0;
     $currentCommand = $70;
     while(1) {
      $71 = $currentCommand;
      $72 = $channel;
      $73 = ((($72)) + 44|0);
      $74 = ($71|0)!=($73|0);
      if (!($74)) {
       break L10;
      }
      $75 = $currentCommand;
      $incomingCommand = $75;
      $76 = $reliableSequenceNumber;
      $77 = $channel;
      $78 = ((($77)) + 38|0);
      $79 = HEAP16[$78>>1]|0;
      $80 = $79&65535;
      $81 = ($76>>>0)>=($80>>>0);
      $82 = $incomingCommand;
      $83 = ((($82)) + 8|0);
      $84 = HEAP16[$83>>1]|0;
      $85 = $84&65535;
      $86 = $channel;
      $87 = ((($86)) + 38|0);
      $88 = HEAP16[$87>>1]|0;
      $89 = $88&65535;
      if ($81) {
       $90 = ($85|0)<($89|0);
       if (!($90)) {
        label = 14;
       }
      } else {
       $91 = ($85|0)>=($89|0);
       if ($91) {
        break L10;
       } else {
        label = 14;
       }
      }
      if ((label|0) == 14) {
       label = 0;
       $92 = $incomingCommand;
       $93 = ((($92)) + 8|0);
       $94 = HEAP16[$93>>1]|0;
       $95 = $94&65535;
       $96 = $reliableSequenceNumber;
       $97 = ($95>>>0)<=($96>>>0);
       if ($97) {
        break;
       }
      }
      $104 = $currentCommand;
      $105 = ((($104)) + 4|0);
      $106 = HEAP32[$105>>2]|0;
      $currentCommand = $106;
     }
     $98 = $incomingCommand;
     $99 = ((($98)) + 8|0);
     $100 = HEAP16[$99>>1]|0;
     $101 = $100&65535;
     $102 = $reliableSequenceNumber;
     $103 = ($101>>>0)<($102>>>0);
     if (!($103)) {
      label = 44;
      break L1;
     }
     break;
    }
    case 12: case 7:  {
     $107 = $2;
     $108 = ((($107)) + 4|0);
     $109 = HEAPU8[$108>>0]|(HEAPU8[$108+1>>0]<<8);
     $110 = (_ntohs($109)|0);
     $111 = $110&65535;
     $unreliableSequenceNumber = $111;
     $112 = $reliableSequenceNumber;
     $113 = $channel;
     $114 = ((($113)) + 38|0);
     $115 = HEAP16[$114>>1]|0;
     $116 = $115&65535;
     $117 = ($112|0)==($116|0);
     if ($117) {
      $118 = $unreliableSequenceNumber;
      $119 = $channel;
      $120 = ((($119)) + 40|0);
      $121 = HEAP16[$120>>1]|0;
      $122 = $121&65535;
      $123 = ($118>>>0)<=($122>>>0);
      if ($123) {
       label = 44;
       break L1;
      }
     }
     $124 = $channel;
     $125 = ((($124)) + 52|0);
     $126 = ((($125)) + 4|0);
     $127 = HEAP32[$126>>2]|0;
     $currentCommand = $127;
     L27: while(1) {
      $128 = $currentCommand;
      $129 = $channel;
      $130 = ((($129)) + 52|0);
      $131 = ($128|0)!=($130|0);
      if (!($131)) {
       break L10;
      }
      $132 = $currentCommand;
      $incomingCommand = $132;
      $133 = $2;
      $134 = HEAP8[$133>>0]|0;
      $135 = $134&255;
      $136 = $135 & 15;
      $137 = ($136|0)==(9);
      do {
       if (!($137)) {
        $138 = $reliableSequenceNumber;
        $139 = $channel;
        $140 = ((($139)) + 38|0);
        $141 = HEAP16[$140>>1]|0;
        $142 = $141&65535;
        $143 = ($138>>>0)>=($142>>>0);
        $144 = $incomingCommand;
        $145 = ((($144)) + 8|0);
        $146 = HEAP16[$145>>1]|0;
        $147 = $146&65535;
        $148 = $channel;
        $149 = ((($148)) + 38|0);
        $150 = HEAP16[$149>>1]|0;
        $151 = $150&65535;
        if ($143) {
         $152 = ($147|0)<($151|0);
         if ($152) {
          break;
         }
        } else {
         $153 = ($147|0)>=($151|0);
         if ($153) {
          break L10;
         }
        }
        $154 = $incomingCommand;
        $155 = ((($154)) + 8|0);
        $156 = HEAP16[$155>>1]|0;
        $157 = $156&65535;
        $158 = $reliableSequenceNumber;
        $159 = ($157>>>0)<($158>>>0);
        if ($159) {
         break L10;
        }
        $160 = $incomingCommand;
        $161 = ((($160)) + 8|0);
        $162 = HEAP16[$161>>1]|0;
        $163 = $162&65535;
        $164 = $reliableSequenceNumber;
        $165 = ($163>>>0)>($164>>>0);
        if (!($165)) {
         $166 = $incomingCommand;
         $167 = ((($166)) + 10|0);
         $168 = HEAP16[$167>>1]|0;
         $169 = $168&65535;
         $170 = $unreliableSequenceNumber;
         $171 = ($169>>>0)<=($170>>>0);
         if ($171) {
          break L27;
         }
        }
       }
      } while(0);
      $178 = $currentCommand;
      $179 = ((($178)) + 4|0);
      $180 = HEAP32[$179>>2]|0;
      $currentCommand = $180;
     }
     $172 = $incomingCommand;
     $173 = ((($172)) + 10|0);
     $174 = HEAP16[$173>>1]|0;
     $175 = $174&65535;
     $176 = $unreliableSequenceNumber;
     $177 = ($175>>>0)<($176>>>0);
     if (!($177)) {
      label = 44;
      break L1;
     }
     break;
    }
    case 9:  {
     $181 = $channel;
     $182 = ((($181)) + 52|0);
     $currentCommand = $182;
     break;
    }
    default: {
     label = 44;
     break L1;
    }
    }
   } while(0);
   $183 = (_enet_malloc(76)|0);
   $incomingCommand = $183;
   $184 = $incomingCommand;
   $185 = ($184|0)==(0|0);
   if (!($185)) {
    $186 = $2;
    $187 = ((($186)) + 2|0);
    $188 = HEAPU8[$187>>0]|(HEAPU8[$187+1>>0]<<8);
    $189 = $incomingCommand;
    $190 = ((($189)) + 8|0);
    HEAP16[$190>>1] = $188;
    $191 = $unreliableSequenceNumber;
    $192 = $191 & 65535;
    $193 = $192&65535;
    $194 = $incomingCommand;
    $195 = ((($194)) + 10|0);
    HEAP16[$195>>1] = $193;
    $196 = $incomingCommand;
    $197 = ((($196)) + 12|0);
    $198 = $2;
    dest=$197; src=$198; stop=dest+48|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
    $199 = $4;
    $200 = $incomingCommand;
    $201 = ((($200)) + 60|0);
    HEAP32[$201>>2] = $199;
    $202 = $4;
    $203 = $incomingCommand;
    $204 = ((($203)) + 64|0);
    HEAP32[$204>>2] = $202;
    $205 = $3;
    $206 = $incomingCommand;
    $207 = ((($206)) + 72|0);
    HEAP32[$207>>2] = $205;
    $208 = $incomingCommand;
    $209 = ((($208)) + 68|0);
    HEAP32[$209>>2] = 0;
    $210 = $4;
    $211 = ($210>>>0)>(0);
    do {
     if ($211) {
      $212 = $4;
      $213 = ($212>>>0)<=(1048576);
      if ($213) {
       $214 = $4;
       $215 = (($214) + 31)|0;
       $216 = (($215>>>0) / 32)&-1;
       $217 = $216<<2;
       $218 = (_enet_malloc($217)|0);
       $219 = $incomingCommand;
       $220 = ((($219)) + 68|0);
       HEAP32[$220>>2] = $218;
      }
      $221 = $incomingCommand;
      $222 = ((($221)) + 68|0);
      $223 = HEAP32[$222>>2]|0;
      $224 = ($223|0)==(0|0);
      $225 = $incomingCommand;
      if ($224) {
       _enet_free($225);
       break L1;
      } else {
       $226 = ((($225)) + 68|0);
       $227 = HEAP32[$226>>2]|0;
       $228 = $4;
       $229 = (($228) + 31)|0;
       $230 = (($229>>>0) / 32)&-1;
       $231 = $230<<2;
       _memset(($227|0),0,($231|0))|0;
       break;
      }
     }
    } while(0);
    $232 = $3;
    $233 = ($232|0)!=(0|0);
    if ($233) {
     $234 = $3;
     $235 = HEAP32[$234>>2]|0;
     $236 = (($235) + 1)|0;
     HEAP32[$234>>2] = $236;
    }
    $237 = $currentCommand;
    $238 = HEAP32[$237>>2]|0;
    $239 = $incomingCommand;
    (_enet_list_insert($238,$239)|0);
    $240 = $2;
    $241 = HEAP8[$240>>0]|0;
    $242 = $241&255;
    $243 = $242 & 15;
    switch ($243|0) {
    case 6: case 8:  {
     $244 = $1;
     $245 = $channel;
     _enet_peer_dispatch_incoming_reliable_commands($244,$245);
     break;
    }
    default: {
     $246 = $1;
     $247 = $channel;
     _enet_peer_dispatch_incoming_unreliable_commands($246,$247);
    }
    }
    $248 = $incomingCommand;
    $0 = $248;
    $263 = $0;
    STACKTOP = sp;return ($263|0);
   }
  }
 } while(0);
 if ((label|0) == 44) {
  $249 = $4;
  $250 = ($249>>>0)>(0);
  if (!($250)) {
   $251 = $3;
   $252 = ($251|0)!=(0|0);
   if ($252) {
    $253 = $3;
    $254 = HEAP32[$253>>2]|0;
    $255 = ($254|0)==(0);
    if ($255) {
     $256 = $3;
     _enet_packet_destroy($256);
    }
   }
   $0 = 625936;
   $263 = $0;
   STACKTOP = sp;return ($263|0);
  }
 }
 $257 = $3;
 $258 = ($257|0)!=(0|0);
 if ($258) {
  $259 = $3;
  $260 = HEAP32[$259>>2]|0;
  $261 = ($260|0)==(0);
  if ($261) {
   $262 = $3;
   _enet_packet_destroy($262);
  }
 }
 $0 = 0;
 $263 = $0;
 STACKTOP = sp;return ($263|0);
}
function _enet_protocol_command_size($commandNumber) {
 $commandNumber = $commandNumber|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $commandNumber;
 $1 = $0;
 $2 = $1&255;
 $3 = $2 & 15;
 $4 = (12128 + ($3<<2)|0);
 $5 = HEAP32[$4>>2]|0;
 STACKTOP = sp;return ($5|0);
}
function _enet_host_flush($host) {
 $host = $host|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $host;
 $1 = (_enet_time_get()|0);
 $2 = $0;
 $3 = ((($2)) + 48|0);
 HEAP32[$3>>2] = $1;
 $4 = $0;
 (_enet_protocol_send_outgoing_commands($4,0,0)|0);
 STACKTOP = sp;return;
}
function _enet_protocol_send_outgoing_commands($host,$event,$checkForTimeouts) {
 $host = $host|0;
 $event = $event|0;
 $checkForTimeouts = $checkForTimeouts|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0;
 var $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0;
 var $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0;
 var $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0;
 var $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0;
 var $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $checksum = 0, $compressedSize = 0, $currentPeer = 0, $header = 0, $headerData = 0;
 var $originalSize = 0, $packetLoss = 0, $sentLength = 0, $shouldCompress = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $headerData = sp + 48|0;
 $1 = $host;
 $2 = $event;
 $3 = $checkForTimeouts;
 $header = $headerData;
 $shouldCompress = 0;
 $4 = $1;
 $5 = ((($4)) + 60|0);
 HEAP32[$5>>2] = 1;
 L1: while(1) {
  $6 = $1;
  $7 = ((($6)) + 60|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8|0)!=(0);
  if (!($9)) {
   label = 59;
   break;
  }
  $10 = $1;
  $11 = ((($10)) + 60|0);
  HEAP32[$11>>2] = 0;
  $12 = $1;
  $13 = ((($12)) + 36|0);
  $14 = HEAP32[$13>>2]|0;
  $currentPeer = $14;
  while(1) {
   $15 = $currentPeer;
   $16 = $1;
   $17 = ((($16)) + 40|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = $1;
   $20 = ((($19)) + 36|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = (($21) + (($18*380)|0)|0);
   $23 = ($15>>>0)<($22>>>0);
   if (!($23)) {
    continue L1;
   }
   $24 = $currentPeer;
   $25 = ((($24)) + 36|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = ($26|0)==(0);
   do {
    if (!($27)) {
     $28 = $currentPeer;
     $29 = ((($28)) + 36|0);
     $30 = HEAP32[$29>>2]|0;
     $31 = ($30|0)==(9);
     if (!($31)) {
      $32 = $1;
      $33 = ((($32)) + 68|0);
      HEAP16[$33>>1] = 0;
      $34 = $1;
      $35 = ((($34)) + 1608|0);
      HEAP32[$35>>2] = 0;
      $36 = $1;
      $37 = ((($36)) + 2132|0);
      HEAP32[$37>>2] = 1;
      $38 = $1;
      $39 = ((($38)) + 64|0);
      HEAP32[$39>>2] = 4;
      $40 = $currentPeer;
      $41 = ((($40)) + 192|0);
      $42 = HEAP32[$41>>2]|0;
      $43 = $currentPeer;
      $44 = ((($43)) + 192|0);
      $45 = ($42|0)==($44|0);
      if (!($45)) {
       $46 = $1;
       $47 = $currentPeer;
       _enet_protocol_send_acknowledgements($46,$47);
      }
      $48 = $3;
      $49 = ($48|0)!=(0);
      if ($49) {
       $50 = $currentPeer;
       $51 = ((($50)) + 200|0);
       $52 = HEAP32[$51>>2]|0;
       $53 = $currentPeer;
       $54 = ((($53)) + 200|0);
       $55 = ($52|0)==($54|0);
       if (!($55)) {
        $56 = $1;
        $57 = ((($56)) + 48|0);
        $58 = HEAP32[$57>>2]|0;
        $59 = $currentPeer;
        $60 = ((($59)) + 80|0);
        $61 = HEAP32[$60>>2]|0;
        $62 = (($58) - ($61))|0;
        $63 = ($62>>>0)>=(86400000);
        if (!($63)) {
         $64 = $1;
         $65 = $currentPeer;
         $66 = $2;
         $67 = (_enet_protocol_check_timeouts($64,$65,$66)|0);
         $68 = ($67|0)==(1);
         if ($68) {
          $69 = $2;
          $70 = ($69|0)!=(0|0);
          if (!($70)) {
           break;
          }
          $71 = $2;
          $72 = HEAP32[$71>>2]|0;
          $73 = ($72|0)!=(0);
          if ($73) {
           label = 15;
           break L1;
          } else {
           break;
          }
         }
        }
       }
      }
      $74 = $currentPeer;
      $75 = ((($74)) + 216|0);
      $76 = HEAP32[$75>>2]|0;
      $77 = $currentPeer;
      $78 = ((($77)) + 216|0);
      $79 = ($76|0)==($78|0);
      if ($79) {
       label = 18;
      } else {
       $80 = $1;
       $81 = $currentPeer;
       $82 = (_enet_protocol_send_reliable_outgoing_commands($80,$81)|0);
       $83 = ($82|0)!=(0);
       if ($83) {
        label = 18;
       }
      }
      if ((label|0) == 18) {
       label = 0;
       $84 = $currentPeer;
       $85 = ((($84)) + 200|0);
       $86 = HEAP32[$85>>2]|0;
       $87 = $currentPeer;
       $88 = ((($87)) + 200|0);
       $89 = ($86|0)==($88|0);
       if ($89) {
        $90 = $1;
        $91 = ((($90)) + 48|0);
        $92 = HEAP32[$91>>2]|0;
        $93 = $currentPeer;
        $94 = ((($93)) + 76|0);
        $95 = HEAP32[$94>>2]|0;
        $96 = (($92) - ($95))|0;
        $97 = ($96>>>0)>=(86400000);
        if ($97) {
         $98 = $currentPeer;
         $99 = ((($98)) + 76|0);
         $100 = HEAP32[$99>>2]|0;
         $101 = $1;
         $102 = ((($101)) + 48|0);
         $103 = HEAP32[$102>>2]|0;
         $104 = (($100) - ($103))|0;
         $116 = $104;
        } else {
         $105 = $1;
         $106 = ((($105)) + 48|0);
         $107 = HEAP32[$106>>2]|0;
         $108 = $currentPeer;
         $109 = ((($108)) + 76|0);
         $110 = HEAP32[$109>>2]|0;
         $111 = (($107) - ($110))|0;
         $116 = $111;
        }
        $112 = $currentPeer;
        $113 = ((($112)) + 136|0);
        $114 = HEAP32[$113>>2]|0;
        $115 = ($116>>>0)>=($114>>>0);
        if ($115) {
         $117 = $currentPeer;
         $118 = ((($117)) + 176|0);
         $119 = HEAP32[$118>>2]|0;
         $120 = $1;
         $121 = ((($120)) + 64|0);
         $122 = HEAP32[$121>>2]|0;
         $123 = (($119) - ($122))|0;
         $124 = ($123>>>0)>=(4);
         if ($124) {
          $125 = $currentPeer;
          _enet_peer_ping($125);
          $126 = $1;
          $127 = $currentPeer;
          (_enet_protocol_send_reliable_outgoing_commands($126,$127)|0);
         }
        }
       }
      }
      $128 = $currentPeer;
      $129 = ((($128)) + 224|0);
      $130 = HEAP32[$129>>2]|0;
      $131 = $currentPeer;
      $132 = ((($131)) + 224|0);
      $133 = ($130|0)==($132|0);
      if (!($133)) {
       $134 = $1;
       $135 = $currentPeer;
       _enet_protocol_send_unreliable_outgoing_commands($134,$135);
      }
      $136 = $1;
      $137 = ((($136)) + 1608|0);
      $138 = HEAP32[$137>>2]|0;
      $139 = ($138|0)==(0);
      if (!($139)) {
       $140 = $currentPeer;
       $141 = ((($140)) + 88|0);
       $142 = HEAP32[$141>>2]|0;
       $143 = ($142|0)==(0);
       $144 = $1;
       $145 = ((($144)) + 48|0);
       $146 = HEAP32[$145>>2]|0;
       $147 = $currentPeer;
       $148 = ((($147)) + 88|0);
       if ($143) {
        HEAP32[$148>>2] = $146;
       } else {
        $149 = HEAP32[$148>>2]|0;
        $150 = (($146) - ($149))|0;
        $151 = ($150>>>0)>=(86400000);
        if ($151) {
         $152 = $currentPeer;
         $153 = ((($152)) + 88|0);
         $154 = HEAP32[$153>>2]|0;
         $155 = $1;
         $156 = ((($155)) + 48|0);
         $157 = HEAP32[$156>>2]|0;
         $158 = (($154) - ($157))|0;
         $167 = $158;
        } else {
         $159 = $1;
         $160 = ((($159)) + 48|0);
         $161 = HEAP32[$160>>2]|0;
         $162 = $currentPeer;
         $163 = ((($162)) + 88|0);
         $164 = HEAP32[$163>>2]|0;
         $165 = (($161) - ($164))|0;
         $167 = $165;
        }
        $166 = ($167>>>0)>=(10000);
        if ($166) {
         $168 = $currentPeer;
         $169 = ((($168)) + 92|0);
         $170 = HEAP32[$169>>2]|0;
         $171 = ($170>>>0)>(0);
         if ($171) {
          $172 = $currentPeer;
          $173 = ((($172)) + 96|0);
          $174 = HEAP32[$173>>2]|0;
          $175 = $174<<16;
          $176 = $currentPeer;
          $177 = ((($176)) + 92|0);
          $178 = HEAP32[$177>>2]|0;
          $179 = (($175>>>0) / ($178>>>0))&-1;
          $packetLoss = $179;
          $180 = $currentPeer;
          $181 = ((($180)) + 104|0);
          $182 = HEAP32[$181>>2]|0;
          $183 = (($182>>>0) / 4)&-1;
          $184 = $currentPeer;
          $185 = ((($184)) + 104|0);
          $186 = HEAP32[$185>>2]|0;
          $187 = (($186) - ($183))|0;
          HEAP32[$185>>2] = $187;
          $188 = $packetLoss;
          $189 = $currentPeer;
          $190 = ((($189)) + 100|0);
          $191 = HEAP32[$190>>2]|0;
          $192 = ($188>>>0)>=($191>>>0);
          if ($192) {
           $193 = $packetLoss;
           $194 = $currentPeer;
           $195 = ((($194)) + 100|0);
           $196 = HEAP32[$195>>2]|0;
           $197 = (($193) - ($196))|0;
           $198 = (($197>>>0) / 8)&-1;
           $199 = $currentPeer;
           $200 = ((($199)) + 100|0);
           $201 = HEAP32[$200>>2]|0;
           $202 = (($201) + ($198))|0;
           HEAP32[$200>>2] = $202;
           $203 = $packetLoss;
           $204 = $currentPeer;
           $205 = ((($204)) + 100|0);
           $206 = HEAP32[$205>>2]|0;
           $207 = (($203) - ($206))|0;
           $208 = (($207>>>0) / 4)&-1;
           $209 = $currentPeer;
           $210 = ((($209)) + 104|0);
           $211 = HEAP32[$210>>2]|0;
           $212 = (($211) + ($208))|0;
           HEAP32[$210>>2] = $212;
          } else {
           $213 = $currentPeer;
           $214 = ((($213)) + 100|0);
           $215 = HEAP32[$214>>2]|0;
           $216 = $packetLoss;
           $217 = (($215) - ($216))|0;
           $218 = (($217>>>0) / 8)&-1;
           $219 = $currentPeer;
           $220 = ((($219)) + 100|0);
           $221 = HEAP32[$220>>2]|0;
           $222 = (($221) - ($218))|0;
           HEAP32[$220>>2] = $222;
           $223 = $currentPeer;
           $224 = ((($223)) + 100|0);
           $225 = HEAP32[$224>>2]|0;
           $226 = $packetLoss;
           $227 = (($225) - ($226))|0;
           $228 = (($227>>>0) / 4)&-1;
           $229 = $currentPeer;
           $230 = ((($229)) + 104|0);
           $231 = HEAP32[$230>>2]|0;
           $232 = (($231) + ($228))|0;
           HEAP32[$230>>2] = $232;
          }
          $233 = $1;
          $234 = ((($233)) + 48|0);
          $235 = HEAP32[$234>>2]|0;
          $236 = $currentPeer;
          $237 = ((($236)) + 88|0);
          HEAP32[$237>>2] = $235;
          $238 = $currentPeer;
          $239 = ((($238)) + 92|0);
          HEAP32[$239>>2] = 0;
          $240 = $currentPeer;
          $241 = ((($240)) + 96|0);
          HEAP32[$241>>2] = 0;
         }
        }
       }
       $242 = $1;
       $243 = ((($242)) + 1612|0);
       HEAP32[$243>>2] = $headerData;
       $244 = $1;
       $245 = ((($244)) + 68|0);
       $246 = HEAP16[$245>>1]|0;
       $247 = $246&65535;
       $248 = $247 & 32768;
       $249 = ($248|0)!=(0);
       $250 = $1;
       if ($249) {
        $251 = ((($250)) + 48|0);
        $252 = HEAP32[$251>>2]|0;
        $253 = $252 & 65535;
        $254 = $253&65535;
        $255 = (_htons($254)|0);
        $256 = $header;
        $257 = ((($256)) + 2|0);
        HEAP8[$257>>0]=$255&255;HEAP8[$257+1>>0]=$255>>8;
        $258 = $1;
        $259 = ((($258)) + 1612|0);
        $260 = ((($259)) + 4|0);
        HEAP32[$260>>2] = 4;
       } else {
        $261 = ((($250)) + 1612|0);
        $262 = ((($261)) + 4|0);
        HEAP32[$262>>2] = ((((0) + 2|0)));
       }
       $shouldCompress = 0;
       $263 = $1;
       $264 = ((($263)) + 2140|0);
       $265 = HEAP32[$264>>2]|0;
       $266 = ($265|0)!=(0|0);
       if ($266) {
        $267 = $1;
        $268 = ((($267)) + 2140|0);
        $269 = ((($268)) + 4|0);
        $270 = HEAP32[$269>>2]|0;
        $271 = ($270|0)!=(0|0);
        if ($271) {
         $272 = $1;
         $273 = ((($272)) + 64|0);
         $274 = HEAP32[$273>>2]|0;
         $275 = (($274) - 4)|0;
         $originalSize = $275;
         $276 = $1;
         $277 = ((($276)) + 2140|0);
         $278 = ((($277)) + 4|0);
         $279 = HEAP32[$278>>2]|0;
         $280 = $1;
         $281 = ((($280)) + 2140|0);
         $282 = HEAP32[$281>>2]|0;
         $283 = $1;
         $284 = ((($283)) + 1612|0);
         $285 = ((($284)) + 8|0);
         $286 = $1;
         $287 = ((($286)) + 2132|0);
         $288 = HEAP32[$287>>2]|0;
         $289 = (($288) - 1)|0;
         $290 = $originalSize;
         $291 = $1;
         $292 = ((($291)) + 2156|0);
         $293 = ((($292)) + 4096|0);
         $294 = $originalSize;
         $295 = (FUNCTION_TABLE_iiiiiii[$279 & 8191]($282,$285,$289,$290,$293,$294)|0);
         $compressedSize = $295;
         $296 = $compressedSize;
         $297 = ($296>>>0)>(0);
         if ($297) {
          $298 = $compressedSize;
          $299 = $originalSize;
          $300 = ($298>>>0)<($299>>>0);
          if ($300) {
           $301 = $1;
           $302 = ((($301)) + 68|0);
           $303 = HEAP16[$302>>1]|0;
           $304 = $303&65535;
           $305 = $304 | 16384;
           $306 = $305&65535;
           HEAP16[$302>>1] = $306;
           $307 = $compressedSize;
           $shouldCompress = $307;
          }
         }
        }
       }
       $308 = $currentPeer;
       $309 = ((($308)) + 12|0);
       $310 = HEAP16[$309>>1]|0;
       $311 = $310&65535;
       $312 = ($311|0)<(4095);
       if ($312) {
        $313 = $currentPeer;
        $314 = ((($313)) + 20|0);
        $315 = HEAP8[$314>>0]|0;
        $316 = $315&255;
        $317 = $316 << 12;
        $318 = $1;
        $319 = ((($318)) + 68|0);
        $320 = HEAP16[$319>>1]|0;
        $321 = $320&65535;
        $322 = $321 | $317;
        $323 = $322&65535;
        HEAP16[$319>>1] = $323;
       }
       $324 = $currentPeer;
       $325 = ((($324)) + 12|0);
       $326 = HEAP16[$325>>1]|0;
       $327 = $326&65535;
       $328 = $1;
       $329 = ((($328)) + 68|0);
       $330 = HEAP16[$329>>1]|0;
       $331 = $330&65535;
       $332 = $327 | $331;
       $333 = $332&65535;
       $334 = (_htons($333)|0);
       $335 = $header;
       HEAP8[$335>>0]=$334&255;HEAP8[$335+1>>0]=$334>>8;
       $336 = $1;
       $337 = ((($336)) + 2136|0);
       $338 = HEAP32[$337>>2]|0;
       $339 = ($338|0)!=(0|0);
       if ($339) {
        $340 = $1;
        $341 = ((($340)) + 1612|0);
        $342 = ((($341)) + 4|0);
        $343 = HEAP32[$342>>2]|0;
        $344 = (($headerData) + ($343)|0);
        $checksum = $344;
        $345 = $currentPeer;
        $346 = ((($345)) + 12|0);
        $347 = HEAP16[$346>>1]|0;
        $348 = $347&65535;
        $349 = ($348|0)<(4095);
        if ($349) {
         $350 = $currentPeer;
         $351 = ((($350)) + 16|0);
         $352 = HEAP32[$351>>2]|0;
         $354 = $352;
        } else {
         $354 = 0;
        }
        $353 = $checksum;
        HEAP32[$353>>2] = $354;
        $355 = $1;
        $356 = ((($355)) + 1612|0);
        $357 = ((($356)) + 4|0);
        $358 = HEAP32[$357>>2]|0;
        $359 = (($358) + 4)|0;
        HEAP32[$357>>2] = $359;
        $360 = $1;
        $361 = ((($360)) + 2136|0);
        $362 = HEAP32[$361>>2]|0;
        $363 = $1;
        $364 = ((($363)) + 1612|0);
        $365 = $1;
        $366 = ((($365)) + 2132|0);
        $367 = HEAP32[$366>>2]|0;
        $368 = (FUNCTION_TABLE_iii[$362 & 8191]($364,$367)|0);
        $369 = $checksum;
        HEAP32[$369>>2] = $368;
       }
       $370 = $shouldCompress;
       $371 = ($370>>>0)>(0);
       if ($371) {
        $372 = $1;
        $373 = ((($372)) + 2156|0);
        $374 = ((($373)) + 4096|0);
        $375 = $1;
        $376 = ((($375)) + 1612|0);
        $377 = ((($376)) + 8|0);
        HEAP32[$377>>2] = $374;
        $378 = $shouldCompress;
        $379 = $1;
        $380 = ((($379)) + 1612|0);
        $381 = ((($380)) + 8|0);
        $382 = ((($381)) + 4|0);
        HEAP32[$382>>2] = $378;
        $383 = $1;
        $384 = ((($383)) + 2132|0);
        HEAP32[$384>>2] = 2;
       }
       $385 = $1;
       $386 = ((($385)) + 48|0);
       $387 = HEAP32[$386>>2]|0;
       $388 = $currentPeer;
       $389 = ((($388)) + 72|0);
       HEAP32[$389>>2] = $387;
       $390 = $1;
       $391 = HEAP32[$390>>2]|0;
       $392 = $currentPeer;
       $393 = ((($392)) + 24|0);
       $394 = $1;
       $395 = ((($394)) + 1612|0);
       $396 = $1;
       $397 = ((($396)) + 2132|0);
       $398 = HEAP32[$397>>2]|0;
       $399 = (_enet_socket_send($391,$393,$395,$398)|0);
       $sentLength = $399;
       $400 = $currentPeer;
       _enet_protocol_remove_sent_unreliable_commands($400);
       $401 = $sentLength;
       $402 = ($401|0)<(0);
       if ($402) {
        label = 56;
        break L1;
       }
       $403 = $sentLength;
       $404 = $1;
       $405 = ((($404)) + 10364|0);
       $406 = HEAP32[$405>>2]|0;
       $407 = (($406) + ($403))|0;
       HEAP32[$405>>2] = $407;
       $408 = $1;
       $409 = ((($408)) + 10368|0);
       $410 = HEAP32[$409>>2]|0;
       $411 = (($410) + 1)|0;
       HEAP32[$409>>2] = $411;
      }
     }
    }
   } while(0);
   $412 = $currentPeer;
   $413 = ((($412)) + 380|0);
   $currentPeer = $413;
  }
 }
 if ((label|0) == 15) {
  $0 = 1;
  $414 = $0;
  STACKTOP = sp;return ($414|0);
 }
 else if ((label|0) == 56) {
  $0 = -1;
  $414 = $0;
  STACKTOP = sp;return ($414|0);
 }
 else if ((label|0) == 59) {
  $0 = 0;
  $414 = $0;
  STACKTOP = sp;return ($414|0);
 }
 return (0)|0;
}
function _enet_protocol_send_acknowledgements($host,$peer) {
 $host = $host|0;
 $peer = $peer|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $acknowledgement = 0, $buffer = 0, $command = 0, $currentAcknowledgement = 0, $reliableSequenceNumber = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $host;
 $1 = $peer;
 $2 = $0;
 $3 = ((($2)) + 1608|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $0;
 $6 = ((($5)) + 70|0);
 $7 = (($6) + (($4*48)|0)|0);
 $command = $7;
 $8 = $0;
 $9 = ((($8)) + 2132|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $0;
 $12 = ((($11)) + 1612|0);
 $13 = (($12) + ($10<<3)|0);
 $buffer = $13;
 $14 = $1;
 $15 = ((($14)) + 192|0);
 $16 = HEAP32[$15>>2]|0;
 $currentAcknowledgement = $16;
 while(1) {
  $17 = $currentAcknowledgement;
  $18 = $1;
  $19 = ((($18)) + 192|0);
  $20 = ($17|0)!=($19|0);
  if (!($20)) {
   label = 10;
   break;
  }
  $21 = $command;
  $22 = $0;
  $23 = ((($22)) + 70|0);
  $24 = ((($23)) + 1536|0);
  $25 = ($21>>>0)>=($24>>>0);
  if ($25) {
   break;
  }
  $26 = $buffer;
  $27 = $0;
  $28 = ((($27)) + 1612|0);
  $29 = ((($28)) + 520|0);
  $30 = ($26>>>0)>=($29>>>0);
  if ($30) {
   break;
  }
  $31 = $1;
  $32 = ((($31)) + 176|0);
  $33 = HEAP32[$32>>2]|0;
  $34 = $0;
  $35 = ((($34)) + 64|0);
  $36 = HEAP32[$35>>2]|0;
  $37 = (($33) - ($36))|0;
  $38 = ($37>>>0)<(8);
  if ($38) {
   break;
  }
  $41 = $currentAcknowledgement;
  $acknowledgement = $41;
  $42 = $currentAcknowledgement;
  $43 = HEAP32[$42>>2]|0;
  $currentAcknowledgement = $43;
  $44 = $command;
  $45 = $buffer;
  HEAP32[$45>>2] = $44;
  $46 = $buffer;
  $47 = ((($46)) + 4|0);
  HEAP32[$47>>2] = 8;
  $48 = $buffer;
  $49 = ((($48)) + 4|0);
  $50 = HEAP32[$49>>2]|0;
  $51 = $0;
  $52 = ((($51)) + 64|0);
  $53 = HEAP32[$52>>2]|0;
  $54 = (($53) + ($50))|0;
  HEAP32[$52>>2] = $54;
  $55 = $acknowledgement;
  $56 = ((($55)) + 12|0);
  $57 = ((($56)) + 2|0);
  $58 = HEAP16[$57>>1]|0;
  $59 = (_htons($58)|0);
  $reliableSequenceNumber = $59;
  $60 = $command;
  HEAP8[$60>>0] = 1;
  $61 = $acknowledgement;
  $62 = ((($61)) + 12|0);
  $63 = ((($62)) + 1|0);
  $64 = HEAP8[$63>>0]|0;
  $65 = $command;
  $66 = ((($65)) + 1|0);
  HEAP8[$66>>0] = $64;
  $67 = $reliableSequenceNumber;
  $68 = $command;
  $69 = ((($68)) + 2|0);
  HEAP8[$69>>0]=$67&255;HEAP8[$69+1>>0]=$67>>8;
  $70 = $reliableSequenceNumber;
  $71 = $command;
  $72 = ((($71)) + 4|0);
  HEAP8[$72>>0]=$70&255;HEAP8[$72+1>>0]=$70>>8;
  $73 = $acknowledgement;
  $74 = ((($73)) + 8|0);
  $75 = HEAP32[$74>>2]|0;
  $76 = $75&65535;
  $77 = (_htons($76)|0);
  $78 = $command;
  $79 = ((($78)) + 6|0);
  HEAP8[$79>>0]=$77&255;HEAP8[$79+1>>0]=$77>>8;
  $80 = $acknowledgement;
  $81 = ((($80)) + 12|0);
  $82 = HEAP8[$81>>0]|0;
  $83 = $82&255;
  $84 = $83 & 15;
  $85 = ($84|0)==(4);
  if ($85) {
   $86 = $0;
   $87 = $1;
   _enet_protocol_dispatch_state($86,$87,9);
  }
  $88 = $acknowledgement;
  (_enet_list_remove($88)|0);
  $89 = $acknowledgement;
  _enet_free($89);
  $90 = $command;
  $91 = ((($90)) + 48|0);
  $command = $91;
  $92 = $buffer;
  $93 = ((($92)) + 8|0);
  $buffer = $93;
 }
 if ((label|0) == 10) {
  $94 = $command;
  $95 = $0;
  $96 = ((($95)) + 70|0);
  $97 = $94;
  $98 = $96;
  $99 = (($97) - ($98))|0;
  $100 = (($99|0) / 48)&-1;
  $101 = $0;
  $102 = ((($101)) + 1608|0);
  HEAP32[$102>>2] = $100;
  $103 = $buffer;
  $104 = $0;
  $105 = ((($104)) + 1612|0);
  $106 = $103;
  $107 = $105;
  $108 = (($106) - ($107))|0;
  $109 = (($108|0) / 8)&-1;
  $110 = $0;
  $111 = ((($110)) + 2132|0);
  HEAP32[$111>>2] = $109;
  STACKTOP = sp;return;
 }
 $39 = $0;
 $40 = ((($39)) + 60|0);
 HEAP32[$40>>2] = 1;
 $94 = $command;
 $95 = $0;
 $96 = ((($95)) + 70|0);
 $97 = $94;
 $98 = $96;
 $99 = (($97) - ($98))|0;
 $100 = (($99|0) / 48)&-1;
 $101 = $0;
 $102 = ((($101)) + 1608|0);
 HEAP32[$102>>2] = $100;
 $103 = $buffer;
 $104 = $0;
 $105 = ((($104)) + 1612|0);
 $106 = $103;
 $107 = $105;
 $108 = (($106) - ($107))|0;
 $109 = (($108|0) / 8)&-1;
 $110 = $0;
 $111 = ((($110)) + 2132|0);
 HEAP32[$111>>2] = $109;
 STACKTOP = sp;return;
}
function _enet_protocol_dispatch_state($host,$peer,$state) {
 $host = $host|0;
 $peer = $peer|0;
 $state = $state|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $host;
 $1 = $peer;
 $2 = $state;
 $3 = $0;
 $4 = $1;
 $5 = $2;
 _enet_protocol_change_state($3,$4,$5);
 $6 = $1;
 $7 = ((($6)) + 240|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)!=(0);
 if ($9) {
  STACKTOP = sp;return;
 }
 $10 = $0;
 $11 = ((($10)) + 52|0);
 $12 = $1;
 (_enet_list_insert($11,$12)|0);
 $13 = $1;
 $14 = ((($13)) + 240|0);
 HEAP32[$14>>2] = 1;
 STACKTOP = sp;return;
}
function _enet_protocol_change_state($host,$peer,$state) {
 $host = $host|0;
 $peer = $peer|0;
 $state = $state|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $host;
 $1 = $peer;
 $2 = $state;
 $3 = $2;
 $4 = ($3|0)==(5);
 $5 = $2;
 $6 = ($5|0)==(6);
 $or$cond = $4 | $6;
 $7 = $1;
 if ($or$cond) {
  _enet_peer_on_connect($7);
 } else {
  _enet_peer_on_disconnect($7);
 }
 $8 = $2;
 $9 = $1;
 $10 = ((($9)) + 36|0);
 HEAP32[$10>>2] = $8;
 STACKTOP = sp;return;
}
function _enet_protocol_check_timeouts($host,$peer,$event) {
 $host = $host|0;
 $peer = $peer|0;
 $event = $event|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $currentCommand = 0, $insertPosition = 0, $outgoingCommand = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $host;
 $2 = $peer;
 $3 = $event;
 $4 = $2;
 $5 = ((($4)) + 200|0);
 $6 = HEAP32[$5>>2]|0;
 $currentCommand = $6;
 $7 = $2;
 $8 = ((($7)) + 216|0);
 $9 = HEAP32[$8>>2]|0;
 $insertPosition = $9;
 while(1) {
  $10 = $currentCommand;
  $11 = $2;
  $12 = ((($11)) + 200|0);
  $13 = ($10|0)!=($12|0);
  if (!($13)) {
   label = 26;
   break;
  }
  $14 = $currentCommand;
  $outgoingCommand = $14;
  $15 = $currentCommand;
  $16 = HEAP32[$15>>2]|0;
  $currentCommand = $16;
  $17 = $1;
  $18 = ((($17)) + 48|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = $outgoingCommand;
  $21 = ((($20)) + 12|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = (($19) - ($22))|0;
  $24 = ($23>>>0)>=(86400000);
  if ($24) {
   $25 = $outgoingCommand;
   $26 = ((($25)) + 12|0);
   $27 = HEAP32[$26>>2]|0;
   $28 = $1;
   $29 = ((($28)) + 48|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = (($27) - ($30))|0;
   $43 = $31;
  } else {
   $32 = $1;
   $33 = ((($32)) + 48|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = $outgoingCommand;
   $36 = ((($35)) + 12|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = (($34) - ($37))|0;
   $43 = $38;
  }
  $39 = $outgoingCommand;
  $40 = ((($39)) + 16|0);
  $41 = HEAP32[$40>>2]|0;
  $42 = ($43>>>0)<($41>>>0);
  if ($42) {
   continue;
  }
  $44 = $2;
  $45 = ((($44)) + 84|0);
  $46 = HEAP32[$45>>2]|0;
  $47 = ($46|0)==(0);
  if ($47) {
   label = 9;
  } else {
   $48 = $outgoingCommand;
   $49 = ((($48)) + 12|0);
   $50 = HEAP32[$49>>2]|0;
   $51 = $2;
   $52 = ((($51)) + 84|0);
   $53 = HEAP32[$52>>2]|0;
   $54 = (($50) - ($53))|0;
   $55 = ($54>>>0)>=(86400000);
   if ($55) {
    label = 9;
   }
  }
  if ((label|0) == 9) {
   label = 0;
   $56 = $outgoingCommand;
   $57 = ((($56)) + 12|0);
   $58 = HEAP32[$57>>2]|0;
   $59 = $2;
   $60 = ((($59)) + 84|0);
   HEAP32[$60>>2] = $58;
  }
  $61 = $2;
  $62 = ((($61)) + 84|0);
  $63 = HEAP32[$62>>2]|0;
  $64 = ($63|0)!=(0);
  if ($64) {
   $65 = $1;
   $66 = ((($65)) + 48|0);
   $67 = HEAP32[$66>>2]|0;
   $68 = $2;
   $69 = ((($68)) + 84|0);
   $70 = HEAP32[$69>>2]|0;
   $71 = (($67) - ($70))|0;
   $72 = ($71>>>0)>=(86400000);
   if ($72) {
    $73 = $2;
    $74 = ((($73)) + 84|0);
    $75 = HEAP32[$74>>2]|0;
    $76 = $1;
    $77 = ((($76)) + 48|0);
    $78 = HEAP32[$77>>2]|0;
    $79 = (($75) - ($78))|0;
    $91 = $79;
   } else {
    $80 = $1;
    $81 = ((($80)) + 48|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = $2;
    $84 = ((($83)) + 84|0);
    $85 = HEAP32[$84>>2]|0;
    $86 = (($82) - ($85))|0;
    $91 = $86;
   }
   $87 = $2;
   $88 = ((($87)) + 148|0);
   $89 = HEAP32[$88>>2]|0;
   $90 = ($91>>>0)>=($89>>>0);
   if ($90) {
    label = 20;
    break;
   }
   $92 = $outgoingCommand;
   $93 = ((($92)) + 16|0);
   $94 = HEAP32[$93>>2]|0;
   $95 = $outgoingCommand;
   $96 = ((($95)) + 20|0);
   $97 = HEAP32[$96>>2]|0;
   $98 = ($94>>>0)>=($97>>>0);
   if ($98) {
    $99 = $1;
    $100 = ((($99)) + 48|0);
    $101 = HEAP32[$100>>2]|0;
    $102 = $2;
    $103 = ((($102)) + 84|0);
    $104 = HEAP32[$103>>2]|0;
    $105 = (($101) - ($104))|0;
    $106 = ($105>>>0)>=(86400000);
    if ($106) {
     $107 = $2;
     $108 = ((($107)) + 84|0);
     $109 = HEAP32[$108>>2]|0;
     $110 = $1;
     $111 = ((($110)) + 48|0);
     $112 = HEAP32[$111>>2]|0;
     $113 = (($109) - ($112))|0;
     $125 = $113;
    } else {
     $114 = $1;
     $115 = ((($114)) + 48|0);
     $116 = HEAP32[$115>>2]|0;
     $117 = $2;
     $118 = ((($117)) + 84|0);
     $119 = HEAP32[$118>>2]|0;
     $120 = (($116) - ($119))|0;
     $125 = $120;
    }
    $121 = $2;
    $122 = ((($121)) + 144|0);
    $123 = HEAP32[$122>>2]|0;
    $124 = ($125>>>0)>=($123>>>0);
    if ($124) {
     label = 20;
     break;
    }
   }
  }
  $129 = $outgoingCommand;
  $130 = ((($129)) + 80|0);
  $131 = HEAP32[$130>>2]|0;
  $132 = ($131|0)!=(0|0);
  if ($132) {
   $133 = $outgoingCommand;
   $134 = ((($133)) + 28|0);
   $135 = HEAP16[$134>>1]|0;
   $136 = $135&65535;
   $137 = $2;
   $138 = ((($137)) + 184|0);
   $139 = HEAP32[$138>>2]|0;
   $140 = (($139) - ($136))|0;
   HEAP32[$138>>2] = $140;
  }
  $141 = $2;
  $142 = ((($141)) + 96|0);
  $143 = HEAP32[$142>>2]|0;
  $144 = (($143) + 1)|0;
  HEAP32[$142>>2] = $144;
  $145 = $outgoingCommand;
  $146 = ((($145)) + 16|0);
  $147 = HEAP32[$146>>2]|0;
  $148 = $147<<1;
  HEAP32[$146>>2] = $148;
  $149 = $insertPosition;
  $150 = $outgoingCommand;
  $151 = (_enet_list_remove($150)|0);
  (_enet_list_insert($149,$151)|0);
  $152 = $currentCommand;
  $153 = $2;
  $154 = ((($153)) + 200|0);
  $155 = HEAP32[$154>>2]|0;
  $156 = ($152|0)==($155|0);
  if (!($156)) {
   continue;
  }
  $157 = $2;
  $158 = ((($157)) + 200|0);
  $159 = HEAP32[$158>>2]|0;
  $160 = $2;
  $161 = ((($160)) + 200|0);
  $162 = ($159|0)==($161|0);
  if ($162) {
   continue;
  }
  $163 = $currentCommand;
  $outgoingCommand = $163;
  $164 = $outgoingCommand;
  $165 = ((($164)) + 12|0);
  $166 = HEAP32[$165>>2]|0;
  $167 = $outgoingCommand;
  $168 = ((($167)) + 16|0);
  $169 = HEAP32[$168>>2]|0;
  $170 = (($166) + ($169))|0;
  $171 = $2;
  $172 = ((($171)) + 80|0);
  HEAP32[$172>>2] = $170;
 }
 if ((label|0) == 20) {
  $126 = $1;
  $127 = $2;
  $128 = $3;
  _enet_protocol_notify_disconnect($126,$127,$128);
  $0 = 1;
  $173 = $0;
  STACKTOP = sp;return ($173|0);
 }
 else if ((label|0) == 26) {
  $0 = 0;
  $173 = $0;
  STACKTOP = sp;return ($173|0);
 }
 return (0)|0;
}
function _enet_protocol_notify_disconnect($host,$peer,$event) {
 $host = $host|0;
 $peer = $peer|0;
 $event = $event|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $host;
 $1 = $peer;
 $2 = $event;
 $3 = $1;
 $4 = ((($3)) + 36|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5>>>0)>=(3);
 if ($6) {
  $7 = $0;
  $8 = ((($7)) + 32|0);
  HEAP32[$8>>2] = 1;
 }
 $9 = $1;
 $10 = ((($9)) + 36|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)!=(1);
 if ($12) {
  $13 = $1;
  $14 = ((($13)) + 36|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ($15>>>0)<(4);
  if ($16) {
   $17 = $1;
   _enet_peer_reset($17);
   STACKTOP = sp;return;
  }
 }
 $18 = $2;
 $19 = ($18|0)!=(0|0);
 if ($19) {
  $20 = $2;
  HEAP32[$20>>2] = 2;
  $21 = $1;
  $22 = $2;
  $23 = ((($22)) + 4|0);
  HEAP32[$23>>2] = $21;
  $24 = $2;
  $25 = ((($24)) + 12|0);
  HEAP32[$25>>2] = 0;
  $26 = $1;
  _enet_peer_reset($26);
  STACKTOP = sp;return;
 } else {
  $27 = $1;
  $28 = ((($27)) + 376|0);
  HEAP32[$28>>2] = 0;
  $29 = $0;
  $30 = $1;
  _enet_protocol_dispatch_state($29,$30,9);
  STACKTOP = sp;return;
 }
}
function _enet_protocol_send_reliable_outgoing_commands($host,$peer) {
 $host = $host|0;
 $peer = $peer|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0;
 var $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $buffer = 0, $canPing = 0, $channel = 0, $command = 0, $commandSize = 0, $currentCommand = 0, $outgoingCommand = 0, $reliableWindow = 0, $windowExceeded = 0, $windowSize = 0, $windowWrap = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $host;
 $1 = $peer;
 $2 = $0;
 $3 = ((($2)) + 1608|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $0;
 $6 = ((($5)) + 70|0);
 $7 = (($6) + (($4*48)|0)|0);
 $command = $7;
 $8 = $0;
 $9 = ((($8)) + 2132|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $0;
 $12 = ((($11)) + 1612|0);
 $13 = (($12) + ($10<<3)|0);
 $buffer = $13;
 $windowExceeded = 0;
 $windowWrap = 0;
 $canPing = 1;
 $14 = $1;
 $15 = ((($14)) + 216|0);
 $16 = HEAP32[$15>>2]|0;
 $currentCommand = $16;
 while(1) {
  $17 = $currentCommand;
  $18 = $1;
  $19 = ((($18)) + 216|0);
  $20 = ($17|0)!=($19|0);
  if (!($20)) {
   label = 39;
   break;
  }
  $21 = $currentCommand;
  $outgoingCommand = $21;
  $22 = $outgoingCommand;
  $23 = ((($22)) + 32|0);
  $24 = ((($23)) + 1|0);
  $25 = HEAP8[$24>>0]|0;
  $26 = $25&255;
  $27 = $1;
  $28 = ((($27)) + 44|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = ($26>>>0)<($29>>>0);
  if ($30) {
   $31 = $outgoingCommand;
   $32 = ((($31)) + 32|0);
   $33 = ((($32)) + 1|0);
   $34 = HEAP8[$33>>0]|0;
   $35 = $34&255;
   $36 = $1;
   $37 = ((($36)) + 40|0);
   $38 = HEAP32[$37>>2]|0;
   $39 = (($38) + (($35*60)|0)|0);
   $40 = $39;
  } else {
   $40 = 0;
  }
  $channel = $40;
  $41 = $outgoingCommand;
  $42 = ((($41)) + 8|0);
  $43 = HEAP16[$42>>1]|0;
  $44 = $43&65535;
  $45 = (($44|0) / 4096)&-1;
  $46 = $45&65535;
  $reliableWindow = $46;
  $47 = $channel;
  $48 = ($47|0)!=(0|0);
  if ($48) {
   $49 = $windowWrap;
   $50 = ($49|0)!=(0);
   do {
    if (!($50)) {
     $51 = $outgoingCommand;
     $52 = ((($51)) + 30|0);
     $53 = HEAP16[$52>>1]|0;
     $54 = $53&65535;
     $55 = ($54|0)<(1);
     if ($55) {
      $56 = $outgoingCommand;
      $57 = ((($56)) + 8|0);
      $58 = HEAP16[$57>>1]|0;
      $59 = $58&65535;
      $60 = (($59|0) % 4096)&-1;
      $61 = ($60|0)!=(0);
      if (!($61)) {
       $62 = $reliableWindow;
       $63 = $62&65535;
       $64 = (($63) + 16)|0;
       $65 = (($64) - 1)|0;
       $66 = (($65|0) % 16)&-1;
       $67 = $channel;
       $68 = ((($67)) + 6|0);
       $69 = (($68) + ($66<<1)|0);
       $70 = HEAP16[$69>>1]|0;
       $71 = $70&65535;
       $72 = ($71|0)>=(4096);
       if (!($72)) {
        $73 = $channel;
        $74 = ((($73)) + 4|0);
        $75 = HEAP16[$74>>1]|0;
        $76 = $75&65535;
        $77 = $reliableWindow;
        $78 = $77&65535;
        $79 = 255 << $78;
        $80 = $reliableWindow;
        $81 = $80&65535;
        $82 = (4096 - ($81))|0;
        $83 = 255 >> $82;
        $84 = $79 | $83;
        $85 = $76 & $84;
        $86 = ($85|0)!=(0);
        if (!($86)) {
         break;
        }
       }
       $windowWrap = 1;
      }
     }
    }
   } while(0);
   $87 = $windowWrap;
   $88 = ($87|0)!=(0);
   if ($88) {
    $89 = $currentCommand;
    $90 = HEAP32[$89>>2]|0;
    $currentCommand = $90;
    continue;
   }
  }
  $91 = $outgoingCommand;
  $92 = ((($91)) + 80|0);
  $93 = HEAP32[$92>>2]|0;
  $94 = ($93|0)!=(0|0);
  if ($94) {
   $95 = $windowExceeded;
   $96 = ($95|0)!=(0);
   if (!($96)) {
    $97 = $1;
    $98 = ((($97)) + 108|0);
    $99 = HEAP32[$98>>2]|0;
    $100 = $1;
    $101 = ((($100)) + 180|0);
    $102 = HEAP32[$101>>2]|0;
    $103 = Math_imul($99, $102)|0;
    $104 = (($103>>>0) / 32)&-1;
    $windowSize = $104;
    $105 = $1;
    $106 = ((($105)) + 184|0);
    $107 = HEAP32[$106>>2]|0;
    $108 = $outgoingCommand;
    $109 = ((($108)) + 28|0);
    $110 = HEAP16[$109>>1]|0;
    $111 = $110&65535;
    $112 = (($107) + ($111))|0;
    $113 = $windowSize;
    $114 = $1;
    $115 = ((($114)) + 176|0);
    $116 = HEAP32[$115>>2]|0;
    $117 = ($113>>>0)>($116>>>0);
    if ($117) {
     $118 = $windowSize;
     $123 = $118;
    } else {
     $119 = $1;
     $120 = ((($119)) + 176|0);
     $121 = HEAP32[$120>>2]|0;
     $123 = $121;
    }
    $122 = ($112>>>0)>($123>>>0);
    if ($122) {
     $windowExceeded = 1;
    }
   }
   $124 = $windowExceeded;
   $125 = ($124|0)!=(0);
   if ($125) {
    $126 = $currentCommand;
    $127 = HEAP32[$126>>2]|0;
    $currentCommand = $127;
    continue;
   }
  }
  $canPing = 0;
  $128 = $outgoingCommand;
  $129 = ((($128)) + 32|0);
  $130 = HEAP8[$129>>0]|0;
  $131 = $130&255;
  $132 = $131 & 15;
  $133 = (12128 + ($132<<2)|0);
  $134 = HEAP32[$133>>2]|0;
  $commandSize = $134;
  $135 = $command;
  $136 = $0;
  $137 = ((($136)) + 70|0);
  $138 = ((($137)) + 1536|0);
  $139 = ($135>>>0)>=($138>>>0);
  if ($139) {
   break;
  }
  $140 = $buffer;
  $141 = ((($140)) + 8|0);
  $142 = $0;
  $143 = ((($142)) + 1612|0);
  $144 = ((($143)) + 520|0);
  $145 = ($141>>>0)>=($144>>>0);
  if ($145) {
   break;
  }
  $146 = $1;
  $147 = ((($146)) + 176|0);
  $148 = HEAP32[$147>>2]|0;
  $149 = $0;
  $150 = ((($149)) + 64|0);
  $151 = HEAP32[$150>>2]|0;
  $152 = (($148) - ($151))|0;
  $153 = $commandSize;
  $154 = ($152>>>0)<($153>>>0);
  if ($154) {
   break;
  }
  $155 = $outgoingCommand;
  $156 = ((($155)) + 80|0);
  $157 = HEAP32[$156>>2]|0;
  $158 = ($157|0)!=(0|0);
  if ($158) {
   $159 = $1;
   $160 = ((($159)) + 176|0);
   $161 = HEAP32[$160>>2]|0;
   $162 = $0;
   $163 = ((($162)) + 64|0);
   $164 = HEAP32[$163>>2]|0;
   $165 = (($161) - ($164))|0;
   $166 = $165&65535;
   $167 = $166&65535;
   $168 = $commandSize;
   $169 = $outgoingCommand;
   $170 = ((($169)) + 28|0);
   $171 = HEAP16[$170>>1]|0;
   $172 = $171&65535;
   $173 = (($168) + ($172))|0;
   $174 = $173&65535;
   $175 = $174&65535;
   $176 = ($167|0)<($175|0);
   if ($176) {
    break;
   }
  }
  $179 = $currentCommand;
  $180 = HEAP32[$179>>2]|0;
  $currentCommand = $180;
  $181 = $channel;
  $182 = ($181|0)!=(0|0);
  if ($182) {
   $183 = $outgoingCommand;
   $184 = ((($183)) + 30|0);
   $185 = HEAP16[$184>>1]|0;
   $186 = $185&65535;
   $187 = ($186|0)<(1);
   if ($187) {
    $188 = $reliableWindow;
    $189 = $188&65535;
    $190 = 1 << $189;
    $191 = $channel;
    $192 = ((($191)) + 4|0);
    $193 = HEAP16[$192>>1]|0;
    $194 = $193&65535;
    $195 = $194 | $190;
    $196 = $195&65535;
    HEAP16[$192>>1] = $196;
    $197 = $reliableWindow;
    $198 = $197&65535;
    $199 = $channel;
    $200 = ((($199)) + 6|0);
    $201 = (($200) + ($198<<1)|0);
    $202 = HEAP16[$201>>1]|0;
    $203 = (($202) + 1)<<16>>16;
    HEAP16[$201>>1] = $203;
   }
  }
  $204 = $outgoingCommand;
  $205 = ((($204)) + 30|0);
  $206 = HEAP16[$205>>1]|0;
  $207 = (($206) + 1)<<16>>16;
  HEAP16[$205>>1] = $207;
  $208 = $outgoingCommand;
  $209 = ((($208)) + 16|0);
  $210 = HEAP32[$209>>2]|0;
  $211 = ($210|0)==(0);
  if ($211) {
   $212 = $1;
   $213 = ((($212)) + 168|0);
   $214 = HEAP32[$213>>2]|0;
   $215 = $1;
   $216 = ((($215)) + 172|0);
   $217 = HEAP32[$216>>2]|0;
   $218 = $217<<2;
   $219 = (($214) + ($218))|0;
   $220 = $outgoingCommand;
   $221 = ((($220)) + 16|0);
   HEAP32[$221>>2] = $219;
   $222 = $1;
   $223 = ((($222)) + 140|0);
   $224 = HEAP32[$223>>2]|0;
   $225 = $outgoingCommand;
   $226 = ((($225)) + 16|0);
   $227 = HEAP32[$226>>2]|0;
   $228 = Math_imul($224, $227)|0;
   $229 = $outgoingCommand;
   $230 = ((($229)) + 20|0);
   HEAP32[$230>>2] = $228;
  }
  $231 = $1;
  $232 = ((($231)) + 200|0);
  $233 = HEAP32[$232>>2]|0;
  $234 = $1;
  $235 = ((($234)) + 200|0);
  $236 = ($233|0)==($235|0);
  if ($236) {
   $237 = $0;
   $238 = ((($237)) + 48|0);
   $239 = HEAP32[$238>>2]|0;
   $240 = $outgoingCommand;
   $241 = ((($240)) + 16|0);
   $242 = HEAP32[$241>>2]|0;
   $243 = (($239) + ($242))|0;
   $244 = $1;
   $245 = ((($244)) + 80|0);
   HEAP32[$245>>2] = $243;
  }
  $246 = $1;
  $247 = ((($246)) + 200|0);
  $248 = $outgoingCommand;
  $249 = (_enet_list_remove($248)|0);
  (_enet_list_insert($247,$249)|0);
  $250 = $0;
  $251 = ((($250)) + 48|0);
  $252 = HEAP32[$251>>2]|0;
  $253 = $outgoingCommand;
  $254 = ((($253)) + 12|0);
  HEAP32[$254>>2] = $252;
  $255 = $command;
  $256 = $buffer;
  HEAP32[$256>>2] = $255;
  $257 = $commandSize;
  $258 = $buffer;
  $259 = ((($258)) + 4|0);
  HEAP32[$259>>2] = $257;
  $260 = $buffer;
  $261 = ((($260)) + 4|0);
  $262 = HEAP32[$261>>2]|0;
  $263 = $0;
  $264 = ((($263)) + 64|0);
  $265 = HEAP32[$264>>2]|0;
  $266 = (($265) + ($262))|0;
  HEAP32[$264>>2] = $266;
  $267 = $0;
  $268 = ((($267)) + 68|0);
  $269 = HEAP16[$268>>1]|0;
  $270 = $269&65535;
  $271 = $270 | 32768;
  $272 = $271&65535;
  HEAP16[$268>>1] = $272;
  $273 = $command;
  $274 = $outgoingCommand;
  $275 = ((($274)) + 32|0);
  dest=$273; src=$275; stop=dest+48|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
  $276 = $outgoingCommand;
  $277 = ((($276)) + 80|0);
  $278 = HEAP32[$277>>2]|0;
  $279 = ($278|0)!=(0|0);
  if ($279) {
   $280 = $buffer;
   $281 = ((($280)) + 8|0);
   $buffer = $281;
   $282 = $outgoingCommand;
   $283 = ((($282)) + 80|0);
   $284 = HEAP32[$283>>2]|0;
   $285 = ((($284)) + 8|0);
   $286 = HEAP32[$285>>2]|0;
   $287 = $outgoingCommand;
   $288 = ((($287)) + 24|0);
   $289 = HEAP32[$288>>2]|0;
   $290 = (($286) + ($289)|0);
   $291 = $buffer;
   HEAP32[$291>>2] = $290;
   $292 = $outgoingCommand;
   $293 = ((($292)) + 28|0);
   $294 = HEAP16[$293>>1]|0;
   $295 = $294&65535;
   $296 = $buffer;
   $297 = ((($296)) + 4|0);
   HEAP32[$297>>2] = $295;
   $298 = $outgoingCommand;
   $299 = ((($298)) + 28|0);
   $300 = HEAP16[$299>>1]|0;
   $301 = $300&65535;
   $302 = $0;
   $303 = ((($302)) + 64|0);
   $304 = HEAP32[$303>>2]|0;
   $305 = (($304) + ($301))|0;
   HEAP32[$303>>2] = $305;
   $306 = $outgoingCommand;
   $307 = ((($306)) + 28|0);
   $308 = HEAP16[$307>>1]|0;
   $309 = $308&65535;
   $310 = $1;
   $311 = ((($310)) + 184|0);
   $312 = HEAP32[$311>>2]|0;
   $313 = (($312) + ($309))|0;
   HEAP32[$311>>2] = $313;
  }
  $314 = $1;
  $315 = ((($314)) + 92|0);
  $316 = HEAP32[$315>>2]|0;
  $317 = (($316) + 1)|0;
  HEAP32[$315>>2] = $317;
  $318 = $command;
  $319 = ((($318)) + 48|0);
  $command = $319;
  $320 = $buffer;
  $321 = ((($320)) + 8|0);
  $buffer = $321;
 }
 if ((label|0) == 39) {
  $322 = $command;
  $323 = $0;
  $324 = ((($323)) + 70|0);
  $325 = $322;
  $326 = $324;
  $327 = (($325) - ($326))|0;
  $328 = (($327|0) / 48)&-1;
  $329 = $0;
  $330 = ((($329)) + 1608|0);
  HEAP32[$330>>2] = $328;
  $331 = $buffer;
  $332 = $0;
  $333 = ((($332)) + 1612|0);
  $334 = $331;
  $335 = $333;
  $336 = (($334) - ($335))|0;
  $337 = (($336|0) / 8)&-1;
  $338 = $0;
  $339 = ((($338)) + 2132|0);
  HEAP32[$339>>2] = $337;
  $340 = $canPing;
  STACKTOP = sp;return ($340|0);
 }
 $177 = $0;
 $178 = ((($177)) + 60|0);
 HEAP32[$178>>2] = 1;
 $322 = $command;
 $323 = $0;
 $324 = ((($323)) + 70|0);
 $325 = $322;
 $326 = $324;
 $327 = (($325) - ($326))|0;
 $328 = (($327|0) / 48)&-1;
 $329 = $0;
 $330 = ((($329)) + 1608|0);
 HEAP32[$330>>2] = $328;
 $331 = $buffer;
 $332 = $0;
 $333 = ((($332)) + 1612|0);
 $334 = $331;
 $335 = $333;
 $336 = (($334) - ($335))|0;
 $337 = (($336|0) / 8)&-1;
 $338 = $0;
 $339 = ((($338)) + 2132|0);
 HEAP32[$339>>2] = $337;
 $340 = $canPing;
 STACKTOP = sp;return ($340|0);
}
function _enet_protocol_send_unreliable_outgoing_commands($host,$peer) {
 $host = $host|0;
 $peer = $peer|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $buffer = 0, $command = 0, $commandSize = 0, $currentCommand = 0, $outgoingCommand = 0, $reliableSequenceNumber = 0, $unreliableSequenceNumber = 0;
 var dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $host;
 $1 = $peer;
 $2 = $0;
 $3 = ((($2)) + 1608|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $0;
 $6 = ((($5)) + 70|0);
 $7 = (($6) + (($4*48)|0)|0);
 $command = $7;
 $8 = $0;
 $9 = ((($8)) + 2132|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $0;
 $12 = ((($11)) + 1612|0);
 $13 = (($12) + ($10<<3)|0);
 $buffer = $13;
 $14 = $1;
 $15 = ((($14)) + 224|0);
 $16 = HEAP32[$15>>2]|0;
 $currentCommand = $16;
 L1: while(1) {
  $17 = $currentCommand;
  $18 = $1;
  $19 = ((($18)) + 224|0);
  $20 = ($17|0)!=($19|0);
  if (!($20)) {
   break;
  }
  $21 = $currentCommand;
  $outgoingCommand = $21;
  $22 = $outgoingCommand;
  $23 = ((($22)) + 32|0);
  $24 = HEAP8[$23>>0]|0;
  $25 = $24&255;
  $26 = $25 & 15;
  $27 = (12128 + ($26<<2)|0);
  $28 = HEAP32[$27>>2]|0;
  $commandSize = $28;
  $29 = $command;
  $30 = $0;
  $31 = ((($30)) + 70|0);
  $32 = ((($31)) + 1536|0);
  $33 = ($29>>>0)>=($32>>>0);
  if ($33) {
   label = 8;
   break;
  }
  $34 = $buffer;
  $35 = ((($34)) + 8|0);
  $36 = $0;
  $37 = ((($36)) + 1612|0);
  $38 = ((($37)) + 520|0);
  $39 = ($35>>>0)>=($38>>>0);
  if ($39) {
   label = 8;
   break;
  }
  $40 = $1;
  $41 = ((($40)) + 176|0);
  $42 = HEAP32[$41>>2]|0;
  $43 = $0;
  $44 = ((($43)) + 64|0);
  $45 = HEAP32[$44>>2]|0;
  $46 = (($42) - ($45))|0;
  $47 = $commandSize;
  $48 = ($46>>>0)<($47>>>0);
  if ($48) {
   label = 8;
   break;
  }
  $49 = $outgoingCommand;
  $50 = ((($49)) + 80|0);
  $51 = HEAP32[$50>>2]|0;
  $52 = ($51|0)!=(0|0);
  if ($52) {
   $53 = $1;
   $54 = ((($53)) + 176|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = $0;
   $57 = ((($56)) + 64|0);
   $58 = HEAP32[$57>>2]|0;
   $59 = (($55) - ($58))|0;
   $60 = $commandSize;
   $61 = $outgoingCommand;
   $62 = ((($61)) + 28|0);
   $63 = HEAP16[$62>>1]|0;
   $64 = $63&65535;
   $65 = (($60) + ($64))|0;
   $66 = ($59>>>0)<($65>>>0);
   if ($66) {
    label = 8;
    break;
   }
  }
  $69 = $currentCommand;
  $70 = HEAP32[$69>>2]|0;
  $currentCommand = $70;
  $71 = $outgoingCommand;
  $72 = ((($71)) + 80|0);
  $73 = HEAP32[$72>>2]|0;
  $74 = ($73|0)!=(0|0);
  if ($74) {
   $75 = $outgoingCommand;
   $76 = ((($75)) + 24|0);
   $77 = HEAP32[$76>>2]|0;
   $78 = ($77|0)==(0);
   if ($78) {
    $79 = $1;
    $80 = ((($79)) + 116|0);
    $81 = HEAP32[$80>>2]|0;
    $82 = (($81) + 7)|0;
    HEAP32[$80>>2] = $82;
    $83 = $1;
    $84 = ((($83)) + 116|0);
    $85 = HEAP32[$84>>2]|0;
    $86 = (($85>>>0) % 32)&-1;
    HEAP32[$84>>2] = $86;
    $87 = $1;
    $88 = ((($87)) + 116|0);
    $89 = HEAP32[$88>>2]|0;
    $90 = $1;
    $91 = ((($90)) + 108|0);
    $92 = HEAP32[$91>>2]|0;
    $93 = ($89>>>0)>($92>>>0);
    if ($93) {
     $94 = $outgoingCommand;
     $95 = ((($94)) + 8|0);
     $96 = HEAP16[$95>>1]|0;
     $reliableSequenceNumber = $96;
     $97 = $outgoingCommand;
     $98 = ((($97)) + 10|0);
     $99 = HEAP16[$98>>1]|0;
     $unreliableSequenceNumber = $99;
     while(1) {
      $100 = $outgoingCommand;
      $101 = ((($100)) + 80|0);
      $102 = HEAP32[$101>>2]|0;
      $103 = HEAP32[$102>>2]|0;
      $104 = (($103) + -1)|0;
      HEAP32[$102>>2] = $104;
      $105 = $outgoingCommand;
      $106 = ((($105)) + 80|0);
      $107 = HEAP32[$106>>2]|0;
      $108 = HEAP32[$107>>2]|0;
      $109 = ($108|0)==(0);
      if ($109) {
       $110 = $outgoingCommand;
       $111 = ((($110)) + 80|0);
       $112 = HEAP32[$111>>2]|0;
       _enet_packet_destroy($112);
      }
      $113 = $outgoingCommand;
      (_enet_list_remove($113)|0);
      $114 = $outgoingCommand;
      _enet_free($114);
      $115 = $currentCommand;
      $116 = $1;
      $117 = ((($116)) + 224|0);
      $118 = ($115|0)==($117|0);
      if ($118) {
       continue L1;
      }
      $119 = $currentCommand;
      $outgoingCommand = $119;
      $120 = $outgoingCommand;
      $121 = ((($120)) + 8|0);
      $122 = HEAP16[$121>>1]|0;
      $123 = $122&65535;
      $124 = $reliableSequenceNumber;
      $125 = $124&65535;
      $126 = ($123|0)!=($125|0);
      if ($126) {
       continue L1;
      }
      $127 = $outgoingCommand;
      $128 = ((($127)) + 10|0);
      $129 = HEAP16[$128>>1]|0;
      $130 = $129&65535;
      $131 = $unreliableSequenceNumber;
      $132 = $131&65535;
      $133 = ($130|0)!=($132|0);
      if ($133) {
       continue L1;
      }
      $134 = $currentCommand;
      $135 = HEAP32[$134>>2]|0;
      $currentCommand = $135;
     }
    }
   }
  }
  $136 = $command;
  $137 = $buffer;
  HEAP32[$137>>2] = $136;
  $138 = $commandSize;
  $139 = $buffer;
  $140 = ((($139)) + 4|0);
  HEAP32[$140>>2] = $138;
  $141 = $buffer;
  $142 = ((($141)) + 4|0);
  $143 = HEAP32[$142>>2]|0;
  $144 = $0;
  $145 = ((($144)) + 64|0);
  $146 = HEAP32[$145>>2]|0;
  $147 = (($146) + ($143))|0;
  HEAP32[$145>>2] = $147;
  $148 = $command;
  $149 = $outgoingCommand;
  $150 = ((($149)) + 32|0);
  dest=$148; src=$150; stop=dest+48|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
  $151 = $outgoingCommand;
  (_enet_list_remove($151)|0);
  $152 = $outgoingCommand;
  $153 = ((($152)) + 80|0);
  $154 = HEAP32[$153>>2]|0;
  $155 = ($154|0)!=(0|0);
  if ($155) {
   $156 = $buffer;
   $157 = ((($156)) + 8|0);
   $buffer = $157;
   $158 = $outgoingCommand;
   $159 = ((($158)) + 80|0);
   $160 = HEAP32[$159>>2]|0;
   $161 = ((($160)) + 8|0);
   $162 = HEAP32[$161>>2]|0;
   $163 = $outgoingCommand;
   $164 = ((($163)) + 24|0);
   $165 = HEAP32[$164>>2]|0;
   $166 = (($162) + ($165)|0);
   $167 = $buffer;
   HEAP32[$167>>2] = $166;
   $168 = $outgoingCommand;
   $169 = ((($168)) + 28|0);
   $170 = HEAP16[$169>>1]|0;
   $171 = $170&65535;
   $172 = $buffer;
   $173 = ((($172)) + 4|0);
   HEAP32[$173>>2] = $171;
   $174 = $buffer;
   $175 = ((($174)) + 4|0);
   $176 = HEAP32[$175>>2]|0;
   $177 = $0;
   $178 = ((($177)) + 64|0);
   $179 = HEAP32[$178>>2]|0;
   $180 = (($179) + ($176))|0;
   HEAP32[$178>>2] = $180;
   $181 = $1;
   $182 = ((($181)) + 208|0);
   $183 = $outgoingCommand;
   (_enet_list_insert($182,$183)|0);
  } else {
   $184 = $outgoingCommand;
   _enet_free($184);
  }
  $185 = $command;
  $186 = ((($185)) + 48|0);
  $command = $186;
  $187 = $buffer;
  $188 = ((($187)) + 8|0);
  $buffer = $188;
 }
 if ((label|0) == 8) {
  $67 = $0;
  $68 = ((($67)) + 60|0);
  HEAP32[$68>>2] = 1;
 }
 $189 = $command;
 $190 = $0;
 $191 = ((($190)) + 70|0);
 $192 = $189;
 $193 = $191;
 $194 = (($192) - ($193))|0;
 $195 = (($194|0) / 48)&-1;
 $196 = $0;
 $197 = ((($196)) + 1608|0);
 HEAP32[$197>>2] = $195;
 $198 = $buffer;
 $199 = $0;
 $200 = ((($199)) + 1612|0);
 $201 = $198;
 $202 = $200;
 $203 = (($201) - ($202))|0;
 $204 = (($203|0) / 8)&-1;
 $205 = $0;
 $206 = ((($205)) + 2132|0);
 HEAP32[$206>>2] = $204;
 $207 = $1;
 $208 = ((($207)) + 36|0);
 $209 = HEAP32[$208>>2]|0;
 $210 = ($209|0)==(6);
 if (!($210)) {
  STACKTOP = sp;return;
 }
 $211 = $1;
 $212 = ((($211)) + 216|0);
 $213 = HEAP32[$212>>2]|0;
 $214 = $1;
 $215 = ((($214)) + 216|0);
 $216 = ($213|0)==($215|0);
 if (!($216)) {
  STACKTOP = sp;return;
 }
 $217 = $1;
 $218 = ((($217)) + 224|0);
 $219 = HEAP32[$218>>2]|0;
 $220 = $1;
 $221 = ((($220)) + 224|0);
 $222 = ($219|0)==($221|0);
 if (!($222)) {
  STACKTOP = sp;return;
 }
 $223 = $1;
 $224 = ((($223)) + 200|0);
 $225 = HEAP32[$224>>2]|0;
 $226 = $1;
 $227 = ((($226)) + 200|0);
 $228 = ($225|0)==($227|0);
 if (!($228)) {
  STACKTOP = sp;return;
 }
 $229 = $1;
 $230 = $1;
 $231 = ((($230)) + 376|0);
 $232 = HEAP32[$231>>2]|0;
 _enet_peer_disconnect($229,$232);
 STACKTOP = sp;return;
}
function _enet_protocol_remove_sent_unreliable_commands($peer) {
 $peer = $peer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $outgoingCommand = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $peer;
 while(1) {
  $1 = $0;
  $2 = ((($1)) + 208|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = $0;
  $5 = ((($4)) + 208|0);
  $6 = ($3|0)==($5|0);
  $7 = $6 ^ 1;
  if (!($7)) {
   break;
  }
  $8 = $0;
  $9 = ((($8)) + 208|0);
  $10 = HEAP32[$9>>2]|0;
  $outgoingCommand = $10;
  $11 = $outgoingCommand;
  (_enet_list_remove($11)|0);
  $12 = $outgoingCommand;
  $13 = ((($12)) + 80|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ($14|0)!=(0|0);
  if ($15) {
   $16 = $outgoingCommand;
   $17 = ((($16)) + 80|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = HEAP32[$18>>2]|0;
   $20 = (($19) + -1)|0;
   HEAP32[$18>>2] = $20;
   $21 = $outgoingCommand;
   $22 = ((($21)) + 80|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = HEAP32[$23>>2]|0;
   $25 = ($24|0)==(0);
   if ($25) {
    $26 = $outgoingCommand;
    $27 = ((($26)) + 80|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = ((($28)) + 4|0);
    $30 = HEAP32[$29>>2]|0;
    $31 = $30 | 256;
    HEAP32[$29>>2] = $31;
    $32 = $outgoingCommand;
    $33 = ((($32)) + 80|0);
    $34 = HEAP32[$33>>2]|0;
    _enet_packet_destroy($34);
   }
  }
  $35 = $outgoingCommand;
  _enet_free($35);
 }
 STACKTOP = sp;return;
}
function _enet_host_check_events($host,$event) {
 $host = $host|0;
 $event = $event|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $host;
 $2 = $event;
 $3 = $2;
 $4 = ($3|0)==(0|0);
 if ($4) {
  $0 = -1;
  $13 = $0;
  STACKTOP = sp;return ($13|0);
 } else {
  $5 = $2;
  HEAP32[$5>>2] = 0;
  $6 = $2;
  $7 = ((($6)) + 4|0);
  HEAP32[$7>>2] = 0;
  $8 = $2;
  $9 = ((($8)) + 16|0);
  HEAP32[$9>>2] = 0;
  $10 = $1;
  $11 = $2;
  $12 = (_enet_protocol_dispatch_incoming_commands($10,$11)|0);
  $0 = $12;
  $13 = $0;
  STACKTOP = sp;return ($13|0);
 }
 return (0)|0;
}
function _enet_protocol_dispatch_incoming_commands($host,$event) {
 $host = $host|0;
 $event = $event|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $8 = 0, $9 = 0, $peer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $host;
 $2 = $event;
 L1: while(1) {
  $3 = $1;
  $4 = ((($3)) + 52|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = $1;
  $7 = ((($6)) + 52|0);
  $8 = ($5|0)==($7|0);
  $9 = $8 ^ 1;
  if (!($9)) {
   label = 11;
   break;
  }
  $10 = $1;
  $11 = ((($10)) + 52|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = (_enet_list_remove($12)|0);
  $peer = $13;
  $14 = $peer;
  $15 = ((($14)) + 240|0);
  HEAP32[$15>>2] = 0;
  $16 = $peer;
  $17 = ((($16)) + 36|0);
  $18 = HEAP32[$17>>2]|0;
  switch ($18|0) {
  case 4: case 3:  {
   label = 4;
   break L1;
   break;
  }
  case 9:  {
   label = 5;
   break L1;
   break;
  }
  case 5:  {
   break;
  }
  default: {
   continue L1;
  }
  }
  $42 = $peer;
  $43 = ((($42)) + 232|0);
  $44 = HEAP32[$43>>2]|0;
  $45 = $peer;
  $46 = ((($45)) + 232|0);
  $47 = ($44|0)==($46|0);
  if ($47) {
   continue;
  }
  $48 = $peer;
  $49 = $2;
  $50 = ((($49)) + 8|0);
  $51 = (_enet_peer_receive($48,$50)|0);
  $52 = $2;
  $53 = ((($52)) + 16|0);
  HEAP32[$53>>2] = $51;
  $54 = $2;
  $55 = ((($54)) + 16|0);
  $56 = HEAP32[$55>>2]|0;
  $57 = ($56|0)==(0|0);
  if (!($57)) {
   label = 8;
   break;
  }
 }
 if ((label|0) == 4) {
  $19 = $1;
  $20 = $peer;
  _enet_protocol_change_state($19,$20,5);
  $21 = $2;
  HEAP32[$21>>2] = 1;
  $22 = $peer;
  $23 = $2;
  $24 = ((($23)) + 4|0);
  HEAP32[$24>>2] = $22;
  $25 = $peer;
  $26 = ((($25)) + 376|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = $2;
  $29 = ((($28)) + 12|0);
  HEAP32[$29>>2] = $27;
  $0 = 1;
  $73 = $0;
  STACKTOP = sp;return ($73|0);
 }
 else if ((label|0) == 5) {
  $30 = $1;
  $31 = ((($30)) + 32|0);
  HEAP32[$31>>2] = 1;
  $32 = $2;
  HEAP32[$32>>2] = 2;
  $33 = $peer;
  $34 = $2;
  $35 = ((($34)) + 4|0);
  HEAP32[$35>>2] = $33;
  $36 = $peer;
  $37 = ((($36)) + 376|0);
  $38 = HEAP32[$37>>2]|0;
  $39 = $2;
  $40 = ((($39)) + 12|0);
  HEAP32[$40>>2] = $38;
  $41 = $peer;
  _enet_peer_reset($41);
  $0 = 1;
  $73 = $0;
  STACKTOP = sp;return ($73|0);
 }
 else if ((label|0) == 8) {
  $58 = $2;
  HEAP32[$58>>2] = 3;
  $59 = $peer;
  $60 = $2;
  $61 = ((($60)) + 4|0);
  HEAP32[$61>>2] = $59;
  $62 = $peer;
  $63 = ((($62)) + 232|0);
  $64 = HEAP32[$63>>2]|0;
  $65 = $peer;
  $66 = ((($65)) + 232|0);
  $67 = ($64|0)==($66|0);
  if (!($67)) {
   $68 = $peer;
   $69 = ((($68)) + 240|0);
   HEAP32[$69>>2] = 1;
   $70 = $1;
   $71 = ((($70)) + 52|0);
   $72 = $peer;
   (_enet_list_insert($71,$72)|0);
  }
  $0 = 1;
  $73 = $0;
  STACKTOP = sp;return ($73|0);
 }
 else if ((label|0) == 11) {
  $0 = 0;
  $73 = $0;
  STACKTOP = sp;return ($73|0);
 }
 return (0)|0;
}
function _enet_host_service($host,$event,$timeout) {
 $host = $host|0;
 $event = $event|0;
 $timeout = $timeout|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $waitCondition = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $waitCondition = sp;
 $1 = $host;
 $2 = $event;
 $3 = $timeout;
 $4 = $2;
 $5 = ($4|0)!=(0|0);
 L1: do {
  if ($5) {
   $6 = $2;
   HEAP32[$6>>2] = 0;
   $7 = $2;
   $8 = ((($7)) + 4|0);
   HEAP32[$8>>2] = 0;
   $9 = $2;
   $10 = ((($9)) + 16|0);
   HEAP32[$10>>2] = 0;
   $11 = $1;
   $12 = $2;
   $13 = (_enet_protocol_dispatch_incoming_commands($11,$12)|0);
   switch ($13|0) {
   case 1:  {
    $0 = 1;
    break L1;
    break;
   }
   case -1:  {
    $0 = -1;
    break L1;
    break;
   }
   default: {
    label = 5;
    break L1;
   }
   }
  } else {
   label = 5;
  }
 } while(0);
 L6: do {
  if ((label|0) == 5) {
   $14 = (_enet_time_get()|0);
   $15 = $1;
   $16 = ((($15)) + 48|0);
   HEAP32[$16>>2] = $14;
   $17 = $1;
   $18 = ((($17)) + 48|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = $3;
   $21 = (($20) + ($19))|0;
   $3 = $21;
   L8: while(1) {
    $22 = $1;
    $23 = ((($22)) + 48|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = $1;
    $26 = ((($25)) + 20|0);
    $27 = HEAP32[$26>>2]|0;
    $28 = (($24) - ($27))|0;
    $29 = ($28>>>0)>=(86400000);
    $30 = $1;
    if ($29) {
     $31 = ((($30)) + 20|0);
     $32 = HEAP32[$31>>2]|0;
     $33 = $1;
     $34 = ((($33)) + 48|0);
     $35 = HEAP32[$34>>2]|0;
     $36 = (($32) - ($35))|0;
     $44 = $36;
    } else {
     $37 = ((($30)) + 48|0);
     $38 = HEAP32[$37>>2]|0;
     $39 = $1;
     $40 = ((($39)) + 20|0);
     $41 = HEAP32[$40>>2]|0;
     $42 = (($38) - ($41))|0;
     $44 = $42;
    }
    $43 = ($44>>>0)>=(1000);
    if ($43) {
     $45 = $1;
     _enet_host_bandwidth_throttle($45);
    }
    $46 = $1;
    $47 = $2;
    $48 = (_enet_protocol_send_outgoing_commands($46,$47,1)|0);
    switch ($48|0) {
    case 1:  {
     label = 12;
     break L8;
     break;
    }
    case -1:  {
     label = 13;
     break L8;
     break;
    }
    default: {
    }
    }
    $49 = $1;
    $50 = $2;
    $51 = (_enet_protocol_receive_incoming_commands($49,$50)|0);
    switch ($51|0) {
    case 1:  {
     label = 15;
     break L8;
     break;
    }
    case -1:  {
     label = 16;
     break L8;
     break;
    }
    default: {
    }
    }
    $52 = $1;
    $53 = $2;
    $54 = (_enet_protocol_send_outgoing_commands($52,$53,1)|0);
    switch ($54|0) {
    case 1:  {
     label = 18;
     break L8;
     break;
    }
    case -1:  {
     label = 19;
     break L8;
     break;
    }
    default: {
    }
    }
    $55 = $2;
    $56 = ($55|0)!=(0|0);
    if ($56) {
     $57 = $1;
     $58 = $2;
     $59 = (_enet_protocol_dispatch_incoming_commands($57,$58)|0);
     switch ($59|0) {
     case 1:  {
      label = 22;
      break L8;
      break;
     }
     case -1:  {
      label = 23;
      break L8;
      break;
     }
     default: {
     }
     }
    }
    while(1) {
     $60 = (_enet_time_get()|0);
     $61 = $1;
     $62 = ((($61)) + 48|0);
     HEAP32[$62>>2] = $60;
     $63 = $1;
     $64 = ((($63)) + 48|0);
     $65 = HEAP32[$64>>2]|0;
     $66 = $3;
     $67 = (($65) - ($66))|0;
     $68 = ($67>>>0)>=(86400000);
     if (!($68)) {
      label = 25;
      break L8;
     }
     HEAP32[$waitCondition>>2] = 6;
     $69 = $1;
     $70 = HEAP32[$69>>2]|0;
     $71 = $3;
     $72 = $1;
     $73 = ((($72)) + 48|0);
     $74 = HEAP32[$73>>2]|0;
     $75 = (($71) - ($74))|0;
     $76 = ($75>>>0)>=(86400000);
     if ($76) {
      $77 = $1;
      $78 = ((($77)) + 48|0);
      $79 = HEAP32[$78>>2]|0;
      $80 = $3;
      $81 = (($79) - ($80))|0;
      $87 = $81;
     } else {
      $82 = $3;
      $83 = $1;
      $84 = ((($83)) + 48|0);
      $85 = HEAP32[$84>>2]|0;
      $86 = (($82) - ($85))|0;
      $87 = $86;
     }
     $88 = (_enet_socket_wait($70,$waitCondition,$87)|0);
     $89 = ($88|0)!=(0);
     if ($89) {
      label = 30;
      break L8;
     }
     $90 = HEAP32[$waitCondition>>2]|0;
     $91 = $90 & 4;
     $92 = ($91|0)!=(0);
     if (!($92)) {
      break;
     }
    }
    $93 = (_enet_time_get()|0);
    $94 = $1;
    $95 = ((($94)) + 48|0);
    HEAP32[$95>>2] = $93;
    $96 = HEAP32[$waitCondition>>2]|0;
    $97 = $96 & 2;
    $98 = ($97|0)!=(0);
    if (!($98)) {
     label = 33;
     break;
    }
   }
   switch (label|0) {
    case 12: {
     $0 = 1;
     break L6;
     break;
    }
    case 13: {
     $0 = -1;
     break L6;
     break;
    }
    case 15: {
     $0 = 1;
     break L6;
     break;
    }
    case 16: {
     $0 = -1;
     break L6;
     break;
    }
    case 18: {
     $0 = 1;
     break L6;
     break;
    }
    case 19: {
     $0 = -1;
     break L6;
     break;
    }
    case 22: {
     $0 = 1;
     break L6;
     break;
    }
    case 23: {
     $0 = -1;
     break L6;
     break;
    }
    case 25: {
     $0 = 0;
     break L6;
     break;
    }
    case 30: {
     $0 = -1;
     break L6;
     break;
    }
    case 33: {
     $0 = 0;
     break L6;
     break;
    }
   }
  }
 } while(0);
 $99 = $0;
 STACKTOP = sp;return ($99|0);
}
function _enet_protocol_receive_incoming_commands($host,$event) {
 $host = $host|0;
 $event = $event|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buffer = 0, $receivedLength = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $buffer = sp;
 $1 = $host;
 $2 = $event;
 L1: while(1) {
  $3 = $1;
  $4 = ((($3)) + 2156|0);
  HEAP32[$buffer>>2] = $4;
  $5 = ((($buffer)) + 4|0);
  HEAP32[$5>>2] = 4096;
  $6 = $1;
  $7 = HEAP32[$6>>2]|0;
  $8 = $1;
  $9 = ((($8)) + 10348|0);
  $10 = (_enet_socket_receive($7,$9,$buffer,1)|0);
  $receivedLength = $10;
  $11 = $receivedLength;
  $12 = ($11|0)<(0);
  if ($12) {
   label = 3;
   break;
  }
  $13 = $receivedLength;
  $14 = ($13|0)==(0);
  if ($14) {
   label = 5;
   break;
  }
  $15 = $1;
  $16 = ((($15)) + 2156|0);
  $17 = $1;
  $18 = ((($17)) + 10356|0);
  HEAP32[$18>>2] = $16;
  $19 = $receivedLength;
  $20 = $1;
  $21 = ((($20)) + 10360|0);
  HEAP32[$21>>2] = $19;
  $22 = $receivedLength;
  $23 = $1;
  $24 = ((($23)) + 10372|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = (($25) + ($22))|0;
  HEAP32[$24>>2] = $26;
  $27 = $1;
  $28 = ((($27)) + 10376|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = (($29) + 1)|0;
  HEAP32[$28>>2] = $30;
  $31 = $1;
  $32 = ((($31)) + 10380|0);
  $33 = HEAP32[$32>>2]|0;
  $34 = ($33|0)!=(0|0);
  L5: do {
   if ($34) {
    $35 = $1;
    $36 = ((($35)) + 10380|0);
    $37 = HEAP32[$36>>2]|0;
    $38 = $1;
    $39 = $2;
    $40 = (FUNCTION_TABLE_iii[$37 & 8191]($38,$39)|0);
    switch ($40|0) {
    case -1:  {
     label = 11;
     break L1;
     break;
    }
    case 1:  {
     break;
    }
    default: {
     break L5;
    }
    }
    $41 = $2;
    $42 = ($41|0)!=(0|0);
    if (!($42)) {
     continue L1;
    }
    $43 = $2;
    $44 = HEAP32[$43>>2]|0;
    $45 = ($44|0)!=(0);
    if ($45) {
     label = 10;
     break L1;
    } else {
     continue L1;
    }
   }
  } while(0);
  $46 = $1;
  $47 = $2;
  $48 = (_enet_protocol_handle_incoming_commands($46,$47)|0);
  switch ($48|0) {
  case 1:  {
   label = 13;
   break L1;
   break;
  }
  case -1:  {
   label = 14;
   break L1;
   break;
  }
  default: {
  }
  }
 }
 if ((label|0) == 3) {
  $0 = -1;
  $49 = $0;
  STACKTOP = sp;return ($49|0);
 }
 else if ((label|0) == 5) {
  $0 = 0;
  $49 = $0;
  STACKTOP = sp;return ($49|0);
 }
 else if ((label|0) == 10) {
  $0 = 1;
  $49 = $0;
  STACKTOP = sp;return ($49|0);
 }
 else if ((label|0) == 11) {
  $0 = -1;
  $49 = $0;
  STACKTOP = sp;return ($49|0);
 }
 else if ((label|0) == 13) {
  $0 = 1;
  $49 = $0;
  STACKTOP = sp;return ($49|0);
 }
 else if ((label|0) == 14) {
  $0 = -1;
  $49 = $0;
  STACKTOP = sp;return ($49|0);
 }
 return (0)|0;
}
function _enet_protocol_handle_incoming_commands($host,$event) {
 $host = $host|0;
 $event = $event|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0;
 var $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0;
 var $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0;
 var $369 = 0, $37 = 0, $370 = 0, $371 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $buffer = 0, $checksum = 0, $command = 0, $commandNumber = 0, $commandSize = 0, $currentData = 0, $desiredChecksum = 0, $flags = 0;
 var $header = 0, $headerSize = 0, $originalSize = 0, $peer = 0, $peerID = 0, $sentTime = 0, $sessionID = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $currentData = sp + 32|0;
 $buffer = sp + 8|0;
 $1 = $host;
 $2 = $event;
 $3 = $1;
 $4 = ((($3)) + 10360|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5>>>0)<(((((0) + 2|0)))>>>0);
 if ($6) {
  $0 = 0;
  $371 = $0;
  STACKTOP = sp;return ($371|0);
 }
 $7 = $1;
 $8 = ((($7)) + 10356|0);
 $9 = HEAP32[$8>>2]|0;
 $header = $9;
 $10 = $header;
 $11 = HEAPU8[$10>>0]|(HEAPU8[$10+1>>0]<<8);
 $12 = (_ntohs($11)|0);
 $peerID = $12;
 $13 = $peerID;
 $14 = $13&65535;
 $15 = $14 & 12288;
 $16 = $15 >> 12;
 $17 = $16&255;
 $sessionID = $17;
 $18 = $peerID;
 $19 = $18&65535;
 $20 = $19 & 49152;
 $21 = $20&65535;
 $flags = $21;
 $22 = $peerID;
 $23 = $22&65535;
 $24 = $23 & -61441;
 $25 = $24&65535;
 $peerID = $25;
 $26 = $flags;
 $27 = $26&65535;
 $28 = $27 & 32768;
 $29 = ($28|0)!=(0);
 $30 = $29 ? 4 : ((((0) + 2|0)));
 $headerSize = $30;
 $31 = $1;
 $32 = ((($31)) + 2136|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ($33|0)!=(0|0);
 if ($34) {
  $35 = $headerSize;
  $36 = (($35) + 4)|0;
  $headerSize = $36;
 }
 $37 = $peerID;
 $38 = $37&65535;
 $39 = ($38|0)==(4095);
 L8: do {
  if ($39) {
   $peer = 0;
  } else {
   $40 = $peerID;
   $41 = $40&65535;
   $42 = $1;
   $43 = ((($42)) + 40|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = ($41>>>0)>=($44>>>0);
   if ($45) {
    $0 = 0;
    $371 = $0;
    STACKTOP = sp;return ($371|0);
   }
   $46 = $peerID;
   $47 = $46&65535;
   $48 = $1;
   $49 = ((($48)) + 36|0);
   $50 = HEAP32[$49>>2]|0;
   $51 = (($50) + (($47*380)|0)|0);
   $peer = $51;
   $52 = $peer;
   $53 = ((($52)) + 36|0);
   $54 = HEAP32[$53>>2]|0;
   $55 = ($54|0)==(0);
   do {
    if (!($55)) {
     $56 = $peer;
     $57 = ((($56)) + 36|0);
     $58 = HEAP32[$57>>2]|0;
     $59 = ($58|0)==(9);
     if (!($59)) {
      $60 = $1;
      $61 = ((($60)) + 10348|0);
      $62 = HEAP32[$61>>2]|0;
      $63 = $peer;
      $64 = ((($63)) + 24|0);
      $65 = HEAP32[$64>>2]|0;
      $66 = ($62|0)!=($65|0);
      if ($66) {
       label = 13;
      } else {
       $67 = $1;
       $68 = ((($67)) + 10348|0);
       $69 = ((($68)) + 4|0);
       $70 = HEAP16[$69>>1]|0;
       $71 = $70&65535;
       $72 = $peer;
       $73 = ((($72)) + 24|0);
       $74 = ((($73)) + 4|0);
       $75 = HEAP16[$74>>1]|0;
       $76 = $75&65535;
       $77 = ($71|0)!=($76|0);
       if ($77) {
        label = 13;
       }
      }
      if ((label|0) == 13) {
       $78 = $peer;
       $79 = ((($78)) + 24|0);
       $80 = HEAP32[$79>>2]|0;
       $81 = ($80|0)!=(-1);
       if ($81) {
        break;
       }
      }
      $82 = $peer;
      $83 = ((($82)) + 12|0);
      $84 = HEAP16[$83>>1]|0;
      $85 = $84&65535;
      $86 = ($85|0)<(4095);
      if (!($86)) {
       break L8;
      }
      $87 = $sessionID;
      $88 = $87&255;
      $89 = $peer;
      $90 = ((($89)) + 21|0);
      $91 = HEAP8[$90>>0]|0;
      $92 = $91&255;
      $93 = ($88|0)!=($92|0);
      if (!($93)) {
       break L8;
      }
     }
    }
   } while(0);
   $0 = 0;
   $371 = $0;
   STACKTOP = sp;return ($371|0);
  }
 } while(0);
 $94 = $flags;
 $95 = $94&65535;
 $96 = $95 & 16384;
 $97 = ($96|0)!=(0);
 do {
  if ($97) {
   $98 = $1;
   $99 = ((($98)) + 2140|0);
   $100 = HEAP32[$99>>2]|0;
   $101 = ($100|0)==(0|0);
   if (!($101)) {
    $102 = $1;
    $103 = ((($102)) + 2140|0);
    $104 = ((($103)) + 8|0);
    $105 = HEAP32[$104>>2]|0;
    $106 = ($105|0)==(0|0);
    if (!($106)) {
     $107 = $1;
     $108 = ((($107)) + 2140|0);
     $109 = ((($108)) + 8|0);
     $110 = HEAP32[$109>>2]|0;
     $111 = $1;
     $112 = ((($111)) + 2140|0);
     $113 = HEAP32[$112>>2]|0;
     $114 = $1;
     $115 = ((($114)) + 10356|0);
     $116 = HEAP32[$115>>2]|0;
     $117 = $headerSize;
     $118 = (($116) + ($117)|0);
     $119 = $1;
     $120 = ((($119)) + 10360|0);
     $121 = HEAP32[$120>>2]|0;
     $122 = $headerSize;
     $123 = (($121) - ($122))|0;
     $124 = $1;
     $125 = ((($124)) + 2156|0);
     $126 = ((($125)) + 4096|0);
     $127 = $headerSize;
     $128 = (($126) + ($127)|0);
     $129 = $headerSize;
     $130 = (4096 - ($129))|0;
     $131 = (FUNCTION_TABLE_iiiiii[$110 & 8191]($113,$118,$123,$128,$130)|0);
     $originalSize = $131;
     $132 = $originalSize;
     $133 = ($132>>>0)<=(0);
     if (!($133)) {
      $134 = $originalSize;
      $135 = $headerSize;
      $136 = (4096 - ($135))|0;
      $137 = ($134>>>0)>($136>>>0);
      if (!($137)) {
       $138 = $1;
       $139 = ((($138)) + 2156|0);
       $140 = ((($139)) + 4096|0);
       $141 = $header;
       $142 = $headerSize;
       _memcpy(($140|0),($141|0),($142|0))|0;
       $143 = $1;
       $144 = ((($143)) + 2156|0);
       $145 = ((($144)) + 4096|0);
       $146 = $1;
       $147 = ((($146)) + 10356|0);
       HEAP32[$147>>2] = $145;
       $148 = $headerSize;
       $149 = $originalSize;
       $150 = (($148) + ($149))|0;
       $151 = $1;
       $152 = ((($151)) + 10360|0);
       HEAP32[$152>>2] = $150;
       break;
      }
     }
     $0 = 0;
     $371 = $0;
     STACKTOP = sp;return ($371|0);
    }
   }
   $0 = 0;
   $371 = $0;
   STACKTOP = sp;return ($371|0);
  }
 } while(0);
 $153 = $1;
 $154 = ((($153)) + 2136|0);
 $155 = HEAP32[$154>>2]|0;
 $156 = ($155|0)!=(0|0);
 if ($156) {
  $157 = $headerSize;
  $158 = (($157) - 4)|0;
  $159 = $1;
  $160 = ((($159)) + 10356|0);
  $161 = HEAP32[$160>>2]|0;
  $162 = (($161) + ($158)|0);
  $checksum = $162;
  $163 = $checksum;
  $164 = HEAP32[$163>>2]|0;
  $desiredChecksum = $164;
  $165 = $peer;
  $166 = ($165|0)!=(0|0);
  if ($166) {
   $167 = $peer;
   $168 = ((($167)) + 16|0);
   $169 = HEAP32[$168>>2]|0;
   $171 = $169;
  } else {
   $171 = 0;
  }
  $170 = $checksum;
  HEAP32[$170>>2] = $171;
  $172 = $1;
  $173 = ((($172)) + 10356|0);
  $174 = HEAP32[$173>>2]|0;
  HEAP32[$buffer>>2] = $174;
  $175 = $1;
  $176 = ((($175)) + 10360|0);
  $177 = HEAP32[$176>>2]|0;
  $178 = ((($buffer)) + 4|0);
  HEAP32[$178>>2] = $177;
  $179 = $1;
  $180 = ((($179)) + 2136|0);
  $181 = HEAP32[$180>>2]|0;
  $182 = (FUNCTION_TABLE_iii[$181 & 8191]($buffer,1)|0);
  $183 = $desiredChecksum;
  $184 = ($182|0)!=($183|0);
  if ($184) {
   $0 = 0;
   $371 = $0;
   STACKTOP = sp;return ($371|0);
  }
 }
 $185 = $peer;
 $186 = ($185|0)!=(0|0);
 if ($186) {
  $187 = $1;
  $188 = ((($187)) + 10348|0);
  $189 = HEAP32[$188>>2]|0;
  $190 = $peer;
  $191 = ((($190)) + 24|0);
  HEAP32[$191>>2] = $189;
  $192 = $1;
  $193 = ((($192)) + 10348|0);
  $194 = ((($193)) + 4|0);
  $195 = HEAP16[$194>>1]|0;
  $196 = $peer;
  $197 = ((($196)) + 24|0);
  $198 = ((($197)) + 4|0);
  HEAP16[$198>>1] = $195;
  $199 = $1;
  $200 = ((($199)) + 10360|0);
  $201 = HEAP32[$200>>2]|0;
  $202 = $peer;
  $203 = ((($202)) + 64|0);
  $204 = HEAP32[$203>>2]|0;
  $205 = (($204) + ($201))|0;
  HEAP32[$203>>2] = $205;
 }
 $206 = $1;
 $207 = ((($206)) + 10356|0);
 $208 = HEAP32[$207>>2]|0;
 $209 = $headerSize;
 $210 = (($208) + ($209)|0);
 HEAP32[$currentData>>2] = $210;
 L51: while(1) {
  $211 = HEAP32[$currentData>>2]|0;
  $212 = $1;
  $213 = ((($212)) + 10360|0);
  $214 = HEAP32[$213>>2]|0;
  $215 = $1;
  $216 = ((($215)) + 10356|0);
  $217 = HEAP32[$216>>2]|0;
  $218 = (($217) + ($214)|0);
  $219 = ($211>>>0)<($218>>>0);
  if (!($219)) {
   break;
  }
  $220 = HEAP32[$currentData>>2]|0;
  $command = $220;
  $221 = HEAP32[$currentData>>2]|0;
  $222 = ((($221)) + 4|0);
  $223 = $1;
  $224 = ((($223)) + 10360|0);
  $225 = HEAP32[$224>>2]|0;
  $226 = $1;
  $227 = ((($226)) + 10356|0);
  $228 = HEAP32[$227>>2]|0;
  $229 = (($228) + ($225)|0);
  $230 = ($222>>>0)>($229>>>0);
  if ($230) {
   break;
  }
  $231 = $command;
  $232 = HEAP8[$231>>0]|0;
  $233 = $232&255;
  $234 = $233 & 15;
  $235 = $234&255;
  $commandNumber = $235;
  $236 = $commandNumber;
  $237 = $236&255;
  $238 = ($237|0)>=(13);
  if ($238) {
   break;
  }
  $239 = $commandNumber;
  $240 = $239&255;
  $241 = (12128 + ($240<<2)|0);
  $242 = HEAP32[$241>>2]|0;
  $commandSize = $242;
  $243 = $commandSize;
  $244 = ($243|0)==(0);
  if ($244) {
   break;
  }
  $245 = HEAP32[$currentData>>2]|0;
  $246 = $commandSize;
  $247 = (($245) + ($246)|0);
  $248 = $1;
  $249 = ((($248)) + 10360|0);
  $250 = HEAP32[$249>>2]|0;
  $251 = $1;
  $252 = ((($251)) + 10356|0);
  $253 = HEAP32[$252>>2]|0;
  $254 = (($253) + ($250)|0);
  $255 = ($247>>>0)>($254>>>0);
  if ($255) {
   break;
  }
  $256 = $commandSize;
  $257 = HEAP32[$currentData>>2]|0;
  $258 = (($257) + ($256)|0);
  HEAP32[$currentData>>2] = $258;
  $259 = $peer;
  $260 = ($259|0)==(0|0);
  if ($260) {
   $261 = $commandNumber;
   $262 = $261&255;
   $263 = ($262|0)!=(2);
   if ($263) {
    break;
   }
  }
  $264 = $command;
  $265 = ((($264)) + 2|0);
  $266 = HEAPU8[$265>>0]|(HEAPU8[$265+1>>0]<<8);
  $267 = (_ntohs($266)|0);
  $268 = $command;
  $269 = ((($268)) + 2|0);
  HEAP8[$269>>0]=$267&255;HEAP8[$269+1>>0]=$267>>8;
  $270 = $commandNumber;
  $271 = $270&255;
  do {
   switch ($271|0) {
   case 1:  {
    $272 = $1;
    $273 = $2;
    $274 = $peer;
    $275 = $command;
    $276 = (_enet_protocol_handle_acknowledge($272,$273,$274,$275)|0);
    $277 = ($276|0)!=(0);
    if ($277) {
     break L51;
    }
    break;
   }
   case 2:  {
    $278 = $peer;
    $279 = ($278|0)!=(0|0);
    if ($279) {
     break L51;
    }
    $280 = $1;
    $281 = $header;
    $282 = $command;
    $283 = (_enet_protocol_handle_connect($280,$281,$282)|0);
    $peer = $283;
    $284 = $peer;
    $285 = ($284|0)==(0|0);
    if ($285) {
     break L51;
    }
    break;
   }
   case 3:  {
    $286 = $1;
    $287 = $2;
    $288 = $peer;
    $289 = $command;
    $290 = (_enet_protocol_handle_verify_connect($286,$287,$288,$289)|0);
    $291 = ($290|0)!=(0);
    if ($291) {
     break L51;
    }
    break;
   }
   case 4:  {
    $292 = $1;
    $293 = $peer;
    $294 = $command;
    $295 = (_enet_protocol_handle_disconnect($292,$293,$294)|0);
    $296 = ($295|0)!=(0);
    if ($296) {
     break L51;
    }
    break;
   }
   case 5:  {
    $297 = $1;
    $298 = $peer;
    $299 = $command;
    $300 = (_enet_protocol_handle_ping($297,$298,$299)|0);
    $301 = ($300|0)!=(0);
    if ($301) {
     break L51;
    }
    break;
   }
   case 6:  {
    $302 = $1;
    $303 = $peer;
    $304 = $command;
    $305 = (_enet_protocol_handle_send_reliable($302,$303,$304,$currentData)|0);
    $306 = ($305|0)!=(0);
    if ($306) {
     break L51;
    }
    break;
   }
   case 7:  {
    $307 = $1;
    $308 = $peer;
    $309 = $command;
    $310 = (_enet_protocol_handle_send_unreliable($307,$308,$309,$currentData)|0);
    $311 = ($310|0)!=(0);
    if ($311) {
     break L51;
    }
    break;
   }
   case 9:  {
    $312 = $1;
    $313 = $peer;
    $314 = $command;
    $315 = (_enet_protocol_handle_send_unsequenced($312,$313,$314,$currentData)|0);
    $316 = ($315|0)!=(0);
    if ($316) {
     break L51;
    }
    break;
   }
   case 8:  {
    $317 = $1;
    $318 = $peer;
    $319 = $command;
    $320 = (_enet_protocol_handle_send_fragment($317,$318,$319,$currentData)|0);
    $321 = ($320|0)!=(0);
    if ($321) {
     break L51;
    }
    break;
   }
   case 10:  {
    $322 = $1;
    $323 = $peer;
    $324 = $command;
    $325 = (_enet_protocol_handle_bandwidth_limit($322,$323,$324)|0);
    $326 = ($325|0)!=(0);
    if ($326) {
     break L51;
    }
    break;
   }
   case 11:  {
    $327 = $1;
    $328 = $peer;
    $329 = $command;
    $330 = (_enet_protocol_handle_throttle_configure($327,$328,$329)|0);
    $331 = ($330|0)!=(0);
    if ($331) {
     break L51;
    }
    break;
   }
   case 12:  {
    $332 = $1;
    $333 = $peer;
    $334 = $command;
    $335 = (_enet_protocol_handle_send_unreliable_fragment($332,$333,$334,$currentData)|0);
    $336 = ($335|0)!=(0);
    if ($336) {
     break L51;
    }
    break;
   }
   default: {
    break L51;
   }
   }
  } while(0);
  $337 = $peer;
  $338 = ($337|0)!=(0|0);
  if (!($338)) {
   continue;
  }
  $339 = $command;
  $340 = HEAP8[$339>>0]|0;
  $341 = $340&255;
  $342 = $341 & 128;
  $343 = ($342|0)!=(0);
  if (!($343)) {
   continue;
  }
  $344 = $flags;
  $345 = $344&65535;
  $346 = $345 & 32768;
  $347 = ($346|0)!=(0);
  if (!($347)) {
   break;
  }
  $348 = $header;
  $349 = ((($348)) + 2|0);
  $350 = HEAPU8[$349>>0]|(HEAPU8[$349+1>>0]<<8);
  $351 = (_ntohs($350)|0);
  $sentTime = $351;
  $352 = $peer;
  $353 = ((($352)) + 36|0);
  $354 = HEAP32[$353>>2]|0;
  switch ($354|0) {
  case 9: case 0: case 2: case 7:  {
   continue L51;
   break;
  }
  case 8:  {
   break;
  }
  default: {
   $363 = $peer;
   $364 = $command;
   $365 = $sentTime;
   (_enet_peer_queue_acknowledgement($363,$364,$365)|0);
   continue L51;
  }
  }
  $355 = $command;
  $356 = HEAP8[$355>>0]|0;
  $357 = $356&255;
  $358 = $357 & 15;
  $359 = ($358|0)==(4);
  if (!($359)) {
   continue;
  }
  $360 = $peer;
  $361 = $command;
  $362 = $sentTime;
  (_enet_peer_queue_acknowledgement($360,$361,$362)|0);
 }
 $366 = $2;
 $367 = ($366|0)!=(0|0);
 if ($367) {
  $368 = $2;
  $369 = HEAP32[$368>>2]|0;
  $370 = ($369|0)!=(0);
  if ($370) {
   $0 = 1;
   $371 = $0;
   STACKTOP = sp;return ($371|0);
  }
 }
 $0 = 0;
 $371 = $0;
 STACKTOP = sp;return ($371|0);
}
function _enet_protocol_handle_acknowledge($host,$event,$peer,$command) {
 $host = $host|0;
 $event = $event|0;
 $peer = $peer|0;
 $command = $command|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $commandNumber = 0, $receivedReliableSequenceNumber = 0, $receivedSentTime = 0, $roundTripTime = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $host;
 $2 = $event;
 $3 = $peer;
 $4 = $command;
 $5 = $3;
 $6 = ((($5)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0);
 if (!($8)) {
  $9 = $3;
  $10 = ((($9)) + 36|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($11|0)==(9);
  if (!($12)) {
   $13 = $4;
   $14 = ((($13)) + 6|0);
   $15 = HEAPU8[$14>>0]|(HEAPU8[$14+1>>0]<<8);
   $16 = (_ntohs($15)|0);
   $17 = $16&65535;
   $receivedSentTime = $17;
   $18 = $1;
   $19 = ((($18)) + 48|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = $20 & -65536;
   $22 = $receivedSentTime;
   $23 = $22 | $21;
   $receivedSentTime = $23;
   $24 = $receivedSentTime;
   $25 = $24 & 32768;
   $26 = $1;
   $27 = ((($26)) + 48|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = $28 & 32768;
   $30 = ($25>>>0)>($29>>>0);
   if ($30) {
    $31 = $receivedSentTime;
    $32 = (($31) - 65536)|0;
    $receivedSentTime = $32;
   }
   $33 = $1;
   $34 = ((($33)) + 48|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = $receivedSentTime;
   $37 = (($35) - ($36))|0;
   $38 = ($37>>>0)>=(86400000);
   if ($38) {
    $0 = 0;
    $245 = $0;
    STACKTOP = sp;return ($245|0);
   }
   $39 = $1;
   $40 = ((($39)) + 48|0);
   $41 = HEAP32[$40>>2]|0;
   $42 = $3;
   $43 = ((($42)) + 76|0);
   HEAP32[$43>>2] = $41;
   $44 = $3;
   $45 = ((($44)) + 84|0);
   HEAP32[$45>>2] = 0;
   $46 = $1;
   $47 = ((($46)) + 48|0);
   $48 = HEAP32[$47>>2]|0;
   $49 = $receivedSentTime;
   $50 = (($48) - ($49))|0;
   $51 = ($50>>>0)>=(86400000);
   if ($51) {
    $52 = $receivedSentTime;
    $53 = $1;
    $54 = ((($53)) + 48|0);
    $55 = HEAP32[$54>>2]|0;
    $56 = (($52) - ($55))|0;
    $62 = $56;
   } else {
    $57 = $1;
    $58 = ((($57)) + 48|0);
    $59 = HEAP32[$58>>2]|0;
    $60 = $receivedSentTime;
    $61 = (($59) - ($60))|0;
    $62 = $61;
   }
   $roundTripTime = $62;
   $63 = $3;
   $64 = $roundTripTime;
   (_enet_peer_throttle($63,$64)|0);
   $65 = $3;
   $66 = ((($65)) + 172|0);
   $67 = HEAP32[$66>>2]|0;
   $68 = (($67>>>0) / 4)&-1;
   $69 = $3;
   $70 = ((($69)) + 172|0);
   $71 = HEAP32[$70>>2]|0;
   $72 = (($71) - ($68))|0;
   HEAP32[$70>>2] = $72;
   $73 = $roundTripTime;
   $74 = $3;
   $75 = ((($74)) + 168|0);
   $76 = HEAP32[$75>>2]|0;
   $77 = ($73>>>0)>=($76>>>0);
   if ($77) {
    $78 = $roundTripTime;
    $79 = $3;
    $80 = ((($79)) + 168|0);
    $81 = HEAP32[$80>>2]|0;
    $82 = (($78) - ($81))|0;
    $83 = (($82>>>0) / 8)&-1;
    $84 = $3;
    $85 = ((($84)) + 168|0);
    $86 = HEAP32[$85>>2]|0;
    $87 = (($86) + ($83))|0;
    HEAP32[$85>>2] = $87;
    $88 = $roundTripTime;
    $89 = $3;
    $90 = ((($89)) + 168|0);
    $91 = HEAP32[$90>>2]|0;
    $92 = (($88) - ($91))|0;
    $93 = (($92>>>0) / 4)&-1;
    $94 = $3;
    $95 = ((($94)) + 172|0);
    $96 = HEAP32[$95>>2]|0;
    $97 = (($96) + ($93))|0;
    HEAP32[$95>>2] = $97;
   } else {
    $98 = $3;
    $99 = ((($98)) + 168|0);
    $100 = HEAP32[$99>>2]|0;
    $101 = $roundTripTime;
    $102 = (($100) - ($101))|0;
    $103 = (($102>>>0) / 8)&-1;
    $104 = $3;
    $105 = ((($104)) + 168|0);
    $106 = HEAP32[$105>>2]|0;
    $107 = (($106) - ($103))|0;
    HEAP32[$105>>2] = $107;
    $108 = $3;
    $109 = ((($108)) + 168|0);
    $110 = HEAP32[$109>>2]|0;
    $111 = $roundTripTime;
    $112 = (($110) - ($111))|0;
    $113 = (($112>>>0) / 4)&-1;
    $114 = $3;
    $115 = ((($114)) + 172|0);
    $116 = HEAP32[$115>>2]|0;
    $117 = (($116) + ($113))|0;
    HEAP32[$115>>2] = $117;
   }
   $118 = $3;
   $119 = ((($118)) + 168|0);
   $120 = HEAP32[$119>>2]|0;
   $121 = $3;
   $122 = ((($121)) + 156|0);
   $123 = HEAP32[$122>>2]|0;
   $124 = ($120>>>0)<($123>>>0);
   if ($124) {
    $125 = $3;
    $126 = ((($125)) + 168|0);
    $127 = HEAP32[$126>>2]|0;
    $128 = $3;
    $129 = ((($128)) + 156|0);
    HEAP32[$129>>2] = $127;
   }
   $130 = $3;
   $131 = ((($130)) + 172|0);
   $132 = HEAP32[$131>>2]|0;
   $133 = $3;
   $134 = ((($133)) + 164|0);
   $135 = HEAP32[$134>>2]|0;
   $136 = ($132>>>0)>($135>>>0);
   if ($136) {
    $137 = $3;
    $138 = ((($137)) + 172|0);
    $139 = HEAP32[$138>>2]|0;
    $140 = $3;
    $141 = ((($140)) + 164|0);
    HEAP32[$141>>2] = $139;
   }
   $142 = $3;
   $143 = ((($142)) + 120|0);
   $144 = HEAP32[$143>>2]|0;
   $145 = ($144|0)==(0);
   if ($145) {
    label = 23;
   } else {
    $146 = $1;
    $147 = ((($146)) + 48|0);
    $148 = HEAP32[$147>>2]|0;
    $149 = $3;
    $150 = ((($149)) + 120|0);
    $151 = HEAP32[$150>>2]|0;
    $152 = (($148) - ($151))|0;
    $153 = ($152>>>0)>=(86400000);
    if ($153) {
     $154 = $3;
     $155 = ((($154)) + 120|0);
     $156 = HEAP32[$155>>2]|0;
     $157 = $1;
     $158 = ((($157)) + 48|0);
     $159 = HEAP32[$158>>2]|0;
     $160 = (($156) - ($159))|0;
     $172 = $160;
    } else {
     $161 = $1;
     $162 = ((($161)) + 48|0);
     $163 = HEAP32[$162>>2]|0;
     $164 = $3;
     $165 = ((($164)) + 120|0);
     $166 = HEAP32[$165>>2]|0;
     $167 = (($163) - ($166))|0;
     $172 = $167;
    }
    $168 = $3;
    $169 = ((($168)) + 132|0);
    $170 = HEAP32[$169>>2]|0;
    $171 = ($172>>>0)>=($170>>>0);
    if ($171) {
     label = 23;
    }
   }
   if ((label|0) == 23) {
    $173 = $3;
    $174 = ((($173)) + 156|0);
    $175 = HEAP32[$174>>2]|0;
    $176 = $3;
    $177 = ((($176)) + 152|0);
    HEAP32[$177>>2] = $175;
    $178 = $3;
    $179 = ((($178)) + 164|0);
    $180 = HEAP32[$179>>2]|0;
    $181 = $3;
    $182 = ((($181)) + 160|0);
    HEAP32[$182>>2] = $180;
    $183 = $3;
    $184 = ((($183)) + 168|0);
    $185 = HEAP32[$184>>2]|0;
    $186 = $3;
    $187 = ((($186)) + 156|0);
    HEAP32[$187>>2] = $185;
    $188 = $3;
    $189 = ((($188)) + 172|0);
    $190 = HEAP32[$189>>2]|0;
    $191 = $3;
    $192 = ((($191)) + 164|0);
    HEAP32[$192>>2] = $190;
    $193 = $1;
    $194 = ((($193)) + 48|0);
    $195 = HEAP32[$194>>2]|0;
    $196 = $3;
    $197 = ((($196)) + 120|0);
    HEAP32[$197>>2] = $195;
   }
   $198 = $4;
   $199 = ((($198)) + 4|0);
   $200 = HEAPU8[$199>>0]|(HEAPU8[$199+1>>0]<<8);
   $201 = (_ntohs($200)|0);
   $202 = $201&65535;
   $receivedReliableSequenceNumber = $202;
   $203 = $3;
   $204 = $receivedReliableSequenceNumber;
   $205 = $204&65535;
   $206 = $4;
   $207 = ((($206)) + 1|0);
   $208 = HEAP8[$207>>0]|0;
   $209 = (_enet_protocol_remove_sent_reliable_command($203,$205,$208)|0);
   $commandNumber = $209;
   $210 = $3;
   $211 = ((($210)) + 36|0);
   $212 = HEAP32[$211>>2]|0;
   L34: do {
    switch ($212|0) {
    case 2:  {
     $213 = $commandNumber;
     $214 = ($213|0)!=(3);
     if (!($214)) {
      $215 = $1;
      $216 = $3;
      $217 = $2;
      _enet_protocol_notify_connect($215,$216,$217);
      break L34;
     }
     $0 = -1;
     $245 = $0;
     STACKTOP = sp;return ($245|0);
     break;
    }
    case 7:  {
     $218 = $commandNumber;
     $219 = ($218|0)!=(4);
     if (!($219)) {
      $220 = $1;
      $221 = $3;
      $222 = $2;
      _enet_protocol_notify_disconnect($220,$221,$222);
      break L34;
     }
     $0 = -1;
     $245 = $0;
     STACKTOP = sp;return ($245|0);
     break;
    }
    case 6:  {
     $223 = $3;
     $224 = ((($223)) + 216|0);
     $225 = HEAP32[$224>>2]|0;
     $226 = $3;
     $227 = ((($226)) + 216|0);
     $228 = ($225|0)==($227|0);
     if ($228) {
      $229 = $3;
      $230 = ((($229)) + 224|0);
      $231 = HEAP32[$230>>2]|0;
      $232 = $3;
      $233 = ((($232)) + 224|0);
      $234 = ($231|0)==($233|0);
      if ($234) {
       $235 = $3;
       $236 = ((($235)) + 200|0);
       $237 = HEAP32[$236>>2]|0;
       $238 = $3;
       $239 = ((($238)) + 200|0);
       $240 = ($237|0)==($239|0);
       if ($240) {
        $241 = $3;
        $242 = $3;
        $243 = ((($242)) + 376|0);
        $244 = HEAP32[$243>>2]|0;
        _enet_peer_disconnect($241,$244);
       }
      }
     }
     break;
    }
    default: {
    }
    }
   } while(0);
   $0 = 0;
   $245 = $0;
   STACKTOP = sp;return ($245|0);
  }
 }
 $0 = 0;
 $245 = $0;
 STACKTOP = sp;return ($245|0);
}
function _enet_protocol_remove_sent_reliable_command($peer,$reliableSequenceNumber,$channelID) {
 $peer = $peer|0;
 $reliableSequenceNumber = $reliableSequenceNumber|0;
 $channelID = $channelID|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $channel = 0, $commandNumber = 0;
 var $currentCommand = 0, $outgoingCommand = 0, $reliableWindow = 0, $wasSent = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $peer;
 $2 = $reliableSequenceNumber;
 $3 = $channelID;
 $outgoingCommand = 0;
 $wasSent = 1;
 $4 = $1;
 $5 = ((($4)) + 200|0);
 $6 = HEAP32[$5>>2]|0;
 $currentCommand = $6;
 while(1) {
  $7 = $currentCommand;
  $8 = $1;
  $9 = ((($8)) + 200|0);
  $10 = ($7|0)!=($9|0);
  if (!($10)) {
   break;
  }
  $11 = $currentCommand;
  $outgoingCommand = $11;
  $12 = $outgoingCommand;
  $13 = ((($12)) + 8|0);
  $14 = HEAP16[$13>>1]|0;
  $15 = $14&65535;
  $16 = $2;
  $17 = $16&65535;
  $18 = ($15|0)==($17|0);
  if ($18) {
   $19 = $outgoingCommand;
   $20 = ((($19)) + 32|0);
   $21 = ((($20)) + 1|0);
   $22 = HEAP8[$21>>0]|0;
   $23 = $22&255;
   $24 = $3;
   $25 = $24&255;
   $26 = ($23|0)==($25|0);
   if ($26) {
    break;
   }
  }
  $27 = $currentCommand;
  $28 = HEAP32[$27>>2]|0;
  $currentCommand = $28;
 }
 $29 = $currentCommand;
 $30 = $1;
 $31 = ((($30)) + 200|0);
 $32 = ($29|0)==($31|0);
 do {
  if ($32) {
   $33 = $1;
   $34 = ((($33)) + 216|0);
   $35 = HEAP32[$34>>2]|0;
   $currentCommand = $35;
   while(1) {
    $36 = $currentCommand;
    $37 = $1;
    $38 = ((($37)) + 216|0);
    $39 = ($36|0)!=($38|0);
    if (!($39)) {
     break;
    }
    $40 = $currentCommand;
    $outgoingCommand = $40;
    $41 = $outgoingCommand;
    $42 = ((($41)) + 30|0);
    $43 = HEAP16[$42>>1]|0;
    $44 = $43&65535;
    $45 = ($44|0)<(1);
    if ($45) {
     label = 10;
     break;
    }
    $46 = $outgoingCommand;
    $47 = ((($46)) + 8|0);
    $48 = HEAP16[$47>>1]|0;
    $49 = $48&65535;
    $50 = $2;
    $51 = $50&65535;
    $52 = ($49|0)==($51|0);
    if ($52) {
     $53 = $outgoingCommand;
     $54 = ((($53)) + 32|0);
     $55 = ((($54)) + 1|0);
     $56 = HEAP8[$55>>0]|0;
     $57 = $56&255;
     $58 = $3;
     $59 = $58&255;
     $60 = ($57|0)==($59|0);
     if ($60) {
      break;
     }
    }
    $61 = $currentCommand;
    $62 = HEAP32[$61>>2]|0;
    $currentCommand = $62;
   }
   if ((label|0) == 10) {
    $0 = 0;
    $177 = $0;
    STACKTOP = sp;return ($177|0);
   }
   $63 = $currentCommand;
   $64 = $1;
   $65 = ((($64)) + 216|0);
   $66 = ($63|0)==($65|0);
   if (!($66)) {
    $wasSent = 0;
    break;
   }
   $0 = 0;
   $177 = $0;
   STACKTOP = sp;return ($177|0);
  }
 } while(0);
 $67 = $outgoingCommand;
 $68 = ($67|0)==(0|0);
 if ($68) {
  $0 = 0;
  $177 = $0;
  STACKTOP = sp;return ($177|0);
 }
 $69 = $3;
 $70 = $69&255;
 $71 = $1;
 $72 = ((($71)) + 44|0);
 $73 = HEAP32[$72>>2]|0;
 $74 = ($70>>>0)<($73>>>0);
 if ($74) {
  $75 = $3;
  $76 = $75&255;
  $77 = $1;
  $78 = ((($77)) + 40|0);
  $79 = HEAP32[$78>>2]|0;
  $80 = (($79) + (($76*60)|0)|0);
  $channel = $80;
  $81 = $2;
  $82 = $81&65535;
  $83 = (($82|0) / 4096)&-1;
  $84 = $83&65535;
  $reliableWindow = $84;
  $85 = $reliableWindow;
  $86 = $85&65535;
  $87 = $channel;
  $88 = ((($87)) + 6|0);
  $89 = (($88) + ($86<<1)|0);
  $90 = HEAP16[$89>>1]|0;
  $91 = $90&65535;
  $92 = ($91|0)>(0);
  if ($92) {
   $93 = $reliableWindow;
   $94 = $93&65535;
   $95 = $channel;
   $96 = ((($95)) + 6|0);
   $97 = (($96) + ($94<<1)|0);
   $98 = HEAP16[$97>>1]|0;
   $99 = (($98) + -1)<<16>>16;
   HEAP16[$97>>1] = $99;
   $100 = $reliableWindow;
   $101 = $100&65535;
   $102 = $channel;
   $103 = ((($102)) + 6|0);
   $104 = (($103) + ($101<<1)|0);
   $105 = HEAP16[$104>>1]|0;
   $106 = ($105<<16>>16)!=(0);
   if (!($106)) {
    $107 = $reliableWindow;
    $108 = $107&65535;
    $109 = 1 << $108;
    $110 = $109 ^ -1;
    $111 = $channel;
    $112 = ((($111)) + 4|0);
    $113 = HEAP16[$112>>1]|0;
    $114 = $113&65535;
    $115 = $114 & $110;
    $116 = $115&65535;
    HEAP16[$112>>1] = $116;
   }
  }
 }
 $117 = $outgoingCommand;
 $118 = ((($117)) + 32|0);
 $119 = HEAP8[$118>>0]|0;
 $120 = $119&255;
 $121 = $120 & 15;
 $commandNumber = $121;
 $122 = $outgoingCommand;
 (_enet_list_remove($122)|0);
 $123 = $outgoingCommand;
 $124 = ((($123)) + 80|0);
 $125 = HEAP32[$124>>2]|0;
 $126 = ($125|0)!=(0|0);
 if ($126) {
  $127 = $wasSent;
  $128 = ($127|0)!=(0);
  if ($128) {
   $129 = $outgoingCommand;
   $130 = ((($129)) + 28|0);
   $131 = HEAP16[$130>>1]|0;
   $132 = $131&65535;
   $133 = $1;
   $134 = ((($133)) + 184|0);
   $135 = HEAP32[$134>>2]|0;
   $136 = (($135) - ($132))|0;
   HEAP32[$134>>2] = $136;
  }
  $137 = $outgoingCommand;
  $138 = ((($137)) + 80|0);
  $139 = HEAP32[$138>>2]|0;
  $140 = HEAP32[$139>>2]|0;
  $141 = (($140) + -1)|0;
  HEAP32[$139>>2] = $141;
  $142 = $outgoingCommand;
  $143 = ((($142)) + 80|0);
  $144 = HEAP32[$143>>2]|0;
  $145 = HEAP32[$144>>2]|0;
  $146 = ($145|0)==(0);
  if ($146) {
   $147 = $outgoingCommand;
   $148 = ((($147)) + 80|0);
   $149 = HEAP32[$148>>2]|0;
   $150 = ((($149)) + 4|0);
   $151 = HEAP32[$150>>2]|0;
   $152 = $151 | 256;
   HEAP32[$150>>2] = $152;
   $153 = $outgoingCommand;
   $154 = ((($153)) + 80|0);
   $155 = HEAP32[$154>>2]|0;
   _enet_packet_destroy($155);
  }
 }
 $156 = $outgoingCommand;
 _enet_free($156);
 $157 = $1;
 $158 = ((($157)) + 200|0);
 $159 = HEAP32[$158>>2]|0;
 $160 = $1;
 $161 = ((($160)) + 200|0);
 $162 = ($159|0)==($161|0);
 if ($162) {
  $163 = $commandNumber;
  $0 = $163;
  $177 = $0;
  STACKTOP = sp;return ($177|0);
 } else {
  $164 = $1;
  $165 = ((($164)) + 200|0);
  $166 = HEAP32[$165>>2]|0;
  $outgoingCommand = $166;
  $167 = $outgoingCommand;
  $168 = ((($167)) + 12|0);
  $169 = HEAP32[$168>>2]|0;
  $170 = $outgoingCommand;
  $171 = ((($170)) + 16|0);
  $172 = HEAP32[$171>>2]|0;
  $173 = (($169) + ($172))|0;
  $174 = $1;
  $175 = ((($174)) + 80|0);
  HEAP32[$175>>2] = $173;
  $176 = $commandNumber;
  $0 = $176;
  $177 = $0;
  STACKTOP = sp;return ($177|0);
 }
 return (0)|0;
}
function _enet_protocol_notify_connect($host,$peer,$event) {
 $host = $host|0;
 $peer = $peer|0;
 $event = $event|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $host;
 $1 = $peer;
 $2 = $event;
 $3 = $0;
 $4 = ((($3)) + 32|0);
 HEAP32[$4>>2] = 1;
 $5 = $2;
 $6 = ($5|0)!=(0|0);
 $7 = $0;
 $8 = $1;
 if ($6) {
  _enet_protocol_change_state($7,$8,5);
  $9 = $2;
  HEAP32[$9>>2] = 1;
  $10 = $1;
  $11 = $2;
  $12 = ((($11)) + 4|0);
  HEAP32[$12>>2] = $10;
  $13 = $1;
  $14 = ((($13)) + 376|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = $2;
  $17 = ((($16)) + 12|0);
  HEAP32[$17>>2] = $15;
  STACKTOP = sp;return;
 } else {
  $18 = $1;
  $19 = ((($18)) + 36|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = ($20|0)==(1);
  $22 = $21 ? 4 : 3;
  _enet_protocol_dispatch_state($7,$8,$22);
  STACKTOP = sp;return;
 }
}
function _enet_protocol_handle_connect($host,$header,$command) {
 $host = $host|0;
 $header = $header|0;
 $command = $command|0;
 var $$sink = 0, $$sink2 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0;
 var $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0;
 var $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0;
 var $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0;
 var $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0;
 var $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0;
 var $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0;
 var $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0;
 var $385 = 0, $386 = 0, $387 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $channel = 0, $channelCount = 0, $currentPeer = 0, $duplicatePeers = 0, $incomingSessionID = 0, $mtu = 0, $or$cond = 0, $outgoingSessionID = 0, $peer = 0, $verifyCommand = 0;
 var $windowSize = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $verifyCommand = sp + 48|0;
 $1 = $host;
 $2 = $header;
 $3 = $command;
 $duplicatePeers = 0;
 $peer = 0;
 $4 = $3;
 $5 = ((($4)) + 16|0);
 $6 = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8)|(HEAPU8[$5+2>>0]<<16)|(HEAPU8[$5+3>>0]<<24);
 $7 = (_ntohl($6)|0);
 $channelCount = $7;
 $8 = $channelCount;
 $9 = ($8>>>0)<(1);
 $10 = $channelCount;
 $11 = ($10>>>0)>(255);
 $or$cond = $9 | $11;
 if ($or$cond) {
  $0 = 0;
  $387 = $0;
  STACKTOP = sp;return ($387|0);
 }
 $12 = $1;
 $13 = ((($12)) + 36|0);
 $14 = HEAP32[$13>>2]|0;
 $currentPeer = $14;
 while(1) {
  $15 = $currentPeer;
  $16 = $1;
  $17 = ((($16)) + 40|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = $1;
  $20 = ((($19)) + 36|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = (($21) + (($18*380)|0)|0);
  $23 = ($15>>>0)<($22>>>0);
  if (!($23)) {
   break;
  }
  $24 = $currentPeer;
  $25 = ((($24)) + 36|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = ($26|0)==(0);
  if ($27) {
   $28 = $peer;
   $29 = ($28|0)==(0|0);
   if ($29) {
    $30 = $currentPeer;
    $peer = $30;
   }
  } else {
   $31 = $currentPeer;
   $32 = ((($31)) + 36|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ($33|0)!=(1);
   if ($34) {
    $35 = $currentPeer;
    $36 = ((($35)) + 24|0);
    $37 = HEAP32[$36>>2]|0;
    $38 = $1;
    $39 = ((($38)) + 10348|0);
    $40 = HEAP32[$39>>2]|0;
    $41 = ($37|0)==($40|0);
    if ($41) {
     $42 = $currentPeer;
     $43 = ((($42)) + 24|0);
     $44 = ((($43)) + 4|0);
     $45 = HEAP16[$44>>1]|0;
     $46 = $45&65535;
     $47 = $1;
     $48 = ((($47)) + 10348|0);
     $49 = ((($48)) + 4|0);
     $50 = HEAP16[$49>>1]|0;
     $51 = $50&65535;
     $52 = ($46|0)==($51|0);
     if ($52) {
      $53 = $currentPeer;
      $54 = ((($53)) + 16|0);
      $55 = HEAP32[$54>>2]|0;
      $56 = $3;
      $57 = ((($56)) + 40|0);
      $58 = HEAPU8[$57>>0]|(HEAPU8[$57+1>>0]<<8)|(HEAPU8[$57+2>>0]<<16)|(HEAPU8[$57+3>>0]<<24);
      $59 = ($55|0)==($58|0);
      if ($59) {
       label = 12;
       break;
      }
     }
     $60 = $duplicatePeers;
     $61 = (($60) + 1)|0;
     $duplicatePeers = $61;
    }
   }
  }
  $62 = $currentPeer;
  $63 = ((($62)) + 380|0);
  $currentPeer = $63;
 }
 if ((label|0) == 12) {
  $0 = 0;
  $387 = $0;
  STACKTOP = sp;return ($387|0);
 }
 $64 = $peer;
 $65 = ($64|0)==(0|0);
 if (!($65)) {
  $66 = $duplicatePeers;
  $67 = $1;
  $68 = ((($67)) + 10392|0);
  $69 = HEAP32[$68>>2]|0;
  $70 = ($66>>>0)>=($69>>>0);
  if (!($70)) {
   $71 = $channelCount;
   $72 = $1;
   $73 = ((($72)) + 44|0);
   $74 = HEAP32[$73>>2]|0;
   $75 = ($71>>>0)>($74>>>0);
   if ($75) {
    $76 = $1;
    $77 = ((($76)) + 44|0);
    $78 = HEAP32[$77>>2]|0;
    $channelCount = $78;
   }
   $79 = $channelCount;
   $80 = ($79*60)|0;
   $81 = (_enet_malloc($80)|0);
   $82 = $peer;
   $83 = ((($82)) + 40|0);
   HEAP32[$83>>2] = $81;
   $84 = $peer;
   $85 = ((($84)) + 40|0);
   $86 = HEAP32[$85>>2]|0;
   $87 = ($86|0)==(0|0);
   if ($87) {
    $0 = 0;
    $387 = $0;
    STACKTOP = sp;return ($387|0);
   }
   $88 = $channelCount;
   $89 = $peer;
   $90 = ((($89)) + 44|0);
   HEAP32[$90>>2] = $88;
   $91 = $peer;
   $92 = ((($91)) + 36|0);
   HEAP32[$92>>2] = 2;
   $93 = $3;
   $94 = ((($93)) + 40|0);
   $95 = HEAPU8[$94>>0]|(HEAPU8[$94+1>>0]<<8)|(HEAPU8[$94+2>>0]<<16)|(HEAPU8[$94+3>>0]<<24);
   $96 = $peer;
   $97 = ((($96)) + 16|0);
   HEAP32[$97>>2] = $95;
   $98 = $peer;
   $99 = ((($98)) + 24|0);
   $100 = $1;
   $101 = ((($100)) + 10348|0);
   ;HEAP32[$99>>2]=HEAP32[$101>>2]|0;HEAP32[$99+4>>2]=HEAP32[$101+4>>2]|0;
   $102 = $3;
   $103 = ((($102)) + 4|0);
   $104 = HEAPU8[$103>>0]|(HEAPU8[$103+1>>0]<<8);
   $105 = (_ntohs($104)|0);
   $106 = $peer;
   $107 = ((($106)) + 12|0);
   HEAP16[$107>>1] = $105;
   $108 = $3;
   $109 = ((($108)) + 20|0);
   $110 = HEAPU8[$109>>0]|(HEAPU8[$109+1>>0]<<8)|(HEAPU8[$109+2>>0]<<16)|(HEAPU8[$109+3>>0]<<24);
   $111 = (_ntohl($110)|0);
   $112 = $peer;
   $113 = ((($112)) + 48|0);
   HEAP32[$113>>2] = $111;
   $114 = $3;
   $115 = ((($114)) + 24|0);
   $116 = HEAPU8[$115>>0]|(HEAPU8[$115+1>>0]<<8)|(HEAPU8[$115+2>>0]<<16)|(HEAPU8[$115+3>>0]<<24);
   $117 = (_ntohl($116)|0);
   $118 = $peer;
   $119 = ((($118)) + 52|0);
   HEAP32[$119>>2] = $117;
   $120 = $3;
   $121 = ((($120)) + 28|0);
   $122 = HEAPU8[$121>>0]|(HEAPU8[$121+1>>0]<<8)|(HEAPU8[$121+2>>0]<<16)|(HEAPU8[$121+3>>0]<<24);
   $123 = (_ntohl($122)|0);
   $124 = $peer;
   $125 = ((($124)) + 132|0);
   HEAP32[$125>>2] = $123;
   $126 = $3;
   $127 = ((($126)) + 32|0);
   $128 = HEAPU8[$127>>0]|(HEAPU8[$127+1>>0]<<8)|(HEAPU8[$127+2>>0]<<16)|(HEAPU8[$127+3>>0]<<24);
   $129 = (_ntohl($128)|0);
   $130 = $peer;
   $131 = ((($130)) + 124|0);
   HEAP32[$131>>2] = $129;
   $132 = $3;
   $133 = ((($132)) + 36|0);
   $134 = HEAPU8[$133>>0]|(HEAPU8[$133+1>>0]<<8)|(HEAPU8[$133+2>>0]<<16)|(HEAPU8[$133+3>>0]<<24);
   $135 = (_ntohl($134)|0);
   $136 = $peer;
   $137 = ((($136)) + 128|0);
   HEAP32[$137>>2] = $135;
   $138 = $3;
   $139 = ((($138)) + 44|0);
   $140 = HEAPU8[$139>>0]|(HEAPU8[$139+1>>0]<<8)|(HEAPU8[$139+2>>0]<<16)|(HEAPU8[$139+3>>0]<<24);
   $141 = (_ntohl($140)|0);
   $142 = $peer;
   $143 = ((($142)) + 376|0);
   HEAP32[$143>>2] = $141;
   $144 = $3;
   $145 = ((($144)) + 6|0);
   $146 = HEAP8[$145>>0]|0;
   $147 = $146&255;
   $148 = ($147|0)==(255);
   if ($148) {
    $149 = $peer;
    $150 = ((($149)) + 20|0);
    $151 = HEAP8[$150>>0]|0;
    $$sink = $151;
   } else {
    $152 = $3;
    $153 = ((($152)) + 6|0);
    $154 = HEAP8[$153>>0]|0;
    $$sink = $154;
   }
   $155 = $$sink&255;
   $156 = $155&255;
   $incomingSessionID = $156;
   $157 = $incomingSessionID;
   $158 = $157&255;
   $159 = (($158) + 1)|0;
   $160 = $159 & 3;
   $161 = $160&255;
   $incomingSessionID = $161;
   $162 = $incomingSessionID;
   $163 = $162&255;
   $164 = $peer;
   $165 = ((($164)) + 20|0);
   $166 = HEAP8[$165>>0]|0;
   $167 = $166&255;
   $168 = ($163|0)==($167|0);
   if ($168) {
    $169 = $incomingSessionID;
    $170 = $169&255;
    $171 = (($170) + 1)|0;
    $172 = $171 & 3;
    $173 = $172&255;
    $incomingSessionID = $173;
   }
   $174 = $incomingSessionID;
   $175 = $peer;
   $176 = ((($175)) + 20|0);
   HEAP8[$176>>0] = $174;
   $177 = $3;
   $178 = ((($177)) + 7|0);
   $179 = HEAP8[$178>>0]|0;
   $180 = $179&255;
   $181 = ($180|0)==(255);
   if ($181) {
    $182 = $peer;
    $183 = ((($182)) + 21|0);
    $184 = HEAP8[$183>>0]|0;
    $$sink2 = $184;
   } else {
    $185 = $3;
    $186 = ((($185)) + 7|0);
    $187 = HEAP8[$186>>0]|0;
    $$sink2 = $187;
   }
   $188 = $$sink2&255;
   $189 = $188&255;
   $outgoingSessionID = $189;
   $190 = $outgoingSessionID;
   $191 = $190&255;
   $192 = (($191) + 1)|0;
   $193 = $192 & 3;
   $194 = $193&255;
   $outgoingSessionID = $194;
   $195 = $outgoingSessionID;
   $196 = $195&255;
   $197 = $peer;
   $198 = ((($197)) + 21|0);
   $199 = HEAP8[$198>>0]|0;
   $200 = $199&255;
   $201 = ($196|0)==($200|0);
   if ($201) {
    $202 = $outgoingSessionID;
    $203 = $202&255;
    $204 = (($203) + 1)|0;
    $205 = $204 & 3;
    $206 = $205&255;
    $outgoingSessionID = $206;
   }
   $207 = $outgoingSessionID;
   $208 = $peer;
   $209 = ((($208)) + 21|0);
   HEAP8[$209>>0] = $207;
   $210 = $peer;
   $211 = ((($210)) + 40|0);
   $212 = HEAP32[$211>>2]|0;
   $channel = $212;
   while(1) {
    $213 = $channel;
    $214 = $channelCount;
    $215 = $peer;
    $216 = ((($215)) + 40|0);
    $217 = HEAP32[$216>>2]|0;
    $218 = (($217) + (($214*60)|0)|0);
    $219 = ($213>>>0)<($218>>>0);
    if (!($219)) {
     break;
    }
    $220 = $channel;
    HEAP16[$220>>1] = 0;
    $221 = $channel;
    $222 = ((($221)) + 2|0);
    HEAP16[$222>>1] = 0;
    $223 = $channel;
    $224 = ((($223)) + 38|0);
    HEAP16[$224>>1] = 0;
    $225 = $channel;
    $226 = ((($225)) + 40|0);
    HEAP16[$226>>1] = 0;
    $227 = $channel;
    $228 = ((($227)) + 44|0);
    _enet_list_clear($228);
    $229 = $channel;
    $230 = ((($229)) + 52|0);
    _enet_list_clear($230);
    $231 = $channel;
    $232 = ((($231)) + 4|0);
    HEAP16[$232>>1] = 0;
    $233 = $channel;
    $234 = ((($233)) + 6|0);
    dest=$234; stop=dest+32|0; do { HEAP16[dest>>1]=0|0; dest=dest+2|0; } while ((dest|0) < (stop|0));
    $235 = $channel;
    $236 = ((($235)) + 60|0);
    $channel = $236;
   }
   $237 = $3;
   $238 = ((($237)) + 8|0);
   $239 = HEAPU8[$238>>0]|(HEAPU8[$238+1>>0]<<8)|(HEAPU8[$238+2>>0]<<16)|(HEAPU8[$238+3>>0]<<24);
   $240 = (_ntohl($239)|0);
   $mtu = $240;
   $241 = $mtu;
   $242 = ($241>>>0)<(576);
   if ($242) {
    $mtu = 576;
   } else {
    $243 = $mtu;
    $244 = ($243>>>0)>(4096);
    if ($244) {
     $mtu = 4096;
    }
   }
   $245 = $mtu;
   $246 = $peer;
   $247 = ((($246)) + 176|0);
   HEAP32[$247>>2] = $245;
   $248 = $1;
   $249 = ((($248)) + 16|0);
   $250 = HEAP32[$249>>2]|0;
   $251 = ($250|0)==(0);
   if ($251) {
    $252 = $peer;
    $253 = ((($252)) + 48|0);
    $254 = HEAP32[$253>>2]|0;
    $255 = ($254|0)==(0);
    if ($255) {
     $256 = $peer;
     $257 = ((($256)) + 180|0);
     HEAP32[$257>>2] = 65536;
    } else {
     label = 42;
    }
   } else {
    label = 42;
   }
   do {
    if ((label|0) == 42) {
     $258 = $1;
     $259 = ((($258)) + 16|0);
     $260 = HEAP32[$259>>2]|0;
     $261 = ($260|0)==(0);
     if (!($261)) {
      $262 = $peer;
      $263 = ((($262)) + 48|0);
      $264 = HEAP32[$263>>2]|0;
      $265 = ($264|0)==(0);
      if (!($265)) {
       $284 = $1;
       $285 = ((($284)) + 16|0);
       $286 = HEAP32[$285>>2]|0;
       $287 = $peer;
       $288 = ((($287)) + 48|0);
       $289 = HEAP32[$288>>2]|0;
       $290 = ($286>>>0)<($289>>>0);
       if ($290) {
        $291 = $1;
        $292 = ((($291)) + 16|0);
        $293 = HEAP32[$292>>2]|0;
        $298 = $293;
       } else {
        $294 = $peer;
        $295 = ((($294)) + 48|0);
        $296 = HEAP32[$295>>2]|0;
        $298 = $296;
       }
       $297 = (($298>>>0) / 65536)&-1;
       $299 = $297<<12;
       $300 = $peer;
       $301 = ((($300)) + 180|0);
       HEAP32[$301>>2] = $299;
       break;
      }
     }
     $266 = $1;
     $267 = ((($266)) + 16|0);
     $268 = HEAP32[$267>>2]|0;
     $269 = $peer;
     $270 = ((($269)) + 48|0);
     $271 = HEAP32[$270>>2]|0;
     $272 = ($268>>>0)>($271>>>0);
     if ($272) {
      $273 = $1;
      $274 = ((($273)) + 16|0);
      $275 = HEAP32[$274>>2]|0;
      $280 = $275;
     } else {
      $276 = $peer;
      $277 = ((($276)) + 48|0);
      $278 = HEAP32[$277>>2]|0;
      $280 = $278;
     }
     $279 = (($280>>>0) / 65536)&-1;
     $281 = $279<<12;
     $282 = $peer;
     $283 = ((($282)) + 180|0);
     HEAP32[$283>>2] = $281;
    }
   } while(0);
   $302 = $peer;
   $303 = ((($302)) + 180|0);
   $304 = HEAP32[$303>>2]|0;
   $305 = ($304>>>0)<(4096);
   $306 = $peer;
   $307 = ((($306)) + 180|0);
   if ($305) {
    HEAP32[$307>>2] = 4096;
   } else {
    $308 = HEAP32[$307>>2]|0;
    $309 = ($308>>>0)>(65536);
    if ($309) {
     $310 = $peer;
     $311 = ((($310)) + 180|0);
     HEAP32[$311>>2] = 65536;
    }
   }
   $312 = $1;
   $313 = ((($312)) + 12|0);
   $314 = HEAP32[$313>>2]|0;
   $315 = ($314|0)==(0);
   if ($315) {
    $windowSize = 65536;
   } else {
    $316 = $1;
    $317 = ((($316)) + 12|0);
    $318 = HEAP32[$317>>2]|0;
    $319 = (($318>>>0) / 65536)&-1;
    $320 = $319<<12;
    $windowSize = $320;
   }
   $321 = $windowSize;
   $322 = $3;
   $323 = ((($322)) + 12|0);
   $324 = HEAPU8[$323>>0]|(HEAPU8[$323+1>>0]<<8)|(HEAPU8[$323+2>>0]<<16)|(HEAPU8[$323+3>>0]<<24);
   $325 = (_ntohl($324)|0);
   $326 = ($321>>>0)>($325>>>0);
   if ($326) {
    $327 = $3;
    $328 = ((($327)) + 12|0);
    $329 = HEAPU8[$328>>0]|(HEAPU8[$328+1>>0]<<8)|(HEAPU8[$328+2>>0]<<16)|(HEAPU8[$328+3>>0]<<24);
    $330 = (_ntohl($329)|0);
    $windowSize = $330;
   }
   $331 = $windowSize;
   $332 = ($331>>>0)<(4096);
   if ($332) {
    $windowSize = 4096;
   } else {
    $333 = $windowSize;
    $334 = ($333>>>0)>(65536);
    if ($334) {
     $windowSize = 65536;
    }
   }
   HEAP8[$verifyCommand>>0] = -125;
   $335 = ((($verifyCommand)) + 1|0);
   HEAP8[$335>>0] = -1;
   $336 = $peer;
   $337 = ((($336)) + 14|0);
   $338 = HEAP16[$337>>1]|0;
   $339 = (_htons($338)|0);
   $340 = ((($verifyCommand)) + 4|0);
   HEAP8[$340>>0]=$339&255;HEAP8[$340+1>>0]=$339>>8;
   $341 = $incomingSessionID;
   $342 = ((($verifyCommand)) + 6|0);
   HEAP8[$342>>0] = $341;
   $343 = $outgoingSessionID;
   $344 = ((($verifyCommand)) + 7|0);
   HEAP8[$344>>0] = $343;
   $345 = $peer;
   $346 = ((($345)) + 176|0);
   $347 = HEAP32[$346>>2]|0;
   $348 = (_htonl($347)|0);
   $349 = ((($verifyCommand)) + 8|0);
   HEAP8[$349>>0]=$348&255;HEAP8[$349+1>>0]=($348>>8)&255;HEAP8[$349+2>>0]=($348>>16)&255;HEAP8[$349+3>>0]=$348>>24;
   $350 = $windowSize;
   $351 = (_htonl($350)|0);
   $352 = ((($verifyCommand)) + 12|0);
   HEAP8[$352>>0]=$351&255;HEAP8[$352+1>>0]=($351>>8)&255;HEAP8[$352+2>>0]=($351>>16)&255;HEAP8[$352+3>>0]=$351>>24;
   $353 = $channelCount;
   $354 = (_htonl($353)|0);
   $355 = ((($verifyCommand)) + 16|0);
   HEAP8[$355>>0]=$354&255;HEAP8[$355+1>>0]=($354>>8)&255;HEAP8[$355+2>>0]=($354>>16)&255;HEAP8[$355+3>>0]=$354>>24;
   $356 = $1;
   $357 = ((($356)) + 12|0);
   $358 = HEAP32[$357>>2]|0;
   $359 = (_htonl($358)|0);
   $360 = ((($verifyCommand)) + 20|0);
   HEAP8[$360>>0]=$359&255;HEAP8[$360+1>>0]=($359>>8)&255;HEAP8[$360+2>>0]=($359>>16)&255;HEAP8[$360+3>>0]=$359>>24;
   $361 = $1;
   $362 = ((($361)) + 16|0);
   $363 = HEAP32[$362>>2]|0;
   $364 = (_htonl($363)|0);
   $365 = ((($verifyCommand)) + 24|0);
   HEAP8[$365>>0]=$364&255;HEAP8[$365+1>>0]=($364>>8)&255;HEAP8[$365+2>>0]=($364>>16)&255;HEAP8[$365+3>>0]=$364>>24;
   $366 = $peer;
   $367 = ((($366)) + 132|0);
   $368 = HEAP32[$367>>2]|0;
   $369 = (_htonl($368)|0);
   $370 = ((($verifyCommand)) + 28|0);
   HEAP8[$370>>0]=$369&255;HEAP8[$370+1>>0]=($369>>8)&255;HEAP8[$370+2>>0]=($369>>16)&255;HEAP8[$370+3>>0]=$369>>24;
   $371 = $peer;
   $372 = ((($371)) + 124|0);
   $373 = HEAP32[$372>>2]|0;
   $374 = (_htonl($373)|0);
   $375 = ((($verifyCommand)) + 32|0);
   HEAP8[$375>>0]=$374&255;HEAP8[$375+1>>0]=($374>>8)&255;HEAP8[$375+2>>0]=($374>>16)&255;HEAP8[$375+3>>0]=$374>>24;
   $376 = $peer;
   $377 = ((($376)) + 128|0);
   $378 = HEAP32[$377>>2]|0;
   $379 = (_htonl($378)|0);
   $380 = ((($verifyCommand)) + 36|0);
   HEAP8[$380>>0]=$379&255;HEAP8[$380+1>>0]=($379>>8)&255;HEAP8[$380+2>>0]=($379>>16)&255;HEAP8[$380+3>>0]=$379>>24;
   $381 = $peer;
   $382 = ((($381)) + 16|0);
   $383 = HEAP32[$382>>2]|0;
   $384 = ((($verifyCommand)) + 40|0);
   HEAP8[$384>>0]=$383&255;HEAP8[$384+1>>0]=($383>>8)&255;HEAP8[$384+2>>0]=($383>>16)&255;HEAP8[$384+3>>0]=$383>>24;
   $385 = $peer;
   (_enet_peer_queue_outgoing_command($385,$verifyCommand,0,0,0)|0);
   $386 = $peer;
   $0 = $386;
   $387 = $0;
   STACKTOP = sp;return ($387|0);
  }
 }
 $0 = 0;
 $387 = $0;
 STACKTOP = sp;return ($387|0);
}
function _enet_protocol_handle_verify_connect($host,$event,$peer,$command) {
 $host = $host|0;
 $event = $event|0;
 $peer = $peer|0;
 $command = $command|0;
 var $$ = 0, $$$ = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $channelCount = 0, $mtu = 0, $or$cond = 0, $windowSize = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $host;
 $2 = $event;
 $3 = $peer;
 $4 = $command;
 $5 = $3;
 $6 = ((($5)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)!=(1);
 if ($8) {
  $0 = 0;
  $124 = $0;
  STACKTOP = sp;return ($124|0);
 }
 $9 = $4;
 $10 = ((($9)) + 16|0);
 $11 = HEAPU8[$10>>0]|(HEAPU8[$10+1>>0]<<8)|(HEAPU8[$10+2>>0]<<16)|(HEAPU8[$10+3>>0]<<24);
 $12 = (_ntohl($11)|0);
 $channelCount = $12;
 $13 = $channelCount;
 $14 = ($13>>>0)<(1);
 $15 = $channelCount;
 $16 = ($15>>>0)>(255);
 $or$cond = $14 | $16;
 if (!($or$cond)) {
  $17 = $4;
  $18 = ((($17)) + 28|0);
  $19 = HEAPU8[$18>>0]|(HEAPU8[$18+1>>0]<<8)|(HEAPU8[$18+2>>0]<<16)|(HEAPU8[$18+3>>0]<<24);
  $20 = (_ntohl($19)|0);
  $21 = $3;
  $22 = ((($21)) + 132|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = ($20|0)!=($23|0);
  if (!($24)) {
   $25 = $4;
   $26 = ((($25)) + 32|0);
   $27 = HEAPU8[$26>>0]|(HEAPU8[$26+1>>0]<<8)|(HEAPU8[$26+2>>0]<<16)|(HEAPU8[$26+3>>0]<<24);
   $28 = (_ntohl($27)|0);
   $29 = $3;
   $30 = ((($29)) + 124|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = ($28|0)!=($31|0);
   if (!($32)) {
    $33 = $4;
    $34 = ((($33)) + 36|0);
    $35 = HEAPU8[$34>>0]|(HEAPU8[$34+1>>0]<<8)|(HEAPU8[$34+2>>0]<<16)|(HEAPU8[$34+3>>0]<<24);
    $36 = (_ntohl($35)|0);
    $37 = $3;
    $38 = ((($37)) + 128|0);
    $39 = HEAP32[$38>>2]|0;
    $40 = ($36|0)!=($39|0);
    if (!($40)) {
     $41 = $4;
     $42 = ((($41)) + 40|0);
     $43 = HEAPU8[$42>>0]|(HEAPU8[$42+1>>0]<<8)|(HEAPU8[$42+2>>0]<<16)|(HEAPU8[$42+3>>0]<<24);
     $44 = $3;
     $45 = ((($44)) + 16|0);
     $46 = HEAP32[$45>>2]|0;
     $47 = ($43|0)!=($46|0);
     if (!($47)) {
      $52 = $3;
      (_enet_protocol_remove_sent_reliable_command($52,1,-1)|0);
      $53 = $channelCount;
      $54 = $3;
      $55 = ((($54)) + 44|0);
      $56 = HEAP32[$55>>2]|0;
      $57 = ($53>>>0)<($56>>>0);
      if ($57) {
       $58 = $channelCount;
       $59 = $3;
       $60 = ((($59)) + 44|0);
       HEAP32[$60>>2] = $58;
      }
      $61 = $4;
      $62 = ((($61)) + 4|0);
      $63 = HEAPU8[$62>>0]|(HEAPU8[$62+1>>0]<<8);
      $64 = (_ntohs($63)|0);
      $65 = $3;
      $66 = ((($65)) + 12|0);
      HEAP16[$66>>1] = $64;
      $67 = $4;
      $68 = ((($67)) + 6|0);
      $69 = HEAP8[$68>>0]|0;
      $70 = $3;
      $71 = ((($70)) + 21|0);
      HEAP8[$71>>0] = $69;
      $72 = $4;
      $73 = ((($72)) + 7|0);
      $74 = HEAP8[$73>>0]|0;
      $75 = $3;
      $76 = ((($75)) + 20|0);
      HEAP8[$76>>0] = $74;
      $77 = $4;
      $78 = ((($77)) + 8|0);
      $79 = HEAPU8[$78>>0]|(HEAPU8[$78+1>>0]<<8)|(HEAPU8[$78+2>>0]<<16)|(HEAPU8[$78+3>>0]<<24);
      $80 = (_ntohl($79)|0);
      $mtu = $80;
      $81 = $mtu;
      $82 = ($81>>>0)<(576);
      if ($82) {
       $mtu = 576;
      } else {
       $83 = $mtu;
       $84 = ($83>>>0)>(4096);
       if ($84) {
        $mtu = 4096;
       }
      }
      $85 = $mtu;
      $86 = $3;
      $87 = ((($86)) + 176|0);
      $88 = HEAP32[$87>>2]|0;
      $89 = ($85>>>0)<($88>>>0);
      if ($89) {
       $90 = $mtu;
       $91 = $3;
       $92 = ((($91)) + 176|0);
       HEAP32[$92>>2] = $90;
      }
      $93 = $4;
      $94 = ((($93)) + 12|0);
      $95 = HEAPU8[$94>>0]|(HEAPU8[$94+1>>0]<<8)|(HEAPU8[$94+2>>0]<<16)|(HEAPU8[$94+3>>0]<<24);
      $96 = (_ntohl($95)|0);
      $windowSize = $96;
      $97 = $windowSize;
      $98 = ($97>>>0)<(4096);
      $$ = $98 ? 4096 : $96;
      $windowSize = $$;
      $99 = $windowSize;
      $100 = ($99>>>0)>(65536);
      $$$ = $100 ? 65536 : $$;
      $windowSize = $$$;
      $101 = $windowSize;
      $102 = $3;
      $103 = ((($102)) + 180|0);
      $104 = HEAP32[$103>>2]|0;
      $105 = ($101>>>0)<($104>>>0);
      if ($105) {
       $106 = $windowSize;
       $107 = $3;
       $108 = ((($107)) + 180|0);
       HEAP32[$108>>2] = $106;
      }
      $109 = $4;
      $110 = ((($109)) + 20|0);
      $111 = HEAPU8[$110>>0]|(HEAPU8[$110+1>>0]<<8)|(HEAPU8[$110+2>>0]<<16)|(HEAPU8[$110+3>>0]<<24);
      $112 = (_ntohl($111)|0);
      $113 = $3;
      $114 = ((($113)) + 48|0);
      HEAP32[$114>>2] = $112;
      $115 = $4;
      $116 = ((($115)) + 24|0);
      $117 = HEAPU8[$116>>0]|(HEAPU8[$116+1>>0]<<8)|(HEAPU8[$116+2>>0]<<16)|(HEAPU8[$116+3>>0]<<24);
      $118 = (_ntohl($117)|0);
      $119 = $3;
      $120 = ((($119)) + 52|0);
      HEAP32[$120>>2] = $118;
      $121 = $1;
      $122 = $3;
      $123 = $2;
      _enet_protocol_notify_connect($121,$122,$123);
      $0 = 0;
      $124 = $0;
      STACKTOP = sp;return ($124|0);
     }
    }
   }
  }
 }
 $48 = $3;
 $49 = ((($48)) + 376|0);
 HEAP32[$49>>2] = 0;
 $50 = $1;
 $51 = $3;
 _enet_protocol_dispatch_state($50,$51,9);
 $0 = -1;
 $124 = $0;
 STACKTOP = sp;return ($124|0);
}
function _enet_protocol_handle_disconnect($host,$peer,$command) {
 $host = $host|0;
 $peer = $peer|0;
 $command = $command|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $host;
 $2 = $peer;
 $3 = $command;
 $4 = $2;
 $5 = ((($4)) + 36|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = $2;
  $9 = ((($8)) + 36|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ($10|0)==(9);
  if (!($11)) {
   $12 = $2;
   $13 = ((($12)) + 36|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ($14|0)==(8);
   if (!($15)) {
    $16 = $2;
    _enet_peer_reset_queues($16);
    $17 = $2;
    $18 = ((($17)) + 36|0);
    $19 = HEAP32[$18>>2]|0;
    $20 = ($19|0)==(4);
    do {
     if ($20) {
      label = 8;
     } else {
      $21 = $2;
      $22 = ((($21)) + 36|0);
      $23 = HEAP32[$22>>2]|0;
      $24 = ($23|0)==(7);
      if ($24) {
       label = 8;
      } else {
       $25 = $2;
       $26 = ((($25)) + 36|0);
       $27 = HEAP32[$26>>2]|0;
       $28 = ($27|0)==(1);
       if ($28) {
        label = 8;
       } else {
        $31 = $2;
        $32 = ((($31)) + 36|0);
        $33 = HEAP32[$32>>2]|0;
        $34 = ($33|0)!=(5);
        if ($34) {
         $35 = $2;
         $36 = ((($35)) + 36|0);
         $37 = HEAP32[$36>>2]|0;
         $38 = ($37|0)!=(6);
         if ($38) {
          $39 = $2;
          $40 = ((($39)) + 36|0);
          $41 = HEAP32[$40>>2]|0;
          $42 = ($41|0)==(3);
          if ($42) {
           $43 = $1;
           $44 = ((($43)) + 32|0);
           HEAP32[$44>>2] = 1;
          }
          $45 = $2;
          _enet_peer_reset($45);
          break;
         }
        }
        $46 = $3;
        $47 = HEAP8[$46>>0]|0;
        $48 = $47&255;
        $49 = $48 & 128;
        $50 = ($49|0)!=(0);
        $51 = $1;
        $52 = $2;
        if ($50) {
         _enet_protocol_change_state($51,$52,8);
         break;
        } else {
         _enet_protocol_dispatch_state($51,$52,9);
         break;
        }
       }
      }
     }
    } while(0);
    if ((label|0) == 8) {
     $29 = $1;
     $30 = $2;
     _enet_protocol_dispatch_state($29,$30,9);
    }
    $53 = $2;
    $54 = ((($53)) + 36|0);
    $55 = HEAP32[$54>>2]|0;
    $56 = ($55|0)!=(0);
    if ($56) {
     $57 = $3;
     $58 = ((($57)) + 4|0);
     $59 = HEAPU8[$58>>0]|(HEAPU8[$58+1>>0]<<8)|(HEAPU8[$58+2>>0]<<16)|(HEAPU8[$58+3>>0]<<24);
     $60 = (_ntohl($59)|0);
     $61 = $2;
     $62 = ((($61)) + 376|0);
     HEAP32[$62>>2] = $60;
    }
    $0 = 0;
    $63 = $0;
    STACKTOP = sp;return ($63|0);
   }
  }
 }
 $0 = 0;
 $63 = $0;
 STACKTOP = sp;return ($63|0);
}
function _enet_protocol_handle_ping($host,$peer,$command) {
 $host = $host|0;
 $peer = $peer|0;
 $command = $command|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $host;
 $2 = $peer;
 $3 = $command;
 $4 = $2;
 $5 = ((($4)) + 36|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)!=(5);
 if ($7) {
  $8 = $2;
  $9 = ((($8)) + 36|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ($10|0)!=(6);
  if ($11) {
   $0 = -1;
   $12 = $0;
   STACKTOP = sp;return ($12|0);
  }
 }
 $0 = 0;
 $12 = $0;
 STACKTOP = sp;return ($12|0);
}
function _enet_protocol_handle_send_reliable($host,$peer,$command,$currentData) {
 $host = $host|0;
 $peer = $peer|0;
 $command = $command|0;
 $currentData = $currentData|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $dataLength = 0, $packet = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $host;
 $2 = $peer;
 $3 = $command;
 $4 = $currentData;
 $5 = $3;
 $6 = ((($5)) + 1|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $9 = $2;
 $10 = ((($9)) + 44|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($8>>>0)>=($11>>>0);
 do {
  if (!($12)) {
   $13 = $2;
   $14 = ((($13)) + 36|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)!=(5);
   if ($16) {
    $17 = $2;
    $18 = ((($17)) + 36|0);
    $19 = HEAP32[$18>>2]|0;
    $20 = ($19|0)!=(6);
    if ($20) {
     break;
    }
   }
   $21 = $3;
   $22 = ((($21)) + 4|0);
   $23 = HEAPU8[$22>>0]|(HEAPU8[$22+1>>0]<<8);
   $24 = (_ntohs($23)|0);
   $25 = $24&65535;
   $dataLength = $25;
   $26 = $dataLength;
   $27 = $4;
   $28 = HEAP32[$27>>2]|0;
   $29 = (($28) + ($26)|0);
   HEAP32[$27>>2] = $29;
   $30 = $dataLength;
   $31 = ($30>>>0)>(1073741824);
   if (!($31)) {
    $32 = $4;
    $33 = HEAP32[$32>>2]|0;
    $34 = $1;
    $35 = ((($34)) + 10356|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($33>>>0)<($36>>>0);
    if (!($37)) {
     $38 = $4;
     $39 = HEAP32[$38>>2]|0;
     $40 = $1;
     $41 = ((($40)) + 10360|0);
     $42 = HEAP32[$41>>2]|0;
     $43 = $1;
     $44 = ((($43)) + 10356|0);
     $45 = HEAP32[$44>>2]|0;
     $46 = (($45) + ($42)|0);
     $47 = ($39>>>0)>($46>>>0);
     if (!($47)) {
      $48 = $3;
      $49 = ((($48)) + 6|0);
      $50 = $dataLength;
      $51 = (_enet_packet_create($49,$50,1)|0);
      $packet = $51;
      $52 = $packet;
      $53 = ($52|0)==(0|0);
      if (!($53)) {
       $54 = $2;
       $55 = $3;
       $56 = $packet;
       $57 = (_enet_peer_queue_incoming_command($54,$55,$56,0)|0);
       $58 = ($57|0)==(0|0);
       if (!($58)) {
        $0 = 0;
        $59 = $0;
        STACKTOP = sp;return ($59|0);
       }
      }
      $0 = -1;
      $59 = $0;
      STACKTOP = sp;return ($59|0);
     }
    }
   }
   $0 = -1;
   $59 = $0;
   STACKTOP = sp;return ($59|0);
  }
 } while(0);
 $0 = -1;
 $59 = $0;
 STACKTOP = sp;return ($59|0);
}
function _enet_protocol_handle_send_unreliable($host,$peer,$command,$currentData) {
 $host = $host|0;
 $peer = $peer|0;
 $command = $command|0;
 $currentData = $currentData|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $dataLength = 0, $packet = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $host;
 $2 = $peer;
 $3 = $command;
 $4 = $currentData;
 $5 = $3;
 $6 = ((($5)) + 1|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $9 = $2;
 $10 = ((($9)) + 44|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($8>>>0)>=($11>>>0);
 do {
  if (!($12)) {
   $13 = $2;
   $14 = ((($13)) + 36|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)!=(5);
   if ($16) {
    $17 = $2;
    $18 = ((($17)) + 36|0);
    $19 = HEAP32[$18>>2]|0;
    $20 = ($19|0)!=(6);
    if ($20) {
     break;
    }
   }
   $21 = $3;
   $22 = ((($21)) + 6|0);
   $23 = HEAPU8[$22>>0]|(HEAPU8[$22+1>>0]<<8);
   $24 = (_ntohs($23)|0);
   $25 = $24&65535;
   $dataLength = $25;
   $26 = $dataLength;
   $27 = $4;
   $28 = HEAP32[$27>>2]|0;
   $29 = (($28) + ($26)|0);
   HEAP32[$27>>2] = $29;
   $30 = $dataLength;
   $31 = ($30>>>0)>(1073741824);
   if (!($31)) {
    $32 = $4;
    $33 = HEAP32[$32>>2]|0;
    $34 = $1;
    $35 = ((($34)) + 10356|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($33>>>0)<($36>>>0);
    if (!($37)) {
     $38 = $4;
     $39 = HEAP32[$38>>2]|0;
     $40 = $1;
     $41 = ((($40)) + 10360|0);
     $42 = HEAP32[$41>>2]|0;
     $43 = $1;
     $44 = ((($43)) + 10356|0);
     $45 = HEAP32[$44>>2]|0;
     $46 = (($45) + ($42)|0);
     $47 = ($39>>>0)>($46>>>0);
     if (!($47)) {
      $48 = $3;
      $49 = ((($48)) + 8|0);
      $50 = $dataLength;
      $51 = (_enet_packet_create($49,$50,0)|0);
      $packet = $51;
      $52 = $packet;
      $53 = ($52|0)==(0|0);
      if (!($53)) {
       $54 = $2;
       $55 = $3;
       $56 = $packet;
       $57 = (_enet_peer_queue_incoming_command($54,$55,$56,0)|0);
       $58 = ($57|0)==(0|0);
       if (!($58)) {
        $0 = 0;
        $59 = $0;
        STACKTOP = sp;return ($59|0);
       }
      }
      $0 = -1;
      $59 = $0;
      STACKTOP = sp;return ($59|0);
     }
    }
   }
   $0 = -1;
   $59 = $0;
   STACKTOP = sp;return ($59|0);
  }
 } while(0);
 $0 = -1;
 $59 = $0;
 STACKTOP = sp;return ($59|0);
}
function _enet_protocol_handle_send_unsequenced($host,$peer,$command,$currentData) {
 $host = $host|0;
 $peer = $peer|0;
 $command = $command|0;
 $currentData = $currentData|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $dataLength = 0, $index = 0, $packet = 0, $unsequencedGroup = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $host;
 $2 = $peer;
 $3 = $command;
 $4 = $currentData;
 $5 = $3;
 $6 = ((($5)) + 1|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $9 = $2;
 $10 = ((($9)) + 44|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($8>>>0)>=($11>>>0);
 do {
  if (!($12)) {
   $13 = $2;
   $14 = ((($13)) + 36|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)!=(5);
   if ($16) {
    $17 = $2;
    $18 = ((($17)) + 36|0);
    $19 = HEAP32[$18>>2]|0;
    $20 = ($19|0)!=(6);
    if ($20) {
     break;
    }
   }
   $21 = $3;
   $22 = ((($21)) + 6|0);
   $23 = HEAPU8[$22>>0]|(HEAPU8[$22+1>>0]<<8);
   $24 = (_ntohs($23)|0);
   $25 = $24&65535;
   $dataLength = $25;
   $26 = $dataLength;
   $27 = $4;
   $28 = HEAP32[$27>>2]|0;
   $29 = (($28) + ($26)|0);
   HEAP32[$27>>2] = $29;
   $30 = $dataLength;
   $31 = ($30>>>0)>(1073741824);
   if (!($31)) {
    $32 = $4;
    $33 = HEAP32[$32>>2]|0;
    $34 = $1;
    $35 = ((($34)) + 10356|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($33>>>0)<($36>>>0);
    if (!($37)) {
     $38 = $4;
     $39 = HEAP32[$38>>2]|0;
     $40 = $1;
     $41 = ((($40)) + 10360|0);
     $42 = HEAP32[$41>>2]|0;
     $43 = $1;
     $44 = ((($43)) + 10356|0);
     $45 = HEAP32[$44>>2]|0;
     $46 = (($45) + ($42)|0);
     $47 = ($39>>>0)>($46>>>0);
     if (!($47)) {
      $48 = $3;
      $49 = ((($48)) + 4|0);
      $50 = HEAPU8[$49>>0]|(HEAPU8[$49+1>>0]<<8);
      $51 = (_ntohs($50)|0);
      $52 = $51&65535;
      $unsequencedGroup = $52;
      $53 = $unsequencedGroup;
      $54 = (($53>>>0) % 1024)&-1;
      $index = $54;
      $55 = $unsequencedGroup;
      $56 = $2;
      $57 = ((($56)) + 244|0);
      $58 = HEAP16[$57>>1]|0;
      $59 = $58&65535;
      $60 = ($55>>>0)<($59>>>0);
      if ($60) {
       $61 = $unsequencedGroup;
       $62 = (($61) + 65536)|0;
       $unsequencedGroup = $62;
      }
      $63 = $unsequencedGroup;
      $64 = $2;
      $65 = ((($64)) + 244|0);
      $66 = HEAP16[$65>>1]|0;
      $67 = $66&65535;
      $68 = (($67) + 32768)|0;
      $69 = ($63>>>0)>=($68>>>0);
      if ($69) {
       $0 = 0;
       $120 = $0;
       STACKTOP = sp;return ($120|0);
      }
      $70 = $unsequencedGroup;
      $71 = $70 & 65535;
      $unsequencedGroup = $71;
      $72 = $unsequencedGroup;
      $73 = $index;
      $74 = (($72) - ($73))|0;
      $75 = $2;
      $76 = ((($75)) + 244|0);
      $77 = HEAP16[$76>>1]|0;
      $78 = $77&65535;
      $79 = ($74|0)!=($78|0);
      if ($79) {
       $80 = $unsequencedGroup;
       $81 = $index;
       $82 = (($80) - ($81))|0;
       $83 = $82&65535;
       $84 = $2;
       $85 = ((($84)) + 244|0);
       HEAP16[$85>>1] = $83;
       $86 = $2;
       $87 = ((($86)) + 248|0);
       dest=$87; stop=dest+128|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
      } else {
       $88 = $index;
       $89 = (($88>>>0) / 32)&-1;
       $90 = $2;
       $91 = ((($90)) + 248|0);
       $92 = (($91) + ($89<<2)|0);
       $93 = HEAP32[$92>>2]|0;
       $94 = $index;
       $95 = (($94>>>0) % 32)&-1;
       $96 = 1 << $95;
       $97 = $93 & $96;
       $98 = ($97|0)!=(0);
       if ($98) {
        $0 = 0;
        $120 = $0;
        STACKTOP = sp;return ($120|0);
       }
      }
      $99 = $3;
      $100 = ((($99)) + 8|0);
      $101 = $dataLength;
      $102 = (_enet_packet_create($100,$101,2)|0);
      $packet = $102;
      $103 = $packet;
      $104 = ($103|0)==(0|0);
      if (!($104)) {
       $105 = $2;
       $106 = $3;
       $107 = $packet;
       $108 = (_enet_peer_queue_incoming_command($105,$106,$107,0)|0);
       $109 = ($108|0)==(0|0);
       if (!($109)) {
        $110 = $index;
        $111 = (($110>>>0) % 32)&-1;
        $112 = 1 << $111;
        $113 = $index;
        $114 = (($113>>>0) / 32)&-1;
        $115 = $2;
        $116 = ((($115)) + 248|0);
        $117 = (($116) + ($114<<2)|0);
        $118 = HEAP32[$117>>2]|0;
        $119 = $118 | $112;
        HEAP32[$117>>2] = $119;
        $0 = 0;
        $120 = $0;
        STACKTOP = sp;return ($120|0);
       }
      }
      $0 = -1;
      $120 = $0;
      STACKTOP = sp;return ($120|0);
     }
    }
   }
   $0 = -1;
   $120 = $0;
   STACKTOP = sp;return ($120|0);
  }
 } while(0);
 $0 = -1;
 $120 = $0;
 STACKTOP = sp;return ($120|0);
}
function _enet_protocol_handle_send_fragment($host,$peer,$command,$currentData) {
 $host = $host|0;
 $peer = $peer|0;
 $command = $command|0;
 $currentData = $currentData|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $channel = 0, $currentCommand = 0;
 var $currentWindow = 0, $fragmentCount = 0, $fragmentLength = 0, $fragmentNumber = 0, $fragmentOffset = 0, $hostCommand = 0, $incomingCommand = 0, $or$cond = 0, $packet = 0, $startCommand = 0, $startSequenceNumber = 0, $startWindow = 0, $totalLength = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $hostCommand = sp + 72|0;
 $1 = $host;
 $2 = $peer;
 $3 = $command;
 $4 = $currentData;
 $startCommand = 0;
 $5 = $3;
 $6 = ((($5)) + 1|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $9 = $2;
 $10 = ((($9)) + 44|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($8>>>0)>=($11>>>0);
 do {
  if (!($12)) {
   $13 = $2;
   $14 = ((($13)) + 36|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)!=(5);
   if ($16) {
    $17 = $2;
    $18 = ((($17)) + 36|0);
    $19 = HEAP32[$18>>2]|0;
    $20 = ($19|0)!=(6);
    if ($20) {
     break;
    }
   }
   $21 = $3;
   $22 = ((($21)) + 6|0);
   $23 = HEAPU8[$22>>0]|(HEAPU8[$22+1>>0]<<8);
   $24 = (_ntohs($23)|0);
   $25 = $24&65535;
   $fragmentLength = $25;
   $26 = $fragmentLength;
   $27 = $4;
   $28 = HEAP32[$27>>2]|0;
   $29 = (($28) + ($26)|0);
   HEAP32[$27>>2] = $29;
   $30 = $fragmentLength;
   $31 = ($30>>>0)>(1073741824);
   if (!($31)) {
    $32 = $4;
    $33 = HEAP32[$32>>2]|0;
    $34 = $1;
    $35 = ((($34)) + 10356|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($33>>>0)<($36>>>0);
    if (!($37)) {
     $38 = $4;
     $39 = HEAP32[$38>>2]|0;
     $40 = $1;
     $41 = ((($40)) + 10360|0);
     $42 = HEAP32[$41>>2]|0;
     $43 = $1;
     $44 = ((($43)) + 10356|0);
     $45 = HEAP32[$44>>2]|0;
     $46 = (($45) + ($42)|0);
     $47 = ($39>>>0)>($46>>>0);
     if (!($47)) {
      $48 = $3;
      $49 = ((($48)) + 1|0);
      $50 = HEAP8[$49>>0]|0;
      $51 = $50&255;
      $52 = $2;
      $53 = ((($52)) + 40|0);
      $54 = HEAP32[$53>>2]|0;
      $55 = (($54) + (($51*60)|0)|0);
      $channel = $55;
      $56 = $3;
      $57 = ((($56)) + 4|0);
      $58 = HEAPU8[$57>>0]|(HEAPU8[$57+1>>0]<<8);
      $59 = (_ntohs($58)|0);
      $60 = $59&65535;
      $startSequenceNumber = $60;
      $61 = $startSequenceNumber;
      $62 = (($61>>>0) / 4096)&-1;
      $63 = $62&65535;
      $startWindow = $63;
      $64 = $channel;
      $65 = ((($64)) + 38|0);
      $66 = HEAP16[$65>>1]|0;
      $67 = $66&65535;
      $68 = (($67|0) / 4096)&-1;
      $69 = $68&65535;
      $currentWindow = $69;
      $70 = $startSequenceNumber;
      $71 = $channel;
      $72 = ((($71)) + 38|0);
      $73 = HEAP16[$72>>1]|0;
      $74 = $73&65535;
      $75 = ($70>>>0)<($74>>>0);
      if ($75) {
       $76 = $startWindow;
       $77 = $76&65535;
       $78 = (($77) + 16)|0;
       $79 = $78&65535;
       $startWindow = $79;
      }
      $80 = $startWindow;
      $81 = $80&65535;
      $82 = $currentWindow;
      $83 = $82&65535;
      $84 = ($81|0)<($83|0);
      if (!($84)) {
       $85 = $startWindow;
       $86 = $85&65535;
       $87 = $currentWindow;
       $88 = $87&65535;
       $89 = (($88) + 8)|0;
       $90 = (($89) - 1)|0;
       $91 = ($86|0)>=($90|0);
       if (!($91)) {
        $92 = $3;
        $93 = ((($92)) + 12|0);
        $94 = HEAPU8[$93>>0]|(HEAPU8[$93+1>>0]<<8)|(HEAPU8[$93+2>>0]<<16)|(HEAPU8[$93+3>>0]<<24);
        $95 = (_ntohl($94)|0);
        $fragmentNumber = $95;
        $96 = $3;
        $97 = ((($96)) + 8|0);
        $98 = HEAPU8[$97>>0]|(HEAPU8[$97+1>>0]<<8)|(HEAPU8[$97+2>>0]<<16)|(HEAPU8[$97+3>>0]<<24);
        $99 = (_ntohl($98)|0);
        $fragmentCount = $99;
        $100 = $3;
        $101 = ((($100)) + 20|0);
        $102 = HEAPU8[$101>>0]|(HEAPU8[$101+1>>0]<<8)|(HEAPU8[$101+2>>0]<<16)|(HEAPU8[$101+3>>0]<<24);
        $103 = (_ntohl($102)|0);
        $fragmentOffset = $103;
        $104 = $3;
        $105 = ((($104)) + 16|0);
        $106 = HEAPU8[$105>>0]|(HEAPU8[$105+1>>0]<<8)|(HEAPU8[$105+2>>0]<<16)|(HEAPU8[$105+3>>0]<<24);
        $107 = (_ntohl($106)|0);
        $totalLength = $107;
        $108 = $fragmentCount;
        $109 = ($108>>>0)>(1048576);
        if (!($109)) {
         $110 = $fragmentNumber;
         $111 = $fragmentCount;
         $112 = ($110>>>0)>=($111>>>0);
         $113 = $totalLength;
         $114 = ($113>>>0)>(1073741824);
         $or$cond = $112 | $114;
         if (!($or$cond)) {
          $115 = $fragmentOffset;
          $116 = $totalLength;
          $117 = ($115>>>0)>=($116>>>0);
          if (!($117)) {
           $118 = $fragmentLength;
           $119 = $totalLength;
           $120 = $fragmentOffset;
           $121 = (($119) - ($120))|0;
           $122 = ($118>>>0)>($121>>>0);
           if (!($122)) {
            $123 = $channel;
            $124 = ((($123)) + 44|0);
            $125 = ((($124)) + 4|0);
            $126 = HEAP32[$125>>2]|0;
            $currentCommand = $126;
            while(1) {
             $127 = $currentCommand;
             $128 = $channel;
             $129 = ((($128)) + 44|0);
             $130 = ($127|0)!=($129|0);
             if (!($130)) {
              break;
             }
             $131 = $currentCommand;
             $incomingCommand = $131;
             $132 = $startSequenceNumber;
             $133 = $channel;
             $134 = ((($133)) + 38|0);
             $135 = HEAP16[$134>>1]|0;
             $136 = $135&65535;
             $137 = ($132>>>0)>=($136>>>0);
             $138 = $incomingCommand;
             $139 = ((($138)) + 8|0);
             $140 = HEAP16[$139>>1]|0;
             $141 = $140&65535;
             $142 = $channel;
             $143 = ((($142)) + 38|0);
             $144 = HEAP16[$143>>1]|0;
             $145 = $144&65535;
             if ($137) {
              $146 = ($141|0)<($145|0);
              if (!($146)) {
               label = 24;
              }
             } else {
              $147 = ($141|0)>=($145|0);
              if ($147) {
               break;
              } else {
               label = 24;
              }
             }
             if ((label|0) == 24) {
              label = 0;
              $148 = $incomingCommand;
              $149 = ((($148)) + 8|0);
              $150 = HEAP16[$149>>1]|0;
              $151 = $150&65535;
              $152 = $startSequenceNumber;
              $153 = ($151>>>0)<=($152>>>0);
              if ($153) {
               label = 25;
               break;
              }
             }
             $179 = $currentCommand;
             $180 = ((($179)) + 4|0);
             $181 = HEAP32[$180>>2]|0;
             $currentCommand = $181;
            }
            L30: do {
             if ((label|0) == 25) {
              $154 = $incomingCommand;
              $155 = ((($154)) + 8|0);
              $156 = HEAP16[$155>>1]|0;
              $157 = $156&65535;
              $158 = $startSequenceNumber;
              $159 = ($157>>>0)<($158>>>0);
              if (!($159)) {
               $160 = $incomingCommand;
               $161 = ((($160)) + 12|0);
               $162 = HEAP8[$161>>0]|0;
               $163 = $162&255;
               $164 = $163 & 15;
               $165 = ($164|0)!=(8);
               do {
                if (!($165)) {
                 $166 = $totalLength;
                 $167 = $incomingCommand;
                 $168 = ((($167)) + 72|0);
                 $169 = HEAP32[$168>>2]|0;
                 $170 = ((($169)) + 12|0);
                 $171 = HEAP32[$170>>2]|0;
                 $172 = ($166|0)!=($171|0);
                 if ($172) {
                  break;
                 }
                 $173 = $fragmentCount;
                 $174 = $incomingCommand;
                 $175 = ((($174)) + 60|0);
                 $176 = HEAP32[$175>>2]|0;
                 $177 = ($173|0)!=($176|0);
                 if ($177) {
                  break;
                 }
                 $178 = $incomingCommand;
                 $startCommand = $178;
                 break L30;
                }
               } while(0);
               $0 = -1;
               $257 = $0;
               STACKTOP = sp;return ($257|0);
              }
             }
            } while(0);
            $182 = $startCommand;
            $183 = ($182|0)==(0|0);
            if ($183) {
             $184 = $3;
             dest=$hostCommand; src=$184; stop=dest+48|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
             $185 = $totalLength;
             $186 = (_enet_packet_create(0,$185,1)|0);
             $packet = $186;
             $187 = $packet;
             $188 = ($187|0)==(0|0);
             if ($188) {
              $0 = -1;
              $257 = $0;
              STACKTOP = sp;return ($257|0);
             }
             $189 = $startSequenceNumber;
             $190 = $189&65535;
             $191 = ((($hostCommand)) + 2|0);
             HEAP8[$191>>0]=$190&255;HEAP8[$191+1>>0]=$190>>8;
             $192 = $2;
             $193 = $packet;
             $194 = $fragmentCount;
             $195 = (_enet_peer_queue_incoming_command($192,$hostCommand,$193,$194)|0);
             $startCommand = $195;
             $196 = $startCommand;
             $197 = ($196|0)==(0|0);
             if ($197) {
              $0 = -1;
              $257 = $0;
              STACKTOP = sp;return ($257|0);
             }
            }
            $198 = $fragmentNumber;
            $199 = (($198>>>0) / 32)&-1;
            $200 = $startCommand;
            $201 = ((($200)) + 68|0);
            $202 = HEAP32[$201>>2]|0;
            $203 = (($202) + ($199<<2)|0);
            $204 = HEAP32[$203>>2]|0;
            $205 = $fragmentNumber;
            $206 = (($205>>>0) % 32)&-1;
            $207 = 1 << $206;
            $208 = $204 & $207;
            $209 = ($208|0)==(0);
            do {
             if ($209) {
              $210 = $startCommand;
              $211 = ((($210)) + 64|0);
              $212 = HEAP32[$211>>2]|0;
              $213 = (($212) + -1)|0;
              HEAP32[$211>>2] = $213;
              $214 = $fragmentNumber;
              $215 = (($214>>>0) % 32)&-1;
              $216 = 1 << $215;
              $217 = $fragmentNumber;
              $218 = (($217>>>0) / 32)&-1;
              $219 = $startCommand;
              $220 = ((($219)) + 68|0);
              $221 = HEAP32[$220>>2]|0;
              $222 = (($221) + ($218<<2)|0);
              $223 = HEAP32[$222>>2]|0;
              $224 = $223 | $216;
              HEAP32[$222>>2] = $224;
              $225 = $fragmentOffset;
              $226 = $fragmentLength;
              $227 = (($225) + ($226))|0;
              $228 = $startCommand;
              $229 = ((($228)) + 72|0);
              $230 = HEAP32[$229>>2]|0;
              $231 = ((($230)) + 12|0);
              $232 = HEAP32[$231>>2]|0;
              $233 = ($227>>>0)>($232>>>0);
              if ($233) {
               $234 = $startCommand;
               $235 = ((($234)) + 72|0);
               $236 = HEAP32[$235>>2]|0;
               $237 = ((($236)) + 12|0);
               $238 = HEAP32[$237>>2]|0;
               $239 = $fragmentOffset;
               $240 = (($238) - ($239))|0;
               $fragmentLength = $240;
              }
              $241 = $startCommand;
              $242 = ((($241)) + 72|0);
              $243 = HEAP32[$242>>2]|0;
              $244 = ((($243)) + 8|0);
              $245 = HEAP32[$244>>2]|0;
              $246 = $fragmentOffset;
              $247 = (($245) + ($246)|0);
              $248 = $3;
              $249 = ((($248)) + 24|0);
              $250 = $fragmentLength;
              _memcpy(($247|0),($249|0),($250|0))|0;
              $251 = $startCommand;
              $252 = ((($251)) + 64|0);
              $253 = HEAP32[$252>>2]|0;
              $254 = ($253>>>0)<=(0);
              if (!($254)) {
               break;
              }
              $255 = $2;
              $256 = $channel;
              _enet_peer_dispatch_incoming_reliable_commands($255,$256);
             }
            } while(0);
            $0 = 0;
            $257 = $0;
            STACKTOP = sp;return ($257|0);
           }
          }
         }
        }
        $0 = -1;
        $257 = $0;
        STACKTOP = sp;return ($257|0);
       }
      }
      $0 = 0;
      $257 = $0;
      STACKTOP = sp;return ($257|0);
     }
    }
   }
   $0 = -1;
   $257 = $0;
   STACKTOP = sp;return ($257|0);
  }
 } while(0);
 $0 = -1;
 $257 = $0;
 STACKTOP = sp;return ($257|0);
}
function _enet_protocol_handle_bandwidth_limit($host,$peer,$command) {
 $host = $host|0;
 $peer = $peer|0;
 $command = $command|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $host;
 $2 = $peer;
 $3 = $command;
 $4 = $2;
 $5 = ((($4)) + 36|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)!=(5);
 if ($7) {
  $8 = $2;
  $9 = ((($8)) + 36|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ($10|0)!=(6);
  if ($11) {
   $0 = -1;
   $78 = $0;
   STACKTOP = sp;return ($78|0);
  }
 }
 $12 = $2;
 $13 = ((($12)) + 48|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ($14|0)!=(0);
 if ($15) {
  $16 = $1;
  $17 = ((($16)) + 10388|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = (($18) + -1)|0;
  HEAP32[$17>>2] = $19;
 }
 $20 = $3;
 $21 = ((($20)) + 4|0);
 $22 = HEAPU8[$21>>0]|(HEAPU8[$21+1>>0]<<8)|(HEAPU8[$21+2>>0]<<16)|(HEAPU8[$21+3>>0]<<24);
 $23 = (_ntohl($22)|0);
 $24 = $2;
 $25 = ((($24)) + 48|0);
 HEAP32[$25>>2] = $23;
 $26 = $3;
 $27 = ((($26)) + 8|0);
 $28 = HEAPU8[$27>>0]|(HEAPU8[$27+1>>0]<<8)|(HEAPU8[$27+2>>0]<<16)|(HEAPU8[$27+3>>0]<<24);
 $29 = (_ntohl($28)|0);
 $30 = $2;
 $31 = ((($30)) + 52|0);
 HEAP32[$31>>2] = $29;
 $32 = $2;
 $33 = ((($32)) + 48|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = ($34|0)!=(0);
 if ($35) {
  $36 = $1;
  $37 = ((($36)) + 10388|0);
  $38 = HEAP32[$37>>2]|0;
  $39 = (($38) + 1)|0;
  HEAP32[$37>>2] = $39;
 }
 $40 = $2;
 $41 = ((($40)) + 48|0);
 $42 = HEAP32[$41>>2]|0;
 $43 = ($42|0)==(0);
 if ($43) {
  $44 = $1;
  $45 = ((($44)) + 16|0);
  $46 = HEAP32[$45>>2]|0;
  $47 = ($46|0)==(0);
  if ($47) {
   $48 = $2;
   $49 = ((($48)) + 180|0);
   HEAP32[$49>>2] = 65536;
  } else {
   label = 11;
  }
 } else {
  label = 11;
 }
 if ((label|0) == 11) {
  $50 = $2;
  $51 = ((($50)) + 48|0);
  $52 = HEAP32[$51>>2]|0;
  $53 = $1;
  $54 = ((($53)) + 16|0);
  $55 = HEAP32[$54>>2]|0;
  $56 = ($52>>>0)<($55>>>0);
  if ($56) {
   $57 = $2;
   $58 = ((($57)) + 48|0);
   $59 = HEAP32[$58>>2]|0;
   $64 = $59;
  } else {
   $60 = $1;
   $61 = ((($60)) + 16|0);
   $62 = HEAP32[$61>>2]|0;
   $64 = $62;
  }
  $63 = (($64>>>0) / 65536)&-1;
  $65 = $63<<12;
  $66 = $2;
  $67 = ((($66)) + 180|0);
  HEAP32[$67>>2] = $65;
 }
 $68 = $2;
 $69 = ((($68)) + 180|0);
 $70 = HEAP32[$69>>2]|0;
 $71 = ($70>>>0)<(4096);
 $72 = $2;
 $73 = ((($72)) + 180|0);
 if ($71) {
  HEAP32[$73>>2] = 4096;
 } else {
  $74 = HEAP32[$73>>2]|0;
  $75 = ($74>>>0)>(65536);
  if ($75) {
   $76 = $2;
   $77 = ((($76)) + 180|0);
   HEAP32[$77>>2] = 65536;
  }
 }
 $0 = 0;
 $78 = $0;
 STACKTOP = sp;return ($78|0);
}
function _enet_protocol_handle_throttle_configure($host,$peer,$command) {
 $host = $host|0;
 $peer = $peer|0;
 $command = $command|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $host;
 $2 = $peer;
 $3 = $command;
 $4 = $2;
 $5 = ((($4)) + 36|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)!=(5);
 if ($7) {
  $8 = $2;
  $9 = ((($8)) + 36|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ($10|0)!=(6);
  if ($11) {
   $0 = -1;
   $30 = $0;
   STACKTOP = sp;return ($30|0);
  }
 }
 $12 = $3;
 $13 = ((($12)) + 4|0);
 $14 = HEAPU8[$13>>0]|(HEAPU8[$13+1>>0]<<8)|(HEAPU8[$13+2>>0]<<16)|(HEAPU8[$13+3>>0]<<24);
 $15 = (_ntohl($14)|0);
 $16 = $2;
 $17 = ((($16)) + 132|0);
 HEAP32[$17>>2] = $15;
 $18 = $3;
 $19 = ((($18)) + 8|0);
 $20 = HEAPU8[$19>>0]|(HEAPU8[$19+1>>0]<<8)|(HEAPU8[$19+2>>0]<<16)|(HEAPU8[$19+3>>0]<<24);
 $21 = (_ntohl($20)|0);
 $22 = $2;
 $23 = ((($22)) + 124|0);
 HEAP32[$23>>2] = $21;
 $24 = $3;
 $25 = ((($24)) + 12|0);
 $26 = HEAPU8[$25>>0]|(HEAPU8[$25+1>>0]<<8)|(HEAPU8[$25+2>>0]<<16)|(HEAPU8[$25+3>>0]<<24);
 $27 = (_ntohl($26)|0);
 $28 = $2;
 $29 = ((($28)) + 128|0);
 HEAP32[$29>>2] = $27;
 $0 = 0;
 $30 = $0;
 STACKTOP = sp;return ($30|0);
}
function _enet_protocol_handle_send_unreliable_fragment($host,$peer,$command,$currentData) {
 $host = $host|0;
 $peer = $peer|0;
 $command = $command|0;
 $currentData = $currentData|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $channel = 0, $currentCommand = 0, $currentWindow = 0, $fragmentCount = 0, $fragmentLength = 0, $fragmentNumber = 0, $fragmentOffset = 0, $incomingCommand = 0, $or$cond = 0, $packet = 0, $reliableSequenceNumber = 0, $reliableWindow = 0, $startCommand = 0, $startSequenceNumber = 0, $totalLength = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $host;
 $2 = $peer;
 $3 = $command;
 $4 = $currentData;
 $startCommand = 0;
 $5 = $3;
 $6 = ((($5)) + 1|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $9 = $2;
 $10 = ((($9)) + 44|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($8>>>0)>=($11>>>0);
 do {
  if (!($12)) {
   $13 = $2;
   $14 = ((($13)) + 36|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)!=(5);
   if ($16) {
    $17 = $2;
    $18 = ((($17)) + 36|0);
    $19 = HEAP32[$18>>2]|0;
    $20 = ($19|0)!=(6);
    if ($20) {
     break;
    }
   }
   $21 = $3;
   $22 = ((($21)) + 6|0);
   $23 = HEAPU8[$22>>0]|(HEAPU8[$22+1>>0]<<8);
   $24 = (_ntohs($23)|0);
   $25 = $24&65535;
   $fragmentLength = $25;
   $26 = $fragmentLength;
   $27 = $4;
   $28 = HEAP32[$27>>2]|0;
   $29 = (($28) + ($26)|0);
   HEAP32[$27>>2] = $29;
   $30 = $fragmentLength;
   $31 = ($30>>>0)>(1073741824);
   if (!($31)) {
    $32 = $4;
    $33 = HEAP32[$32>>2]|0;
    $34 = $1;
    $35 = ((($34)) + 10356|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($33>>>0)<($36>>>0);
    if (!($37)) {
     $38 = $4;
     $39 = HEAP32[$38>>2]|0;
     $40 = $1;
     $41 = ((($40)) + 10360|0);
     $42 = HEAP32[$41>>2]|0;
     $43 = $1;
     $44 = ((($43)) + 10356|0);
     $45 = HEAP32[$44>>2]|0;
     $46 = (($45) + ($42)|0);
     $47 = ($39>>>0)>($46>>>0);
     if (!($47)) {
      $48 = $3;
      $49 = ((($48)) + 1|0);
      $50 = HEAP8[$49>>0]|0;
      $51 = $50&255;
      $52 = $2;
      $53 = ((($52)) + 40|0);
      $54 = HEAP32[$53>>2]|0;
      $55 = (($54) + (($51*60)|0)|0);
      $channel = $55;
      $56 = $3;
      $57 = ((($56)) + 2|0);
      $58 = HEAPU8[$57>>0]|(HEAPU8[$57+1>>0]<<8);
      $59 = $58&65535;
      $reliableSequenceNumber = $59;
      $60 = $3;
      $61 = ((($60)) + 4|0);
      $62 = HEAPU8[$61>>0]|(HEAPU8[$61+1>>0]<<8);
      $63 = (_ntohs($62)|0);
      $64 = $63&65535;
      $startSequenceNumber = $64;
      $65 = $reliableSequenceNumber;
      $66 = (($65>>>0) / 4096)&-1;
      $67 = $66&65535;
      $reliableWindow = $67;
      $68 = $channel;
      $69 = ((($68)) + 38|0);
      $70 = HEAP16[$69>>1]|0;
      $71 = $70&65535;
      $72 = (($71|0) / 4096)&-1;
      $73 = $72&65535;
      $currentWindow = $73;
      $74 = $reliableSequenceNumber;
      $75 = $channel;
      $76 = ((($75)) + 38|0);
      $77 = HEAP16[$76>>1]|0;
      $78 = $77&65535;
      $79 = ($74>>>0)<($78>>>0);
      if ($79) {
       $80 = $reliableWindow;
       $81 = $80&65535;
       $82 = (($81) + 16)|0;
       $83 = $82&65535;
       $reliableWindow = $83;
      }
      $84 = $reliableWindow;
      $85 = $84&65535;
      $86 = $currentWindow;
      $87 = $86&65535;
      $88 = ($85|0)<($87|0);
      if (!($88)) {
       $89 = $reliableWindow;
       $90 = $89&65535;
       $91 = $currentWindow;
       $92 = $91&65535;
       $93 = (($92) + 8)|0;
       $94 = (($93) - 1)|0;
       $95 = ($90|0)>=($94|0);
       if (!($95)) {
        $96 = $reliableSequenceNumber;
        $97 = $channel;
        $98 = ((($97)) + 38|0);
        $99 = HEAP16[$98>>1]|0;
        $100 = $99&65535;
        $101 = ($96|0)==($100|0);
        if ($101) {
         $102 = $startSequenceNumber;
         $103 = $channel;
         $104 = ((($103)) + 40|0);
         $105 = HEAP16[$104>>1]|0;
         $106 = $105&65535;
         $107 = ($102>>>0)<=($106>>>0);
         if ($107) {
          $0 = 0;
          $282 = $0;
          STACKTOP = sp;return ($282|0);
         }
        }
        $108 = $3;
        $109 = ((($108)) + 12|0);
        $110 = HEAPU8[$109>>0]|(HEAPU8[$109+1>>0]<<8)|(HEAPU8[$109+2>>0]<<16)|(HEAPU8[$109+3>>0]<<24);
        $111 = (_ntohl($110)|0);
        $fragmentNumber = $111;
        $112 = $3;
        $113 = ((($112)) + 8|0);
        $114 = HEAPU8[$113>>0]|(HEAPU8[$113+1>>0]<<8)|(HEAPU8[$113+2>>0]<<16)|(HEAPU8[$113+3>>0]<<24);
        $115 = (_ntohl($114)|0);
        $fragmentCount = $115;
        $116 = $3;
        $117 = ((($116)) + 20|0);
        $118 = HEAPU8[$117>>0]|(HEAPU8[$117+1>>0]<<8)|(HEAPU8[$117+2>>0]<<16)|(HEAPU8[$117+3>>0]<<24);
        $119 = (_ntohl($118)|0);
        $fragmentOffset = $119;
        $120 = $3;
        $121 = ((($120)) + 16|0);
        $122 = HEAPU8[$121>>0]|(HEAPU8[$121+1>>0]<<8)|(HEAPU8[$121+2>>0]<<16)|(HEAPU8[$121+3>>0]<<24);
        $123 = (_ntohl($122)|0);
        $totalLength = $123;
        $124 = $fragmentCount;
        $125 = ($124>>>0)>(1048576);
        if (!($125)) {
         $126 = $fragmentNumber;
         $127 = $fragmentCount;
         $128 = ($126>>>0)>=($127>>>0);
         $129 = $totalLength;
         $130 = ($129>>>0)>(1073741824);
         $or$cond = $128 | $130;
         if (!($or$cond)) {
          $131 = $fragmentOffset;
          $132 = $totalLength;
          $133 = ($131>>>0)>=($132>>>0);
          if (!($133)) {
           $134 = $fragmentLength;
           $135 = $totalLength;
           $136 = $fragmentOffset;
           $137 = (($135) - ($136))|0;
           $138 = ($134>>>0)>($137>>>0);
           if (!($138)) {
            $139 = $channel;
            $140 = ((($139)) + 52|0);
            $141 = ((($140)) + 4|0);
            $142 = HEAP32[$141>>2]|0;
            $currentCommand = $142;
            L26: while(1) {
             $143 = $currentCommand;
             $144 = $channel;
             $145 = ((($144)) + 52|0);
             $146 = ($143|0)!=($145|0);
             if (!($146)) {
              break;
             }
             $147 = $currentCommand;
             $incomingCommand = $147;
             $148 = $reliableSequenceNumber;
             $149 = $channel;
             $150 = ((($149)) + 38|0);
             $151 = HEAP16[$150>>1]|0;
             $152 = $151&65535;
             $153 = ($148>>>0)>=($152>>>0);
             $154 = $incomingCommand;
             $155 = ((($154)) + 8|0);
             $156 = HEAP16[$155>>1]|0;
             $157 = $156&65535;
             $158 = $channel;
             $159 = ((($158)) + 38|0);
             $160 = HEAP16[$159>>1]|0;
             $161 = $160&65535;
             if ($153) {
              $162 = ($157|0)<($161|0);
              if (!($162)) {
               label = 27;
              }
             } else {
              $163 = ($157|0)>=($161|0);
              if ($163) {
               break;
              } else {
               label = 27;
              }
             }
             do {
              if ((label|0) == 27) {
               label = 0;
               $164 = $incomingCommand;
               $165 = ((($164)) + 8|0);
               $166 = HEAP16[$165>>1]|0;
               $167 = $166&65535;
               $168 = $reliableSequenceNumber;
               $169 = ($167>>>0)<($168>>>0);
               if ($169) {
                break L26;
               }
               $170 = $incomingCommand;
               $171 = ((($170)) + 8|0);
               $172 = HEAP16[$171>>1]|0;
               $173 = $172&65535;
               $174 = $reliableSequenceNumber;
               $175 = ($173>>>0)>($174>>>0);
               if ($175) {
                break;
               }
               $176 = $incomingCommand;
               $177 = ((($176)) + 10|0);
               $178 = HEAP16[$177>>1]|0;
               $179 = $178&65535;
               $180 = $startSequenceNumber;
               $181 = ($179>>>0)<=($180>>>0);
               if ($181) {
                label = 30;
                break L26;
               }
              }
             } while(0);
             $207 = $currentCommand;
             $208 = ((($207)) + 4|0);
             $209 = HEAP32[$208>>2]|0;
             $currentCommand = $209;
            }
            L37: do {
             if ((label|0) == 30) {
              $182 = $incomingCommand;
              $183 = ((($182)) + 10|0);
              $184 = HEAP16[$183>>1]|0;
              $185 = $184&65535;
              $186 = $startSequenceNumber;
              $187 = ($185>>>0)<($186>>>0);
              if (!($187)) {
               $188 = $incomingCommand;
               $189 = ((($188)) + 12|0);
               $190 = HEAP8[$189>>0]|0;
               $191 = $190&255;
               $192 = $191 & 15;
               $193 = ($192|0)!=(12);
               do {
                if (!($193)) {
                 $194 = $totalLength;
                 $195 = $incomingCommand;
                 $196 = ((($195)) + 72|0);
                 $197 = HEAP32[$196>>2]|0;
                 $198 = ((($197)) + 12|0);
                 $199 = HEAP32[$198>>2]|0;
                 $200 = ($194|0)!=($199|0);
                 if ($200) {
                  break;
                 }
                 $201 = $fragmentCount;
                 $202 = $incomingCommand;
                 $203 = ((($202)) + 60|0);
                 $204 = HEAP32[$203>>2]|0;
                 $205 = ($201|0)!=($204|0);
                 if ($205) {
                  break;
                 }
                 $206 = $incomingCommand;
                 $startCommand = $206;
                 break L37;
                }
               } while(0);
               $0 = -1;
               $282 = $0;
               STACKTOP = sp;return ($282|0);
              }
             }
            } while(0);
            $210 = $startCommand;
            $211 = ($210|0)==(0|0);
            do {
             if ($211) {
              $212 = $totalLength;
              $213 = (_enet_packet_create(0,$212,8)|0);
              $packet = $213;
              $214 = $packet;
              $215 = ($214|0)==(0|0);
              if ($215) {
               $0 = -1;
               $282 = $0;
               STACKTOP = sp;return ($282|0);
              }
              $216 = $2;
              $217 = $3;
              $218 = $packet;
              $219 = $fragmentCount;
              $220 = (_enet_peer_queue_incoming_command($216,$217,$218,$219)|0);
              $startCommand = $220;
              $221 = $startCommand;
              $222 = ($221|0)==(0|0);
              if (!($222)) {
               break;
              }
              $0 = -1;
              $282 = $0;
              STACKTOP = sp;return ($282|0);
             }
            } while(0);
            $223 = $fragmentNumber;
            $224 = (($223>>>0) / 32)&-1;
            $225 = $startCommand;
            $226 = ((($225)) + 68|0);
            $227 = HEAP32[$226>>2]|0;
            $228 = (($227) + ($224<<2)|0);
            $229 = HEAP32[$228>>2]|0;
            $230 = $fragmentNumber;
            $231 = (($230>>>0) % 32)&-1;
            $232 = 1 << $231;
            $233 = $229 & $232;
            $234 = ($233|0)==(0);
            do {
             if ($234) {
              $235 = $startCommand;
              $236 = ((($235)) + 64|0);
              $237 = HEAP32[$236>>2]|0;
              $238 = (($237) + -1)|0;
              HEAP32[$236>>2] = $238;
              $239 = $fragmentNumber;
              $240 = (($239>>>0) % 32)&-1;
              $241 = 1 << $240;
              $242 = $fragmentNumber;
              $243 = (($242>>>0) / 32)&-1;
              $244 = $startCommand;
              $245 = ((($244)) + 68|0);
              $246 = HEAP32[$245>>2]|0;
              $247 = (($246) + ($243<<2)|0);
              $248 = HEAP32[$247>>2]|0;
              $249 = $248 | $241;
              HEAP32[$247>>2] = $249;
              $250 = $fragmentOffset;
              $251 = $fragmentLength;
              $252 = (($250) + ($251))|0;
              $253 = $startCommand;
              $254 = ((($253)) + 72|0);
              $255 = HEAP32[$254>>2]|0;
              $256 = ((($255)) + 12|0);
              $257 = HEAP32[$256>>2]|0;
              $258 = ($252>>>0)>($257>>>0);
              if ($258) {
               $259 = $startCommand;
               $260 = ((($259)) + 72|0);
               $261 = HEAP32[$260>>2]|0;
               $262 = ((($261)) + 12|0);
               $263 = HEAP32[$262>>2]|0;
               $264 = $fragmentOffset;
               $265 = (($263) - ($264))|0;
               $fragmentLength = $265;
              }
              $266 = $startCommand;
              $267 = ((($266)) + 72|0);
              $268 = HEAP32[$267>>2]|0;
              $269 = ((($268)) + 8|0);
              $270 = HEAP32[$269>>2]|0;
              $271 = $fragmentOffset;
              $272 = (($270) + ($271)|0);
              $273 = $3;
              $274 = ((($273)) + 24|0);
              $275 = $fragmentLength;
              _memcpy(($272|0),($274|0),($275|0))|0;
              $276 = $startCommand;
              $277 = ((($276)) + 64|0);
              $278 = HEAP32[$277>>2]|0;
              $279 = ($278>>>0)<=(0);
              if (!($279)) {
               break;
              }
              $280 = $2;
              $281 = $channel;
              _enet_peer_dispatch_incoming_unreliable_commands($280,$281);
             }
            } while(0);
            $0 = 0;
            $282 = $0;
            STACKTOP = sp;return ($282|0);
           }
          }
         }
        }
        $0 = -1;
        $282 = $0;
        STACKTOP = sp;return ($282|0);
       }
      }
      $0 = 0;
      $282 = $0;
      STACKTOP = sp;return ($282|0);
     }
    }
   }
   $0 = -1;
   $282 = $0;
   STACKTOP = sp;return ($282|0);
  }
 } while(0);
 $0 = -1;
 $282 = $0;
 STACKTOP = sp;return ($282|0);
}
function _enet_initialize() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _enet_deinitialize() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _enet_host_random_seed() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_time((0|0))|0);
 return ($0|0);
}
function _enet_time_get() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $timeVal = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $timeVal = sp;
 (_gettimeofday(($timeVal|0),(0|0))|0);
 $0 = HEAP32[$timeVal>>2]|0;
 $1 = ($0*1000)|0;
 $2 = ((($timeVal)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3|0) / 1000)&-1;
 $5 = (($1) + ($4))|0;
 $6 = HEAP32[156503]|0;
 $7 = (($5) - ($6))|0;
 STACKTOP = sp;return ($7|0);
}
function _enet_address_set_host($address,$name) {
 $address = $address|0;
 $name = $name|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $hostEntry = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $address;
 $2 = $name;
 $hostEntry = 0;
 $3 = $2;
 $4 = (_gethostbyname(($3|0))|0);
 $hostEntry = $4;
 $5 = $hostEntry;
 $6 = ($5|0)==(0|0);
 if (!($6)) {
  $7 = $hostEntry;
  $8 = ((($7)) + 8|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)!=(2);
  if (!($10)) {
   $15 = $hostEntry;
   $16 = ((($15)) + 16|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = HEAP32[$17>>2]|0;
   $19 = HEAP32[$18>>2]|0;
   $20 = $1;
   HEAP32[$20>>2] = $19;
   $0 = 0;
   $21 = $0;
   STACKTOP = sp;return ($21|0);
  }
 }
 $11 = $2;
 $12 = $1;
 $13 = (_inet_aton($11,$12)|0);
 $14 = ($13|0)!=(0);
 if ($14) {
  $0 = 0;
  $21 = $0;
  STACKTOP = sp;return ($21|0);
 } else {
  $0 = -1;
  $21 = $0;
  STACKTOP = sp;return ($21|0);
 }
 return (0)|0;
}
function _enet_address_get_host_ip($address,$name,$nameLength) {
 $address = $address|0;
 $name = $name|0;
 $nameLength = $nameLength|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $addr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 20|0;
 $1 = $address;
 $2 = $name;
 $3 = $nameLength;
 $4 = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$4>>2]|0;
 $5 = (_inet_ntoa($$byval_copy)|0);
 $addr = $5;
 $6 = $addr;
 $7 = ($6|0)!=(0|0);
 if ($7) {
  $8 = $2;
  $9 = $addr;
  $10 = $3;
  (_strncpy($8,$9,$10)|0);
  $0 = 0;
  $11 = $0;
  STACKTOP = sp;return ($11|0);
 } else {
  $0 = -1;
  $11 = $0;
  STACKTOP = sp;return ($11|0);
 }
 return (0)|0;
}
function _enet_socket_bind($socket,$address) {
 $socket = $socket|0;
 $address = $address|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $sin = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $sin = sp;
 $0 = $socket;
 $1 = $address;
 ;HEAP32[$sin>>2]=0|0;HEAP32[$sin+4>>2]=0|0;HEAP32[$sin+8>>2]=0|0;HEAP32[$sin+12>>2]=0|0;
 HEAP16[$sin>>1] = 2;
 $2 = $1;
 $3 = ($2|0)!=(0|0);
 if ($3) {
  $4 = $1;
  $5 = ((($4)) + 4|0);
  $6 = HEAP16[$5>>1]|0;
  $7 = (_htons($6)|0);
  $8 = ((($sin)) + 2|0);
  HEAP16[$8>>1] = $7;
  $9 = $1;
  $10 = HEAP32[$9>>2]|0;
  $11 = ((($sin)) + 4|0);
  HEAP32[$11>>2] = $10;
  $14 = $0;
  $15 = (_bind($14,$sin,16)|0);
  STACKTOP = sp;return ($15|0);
 } else {
  $12 = ((($sin)) + 2|0);
  HEAP16[$12>>1] = 0;
  $13 = ((($sin)) + 4|0);
  HEAP32[$13>>2] = 0;
  $14 = $0;
  $15 = (_bind($14,$sin,16)|0);
  STACKTOP = sp;return ($15|0);
 }
 return (0)|0;
}
function _enet_socket_get_address($socket,$address) {
 $socket = $socket|0;
 $address = $address|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $sin = 0, $sinLength = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $sin = sp + 8|0;
 $sinLength = sp;
 $1 = $socket;
 $2 = $address;
 HEAP32[$sinLength>>2] = 16;
 $3 = $1;
 $4 = (_getsockname($3,$sin,$sinLength)|0);
 $5 = ($4|0)==(-1);
 if ($5) {
  $0 = -1;
  $14 = $0;
  STACKTOP = sp;return ($14|0);
 } else {
  $6 = ((($sin)) + 4|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = $2;
  HEAP32[$8>>2] = $7;
  $9 = ((($sin)) + 2|0);
  $10 = HEAP16[$9>>1]|0;
  $11 = (_ntohs($10)|0);
  $12 = $2;
  $13 = ((($12)) + 4|0);
  HEAP16[$13>>1] = $11;
  $0 = 0;
  $14 = $0;
  STACKTOP = sp;return ($14|0);
 }
 return (0)|0;
}
function _enet_socket_create($type) {
 $type = $type|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $type;
 $1 = $0;
 $2 = ($1|0)==(2);
 $3 = $2 ? 2 : 1;
 $4 = (_socket(2,$3,0)|0);
 STACKTOP = sp;return ($4|0);
}
function _enet_socket_set_option($socket,$option,$value) {
 $socket = $socket|0;
 $option = $option|0;
 $value = $value|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $result = 0, $timeVal = 0, $timeVal1 = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $2 = sp + 28|0;
 $timeVal = sp + 16|0;
 $timeVal1 = sp + 8|0;
 $0 = $socket;
 $1 = $option;
 HEAP32[$2>>2] = $value;
 $result = -1;
 $3 = $1;
 switch ($3|0) {
 case 1:  {
  $4 = $0;
  HEAP32[$vararg_buffer>>2] = $2;
  $5 = (_ioctl($4,21537,$vararg_buffer)|0);
  $result = $5;
  break;
 }
 case 2:  {
  $6 = $0;
  $7 = (_setsockopt($6,1,6,$2,4)|0);
  $result = $7;
  break;
 }
 case 5:  {
  $8 = $0;
  $9 = (_setsockopt($8,1,2,$2,4)|0);
  $result = $9;
  break;
 }
 case 3:  {
  $10 = $0;
  $11 = (_setsockopt($10,1,8,$2,4)|0);
  $result = $11;
  break;
 }
 case 4:  {
  $12 = $0;
  $13 = (_setsockopt($12,1,7,$2,4)|0);
  $result = $13;
  break;
 }
 case 6:  {
  $14 = HEAP32[$2>>2]|0;
  $15 = (($14|0) / 1000)&-1;
  HEAP32[$timeVal>>2] = $15;
  $16 = HEAP32[$2>>2]|0;
  $17 = (($16|0) % 1000)&-1;
  $18 = ($17*1000)|0;
  $19 = ((($timeVal)) + 4|0);
  HEAP32[$19>>2] = $18;
  $20 = $0;
  $21 = (_setsockopt($20,1,20,$timeVal,8)|0);
  $result = $21;
  break;
 }
 case 7:  {
  $22 = HEAP32[$2>>2]|0;
  $23 = (($22|0) / 1000)&-1;
  HEAP32[$timeVal1>>2] = $23;
  $24 = HEAP32[$2>>2]|0;
  $25 = (($24|0) % 1000)&-1;
  $26 = ($25*1000)|0;
  $27 = ((($timeVal1)) + 4|0);
  HEAP32[$27>>2] = $26;
  $28 = $0;
  $29 = (_setsockopt($28,1,21,$timeVal1,8)|0);
  $result = $29;
  break;
 }
 case 9:  {
  $30 = $0;
  $31 = (_setsockopt($30,6,1,$2,4)|0);
  $result = $31;
  break;
 }
 default: {
 }
 }
 $32 = $result;
 $33 = ($32|0)==(-1);
 $34 = $33 ? -1 : 0;
 STACKTOP = sp;return ($34|0);
}
function _enet_socket_destroy($socket) {
 $socket = $socket|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $socket;
 $1 = $0;
 $2 = ($1|0)!=(-1);
 if ($2) {
  $3 = $0;
  (_close($3)|0);
 }
 STACKTOP = sp;return;
}
function _enet_socket_send($socket,$address,$buffers,$bufferCount) {
 $socket = $socket|0;
 $address = $address|0;
 $buffers = $buffers|0;
 $bufferCount = $bufferCount|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $msgHdr = 0, $sentLength = 0, $sin = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $msgHdr = sp + 24|0;
 $sin = sp + 8|0;
 $1 = $socket;
 $2 = $address;
 $3 = $buffers;
 $4 = $bufferCount;
 ;HEAP32[$msgHdr>>2]=0|0;HEAP32[$msgHdr+4>>2]=0|0;HEAP32[$msgHdr+8>>2]=0|0;HEAP32[$msgHdr+12>>2]=0|0;HEAP32[$msgHdr+16>>2]=0|0;HEAP32[$msgHdr+20>>2]=0|0;HEAP32[$msgHdr+24>>2]=0|0;
 $5 = $2;
 $6 = ($5|0)!=(0|0);
 if ($6) {
  ;HEAP32[$sin>>2]=0|0;HEAP32[$sin+4>>2]=0|0;HEAP32[$sin+8>>2]=0|0;HEAP32[$sin+12>>2]=0|0;
  HEAP16[$sin>>1] = 2;
  $7 = $2;
  $8 = ((($7)) + 4|0);
  $9 = HEAP16[$8>>1]|0;
  $10 = (_htons($9)|0);
  $11 = ((($sin)) + 2|0);
  HEAP16[$11>>1] = $10;
  $12 = $2;
  $13 = HEAP32[$12>>2]|0;
  $14 = ((($sin)) + 4|0);
  HEAP32[$14>>2] = $13;
  HEAP32[$msgHdr>>2] = $sin;
  $15 = ((($msgHdr)) + 4|0);
  HEAP32[$15>>2] = 16;
 }
 $16 = $3;
 $17 = ((($msgHdr)) + 8|0);
 HEAP32[$17>>2] = $16;
 $18 = $4;
 $19 = ((($msgHdr)) + 12|0);
 HEAP32[$19>>2] = $18;
 $20 = $1;
 $21 = (_sendmsg($20,$msgHdr,16384)|0);
 $sentLength = $21;
 $22 = $sentLength;
 $23 = ($22|0)==(-1);
 if (!($23)) {
  $27 = $sentLength;
  $0 = $27;
  $28 = $0;
  STACKTOP = sp;return ($28|0);
 }
 $24 = (___errno_location()|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = ($25|0)==(11);
 if ($26) {
  $0 = 0;
  $28 = $0;
  STACKTOP = sp;return ($28|0);
 } else {
  $0 = -1;
  $28 = $0;
  STACKTOP = sp;return ($28|0);
 }
 return (0)|0;
}
function _enet_socket_receive($socket,$address,$buffers,$bufferCount) {
 $socket = $socket|0;
 $address = $address|0;
 $buffers = $buffers|0;
 $bufferCount = $bufferCount|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $msgHdr = 0, $recvLength = 0, $sin = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $msgHdr = sp + 24|0;
 $sin = sp + 8|0;
 $1 = $socket;
 $2 = $address;
 $3 = $buffers;
 $4 = $bufferCount;
 ;HEAP32[$msgHdr>>2]=0|0;HEAP32[$msgHdr+4>>2]=0|0;HEAP32[$msgHdr+8>>2]=0|0;HEAP32[$msgHdr+12>>2]=0|0;HEAP32[$msgHdr+16>>2]=0|0;HEAP32[$msgHdr+20>>2]=0|0;HEAP32[$msgHdr+24>>2]=0|0;
 $5 = $2;
 $6 = ($5|0)!=(0|0);
 if ($6) {
  HEAP32[$msgHdr>>2] = $sin;
  $7 = ((($msgHdr)) + 4|0);
  HEAP32[$7>>2] = 16;
 }
 $8 = $3;
 $9 = ((($msgHdr)) + 8|0);
 HEAP32[$9>>2] = $8;
 $10 = $4;
 $11 = ((($msgHdr)) + 12|0);
 HEAP32[$11>>2] = $10;
 $12 = $1;
 $13 = (_recvmsg($12,$msgHdr,16384)|0);
 $recvLength = $13;
 $14 = $recvLength;
 $15 = ($14|0)==(-1);
 if (!($15)) {
  $19 = $2;
  $20 = ($19|0)!=(0|0);
  if ($20) {
   $21 = ((($sin)) + 4|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = $2;
   HEAP32[$23>>2] = $22;
   $24 = ((($sin)) + 2|0);
   $25 = HEAP16[$24>>1]|0;
   $26 = (_ntohs($25)|0);
   $27 = $2;
   $28 = ((($27)) + 4|0);
   HEAP16[$28>>1] = $26;
  }
  $29 = $recvLength;
  $0 = $29;
  $30 = $0;
  STACKTOP = sp;return ($30|0);
 }
 $16 = (___errno_location()|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ($17|0)==(11);
 if ($18) {
  $0 = 0;
  $30 = $0;
  STACKTOP = sp;return ($30|0);
 } else {
  $0 = -1;
  $30 = $0;
  STACKTOP = sp;return ($30|0);
 }
 return (0)|0;
}
function _enet_socket_wait($socket,$condition,$timeout) {
 $socket = $socket|0;
 $condition = $condition|0;
 $timeout = $timeout|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $__b = 0, $__b2 = 0, $__i = 0, $__i1 = 0, $readSet = 0, $selectCount = 0, $timeVal = 0, $writeSet = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $readSet = sp + 160|0;
 $writeSet = sp + 32|0;
 $timeVal = sp + 24|0;
 $1 = $socket;
 $2 = $condition;
 $3 = $timeout;
 $4 = $3;
 $5 = (($4>>>0) / 1000)&-1;
 HEAP32[$timeVal>>2] = $5;
 $6 = $3;
 $7 = (($6>>>0) % 1000)&-1;
 $8 = ($7*1000)|0;
 $9 = ((($timeVal)) + 4|0);
 HEAP32[$9>>2] = $8;
 $__b = $readSet;
 $__i = 32;
 while(1) {
  $10 = $__i;
  $11 = ($10|0)!=(0);
  if (!($11)) {
   break;
  }
  $12 = $__b;
  $13 = ((($12)) + 4|0);
  $__b = $13;
  HEAP32[$12>>2] = 0;
  $14 = $__i;
  $15 = (($14) + -1)|0;
  $__i = $15;
 }
 $__b2 = $writeSet;
 $__i1 = 32;
 while(1) {
  $16 = $__i1;
  $17 = ($16|0)!=(0);
  if (!($17)) {
   break;
  }
  $18 = $__b2;
  $19 = ((($18)) + 4|0);
  $__b2 = $19;
  HEAP32[$18>>2] = 0;
  $20 = $__i1;
  $21 = (($20) + -1)|0;
  $__i1 = $21;
 }
 $22 = $2;
 $23 = HEAP32[$22>>2]|0;
 $24 = $23 & 1;
 $25 = ($24|0)!=(0);
 if ($25) {
  $26 = $1;
  $27 = (($26>>>0) % 32)&-1;
  $28 = 1 << $27;
  $29 = $1;
  $30 = (($29>>>0) / 32)&-1;
  $31 = (($writeSet) + ($30<<2)|0);
  $32 = HEAP32[$31>>2]|0;
  $33 = $32 | $28;
  HEAP32[$31>>2] = $33;
 }
 $34 = $2;
 $35 = HEAP32[$34>>2]|0;
 $36 = $35 & 2;
 $37 = ($36|0)!=(0);
 if ($37) {
  $38 = $1;
  $39 = (($38>>>0) % 32)&-1;
  $40 = 1 << $39;
  $41 = $1;
  $42 = (($41>>>0) / 32)&-1;
  $43 = (($readSet) + ($42<<2)|0);
  $44 = HEAP32[$43>>2]|0;
  $45 = $44 | $40;
  HEAP32[$43>>2] = $45;
 }
 $46 = $1;
 $47 = (($46) + 1)|0;
 $48 = (_select($47,$readSet,$writeSet,0,$timeVal)|0);
 $selectCount = $48;
 $49 = $selectCount;
 $50 = ($49|0)<(0);
 if ($50) {
  $51 = (___errno_location()|0);
  $52 = HEAP32[$51>>2]|0;
  $53 = ($52|0)==(4);
  if ($53) {
   $54 = $2;
   $55 = HEAP32[$54>>2]|0;
   $56 = $55 & 4;
   $57 = ($56|0)!=(0);
   if ($57) {
    $58 = $2;
    HEAP32[$58>>2] = 4;
    $0 = 0;
    $86 = $0;
    STACKTOP = sp;return ($86|0);
   }
  }
  $0 = -1;
  $86 = $0;
  STACKTOP = sp;return ($86|0);
 }
 $59 = $2;
 HEAP32[$59>>2] = 0;
 $60 = $selectCount;
 $61 = ($60|0)==(0);
 if ($61) {
  $0 = 0;
  $86 = $0;
  STACKTOP = sp;return ($86|0);
 }
 $62 = $1;
 $63 = (($62>>>0) / 32)&-1;
 $64 = (($writeSet) + ($63<<2)|0);
 $65 = HEAP32[$64>>2]|0;
 $66 = $1;
 $67 = (($66>>>0) % 32)&-1;
 $68 = 1 << $67;
 $69 = $65 & $68;
 $70 = ($69|0)!=(0);
 if ($70) {
  $71 = $2;
  $72 = HEAP32[$71>>2]|0;
  $73 = $72 | 1;
  HEAP32[$71>>2] = $73;
 }
 $74 = $1;
 $75 = (($74>>>0) / 32)&-1;
 $76 = (($readSet) + ($75<<2)|0);
 $77 = HEAP32[$76>>2]|0;
 $78 = $1;
 $79 = (($78>>>0) % 32)&-1;
 $80 = 1 << $79;
 $81 = $77 & $80;
 $82 = ($81|0)!=(0);
 if ($82) {
  $83 = $2;
  $84 = HEAP32[$83>>2]|0;
  $85 = $84 | 2;
  HEAP32[$83>>2] = $85;
 }
 $0 = 0;
 $86 = $0;
 STACKTOP = sp;return ($86|0);
}
function __ZN4love9luasocket6__openEP9lua_State($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _lua_getfield($L,-10002,523111); //@line 47 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-1,523223); //@line 47 "libraries/luasocket/luasocket.cpp"
 _lua_pushcclosure($L,3481,0); //@line 47 "libraries/luasocket/luasocket.cpp"
 _lua_setfield($L,-2,267565); //@line 47 "libraries/luasocket/luasocket.cpp"
 _lua_settop($L,-3); //@line 47 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-10002,523111); //@line 48 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-1,523223); //@line 48 "libraries/luasocket/luasocket.cpp"
 _lua_pushcclosure($L,3482,0); //@line 48 "libraries/luasocket/luasocket.cpp"
 _lua_setfield($L,-2,267577); //@line 48 "libraries/luasocket/luasocket.cpp"
 _lua_settop($L,-3); //@line 48 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-10002,523111); //@line 50 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-1,523223); //@line 50 "libraries/luasocket/luasocket.cpp"
 _lua_pushcclosure($L,3483,0); //@line 50 "libraries/luasocket/luasocket.cpp"
 _lua_setfield($L,-2,363666); //@line 50 "libraries/luasocket/luasocket.cpp"
 _lua_settop($L,-3); //@line 50 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-10002,523111); //@line 51 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-1,523223); //@line 51 "libraries/luasocket/luasocket.cpp"
 _lua_pushcclosure($L,3484,0); //@line 51 "libraries/luasocket/luasocket.cpp"
 _lua_setfield($L,-2,280787); //@line 51 "libraries/luasocket/luasocket.cpp"
 _lua_settop($L,-3); //@line 51 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-10002,523111); //@line 52 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-1,523223); //@line 52 "libraries/luasocket/luasocket.cpp"
 _lua_pushcclosure($L,3485,0); //@line 52 "libraries/luasocket/luasocket.cpp"
 _lua_setfield($L,-2,293000); //@line 52 "libraries/luasocket/luasocket.cpp"
 _lua_settop($L,-3); //@line 52 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-10002,523111); //@line 53 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-1,523223); //@line 53 "libraries/luasocket/luasocket.cpp"
 _lua_pushcclosure($L,3486,0); //@line 53 "libraries/luasocket/luasocket.cpp"
 _lua_setfield($L,-2,301199); //@line 53 "libraries/luasocket/luasocket.cpp"
 _lua_settop($L,-3); //@line 53 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-10002,523111); //@line 54 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-1,523223); //@line 54 "libraries/luasocket/luasocket.cpp"
 _lua_pushcclosure($L,3487,0); //@line 54 "libraries/luasocket/luasocket.cpp"
 _lua_setfield($L,-2,363760); //@line 54 "libraries/luasocket/luasocket.cpp"
 _lua_settop($L,-3); //@line 54 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-10002,523111); //@line 55 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-1,523223); //@line 55 "libraries/luasocket/luasocket.cpp"
 _lua_pushcclosure($L,3488,0); //@line 55 "libraries/luasocket/luasocket.cpp"
 _lua_setfield($L,-2,311617); //@line 55 "libraries/luasocket/luasocket.cpp"
 _lua_settop($L,-3); //@line 55 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-10002,523111); //@line 56 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-1,523223); //@line 56 "libraries/luasocket/luasocket.cpp"
 _lua_pushcclosure($L,3489,0); //@line 56 "libraries/luasocket/luasocket.cpp"
 _lua_setfield($L,-2,315244); //@line 56 "libraries/luasocket/luasocket.cpp"
 _lua_settop($L,-3); //@line 56 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-10002,523111); //@line 57 "libraries/luasocket/luasocket.cpp"
 _lua_getfield($L,-1,523223); //@line 57 "libraries/luasocket/luasocket.cpp"
 _lua_pushcclosure($L,3490,0); //@line 57 "libraries/luasocket/luasocket.cpp"
 _lua_setfield($L,-2,325791); //@line 57 "libraries/luasocket/luasocket.cpp"
 _lua_settop($L,-3); //@line 57 "libraries/luasocket/luasocket.cpp"
 return 0; //@line 61 "libraries/luasocket/luasocket.cpp"
}
function __ZN4love9luasocket23__open_luasocket_socketEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_loadbuffer($L,267587,4061,271648)|0); //@line 214 "libraries/luasocket/libluasocket/socket.lua.h"
 $1 = ($0|0)==(0); //@line 214 "libraries/luasocket/libluasocket/socket.lua.h"
 if ($1) {
  _lua_call($L,0,0); //@line 214 "libraries/luasocket/libluasocket/socket.lua.h"
 }
 _lua_getfield($L,-10002,363666); //@line 67 "libraries/luasocket/luasocket.cpp"
 return 1; //@line 68 "libraries/luasocket/luasocket.cpp"
}
function __ZN4love9luasocket20__open_luasocket_ftpEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_loadbuffer($L,271659,9120,280779)|0); //@line 467 "libraries/luasocket/libluasocket/ftp.lua.h"
 $1 = ($0|0)==(0); //@line 467 "libraries/luasocket/libluasocket/ftp.lua.h"
 if ($1) {
  _lua_call($L,0,0); //@line 467 "libraries/luasocket/libluasocket/ftp.lua.h"
 }
 _lua_getfield($L,-10002,280787); //@line 74 "libraries/luasocket/luasocket.cpp"
 return 1; //@line 75 "libraries/luasocket/luasocket.cpp"
}
function __ZN4love9luasocket21__open_luasocket_httpEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_loadbuffer($L,280798,12193,292991)|0); //@line 620 "libraries/luasocket/libluasocket/http.lua.h"
 $1 = ($0|0)==(0); //@line 620 "libraries/luasocket/libluasocket/http.lua.h"
 if ($1) {
  _lua_call($L,0,0); //@line 620 "libraries/luasocket/libluasocket/http.lua.h"
 }
 _lua_getfield($L,-10002,293000); //@line 81 "libraries/luasocket/luasocket.cpp"
 return 1; //@line 82 "libraries/luasocket/luasocket.cpp"
}
function __ZN4love9luasocket22__open_luasocket_ltn12EP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_loadbuffer($L,293012,8177,301189)|0); //@line 419 "libraries/luasocket/libluasocket/ltn12.lua.h"
 $1 = ($0|0)==(0); //@line 419 "libraries/luasocket/libluasocket/ltn12.lua.h"
 if ($1) {
  _lua_call($L,0,0); //@line 419 "libraries/luasocket/libluasocket/ltn12.lua.h"
 }
 _lua_getfield($L,-10002,301199); //@line 88 "libraries/luasocket/luasocket.cpp"
 return 1; //@line 89 "libraries/luasocket/luasocket.cpp"
}
function __ZN4love9luasocket21__open_luasocket_mimeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_loadbuffer($L,301205,2433,303638)|0); //@line 132 "libraries/luasocket/libluasocket/mime.lua.h"
 $1 = ($0|0)==(0); //@line 132 "libraries/luasocket/libluasocket/mime.lua.h"
 if ($1) {
  _lua_call($L,0,0); //@line 132 "libraries/luasocket/libluasocket/mime.lua.h"
 }
 _lua_getfield($L,-10002,363760); //@line 95 "libraries/luasocket/luasocket.cpp"
 return 1; //@line 96 "libraries/luasocket/luasocket.cpp"
}
function __ZN4love9luasocket21__open_luasocket_smtpEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_loadbuffer($L,303647,7961,311608)|0); //@line 409 "libraries/luasocket/libluasocket/smtp.lua.h"
 $1 = ($0|0)==(0); //@line 409 "libraries/luasocket/libluasocket/smtp.lua.h"
 if ($1) {
  _lua_call($L,0,0); //@line 409 "libraries/luasocket/libluasocket/smtp.lua.h"
 }
 _lua_getfield($L,-10002,311617); //@line 102 "libraries/luasocket/luasocket.cpp"
 return 1; //@line 103 "libraries/luasocket/luasocket.cpp"
}
function __ZN4love9luasocket19__open_luasocket_tpEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_loadbuffer($L,311629,3608,315237)|0); //@line 191 "libraries/luasocket/libluasocket/tp.lua.h"
 $1 = ($0|0)==(0); //@line 191 "libraries/luasocket/libluasocket/tp.lua.h"
 if ($1) {
  _lua_call($L,0,0); //@line 191 "libraries/luasocket/libluasocket/tp.lua.h"
 }
 _lua_getfield($L,-10002,315244); //@line 109 "libraries/luasocket/luasocket.cpp"
 return 1; //@line 110 "libraries/luasocket/luasocket.cpp"
}
function __ZN4love9luasocket20__open_luasocket_urlEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_loadbuffer($L,315254,10529,325783)|0); //@line 537 "libraries/luasocket/libluasocket/url.lua.h"
 $1 = ($0|0)==(0); //@line 537 "libraries/luasocket/libluasocket/url.lua.h"
 if ($1) {
  _lua_call($L,0,0); //@line 537 "libraries/luasocket/libluasocket/url.lua.h"
 }
 _lua_getfield($L,-10002,325791); //@line 116 "libraries/luasocket/luasocket.cpp"
 return 1; //@line 117 "libraries/luasocket/luasocket.cpp"
}
function _luaopen_luautf8($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 _lua_createtable($1,0,6);
 $l = 12180;
 while(1) {
  $2 = $l;
  $3 = HEAP32[$2>>2]|0;
  $4 = ($3|0)!=(0|0);
  if (!($4)) {
   break;
  }
  $5 = $l;
  $6 = ((($5)) + 4|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)!=(0|0);
  if ($8) {
   $9 = $0;
   $10 = $l;
   $11 = ((($10)) + 4|0);
   $12 = HEAP32[$11>>2]|0;
   _lua_pushcclosure($9,$12,0);
   $13 = $0;
   $14 = $l;
   $15 = HEAP32[$14>>2]|0;
   _lua_setfield($13,-2,$15);
  }
  $16 = $l;
  $17 = ((($16)) + 8|0);
  $l = $17;
 }
 $18 = $0;
 _lua_pushlstring($18,326010,16);
 $19 = $0;
 _lua_setfield($19,-2,325998);
 STACKTOP = sp;return 1;
}
function _byteoffset($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $9 = 0, $len = 0, $n = 0, $posi = 0, $s = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $len = sp + 16|0;
 $0 = $L;
 $1 = $0;
 $2 = (_luaL_checklstring($1,1,$len)|0);
 $s = $2;
 $3 = $0;
 $4 = (_luaL_checkinteger($3,2)|0);
 $n = $4;
 $5 = $n;
 $6 = ($5|0)>=(0);
 $7 = HEAP32[$len>>2]|0;
 $8 = (($7) + 1)|0;
 $9 = $6 ? 1 : $8;
 $posi = $9;
 $10 = $0;
 $11 = $posi;
 $12 = (_luaL_optinteger($10,3,$11)|0);
 $13 = HEAP32[$len>>2]|0;
 $14 = (_u_posrelat($12,$13)|0);
 $posi = $14;
 $15 = $posi;
 $16 = (1)<=($15|0);
 if ($16) {
  $17 = $posi;
  $18 = (($17) + -1)|0;
  $posi = $18;
  $19 = HEAP32[$len>>2]|0;
  $20 = ($18|0)<=($19|0);
  if (!($20)) {
   label = 3;
  }
 } else {
  label = 3;
 }
 if ((label|0) == 3) {
  $21 = $0;
  (_luaL_argerror($21,3,325809)|0);
 }
 $22 = $n;
 $23 = ($22|0)==(0);
 L6: do {
  if ($23) {
   while(1) {
    $24 = $posi;
    $25 = ($24|0)>(0);
    if (!($25)) {
     break L6;
    }
    $26 = $s;
    $27 = $posi;
    $28 = (($26) + ($27)|0);
    $29 = HEAP8[$28>>0]|0;
    $30 = $29 << 24 >> 24;
    $31 = $30 & 192;
    $32 = ($31|0)==(128);
    if (!($32)) {
     break L6;
    }
    $33 = $posi;
    $34 = (($33) + -1)|0;
    $posi = $34;
   }
  } else {
   $35 = $s;
   $36 = $posi;
   $37 = (($35) + ($36)|0);
   $38 = HEAP8[$37>>0]|0;
   $39 = $38 << 24 >> 24;
   $40 = $39 & 192;
   $41 = ($40|0)==(128);
   if ($41) {
    $42 = $0;
    (_luaL_error($42,325831,$vararg_buffer)|0);
   }
   $43 = $n;
   $44 = ($43|0)<(0);
   if ($44) {
    while(1) {
     $45 = $n;
     $46 = ($45|0)<(0);
     $47 = $posi;
     $48 = ($47|0)>(0);
     $49 = $46 ? $48 : 0;
     if (!($49)) {
      break L6;
     }
     while(1) {
      $50 = $posi;
      $51 = (($50) + -1)|0;
      $posi = $51;
      $52 = $posi;
      $53 = ($52|0)>(0);
      if (!($53)) {
       break;
      }
      $54 = $s;
      $55 = $posi;
      $56 = (($54) + ($55)|0);
      $57 = HEAP8[$56>>0]|0;
      $58 = $57 << 24 >> 24;
      $59 = $58 & 192;
      $60 = ($59|0)==(128);
      if (!($60)) {
       break;
      }
     }
     $61 = $n;
     $62 = (($61) + 1)|0;
     $n = $62;
    }
   }
   $63 = $n;
   $64 = (($63) + -1)|0;
   $n = $64;
   while(1) {
    $65 = $n;
    $66 = ($65|0)>(0);
    if (!($66)) {
     break L6;
    }
    $67 = $posi;
    $68 = HEAP32[$len>>2]|0;
    $69 = ($67|0)<($68|0);
    if (!($69)) {
     break L6;
    }
    while(1) {
     $70 = $posi;
     $71 = (($70) + 1)|0;
     $posi = $71;
     $72 = $s;
     $73 = $posi;
     $74 = (($72) + ($73)|0);
     $75 = HEAP8[$74>>0]|0;
     $76 = $75 << 24 >> 24;
     $77 = $76 & 192;
     $78 = ($77|0)==(128);
     if (!($78)) {
      break;
     }
    }
    $79 = $n;
    $80 = (($79) + -1)|0;
    $n = $80;
   }
  }
 } while(0);
 $81 = $n;
 $82 = ($81|0)==(0);
 $83 = $0;
 if ($82) {
  $84 = $posi;
  $85 = (($84) + 1)|0;
  _lua_pushinteger($83,$85);
  STACKTOP = sp;return 1;
 } else {
  _lua_pushnil($83);
  STACKTOP = sp;return 1;
 }
 return (0)|0;
}
function _u_posrelat($pos,$len) {
 $pos = $pos|0;
 $len = $len|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $pos;
 $2 = $len;
 $3 = $1;
 $4 = ($3|0)>=(0);
 $5 = $1;
 do {
  if ($4) {
   $0 = $5;
  } else {
   $6 = (0 - ($5))|0;
   $7 = $2;
   $8 = ($6>>>0)>($7>>>0);
   if ($8) {
    $0 = 0;
    break;
   } else {
    $9 = $2;
    $10 = $1;
    $11 = (($9) + ($10))|0;
    $12 = (($11) + 1)|0;
    $0 = $12;
    break;
   }
  }
 } while(0);
 $13 = $0;
 STACKTOP = sp;return ($13|0);
}
function _codepoint($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $code = 0, $len = 0, $n = 0, $pose = 0;
 var $posi = 0, $s = 0, $se = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $len = sp + 36|0;
 $code = sp + 12|0;
 $1 = $L;
 $2 = $1;
 $3 = (_luaL_checklstring($2,1,$len)|0);
 $s = $3;
 $4 = $1;
 $5 = (_luaL_optinteger($4,2,1)|0);
 $6 = HEAP32[$len>>2]|0;
 $7 = (_u_posrelat($5,$6)|0);
 $posi = $7;
 $8 = $1;
 $9 = $posi;
 $10 = (_luaL_optinteger($8,3,$9)|0);
 $11 = HEAP32[$len>>2]|0;
 $12 = (_u_posrelat($10,$11)|0);
 $pose = $12;
 $13 = $posi;
 $14 = ($13|0)>=(1);
 if (!($14)) {
  $15 = $1;
  (_luaL_argerror($15,2,325881)|0);
 }
 $16 = $pose;
 $17 = HEAP32[$len>>2]|0;
 $18 = ($16|0)<=($17|0);
 if (!($18)) {
  $19 = $1;
  (_luaL_argerror($19,3,325881)|0);
 }
 $20 = $posi;
 $21 = $pose;
 $22 = ($20|0)>($21|0);
 if ($22) {
  $0 = 0;
  $55 = $0;
  STACKTOP = sp;return ($55|0);
 }
 $23 = $pose;
 $24 = $posi;
 $25 = (($23) - ($24))|0;
 $26 = (($25) + 1)|0;
 $n = $26;
 $27 = $posi;
 $28 = $n;
 $29 = (($27) + ($28))|0;
 $30 = $pose;
 $31 = ($29|0)<=($30|0);
 $32 = $1;
 if ($31) {
  $33 = (_luaL_error($32,522501,$vararg_buffer)|0);
  $0 = $33;
  $55 = $0;
  STACKTOP = sp;return ($55|0);
 }
 $34 = $n;
 _luaL_checkstack($32,$34,522501);
 $n = 0;
 $35 = $s;
 $36 = $pose;
 $37 = (($35) + ($36)|0);
 $se = $37;
 $38 = $posi;
 $39 = (($38) - 1)|0;
 $40 = $s;
 $41 = (($40) + ($39)|0);
 $s = $41;
 while(1) {
  $42 = $s;
  $43 = $se;
  $44 = ($42>>>0)<($43>>>0);
  if (!($44)) {
   label = 14;
   break;
  }
  $45 = $s;
  $46 = (_utf8_decode($45,$code)|0);
  $s = $46;
  $47 = $s;
  $48 = ($47|0)==(0|0);
  $49 = $1;
  if ($48) {
   label = 12;
   break;
  }
  $51 = HEAP32[$code>>2]|0;
  _lua_pushinteger($49,$51);
  $52 = $n;
  $53 = (($52) + 1)|0;
  $n = $53;
 }
 if ((label|0) == 12) {
  $50 = (_luaL_error($49,325894,$vararg_buffer1)|0);
  $0 = $50;
  $55 = $0;
  STACKTOP = sp;return ($55|0);
 }
 else if ((label|0) == 14) {
  $54 = $n;
  $0 = $54;
  $55 = $0;
  STACKTOP = sp;return ($55|0);
 }
 return (0)|0;
}
function _utf8_decode($o,$val) {
 $o = $o|0;
 $val = $val|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $c = 0, $cc = 0, $count = 0, $or$cond = 0, $res = 0;
 var $s = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $o;
 $2 = $val;
 $3 = $1;
 $s = $3;
 $4 = $s;
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $c = $6;
 $res = 0;
 $7 = $c;
 $8 = ($7>>>0)<(128);
 do {
  if ($8) {
   $9 = $c;
   $res = $9;
  } else {
   $count = 0;
   while(1) {
    $10 = $c;
    $11 = $10 & 64;
    $12 = ($11|0)!=(0);
    if (!($12)) {
     break;
    }
    $13 = $count;
    $14 = (($13) + 1)|0;
    $count = $14;
    $15 = $s;
    $16 = (($15) + ($14)|0);
    $17 = HEAP8[$16>>0]|0;
    $18 = $17&255;
    $cc = $18;
    $19 = $cc;
    $20 = $19 & 192;
    $21 = ($20|0)!=(128);
    if ($21) {
     label = 6;
     break;
    }
    $22 = $res;
    $23 = $22 << 6;
    $24 = $cc;
    $25 = $24 & 63;
    $26 = $23 | $25;
    $res = $26;
    $27 = $c;
    $28 = $27 << 1;
    $c = $28;
   }
   if ((label|0) == 6) {
    $0 = 0;
    $54 = $0;
    STACKTOP = sp;return ($54|0);
   }
   $29 = $c;
   $30 = $29 & 127;
   $31 = $count;
   $32 = ($31*5)|0;
   $33 = $30 << $32;
   $34 = $res;
   $35 = $34 | $33;
   $res = $35;
   $36 = $count;
   $37 = ($36|0)>(3);
   $38 = $res;
   $39 = ($38>>>0)>(1114111);
   $or$cond = $37 | $39;
   if (!($or$cond)) {
    $40 = $res;
    $41 = $count;
    $42 = (12236 + ($41<<2)|0);
    $43 = HEAP32[$42>>2]|0;
    $44 = ($40>>>0)<=($43>>>0);
    if (!($44)) {
     $45 = $count;
     $46 = $s;
     $47 = (($46) + ($45)|0);
     $s = $47;
     break;
    }
   }
   $0 = 0;
   $54 = $0;
   STACKTOP = sp;return ($54|0);
  }
 } while(0);
 $48 = $2;
 $49 = ($48|0)!=(0|0);
 if ($49) {
  $50 = $res;
  $51 = $2;
  HEAP32[$51>>2] = $50;
 }
 $52 = $s;
 $53 = ((($52)) + 1|0);
 $0 = $53;
 $54 = $0;
 STACKTOP = sp;return ($54|0);
}
function _utfchar($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b = 0, $i = 0, $n = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1056|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $b = sp;
 $0 = $L;
 $1 = $0;
 $2 = (_lua_gettop($1)|0);
 $n = $2;
 $3 = $n;
 $4 = ($3|0)==(1);
 $5 = $0;
 if ($4) {
  _pushutfchar($5,1);
  STACKTOP = sp;return 1;
 }
 _luaL_buffinit($5,$b);
 $i = 1;
 while(1) {
  $6 = $i;
  $7 = $n;
  $8 = ($6|0)<=($7|0);
  if (!($8)) {
   break;
  }
  $9 = $0;
  $10 = $i;
  _pushutfchar($9,$10);
  _luaL_addvalue($b);
  $11 = $i;
  $12 = (($11) + 1)|0;
  $i = $12;
 }
 _luaL_pushresult($b);
 STACKTOP = sp;return 1;
}
function _pushutfchar($L,$arg) {
 $L = $L|0;
 $arg = $arg|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buff = 0;
 var $code = 0, $l = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $buff = sp + 16|0;
 $0 = $L;
 $1 = $arg;
 $2 = $0;
 $3 = $1;
 $4 = (_luaL_checkinteger($2,$3)|0);
 $code = $4;
 $5 = $code;
 $6 = (0)<=($5|0);
 $7 = $code;
 $8 = ($7|0)<=(1114111);
 $or$cond = $6 & $8;
 if (!($or$cond)) {
  $9 = $0;
  $10 = $1;
  (_luaL_argerror($9,$10,325913)|0);
 }
 $11 = $code;
 $12 = (_utf8esc($buff,$11)|0);
 $l = $12;
 $13 = $0;
 $14 = ((($buff)) + 8|0);
 $15 = $l;
 $16 = (0 - ($15))|0;
 $17 = (($14) + ($16)|0);
 $18 = $l;
 _lua_pushlstring($13,$17,$18);
 STACKTOP = sp;return;
}
function _utf8esc($buff,$x) {
 $buff = $buff|0;
 $x = $x|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $mfb = 0, $n = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $buff;
 $1 = $x;
 $n = 1;
 $2 = $1;
 $3 = ($2>>>0)<(128);
 if ($3) {
  $4 = $1;
  $5 = $4&255;
  $6 = $0;
  $7 = ((($6)) + 7|0);
  HEAP8[$7>>0] = $5;
  $34 = $n;
  STACKTOP = sp;return ($34|0);
 }
 $mfb = 63;
 while(1) {
  $8 = $1;
  $9 = $8 & 63;
  $10 = 128 | $9;
  $11 = $10&255;
  $12 = $n;
  $13 = (($12) + 1)|0;
  $n = $13;
  $14 = (8 - ($12))|0;
  $15 = $0;
  $16 = (($15) + ($14)|0);
  HEAP8[$16>>0] = $11;
  $17 = $1;
  $18 = $17 >>> 6;
  $1 = $18;
  $19 = $mfb;
  $20 = $19 >>> 1;
  $mfb = $20;
  $21 = $1;
  $22 = $mfb;
  $23 = ($21>>>0)>($22>>>0);
  if (!($23)) {
   break;
  }
 }
 $24 = $mfb;
 $25 = $24 ^ -1;
 $26 = $25 << 1;
 $27 = $1;
 $28 = $26 | $27;
 $29 = $28&255;
 $30 = $n;
 $31 = (8 - ($30))|0;
 $32 = $0;
 $33 = (($32) + ($31)|0);
 HEAP8[$33>>0] = $29;
 $34 = $n;
 STACKTOP = sp;return ($34|0);
}
function _utflen($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $len = 0, $n = 0, $posi = 0, $posj = 0, $s = 0, $s1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $len = sp + 16|0;
 $1 = $L;
 $n = 0;
 $2 = $1;
 $3 = (_luaL_checklstring($2,1,$len)|0);
 $s = $3;
 $4 = $1;
 $5 = (_luaL_optinteger($4,2,1)|0);
 $6 = HEAP32[$len>>2]|0;
 $7 = (_u_posrelat($5,$6)|0);
 $posi = $7;
 $8 = $1;
 $9 = (_luaL_optinteger($8,3,-1)|0);
 $10 = HEAP32[$len>>2]|0;
 $11 = (_u_posrelat($9,$10)|0);
 $posj = $11;
 $12 = $posi;
 $13 = (1)<=($12|0);
 if ($13) {
  $14 = $posi;
  $15 = (($14) + -1)|0;
  $posi = $15;
  $16 = HEAP32[$len>>2]|0;
  $17 = ($15|0)<=($16|0);
  if (!($17)) {
   label = 3;
  }
 } else {
  label = 3;
 }
 if ((label|0) == 3) {
  $18 = $1;
  (_luaL_argerror($18,2,325932)|0);
 }
 $19 = $posj;
 $20 = (($19) + -1)|0;
 $posj = $20;
 $21 = HEAP32[$len>>2]|0;
 $22 = ($20|0)<($21|0);
 if (!($22)) {
  $23 = $1;
  (_luaL_argerror($23,3,325963)|0);
 }
 while(1) {
  $24 = $posi;
  $25 = $posj;
  $26 = ($24|0)<=($25|0);
  if (!($26)) {
   label = 11;
   break;
  }
  $27 = $s;
  $28 = $posi;
  $29 = (($27) + ($28)|0);
  $30 = (_utf8_decode($29,0)|0);
  $s1 = $30;
  $31 = $s1;
  $32 = ($31|0)==(0|0);
  if ($32) {
   label = 9;
   break;
  }
  $37 = $s1;
  $38 = $s;
  $39 = $37;
  $40 = $38;
  $41 = (($39) - ($40))|0;
  $posi = $41;
  $42 = $n;
  $43 = (($42) + 1)|0;
  $n = $43;
 }
 if ((label|0) == 9) {
  $33 = $1;
  _lua_pushnil($33);
  $34 = $1;
  $35 = $posi;
  $36 = (($35) + 1)|0;
  _lua_pushinteger($34,$36);
  $0 = 2;
  $46 = $0;
  STACKTOP = sp;return ($46|0);
 }
 else if ((label|0) == 11) {
  $44 = $1;
  $45 = $n;
  _lua_pushinteger($44,$45);
  $0 = 1;
  $46 = $0;
  STACKTOP = sp;return ($46|0);
 }
 return (0)|0;
}
function _iter_codes($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 (_luaL_checklstring($1,1,0)|0);
 $2 = $0;
 _lua_pushcclosure($2,3491,0);
 $3 = $0;
 _lua_pushvalue($3,1);
 $4 = $0;
 _lua_pushinteger($4,0);
 STACKTOP = sp;return 3;
}
function _iter_aux($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $code = 0, $len = 0, $n = 0, $next = 0, $s = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $len = sp + 20|0;
 $code = sp + 8|0;
 $1 = $L;
 $2 = $1;
 $3 = (_luaL_checklstring($2,1,$len)|0);
 $s = $3;
 $4 = $1;
 $5 = (_lua_tointeger($4,2)|0);
 $6 = (($5) - 1)|0;
 $n = $6;
 $7 = $n;
 $8 = ($7|0)<(0);
 L1: do {
  if ($8) {
   $n = 0;
  } else {
   $9 = $n;
   $10 = HEAP32[$len>>2]|0;
   $11 = ($9|0)<($10|0);
   if ($11) {
    $12 = $n;
    $13 = (($12) + 1)|0;
    $n = $13;
    while(1) {
     $14 = $s;
     $15 = $n;
     $16 = (($14) + ($15)|0);
     $17 = HEAP8[$16>>0]|0;
     $18 = $17 << 24 >> 24;
     $19 = $18 & 192;
     $20 = ($19|0)==(128);
     if (!($20)) {
      break L1;
     }
     $21 = $n;
     $22 = (($21) + 1)|0;
     $n = $22;
    }
   }
  }
 } while(0);
 $23 = $n;
 $24 = HEAP32[$len>>2]|0;
 $25 = ($23|0)>=($24|0);
 if ($25) {
  $0 = 0;
  $44 = $0;
  STACKTOP = sp;return ($44|0);
 }
 $26 = $s;
 $27 = $n;
 $28 = (($26) + ($27)|0);
 $29 = (_utf8_decode($28,$code)|0);
 $next = $29;
 $30 = $next;
 $31 = ($30|0)==(0|0);
 if (!($31)) {
  $32 = $next;
  $33 = HEAP8[$32>>0]|0;
  $34 = $33 << 24 >> 24;
  $35 = $34 & 192;
  $36 = ($35|0)==(128);
  if (!($36)) {
   $39 = $1;
   $40 = $n;
   $41 = (($40) + 1)|0;
   _lua_pushinteger($39,$41);
   $42 = $1;
   $43 = HEAP32[$code>>2]|0;
   _lua_pushinteger($42,$43);
   $0 = 2;
   $44 = $0;
   STACKTOP = sp;return ($44|0);
  }
 }
 $37 = $1;
 $38 = (_luaL_error($37,325894,$vararg_buffer)|0);
 $0 = $38;
 $44 = $0;
 STACKTOP = sp;return ($44|0);
}
function __ZN9Noise12345noiseEfff($x,$y,$z) {
 $x = +$x;
 $y = +$y;
 $z = +$z;
 var $0 = 0, $1 = 0, $10 = 0.0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0, $108 = 0, $109 = 0, $11 = 0.0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0.0, $118 = 0, $119 = 0, $12 = 0.0, $120 = 0, $121 = 0.0, $122 = 0, $123 = 0, $124 = 0.0, $125 = 0.0, $126 = 0, $127 = 0, $128 = 0.0, $129 = 0.0, $13 = 0.0, $130 = 0.0, $131 = 0.0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0.0, $140 = 0, $141 = 0, $142 = 0.0, $143 = 0, $144 = 0, $145 = 0, $146 = 0.0, $147 = 0, $148 = 0, $149 = 0.0, $15 = 0.0, $150 = 0.0, $151 = 0;
 var $152 = 0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0.0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0.0, $17 = 0.0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0.0, $174 = 0, $175 = 0, $176 = 0.0, $177 = 0.0, $178 = 0, $179 = 0, $18 = 0, $180 = 0.0, $181 = 0.0, $182 = 0.0, $183 = 0.0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0, $193 = 0, $194 = 0.0, $195 = 0, $196 = 0, $197 = 0.0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0.0, $202 = 0.0, $203 = 0.0, $204 = 0.0, $205 = 0.0;
 var $206 = 0.0, $207 = 0.0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0.0, $215 = 0, $216 = 0, $217 = 0, $218 = 0.0, $219 = 0, $22 = 0, $220 = 0, $221 = 0.0, $222 = 0.0, $223 = 0;
 var $224 = 0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0.0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0.0, $236 = 0, $237 = 0, $238 = 0, $239 = 0.0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0.0, $243 = 0.0, $244 = 0, $245 = 0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0.0, $254 = 0.0, $255 = 0.0, $256 = 0.0, $257 = 0.0, $258 = 0.0, $259 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0;
 var $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0.0, $9 = 0.0, $90 = 0, $91 = 0, $92 = 0, $93 = 0.0, $94 = 0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0, $99 = 0;
 var $not$ = 0, $not$1 = 0, $not$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (~~(($x))); //@line 262 "libraries/noise1234/noise1234.cpp"
 $not$ = !($x > 0.0); //@line 262 "libraries/noise1234/noise1234.cpp"
 $1 = $not$ << 31 >> 31; //@line 262 "libraries/noise1234/noise1234.cpp"
 $2 = (($1) + ($0))|0; //@line 262 "libraries/noise1234/noise1234.cpp"
 $3 = (~~(($y))); //@line 263 "libraries/noise1234/noise1234.cpp"
 $not$1 = !($y > 0.0); //@line 263 "libraries/noise1234/noise1234.cpp"
 $4 = $not$1 << 31 >> 31; //@line 263 "libraries/noise1234/noise1234.cpp"
 $5 = (($4) + ($3))|0; //@line 263 "libraries/noise1234/noise1234.cpp"
 $6 = (~~(($z))); //@line 264 "libraries/noise1234/noise1234.cpp"
 $not$2 = !($z > 0.0); //@line 264 "libraries/noise1234/noise1234.cpp"
 $7 = $not$2 << 31 >> 31; //@line 264 "libraries/noise1234/noise1234.cpp"
 $8 = (($7) + ($6))|0; //@line 264 "libraries/noise1234/noise1234.cpp"
 $9 = (+($2|0)); //@line 265 "libraries/noise1234/noise1234.cpp"
 $10 = $x - $9; //@line 265 "libraries/noise1234/noise1234.cpp"
 $11 = (+($5|0)); //@line 266 "libraries/noise1234/noise1234.cpp"
 $12 = $y - $11; //@line 266 "libraries/noise1234/noise1234.cpp"
 $13 = (+($8|0)); //@line 267 "libraries/noise1234/noise1234.cpp"
 $14 = $z - $13; //@line 267 "libraries/noise1234/noise1234.cpp"
 $15 = $10 + -1.0; //@line 268 "libraries/noise1234/noise1234.cpp"
 $16 = $12 + -1.0; //@line 269 "libraries/noise1234/noise1234.cpp"
 $17 = $14 + -1.0; //@line 270 "libraries/noise1234/noise1234.cpp"
 $18 = (($2) + 1)|0; //@line 271 "libraries/noise1234/noise1234.cpp"
 $19 = $18 & 255; //@line 271 "libraries/noise1234/noise1234.cpp"
 $20 = (($5) + 1)|0; //@line 272 "libraries/noise1234/noise1234.cpp"
 $21 = $20 & 255; //@line 272 "libraries/noise1234/noise1234.cpp"
 $22 = (($8) + 1)|0; //@line 273 "libraries/noise1234/noise1234.cpp"
 $23 = $22 & 255; //@line 273 "libraries/noise1234/noise1234.cpp"
 $24 = $2 & 255; //@line 274 "libraries/noise1234/noise1234.cpp"
 $25 = $5 & 255; //@line 275 "libraries/noise1234/noise1234.cpp"
 $26 = $8 & 255; //@line 276 "libraries/noise1234/noise1234.cpp"
 $27 = $14 * $14; //@line 278 "libraries/noise1234/noise1234.cpp"
 $28 = $14 * $27; //@line 278 "libraries/noise1234/noise1234.cpp"
 $29 = $14 * 6.0; //@line 278 "libraries/noise1234/noise1234.cpp"
 $30 = $29 + -15.0; //@line 278 "libraries/noise1234/noise1234.cpp"
 $31 = $14 * $30; //@line 278 "libraries/noise1234/noise1234.cpp"
 $32 = $31 + 10.0; //@line 278 "libraries/noise1234/noise1234.cpp"
 $33 = $28 * $32; //@line 278 "libraries/noise1234/noise1234.cpp"
 $34 = $12 * $12; //@line 279 "libraries/noise1234/noise1234.cpp"
 $35 = $12 * $34; //@line 279 "libraries/noise1234/noise1234.cpp"
 $36 = $12 * 6.0; //@line 279 "libraries/noise1234/noise1234.cpp"
 $37 = $36 + -15.0; //@line 279 "libraries/noise1234/noise1234.cpp"
 $38 = $12 * $37; //@line 279 "libraries/noise1234/noise1234.cpp"
 $39 = $38 + 10.0; //@line 279 "libraries/noise1234/noise1234.cpp"
 $40 = $35 * $39; //@line 279 "libraries/noise1234/noise1234.cpp"
 $41 = $10 * $10; //@line 280 "libraries/noise1234/noise1234.cpp"
 $42 = $10 * $41; //@line 280 "libraries/noise1234/noise1234.cpp"
 $43 = $10 * 6.0; //@line 280 "libraries/noise1234/noise1234.cpp"
 $44 = $43 + -15.0; //@line 280 "libraries/noise1234/noise1234.cpp"
 $45 = $10 * $44; //@line 280 "libraries/noise1234/noise1234.cpp"
 $46 = $45 + 10.0; //@line 280 "libraries/noise1234/noise1234.cpp"
 $47 = $42 * $46; //@line 280 "libraries/noise1234/noise1234.cpp"
 $48 = (326027 + ($26)|0); //@line 282 "libraries/noise1234/noise1234.cpp"
 $49 = HEAP8[$48>>0]|0; //@line 296 "libraries/noise1234/noise1234.cpp"
 $50 = $49&255; //@line 282 "libraries/noise1234/noise1234.cpp"
 $51 = (($50) + ($25))|0; //@line 282 "libraries/noise1234/noise1234.cpp"
 $52 = (326027 + ($51)|0); //@line 282 "libraries/noise1234/noise1234.cpp"
 $53 = HEAP8[$52>>0]|0; //@line 292 "libraries/noise1234/noise1234.cpp"
 $54 = $53&255; //@line 282 "libraries/noise1234/noise1234.cpp"
 $55 = (($54) + ($24))|0; //@line 282 "libraries/noise1234/noise1234.cpp"
 $56 = (326027 + ($55)|0); //@line 282 "libraries/noise1234/noise1234.cpp"
 $57 = HEAP8[$56>>0]|0; //@line 282 "libraries/noise1234/noise1234.cpp"
 $58 = $57&255; //@line 282 "libraries/noise1234/noise1234.cpp"
 $59 = $58 & 15; //@line 124 "libraries/noise1234/noise1234.cpp"
 $60 = ($59>>>0)<(8); //@line 125 "libraries/noise1234/noise1234.cpp"
 $61 = $60 ? $10 : $12; //@line 125 "libraries/noise1234/noise1234.cpp"
 $62 = ($59>>>0)<(4); //@line 126 "libraries/noise1234/noise1234.cpp"
 if ($62) {
  $73 = $12;
 } else {
  $63 = $59 | 2; //@line 126 "libraries/noise1234/noise1234.cpp"
  $64 = ($63|0)==(14); //@line 126 "libraries/noise1234/noise1234.cpp"
  $65 = $64 ? $10 : $14; //@line 126 "libraries/noise1234/noise1234.cpp"
  $73 = $65;
 }
 $66 = $58 & 1; //@line 127 "libraries/noise1234/noise1234.cpp"
 $67 = ($66|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
 $68 = -$61; //@line 127 "libraries/noise1234/noise1234.cpp"
 $69 = $67 ? $68 : $61; //@line 127 "libraries/noise1234/noise1234.cpp"
 $70 = $58 & 2; //@line 127 "libraries/noise1234/noise1234.cpp"
 $71 = ($70|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
 $72 = -$73; //@line 127 "libraries/noise1234/noise1234.cpp"
 $74 = $71 ? $72 : $73; //@line 127 "libraries/noise1234/noise1234.cpp"
 $75 = $69 + $74; //@line 127 "libraries/noise1234/noise1234.cpp"
 $76 = (326027 + ($23)|0); //@line 283 "libraries/noise1234/noise1234.cpp"
 $77 = HEAP8[$76>>0]|0; //@line 297 "libraries/noise1234/noise1234.cpp"
 $78 = $77&255; //@line 283 "libraries/noise1234/noise1234.cpp"
 $79 = (($78) + ($25))|0; //@line 283 "libraries/noise1234/noise1234.cpp"
 $80 = (326027 + ($79)|0); //@line 283 "libraries/noise1234/noise1234.cpp"
 $81 = HEAP8[$80>>0]|0; //@line 293 "libraries/noise1234/noise1234.cpp"
 $82 = $81&255; //@line 283 "libraries/noise1234/noise1234.cpp"
 $83 = (($82) + ($24))|0; //@line 283 "libraries/noise1234/noise1234.cpp"
 $84 = (326027 + ($83)|0); //@line 283 "libraries/noise1234/noise1234.cpp"
 $85 = HEAP8[$84>>0]|0; //@line 283 "libraries/noise1234/noise1234.cpp"
 $86 = $85&255; //@line 283 "libraries/noise1234/noise1234.cpp"
 $87 = $86 & 15; //@line 124 "libraries/noise1234/noise1234.cpp"
 $88 = ($87>>>0)<(8); //@line 125 "libraries/noise1234/noise1234.cpp"
 $89 = $88 ? $10 : $12; //@line 125 "libraries/noise1234/noise1234.cpp"
 $90 = ($87>>>0)<(4); //@line 126 "libraries/noise1234/noise1234.cpp"
 if ($90) {
  $101 = $12;
 } else {
  $91 = $87 | 2; //@line 126 "libraries/noise1234/noise1234.cpp"
  $92 = ($91|0)==(14); //@line 126 "libraries/noise1234/noise1234.cpp"
  $93 = $92 ? $10 : $17; //@line 126 "libraries/noise1234/noise1234.cpp"
  $101 = $93;
 }
 $94 = $86 & 1; //@line 127 "libraries/noise1234/noise1234.cpp"
 $95 = ($94|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
 $96 = -$89; //@line 127 "libraries/noise1234/noise1234.cpp"
 $97 = $95 ? $96 : $89; //@line 127 "libraries/noise1234/noise1234.cpp"
 $98 = $86 & 2; //@line 127 "libraries/noise1234/noise1234.cpp"
 $99 = ($98|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
 $100 = -$101; //@line 127 "libraries/noise1234/noise1234.cpp"
 $102 = $99 ? $100 : $101; //@line 127 "libraries/noise1234/noise1234.cpp"
 $103 = $97 + $102; //@line 127 "libraries/noise1234/noise1234.cpp"
 $104 = $103 - $75; //@line 284 "libraries/noise1234/noise1234.cpp"
 $105 = $33 * $104; //@line 284 "libraries/noise1234/noise1234.cpp"
 $106 = $75 + $105; //@line 284 "libraries/noise1234/noise1234.cpp"
 $107 = (($50) + ($21))|0; //@line 286 "libraries/noise1234/noise1234.cpp"
 $108 = (326027 + ($107)|0); //@line 286 "libraries/noise1234/noise1234.cpp"
 $109 = HEAP8[$108>>0]|0; //@line 296 "libraries/noise1234/noise1234.cpp"
 $110 = $109&255; //@line 286 "libraries/noise1234/noise1234.cpp"
 $111 = (($110) + ($24))|0; //@line 286 "libraries/noise1234/noise1234.cpp"
 $112 = (326027 + ($111)|0); //@line 286 "libraries/noise1234/noise1234.cpp"
 $113 = HEAP8[$112>>0]|0; //@line 286 "libraries/noise1234/noise1234.cpp"
 $114 = $113&255; //@line 286 "libraries/noise1234/noise1234.cpp"
 $115 = $114 & 15; //@line 124 "libraries/noise1234/noise1234.cpp"
 $116 = ($115>>>0)<(8); //@line 125 "libraries/noise1234/noise1234.cpp"
 $117 = $116 ? $10 : $16; //@line 125 "libraries/noise1234/noise1234.cpp"
 $118 = ($115>>>0)<(4); //@line 126 "libraries/noise1234/noise1234.cpp"
 if ($118) {
  $129 = $16;
 } else {
  $119 = $115 | 2; //@line 126 "libraries/noise1234/noise1234.cpp"
  $120 = ($119|0)==(14); //@line 126 "libraries/noise1234/noise1234.cpp"
  $121 = $120 ? $10 : $14; //@line 126 "libraries/noise1234/noise1234.cpp"
  $129 = $121;
 }
 $122 = $114 & 1; //@line 127 "libraries/noise1234/noise1234.cpp"
 $123 = ($122|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
 $124 = -$117; //@line 127 "libraries/noise1234/noise1234.cpp"
 $125 = $123 ? $124 : $117; //@line 127 "libraries/noise1234/noise1234.cpp"
 $126 = $114 & 2; //@line 127 "libraries/noise1234/noise1234.cpp"
 $127 = ($126|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
 $128 = -$129; //@line 127 "libraries/noise1234/noise1234.cpp"
 $130 = $127 ? $128 : $129; //@line 127 "libraries/noise1234/noise1234.cpp"
 $131 = $125 + $130; //@line 127 "libraries/noise1234/noise1234.cpp"
 $132 = (($78) + ($21))|0; //@line 287 "libraries/noise1234/noise1234.cpp"
 $133 = (326027 + ($132)|0); //@line 287 "libraries/noise1234/noise1234.cpp"
 $134 = HEAP8[$133>>0]|0; //@line 297 "libraries/noise1234/noise1234.cpp"
 $135 = $134&255; //@line 287 "libraries/noise1234/noise1234.cpp"
 $136 = (($135) + ($24))|0; //@line 287 "libraries/noise1234/noise1234.cpp"
 $137 = (326027 + ($136)|0); //@line 287 "libraries/noise1234/noise1234.cpp"
 $138 = HEAP8[$137>>0]|0; //@line 287 "libraries/noise1234/noise1234.cpp"
 $139 = $138&255; //@line 287 "libraries/noise1234/noise1234.cpp"
 $140 = $139 & 15; //@line 124 "libraries/noise1234/noise1234.cpp"
 $141 = ($140>>>0)<(8); //@line 125 "libraries/noise1234/noise1234.cpp"
 $142 = $141 ? $10 : $16; //@line 125 "libraries/noise1234/noise1234.cpp"
 $143 = ($140>>>0)<(4); //@line 126 "libraries/noise1234/noise1234.cpp"
 if ($143) {
  $154 = $16;
 } else {
  $144 = $140 | 2; //@line 126 "libraries/noise1234/noise1234.cpp"
  $145 = ($144|0)==(14); //@line 126 "libraries/noise1234/noise1234.cpp"
  $146 = $145 ? $10 : $17; //@line 126 "libraries/noise1234/noise1234.cpp"
  $154 = $146;
 }
 $147 = $139 & 1; //@line 127 "libraries/noise1234/noise1234.cpp"
 $148 = ($147|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
 $149 = -$142; //@line 127 "libraries/noise1234/noise1234.cpp"
 $150 = $148 ? $149 : $142; //@line 127 "libraries/noise1234/noise1234.cpp"
 $151 = $139 & 2; //@line 127 "libraries/noise1234/noise1234.cpp"
 $152 = ($151|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
 $153 = -$154; //@line 127 "libraries/noise1234/noise1234.cpp"
 $155 = $152 ? $153 : $154; //@line 127 "libraries/noise1234/noise1234.cpp"
 $156 = $150 + $155; //@line 127 "libraries/noise1234/noise1234.cpp"
 $157 = $156 - $131; //@line 288 "libraries/noise1234/noise1234.cpp"
 $158 = $33 * $157; //@line 288 "libraries/noise1234/noise1234.cpp"
 $159 = $131 + $158; //@line 288 "libraries/noise1234/noise1234.cpp"
 $160 = $159 - $106; //@line 290 "libraries/noise1234/noise1234.cpp"
 $161 = $40 * $160; //@line 290 "libraries/noise1234/noise1234.cpp"
 $162 = $106 + $161; //@line 290 "libraries/noise1234/noise1234.cpp"
 $163 = (($54) + ($19))|0; //@line 292 "libraries/noise1234/noise1234.cpp"
 $164 = (326027 + ($163)|0); //@line 292 "libraries/noise1234/noise1234.cpp"
 $165 = HEAP8[$164>>0]|0; //@line 292 "libraries/noise1234/noise1234.cpp"
 $166 = $165&255; //@line 292 "libraries/noise1234/noise1234.cpp"
 $167 = $166 & 15; //@line 124 "libraries/noise1234/noise1234.cpp"
 $168 = ($167>>>0)<(8); //@line 125 "libraries/noise1234/noise1234.cpp"
 $169 = $168 ? $15 : $12; //@line 125 "libraries/noise1234/noise1234.cpp"
 $170 = ($167>>>0)<(4); //@line 126 "libraries/noise1234/noise1234.cpp"
 if ($170) {
  $181 = $12;
 } else {
  $171 = $167 | 2; //@line 126 "libraries/noise1234/noise1234.cpp"
  $172 = ($171|0)==(14); //@line 126 "libraries/noise1234/noise1234.cpp"
  $173 = $172 ? $15 : $14; //@line 126 "libraries/noise1234/noise1234.cpp"
  $181 = $173;
 }
 $174 = $166 & 1; //@line 127 "libraries/noise1234/noise1234.cpp"
 $175 = ($174|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
 $176 = -$169; //@line 127 "libraries/noise1234/noise1234.cpp"
 $177 = $175 ? $176 : $169; //@line 127 "libraries/noise1234/noise1234.cpp"
 $178 = $166 & 2; //@line 127 "libraries/noise1234/noise1234.cpp"
 $179 = ($178|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
 $180 = -$181; //@line 127 "libraries/noise1234/noise1234.cpp"
 $182 = $179 ? $180 : $181; //@line 127 "libraries/noise1234/noise1234.cpp"
 $183 = $177 + $182; //@line 127 "libraries/noise1234/noise1234.cpp"
 $184 = (($82) + ($19))|0; //@line 293 "libraries/noise1234/noise1234.cpp"
 $185 = (326027 + ($184)|0); //@line 293 "libraries/noise1234/noise1234.cpp"
 $186 = HEAP8[$185>>0]|0; //@line 293 "libraries/noise1234/noise1234.cpp"
 $187 = $186&255; //@line 293 "libraries/noise1234/noise1234.cpp"
 $188 = $187 & 15; //@line 124 "libraries/noise1234/noise1234.cpp"
 $189 = ($188>>>0)<(8); //@line 125 "libraries/noise1234/noise1234.cpp"
 $190 = $189 ? $15 : $12; //@line 125 "libraries/noise1234/noise1234.cpp"
 $191 = ($188>>>0)<(4); //@line 126 "libraries/noise1234/noise1234.cpp"
 if ($191) {
  $202 = $12;
 } else {
  $192 = $188 | 2; //@line 126 "libraries/noise1234/noise1234.cpp"
  $193 = ($192|0)==(14); //@line 126 "libraries/noise1234/noise1234.cpp"
  $194 = $193 ? $15 : $17; //@line 126 "libraries/noise1234/noise1234.cpp"
  $202 = $194;
 }
 $195 = $187 & 1; //@line 127 "libraries/noise1234/noise1234.cpp"
 $196 = ($195|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
 $197 = -$190; //@line 127 "libraries/noise1234/noise1234.cpp"
 $198 = $196 ? $197 : $190; //@line 127 "libraries/noise1234/noise1234.cpp"
 $199 = $187 & 2; //@line 127 "libraries/noise1234/noise1234.cpp"
 $200 = ($199|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
 $201 = -$202; //@line 127 "libraries/noise1234/noise1234.cpp"
 $203 = $200 ? $201 : $202; //@line 127 "libraries/noise1234/noise1234.cpp"
 $204 = $198 + $203; //@line 127 "libraries/noise1234/noise1234.cpp"
 $205 = $204 - $183; //@line 294 "libraries/noise1234/noise1234.cpp"
 $206 = $33 * $205; //@line 294 "libraries/noise1234/noise1234.cpp"
 $207 = $183 + $206; //@line 294 "libraries/noise1234/noise1234.cpp"
 $208 = (($110) + ($19))|0; //@line 296 "libraries/noise1234/noise1234.cpp"
 $209 = (326027 + ($208)|0); //@line 296 "libraries/noise1234/noise1234.cpp"
 $210 = HEAP8[$209>>0]|0; //@line 296 "libraries/noise1234/noise1234.cpp"
 $211 = $210&255; //@line 296 "libraries/noise1234/noise1234.cpp"
 $212 = $211 & 15; //@line 124 "libraries/noise1234/noise1234.cpp"
 $213 = ($212>>>0)<(8); //@line 125 "libraries/noise1234/noise1234.cpp"
 $214 = $213 ? $15 : $16; //@line 125 "libraries/noise1234/noise1234.cpp"
 $215 = ($212>>>0)<(4); //@line 126 "libraries/noise1234/noise1234.cpp"
 if ($215) {
  $226 = $16;
 } else {
  $216 = $212 | 2; //@line 126 "libraries/noise1234/noise1234.cpp"
  $217 = ($216|0)==(14); //@line 126 "libraries/noise1234/noise1234.cpp"
  $218 = $217 ? $15 : $14; //@line 126 "libraries/noise1234/noise1234.cpp"
  $226 = $218;
 }
 $219 = $211 & 1; //@line 127 "libraries/noise1234/noise1234.cpp"
 $220 = ($219|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
 $221 = -$214; //@line 127 "libraries/noise1234/noise1234.cpp"
 $222 = $220 ? $221 : $214; //@line 127 "libraries/noise1234/noise1234.cpp"
 $223 = $211 & 2; //@line 127 "libraries/noise1234/noise1234.cpp"
 $224 = ($223|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
 $225 = -$226; //@line 127 "libraries/noise1234/noise1234.cpp"
 $227 = $224 ? $225 : $226; //@line 127 "libraries/noise1234/noise1234.cpp"
 $228 = $222 + $227; //@line 127 "libraries/noise1234/noise1234.cpp"
 $229 = (($135) + ($19))|0; //@line 297 "libraries/noise1234/noise1234.cpp"
 $230 = (326027 + ($229)|0); //@line 297 "libraries/noise1234/noise1234.cpp"
 $231 = HEAP8[$230>>0]|0; //@line 297 "libraries/noise1234/noise1234.cpp"
 $232 = $231&255; //@line 297 "libraries/noise1234/noise1234.cpp"
 $233 = $232 & 15; //@line 124 "libraries/noise1234/noise1234.cpp"
 $234 = ($233>>>0)<(8); //@line 125 "libraries/noise1234/noise1234.cpp"
 $235 = $234 ? $15 : $16; //@line 125 "libraries/noise1234/noise1234.cpp"
 $236 = ($233>>>0)<(4); //@line 126 "libraries/noise1234/noise1234.cpp"
 if ($236) {
  $247 = $16;
  $240 = $232 & 1; //@line 127 "libraries/noise1234/noise1234.cpp"
  $241 = ($240|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
  $242 = -$235; //@line 127 "libraries/noise1234/noise1234.cpp"
  $243 = $241 ? $242 : $235; //@line 127 "libraries/noise1234/noise1234.cpp"
  $244 = $232 & 2; //@line 127 "libraries/noise1234/noise1234.cpp"
  $245 = ($244|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
  $246 = -$247; //@line 127 "libraries/noise1234/noise1234.cpp"
  $248 = $245 ? $246 : $247; //@line 127 "libraries/noise1234/noise1234.cpp"
  $249 = $243 + $248; //@line 127 "libraries/noise1234/noise1234.cpp"
  $250 = $249 - $228; //@line 298 "libraries/noise1234/noise1234.cpp"
  $251 = $33 * $250; //@line 298 "libraries/noise1234/noise1234.cpp"
  $252 = $228 + $251; //@line 298 "libraries/noise1234/noise1234.cpp"
  $253 = $252 - $207; //@line 300 "libraries/noise1234/noise1234.cpp"
  $254 = $40 * $253; //@line 300 "libraries/noise1234/noise1234.cpp"
  $255 = $207 + $254; //@line 300 "libraries/noise1234/noise1234.cpp"
  $256 = $255 - $162; //@line 302 "libraries/noise1234/noise1234.cpp"
  $257 = $47 * $256; //@line 302 "libraries/noise1234/noise1234.cpp"
  $258 = $162 + $257; //@line 302 "libraries/noise1234/noise1234.cpp"
  $259 = $258 * 0.93599998950958251; //@line 302 "libraries/noise1234/noise1234.cpp"
  return (+$259); //@line 302 "libraries/noise1234/noise1234.cpp"
 }
 $237 = $233 | 2; //@line 126 "libraries/noise1234/noise1234.cpp"
 $238 = ($237|0)==(14); //@line 126 "libraries/noise1234/noise1234.cpp"
 $239 = $238 ? $15 : $17; //@line 126 "libraries/noise1234/noise1234.cpp"
 $247 = $239;
 $240 = $232 & 1; //@line 127 "libraries/noise1234/noise1234.cpp"
 $241 = ($240|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
 $242 = -$235; //@line 127 "libraries/noise1234/noise1234.cpp"
 $243 = $241 ? $242 : $235; //@line 127 "libraries/noise1234/noise1234.cpp"
 $244 = $232 & 2; //@line 127 "libraries/noise1234/noise1234.cpp"
 $245 = ($244|0)!=(0); //@line 127 "libraries/noise1234/noise1234.cpp"
 $246 = -$247; //@line 127 "libraries/noise1234/noise1234.cpp"
 $248 = $245 ? $246 : $247; //@line 127 "libraries/noise1234/noise1234.cpp"
 $249 = $243 + $248; //@line 127 "libraries/noise1234/noise1234.cpp"
 $250 = $249 - $228; //@line 298 "libraries/noise1234/noise1234.cpp"
 $251 = $33 * $250; //@line 298 "libraries/noise1234/noise1234.cpp"
 $252 = $228 + $251; //@line 298 "libraries/noise1234/noise1234.cpp"
 $253 = $252 - $207; //@line 300 "libraries/noise1234/noise1234.cpp"
 $254 = $40 * $253; //@line 300 "libraries/noise1234/noise1234.cpp"
 $255 = $207 + $254; //@line 300 "libraries/noise1234/noise1234.cpp"
 $256 = $255 - $162; //@line 302 "libraries/noise1234/noise1234.cpp"
 $257 = $47 * $256; //@line 302 "libraries/noise1234/noise1234.cpp"
 $258 = $162 + $257; //@line 302 "libraries/noise1234/noise1234.cpp"
 $259 = $258 * 0.93599998950958251; //@line 302 "libraries/noise1234/noise1234.cpp"
 return (+$259); //@line 302 "libraries/noise1234/noise1234.cpp"
}
function __ZN9Noise12345noiseEffff($x,$y,$z,$w) {
 $x = +$x;
 $y = +$y;
 $z = +$z;
 $w = +$w;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0.0, $118 = 0, $119 = 0.0, $12 = 0.0, $120 = 0, $121 = 0.0, $122 = 0, $123 = 0, $124 = 0.0, $125 = 0.0, $126 = 0, $127 = 0, $128 = 0.0, $129 = 0.0, $13 = 0.0, $130 = 0.0, $131 = 0, $132 = 0, $133 = 0.0;
 var $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0, $14 = 0.0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0.0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0.0, $154 = 0, $155 = 0.0, $156 = 0, $157 = 0.0, $158 = 0, $159 = 0, $16 = 0.0, $160 = 0.0, $161 = 0.0, $162 = 0, $163 = 0, $164 = 0.0, $165 = 0.0, $166 = 0.0, $167 = 0, $168 = 0, $169 = 0.0, $17 = 0.0;
 var $170 = 0.0, $171 = 0.0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0.0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0.0, $187 = 0, $188 = 0.0;
 var $189 = 0, $19 = 0.0, $190 = 0.0, $191 = 0, $192 = 0, $193 = 0.0, $194 = 0.0, $195 = 0, $196 = 0, $197 = 0.0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0.0, $200 = 0, $201 = 0, $202 = 0.0, $203 = 0.0, $204 = 0.0, $205 = 0.0;
 var $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0.0, $21 = 0.0, $210 = 0.0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0.0, $220 = 0, $221 = 0.0, $222 = 0, $223 = 0.0;
 var $224 = 0, $225 = 0.0, $226 = 0, $227 = 0, $228 = 0.0, $229 = 0.0, $23 = 0.0, $230 = 0, $231 = 0, $232 = 0.0, $233 = 0.0, $234 = 0.0, $235 = 0, $236 = 0, $237 = 0.0, $238 = 0.0, $239 = 0.0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0.0, $251 = 0, $252 = 0.0, $253 = 0, $254 = 0.0, $255 = 0, $256 = 0, $257 = 0.0, $258 = 0.0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0.0, $262 = 0.0, $263 = 0.0, $264 = 0, $265 = 0, $266 = 0.0, $267 = 0.0, $268 = 0.0, $269 = 0.0, $27 = 0, $270 = 0.0, $271 = 0.0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0.0, $283 = 0, $284 = 0.0, $285 = 0, $286 = 0.0, $287 = 0, $288 = 0, $289 = 0.0, $29 = 0, $290 = 0.0, $291 = 0, $292 = 0, $293 = 0.0, $294 = 0.0, $295 = 0.0, $296 = 0;
 var $297 = 0, $298 = 0.0, $299 = 0.0, $3 = 0, $30 = 0, $300 = 0.0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0.0, $312 = 0, $313 = 0.0;
 var $314 = 0, $315 = 0.0, $316 = 0, $317 = 0, $318 = 0.0, $319 = 0.0, $32 = 0, $320 = 0, $321 = 0, $322 = 0.0, $323 = 0.0, $324 = 0.0, $325 = 0, $326 = 0, $327 = 0.0, $328 = 0.0, $329 = 0.0, $33 = 0, $330 = 0.0, $331 = 0.0;
 var $332 = 0.0, $333 = 0.0, $334 = 0.0, $335 = 0.0, $336 = 0.0, $337 = 0.0, $338 = 0.0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0.0, $346 = 0, $347 = 0.0, $348 = 0, $349 = 0.0, $35 = 0;
 var $350 = 0, $351 = 0, $352 = 0.0, $353 = 0.0, $354 = 0, $355 = 0, $356 = 0.0, $357 = 0.0, $358 = 0.0, $359 = 0, $36 = 0.0, $360 = 0, $361 = 0.0, $362 = 0.0, $363 = 0.0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0;
 var $369 = 0, $37 = 0.0, $370 = 0.0, $371 = 0, $372 = 0.0, $373 = 0, $374 = 0.0, $375 = 0, $376 = 0, $377 = 0.0, $378 = 0.0, $379 = 0, $38 = 0.0, $380 = 0, $381 = 0.0, $382 = 0.0, $383 = 0.0, $384 = 0, $385 = 0, $386 = 0.0;
 var $387 = 0.0, $388 = 0.0, $389 = 0.0, $39 = 0.0, $390 = 0.0, $391 = 0.0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0.0, $399 = 0, $4 = 0, $40 = 0.0, $400 = 0.0, $401 = 0, $402 = 0.0, $403 = 0;
 var $404 = 0, $405 = 0.0, $406 = 0.0, $407 = 0, $408 = 0, $409 = 0.0, $41 = 0.0, $410 = 0.0, $411 = 0.0, $412 = 0, $413 = 0, $414 = 0.0, $415 = 0.0, $416 = 0.0, $417 = 0, $418 = 0, $419 = 0, $42 = 0.0, $420 = 0, $421 = 0;
 var $422 = 0, $423 = 0.0, $424 = 0, $425 = 0.0, $426 = 0, $427 = 0.0, $428 = 0, $429 = 0, $43 = 0.0, $430 = 0.0, $431 = 0.0, $432 = 0, $433 = 0, $434 = 0.0, $435 = 0.0, $436 = 0.0, $437 = 0, $438 = 0, $439 = 0.0, $44 = 0.0;
 var $440 = 0.0, $441 = 0.0, $442 = 0.0, $443 = 0.0, $444 = 0.0, $445 = 0.0, $446 = 0.0, $447 = 0.0, $448 = 0, $449 = 0, $45 = 0.0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0.0, $455 = 0, $456 = 0.0, $457 = 0, $458 = 0.0;
 var $459 = 0, $46 = 0.0, $460 = 0, $461 = 0.0, $462 = 0.0, $463 = 0, $464 = 0, $465 = 0.0, $466 = 0.0, $467 = 0.0, $468 = 0, $469 = 0, $47 = 0.0, $470 = 0.0, $471 = 0.0, $472 = 0.0, $473 = 0, $474 = 0, $475 = 0, $476 = 0;
 var $477 = 0, $478 = 0, $479 = 0.0, $48 = 0.0, $480 = 0, $481 = 0.0, $482 = 0, $483 = 0.0, $484 = 0, $485 = 0, $486 = 0.0, $487 = 0.0, $488 = 0, $489 = 0, $49 = 0.0, $490 = 0.0, $491 = 0.0, $492 = 0.0, $493 = 0, $494 = 0;
 var $495 = 0.0, $496 = 0.0, $497 = 0.0, $498 = 0.0, $499 = 0.0, $5 = 0, $50 = 0.0, $500 = 0.0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0.0, $508 = 0, $509 = 0.0, $51 = 0.0, $510 = 0, $511 = 0.0;
 var $512 = 0, $513 = 0, $514 = 0.0, $515 = 0.0, $516 = 0, $517 = 0, $518 = 0.0, $519 = 0.0, $52 = 0.0, $520 = 0.0, $521 = 0, $522 = 0, $523 = 0.0, $524 = 0.0, $525 = 0.0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0.0;
 var $530 = 0, $531 = 0, $532 = 0.0, $533 = 0, $534 = 0.0, $535 = 0, $536 = 0.0, $537 = 0, $538 = 0, $539 = 0.0, $54 = 0.0, $540 = 0.0, $541 = 0, $542 = 0, $543 = 0.0, $544 = 0.0, $545 = 0.0, $546 = 0, $547 = 0, $548 = 0.0;
 var $549 = 0.0, $55 = 0.0, $550 = 0.0, $551 = 0.0, $552 = 0.0, $553 = 0.0, $554 = 0.0, $555 = 0.0, $556 = 0.0, $557 = 0.0, $558 = 0.0, $559 = 0.0, $56 = 0.0, $560 = 0.0, $561 = 0.0, $562 = 0.0, $563 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0;
 var $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0, $87 = 0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0, $91 = 0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0;
 var $96 = 0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $not$ = 0, $not$1 = 0, $not$2 = 0, $not$3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (~~(($x))); //@line 370 "libraries/noise1234/noise1234.cpp"
 $not$ = !($x > 0.0); //@line 370 "libraries/noise1234/noise1234.cpp"
 $1 = $not$ << 31 >> 31; //@line 370 "libraries/noise1234/noise1234.cpp"
 $2 = (($1) + ($0))|0; //@line 370 "libraries/noise1234/noise1234.cpp"
 $3 = (~~(($y))); //@line 371 "libraries/noise1234/noise1234.cpp"
 $not$1 = !($y > 0.0); //@line 371 "libraries/noise1234/noise1234.cpp"
 $4 = $not$1 << 31 >> 31; //@line 371 "libraries/noise1234/noise1234.cpp"
 $5 = (($4) + ($3))|0; //@line 371 "libraries/noise1234/noise1234.cpp"
 $6 = (~~(($z))); //@line 372 "libraries/noise1234/noise1234.cpp"
 $not$2 = !($z > 0.0); //@line 372 "libraries/noise1234/noise1234.cpp"
 $7 = $not$2 << 31 >> 31; //@line 372 "libraries/noise1234/noise1234.cpp"
 $8 = (($7) + ($6))|0; //@line 372 "libraries/noise1234/noise1234.cpp"
 $9 = (~~(($w))); //@line 373 "libraries/noise1234/noise1234.cpp"
 $not$3 = !($w > 0.0); //@line 373 "libraries/noise1234/noise1234.cpp"
 $10 = $not$3 << 31 >> 31; //@line 373 "libraries/noise1234/noise1234.cpp"
 $11 = (($10) + ($9))|0; //@line 373 "libraries/noise1234/noise1234.cpp"
 $12 = (+($2|0)); //@line 374 "libraries/noise1234/noise1234.cpp"
 $13 = $x - $12; //@line 374 "libraries/noise1234/noise1234.cpp"
 $14 = (+($5|0)); //@line 375 "libraries/noise1234/noise1234.cpp"
 $15 = $y - $14; //@line 375 "libraries/noise1234/noise1234.cpp"
 $16 = (+($8|0)); //@line 376 "libraries/noise1234/noise1234.cpp"
 $17 = $z - $16; //@line 376 "libraries/noise1234/noise1234.cpp"
 $18 = (+($11|0)); //@line 377 "libraries/noise1234/noise1234.cpp"
 $19 = $w - $18; //@line 377 "libraries/noise1234/noise1234.cpp"
 $20 = $13 + -1.0; //@line 378 "libraries/noise1234/noise1234.cpp"
 $21 = $15 + -1.0; //@line 379 "libraries/noise1234/noise1234.cpp"
 $22 = $17 + -1.0; //@line 380 "libraries/noise1234/noise1234.cpp"
 $23 = $19 + -1.0; //@line 381 "libraries/noise1234/noise1234.cpp"
 $24 = (($2) + 1)|0; //@line 382 "libraries/noise1234/noise1234.cpp"
 $25 = $24 & 255; //@line 382 "libraries/noise1234/noise1234.cpp"
 $26 = (($5) + 1)|0; //@line 383 "libraries/noise1234/noise1234.cpp"
 $27 = $26 & 255; //@line 383 "libraries/noise1234/noise1234.cpp"
 $28 = (($8) + 1)|0; //@line 384 "libraries/noise1234/noise1234.cpp"
 $29 = $28 & 255; //@line 384 "libraries/noise1234/noise1234.cpp"
 $30 = (($11) + 1)|0; //@line 385 "libraries/noise1234/noise1234.cpp"
 $31 = $30 & 255; //@line 385 "libraries/noise1234/noise1234.cpp"
 $32 = $2 & 255; //@line 386 "libraries/noise1234/noise1234.cpp"
 $33 = $5 & 255; //@line 387 "libraries/noise1234/noise1234.cpp"
 $34 = $8 & 255; //@line 388 "libraries/noise1234/noise1234.cpp"
 $35 = $11 & 255; //@line 389 "libraries/noise1234/noise1234.cpp"
 $36 = $19 * $19; //@line 391 "libraries/noise1234/noise1234.cpp"
 $37 = $19 * $36; //@line 391 "libraries/noise1234/noise1234.cpp"
 $38 = $19 * 6.0; //@line 391 "libraries/noise1234/noise1234.cpp"
 $39 = $38 + -15.0; //@line 391 "libraries/noise1234/noise1234.cpp"
 $40 = $19 * $39; //@line 391 "libraries/noise1234/noise1234.cpp"
 $41 = $40 + 10.0; //@line 391 "libraries/noise1234/noise1234.cpp"
 $42 = $37 * $41; //@line 391 "libraries/noise1234/noise1234.cpp"
 $43 = $17 * $17; //@line 392 "libraries/noise1234/noise1234.cpp"
 $44 = $17 * $43; //@line 392 "libraries/noise1234/noise1234.cpp"
 $45 = $17 * 6.0; //@line 392 "libraries/noise1234/noise1234.cpp"
 $46 = $45 + -15.0; //@line 392 "libraries/noise1234/noise1234.cpp"
 $47 = $17 * $46; //@line 392 "libraries/noise1234/noise1234.cpp"
 $48 = $47 + 10.0; //@line 392 "libraries/noise1234/noise1234.cpp"
 $49 = $44 * $48; //@line 392 "libraries/noise1234/noise1234.cpp"
 $50 = $15 * $15; //@line 393 "libraries/noise1234/noise1234.cpp"
 $51 = $15 * $50; //@line 393 "libraries/noise1234/noise1234.cpp"
 $52 = $15 * 6.0; //@line 393 "libraries/noise1234/noise1234.cpp"
 $53 = $52 + -15.0; //@line 393 "libraries/noise1234/noise1234.cpp"
 $54 = $15 * $53; //@line 393 "libraries/noise1234/noise1234.cpp"
 $55 = $54 + 10.0; //@line 393 "libraries/noise1234/noise1234.cpp"
 $56 = $51 * $55; //@line 393 "libraries/noise1234/noise1234.cpp"
 $57 = $13 * $13; //@line 394 "libraries/noise1234/noise1234.cpp"
 $58 = $13 * $57; //@line 394 "libraries/noise1234/noise1234.cpp"
 $59 = $13 * 6.0; //@line 394 "libraries/noise1234/noise1234.cpp"
 $60 = $59 + -15.0; //@line 394 "libraries/noise1234/noise1234.cpp"
 $61 = $13 * $60; //@line 394 "libraries/noise1234/noise1234.cpp"
 $62 = $61 + 10.0; //@line 394 "libraries/noise1234/noise1234.cpp"
 $63 = $58 * $62; //@line 394 "libraries/noise1234/noise1234.cpp"
 $64 = (326027 + ($35)|0); //@line 396 "libraries/noise1234/noise1234.cpp"
 $65 = HEAP8[$64>>0]|0; //@line 396 "libraries/noise1234/noise1234.cpp"
 $66 = $65&255; //@line 396 "libraries/noise1234/noise1234.cpp"
 $67 = (($66) + ($34))|0; //@line 396 "libraries/noise1234/noise1234.cpp"
 $68 = (326027 + ($67)|0); //@line 396 "libraries/noise1234/noise1234.cpp"
 $69 = HEAP8[$68>>0]|0; //@line 396 "libraries/noise1234/noise1234.cpp"
 $70 = $69&255; //@line 396 "libraries/noise1234/noise1234.cpp"
 $71 = (($70) + ($33))|0; //@line 396 "libraries/noise1234/noise1234.cpp"
 $72 = (326027 + ($71)|0); //@line 396 "libraries/noise1234/noise1234.cpp"
 $73 = HEAP8[$72>>0]|0; //@line 396 "libraries/noise1234/noise1234.cpp"
 $74 = $73&255; //@line 396 "libraries/noise1234/noise1234.cpp"
 $75 = (($74) + ($32))|0; //@line 396 "libraries/noise1234/noise1234.cpp"
 $76 = (326027 + ($75)|0); //@line 396 "libraries/noise1234/noise1234.cpp"
 $77 = HEAP8[$76>>0]|0; //@line 396 "libraries/noise1234/noise1234.cpp"
 $78 = $77&255; //@line 396 "libraries/noise1234/noise1234.cpp"
 $79 = $78 & 31; //@line 131 "libraries/noise1234/noise1234.cpp"
 $80 = ($79>>>0)<(24); //@line 132 "libraries/noise1234/noise1234.cpp"
 $81 = $80 ? $13 : $15; //@line 132 "libraries/noise1234/noise1234.cpp"
 $82 = ($79>>>0)<(16); //@line 133 "libraries/noise1234/noise1234.cpp"
 $83 = $82 ? $15 : $17; //@line 133 "libraries/noise1234/noise1234.cpp"
 $84 = ($79>>>0)<(8); //@line 134 "libraries/noise1234/noise1234.cpp"
 $85 = $84 ? $17 : $19; //@line 134 "libraries/noise1234/noise1234.cpp"
 $86 = $78 & 1; //@line 135 "libraries/noise1234/noise1234.cpp"
 $87 = ($86|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $88 = -$81; //@line 135 "libraries/noise1234/noise1234.cpp"
 $89 = $87 ? $88 : $81; //@line 135 "libraries/noise1234/noise1234.cpp"
 $90 = $78 & 2; //@line 135 "libraries/noise1234/noise1234.cpp"
 $91 = ($90|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $92 = -$83; //@line 135 "libraries/noise1234/noise1234.cpp"
 $93 = $91 ? $92 : $83; //@line 135 "libraries/noise1234/noise1234.cpp"
 $94 = $89 + $93; //@line 135 "libraries/noise1234/noise1234.cpp"
 $95 = $78 & 4; //@line 135 "libraries/noise1234/noise1234.cpp"
 $96 = ($95|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $97 = -$85; //@line 135 "libraries/noise1234/noise1234.cpp"
 $98 = $96 ? $97 : $85; //@line 135 "libraries/noise1234/noise1234.cpp"
 $99 = $98 + $94; //@line 135 "libraries/noise1234/noise1234.cpp"
 $100 = (326027 + ($31)|0); //@line 397 "libraries/noise1234/noise1234.cpp"
 $101 = HEAP8[$100>>0]|0; //@line 397 "libraries/noise1234/noise1234.cpp"
 $102 = $101&255; //@line 397 "libraries/noise1234/noise1234.cpp"
 $103 = (($102) + ($34))|0; //@line 397 "libraries/noise1234/noise1234.cpp"
 $104 = (326027 + ($103)|0); //@line 397 "libraries/noise1234/noise1234.cpp"
 $105 = HEAP8[$104>>0]|0; //@line 397 "libraries/noise1234/noise1234.cpp"
 $106 = $105&255; //@line 397 "libraries/noise1234/noise1234.cpp"
 $107 = (($106) + ($33))|0; //@line 397 "libraries/noise1234/noise1234.cpp"
 $108 = (326027 + ($107)|0); //@line 397 "libraries/noise1234/noise1234.cpp"
 $109 = HEAP8[$108>>0]|0; //@line 397 "libraries/noise1234/noise1234.cpp"
 $110 = $109&255; //@line 397 "libraries/noise1234/noise1234.cpp"
 $111 = (($110) + ($32))|0; //@line 397 "libraries/noise1234/noise1234.cpp"
 $112 = (326027 + ($111)|0); //@line 397 "libraries/noise1234/noise1234.cpp"
 $113 = HEAP8[$112>>0]|0; //@line 397 "libraries/noise1234/noise1234.cpp"
 $114 = $113&255; //@line 397 "libraries/noise1234/noise1234.cpp"
 $115 = $114 & 31; //@line 131 "libraries/noise1234/noise1234.cpp"
 $116 = ($115>>>0)<(24); //@line 132 "libraries/noise1234/noise1234.cpp"
 $117 = $116 ? $13 : $15; //@line 132 "libraries/noise1234/noise1234.cpp"
 $118 = ($115>>>0)<(16); //@line 133 "libraries/noise1234/noise1234.cpp"
 $119 = $118 ? $15 : $17; //@line 133 "libraries/noise1234/noise1234.cpp"
 $120 = ($115>>>0)<(8); //@line 134 "libraries/noise1234/noise1234.cpp"
 $121 = $120 ? $17 : $23; //@line 134 "libraries/noise1234/noise1234.cpp"
 $122 = $114 & 1; //@line 135 "libraries/noise1234/noise1234.cpp"
 $123 = ($122|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $124 = -$117; //@line 135 "libraries/noise1234/noise1234.cpp"
 $125 = $123 ? $124 : $117; //@line 135 "libraries/noise1234/noise1234.cpp"
 $126 = $114 & 2; //@line 135 "libraries/noise1234/noise1234.cpp"
 $127 = ($126|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $128 = -$119; //@line 135 "libraries/noise1234/noise1234.cpp"
 $129 = $127 ? $128 : $119; //@line 135 "libraries/noise1234/noise1234.cpp"
 $130 = $125 + $129; //@line 135 "libraries/noise1234/noise1234.cpp"
 $131 = $114 & 4; //@line 135 "libraries/noise1234/noise1234.cpp"
 $132 = ($131|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $133 = -$121; //@line 135 "libraries/noise1234/noise1234.cpp"
 $134 = $132 ? $133 : $121; //@line 135 "libraries/noise1234/noise1234.cpp"
 $135 = $134 + $130; //@line 135 "libraries/noise1234/noise1234.cpp"
 $136 = $135 - $99; //@line 398 "libraries/noise1234/noise1234.cpp"
 $137 = $42 * $136; //@line 398 "libraries/noise1234/noise1234.cpp"
 $138 = $99 + $137; //@line 398 "libraries/noise1234/noise1234.cpp"
 $139 = (($66) + ($29))|0; //@line 400 "libraries/noise1234/noise1234.cpp"
 $140 = (326027 + ($139)|0); //@line 400 "libraries/noise1234/noise1234.cpp"
 $141 = HEAP8[$140>>0]|0; //@line 400 "libraries/noise1234/noise1234.cpp"
 $142 = $141&255; //@line 400 "libraries/noise1234/noise1234.cpp"
 $143 = (($142) + ($33))|0; //@line 400 "libraries/noise1234/noise1234.cpp"
 $144 = (326027 + ($143)|0); //@line 400 "libraries/noise1234/noise1234.cpp"
 $145 = HEAP8[$144>>0]|0; //@line 400 "libraries/noise1234/noise1234.cpp"
 $146 = $145&255; //@line 400 "libraries/noise1234/noise1234.cpp"
 $147 = (($146) + ($32))|0; //@line 400 "libraries/noise1234/noise1234.cpp"
 $148 = (326027 + ($147)|0); //@line 400 "libraries/noise1234/noise1234.cpp"
 $149 = HEAP8[$148>>0]|0; //@line 400 "libraries/noise1234/noise1234.cpp"
 $150 = $149&255; //@line 400 "libraries/noise1234/noise1234.cpp"
 $151 = $150 & 31; //@line 131 "libraries/noise1234/noise1234.cpp"
 $152 = ($151>>>0)<(24); //@line 132 "libraries/noise1234/noise1234.cpp"
 $153 = $152 ? $13 : $15; //@line 132 "libraries/noise1234/noise1234.cpp"
 $154 = ($151>>>0)<(16); //@line 133 "libraries/noise1234/noise1234.cpp"
 $155 = $154 ? $15 : $22; //@line 133 "libraries/noise1234/noise1234.cpp"
 $156 = ($151>>>0)<(8); //@line 134 "libraries/noise1234/noise1234.cpp"
 $157 = $156 ? $22 : $19; //@line 134 "libraries/noise1234/noise1234.cpp"
 $158 = $150 & 1; //@line 135 "libraries/noise1234/noise1234.cpp"
 $159 = ($158|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $160 = -$153; //@line 135 "libraries/noise1234/noise1234.cpp"
 $161 = $159 ? $160 : $153; //@line 135 "libraries/noise1234/noise1234.cpp"
 $162 = $150 & 2; //@line 135 "libraries/noise1234/noise1234.cpp"
 $163 = ($162|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $164 = -$155; //@line 135 "libraries/noise1234/noise1234.cpp"
 $165 = $163 ? $164 : $155; //@line 135 "libraries/noise1234/noise1234.cpp"
 $166 = $161 + $165; //@line 135 "libraries/noise1234/noise1234.cpp"
 $167 = $150 & 4; //@line 135 "libraries/noise1234/noise1234.cpp"
 $168 = ($167|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $169 = -$157; //@line 135 "libraries/noise1234/noise1234.cpp"
 $170 = $168 ? $169 : $157; //@line 135 "libraries/noise1234/noise1234.cpp"
 $171 = $170 + $166; //@line 135 "libraries/noise1234/noise1234.cpp"
 $172 = (($102) + ($29))|0; //@line 401 "libraries/noise1234/noise1234.cpp"
 $173 = (326027 + ($172)|0); //@line 401 "libraries/noise1234/noise1234.cpp"
 $174 = HEAP8[$173>>0]|0; //@line 401 "libraries/noise1234/noise1234.cpp"
 $175 = $174&255; //@line 401 "libraries/noise1234/noise1234.cpp"
 $176 = (($175) + ($33))|0; //@line 401 "libraries/noise1234/noise1234.cpp"
 $177 = (326027 + ($176)|0); //@line 401 "libraries/noise1234/noise1234.cpp"
 $178 = HEAP8[$177>>0]|0; //@line 401 "libraries/noise1234/noise1234.cpp"
 $179 = $178&255; //@line 401 "libraries/noise1234/noise1234.cpp"
 $180 = (($179) + ($32))|0; //@line 401 "libraries/noise1234/noise1234.cpp"
 $181 = (326027 + ($180)|0); //@line 401 "libraries/noise1234/noise1234.cpp"
 $182 = HEAP8[$181>>0]|0; //@line 401 "libraries/noise1234/noise1234.cpp"
 $183 = $182&255; //@line 401 "libraries/noise1234/noise1234.cpp"
 $184 = $183 & 31; //@line 131 "libraries/noise1234/noise1234.cpp"
 $185 = ($184>>>0)<(24); //@line 132 "libraries/noise1234/noise1234.cpp"
 $186 = $185 ? $13 : $15; //@line 132 "libraries/noise1234/noise1234.cpp"
 $187 = ($184>>>0)<(16); //@line 133 "libraries/noise1234/noise1234.cpp"
 $188 = $187 ? $15 : $22; //@line 133 "libraries/noise1234/noise1234.cpp"
 $189 = ($184>>>0)<(8); //@line 134 "libraries/noise1234/noise1234.cpp"
 $190 = $189 ? $22 : $23; //@line 134 "libraries/noise1234/noise1234.cpp"
 $191 = $183 & 1; //@line 135 "libraries/noise1234/noise1234.cpp"
 $192 = ($191|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $193 = -$186; //@line 135 "libraries/noise1234/noise1234.cpp"
 $194 = $192 ? $193 : $186; //@line 135 "libraries/noise1234/noise1234.cpp"
 $195 = $183 & 2; //@line 135 "libraries/noise1234/noise1234.cpp"
 $196 = ($195|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $197 = -$188; //@line 135 "libraries/noise1234/noise1234.cpp"
 $198 = $196 ? $197 : $188; //@line 135 "libraries/noise1234/noise1234.cpp"
 $199 = $194 + $198; //@line 135 "libraries/noise1234/noise1234.cpp"
 $200 = $183 & 4; //@line 135 "libraries/noise1234/noise1234.cpp"
 $201 = ($200|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $202 = -$190; //@line 135 "libraries/noise1234/noise1234.cpp"
 $203 = $201 ? $202 : $190; //@line 135 "libraries/noise1234/noise1234.cpp"
 $204 = $203 + $199; //@line 135 "libraries/noise1234/noise1234.cpp"
 $205 = $204 - $171; //@line 402 "libraries/noise1234/noise1234.cpp"
 $206 = $42 * $205; //@line 402 "libraries/noise1234/noise1234.cpp"
 $207 = $171 + $206; //@line 402 "libraries/noise1234/noise1234.cpp"
 $208 = $207 - $138; //@line 404 "libraries/noise1234/noise1234.cpp"
 $209 = $49 * $208; //@line 404 "libraries/noise1234/noise1234.cpp"
 $210 = $138 + $209; //@line 404 "libraries/noise1234/noise1234.cpp"
 $211 = (($70) + ($27))|0; //@line 406 "libraries/noise1234/noise1234.cpp"
 $212 = (326027 + ($211)|0); //@line 406 "libraries/noise1234/noise1234.cpp"
 $213 = HEAP8[$212>>0]|0; //@line 406 "libraries/noise1234/noise1234.cpp"
 $214 = $213&255; //@line 406 "libraries/noise1234/noise1234.cpp"
 $215 = (($214) + ($32))|0; //@line 406 "libraries/noise1234/noise1234.cpp"
 $216 = (326027 + ($215)|0); //@line 406 "libraries/noise1234/noise1234.cpp"
 $217 = HEAP8[$216>>0]|0; //@line 406 "libraries/noise1234/noise1234.cpp"
 $218 = $217&255; //@line 406 "libraries/noise1234/noise1234.cpp"
 $219 = $218 & 31; //@line 131 "libraries/noise1234/noise1234.cpp"
 $220 = ($219>>>0)<(24); //@line 132 "libraries/noise1234/noise1234.cpp"
 $221 = $220 ? $13 : $21; //@line 132 "libraries/noise1234/noise1234.cpp"
 $222 = ($219>>>0)<(16); //@line 133 "libraries/noise1234/noise1234.cpp"
 $223 = $222 ? $21 : $17; //@line 133 "libraries/noise1234/noise1234.cpp"
 $224 = ($219>>>0)<(8); //@line 134 "libraries/noise1234/noise1234.cpp"
 $225 = $224 ? $17 : $19; //@line 134 "libraries/noise1234/noise1234.cpp"
 $226 = $218 & 1; //@line 135 "libraries/noise1234/noise1234.cpp"
 $227 = ($226|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $228 = -$221; //@line 135 "libraries/noise1234/noise1234.cpp"
 $229 = $227 ? $228 : $221; //@line 135 "libraries/noise1234/noise1234.cpp"
 $230 = $218 & 2; //@line 135 "libraries/noise1234/noise1234.cpp"
 $231 = ($230|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $232 = -$223; //@line 135 "libraries/noise1234/noise1234.cpp"
 $233 = $231 ? $232 : $223; //@line 135 "libraries/noise1234/noise1234.cpp"
 $234 = $229 + $233; //@line 135 "libraries/noise1234/noise1234.cpp"
 $235 = $218 & 4; //@line 135 "libraries/noise1234/noise1234.cpp"
 $236 = ($235|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $237 = -$225; //@line 135 "libraries/noise1234/noise1234.cpp"
 $238 = $236 ? $237 : $225; //@line 135 "libraries/noise1234/noise1234.cpp"
 $239 = $238 + $234; //@line 135 "libraries/noise1234/noise1234.cpp"
 $240 = (($106) + ($27))|0; //@line 407 "libraries/noise1234/noise1234.cpp"
 $241 = (326027 + ($240)|0); //@line 407 "libraries/noise1234/noise1234.cpp"
 $242 = HEAP8[$241>>0]|0; //@line 407 "libraries/noise1234/noise1234.cpp"
 $243 = $242&255; //@line 407 "libraries/noise1234/noise1234.cpp"
 $244 = (($243) + ($32))|0; //@line 407 "libraries/noise1234/noise1234.cpp"
 $245 = (326027 + ($244)|0); //@line 407 "libraries/noise1234/noise1234.cpp"
 $246 = HEAP8[$245>>0]|0; //@line 407 "libraries/noise1234/noise1234.cpp"
 $247 = $246&255; //@line 407 "libraries/noise1234/noise1234.cpp"
 $248 = $247 & 31; //@line 131 "libraries/noise1234/noise1234.cpp"
 $249 = ($248>>>0)<(24); //@line 132 "libraries/noise1234/noise1234.cpp"
 $250 = $249 ? $13 : $21; //@line 132 "libraries/noise1234/noise1234.cpp"
 $251 = ($248>>>0)<(16); //@line 133 "libraries/noise1234/noise1234.cpp"
 $252 = $251 ? $21 : $17; //@line 133 "libraries/noise1234/noise1234.cpp"
 $253 = ($248>>>0)<(8); //@line 134 "libraries/noise1234/noise1234.cpp"
 $254 = $253 ? $17 : $23; //@line 134 "libraries/noise1234/noise1234.cpp"
 $255 = $247 & 1; //@line 135 "libraries/noise1234/noise1234.cpp"
 $256 = ($255|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $257 = -$250; //@line 135 "libraries/noise1234/noise1234.cpp"
 $258 = $256 ? $257 : $250; //@line 135 "libraries/noise1234/noise1234.cpp"
 $259 = $247 & 2; //@line 135 "libraries/noise1234/noise1234.cpp"
 $260 = ($259|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $261 = -$252; //@line 135 "libraries/noise1234/noise1234.cpp"
 $262 = $260 ? $261 : $252; //@line 135 "libraries/noise1234/noise1234.cpp"
 $263 = $258 + $262; //@line 135 "libraries/noise1234/noise1234.cpp"
 $264 = $247 & 4; //@line 135 "libraries/noise1234/noise1234.cpp"
 $265 = ($264|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $266 = -$254; //@line 135 "libraries/noise1234/noise1234.cpp"
 $267 = $265 ? $266 : $254; //@line 135 "libraries/noise1234/noise1234.cpp"
 $268 = $267 + $263; //@line 135 "libraries/noise1234/noise1234.cpp"
 $269 = $268 - $239; //@line 408 "libraries/noise1234/noise1234.cpp"
 $270 = $42 * $269; //@line 408 "libraries/noise1234/noise1234.cpp"
 $271 = $239 + $270; //@line 408 "libraries/noise1234/noise1234.cpp"
 $272 = (($142) + ($27))|0; //@line 410 "libraries/noise1234/noise1234.cpp"
 $273 = (326027 + ($272)|0); //@line 410 "libraries/noise1234/noise1234.cpp"
 $274 = HEAP8[$273>>0]|0; //@line 410 "libraries/noise1234/noise1234.cpp"
 $275 = $274&255; //@line 410 "libraries/noise1234/noise1234.cpp"
 $276 = (($275) + ($32))|0; //@line 410 "libraries/noise1234/noise1234.cpp"
 $277 = (326027 + ($276)|0); //@line 410 "libraries/noise1234/noise1234.cpp"
 $278 = HEAP8[$277>>0]|0; //@line 410 "libraries/noise1234/noise1234.cpp"
 $279 = $278&255; //@line 410 "libraries/noise1234/noise1234.cpp"
 $280 = $279 & 31; //@line 131 "libraries/noise1234/noise1234.cpp"
 $281 = ($280>>>0)<(24); //@line 132 "libraries/noise1234/noise1234.cpp"
 $282 = $281 ? $13 : $21; //@line 132 "libraries/noise1234/noise1234.cpp"
 $283 = ($280>>>0)<(16); //@line 133 "libraries/noise1234/noise1234.cpp"
 $284 = $283 ? $21 : $22; //@line 133 "libraries/noise1234/noise1234.cpp"
 $285 = ($280>>>0)<(8); //@line 134 "libraries/noise1234/noise1234.cpp"
 $286 = $285 ? $22 : $19; //@line 134 "libraries/noise1234/noise1234.cpp"
 $287 = $279 & 1; //@line 135 "libraries/noise1234/noise1234.cpp"
 $288 = ($287|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $289 = -$282; //@line 135 "libraries/noise1234/noise1234.cpp"
 $290 = $288 ? $289 : $282; //@line 135 "libraries/noise1234/noise1234.cpp"
 $291 = $279 & 2; //@line 135 "libraries/noise1234/noise1234.cpp"
 $292 = ($291|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $293 = -$284; //@line 135 "libraries/noise1234/noise1234.cpp"
 $294 = $292 ? $293 : $284; //@line 135 "libraries/noise1234/noise1234.cpp"
 $295 = $290 + $294; //@line 135 "libraries/noise1234/noise1234.cpp"
 $296 = $279 & 4; //@line 135 "libraries/noise1234/noise1234.cpp"
 $297 = ($296|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $298 = -$286; //@line 135 "libraries/noise1234/noise1234.cpp"
 $299 = $297 ? $298 : $286; //@line 135 "libraries/noise1234/noise1234.cpp"
 $300 = $299 + $295; //@line 135 "libraries/noise1234/noise1234.cpp"
 $301 = (($175) + ($27))|0; //@line 411 "libraries/noise1234/noise1234.cpp"
 $302 = (326027 + ($301)|0); //@line 411 "libraries/noise1234/noise1234.cpp"
 $303 = HEAP8[$302>>0]|0; //@line 411 "libraries/noise1234/noise1234.cpp"
 $304 = $303&255; //@line 411 "libraries/noise1234/noise1234.cpp"
 $305 = (($304) + ($32))|0; //@line 411 "libraries/noise1234/noise1234.cpp"
 $306 = (326027 + ($305)|0); //@line 411 "libraries/noise1234/noise1234.cpp"
 $307 = HEAP8[$306>>0]|0; //@line 411 "libraries/noise1234/noise1234.cpp"
 $308 = $307&255; //@line 411 "libraries/noise1234/noise1234.cpp"
 $309 = $308 & 31; //@line 131 "libraries/noise1234/noise1234.cpp"
 $310 = ($309>>>0)<(24); //@line 132 "libraries/noise1234/noise1234.cpp"
 $311 = $310 ? $13 : $21; //@line 132 "libraries/noise1234/noise1234.cpp"
 $312 = ($309>>>0)<(16); //@line 133 "libraries/noise1234/noise1234.cpp"
 $313 = $312 ? $21 : $22; //@line 133 "libraries/noise1234/noise1234.cpp"
 $314 = ($309>>>0)<(8); //@line 134 "libraries/noise1234/noise1234.cpp"
 $315 = $314 ? $22 : $23; //@line 134 "libraries/noise1234/noise1234.cpp"
 $316 = $308 & 1; //@line 135 "libraries/noise1234/noise1234.cpp"
 $317 = ($316|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $318 = -$311; //@line 135 "libraries/noise1234/noise1234.cpp"
 $319 = $317 ? $318 : $311; //@line 135 "libraries/noise1234/noise1234.cpp"
 $320 = $308 & 2; //@line 135 "libraries/noise1234/noise1234.cpp"
 $321 = ($320|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $322 = -$313; //@line 135 "libraries/noise1234/noise1234.cpp"
 $323 = $321 ? $322 : $313; //@line 135 "libraries/noise1234/noise1234.cpp"
 $324 = $319 + $323; //@line 135 "libraries/noise1234/noise1234.cpp"
 $325 = $308 & 4; //@line 135 "libraries/noise1234/noise1234.cpp"
 $326 = ($325|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $327 = -$315; //@line 135 "libraries/noise1234/noise1234.cpp"
 $328 = $326 ? $327 : $315; //@line 135 "libraries/noise1234/noise1234.cpp"
 $329 = $328 + $324; //@line 135 "libraries/noise1234/noise1234.cpp"
 $330 = $329 - $300; //@line 412 "libraries/noise1234/noise1234.cpp"
 $331 = $42 * $330; //@line 412 "libraries/noise1234/noise1234.cpp"
 $332 = $300 + $331; //@line 412 "libraries/noise1234/noise1234.cpp"
 $333 = $332 - $271; //@line 414 "libraries/noise1234/noise1234.cpp"
 $334 = $49 * $333; //@line 414 "libraries/noise1234/noise1234.cpp"
 $335 = $271 + $334; //@line 414 "libraries/noise1234/noise1234.cpp"
 $336 = $335 - $210; //@line 416 "libraries/noise1234/noise1234.cpp"
 $337 = $56 * $336; //@line 416 "libraries/noise1234/noise1234.cpp"
 $338 = $210 + $337; //@line 416 "libraries/noise1234/noise1234.cpp"
 $339 = (($74) + ($25))|0; //@line 418 "libraries/noise1234/noise1234.cpp"
 $340 = (326027 + ($339)|0); //@line 418 "libraries/noise1234/noise1234.cpp"
 $341 = HEAP8[$340>>0]|0; //@line 418 "libraries/noise1234/noise1234.cpp"
 $342 = $341&255; //@line 418 "libraries/noise1234/noise1234.cpp"
 $343 = $342 & 31; //@line 131 "libraries/noise1234/noise1234.cpp"
 $344 = ($343>>>0)<(24); //@line 132 "libraries/noise1234/noise1234.cpp"
 $345 = $344 ? $20 : $15; //@line 132 "libraries/noise1234/noise1234.cpp"
 $346 = ($343>>>0)<(16); //@line 133 "libraries/noise1234/noise1234.cpp"
 $347 = $346 ? $15 : $17; //@line 133 "libraries/noise1234/noise1234.cpp"
 $348 = ($343>>>0)<(8); //@line 134 "libraries/noise1234/noise1234.cpp"
 $349 = $348 ? $17 : $19; //@line 134 "libraries/noise1234/noise1234.cpp"
 $350 = $342 & 1; //@line 135 "libraries/noise1234/noise1234.cpp"
 $351 = ($350|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $352 = -$345; //@line 135 "libraries/noise1234/noise1234.cpp"
 $353 = $351 ? $352 : $345; //@line 135 "libraries/noise1234/noise1234.cpp"
 $354 = $342 & 2; //@line 135 "libraries/noise1234/noise1234.cpp"
 $355 = ($354|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $356 = -$347; //@line 135 "libraries/noise1234/noise1234.cpp"
 $357 = $355 ? $356 : $347; //@line 135 "libraries/noise1234/noise1234.cpp"
 $358 = $353 + $357; //@line 135 "libraries/noise1234/noise1234.cpp"
 $359 = $342 & 4; //@line 135 "libraries/noise1234/noise1234.cpp"
 $360 = ($359|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $361 = -$349; //@line 135 "libraries/noise1234/noise1234.cpp"
 $362 = $360 ? $361 : $349; //@line 135 "libraries/noise1234/noise1234.cpp"
 $363 = $362 + $358; //@line 135 "libraries/noise1234/noise1234.cpp"
 $364 = (($110) + ($25))|0; //@line 419 "libraries/noise1234/noise1234.cpp"
 $365 = (326027 + ($364)|0); //@line 419 "libraries/noise1234/noise1234.cpp"
 $366 = HEAP8[$365>>0]|0; //@line 419 "libraries/noise1234/noise1234.cpp"
 $367 = $366&255; //@line 419 "libraries/noise1234/noise1234.cpp"
 $368 = $367 & 31; //@line 131 "libraries/noise1234/noise1234.cpp"
 $369 = ($368>>>0)<(24); //@line 132 "libraries/noise1234/noise1234.cpp"
 $370 = $369 ? $20 : $15; //@line 132 "libraries/noise1234/noise1234.cpp"
 $371 = ($368>>>0)<(16); //@line 133 "libraries/noise1234/noise1234.cpp"
 $372 = $371 ? $15 : $17; //@line 133 "libraries/noise1234/noise1234.cpp"
 $373 = ($368>>>0)<(8); //@line 134 "libraries/noise1234/noise1234.cpp"
 $374 = $373 ? $17 : $23; //@line 134 "libraries/noise1234/noise1234.cpp"
 $375 = $367 & 1; //@line 135 "libraries/noise1234/noise1234.cpp"
 $376 = ($375|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $377 = -$370; //@line 135 "libraries/noise1234/noise1234.cpp"
 $378 = $376 ? $377 : $370; //@line 135 "libraries/noise1234/noise1234.cpp"
 $379 = $367 & 2; //@line 135 "libraries/noise1234/noise1234.cpp"
 $380 = ($379|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $381 = -$372; //@line 135 "libraries/noise1234/noise1234.cpp"
 $382 = $380 ? $381 : $372; //@line 135 "libraries/noise1234/noise1234.cpp"
 $383 = $378 + $382; //@line 135 "libraries/noise1234/noise1234.cpp"
 $384 = $367 & 4; //@line 135 "libraries/noise1234/noise1234.cpp"
 $385 = ($384|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $386 = -$374; //@line 135 "libraries/noise1234/noise1234.cpp"
 $387 = $385 ? $386 : $374; //@line 135 "libraries/noise1234/noise1234.cpp"
 $388 = $387 + $383; //@line 135 "libraries/noise1234/noise1234.cpp"
 $389 = $388 - $363; //@line 420 "libraries/noise1234/noise1234.cpp"
 $390 = $42 * $389; //@line 420 "libraries/noise1234/noise1234.cpp"
 $391 = $363 + $390; //@line 420 "libraries/noise1234/noise1234.cpp"
 $392 = (($146) + ($25))|0; //@line 422 "libraries/noise1234/noise1234.cpp"
 $393 = (326027 + ($392)|0); //@line 422 "libraries/noise1234/noise1234.cpp"
 $394 = HEAP8[$393>>0]|0; //@line 422 "libraries/noise1234/noise1234.cpp"
 $395 = $394&255; //@line 422 "libraries/noise1234/noise1234.cpp"
 $396 = $395 & 31; //@line 131 "libraries/noise1234/noise1234.cpp"
 $397 = ($396>>>0)<(24); //@line 132 "libraries/noise1234/noise1234.cpp"
 $398 = $397 ? $20 : $15; //@line 132 "libraries/noise1234/noise1234.cpp"
 $399 = ($396>>>0)<(16); //@line 133 "libraries/noise1234/noise1234.cpp"
 $400 = $399 ? $15 : $22; //@line 133 "libraries/noise1234/noise1234.cpp"
 $401 = ($396>>>0)<(8); //@line 134 "libraries/noise1234/noise1234.cpp"
 $402 = $401 ? $22 : $19; //@line 134 "libraries/noise1234/noise1234.cpp"
 $403 = $395 & 1; //@line 135 "libraries/noise1234/noise1234.cpp"
 $404 = ($403|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $405 = -$398; //@line 135 "libraries/noise1234/noise1234.cpp"
 $406 = $404 ? $405 : $398; //@line 135 "libraries/noise1234/noise1234.cpp"
 $407 = $395 & 2; //@line 135 "libraries/noise1234/noise1234.cpp"
 $408 = ($407|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $409 = -$400; //@line 135 "libraries/noise1234/noise1234.cpp"
 $410 = $408 ? $409 : $400; //@line 135 "libraries/noise1234/noise1234.cpp"
 $411 = $406 + $410; //@line 135 "libraries/noise1234/noise1234.cpp"
 $412 = $395 & 4; //@line 135 "libraries/noise1234/noise1234.cpp"
 $413 = ($412|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $414 = -$402; //@line 135 "libraries/noise1234/noise1234.cpp"
 $415 = $413 ? $414 : $402; //@line 135 "libraries/noise1234/noise1234.cpp"
 $416 = $415 + $411; //@line 135 "libraries/noise1234/noise1234.cpp"
 $417 = (($179) + ($25))|0; //@line 423 "libraries/noise1234/noise1234.cpp"
 $418 = (326027 + ($417)|0); //@line 423 "libraries/noise1234/noise1234.cpp"
 $419 = HEAP8[$418>>0]|0; //@line 423 "libraries/noise1234/noise1234.cpp"
 $420 = $419&255; //@line 423 "libraries/noise1234/noise1234.cpp"
 $421 = $420 & 31; //@line 131 "libraries/noise1234/noise1234.cpp"
 $422 = ($421>>>0)<(24); //@line 132 "libraries/noise1234/noise1234.cpp"
 $423 = $422 ? $20 : $15; //@line 132 "libraries/noise1234/noise1234.cpp"
 $424 = ($421>>>0)<(16); //@line 133 "libraries/noise1234/noise1234.cpp"
 $425 = $424 ? $15 : $22; //@line 133 "libraries/noise1234/noise1234.cpp"
 $426 = ($421>>>0)<(8); //@line 134 "libraries/noise1234/noise1234.cpp"
 $427 = $426 ? $22 : $23; //@line 134 "libraries/noise1234/noise1234.cpp"
 $428 = $420 & 1; //@line 135 "libraries/noise1234/noise1234.cpp"
 $429 = ($428|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $430 = -$423; //@line 135 "libraries/noise1234/noise1234.cpp"
 $431 = $429 ? $430 : $423; //@line 135 "libraries/noise1234/noise1234.cpp"
 $432 = $420 & 2; //@line 135 "libraries/noise1234/noise1234.cpp"
 $433 = ($432|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $434 = -$425; //@line 135 "libraries/noise1234/noise1234.cpp"
 $435 = $433 ? $434 : $425; //@line 135 "libraries/noise1234/noise1234.cpp"
 $436 = $431 + $435; //@line 135 "libraries/noise1234/noise1234.cpp"
 $437 = $420 & 4; //@line 135 "libraries/noise1234/noise1234.cpp"
 $438 = ($437|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $439 = -$427; //@line 135 "libraries/noise1234/noise1234.cpp"
 $440 = $438 ? $439 : $427; //@line 135 "libraries/noise1234/noise1234.cpp"
 $441 = $440 + $436; //@line 135 "libraries/noise1234/noise1234.cpp"
 $442 = $441 - $416; //@line 424 "libraries/noise1234/noise1234.cpp"
 $443 = $42 * $442; //@line 424 "libraries/noise1234/noise1234.cpp"
 $444 = $416 + $443; //@line 424 "libraries/noise1234/noise1234.cpp"
 $445 = $444 - $391; //@line 426 "libraries/noise1234/noise1234.cpp"
 $446 = $49 * $445; //@line 426 "libraries/noise1234/noise1234.cpp"
 $447 = $391 + $446; //@line 426 "libraries/noise1234/noise1234.cpp"
 $448 = (($214) + ($25))|0; //@line 428 "libraries/noise1234/noise1234.cpp"
 $449 = (326027 + ($448)|0); //@line 428 "libraries/noise1234/noise1234.cpp"
 $450 = HEAP8[$449>>0]|0; //@line 428 "libraries/noise1234/noise1234.cpp"
 $451 = $450&255; //@line 428 "libraries/noise1234/noise1234.cpp"
 $452 = $451 & 31; //@line 131 "libraries/noise1234/noise1234.cpp"
 $453 = ($452>>>0)<(24); //@line 132 "libraries/noise1234/noise1234.cpp"
 $454 = $453 ? $20 : $21; //@line 132 "libraries/noise1234/noise1234.cpp"
 $455 = ($452>>>0)<(16); //@line 133 "libraries/noise1234/noise1234.cpp"
 $456 = $455 ? $21 : $17; //@line 133 "libraries/noise1234/noise1234.cpp"
 $457 = ($452>>>0)<(8); //@line 134 "libraries/noise1234/noise1234.cpp"
 $458 = $457 ? $17 : $19; //@line 134 "libraries/noise1234/noise1234.cpp"
 $459 = $451 & 1; //@line 135 "libraries/noise1234/noise1234.cpp"
 $460 = ($459|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $461 = -$454; //@line 135 "libraries/noise1234/noise1234.cpp"
 $462 = $460 ? $461 : $454; //@line 135 "libraries/noise1234/noise1234.cpp"
 $463 = $451 & 2; //@line 135 "libraries/noise1234/noise1234.cpp"
 $464 = ($463|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $465 = -$456; //@line 135 "libraries/noise1234/noise1234.cpp"
 $466 = $464 ? $465 : $456; //@line 135 "libraries/noise1234/noise1234.cpp"
 $467 = $462 + $466; //@line 135 "libraries/noise1234/noise1234.cpp"
 $468 = $451 & 4; //@line 135 "libraries/noise1234/noise1234.cpp"
 $469 = ($468|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $470 = -$458; //@line 135 "libraries/noise1234/noise1234.cpp"
 $471 = $469 ? $470 : $458; //@line 135 "libraries/noise1234/noise1234.cpp"
 $472 = $471 + $467; //@line 135 "libraries/noise1234/noise1234.cpp"
 $473 = (($243) + ($25))|0; //@line 429 "libraries/noise1234/noise1234.cpp"
 $474 = (326027 + ($473)|0); //@line 429 "libraries/noise1234/noise1234.cpp"
 $475 = HEAP8[$474>>0]|0; //@line 429 "libraries/noise1234/noise1234.cpp"
 $476 = $475&255; //@line 429 "libraries/noise1234/noise1234.cpp"
 $477 = $476 & 31; //@line 131 "libraries/noise1234/noise1234.cpp"
 $478 = ($477>>>0)<(24); //@line 132 "libraries/noise1234/noise1234.cpp"
 $479 = $478 ? $20 : $21; //@line 132 "libraries/noise1234/noise1234.cpp"
 $480 = ($477>>>0)<(16); //@line 133 "libraries/noise1234/noise1234.cpp"
 $481 = $480 ? $21 : $17; //@line 133 "libraries/noise1234/noise1234.cpp"
 $482 = ($477>>>0)<(8); //@line 134 "libraries/noise1234/noise1234.cpp"
 $483 = $482 ? $17 : $23; //@line 134 "libraries/noise1234/noise1234.cpp"
 $484 = $476 & 1; //@line 135 "libraries/noise1234/noise1234.cpp"
 $485 = ($484|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $486 = -$479; //@line 135 "libraries/noise1234/noise1234.cpp"
 $487 = $485 ? $486 : $479; //@line 135 "libraries/noise1234/noise1234.cpp"
 $488 = $476 & 2; //@line 135 "libraries/noise1234/noise1234.cpp"
 $489 = ($488|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $490 = -$481; //@line 135 "libraries/noise1234/noise1234.cpp"
 $491 = $489 ? $490 : $481; //@line 135 "libraries/noise1234/noise1234.cpp"
 $492 = $487 + $491; //@line 135 "libraries/noise1234/noise1234.cpp"
 $493 = $476 & 4; //@line 135 "libraries/noise1234/noise1234.cpp"
 $494 = ($493|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $495 = -$483; //@line 135 "libraries/noise1234/noise1234.cpp"
 $496 = $494 ? $495 : $483; //@line 135 "libraries/noise1234/noise1234.cpp"
 $497 = $496 + $492; //@line 135 "libraries/noise1234/noise1234.cpp"
 $498 = $497 - $472; //@line 430 "libraries/noise1234/noise1234.cpp"
 $499 = $42 * $498; //@line 430 "libraries/noise1234/noise1234.cpp"
 $500 = $472 + $499; //@line 430 "libraries/noise1234/noise1234.cpp"
 $501 = (($275) + ($25))|0; //@line 432 "libraries/noise1234/noise1234.cpp"
 $502 = (326027 + ($501)|0); //@line 432 "libraries/noise1234/noise1234.cpp"
 $503 = HEAP8[$502>>0]|0; //@line 432 "libraries/noise1234/noise1234.cpp"
 $504 = $503&255; //@line 432 "libraries/noise1234/noise1234.cpp"
 $505 = $504 & 31; //@line 131 "libraries/noise1234/noise1234.cpp"
 $506 = ($505>>>0)<(24); //@line 132 "libraries/noise1234/noise1234.cpp"
 $507 = $506 ? $20 : $21; //@line 132 "libraries/noise1234/noise1234.cpp"
 $508 = ($505>>>0)<(16); //@line 133 "libraries/noise1234/noise1234.cpp"
 $509 = $508 ? $21 : $22; //@line 133 "libraries/noise1234/noise1234.cpp"
 $510 = ($505>>>0)<(8); //@line 134 "libraries/noise1234/noise1234.cpp"
 $511 = $510 ? $22 : $19; //@line 134 "libraries/noise1234/noise1234.cpp"
 $512 = $504 & 1; //@line 135 "libraries/noise1234/noise1234.cpp"
 $513 = ($512|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $514 = -$507; //@line 135 "libraries/noise1234/noise1234.cpp"
 $515 = $513 ? $514 : $507; //@line 135 "libraries/noise1234/noise1234.cpp"
 $516 = $504 & 2; //@line 135 "libraries/noise1234/noise1234.cpp"
 $517 = ($516|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $518 = -$509; //@line 135 "libraries/noise1234/noise1234.cpp"
 $519 = $517 ? $518 : $509; //@line 135 "libraries/noise1234/noise1234.cpp"
 $520 = $515 + $519; //@line 135 "libraries/noise1234/noise1234.cpp"
 $521 = $504 & 4; //@line 135 "libraries/noise1234/noise1234.cpp"
 $522 = ($521|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $523 = -$511; //@line 135 "libraries/noise1234/noise1234.cpp"
 $524 = $522 ? $523 : $511; //@line 135 "libraries/noise1234/noise1234.cpp"
 $525 = $524 + $520; //@line 135 "libraries/noise1234/noise1234.cpp"
 $526 = (($304) + ($25))|0; //@line 433 "libraries/noise1234/noise1234.cpp"
 $527 = (326027 + ($526)|0); //@line 433 "libraries/noise1234/noise1234.cpp"
 $528 = HEAP8[$527>>0]|0; //@line 433 "libraries/noise1234/noise1234.cpp"
 $529 = $528&255; //@line 433 "libraries/noise1234/noise1234.cpp"
 $530 = $529 & 31; //@line 131 "libraries/noise1234/noise1234.cpp"
 $531 = ($530>>>0)<(24); //@line 132 "libraries/noise1234/noise1234.cpp"
 $532 = $531 ? $20 : $21; //@line 132 "libraries/noise1234/noise1234.cpp"
 $533 = ($530>>>0)<(16); //@line 133 "libraries/noise1234/noise1234.cpp"
 $534 = $533 ? $21 : $22; //@line 133 "libraries/noise1234/noise1234.cpp"
 $535 = ($530>>>0)<(8); //@line 134 "libraries/noise1234/noise1234.cpp"
 $536 = $535 ? $22 : $23; //@line 134 "libraries/noise1234/noise1234.cpp"
 $537 = $529 & 1; //@line 135 "libraries/noise1234/noise1234.cpp"
 $538 = ($537|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $539 = -$532; //@line 135 "libraries/noise1234/noise1234.cpp"
 $540 = $538 ? $539 : $532; //@line 135 "libraries/noise1234/noise1234.cpp"
 $541 = $529 & 2; //@line 135 "libraries/noise1234/noise1234.cpp"
 $542 = ($541|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $543 = -$534; //@line 135 "libraries/noise1234/noise1234.cpp"
 $544 = $542 ? $543 : $534; //@line 135 "libraries/noise1234/noise1234.cpp"
 $545 = $540 + $544; //@line 135 "libraries/noise1234/noise1234.cpp"
 $546 = $529 & 4; //@line 135 "libraries/noise1234/noise1234.cpp"
 $547 = ($546|0)!=(0); //@line 135 "libraries/noise1234/noise1234.cpp"
 $548 = -$536; //@line 135 "libraries/noise1234/noise1234.cpp"
 $549 = $547 ? $548 : $536; //@line 135 "libraries/noise1234/noise1234.cpp"
 $550 = $549 + $545; //@line 135 "libraries/noise1234/noise1234.cpp"
 $551 = $550 - $525; //@line 434 "libraries/noise1234/noise1234.cpp"
 $552 = $42 * $551; //@line 434 "libraries/noise1234/noise1234.cpp"
 $553 = $525 + $552; //@line 434 "libraries/noise1234/noise1234.cpp"
 $554 = $553 - $500; //@line 436 "libraries/noise1234/noise1234.cpp"
 $555 = $49 * $554; //@line 436 "libraries/noise1234/noise1234.cpp"
 $556 = $500 + $555; //@line 436 "libraries/noise1234/noise1234.cpp"
 $557 = $556 - $447; //@line 438 "libraries/noise1234/noise1234.cpp"
 $558 = $56 * $557; //@line 438 "libraries/noise1234/noise1234.cpp"
 $559 = $447 + $558; //@line 438 "libraries/noise1234/noise1234.cpp"
 $560 = $559 - $338; //@line 440 "libraries/noise1234/noise1234.cpp"
 $561 = $63 * $560; //@line 440 "libraries/noise1234/noise1234.cpp"
 $562 = $338 + $561; //@line 440 "libraries/noise1234/noise1234.cpp"
 $563 = $562 * 0.87000000476837158; //@line 440 "libraries/noise1234/noise1234.cpp"
 return (+$563); //@line 440 "libraries/noise1234/noise1234.cpp"
}
function __ZN16SimplexNoise12345noiseEf($x) {
 $x = +$x;
 var $$$i = 0.0, $$$i2 = 0.0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0.0;
 var $25 = 0.0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0;
 var $9 = 0.0, $addconv$i = 0, $addconv$i1 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (~~(($x))); //@line 123 "libraries/noise1234/simplexnoise1234.cpp"
 $not$ = !($x > 0.0); //@line 123 "libraries/noise1234/simplexnoise1234.cpp"
 $1 = $not$ << 31 >> 31; //@line 123 "libraries/noise1234/simplexnoise1234.cpp"
 $2 = (($1) + ($0))|0; //@line 123 "libraries/noise1234/simplexnoise1234.cpp"
 $3 = (($2) + 1)|0; //@line 124 "libraries/noise1234/simplexnoise1234.cpp"
 $4 = (+($2|0)); //@line 125 "libraries/noise1234/simplexnoise1234.cpp"
 $5 = $x - $4; //@line 125 "libraries/noise1234/simplexnoise1234.cpp"
 $6 = $5 + -1.0; //@line 126 "libraries/noise1234/simplexnoise1234.cpp"
 $7 = $5 * $5; //@line 130 "libraries/noise1234/simplexnoise1234.cpp"
 $8 = 1.0 - $7; //@line 130 "libraries/noise1234/simplexnoise1234.cpp"
 $9 = $8 * $8; //@line 132 "libraries/noise1234/simplexnoise1234.cpp"
 $10 = $9 * $9; //@line 133 "libraries/noise1234/simplexnoise1234.cpp"
 $11 = $2 & 255; //@line 133 "libraries/noise1234/simplexnoise1234.cpp"
 $12 = (326539 + ($11)|0); //@line 133 "libraries/noise1234/simplexnoise1234.cpp"
 $13 = HEAP8[$12>>0]|0; //@line 133 "libraries/noise1234/simplexnoise1234.cpp"
 $14 = $13&255; //@line 133 "libraries/noise1234/simplexnoise1234.cpp"
 $15 = $14 & 7; //@line 108 "libraries/noise1234/simplexnoise1234.cpp"
 $addconv$i = (($15) + 1)|0; //@line 108 "libraries/noise1234/simplexnoise1234.cpp"
 $16 = (+($addconv$i|0)); //@line 108 "libraries/noise1234/simplexnoise1234.cpp"
 $17 = $14 & 8; //@line 109 "libraries/noise1234/simplexnoise1234.cpp"
 $18 = ($17|0)==(0); //@line 109 "libraries/noise1234/simplexnoise1234.cpp"
 $19 = -$16; //@line 109 "libraries/noise1234/simplexnoise1234.cpp"
 $$$i = $18 ? $16 : $19; //@line 109 "libraries/noise1234/simplexnoise1234.cpp"
 $20 = $5 * $$$i; //@line 110 "libraries/noise1234/simplexnoise1234.cpp"
 $21 = $10 * $20; //@line 133 "libraries/noise1234/simplexnoise1234.cpp"
 $22 = $6 * $6; //@line 135 "libraries/noise1234/simplexnoise1234.cpp"
 $23 = 1.0 - $22; //@line 135 "libraries/noise1234/simplexnoise1234.cpp"
 $24 = $23 * $23; //@line 137 "libraries/noise1234/simplexnoise1234.cpp"
 $25 = $24 * $24; //@line 138 "libraries/noise1234/simplexnoise1234.cpp"
 $26 = $3 & 255; //@line 138 "libraries/noise1234/simplexnoise1234.cpp"
 $27 = (326539 + ($26)|0); //@line 138 "libraries/noise1234/simplexnoise1234.cpp"
 $28 = HEAP8[$27>>0]|0; //@line 138 "libraries/noise1234/simplexnoise1234.cpp"
 $29 = $28&255; //@line 138 "libraries/noise1234/simplexnoise1234.cpp"
 $30 = $29 & 7; //@line 108 "libraries/noise1234/simplexnoise1234.cpp"
 $addconv$i1 = (($30) + 1)|0; //@line 108 "libraries/noise1234/simplexnoise1234.cpp"
 $31 = (+($addconv$i1|0)); //@line 108 "libraries/noise1234/simplexnoise1234.cpp"
 $32 = $29 & 8; //@line 109 "libraries/noise1234/simplexnoise1234.cpp"
 $33 = ($32|0)==(0); //@line 109 "libraries/noise1234/simplexnoise1234.cpp"
 $34 = -$31; //@line 109 "libraries/noise1234/simplexnoise1234.cpp"
 $$$i2 = $33 ? $31 : $34; //@line 109 "libraries/noise1234/simplexnoise1234.cpp"
 $35 = $6 * $$$i2; //@line 110 "libraries/noise1234/simplexnoise1234.cpp"
 $36 = $25 * $35; //@line 138 "libraries/noise1234/simplexnoise1234.cpp"
 $37 = $21 + $36; //@line 141 "libraries/noise1234/simplexnoise1234.cpp"
 $38 = $37 * 0.39500001072883606; //@line 141 "libraries/noise1234/simplexnoise1234.cpp"
 return (+$38); //@line 141 "libraries/noise1234/simplexnoise1234.cpp"
}
function __ZN16SimplexNoise12345noiseEff($x,$y) {
 $x = +$x;
 $y = +$y;
 var $$ = 0, $$v$i = 0.0, $$v$i3 = 0.0, $$v$i4 = 0.0, $0 = 0.0, $1 = 0.0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0, $106 = 0.0, $107 = 0.0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0, $123 = 0.0, $124 = 0.0, $125 = 0, $126 = 0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0.0;
 var $130 = 0.0, $131 = 0.0, $132 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0, $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0.0;
 var $3 = 0.0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0.0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0, $45 = 0.0, $46 = 0.0, $47 = 0.0;
 var $48 = 0.0, $49 = 0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0.0, $63 = 0, $64 = 0, $65 = 0.0;
 var $66 = 0.0, $67 = 0, $68 = 0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0.0, $78 = 0.0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0, $93 = 0, $94 = 0.0, $95 = 0.0, $96 = 0, $97 = 0, $98 = 0.0, $99 = 0.0, $n0$0 = 0.0, $n1$0 = 0.0, $n2$0 = 0.0;
 var $not$ = 0, $not$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $x + $y; //@line 154 "libraries/noise1234/simplexnoise1234.cpp"
 $1 = $0; //@line 154 "libraries/noise1234/simplexnoise1234.cpp"
 $2 = $1 * 0.36602540300000003; //@line 154 "libraries/noise1234/simplexnoise1234.cpp"
 $3 = $2; //@line 154 "libraries/noise1234/simplexnoise1234.cpp"
 $4 = $3 + $x; //@line 155 "libraries/noise1234/simplexnoise1234.cpp"
 $5 = $3 + $y; //@line 156 "libraries/noise1234/simplexnoise1234.cpp"
 $6 = (~~(($4))); //@line 157 "libraries/noise1234/simplexnoise1234.cpp"
 $not$ = !($4 > 0.0); //@line 157 "libraries/noise1234/simplexnoise1234.cpp"
 $7 = $not$ << 31 >> 31; //@line 157 "libraries/noise1234/simplexnoise1234.cpp"
 $8 = (($7) + ($6))|0; //@line 157 "libraries/noise1234/simplexnoise1234.cpp"
 $9 = (~~(($5))); //@line 158 "libraries/noise1234/simplexnoise1234.cpp"
 $not$1 = !($5 > 0.0); //@line 158 "libraries/noise1234/simplexnoise1234.cpp"
 $10 = $not$1 << 31 >> 31; //@line 158 "libraries/noise1234/simplexnoise1234.cpp"
 $11 = (($10) + ($9))|0; //@line 158 "libraries/noise1234/simplexnoise1234.cpp"
 $12 = (($8) + ($11))|0; //@line 160 "libraries/noise1234/simplexnoise1234.cpp"
 $13 = (+($12|0)); //@line 160 "libraries/noise1234/simplexnoise1234.cpp"
 $14 = $13; //@line 160 "libraries/noise1234/simplexnoise1234.cpp"
 $15 = $14 * 0.211324865; //@line 160 "libraries/noise1234/simplexnoise1234.cpp"
 $16 = $15; //@line 160 "libraries/noise1234/simplexnoise1234.cpp"
 $17 = (+($8|0)); //@line 161 "libraries/noise1234/simplexnoise1234.cpp"
 $18 = $17 - $16; //@line 161 "libraries/noise1234/simplexnoise1234.cpp"
 $19 = (+($11|0)); //@line 162 "libraries/noise1234/simplexnoise1234.cpp"
 $20 = $19 - $16; //@line 162 "libraries/noise1234/simplexnoise1234.cpp"
 $21 = $x - $18; //@line 163 "libraries/noise1234/simplexnoise1234.cpp"
 $22 = $y - $20; //@line 164 "libraries/noise1234/simplexnoise1234.cpp"
 $23 = $21 > $22; //@line 169 "libraries/noise1234/simplexnoise1234.cpp"
 $24 = $23&1; //@line 169 "libraries/noise1234/simplexnoise1234.cpp"
 $$ = $24 ^ 1; //@line 169 "libraries/noise1234/simplexnoise1234.cpp"
 $25 = (+($23&1)); //@line 176 "libraries/noise1234/simplexnoise1234.cpp"
 $26 = $21 - $25; //@line 176 "libraries/noise1234/simplexnoise1234.cpp"
 $27 = $26; //@line 176 "libraries/noise1234/simplexnoise1234.cpp"
 $28 = $27 + 0.211324865; //@line 176 "libraries/noise1234/simplexnoise1234.cpp"
 $29 = $28; //@line 176 "libraries/noise1234/simplexnoise1234.cpp"
 $30 = (+($$|0)); //@line 177 "libraries/noise1234/simplexnoise1234.cpp"
 $31 = $22 - $30; //@line 177 "libraries/noise1234/simplexnoise1234.cpp"
 $32 = $31; //@line 177 "libraries/noise1234/simplexnoise1234.cpp"
 $33 = $32 + 0.211324865; //@line 177 "libraries/noise1234/simplexnoise1234.cpp"
 $34 = $33; //@line 177 "libraries/noise1234/simplexnoise1234.cpp"
 $35 = $21 + -1.0; //@line 178 "libraries/noise1234/simplexnoise1234.cpp"
 $36 = $35; //@line 178 "libraries/noise1234/simplexnoise1234.cpp"
 $37 = $36 + 0.42264973; //@line 178 "libraries/noise1234/simplexnoise1234.cpp"
 $38 = $37; //@line 178 "libraries/noise1234/simplexnoise1234.cpp"
 $39 = $22 + -1.0; //@line 179 "libraries/noise1234/simplexnoise1234.cpp"
 $40 = $39; //@line 179 "libraries/noise1234/simplexnoise1234.cpp"
 $41 = $40 + 0.42264973; //@line 179 "libraries/noise1234/simplexnoise1234.cpp"
 $42 = $41; //@line 179 "libraries/noise1234/simplexnoise1234.cpp"
 $43 = $8 & 255; //@line 182 "libraries/noise1234/simplexnoise1234.cpp"
 $44 = $11 & 255; //@line 183 "libraries/noise1234/simplexnoise1234.cpp"
 $45 = $21 * $21; //@line 186 "libraries/noise1234/simplexnoise1234.cpp"
 $46 = 0.5 - $45; //@line 186 "libraries/noise1234/simplexnoise1234.cpp"
 $47 = $22 * $22; //@line 186 "libraries/noise1234/simplexnoise1234.cpp"
 $48 = $46 - $47; //@line 186 "libraries/noise1234/simplexnoise1234.cpp"
 $49 = $48 < 0.0; //@line 187 "libraries/noise1234/simplexnoise1234.cpp"
 if ($49) {
  $n0$0 = 0.0;
 } else {
  $50 = $48 * $48; //@line 189 "libraries/noise1234/simplexnoise1234.cpp"
  $51 = $50 * $50; //@line 190 "libraries/noise1234/simplexnoise1234.cpp"
  $52 = (326539 + ($44)|0); //@line 190 "libraries/noise1234/simplexnoise1234.cpp"
  $53 = HEAP8[$52>>0]|0; //@line 190 "libraries/noise1234/simplexnoise1234.cpp"
  $54 = $53&255; //@line 190 "libraries/noise1234/simplexnoise1234.cpp"
  $55 = (($54) + ($43))|0; //@line 190 "libraries/noise1234/simplexnoise1234.cpp"
  $56 = (326539 + ($55)|0); //@line 190 "libraries/noise1234/simplexnoise1234.cpp"
  $57 = HEAP8[$56>>0]|0; //@line 190 "libraries/noise1234/simplexnoise1234.cpp"
  $58 = $57&255; //@line 190 "libraries/noise1234/simplexnoise1234.cpp"
  $59 = $58 & 4; //@line 114 "libraries/noise1234/simplexnoise1234.cpp"
  $60 = ($59>>>0)<(4); //@line 115 "libraries/noise1234/simplexnoise1234.cpp"
  $61 = $60 ? $21 : $22; //@line 115 "libraries/noise1234/simplexnoise1234.cpp"
  $62 = $60 ? $22 : $21; //@line 116 "libraries/noise1234/simplexnoise1234.cpp"
  $63 = $58 & 1; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $64 = ($63|0)!=(0); //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $65 = -$61; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $66 = $64 ? $65 : $61; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $67 = $58 & 2; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $68 = ($67|0)!=(0); //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $$v$i = $68 ? -2.0 : 2.0; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $69 = $62 * $$v$i; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $70 = $66 + $69; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $71 = $51 * $70; //@line 190 "libraries/noise1234/simplexnoise1234.cpp"
  $n0$0 = $71;
 }
 $72 = $29 * $29; //@line 193 "libraries/noise1234/simplexnoise1234.cpp"
 $73 = 0.5 - $72; //@line 193 "libraries/noise1234/simplexnoise1234.cpp"
 $74 = $34 * $34; //@line 193 "libraries/noise1234/simplexnoise1234.cpp"
 $75 = $73 - $74; //@line 193 "libraries/noise1234/simplexnoise1234.cpp"
 $76 = $75 < 0.0; //@line 194 "libraries/noise1234/simplexnoise1234.cpp"
 if ($76) {
  $n1$0 = 0.0;
 } else {
  $77 = $75 * $75; //@line 196 "libraries/noise1234/simplexnoise1234.cpp"
  $78 = $77 * $77; //@line 197 "libraries/noise1234/simplexnoise1234.cpp"
  $79 = (($24) + ($43))|0; //@line 197 "libraries/noise1234/simplexnoise1234.cpp"
  $80 = (($$) + ($44))|0; //@line 197 "libraries/noise1234/simplexnoise1234.cpp"
  $81 = (326539 + ($80)|0); //@line 197 "libraries/noise1234/simplexnoise1234.cpp"
  $82 = HEAP8[$81>>0]|0; //@line 197 "libraries/noise1234/simplexnoise1234.cpp"
  $83 = $82&255; //@line 197 "libraries/noise1234/simplexnoise1234.cpp"
  $84 = (($79) + ($83))|0; //@line 197 "libraries/noise1234/simplexnoise1234.cpp"
  $85 = (326539 + ($84)|0); //@line 197 "libraries/noise1234/simplexnoise1234.cpp"
  $86 = HEAP8[$85>>0]|0; //@line 197 "libraries/noise1234/simplexnoise1234.cpp"
  $87 = $86&255; //@line 197 "libraries/noise1234/simplexnoise1234.cpp"
  $88 = $87 & 4; //@line 114 "libraries/noise1234/simplexnoise1234.cpp"
  $89 = ($88>>>0)<(4); //@line 115 "libraries/noise1234/simplexnoise1234.cpp"
  $90 = $89 ? $29 : $34; //@line 115 "libraries/noise1234/simplexnoise1234.cpp"
  $91 = $89 ? $34 : $29; //@line 116 "libraries/noise1234/simplexnoise1234.cpp"
  $92 = $87 & 1; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $93 = ($92|0)!=(0); //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $94 = -$90; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $95 = $93 ? $94 : $90; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $96 = $87 & 2; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $97 = ($96|0)!=(0); //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $$v$i4 = $97 ? -2.0 : 2.0; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $98 = $91 * $$v$i4; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $99 = $95 + $98; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
  $100 = $78 * $99; //@line 197 "libraries/noise1234/simplexnoise1234.cpp"
  $n1$0 = $100;
 }
 $101 = $38 * $38; //@line 200 "libraries/noise1234/simplexnoise1234.cpp"
 $102 = 0.5 - $101; //@line 200 "libraries/noise1234/simplexnoise1234.cpp"
 $103 = $42 * $42; //@line 200 "libraries/noise1234/simplexnoise1234.cpp"
 $104 = $102 - $103; //@line 200 "libraries/noise1234/simplexnoise1234.cpp"
 $105 = $104 < 0.0; //@line 201 "libraries/noise1234/simplexnoise1234.cpp"
 if ($105) {
  $n2$0 = 0.0;
  $130 = $n0$0 + $n1$0; //@line 209 "libraries/noise1234/simplexnoise1234.cpp"
  $131 = $130 + $n2$0; //@line 209 "libraries/noise1234/simplexnoise1234.cpp"
  $132 = $131 * 45.229999542236328; //@line 209 "libraries/noise1234/simplexnoise1234.cpp"
  return (+$132); //@line 209 "libraries/noise1234/simplexnoise1234.cpp"
 }
 $106 = $104 * $104; //@line 203 "libraries/noise1234/simplexnoise1234.cpp"
 $107 = $106 * $106; //@line 204 "libraries/noise1234/simplexnoise1234.cpp"
 $108 = (($43) + 1)|0; //@line 204 "libraries/noise1234/simplexnoise1234.cpp"
 $109 = (($44) + 1)|0; //@line 204 "libraries/noise1234/simplexnoise1234.cpp"
 $110 = (326539 + ($109)|0); //@line 204 "libraries/noise1234/simplexnoise1234.cpp"
 $111 = HEAP8[$110>>0]|0; //@line 204 "libraries/noise1234/simplexnoise1234.cpp"
 $112 = $111&255; //@line 204 "libraries/noise1234/simplexnoise1234.cpp"
 $113 = (($108) + ($112))|0; //@line 204 "libraries/noise1234/simplexnoise1234.cpp"
 $114 = (326539 + ($113)|0); //@line 204 "libraries/noise1234/simplexnoise1234.cpp"
 $115 = HEAP8[$114>>0]|0; //@line 204 "libraries/noise1234/simplexnoise1234.cpp"
 $116 = $115&255; //@line 204 "libraries/noise1234/simplexnoise1234.cpp"
 $117 = $116 & 4; //@line 114 "libraries/noise1234/simplexnoise1234.cpp"
 $118 = ($117>>>0)<(4); //@line 115 "libraries/noise1234/simplexnoise1234.cpp"
 $119 = $118 ? $38 : $42; //@line 115 "libraries/noise1234/simplexnoise1234.cpp"
 $120 = $118 ? $42 : $38; //@line 116 "libraries/noise1234/simplexnoise1234.cpp"
 $121 = $116 & 1; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
 $122 = ($121|0)!=(0); //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
 $123 = -$119; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
 $124 = $122 ? $123 : $119; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
 $125 = $116 & 2; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
 $126 = ($125|0)!=(0); //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
 $$v$i3 = $126 ? -2.0 : 2.0; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
 $127 = $120 * $$v$i3; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
 $128 = $124 + $127; //@line 117 "libraries/noise1234/simplexnoise1234.cpp"
 $129 = $107 * $128; //@line 204 "libraries/noise1234/simplexnoise1234.cpp"
 $n2$0 = $129;
 $130 = $n0$0 + $n1$0; //@line 209 "libraries/noise1234/simplexnoise1234.cpp"
 $131 = $130 + $n2$0; //@line 209 "libraries/noise1234/simplexnoise1234.cpp"
 $132 = $131 * 45.229999542236328; //@line 209 "libraries/noise1234/simplexnoise1234.cpp"
 return (+$132); //@line 209 "libraries/noise1234/simplexnoise1234.cpp"
}
function __ZN4love9ExceptionD2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (12260); //@line 64 "common/Exception.cpp"
 $0 = ((($this)) + 4|0); //@line 65 "common/Exception.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 65 "common/Exception.cpp"
 return; //@line 65 "common/Exception.cpp"
}
function __ZN4love9ExceptionD0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (12260); //@line 64 "common/Exception.cpp"
 $0 = ((($this)) + 4|0); //@line 65 "common/Exception.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 65 "common/Exception.cpp"
 __ZdlPv($this); //@line 64 "common/Exception.cpp"
 return; //@line 65 "common/Exception.cpp"
}
function __ZNK4love9Exception4whatEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 56 "common/Exception.h"
 $1 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = $1 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ($2<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($3) {
  $6 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = $6;
  return ($7|0); //@line 56 "common/Exception.h"
 } else {
  $4 = ((($this)) + 12|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $5 = HEAP32[$4>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = $5;
  return ($7|0); //@line 56 "common/Exception.h"
 }
 return (0)|0;
}
function __ZN4love9ExceptionC2EPKcz($this,$fmt,$varargs) {
 $this = $this|0;
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $args = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $or$cond = 0, $or$cond3 = 0, $size_buffer$0 = 0;
 var $size_buffer$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $args = sp + 16|0;
 $0 = sp;
 HEAP32[$this>>2] = (12260); //@line 30 "common/Exception.cpp"
 $1 = ((($this)) + 4|0); //@line 29 "common/Exception.cpp"
 ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $size_buffer$0 = 256;
 while(1) {
  $2 = ($size_buffer$0|0)>(-1); //@line 36 "common/Exception.cpp"
  $3 = $2 ? $size_buffer$0 : -1; //@line 36 "common/Exception.cpp"
  __THREW__ = 0;
  $4 = (invoke_ii(3492,($3|0))|0); //@line 36 "common/Exception.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if ($6) {
   label = 5;
   break;
  }
  _memset(($4|0),0,($size_buffer$0|0))|0; //@line 37 "common/Exception.cpp"
  HEAP32[$args>>2] = $varargs; //@line 39 "common/Exception.cpp"
  $7 = (_vsnprintf($4,$size_buffer$0,$fmt,$args)|0); //@line 40 "common/Exception.cpp"
  $8 = ($7|0)==($size_buffer$0|0); //@line 50 "common/Exception.cpp"
  $9 = ($7|0)==(-1); //@line 50 "common/Exception.cpp"
  $or$cond = $8 | $9; //@line 50 "common/Exception.cpp"
  $10 = (($size_buffer$0) + -1)|0; //@line 50 "common/Exception.cpp"
  $11 = ($7|0)==($10|0); //@line 50 "common/Exception.cpp"
  $or$cond3 = $11 | $or$cond; //@line 50 "common/Exception.cpp"
  if ($or$cond3) {
   $12 = $size_buffer$0 << 1; //@line 51 "common/Exception.cpp"
   $size_buffer$1 = $12;
  } else {
   $17 = ($7|0)>($size_buffer$0|0); //@line 52 "common/Exception.cpp"
   if (!($17)) {
    $$lcssa = $4;
    break;
   }
   $18 = (($7) + 2)|0; //@line 53 "common/Exception.cpp"
   $size_buffer$1 = $18;
  }
  __ZdaPv($4); //@line 57 "common/Exception.cpp"
  $size_buffer$0 = $size_buffer$1;
 }
 if ((label|0) == 5) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  $lpad$phi$index = $13;$lpad$phi$index2 = $14;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 61 "common/Exception.cpp"
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $19 = (_strlen($$lcssa)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viii(3460,($0|0),($$lcssa|0),($19|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  $lpad$phi$index = $15;$lpad$phi$index2 = $16;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 61 "common/Exception.cpp"
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $22 = HEAP8[$1>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $23 = $22 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $24 = ($23<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($24) {
  $28 = ((($1)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$28>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$1>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 } else {
  $25 = ((($this)) + 12|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $26 = HEAP32[$25>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$26>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $27 = ((($this)) + 8|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$27>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 __THREW__ = 0;
 invoke_vii(3368,($1|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $29 = __THREW__; __THREW__ = 0;
 $30 = $29&1;
 if ($30) {
  $31 = ___cxa_find_matching_catch_3(0|0)|0;
  $32 = tempRet0;
  ___clang_call_terminate($31); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 } else {
  ;HEAP32[$1>>2]=HEAP32[$0>>2]|0;HEAP32[$1+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$1+8>>2]=HEAP32[$0+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 59 "common/Exception.cpp"
  __ZdaPv($$lcssa); //@line 60 "common/Exception.cpp"
  STACKTOP = sp;return; //@line 61 "common/Exception.cpp"
 }
}
function __ZN4love10luax_refifEP9lua_Statei($L,$type) {
 $L = $L|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $r$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_type($L,-1)|0); //@line 84 "common/runtime.cpp"
 $1 = ($0|0)==($type|0); //@line 84 "common/runtime.cpp"
 if (!($1)) {
  _lua_settop($L,-2); //@line 87 "common/runtime.cpp"
  $r$0 = 0;
  return ($r$0|0); //@line 89 "common/runtime.cpp"
 }
 $2 = (__Znwj(12)|0); //@line 85 "common/runtime.cpp"
 __THREW__ = 0;
 invoke_vii(3493,($2|0),($L|0)); //@line 85 "common/runtime.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  __ZdlPv($2); //@line 85 "common/runtime.cpp"
  ___resumeException($5|0);
  // unreachable;
 } else {
  $r$0 = $2;
  return ($r$0|0); //@line 89 "common/runtime.cpp"
 }
 return (0)|0;
}
function __ZN4love14luax_tobooleanEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_toboolean($L,$idx)|0); //@line 100 "common/runtime.cpp"
 $1 = ($0|0)!=(0); //@line 100 "common/runtime.cpp"
 return ($1|0); //@line 100 "common/runtime.cpp"
}
function __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$b) {
 $L = $L|0;
 $b = $b|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $b&1; //@line 105 "common/runtime.cpp"
 _lua_pushboolean($L,$0); //@line 105 "common/runtime.cpp"
 return; //@line 106 "common/runtime.cpp"
}
function __ZN4love15luax_optbooleanEP9lua_Stateib($L,$idx,$b) {
 $L = $L|0;
 $idx = $idx|0;
 $b = $b|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_type($L,$idx)|0); //@line 110 "common/runtime.cpp"
 $1 = ($0|0)==(1); //@line 110 "common/runtime.cpp"
 if (!($1)) {
  $$0 = $b;
  return ($$0|0); //@line 113 "common/runtime.cpp"
 }
 $2 = (_lua_toboolean($L,$idx)|0); //@line 111 "common/runtime.cpp"
 $3 = ($2|0)==(1); //@line 111 "common/runtime.cpp"
 $$0 = $3;
 return ($$0|0); //@line 113 "common/runtime.cpp"
}
function __ZN4love13luax_tostringEP9lua_Statei($agg$result,$L,$idx) {
 $agg$result = $agg$result|0;
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, $1 = 0, $len = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $len = sp;
 $0 = (_lua_tolstring($L,$idx,$len)|0); //@line 118 "common/runtime.cpp"
 $1 = HEAP32[$len>>2]|0; //@line 119 "common/runtime.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($agg$result,$0,$1); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 STACKTOP = sp;return; //@line 120 "common/runtime.cpp"
}
function __ZN4love16luax_checkstringEP9lua_Statei($agg$result,$L,$idx) {
 $agg$result = $agg$result|0;
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, $1 = 0, $len = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $len = sp;
 $0 = (_luaL_checklstring($L,$idx,$len)|0); //@line 125 "common/runtime.cpp"
 $1 = HEAP32[$len>>2]|0; //@line 126 "common/runtime.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($agg$result,$0,$1); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 STACKTOP = sp;return; //@line 127 "common/runtime.cpp"
}
function __ZN4love15luax_pushstringEP9lua_StateRKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE($L,$str) {
 $L = $L|0;
 $str = $str|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$str>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $1 = $0 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = ($1<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ((($str)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = HEAP32[$3>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($str)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = $2 ? $5 : $4; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($str)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = HEAP32[$7>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $0&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $9 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $2 ? $10 : $8; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 _lua_pushlstring($L,$6,$11); //@line 131 "common/runtime.cpp"
 return; //@line 132 "common/runtime.cpp"
}
function __ZN4love13luax_boolflagEP9lua_StateiPKcb($L,$table_index,$key,$defaultValue) {
 $L = $L|0;
 $table_index = $table_index|0;
 $key = $key|0;
 $defaultValue = $defaultValue|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $retval$0$in = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _lua_getfield($L,$table_index,$key); //@line 136 "common/runtime.cpp"
 $0 = (_lua_type($L,-1)|0); //@line 139 "common/runtime.cpp"
 $1 = ($0|0)<(1); //@line 139 "common/runtime.cpp"
 if ($1) {
  $retval$0$in = $defaultValue;
 } else {
  $2 = (_lua_toboolean($L,-1)|0); //@line 142 "common/runtime.cpp"
  $3 = ($2|0)!=(0); //@line 142 "common/runtime.cpp"
  $retval$0$in = $3;
 }
 _lua_settop($L,-2); //@line 144 "common/runtime.cpp"
 return ($retval$0$in|0); //@line 145 "common/runtime.cpp"
}
function __ZN4love12luax_intflagEP9lua_StateiPKci($L,$table_index,$key,$defaultValue) {
 $L = $L|0;
 $table_index = $table_index|0;
 $key = $key|0;
 $defaultValue = $defaultValue|0;
 var $0 = 0, $1 = 0, $2 = 0, $retval$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _lua_getfield($L,$table_index,$key); //@line 150 "common/runtime.cpp"
 $0 = (_lua_isnumber($L,-1)|0); //@line 153 "common/runtime.cpp"
 $1 = ($0|0)==(0); //@line 153 "common/runtime.cpp"
 if ($1) {
  $retval$0 = $defaultValue;
 } else {
  $2 = (_lua_tointeger($L,-1)|0); //@line 156 "common/runtime.cpp"
  $retval$0 = $2;
 }
 _lua_settop($L,-2); //@line 158 "common/runtime.cpp"
 return ($retval$0|0); //@line 159 "common/runtime.cpp"
}
function __ZN4love16luax_assert_argcEP9lua_Statei($L,$min) {
 $L = $L|0;
 $min = $min|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 162 "common/runtime.cpp"
 $0 = (_lua_gettop($L)|0); //@line 164 "common/runtime.cpp"
 $1 = ($0|0)<($min|0); //@line 165 "common/runtime.cpp"
 if (!($1)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 168 "common/runtime.cpp"
 }
 HEAP32[$vararg_buffer>>2] = $0; //@line 166 "common/runtime.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 166 "common/runtime.cpp"
 HEAP32[$vararg_ptr1>>2] = $min; //@line 166 "common/runtime.cpp"
 $2 = (_luaL_error($L,327069,$vararg_buffer)|0); //@line 166 "common/runtime.cpp"
 $$0 = $2;
 STACKTOP = sp;return ($$0|0); //@line 168 "common/runtime.cpp"
}
function __ZN4love16luax_assert_argcEP9lua_Stateii($L,$min,$max) {
 $L = $L|0;
 $min = $min|0;
 $max = $max|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 170 "common/runtime.cpp"
 $0 = (_lua_gettop($L)|0); //@line 172 "common/runtime.cpp"
 $1 = ($0|0)<($min|0); //@line 173 "common/runtime.cpp"
 $2 = ($0|0)>($max|0); //@line 173 "common/runtime.cpp"
 $or$cond = $1 | $2; //@line 173 "common/runtime.cpp"
 if (!($or$cond)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 176 "common/runtime.cpp"
 }
 HEAP32[$vararg_buffer>>2] = $0; //@line 174 "common/runtime.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 174 "common/runtime.cpp"
 HEAP32[$vararg_ptr1>>2] = $min; //@line 174 "common/runtime.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 174 "common/runtime.cpp"
 HEAP32[$vararg_ptr2>>2] = $max; //@line 174 "common/runtime.cpp"
 $3 = (_luaL_error($L,327133,$vararg_buffer)|0); //@line 174 "common/runtime.cpp"
 $$0 = $3;
 STACKTOP = sp;return ($$0|0); //@line 176 "common/runtime.cpp"
}
function __ZN4love12luax_requireEP9lua_StatePKc($L,$name) {
 $L = $L|0;
 $name = $name|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _lua_getfield($L,-10002,523817); //@line 211 "common/runtime.cpp"
 _lua_pushstring($L,$name); //@line 212 "common/runtime.cpp"
 _lua_call($L,1,1); //@line 213 "common/runtime.cpp"
 return 1; //@line 214 "common/runtime.cpp"
}
function __ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$m) {
 $L = $L|0;
 $m = $m|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $t$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($m)) + 8|0); //@line 219 "common/runtime.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 219 "common/runtime.cpp"
 $2 = ((($m)) + 4|0); //@line 219 "common/runtime.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 219 "common/runtime.cpp"
 __ZN4love11addTypeNameENS_4TypeEPKc($1,$3); //@line 219 "common/runtime.cpp"
 (__ZN4love15luax_insistloveEP9lua_StatePKc($L,327191)|0); //@line 614 "common/runtime.cpp"
 $4 = (_lua_newuserdata($L,8)|0); //@line 224 "common/runtime.cpp"
 $5 = HEAP32[$m>>2]|0; //@line 225 "common/runtime.cpp"
 $6 = ((($4)) + 4|0); //@line 225 "common/runtime.cpp"
 HEAP32[$6>>2] = $5; //@line 225 "common/runtime.cpp"
 $7 = HEAP32[$0>>2]|0; //@line 226 "common/runtime.cpp"
 HEAP32[$4>>2] = $7; //@line 226 "common/runtime.cpp"
 $8 = $5; //@line 228 "common/runtime.cpp"
 $9 = $5; //@line 228 "common/runtime.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 228 "common/runtime.cpp"
 $11 = ((($10)) + 20|0); //@line 228 "common/runtime.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 228 "common/runtime.cpp"
 $13 = (FUNCTION_TABLE_ii[$12 & 8191]($8)|0); //@line 228 "common/runtime.cpp"
 (_luaL_newmetatable($L,$13)|0); //@line 228 "common/runtime.cpp"
 _lua_pushvalue($L,-1); //@line 229 "common/runtime.cpp"
 _lua_setfield($L,-2,523215); //@line 230 "common/runtime.cpp"
 _lua_pushcclosure($L,3494,0); //@line 231 "common/runtime.cpp"
 _lua_setfield($L,-2,523106); //@line 232 "common/runtime.cpp"
 (_lua_setmetatable($L,-2)|0); //@line 234 "common/runtime.cpp"
 $14 = HEAP32[$2>>2]|0; //@line 235 "common/runtime.cpp"
 _lua_setfield($L,-2,$14); //@line 235 "common/runtime.cpp"
 _lua_settop($L,-2); //@line 236 "common/runtime.cpp"
 _lua_getfield($L,-10002,445428); //@line 566 "common/runtime.cpp"
 $15 = (_lua_type($L,-1)|0); //@line 568 "common/runtime.cpp"
 $16 = ($15|0)==(5); //@line 568 "common/runtime.cpp"
 if (!($16)) {
  _lua_settop($L,-2); //@line 570 "common/runtime.cpp"
  _lua_createtable($L,0,0); //@line 571 "common/runtime.cpp"
  _lua_pushvalue($L,-1); //@line 572 "common/runtime.cpp"
  _lua_setfield($L,-10002,445428); //@line 573 "common/runtime.cpp"
 }
 _lua_createtable($L,0,0); //@line 242 "common/runtime.cpp"
 $17 = ((($m)) + 12|0); //@line 245 "common/runtime.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 245 "common/runtime.cpp"
 $19 = ($18|0)==(0|0); //@line 245 "common/runtime.cpp"
 if (!($19)) {
  $20 = HEAP32[$18>>2]|0; //@line 202 "common/runtime.cpp"
  $21 = ($20|0)==(0|0); //@line 202 "common/runtime.cpp"
  if (!($21)) {
   $23 = $18;
   while(1) {
    $22 = ((($23)) + 4|0); //@line 204 "common/runtime.cpp"
    $24 = HEAP32[$22>>2]|0; //@line 204 "common/runtime.cpp"
    _lua_pushcclosure($L,$24,0); //@line 204 "common/runtime.cpp"
    $25 = HEAP32[$23>>2]|0; //@line 205 "common/runtime.cpp"
    _lua_setfield($L,-2,$25); //@line 205 "common/runtime.cpp"
    $26 = ((($23)) + 8|0); //@line 202 "common/runtime.cpp"
    $27 = HEAP32[$26>>2]|0; //@line 202 "common/runtime.cpp"
    $28 = ($27|0)==(0|0); //@line 202 "common/runtime.cpp"
    if ($28) {
     break;
    } else {
     $23 = $26;
    }
   }
  }
 }
 $29 = ((($m)) + 16|0); //@line 249 "common/runtime.cpp"
 $30 = HEAP32[$29>>2]|0; //@line 249 "common/runtime.cpp"
 $31 = ($30|0)==(0|0); //@line 249 "common/runtime.cpp"
 if (!($31)) {
  $32 = HEAP32[$30>>2]|0; //@line 251 "common/runtime.cpp"
  $33 = ($32|0)==(0|0); //@line 251 "common/runtime.cpp"
  if (!($33)) {
   $34 = $32;$t$01 = $30;
   while(1) {
    (FUNCTION_TABLE_ii[$34 & 8191]($L)|0); //@line 252 "common/runtime.cpp"
    $35 = ((($t$01)) + 4|0); //@line 251 "common/runtime.cpp"
    $36 = HEAP32[$35>>2]|0; //@line 251 "common/runtime.cpp"
    $37 = ($36|0)==(0|0); //@line 251 "common/runtime.cpp"
    if ($37) {
     break;
    } else {
     $34 = $36;$t$01 = $35;
    }
   }
  }
 }
 _lua_pushvalue($L,-1); //@line 255 "common/runtime.cpp"
 $38 = HEAP32[$2>>2]|0; //@line 256 "common/runtime.cpp"
 _lua_setfield($L,-3,$38); //@line 256 "common/runtime.cpp"
 _lua_remove($L,-2); //@line 257 "common/runtime.cpp"
 $39 = HEAP32[$m>>2]|0; //@line 260 "common/runtime.cpp"
 __ZN4love6Module16registerInstanceEPS0_($39); //@line 260 "common/runtime.cpp"
 return 1; //@line 262 "common/runtime.cpp"
}
function __ZN4love15luax_insistloveEP9lua_StatePKc($L,$k) {
 $L = $L|0;
 $k = $k|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _lua_getfield($L,-10002,445428); //@line 566 "common/runtime.cpp"
 $0 = (_lua_type($L,-1)|0); //@line 568 "common/runtime.cpp"
 $1 = ($0|0)==(5); //@line 568 "common/runtime.cpp"
 if (!($1)) {
  _lua_settop($L,-2); //@line 570 "common/runtime.cpp"
  _lua_createtable($L,0,0); //@line 571 "common/runtime.cpp"
  _lua_pushvalue($L,-1); //@line 572 "common/runtime.cpp"
  _lua_setfield($L,-10002,445428); //@line 573 "common/runtime.cpp"
 }
 (__ZN4love11luax_insistEP9lua_StateiPKc($L,-1,$k)|0); //@line 587 "common/runtime.cpp"
 _lua_replace($L,-2); //@line 591 "common/runtime.cpp"
 return 1; //@line 593 "common/runtime.cpp"
}
function __ZN4love11luax_insistEP9lua_StateiPKc($L,$idx,$k) {
 $L = $L|0;
 $idx = $idx|0;
 $k = $k|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx>>>0)>(4294957296); //@line 547 "common/runtime.cpp"
 if ($0) {
  $1 = (_lua_gettop($L)|0); //@line 548 "common/runtime.cpp"
  $2 = (($idx) + 1)|0; //@line 548 "common/runtime.cpp"
  $3 = (($2) + ($1))|0; //@line 548 "common/runtime.cpp"
  $$0 = $3;
 } else {
  $$0 = $idx;
 }
 _lua_getfield($L,$$0,$k); //@line 550 "common/runtime.cpp"
 $4 = (_lua_type($L,-1)|0); //@line 553 "common/runtime.cpp"
 $5 = ($4|0)==(5); //@line 553 "common/runtime.cpp"
 if ($5) {
  return 1; //@line 561 "common/runtime.cpp"
 }
 _lua_settop($L,-2); //@line 555 "common/runtime.cpp"
 _lua_createtable($L,0,0); //@line 556 "common/runtime.cpp"
 _lua_pushvalue($L,-1); //@line 557 "common/runtime.cpp"
 _lua_setfield($L,$$0,$k); //@line 558 "common/runtime.cpp"
 return 1; //@line 561 "common/runtime.cpp"
}
function __ZN4loveL5w__gcEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_touserdata($L,1)|0); //@line 44 "common/runtime.cpp"
 $1 = ((($0)) + 4|0); //@line 45 "common/runtime.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 45 "common/runtime.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 45 "common/runtime.cpp"
 $4 = ((($3)) + 12|0); //@line 45 "common/runtime.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 45 "common/runtime.cpp"
 FUNCTION_TABLE_vi[$5 & 8191]($2); //@line 45 "common/runtime.cpp"
 return 0; //@line 46 "common/runtime.cpp"
}
function __ZN4love19luax_insistregistryEP9lua_StateNS_8RegistryE($L,$r) {
 $L = $L|0;
 $r = $r|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 609 "common/runtime.cpp"
 switch ($r|0) {
 case 0:  {
  (__ZN4love15luax_insistloveEP9lua_StatePKc($L,327191)|0); //@line 614 "common/runtime.cpp"
  $$0 = 1;
  break;
 }
 case 1:  {
  _lua_getfield($L,-10000,327200); //@line 550 "common/runtime.cpp"
  $0 = (_lua_type($L,-1)|0); //@line 553 "common/runtime.cpp"
  $1 = ($0|0)==(5); //@line 553 "common/runtime.cpp"
  if ($1) {
   $$0 = 1;
  } else {
   _lua_settop($L,-2); //@line 555 "common/runtime.cpp"
   _lua_createtable($L,0,0); //@line 556 "common/runtime.cpp"
   _lua_pushvalue($L,-1); //@line 557 "common/runtime.cpp"
   _lua_setfield($L,-10000,327200); //@line 558 "common/runtime.cpp"
   $$0 = 1;
  }
  break;
 }
 default: {
  $2 = (_luaL_error($L,327213,$vararg_buffer)|0); //@line 618 "common/runtime.cpp"
  $$0 = $2;
 }
 }
 STACKTOP = sp;return ($$0|0); //@line 620 "common/runtime.cpp"
}
function __ZN4love17luax_insistglobalEP9lua_StatePKc($L,$k) {
 $L = $L|0;
 $k = $k|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _lua_getfield($L,-10002,$k); //@line 566 "common/runtime.cpp"
 $0 = (_lua_type($L,-1)|0); //@line 568 "common/runtime.cpp"
 $1 = ($0|0)==(5); //@line 568 "common/runtime.cpp"
 if ($1) {
  return 1; //@line 576 "common/runtime.cpp"
 }
 _lua_settop($L,-2); //@line 570 "common/runtime.cpp"
 _lua_createtable($L,0,0); //@line 571 "common/runtime.cpp"
 _lua_pushvalue($L,-1); //@line 572 "common/runtime.cpp"
 _lua_setfield($L,-10002,$k); //@line 573 "common/runtime.cpp"
 return 1; //@line 576 "common/runtime.cpp"
}
function __ZN4love12luax_preloadEP9lua_StatePFiS1_EPKc($L,$f,$name) {
 $L = $L|0;
 $f = $f|0;
 $name = $name|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _lua_getfield($L,-10002,523111); //@line 267 "common/runtime.cpp"
 _lua_getfield($L,-1,523223); //@line 268 "common/runtime.cpp"
 _lua_pushcclosure($L,$f,0); //@line 269 "common/runtime.cpp"
 _lua_setfield($L,-2,$name); //@line 270 "common/runtime.cpp"
 _lua_settop($L,-3); //@line 271 "common/runtime.cpp"
 return 0; //@line 272 "common/runtime.cpp"
}
function __ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,$type,$name,$varargs) {
 $L = $L|0;
 $type = $type|0;
 $name = $name|0;
 $varargs = $varargs|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $arglist_current = 0, $arglist_current2 = 0, $arglist_next = 0, $arglist_next3 = 0, $expanded = 0, $expanded10 = 0, $expanded12 = 0, $expanded13 = 0, $expanded14 = 0, $expanded3 = 0, $expanded5 = 0, $expanded6 = 0, $expanded7 = 0;
 var $expanded9 = 0, $f$01 = 0, $fs = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $fs = sp;
 __ZN4love11addTypeNameENS_4TypeEPKc($type,$name); //@line 277 "common/runtime.cpp"
 _lua_getfield($L,-10000,327200); //@line 629 "common/runtime.cpp"
 $0 = (_lua_type($L,-1)|0); //@line 283 "common/runtime.cpp"
 $1 = ($0|0)==(5); //@line 283 "common/runtime.cpp"
 if ($1) {
  _lua_settop($L,-2); //@line 302 "common/runtime.cpp"
 } else {
  _lua_createtable($L,0,0); //@line 285 "common/runtime.cpp"
  _lua_replace($L,-2); //@line 286 "common/runtime.cpp"
  _lua_createtable($L,0,0); //@line 289 "common/runtime.cpp"
  _lua_pushlstring($L,442531,1); //@line 292 "common/runtime.cpp"
  _lua_setfield($L,-2,520718); //@line 293 "common/runtime.cpp"
  (_lua_setmetatable($L,-2)|0); //@line 296 "common/runtime.cpp"
  _lua_setfield($L,-10000,327200); //@line 299 "common/runtime.cpp"
 }
 (_luaL_newmetatable($L,$name)|0); //@line 304 "common/runtime.cpp"
 _lua_pushvalue($L,-1); //@line 307 "common/runtime.cpp"
 _lua_setfield($L,-2,523215); //@line 308 "common/runtime.cpp"
 _lua_pushcclosure($L,3494,0); //@line 311 "common/runtime.cpp"
 _lua_setfield($L,-2,523106); //@line 312 "common/runtime.cpp"
 _lua_pushcclosure($L,3495,0); //@line 315 "common/runtime.cpp"
 _lua_setfield($L,-2,519021); //@line 316 "common/runtime.cpp"
 _lua_pushstring($L,$name); //@line 319 "common/runtime.cpp"
 _lua_pushcclosure($L,3496,1); //@line 320 "common/runtime.cpp"
 _lua_setfield($L,-2,521666); //@line 321 "common/runtime.cpp"
 _lua_pushstring($L,$name); //@line 324 "common/runtime.cpp"
 _lua_pushcclosure($L,3497,1); //@line 325 "common/runtime.cpp"
 _lua_setfield($L,-2,521785); //@line 326 "common/runtime.cpp"
 _lua_pushcclosure($L,3498,0); //@line 329 "common/runtime.cpp"
 _lua_setfield($L,-2,327248); //@line 330 "common/runtime.cpp"
 HEAP32[$fs>>2] = $varargs; //@line 333 "common/runtime.cpp"
 $arglist_current = HEAP32[$fs>>2]|0; //@line 334 "common/runtime.cpp"
 $2 = $arglist_current; //@line 334 "common/runtime.cpp"
 $3 = ((0) + 4|0);
 $expanded3 = $3;
 $expanded = (($expanded3) - 1)|0;
 $4 = (($2) + ($expanded))|0; //@line 334 "common/runtime.cpp"
 $5 = ((0) + 4|0);
 $expanded7 = $5;
 $expanded6 = (($expanded7) - 1)|0;
 $expanded5 = $expanded6 ^ -1;
 $6 = $4 & $expanded5; //@line 334 "common/runtime.cpp"
 $7 = $6; //@line 334 "common/runtime.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 334 "common/runtime.cpp"
 $arglist_next = ((($7)) + 4|0); //@line 334 "common/runtime.cpp"
 HEAP32[$fs>>2] = $arglist_next; //@line 334 "common/runtime.cpp"
 $9 = ($8|0)==(0|0); //@line 334 "common/runtime.cpp"
 if ($9) {
  _lua_settop($L,-2); //@line 338 "common/runtime.cpp"
  STACKTOP = sp;return 0; //@line 339 "common/runtime.cpp"
 } else {
  $f$01 = $8;
 }
 while(1) {
  $10 = HEAP32[$f$01>>2]|0; //@line 202 "common/runtime.cpp"
  $11 = ($10|0)==(0|0); //@line 202 "common/runtime.cpp"
  if (!($11)) {
   $13 = $f$01;
   while(1) {
    $12 = ((($13)) + 4|0); //@line 204 "common/runtime.cpp"
    $14 = HEAP32[$12>>2]|0; //@line 204 "common/runtime.cpp"
    _lua_pushcclosure($L,$14,0); //@line 204 "common/runtime.cpp"
    $15 = HEAP32[$13>>2]|0; //@line 205 "common/runtime.cpp"
    _lua_setfield($L,-2,$15); //@line 205 "common/runtime.cpp"
    $16 = ((($13)) + 8|0); //@line 202 "common/runtime.cpp"
    $17 = HEAP32[$16>>2]|0; //@line 202 "common/runtime.cpp"
    $18 = ($17|0)==(0|0); //@line 202 "common/runtime.cpp"
    if ($18) {
     break;
    } else {
     $13 = $16;
    }
   }
  }
  $arglist_current2 = HEAP32[$fs>>2]|0; //@line 334 "common/runtime.cpp"
  $19 = $arglist_current2; //@line 334 "common/runtime.cpp"
  $20 = ((0) + 4|0);
  $expanded10 = $20;
  $expanded9 = (($expanded10) - 1)|0;
  $21 = (($19) + ($expanded9))|0; //@line 334 "common/runtime.cpp"
  $22 = ((0) + 4|0);
  $expanded14 = $22;
  $expanded13 = (($expanded14) - 1)|0;
  $expanded12 = $expanded13 ^ -1;
  $23 = $21 & $expanded12; //@line 334 "common/runtime.cpp"
  $24 = $23; //@line 334 "common/runtime.cpp"
  $25 = HEAP32[$24>>2]|0; //@line 334 "common/runtime.cpp"
  $arglist_next3 = ((($24)) + 4|0); //@line 334 "common/runtime.cpp"
  HEAP32[$fs>>2] = $arglist_next3; //@line 334 "common/runtime.cpp"
  $26 = ($25|0)==(0|0); //@line 334 "common/runtime.cpp"
  if ($26) {
   break;
  } else {
   $f$01 = $25;
  }
 }
 _lua_settop($L,-2); //@line 338 "common/runtime.cpp"
 STACKTOP = sp;return 0; //@line 339 "common/runtime.cpp"
}
function __ZN4loveL5w__eqEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_touserdata($L,1)|0); //@line 73 "common/runtime.cpp"
 $1 = (_lua_touserdata($L,2)|0); //@line 74 "common/runtime.cpp"
 $2 = ((($0)) + 4|0); //@line 75 "common/runtime.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 75 "common/runtime.cpp"
 $4 = ((($1)) + 4|0); //@line 75 "common/runtime.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 75 "common/runtime.cpp"
 $6 = ($3|0)==($5|0); //@line 75 "common/runtime.cpp"
 $7 = $6&1; //@line 105 "common/runtime.cpp"
 _lua_pushboolean($L,$7); //@line 105 "common/runtime.cpp"
 return 1; //@line 76 "common/runtime.cpp"
}
function __ZN4loveL11w__tostringEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 49 "common/runtime.cpp"
 $0 = (_lua_touserdata($L,1)|0); //@line 51 "common/runtime.cpp"
 $1 = (_lua_tolstring($L,-10003,0)|0); //@line 52 "common/runtime.cpp"
 $2 = ((($0)) + 4|0); //@line 53 "common/runtime.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 53 "common/runtime.cpp"
 HEAP32[$vararg_buffer>>2] = $1; //@line 53 "common/runtime.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 53 "common/runtime.cpp"
 HEAP32[$vararg_ptr1>>2] = $3; //@line 53 "common/runtime.cpp"
 (_lua_pushfstring($L,520637,$vararg_buffer)|0); //@line 53 "common/runtime.cpp"
 STACKTOP = sp;return 1; //@line 54 "common/runtime.cpp"
}
function __ZN4loveL7w__typeEP9lua_State($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _lua_pushvalue($L,-10003); //@line 59 "common/runtime.cpp"
 return 1; //@line 60 "common/runtime.cpp"
}
function __ZN4loveL9w__typeOfEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $t$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $t$i = sp;
 $0 = (_lua_touserdata($L,1)|0); //@line 65 "common/runtime.cpp"
 HEAP32[$t$i>>2] = 0; //@line 721 "common/runtime.cpp"
 $1 = (_luaL_checklstring($L,2,0)|0); //@line 722 "common/runtime.cpp"
 (__ZN4love11getTypeNameEPKcRNS_4TypeE($1,$t$i)|0); //@line 722 "common/runtime.cpp"
 $2 = HEAP32[$t$i>>2]|0; //@line 723 "common/runtime.cpp"
 $3 = HEAP32[$0>>2]|0; //@line 67 "common/runtime.cpp"
 $4 = HEAP32[156504]|0; //@line 67 "common/runtime.cpp"
 $5 = $2 >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $6 = ((($4) + ($3<<3)|0) + ($5<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = $2 & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = 1 << $7; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = HEAP32[$6>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $10 = $9 & $8; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = ($10|0)!=(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = $11&1; //@line 105 "common/runtime.cpp"
 _lua_pushboolean($L,$12); //@line 105 "common/runtime.cpp"
 STACKTOP = sp;return 1; //@line 68 "common/runtime.cpp"
}
function __ZN4love21luax_gettypemetatableEP9lua_StateNS_4TypeE($L,$type) {
 $L = $L|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $name = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp;
 HEAP32[$name>>2] = 0; //@line 344 "common/runtime.cpp"
 $0 = (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 345 "common/runtime.cpp"
 if ($0) {
  $1 = HEAP32[$name>>2]|0; //@line 346 "common/runtime.cpp"
  _lua_getfield($L,-10000,$1); //@line 346 "common/runtime.cpp"
  STACKTOP = sp;return; //@line 349 "common/runtime.cpp"
 } else {
  _lua_pushnil($L); //@line 348 "common/runtime.cpp"
  STACKTOP = sp;return; //@line 349 "common/runtime.cpp"
 }
}
function __ZN4love17luax_table_insertEP9lua_Stateiii($L,$tindex,$vindex,$pos) {
 $L = $L|0;
 $tindex = $tindex|0;
 $vindex = $vindex|0;
 $pos = $pos|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $i$04 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($tindex|0)<(0); //@line 353 "common/runtime.cpp"
 if ($0) {
  $1 = (_lua_gettop($L)|0); //@line 354 "common/runtime.cpp"
  $2 = (($tindex) + 1)|0; //@line 354 "common/runtime.cpp"
  $3 = (($2) + ($1))|0; //@line 354 "common/runtime.cpp"
  $$02 = $3;
 } else {
  $$02 = $tindex;
 }
 $4 = ($vindex|0)<(0); //@line 355 "common/runtime.cpp"
 if ($4) {
  $5 = (_lua_gettop($L)|0); //@line 356 "common/runtime.cpp"
  $6 = (($vindex) + 1)|0; //@line 356 "common/runtime.cpp"
  $7 = (($6) + ($5))|0; //@line 356 "common/runtime.cpp"
  $$01 = $7;
 } else {
  $$01 = $vindex;
 }
 $8 = ($pos|0)==(-1); //@line 358 "common/runtime.cpp"
 if ($8) {
  _lua_pushvalue($L,$$01); //@line 360 "common/runtime.cpp"
  $9 = (_lua_objlen($L,$$02)|0); //@line 700 "common/runtime.cpp"
  $10 = (($9) + 1)|0; //@line 361 "common/runtime.cpp"
  _lua_rawseti($L,$$02,$10); //@line 361 "common/runtime.cpp"
  return 0; //@line 376 "common/runtime.cpp"
 }
 $11 = ($pos|0)<(0); //@line 364 "common/runtime.cpp"
 if ($11) {
  $12 = (_lua_objlen($L,$$02)|0); //@line 700 "common/runtime.cpp"
  $13 = (($pos) + 1)|0; //@line 365 "common/runtime.cpp"
  $14 = (($13) + ($12))|0; //@line 365 "common/runtime.cpp"
  $$0 = $14;
 } else {
  $$0 = $pos;
 }
 $15 = (_lua_objlen($L,$$02)|0); //@line 700 "common/runtime.cpp"
 $16 = ($15|0)<($$0|0); //@line 367 "common/runtime.cpp"
 if (!($16)) {
  $17 = (($15) + 1)|0; //@line 367 "common/runtime.cpp"
  $i$04 = $17;
  while(1) {
   $18 = (($i$04) + -1)|0; //@line 369 "common/runtime.cpp"
   _lua_rawgeti($L,$$02,$18); //@line 369 "common/runtime.cpp"
   _lua_rawseti($L,$$02,$i$04); //@line 370 "common/runtime.cpp"
   $19 = ($18|0)>($$0|0); //@line 367 "common/runtime.cpp"
   if ($19) {
    $i$04 = $18;
   } else {
    break;
   }
  }
 }
 _lua_pushvalue($L,$$01); //@line 373 "common/runtime.cpp"
 _lua_rawseti($L,$$02,$$0); //@line 374 "common/runtime.cpp"
 return 0; //@line 376 "common/runtime.cpp"
}
function __ZN4love11luax_objlenEP9lua_Statei($L,$ndx) {
 $L = $L|0;
 $ndx = $ndx|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_objlen($L,$ndx)|0); //@line 700 "common/runtime.cpp"
 return ($0|0); //@line 700 "common/runtime.cpp"
}
function __ZN4love22luax_register_searcherEP9lua_StatePFiS1_Ei($L,$f,$pos) {
 $L = $L|0;
 $f = $f|0;
 $pos = $pos|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 378 "common/runtime.cpp"
 $vararg_buffer = sp; //@line 378 "common/runtime.cpp"
 _lua_getfield($L,-10002,523111); //@line 381 "common/runtime.cpp"
 $0 = (_lua_type($L,-1)|0); //@line 383 "common/runtime.cpp"
 $1 = ($0|0)==(0); //@line 383 "common/runtime.cpp"
 if ($1) {
  $2 = (_luaL_error($L,327255,$vararg_buffer)|0); //@line 384 "common/runtime.cpp"
  $$0 = $2;
  STACKTOP = sp;return ($$0|0); //@line 402 "common/runtime.cpp"
 }
 _lua_getfield($L,-1,523447); //@line 386 "common/runtime.cpp"
 $3 = (_lua_type($L,-1)|0); //@line 389 "common/runtime.cpp"
 $4 = ($3|0)==(0); //@line 389 "common/runtime.cpp"
 if ($4) {
  _lua_settop($L,-2); //@line 391 "common/runtime.cpp"
  _lua_getfield($L,-1,327310); //@line 392 "common/runtime.cpp"
 }
 $5 = (_lua_type($L,-1)|0); //@line 395 "common/runtime.cpp"
 $6 = ($5|0)==(0); //@line 395 "common/runtime.cpp"
 if ($6) {
  $7 = (_luaL_error($L,327320,$vararg_buffer1)|0); //@line 396 "common/runtime.cpp"
  $$0 = $7;
  STACKTOP = sp;return ($$0|0); //@line 402 "common/runtime.cpp"
 } else {
  _lua_pushcclosure($L,$f,0); //@line 398 "common/runtime.cpp"
  (__ZN4love17luax_table_insertEP9lua_Stateiii($L,-2,-1,$pos)|0); //@line 399 "common/runtime.cpp"
  _lua_settop($L,-4); //@line 400 "common/runtime.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 402 "common/runtime.cpp"
 }
 return (0)|0;
}
function __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,$type,$object) {
 $L = $L|0;
 $type = $type|0;
 $object = $object|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name$i = 0, $name$i1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name$i1 = sp + 4|0;
 $name$i = sp;
 $0 = ($object|0)==(0|0); //@line 422 "common/runtime.cpp"
 if ($0) {
  _lua_pushnil($L); //@line 424 "common/runtime.cpp"
  STACKTOP = sp;return; //@line 460 "common/runtime.cpp"
 }
 _lua_getfield($L,-10000,327200); //@line 629 "common/runtime.cpp"
 $1 = (_lua_type($L,-1)|0); //@line 432 "common/runtime.cpp"
 $2 = ($1|0)==(5); //@line 432 "common/runtime.cpp"
 if (!($2)) {
  _lua_settop($L,-2); //@line 434 "common/runtime.cpp"
  $3 = (_lua_newuserdata($L,8)|0); //@line 406 "common/runtime.cpp"
  $4 = HEAP32[$object>>2]|0; //@line 408 "common/runtime.cpp"
  $5 = ((($4)) + 8|0); //@line 408 "common/runtime.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 408 "common/runtime.cpp"
  FUNCTION_TABLE_vi[$6 & 8191]($object); //@line 408 "common/runtime.cpp"
  $7 = ((($3)) + 4|0); //@line 410 "common/runtime.cpp"
  HEAP32[$7>>2] = $object; //@line 410 "common/runtime.cpp"
  HEAP32[$3>>2] = $type; //@line 411 "common/runtime.cpp"
  HEAP32[$name$i>>2] = 442055; //@line 413 "common/runtime.cpp"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name$i)|0); //@line 414 "common/runtime.cpp"
  $8 = HEAP32[$name$i>>2]|0; //@line 416 "common/runtime.cpp"
  (_luaL_newmetatable($L,$8)|0); //@line 416 "common/runtime.cpp"
  (_lua_setmetatable($L,-2)|0); //@line 417 "common/runtime.cpp"
  STACKTOP = sp;return; //@line 460 "common/runtime.cpp"
 }
 _lua_pushlightuserdata($L,$object); //@line 439 "common/runtime.cpp"
 _lua_gettable($L,-2); //@line 440 "common/runtime.cpp"
 $9 = (_lua_type($L,-1)|0); //@line 443 "common/runtime.cpp"
 $10 = ($9|0)==(7); //@line 443 "common/runtime.cpp"
 if (!($10)) {
  _lua_settop($L,-2); //@line 445 "common/runtime.cpp"
  $11 = (_lua_newuserdata($L,8)|0); //@line 406 "common/runtime.cpp"
  $12 = HEAP32[$object>>2]|0; //@line 408 "common/runtime.cpp"
  $13 = ((($12)) + 8|0); //@line 408 "common/runtime.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 408 "common/runtime.cpp"
  FUNCTION_TABLE_vi[$14 & 8191]($object); //@line 408 "common/runtime.cpp"
  $15 = ((($11)) + 4|0); //@line 410 "common/runtime.cpp"
  HEAP32[$15>>2] = $object; //@line 410 "common/runtime.cpp"
  HEAP32[$11>>2] = $type; //@line 411 "common/runtime.cpp"
  HEAP32[$name$i1>>2] = 442055; //@line 413 "common/runtime.cpp"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name$i1)|0); //@line 414 "common/runtime.cpp"
  $16 = HEAP32[$name$i1>>2]|0; //@line 416 "common/runtime.cpp"
  (_luaL_newmetatable($L,$16)|0); //@line 416 "common/runtime.cpp"
  (_lua_setmetatable($L,-2)|0); //@line 417 "common/runtime.cpp"
  _lua_pushlightuserdata($L,$object); //@line 449 "common/runtime.cpp"
  _lua_pushvalue($L,-2); //@line 450 "common/runtime.cpp"
  _lua_settable($L,-4); //@line 453 "common/runtime.cpp"
 }
 _lua_remove($L,-2); //@line 457 "common/runtime.cpp"
 STACKTOP = sp;return; //@line 460 "common/runtime.cpp"
}
function __ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $$1 = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_type($L,$idx)|0); //@line 464 "common/runtime.cpp"
 $1 = ($0|0)==(7); //@line 464 "common/runtime.cpp"
 if (!($1)) {
  $$1 = 0;
  return ($$1|0); //@line 473 "common/runtime.cpp"
 }
 $2 = (_lua_touserdata($L,$idx)|0); //@line 467 "common/runtime.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 469 "common/runtime.cpp"
 $$off = (($3) + -1)|0; //@line 469 "common/runtime.cpp"
 $4 = ($$off>>>0)<(59); //@line 469 "common/runtime.cpp"
 if (!($4)) {
  $$1 = 0;
  return ($$1|0); //@line 473 "common/runtime.cpp"
 }
 $5 = HEAP32[156504]|0; //@line 470 "common/runtime.cpp"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($3<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)!=(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $$1 = $12;
 return ($$1|0); //@line 473 "common/runtime.cpp"
}
function __ZN4love16luax_getfunctionEP9lua_StatePKcS3_($L,$mod,$fn) {
 $L = $L|0;
 $mod = $mod|0;
 $fn = $fn|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_ptr6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0; //@line 475 "common/runtime.cpp"
 $vararg_buffer1 = sp + 8|0; //@line 475 "common/runtime.cpp"
 $vararg_buffer = sp; //@line 475 "common/runtime.cpp"
 _lua_getfield($L,-10002,445428); //@line 477 "common/runtime.cpp"
 $0 = (_lua_type($L,-1)|0); //@line 478 "common/runtime.cpp"
 $1 = ($0|0)==(0); //@line 478 "common/runtime.cpp"
 if ($1) {
  $2 = (_luaL_error($L,327383,$vararg_buffer)|0); //@line 478 "common/runtime.cpp"
  $$0 = $2;
  STACKTOP = sp;return ($$0|0); //@line 487 "common/runtime.cpp"
 }
 _lua_getfield($L,-1,$mod); //@line 479 "common/runtime.cpp"
 $3 = (_lua_type($L,-1)|0); //@line 480 "common/runtime.cpp"
 $4 = ($3|0)==(0); //@line 480 "common/runtime.cpp"
 if ($4) {
  HEAP32[$vararg_buffer1>>2] = $mod; //@line 480 "common/runtime.cpp"
  $5 = (_luaL_error($L,327411,$vararg_buffer1)|0); //@line 480 "common/runtime.cpp"
  $$0 = $5;
  STACKTOP = sp;return ($$0|0); //@line 487 "common/runtime.cpp"
 }
 _lua_getfield($L,-1,$fn); //@line 481 "common/runtime.cpp"
 $6 = (_lua_type($L,-1)|0); //@line 482 "common/runtime.cpp"
 $7 = ($6|0)==(0); //@line 482 "common/runtime.cpp"
 if ($7) {
  HEAP32[$vararg_buffer3>>2] = $mod; //@line 482 "common/runtime.cpp"
  $vararg_ptr6 = ((($vararg_buffer3)) + 4|0); //@line 482 "common/runtime.cpp"
  HEAP32[$vararg_ptr6>>2] = $fn; //@line 482 "common/runtime.cpp"
  $8 = (_luaL_error($L,327435,$vararg_buffer3)|0); //@line 482 "common/runtime.cpp"
  $$0 = $8;
  STACKTOP = sp;return ($$0|0); //@line 487 "common/runtime.cpp"
 } else {
  _lua_remove($L,-2); //@line 484 "common/runtime.cpp"
  _lua_remove($L,-2); //@line 485 "common/runtime.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 487 "common/runtime.cpp"
 }
 return (0)|0;
}
function __ZN4love12luax_convobjEP9lua_StateiPKcS3_($L,$idx,$mod,$fn) {
 $L = $L|0;
 $idx = $idx|0;
 $mod = $mod|0;
 $fn = $fn|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 489 "common/runtime.cpp"
 $vararg_buffer = sp; //@line 489 "common/runtime.cpp"
 $0 = ($idx>>>0)>(4294957296); //@line 492 "common/runtime.cpp"
 if ($0) {
  $1 = (_lua_gettop($L)|0); //@line 493 "common/runtime.cpp"
  $2 = (($idx) + 1)|0; //@line 493 "common/runtime.cpp"
  $3 = (($2) + ($1))|0; //@line 493 "common/runtime.cpp"
  $$0 = $3;
 } else {
  $$0 = $idx;
 }
 (__ZN4love16luax_getfunctionEP9lua_StatePKcS3_($L,$mod,$fn)|0); //@line 496 "common/runtime.cpp"
 _lua_pushvalue($L,$$0); //@line 497 "common/runtime.cpp"
 _lua_call($L,1,2); //@line 498 "common/runtime.cpp"
 $4 = (_lua_type($L,-2)|0); //@line 187 "common/runtime.cpp"
 $5 = ($4|0)<(1); //@line 187 "common/runtime.cpp"
 do {
  if ($5) {
   $6 = (_lua_isstring($L,-1)|0); //@line 189 "common/runtime.cpp"
   $7 = ($6|0)==(0); //@line 189 "common/runtime.cpp"
   if ($7) {
    (_luaL_error($L,520154,$vararg_buffer1)|0); //@line 192 "common/runtime.cpp"
    break;
   } else {
    $8 = (_lua_tolstring($L,-1,0)|0); //@line 190 "common/runtime.cpp"
    (_luaL_error($L,$8,$vararg_buffer)|0); //@line 190 "common/runtime.cpp"
    break;
   }
  }
 } while(0);
 _lua_settop($L,-2); //@line 500 "common/runtime.cpp"
 _lua_replace($L,$$0); //@line 501 "common/runtime.cpp"
 STACKTOP = sp;return 0; //@line 502 "common/runtime.cpp"
}
function __ZN4love12luax_convobjEP9lua_StatePiiPKcS4_($L,$idxs,$n,$mod,$fn) {
 $L = $L|0;
 $idxs = $idxs|0;
 $n = $n|0;
 $mod = $mod|0;
 $fn = $fn|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 505 "common/runtime.cpp"
 $vararg_buffer = sp; //@line 505 "common/runtime.cpp"
 (__ZN4love16luax_getfunctionEP9lua_StatePKcS3_($L,$mod,$fn)|0); //@line 507 "common/runtime.cpp"
 $0 = ($n|0)>(0); //@line 508 "common/runtime.cpp"
 if ($0) {
  $i$01 = 0;
  while(1) {
   $6 = (($idxs) + ($i$01<<2)|0); //@line 510 "common/runtime.cpp"
   $7 = HEAP32[$6>>2]|0; //@line 510 "common/runtime.cpp"
   _lua_pushvalue($L,$7); //@line 510 "common/runtime.cpp"
   $8 = (($i$01) + 1)|0; //@line 508 "common/runtime.cpp"
   $exitcond = ($8|0)==($n|0); //@line 508 "common/runtime.cpp"
   if ($exitcond) {
    break;
   } else {
    $i$01 = $8;
   }
  }
 }
 _lua_call($L,$n,2); //@line 512 "common/runtime.cpp"
 $1 = (_lua_type($L,-2)|0); //@line 187 "common/runtime.cpp"
 $2 = ($1|0)<(1); //@line 187 "common/runtime.cpp"
 do {
  if ($2) {
   $3 = (_lua_isstring($L,-1)|0); //@line 189 "common/runtime.cpp"
   $4 = ($3|0)==(0); //@line 189 "common/runtime.cpp"
   if ($4) {
    (_luaL_error($L,520154,$vararg_buffer1)|0); //@line 192 "common/runtime.cpp"
    break;
   } else {
    $5 = (_lua_tolstring($L,-1,0)|0); //@line 190 "common/runtime.cpp"
    (_luaL_error($L,$5,$vararg_buffer)|0); //@line 190 "common/runtime.cpp"
    break;
   }
  }
 } while(0);
 _lua_settop($L,-2); //@line 514 "common/runtime.cpp"
 if (!($0)) {
  STACKTOP = sp;return 0; //@line 517 "common/runtime.cpp"
 }
 $9 = HEAP32[$idxs>>2]|0; //@line 516 "common/runtime.cpp"
 _lua_replace($L,$9); //@line 516 "common/runtime.cpp"
 STACKTOP = sp;return 0; //@line 517 "common/runtime.cpp"
}
function __ZN4love23luax_insistpinnedthreadEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _lua_getfield($L,-10000,327462); //@line 640 "common/runtime.cpp"
 $0 = (_lua_type($L,-1)|0); //@line 642 "common/runtime.cpp"
 $1 = ($0|0)<(1); //@line 642 "common/runtime.cpp"
 if ($1) {
  _lua_settop($L,-2); //@line 644 "common/runtime.cpp"
  (_lua_pushthread($L)|0); //@line 651 "common/runtime.cpp"
  _lua_pushvalue($L,-1); //@line 652 "common/runtime.cpp"
  _lua_setfield($L,-10000,327462); //@line 653 "common/runtime.cpp"
 }
 $2 = (_lua_tothread($L,-1)|0); //@line 656 "common/runtime.cpp"
 _lua_settop($L,-2); //@line 657 "common/runtime.cpp"
 return ($2|0); //@line 658 "common/runtime.cpp"
}
function __ZN4love20luax_getpinnedthreadEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _lua_getfield($L,-10000,327462); //@line 663 "common/runtime.cpp"
 $0 = (_lua_tothread($L,-1)|0); //@line 664 "common/runtime.cpp"
 _lua_settop($L,-2); //@line 665 "common/runtime.cpp"
 return ($0|0); //@line 666 "common/runtime.cpp"
}
function _luax_typerror($L,$narg,$tname) {
 $L = $L|0;
 $narg = $narg|0;
 $tname = $tname|0;
 var $$not = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $argtname$2 = 0, $or$cond = 0, $t = 0, $vararg_buffer = 0, $vararg_ptr1 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $t = sp + 8|0;
 $0 = (_lua_type($L,$narg)|0); //@line 671 "common/runtime.cpp"
 $1 = ($0|0)==(7); //@line 675 "common/runtime.cpp"
 if ($1) {
  $2 = (_luaL_getmetafield($L,$narg,521785)|0); //@line 675 "common/runtime.cpp"
  $3 = ($2|0)==(0); //@line 675 "common/runtime.cpp"
  if ($3) {
   label = 6;
  } else {
   _lua_pushvalue($L,$narg); //@line 677 "common/runtime.cpp"
   $4 = (_lua_pcall($L,1,1,0)|0); //@line 678 "common/runtime.cpp"
   $5 = ($4|0)==(0); //@line 678 "common/runtime.cpp"
   if ($5) {
    $6 = (_lua_type($L,-1)|0); //@line 678 "common/runtime.cpp"
    $7 = ($6|0)==(4); //@line 678 "common/runtime.cpp"
    if ($7) {
     $8 = (_lua_tolstring($L,-1,0)|0); //@line 680 "common/runtime.cpp"
     $9 = (__ZN4love11getTypeNameEPKcRNS_4TypeE($8,$t)|0); //@line 685 "common/runtime.cpp"
     $$not = $9 ^ 1;
     $10 = ($8|0)==(0|0); //@line 690 "common/runtime.cpp"
     $or$cond = $10 | $$not;
     if ($or$cond) {
      label = 6;
     } else {
      $argtname$2 = $8;
     }
    } else {
     label = 6;
    }
   } else {
    label = 6;
   }
  }
 } else {
  label = 6;
 }
 if ((label|0) == 6) {
  $11 = (_lua_typename($L,$0)|0); //@line 691 "common/runtime.cpp"
  $argtname$2 = $11;
 }
 HEAP32[$vararg_buffer>>2] = $tname; //@line 693 "common/runtime.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 693 "common/runtime.cpp"
 HEAP32[$vararg_ptr1>>2] = $argtname$2; //@line 693 "common/runtime.cpp"
 $12 = (_lua_pushfstring($L,519492,$vararg_buffer)|0); //@line 693 "common/runtime.cpp"
 $13 = (_luaL_argerror($L,$narg,$12)|0); //@line 694 "common/runtime.cpp"
 STACKTOP = sp;return ($13|0); //@line 694 "common/runtime.cpp"
}
function _luax_register($L,$name,$l) {
 $L = $L|0;
 $name = $name|0;
 $l = $l|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($name|0)==(0|0); //@line 708 "common/runtime.cpp"
 if (!($0)) {
  _lua_createtable($L,0,0); //@line 709 "common/runtime.cpp"
  $1 = ($l|0)==(0|0); //@line 199 "common/runtime.cpp"
  if (!($1)) {
   $2 = HEAP32[$l>>2]|0; //@line 202 "common/runtime.cpp"
   $3 = ($2|0)==(0|0); //@line 202 "common/runtime.cpp"
   if (!($3)) {
    $5 = $l;
    while(1) {
     $4 = ((($5)) + 4|0); //@line 204 "common/runtime.cpp"
     $6 = HEAP32[$4>>2]|0; //@line 204 "common/runtime.cpp"
     _lua_pushcclosure($L,$6,0); //@line 204 "common/runtime.cpp"
     $7 = HEAP32[$5>>2]|0; //@line 205 "common/runtime.cpp"
     _lua_setfield($L,-2,$7); //@line 205 "common/runtime.cpp"
     $8 = ((($5)) + 8|0); //@line 202 "common/runtime.cpp"
     $9 = HEAP32[$8>>2]|0; //@line 202 "common/runtime.cpp"
     $10 = ($9|0)==(0|0); //@line 202 "common/runtime.cpp"
     if ($10) {
      break;
     } else {
      $5 = $8;
     }
    }
   }
  }
  _lua_pushvalue($L,-1); //@line 714 "common/runtime.cpp"
  _lua_setfield($L,-10002,$name); //@line 715 "common/runtime.cpp"
  return; //@line 717 "common/runtime.cpp"
 }
 $11 = ($l|0)==(0|0); //@line 199 "common/runtime.cpp"
 if ($11) {
  return; //@line 717 "common/runtime.cpp"
 }
 $12 = HEAP32[$l>>2]|0; //@line 202 "common/runtime.cpp"
 $13 = ($12|0)==(0|0); //@line 202 "common/runtime.cpp"
 if ($13) {
  return; //@line 717 "common/runtime.cpp"
 } else {
  $15 = $l;
 }
 while(1) {
  $14 = ((($15)) + 4|0); //@line 204 "common/runtime.cpp"
  $16 = HEAP32[$14>>2]|0; //@line 204 "common/runtime.cpp"
  _lua_pushcclosure($L,$16,0); //@line 204 "common/runtime.cpp"
  $17 = HEAP32[$15>>2]|0; //@line 205 "common/runtime.cpp"
  _lua_setfield($L,-2,$17); //@line 205 "common/runtime.cpp"
  $18 = ((($15)) + 8|0); //@line 202 "common/runtime.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 202 "common/runtime.cpp"
  $20 = ($19|0)==(0|0); //@line 202 "common/runtime.cpp"
  if ($20) {
   break;
  } else {
   $15 = $18;
  }
 }
 return; //@line 717 "common/runtime.cpp"
}
function __GLOBAL__sub_I_types_cpp() {
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0;
 var $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0;
 var $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0;
 var $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0;
 var $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0;
 var $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0;
 var $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0;
 var $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0;
 var $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0;
 var $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0;
 var $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0;
 var $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0;
 var $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0;
 var $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0;
 var $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0;
 var $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0;
 var $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0;
 var $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0;
 var $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0;
 var $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0;
 var $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0;
 var $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0;
 var $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0;
 var $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0;
 var $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0;
 var $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0;
 var $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0;
 var $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0;
 var $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0;
 var $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0;
 var $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0;
 var $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0;
 var $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0;
 var $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0;
 var $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0;
 var $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0;
 var $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0;
 var $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0;
 var $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0;
 var $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0;
 var $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0;
 var $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0;
 var $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0;
 var $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0;
 var $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0;
 var $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0;
 var $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0;
 var $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0;
 var $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0;
 var $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0;
 var $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0;
 var $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0;
 var $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 480|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 472|0;
 $1 = sp + 464|0;
 $2 = sp + 456|0;
 $3 = sp + 448|0;
 $4 = sp + 440|0;
 $5 = sp + 432|0;
 $6 = sp + 424|0;
 $7 = sp + 416|0;
 $8 = sp + 408|0;
 $9 = sp + 400|0;
 $10 = sp + 392|0;
 $11 = sp + 384|0;
 $12 = sp + 376|0;
 $13 = sp + 368|0;
 $14 = sp + 360|0;
 $15 = sp + 352|0;
 $16 = sp + 344|0;
 $17 = sp + 336|0;
 $18 = sp + 328|0;
 $19 = sp + 320|0;
 $20 = sp + 312|0;
 $21 = sp + 304|0;
 $22 = sp + 296|0;
 $23 = sp + 288|0;
 $24 = sp + 280|0;
 $25 = sp + 272|0;
 $26 = sp + 264|0;
 $27 = sp + 256|0;
 $28 = sp + 248|0;
 $29 = sp + 240|0;
 $30 = sp + 232|0;
 $31 = sp + 224|0;
 $32 = sp + 216|0;
 $33 = sp + 208|0;
 $34 = sp + 200|0;
 $35 = sp + 192|0;
 $36 = sp + 184|0;
 $37 = sp + 176|0;
 $38 = sp + 168|0;
 $39 = sp + 160|0;
 $40 = sp + 152|0;
 $41 = sp + 144|0;
 $42 = sp + 136|0;
 $43 = sp + 128|0;
 $44 = sp + 120|0;
 $45 = sp + 112|0;
 $46 = sp + 104|0;
 $47 = sp + 96|0;
 $48 = sp + 88|0;
 $49 = sp + 80|0;
 $50 = sp + 72|0;
 $51 = sp + 64|0;
 $52 = sp + 56|0;
 $53 = sp + 48|0;
 $54 = sp + 40|0;
 $55 = sp + 32|0;
 $56 = sp + 24|0;
 $57 = sp + 16|0;
 $58 = sp + 8|0;
 $59 = sp;
 $60 = $0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $61 = $60; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$61>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $62 = (($60) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $63 = $62; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$63>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($0,0)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $64 = $0; //@line 32 "common/types.cpp"
 $65 = $64; //@line 32 "common/types.cpp"
 $66 = HEAP32[$65>>2]|0; //@line 32 "common/types.cpp"
 $67 = (($64) + 4)|0; //@line 32 "common/types.cpp"
 $68 = $67; //@line 32 "common/types.cpp"
 $69 = HEAP32[$68>>2]|0; //@line 32 "common/types.cpp"
 $70 = 623456; //@line 32 "common/types.cpp"
 $71 = $70; //@line 32 "common/types.cpp"
 HEAP32[$71>>2] = $66; //@line 32 "common/types.cpp"
 $72 = (($70) + 4)|0; //@line 32 "common/types.cpp"
 $73 = $72; //@line 32 "common/types.cpp"
 HEAP32[$73>>2] = $69; //@line 32 "common/types.cpp"
 $74 = $1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $75 = $74; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$75>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $76 = (($74) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $77 = $76; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$77>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($1,1)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $78 = $1; //@line 34 "common/types.cpp"
 $79 = $78; //@line 34 "common/types.cpp"
 $80 = HEAP32[$79>>2]|0; //@line 34 "common/types.cpp"
 $81 = (($78) + 4)|0; //@line 34 "common/types.cpp"
 $82 = $81; //@line 34 "common/types.cpp"
 $83 = HEAP32[$82>>2]|0; //@line 34 "common/types.cpp"
 $84 = (623464); //@line 34 "common/types.cpp"
 $85 = $84; //@line 34 "common/types.cpp"
 HEAP32[$85>>2] = $80; //@line 34 "common/types.cpp"
 $86 = (($84) + 4)|0; //@line 34 "common/types.cpp"
 $87 = $86; //@line 34 "common/types.cpp"
 HEAP32[$87>>2] = $83; //@line 34 "common/types.cpp"
 $88 = $2; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $89 = $88; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$89>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $90 = (($88) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $91 = $90; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$91>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($2,2)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $92 = $2; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $93 = $92; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $94 = HEAP32[$93>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $95 = (($92) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $96 = $95; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $97 = HEAP32[$96>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $98 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $99 = $94 | $98; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $100 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $101 = $97 | $100; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $102 = (623472); //@line 35 "common/types.cpp"
 $103 = $102; //@line 35 "common/types.cpp"
 HEAP32[$103>>2] = $99; //@line 35 "common/types.cpp"
 $104 = (($102) + 4)|0; //@line 35 "common/types.cpp"
 $105 = $104; //@line 35 "common/types.cpp"
 HEAP32[$105>>2] = $101; //@line 35 "common/types.cpp"
 $106 = $3; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $107 = $106; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$107>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $108 = (($106) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $109 = $108; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$109>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($3,3)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $110 = $3; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $111 = $110; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $112 = HEAP32[$111>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $113 = (($110) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $114 = $113; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $115 = HEAP32[$114>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $116 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $117 = $112 | $116; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $118 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $119 = $115 | $118; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $120 = (623480); //@line 36 "common/types.cpp"
 $121 = $120; //@line 36 "common/types.cpp"
 HEAP32[$121>>2] = $117; //@line 36 "common/types.cpp"
 $122 = (($120) + 4)|0; //@line 36 "common/types.cpp"
 $123 = $122; //@line 36 "common/types.cpp"
 HEAP32[$123>>2] = $119; //@line 36 "common/types.cpp"
 $124 = $4; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $125 = $124; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$125>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $126 = (($124) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $127 = $126; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$127>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($4,4)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $128 = $4; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $129 = $128; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $130 = HEAP32[$129>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $131 = (($128) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $132 = $131; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $133 = HEAP32[$132>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $134 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $135 = $130 | $134; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $136 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $137 = $133 | $136; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $138 = (623488); //@line 37 "common/types.cpp"
 $139 = $138; //@line 37 "common/types.cpp"
 HEAP32[$139>>2] = $135; //@line 37 "common/types.cpp"
 $140 = (($138) + 4)|0; //@line 37 "common/types.cpp"
 $141 = $140; //@line 37 "common/types.cpp"
 HEAP32[$141>>2] = $137; //@line 37 "common/types.cpp"
 $142 = $5; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $143 = $142; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$143>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $144 = (($142) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $145 = $144; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$145>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($5,5)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $146 = $5; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $147 = $146; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $148 = HEAP32[$147>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $149 = (($146) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $150 = $149; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $151 = HEAP32[$150>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $152 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $153 = $148 | $152; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $154 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $155 = $151 | $154; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $156 = (623496); //@line 40 "common/types.cpp"
 $157 = $156; //@line 40 "common/types.cpp"
 HEAP32[$157>>2] = $153; //@line 40 "common/types.cpp"
 $158 = (($156) + 4)|0; //@line 40 "common/types.cpp"
 $159 = $158; //@line 40 "common/types.cpp"
 HEAP32[$159>>2] = $155; //@line 40 "common/types.cpp"
 $160 = $6; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $161 = $160; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$161>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $162 = (($160) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $163 = $162; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$163>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($6,6)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $164 = $6; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $165 = $164; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $166 = HEAP32[$165>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $167 = (($164) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $168 = $167; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $169 = HEAP32[$168>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $170 = HEAP32[(623496)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $171 = $166 | $170; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $172 = HEAP32[(623500)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $173 = $169 | $172; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $174 = (623504); //@line 41 "common/types.cpp"
 $175 = $174; //@line 41 "common/types.cpp"
 HEAP32[$175>>2] = $171; //@line 41 "common/types.cpp"
 $176 = (($174) + 4)|0; //@line 41 "common/types.cpp"
 $177 = $176; //@line 41 "common/types.cpp"
 HEAP32[$177>>2] = $173; //@line 41 "common/types.cpp"
 $178 = $7; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $179 = $178; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$179>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $180 = (($178) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $181 = $180; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$181>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($7,7)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $182 = $7; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $183 = $182; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $184 = HEAP32[$183>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $185 = (($182) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $186 = $185; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $187 = HEAP32[$186>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $188 = HEAP32[(623472)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $189 = $184 | $188; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $190 = HEAP32[(623476)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $191 = $187 | $190; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $192 = (623512); //@line 42 "common/types.cpp"
 $193 = $192; //@line 42 "common/types.cpp"
 HEAP32[$193>>2] = $189; //@line 42 "common/types.cpp"
 $194 = (($192) + 4)|0; //@line 42 "common/types.cpp"
 $195 = $194; //@line 42 "common/types.cpp"
 HEAP32[$195>>2] = $191; //@line 42 "common/types.cpp"
 $196 = $8; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $197 = $196; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$197>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $198 = (($196) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $199 = $198; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$199>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($8,8)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $200 = $8; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $201 = $200; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $202 = HEAP32[$201>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $203 = (($200) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $204 = $203; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $205 = HEAP32[$204>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $206 = HEAP32[(623472)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $207 = $202 | $206; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $208 = HEAP32[(623476)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $209 = $205 | $208; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $210 = (623520); //@line 44 "common/types.cpp"
 $211 = $210; //@line 44 "common/types.cpp"
 HEAP32[$211>>2] = $207; //@line 44 "common/types.cpp"
 $212 = (($210) + 4)|0; //@line 44 "common/types.cpp"
 $213 = $212; //@line 44 "common/types.cpp"
 HEAP32[$213>>2] = $209; //@line 44 "common/types.cpp"
 $214 = $9; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $215 = $214; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$215>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $216 = (($214) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $217 = $216; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$217>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($9,9)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $218 = $9; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $219 = $218; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $220 = HEAP32[$219>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $221 = (($218) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $222 = $221; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $223 = HEAP32[$222>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $224 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $225 = $220 | $224; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $226 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $227 = $223 | $226; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $228 = (623528); //@line 45 "common/types.cpp"
 $229 = $228; //@line 45 "common/types.cpp"
 HEAP32[$229>>2] = $225; //@line 45 "common/types.cpp"
 $230 = (($228) + 4)|0; //@line 45 "common/types.cpp"
 $231 = $230; //@line 45 "common/types.cpp"
 HEAP32[$231>>2] = $227; //@line 45 "common/types.cpp"
 $232 = $10; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $233 = $232; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$233>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $234 = (($232) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $235 = $234; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$235>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($10,10)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $236 = $10; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $237 = $236; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $238 = HEAP32[$237>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $239 = (($236) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $240 = $239; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $241 = HEAP32[$240>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $242 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $243 = $238 | $242; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $244 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $245 = $241 | $244; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $246 = (623536); //@line 48 "common/types.cpp"
 $247 = $246; //@line 48 "common/types.cpp"
 HEAP32[$247>>2] = $243; //@line 48 "common/types.cpp"
 $248 = (($246) + 4)|0; //@line 48 "common/types.cpp"
 $249 = $248; //@line 48 "common/types.cpp"
 HEAP32[$249>>2] = $245; //@line 48 "common/types.cpp"
 $250 = $11; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $251 = $250; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$251>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $252 = (($250) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $253 = $252; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$253>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($11,11)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $254 = $11; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $255 = $254; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $256 = HEAP32[$255>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $257 = (($254) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $258 = $257; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $259 = HEAP32[$258>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $260 = HEAP32[(623536)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $261 = $256 | $260; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $262 = HEAP32[(623540)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $263 = $259 | $262; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $264 = (623544); //@line 49 "common/types.cpp"
 $265 = $264; //@line 49 "common/types.cpp"
 HEAP32[$265>>2] = $261; //@line 49 "common/types.cpp"
 $266 = (($264) + 4)|0; //@line 49 "common/types.cpp"
 $267 = $266; //@line 49 "common/types.cpp"
 HEAP32[$267>>2] = $263; //@line 49 "common/types.cpp"
 $268 = $12; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $269 = $268; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$269>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $270 = (($268) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $271 = $270; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$271>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($12,12)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $272 = $12; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $273 = $272; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $274 = HEAP32[$273>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $275 = (($272) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $276 = $275; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $277 = HEAP32[$276>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $278 = HEAP32[(623544)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $279 = $274 | $278; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $280 = HEAP32[(623548)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $281 = $277 | $280; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $282 = (623552); //@line 50 "common/types.cpp"
 $283 = $282; //@line 50 "common/types.cpp"
 HEAP32[$283>>2] = $279; //@line 50 "common/types.cpp"
 $284 = (($282) + 4)|0; //@line 50 "common/types.cpp"
 $285 = $284; //@line 50 "common/types.cpp"
 HEAP32[$285>>2] = $281; //@line 50 "common/types.cpp"
 $286 = $13; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $287 = $286; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$287>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $288 = (($286) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $289 = $288; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$289>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($13,13)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $290 = $13; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $291 = $290; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $292 = HEAP32[$291>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $293 = (($290) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $294 = $293; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $295 = HEAP32[$294>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $296 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $297 = $292 | $296; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $298 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $299 = $295 | $298; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $300 = (623560); //@line 51 "common/types.cpp"
 $301 = $300; //@line 51 "common/types.cpp"
 HEAP32[$301>>2] = $297; //@line 51 "common/types.cpp"
 $302 = (($300) + 4)|0; //@line 51 "common/types.cpp"
 $303 = $302; //@line 51 "common/types.cpp"
 HEAP32[$303>>2] = $299; //@line 51 "common/types.cpp"
 $304 = $14; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $305 = $304; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$305>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $306 = (($304) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $307 = $306; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$307>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($14,14)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $308 = $14; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $309 = $308; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $310 = HEAP32[$309>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $311 = (($308) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $312 = $311; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $313 = HEAP32[$312>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $314 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $315 = $310 | $314; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $316 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $317 = $313 | $316; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $318 = (623568); //@line 52 "common/types.cpp"
 $319 = $318; //@line 52 "common/types.cpp"
 HEAP32[$319>>2] = $315; //@line 52 "common/types.cpp"
 $320 = (($318) + 4)|0; //@line 52 "common/types.cpp"
 $321 = $320; //@line 52 "common/types.cpp"
 HEAP32[$321>>2] = $317; //@line 52 "common/types.cpp"
 $322 = $15; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $323 = $322; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$323>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $324 = (($322) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $325 = $324; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$325>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($15,15)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $326 = $15; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $327 = $326; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $328 = HEAP32[$327>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $329 = (($326) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $330 = $329; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $331 = HEAP32[$330>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $332 = HEAP32[(623536)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $333 = $328 | $332; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $334 = HEAP32[(623540)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $335 = $331 | $334; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $336 = (623576); //@line 53 "common/types.cpp"
 $337 = $336; //@line 53 "common/types.cpp"
 HEAP32[$337>>2] = $333; //@line 53 "common/types.cpp"
 $338 = (($336) + 4)|0; //@line 53 "common/types.cpp"
 $339 = $338; //@line 53 "common/types.cpp"
 HEAP32[$339>>2] = $335; //@line 53 "common/types.cpp"
 $340 = $16; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $341 = $340; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$341>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $342 = (($340) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $343 = $342; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$343>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($16,16)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $344 = $16; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $345 = $344; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $346 = HEAP32[$345>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $347 = (($344) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $348 = $347; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $349 = HEAP32[$348>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $350 = HEAP32[(623536)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $351 = $346 | $350; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $352 = HEAP32[(623540)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $353 = $349 | $352; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $354 = (623584); //@line 54 "common/types.cpp"
 $355 = $354; //@line 54 "common/types.cpp"
 HEAP32[$355>>2] = $351; //@line 54 "common/types.cpp"
 $356 = (($354) + 4)|0; //@line 54 "common/types.cpp"
 $357 = $356; //@line 54 "common/types.cpp"
 HEAP32[$357>>2] = $353; //@line 54 "common/types.cpp"
 $358 = $17; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $359 = $358; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$359>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $360 = (($358) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $361 = $360; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$361>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($17,17)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $362 = $17; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $363 = $362; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $364 = HEAP32[$363>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $365 = (($362) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $366 = $365; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $367 = HEAP32[$366>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $368 = HEAP32[(623544)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $369 = $364 | $368; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $370 = HEAP32[(623548)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $371 = $367 | $370; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $372 = (623592); //@line 55 "common/types.cpp"
 $373 = $372; //@line 55 "common/types.cpp"
 HEAP32[$373>>2] = $369; //@line 55 "common/types.cpp"
 $374 = (($372) + 4)|0; //@line 55 "common/types.cpp"
 $375 = $374; //@line 55 "common/types.cpp"
 HEAP32[$375>>2] = $371; //@line 55 "common/types.cpp"
 $376 = $18; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $377 = $376; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$377>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $378 = (($376) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $379 = $378; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$379>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($18,18)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $380 = $18; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $381 = $380; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $382 = HEAP32[$381>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $383 = (($380) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $384 = $383; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $385 = HEAP32[$384>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $386 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $387 = $382 | $386; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $388 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $389 = $385 | $388; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $390 = (623600); //@line 56 "common/types.cpp"
 $391 = $390; //@line 56 "common/types.cpp"
 HEAP32[$391>>2] = $387; //@line 56 "common/types.cpp"
 $392 = (($390) + 4)|0; //@line 56 "common/types.cpp"
 $393 = $392; //@line 56 "common/types.cpp"
 HEAP32[$393>>2] = $389; //@line 56 "common/types.cpp"
 $394 = $19; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $395 = $394; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$395>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $396 = (($394) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $397 = $396; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$397>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($19,19)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $398 = $19; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $399 = $398; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $400 = HEAP32[$399>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $401 = (($398) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $402 = $401; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $403 = HEAP32[$402>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $404 = HEAP32[(623536)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $405 = $400 | $404; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $406 = HEAP32[(623540)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $407 = $403 | $406; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $408 = (623608); //@line 57 "common/types.cpp"
 $409 = $408; //@line 57 "common/types.cpp"
 HEAP32[$409>>2] = $405; //@line 57 "common/types.cpp"
 $410 = (($408) + 4)|0; //@line 57 "common/types.cpp"
 $411 = $410; //@line 57 "common/types.cpp"
 HEAP32[$411>>2] = $407; //@line 57 "common/types.cpp"
 $412 = $20; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $413 = $412; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$413>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $414 = (($412) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $415 = $414; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$415>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($20,20)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $416 = $20; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $417 = $416; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $418 = HEAP32[$417>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $419 = (($416) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $420 = $419; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $421 = HEAP32[$420>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $422 = HEAP32[(623536)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $423 = $418 | $422; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $424 = HEAP32[(623540)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $425 = $421 | $424; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $426 = (623616); //@line 58 "common/types.cpp"
 $427 = $426; //@line 58 "common/types.cpp"
 HEAP32[$427>>2] = $423; //@line 58 "common/types.cpp"
 $428 = (($426) + 4)|0; //@line 58 "common/types.cpp"
 $429 = $428; //@line 58 "common/types.cpp"
 HEAP32[$429>>2] = $425; //@line 58 "common/types.cpp"
 $430 = $21; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $431 = $430; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$431>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $432 = (($430) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $433 = $432; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$433>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($21,21)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $434 = $21; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $435 = $434; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $436 = HEAP32[$435>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $437 = (($434) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $438 = $437; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $439 = HEAP32[$438>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $440 = HEAP32[(623536)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $441 = $436 | $440; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $442 = HEAP32[(623540)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $443 = $439 | $442; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $444 = (623624); //@line 59 "common/types.cpp"
 $445 = $444; //@line 59 "common/types.cpp"
 HEAP32[$445>>2] = $441; //@line 59 "common/types.cpp"
 $446 = (($444) + 4)|0; //@line 59 "common/types.cpp"
 $447 = $446; //@line 59 "common/types.cpp"
 HEAP32[$447>>2] = $443; //@line 59 "common/types.cpp"
 $448 = $22; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $449 = $448; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$449>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $450 = (($448) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $451 = $450; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$451>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($22,22)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $452 = $22; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $453 = $452; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $454 = HEAP32[$453>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $455 = (($452) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $456 = $455; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $457 = HEAP32[$456>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $458 = HEAP32[(623472)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $459 = $454 | $458; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $460 = HEAP32[(623476)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $461 = $457 | $460; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $462 = (623632); //@line 62 "common/types.cpp"
 $463 = $462; //@line 62 "common/types.cpp"
 HEAP32[$463>>2] = $459; //@line 62 "common/types.cpp"
 $464 = (($462) + 4)|0; //@line 62 "common/types.cpp"
 $465 = $464; //@line 62 "common/types.cpp"
 HEAP32[$465>>2] = $461; //@line 62 "common/types.cpp"
 $466 = $23; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $467 = $466; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$467>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $468 = (($466) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $469 = $468; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$469>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($23,23)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $470 = $23; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $471 = $470; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $472 = HEAP32[$471>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $473 = (($470) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $474 = $473; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $475 = HEAP32[$474>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $476 = HEAP32[(623472)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $477 = $472 | $476; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $478 = HEAP32[(623476)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $479 = $475 | $478; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $480 = (623640); //@line 63 "common/types.cpp"
 $481 = $480; //@line 63 "common/types.cpp"
 HEAP32[$481>>2] = $477; //@line 63 "common/types.cpp"
 $482 = (($480) + 4)|0; //@line 63 "common/types.cpp"
 $483 = $482; //@line 63 "common/types.cpp"
 HEAP32[$483>>2] = $479; //@line 63 "common/types.cpp"
 $484 = $24; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $485 = $484; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$485>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $486 = (($484) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $487 = $486; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$487>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($24,24)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $488 = $24; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $489 = $488; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $490 = HEAP32[$489>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $491 = (($488) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $492 = $491; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $493 = HEAP32[$492>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $494 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $495 = $490 | $494; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $496 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $497 = $493 | $496; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $498 = (623648); //@line 66 "common/types.cpp"
 $499 = $498; //@line 66 "common/types.cpp"
 HEAP32[$499>>2] = $495; //@line 66 "common/types.cpp"
 $500 = (($498) + 4)|0; //@line 66 "common/types.cpp"
 $501 = $500; //@line 66 "common/types.cpp"
 HEAP32[$501>>2] = $497; //@line 66 "common/types.cpp"
 $502 = $25; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $503 = $502; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$503>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $504 = (($502) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $505 = $504; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$505>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($25,25)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $506 = $25; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $507 = $506; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $508 = HEAP32[$507>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $509 = (($506) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $510 = $509; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $511 = HEAP32[$510>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $512 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $513 = $508 | $512; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $514 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $515 = $511 | $514; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $516 = (623656); //@line 69 "common/types.cpp"
 $517 = $516; //@line 69 "common/types.cpp"
 HEAP32[$517>>2] = $513; //@line 69 "common/types.cpp"
 $518 = (($516) + 4)|0; //@line 69 "common/types.cpp"
 $519 = $518; //@line 69 "common/types.cpp"
 HEAP32[$519>>2] = $515; //@line 69 "common/types.cpp"
 $520 = $26; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $521 = $520; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$521>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $522 = (($520) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $523 = $522; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$523>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($26,26)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $524 = $26; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $525 = $524; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $526 = HEAP32[$525>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $527 = (($524) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $528 = $527; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $529 = HEAP32[$528>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $530 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $531 = $526 | $530; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $532 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $533 = $529 | $532; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $534 = (623664); //@line 70 "common/types.cpp"
 $535 = $534; //@line 70 "common/types.cpp"
 HEAP32[$535>>2] = $531; //@line 70 "common/types.cpp"
 $536 = (($534) + 4)|0; //@line 70 "common/types.cpp"
 $537 = $536; //@line 70 "common/types.cpp"
 HEAP32[$537>>2] = $533; //@line 70 "common/types.cpp"
 $538 = $27; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $539 = $538; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$539>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $540 = (($538) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $541 = $540; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$541>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($27,27)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $542 = $27; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $543 = $542; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $544 = HEAP32[$543>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $545 = (($542) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $546 = $545; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $547 = HEAP32[$546>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $548 = HEAP32[(623472)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $549 = $544 | $548; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $550 = HEAP32[(623476)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $551 = $547 | $550; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $552 = (623672); //@line 71 "common/types.cpp"
 $553 = $552; //@line 71 "common/types.cpp"
 HEAP32[$553>>2] = $549; //@line 71 "common/types.cpp"
 $554 = (($552) + 4)|0; //@line 71 "common/types.cpp"
 $555 = $554; //@line 71 "common/types.cpp"
 HEAP32[$555>>2] = $551; //@line 71 "common/types.cpp"
 $556 = $28; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $557 = $556; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$557>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $558 = (($556) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $559 = $558; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$559>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($28,28)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $560 = $28; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $561 = $560; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $562 = HEAP32[$561>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $563 = (($560) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $564 = $563; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $565 = HEAP32[$564>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $566 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $567 = $562 | $566; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $568 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $569 = $565 | $568; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $570 = (623680); //@line 74 "common/types.cpp"
 $571 = $570; //@line 74 "common/types.cpp"
 HEAP32[$571>>2] = $567; //@line 74 "common/types.cpp"
 $572 = (($570) + 4)|0; //@line 74 "common/types.cpp"
 $573 = $572; //@line 74 "common/types.cpp"
 HEAP32[$573>>2] = $569; //@line 74 "common/types.cpp"
 $574 = $29; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $575 = $574; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$575>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $576 = (($574) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $577 = $576; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$577>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($29,29)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $578 = $29; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $579 = $578; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $580 = HEAP32[$579>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $581 = (($578) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $582 = $581; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $583 = HEAP32[$582>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $584 = HEAP32[(623472)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $585 = $580 | $584; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $586 = HEAP32[(623476)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $587 = $583 | $586; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $588 = (623688); //@line 77 "common/types.cpp"
 $589 = $588; //@line 77 "common/types.cpp"
 HEAP32[$589>>2] = $585; //@line 77 "common/types.cpp"
 $590 = (($588) + 4)|0; //@line 77 "common/types.cpp"
 $591 = $590; //@line 77 "common/types.cpp"
 HEAP32[$591>>2] = $587; //@line 77 "common/types.cpp"
 $592 = $30; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $593 = $592; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$593>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $594 = (($592) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $595 = $594; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$595>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($30,30)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $596 = $30; //@line 78 "common/types.cpp"
 $597 = $596; //@line 78 "common/types.cpp"
 $598 = HEAP32[$597>>2]|0; //@line 78 "common/types.cpp"
 $599 = (($596) + 4)|0; //@line 78 "common/types.cpp"
 $600 = $599; //@line 78 "common/types.cpp"
 $601 = HEAP32[$600>>2]|0; //@line 78 "common/types.cpp"
 $602 = (623696); //@line 78 "common/types.cpp"
 $603 = $602; //@line 78 "common/types.cpp"
 HEAP32[$603>>2] = $598; //@line 78 "common/types.cpp"
 $604 = (($602) + 4)|0; //@line 78 "common/types.cpp"
 $605 = $604; //@line 78 "common/types.cpp"
 HEAP32[$605>>2] = $601; //@line 78 "common/types.cpp"
 $606 = $31; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $607 = $606; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$607>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $608 = (($606) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $609 = $608; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$609>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($31,31)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $610 = $31; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $611 = $610; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $612 = HEAP32[$611>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $613 = (($610) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $614 = $613; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $615 = HEAP32[$614>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $616 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $617 = $612 | $616; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $618 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $619 = $615 | $618; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $620 = (623704); //@line 81 "common/types.cpp"
 $621 = $620; //@line 81 "common/types.cpp"
 HEAP32[$621>>2] = $617; //@line 81 "common/types.cpp"
 $622 = (($620) + 4)|0; //@line 81 "common/types.cpp"
 $623 = $622; //@line 81 "common/types.cpp"
 HEAP32[$623>>2] = $619; //@line 81 "common/types.cpp"
 $624 = $32; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $625 = $624; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$625>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $626 = (($624) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $627 = $626; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$627>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($32,32)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $628 = $32; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $629 = $628; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $630 = HEAP32[$629>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $631 = (($628) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $632 = $631; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $633 = HEAP32[$632>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $634 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $635 = $630 | $634; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $636 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $637 = $633 | $636; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $638 = (623712); //@line 84 "common/types.cpp"
 $639 = $638; //@line 84 "common/types.cpp"
 HEAP32[$639>>2] = $635; //@line 84 "common/types.cpp"
 $640 = (($638) + 4)|0; //@line 84 "common/types.cpp"
 $641 = $640; //@line 84 "common/types.cpp"
 HEAP32[$641>>2] = $637; //@line 84 "common/types.cpp"
 $642 = $33; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $643 = $642; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$643>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $644 = (($642) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $645 = $644; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$645>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($33,33)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $646 = $33; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $647 = $646; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $648 = HEAP32[$647>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $649 = (($646) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $650 = $649; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $651 = HEAP32[$650>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $652 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $653 = $648 | $652; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $654 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $655 = $651 | $654; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $656 = (623720); //@line 85 "common/types.cpp"
 $657 = $656; //@line 85 "common/types.cpp"
 HEAP32[$657>>2] = $653; //@line 85 "common/types.cpp"
 $658 = (($656) + 4)|0; //@line 85 "common/types.cpp"
 $659 = $658; //@line 85 "common/types.cpp"
 HEAP32[$659>>2] = $655; //@line 85 "common/types.cpp"
 $660 = $34; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $661 = $660; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$661>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $662 = (($660) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $663 = $662; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$663>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($34,34)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $664 = $34; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $665 = $664; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $666 = HEAP32[$665>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $667 = (($664) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $668 = $667; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $669 = HEAP32[$668>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $670 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $671 = $666 | $670; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $672 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $673 = $669 | $672; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $674 = (623728); //@line 86 "common/types.cpp"
 $675 = $674; //@line 86 "common/types.cpp"
 HEAP32[$675>>2] = $671; //@line 86 "common/types.cpp"
 $676 = (($674) + 4)|0; //@line 86 "common/types.cpp"
 $677 = $676; //@line 86 "common/types.cpp"
 HEAP32[$677>>2] = $673; //@line 86 "common/types.cpp"
 $678 = $35; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $679 = $678; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$679>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $680 = (($678) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $681 = $680; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$681>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($35,35)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $682 = $35; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $683 = $682; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $684 = HEAP32[$683>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $685 = (($682) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $686 = $685; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $687 = HEAP32[$686>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $688 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $689 = $684 | $688; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $690 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $691 = $687 | $690; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $692 = (623736); //@line 87 "common/types.cpp"
 $693 = $692; //@line 87 "common/types.cpp"
 HEAP32[$693>>2] = $689; //@line 87 "common/types.cpp"
 $694 = (($692) + 4)|0; //@line 87 "common/types.cpp"
 $695 = $694; //@line 87 "common/types.cpp"
 HEAP32[$695>>2] = $691; //@line 87 "common/types.cpp"
 $696 = $36; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $697 = $696; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$697>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $698 = (($696) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $699 = $698; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$699>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($36,36)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $700 = $36; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $701 = $700; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $702 = HEAP32[$701>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $703 = (($700) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $704 = $703; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $705 = HEAP32[$704>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $706 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $707 = $702 | $706; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $708 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $709 = $705 | $708; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $710 = (623744); //@line 88 "common/types.cpp"
 $711 = $710; //@line 88 "common/types.cpp"
 HEAP32[$711>>2] = $707; //@line 88 "common/types.cpp"
 $712 = (($710) + 4)|0; //@line 88 "common/types.cpp"
 $713 = $712; //@line 88 "common/types.cpp"
 HEAP32[$713>>2] = $709; //@line 88 "common/types.cpp"
 $714 = $37; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $715 = $714; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$715>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $716 = (($714) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $717 = $716; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$717>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($37,37)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $718 = $37; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $719 = $718; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $720 = HEAP32[$719>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $721 = (($718) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $722 = $721; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $723 = HEAP32[$722>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $724 = HEAP32[(623744)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $725 = $720 | $724; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $726 = HEAP32[(623748)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $727 = $723 | $726; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $728 = (623752); //@line 89 "common/types.cpp"
 $729 = $728; //@line 89 "common/types.cpp"
 HEAP32[$729>>2] = $725; //@line 89 "common/types.cpp"
 $730 = (($728) + 4)|0; //@line 89 "common/types.cpp"
 $731 = $730; //@line 89 "common/types.cpp"
 HEAP32[$731>>2] = $727; //@line 89 "common/types.cpp"
 $732 = $38; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $733 = $732; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$733>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $734 = (($732) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $735 = $734; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$735>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($38,38)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $736 = $38; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $737 = $736; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $738 = HEAP32[$737>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $739 = (($736) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $740 = $739; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $741 = HEAP32[$740>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $742 = HEAP32[(623744)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $743 = $738 | $742; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $744 = HEAP32[(623748)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $745 = $741 | $744; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $746 = (623760); //@line 90 "common/types.cpp"
 $747 = $746; //@line 90 "common/types.cpp"
 HEAP32[$747>>2] = $743; //@line 90 "common/types.cpp"
 $748 = (($746) + 4)|0; //@line 90 "common/types.cpp"
 $749 = $748; //@line 90 "common/types.cpp"
 HEAP32[$749>>2] = $745; //@line 90 "common/types.cpp"
 $750 = $39; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $751 = $750; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$751>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $752 = (($750) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $753 = $752; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$753>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($39,39)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $754 = $39; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $755 = $754; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $756 = HEAP32[$755>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $757 = (($754) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $758 = $757; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $759 = HEAP32[$758>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $760 = HEAP32[(623744)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $761 = $756 | $760; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $762 = HEAP32[(623748)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $763 = $759 | $762; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $764 = (623768); //@line 91 "common/types.cpp"
 $765 = $764; //@line 91 "common/types.cpp"
 HEAP32[$765>>2] = $761; //@line 91 "common/types.cpp"
 $766 = (($764) + 4)|0; //@line 91 "common/types.cpp"
 $767 = $766; //@line 91 "common/types.cpp"
 HEAP32[$767>>2] = $763; //@line 91 "common/types.cpp"
 $768 = $40; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $769 = $768; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$769>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $770 = (($768) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $771 = $770; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$771>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($40,40)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $772 = $40; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $773 = $772; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $774 = HEAP32[$773>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $775 = (($772) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $776 = $775; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $777 = HEAP32[$776>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $778 = HEAP32[(623744)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $779 = $774 | $778; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $780 = HEAP32[(623748)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $781 = $777 | $780; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $782 = (623776); //@line 92 "common/types.cpp"
 $783 = $782; //@line 92 "common/types.cpp"
 HEAP32[$783>>2] = $779; //@line 92 "common/types.cpp"
 $784 = (($782) + 4)|0; //@line 92 "common/types.cpp"
 $785 = $784; //@line 92 "common/types.cpp"
 HEAP32[$785>>2] = $781; //@line 92 "common/types.cpp"
 $786 = $41; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $787 = $786; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$787>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $788 = (($786) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $789 = $788; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$789>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($41,41)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $790 = $41; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $791 = $790; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $792 = HEAP32[$791>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $793 = (($790) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $794 = $793; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $795 = HEAP32[$794>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $796 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $797 = $792 | $796; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $798 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $799 = $795 | $798; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $800 = (623784); //@line 93 "common/types.cpp"
 $801 = $800; //@line 93 "common/types.cpp"
 HEAP32[$801>>2] = $797; //@line 93 "common/types.cpp"
 $802 = (($800) + 4)|0; //@line 93 "common/types.cpp"
 $803 = $802; //@line 93 "common/types.cpp"
 HEAP32[$803>>2] = $799; //@line 93 "common/types.cpp"
 $804 = $42; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $805 = $804; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$805>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $806 = (($804) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $807 = $806; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$807>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($42,42)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $808 = $42; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $809 = $808; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $810 = HEAP32[$809>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $811 = (($808) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $812 = $811; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $813 = HEAP32[$812>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $814 = HEAP32[(623784)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $815 = $810 | $814; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $816 = HEAP32[(623788)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $817 = $813 | $816; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $818 = (623792); //@line 94 "common/types.cpp"
 $819 = $818; //@line 94 "common/types.cpp"
 HEAP32[$819>>2] = $815; //@line 94 "common/types.cpp"
 $820 = (($818) + 4)|0; //@line 94 "common/types.cpp"
 $821 = $820; //@line 94 "common/types.cpp"
 HEAP32[$821>>2] = $817; //@line 94 "common/types.cpp"
 $822 = $43; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $823 = $822; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$823>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $824 = (($822) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $825 = $824; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$825>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($43,43)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $826 = $43; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $827 = $826; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $828 = HEAP32[$827>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $829 = (($826) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $830 = $829; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $831 = HEAP32[$830>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $832 = HEAP32[(623784)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $833 = $828 | $832; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $834 = HEAP32[(623788)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $835 = $831 | $834; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $836 = (623800); //@line 95 "common/types.cpp"
 $837 = $836; //@line 95 "common/types.cpp"
 HEAP32[$837>>2] = $833; //@line 95 "common/types.cpp"
 $838 = (($836) + 4)|0; //@line 95 "common/types.cpp"
 $839 = $838; //@line 95 "common/types.cpp"
 HEAP32[$839>>2] = $835; //@line 95 "common/types.cpp"
 $840 = $44; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $841 = $840; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$841>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $842 = (($840) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $843 = $842; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$843>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($44,44)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $844 = $44; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $845 = $844; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $846 = HEAP32[$845>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $847 = (($844) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $848 = $847; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $849 = HEAP32[$848>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $850 = HEAP32[(623784)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $851 = $846 | $850; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $852 = HEAP32[(623788)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $853 = $849 | $852; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $854 = (623808); //@line 96 "common/types.cpp"
 $855 = $854; //@line 96 "common/types.cpp"
 HEAP32[$855>>2] = $851; //@line 96 "common/types.cpp"
 $856 = (($854) + 4)|0; //@line 96 "common/types.cpp"
 $857 = $856; //@line 96 "common/types.cpp"
 HEAP32[$857>>2] = $853; //@line 96 "common/types.cpp"
 $858 = $45; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $859 = $858; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$859>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $860 = (($858) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $861 = $860; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$861>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($45,45)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $862 = $45; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $863 = $862; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $864 = HEAP32[$863>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $865 = (($862) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $866 = $865; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $867 = HEAP32[$866>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $868 = HEAP32[(623784)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $869 = $864 | $868; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $870 = HEAP32[(623788)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $871 = $867 | $870; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $872 = (623816); //@line 97 "common/types.cpp"
 $873 = $872; //@line 97 "common/types.cpp"
 HEAP32[$873>>2] = $869; //@line 97 "common/types.cpp"
 $874 = (($872) + 4)|0; //@line 97 "common/types.cpp"
 $875 = $874; //@line 97 "common/types.cpp"
 HEAP32[$875>>2] = $871; //@line 97 "common/types.cpp"
 $876 = $46; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $877 = $876; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$877>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $878 = (($876) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $879 = $878; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$879>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($46,46)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $880 = $46; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $881 = $880; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $882 = HEAP32[$881>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $883 = (($880) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $884 = $883; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $885 = HEAP32[$884>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $886 = HEAP32[(623784)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $887 = $882 | $886; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $888 = HEAP32[(623788)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $889 = $885 | $888; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $890 = (623824); //@line 98 "common/types.cpp"
 $891 = $890; //@line 98 "common/types.cpp"
 HEAP32[$891>>2] = $887; //@line 98 "common/types.cpp"
 $892 = (($890) + 4)|0; //@line 98 "common/types.cpp"
 $893 = $892; //@line 98 "common/types.cpp"
 HEAP32[$893>>2] = $889; //@line 98 "common/types.cpp"
 $894 = $47; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $895 = $894; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$895>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $896 = (($894) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $897 = $896; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$897>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($47,47)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $898 = $47; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $899 = $898; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $900 = HEAP32[$899>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $901 = (($898) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $902 = $901; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $903 = HEAP32[$902>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $904 = HEAP32[(623784)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $905 = $900 | $904; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $906 = HEAP32[(623788)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $907 = $903 | $906; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $908 = (623832); //@line 99 "common/types.cpp"
 $909 = $908; //@line 99 "common/types.cpp"
 HEAP32[$909>>2] = $905; //@line 99 "common/types.cpp"
 $910 = (($908) + 4)|0; //@line 99 "common/types.cpp"
 $911 = $910; //@line 99 "common/types.cpp"
 HEAP32[$911>>2] = $907; //@line 99 "common/types.cpp"
 $912 = $48; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $913 = $912; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$913>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $914 = (($912) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $915 = $914; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$915>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($48,48)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $916 = $48; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $917 = $916; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $918 = HEAP32[$917>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $919 = (($916) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $920 = $919; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $921 = HEAP32[$920>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $922 = HEAP32[(623784)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $923 = $918 | $922; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $924 = HEAP32[(623788)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $925 = $921 | $924; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $926 = (623840); //@line 100 "common/types.cpp"
 $927 = $926; //@line 100 "common/types.cpp"
 HEAP32[$927>>2] = $923; //@line 100 "common/types.cpp"
 $928 = (($926) + 4)|0; //@line 100 "common/types.cpp"
 $929 = $928; //@line 100 "common/types.cpp"
 HEAP32[$929>>2] = $925; //@line 100 "common/types.cpp"
 $930 = $49; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $931 = $930; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$931>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $932 = (($930) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $933 = $932; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$933>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($49,49)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $934 = $49; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $935 = $934; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $936 = HEAP32[$935>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $937 = (($934) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $938 = $937; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $939 = HEAP32[$938>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $940 = HEAP32[(623784)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $941 = $936 | $940; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $942 = HEAP32[(623788)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $943 = $939 | $942; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $944 = (623848); //@line 101 "common/types.cpp"
 $945 = $944; //@line 101 "common/types.cpp"
 HEAP32[$945>>2] = $941; //@line 101 "common/types.cpp"
 $946 = (($944) + 4)|0; //@line 101 "common/types.cpp"
 $947 = $946; //@line 101 "common/types.cpp"
 HEAP32[$947>>2] = $943; //@line 101 "common/types.cpp"
 $948 = $50; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $949 = $948; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$949>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $950 = (($948) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $951 = $950; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$951>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($50,50)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $952 = $50; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $953 = $952; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $954 = HEAP32[$953>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $955 = (($952) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $956 = $955; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $957 = HEAP32[$956>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $958 = HEAP32[(623784)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $959 = $954 | $958; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $960 = HEAP32[(623788)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $961 = $957 | $960; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $962 = (623856); //@line 102 "common/types.cpp"
 $963 = $962; //@line 102 "common/types.cpp"
 HEAP32[$963>>2] = $959; //@line 102 "common/types.cpp"
 $964 = (($962) + 4)|0; //@line 102 "common/types.cpp"
 $965 = $964; //@line 102 "common/types.cpp"
 HEAP32[$965>>2] = $961; //@line 102 "common/types.cpp"
 $966 = $51; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $967 = $966; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$967>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $968 = (($966) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $969 = $968; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$969>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($51,51)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $970 = $51; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $971 = $970; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $972 = HEAP32[$971>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $973 = (($970) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $974 = $973; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $975 = HEAP32[$974>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $976 = HEAP32[(623784)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $977 = $972 | $976; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $978 = HEAP32[(623788)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $979 = $975 | $978; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $980 = (623864); //@line 103 "common/types.cpp"
 $981 = $980; //@line 103 "common/types.cpp"
 HEAP32[$981>>2] = $977; //@line 103 "common/types.cpp"
 $982 = (($980) + 4)|0; //@line 103 "common/types.cpp"
 $983 = $982; //@line 103 "common/types.cpp"
 HEAP32[$983>>2] = $979; //@line 103 "common/types.cpp"
 $984 = $52; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $985 = $984; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$985>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $986 = (($984) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $987 = $986; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$987>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($52,52)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $988 = $52; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $989 = $988; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $990 = HEAP32[$989>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $991 = (($988) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $992 = $991; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $993 = HEAP32[$992>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $994 = HEAP32[(623784)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $995 = $990 | $994; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $996 = HEAP32[(623788)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $997 = $993 | $996; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $998 = (623872); //@line 104 "common/types.cpp"
 $999 = $998; //@line 104 "common/types.cpp"
 HEAP32[$999>>2] = $995; //@line 104 "common/types.cpp"
 $1000 = (($998) + 4)|0; //@line 104 "common/types.cpp"
 $1001 = $1000; //@line 104 "common/types.cpp"
 HEAP32[$1001>>2] = $997; //@line 104 "common/types.cpp"
 $1002 = $53; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1003 = $1002; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$1003>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1004 = (($1002) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1005 = $1004; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$1005>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($53,53)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1006 = $53; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1007 = $1006; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1008 = HEAP32[$1007>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1009 = (($1006) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1010 = $1009; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1011 = HEAP32[$1010>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1012 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1013 = $1008 | $1012; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1014 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1015 = $1011 | $1014; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1016 = (623880); //@line 107 "common/types.cpp"
 $1017 = $1016; //@line 107 "common/types.cpp"
 HEAP32[$1017>>2] = $1013; //@line 107 "common/types.cpp"
 $1018 = (($1016) + 4)|0; //@line 107 "common/types.cpp"
 $1019 = $1018; //@line 107 "common/types.cpp"
 HEAP32[$1019>>2] = $1015; //@line 107 "common/types.cpp"
 $1020 = $54; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1021 = $1020; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$1021>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1022 = (($1020) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1023 = $1022; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$1023>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($54,54)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1024 = $54; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1025 = $1024; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1026 = HEAP32[$1025>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1027 = (($1024) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1028 = $1027; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1029 = HEAP32[$1028>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1030 = HEAP32[(623464)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1031 = $1026 | $1030; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1032 = HEAP32[(623468)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1033 = $1029 | $1032; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1034 = (623888); //@line 108 "common/types.cpp"
 $1035 = $1034; //@line 108 "common/types.cpp"
 HEAP32[$1035>>2] = $1031; //@line 108 "common/types.cpp"
 $1036 = (($1034) + 4)|0; //@line 108 "common/types.cpp"
 $1037 = $1036; //@line 108 "common/types.cpp"
 HEAP32[$1037>>2] = $1033; //@line 108 "common/types.cpp"
 $1038 = $55; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1039 = $1038; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$1039>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1040 = (($1038) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1041 = $1040; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$1041>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($55,55)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1042 = $55; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1043 = $1042; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1044 = HEAP32[$1043>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1045 = (($1042) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1046 = $1045; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1047 = HEAP32[$1046>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1048 = HEAP32[(623488)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1049 = $1044 | $1048; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1050 = HEAP32[(623492)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1051 = $1047 | $1050; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1052 = (623896); //@line 111 "common/types.cpp"
 $1053 = $1052; //@line 111 "common/types.cpp"
 HEAP32[$1053>>2] = $1049; //@line 111 "common/types.cpp"
 $1054 = (($1052) + 4)|0; //@line 111 "common/types.cpp"
 $1055 = $1054; //@line 111 "common/types.cpp"
 HEAP32[$1055>>2] = $1051; //@line 111 "common/types.cpp"
 $1056 = $56; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1057 = $1056; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$1057>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1058 = (($1056) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1059 = $1058; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$1059>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($56,56)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1060 = $56; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1061 = $1060; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1062 = HEAP32[$1061>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1063 = (($1060) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1064 = $1063; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1065 = HEAP32[$1064>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1066 = HEAP32[(623480)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1067 = $1062 | $1066; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1068 = HEAP32[(623484)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1069 = $1065 | $1068; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1070 = (623904); //@line 114 "common/types.cpp"
 $1071 = $1070; //@line 114 "common/types.cpp"
 HEAP32[$1071>>2] = $1067; //@line 114 "common/types.cpp"
 $1072 = (($1070) + 4)|0; //@line 114 "common/types.cpp"
 $1073 = $1072; //@line 114 "common/types.cpp"
 HEAP32[$1073>>2] = $1069; //@line 114 "common/types.cpp"
 $1074 = $57; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1075 = $1074; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$1075>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1076 = (($1074) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1077 = $1076; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$1077>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($57,57)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1078 = $57; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1079 = $1078; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1080 = HEAP32[$1079>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1081 = (($1078) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1082 = $1081; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1083 = HEAP32[$1082>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1084 = HEAP32[(623480)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1085 = $1080 | $1084; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1086 = HEAP32[(623484)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1087 = $1083 | $1086; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1088 = (623912); //@line 115 "common/types.cpp"
 $1089 = $1088; //@line 115 "common/types.cpp"
 HEAP32[$1089>>2] = $1085; //@line 115 "common/types.cpp"
 $1090 = (($1088) + 4)|0; //@line 115 "common/types.cpp"
 $1091 = $1090; //@line 115 "common/types.cpp"
 HEAP32[$1091>>2] = $1087; //@line 115 "common/types.cpp"
 $1092 = $58; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1093 = $1092; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$1093>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1094 = (($1092) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1095 = $1094; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$1095>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($58,58)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1096 = $58; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1097 = $1096; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1098 = HEAP32[$1097>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1099 = (($1096) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1100 = $1099; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1101 = HEAP32[$1100>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1102 = HEAP32[(623480)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1103 = $1098 | $1102; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1104 = HEAP32[(623484)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1105 = $1101 | $1104; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1106 = (623920); //@line 116 "common/types.cpp"
 $1107 = $1106; //@line 116 "common/types.cpp"
 HEAP32[$1107>>2] = $1103; //@line 116 "common/types.cpp"
 $1108 = (($1106) + 4)|0; //@line 116 "common/types.cpp"
 $1109 = $1108; //@line 116 "common/types.cpp"
 HEAP32[$1109>>2] = $1105; //@line 116 "common/types.cpp"
 $1110 = $59; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1111 = $1110; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$1111>>2] = 1; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1112 = (($1110) + 4)|0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1113 = $1112; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 HEAP32[$1113>>2] = 0; //@line 1017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 (__ZNSt3__16bitsetILj60EElSEj($59,59)|0); //@line 1018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1114 = $59; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1115 = $1114; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1116 = HEAP32[$1115>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1117 = (($1114) + 4)|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1118 = $1117; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1119 = HEAP32[$1118>>2]|0; //@line 1047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1120 = HEAP32[(623480)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1121 = $1116 | $1120; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1122 = HEAP32[(623484)>>2]|0; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1123 = $1119 | $1122; //@line 281 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $1124 = (623928); //@line 117 "common/types.cpp"
 $1125 = $1124; //@line 117 "common/types.cpp"
 HEAP32[$1125>>2] = $1121; //@line 117 "common/types.cpp"
 $1126 = (($1124) + 4)|0; //@line 117 "common/types.cpp"
 $1127 = $1126; //@line 117 "common/types.cpp"
 HEAP32[$1127>>2] = $1123; //@line 117 "common/types.cpp"
 HEAP32[156504] = 623456; //@line 122 "common/types.cpp"
 __ZN4love9StringMapINS_4TypeELj60EEC2EPKNS2_5EntryEj(626020,0,0); //@line 124 "common/types.cpp"
 STACKTOP = sp;return;
}
function __ZNSt3__16bitsetILj60EElSEj($this,$__pos) {
 $this = $this|0;
 $__pos = $__pos|0;
 var $$idx = 0, $$sroa$7$1$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fold = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 24|0;
 $1 = sp + 16|0;
 $2 = sp + 8|0;
 $3 = sp;
 $4 = ($__pos>>>0)>(60); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $5 = $4 ? 60 : $__pos; //@line 807 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $6 = (60 - ($5))|0; //@line 808 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = $6 >>> 5; //@line 179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = (($this) + ($7<<2)|0); //@line 179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = $6 & 31; //@line 179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = ((($this)) + 4|0); //@line 179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $11 = $10; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($9|0)==(28); //@line 686 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if ($12) {
  $$idx = $7 << 5; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $13 = $$idx | 28; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $14 = ($13|0)>(0); //@line 556 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  if ($14) {
   $15 = ($13|0)<(28); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $16 = $15 ? $13 : 28; //@line 561 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $17 = (($13) - ($16))|0; //@line 562 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $18 = HEAP32[$8>>2]|0; //@line 565 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $19 = $18 & 268435455; //@line 565 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $20 = HEAP32[$10>>2]|0; //@line 566 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $21 = $20 & -268435456; //@line 566 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $22 = $21 | $19; //@line 567 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$10>>2] = $22; //@line 567 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $fold = (($6) - ($16))|0; //@line 569 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $23 = $fold & 31; //@line 569 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $24 = $17 >>> 5; //@line 575 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $25 = (0 - ($24))|0; //@line 576 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $26 = (($10) + ($25<<2)|0); //@line 587 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $27 = (($8) + ($25<<2)|0); //@line 577 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $28 = $17 >>> 3; //@line 580 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   _memmove(($26|0),($27|0),($28|0))|0; //@line 578 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $29 = $26; //@line 583 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $$sroa$7$1$i = $23;$30 = $29;
  } else {
   $$sroa$7$1$i = 28;$30 = $11;
  }
  HEAP32[$3>>2] = $30; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $31 = ((($3)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$31>>2] = $$sroa$7$1$i; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 } else {
  $32 = $8; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $33 = $this; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$0>>2] = $33; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $34 = ((($0)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$34>>2] = 0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$1>>2] = $32; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $35 = ((($1)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$35>>2] = $9; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$2>>2] = $11; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $36 = ((($2)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$36>>2] = 28; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  __ZNSt3__125__copy_backward_unalignedINS_8__bitsetILj2ELj60EEELb0EEENS_14__bit_iteratorIT_Lb0EXLi0EEEENS3_IS4_XT0_EXLi0EEEES6_S5_($3,$0,$1,$2); //@line 688 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 }
 $37 = ($5|0)==(0); //@line 386 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if ($37) {
  STACKTOP = sp;return ($this|0); //@line 810 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 }
 $38 = $5 >>> 5; //@line 339 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $39 = $38 << 2; //@line 340 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 _memset(($this|0),0,($39|0))|0; //@line 340 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $40 = $38 << 5; //@line 341 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $41 = ($40|0)==($5|0); //@line 343 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if ($41) {
  STACKTOP = sp;return ($this|0); //@line 810 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 }
 $42 = (($this) + ($38<<2)|0); //@line 345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $43 = (32 - ($5))|0; //@line 346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $44 = (($43) + ($40))|0; //@line 346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $45 = -1 >>> $44; //@line 346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $46 = $45 ^ -1; //@line 347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $47 = HEAP32[$42>>2]|0; //@line 347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $48 = $47 & $46; //@line 347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$42>>2] = $48; //@line 347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 STACKTOP = sp;return ($this|0); //@line 810 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
}
function __ZNSt3__125__copy_backward_unalignedINS_8__bitsetILj2ELj60EEELb0EEENS_14__bit_iteratorIT_Lb0EXLi0EEEENS3_IS4_XT0_EXLi0EEEES6_S5_($agg$result,$__first,$__last,$__result) {
 $agg$result = $agg$result|0;
 $__first = $__first|0;
 $__last = $__last|0;
 $__result = $__result|0;
 var $$pn = 0, $$pre = 0, $$pre$phi7Z2D = 0, $$pre$phiZ2D = 0, $$pre6 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__dn$0 = 0, $__n$0 = 0, $__n$1$lcssa = 0, $__n$13 = 0, $storemerge = 0, $sum = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$__last>>2]|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $1 = HEAP32[$__first>>2]|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $2 = (($0) - ($1))|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $3 = $2 << 3; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $4 = ((($__last)) + 4|0); //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $5 = HEAP32[$4>>2]|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $6 = (($3) + ($5))|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $7 = ((($__first)) + 4|0); //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $8 = HEAP32[$7>>2]|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $9 = (($6) - ($8))|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $10 = ($9|0)>(0); //@line 605 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $0; //@line 614 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($10)) {
  $$pre6 = ((($__result)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $$pre$phi7Z2D = $$pre6;
  $126 = HEAP32[$__result>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$agg$result>>2] = $126; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $127 = ((($agg$result)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $128 = HEAP32[$$pre$phi7Z2D>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$127>>2] = $128; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  return; //@line 679 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 }
 $12 = ($5|0)==(0); //@line 608 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if ($12) {
  $$pre = ((($__result)) + 4|0); //@line 644 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $$pre$phiZ2D = $$pre;$__n$0 = $9;
 } else {
  $13 = ($9|0)<($5|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $14 = $13 ? $9 : $5; //@line 610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $15 = (($9) - ($14))|0; //@line 611 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $16 = (32 - ($5))|0; //@line 612 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $17 = (($5) - ($14))|0; //@line 613 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $18 = -1 << $17; //@line 613 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $19 = -1 >>> $16; //@line 613 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $20 = $18 & $19; //@line 613 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $21 = HEAP32[$11>>2]|0; //@line 614 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $22 = $20 & $21; //@line 614 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $23 = ((($__result)) + 4|0); //@line 615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $24 = HEAP32[$23>>2]|0; //@line 615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $25 = ($24|0)<($14|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $26 = $25 ? $24 : $14; //@line 616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $27 = ($26|0)==(0); //@line 617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  if ($27) {
   $__dn$0 = $14;
  } else {
   $28 = (32 - ($24))|0; //@line 615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $29 = (($24) - ($26))|0; //@line 619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $30 = -1 << $29; //@line 619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $31 = -1 >>> $28; //@line 619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $32 = $30 & $31; //@line 619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $33 = $32 ^ -1; //@line 620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $34 = HEAP32[$__result>>2]|0; //@line 622 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $35 = HEAP32[$34>>2]|0; //@line 620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $36 = $35 & $33; //@line 622 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$34>>2] = $36; //@line 620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $37 = HEAP32[$23>>2]|0; //@line 621 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $38 = HEAP32[$4>>2]|0; //@line 621 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $39 = ($37>>>0)>($38>>>0); //@line 621 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $40 = (($38) - ($37))|0; //@line 624 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $41 = $22 >>> $40; //@line 624 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $42 = (($37) - ($38))|0; //@line 622 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $43 = $22 << $42; //@line 622 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $$pn = $39 ? $43 : $41; //@line 621 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $storemerge = $36 | $$pn; //@line 624 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$34>>2] = $storemerge; //@line 622 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $44 = HEAP32[$23>>2]|0; //@line 626 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $45 = (($44) - ($26))|0; //@line 625 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $46 = $45 & 31; //@line 626 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$23>>2] = $46; //@line 625 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $47 = (($14) - ($26))|0; //@line 627 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $__dn$0 = $47;
  }
  $48 = ($__dn$0|0)>(0); //@line 629 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  if ($48) {
   $49 = (0 - ($__dn$0))|0; //@line 633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $50 = HEAP32[$__result>>2]|0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $51 = ((($50)) + -4|0); //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$__result>>2] = $51; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $52 = $49 & 31; //@line 633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$23>>2] = $52; //@line 633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $53 = -1 << $52; //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $54 = $53 ^ -1; //@line 635 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $55 = HEAP32[$51>>2]|0; //@line 635 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $56 = $55 & $54; //@line 635 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$51>>2] = $56; //@line 635 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $57 = HEAP32[$4>>2]|0; //@line 636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $sum = (($__dn$0) + ($26))|0; //@line 636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $58 = (($57) - ($sum))|0; //@line 636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$4>>2] = $58; //@line 636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $59 = HEAP32[$23>>2]|0; //@line 637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $60 = (($59) - ($58))|0; //@line 637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $61 = $22 << $60; //@line 637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $62 = HEAP32[$51>>2]|0; //@line 637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $63 = $62 | $61; //@line 637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$51>>2] = $63; //@line 637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $$pre$phiZ2D = $23;$__n$0 = $15;
  } else {
   $$pre$phiZ2D = $23;$__n$0 = $15;
  }
 }
 $64 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 644 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $65 = (32 - ($64))|0; //@line 644 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $66 = -1 >>> $65; //@line 645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $67 = ($__n$0>>>0)>(31); //@line 646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if ($67) {
  $68 = $66 ^ -1; //@line 649 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $__n$13 = $__n$0;
  while(1) {
   $69 = HEAP32[$__last>>2]|0; //@line 648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $70 = ((($69)) + -4|0); //@line 648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$__last>>2] = $70; //@line 648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $71 = HEAP32[$70>>2]|0; //@line 648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $72 = HEAP32[$__result>>2]|0; //@line 649 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $73 = HEAP32[$72>>2]|0; //@line 649 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $74 = $73 & $68; //@line 649 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $75 = $71 >>> $65; //@line 650 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $76 = $74 | $75; //@line 650 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$72>>2] = $76; //@line 650 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $77 = ((($72)) + -4|0); //@line 651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$__result>>2] = $77; //@line 651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $78 = HEAP32[$77>>2]|0; //@line 651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $79 = $78 & $66; //@line 651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$77>>2] = $79; //@line 651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $80 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 652 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $81 = $71 << $80; //@line 652 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $82 = $79 | $81; //@line 652 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$77>>2] = $82; //@line 652 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $83 = (($__n$13) + -32)|0; //@line 646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $84 = ($83>>>0)>(31); //@line 646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   if ($84) {
    $__n$13 = $83;
   } else {
    break;
   }
  }
  $85 = $__n$0 & 31; //@line 646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $__n$1$lcssa = $85;
 } else {
  $__n$1$lcssa = $__n$0;
 }
 $86 = ($__n$1$lcssa|0)>(0); //@line 655 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($86)) {
  $$pre$phi7Z2D = $$pre$phiZ2D;
  $126 = HEAP32[$__result>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$agg$result>>2] = $126; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $127 = ((($agg$result)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $128 = HEAP32[$$pre$phi7Z2D>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$127>>2] = $128; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  return; //@line 679 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 }
 $87 = (32 - ($__n$1$lcssa))|0; //@line 657 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $88 = -1 << $87; //@line 657 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $89 = HEAP32[$__last>>2]|0; //@line 658 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $90 = ((($89)) + -4|0); //@line 658 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$__last>>2] = $90; //@line 658 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $91 = HEAP32[$90>>2]|0; //@line 658 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $92 = $91 & $88; //@line 658 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $93 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $94 = (32 - ($93))|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $95 = ($93|0)<($__n$1$lcssa|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $96 = $95 ? $93 : $__n$1$lcssa; //@line 660 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $97 = (($93) - ($96))|0; //@line 661 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $98 = -1 << $97; //@line 661 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $99 = -1 >>> $94; //@line 661 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $100 = $98 & $99; //@line 661 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $101 = $100 ^ -1; //@line 662 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $102 = HEAP32[$__result>>2]|0; //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $103 = HEAP32[$102>>2]|0; //@line 662 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $104 = $103 & $101; //@line 662 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$102>>2] = $104; //@line 662 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $105 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 663 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $106 = (32 - ($105))|0; //@line 663 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $107 = $92 >>> $106; //@line 663 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $108 = $104 | $107; //@line 663 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$102>>2] = $108; //@line 663 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $109 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 665 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $110 = (($109) - ($96))|0; //@line 664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $111 = $110 & 31; //@line 665 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$$pre$phiZ2D>>2] = $111; //@line 664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $112 = (($__n$1$lcssa) - ($96))|0; //@line 666 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $113 = ($112|0)>(0); //@line 667 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($113)) {
  $$pre$phi7Z2D = $$pre$phiZ2D;
  $126 = HEAP32[$__result>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$agg$result>>2] = $126; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $127 = ((($agg$result)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $128 = HEAP32[$$pre$phi7Z2D>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$127>>2] = $128; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  return; //@line 679 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 }
 $114 = ((($102)) + -4|0); //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$__result>>2] = $114; //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $115 = (0 - ($112))|0; //@line 671 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $116 = $115 & 31; //@line 671 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$$pre$phiZ2D>>2] = $116; //@line 671 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $117 = -1 << $116; //@line 672 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $118 = $117 ^ -1; //@line 673 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $119 = HEAP32[$114>>2]|0; //@line 673 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $120 = $119 & $118; //@line 673 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$114>>2] = $120; //@line 673 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $121 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 674 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $122 = (($__n$1$lcssa) + -32)|0; //@line 674 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $123 = (($122) + ($121))|0; //@line 674 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $124 = $92 << $123; //@line 674 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $125 = $120 | $124; //@line 674 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$114>>2] = $125; //@line 674 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $$pre$phi7Z2D = $$pre$phiZ2D;
 $126 = HEAP32[$__result>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$agg$result>>2] = $126; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $127 = ((($agg$result)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $128 = HEAP32[$$pre$phi7Z2D>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$127>>2] = $128; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 return; //@line 679 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
}
function __ZN4love9StringMapINS_4TypeELj60EEC2EPKNS2_5EntryEj($this,$entries,$num) {
 $this = $this|0;
 $entries = $entries|0;
 $num = $num|0;
 var $$02$i$i = 0, $$lcssa26 = 0, $$lcssa28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$08$i = 0, $i1$05 = 0, $scevgep = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 8|0); //@line 155 "common/StringMap.h"
 HEAP8[$0>>0] = 0; //@line 155 "common/StringMap.h"
 $1 = ((($this)) + 20|0); //@line 155 "common/StringMap.h"
 HEAP8[$1>>0] = 0; //@line 155 "common/StringMap.h"
 $2 = ((($this)) + 32|0); //@line 155 "common/StringMap.h"
 HEAP8[$2>>0] = 0; //@line 155 "common/StringMap.h"
 $3 = ((($this)) + 44|0); //@line 155 "common/StringMap.h"
 HEAP8[$3>>0] = 0; //@line 155 "common/StringMap.h"
 $4 = ((($this)) + 56|0); //@line 155 "common/StringMap.h"
 HEAP8[$4>>0] = 0; //@line 155 "common/StringMap.h"
 $5 = ((($this)) + 68|0); //@line 155 "common/StringMap.h"
 HEAP8[$5>>0] = 0; //@line 155 "common/StringMap.h"
 $6 = ((($this)) + 80|0); //@line 155 "common/StringMap.h"
 HEAP8[$6>>0] = 0; //@line 155 "common/StringMap.h"
 $7 = ((($this)) + 92|0); //@line 155 "common/StringMap.h"
 HEAP8[$7>>0] = 0; //@line 155 "common/StringMap.h"
 $8 = ((($this)) + 104|0); //@line 155 "common/StringMap.h"
 HEAP8[$8>>0] = 0; //@line 155 "common/StringMap.h"
 $9 = ((($this)) + 116|0); //@line 155 "common/StringMap.h"
 HEAP8[$9>>0] = 0; //@line 155 "common/StringMap.h"
 $10 = ((($this)) + 128|0); //@line 155 "common/StringMap.h"
 HEAP8[$10>>0] = 0; //@line 155 "common/StringMap.h"
 $11 = ((($this)) + 140|0); //@line 155 "common/StringMap.h"
 HEAP8[$11>>0] = 0; //@line 155 "common/StringMap.h"
 $12 = ((($this)) + 152|0); //@line 155 "common/StringMap.h"
 HEAP8[$12>>0] = 0; //@line 155 "common/StringMap.h"
 $13 = ((($this)) + 164|0); //@line 155 "common/StringMap.h"
 HEAP8[$13>>0] = 0; //@line 155 "common/StringMap.h"
 $14 = ((($this)) + 176|0); //@line 155 "common/StringMap.h"
 HEAP8[$14>>0] = 0; //@line 155 "common/StringMap.h"
 $15 = ((($this)) + 188|0); //@line 155 "common/StringMap.h"
 HEAP8[$15>>0] = 0; //@line 155 "common/StringMap.h"
 $16 = ((($this)) + 200|0); //@line 155 "common/StringMap.h"
 HEAP8[$16>>0] = 0; //@line 155 "common/StringMap.h"
 $17 = ((($this)) + 212|0); //@line 155 "common/StringMap.h"
 HEAP8[$17>>0] = 0; //@line 155 "common/StringMap.h"
 $18 = ((($this)) + 224|0); //@line 155 "common/StringMap.h"
 HEAP8[$18>>0] = 0; //@line 155 "common/StringMap.h"
 $19 = ((($this)) + 236|0); //@line 155 "common/StringMap.h"
 HEAP8[$19>>0] = 0; //@line 155 "common/StringMap.h"
 $20 = ((($this)) + 248|0); //@line 155 "common/StringMap.h"
 HEAP8[$20>>0] = 0; //@line 155 "common/StringMap.h"
 $21 = ((($this)) + 260|0); //@line 155 "common/StringMap.h"
 HEAP8[$21>>0] = 0; //@line 155 "common/StringMap.h"
 $22 = ((($this)) + 272|0); //@line 155 "common/StringMap.h"
 HEAP8[$22>>0] = 0; //@line 155 "common/StringMap.h"
 $23 = ((($this)) + 284|0); //@line 155 "common/StringMap.h"
 HEAP8[$23>>0] = 0; //@line 155 "common/StringMap.h"
 $24 = ((($this)) + 296|0); //@line 155 "common/StringMap.h"
 HEAP8[$24>>0] = 0; //@line 155 "common/StringMap.h"
 $25 = ((($this)) + 308|0); //@line 155 "common/StringMap.h"
 HEAP8[$25>>0] = 0; //@line 155 "common/StringMap.h"
 $26 = ((($this)) + 320|0); //@line 155 "common/StringMap.h"
 HEAP8[$26>>0] = 0; //@line 155 "common/StringMap.h"
 $27 = ((($this)) + 332|0); //@line 155 "common/StringMap.h"
 HEAP8[$27>>0] = 0; //@line 155 "common/StringMap.h"
 $28 = ((($this)) + 344|0); //@line 155 "common/StringMap.h"
 HEAP8[$28>>0] = 0; //@line 155 "common/StringMap.h"
 $29 = ((($this)) + 356|0); //@line 155 "common/StringMap.h"
 HEAP8[$29>>0] = 0; //@line 155 "common/StringMap.h"
 $30 = ((($this)) + 368|0); //@line 155 "common/StringMap.h"
 HEAP8[$30>>0] = 0; //@line 155 "common/StringMap.h"
 $31 = ((($this)) + 380|0); //@line 155 "common/StringMap.h"
 HEAP8[$31>>0] = 0; //@line 155 "common/StringMap.h"
 $32 = ((($this)) + 392|0); //@line 155 "common/StringMap.h"
 HEAP8[$32>>0] = 0; //@line 155 "common/StringMap.h"
 $33 = ((($this)) + 404|0); //@line 155 "common/StringMap.h"
 HEAP8[$33>>0] = 0; //@line 155 "common/StringMap.h"
 $34 = ((($this)) + 416|0); //@line 155 "common/StringMap.h"
 HEAP8[$34>>0] = 0; //@line 155 "common/StringMap.h"
 $35 = ((($this)) + 428|0); //@line 155 "common/StringMap.h"
 HEAP8[$35>>0] = 0; //@line 155 "common/StringMap.h"
 $36 = ((($this)) + 440|0); //@line 155 "common/StringMap.h"
 HEAP8[$36>>0] = 0; //@line 155 "common/StringMap.h"
 $37 = ((($this)) + 452|0); //@line 155 "common/StringMap.h"
 HEAP8[$37>>0] = 0; //@line 155 "common/StringMap.h"
 $38 = ((($this)) + 464|0); //@line 155 "common/StringMap.h"
 HEAP8[$38>>0] = 0; //@line 155 "common/StringMap.h"
 $39 = ((($this)) + 476|0); //@line 155 "common/StringMap.h"
 HEAP8[$39>>0] = 0; //@line 155 "common/StringMap.h"
 $40 = ((($this)) + 488|0); //@line 155 "common/StringMap.h"
 HEAP8[$40>>0] = 0; //@line 155 "common/StringMap.h"
 $41 = ((($this)) + 500|0); //@line 155 "common/StringMap.h"
 HEAP8[$41>>0] = 0; //@line 155 "common/StringMap.h"
 $42 = ((($this)) + 512|0); //@line 155 "common/StringMap.h"
 HEAP8[$42>>0] = 0; //@line 155 "common/StringMap.h"
 $43 = ((($this)) + 524|0); //@line 155 "common/StringMap.h"
 HEAP8[$43>>0] = 0; //@line 155 "common/StringMap.h"
 $44 = ((($this)) + 536|0); //@line 155 "common/StringMap.h"
 HEAP8[$44>>0] = 0; //@line 155 "common/StringMap.h"
 $45 = ((($this)) + 548|0); //@line 155 "common/StringMap.h"
 HEAP8[$45>>0] = 0; //@line 155 "common/StringMap.h"
 $46 = ((($this)) + 560|0); //@line 155 "common/StringMap.h"
 HEAP8[$46>>0] = 0; //@line 155 "common/StringMap.h"
 $47 = ((($this)) + 572|0); //@line 155 "common/StringMap.h"
 HEAP8[$47>>0] = 0; //@line 155 "common/StringMap.h"
 $48 = ((($this)) + 584|0); //@line 155 "common/StringMap.h"
 HEAP8[$48>>0] = 0; //@line 155 "common/StringMap.h"
 $49 = ((($this)) + 596|0); //@line 155 "common/StringMap.h"
 HEAP8[$49>>0] = 0; //@line 155 "common/StringMap.h"
 $50 = ((($this)) + 608|0); //@line 155 "common/StringMap.h"
 HEAP8[$50>>0] = 0; //@line 155 "common/StringMap.h"
 $51 = ((($this)) + 620|0); //@line 155 "common/StringMap.h"
 HEAP8[$51>>0] = 0; //@line 155 "common/StringMap.h"
 $52 = ((($this)) + 632|0); //@line 155 "common/StringMap.h"
 HEAP8[$52>>0] = 0; //@line 155 "common/StringMap.h"
 $53 = ((($this)) + 644|0); //@line 155 "common/StringMap.h"
 HEAP8[$53>>0] = 0; //@line 155 "common/StringMap.h"
 $54 = ((($this)) + 656|0); //@line 155 "common/StringMap.h"
 HEAP8[$54>>0] = 0; //@line 155 "common/StringMap.h"
 $55 = ((($this)) + 668|0); //@line 155 "common/StringMap.h"
 HEAP8[$55>>0] = 0; //@line 155 "common/StringMap.h"
 $56 = ((($this)) + 680|0); //@line 155 "common/StringMap.h"
 HEAP8[$56>>0] = 0; //@line 155 "common/StringMap.h"
 $57 = ((($this)) + 692|0); //@line 155 "common/StringMap.h"
 HEAP8[$57>>0] = 0; //@line 155 "common/StringMap.h"
 $58 = ((($this)) + 704|0); //@line 155 "common/StringMap.h"
 HEAP8[$58>>0] = 0; //@line 155 "common/StringMap.h"
 $59 = ((($this)) + 716|0); //@line 155 "common/StringMap.h"
 HEAP8[$59>>0] = 0; //@line 155 "common/StringMap.h"
 $60 = ((($this)) + 728|0); //@line 155 "common/StringMap.h"
 HEAP8[$60>>0] = 0; //@line 155 "common/StringMap.h"
 $61 = ((($this)) + 740|0); //@line 155 "common/StringMap.h"
 HEAP8[$61>>0] = 0; //@line 155 "common/StringMap.h"
 $62 = ((($this)) + 752|0); //@line 155 "common/StringMap.h"
 HEAP8[$62>>0] = 0; //@line 155 "common/StringMap.h"
 $63 = ((($this)) + 764|0); //@line 155 "common/StringMap.h"
 HEAP8[$63>>0] = 0; //@line 155 "common/StringMap.h"
 $64 = ((($this)) + 776|0); //@line 155 "common/StringMap.h"
 HEAP8[$64>>0] = 0; //@line 155 "common/StringMap.h"
 $65 = ((($this)) + 788|0); //@line 155 "common/StringMap.h"
 HEAP8[$65>>0] = 0; //@line 155 "common/StringMap.h"
 $66 = ((($this)) + 800|0); //@line 155 "common/StringMap.h"
 HEAP8[$66>>0] = 0; //@line 155 "common/StringMap.h"
 $67 = ((($this)) + 812|0); //@line 155 "common/StringMap.h"
 HEAP8[$67>>0] = 0; //@line 155 "common/StringMap.h"
 $68 = ((($this)) + 824|0); //@line 155 "common/StringMap.h"
 HEAP8[$68>>0] = 0; //@line 155 "common/StringMap.h"
 $69 = ((($this)) + 836|0); //@line 155 "common/StringMap.h"
 HEAP8[$69>>0] = 0; //@line 155 "common/StringMap.h"
 $70 = ((($this)) + 848|0); //@line 155 "common/StringMap.h"
 HEAP8[$70>>0] = 0; //@line 155 "common/StringMap.h"
 $71 = ((($this)) + 860|0); //@line 155 "common/StringMap.h"
 HEAP8[$71>>0] = 0; //@line 155 "common/StringMap.h"
 $72 = ((($this)) + 872|0); //@line 155 "common/StringMap.h"
 HEAP8[$72>>0] = 0; //@line 155 "common/StringMap.h"
 $73 = ((($this)) + 884|0); //@line 155 "common/StringMap.h"
 HEAP8[$73>>0] = 0; //@line 155 "common/StringMap.h"
 $74 = ((($this)) + 896|0); //@line 155 "common/StringMap.h"
 HEAP8[$74>>0] = 0; //@line 155 "common/StringMap.h"
 $75 = ((($this)) + 908|0); //@line 155 "common/StringMap.h"
 HEAP8[$75>>0] = 0; //@line 155 "common/StringMap.h"
 $76 = ((($this)) + 920|0); //@line 155 "common/StringMap.h"
 HEAP8[$76>>0] = 0; //@line 155 "common/StringMap.h"
 $77 = ((($this)) + 932|0); //@line 155 "common/StringMap.h"
 HEAP8[$77>>0] = 0; //@line 155 "common/StringMap.h"
 $78 = ((($this)) + 944|0); //@line 155 "common/StringMap.h"
 HEAP8[$78>>0] = 0; //@line 155 "common/StringMap.h"
 $79 = ((($this)) + 956|0); //@line 155 "common/StringMap.h"
 HEAP8[$79>>0] = 0; //@line 155 "common/StringMap.h"
 $80 = ((($this)) + 968|0); //@line 155 "common/StringMap.h"
 HEAP8[$80>>0] = 0; //@line 155 "common/StringMap.h"
 $81 = ((($this)) + 980|0); //@line 155 "common/StringMap.h"
 HEAP8[$81>>0] = 0; //@line 155 "common/StringMap.h"
 $82 = ((($this)) + 992|0); //@line 155 "common/StringMap.h"
 HEAP8[$82>>0] = 0; //@line 155 "common/StringMap.h"
 $83 = ((($this)) + 1004|0); //@line 155 "common/StringMap.h"
 HEAP8[$83>>0] = 0; //@line 155 "common/StringMap.h"
 $84 = ((($this)) + 1016|0); //@line 155 "common/StringMap.h"
 HEAP8[$84>>0] = 0; //@line 155 "common/StringMap.h"
 $85 = ((($this)) + 1028|0); //@line 155 "common/StringMap.h"
 HEAP8[$85>>0] = 0; //@line 155 "common/StringMap.h"
 $86 = ((($this)) + 1040|0); //@line 155 "common/StringMap.h"
 HEAP8[$86>>0] = 0; //@line 155 "common/StringMap.h"
 $87 = ((($this)) + 1052|0); //@line 155 "common/StringMap.h"
 HEAP8[$87>>0] = 0; //@line 155 "common/StringMap.h"
 $88 = ((($this)) + 1064|0); //@line 155 "common/StringMap.h"
 HEAP8[$88>>0] = 0; //@line 155 "common/StringMap.h"
 $89 = ((($this)) + 1076|0); //@line 155 "common/StringMap.h"
 HEAP8[$89>>0] = 0; //@line 155 "common/StringMap.h"
 $90 = ((($this)) + 1088|0); //@line 155 "common/StringMap.h"
 HEAP8[$90>>0] = 0; //@line 155 "common/StringMap.h"
 $91 = ((($this)) + 1100|0); //@line 155 "common/StringMap.h"
 HEAP8[$91>>0] = 0; //@line 155 "common/StringMap.h"
 $92 = ((($this)) + 1112|0); //@line 155 "common/StringMap.h"
 HEAP8[$92>>0] = 0; //@line 155 "common/StringMap.h"
 $93 = ((($this)) + 1124|0); //@line 155 "common/StringMap.h"
 HEAP8[$93>>0] = 0; //@line 155 "common/StringMap.h"
 $94 = ((($this)) + 1136|0); //@line 155 "common/StringMap.h"
 HEAP8[$94>>0] = 0; //@line 155 "common/StringMap.h"
 $95 = ((($this)) + 1148|0); //@line 155 "common/StringMap.h"
 HEAP8[$95>>0] = 0; //@line 155 "common/StringMap.h"
 $96 = ((($this)) + 1160|0); //@line 155 "common/StringMap.h"
 HEAP8[$96>>0] = 0; //@line 155 "common/StringMap.h"
 $97 = ((($this)) + 1172|0); //@line 155 "common/StringMap.h"
 HEAP8[$97>>0] = 0; //@line 155 "common/StringMap.h"
 $98 = ((($this)) + 1184|0); //@line 155 "common/StringMap.h"
 HEAP8[$98>>0] = 0; //@line 155 "common/StringMap.h"
 $99 = ((($this)) + 1196|0); //@line 155 "common/StringMap.h"
 HEAP8[$99>>0] = 0; //@line 155 "common/StringMap.h"
 $100 = ((($this)) + 1208|0); //@line 155 "common/StringMap.h"
 HEAP8[$100>>0] = 0; //@line 155 "common/StringMap.h"
 $101 = ((($this)) + 1220|0); //@line 155 "common/StringMap.h"
 HEAP8[$101>>0] = 0; //@line 155 "common/StringMap.h"
 $102 = ((($this)) + 1232|0); //@line 155 "common/StringMap.h"
 HEAP8[$102>>0] = 0; //@line 155 "common/StringMap.h"
 $103 = ((($this)) + 1244|0); //@line 155 "common/StringMap.h"
 HEAP8[$103>>0] = 0; //@line 155 "common/StringMap.h"
 $104 = ((($this)) + 1256|0); //@line 155 "common/StringMap.h"
 HEAP8[$104>>0] = 0; //@line 155 "common/StringMap.h"
 $105 = ((($this)) + 1268|0); //@line 155 "common/StringMap.h"
 HEAP8[$105>>0] = 0; //@line 155 "common/StringMap.h"
 $106 = ((($this)) + 1280|0); //@line 155 "common/StringMap.h"
 HEAP8[$106>>0] = 0; //@line 155 "common/StringMap.h"
 $107 = ((($this)) + 1292|0); //@line 155 "common/StringMap.h"
 HEAP8[$107>>0] = 0; //@line 155 "common/StringMap.h"
 $108 = ((($this)) + 1304|0); //@line 155 "common/StringMap.h"
 HEAP8[$108>>0] = 0; //@line 155 "common/StringMap.h"
 $109 = ((($this)) + 1316|0); //@line 155 "common/StringMap.h"
 HEAP8[$109>>0] = 0; //@line 155 "common/StringMap.h"
 $110 = ((($this)) + 1328|0); //@line 155 "common/StringMap.h"
 HEAP8[$110>>0] = 0; //@line 155 "common/StringMap.h"
 $111 = ((($this)) + 1340|0); //@line 155 "common/StringMap.h"
 HEAP8[$111>>0] = 0; //@line 155 "common/StringMap.h"
 $112 = ((($this)) + 1352|0); //@line 155 "common/StringMap.h"
 HEAP8[$112>>0] = 0; //@line 155 "common/StringMap.h"
 $113 = ((($this)) + 1364|0); //@line 155 "common/StringMap.h"
 HEAP8[$113>>0] = 0; //@line 155 "common/StringMap.h"
 $114 = ((($this)) + 1376|0); //@line 155 "common/StringMap.h"
 HEAP8[$114>>0] = 0; //@line 155 "common/StringMap.h"
 $115 = ((($this)) + 1388|0); //@line 155 "common/StringMap.h"
 HEAP8[$115>>0] = 0; //@line 155 "common/StringMap.h"
 $116 = ((($this)) + 1400|0); //@line 155 "common/StringMap.h"
 HEAP8[$116>>0] = 0; //@line 155 "common/StringMap.h"
 $117 = ((($this)) + 1412|0); //@line 155 "common/StringMap.h"
 HEAP8[$117>>0] = 0; //@line 155 "common/StringMap.h"
 $118 = ((($this)) + 1424|0); //@line 155 "common/StringMap.h"
 HEAP8[$118>>0] = 0; //@line 155 "common/StringMap.h"
 $119 = ((($this)) + 1436|0); //@line 155 "common/StringMap.h"
 HEAP8[$119>>0] = 0; //@line 155 "common/StringMap.h"
 $scevgep = ((($this)) + 1440|0); //@line 44 "common/StringMap.h"
 _memset(($scevgep|0),0,240)|0; //@line 44 "common/StringMap.h"
 $120 = $num >>> 3; //@line 46 "common/StringMap.h"
 $121 = ($120|0)==(0); //@line 48 "common/StringMap.h"
 if ($121) {
  STACKTOP = sp;return; //@line 50 "common/StringMap.h"
 } else {
  $i1$05 = 0;
 }
 while(1) {
  $122 = (($entries) + ($i1$05<<3)|0); //@line 49 "common/StringMap.h"
  $123 = HEAP32[$122>>2]|0; //@line 49 "common/StringMap.h"
  $124 = (((($entries) + ($i1$05<<3)|0)) + 4|0); //@line 49 "common/StringMap.h"
  $125 = HEAP32[$124>>2]|0; //@line 49 "common/StringMap.h"
  $126 = HEAP8[$123>>0]|0; //@line 142 "common/StringMap.h"
  $127 = ($126<<24>>24)==(0); //@line 142 "common/StringMap.h"
  if ($127) {
   $hash$0$lcssa$i$i = 5381;
  } else {
   $$02$i$i = $123;$129 = $126;$hash$01$i$i = 5381;
   while(1) {
    $128 = $129 << 24 >> 24; //@line 142 "common/StringMap.h"
    $130 = ((($$02$i$i)) + 1|0); //@line 142 "common/StringMap.h"
    $131 = ($hash$01$i$i*33)|0; //@line 143 "common/StringMap.h"
    $132 = (($131) + ($128))|0; //@line 143 "common/StringMap.h"
    $133 = HEAP8[$130>>0]|0; //@line 142 "common/StringMap.h"
    $134 = ($133<<24>>24)==(0); //@line 142 "common/StringMap.h"
    if ($134) {
     $hash$0$lcssa$i$i = $132;
     break;
    } else {
     $$02$i$i = $130;$129 = $133;$hash$01$i$i = $132;
    }
   }
  }
  $i$08$i = 0;
  while(1) {
   $137 = (($i$08$i) + ($hash$0$lcssa$i$i))|0; //@line 112 "common/StringMap.h"
   $138 = (($137>>>0) % 120)&-1; //@line 112 "common/StringMap.h"
   $139 = (((($this) + (($138*12)|0)|0)) + 8|0); //@line 114 "common/StringMap.h"
   $140 = HEAP8[$139>>0]|0; //@line 114 "common/StringMap.h"
   $141 = ($140<<24>>24)==(0); //@line 114 "common/StringMap.h"
   $136 = (($i$08$i) + 1)|0; //@line 110 "common/StringMap.h"
   if ($141) {
    $$lcssa26 = $138;$$lcssa28 = $139;
    label = 8;
    break;
   }
   $135 = ($136>>>0)<(120); //@line 110 "common/StringMap.h"
   if ($135) {
    $i$08$i = $136;
   } else {
    break;
   }
  }
  if ((label|0) == 8) {
   label = 0;
   HEAP8[$$lcssa28>>0] = 1; //@line 117 "common/StringMap.h"
   $142 = (($this) + (($$lcssa26*12)|0)|0); //@line 118 "common/StringMap.h"
   HEAP32[$142>>2] = $123; //@line 118 "common/StringMap.h"
   $143 = (((($this) + (($$lcssa26*12)|0)|0)) + 4|0); //@line 119 "common/StringMap.h"
   HEAP32[$143>>2] = $125; //@line 119 "common/StringMap.h"
  }
  $144 = ($125>>>0)>(59); //@line 126 "common/StringMap.h"
  if ($144) {
   HEAP32[$vararg_buffer>>2] = $123; //@line 128 "common/StringMap.h"
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "common/StringMap.h"
   HEAP32[$vararg_ptr1>>2] = $125; //@line 128 "common/StringMap.h"
   (_printf(443858,$vararg_buffer)|0); //@line 128 "common/StringMap.h"
  } else {
   $145 = (((($this)) + 1440|0) + ($125<<2)|0); //@line 132 "common/StringMap.h"
   HEAP32[$145>>2] = $123; //@line 132 "common/StringMap.h"
  }
  $146 = (($i1$05) + 1)|0; //@line 48 "common/StringMap.h"
  $exitcond = ($146|0)==($120|0); //@line 48 "common/StringMap.h"
  if ($exitcond) {
   break;
  } else {
   $i1$05 = $146;
  }
 }
 STACKTOP = sp;return; //@line 50 "common/StringMap.h"
}
function __ZN4love11addTypeNameENS_4TypeEPKc($type,$name) {
 $type = $type|0;
 $name = $name|0;
 var $$02$i$i = 0, $$lcssa = 0, $$lcssa23 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$08$i = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 126 "common/types.cpp"
 $0 = ($type>>>0)>(59); //@line 91 "common/StringMap.h"
 if (!($0)) {
  $1 = (627460 + ($type<<2)|0); //@line 94 "common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "common/StringMap.h"
  if (!($3)) {
   STACKTOP = sp;return; //@line 131 "common/types.cpp"
  }
 }
 $4 = HEAP8[$name>>0]|0; //@line 142 "common/StringMap.h"
 $5 = ($4<<24>>24)==(0); //@line 142 "common/StringMap.h"
 if ($5) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $name;$7 = $4;$hash$01$i$i = 5381;
  while(1) {
   $6 = $7 << 24 >> 24; //@line 142 "common/StringMap.h"
   $8 = ((($$02$i$i)) + 1|0); //@line 142 "common/StringMap.h"
   $9 = ($hash$01$i$i*33)|0; //@line 143 "common/StringMap.h"
   $10 = (($9) + ($6))|0; //@line 143 "common/StringMap.h"
   $11 = HEAP8[$8>>0]|0; //@line 142 "common/StringMap.h"
   $12 = ($11<<24>>24)==(0); //@line 142 "common/StringMap.h"
   if ($12) {
    $hash$0$lcssa$i$i = $10;
    break;
   } else {
    $$02$i$i = $8;$7 = $11;$hash$01$i$i = $10;
   }
  }
 }
 $i$08$i = 0;
 while(1) {
  $15 = (($i$08$i) + ($hash$0$lcssa$i$i))|0; //@line 112 "common/StringMap.h"
  $16 = (($15>>>0) % 120)&-1; //@line 112 "common/StringMap.h"
  $17 = (((626020 + (($16*12)|0)|0)) + 8|0); //@line 114 "common/StringMap.h"
  $18 = HEAP8[$17>>0]|0; //@line 114 "common/StringMap.h"
  $19 = ($18<<24>>24)==(0); //@line 114 "common/StringMap.h"
  $14 = (($i$08$i) + 1)|0; //@line 110 "common/StringMap.h"
  if ($19) {
   $$lcssa = $16;$$lcssa23 = $17;
   label = 8;
   break;
  }
  $13 = ($14>>>0)<(120); //@line 110 "common/StringMap.h"
  if ($13) {
   $i$08$i = $14;
  } else {
   break;
  }
 }
 if ((label|0) == 8) {
  HEAP8[$$lcssa23>>0] = 1; //@line 117 "common/StringMap.h"
  $20 = (626020 + (($$lcssa*12)|0)|0); //@line 118 "common/StringMap.h"
  HEAP32[$20>>2] = $name; //@line 118 "common/StringMap.h"
  $21 = (((626020 + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "common/StringMap.h"
  HEAP32[$21>>2] = $type; //@line 119 "common/StringMap.h"
 }
 if ($0) {
  HEAP32[$vararg_buffer>>2] = $name; //@line 128 "common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $type; //@line 128 "common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "common/StringMap.h"
  STACKTOP = sp;return; //@line 131 "common/types.cpp"
 } else {
  $22 = (627460 + ($type<<2)|0); //@line 132 "common/StringMap.h"
  HEAP32[$22>>2] = $name; //@line 132 "common/StringMap.h"
  STACKTOP = sp;return; //@line 131 "common/types.cpp"
 }
}
function __ZN4love11getTypeNameEPKcRNS_4TypeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "common/StringMap.h"
  $10 = (($9>>>0) % 120)&-1; //@line 72 "common/StringMap.h"
  $11 = (((626020 + (($10*12)|0)|0)) + 8|0); //@line 74 "common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (626020 + (($10*12)|0)|0); //@line 77 "common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "common/StringMap.h"
  $31 = ($30>>>0)<(120); //@line 70 "common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((626020 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 135 "common/types.cpp"
  return ($$$2$i|0); //@line 135 "common/types.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 135 "common/types.cpp"
  return ($$$2$i|0); //@line 135 "common/types.cpp"
 }
 return (0)|0;
}
function __ZN4love11getTypeNameENS_4TypeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(59); //@line 91 "common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (627460 + ($in<<2)|0); //@line 94 "common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 140 "common/types.cpp"
}
function __ZN4love16w_Data_getStringEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4DataEEEPT_P9lua_StateiNS_4TypeE($L,1,2)|0); //@line 28 "common/wrap_Data.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 34 "common/wrap_Data.cpp"
 $2 = ((($1)) + 16|0); //@line 34 "common/wrap_Data.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 34 "common/wrap_Data.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 34 "common/wrap_Data.cpp"
 $5 = HEAP32[$0>>2]|0; //@line 34 "common/wrap_Data.cpp"
 $6 = ((($5)) + 20|0); //@line 34 "common/wrap_Data.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 34 "common/wrap_Data.cpp"
 $8 = (FUNCTION_TABLE_ii[$7 & 8191]($0)|0); //@line 34 "common/wrap_Data.cpp"
 _lua_pushlstring($L,$4,$8); //@line 34 "common/wrap_Data.cpp"
 return 1; //@line 35 "common/wrap_Data.cpp"
}
function __ZN4love17w_Data_getPointerEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4DataEEEPT_P9lua_StateiNS_4TypeE($L,1,2)|0); //@line 28 "common/wrap_Data.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 41 "common/wrap_Data.cpp"
 $2 = ((($1)) + 16|0); //@line 41 "common/wrap_Data.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 41 "common/wrap_Data.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 41 "common/wrap_Data.cpp"
 _lua_pushlightuserdata($L,$4); //@line 41 "common/wrap_Data.cpp"
 return 1; //@line 42 "common/wrap_Data.cpp"
}
function __ZN4love14w_Data_getSizeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4DataEEEPT_P9lua_StateiNS_4TypeE($L,1,2)|0); //@line 28 "common/wrap_Data.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 48 "common/wrap_Data.cpp"
 $2 = ((($1)) + 20|0); //@line 48 "common/wrap_Data.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 48 "common/wrap_Data.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 48 "common/wrap_Data.cpp"
 $5 = (+($4>>>0)); //@line 48 "common/wrap_Data.cpp"
 _lua_pushnumber($L,$5); //@line 48 "common/wrap_Data.cpp"
 return 1; //@line 49 "common/wrap_Data.cpp"
}
function __ZN4love14luax_checkdataEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4DataEEEPT_P9lua_StateiNS_4TypeE($L,$idx,2)|0); //@line 28 "common/wrap_Data.cpp"
 return ($0|0); //@line 28 "common/wrap_Data.cpp"
}
function __ZN4love11w_Data_openEP9lua_State($L) {
 $L = $L|0;
 var $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 60 "common/wrap_Data.cpp"
 HEAP32[$vararg_buffer>>2] = 12272; //@line 62 "common/wrap_Data.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 62 "common/wrap_Data.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 62 "common/wrap_Data.cpp"
 (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,2,327500,$vararg_buffer)|0); //@line 62 "common/wrap_Data.cpp"
 STACKTOP = sp;return 0; //@line 63 "common/wrap_Data.cpp"
}
function __ZN4love5audio16w_getSourceCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 41 "modules/audio/wrap_Audio.cpp"
 $2 = ((($1)) + 32|0); //@line 41 "modules/audio/wrap_Audio.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 41 "modules/audio/wrap_Audio.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 41 "modules/audio/wrap_Audio.cpp"
 _lua_pushinteger($L,$4); //@line 41 "modules/audio/wrap_Audio.cpp"
 return 1; //@line 42 "modules/audio/wrap_Audio.cpp"
}
function __ZN4love5audio11w_newSourceEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $$1 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, $stype = 0, $t$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $stype = sp + 12|0;
 $0 = (_lua_isstring($L,1)|0); //@line 47 "modules/audio/wrap_Audio.cpp"
 $1 = ($0|0)==(0); //@line 47 "modules/audio/wrap_Audio.cpp"
 if ($1) {
  $2 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,1,5)|0); //@line 47 "modules/audio/wrap_Audio.cpp"
  if ($2) {
   label = 4;
  } else {
   $3 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,1,7)|0); //@line 47 "modules/audio/wrap_Audio.cpp"
   if ($3) {
    label = 4;
   }
  }
 } else {
  label = 4;
 }
 if ((label|0) == 4) {
  (__ZN4love12luax_convobjEP9lua_StateiPKcS3_($L,1,327505,327511)|0); //@line 48 "modules/audio/wrap_Audio.cpp"
 }
 HEAP32[$stype>>2] = 1; //@line 50 "modules/audio/wrap_Audio.cpp"
 $4 = (_lua_type($L,2)|0); //@line 52 "modules/audio/wrap_Audio.cpp"
 $5 = ($4|0)<(1); //@line 52 "modules/audio/wrap_Audio.cpp"
 if (!($5)) {
  $6 = (_lua_tolstring($L,2,0)|0); //@line 52 "modules/audio/wrap_Audio.cpp"
  $7 = ($6|0)==(0|0); //@line 53 "modules/audio/wrap_Audio.cpp"
  if (!($7)) {
   $8 = (__ZN4love5audio6Source11getConstantEPKcRNS1_4TypeE($6,$stype)|0); //@line 53 "modules/audio/wrap_Audio.cpp"
   if (!($8)) {
    HEAP32[$vararg_buffer>>2] = $6; //@line 54 "modules/audio/wrap_Audio.cpp"
    $9 = (_luaL_error($L,327522,$vararg_buffer)|0); //@line 54 "modules/audio/wrap_Audio.cpp"
    $$1 = $9;
    STACKTOP = sp;return ($$1|0); //@line 76 "modules/audio/wrap_Audio.cpp"
   }
   $$pre = HEAP32[$stype>>2]|0; //@line 56 "modules/audio/wrap_Audio.cpp"
   $phitmp = ($$pre|0)==(0); //@line 53 "modules/audio/wrap_Audio.cpp"
   if ($phitmp) {
    $10 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,1,30)|0); //@line 56 "modules/audio/wrap_Audio.cpp"
    if ($10) {
     (__ZN4love12luax_convobjEP9lua_StateiPKcS3_($L,1,327505,327546)|0); //@line 57 "modules/audio/wrap_Audio.cpp"
    }
   }
  }
 }
 __THREW__ = 0;
 $11 = (invoke_iiii(3458,($L|0),1,29)|0); //@line 62 "modules/audio/wrap_Audio.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 do {
  if ($13) {
   label = 20;
  } else {
   if ($11) {
    $14 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
    $15 = HEAP32[$14>>2]|0; //@line 63 "modules/audio/wrap_Audio.cpp"
    $16 = ((($15)) + 28|0); //@line 63 "modules/audio/wrap_Audio.cpp"
    $17 = HEAP32[$16>>2]|0; //@line 63 "modules/audio/wrap_Audio.cpp"
    __THREW__ = 0;
    $18 = (invoke_iii(3499,($L|0),1)|0); //@line 516 "./common/runtime.h"
    $19 = __THREW__; __THREW__ = 0;
    $20 = $19&1;
    if ($20) {
     label = 20;
     break;
    }
    $21 = ((($18)) + 4|0); //@line 516 "./common/runtime.h"
    $22 = HEAP32[$21>>2]|0; //@line 516 "./common/runtime.h"
    __THREW__ = 0;
    $23 = (invoke_iii($17|0,($14|0),($22|0))|0); //@line 63 "modules/audio/wrap_Audio.cpp"
    $24 = __THREW__; __THREW__ = 0;
    $25 = $24&1;
    if ($25) {
     label = 20;
     break;
    } else {
     $t$0 = $23;
    }
   } else {
    __THREW__ = 0;
    $26 = (invoke_iiii(3458,($L|0),1,30)|0); //@line 64 "modules/audio/wrap_Audio.cpp"
    $27 = __THREW__; __THREW__ = 0;
    $28 = $27&1;
    if ($28) {
     label = 20;
     break;
    }
    if (!($26)) {
     break;
    }
    $29 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
    $30 = HEAP32[$29>>2]|0; //@line 65 "modules/audio/wrap_Audio.cpp"
    $31 = ((($30)) + 24|0); //@line 65 "modules/audio/wrap_Audio.cpp"
    $32 = HEAP32[$31>>2]|0; //@line 65 "modules/audio/wrap_Audio.cpp"
    __THREW__ = 0;
    $33 = (invoke_iii(3499,($L|0),1)|0); //@line 516 "./common/runtime.h"
    $34 = __THREW__; __THREW__ = 0;
    $35 = $34&1;
    if ($35) {
     label = 20;
     break;
    }
    $36 = ((($33)) + 4|0); //@line 516 "./common/runtime.h"
    $37 = HEAP32[$36>>2]|0; //@line 516 "./common/runtime.h"
    __THREW__ = 0;
    $38 = (invoke_iii($32|0,($29|0),($37|0))|0); //@line 65 "modules/audio/wrap_Audio.cpp"
    $39 = __THREW__; __THREW__ = 0;
    $40 = $39&1;
    if ($40) {
     label = 20;
     break;
    } else {
     $t$0 = $38;
    }
   }
   $59 = ($t$0|0)==(0|0); //@line 68 "modules/audio/wrap_Audio.cpp"
   if (!($59)) {
    __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,28,$t$0); //@line 70 "modules/audio/wrap_Audio.cpp"
    $60 = HEAP32[$t$0>>2]|0; //@line 71 "modules/audio/wrap_Audio.cpp"
    $61 = ((($60)) + 12|0); //@line 71 "modules/audio/wrap_Audio.cpp"
    $62 = HEAP32[$61>>2]|0; //@line 71 "modules/audio/wrap_Audio.cpp"
    FUNCTION_TABLE_vi[$62 & 8191]($t$0); //@line 71 "modules/audio/wrap_Audio.cpp"
    $$1 = 1;
    STACKTOP = sp;return ($$1|0); //@line 76 "modules/audio/wrap_Audio.cpp"
   }
  }
 } while(0);
 do {
  if ((label|0) == 20) {
   $41 = ___cxa_find_matching_catch_3(8992|0)|0;
   $42 = tempRet0;
   $43 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $44 = ($42|0)==($43|0); //@line 535 "./common/runtime.h"
   if (!($44)) {
    $$01$i = $42;$$02$i = $41;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $45 = (___cxa_begin_catch(($41|0))|0); //@line 535 "./common/runtime.h"
   $46 = HEAP32[$45>>2]|0; //@line 539 "./common/runtime.h"
   $47 = ((($46)) + 8|0); //@line 539 "./common/runtime.h"
   $48 = HEAP32[$47>>2]|0; //@line 539 "./common/runtime.h"
   $49 = (FUNCTION_TABLE_ii[$48 & 8191]($45)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($49|0)); //@line 539 "./common/runtime.h"
   $50 = __THREW__; __THREW__ = 0;
   $51 = $50&1;
   if (!($51)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $52 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer1>>2] = $52; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
    break;
   }
   $53 = ___cxa_find_matching_catch_2()|0;
   $54 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $55 = __THREW__; __THREW__ = 0;
   $56 = $55&1;
   if ($56) {
    $57 = ___cxa_find_matching_catch_3(0|0)|0;
    $58 = tempRet0;
    ___clang_call_terminate($57); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $54;$$02$i = $53;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  }
 } while(0);
 $63 = (_luax_typerror($L,1,327559)|0); //@line 75 "modules/audio/wrap_Audio.cpp"
 $$1 = $63;
 STACKTOP = sp;return ($$1|0); //@line 76 "modules/audio/wrap_Audio.cpp"
}
function __ZN4love5audio6w_playEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love5audio16luax_checksourceEP9lua_Statei($L,1)|0); //@line 80 "modules/audio/wrap_Audio.cpp"
 $1 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 81 "modules/audio/wrap_Audio.cpp"
 $3 = ((($2)) + 44|0); //@line 81 "modules/audio/wrap_Audio.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 81 "modules/audio/wrap_Audio.cpp"
 $5 = (FUNCTION_TABLE_iii[$4 & 8191]($1,$0)|0); //@line 81 "modules/audio/wrap_Audio.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$5); //@line 81 "modules/audio/wrap_Audio.cpp"
 return 1; //@line 82 "modules/audio/wrap_Audio.cpp"
}
function __ZN4love5audio6w_stopEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_gettop($L)|0); //@line 87 "modules/audio/wrap_Audio.cpp"
 $1 = ($0|0)==(0); //@line 87 "modules/audio/wrap_Audio.cpp"
 if ($1) {
  $2 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
  $3 = HEAP32[$2>>2]|0; //@line 89 "modules/audio/wrap_Audio.cpp"
  $4 = ((($3)) + 52|0); //@line 89 "modules/audio/wrap_Audio.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 89 "modules/audio/wrap_Audio.cpp"
  FUNCTION_TABLE_vi[$5 & 8191]($2); //@line 89 "modules/audio/wrap_Audio.cpp"
  return 0; //@line 96 "modules/audio/wrap_Audio.cpp"
 } else {
  $6 = (__ZN4love5audio16luax_checksourceEP9lua_Statei($L,1)|0); //@line 93 "modules/audio/wrap_Audio.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 94 "modules/audio/wrap_Audio.cpp"
  $8 = ((($7)) + 24|0); //@line 94 "modules/audio/wrap_Audio.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 94 "modules/audio/wrap_Audio.cpp"
  FUNCTION_TABLE_vi[$9 & 8191]($6); //@line 94 "modules/audio/wrap_Audio.cpp"
  return 0; //@line 96 "modules/audio/wrap_Audio.cpp"
 }
 return (0)|0;
}
function __ZN4love5audio7w_pauseEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_gettop($L)|0); //@line 101 "modules/audio/wrap_Audio.cpp"
 $1 = ($0|0)==(0); //@line 101 "modules/audio/wrap_Audio.cpp"
 if ($1) {
  $2 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
  $3 = HEAP32[$2>>2]|0; //@line 103 "modules/audio/wrap_Audio.cpp"
  $4 = ((($3)) + 60|0); //@line 103 "modules/audio/wrap_Audio.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 103 "modules/audio/wrap_Audio.cpp"
  FUNCTION_TABLE_vi[$5 & 8191]($2); //@line 103 "modules/audio/wrap_Audio.cpp"
  return 0; //@line 111 "modules/audio/wrap_Audio.cpp"
 } else {
  $6 = (__ZN4love5audio16luax_checksourceEP9lua_Statei($L,1)|0); //@line 107 "modules/audio/wrap_Audio.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 108 "modules/audio/wrap_Audio.cpp"
  $8 = ((($7)) + 28|0); //@line 108 "modules/audio/wrap_Audio.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 108 "modules/audio/wrap_Audio.cpp"
  FUNCTION_TABLE_vi[$9 & 8191]($6); //@line 108 "modules/audio/wrap_Audio.cpp"
  return 0; //@line 111 "modules/audio/wrap_Audio.cpp"
 }
 return (0)|0;
}
function __ZN4love5audio8w_resumeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_gettop($L)|0); //@line 116 "modules/audio/wrap_Audio.cpp"
 $1 = ($0|0)==(0); //@line 116 "modules/audio/wrap_Audio.cpp"
 if ($1) {
  $2 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
  $3 = HEAP32[$2>>2]|0; //@line 118 "modules/audio/wrap_Audio.cpp"
  $4 = ((($3)) + 68|0); //@line 118 "modules/audio/wrap_Audio.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 118 "modules/audio/wrap_Audio.cpp"
  FUNCTION_TABLE_vi[$5 & 8191]($2); //@line 118 "modules/audio/wrap_Audio.cpp"
  return 0; //@line 125 "modules/audio/wrap_Audio.cpp"
 } else {
  $6 = (__ZN4love5audio16luax_checksourceEP9lua_Statei($L,1)|0); //@line 122 "modules/audio/wrap_Audio.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 123 "modules/audio/wrap_Audio.cpp"
  $8 = ((($7)) + 32|0); //@line 123 "modules/audio/wrap_Audio.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 123 "modules/audio/wrap_Audio.cpp"
  FUNCTION_TABLE_vi[$9 & 8191]($6); //@line 123 "modules/audio/wrap_Audio.cpp"
  return 0; //@line 125 "modules/audio/wrap_Audio.cpp"
 }
 return (0)|0;
}
function __ZN4love5audio8w_rewindEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_gettop($L)|0); //@line 130 "modules/audio/wrap_Audio.cpp"
 $1 = ($0|0)==(0); //@line 130 "modules/audio/wrap_Audio.cpp"
 if ($1) {
  $2 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
  $3 = HEAP32[$2>>2]|0; //@line 132 "modules/audio/wrap_Audio.cpp"
  $4 = ((($3)) + 76|0); //@line 132 "modules/audio/wrap_Audio.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 132 "modules/audio/wrap_Audio.cpp"
  FUNCTION_TABLE_vi[$5 & 8191]($2); //@line 132 "modules/audio/wrap_Audio.cpp"
  return 0; //@line 139 "modules/audio/wrap_Audio.cpp"
 } else {
  $6 = (__ZN4love5audio16luax_checksourceEP9lua_Statei($L,1)|0); //@line 136 "modules/audio/wrap_Audio.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 137 "modules/audio/wrap_Audio.cpp"
  $8 = ((($7)) + 36|0); //@line 137 "modules/audio/wrap_Audio.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 137 "modules/audio/wrap_Audio.cpp"
  FUNCTION_TABLE_vi[$9 & 8191]($6); //@line 137 "modules/audio/wrap_Audio.cpp"
  return 0; //@line 139 "modules/audio/wrap_Audio.cpp"
 }
 return (0)|0;
}
function __ZN4love5audio11w_setVolumeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1)); //@line 144 "modules/audio/wrap_Audio.cpp"
 $1 = $0; //@line 144 "modules/audio/wrap_Audio.cpp"
 $2 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
 $3 = HEAP32[$2>>2]|0; //@line 145 "modules/audio/wrap_Audio.cpp"
 $4 = ((($3)) + 80|0); //@line 145 "modules/audio/wrap_Audio.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 145 "modules/audio/wrap_Audio.cpp"
 FUNCTION_TABLE_vid[$5 & 8191]($2,$1); //@line 145 "modules/audio/wrap_Audio.cpp"
 return 0; //@line 146 "modules/audio/wrap_Audio.cpp"
}
function __ZN4love5audio11w_getVolumeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 151 "modules/audio/wrap_Audio.cpp"
 $2 = ((($1)) + 84|0); //@line 151 "modules/audio/wrap_Audio.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 151 "modules/audio/wrap_Audio.cpp"
 $4 = (+FUNCTION_TABLE_di[$3 & 4095]($0)); //@line 151 "modules/audio/wrap_Audio.cpp"
 $5 = $4; //@line 151 "modules/audio/wrap_Audio.cpp"
 _lua_pushnumber($L,$5); //@line 151 "modules/audio/wrap_Audio.cpp"
 return 1; //@line 152 "modules/audio/wrap_Audio.cpp"
}
function __ZN4love5audio13w_setPositionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, $10 = 0, $11 = 0, $2 = 0.0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp;
 $0 = (+_luaL_checknumber($L,1)); //@line 158 "modules/audio/wrap_Audio.cpp"
 $1 = $0; //@line 158 "modules/audio/wrap_Audio.cpp"
 HEAPF32[$v>>2] = $1; //@line 158 "modules/audio/wrap_Audio.cpp"
 $2 = (+_luaL_checknumber($L,2)); //@line 159 "modules/audio/wrap_Audio.cpp"
 $3 = $2; //@line 159 "modules/audio/wrap_Audio.cpp"
 $4 = ((($v)) + 4|0); //@line 159 "modules/audio/wrap_Audio.cpp"
 HEAPF32[$4>>2] = $3; //@line 159 "modules/audio/wrap_Audio.cpp"
 $5 = (+_luaL_optnumber($L,3,0.0)); //@line 160 "modules/audio/wrap_Audio.cpp"
 $6 = $5; //@line 160 "modules/audio/wrap_Audio.cpp"
 $7 = ((($v)) + 8|0); //@line 160 "modules/audio/wrap_Audio.cpp"
 HEAPF32[$7>>2] = $6; //@line 160 "modules/audio/wrap_Audio.cpp"
 $8 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
 $9 = HEAP32[$8>>2]|0; //@line 161 "modules/audio/wrap_Audio.cpp"
 $10 = ((($9)) + 92|0); //@line 161 "modules/audio/wrap_Audio.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 161 "modules/audio/wrap_Audio.cpp"
 FUNCTION_TABLE_vii[$11 & 8191]($8,$v); //@line 161 "modules/audio/wrap_Audio.cpp"
 STACKTOP = sp;return 0; //@line 162 "modules/audio/wrap_Audio.cpp"
}
function __ZN4love5audio13w_getPositionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp;
 $0 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 168 "modules/audio/wrap_Audio.cpp"
 $2 = ((($1)) + 88|0); //@line 168 "modules/audio/wrap_Audio.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 168 "modules/audio/wrap_Audio.cpp"
 FUNCTION_TABLE_vii[$3 & 8191]($0,$v); //@line 168 "modules/audio/wrap_Audio.cpp"
 $4 = +HEAPF32[$v>>2]; //@line 169 "modules/audio/wrap_Audio.cpp"
 $5 = $4; //@line 169 "modules/audio/wrap_Audio.cpp"
 _lua_pushnumber($L,$5); //@line 169 "modules/audio/wrap_Audio.cpp"
 $6 = ((($v)) + 4|0); //@line 170 "modules/audio/wrap_Audio.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 170 "modules/audio/wrap_Audio.cpp"
 $8 = $7; //@line 170 "modules/audio/wrap_Audio.cpp"
 _lua_pushnumber($L,$8); //@line 170 "modules/audio/wrap_Audio.cpp"
 $9 = ((($v)) + 8|0); //@line 171 "modules/audio/wrap_Audio.cpp"
 $10 = +HEAPF32[$9>>2]; //@line 171 "modules/audio/wrap_Audio.cpp"
 $11 = $10; //@line 171 "modules/audio/wrap_Audio.cpp"
 _lua_pushnumber($L,$11); //@line 171 "modules/audio/wrap_Audio.cpp"
 STACKTOP = sp;return 3; //@line 172 "modules/audio/wrap_Audio.cpp"
}
function __ZN4love5audio16w_setOrientationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0.0;
 var $9 = 0.0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp;
 $0 = (+_luaL_checknumber($L,1)); //@line 178 "modules/audio/wrap_Audio.cpp"
 $1 = $0; //@line 178 "modules/audio/wrap_Audio.cpp"
 HEAPF32[$v>>2] = $1; //@line 178 "modules/audio/wrap_Audio.cpp"
 $2 = (+_luaL_checknumber($L,2)); //@line 179 "modules/audio/wrap_Audio.cpp"
 $3 = $2; //@line 179 "modules/audio/wrap_Audio.cpp"
 $4 = ((($v)) + 4|0); //@line 179 "modules/audio/wrap_Audio.cpp"
 HEAPF32[$4>>2] = $3; //@line 179 "modules/audio/wrap_Audio.cpp"
 $5 = (+_luaL_checknumber($L,3)); //@line 180 "modules/audio/wrap_Audio.cpp"
 $6 = $5; //@line 180 "modules/audio/wrap_Audio.cpp"
 $7 = ((($v)) + 8|0); //@line 180 "modules/audio/wrap_Audio.cpp"
 HEAPF32[$7>>2] = $6; //@line 180 "modules/audio/wrap_Audio.cpp"
 $8 = (+_luaL_checknumber($L,4)); //@line 181 "modules/audio/wrap_Audio.cpp"
 $9 = $8; //@line 181 "modules/audio/wrap_Audio.cpp"
 $10 = ((($v)) + 12|0); //@line 181 "modules/audio/wrap_Audio.cpp"
 HEAPF32[$10>>2] = $9; //@line 181 "modules/audio/wrap_Audio.cpp"
 $11 = (+_luaL_checknumber($L,5)); //@line 182 "modules/audio/wrap_Audio.cpp"
 $12 = $11; //@line 182 "modules/audio/wrap_Audio.cpp"
 $13 = ((($v)) + 16|0); //@line 182 "modules/audio/wrap_Audio.cpp"
 HEAPF32[$13>>2] = $12; //@line 182 "modules/audio/wrap_Audio.cpp"
 $14 = (+_luaL_checknumber($L,6)); //@line 183 "modules/audio/wrap_Audio.cpp"
 $15 = $14; //@line 183 "modules/audio/wrap_Audio.cpp"
 $16 = ((($v)) + 20|0); //@line 183 "modules/audio/wrap_Audio.cpp"
 HEAPF32[$16>>2] = $15; //@line 183 "modules/audio/wrap_Audio.cpp"
 $17 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
 $18 = HEAP32[$17>>2]|0; //@line 184 "modules/audio/wrap_Audio.cpp"
 $19 = ((($18)) + 100|0); //@line 184 "modules/audio/wrap_Audio.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 184 "modules/audio/wrap_Audio.cpp"
 FUNCTION_TABLE_vii[$20 & 8191]($17,$v); //@line 184 "modules/audio/wrap_Audio.cpp"
 STACKTOP = sp;return 0; //@line 185 "modules/audio/wrap_Audio.cpp"
}
function __ZN4love5audio16w_getOrientationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0.0;
 var $9 = 0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp;
 $0 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 191 "modules/audio/wrap_Audio.cpp"
 $2 = ((($1)) + 96|0); //@line 191 "modules/audio/wrap_Audio.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 191 "modules/audio/wrap_Audio.cpp"
 FUNCTION_TABLE_vii[$3 & 8191]($0,$v); //@line 191 "modules/audio/wrap_Audio.cpp"
 $4 = +HEAPF32[$v>>2]; //@line 192 "modules/audio/wrap_Audio.cpp"
 $5 = $4; //@line 192 "modules/audio/wrap_Audio.cpp"
 _lua_pushnumber($L,$5); //@line 192 "modules/audio/wrap_Audio.cpp"
 $6 = ((($v)) + 4|0); //@line 193 "modules/audio/wrap_Audio.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 193 "modules/audio/wrap_Audio.cpp"
 $8 = $7; //@line 193 "modules/audio/wrap_Audio.cpp"
 _lua_pushnumber($L,$8); //@line 193 "modules/audio/wrap_Audio.cpp"
 $9 = ((($v)) + 8|0); //@line 194 "modules/audio/wrap_Audio.cpp"
 $10 = +HEAPF32[$9>>2]; //@line 194 "modules/audio/wrap_Audio.cpp"
 $11 = $10; //@line 194 "modules/audio/wrap_Audio.cpp"
 _lua_pushnumber($L,$11); //@line 194 "modules/audio/wrap_Audio.cpp"
 $12 = ((($v)) + 12|0); //@line 195 "modules/audio/wrap_Audio.cpp"
 $13 = +HEAPF32[$12>>2]; //@line 195 "modules/audio/wrap_Audio.cpp"
 $14 = $13; //@line 195 "modules/audio/wrap_Audio.cpp"
 _lua_pushnumber($L,$14); //@line 195 "modules/audio/wrap_Audio.cpp"
 $15 = ((($v)) + 16|0); //@line 196 "modules/audio/wrap_Audio.cpp"
 $16 = +HEAPF32[$15>>2]; //@line 196 "modules/audio/wrap_Audio.cpp"
 $17 = $16; //@line 196 "modules/audio/wrap_Audio.cpp"
 _lua_pushnumber($L,$17); //@line 196 "modules/audio/wrap_Audio.cpp"
 $18 = ((($v)) + 20|0); //@line 197 "modules/audio/wrap_Audio.cpp"
 $19 = +HEAPF32[$18>>2]; //@line 197 "modules/audio/wrap_Audio.cpp"
 $20 = $19; //@line 197 "modules/audio/wrap_Audio.cpp"
 _lua_pushnumber($L,$20); //@line 197 "modules/audio/wrap_Audio.cpp"
 STACKTOP = sp;return 6; //@line 198 "modules/audio/wrap_Audio.cpp"
}
function __ZN4love5audio13w_setVelocityEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, $10 = 0, $11 = 0, $2 = 0.0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp;
 $0 = (+_luaL_checknumber($L,1)); //@line 204 "modules/audio/wrap_Audio.cpp"
 $1 = $0; //@line 204 "modules/audio/wrap_Audio.cpp"
 HEAPF32[$v>>2] = $1; //@line 204 "modules/audio/wrap_Audio.cpp"
 $2 = (+_luaL_checknumber($L,2)); //@line 205 "modules/audio/wrap_Audio.cpp"
 $3 = $2; //@line 205 "modules/audio/wrap_Audio.cpp"
 $4 = ((($v)) + 4|0); //@line 205 "modules/audio/wrap_Audio.cpp"
 HEAPF32[$4>>2] = $3; //@line 205 "modules/audio/wrap_Audio.cpp"
 $5 = (+_luaL_optnumber($L,3,0.0)); //@line 206 "modules/audio/wrap_Audio.cpp"
 $6 = $5; //@line 206 "modules/audio/wrap_Audio.cpp"
 $7 = ((($v)) + 8|0); //@line 206 "modules/audio/wrap_Audio.cpp"
 HEAPF32[$7>>2] = $6; //@line 206 "modules/audio/wrap_Audio.cpp"
 $8 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
 $9 = HEAP32[$8>>2]|0; //@line 207 "modules/audio/wrap_Audio.cpp"
 $10 = ((($9)) + 108|0); //@line 207 "modules/audio/wrap_Audio.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 207 "modules/audio/wrap_Audio.cpp"
 FUNCTION_TABLE_vii[$11 & 8191]($8,$v); //@line 207 "modules/audio/wrap_Audio.cpp"
 STACKTOP = sp;return 0; //@line 208 "modules/audio/wrap_Audio.cpp"
}
function __ZN4love5audio13w_getVelocityEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp;
 $0 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 214 "modules/audio/wrap_Audio.cpp"
 $2 = ((($1)) + 104|0); //@line 214 "modules/audio/wrap_Audio.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 214 "modules/audio/wrap_Audio.cpp"
 FUNCTION_TABLE_vii[$3 & 8191]($0,$v); //@line 214 "modules/audio/wrap_Audio.cpp"
 $4 = +HEAPF32[$v>>2]; //@line 215 "modules/audio/wrap_Audio.cpp"
 $5 = $4; //@line 215 "modules/audio/wrap_Audio.cpp"
 _lua_pushnumber($L,$5); //@line 215 "modules/audio/wrap_Audio.cpp"
 $6 = ((($v)) + 4|0); //@line 216 "modules/audio/wrap_Audio.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 216 "modules/audio/wrap_Audio.cpp"
 $8 = $7; //@line 216 "modules/audio/wrap_Audio.cpp"
 _lua_pushnumber($L,$8); //@line 216 "modules/audio/wrap_Audio.cpp"
 $9 = ((($v)) + 8|0); //@line 217 "modules/audio/wrap_Audio.cpp"
 $10 = +HEAPF32[$9>>2]; //@line 217 "modules/audio/wrap_Audio.cpp"
 $11 = $10; //@line 217 "modules/audio/wrap_Audio.cpp"
 _lua_pushnumber($L,$11); //@line 217 "modules/audio/wrap_Audio.cpp"
 STACKTOP = sp;return 3; //@line 218 "modules/audio/wrap_Audio.cpp"
}
function __ZN4love5audio17w_setDopplerScaleEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 223 "modules/audio/wrap_Audio.cpp"
 $2 = ((($1)) + 112|0); //@line 223 "modules/audio/wrap_Audio.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 223 "modules/audio/wrap_Audio.cpp"
 $4 = (+_luaL_checknumber($L,1)); //@line 181 "./common/runtime.h"
 $5 = $4; //@line 181 "./common/runtime.h"
 FUNCTION_TABLE_vid[$3 & 8191]($0,$5); //@line 223 "modules/audio/wrap_Audio.cpp"
 return 0; //@line 224 "modules/audio/wrap_Audio.cpp"
}
function __ZN4love5audio17w_getDopplerScaleEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 229 "modules/audio/wrap_Audio.cpp"
 $2 = ((($1)) + 116|0); //@line 229 "modules/audio/wrap_Audio.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 229 "modules/audio/wrap_Audio.cpp"
 $4 = (+FUNCTION_TABLE_di[$3 & 4095]($0)); //@line 229 "modules/audio/wrap_Audio.cpp"
 $5 = $4; //@line 229 "modules/audio/wrap_Audio.cpp"
 _lua_pushnumber($L,$5); //@line 229 "modules/audio/wrap_Audio.cpp"
 return 1; //@line 230 "modules/audio/wrap_Audio.cpp"
}
function __ZN4love5audio18w_setDistanceModelEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $distanceModel = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $distanceModel = sp + 4|0;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 278 "modules/audio/wrap_Audio.cpp"
 $1 = (__ZN4love5audio5Audio11getConstantEPKcRNS1_13DistanceModelE($0,$distanceModel)|0); //@line 280 "modules/audio/wrap_Audio.cpp"
 if ($1) {
  $3 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
  $4 = HEAP32[$3>>2]|0; //@line 282 "modules/audio/wrap_Audio.cpp"
  $5 = ((($4)) + 140|0); //@line 282 "modules/audio/wrap_Audio.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 282 "modules/audio/wrap_Audio.cpp"
  $7 = HEAP32[$distanceModel>>2]|0; //@line 282 "modules/audio/wrap_Audio.cpp"
  FUNCTION_TABLE_vii[$6 & 8191]($3,$7); //@line 282 "modules/audio/wrap_Audio.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 284 "modules/audio/wrap_Audio.cpp"
 } else {
  HEAP32[$vararg_buffer>>2] = $0; //@line 281 "modules/audio/wrap_Audio.cpp"
  $2 = (_luaL_error($L,327580,$vararg_buffer)|0); //@line 281 "modules/audio/wrap_Audio.cpp"
  $$0 = $2;
  STACKTOP = sp;return ($$0|0); //@line 284 "modules/audio/wrap_Audio.cpp"
 }
 return (0)|0;
}
function __ZN4love5audio18w_getDistanceModelEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $modelStr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $modelStr = sp;
 $0 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 288 "modules/audio/wrap_Audio.cpp"
 $2 = ((($1)) + 136|0); //@line 288 "modules/audio/wrap_Audio.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 288 "modules/audio/wrap_Audio.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 288 "modules/audio/wrap_Audio.cpp"
 $5 = (__ZN4love5audio5Audio11getConstantENS1_13DistanceModelERPKc($4,$modelStr)|0); //@line 290 "modules/audio/wrap_Audio.cpp"
 if (!($5)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 294 "modules/audio/wrap_Audio.cpp"
 }
 $6 = HEAP32[$modelStr>>2]|0; //@line 292 "modules/audio/wrap_Audio.cpp"
 _lua_pushstring($L,$6); //@line 292 "modules/audio/wrap_Audio.cpp"
 $$0 = 1;
 STACKTOP = sp;return ($$0|0); //@line 294 "modules/audio/wrap_Audio.cpp"
}
function _luaopen_love_audio($L) {
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $$1 = 0, $$13 = 0, $$2 = 0, $$24 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body18$index7Z2D = 0, $eh$lpad$body18$indexZ2D = 0, $instance$122 = 0, $vararg_buffer = 0, $w = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 28|0;
 $1 = sp + 24|0;
 $w = sp + 4|0;
 $2 = HEAP32[156402]|0; //@line 98 "./common/Module.h"
 $3 = ($2|0)==(0|0); //@line 335 "modules/audio/wrap_Audio.cpp"
 L1: do {
  if ($3) {
   __THREW__ = 0;
   $4 = (invoke_ii(3348,32)|0); //@line 340 "modules/audio/wrap_Audio.cpp"
   $5 = __THREW__; __THREW__ = 0;
   $6 = $5&1;
   do {
    if ($6) {
     $9 = ___cxa_find_matching_catch_3(96|0)|0;
     $10 = tempRet0;
     $$01 = $9;$$02 = $10;
    } else {
     __THREW__ = 0;
     invoke_vi(3500,($4|0)); //@line 340 "modules/audio/wrap_Audio.cpp"
     $7 = __THREW__; __THREW__ = 0;
     $8 = $7&1;
     if ($8) {
      $11 = ___cxa_find_matching_catch_3(96|0)|0;
      $12 = tempRet0;
      __ZdlPv($4); //@line 340 "modules/audio/wrap_Audio.cpp"
      $$01 = $11;$$02 = $12;
      break;
     } else {
      $instance$122 = $4;
      break L1;
     }
    }
   } while(0);
   $13 = (_llvm_eh_typeid_for((96|0))|0); //@line 341 "modules/audio/wrap_Audio.cpp"
   $14 = ($$02|0)==($13|0); //@line 341 "modules/audio/wrap_Audio.cpp"
   if (!($14)) {
    $$2 = $$01;$$24 = $$02;
    ___resumeException($$2|0);
    // unreachable;
   }
   $15 = (___cxa_begin_catch(($$01|0))|0); //@line 341 "modules/audio/wrap_Audio.cpp"
   $16 = HEAP32[$15>>2]|0; //@line 344 "modules/audio/wrap_Audio.cpp"
   $17 = ((($16)) + 8|0); //@line 344 "modules/audio/wrap_Audio.cpp"
   $18 = HEAP32[$17>>2]|0; //@line 344 "modules/audio/wrap_Audio.cpp"
   $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 344 "modules/audio/wrap_Audio.cpp"
   $20 = (_strlen($19)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   $21 = (invoke_iiii(3366,(1206612|0),($19|0),($20|0))|0); //@line 894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
   $22 = __THREW__; __THREW__ = 0;
   $23 = $22&1;
   do {
    if ($23) {
     label = 15;
    } else {
     $24 = HEAP32[$21>>2]|0; //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
     $25 = ((($24)) + -12|0); //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
     $26 = HEAP32[$25>>2]|0; //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
     $27 = (($21) + ($26)|0); //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
     __THREW__ = 0;
     $28 = (invoke_ii(3373,($27|0))|0); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     $29 = __THREW__; __THREW__ = 0;
     $30 = $29&1;
     if ($30) {
      label = 15;
     } else {
      HEAP32[$1>>2] = $28; //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
      __THREW__ = 0;
      $31 = (invoke_iii(3374,($1|0),(1207576|0))|0); //@line 181 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if (!($33)) {
       $34 = HEAP32[$31>>2]|0; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
       $35 = ((($34)) + 28|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
       $36 = HEAP32[$35>>2]|0; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
       __THREW__ = 0;
       $37 = (invoke_iii($36|0,($31|0),10)|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
       $38 = __THREW__; __THREW__ = 0;
       $39 = $38&1;
       if (!($39)) {
        __ZNSt3__16localeD2Ev($1); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
        __THREW__ = 0;
        (invoke_iii(3501,($21|0),($37|0))|0); //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
        $42 = __THREW__; __THREW__ = 0;
        $43 = $42&1;
        if ($43) {
         label = 15;
         break;
        }
        __THREW__ = 0;
        (invoke_ii(3502,($21|0))|0); //@line 1035 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
        $44 = __THREW__; __THREW__ = 0;
        $45 = $44&1;
        if ($45) {
         label = 15;
         break;
        }
        ___cxa_end_catch(); //@line 345 "modules/audio/wrap_Audio.cpp"
        __THREW__ = 0;
        $53 = (invoke_ii(3348,16)|0); //@line 357 "modules/audio/wrap_Audio.cpp"
        $54 = __THREW__; __THREW__ = 0;
        $55 = $54&1;
        do {
         if ($55) {
          $58 = ___cxa_find_matching_catch_3(96|0)|0;
          $59 = tempRet0;
          $$1 = $58;$$13 = $59;
         } else {
          __THREW__ = 0;
          invoke_vi(3503,($53|0)); //@line 357 "modules/audio/wrap_Audio.cpp"
          $56 = __THREW__; __THREW__ = 0;
          $57 = $56&1;
          if ($57) {
           $60 = ___cxa_find_matching_catch_3(96|0)|0;
           $61 = tempRet0;
           __ZdlPv($53); //@line 357 "modules/audio/wrap_Audio.cpp"
           $$1 = $60;$$13 = $61;
           break;
          } else {
           $instance$122 = $53;
           break L1;
          }
         }
        } while(0);
        $62 = ($$13|0)==($$02|0); //@line 358 "modules/audio/wrap_Audio.cpp"
        if (!($62)) {
         $$2 = $$1;$$24 = $$13;
         ___resumeException($$2|0);
         // unreachable;
        }
        $63 = (___cxa_begin_catch(($$1|0))|0); //@line 358 "modules/audio/wrap_Audio.cpp"
        $64 = HEAP32[$63>>2]|0; //@line 361 "modules/audio/wrap_Audio.cpp"
        $65 = ((($64)) + 8|0); //@line 361 "modules/audio/wrap_Audio.cpp"
        $66 = HEAP32[$65>>2]|0; //@line 361 "modules/audio/wrap_Audio.cpp"
        $67 = (FUNCTION_TABLE_ii[$66 & 8191]($63)|0); //@line 361 "modules/audio/wrap_Audio.cpp"
        $68 = (_strlen($67)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        __THREW__ = 0;
        $69 = (invoke_iiii(3366,(1206612|0),($67|0),($68|0))|0); //@line 894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
        $70 = __THREW__; __THREW__ = 0;
        $71 = $70&1;
        do {
         if ($71) {
          label = 31;
         } else {
          $72 = HEAP32[$69>>2]|0; //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
          $73 = ((($72)) + -12|0); //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
          $74 = HEAP32[$73>>2]|0; //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
          $75 = (($69) + ($74)|0); //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
          __THREW__ = 0;
          $76 = (invoke_ii(3373,($75|0))|0); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
          $77 = __THREW__; __THREW__ = 0;
          $78 = $77&1;
          if ($78) {
           label = 31;
          } else {
           HEAP32[$0>>2] = $76; //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
           __THREW__ = 0;
           $79 = (invoke_iii(3374,($0|0),(1207576|0))|0); //@line 181 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
           $80 = __THREW__; __THREW__ = 0;
           $81 = $80&1;
           if (!($81)) {
            $82 = HEAP32[$79>>2]|0; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
            $83 = ((($82)) + 28|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
            $84 = HEAP32[$83>>2]|0; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
            __THREW__ = 0;
            $85 = (invoke_iii($84|0,($79|0),10)|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
            $86 = __THREW__; __THREW__ = 0;
            $87 = $86&1;
            if (!($87)) {
             __ZNSt3__16localeD2Ev($0); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
             __THREW__ = 0;
             (invoke_iii(3501,($69|0),($85|0))|0); //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
             $90 = __THREW__; __THREW__ = 0;
             $91 = $90&1;
             if ($91) {
              label = 31;
              break;
             }
             __THREW__ = 0;
             (invoke_ii(3502,($69|0))|0); //@line 1035 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
             $92 = __THREW__; __THREW__ = 0;
             $93 = $92&1;
             if ($93) {
              label = 31;
              break;
             }
             ___cxa_end_catch(); //@line 362 "modules/audio/wrap_Audio.cpp"
             $98 = (_luaL_error($L,327607,$vararg_buffer)|0); //@line 367 "modules/audio/wrap_Audio.cpp"
             $$0 = $98;
             STACKTOP = sp;return ($$0|0); //@line 379 "modules/audio/wrap_Audio.cpp"
            }
           }
           $88 = ___cxa_find_matching_catch_2()|0;
           $89 = tempRet0;
           __ZNSt3__16localeD2Ev($0); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
           $eh$lpad$body18$index7Z2D = $89;$eh$lpad$body18$indexZ2D = $88;
          }
         }
        } while(0);
        if ((label|0) == 31) {
         $94 = ___cxa_find_matching_catch_2()|0;
         $95 = tempRet0;
         $eh$lpad$body18$index7Z2D = $95;$eh$lpad$body18$indexZ2D = $94;
        }
        __THREW__ = 0;
        invoke_v(3378); //@line 362 "modules/audio/wrap_Audio.cpp"
        $96 = __THREW__; __THREW__ = 0;
        $97 = $96&1;
        if ($97) {
         $104 = ___cxa_find_matching_catch_3(0|0)|0;
         $105 = tempRet0;
         ___clang_call_terminate($104); //@line 345 "modules/audio/wrap_Audio.cpp"
         // unreachable; //@line 345 "modules/audio/wrap_Audio.cpp"
        } else {
         $$2 = $eh$lpad$body18$indexZ2D;$$24 = $eh$lpad$body18$index7Z2D;
         ___resumeException($$2|0);
         // unreachable;
        }
       }
      }
      $40 = ___cxa_find_matching_catch_2()|0;
      $41 = tempRet0;
      __ZNSt3__16localeD2Ev($1); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
      $eh$lpad$body$index2Z2D = $41;$eh$lpad$body$indexZ2D = $40;
     }
    }
   } while(0);
   if ((label|0) == 15) {
    $46 = ___cxa_find_matching_catch_2()|0;
    $47 = tempRet0;
    $eh$lpad$body$index2Z2D = $47;$eh$lpad$body$indexZ2D = $46;
   }
   __THREW__ = 0;
   invoke_v(3378); //@line 345 "modules/audio/wrap_Audio.cpp"
   $48 = __THREW__; __THREW__ = 0;
   $49 = $48&1;
   if ($49) {
    $104 = ___cxa_find_matching_catch_3(0|0)|0;
    $105 = tempRet0;
    ___clang_call_terminate($104); //@line 345 "modules/audio/wrap_Audio.cpp"
    // unreachable; //@line 345 "modules/audio/wrap_Audio.cpp"
   } else {
    $$2 = $eh$lpad$body$indexZ2D;$$24 = $eh$lpad$body$index2Z2D;
    ___resumeException($$2|0);
    // unreachable;
   }
  } else {
   $50 = HEAP32[$2>>2]|0; //@line 348 "modules/audio/wrap_Audio.cpp"
   $51 = ((($50)) + 8|0); //@line 348 "modules/audio/wrap_Audio.cpp"
   $52 = HEAP32[$51>>2]|0; //@line 348 "modules/audio/wrap_Audio.cpp"
   FUNCTION_TABLE_vi[$52 & 8191]($2); //@line 348 "modules/audio/wrap_Audio.cpp"
   $instance$122 = $2;
  }
 } while(0);
 HEAP32[$w>>2] = $instance$122; //@line 370 "modules/audio/wrap_Audio.cpp"
 $99 = ((($w)) + 4|0); //@line 371 "modules/audio/wrap_Audio.cpp"
 HEAP32[$99>>2] = 327640; //@line 371 "modules/audio/wrap_Audio.cpp"
 $100 = ((($w)) + 8|0); //@line 372 "modules/audio/wrap_Audio.cpp"
 HEAP32[$100>>2] = 3; //@line 372 "modules/audio/wrap_Audio.cpp"
 $101 = ((($w)) + 12|0); //@line 373 "modules/audio/wrap_Audio.cpp"
 HEAP32[$101>>2] = 12304; //@line 373 "modules/audio/wrap_Audio.cpp"
 $102 = ((($w)) + 16|0); //@line 374 "modules/audio/wrap_Audio.cpp"
 HEAP32[$102>>2] = 12464; //@line 374 "modules/audio/wrap_Audio.cpp"
 $103 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 376 "modules/audio/wrap_Audio.cpp"
 $$0 = $103;
 STACKTOP = sp;return ($$0|0); //@line 379 "modules/audio/wrap_Audio.cpp"
}
function __ZN4love5audio16luax_checksourceEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,$idx,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 return ($0|0); //@line 32 "modules/audio/wrap_Source.cpp"
}
function __ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love5audio14w_Source_cloneEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $clone$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 35 "modules/audio/wrap_Source.cpp"
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 39 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 16|0); //@line 39 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 39 "modules/audio/wrap_Source.cpp"
 __THREW__ = 0;
 $4 = (invoke_ii($3|0,($0|0))|0); //@line 39 "modules/audio/wrap_Source.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 do {
  if ($6) {
   $7 = ___cxa_find_matching_catch_3(8992|0)|0;
   $8 = tempRet0;
   $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
   if (!($10)) {
    $$01$i = $8;$$02$i = $7;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
   $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
   $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
   $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
   $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
   $16 = __THREW__; __THREW__ = 0;
   $17 = $16&1;
   if (!($17)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $18 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $18; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $clone$0 = 0;
    break;
   }
   $19 = ___cxa_find_matching_catch_2()|0;
   $20 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $21 = __THREW__; __THREW__ = 0;
   $22 = $21&1;
   if ($22) {
    $23 = ___cxa_find_matching_catch_3(0|0)|0;
    $24 = tempRet0;
    ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $20;$$02$i = $19;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $clone$0 = $4;
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,28,$clone$0); //@line 40 "modules/audio/wrap_Source.cpp"
 $25 = HEAP32[$clone$0>>2]|0; //@line 41 "modules/audio/wrap_Source.cpp"
 $26 = ((($25)) + 12|0); //@line 41 "modules/audio/wrap_Source.cpp"
 $27 = HEAP32[$26>>2]|0; //@line 41 "modules/audio/wrap_Source.cpp"
 FUNCTION_TABLE_vi[$27 & 8191]($clone$0); //@line 41 "modules/audio/wrap_Source.cpp"
 STACKTOP = sp;return 1; //@line 42 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio13w_Source_playEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 48 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 20|0); //@line 48 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 48 "modules/audio/wrap_Source.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 48 "modules/audio/wrap_Source.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 48 "modules/audio/wrap_Source.cpp"
 return 1; //@line 49 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio13w_Source_stopEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 55 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 24|0); //@line 55 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 55 "modules/audio/wrap_Source.cpp"
 FUNCTION_TABLE_vi[$3 & 8191]($0); //@line 55 "modules/audio/wrap_Source.cpp"
 return 0; //@line 56 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio14w_Source_pauseEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 62 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 28|0); //@line 62 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 62 "modules/audio/wrap_Source.cpp"
 FUNCTION_TABLE_vi[$3 & 8191]($0); //@line 62 "modules/audio/wrap_Source.cpp"
 return 0; //@line 63 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio15w_Source_resumeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 69 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 32|0); //@line 69 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 69 "modules/audio/wrap_Source.cpp"
 FUNCTION_TABLE_vi[$3 & 8191]($0); //@line 69 "modules/audio/wrap_Source.cpp"
 return 0; //@line 70 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio15w_Source_rewindEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 76 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 36|0); //@line 76 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 76 "modules/audio/wrap_Source.cpp"
 FUNCTION_TABLE_vi[$3 & 8191]($0); //@line 76 "modules/audio/wrap_Source.cpp"
 return 0; //@line 77 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio17w_Source_setPitchEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0.0, $10 = 0, $2 = 0.0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 80 "modules/audio/wrap_Source.cpp"
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 83 "modules/audio/wrap_Source.cpp"
 $2 = $1; //@line 83 "modules/audio/wrap_Source.cpp"
 $3 = $2; //@line 84 "modules/audio/wrap_Source.cpp"
 $4 = $3 > 1.7976931348623157E+308; //@line 84 "modules/audio/wrap_Source.cpp"
 if (!($4)) {
  $5 = $3 < 2.2250738585072014E-308; //@line 85 "modules/audio/wrap_Source.cpp"
  $6 = ($2 != $2) | (0.0 != 0.0); //@line 86 "modules/audio/wrap_Source.cpp"
  $or$cond = $6 | $5; //@line 85 "modules/audio/wrap_Source.cpp"
  if (!($or$cond)) {
   $8 = HEAP32[$0>>2]|0; //@line 88 "modules/audio/wrap_Source.cpp"
   $9 = ((($8)) + 56|0); //@line 88 "modules/audio/wrap_Source.cpp"
   $10 = HEAP32[$9>>2]|0; //@line 88 "modules/audio/wrap_Source.cpp"
   FUNCTION_TABLE_vid[$10 & 8191]($0,$2); //@line 88 "modules/audio/wrap_Source.cpp"
   $$0 = 0;
   STACKTOP = sp;return ($$0|0); //@line 90 "modules/audio/wrap_Source.cpp"
  }
 }
 $7 = (_luaL_error($L,327767,$vararg_buffer)|0); //@line 87 "modules/audio/wrap_Source.cpp"
 $$0 = $7;
 STACKTOP = sp;return ($$0|0); //@line 90 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio17w_Source_getPitchEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 95 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 60|0); //@line 95 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 95 "modules/audio/wrap_Source.cpp"
 $4 = (+FUNCTION_TABLE_di[$3 & 4095]($0)); //@line 95 "modules/audio/wrap_Source.cpp"
 $5 = $4; //@line 95 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$5); //@line 95 "modules/audio/wrap_Source.cpp"
 return 1; //@line 96 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio18w_Source_setVolumeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 102 "modules/audio/wrap_Source.cpp"
 $2 = $1; //@line 102 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$0>>2]|0; //@line 103 "modules/audio/wrap_Source.cpp"
 $4 = ((($3)) + 64|0); //@line 103 "modules/audio/wrap_Source.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 103 "modules/audio/wrap_Source.cpp"
 FUNCTION_TABLE_vid[$5 & 8191]($0,$2); //@line 103 "modules/audio/wrap_Source.cpp"
 return 0; //@line 104 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio18w_Source_getVolumeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 110 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 68|0); //@line 110 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 110 "modules/audio/wrap_Source.cpp"
 $4 = (+FUNCTION_TABLE_di[$3 & 4095]($0)); //@line 110 "modules/audio/wrap_Source.cpp"
 $5 = $4; //@line 110 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$5); //@line 110 "modules/audio/wrap_Source.cpp"
 return 1; //@line 111 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio13w_Source_seekEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$1 = 0, $$pre = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0.0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $u = 0, $vararg_buffer = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $u = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 117 "modules/audio/wrap_Source.cpp"
 $2 = $1; //@line 117 "modules/audio/wrap_Source.cpp"
 $3 = $2 < 0.0; //@line 118 "modules/audio/wrap_Source.cpp"
 if ($3) {
  $4 = (_luaL_argerror($L,2,327803)|0); //@line 119 "modules/audio/wrap_Source.cpp"
  $$1 = $4;
  STACKTOP = sp;return ($$1|0); //@line 128 "modules/audio/wrap_Source.cpp"
 }
 HEAP32[$u>>2] = 0; //@line 121 "modules/audio/wrap_Source.cpp"
 $5 = (_lua_type($L,3)|0); //@line 122 "modules/audio/wrap_Source.cpp"
 $6 = ($5|0)<(1); //@line 122 "modules/audio/wrap_Source.cpp"
 do {
  if ($6) {
   $14 = 0;
   label = 8;
  } else {
   $7 = (_lua_tolstring($L,3,0)|0); //@line 122 "modules/audio/wrap_Source.cpp"
   $8 = ($7|0)==(0|0); //@line 123 "modules/audio/wrap_Source.cpp"
   if ($8) {
    $14 = 0;
    label = 8;
   } else {
    $9 = (__ZN4love5audio6Source11getConstantEPKcRNS1_4UnitE($7,$u)|0); //@line 123 "modules/audio/wrap_Source.cpp"
    if ($9) {
     $$pre = HEAP32[$u>>2]|0; //@line 126 "modules/audio/wrap_Source.cpp"
     $14 = $$pre;
     label = 8;
     break;
    } else {
     HEAP32[$vararg_buffer>>2] = $7; //@line 124 "modules/audio/wrap_Source.cpp"
     $10 = (_luaL_error($L,327837,$vararg_buffer)|0); //@line 124 "modules/audio/wrap_Source.cpp"
     $$0 = $10;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $11 = HEAP32[$0>>2]|0; //@line 126 "modules/audio/wrap_Source.cpp"
  $12 = ((($11)) + 72|0); //@line 126 "modules/audio/wrap_Source.cpp"
  $13 = HEAP32[$12>>2]|0; //@line 126 "modules/audio/wrap_Source.cpp"
  FUNCTION_TABLE_vidi[$13 & 2047]($0,$2,$14); //@line 126 "modules/audio/wrap_Source.cpp"
  $$0 = 0;
 }
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0); //@line 128 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio13w_Source_tellEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $u = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $u = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 HEAP32[$u>>2] = 0; //@line 134 "modules/audio/wrap_Source.cpp"
 $1 = (_lua_type($L,2)|0); //@line 135 "modules/audio/wrap_Source.cpp"
 $2 = ($1|0)<(1); //@line 135 "modules/audio/wrap_Source.cpp"
 do {
  if ($2) {
   $10 = 0;
  } else {
   $3 = (_lua_tolstring($L,2,0)|0); //@line 135 "modules/audio/wrap_Source.cpp"
   $4 = ($3|0)==(0|0); //@line 136 "modules/audio/wrap_Source.cpp"
   if ($4) {
    $10 = 0;
   } else {
    $5 = (__ZN4love5audio6Source11getConstantEPKcRNS1_4UnitE($3,$u)|0); //@line 136 "modules/audio/wrap_Source.cpp"
    if ($5) {
     $$pre = HEAP32[$u>>2]|0; //@line 139 "modules/audio/wrap_Source.cpp"
     $10 = $$pre;
     break;
    }
    HEAP32[$vararg_buffer>>2] = $3; //@line 137 "modules/audio/wrap_Source.cpp"
    $6 = (_luaL_error($L,327837,$vararg_buffer)|0); //@line 137 "modules/audio/wrap_Source.cpp"
    $$0 = $6;
    STACKTOP = sp;return ($$0|0); //@line 141 "modules/audio/wrap_Source.cpp"
   }
  }
 } while(0);
 $7 = HEAP32[$0>>2]|0; //@line 139 "modules/audio/wrap_Source.cpp"
 $8 = ((($7)) + 76|0); //@line 139 "modules/audio/wrap_Source.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 139 "modules/audio/wrap_Source.cpp"
 $11 = (+FUNCTION_TABLE_dii[$9 & 8191]($0,$10)); //@line 139 "modules/audio/wrap_Source.cpp"
 $12 = $11; //@line 139 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$12); //@line 139 "modules/audio/wrap_Source.cpp"
 $$0 = 1;
 STACKTOP = sp;return ($$0|0); //@line 141 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio20w_Source_getDurationEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $u = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $u = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 HEAP32[$u>>2] = 0; //@line 147 "modules/audio/wrap_Source.cpp"
 $1 = (_lua_type($L,2)|0); //@line 148 "modules/audio/wrap_Source.cpp"
 $2 = ($1|0)<(1); //@line 148 "modules/audio/wrap_Source.cpp"
 do {
  if ($2) {
   $10 = 0;
  } else {
   $3 = (_lua_tolstring($L,2,0)|0); //@line 148 "modules/audio/wrap_Source.cpp"
   $4 = ($3|0)==(0|0); //@line 149 "modules/audio/wrap_Source.cpp"
   if ($4) {
    $10 = 0;
   } else {
    $5 = (__ZN4love5audio6Source11getConstantEPKcRNS1_4UnitE($3,$u)|0); //@line 149 "modules/audio/wrap_Source.cpp"
    if ($5) {
     $$pre = HEAP32[$u>>2]|0; //@line 152 "modules/audio/wrap_Source.cpp"
     $10 = $$pre;
     break;
    }
    HEAP32[$vararg_buffer>>2] = $3; //@line 150 "modules/audio/wrap_Source.cpp"
    $6 = (_luaL_error($L,327837,$vararg_buffer)|0); //@line 150 "modules/audio/wrap_Source.cpp"
    $$0 = $6;
    STACKTOP = sp;return ($$0|0); //@line 154 "modules/audio/wrap_Source.cpp"
   }
  }
 } while(0);
 $7 = HEAP32[$0>>2]|0; //@line 152 "modules/audio/wrap_Source.cpp"
 $8 = ((($7)) + 80|0); //@line 152 "modules/audio/wrap_Source.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 152 "modules/audio/wrap_Source.cpp"
 $11 = (+FUNCTION_TABLE_dii[$9 & 8191]($0,$10)); //@line 152 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$11); //@line 152 "modules/audio/wrap_Source.cpp"
 $$0 = 1;
 STACKTOP = sp;return ($$0|0); //@line 154 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio20w_Source_setPositionEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0, $9 = 0, $v = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $v = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 160 "modules/audio/wrap_Source.cpp"
 $2 = $1; //@line 160 "modules/audio/wrap_Source.cpp"
 HEAPF32[$v>>2] = $2; //@line 160 "modules/audio/wrap_Source.cpp"
 $3 = (+_luaL_checknumber($L,3)); //@line 161 "modules/audio/wrap_Source.cpp"
 $4 = $3; //@line 161 "modules/audio/wrap_Source.cpp"
 $5 = ((($v)) + 4|0); //@line 161 "modules/audio/wrap_Source.cpp"
 HEAPF32[$5>>2] = $4; //@line 161 "modules/audio/wrap_Source.cpp"
 $6 = (+_luaL_optnumber($L,4,0.0)); //@line 162 "modules/audio/wrap_Source.cpp"
 $7 = $6; //@line 162 "modules/audio/wrap_Source.cpp"
 $8 = ((($v)) + 8|0); //@line 162 "modules/audio/wrap_Source.cpp"
 HEAPF32[$8>>2] = $7; //@line 162 "modules/audio/wrap_Source.cpp"
 $9 = HEAP32[$0>>2]|0; //@line 163 "modules/audio/wrap_Source.cpp"
 $10 = ((($9)) + 84|0); //@line 163 "modules/audio/wrap_Source.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 163 "modules/audio/wrap_Source.cpp"
 __THREW__ = 0;
 invoke_vii($11|0,($0|0),($v|0)); //@line 163 "modules/audio/wrap_Source.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if (!($13)) {
  STACKTOP = sp;return 0; //@line 164 "modules/audio/wrap_Source.cpp"
 }
 $14 = ___cxa_find_matching_catch_3(8992|0)|0;
 $15 = tempRet0;
 $16 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $17 = ($15|0)==($16|0); //@line 535 "./common/runtime.h"
 if (!($17)) {
  $$01$i = $15;$$02$i = $14;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $18 = (___cxa_begin_catch(($14|0))|0); //@line 535 "./common/runtime.h"
 $19 = HEAP32[$18>>2]|0; //@line 539 "./common/runtime.h"
 $20 = ((($19)) + 8|0); //@line 539 "./common/runtime.h"
 $21 = HEAP32[$20>>2]|0; //@line 539 "./common/runtime.h"
 $22 = (FUNCTION_TABLE_ii[$21 & 8191]($18)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($22|0)); //@line 539 "./common/runtime.h"
 $23 = __THREW__; __THREW__ = 0;
 $24 = $23&1;
 if (!($24)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $25 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $25; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 164 "modules/audio/wrap_Source.cpp"
 }
 $26 = ___cxa_find_matching_catch_2()|0;
 $27 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if ($29) {
  $30 = ___cxa_find_matching_catch_3(0|0)|0;
  $31 = tempRet0;
  ___clang_call_terminate($30); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $27;$$02$i = $26;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5audio20w_Source_getPositionEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0.0;
 var $25 = 0.0, $26 = 0, $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $v = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $v = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 171 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 88|0); //@line 171 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 171 "modules/audio/wrap_Source.cpp"
 __THREW__ = 0;
 invoke_vii($3|0,($0|0),($v|0)); //@line 171 "modules/audio/wrap_Source.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 do {
  if ($5) {
   $6 = ___cxa_find_matching_catch_3(8992|0)|0;
   $7 = tempRet0;
   $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
   if (!($9)) {
    $$01$i = $7;$$02$i = $6;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
   $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
   $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
   $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
   $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if (!($16)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $17; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    break;
   }
   $18 = ___cxa_find_matching_catch_2()|0;
   $19 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if ($21) {
    $22 = ___cxa_find_matching_catch_3(0|0)|0;
    $23 = tempRet0;
    ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $19;$$02$i = $18;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  }
 } while(0);
 $24 = +HEAPF32[$v>>2]; //@line 172 "modules/audio/wrap_Source.cpp"
 $25 = $24; //@line 172 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$25); //@line 172 "modules/audio/wrap_Source.cpp"
 $26 = ((($v)) + 4|0); //@line 173 "modules/audio/wrap_Source.cpp"
 $27 = +HEAPF32[$26>>2]; //@line 173 "modules/audio/wrap_Source.cpp"
 $28 = $27; //@line 173 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$28); //@line 173 "modules/audio/wrap_Source.cpp"
 $29 = ((($v)) + 8|0); //@line 174 "modules/audio/wrap_Source.cpp"
 $30 = +HEAPF32[$29>>2]; //@line 174 "modules/audio/wrap_Source.cpp"
 $31 = $30; //@line 174 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$31); //@line 174 "modules/audio/wrap_Source.cpp"
 STACKTOP = sp;return 3; //@line 175 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio20w_Source_setVelocityEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0, $9 = 0, $v = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $v = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 182 "modules/audio/wrap_Source.cpp"
 $2 = $1; //@line 182 "modules/audio/wrap_Source.cpp"
 HEAPF32[$v>>2] = $2; //@line 182 "modules/audio/wrap_Source.cpp"
 $3 = (+_luaL_checknumber($L,3)); //@line 183 "modules/audio/wrap_Source.cpp"
 $4 = $3; //@line 183 "modules/audio/wrap_Source.cpp"
 $5 = ((($v)) + 4|0); //@line 183 "modules/audio/wrap_Source.cpp"
 HEAPF32[$5>>2] = $4; //@line 183 "modules/audio/wrap_Source.cpp"
 $6 = (+_luaL_optnumber($L,4,0.0)); //@line 184 "modules/audio/wrap_Source.cpp"
 $7 = $6; //@line 184 "modules/audio/wrap_Source.cpp"
 $8 = ((($v)) + 8|0); //@line 184 "modules/audio/wrap_Source.cpp"
 HEAPF32[$8>>2] = $7; //@line 184 "modules/audio/wrap_Source.cpp"
 $9 = HEAP32[$0>>2]|0; //@line 185 "modules/audio/wrap_Source.cpp"
 $10 = ((($9)) + 92|0); //@line 185 "modules/audio/wrap_Source.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 185 "modules/audio/wrap_Source.cpp"
 __THREW__ = 0;
 invoke_vii($11|0,($0|0),($v|0)); //@line 185 "modules/audio/wrap_Source.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if (!($13)) {
  STACKTOP = sp;return 0; //@line 186 "modules/audio/wrap_Source.cpp"
 }
 $14 = ___cxa_find_matching_catch_3(8992|0)|0;
 $15 = tempRet0;
 $16 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $17 = ($15|0)==($16|0); //@line 535 "./common/runtime.h"
 if (!($17)) {
  $$01$i = $15;$$02$i = $14;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $18 = (___cxa_begin_catch(($14|0))|0); //@line 535 "./common/runtime.h"
 $19 = HEAP32[$18>>2]|0; //@line 539 "./common/runtime.h"
 $20 = ((($19)) + 8|0); //@line 539 "./common/runtime.h"
 $21 = HEAP32[$20>>2]|0; //@line 539 "./common/runtime.h"
 $22 = (FUNCTION_TABLE_ii[$21 & 8191]($18)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($22|0)); //@line 539 "./common/runtime.h"
 $23 = __THREW__; __THREW__ = 0;
 $24 = $23&1;
 if (!($24)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $25 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $25; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 186 "modules/audio/wrap_Source.cpp"
 }
 $26 = ___cxa_find_matching_catch_2()|0;
 $27 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if ($29) {
  $30 = ___cxa_find_matching_catch_3(0|0)|0;
  $31 = tempRet0;
  ___clang_call_terminate($30); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $27;$$02$i = $26;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5audio20w_Source_getVelocityEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0.0;
 var $25 = 0.0, $26 = 0, $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $v = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $v = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 193 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 96|0); //@line 193 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 193 "modules/audio/wrap_Source.cpp"
 __THREW__ = 0;
 invoke_vii($3|0,($0|0),($v|0)); //@line 193 "modules/audio/wrap_Source.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 do {
  if ($5) {
   $6 = ___cxa_find_matching_catch_3(8992|0)|0;
   $7 = tempRet0;
   $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
   if (!($9)) {
    $$01$i = $7;$$02$i = $6;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
   $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
   $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
   $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
   $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if (!($16)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $17; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    break;
   }
   $18 = ___cxa_find_matching_catch_2()|0;
   $19 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if ($21) {
    $22 = ___cxa_find_matching_catch_3(0|0)|0;
    $23 = tempRet0;
    ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $19;$$02$i = $18;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  }
 } while(0);
 $24 = +HEAPF32[$v>>2]; //@line 194 "modules/audio/wrap_Source.cpp"
 $25 = $24; //@line 194 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$25); //@line 194 "modules/audio/wrap_Source.cpp"
 $26 = ((($v)) + 4|0); //@line 195 "modules/audio/wrap_Source.cpp"
 $27 = +HEAPF32[$26>>2]; //@line 195 "modules/audio/wrap_Source.cpp"
 $28 = $27; //@line 195 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$28); //@line 195 "modules/audio/wrap_Source.cpp"
 $29 = ((($v)) + 8|0); //@line 196 "modules/audio/wrap_Source.cpp"
 $30 = +HEAPF32[$29>>2]; //@line 196 "modules/audio/wrap_Source.cpp"
 $31 = $30; //@line 196 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$31); //@line 196 "modules/audio/wrap_Source.cpp"
 STACKTOP = sp;return 3; //@line 197 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio21w_Source_setDirectionEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0, $9 = 0, $v = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $v = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 204 "modules/audio/wrap_Source.cpp"
 $2 = $1; //@line 204 "modules/audio/wrap_Source.cpp"
 HEAPF32[$v>>2] = $2; //@line 204 "modules/audio/wrap_Source.cpp"
 $3 = (+_luaL_checknumber($L,3)); //@line 205 "modules/audio/wrap_Source.cpp"
 $4 = $3; //@line 205 "modules/audio/wrap_Source.cpp"
 $5 = ((($v)) + 4|0); //@line 205 "modules/audio/wrap_Source.cpp"
 HEAPF32[$5>>2] = $4; //@line 205 "modules/audio/wrap_Source.cpp"
 $6 = (+_luaL_optnumber($L,4,0.0)); //@line 206 "modules/audio/wrap_Source.cpp"
 $7 = $6; //@line 206 "modules/audio/wrap_Source.cpp"
 $8 = ((($v)) + 8|0); //@line 206 "modules/audio/wrap_Source.cpp"
 HEAPF32[$8>>2] = $7; //@line 206 "modules/audio/wrap_Source.cpp"
 $9 = HEAP32[$0>>2]|0; //@line 207 "modules/audio/wrap_Source.cpp"
 $10 = ((($9)) + 100|0); //@line 207 "modules/audio/wrap_Source.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 207 "modules/audio/wrap_Source.cpp"
 __THREW__ = 0;
 invoke_vii($11|0,($0|0),($v|0)); //@line 207 "modules/audio/wrap_Source.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if (!($13)) {
  STACKTOP = sp;return 0; //@line 208 "modules/audio/wrap_Source.cpp"
 }
 $14 = ___cxa_find_matching_catch_3(8992|0)|0;
 $15 = tempRet0;
 $16 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $17 = ($15|0)==($16|0); //@line 535 "./common/runtime.h"
 if (!($17)) {
  $$01$i = $15;$$02$i = $14;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $18 = (___cxa_begin_catch(($14|0))|0); //@line 535 "./common/runtime.h"
 $19 = HEAP32[$18>>2]|0; //@line 539 "./common/runtime.h"
 $20 = ((($19)) + 8|0); //@line 539 "./common/runtime.h"
 $21 = HEAP32[$20>>2]|0; //@line 539 "./common/runtime.h"
 $22 = (FUNCTION_TABLE_ii[$21 & 8191]($18)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($22|0)); //@line 539 "./common/runtime.h"
 $23 = __THREW__; __THREW__ = 0;
 $24 = $23&1;
 if (!($24)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $25 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $25; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 208 "modules/audio/wrap_Source.cpp"
 }
 $26 = ___cxa_find_matching_catch_2()|0;
 $27 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if ($29) {
  $30 = ___cxa_find_matching_catch_3(0|0)|0;
  $31 = tempRet0;
  ___clang_call_terminate($30); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $27;$$02$i = $26;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5audio21w_Source_getDirectionEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0.0;
 var $25 = 0.0, $26 = 0, $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $v = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $v = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 215 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 104|0); //@line 215 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 215 "modules/audio/wrap_Source.cpp"
 __THREW__ = 0;
 invoke_vii($3|0,($0|0),($v|0)); //@line 215 "modules/audio/wrap_Source.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 do {
  if ($5) {
   $6 = ___cxa_find_matching_catch_3(8992|0)|0;
   $7 = tempRet0;
   $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
   if (!($9)) {
    $$01$i = $7;$$02$i = $6;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
   $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
   $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
   $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
   $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if (!($16)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $17; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    break;
   }
   $18 = ___cxa_find_matching_catch_2()|0;
   $19 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if ($21) {
    $22 = ___cxa_find_matching_catch_3(0|0)|0;
    $23 = tempRet0;
    ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $19;$$02$i = $18;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  }
 } while(0);
 $24 = +HEAPF32[$v>>2]; //@line 216 "modules/audio/wrap_Source.cpp"
 $25 = $24; //@line 216 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$25); //@line 216 "modules/audio/wrap_Source.cpp"
 $26 = ((($v)) + 4|0); //@line 217 "modules/audio/wrap_Source.cpp"
 $27 = +HEAPF32[$26>>2]; //@line 217 "modules/audio/wrap_Source.cpp"
 $28 = $27; //@line 217 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$28); //@line 217 "modules/audio/wrap_Source.cpp"
 $29 = ((($v)) + 8|0); //@line 218 "modules/audio/wrap_Source.cpp"
 $30 = +HEAPF32[$29>>2]; //@line 218 "modules/audio/wrap_Source.cpp"
 $31 = $30; //@line 218 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$31); //@line 218 "modules/audio/wrap_Source.cpp"
 STACKTOP = sp;return 3; //@line 219 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio16w_Source_setConeEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 222 "modules/audio/wrap_Source.cpp"
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 225 "modules/audio/wrap_Source.cpp"
 $2 = $1; //@line 225 "modules/audio/wrap_Source.cpp"
 $3 = (+_luaL_checknumber($L,3)); //@line 226 "modules/audio/wrap_Source.cpp"
 $4 = $3; //@line 226 "modules/audio/wrap_Source.cpp"
 $5 = (+_luaL_optnumber($L,4,0.0)); //@line 227 "modules/audio/wrap_Source.cpp"
 $6 = $5; //@line 227 "modules/audio/wrap_Source.cpp"
 $7 = HEAP32[$0>>2]|0; //@line 228 "modules/audio/wrap_Source.cpp"
 $8 = ((($7)) + 108|0); //@line 228 "modules/audio/wrap_Source.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 228 "modules/audio/wrap_Source.cpp"
 __THREW__ = 0;
 invoke_viddd($9|0,($0|0),(+$2),(+$4),(+$6)); //@line 228 "modules/audio/wrap_Source.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if (!($11)) {
  STACKTOP = sp;return 0; //@line 229 "modules/audio/wrap_Source.cpp"
 }
 $12 = ___cxa_find_matching_catch_3(8992|0)|0;
 $13 = tempRet0;
 $14 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $15 = ($13|0)==($14|0); //@line 535 "./common/runtime.h"
 if (!($15)) {
  $$01$i = $13;$$02$i = $12;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $16 = (___cxa_begin_catch(($12|0))|0); //@line 535 "./common/runtime.h"
 $17 = HEAP32[$16>>2]|0; //@line 539 "./common/runtime.h"
 $18 = ((($17)) + 8|0); //@line 539 "./common/runtime.h"
 $19 = HEAP32[$18>>2]|0; //@line 539 "./common/runtime.h"
 $20 = (FUNCTION_TABLE_ii[$19 & 8191]($16)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($20|0)); //@line 539 "./common/runtime.h"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if (!($22)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $23 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $23; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 229 "modules/audio/wrap_Source.cpp"
 }
 $24 = ___cxa_find_matching_catch_2()|0;
 $25 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  $28 = ___cxa_find_matching_catch_3(0|0)|0;
  $29 = tempRet0;
  ___clang_call_terminate($28); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $25;$$02$i = $24;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5audio16w_Source_getConeEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0.0;
 var $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $innerAngle = 0, $outerAngle = 0, $outerVolume = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $innerAngle = sp + 12|0;
 $outerAngle = sp + 8|0;
 $outerVolume = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 236 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 112|0); //@line 236 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 236 "modules/audio/wrap_Source.cpp"
 __THREW__ = 0;
 invoke_viiii($3|0,($0|0),($innerAngle|0),($outerAngle|0),($outerVolume|0)); //@line 236 "modules/audio/wrap_Source.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 do {
  if ($5) {
   $6 = ___cxa_find_matching_catch_3(8992|0)|0;
   $7 = tempRet0;
   $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
   if (!($9)) {
    $$01$i = $7;$$02$i = $6;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
   $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
   $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
   $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
   $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if (!($16)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $17; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    break;
   }
   $18 = ___cxa_find_matching_catch_2()|0;
   $19 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if ($21) {
    $22 = ___cxa_find_matching_catch_3(0|0)|0;
    $23 = tempRet0;
    ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $19;$$02$i = $18;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  }
 } while(0);
 $24 = +HEAPF32[$innerAngle>>2]; //@line 237 "modules/audio/wrap_Source.cpp"
 $25 = $24; //@line 237 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$25); //@line 237 "modules/audio/wrap_Source.cpp"
 $26 = +HEAPF32[$outerAngle>>2]; //@line 238 "modules/audio/wrap_Source.cpp"
 $27 = $26; //@line 238 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$27); //@line 238 "modules/audio/wrap_Source.cpp"
 $28 = +HEAPF32[$outerVolume>>2]; //@line 239 "modules/audio/wrap_Source.cpp"
 $29 = $28; //@line 239 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$29); //@line 239 "modules/audio/wrap_Source.cpp"
 STACKTOP = sp;return 3; //@line 240 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio20w_Source_setRelativeEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 243 "modules/audio/wrap_Source.cpp"
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 246 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 116|0); //@line 246 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 246 "modules/audio/wrap_Source.cpp"
 __THREW__ = 0;
 $4 = (invoke_iii(3504,($L|0),2)|0); //@line 246 "modules/audio/wrap_Source.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  __THREW__ = 0;
  invoke_vii($3|0,($0|0),($4|0)); //@line 246 "modules/audio/wrap_Source.cpp"
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if (!($8)) {
   STACKTOP = sp;return 0; //@line 247 "modules/audio/wrap_Source.cpp"
  }
 }
 $9 = ___cxa_find_matching_catch_3(8992|0)|0;
 $10 = tempRet0;
 $11 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $12 = ($10|0)==($11|0); //@line 535 "./common/runtime.h"
 if (!($12)) {
  $$01$i = $10;$$02$i = $9;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $13 = (___cxa_begin_catch(($9|0))|0); //@line 535 "./common/runtime.h"
 $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
 $15 = ((($14)) + 8|0); //@line 539 "./common/runtime.h"
 $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
 $17 = (FUNCTION_TABLE_ii[$16 & 8191]($13)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($17|0)); //@line 539 "./common/runtime.h"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if (!($19)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $20 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $20; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 247 "modules/audio/wrap_Source.cpp"
 }
 $21 = ___cxa_find_matching_catch_2()|0;
 $22 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $23 = __THREW__; __THREW__ = 0;
 $24 = $23&1;
 if ($24) {
  $25 = ___cxa_find_matching_catch_3(0|0)|0;
  $26 = tempRet0;
  ___clang_call_terminate($25); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $22;$$02$i = $21;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5audio19w_Source_isRelativeEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 250 "modules/audio/wrap_Source.cpp"
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 253 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 120|0); //@line 253 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 253 "modules/audio/wrap_Source.cpp"
 __THREW__ = 0;
 $4 = (invoke_ii($3|0,($0|0))|0); //@line 253 "modules/audio/wrap_Source.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  __THREW__ = 0;
  invoke_vii(3397,($L|0),($4|0)); //@line 253 "modules/audio/wrap_Source.cpp"
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if (!($8)) {
   STACKTOP = sp;return 1; //@line 254 "modules/audio/wrap_Source.cpp"
  }
 }
 $9 = ___cxa_find_matching_catch_3(8992|0)|0;
 $10 = tempRet0;
 $11 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $12 = ($10|0)==($11|0); //@line 535 "./common/runtime.h"
 if (!($12)) {
  $$01$i = $10;$$02$i = $9;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $13 = (___cxa_begin_catch(($9|0))|0); //@line 535 "./common/runtime.h"
 $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
 $15 = ((($14)) + 8|0); //@line 539 "./common/runtime.h"
 $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
 $17 = (FUNCTION_TABLE_ii[$16 & 8191]($13)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($17|0)); //@line 539 "./common/runtime.h"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if (!($19)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $20 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $20; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 1; //@line 254 "modules/audio/wrap_Source.cpp"
 }
 $21 = ___cxa_find_matching_catch_2()|0;
 $22 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $23 = __THREW__; __THREW__ = 0;
 $24 = $23&1;
 if ($24) {
  $25 = ___cxa_find_matching_catch_3(0|0)|0;
  $26 = tempRet0;
  ___clang_call_terminate($25); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $22;$$02$i = $21;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5audio19w_Source_setLoopingEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 260 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 124|0); //@line 260 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 260 "modules/audio/wrap_Source.cpp"
 $4 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,2)|0); //@line 260 "modules/audio/wrap_Source.cpp"
 FUNCTION_TABLE_vii[$3 & 8191]($0,$4); //@line 260 "modules/audio/wrap_Source.cpp"
 return 0; //@line 261 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio18w_Source_isLoopingEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 267 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 128|0); //@line 267 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 267 "modules/audio/wrap_Source.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 267 "modules/audio/wrap_Source.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 267 "modules/audio/wrap_Source.cpp"
 return 1; //@line 268 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio18w_Source_isStoppedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 274 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 40|0); //@line 274 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 274 "modules/audio/wrap_Source.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 274 "modules/audio/wrap_Source.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 274 "modules/audio/wrap_Source.cpp"
 return 1; //@line 275 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio17w_Source_isPausedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 281 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 44|0); //@line 281 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 281 "modules/audio/wrap_Source.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 281 "modules/audio/wrap_Source.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 281 "modules/audio/wrap_Source.cpp"
 return 1; //@line 282 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio18w_Source_isPlayingEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 288 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 40|0); //@line 288 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 288 "modules/audio/wrap_Source.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 288 "modules/audio/wrap_Source.cpp"
 if ($4) {
  $10 = 0;
  __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$10); //@line 288 "modules/audio/wrap_Source.cpp"
  return 1; //@line 289 "modules/audio/wrap_Source.cpp"
 }
 $5 = HEAP32[$0>>2]|0; //@line 288 "modules/audio/wrap_Source.cpp"
 $6 = ((($5)) + 44|0); //@line 288 "modules/audio/wrap_Source.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 288 "modules/audio/wrap_Source.cpp"
 $8 = (FUNCTION_TABLE_ii[$7 & 8191]($0)|0); //@line 288 "modules/audio/wrap_Source.cpp"
 $9 = $8 ^ 1; //@line 288 "modules/audio/wrap_Source.cpp"
 $10 = $9;
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$10); //@line 288 "modules/audio/wrap_Source.cpp"
 return 1; //@line 289 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio24w_Source_setVolumeLimitsEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, $or$cond = 0;
 var $or$cond3 = 0, $or$cond5 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 292 "modules/audio/wrap_Source.cpp"
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 295 "modules/audio/wrap_Source.cpp"
 $2 = $1; //@line 295 "modules/audio/wrap_Source.cpp"
 $3 = (+_luaL_checknumber($L,3)); //@line 296 "modules/audio/wrap_Source.cpp"
 $4 = $3; //@line 296 "modules/audio/wrap_Source.cpp"
 $5 = $2 < 0.0; //@line 297 "modules/audio/wrap_Source.cpp"
 $6 = $2 > 1.0; //@line 297 "modules/audio/wrap_Source.cpp"
 $or$cond = $5 | $6; //@line 297 "modules/audio/wrap_Source.cpp"
 $7 = $4 < 0.0; //@line 297 "modules/audio/wrap_Source.cpp"
 $or$cond3 = $or$cond | $7; //@line 297 "modules/audio/wrap_Source.cpp"
 $8 = $4 > 1.0; //@line 297 "modules/audio/wrap_Source.cpp"
 $or$cond5 = $8 | $or$cond3; //@line 297 "modules/audio/wrap_Source.cpp"
 if ($or$cond5) {
  $9 = $2; //@line 298 "modules/audio/wrap_Source.cpp"
  $10 = $4; //@line 298 "modules/audio/wrap_Source.cpp"
  HEAPF64[$vararg_buffer>>3] = $9; //@line 298 "modules/audio/wrap_Source.cpp"
  $vararg_ptr1 = ((($vararg_buffer)) + 8|0); //@line 298 "modules/audio/wrap_Source.cpp"
  HEAPF64[$vararg_ptr1>>3] = $10; //@line 298 "modules/audio/wrap_Source.cpp"
  $11 = (_luaL_error($L,327866,$vararg_buffer)|0); //@line 298 "modules/audio/wrap_Source.cpp"
  $$0 = $11;
  STACKTOP = sp;return ($$0|0); //@line 302 "modules/audio/wrap_Source.cpp"
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 299 "modules/audio/wrap_Source.cpp"
  $13 = ((($12)) + 132|0); //@line 299 "modules/audio/wrap_Source.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 299 "modules/audio/wrap_Source.cpp"
  FUNCTION_TABLE_vid[$14 & 8191]($0,$2); //@line 299 "modules/audio/wrap_Source.cpp"
  $15 = HEAP32[$0>>2]|0; //@line 300 "modules/audio/wrap_Source.cpp"
  $16 = ((($15)) + 140|0); //@line 300 "modules/audio/wrap_Source.cpp"
  $17 = HEAP32[$16>>2]|0; //@line 300 "modules/audio/wrap_Source.cpp"
  FUNCTION_TABLE_vid[$17 & 8191]($0,$4); //@line 300 "modules/audio/wrap_Source.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 302 "modules/audio/wrap_Source.cpp"
 }
 return (0)|0;
}
function __ZN4love5audio24w_Source_getVolumeLimitsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 307 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 136|0); //@line 307 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 307 "modules/audio/wrap_Source.cpp"
 $4 = (+FUNCTION_TABLE_di[$3 & 4095]($0)); //@line 307 "modules/audio/wrap_Source.cpp"
 $5 = $4; //@line 307 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$5); //@line 307 "modules/audio/wrap_Source.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 308 "modules/audio/wrap_Source.cpp"
 $7 = ((($6)) + 144|0); //@line 308 "modules/audio/wrap_Source.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 308 "modules/audio/wrap_Source.cpp"
 $9 = (+FUNCTION_TABLE_di[$8 & 4095]($0)); //@line 308 "modules/audio/wrap_Source.cpp"
 $10 = $9; //@line 308 "modules/audio/wrap_Source.cpp"
 _lua_pushnumber($L,$10); //@line 308 "modules/audio/wrap_Source.cpp"
 return 2; //@line 309 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio32w_Source_setAttenuationDistancesEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0.0;
 var $9 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 16|0; //@line 312 "modules/audio/wrap_Source.cpp"
 $vararg_buffer = sp; //@line 312 "modules/audio/wrap_Source.cpp"
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 315 "modules/audio/wrap_Source.cpp"
 $2 = $1; //@line 315 "modules/audio/wrap_Source.cpp"
 $3 = (+_luaL_checknumber($L,3)); //@line 316 "modules/audio/wrap_Source.cpp"
 $4 = $3; //@line 316 "modules/audio/wrap_Source.cpp"
 $5 = $2 < 0.0; //@line 317 "modules/audio/wrap_Source.cpp"
 $6 = $4 < 0.0; //@line 317 "modules/audio/wrap_Source.cpp"
 $or$cond = $5 | $6; //@line 317 "modules/audio/wrap_Source.cpp"
 if ($or$cond) {
  $7 = $2; //@line 318 "modules/audio/wrap_Source.cpp"
  $8 = $4; //@line 318 "modules/audio/wrap_Source.cpp"
  HEAPF64[$vararg_buffer>>3] = $7; //@line 318 "modules/audio/wrap_Source.cpp"
  $vararg_ptr1 = ((($vararg_buffer)) + 8|0); //@line 318 "modules/audio/wrap_Source.cpp"
  HEAPF64[$vararg_ptr1>>3] = $8; //@line 318 "modules/audio/wrap_Source.cpp"
  $9 = (_luaL_error($L,327915,$vararg_buffer)|0); //@line 318 "modules/audio/wrap_Source.cpp"
  $$0 = $9;
  STACKTOP = sp;return ($$0|0); //@line 324 "modules/audio/wrap_Source.cpp"
 }
 $10 = HEAP32[$0>>2]|0; //@line 320 "modules/audio/wrap_Source.cpp"
 $11 = ((($10)) + 148|0); //@line 320 "modules/audio/wrap_Source.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 320 "modules/audio/wrap_Source.cpp"
 __THREW__ = 0;
 invoke_vid($12|0,($0|0),(+$2)); //@line 320 "modules/audio/wrap_Source.cpp"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if (!($14)) {
  $15 = HEAP32[$0>>2]|0; //@line 321 "modules/audio/wrap_Source.cpp"
  $16 = ((($15)) + 164|0); //@line 321 "modules/audio/wrap_Source.cpp"
  $17 = HEAP32[$16>>2]|0; //@line 321 "modules/audio/wrap_Source.cpp"
  __THREW__ = 0;
  invoke_vid($17|0,($0|0),(+$4)); //@line 321 "modules/audio/wrap_Source.cpp"
  $18 = __THREW__; __THREW__ = 0;
  $19 = $18&1;
  if (!($19)) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0); //@line 324 "modules/audio/wrap_Source.cpp"
  }
 }
 $20 = ___cxa_find_matching_catch_3(8992|0)|0;
 $21 = tempRet0;
 $22 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $23 = ($21|0)==($22|0); //@line 535 "./common/runtime.h"
 if (!($23)) {
  $$01$i = $21;$$02$i = $20;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $24 = (___cxa_begin_catch(($20|0))|0); //@line 535 "./common/runtime.h"
 $25 = HEAP32[$24>>2]|0; //@line 539 "./common/runtime.h"
 $26 = ((($25)) + 8|0); //@line 539 "./common/runtime.h"
 $27 = HEAP32[$26>>2]|0; //@line 539 "./common/runtime.h"
 $28 = (FUNCTION_TABLE_ii[$27 & 8191]($24)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($28|0)); //@line 539 "./common/runtime.h"
 $29 = __THREW__; __THREW__ = 0;
 $30 = $29&1;
 if (!($30)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $31 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer2>>2] = $31; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer2)|0); //@line 543 "./common/runtime.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 324 "modules/audio/wrap_Source.cpp"
 }
 $32 = ___cxa_find_matching_catch_2()|0;
 $33 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $34 = __THREW__; __THREW__ = 0;
 $35 = $34&1;
 if ($35) {
  $36 = ___cxa_find_matching_catch_3(0|0)|0;
  $37 = tempRet0;
  ___clang_call_terminate($36); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $33;$$02$i = $32;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5audio32w_Source_getAttenuationDistancesEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0, $vararg_buffer = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 326 "modules/audio/wrap_Source.cpp"
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 330 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 152|0); //@line 330 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 330 "modules/audio/wrap_Source.cpp"
 __THREW__ = 0;
 $4 = (+invoke_di($3|0,($0|0))); //@line 330 "modules/audio/wrap_Source.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  $7 = $4; //@line 330 "modules/audio/wrap_Source.cpp"
  __THREW__ = 0;
  invoke_vid(3394,($L|0),(+$7)); //@line 330 "modules/audio/wrap_Source.cpp"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if (!($9)) {
   $10 = HEAP32[$0>>2]|0; //@line 331 "modules/audio/wrap_Source.cpp"
   $11 = ((($10)) + 168|0); //@line 331 "modules/audio/wrap_Source.cpp"
   $12 = HEAP32[$11>>2]|0; //@line 331 "modules/audio/wrap_Source.cpp"
   __THREW__ = 0;
   $13 = (+invoke_di($12|0,($0|0))); //@line 331 "modules/audio/wrap_Source.cpp"
   $14 = __THREW__; __THREW__ = 0;
   $15 = $14&1;
   if (!($15)) {
    $16 = $13; //@line 331 "modules/audio/wrap_Source.cpp"
    __THREW__ = 0;
    invoke_vid(3394,($L|0),(+$16)); //@line 331 "modules/audio/wrap_Source.cpp"
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if (!($18)) {
     STACKTOP = sp;return 2; //@line 333 "modules/audio/wrap_Source.cpp"
    }
   }
  }
 }
 $19 = ___cxa_find_matching_catch_3(8992|0)|0;
 $20 = tempRet0;
 $21 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $22 = ($20|0)==($21|0); //@line 535 "./common/runtime.h"
 if (!($22)) {
  $$01$i = $20;$$02$i = $19;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $23 = (___cxa_begin_catch(($19|0))|0); //@line 535 "./common/runtime.h"
 $24 = HEAP32[$23>>2]|0; //@line 539 "./common/runtime.h"
 $25 = ((($24)) + 8|0); //@line 539 "./common/runtime.h"
 $26 = HEAP32[$25>>2]|0; //@line 539 "./common/runtime.h"
 $27 = (FUNCTION_TABLE_ii[$26 & 8191]($23)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($27|0)); //@line 539 "./common/runtime.h"
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if (!($29)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $30 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $30; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 2; //@line 333 "modules/audio/wrap_Source.cpp"
 }
 $31 = ___cxa_find_matching_catch_2()|0;
 $32 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $33 = __THREW__; __THREW__ = 0;
 $34 = $33&1;
 if ($34) {
  $35 = ___cxa_find_matching_catch_3(0|0)|0;
  $36 = tempRet0;
  ___clang_call_terminate($35); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $32;$$02$i = $31;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5audio19w_Source_setRolloffEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 336 "modules/audio/wrap_Source.cpp"
 $vararg_buffer = sp; //@line 336 "modules/audio/wrap_Source.cpp"
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 339 "modules/audio/wrap_Source.cpp"
 $2 = $1; //@line 339 "modules/audio/wrap_Source.cpp"
 $3 = $2 < 0.0; //@line 340 "modules/audio/wrap_Source.cpp"
 if ($3) {
  $4 = $2; //@line 341 "modules/audio/wrap_Source.cpp"
  HEAPF64[$vararg_buffer>>3] = $4; //@line 341 "modules/audio/wrap_Source.cpp"
  $5 = (_luaL_error($L,327954,$vararg_buffer)|0); //@line 341 "modules/audio/wrap_Source.cpp"
  $$0 = $5;
  STACKTOP = sp;return ($$0|0); //@line 344 "modules/audio/wrap_Source.cpp"
 }
 $6 = HEAP32[$0>>2]|0; //@line 342 "modules/audio/wrap_Source.cpp"
 $7 = ((($6)) + 156|0); //@line 342 "modules/audio/wrap_Source.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 342 "modules/audio/wrap_Source.cpp"
 __THREW__ = 0;
 invoke_vid($8|0,($0|0),(+$2)); //@line 342 "modules/audio/wrap_Source.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if (!($10)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 344 "modules/audio/wrap_Source.cpp"
 }
 $11 = ___cxa_find_matching_catch_3(8992|0)|0;
 $12 = tempRet0;
 $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $14 = ($12|0)==($13|0); //@line 535 "./common/runtime.h"
 if (!($14)) {
  $$01$i = $12;$$02$i = $11;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $15 = (___cxa_begin_catch(($11|0))|0); //@line 535 "./common/runtime.h"
 $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
 $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
 $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
 $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if (!($21)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $22; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 344 "modules/audio/wrap_Source.cpp"
 }
 $23 = ___cxa_find_matching_catch_2()|0;
 $24 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if ($26) {
  $27 = ___cxa_find_matching_catch_3(0|0)|0;
  $28 = tempRet0;
  ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $24;$$02$i = $23;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5audio19w_Source_getRolloffEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 346 "modules/audio/wrap_Source.cpp"
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 349 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 160|0); //@line 349 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 349 "modules/audio/wrap_Source.cpp"
 __THREW__ = 0;
 $4 = (+invoke_di($3|0,($0|0))); //@line 349 "modules/audio/wrap_Source.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  $7 = $4; //@line 349 "modules/audio/wrap_Source.cpp"
  __THREW__ = 0;
  invoke_vid(3394,($L|0),(+$7)); //@line 349 "modules/audio/wrap_Source.cpp"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if (!($9)) {
   STACKTOP = sp;return 1; //@line 350 "modules/audio/wrap_Source.cpp"
  }
 }
 $10 = ___cxa_find_matching_catch_3(8992|0)|0;
 $11 = tempRet0;
 $12 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $13 = ($11|0)==($12|0); //@line 535 "./common/runtime.h"
 if (!($13)) {
  $$01$i = $11;$$02$i = $10;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $14 = (___cxa_begin_catch(($10|0))|0); //@line 535 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = ((($15)) + 8|0); //@line 539 "./common/runtime.h"
 $17 = HEAP32[$16>>2]|0; //@line 539 "./common/runtime.h"
 $18 = (FUNCTION_TABLE_ii[$17 & 8191]($14)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($18|0)); //@line 539 "./common/runtime.h"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if (!($20)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $21 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $21; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 1; //@line 350 "modules/audio/wrap_Source.cpp"
 }
 $22 = ___cxa_find_matching_catch_2()|0;
 $23 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $26 = ___cxa_find_matching_catch_3(0|0)|0;
  $27 = tempRet0;
  ___clang_call_terminate($26); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $23;$$02$i = $22;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5audio20w_Source_getChannelsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 356 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 172|0); //@line 356 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 356 "modules/audio/wrap_Source.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 356 "modules/audio/wrap_Source.cpp"
 _lua_pushinteger($L,$4); //@line 356 "modules/audio/wrap_Source.cpp"
 return 1; //@line 357 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5audio16w_Source_getTypeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $str = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $str = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,1,28)|0); //@line 32 "modules/audio/wrap_Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 363 "modules/audio/wrap_Source.cpp"
 $2 = ((($1)) + 176|0); //@line 363 "modules/audio/wrap_Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 363 "modules/audio/wrap_Source.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 363 "modules/audio/wrap_Source.cpp"
 HEAP32[$str>>2] = 0; //@line 364 "modules/audio/wrap_Source.cpp"
 $5 = (__ZN4love5audio6Source11getConstantENS1_4TypeERPKc($4,$str)|0); //@line 366 "modules/audio/wrap_Source.cpp"
 if ($5) {
  $7 = HEAP32[$str>>2]|0; //@line 369 "modules/audio/wrap_Source.cpp"
  _lua_pushstring($L,$7); //@line 369 "modules/audio/wrap_Source.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 371 "modules/audio/wrap_Source.cpp"
 } else {
  $6 = (_luaL_error($L,327988,$vararg_buffer)|0); //@line 367 "modules/audio/wrap_Source.cpp"
  $$0 = $6;
  STACKTOP = sp;return ($$0|0); //@line 371 "modules/audio/wrap_Source.cpp"
 }
 return (0)|0;
}
function _luaopen_source($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 421 "modules/audio/wrap_Source.cpp"
 HEAP32[$vararg_buffer>>2] = 12472; //@line 423 "modules/audio/wrap_Source.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 423 "modules/audio/wrap_Source.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 423 "modules/audio/wrap_Source.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,28,328009,$vararg_buffer)|0); //@line 423 "modules/audio/wrap_Source.cpp"
 STACKTOP = sp;return ($0|0); //@line 423 "modules/audio/wrap_Source.cpp"
}
function __ZN4love5event6w_pollEP9lua_State($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _lua_pushcclosure($L,3505,0); //@line 52 "modules/event/wrap_Event.cpp"
 return 1; //@line 53 "modules/event/wrap_Event.cpp"
}
function __ZN4love5eventL8w_poll_iEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $m = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $m = sp;
 HEAP32[$m>>2] = 0; //@line 37 "modules/event/wrap_Event.cpp"
 $0 = HEAP32[(625612)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZN4love5event5Event4pollERPNS0_7MessageE($0,$m)|0); //@line 39 "modules/event/wrap_Event.cpp"
 if (!($1)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 48 "modules/event/wrap_Event.cpp"
 }
 $2 = HEAP32[$m>>2]|0; //@line 41 "modules/event/wrap_Event.cpp"
 $3 = (__ZN4love5event7Message5toLuaEP9lua_State($2,$L)|0); //@line 41 "modules/event/wrap_Event.cpp"
 $4 = HEAP32[$m>>2]|0; //@line 42 "modules/event/wrap_Event.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 42 "modules/event/wrap_Event.cpp"
 $6 = ((($5)) + 12|0); //@line 42 "modules/event/wrap_Event.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 42 "modules/event/wrap_Event.cpp"
 FUNCTION_TABLE_vi[$7 & 8191]($4); //@line 42 "modules/event/wrap_Event.cpp"
 $$0 = $3;
 STACKTOP = sp;return ($$0|0); //@line 48 "modules/event/wrap_Event.cpp"
}
function __ZN4love5event6w_pumpEP9lua_State($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[(625612)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 58 "modules/event/wrap_Event.cpp"
 $3 = ((($2)) + 28|0); //@line 58 "modules/event/wrap_Event.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 58 "modules/event/wrap_Event.cpp"
 FUNCTION_TABLE_vi[$4 & 8191]($1); //@line 58 "modules/event/wrap_Event.cpp"
 return 0; //@line 59 "modules/event/wrap_Event.cpp"
}
function __ZN4love5event6w_waitEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625612)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 64 "modules/event/wrap_Event.cpp"
 $2 = ((($1)) + 32|0); //@line 64 "modules/event/wrap_Event.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 64 "modules/event/wrap_Event.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 64 "modules/event/wrap_Event.cpp"
 $5 = ($4|0)==(0|0); //@line 65 "modules/event/wrap_Event.cpp"
 if ($5) {
  $$0 = 0;
  return ($$0|0); //@line 73 "modules/event/wrap_Event.cpp"
 }
 $6 = (__ZN4love5event7Message5toLuaEP9lua_State($4,$L)|0); //@line 67 "modules/event/wrap_Event.cpp"
 $7 = HEAP32[$4>>2]|0; //@line 68 "modules/event/wrap_Event.cpp"
 $8 = ((($7)) + 12|0); //@line 68 "modules/event/wrap_Event.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 68 "modules/event/wrap_Event.cpp"
 FUNCTION_TABLE_vi[$9 & 8191]($4); //@line 68 "modules/event/wrap_Event.cpp"
 $$0 = $6;
 return ($$0|0); //@line 73 "modules/event/wrap_Event.cpp"
}
function __ZN4love5event6w_pushEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love5event7Message7fromLuaEP9lua_Statei($L,1)|0); //@line 77 "modules/event/wrap_Event.cpp"
 $1 = ($0|0)!=(0|0); //@line 79 "modules/event/wrap_Event.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$1); //@line 79 "modules/event/wrap_Event.cpp"
 $2 = ($0|0)==(0|0); //@line 81 "modules/event/wrap_Event.cpp"
 if ($2) {
  return 1; //@line 88 "modules/event/wrap_Event.cpp"
 }
 $3 = HEAP32[(625612)>>2]|0; //@line 98 "./common/Module.h"
 __ZN4love5event5Event4pushEPNS0_7MessageE($3,$0); //@line 84 "modules/event/wrap_Event.cpp"
 $4 = HEAP32[$0>>2]|0; //@line 85 "modules/event/wrap_Event.cpp"
 $5 = ((($4)) + 12|0); //@line 85 "modules/event/wrap_Event.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 85 "modules/event/wrap_Event.cpp"
 FUNCTION_TABLE_vi[$6 & 8191]($0); //@line 85 "modules/event/wrap_Event.cpp"
 return 1; //@line 88 "modules/event/wrap_Event.cpp"
}
function __ZN4love5event7w_clearEP9lua_State($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[(625612)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 92 "modules/event/wrap_Event.cpp"
 $3 = ((($2)) + 24|0); //@line 92 "modules/event/wrap_Event.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 92 "modules/event/wrap_Event.cpp"
 FUNCTION_TABLE_vi[$4 & 8191]($1); //@line 92 "modules/event/wrap_Event.cpp"
 return 0; //@line 93 "modules/event/wrap_Event.cpp"
}
function __ZN4love5event6w_quitEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$pre$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i10 = 0, $$pre$i13 = 0, $$pre$i3 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $11 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $args = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $args = sp + 16|0;
 $0 = sp + 12|0;
 $1 = sp;
 HEAP32[$args>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($args)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($args)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$3>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 $4 = (invoke_iiii(3506,($L|0),1,1)|0); //@line 106 "./common/Object.h"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 L1: do {
  if ($6) {
   label = 11;
  } else {
   $7 = ($4|0)==(0|0); //@line 101 "modules/event/wrap_Event.cpp"
   do {
    if (!($7)) {
     HEAP32[$0>>2] = $4; //@line 93 "./common/Object.h"
     $8 = HEAP32[$4>>2]|0; //@line 95 "./common/Object.h"
     $9 = ((($8)) + 8|0); //@line 95 "./common/Object.h"
     $10 = HEAP32[$9>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($10|0,($4|0)); //@line 95 "./common/Object.h"
     $11 = __THREW__; __THREW__ = 0;
     $12 = $11&1;
     if ($12) {
      label = 11;
      break L1;
     }
     $13 = HEAP32[$2>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $14 = ((($args)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $15 = HEAP32[$14>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $16 = ($13>>>0)<($15>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($16) {
      HEAP32[$13>>2] = $4; //@line 99 "./common/Object.h"
      $17 = HEAP32[$4>>2]|0; //@line 101 "./common/Object.h"
      $18 = ((($17)) + 8|0); //@line 101 "./common/Object.h"
      $19 = HEAP32[$18>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($19|0,($4|0)); //@line 101 "./common/Object.h"
      $20 = __THREW__; __THREW__ = 0;
      $21 = $20&1;
      if ($21) {
       label = 12;
      } else {
       $$pre$i3 = HEAP32[$2>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $22 = ((($$pre$i3)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$2>>2] = $22; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($args|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $23 = __THREW__; __THREW__ = 0;
      $24 = $23&1;
      if ($24) {
       label = 12;
      }
     }
     if ((label|0) == 12) {
      $39 = ___cxa_find_matching_catch_2()|0;
      $40 = tempRet0;
      $41 = HEAP32[$4>>2]|0; //@line 106 "./common/Object.h"
      $42 = ((($41)) + 12|0); //@line 106 "./common/Object.h"
      $43 = HEAP32[$42>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($43|0,($4|0)); //@line 106 "./common/Object.h"
      $44 = __THREW__; __THREW__ = 0;
      $45 = $44&1;
      if (!($45)) {
       $$1 = $39;$$12 = $40;
       break L1;
      }
      $46 = ___cxa_find_matching_catch_3(0|0)|0;
      $47 = tempRet0;
      ___clang_call_terminate($46); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $25 = HEAP32[$4>>2]|0; //@line 106 "./common/Object.h"
     $26 = ((($25)) + 12|0); //@line 106 "./common/Object.h"
     $27 = HEAP32[$26>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($27|0,($4|0)); //@line 106 "./common/Object.h"
     $28 = __THREW__; __THREW__ = 0;
     $29 = $28&1;
     if ($29) {
      $30 = ___cxa_find_matching_catch_3(0|0)|0;
      $31 = tempRet0;
      ___clang_call_terminate($30); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     } else {
      $32 = HEAP32[$4>>2]|0; //@line 104 "modules/event/wrap_Event.cpp"
      $33 = ((($32)) + 12|0); //@line 104 "modules/event/wrap_Event.cpp"
      $34 = HEAP32[$33>>2]|0; //@line 104 "modules/event/wrap_Event.cpp"
      __THREW__ = 0;
      invoke_vi($34|0,($4|0)); //@line 104 "modules/event/wrap_Event.cpp"
      $35 = __THREW__; __THREW__ = 0;
      $36 = $35&1;
      if ($36) {
       label = 11;
       break L1;
      } else {
       break;
      }
     }
    }
   } while(0);
   __THREW__ = 0;
   $48 = (invoke_ii(3348,32)|0); //@line 107 "modules/event/wrap_Event.cpp"
   $49 = __THREW__; __THREW__ = 0;
   $50 = $49&1;
   if ($50) {
    label = 11;
   } else {
    __THREW__ = 0;
    invoke_viii(3460,($1|0),(365855|0),4); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $51 = __THREW__; __THREW__ = 0;
    $52 = $51&1;
    do {
     if ($52) {
      $83 = ___cxa_find_matching_catch_2()|0;
      $84 = tempRet0;
      $$0 = $83;$$01 = $84;
     } else {
      __THREW__ = 0;
      invoke_viii(3508,($48|0),($1|0),($args|0)); //@line 107 "modules/event/wrap_Event.cpp"
      $53 = __THREW__; __THREW__ = 0;
      $54 = $53&1;
      if ($54) {
       $85 = ___cxa_find_matching_catch_2()|0;
       $86 = tempRet0;
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 107 "modules/event/wrap_Event.cpp"
       $$0 = $85;$$01 = $86;
       break;
      }
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 107 "modules/event/wrap_Event.cpp"
      $55 = HEAP32[(625612)>>2]|0; //@line 98 "./common/Module.h"
      __THREW__ = 0;
      invoke_vii(3509,($55|0),($48|0)); //@line 108 "modules/event/wrap_Event.cpp"
      $56 = __THREW__; __THREW__ = 0;
      $57 = $56&1;
      if ($57) {
       label = 11;
       break L1;
      }
      $58 = HEAP32[$48>>2]|0; //@line 109 "modules/event/wrap_Event.cpp"
      $59 = ((($58)) + 12|0); //@line 109 "modules/event/wrap_Event.cpp"
      $60 = HEAP32[$59>>2]|0; //@line 109 "modules/event/wrap_Event.cpp"
      __THREW__ = 0;
      invoke_vi($60|0,($48|0)); //@line 109 "modules/event/wrap_Event.cpp"
      $61 = __THREW__; __THREW__ = 0;
      $62 = $61&1;
      if ($62) {
       label = 11;
       break L1;
      }
      __THREW__ = 0;
      invoke_vii(3397,($L|0),1); //@line 111 "modules/event/wrap_Event.cpp"
      $63 = __THREW__; __THREW__ = 0;
      $64 = $63&1;
      if ($64) {
       label = 11;
       break L1;
      }
      $65 = HEAP32[$args>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $66 = ($65|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($66) {
       STACKTOP = sp;return 1; //@line 113 "modules/event/wrap_Event.cpp"
      }
      $67 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $68 = ($67|0)==($65|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      do {
       if ($68) {
        $82 = $65;
       } else {
        $70 = $67;
        while(1) {
         $69 = ((($70)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$2>>2] = $69; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $71 = HEAP32[$69>>2]|0; //@line 106 "./common/Object.h"
         $72 = ($71|0)==(0|0); //@line 106 "./common/Object.h"
         if ($72) {
          $74 = $69;
         } else {
          $75 = HEAP32[$71>>2]|0; //@line 106 "./common/Object.h"
          $76 = ((($75)) + 12|0); //@line 106 "./common/Object.h"
          $77 = HEAP32[$76>>2]|0; //@line 106 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($77|0,($71|0)); //@line 106 "./common/Object.h"
          $78 = __THREW__; __THREW__ = 0;
          $79 = $78&1;
          if ($79) {
           label = 26;
           break;
          }
          $$pre$i$i$i10 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $74 = $$pre$i$i$i10;
         }
         $73 = ($74|0)==($65|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($73) {
          label = 27;
          break;
         } else {
          $70 = $74;
         }
        }
        if ((label|0) == 26) {
         $80 = ___cxa_find_matching_catch_3(0|0)|0;
         $81 = tempRet0;
         ___clang_call_terminate($80); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        else if ((label|0) == 27) {
         $$pre$i13 = HEAP32[$args>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $82 = $$pre$i13;
         break;
        }
       }
      } while(0);
      __ZdlPv($82); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      STACKTOP = sp;return 1; //@line 113 "modules/event/wrap_Event.cpp"
     }
    } while(0);
    __ZdlPv($48); //@line 107 "modules/event/wrap_Event.cpp"
    $$1 = $$0;$$12 = $$01;
   }
  }
 } while(0);
 if ((label|0) == 11) {
  $37 = ___cxa_find_matching_catch_2()|0;
  $38 = tempRet0;
  $$1 = $37;$$12 = $38;
 }
 $87 = HEAP32[$args>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $88 = ($87|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($88) {
  ___resumeException($$1|0);
  // unreachable;
 }
 $89 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $90 = ($89|0)==($87|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($90) {
   $104 = $87;
  } else {
   $92 = $89;
   while(1) {
    $91 = ((($92)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$2>>2] = $91; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $93 = HEAP32[$91>>2]|0; //@line 106 "./common/Object.h"
    $94 = ($93|0)==(0|0); //@line 106 "./common/Object.h"
    if ($94) {
     $96 = $91;
    } else {
     $97 = HEAP32[$93>>2]|0; //@line 106 "./common/Object.h"
     $98 = ((($97)) + 12|0); //@line 106 "./common/Object.h"
     $99 = HEAP32[$98>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($99|0,($93|0)); //@line 106 "./common/Object.h"
     $100 = __THREW__; __THREW__ = 0;
     $101 = $100&1;
     if ($101) {
      label = 39;
      break;
     }
     $$pre$i$i$i = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $96 = $$pre$i$i$i;
    }
    $95 = ($96|0)==($87|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($95) {
     label = 40;
     break;
    } else {
     $92 = $96;
    }
   }
   if ((label|0) == 39) {
    $102 = ___cxa_find_matching_catch_3(0|0)|0;
    $103 = tempRet0;
    ___clang_call_terminate($102); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 40) {
    $$pre$i = HEAP32[$args>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $104 = $$pre$i;
    break;
   }
  }
 } while(0);
 __ZdlPv($104); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__16vectorIN4love9StrongRefINS1_7VariantEEENS_9allocatorIS4_EEE21__push_back_slow_pathIS4_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i8 = 0, $$0$i9 = 0, $$01$i$i = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre12 = 0, $$pre13 = 0, $$pre14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__v$sroa$15$0 = 0, $__v$sroa$6$0 = 0, $__v$sroa$6$0$lcssa = 0, $__v$sroa$6$1 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = $$in; //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = HEAP32[$8>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $10 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($11>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $17 = $10 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ($17>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $19 = $18 ? $5 : $17; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = (($20) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = $21 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = ($19|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $24 = $20; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($23) {
   $$0$i9 = 0;$101 = $24;$28 = 0;$29 = $22;
  } else {
   $$0$i8 = $19;$100 = $22;$99 = $24;
   label = 6;
  }
 } else {
  $13 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = $14 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = $13; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $$0$i8 = 1073741823;$100 = $15;$99 = $16;
  label = 6;
 }
 if ((label|0) == 6) {
  $25 = $$0$i8 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (__Znwj($25)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i9 = $$0$i8;$101 = $99;$28 = $26;$29 = $100;
 }
 $27 = (($28) + ($29<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = $27; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $31 = (($28) + ($$0$i9<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $32 = $31; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $33 = HEAP32[$__x>>2]|0; //@line 139 "./common/Object.h"
 HEAP32[$27>>2] = $33; //@line 99 "./common/Object.h"
 $34 = ($33|0)==(0|0); //@line 101 "./common/Object.h"
 do {
  if ($34) {
   $43 = $101;$44 = $7;
   label = 10;
  } else {
   $35 = HEAP32[$33>>2]|0; //@line 101 "./common/Object.h"
   $36 = ((($35)) + 8|0); //@line 101 "./common/Object.h"
   $37 = HEAP32[$36>>2]|0; //@line 101 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($37|0,($33|0)); //@line 101 "./common/Object.h"
   $38 = __THREW__; __THREW__ = 0;
   $39 = $38&1;
   if ($39) {
    $81 = ___cxa_find_matching_catch_2()|0;
    $82 = tempRet0;
    $__v$sroa$15$0 = $30;$__v$sroa$6$1 = $30;$lpad$phi$index = $81;$lpad$phi$index2 = $82;
    break;
   } else {
    $$pre12 = HEAP32[$this>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre13 = HEAP32[$0>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $43 = $$pre13;$44 = $$pre12;
    label = 10;
    break;
   }
  }
 } while(0);
 L15: do {
  if ((label|0) == 10) {
   $40 = ((($27)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $41 = $40; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $42 = ($43|0)==($44|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   do {
    if ($42) {
     $45 = $43; //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $60 = $30;$62 = $45;$64 = $43;
    } else {
     $$01$i$i = $43;$48 = $27;$__v$sroa$6$0 = $30;
     while(1) {
      $46 = ((($$01$i$i)) + -4|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $47 = ((($48)) + -4|0); //@line 99 "./common/Object.h"
      $49 = HEAP32[$46>>2]|0; //@line 139 "./common/Object.h"
      HEAP32[$47>>2] = $49; //@line 99 "./common/Object.h"
      $50 = ($49|0)==(0|0); //@line 101 "./common/Object.h"
      if (!($50)) {
       $51 = HEAP32[$49>>2]|0; //@line 101 "./common/Object.h"
       $52 = ((($51)) + 8|0); //@line 101 "./common/Object.h"
       $53 = HEAP32[$52>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($53|0,($49|0)); //@line 101 "./common/Object.h"
       $54 = __THREW__; __THREW__ = 0;
       $55 = $54&1;
       if ($55) {
        $__v$sroa$6$0$lcssa = $__v$sroa$6$0;
        label = 24;
        break;
       }
      }
      $56 = $__v$sroa$6$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $57 = ((($56)) + -4|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $58 = $57; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $59 = ($46|0)==($44|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      if ($59) {
       $$lcssa = $58;
       label = 15;
       break;
      } else {
       $$01$i$i = $46;$48 = $57;$__v$sroa$6$0 = $58;
      }
     }
     if ((label|0) == 15) {
      $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $$pre14 = HEAP32[$0>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $60 = $$lcssa;$62 = $$pre$i;$64 = $$pre14;
      break;
     }
     else if ((label|0) == 24) {
      $79 = ___cxa_find_matching_catch_2()|0;
      $80 = tempRet0;
      $__v$sroa$15$0 = $41;$__v$sroa$6$1 = $__v$sroa$6$0$lcssa;$lpad$phi$index = $79;$lpad$phi$index2 = $80;
      break L15;
     }
    }
   } while(0);
   HEAP32[$this>>2] = $60; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$0>>2] = $41; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$8>>2] = $32; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $61 = $62; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $63 = ($64|0)==($61|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   L28: do {
    if (!($63)) {
     $66 = $64;
     while(1) {
      $65 = ((($66)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $67 = HEAP32[$65>>2]|0; //@line 106 "./common/Object.h"
      $68 = ($67|0)==(0|0); //@line 106 "./common/Object.h"
      if (!($68)) {
       $70 = HEAP32[$67>>2]|0; //@line 106 "./common/Object.h"
       $71 = ((($70)) + 12|0); //@line 106 "./common/Object.h"
       $72 = HEAP32[$71>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($72|0,($67|0)); //@line 106 "./common/Object.h"
       $73 = __THREW__; __THREW__ = 0;
       $74 = $73&1;
       if ($74) {
        break;
       }
      }
      $69 = ($65|0)==($61|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if ($69) {
       break L28;
      } else {
       $66 = $65;
      }
     }
     $75 = ___cxa_find_matching_catch_3(0|0)|0;
     $76 = tempRet0;
     ___clang_call_terminate($75); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   } while(0);
   $77 = ($62|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($77) {
    return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $78 = $62; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __ZdlPv($78); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
 } while(0);
 $83 = $__v$sroa$6$1; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $84 = $__v$sroa$15$0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $85 = ($84|0)==($83|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 L41: do {
  if (!($85)) {
   $87 = $84;
   while(1) {
    $86 = ((($87)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $88 = HEAP32[$86>>2]|0; //@line 106 "./common/Object.h"
    $89 = ($88|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($89)) {
     $91 = HEAP32[$88>>2]|0; //@line 106 "./common/Object.h"
     $92 = ((($91)) + 12|0); //@line 106 "./common/Object.h"
     $93 = HEAP32[$92>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($93|0,($88|0)); //@line 106 "./common/Object.h"
     $94 = __THREW__; __THREW__ = 0;
     $95 = $94&1;
     if ($95) {
      break;
     }
    }
    $90 = ($86|0)==($83|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($90) {
     break L41;
    } else {
     $87 = $86;
    }
   }
   $96 = ___cxa_find_matching_catch_3(0|0)|0;
   $97 = tempRet0;
   ___clang_call_terminate($96); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 } while(0);
 $98 = ($28|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($98) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 __ZdlPv($28); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function _luaopen_love_event($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0;
 var $eh$lpad$body$i$indexZ2D = 0, $instance$sroa$0$1 = 0, $vararg_buffer = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $w = sp + 4|0;
 $0 = HEAP32[(625612)>>2]|0; //@line 98 "./common/Module.h"
 $1 = $0; //@line 129 "modules/event/wrap_Event.cpp"
 $2 = ($0|0)==(0|0); //@line 130 "modules/event/wrap_Event.cpp"
 L1: do {
  if ($2) {
   __THREW__ = 0;
   $3 = (invoke_ii(3348,36)|0); //@line 132 "modules/event/wrap_Event.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   do {
    if ($5) {
     $11 = ___cxa_find_matching_catch_3(8992|0)|0;
     $12 = tempRet0;
     $eh$lpad$body$i$index2Z2D = $12;$eh$lpad$body$i$indexZ2D = $11;
    } else {
     __THREW__ = 0;
     invoke_vi(3510,($3|0)); //@line 132 "modules/event/wrap_Event.cpp"
     $6 = __THREW__; __THREW__ = 0;
     $7 = $6&1;
     if ($7) {
      $8 = ___cxa_find_matching_catch_3(8992|0)|0;
      $9 = tempRet0;
      __ZdlPv($3); //@line 132 "modules/event/wrap_Event.cpp"
      $eh$lpad$body$i$index2Z2D = $9;$eh$lpad$body$i$indexZ2D = $8;
      break;
     } else {
      $10 = $3; //@line 132 "modules/event/wrap_Event.cpp"
      $instance$sroa$0$1 = $10;
      break L1;
     }
    }
   } while(0);
   $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $14 = ($eh$lpad$body$i$index2Z2D|0)==($13|0); //@line 535 "./common/runtime.h"
   if (!($14)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $15 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
   $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
   $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
   $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $instance$sroa$0$1 = $1;
    break;
   }
   $23 = ___cxa_find_matching_catch_2()|0;
   $24 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    $27 = ___cxa_find_matching_catch_3(0|0)|0;
    $28 = tempRet0;
    ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $24;$$02$i = $23;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $29 = HEAP32[$0>>2]|0; //@line 135 "modules/event/wrap_Event.cpp"
   $30 = ((($29)) + 8|0); //@line 135 "modules/event/wrap_Event.cpp"
   $31 = HEAP32[$30>>2]|0; //@line 135 "modules/event/wrap_Event.cpp"
   FUNCTION_TABLE_vi[$31 & 8191]($0); //@line 135 "modules/event/wrap_Event.cpp"
   $instance$sroa$0$1 = $1;
  }
 } while(0);
 HEAP32[$w>>2] = $instance$sroa$0$1; //@line 138 "modules/event/wrap_Event.cpp"
 $32 = ((($w)) + 4|0); //@line 139 "modules/event/wrap_Event.cpp"
 HEAP32[$32>>2] = 328262; //@line 139 "modules/event/wrap_Event.cpp"
 $33 = ((($w)) + 8|0); //@line 140 "modules/event/wrap_Event.cpp"
 HEAP32[$33>>2] = 3; //@line 140 "modules/event/wrap_Event.cpp"
 $34 = ((($w)) + 12|0); //@line 141 "modules/event/wrap_Event.cpp"
 HEAP32[$34>>2] = 12768; //@line 141 "modules/event/wrap_Event.cpp"
 $35 = ((($w)) + 16|0); //@line 142 "modules/event/wrap_Event.cpp"
 HEAP32[$35>>2] = 0; //@line 142 "modules/event/wrap_Event.cpp"
 $36 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 144 "modules/event/wrap_Event.cpp"
 STACKTOP = sp;return ($36|0); //@line 146 "modules/event/wrap_Event.cpp"
}
function __ZN4love10filesystem8FileDataD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (12832); //@line 51 "modules/filesystem/FileData.cpp"
 $0 = ((($this)) + 8|0); //@line 52 "modules/filesystem/FileData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 52 "modules/filesystem/FileData.cpp"
 $2 = ($1|0)==(0|0); //@line 52 "modules/filesystem/FileData.cpp"
 if (!($2)) {
  __ZdaPv($1); //@line 52 "modules/filesystem/FileData.cpp"
 }
 $3 = ((($this)) + 36|0); //@line 53 "modules/filesystem/FileData.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 53 "modules/filesystem/FileData.cpp"
 $4 = ((($this)) + 24|0); //@line 53 "modules/filesystem/FileData.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 53 "modules/filesystem/FileData.cpp"
 return; //@line 53 "modules/filesystem/FileData.cpp"
}
function __ZN4love10filesystem8FileDataD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (12832); //@line 51 "modules/filesystem/FileData.cpp"
 $0 = ((($this)) + 8|0); //@line 52 "modules/filesystem/FileData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 52 "modules/filesystem/FileData.cpp"
 $2 = ($1|0)==(0|0); //@line 52 "modules/filesystem/FileData.cpp"
 if (!($2)) {
  __ZdaPv($1); //@line 52 "modules/filesystem/FileData.cpp"
 }
 $3 = ((($this)) + 36|0); //@line 53 "modules/filesystem/FileData.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 53 "modules/filesystem/FileData.cpp"
 $4 = ((($this)) + 24|0); //@line 53 "modules/filesystem/FileData.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 53 "modules/filesystem/FileData.cpp"
 __ZdlPv($this); //@line 51 "modules/filesystem/FileData.cpp"
 return; //@line 53 "modules/filesystem/FileData.cpp"
}
function __ZNK4love10filesystem8FileData7getDataEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 57 "modules/filesystem/FileData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 57 "modules/filesystem/FileData.cpp"
 return ($1|0); //@line 57 "modules/filesystem/FileData.cpp"
}
function __ZNK4love10filesystem8FileData7getSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 63 "modules/filesystem/FileData.cpp"
 $1 = $0; //@line 63 "modules/filesystem/FileData.cpp"
 $2 = $1; //@line 63 "modules/filesystem/FileData.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 63 "modules/filesystem/FileData.cpp"
 $4 = (($1) + 4)|0; //@line 63 "modules/filesystem/FileData.cpp"
 $5 = $4; //@line 63 "modules/filesystem/FileData.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 63 "modules/filesystem/FileData.cpp"
 $7 = ($6>>>0)<(0); //@line 63 "modules/filesystem/FileData.cpp"
 $8 = ($3>>>0)<(4294967295); //@line 63 "modules/filesystem/FileData.cpp"
 $9 = ($6|0)==(0); //@line 63 "modules/filesystem/FileData.cpp"
 $10 = $9 & $8; //@line 63 "modules/filesystem/FileData.cpp"
 $11 = $7 | $10; //@line 63 "modules/filesystem/FileData.cpp"
 $12 = $11 ? $3 : -1; //@line 63 "modules/filesystem/FileData.cpp"
 $11 ? $6 : 0; //@line 63 "modules/filesystem/FileData.cpp"
 return ($12|0); //@line 63 "modules/filesystem/FileData.cpp"
}
function __GLOBAL__sub_I_FileData_cpp() {
 var $$02$i$i$1$i = 0, $$02$i$i$i = 0, $$lcssa$i$1$i = 0, $$lcssa$i$i = 0, $$lcssa10$i$1$i = 0, $$lcssa10$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i$i$1$i = 0, $hash$0$lcssa$i$i$i = 0;
 var $hash$01$i$i$1$i = 0, $hash$01$i$i$i = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 HEAP8[(627708)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627720)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627732)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627744)>>0] = 0; //@line 155 "./common/StringMap.h"
 $0 = (627748); //@line 44 "./common/StringMap.h"
 $1 = $0; //@line 44 "./common/StringMap.h"
 HEAP32[$1>>2] = 0; //@line 44 "./common/StringMap.h"
 $2 = (($0) + 4)|0; //@line 44 "./common/StringMap.h"
 $3 = $2; //@line 44 "./common/StringMap.h"
 HEAP32[$3>>2] = 0; //@line 44 "./common/StringMap.h"
 $4 = HEAP32[3214]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(12860)>>2]|0; //@line 49 "./common/StringMap.h"
 $6 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
 $7 = ($6<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($7) {
  $hash$0$lcssa$i$i$i = 5381;
 } else {
  $$02$i$i$i = $4;$9 = $6;$hash$01$i$i$i = 5381;
  while(1) {
   $8 = $9 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $10 = ((($$02$i$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $11 = ($hash$01$i$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $12 = (($11) + ($8))|0; //@line 143 "./common/StringMap.h"
   $13 = HEAP8[$10>>0]|0; //@line 142 "./common/StringMap.h"
   $14 = ($13<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($14) {
    $hash$0$lcssa$i$i$i = $12;
    break;
   } else {
    $$02$i$i$i = $10;$9 = $13;$hash$01$i$i$i = $12;
   }
  }
 }
 $15 = $hash$0$lcssa$i$i$i & 3; //@line 112 "./common/StringMap.h"
 $16 = (((627700 + (($15*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $17 = HEAP8[$16>>0]|0; //@line 114 "./common/StringMap.h"
 $18 = ($17<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($18) {
  $$lcssa$i$i = $15;$$lcssa10$i$i = $16;
  label = 5;
 } else {
  $19 = (($hash$0$lcssa$i$i$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $20 = $19 & 3; //@line 112 "./common/StringMap.h"
  $21 = (((627700 + (($20*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $22 = HEAP8[$21>>0]|0; //@line 114 "./common/StringMap.h"
  $23 = ($22<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($23) {
   $$lcssa$i$i = $20;$$lcssa10$i$i = $21;
   label = 5;
  } else {
   $28 = (($hash$0$lcssa$i$i$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $29 = $28 & 3; //@line 112 "./common/StringMap.h"
   $30 = (((627700 + (($29*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $31 = HEAP8[$30>>0]|0; //@line 114 "./common/StringMap.h"
   $32 = ($31<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($32) {
    $$lcssa$i$i = $29;$$lcssa10$i$i = $30;
    label = 5;
   } else {
    $33 = (($hash$0$lcssa$i$i$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $34 = $33 & 3; //@line 112 "./common/StringMap.h"
    $35 = (((627700 + (($34*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $36 = HEAP8[$35>>0]|0; //@line 114 "./common/StringMap.h"
    $37 = ($36<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($37) {
     $$lcssa$i$i = $34;$$lcssa10$i$i = $35;
     label = 5;
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10$i$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $24 = (627700 + (($$lcssa$i$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$24>>2] = $4; //@line 118 "./common/StringMap.h"
  $25 = (((627700 + (($$lcssa$i$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$25>>2] = $5; //@line 119 "./common/StringMap.h"
 }
 $26 = ($5>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($26) {
  HEAP32[$vararg_buffer>>2] = $4; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $5; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
 } else {
  $27 = (627748 + ($5<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$27>>2] = $4; //@line 132 "./common/StringMap.h"
 }
 $38 = HEAP32[(12864)>>2]|0; //@line 49 "./common/StringMap.h"
 $39 = HEAP32[(12868)>>2]|0; //@line 49 "./common/StringMap.h"
 $40 = HEAP8[$38>>0]|0; //@line 142 "./common/StringMap.h"
 $41 = ($40<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($41) {
  $hash$0$lcssa$i$i$1$i = 5381;
 } else {
  $$02$i$i$1$i = $38;$43 = $40;$hash$01$i$i$1$i = 5381;
  while(1) {
   $42 = $43 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $44 = ((($$02$i$i$1$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $45 = ($hash$01$i$i$1$i*33)|0; //@line 143 "./common/StringMap.h"
   $46 = (($45) + ($42))|0; //@line 143 "./common/StringMap.h"
   $47 = HEAP8[$44>>0]|0; //@line 142 "./common/StringMap.h"
   $48 = ($47<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($48) {
    $hash$0$lcssa$i$i$1$i = $46;
    break;
   } else {
    $$02$i$i$1$i = $44;$43 = $47;$hash$01$i$i$1$i = $46;
   }
  }
 }
 $49 = $hash$0$lcssa$i$i$1$i & 3; //@line 112 "./common/StringMap.h"
 $50 = (((627700 + (($49*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $51 = HEAP8[$50>>0]|0; //@line 114 "./common/StringMap.h"
 $52 = ($51<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($52) {
  $$lcssa$i$1$i = $49;$$lcssa10$i$1$i = $50;
  label = 17;
 } else {
  $53 = (($hash$0$lcssa$i$i$1$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $54 = $53 & 3; //@line 112 "./common/StringMap.h"
  $55 = (((627700 + (($54*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $56 = HEAP8[$55>>0]|0; //@line 114 "./common/StringMap.h"
  $57 = ($56<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($57) {
   $$lcssa$i$1$i = $54;$$lcssa10$i$1$i = $55;
   label = 17;
  } else {
   $58 = (($hash$0$lcssa$i$i$1$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $59 = $58 & 3; //@line 112 "./common/StringMap.h"
   $60 = (((627700 + (($59*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $61 = HEAP8[$60>>0]|0; //@line 114 "./common/StringMap.h"
   $62 = ($61<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($62) {
    $$lcssa$i$1$i = $59;$$lcssa10$i$1$i = $60;
    label = 17;
   } else {
    $63 = (($hash$0$lcssa$i$i$1$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $64 = $63 & 3; //@line 112 "./common/StringMap.h"
    $65 = (((627700 + (($64*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $66 = HEAP8[$65>>0]|0; //@line 114 "./common/StringMap.h"
    $67 = ($66<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($67) {
     $$lcssa$i$1$i = $64;$$lcssa10$i$1$i = $65;
     label = 17;
    }
   }
  }
 }
 if ((label|0) == 17) {
  HEAP8[$$lcssa10$i$1$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $68 = (627700 + (($$lcssa$i$1$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$68>>2] = $38; //@line 118 "./common/StringMap.h"
  $69 = (((627700 + (($$lcssa$i$1$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$69>>2] = $39; //@line 119 "./common/StringMap.h"
 }
 $70 = ($39>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($70) {
  HEAP32[$vararg_buffer2>>2] = $38; //@line 128 "./common/StringMap.h"
  $vararg_ptr5 = ((($vararg_buffer2)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr5>>2] = $39; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer2)|0); //@line 128 "./common/StringMap.h"
  STACKTOP = sp;return;
 } else {
  $71 = (627748 + ($39<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$71>>2] = $38; //@line 132 "./common/StringMap.h"
  STACKTOP = sp;return;
 }
}
function __ZN4love10filesystem8FileDataC2EyRKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE($this,$0,$1,$filename) {
 $this = $this|0;
 $0 = $0|0;
 $1 = $1|0;
 $filename = $filename|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $2 = sp + 4|0;
 __ZN4love6ObjectC2Ev($this); //@line 37 "./common/Data.h"
 HEAP32[$this>>2] = (12832); //@line 36 "modules/filesystem/FileData.cpp"
 $3 = ((($this)) + 8|0); //@line 33 "modules/filesystem/FileData.cpp"
 HEAP32[$3>>2] = 0; //@line 33 "modules/filesystem/FileData.cpp"
 $4 = ((($this)) + 16|0); //@line 34 "modules/filesystem/FileData.cpp"
 $5 = $4; //@line 34 "modules/filesystem/FileData.cpp"
 $6 = $5; //@line 34 "modules/filesystem/FileData.cpp"
 HEAP32[$6>>2] = $0; //@line 34 "modules/filesystem/FileData.cpp"
 $7 = (($5) + 4)|0; //@line 34 "modules/filesystem/FileData.cpp"
 $8 = $7; //@line 34 "modules/filesystem/FileData.cpp"
 HEAP32[$8>>2] = 0; //@line 34 "modules/filesystem/FileData.cpp"
 $9 = ((($this)) + 24|0); //@line 35 "modules/filesystem/FileData.cpp"
 __THREW__ = 0;
 invoke_vii(3351,($9|0),($filename|0)); //@line 35 "modules/filesystem/FileData.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $18 = ___cxa_find_matching_catch_2()|0;
  $19 = tempRet0;
  $$2 = $18;$$23 = $19;
  ___resumeException($$2|0);
  // unreachable;
 }
 $12 = ((($this)) + 36|0); //@line 32 "modules/filesystem/FileData.cpp"
 ;HEAP32[$12>>2]=0|0;HEAP32[$12+4>>2]=0|0;HEAP32[$12+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 $13 = (invoke_ii(3492,($0|0))|0); //@line 39 "modules/filesystem/FileData.cpp"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 do {
  if ($15) {
   $20 = ___cxa_find_matching_catch_3(9056|0)|0;
   $21 = tempRet0;
   $22 = (_llvm_eh_typeid_for((9056|0))|0); //@line 40 "modules/filesystem/FileData.cpp"
   $23 = ($21|0)==($22|0); //@line 40 "modules/filesystem/FileData.cpp"
   if ($23) {
    (___cxa_begin_catch(($20|0))|0); //@line 40 "modules/filesystem/FileData.cpp"
    $24 = (___cxa_allocate_exception(16)|0); //@line 43 "modules/filesystem/FileData.cpp"
    __THREW__ = 0;
    invoke_viii(3349,($24|0),(507610|0),($vararg_buffer|0)); //@line 43 "modules/filesystem/FileData.cpp"
    $25 = __THREW__; __THREW__ = 0;
    $26 = $25&1;
    if ($26) {
     $30 = ___cxa_find_matching_catch_2()|0;
     $31 = tempRet0;
     ___cxa_free_exception(($24|0)); //@line 43 "modules/filesystem/FileData.cpp"
     $$0 = $30;$$01 = $31;
    } else {
     __THREW__ = 0;
     invoke_viii(3350,($24|0),(96|0),(344|0)); //@line 43 "modules/filesystem/FileData.cpp"
     $27 = __THREW__; __THREW__ = 0;
     $28 = ___cxa_find_matching_catch_2()|0;
     $29 = tempRet0;
     $$0 = $28;$$01 = $29;
    }
    __THREW__ = 0;
    invoke_v(3378); //@line 44 "modules/filesystem/FileData.cpp"
    $32 = __THREW__; __THREW__ = 0;
    $33 = $32&1;
    if ($33) {
     $51 = ___cxa_find_matching_catch_3(0|0)|0;
     $52 = tempRet0;
     ___clang_call_terminate($51); //@line 44 "modules/filesystem/FileData.cpp"
     // unreachable; //@line 44 "modules/filesystem/FileData.cpp"
    } else {
     $$1 = $$0;$$12 = $$01;
    }
   } else {
    $$1 = $20;$$12 = $21;
   }
  } else {
   HEAP32[$3>>2] = $13; //@line 39 "modules/filesystem/FileData.cpp"
   $16 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE5rfindEcj($filename,46,-1)|0); //@line 46 "modules/filesystem/FileData.cpp"
   $17 = ($16|0)==(-1); //@line 46 "modules/filesystem/FileData.cpp"
   if ($17) {
    STACKTOP = sp;return; //@line 48 "modules/filesystem/FileData.cpp"
   }
   $34 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE5rfindEcj($filename,46,-1)|0); //@line 47 "modules/filesystem/FileData.cpp"
   $35 = (($34) + 1)|0; //@line 47 "modules/filesystem/FileData.cpp"
   __THREW__ = 0;
   invoke_viiiii(3511,($2|0),($filename|0),($35|0),-1,($filename|0)); //@line 3345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $36 = __THREW__; __THREW__ = 0;
   $37 = $36&1;
   if ($37) {
    $49 = ___cxa_find_matching_catch_2()|0;
    $50 = tempRet0;
    $$1 = $49;$$12 = $50;
    break;
   }
   $38 = HEAP8[$12>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $39 = $38 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $40 = ($39<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if ($40) {
    $44 = ((($12)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$44>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$12>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   } else {
    $41 = ((($this)) + 44|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $42 = HEAP32[$41>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$42>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $43 = ((($this)) + 40|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP32[$43>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   }
   __THREW__ = 0;
   invoke_vii(3368,($12|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $45 = __THREW__; __THREW__ = 0;
   $46 = $45&1;
   if ($46) {
    $47 = ___cxa_find_matching_catch_3(0|0)|0;
    $48 = tempRet0;
    ___clang_call_terminate($47); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   }
   ;HEAP32[$12>>2]=HEAP32[$2>>2]|0;HEAP32[$12+4>>2]=HEAP32[$2+4>>2]|0;HEAP32[$12+8>>2]=HEAP32[$2+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   ;HEAP32[$2>>2]=0|0;HEAP32[$2+4>>2]=0|0;HEAP32[$2+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 47 "modules/filesystem/FileData.cpp"
   STACKTOP = sp;return; //@line 48 "modules/filesystem/FileData.cpp"
  }
 } while(0);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12); //@line 48 "modules/filesystem/FileData.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9); //@line 48 "modules/filesystem/FileData.cpp"
 $$2 = $$1;$$23 = $$12;
 ___resumeException($$2|0);
 // unreachable;
}
function __ZNK4love10filesystem8FileData11getFilenameEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 68 "modules/filesystem/FileData.cpp"
 return ($0|0); //@line 68 "modules/filesystem/FileData.cpp"
}
function __ZNK4love10filesystem8FileData12getExtensionEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 36|0); //@line 73 "modules/filesystem/FileData.cpp"
 return ($0|0); //@line 73 "modules/filesystem/FileData.cpp"
}
function __ZN4love10filesystem8FileData11getConstantEPKcRNS1_7DecoderE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love9StringMapINS_10filesystem8FileData7DecoderELj2EE4findEPKcRS3_(627700,$in,$out)|0); //@line 78 "modules/filesystem/FileData.cpp"
 return ($0|0); //@line 78 "modules/filesystem/FileData.cpp"
}
function __ZN4love9StringMapINS_10filesystem8FileData7DecoderELj2EE4findEPKcRS3_($this,$key,$t) {
 $this = $this|0;
 $key = $key|0;
 $t = $t|0;
 var $$$2 = 0, $$011$i$137 = 0, $$011$i$233 = 0, $$011$i$330 = 0, $$011$i41 = 0, $$0110$i$138 = 0, $$0110$i$234 = 0, $$0110$i$331 = 0, $$0110$i42 = 0, $$02$i = 0, $$13 = 0, $$2 = 0, $$lcssa12 = 0, $$pre14 = 0, $$pre16 = 0, $$pre18 = 0, $$pre20 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $or$cond$i = 0, $or$cond$i$1 = 0, $or$cond$i$136 = 0, $or$cond$i$2 = 0, $or$cond$i$232 = 0, $or$cond$i$3 = 0, $or$cond$i$329 = 0, $or$cond$i40 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 3; //@line 72 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 74 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 74 "./common/StringMap.h"
 if ($12) {
  $$13 = 1;$$2 = 0;
  $$$2 = $$13 & $$2;
  return ($$$2|0); //@line 85 "./common/StringMap.h"
 }
 $13 = (($this) + (($9*12)|0)|0); //@line 77 "./common/StringMap.h"
 $14 = HEAP32[$13>>2]|0; //@line 77 "./common/StringMap.h"
 $15 = HEAP8[$14>>0]|0; //@line 54 "./common/StringMap.h"
 $16 = ($15<<24>>24)==(0); //@line 54 "./common/StringMap.h"
 L8: do {
  if ($16) {
   $26 = $0;
   label = 8;
  } else {
   $17 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
   $18 = ($15<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
   $or$cond$i40 = $17 & $18; //@line 54 "./common/StringMap.h"
   if ($or$cond$i40) {
    $$011$i41 = $key;$$0110$i42 = $14;
    while(1) {
     $19 = ((($$0110$i42)) + 1|0); //@line 59 "./common/StringMap.h"
     $20 = ((($$011$i41)) + 1|0); //@line 60 "./common/StringMap.h"
     $21 = HEAP8[$19>>0]|0; //@line 54 "./common/StringMap.h"
     $22 = ($21<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     $$pre14 = HEAP8[$20>>0]|0; //@line 63 "./common/StringMap.h"
     if ($22) {
      $26 = $$pre14;
      label = 8;
      break L8;
     }
     $23 = ($$pre14<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $24 = ($21<<24>>24)==($$pre14<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i = $23 & $24; //@line 54 "./common/StringMap.h"
     if ($or$cond$i) {
      $$011$i41 = $20;$$0110$i42 = $19;
     } else {
      label = 10;
      break;
     }
    }
   } else {
    label = 10;
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $25 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
  if ($25) {
   $$lcssa12 = $9;
  } else {
   label = 10;
  }
 }
 do {
  if ((label|0) == 10) {
   $29 = (($hash$0$lcssa$i) + 1)|0; //@line 72 "./common/StringMap.h"
   $30 = $29 & 3; //@line 72 "./common/StringMap.h"
   $31 = (((($this) + (($30*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $32 = HEAP8[$31>>0]|0; //@line 74 "./common/StringMap.h"
   $33 = ($32<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($33) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $34 = (($this) + (($30*12)|0)|0); //@line 77 "./common/StringMap.h"
   $35 = HEAP32[$34>>2]|0; //@line 77 "./common/StringMap.h"
   $36 = HEAP8[$35>>0]|0; //@line 54 "./common/StringMap.h"
   $37 = ($36<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L20: do {
    if ($37) {
     $47 = $0;
     label = 16;
    } else {
     $38 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $39 = ($36<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$136 = $38 & $39; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$136) {
      $$011$i$137 = $key;$$0110$i$138 = $35;
      while(1) {
       $40 = ((($$0110$i$138)) + 1|0); //@line 59 "./common/StringMap.h"
       $41 = ((($$011$i$137)) + 1|0); //@line 60 "./common/StringMap.h"
       $42 = HEAP8[$40>>0]|0; //@line 54 "./common/StringMap.h"
       $43 = ($42<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre16 = HEAP8[$41>>0]|0; //@line 63 "./common/StringMap.h"
       if ($43) {
        $47 = $$pre16;
        label = 16;
        break L20;
       }
       $44 = ($$pre16<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $45 = ($42<<24>>24)==($$pre16<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$1 = $44 & $45; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$1) {
        $$011$i$137 = $41;$$0110$i$138 = $40;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 16) {
    $46 = ($47<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($46) {
     $$lcssa12 = $30;
     break;
    }
   }
   $48 = (($hash$0$lcssa$i) + 2)|0; //@line 72 "./common/StringMap.h"
   $49 = $48 & 3; //@line 72 "./common/StringMap.h"
   $50 = (((($this) + (($49*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $51 = HEAP8[$50>>0]|0; //@line 74 "./common/StringMap.h"
   $52 = ($51<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($52) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $53 = (($this) + (($49*12)|0)|0); //@line 77 "./common/StringMap.h"
   $54 = HEAP32[$53>>2]|0; //@line 77 "./common/StringMap.h"
   $55 = HEAP8[$54>>0]|0; //@line 54 "./common/StringMap.h"
   $56 = ($55<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L31: do {
    if ($56) {
     $66 = $0;
     label = 22;
    } else {
     $57 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $58 = ($55<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$232 = $57 & $58; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$232) {
      $$011$i$233 = $key;$$0110$i$234 = $54;
      while(1) {
       $59 = ((($$0110$i$234)) + 1|0); //@line 59 "./common/StringMap.h"
       $60 = ((($$011$i$233)) + 1|0); //@line 60 "./common/StringMap.h"
       $61 = HEAP8[$59>>0]|0; //@line 54 "./common/StringMap.h"
       $62 = ($61<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre18 = HEAP8[$60>>0]|0; //@line 63 "./common/StringMap.h"
       if ($62) {
        $66 = $$pre18;
        label = 22;
        break L31;
       }
       $63 = ($$pre18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $64 = ($61<<24>>24)==($$pre18<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$2 = $63 & $64; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$2) {
        $$011$i$233 = $60;$$0110$i$234 = $59;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 22) {
    $65 = ($66<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($65) {
     $$lcssa12 = $49;
     break;
    }
   }
   $67 = (($hash$0$lcssa$i) + 3)|0; //@line 72 "./common/StringMap.h"
   $68 = $67 & 3; //@line 72 "./common/StringMap.h"
   $69 = (((($this) + (($68*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $70 = HEAP8[$69>>0]|0; //@line 74 "./common/StringMap.h"
   $71 = ($70<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($71) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $72 = (($this) + (($68*12)|0)|0); //@line 77 "./common/StringMap.h"
   $73 = HEAP32[$72>>2]|0; //@line 77 "./common/StringMap.h"
   $74 = HEAP8[$73>>0]|0; //@line 54 "./common/StringMap.h"
   $75 = ($74<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L42: do {
    if ($75) {
     $85 = $0;
    } else {
     $76 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $77 = ($74<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$329 = $76 & $77; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$329) {
      $$011$i$330 = $key;$$0110$i$331 = $73;
     } else {
      $$13 = 0;$$2 = 0;
      $$$2 = $$13 & $$2;
      return ($$$2|0); //@line 85 "./common/StringMap.h"
     }
     while(1) {
      $78 = ((($$0110$i$331)) + 1|0); //@line 59 "./common/StringMap.h"
      $79 = ((($$011$i$330)) + 1|0); //@line 60 "./common/StringMap.h"
      $80 = HEAP8[$78>>0]|0; //@line 54 "./common/StringMap.h"
      $81 = ($80<<24>>24)==(0); //@line 54 "./common/StringMap.h"
      $$pre20 = HEAP8[$79>>0]|0; //@line 63 "./common/StringMap.h"
      if ($81) {
       $85 = $$pre20;
       break L42;
      }
      $82 = ($$pre20<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
      $83 = ($80<<24>>24)==($$pre20<<24>>24); //@line 56 "./common/StringMap.h"
      $or$cond$i$3 = $82 & $83; //@line 54 "./common/StringMap.h"
      if ($or$cond$i$3) {
       $$011$i$330 = $79;$$0110$i$331 = $78;
      } else {
       $$13 = 0;$$2 = 0;
       break;
      }
     }
     $$$2 = $$13 & $$2;
     return ($$$2|0); //@line 85 "./common/StringMap.h"
    }
   } while(0);
   $84 = ($85<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($84) {
    $$lcssa12 = $68;
   } else {
    $$13 = 0;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
  }
 } while(0);
 $27 = (((($this) + (($$lcssa12*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
 $28 = HEAP32[$27>>2]|0; //@line 79 "./common/StringMap.h"
 HEAP32[$t>>2] = $28; //@line 79 "./common/StringMap.h"
 $$13 = 1;$$2 = 1;
 $$$2 = $$13 & $$2;
 return ($$$2|0); //@line 85 "./common/StringMap.h"
}
function __ZN4love10filesystem24hack_setupWriteDirectoryEv() {
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $1 = ($0|0)==(0|0); //@line 47 "modules/filesystem/wrap_Filesystem.cpp"
 if ($1) {
  $$0 = 0;
  return ($$0|0); //@line 50 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $2 = HEAP32[$0>>2]|0; //@line 48 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 36|0); //@line 48 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 48 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($0)|0); //@line 48 "modules/filesystem/wrap_Filesystem.cpp"
 $$0 = $5;
 return ($$0|0); //@line 50 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem6w_initEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 52 "modules/filesystem/wrap_Filesystem.cpp"
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 54 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 55 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 24|0); //@line 55 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 55 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 invoke_vii($4|0,($1|0),($0|0)); //@line 55 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  STACKTOP = sp;return 0; //@line 56 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $7 = ___cxa_find_matching_catch_3(8992|0)|0;
 $8 = tempRet0;
 $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
 if (!($10)) {
  $$01$i = $8;$$02$i = $7;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
 $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
 $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if (!($17)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $18 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $18; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 56 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $19 = ___cxa_find_matching_catch_2()|0;
 $20 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $23 = ___cxa_find_matching_catch_3(0|0)|0;
  $24 = tempRet0;
  ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $20;$$02$i = $19;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love10filesystem10w_setFusedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 63 "modules/filesystem/wrap_Filesystem.cpp"
 $2 = ((($1)) + 28|0); //@line 63 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 63 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,1)|0); //@line 63 "modules/filesystem/wrap_Filesystem.cpp"
 FUNCTION_TABLE_vii[$3 & 8191]($0,$4); //@line 63 "modules/filesystem/wrap_Filesystem.cpp"
 return 0; //@line 64 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem9w_isFusedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 69 "modules/filesystem/wrap_Filesystem.cpp"
 $2 = ((($1)) + 32|0); //@line 69 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 69 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 69 "modules/filesystem/wrap_Filesystem.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 69 "modules/filesystem/wrap_Filesystem.cpp"
 return 1; //@line 70 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem24w_setAndroidSaveExternalEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,1,0)|0); //@line 75 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 76 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 40|0); //@line 76 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 76 "modules/filesystem/wrap_Filesystem.cpp"
 FUNCTION_TABLE_vii[$4 & 8191]($1,$0); //@line 76 "modules/filesystem/wrap_Filesystem.cpp"
 return 0; //@line 77 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem13w_setIdentityEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 80 "modules/filesystem/wrap_Filesystem.cpp"
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 82 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,2,0)|0); //@line 83 "modules/filesystem/wrap_Filesystem.cpp"
 $2 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $3 = HEAP32[$2>>2]|0; //@line 85 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = ((($3)) + 48|0); //@line 85 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 85 "modules/filesystem/wrap_Filesystem.cpp"
 $6 = (FUNCTION_TABLE_iiii[$5 & 8191]($2,$0,$1)|0); //@line 85 "modules/filesystem/wrap_Filesystem.cpp"
 if ($6) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 89 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $7 = (_luaL_error($L,507182,$vararg_buffer)|0); //@line 86 "modules/filesystem/wrap_Filesystem.cpp"
 $$0 = $7;
 STACKTOP = sp;return ($$0|0); //@line 89 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem13w_getIdentityEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 93 "modules/filesystem/wrap_Filesystem.cpp"
 $2 = ((($1)) + 52|0); //@line 93 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 93 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 93 "modules/filesystem/wrap_Filesystem.cpp"
 _lua_pushstring($L,$4); //@line 93 "modules/filesystem/wrap_Filesystem.cpp"
 return 1; //@line 94 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem11w_setSourceEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 97 "modules/filesystem/wrap_Filesystem.cpp"
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 99 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 101 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 56|0); //@line 101 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 101 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = (FUNCTION_TABLE_iii[$4 & 8191]($1,$0)|0); //@line 101 "modules/filesystem/wrap_Filesystem.cpp"
 if ($5) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 105 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $6 = (_luaL_error($L,328327,$vararg_buffer)|0); //@line 102 "modules/filesystem/wrap_Filesystem.cpp"
 $$0 = $6;
 STACKTOP = sp;return ($$0|0); //@line 105 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem11w_getSourceEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 109 "modules/filesystem/wrap_Filesystem.cpp"
 $2 = ((($1)) + 60|0); //@line 109 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 109 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 109 "modules/filesystem/wrap_Filesystem.cpp"
 _lua_pushstring($L,$4); //@line 109 "modules/filesystem/wrap_Filesystem.cpp"
 return 1; //@line 110 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem7w_mountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 115 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = (_luaL_checklstring($L,2,0)|0); //@line 116 "modules/filesystem/wrap_Filesystem.cpp"
 $2 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,3,0)|0); //@line 117 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $4 = HEAP32[$3>>2]|0; //@line 119 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = ((($4)) + 64|0); //@line 119 "modules/filesystem/wrap_Filesystem.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 119 "modules/filesystem/wrap_Filesystem.cpp"
 $7 = (FUNCTION_TABLE_iiiii[$6 & 8191]($3,$0,$1,$2)|0); //@line 119 "modules/filesystem/wrap_Filesystem.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$7); //@line 119 "modules/filesystem/wrap_Filesystem.cpp"
 return 1; //@line 120 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem9w_unmountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 125 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 127 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 68|0); //@line 127 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 127 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = (FUNCTION_TABLE_iii[$4 & 8191]($1,$0)|0); //@line 127 "modules/filesystem/wrap_Filesystem.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$5); //@line 127 "modules/filesystem/wrap_Filesystem.cpp"
 return 1; //@line 128 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem9w_newFileEP9lua_State($L) {
 $L = $L|0;
 var $$01 = 0, $$03 = 0, $$1 = 0, $$12 = 0, $$14 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $mode = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $mode = sp + 20|0;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 133 "modules/filesystem/wrap_Filesystem.cpp"
 HEAP32[$mode>>2] = 0; //@line 136 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = (_lua_isstring($L,2)|0); //@line 138 "modules/filesystem/wrap_Filesystem.cpp"
 $2 = ($1|0)==(0); //@line 138 "modules/filesystem/wrap_Filesystem.cpp"
 if (!($2)) {
  $3 = (_luaL_checklstring($L,2,0)|0); //@line 140 "modules/filesystem/wrap_Filesystem.cpp"
  $4 = (__ZN4love10filesystem4File11getConstantEPKcRNS1_4ModeE($3,$mode)|0); //@line 141 "modules/filesystem/wrap_Filesystem.cpp"
  if (!($4)) {
   HEAP32[$vararg_buffer>>2] = $3; //@line 142 "modules/filesystem/wrap_Filesystem.cpp"
   $5 = (_luaL_error($L,366404,$vararg_buffer)|0); //@line 142 "modules/filesystem/wrap_Filesystem.cpp"
   $$1 = $5;
   STACKTOP = sp;return ($$1|0); //@line 164 "modules/filesystem/wrap_Filesystem.cpp"
  }
 }
 $6 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $7 = HEAP32[$6>>2]|0; //@line 145 "modules/filesystem/wrap_Filesystem.cpp"
 $8 = ((($7)) + 72|0); //@line 145 "modules/filesystem/wrap_Filesystem.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 145 "modules/filesystem/wrap_Filesystem.cpp"
 $10 = (FUNCTION_TABLE_iii[$9 & 8191]($6,$0)|0); //@line 145 "modules/filesystem/wrap_Filesystem.cpp"
 $11 = HEAP32[$mode>>2]|0; //@line 147 "modules/filesystem/wrap_Filesystem.cpp"
 $12 = ($11|0)==(0); //@line 147 "modules/filesystem/wrap_Filesystem.cpp"
 L6: do {
  if (!($12)) {
   $13 = HEAP32[$10>>2]|0; //@line 151 "modules/filesystem/wrap_Filesystem.cpp"
   $14 = ((($13)) + 16|0); //@line 151 "modules/filesystem/wrap_Filesystem.cpp"
   $15 = HEAP32[$14>>2]|0; //@line 151 "modules/filesystem/wrap_Filesystem.cpp"
   __THREW__ = 0;
   $16 = (invoke_iii($15|0,($10|0),($11|0))|0); //@line 151 "modules/filesystem/wrap_Filesystem.cpp"
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   do {
    if ($18) {
     label = 9;
    } else {
     if ($16) {
      break L6;
     }
     $19 = (___cxa_allocate_exception(16)|0); //@line 152 "modules/filesystem/wrap_Filesystem.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($19|0),(366514|0),($vararg_buffer1|0)); //@line 152 "modules/filesystem/wrap_Filesystem.cpp"
     $20 = __THREW__; __THREW__ = 0;
     $21 = $20&1;
     if ($21) {
      $25 = ___cxa_find_matching_catch_3(96|0)|0;
      $26 = tempRet0;
      ___cxa_free_exception(($19|0)); //@line 152 "modules/filesystem/wrap_Filesystem.cpp"
      $$01 = $26;$$03 = $25;
      break;
     } else {
      __THREW__ = 0;
      invoke_viii(3350,($19|0),(96|0),(344|0)); //@line 152 "modules/filesystem/wrap_Filesystem.cpp"
      $22 = __THREW__; __THREW__ = 0;
      label = 9;
      break;
     }
    }
   } while(0);
   if ((label|0) == 9) {
    $23 = ___cxa_find_matching_catch_3(96|0)|0;
    $24 = tempRet0;
    $$01 = $24;$$03 = $23;
   }
   $27 = (_llvm_eh_typeid_for((96|0))|0); //@line 153 "modules/filesystem/wrap_Filesystem.cpp"
   $28 = ($$01|0)==($27|0); //@line 153 "modules/filesystem/wrap_Filesystem.cpp"
   if (!($28)) {
    $$12 = $$01;$$14 = $$03;
    ___resumeException($$14|0);
    // unreachable;
   }
   $29 = (___cxa_begin_catch(($$03|0))|0); //@line 153 "modules/filesystem/wrap_Filesystem.cpp"
   $30 = HEAP32[$10>>2]|0; //@line 156 "modules/filesystem/wrap_Filesystem.cpp"
   $31 = ((($30)) + 12|0); //@line 156 "modules/filesystem/wrap_Filesystem.cpp"
   $32 = HEAP32[$31>>2]|0; //@line 156 "modules/filesystem/wrap_Filesystem.cpp"
   __THREW__ = 0;
   invoke_vi($32|0,($10|0)); //@line 156 "modules/filesystem/wrap_Filesystem.cpp"
   $33 = __THREW__; __THREW__ = 0;
   $34 = $33&1;
   if (!($34)) {
    $35 = HEAP32[$29>>2]|0; //@line 157 "modules/filesystem/wrap_Filesystem.cpp"
    $36 = ((($35)) + 8|0); //@line 157 "modules/filesystem/wrap_Filesystem.cpp"
    $37 = HEAP32[$36>>2]|0; //@line 157 "modules/filesystem/wrap_Filesystem.cpp"
    $38 = (FUNCTION_TABLE_ii[$37 & 8191]($29)|0); //@line 157 "modules/filesystem/wrap_Filesystem.cpp"
    __THREW__ = 0;
    HEAP32[$vararg_buffer3>>2] = $38; //@line 157 "modules/filesystem/wrap_Filesystem.cpp"
    $39 = (invoke_iiii(3512,($L|0),(595867|0),($vararg_buffer3|0))|0); //@line 157 "modules/filesystem/wrap_Filesystem.cpp"
    $40 = __THREW__; __THREW__ = 0;
    $41 = $40&1;
    if (!($41)) {
     ___cxa_end_catch(); //@line 158 "modules/filesystem/wrap_Filesystem.cpp"
     $$1 = $39;
     STACKTOP = sp;return ($$1|0); //@line 164 "modules/filesystem/wrap_Filesystem.cpp"
    }
   }
   $42 = ___cxa_find_matching_catch_2()|0;
   $43 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 158 "modules/filesystem/wrap_Filesystem.cpp"
   $44 = __THREW__; __THREW__ = 0;
   $45 = $44&1;
   if ($45) {
    $49 = ___cxa_find_matching_catch_3(0|0)|0;
    $50 = tempRet0;
    ___clang_call_terminate($49); //@line 158 "modules/filesystem/wrap_Filesystem.cpp"
    // unreachable; //@line 158 "modules/filesystem/wrap_Filesystem.cpp"
   } else {
    $$12 = $43;$$14 = $42;
    ___resumeException($$14|0);
    // unreachable;
   }
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,5,$10); //@line 161 "modules/filesystem/wrap_Filesystem.cpp"
 $46 = HEAP32[$10>>2]|0; //@line 162 "modules/filesystem/wrap_Filesystem.cpp"
 $47 = ((($46)) + 12|0); //@line 162 "modules/filesystem/wrap_Filesystem.cpp"
 $48 = HEAP32[$47>>2]|0; //@line 162 "modules/filesystem/wrap_Filesystem.cpp"
 FUNCTION_TABLE_vi[$48 & 8191]($10); //@line 162 "modules/filesystem/wrap_Filesystem.cpp"
 $$1 = 1;
 STACKTOP = sp;return ($$1|0); //@line 164 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem12luax_getfileEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $file$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_isstring($L,$idx)|0); //@line 169 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = ($0|0)==(0); //@line 169 "modules/filesystem/wrap_Filesystem.cpp"
 if ($1) {
  $8 = (__ZN4love10filesystem14luax_checkfileEP9lua_Statei($L,$idx)|0); //@line 175 "modules/filesystem/wrap_Filesystem.cpp"
  $file$0 = $8;
  return ($file$0|0); //@line 177 "modules/filesystem/wrap_Filesystem.cpp"
 } else {
  $2 = (_luaL_checklstring($L,$idx,0)|0); //@line 171 "modules/filesystem/wrap_Filesystem.cpp"
  $3 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
  $4 = HEAP32[$3>>2]|0; //@line 172 "modules/filesystem/wrap_Filesystem.cpp"
  $5 = ((($4)) + 72|0); //@line 172 "modules/filesystem/wrap_Filesystem.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 172 "modules/filesystem/wrap_Filesystem.cpp"
  $7 = (FUNCTION_TABLE_iii[$6 & 8191]($3,$2)|0); //@line 172 "modules/filesystem/wrap_Filesystem.cpp"
  $file$0 = $7;
  return ($file$0|0); //@line 177 "modules/filesystem/wrap_Filesystem.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem16luax_getfiledataEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$0 = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $data$0 = 0, $file$0 = 0, $file$0$i = 0, $or$cond = 0, $vararg_buffer = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 180 "modules/filesystem/wrap_Filesystem.cpp"
 $0 = (_lua_isstring($L,$idx)|0); //@line 185 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = ($0|0)==(0); //@line 185 "modules/filesystem/wrap_Filesystem.cpp"
 if ($1) {
  $2 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,$idx,5)|0); //@line 185 "modules/filesystem/wrap_Filesystem.cpp"
  if ($2) {
   label = 3;
  } else {
   $15 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,$idx,7)|0); //@line 190 "modules/filesystem/wrap_Filesystem.cpp"
   if ($15) {
    $16 = (__ZN4love10filesystem18luax_checkfiledataEP9lua_Statei($L,$idx)|0); //@line 192 "modules/filesystem/wrap_Filesystem.cpp"
    $17 = HEAP32[$16>>2]|0; //@line 193 "modules/filesystem/wrap_Filesystem.cpp"
    $18 = ((($17)) + 8|0); //@line 193 "modules/filesystem/wrap_Filesystem.cpp"
    $19 = HEAP32[$18>>2]|0; //@line 193 "modules/filesystem/wrap_Filesystem.cpp"
    FUNCTION_TABLE_vi[$19 & 8191]($16); //@line 193 "modules/filesystem/wrap_Filesystem.cpp"
    $data$0 = $16;$file$0 = 0;
   } else {
    $data$0 = 0;$file$0 = 0;
   }
  }
 } else {
  label = 3;
 }
 if ((label|0) == 3) {
  $3 = (_lua_isstring($L,$idx)|0); //@line 169 "modules/filesystem/wrap_Filesystem.cpp"
  $4 = ($3|0)==(0); //@line 169 "modules/filesystem/wrap_Filesystem.cpp"
  if ($4) {
   $11 = (__ZN4love10filesystem14luax_checkfileEP9lua_Statei($L,$idx)|0); //@line 175 "modules/filesystem/wrap_Filesystem.cpp"
   $file$0$i = $11;
  } else {
   $5 = (_luaL_checklstring($L,$idx,0)|0); //@line 171 "modules/filesystem/wrap_Filesystem.cpp"
   $6 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
   $7 = HEAP32[$6>>2]|0; //@line 172 "modules/filesystem/wrap_Filesystem.cpp"
   $8 = ((($7)) + 72|0); //@line 172 "modules/filesystem/wrap_Filesystem.cpp"
   $9 = HEAP32[$8>>2]|0; //@line 172 "modules/filesystem/wrap_Filesystem.cpp"
   $10 = (FUNCTION_TABLE_iii[$9 & 8191]($6,$5)|0); //@line 172 "modules/filesystem/wrap_Filesystem.cpp"
   $file$0$i = $10;
  }
  $12 = HEAP32[$file$0$i>>2]|0; //@line 188 "modules/filesystem/wrap_Filesystem.cpp"
  $13 = ((($12)) + 8|0); //@line 188 "modules/filesystem/wrap_Filesystem.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 188 "modules/filesystem/wrap_Filesystem.cpp"
  FUNCTION_TABLE_vi[$14 & 8191]($file$0$i); //@line 188 "modules/filesystem/wrap_Filesystem.cpp"
  $data$0 = 0;$file$0 = $file$0$i;
 }
 $20 = ($data$0|0)!=(0|0); //@line 196 "modules/filesystem/wrap_Filesystem.cpp"
 $21 = ($file$0|0)!=(0|0); //@line 196 "modules/filesystem/wrap_Filesystem.cpp"
 $or$cond = $21 | $20; //@line 196 "modules/filesystem/wrap_Filesystem.cpp"
 if (!($or$cond)) {
  (_luaL_argerror($L,$idx,328349)|0); //@line 198 "modules/filesystem/wrap_Filesystem.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 211 "modules/filesystem/wrap_Filesystem.cpp"
 }
 if (!($21)) {
  $$0 = $data$0;
  STACKTOP = sp;return ($$0|0); //@line 211 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $22 = HEAP32[$file$0>>2]|0; //@line 205 "modules/filesystem/wrap_Filesystem.cpp"
 $23 = ((($22)) + 32|0); //@line 205 "modules/filesystem/wrap_Filesystem.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 205 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 $25 = (invoke_iiii($24|0,($file$0|0),-1,-1)|0); //@line 205 "modules/filesystem/wrap_Filesystem.cpp"
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if (!($27)) {
  $47 = HEAP32[$file$0>>2]|0; //@line 206 "modules/filesystem/wrap_Filesystem.cpp"
  $48 = ((($47)) + 12|0); //@line 206 "modules/filesystem/wrap_Filesystem.cpp"
  $49 = HEAP32[$48>>2]|0; //@line 206 "modules/filesystem/wrap_Filesystem.cpp"
  FUNCTION_TABLE_vi[$49 & 8191]($file$0); //@line 206 "modules/filesystem/wrap_Filesystem.cpp"
  $$0 = $25;
  STACKTOP = sp;return ($$0|0); //@line 211 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $28 = ___cxa_find_matching_catch_3(8992|0)|0;
 $29 = tempRet0;
 $30 = (_llvm_eh_typeid_for((8992|0))|0); //@line 556 "./common/runtime.h"
 $31 = ($29|0)==($30|0); //@line 556 "./common/runtime.h"
 if (!($31)) {
  $$01$i = $29;$$02$i = $28;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $32 = (___cxa_begin_catch(($28|0))|0); //@line 556 "./common/runtime.h"
 $33 = HEAP32[$32>>2]|0; //@line 560 "./common/runtime.h"
 $34 = ((($33)) + 8|0); //@line 560 "./common/runtime.h"
 $35 = HEAP32[$34>>2]|0; //@line 560 "./common/runtime.h"
 $36 = (FUNCTION_TABLE_ii[$35 & 8191]($32)|0); //@line 560 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($36|0)); //@line 560 "./common/runtime.h"
 $37 = __THREW__; __THREW__ = 0;
 $38 = $37&1;
 if (!($38)) {
  ___cxa_end_catch(); //@line 561 "./common/runtime.h"
  $39 = HEAP32[$file$0>>2]|0; //@line 206 "modules/filesystem/wrap_Filesystem.cpp"
  $40 = ((($39)) + 12|0); //@line 206 "modules/filesystem/wrap_Filesystem.cpp"
  $41 = HEAP32[$40>>2]|0; //@line 206 "modules/filesystem/wrap_Filesystem.cpp"
  FUNCTION_TABLE_vi[$41 & 8191]($file$0); //@line 206 "modules/filesystem/wrap_Filesystem.cpp"
  $42 = (_lua_tolstring($L,-1,0)|0); //@line 566 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $42; //@line 566 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 566 "./common/runtime.h"
  $$0 = $data$0;
  STACKTOP = sp;return ($$0|0); //@line 211 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $43 = ___cxa_find_matching_catch_2()|0;
 $44 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 561 "./common/runtime.h"
 $45 = __THREW__; __THREW__ = 0;
 $46 = $45&1;
 if ($46) {
  $50 = ___cxa_find_matching_catch_3(0|0)|0;
  $51 = tempRet0;
  ___clang_call_terminate($50); //@line 561 "./common/runtime.h"
  // unreachable; //@line 561 "./common/runtime.h"
 } else {
  $$01$i = $44;$$02$i = $43;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love10filesystem13w_newFileDataEP9lua_State($L) {
 $L = $L|0;
 var $$01 = 0, $$02 = 0, $$2 = 0, $$3 = 0, $$pr = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $decoder = 0, $length = 0, $t$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $length = sp + 24|0;
 $decoder = sp + 20|0;
 $0 = (_lua_gettop($L)|0); //@line 216 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = ($0|0)==(1); //@line 216 "modules/filesystem/wrap_Filesystem.cpp"
 if (!($1)) {
  HEAP32[$length>>2] = 0; //@line 244 "modules/filesystem/wrap_Filesystem.cpp"
  $32 = (_luaL_checklstring($L,1,$length)|0); //@line 245 "modules/filesystem/wrap_Filesystem.cpp"
  $33 = (_luaL_checklstring($L,2,0)|0); //@line 246 "modules/filesystem/wrap_Filesystem.cpp"
  $34 = (_lua_isstring($L,3)|0); //@line 247 "modules/filesystem/wrap_Filesystem.cpp"
  $35 = ($34|0)==(0); //@line 247 "modules/filesystem/wrap_Filesystem.cpp"
  L3: do {
   if ($35) {
    HEAP32[$decoder>>2] = 0; //@line 249 "modules/filesystem/wrap_Filesystem.cpp"
    $57 = $decoder;
    label = 18;
   } else {
    $36 = (_lua_tolstring($L,3,0)|0); //@line 247 "modules/filesystem/wrap_Filesystem.cpp"
    HEAP32[$decoder>>2] = 0; //@line 249 "modules/filesystem/wrap_Filesystem.cpp"
    $37 = ($36|0)==(0|0); //@line 251 "modules/filesystem/wrap_Filesystem.cpp"
    if ($37) {
     $57 = $decoder;
     label = 18;
    } else {
     $38 = (__ZN4love10filesystem8FileData11getConstantEPKcRNS1_7DecoderE($36,$decoder)|0); //@line 251 "modules/filesystem/wrap_Filesystem.cpp"
     if (!($38)) {
      HEAP32[$vararg_buffer1>>2] = $36; //@line 252 "modules/filesystem/wrap_Filesystem.cpp"
      $39 = (_luaL_error($L,328420,$vararg_buffer1)|0); //@line 252 "modules/filesystem/wrap_Filesystem.cpp"
      $$2 = $39;$58 = $decoder;
      break;
     }
     $$pr = HEAP32[$decoder>>2]|0; //@line 256 "modules/filesystem/wrap_Filesystem.cpp"
     switch ($$pr|0) {
     case 0:  {
      $57 = $decoder;
      label = 18;
      break L3;
      break;
     }
     case 1:  {
      $46 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
      $47 = HEAP32[$46>>2]|0; //@line 262 "modules/filesystem/wrap_Filesystem.cpp"
      $48 = ((($47)) + 80|0); //@line 262 "modules/filesystem/wrap_Filesystem.cpp"
      $49 = HEAP32[$48>>2]|0; //@line 262 "modules/filesystem/wrap_Filesystem.cpp"
      $50 = (FUNCTION_TABLE_iiii[$49 & 8191]($46,$32,$33)|0); //@line 262 "modules/filesystem/wrap_Filesystem.cpp"
      $59 = $decoder;$t$0 = $50;
      label = 21;
      break L3;
      break;
     }
     default: {
      HEAP32[$vararg_buffer4>>2] = $36; //@line 265 "modules/filesystem/wrap_Filesystem.cpp"
      $51 = (_luaL_error($L,328420,$vararg_buffer4)|0); //@line 265 "modules/filesystem/wrap_Filesystem.cpp"
      $$2 = $51;$58 = $decoder;
      break L3;
     }
     }
    }
   }
  } while(0);
  if ((label|0) == 18) {
   $40 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
   $41 = HEAP32[$40>>2]|0; //@line 259 "modules/filesystem/wrap_Filesystem.cpp"
   $42 = ((($41)) + 76|0); //@line 259 "modules/filesystem/wrap_Filesystem.cpp"
   $43 = HEAP32[$42>>2]|0; //@line 259 "modules/filesystem/wrap_Filesystem.cpp"
   $44 = HEAP32[$length>>2]|0; //@line 259 "modules/filesystem/wrap_Filesystem.cpp"
   $45 = (FUNCTION_TABLE_iiiii[$43 & 8191]($40,$32,$44,$33)|0); //@line 259 "modules/filesystem/wrap_Filesystem.cpp"
   $59 = $57;$t$0 = $45;
   label = 21;
  }
  if ((label|0) == 21) {
   __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,7,$t$0); //@line 268 "modules/filesystem/wrap_Filesystem.cpp"
   $52 = HEAP32[$t$0>>2]|0; //@line 269 "modules/filesystem/wrap_Filesystem.cpp"
   $53 = ((($52)) + 12|0); //@line 269 "modules/filesystem/wrap_Filesystem.cpp"
   $54 = HEAP32[$53>>2]|0; //@line 269 "modules/filesystem/wrap_Filesystem.cpp"
   FUNCTION_TABLE_vi[$54 & 8191]($t$0); //@line 269 "modules/filesystem/wrap_Filesystem.cpp"
   $$2 = 1;$58 = $59;
  }
  $$3 = $$2;
  STACKTOP = sp;return ($$3|0); //@line 271 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $2 = (_lua_isstring($L,1)|0); //@line 219 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ($2|0)==(0); //@line 219 "modules/filesystem/wrap_Filesystem.cpp"
 if (!($3)) {
  (__ZN4love12luax_convobjEP9lua_StateiPKcS3_($L,1,354839,328386)|0); //@line 220 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $4 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,1,5)|0); //@line 223 "modules/filesystem/wrap_Filesystem.cpp"
 if (!($4)) {
  $31 = (_luaL_argerror($L,1,328394)|0); //@line 241 "modules/filesystem/wrap_Filesystem.cpp"
  $$3 = $31;
  STACKTOP = sp;return ($$3|0); //@line 271 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $5 = (__ZN4love10filesystem14luax_checkfileEP9lua_Statei($L,1)|0); //@line 225 "modules/filesystem/wrap_Filesystem.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 230 "modules/filesystem/wrap_Filesystem.cpp"
 $7 = ((($6)) + 32|0); //@line 230 "modules/filesystem/wrap_Filesystem.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 230 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 $9 = (invoke_iiii($8|0,($5|0),-1,-1)|0); //@line 230 "modules/filesystem/wrap_Filesystem.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if (!($11)) {
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,7,$9); //@line 236 "modules/filesystem/wrap_Filesystem.cpp"
  $12 = HEAP32[$9>>2]|0; //@line 237 "modules/filesystem/wrap_Filesystem.cpp"
  $13 = ((($12)) + 12|0); //@line 237 "modules/filesystem/wrap_Filesystem.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 237 "modules/filesystem/wrap_Filesystem.cpp"
  FUNCTION_TABLE_vi[$14 & 8191]($9); //@line 237 "modules/filesystem/wrap_Filesystem.cpp"
  $$3 = 1;
  STACKTOP = sp;return ($$3|0); //@line 271 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $15 = ___cxa_find_matching_catch_3(96|0)|0;
 $16 = tempRet0;
 $17 = (_llvm_eh_typeid_for((96|0))|0); //@line 231 "modules/filesystem/wrap_Filesystem.cpp"
 $18 = ($16|0)==($17|0); //@line 231 "modules/filesystem/wrap_Filesystem.cpp"
 if (!($18)) {
  $$01 = $15;$$02 = $16;
  ___resumeException($$01|0);
  // unreachable;
 }
 $19 = (___cxa_begin_catch(($15|0))|0); //@line 231 "modules/filesystem/wrap_Filesystem.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 234 "modules/filesystem/wrap_Filesystem.cpp"
 $21 = ((($20)) + 8|0); //@line 234 "modules/filesystem/wrap_Filesystem.cpp"
 $22 = HEAP32[$21>>2]|0; //@line 234 "modules/filesystem/wrap_Filesystem.cpp"
 $23 = (FUNCTION_TABLE_ii[$22 & 8191]($19)|0); //@line 234 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $23; //@line 234 "modules/filesystem/wrap_Filesystem.cpp"
 $24 = (invoke_iiii(3512,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 234 "modules/filesystem/wrap_Filesystem.cpp"
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if (!($26)) {
  ___cxa_end_catch(); //@line 235 "modules/filesystem/wrap_Filesystem.cpp"
  $$3 = $24;
  STACKTOP = sp;return ($$3|0); //@line 271 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $27 = ___cxa_find_matching_catch_2()|0;
 $28 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 235 "modules/filesystem/wrap_Filesystem.cpp"
 $29 = __THREW__; __THREW__ = 0;
 $30 = $29&1;
 if ($30) {
  $55 = ___cxa_find_matching_catch_3(0|0)|0;
  $56 = tempRet0;
  ___clang_call_terminate($55); //@line 235 "modules/filesystem/wrap_Filesystem.cpp"
  // unreachable; //@line 235 "modules/filesystem/wrap_Filesystem.cpp"
 } else {
  $$01 = $27;$$02 = $28;
  ___resumeException($$01|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love10filesystem21w_getWorkingDirectoryEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 275 "modules/filesystem/wrap_Filesystem.cpp"
 $2 = ((($1)) + 84|0); //@line 275 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 275 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 275 "modules/filesystem/wrap_Filesystem.cpp"
 _lua_pushstring($L,$4); //@line 275 "modules/filesystem/wrap_Filesystem.cpp"
 return 1; //@line 276 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem18w_getUserDirectoryEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 281 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 88|0); //@line 281 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 281 "modules/filesystem/wrap_Filesystem.cpp"
 FUNCTION_TABLE_vii[$4 & 8191]($0,$1); //@line 281 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 invoke_vii(3405,($L|0),($0|0)); //@line 281 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 281 "modules/filesystem/wrap_Filesystem.cpp"
  ___resumeException($7|0);
  // unreachable;
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 281 "modules/filesystem/wrap_Filesystem.cpp"
  STACKTOP = sp;return 1; //@line 282 "modules/filesystem/wrap_Filesystem.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem21w_getAppdataDirectoryEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 287 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 92|0); //@line 287 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 287 "modules/filesystem/wrap_Filesystem.cpp"
 FUNCTION_TABLE_vii[$4 & 8191]($0,$1); //@line 287 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 invoke_vii(3405,($L|0),($0|0)); //@line 287 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 287 "modules/filesystem/wrap_Filesystem.cpp"
  ___resumeException($7|0);
  // unreachable;
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 287 "modules/filesystem/wrap_Filesystem.cpp"
  STACKTOP = sp;return 1; //@line 288 "modules/filesystem/wrap_Filesystem.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem18w_getSaveDirectoryEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 293 "modules/filesystem/wrap_Filesystem.cpp"
 $2 = ((($1)) + 96|0); //@line 293 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 293 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 293 "modules/filesystem/wrap_Filesystem.cpp"
 _lua_pushstring($L,$4); //@line 293 "modules/filesystem/wrap_Filesystem.cpp"
 return 1; //@line 294 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem24w_getSourceBaseDirectoryEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 299 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 100|0); //@line 299 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 299 "modules/filesystem/wrap_Filesystem.cpp"
 FUNCTION_TABLE_vii[$4 & 8191]($0,$1); //@line 299 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 invoke_vii(3405,($L|0),($0|0)); //@line 299 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 299 "modules/filesystem/wrap_Filesystem.cpp"
  ___resumeException($7|0);
  // unreachable;
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 299 "modules/filesystem/wrap_Filesystem.cpp"
  STACKTOP = sp;return 1; //@line 300 "modules/filesystem/wrap_Filesystem.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem18w_getRealDirectoryEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dir = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $dir = sp + 16|0;
 $0 = sp + 4|0;
 $1 = (_luaL_checklstring($L,1,0)|0); //@line 305 "modules/filesystem/wrap_Filesystem.cpp"
 ;HEAP32[$dir>>2]=0|0;HEAP32[$dir+4>>2]=0|0;HEAP32[$dir+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $3 = HEAP32[$2>>2]|0; //@line 310 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = ((($3)) + 104|0); //@line 310 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 310 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 invoke_viii($5|0,($0|0),($2|0),($1|0)); //@line 310 "modules/filesystem/wrap_Filesystem.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 do {
  if ($7) {
   $28 = ___cxa_find_matching_catch_3(96|0)|0;
   $29 = tempRet0;
   $30 = (_llvm_eh_typeid_for((96|0))|0); //@line 311 "modules/filesystem/wrap_Filesystem.cpp"
   $31 = ($29|0)==($30|0); //@line 311 "modules/filesystem/wrap_Filesystem.cpp"
   if (!($31)) {
    $$01 = $29;$$02 = $28;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($dir); //@line 319 "modules/filesystem/wrap_Filesystem.cpp"
    ___resumeException($$02|0);
    // unreachable;
   }
   $32 = (___cxa_begin_catch(($28|0))|0); //@line 311 "modules/filesystem/wrap_Filesystem.cpp"
   $33 = HEAP32[$32>>2]|0; //@line 314 "modules/filesystem/wrap_Filesystem.cpp"
   $34 = ((($33)) + 8|0); //@line 314 "modules/filesystem/wrap_Filesystem.cpp"
   $35 = HEAP32[$34>>2]|0; //@line 314 "modules/filesystem/wrap_Filesystem.cpp"
   $36 = (FUNCTION_TABLE_ii[$35 & 8191]($32)|0); //@line 314 "modules/filesystem/wrap_Filesystem.cpp"
   __THREW__ = 0;
   HEAP32[$vararg_buffer>>2] = $36; //@line 314 "modules/filesystem/wrap_Filesystem.cpp"
   $37 = (invoke_iiii(3512,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 314 "modules/filesystem/wrap_Filesystem.cpp"
   $38 = __THREW__; __THREW__ = 0;
   $39 = $38&1;
   if (!($39)) {
    __THREW__ = 0;
    invoke_v(3378); //@line 315 "modules/filesystem/wrap_Filesystem.cpp"
    $40 = __THREW__; __THREW__ = 0;
    $41 = $40&1;
    if ($41) {
     break;
    } else {
     $$0 = $37;
    }
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($dir); //@line 319 "modules/filesystem/wrap_Filesystem.cpp"
    STACKTOP = sp;return ($$0|0); //@line 319 "modules/filesystem/wrap_Filesystem.cpp"
   }
   $42 = ___cxa_find_matching_catch_2()|0;
   $43 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 315 "modules/filesystem/wrap_Filesystem.cpp"
   $44 = __THREW__; __THREW__ = 0;
   $45 = $44&1;
   if ($45) {
    $48 = ___cxa_find_matching_catch_3(0|0)|0;
    $49 = tempRet0;
    ___clang_call_terminate($48); //@line 315 "modules/filesystem/wrap_Filesystem.cpp"
    // unreachable; //@line 315 "modules/filesystem/wrap_Filesystem.cpp"
   } else {
    $$01 = $43;$$02 = $42;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($dir); //@line 319 "modules/filesystem/wrap_Filesystem.cpp"
    ___resumeException($$02|0);
    // unreachable;
   }
  } else {
   $8 = HEAP8[$dir>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $9 = $8 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $10 = ($9<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if ($10) {
    $14 = ((($dir)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$14>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$dir>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   } else {
    $11 = ((($dir)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $12 = HEAP32[$11>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$12>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $13 = ((($dir)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP32[$13>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   }
   __THREW__ = 0;
   invoke_vii(3368,($dir|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if ($16) {
    $17 = ___cxa_find_matching_catch_3(0|0)|0;
    $18 = tempRet0;
    ___clang_call_terminate($17); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   }
   ;HEAP32[$dir>>2]=HEAP32[$0>>2]|0;HEAP32[$dir+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$dir+8>>2]=HEAP32[$0+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 310 "modules/filesystem/wrap_Filesystem.cpp"
   $19 = HEAP8[$dir>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $20 = $19 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $21 = ($20<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $22 = ((($dir)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $23 = HEAP32[$22>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $24 = ((($dir)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $25 = $21 ? $24 : $23; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($25|0)); //@line 317 "modules/filesystem/wrap_Filesystem.cpp"
   $26 = __THREW__; __THREW__ = 0;
   $27 = $26&1;
   if (!($27)) {
    $$0 = 1;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($dir); //@line 319 "modules/filesystem/wrap_Filesystem.cpp"
    STACKTOP = sp;return ($$0|0); //@line 319 "modules/filesystem/wrap_Filesystem.cpp"
   }
  }
 } while(0);
 $46 = ___cxa_find_matching_catch_2()|0;
 $47 = tempRet0;
 $$01 = $47;$$02 = $46;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($dir); //@line 319 "modules/filesystem/wrap_Filesystem.cpp"
 ___resumeException($$02|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love10filesystem19w_getExecutablePathEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 323 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 176|0); //@line 323 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 323 "modules/filesystem/wrap_Filesystem.cpp"
 FUNCTION_TABLE_vii[$4 & 8191]($0,$1); //@line 323 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 invoke_vii(3405,($L|0),($0|0)); //@line 323 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 323 "modules/filesystem/wrap_Filesystem.cpp"
  ___resumeException($7|0);
  // unreachable;
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 323 "modules/filesystem/wrap_Filesystem.cpp"
  STACKTOP = sp;return 1; //@line 324 "modules/filesystem/wrap_Filesystem.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem8w_existsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 329 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 330 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 108|0); //@line 330 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 330 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = (FUNCTION_TABLE_iii[$4 & 8191]($1,$0)|0); //@line 330 "modules/filesystem/wrap_Filesystem.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$5); //@line 330 "modules/filesystem/wrap_Filesystem.cpp"
 return 1; //@line 331 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem13w_isDirectoryEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 336 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 337 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 112|0); //@line 337 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 337 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = (FUNCTION_TABLE_iii[$4 & 8191]($1,$0)|0); //@line 337 "modules/filesystem/wrap_Filesystem.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$5); //@line 337 "modules/filesystem/wrap_Filesystem.cpp"
 return 1; //@line 338 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem8w_isFileEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 343 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 344 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 116|0); //@line 344 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 344 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = (FUNCTION_TABLE_iii[$4 & 8191]($1,$0)|0); //@line 344 "modules/filesystem/wrap_Filesystem.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$5); //@line 344 "modules/filesystem/wrap_Filesystem.cpp"
 return 1; //@line 345 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem11w_isSymlinkEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 350 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 351 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 120|0); //@line 351 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 351 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = (FUNCTION_TABLE_iii[$4 & 8191]($1,$0)|0); //@line 351 "modules/filesystem/wrap_Filesystem.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$5); //@line 351 "modules/filesystem/wrap_Filesystem.cpp"
 return 1; //@line 352 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem17w_createDirectoryEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 357 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 358 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 124|0); //@line 358 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 358 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = (FUNCTION_TABLE_iii[$4 & 8191]($1,$0)|0); //@line 358 "modules/filesystem/wrap_Filesystem.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$5); //@line 358 "modules/filesystem/wrap_Filesystem.cpp"
 return 1; //@line 359 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem8w_removeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 364 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 365 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 128|0); //@line 365 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 365 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = (FUNCTION_TABLE_iii[$4 & 8191]($1,$0)|0); //@line 365 "modules/filesystem/wrap_Filesystem.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$5); //@line 365 "modules/filesystem/wrap_Filesystem.cpp"
 return 1; //@line 366 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem6w_readEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 369 "modules/filesystem/wrap_Filesystem.cpp"
 $vararg_buffer = sp; //@line 369 "modules/filesystem/wrap_Filesystem.cpp"
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 371 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = (_luaL_optinteger($L,2,-1)|0); //@line 372 "modules/filesystem/wrap_Filesystem.cpp"
 $2 = ($1|0)<(0); //@line 372 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = $2 << 31 >> 31; //@line 372 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $5 = HEAP32[$4>>2]|0; //@line 377 "modules/filesystem/wrap_Filesystem.cpp"
 $6 = ((($5)) + 132|0); //@line 377 "modules/filesystem/wrap_Filesystem.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 377 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 $8 = (invoke_iiiii($7|0,($4|0),($0|0),($1|0),($3|0))|0); //@line 377 "modules/filesystem/wrap_Filesystem.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if (!($10)) {
  $11 = ($8|0)==(0|0); //@line 384 "modules/filesystem/wrap_Filesystem.cpp"
  if ($11) {
   $28 = (__ZN4love10filesystem12luax_ioErrorEP9lua_StatePKcz($L,328449,$vararg_buffer1)|0); //@line 385 "modules/filesystem/wrap_Filesystem.cpp"
   $$0 = $28;
   STACKTOP = sp;return ($$0|0); //@line 397 "modules/filesystem/wrap_Filesystem.cpp"
  } else {
   $29 = HEAP32[$8>>2]|0; //@line 388 "modules/filesystem/wrap_Filesystem.cpp"
   $30 = ((($29)) + 16|0); //@line 388 "modules/filesystem/wrap_Filesystem.cpp"
   $31 = HEAP32[$30>>2]|0; //@line 388 "modules/filesystem/wrap_Filesystem.cpp"
   $32 = (FUNCTION_TABLE_ii[$31 & 8191]($8)|0); //@line 388 "modules/filesystem/wrap_Filesystem.cpp"
   $33 = HEAP32[$8>>2]|0; //@line 388 "modules/filesystem/wrap_Filesystem.cpp"
   $34 = ((($33)) + 20|0); //@line 388 "modules/filesystem/wrap_Filesystem.cpp"
   $35 = HEAP32[$34>>2]|0; //@line 388 "modules/filesystem/wrap_Filesystem.cpp"
   $36 = (FUNCTION_TABLE_ii[$35 & 8191]($8)|0); //@line 388 "modules/filesystem/wrap_Filesystem.cpp"
   _lua_pushlstring($L,$32,$36); //@line 388 "modules/filesystem/wrap_Filesystem.cpp"
   $37 = HEAP32[$8>>2]|0; //@line 391 "modules/filesystem/wrap_Filesystem.cpp"
   $38 = ((($37)) + 20|0); //@line 391 "modules/filesystem/wrap_Filesystem.cpp"
   $39 = HEAP32[$38>>2]|0; //@line 391 "modules/filesystem/wrap_Filesystem.cpp"
   $40 = (FUNCTION_TABLE_ii[$39 & 8191]($8)|0); //@line 391 "modules/filesystem/wrap_Filesystem.cpp"
   _lua_pushinteger($L,$40); //@line 391 "modules/filesystem/wrap_Filesystem.cpp"
   $41 = HEAP32[$8>>2]|0; //@line 394 "modules/filesystem/wrap_Filesystem.cpp"
   $42 = ((($41)) + 12|0); //@line 394 "modules/filesystem/wrap_Filesystem.cpp"
   $43 = HEAP32[$42>>2]|0; //@line 394 "modules/filesystem/wrap_Filesystem.cpp"
   FUNCTION_TABLE_vi[$43 & 8191]($8); //@line 394 "modules/filesystem/wrap_Filesystem.cpp"
   $$0 = 2;
   STACKTOP = sp;return ($$0|0); //@line 397 "modules/filesystem/wrap_Filesystem.cpp"
  }
 }
 $12 = ___cxa_find_matching_catch_3(96|0)|0;
 $13 = tempRet0;
 $14 = (_llvm_eh_typeid_for((96|0))|0); //@line 378 "modules/filesystem/wrap_Filesystem.cpp"
 $15 = ($13|0)==($14|0); //@line 378 "modules/filesystem/wrap_Filesystem.cpp"
 if (!($15)) {
  $$01 = $13;$$02 = $12;
  ___resumeException($$02|0);
  // unreachable;
 }
 $16 = (___cxa_begin_catch(($12|0))|0); //@line 378 "modules/filesystem/wrap_Filesystem.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 381 "modules/filesystem/wrap_Filesystem.cpp"
 $18 = ((($17)) + 8|0); //@line 381 "modules/filesystem/wrap_Filesystem.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 381 "modules/filesystem/wrap_Filesystem.cpp"
 $20 = (FUNCTION_TABLE_ii[$19 & 8191]($16)|0); //@line 381 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $20; //@line 381 "modules/filesystem/wrap_Filesystem.cpp"
 $21 = (invoke_iiii(3512,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 381 "modules/filesystem/wrap_Filesystem.cpp"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if (!($23)) {
  ___cxa_end_catch(); //@line 382 "modules/filesystem/wrap_Filesystem.cpp"
  $$0 = $21;
  STACKTOP = sp;return ($$0|0); //@line 397 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $24 = ___cxa_find_matching_catch_2()|0;
 $25 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 382 "modules/filesystem/wrap_Filesystem.cpp"
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  $44 = ___cxa_find_matching_catch_3(0|0)|0;
  $45 = tempRet0;
  ___clang_call_terminate($44); //@line 382 "modules/filesystem/wrap_Filesystem.cpp"
  // unreachable; //@line 382 "modules/filesystem/wrap_Filesystem.cpp"
 } else {
  $$01 = $25;$$02 = $24;
  ___resumeException($$02|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love10filesystem7w_writeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love10filesystemL17w_write_or_appendEP9lua_StateNS0_4File4ModeE($L,2)|0); //@line 438 "modules/filesystem/wrap_Filesystem.cpp"
 return ($0|0); //@line 438 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystemL17w_write_or_appendEP9lua_StateNS0_4File4ModeE($L,$mode) {
 $L = $L|0;
 $mode = $mode|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $input$0 = 0, $len = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $len = sp + 4|0;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 401 "modules/filesystem/wrap_Filesystem.cpp"
 HEAP32[$len>>2] = 0; //@line 404 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,2,2)|0); //@line 406 "modules/filesystem/wrap_Filesystem.cpp"
 do {
  if ($1) {
   $2 = (_lua_touserdata($L,2)|0); //@line 516 "./common/runtime.h"
   $3 = ((($2)) + 4|0); //@line 516 "./common/runtime.h"
   $4 = HEAP32[$3>>2]|0; //@line 516 "./common/runtime.h"
   $5 = HEAP32[$4>>2]|0; //@line 409 "modules/filesystem/wrap_Filesystem.cpp"
   $6 = ((($5)) + 16|0); //@line 409 "modules/filesystem/wrap_Filesystem.cpp"
   $7 = HEAP32[$6>>2]|0; //@line 409 "modules/filesystem/wrap_Filesystem.cpp"
   $8 = (FUNCTION_TABLE_ii[$7 & 8191]($4)|0); //@line 409 "modules/filesystem/wrap_Filesystem.cpp"
   $9 = HEAP32[$4>>2]|0; //@line 410 "modules/filesystem/wrap_Filesystem.cpp"
   $10 = ((($9)) + 20|0); //@line 410 "modules/filesystem/wrap_Filesystem.cpp"
   $11 = HEAP32[$10>>2]|0; //@line 410 "modules/filesystem/wrap_Filesystem.cpp"
   $12 = (FUNCTION_TABLE_ii[$11 & 8191]($4)|0); //@line 410 "modules/filesystem/wrap_Filesystem.cpp"
   HEAP32[$len>>2] = $12; //@line 410 "modules/filesystem/wrap_Filesystem.cpp"
   $17 = $12;$input$0 = $8;
  } else {
   $13 = (_lua_isstring($L,2)|0); //@line 412 "modules/filesystem/wrap_Filesystem.cpp"
   $14 = ($13|0)==(0); //@line 412 "modules/filesystem/wrap_Filesystem.cpp"
   if (!($14)) {
    $15 = (_lua_tolstring($L,2,$len)|0); //@line 413 "modules/filesystem/wrap_Filesystem.cpp"
    $$pre = HEAP32[$len>>2]|0; //@line 418 "modules/filesystem/wrap_Filesystem.cpp"
    $17 = $$pre;$input$0 = $15;
    break;
   }
   $16 = (_luaL_argerror($L,2,328473)|0); //@line 415 "modules/filesystem/wrap_Filesystem.cpp"
   $$0 = $16;
   STACKTOP = sp;return ($$0|0); //@line 434 "modules/filesystem/wrap_Filesystem.cpp"
  }
 } while(0);
 $18 = (_luaL_optinteger($L,3,$17)|0); //@line 418 "modules/filesystem/wrap_Filesystem.cpp"
 HEAP32[$len>>2] = $18; //@line 418 "modules/filesystem/wrap_Filesystem.cpp"
 $19 = ($mode|0)==(3); //@line 422 "modules/filesystem/wrap_Filesystem.cpp"
 $20 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $21 = HEAP32[$20>>2]|0; //@line 423 "modules/filesystem/wrap_Filesystem.cpp"
 if ($19) {
  $22 = ((($21)) + 140|0); //@line 423 "modules/filesystem/wrap_Filesystem.cpp"
  $23 = HEAP32[$22>>2]|0; //@line 423 "modules/filesystem/wrap_Filesystem.cpp"
  __THREW__ = 0;
  invoke_viiiii($23|0,($20|0),($0|0),($input$0|0),($18|0),0); //@line 423 "modules/filesystem/wrap_Filesystem.cpp"
  $24 = __THREW__; __THREW__ = 0;
  $25 = $24&1;
  if (!($25)) {
   label = 13;
  }
 } else {
  $38 = ((($21)) + 136|0); //@line 425 "modules/filesystem/wrap_Filesystem.cpp"
  $39 = HEAP32[$38>>2]|0; //@line 425 "modules/filesystem/wrap_Filesystem.cpp"
  __THREW__ = 0;
  invoke_viiiii($39|0,($20|0),($0|0),($input$0|0),($18|0),0); //@line 425 "modules/filesystem/wrap_Filesystem.cpp"
  $40 = __THREW__; __THREW__ = 0;
  $41 = $40&1;
  if (!($41)) {
   label = 13;
  }
 }
 if ((label|0) == 13) {
  __ZN4love16luax_pushbooleanEP9lua_Stateb($L,1); //@line 432 "modules/filesystem/wrap_Filesystem.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 434 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $26 = ___cxa_find_matching_catch_3(96|0)|0;
 $27 = tempRet0;
 $28 = (_llvm_eh_typeid_for((96|0))|0); //@line 426 "modules/filesystem/wrap_Filesystem.cpp"
 $29 = ($27|0)==($28|0); //@line 426 "modules/filesystem/wrap_Filesystem.cpp"
 if (!($29)) {
  $$01 = $27;$$02 = $26;
  ___resumeException($$02|0);
  // unreachable;
 }
 $30 = (___cxa_begin_catch(($26|0))|0); //@line 426 "modules/filesystem/wrap_Filesystem.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 429 "modules/filesystem/wrap_Filesystem.cpp"
 $32 = ((($31)) + 8|0); //@line 429 "modules/filesystem/wrap_Filesystem.cpp"
 $33 = HEAP32[$32>>2]|0; //@line 429 "modules/filesystem/wrap_Filesystem.cpp"
 $34 = (FUNCTION_TABLE_ii[$33 & 8191]($30)|0); //@line 429 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $34; //@line 429 "modules/filesystem/wrap_Filesystem.cpp"
 $35 = (invoke_iiii(3512,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 429 "modules/filesystem/wrap_Filesystem.cpp"
 $36 = __THREW__; __THREW__ = 0;
 $37 = $36&1;
 if (!($37)) {
  ___cxa_end_catch(); //@line 430 "modules/filesystem/wrap_Filesystem.cpp"
  $$0 = $35;
  STACKTOP = sp;return ($$0|0); //@line 434 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $42 = ___cxa_find_matching_catch_2()|0;
 $43 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 430 "modules/filesystem/wrap_Filesystem.cpp"
 $44 = __THREW__; __THREW__ = 0;
 $45 = $44&1;
 if ($45) {
  $46 = ___cxa_find_matching_catch_3(0|0)|0;
  $47 = tempRet0;
  ___clang_call_terminate($46); //@line 430 "modules/filesystem/wrap_Filesystem.cpp"
  // unreachable; //@line 430 "modules/filesystem/wrap_Filesystem.cpp"
 } else {
  $$01 = $43;$$02 = $42;
  ___resumeException($$02|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love10filesystem8w_appendEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love10filesystemL17w_write_or_appendEP9lua_StateNS0_4File4ModeE($L,3)|0); //@line 443 "modules/filesystem/wrap_Filesystem.cpp"
 return ($0|0); //@line 443 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem19w_getDirectoryItemsEP9lua_State($L) {
 $L = $L|0;
 var $$lcssa = 0, $$lcssa24 = 0, $$lcssa26 = 0, $$pre$i = 0, $$pre$i3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $i$0 = 0, $items = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $items = sp;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 448 "modules/filesystem/wrap_Filesystem.cpp"
 HEAP32[$items>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($items)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($items)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $4 = HEAP32[$3>>2]|0; //@line 451 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = ((($4)) + 144|0); //@line 451 "modules/filesystem/wrap_Filesystem.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 451 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 invoke_viii($6|0,($3|0),($0|0),($items|0)); //@line 451 "modules/filesystem/wrap_Filesystem.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 do {
  if ($8) {
   label = 11;
  } else {
   $9 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $10 = HEAP32[$items>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $11 = (($9) - ($10))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $12 = (($11|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __THREW__ = 0;
   invoke_viii(3393,($L|0),($12|0),0); //@line 453 "modules/filesystem/wrap_Filesystem.cpp"
   $13 = __THREW__; __THREW__ = 0;
   $14 = $13&1;
   if ($14) {
    label = 11;
   } else {
    $i$0 = 0;
    while(1) {
     $15 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $16 = HEAP32[$items>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $17 = (($15) - ($16))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $18 = (($17|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $19 = ($i$0|0)<($18|0); //@line 455 "modules/filesystem/wrap_Filesystem.cpp"
     $20 = $16; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($19)) {
      $$lcssa = $15;$$lcssa24 = $16;$$lcssa26 = $20;
      break;
     }
     $43 = (($20) + (($i$0*12)|0)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $44 = HEAP8[$43>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $45 = $44 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $46 = ($45<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($46) {
      $49 = ((($43)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $50 = $49;
     } else {
      $47 = (((($20) + (($i$0*12)|0)|0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $48 = HEAP32[$47>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $50 = $48;
     }
     __THREW__ = 0;
     invoke_vii(3380,($L|0),($50|0)); //@line 457 "modules/filesystem/wrap_Filesystem.cpp"
     $51 = __THREW__; __THREW__ = 0;
     $52 = $51&1;
     if ($52) {
      label = 10;
      break;
     }
     $53 = (($i$0) + 1)|0; //@line 458 "modules/filesystem/wrap_Filesystem.cpp"
     __THREW__ = 0;
     invoke_viii(3395,($L|0),-2,($53|0)); //@line 458 "modules/filesystem/wrap_Filesystem.cpp"
     $54 = __THREW__; __THREW__ = 0;
     $55 = $54&1;
     if ($55) {
      label = 10;
      break;
     } else {
      $i$0 = $53;
     }
    }
    if ((label|0) == 10) {
     $30 = ___cxa_find_matching_catch_2()|0;
     $31 = tempRet0;
     $lpad$phi$index = $30;$lpad$phi$index2 = $31;
     break;
    }
    $21 = $$lcssa; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $22 = ($$lcssa24|0)==(0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($22) {
     STACKTOP = sp;return 1; //@line 463 "modules/filesystem/wrap_Filesystem.cpp"
    }
    $23 = $$lcssa24; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $24 = ($21|0)==($$lcssa26|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($24) {
     $29 = $23;
    } else {
     $26 = $21;
     while(1) {
      $25 = ((($26)) + -12|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$1>>2] = $25; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($25); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $27 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $28 = ($27|0)==($$lcssa26|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($28) {
       break;
      } else {
       $26 = $27;
      }
     }
     $$pre$i3 = HEAP32[$items>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $29 = $$pre$i3;
    }
    __ZdlPv($29); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    STACKTOP = sp;return 1; //@line 463 "modules/filesystem/wrap_Filesystem.cpp"
   }
  }
 } while(0);
 if ((label|0) == 11) {
  $32 = ___cxa_find_matching_catch_2()|0;
  $33 = tempRet0;
  $lpad$phi$index = $32;$lpad$phi$index2 = $33;
 }
 $34 = HEAP32[$items>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $35 = ($34|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($35) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $36 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $37 = ($36|0)==($34|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($37) {
  $42 = $34;
 } else {
  $39 = $36;
  while(1) {
   $38 = ((($39)) + -12|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$1>>2] = $38; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($38); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $40 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $41 = ($40|0)==($34|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($41) {
    break;
   } else {
    $39 = $40;
   }
  }
  $$pre$i = HEAP32[$items>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $42 = $$pre$i;
 }
 __ZdlPv($42); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love10filesystem7w_linesEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 465 "modules/filesystem/wrap_Filesystem.cpp"
 $vararg_buffer = sp; //@line 465 "modules/filesystem/wrap_Filesystem.cpp"
 $0 = (_lua_isstring($L,1)|0); //@line 469 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = ($0|0)==(0); //@line 469 "modules/filesystem/wrap_Filesystem.cpp"
 if ($1) {
  $37 = (_luaL_argerror($L,1,328497)|0); //@line 486 "modules/filesystem/wrap_Filesystem.cpp"
  $$1 = $37;
  STACKTOP = sp;return ($$1|0); //@line 490 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $2 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $3 = HEAP32[$2>>2]|0; //@line 471 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = ((($3)) + 72|0); //@line 471 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 471 "modules/filesystem/wrap_Filesystem.cpp"
 $6 = (_lua_tolstring($L,1,0)|0); //@line 471 "modules/filesystem/wrap_Filesystem.cpp"
 $7 = (FUNCTION_TABLE_iii[$5 & 8191]($2,$6)|0); //@line 471 "modules/filesystem/wrap_Filesystem.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 474 "modules/filesystem/wrap_Filesystem.cpp"
 $9 = ((($8)) + 16|0); //@line 474 "modules/filesystem/wrap_Filesystem.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 474 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 $11 = (invoke_iii($10|0,($7|0),1)|0); //@line 474 "modules/filesystem/wrap_Filesystem.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 do {
  if ($13) {
   $14 = ___cxa_find_matching_catch_3(8992|0)|0;
   $15 = tempRet0;
   $16 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $17 = ($15|0)==($16|0); //@line 535 "./common/runtime.h"
   if (!($17)) {
    $$01$i = $15;$$02$i = $14;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $18 = (___cxa_begin_catch(($14|0))|0); //@line 535 "./common/runtime.h"
   $19 = HEAP32[$18>>2]|0; //@line 539 "./common/runtime.h"
   $20 = ((($19)) + 8|0); //@line 539 "./common/runtime.h"
   $21 = HEAP32[$20>>2]|0; //@line 539 "./common/runtime.h"
   $22 = (FUNCTION_TABLE_ii[$21 & 8191]($18)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($22|0)); //@line 539 "./common/runtime.h"
   $23 = __THREW__; __THREW__ = 0;
   $24 = $23&1;
   if (!($24)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $25 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $25; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $33 = $7;
    break;
   }
   $26 = ___cxa_find_matching_catch_2()|0;
   $27 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $28 = __THREW__; __THREW__ = 0;
   $29 = $28&1;
   if ($29) {
    $30 = ___cxa_find_matching_catch_3(0|0)|0;
    $31 = tempRet0;
    ___clang_call_terminate($30); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $27;$$02$i = $26;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   if ($11) {
    __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,5,$7); //@line 482 "modules/filesystem/wrap_Filesystem.cpp"
    $38 = HEAP32[$7>>2]|0; //@line 483 "modules/filesystem/wrap_Filesystem.cpp"
    $39 = ((($38)) + 12|0); //@line 483 "modules/filesystem/wrap_Filesystem.cpp"
    $40 = HEAP32[$39>>2]|0; //@line 483 "modules/filesystem/wrap_Filesystem.cpp"
    FUNCTION_TABLE_vi[$40 & 8191]($7); //@line 483 "modules/filesystem/wrap_Filesystem.cpp"
    _lua_pushcclosure($L,3513,1); //@line 488 "modules/filesystem/wrap_Filesystem.cpp"
    $$1 = 1;
    STACKTOP = sp;return ($$1|0); //@line 490 "modules/filesystem/wrap_Filesystem.cpp"
   } else {
    $33 = $7;
   }
  }
 } while(0);
 $32 = HEAP32[$33>>2]|0; //@line 478 "modules/filesystem/wrap_Filesystem.cpp"
 $34 = ((($32)) + 12|0); //@line 478 "modules/filesystem/wrap_Filesystem.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 478 "modules/filesystem/wrap_Filesystem.cpp"
 FUNCTION_TABLE_vi[$35 & 8191]($33); //@line 478 "modules/filesystem/wrap_Filesystem.cpp"
 $36 = (_luaL_error($L,366514,$vararg_buffer1)|0); //@line 479 "modules/filesystem/wrap_Filesystem.cpp"
 $$1 = $36;
 STACKTOP = sp;return ($$1|0); //@line 490 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem6w_loadEP9lua_State($L) {
 $L = $L|0;
 var $$01 = 0, $$02 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index8Z2D = 0, $eh$lpad$body$indexZ2D = 0, $filename = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $filename = sp + 32|0;
 $0 = sp + 20|0;
 $1 = (_luaL_checklstring($L,1,0)|0); //@line 494 "modules/filesystem/wrap_Filesystem.cpp"
 $2 = (_strlen($1)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($filename,$1,$2); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $4 = HEAP32[$3>>2]|0; //@line 499 "modules/filesystem/wrap_Filesystem.cpp"
 $5 = ((($4)) + 132|0); //@line 499 "modules/filesystem/wrap_Filesystem.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 499 "modules/filesystem/wrap_Filesystem.cpp"
 $7 = HEAP8[$filename>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $7 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = ($8<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = ((($filename)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = HEAP32[$10>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = ((($filename)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = $9 ? $12 : $11; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 $14 = (invoke_iiiii($6|0,($3|0),($13|0),-1,-1)|0); //@line 499 "modules/filesystem/wrap_Filesystem.cpp"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 L1: do {
  if ($16) {
   $23 = ___cxa_find_matching_catch_3(96|0)|0;
   $24 = tempRet0;
   $25 = (_llvm_eh_typeid_for((96|0))|0); //@line 500 "modules/filesystem/wrap_Filesystem.cpp"
   $26 = ($24|0)==($25|0); //@line 500 "modules/filesystem/wrap_Filesystem.cpp"
   if (!($26)) {
    $$01 = $24;$$02 = $23;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($filename); //@line 520 "modules/filesystem/wrap_Filesystem.cpp"
    ___resumeException($$02|0);
    // unreachable;
   }
   $27 = (___cxa_begin_catch(($23|0))|0); //@line 500 "modules/filesystem/wrap_Filesystem.cpp"
   $28 = HEAP32[$27>>2]|0; //@line 503 "modules/filesystem/wrap_Filesystem.cpp"
   $29 = ((($28)) + 8|0); //@line 503 "modules/filesystem/wrap_Filesystem.cpp"
   $30 = HEAP32[$29>>2]|0; //@line 503 "modules/filesystem/wrap_Filesystem.cpp"
   $31 = (FUNCTION_TABLE_ii[$30 & 8191]($27)|0); //@line 503 "modules/filesystem/wrap_Filesystem.cpp"
   __THREW__ = 0;
   HEAP32[$vararg_buffer>>2] = $31; //@line 503 "modules/filesystem/wrap_Filesystem.cpp"
   $32 = (invoke_iiii(3512,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 503 "modules/filesystem/wrap_Filesystem.cpp"
   $33 = __THREW__; __THREW__ = 0;
   $34 = $33&1;
   if (!($34)) {
    __THREW__ = 0;
    invoke_v(3378); //@line 504 "modules/filesystem/wrap_Filesystem.cpp"
    $35 = __THREW__; __THREW__ = 0;
    $36 = $35&1;
    if ($36) {
     label = 7;
     break;
    } else {
     $$1 = $32;
    }
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($filename); //@line 520 "modules/filesystem/wrap_Filesystem.cpp"
    STACKTOP = sp;return ($$1|0); //@line 520 "modules/filesystem/wrap_Filesystem.cpp"
   }
   $37 = ___cxa_find_matching_catch_2()|0;
   $38 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 504 "modules/filesystem/wrap_Filesystem.cpp"
   $39 = __THREW__; __THREW__ = 0;
   $40 = $39&1;
   if ($40) {
    $98 = ___cxa_find_matching_catch_3(0|0)|0;
    $99 = tempRet0;
    ___clang_call_terminate($98); //@line 504 "modules/filesystem/wrap_Filesystem.cpp"
    // unreachable; //@line 504 "modules/filesystem/wrap_Filesystem.cpp"
   } else {
    $$01 = $38;$$02 = $37;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($filename); //@line 520 "modules/filesystem/wrap_Filesystem.cpp"
    ___resumeException($$02|0);
    // unreachable;
   }
  } else {
   $17 = HEAP32[$14>>2]|0; //@line 506 "modules/filesystem/wrap_Filesystem.cpp"
   $18 = ((($17)) + 16|0); //@line 506 "modules/filesystem/wrap_Filesystem.cpp"
   $19 = HEAP32[$18>>2]|0; //@line 506 "modules/filesystem/wrap_Filesystem.cpp"
   __THREW__ = 0;
   $20 = (invoke_ii($19|0,($14|0))|0); //@line 506 "modules/filesystem/wrap_Filesystem.cpp"
   $21 = __THREW__; __THREW__ = 0;
   $22 = $21&1;
   if ($22) {
    label = 7;
   } else {
    $43 = HEAP32[$14>>2]|0; //@line 506 "modules/filesystem/wrap_Filesystem.cpp"
    $44 = ((($43)) + 20|0); //@line 506 "modules/filesystem/wrap_Filesystem.cpp"
    $45 = HEAP32[$44>>2]|0; //@line 506 "modules/filesystem/wrap_Filesystem.cpp"
    __THREW__ = 0;
    $46 = (invoke_ii($45|0,($14|0))|0); //@line 506 "modules/filesystem/wrap_Filesystem.cpp"
    $47 = __THREW__; __THREW__ = 0;
    $48 = $47&1;
    if ($48) {
     label = 7;
    } else {
     ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $49 = HEAP8[$filename>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $50 = $49 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $51 = ($50<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $52 = ((($filename)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $53 = HEAP32[$52>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $54 = $49&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $55 = $54 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $56 = $51 ? $55 : $53; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $57 = (($56) + 1)|0; //@line 3957 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     __THREW__ = 0;
     invoke_viiii(3514,($0|0),(442495|0),1,($57|0)); //@line 3957 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $58 = __THREW__; __THREW__ = 0;
     $59 = $58&1;
     if (!($59)) {
      $60 = HEAP8[$filename>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $61 = $60 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $62 = ($61<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $63 = HEAP32[$10>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $64 = $62 ? $12 : $63; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __THREW__ = 0;
      (invoke_iiii(3462,($0|0),($64|0),($56|0))|0); //@line 3958 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $65 = __THREW__; __THREW__ = 0;
      $66 = $65&1;
      if (!($66)) {
       $69 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $70 = $69 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $71 = ($70<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $72 = ((($0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $73 = HEAP32[$72>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $74 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $75 = $71 ? $74 : $73; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       __THREW__ = 0;
       $76 = (invoke_iiiii(3515,($L|0),($20|0),($46|0),($75|0))|0); //@line 506 "modules/filesystem/wrap_Filesystem.cpp"
       $77 = __THREW__; __THREW__ = 0;
       $78 = $77&1;
       if ($78) {
        $84 = ___cxa_find_matching_catch_2()|0;
        $85 = tempRet0;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 506 "modules/filesystem/wrap_Filesystem.cpp"
        $$01 = $85;$$02 = $84;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($filename); //@line 520 "modules/filesystem/wrap_Filesystem.cpp"
        ___resumeException($$02|0);
        // unreachable;
       }
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 506 "modules/filesystem/wrap_Filesystem.cpp"
       $79 = HEAP32[$14>>2]|0; //@line 508 "modules/filesystem/wrap_Filesystem.cpp"
       $80 = ((($79)) + 12|0); //@line 508 "modules/filesystem/wrap_Filesystem.cpp"
       $81 = HEAP32[$80>>2]|0; //@line 508 "modules/filesystem/wrap_Filesystem.cpp"
       __THREW__ = 0;
       invoke_vi($81|0,($14|0)); //@line 508 "modules/filesystem/wrap_Filesystem.cpp"
       $82 = __THREW__; __THREW__ = 0;
       $83 = $82&1;
       if ($83) {
        label = 7;
        break;
       }
       switch ($76|0) {
       case 4:  {
        __THREW__ = 0;
        $86 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 514 "modules/filesystem/wrap_Filesystem.cpp"
        $87 = __THREW__; __THREW__ = 0;
        $88 = $87&1;
        if ($88) {
         label = 7;
         break L1;
        }
        __THREW__ = 0;
        HEAP32[$vararg_buffer1>>2] = $86; //@line 514 "modules/filesystem/wrap_Filesystem.cpp"
        $89 = (invoke_iiii(3389,($L|0),(328516|0),($vararg_buffer1|0))|0); //@line 514 "modules/filesystem/wrap_Filesystem.cpp"
        $90 = __THREW__; __THREW__ = 0;
        $91 = $90&1;
        if ($91) {
         label = 7;
         break L1;
        } else {
         $$1 = $89;
        }
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($filename); //@line 520 "modules/filesystem/wrap_Filesystem.cpp"
        STACKTOP = sp;return ($$1|0); //@line 520 "modules/filesystem/wrap_Filesystem.cpp"
        break;
       }
       case 3:  {
        __THREW__ = 0;
        $92 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 516 "modules/filesystem/wrap_Filesystem.cpp"
        $93 = __THREW__; __THREW__ = 0;
        $94 = $93&1;
        if ($94) {
         label = 7;
         break L1;
        }
        __THREW__ = 0;
        HEAP32[$vararg_buffer4>>2] = $92; //@line 516 "modules/filesystem/wrap_Filesystem.cpp"
        $95 = (invoke_iiii(3389,($L|0),(328545|0),($vararg_buffer4|0))|0); //@line 516 "modules/filesystem/wrap_Filesystem.cpp"
        $96 = __THREW__; __THREW__ = 0;
        $97 = $96&1;
        if ($97) {
         label = 7;
         break L1;
        } else {
         $$1 = $95;
        }
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($filename); //@line 520 "modules/filesystem/wrap_Filesystem.cpp"
        STACKTOP = sp;return ($$1|0); //@line 520 "modules/filesystem/wrap_Filesystem.cpp"
        break;
       }
       default: {
        $$1 = 1;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($filename); //@line 520 "modules/filesystem/wrap_Filesystem.cpp"
        STACKTOP = sp;return ($$1|0); //@line 520 "modules/filesystem/wrap_Filesystem.cpp"
       }
       }
      }
     }
     $67 = ___cxa_find_matching_catch_2()|0;
     $68 = tempRet0;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 3960 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $eh$lpad$body$index8Z2D = $68;$eh$lpad$body$indexZ2D = $67;
    }
   }
  }
 } while(0);
 if ((label|0) == 7) {
  $41 = ___cxa_find_matching_catch_2()|0;
  $42 = tempRet0;
  $eh$lpad$body$index8Z2D = $42;$eh$lpad$body$indexZ2D = $41;
 }
 $$01 = $eh$lpad$body$index8Z2D;$$02 = $eh$lpad$body$indexZ2D;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($filename); //@line 520 "modules/filesystem/wrap_Filesystem.cpp"
 ___resumeException($$02|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love10filesystem17w_getLastModifiedEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 522 "modules/filesystem/wrap_Filesystem.cpp"
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 524 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 529 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 148|0); //@line 529 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 529 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 $5 = (invoke_iii($4|0,($1|0),($0|0))|0); //@line 529 "modules/filesystem/wrap_Filesystem.cpp"
 $6 = tempRet0; //@line 529 "modules/filesystem/wrap_Filesystem.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if (!($8)) {
  $9 = (+($5>>>0)) + (4294967296.0*(+($6|0))); //@line 536 "modules/filesystem/wrap_Filesystem.cpp"
  _lua_pushnumber($L,$9); //@line 536 "modules/filesystem/wrap_Filesystem.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 538 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $10 = ___cxa_find_matching_catch_3(96|0)|0;
 $11 = tempRet0;
 $12 = (_llvm_eh_typeid_for((96|0))|0); //@line 530 "modules/filesystem/wrap_Filesystem.cpp"
 $13 = ($11|0)==($12|0); //@line 530 "modules/filesystem/wrap_Filesystem.cpp"
 if (!($13)) {
  $$01 = $11;$$02 = $10;
  ___resumeException($$02|0);
  // unreachable;
 }
 $14 = (___cxa_begin_catch(($10|0))|0); //@line 530 "modules/filesystem/wrap_Filesystem.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 533 "modules/filesystem/wrap_Filesystem.cpp"
 $16 = ((($15)) + 8|0); //@line 533 "modules/filesystem/wrap_Filesystem.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 533 "modules/filesystem/wrap_Filesystem.cpp"
 $18 = (FUNCTION_TABLE_ii[$17 & 8191]($14)|0); //@line 533 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $18; //@line 533 "modules/filesystem/wrap_Filesystem.cpp"
 $19 = (invoke_iiii(3512,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 533 "modules/filesystem/wrap_Filesystem.cpp"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if (!($21)) {
  ___cxa_end_catch(); //@line 534 "modules/filesystem/wrap_Filesystem.cpp"
  $$0 = $19;
  STACKTOP = sp;return ($$0|0); //@line 538 "modules/filesystem/wrap_Filesystem.cpp"
 }
 $22 = ___cxa_find_matching_catch_2()|0;
 $23 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 534 "modules/filesystem/wrap_Filesystem.cpp"
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $26 = ___cxa_find_matching_catch_3(0|0)|0;
  $27 = tempRet0;
  ___clang_call_terminate($26); //@line 534 "modules/filesystem/wrap_Filesystem.cpp"
  // unreachable; //@line 534 "modules/filesystem/wrap_Filesystem.cpp"
 } else {
  $$01 = $23;$$02 = $22;
  ___resumeException($$02|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love10filesystem9w_getSizeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0; //@line 540 "modules/filesystem/wrap_Filesystem.cpp"
 $vararg_buffer1 = sp + 8|0; //@line 540 "modules/filesystem/wrap_Filesystem.cpp"
 $vararg_buffer = sp; //@line 540 "modules/filesystem/wrap_Filesystem.cpp"
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 542 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 547 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 152|0); //@line 547 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 547 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 $5 = (invoke_iii($4|0,($1|0),($0|0))|0); //@line 547 "modules/filesystem/wrap_Filesystem.cpp"
 $6 = tempRet0; //@line 547 "modules/filesystem/wrap_Filesystem.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $12 = ___cxa_find_matching_catch_3(96|0)|0;
  $13 = tempRet0;
  $14 = (_llvm_eh_typeid_for((96|0))|0); //@line 548 "modules/filesystem/wrap_Filesystem.cpp"
  $15 = ($13|0)==($14|0); //@line 548 "modules/filesystem/wrap_Filesystem.cpp"
  if (!($15)) {
   $$01 = $13;$$02 = $12;
   ___resumeException($$02|0);
   // unreachable;
  }
  $16 = (___cxa_begin_catch(($12|0))|0); //@line 548 "modules/filesystem/wrap_Filesystem.cpp"
  $17 = HEAP32[$16>>2]|0; //@line 551 "modules/filesystem/wrap_Filesystem.cpp"
  $18 = ((($17)) + 8|0); //@line 551 "modules/filesystem/wrap_Filesystem.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 551 "modules/filesystem/wrap_Filesystem.cpp"
  $20 = (FUNCTION_TABLE_ii[$19 & 8191]($16)|0); //@line 551 "modules/filesystem/wrap_Filesystem.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer>>2] = $20; //@line 551 "modules/filesystem/wrap_Filesystem.cpp"
  $21 = (invoke_iiii(3512,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 551 "modules/filesystem/wrap_Filesystem.cpp"
  $22 = __THREW__; __THREW__ = 0;
  $23 = $22&1;
  if (!($23)) {
   ___cxa_end_catch(); //@line 552 "modules/filesystem/wrap_Filesystem.cpp"
   $$0 = $21;
   STACKTOP = sp;return ($$0|0); //@line 562 "modules/filesystem/wrap_Filesystem.cpp"
  }
  $24 = ___cxa_find_matching_catch_2()|0;
  $25 = tempRet0;
  __THREW__ = 0;
  invoke_v(3378); //@line 552 "modules/filesystem/wrap_Filesystem.cpp"
  $26 = __THREW__; __THREW__ = 0;
  $27 = $26&1;
  if ($27) {
   $36 = ___cxa_find_matching_catch_3(0|0)|0;
   $37 = tempRet0;
   ___clang_call_terminate($36); //@line 552 "modules/filesystem/wrap_Filesystem.cpp"
   // unreachable; //@line 552 "modules/filesystem/wrap_Filesystem.cpp"
  } else {
   $$01 = $25;$$02 = $24;
   ___resumeException($$02|0);
   // unreachable;
  }
 } else {
  $9 = ($5|0)==(-1); //@line 555 "modules/filesystem/wrap_Filesystem.cpp"
  $10 = ($6|0)==(-1); //@line 555 "modules/filesystem/wrap_Filesystem.cpp"
  $11 = $9 & $10; //@line 555 "modules/filesystem/wrap_Filesystem.cpp"
  if ($11) {
   $28 = (__ZN4love10filesystem12luax_ioErrorEP9lua_StatePKcz($L,366354,$vararg_buffer1)|0); //@line 556 "modules/filesystem/wrap_Filesystem.cpp"
   $$0 = $28;
   STACKTOP = sp;return ($$0|0); //@line 562 "modules/filesystem/wrap_Filesystem.cpp"
  }
  $29 = ($6|0)>(2097151); //@line 557 "modules/filesystem/wrap_Filesystem.cpp"
  $30 = ($5>>>0)>(4294967295); //@line 557 "modules/filesystem/wrap_Filesystem.cpp"
  $31 = ($6|0)==(2097151); //@line 557 "modules/filesystem/wrap_Filesystem.cpp"
  $32 = $31 & $30; //@line 557 "modules/filesystem/wrap_Filesystem.cpp"
  $33 = $29 | $32; //@line 557 "modules/filesystem/wrap_Filesystem.cpp"
  if ($33) {
   $34 = (__ZN4love10filesystem12luax_ioErrorEP9lua_StatePKcz($L,328563,$vararg_buffer3)|0); //@line 558 "modules/filesystem/wrap_Filesystem.cpp"
   $$0 = $34;
   STACKTOP = sp;return ($$0|0); //@line 562 "modules/filesystem/wrap_Filesystem.cpp"
  } else {
   $35 = (+($5>>>0)) + (4294967296.0*(+($6|0))); //@line 560 "modules/filesystem/wrap_Filesystem.cpp"
   _lua_pushnumber($L,$35); //@line 560 "modules/filesystem/wrap_Filesystem.cpp"
   $$0 = 1;
   STACKTOP = sp;return ($$0|0); //@line 562 "modules/filesystem/wrap_Filesystem.cpp"
  }
 }
 return (0)|0;
}
function __ZN4love10filesystem20w_setSymlinksEnabledEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 566 "modules/filesystem/wrap_Filesystem.cpp"
 $2 = ((($1)) + 156|0); //@line 566 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 566 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,1)|0); //@line 566 "modules/filesystem/wrap_Filesystem.cpp"
 FUNCTION_TABLE_vii[$3 & 8191]($0,$4); //@line 566 "modules/filesystem/wrap_Filesystem.cpp"
 return 0; //@line 567 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem20w_areSymlinksEnabledEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 572 "modules/filesystem/wrap_Filesystem.cpp"
 $2 = ((($1)) + 160|0); //@line 572 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 572 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 572 "modules/filesystem/wrap_Filesystem.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 572 "modules/filesystem/wrap_Filesystem.cpp"
 return 1; //@line 573 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love10filesystem16w_getRequirePathEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$0$i = 0, $$01 = 0, $$01$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$03 = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $lpad$phi$index = 0, $lpad$phi$index7 = 0, $path = 0, $seperator$0$off04 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 160|0;
 $path = sp + 16|0;
 $1 = sp;
 $2 = ((($path)) + 64|0); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $3 = ((($path)) + 8|0); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$3>>2] = (9320); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $4 = ((($path)) + 12|0); //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$path>>2] = ((9524)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 HEAP32[$2>>2] = ((9544)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $5 = ((($path)) + 4|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 HEAP32[$5>>2] = 0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $6 = ((($path)) + 64|0); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 __THREW__ = 0;
 invoke_vii(3362,($6|0),($4|0)); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  $$0$i = $20;$$01$i = $19;
  __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($2); //@line 900 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  ___resumeException($$01$i|0);
  // unreachable;
 }
 $9 = ((($path)) + 136|0); //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$9>>2] = 0; //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 $10 = ((($path)) + 140|0); //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$10>>2] = -1; //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$path>>2] = (9300); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$2>>2] = (9340); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$3>>2] = (9320); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 __THREW__ = 0;
 invoke_vi(3363,($4|0)); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 do {
  if ($12) {
   $21 = ___cxa_find_matching_catch_2()|0;
   $22 = tempRet0;
   $eh$lpad$body$i$index2Z2D = $22;$eh$lpad$body$i$indexZ2D = $21;
  } else {
   HEAP32[$4>>2] = (9356); //@line 243 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $13 = ((($path)) + 44|0); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $14 = ((($path)) + 60|0); //@line 242 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   ;HEAP32[$13>>2]=0|0;HEAP32[$13+4>>2]=0|0;HEAP32[$13+8>>2]=0|0;HEAP32[$13+12>>2]=0|0; //@line 241 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   HEAP32[$14>>2] = 24; //@line 242 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   invoke_vii(3364,($4|0),($0|0)); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if ($16) {
    $17 = ___cxa_find_matching_catch_2()|0;
    $18 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13); //@line 245 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($4); //@line 245 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $eh$lpad$body$i$index2Z2D = $18;$eh$lpad$body$i$indexZ2D = $17;
    break;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $23 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
   $24 = HEAP32[$23>>2]|0; //@line 580 "modules/filesystem/wrap_Filesystem.cpp"
   $25 = ((($24)) + 164|0); //@line 580 "modules/filesystem/wrap_Filesystem.cpp"
   $26 = HEAP32[$25>>2]|0; //@line 580 "modules/filesystem/wrap_Filesystem.cpp"
   __THREW__ = 0;
   $27 = (invoke_ii($26|0,($23|0))|0); //@line 580 "modules/filesystem/wrap_Filesystem.cpp"
   $28 = __THREW__; __THREW__ = 0;
   $29 = $28&1;
   L11: do {
    if ($29) {
     label = 14;
    } else {
     $30 = HEAP32[$27>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $31 = ((($27)) + 4|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $32 = HEAP32[$31>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $33 = ($30|0)==($32|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     L13: do {
      if (!($33)) {
       $34 = ((($path)) + 8|0); //@line 583 "modules/filesystem/wrap_Filesystem.cpp"
       $__begin$sroa$0$03 = $30;$seperator$0$off04 = 0;
       while(1) {
        if ($seperator$0$off04) {
         __THREW__ = 0;
         (invoke_iiii(3366,($34|0),(442487|0),1)|0); //@line 894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
         $41 = __THREW__; __THREW__ = 0;
         $42 = $41&1;
         if ($42) {
          break;
         }
        }
        $43 = HEAP8[$__begin$sroa$0$03>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $44 = $43 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $45 = ($44<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $46 = ((($__begin$sroa$0$03)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $47 = HEAP32[$46>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $48 = ((($__begin$sroa$0$03)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $49 = $45 ? $48 : $47; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $50 = ((($__begin$sroa$0$03)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $51 = HEAP32[$50>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $52 = $43&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $53 = $52 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $54 = $45 ? $53 : $51; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        __THREW__ = 0;
        (invoke_iiii(3366,($34|0),($49|0),($54|0))|0); //@line 1080 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
        $55 = __THREW__; __THREW__ = 0;
        $56 = $55&1;
        if ($56) {
         break;
        }
        $57 = ((($__begin$sroa$0$03)) + 12|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        $58 = ($57|0)==($32|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        if ($58) {
         break L13;
        } else {
         $__begin$sroa$0$03 = $57;$seperator$0$off04 = 1;
        }
       }
       $37 = ___cxa_find_matching_catch_2()|0;
       $38 = tempRet0;
       $lpad$phi$index = $37;$lpad$phi$index7 = $38;
       break L11;
      }
     } while(0);
     __THREW__ = 0;
     invoke_vii(3367,($1|0),($4|0)); //@line 964 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     $35 = __THREW__; __THREW__ = 0;
     $36 = $35&1;
     if ($36) {
      label = 14;
     } else {
      __THREW__ = 0;
      invoke_vii(3405,($L|0),($1|0)); //@line 590 "modules/filesystem/wrap_Filesystem.cpp"
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if (!($60)) {
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 590 "modules/filesystem/wrap_Filesystem.cpp"
       HEAP32[$path>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
       HEAP32[$2>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
       $61 = ((($path)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
       HEAP32[$61>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
       HEAP32[$4>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
       __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($4); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
       __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($2); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
       STACKTOP = sp;return 1; //@line 592 "modules/filesystem/wrap_Filesystem.cpp"
      }
      $62 = ___cxa_find_matching_catch_2()|0;
      $63 = tempRet0;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 590 "modules/filesystem/wrap_Filesystem.cpp"
      $$0 = $63;$$01 = $62;
      HEAP32[$path>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      HEAP32[$2>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      $64 = ((($path)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      HEAP32[$64>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      HEAP32[$4>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($4); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($2); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      ___resumeException($$01|0);
      // unreachable;
     }
    }
   } while(0);
   if ((label|0) == 14) {
    $39 = ___cxa_find_matching_catch_2()|0;
    $40 = tempRet0;
    $lpad$phi$index = $39;$lpad$phi$index7 = $40;
   }
   $$0 = $lpad$phi$index7;$$01 = $lpad$phi$index;
   HEAP32[$path>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$2>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   $64 = ((($path)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$64>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$4>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($4); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($2); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   ___resumeException($$01|0);
   // unreachable;
  }
 } while(0);
 $$0$i = $eh$lpad$body$i$index2Z2D;$$01$i = $eh$lpad$body$i$indexZ2D;
 __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($2); //@line 900 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 ___resumeException($$01$i|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love10filesystem16w_setRequirePathEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$0$i = 0, $$01 = 0, $$01$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $eh$lpad$body$index7Z2D = 0, $eh$lpad$body$indexZ2D = 0, $element = 0, $lpad$phi$index = 0, $lpad$phi$index12 = 0, $path = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 156|0;
 $element = sp + 144|0;
 $path = sp;
 __ZN4love16luax_checkstringEP9lua_Statei($element,$L,1); //@line 596 "modules/filesystem/wrap_Filesystem.cpp"
 $1 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 597 "modules/filesystem/wrap_Filesystem.cpp"
 $3 = ((($2)) + 164|0); //@line 597 "modules/filesystem/wrap_Filesystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 597 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 $5 = (invoke_ii($4|0,($1|0))|0); //@line 597 "modules/filesystem/wrap_Filesystem.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $73 = ___cxa_find_matching_catch_2()|0;
  $74 = tempRet0;
  $eh$lpad$body$index7Z2D = $74;$eh$lpad$body$indexZ2D = $73;
 } else {
  $8 = HEAP32[$5>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $9 = ((($5)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $10 = HEAP32[$9>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $11 = ($10|0)==($8|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($11)) {
   $13 = $10;
   while(1) {
    $12 = ((($13)) + -12|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$9>>2] = $12; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $14 = HEAP32[$9>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $15 = ($14|0)==($8|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($15) {
     break;
    } else {
     $13 = $14;
    }
   }
  }
  $16 = ((($path)) + 64|0); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $17 = ((($path)) + 8|0); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  HEAP32[$17>>2] = (9320); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $18 = ((($path)) + 12|0); //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  HEAP32[$path>>2] = ((9524)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
  HEAP32[$16>>2] = ((9544)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
  $19 = ((($path)) + 4|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
  HEAP32[$19>>2] = 0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
  $20 = ((($path)) + 64|0); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
  __THREW__ = 0;
  invoke_vii(3362,($20|0),($18|0)); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
  $21 = __THREW__; __THREW__ = 0;
  $22 = $21&1;
  if ($22) {
   $33 = ___cxa_find_matching_catch_2()|0;
   $34 = tempRet0;
   $$0$i = $34;$$01$i = $33;
  } else {
   $23 = ((($path)) + 136|0); //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
   HEAP32[$23>>2] = 0; //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
   $24 = ((($path)) + 140|0); //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
   HEAP32[$24>>2] = -1; //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
   HEAP32[$path>>2] = (9300); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   HEAP32[$16>>2] = (9340); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   HEAP32[$17>>2] = (9320); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   __THREW__ = 0;
   invoke_vi(3363,($18|0)); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   do {
    if ($26) {
     $35 = ___cxa_find_matching_catch_2()|0;
     $36 = tempRet0;
     $eh$lpad$body$i$index2Z2D = $36;$eh$lpad$body$i$indexZ2D = $35;
    } else {
     HEAP32[$18>>2] = (9356); //@line 243 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     $27 = ((($path)) + 44|0); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     $28 = ((($path)) + 60|0); //@line 242 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     ;HEAP32[$27>>2]=0|0;HEAP32[$27+4>>2]=0|0;HEAP32[$27+8>>2]=0|0;HEAP32[$27+12>>2]=0|0; //@line 241 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     HEAP32[$28>>2] = 24; //@line 242 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     __THREW__ = 0;
     invoke_vii(3364,($18|0),($0|0)); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     $29 = __THREW__; __THREW__ = 0;
     $30 = $29&1;
     if ($30) {
      $31 = ___cxa_find_matching_catch_2()|0;
      $32 = tempRet0;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($27); //@line 245 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
      __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($18); //@line 245 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
      $eh$lpad$body$i$index2Z2D = $32;$eh$lpad$body$i$indexZ2D = $31;
      break;
     }
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     $37 = ((($path)) + 8|0); //@line 601 "modules/filesystem/wrap_Filesystem.cpp"
     $38 = HEAP8[$element>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $39 = $38 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $40 = ($39<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $41 = ((($element)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $42 = HEAP32[$41>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $43 = ((($element)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $44 = $40 ? $43 : $42; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $45 = ((($element)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $46 = HEAP32[$45>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $47 = $38&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $48 = $47 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $49 = $40 ? $48 : $46; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     __THREW__ = 0;
     (invoke_iiii(3366,($37|0),($44|0),($49|0))|0); //@line 1080 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
     $50 = __THREW__; __THREW__ = 0;
     $51 = $50&1;
     do {
      if ($51) {
       $77 = ___cxa_find_matching_catch_2()|0;
       $78 = tempRet0;
       $lpad$phi$index = $77;$lpad$phi$index12 = $78;
      } else {
       $52 = ((($5)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       while(1) {
        __THREW__ = 0;
        $53 = (invoke_iiii(3516,($path|0),($element|0),59)|0); //@line 603 "modules/filesystem/wrap_Filesystem.cpp"
        $54 = __THREW__; __THREW__ = 0;
        $55 = $54&1;
        if ($55) {
         label = 22;
         break;
        }
        $56 = HEAP32[$53>>2]|0; //@line 603 "modules/filesystem/wrap_Filesystem.cpp"
        $57 = ((($56)) + -12|0); //@line 603 "modules/filesystem/wrap_Filesystem.cpp"
        $58 = HEAP32[$57>>2]|0; //@line 603 "modules/filesystem/wrap_Filesystem.cpp"
        $59 = (($53) + ($58)|0); //@line 603 "modules/filesystem/wrap_Filesystem.cpp"
        $60 = ((($59)) + 16|0); //@line 545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
        $61 = HEAP32[$60>>2]|0; //@line 545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
        $62 = $61 & 5; //@line 545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
        $63 = ($62|0)==(0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
        if (!($63)) {
         label = 25;
         break;
        }
        $64 = HEAP32[$9>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $65 = HEAP32[$52>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $66 = ($64|0)==($65|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($66) {
         __THREW__ = 0;
         invoke_vii(3517,($5|0),($element|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $71 = __THREW__; __THREW__ = 0;
         $72 = $71&1;
         if ($72) {
          label = 22;
          break;
         } else {
          continue;
         }
        }
        __THREW__ = 0;
        invoke_vii(3351,($64|0),($element|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $67 = __THREW__; __THREW__ = 0;
        $68 = $67&1;
        if ($68) {
         label = 22;
         break;
        }
        $69 = HEAP32[$9>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $70 = ((($69)) + 12|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$9>>2] = $70; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       if ((label|0) == 22) {
        $75 = ___cxa_find_matching_catch_2()|0;
        $76 = tempRet0;
        $lpad$phi$index = $75;$lpad$phi$index12 = $76;
        break;
       }
       else if ((label|0) == 25) {
        HEAP32[$path>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$16>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$37>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$18>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($27); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($18); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($16); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($element); //@line 607 "modules/filesystem/wrap_Filesystem.cpp"
        STACKTOP = sp;return 0; //@line 607 "modules/filesystem/wrap_Filesystem.cpp"
       }
      }
     } while(0);
     HEAP32[$path>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     HEAP32[$16>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     HEAP32[$37>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     HEAP32[$18>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($27); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($18); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($16); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     $$0 = $lpad$phi$index12;$$01 = $lpad$phi$index;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($element); //@line 607 "modules/filesystem/wrap_Filesystem.cpp"
     ___resumeException($$01|0);
     // unreachable;
    }
   } while(0);
   $$0$i = $eh$lpad$body$i$index2Z2D;$$01$i = $eh$lpad$body$i$indexZ2D;
  }
  __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($16); //@line 900 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  $eh$lpad$body$index7Z2D = $$0$i;$eh$lpad$body$indexZ2D = $$01$i;
 }
 $$0 = $eh$lpad$body$index7Z2D;$$01 = $eh$lpad$body$indexZ2D;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($element); //@line 607 "modules/filesystem/wrap_Filesystem.cpp"
 ___resumeException($$01|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__17getlineIcNS_11char_traitsIcEENS_9allocatorIcEEEERNS_13basic_istreamIT_T0_EES9_RNS_12basic_stringIS6_S7_T1_EES6_($__is,$__str,$__dlm) {
 $__is = $__is|0;
 $__str = $__str|0;
 $__dlm = $__dlm|0;
 var $$0$i1 = 0, $$__err$2 = 0, $$pre = 0, $$pre$phi10Z2D = 0, $$pre$phi12Z2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $__err$2$ph = 0, $__extr$0 = 0, $__extr$1$ph = 0, $__sen = 0, $lpad$phi$index = 0;
 var $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__sen = sp;
 __THREW__ = 0;
 invoke_viii(3518,($__sen|0),($__is|0),1); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 L1: do {
  if ($1) {
   label = 12;
  } else {
   $2 = HEAP8[$__sen>>0]|0; //@line 276 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
   $3 = ($2<<24>>24)==(0); //@line 276 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
   do {
    if (!($3)) {
     $4 = HEAP8[$__str>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $5 = $4 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $6 = ($5<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($6) {
      $10 = ((($__str)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$10>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$__str>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $$pre = ((($__str)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $$pre$phiZ2D = $$pre;
     } else {
      $7 = ((($__str)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $8 = HEAP32[$7>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$8>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $9 = ((($__str)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP32[$9>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $$pre$phiZ2D = $9;
     }
     $__extr$0 = 0;
     while(1) {
      $11 = HEAP32[$__is>>2]|0; //@line 1615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
      $12 = ((($11)) + -12|0); //@line 1615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
      $13 = HEAP32[$12>>2]|0; //@line 1615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
      $14 = (($__is) + ($13)|0); //@line 1615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
      $15 = ((($14)) + 24|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
      $16 = HEAP32[$15>>2]|0; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
      $17 = ((($16)) + 12|0); //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
      $18 = HEAP32[$17>>2]|0; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
      $19 = ((($16)) + 16|0); //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
      $20 = HEAP32[$19>>2]|0; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
      $21 = ($18|0)==($20|0); //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
      if ($21) {
       $22 = HEAP32[$16>>2]|0; //@line 302 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
       $23 = ((($22)) + 40|0); //@line 302 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
       $24 = HEAP32[$23>>2]|0; //@line 302 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
       __THREW__ = 0;
       $25 = (invoke_ii($24|0,($16|0))|0); //@line 302 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
       $26 = __THREW__; __THREW__ = 0;
       $27 = $26&1;
       if ($27) {
        label = 11;
        break;
       }
       $31 = ($25|0)==(-1); //@line 660 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($31) {
        $__err$2$ph = 2;$__extr$1$ph = $__extr$0;
        label = 19;
        break;
       } else {
        $$0$i1 = $25;
       }
      } else {
       $28 = ((($18)) + 1|0); //@line 303 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
       HEAP32[$17>>2] = $28; //@line 303 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
       $29 = HEAP8[$18>>0]|0; //@line 303 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/streambuf"
       $30 = $29&255; //@line 658 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$0$i1 = $30;
      }
      $42 = (($__extr$0) + 1)|0; //@line 1621 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
      $43 = $$0$i1&255; //@line 656 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $44 = ($43<<24>>24)==($__dlm<<24>>24); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($44) {
       $__err$2$ph = 0;$__extr$1$ph = $42;
       label = 19;
       break;
      }
      __THREW__ = 0;
      invoke_vii(3370,($__str|0),($43|0)); //@line 1625 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
      $45 = __THREW__; __THREW__ = 0;
      $46 = $45&1;
      if ($46) {
       label = 11;
       break;
      }
      $47 = HEAP8[$__str>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $48 = $47 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $49 = ($48<<24>>24)!=(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $50 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $51 = ($50|0)==(-17); //@line 1626 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
      $or$cond = $49 & $51; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($or$cond) {
       $__err$2$ph = 4;$__extr$1$ph = $42;
       label = 19;
       break;
      } else {
       $__extr$0 = $42;
      }
     }
     if ((label|0) == 11) {
      $32 = ___cxa_find_matching_catch_3(0|0)|0;
      $33 = tempRet0;
      $$pre$phi10Z2D = $__is;$$pre$phi12Z2D = $__is;$lpad$phi$index = $32;
      break L1;
     }
     else if ((label|0) == 19) {
      $52 = ($__extr$1$ph|0)==(0); //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
      $53 = $__err$2$ph | 4; //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
      $$__err$2 = $52 ? $53 : $__err$2$ph; //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
      $54 = HEAP32[$__is>>2]|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
      $55 = ((($54)) + -12|0); //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
      $56 = HEAP32[$55>>2]|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
      $57 = (($__is) + ($56)|0); //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
      $58 = ((($57)) + 16|0); //@line 524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
      $59 = HEAP32[$58>>2]|0; //@line 524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
      $60 = $59 | $$__err$2; //@line 524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
      __THREW__ = 0;
      invoke_vii(3376,($57|0),($60|0)); //@line 524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
      $61 = __THREW__; __THREW__ = 0;
      $62 = $61&1;
      if ($62) {
       label = 12;
       break L1;
      } else {
       break;
      }
     }
    }
   } while(0);
   STACKTOP = sp;return ($__is|0); //@line 1643 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
  }
 } while(0);
 if ((label|0) == 12) {
  $34 = ___cxa_find_matching_catch_3(0|0)|0;
  $35 = tempRet0;
  $$pre$phi10Z2D = $__is;$$pre$phi12Z2D = $__is;$lpad$phi$index = $34;
 }
 (___cxa_begin_catch(($lpad$phi$index|0))|0); //@line 1637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $36 = HEAP32[$$pre$phi10Z2D>>2]|0; //@line 1640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $37 = ((($36)) + -12|0); //@line 1640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $38 = HEAP32[$37>>2]|0; //@line 1640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $39 = (($$pre$phi12Z2D) + ($38)|0); //@line 1640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 __THREW__ = 0;
 invoke_vi(3377,($39|0)); //@line 1640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $40 = __THREW__; __THREW__ = 0;
 $41 = $40&1;
 if (!($41)) {
  ___cxa_end_catch(); //@line 1641 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
  STACKTOP = sp;return ($__is|0); //@line 1643 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 }
 $63 = ___cxa_find_matching_catch_2()|0;
 $64 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 1641 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $65 = __THREW__; __THREW__ = 0;
 $66 = $65&1;
 if ($66) {
  $67 = ___cxa_find_matching_catch_3(0|0)|0;
  $68 = tempRet0;
  ___clang_call_terminate($67); //@line 1641 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
  // unreachable; //@line 1641 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 } else {
  ___resumeException($63|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZNSt3__16vectorINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEENS4_IS6_EEE21__push_back_slow_pathIRKS6_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i4 = 0, $$0$i5 = 0, $$01$i$i = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__v$sroa$6$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($3|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(357913941); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $10 = $$pre;
 } else {
  $10 = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($10))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = (($9|0) / 12)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($11>>>0)<(178956970); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $16 = $11 << 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = ($16>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $18 = $17 ? $5 : $16; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = (($19) - ($10))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = (($20|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = ($18|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($22) {
   $$0$i5 = 0;$26 = 0;$27 = $21;
  } else {
   $$0$i4 = $18;$59 = $21;
   label = 6;
  }
 } else {
  $13 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) - ($10))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = (($14|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i4 = 357913941;$59 = $15;
  label = 6;
 }
 if ((label|0) == 6) {
  $23 = ($$0$i4*12)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = (__Znwj($23)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i5 = $$0$i4;$26 = $24;$27 = $59;
 }
 $25 = (($26) + (($27*12)|0)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $25; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = (($26) + (($$0$i5*12)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = $29; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 __THREW__ = 0;
 invoke_vii(3351,($25|0),($__x|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $31 = __THREW__; __THREW__ = 0;
 $32 = $31&1;
 if ($32) {
  $56 = ___cxa_find_matching_catch_2()|0;
  $57 = tempRet0;
  $58 = ($26|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($58) {
   ___resumeException($56|0);
   // unreachable;
  }
  __ZdlPv($26); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($56|0);
  // unreachable;
 }
 $33 = ((($25)) + 12|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $34 = $33; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $35 = HEAP32[$this>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $36 = HEAP32[$0>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $37 = ($36|0)==($35|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($37) {
  $38 = $35; //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $46 = $28;$48 = $38;$50 = $35;
 } else {
  $$01$i$i = $36;$40 = $25;$__v$sroa$6$0 = $28;
  while(1) {
   $39 = ((($40)) + -12|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $41 = ((($$01$i$i)) + -12|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   ;HEAP32[$39>>2]=HEAP32[$41>>2]|0;HEAP32[$39+4>>2]=HEAP32[$41+4>>2]|0;HEAP32[$39+8>>2]=HEAP32[$41+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   ;HEAP32[$41>>2]=0|0;HEAP32[$41+4>>2]=0|0;HEAP32[$41+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $42 = $__v$sroa$6$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $43 = ((($42)) + -12|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $44 = $43; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $45 = ($41|0)==($35|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if ($45) {
    $$lcssa = $44;
    break;
   } else {
    $$01$i$i = $41;$40 = $43;$__v$sroa$6$0 = $44;
   }
  }
  $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $$pre6 = HEAP32[$0>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $46 = $$lcssa;$48 = $$pre$i;$50 = $$pre6;
 }
 HEAP32[$this>>2] = $46; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $34; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $47 = $48; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $49 = ($50|0)==($47|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if (!($49)) {
  $52 = $50;
  while(1) {
   $51 = ((($52)) + -12|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($51); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $53 = ($51|0)==($47|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($53) {
    break;
   } else {
    $52 = $51;
   }
  }
 }
 $54 = ($48|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($54) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $55 = $48; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($55); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love10filesystem6loaderEP9lua_State($L) {
 $L = $L|0;
 var $$014 = 0, $$03 = 0, $$06 = 0, $$07 = 0, $$1 = 0, $$3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $__begin$sroa$0$015 = 0, $__begin2$sroa$0$013 = 0, $element = 0, $errstr = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $modulename = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $modulename = sp + 28|0;
 $element = sp + 16|0;
 $errstr = sp + 4|0;
 __ZN4love13luax_tostringEP9lua_Statei($modulename,$L,1); //@line 611 "modules/filesystem/wrap_Filesystem.cpp"
 $0 = HEAP8[$modulename>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $1 = $0 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = ($1<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ((($modulename)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = HEAP32[$3>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($modulename)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = $2 ? $5 : $4; //@line 1755 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($modulename)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = HEAP32[$7>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $0&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $9 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $2 ? $10 : $8; //@line 1755 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = (($6) + ($11)|0); //@line 1399 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ($11|0)==(0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if (!($13)) {
  $__begin$sroa$0$015 = $6;
  while(1) {
   $21 = HEAP8[$__begin$sroa$0$015>>0]|0; //@line 615 "modules/filesystem/wrap_Filesystem.cpp"
   $22 = ($21<<24>>24)==(46); //@line 615 "modules/filesystem/wrap_Filesystem.cpp"
   if ($22) {
    HEAP8[$__begin$sroa$0$015>>0] = 47; //@line 616 "modules/filesystem/wrap_Filesystem.cpp"
   }
   $23 = ((($__begin$sroa$0$015)) + 1|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $24 = ($23|0)==($12|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($24) {
    break;
   } else {
    $__begin$sroa$0$015 = $23;
   }
  }
 }
 $14 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $15 = HEAP32[$14>>2]|0; //@line 620 "modules/filesystem/wrap_Filesystem.cpp"
 $16 = ((($15)) + 164|0); //@line 620 "modules/filesystem/wrap_Filesystem.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 620 "modules/filesystem/wrap_Filesystem.cpp"
 __THREW__ = 0;
 $18 = (invoke_ii($17|0,($14|0))|0); //@line 620 "modules/filesystem/wrap_Filesystem.cpp"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 L8: do {
  if ($20) {
   label = 10;
  } else {
   $25 = HEAP32[$18>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $26 = ((($18)) + 4|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $27 = HEAP32[$26>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $28 = ($25|0)==($27|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   L10: do {
    if (!($28)) {
     $29 = ((($element)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $30 = ((($element)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $$014 = 0;$__begin2$sroa$0$013 = $25;
     while(1) {
      __THREW__ = 0;
      invoke_vii(3351,($element|0),($__begin2$sroa$0$013|0)); //@line 620 "modules/filesystem/wrap_Filesystem.cpp"
      $37 = __THREW__; __THREW__ = 0;
      $38 = $37&1;
      if ($38) {
       label = 9;
       break;
      }
      $39 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE4findEcj($element,63,0)|0); //@line 622 "modules/filesystem/wrap_Filesystem.cpp"
      $40 = ($39|0)==(-1); //@line 623 "modules/filesystem/wrap_Filesystem.cpp"
      if (!($40)) {
       $41 = HEAP8[$modulename>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $42 = $41 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $43 = ($42<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $44 = HEAP32[$3>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $45 = $43 ? $5 : $44; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $46 = HEAP32[$7>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $47 = $41&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $48 = $47 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $49 = $43 ? $48 : $46; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       __THREW__ = 0;
       (invoke_iiiiii(3519,($element|0),($39|0),1,($45|0),($49|0))|0); //@line 2997 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $50 = __THREW__; __THREW__ = 0;
       $51 = $50&1;
       if ($51) {
        label = 15;
        break;
       }
      }
      $54 = HEAP32[$14>>2]|0; //@line 626 "modules/filesystem/wrap_Filesystem.cpp"
      $55 = ((($54)) + 116|0); //@line 626 "modules/filesystem/wrap_Filesystem.cpp"
      $56 = HEAP32[$55>>2]|0; //@line 626 "modules/filesystem/wrap_Filesystem.cpp"
      $57 = HEAP8[$element>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $58 = $57 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $59 = ($58<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $60 = HEAP32[$29>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $61 = $59 ? $30 : $60; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __THREW__ = 0;
      $62 = (invoke_iii($56|0,($14|0),($61|0))|0); //@line 626 "modules/filesystem/wrap_Filesystem.cpp"
      $63 = __THREW__; __THREW__ = 0;
      $64 = $63&1;
      if ($64) {
       label = 15;
       break;
      }
      if ($62) {
       __THREW__ = 0;
       invoke_vii(3384,($L|0),-2); //@line 628 "modules/filesystem/wrap_Filesystem.cpp"
       $65 = __THREW__; __THREW__ = 0;
       $66 = $65&1;
       if ($66) {
        label = 15;
        break;
       }
       $67 = HEAP8[$element>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $68 = $67 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $69 = ($68<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $70 = HEAP32[$29>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $71 = $69 ? $30 : $70; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       __THREW__ = 0;
       invoke_vii(3380,($L|0),($71|0)); //@line 629 "modules/filesystem/wrap_Filesystem.cpp"
       $72 = __THREW__; __THREW__ = 0;
       $73 = $72&1;
       if ($73) {
        label = 15;
        break;
       }
       __THREW__ = 0;
       $74 = (invoke_ii(445,($L|0))|0); //@line 630 "modules/filesystem/wrap_Filesystem.cpp"
       $75 = __THREW__; __THREW__ = 0;
       $76 = $75&1;
       if ($76) {
        label = 15;
        break;
       } else {
        $$03 = 0;$$1 = $74;
       }
      } else {
       $$03 = 1;$$1 = $$014;
      }
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($element); //@line 632 "modules/filesystem/wrap_Filesystem.cpp"
      $32 = ((($__begin2$sroa$0$013)) + 12|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if (!($$03)) {
       $$3 = $$1;
       label = 26;
       break;
      }
      $31 = ($32|0)==($27|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($31) {
       break L10;
      } else {
       $$014 = $$1;$__begin2$sroa$0$013 = $32;
      }
     }
     if ((label|0) == 9) {
      $33 = ___cxa_find_matching_catch_2()|0;
      $34 = tempRet0;
      $lpad$phi$index = $33;$lpad$phi$index2 = $34;
      break L8;
     }
     else if ((label|0) == 15) {
      $52 = ___cxa_find_matching_catch_2()|0;
      $53 = tempRet0;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($element); //@line 632 "modules/filesystem/wrap_Filesystem.cpp"
      $$06 = $52;$$07 = $53;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($modulename); //@line 638 "modules/filesystem/wrap_Filesystem.cpp"
      ___resumeException($$06|0);
      // unreachable;
     }
     else if ((label|0) == 26) {
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($modulename); //@line 638 "modules/filesystem/wrap_Filesystem.cpp"
      STACKTOP = sp;return ($$3|0); //@line 638 "modules/filesystem/wrap_Filesystem.cpp"
     }
    }
   } while(0);
   __THREW__ = 0;
   invoke_viii(3460,($errstr|0),(328604|0),35); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $77 = __THREW__; __THREW__ = 0;
   $78 = $77&1;
   if ($78) {
    label = 10;
   } else {
    $79 = HEAP8[$errstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $80 = $79 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $81 = ($80<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $82 = ((($errstr)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $83 = HEAP32[$82>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $84 = ((($errstr)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $85 = $81 ? $84 : $83; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $86 = HEAP8[$modulename>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $87 = $86 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $88 = ($87<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $89 = HEAP32[$3>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $90 = $88 ? $5 : $89; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    __THREW__ = 0;
    HEAP32[$vararg_buffer>>2] = $90; //@line 636 "modules/filesystem/wrap_Filesystem.cpp"
    (invoke_iiii(3520,($L|0),($85|0),($vararg_buffer|0))|0); //@line 636 "modules/filesystem/wrap_Filesystem.cpp"
    $91 = __THREW__; __THREW__ = 0;
    $92 = $91&1;
    if ($92) {
     $93 = ___cxa_find_matching_catch_2()|0;
     $94 = tempRet0;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($errstr); //@line 638 "modules/filesystem/wrap_Filesystem.cpp"
     $$06 = $93;$$07 = $94;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($modulename); //@line 638 "modules/filesystem/wrap_Filesystem.cpp"
     ___resumeException($$06|0);
     // unreachable;
    } else {
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($errstr); //@line 638 "modules/filesystem/wrap_Filesystem.cpp"
     $$3 = 1;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($modulename); //@line 638 "modules/filesystem/wrap_Filesystem.cpp"
     STACKTOP = sp;return ($$3|0); //@line 638 "modules/filesystem/wrap_Filesystem.cpp"
    }
   }
  }
 } while(0);
 if ((label|0) == 10) {
  $35 = ___cxa_find_matching_catch_2()|0;
  $36 = tempRet0;
  $lpad$phi$index = $35;$lpad$phi$index2 = $36;
 }
 $$06 = $lpad$phi$index;$$07 = $lpad$phi$index2;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($modulename); //@line 638 "modules/filesystem/wrap_Filesystem.cpp"
 ___resumeException($$06|0);
 // unreachable;
 return (0)|0;
}
function _luaopen_love_filesystem($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0;
 var $eh$lpad$body$i$indexZ2D = 0, $instance$sroa$0$1 = 0, $vararg_buffer = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $w = sp + 4|0;
 $0 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
 $1 = $0; //@line 766 "modules/filesystem/wrap_Filesystem.cpp"
 $2 = ($0|0)==(0|0); //@line 767 "modules/filesystem/wrap_Filesystem.cpp"
 L1: do {
  if ($2) {
   __THREW__ = 0;
   $3 = (invoke_ii(3348,112)|0); //@line 769 "modules/filesystem/wrap_Filesystem.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   do {
    if ($5) {
     $11 = ___cxa_find_matching_catch_3(8992|0)|0;
     $12 = tempRet0;
     $eh$lpad$body$i$index2Z2D = $12;$eh$lpad$body$i$indexZ2D = $11;
    } else {
     __THREW__ = 0;
     invoke_vi(3521,($3|0)); //@line 769 "modules/filesystem/wrap_Filesystem.cpp"
     $6 = __THREW__; __THREW__ = 0;
     $7 = $6&1;
     if ($7) {
      $8 = ___cxa_find_matching_catch_3(8992|0)|0;
      $9 = tempRet0;
      __ZdlPv($3); //@line 769 "modules/filesystem/wrap_Filesystem.cpp"
      $eh$lpad$body$i$index2Z2D = $9;$eh$lpad$body$i$indexZ2D = $8;
      break;
     } else {
      $10 = $3; //@line 769 "modules/filesystem/wrap_Filesystem.cpp"
      $instance$sroa$0$1 = $10;
      break L1;
     }
    }
   } while(0);
   $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $14 = ($eh$lpad$body$i$index2Z2D|0)==($13|0); //@line 535 "./common/runtime.h"
   if (!($14)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $15 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
   $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
   $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
   $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $instance$sroa$0$1 = $1;
    break;
   }
   $23 = ___cxa_find_matching_catch_2()|0;
   $24 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    $27 = ___cxa_find_matching_catch_3(0|0)|0;
    $28 = tempRet0;
    ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $24;$$02$i = $23;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $29 = HEAP32[$0>>2]|0; //@line 772 "modules/filesystem/wrap_Filesystem.cpp"
   $30 = ((($29)) + 8|0); //@line 772 "modules/filesystem/wrap_Filesystem.cpp"
   $31 = HEAP32[$30>>2]|0; //@line 772 "modules/filesystem/wrap_Filesystem.cpp"
   FUNCTION_TABLE_vi[$31 & 8191]($0); //@line 772 "modules/filesystem/wrap_Filesystem.cpp"
   $instance$sroa$0$1 = $1;
  }
 } while(0);
 (__ZN4love22luax_register_searcherEP9lua_StatePFiS1_Ei($L,3522,2)|0); //@line 775 "modules/filesystem/wrap_Filesystem.cpp"
 HEAP32[$w>>2] = $instance$sroa$0$1; //@line 783 "modules/filesystem/wrap_Filesystem.cpp"
 $32 = ((($w)) + 4|0); //@line 784 "modules/filesystem/wrap_Filesystem.cpp"
 HEAP32[$32>>2] = 354839; //@line 784 "modules/filesystem/wrap_Filesystem.cpp"
 $33 = ((($w)) + 8|0); //@line 785 "modules/filesystem/wrap_Filesystem.cpp"
 HEAP32[$33>>2] = 56; //@line 785 "modules/filesystem/wrap_Filesystem.cpp"
 $34 = ((($w)) + 12|0); //@line 786 "modules/filesystem/wrap_Filesystem.cpp"
 HEAP32[$34>>2] = 12872; //@line 786 "modules/filesystem/wrap_Filesystem.cpp"
 $35 = ((($w)) + 16|0); //@line 787 "modules/filesystem/wrap_Filesystem.cpp"
 HEAP32[$35>>2] = 13176; //@line 787 "modules/filesystem/wrap_Filesystem.cpp"
 $36 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 789 "modules/filesystem/wrap_Filesystem.cpp"
 STACKTOP = sp;return ($36|0); //@line 789 "modules/filesystem/wrap_Filesystem.cpp"
}
function __ZN4love4font15w_newRasterizerEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $t$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 38 "modules/font/wrap_Font.cpp"
 $0 = (_lua_type($L,1)|0); //@line 40 "modules/font/wrap_Font.cpp"
 $1 = ($0|0)==(3); //@line 40 "modules/font/wrap_Font.cpp"
 if (!($1)) {
  $2 = (_lua_type($L,2)|0); //@line 40 "modules/font/wrap_Font.cpp"
  $3 = ($2|0)==(3); //@line 40 "modules/font/wrap_Font.cpp"
  if (!($3)) {
   $4 = (_lua_type($L,1)|0); //@line 40 "modules/font/wrap_Font.cpp"
   $5 = ($4|0)==(-1); //@line 40 "modules/font/wrap_Font.cpp"
   if (!($5)) {
    $7 = (_lua_type($L,2)|0); //@line 45 "modules/font/wrap_Font.cpp"
    $8 = ($7|0)<(1); //@line 45 "modules/font/wrap_Font.cpp"
    if (!($8)) {
     (__ZN4love4font21w_newBMFontRasterizerEP9lua_State($L)|0); //@line 63 "modules/font/wrap_Font.cpp"
     $$0 = 1;
     STACKTOP = sp;return ($$0|0); //@line 65 "modules/font/wrap_Font.cpp"
    }
    $9 = (__ZN4love10filesystem16luax_getfiledataEP9lua_Statei($L,1)|0); //@line 49 "modules/font/wrap_Font.cpp"
    $10 = HEAP32[(625620)>>2]|0; //@line 98 "./common/Module.h"
    $11 = HEAP32[$10>>2]|0; //@line 52 "modules/font/wrap_Font.cpp"
    $12 = ((($11)) + 24|0); //@line 52 "modules/font/wrap_Font.cpp"
    $13 = HEAP32[$12>>2]|0; //@line 52 "modules/font/wrap_Font.cpp"
    __THREW__ = 0;
    $14 = (invoke_iii($13|0,($10|0),($9|0))|0); //@line 52 "modules/font/wrap_Font.cpp"
    $15 = __THREW__; __THREW__ = 0;
    $16 = $15&1;
    do {
     if ($16) {
      $17 = ___cxa_find_matching_catch_3(8992|0)|0;
      $18 = tempRet0;
      $19 = (_llvm_eh_typeid_for((8992|0))|0); //@line 556 "./common/runtime.h"
      $20 = ($18|0)==($19|0); //@line 556 "./common/runtime.h"
      if (!($20)) {
       $$01$i = $18;$$02$i = $17;
       ___resumeException($$02$i|0);
       // unreachable;
      }
      $21 = (___cxa_begin_catch(($17|0))|0); //@line 556 "./common/runtime.h"
      $22 = HEAP32[$21>>2]|0; //@line 560 "./common/runtime.h"
      $23 = ((($22)) + 8|0); //@line 560 "./common/runtime.h"
      $24 = HEAP32[$23>>2]|0; //@line 560 "./common/runtime.h"
      $25 = (FUNCTION_TABLE_ii[$24 & 8191]($21)|0); //@line 560 "./common/runtime.h"
      __THREW__ = 0;
      invoke_vii(3380,($L|0),($25|0)); //@line 560 "./common/runtime.h"
      $26 = __THREW__; __THREW__ = 0;
      $27 = $26&1;
      if (!($27)) {
       ___cxa_end_catch(); //@line 561 "./common/runtime.h"
       $28 = HEAP32[$9>>2]|0; //@line 53 "modules/font/wrap_Font.cpp"
       $29 = ((($28)) + 12|0); //@line 53 "modules/font/wrap_Font.cpp"
       $30 = HEAP32[$29>>2]|0; //@line 53 "modules/font/wrap_Font.cpp"
       FUNCTION_TABLE_vi[$30 & 8191]($9); //@line 53 "modules/font/wrap_Font.cpp"
       $31 = (_lua_tolstring($L,-1,0)|0); //@line 566 "./common/runtime.h"
       HEAP32[$vararg_buffer>>2] = $31; //@line 566 "./common/runtime.h"
       (_luaL_error($L,595867,$vararg_buffer)|0); //@line 566 "./common/runtime.h"
       $t$0 = 0;
       break;
      }
      $32 = ___cxa_find_matching_catch_2()|0;
      $33 = tempRet0;
      __THREW__ = 0;
      invoke_v(3378); //@line 561 "./common/runtime.h"
      $34 = __THREW__; __THREW__ = 0;
      $35 = $34&1;
      if ($35) {
       $39 = ___cxa_find_matching_catch_3(0|0)|0;
       $40 = tempRet0;
       ___clang_call_terminate($39); //@line 561 "./common/runtime.h"
       // unreachable; //@line 561 "./common/runtime.h"
      } else {
       $$01$i = $33;$$02$i = $32;
       ___resumeException($$02$i|0);
       // unreachable;
      }
     } else {
      $36 = HEAP32[$9>>2]|0; //@line 53 "modules/font/wrap_Font.cpp"
      $37 = ((($36)) + 12|0); //@line 53 "modules/font/wrap_Font.cpp"
      $38 = HEAP32[$37>>2]|0; //@line 53 "modules/font/wrap_Font.cpp"
      FUNCTION_TABLE_vi[$38 & 8191]($9); //@line 53 "modules/font/wrap_Font.cpp"
      $t$0 = $14;
     }
    } while(0);
    __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,9,$t$0); //@line 56 "modules/font/wrap_Font.cpp"
    $41 = HEAP32[$t$0>>2]|0; //@line 57 "modules/font/wrap_Font.cpp"
    $42 = ((($41)) + 12|0); //@line 57 "modules/font/wrap_Font.cpp"
    $43 = HEAP32[$42>>2]|0; //@line 57 "modules/font/wrap_Font.cpp"
    FUNCTION_TABLE_vi[$43 & 8191]($t$0); //@line 57 "modules/font/wrap_Font.cpp"
    $$0 = 1;
    STACKTOP = sp;return ($$0|0); //@line 65 "modules/font/wrap_Font.cpp"
   }
  }
 }
 $6 = (__ZN4love4font23w_newTrueTypeRasterizerEP9lua_State($L)|0); //@line 43 "modules/font/wrap_Font.cpp"
 $$0 = $6;
 STACKTOP = sp;return ($$0|0); //@line 65 "modules/font/wrap_Font.cpp"
}
function __ZN4love4font23w_newTrueTypeRasterizerEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$01$i5 = 0, $$02$i = 0, $$02$i4 = 0, $$2 = 0, $$pre = 0, $$pre11 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0.0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $d$0 = 0, $hinting = 0, $t$4 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0;
 var $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $hinting = sp + 28|0;
 HEAP32[$hinting>>2] = 0; //@line 70 "modules/font/wrap_Font.cpp"
 $0 = (_lua_type($L,1)|0); //@line 72 "modules/font/wrap_Font.cpp"
 $1 = ($0|0)==(3); //@line 72 "modules/font/wrap_Font.cpp"
 do {
  if ($1) {
   label = 3;
  } else {
   $2 = (_lua_type($L,1)|0); //@line 72 "modules/font/wrap_Font.cpp"
   $3 = ($2|0)==(-1); //@line 72 "modules/font/wrap_Font.cpp"
   if ($3) {
    label = 3;
   } else {
    $38 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,1,2)|0); //@line 87 "modules/font/wrap_Font.cpp"
    if ($38) {
     $39 = (__ZN4love14luax_checkdataEP9lua_Statei($L,1)|0); //@line 89 "modules/font/wrap_Font.cpp"
     $40 = HEAP32[$39>>2]|0; //@line 90 "modules/font/wrap_Font.cpp"
     $41 = ((($40)) + 8|0); //@line 90 "modules/font/wrap_Font.cpp"
     $42 = HEAP32[$41>>2]|0; //@line 90 "modules/font/wrap_Font.cpp"
     FUNCTION_TABLE_vi[$42 & 8191]($39); //@line 90 "modules/font/wrap_Font.cpp"
     $d$0 = $39;
    } else {
     $43 = (__ZN4love10filesystem16luax_getfiledataEP9lua_Statei($L,1)|0); //@line 93 "modules/font/wrap_Font.cpp"
     $d$0 = $43;
    }
    $44 = (+_luaL_optnumber($L,2,12.0)); //@line 95 "modules/font/wrap_Font.cpp"
    $45 = (~~(($44))); //@line 95 "modules/font/wrap_Font.cpp"
    $46 = (_lua_type($L,3)|0); //@line 97 "modules/font/wrap_Font.cpp"
    $47 = ($46|0)<(1); //@line 97 "modules/font/wrap_Font.cpp"
    do {
     if ($47) {
      $55 = 0;
     } else {
      $48 = (_luaL_checklstring($L,3,0)|0); //@line 97 "modules/font/wrap_Font.cpp"
      $49 = ($48|0)==(0|0); //@line 98 "modules/font/wrap_Font.cpp"
      if ($49) {
       $55 = 0;
      } else {
       $50 = (__ZN4love4font18TrueTypeRasterizer11getConstantEPKcRNS1_7HintingE($48,$hinting)|0); //@line 98 "modules/font/wrap_Font.cpp"
       if ($50) {
        $$pre = HEAP32[$hinting>>2]|0; //@line 102 "modules/font/wrap_Font.cpp"
        $55 = $$pre;
        break;
       }
       HEAP32[$vararg_buffer7>>2] = $48; //@line 99 "modules/font/wrap_Font.cpp"
       $83 = (_luaL_error($L,329027,$vararg_buffer7)|0); //@line 99 "modules/font/wrap_Font.cpp"
       $$2 = $83;
       STACKTOP = sp;return ($$2|0); //@line 110 "modules/font/wrap_Font.cpp"
      }
     }
    } while(0);
    $51 = HEAP32[(625620)>>2]|0; //@line 98 "./common/Module.h"
    $52 = HEAP32[$51>>2]|0; //@line 102 "modules/font/wrap_Font.cpp"
    $53 = ((($52)) + 32|0); //@line 102 "modules/font/wrap_Font.cpp"
    $54 = HEAP32[$53>>2]|0; //@line 102 "modules/font/wrap_Font.cpp"
    __THREW__ = 0;
    $56 = (invoke_iiiii($54|0,($51|0),($d$0|0),($45|0),($55|0))|0); //@line 102 "modules/font/wrap_Font.cpp"
    $57 = __THREW__; __THREW__ = 0;
    $58 = $57&1;
    if (!($58)) {
     $78 = HEAP32[$d$0>>2]|0; //@line 103 "modules/font/wrap_Font.cpp"
     $79 = ((($78)) + 12|0); //@line 103 "modules/font/wrap_Font.cpp"
     $80 = HEAP32[$79>>2]|0; //@line 103 "modules/font/wrap_Font.cpp"
     FUNCTION_TABLE_vi[$80 & 8191]($d$0); //@line 103 "modules/font/wrap_Font.cpp"
     $t$4 = $56;
     break;
    }
    $59 = ___cxa_find_matching_catch_3(8992|0)|0;
    $60 = tempRet0;
    $61 = (_llvm_eh_typeid_for((8992|0))|0); //@line 556 "./common/runtime.h"
    $62 = ($60|0)==($61|0); //@line 556 "./common/runtime.h"
    if (!($62)) {
     $$01$i5 = $60;$$02$i4 = $59;
     ___resumeException($$02$i4|0);
     // unreachable;
    }
    $63 = (___cxa_begin_catch(($59|0))|0); //@line 556 "./common/runtime.h"
    $64 = HEAP32[$63>>2]|0; //@line 560 "./common/runtime.h"
    $65 = ((($64)) + 8|0); //@line 560 "./common/runtime.h"
    $66 = HEAP32[$65>>2]|0; //@line 560 "./common/runtime.h"
    $67 = (FUNCTION_TABLE_ii[$66 & 8191]($63)|0); //@line 560 "./common/runtime.h"
    __THREW__ = 0;
    invoke_vii(3380,($L|0),($67|0)); //@line 560 "./common/runtime.h"
    $68 = __THREW__; __THREW__ = 0;
    $69 = $68&1;
    if (!($69)) {
     ___cxa_end_catch(); //@line 561 "./common/runtime.h"
     $70 = HEAP32[$d$0>>2]|0; //@line 103 "modules/font/wrap_Font.cpp"
     $71 = ((($70)) + 12|0); //@line 103 "modules/font/wrap_Font.cpp"
     $72 = HEAP32[$71>>2]|0; //@line 103 "modules/font/wrap_Font.cpp"
     FUNCTION_TABLE_vi[$72 & 8191]($d$0); //@line 103 "modules/font/wrap_Font.cpp"
     $73 = (_lua_tolstring($L,-1,0)|0); //@line 566 "./common/runtime.h"
     HEAP32[$vararg_buffer4>>2] = $73; //@line 566 "./common/runtime.h"
     (_luaL_error($L,595867,$vararg_buffer4)|0); //@line 566 "./common/runtime.h"
     $t$4 = 0;
     break;
    }
    $74 = ___cxa_find_matching_catch_2()|0;
    $75 = tempRet0;
    __THREW__ = 0;
    invoke_v(3378); //@line 561 "./common/runtime.h"
    $76 = __THREW__; __THREW__ = 0;
    $77 = $76&1;
    if ($77) {
     $81 = ___cxa_find_matching_catch_3(0|0)|0;
     $82 = tempRet0;
     ___clang_call_terminate($81); //@line 561 "./common/runtime.h"
     // unreachable; //@line 561 "./common/runtime.h"
    } else {
     $$01$i5 = $75;$$02$i4 = $74;
     ___resumeException($$02$i4|0);
     // unreachable;
    }
   }
  }
 } while(0);
 do {
  if ((label|0) == 3) {
   $4 = (+_luaL_optnumber($L,1,12.0)); //@line 75 "modules/font/wrap_Font.cpp"
   $5 = (~~(($4))); //@line 75 "modules/font/wrap_Font.cpp"
   $6 = (_lua_type($L,2)|0); //@line 77 "modules/font/wrap_Font.cpp"
   $7 = ($6|0)<(1); //@line 77 "modules/font/wrap_Font.cpp"
   do {
    if ($7) {
     $15 = 0;
    } else {
     $8 = (_luaL_checklstring($L,2,0)|0); //@line 77 "modules/font/wrap_Font.cpp"
     $9 = ($8|0)==(0|0); //@line 78 "modules/font/wrap_Font.cpp"
     if ($9) {
      $15 = 0;
     } else {
      $10 = (__ZN4love4font18TrueTypeRasterizer11getConstantEPKcRNS1_7HintingE($8,$hinting)|0); //@line 78 "modules/font/wrap_Font.cpp"
      if ($10) {
       $$pre11 = HEAP32[$hinting>>2]|0; //@line 81 "modules/font/wrap_Font.cpp"
       $15 = $$pre11;
       break;
      }
      HEAP32[$vararg_buffer1>>2] = $8; //@line 79 "modules/font/wrap_Font.cpp"
      $37 = (_luaL_error($L,329027,$vararg_buffer1)|0); //@line 79 "modules/font/wrap_Font.cpp"
      $$2 = $37;
      STACKTOP = sp;return ($$2|0); //@line 110 "modules/font/wrap_Font.cpp"
     }
    }
   } while(0);
   $11 = HEAP32[(625620)>>2]|0; //@line 98 "./common/Module.h"
   $12 = HEAP32[$11>>2]|0; //@line 81 "modules/font/wrap_Font.cpp"
   $13 = ((($12)) + 28|0); //@line 81 "modules/font/wrap_Font.cpp"
   $14 = HEAP32[$13>>2]|0; //@line 81 "modules/font/wrap_Font.cpp"
   __THREW__ = 0;
   $16 = (invoke_iiii($14|0,($11|0),($5|0),($15|0))|0); //@line 81 "modules/font/wrap_Font.cpp"
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if ($18) {
    $19 = ___cxa_find_matching_catch_3(8992|0)|0;
    $20 = tempRet0;
    $21 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
    $22 = ($20|0)==($21|0); //@line 535 "./common/runtime.h"
    if (!($22)) {
     $$01$i = $20;$$02$i = $19;
     ___resumeException($$02$i|0);
     // unreachable;
    }
    $23 = (___cxa_begin_catch(($19|0))|0); //@line 535 "./common/runtime.h"
    $24 = HEAP32[$23>>2]|0; //@line 539 "./common/runtime.h"
    $25 = ((($24)) + 8|0); //@line 539 "./common/runtime.h"
    $26 = HEAP32[$25>>2]|0; //@line 539 "./common/runtime.h"
    $27 = (FUNCTION_TABLE_ii[$26 & 8191]($23)|0); //@line 539 "./common/runtime.h"
    __THREW__ = 0;
    invoke_vii(3380,($L|0),($27|0)); //@line 539 "./common/runtime.h"
    $28 = __THREW__; __THREW__ = 0;
    $29 = $28&1;
    if (!($29)) {
     ___cxa_end_catch(); //@line 540 "./common/runtime.h"
     $30 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
     HEAP32[$vararg_buffer>>2] = $30; //@line 543 "./common/runtime.h"
     (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
     $t$4 = 0;
     break;
    }
    $31 = ___cxa_find_matching_catch_2()|0;
    $32 = tempRet0;
    __THREW__ = 0;
    invoke_v(3378); //@line 540 "./common/runtime.h"
    $33 = __THREW__; __THREW__ = 0;
    $34 = $33&1;
    if ($34) {
     $35 = ___cxa_find_matching_catch_3(0|0)|0;
     $36 = tempRet0;
     ___clang_call_terminate($35); //@line 540 "./common/runtime.h"
     // unreachable; //@line 540 "./common/runtime.h"
    } else {
     $$01$i = $32;$$02$i = $31;
     ___resumeException($$02$i|0);
     // unreachable;
    }
   } else {
    $t$4 = $16;
   }
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,9,$t$4); //@line 107 "modules/font/wrap_Font.cpp"
 $84 = HEAP32[$t$4>>2]|0; //@line 108 "modules/font/wrap_Font.cpp"
 $85 = ((($84)) + 12|0); //@line 108 "modules/font/wrap_Font.cpp"
 $86 = HEAP32[$85>>2]|0; //@line 108 "modules/font/wrap_Font.cpp"
 FUNCTION_TABLE_vi[$86 & 8191]($t$4); //@line 108 "modules/font/wrap_Font.cpp"
 $$2 = 1;
 STACKTOP = sp;return ($$2|0); //@line 110 "modules/font/wrap_Font.cpp"
}
function __ZN4love4font21w_newBMFontRasterizerEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$01$i$i = 0;
 var $__begin$sroa$0$01$i1$i = 0, $eh$lpad$body$index6Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i$0 = 0, $i1$0 = 0, $id = 0, $id2 = 0, $images = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i1 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i2 = 0, $t$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $images = sp + 12|0;
 $id = sp + 8|0;
 $id2 = sp + 4|0;
 $0 = (__ZN4love10filesystem16luax_getfiledataEP9lua_Statei($L,1)|0); //@line 122 "modules/font/wrap_Font.cpp"
 HEAP32[$images>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($images)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($images)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 $3 = (invoke_iii(3358,($L|0),2)|0); //@line 125 "modules/font/wrap_Font.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 L1: do {
  if ($5) {
   label = 11;
  } else {
   $6 = ($3|0)==(5); //@line 125 "modules/font/wrap_Font.cpp"
   L3: do {
    if ($6) {
     $8 = ((($images)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $i$0 = 1;
     while(1) {
      __THREW__ = 0;
      $9 = (invoke_iii(3381,($L|0),2)|0); //@line 127 "modules/font/wrap_Font.cpp"
      $10 = __THREW__; __THREW__ = 0;
      $11 = $10&1;
      if ($11) {
       break;
      }
      $12 = ($i$0|0)>($9|0); //@line 127 "modules/font/wrap_Font.cpp"
      if ($12) {
       break L3;
      }
      __THREW__ = 0;
      invoke_viii(3382,($L|0),2,($i$0|0)); //@line 129 "modules/font/wrap_Font.cpp"
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if ($33) {
       break;
      }
      __THREW__ = 0;
      $34 = (invoke_iii(3358,($L|0),1)|0); //@line 114 "modules/font/wrap_Font.cpp"
      $35 = __THREW__; __THREW__ = 0;
      $36 = $35&1;
      if ($36) {
       break;
      }
      $37 = ($34|0)==(4); //@line 114 "modules/font/wrap_Font.cpp"
      if ($37) {
       label = 24;
      } else {
       __THREW__ = 0;
       $38 = (invoke_iiii(3458,($L|0),-1,5)|0); //@line 114 "modules/font/wrap_Font.cpp"
       $39 = __THREW__; __THREW__ = 0;
       $40 = $39&1;
       if ($40) {
        break;
       }
       if ($38) {
        label = 24;
       } else {
        __THREW__ = 0;
        $41 = (invoke_iiii(3458,($L|0),-1,7)|0); //@line 114 "modules/font/wrap_Font.cpp"
        $42 = __THREW__; __THREW__ = 0;
        $43 = $42&1;
        if ($43) {
         break;
        }
        if ($41) {
         label = 24;
        }
       }
      }
      if ((label|0) == 24) {
       label = 0;
       __THREW__ = 0;
       (invoke_iiiii(3457,($L|0),-1,(438231|0),(439339|0))|0); //@line 115 "modules/font/wrap_Font.cpp"
       $44 = __THREW__; __THREW__ = 0;
       $45 = $44&1;
       if ($45) {
        break;
       }
      }
      __THREW__ = 0;
      $46 = (invoke_iiii(3523,($L|0),-1,22)|0); //@line 132 "modules/font/wrap_Font.cpp"
      $47 = __THREW__; __THREW__ = 0;
      $48 = $47&1;
      if ($48) {
       break;
      }
      HEAP32[$id>>2] = $46; //@line 132 "modules/font/wrap_Font.cpp"
      $49 = HEAP32[$1>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $50 = HEAP32[$8>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $51 = ($49|0)==($50|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($51) {
       __THREW__ = 0;
       invoke_vii(3524,($images|0),($id|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $55 = __THREW__; __THREW__ = 0;
       $56 = $55&1;
       if ($56) {
        break;
       }
      } else {
       $52 = $46; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       HEAP32[$49>>2] = $52; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $53 = HEAP32[$1>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $54 = ((($53)) + 4|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$1>>2] = $54; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      $57 = HEAP32[$46>>2]|0; //@line 134 "modules/font/wrap_Font.cpp"
      $58 = ((($57)) + 8|0); //@line 134 "modules/font/wrap_Font.cpp"
      $59 = HEAP32[$58>>2]|0; //@line 134 "modules/font/wrap_Font.cpp"
      __THREW__ = 0;
      invoke_vi($59|0,($46|0)); //@line 134 "modules/font/wrap_Font.cpp"
      $60 = __THREW__; __THREW__ = 0;
      $61 = $60&1;
      if ($61) {
       break;
      }
      __THREW__ = 0;
      invoke_vii(3384,($L|0),-2); //@line 136 "modules/font/wrap_Font.cpp"
      $62 = __THREW__; __THREW__ = 0;
      $63 = $62&1;
      if ($63) {
       break;
      }
      $64 = (($i$0) + 1)|0; //@line 127 "modules/font/wrap_Font.cpp"
      $i$0 = $64;
     }
     $17 = ___cxa_find_matching_catch_2()|0;
     $18 = tempRet0;
     $eh$lpad$body$index6Z2D = $18;$eh$lpad$body$indexZ2D = $17;
     break L1;
    } else {
     $7 = ((($images)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $i1$0 = 2;
     while(1) {
      __THREW__ = 0;
      $65 = (invoke_ii(3385,($L|0))|0); //@line 141 "modules/font/wrap_Font.cpp"
      $66 = __THREW__; __THREW__ = 0;
      $67 = $66&1;
      if ($67) {
       break;
      }
      $68 = ($i1$0|0)>($65|0); //@line 141 "modules/font/wrap_Font.cpp"
      if ($68) {
       break L3;
      }
      __THREW__ = 0;
      $69 = (invoke_iii(3358,($L|0),1)|0); //@line 114 "modules/font/wrap_Font.cpp"
      $70 = __THREW__; __THREW__ = 0;
      $71 = $70&1;
      if ($71) {
       break;
      }
      $72 = ($69|0)==(4); //@line 114 "modules/font/wrap_Font.cpp"
      if ($72) {
       label = 40;
      } else {
       __THREW__ = 0;
       $73 = (invoke_iiii(3458,($L|0),($i1$0|0),5)|0); //@line 114 "modules/font/wrap_Font.cpp"
       $74 = __THREW__; __THREW__ = 0;
       $75 = $74&1;
       if ($75) {
        break;
       }
       if ($73) {
        label = 40;
       } else {
        __THREW__ = 0;
        $76 = (invoke_iiii(3458,($L|0),($i1$0|0),7)|0); //@line 114 "modules/font/wrap_Font.cpp"
        $77 = __THREW__; __THREW__ = 0;
        $78 = $77&1;
        if ($78) {
         break;
        }
        if ($76) {
         label = 40;
        }
       }
      }
      if ((label|0) == 40) {
       label = 0;
       __THREW__ = 0;
       (invoke_iiiii(3457,($L|0),($i1$0|0),(438231|0),(439339|0))|0); //@line 115 "modules/font/wrap_Font.cpp"
       $79 = __THREW__; __THREW__ = 0;
       $80 = $79&1;
       if ($80) {
        break;
       }
      }
      __THREW__ = 0;
      $81 = (invoke_iiii(3523,($L|0),($i1$0|0),22)|0); //@line 144 "modules/font/wrap_Font.cpp"
      $82 = __THREW__; __THREW__ = 0;
      $83 = $82&1;
      if ($83) {
       break;
      }
      HEAP32[$id2>>2] = $81; //@line 144 "modules/font/wrap_Font.cpp"
      $84 = HEAP32[$1>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $85 = HEAP32[$7>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $86 = ($84|0)==($85|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($86) {
       __THREW__ = 0;
       invoke_vii(3524,($images|0),($id2|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $90 = __THREW__; __THREW__ = 0;
       $91 = $90&1;
       if ($91) {
        break;
       }
      } else {
       $87 = $81; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       HEAP32[$84>>2] = $87; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $88 = HEAP32[$1>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $89 = ((($88)) + 4|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$1>>2] = $89; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      $92 = HEAP32[$81>>2]|0; //@line 146 "modules/font/wrap_Font.cpp"
      $93 = ((($92)) + 8|0); //@line 146 "modules/font/wrap_Font.cpp"
      $94 = HEAP32[$93>>2]|0; //@line 146 "modules/font/wrap_Font.cpp"
      __THREW__ = 0;
      invoke_vi($94|0,($81|0)); //@line 146 "modules/font/wrap_Font.cpp"
      $95 = __THREW__; __THREW__ = 0;
      $96 = $95&1;
      if ($96) {
       break;
      }
      $97 = (($i1$0) + 1)|0; //@line 141 "modules/font/wrap_Font.cpp"
      $i1$0 = $97;
     }
     $19 = ___cxa_find_matching_catch_2()|0;
     $20 = tempRet0;
     $eh$lpad$body$index6Z2D = $20;$eh$lpad$body$indexZ2D = $19;
     break L1;
    }
   } while(0);
   $98 = HEAP32[(625620)>>2]|0; //@line 98 "./common/Module.h"
   $99 = HEAP32[$98>>2]|0; //@line 151 "modules/font/wrap_Font.cpp"
   $100 = ((($99)) + 36|0); //@line 151 "modules/font/wrap_Font.cpp"
   $101 = HEAP32[$100>>2]|0; //@line 151 "modules/font/wrap_Font.cpp"
   __THREW__ = 0;
   $102 = (invoke_iiii($101|0,($98|0),($0|0),($images|0))|0); //@line 151 "modules/font/wrap_Font.cpp"
   $103 = __THREW__; __THREW__ = 0;
   $104 = $103&1;
   L49: do {
    if ($104) {
     $105 = ___cxa_find_matching_catch_3(8992|0)|0;
     $106 = tempRet0;
     $107 = (_llvm_eh_typeid_for((8992|0))|0); //@line 556 "./common/runtime.h"
     $108 = ($106|0)==($107|0); //@line 556 "./common/runtime.h"
     do {
      if ($108) {
       $109 = (___cxa_begin_catch(($105|0))|0); //@line 556 "./common/runtime.h"
       $110 = HEAP32[$109>>2]|0; //@line 560 "./common/runtime.h"
       $111 = ((($110)) + 8|0); //@line 560 "./common/runtime.h"
       $112 = HEAP32[$111>>2]|0; //@line 560 "./common/runtime.h"
       $113 = (FUNCTION_TABLE_ii[$112 & 8191]($109)|0); //@line 560 "./common/runtime.h"
       __THREW__ = 0;
       invoke_vii(3380,($L|0),($113|0)); //@line 560 "./common/runtime.h"
       $114 = __THREW__; __THREW__ = 0;
       $115 = $114&1;
       if ($115) {
        $139 = ___cxa_find_matching_catch_2()|0;
        $140 = tempRet0;
        __THREW__ = 0;
        invoke_v(3378); //@line 561 "./common/runtime.h"
        $141 = __THREW__; __THREW__ = 0;
        $142 = $141&1;
        if (!($142)) {
         $$01$i = $140;$$02$i = $139;
         break;
        }
        $159 = ___cxa_find_matching_catch_3(0|0)|0;
        $160 = tempRet0;
        ___clang_call_terminate($159); //@line 561 "./common/runtime.h"
        // unreachable; //@line 561 "./common/runtime.h"
       }
       __THREW__ = 0;
       invoke_v(3378); //@line 561 "./common/runtime.h"
       $116 = __THREW__; __THREW__ = 0;
       $117 = $116&1;
       if ($117) {
        label = 11;
        break L1;
       }
       $118 = HEAP32[$0>>2]|0; //@line 152 "modules/font/wrap_Font.cpp"
       $119 = ((($118)) + 12|0); //@line 152 "modules/font/wrap_Font.cpp"
       $120 = HEAP32[$119>>2]|0; //@line 152 "modules/font/wrap_Font.cpp"
       __THREW__ = 0;
       invoke_vi($120|0,($0|0)); //@line 152 "modules/font/wrap_Font.cpp"
       $121 = __THREW__; __THREW__ = 0;
       $122 = $121&1;
       if ($122) {
        label = 11;
        break L1;
       }
       $123 = HEAP32[$images>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $124 = HEAP32[$1>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $125 = ($123|0)==($124|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       L59: do {
        if (!($125)) {
         $__begin$sroa$0$01$i$i = $123;
         while(1) {
          $126 = HEAP32[$__begin$sroa$0$01$i$i>>2]|0; //@line 152 "modules/font/wrap_Font.cpp"
          $127 = HEAP32[$126>>2]|0; //@line 152 "modules/font/wrap_Font.cpp"
          $128 = ((($127)) + 12|0); //@line 152 "modules/font/wrap_Font.cpp"
          $129 = HEAP32[$128>>2]|0; //@line 152 "modules/font/wrap_Font.cpp"
          __THREW__ = 0;
          invoke_vi($129|0,($126|0)); //@line 152 "modules/font/wrap_Font.cpp"
          $130 = __THREW__; __THREW__ = 0;
          $131 = $130&1;
          if ($131) {
           break;
          }
          $132 = ((($__begin$sroa$0$01$i$i)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
          $133 = ($132|0)==($124|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
          if ($133) {
           break L59;
          } else {
           $__begin$sroa$0$01$i$i = $132;
          }
         }
         $15 = ___cxa_find_matching_catch_2()|0;
         $16 = tempRet0;
         $eh$lpad$body$index6Z2D = $16;$eh$lpad$body$indexZ2D = $15;
         break L1;
        }
       } while(0);
       __THREW__ = 0;
       $134 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 566 "./common/runtime.h"
       $135 = __THREW__; __THREW__ = 0;
       $136 = $135&1;
       if ($136) {
        label = 11;
        break L1;
       }
       __THREW__ = 0;
       HEAP32[$vararg_buffer>>2] = $134; //@line 566 "./common/runtime.h"
       (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 566 "./common/runtime.h"
       $137 = __THREW__; __THREW__ = 0;
       $138 = $137&1;
       if ($138) {
        label = 11;
        break L1;
       } else {
        $t$0 = 0;
        break L49;
       }
      } else {
       $$01$i = $106;$$02$i = $105;
      }
     } while(0);
     $eh$lpad$body$index6Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
     break L1;
    } else {
     $143 = HEAP32[$0>>2]|0; //@line 152 "modules/font/wrap_Font.cpp"
     $144 = ((($143)) + 12|0); //@line 152 "modules/font/wrap_Font.cpp"
     $145 = HEAP32[$144>>2]|0; //@line 152 "modules/font/wrap_Font.cpp"
     __THREW__ = 0;
     invoke_vi($145|0,($0|0)); //@line 152 "modules/font/wrap_Font.cpp"
     $146 = __THREW__; __THREW__ = 0;
     $147 = $146&1;
     if ($147) {
      label = 11;
      break L1;
     }
     $148 = HEAP32[$images>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $149 = HEAP32[$1>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $150 = ($148|0)==($149|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($150) {
      $t$0 = $102;
     } else {
      $__begin$sroa$0$01$i1$i = $148;
      while(1) {
       $151 = HEAP32[$__begin$sroa$0$01$i1$i>>2]|0; //@line 152 "modules/font/wrap_Font.cpp"
       $152 = HEAP32[$151>>2]|0; //@line 152 "modules/font/wrap_Font.cpp"
       $153 = ((($152)) + 12|0); //@line 152 "modules/font/wrap_Font.cpp"
       $154 = HEAP32[$153>>2]|0; //@line 152 "modules/font/wrap_Font.cpp"
       __THREW__ = 0;
       invoke_vi($154|0,($151|0)); //@line 152 "modules/font/wrap_Font.cpp"
       $155 = __THREW__; __THREW__ = 0;
       $156 = $155&1;
       if ($156) {
        break;
       }
       $157 = ((($__begin$sroa$0$01$i1$i)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       $158 = ($157|0)==($149|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       if ($158) {
        $t$0 = $102;
        break L49;
       } else {
        $__begin$sroa$0$01$i1$i = $157;
       }
      }
      $13 = ___cxa_find_matching_catch_2()|0;
      $14 = tempRet0;
      $eh$lpad$body$index6Z2D = $14;$eh$lpad$body$indexZ2D = $13;
      break L1;
     }
    }
   } while(0);
   __THREW__ = 0;
   invoke_viii(3387,($L|0),9,($t$0|0)); //@line 155 "modules/font/wrap_Font.cpp"
   $161 = __THREW__; __THREW__ = 0;
   $162 = $161&1;
   if ($162) {
    label = 11;
   } else {
    $163 = HEAP32[$t$0>>2]|0; //@line 156 "modules/font/wrap_Font.cpp"
    $164 = ((($163)) + 12|0); //@line 156 "modules/font/wrap_Font.cpp"
    $165 = HEAP32[$164>>2]|0; //@line 156 "modules/font/wrap_Font.cpp"
    __THREW__ = 0;
    invoke_vi($165|0,($t$0|0)); //@line 156 "modules/font/wrap_Font.cpp"
    $166 = __THREW__; __THREW__ = 0;
    $167 = $166&1;
    if ($167) {
     label = 11;
    } else {
     $168 = HEAP32[$images>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $169 = ($168|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $170 = $168; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($169) {
      STACKTOP = sp;return 1; //@line 158 "modules/font/wrap_Font.cpp"
     }
     $171 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $172 = ($171|0)==($168|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($172)) {
      $scevgep$i$i$i = ((($171)) + -4|0);
      $173 = $scevgep$i$i$i;
      $174 = (($173) - ($170))|0;
      $175 = $174 >>> 2;
      $176 = $175 ^ -1;
      $scevgep4$i$i$i = (($171) + ($176<<2)|0);
      HEAP32[$1>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __ZdlPv($168); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     STACKTOP = sp;return 1; //@line 158 "modules/font/wrap_Font.cpp"
    }
   }
  }
 } while(0);
 if ((label|0) == 11) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  $eh$lpad$body$index6Z2D = $22;$eh$lpad$body$indexZ2D = $21;
 }
 $23 = HEAP32[$images>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $24 = ($23|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $25 = $23; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($24) {
  ___resumeException($eh$lpad$body$indexZ2D|0);
  // unreachable;
 }
 $26 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = ($26|0)==($23|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($27)) {
  $scevgep$i$i$i1 = ((($26)) + -4|0);
  $28 = $scevgep$i$i$i1;
  $29 = (($28) - ($25))|0;
  $30 = $29 >>> 2;
  $31 = $30 ^ -1;
  $scevgep4$i$i$i2 = (($26) + ($31<<2)|0);
  HEAP32[$1>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($23); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__16vectorIPN4love5image9ImageDataENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love4font20w_newImageRasterizerEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $glyphs = 0, $t$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $glyphs = sp + 4|0;
 $0 = (_lua_type($L,1)|0); //@line 114 "modules/font/wrap_Font.cpp"
 $1 = ($0|0)==(4); //@line 114 "modules/font/wrap_Font.cpp"
 if ($1) {
  label = 4;
 } else {
  $2 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,1,5)|0); //@line 114 "modules/font/wrap_Font.cpp"
  if ($2) {
   label = 4;
  } else {
   $3 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,1,7)|0); //@line 114 "modules/font/wrap_Font.cpp"
   if ($3) {
    label = 4;
   }
  }
 }
 if ((label|0) == 4) {
  (__ZN4love12luax_convobjEP9lua_StateiPKcS3_($L,1,438231,439339)|0); //@line 115 "modules/font/wrap_Font.cpp"
 }
 $4 = (__ZN4love14luax_checktypeINS_5image9ImageDataEEEPT_P9lua_StateiNS_4TypeE($L,1,22)|0); //@line 166 "modules/font/wrap_Font.cpp"
 __ZN4love16luax_checkstringEP9lua_Statei($glyphs,$L,2); //@line 167 "modules/font/wrap_Font.cpp"
 __THREW__ = 0;
 $5 = (+invoke_diid(3525,($L|0),3,0.0)); //@line 168 "modules/font/wrap_Font.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 L7: do {
  if (!($7)) {
   $8 = (~~(($5))); //@line 168 "modules/font/wrap_Font.cpp"
   $9 = HEAP32[(625620)>>2]|0; //@line 98 "./common/Module.h"
   $10 = HEAP32[$9>>2]|0; //@line 170 "modules/font/wrap_Font.cpp"
   $11 = ((($10)) + 40|0); //@line 170 "modules/font/wrap_Font.cpp"
   $12 = HEAP32[$11>>2]|0; //@line 170 "modules/font/wrap_Font.cpp"
   __THREW__ = 0;
   $13 = (invoke_iiiii($12|0,($9|0),($4|0),($glyphs|0),($8|0))|0); //@line 170 "modules/font/wrap_Font.cpp"
   $14 = __THREW__; __THREW__ = 0;
   $15 = $14&1;
   L9: do {
    if ($15) {
     $16 = ___cxa_find_matching_catch_3(8992|0)|0;
     $17 = tempRet0;
     $18 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
     $19 = ($17|0)==($18|0); //@line 535 "./common/runtime.h"
     do {
      if ($19) {
       $20 = (___cxa_begin_catch(($16|0))|0); //@line 535 "./common/runtime.h"
       $21 = HEAP32[$20>>2]|0; //@line 539 "./common/runtime.h"
       $22 = ((($21)) + 8|0); //@line 539 "./common/runtime.h"
       $23 = HEAP32[$22>>2]|0; //@line 539 "./common/runtime.h"
       $24 = (FUNCTION_TABLE_ii[$23 & 8191]($20)|0); //@line 539 "./common/runtime.h"
       __THREW__ = 0;
       invoke_vii(3380,($L|0),($24|0)); //@line 539 "./common/runtime.h"
       $25 = __THREW__; __THREW__ = 0;
       $26 = $25&1;
       if ($26) {
        $34 = ___cxa_find_matching_catch_2()|0;
        $35 = tempRet0;
        __THREW__ = 0;
        invoke_v(3378); //@line 540 "./common/runtime.h"
        $36 = __THREW__; __THREW__ = 0;
        $37 = $36&1;
        if (!($37)) {
         $$01$i = $35;$$02$i = $34;
         break;
        }
        $38 = ___cxa_find_matching_catch_3(0|0)|0;
        $39 = tempRet0;
        ___clang_call_terminate($38); //@line 540 "./common/runtime.h"
        // unreachable; //@line 540 "./common/runtime.h"
       }
       __THREW__ = 0;
       invoke_v(3378); //@line 540 "./common/runtime.h"
       $27 = __THREW__; __THREW__ = 0;
       $28 = $27&1;
       if ($28) {
        break L7;
       }
       __THREW__ = 0;
       $29 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
       $30 = __THREW__; __THREW__ = 0;
       $31 = $30&1;
       if ($31) {
        break L7;
       }
       __THREW__ = 0;
       HEAP32[$vararg_buffer>>2] = $29; //@line 543 "./common/runtime.h"
       (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 543 "./common/runtime.h"
       $32 = __THREW__; __THREW__ = 0;
       $33 = $32&1;
       if ($33) {
        break L7;
       } else {
        $t$0 = 0;
        break L9;
       }
      } else {
       $$01$i = $17;$$02$i = $16;
      }
     } while(0);
     $eh$lpad$body$index2Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($glyphs); //@line 175 "modules/font/wrap_Font.cpp"
     ___resumeException($eh$lpad$body$indexZ2D|0);
     // unreachable;
    } else {
     $t$0 = $13;
    }
   } while(0);
   __THREW__ = 0;
   invoke_viii(3387,($L|0),9,($t$0|0)); //@line 172 "modules/font/wrap_Font.cpp"
   $40 = __THREW__; __THREW__ = 0;
   $41 = $40&1;
   if (!($41)) {
    $42 = HEAP32[$t$0>>2]|0; //@line 173 "modules/font/wrap_Font.cpp"
    $43 = ((($42)) + 12|0); //@line 173 "modules/font/wrap_Font.cpp"
    $44 = HEAP32[$43>>2]|0; //@line 173 "modules/font/wrap_Font.cpp"
    __THREW__ = 0;
    invoke_vi($44|0,($t$0|0)); //@line 173 "modules/font/wrap_Font.cpp"
    $45 = __THREW__; __THREW__ = 0;
    $46 = $45&1;
    if (!($46)) {
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($glyphs); //@line 175 "modules/font/wrap_Font.cpp"
     STACKTOP = sp;return 1; //@line 175 "modules/font/wrap_Font.cpp"
    }
   }
  }
 } while(0);
 $47 = ___cxa_find_matching_catch_2()|0;
 $48 = tempRet0;
 $eh$lpad$body$index2Z2D = $48;$eh$lpad$body$indexZ2D = $47;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($glyphs); //@line 175 "modules/font/wrap_Font.cpp"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love4font14w_newGlyphDataEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $glyph = 0, $t$0 = 0, $t$1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $glyph = sp + 4|0;
 $0 = (__ZN4love4font20luax_checkrasterizerEP9lua_Statei($L,1)|0); //@line 179 "modules/font/wrap_Font.cpp"
 $1 = (_lua_type($L,2)|0); //@line 183 "modules/font/wrap_Font.cpp"
 $2 = ($1|0)==(4); //@line 183 "modules/font/wrap_Font.cpp"
 if (!($2)) {
  $36 = (+_luaL_checknumber($L,2)); //@line 190 "modules/font/wrap_Font.cpp"
  $37 = (~~(($36))>>>0); //@line 190 "modules/font/wrap_Font.cpp"
  $38 = HEAP32[(625620)>>2]|0; //@line 98 "./common/Module.h"
  $39 = HEAP32[$38>>2]|0; //@line 191 "modules/font/wrap_Font.cpp"
  $40 = ((($39)) + 52|0); //@line 191 "modules/font/wrap_Font.cpp"
  $41 = HEAP32[$40>>2]|0; //@line 191 "modules/font/wrap_Font.cpp"
  $42 = (FUNCTION_TABLE_iiii[$41 & 8191]($38,$0,$37)|0); //@line 191 "modules/font/wrap_Font.cpp"
  $t$1 = $42;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,8,$t$1); //@line 194 "modules/font/wrap_Font.cpp"
  $43 = HEAP32[$t$1>>2]|0; //@line 195 "modules/font/wrap_Font.cpp"
  $44 = ((($43)) + 12|0); //@line 195 "modules/font/wrap_Font.cpp"
  $45 = HEAP32[$44>>2]|0; //@line 195 "modules/font/wrap_Font.cpp"
  FUNCTION_TABLE_vi[$45 & 8191]($t$1); //@line 195 "modules/font/wrap_Font.cpp"
  STACKTOP = sp;return 1; //@line 196 "modules/font/wrap_Font.cpp"
 }
 __ZN4love16luax_checkstringEP9lua_Statei($glyph,$L,2); //@line 185 "modules/font/wrap_Font.cpp"
 $3 = HEAP32[(625620)>>2]|0; //@line 98 "./common/Module.h"
 $4 = HEAP32[$3>>2]|0; //@line 186 "modules/font/wrap_Font.cpp"
 $5 = ((($4)) + 48|0); //@line 186 "modules/font/wrap_Font.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 186 "modules/font/wrap_Font.cpp"
 __THREW__ = 0;
 $7 = (invoke_iiii($6|0,($3|0),($0|0),($glyph|0))|0); //@line 186 "modules/font/wrap_Font.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 L5: do {
  if ($9) {
   $10 = ___cxa_find_matching_catch_3(8992|0)|0;
   $11 = tempRet0;
   $12 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $13 = ($11|0)==($12|0); //@line 535 "./common/runtime.h"
   do {
    if ($13) {
     $14 = (___cxa_begin_catch(($10|0))|0); //@line 535 "./common/runtime.h"
     $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
     $16 = ((($15)) + 8|0); //@line 539 "./common/runtime.h"
     $17 = HEAP32[$16>>2]|0; //@line 539 "./common/runtime.h"
     $18 = (FUNCTION_TABLE_ii[$17 & 8191]($14)|0); //@line 539 "./common/runtime.h"
     __THREW__ = 0;
     invoke_vii(3380,($L|0),($18|0)); //@line 539 "./common/runtime.h"
     $19 = __THREW__; __THREW__ = 0;
     $20 = $19&1;
     if ($20) {
      $28 = ___cxa_find_matching_catch_2()|0;
      $29 = tempRet0;
      __THREW__ = 0;
      invoke_v(3378); //@line 540 "./common/runtime.h"
      $30 = __THREW__; __THREW__ = 0;
      $31 = $30&1;
      if (!($31)) {
       $$01$i = $29;$$02$i = $28;
       break;
      }
      $32 = ___cxa_find_matching_catch_3(0|0)|0;
      $33 = tempRet0;
      ___clang_call_terminate($32); //@line 540 "./common/runtime.h"
      // unreachable; //@line 540 "./common/runtime.h"
     }
     __THREW__ = 0;
     invoke_v(3378); //@line 540 "./common/runtime.h"
     $21 = __THREW__; __THREW__ = 0;
     $22 = $21&1;
     if (!($22)) {
      __THREW__ = 0;
      $23 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
      $24 = __THREW__; __THREW__ = 0;
      $25 = $24&1;
      if (!($25)) {
       __THREW__ = 0;
       HEAP32[$vararg_buffer>>2] = $23; //@line 543 "./common/runtime.h"
       (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 543 "./common/runtime.h"
       $26 = __THREW__; __THREW__ = 0;
       $27 = $26&1;
       if (!($27)) {
        $t$0 = 0;
        break L5;
       }
      }
     }
     $34 = ___cxa_find_matching_catch_2()|0;
     $35 = tempRet0;
     $eh$lpad$body$index2Z2D = $35;$eh$lpad$body$indexZ2D = $34;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($glyph); //@line 187 "modules/font/wrap_Font.cpp"
     ___resumeException($eh$lpad$body$indexZ2D|0);
     // unreachable;
    } else {
     $$01$i = $11;$$02$i = $10;
    }
   } while(0);
   $eh$lpad$body$index2Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($glyph); //@line 187 "modules/font/wrap_Font.cpp"
   ___resumeException($eh$lpad$body$indexZ2D|0);
   // unreachable;
  } else {
   $t$0 = $7;
  }
 } while(0);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($glyph); //@line 187 "modules/font/wrap_Font.cpp"
 $t$1 = $t$0;
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,8,$t$1); //@line 194 "modules/font/wrap_Font.cpp"
 $43 = HEAP32[$t$1>>2]|0; //@line 195 "modules/font/wrap_Font.cpp"
 $44 = ((($43)) + 12|0); //@line 195 "modules/font/wrap_Font.cpp"
 $45 = HEAP32[$44>>2]|0; //@line 195 "modules/font/wrap_Font.cpp"
 FUNCTION_TABLE_vi[$45 & 8191]($t$1); //@line 195 "modules/font/wrap_Font.cpp"
 STACKTOP = sp;return 1; //@line 196 "modules/font/wrap_Font.cpp"
}
function _luaopen_love_font($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0;
 var $eh$lpad$body$i$indexZ2D = 0, $instance$sroa$0$1 = 0, $vararg_buffer = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $w = sp + 4|0;
 $0 = HEAP32[(625620)>>2]|0; //@line 98 "./common/Module.h"
 $1 = $0; //@line 219 "modules/font/wrap_Font.cpp"
 $2 = ($0|0)==(0|0); //@line 220 "modules/font/wrap_Font.cpp"
 L1: do {
  if ($2) {
   __THREW__ = 0;
   $3 = (invoke_ii(3348,12)|0); //@line 222 "modules/font/wrap_Font.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   do {
    if ($5) {
     $11 = ___cxa_find_matching_catch_3(8992|0)|0;
     $12 = tempRet0;
     $eh$lpad$body$i$index2Z2D = $12;$eh$lpad$body$i$indexZ2D = $11;
    } else {
     __THREW__ = 0;
     invoke_vi(3526,($3|0)); //@line 222 "modules/font/wrap_Font.cpp"
     $6 = __THREW__; __THREW__ = 0;
     $7 = $6&1;
     if ($7) {
      $8 = ___cxa_find_matching_catch_3(8992|0)|0;
      $9 = tempRet0;
      __ZdlPv($3); //@line 222 "modules/font/wrap_Font.cpp"
      $eh$lpad$body$i$index2Z2D = $9;$eh$lpad$body$i$indexZ2D = $8;
      break;
     } else {
      $10 = $3; //@line 222 "modules/font/wrap_Font.cpp"
      $instance$sroa$0$1 = $10;
      break L1;
     }
    }
   } while(0);
   $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $14 = ($eh$lpad$body$i$index2Z2D|0)==($13|0); //@line 535 "./common/runtime.h"
   if (!($14)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $15 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
   $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
   $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
   $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $instance$sroa$0$1 = $1;
    break;
   }
   $23 = ___cxa_find_matching_catch_2()|0;
   $24 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    $27 = ___cxa_find_matching_catch_3(0|0)|0;
    $28 = tempRet0;
    ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $24;$$02$i = $23;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $29 = HEAP32[$0>>2]|0; //@line 225 "modules/font/wrap_Font.cpp"
   $30 = ((($29)) + 8|0); //@line 225 "modules/font/wrap_Font.cpp"
   $31 = HEAP32[$30>>2]|0; //@line 225 "modules/font/wrap_Font.cpp"
   FUNCTION_TABLE_vi[$31 & 8191]($0); //@line 225 "modules/font/wrap_Font.cpp"
   $instance$sroa$0$1 = $1;
  }
 } while(0);
 HEAP32[$w>>2] = $instance$sroa$0$1; //@line 228 "modules/font/wrap_Font.cpp"
 $32 = ((($w)) + 4|0); //@line 229 "modules/font/wrap_Font.cpp"
 HEAP32[$32>>2] = 329493; //@line 229 "modules/font/wrap_Font.cpp"
 $33 = ((($w)) + 8|0); //@line 230 "modules/font/wrap_Font.cpp"
 HEAP32[$33>>2] = 3; //@line 230 "modules/font/wrap_Font.cpp"
 $34 = ((($w)) + 12|0); //@line 231 "modules/font/wrap_Font.cpp"
 HEAP32[$34>>2] = 13192; //@line 231 "modules/font/wrap_Font.cpp"
 $35 = ((($w)) + 16|0); //@line 232 "modules/font/wrap_Font.cpp"
 HEAP32[$35>>2] = 13240; //@line 232 "modules/font/wrap_Font.cpp"
 $36 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 234 "modules/font/wrap_Font.cpp"
 STACKTOP = sp;return ($36|0); //@line 234 "modules/font/wrap_Font.cpp"
}
function __ZN4love14luax_checktypeINS_4font9GlyphDataEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love4font20w_GlyphData_getWidthEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4font9GlyphDataEEEPT_P9lua_StateiNS_4TypeE($L,1,8)|0); //@line 30 "modules/font/wrap_GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 36 "modules/font/wrap_GlyphData.cpp"
 $2 = ((($1)) + 28|0); //@line 36 "modules/font/wrap_GlyphData.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 36 "modules/font/wrap_GlyphData.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 36 "modules/font/wrap_GlyphData.cpp"
 _lua_pushinteger($L,$4); //@line 36 "modules/font/wrap_GlyphData.cpp"
 return 1; //@line 37 "modules/font/wrap_GlyphData.cpp"
}
function __ZN4love4font21w_GlyphData_getHeightEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4font9GlyphDataEEEPT_P9lua_StateiNS_4TypeE($L,1,8)|0); //@line 30 "modules/font/wrap_GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 43 "modules/font/wrap_GlyphData.cpp"
 $2 = ((($1)) + 24|0); //@line 43 "modules/font/wrap_GlyphData.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 43 "modules/font/wrap_GlyphData.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 43 "modules/font/wrap_GlyphData.cpp"
 _lua_pushinteger($L,$4); //@line 43 "modules/font/wrap_GlyphData.cpp"
 return 1; //@line 44 "modules/font/wrap_GlyphData.cpp"
}
function __ZN4love4font25w_GlyphData_getDimensionsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4font9GlyphDataEEEPT_P9lua_StateiNS_4TypeE($L,1,8)|0); //@line 30 "modules/font/wrap_GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 50 "modules/font/wrap_GlyphData.cpp"
 $2 = ((($1)) + 28|0); //@line 50 "modules/font/wrap_GlyphData.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 50 "modules/font/wrap_GlyphData.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 50 "modules/font/wrap_GlyphData.cpp"
 _lua_pushinteger($L,$4); //@line 50 "modules/font/wrap_GlyphData.cpp"
 $5 = HEAP32[$0>>2]|0; //@line 51 "modules/font/wrap_GlyphData.cpp"
 $6 = ((($5)) + 24|0); //@line 51 "modules/font/wrap_GlyphData.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 51 "modules/font/wrap_GlyphData.cpp"
 $8 = (FUNCTION_TABLE_ii[$7 & 8191]($0)|0); //@line 51 "modules/font/wrap_GlyphData.cpp"
 _lua_pushinteger($L,$8); //@line 51 "modules/font/wrap_GlyphData.cpp"
 return 2; //@line 52 "modules/font/wrap_GlyphData.cpp"
}
function __ZN4love4font20w_GlyphData_getGlyphEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4font9GlyphDataEEEPT_P9lua_StateiNS_4TypeE($L,1,8)|0); //@line 30 "modules/font/wrap_GlyphData.cpp"
 $1 = (__ZNK4love4font9GlyphData8getGlyphEv($0)|0); //@line 58 "modules/font/wrap_GlyphData.cpp"
 $2 = (+($1>>>0)); //@line 59 "modules/font/wrap_GlyphData.cpp"
 _lua_pushnumber($L,$2); //@line 59 "modules/font/wrap_GlyphData.cpp"
 return 1; //@line 60 "modules/font/wrap_GlyphData.cpp"
}
function __ZN4love4font26w_GlyphData_getGlyphStringEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 $1 = (__ZN4love14luax_checktypeINS_4font9GlyphDataEEEPT_P9lua_StateiNS_4TypeE($L,1,8)|0); //@line 30 "modules/font/wrap_GlyphData.cpp"
 __THREW__ = 0;
 invoke_vii(3527,($0|0),($1|0)); //@line 67 "modules/font/wrap_GlyphData.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 do {
  if ($3) {
   $8 = ___cxa_find_matching_catch_3(8992|0)|0;
   $9 = tempRet0;
   $eh$lpad$body$i$index2Z2D = $9;$eh$lpad$body$i$indexZ2D = $8;
  } else {
   __THREW__ = 0;
   invoke_vii(3405,($L|0),($0|0)); //@line 67 "modules/font/wrap_GlyphData.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   if ($5) {
    $6 = ___cxa_find_matching_catch_3(8992|0)|0;
    $7 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 67 "modules/font/wrap_GlyphData.cpp"
    $eh$lpad$body$i$index2Z2D = $7;$eh$lpad$body$i$indexZ2D = $6;
    break;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 67 "modules/font/wrap_GlyphData.cpp"
   STACKTOP = sp;return 1; //@line 68 "modules/font/wrap_GlyphData.cpp"
  }
 } while(0);
 $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $11 = ($eh$lpad$body$i$index2Z2D|0)==($10|0); //@line 535 "./common/runtime.h"
 if (!($11)) {
  $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $12 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = ((($13)) + 8|0); //@line 539 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($16|0)); //@line 539 "./common/runtime.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if (!($18)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $19 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $19; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 1; //@line 68 "modules/font/wrap_GlyphData.cpp"
 }
 $20 = ___cxa_find_matching_catch_2()|0;
 $21 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if ($23) {
  $24 = ___cxa_find_matching_catch_3(0|0)|0;
  $25 = tempRet0;
  ___clang_call_terminate($24); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $21;$$02$i = $20;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love4font22w_GlyphData_getAdvanceEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4font9GlyphDataEEEPT_P9lua_StateiNS_4TypeE($L,1,8)|0); //@line 30 "modules/font/wrap_GlyphData.cpp"
 $1 = (__ZNK4love4font9GlyphData10getAdvanceEv($0)|0); //@line 74 "modules/font/wrap_GlyphData.cpp"
 _lua_pushinteger($L,$1); //@line 74 "modules/font/wrap_GlyphData.cpp"
 return 1; //@line 75 "modules/font/wrap_GlyphData.cpp"
}
function __ZN4love4font22w_GlyphData_getBearingEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4font9GlyphDataEEEPT_P9lua_StateiNS_4TypeE($L,1,8)|0); //@line 30 "modules/font/wrap_GlyphData.cpp"
 $1 = (__ZNK4love4font9GlyphData11getBearingXEv($0)|0); //@line 81 "modules/font/wrap_GlyphData.cpp"
 _lua_pushinteger($L,$1); //@line 81 "modules/font/wrap_GlyphData.cpp"
 $2 = (__ZNK4love4font9GlyphData11getBearingYEv($0)|0); //@line 82 "modules/font/wrap_GlyphData.cpp"
 _lua_pushinteger($L,$2); //@line 82 "modules/font/wrap_GlyphData.cpp"
 return 2; //@line 83 "modules/font/wrap_GlyphData.cpp"
}
function __ZN4love4font26w_GlyphData_getBoundingBoxEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4font9GlyphDataEEEPT_P9lua_StateiNS_4TypeE($L,1,8)|0); //@line 30 "modules/font/wrap_GlyphData.cpp"
 $1 = (__ZNK4love4font9GlyphData7getMinXEv($0)|0); //@line 90 "modules/font/wrap_GlyphData.cpp"
 $2 = (__ZNK4love4font9GlyphData7getMinYEv($0)|0); //@line 91 "modules/font/wrap_GlyphData.cpp"
 $3 = (__ZNK4love4font9GlyphData7getMaxXEv($0)|0); //@line 92 "modules/font/wrap_GlyphData.cpp"
 $4 = (__ZNK4love4font9GlyphData7getMaxYEv($0)|0); //@line 93 "modules/font/wrap_GlyphData.cpp"
 $5 = (($3) - ($1))|0; //@line 95 "modules/font/wrap_GlyphData.cpp"
 $6 = (($4) - ($2))|0; //@line 96 "modules/font/wrap_GlyphData.cpp"
 _lua_pushinteger($L,$1); //@line 98 "modules/font/wrap_GlyphData.cpp"
 _lua_pushinteger($L,$2); //@line 99 "modules/font/wrap_GlyphData.cpp"
 _lua_pushinteger($L,$5); //@line 100 "modules/font/wrap_GlyphData.cpp"
 _lua_pushinteger($L,$6); //@line 101 "modules/font/wrap_GlyphData.cpp"
 return 4; //@line 103 "modules/font/wrap_GlyphData.cpp"
}
function __ZN4love4font21w_GlyphData_getFormatEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $str = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $str = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_4font9GlyphDataEEEPT_P9lua_StateiNS_4TypeE($L,1,8)|0); //@line 30 "modules/font/wrap_GlyphData.cpp"
 $1 = (__ZNK4love4font9GlyphData9getFormatEv($0)|0); //@line 111 "modules/font/wrap_GlyphData.cpp"
 $2 = (__ZN4love4font9GlyphData11getConstantENS1_6FormatERPKc($1,$str)|0); //@line 111 "modules/font/wrap_GlyphData.cpp"
 if ($2) {
  $4 = HEAP32[$str>>2]|0; //@line 114 "modules/font/wrap_GlyphData.cpp"
  _lua_pushstring($L,$4); //@line 114 "modules/font/wrap_GlyphData.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 116 "modules/font/wrap_GlyphData.cpp"
 } else {
  $3 = (_luaL_error($L,329121,$vararg_buffer)|0); //@line 112 "modules/font/wrap_GlyphData.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 116 "modules/font/wrap_GlyphData.cpp"
 }
 return (0)|0;
}
function _luaopen_glyphdata($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 132 "modules/font/wrap_GlyphData.cpp"
 HEAP32[$vararg_buffer>>2] = 12272; //@line 134 "modules/font/wrap_GlyphData.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 134 "modules/font/wrap_GlyphData.cpp"
 HEAP32[$vararg_ptr1>>2] = 13252; //@line 134 "modules/font/wrap_GlyphData.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 134 "modules/font/wrap_GlyphData.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 134 "modules/font/wrap_GlyphData.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,8,329147,$vararg_buffer)|0); //@line 134 "modules/font/wrap_GlyphData.cpp"
 STACKTOP = sp;return ($0|0); //@line 134 "modules/font/wrap_GlyphData.cpp"
}
function __ZN4love4font20luax_checkrasterizerEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4font10RasterizerEEEPT_P9lua_StateiNS_4TypeE($L,$idx,9)|0); //@line 32 "modules/font/wrap_Rasterizer.cpp"
 return ($0|0); //@line 32 "modules/font/wrap_Rasterizer.cpp"
}
function __ZN4love14luax_checktypeINS_4font10RasterizerEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love4font22w_Rasterizer_getHeightEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4font10RasterizerEEEPT_P9lua_StateiNS_4TypeE($L,1,9)|0); //@line 32 "modules/font/wrap_Rasterizer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 38 "modules/font/wrap_Rasterizer.cpp"
 $2 = ((($1)) + 16|0); //@line 38 "modules/font/wrap_Rasterizer.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 38 "modules/font/wrap_Rasterizer.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 38 "modules/font/wrap_Rasterizer.cpp"
 _lua_pushinteger($L,$4); //@line 38 "modules/font/wrap_Rasterizer.cpp"
 return 1; //@line 39 "modules/font/wrap_Rasterizer.cpp"
}
function __ZN4love4font23w_Rasterizer_getAdvanceEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4font10RasterizerEEEPT_P9lua_StateiNS_4TypeE($L,1,9)|0); //@line 32 "modules/font/wrap_Rasterizer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 45 "modules/font/wrap_Rasterizer.cpp"
 $2 = ((($1)) + 20|0); //@line 45 "modules/font/wrap_Rasterizer.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 45 "modules/font/wrap_Rasterizer.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 45 "modules/font/wrap_Rasterizer.cpp"
 _lua_pushinteger($L,$4); //@line 45 "modules/font/wrap_Rasterizer.cpp"
 return 1; //@line 46 "modules/font/wrap_Rasterizer.cpp"
}
function __ZN4love4font22w_Rasterizer_getAscentEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4font10RasterizerEEEPT_P9lua_StateiNS_4TypeE($L,1,9)|0); //@line 32 "modules/font/wrap_Rasterizer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 52 "modules/font/wrap_Rasterizer.cpp"
 $2 = ((($1)) + 24|0); //@line 52 "modules/font/wrap_Rasterizer.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 52 "modules/font/wrap_Rasterizer.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 52 "modules/font/wrap_Rasterizer.cpp"
 _lua_pushinteger($L,$4); //@line 52 "modules/font/wrap_Rasterizer.cpp"
 return 1; //@line 53 "modules/font/wrap_Rasterizer.cpp"
}
function __ZN4love4font23w_Rasterizer_getDescentEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4font10RasterizerEEEPT_P9lua_StateiNS_4TypeE($L,1,9)|0); //@line 32 "modules/font/wrap_Rasterizer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 59 "modules/font/wrap_Rasterizer.cpp"
 $2 = ((($1)) + 28|0); //@line 59 "modules/font/wrap_Rasterizer.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 59 "modules/font/wrap_Rasterizer.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 59 "modules/font/wrap_Rasterizer.cpp"
 _lua_pushinteger($L,$4); //@line 59 "modules/font/wrap_Rasterizer.cpp"
 return 1; //@line 60 "modules/font/wrap_Rasterizer.cpp"
}
function __ZN4love4font26w_Rasterizer_getLineHeightEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4font10RasterizerEEEPT_P9lua_StateiNS_4TypeE($L,1,9)|0); //@line 32 "modules/font/wrap_Rasterizer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 66 "modules/font/wrap_Rasterizer.cpp"
 $2 = ((($1)) + 32|0); //@line 66 "modules/font/wrap_Rasterizer.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 66 "modules/font/wrap_Rasterizer.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 66 "modules/font/wrap_Rasterizer.cpp"
 _lua_pushinteger($L,$4); //@line 66 "modules/font/wrap_Rasterizer.cpp"
 return 1; //@line 67 "modules/font/wrap_Rasterizer.cpp"
}
function __ZN4love4font25w_Rasterizer_getGlyphDataEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $g$0 = 0, $glyph$i$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $glyph$i$i = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_4font10RasterizerEEEPT_P9lua_StateiNS_4TypeE($L,1,9)|0); //@line 32 "modules/font/wrap_Rasterizer.cpp"
 __THREW__ = 0;
 $1 = (invoke_iii(3358,($L|0),2)|0); //@line 77 "modules/font/wrap_Rasterizer.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 do {
  if ($3) {
   label = 9;
  } else {
   $4 = ($1|0)==(4); //@line 77 "modules/font/wrap_Rasterizer.cpp"
   if (!($4)) {
    __THREW__ = 0;
    $15 = (+invoke_dii(3361,($L|0),2)); //@line 84 "modules/font/wrap_Rasterizer.cpp"
    $16 = __THREW__; __THREW__ = 0;
    $17 = $16&1;
    if ($17) {
     label = 9;
     break;
    }
    $18 = (~~(($15))>>>0); //@line 84 "modules/font/wrap_Rasterizer.cpp"
    $19 = HEAP32[$0>>2]|0; //@line 85 "modules/font/wrap_Rasterizer.cpp"
    $20 = ((($19)) + 36|0); //@line 85 "modules/font/wrap_Rasterizer.cpp"
    $21 = HEAP32[$20>>2]|0; //@line 85 "modules/font/wrap_Rasterizer.cpp"
    __THREW__ = 0;
    $22 = (invoke_iii($21|0,($0|0),($18|0))|0); //@line 85 "modules/font/wrap_Rasterizer.cpp"
    $23 = __THREW__; __THREW__ = 0;
    $24 = $23&1;
    if ($24) {
     label = 9;
     break;
    } else {
     $g$0 = $22;
     break;
    }
   }
   __THREW__ = 0;
   invoke_viii(3403,($glyph$i$i|0),($L|0),2); //@line 79 "modules/font/wrap_Rasterizer.cpp"
   $5 = __THREW__; __THREW__ = 0;
   $6 = $5&1;
   if ($6) {
    label = 9;
   } else {
    $7 = HEAP32[$0>>2]|0; //@line 80 "modules/font/wrap_Rasterizer.cpp"
    $8 = ((($7)) + 40|0); //@line 80 "modules/font/wrap_Rasterizer.cpp"
    $9 = HEAP32[$8>>2]|0; //@line 80 "modules/font/wrap_Rasterizer.cpp"
    __THREW__ = 0;
    $10 = (invoke_iii($9|0,($0|0),($glyph$i$i|0))|0); //@line 80 "modules/font/wrap_Rasterizer.cpp"
    $11 = __THREW__; __THREW__ = 0;
    $12 = $11&1;
    if ($12) {
     $13 = ___cxa_find_matching_catch_3(8992|0)|0;
     $14 = tempRet0;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($glyph$i$i); //@line 81 "modules/font/wrap_Rasterizer.cpp"
     $eh$lpad$body$i$index2Z2D = $14;$eh$lpad$body$i$indexZ2D = $13;
     label = 10;
     break;
    } else {
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($glyph$i$i); //@line 81 "modules/font/wrap_Rasterizer.cpp"
     $g$0 = $10;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 9) {
  $25 = ___cxa_find_matching_catch_3(8992|0)|0;
  $26 = tempRet0;
  $eh$lpad$body$i$index2Z2D = $26;$eh$lpad$body$i$indexZ2D = $25;
  label = 10;
 }
 do {
  if ((label|0) == 10) {
   $27 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $28 = ($eh$lpad$body$i$index2Z2D|0)==($27|0); //@line 535 "./common/runtime.h"
   if (!($28)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $29 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $30 = HEAP32[$29>>2]|0; //@line 539 "./common/runtime.h"
   $31 = ((($30)) + 8|0); //@line 539 "./common/runtime.h"
   $32 = HEAP32[$31>>2]|0; //@line 539 "./common/runtime.h"
   $33 = (FUNCTION_TABLE_ii[$32 & 8191]($29)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($33|0)); //@line 539 "./common/runtime.h"
   $34 = __THREW__; __THREW__ = 0;
   $35 = $34&1;
   if (!($35)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $36 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $36; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $g$0 = 0;
    break;
   }
   $37 = ___cxa_find_matching_catch_2()|0;
   $38 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $39 = __THREW__; __THREW__ = 0;
   $40 = $39&1;
   if ($40) {
    $41 = ___cxa_find_matching_catch_3(0|0)|0;
    $42 = tempRet0;
    ___clang_call_terminate($41); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $38;$$02$i = $37;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,8,$g$0); //@line 89 "modules/font/wrap_Rasterizer.cpp"
 $43 = HEAP32[$g$0>>2]|0; //@line 90 "modules/font/wrap_Rasterizer.cpp"
 $44 = ((($43)) + 12|0); //@line 90 "modules/font/wrap_Rasterizer.cpp"
 $45 = HEAP32[$44>>2]|0; //@line 90 "modules/font/wrap_Rasterizer.cpp"
 FUNCTION_TABLE_vi[$45 & 8191]($g$0); //@line 90 "modules/font/wrap_Rasterizer.cpp"
 STACKTOP = sp;return 1; //@line 91 "modules/font/wrap_Rasterizer.cpp"
}
function __ZN4love4font26w_Rasterizer_getGlyphCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4font10RasterizerEEEPT_P9lua_StateiNS_4TypeE($L,1,9)|0); //@line 32 "modules/font/wrap_Rasterizer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 97 "modules/font/wrap_Rasterizer.cpp"
 $2 = ((($1)) + 44|0); //@line 97 "modules/font/wrap_Rasterizer.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 97 "modules/font/wrap_Rasterizer.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 97 "modules/font/wrap_Rasterizer.cpp"
 _lua_pushinteger($L,$4); //@line 97 "modules/font/wrap_Rasterizer.cpp"
 return 1; //@line 98 "modules/font/wrap_Rasterizer.cpp"
}
function __ZN4love4font22w_Rasterizer_hasGlyphsEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0.0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $hasglyph$0 = 0, $hasglyph$0$lcssa = 0, $hasglyph$0$lcssa19 = 0, $hasglyph$020 = 0;
 var $hasglyph$18 = 0, $hasglyph$18$in = 0, $hasglyph$2 = 0, $hasglyph$3 = 0, $i$01$i$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 $1 = (__ZN4love14luax_checktypeINS_4font10RasterizerEEEPT_P9lua_StateiNS_4TypeE($L,1,9)|0); //@line 32 "modules/font/wrap_Rasterizer.cpp"
 $2 = (_lua_gettop($L)|0); //@line 107 "modules/font/wrap_Rasterizer.cpp"
 $3 = (($2) + -1)|0; //@line 107 "modules/font/wrap_Rasterizer.cpp"
 $4 = ($3|0)<(1); //@line 108 "modules/font/wrap_Rasterizer.cpp"
 $5 = $4 ? 1 : $3; //@line 108 "modules/font/wrap_Rasterizer.cpp"
 $6 = ($5|0)>(0); //@line 111 "modules/font/wrap_Rasterizer.cpp"
 L1: do {
  if ($6) {
   $hasglyph$0 = 0;$i$01$i$i = 2;
   while(1) {
    __THREW__ = 0;
    $9 = (invoke_iii(3358,($L|0),($i$01$i$i|0))|0); //@line 113 "modules/font/wrap_Rasterizer.cpp"
    $10 = __THREW__; __THREW__ = 0;
    $11 = $10&1;
    if ($11) {
     $hasglyph$0$lcssa = $hasglyph$0;
     label = 14;
     break;
    }
    $12 = ($9|0)==(4); //@line 113 "modules/font/wrap_Rasterizer.cpp"
    if ($12) {
     $13 = HEAP32[$1>>2]|0; //@line 114 "modules/font/wrap_Rasterizer.cpp"
     $14 = ((($13)) + 52|0); //@line 114 "modules/font/wrap_Rasterizer.cpp"
     $15 = HEAP32[$14>>2]|0; //@line 114 "modules/font/wrap_Rasterizer.cpp"
     __THREW__ = 0;
     invoke_viii(3403,($0|0),($L|0),($i$01$i$i|0)); //@line 114 "modules/font/wrap_Rasterizer.cpp"
     $16 = __THREW__; __THREW__ = 0;
     $17 = $16&1;
     if ($17) {
      $hasglyph$0$lcssa = $hasglyph$0;
      label = 14;
      break;
     }
     __THREW__ = 0;
     $18 = (invoke_iii($15|0,($1|0),($0|0))|0); //@line 114 "modules/font/wrap_Rasterizer.cpp"
     $19 = __THREW__; __THREW__ = 0;
     $20 = $19&1;
     if ($20) {
      $hasglyph$0$lcssa19 = $hasglyph$0;
      label = 9;
      break;
     }
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 114 "modules/font/wrap_Rasterizer.cpp"
     if ($18) {
      $hasglyph$18$in = $18;
     } else {
      $hasglyph$2 = 0;
      break L1;
     }
    } else {
     $23 = HEAP32[$1>>2]|0; //@line 116 "modules/font/wrap_Rasterizer.cpp"
     $24 = ((($23)) + 48|0); //@line 116 "modules/font/wrap_Rasterizer.cpp"
     $25 = HEAP32[$24>>2]|0; //@line 116 "modules/font/wrap_Rasterizer.cpp"
     __THREW__ = 0;
     $26 = (+invoke_dii(3361,($L|0),($i$01$i$i|0))); //@line 116 "modules/font/wrap_Rasterizer.cpp"
     $27 = __THREW__; __THREW__ = 0;
     $28 = $27&1;
     if ($28) {
      $hasglyph$0$lcssa = $hasglyph$0;
      label = 14;
      break;
     }
     $29 = (~~(($26))>>>0); //@line 116 "modules/font/wrap_Rasterizer.cpp"
     __THREW__ = 0;
     $30 = (invoke_iii($25|0,($1|0),($29|0))|0); //@line 116 "modules/font/wrap_Rasterizer.cpp"
     $31 = __THREW__; __THREW__ = 0;
     $32 = $31&1;
     if ($32) {
      $hasglyph$0$lcssa = $hasglyph$0;
      label = 14;
      break;
     }
     if ($30) {
      $hasglyph$18$in = $30;
     } else {
      $hasglyph$2 = 0;
      break L1;
     }
    }
    $hasglyph$18 = $hasglyph$18$in&1; //@line 114 "modules/font/wrap_Rasterizer.cpp"
    $7 = (($i$01$i$i) + 1)|0; //@line 111 "modules/font/wrap_Rasterizer.cpp"
    $8 = ($i$01$i$i|0)>($5|0); //@line 111 "modules/font/wrap_Rasterizer.cpp"
    if ($8) {
     $hasglyph$2 = $hasglyph$18;
     break L1;
    } else {
     $hasglyph$0 = $hasglyph$18;$i$01$i$i = $7;
    }
   }
   if ((label|0) == 9) {
    $21 = ___cxa_find_matching_catch_3(8992|0)|0;
    $22 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 114 "modules/font/wrap_Rasterizer.cpp"
    $eh$lpad$body$i$index2Z2D = $22;$eh$lpad$body$i$indexZ2D = $21;$hasglyph$020 = $hasglyph$0$lcssa19;
   }
   else if ((label|0) == 14) {
    $33 = ___cxa_find_matching_catch_3(8992|0)|0;
    $34 = tempRet0;
    $eh$lpad$body$i$index2Z2D = $34;$eh$lpad$body$i$indexZ2D = $33;$hasglyph$020 = $hasglyph$0$lcssa;
   }
   $35 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $36 = ($eh$lpad$body$i$index2Z2D|0)==($35|0); //@line 535 "./common/runtime.h"
   if (!($36)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $37 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $38 = HEAP32[$37>>2]|0; //@line 539 "./common/runtime.h"
   $39 = ((($38)) + 8|0); //@line 539 "./common/runtime.h"
   $40 = HEAP32[$39>>2]|0; //@line 539 "./common/runtime.h"
   $41 = (FUNCTION_TABLE_ii[$40 & 8191]($37)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($41|0)); //@line 539 "./common/runtime.h"
   $42 = __THREW__; __THREW__ = 0;
   $43 = $42&1;
   if (!($43)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $44 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $44; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $hasglyph$3 = $hasglyph$020;
    $51 = ($hasglyph$3<<24>>24)!=(0); //@line 123 "modules/font/wrap_Rasterizer.cpp"
    __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$51); //@line 123 "modules/font/wrap_Rasterizer.cpp"
    STACKTOP = sp;return 1; //@line 124 "modules/font/wrap_Rasterizer.cpp"
   }
   $45 = ___cxa_find_matching_catch_2()|0;
   $46 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $47 = __THREW__; __THREW__ = 0;
   $48 = $47&1;
   if ($48) {
    $49 = ___cxa_find_matching_catch_3(0|0)|0;
    $50 = tempRet0;
    ___clang_call_terminate($49); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $46;$$02$i = $45;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $hasglyph$2 = 0;
  }
 } while(0);
 $hasglyph$3 = $hasglyph$2;
 $51 = ($hasglyph$3<<24>>24)!=(0); //@line 123 "modules/font/wrap_Rasterizer.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$51); //@line 123 "modules/font/wrap_Rasterizer.cpp"
 STACKTOP = sp;return 1; //@line 124 "modules/font/wrap_Rasterizer.cpp"
}
function _luaopen_rasterizer($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 140 "modules/font/wrap_Rasterizer.cpp"
 HEAP32[$vararg_buffer>>2] = 13332; //@line 142 "modules/font/wrap_Rasterizer.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 142 "modules/font/wrap_Rasterizer.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 142 "modules/font/wrap_Rasterizer.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,9,329192,$vararg_buffer)|0); //@line 142 "modules/font/wrap_Rasterizer.cpp"
 STACKTOP = sp;return ($0|0); //@line 142 "modules/font/wrap_Rasterizer.cpp"
}
function __ZN4love8graphics6opengl7w_resetEP9lua_State($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZN4love8graphics6opengl8Graphics5resetEv($1); //@line 59 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 0; //@line 60 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl7w_clearEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$01$i3 = 0, $$02$i = 0, $$02$i2 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0.0, $103 = 0, $104 = 0, $105 = 0.0, $106 = 0, $107 = 0, $108 = 0.0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0, $127 = 0.0, $128 = 0.0, $129 = 0, $13 = 0;
 var $130 = 0.0, $131 = 0.0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0, $92 = 0, $93 = 0.0;
 var $94 = 0, $95 = 0, $96 = 0.0, $97 = 0, $98 = 0, $99 = 0.0, $color = 0, $color$byval_copy = 0, $colors = 0, $eh$lpad$body$index7Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i$0 = 0, $scevgep = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i$i = 0, $scevgep$i$i$i6 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i$i = 0, $scevgep4$i$i$i7 = 0, $vararg_buffer = 0;
 var $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $color$byval_copy = sp + 40|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $color = sp + 24|0;
 $colors = sp + 12|0;
 ;HEAP32[$color>>2]=0|0;HEAP32[$color+4>>2]=0|0;HEAP32[$color+8>>2]=0|0;HEAP32[$color+12>>2]=0|0; //@line 37 "./modules/graphics/Color.h"
 $0 = (_lua_type($L,1)|0); //@line 67 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = ($0|0)<(1); //@line 67 "modules/graphics/opengl/wrap_Graphics.cpp"
 do {
  if ($1) {
   ;HEAP32[$color>>2]=0|0;HEAP32[$color+4>>2]=0|0;HEAP32[$color+8>>2]=0|0;HEAP32[$color+12>>2]=0|0; //@line 42 "./modules/graphics/Color.h"
  } else {
   $2 = (_lua_type($L,1)|0); //@line 69 "modules/graphics/opengl/wrap_Graphics.cpp"
   $3 = ($2|0)==(5); //@line 69 "modules/graphics/opengl/wrap_Graphics.cpp"
   if (!($3)) {
    $122 = (+_luaL_checknumber($L,1)); //@line 98 "modules/graphics/opengl/wrap_Graphics.cpp"
    $123 = $122; //@line 98 "modules/graphics/opengl/wrap_Graphics.cpp"
    HEAPF32[$color>>2] = $123; //@line 98 "modules/graphics/opengl/wrap_Graphics.cpp"
    $124 = (+_luaL_checknumber($L,2)); //@line 99 "modules/graphics/opengl/wrap_Graphics.cpp"
    $125 = $124; //@line 99 "modules/graphics/opengl/wrap_Graphics.cpp"
    $126 = ((($color)) + 4|0); //@line 99 "modules/graphics/opengl/wrap_Graphics.cpp"
    HEAPF32[$126>>2] = $125; //@line 99 "modules/graphics/opengl/wrap_Graphics.cpp"
    $127 = (+_luaL_checknumber($L,3)); //@line 100 "modules/graphics/opengl/wrap_Graphics.cpp"
    $128 = $127; //@line 100 "modules/graphics/opengl/wrap_Graphics.cpp"
    $129 = ((($color)) + 8|0); //@line 100 "modules/graphics/opengl/wrap_Graphics.cpp"
    HEAPF32[$129>>2] = $128; //@line 100 "modules/graphics/opengl/wrap_Graphics.cpp"
    $130 = (+_luaL_optnumber($L,4,255.0)); //@line 101 "modules/graphics/opengl/wrap_Graphics.cpp"
    $131 = $130; //@line 101 "modules/graphics/opengl/wrap_Graphics.cpp"
    $132 = ((($color)) + 12|0); //@line 101 "modules/graphics/opengl/wrap_Graphics.cpp"
    HEAPF32[$132>>2] = $131; //@line 101 "modules/graphics/opengl/wrap_Graphics.cpp"
    break;
   }
   $4 = (_lua_gettop($L)|0); //@line 71 "modules/graphics/opengl/wrap_Graphics.cpp"
   HEAP32[$colors>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $5 = ((($colors)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$5>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $6 = ((($colors)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$6>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $7 = ($4|0)==(0); //@line 1080 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   do {
    if ($7) {
     $i$0 = 0;
    } else {
     $8 = ($4>>>0)>(214748364); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($8) {
      __THREW__ = 0;
      invoke_vi(3528,($colors|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $9 = __THREW__; __THREW__ = 0;
      $10 = $9&1;
      if (!($10)) {
       label = 7;
      }
     } else {
      label = 7;
     }
     if ((label|0) == 7) {
      $11 = ($4*20)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      __THREW__ = 0;
      $12 = (invoke_ii(3348,($11|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $13 = __THREW__; __THREW__ = 0;
      $14 = $13&1;
      if (!($14)) {
       HEAP32[$colors>>2] = $12; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $15 = (($12) + (($4*20)|0)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $16 = ((($colors)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       HEAP32[$16>>2] = $15; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $scevgep = (($12) + ($11)|0); //@line 985 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       _memset(($12|0),0,($11|0))|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       HEAP32[$5>>2] = $scevgep; //@line 989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $i$0 = 0;
       break;
      }
     }
     $17 = ___cxa_find_matching_catch_2()|0;
     $18 = tempRet0;
     $19 = HEAP32[$colors>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $20 = ($19|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $21 = $19; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($20) {
      ___resumeException($17|0);
      // unreachable;
     }
     $22 = HEAP32[$5>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $23 = ($22|0)==($19|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($23)) {
      $scevgep$i$i$i$i = ((($22)) + -20|0);
      $24 = $scevgep$i$i$i$i;
      $25 = (($24) - ($21))|0;
      $26 = (($25>>>0) / 20)&-1;
      $27 = $26 ^ -1;
      $scevgep4$i$i$i$i = (($22) + (($27*20)|0)|0);
      HEAP32[$5>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __ZdlPv($19); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     ___resumeException($17|0);
     // unreachable;
    }
   } while(0);
   while(1) {
    __THREW__ = 0;
    $28 = (invoke_ii(3385,($L|0))|0); //@line 73 "modules/graphics/opengl/wrap_Graphics.cpp"
    $29 = __THREW__; __THREW__ = 0;
    $30 = $29&1;
    if ($30) {
     label = 26;
     break;
    }
    $31 = ($i$0|0)<($28|0); //@line 73 "modules/graphics/opengl/wrap_Graphics.cpp"
    if (!($31)) {
     label = 16;
     break;
    }
    $74 = (($i$0) + 1)|0; //@line 75 "modules/graphics/opengl/wrap_Graphics.cpp"
    __THREW__ = 0;
    $75 = (invoke_iii(3358,($L|0),($74|0))|0); //@line 75 "modules/graphics/opengl/wrap_Graphics.cpp"
    $76 = __THREW__; __THREW__ = 0;
    $77 = $76&1;
    if ($77) {
     label = 26;
     break;
    }
    $78 = ($75|0)<(1); //@line 75 "modules/graphics/opengl/wrap_Graphics.cpp"
    if (!($78)) {
     __THREW__ = 0;
     $79 = (invoke_iii(3381,($L|0),($74|0))|0); //@line 75 "modules/graphics/opengl/wrap_Graphics.cpp"
     $80 = __THREW__; __THREW__ = 0;
     $81 = $80&1;
     if ($81) {
      label = 26;
      break;
     }
     $82 = ($79|0)==(0); //@line 75 "modules/graphics/opengl/wrap_Graphics.cpp"
     if (!($82)) {
      __THREW__ = 0;
      invoke_viii(3382,($L|0),($74|0),1); //@line 82 "modules/graphics/opengl/wrap_Graphics.cpp"
      $83 = __THREW__; __THREW__ = 0;
      $84 = $83&1;
      if ($84) {
       label = 25;
       break;
      }
      __THREW__ = 0;
      invoke_viii(3382,($L|0),($74|0),2); //@line 82 "modules/graphics/opengl/wrap_Graphics.cpp"
      $87 = __THREW__; __THREW__ = 0;
      $88 = $87&1;
      if ($88) {
       label = 25;
       break;
      }
      __THREW__ = 0;
      invoke_viii(3382,($L|0),($74|0),3); //@line 82 "modules/graphics/opengl/wrap_Graphics.cpp"
      $154 = __THREW__; __THREW__ = 0;
      $155 = $154&1;
      if ($155) {
       label = 25;
       break;
      }
      __THREW__ = 0;
      invoke_viii(3382,($L|0),($74|0),4); //@line 82 "modules/graphics/opengl/wrap_Graphics.cpp"
      $156 = __THREW__; __THREW__ = 0;
      $157 = $156&1;
      if ($157) {
       label = 25;
       break;
      }
      $158 = HEAP32[$colors>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $159 = (((($158) + (($i$0*20)|0)|0)) + 16|0); //@line 84 "modules/graphics/opengl/wrap_Graphics.cpp"
      HEAP8[$159>>0] = 1; //@line 84 "modules/graphics/opengl/wrap_Graphics.cpp"
      __THREW__ = 0;
      $90 = (+invoke_dii(3361,($L|0),-4)); //@line 85 "modules/graphics/opengl/wrap_Graphics.cpp"
      $160 = __THREW__; __THREW__ = 0;
      $161 = $160&1;
      if ($161) {
       label = 26;
       break;
      }
      $89 = $90; //@line 85 "modules/graphics/opengl/wrap_Graphics.cpp"
      $91 = HEAP32[$colors>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $92 = (($91) + (($i$0*20)|0)|0); //@line 85 "modules/graphics/opengl/wrap_Graphics.cpp"
      HEAPF32[$92>>2] = $89; //@line 85 "modules/graphics/opengl/wrap_Graphics.cpp"
      __THREW__ = 0;
      $93 = (+invoke_dii(3361,($L|0),-3)); //@line 86 "modules/graphics/opengl/wrap_Graphics.cpp"
      $94 = __THREW__; __THREW__ = 0;
      $95 = $94&1;
      if ($95) {
       label = 26;
       break;
      }
      $96 = $93; //@line 86 "modules/graphics/opengl/wrap_Graphics.cpp"
      $97 = HEAP32[$colors>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $98 = (((($97) + (($i$0*20)|0)|0)) + 4|0); //@line 86 "modules/graphics/opengl/wrap_Graphics.cpp"
      HEAPF32[$98>>2] = $96; //@line 86 "modules/graphics/opengl/wrap_Graphics.cpp"
      __THREW__ = 0;
      $99 = (+invoke_dii(3361,($L|0),-2)); //@line 87 "modules/graphics/opengl/wrap_Graphics.cpp"
      $100 = __THREW__; __THREW__ = 0;
      $101 = $100&1;
      if ($101) {
       label = 26;
       break;
      }
      $102 = $99; //@line 87 "modules/graphics/opengl/wrap_Graphics.cpp"
      $103 = HEAP32[$colors>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $104 = (((($103) + (($i$0*20)|0)|0)) + 8|0); //@line 87 "modules/graphics/opengl/wrap_Graphics.cpp"
      HEAPF32[$104>>2] = $102; //@line 87 "modules/graphics/opengl/wrap_Graphics.cpp"
      __THREW__ = 0;
      $105 = (+invoke_diid(3525,($L|0),-1,255.0)); //@line 88 "modules/graphics/opengl/wrap_Graphics.cpp"
      $106 = __THREW__; __THREW__ = 0;
      $107 = $106&1;
      if ($107) {
       label = 26;
       break;
      }
      $108 = $105; //@line 88 "modules/graphics/opengl/wrap_Graphics.cpp"
      $109 = HEAP32[$colors>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $110 = (((($109) + (($i$0*20)|0)|0)) + 12|0); //@line 88 "modules/graphics/opengl/wrap_Graphics.cpp"
      HEAPF32[$110>>2] = $108; //@line 88 "modules/graphics/opengl/wrap_Graphics.cpp"
      __THREW__ = 0;
      invoke_vii(3384,($L|0),-5); //@line 90 "modules/graphics/opengl/wrap_Graphics.cpp"
      $111 = __THREW__; __THREW__ = 0;
      $112 = $111&1;
      if ($112) {
       label = 26;
       break;
      } else {
       $i$0 = $74;
       continue;
      }
     }
    }
    $85 = HEAP32[$colors>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $86 = (((($85) + (($i$0*20)|0)|0)) + 16|0); //@line 77 "modules/graphics/opengl/wrap_Graphics.cpp"
    HEAP8[$86>>0] = 0; //@line 77 "modules/graphics/opengl/wrap_Graphics.cpp"
    $i$0 = $74;
   }
   L40: do {
    if ((label|0) == 16) {
     $32 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
     __THREW__ = 0;
     invoke_vii(3529,($32|0),($colors|0)); //@line 93 "modules/graphics/opengl/wrap_Graphics.cpp"
     $33 = __THREW__; __THREW__ = 0;
     $34 = $33&1;
     L43: do {
      if ($34) {
       $35 = ___cxa_find_matching_catch_3(8992|0)|0;
       $36 = tempRet0;
       $37 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
       $38 = ($36|0)==($37|0); //@line 535 "./common/runtime.h"
       do {
        if ($38) {
         $39 = (___cxa_begin_catch(($35|0))|0); //@line 535 "./common/runtime.h"
         $40 = HEAP32[$39>>2]|0; //@line 539 "./common/runtime.h"
         $41 = ((($40)) + 8|0); //@line 539 "./common/runtime.h"
         $42 = HEAP32[$41>>2]|0; //@line 539 "./common/runtime.h"
         $43 = (FUNCTION_TABLE_ii[$42 & 8191]($39)|0); //@line 539 "./common/runtime.h"
         __THREW__ = 0;
         invoke_vii(3380,($L|0),($43|0)); //@line 539 "./common/runtime.h"
         $44 = __THREW__; __THREW__ = 0;
         $45 = $44&1;
         if ($45) {
          $53 = ___cxa_find_matching_catch_2()|0;
          $54 = tempRet0;
          __THREW__ = 0;
          invoke_v(3378); //@line 540 "./common/runtime.h"
          $55 = __THREW__; __THREW__ = 0;
          $56 = $55&1;
          if (!($56)) {
           $$01$i3 = $54;$$02$i2 = $53;
           break;
          }
          $57 = ___cxa_find_matching_catch_3(0|0)|0;
          $58 = tempRet0;
          ___clang_call_terminate($57); //@line 540 "./common/runtime.h"
          // unreachable; //@line 540 "./common/runtime.h"
         }
         __THREW__ = 0;
         invoke_v(3378); //@line 540 "./common/runtime.h"
         $46 = __THREW__; __THREW__ = 0;
         $47 = $46&1;
         if (!($47)) {
          __THREW__ = 0;
          $48 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
          $49 = __THREW__; __THREW__ = 0;
          $50 = $49&1;
          if (!($50)) {
           __THREW__ = 0;
           HEAP32[$vararg_buffer>>2] = $48; //@line 543 "./common/runtime.h"
           (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 543 "./common/runtime.h"
           $51 = __THREW__; __THREW__ = 0;
           $52 = $51&1;
           if (!($52)) {
            break L43;
           }
          }
         }
         $63 = ___cxa_find_matching_catch_2()|0;
         $64 = tempRet0;
         $eh$lpad$body$index7Z2D = $64;$eh$lpad$body$indexZ2D = $63;
         break L40;
        } else {
         $$01$i3 = $36;$$02$i2 = $35;
        }
       } while(0);
       $eh$lpad$body$index7Z2D = $$01$i3;$eh$lpad$body$indexZ2D = $$02$i2;
       break L40;
      }
     } while(0);
     $113 = HEAP32[$colors>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $114 = ($113|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $115 = $113; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($114)) {
      $116 = HEAP32[$5>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $117 = ($116|0)==($113|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($117)) {
       $scevgep$i$i$i6 = ((($116)) + -20|0);
       $118 = $scevgep$i$i$i6;
       $119 = (($118) - ($115))|0;
       $120 = (($119>>>0) / 20)&-1;
       $121 = $120 ^ -1;
       $scevgep4$i$i$i7 = (($116) + (($121*20)|0)|0);
       HEAP32[$5>>2] = $scevgep4$i$i$i7; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($113); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     STACKTOP = sp;return 0; //@line 106 "modules/graphics/opengl/wrap_Graphics.cpp"
    }
    else if ((label|0) == 25) {
     $59 = ___cxa_find_matching_catch_2()|0;
     $60 = tempRet0;
     $eh$lpad$body$index7Z2D = $60;$eh$lpad$body$indexZ2D = $59;
    }
    else if ((label|0) == 26) {
     $61 = ___cxa_find_matching_catch_2()|0;
     $62 = tempRet0;
     $eh$lpad$body$index7Z2D = $62;$eh$lpad$body$indexZ2D = $61;
    }
   } while(0);
   $65 = HEAP32[$colors>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $66 = ($65|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $67 = $65; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($66) {
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
   $68 = HEAP32[$5>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $69 = ($68|0)==($65|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($69)) {
    $scevgep$i$i$i = ((($68)) + -20|0);
    $70 = $scevgep$i$i$i;
    $71 = (($70) - ($67))|0;
    $72 = (($71>>>0) / 20)&-1;
    $73 = $72 ^ -1;
    $scevgep4$i$i$i = (($68) + (($73*20)|0)|0);
    HEAP32[$5>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($65); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($eh$lpad$body$indexZ2D|0);
   // unreachable;
  }
 } while(0);
 $133 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 ;HEAP32[$color$byval_copy>>2]=HEAP32[$color>>2]|0;HEAP32[$color$byval_copy+4>>2]=HEAP32[$color+4>>2]|0;HEAP32[$color$byval_copy+8>>2]=HEAP32[$color+8>>2]|0;HEAP32[$color$byval_copy+12>>2]=HEAP32[$color+12>>2]|0; //@line 104 "modules/graphics/opengl/wrap_Graphics.cpp"
 invoke_vii(3530,($133|0),($color$byval_copy|0)); //@line 104 "modules/graphics/opengl/wrap_Graphics.cpp"
 $134 = __THREW__; __THREW__ = 0;
 $135 = $134&1;
 if (!($135)) {
  STACKTOP = sp;return 0; //@line 106 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $136 = ___cxa_find_matching_catch_3(8992|0)|0;
 $137 = tempRet0;
 $138 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $139 = ($137|0)==($138|0); //@line 535 "./common/runtime.h"
 if (!($139)) {
  $$01$i = $137;$$02$i = $136;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $140 = (___cxa_begin_catch(($136|0))|0); //@line 535 "./common/runtime.h"
 $141 = HEAP32[$140>>2]|0; //@line 539 "./common/runtime.h"
 $142 = ((($141)) + 8|0); //@line 539 "./common/runtime.h"
 $143 = HEAP32[$142>>2]|0; //@line 539 "./common/runtime.h"
 $144 = (FUNCTION_TABLE_ii[$143 & 8191]($140)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($144|0)); //@line 539 "./common/runtime.h"
 $145 = __THREW__; __THREW__ = 0;
 $146 = $145&1;
 if (!($146)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $147 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $147; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 106 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $148 = ___cxa_find_matching_catch_2()|0;
 $149 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $150 = __THREW__; __THREW__ = 0;
 $151 = $150&1;
 if ($151) {
  $152 = ___cxa_find_matching_catch_3(0|0)|0;
  $153 = tempRet0;
  ___clang_call_terminate($152); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $149;$$02$i = $148;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl9w_discardEP9lua_State($L) {
 $L = $L|0;
 var $$0$i$i = 0, $$cast = 0, $$pre = 0, $$pre$i = 0, $$pre1$i = 0, $$pre7 = 0, $$pre8 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $colorbuffers = 0, $i$0 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, $storemerge$i$i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $colorbuffers = sp + 24|0;
 $0 = sp + 12|0;
 $1 = sp;
 HEAP32[$colorbuffers>>2] = 0; //@line 2563 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($colorbuffers)) + 4|0); //@line 2564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($colorbuffers)) + 8|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$3>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 $4 = (invoke_iii(3358,($L|0),1)|0); //@line 112 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 L1: do {
  if ($6) {
   label = 45;
  } else {
   $7 = ($4|0)==(5); //@line 112 "modules/graphics/opengl/wrap_Graphics.cpp"
   L3: do {
    if ($7) {
     $i$0 = 1;
     while(1) {
      __THREW__ = 0;
      $8 = (invoke_iii(3381,($L|0),1)|0); //@line 114 "modules/graphics/opengl/wrap_Graphics.cpp"
      $9 = __THREW__; __THREW__ = 0;
      $10 = $9&1;
      if ($10) {
       break;
      }
      $11 = ($i$0>>>0)>($8>>>0); //@line 114 "modules/graphics/opengl/wrap_Graphics.cpp"
      if ($11) {
       break L3;
      }
      __THREW__ = 0;
      invoke_viii(3382,($L|0),1,($i$0|0)); //@line 116 "modules/graphics/opengl/wrap_Graphics.cpp"
      $12 = __THREW__; __THREW__ = 0;
      $13 = $12&1;
      if ($13) {
       break;
      }
      __THREW__ = 0;
      $14 = (invoke_iiii(3474,($L|0),-1,1)|0); //@line 117 "modules/graphics/opengl/wrap_Graphics.cpp"
      $15 = __THREW__; __THREW__ = 0;
      $16 = $15&1;
      if ($16) {
       break;
      }
      $17 = HEAP32[$2>>2]|0; //@line 2994 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $18 = HEAP32[$3>>2]|0; //@line 2252 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $19 = $18 << 5; //@line 2166 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $20 = ($17|0)==($19|0); //@line 2994 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($20) {
       $21 = (($17) + 1)|0; //@line 2995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $22 = ($21|0)<(0); //@line 2523 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($22) {
        __THREW__ = 0;
        invoke_vi(3528,($colorbuffers|0)); //@line 2524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $23 = __THREW__; __THREW__ = 0;
        $24 = $23&1;
        if ($24) {
         break;
        }
        $$pre$i = HEAP32[$3>>2]|0; //@line 2252 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $26 = $$pre$i;
       } else {
        $26 = $18;
       }
       $25 = $26 << 5; //@line 2166 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $27 = ($25>>>0)>(1073741822); //@line 2526 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($27) {
        $$0$i$i = 2147483647;
       } else {
        $28 = $26 << 6; //@line 2528 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $29 = (($17) + 32)|0; //@line 2371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $30 = $29 & -32; //@line 2371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $31 = ($28>>>0)<($30>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $32 = $31 ? $30 : $28; //@line 2528 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $$0$i$i = $32;
       }
       __THREW__ = 0;
       invoke_vii(3531,($colorbuffers|0),($$0$i$i|0)); //@line 2995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $33 = __THREW__; __THREW__ = 0;
       $34 = $33&1;
       if ($34) {
        break;
       }
       $$pre1$i = HEAP32[$2>>2]|0; //@line 2996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $36 = $$pre1$i;
      } else {
       $36 = $17;
      }
      $35 = (($36) + 1)|0; //@line 2996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$2>>2] = $35; //@line 2996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $37 = HEAP32[$colorbuffers>>2]|0; //@line 2384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $38 = $36 >>> 5; //@line 2384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $39 = (($37) + ($38<<2)|0); //@line 2384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $40 = $36 & 31; //@line 2384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $41 = 1 << $40; //@line 2384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($14) {
       $42 = HEAP32[$39>>2]|0; //@line 60 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
       $43 = $42 | $41; //@line 60 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
       $storemerge$i$i = $43;
      } else {
       $44 = $41 ^ -1; //@line 62 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
       $45 = HEAP32[$39>>2]|0; //@line 62 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
       $46 = $45 & $44; //@line 62 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
       $storemerge$i$i = $46;
      }
      HEAP32[$39>>2] = $storemerge$i$i; //@line 60 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
      __THREW__ = 0;
      invoke_vii(3384,($L|0),-2); //@line 118 "modules/graphics/opengl/wrap_Graphics.cpp"
      $47 = __THREW__; __THREW__ = 0;
      $48 = $47&1;
      if ($48) {
       break;
      }
      $49 = (($i$0) + 1)|0; //@line 114 "modules/graphics/opengl/wrap_Graphics.cpp"
      $i$0 = $49;
     }
     $117 = ___cxa_find_matching_catch_2()|0;
     $118 = tempRet0;
     $lpad$phi$index = $117;$lpad$phi$index2 = $118;
     break L1;
    } else {
     __THREW__ = 0;
     $50 = (invoke_iiii(3474,($L|0),1,1)|0); //@line 123 "modules/graphics/opengl/wrap_Graphics.cpp"
     $51 = __THREW__; __THREW__ = 0;
     $52 = $51&1;
     if ($52) {
      label = 45;
      break L1;
     }
     $53 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
     __THREW__ = 0;
     invoke_vii(3532,($0|0),($53|0)); //@line 124 "modules/graphics/opengl/wrap_Graphics.cpp"
     $54 = __THREW__; __THREW__ = 0;
     $55 = $54&1;
     if ($55) {
      label = 45;
      break L1;
     }
     $56 = ((($0)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $57 = HEAP32[$56>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $58 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $59 = (($57) - ($58))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $60 = $59 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $61 = ($60>>>0)>(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $62 = $61 ? $60 : 1; //@line 124 "modules/graphics/opengl/wrap_Graphics.cpp"
     $63 = ($58|0)==(0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $64 = $58; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $65 = $57; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($63)) {
      $$cast = $58; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $66 = ($65|0)==($$cast|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($66)) {
       $scevgep$i$i$i = ((($65)) + -4|0);
       $67 = $scevgep$i$i$i;
       $68 = (($67) - ($58))|0;
       $69 = $68 >>> 2;
       $70 = $69 ^ -1;
       $scevgep4$i$i$i = (($65) + ($70<<2)|0);
       HEAP32[$56>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($64); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     HEAP32[$1>>2] = 0; //@line 2608 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $71 = ((($1)) + 4|0); //@line 2609 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$71>>2] = 0; //@line 2609 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $72 = ((($1)) + 8|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$72>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $73 = ($62|0)==(0); //@line 2612 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     do {
      if ($73) {
       $123 = 0;$124 = 0;$125 = 0;
      } else {
       $74 = ($62|0)<(0); //@line 2484 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($74) {
        __THREW__ = 0;
        invoke_vi(3528,($1|0)); //@line 2485 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $75 = __THREW__; __THREW__ = 0;
        $76 = $75&1;
        if ($76) {
         label = 45;
         break L1;
        }
       }
       $77 = (($62) + -1)|0; //@line 2169 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $78 = $77 >>> 5; //@line 2169 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $79 = (($78) + 1)|0; //@line 2169 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $80 = $79 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       __THREW__ = 0;
       $81 = (invoke_ii(3348,($80|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       $82 = __THREW__; __THREW__ = 0;
       $83 = $82&1;
       if ($83) {
        label = 45;
        break L1;
       }
       HEAP32[$1>>2] = $81; //@line 2487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$72>>2] = $79; //@line 2489 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$71>>2] = $62; //@line 2541 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $84 = $62 >>> 5; //@line 339 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
       $85 = $84 << 2; //@line 340 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
       $86 = $81; //@line 388 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
       if ($50) {
        _memset(($81|0),-1,($85|0))|0; //@line 370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $87 = $84 << 5; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $88 = ($87|0)==($62|0); //@line 373 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        if ($88) {
         $123 = $79;$124 = $62;$125 = $86;
         break;
        }
        $89 = (($81) + ($84<<2)|0); //@line 375 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $90 = (32 - ($62))|0; //@line 376 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $91 = (($90) + ($87))|0; //@line 376 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $92 = -1 >>> $91; //@line 376 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $93 = HEAP32[$89>>2]|0; //@line 377 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $94 = $93 | $92; //@line 377 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        HEAP32[$89>>2] = $94; //@line 377 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $123 = $79;$124 = $62;$125 = $86;
        break;
       } else {
        _memset(($81|0),0,($85|0))|0; //@line 340 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $95 = $84 << 5; //@line 341 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $96 = ($95|0)==($62|0); //@line 343 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        if ($96) {
         $123 = $79;$124 = $62;$125 = $86;
         break;
        }
        $97 = (($81) + ($84<<2)|0); //@line 345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $98 = (32 - ($62))|0; //@line 346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $99 = (($98) + ($95))|0; //@line 346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $100 = -1 >>> $99; //@line 346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $101 = $100 ^ -1; //@line 347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $102 = HEAP32[$97>>2]|0; //@line 347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $103 = $102 & $101; //@line 347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        HEAP32[$97>>2] = $103; //@line 347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $123 = $79;$124 = $62;$125 = $86;
        break;
       }
      }
     } while(0);
     $104 = HEAP32[$colorbuffers>>2]|0; //@line 2496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $105 = ($104|0)==(0|0); //@line 2496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($105) {
      $106 = $125;$107 = $124;$108 = $123;
     } else {
      __ZdlPv($104); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      HEAP32[$colorbuffers>>2] = 0; //@line 2500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$3>>2] = 0; //@line 2501 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$2>>2] = 0; //@line 2501 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$pre = HEAP32[$1>>2]|0; //@line 2871 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$pre7 = HEAP32[$71>>2]|0; //@line 2872 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$pre8 = HEAP32[$72>>2]|0; //@line 2873 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $106 = $$pre;$107 = $$pre7;$108 = $$pre8;
     }
     HEAP32[$colorbuffers>>2] = $106; //@line 2871 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$2>>2] = $107; //@line 2872 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$3>>2] = $108; //@line 2873 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$1>>2] = 0; //@line 2874 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$71>>2] = 0; //@line 2875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$72>>2] = 0; //@line 2875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
   } while(0);
   __THREW__ = 0;
   $109 = (invoke_iiii(3474,($L|0),2,1)|0); //@line 128 "modules/graphics/opengl/wrap_Graphics.cpp"
   $110 = __THREW__; __THREW__ = 0;
   $111 = $110&1;
   if ($111) {
    label = 45;
   } else {
    $112 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
    __THREW__ = 0;
    invoke_viii(3533,($112|0),($colorbuffers|0),($109|0)); //@line 129 "modules/graphics/opengl/wrap_Graphics.cpp"
    $113 = __THREW__; __THREW__ = 0;
    $114 = $113&1;
    if ($114) {
     label = 45;
    } else {
     $115 = HEAP32[$colorbuffers>>2]|0; //@line 2753 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $116 = ($115|0)==(0|0); //@line 2753 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($116) {
      STACKTOP = sp;return 0; //@line 131 "modules/graphics/opengl/wrap_Graphics.cpp"
     }
     __ZdlPv($115); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     STACKTOP = sp;return 0; //@line 131 "modules/graphics/opengl/wrap_Graphics.cpp"
    }
   }
  }
 } while(0);
 if ((label|0) == 45) {
  $119 = ___cxa_find_matching_catch_2()|0;
  $120 = tempRet0;
  $lpad$phi$index = $119;$lpad$phi$index2 = $120;
 }
 $121 = HEAP32[$colorbuffers>>2]|0; //@line 2753 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $122 = ($121|0)==(0|0); //@line 2753 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($122) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 __ZdlPv($121); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__16vectorIbNS_9allocatorIbEEE7reserveEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__v = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__v = sp + 16|0;
 $0 = sp + 8|0;
 $1 = sp;
 $2 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 2252 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 << 5; //@line 2166 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($4>>>0)<($__n>>>0); //@line 2942 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($5)) {
  STACKTOP = sp;return; //@line 2950 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 HEAP32[$__v>>2] = 0; //@line 2572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ((($__v)) + 4|0); //@line 2573 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$6>>2] = 0; //@line 2573 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ((($__v)) + 8|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$7>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = ($__n|0)<(0); //@line 2484 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($8) {
  __THREW__ = 0;
  invoke_vi(3528,($__v|0)); //@line 2485 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if (!($10)) {
   label = 4;
  }
 } else {
  label = 4;
 }
 if ((label|0) == 4) {
  $11 = (($__n) + -1)|0; //@line 2169 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $12 = $11 >>> 5; //@line 2169 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) + 1)|0; //@line 2169 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  __THREW__ = 0;
  $15 = (invoke_ii(3348,($14|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $16 = __THREW__; __THREW__ = 0;
  $17 = $16&1;
  if (!($17)) {
   HEAP32[$__v>>2] = $15; //@line 2487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$6>>2] = 0; //@line 2488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$7>>2] = $13; //@line 2489 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $18 = HEAP32[$this>>2]|0; //@line 2390 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$0>>2] = $18; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $19 = ((($0)) + 4|0); //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$19>>2] = 0; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $20 = ((($this)) + 4|0); //@line 2270 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $21 = HEAP32[$20>>2]|0; //@line 2270 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $22 = $18; //@line 2390 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $23 = $21 >>> 5; //@line 2390 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $24 = (($22) + ($23<<2)|0); //@line 2390 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $25 = $21 & 31; //@line 2390 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$1>>2] = $24; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $26 = ((($1)) + 4|0); //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$26>>2] = $25; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   __ZNSt3__16vectorIbNS_9allocatorIbEEE18__construct_at_endINS_14__bit_iteratorIS3_Lb0ELj0EEEEENS_9enable_ifIXsr21__is_forward_iteratorIT_EE5valueEvE4typeES8_S8_($__v,$0,$1); //@line 2946 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $27 = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $28 = HEAP32[$__v>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$this>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$__v>>2] = $27; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $29 = HEAP32[$20>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $30 = HEAP32[$6>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$20>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$6>>2] = $29; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $31 = HEAP32[$2>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $32 = HEAP32[$7>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$2>>2] = $32; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$7>>2] = $31; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $33 = ($27|0)==(0); //@line 2753 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($33)) {
    $34 = $27; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    __ZdlPv($34); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   STACKTOP = sp;return; //@line 2950 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
 }
 $35 = ___cxa_find_matching_catch_2()|0;
 $36 = tempRet0;
 $37 = HEAP32[$__v>>2]|0; //@line 2753 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $38 = ($37|0)==(0|0); //@line 2753 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($38) {
  ___resumeException($35|0);
  // unreachable;
 }
 __ZdlPv($37); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($35|0);
 // unreachable;
}
function __ZNSt3__16vectorIbNS_9allocatorIbEEE18__construct_at_endINS_14__bit_iteratorIS3_Lb0ELj0EEEEENS_9enable_ifIXsr21__is_forward_iteratorIT_EE5valueEvE4typeES8_S8_($this,$__first,$__last) {
 $this = $this|0;
 $__first = $__first|0;
 $__last = $__last|0;
 var $$in = 0, $$sroa$7$0$i = 0, $$sroa$7$1$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $8 = 0, $9 = 0, $__n$0$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 24|0;
 $1 = sp + 16|0;
 $2 = sp + 8|0;
 $3 = sp;
 $4 = ((($this)) + 4|0); //@line 2554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$4>>2]|0; //@line 2554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$__first>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $7 = ((($__first)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $8 = HEAP32[$7>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $9 = HEAP32[$__last>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $10 = ((($__last)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = HEAP32[$10>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = (($9) - ($6))|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $13 = $12 << 3; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $14 = (($5) - ($8))|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $15 = (($14) + ($11))|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $16 = (($15) + ($13))|0; //@line 2555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$4>>2] = $16; //@line 2555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = HEAP32[$this>>2]|0; //@line 2390 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $18 = $5 >>> 5; //@line 2390 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $19 = (($17) + ($18<<2)|0); //@line 2390 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = $5 & 31; //@line 2390 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = $19; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $22 = ($8|0)==($20|0); //@line 539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($22)) {
  HEAP32[$0>>2] = $6; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $68 = ((($0)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$68>>2] = $8; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$1>>2] = $9; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $69 = ((($1)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$69>>2] = $11; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$2>>2] = $21; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $70 = ((($2)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$70>>2] = $20; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  __ZNSt3__116__copy_unalignedINS_6vectorIbNS_9allocatorIbEEEELb0EEENS_14__bit_iteratorIT_Lb0EXLi0EEEENS5_IS6_XT0_EXLi0EEEES8_S7_($3,$0,$1,$2); //@line 541 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  STACKTOP = sp;return; //@line 2557 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $23 = (($11) - ($8))|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $24 = (($23) + ($13))|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $25 = ($24|0)>(0); //@line 417 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $26 = $6; //@line 426 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if ($25) {
  $27 = ($8|0)==(0); //@line 420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  if ($27) {
   $28 = $6; //@line 417 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $$in = $19;$$sroa$7$0$i = 0;$50 = $28;$57 = $26;$__n$0$i$i = $24;
  } else {
   $29 = (32 - ($8))|0; //@line 422 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $30 = ($24|0)<($29|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $31 = $30 ? $24 : $29; //@line 423 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $32 = (($24) - ($31))|0; //@line 424 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $33 = -1 << $8; //@line 425 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $34 = (($29) - ($31))|0; //@line 425 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $35 = -1 >>> $34; //@line 425 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $36 = $35 & $33; //@line 425 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $37 = HEAP32[$26>>2]|0; //@line 426 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $38 = $37 & $36; //@line 426 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $39 = $36 ^ -1; //@line 427 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $40 = HEAP32[$19>>2]|0; //@line 427 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $41 = $40 & $39; //@line 427 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $42 = $41 | $38; //@line 428 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$19>>2] = $42; //@line 428 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $43 = (($31) + ($8))|0; //@line 429 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $44 = $43 >>> 5; //@line 429 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $45 = (($19) + ($44<<2)|0); //@line 429 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $46 = $43 & 31; //@line 430 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $47 = ((($26)) + 4|0); //@line 431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $$in = $45;$$sroa$7$0$i = $46;$50 = $47;$57 = $47;$__n$0$i$i = $32;
  }
  $48 = $__n$0$i$i >>> 5; //@line 436 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $49 = $48 << 2; //@line 439 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  _memmove(($$in|0),($50|0),($49|0))|0; //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $51 = $48 << 5; //@line 440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $52 = (($__n$0$i$i) - ($51))|0; //@line 440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $53 = (($$in) + ($48<<2)|0); //@line 448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $54 = ($52|0)>(0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $55 = $53; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  if ($54) {
   $56 = (($57) + ($48<<2)|0); //@line 445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $58 = (32 - ($52))|0; //@line 446 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $59 = -1 >>> $58; //@line 446 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $60 = HEAP32[$56>>2]|0; //@line 447 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $61 = $60 & $59; //@line 447 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $62 = $59 ^ -1; //@line 448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $63 = HEAP32[$53>>2]|0; //@line 448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $64 = $63 & $62; //@line 448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $65 = $64 | $61; //@line 449 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$53>>2] = $65; //@line 449 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $$sroa$7$1$i = $52;$66 = $55;
  } else {
   $$sroa$7$1$i = $$sroa$7$0$i;$66 = $55;
  }
 } else {
  $$sroa$7$1$i = $8;$66 = $21;
 }
 HEAP32[$3>>2] = $66; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $67 = ((($3)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$67>>2] = $$sroa$7$1$i; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 STACKTOP = sp;return; //@line 2557 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNSt3__116__copy_unalignedINS_6vectorIbNS_9allocatorIbEEEELb0EEENS_14__bit_iteratorIT_Lb0EXLi0EEEENS5_IS6_XT0_EXLi0EEEES8_S7_($agg$result,$__first,$__last,$__result) {
 $agg$result = $agg$result|0;
 $__first = $__first|0;
 $__last = $__last|0;
 $__result = $__result|0;
 var $$lcssa = 0, $$phi$trans$insert = 0, $$pn = 0, $$pre = 0, $$pre$phi5Z2D = 0, $$pre$phiZ2D = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__n$0 = 0, $__n$1$lcssa = 0, $__n$11 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$__last>>2]|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $1 = HEAP32[$__first>>2]|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $2 = (($0) - ($1))|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $3 = $2 << 3; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $4 = ((($__last)) + 4|0); //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $5 = HEAP32[$4>>2]|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $6 = (($3) + ($5))|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $7 = ((($__first)) + 4|0); //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $8 = HEAP32[$7>>2]|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $9 = (($6) - ($8))|0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $10 = ($9|0)>(0); //@line 466 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $1; //@line 475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($10)) {
  $$pre4 = ((($__result)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $$pre$phi5Z2D = $$pre4;
  $118 = HEAP32[$__result>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$agg$result>>2] = $118; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $119 = ((($agg$result)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $120 = HEAP32[$$pre$phi5Z2D>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$119>>2] = $120; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  return; //@line 532 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 }
 $12 = ($8|0)==(0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if ($12) {
  $$phi$trans$insert = ((($__result)) + 4|0); //@line 499 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 499 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $$pre$phiZ2D = $$phi$trans$insert;$121 = $11;$62 = $$pre;$__n$0 = $9;
 } else {
  $13 = (32 - ($8))|0; //@line 471 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $14 = ($9|0)<($13|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $15 = $14 ? $9 : $13; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $16 = (($9) - ($15))|0; //@line 473 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $17 = -1 << $8; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $18 = (($13) - ($15))|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $19 = -1 >>> $18; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $20 = $19 & $17; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $21 = HEAP32[$11>>2]|0; //@line 475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $22 = $20 & $21; //@line 475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $23 = ((($__result)) + 4|0); //@line 476 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $24 = HEAP32[$23>>2]|0; //@line 476 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $25 = (32 - ($24))|0; //@line 476 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $26 = ($25>>>0)<($15>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $27 = $26 ? $25 : $15; //@line 477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $28 = -1 << $24; //@line 478 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $29 = (($25) - ($27))|0; //@line 478 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $30 = -1 >>> $29; //@line 478 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $31 = $30 & $28; //@line 478 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $32 = $31 ^ -1; //@line 479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $33 = HEAP32[$__result>>2]|0; //@line 484 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $34 = HEAP32[$33>>2]|0; //@line 479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $35 = $34 & $32; //@line 481 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$33>>2] = $35; //@line 479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $36 = HEAP32[$23>>2]|0; //@line 480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $37 = HEAP32[$7>>2]|0; //@line 480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $38 = ($36>>>0)>($37>>>0); //@line 480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $39 = (($37) - ($36))|0; //@line 483 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $40 = $22 >>> $39; //@line 483 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $41 = (($36) - ($37))|0; //@line 481 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $42 = $22 << $41; //@line 481 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $$pn = $38 ? $42 : $40; //@line 480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $storemerge = $35 | $$pn; //@line 483 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$33>>2] = $storemerge; //@line 481 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $43 = HEAP32[$23>>2]|0; //@line 484 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $44 = (($43) + ($27))|0; //@line 484 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $45 = $44 >>> 5; //@line 484 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $46 = (($33) + ($45<<2)|0); //@line 491 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$__result>>2] = $46; //@line 484 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $47 = $44 & 31; //@line 485 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$23>>2] = $47; //@line 485 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $48 = (($15) - ($27))|0; //@line 486 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $49 = ($48|0)>(0); //@line 487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  if ($49) {
   $50 = (32 - ($48))|0; //@line 489 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $51 = -1 >>> $50; //@line 489 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $52 = $51 ^ -1; //@line 490 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $53 = HEAP32[$46>>2]|0; //@line 490 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $54 = $53 & $52; //@line 490 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$46>>2] = $54; //@line 490 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $55 = HEAP32[$7>>2]|0; //@line 491 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $56 = (($55) + ($27))|0; //@line 491 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $57 = $22 >>> $56; //@line 491 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $58 = $54 | $57; //@line 491 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$46>>2] = $58; //@line 491 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$23>>2] = $48; //@line 492 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $122 = $48;
  } else {
   $122 = $47;
  }
  $59 = HEAP32[$__first>>2]|0; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $60 = ((($59)) + 4|0); //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$__first>>2] = $60; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $$pre$phiZ2D = $23;$121 = $60;$62 = $122;$__n$0 = $16;
 }
 $61 = (32 - ($62))|0; //@line 499 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $63 = -1 << $62; //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $64 = ($__n$0>>>0)>(31); //@line 501 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if ($64) {
  $65 = $63 ^ -1; //@line 504 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $67 = $121;$__n$11 = $__n$0;
  while(1) {
   $66 = HEAP32[$67>>2]|0; //@line 503 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $68 = HEAP32[$__result>>2]|0; //@line 504 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $69 = HEAP32[$68>>2]|0; //@line 504 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $70 = $69 & $65; //@line 504 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$68>>2] = $70; //@line 504 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $71 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $72 = $66 << $71; //@line 505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $73 = $70 | $72; //@line 505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$68>>2] = $73; //@line 505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $74 = ((($68)) + 4|0); //@line 506 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$__result>>2] = $74; //@line 506 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $75 = HEAP32[$74>>2]|0; //@line 507 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $76 = $75 & $63; //@line 507 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $77 = $66 >>> $61; //@line 508 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $78 = $76 | $77; //@line 508 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$74>>2] = $78; //@line 508 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $79 = (($__n$11) + -32)|0; //@line 501 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $80 = HEAP32[$__first>>2]|0; //@line 501 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $81 = ((($80)) + 4|0); //@line 501 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   HEAP32[$__first>>2] = $81; //@line 501 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   $82 = ($79>>>0)>(31); //@line 501 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
   if ($82) {
    $67 = $81;$__n$11 = $79;
   } else {
    $$lcssa = $81;
    break;
   }
  }
  $83 = $__n$0 & 31; //@line 501 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $88 = $$lcssa;$__n$1$lcssa = $83;
 } else {
  $88 = $121;$__n$1$lcssa = $__n$0;
 }
 $84 = ($__n$1$lcssa|0)>(0); //@line 511 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($84)) {
  $$pre$phi5Z2D = $$pre$phiZ2D;
  $118 = HEAP32[$__result>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$agg$result>>2] = $118; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $119 = ((($agg$result)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $120 = HEAP32[$$pre$phi5Z2D>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$119>>2] = $120; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  return; //@line 532 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 }
 $85 = (32 - ($__n$1$lcssa))|0; //@line 513 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $86 = -1 >>> $85; //@line 513 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $87 = HEAP32[$88>>2]|0; //@line 514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $89 = $87 & $86; //@line 514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $90 = ($61|0)<($__n$1$lcssa|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $91 = $90 ? $61 : $__n$1$lcssa; //@line 515 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $92 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 516 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $93 = -1 << $92; //@line 516 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $94 = (($61) - ($91))|0; //@line 516 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $95 = -1 >>> $94; //@line 516 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $96 = $93 & $95; //@line 516 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $97 = $96 ^ -1; //@line 517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $98 = HEAP32[$__result>>2]|0; //@line 517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $99 = HEAP32[$98>>2]|0; //@line 517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $100 = $99 & $97; //@line 517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$98>>2] = $100; //@line 517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $101 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 518 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $102 = $89 << $101; //@line 518 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $103 = $100 | $102; //@line 518 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$98>>2] = $103; //@line 518 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $104 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $105 = (($104) + ($91))|0; //@line 519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $106 = $105 >>> 5; //@line 519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $107 = (($98) + ($106<<2)|0); //@line 526 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$__result>>2] = $107; //@line 519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $108 = $105 & 31; //@line 520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$$pre$phiZ2D>>2] = $108; //@line 520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $109 = (($__n$1$lcssa) - ($91))|0; //@line 521 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $110 = ($109|0)>(0); //@line 522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($110)) {
  $$pre$phi5Z2D = $$pre$phiZ2D;
  $118 = HEAP32[$__result>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$agg$result>>2] = $118; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $119 = ((($agg$result)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $120 = HEAP32[$$pre$phi5Z2D>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  HEAP32[$119>>2] = $120; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  return; //@line 532 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 }
 $111 = (32 - ($109))|0; //@line 524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $112 = -1 >>> $111; //@line 524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $113 = $112 ^ -1; //@line 525 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $114 = HEAP32[$107>>2]|0; //@line 525 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $115 = $114 & $113; //@line 525 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $116 = $89 >>> $91; //@line 526 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $117 = $115 | $116; //@line 526 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$107>>2] = $117; //@line 526 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$$pre$phiZ2D>>2] = $109; //@line 527 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $$pre$phi5Z2D = $$pre$phiZ2D;
 $118 = HEAP32[$__result>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$agg$result>>2] = $118; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $119 = ((($agg$result)) + 4|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $120 = HEAP32[$$pre$phi5Z2D>>2]|0; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 HEAP32[$119>>2] = $120; //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 return; //@line 532 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
}
function __ZN4love8graphics6opengl9w_presentEP9lua_State($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZN4love8graphics6opengl8Graphics7presentEv($1); //@line 135 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 0; //@line 136 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl11w_isCreatedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZNK4love8graphics6opengl8Graphics9isCreatedEv($0)|0); //@line 141 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$1); //@line 141 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 1; //@line 142 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl10w_isActiveEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 147 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = ((($1)) + 40|0); //@line 147 "modules/graphics/opengl/wrap_Graphics.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 147 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 147 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 147 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 1; //@line 148 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl16w_isGammaCorrectEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZNK4love8graphics6opengl8Graphics14isGammaCorrectEv($0)|0); //@line 153 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$1); //@line 153 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 1; //@line 154 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl10w_getWidthEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZNK4love8graphics6opengl8Graphics8getWidthEv($0)|0); //@line 159 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushinteger($L,$1); //@line 159 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 1; //@line 160 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl11w_getHeightEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZNK4love8graphics6opengl8Graphics9getHeightEv($0)|0); //@line 165 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushinteger($L,$1); //@line 165 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 1; //@line 166 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl15w_getDimensionsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZNK4love8graphics6opengl8Graphics8getWidthEv($0)|0); //@line 171 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushinteger($L,$1); //@line 171 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $3 = (__ZNK4love8graphics6opengl8Graphics9getHeightEv($2)|0); //@line 172 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushinteger($L,$3); //@line 172 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 2; //@line 173 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl12w_setScissorEP9lua_State($L) {
 $L = $L|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 176 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = (_lua_gettop($L)|0); //@line 178 "modules/graphics/opengl/wrap_Graphics.cpp"
 switch ($0|0) {
 case 0:  {
  label = 6;
  break;
 }
 case 4:  {
  $1 = (_lua_type($L,1)|0); //@line 180 "modules/graphics/opengl/wrap_Graphics.cpp"
  $2 = ($1|0)==(0); //@line 180 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($2) {
   $3 = (_lua_type($L,2)|0); //@line 180 "modules/graphics/opengl/wrap_Graphics.cpp"
   $4 = ($3|0)==(0); //@line 180 "modules/graphics/opengl/wrap_Graphics.cpp"
   if ($4) {
    $5 = (_lua_type($L,3)|0); //@line 181 "modules/graphics/opengl/wrap_Graphics.cpp"
    $6 = ($5|0)==(0); //@line 181 "modules/graphics/opengl/wrap_Graphics.cpp"
    if ($6) {
     $7 = (_lua_type($L,4)|0); //@line 181 "modules/graphics/opengl/wrap_Graphics.cpp"
     $8 = ($7|0)==(0); //@line 181 "modules/graphics/opengl/wrap_Graphics.cpp"
     if ($8) {
      label = 6;
     }
    }
   }
  }
  break;
 }
 default: {
 }
 }
 if ((label|0) == 6) {
  $9 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  __ZN4love8graphics6opengl8Graphics10setScissorEv($9); //@line 183 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 197 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $10 = (+_luaL_checknumber($L,1)); //@line 187 "modules/graphics/opengl/wrap_Graphics.cpp"
 $11 = (+_luaL_checknumber($L,2)); //@line 188 "modules/graphics/opengl/wrap_Graphics.cpp"
 $12 = (+_luaL_checknumber($L,3)); //@line 189 "modules/graphics/opengl/wrap_Graphics.cpp"
 $13 = (~~(($12))); //@line 189 "modules/graphics/opengl/wrap_Graphics.cpp"
 $14 = (+_luaL_checknumber($L,4)); //@line 190 "modules/graphics/opengl/wrap_Graphics.cpp"
 $15 = (~~(($14))); //@line 190 "modules/graphics/opengl/wrap_Graphics.cpp"
 $16 = $15 | $13; //@line 192 "modules/graphics/opengl/wrap_Graphics.cpp"
 $17 = ($16|0)<(0); //@line 192 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($17) {
  $18 = (_luaL_error($L,329241,$vararg_buffer)|0); //@line 193 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$1 = $18;
  STACKTOP = sp;return ($$1|0); //@line 197 "modules/graphics/opengl/wrap_Graphics.cpp"
 } else {
  $19 = (~~(($11))); //@line 188 "modules/graphics/opengl/wrap_Graphics.cpp"
  $20 = (~~(($10))); //@line 187 "modules/graphics/opengl/wrap_Graphics.cpp"
  $21 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  __ZN4love8graphics6opengl8Graphics10setScissorEiiii($21,$20,$19,$13,$15); //@line 195 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 197 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl18w_intersectScissorEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0.0, $1 = 0.0, $10 = 0, $11 = 0, $2 = 0.0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 199 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = (+_luaL_checknumber($L,1)); //@line 201 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 202 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = (+_luaL_checknumber($L,3)); //@line 203 "modules/graphics/opengl/wrap_Graphics.cpp"
 $3 = (~~(($2))); //@line 203 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = (+_luaL_checknumber($L,4)); //@line 204 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = (~~(($4))); //@line 204 "modules/graphics/opengl/wrap_Graphics.cpp"
 $6 = $5 | $3; //@line 206 "modules/graphics/opengl/wrap_Graphics.cpp"
 $7 = ($6|0)<(0); //@line 206 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($7) {
  $8 = (_luaL_error($L,329241,$vararg_buffer)|0); //@line 207 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = $8;
  STACKTOP = sp;return ($$0|0); //@line 211 "modules/graphics/opengl/wrap_Graphics.cpp"
 } else {
  $9 = (~~(($1))); //@line 202 "modules/graphics/opengl/wrap_Graphics.cpp"
  $10 = (~~(($0))); //@line 201 "modules/graphics/opengl/wrap_Graphics.cpp"
  $11 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  __ZN4love8graphics6opengl8Graphics16intersectScissorEiiii($11,$10,$9,$3,$5); //@line 209 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 211 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl12w_getScissorEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $h = 0, $w = 0, $x = 0, $y = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $x = sp + 12|0;
 $y = sp + 8|0;
 $w = sp + 4|0;
 $h = sp;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZNK4love8graphics6opengl8Graphics10getScissorERiS3_S3_S3_($0,$x,$y,$w,$h)|0); //@line 216 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($1)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 225 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $2 = HEAP32[$x>>2]|0; //@line 219 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushinteger($L,$2); //@line 219 "modules/graphics/opengl/wrap_Graphics.cpp"
 $3 = HEAP32[$y>>2]|0; //@line 220 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushinteger($L,$3); //@line 220 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = HEAP32[$w>>2]|0; //@line 221 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushinteger($L,$4); //@line 221 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = HEAP32[$h>>2]|0; //@line 222 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushinteger($L,$5); //@line 222 "modules/graphics/opengl/wrap_Graphics.cpp"
 $$0 = 4;
 STACKTOP = sp;return ($$0|0); //@line 225 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl9w_stencilEP9lua_State($L) {
 $L = $L|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $action = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $action = sp + 4|0;
 _luaL_checktype($L,1,6); //@line 229 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAP32[$action>>2] = 0; //@line 231 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = (_lua_type($L,2)|0); //@line 233 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = ($0|0)<(1); //@line 233 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($1)) {
  $2 = (_luaL_checklstring($L,2,0)|0); //@line 235 "modules/graphics/opengl/wrap_Graphics.cpp"
  $3 = (__ZN4love8graphics8Graphics11getConstantEPKcRNS1_13StencilActionE($2,$action)|0); //@line 236 "modules/graphics/opengl/wrap_Graphics.cpp"
  if (!($3)) {
   HEAP32[$vararg_buffer>>2] = $2; //@line 237 "modules/graphics/opengl/wrap_Graphics.cpp"
   $4 = (_luaL_error($L,329294,$vararg_buffer)|0); //@line 237 "modules/graphics/opengl/wrap_Graphics.cpp"
   $$1 = $4;
   STACKTOP = sp;return ($$1|0); //@line 254 "modules/graphics/opengl/wrap_Graphics.cpp"
  }
 }
 $5 = (+_luaL_optnumber($L,3,1.0)); //@line 240 "modules/graphics/opengl/wrap_Graphics.cpp"
 $6 = (~~(($5))); //@line 240 "modules/graphics/opengl/wrap_Graphics.cpp"
 $7 = (_lua_toboolean($L,4)|0); //@line 243 "modules/graphics/opengl/wrap_Graphics.cpp"
 $8 = ($7|0)==(0); //@line 243 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($8) {
  $9 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  __ZN4love8graphics6opengl8Graphics12clearStencilEv($9); //@line 244 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $10 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $11 = HEAP32[$action>>2]|0; //@line 246 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love8graphics6opengl8Graphics19drawToStencilBufferENS0_8Graphics13StencilActionEi($10,$11,$6); //@line 246 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushvalue($L,1); //@line 249 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_call($L,0,0); //@line 250 "modules/graphics/opengl/wrap_Graphics.cpp"
 $12 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZN4love8graphics6opengl8Graphics23stopDrawToStencilBufferEv($12); //@line 252 "modules/graphics/opengl/wrap_Graphics.cpp"
 $$1 = 0;
 STACKTOP = sp;return ($$1|0); //@line 254 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl16w_setStencilTestEP9lua_State($L) {
 $L = $L|0;
 var $$1 = 0, $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $compare = 0, $comparevalue$1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $compare = sp + 4|0;
 HEAP32[$compare>>2] = 6; //@line 259 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = (_lua_type($L,1)|0); //@line 262 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = ($0|0)<(1); //@line 262 "modules/graphics/opengl/wrap_Graphics.cpp"
 do {
  if ($1) {
   $8 = 6;$comparevalue$1 = 0;
  } else {
   $2 = (_luaL_checklstring($L,1,0)|0); //@line 264 "modules/graphics/opengl/wrap_Graphics.cpp"
   $3 = (__ZN4love8graphics8Graphics11getConstantEPKcRNS1_11CompareModeE($2,$compare)|0); //@line 265 "modules/graphics/opengl/wrap_Graphics.cpp"
   if ($3) {
    $4 = (+_luaL_checknumber($L,2)); //@line 268 "modules/graphics/opengl/wrap_Graphics.cpp"
    $5 = (~~(($4))); //@line 268 "modules/graphics/opengl/wrap_Graphics.cpp"
    $$pre = HEAP32[$compare>>2]|0; //@line 271 "modules/graphics/opengl/wrap_Graphics.cpp"
    $8 = $$pre;$comparevalue$1 = $5;
    break;
   }
   HEAP32[$vararg_buffer>>2] = $2; //@line 266 "modules/graphics/opengl/wrap_Graphics.cpp"
   $6 = (_luaL_error($L,329326,$vararg_buffer)|0); //@line 266 "modules/graphics/opengl/wrap_Graphics.cpp"
   $$1 = $6;
   STACKTOP = sp;return ($$1|0); //@line 273 "modules/graphics/opengl/wrap_Graphics.cpp"
  }
 } while(0);
 $7 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZN4love8graphics6opengl8Graphics14setStencilTestENS0_8Graphics11CompareModeEi($7,$8,$comparevalue$1); //@line 271 "modules/graphics/opengl/wrap_Graphics.cpp"
 $$1 = 0;
 STACKTOP = sp;return ($$1|0); //@line 273 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl16w_getStencilTestEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $compare = 0, $comparestr = 0, $comparevalue = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $compare = sp + 12|0;
 $comparevalue = sp + 8|0;
 $comparestr = sp + 4|0;
 HEAP32[$compare>>2] = 6; //@line 277 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAP32[$comparevalue>>2] = 1; //@line 278 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZN4love8graphics6opengl8Graphics14getStencilTestERNS0_8Graphics11CompareModeERi($0,$compare,$comparevalue); //@line 280 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = HEAP32[$compare>>2]|0; //@line 283 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = (__ZN4love8graphics8Graphics11getConstantENS1_11CompareModeERPKc($1,$comparestr)|0); //@line 283 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($2) {
  $4 = HEAP32[$comparestr>>2]|0; //@line 286 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_pushstring($L,$4); //@line 286 "modules/graphics/opengl/wrap_Graphics.cpp"
  $5 = HEAP32[$comparevalue>>2]|0; //@line 287 "modules/graphics/opengl/wrap_Graphics.cpp"
  $6 = (+($5|0)); //@line 287 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_pushnumber($L,$6); //@line 287 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = 2;
  STACKTOP = sp;return ($$0|0); //@line 289 "modules/graphics/opengl/wrap_Graphics.cpp"
 } else {
  $3 = (_luaL_error($L,329351,$vararg_buffer)|0); //@line 284 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 289 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl10w_newImageEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$01$i29 = 0, $$01$i49 = 0, $$02$i = 0, $$02$i28 = 0, $$02$i48 = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0;
 var $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0;
 var $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0;
 var $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0;
 var $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0;
 var $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0;
 var $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0;
 var $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0;
 var $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0;
 var $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0;
 var $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0;
 var $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0;
 var $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $__begin$sroa$0$02$i$i = 0, $__begin$sroa$0$02$i3$i = 0, $__begin2$sroa$0$01$i$i = 0, $__begin2$sroa$0$01$i5$i = 0, $cdata = 0, $data = 0, $eh$lpad$body$index21Z2D = 0, $eh$lpad$body$indexZ2D = 0, $flags = 0, $i$0 = 0, $image$0 = 0, $name$i = 0, $name$i10 = 0, $name$i39 = 0, $releasedata$1 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i12 = 0, $scevgep$i$i$i17 = 0, $scevgep$i$i$i7 = 0, $scevgep4$i$i$i = 0;
 var $scevgep4$i$i$i13 = 0, $scevgep4$i$i$i18 = 0, $scevgep4$i$i$i8 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer11 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer11 = sp + 40|0;
 $vararg_buffer8 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $name$i39 = sp + 100|0;
 $0 = sp + 96|0;
 $1 = sp + 92|0;
 $name$i10 = sp + 88|0;
 $name$i = sp + 84|0;
 $data = sp + 72|0;
 $cdata = sp + 60|0;
 $flags = sp + 104|0;
 $2 = sp + 56|0;
 $3 = sp + 52|0;
 $4 = sp + 48|0;
 $5 = sp + 44|0;
 $6 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $7 = (__ZNK4love8graphics6opengl8Graphics9isCreatedEv($6)|0); //@line 52 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($7)) {
  (_luaL_error($L,329373,$vararg_buffer)|0); //@line 53 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 HEAP32[$data>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($data)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$8>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ((($data)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$9>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$cdata>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = ((($cdata)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$10>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ((($cdata)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$11>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP8[$flags>>0] = 0; //@line 64 "modules/graphics/opengl/Image.h"
 $12 = ((($flags)) + 1|0); //@line 65 "modules/graphics/opengl/Image.h"
 HEAP8[$12>>0] = 0; //@line 65 "modules/graphics/opengl/Image.h"
 __THREW__ = 0;
 $13 = (invoke_iii(3358,($L|0),2)|0); //@line 306 "modules/graphics/opengl/wrap_Graphics.cpp"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 L4: do {
  if ($15) {
   label = 16;
  } else {
   $16 = ($13|0)<(1); //@line 306 "modules/graphics/opengl/wrap_Graphics.cpp"
   if (!($16)) {
    __THREW__ = 0;
    invoke_viii(3534,($L|0),2,5); //@line 308 "modules/graphics/opengl/wrap_Graphics.cpp"
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if ($18) {
     label = 16;
     break;
    }
    HEAP32[$name$i>>2] = 0; //@line 293 "modules/graphics/opengl/wrap_Graphics.cpp"
    __THREW__ = 0;
    (invoke_iii(3535,0,($name$i|0))|0); //@line 294 "modules/graphics/opengl/wrap_Graphics.cpp"
    $19 = __THREW__; __THREW__ = 0;
    $20 = $19&1;
    if ($20) {
     label = 16;
     break;
    }
    $21 = HEAP32[$name$i>>2]|0; //@line 295 "modules/graphics/opengl/wrap_Graphics.cpp"
    $22 = HEAP8[$flags>>0]|0; //@line 309 "modules/graphics/opengl/wrap_Graphics.cpp"
    $23 = ($22<<24>>24)!=(0); //@line 309 "modules/graphics/opengl/wrap_Graphics.cpp"
    __THREW__ = 0;
    $24 = (invoke_iiiii(3536,($L|0),2,($21|0),($23|0))|0); //@line 309 "modules/graphics/opengl/wrap_Graphics.cpp"
    $25 = __THREW__; __THREW__ = 0;
    $26 = $25&1;
    if ($26) {
     label = 16;
     break;
    }
    $27 = $24&1; //@line 309 "modules/graphics/opengl/wrap_Graphics.cpp"
    HEAP8[$flags>>0] = $27; //@line 309 "modules/graphics/opengl/wrap_Graphics.cpp"
    HEAP32[$name$i10>>2] = 0; //@line 293 "modules/graphics/opengl/wrap_Graphics.cpp"
    __THREW__ = 0;
    (invoke_iii(3535,1,($name$i10|0))|0); //@line 294 "modules/graphics/opengl/wrap_Graphics.cpp"
    $28 = __THREW__; __THREW__ = 0;
    $29 = $28&1;
    if ($29) {
     label = 16;
     break;
    }
    $30 = HEAP32[$name$i10>>2]|0; //@line 295 "modules/graphics/opengl/wrap_Graphics.cpp"
    $31 = HEAP8[$12>>0]|0; //@line 310 "modules/graphics/opengl/wrap_Graphics.cpp"
    $32 = ($31<<24>>24)!=(0); //@line 310 "modules/graphics/opengl/wrap_Graphics.cpp"
    __THREW__ = 0;
    $33 = (invoke_iiiii(3536,($L|0),2,($30|0),($32|0))|0); //@line 310 "modules/graphics/opengl/wrap_Graphics.cpp"
    $34 = __THREW__; __THREW__ = 0;
    $35 = $34&1;
    if ($35) {
     label = 16;
     break;
    }
    $36 = $33&1; //@line 310 "modules/graphics/opengl/wrap_Graphics.cpp"
    HEAP8[$12>>0] = $36; //@line 310 "modules/graphics/opengl/wrap_Graphics.cpp"
   }
   __THREW__ = 0;
   $67 = (invoke_iii(3453,($L|0),1)|0); //@line 316 "modules/graphics/opengl/wrap_Graphics.cpp"
   $68 = __THREW__; __THREW__ = 0;
   $69 = $68&1;
   if ($69) {
    label = 16;
   } else {
    $70 = ($67|0)==(0); //@line 316 "modules/graphics/opengl/wrap_Graphics.cpp"
    do {
     if ($70) {
      __THREW__ = 0;
      $71 = (invoke_iiii(3458,($L|0),1,5)|0); //@line 316 "modules/graphics/opengl/wrap_Graphics.cpp"
      $72 = __THREW__; __THREW__ = 0;
      $73 = $72&1;
      if ($73) {
       label = 16;
       break L4;
      }
      if ($71) {
       label = 32;
      } else {
       __THREW__ = 0;
       $74 = (invoke_iiii(3458,($L|0),1,7)|0); //@line 316 "modules/graphics/opengl/wrap_Graphics.cpp"
       $75 = __THREW__; __THREW__ = 0;
       $76 = $75&1;
       if ($76) {
        label = 16;
        break L4;
       }
       if ($74) {
        label = 32;
       } else {
        __THREW__ = 0;
        $189 = (invoke_iiii(3458,($L|0),1,23)|0); //@line 342 "modules/graphics/opengl/wrap_Graphics.cpp"
        $190 = __THREW__; __THREW__ = 0;
        $191 = $190&1;
        if ($191) {
         label = 16;
         break L4;
        }
        if ($189) {
         __THREW__ = 0;
         $192 = (invoke_iii(3540,($L|0),1)|0); //@line 343 "modules/graphics/opengl/wrap_Graphics.cpp"
         $193 = __THREW__; __THREW__ = 0;
         $194 = $193&1;
         if ($194) {
          label = 16;
          break L4;
         }
         HEAP32[$2>>2] = $192; //@line 343 "modules/graphics/opengl/wrap_Graphics.cpp"
         $195 = HEAP32[$10>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $196 = ((($cdata)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $197 = HEAP32[$196>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $198 = ($195>>>0)<($197>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($198) {
          $199 = $192; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          HEAP32[$195>>2] = $199; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $200 = HEAP32[$10>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $201 = ((($200)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$10>>2] = $201; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $releasedata$1 = 0;
          label = 75;
          break;
         } else {
          __THREW__ = 0;
          invoke_vii(3538,($cdata|0),($2|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $202 = __THREW__; __THREW__ = 0;
          $203 = $202&1;
          if ($203) {
           label = 16;
           break L4;
          } else {
           $releasedata$1 = 0;
           label = 75;
           break;
          }
         }
        } else {
         __THREW__ = 0;
         $204 = (invoke_iii(3541,($L|0),1)|0); //@line 345 "modules/graphics/opengl/wrap_Graphics.cpp"
         $205 = __THREW__; __THREW__ = 0;
         $206 = $205&1;
         if ($206) {
          label = 16;
          break L4;
         }
         HEAP32[$3>>2] = $204; //@line 345 "modules/graphics/opengl/wrap_Graphics.cpp"
         $207 = HEAP32[$8>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $208 = ((($data)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $209 = HEAP32[$208>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $210 = ($207>>>0)<($209>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($210) {
          $211 = $204; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          HEAP32[$207>>2] = $211; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $212 = HEAP32[$8>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $213 = ((($212)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$8>>2] = $213; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $releasedata$1 = 0;
          label = 75;
          break;
         } else {
          __THREW__ = 0;
          invoke_vii(3539,($data|0),($3|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $214 = __THREW__; __THREW__ = 0;
          $215 = $214&1;
          if ($215) {
           label = 16;
           break L4;
          } else {
           $releasedata$1 = 0;
           label = 75;
           break;
          }
         }
        }
       }
      }
     } else {
      label = 32;
     }
    } while(0);
    L33: do {
     if ((label|0) == 32) {
      $77 = HEAP32[(625628)>>2]|0; //@line 98 "./common/Module.h"
      $78 = ($77|0)==(0|0); //@line 319 "modules/graphics/opengl/wrap_Graphics.cpp"
      if ($78) {
       __THREW__ = 0;
       $79 = (invoke_iiii(3389,($L|0),(329421|0),($vararg_buffer1|0))|0); //@line 320 "modules/graphics/opengl/wrap_Graphics.cpp"
       $80 = __THREW__; __THREW__ = 0;
       $81 = $80&1;
       if ($81) {
        label = 16;
        break L4;
       } else {
        $$2 = $79;
        break;
       }
      }
      __THREW__ = 0;
      $82 = (invoke_iii(3537,($L|0),1)|0); //@line 322 "modules/graphics/opengl/wrap_Graphics.cpp"
      $83 = __THREW__; __THREW__ = 0;
      $84 = $83&1;
      if ($84) {
       label = 16;
       break L4;
      }
      $85 = HEAP32[$77>>2]|0; //@line 324 "modules/graphics/opengl/wrap_Graphics.cpp"
      $86 = ((($85)) + 40|0); //@line 324 "modules/graphics/opengl/wrap_Graphics.cpp"
      $87 = HEAP32[$86>>2]|0; //@line 324 "modules/graphics/opengl/wrap_Graphics.cpp"
      __THREW__ = 0;
      $88 = (invoke_iii($87|0,($77|0),($82|0))|0); //@line 324 "modules/graphics/opengl/wrap_Graphics.cpp"
      $89 = __THREW__; __THREW__ = 0;
      $90 = $89&1;
      if ($90) {
       label = 16;
       break L4;
      }
      if ($88) {
       $91 = HEAP32[$77>>2]|0; //@line 327 "modules/graphics/opengl/wrap_Graphics.cpp"
       $92 = ((($91)) + 36|0); //@line 327 "modules/graphics/opengl/wrap_Graphics.cpp"
       $93 = HEAP32[$92>>2]|0; //@line 327 "modules/graphics/opengl/wrap_Graphics.cpp"
       __THREW__ = 0;
       $94 = (invoke_iii($93|0,($77|0),($82|0))|0); //@line 327 "modules/graphics/opengl/wrap_Graphics.cpp"
       $95 = __THREW__; __THREW__ = 0;
       $96 = $95&1;
       do {
        if (!($96)) {
         HEAP32[$1>>2] = $94; //@line 327 "modules/graphics/opengl/wrap_Graphics.cpp"
         $97 = HEAP32[$10>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $98 = ((($cdata)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $99 = HEAP32[$98>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $100 = ($97>>>0)<($99>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($100) {
          $101 = $94; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          HEAP32[$97>>2] = $101; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $102 = HEAP32[$10>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $103 = ((($102)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$10>>2] = $103; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         } else {
          __THREW__ = 0;
          invoke_vii(3538,($cdata|0),($1|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $104 = __THREW__; __THREW__ = 0;
          $105 = $104&1;
          if ($105) {
           break;
          }
         }
         $133 = HEAP32[$82>>2]|0; //@line 328 "modules/graphics/opengl/wrap_Graphics.cpp"
         $134 = ((($133)) + 12|0); //@line 328 "modules/graphics/opengl/wrap_Graphics.cpp"
         $135 = HEAP32[$134>>2]|0; //@line 328 "modules/graphics/opengl/wrap_Graphics.cpp"
         __THREW__ = 0;
         invoke_vi($135|0,($82|0)); //@line 328 "modules/graphics/opengl/wrap_Graphics.cpp"
         $136 = __THREW__; __THREW__ = 0;
         $137 = $136&1;
         if ($137) {
          label = 16;
          break L4;
         } else {
          $releasedata$1 = 1;
          label = 75;
          break L33;
         }
        }
       } while(0);
       $106 = ___cxa_find_matching_catch_3(8992|0)|0;
       $107 = tempRet0;
       $108 = (_llvm_eh_typeid_for((8992|0))|0); //@line 556 "./common/runtime.h"
       $109 = ($107|0)==($108|0); //@line 556 "./common/runtime.h"
       do {
        if ($109) {
         $110 = (___cxa_begin_catch(($106|0))|0); //@line 556 "./common/runtime.h"
         $111 = HEAP32[$110>>2]|0; //@line 560 "./common/runtime.h"
         $112 = ((($111)) + 8|0); //@line 560 "./common/runtime.h"
         $113 = HEAP32[$112>>2]|0; //@line 560 "./common/runtime.h"
         $114 = (FUNCTION_TABLE_ii[$113 & 8191]($110)|0); //@line 560 "./common/runtime.h"
         __THREW__ = 0;
         invoke_vii(3380,($L|0),($114|0)); //@line 560 "./common/runtime.h"
         $115 = __THREW__; __THREW__ = 0;
         $116 = $115&1;
         if ($116) {
          $129 = ___cxa_find_matching_catch_2()|0;
          $130 = tempRet0;
          __THREW__ = 0;
          invoke_v(3378); //@line 561 "./common/runtime.h"
          $131 = __THREW__; __THREW__ = 0;
          $132 = $131&1;
          if (!($132)) {
           $$01$i = $130;$$02$i = $129;
           break;
          }
          $138 = ___cxa_find_matching_catch_3(0|0)|0;
          $139 = tempRet0;
          ___clang_call_terminate($138); //@line 561 "./common/runtime.h"
          // unreachable; //@line 561 "./common/runtime.h"
         }
         __THREW__ = 0;
         invoke_v(3378); //@line 561 "./common/runtime.h"
         $117 = __THREW__; __THREW__ = 0;
         $118 = $117&1;
         if ($118) {
          label = 16;
          break L4;
         }
         $119 = HEAP32[$82>>2]|0; //@line 328 "modules/graphics/opengl/wrap_Graphics.cpp"
         $120 = ((($119)) + 12|0); //@line 328 "modules/graphics/opengl/wrap_Graphics.cpp"
         $121 = HEAP32[$120>>2]|0; //@line 328 "modules/graphics/opengl/wrap_Graphics.cpp"
         __THREW__ = 0;
         invoke_vi($121|0,($82|0)); //@line 328 "modules/graphics/opengl/wrap_Graphics.cpp"
         $122 = __THREW__; __THREW__ = 0;
         $123 = $122&1;
         if ($123) {
          label = 16;
          break L4;
         }
         __THREW__ = 0;
         $124 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 566 "./common/runtime.h"
         $125 = __THREW__; __THREW__ = 0;
         $126 = $125&1;
         if ($126) {
          label = 16;
          break L4;
         }
         __THREW__ = 0;
         HEAP32[$vararg_buffer3>>2] = $124; //@line 566 "./common/runtime.h"
         (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer3|0))|0); //@line 566 "./common/runtime.h"
         $127 = __THREW__; __THREW__ = 0;
         $128 = $127&1;
         if ($128) {
          label = 16;
          break L4;
         } else {
          $releasedata$1 = 1;
          label = 75;
          break L33;
         }
        } else {
         $$01$i = $107;$$02$i = $106;
        }
       } while(0);
       $eh$lpad$body$index21Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
       break L4;
      } else {
       $140 = HEAP32[$77>>2]|0; //@line 334 "modules/graphics/opengl/wrap_Graphics.cpp"
       $141 = ((($140)) + 24|0); //@line 334 "modules/graphics/opengl/wrap_Graphics.cpp"
       $142 = HEAP32[$141>>2]|0; //@line 334 "modules/graphics/opengl/wrap_Graphics.cpp"
       __THREW__ = 0;
       $143 = (invoke_iii($142|0,($77|0),($82|0))|0); //@line 334 "modules/graphics/opengl/wrap_Graphics.cpp"
       $144 = __THREW__; __THREW__ = 0;
       $145 = $144&1;
       do {
        if (!($145)) {
         HEAP32[$0>>2] = $143; //@line 334 "modules/graphics/opengl/wrap_Graphics.cpp"
         $146 = HEAP32[$8>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $147 = ((($data)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $148 = HEAP32[$147>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $149 = ($146>>>0)<($148>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($149) {
          $150 = $143; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          HEAP32[$146>>2] = $150; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $151 = HEAP32[$8>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $152 = ((($151)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$8>>2] = $152; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         } else {
          __THREW__ = 0;
          invoke_vii(3539,($data|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $153 = __THREW__; __THREW__ = 0;
          $154 = $153&1;
          if ($154) {
           break;
          }
         }
         $182 = HEAP32[$82>>2]|0; //@line 335 "modules/graphics/opengl/wrap_Graphics.cpp"
         $183 = ((($182)) + 12|0); //@line 335 "modules/graphics/opengl/wrap_Graphics.cpp"
         $184 = HEAP32[$183>>2]|0; //@line 335 "modules/graphics/opengl/wrap_Graphics.cpp"
         __THREW__ = 0;
         invoke_vi($184|0,($82|0)); //@line 335 "modules/graphics/opengl/wrap_Graphics.cpp"
         $185 = __THREW__; __THREW__ = 0;
         $186 = $185&1;
         if ($186) {
          label = 16;
          break L4;
         } else {
          $releasedata$1 = 1;
          label = 75;
          break L33;
         }
        }
       } while(0);
       $155 = ___cxa_find_matching_catch_3(8992|0)|0;
       $156 = tempRet0;
       $157 = (_llvm_eh_typeid_for((8992|0))|0); //@line 556 "./common/runtime.h"
       $158 = ($156|0)==($157|0); //@line 556 "./common/runtime.h"
       do {
        if ($158) {
         $159 = (___cxa_begin_catch(($155|0))|0); //@line 556 "./common/runtime.h"
         $160 = HEAP32[$159>>2]|0; //@line 560 "./common/runtime.h"
         $161 = ((($160)) + 8|0); //@line 560 "./common/runtime.h"
         $162 = HEAP32[$161>>2]|0; //@line 560 "./common/runtime.h"
         $163 = (FUNCTION_TABLE_ii[$162 & 8191]($159)|0); //@line 560 "./common/runtime.h"
         __THREW__ = 0;
         invoke_vii(3380,($L|0),($163|0)); //@line 560 "./common/runtime.h"
         $164 = __THREW__; __THREW__ = 0;
         $165 = $164&1;
         if ($165) {
          $178 = ___cxa_find_matching_catch_2()|0;
          $179 = tempRet0;
          __THREW__ = 0;
          invoke_v(3378); //@line 561 "./common/runtime.h"
          $180 = __THREW__; __THREW__ = 0;
          $181 = $180&1;
          if (!($181)) {
           $$01$i29 = $179;$$02$i28 = $178;
           break;
          }
          $187 = ___cxa_find_matching_catch_3(0|0)|0;
          $188 = tempRet0;
          ___clang_call_terminate($187); //@line 561 "./common/runtime.h"
          // unreachable; //@line 561 "./common/runtime.h"
         }
         __THREW__ = 0;
         invoke_v(3378); //@line 561 "./common/runtime.h"
         $166 = __THREW__; __THREW__ = 0;
         $167 = $166&1;
         if ($167) {
          label = 16;
          break L4;
         }
         $168 = HEAP32[$82>>2]|0; //@line 335 "modules/graphics/opengl/wrap_Graphics.cpp"
         $169 = ((($168)) + 12|0); //@line 335 "modules/graphics/opengl/wrap_Graphics.cpp"
         $170 = HEAP32[$169>>2]|0; //@line 335 "modules/graphics/opengl/wrap_Graphics.cpp"
         __THREW__ = 0;
         invoke_vi($170|0,($82|0)); //@line 335 "modules/graphics/opengl/wrap_Graphics.cpp"
         $171 = __THREW__; __THREW__ = 0;
         $172 = $171&1;
         if ($172) {
          label = 16;
          break L4;
         }
         __THREW__ = 0;
         $173 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 566 "./common/runtime.h"
         $174 = __THREW__; __THREW__ = 0;
         $175 = $174&1;
         if ($175) {
          label = 16;
          break L4;
         }
         __THREW__ = 0;
         HEAP32[$vararg_buffer5>>2] = $173; //@line 566 "./common/runtime.h"
         (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer5|0))|0); //@line 566 "./common/runtime.h"
         $176 = __THREW__; __THREW__ = 0;
         $177 = $176&1;
         if ($177) {
          label = 16;
          break L4;
         } else {
          $releasedata$1 = 1;
          label = 75;
          break L33;
         }
        } else {
         $$01$i29 = $156;$$02$i28 = $155;
        }
       } while(0);
       $eh$lpad$body$index21Z2D = $$01$i29;$eh$lpad$body$indexZ2D = $$02$i28;
       break L4;
      }
     }
    } while(0);
    L77: do {
     if ((label|0) == 75) {
      __THREW__ = 0;
      $216 = (invoke_iii(3358,($L|0),2)|0); //@line 347 "modules/graphics/opengl/wrap_Graphics.cpp"
      $217 = __THREW__; __THREW__ = 0;
      $218 = $217&1;
      if ($218) {
       label = 16;
       break L4;
      }
      $219 = ($216|0)==(5); //@line 347 "modules/graphics/opengl/wrap_Graphics.cpp"
      if ($219) {
       HEAP32[$name$i39>>2] = 0; //@line 293 "modules/graphics/opengl/wrap_Graphics.cpp"
       __THREW__ = 0;
       (invoke_iii(3535,0,($name$i39|0))|0); //@line 294 "modules/graphics/opengl/wrap_Graphics.cpp"
       $220 = __THREW__; __THREW__ = 0;
       $221 = $220&1;
       if ($221) {
        label = 16;
        break L4;
       }
       $222 = HEAP32[$name$i39>>2]|0; //@line 295 "modules/graphics/opengl/wrap_Graphics.cpp"
       __THREW__ = 0;
       invoke_viii(3471,($L|0),2,($222|0)); //@line 349 "modules/graphics/opengl/wrap_Graphics.cpp"
       $223 = __THREW__; __THREW__ = 0;
       $224 = $223&1;
       if ($224) {
        label = 16;
        break L4;
       }
       __THREW__ = 0;
       $225 = (invoke_iii(3358,($L|0),-1)|0); //@line 353 "modules/graphics/opengl/wrap_Graphics.cpp"
       $226 = __THREW__; __THREW__ = 0;
       $227 = $226&1;
       if ($227) {
        label = 16;
        break L4;
       }
       $228 = ($225|0)==(5); //@line 353 "modules/graphics/opengl/wrap_Graphics.cpp"
       L85: do {
        if ($228) {
         $229 = ((($cdata)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $230 = ((($data)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $i$0 = 1;
         L87: while(1) {
          __THREW__ = 0;
          $231 = (invoke_iii(3381,($L|0),-1)|0); //@line 355 "modules/graphics/opengl/wrap_Graphics.cpp"
          $232 = __THREW__; __THREW__ = 0;
          $233 = $232&1;
          if ($233) {
           break;
          }
          $234 = ($i$0>>>0)>($231>>>0); //@line 355 "modules/graphics/opengl/wrap_Graphics.cpp"
          if ($234) {
           break L85;
          }
          __THREW__ = 0;
          invoke_viii(3382,($L|0),-1,($i$0|0)); //@line 357 "modules/graphics/opengl/wrap_Graphics.cpp"
          $235 = __THREW__; __THREW__ = 0;
          $236 = $235&1;
          if ($236) {
           break;
          }
          $237 = HEAP32[$data>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $238 = HEAP32[$8>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $239 = ($237|0)==($238|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          do {
           if ($239) {
            $256 = HEAP32[$cdata>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $257 = HEAP32[$10>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $258 = ($256|0)==($257|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            if ($258) {
             break;
            }
            __THREW__ = 0;
            $259 = (invoke_iiii(3458,($L|0),-1,23)|0); //@line 368 "modules/graphics/opengl/wrap_Graphics.cpp"
            $260 = __THREW__; __THREW__ = 0;
            $261 = $260&1;
            if ($261) {
             break L87;
            }
            if (!($259)) {
             __THREW__ = 0;
             (invoke_iiiii(3457,($L|0),-1,(438231|0),(348117|0))|0); //@line 369 "modules/graphics/opengl/wrap_Graphics.cpp"
             $262 = __THREW__; __THREW__ = 0;
             $263 = $262&1;
             if ($263) {
              break L87;
             }
            }
            __THREW__ = 0;
            $264 = (invoke_iii(3540,($L|0),-1)|0); //@line 371 "modules/graphics/opengl/wrap_Graphics.cpp"
            $265 = __THREW__; __THREW__ = 0;
            $266 = $265&1;
            if ($266) {
             break L87;
            }
            HEAP32[$5>>2] = $264; //@line 371 "modules/graphics/opengl/wrap_Graphics.cpp"
            $267 = HEAP32[$10>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $268 = HEAP32[$229>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $269 = ($267>>>0)<($268>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            if ($269) {
             $270 = $264; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
             HEAP32[$267>>2] = $270; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
             $271 = HEAP32[$10>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $272 = ((($271)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             HEAP32[$10>>2] = $272; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             break;
            } else {
             __THREW__ = 0;
             invoke_vii(3538,($cdata|0),($5|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $273 = __THREW__; __THREW__ = 0;
             $274 = $273&1;
             if ($274) {
              break L87;
             } else {
              break;
             }
            }
           } else {
            __THREW__ = 0;
            $240 = (invoke_iiii(3458,($L|0),-1,22)|0); //@line 361 "modules/graphics/opengl/wrap_Graphics.cpp"
            $241 = __THREW__; __THREW__ = 0;
            $242 = $241&1;
            if ($242) {
             break L87;
            }
            if (!($240)) {
             __THREW__ = 0;
             (invoke_iiiii(3457,($L|0),-1,(438231|0),(439339|0))|0); //@line 362 "modules/graphics/opengl/wrap_Graphics.cpp"
             $243 = __THREW__; __THREW__ = 0;
             $244 = $243&1;
             if ($244) {
              break L87;
             }
            }
            __THREW__ = 0;
            $245 = (invoke_iii(3541,($L|0),-1)|0); //@line 364 "modules/graphics/opengl/wrap_Graphics.cpp"
            $246 = __THREW__; __THREW__ = 0;
            $247 = $246&1;
            if ($247) {
             break L87;
            }
            HEAP32[$4>>2] = $245; //@line 364 "modules/graphics/opengl/wrap_Graphics.cpp"
            $248 = HEAP32[$8>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $249 = HEAP32[$230>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $250 = ($248>>>0)<($249>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            if ($250) {
             $251 = $245; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
             HEAP32[$248>>2] = $251; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
             $252 = HEAP32[$8>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $253 = ((($252)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             HEAP32[$8>>2] = $253; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             break;
            } else {
             __THREW__ = 0;
             invoke_vii(3539,($data|0),($4|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $254 = __THREW__; __THREW__ = 0;
             $255 = $254&1;
             if ($255) {
              break L87;
             } else {
              break;
             }
            }
           }
          } while(0);
          __THREW__ = 0;
          invoke_vii(3384,($L|0),-2); //@line 374 "modules/graphics/opengl/wrap_Graphics.cpp"
          $275 = __THREW__; __THREW__ = 0;
          $276 = $275&1;
          if ($276) {
           break;
          }
          $277 = (($i$0) + 1)|0; //@line 355 "modules/graphics/opengl/wrap_Graphics.cpp"
          $i$0 = $277;
         }
         $45 = ___cxa_find_matching_catch_2()|0;
         $46 = tempRet0;
         $eh$lpad$body$index21Z2D = $46;$eh$lpad$body$indexZ2D = $45;
         break L4;
        }
       } while(0);
       __THREW__ = 0;
       invoke_vii(3384,($L|0),-2); //@line 378 "modules/graphics/opengl/wrap_Graphics.cpp"
       $278 = __THREW__; __THREW__ = 0;
       $279 = $278&1;
       if ($279) {
        label = 16;
        break L4;
       }
      }
      $280 = HEAP32[$cdata>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $281 = HEAP32[$10>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $282 = ($280|0)==($281|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($282) {
       $287 = HEAP32[$data>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $288 = HEAP32[$8>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $289 = ($287|0)==($288|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($289) {
        $image$0 = 0;
        label = 121;
       } else {
        $290 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
        __THREW__ = 0;
        $291 = (invoke_iiii(3543,($290|0),($data|0),($flags|0))|0); //@line 388 "modules/graphics/opengl/wrap_Graphics.cpp"
        $292 = __THREW__; __THREW__ = 0;
        $293 = $292&1;
        if ($293) {
         label = 108;
        } else {
         $image$0 = $291;
         label = 121;
        }
       }
      } else {
       $283 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
       __THREW__ = 0;
       $284 = (invoke_iiii(3542,($283|0),($cdata|0),($flags|0))|0); //@line 386 "modules/graphics/opengl/wrap_Graphics.cpp"
       $285 = __THREW__; __THREW__ = 0;
       $286 = $285&1;
       if ($286) {
        label = 108;
       } else {
        $image$0 = $284;
        label = 121;
       }
      }
      L121: do {
       if ((label|0) == 108) {
        $294 = ___cxa_find_matching_catch_3(8992|0)|0;
        $295 = tempRet0;
        $296 = (_llvm_eh_typeid_for((8992|0))|0); //@line 556 "./common/runtime.h"
        $297 = ($295|0)==($296|0); //@line 556 "./common/runtime.h"
        do {
         if ($297) {
          $298 = (___cxa_begin_catch(($294|0))|0); //@line 556 "./common/runtime.h"
          $299 = HEAP32[$298>>2]|0; //@line 560 "./common/runtime.h"
          $300 = ((($299)) + 8|0); //@line 560 "./common/runtime.h"
          $301 = HEAP32[$300>>2]|0; //@line 560 "./common/runtime.h"
          $302 = (FUNCTION_TABLE_ii[$301 & 8191]($298)|0); //@line 560 "./common/runtime.h"
          __THREW__ = 0;
          invoke_vii(3380,($L|0),($302|0)); //@line 560 "./common/runtime.h"
          $303 = __THREW__; __THREW__ = 0;
          $304 = $303&1;
          if ($304) {
           $335 = ___cxa_find_matching_catch_2()|0;
           $336 = tempRet0;
           __THREW__ = 0;
           invoke_v(3378); //@line 561 "./common/runtime.h"
           $337 = __THREW__; __THREW__ = 0;
           $338 = $337&1;
           if (!($338)) {
            $$01$i49 = $336;$$02$i48 = $335;
            break;
           }
           $362 = ___cxa_find_matching_catch_3(0|0)|0;
           $363 = tempRet0;
           ___clang_call_terminate($362); //@line 561 "./common/runtime.h"
           // unreachable; //@line 561 "./common/runtime.h"
          }
          __THREW__ = 0;
          invoke_v(3378); //@line 561 "./common/runtime.h"
          $305 = __THREW__; __THREW__ = 0;
          $306 = $305&1;
          if ($306) {
           label = 16;
           break L4;
          }
          $307 = ($releasedata$1<<24>>24)==(0); //@line 391 "modules/graphics/opengl/wrap_Graphics.cpp"
          L146: do {
           if (!($307)) {
            $308 = HEAP32[$data>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $309 = HEAP32[$8>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $310 = ($308|0)==($309|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
            L148: do {
             if (!($310)) {
              $__begin$sroa$0$02$i$i = $308;
              while(1) {
               $314 = HEAP32[$__begin$sroa$0$02$i$i>>2]|0; //@line 393 "modules/graphics/opengl/wrap_Graphics.cpp"
               $315 = HEAP32[$314>>2]|0; //@line 394 "modules/graphics/opengl/wrap_Graphics.cpp"
               $316 = ((($315)) + 12|0); //@line 394 "modules/graphics/opengl/wrap_Graphics.cpp"
               $317 = HEAP32[$316>>2]|0; //@line 394 "modules/graphics/opengl/wrap_Graphics.cpp"
               __THREW__ = 0;
               invoke_vi($317|0,($314|0)); //@line 394 "modules/graphics/opengl/wrap_Graphics.cpp"
               $318 = __THREW__; __THREW__ = 0;
               $319 = $318&1;
               if ($319) {
                break;
               }
               $320 = ((($__begin$sroa$0$02$i$i)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
               $321 = ($320|0)==($309|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
               if ($321) {
                break L148;
               } else {
                $__begin$sroa$0$02$i$i = $320;
               }
              }
              $43 = ___cxa_find_matching_catch_2()|0;
              $44 = tempRet0;
              $eh$lpad$body$index21Z2D = $44;$eh$lpad$body$indexZ2D = $43;
              break L4;
             }
            } while(0);
            $311 = HEAP32[$cdata>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $312 = HEAP32[$10>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $313 = ($311|0)==($312|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
            if (!($313)) {
             $__begin2$sroa$0$01$i$i = $311;
             while(1) {
              $322 = HEAP32[$__begin2$sroa$0$01$i$i>>2]|0; //@line 395 "modules/graphics/opengl/wrap_Graphics.cpp"
              $323 = HEAP32[$322>>2]|0; //@line 396 "modules/graphics/opengl/wrap_Graphics.cpp"
              $324 = ((($323)) + 12|0); //@line 396 "modules/graphics/opengl/wrap_Graphics.cpp"
              $325 = HEAP32[$324>>2]|0; //@line 396 "modules/graphics/opengl/wrap_Graphics.cpp"
              __THREW__ = 0;
              invoke_vi($325|0,($322|0)); //@line 396 "modules/graphics/opengl/wrap_Graphics.cpp"
              $326 = __THREW__; __THREW__ = 0;
              $327 = $326&1;
              if ($327) {
               break;
              }
              $328 = ((($__begin2$sroa$0$01$i$i)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
              $329 = ($328|0)==($312|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
              if ($329) {
               break L146;
              } else {
               $__begin2$sroa$0$01$i$i = $328;
              }
             }
             $41 = ___cxa_find_matching_catch_2()|0;
             $42 = tempRet0;
             $eh$lpad$body$index21Z2D = $42;$eh$lpad$body$indexZ2D = $41;
             break L4;
            }
           }
          } while(0);
          __THREW__ = 0;
          $330 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 566 "./common/runtime.h"
          $331 = __THREW__; __THREW__ = 0;
          $332 = $331&1;
          if ($332) {
           label = 16;
           break L4;
          }
          __THREW__ = 0;
          HEAP32[$vararg_buffer8>>2] = $330; //@line 566 "./common/runtime.h"
          (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer8|0))|0); //@line 566 "./common/runtime.h"
          $333 = __THREW__; __THREW__ = 0;
          $334 = $333&1;
          if ($334) {
           label = 16;
           break L4;
          } else {
           break L121;
          }
         } else {
          $$01$i49 = $295;$$02$i48 = $294;
         }
        } while(0);
        $eh$lpad$body$index21Z2D = $$01$i49;$eh$lpad$body$indexZ2D = $$02$i48;
        break L4;
       }
       else if ((label|0) == 121) {
        $339 = ($releasedata$1<<24>>24)==(0); //@line 391 "modules/graphics/opengl/wrap_Graphics.cpp"
        L123: do {
         if (!($339)) {
          $340 = HEAP32[$data>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $341 = HEAP32[$8>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $342 = ($340|0)==($341|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
          L125: do {
           if (!($342)) {
            $__begin$sroa$0$02$i3$i = $340;
            while(1) {
             $346 = HEAP32[$__begin$sroa$0$02$i3$i>>2]|0; //@line 393 "modules/graphics/opengl/wrap_Graphics.cpp"
             $347 = HEAP32[$346>>2]|0; //@line 394 "modules/graphics/opengl/wrap_Graphics.cpp"
             $348 = ((($347)) + 12|0); //@line 394 "modules/graphics/opengl/wrap_Graphics.cpp"
             $349 = HEAP32[$348>>2]|0; //@line 394 "modules/graphics/opengl/wrap_Graphics.cpp"
             __THREW__ = 0;
             invoke_vi($349|0,($346|0)); //@line 394 "modules/graphics/opengl/wrap_Graphics.cpp"
             $350 = __THREW__; __THREW__ = 0;
             $351 = $350&1;
             if ($351) {
              break;
             }
             $352 = ((($__begin$sroa$0$02$i3$i)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
             $353 = ($352|0)==($341|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
             if ($353) {
              break L125;
             } else {
              $__begin$sroa$0$02$i3$i = $352;
             }
            }
            $39 = ___cxa_find_matching_catch_2()|0;
            $40 = tempRet0;
            $eh$lpad$body$index21Z2D = $40;$eh$lpad$body$indexZ2D = $39;
            break L4;
           }
          } while(0);
          $343 = HEAP32[$cdata>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $344 = HEAP32[$10>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $345 = ($343|0)==($344|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
          if (!($345)) {
           $__begin2$sroa$0$01$i5$i = $343;
           while(1) {
            $354 = HEAP32[$__begin2$sroa$0$01$i5$i>>2]|0; //@line 395 "modules/graphics/opengl/wrap_Graphics.cpp"
            $355 = HEAP32[$354>>2]|0; //@line 396 "modules/graphics/opengl/wrap_Graphics.cpp"
            $356 = ((($355)) + 12|0); //@line 396 "modules/graphics/opengl/wrap_Graphics.cpp"
            $357 = HEAP32[$356>>2]|0; //@line 396 "modules/graphics/opengl/wrap_Graphics.cpp"
            __THREW__ = 0;
            invoke_vi($357|0,($354|0)); //@line 396 "modules/graphics/opengl/wrap_Graphics.cpp"
            $358 = __THREW__; __THREW__ = 0;
            $359 = $358&1;
            if ($359) {
             break;
            }
            $360 = ((($__begin2$sroa$0$01$i5$i)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
            $361 = ($360|0)==($344|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
            if ($361) {
             break L123;
            } else {
             $__begin2$sroa$0$01$i5$i = $360;
            }
           }
           $37 = ___cxa_find_matching_catch_2()|0;
           $38 = tempRet0;
           $eh$lpad$body$index21Z2D = $38;$eh$lpad$body$indexZ2D = $37;
           break L4;
          }
         }
        } while(0);
        $364 = ($image$0|0)==(0|0); //@line 401 "modules/graphics/opengl/wrap_Graphics.cpp"
        if (!($364)) {
         __THREW__ = 0;
         invoke_viii(3387,($L|0),12,($image$0|0)); //@line 405 "modules/graphics/opengl/wrap_Graphics.cpp"
         $368 = __THREW__; __THREW__ = 0;
         $369 = $368&1;
         if ($369) {
          label = 16;
          break L4;
         }
         $370 = HEAP32[$image$0>>2]|0; //@line 406 "modules/graphics/opengl/wrap_Graphics.cpp"
         $371 = ((($370)) + 12|0); //@line 406 "modules/graphics/opengl/wrap_Graphics.cpp"
         $372 = HEAP32[$371>>2]|0; //@line 406 "modules/graphics/opengl/wrap_Graphics.cpp"
         __THREW__ = 0;
         invoke_vi($372|0,($image$0|0)); //@line 406 "modules/graphics/opengl/wrap_Graphics.cpp"
         $373 = __THREW__; __THREW__ = 0;
         $374 = $373&1;
         if ($374) {
          label = 16;
          break L4;
         } else {
          $$2 = 1;
          break L77;
         }
        }
       }
      } while(0);
      __THREW__ = 0;
      $365 = (invoke_iiii(3389,($L|0),(329471|0),($vararg_buffer11|0))|0); //@line 402 "modules/graphics/opengl/wrap_Graphics.cpp"
      $366 = __THREW__; __THREW__ = 0;
      $367 = $366&1;
      if ($367) {
       label = 16;
       break L4;
      } else {
       $$2 = $365;
      }
     }
    } while(0);
    $375 = HEAP32[$cdata>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $376 = ($375|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $377 = $375; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($376)) {
     $378 = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $379 = ($378|0)==($375|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($379)) {
      $scevgep$i$i$i7 = ((($378)) + -4|0);
      $380 = $scevgep$i$i$i7;
      $381 = (($380) - ($377))|0;
      $382 = $381 >>> 2;
      $383 = $382 ^ -1;
      $scevgep4$i$i$i8 = (($378) + ($383<<2)|0);
      HEAP32[$10>>2] = $scevgep4$i$i$i8; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __ZdlPv($375); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    }
    $384 = HEAP32[$data>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $385 = ($384|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $386 = $384; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($385) {
     STACKTOP = sp;return ($$2|0); //@line 408 "modules/graphics/opengl/wrap_Graphics.cpp"
    }
    $387 = HEAP32[$8>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $388 = ($387|0)==($384|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($388)) {
     $scevgep$i$i$i = ((($387)) + -4|0);
     $389 = $scevgep$i$i$i;
     $390 = (($389) - ($386))|0;
     $391 = $390 >>> 2;
     $392 = $391 ^ -1;
     $scevgep4$i$i$i = (($387) + ($392<<2)|0);
     HEAP32[$8>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($384); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    STACKTOP = sp;return ($$2|0); //@line 408 "modules/graphics/opengl/wrap_Graphics.cpp"
   }
  }
 } while(0);
 if ((label|0) == 16) {
  $47 = ___cxa_find_matching_catch_2()|0;
  $48 = tempRet0;
  $eh$lpad$body$index21Z2D = $48;$eh$lpad$body$indexZ2D = $47;
 }
 $49 = HEAP32[$cdata>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $50 = ($49|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $51 = $49; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($50)) {
  $52 = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $53 = ($52|0)==($49|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($53)) {
   $scevgep$i$i$i12 = ((($52)) + -4|0);
   $54 = $scevgep$i$i$i12;
   $55 = (($54) - ($51))|0;
   $56 = $55 >>> 2;
   $57 = $56 ^ -1;
   $scevgep4$i$i$i13 = (($52) + ($57<<2)|0);
   HEAP32[$10>>2] = $scevgep4$i$i$i13; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($49); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $58 = HEAP32[$data>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $59 = ($58|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $60 = $58; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($59) {
  ___resumeException($eh$lpad$body$indexZ2D|0);
  // unreachable;
 }
 $61 = HEAP32[$8>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $62 = ($61|0)==($58|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($62)) {
  $scevgep$i$i$i17 = ((($61)) + -4|0);
  $63 = $scevgep$i$i$i17;
  $64 = (($63) - ($60))|0;
  $65 = $64 >>> 2;
  $66 = $65 ^ -1;
  $scevgep4$i$i$i18 = (($61) + ($66<<2)|0);
  HEAP32[$8>>2] = $scevgep4$i$i$i18; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($58); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__16vectorIPN4love5image19CompressedImageDataENS_9allocatorIS4_EEE21__push_back_slow_pathIS4_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNSt3__16vectorIPN4love5image9ImageDataENS_9allocatorIS4_EEE21__push_back_slow_pathIS4_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl9w_newQuadEP9lua_State($L) {
 $L = $L|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 40|0;
 $vararg_buffer = sp + 32|0;
 $0 = sp;
 $1 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $2 = (__ZNK4love8graphics6opengl8Graphics9isCreatedEv($1)|0); //@line 52 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($2)) {
  (_luaL_error($L,329373,$vararg_buffer)|0); //@line 53 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $3 = (+_luaL_checknumber($L,1)); //@line 415 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = (+_luaL_checknumber($L,2)); //@line 416 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = (+_luaL_checknumber($L,3)); //@line 417 "modules/graphics/opengl/wrap_Graphics.cpp"
 $6 = (+_luaL_checknumber($L,4)); //@line 418 "modules/graphics/opengl/wrap_Graphics.cpp"
 $7 = (+_luaL_checknumber($L,5)); //@line 420 "modules/graphics/opengl/wrap_Graphics.cpp"
 $8 = (+_luaL_checknumber($L,6)); //@line 421 "modules/graphics/opengl/wrap_Graphics.cpp"
 $9 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 HEAPF64[$0>>3] = $3; //@line 423 "modules/graphics/opengl/wrap_Graphics.cpp"
 $10 = ((($0)) + 8|0); //@line 423 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAPF64[$10>>3] = $4; //@line 423 "modules/graphics/opengl/wrap_Graphics.cpp"
 $11 = ((($0)) + 16|0); //@line 423 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAPF64[$11>>3] = $5; //@line 423 "modules/graphics/opengl/wrap_Graphics.cpp"
 $12 = ((($0)) + 24|0); //@line 423 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAPF64[$12>>3] = $6; //@line 423 "modules/graphics/opengl/wrap_Graphics.cpp"
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;HEAP32[$$byval_copy+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$$byval_copy+8>>2]=HEAP32[$0+8>>2]|0;HEAP32[$$byval_copy+12>>2]=HEAP32[$0+12>>2]|0;HEAP32[$$byval_copy+16>>2]=HEAP32[$0+16>>2]|0;HEAP32[$$byval_copy+20>>2]=HEAP32[$0+20>>2]|0;HEAP32[$$byval_copy+24>>2]=HEAP32[$0+24>>2]|0;HEAP32[$$byval_copy+28>>2]=HEAP32[$0+28>>2]|0; //@line 423 "modules/graphics/opengl/wrap_Graphics.cpp"
 $13 = (__ZN4love8graphics6opengl8Graphics7newQuadENS0_4Quad8ViewportEdd($9,$$byval_copy,$7,$8)|0); //@line 423 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,13,$13); //@line 424 "modules/graphics/opengl/wrap_Graphics.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 425 "modules/graphics/opengl/wrap_Graphics.cpp"
 $15 = ((($14)) + 12|0); //@line 425 "modules/graphics/opengl/wrap_Graphics.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 425 "modules/graphics/opengl/wrap_Graphics.cpp"
 FUNCTION_TABLE_vi[$16 & 8191]($13); //@line 425 "modules/graphics/opengl/wrap_Graphics.cpp"
 STACKTOP = sp;return 1; //@line 426 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl9w_newFontEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $$cast = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $8 = 0, $9 = 0, $font$0 = 0, $i$0 = 0;
 var $idxs = 0, $lpad$phi$index = 0, $lpad$phi$index4 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i1 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i2 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $idxs = sp + 16|0;
 $0 = sp + 12|0;
 $1 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $2 = (__ZNK4love8graphics6opengl8Graphics9isCreatedEv($1)|0); //@line 52 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($2)) {
  (_luaL_error($L,329373,$vararg_buffer)|0); //@line 53 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $3 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,1,9)|0); //@line 436 "modules/graphics/opengl/wrap_Graphics.cpp"
 L4: do {
  if (!($3)) {
   HEAP32[$idxs>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $4 = ((($idxs)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$4>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $5 = ((($idxs)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$5>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $6 = ((($idxs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $i$0 = 0;
   while(1) {
    __THREW__ = 0;
    $7 = (invoke_ii(3385,($L|0))|0); //@line 439 "modules/graphics/opengl/wrap_Graphics.cpp"
    $8 = __THREW__; __THREW__ = 0;
    $9 = $8&1;
    if ($9) {
     label = 8;
     break;
    }
    $10 = ($i$0|0)<($7|0); //@line 439 "modules/graphics/opengl/wrap_Graphics.cpp"
    if (!($10)) {
     label = 7;
     break;
    }
    $30 = (($i$0) + 1)|0; //@line 440 "modules/graphics/opengl/wrap_Graphics.cpp"
    HEAP32[$0>>2] = $30; //@line 440 "modules/graphics/opengl/wrap_Graphics.cpp"
    $31 = HEAP32[$4>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $32 = HEAP32[$6>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $33 = ($31>>>0)<($32>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($33) {
     HEAP32[$31>>2] = $30; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $34 = ((($31)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$4>>2] = $34; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $i$0 = $30;
     continue;
    } else {
     __THREW__ = 0;
     invoke_vii(3406,($idxs|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $35 = __THREW__; __THREW__ = 0;
     $36 = $35&1;
     if ($36) {
      label = 8;
      break;
     } else {
      $i$0 = $30;
      continue;
     }
    }
   }
   do {
    if ((label|0) == 7) {
     $11 = HEAP32[$idxs>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $12 = HEAP32[$4>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $$cast = $11; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $13 = (($12) - ($$cast))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     __THREW__ = 0;
     (invoke_iiiiii(3455,($L|0),($11|0),($14|0),(329493|0),(329498|0))|0); //@line 442 "modules/graphics/opengl/wrap_Graphics.cpp"
     $15 = __THREW__; __THREW__ = 0;
     $16 = $15&1;
     if ($16) {
      $19 = ___cxa_find_matching_catch_2()|0;
      $20 = tempRet0;
      $lpad$phi$index = $19;$lpad$phi$index4 = $20;
      break;
     }
     $37 = HEAP32[$idxs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $38 = ($37|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $39 = $37; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($38)) {
      $40 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $41 = ($40|0)==($37|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($41)) {
       $scevgep$i$i$i1 = ((($40)) + -4|0);
       $42 = $scevgep$i$i$i1;
       $43 = (($42) - ($39))|0;
       $44 = $43 >>> 2;
       $45 = $44 ^ -1;
       $scevgep4$i$i$i2 = (($40) + ($45<<2)|0);
       HEAP32[$4>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($37); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     break L4;
    }
    else if ((label|0) == 8) {
     $17 = ___cxa_find_matching_catch_2()|0;
     $18 = tempRet0;
     $lpad$phi$index = $17;$lpad$phi$index4 = $18;
    }
   } while(0);
   $21 = HEAP32[$idxs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $22 = ($21|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $23 = $21; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($22) {
    ___resumeException($lpad$phi$index|0);
    // unreachable;
   }
   $24 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $25 = ($24|0)==($21|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($25)) {
    $scevgep$i$i$i = ((($24)) + -4|0);
    $26 = $scevgep$i$i$i;
    $27 = (($26) - ($23))|0;
    $28 = $27 >>> 2;
    $29 = $28 ^ -1;
    $scevgep4$i$i$i = (($24) + ($29<<2)|0);
    HEAP32[$4>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($21); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($lpad$phi$index|0);
   // unreachable;
  }
 } while(0);
 $46 = (__ZN4love14luax_checktypeINS_4font10RasterizerEEEPT_P9lua_StateiNS_4TypeE($L,1,9)|0); //@line 445 "modules/graphics/opengl/wrap_Graphics.cpp"
 $47 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $48 = (invoke_ii(3544,($47|0))|0); //@line 448 "modules/graphics/opengl/wrap_Graphics.cpp"
 $49 = __THREW__; __THREW__ = 0;
 $50 = $49&1;
 if (!($50)) {
  __THREW__ = 0;
  $51 = (invoke_iiii(3545,($47|0),($46|0),($48|0))|0); //@line 448 "modules/graphics/opengl/wrap_Graphics.cpp"
  $52 = __THREW__; __THREW__ = 0;
  $53 = $52&1;
  if (!($53)) {
   $font$0 = $51;
   __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,14,$font$0); //@line 452 "modules/graphics/opengl/wrap_Graphics.cpp"
   $72 = HEAP32[$font$0>>2]|0; //@line 453 "modules/graphics/opengl/wrap_Graphics.cpp"
   $73 = ((($72)) + 12|0); //@line 453 "modules/graphics/opengl/wrap_Graphics.cpp"
   $74 = HEAP32[$73>>2]|0; //@line 453 "modules/graphics/opengl/wrap_Graphics.cpp"
   FUNCTION_TABLE_vi[$74 & 8191]($font$0); //@line 453 "modules/graphics/opengl/wrap_Graphics.cpp"
   STACKTOP = sp;return 1; //@line 454 "modules/graphics/opengl/wrap_Graphics.cpp"
  }
 }
 $54 = ___cxa_find_matching_catch_3(8992|0)|0;
 $55 = tempRet0;
 $56 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $57 = ($55|0)==($56|0); //@line 535 "./common/runtime.h"
 if (!($57)) {
  $$01$i = $55;$$02$i = $54;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $58 = (___cxa_begin_catch(($54|0))|0); //@line 535 "./common/runtime.h"
 $59 = HEAP32[$58>>2]|0; //@line 539 "./common/runtime.h"
 $60 = ((($59)) + 8|0); //@line 539 "./common/runtime.h"
 $61 = HEAP32[$60>>2]|0; //@line 539 "./common/runtime.h"
 $62 = (FUNCTION_TABLE_ii[$61 & 8191]($58)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($62|0)); //@line 539 "./common/runtime.h"
 $63 = __THREW__; __THREW__ = 0;
 $64 = $63&1;
 if (!($64)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $65 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $65; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  $font$0 = 0;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,14,$font$0); //@line 452 "modules/graphics/opengl/wrap_Graphics.cpp"
  $72 = HEAP32[$font$0>>2]|0; //@line 453 "modules/graphics/opengl/wrap_Graphics.cpp"
  $73 = ((($72)) + 12|0); //@line 453 "modules/graphics/opengl/wrap_Graphics.cpp"
  $74 = HEAP32[$73>>2]|0; //@line 453 "modules/graphics/opengl/wrap_Graphics.cpp"
  FUNCTION_TABLE_vi[$74 & 8191]($font$0); //@line 453 "modules/graphics/opengl/wrap_Graphics.cpp"
  STACKTOP = sp;return 1; //@line 454 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $66 = ___cxa_find_matching_catch_2()|0;
 $67 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $68 = __THREW__; __THREW__ = 0;
 $69 = $68&1;
 if ($69) {
  $70 = ___cxa_find_matching_catch_3(0|0)|0;
  $71 = tempRet0;
  ___clang_call_terminate($70); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $67;$$02$i = $66;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl14w_newImageFontEP9lua_State($L) {
 $L = $L|0;
 var $$1 = 0, $$cast = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, $filter = 0, $i1$0 = 0, $idxs = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i2 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i3 = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $filter = sp + 24|0;
 $idxs = sp + 8|0;
 $0 = sp + 4|0;
 $1 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $2 = (__ZNK4love8graphics6opengl8Graphics9isCreatedEv($1)|0); //@line 52 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($2)) {
  (_luaL_error($L,329373,$vararg_buffer)|0); //@line 53 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $3 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $4 = (__ZNK4love8graphics6opengl8Graphics16getDefaultFilterEv($3)|0); //@line 462 "modules/graphics/opengl/wrap_Graphics.cpp"
 ;HEAP32[$filter>>2]=HEAP32[$4>>2]|0;HEAP32[$filter+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$filter+8>>2]=HEAP32[$4+8>>2]|0;HEAP32[$filter+12>>2]=HEAP32[$4+12>>2]|0; //@line 462 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,1,12)|0); //@line 465 "modules/graphics/opengl/wrap_Graphics.cpp"
 do {
  if ($5) {
   $6 = (__ZN4love14luax_checktypeINS_8graphics6opengl5ImageEEEPT_P9lua_StateiNS_4TypeE($L,1,12)|0); //@line 467 "modules/graphics/opengl/wrap_Graphics.cpp"
   $7 = HEAP32[$6>>2]|0; //@line 468 "modules/graphics/opengl/wrap_Graphics.cpp"
   $8 = ((($7)) + 36|0); //@line 468 "modules/graphics/opengl/wrap_Graphics.cpp"
   $9 = HEAP32[$8>>2]|0; //@line 468 "modules/graphics/opengl/wrap_Graphics.cpp"
   $10 = (FUNCTION_TABLE_ii[$9 & 8191]($6)|0); //@line 468 "modules/graphics/opengl/wrap_Graphics.cpp"
   ;HEAP32[$filter>>2]=HEAP32[$10>>2]|0;HEAP32[$filter+4>>2]=HEAP32[$10+4>>2]|0;HEAP32[$filter+8>>2]=HEAP32[$10+8>>2]|0;HEAP32[$filter+12>>2]=HEAP32[$10+12>>2]|0; //@line 468 "modules/graphics/opengl/wrap_Graphics.cpp"
   $11 = (__ZNK4love8graphics6opengl5Image12getImageDataEv($6)|0); //@line 469 "modules/graphics/opengl/wrap_Graphics.cpp"
   $12 = HEAP32[$11>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $13 = ((($11)) + 4|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $14 = HEAP32[$13>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $15 = ($12|0)==($14|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($15)) {
    $16 = HEAP32[$12>>2]|0; //@line 139 "./common/Object.h"
    __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,22,$16); //@line 472 "modules/graphics/opengl/wrap_Graphics.cpp"
    _lua_replace($L,1); //@line 473 "modules/graphics/opengl/wrap_Graphics.cpp"
    break;
   }
   $17 = (_luaL_argerror($L,1,329512)|0); //@line 471 "modules/graphics/opengl/wrap_Graphics.cpp"
   $$1 = $17;
   STACKTOP = sp;return ($$1|0); //@line 497 "modules/graphics/opengl/wrap_Graphics.cpp"
  }
 } while(0);
 $18 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,1,9)|0); //@line 477 "modules/graphics/opengl/wrap_Graphics.cpp"
 L11: do {
  if (!($18)) {
   _luaL_checktype($L,2,4); //@line 479 "modules/graphics/opengl/wrap_Graphics.cpp"
   HEAP32[$idxs>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $19 = ((($idxs)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$19>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $20 = ((($idxs)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$20>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $21 = ((($idxs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $i1$0 = 0;
   while(1) {
    __THREW__ = 0;
    $22 = (invoke_ii(3385,($L|0))|0); //@line 482 "modules/graphics/opengl/wrap_Graphics.cpp"
    $23 = __THREW__; __THREW__ = 0;
    $24 = $23&1;
    if ($24) {
     label = 12;
     break;
    }
    $25 = ($i1$0|0)<($22|0); //@line 482 "modules/graphics/opengl/wrap_Graphics.cpp"
    if (!($25)) {
     label = 11;
     break;
    }
    $45 = (($i1$0) + 1)|0; //@line 483 "modules/graphics/opengl/wrap_Graphics.cpp"
    HEAP32[$0>>2] = $45; //@line 483 "modules/graphics/opengl/wrap_Graphics.cpp"
    $46 = HEAP32[$19>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $47 = HEAP32[$21>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $48 = ($46>>>0)<($47>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($48) {
     HEAP32[$46>>2] = $45; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $49 = ((($46)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$19>>2] = $49; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $i1$0 = $45;
     continue;
    } else {
     __THREW__ = 0;
     invoke_vii(3406,($idxs|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $50 = __THREW__; __THREW__ = 0;
     $51 = $50&1;
     if ($51) {
      label = 12;
      break;
     } else {
      $i1$0 = $45;
      continue;
     }
    }
   }
   do {
    if ((label|0) == 11) {
     $26 = HEAP32[$idxs>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $27 = HEAP32[$19>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $$cast = $26; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $28 = (($27) - ($$cast))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $29 = $28 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     __THREW__ = 0;
     (invoke_iiiiii(3455,($L|0),($26|0),($29|0),(329493|0),(329542|0))|0); //@line 485 "modules/graphics/opengl/wrap_Graphics.cpp"
     $30 = __THREW__; __THREW__ = 0;
     $31 = $30&1;
     if ($31) {
      $34 = ___cxa_find_matching_catch_2()|0;
      $35 = tempRet0;
      $lpad$phi$index = $34;$lpad$phi$index2 = $35;
      break;
     }
     $52 = HEAP32[$idxs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $53 = ($52|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $54 = $52; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($53)) {
      $55 = HEAP32[$19>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $56 = ($55|0)==($52|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($56)) {
       $scevgep$i$i$i = ((($55)) + -4|0);
       $57 = $scevgep$i$i$i;
       $58 = (($57) - ($54))|0;
       $59 = $58 >>> 2;
       $60 = $59 ^ -1;
       $scevgep4$i$i$i = (($55) + ($60<<2)|0);
       HEAP32[$19>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($52); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     break L11;
    }
    else if ((label|0) == 12) {
     $32 = ___cxa_find_matching_catch_2()|0;
     $33 = tempRet0;
     $lpad$phi$index = $32;$lpad$phi$index2 = $33;
    }
   } while(0);
   $36 = HEAP32[$idxs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $37 = ($36|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $38 = $36; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($37) {
    ___resumeException($lpad$phi$index|0);
    // unreachable;
   }
   $39 = HEAP32[$19>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $40 = ($39|0)==($36|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($40)) {
    $scevgep$i$i$i2 = ((($39)) + -4|0);
    $41 = $scevgep$i$i$i2;
    $42 = (($41) - ($38))|0;
    $43 = $42 >>> 2;
    $44 = $43 ^ -1;
    $scevgep4$i$i$i3 = (($39) + ($44<<2)|0);
    HEAP32[$19>>2] = $scevgep4$i$i$i3; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($36); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($lpad$phi$index|0);
   // unreachable;
  }
 } while(0);
 $61 = (__ZN4love14luax_checktypeINS_4font10RasterizerEEEPT_P9lua_StateiNS_4TypeE($L,1,9)|0); //@line 488 "modules/graphics/opengl/wrap_Graphics.cpp"
 $62 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $63 = (__ZN4love8graphics6opengl8Graphics7newFontEPNS_4font10RasterizerERKNS0_7Texture6FilterE($62,$61,$filter)|0); //@line 491 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,14,$63); //@line 494 "modules/graphics/opengl/wrap_Graphics.cpp"
 $64 = HEAP32[$63>>2]|0; //@line 495 "modules/graphics/opengl/wrap_Graphics.cpp"
 $65 = ((($64)) + 12|0); //@line 495 "modules/graphics/opengl/wrap_Graphics.cpp"
 $66 = HEAP32[$65>>2]|0; //@line 495 "modules/graphics/opengl/wrap_Graphics.cpp"
 FUNCTION_TABLE_vi[$66 & 8191]($63); //@line 495 "modules/graphics/opengl/wrap_Graphics.cpp"
 $$1 = 1;
 STACKTOP = sp;return ($$1|0); //@line 497 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love14luax_checktypeINS_8graphics6opengl5ImageEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love8graphics6opengl16w_newSpriteBatchEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $$1 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $t$0 = 0, $usage = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $usage = sp + 20|0;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZNK4love8graphics6opengl8Graphics9isCreatedEv($0)|0); //@line 52 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($1)) {
  (_luaL_error($L,329373,$vararg_buffer)|0); //@line 53 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $2 = (__ZN4love8graphics17luax_checktextureEP9lua_Statei($L,1)|0); //@line 503 "modules/graphics/opengl/wrap_Graphics.cpp"
 $3 = (+_luaL_optnumber($L,2,1000.0)); //@line 504 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = (~~(($3))); //@line 504 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAP32[$usage>>2] = 1; //@line 505 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = (_lua_gettop($L)|0); //@line 506 "modules/graphics/opengl/wrap_Graphics.cpp"
 $6 = ($5|0)>(2); //@line 506 "modules/graphics/opengl/wrap_Graphics.cpp"
 do {
  if ($6) {
   $7 = (_luaL_checklstring($L,3,0)|0); //@line 508 "modules/graphics/opengl/wrap_Graphics.cpp"
   $8 = (__ZN4love8graphics6opengl4Mesh11getConstantEPKcRNS2_5UsageE($7,$usage)|0); //@line 509 "modules/graphics/opengl/wrap_Graphics.cpp"
   if ($8) {
    $$pre = HEAP32[$usage>>2]|0; //@line 515 "modules/graphics/opengl/wrap_Graphics.cpp"
    $11 = $$pre;
    break;
   }
   HEAP32[$vararg_buffer1>>2] = $7; //@line 510 "modules/graphics/opengl/wrap_Graphics.cpp"
   $9 = (_luaL_error($L,329561,$vararg_buffer1)|0); //@line 510 "modules/graphics/opengl/wrap_Graphics.cpp"
   $$1 = $9;
   STACKTOP = sp;return ($$1|0); //@line 521 "modules/graphics/opengl/wrap_Graphics.cpp"
  } else {
   $11 = 1;
  }
 } while(0);
 $10 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $12 = (invoke_iiiii(3546,($10|0),($2|0),($4|0),($11|0))|0); //@line 515 "modules/graphics/opengl/wrap_Graphics.cpp"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 do {
  if ($14) {
   $15 = ___cxa_find_matching_catch_3(8992|0)|0;
   $16 = tempRet0;
   $17 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $18 = ($16|0)==($17|0); //@line 535 "./common/runtime.h"
   if (!($18)) {
    $$01$i = $16;$$02$i = $15;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $19 = (___cxa_begin_catch(($15|0))|0); //@line 535 "./common/runtime.h"
   $20 = HEAP32[$19>>2]|0; //@line 539 "./common/runtime.h"
   $21 = ((($20)) + 8|0); //@line 539 "./common/runtime.h"
   $22 = HEAP32[$21>>2]|0; //@line 539 "./common/runtime.h"
   $23 = (FUNCTION_TABLE_ii[$22 & 8191]($19)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($23|0)); //@line 539 "./common/runtime.h"
   $24 = __THREW__; __THREW__ = 0;
   $25 = $24&1;
   if (!($25)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $26 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer3>>2] = $26; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer3)|0); //@line 543 "./common/runtime.h"
    $t$0 = 0;
    break;
   }
   $27 = ___cxa_find_matching_catch_2()|0;
   $28 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $29 = __THREW__; __THREW__ = 0;
   $30 = $29&1;
   if ($30) {
    $31 = ___cxa_find_matching_catch_3(0|0)|0;
    $32 = tempRet0;
    ___clang_call_terminate($31); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $28;$$02$i = $27;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $t$0 = $12;
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,16,$t$0); //@line 518 "modules/graphics/opengl/wrap_Graphics.cpp"
 $33 = HEAP32[$t$0>>2]|0; //@line 519 "modules/graphics/opengl/wrap_Graphics.cpp"
 $34 = ((($33)) + 12|0); //@line 519 "modules/graphics/opengl/wrap_Graphics.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 519 "modules/graphics/opengl/wrap_Graphics.cpp"
 FUNCTION_TABLE_vi[$35 & 8191]($t$0); //@line 519 "modules/graphics/opengl/wrap_Graphics.cpp"
 $$1 = 1;
 STACKTOP = sp;return ($$1|0); //@line 521 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl19w_newParticleSystemEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $t$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0;
 var $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0; //@line 523 "modules/graphics/opengl/wrap_Graphics.cpp"
 $vararg_buffer1 = sp + 8|0; //@line 523 "modules/graphics/opengl/wrap_Graphics.cpp"
 $vararg_buffer = sp; //@line 523 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZNK4love8graphics6opengl8Graphics9isCreatedEv($0)|0); //@line 52 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($1)) {
  (_luaL_error($L,329373,$vararg_buffer)|0); //@line 53 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $2 = (__ZN4love8graphics17luax_checktextureEP9lua_Statei($L,1)|0); //@line 527 "modules/graphics/opengl/wrap_Graphics.cpp"
 $3 = (+_luaL_optnumber($L,2,1000.0)); //@line 528 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = $3 < 1.0; //@line 530 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = $3 > 536870911.0; //@line 530 "modules/graphics/opengl/wrap_Graphics.cpp"
 $or$cond = $4 | $5; //@line 530 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($or$cond) {
  $6 = (_luaL_error($L,329596,$vararg_buffer1)|0); //@line 531 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = $6;
  STACKTOP = sp;return ($$0|0); //@line 540 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $7 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $8 = (~~(($3))); //@line 534 "modules/graphics/opengl/wrap_Graphics.cpp"
 __THREW__ = 0;
 $9 = (invoke_iiii(3547,($7|0),($2|0),($8|0))|0); //@line 534 "modules/graphics/opengl/wrap_Graphics.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 do {
  if ($11) {
   $12 = ___cxa_find_matching_catch_3(8992|0)|0;
   $13 = tempRet0;
   $14 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $15 = ($13|0)==($14|0); //@line 535 "./common/runtime.h"
   if (!($15)) {
    $$01$i = $13;$$02$i = $12;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $16 = (___cxa_begin_catch(($12|0))|0); //@line 535 "./common/runtime.h"
   $17 = HEAP32[$16>>2]|0; //@line 539 "./common/runtime.h"
   $18 = ((($17)) + 8|0); //@line 539 "./common/runtime.h"
   $19 = HEAP32[$18>>2]|0; //@line 539 "./common/runtime.h"
   $20 = (FUNCTION_TABLE_ii[$19 & 8191]($16)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($20|0)); //@line 539 "./common/runtime.h"
   $21 = __THREW__; __THREW__ = 0;
   $22 = $21&1;
   if (!($22)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $23 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer3>>2] = $23; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer3)|0); //@line 543 "./common/runtime.h"
    $t$0 = 0;
    break;
   }
   $24 = ___cxa_find_matching_catch_2()|0;
   $25 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $26 = __THREW__; __THREW__ = 0;
   $27 = $26&1;
   if ($27) {
    $28 = ___cxa_find_matching_catch_3(0|0)|0;
    $29 = tempRet0;
    ___clang_call_terminate($28); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $25;$$02$i = $24;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $t$0 = $9;
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,15,$t$0); //@line 537 "modules/graphics/opengl/wrap_Graphics.cpp"
 $30 = HEAP32[$t$0>>2]|0; //@line 538 "modules/graphics/opengl/wrap_Graphics.cpp"
 $31 = ((($30)) + 12|0); //@line 538 "modules/graphics/opengl/wrap_Graphics.cpp"
 $32 = HEAP32[$31>>2]|0; //@line 538 "modules/graphics/opengl/wrap_Graphics.cpp"
 FUNCTION_TABLE_vi[$32 & 8191]($t$0); //@line 538 "modules/graphics/opengl/wrap_Graphics.cpp"
 $$0 = 1;
 STACKTOP = sp;return ($$0|0); //@line 540 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl11w_newCanvasEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0.0, $format = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $format = sp + 28|0;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZNK4love8graphics6opengl8Graphics9isCreatedEv($0)|0); //@line 52 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($1)) {
  (_luaL_error($L,329373,$vararg_buffer)|0); //@line 53 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $2 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $3 = (__ZNK4love8graphics6opengl8Graphics8getWidthEv($2)|0); //@line 547 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = (+($3|0)); //@line 547 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = (+_luaL_optnumber($L,1,$4)); //@line 547 "modules/graphics/opengl/wrap_Graphics.cpp"
 $6 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $7 = (__ZNK4love8graphics6opengl8Graphics9getHeightEv($6)|0); //@line 548 "modules/graphics/opengl/wrap_Graphics.cpp"
 $8 = (+($7|0)); //@line 548 "modules/graphics/opengl/wrap_Graphics.cpp"
 $9 = (+_luaL_optnumber($L,2,$8)); //@line 548 "modules/graphics/opengl/wrap_Graphics.cpp"
 $10 = (_luaL_optlstring($L,3,520872,0)|0); //@line 549 "modules/graphics/opengl/wrap_Graphics.cpp"
 $11 = (+_luaL_optnumber($L,4,0.0)); //@line 550 "modules/graphics/opengl/wrap_Graphics.cpp"
 $12 = (__ZN4love8graphics6opengl6Canvas11getConstantEPKcRNS2_6FormatE($10,$format)|0); //@line 553 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($12)) {
  HEAP32[$vararg_buffer1>>2] = $10; //@line 554 "modules/graphics/opengl/wrap_Graphics.cpp"
  $13 = (_luaL_error($L,329624,$vararg_buffer1)|0); //@line 554 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$1 = $13;
  STACKTOP = sp;return ($$1|0); //@line 567 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $14 = (~~(($11))); //@line 550 "modules/graphics/opengl/wrap_Graphics.cpp"
 $15 = (~~(($9))); //@line 548 "modules/graphics/opengl/wrap_Graphics.cpp"
 $16 = (~~(($5))); //@line 547 "modules/graphics/opengl/wrap_Graphics.cpp"
 $17 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $18 = HEAP32[$format>>2]|0; //@line 558 "modules/graphics/opengl/wrap_Graphics.cpp"
 __THREW__ = 0;
 $19 = (invoke_iiiiii(3548,($17|0),($16|0),($15|0),($18|0),($14|0))|0); //@line 558 "modules/graphics/opengl/wrap_Graphics.cpp"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 do {
  if ($21) {
   $22 = ___cxa_find_matching_catch_3(8992|0)|0;
   $23 = tempRet0;
   $24 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $25 = ($23|0)==($24|0); //@line 535 "./common/runtime.h"
   if (!($25)) {
    $$01$i = $23;$$02$i = $22;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $26 = (___cxa_begin_catch(($22|0))|0); //@line 535 "./common/runtime.h"
   $27 = HEAP32[$26>>2]|0; //@line 539 "./common/runtime.h"
   $28 = ((($27)) + 8|0); //@line 539 "./common/runtime.h"
   $29 = HEAP32[$28>>2]|0; //@line 539 "./common/runtime.h"
   $30 = (FUNCTION_TABLE_ii[$29 & 8191]($26)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($30|0)); //@line 539 "./common/runtime.h"
   $31 = __THREW__; __THREW__ = 0;
   $32 = $31&1;
   if (!($32)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $33 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer3>>2] = $33; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer3)|0); //@line 543 "./common/runtime.h"
    break;
   }
   $34 = ___cxa_find_matching_catch_2()|0;
   $35 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $36 = __THREW__; __THREW__ = 0;
   $37 = $36&1;
   if ($37) {
    $38 = ___cxa_find_matching_catch_3(0|0)|0;
    $39 = tempRet0;
    ___clang_call_terminate($38); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $35;$$02$i = $34;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $40 = ($19|0)==(0|0); //@line 561 "modules/graphics/opengl/wrap_Graphics.cpp"
   if (!($40)) {
    __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,17,$19); //@line 564 "modules/graphics/opengl/wrap_Graphics.cpp"
    $42 = HEAP32[$19>>2]|0; //@line 565 "modules/graphics/opengl/wrap_Graphics.cpp"
    $43 = ((($42)) + 12|0); //@line 565 "modules/graphics/opengl/wrap_Graphics.cpp"
    $44 = HEAP32[$43>>2]|0; //@line 565 "modules/graphics/opengl/wrap_Graphics.cpp"
    FUNCTION_TABLE_vi[$44 & 8191]($19); //@line 565 "modules/graphics/opengl/wrap_Graphics.cpp"
    $$1 = 1;
    STACKTOP = sp;return ($$1|0); //@line 567 "modules/graphics/opengl/wrap_Graphics.cpp"
   }
  }
 } while(0);
 $41 = (_luaL_error($L,329650,$vararg_buffer6)|0); //@line 562 "modules/graphics/opengl/wrap_Graphics.cpp"
 $$1 = $41;
 STACKTOP = sp;return ($$1|0); //@line 567 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl11w_newShaderEP9lua_State($L) {
 $L = $L|0;
 var $$05 = 0, $$06 = 0, $$10 = 0, $$12 = 0, $$lcssa = 0, $$off = 0, $$off$1 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $brmerge13$demorgan = 0, $brmerge16$demorgan = 0, $i4$034$lcssa36 = 0, $lpad$phi$index = 0, $lpad$phi$index11 = 0, $slen = 0, $source = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer6 = 0, $vararg_buffer8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer8 = sp + 32|0;
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $slen = sp + 88|0;
 $source = sp + 64|0;
 $0 = sp + 48|0;
 $1 = sp + 36|0;
 $2 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $3 = (__ZNK4love8graphics6opengl8Graphics9isCreatedEv($2)|0); //@line 52 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($3)) {
  (_luaL_error($L,329373,$vararg_buffer)|0); //@line 53 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 _lua_settop($L,2); //@line 574 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = (_lua_isstring($L,1)|0); //@line 579 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = ($4|0)==(0); //@line 579 "modules/graphics/opengl/wrap_Graphics.cpp"
 do {
  if ($5) {
   label = 12;
  } else {
   (__ZN4love16luax_getfunctionEP9lua_StatePKcS3_($L,354839,354850)|0); //@line 583 "modules/graphics/opengl/wrap_Graphics.cpp"
   _lua_pushvalue($L,1); //@line 584 "modules/graphics/opengl/wrap_Graphics.cpp"
   _lua_call($L,1,1); //@line 585 "modules/graphics/opengl/wrap_Graphics.cpp"
   $6 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,-1)|0); //@line 587 "modules/graphics/opengl/wrap_Graphics.cpp"
   _lua_settop($L,-2); //@line 588 "modules/graphics/opengl/wrap_Graphics.cpp"
   if ($6) {
    (__ZN4love16luax_getfunctionEP9lua_StatePKcS3_($L,354839,521561)|0); //@line 592 "modules/graphics/opengl/wrap_Graphics.cpp"
    _lua_pushvalue($L,1); //@line 593 "modules/graphics/opengl/wrap_Graphics.cpp"
    _lua_call($L,1,1); //@line 594 "modules/graphics/opengl/wrap_Graphics.cpp"
    _lua_replace($L,1); //@line 595 "modules/graphics/opengl/wrap_Graphics.cpp"
    label = 12;
    break;
   }
   HEAP32[$slen>>2] = 0; //@line 601 "modules/graphics/opengl/wrap_Graphics.cpp"
   $7 = (_lua_tolstring($L,1,$slen)|0); //@line 602 "modules/graphics/opengl/wrap_Graphics.cpp"
   $8 = HEAP32[$slen>>2]|0; //@line 603 "modules/graphics/opengl/wrap_Graphics.cpp"
   $$off = (($8) + -1)|0; //@line 603 "modules/graphics/opengl/wrap_Graphics.cpp"
   $9 = ($$off>>>0)<(255); //@line 603 "modules/graphics/opengl/wrap_Graphics.cpp"
   if ($9) {
    $10 = (_strchr($7,10)|0); //@line 603 "modules/graphics/opengl/wrap_Graphics.cpp"
    $11 = ($10|0)==(0|0); //@line 603 "modules/graphics/opengl/wrap_Graphics.cpp"
    if ($11) {
     $12 = (_strchr($7,46)|0); //@line 605 "modules/graphics/opengl/wrap_Graphics.cpp"
     $13 = ($12|0)==(0|0); //@line 606 "modules/graphics/opengl/wrap_Graphics.cpp"
     if (!($13)) {
      $14 = (_strchr($12,59)|0); //@line 606 "modules/graphics/opengl/wrap_Graphics.cpp"
      $15 = ($14|0)==(0|0); //@line 606 "modules/graphics/opengl/wrap_Graphics.cpp"
      if ($15) {
       $16 = (_strchr($12,32)|0); //@line 606 "modules/graphics/opengl/wrap_Graphics.cpp"
       $17 = ($16|0)==(0|0); //@line 606 "modules/graphics/opengl/wrap_Graphics.cpp"
       if ($17) {
        $$lcssa = $7;
        break;
       }
      }
     }
    }
   }
   label = 12;
  }
 } while(0);
 L15: do {
  if ((label|0) == 12) {
   $18 = (_lua_isstring($L,2)|0); //@line 579 "modules/graphics/opengl/wrap_Graphics.cpp"
   $19 = ($18|0)==(0); //@line 579 "modules/graphics/opengl/wrap_Graphics.cpp"
   do {
    if (!($19)) {
     (__ZN4love16luax_getfunctionEP9lua_StatePKcS3_($L,354839,354850)|0); //@line 583 "modules/graphics/opengl/wrap_Graphics.cpp"
     _lua_pushvalue($L,2); //@line 584 "modules/graphics/opengl/wrap_Graphics.cpp"
     _lua_call($L,1,1); //@line 585 "modules/graphics/opengl/wrap_Graphics.cpp"
     $141 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,-1)|0); //@line 587 "modules/graphics/opengl/wrap_Graphics.cpp"
     _lua_settop($L,-2); //@line 588 "modules/graphics/opengl/wrap_Graphics.cpp"
     if ($141) {
      (__ZN4love16luax_getfunctionEP9lua_StatePKcS3_($L,354839,521561)|0); //@line 592 "modules/graphics/opengl/wrap_Graphics.cpp"
      _lua_pushvalue($L,2); //@line 593 "modules/graphics/opengl/wrap_Graphics.cpp"
      _lua_call($L,1,1); //@line 594 "modules/graphics/opengl/wrap_Graphics.cpp"
      _lua_replace($L,2); //@line 595 "modules/graphics/opengl/wrap_Graphics.cpp"
      break;
     }
     HEAP32[$slen>>2] = 0; //@line 601 "modules/graphics/opengl/wrap_Graphics.cpp"
     $142 = (_lua_tolstring($L,2,$slen)|0); //@line 602 "modules/graphics/opengl/wrap_Graphics.cpp"
     $143 = HEAP32[$slen>>2]|0; //@line 603 "modules/graphics/opengl/wrap_Graphics.cpp"
     $$off$1 = (($143) + -1)|0; //@line 603 "modules/graphics/opengl/wrap_Graphics.cpp"
     $144 = ($$off$1>>>0)<(255); //@line 603 "modules/graphics/opengl/wrap_Graphics.cpp"
     if ($144) {
      $145 = (_strchr($142,10)|0); //@line 603 "modules/graphics/opengl/wrap_Graphics.cpp"
      $146 = ($145|0)==(0|0); //@line 603 "modules/graphics/opengl/wrap_Graphics.cpp"
      if ($146) {
       $147 = (_strchr($142,46)|0); //@line 605 "modules/graphics/opengl/wrap_Graphics.cpp"
       $148 = ($147|0)==(0|0); //@line 606 "modules/graphics/opengl/wrap_Graphics.cpp"
       if (!($148)) {
        $149 = (_strchr($147,59)|0); //@line 606 "modules/graphics/opengl/wrap_Graphics.cpp"
        $150 = ($149|0)==(0|0); //@line 606 "modules/graphics/opengl/wrap_Graphics.cpp"
        if ($150) {
         $151 = (_strchr($147,32)|0); //@line 606 "modules/graphics/opengl/wrap_Graphics.cpp"
         $152 = ($151|0)==(0|0); //@line 606 "modules/graphics/opengl/wrap_Graphics.cpp"
         if ($152) {
          $$lcssa = $142;
          break L15;
         }
        }
       }
      }
     }
    }
   } while(0);
   $153 = (_lua_isstring($L,1)|0); //@line 612 "modules/graphics/opengl/wrap_Graphics.cpp"
   $46 = ($153|0)!=(0); //@line 612 "modules/graphics/opengl/wrap_Graphics.cpp"
   $154 = (_lua_isstring($L,2)|0); //@line 613 "modules/graphics/opengl/wrap_Graphics.cpp"
   $47 = ($154|0)!=(0); //@line 613 "modules/graphics/opengl/wrap_Graphics.cpp"
   $155 = $154 | $153; //@line 616 "modules/graphics/opengl/wrap_Graphics.cpp"
   $156 = ($155|0)==(0); //@line 616 "modules/graphics/opengl/wrap_Graphics.cpp"
   if ($156) {
    (_luaL_checklstring($L,1,0)|0); //@line 617 "modules/graphics/opengl/wrap_Graphics.cpp"
   }
   (__ZN4love16luax_getfunctionEP9lua_StatePKcS3_($L,329707,329716)|0); //@line 619 "modules/graphics/opengl/wrap_Graphics.cpp"
   _lua_pushvalue($L,1); //@line 622 "modules/graphics/opengl/wrap_Graphics.cpp"
   _lua_pushvalue($L,2); //@line 623 "modules/graphics/opengl/wrap_Graphics.cpp"
   $21 = (_lua_pcall($L,2,2,0)|0); //@line 626 "modules/graphics/opengl/wrap_Graphics.cpp"
   $22 = ($21|0)==(0); //@line 626 "modules/graphics/opengl/wrap_Graphics.cpp"
   if (!($22)) {
    $23 = (_lua_tolstring($L,-1,0)|0); //@line 627 "modules/graphics/opengl/wrap_Graphics.cpp"
    HEAP32[$vararg_buffer3>>2] = $23; //@line 627 "modules/graphics/opengl/wrap_Graphics.cpp"
    $24 = (_luaL_error($L,595867,$vararg_buffer3)|0); //@line 627 "modules/graphics/opengl/wrap_Graphics.cpp"
    $$12 = $24;
    STACKTOP = sp;return ($$12|0); //@line 674 "modules/graphics/opengl/wrap_Graphics.cpp"
   }
   ;HEAP32[$source>>2]=0|0;HEAP32[$source+4>>2]=0|0;HEAP32[$source+8>>2]=0|0;HEAP32[$source+12>>2]=0|0;HEAP32[$source+16>>2]=0|0;HEAP32[$source+20>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   $25 = (invoke_iii(3453,($L|0),-2)|0); //@line 632 "modules/graphics/opengl/wrap_Graphics.cpp"
   $26 = __THREW__; __THREW__ = 0;
   $27 = $26&1;
   L36: do {
    if ($27) {
     label = 27;
    } else {
     $28 = ($25|0)==(0); //@line 632 "modules/graphics/opengl/wrap_Graphics.cpp"
     do {
      if ($28) {
       $brmerge13$demorgan = $46 & $47; //@line 634 "modules/graphics/opengl/wrap_Graphics.cpp"
       if ($brmerge13$demorgan) {
        __THREW__ = 0;
        $48 = (invoke_iiii(3389,($L|0),(329734|0),($vararg_buffer6|0))|0); //@line 635 "modules/graphics/opengl/wrap_Graphics.cpp"
        $49 = __THREW__; __THREW__ = 0;
        $50 = $49&1;
        if ($50) {
         label = 27;
         break L36;
        } else {
         $$10 = $48;
        }
       } else {
        label = 31;
       }
      } else {
       __THREW__ = 0;
       invoke_viii(3403,($0|0),($L|0),-2); //@line 633 "modules/graphics/opengl/wrap_Graphics.cpp"
       $29 = __THREW__; __THREW__ = 0;
       $30 = $29&1;
       if ($30) {
        label = 27;
        break L36;
       }
       $31 = HEAP8[$source>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $32 = $31 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $33 = ($32<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($33) {
        $37 = ((($source)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP8[$37>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP8[$source>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       } else {
        $34 = ((($source)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $35 = HEAP32[$34>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP8[$35>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $36 = ((($source)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP32[$36>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       }
       __THREW__ = 0;
       invoke_vii(3368,($source|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $38 = __THREW__; __THREW__ = 0;
       $39 = $38&1;
       if ($39) {
        $40 = ___cxa_find_matching_catch_3(0|0)|0;
        $41 = tempRet0;
        ___clang_call_terminate($40); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       } else {
        ;HEAP32[$source>>2]=HEAP32[$0>>2]|0;HEAP32[$source+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$source+8>>2]=HEAP32[$0+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 633 "modules/graphics/opengl/wrap_Graphics.cpp"
        label = 31;
        break;
       }
      }
     } while(0);
     L50: do {
      if ((label|0) == 31) {
       __THREW__ = 0;
       $51 = (invoke_iii(3453,($L|0),-1)|0); //@line 638 "modules/graphics/opengl/wrap_Graphics.cpp"
       $52 = __THREW__; __THREW__ = 0;
       $53 = $52&1;
       if ($53) {
        label = 27;
        break L36;
       }
       $54 = ($51|0)==(0); //@line 638 "modules/graphics/opengl/wrap_Graphics.cpp"
       do {
        if ($54) {
         $brmerge16$demorgan = $46 & $47; //@line 640 "modules/graphics/opengl/wrap_Graphics.cpp"
         if ($brmerge16$demorgan) {
          __THREW__ = 0;
          $69 = (invoke_iiii(3389,($L|0),(329800|0),($vararg_buffer8|0))|0); //@line 641 "modules/graphics/opengl/wrap_Graphics.cpp"
          $70 = __THREW__; __THREW__ = 0;
          $71 = $70&1;
          if ($71) {
           label = 27;
           break L36;
          } else {
           $$10 = $69;
           break L50;
          }
         }
        } else {
         __THREW__ = 0;
         invoke_viii(3403,($1|0),($L|0),-1); //@line 639 "modules/graphics/opengl/wrap_Graphics.cpp"
         $55 = __THREW__; __THREW__ = 0;
         $56 = $55&1;
         if ($56) {
          label = 27;
          break L36;
         }
         $57 = ((($source)) + 12|0); //@line 639 "modules/graphics/opengl/wrap_Graphics.cpp"
         $58 = HEAP8[$57>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $59 = $58 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $60 = ($59<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         if ($60) {
          $64 = ((($57)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          HEAP8[$64>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          HEAP8[$57>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         } else {
          $61 = ((($source)) + 20|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $62 = HEAP32[$61>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          HEAP8[$62>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $63 = ((($source)) + 16|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          HEAP32[$63>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         }
         __THREW__ = 0;
         invoke_vii(3368,($57|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $65 = __THREW__; __THREW__ = 0;
         $66 = $65&1;
         if ($66) {
          $67 = ___cxa_find_matching_catch_3(0|0)|0;
          $68 = tempRet0;
          ___clang_call_terminate($67); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         } else {
          ;HEAP32[$57>>2]=HEAP32[$1>>2]|0;HEAP32[$57+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$57+8>>2]=HEAP32[$1+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 639 "modules/graphics/opengl/wrap_Graphics.cpp"
          break;
         }
        }
       } while(0);
       $72 = HEAP8[$source>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $73 = $72 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $74 = ($73<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $75 = ((($source)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $76 = HEAP32[$75>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $77 = $72&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $78 = $77 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $79 = $74 ? $78 : $76; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $80 = ($79|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       L66: do {
        if ($80) {
         $81 = ((($source)) + 12|0); //@line 643 "modules/graphics/opengl/wrap_Graphics.cpp"
         $82 = HEAP8[$81>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $83 = $82 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $84 = ($83<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $85 = ((($source)) + 16|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $86 = HEAP32[$85>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $87 = $82&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $88 = $87 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $89 = $84 ? $88 : $86; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $90 = ($89|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         if ($90) {
          __THREW__ = 0;
          $91 = (invoke_iii(3453,($L|0),1)|0); //@line 648 "modules/graphics/opengl/wrap_Graphics.cpp"
          $92 = __THREW__; __THREW__ = 0;
          $93 = $92&1;
          do {
           if (!($93)) {
            $94 = ($91|0)==(0); //@line 648 "modules/graphics/opengl/wrap_Graphics.cpp"
            if ($94) {
             __THREW__ = 0;
             $98 = (invoke_iii(3453,($L|0),2)|0); //@line 648 "modules/graphics/opengl/wrap_Graphics.cpp"
             $99 = __THREW__; __THREW__ = 0;
             $100 = $99&1;
             if ($100) {
              break;
             }
             $140 = ($98|0)==(0); //@line 648 "modules/graphics/opengl/wrap_Graphics.cpp"
             if ($140) {
              break L66;
             } else {
              $i4$034$lcssa36 = 2;
             }
            } else {
             $i4$034$lcssa36 = 1;
            }
            __THREW__ = 0;
            $95 = (invoke_iiii(3549,($L|0),($i4$034$lcssa36|0),(329863|0))|0); //@line 649 "modules/graphics/opengl/wrap_Graphics.cpp"
            $96 = __THREW__; __THREW__ = 0;
            $97 = $96&1;
            if ($97) {
             label = 27;
             break L36;
            } else {
             $$10 = $95;
             break L50;
            }
           }
          } while(0);
          $42 = ___cxa_find_matching_catch_2()|0;
          $43 = tempRet0;
          $lpad$phi$index = $42;$lpad$phi$index11 = $43;
          break L36;
         }
        }
       } while(0);
       $101 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
       __THREW__ = 0;
       $102 = (invoke_iii(3550,($101|0),($source|0))|0); //@line 656 "modules/graphics/opengl/wrap_Graphics.cpp"
       $103 = __THREW__; __THREW__ = 0;
       $104 = $103&1;
       if (!($104)) {
        __THREW__ = 0;
        invoke_viii(3387,($L|0),18,($102|0)); //@line 657 "modules/graphics/opengl/wrap_Graphics.cpp"
        $105 = __THREW__; __THREW__ = 0;
        $106 = $105&1;
        if (!($106)) {
         $107 = HEAP32[$102>>2]|0; //@line 658 "modules/graphics/opengl/wrap_Graphics.cpp"
         $108 = ((($107)) + 12|0); //@line 658 "modules/graphics/opengl/wrap_Graphics.cpp"
         $109 = HEAP32[$108>>2]|0; //@line 658 "modules/graphics/opengl/wrap_Graphics.cpp"
         __THREW__ = 0;
         invoke_vi($109|0,($102|0)); //@line 658 "modules/graphics/opengl/wrap_Graphics.cpp"
         $110 = __THREW__; __THREW__ = 0;
         $111 = $110&1;
         if (!($111)) {
          $$10 = 1;
          break;
         }
        }
       }
       $112 = ___cxa_find_matching_catch_3(96|0)|0;
       $113 = tempRet0;
       $114 = (_llvm_eh_typeid_for((96|0))|0); //@line 659 "modules/graphics/opengl/wrap_Graphics.cpp"
       $115 = ($113|0)==($114|0); //@line 659 "modules/graphics/opengl/wrap_Graphics.cpp"
       if (!($115)) {
        $$05 = $113;$$06 = $112;
        $137 = ((($source)) + 12|0); //@line 84 "modules/graphics/opengl/Shader.h"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($137); //@line 84 "modules/graphics/opengl/Shader.h"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($source); //@line 84 "modules/graphics/opengl/Shader.h"
        ___resumeException($$06|0);
        // unreachable;
       }
       $116 = (___cxa_begin_catch(($112|0))|0); //@line 659 "modules/graphics/opengl/wrap_Graphics.cpp"
       __THREW__ = 0;
       (invoke_iiii(3551,($L|0),(329707|0),(329904|0))|0); //@line 662 "modules/graphics/opengl/wrap_Graphics.cpp"
       $117 = __THREW__; __THREW__ = 0;
       $118 = $117&1;
       if (!($118)) {
        $119 = HEAP32[$116>>2]|0; //@line 663 "modules/graphics/opengl/wrap_Graphics.cpp"
        $120 = ((($119)) + 8|0); //@line 663 "modules/graphics/opengl/wrap_Graphics.cpp"
        $121 = HEAP32[$120>>2]|0; //@line 663 "modules/graphics/opengl/wrap_Graphics.cpp"
        $122 = (FUNCTION_TABLE_ii[$121 & 8191]($116)|0); //@line 663 "modules/graphics/opengl/wrap_Graphics.cpp"
        __THREW__ = 0;
        invoke_vii(3380,($L|0),($122|0)); //@line 663 "modules/graphics/opengl/wrap_Graphics.cpp"
        $123 = __THREW__; __THREW__ = 0;
        $124 = $123&1;
        if (!($124)) {
         __THREW__ = 0;
         (invoke_iiiii(3552,($L|0),1,1,0)|0); //@line 666 "modules/graphics/opengl/wrap_Graphics.cpp"
         $125 = __THREW__; __THREW__ = 0;
         $126 = $125&1;
         if (!($126)) {
          __THREW__ = 0;
          invoke_v(3378); //@line 668 "modules/graphics/opengl/wrap_Graphics.cpp"
          $127 = __THREW__; __THREW__ = 0;
          $128 = $127&1;
          if ($128) {
           label = 27;
           break L36;
          }
          __THREW__ = 0;
          $129 = (invoke_ii(3553,($L|0))|0); //@line 671 "modules/graphics/opengl/wrap_Graphics.cpp"
          $130 = __THREW__; __THREW__ = 0;
          $131 = $130&1;
          if ($131) {
           label = 27;
           break L36;
          } else {
           $$10 = $129;
           break;
          }
         }
        }
       }
       $132 = ___cxa_find_matching_catch_2()|0;
       $133 = tempRet0;
       __THREW__ = 0;
       invoke_v(3378); //@line 668 "modules/graphics/opengl/wrap_Graphics.cpp"
       $134 = __THREW__; __THREW__ = 0;
       $135 = $134&1;
       if ($135) {
        $138 = ___cxa_find_matching_catch_3(0|0)|0;
        $139 = tempRet0;
        ___clang_call_terminate($138); //@line 668 "modules/graphics/opengl/wrap_Graphics.cpp"
        // unreachable; //@line 668 "modules/graphics/opengl/wrap_Graphics.cpp"
       } else {
        $$05 = $133;$$06 = $132;
        $137 = ((($source)) + 12|0); //@line 84 "modules/graphics/opengl/Shader.h"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($137); //@line 84 "modules/graphics/opengl/Shader.h"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($source); //@line 84 "modules/graphics/opengl/Shader.h"
        ___resumeException($$06|0);
        // unreachable;
       }
      }
     } while(0);
     $136 = ((($source)) + 12|0); //@line 84 "modules/graphics/opengl/Shader.h"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($136); //@line 84 "modules/graphics/opengl/Shader.h"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($source); //@line 84 "modules/graphics/opengl/Shader.h"
     $$12 = $$10;
     STACKTOP = sp;return ($$12|0); //@line 674 "modules/graphics/opengl/wrap_Graphics.cpp"
    }
   } while(0);
   if ((label|0) == 27) {
    $44 = ___cxa_find_matching_catch_2()|0;
    $45 = tempRet0;
    $lpad$phi$index = $44;$lpad$phi$index11 = $45;
   }
   $$05 = $lpad$phi$index11;$$06 = $lpad$phi$index;
   $137 = ((($source)) + 12|0); //@line 84 "modules/graphics/opengl/Shader.h"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($137); //@line 84 "modules/graphics/opengl/Shader.h"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($source); //@line 84 "modules/graphics/opengl/Shader.h"
   ___resumeException($$06|0);
   // unreachable;
  }
 } while(0);
 HEAP32[$vararg_buffer1>>2] = $$lcssa; //@line 607 "modules/graphics/opengl/wrap_Graphics.cpp"
 $20 = (_luaL_error($L,507142,$vararg_buffer1)|0); //@line 607 "modules/graphics/opengl/wrap_Graphics.cpp"
 $$12 = $20;
 STACKTOP = sp;return ($$12|0); //@line 674 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl9w_newMeshEP9lua_State($L) {
 $L = $L|0;
 var $$01$i$i = 0, $$01$i$i13 = 0, $$01$i14$i = 0, $$01$i21$i = 0, $$01$i28$i = 0, $$01$i35$i = 0, $$02$i = 0, $$02$i$i = 0, $$02$i$i12 = 0, $$02$i13$i = 0, $$02$i20$i = 0, $$02$i27$i = 0, $$02$i34$i = 0, $$03$i = 0, $$1$i = 0, $$3$i = 0, $$in$i = 0, $$lcssa49 = 0, $$lcssa53 = 0, $$lcssa55 = 0;
 var $$off$i = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i44$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0.0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0;
 var $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0;
 var $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0;
 var $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0;
 var $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0;
 var $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0;
 var $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0;
 var $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0.0, $398 = 0.0, $399 = 0.0, $4 = 0, $40 = 0;
 var $400 = 0, $401 = 0, $402 = 0.0, $403 = 0.0, $404 = 0, $405 = 0, $406 = 0.0, $407 = 0.0, $408 = 0, $409 = 0, $41 = 0, $410 = 0.0, $411 = 0.0, $412 = 0, $413 = 0, $414 = 0, $415 = 0.0, $416 = 0, $417 = 0, $418 = 0;
 var $419 = 0.0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0.0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0;
 var $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0;
 var $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0;
 var $473 = 0, $474 = 0.0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0;
 var $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0;
 var $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0.0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $c$094$i = 0;
 var $cond1$i = 0, $data3$i = 0, $eh$lpad$body$i$index45Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $eh$lpad$body$i10$index65Z2D = 0, $eh$lpad$body$i10$indexZ2D = 0, $format$i = 0, $i$0$i = 0, $i$024$i = 0, $lpad$phi80$i$index = 0, $lpad$phi80$i$index58 = 0, $n$096$i = 0, $n$1$lcssa$i = 0, $n$193$i = 0, $scevgep$i$i$i$i = 0, $scevgep$i$i$i8$i = 0, $scevgep4$i$i$i$i = 0, $scevgep4$i$i$i9$i = 0, $storemerge$i = 0, $storemerge695$i = 0;
 var $t$0 = 0, $t$1$i = 0, $t$2$i = 0, $t$3$i = 0, $v$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer12 = 0, $vararg_buffer15 = 0, $vararg_buffer18 = 0, $vararg_buffer21 = 0, $vararg_buffer24 = 0, $vararg_buffer27 = 0, $vararg_buffer3 = 0, $vararg_buffer30 = 0, $vararg_buffer33 = 0, $vararg_buffer6 = 0, $vararg_buffer9 = 0, $vertexformat$i = 0, $vertices$i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 208|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer33 = sp + 96|0;
 $vararg_buffer30 = sp + 88|0;
 $vararg_buffer27 = sp + 80|0;
 $vararg_buffer24 = sp + 72|0;
 $vararg_buffer21 = sp + 64|0;
 $vararg_buffer18 = sp + 56|0;
 $vararg_buffer15 = sp + 48|0;
 $vararg_buffer12 = sp + 40|0;
 $vararg_buffer9 = sp + 32|0;
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 176|0;
 $1 = sp + 172|0;
 $vertices$i = sp + 160|0;
 $v$i = sp + 140|0;
 $2 = sp + 136|0;
 $3 = sp + 132|0;
 $vertexformat$i = sp + 120|0;
 $format$i = sp + 100|0;
 $data3$i = sp + 184|0;
 $4 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $5 = (__ZNK4love8graphics6opengl8Graphics9isCreatedEv($4)|0); //@line 52 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($5)) {
  (_luaL_error($L,329373,$vararg_buffer)|0); //@line 53 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $6 = (_lua_type($L,1)|0); //@line 877 "modules/graphics/opengl/wrap_Graphics.cpp"
 switch ($6|0) {
 case 3: case 5:  {
  break;
 }
 default: {
  (_luaL_argerror($L,1,329932)|0); //@line 879 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 }
 $7 = (_lua_type($L,2)|0); //@line 883 "modules/graphics/opengl/wrap_Graphics.cpp"
 $8 = ($6|0)==(5); //@line 884 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($8) {
  $9 = ($7|0)==(3); //@line 884 "modules/graphics/opengl/wrap_Graphics.cpp"
  $10 = $7 | 2; //@line 884 "modules/graphics/opengl/wrap_Graphics.cpp"
  $11 = ($10|0)==(7); //@line 884 "modules/graphics/opengl/wrap_Graphics.cpp"
  $12 = $9 | $11; //@line 884 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($12) {
   HEAP32[$vertexformat$i>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $13 = ((($vertexformat$i)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$13>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $14 = ((($vertexformat$i)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$14>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$3>>2] = 0; //@line 760 "modules/graphics/opengl/wrap_Graphics.cpp"
   __THREW__ = 0;
   $15 = (invoke_iii(3358,($L|0),3)|0); //@line 688 "modules/graphics/opengl/wrap_Graphics.cpp"
   $16 = __THREW__; __THREW__ = 0;
   $17 = $16&1;
   L10: do {
    if ($17) {
     label = 30;
    } else {
     $18 = ($15|0)<(1); //@line 688 "modules/graphics/opengl/wrap_Graphics.cpp"
     if (!($18)) {
      __THREW__ = 0;
      $19 = (invoke_iiii(3359,($L|0),3,(0|0))|0); //@line 688 "modules/graphics/opengl/wrap_Graphics.cpp"
      $20 = __THREW__; __THREW__ = 0;
      $21 = $20&1;
      if ($21) {
       label = 30;
       break;
      }
      $22 = ($19|0)==(0|0); //@line 690 "modules/graphics/opengl/wrap_Graphics.cpp"
      if (!($22)) {
       __THREW__ = 0;
       $23 = (invoke_iii(3554,($19|0),($3|0))|0); //@line 690 "modules/graphics/opengl/wrap_Graphics.cpp"
       $24 = __THREW__; __THREW__ = 0;
       $25 = $24&1;
       if ($25) {
        label = 30;
        break;
       }
       if (!($23)) {
        __THREW__ = 0;
        HEAP32[$vararg_buffer1>>2] = $19; //@line 691 "modules/graphics/opengl/wrap_Graphics.cpp"
        (invoke_iiii(3389,($L|0),(343313|0),($vararg_buffer1|0))|0); //@line 691 "modules/graphics/opengl/wrap_Graphics.cpp"
        $26 = __THREW__; __THREW__ = 0;
        $27 = $26&1;
        if ($27) {
         label = 30;
         break;
        }
       }
      }
     }
     $28 = HEAP32[$3>>2]|0; //@line 693 "modules/graphics/opengl/wrap_Graphics.cpp"
     HEAP32[$2>>2] = 1; //@line 761 "modules/graphics/opengl/wrap_Graphics.cpp"
     __THREW__ = 0;
     $29 = (invoke_iii(3358,($L|0),4)|0); //@line 678 "modules/graphics/opengl/wrap_Graphics.cpp"
     $30 = __THREW__; __THREW__ = 0;
     $31 = $30&1;
     if ($31) {
      label = 30;
     } else {
      $32 = ($29|0)<(1); //@line 678 "modules/graphics/opengl/wrap_Graphics.cpp"
      if (!($32)) {
       __THREW__ = 0;
       $33 = (invoke_iiii(3359,($L|0),4,(0|0))|0); //@line 678 "modules/graphics/opengl/wrap_Graphics.cpp"
       $34 = __THREW__; __THREW__ = 0;
       $35 = $34&1;
       if ($35) {
        label = 30;
        break;
       }
       $36 = ($33|0)==(0|0); //@line 680 "modules/graphics/opengl/wrap_Graphics.cpp"
       if (!($36)) {
        __THREW__ = 0;
        $37 = (invoke_iii(3555,($33|0),($2|0))|0); //@line 680 "modules/graphics/opengl/wrap_Graphics.cpp"
        $38 = __THREW__; __THREW__ = 0;
        $39 = $38&1;
        if ($39) {
         label = 30;
         break;
        }
        if (!($37)) {
         __THREW__ = 0;
         HEAP32[$vararg_buffer3>>2] = $33; //@line 681 "modules/graphics/opengl/wrap_Graphics.cpp"
         (invoke_iiii(3389,($L|0),(329957|0),($vararg_buffer3|0))|0); //@line 681 "modules/graphics/opengl/wrap_Graphics.cpp"
         $40 = __THREW__; __THREW__ = 0;
         $41 = $40&1;
         if ($41) {
          label = 30;
          break;
         }
        }
       }
      }
      $42 = HEAP32[$2>>2]|0; //@line 683 "modules/graphics/opengl/wrap_Graphics.cpp"
      __THREW__ = 0;
      invoke_viii(3382,($L|0),1,1); //@line 763 "modules/graphics/opengl/wrap_Graphics.cpp"
      $43 = __THREW__; __THREW__ = 0;
      $44 = $43&1;
      if ($44) {
       label = 30;
      } else {
       __THREW__ = 0;
       $45 = (invoke_iii(3358,($L|0),-1)|0); //@line 764 "modules/graphics/opengl/wrap_Graphics.cpp"
       $46 = __THREW__; __THREW__ = 0;
       $47 = $46&1;
       if ($47) {
        label = 30;
       } else {
        $48 = ($45|0)==(5); //@line 764 "modules/graphics/opengl/wrap_Graphics.cpp"
        L29: do {
         if ($48) {
          __THREW__ = 0;
          invoke_vii(3384,($L|0),-2); //@line 769 "modules/graphics/opengl/wrap_Graphics.cpp"
          $63 = __THREW__; __THREW__ = 0;
          $64 = $63&1;
          if ($64) {
           label = 30;
           break L10;
          }
          $65 = ((($format$i)) + 12|0); //@line 784 "modules/graphics/opengl/wrap_Graphics.cpp"
          $66 = ((($format$i)) + 16|0); //@line 790 "modules/graphics/opengl/wrap_Graphics.cpp"
          $67 = ((($vertexformat$i)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $i$0$i = 1;
          while(1) {
           __THREW__ = 0;
           $68 = (invoke_iii(3381,($L|0),1)|0); //@line 772 "modules/graphics/opengl/wrap_Graphics.cpp"
           $69 = __THREW__; __THREW__ = 0;
           $70 = $69&1;
           if ($70) {
            label = 29;
            break;
           }
           $71 = ($i$0$i|0)>($68|0); //@line 772 "modules/graphics/opengl/wrap_Graphics.cpp"
           if ($71) {
            label = 56;
            break;
           }
           __THREW__ = 0;
           invoke_viii(3382,($L|0),1,($i$0$i|0)); //@line 774 "modules/graphics/opengl/wrap_Graphics.cpp"
           $72 = __THREW__; __THREW__ = 0;
           $73 = $72&1;
           if ($73) {
            label = 29;
            break;
           }
           __THREW__ = 0;
           invoke_viii(3382,($L|0),-1,1); //@line 778 "modules/graphics/opengl/wrap_Graphics.cpp"
           $74 = __THREW__; __THREW__ = 0;
           $75 = $74&1;
           if ($75) {
            label = 28;
            break;
           }
           __THREW__ = 0;
           invoke_viii(3382,($L|0),-2,2); //@line 778 "modules/graphics/opengl/wrap_Graphics.cpp"
           $76 = __THREW__; __THREW__ = 0;
           $77 = $76&1;
           if ($77) {
            label = 28;
            break;
           }
           __THREW__ = 0;
           invoke_viii(3382,($L|0),-3,3); //@line 778 "modules/graphics/opengl/wrap_Graphics.cpp"
           $333 = __THREW__; __THREW__ = 0;
           $334 = $333&1;
           if ($334) {
            label = 28;
            break;
           }
           ;HEAP32[$format$i>>2]=0|0;HEAP32[$format$i+4>>2]=0|0;HEAP32[$format$i+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           __THREW__ = 0;
           $78 = (invoke_iiii(3359,($L|0),-3,(0|0))|0); //@line 781 "modules/graphics/opengl/wrap_Graphics.cpp"
           $335 = __THREW__; __THREW__ = 0;
           $336 = $335&1;
           if ($336) {
            label = 44;
            break;
           }
           __THREW__ = 0;
           (invoke_iii(3360,($format$i|0),($78|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $79 = __THREW__; __THREW__ = 0;
           $80 = $79&1;
           if ($80) {
            label = 44;
            break;
           }
           __THREW__ = 0;
           $81 = (invoke_iiii(3359,($L|0),-2,(0|0))|0); //@line 783 "modules/graphics/opengl/wrap_Graphics.cpp"
           $82 = __THREW__; __THREW__ = 0;
           $83 = $82&1;
           if ($83) {
            label = 44;
            break;
           }
           __THREW__ = 0;
           $84 = (invoke_iii(3556,($81|0),($65|0))|0); //@line 784 "modules/graphics/opengl/wrap_Graphics.cpp"
           $85 = __THREW__; __THREW__ = 0;
           $86 = $85&1;
           if ($86) {
            label = 44;
            break;
           }
           if (!($84)) {
            $$lcssa53 = $81;
            label = 43;
            break;
           }
           __THREW__ = 0;
           $93 = (+invoke_dii(3361,($L|0),-1)); //@line 790 "modules/graphics/opengl/wrap_Graphics.cpp"
           $94 = __THREW__; __THREW__ = 0;
           $95 = $94&1;
           if ($95) {
            label = 44;
            break;
           }
           $96 = (~~(($93))); //@line 790 "modules/graphics/opengl/wrap_Graphics.cpp"
           HEAP32[$66>>2] = $96; //@line 790 "modules/graphics/opengl/wrap_Graphics.cpp"
           $$off$i = (($96) + -1)|0; //@line 791 "modules/graphics/opengl/wrap_Graphics.cpp"
           $97 = ($$off$i>>>0)>(3); //@line 791 "modules/graphics/opengl/wrap_Graphics.cpp"
           if ($97) {
            $$lcssa55 = $96;
            label = 49;
            break;
           }
           __THREW__ = 0;
           invoke_vii(3384,($L|0),-5); //@line 797 "modules/graphics/opengl/wrap_Graphics.cpp"
           $100 = __THREW__; __THREW__ = 0;
           $101 = $100&1;
           if ($101) {
            label = 44;
            break;
           }
           $102 = HEAP32[$13>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $103 = HEAP32[$67>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $104 = ($102|0)==($103|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           if ($104) {
            __THREW__ = 0;
            invoke_vii(3557,($vertexformat$i|0),($format$i|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $120 = __THREW__; __THREW__ = 0;
            $121 = $120&1;
            if ($121) {
             label = 44;
             break;
            }
           } else {
            __THREW__ = 0;
            invoke_vii(3351,($102|0),($format$i|0)); //@line 81 "modules/graphics/opengl/Mesh.h"
            $105 = __THREW__; __THREW__ = 0;
            $106 = $105&1;
            if ($106) {
             label = 44;
             break;
            }
            $107 = ((($102)) + 12|0); //@line 81 "modules/graphics/opengl/Mesh.h"
            $108 = $65; //@line 81 "modules/graphics/opengl/Mesh.h"
            $109 = $108; //@line 81 "modules/graphics/opengl/Mesh.h"
            $110 = HEAP32[$109>>2]|0; //@line 81 "modules/graphics/opengl/Mesh.h"
            $111 = (($108) + 4)|0; //@line 81 "modules/graphics/opengl/Mesh.h"
            $112 = $111; //@line 81 "modules/graphics/opengl/Mesh.h"
            $113 = HEAP32[$112>>2]|0; //@line 81 "modules/graphics/opengl/Mesh.h"
            $114 = $107; //@line 81 "modules/graphics/opengl/Mesh.h"
            $115 = $114; //@line 81 "modules/graphics/opengl/Mesh.h"
            HEAP32[$115>>2] = $110; //@line 81 "modules/graphics/opengl/Mesh.h"
            $116 = (($114) + 4)|0; //@line 81 "modules/graphics/opengl/Mesh.h"
            $117 = $116; //@line 81 "modules/graphics/opengl/Mesh.h"
            HEAP32[$117>>2] = $113; //@line 81 "modules/graphics/opengl/Mesh.h"
            $118 = HEAP32[$13>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $119 = ((($118)) + 20|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            HEAP32[$13>>2] = $119; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           }
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($format$i); //@line 81 "modules/graphics/opengl/Mesh.h"
           $122 = (($i$0$i) + 1)|0; //@line 772 "modules/graphics/opengl/wrap_Graphics.cpp"
           $i$0$i = $122;
          }
          if ((label|0) == 28) {
           $57 = ___cxa_find_matching_catch_2()|0;
           $58 = tempRet0;
           $eh$lpad$body$i$index45Z2D = $58;$eh$lpad$body$i$indexZ2D = $57;
           label = 31;
           break L10;
          }
          else if ((label|0) == 29) {
           $59 = ___cxa_find_matching_catch_2()|0;
           $60 = tempRet0;
           $eh$lpad$body$i$index45Z2D = $60;$eh$lpad$body$i$indexZ2D = $59;
           label = 31;
           break L10;
          }
          else if ((label|0) == 43) {
           __THREW__ = 0;
           HEAP32[$vararg_buffer6>>2] = $$lcssa53; //@line 786 "modules/graphics/opengl/wrap_Graphics.cpp"
           (invoke_iiii(3389,($L|0),(330010|0),($vararg_buffer6|0))|0); //@line 786 "modules/graphics/opengl/wrap_Graphics.cpp"
           $87 = __THREW__; __THREW__ = 0;
           $88 = $87&1;
           if ($88) {
            label = 45;
           } else {
            __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($format$i); //@line 81 "modules/graphics/opengl/Mesh.h"
            $$3$i = 0;
            break;
           }
          }
          else if ((label|0) == 44) {
           $89 = ___cxa_find_matching_catch_2()|0;
           $90 = tempRet0;
           $lpad$phi80$i$index = $89;$lpad$phi80$i$index58 = $90;
          }
          else if ((label|0) == 49) {
           __THREW__ = 0;
           HEAP32[$vararg_buffer9>>2] = $$lcssa55; //@line 793 "modules/graphics/opengl/wrap_Graphics.cpp"
           (invoke_iiii(3389,($L|0),(330049|0),($vararg_buffer9|0))|0); //@line 793 "modules/graphics/opengl/wrap_Graphics.cpp"
           $98 = __THREW__; __THREW__ = 0;
           $99 = $98&1;
           if ($99) {
            label = 45;
           } else {
            __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($format$i); //@line 81 "modules/graphics/opengl/Mesh.h"
            $$3$i = 0;
            break;
           }
          }
          else if ((label|0) == 56) {
           __THREW__ = 0;
           $123 = (invoke_iii(3558,($L|0),2)|0); //@line 801 "modules/graphics/opengl/wrap_Graphics.cpp"
           $124 = __THREW__; __THREW__ = 0;
           $125 = $124&1;
           if ($125) {
            label = 30;
            break L10;
           }
           $126 = ($123|0)==(0); //@line 801 "modules/graphics/opengl/wrap_Graphics.cpp"
           if (!($126)) {
            __THREW__ = 0;
            $127 = (+invoke_dii(3361,($L|0),2)); //@line 803 "modules/graphics/opengl/wrap_Graphics.cpp"
            $128 = __THREW__; __THREW__ = 0;
            $129 = $128&1;
            if ($129) {
             label = 30;
             break L10;
            }
            $130 = (~~(($127))); //@line 803 "modules/graphics/opengl/wrap_Graphics.cpp"
            $131 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
            __THREW__ = 0;
            $132 = (invoke_iiiiii(3559,($131|0),($vertexformat$i|0),($130|0),($28|0),($42|0))|0); //@line 804 "modules/graphics/opengl/wrap_Graphics.cpp"
            $133 = __THREW__; __THREW__ = 0;
            $134 = $133&1;
            if (!($134)) {
             $$3$i = $132;
             break;
            }
            $135 = ___cxa_find_matching_catch_3(8992|0)|0;
            $136 = tempRet0;
            $137 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
            $138 = ($136|0)==($137|0); //@line 535 "./common/runtime.h"
            do {
             if ($138) {
              $139 = (___cxa_begin_catch(($135|0))|0); //@line 535 "./common/runtime.h"
              $140 = HEAP32[$139>>2]|0; //@line 539 "./common/runtime.h"
              $141 = ((($140)) + 8|0); //@line 539 "./common/runtime.h"
              $142 = HEAP32[$141>>2]|0; //@line 539 "./common/runtime.h"
              $143 = (FUNCTION_TABLE_ii[$142 & 8191]($139)|0); //@line 539 "./common/runtime.h"
              __THREW__ = 0;
              invoke_vii(3380,($L|0),($143|0)); //@line 539 "./common/runtime.h"
              $144 = __THREW__; __THREW__ = 0;
              $145 = $144&1;
              if ($145) {
               $153 = ___cxa_find_matching_catch_2()|0;
               $154 = tempRet0;
               __THREW__ = 0;
               invoke_v(3378); //@line 540 "./common/runtime.h"
               $155 = __THREW__; __THREW__ = 0;
               $156 = $155&1;
               if (!($156)) {
                $$01$i$i = $154;$$02$i$i = $153;
                break;
               }
               $157 = ___cxa_find_matching_catch_3(0|0)|0;
               $158 = tempRet0;
               ___clang_call_terminate($157); //@line 540 "./common/runtime.h"
               // unreachable; //@line 540 "./common/runtime.h"
              }
              __THREW__ = 0;
              invoke_v(3378); //@line 540 "./common/runtime.h"
              $146 = __THREW__; __THREW__ = 0;
              $147 = $146&1;
              if ($147) {
               label = 30;
               break L10;
              }
              __THREW__ = 0;
              $148 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
              $149 = __THREW__; __THREW__ = 0;
              $150 = $149&1;
              if ($150) {
               label = 30;
               break L10;
              }
              __THREW__ = 0;
              HEAP32[$vararg_buffer12>>2] = $148; //@line 543 "./common/runtime.h"
              (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer12|0))|0); //@line 543 "./common/runtime.h"
              $151 = __THREW__; __THREW__ = 0;
              $152 = $151&1;
              if ($152) {
               label = 30;
               break L10;
              } else {
               $$3$i = 0;
               break L29;
              }
             } else {
              $$01$i$i = $136;$$02$i$i = $135;
             }
            } while(0);
            $eh$lpad$body$i$index45Z2D = $$01$i$i;$eh$lpad$body$i$indexZ2D = $$02$i$i;
            label = 31;
            break L10;
           }
           __THREW__ = 0;
           $159 = (invoke_iiii(3458,($L|0),2,2)|0); //@line 806 "modules/graphics/opengl/wrap_Graphics.cpp"
           $160 = __THREW__; __THREW__ = 0;
           $161 = $160&1;
           if ($161) {
            label = 30;
            break L10;
           }
           if ($159) {
            __THREW__ = 0;
            $162 = (invoke_iiii(3463,($L|0),2,2)|0); //@line 809 "modules/graphics/opengl/wrap_Graphics.cpp"
            $163 = __THREW__; __THREW__ = 0;
            $164 = $163&1;
            if ($164) {
             label = 30;
             break L10;
            }
            $165 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
            $166 = HEAP32[$162>>2]|0; //@line 810 "modules/graphics/opengl/wrap_Graphics.cpp"
            $167 = ((($166)) + 16|0); //@line 810 "modules/graphics/opengl/wrap_Graphics.cpp"
            $168 = HEAP32[$167>>2]|0; //@line 810 "modules/graphics/opengl/wrap_Graphics.cpp"
            __THREW__ = 0;
            $169 = (invoke_ii($168|0,($162|0))|0); //@line 810 "modules/graphics/opengl/wrap_Graphics.cpp"
            $170 = __THREW__; __THREW__ = 0;
            $171 = $170&1;
            do {
             if (!($171)) {
              $172 = HEAP32[$162>>2]|0; //@line 810 "modules/graphics/opengl/wrap_Graphics.cpp"
              $173 = ((($172)) + 20|0); //@line 810 "modules/graphics/opengl/wrap_Graphics.cpp"
              $174 = HEAP32[$173>>2]|0; //@line 810 "modules/graphics/opengl/wrap_Graphics.cpp"
              __THREW__ = 0;
              $175 = (invoke_ii($174|0,($162|0))|0); //@line 810 "modules/graphics/opengl/wrap_Graphics.cpp"
              $176 = __THREW__; __THREW__ = 0;
              $177 = $176&1;
              if ($177) {
               break;
              }
              __THREW__ = 0;
              $178 = (invoke_iiiiiii(3560,($165|0),($vertexformat$i|0),($169|0),($175|0),($28|0),($42|0))|0); //@line 810 "modules/graphics/opengl/wrap_Graphics.cpp"
              $179 = __THREW__; __THREW__ = 0;
              $180 = $179&1;
              if (!($180)) {
               $$3$i = $178;
               break L29;
              }
             }
            } while(0);
            $181 = ___cxa_find_matching_catch_3(8992|0)|0;
            $182 = tempRet0;
            $183 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
            $184 = ($182|0)==($183|0); //@line 535 "./common/runtime.h"
            do {
             if ($184) {
              $185 = (___cxa_begin_catch(($181|0))|0); //@line 535 "./common/runtime.h"
              $186 = HEAP32[$185>>2]|0; //@line 539 "./common/runtime.h"
              $187 = ((($186)) + 8|0); //@line 539 "./common/runtime.h"
              $188 = HEAP32[$187>>2]|0; //@line 539 "./common/runtime.h"
              $189 = (FUNCTION_TABLE_ii[$188 & 8191]($185)|0); //@line 539 "./common/runtime.h"
              __THREW__ = 0;
              invoke_vii(3380,($L|0),($189|0)); //@line 539 "./common/runtime.h"
              $190 = __THREW__; __THREW__ = 0;
              $191 = $190&1;
              if ($191) {
               $199 = ___cxa_find_matching_catch_2()|0;
               $200 = tempRet0;
               __THREW__ = 0;
               invoke_v(3378); //@line 540 "./common/runtime.h"
               $201 = __THREW__; __THREW__ = 0;
               $202 = $201&1;
               if (!($202)) {
                $$01$i21$i = $200;$$02$i20$i = $199;
                break;
               }
               $203 = ___cxa_find_matching_catch_3(0|0)|0;
               $204 = tempRet0;
               ___clang_call_terminate($203); //@line 540 "./common/runtime.h"
               // unreachable; //@line 540 "./common/runtime.h"
              }
              __THREW__ = 0;
              invoke_v(3378); //@line 540 "./common/runtime.h"
              $192 = __THREW__; __THREW__ = 0;
              $193 = $192&1;
              if ($193) {
               label = 30;
               break L10;
              }
              __THREW__ = 0;
              $194 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
              $195 = __THREW__; __THREW__ = 0;
              $196 = $195&1;
              if ($196) {
               label = 30;
               break L10;
              }
              __THREW__ = 0;
              HEAP32[$vararg_buffer15>>2] = $194; //@line 543 "./common/runtime.h"
              (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer15|0))|0); //@line 543 "./common/runtime.h"
              $197 = __THREW__; __THREW__ = 0;
              $198 = $197&1;
              if ($198) {
               label = 30;
               break L10;
              } else {
               $$3$i = 0;
               break L29;
              }
             } else {
              $$01$i21$i = $182;$$02$i20$i = $181;
             }
            } while(0);
            $eh$lpad$body$i$index45Z2D = $$01$i21$i;$eh$lpad$body$i$indexZ2D = $$02$i20$i;
            label = 31;
            break L10;
           }
           __THREW__ = 0;
           invoke_viii(3382,($L|0),2,1); //@line 815 "modules/graphics/opengl/wrap_Graphics.cpp"
           $205 = __THREW__; __THREW__ = 0;
           $206 = $205&1;
           if ($206) {
            label = 30;
            break L10;
           }
           __THREW__ = 0;
           $207 = (invoke_iii(3358,($L|0),-1)|0); //@line 816 "modules/graphics/opengl/wrap_Graphics.cpp"
           $208 = __THREW__; __THREW__ = 0;
           $209 = $208&1;
           if ($209) {
            label = 30;
            break L10;
           }
           $210 = ($207|0)==(5); //@line 816 "modules/graphics/opengl/wrap_Graphics.cpp"
           if (!($210)) {
            __THREW__ = 0;
            (invoke_iiii(3549,($L|0),2,(330120|0))|0); //@line 818 "modules/graphics/opengl/wrap_Graphics.cpp"
            $211 = __THREW__; __THREW__ = 0;
            $212 = $211&1;
            if ($212) {
             label = 30;
             break L10;
            } else {
             $$3$i = 0;
             break;
            }
           }
           __THREW__ = 0;
           invoke_vii(3384,($L|0),-2); //@line 821 "modules/graphics/opengl/wrap_Graphics.cpp"
           $213 = __THREW__; __THREW__ = 0;
           $214 = $213&1;
           if ($214) {
            label = 30;
            break L10;
           }
           __THREW__ = 0;
           $215 = (invoke_iii(3381,($L|0),2)|0); //@line 827 "modules/graphics/opengl/wrap_Graphics.cpp"
           $216 = __THREW__; __THREW__ = 0;
           $217 = $216&1;
           if ($217) {
            label = 30;
            break L10;
           }
           $218 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
           __THREW__ = 0;
           $219 = (invoke_iiiiii(3559,($218|0),($vertexformat$i|0),($215|0),($28|0),($42|0))|0); //@line 829 "modules/graphics/opengl/wrap_Graphics.cpp"
           $220 = __THREW__; __THREW__ = 0;
           $221 = $220&1;
           L102: do {
            if ($221) {
             $222 = ___cxa_find_matching_catch_3(8992|0)|0;
             $223 = tempRet0;
             $224 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
             $225 = ($223|0)==($224|0); //@line 535 "./common/runtime.h"
             do {
              if ($225) {
               $226 = (___cxa_begin_catch(($222|0))|0); //@line 535 "./common/runtime.h"
               $227 = HEAP32[$226>>2]|0; //@line 539 "./common/runtime.h"
               $228 = ((($227)) + 8|0); //@line 539 "./common/runtime.h"
               $229 = HEAP32[$228>>2]|0; //@line 539 "./common/runtime.h"
               $230 = (FUNCTION_TABLE_ii[$229 & 8191]($226)|0); //@line 539 "./common/runtime.h"
               __THREW__ = 0;
               invoke_vii(3380,($L|0),($230|0)); //@line 539 "./common/runtime.h"
               $231 = __THREW__; __THREW__ = 0;
               $232 = $231&1;
               if ($232) {
                $240 = ___cxa_find_matching_catch_2()|0;
                $241 = tempRet0;
                __THREW__ = 0;
                invoke_v(3378); //@line 540 "./common/runtime.h"
                $242 = __THREW__; __THREW__ = 0;
                $243 = $242&1;
                if (!($243)) {
                 $$01$i28$i = $241;$$02$i27$i = $240;
                 break;
                }
                $244 = ___cxa_find_matching_catch_3(0|0)|0;
                $245 = tempRet0;
                ___clang_call_terminate($244); //@line 540 "./common/runtime.h"
                // unreachable; //@line 540 "./common/runtime.h"
               }
               __THREW__ = 0;
               invoke_v(3378); //@line 540 "./common/runtime.h"
               $233 = __THREW__; __THREW__ = 0;
               $234 = $233&1;
               if ($234) {
                label = 30;
                break L10;
               }
               __THREW__ = 0;
               $235 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
               $236 = __THREW__; __THREW__ = 0;
               $237 = $236&1;
               if ($237) {
                label = 30;
                break L10;
               }
               __THREW__ = 0;
               HEAP32[$vararg_buffer18>>2] = $235; //@line 543 "./common/runtime.h"
               (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer18|0))|0); //@line 543 "./common/runtime.h"
               $238 = __THREW__; __THREW__ = 0;
               $239 = $238&1;
               if ($239) {
                label = 30;
                break L10;
               } else {
                $t$2$i = 0;
                break L102;
               }
              } else {
               $$01$i28$i = $223;$$02$i27$i = $222;
              }
             } while(0);
             $eh$lpad$body$i$index45Z2D = $$01$i28$i;$eh$lpad$body$i$indexZ2D = $$02$i27$i;
             label = 31;
             break L10;
            } else {
             $t$2$i = $219;
            }
           } while(0);
           $storemerge$i = 0;
           L114: while(1) {
            $246 = ($storemerge$i>>>0)<($215>>>0); //@line 834 "modules/graphics/opengl/wrap_Graphics.cpp"
            if (!($246)) {
             label = 99;
             break;
            }
            $249 = (($storemerge$i) + 1)|0; //@line 837 "modules/graphics/opengl/wrap_Graphics.cpp"
            __THREW__ = 0;
            invoke_viii(3382,($L|0),2,($249|0)); //@line 837 "modules/graphics/opengl/wrap_Graphics.cpp"
            $250 = __THREW__; __THREW__ = 0;
            $251 = $250&1;
            if ($251) {
             label = 27;
             break;
            }
            __THREW__ = 0;
            invoke_viii(3534,($L|0),-1,5); //@line 838 "modules/graphics/opengl/wrap_Graphics.cpp"
            $252 = __THREW__; __THREW__ = 0;
            $253 = $252&1;
            if ($253) {
             label = 27;
             break;
            }
            $254 = HEAP32[$13>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $255 = HEAP32[$vertexformat$i>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $256 = ($254|0)==($255|0); //@line 841 "modules/graphics/opengl/wrap_Graphics.cpp"
            if (!($256)) {
             $$in$i = $255;$n$096$i = 0;$storemerge695$i = 0;
             while(1) {
              $259 = $$in$i; //@line 841 "modules/graphics/opengl/wrap_Graphics.cpp"
              $260 = (((($259) + (($storemerge695$i*20)|0)|0)) + 16|0); //@line 843 "modules/graphics/opengl/wrap_Graphics.cpp"
              $261 = HEAP32[$260>>2]|0; //@line 843 "modules/graphics/opengl/wrap_Graphics.cpp"
              $262 = ($261|0)>(0); //@line 846 "modules/graphics/opengl/wrap_Graphics.cpp"
              if ($262) {
               $c$094$i = 0;$n$193$i = $n$096$i;
               while(1) {
                $269 = (($n$193$i) + 1)|0; //@line 848 "modules/graphics/opengl/wrap_Graphics.cpp"
                $270 = $c$094$i ^ -1; //@line 849 "modules/graphics/opengl/wrap_Graphics.cpp"
                __THREW__ = 0;
                invoke_viii(3382,($L|0),($270|0),($269|0)); //@line 849 "modules/graphics/opengl/wrap_Graphics.cpp"
                $271 = __THREW__; __THREW__ = 0;
                $272 = $271&1;
                if ($272) {
                 label = 25;
                 break L114;
                }
                $273 = (($c$094$i) + 1)|0; //@line 849 "modules/graphics/opengl/wrap_Graphics.cpp"
                $274 = ($273|0)<($261|0); //@line 846 "modules/graphics/opengl/wrap_Graphics.cpp"
                if ($274) {
                 $c$094$i = $273;$n$193$i = $269;
                } else {
                 $$lcssa49 = $269;
                 break;
                }
               }
               $$pre$i = HEAP32[$vertexformat$i>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
               $265 = $$pre$i;$n$1$lcssa$i = $$lcssa49;
              } else {
               $265 = $259;$n$1$lcssa$i = $n$096$i;
              }
              $263 = (0 - ($261))|0; //@line 853 "modules/graphics/opengl/wrap_Graphics.cpp"
              $264 = (((($265) + (($storemerge695$i*20)|0)|0)) + 12|0); //@line 853 "modules/graphics/opengl/wrap_Graphics.cpp"
              $266 = HEAP32[$264>>2]|0; //@line 853 "modules/graphics/opengl/wrap_Graphics.cpp"
              __THREW__ = 0;
              (invoke_iiiiii(3562,($L|0),($263|0),($266|0),($261|0),($data3$i|0))|0); //@line 853 "modules/graphics/opengl/wrap_Graphics.cpp"
              $267 = __THREW__; __THREW__ = 0;
              $268 = $267&1;
              if ($268) {
               label = 26;
               break L114;
              }
              $275 = $261 ^ -1; //@line 855 "modules/graphics/opengl/wrap_Graphics.cpp"
              __THREW__ = 0;
              invoke_vii(3384,($L|0),($275|0)); //@line 855 "modules/graphics/opengl/wrap_Graphics.cpp"
              $276 = __THREW__; __THREW__ = 0;
              $277 = $276&1;
              if ($277) {
               label = 26;
               break L114;
              }
              __THREW__ = 0;
              invoke_viiiii(3563,($t$2$i|0),($storemerge$i|0),($storemerge695$i|0),($data3$i|0),16); //@line 858 "modules/graphics/opengl/wrap_Graphics.cpp"
              $278 = __THREW__; __THREW__ = 0;
              $279 = $278&1;
              if ($279) {
               $280 = ___cxa_find_matching_catch_3(8992|0)|0;
               $281 = tempRet0;
               $282 = (_llvm_eh_typeid_for((8992|0))|0); //@line 556 "./common/runtime.h"
               $283 = ($281|0)==($282|0); //@line 556 "./common/runtime.h"
               if (!($283)) {
                $$01$i35$i = $281;$$02$i34$i = $280;
                break L114;
               }
               $284 = (___cxa_begin_catch(($280|0))|0); //@line 556 "./common/runtime.h"
               $285 = HEAP32[$284>>2]|0; //@line 560 "./common/runtime.h"
               $286 = ((($285)) + 8|0); //@line 560 "./common/runtime.h"
               $287 = HEAP32[$286>>2]|0; //@line 560 "./common/runtime.h"
               $288 = (FUNCTION_TABLE_ii[$287 & 8191]($284)|0); //@line 560 "./common/runtime.h"
               __THREW__ = 0;
               invoke_vii(3380,($L|0),($288|0)); //@line 560 "./common/runtime.h"
               $289 = __THREW__; __THREW__ = 0;
               $290 = $289&1;
               if ($290) {
                label = 117;
                break L114;
               }
               __THREW__ = 0;
               invoke_v(3378); //@line 561 "./common/runtime.h"
               $291 = __THREW__; __THREW__ = 0;
               $292 = $291&1;
               if ($292) {
                label = 26;
                break L114;
               }
               $293 = HEAP32[$t$2$i>>2]|0; //@line 859 "modules/graphics/opengl/wrap_Graphics.cpp"
               $294 = ((($293)) + 12|0); //@line 859 "modules/graphics/opengl/wrap_Graphics.cpp"
               $295 = HEAP32[$294>>2]|0; //@line 859 "modules/graphics/opengl/wrap_Graphics.cpp"
               __THREW__ = 0;
               invoke_vi($295|0,($t$2$i|0)); //@line 859 "modules/graphics/opengl/wrap_Graphics.cpp"
               $296 = __THREW__; __THREW__ = 0;
               $297 = $296&1;
               if ($297) {
                label = 26;
                break L114;
               }
               __THREW__ = 0;
               $298 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 566 "./common/runtime.h"
               $299 = __THREW__; __THREW__ = 0;
               $300 = $299&1;
               if ($300) {
                label = 26;
                break L114;
               }
               __THREW__ = 0;
               HEAP32[$vararg_buffer21>>2] = $298; //@line 566 "./common/runtime.h"
               (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer21|0))|0); //@line 566 "./common/runtime.h"
               $301 = __THREW__; __THREW__ = 0;
               $302 = $301&1;
               if ($302) {
                label = 26;
                break L114;
               }
              }
              $309 = (($storemerge695$i) + 1)|0; //@line 841 "modules/graphics/opengl/wrap_Graphics.cpp"
              $310 = HEAP32[$13>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $311 = HEAP32[$vertexformat$i>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $312 = (($310) - ($311))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $313 = (($312|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $314 = ($309>>>0)<($313>>>0); //@line 841 "modules/graphics/opengl/wrap_Graphics.cpp"
              if ($314) {
               $$in$i = $311;$n$096$i = $n$1$lcssa$i;$storemerge695$i = $309;
              } else {
               break;
              }
             }
            }
            __THREW__ = 0;
            invoke_vii(3384,($L|0),-2); //@line 863 "modules/graphics/opengl/wrap_Graphics.cpp"
            $257 = __THREW__; __THREW__ = 0;
            $258 = $257&1;
            if ($258) {
             label = 27;
             break;
            } else {
             $storemerge$i = $249;
            }
           }
           do {
            if ((label|0) == 25) {
             $51 = ___cxa_find_matching_catch_2()|0;
             $52 = tempRet0;
             $eh$lpad$body$i$index45Z2D = $52;$eh$lpad$body$i$indexZ2D = $51;
             label = 31;
             break L10;
            }
            else if ((label|0) == 26) {
             $53 = ___cxa_find_matching_catch_2()|0;
             $54 = tempRet0;
             $eh$lpad$body$i$index45Z2D = $54;$eh$lpad$body$i$indexZ2D = $53;
             label = 31;
             break L10;
            }
            else if ((label|0) == 27) {
             $55 = ___cxa_find_matching_catch_2()|0;
             $56 = tempRet0;
             $eh$lpad$body$i$index45Z2D = $56;$eh$lpad$body$i$indexZ2D = $55;
             label = 31;
             break L10;
            }
            else if ((label|0) == 99) {
             __THREW__ = 0;
             invoke_vi(3561,($t$2$i|0)); //@line 866 "modules/graphics/opengl/wrap_Graphics.cpp"
             $247 = __THREW__; __THREW__ = 0;
             $248 = $247&1;
             if ($248) {
              label = 30;
              break L10;
             }
             $$3$i = $t$2$i;
             break L29;
            }
            else if ((label|0) == 117) {
             $303 = ___cxa_find_matching_catch_2()|0;
             $304 = tempRet0;
             __THREW__ = 0;
             invoke_v(3378); //@line 561 "./common/runtime.h"
             $305 = __THREW__; __THREW__ = 0;
             $306 = $305&1;
             if (!($306)) {
              $$01$i35$i = $304;$$02$i34$i = $303;
              break;
             }
             $307 = ___cxa_find_matching_catch_3(0|0)|0;
             $308 = tempRet0;
             ___clang_call_terminate($307); //@line 561 "./common/runtime.h"
             // unreachable; //@line 561 "./common/runtime.h"
            }
           } while(0);
           $eh$lpad$body$i$index45Z2D = $$01$i35$i;$eh$lpad$body$i$indexZ2D = $$02$i34$i;
           label = 31;
           break L10;
          }
          if ((label|0) == 45) {
           $91 = ___cxa_find_matching_catch_2()|0;
           $92 = tempRet0;
           $lpad$phi80$i$index = $91;$lpad$phi80$i$index58 = $92;
          }
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($format$i); //@line 81 "modules/graphics/opengl/Mesh.h"
          $$02$i = $lpad$phi80$i$index;$$03$i = $lpad$phi80$i$index58;
          break L10;
         } else {
          __THREW__ = 0;
          (invoke_iiii(3549,($L|0),1,(329985|0))|0); //@line 766 "modules/graphics/opengl/wrap_Graphics.cpp"
          $49 = __THREW__; __THREW__ = 0;
          $50 = $49&1;
          if ($50) {
           label = 30;
           break L10;
          } else {
           $$3$i = 0;
          }
         }
        } while(0);
        $315 = HEAP32[$vertexformat$i>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $316 = ($315|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($316)) {
         $317 = HEAP32[$13>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $318 = ($317|0)==($315|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($318) {
          $323 = $315;
         } else {
          $320 = $317;
          while(1) {
           $319 = ((($320)) + -20|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           HEAP32[$13>>2] = $319; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($319); //@line 81 "modules/graphics/opengl/Mesh.h"
           $321 = HEAP32[$13>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $322 = ($321|0)==($315|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           if ($322) {
            break;
           } else {
            $320 = $321;
           }
          }
          $$pre$i44$i = HEAP32[$vertexformat$i>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $323 = $$pre$i44$i;
         }
         __ZdlPv($323); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        }
        $t$0 = $$3$i;
        __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,19,$t$0); //@line 889 "modules/graphics/opengl/wrap_Graphics.cpp"
        $512 = HEAP32[$t$0>>2]|0; //@line 890 "modules/graphics/opengl/wrap_Graphics.cpp"
        $513 = ((($512)) + 12|0); //@line 890 "modules/graphics/opengl/wrap_Graphics.cpp"
        $514 = HEAP32[$513>>2]|0; //@line 890 "modules/graphics/opengl/wrap_Graphics.cpp"
        FUNCTION_TABLE_vi[$514 & 8191]($t$0); //@line 890 "modules/graphics/opengl/wrap_Graphics.cpp"
        STACKTOP = sp;return 1; //@line 891 "modules/graphics/opengl/wrap_Graphics.cpp"
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 30) {
    $61 = ___cxa_find_matching_catch_2()|0;
    $62 = tempRet0;
    $eh$lpad$body$i$index45Z2D = $62;$eh$lpad$body$i$indexZ2D = $61;
    label = 31;
   }
   if ((label|0) == 31) {
    $$02$i = $eh$lpad$body$i$indexZ2D;$$03$i = $eh$lpad$body$i$index45Z2D;
   }
   $324 = HEAP32[$vertexformat$i>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $325 = ($324|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($325) {
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $326 = HEAP32[$13>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $327 = ($326|0)==($324|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($327) {
    $332 = $324;
   } else {
    $329 = $326;
    while(1) {
     $328 = ((($329)) + -20|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$13>>2] = $328; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($328); //@line 81 "modules/graphics/opengl/Mesh.h"
     $330 = HEAP32[$13>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $331 = ($330|0)==($324|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($331) {
      break;
     } else {
      $329 = $330;
     }
    }
    $$pre$i$i = HEAP32[$vertexformat$i>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $332 = $$pre$i$i;
   }
   __ZdlPv($332); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($$02$i|0);
   // unreachable;
  }
 }
 HEAP32[$1>>2] = 0; //@line 700 "modules/graphics/opengl/wrap_Graphics.cpp"
 $337 = (_lua_type($L,2)|0); //@line 688 "modules/graphics/opengl/wrap_Graphics.cpp"
 $338 = ($337|0)<(1); //@line 688 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($338)) {
  $339 = (_luaL_checklstring($L,2,0)|0); //@line 688 "modules/graphics/opengl/wrap_Graphics.cpp"
  $340 = ($339|0)==(0|0); //@line 690 "modules/graphics/opengl/wrap_Graphics.cpp"
  if (!($340)) {
   $341 = (__ZN4love8graphics6opengl4Mesh11getConstantEPKcRNS2_8DrawModeE($339,$1)|0); //@line 690 "modules/graphics/opengl/wrap_Graphics.cpp"
   if (!($341)) {
    HEAP32[$vararg_buffer24>>2] = $339; //@line 691 "modules/graphics/opengl/wrap_Graphics.cpp"
    (_luaL_error($L,343313,$vararg_buffer24)|0); //@line 691 "modules/graphics/opengl/wrap_Graphics.cpp"
   }
  }
 }
 $342 = HEAP32[$1>>2]|0; //@line 693 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAP32[$0>>2] = 1; //@line 701 "modules/graphics/opengl/wrap_Graphics.cpp"
 $343 = (_lua_type($L,3)|0); //@line 678 "modules/graphics/opengl/wrap_Graphics.cpp"
 $344 = ($343|0)<(1); //@line 678 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($344)) {
  $345 = (_luaL_checklstring($L,3,0)|0); //@line 678 "modules/graphics/opengl/wrap_Graphics.cpp"
  $346 = ($345|0)==(0|0); //@line 680 "modules/graphics/opengl/wrap_Graphics.cpp"
  if (!($346)) {
   $347 = (__ZN4love8graphics6opengl4Mesh11getConstantEPKcRNS2_5UsageE($345,$0)|0); //@line 680 "modules/graphics/opengl/wrap_Graphics.cpp"
   if (!($347)) {
    HEAP32[$vararg_buffer27>>2] = $345; //@line 681 "modules/graphics/opengl/wrap_Graphics.cpp"
    (_luaL_error($L,329957,$vararg_buffer27)|0); //@line 681 "modules/graphics/opengl/wrap_Graphics.cpp"
   }
  }
 }
 $348 = HEAP32[$0>>2]|0; //@line 683 "modules/graphics/opengl/wrap_Graphics.cpp"
 $349 = (_lua_type($L,1)|0); //@line 705 "modules/graphics/opengl/wrap_Graphics.cpp"
 $350 = ($349|0)==(5); //@line 705 "modules/graphics/opengl/wrap_Graphics.cpp"
 do {
  if ($350) {
   $351 = (__ZN4love11luax_objlenEP9lua_Statei($L,1)|0); //@line 707 "modules/graphics/opengl/wrap_Graphics.cpp"
   HEAP32[$vertices$i>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $352 = ((($vertices$i)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$352>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $353 = ((($vertices$i)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$353>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $354 = ((($vertices$i)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $355 = ($351|0)==(0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   L189: do {
    if ($355) {
     label = 176;
    } else {
     $356 = ($351*20)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     __THREW__ = 0;
     $357 = (invoke_ii(3348,($356|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $358 = __THREW__; __THREW__ = 0;
     $359 = $358&1;
     if ($359) {
      label = 152;
     } else {
      $360 = $357; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $361 = (($357) + (($351*20)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $362 = $361; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      HEAP32[$vertices$i>>2] = $360; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$352>>2] = $360; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$354>>2] = $362; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $363 = ((($v$i)) + 4|0); //@line 728 "modules/graphics/opengl/wrap_Graphics.cpp"
      $364 = ((($v$i)) + 8|0); //@line 729 "modules/graphics/opengl/wrap_Graphics.cpp"
      $365 = ((($v$i)) + 12|0); //@line 730 "modules/graphics/opengl/wrap_Graphics.cpp"
      $366 = ((($v$i)) + 16|0); //@line 732 "modules/graphics/opengl/wrap_Graphics.cpp"
      $367 = ((($v$i)) + 17|0); //@line 733 "modules/graphics/opengl/wrap_Graphics.cpp"
      $368 = ((($v$i)) + 18|0); //@line 734 "modules/graphics/opengl/wrap_Graphics.cpp"
      $369 = ((($v$i)) + 19|0); //@line 735 "modules/graphics/opengl/wrap_Graphics.cpp"
      $i$024$i = 1;
      while(1) {
       __THREW__ = 0;
       invoke_viii(3382,($L|0),1,($i$024$i|0)); //@line 714 "modules/graphics/opengl/wrap_Graphics.cpp"
       $385 = __THREW__; __THREW__ = 0;
       $386 = $385&1;
       if ($386) {
        label = 151;
        break;
       }
       __THREW__ = 0;
       $387 = (invoke_iii(3358,($L|0),-1)|0); //@line 716 "modules/graphics/opengl/wrap_Graphics.cpp"
       $388 = __THREW__; __THREW__ = 0;
       $389 = $388&1;
       if ($389) {
        label = 151;
        break;
       }
       $390 = ($387|0)==(5); //@line 716 "modules/graphics/opengl/wrap_Graphics.cpp"
       if (!($390)) {
        label = 162;
        break;
       }
       __THREW__ = 0;
       invoke_viii(3382,($L|0),-1,1); //@line 723 "modules/graphics/opengl/wrap_Graphics.cpp"
       $391 = __THREW__; __THREW__ = 0;
       $392 = $391&1;
       if ($392) {
        label = 150;
        break;
       }
       __THREW__ = 0;
       invoke_viii(3382,($L|0),-2,2); //@line 723 "modules/graphics/opengl/wrap_Graphics.cpp"
       $395 = __THREW__; __THREW__ = 0;
       $396 = $395&1;
       if ($396) {
        label = 150;
        break;
       }
       __THREW__ = 0;
       invoke_viii(3382,($L|0),-3,3); //@line 723 "modules/graphics/opengl/wrap_Graphics.cpp"
       $498 = __THREW__; __THREW__ = 0;
       $499 = $498&1;
       if ($499) {
        label = 150;
        break;
       }
       __THREW__ = 0;
       invoke_viii(3382,($L|0),-4,4); //@line 723 "modules/graphics/opengl/wrap_Graphics.cpp"
       $500 = __THREW__; __THREW__ = 0;
       $501 = $500&1;
       if ($501) {
        label = 150;
        break;
       }
       __THREW__ = 0;
       invoke_viii(3382,($L|0),-5,5); //@line 723 "modules/graphics/opengl/wrap_Graphics.cpp"
       $502 = __THREW__; __THREW__ = 0;
       $503 = $502&1;
       if ($503) {
        label = 150;
        break;
       }
       __THREW__ = 0;
       invoke_viii(3382,($L|0),-6,6); //@line 723 "modules/graphics/opengl/wrap_Graphics.cpp"
       $504 = __THREW__; __THREW__ = 0;
       $505 = $504&1;
       if ($505) {
        label = 150;
        break;
       }
       __THREW__ = 0;
       invoke_viii(3382,($L|0),-7,7); //@line 723 "modules/graphics/opengl/wrap_Graphics.cpp"
       $506 = __THREW__; __THREW__ = 0;
       $507 = $506&1;
       if ($507) {
        label = 150;
        break;
       }
       __THREW__ = 0;
       invoke_viii(3382,($L|0),-8,8); //@line 723 "modules/graphics/opengl/wrap_Graphics.cpp"
       $508 = __THREW__; __THREW__ = 0;
       $509 = $508&1;
       if ($509) {
        label = 150;
        break;
       }
       __THREW__ = 0;
       $398 = (+invoke_dii(3361,($L|0),-8)); //@line 727 "modules/graphics/opengl/wrap_Graphics.cpp"
       $510 = __THREW__; __THREW__ = 0;
       $511 = $510&1;
       if ($511) {
        label = 151;
        break;
       }
       $397 = $398; //@line 727 "modules/graphics/opengl/wrap_Graphics.cpp"
       HEAPF32[$v$i>>2] = $397; //@line 727 "modules/graphics/opengl/wrap_Graphics.cpp"
       __THREW__ = 0;
       $399 = (+invoke_dii(3361,($L|0),-7)); //@line 728 "modules/graphics/opengl/wrap_Graphics.cpp"
       $400 = __THREW__; __THREW__ = 0;
       $401 = $400&1;
       if ($401) {
        label = 151;
        break;
       }
       $402 = $399; //@line 728 "modules/graphics/opengl/wrap_Graphics.cpp"
       HEAPF32[$363>>2] = $402; //@line 728 "modules/graphics/opengl/wrap_Graphics.cpp"
       __THREW__ = 0;
       $403 = (+invoke_diid(3525,($L|0),-6,0.0)); //@line 729 "modules/graphics/opengl/wrap_Graphics.cpp"
       $404 = __THREW__; __THREW__ = 0;
       $405 = $404&1;
       if ($405) {
        label = 151;
        break;
       }
       $406 = $403; //@line 729 "modules/graphics/opengl/wrap_Graphics.cpp"
       HEAPF32[$364>>2] = $406; //@line 729 "modules/graphics/opengl/wrap_Graphics.cpp"
       __THREW__ = 0;
       $407 = (+invoke_diid(3525,($L|0),-5,0.0)); //@line 730 "modules/graphics/opengl/wrap_Graphics.cpp"
       $408 = __THREW__; __THREW__ = 0;
       $409 = $408&1;
       if ($409) {
        label = 151;
        break;
       }
       $410 = $407; //@line 730 "modules/graphics/opengl/wrap_Graphics.cpp"
       HEAPF32[$365>>2] = $410; //@line 730 "modules/graphics/opengl/wrap_Graphics.cpp"
       __THREW__ = 0;
       $411 = (+invoke_diid(3525,($L|0),-4,255.0)); //@line 732 "modules/graphics/opengl/wrap_Graphics.cpp"
       $412 = __THREW__; __THREW__ = 0;
       $413 = $412&1;
       if ($413) {
        label = 151;
        break;
       }
       $414 = (~~(($411))&255); //@line 732 "modules/graphics/opengl/wrap_Graphics.cpp"
       HEAP8[$366>>0] = $414; //@line 732 "modules/graphics/opengl/wrap_Graphics.cpp"
       __THREW__ = 0;
       $415 = (+invoke_diid(3525,($L|0),-3,255.0)); //@line 733 "modules/graphics/opengl/wrap_Graphics.cpp"
       $416 = __THREW__; __THREW__ = 0;
       $417 = $416&1;
       if ($417) {
        label = 151;
        break;
       }
       $418 = (~~(($415))&255); //@line 733 "modules/graphics/opengl/wrap_Graphics.cpp"
       HEAP8[$367>>0] = $418; //@line 733 "modules/graphics/opengl/wrap_Graphics.cpp"
       __THREW__ = 0;
       $419 = (+invoke_diid(3525,($L|0),-2,255.0)); //@line 734 "modules/graphics/opengl/wrap_Graphics.cpp"
       $420 = __THREW__; __THREW__ = 0;
       $421 = $420&1;
       if ($421) {
        label = 151;
        break;
       }
       $422 = (~~(($419))&255); //@line 734 "modules/graphics/opengl/wrap_Graphics.cpp"
       HEAP8[$368>>0] = $422; //@line 734 "modules/graphics/opengl/wrap_Graphics.cpp"
       __THREW__ = 0;
       $423 = (+invoke_diid(3525,($L|0),-1,255.0)); //@line 735 "modules/graphics/opengl/wrap_Graphics.cpp"
       $424 = __THREW__; __THREW__ = 0;
       $425 = $424&1;
       if ($425) {
        label = 151;
        break;
       }
       $426 = (~~(($423))&255); //@line 735 "modules/graphics/opengl/wrap_Graphics.cpp"
       HEAP8[$369>>0] = $426; //@line 735 "modules/graphics/opengl/wrap_Graphics.cpp"
       __THREW__ = 0;
       invoke_vii(3384,($L|0),-10); //@line 737 "modules/graphics/opengl/wrap_Graphics.cpp"
       $427 = __THREW__; __THREW__ = 0;
       $428 = $427&1;
       if ($428) {
        label = 151;
        break;
       }
       $429 = HEAP32[$352>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $430 = HEAP32[$354>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $431 = ($429|0)==($430|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($431) {
        __THREW__ = 0;
        invoke_vii(3388,($vertices$i|0),($v$i|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $433 = __THREW__; __THREW__ = 0;
        $434 = $433&1;
        if ($434) {
         label = 151;
         break;
        }
       } else {
        ;HEAP32[$429>>2]=HEAP32[$v$i>>2]|0;HEAP32[$429+4>>2]=HEAP32[$v$i+4>>2]|0;HEAP32[$429+8>>2]=HEAP32[$v$i+8>>2]|0;HEAP32[$429+12>>2]=HEAP32[$v$i+12>>2]|0;HEAP32[$429+16>>2]=HEAP32[$v$i+16>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $432 = ((($429)) + 20|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$352>>2] = $432; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       $435 = (($i$024$i) + 1)|0; //@line 712 "modules/graphics/opengl/wrap_Graphics.cpp"
       $436 = ($435>>>0)>($351>>>0); //@line 712 "modules/graphics/opengl/wrap_Graphics.cpp"
       if ($436) {
        label = 176;
        break L189;
       } else {
        $i$024$i = $435;
       }
      }
      if ((label|0) == 150) {
       $370 = ___cxa_find_matching_catch_2()|0;
       $371 = tempRet0;
       $eh$lpad$body$i10$index65Z2D = $371;$eh$lpad$body$i10$indexZ2D = $370;
       break;
      }
      else if ((label|0) == 151) {
       $372 = ___cxa_find_matching_catch_2()|0;
       $373 = tempRet0;
       $eh$lpad$body$i10$index65Z2D = $373;$eh$lpad$body$i10$indexZ2D = $372;
       break;
      }
      else if ((label|0) == 162) {
       __THREW__ = 0;
       (invoke_iiii(3564,($L|0),1,(330145|0))|0); //@line 718 "modules/graphics/opengl/wrap_Graphics.cpp"
       $393 = __THREW__; __THREW__ = 0;
       $394 = $393&1;
       if ($394) {
        label = 152;
        break;
       } else {
        $$1$i = 1;$t$1$i = 0;
        label = 185;
        break;
       }
      }
     }
    }
   } while(0);
   L222: do {
    if ((label|0) == 176) {
     $437 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
     __THREW__ = 0;
     $438 = (invoke_iiiii(3565,($437|0),($vertices$i|0),($342|0),($348|0))|0); //@line 741 "modules/graphics/opengl/wrap_Graphics.cpp"
     $439 = __THREW__; __THREW__ = 0;
     $440 = $439&1;
     if ($440) {
      $441 = ___cxa_find_matching_catch_3(8992|0)|0;
      $442 = tempRet0;
      $443 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
      $444 = ($442|0)==($443|0); //@line 535 "./common/runtime.h"
      do {
       if ($444) {
        $445 = (___cxa_begin_catch(($441|0))|0); //@line 535 "./common/runtime.h"
        $446 = HEAP32[$445>>2]|0; //@line 539 "./common/runtime.h"
        $447 = ((($446)) + 8|0); //@line 539 "./common/runtime.h"
        $448 = HEAP32[$447>>2]|0; //@line 539 "./common/runtime.h"
        $449 = (FUNCTION_TABLE_ii[$448 & 8191]($445)|0); //@line 539 "./common/runtime.h"
        __THREW__ = 0;
        invoke_vii(3380,($L|0),($449|0)); //@line 539 "./common/runtime.h"
        $450 = __THREW__; __THREW__ = 0;
        $451 = $450&1;
        if ($451) {
         $459 = ___cxa_find_matching_catch_2()|0;
         $460 = tempRet0;
         __THREW__ = 0;
         invoke_v(3378); //@line 540 "./common/runtime.h"
         $461 = __THREW__; __THREW__ = 0;
         $462 = $461&1;
         if (!($462)) {
          $$01$i$i13 = $460;$$02$i$i12 = $459;
          break;
         }
         $463 = ___cxa_find_matching_catch_3(0|0)|0;
         $464 = tempRet0;
         ___clang_call_terminate($463); //@line 540 "./common/runtime.h"
         // unreachable; //@line 540 "./common/runtime.h"
        }
        __THREW__ = 0;
        invoke_v(3378); //@line 540 "./common/runtime.h"
        $452 = __THREW__; __THREW__ = 0;
        $453 = $452&1;
        if ($453) {
         label = 152;
         break L222;
        }
        __THREW__ = 0;
        $454 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
        $455 = __THREW__; __THREW__ = 0;
        $456 = $455&1;
        if ($456) {
         label = 152;
         break L222;
        }
        __THREW__ = 0;
        HEAP32[$vararg_buffer30>>2] = $454; //@line 543 "./common/runtime.h"
        (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer30|0))|0); //@line 543 "./common/runtime.h"
        $457 = __THREW__; __THREW__ = 0;
        $458 = $457&1;
        if ($458) {
         label = 152;
         break L222;
        } else {
         $$1$i = 0;$t$1$i = 0;
         label = 185;
         break L222;
        }
       } else {
        $$01$i$i13 = $442;$$02$i$i12 = $441;
       }
      } while(0);
      $eh$lpad$body$i10$index65Z2D = $$01$i$i13;$eh$lpad$body$i10$indexZ2D = $$02$i$i12;
     } else {
      $$1$i = 0;$t$1$i = $438;
      label = 185;
     }
    }
   } while(0);
   if ((label|0) == 152) {
    $374 = ___cxa_find_matching_catch_2()|0;
    $375 = tempRet0;
    $eh$lpad$body$i10$index65Z2D = $375;$eh$lpad$body$i10$indexZ2D = $374;
   }
   else if ((label|0) == 185) {
    $465 = HEAP32[$vertices$i>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $466 = ($465|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $467 = $465; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($466)) {
     $468 = HEAP32[$352>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $469 = ($468|0)==($465|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($469)) {
      $scevgep$i$i$i8$i = ((($468)) + -20|0);
      $470 = $scevgep$i$i$i8$i;
      $471 = (($470) - ($467))|0;
      $472 = (($471>>>0) / 20)&-1;
      $473 = $472 ^ -1;
      $scevgep4$i$i$i9$i = (($468) + (($473*20)|0)|0);
      HEAP32[$352>>2] = $scevgep4$i$i$i9$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __ZdlPv($465); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    }
    $cond1$i = ($$1$i|0)==(0); //@line 887 "modules/graphics/opengl/wrap_Graphics.cpp"
    if ($cond1$i) {
     $t$3$i = $t$1$i;
     break;
    } else {
     $t$0 = 0;
    }
    __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,19,$t$0); //@line 889 "modules/graphics/opengl/wrap_Graphics.cpp"
    $512 = HEAP32[$t$0>>2]|0; //@line 890 "modules/graphics/opengl/wrap_Graphics.cpp"
    $513 = ((($512)) + 12|0); //@line 890 "modules/graphics/opengl/wrap_Graphics.cpp"
    $514 = HEAP32[$513>>2]|0; //@line 890 "modules/graphics/opengl/wrap_Graphics.cpp"
    FUNCTION_TABLE_vi[$514 & 8191]($t$0); //@line 890 "modules/graphics/opengl/wrap_Graphics.cpp"
    STACKTOP = sp;return 1; //@line 891 "modules/graphics/opengl/wrap_Graphics.cpp"
   }
   $376 = HEAP32[$vertices$i>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $377 = ($376|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $378 = $376; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($377) {
    ___resumeException($eh$lpad$body$i10$indexZ2D|0);
    // unreachable;
   }
   $379 = HEAP32[$352>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $380 = ($379|0)==($376|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($380)) {
    $scevgep$i$i$i$i = ((($379)) + -20|0);
    $381 = $scevgep$i$i$i$i;
    $382 = (($381) - ($378))|0;
    $383 = (($382>>>0) / 20)&-1;
    $384 = $383 ^ -1;
    $scevgep4$i$i$i$i = (($379) + (($384*20)|0)|0);
    HEAP32[$352>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($376); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($eh$lpad$body$i10$indexZ2D|0);
   // unreachable;
  } else {
   $474 = (+_luaL_checknumber($L,1)); //@line 745 "modules/graphics/opengl/wrap_Graphics.cpp"
   $475 = (~~(($474))); //@line 745 "modules/graphics/opengl/wrap_Graphics.cpp"
   $476 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
   __THREW__ = 0;
   $477 = (invoke_iiiii(3566,($476|0),($475|0),($342|0),($348|0))|0); //@line 746 "modules/graphics/opengl/wrap_Graphics.cpp"
   $478 = __THREW__; __THREW__ = 0;
   $479 = $478&1;
   if ($479) {
    $480 = ___cxa_find_matching_catch_3(8992|0)|0;
    $481 = tempRet0;
    $482 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
    $483 = ($481|0)==($482|0); //@line 535 "./common/runtime.h"
    if (!($483)) {
     $$01$i14$i = $481;$$02$i13$i = $480;
     ___resumeException($$02$i13$i|0);
     // unreachable;
    }
    $484 = (___cxa_begin_catch(($480|0))|0); //@line 535 "./common/runtime.h"
    $485 = HEAP32[$484>>2]|0; //@line 539 "./common/runtime.h"
    $486 = ((($485)) + 8|0); //@line 539 "./common/runtime.h"
    $487 = HEAP32[$486>>2]|0; //@line 539 "./common/runtime.h"
    $488 = (FUNCTION_TABLE_ii[$487 & 8191]($484)|0); //@line 539 "./common/runtime.h"
    __THREW__ = 0;
    invoke_vii(3380,($L|0),($488|0)); //@line 539 "./common/runtime.h"
    $489 = __THREW__; __THREW__ = 0;
    $490 = $489&1;
    if (!($490)) {
     ___cxa_end_catch(); //@line 540 "./common/runtime.h"
     $491 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
     HEAP32[$vararg_buffer33>>2] = $491; //@line 543 "./common/runtime.h"
     (_luaL_error($L,595867,$vararg_buffer33)|0); //@line 543 "./common/runtime.h"
     $t$3$i = 0;
     break;
    }
    $492 = ___cxa_find_matching_catch_2()|0;
    $493 = tempRet0;
    __THREW__ = 0;
    invoke_v(3378); //@line 540 "./common/runtime.h"
    $494 = __THREW__; __THREW__ = 0;
    $495 = $494&1;
    if ($495) {
     $496 = ___cxa_find_matching_catch_3(0|0)|0;
     $497 = tempRet0;
     ___clang_call_terminate($496); //@line 540 "./common/runtime.h"
     // unreachable; //@line 540 "./common/runtime.h"
    } else {
     $$01$i14$i = $493;$$02$i13$i = $492;
     ___resumeException($$02$i13$i|0);
     // unreachable;
    }
   } else {
    $t$3$i = $477;
   }
  }
 } while(0);
 $t$0 = $t$3$i;
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,19,$t$0); //@line 889 "modules/graphics/opengl/wrap_Graphics.cpp"
 $512 = HEAP32[$t$0>>2]|0; //@line 890 "modules/graphics/opengl/wrap_Graphics.cpp"
 $513 = ((($512)) + 12|0); //@line 890 "modules/graphics/opengl/wrap_Graphics.cpp"
 $514 = HEAP32[$513>>2]|0; //@line 890 "modules/graphics/opengl/wrap_Graphics.cpp"
 FUNCTION_TABLE_vi[$514 & 8191]($t$0); //@line 890 "modules/graphics/opengl/wrap_Graphics.cpp"
 STACKTOP = sp;return 1; //@line 891 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZNSt3__16vectorIN4love8graphics6opengl4Mesh12AttribFormatENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i4 = 0, $$0$i5 = 0, $$01$i$i = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $9 = 0, $__v$sroa$6$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($3|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(214748364); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $10 = $$pre;
 } else {
  $10 = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($10))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = (($9|0) / 20)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($11>>>0)<(107374182); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $16 = $11 << 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = ($16>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $18 = $17 ? $5 : $16; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = (($19) - ($10))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = (($20|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = ($18|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($22) {
   $$0$i5 = 0;$26 = 0;$27 = $21;
  } else {
   $$0$i4 = $18;$83 = $21;
   label = 6;
  }
 } else {
  $13 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) - ($10))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = (($14|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i4 = 214748364;$83 = $15;
  label = 6;
 }
 if ((label|0) == 6) {
  $23 = ($$0$i4*20)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = (__Znwj($23)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i5 = $$0$i4;$26 = $24;$27 = $83;
 }
 $25 = (($26) + (($27*20)|0)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $25; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = (($26) + (($$0$i5*20)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = $29; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 __THREW__ = 0;
 invoke_vii(3351,($25|0),($__x|0)); //@line 81 "modules/graphics/opengl/Mesh.h"
 $31 = __THREW__; __THREW__ = 0;
 $32 = $31&1;
 if ($32) {
  $80 = ___cxa_find_matching_catch_2()|0;
  $81 = tempRet0;
  $82 = ($26|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($82) {
   ___resumeException($80|0);
   // unreachable;
  }
  __ZdlPv($26); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($80|0);
  // unreachable;
 }
 $33 = (((($26) + (($27*20)|0)|0)) + 12|0); //@line 81 "modules/graphics/opengl/Mesh.h"
 $34 = ((($__x)) + 12|0); //@line 81 "modules/graphics/opengl/Mesh.h"
 $35 = $34; //@line 81 "modules/graphics/opengl/Mesh.h"
 $36 = $35; //@line 81 "modules/graphics/opengl/Mesh.h"
 $37 = HEAP32[$36>>2]|0; //@line 81 "modules/graphics/opengl/Mesh.h"
 $38 = (($35) + 4)|0; //@line 81 "modules/graphics/opengl/Mesh.h"
 $39 = $38; //@line 81 "modules/graphics/opengl/Mesh.h"
 $40 = HEAP32[$39>>2]|0; //@line 81 "modules/graphics/opengl/Mesh.h"
 $41 = $33; //@line 81 "modules/graphics/opengl/Mesh.h"
 $42 = $41; //@line 81 "modules/graphics/opengl/Mesh.h"
 HEAP32[$42>>2] = $37; //@line 81 "modules/graphics/opengl/Mesh.h"
 $43 = (($41) + 4)|0; //@line 81 "modules/graphics/opengl/Mesh.h"
 $44 = $43; //@line 81 "modules/graphics/opengl/Mesh.h"
 HEAP32[$44>>2] = $40; //@line 81 "modules/graphics/opengl/Mesh.h"
 $45 = ((($25)) + 20|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $46 = $45; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $47 = HEAP32[$this>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $48 = HEAP32[$0>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $49 = ($48|0)==($47|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($49) {
  $50 = $47; //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $70 = $28;$72 = $50;$74 = $47;
 } else {
  $$01$i$i = $48;$52 = $25;$__v$sroa$6$0 = $28;
  while(1) {
   $51 = ((($52)) + -20|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $53 = ((($$01$i$i)) + -20|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   ;HEAP32[$51>>2]=HEAP32[$53>>2]|0;HEAP32[$51+4>>2]=HEAP32[$53+4>>2]|0;HEAP32[$51+8>>2]=HEAP32[$53+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   ;HEAP32[$53>>2]=0|0;HEAP32[$53+4>>2]=0|0;HEAP32[$53+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $54 = ((($52)) + -8|0); //@line 81 "modules/graphics/opengl/Mesh.h"
   $55 = ((($$01$i$i)) + -8|0); //@line 81 "modules/graphics/opengl/Mesh.h"
   $56 = $55; //@line 81 "modules/graphics/opengl/Mesh.h"
   $57 = $56; //@line 81 "modules/graphics/opengl/Mesh.h"
   $58 = HEAP32[$57>>2]|0; //@line 81 "modules/graphics/opengl/Mesh.h"
   $59 = (($56) + 4)|0; //@line 81 "modules/graphics/opengl/Mesh.h"
   $60 = $59; //@line 81 "modules/graphics/opengl/Mesh.h"
   $61 = HEAP32[$60>>2]|0; //@line 81 "modules/graphics/opengl/Mesh.h"
   $62 = $54; //@line 81 "modules/graphics/opengl/Mesh.h"
   $63 = $62; //@line 81 "modules/graphics/opengl/Mesh.h"
   HEAP32[$63>>2] = $58; //@line 81 "modules/graphics/opengl/Mesh.h"
   $64 = (($62) + 4)|0; //@line 81 "modules/graphics/opengl/Mesh.h"
   $65 = $64; //@line 81 "modules/graphics/opengl/Mesh.h"
   HEAP32[$65>>2] = $61; //@line 81 "modules/graphics/opengl/Mesh.h"
   $66 = $__v$sroa$6$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $67 = ((($66)) + -20|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $68 = $67; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $69 = ($53|0)==($47|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if ($69) {
    $$lcssa = $68;
    break;
   } else {
    $$01$i$i = $53;$52 = $67;$__v$sroa$6$0 = $68;
   }
  }
  $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $$pre6 = HEAP32[$0>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $70 = $$lcssa;$72 = $$pre$i;$74 = $$pre6;
 }
 HEAP32[$this>>2] = $70; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $46; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $71 = $72; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $73 = ($74|0)==($71|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if (!($73)) {
  $76 = $74;
  while(1) {
   $75 = ((($76)) + -20|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($75); //@line 81 "modules/graphics/opengl/Mesh.h"
   $77 = ($75|0)==($71|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($77) {
    break;
   } else {
    $76 = $75;
   }
  }
 }
 $78 = ($72|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($78) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $79 = $72; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($79); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl9w_newTextEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$01$i2 = 0, $$02$i = 0, $$02$i1 = 0, $$pre$i = 0, $$pre$i$i$i = 0, $$pre$i3$i$i = 0, $$pre$i7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index7Z2D = 0, $eh$lpad$body$indexZ2D = 0, $t$1 = 0, $t$2 = 0, $text = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 32|0;
 $text = sp + 20|0;
 $1 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $2 = (__ZNK4love8graphics6opengl8Graphics9isCreatedEv($1)|0); //@line 52 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($2)) {
  (_luaL_error($L,329373,$vararg_buffer)|0); //@line 53 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $3 = (__ZN4love8graphics6opengl14luax_checkfontEP9lua_Statei($L,1)|0); //@line 898 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = (_lua_type($L,2)|0); //@line 901 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = ($4|0)<(1); //@line 901 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($5) {
  $6 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  HEAP32[$0>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $7 = ((($0)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$7>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $8 = ((($0)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$8>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  __THREW__ = 0;
  $9 = (invoke_iiii(3567,($6|0),($3|0),($0|0))|0); //@line 902 "modules/graphics/opengl/wrap_Graphics.cpp"
  $10 = __THREW__; __THREW__ = 0;
  $11 = $10&1;
  if (!($11)) {
   $12 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $13 = ($12|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($13)) {
    $14 = HEAP32[$7>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $15 = ($14|0)==($12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($15) {
     $20 = $12;
    } else {
     $17 = $14;
     while(1) {
      $16 = ((($17)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$7>>2] = $16; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($16); //@line 64 "modules/graphics/opengl/Font.h"
      $18 = HEAP32[$7>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $19 = ($18|0)==($12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($19) {
       break;
      } else {
       $17 = $18;
      }
     }
     $$pre$i3$i$i = HEAP32[$0>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $20 = $$pre$i3$i$i;
    }
    __ZdlPv($20); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $t$2 = $9;
   __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,20,$t$2); //@line 911 "modules/graphics/opengl/wrap_Graphics.cpp"
   $100 = HEAP32[$t$2>>2]|0; //@line 912 "modules/graphics/opengl/wrap_Graphics.cpp"
   $101 = ((($100)) + 12|0); //@line 912 "modules/graphics/opengl/wrap_Graphics.cpp"
   $102 = HEAP32[$101>>2]|0; //@line 912 "modules/graphics/opengl/wrap_Graphics.cpp"
   FUNCTION_TABLE_vi[$102 & 8191]($t$2); //@line 912 "modules/graphics/opengl/wrap_Graphics.cpp"
   STACKTOP = sp;return 1; //@line 913 "modules/graphics/opengl/wrap_Graphics.cpp"
  }
  $21 = ___cxa_find_matching_catch_3(8992|0)|0;
  $22 = tempRet0;
  $23 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $24 = ($23|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($24)) {
   $25 = HEAP32[$7>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $26 = ($25|0)==($23|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($26) {
    $31 = $23;
   } else {
    $28 = $25;
    while(1) {
     $27 = ((($28)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$7>>2] = $27; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($27); //@line 64 "modules/graphics/opengl/Font.h"
     $29 = HEAP32[$7>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $30 = ($29|0)==($23|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($30) {
      break;
     } else {
      $28 = $29;
     }
    }
    $$pre$i$i$i = HEAP32[$0>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $31 = $$pre$i$i$i;
   }
   __ZdlPv($31); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  $32 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
  $33 = ($22|0)==($32|0); //@line 535 "./common/runtime.h"
  if (!($33)) {
   $$01$i = $22;$$02$i = $21;
   ___resumeException($$02$i|0);
   // unreachable;
  }
  $34 = (___cxa_begin_catch(($21|0))|0); //@line 535 "./common/runtime.h"
  $35 = HEAP32[$34>>2]|0; //@line 539 "./common/runtime.h"
  $36 = ((($35)) + 8|0); //@line 539 "./common/runtime.h"
  $37 = HEAP32[$36>>2]|0; //@line 539 "./common/runtime.h"
  $38 = (FUNCTION_TABLE_ii[$37 & 8191]($34)|0); //@line 539 "./common/runtime.h"
  __THREW__ = 0;
  invoke_vii(3380,($L|0),($38|0)); //@line 539 "./common/runtime.h"
  $39 = __THREW__; __THREW__ = 0;
  $40 = $39&1;
  if (!($40)) {
   ___cxa_end_catch(); //@line 540 "./common/runtime.h"
   $41 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
   HEAP32[$vararg_buffer1>>2] = $41; //@line 543 "./common/runtime.h"
   (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
   $t$2 = 0;
   __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,20,$t$2); //@line 911 "modules/graphics/opengl/wrap_Graphics.cpp"
   $100 = HEAP32[$t$2>>2]|0; //@line 912 "modules/graphics/opengl/wrap_Graphics.cpp"
   $101 = ((($100)) + 12|0); //@line 912 "modules/graphics/opengl/wrap_Graphics.cpp"
   $102 = HEAP32[$101>>2]|0; //@line 912 "modules/graphics/opengl/wrap_Graphics.cpp"
   FUNCTION_TABLE_vi[$102 & 8191]($t$2); //@line 912 "modules/graphics/opengl/wrap_Graphics.cpp"
   STACKTOP = sp;return 1; //@line 913 "modules/graphics/opengl/wrap_Graphics.cpp"
  }
  $42 = ___cxa_find_matching_catch_2()|0;
  $43 = tempRet0;
  __THREW__ = 0;
  invoke_v(3378); //@line 540 "./common/runtime.h"
  $44 = __THREW__; __THREW__ = 0;
  $45 = $44&1;
  if ($45) {
   $46 = ___cxa_find_matching_catch_3(0|0)|0;
   $47 = tempRet0;
   ___clang_call_terminate($46); //@line 540 "./common/runtime.h"
   // unreachable; //@line 540 "./common/runtime.h"
  } else {
   $$01$i = $43;$$02$i = $42;
   ___resumeException($$02$i|0);
   // unreachable;
  }
 }
 HEAP32[$text>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $48 = ((($text)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$48>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $49 = ((($text)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$49>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 invoke_viii(3568,($L|0),2,($text|0)); //@line 906 "modules/graphics/opengl/wrap_Graphics.cpp"
 $50 = __THREW__; __THREW__ = 0;
 $51 = $50&1;
 L37: do {
  if ($51) {
   label = 38;
  } else {
   $52 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
   __THREW__ = 0;
   $53 = (invoke_iiii(3567,($52|0),($3|0),($text|0))|0); //@line 908 "modules/graphics/opengl/wrap_Graphics.cpp"
   $54 = __THREW__; __THREW__ = 0;
   $55 = $54&1;
   L39: do {
    if ($55) {
     $56 = ___cxa_find_matching_catch_3(8992|0)|0;
     $57 = tempRet0;
     $58 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
     $59 = ($57|0)==($58|0); //@line 535 "./common/runtime.h"
     do {
      if ($59) {
       $60 = (___cxa_begin_catch(($56|0))|0); //@line 535 "./common/runtime.h"
       $61 = HEAP32[$60>>2]|0; //@line 539 "./common/runtime.h"
       $62 = ((($61)) + 8|0); //@line 539 "./common/runtime.h"
       $63 = HEAP32[$62>>2]|0; //@line 539 "./common/runtime.h"
       $64 = (FUNCTION_TABLE_ii[$63 & 8191]($60)|0); //@line 539 "./common/runtime.h"
       __THREW__ = 0;
       invoke_vii(3380,($L|0),($64|0)); //@line 539 "./common/runtime.h"
       $65 = __THREW__; __THREW__ = 0;
       $66 = $65&1;
       if ($66) {
        $74 = ___cxa_find_matching_catch_2()|0;
        $75 = tempRet0;
        __THREW__ = 0;
        invoke_v(3378); //@line 540 "./common/runtime.h"
        $76 = __THREW__; __THREW__ = 0;
        $77 = $76&1;
        if (!($77)) {
         $$01$i2 = $75;$$02$i1 = $74;
         break;
        }
        $78 = ___cxa_find_matching_catch_3(0|0)|0;
        $79 = tempRet0;
        ___clang_call_terminate($78); //@line 540 "./common/runtime.h"
        // unreachable; //@line 540 "./common/runtime.h"
       }
       __THREW__ = 0;
       invoke_v(3378); //@line 540 "./common/runtime.h"
       $67 = __THREW__; __THREW__ = 0;
       $68 = $67&1;
       if ($68) {
        label = 38;
        break L37;
       }
       __THREW__ = 0;
       $69 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
       $70 = __THREW__; __THREW__ = 0;
       $71 = $70&1;
       if ($71) {
        label = 38;
        break L37;
       }
       __THREW__ = 0;
       HEAP32[$vararg_buffer3>>2] = $69; //@line 543 "./common/runtime.h"
       (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer3|0))|0); //@line 543 "./common/runtime.h"
       $72 = __THREW__; __THREW__ = 0;
       $73 = $72&1;
       if ($73) {
        label = 38;
        break L37;
       } else {
        $t$1 = 0;
        break L39;
       }
      } else {
       $$01$i2 = $57;$$02$i1 = $56;
      }
     } while(0);
     $eh$lpad$body$index7Z2D = $$01$i2;$eh$lpad$body$indexZ2D = $$02$i1;
     break L37;
    } else {
     $t$1 = $53;
    }
   } while(0);
   $80 = HEAP32[$text>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $81 = ($80|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($81)) {
    $82 = HEAP32[$48>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $83 = ($82|0)==($80|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($83) {
     $88 = $80;
    } else {
     $85 = $82;
     while(1) {
      $84 = ((($85)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$48>>2] = $84; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($84); //@line 64 "modules/graphics/opengl/Font.h"
      $86 = HEAP32[$48>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $87 = ($86|0)==($80|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($87) {
       break;
      } else {
       $85 = $86;
      }
     }
     $$pre$i = HEAP32[$text>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $88 = $$pre$i;
    }
    __ZdlPv($88); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $t$2 = $t$1;
   __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,20,$t$2); //@line 911 "modules/graphics/opengl/wrap_Graphics.cpp"
   $100 = HEAP32[$t$2>>2]|0; //@line 912 "modules/graphics/opengl/wrap_Graphics.cpp"
   $101 = ((($100)) + 12|0); //@line 912 "modules/graphics/opengl/wrap_Graphics.cpp"
   $102 = HEAP32[$101>>2]|0; //@line 912 "modules/graphics/opengl/wrap_Graphics.cpp"
   FUNCTION_TABLE_vi[$102 & 8191]($t$2); //@line 912 "modules/graphics/opengl/wrap_Graphics.cpp"
   STACKTOP = sp;return 1; //@line 913 "modules/graphics/opengl/wrap_Graphics.cpp"
  }
 } while(0);
 if ((label|0) == 38) {
  $89 = ___cxa_find_matching_catch_2()|0;
  $90 = tempRet0;
  $eh$lpad$body$index7Z2D = $90;$eh$lpad$body$indexZ2D = $89;
 }
 $91 = HEAP32[$text>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $92 = ($91|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($92) {
  ___resumeException($eh$lpad$body$indexZ2D|0);
  // unreachable;
 }
 $93 = HEAP32[$48>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $94 = ($93|0)==($91|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($94) {
  $99 = $91;
 } else {
  $96 = $93;
  while(1) {
   $95 = ((($96)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$48>>2] = $95; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($95); //@line 64 "modules/graphics/opengl/Font.h"
   $97 = HEAP32[$48>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $98 = ($97|0)==($91|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($98) {
    break;
   } else {
    $96 = $97;
   }
  }
  $$pre$i7 = HEAP32[$text>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $99 = $$pre$i7;
 }
 __ZdlPv($99); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8graphics6opengl10w_newVideoEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $video$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 916 "modules/graphics/opengl/wrap_Graphics.cpp"
 $vararg_buffer = sp; //@line 916 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZNK4love8graphics6opengl8Graphics9isCreatedEv($0)|0); //@line 52 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($1)) {
  (_luaL_error($L,329373,$vararg_buffer)|0); //@line 53 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $2 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,1,55)|0); //@line 920 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($2)) {
  (__ZN4love12luax_convobjEP9lua_StateiPKcS3_($L,1,330161,330167)|0); //@line 921 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $3 = (__ZN4love14luax_checktypeINS_5video11VideoStreamEEEPT_P9lua_StateiNS_4TypeE($L,1,55)|0); //@line 923 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $5 = (invoke_iii(3569,($4|0),($3|0))|0); //@line 926 "modules/graphics/opengl/wrap_Graphics.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  $video$0 = $5;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,21,$video$0); //@line 927 "modules/graphics/opengl/wrap_Graphics.cpp"
  $26 = HEAP32[$video$0>>2]|0; //@line 928 "modules/graphics/opengl/wrap_Graphics.cpp"
  $27 = ((($26)) + 12|0); //@line 928 "modules/graphics/opengl/wrap_Graphics.cpp"
  $28 = HEAP32[$27>>2]|0; //@line 928 "modules/graphics/opengl/wrap_Graphics.cpp"
  FUNCTION_TABLE_vi[$28 & 8191]($video$0); //@line 928 "modules/graphics/opengl/wrap_Graphics.cpp"
  STACKTOP = sp;return 1; //@line 929 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $8 = ___cxa_find_matching_catch_3(8992|0)|0;
 $9 = tempRet0;
 $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $11 = ($9|0)==($10|0); //@line 535 "./common/runtime.h"
 if (!($11)) {
  $$01$i = $9;$$02$i = $8;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $12 = (___cxa_begin_catch(($8|0))|0); //@line 535 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = ((($13)) + 8|0); //@line 539 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($16|0)); //@line 539 "./common/runtime.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if (!($18)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $19 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $19; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  $video$0 = 0;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,21,$video$0); //@line 927 "modules/graphics/opengl/wrap_Graphics.cpp"
  $26 = HEAP32[$video$0>>2]|0; //@line 928 "modules/graphics/opengl/wrap_Graphics.cpp"
  $27 = ((($26)) + 12|0); //@line 928 "modules/graphics/opengl/wrap_Graphics.cpp"
  $28 = HEAP32[$27>>2]|0; //@line 928 "modules/graphics/opengl/wrap_Graphics.cpp"
  FUNCTION_TABLE_vi[$28 & 8191]($video$0); //@line 928 "modules/graphics/opengl/wrap_Graphics.cpp"
  STACKTOP = sp;return 1; //@line 929 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $20 = ___cxa_find_matching_catch_2()|0;
 $21 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if ($23) {
  $24 = ___cxa_find_matching_catch_3(0|0)|0;
  $25 = tempRet0;
  ___clang_call_terminate($24); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $21;$$02$i = $20;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl10w_setColorEP9lua_State($L) {
 $L = $L|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $c$sroa$0$0 = 0.0, $c$sroa$0$0$in = 0.0, $c$sroa$10$0 = 0.0, $c$sroa$10$0$in = 0.0;
 var $c$sroa$6$0 = 0.0, $c$sroa$6$0$in = 0.0, $c$sroa$8$0 = 0.0, $c$sroa$8$0$in = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 16|0;
 $0 = sp;
 $1 = (_lua_type($L,1)|0); //@line 935 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = ($1|0)==(5); //@line 935 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($2) {
  _lua_rawgeti($L,1,1); //@line 938 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_rawgeti($L,1,2); //@line 938 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_rawgeti($L,1,3); //@line 938 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_rawgeti($L,1,4); //@line 938 "modules/graphics/opengl/wrap_Graphics.cpp"
  $3 = (+_luaL_checknumber($L,-4)); //@line 940 "modules/graphics/opengl/wrap_Graphics.cpp"
  $4 = (+_luaL_checknumber($L,-3)); //@line 941 "modules/graphics/opengl/wrap_Graphics.cpp"
  $5 = (+_luaL_checknumber($L,-2)); //@line 942 "modules/graphics/opengl/wrap_Graphics.cpp"
  $6 = (+_luaL_optnumber($L,-1,255.0)); //@line 943 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_settop($L,-5); //@line 945 "modules/graphics/opengl/wrap_Graphics.cpp"
  $c$sroa$0$0$in = $3;$c$sroa$10$0$in = $6;$c$sroa$6$0$in = $4;$c$sroa$8$0$in = $5;
 } else {
  $7 = (+_luaL_checknumber($L,1)); //@line 949 "modules/graphics/opengl/wrap_Graphics.cpp"
  $8 = (+_luaL_checknumber($L,2)); //@line 950 "modules/graphics/opengl/wrap_Graphics.cpp"
  $9 = (+_luaL_checknumber($L,3)); //@line 951 "modules/graphics/opengl/wrap_Graphics.cpp"
  $10 = (+_luaL_optnumber($L,4,255.0)); //@line 952 "modules/graphics/opengl/wrap_Graphics.cpp"
  $c$sroa$0$0$in = $7;$c$sroa$10$0$in = $10;$c$sroa$6$0$in = $8;$c$sroa$8$0$in = $9;
 }
 $c$sroa$0$0 = $c$sroa$0$0$in; //@line 940 "modules/graphics/opengl/wrap_Graphics.cpp"
 $c$sroa$6$0 = $c$sroa$6$0$in; //@line 941 "modules/graphics/opengl/wrap_Graphics.cpp"
 $c$sroa$8$0 = $c$sroa$8$0$in; //@line 942 "modules/graphics/opengl/wrap_Graphics.cpp"
 $c$sroa$10$0 = $c$sroa$10$0$in; //@line 943 "modules/graphics/opengl/wrap_Graphics.cpp"
 $11 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 HEAPF32[$0>>2] = $c$sroa$0$0; //@line 954 "modules/graphics/opengl/wrap_Graphics.cpp"
 $12 = ((($0)) + 4|0); //@line 954 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAPF32[$12>>2] = $c$sroa$6$0; //@line 954 "modules/graphics/opengl/wrap_Graphics.cpp"
 $13 = ((($0)) + 8|0); //@line 954 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAPF32[$13>>2] = $c$sroa$8$0; //@line 954 "modules/graphics/opengl/wrap_Graphics.cpp"
 $14 = ((($0)) + 12|0); //@line 954 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAPF32[$14>>2] = $c$sroa$10$0; //@line 954 "modules/graphics/opengl/wrap_Graphics.cpp"
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;HEAP32[$$byval_copy+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$$byval_copy+8>>2]=HEAP32[$0+8>>2]|0;HEAP32[$$byval_copy+12>>2]=HEAP32[$0+12>>2]|0; //@line 954 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love8graphics6opengl8Graphics8setColorENS0_6ColorTIfEE($11,$$byval_copy); //@line 954 "modules/graphics/opengl/wrap_Graphics.cpp"
 STACKTOP = sp;return 0; //@line 955 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl10w_getColorEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0.0, $2 = 0.0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0, $c = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $c = sp;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZNK4love8graphics6opengl8Graphics8getColorEv($c,$0); //@line 960 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = +HEAPF32[$c>>2]; //@line 961 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = $1; //@line 961 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushnumber($L,$2); //@line 961 "modules/graphics/opengl/wrap_Graphics.cpp"
 $3 = ((($c)) + 4|0); //@line 962 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = +HEAPF32[$3>>2]; //@line 962 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = $4; //@line 962 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushnumber($L,$5); //@line 962 "modules/graphics/opengl/wrap_Graphics.cpp"
 $6 = ((($c)) + 8|0); //@line 963 "modules/graphics/opengl/wrap_Graphics.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 963 "modules/graphics/opengl/wrap_Graphics.cpp"
 $8 = $7; //@line 963 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushnumber($L,$8); //@line 963 "modules/graphics/opengl/wrap_Graphics.cpp"
 $9 = ((($c)) + 12|0); //@line 964 "modules/graphics/opengl/wrap_Graphics.cpp"
 $10 = +HEAPF32[$9>>2]; //@line 964 "modules/graphics/opengl/wrap_Graphics.cpp"
 $11 = $10; //@line 964 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushnumber($L,$11); //@line 964 "modules/graphics/opengl/wrap_Graphics.cpp"
 STACKTOP = sp;return 4; //@line 965 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl20w_setBackgroundColorEP9lua_State($L) {
 $L = $L|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $c$sroa$0$0 = 0.0, $c$sroa$0$0$in = 0.0, $c$sroa$10$0 = 0.0, $c$sroa$10$0$in = 0.0;
 var $c$sroa$6$0 = 0.0, $c$sroa$6$0$in = 0.0, $c$sroa$8$0 = 0.0, $c$sroa$8$0$in = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 16|0;
 $0 = sp;
 $1 = (_lua_type($L,1)|0); //@line 971 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = ($1|0)==(5); //@line 971 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($2) {
  _lua_rawgeti($L,1,1); //@line 974 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_rawgeti($L,1,2); //@line 974 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_rawgeti($L,1,3); //@line 974 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_rawgeti($L,1,4); //@line 974 "modules/graphics/opengl/wrap_Graphics.cpp"
  $3 = (+_luaL_checknumber($L,-4)); //@line 976 "modules/graphics/opengl/wrap_Graphics.cpp"
  $4 = (+_luaL_checknumber($L,-3)); //@line 977 "modules/graphics/opengl/wrap_Graphics.cpp"
  $5 = (+_luaL_checknumber($L,-2)); //@line 978 "modules/graphics/opengl/wrap_Graphics.cpp"
  $6 = (+_luaL_optnumber($L,-1,255.0)); //@line 979 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_settop($L,-5); //@line 981 "modules/graphics/opengl/wrap_Graphics.cpp"
  $c$sroa$0$0$in = $3;$c$sroa$10$0$in = $6;$c$sroa$6$0$in = $4;$c$sroa$8$0$in = $5;
 } else {
  $7 = (+_luaL_checknumber($L,1)); //@line 985 "modules/graphics/opengl/wrap_Graphics.cpp"
  $8 = (+_luaL_checknumber($L,2)); //@line 986 "modules/graphics/opengl/wrap_Graphics.cpp"
  $9 = (+_luaL_checknumber($L,3)); //@line 987 "modules/graphics/opengl/wrap_Graphics.cpp"
  $10 = (+_luaL_optnumber($L,4,255.0)); //@line 988 "modules/graphics/opengl/wrap_Graphics.cpp"
  $c$sroa$0$0$in = $7;$c$sroa$10$0$in = $10;$c$sroa$6$0$in = $8;$c$sroa$8$0$in = $9;
 }
 $c$sroa$0$0 = $c$sroa$0$0$in; //@line 976 "modules/graphics/opengl/wrap_Graphics.cpp"
 $c$sroa$6$0 = $c$sroa$6$0$in; //@line 977 "modules/graphics/opengl/wrap_Graphics.cpp"
 $c$sroa$8$0 = $c$sroa$8$0$in; //@line 978 "modules/graphics/opengl/wrap_Graphics.cpp"
 $c$sroa$10$0 = $c$sroa$10$0$in; //@line 979 "modules/graphics/opengl/wrap_Graphics.cpp"
 $11 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 HEAPF32[$0>>2] = $c$sroa$0$0; //@line 990 "modules/graphics/opengl/wrap_Graphics.cpp"
 $12 = ((($0)) + 4|0); //@line 990 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAPF32[$12>>2] = $c$sroa$6$0; //@line 990 "modules/graphics/opengl/wrap_Graphics.cpp"
 $13 = ((($0)) + 8|0); //@line 990 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAPF32[$13>>2] = $c$sroa$8$0; //@line 990 "modules/graphics/opengl/wrap_Graphics.cpp"
 $14 = ((($0)) + 12|0); //@line 990 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAPF32[$14>>2] = $c$sroa$10$0; //@line 990 "modules/graphics/opengl/wrap_Graphics.cpp"
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;HEAP32[$$byval_copy+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$$byval_copy+8>>2]=HEAP32[$0+8>>2]|0;HEAP32[$$byval_copy+12>>2]=HEAP32[$0+12>>2]|0; //@line 990 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love8graphics6opengl8Graphics18setBackgroundColorENS0_6ColorTIfEE($11,$$byval_copy); //@line 990 "modules/graphics/opengl/wrap_Graphics.cpp"
 STACKTOP = sp;return 0; //@line 991 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl20w_getBackgroundColorEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0.0, $2 = 0.0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0, $c = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $c = sp;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZNK4love8graphics6opengl8Graphics18getBackgroundColorEv($c,$0); //@line 996 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = +HEAPF32[$c>>2]; //@line 997 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = $1; //@line 997 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushnumber($L,$2); //@line 997 "modules/graphics/opengl/wrap_Graphics.cpp"
 $3 = ((($c)) + 4|0); //@line 998 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = +HEAPF32[$3>>2]; //@line 998 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = $4; //@line 998 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushnumber($L,$5); //@line 998 "modules/graphics/opengl/wrap_Graphics.cpp"
 $6 = ((($c)) + 8|0); //@line 999 "modules/graphics/opengl/wrap_Graphics.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 999 "modules/graphics/opengl/wrap_Graphics.cpp"
 $8 = $7; //@line 999 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushnumber($L,$8); //@line 999 "modules/graphics/opengl/wrap_Graphics.cpp"
 $9 = ((($c)) + 12|0); //@line 1000 "modules/graphics/opengl/wrap_Graphics.cpp"
 $10 = +HEAPF32[$9>>2]; //@line 1000 "modules/graphics/opengl/wrap_Graphics.cpp"
 $11 = $10; //@line 1000 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushnumber($L,$11); //@line 1000 "modules/graphics/opengl/wrap_Graphics.cpp"
 STACKTOP = sp;return 4; //@line 1001 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl12w_setNewFontEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 (__ZN4love8graphics6opengl9w_newFontEP9lua_State($L)|0); //@line 1006 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE($L,-1,14)|0); //@line 1007 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZN4love8graphics6opengl8Graphics7setFontEPNS1_4FontE($1,$0); //@line 1008 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 1; //@line 1009 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love8graphics6opengl9w_setFontEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE($L,1,14)|0); //@line 1014 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZN4love8graphics6opengl8Graphics7setFontEPNS1_4FontE($1,$0); //@line 1015 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 0; //@line 1016 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl9w_getFontEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 1019 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $1 = (invoke_ii(3570,($0|0))|0); //@line 1022 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  $f$0 = $1;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,14,$f$0); //@line 1024 "modules/graphics/opengl/wrap_Graphics.cpp"
  STACKTOP = sp;return 1; //@line 1025 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $4 = ___cxa_find_matching_catch_3(8992|0)|0;
 $5 = tempRet0;
 $6 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $7 = ($5|0)==($6|0); //@line 535 "./common/runtime.h"
 if (!($7)) {
  $$01$i = $5;$$02$i = $4;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $8 = (___cxa_begin_catch(($4|0))|0); //@line 535 "./common/runtime.h"
 $9 = HEAP32[$8>>2]|0; //@line 539 "./common/runtime.h"
 $10 = ((($9)) + 8|0); //@line 539 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($8)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($12|0)); //@line 539 "./common/runtime.h"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if (!($14)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $15 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $15; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $f$0 = 0;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,14,$f$0); //@line 1024 "modules/graphics/opengl/wrap_Graphics.cpp"
  STACKTOP = sp;return 1; //@line 1025 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $16 = ___cxa_find_matching_catch_2()|0;
 $17 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $20 = ___cxa_find_matching_catch_3(0|0)|0;
  $21 = tempRet0;
  ___clang_call_terminate($20); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $17;$$02$i = $16;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl14w_setColorMaskEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $mask$sroa$0$0 = 0, $mask$sroa$10$0 = 0, $mask$sroa$6$0 = 0;
 var $mask$sroa$8$0 = 0, $phitmp = 0, $phitmp2 = 0, $phitmp3 = 0, $tmpcast$byval_copy = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tmpcast$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = (_lua_gettop($L)|0); //@line 1032 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = ($1|0)<(2); //@line 1032 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($2) {
  $3 = (_lua_type($L,1)|0); //@line 1032 "modules/graphics/opengl/wrap_Graphics.cpp"
  $4 = ($3|0)<(1); //@line 1032 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($4) {
   $mask$sroa$0$0 = 1;$mask$sroa$10$0 = 16777216;$mask$sroa$6$0 = 256;$mask$sroa$8$0 = 65536;
  } else {
   label = 3;
  }
 } else {
  label = 3;
 }
 if ((label|0) == 3) {
  $5 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,1)|0); //@line 1039 "modules/graphics/opengl/wrap_Graphics.cpp"
  $6 = $5&1; //@line 1039 "modules/graphics/opengl/wrap_Graphics.cpp"
  $7 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,2)|0); //@line 1040 "modules/graphics/opengl/wrap_Graphics.cpp"
  $8 = $7&1; //@line 1040 "modules/graphics/opengl/wrap_Graphics.cpp"
  $9 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,3)|0); //@line 1041 "modules/graphics/opengl/wrap_Graphics.cpp"
  $10 = $9&1; //@line 1041 "modules/graphics/opengl/wrap_Graphics.cpp"
  $11 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,4)|0); //@line 1042 "modules/graphics/opengl/wrap_Graphics.cpp"
  $12 = $11&1; //@line 1042 "modules/graphics/opengl/wrap_Graphics.cpp"
  $phitmp = $12 << 24;
  $phitmp2 = $10 << 16;
  $phitmp3 = $8 << 8;
  $mask$sroa$0$0 = $6;$mask$sroa$10$0 = $phitmp;$mask$sroa$6$0 = $phitmp3;$mask$sroa$8$0 = $phitmp2;
 }
 $13 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $14 = $mask$sroa$8$0 | $mask$sroa$10$0; //@line 1045 "modules/graphics/opengl/wrap_Graphics.cpp"
 $15 = $14 | $mask$sroa$6$0; //@line 1045 "modules/graphics/opengl/wrap_Graphics.cpp"
 $16 = $15 | $mask$sroa$0$0; //@line 1045 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAP32[$0>>2] = $16; //@line 1045 "modules/graphics/opengl/wrap_Graphics.cpp"
 ;HEAP8[$tmpcast$byval_copy>>0]=HEAP8[$0>>0]|0;HEAP8[$tmpcast$byval_copy+1>>0]=HEAP8[$0+1>>0]|0;HEAP8[$tmpcast$byval_copy+2>>0]=HEAP8[$0+2>>0]|0;HEAP8[$tmpcast$byval_copy+3>>0]=HEAP8[$0+3>>0]|0; //@line 1045 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love8graphics6opengl8Graphics12setColorMaskENS0_8Graphics9ColorMaskE($13,$tmpcast$byval_copy); //@line 1045 "modules/graphics/opengl/wrap_Graphics.cpp"
 STACKTOP = sp;return 0; //@line 1047 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl14w_getColorMaskEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $mask = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $mask = sp;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZNK4love8graphics6opengl8Graphics12getColorMaskEv($mask,$0); //@line 1052 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = HEAP8[$mask>>0]|0; //@line 1054 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 1054 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$2); //@line 1054 "modules/graphics/opengl/wrap_Graphics.cpp"
 $3 = ((($mask)) + 1|0); //@line 1055 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = HEAP8[$3>>0]|0; //@line 1055 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = ($4<<24>>24)!=(0); //@line 1055 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$5); //@line 1055 "modules/graphics/opengl/wrap_Graphics.cpp"
 $6 = ((($mask)) + 2|0); //@line 1056 "modules/graphics/opengl/wrap_Graphics.cpp"
 $7 = HEAP8[$6>>0]|0; //@line 1056 "modules/graphics/opengl/wrap_Graphics.cpp"
 $8 = ($7<<24>>24)!=(0); //@line 1056 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$8); //@line 1056 "modules/graphics/opengl/wrap_Graphics.cpp"
 $9 = ((($mask)) + 3|0); //@line 1057 "modules/graphics/opengl/wrap_Graphics.cpp"
 $10 = HEAP8[$9>>0]|0; //@line 1057 "modules/graphics/opengl/wrap_Graphics.cpp"
 $11 = ($10<<24>>24)!=(0); //@line 1057 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$11); //@line 1057 "modules/graphics/opengl/wrap_Graphics.cpp"
 STACKTOP = sp;return 4; //@line 1059 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl14w_setBlendModeEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $$1 = 0, $$2 = 0, $$idx$val$val = 0, $$idx2$val$val = 0, $$idx2$val$val$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $alphamode = 0, $mode = 0, $vararg_buffer = 0, $vararg_buffer1 = 0;
 var $vararg_buffer4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $mode = sp + 24|0;
 $alphamode = sp + 20|0;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 1065 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = (__ZN4love8graphics8Graphics11getConstantEPKcRNS1_9BlendModeE($0,$mode)|0); //@line 1066 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($1)) {
  HEAP32[$vararg_buffer>>2] = $0; //@line 1067 "modules/graphics/opengl/wrap_Graphics.cpp"
  $2 = (_luaL_error($L,330182,$vararg_buffer)|0); //@line 1067 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$2 = $2;
  STACKTOP = sp;return ($$2|0); //@line 1079 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 HEAP32[$alphamode>>2] = 0; //@line 1069 "modules/graphics/opengl/wrap_Graphics.cpp"
 $3 = (_lua_type($L,2)|0); //@line 1070 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = ($3|0)<(1); //@line 1070 "modules/graphics/opengl/wrap_Graphics.cpp"
 do {
  if ($4) {
   $$idx2$val$val = 0;
   label = 7;
  } else {
   $5 = (_luaL_checklstring($L,2,0)|0); //@line 1072 "modules/graphics/opengl/wrap_Graphics.cpp"
   $6 = (__ZN4love8graphics8Graphics11getConstantEPKcRNS1_10BlendAlphaE($5,$alphamode)|0); //@line 1073 "modules/graphics/opengl/wrap_Graphics.cpp"
   if ($6) {
    $$idx2$val$val$pre = HEAP32[$alphamode>>2]|0;
    $$idx2$val$val = $$idx2$val$val$pre;
    label = 7;
    break;
   } else {
    HEAP32[$vararg_buffer1>>2] = $5; //@line 1074 "modules/graphics/opengl/wrap_Graphics.cpp"
    $7 = (_luaL_error($L,330205,$vararg_buffer1)|0); //@line 1074 "modules/graphics/opengl/wrap_Graphics.cpp"
    $$1 = $7;
    break;
   }
  }
 } while(0);
 do {
  if ((label|0) == 7) {
   $$idx$val$val = HEAP32[$mode>>2]|0;
   $8 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
   __THREW__ = 0;
   invoke_viii(3571,($8|0),($$idx$val$val|0),($$idx2$val$val|0)); //@line 1077 "modules/graphics/opengl/wrap_Graphics.cpp"
   $9 = __THREW__; __THREW__ = 0;
   $10 = $9&1;
   if ($10) {
    $11 = ___cxa_find_matching_catch_3(8992|0)|0;
    $12 = tempRet0;
    $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
    $14 = ($12|0)==($13|0); //@line 535 "./common/runtime.h"
    if (!($14)) {
     $$01$i = $12;$$02$i = $11;
     ___resumeException($$02$i|0);
     // unreachable;
    }
    $15 = (___cxa_begin_catch(($11|0))|0); //@line 535 "./common/runtime.h"
    $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
    $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
    $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
    $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
    __THREW__ = 0;
    invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
    $20 = __THREW__; __THREW__ = 0;
    $21 = $20&1;
    if (!($21)) {
     ___cxa_end_catch(); //@line 540 "./common/runtime.h"
     $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
     HEAP32[$vararg_buffer4>>2] = $22; //@line 543 "./common/runtime.h"
     (_luaL_error($L,595867,$vararg_buffer4)|0); //@line 543 "./common/runtime.h"
     $$1 = 0;
     break;
    }
    $23 = ___cxa_find_matching_catch_2()|0;
    $24 = tempRet0;
    __THREW__ = 0;
    invoke_v(3378); //@line 540 "./common/runtime.h"
    $25 = __THREW__; __THREW__ = 0;
    $26 = $25&1;
    if ($26) {
     $27 = ___cxa_find_matching_catch_3(0|0)|0;
     $28 = tempRet0;
     ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
     // unreachable; //@line 540 "./common/runtime.h"
    } else {
     $$01$i = $24;$$02$i = $23;
     ___resumeException($$02$i|0);
     // unreachable;
    }
   } else {
    $$1 = 0;
   }
  }
 } while(0);
 $$2 = $$1;
 STACKTOP = sp;return ($$2|0); //@line 1079 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl14w_getBlendModeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $alphamode = 0, $alphastr = 0, $str = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $str = sp + 20|0;
 $alphastr = sp + 16|0;
 $alphamode = sp + 12|0;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZNK4love8graphics6opengl8Graphics12getBlendModeERNS0_8Graphics10BlendAlphaE($0,$alphamode)|0); //@line 1087 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = (__ZN4love8graphics8Graphics11getConstantENS1_9BlendModeERPKc($1,$str)|0); //@line 1089 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($2)) {
  $3 = (_luaL_error($L,330234,$vararg_buffer)|0); //@line 1090 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 1098 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $4 = HEAP32[$alphamode>>2]|0; //@line 1092 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = (__ZN4love8graphics8Graphics11getConstantENS1_10BlendAlphaERPKc($4,$alphastr)|0); //@line 1092 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($5) {
  $7 = HEAP32[$str>>2]|0; //@line 1095 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_pushstring($L,$7); //@line 1095 "modules/graphics/opengl/wrap_Graphics.cpp"
  $8 = HEAP32[$alphastr>>2]|0; //@line 1096 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_pushstring($L,$8); //@line 1096 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = 2;
  STACKTOP = sp;return ($$0|0); //@line 1098 "modules/graphics/opengl/wrap_Graphics.cpp"
 } else {
  $6 = (_luaL_error($L,330253,$vararg_buffer1)|0); //@line 1093 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = $6;
  STACKTOP = sp;return ($$0|0); //@line 1098 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl18w_setDefaultFilterEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, $f = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $f = sp + 16|0;
 HEAP32[$f>>2] = 1; //@line 62 "./modules/graphics/Texture.h"
 $0 = ((($f)) + 4|0); //@line 63 "./modules/graphics/Texture.h"
 HEAP32[$0>>2] = 1; //@line 63 "./modules/graphics/Texture.h"
 $1 = ((($f)) + 8|0); //@line 64 "./modules/graphics/Texture.h"
 HEAP32[$1>>2] = 0; //@line 64 "./modules/graphics/Texture.h"
 $2 = ((($f)) + 12|0); //@line 65 "./modules/graphics/Texture.h"
 HEAPF32[$2>>2] = 1.0; //@line 65 "./modules/graphics/Texture.h"
 $3 = (_luaL_checklstring($L,1,0)|0); //@line 1104 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = (_luaL_optlstring($L,2,$3,0)|0); //@line 1105 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = (__ZN4love8graphics7Texture11getConstantEPKcRNS1_10FilterModeE($3,$f)|0); //@line 1107 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($5)) {
  HEAP32[$vararg_buffer>>2] = $3; //@line 1108 "modules/graphics/opengl/wrap_Graphics.cpp"
  $6 = (_luaL_error($L,439370,$vararg_buffer)|0); //@line 1108 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = $6;
  STACKTOP = sp;return ($$0|0); //@line 1117 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $7 = (__ZN4love8graphics7Texture11getConstantEPKcRNS1_10FilterModeE($4,$0)|0); //@line 1109 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($7) {
  $9 = (+_luaL_optnumber($L,3,1.0)); //@line 1112 "modules/graphics/opengl/wrap_Graphics.cpp"
  $10 = $9; //@line 1112 "modules/graphics/opengl/wrap_Graphics.cpp"
  HEAPF32[$2>>2] = $10; //@line 1112 "modules/graphics/opengl/wrap_Graphics.cpp"
  $11 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  __ZN4love8graphics6opengl8Graphics16setDefaultFilterERKNS0_7Texture6FilterE($11,$f); //@line 1114 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 1117 "modules/graphics/opengl/wrap_Graphics.cpp"
 } else {
  HEAP32[$vararg_buffer1>>2] = $4; //@line 1110 "modules/graphics/opengl/wrap_Graphics.cpp"
  $8 = (_luaL_error($L,439370,$vararg_buffer1)|0); //@line 1110 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = $8;
  STACKTOP = sp;return ($$0|0); //@line 1117 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl18w_getDefaultFilterEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $magstr = 0, $minstr = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $minstr = sp + 16|0;
 $magstr = sp + 12|0;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZNK4love8graphics6opengl8Graphics16getDefaultFilterEv($0)|0); //@line 1121 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 1124 "modules/graphics/opengl/wrap_Graphics.cpp"
 $3 = (__ZN4love8graphics7Texture11getConstantENS1_10FilterModeERPKc($2,$minstr)|0); //@line 1124 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($3)) {
  $4 = (_luaL_error($L,330278,$vararg_buffer)|0); //@line 1125 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = $4;
  STACKTOP = sp;return ($$0|0); //@line 1132 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $5 = ((($1)) + 4|0); //@line 1126 "modules/graphics/opengl/wrap_Graphics.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 1126 "modules/graphics/opengl/wrap_Graphics.cpp"
 $7 = (__ZN4love8graphics7Texture11getConstantENS1_10FilterModeERPKc($6,$magstr)|0); //@line 1126 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($7) {
  $9 = HEAP32[$minstr>>2]|0; //@line 1128 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_pushstring($L,$9); //@line 1128 "modules/graphics/opengl/wrap_Graphics.cpp"
  $10 = HEAP32[$magstr>>2]|0; //@line 1129 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_pushstring($L,$10); //@line 1129 "modules/graphics/opengl/wrap_Graphics.cpp"
  $11 = ((($1)) + 12|0); //@line 1130 "modules/graphics/opengl/wrap_Graphics.cpp"
  $12 = +HEAPF32[$11>>2]; //@line 1130 "modules/graphics/opengl/wrap_Graphics.cpp"
  $13 = $12; //@line 1130 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_pushnumber($L,$13); //@line 1130 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = 3;
  STACKTOP = sp;return ($$0|0); //@line 1132 "modules/graphics/opengl/wrap_Graphics.cpp"
 } else {
  $8 = (_luaL_error($L,330311,$vararg_buffer1)|0); //@line 1127 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = $8;
  STACKTOP = sp;return ($$0|0); //@line 1132 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl24w_setDefaultMipmapFilterEP9lua_State($L) {
 $L = $L|0;
 var $$1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0, $filter = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $filter = sp + 4|0;
 HEAP32[$filter>>2] = 0; //@line 1136 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = (_lua_type($L,1)|0); //@line 1137 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = ($0|0)<(1); //@line 1137 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($1)) {
  $2 = (_luaL_checklstring($L,1,0)|0); //@line 1139 "modules/graphics/opengl/wrap_Graphics.cpp"
  $3 = (__ZN4love8graphics7Texture11getConstantEPKcRNS1_10FilterModeE($2,$filter)|0); //@line 1140 "modules/graphics/opengl/wrap_Graphics.cpp"
  if (!($3)) {
   HEAP32[$vararg_buffer>>2] = $2; //@line 1141 "modules/graphics/opengl/wrap_Graphics.cpp"
   $4 = (_luaL_error($L,439370,$vararg_buffer)|0); //@line 1141 "modules/graphics/opengl/wrap_Graphics.cpp"
   $$1 = $4;
   STACKTOP = sp;return ($$1|0); //@line 1149 "modules/graphics/opengl/wrap_Graphics.cpp"
  }
 }
 $5 = (+_luaL_optnumber($L,2,0.0)); //@line 1144 "modules/graphics/opengl/wrap_Graphics.cpp"
 $6 = $5; //@line 1144 "modules/graphics/opengl/wrap_Graphics.cpp"
 $7 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $8 = HEAP32[$filter>>2]|0; //@line 1146 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love8graphics6opengl8Graphics22setDefaultMipmapFilterENS0_7Texture10FilterModeEf($7,$8,$6); //@line 1146 "modules/graphics/opengl/wrap_Graphics.cpp"
 $$1 = 0;
 STACKTOP = sp;return ($$1|0); //@line 1149 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl24w_getDefaultMipmapFilterEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $filter = 0, $sharpness = 0, $str = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $filter = sp + 8|0;
 $sharpness = sp + 4|0;
 $str = sp;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZNK4love8graphics6opengl8Graphics22getDefaultMipmapFilterEPNS0_7Texture10FilterModeEPf($0,$filter,$sharpness); //@line 1156 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = HEAP32[$filter>>2]|0; //@line 1159 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = (__ZN4love8graphics7Texture11getConstantENS1_10FilterModeERPKc($1,$str)|0); //@line 1159 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($2) {
  $3 = HEAP32[$str>>2]|0; //@line 1160 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_pushstring($L,$3); //@line 1160 "modules/graphics/opengl/wrap_Graphics.cpp"
 } else {
  _lua_pushnil($L); //@line 1162 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $4 = +HEAPF32[$sharpness>>2]; //@line 1164 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = $4; //@line 1164 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushnumber($L,$5); //@line 1164 "modules/graphics/opengl/wrap_Graphics.cpp"
 STACKTOP = sp;return 2; //@line 1166 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl14w_setLineWidthEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1)); //@line 1171 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = $0; //@line 1171 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZN4love8graphics6opengl8Graphics12setLineWidthEf($2,$1); //@line 1172 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 0; //@line 1173 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl14w_setLineStyleEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $style = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $style = sp + 4|0;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 1179 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = (__ZN4love8graphics8Graphics11getConstantEPKcRNS1_9LineStyleE($0,$style)|0); //@line 1180 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($1) {
  $3 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  $4 = HEAP32[$style>>2]|0; //@line 1183 "modules/graphics/opengl/wrap_Graphics.cpp"
  __ZN4love8graphics6opengl8Graphics12setLineStyleENS0_8Graphics9LineStyleE($3,$4); //@line 1183 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 1185 "modules/graphics/opengl/wrap_Graphics.cpp"
 } else {
  HEAP32[$vararg_buffer>>2] = $0; //@line 1181 "modules/graphics/opengl/wrap_Graphics.cpp"
  $2 = (_luaL_error($L,330345,$vararg_buffer)|0); //@line 1181 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = $2;
  STACKTOP = sp;return ($$0|0); //@line 1185 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl13w_setLineJoinEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $join = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $join = sp + 4|0;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 1190 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = (__ZN4love8graphics8Graphics11getConstantEPKcRNS1_8LineJoinE($0,$join)|0); //@line 1191 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($1) {
  $3 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  $4 = HEAP32[$join>>2]|0; //@line 1194 "modules/graphics/opengl/wrap_Graphics.cpp"
  __ZN4love8graphics6opengl8Graphics11setLineJoinENS0_8Graphics8LineJoinE($3,$4); //@line 1194 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 1196 "modules/graphics/opengl/wrap_Graphics.cpp"
 } else {
  HEAP32[$vararg_buffer>>2] = $0; //@line 1192 "modules/graphics/opengl/wrap_Graphics.cpp"
  $2 = (_luaL_error($L,330368,$vararg_buffer)|0); //@line 1192 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = $2;
  STACKTOP = sp;return ($$0|0); //@line 1196 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl14w_getLineWidthEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (+__ZNK4love8graphics6opengl8Graphics12getLineWidthEv($0)); //@line 1200 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = $1; //@line 1200 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushnumber($L,$2); //@line 1200 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 1; //@line 1201 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl14w_getLineStyleEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $str = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $str = sp + 4|0;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZNK4love8graphics6opengl8Graphics12getLineStyleEv($0)|0); //@line 1206 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = (__ZN4love8graphics8Graphics11getConstantENS1_9LineStyleERPKc($1,$str)|0); //@line 1208 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($2) {
  $4 = HEAP32[$str>>2]|0; //@line 1210 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_pushstring($L,$4); //@line 1210 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 1212 "modules/graphics/opengl/wrap_Graphics.cpp"
 } else {
  $3 = (_luaL_error($L,330395,$vararg_buffer)|0); //@line 1209 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 1212 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl13w_getLineJoinEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $str = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $str = sp + 4|0;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZNK4love8graphics6opengl8Graphics11getLineJoinEv($0)|0); //@line 1216 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = (__ZN4love8graphics8Graphics11getConstantENS1_8LineJoinERPKc($1,$str)|0); //@line 1218 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($2) {
  $4 = HEAP32[$str>>2]|0; //@line 1220 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_pushstring($L,$4); //@line 1220 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 1222 "modules/graphics/opengl/wrap_Graphics.cpp"
 } else {
  $3 = (_luaL_error($L,330414,$vararg_buffer)|0); //@line 1219 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 1222 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl14w_setPointSizeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1)); //@line 1226 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = $0; //@line 1226 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZN4love8graphics6opengl8Graphics12setPointSizeEf($2,$1); //@line 1227 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 0; //@line 1228 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl14w_getPointSizeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (+__ZNK4love8graphics6opengl8Graphics12getPointSizeEv($0)); //@line 1233 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = $1; //@line 1233 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushnumber($L,$2); //@line 1233 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 1; //@line 1234 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl14w_setWireframeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,1)|0); //@line 1239 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love8graphics6opengl8Graphics12setWireframeEb($0,$1); //@line 1239 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 0; //@line 1240 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl13w_isWireframeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZNK4love8graphics6opengl8Graphics11isWireframeEv($0)|0); //@line 1245 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$1); //@line 1245 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 1; //@line 1246 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl15w_newScreenshotEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 1249 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = (__ZN4love14luax_getmoduleINS_5image5ImageEEEPT_P9lua_StateNS_4TypeE($L,58)|0); //@line 1251 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,1,0)|0); //@line 1252 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 $3 = (invoke_iiii(3572,($2|0),($0|0),($1|0))|0); //@line 1255 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 do {
  if ($5) {
   $6 = ___cxa_find_matching_catch_3(8992|0)|0;
   $7 = tempRet0;
   $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
   if (!($9)) {
    $$01$i = $7;$$02$i = $6;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
   $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
   $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
   $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
   $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if (!($16)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $17; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $i$0 = 0;
    break;
   }
   $18 = ___cxa_find_matching_catch_2()|0;
   $19 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if ($21) {
    $22 = ___cxa_find_matching_catch_3(0|0)|0;
    $23 = tempRet0;
    ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $19;$$02$i = $18;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $i$0 = $3;
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,22,$i$0); //@line 1257 "modules/graphics/opengl/wrap_Graphics.cpp"
 $24 = HEAP32[$i$0>>2]|0; //@line 1258 "modules/graphics/opengl/wrap_Graphics.cpp"
 $25 = ((($24)) + 12|0); //@line 1258 "modules/graphics/opengl/wrap_Graphics.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 1258 "modules/graphics/opengl/wrap_Graphics.cpp"
 FUNCTION_TABLE_vi[$26 & 8191]($i$0); //@line 1258 "modules/graphics/opengl/wrap_Graphics.cpp"
 STACKTOP = sp;return 1; //@line 1259 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love14luax_getmoduleINS_5image5ImageEEEPT_P9lua_StateNS_4TypeE($L,$type) {
 $L = $L|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $vararg_buffer = 0, $vararg_buffer1 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $name = sp + 12|0;
 HEAP32[$name>>2] = 442055; //@line 461 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 462 "./common/runtime.h"
 (__ZN4love19luax_insistregistryEP9lua_StateNS_8RegistryE($L,0)|0); //@line 464 "./common/runtime.h"
 $0 = HEAP32[$name>>2]|0; //@line 465 "./common/runtime.h"
 _lua_getfield($L,-1,$0); //@line 465 "./common/runtime.h"
 $1 = (_lua_isuserdata($L,-1)|0); //@line 467 "./common/runtime.h"
 $2 = ($1|0)==(0); //@line 467 "./common/runtime.h"
 if ($2) {
  $3 = HEAP32[$name>>2]|0; //@line 468 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $3; //@line 468 "./common/runtime.h"
  (_luaL_error($L,330432,$vararg_buffer)|0); //@line 468 "./common/runtime.h"
 }
 $4 = (_lua_touserdata($L,-1)|0); //@line 470 "./common/runtime.h"
 $5 = HEAP32[$4>>2]|0; //@line 472 "./common/runtime.h"
 $6 = HEAP32[156504]|0; //@line 472 "./common/runtime.h"
 $7 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = ((($6) + ($5<<3)|0) + ($7<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = 1 << $9; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $11 = HEAP32[$8>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = $11 & $10; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $13 = ($12|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($13)) {
  _lua_settop($L,-3); //@line 475 "./common/runtime.h"
  $15 = ((($4)) + 4|0); //@line 477 "./common/runtime.h"
  $16 = HEAP32[$15>>2]|0; //@line 477 "./common/runtime.h"
  STACKTOP = sp;return ($16|0); //@line 477 "./common/runtime.h"
 }
 $14 = HEAP32[$name>>2]|0; //@line 473 "./common/runtime.h"
 HEAP32[$vararg_buffer1>>2] = $14; //@line 473 "./common/runtime.h"
 (_luaL_error($L,330468,$vararg_buffer1)|0); //@line 473 "./common/runtime.h"
 _lua_settop($L,-3); //@line 475 "./common/runtime.h"
 $15 = ((($4)) + 4|0); //@line 477 "./common/runtime.h"
 $16 = HEAP32[$15>>2]|0; //@line 477 "./common/runtime.h"
 STACKTOP = sp;return ($16|0); //@line 477 "./common/runtime.h"
}
function __ZN4love8graphics6opengl11w_setCanvasEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $canvases = 0, $eh$lpad$body$index4Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i$0 = 0, $i1$0 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i3 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $canvases = sp + 12|0;
 $0 = sp + 8|0;
 $1 = sp + 4|0;
 $2 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZN4love8graphics6opengl8Graphics23stopDrawToStencilBufferEv($2); //@line 1265 "modules/graphics/opengl/wrap_Graphics.cpp"
 $3 = (_lua_type($L,1)|0); //@line 1268 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = ($3|0)<(1); //@line 1268 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($4) {
  $5 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  __ZN4love8graphics6opengl8Graphics9setCanvasEv($5); //@line 1270 "modules/graphics/opengl/wrap_Graphics.cpp"
  STACKTOP = sp;return 0; //@line 1300 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $6 = (_lua_type($L,1)|0); //@line 1274 "modules/graphics/opengl/wrap_Graphics.cpp"
 $7 = ($6|0)==(5); //@line 1274 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAP32[$canvases>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($canvases)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$8>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ((($canvases)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$9>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $10 = ((($canvases)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 L5: do {
  if ($7) {
   $i$0 = 1;
   while(1) {
    __THREW__ = 0;
    $11 = (invoke_iii(3381,($L|0),1)|0); //@line 1279 "modules/graphics/opengl/wrap_Graphics.cpp"
    $12 = __THREW__; __THREW__ = 0;
    $13 = $12&1;
    if ($13) {
     break;
    }
    $14 = ($i$0|0)>($11|0); //@line 1279 "modules/graphics/opengl/wrap_Graphics.cpp"
    if ($14) {
     label = 28;
     break L5;
    }
    __THREW__ = 0;
    invoke_viii(3382,($L|0),1,($i$0|0)); //@line 1281 "modules/graphics/opengl/wrap_Graphics.cpp"
    $30 = __THREW__; __THREW__ = 0;
    $31 = $30&1;
    if ($31) {
     break;
    }
    __THREW__ = 0;
    $32 = (invoke_iii(3573,($L|0),-1)|0); //@line 1282 "modules/graphics/opengl/wrap_Graphics.cpp"
    $33 = __THREW__; __THREW__ = 0;
    $34 = $33&1;
    if ($34) {
     break;
    }
    HEAP32[$0>>2] = $32; //@line 1282 "modules/graphics/opengl/wrap_Graphics.cpp"
    $35 = HEAP32[$8>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $36 = HEAP32[$10>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $37 = ($35>>>0)<($36>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($37) {
     $38 = $32; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$35>>2] = $38; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $39 = HEAP32[$8>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $40 = ((($39)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$8>>2] = $40; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    } else {
     __THREW__ = 0;
     invoke_vii(3574,($canvases|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $41 = __THREW__; __THREW__ = 0;
     $42 = $41&1;
     if ($42) {
      break;
     }
    }
    __THREW__ = 0;
    invoke_vii(3384,($L|0),-2); //@line 1283 "modules/graphics/opengl/wrap_Graphics.cpp"
    $43 = __THREW__; __THREW__ = 0;
    $44 = $43&1;
    if ($44) {
     break;
    }
    $45 = (($i$0) + 1)|0; //@line 1279 "modules/graphics/opengl/wrap_Graphics.cpp"
    $i$0 = $45;
   }
   $15 = ___cxa_find_matching_catch_2()|0;
   $16 = tempRet0;
   $eh$lpad$body$index4Z2D = $16;$eh$lpad$body$indexZ2D = $15;
  } else {
   $i1$0 = 1;
   while(1) {
    __THREW__ = 0;
    $46 = (invoke_ii(3385,($L|0))|0); //@line 1288 "modules/graphics/opengl/wrap_Graphics.cpp"
    $47 = __THREW__; __THREW__ = 0;
    $48 = $47&1;
    if ($48) {
     break;
    }
    $49 = ($i1$0|0)>($46|0); //@line 1288 "modules/graphics/opengl/wrap_Graphics.cpp"
    if ($49) {
     label = 28;
     break L5;
    }
    __THREW__ = 0;
    $50 = (invoke_iii(3573,($L|0),($i1$0|0))|0); //@line 1289 "modules/graphics/opengl/wrap_Graphics.cpp"
    $51 = __THREW__; __THREW__ = 0;
    $52 = $51&1;
    if ($52) {
     break;
    }
    HEAP32[$1>>2] = $50; //@line 1289 "modules/graphics/opengl/wrap_Graphics.cpp"
    $53 = HEAP32[$8>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $54 = HEAP32[$10>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $55 = ($53>>>0)<($54>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($55) {
     $56 = $50; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$53>>2] = $56; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $57 = HEAP32[$8>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $58 = ((($57)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$8>>2] = $58; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    } else {
     __THREW__ = 0;
     invoke_vii(3574,($canvases|0),($1|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $59 = __THREW__; __THREW__ = 0;
     $60 = $59&1;
     if ($60) {
      break;
     }
    }
    $61 = (($i1$0) + 1)|0; //@line 1288 "modules/graphics/opengl/wrap_Graphics.cpp"
    $i1$0 = $61;
   }
   $17 = ___cxa_find_matching_catch_2()|0;
   $18 = tempRet0;
   $eh$lpad$body$index4Z2D = $18;$eh$lpad$body$indexZ2D = $17;
  }
 } while(0);
 L28: do {
  if ((label|0) == 28) {
   $62 = HEAP32[$8>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $63 = HEAP32[$canvases>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $64 = ($62|0)==($63|0); //@line 1293 "modules/graphics/opengl/wrap_Graphics.cpp"
   $65 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
   if ($64) {
    __THREW__ = 0;
    invoke_vi(3576,($65|0)); //@line 1296 "modules/graphics/opengl/wrap_Graphics.cpp"
    $68 = __THREW__; __THREW__ = 0;
    $69 = $68&1;
    if ($69) {
     label = 31;
    }
   } else {
    __THREW__ = 0;
    invoke_vii(3575,($65|0),($canvases|0)); //@line 1294 "modules/graphics/opengl/wrap_Graphics.cpp"
    $66 = __THREW__; __THREW__ = 0;
    $67 = $66&1;
    if ($67) {
     label = 31;
    }
   }
   L33: do {
    if ((label|0) == 31) {
     $70 = ___cxa_find_matching_catch_3(8992|0)|0;
     $71 = tempRet0;
     $72 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
     $73 = ($71|0)==($72|0); //@line 535 "./common/runtime.h"
     do {
      if ($73) {
       $74 = (___cxa_begin_catch(($70|0))|0); //@line 535 "./common/runtime.h"
       $75 = HEAP32[$74>>2]|0; //@line 539 "./common/runtime.h"
       $76 = ((($75)) + 8|0); //@line 539 "./common/runtime.h"
       $77 = HEAP32[$76>>2]|0; //@line 539 "./common/runtime.h"
       $78 = (FUNCTION_TABLE_ii[$77 & 8191]($74)|0); //@line 539 "./common/runtime.h"
       __THREW__ = 0;
       invoke_vii(3380,($L|0),($78|0)); //@line 539 "./common/runtime.h"
       $79 = __THREW__; __THREW__ = 0;
       $80 = $79&1;
       if ($80) {
        $88 = ___cxa_find_matching_catch_2()|0;
        $89 = tempRet0;
        __THREW__ = 0;
        invoke_v(3378); //@line 540 "./common/runtime.h"
        $90 = __THREW__; __THREW__ = 0;
        $91 = $90&1;
        if (!($91)) {
         $$01$i = $89;$$02$i = $88;
         break;
        }
        $92 = ___cxa_find_matching_catch_3(0|0)|0;
        $93 = tempRet0;
        ___clang_call_terminate($92); //@line 540 "./common/runtime.h"
        // unreachable; //@line 540 "./common/runtime.h"
       }
       __THREW__ = 0;
       invoke_v(3378); //@line 540 "./common/runtime.h"
       $81 = __THREW__; __THREW__ = 0;
       $82 = $81&1;
       if (!($82)) {
        __THREW__ = 0;
        $83 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
        $84 = __THREW__; __THREW__ = 0;
        $85 = $84&1;
        if (!($85)) {
         __THREW__ = 0;
         HEAP32[$vararg_buffer>>2] = $83; //@line 543 "./common/runtime.h"
         (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 543 "./common/runtime.h"
         $86 = __THREW__; __THREW__ = 0;
         $87 = $86&1;
         if (!($87)) {
          break L33;
         }
        }
       }
       $19 = ___cxa_find_matching_catch_2()|0;
       $20 = tempRet0;
       $eh$lpad$body$index4Z2D = $20;$eh$lpad$body$indexZ2D = $19;
       break L28;
      } else {
       $$01$i = $71;$$02$i = $70;
      }
     } while(0);
     $eh$lpad$body$index4Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
     break L28;
    }
   } while(0);
   $94 = HEAP32[$canvases>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $95 = ($94|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $96 = $94; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($95)) {
    $97 = HEAP32[$8>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $98 = ($97|0)==($94|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($98)) {
     $scevgep$i$i$i = ((($97)) + -4|0);
     $99 = $scevgep$i$i$i;
     $100 = (($99) - ($96))|0;
     $101 = $100 >>> 2;
     $102 = $101 ^ -1;
     $scevgep4$i$i$i = (($97) + ($102<<2)|0);
     HEAP32[$8>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($94); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   STACKTOP = sp;return 0; //@line 1300 "modules/graphics/opengl/wrap_Graphics.cpp"
  }
 } while(0);
 $21 = HEAP32[$canvases>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $22 = ($21|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $23 = $21; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($22) {
  ___resumeException($eh$lpad$body$indexZ2D|0);
  // unreachable;
 }
 $24 = HEAP32[$8>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $25 = ($24|0)==($21|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($25)) {
  $scevgep$i$i$i3 = ((($24)) + -4|0);
  $26 = $scevgep$i$i$i3;
  $27 = (($26) - ($23))|0;
  $28 = $27 >>> 2;
  $29 = $28 ^ -1;
  $scevgep4$i$i$i4 = (($24) + ($29<<2)|0);
  HEAP32[$8>>2] = $scevgep4$i$i$i4; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($21); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__16vectorIPN4love8graphics6opengl6CanvasENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl11w_getCanvasEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$07 = 0, $canvases = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $n$08 = 0;
 var $n$1 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i1 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $canvases = sp;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZNK4love8graphics6opengl8Graphics9getCanvasEv($canvases,$0); //@line 1304 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = HEAP32[$canvases>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($canvases)) + 4|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($1|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 L1: do {
  if ($4) {
   __THREW__ = 0;
   invoke_vi(3577,($L|0)); //@line 1315 "modules/graphics/opengl/wrap_Graphics.cpp"
   $24 = __THREW__; __THREW__ = 0;
   $25 = $24&1;
   if ($25) {
    $13 = ___cxa_find_matching_catch_2()|0;
    $14 = tempRet0;
    $lpad$phi$index = $13;$lpad$phi$index2 = $14;
    label = 6;
   } else {
    $n$1 = 1;
    label = 12;
   }
  } else {
   $__begin$sroa$0$07 = $1;$n$08 = 0;
   while(1) {
    $5 = HEAP32[$__begin$sroa$0$07>>2]|0; //@line 1307 "modules/graphics/opengl/wrap_Graphics.cpp"
    __THREW__ = 0;
    invoke_viii(3387,($L|0),17,($5|0)); //@line 1309 "modules/graphics/opengl/wrap_Graphics.cpp"
    $6 = __THREW__; __THREW__ = 0;
    $7 = $6&1;
    if ($7) {
     break;
    }
    $8 = (($n$08) + 1)|0; //@line 1310 "modules/graphics/opengl/wrap_Graphics.cpp"
    $9 = ((($__begin$sroa$0$07)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $10 = ($9|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($10) {
     $n$1 = $8;
     label = 12;
     break L1;
    } else {
     $__begin$sroa$0$07 = $9;$n$08 = $8;
    }
   }
   $11 = ___cxa_find_matching_catch_2()|0;
   $12 = tempRet0;
   $lpad$phi$index = $11;$lpad$phi$index2 = $12;
   label = 6;
  }
 } while(0);
 if ((label|0) == 6) {
  $15 = HEAP32[$canvases>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = $15; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($16) {
   ___resumeException($lpad$phi$index|0);
   // unreachable;
  }
  $18 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = ($18|0)==($15|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($19)) {
   $scevgep$i$i$i1 = ((($18)) + -4|0);
   $20 = $scevgep$i$i$i1;
   $21 = (($20) - ($17))|0;
   $22 = $21 >>> 2;
   $23 = $22 ^ -1;
   $scevgep4$i$i$i2 = (($18) + ($23<<2)|0);
   HEAP32[$2>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($15); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 else if ((label|0) == 12) {
  $26 = HEAP32[$canvases>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $27 = ($26|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $28 = $26; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($27) {
   STACKTOP = sp;return ($n$1|0); //@line 1320 "modules/graphics/opengl/wrap_Graphics.cpp"
  }
  $29 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $30 = ($29|0)==($26|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($30)) {
   $scevgep$i$i$i = ((($29)) + -4|0);
   $31 = $scevgep$i$i$i;
   $32 = (($31) - ($28))|0;
   $33 = $32 >>> 2;
   $34 = $33 ^ -1;
   $scevgep4$i$i$i = (($29) + ($34<<2)|0);
   HEAP32[$2>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($26); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  STACKTOP = sp;return ($n$1|0); //@line 1320 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl11w_setShaderEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_type($L,1)|0); //@line 1324 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = ($0|0)<(1); //@line 1324 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($1) {
  $2 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  __ZN4love8graphics6opengl8Graphics9setShaderEv($2); //@line 1326 "modules/graphics/opengl/wrap_Graphics.cpp"
  return 0; //@line 1333 "modules/graphics/opengl/wrap_Graphics.cpp"
 } else {
  $3 = (__ZN4love8graphics6opengl16luax_checkshaderEP9lua_Statei($L,1)|0); //@line 1330 "modules/graphics/opengl/wrap_Graphics.cpp"
  $4 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  __ZN4love8graphics6opengl8Graphics9setShaderEPNS1_6ShaderE($4,$3); //@line 1331 "modules/graphics/opengl/wrap_Graphics.cpp"
  return 0; //@line 1333 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl11w_getShaderEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = (__ZNK4love8graphics6opengl8Graphics9getShaderEv($0)|0); //@line 1337 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = ($1|0)==(0|0); //@line 1338 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($2) {
  _lua_pushnil($L); //@line 1341 "modules/graphics/opengl/wrap_Graphics.cpp"
  return 1; //@line 1343 "modules/graphics/opengl/wrap_Graphics.cpp"
 } else {
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,18,$1); //@line 1339 "modules/graphics/opengl/wrap_Graphics.cpp"
  return 1; //@line 1343 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl22w_setDefaultShaderCodeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$pre = 0, $$pre$phi29Z2D = 0, $$pre$phi31Z2D = 0, $$pre$phi33Z2D = 0, $$pre$phi35Z2D = 0, $$pre$phi37Z2D = 0, $$pre$phiZ2D = 0, $$pre30 = 0, $$pre34 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $openglVideocode = 0, $openglcode = 0, $openglesVideocode = 0, $openglescode = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $openglcode = sp + 168|0;
 $0 = sp + 156|0;
 $1 = sp + 144|0;
 $openglVideocode = sp + 120|0;
 $2 = sp + 108|0;
 $3 = sp + 96|0;
 $openglescode = sp + 72|0;
 $4 = sp + 60|0;
 $5 = sp + 48|0;
 $openglesVideocode = sp + 24|0;
 $6 = sp + 12|0;
 $7 = sp;
 _luaL_checktype($L,1,5); //@line 1348 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_getfield($L,1,330488); //@line 1350 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_rawgeti($L,-1,1); //@line 1351 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_rawgeti($L,-2,2); //@line 1352 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_rawgeti($L,-3,3); //@line 1353 "modules/graphics/opengl/wrap_Graphics.cpp"
 ;HEAP32[$openglcode>>2]=0|0;HEAP32[$openglcode+4>>2]=0|0;HEAP32[$openglcode+8>>2]=0|0;HEAP32[$openglcode+12>>2]=0|0;HEAP32[$openglcode+16>>2]=0|0;HEAP32[$openglcode+20>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viii(3403,($0|0),($L|0),-3); //@line 1356 "modules/graphics/opengl/wrap_Graphics.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if (!($9)) {
  $10 = HEAP8[$openglcode>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $11 = $10 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $12 = ($11<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($12) {
   $16 = ((($openglcode)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP8[$16>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP8[$openglcode>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  } else {
   $13 = ((($openglcode)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $14 = HEAP32[$13>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP8[$14>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $15 = ((($openglcode)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP32[$15>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  }
  __THREW__ = 0;
  invoke_vii(3368,($openglcode|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $17 = __THREW__; __THREW__ = 0;
  $18 = $17&1;
  if ($18) {
   $19 = ___cxa_find_matching_catch_3(0|0)|0;
   $20 = tempRet0;
   ___clang_call_terminate($19); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  }
  ;HEAP32[$openglcode>>2]=HEAP32[$0>>2]|0;HEAP32[$openglcode+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$openglcode+8>>2]=HEAP32[$0+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 1356 "modules/graphics/opengl/wrap_Graphics.cpp"
  __THREW__ = 0;
  invoke_viii(3403,($1|0),($L|0),-2); //@line 1357 "modules/graphics/opengl/wrap_Graphics.cpp"
  $21 = __THREW__; __THREW__ = 0;
  $22 = $21&1;
  if (!($22)) {
   $23 = ((($openglcode)) + 12|0); //@line 1357 "modules/graphics/opengl/wrap_Graphics.cpp"
   $24 = HEAP8[$23>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $25 = $24 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $26 = ($25<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if ($26) {
    $30 = ((($23)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$30>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$23>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   } else {
    $27 = ((($openglcode)) + 20|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $28 = HEAP32[$27>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$28>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $29 = ((($openglcode)) + 16|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP32[$29>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   }
   __THREW__ = 0;
   invoke_vii(3368,($23|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $31 = __THREW__; __THREW__ = 0;
   $32 = $31&1;
   if ($32) {
    $33 = ___cxa_find_matching_catch_3(0|0)|0;
    $34 = tempRet0;
    ___clang_call_terminate($33); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   }
   ;HEAP32[$23>>2]=HEAP32[$1>>2]|0;HEAP32[$23+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$23+8>>2]=HEAP32[$1+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 1357 "modules/graphics/opengl/wrap_Graphics.cpp"
   ;HEAP32[$openglVideocode>>2]=0|0;HEAP32[$openglVideocode+4>>2]=0|0;HEAP32[$openglVideocode+8>>2]=0|0;HEAP32[$openglVideocode+12>>2]=0|0;HEAP32[$openglVideocode+16>>2]=0|0;HEAP32[$openglVideocode+20>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   invoke_viii(3403,($2|0),($L|0),-3); //@line 1360 "modules/graphics/opengl/wrap_Graphics.cpp"
   $35 = __THREW__; __THREW__ = 0;
   $36 = $35&1;
   if ($36) {
    label = 65;
   } else {
    $37 = HEAP8[$openglVideocode>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $38 = $37 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $39 = ($38<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if ($39) {
     $43 = ((($openglVideocode)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP8[$43>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP8[$openglVideocode>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    } else {
     $40 = ((($openglVideocode)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $41 = HEAP32[$40>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP8[$41>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $42 = ((($openglVideocode)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP32[$42>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    }
    __THREW__ = 0;
    invoke_vii(3368,($openglVideocode|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $44 = __THREW__; __THREW__ = 0;
    $45 = $44&1;
    if ($45) {
     $46 = ___cxa_find_matching_catch_3(0|0)|0;
     $47 = tempRet0;
     ___clang_call_terminate($46); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    }
    ;HEAP32[$openglVideocode>>2]=HEAP32[$2>>2]|0;HEAP32[$openglVideocode+4>>2]=HEAP32[$2+4>>2]|0;HEAP32[$openglVideocode+8>>2]=HEAP32[$2+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    ;HEAP32[$2>>2]=0|0;HEAP32[$2+4>>2]=0|0;HEAP32[$2+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 1360 "modules/graphics/opengl/wrap_Graphics.cpp"
    __THREW__ = 0;
    invoke_viii(3403,($3|0),($L|0),-1); //@line 1361 "modules/graphics/opengl/wrap_Graphics.cpp"
    $48 = __THREW__; __THREW__ = 0;
    $49 = $48&1;
    if ($49) {
     label = 65;
    } else {
     $50 = ((($openglVideocode)) + 12|0); //@line 1361 "modules/graphics/opengl/wrap_Graphics.cpp"
     $51 = HEAP8[$50>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $52 = $51 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $53 = ($52<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($53) {
      $57 = ((($50)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$57>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$50>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     } else {
      $54 = ((($openglVideocode)) + 20|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $55 = HEAP32[$54>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$55>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $56 = ((($openglVideocode)) + 16|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP32[$56>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     }
     __THREW__ = 0;
     invoke_vii(3368,($50|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $58 = __THREW__; __THREW__ = 0;
     $59 = $58&1;
     if ($59) {
      $60 = ___cxa_find_matching_catch_3(0|0)|0;
      $61 = tempRet0;
      ___clang_call_terminate($60); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     }
     ;HEAP32[$50>>2]=HEAP32[$3>>2]|0;HEAP32[$50+4>>2]=HEAP32[$3+4>>2]|0;HEAP32[$50+8>>2]=HEAP32[$3+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     ;HEAP32[$3>>2]=0|0;HEAP32[$3+4>>2]=0|0;HEAP32[$3+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 1361 "modules/graphics/opengl/wrap_Graphics.cpp"
     __THREW__ = 0;
     invoke_vii(3384,($L|0),-5); //@line 1363 "modules/graphics/opengl/wrap_Graphics.cpp"
     $62 = __THREW__; __THREW__ = 0;
     $63 = $62&1;
     if ($63) {
      label = 65;
     } else {
      __THREW__ = 0;
      invoke_viii(3471,($L|0),1,(330495|0)); //@line 1365 "modules/graphics/opengl/wrap_Graphics.cpp"
      $64 = __THREW__; __THREW__ = 0;
      $65 = $64&1;
      if ($65) {
       label = 65;
      } else {
       __THREW__ = 0;
       invoke_viii(3382,($L|0),-1,1); //@line 1366 "modules/graphics/opengl/wrap_Graphics.cpp"
       $66 = __THREW__; __THREW__ = 0;
       $67 = $66&1;
       if ($67) {
        label = 65;
       } else {
        __THREW__ = 0;
        invoke_viii(3382,($L|0),-2,2); //@line 1367 "modules/graphics/opengl/wrap_Graphics.cpp"
        $68 = __THREW__; __THREW__ = 0;
        $69 = $68&1;
        if ($69) {
         label = 65;
        } else {
         __THREW__ = 0;
         invoke_viii(3382,($L|0),-3,3); //@line 1368 "modules/graphics/opengl/wrap_Graphics.cpp"
         $70 = __THREW__; __THREW__ = 0;
         $71 = $70&1;
         if ($71) {
          label = 65;
         } else {
          ;HEAP32[$openglescode>>2]=0|0;HEAP32[$openglescode+4>>2]=0|0;HEAP32[$openglescode+8>>2]=0|0;HEAP32[$openglescode+12>>2]=0|0;HEAP32[$openglescode+16>>2]=0|0;HEAP32[$openglescode+20>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          __THREW__ = 0;
          invoke_viii(3403,($4|0),($L|0),-3); //@line 1371 "modules/graphics/opengl/wrap_Graphics.cpp"
          $72 = __THREW__; __THREW__ = 0;
          $73 = $72&1;
          do {
           if ($73) {
            label = 66;
           } else {
            $74 = HEAP8[$openglescode>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $75 = $74 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $76 = ($75<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            if ($76) {
             $80 = ((($openglescode)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             HEAP8[$80>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             HEAP8[$openglescode>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            } else {
             $77 = ((($openglescode)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $78 = HEAP32[$77>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             HEAP8[$78>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $79 = ((($openglescode)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             HEAP32[$79>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            }
            __THREW__ = 0;
            invoke_vii(3368,($openglescode|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $81 = __THREW__; __THREW__ = 0;
            $82 = $81&1;
            if ($82) {
             $83 = ___cxa_find_matching_catch_3(0|0)|0;
             $84 = tempRet0;
             ___clang_call_terminate($83); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            }
            ;HEAP32[$openglescode>>2]=HEAP32[$4>>2]|0;HEAP32[$openglescode+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$openglescode+8>>2]=HEAP32[$4+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 1371 "modules/graphics/opengl/wrap_Graphics.cpp"
            __THREW__ = 0;
            invoke_viii(3403,($5|0),($L|0),-2); //@line 1372 "modules/graphics/opengl/wrap_Graphics.cpp"
            $85 = __THREW__; __THREW__ = 0;
            $86 = $85&1;
            if ($86) {
             label = 66;
             break;
            }
            $87 = ((($openglescode)) + 12|0); //@line 1372 "modules/graphics/opengl/wrap_Graphics.cpp"
            $88 = HEAP8[$87>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $89 = $88 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $90 = ($89<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            if ($90) {
             $94 = ((($87)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             HEAP8[$94>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             HEAP8[$87>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            } else {
             $91 = ((($openglescode)) + 20|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $92 = HEAP32[$91>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             HEAP8[$92>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $93 = ((($openglescode)) + 16|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             HEAP32[$93>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            }
            __THREW__ = 0;
            invoke_vii(3368,($87|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $95 = __THREW__; __THREW__ = 0;
            $96 = $95&1;
            if ($96) {
             $97 = ___cxa_find_matching_catch_3(0|0)|0;
             $98 = tempRet0;
             ___clang_call_terminate($97); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            }
            ;HEAP32[$87>>2]=HEAP32[$5>>2]|0;HEAP32[$87+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$87+8>>2]=HEAP32[$5+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5); //@line 1372 "modules/graphics/opengl/wrap_Graphics.cpp"
            ;HEAP32[$openglesVideocode>>2]=0|0;HEAP32[$openglesVideocode+4>>2]=0|0;HEAP32[$openglesVideocode+8>>2]=0|0;HEAP32[$openglesVideocode+12>>2]=0|0;HEAP32[$openglesVideocode+16>>2]=0|0;HEAP32[$openglesVideocode+20>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            __THREW__ = 0;
            invoke_viii(3403,($6|0),($L|0),-3); //@line 1375 "modules/graphics/opengl/wrap_Graphics.cpp"
            $99 = __THREW__; __THREW__ = 0;
            $100 = $99&1;
            do {
             if (!($100)) {
              $101 = HEAP8[$openglesVideocode>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              $102 = $101 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              $103 = ($102<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              if ($103) {
               $107 = ((($openglesVideocode)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               HEAP8[$107>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               HEAP8[$openglesVideocode>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              } else {
               $104 = ((($openglesVideocode)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $105 = HEAP32[$104>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               HEAP8[$105>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $106 = ((($openglesVideocode)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               HEAP32[$106>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              }
              __THREW__ = 0;
              invoke_vii(3368,($openglesVideocode|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              $108 = __THREW__; __THREW__ = 0;
              $109 = $108&1;
              if ($109) {
               $110 = ___cxa_find_matching_catch_3(0|0)|0;
               $111 = tempRet0;
               ___clang_call_terminate($110); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              }
              ;HEAP32[$openglesVideocode>>2]=HEAP32[$6>>2]|0;HEAP32[$openglesVideocode+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$openglesVideocode+8>>2]=HEAP32[$6+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 1375 "modules/graphics/opengl/wrap_Graphics.cpp"
              __THREW__ = 0;
              invoke_viii(3403,($7|0),($L|0),-1); //@line 1376 "modules/graphics/opengl/wrap_Graphics.cpp"
              $112 = __THREW__; __THREW__ = 0;
              $113 = $112&1;
              if ($113) {
               break;
              }
              $114 = ((($openglesVideocode)) + 12|0); //@line 1376 "modules/graphics/opengl/wrap_Graphics.cpp"
              $115 = HEAP8[$114>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              $116 = $115 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              $117 = ($116<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              if ($117) {
               $121 = ((($114)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               HEAP8[$121>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               HEAP8[$114>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              } else {
               $118 = ((($openglesVideocode)) + 20|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $119 = HEAP32[$118>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               HEAP8[$119>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $120 = ((($openglesVideocode)) + 16|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               HEAP32[$120>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              }
              __THREW__ = 0;
              invoke_vii(3368,($114|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              $122 = __THREW__; __THREW__ = 0;
              $123 = $122&1;
              if ($123) {
               $124 = ___cxa_find_matching_catch_3(0|0)|0;
               $125 = tempRet0;
               ___clang_call_terminate($124); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              }
              ;HEAP32[$114>>2]=HEAP32[$7>>2]|0;HEAP32[$114+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$114+8>>2]=HEAP32[$7+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7); //@line 1376 "modules/graphics/opengl/wrap_Graphics.cpp"
              __THREW__ = 0;
              invoke_vii(3384,($L|0),-5); //@line 1378 "modules/graphics/opengl/wrap_Graphics.cpp"
              $126 = __THREW__; __THREW__ = 0;
              $127 = $126&1;
              if ($127) {
               break;
              }
              __THREW__ = 0;
              (invoke_iii(3578,(634336|0),($openglcode|0))|0); //@line 84 "modules/graphics/opengl/Shader.h"
              $128 = __THREW__; __THREW__ = 0;
              $129 = $128&1;
              if ($129) {
               break;
              }
              __THREW__ = 0;
              (invoke_iii(3578,((634348)|0),($23|0))|0); //@line 84 "modules/graphics/opengl/Shader.h"
              $130 = __THREW__; __THREW__ = 0;
              $131 = $130&1;
              if ($131) {
               break;
              }
              __THREW__ = 0;
              (invoke_iii(3578,((634360)|0),($openglescode|0))|0); //@line 84 "modules/graphics/opengl/Shader.h"
              $132 = __THREW__; __THREW__ = 0;
              $133 = $132&1;
              if ($133) {
               break;
              }
              __THREW__ = 0;
              (invoke_iii(3578,((634372)|0),($87|0))|0); //@line 84 "modules/graphics/opengl/Shader.h"
              $134 = __THREW__; __THREW__ = 0;
              $135 = $134&1;
              if ($135) {
               break;
              }
              __THREW__ = 0;
              (invoke_iii(3578,(634384|0),($openglVideocode|0))|0); //@line 84 "modules/graphics/opengl/Shader.h"
              $136 = __THREW__; __THREW__ = 0;
              $137 = $136&1;
              if ($137) {
               break;
              }
              __THREW__ = 0;
              (invoke_iii(3578,((634396)|0),($50|0))|0); //@line 84 "modules/graphics/opengl/Shader.h"
              $138 = __THREW__; __THREW__ = 0;
              $139 = $138&1;
              if ($139) {
               break;
              }
              __THREW__ = 0;
              (invoke_iii(3578,((634408)|0),($openglesVideocode|0))|0); //@line 84 "modules/graphics/opengl/Shader.h"
              $140 = __THREW__; __THREW__ = 0;
              $141 = $140&1;
              if ($141) {
               break;
              }
              __THREW__ = 0;
              (invoke_iii(3578,((634420)|0),($114|0))|0); //@line 84 "modules/graphics/opengl/Shader.h"
              $142 = __THREW__; __THREW__ = 0;
              $143 = $142&1;
              if ($143) {
               break;
              }
              __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($114); //@line 84 "modules/graphics/opengl/Shader.h"
              __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($openglesVideocode); //@line 84 "modules/graphics/opengl/Shader.h"
              __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($87); //@line 84 "modules/graphics/opengl/Shader.h"
              __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($openglescode); //@line 84 "modules/graphics/opengl/Shader.h"
              __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($50); //@line 84 "modules/graphics/opengl/Shader.h"
              __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($openglVideocode); //@line 84 "modules/graphics/opengl/Shader.h"
              __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($23); //@line 84 "modules/graphics/opengl/Shader.h"
              __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($openglcode); //@line 84 "modules/graphics/opengl/Shader.h"
              STACKTOP = sp;return 0; //@line 1386 "modules/graphics/opengl/wrap_Graphics.cpp"
             }
            } while(0);
            $150 = ___cxa_find_matching_catch_2()|0;
            $151 = tempRet0;
            $152 = ((($openglesVideocode)) + 12|0); //@line 84 "modules/graphics/opengl/Shader.h"
            __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($152); //@line 84 "modules/graphics/opengl/Shader.h"
            __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($openglesVideocode); //@line 84 "modules/graphics/opengl/Shader.h"
            $$0 = $151;$$01 = $150;$$pre$phi29Z2D = $openglescode;$$pre$phiZ2D = $87;
           }
          } while(0);
          if ((label|0) == 66) {
           $148 = ___cxa_find_matching_catch_2()|0;
           $149 = tempRet0;
           $$pre = ((($openglescode)) + 12|0); //@line 84 "modules/graphics/opengl/Shader.h"
           $$0 = $149;$$01 = $148;$$pre$phi29Z2D = $openglescode;$$pre$phiZ2D = $$pre;
          }
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre$phiZ2D); //@line 84 "modules/graphics/opengl/Shader.h"
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre$phi29Z2D); //@line 84 "modules/graphics/opengl/Shader.h"
          $$1 = $$0;$$12 = $$01;$$pre$phi31Z2D = $50;$$pre$phi33Z2D = $openglVideocode;
         }
        }
       }
      }
     }
    }
   }
   if ((label|0) == 65) {
    $146 = ___cxa_find_matching_catch_2()|0;
    $147 = tempRet0;
    $$pre30 = ((($openglVideocode)) + 12|0); //@line 84 "modules/graphics/opengl/Shader.h"
    $$1 = $147;$$12 = $146;$$pre$phi31Z2D = $$pre30;$$pre$phi33Z2D = $openglVideocode;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre$phi31Z2D); //@line 84 "modules/graphics/opengl/Shader.h"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre$phi33Z2D); //@line 84 "modules/graphics/opengl/Shader.h"
   $$2 = $$1;$$23 = $$12;$$pre$phi35Z2D = $23;$$pre$phi37Z2D = $openglcode;
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre$phi35Z2D); //@line 84 "modules/graphics/opengl/Shader.h"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre$phi37Z2D); //@line 84 "modules/graphics/opengl/Shader.h"
   ___resumeException($$23|0);
   // unreachable;
  }
 }
 $144 = ___cxa_find_matching_catch_2()|0;
 $145 = tempRet0;
 $$pre34 = ((($openglcode)) + 12|0); //@line 84 "modules/graphics/opengl/Shader.h"
 $$2 = $145;$$23 = $144;$$pre$phi35Z2D = $$pre34;$$pre$phi37Z2D = $openglcode;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre$phi35Z2D); //@line 84 "modules/graphics/opengl/Shader.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre$phi37Z2D); //@line 84 "modules/graphics/opengl/Shader.h"
 ___resumeException($$23|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8graphics6opengl14w_getSupportedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp;
 _lua_createtable($L,0,3); //@line 1390 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAP32[$name>>2] = 0; //@line 1395 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = (__ZN4love8graphics8Graphics11getConstantENS1_7SupportERPKc(0,$name)|0); //@line 1397 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($0) {
  $1 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  $2 = (__ZNK4love8graphics6opengl8Graphics11isSupportedENS0_8Graphics7SupportE($1,0)|0); //@line 1400 "modules/graphics/opengl/wrap_Graphics.cpp"
  __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$2); //@line 1400 "modules/graphics/opengl/wrap_Graphics.cpp"
  $3 = HEAP32[$name>>2]|0; //@line 1401 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_setfield($L,-2,$3); //@line 1401 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 HEAP32[$name>>2] = 0; //@line 1395 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = (__ZN4love8graphics8Graphics11getConstantENS1_7SupportERPKc(1,$name)|0); //@line 1397 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($4) {
  $5 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  $6 = (__ZNK4love8graphics6opengl8Graphics11isSupportedENS0_8Graphics7SupportE($5,1)|0); //@line 1400 "modules/graphics/opengl/wrap_Graphics.cpp"
  __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$6); //@line 1400 "modules/graphics/opengl/wrap_Graphics.cpp"
  $7 = HEAP32[$name>>2]|0; //@line 1401 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_setfield($L,-2,$7); //@line 1401 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 HEAP32[$name>>2] = 0; //@line 1395 "modules/graphics/opengl/wrap_Graphics.cpp"
 $8 = (__ZN4love8graphics8Graphics11getConstantENS1_7SupportERPKc(2,$name)|0); //@line 1397 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($8)) {
  STACKTOP = sp;return 1; //@line 1404 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $9 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $10 = (__ZNK4love8graphics6opengl8Graphics11isSupportedENS0_8Graphics7SupportE($9,2)|0); //@line 1400 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$10); //@line 1400 "modules/graphics/opengl/wrap_Graphics.cpp"
 $11 = HEAP32[$name>>2]|0; //@line 1401 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_setfield($L,-2,$11); //@line 1401 "modules/graphics/opengl/wrap_Graphics.cpp"
 STACKTOP = sp;return 1; //@line 1404 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl18w_getCanvasFormatsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $exitcond = 0, $i$01 = 0, $name = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp;
 _lua_createtable($L,0,17); //@line 1409 "modules/graphics/opengl/wrap_Graphics.cpp"
 $i$01 = 0;
 while(1) {
  HEAP32[$name>>2] = 0; //@line 1414 "modules/graphics/opengl/wrap_Graphics.cpp"
  $0 = (__ZN4love8graphics6opengl6Canvas11getConstantENS2_6FormatERPKc($i$01,$name)|0); //@line 1416 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($0) {
   $1 = (__ZN4love8graphics6opengl6Canvas17isFormatSupportedENS2_6FormatE($i$01)|0); //@line 1419 "modules/graphics/opengl/wrap_Graphics.cpp"
   __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$1); //@line 1419 "modules/graphics/opengl/wrap_Graphics.cpp"
   $2 = HEAP32[$name>>2]|0; //@line 1420 "modules/graphics/opengl/wrap_Graphics.cpp"
   _lua_setfield($L,-2,$2); //@line 1420 "modules/graphics/opengl/wrap_Graphics.cpp"
  }
  $3 = (($i$01) + 1)|0; //@line 1411 "modules/graphics/opengl/wrap_Graphics.cpp"
  $exitcond = ($3|0)==(17); //@line 1411 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($exitcond) {
   break;
  } else {
   $i$01 = $3;
  }
 }
 STACKTOP = sp;return 1; //@line 1423 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl27w_getCompressedImageFormatsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $exitcond = 0, $i$01 = 0, $name = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp;
 _lua_createtable($L,0,37); //@line 1428 "modules/graphics/opengl/wrap_Graphics.cpp"
 $i$01 = 0;
 while(1) {
  HEAP32[$name>>2] = 0; //@line 1433 "modules/graphics/opengl/wrap_Graphics.cpp"
  $0 = ($i$01|0)==(0); //@line 1435 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($0) {
   $i$01 = 1;
   continue;
  }
  $1 = (__ZN4love5image19CompressedImageData11getConstantENS1_6FormatERPKc($i$01,$name)|0); //@line 1438 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($1) {
   $2 = (__ZN4love8graphics6opengl5Image27hasCompressedTextureSupportENS_5image19CompressedImageData6FormatEb($i$01,0)|0); //@line 1441 "modules/graphics/opengl/wrap_Graphics.cpp"
   __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$2); //@line 1441 "modules/graphics/opengl/wrap_Graphics.cpp"
   $3 = HEAP32[$name>>2]|0; //@line 1442 "modules/graphics/opengl/wrap_Graphics.cpp"
   _lua_setfield($L,-2,$3); //@line 1442 "modules/graphics/opengl/wrap_Graphics.cpp"
  }
  $4 = (($i$01) + 1)|0; //@line 1430 "modules/graphics/opengl/wrap_Graphics.cpp"
  $exitcond = ($4|0)==(37); //@line 1430 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($exitcond) {
   break;
  } else {
   $i$01 = $4;
  }
 }
 STACKTOP = sp;return 1; //@line 1445 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl17w_getRendererInfoEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $info = 0, $vararg_buffer = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 56|0;
 $info = sp + 8|0;
 dest=$info; stop=dest+48|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $1 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 invoke_vii(3579,($0|0),($1|0)); //@line 1451 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 L1: do {
  if ($3) {
   $7 = ___cxa_find_matching_catch_3(8992|0)|0;
   $8 = tempRet0;
   $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
   do {
    if ($10) {
     $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
     $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
     $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
     $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
     $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
     __THREW__ = 0;
     invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
     $16 = __THREW__; __THREW__ = 0;
     $17 = $16&1;
     if ($17) {
      $25 = ___cxa_find_matching_catch_2()|0;
      $26 = tempRet0;
      __THREW__ = 0;
      invoke_v(3378); //@line 540 "./common/runtime.h"
      $27 = __THREW__; __THREW__ = 0;
      $28 = $27&1;
      if (!($28)) {
       $$01$i = $26;$$02$i = $25;
       break;
      }
      $29 = ___cxa_find_matching_catch_3(0|0)|0;
      $30 = tempRet0;
      ___clang_call_terminate($29); //@line 540 "./common/runtime.h"
      // unreachable; //@line 540 "./common/runtime.h"
     }
     __THREW__ = 0;
     invoke_v(3378); //@line 540 "./common/runtime.h"
     $18 = __THREW__; __THREW__ = 0;
     $19 = $18&1;
     if ($19) {
      break L1;
     }
     __THREW__ = 0;
     $20 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
     $21 = __THREW__; __THREW__ = 0;
     $22 = $21&1;
     if ($22) {
      break L1;
     }
     __THREW__ = 0;
     HEAP32[$vararg_buffer>>2] = $20; //@line 543 "./common/runtime.h"
     (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 543 "./common/runtime.h"
     $23 = __THREW__; __THREW__ = 0;
     $24 = $23&1;
     if ($24) {
      break L1;
     } else {
      label = 11;
      break L1;
     }
    } else {
     $$01$i = $8;$$02$i = $7;
    }
   } while(0);
   $eh$lpad$body$index2Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
   $44 = ((($info)) + 36|0); //@line 181 "./modules/graphics/Graphics.h"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($44); //@line 181 "./modules/graphics/Graphics.h"
   $45 = ((($info)) + 24|0); //@line 181 "./modules/graphics/Graphics.h"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($45); //@line 181 "./modules/graphics/Graphics.h"
   $46 = ((($info)) + 12|0); //@line 181 "./modules/graphics/Graphics.h"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($46); //@line 181 "./modules/graphics/Graphics.h"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($info); //@line 181 "./modules/graphics/Graphics.h"
   ___resumeException($eh$lpad$body$indexZ2D|0);
   // unreachable;
  } else {
   (__ZN4love8graphics8Graphics12RendererInfoaSEOS2_($info,$0)|0); //@line 1451 "modules/graphics/opengl/wrap_Graphics.cpp"
   $4 = ((($0)) + 36|0); //@line 181 "./modules/graphics/Graphics.h"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 181 "./modules/graphics/Graphics.h"
   $5 = ((($0)) + 24|0); //@line 181 "./modules/graphics/Graphics.h"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5); //@line 181 "./modules/graphics/Graphics.h"
   $6 = ((($0)) + 12|0); //@line 181 "./modules/graphics/Graphics.h"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 181 "./modules/graphics/Graphics.h"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 181 "./modules/graphics/Graphics.h"
   label = 11;
  }
 } while(0);
 if ((label|0) == 11) {
  __THREW__ = 0;
  invoke_vii(3405,($L|0),($info|0)); //@line 1453 "modules/graphics/opengl/wrap_Graphics.cpp"
  $31 = __THREW__; __THREW__ = 0;
  $32 = $31&1;
  if (!($32)) {
   $33 = ((($info)) + 12|0); //@line 1454 "modules/graphics/opengl/wrap_Graphics.cpp"
   __THREW__ = 0;
   invoke_vii(3405,($L|0),($33|0)); //@line 1454 "modules/graphics/opengl/wrap_Graphics.cpp"
   $34 = __THREW__; __THREW__ = 0;
   $35 = $34&1;
   if (!($35)) {
    $36 = ((($info)) + 24|0); //@line 1455 "modules/graphics/opengl/wrap_Graphics.cpp"
    __THREW__ = 0;
    invoke_vii(3405,($L|0),($36|0)); //@line 1455 "modules/graphics/opengl/wrap_Graphics.cpp"
    $37 = __THREW__; __THREW__ = 0;
    $38 = $37&1;
    if (!($38)) {
     $39 = ((($info)) + 36|0); //@line 1456 "modules/graphics/opengl/wrap_Graphics.cpp"
     __THREW__ = 0;
     invoke_vii(3405,($L|0),($39|0)); //@line 1456 "modules/graphics/opengl/wrap_Graphics.cpp"
     $40 = __THREW__; __THREW__ = 0;
     $41 = $40&1;
     if (!($41)) {
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($39); //@line 181 "./modules/graphics/Graphics.h"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($36); //@line 181 "./modules/graphics/Graphics.h"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($33); //@line 181 "./modules/graphics/Graphics.h"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($info); //@line 181 "./modules/graphics/Graphics.h"
      STACKTOP = sp;return 4; //@line 1458 "modules/graphics/opengl/wrap_Graphics.cpp"
     }
    }
   }
  }
 }
 $42 = ___cxa_find_matching_catch_2()|0;
 $43 = tempRet0;
 $eh$lpad$body$index2Z2D = $43;$eh$lpad$body$indexZ2D = $42;
 $44 = ((($info)) + 36|0); //@line 181 "./modules/graphics/Graphics.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($44); //@line 181 "./modules/graphics/Graphics.h"
 $45 = ((($info)) + 24|0); //@line 181 "./modules/graphics/Graphics.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($45); //@line 181 "./modules/graphics/Graphics.h"
 $46 = ((($info)) + 12|0); //@line 181 "./modules/graphics/Graphics.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($46); //@line 181 "./modules/graphics/Graphics.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($info); //@line 181 "./modules/graphics/Graphics.h"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8graphics8Graphics12RendererInfoaSEOS2_($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP8[$this>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = $1 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ($2<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($3) {
  $7 = ((($this)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$7>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$this>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 } else {
  $4 = ((($this)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $5 = HEAP32[$4>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$5>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $6 = ((($this)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$6>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 __THREW__ = 0;
 invoke_vii(3368,($this|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_3(0|0)|0;
  $11 = tempRet0;
  ___clang_call_terminate($10); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 ;HEAP32[$this>>2]=HEAP32[$0>>2]|0;HEAP32[$this+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$this+8>>2]=HEAP32[$0+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = ((($this)) + 12|0); //@line 181 "./modules/graphics/Graphics.h"
 $13 = ((($0)) + 12|0); //@line 181 "./modules/graphics/Graphics.h"
 $14 = HEAP8[$12>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $15 = $14 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $16 = ($15<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($16) {
  $20 = ((($12)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$20>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$12>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 } else {
  $17 = ((($this)) + 20|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $18 = HEAP32[$17>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$18>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $19 = ((($this)) + 16|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$19>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 __THREW__ = 0;
 invoke_vii(3368,($12|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $23 = ___cxa_find_matching_catch_3(0|0)|0;
  $24 = tempRet0;
  ___clang_call_terminate($23); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 ;HEAP32[$12>>2]=HEAP32[$13>>2]|0;HEAP32[$12+4>>2]=HEAP32[$13+4>>2]|0;HEAP32[$12+8>>2]=HEAP32[$13+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 ;HEAP32[$13>>2]=0|0;HEAP32[$13+4>>2]=0|0;HEAP32[$13+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $25 = ((($this)) + 24|0); //@line 181 "./modules/graphics/Graphics.h"
 $26 = ((($0)) + 24|0); //@line 181 "./modules/graphics/Graphics.h"
 $27 = HEAP8[$25>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $28 = $27 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $29 = ($28<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($29) {
  $33 = ((($25)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$33>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$25>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 } else {
  $30 = ((($this)) + 32|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $31 = HEAP32[$30>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$31>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $32 = ((($this)) + 28|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$32>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 __THREW__ = 0;
 invoke_vii(3368,($25|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $34 = __THREW__; __THREW__ = 0;
 $35 = $34&1;
 if ($35) {
  $36 = ___cxa_find_matching_catch_3(0|0)|0;
  $37 = tempRet0;
  ___clang_call_terminate($36); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 ;HEAP32[$25>>2]=HEAP32[$26>>2]|0;HEAP32[$25+4>>2]=HEAP32[$26+4>>2]|0;HEAP32[$25+8>>2]=HEAP32[$26+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 ;HEAP32[$26>>2]=0|0;HEAP32[$26+4>>2]=0|0;HEAP32[$26+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $38 = ((($this)) + 36|0); //@line 181 "./modules/graphics/Graphics.h"
 $39 = ((($0)) + 36|0); //@line 181 "./modules/graphics/Graphics.h"
 $40 = HEAP8[$38>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $41 = $40 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $42 = ($41<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($42) {
  $46 = ((($38)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$46>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$38>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 } else {
  $43 = ((($this)) + 44|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $44 = HEAP32[$43>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$44>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $45 = ((($this)) + 40|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$45>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 __THREW__ = 0;
 invoke_vii(3368,($38|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $47 = __THREW__; __THREW__ = 0;
 $48 = $47&1;
 if ($48) {
  $49 = ___cxa_find_matching_catch_3(0|0)|0;
  $50 = tempRet0;
  ___clang_call_terminate($49); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 } else {
  ;HEAP32[$38>>2]=HEAP32[$39>>2]|0;HEAP32[$38+4>>2]=HEAP32[$39+4>>2]|0;HEAP32[$38+8>>2]=HEAP32[$39+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  ;HEAP32[$39>>2]=0|0;HEAP32[$39+4>>2]=0|0;HEAP32[$39+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  return ($this|0); //@line 181 "./modules/graphics/Graphics.h"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl17w_getSystemLimitsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0, $2 = 0.0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, $name = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp;
 _lua_createtable($L,0,4); //@line 1462 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAP32[$name>>2] = 0; //@line 1467 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = (__ZN4love8graphics8Graphics11getConstantENS1_11SystemLimitERPKc(0,$name)|0); //@line 1469 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($0) {
  $1 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  $2 = (+__ZNK4love8graphics6opengl8Graphics14getSystemLimitENS0_8Graphics11SystemLimitE($1,0)); //@line 1472 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_pushnumber($L,$2); //@line 1472 "modules/graphics/opengl/wrap_Graphics.cpp"
  $3 = HEAP32[$name>>2]|0; //@line 1473 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_setfield($L,-2,$3); //@line 1473 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 HEAP32[$name>>2] = 0; //@line 1467 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = (__ZN4love8graphics8Graphics11getConstantENS1_11SystemLimitERPKc(1,$name)|0); //@line 1469 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($4) {
  $5 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  $6 = (+__ZNK4love8graphics6opengl8Graphics14getSystemLimitENS0_8Graphics11SystemLimitE($5,1)); //@line 1472 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_pushnumber($L,$6); //@line 1472 "modules/graphics/opengl/wrap_Graphics.cpp"
  $7 = HEAP32[$name>>2]|0; //@line 1473 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_setfield($L,-2,$7); //@line 1473 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 HEAP32[$name>>2] = 0; //@line 1467 "modules/graphics/opengl/wrap_Graphics.cpp"
 $8 = (__ZN4love8graphics8Graphics11getConstantENS1_11SystemLimitERPKc(2,$name)|0); //@line 1469 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($8) {
  $9 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  $10 = (+__ZNK4love8graphics6opengl8Graphics14getSystemLimitENS0_8Graphics11SystemLimitE($9,2)); //@line 1472 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_pushnumber($L,$10); //@line 1472 "modules/graphics/opengl/wrap_Graphics.cpp"
  $11 = HEAP32[$name>>2]|0; //@line 1473 "modules/graphics/opengl/wrap_Graphics.cpp"
  _lua_setfield($L,-2,$11); //@line 1473 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 HEAP32[$name>>2] = 0; //@line 1467 "modules/graphics/opengl/wrap_Graphics.cpp"
 $12 = (__ZN4love8graphics8Graphics11getConstantENS1_11SystemLimitERPKc(3,$name)|0); //@line 1469 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($12)) {
  STACKTOP = sp;return 1; //@line 1476 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $13 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $14 = (+__ZNK4love8graphics6opengl8Graphics14getSystemLimitENS0_8Graphics11SystemLimitE($13,3)); //@line 1472 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushnumber($L,$14); //@line 1472 "modules/graphics/opengl/wrap_Graphics.cpp"
 $15 = HEAP32[$name>>2]|0; //@line 1473 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_setfield($L,-2,$15); //@line 1473 "modules/graphics/opengl/wrap_Graphics.cpp"
 STACKTOP = sp;return 1; //@line 1476 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl10w_getStatsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $sname = 0;
 var $stats = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $stats = sp + 8|0;
 $sname = sp;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZNK4love8graphics6opengl8Graphics8getStatsEv($stats,$0); //@line 1481 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_createtable($L,0,6); //@line 1483 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAP32[$sname>>2] = 0; //@line 1485 "modules/graphics/opengl/wrap_Graphics.cpp"
 (__ZN4love8graphics8Graphics11getConstantENS1_8StatTypeERPKc(0,$sname)|0); //@line 1487 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = HEAP32[$stats>>2]|0; //@line 1488 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushinteger($L,$1); //@line 1488 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = HEAP32[$sname>>2]|0; //@line 1489 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_setfield($L,-2,$2); //@line 1489 "modules/graphics/opengl/wrap_Graphics.cpp"
 (__ZN4love8graphics8Graphics11getConstantENS1_8StatTypeERPKc(1,$sname)|0); //@line 1491 "modules/graphics/opengl/wrap_Graphics.cpp"
 $3 = ((($stats)) + 4|0); //@line 1492 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 1492 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushinteger($L,$4); //@line 1492 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = HEAP32[$sname>>2]|0; //@line 1493 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_setfield($L,-2,$5); //@line 1493 "modules/graphics/opengl/wrap_Graphics.cpp"
 (__ZN4love8graphics8Graphics11getConstantENS1_8StatTypeERPKc(2,$sname)|0); //@line 1495 "modules/graphics/opengl/wrap_Graphics.cpp"
 $6 = ((($stats)) + 8|0); //@line 1496 "modules/graphics/opengl/wrap_Graphics.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1496 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushinteger($L,$7); //@line 1496 "modules/graphics/opengl/wrap_Graphics.cpp"
 $8 = HEAP32[$sname>>2]|0; //@line 1497 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_setfield($L,-2,$8); //@line 1497 "modules/graphics/opengl/wrap_Graphics.cpp"
 (__ZN4love8graphics8Graphics11getConstantENS1_8StatTypeERPKc(3,$sname)|0); //@line 1499 "modules/graphics/opengl/wrap_Graphics.cpp"
 $9 = ((($stats)) + 12|0); //@line 1500 "modules/graphics/opengl/wrap_Graphics.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 1500 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushinteger($L,$10); //@line 1500 "modules/graphics/opengl/wrap_Graphics.cpp"
 $11 = HEAP32[$sname>>2]|0; //@line 1501 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_setfield($L,-2,$11); //@line 1501 "modules/graphics/opengl/wrap_Graphics.cpp"
 (__ZN4love8graphics8Graphics11getConstantENS1_8StatTypeERPKc(4,$sname)|0); //@line 1503 "modules/graphics/opengl/wrap_Graphics.cpp"
 $12 = ((($stats)) + 16|0); //@line 1504 "modules/graphics/opengl/wrap_Graphics.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 1504 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushinteger($L,$13); //@line 1504 "modules/graphics/opengl/wrap_Graphics.cpp"
 $14 = HEAP32[$sname>>2]|0; //@line 1505 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_setfield($L,-2,$14); //@line 1505 "modules/graphics/opengl/wrap_Graphics.cpp"
 (__ZN4love8graphics8Graphics11getConstantENS1_8StatTypeERPKc(5,$sname)|0); //@line 1507 "modules/graphics/opengl/wrap_Graphics.cpp"
 $15 = ((($stats)) + 20|0); //@line 1508 "modules/graphics/opengl/wrap_Graphics.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1508 "modules/graphics/opengl/wrap_Graphics.cpp"
 $17 = (+($16>>>0)); //@line 1508 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_pushnumber($L,$17); //@line 1508 "modules/graphics/opengl/wrap_Graphics.cpp"
 $18 = HEAP32[$sname>>2]|0; //@line 1509 "modules/graphics/opengl/wrap_Graphics.cpp"
 _lua_setfield($L,-2,$18); //@line 1509 "modules/graphics/opengl/wrap_Graphics.cpp"
 STACKTOP = sp;return 1; //@line 1511 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl6w_drawEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0;
 var $24 = 0.0, $25 = 0.0, $26 = 0, $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $drawable$0 = 0, $or$cond = 0, $quad$0 = 0, $startidx$0 = 0, $texture$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 1514 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,2,13)|0); //@line 1521 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($0) {
  $1 = (__ZN4love8graphics17luax_checktextureEP9lua_Statei($L,1)|0); //@line 1523 "modules/graphics/opengl/wrap_Graphics.cpp"
  $2 = (_lua_touserdata($L,2)|0); //@line 516 "./common/runtime.h"
  $3 = ((($2)) + 4|0); //@line 516 "./common/runtime.h"
  $4 = HEAP32[$3>>2]|0; //@line 516 "./common/runtime.h"
  $drawable$0 = 0;$quad$0 = $4;$startidx$0 = 3;$texture$0 = $1;
 } else {
  $5 = (_lua_type($L,2)|0); //@line 1527 "modules/graphics/opengl/wrap_Graphics.cpp"
  $6 = ($5|0)==(0); //@line 1527 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($6) {
   $7 = (_lua_type($L,3)|0); //@line 1527 "modules/graphics/opengl/wrap_Graphics.cpp"
   $8 = ($7|0)<(1); //@line 1527 "modules/graphics/opengl/wrap_Graphics.cpp"
   if (!($8)) {
    $9 = (_luax_typerror($L,2,347523)|0); //@line 1529 "modules/graphics/opengl/wrap_Graphics.cpp"
    $$0 = $9;
    STACKTOP = sp;return ($$0|0); //@line 1555 "modules/graphics/opengl/wrap_Graphics.cpp"
   }
  }
  $10 = (__ZN4love14luax_checktypeINS_8graphics8DrawableEEEPT_P9lua_StateiNS_4TypeE($L,1,10)|0); //@line 1533 "modules/graphics/opengl/wrap_Graphics.cpp"
  $drawable$0 = $10;$quad$0 = 0;$startidx$0 = 2;$texture$0 = 0;
 }
 $11 = (+_luaL_optnumber($L,$startidx$0,0.0)); //@line 1537 "modules/graphics/opengl/wrap_Graphics.cpp"
 $12 = $11; //@line 1537 "modules/graphics/opengl/wrap_Graphics.cpp"
 $13 = (($startidx$0) + 1)|0; //@line 1538 "modules/graphics/opengl/wrap_Graphics.cpp"
 $14 = (+_luaL_optnumber($L,$13,0.0)); //@line 1538 "modules/graphics/opengl/wrap_Graphics.cpp"
 $15 = $14; //@line 1538 "modules/graphics/opengl/wrap_Graphics.cpp"
 $16 = (($startidx$0) + 2)|0; //@line 1539 "modules/graphics/opengl/wrap_Graphics.cpp"
 $17 = (+_luaL_optnumber($L,$16,0.0)); //@line 1539 "modules/graphics/opengl/wrap_Graphics.cpp"
 $18 = $17; //@line 1539 "modules/graphics/opengl/wrap_Graphics.cpp"
 $19 = (($startidx$0) + 3)|0; //@line 1540 "modules/graphics/opengl/wrap_Graphics.cpp"
 $20 = (+_luaL_optnumber($L,$19,1.0)); //@line 1540 "modules/graphics/opengl/wrap_Graphics.cpp"
 $21 = $20; //@line 1540 "modules/graphics/opengl/wrap_Graphics.cpp"
 $22 = $startidx$0 | 4; //@line 1541 "modules/graphics/opengl/wrap_Graphics.cpp"
 $23 = $21; //@line 1541 "modules/graphics/opengl/wrap_Graphics.cpp"
 $24 = (+_luaL_optnumber($L,$22,$23)); //@line 1541 "modules/graphics/opengl/wrap_Graphics.cpp"
 $25 = $24; //@line 1541 "modules/graphics/opengl/wrap_Graphics.cpp"
 $26 = (($startidx$0) + 5)|0; //@line 1542 "modules/graphics/opengl/wrap_Graphics.cpp"
 $27 = (+_luaL_optnumber($L,$26,0.0)); //@line 1542 "modules/graphics/opengl/wrap_Graphics.cpp"
 $28 = $27; //@line 1542 "modules/graphics/opengl/wrap_Graphics.cpp"
 $29 = (($startidx$0) + 6)|0; //@line 1543 "modules/graphics/opengl/wrap_Graphics.cpp"
 $30 = (+_luaL_optnumber($L,$29,0.0)); //@line 1543 "modules/graphics/opengl/wrap_Graphics.cpp"
 $31 = $30; //@line 1543 "modules/graphics/opengl/wrap_Graphics.cpp"
 $32 = (($startidx$0) + 7)|0; //@line 1544 "modules/graphics/opengl/wrap_Graphics.cpp"
 $33 = (+_luaL_optnumber($L,$32,0.0)); //@line 1544 "modules/graphics/opengl/wrap_Graphics.cpp"
 $34 = $33; //@line 1544 "modules/graphics/opengl/wrap_Graphics.cpp"
 $35 = $startidx$0 | 8; //@line 1545 "modules/graphics/opengl/wrap_Graphics.cpp"
 $36 = (+_luaL_optnumber($L,$35,0.0)); //@line 1545 "modules/graphics/opengl/wrap_Graphics.cpp"
 $37 = $36; //@line 1545 "modules/graphics/opengl/wrap_Graphics.cpp"
 $38 = ($texture$0|0)==(0|0); //@line 1548 "modules/graphics/opengl/wrap_Graphics.cpp"
 $39 = ($quad$0|0)==(0|0); //@line 1548 "modules/graphics/opengl/wrap_Graphics.cpp"
 $or$cond = $39 | $38; //@line 1548 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($or$cond) {
  $45 = ($drawable$0|0)==(0|0); //@line 1550 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($45) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0); //@line 1555 "modules/graphics/opengl/wrap_Graphics.cpp"
  }
  $46 = HEAP32[$drawable$0>>2]|0; //@line 1551 "modules/graphics/opengl/wrap_Graphics.cpp"
  $47 = ((($46)) + 16|0); //@line 1551 "modules/graphics/opengl/wrap_Graphics.cpp"
  $48 = HEAP32[$47>>2]|0; //@line 1551 "modules/graphics/opengl/wrap_Graphics.cpp"
  __THREW__ = 0;
  invoke_viddddddddd($48|0,($drawable$0|0),(+$12),(+$15),(+$18),(+$21),(+$25),(+$28),(+$31),(+$34),(+$37)); //@line 1551 "modules/graphics/opengl/wrap_Graphics.cpp"
  $49 = __THREW__; __THREW__ = 0;
  $50 = $49&1;
  if (!($50)) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0); //@line 1555 "modules/graphics/opengl/wrap_Graphics.cpp"
  }
 } else {
  $40 = HEAP32[$texture$0>>2]|0; //@line 1549 "modules/graphics/opengl/wrap_Graphics.cpp"
  $41 = ((($40)) + 20|0); //@line 1549 "modules/graphics/opengl/wrap_Graphics.cpp"
  $42 = HEAP32[$41>>2]|0; //@line 1549 "modules/graphics/opengl/wrap_Graphics.cpp"
  __THREW__ = 0;
  invoke_viiddddddddd($42|0,($texture$0|0),($quad$0|0),(+$12),(+$15),(+$18),(+$21),(+$25),(+$28),(+$31),(+$34),(+$37)); //@line 1549 "modules/graphics/opengl/wrap_Graphics.cpp"
  $43 = __THREW__; __THREW__ = 0;
  $44 = $43&1;
  if (!($44)) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0); //@line 1555 "modules/graphics/opengl/wrap_Graphics.cpp"
  }
 }
 $51 = ___cxa_find_matching_catch_3(8992|0)|0;
 $52 = tempRet0;
 $53 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $54 = ($52|0)==($53|0); //@line 535 "./common/runtime.h"
 if (!($54)) {
  $$01$i = $52;$$02$i = $51;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $55 = (___cxa_begin_catch(($51|0))|0); //@line 535 "./common/runtime.h"
 $56 = HEAP32[$55>>2]|0; //@line 539 "./common/runtime.h"
 $57 = ((($56)) + 8|0); //@line 539 "./common/runtime.h"
 $58 = HEAP32[$57>>2]|0; //@line 539 "./common/runtime.h"
 $59 = (FUNCTION_TABLE_ii[$58 & 8191]($55)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($59|0)); //@line 539 "./common/runtime.h"
 $60 = __THREW__; __THREW__ = 0;
 $61 = $60&1;
 if (!($61)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $62 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $62; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 1555 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $63 = ___cxa_find_matching_catch_2()|0;
 $64 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $65 = __THREW__; __THREW__ = 0;
 $66 = $65&1;
 if ($66) {
  $67 = ___cxa_find_matching_catch_3(0|0)|0;
  $68 = tempRet0;
  ___clang_call_terminate($67); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $64;$$02$i = $63;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love14luax_checktypeINS_8graphics8DrawableEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love8graphics6opengl7w_printEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $$pre$i = 0, $$pre$i5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0;
 var $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0, $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0.0, $4 = 0.0, $40 = 0.0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0.0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0.0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $str = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $str = sp + 4|0;
 HEAP32[$str>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $0 = ((($str)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($str)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 invoke_viii(3568,($L|0),1,($str|0)); //@line 1560 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 L1: do {
  if ($3) {
   label = 26;
  } else {
   __THREW__ = 0;
   $4 = (+invoke_diid(3525,($L|0),2,0.0)); //@line 1562 "modules/graphics/opengl/wrap_Graphics.cpp"
   $5 = __THREW__; __THREW__ = 0;
   $6 = $5&1;
   if ($6) {
    label = 26;
   } else {
    $7 = $4; //@line 1562 "modules/graphics/opengl/wrap_Graphics.cpp"
    __THREW__ = 0;
    $8 = (+invoke_diid(3525,($L|0),3,0.0)); //@line 1563 "modules/graphics/opengl/wrap_Graphics.cpp"
    $9 = __THREW__; __THREW__ = 0;
    $10 = $9&1;
    if ($10) {
     label = 26;
    } else {
     $11 = $8; //@line 1563 "modules/graphics/opengl/wrap_Graphics.cpp"
     __THREW__ = 0;
     $12 = (+invoke_diid(3525,($L|0),4,0.0)); //@line 1564 "modules/graphics/opengl/wrap_Graphics.cpp"
     $13 = __THREW__; __THREW__ = 0;
     $14 = $13&1;
     if ($14) {
      label = 26;
     } else {
      $15 = $12; //@line 1564 "modules/graphics/opengl/wrap_Graphics.cpp"
      __THREW__ = 0;
      $16 = (+invoke_diid(3525,($L|0),5,1.0)); //@line 1565 "modules/graphics/opengl/wrap_Graphics.cpp"
      $17 = __THREW__; __THREW__ = 0;
      $18 = $17&1;
      if ($18) {
       label = 26;
      } else {
       $19 = $16; //@line 1565 "modules/graphics/opengl/wrap_Graphics.cpp"
       $20 = $19; //@line 1566 "modules/graphics/opengl/wrap_Graphics.cpp"
       __THREW__ = 0;
       $21 = (+invoke_diid(3525,($L|0),6,(+$20))); //@line 1566 "modules/graphics/opengl/wrap_Graphics.cpp"
       $22 = __THREW__; __THREW__ = 0;
       $23 = $22&1;
       if ($23) {
        label = 26;
       } else {
        $24 = $21; //@line 1566 "modules/graphics/opengl/wrap_Graphics.cpp"
        __THREW__ = 0;
        $25 = (+invoke_diid(3525,($L|0),7,0.0)); //@line 1567 "modules/graphics/opengl/wrap_Graphics.cpp"
        $26 = __THREW__; __THREW__ = 0;
        $27 = $26&1;
        if ($27) {
         label = 26;
        } else {
         $28 = $25; //@line 1567 "modules/graphics/opengl/wrap_Graphics.cpp"
         __THREW__ = 0;
         $29 = (+invoke_diid(3525,($L|0),8,0.0)); //@line 1568 "modules/graphics/opengl/wrap_Graphics.cpp"
         $30 = __THREW__; __THREW__ = 0;
         $31 = $30&1;
         if ($31) {
          label = 26;
         } else {
          $32 = $29; //@line 1568 "modules/graphics/opengl/wrap_Graphics.cpp"
          __THREW__ = 0;
          $33 = (+invoke_diid(3525,($L|0),9,0.0)); //@line 1569 "modules/graphics/opengl/wrap_Graphics.cpp"
          $34 = __THREW__; __THREW__ = 0;
          $35 = $34&1;
          if ($35) {
           label = 26;
          } else {
           __THREW__ = 0;
           $36 = (+invoke_diid(3525,($L|0),10,0.0)); //@line 1570 "modules/graphics/opengl/wrap_Graphics.cpp"
           $37 = __THREW__; __THREW__ = 0;
           $38 = $37&1;
           if ($38) {
            label = 26;
           } else {
            $39 = $33; //@line 1569 "modules/graphics/opengl/wrap_Graphics.cpp"
            $40 = $36; //@line 1570 "modules/graphics/opengl/wrap_Graphics.cpp"
            $41 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
            __THREW__ = 0;
            invoke_viiddddddddd(3580,($41|0),($str|0),(+$7),(+$11),(+$15),(+$19),(+$24),(+$28),(+$32),(+$39),(+$40)); //@line 1573 "modules/graphics/opengl/wrap_Graphics.cpp"
            $42 = __THREW__; __THREW__ = 0;
            $43 = $42&1;
            L12: do {
             if ($43) {
              $44 = ___cxa_find_matching_catch_3(8992|0)|0;
              $45 = tempRet0;
              $46 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
              $47 = ($45|0)==($46|0); //@line 535 "./common/runtime.h"
              do {
               if ($47) {
                $48 = (___cxa_begin_catch(($44|0))|0); //@line 535 "./common/runtime.h"
                $49 = HEAP32[$48>>2]|0; //@line 539 "./common/runtime.h"
                $50 = ((($49)) + 8|0); //@line 539 "./common/runtime.h"
                $51 = HEAP32[$50>>2]|0; //@line 539 "./common/runtime.h"
                $52 = (FUNCTION_TABLE_ii[$51 & 8191]($48)|0); //@line 539 "./common/runtime.h"
                __THREW__ = 0;
                invoke_vii(3380,($L|0),($52|0)); //@line 539 "./common/runtime.h"
                $53 = __THREW__; __THREW__ = 0;
                $54 = $53&1;
                if ($54) {
                 $62 = ___cxa_find_matching_catch_2()|0;
                 $63 = tempRet0;
                 __THREW__ = 0;
                 invoke_v(3378); //@line 540 "./common/runtime.h"
                 $64 = __THREW__; __THREW__ = 0;
                 $65 = $64&1;
                 if (!($65)) {
                  $$01$i = $63;$$02$i = $62;
                  break;
                 }
                 $66 = ___cxa_find_matching_catch_3(0|0)|0;
                 $67 = tempRet0;
                 ___clang_call_terminate($66); //@line 540 "./common/runtime.h"
                 // unreachable; //@line 540 "./common/runtime.h"
                }
                __THREW__ = 0;
                invoke_v(3378); //@line 540 "./common/runtime.h"
                $55 = __THREW__; __THREW__ = 0;
                $56 = $55&1;
                if ($56) {
                 label = 26;
                 break L1;
                }
                __THREW__ = 0;
                $57 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
                $58 = __THREW__; __THREW__ = 0;
                $59 = $58&1;
                if ($59) {
                 label = 26;
                 break L1;
                }
                __THREW__ = 0;
                HEAP32[$vararg_buffer>>2] = $57; //@line 543 "./common/runtime.h"
                (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 543 "./common/runtime.h"
                $60 = __THREW__; __THREW__ = 0;
                $61 = $60&1;
                if ($61) {
                 label = 26;
                 break L1;
                } else {
                 break L12;
                }
               } else {
                $$01$i = $45;$$02$i = $44;
               }
              } while(0);
              $eh$lpad$body$index2Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
              break L1;
             }
            } while(0);
            $68 = HEAP32[$str>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $69 = ($68|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            if ($69) {
             STACKTOP = sp;return 0; //@line 1576 "modules/graphics/opengl/wrap_Graphics.cpp"
            }
            $70 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $71 = ($70|0)==($68|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            if ($71) {
             $76 = $68;
            } else {
             $73 = $70;
             while(1) {
              $72 = ((($73)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              HEAP32[$0>>2] = $72; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($72); //@line 64 "modules/graphics/opengl/Font.h"
              $74 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $75 = ($74|0)==($68|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              if ($75) {
               break;
              } else {
               $73 = $74;
              }
             }
             $$pre$i5 = HEAP32[$str>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $76 = $$pre$i5;
            }
            __ZdlPv($76); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
            STACKTOP = sp;return 0; //@line 1576 "modules/graphics/opengl/wrap_Graphics.cpp"
           }
          }
         }
        }
       }
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 26) {
  $77 = ___cxa_find_matching_catch_2()|0;
  $78 = tempRet0;
  $eh$lpad$body$index2Z2D = $78;$eh$lpad$body$indexZ2D = $77;
 }
 $79 = HEAP32[$str>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $80 = ($79|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($80) {
  ___resumeException($eh$lpad$body$indexZ2D|0);
  // unreachable;
 }
 $81 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $82 = ($81|0)==($79|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($82) {
  $87 = $79;
 } else {
  $84 = $81;
  while(1) {
   $83 = ((($84)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$0>>2] = $83; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($83); //@line 64 "modules/graphics/opengl/Font.h"
   $85 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $86 = ($85|0)==($79|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($86) {
    break;
   } else {
    $84 = $85;
   }
  }
  $$pre$i = HEAP32[$str>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $87 = $$pre$i;
 }
 __ZdlPv($87); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8graphics6opengl8w_printfEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $$1 = 0, $$pre$i = 0, $$pre$i4 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0.0, $12 = 0.0;
 var $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0.0, $45 = 0, $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0, $55 = 0, $56 = 0.0, $57 = 0.0, $58 = 0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0, $64 = 0.0, $65 = 0.0, $66 = 0, $67 = 0;
 var $68 = 0.0, $69 = 0, $7 = 0.0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0.0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $align = 0, $angle$0 = 0.0, $eh$lpad$body$index5Z2D = 0, $eh$lpad$body$indexZ2D = 0, $kx$0 = 0.0;
 var $ky$0 = 0.0, $ox$0 = 0.0, $oy$0 = 0.0, $str = 0, $sx$0 = 0.0, $sy$0 = 0.0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $str = sp + 16|0;
 $align = sp + 12|0;
 HEAP32[$str>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $0 = ((($str)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($str)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 invoke_viii(3568,($L|0),1,($str|0)); //@line 1581 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 L1: do {
  if ($3) {
   label = 13;
  } else {
   __THREW__ = 0;
   $4 = (+invoke_dii(3361,($L|0),2)); //@line 1583 "modules/graphics/opengl/wrap_Graphics.cpp"
   $5 = __THREW__; __THREW__ = 0;
   $6 = $5&1;
   if ($6) {
    label = 13;
   } else {
    $7 = $4; //@line 1583 "modules/graphics/opengl/wrap_Graphics.cpp"
    __THREW__ = 0;
    $8 = (+invoke_dii(3361,($L|0),3)); //@line 1584 "modules/graphics/opengl/wrap_Graphics.cpp"
    $9 = __THREW__; __THREW__ = 0;
    $10 = $9&1;
    if ($10) {
     label = 13;
    } else {
     $11 = $8; //@line 1584 "modules/graphics/opengl/wrap_Graphics.cpp"
     __THREW__ = 0;
     $12 = (+invoke_dii(3361,($L|0),4)); //@line 1585 "modules/graphics/opengl/wrap_Graphics.cpp"
     $13 = __THREW__; __THREW__ = 0;
     $14 = $13&1;
     if ($14) {
      label = 13;
     } else {
      $15 = $12; //@line 1585 "modules/graphics/opengl/wrap_Graphics.cpp"
      HEAP32[$align>>2] = 0; //@line 1592 "modules/graphics/opengl/wrap_Graphics.cpp"
      __THREW__ = 0;
      $16 = (invoke_ii(3385,($L|0))|0); //@line 1594 "modules/graphics/opengl/wrap_Graphics.cpp"
      $17 = __THREW__; __THREW__ = 0;
      $18 = $17&1;
      if ($18) {
       label = 13;
      } else {
       $19 = ($16|0)>(4); //@line 1594 "modules/graphics/opengl/wrap_Graphics.cpp"
       do {
        if ($19) {
         __THREW__ = 0;
         $20 = (invoke_iii(3358,($L|0),5)|0); //@line 1596 "modules/graphics/opengl/wrap_Graphics.cpp"
         $21 = __THREW__; __THREW__ = 0;
         $22 = $21&1;
         if ($22) {
          label = 13;
          break L1;
         }
         $23 = ($20|0)==(0); //@line 1596 "modules/graphics/opengl/wrap_Graphics.cpp"
         if (!($23)) {
          __THREW__ = 0;
          $24 = (invoke_iiii(3359,($L|0),5,(0|0))|0); //@line 1598 "modules/graphics/opengl/wrap_Graphics.cpp"
          $25 = __THREW__; __THREW__ = 0;
          $26 = $25&1;
          if ($26) {
           label = 13;
           break L1;
          }
          __THREW__ = 0;
          $27 = (invoke_iii(3581,($24|0),($align|0))|0); //@line 1599 "modules/graphics/opengl/wrap_Graphics.cpp"
          $28 = __THREW__; __THREW__ = 0;
          $29 = $28&1;
          if ($29) {
           label = 13;
           break L1;
          }
          if (!($27)) {
           __THREW__ = 0;
           HEAP32[$vararg_buffer>>2] = $24; //@line 1600 "modules/graphics/opengl/wrap_Graphics.cpp"
           $30 = (invoke_iiii(3389,($L|0),(330504|0),($vararg_buffer|0))|0); //@line 1600 "modules/graphics/opengl/wrap_Graphics.cpp"
           $31 = __THREW__; __THREW__ = 0;
           $32 = $31&1;
           if ($32) {
            label = 13;
            break L1;
           } else {
            $$1 = $30;
            break;
           }
          }
         }
         __THREW__ = 0;
         $44 = (+invoke_diid(3525,($L|0),6,0.0)); //@line 1603 "modules/graphics/opengl/wrap_Graphics.cpp"
         $45 = __THREW__; __THREW__ = 0;
         $46 = $45&1;
         if ($46) {
          label = 13;
          break L1;
         }
         $47 = $44; //@line 1603 "modules/graphics/opengl/wrap_Graphics.cpp"
         __THREW__ = 0;
         $48 = (+invoke_diid(3525,($L|0),7,1.0)); //@line 1604 "modules/graphics/opengl/wrap_Graphics.cpp"
         $49 = __THREW__; __THREW__ = 0;
         $50 = $49&1;
         if ($50) {
          label = 13;
          break L1;
         }
         $51 = $48; //@line 1604 "modules/graphics/opengl/wrap_Graphics.cpp"
         $52 = $51; //@line 1605 "modules/graphics/opengl/wrap_Graphics.cpp"
         __THREW__ = 0;
         $53 = (+invoke_diid(3525,($L|0),8,(+$52))); //@line 1605 "modules/graphics/opengl/wrap_Graphics.cpp"
         $54 = __THREW__; __THREW__ = 0;
         $55 = $54&1;
         if ($55) {
          label = 13;
          break L1;
         }
         $56 = $53; //@line 1605 "modules/graphics/opengl/wrap_Graphics.cpp"
         __THREW__ = 0;
         $57 = (+invoke_diid(3525,($L|0),9,0.0)); //@line 1606 "modules/graphics/opengl/wrap_Graphics.cpp"
         $58 = __THREW__; __THREW__ = 0;
         $59 = $58&1;
         if ($59) {
          label = 13;
          break L1;
         }
         $60 = $57; //@line 1606 "modules/graphics/opengl/wrap_Graphics.cpp"
         __THREW__ = 0;
         $61 = (+invoke_diid(3525,($L|0),10,0.0)); //@line 1607 "modules/graphics/opengl/wrap_Graphics.cpp"
         $62 = __THREW__; __THREW__ = 0;
         $63 = $62&1;
         if ($63) {
          label = 13;
          break L1;
         }
         $64 = $61; //@line 1607 "modules/graphics/opengl/wrap_Graphics.cpp"
         __THREW__ = 0;
         $65 = (+invoke_diid(3525,($L|0),11,0.0)); //@line 1608 "modules/graphics/opengl/wrap_Graphics.cpp"
         $66 = __THREW__; __THREW__ = 0;
         $67 = $66&1;
         if ($67) {
          label = 13;
          break L1;
         }
         __THREW__ = 0;
         $68 = (+invoke_diid(3525,($L|0),12,0.0)); //@line 1609 "modules/graphics/opengl/wrap_Graphics.cpp"
         $69 = __THREW__; __THREW__ = 0;
         $70 = $69&1;
         if ($70) {
          label = 13;
          break L1;
         }
         $71 = $65; //@line 1608 "modules/graphics/opengl/wrap_Graphics.cpp"
         $72 = $68; //@line 1609 "modules/graphics/opengl/wrap_Graphics.cpp"
         $angle$0 = $47;$kx$0 = $71;$ky$0 = $72;$ox$0 = $60;$oy$0 = $64;$sx$0 = $51;$sy$0 = $56;
         label = 28;
        } else {
         $angle$0 = 0.0;$kx$0 = 0.0;$ky$0 = 0.0;$ox$0 = 0.0;$oy$0 = 0.0;$sx$0 = 1.0;$sy$0 = 1.0;
         label = 28;
        }
       } while(0);
       L23: do {
        if ((label|0) == 28) {
         $73 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
         $74 = HEAP32[$align>>2]|0; //@line 1613 "modules/graphics/opengl/wrap_Graphics.cpp"
         __THREW__ = 0;
         invoke_viidddiddddddd(3582,($73|0),($str|0),(+$7),(+$11),(+$15),($74|0),(+$angle$0),(+$sx$0),(+$sy$0),(+$ox$0),(+$oy$0),(+$kx$0),(+$ky$0)); //@line 1613 "modules/graphics/opengl/wrap_Graphics.cpp"
         $75 = __THREW__; __THREW__ = 0;
         $76 = $75&1;
         if ($76) {
          $77 = ___cxa_find_matching_catch_3(8992|0)|0;
          $78 = tempRet0;
          $79 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
          $80 = ($78|0)==($79|0); //@line 535 "./common/runtime.h"
          do {
           if ($80) {
            $81 = (___cxa_begin_catch(($77|0))|0); //@line 535 "./common/runtime.h"
            $82 = HEAP32[$81>>2]|0; //@line 539 "./common/runtime.h"
            $83 = ((($82)) + 8|0); //@line 539 "./common/runtime.h"
            $84 = HEAP32[$83>>2]|0; //@line 539 "./common/runtime.h"
            $85 = (FUNCTION_TABLE_ii[$84 & 8191]($81)|0); //@line 539 "./common/runtime.h"
            __THREW__ = 0;
            invoke_vii(3380,($L|0),($85|0)); //@line 539 "./common/runtime.h"
            $86 = __THREW__; __THREW__ = 0;
            $87 = $86&1;
            if ($87) {
             $95 = ___cxa_find_matching_catch_2()|0;
             $96 = tempRet0;
             __THREW__ = 0;
             invoke_v(3378); //@line 540 "./common/runtime.h"
             $97 = __THREW__; __THREW__ = 0;
             $98 = $97&1;
             if (!($98)) {
              $$01$i = $96;$$02$i = $95;
              break;
             }
             $99 = ___cxa_find_matching_catch_3(0|0)|0;
             $100 = tempRet0;
             ___clang_call_terminate($99); //@line 540 "./common/runtime.h"
             // unreachable; //@line 540 "./common/runtime.h"
            }
            __THREW__ = 0;
            invoke_v(3378); //@line 540 "./common/runtime.h"
            $88 = __THREW__; __THREW__ = 0;
            $89 = $88&1;
            if ($89) {
             label = 13;
             break L1;
            }
            __THREW__ = 0;
            $90 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
            $91 = __THREW__; __THREW__ = 0;
            $92 = $91&1;
            if ($92) {
             label = 13;
             break L1;
            }
            __THREW__ = 0;
            HEAP32[$vararg_buffer1>>2] = $90; //@line 543 "./common/runtime.h"
            (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer1|0))|0); //@line 543 "./common/runtime.h"
            $93 = __THREW__; __THREW__ = 0;
            $94 = $93&1;
            if ($94) {
             label = 13;
             break L1;
            } else {
             $$1 = 0;
             break L23;
            }
           } else {
            $$01$i = $78;$$02$i = $77;
           }
          } while(0);
          $eh$lpad$body$index5Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
          break L1;
         } else {
          $$1 = 0;
         }
        }
       } while(0);
       $101 = HEAP32[$str>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $102 = ($101|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($102) {
        STACKTOP = sp;return ($$1|0); //@line 1616 "modules/graphics/opengl/wrap_Graphics.cpp"
       }
       $103 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $104 = ($103|0)==($101|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($104) {
        $109 = $101;
       } else {
        $106 = $103;
        while(1) {
         $105 = ((($106)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$0>>2] = $105; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($105); //@line 64 "modules/graphics/opengl/Font.h"
         $107 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $108 = ($107|0)==($101|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($108) {
          break;
         } else {
          $106 = $107;
         }
        }
        $$pre$i = HEAP32[$str>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $109 = $$pre$i;
       }
       __ZdlPv($109); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       STACKTOP = sp;return ($$1|0); //@line 1616 "modules/graphics/opengl/wrap_Graphics.cpp"
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 13) {
  $33 = ___cxa_find_matching_catch_2()|0;
  $34 = tempRet0;
  $eh$lpad$body$index5Z2D = $34;$eh$lpad$body$indexZ2D = $33;
 }
 $35 = HEAP32[$str>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $36 = ($35|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($36) {
  ___resumeException($eh$lpad$body$indexZ2D|0);
  // unreachable;
 }
 $37 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $38 = ($37|0)==($35|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($38) {
  $43 = $35;
 } else {
  $40 = $37;
  while(1) {
   $39 = ((($40)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$0>>2] = $39; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($39); //@line 64 "modules/graphics/opengl/Font.h"
   $41 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $42 = ($41|0)==($35|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($42) {
    break;
   } else {
    $40 = $41;
   }
  }
  $$pre$i4 = HEAP32[$str>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $43 = $$pre$i4;
 }
 __ZdlPv($43); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8graphics6opengl8w_pointsEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$arith = 0, $$overflow = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0;
 var $24 = 0.0, $25 = 0.0, $26 = 0, $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $args$0 = 0, $args$0$ = 0, $brmerge = 0;
 var $colors$0 = 0, $exitcond13 = 0, $exitcond14 = 0, $exitcond15 = 0, $i$06 = 0, $i1$07 = 0, $i2$09 = 0, $is_table$0$off0 = 0, $is_table_of_tables$0$off0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 1618 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = (_lua_gettop($L)|0); //@line 1625 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = ($0|0)==(1); //@line 1628 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($1) {
  $2 = (_lua_type($L,1)|0); //@line 1628 "modules/graphics/opengl/wrap_Graphics.cpp"
  $3 = ($2|0)==(5); //@line 1628 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($3) {
   $4 = (__ZN4love11luax_objlenEP9lua_Statei($L,1)|0); //@line 1631 "modules/graphics/opengl/wrap_Graphics.cpp"
   _lua_rawgeti($L,1,1); //@line 1633 "modules/graphics/opengl/wrap_Graphics.cpp"
   $5 = (_lua_type($L,-1)|0); //@line 1634 "modules/graphics/opengl/wrap_Graphics.cpp"
   $6 = ($5|0)==(5); //@line 1634 "modules/graphics/opengl/wrap_Graphics.cpp"
   _lua_settop($L,-2); //@line 1635 "modules/graphics/opengl/wrap_Graphics.cpp"
   $args$0 = $4;$is_table$0$off0 = 1;$is_table_of_tables$0$off0 = $6;
  } else {
   $args$0 = 1;$is_table$0$off0 = 0;$is_table_of_tables$0$off0 = 0;
  }
 } else {
  $args$0 = $0;$is_table$0$off0 = 0;$is_table_of_tables$0$off0 = 0;
 }
 $7 = $args$0 & 1; //@line 1638 "modules/graphics/opengl/wrap_Graphics.cpp"
 $8 = ($7|0)==(0); //@line 1638 "modules/graphics/opengl/wrap_Graphics.cpp"
 $brmerge = $8 | $is_table_of_tables$0$off0; //@line 1638 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($brmerge)) {
  $9 = (_luaL_error($L,330528,$vararg_buffer)|0); //@line 1639 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = $9;
  STACKTOP = sp;return ($$0|0); //@line 1699 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $10 = (($args$0|0) / 2)&-1; //@line 1641 "modules/graphics/opengl/wrap_Graphics.cpp"
 $args$0$ = $is_table_of_tables$0$off0 ? $args$0 : $10; //@line 1642 "modules/graphics/opengl/wrap_Graphics.cpp"
 $11 = $args$0$ << 1; //@line 1648 "modules/graphics/opengl/wrap_Graphics.cpp"
 $$arith = $11<<2; //@line 1648 "modules/graphics/opengl/wrap_Graphics.cpp"
 $$overflow = ($11>>>0)>(1073741823); //@line 1648 "modules/graphics/opengl/wrap_Graphics.cpp"
 $12 = $$overflow ? -1 : $$arith; //@line 1648 "modules/graphics/opengl/wrap_Graphics.cpp"
 $13 = (__Znaj($12)|0); //@line 1648 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($is_table_of_tables$0$off0) {
  $14 = $args$0$ << 2; //@line 1651 "modules/graphics/opengl/wrap_Graphics.cpp"
  $15 = ($args$0$|0)<(0); //@line 1651 "modules/graphics/opengl/wrap_Graphics.cpp"
  $16 = $15 ? -1 : $14; //@line 1651 "modules/graphics/opengl/wrap_Graphics.cpp"
  $17 = (__Znaj($16)|0); //@line 1651 "modules/graphics/opengl/wrap_Graphics.cpp"
  $colors$0 = $17;
 } else {
  $colors$0 = 0;
 }
 $18 = ($args$0|0)>(0); //@line 1658 "modules/graphics/opengl/wrap_Graphics.cpp"
 do {
  if ($is_table$0$off0) {
   if ($is_table_of_tables$0$off0) {
    if ($18) {
     $i$06 = 0;
    } else {
     break;
    }
    while(1) {
     $19 = (($i$06) + 1)|0; //@line 1660 "modules/graphics/opengl/wrap_Graphics.cpp"
     _lua_rawgeti($L,1,$19); //@line 1660 "modules/graphics/opengl/wrap_Graphics.cpp"
     _lua_rawgeti($L,-1,1); //@line 1662 "modules/graphics/opengl/wrap_Graphics.cpp"
     _lua_rawgeti($L,-2,2); //@line 1662 "modules/graphics/opengl/wrap_Graphics.cpp"
     _lua_rawgeti($L,-3,3); //@line 1662 "modules/graphics/opengl/wrap_Graphics.cpp"
     _lua_rawgeti($L,-4,4); //@line 1662 "modules/graphics/opengl/wrap_Graphics.cpp"
     _lua_rawgeti($L,-5,5); //@line 1662 "modules/graphics/opengl/wrap_Graphics.cpp"
     _lua_rawgeti($L,-6,6); //@line 1662 "modules/graphics/opengl/wrap_Graphics.cpp"
     $20 = (+_lua_tonumber($L,-6)); //@line 171 "./common/runtime.h"
     $21 = $20; //@line 171 "./common/runtime.h"
     $22 = $i$06 << 1; //@line 1664 "modules/graphics/opengl/wrap_Graphics.cpp"
     $23 = (($13) + ($22<<2)|0); //@line 1664 "modules/graphics/opengl/wrap_Graphics.cpp"
     HEAPF32[$23>>2] = $21; //@line 1664 "modules/graphics/opengl/wrap_Graphics.cpp"
     $24 = (+_lua_tonumber($L,-5)); //@line 171 "./common/runtime.h"
     $25 = $24; //@line 171 "./common/runtime.h"
     $26 = $22 | 1; //@line 1665 "modules/graphics/opengl/wrap_Graphics.cpp"
     $27 = (($13) + ($26<<2)|0); //@line 1665 "modules/graphics/opengl/wrap_Graphics.cpp"
     HEAPF32[$27>>2] = $25; //@line 1665 "modules/graphics/opengl/wrap_Graphics.cpp"
     $28 = (+_luaL_optnumber($L,-4,255.0)); //@line 1667 "modules/graphics/opengl/wrap_Graphics.cpp"
     $29 = (~~(($28))&255); //@line 1667 "modules/graphics/opengl/wrap_Graphics.cpp"
     $30 = $i$06 << 2; //@line 1667 "modules/graphics/opengl/wrap_Graphics.cpp"
     $31 = (($colors$0) + ($30)|0); //@line 1667 "modules/graphics/opengl/wrap_Graphics.cpp"
     HEAP8[$31>>0] = $29; //@line 1667 "modules/graphics/opengl/wrap_Graphics.cpp"
     $32 = (+_luaL_optnumber($L,-3,255.0)); //@line 1668 "modules/graphics/opengl/wrap_Graphics.cpp"
     $33 = (~~(($32))&255); //@line 1668 "modules/graphics/opengl/wrap_Graphics.cpp"
     $34 = $30 | 1; //@line 1668 "modules/graphics/opengl/wrap_Graphics.cpp"
     $35 = (($colors$0) + ($34)|0); //@line 1668 "modules/graphics/opengl/wrap_Graphics.cpp"
     HEAP8[$35>>0] = $33; //@line 1668 "modules/graphics/opengl/wrap_Graphics.cpp"
     $36 = (+_luaL_optnumber($L,-2,255.0)); //@line 1669 "modules/graphics/opengl/wrap_Graphics.cpp"
     $37 = (~~(($36))&255); //@line 1669 "modules/graphics/opengl/wrap_Graphics.cpp"
     $38 = $30 | 2; //@line 1669 "modules/graphics/opengl/wrap_Graphics.cpp"
     $39 = (($colors$0) + ($38)|0); //@line 1669 "modules/graphics/opengl/wrap_Graphics.cpp"
     HEAP8[$39>>0] = $37; //@line 1669 "modules/graphics/opengl/wrap_Graphics.cpp"
     $40 = (+_luaL_optnumber($L,-1,255.0)); //@line 1670 "modules/graphics/opengl/wrap_Graphics.cpp"
     $41 = (~~(($40))&255); //@line 1670 "modules/graphics/opengl/wrap_Graphics.cpp"
     $42 = $30 | 3; //@line 1670 "modules/graphics/opengl/wrap_Graphics.cpp"
     $43 = (($colors$0) + ($42)|0); //@line 1670 "modules/graphics/opengl/wrap_Graphics.cpp"
     HEAP8[$43>>0] = $41; //@line 1670 "modules/graphics/opengl/wrap_Graphics.cpp"
     _lua_settop($L,-8); //@line 1672 "modules/graphics/opengl/wrap_Graphics.cpp"
     $exitcond13 = ($19|0)==($args$0|0); //@line 1658 "modules/graphics/opengl/wrap_Graphics.cpp"
     if ($exitcond13) {
      break;
     } else {
      $i$06 = $19;
     }
    }
   } else {
    if ($18) {
     $i1$07 = 0;
    } else {
     break;
    }
    while(1) {
     $44 = (($i1$07) + 1)|0; //@line 1680 "modules/graphics/opengl/wrap_Graphics.cpp"
     _lua_rawgeti($L,1,$44); //@line 1680 "modules/graphics/opengl/wrap_Graphics.cpp"
     $45 = (+_lua_tonumber($L,-1)); //@line 171 "./common/runtime.h"
     $46 = $45; //@line 171 "./common/runtime.h"
     $47 = (($13) + ($i1$07<<2)|0); //@line 1681 "modules/graphics/opengl/wrap_Graphics.cpp"
     HEAPF32[$47>>2] = $46; //@line 1681 "modules/graphics/opengl/wrap_Graphics.cpp"
     _lua_settop($L,-2); //@line 1682 "modules/graphics/opengl/wrap_Graphics.cpp"
     $exitcond14 = ($44|0)==($args$0|0); //@line 1678 "modules/graphics/opengl/wrap_Graphics.cpp"
     if ($exitcond14) {
      break;
     } else {
      $i1$07 = $44;
     }
    }
   }
  } else {
   if ($18) {
    $i2$09 = 0;
    while(1) {
     $48 = (($i2$09) + 1)|0; //@line 1689 "modules/graphics/opengl/wrap_Graphics.cpp"
     $49 = (+_lua_tonumber($L,$48)); //@line 171 "./common/runtime.h"
     $50 = $49; //@line 171 "./common/runtime.h"
     $51 = (($13) + ($i2$09<<2)|0); //@line 1689 "modules/graphics/opengl/wrap_Graphics.cpp"
     HEAPF32[$51>>2] = $50; //@line 1689 "modules/graphics/opengl/wrap_Graphics.cpp"
     $exitcond15 = ($48|0)==($args$0|0); //@line 1688 "modules/graphics/opengl/wrap_Graphics.cpp"
     if ($exitcond15) {
      break;
     } else {
      $i2$09 = $48;
     }
    }
   }
  }
 } while(0);
 $52 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZN4love8graphics6opengl8Graphics6pointsEPKfPKhj($52,$13,$colors$0,$args$0$); //@line 1692 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZdaPv($13); //@line 1694 "modules/graphics/opengl/wrap_Graphics.cpp"
 $53 = ($colors$0|0)==(0|0); //@line 1695 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($53) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 1699 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 __ZdaPv($colors$0); //@line 1696 "modules/graphics/opengl/wrap_Graphics.cpp"
 $$0 = 0;
 STACKTOP = sp;return ($$0|0); //@line 1699 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl6w_lineEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$arith = 0, $$overflow = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $args$0 = 0, $exitcond = 0, $exitcond8 = 0, $i$04 = 0, $i1$05 = 0, $is_table$0$off0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 1701 "modules/graphics/opengl/wrap_Graphics.cpp"
 $vararg_buffer = sp; //@line 1701 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = (_lua_gettop($L)|0); //@line 1703 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = ($0|0)==(1); //@line 1705 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($1) {
  $2 = (_lua_type($L,1)|0); //@line 1705 "modules/graphics/opengl/wrap_Graphics.cpp"
  $3 = ($2|0)==(5); //@line 1705 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($3) {
   $4 = (__ZN4love11luax_objlenEP9lua_Statei($L,1)|0); //@line 1707 "modules/graphics/opengl/wrap_Graphics.cpp"
   $args$0 = $4;$is_table$0$off0 = 1;
   label = 4;
  }
 } else {
  $args$0 = $0;$is_table$0$off0 = 0;
  label = 4;
 }
 if ((label|0) == 4) {
  $5 = $args$0 & 1; //@line 1711 "modules/graphics/opengl/wrap_Graphics.cpp"
  $6 = ($5|0)==(0); //@line 1711 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($6) {
   $8 = ($args$0|0)<(4); //@line 1713 "modules/graphics/opengl/wrap_Graphics.cpp"
   if ($8) {
    $9 = (_luaL_error($L,330582,$vararg_buffer1)|0); //@line 1714 "modules/graphics/opengl/wrap_Graphics.cpp"
    $$0 = $9;
    STACKTOP = sp;return ($$0|0); //@line 1736 "modules/graphics/opengl/wrap_Graphics.cpp"
   }
   $$arith = $args$0<<2; //@line 1716 "modules/graphics/opengl/wrap_Graphics.cpp"
   $$overflow = ($args$0>>>0)>(1073741823); //@line 1716 "modules/graphics/opengl/wrap_Graphics.cpp"
   $10 = $$overflow ? -1 : $$arith; //@line 1716 "modules/graphics/opengl/wrap_Graphics.cpp"
   $11 = (__Znaj($10)|0); //@line 1716 "modules/graphics/opengl/wrap_Graphics.cpp"
   if ($is_table$0$off0) {
    $i$04 = 0;
    while(1) {
     $12 = (($i$04) + 1)|0; //@line 1721 "modules/graphics/opengl/wrap_Graphics.cpp"
     _lua_rawgeti($L,1,$12); //@line 1721 "modules/graphics/opengl/wrap_Graphics.cpp"
     $13 = (+_lua_tonumber($L,-1)); //@line 171 "./common/runtime.h"
     $14 = $13; //@line 171 "./common/runtime.h"
     $15 = (($11) + ($i$04<<2)|0); //@line 1722 "modules/graphics/opengl/wrap_Graphics.cpp"
     HEAPF32[$15>>2] = $14; //@line 1722 "modules/graphics/opengl/wrap_Graphics.cpp"
     _lua_settop($L,-2); //@line 1723 "modules/graphics/opengl/wrap_Graphics.cpp"
     $exitcond = ($12|0)==($args$0|0); //@line 1719 "modules/graphics/opengl/wrap_Graphics.cpp"
     if ($exitcond) {
      break;
     } else {
      $i$04 = $12;
     }
    }
   } else {
    $i1$05 = 0;
    while(1) {
     $16 = (($i1$05) + 1)|0; //@line 1729 "modules/graphics/opengl/wrap_Graphics.cpp"
     $17 = (+_lua_tonumber($L,$16)); //@line 171 "./common/runtime.h"
     $18 = $17; //@line 171 "./common/runtime.h"
     $19 = (($11) + ($i1$05<<2)|0); //@line 1729 "modules/graphics/opengl/wrap_Graphics.cpp"
     HEAPF32[$19>>2] = $18; //@line 1729 "modules/graphics/opengl/wrap_Graphics.cpp"
     $exitcond8 = ($16|0)==($args$0|0); //@line 1728 "modules/graphics/opengl/wrap_Graphics.cpp"
     if ($exitcond8) {
      break;
     } else {
      $i1$05 = $16;
     }
    }
   }
   $20 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
   __ZN4love8graphics6opengl8Graphics8polylineEPKfj($20,$11,$args$0); //@line 1732 "modules/graphics/opengl/wrap_Graphics.cpp"
   __ZdaPv($11); //@line 1734 "modules/graphics/opengl/wrap_Graphics.cpp"
   $$0 = 0;
   STACKTOP = sp;return ($$0|0); //@line 1736 "modules/graphics/opengl/wrap_Graphics.cpp"
  }
 }
 $7 = (_luaL_error($L,330528,$vararg_buffer)|0); //@line 1712 "modules/graphics/opengl/wrap_Graphics.cpp"
 $$0 = $7;
 STACKTOP = sp;return ($$0|0); //@line 1736 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl11w_rectangleEP9lua_State($L) {
 $L = $L|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0;
 var $26 = 0, $27 = 0.0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $mode = 0, $points$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $mode = sp + 4|0;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 1741 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = (__ZN4love8graphics8Graphics11getConstantEPKcRNS1_8DrawModeE($0,$mode)|0); //@line 1742 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($1)) {
  HEAP32[$vararg_buffer>>2] = $0; //@line 1743 "modules/graphics/opengl/wrap_Graphics.cpp"
  $2 = (_luaL_error($L,330624,$vararg_buffer)|0); //@line 1743 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$1 = $2;
  STACKTOP = sp;return ($$1|0); //@line 1767 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $3 = (+_luaL_checknumber($L,2)); //@line 1745 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = $3; //@line 1745 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = (+_luaL_checknumber($L,3)); //@line 1746 "modules/graphics/opengl/wrap_Graphics.cpp"
 $6 = $5; //@line 1746 "modules/graphics/opengl/wrap_Graphics.cpp"
 $7 = (+_luaL_checknumber($L,4)); //@line 1747 "modules/graphics/opengl/wrap_Graphics.cpp"
 $8 = $7; //@line 1747 "modules/graphics/opengl/wrap_Graphics.cpp"
 $9 = (+_luaL_checknumber($L,5)); //@line 1748 "modules/graphics/opengl/wrap_Graphics.cpp"
 $10 = $9; //@line 1748 "modules/graphics/opengl/wrap_Graphics.cpp"
 $11 = (_lua_type($L,6)|0); //@line 1750 "modules/graphics/opengl/wrap_Graphics.cpp"
 $12 = ($11|0)<(1); //@line 1750 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($12) {
  $13 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  $14 = HEAP32[$mode>>2]|0; //@line 1752 "modules/graphics/opengl/wrap_Graphics.cpp"
  __ZN4love8graphics6opengl8Graphics9rectangleENS0_8Graphics8DrawModeEffff($13,$14,$4,$6,$8,$10); //@line 1752 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 1767 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $15 = (+_luaL_optnumber($L,6,0.0)); //@line 1756 "modules/graphics/opengl/wrap_Graphics.cpp"
 $16 = $15; //@line 1756 "modules/graphics/opengl/wrap_Graphics.cpp"
 $17 = $16; //@line 1757 "modules/graphics/opengl/wrap_Graphics.cpp"
 $18 = (+_luaL_optnumber($L,7,$17)); //@line 1757 "modules/graphics/opengl/wrap_Graphics.cpp"
 $19 = $18; //@line 1757 "modules/graphics/opengl/wrap_Graphics.cpp"
 $20 = (_lua_type($L,8)|0); //@line 1760 "modules/graphics/opengl/wrap_Graphics.cpp"
 $21 = ($20|0)<(1); //@line 1760 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($21) {
  $22 = $16 < $19; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $23 = $22 ? $19 : $16; //@line 1761 "modules/graphics/opengl/wrap_Graphics.cpp"
  $24 = $23 > 20.0; //@line 1761 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($24) {
   $25 = $23 * 0.5; //@line 1761 "modules/graphics/opengl/wrap_Graphics.cpp"
   $26 = (~~(($25))); //@line 1761 "modules/graphics/opengl/wrap_Graphics.cpp"
   $points$0 = $26;
  } else {
   $points$0 = 10;
  }
 } else {
  $27 = (+_luaL_checknumber($L,8)); //@line 1763 "modules/graphics/opengl/wrap_Graphics.cpp"
  $28 = (~~(($27))); //@line 1763 "modules/graphics/opengl/wrap_Graphics.cpp"
  $points$0 = $28;
 }
 $29 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $30 = HEAP32[$mode>>2]|0; //@line 1765 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love8graphics6opengl8Graphics9rectangleENS0_8Graphics8DrawModeEffffffi($29,$30,$4,$6,$8,$10,$16,$19,$points$0); //@line 1765 "modules/graphics/opengl/wrap_Graphics.cpp"
 $$1 = 0;
 STACKTOP = sp;return ($$1|0); //@line 1767 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl8w_circleEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0, $mode = 0;
 var $points$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $mode = sp + 4|0;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 1772 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = (__ZN4love8graphics8Graphics11getConstantEPKcRNS1_8DrawModeE($0,$mode)|0); //@line 1773 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($1)) {
  HEAP32[$vararg_buffer>>2] = $0; //@line 1774 "modules/graphics/opengl/wrap_Graphics.cpp"
  $2 = (_luaL_error($L,330624,$vararg_buffer)|0); //@line 1774 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = $2;
  STACKTOP = sp;return ($$0|0); //@line 1787 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $3 = (+_luaL_checknumber($L,2)); //@line 1776 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = $3; //@line 1776 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = (+_luaL_checknumber($L,3)); //@line 1777 "modules/graphics/opengl/wrap_Graphics.cpp"
 $6 = $5; //@line 1777 "modules/graphics/opengl/wrap_Graphics.cpp"
 $7 = (+_luaL_checknumber($L,4)); //@line 1778 "modules/graphics/opengl/wrap_Graphics.cpp"
 $8 = $7; //@line 1778 "modules/graphics/opengl/wrap_Graphics.cpp"
 $9 = (_lua_type($L,5)|0); //@line 1780 "modules/graphics/opengl/wrap_Graphics.cpp"
 $10 = ($9|0)<(1); //@line 1780 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($10) {
  $11 = $8 > 10.0; //@line 1781 "modules/graphics/opengl/wrap_Graphics.cpp"
  $12 = (~~(($8))); //@line 1781 "modules/graphics/opengl/wrap_Graphics.cpp"
  $13 = $11 ? $12 : 10; //@line 1781 "modules/graphics/opengl/wrap_Graphics.cpp"
  $points$0 = $13;
 } else {
  $14 = (+_luaL_checknumber($L,5)); //@line 1783 "modules/graphics/opengl/wrap_Graphics.cpp"
  $15 = (~~(($14))); //@line 1783 "modules/graphics/opengl/wrap_Graphics.cpp"
  $points$0 = $15;
 }
 $16 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $17 = HEAP32[$mode>>2]|0; //@line 1785 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love8graphics6opengl8Graphics6circleENS0_8Graphics8DrawModeEfffi($16,$17,$4,$6,$8,$points$0); //@line 1785 "modules/graphics/opengl/wrap_Graphics.cpp"
 $$0 = 0;
 STACKTOP = sp;return ($$0|0); //@line 1787 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl9w_ellipseEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0;
 var $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $mode = 0, $points$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $mode = sp + 4|0;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 1792 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = (__ZN4love8graphics8Graphics11getConstantEPKcRNS1_8DrawModeE($0,$mode)|0); //@line 1793 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($1)) {
  HEAP32[$vararg_buffer>>2] = $0; //@line 1794 "modules/graphics/opengl/wrap_Graphics.cpp"
  $2 = (_luaL_error($L,330624,$vararg_buffer)|0); //@line 1794 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$0 = $2;
  STACKTOP = sp;return ($$0|0); //@line 1809 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $3 = (+_luaL_checknumber($L,2)); //@line 1796 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = $3; //@line 1796 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = (+_luaL_checknumber($L,3)); //@line 1797 "modules/graphics/opengl/wrap_Graphics.cpp"
 $6 = $5; //@line 1797 "modules/graphics/opengl/wrap_Graphics.cpp"
 $7 = (+_luaL_checknumber($L,4)); //@line 1798 "modules/graphics/opengl/wrap_Graphics.cpp"
 $8 = $7; //@line 1798 "modules/graphics/opengl/wrap_Graphics.cpp"
 $9 = $8; //@line 1799 "modules/graphics/opengl/wrap_Graphics.cpp"
 $10 = (+_luaL_optnumber($L,5,$9)); //@line 1799 "modules/graphics/opengl/wrap_Graphics.cpp"
 $11 = $10; //@line 1799 "modules/graphics/opengl/wrap_Graphics.cpp"
 $12 = (_lua_type($L,6)|0); //@line 1802 "modules/graphics/opengl/wrap_Graphics.cpp"
 $13 = ($12|0)<(1); //@line 1802 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($13) {
  $14 = $8 + $11; //@line 1803 "modules/graphics/opengl/wrap_Graphics.cpp"
  $15 = $14 > 30.0; //@line 1803 "modules/graphics/opengl/wrap_Graphics.cpp"
  $16 = $14 * 0.5; //@line 1803 "modules/graphics/opengl/wrap_Graphics.cpp"
  $17 = (~~(($16))); //@line 1803 "modules/graphics/opengl/wrap_Graphics.cpp"
  $18 = $15 ? $17 : 15; //@line 1803 "modules/graphics/opengl/wrap_Graphics.cpp"
  $points$0 = $18;
 } else {
  $19 = (+_luaL_checknumber($L,6)); //@line 1805 "modules/graphics/opengl/wrap_Graphics.cpp"
  $20 = (~~(($19))); //@line 1805 "modules/graphics/opengl/wrap_Graphics.cpp"
  $points$0 = $20;
 }
 $21 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $22 = HEAP32[$mode>>2]|0; //@line 1807 "modules/graphics/opengl/wrap_Graphics.cpp"
 __ZN4love8graphics6opengl8Graphics7ellipseENS0_8Graphics8DrawModeEffffi($21,$22,$4,$6,$8,$11,$points$0); //@line 1807 "modules/graphics/opengl/wrap_Graphics.cpp"
 $$0 = 0;
 STACKTOP = sp;return ($$0|0); //@line 1809 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl5w_arcEP9lua_State($L) {
 $L = $L|0;
 var $$1 = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0;
 var $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0.0;
 var $arcmode = 0, $drawmode = 0, $fabsf = 0.0, $startidx$1 = 0, $storemerge = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $drawmode = sp + 16|0;
 $arcmode = sp + 12|0;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 1814 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = (__ZN4love8graphics8Graphics11getConstantEPKcRNS1_8DrawModeE($0,$drawmode)|0); //@line 1815 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($1)) {
  HEAP32[$vararg_buffer>>2] = $0; //@line 1816 "modules/graphics/opengl/wrap_Graphics.cpp"
  $2 = (_luaL_error($L,330624,$vararg_buffer)|0); //@line 1816 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$2 = $2;
  STACKTOP = sp;return ($$2|0); //@line 1849 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 HEAP32[$arcmode>>2] = 2; //@line 1820 "modules/graphics/opengl/wrap_Graphics.cpp"
 $3 = (_lua_type($L,2)|0); //@line 1822 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = ($3|0)==(4); //@line 1822 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($4) {
  $5 = (_luaL_checklstring($L,2,0)|0); //@line 1824 "modules/graphics/opengl/wrap_Graphics.cpp"
  $6 = (__ZN4love8graphics8Graphics11getConstantEPKcRNS1_7ArcModeE($5,$arcmode)|0); //@line 1825 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($6) {
   $startidx$1 = 3;
   label = 6;
  } else {
   HEAP32[$vararg_buffer1>>2] = $5; //@line 1826 "modules/graphics/opengl/wrap_Graphics.cpp"
   $7 = (_luaL_error($L,330646,$vararg_buffer1)|0); //@line 1826 "modules/graphics/opengl/wrap_Graphics.cpp"
   $$1 = $7;
  }
 } else {
  $startidx$1 = 2;
  label = 6;
 }
 if ((label|0) == 6) {
  $8 = (+_luaL_checknumber($L,$startidx$1)); //@line 1831 "modules/graphics/opengl/wrap_Graphics.cpp"
  $9 = $8; //@line 1831 "modules/graphics/opengl/wrap_Graphics.cpp"
  $10 = (($startidx$1) + 1)|0; //@line 1832 "modules/graphics/opengl/wrap_Graphics.cpp"
  $11 = (+_luaL_checknumber($L,$10)); //@line 1832 "modules/graphics/opengl/wrap_Graphics.cpp"
  $12 = $11; //@line 1832 "modules/graphics/opengl/wrap_Graphics.cpp"
  $13 = (($startidx$1) + 2)|0; //@line 1833 "modules/graphics/opengl/wrap_Graphics.cpp"
  $14 = (+_luaL_checknumber($L,$13)); //@line 1833 "modules/graphics/opengl/wrap_Graphics.cpp"
  $15 = $14; //@line 1833 "modules/graphics/opengl/wrap_Graphics.cpp"
  $16 = (($startidx$1) + 3)|0; //@line 1834 "modules/graphics/opengl/wrap_Graphics.cpp"
  $17 = (+_luaL_checknumber($L,$16)); //@line 1834 "modules/graphics/opengl/wrap_Graphics.cpp"
  $18 = $17; //@line 1834 "modules/graphics/opengl/wrap_Graphics.cpp"
  $19 = $startidx$1 | 4; //@line 1835 "modules/graphics/opengl/wrap_Graphics.cpp"
  $20 = (+_luaL_checknumber($L,$19)); //@line 1835 "modules/graphics/opengl/wrap_Graphics.cpp"
  $21 = $20; //@line 1835 "modules/graphics/opengl/wrap_Graphics.cpp"
  $22 = (~~(($15))); //@line 1837 "modules/graphics/opengl/wrap_Graphics.cpp"
  $23 = $18 - $21; //@line 1838 "modules/graphics/opengl/wrap_Graphics.cpp"
  $fabsf = (+Math_abs((+$23))); //@line 1838 "modules/graphics/opengl/wrap_Graphics.cpp"
  $24 = $fabsf < 6.2831854820251465; //@line 1841 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($24) {
   $25 = $fabsf / 6.2831854820251465; //@line 1842 "modules/graphics/opengl/wrap_Graphics.cpp"
   $26 = (+($22|0)); //@line 1842 "modules/graphics/opengl/wrap_Graphics.cpp"
   $27 = $26 * $25; //@line 1842 "modules/graphics/opengl/wrap_Graphics.cpp"
   $28 = (~~(($27))); //@line 1842 "modules/graphics/opengl/wrap_Graphics.cpp"
   $storemerge = $28;
  } else {
   $storemerge = $22;
  }
  $29 = ($storemerge|0)<(10); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $30 = $29 ? 10 : $storemerge; //@line 1844 "modules/graphics/opengl/wrap_Graphics.cpp"
  $31 = (($startidx$1) + 5)|0; //@line 1845 "modules/graphics/opengl/wrap_Graphics.cpp"
  $32 = (+($30|0)); //@line 1845 "modules/graphics/opengl/wrap_Graphics.cpp"
  $33 = (+_luaL_optnumber($L,$31,$32)); //@line 1845 "modules/graphics/opengl/wrap_Graphics.cpp"
  $34 = (~~(($33))); //@line 1845 "modules/graphics/opengl/wrap_Graphics.cpp"
  $35 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  $36 = HEAP32[$drawmode>>2]|0; //@line 1847 "modules/graphics/opengl/wrap_Graphics.cpp"
  $37 = HEAP32[$arcmode>>2]|0; //@line 1847 "modules/graphics/opengl/wrap_Graphics.cpp"
  __ZN4love8graphics6opengl8Graphics3arcENS0_8Graphics8DrawModeENS3_7ArcModeEfffffi($35,$36,$37,$9,$12,$15,$18,$21,$34); //@line 1847 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$1 = 0;
 }
 $$2 = $$1;
 STACKTOP = sp;return ($$2|0); //@line 1849 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl9w_polygonEP9lua_State($L) {
 $L = $L|0;
 var $$1 = 0, $$arith = 0, $$overflow = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0.0, $23 = 0.0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $args$0 = 0, $exitcond = 0, $exitcond8 = 0;
 var $i$04 = 0, $i1$05 = 0, $is_table$0$off0 = 0, $mode = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $mode = sp + 20|0;
 $0 = (_lua_gettop($L)|0); //@line 1853 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = (($0) + -1)|0; //@line 1853 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = (_luaL_checklstring($L,1,0)|0); //@line 1856 "modules/graphics/opengl/wrap_Graphics.cpp"
 $3 = (__ZN4love8graphics8Graphics11getConstantEPKcRNS1_8DrawModeE($2,$mode)|0); //@line 1857 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($3)) {
  HEAP32[$vararg_buffer>>2] = $2; //@line 1858 "modules/graphics/opengl/wrap_Graphics.cpp"
  $4 = (_luaL_error($L,330624,$vararg_buffer)|0); //@line 1858 "modules/graphics/opengl/wrap_Graphics.cpp"
  $$1 = $4;
  STACKTOP = sp;return ($$1|0); //@line 1897 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $5 = ($1|0)==(1); //@line 1862 "modules/graphics/opengl/wrap_Graphics.cpp"
 if ($5) {
  $6 = (_lua_type($L,2)|0); //@line 1862 "modules/graphics/opengl/wrap_Graphics.cpp"
  $7 = ($6|0)==(5); //@line 1862 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($7) {
   $8 = (__ZN4love11luax_objlenEP9lua_Statei($L,2)|0); //@line 1864 "modules/graphics/opengl/wrap_Graphics.cpp"
   $args$0 = $8;$is_table$0$off0 = 1;
   label = 6;
  }
 } else {
  $args$0 = $1;$is_table$0$off0 = 0;
  label = 6;
 }
 if ((label|0) == 6) {
  $9 = $args$0 & 1; //@line 1868 "modules/graphics/opengl/wrap_Graphics.cpp"
  $10 = ($9|0)==(0); //@line 1868 "modules/graphics/opengl/wrap_Graphics.cpp"
  if ($10) {
   $12 = ($args$0|0)<(6); //@line 1870 "modules/graphics/opengl/wrap_Graphics.cpp"
   if ($12) {
    $13 = (_luaL_error($L,330667,$vararg_buffer3)|0); //@line 1871 "modules/graphics/opengl/wrap_Graphics.cpp"
    $$1 = $13;
    STACKTOP = sp;return ($$1|0); //@line 1897 "modules/graphics/opengl/wrap_Graphics.cpp"
   }
   $14 = (($args$0) + 2)|0; //@line 1874 "modules/graphics/opengl/wrap_Graphics.cpp"
   $$arith = $14<<2; //@line 1874 "modules/graphics/opengl/wrap_Graphics.cpp"
   $$overflow = ($14>>>0)>(1073741823); //@line 1874 "modules/graphics/opengl/wrap_Graphics.cpp"
   $15 = $$overflow ? -1 : $$arith; //@line 1874 "modules/graphics/opengl/wrap_Graphics.cpp"
   $16 = (__Znaj($15)|0); //@line 1874 "modules/graphics/opengl/wrap_Graphics.cpp"
   if ($is_table$0$off0) {
    $i$04 = 0;
    while(1) {
     $17 = (($i$04) + 1)|0; //@line 1879 "modules/graphics/opengl/wrap_Graphics.cpp"
     _lua_rawgeti($L,2,$17); //@line 1879 "modules/graphics/opengl/wrap_Graphics.cpp"
     $18 = (+_lua_tonumber($L,-1)); //@line 171 "./common/runtime.h"
     $19 = $18; //@line 171 "./common/runtime.h"
     $20 = (($16) + ($i$04<<2)|0); //@line 1880 "modules/graphics/opengl/wrap_Graphics.cpp"
     HEAPF32[$20>>2] = $19; //@line 1880 "modules/graphics/opengl/wrap_Graphics.cpp"
     _lua_settop($L,-2); //@line 1881 "modules/graphics/opengl/wrap_Graphics.cpp"
     $exitcond = ($17|0)==($args$0|0); //@line 1877 "modules/graphics/opengl/wrap_Graphics.cpp"
     if ($exitcond) {
      break;
     } else {
      $i$04 = $17;
     }
    }
   } else {
    $i1$05 = 0;
    while(1) {
     $21 = (($i1$05) + 2)|0; //@line 1887 "modules/graphics/opengl/wrap_Graphics.cpp"
     $22 = (+_lua_tonumber($L,$21)); //@line 171 "./common/runtime.h"
     $23 = $22; //@line 171 "./common/runtime.h"
     $24 = (($16) + ($i1$05<<2)|0); //@line 1887 "modules/graphics/opengl/wrap_Graphics.cpp"
     HEAPF32[$24>>2] = $23; //@line 1887 "modules/graphics/opengl/wrap_Graphics.cpp"
     $25 = (($i1$05) + 1)|0; //@line 1886 "modules/graphics/opengl/wrap_Graphics.cpp"
     $exitcond8 = ($25|0)==($args$0|0); //@line 1886 "modules/graphics/opengl/wrap_Graphics.cpp"
     if ($exitcond8) {
      break;
     } else {
      $i1$05 = $25;
     }
    }
   }
   $26 = HEAP32[$16>>2]|0; //@line 1891 "modules/graphics/opengl/wrap_Graphics.cpp"
   $27 = (($16) + ($args$0<<2)|0); //@line 1891 "modules/graphics/opengl/wrap_Graphics.cpp"
   HEAP32[$27>>2] = $26; //@line 1891 "modules/graphics/opengl/wrap_Graphics.cpp"
   $28 = ((($16)) + 4|0); //@line 1892 "modules/graphics/opengl/wrap_Graphics.cpp"
   $29 = HEAP32[$28>>2]|0; //@line 1892 "modules/graphics/opengl/wrap_Graphics.cpp"
   $30 = (($args$0) + 1)|0; //@line 1892 "modules/graphics/opengl/wrap_Graphics.cpp"
   $31 = (($16) + ($30<<2)|0); //@line 1892 "modules/graphics/opengl/wrap_Graphics.cpp"
   HEAP32[$31>>2] = $29; //@line 1892 "modules/graphics/opengl/wrap_Graphics.cpp"
   $32 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
   $33 = HEAP32[$mode>>2]|0; //@line 1893 "modules/graphics/opengl/wrap_Graphics.cpp"
   __ZN4love8graphics6opengl8Graphics7polygonENS0_8Graphics8DrawModeEPKfj($32,$33,$16,$14); //@line 1893 "modules/graphics/opengl/wrap_Graphics.cpp"
   __ZdaPv($16); //@line 1894 "modules/graphics/opengl/wrap_Graphics.cpp"
   $$1 = 0;
   STACKTOP = sp;return ($$1|0); //@line 1897 "modules/graphics/opengl/wrap_Graphics.cpp"
  }
 }
 $11 = (_luaL_error($L,330528,$vararg_buffer1)|0); //@line 1869 "modules/graphics/opengl/wrap_Graphics.cpp"
 $$1 = $11;
 STACKTOP = sp;return ($$1|0); //@line 1897 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl6w_pushEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$02$i = 0, $$idx$val$val = 0, $$idx$val$val$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $stype = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $stype = sp + 12|0;
 HEAP32[$stype>>2] = 1; //@line 1901 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = (_lua_type($L,1)|0); //@line 1902 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = ($0|0)<(1); //@line 1902 "modules/graphics/opengl/wrap_Graphics.cpp"
 do {
  if ($1) {
   $$idx$val$val = 1;
  } else {
   $2 = (_luaL_checklstring($L,1,0)|0); //@line 1902 "modules/graphics/opengl/wrap_Graphics.cpp"
   $3 = ($2|0)==(0|0); //@line 1903 "modules/graphics/opengl/wrap_Graphics.cpp"
   if ($3) {
    $$idx$val$val = 1;
   } else {
    $4 = (__ZN4love8graphics8Graphics11getConstantEPKcRNS1_9StackTypeE($2,$stype)|0); //@line 1903 "modules/graphics/opengl/wrap_Graphics.cpp"
    if ($4) {
     $$idx$val$val$pre = HEAP32[$stype>>2]|0;
     $$idx$val$val = $$idx$val$val$pre;
     break;
    }
    HEAP32[$vararg_buffer>>2] = $2; //@line 1904 "modules/graphics/opengl/wrap_Graphics.cpp"
    $5 = (_luaL_error($L,330714,$vararg_buffer)|0); //@line 1904 "modules/graphics/opengl/wrap_Graphics.cpp"
    $$0 = $5;
    STACKTOP = sp;return ($$0|0); //@line 1908 "modules/graphics/opengl/wrap_Graphics.cpp"
   }
  }
 } while(0);
 $6 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 invoke_vii(3583,($6|0),($$idx$val$val|0)); //@line 1906 "modules/graphics/opengl/wrap_Graphics.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if (!($8)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 1908 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $9 = ___cxa_find_matching_catch_3(8992|0)|0;
 $10 = tempRet0;
 $11 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $12 = ($10|0)==($11|0); //@line 535 "./common/runtime.h"
 if (!($12)) {
  $$01$i = $10;$$02$i = $9;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $13 = (___cxa_begin_catch(($9|0))|0); //@line 535 "./common/runtime.h"
 $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
 $15 = ((($14)) + 8|0); //@line 539 "./common/runtime.h"
 $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
 $17 = (FUNCTION_TABLE_ii[$16 & 8191]($13)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($17|0)); //@line 539 "./common/runtime.h"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if (!($19)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $20 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $20; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 1908 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $21 = ___cxa_find_matching_catch_2()|0;
 $22 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $23 = __THREW__; __THREW__ = 0;
 $24 = $23&1;
 if ($24) {
  $25 = ___cxa_find_matching_catch_3(0|0)|0;
  $26 = tempRet0;
  ___clang_call_terminate($25); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $22;$$02$i = $21;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl5w_popEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 1910 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __THREW__ = 0;
 invoke_vi(3584,($0|0)); //@line 1912 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if (!($2)) {
  STACKTOP = sp;return 0; //@line 1913 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $3 = ___cxa_find_matching_catch_3(8992|0)|0;
 $4 = tempRet0;
 $5 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $6 = ($4|0)==($5|0); //@line 535 "./common/runtime.h"
 if (!($6)) {
  $$01$i = $4;$$02$i = $3;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $7 = (___cxa_begin_catch(($3|0))|0); //@line 535 "./common/runtime.h"
 $8 = HEAP32[$7>>2]|0; //@line 539 "./common/runtime.h"
 $9 = ((($8)) + 8|0); //@line 539 "./common/runtime.h"
 $10 = HEAP32[$9>>2]|0; //@line 539 "./common/runtime.h"
 $11 = (FUNCTION_TABLE_ii[$10 & 8191]($7)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($11|0)); //@line 539 "./common/runtime.h"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if (!($13)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $14 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $14; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 1913 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 $15 = ___cxa_find_matching_catch_2()|0;
 $16 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if ($18) {
  $19 = ___cxa_find_matching_catch_3(0|0)|0;
  $20 = tempRet0;
  ___clang_call_terminate($19); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $16;$$02$i = $15;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl8w_rotateEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1)); //@line 1918 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = $0; //@line 1918 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZN4love8graphics6opengl8Graphics6rotateEf($2,$1); //@line 1919 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 0; //@line 1920 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl7w_scaleEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_optnumber($L,1,1.0)); //@line 1925 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = $0; //@line 1925 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = $1; //@line 1926 "modules/graphics/opengl/wrap_Graphics.cpp"
 $3 = (+_luaL_optnumber($L,2,$2)); //@line 1926 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = $3; //@line 1926 "modules/graphics/opengl/wrap_Graphics.cpp"
 $5 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZN4love8graphics6opengl8Graphics5scaleEff($5,$1,$4); //@line 1927 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 0; //@line 1928 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl11w_translateEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1)); //@line 1933 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = $0; //@line 1933 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = (+_luaL_checknumber($L,2)); //@line 1934 "modules/graphics/opengl/wrap_Graphics.cpp"
 $3 = $2; //@line 1934 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZN4love8graphics6opengl8Graphics9translateEff($4,$1,$3); //@line 1935 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 0; //@line 1936 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl7w_shearEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1)); //@line 1941 "modules/graphics/opengl/wrap_Graphics.cpp"
 $1 = $0; //@line 1941 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = (+_luaL_checknumber($L,2)); //@line 1942 "modules/graphics/opengl/wrap_Graphics.cpp"
 $3 = $2; //@line 1942 "modules/graphics/opengl/wrap_Graphics.cpp"
 $4 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZN4love8graphics6opengl8Graphics5shearEff($4,$1,$3); //@line 1943 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 0; //@line 1944 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl8w_originEP9lua_State($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 __ZN4love8graphics6opengl8Graphics6originEv($1); //@line 1949 "modules/graphics/opengl/wrap_Graphics.cpp"
 return 0; //@line 1950 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function _luaopen_love_graphics($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $instance$sroa$0$1 = 0, $vararg_buffer = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $w = sp + 4|0;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = $0; //@line 2081 "modules/graphics/opengl/wrap_Graphics.cpp"
 $2 = ($0|0)==(0|0); //@line 2082 "modules/graphics/opengl/wrap_Graphics.cpp"
 L1: do {
  if ($2) {
   __THREW__ = 0;
   $3 = (invoke_ii(3348,68)|0); //@line 2084 "modules/graphics/opengl/wrap_Graphics.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   do {
    if ($5) {
     $11 = ___cxa_find_matching_catch_3(8992|0)|0;
     $12 = tempRet0;
     $eh$lpad$body$i$index2Z2D = $12;$eh$lpad$body$i$indexZ2D = $11;
    } else {
     __THREW__ = 0;
     invoke_vi(3585,($3|0)); //@line 2084 "modules/graphics/opengl/wrap_Graphics.cpp"
     $6 = __THREW__; __THREW__ = 0;
     $7 = $6&1;
     if ($7) {
      $8 = ___cxa_find_matching_catch_3(8992|0)|0;
      $9 = tempRet0;
      __ZdlPv($3); //@line 2084 "modules/graphics/opengl/wrap_Graphics.cpp"
      $eh$lpad$body$i$index2Z2D = $9;$eh$lpad$body$i$indexZ2D = $8;
      break;
     } else {
      $10 = $3; //@line 2084 "modules/graphics/opengl/wrap_Graphics.cpp"
      $instance$sroa$0$1 = $10;
      break L1;
     }
    }
   } while(0);
   $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $14 = ($eh$lpad$body$i$index2Z2D|0)==($13|0); //@line 535 "./common/runtime.h"
   if (!($14)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $15 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
   $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
   $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
   $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $instance$sroa$0$1 = $1;
    break;
   }
   $23 = ___cxa_find_matching_catch_2()|0;
   $24 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    $27 = ___cxa_find_matching_catch_3(0|0)|0;
    $28 = tempRet0;
    ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $24;$$02$i = $23;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $29 = HEAP32[$0>>2]|0; //@line 2087 "modules/graphics/opengl/wrap_Graphics.cpp"
   $30 = ((($29)) + 8|0); //@line 2087 "modules/graphics/opengl/wrap_Graphics.cpp"
   $31 = HEAP32[$30>>2]|0; //@line 2087 "modules/graphics/opengl/wrap_Graphics.cpp"
   FUNCTION_TABLE_vi[$31 & 8191]($0); //@line 2087 "modules/graphics/opengl/wrap_Graphics.cpp"
   $instance$sroa$0$1 = $1;
  }
 } while(0);
 HEAP32[$w>>2] = $instance$sroa$0$1; //@line 2090 "modules/graphics/opengl/wrap_Graphics.cpp"
 $32 = ((($w)) + 4|0); //@line 2091 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAP32[$32>>2] = 329707; //@line 2091 "modules/graphics/opengl/wrap_Graphics.cpp"
 $33 = ((($w)) + 8|0); //@line 2092 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAP32[$33>>2] = 57; //@line 2092 "modules/graphics/opengl/wrap_Graphics.cpp"
 $34 = ((($w)) + 12|0); //@line 2093 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAP32[$34>>2] = 13404; //@line 2093 "modules/graphics/opengl/wrap_Graphics.cpp"
 $35 = ((($w)) + 16|0); //@line 2094 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAP32[$35>>2] = 14060; //@line 2094 "modules/graphics/opengl/wrap_Graphics.cpp"
 $36 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 2096 "modules/graphics/opengl/wrap_Graphics.cpp"
 $37 = (_luaL_loadbuffer($L,331533,11561,343094)|0); //@line 2098 "modules/graphics/opengl/wrap_Graphics.cpp"
 $38 = ($37|0)==(0); //@line 2098 "modules/graphics/opengl/wrap_Graphics.cpp"
 if (!($38)) {
  STACKTOP = sp;return ($36|0); //@line 2101 "modules/graphics/opengl/wrap_Graphics.cpp"
 }
 _lua_call($L,0,0); //@line 2099 "modules/graphics/opengl/wrap_Graphics.cpp"
 STACKTOP = sp;return ($36|0); //@line 2101 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6openglL16luaopen_drawableEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 2056 "modules/graphics/opengl/wrap_Graphics.cpp"
 HEAP32[$vararg_buffer>>2] = 0; //@line 2058 "modules/graphics/opengl/wrap_Graphics.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,10,331524,$vararg_buffer)|0); //@line 2058 "modules/graphics/opengl/wrap_Graphics.cpp"
 STACKTOP = sp;return ($0|0); //@line 2058 "modules/graphics/opengl/wrap_Graphics.cpp"
}
function __ZN4love8graphics6opengl23w_Image_setMipmapFilterEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $f = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $f = sp + 16|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl5ImageEEEPT_P9lua_StateiNS_4TypeE($L,1,12)|0); //@line 33 "modules/graphics/opengl/wrap_Image.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 39 "modules/graphics/opengl/wrap_Image.cpp"
 $2 = ((($1)) + 36|0); //@line 39 "modules/graphics/opengl/wrap_Image.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 39 "modules/graphics/opengl/wrap_Image.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 39 "modules/graphics/opengl/wrap_Image.cpp"
 ;HEAP32[$f>>2]=HEAP32[$4>>2]|0;HEAP32[$f+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$f+8>>2]=HEAP32[$4+8>>2]|0;HEAP32[$f+12>>2]=HEAP32[$4+12>>2]|0; //@line 39 "modules/graphics/opengl/wrap_Image.cpp"
 $5 = (_lua_type($L,2)|0); //@line 41 "modules/graphics/opengl/wrap_Image.cpp"
 $6 = ($5|0)<(1); //@line 41 "modules/graphics/opengl/wrap_Image.cpp"
 if ($6) {
  $7 = ((($f)) + 8|0); //@line 42 "modules/graphics/opengl/wrap_Image.cpp"
  HEAP32[$7>>2] = 0; //@line 42 "modules/graphics/opengl/wrap_Image.cpp"
 } else {
  $8 = (_luaL_checklstring($L,2,0)|0); //@line 45 "modules/graphics/opengl/wrap_Image.cpp"
  $9 = ((($f)) + 8|0); //@line 46 "modules/graphics/opengl/wrap_Image.cpp"
  $10 = (__ZN4love8graphics7Texture11getConstantEPKcRNS1_10FilterModeE($8,$9)|0); //@line 46 "modules/graphics/opengl/wrap_Image.cpp"
  if (!($10)) {
   HEAP32[$vararg_buffer>>2] = $8; //@line 47 "modules/graphics/opengl/wrap_Image.cpp"
   $11 = (_luaL_error($L,439370,$vararg_buffer)|0); //@line 47 "modules/graphics/opengl/wrap_Image.cpp"
   $$1 = $11;
   STACKTOP = sp;return ($$1|0); //@line 54 "modules/graphics/opengl/wrap_Image.cpp"
  }
 }
 $12 = HEAP32[$0>>2]|0; //@line 50 "modules/graphics/opengl/wrap_Image.cpp"
 $13 = ((($12)) + 32|0); //@line 50 "modules/graphics/opengl/wrap_Image.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 50 "modules/graphics/opengl/wrap_Image.cpp"
 __THREW__ = 0;
 invoke_vii($14|0,($0|0),($f|0)); //@line 50 "modules/graphics/opengl/wrap_Image.cpp"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 do {
  if ($16) {
   $17 = ___cxa_find_matching_catch_3(8992|0)|0;
   $18 = tempRet0;
   $19 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $20 = ($18|0)==($19|0); //@line 535 "./common/runtime.h"
   if (!($20)) {
    $$01$i = $18;$$02$i = $17;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $21 = (___cxa_begin_catch(($17|0))|0); //@line 535 "./common/runtime.h"
   $22 = HEAP32[$21>>2]|0; //@line 539 "./common/runtime.h"
   $23 = ((($22)) + 8|0); //@line 539 "./common/runtime.h"
   $24 = HEAP32[$23>>2]|0; //@line 539 "./common/runtime.h"
   $25 = (FUNCTION_TABLE_ii[$24 & 8191]($21)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($25|0)); //@line 539 "./common/runtime.h"
   $26 = __THREW__; __THREW__ = 0;
   $27 = $26&1;
   if (!($27)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $28 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer1>>2] = $28; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
    break;
   }
   $29 = ___cxa_find_matching_catch_2()|0;
   $30 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $31 = __THREW__; __THREW__ = 0;
   $32 = $31&1;
   if ($32) {
    $33 = ___cxa_find_matching_catch_3(0|0)|0;
    $34 = tempRet0;
    ___clang_call_terminate($33); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $30;$$02$i = $29;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  }
 } while(0);
 $35 = (+_luaL_optnumber($L,3,0.0)); //@line 51 "modules/graphics/opengl/wrap_Image.cpp"
 $36 = $35; //@line 51 "modules/graphics/opengl/wrap_Image.cpp"
 __ZN4love8graphics6opengl5Image18setMipmapSharpnessEf($0,$36); //@line 51 "modules/graphics/opengl/wrap_Image.cpp"
 $$1 = 0;
 STACKTOP = sp;return ($$1|0); //@line 54 "modules/graphics/opengl/wrap_Image.cpp"
}
function __ZN4love8graphics6opengl23w_Image_getMipmapFilterEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, $mipmapstr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $mipmapstr = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl5ImageEEEPT_P9lua_StateiNS_4TypeE($L,1,12)|0); //@line 33 "modules/graphics/opengl/wrap_Image.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 60 "modules/graphics/opengl/wrap_Image.cpp"
 $2 = ((($1)) + 36|0); //@line 60 "modules/graphics/opengl/wrap_Image.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 60 "modules/graphics/opengl/wrap_Image.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 60 "modules/graphics/opengl/wrap_Image.cpp"
 $5 = ((($4)) + 8|0); //@line 63 "modules/graphics/opengl/wrap_Image.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 63 "modules/graphics/opengl/wrap_Image.cpp"
 $7 = (__ZN4love8graphics7Texture11getConstantENS1_10FilterModeERPKc($6,$mipmapstr)|0); //@line 63 "modules/graphics/opengl/wrap_Image.cpp"
 if ($7) {
  $8 = HEAP32[$mipmapstr>>2]|0; //@line 64 "modules/graphics/opengl/wrap_Image.cpp"
  _lua_pushstring($L,$8); //@line 64 "modules/graphics/opengl/wrap_Image.cpp"
 } else {
  _lua_pushnil($L); //@line 66 "modules/graphics/opengl/wrap_Image.cpp"
 }
 $9 = (+__ZNK4love8graphics6opengl5Image18getMipmapSharpnessEv($0)); //@line 68 "modules/graphics/opengl/wrap_Image.cpp"
 $10 = $9; //@line 68 "modules/graphics/opengl/wrap_Image.cpp"
 _lua_pushnumber($L,$10); //@line 68 "modules/graphics/opengl/wrap_Image.cpp"
 STACKTOP = sp;return 2; //@line 69 "modules/graphics/opengl/wrap_Image.cpp"
}
function __ZN4love8graphics6opengl20w_Image_isCompressedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl5ImageEEEPT_P9lua_StateiNS_4TypeE($L,1,12)|0); //@line 33 "modules/graphics/opengl/wrap_Image.cpp"
 $1 = (__ZNK4love8graphics6opengl5Image12isCompressedEv($0)|0); //@line 75 "modules/graphics/opengl/wrap_Image.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$1); //@line 75 "modules/graphics/opengl/wrap_Image.cpp"
 return 1; //@line 76 "modules/graphics/opengl/wrap_Image.cpp"
}
function __ZN4love8graphics6opengl15w_Image_refreshEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 79 "modules/graphics/opengl/wrap_Image.cpp"
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl5ImageEEEPT_P9lua_StateiNS_4TypeE($L,1,12)|0); //@line 33 "modules/graphics/opengl/wrap_Image.cpp"
 $1 = (+_luaL_optnumber($L,2,0.0)); //@line 83 "modules/graphics/opengl/wrap_Image.cpp"
 $2 = (~~(($1))); //@line 83 "modules/graphics/opengl/wrap_Image.cpp"
 $3 = (+_luaL_optnumber($L,3,0.0)); //@line 84 "modules/graphics/opengl/wrap_Image.cpp"
 $4 = (~~(($3))); //@line 84 "modules/graphics/opengl/wrap_Image.cpp"
 $5 = HEAP32[$0>>2]|0; //@line 85 "modules/graphics/opengl/wrap_Image.cpp"
 $6 = ((($5)) + 24|0); //@line 85 "modules/graphics/opengl/wrap_Image.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 85 "modules/graphics/opengl/wrap_Image.cpp"
 $8 = (FUNCTION_TABLE_ii[$7 & 8191]($0)|0); //@line 85 "modules/graphics/opengl/wrap_Image.cpp"
 $9 = (+($8|0)); //@line 85 "modules/graphics/opengl/wrap_Image.cpp"
 $10 = (+_luaL_optnumber($L,4,$9)); //@line 85 "modules/graphics/opengl/wrap_Image.cpp"
 $11 = (~~(($10))); //@line 85 "modules/graphics/opengl/wrap_Image.cpp"
 $12 = HEAP32[$0>>2]|0; //@line 86 "modules/graphics/opengl/wrap_Image.cpp"
 $13 = ((($12)) + 28|0); //@line 86 "modules/graphics/opengl/wrap_Image.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 86 "modules/graphics/opengl/wrap_Image.cpp"
 $15 = (FUNCTION_TABLE_ii[$14 & 8191]($0)|0); //@line 86 "modules/graphics/opengl/wrap_Image.cpp"
 $16 = (+($15|0)); //@line 86 "modules/graphics/opengl/wrap_Image.cpp"
 $17 = (+_luaL_optnumber($L,5,$16)); //@line 86 "modules/graphics/opengl/wrap_Image.cpp"
 $18 = (~~(($17))); //@line 86 "modules/graphics/opengl/wrap_Image.cpp"
 __THREW__ = 0;
 (invoke_iiiiii(3586,($0|0),($2|0),($4|0),($11|0),($18|0))|0); //@line 88 "modules/graphics/opengl/wrap_Image.cpp"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if (!($20)) {
  STACKTOP = sp;return 0; //@line 89 "modules/graphics/opengl/wrap_Image.cpp"
 }
 $21 = ___cxa_find_matching_catch_3(8992|0)|0;
 $22 = tempRet0;
 $23 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $24 = ($22|0)==($23|0); //@line 535 "./common/runtime.h"
 if (!($24)) {
  $$01$i = $22;$$02$i = $21;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $25 = (___cxa_begin_catch(($21|0))|0); //@line 535 "./common/runtime.h"
 $26 = HEAP32[$25>>2]|0; //@line 539 "./common/runtime.h"
 $27 = ((($26)) + 8|0); //@line 539 "./common/runtime.h"
 $28 = HEAP32[$27>>2]|0; //@line 539 "./common/runtime.h"
 $29 = (FUNCTION_TABLE_ii[$28 & 8191]($25)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($29|0)); //@line 539 "./common/runtime.h"
 $30 = __THREW__; __THREW__ = 0;
 $31 = $30&1;
 if (!($31)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $32 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $32; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 89 "modules/graphics/opengl/wrap_Image.cpp"
 }
 $33 = ___cxa_find_matching_catch_2()|0;
 $34 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $35 = __THREW__; __THREW__ = 0;
 $36 = $35&1;
 if ($36) {
  $37 = ___cxa_find_matching_catch_3(0|0)|0;
  $38 = tempRet0;
  ___clang_call_terminate($37); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $34;$$02$i = $33;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl15w_Image_getDataEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$02 = 0, $__begin2$sroa$0$04 = 0, $n$2 = 0, $scevgep = 0, $scevgep12 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl5ImageEEEPT_P9lua_StateiNS_4TypeE($L,1,12)|0); //@line 33 "modules/graphics/opengl/wrap_Image.cpp"
 $1 = (__ZNK4love8graphics6opengl5Image12isCompressedEv($0)|0); //@line 97 "modules/graphics/opengl/wrap_Image.cpp"
 if ($1) {
  $2 = (__ZNK4love8graphics6opengl5Image17getCompressedDataEv($0)|0); //@line 99 "modules/graphics/opengl/wrap_Image.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = ((($2)) + 4|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $5 = HEAP32[$4>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $6 = ($3|0)==($5|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($6) {
   $n$2 = 0;
   return ($n$2|0); //@line 114 "modules/graphics/opengl/wrap_Image.cpp"
  }
  $7 = $3;
  $scevgep = ((($5)) + -4|0);
  $8 = $scevgep;
  $9 = (($8) - ($7))|0;
  $10 = $9 >>> 2;
  $__begin$sroa$0$02 = $3;
  while(1) {
   $11 = HEAP32[$__begin$sroa$0$02>>2]|0; //@line 139 "./common/Object.h"
   __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,23,$11); //@line 101 "modules/graphics/opengl/wrap_Image.cpp"
   $12 = ((($__begin$sroa$0$02)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $13 = ($12|0)==($5|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($13) {
    break;
   } else {
    $__begin$sroa$0$02 = $12;
   }
  }
  $26 = (($10) + 1)|0;
  $n$2 = $26;
  return ($n$2|0); //@line 114 "modules/graphics/opengl/wrap_Image.cpp"
 } else {
  $14 = (__ZNK4love8graphics6opengl5Image12getImageDataEv($0)|0); //@line 107 "modules/graphics/opengl/wrap_Image.cpp"
  $15 = HEAP32[$14>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ((($14)) + 4|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = HEAP32[$16>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ($15|0)==($17|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($18) {
   $n$2 = 0;
   return ($n$2|0); //@line 114 "modules/graphics/opengl/wrap_Image.cpp"
  }
  $19 = $15;
  $scevgep12 = ((($17)) + -4|0);
  $20 = $scevgep12;
  $21 = (($20) - ($19))|0;
  $22 = $21 >>> 2;
  $__begin2$sroa$0$04 = $15;
  while(1) {
   $23 = HEAP32[$__begin2$sroa$0$04>>2]|0; //@line 139 "./common/Object.h"
   __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,22,$23); //@line 109 "modules/graphics/opengl/wrap_Image.cpp"
   $24 = ((($__begin2$sroa$0$04)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $25 = ($24|0)==($17|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($25) {
    break;
   } else {
    $__begin2$sroa$0$04 = $24;
   }
  }
  $27 = (($22) + 1)|0;
  $n$2 = $27;
  return ($n$2|0); //@line 114 "modules/graphics/opengl/wrap_Image.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl16w_Image_getFlagsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name$i = 0, $name$i1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name$i1 = sp + 4|0;
 $name$i = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl5ImageEEEPT_P9lua_StateiNS_4TypeE($L,1,12)|0); //@line 33 "modules/graphics/opengl/wrap_Image.cpp"
 $1 = (__ZNK4love8graphics6opengl5Image8getFlagsEv($0)|0); //@line 127 "modules/graphics/opengl/wrap_Image.cpp"
 $2 = HEAP8[$1>>0]|0; //@line 127 "modules/graphics/opengl/wrap_Image.cpp"
 $3 = ((($1)) + 1|0); //@line 127 "modules/graphics/opengl/wrap_Image.cpp"
 $4 = HEAP8[$3>>0]|0; //@line 127 "modules/graphics/opengl/wrap_Image.cpp"
 _lua_createtable($L,0,2); //@line 129 "modules/graphics/opengl/wrap_Image.cpp"
 $5 = $2 & 1; //@line 131 "modules/graphics/opengl/wrap_Image.cpp"
 $6 = $5&255; //@line 131 "modules/graphics/opengl/wrap_Image.cpp"
 _lua_pushboolean($L,$6); //@line 131 "modules/graphics/opengl/wrap_Image.cpp"
 HEAP32[$name$i>>2] = 0; //@line 119 "modules/graphics/opengl/wrap_Image.cpp"
 (__ZN4love8graphics6opengl5Image11getConstantENS2_8FlagTypeERPKc(0,$name$i)|0); //@line 120 "modules/graphics/opengl/wrap_Image.cpp"
 $7 = HEAP32[$name$i>>2]|0; //@line 121 "modules/graphics/opengl/wrap_Image.cpp"
 _lua_setfield($L,-2,$7); //@line 132 "modules/graphics/opengl/wrap_Image.cpp"
 $8 = $4 & 1; //@line 134 "modules/graphics/opengl/wrap_Image.cpp"
 $9 = $8&255; //@line 134 "modules/graphics/opengl/wrap_Image.cpp"
 _lua_pushboolean($L,$9); //@line 134 "modules/graphics/opengl/wrap_Image.cpp"
 HEAP32[$name$i1>>2] = 0; //@line 119 "modules/graphics/opengl/wrap_Image.cpp"
 (__ZN4love8graphics6opengl5Image11getConstantENS2_8FlagTypeERPKc(1,$name$i1)|0); //@line 120 "modules/graphics/opengl/wrap_Image.cpp"
 $10 = HEAP32[$name$i1>>2]|0; //@line 121 "modules/graphics/opengl/wrap_Image.cpp"
 _lua_setfield($L,-2,$10); //@line 135 "modules/graphics/opengl/wrap_Image.cpp"
 STACKTOP = sp;return 1; //@line 137 "modules/graphics/opengl/wrap_Image.cpp"
}
function _luaopen_image($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 151 "modules/graphics/opengl/wrap_Image.cpp"
 HEAP32[$vararg_buffer>>2] = 15812; //@line 153 "modules/graphics/opengl/wrap_Image.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 153 "modules/graphics/opengl/wrap_Image.cpp"
 HEAP32[$vararg_ptr1>>2] = 14112; //@line 153 "modules/graphics/opengl/wrap_Image.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 153 "modules/graphics/opengl/wrap_Image.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 153 "modules/graphics/opengl/wrap_Image.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,12,343112,$vararg_buffer)|0); //@line 153 "modules/graphics/opengl/wrap_Image.cpp"
 STACKTOP = sp;return ($0|0); //@line 153 "modules/graphics/opengl/wrap_Image.cpp"
}
function __ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love8graphics6opengl23luax_writeAttributeDataEP9lua_StateiNS1_4Mesh8DataTypeEiPc($L,$startidx,$type,$components,$data) {
 $L = $L|0;
 $startidx = $startidx|0;
 $type = $type|0;
 $components = $components|0;
 $data = $data|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0.0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, $exitcond$i = 0, $exitcond$i2 = 0, $i$01$i = 0, $i$01$i1 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 switch ($type|0) {
 case 0:  {
  $0 = ($components|0)>(0); //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
  if ($0) {
   $i$01$i = 0;
   while(1) {
    $1 = (($i$01$i) + ($startidx))|0; //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
    $2 = (+_luaL_optnumber($L,$1,255.0)); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
    $3 = (~~(($2))&255); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
    $4 = (($data) + ($i$01$i)|0); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
    HEAP8[$4>>0] = $3; //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
    $5 = (($i$01$i) + 1)|0; //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
    $exitcond$i = ($5|0)==($components|0); //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
    if ($exitcond$i) {
     break;
    } else {
     $i$01$i = $5;
    }
   }
  }
  $6 = (($data) + ($components)|0); //@line 68 "modules/graphics/opengl/wrap_Mesh.cpp"
  $$0 = $6;
  return ($$0|0); //@line 74 "modules/graphics/opengl/wrap_Mesh.cpp"
  break;
 }
 case 1:  {
  $7 = ($components|0)>(0); //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
  if ($7) {
   $i$01$i1 = 0;
   while(1) {
    $8 = (($i$01$i1) + ($startidx))|0; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
    $9 = (+_luaL_optnumber($L,$8,0.0)); //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
    $10 = $9; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
    $11 = (($data) + ($i$01$i1<<2)|0); //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
    HEAPF32[$11>>2] = $10; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
    $12 = (($i$01$i1) + 1)|0; //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
    $exitcond$i2 = ($12|0)==($components|0); //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
    if ($exitcond$i2) {
     break;
    } else {
     $i$01$i1 = $12;
    }
   }
  }
  $13 = $components << 2; //@line 60 "modules/graphics/opengl/wrap_Mesh.cpp"
  $14 = (($data) + ($13)|0); //@line 70 "modules/graphics/opengl/wrap_Mesh.cpp"
  $$0 = $14;
  return ($$0|0); //@line 74 "modules/graphics/opengl/wrap_Mesh.cpp"
  break;
 }
 default: {
  $$0 = $data;
  return ($$0|0); //@line 74 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl18w_Mesh_setVerticesEP9lua_State($L) {
 $L = $L|0;
 var $$0$i = 0, $$0$i$us = 0, $$2 = 0, $0 = 0, $1 = 0.0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0.0, $56 = 0.0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0.0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0.0, $98 = 0.0, $99 = 0, $__begin$sroa$0$014 = 0, $__begin2$sroa$0$02 = 0, $__begin2$sroa$0$02$us = 0, $data$09 = 0, $data$09$us = 0, $data$1$lcssa = 0, $data$1$lcssa$us = 0, $data$13 = 0, $data$13$us = 0, $exitcond = 0;
 var $exitcond$i$i = 0, $exitcond$i$i$us = 0, $exitcond$i2$i = 0, $exitcond$i2$i$us = 0, $exitcond21 = 0, $exitcond22 = 0, $i$01$i$i = 0, $i$01$i$i$us = 0, $i$01$i1$i = 0, $i$01$i1$i$us = 0, $i$010 = 0, $i$010$us = 0, $idx$05 = 0, $idx$05$us = 0, $j$01 = 0, $ncomponents$0$lcssa = 0, $ncomponents$015 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_ptr4 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 109 "modules/graphics/opengl/wrap_Mesh.cpp"
 $vararg_buffer = sp; //@line 109 "modules/graphics/opengl/wrap_Mesh.cpp"
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 $1 = (+_luaL_optnumber($L,3,1.0)); //@line 112 "modules/graphics/opengl/wrap_Mesh.cpp"
 $2 = (~~(($1))>>>0); //@line 112 "modules/graphics/opengl/wrap_Mesh.cpp"
 $3 = (($2) + -1)|0; //@line 112 "modules/graphics/opengl/wrap_Mesh.cpp"
 $4 = (__ZNK4love8graphics6opengl4Mesh14getVertexCountEv($0)|0); //@line 114 "modules/graphics/opengl/wrap_Mesh.cpp"
 $5 = ($3>>>0)<($4>>>0); //@line 114 "modules/graphics/opengl/wrap_Mesh.cpp"
 if (!($5)) {
  $6 = (__ZNK4love8graphics6opengl4Mesh14getVertexCountEv($0)|0); //@line 115 "modules/graphics/opengl/wrap_Mesh.cpp"
  HEAP32[$vararg_buffer>>2] = $6; //@line 115 "modules/graphics/opengl/wrap_Mesh.cpp"
  $7 = (_luaL_error($L,343175,$vararg_buffer)|0); //@line 115 "modules/graphics/opengl/wrap_Mesh.cpp"
  $$2 = $7;
  STACKTOP = sp;return ($$2|0); //@line 172 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 $8 = (__ZNK4love8graphics6opengl4Mesh15getVertexStrideEv($0)|0); //@line 117 "modules/graphics/opengl/wrap_Mesh.cpp"
 $9 = Math_imul($8, $3)|0; //@line 118 "modules/graphics/opengl/wrap_Mesh.cpp"
 $10 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,2,2)|0); //@line 120 "modules/graphics/opengl/wrap_Mesh.cpp"
 if ($10) {
  $11 = (__ZN4love14luax_checktypeINS_4DataEEEPT_P9lua_StateiNS_4TypeE($L,2,2)|0); //@line 122 "modules/graphics/opengl/wrap_Mesh.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 124 "modules/graphics/opengl/wrap_Mesh.cpp"
  $13 = ((($12)) + 20|0); //@line 124 "modules/graphics/opengl/wrap_Mesh.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 124 "modules/graphics/opengl/wrap_Mesh.cpp"
  $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 124 "modules/graphics/opengl/wrap_Mesh.cpp"
  $16 = (__ZNK4love8graphics6opengl4Mesh14getVertexCountEv($0)|0); //@line 124 "modules/graphics/opengl/wrap_Mesh.cpp"
  $17 = (($16) - ($3))|0; //@line 124 "modules/graphics/opengl/wrap_Mesh.cpp"
  $18 = Math_imul($17, $8)|0; //@line 124 "modules/graphics/opengl/wrap_Mesh.cpp"
  $19 = ($18>>>0)<($15>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $20 = $19 ? $18 : $15; //@line 124 "modules/graphics/opengl/wrap_Mesh.cpp"
  $21 = (__ZN4love8graphics6opengl4Mesh13mapVertexDataEv($0)|0); //@line 125 "modules/graphics/opengl/wrap_Mesh.cpp"
  $22 = (($21) + ($9)|0); //@line 125 "modules/graphics/opengl/wrap_Mesh.cpp"
  $23 = HEAP32[$11>>2]|0; //@line 127 "modules/graphics/opengl/wrap_Mesh.cpp"
  $24 = ((($23)) + 16|0); //@line 127 "modules/graphics/opengl/wrap_Mesh.cpp"
  $25 = HEAP32[$24>>2]|0; //@line 127 "modules/graphics/opengl/wrap_Mesh.cpp"
  $26 = (FUNCTION_TABLE_ii[$25 & 8191]($11)|0); //@line 127 "modules/graphics/opengl/wrap_Mesh.cpp"
  _memcpy(($22|0),($26|0),($20|0))|0; //@line 127 "modules/graphics/opengl/wrap_Mesh.cpp"
  __ZN4love8graphics6opengl4Mesh15unmapVertexDataEjj($0,$9,$20); //@line 129 "modules/graphics/opengl/wrap_Mesh.cpp"
  $$2 = 0;
  STACKTOP = sp;return ($$2|0); //@line 172 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 _luaL_checktype($L,2,5); //@line 133 "modules/graphics/opengl/wrap_Mesh.cpp"
 $27 = (__ZN4love11luax_objlenEP9lua_Statei($L,2)|0); //@line 134 "modules/graphics/opengl/wrap_Mesh.cpp"
 $28 = (($27) + ($3))|0; //@line 136 "modules/graphics/opengl/wrap_Mesh.cpp"
 $29 = (__ZNK4love8graphics6opengl4Mesh14getVertexCountEv($0)|0); //@line 136 "modules/graphics/opengl/wrap_Mesh.cpp"
 $30 = ($28>>>0)>($29>>>0); //@line 136 "modules/graphics/opengl/wrap_Mesh.cpp"
 if ($30) {
  $31 = (__ZNK4love8graphics6opengl4Mesh14getVertexCountEv($0)|0); //@line 137 "modules/graphics/opengl/wrap_Mesh.cpp"
  $32 = (($31) - ($3))|0; //@line 137 "modules/graphics/opengl/wrap_Mesh.cpp"
  HEAP32[$vararg_buffer1>>2] = $32; //@line 137 "modules/graphics/opengl/wrap_Mesh.cpp"
  $vararg_ptr4 = ((($vararg_buffer1)) + 4|0); //@line 137 "modules/graphics/opengl/wrap_Mesh.cpp"
  HEAP32[$vararg_ptr4>>2] = $27; //@line 137 "modules/graphics/opengl/wrap_Mesh.cpp"
  $33 = (_luaL_error($L,343229,$vararg_buffer1)|0); //@line 137 "modules/graphics/opengl/wrap_Mesh.cpp"
  $$2 = $33;
  STACKTOP = sp;return ($$2|0); //@line 172 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 $34 = (__ZNK4love8graphics6opengl4Mesh15getVertexFormatEv($0)|0); //@line 139 "modules/graphics/opengl/wrap_Mesh.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $36 = ((($34)) + 4|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $37 = HEAP32[$36>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $38 = ($35|0)==($37|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($38) {
  $ncomponents$0$lcssa = 0;
 } else {
  $__begin$sroa$0$014 = $35;$ncomponents$015 = 0;
  while(1) {
   $72 = ((($__begin$sroa$0$014)) + 16|0); //@line 143 "modules/graphics/opengl/wrap_Mesh.cpp"
   $73 = HEAP32[$72>>2]|0; //@line 143 "modules/graphics/opengl/wrap_Mesh.cpp"
   $74 = (($73) + ($ncomponents$015))|0; //@line 143 "modules/graphics/opengl/wrap_Mesh.cpp"
   $75 = ((($__begin$sroa$0$014)) + 20|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $76 = ($75|0)==($37|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($76) {
    $ncomponents$0$lcssa = $74;
    break;
   } else {
    $__begin$sroa$0$014 = $75;$ncomponents$015 = $74;
   }
  }
 }
 $39 = (__ZN4love8graphics6opengl4Mesh13mapVertexDataEv($0)|0); //@line 145 "modules/graphics/opengl/wrap_Mesh.cpp"
 $40 = (($39) + ($9)|0); //@line 145 "modules/graphics/opengl/wrap_Mesh.cpp"
 $41 = ($27|0)==(0); //@line 147 "modules/graphics/opengl/wrap_Mesh.cpp"
 L17: do {
  if (!($41)) {
   $42 = ($ncomponents$0$lcssa|0)<(1); //@line 154 "modules/graphics/opengl/wrap_Mesh.cpp"
   $43 = (0 - ($ncomponents$0$lcssa))|0; //@line 157 "modules/graphics/opengl/wrap_Mesh.cpp"
   $44 = (-2 - ($ncomponents$0$lcssa))|0; //@line 167 "modules/graphics/opengl/wrap_Mesh.cpp"
   if ($42) {
    $data$09$us = $40;$i$010$us = 0;
    while(1) {
     $45 = (($i$010$us) + 1)|0; //@line 150 "modules/graphics/opengl/wrap_Mesh.cpp"
     _lua_rawgeti($L,2,$45); //@line 150 "modules/graphics/opengl/wrap_Mesh.cpp"
     _luaL_checktype($L,-1,5); //@line 151 "modules/graphics/opengl/wrap_Mesh.cpp"
     $46 = HEAP32[$34>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $47 = HEAP32[$36>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $48 = ($46|0)==($47|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     L22: do {
      if ($48) {
       $data$1$lcssa$us = $data$09$us;
      } else {
       $__begin2$sroa$0$02$us = $46;$data$13$us = $data$09$us;$idx$05$us = $43;
       while(1) {
        $49 = ((($__begin2$sroa$0$02$us)) + 12|0); //@line 162 "modules/graphics/opengl/wrap_Mesh.cpp"
        $50 = HEAP32[$49>>2]|0; //@line 162 "modules/graphics/opengl/wrap_Mesh.cpp"
        $51 = ((($__begin2$sroa$0$02$us)) + 16|0); //@line 162 "modules/graphics/opengl/wrap_Mesh.cpp"
        $52 = HEAP32[$51>>2]|0; //@line 162 "modules/graphics/opengl/wrap_Mesh.cpp"
        switch ($50|0) {
        case 0:  {
         $61 = ($52|0)>(0); //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
         if ($61) {
          $i$01$i$i$us = 0;
          while(1) {
           $62 = (($i$01$i$i$us) + ($idx$05$us))|0; //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
           $63 = (+_luaL_optnumber($L,$62,255.0)); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
           $64 = (~~(($63))&255); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
           $65 = (($data$13$us) + ($i$01$i$i$us)|0); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
           HEAP8[$65>>0] = $64; //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
           $66 = (($i$01$i$i$us) + 1)|0; //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
           $exitcond$i$i$us = ($66|0)==($52|0); //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
           if ($exitcond$i$i$us) {
            break;
           } else {
            $i$01$i$i$us = $66;
           }
          }
         }
         $67 = (($data$13$us) + ($52)|0); //@line 68 "modules/graphics/opengl/wrap_Mesh.cpp"
         $$0$i$us = $67;
         break;
        }
        case 1:  {
         $53 = ($52|0)>(0); //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
         if ($53) {
          $i$01$i1$i$us = 0;
          while(1) {
           $54 = (($i$01$i1$i$us) + ($idx$05$us))|0; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
           $55 = (+_luaL_optnumber($L,$54,0.0)); //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
           $56 = $55; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
           $57 = (($data$13$us) + ($i$01$i1$i$us<<2)|0); //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
           HEAPF32[$57>>2] = $56; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
           $58 = (($i$01$i1$i$us) + 1)|0; //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
           $exitcond$i2$i$us = ($58|0)==($52|0); //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
           if ($exitcond$i2$i$us) {
            break;
           } else {
            $i$01$i1$i$us = $58;
           }
          }
         }
         $59 = $52 << 2; //@line 60 "modules/graphics/opengl/wrap_Mesh.cpp"
         $60 = (($data$13$us) + ($59)|0); //@line 70 "modules/graphics/opengl/wrap_Mesh.cpp"
         $$0$i$us = $60;
         break;
        }
        default: {
         $$0$i$us = $data$13$us;
        }
        }
        $68 = HEAP32[$51>>2]|0; //@line 164 "modules/graphics/opengl/wrap_Mesh.cpp"
        $69 = (($68) + ($idx$05$us))|0; //@line 164 "modules/graphics/opengl/wrap_Mesh.cpp"
        $70 = ((($__begin2$sroa$0$02$us)) + 20|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        $71 = ($70|0)==($47|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        if ($71) {
         $data$1$lcssa$us = $$0$i$us;
         break L22;
        } else {
         $__begin2$sroa$0$02$us = $70;$data$13$us = $$0$i$us;$idx$05$us = $69;
        }
       }
      }
     } while(0);
     _lua_settop($L,$44); //@line 167 "modules/graphics/opengl/wrap_Mesh.cpp"
     $exitcond = ($45|0)==($27|0); //@line 147 "modules/graphics/opengl/wrap_Mesh.cpp"
     if ($exitcond) {
      break L17;
     } else {
      $data$09$us = $data$1$lcssa$us;$i$010$us = $45;
     }
    }
   } else {
    $data$09 = $40;$i$010 = 0;
   }
   while(1) {
    $78 = (($i$010) + 1)|0; //@line 150 "modules/graphics/opengl/wrap_Mesh.cpp"
    _lua_rawgeti($L,2,$78); //@line 150 "modules/graphics/opengl/wrap_Mesh.cpp"
    _luaL_checktype($L,-1,5); //@line 151 "modules/graphics/opengl/wrap_Mesh.cpp"
    $j$01 = 1;
    while(1) {
     $82 = (0 - ($j$01))|0; //@line 155 "modules/graphics/opengl/wrap_Mesh.cpp"
     _lua_rawgeti($L,$82,$j$01); //@line 155 "modules/graphics/opengl/wrap_Mesh.cpp"
     $83 = (($j$01) + 1)|0; //@line 154 "modules/graphics/opengl/wrap_Mesh.cpp"
     $exitcond21 = ($j$01|0)==($ncomponents$0$lcssa|0); //@line 154 "modules/graphics/opengl/wrap_Mesh.cpp"
     if ($exitcond21) {
      break;
     } else {
      $j$01 = $83;
     }
    }
    $79 = HEAP32[$34>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $80 = HEAP32[$36>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $81 = ($79|0)==($80|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    L43: do {
     if ($81) {
      $data$1$lcssa = $data$09;
     } else {
      $__begin2$sroa$0$02 = $79;$data$13 = $data$09;$idx$05 = $43;
      while(1) {
       $84 = ((($__begin2$sroa$0$02)) + 12|0); //@line 162 "modules/graphics/opengl/wrap_Mesh.cpp"
       $85 = HEAP32[$84>>2]|0; //@line 162 "modules/graphics/opengl/wrap_Mesh.cpp"
       $86 = ((($__begin2$sroa$0$02)) + 16|0); //@line 162 "modules/graphics/opengl/wrap_Mesh.cpp"
       $87 = HEAP32[$86>>2]|0; //@line 162 "modules/graphics/opengl/wrap_Mesh.cpp"
       switch ($85|0) {
       case 0:  {
        $88 = ($87|0)>(0); //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
        if ($88) {
         $i$01$i$i = 0;
         while(1) {
          $89 = (($i$01$i$i) + ($idx$05))|0; //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
          $90 = (+_luaL_optnumber($L,$89,255.0)); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
          $91 = (~~(($90))&255); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
          $92 = (($data$13) + ($i$01$i$i)|0); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
          HEAP8[$92>>0] = $91; //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
          $93 = (($i$01$i$i) + 1)|0; //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
          $exitcond$i$i = ($93|0)==($87|0); //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
          if ($exitcond$i$i) {
           break;
          } else {
           $i$01$i$i = $93;
          }
         }
        }
        $94 = (($data$13) + ($87)|0); //@line 68 "modules/graphics/opengl/wrap_Mesh.cpp"
        $$0$i = $94;
        break;
       }
       case 1:  {
        $95 = ($87|0)>(0); //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
        if ($95) {
         $i$01$i1$i = 0;
         while(1) {
          $96 = (($i$01$i1$i) + ($idx$05))|0; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
          $97 = (+_luaL_optnumber($L,$96,0.0)); //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
          $98 = $97; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
          $99 = (($data$13) + ($i$01$i1$i<<2)|0); //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
          HEAPF32[$99>>2] = $98; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
          $100 = (($i$01$i1$i) + 1)|0; //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
          $exitcond$i2$i = ($100|0)==($87|0); //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
          if ($exitcond$i2$i) {
           break;
          } else {
           $i$01$i1$i = $100;
          }
         }
        }
        $101 = $87 << 2; //@line 60 "modules/graphics/opengl/wrap_Mesh.cpp"
        $102 = (($data$13) + ($101)|0); //@line 70 "modules/graphics/opengl/wrap_Mesh.cpp"
        $$0$i = $102;
        break;
       }
       default: {
        $$0$i = $data$13;
       }
       }
       $103 = HEAP32[$86>>2]|0; //@line 164 "modules/graphics/opengl/wrap_Mesh.cpp"
       $104 = (($103) + ($idx$05))|0; //@line 164 "modules/graphics/opengl/wrap_Mesh.cpp"
       $105 = ((($__begin2$sroa$0$02)) + 20|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       $106 = ($105|0)==($80|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       if ($106) {
        $data$1$lcssa = $$0$i;
        break L43;
       } else {
        $__begin2$sroa$0$02 = $105;$data$13 = $$0$i;$idx$05 = $104;
       }
      }
     }
    } while(0);
    _lua_settop($L,$44); //@line 167 "modules/graphics/opengl/wrap_Mesh.cpp"
    $exitcond22 = ($78|0)==($27|0); //@line 147 "modules/graphics/opengl/wrap_Mesh.cpp"
    if ($exitcond22) {
     break;
    } else {
     $data$09 = $data$1$lcssa;$i$010 = $78;
    }
   }
  }
 } while(0);
 $77 = Math_imul($27, $8)|0; //@line 170 "modules/graphics/opengl/wrap_Mesh.cpp"
 __ZN4love8graphics6opengl4Mesh15unmapVertexDataEjj($0,$9,$77); //@line 170 "modules/graphics/opengl/wrap_Mesh.cpp"
 $$2 = 0;
 STACKTOP = sp;return ($$2|0); //@line 172 "modules/graphics/opengl/wrap_Mesh.cpp"
}
function __ZN4love8graphics6opengl16w_Mesh_setVertexEP9lua_State($L) {
 $L = $L|0;
 var $$0$i = 0, $$0$i11 = 0, $$01$i = 0, $$02$i = 0, $$lcssa16 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0.0, $26 = 0.0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0.0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0.0, $54 = 0.0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $9 = 0, $__begin$sroa$0$019 = 0, $__begin2$sroa$0$023 = 0, $exitcond$i$i = 0, $exitcond$i$i4 = 0, $exitcond$i2$i = 0, $exitcond$i2$i8 = 0, $i$01$i$i = 0, $i$01$i$i3 = 0, $i$01$i1$i = 0;
 var $i$01$i1$i7 = 0, $i$017 = 0, $idx$020 = 0, $idx$124 = 0, $vararg_buffer = 0, $writtendata$021 = 0, $writtendata$125 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 174 "modules/graphics/opengl/wrap_Mesh.cpp"
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 $1 = (_luaL_checkinteger($L,2)|0); //@line 177 "modules/graphics/opengl/wrap_Mesh.cpp"
 $2 = (($1) + -1)|0; //@line 177 "modules/graphics/opengl/wrap_Mesh.cpp"
 $3 = (_lua_type($L,3)|0); //@line 179 "modules/graphics/opengl/wrap_Mesh.cpp"
 $4 = ($3|0)==(5); //@line 179 "modules/graphics/opengl/wrap_Mesh.cpp"
 $5 = (__ZNK4love8graphics6opengl4Mesh15getVertexFormatEv($0)|0); //@line 181 "modules/graphics/opengl/wrap_Mesh.cpp"
 $6 = (__ZN4love8graphics6opengl4Mesh22getVertexScratchBufferEv($0)|0); //@line 183 "modules/graphics/opengl/wrap_Mesh.cpp"
 $7 = HEAP32[$5>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($5)) + 4|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = HEAP32[$8>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = ($7|0)==($9|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 L1: do {
  if ($4) {
   if (!($10)) {
    $__begin$sroa$0$019 = $7;$idx$020 = 1;$writtendata$021 = $6;
    while(1) {
     $11 = ((($__begin$sroa$0$019)) + 16|0); //@line 192 "modules/graphics/opengl/wrap_Mesh.cpp"
     $12 = HEAP32[$11>>2]|0; //@line 192 "modules/graphics/opengl/wrap_Mesh.cpp"
     $13 = ($12|0)>(0); //@line 192 "modules/graphics/opengl/wrap_Mesh.cpp"
     if ($13) {
      $i$017 = $idx$020;
      while(1) {
       _lua_rawgeti($L,3,$i$017); //@line 193 "modules/graphics/opengl/wrap_Mesh.cpp"
       $36 = (($i$017) + 1)|0; //@line 192 "modules/graphics/opengl/wrap_Mesh.cpp"
       $37 = HEAP32[$11>>2]|0; //@line 192 "modules/graphics/opengl/wrap_Mesh.cpp"
       $38 = (($37) + ($idx$020))|0; //@line 192 "modules/graphics/opengl/wrap_Mesh.cpp"
       $39 = ($36|0)<($38|0); //@line 192 "modules/graphics/opengl/wrap_Mesh.cpp"
       if ($39) {
        $i$017 = $36;
       } else {
        $$lcssa16 = $37;
        break;
       }
      }
     } else {
      $$lcssa16 = $12;
     }
     $14 = ((($__begin$sroa$0$019)) + 12|0); //@line 196 "modules/graphics/opengl/wrap_Mesh.cpp"
     $15 = HEAP32[$14>>2]|0; //@line 196 "modules/graphics/opengl/wrap_Mesh.cpp"
     switch ($15|0) {
     case 0:  {
      $16 = ($$lcssa16|0)>(0); //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
      if ($16) {
       $i$01$i$i = 0;
       while(1) {
        $17 = (($i$01$i$i) - ($$lcssa16))|0; //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
        $18 = (+_luaL_optnumber($L,$17,255.0)); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
        $19 = (~~(($18))&255); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
        $20 = (($writtendata$021) + ($i$01$i$i)|0); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
        HEAP8[$20>>0] = $19; //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
        $21 = (($i$01$i$i) + 1)|0; //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
        $exitcond$i$i = ($21|0)==($$lcssa16|0); //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
        if ($exitcond$i$i) {
         break;
        } else {
         $i$01$i$i = $21;
        }
       }
      }
      $22 = (($writtendata$021) + ($$lcssa16)|0); //@line 68 "modules/graphics/opengl/wrap_Mesh.cpp"
      $$0$i = $22;
      break;
     }
     case 1:  {
      $23 = ($$lcssa16|0)>(0); //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
      if ($23) {
       $i$01$i1$i = 0;
       while(1) {
        $24 = (($i$01$i1$i) - ($$lcssa16))|0; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
        $25 = (+_luaL_optnumber($L,$24,0.0)); //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
        $26 = $25; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
        $27 = (($writtendata$021) + ($i$01$i1$i<<2)|0); //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
        HEAPF32[$27>>2] = $26; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
        $28 = (($i$01$i1$i) + 1)|0; //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
        $exitcond$i2$i = ($28|0)==($$lcssa16|0); //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
        if ($exitcond$i2$i) {
         break;
        } else {
         $i$01$i1$i = $28;
        }
       }
      }
      $29 = $$lcssa16 << 2; //@line 60 "modules/graphics/opengl/wrap_Mesh.cpp"
      $30 = (($writtendata$021) + ($29)|0); //@line 70 "modules/graphics/opengl/wrap_Mesh.cpp"
      $$0$i = $30;
      break;
     }
     default: {
      $$0$i = $writtendata$021;
     }
     }
     $31 = HEAP32[$11>>2]|0; //@line 198 "modules/graphics/opengl/wrap_Mesh.cpp"
     $32 = (($31) + ($idx$020))|0; //@line 198 "modules/graphics/opengl/wrap_Mesh.cpp"
     $33 = $31 ^ -1; //@line 199 "modules/graphics/opengl/wrap_Mesh.cpp"
     _lua_settop($L,$33); //@line 199 "modules/graphics/opengl/wrap_Mesh.cpp"
     $34 = ((($__begin$sroa$0$019)) + 20|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $35 = ($34|0)==($9|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($35) {
      break L1;
     } else {
      $__begin$sroa$0$019 = $34;$idx$020 = $32;$writtendata$021 = $$0$i;
     }
    }
   }
  } else {
   if (!($10)) {
    $__begin2$sroa$0$023 = $7;$idx$124 = 3;$writtendata$125 = $6;
    while(1) {
     $40 = ((($__begin2$sroa$0$023)) + 12|0); //@line 207 "modules/graphics/opengl/wrap_Mesh.cpp"
     $41 = HEAP32[$40>>2]|0; //@line 207 "modules/graphics/opengl/wrap_Mesh.cpp"
     $42 = ((($__begin2$sroa$0$023)) + 16|0); //@line 207 "modules/graphics/opengl/wrap_Mesh.cpp"
     $43 = HEAP32[$42>>2]|0; //@line 207 "modules/graphics/opengl/wrap_Mesh.cpp"
     switch ($41|0) {
     case 0:  {
      $44 = ($43|0)>(0); //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
      if ($44) {
       $i$01$i$i3 = 0;
       while(1) {
        $45 = (($i$01$i$i3) + ($idx$124))|0; //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
        $46 = (+_luaL_optnumber($L,$45,255.0)); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
        $47 = (~~(($46))&255); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
        $48 = (($writtendata$125) + ($i$01$i$i3)|0); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
        HEAP8[$48>>0] = $47; //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
        $49 = (($i$01$i$i3) + 1)|0; //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
        $exitcond$i$i4 = ($49|0)==($43|0); //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
        if ($exitcond$i$i4) {
         break;
        } else {
         $i$01$i$i3 = $49;
        }
       }
      }
      $50 = (($writtendata$125) + ($43)|0); //@line 68 "modules/graphics/opengl/wrap_Mesh.cpp"
      $$0$i11 = $50;
      break;
     }
     case 1:  {
      $51 = ($43|0)>(0); //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
      if ($51) {
       $i$01$i1$i7 = 0;
       while(1) {
        $52 = (($i$01$i1$i7) + ($idx$124))|0; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
        $53 = (+_luaL_optnumber($L,$52,0.0)); //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
        $54 = $53; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
        $55 = (($writtendata$125) + ($i$01$i1$i7<<2)|0); //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
        HEAPF32[$55>>2] = $54; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
        $56 = (($i$01$i1$i7) + 1)|0; //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
        $exitcond$i2$i8 = ($56|0)==($43|0); //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
        if ($exitcond$i2$i8) {
         break;
        } else {
         $i$01$i1$i7 = $56;
        }
       }
      }
      $57 = $43 << 2; //@line 60 "modules/graphics/opengl/wrap_Mesh.cpp"
      $58 = (($writtendata$125) + ($57)|0); //@line 70 "modules/graphics/opengl/wrap_Mesh.cpp"
      $$0$i11 = $58;
      break;
     }
     default: {
      $$0$i11 = $writtendata$125;
     }
     }
     $59 = HEAP32[$42>>2]|0; //@line 208 "modules/graphics/opengl/wrap_Mesh.cpp"
     $60 = (($59) + ($idx$124))|0; //@line 208 "modules/graphics/opengl/wrap_Mesh.cpp"
     $61 = ((($__begin2$sroa$0$023)) + 20|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $62 = ($61|0)==($9|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($62) {
      break L1;
     } else {
      $__begin2$sroa$0$023 = $61;$idx$124 = $60;$writtendata$125 = $$0$i11;
     }
    }
   }
  }
 } while(0);
 __THREW__ = 0;
 $63 = (invoke_ii(3587,($0|0))|0); //@line 212 "modules/graphics/opengl/wrap_Mesh.cpp"
 $64 = __THREW__; __THREW__ = 0;
 $65 = $64&1;
 if (!($65)) {
  __THREW__ = 0;
  invoke_viiii(3588,($0|0),($2|0),($6|0),($63|0)); //@line 212 "modules/graphics/opengl/wrap_Mesh.cpp"
  $66 = __THREW__; __THREW__ = 0;
  $67 = $66&1;
  if (!($67)) {
   STACKTOP = sp;return 0; //@line 213 "modules/graphics/opengl/wrap_Mesh.cpp"
  }
 }
 $68 = ___cxa_find_matching_catch_3(8992|0)|0;
 $69 = tempRet0;
 $70 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $71 = ($69|0)==($70|0); //@line 535 "./common/runtime.h"
 if (!($71)) {
  $$01$i = $69;$$02$i = $68;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $72 = (___cxa_begin_catch(($68|0))|0); //@line 535 "./common/runtime.h"
 $73 = HEAP32[$72>>2]|0; //@line 539 "./common/runtime.h"
 $74 = ((($73)) + 8|0); //@line 539 "./common/runtime.h"
 $75 = HEAP32[$74>>2]|0; //@line 539 "./common/runtime.h"
 $76 = (FUNCTION_TABLE_ii[$75 & 8191]($72)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($76|0)); //@line 539 "./common/runtime.h"
 $77 = __THREW__; __THREW__ = 0;
 $78 = $77&1;
 if (!($78)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $79 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $79; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 213 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 $80 = ___cxa_find_matching_catch_2()|0;
 $81 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $82 = __THREW__; __THREW__ = 0;
 $83 = $82&1;
 if ($83) {
  $84 = ___cxa_find_matching_catch_3(0|0)|0;
  $85 = tempRet0;
  ___clang_call_terminate($84); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $81;$$02$i = $80;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl16w_Mesh_getVertexEP9lua_State($L) {
 $L = $L|0;
 var $$0$i = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0.0, $45 = 0.0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$03 = 0, $exitcond$i$i = 0, $exitcond$i2$i = 0, $i$01$i$i = 0;
 var $i$01$i1$i = 0, $n$0$lcssa = 0, $n$04 = 0, $readdata$05 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 216 "modules/graphics/opengl/wrap_Mesh.cpp"
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 $1 = (_luaL_checkinteger($L,2)|0); //@line 219 "modules/graphics/opengl/wrap_Mesh.cpp"
 $2 = (__ZNK4love8graphics6opengl4Mesh15getVertexFormatEv($0)|0); //@line 221 "modules/graphics/opengl/wrap_Mesh.cpp"
 $3 = (__ZN4love8graphics6opengl4Mesh22getVertexScratchBufferEv($0)|0); //@line 223 "modules/graphics/opengl/wrap_Mesh.cpp"
 __THREW__ = 0;
 $4 = (invoke_ii(3587,($0|0))|0); //@line 226 "modules/graphics/opengl/wrap_Mesh.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  label = 3;
 } else {
  $7 = (($1) + -1)|0; //@line 219 "modules/graphics/opengl/wrap_Mesh.cpp"
  __THREW__ = 0;
  (invoke_iiiii(3589,($0|0),($7|0),($3|0),($4|0))|0); //@line 226 "modules/graphics/opengl/wrap_Mesh.cpp"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if ($9) {
   label = 3;
  }
 }
 do {
  if ((label|0) == 3) {
   $10 = ___cxa_find_matching_catch_3(8992|0)|0;
   $11 = tempRet0;
   $12 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $13 = ($11|0)==($12|0); //@line 535 "./common/runtime.h"
   if (!($13)) {
    $$01$i = $11;$$02$i = $10;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $14 = (___cxa_begin_catch(($10|0))|0); //@line 535 "./common/runtime.h"
   $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
   $16 = ((($15)) + 8|0); //@line 539 "./common/runtime.h"
   $17 = HEAP32[$16>>2]|0; //@line 539 "./common/runtime.h"
   $18 = (FUNCTION_TABLE_ii[$17 & 8191]($14)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($18|0)); //@line 539 "./common/runtime.h"
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   if (!($20)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $21 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $21; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    break;
   }
   $22 = ___cxa_find_matching_catch_2()|0;
   $23 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $24 = __THREW__; __THREW__ = 0;
   $25 = $24&1;
   if ($25) {
    $26 = ___cxa_find_matching_catch_3(0|0)|0;
    $27 = tempRet0;
    ___clang_call_terminate($26); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $23;$$02$i = $22;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  }
 } while(0);
 $28 = HEAP32[$2>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $29 = ((($2)) + 4|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = HEAP32[$29>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = ($28|0)==($30|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($31) {
  $n$0$lcssa = 0;
  STACKTOP = sp;return ($n$0$lcssa|0); //@line 236 "modules/graphics/opengl/wrap_Mesh.cpp"
 } else {
  $__begin$sroa$0$03 = $28;$n$04 = 0;$readdata$05 = $3;
 }
 while(1) {
  $32 = ((($__begin$sroa$0$03)) + 12|0); //@line 232 "modules/graphics/opengl/wrap_Mesh.cpp"
  $33 = HEAP32[$32>>2]|0; //@line 232 "modules/graphics/opengl/wrap_Mesh.cpp"
  $34 = ((($__begin$sroa$0$03)) + 16|0); //@line 232 "modules/graphics/opengl/wrap_Mesh.cpp"
  $35 = HEAP32[$34>>2]|0; //@line 232 "modules/graphics/opengl/wrap_Mesh.cpp"
  switch ($33|0) {
  case 0:  {
   $36 = ($35|0)>(0); //@line 80 "modules/graphics/opengl/wrap_Mesh.cpp"
   if ($36) {
    $i$01$i$i = 0;
    while(1) {
     $37 = (($readdata$05) + ($i$01$i$i)|0); //@line 81 "modules/graphics/opengl/wrap_Mesh.cpp"
     $38 = HEAP8[$37>>0]|0; //@line 81 "modules/graphics/opengl/wrap_Mesh.cpp"
     $39 = (+($38&255)); //@line 81 "modules/graphics/opengl/wrap_Mesh.cpp"
     _lua_pushnumber($L,$39); //@line 81 "modules/graphics/opengl/wrap_Mesh.cpp"
     $40 = (($i$01$i$i) + 1)|0; //@line 80 "modules/graphics/opengl/wrap_Mesh.cpp"
     $exitcond$i$i = ($40|0)==($35|0); //@line 80 "modules/graphics/opengl/wrap_Mesh.cpp"
     if ($exitcond$i$i) {
      break;
     } else {
      $i$01$i$i = $40;
     }
    }
   }
   $41 = (($readdata$05) + ($35)|0); //@line 101 "modules/graphics/opengl/wrap_Mesh.cpp"
   $$0$i = $41;
   break;
  }
  case 1:  {
   $42 = ($35|0)>(0); //@line 90 "modules/graphics/opengl/wrap_Mesh.cpp"
   if ($42) {
    $i$01$i1$i = 0;
    while(1) {
     $43 = (($readdata$05) + ($i$01$i1$i<<2)|0); //@line 91 "modules/graphics/opengl/wrap_Mesh.cpp"
     $44 = +HEAPF32[$43>>2]; //@line 91 "modules/graphics/opengl/wrap_Mesh.cpp"
     $45 = $44; //@line 91 "modules/graphics/opengl/wrap_Mesh.cpp"
     _lua_pushnumber($L,$45); //@line 91 "modules/graphics/opengl/wrap_Mesh.cpp"
     $46 = (($i$01$i1$i) + 1)|0; //@line 90 "modules/graphics/opengl/wrap_Mesh.cpp"
     $exitcond$i2$i = ($46|0)==($35|0); //@line 90 "modules/graphics/opengl/wrap_Mesh.cpp"
     if ($exitcond$i2$i) {
      break;
     } else {
      $i$01$i1$i = $46;
     }
    }
   }
   $47 = $35 << 2; //@line 93 "modules/graphics/opengl/wrap_Mesh.cpp"
   $48 = (($readdata$05) + ($47)|0); //@line 103 "modules/graphics/opengl/wrap_Mesh.cpp"
   $$0$i = $48;
   break;
  }
  default: {
   $$0$i = $readdata$05;
  }
  }
  $49 = HEAP32[$34>>2]|0; //@line 233 "modules/graphics/opengl/wrap_Mesh.cpp"
  $50 = (($49) + ($n$04))|0; //@line 233 "modules/graphics/opengl/wrap_Mesh.cpp"
  $51 = ((($__begin$sroa$0$03)) + 20|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $52 = ($51|0)==($30|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($52) {
   $n$0$lcssa = $50;
   break;
  } else {
   $__begin$sroa$0$03 = $51;$n$04 = $50;$readdata$05 = $$0$i;
  }
 }
 STACKTOP = sp;return ($n$0$lcssa|0); //@line 236 "modules/graphics/opengl/wrap_Mesh.cpp"
}
function __ZN4love8graphics6opengl25w_Mesh_setVertexAttributeEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$01$i2 = 0, $$02$i = 0, $$02$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $7 = 0, $8 = 0, $9 = 0, $components = 0, $data = 0, $exitcond$i$i = 0, $exitcond$i2$i = 0, $i$01$i$i = 0, $i$01$i1$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $components = sp + 12|0;
 $data = sp + 16|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 $1 = (_luaL_checkinteger($L,2)|0); //@line 242 "modules/graphics/opengl/wrap_Mesh.cpp"
 $2 = (($1) + -1)|0; //@line 242 "modules/graphics/opengl/wrap_Mesh.cpp"
 $3 = (_luaL_checkinteger($L,3)|0); //@line 243 "modules/graphics/opengl/wrap_Mesh.cpp"
 $4 = (($3) + -1)|0; //@line 243 "modules/graphics/opengl/wrap_Mesh.cpp"
 __THREW__ = 0;
 $5 = (invoke_iiii(3590,($0|0),($4|0),($components|0))|0); //@line 247 "modules/graphics/opengl/wrap_Mesh.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 L1: do {
  if ($7) {
   $8 = ___cxa_find_matching_catch_3(8992|0)|0;
   $9 = tempRet0;
   $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $11 = ($9|0)==($10|0); //@line 535 "./common/runtime.h"
   if (!($11)) {
    $$01$i = $9;$$02$i = $8;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $12 = (___cxa_begin_catch(($8|0))|0); //@line 535 "./common/runtime.h"
   $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
   $14 = ((($13)) + 8|0); //@line 539 "./common/runtime.h"
   $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
   $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($16|0)); //@line 539 "./common/runtime.h"
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if (!($18)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $19 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $19; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $20 = HEAP32[$components>>2]|0; //@line 253 "modules/graphics/opengl/wrap_Mesh.cpp"
    $29 = $20;$62 = $data;
    label = 9;
    break;
   }
   $21 = ___cxa_find_matching_catch_2()|0;
   $22 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $23 = __THREW__; __THREW__ = 0;
   $24 = $23&1;
   if ($24) {
    $25 = ___cxa_find_matching_catch_3(0|0)|0;
    $26 = tempRet0;
    ___clang_call_terminate($25); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $22;$$02$i = $21;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $27 = HEAP32[$components>>2]|0; //@line 253 "modules/graphics/opengl/wrap_Mesh.cpp"
   switch ($5|0) {
   case 0:  {
    $29 = $27;$62 = $data;
    label = 9;
    break L1;
    break;
   }
   case 1:  {
    break;
   }
   default: {
    $41 = $data;
    break L1;
   }
   }
   $35 = ($27|0)>(0); //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
   if ($35) {
    $i$01$i1$i = 0;
    while(1) {
     $36 = (($i$01$i1$i) + 4)|0; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
     $37 = (+_luaL_optnumber($L,$36,0.0)); //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
     $38 = $37; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
     $39 = (($data) + ($i$01$i1$i<<2)|0); //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
     HEAPF32[$39>>2] = $38; //@line 58 "modules/graphics/opengl/wrap_Mesh.cpp"
     $40 = (($i$01$i1$i) + 1)|0; //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
     $exitcond$i2$i = ($40|0)==($27|0); //@line 57 "modules/graphics/opengl/wrap_Mesh.cpp"
     if ($exitcond$i2$i) {
      $41 = $data;
      break;
     } else {
      $i$01$i1$i = $40;
     }
    }
   } else {
    $41 = $data;
   }
  }
 } while(0);
 if ((label|0) == 9) {
  $28 = ($29|0)>(0); //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
  if ($28) {
   $i$01$i$i = 0;
   while(1) {
    $30 = (($i$01$i$i) + 4)|0; //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
    $31 = (+_luaL_optnumber($L,$30,255.0)); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
    $32 = (~~(($31))&255); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
    $33 = (($data) + ($i$01$i$i)|0); //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
    HEAP8[$33>>0] = $32; //@line 48 "modules/graphics/opengl/wrap_Mesh.cpp"
    $34 = (($i$01$i$i) + 1)|0; //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
    $exitcond$i$i = ($34|0)==($29|0); //@line 47 "modules/graphics/opengl/wrap_Mesh.cpp"
    if ($exitcond$i$i) {
     $41 = $62;
     break;
    } else {
     $i$01$i$i = $34;
    }
   }
  } else {
   $41 = $62;
  }
 }
 __THREW__ = 0;
 invoke_viiiii(3563,($0|0),($2|0),($4|0),($41|0),16); //@line 255 "modules/graphics/opengl/wrap_Mesh.cpp"
 $42 = __THREW__; __THREW__ = 0;
 $43 = $42&1;
 if (!($43)) {
  STACKTOP = sp;return 0; //@line 256 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 $44 = ___cxa_find_matching_catch_3(8992|0)|0;
 $45 = tempRet0;
 $46 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $47 = ($45|0)==($46|0); //@line 535 "./common/runtime.h"
 if (!($47)) {
  $$01$i2 = $45;$$02$i1 = $44;
  ___resumeException($$02$i1|0);
  // unreachable;
 }
 $48 = (___cxa_begin_catch(($44|0))|0); //@line 535 "./common/runtime.h"
 $49 = HEAP32[$48>>2]|0; //@line 539 "./common/runtime.h"
 $50 = ((($49)) + 8|0); //@line 539 "./common/runtime.h"
 $51 = HEAP32[$50>>2]|0; //@line 539 "./common/runtime.h"
 $52 = (FUNCTION_TABLE_ii[$51 & 8191]($48)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($52|0)); //@line 539 "./common/runtime.h"
 $53 = __THREW__; __THREW__ = 0;
 $54 = $53&1;
 if (!($54)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $55 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $55; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 256 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 $56 = ___cxa_find_matching_catch_2()|0;
 $57 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $58 = __THREW__; __THREW__ = 0;
 $59 = $58&1;
 if ($59) {
  $60 = ___cxa_find_matching_catch_3(0|0)|0;
  $61 = tempRet0;
  ___clang_call_terminate($60); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i2 = $57;$$02$i1 = $56;
  ___resumeException($$02$i1|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl25w_Mesh_getVertexAttributeEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$01$i2 = 0, $$02$i = 0, $$02$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0, $53 = 0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $components = 0, $data = 0, $exitcond$i$i = 0, $exitcond$i2$i = 0, $i$01$i$i = 0, $i$01$i1$i = 0, $type$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $components = sp + 12|0;
 $data = sp + 16|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 $1 = (_luaL_checkinteger($L,2)|0); //@line 262 "modules/graphics/opengl/wrap_Mesh.cpp"
 $2 = (($1) + -1)|0; //@line 262 "modules/graphics/opengl/wrap_Mesh.cpp"
 $3 = (_luaL_checkinteger($L,3)|0); //@line 263 "modules/graphics/opengl/wrap_Mesh.cpp"
 $4 = (($3) + -1)|0; //@line 263 "modules/graphics/opengl/wrap_Mesh.cpp"
 __THREW__ = 0;
 $5 = (invoke_iiii(3590,($0|0),($4|0),($components|0))|0); //@line 267 "modules/graphics/opengl/wrap_Mesh.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 do {
  if ($7) {
   $8 = ___cxa_find_matching_catch_3(8992|0)|0;
   $9 = tempRet0;
   $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $11 = ($9|0)==($10|0); //@line 535 "./common/runtime.h"
   if (!($11)) {
    $$01$i = $9;$$02$i = $8;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $12 = (___cxa_begin_catch(($8|0))|0); //@line 535 "./common/runtime.h"
   $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
   $14 = ((($13)) + 8|0); //@line 539 "./common/runtime.h"
   $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
   $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($16|0)); //@line 539 "./common/runtime.h"
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if (!($18)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $19 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $19; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $type$0 = 0;
    break;
   }
   $20 = ___cxa_find_matching_catch_2()|0;
   $21 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $22 = __THREW__; __THREW__ = 0;
   $23 = $22&1;
   if ($23) {
    $24 = ___cxa_find_matching_catch_3(0|0)|0;
    $25 = tempRet0;
    ___clang_call_terminate($24); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $21;$$02$i = $20;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $type$0 = $5;
  }
 } while(0);
 __THREW__ = 0;
 (invoke_iiiiii(3591,($0|0),($2|0),($4|0),($data|0),16)|0); //@line 272 "modules/graphics/opengl/wrap_Mesh.cpp"
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 do {
  if ($27) {
   $28 = ___cxa_find_matching_catch_3(8992|0)|0;
   $29 = tempRet0;
   $30 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $31 = ($29|0)==($30|0); //@line 535 "./common/runtime.h"
   if (!($31)) {
    $$01$i2 = $29;$$02$i1 = $28;
    ___resumeException($$02$i1|0);
    // unreachable;
   }
   $32 = (___cxa_begin_catch(($28|0))|0); //@line 535 "./common/runtime.h"
   $33 = HEAP32[$32>>2]|0; //@line 539 "./common/runtime.h"
   $34 = ((($33)) + 8|0); //@line 539 "./common/runtime.h"
   $35 = HEAP32[$34>>2]|0; //@line 539 "./common/runtime.h"
   $36 = (FUNCTION_TABLE_ii[$35 & 8191]($32)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($36|0)); //@line 539 "./common/runtime.h"
   $37 = __THREW__; __THREW__ = 0;
   $38 = $37&1;
   if (!($38)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $39 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer1>>2] = $39; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
    break;
   }
   $40 = ___cxa_find_matching_catch_2()|0;
   $41 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $42 = __THREW__; __THREW__ = 0;
   $43 = $42&1;
   if ($43) {
    $44 = ___cxa_find_matching_catch_3(0|0)|0;
    $45 = tempRet0;
    ___clang_call_terminate($44); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i2 = $41;$$02$i1 = $40;
    ___resumeException($$02$i1|0);
    // unreachable;
   }
  }
 } while(0);
 $46 = HEAP32[$components>>2]|0; //@line 274 "modules/graphics/opengl/wrap_Mesh.cpp"
 switch ($type$0|0) {
 case 0:  {
  $47 = ($46|0)>(0); //@line 80 "modules/graphics/opengl/wrap_Mesh.cpp"
  if ($47) {
   $i$01$i$i = 0;
  } else {
   $57 = HEAP32[$components>>2]|0; //@line 275 "modules/graphics/opengl/wrap_Mesh.cpp"
   STACKTOP = sp;return ($57|0); //@line 275 "modules/graphics/opengl/wrap_Mesh.cpp"
  }
  while(1) {
   $48 = (($data) + ($i$01$i$i)|0); //@line 81 "modules/graphics/opengl/wrap_Mesh.cpp"
   $49 = HEAP8[$48>>0]|0; //@line 81 "modules/graphics/opengl/wrap_Mesh.cpp"
   $50 = (+($49&255)); //@line 81 "modules/graphics/opengl/wrap_Mesh.cpp"
   _lua_pushnumber($L,$50); //@line 81 "modules/graphics/opengl/wrap_Mesh.cpp"
   $51 = (($i$01$i$i) + 1)|0; //@line 80 "modules/graphics/opengl/wrap_Mesh.cpp"
   $exitcond$i$i = ($51|0)==($46|0); //@line 80 "modules/graphics/opengl/wrap_Mesh.cpp"
   if ($exitcond$i$i) {
    break;
   } else {
    $i$01$i$i = $51;
   }
  }
  $57 = HEAP32[$components>>2]|0; //@line 275 "modules/graphics/opengl/wrap_Mesh.cpp"
  STACKTOP = sp;return ($57|0); //@line 275 "modules/graphics/opengl/wrap_Mesh.cpp"
  break;
 }
 case 1:  {
  $52 = ($46|0)>(0); //@line 90 "modules/graphics/opengl/wrap_Mesh.cpp"
  if ($52) {
   $i$01$i1$i = 0;
  } else {
   $57 = HEAP32[$components>>2]|0; //@line 275 "modules/graphics/opengl/wrap_Mesh.cpp"
   STACKTOP = sp;return ($57|0); //@line 275 "modules/graphics/opengl/wrap_Mesh.cpp"
  }
  while(1) {
   $53 = (($data) + ($i$01$i1$i<<2)|0); //@line 91 "modules/graphics/opengl/wrap_Mesh.cpp"
   $54 = +HEAPF32[$53>>2]; //@line 91 "modules/graphics/opengl/wrap_Mesh.cpp"
   $55 = $54; //@line 91 "modules/graphics/opengl/wrap_Mesh.cpp"
   _lua_pushnumber($L,$55); //@line 91 "modules/graphics/opengl/wrap_Mesh.cpp"
   $56 = (($i$01$i1$i) + 1)|0; //@line 90 "modules/graphics/opengl/wrap_Mesh.cpp"
   $exitcond$i2$i = ($56|0)==($46|0); //@line 90 "modules/graphics/opengl/wrap_Mesh.cpp"
   if ($exitcond$i2$i) {
    break;
   } else {
    $i$01$i1$i = $56;
   }
  }
  $57 = HEAP32[$components>>2]|0; //@line 275 "modules/graphics/opengl/wrap_Mesh.cpp"
  STACKTOP = sp;return ($57|0); //@line 275 "modules/graphics/opengl/wrap_Mesh.cpp"
  break;
 }
 default: {
  $57 = HEAP32[$components>>2]|0; //@line 275 "modules/graphics/opengl/wrap_Mesh.cpp"
  STACKTOP = sp;return ($57|0); //@line 275 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl21w_Mesh_getVertexCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 $1 = (__ZNK4love8graphics6opengl4Mesh14getVertexCountEv($0)|0); //@line 281 "modules/graphics/opengl/wrap_Mesh.cpp"
 _lua_pushinteger($L,$1); //@line 281 "modules/graphics/opengl/wrap_Mesh.cpp"
 return 1; //@line 282 "modules/graphics/opengl/wrap_Mesh.cpp"
}
function __ZN4love8graphics6opengl22w_Mesh_getVertexFormatEP9lua_State($L) {
 $L = $L|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$02 = 0, $tname = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $tname = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 $1 = (__ZNK4love8graphics6opengl4Mesh15getVertexFormatEv($0)|0); //@line 289 "modules/graphics/opengl/wrap_Mesh.cpp"
 $2 = ((($1)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = (($5|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 _lua_createtable($L,$6,0); //@line 290 "modules/graphics/opengl/wrap_Mesh.cpp"
 HEAP32[$tname>>2] = 0; //@line 292 "modules/graphics/opengl/wrap_Mesh.cpp"
 $7 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ($7|0)==($8|0); //@line 294 "modules/graphics/opengl/wrap_Mesh.cpp"
 if ($9) {
  $34 = 1;
  STACKTOP = sp;return ($34|0); //@line 315 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 $$in = $8;$i$02 = 0;
 while(1) {
  $10 = $$in; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $11 = (((($10) + (($i$02*20)|0)|0)) + 12|0); //@line 296 "modules/graphics/opengl/wrap_Mesh.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 296 "modules/graphics/opengl/wrap_Mesh.cpp"
  $13 = (__ZN4love8graphics6opengl4Mesh11getConstantENS2_8DataTypeERPKc($12,$tname)|0); //@line 296 "modules/graphics/opengl/wrap_Mesh.cpp"
  if (!($13)) {
   break;
  }
  _lua_createtable($L,3,0); //@line 299 "modules/graphics/opengl/wrap_Mesh.cpp"
  $14 = HEAP32[$1>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = (($14) + (($i$02*20)|0)|0); //@line 301 "modules/graphics/opengl/wrap_Mesh.cpp"
  $16 = HEAP8[$15>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $17 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $18 = ($17<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($18) {
   $21 = ((($15)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $22 = $21;
  } else {
   $19 = (((($14) + (($i$02*20)|0)|0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $20 = HEAP32[$19>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $22 = $20;
  }
  _lua_pushstring($L,$22); //@line 301 "modules/graphics/opengl/wrap_Mesh.cpp"
  _lua_rawseti($L,-2,1); //@line 302 "modules/graphics/opengl/wrap_Mesh.cpp"
  $23 = HEAP32[$tname>>2]|0; //@line 304 "modules/graphics/opengl/wrap_Mesh.cpp"
  _lua_pushstring($L,$23); //@line 304 "modules/graphics/opengl/wrap_Mesh.cpp"
  _lua_rawseti($L,-2,2); //@line 305 "modules/graphics/opengl/wrap_Mesh.cpp"
  $24 = HEAP32[$1>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $25 = (((($24) + (($i$02*20)|0)|0)) + 16|0); //@line 307 "modules/graphics/opengl/wrap_Mesh.cpp"
  $26 = HEAP32[$25>>2]|0; //@line 307 "modules/graphics/opengl/wrap_Mesh.cpp"
  _lua_pushinteger($L,$26); //@line 307 "modules/graphics/opengl/wrap_Mesh.cpp"
  _lua_rawseti($L,-2,3); //@line 308 "modules/graphics/opengl/wrap_Mesh.cpp"
  $27 = (($i$02) + 1)|0; //@line 311 "modules/graphics/opengl/wrap_Mesh.cpp"
  _lua_rawseti($L,-2,$27); //@line 311 "modules/graphics/opengl/wrap_Mesh.cpp"
  $28 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $29 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $30 = (($28) - ($29))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $31 = (($30|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $32 = ($27>>>0)<($31>>>0); //@line 294 "modules/graphics/opengl/wrap_Mesh.cpp"
  if ($32) {
   $$in = $29;$i$02 = $27;
  } else {
   $34 = 1;
   label = 9;
   break;
  }
 }
 if ((label|0) == 9) {
  STACKTOP = sp;return ($34|0); //@line 315 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 $33 = (_luaL_error($L,343277,$vararg_buffer)|0); //@line 297 "modules/graphics/opengl/wrap_Mesh.cpp"
 $34 = $33;
 STACKTOP = sp;return ($34|0); //@line 315 "modules/graphics/opengl/wrap_Mesh.cpp"
}
function __ZN4love8graphics6opengl26w_Mesh_setAttributeEnabledEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 $1 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 $2 = (_luaL_checklstring($L,2,0)|0); //@line 320 "modules/graphics/opengl/wrap_Mesh.cpp"
 $3 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,3)|0); //@line 321 "modules/graphics/opengl/wrap_Mesh.cpp"
 $4 = (_strlen($2)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viii(3460,($0|0),($2|0),($4|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 do {
  if ($6) {
   $11 = ___cxa_find_matching_catch_3(8992|0)|0;
   $12 = tempRet0;
   $eh$lpad$body$i$index2Z2D = $12;$eh$lpad$body$i$indexZ2D = $11;
  } else {
   __THREW__ = 0;
   invoke_viii(3592,($1|0),($0|0),($3|0)); //@line 322 "modules/graphics/opengl/wrap_Mesh.cpp"
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if ($8) {
    $9 = ___cxa_find_matching_catch_3(8992|0)|0;
    $10 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 322 "modules/graphics/opengl/wrap_Mesh.cpp"
    $eh$lpad$body$i$index2Z2D = $10;$eh$lpad$body$i$indexZ2D = $9;
    break;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 322 "modules/graphics/opengl/wrap_Mesh.cpp"
   STACKTOP = sp;return 0; //@line 323 "modules/graphics/opengl/wrap_Mesh.cpp"
  }
 } while(0);
 $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $14 = ($eh$lpad$body$i$index2Z2D|0)==($13|0); //@line 535 "./common/runtime.h"
 if (!($14)) {
  $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $15 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
 $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
 $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
 $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
 $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if (!($21)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 323 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 $23 = ___cxa_find_matching_catch_2()|0;
 $24 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if ($26) {
  $27 = ___cxa_find_matching_catch_3(0|0)|0;
  $28 = tempRet0;
  ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $24;$$02$i = $23;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl25w_Mesh_isAttributeEnabledEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $enabled$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 $1 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 $2 = (_luaL_checklstring($L,2,0)|0); //@line 329 "modules/graphics/opengl/wrap_Mesh.cpp"
 $3 = (_strlen($2)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viii(3460,($0|0),($2|0),($3|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 do {
  if ($5) {
   $12 = ___cxa_find_matching_catch_3(8992|0)|0;
   $13 = tempRet0;
   $eh$lpad$body$i$index2Z2D = $13;$eh$lpad$body$i$indexZ2D = $12;
  } else {
   __THREW__ = 0;
   $6 = (invoke_iii(3593,($1|0),($0|0))|0); //@line 331 "modules/graphics/opengl/wrap_Mesh.cpp"
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if ($8) {
    $9 = ___cxa_find_matching_catch_3(8992|0)|0;
    $10 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 331 "modules/graphics/opengl/wrap_Mesh.cpp"
    $eh$lpad$body$i$index2Z2D = $10;$eh$lpad$body$i$indexZ2D = $9;
    break;
   }
   $11 = $6&1; //@line 331 "modules/graphics/opengl/wrap_Mesh.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 331 "modules/graphics/opengl/wrap_Mesh.cpp"
   $enabled$0 = $11;
   _lua_pushboolean($L,$enabled$0); //@line 332 "modules/graphics/opengl/wrap_Mesh.cpp"
   STACKTOP = sp;return 1; //@line 333 "modules/graphics/opengl/wrap_Mesh.cpp"
  }
 } while(0);
 $14 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $15 = ($eh$lpad$body$i$index2Z2D|0)==($14|0); //@line 535 "./common/runtime.h"
 if (!($15)) {
  $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $16 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
 $17 = HEAP32[$16>>2]|0; //@line 539 "./common/runtime.h"
 $18 = ((($17)) + 8|0); //@line 539 "./common/runtime.h"
 $19 = HEAP32[$18>>2]|0; //@line 539 "./common/runtime.h"
 $20 = (FUNCTION_TABLE_ii[$19 & 8191]($16)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($20|0)); //@line 539 "./common/runtime.h"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if (!($22)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $23 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $23; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $enabled$0 = 0;
  _lua_pushboolean($L,$enabled$0); //@line 332 "modules/graphics/opengl/wrap_Mesh.cpp"
  STACKTOP = sp;return 1; //@line 333 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 $24 = ___cxa_find_matching_catch_2()|0;
 $25 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  $28 = ___cxa_find_matching_catch_3(0|0)|0;
  $29 = tempRet0;
  ___clang_call_terminate($28); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $25;$$02$i = $24;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl22w_Mesh_attachAttributeEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 $1 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 $2 = (_luaL_checklstring($L,2,0)|0); //@line 339 "modules/graphics/opengl/wrap_Mesh.cpp"
 $3 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,3,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 $4 = (_strlen($2)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viii(3460,($0|0),($2|0),($4|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 do {
  if ($6) {
   $11 = ___cxa_find_matching_catch_3(8992|0)|0;
   $12 = tempRet0;
   $eh$lpad$body$i$index2Z2D = $12;$eh$lpad$body$i$indexZ2D = $11;
  } else {
   __THREW__ = 0;
   invoke_viii(3594,($1|0),($0|0),($3|0)); //@line 341 "modules/graphics/opengl/wrap_Mesh.cpp"
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if ($8) {
    $9 = ___cxa_find_matching_catch_3(8992|0)|0;
    $10 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 341 "modules/graphics/opengl/wrap_Mesh.cpp"
    $eh$lpad$body$i$index2Z2D = $10;$eh$lpad$body$i$indexZ2D = $9;
    break;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 341 "modules/graphics/opengl/wrap_Mesh.cpp"
   STACKTOP = sp;return 0; //@line 342 "modules/graphics/opengl/wrap_Mesh.cpp"
  }
 } while(0);
 $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $14 = ($eh$lpad$body$i$index2Z2D|0)==($13|0); //@line 535 "./common/runtime.h"
 if (!($14)) {
  $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $15 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
 $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
 $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
 $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
 $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if (!($21)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 342 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 $23 = ___cxa_find_matching_catch_2()|0;
 $24 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if ($26) {
  $27 = ___cxa_find_matching_catch_3(0|0)|0;
  $28 = tempRet0;
  ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $24;$$02$i = $23;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl12w_Mesh_flushEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 __ZN4love8graphics6opengl4Mesh5flushEv($0); //@line 348 "modules/graphics/opengl/wrap_Mesh.cpp"
 return 0; //@line 349 "modules/graphics/opengl/wrap_Mesh.cpp"
}
function __ZN4love8graphics6opengl19w_Mesh_setVertexMapEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$01$i7 = 0, $$02$i = 0, $$02$i6 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index7Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i$0 = 0, $i1$023 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i11 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i12 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vertexmap = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $vertexmap = sp + 20|0;
 $0 = sp + 16|0;
 $1 = sp + 12|0;
 $2 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 $3 = (_lua_type($L,2)|0); //@line 356 "modules/graphics/opengl/wrap_Mesh.cpp"
 $4 = ($3|0)<(1); //@line 356 "modules/graphics/opengl/wrap_Mesh.cpp"
 if ($4) {
  __THREW__ = 0;
  invoke_vi(3595,($2|0)); //@line 359 "modules/graphics/opengl/wrap_Mesh.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   STACKTOP = sp;return 0; //@line 386 "modules/graphics/opengl/wrap_Mesh.cpp"
  }
  $7 = ___cxa_find_matching_catch_3(8992|0)|0;
  $8 = tempRet0;
  $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
  $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
  if (!($10)) {
   $$01$i = $8;$$02$i = $7;
   ___resumeException($$02$i|0);
   // unreachable;
  }
  $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
  $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
  $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
  $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
  $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
  __THREW__ = 0;
  invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
  $16 = __THREW__; __THREW__ = 0;
  $17 = $16&1;
  if (!($17)) {
   ___cxa_end_catch(); //@line 540 "./common/runtime.h"
   $18 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
   HEAP32[$vararg_buffer>>2] = $18; //@line 543 "./common/runtime.h"
   (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
   STACKTOP = sp;return 0; //@line 386 "modules/graphics/opengl/wrap_Mesh.cpp"
  }
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  __THREW__ = 0;
  invoke_v(3378); //@line 540 "./common/runtime.h"
  $21 = __THREW__; __THREW__ = 0;
  $22 = $21&1;
  if ($22) {
   $23 = ___cxa_find_matching_catch_3(0|0)|0;
   $24 = tempRet0;
   ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
   // unreachable; //@line 540 "./common/runtime.h"
  } else {
   $$01$i = $20;$$02$i = $19;
   ___resumeException($$02$i|0);
   // unreachable;
  }
 }
 $25 = (_lua_type($L,2)|0); //@line 363 "modules/graphics/opengl/wrap_Mesh.cpp"
 $26 = ($25|0)==(5); //@line 363 "modules/graphics/opengl/wrap_Mesh.cpp"
 if ($26) {
  $27 = (__ZN4love11luax_objlenEP9lua_Statei($L,2)|0); //@line 364 "modules/graphics/opengl/wrap_Mesh.cpp"
  $34 = $27;
 } else {
  $28 = (_lua_gettop($L)|0); //@line 364 "modules/graphics/opengl/wrap_Mesh.cpp"
  $29 = (($28) + -1)|0; //@line 364 "modules/graphics/opengl/wrap_Mesh.cpp"
  $34 = $29;
 }
 HEAP32[$vertexmap>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = ((($vertexmap)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$30>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = ((($vertexmap)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$31>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $32 = ((($vertexmap)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = ($34|0)==(0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($33) {
  label = 15;
 } else {
  $35 = $34 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  __THREW__ = 0;
  $36 = (invoke_ii(3348,($35|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $37 = __THREW__; __THREW__ = 0;
  $38 = $37&1;
  if ($38) {
   label = 20;
  } else {
   $39 = $36; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $40 = (($36) + ($34<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $41 = $40; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   HEAP32[$vertexmap>>2] = $39; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$30>>2] = $39; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$32>>2] = $41; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   label = 15;
  }
 }
 L24: do {
  if ((label|0) == 15) {
   L26: do {
    if ($26) {
     $i$0 = 0;
     while(1) {
      $43 = ($i$0|0)<($34|0); //@line 371 "modules/graphics/opengl/wrap_Mesh.cpp"
      if (!($43)) {
       break L26;
      }
      $59 = (($i$0) + 1)|0; //@line 373 "modules/graphics/opengl/wrap_Mesh.cpp"
      __THREW__ = 0;
      invoke_viii(3382,($L|0),2,($59|0)); //@line 373 "modules/graphics/opengl/wrap_Mesh.cpp"
      $60 = __THREW__; __THREW__ = 0;
      $61 = $60&1;
      if ($61) {
       break;
      }
      __THREW__ = 0;
      $62 = (invoke_iii(3596,($L|0),-1)|0); //@line 374 "modules/graphics/opengl/wrap_Mesh.cpp"
      $63 = __THREW__; __THREW__ = 0;
      $64 = $63&1;
      if ($64) {
       break;
      }
      $65 = (($62) + -1)|0; //@line 374 "modules/graphics/opengl/wrap_Mesh.cpp"
      HEAP32[$0>>2] = $65; //@line 374 "modules/graphics/opengl/wrap_Mesh.cpp"
      $66 = HEAP32[$30>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $67 = HEAP32[$32>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $68 = ($66>>>0)<($67>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($68) {
       HEAP32[$66>>2] = $65; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $69 = ((($66)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$30>>2] = $69; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      } else {
       __THREW__ = 0;
       invoke_vii(3597,($vertexmap|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $70 = __THREW__; __THREW__ = 0;
       $71 = $70&1;
       if ($71) {
        break;
       }
      }
      __THREW__ = 0;
      invoke_vii(3384,($L|0),-2); //@line 375 "modules/graphics/opengl/wrap_Mesh.cpp"
      $72 = __THREW__; __THREW__ = 0;
      $73 = $72&1;
      if ($73) {
       break;
      } else {
       $i$0 = $59;
      }
     }
     $44 = ___cxa_find_matching_catch_2()|0;
     $45 = tempRet0;
     $eh$lpad$body$index7Z2D = $45;$eh$lpad$body$indexZ2D = $44;
     break L24;
    } else {
     $42 = ($34|0)>(0); //@line 380 "modules/graphics/opengl/wrap_Mesh.cpp"
     if ($42) {
      $i1$023 = 0;
      while(1) {
       $74 = (($i1$023) + 2)|0; //@line 381 "modules/graphics/opengl/wrap_Mesh.cpp"
       __THREW__ = 0;
       $75 = (invoke_iii(3596,($L|0),($74|0))|0); //@line 381 "modules/graphics/opengl/wrap_Mesh.cpp"
       $76 = __THREW__; __THREW__ = 0;
       $77 = $76&1;
       if ($77) {
        break;
       }
       $78 = (($75) + -1)|0; //@line 381 "modules/graphics/opengl/wrap_Mesh.cpp"
       HEAP32[$1>>2] = $78; //@line 381 "modules/graphics/opengl/wrap_Mesh.cpp"
       $79 = HEAP32[$30>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $80 = HEAP32[$32>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $81 = ($79>>>0)<($80>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($81) {
        HEAP32[$79>>2] = $78; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $82 = ((($79)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$30>>2] = $82; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       } else {
        __THREW__ = 0;
        invoke_vii(3597,($vertexmap|0),($1|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $83 = __THREW__; __THREW__ = 0;
        $84 = $83&1;
        if ($84) {
         break;
        }
       }
       $85 = (($i1$023) + 1)|0; //@line 380 "modules/graphics/opengl/wrap_Mesh.cpp"
       $86 = ($85|0)<($34|0); //@line 380 "modules/graphics/opengl/wrap_Mesh.cpp"
       if ($86) {
        $i1$023 = $85;
       } else {
        break L26;
       }
      }
      $46 = ___cxa_find_matching_catch_2()|0;
      $47 = tempRet0;
      $eh$lpad$body$index7Z2D = $47;$eh$lpad$body$indexZ2D = $46;
      break L24;
     }
    }
   } while(0);
   __THREW__ = 0;
   invoke_vii(3598,($2|0),($vertexmap|0)); //@line 384 "modules/graphics/opengl/wrap_Mesh.cpp"
   $87 = __THREW__; __THREW__ = 0;
   $88 = $87&1;
   L47: do {
    if ($88) {
     $89 = ___cxa_find_matching_catch_3(8992|0)|0;
     $90 = tempRet0;
     $91 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
     $92 = ($90|0)==($91|0); //@line 535 "./common/runtime.h"
     do {
      if ($92) {
       $93 = (___cxa_begin_catch(($89|0))|0); //@line 535 "./common/runtime.h"
       $94 = HEAP32[$93>>2]|0; //@line 539 "./common/runtime.h"
       $95 = ((($94)) + 8|0); //@line 539 "./common/runtime.h"
       $96 = HEAP32[$95>>2]|0; //@line 539 "./common/runtime.h"
       $97 = (FUNCTION_TABLE_ii[$96 & 8191]($93)|0); //@line 539 "./common/runtime.h"
       __THREW__ = 0;
       invoke_vii(3380,($L|0),($97|0)); //@line 539 "./common/runtime.h"
       $98 = __THREW__; __THREW__ = 0;
       $99 = $98&1;
       if ($99) {
        $107 = ___cxa_find_matching_catch_2()|0;
        $108 = tempRet0;
        __THREW__ = 0;
        invoke_v(3378); //@line 540 "./common/runtime.h"
        $109 = __THREW__; __THREW__ = 0;
        $110 = $109&1;
        if (!($110)) {
         $$01$i7 = $108;$$02$i6 = $107;
         break;
        }
        $111 = ___cxa_find_matching_catch_3(0|0)|0;
        $112 = tempRet0;
        ___clang_call_terminate($111); //@line 540 "./common/runtime.h"
        // unreachable; //@line 540 "./common/runtime.h"
       }
       __THREW__ = 0;
       invoke_v(3378); //@line 540 "./common/runtime.h"
       $100 = __THREW__; __THREW__ = 0;
       $101 = $100&1;
       if ($101) {
        label = 20;
        break L24;
       }
       __THREW__ = 0;
       $102 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
       $103 = __THREW__; __THREW__ = 0;
       $104 = $103&1;
       if ($104) {
        label = 20;
        break L24;
       }
       __THREW__ = 0;
       HEAP32[$vararg_buffer1>>2] = $102; //@line 543 "./common/runtime.h"
       (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer1|0))|0); //@line 543 "./common/runtime.h"
       $105 = __THREW__; __THREW__ = 0;
       $106 = $105&1;
       if ($106) {
        label = 20;
        break L24;
       } else {
        break L47;
       }
      } else {
       $$01$i7 = $90;$$02$i6 = $89;
      }
     } while(0);
     $eh$lpad$body$index7Z2D = $$01$i7;$eh$lpad$body$indexZ2D = $$02$i6;
     break L24;
    }
   } while(0);
   $113 = HEAP32[$vertexmap>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $114 = ($113|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $115 = $113; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($114)) {
    $116 = HEAP32[$30>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $117 = ($116|0)==($113|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($117)) {
     $scevgep$i$i$i11 = ((($116)) + -4|0);
     $118 = $scevgep$i$i$i11;
     $119 = (($118) - ($115))|0;
     $120 = $119 >>> 2;
     $121 = $120 ^ -1;
     $scevgep4$i$i$i12 = (($116) + ($121<<2)|0);
     HEAP32[$30>>2] = $scevgep4$i$i$i12; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($113); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   STACKTOP = sp;return 0; //@line 386 "modules/graphics/opengl/wrap_Mesh.cpp"
  }
 } while(0);
 if ((label|0) == 20) {
  $48 = ___cxa_find_matching_catch_2()|0;
  $49 = tempRet0;
  $eh$lpad$body$index7Z2D = $49;$eh$lpad$body$indexZ2D = $48;
 }
 $50 = HEAP32[$vertexmap>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $51 = ($50|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $52 = $50; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($51) {
  ___resumeException($eh$lpad$body$indexZ2D|0);
  // unreachable;
 }
 $53 = HEAP32[$30>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $54 = ($53|0)==($50|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($54)) {
  $scevgep$i$i$i = ((($53)) + -4|0);
  $55 = $scevgep$i$i$i;
  $56 = (($55) - ($52))|0;
  $57 = $56 >>> 2;
  $58 = $57 ^ -1;
  $scevgep4$i$i$i = (($53) + ($58<<2)|0);
  HEAP32[$30>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($50); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__16vectorIjNS_9allocatorIjEEE21__push_back_slow_pathIjEEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl19w_Mesh_getVertexMapEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index3Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i$0 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i3 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i4 = 0, $vararg_buffer = 0;
 var $vertex_map = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $vertex_map = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 HEAP32[$vertex_map>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($vertex_map)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($vertex_map)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 $3 = (invoke_iii(3599,($0|0),($vertex_map|0))|0); //@line 394 "modules/graphics/opengl/wrap_Mesh.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 L1: do {
  if ($5) {
   $6 = ___cxa_find_matching_catch_3(8992|0)|0;
   $7 = tempRet0;
   $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
   do {
    if ($9) {
     $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
     $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
     $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
     $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
     $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
     __THREW__ = 0;
     invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
     $15 = __THREW__; __THREW__ = 0;
     $16 = $15&1;
     if ($16) {
      $24 = ___cxa_find_matching_catch_2()|0;
      $25 = tempRet0;
      __THREW__ = 0;
      invoke_v(3378); //@line 540 "./common/runtime.h"
      $26 = __THREW__; __THREW__ = 0;
      $27 = $26&1;
      if (!($27)) {
       $$01$i = $25;$$02$i = $24;
       break;
      }
      $28 = ___cxa_find_matching_catch_3(0|0)|0;
      $29 = tempRet0;
      ___clang_call_terminate($28); //@line 540 "./common/runtime.h"
      // unreachable; //@line 540 "./common/runtime.h"
     }
     __THREW__ = 0;
     invoke_v(3378); //@line 540 "./common/runtime.h"
     $17 = __THREW__; __THREW__ = 0;
     $18 = $17&1;
     if ($18) {
      label = 13;
      break L1;
     }
     __THREW__ = 0;
     $19 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
     $20 = __THREW__; __THREW__ = 0;
     $21 = $20&1;
     if ($21) {
      label = 13;
      break L1;
     }
     __THREW__ = 0;
     HEAP32[$vararg_buffer>>2] = $19; //@line 543 "./common/runtime.h"
     (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 543 "./common/runtime.h"
     $22 = __THREW__; __THREW__ = 0;
     $23 = $22&1;
     if ($23) {
      label = 13;
      break L1;
     } else {
      label = 11;
      break L1;
     }
    } else {
     $$01$i = $7;$$02$i = $6;
    }
   } while(0);
   $eh$lpad$body$index3Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
  } else {
   if ($3) {
    $45 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $46 = HEAP32[$vertex_map>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $47 = (($45) - ($46))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $48 = $47 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __THREW__ = 0;
    invoke_viii(3393,($L|0),($48|0),0); //@line 404 "modules/graphics/opengl/wrap_Mesh.cpp"
    $49 = __THREW__; __THREW__ = 0;
    $50 = $49&1;
    if ($50) {
     label = 13;
    } else {
     $i$0 = 0;
     while(1) {
      $51 = ($i$0|0)<($48|0); //@line 406 "modules/graphics/opengl/wrap_Mesh.cpp"
      if (!($51)) {
       label = 23;
       break L1;
      }
      $52 = HEAP32[$vertex_map>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $53 = (($52) + ($i$0<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $54 = HEAP32[$53>>2]|0; //@line 408 "modules/graphics/opengl/wrap_Mesh.cpp"
      $55 = (($54) + 1)|0; //@line 408 "modules/graphics/opengl/wrap_Mesh.cpp"
      __THREW__ = 0;
      invoke_vii(3468,($L|0),($55|0)); //@line 408 "modules/graphics/opengl/wrap_Mesh.cpp"
      $56 = __THREW__; __THREW__ = 0;
      $57 = $56&1;
      if ($57) {
       break;
      }
      $58 = (($i$0) + 1)|0; //@line 409 "modules/graphics/opengl/wrap_Mesh.cpp"
      __THREW__ = 0;
      invoke_viii(3395,($L|0),-2,($58|0)); //@line 409 "modules/graphics/opengl/wrap_Mesh.cpp"
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if ($60) {
       break;
      } else {
       $i$0 = $58;
      }
     }
     $32 = ___cxa_find_matching_catch_2()|0;
     $33 = tempRet0;
     $eh$lpad$body$index3Z2D = $33;$eh$lpad$body$indexZ2D = $32;
    }
   } else {
    label = 11;
   }
  }
 } while(0);
 if ((label|0) == 11) {
  __THREW__ = 0;
  invoke_vi(3577,($L|0)); //@line 398 "modules/graphics/opengl/wrap_Mesh.cpp"
  $30 = __THREW__; __THREW__ = 0;
  $31 = $30&1;
  if ($31) {
   label = 13;
  } else {
   label = 23;
  }
 }
 if ((label|0) == 13) {
  $34 = ___cxa_find_matching_catch_2()|0;
  $35 = tempRet0;
  $eh$lpad$body$index3Z2D = $35;$eh$lpad$body$indexZ2D = $34;
 }
 else if ((label|0) == 23) {
  $61 = HEAP32[$vertex_map>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $62 = ($61|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $63 = $61; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($62) {
   STACKTOP = sp;return 1; //@line 413 "modules/graphics/opengl/wrap_Mesh.cpp"
  }
  $64 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $65 = ($64|0)==($61|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($65)) {
   $scevgep$i$i$i = ((($64)) + -4|0);
   $66 = $scevgep$i$i$i;
   $67 = (($66) - ($63))|0;
   $68 = $67 >>> 2;
   $69 = $68 ^ -1;
   $scevgep4$i$i$i = (($64) + ($69<<2)|0);
   HEAP32[$1>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($61); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  STACKTOP = sp;return 1; //@line 413 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 $36 = HEAP32[$vertex_map>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $37 = ($36|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $38 = $36; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($37) {
  ___resumeException($eh$lpad$body$indexZ2D|0);
  // unreachable;
 }
 $39 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $40 = ($39|0)==($36|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($40)) {
  $scevgep$i$i$i3 = ((($39)) + -4|0);
  $41 = $scevgep$i$i$i3;
  $42 = (($41) - ($38))|0;
  $43 = $42 >>> 2;
  $44 = $43 ^ -1;
  $scevgep4$i$i$i4 = (($39) + ($44<<2)|0);
  HEAP32[$1>>2] = $scevgep4$i$i$i4; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($36); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8graphics6opengl17w_Mesh_setTextureEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 $1 = (_lua_type($L,2)|0); //@line 419 "modules/graphics/opengl/wrap_Mesh.cpp"
 $2 = ($1|0)<(1); //@line 419 "modules/graphics/opengl/wrap_Mesh.cpp"
 if ($2) {
  __ZN4love8graphics6opengl4Mesh10setTextureEv($0); //@line 420 "modules/graphics/opengl/wrap_Mesh.cpp"
  return 0; //@line 427 "modules/graphics/opengl/wrap_Mesh.cpp"
 } else {
  $3 = (__ZN4love8graphics17luax_checktextureEP9lua_Statei($L,2)|0); //@line 423 "modules/graphics/opengl/wrap_Mesh.cpp"
  __ZN4love8graphics6opengl4Mesh10setTextureEPNS0_7TextureE($0,$3); //@line 424 "modules/graphics/opengl/wrap_Mesh.cpp"
  return 0; //@line 427 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl17w_Mesh_getTextureEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 430 "modules/graphics/opengl/wrap_Mesh.cpp"
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 $1 = (__ZNK4love8graphics6opengl4Mesh10getTextureEv($0)|0); //@line 433 "modules/graphics/opengl/wrap_Mesh.cpp"
 $2 = ($1|0)==(0|0); //@line 435 "modules/graphics/opengl/wrap_Mesh.cpp"
 if ($2) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 447 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 $3 = HEAP32[$1>>2]|0; //@line 439 "modules/graphics/opengl/wrap_Mesh.cpp"
 $4 = ((($3)) + -4|0); //@line 439 "modules/graphics/opengl/wrap_Mesh.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 439 "modules/graphics/opengl/wrap_Mesh.cpp"
 $6 = ((($5)) + 4|0); //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 $7 = HEAP32[$6>>2]|0; //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 $8 = HEAP32[(1972)>>2]|0; //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 $9 = ($7|0)==($8|0); //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 if ($9) {
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,12,$1); //@line 440 "modules/graphics/opengl/wrap_Mesh.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 447 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 $10 = HEAP32[(1860)>>2]|0; //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 $11 = ($7|0)==($10|0); //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 if ($11) {
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,17,$1); //@line 442 "modules/graphics/opengl/wrap_Mesh.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 447 "modules/graphics/opengl/wrap_Mesh.cpp"
 } else {
  $12 = (_luaL_error($L,345068,$vararg_buffer)|0); //@line 444 "modules/graphics/opengl/wrap_Mesh.cpp"
  $$0 = $12;
  STACKTOP = sp;return ($$0|0); //@line 447 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl18w_Mesh_setDrawModeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $mode = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $mode = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 $1 = (_luaL_checklstring($L,2,0)|0); //@line 452 "modules/graphics/opengl/wrap_Mesh.cpp"
 $2 = (__ZN4love8graphics6opengl4Mesh11getConstantEPKcRNS2_8DrawModeE($1,$mode)|0); //@line 455 "modules/graphics/opengl/wrap_Mesh.cpp"
 if ($2) {
  $4 = HEAP32[$mode>>2]|0; //@line 458 "modules/graphics/opengl/wrap_Mesh.cpp"
  __ZN4love8graphics6opengl4Mesh11setDrawModeENS2_8DrawModeE($0,$4); //@line 458 "modules/graphics/opengl/wrap_Mesh.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 460 "modules/graphics/opengl/wrap_Mesh.cpp"
 } else {
  HEAP32[$vararg_buffer>>2] = $1; //@line 456 "modules/graphics/opengl/wrap_Mesh.cpp"
  $3 = (_luaL_error($L,343313,$vararg_buffer)|0); //@line 456 "modules/graphics/opengl/wrap_Mesh.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 460 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl18w_Mesh_getDrawModeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $str = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $str = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 $1 = (__ZNK4love8graphics6opengl4Mesh11getDrawModeEv($0)|0); //@line 465 "modules/graphics/opengl/wrap_Mesh.cpp"
 $2 = (__ZN4love8graphics6opengl4Mesh11getConstantENS2_8DrawModeERPKc($1,$str)|0); //@line 468 "modules/graphics/opengl/wrap_Mesh.cpp"
 if ($2) {
  $4 = HEAP32[$str>>2]|0; //@line 471 "modules/graphics/opengl/wrap_Mesh.cpp"
  _lua_pushstring($L,$4); //@line 471 "modules/graphics/opengl/wrap_Mesh.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 473 "modules/graphics/opengl/wrap_Mesh.cpp"
 } else {
  $3 = (_luaL_error($L,343340,$vararg_buffer)|0); //@line 469 "modules/graphics/opengl/wrap_Mesh.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 473 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl19w_Mesh_setDrawRangeEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 475 "modules/graphics/opengl/wrap_Mesh.cpp"
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 $1 = (_lua_type($L,2)|0); //@line 479 "modules/graphics/opengl/wrap_Mesh.cpp"
 $2 = ($1|0)<(1); //@line 479 "modules/graphics/opengl/wrap_Mesh.cpp"
 if ($2) {
  __ZN4love8graphics6opengl4Mesh12setDrawRangeEv($0); //@line 480 "modules/graphics/opengl/wrap_Mesh.cpp"
  STACKTOP = sp;return 0; //@line 488 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 $3 = (+_luaL_checknumber($L,2)); //@line 483 "modules/graphics/opengl/wrap_Mesh.cpp"
 $4 = (~~(($3))); //@line 483 "modules/graphics/opengl/wrap_Mesh.cpp"
 $5 = (($4) + -1)|0; //@line 483 "modules/graphics/opengl/wrap_Mesh.cpp"
 $6 = (+_luaL_checknumber($L,3)); //@line 484 "modules/graphics/opengl/wrap_Mesh.cpp"
 $7 = (~~(($6))); //@line 484 "modules/graphics/opengl/wrap_Mesh.cpp"
 $8 = (($7) + -1)|0; //@line 484 "modules/graphics/opengl/wrap_Mesh.cpp"
 __THREW__ = 0;
 invoke_viii(3600,($0|0),($5|0),($8|0)); //@line 485 "modules/graphics/opengl/wrap_Mesh.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if (!($10)) {
  STACKTOP = sp;return 0; //@line 488 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 $11 = ___cxa_find_matching_catch_3(8992|0)|0;
 $12 = tempRet0;
 $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $14 = ($12|0)==($13|0); //@line 535 "./common/runtime.h"
 if (!($14)) {
  $$01$i = $12;$$02$i = $11;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $15 = (___cxa_begin_catch(($11|0))|0); //@line 535 "./common/runtime.h"
 $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
 $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
 $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
 $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if (!($21)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 488 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 $23 = ___cxa_find_matching_catch_2()|0;
 $24 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if ($26) {
  $27 = ___cxa_find_matching_catch_3(0|0)|0;
  $28 = tempRet0;
  ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $24;$$02$i = $23;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl19w_Mesh_getDrawRangeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $rangemax = 0, $rangemin = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $rangemin = sp + 4|0;
 $rangemax = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,1,19)|0); //@line 40 "modules/graphics/opengl/wrap_Mesh.cpp"
 HEAP32[$rangemin>>2] = -1; //@line 495 "modules/graphics/opengl/wrap_Mesh.cpp"
 HEAP32[$rangemax>>2] = -1; //@line 496 "modules/graphics/opengl/wrap_Mesh.cpp"
 __ZNK4love8graphics6opengl4Mesh12getDrawRangeERiS3_($0,$rangemin,$rangemax); //@line 497 "modules/graphics/opengl/wrap_Mesh.cpp"
 $1 = HEAP32[$rangemin>>2]|0; //@line 499 "modules/graphics/opengl/wrap_Mesh.cpp"
 $2 = HEAP32[$rangemax>>2]|0; //@line 499 "modules/graphics/opengl/wrap_Mesh.cpp"
 $3 = $2 | $1; //@line 499 "modules/graphics/opengl/wrap_Mesh.cpp"
 $4 = ($3|0)<(0); //@line 499 "modules/graphics/opengl/wrap_Mesh.cpp"
 if ($4) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 505 "modules/graphics/opengl/wrap_Mesh.cpp"
 }
 $5 = (($1) + 1)|0; //@line 502 "modules/graphics/opengl/wrap_Mesh.cpp"
 _lua_pushinteger($L,$5); //@line 502 "modules/graphics/opengl/wrap_Mesh.cpp"
 $6 = HEAP32[$rangemax>>2]|0; //@line 503 "modules/graphics/opengl/wrap_Mesh.cpp"
 $7 = (($6) + 1)|0; //@line 503 "modules/graphics/opengl/wrap_Mesh.cpp"
 _lua_pushinteger($L,$7); //@line 503 "modules/graphics/opengl/wrap_Mesh.cpp"
 $$0 = 2;
 STACKTOP = sp;return ($$0|0); //@line 505 "modules/graphics/opengl/wrap_Mesh.cpp"
}
function _luaopen_mesh($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 531 "modules/graphics/opengl/wrap_Mesh.cpp"
 HEAP32[$vararg_buffer>>2] = 14168; //@line 533 "modules/graphics/opengl/wrap_Mesh.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 533 "modules/graphics/opengl/wrap_Mesh.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 533 "modules/graphics/opengl/wrap_Mesh.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,19,343364,$vararg_buffer)|0); //@line 533 "modules/graphics/opengl/wrap_Mesh.cpp"
 STACKTOP = sp;return ($0|0); //@line 533 "modules/graphics/opengl/wrap_Mesh.cpp"
}
function __ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love8graphics6opengl22w_ParticleSystem_cloneEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $clone$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 47 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 52 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = ((($1)) + 20|0); //@line 52 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 52 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __THREW__ = 0;
 $4 = (invoke_ii($3|0,($0|0))|0); //@line 52 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 do {
  if ($6) {
   $7 = ___cxa_find_matching_catch_3(8992|0)|0;
   $8 = tempRet0;
   $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
   if (!($10)) {
    $$01$i = $8;$$02$i = $7;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
   $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
   $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
   $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
   $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
   $16 = __THREW__; __THREW__ = 0;
   $17 = $16&1;
   if (!($17)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $18 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $18; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $clone$0 = 0;
    break;
   }
   $19 = ___cxa_find_matching_catch_2()|0;
   $20 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $21 = __THREW__; __THREW__ = 0;
   $22 = $21&1;
   if ($22) {
    $23 = ___cxa_find_matching_catch_3(0|0)|0;
    $24 = tempRet0;
    ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $20;$$02$i = $19;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $clone$0 = $4;
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,15,$clone$0); //@line 54 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $25 = HEAP32[$clone$0>>2]|0; //@line 55 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $26 = ((($25)) + 12|0); //@line 55 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $27 = HEAP32[$26>>2]|0; //@line 55 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 FUNCTION_TABLE_vi[$27 & 8191]($clone$0); //@line 55 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 STACKTOP = sp;return 1; //@line 56 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl27w_ParticleSystem_setTextureEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (__ZN4love8graphics17luax_checktextureEP9lua_Statei($L,2)|0); //@line 62 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem10setTextureEPNS0_7TextureE($0,$1); //@line 63 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 64 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl27w_ParticleSystem_getTextureEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 67 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (__ZNK4love8graphics14ParticleSystem10getTextureEv($0)|0); //@line 70 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = ($1|0)==(0|0); //@line 73 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 if ($2) {
  ___cxa_bad_typeid(); //@line 73 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  // unreachable; //@line 73 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 }
 $3 = HEAP32[$1>>2]|0; //@line 73 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = ((($3)) + -4|0); //@line 73 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 73 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $6 = ((($5)) + 4|0); //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 $7 = HEAP32[$6>>2]|0; //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 $8 = HEAP32[(1972)>>2]|0; //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 $9 = ($7|0)==($8|0); //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 if ($9) {
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,12,$1); //@line 74 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 81 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 }
 $10 = HEAP32[(1860)>>2]|0; //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 $11 = ($7|0)==($10|0); //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 if ($11) {
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,17,$1); //@line 76 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 81 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 } else {
  $12 = (_luaL_error($L,345068,$vararg_buffer)|0); //@line 78 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $$0 = $12;
  STACKTOP = sp;return ($$0|0); //@line 81 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl30w_ParticleSystem_setBufferSizeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 83 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $vararg_buffer = sp; //@line 83 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 86 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1 < 1.0; //@line 87 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = $1 > 536870911.0; //@line 87 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $or$cond = $2 | $3; //@line 87 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 if ($or$cond) {
  $4 = (_luaL_error($L,347355,$vararg_buffer)|0); //@line 88 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $$0 = $4;
  STACKTOP = sp;return ($$0|0); //@line 92 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 }
 $5 = HEAP32[$0>>2]|0; //@line 90 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $6 = ((($5)) + 24|0); //@line 90 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 90 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $8 = (~~(($1))>>>0); //@line 90 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __THREW__ = 0;
 invoke_vii($7|0,($0|0),($8|0)); //@line 90 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if (!($10)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 92 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 }
 $11 = ___cxa_find_matching_catch_3(8992|0)|0;
 $12 = tempRet0;
 $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $14 = ($12|0)==($13|0); //@line 535 "./common/runtime.h"
 if (!($14)) {
  $$01$i = $12;$$02$i = $11;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $15 = (___cxa_begin_catch(($11|0))|0); //@line 535 "./common/runtime.h"
 $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
 $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
 $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
 $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if (!($21)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $22; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 92 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 }
 $23 = ___cxa_find_matching_catch_2()|0;
 $24 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if ($26) {
  $27 = ___cxa_find_matching_catch_3(0|0)|0;
  $28 = tempRet0;
  ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $24;$$02$i = $23;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl30w_ParticleSystem_getBufferSizeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (__ZNK4love8graphics14ParticleSystem13getBufferSizeEv($0)|0); //@line 97 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushinteger($L,$1); //@line 97 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 1; //@line 98 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl30w_ParticleSystem_setInsertModeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $mode = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $mode = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (_luaL_checklstring($L,2,0)|0); //@line 105 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = (__ZN4love8graphics14ParticleSystem11getConstantEPKcRNS1_10InsertModeE($1,$mode)|0); //@line 106 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 if ($2) {
  $4 = HEAP32[$mode>>2]|0; //@line 108 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  __ZN4love8graphics14ParticleSystem13setInsertModeENS1_10InsertModeE($0,$4); //@line 108 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 110 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 } else {
  HEAP32[$vararg_buffer>>2] = $1; //@line 107 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $3 = (_luaL_error($L,343570,$vararg_buffer)|0); //@line 107 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 110 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl30w_ParticleSystem_getInsertModeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $str = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $str = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (__ZNK4love8graphics14ParticleSystem13getInsertModeEv($0)|0); //@line 116 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = (__ZN4love8graphics14ParticleSystem11getConstantENS1_10InsertModeERPKc($1,$str)|0); //@line 118 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 if ($2) {
  $4 = HEAP32[$str>>2]|0; //@line 120 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  _lua_pushstring($L,$4); //@line 120 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 122 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 } else {
  $3 = (_luaL_error($L,343596,$vararg_buffer)|0); //@line 119 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 122 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl32w_ParticleSystem_setEmissionRateEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 124 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 127 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 127 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __THREW__ = 0;
 invoke_vid(3601,($0|0),(+$2)); //@line 128 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  STACKTOP = sp;return 0; //@line 129 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 }
 $5 = ___cxa_find_matching_catch_3(8992|0)|0;
 $6 = tempRet0;
 $7 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $8 = ($6|0)==($7|0); //@line 535 "./common/runtime.h"
 if (!($8)) {
  $$01$i = $6;$$02$i = $5;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $9 = (___cxa_begin_catch(($5|0))|0); //@line 535 "./common/runtime.h"
 $10 = HEAP32[$9>>2]|0; //@line 539 "./common/runtime.h"
 $11 = ((($10)) + 8|0); //@line 539 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = (FUNCTION_TABLE_ii[$12 & 8191]($9)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($13|0)); //@line 539 "./common/runtime.h"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if (!($15)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $16 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $16; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 129 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 }
 $17 = ___cxa_find_matching_catch_2()|0;
 $18 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $21 = ___cxa_find_matching_catch_3(0|0)|0;
  $22 = tempRet0;
  ___clang_call_terminate($21); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $18;$$02$i = $17;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl32w_ParticleSystem_getEmissionRateEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+__ZNK4love8graphics14ParticleSystem15getEmissionRateEv($0)); //@line 135 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 135 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$2); //@line 135 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 1; //@line 136 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl35w_ParticleSystem_setEmitterLifetimeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 142 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 142 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem18setEmitterLifetimeEf($0,$2); //@line 143 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 144 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl35w_ParticleSystem_getEmitterLifetimeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+__ZNK4love8graphics14ParticleSystem18getEmitterLifetimeEv($0)); //@line 150 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 150 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$2); //@line 150 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 1; //@line 151 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl36w_ParticleSystem_setParticleLifetimeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 157 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 157 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = $2; //@line 158 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = (+_luaL_optnumber($L,3,$3)); //@line 158 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $5 = $4; //@line 158 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem19setParticleLifetimeEff($0,$2,$5); //@line 159 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 160 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl36w_ParticleSystem_getParticleLifetimeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $max = 0, $min = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $min = sp + 4|0;
 $max = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZNK4love8graphics14ParticleSystem19getParticleLifetimeERfS2_($0,$min,$max); //@line 167 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = +HEAPF32[$min>>2]; //@line 168 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 168 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$2); //@line 168 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = +HEAPF32[$max>>2]; //@line 169 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = $3; //@line 169 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$4); //@line 169 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 STACKTOP = sp;return 2; //@line 170 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl28w_ParticleSystem_setPositionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 176 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 176 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = (+_luaL_checknumber($L,3)); //@line 177 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = $3; //@line 177 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem11setPositionEff($0,$2,$4); //@line 178 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 179 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl28w_ParticleSystem_getPositionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (__ZNK4love8graphics14ParticleSystem11getPositionEv($0)|0); //@line 185 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = +HEAPF32[$1>>2]; //@line 185 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = ((($1)) + 4|0); //@line 185 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = +HEAPF32[$3>>2]; //@line 185 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $5 = $2; //@line 186 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$5); //@line 186 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $6 = $4; //@line 187 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$6); //@line 187 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 2; //@line 188 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl23w_ParticleSystem_moveToEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 194 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 194 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = (+_luaL_checknumber($L,3)); //@line 195 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = $3; //@line 195 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem6moveToEff($0,$2,$4); //@line 196 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 197 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl30w_ParticleSystem_setAreaSpreadEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$pre = 0, $$pre2 = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $distribution = 0, $or$cond = 0;
 var $phitmp = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $x$0 = 0.0, $y$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $distribution = sp + 12|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 HEAP32[$distribution>>2] = 0; //@line 204 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (_lua_type($L,2)|0); //@line 207 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = ($1|0)<(1); //@line 207 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 do {
  if ($2) {
   $14 = 0;$x$0 = 0.0;$y$0 = 0.0;
  } else {
   $3 = (_luaL_checklstring($L,2,0)|0); //@line 207 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
   $4 = ($3|0)==(0|0); //@line 208 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
   if ($4) {
    $14 = 0;$x$0 = 0.0;$y$0 = 0.0;
   } else {
    $5 = (__ZN4love8graphics14ParticleSystem11getConstantEPKcRNS1_22AreaSpreadDistributionE($3,$distribution)|0); //@line 208 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    if (!($5)) {
     HEAP32[$vararg_buffer>>2] = $3; //@line 209 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
     $6 = (_luaL_error($L,343616,$vararg_buffer)|0); //@line 209 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
     $$0 = $6;
     STACKTOP = sp;return ($$0|0); //@line 221 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    }
    $$pre = HEAP32[$distribution>>2]|0; //@line 211 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $phitmp = ($$pre|0)==(0); //@line 208 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    if ($phitmp) {
     $14 = 0;$x$0 = 0.0;$y$0 = 0.0;
    } else {
     $7 = (+_luaL_checknumber($L,3)); //@line 213 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
     $8 = $7; //@line 213 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
     $9 = (+_luaL_checknumber($L,4)); //@line 214 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
     $10 = $9; //@line 214 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
     $11 = $8 < 0.0; //@line 215 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
     $12 = $10 < 0.0; //@line 215 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
     $or$cond = $11 | $12; //@line 215 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
     if (!($or$cond)) {
      $$pre2 = HEAP32[$distribution>>2]|0; //@line 219 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $14 = $$pre2;$x$0 = $8;$y$0 = $10;
      break;
     }
     $13 = (_luaL_error($L,343650,$vararg_buffer1)|0); //@line 216 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
     $$0 = $13;
     STACKTOP = sp;return ($$0|0); //@line 221 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    }
   }
  }
 } while(0);
 __ZN4love8graphics14ParticleSystem13setAreaSpreadENS1_22AreaSpreadDistributionEff($0,$14,$x$0,$y$0); //@line 219 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $$0 = 0;
 STACKTOP = sp;return ($$0|0); //@line 221 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl30w_ParticleSystem_getAreaSpreadEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0.0, $str = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $str = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (__ZNK4love8graphics14ParticleSystem25getAreaSpreadDistributionEv($0)|0); //@line 226 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 (__ZN4love8graphics14ParticleSystem11getConstantENS1_22AreaSpreadDistributionERPKc($1,$str)|0); //@line 228 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = (__ZNK4love8graphics14ParticleSystem23getAreaSpreadParametersEv($0)|0); //@line 229 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = HEAP32[$str>>2]|0; //@line 231 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushstring($L,$3); //@line 231 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = +HEAPF32[$2>>2]; //@line 232 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $5 = $4; //@line 232 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$5); //@line 232 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $6 = ((($2)) + 4|0); //@line 233 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 233 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $8 = $7; //@line 233 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$8); //@line 233 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 STACKTOP = sp;return 3; //@line 235 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl29w_ParticleSystem_setDirectionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 241 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 241 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem12setDirectionEf($0,$2); //@line 242 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 243 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl29w_ParticleSystem_getDirectionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+__ZNK4love8graphics14ParticleSystem12getDirectionEv($0)); //@line 249 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 249 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$2); //@line 249 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 1; //@line 250 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl26w_ParticleSystem_setSpreadEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 256 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 256 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem9setSpreadEf($0,$2); //@line 257 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 258 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl26w_ParticleSystem_getSpreadEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+__ZNK4love8graphics14ParticleSystem9getSpreadEv($0)); //@line 264 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 264 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$2); //@line 264 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 1; //@line 265 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl25w_ParticleSystem_setSpeedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 271 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 271 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = $2; //@line 272 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = (+_luaL_optnumber($L,3,$3)); //@line 272 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $5 = $4; //@line 272 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem8setSpeedEff($0,$2,$5); //@line 273 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 274 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl25w_ParticleSystem_getSpeedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $max = 0, $min = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $min = sp + 4|0;
 $max = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZNK4love8graphics14ParticleSystem8getSpeedERfS2_($0,$min,$max); //@line 281 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = +HEAPF32[$min>>2]; //@line 282 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 282 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$2); //@line 282 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = +HEAPF32[$max>>2]; //@line 283 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = $3; //@line 283 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$4); //@line 283 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 STACKTOP = sp;return 2; //@line 284 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl38w_ParticleSystem_setLinearAccelerationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $10 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 290 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 290 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = (+_luaL_checknumber($L,3)); //@line 291 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = $3; //@line 291 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $5 = $2; //@line 292 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $6 = (+_luaL_optnumber($L,4,$5)); //@line 292 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $7 = $6; //@line 292 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $8 = $4; //@line 293 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $9 = (+_luaL_optnumber($L,5,$8)); //@line 293 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $10 = $9; //@line 293 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem21setLinearAccelerationEffff($0,$2,$4,$7,$10); //@line 294 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 295 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl38w_ParticleSystem_getLinearAccelerationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $max = 0, $min = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $min = sp + 8|0;
 $max = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 HEAPF32[$min>>2] = 0.0; //@line 219 "./common/Vector.h"
 $1 = ((($min)) + 4|0); //@line 220 "./common/Vector.h"
 HEAPF32[$1>>2] = 0.0; //@line 220 "./common/Vector.h"
 HEAPF32[$max>>2] = 0.0; //@line 219 "./common/Vector.h"
 $2 = ((($max)) + 4|0); //@line 220 "./common/Vector.h"
 HEAPF32[$2>>2] = 0.0; //@line 220 "./common/Vector.h"
 __ZNK4love8graphics14ParticleSystem21getLinearAccelerationERNS_6VectorES3_($0,$min,$max); //@line 302 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = +HEAPF32[$min>>2]; //@line 303 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = $3; //@line 303 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$4); //@line 303 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $5 = +HEAPF32[$1>>2]; //@line 304 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $6 = $5; //@line 304 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$6); //@line 304 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $7 = +HEAPF32[$max>>2]; //@line 305 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $8 = $7; //@line 305 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$8); //@line 305 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $9 = +HEAPF32[$2>>2]; //@line 306 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $10 = $9; //@line 306 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$10); //@line 306 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 STACKTOP = sp;return 4; //@line 307 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl38w_ParticleSystem_setRadialAccelerationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 313 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 313 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = $2; //@line 314 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = (+_luaL_optnumber($L,3,$3)); //@line 314 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $5 = $4; //@line 314 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem21setRadialAccelerationEff($0,$2,$5); //@line 315 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 316 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl38w_ParticleSystem_getRadialAccelerationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $max = 0, $min = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $min = sp + 4|0;
 $max = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZNK4love8graphics14ParticleSystem21getRadialAccelerationERfS2_($0,$min,$max); //@line 323 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = +HEAPF32[$min>>2]; //@line 324 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 324 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$2); //@line 324 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = +HEAPF32[$max>>2]; //@line 325 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = $3; //@line 325 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$4); //@line 325 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 STACKTOP = sp;return 2; //@line 326 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl42w_ParticleSystem_setTangentialAccelerationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 332 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 332 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = $2; //@line 333 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = (+_luaL_optnumber($L,3,$3)); //@line 333 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $5 = $4; //@line 333 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem25setTangentialAccelerationEff($0,$2,$5); //@line 334 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 335 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl42w_ParticleSystem_getTangentialAccelerationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $max = 0, $min = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $min = sp + 4|0;
 $max = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZNK4love8graphics14ParticleSystem25getTangentialAccelerationERfS2_($0,$min,$max); //@line 342 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = +HEAPF32[$min>>2]; //@line 343 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 343 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$2); //@line 343 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = +HEAPF32[$max>>2]; //@line 344 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = $3; //@line 344 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$4); //@line 344 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 STACKTOP = sp;return 2; //@line 345 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl33w_ParticleSystem_setLinearDampingEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 351 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 351 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = $2; //@line 352 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = (+_luaL_optnumber($L,3,$3)); //@line 352 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $5 = $4; //@line 352 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem16setLinearDampingEff($0,$2,$5); //@line 353 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 354 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl33w_ParticleSystem_getLinearDampingEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $max = 0, $min = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $min = sp + 4|0;
 $max = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZNK4love8graphics14ParticleSystem16getLinearDampingERfS2_($0,$min,$max); //@line 361 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = +HEAPF32[$min>>2]; //@line 362 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 362 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$2); //@line 362 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = +HEAPF32[$max>>2]; //@line 363 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = $3; //@line 363 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$4); //@line 363 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 STACKTOP = sp;return 2; //@line 364 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl25w_ParticleSystem_setSizesEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0, $9 = 0, $i$06 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i1 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i2 = 0, $sizes = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $sizes = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (_lua_gettop($L)|0); //@line 370 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = (($1) + -1)|0; //@line 370 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = ($2>>>0)>(8); //@line 372 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 if ($3) {
  $4 = (_luaL_error($L,343696,$vararg_buffer)|0); //@line 373 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $$0 = $4;
  STACKTOP = sp;return ($$0|0); //@line 389 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 }
 $5 = ($2>>>0)<(2); //@line 375 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 if ($5) {
  $6 = (+_luaL_checknumber($L,2)); //@line 181 "./common/runtime.h"
  $7 = $6; //@line 181 "./common/runtime.h"
  __ZN4love8graphics14ParticleSystem7setSizeEf($0,$7); //@line 378 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 389 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 }
 HEAP32[$sizes>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($sizes)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$8>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ((($sizes)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$9>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $10 = $2 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $11 = (__Znwj($10)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 HEAP32[$sizes>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = (($11) + ($2<<2)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ((($sizes)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$13>>2] = $12; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 _memset(($11|0),0,($10|0))|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$8>>2] = $12; //@line 988 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $i$06 = 0;
 while(1) {
  $16 = (($i$06) + 2)|0; //@line 384 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  __THREW__ = 0;
  $17 = (+invoke_dii(3361,($L|0),($16|0))); //@line 181 "./common/runtime.h"
  $18 = __THREW__; __THREW__ = 0;
  $19 = $18&1;
  if ($19) {
   label = 9;
   break;
  }
  $20 = $17; //@line 181 "./common/runtime.h"
  $21 = (($i$06) + 1)|0; //@line 384 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $22 = HEAP32[$sizes>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = (($22) + ($i$06<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAPF32[$23>>2] = $20; //@line 384 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $24 = ($21>>>0)<($2>>>0); //@line 383 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  if ($24) {
   $i$06 = $21;
  } else {
   label = 6;
   break;
  }
 }
 do {
  if ((label|0) == 6) {
   __THREW__ = 0;
   invoke_vii(3602,($0|0),($sizes|0)); //@line 386 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
   $14 = __THREW__; __THREW__ = 0;
   $15 = $14&1;
   if ($15) {
    $27 = ___cxa_find_matching_catch_2()|0;
    $28 = tempRet0;
    $lpad$phi$index = $27;$lpad$phi$index2 = $28;
    break;
   }
   $38 = HEAP32[$sizes>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $39 = ($38|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $40 = $38; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($39)) {
    $41 = HEAP32[$8>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $42 = ($41|0)==($38|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($42)) {
     $scevgep$i$i$i1 = ((($41)) + -4|0);
     $43 = $scevgep$i$i$i1;
     $44 = (($43) - ($40))|0;
     $45 = $44 >>> 2;
     $46 = $45 ^ -1;
     $scevgep4$i$i$i2 = (($41) + ($46<<2)|0);
     HEAP32[$8>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($38); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $$0 = 0;
   STACKTOP = sp;return ($$0|0); //@line 389 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  }
  else if ((label|0) == 9) {
   $25 = ___cxa_find_matching_catch_2()|0;
   $26 = tempRet0;
   $lpad$phi$index = $25;$lpad$phi$index2 = $26;
  }
 } while(0);
 $29 = HEAP32[$sizes>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = ($29|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $29; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($30) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $32 = HEAP32[$8>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $33 = ($32|0)==($29|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($33)) {
  $scevgep$i$i$i = ((($32)) + -4|0);
  $34 = $scevgep$i$i$i;
  $35 = (($34) - ($31))|0;
  $36 = $35 >>> 2;
  $37 = $36 ^ -1;
  $scevgep4$i$i$i = (($32) + ($37<<2)|0);
  HEAP32[$8>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($29); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8graphics6opengl25w_ParticleSystem_getSizesEP9lua_State($L) {
 $L = $L|0;
 var $$in = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0.0, $i$01 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (__ZNK4love8graphics14ParticleSystem8getSizesEv($0)|0); //@line 394 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = ((($1)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($3|0)==($4|0); //@line 396 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 if ($5) {
  $$lcssa = 0;
  return ($$lcssa|0); //@line 399 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 } else {
  $$in = $4;$i$01 = 0;
 }
 while(1) {
  $6 = $$in; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $7 = (($6) + ($i$01<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $8 = +HEAPF32[$7>>2]; //@line 397 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $9 = $8; //@line 397 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  _lua_pushnumber($L,$9); //@line 397 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $10 = (($i$01) + 1)|0; //@line 396 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $11 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $12 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($11) - ($12))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = ($10>>>0)<($14>>>0); //@line 396 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  if ($15) {
   $$in = $12;$i$01 = $10;
  } else {
   $$lcssa = $14;
   break;
  }
 }
 return ($$lcssa|0); //@line 399 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl33w_ParticleSystem_setSizeVariationEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0, $4 = 0, $5 = 0, $or$cond = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 402 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 405 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 405 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = $2 < 0.0; //@line 406 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = $2 > 1.0; //@line 406 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $or$cond = $3 | $4; //@line 406 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 if ($or$cond) {
  $5 = (_luaL_error($L,343733,$vararg_buffer)|0); //@line 407 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $$0 = $5;
  STACKTOP = sp;return ($$0|0); //@line 411 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 } else {
  __ZN4love8graphics14ParticleSystem16setSizeVariationEf($0,$2); //@line 409 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 411 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl33w_ParticleSystem_getSizeVariationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+__ZNK4love8graphics14ParticleSystem16getSizeVariationEv($0)); //@line 416 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 416 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$2); //@line 416 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 1; //@line 417 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl28w_ParticleSystem_setRotationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 423 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 423 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = $2; //@line 424 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = (+_luaL_optnumber($L,3,$3)); //@line 424 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $5 = $4; //@line 424 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem11setRotationEff($0,$2,$5); //@line 425 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 426 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl28w_ParticleSystem_getRotationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $max = 0, $min = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $min = sp + 4|0;
 $max = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZNK4love8graphics14ParticleSystem11getRotationERfS2_($0,$min,$max); //@line 433 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = +HEAPF32[$min>>2]; //@line 434 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 434 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$2); //@line 434 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = +HEAPF32[$max>>2]; //@line 435 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = $3; //@line 435 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$4); //@line 435 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 STACKTOP = sp;return 2; //@line 436 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl24w_ParticleSystem_setSpinEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 442 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 442 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = $2; //@line 443 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = (+_luaL_optnumber($L,3,$3)); //@line 443 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $5 = $4; //@line 443 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem7setSpinEff($0,$2,$5); //@line 444 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 445 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl24w_ParticleSystem_getSpinEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $end = 0, $start = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $start = sp + 4|0;
 $end = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZNK4love8graphics14ParticleSystem7getSpinERfS2_($0,$start,$end); //@line 452 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = +HEAPF32[$start>>2]; //@line 453 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 453 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$2); //@line 453 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = +HEAPF32[$end>>2]; //@line 454 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = $3; //@line 454 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$4); //@line 454 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 STACKTOP = sp;return 2; //@line 455 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl33w_ParticleSystem_setSpinVariationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 461 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 461 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem16setSpinVariationEf($0,$2); //@line 462 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 463 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl33w_ParticleSystem_getSpinVariationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+__ZNK4love8graphics14ParticleSystem16getSpinVariationEv($0)); //@line 469 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 469 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$2); //@line 469 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 1; //@line 470 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl26w_ParticleSystem_setOffsetEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 476 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 476 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = (+_luaL_checknumber($L,3)); //@line 477 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = $3; //@line 477 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem9setOffsetEff($0,$2,$4); //@line 478 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 479 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl26w_ParticleSystem_getOffsetEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0, $4 = 0.0, $5 = 0.0, $offset = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $offset = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZNK4love8graphics14ParticleSystem9getOffsetEv($offset,$0); //@line 485 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = +HEAPF32[$offset>>2]; //@line 305 "./common/Vector.h"
 $2 = $1; //@line 486 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$2); //@line 486 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = ((($offset)) + 4|0); //@line 310 "./common/Vector.h"
 $4 = +HEAPF32[$3>>2]; //@line 310 "./common/Vector.h"
 $5 = $4; //@line 487 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$5); //@line 487 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 STACKTOP = sp;return 2; //@line 488 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl26w_ParticleSystem_setColorsEP9lua_State($L) {
 $L = $L|0;
 var $$0$i$i = 0, $$0$i$i26 = 0, $$034 = 0, $$08 = 0, $$09 = 0, $$15 = 0, $$3 = 0, $$lcssa70 = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0.0, $148 = 0, $149 = 0, $15 = 0, $150 = 0.0, $151 = 0, $152 = 0, $153 = 0, $154 = 0.0, $155 = 0, $156 = 0, $157 = 0.0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0.0;
 var $162 = 0, $163 = 0, $164 = 0.0, $165 = 0, $166 = 0, $167 = 0, $168 = 0.0, $169 = 0, $17 = 0, $170 = 0, $171 = 0.0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0, $72 = 0.0, $73 = 0, $74 = 0;
 var $75 = 0.0, $76 = 0, $77 = 0, $78 = 0.0, $79 = 0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0, $83 = 0, $84 = 0.0, $85 = 0, $86 = 0, $87 = 0.0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $colors = 0, $colors3 = 0, $cond2 = 0, $i$047 = 0, $i4$048 = 0, $j$0 = 0, $lpad$phi$index = 0, $lpad$phi$index7 = 0, $lpad$phi44$index = 0, $lpad$phi44$index13 = 0, $or$cond = 0, $scevgep = 0, $scevgep$i$i$i = 0;
 var $scevgep$i$i$i$i = 0, $scevgep$i$i$i$i27 = 0, $scevgep$i$i$i10 = 0, $scevgep$i$i$i15 = 0, $scevgep$i$i$i20 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i$i = 0, $scevgep4$i$i$i$i28 = 0, $scevgep4$i$i$i11 = 0, $scevgep4$i$i$i16 = 0, $scevgep4$i$i$i21 = 0, $scevgep58 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $colors = sp + 32|0;
 $colors3 = sp + 20|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (_lua_type($L,2)|0); //@line 495 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = ($1|0)==(5); //@line 495 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $3 = (_lua_gettop($L)|0); //@line 497 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = (($3) + -1)|0; //@line 497 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 L1: do {
  if ($2) {
   $5 = ($4|0)>(8); //@line 499 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
   if ($5) {
    $6 = (_luaL_error($L,343786,$vararg_buffer)|0); //@line 500 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $$3 = $6;
    STACKTOP = sp;return ($$3|0); //@line 551 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
   }
   HEAP32[$colors>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $7 = ((($colors)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$7>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $8 = ((($colors)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$8>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $9 = ($4|0)==(0); //@line 1080 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   do {
    if (!($9)) {
     $10 = ($4>>>0)>(268435455); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($10) {
      __THREW__ = 0;
      invoke_vi(3528,($colors|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $11 = __THREW__; __THREW__ = 0;
      $12 = $11&1;
      if (!($12)) {
       label = 7;
      }
     } else {
      label = 7;
     }
     if ((label|0) == 7) {
      $13 = $4 << 4; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      __THREW__ = 0;
      $14 = (invoke_ii(3348,($13|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $15 = __THREW__; __THREW__ = 0;
      $16 = $15&1;
      if (!($16)) {
       HEAP32[$7>>2] = $14; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$colors>>2] = $14; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $17 = (($14) + ($4<<4)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $18 = ((($colors)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       HEAP32[$18>>2] = $17; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $19 = $3 << 4; //@line 985 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $20 = (($19) + -16)|0; //@line 985 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $scevgep = (($14) + ($20)|0); //@line 985 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $$0$i$i = $4;$21 = $14;
       while(1) {
        ;HEAP32[$21>>2]=0|0;HEAP32[$21+4>>2]=0|0;HEAP32[$21+8>>2]=0|0;HEAP32[$21+12>>2]=0|0; //@line 37 "./modules/graphics/Color.h"
        $22 = ((($21)) + 16|0); //@line 989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $23 = (($$0$i$i) + -1)|0; //@line 990 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $24 = ($23|0)==(0); //@line 992 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($24) {
         break;
        } else {
         $$0$i$i = $23;$21 = $22;
        }
       }
       HEAP32[$7>>2] = $scevgep; //@line 989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       break;
      }
     }
     $26 = ___cxa_find_matching_catch_2()|0;
     $27 = tempRet0;
     $28 = HEAP32[$colors>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $29 = ($28|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $30 = $28; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($29) {
      ___resumeException($26|0);
      // unreachable;
     }
     $31 = HEAP32[$7>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $32 = ($31|0)==($28|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($32)) {
      $scevgep$i$i$i$i = ((($31)) + -16|0);
      $33 = $scevgep$i$i$i$i;
      $34 = (($33) - ($30))|0;
      $35 = $34 >>> 4;
      $36 = $35 ^ -1;
      $scevgep4$i$i$i$i = (($31) + ($36<<4)|0);
      HEAP32[$7>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __ZdlPv($28); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     ___resumeException($26|0);
     // unreachable;
    }
   } while(0);
   $25 = ($3|0)>(1); //@line 504 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
   L26: do {
    if ($25) {
     $i$047 = 0;
     L27: while(1) {
      $37 = (($i$047) + 2)|0; //@line 506 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      __THREW__ = 0;
      invoke_viii(3534,($L|0),($37|0),5); //@line 506 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $38 = __THREW__; __THREW__ = 0;
      $39 = $38&1;
      if ($39) {
       label = 22;
       break;
      }
      __THREW__ = 0;
      $40 = (invoke_iii(3381,($L|0),($37|0))|0); //@line 508 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $41 = __THREW__; __THREW__ = 0;
      $42 = $41&1;
      if ($42) {
       label = 22;
       break;
      }
      $43 = ($40>>>0)<(3); //@line 508 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      if ($43) {
       $$lcssa70 = $37;
       label = 20;
       break;
      } else {
       $j$0 = 0;
      }
      while(1) {
       $62 = ($j$0|0)<(4); //@line 511 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
       if (!($62)) {
        break;
       }
       $66 = (($j$0) + 1)|0; //@line 513 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
       __THREW__ = 0;
       invoke_viii(3382,($L|0),($37|0),($66|0)); //@line 513 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
       $67 = __THREW__; __THREW__ = 0;
       $68 = $67&1;
       if ($68) {
        label = 21;
        break L27;
       } else {
        $j$0 = $66;
       }
      }
      __THREW__ = 0;
      $63 = (+invoke_dii(3361,($L|0),-4)); //@line 515 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $64 = __THREW__; __THREW__ = 0;
      $65 = $64&1;
      if ($65) {
       label = 22;
       break;
      }
      $69 = $63; //@line 515 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $70 = HEAP32[$colors>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $71 = (($70) + ($i$047<<4)|0); //@line 515 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      HEAPF32[$71>>2] = $69; //@line 515 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      __THREW__ = 0;
      $72 = (+invoke_dii(3361,($L|0),-3)); //@line 516 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $73 = __THREW__; __THREW__ = 0;
      $74 = $73&1;
      if ($74) {
       label = 22;
       break;
      }
      $75 = $72; //@line 516 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $76 = HEAP32[$colors>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $77 = (((($76) + ($i$047<<4)|0)) + 4|0); //@line 516 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      HEAPF32[$77>>2] = $75; //@line 516 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      __THREW__ = 0;
      $78 = (+invoke_dii(3361,($L|0),-2)); //@line 517 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $79 = __THREW__; __THREW__ = 0;
      $80 = $79&1;
      if ($80) {
       label = 22;
       break;
      }
      $81 = $78; //@line 517 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $82 = HEAP32[$colors>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $83 = (((($82) + ($i$047<<4)|0)) + 8|0); //@line 517 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      HEAPF32[$83>>2] = $81; //@line 517 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      __THREW__ = 0;
      $84 = (+invoke_diid(3525,($L|0),-1,255.0)); //@line 518 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $85 = __THREW__; __THREW__ = 0;
      $86 = $85&1;
      if ($86) {
       label = 22;
       break;
      }
      $87 = $84; //@line 518 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $88 = HEAP32[$colors>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $89 = (((($88) + ($i$047<<4)|0)) + 12|0); //@line 518 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      HEAPF32[$89>>2] = $87; //@line 518 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      __THREW__ = 0;
      invoke_vii(3384,($L|0),-5); //@line 521 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $90 = __THREW__; __THREW__ = 0;
      $91 = $90&1;
      if ($91) {
       label = 22;
       break;
      }
      $92 = (($i$047) + 1)|0; //@line 504 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $93 = ($92|0)<($4|0); //@line 504 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      if ($93) {
       $i$047 = $92;
      } else {
       label = 36;
       break L26;
      }
     }
     if ((label|0) == 20) {
      __THREW__ = 0;
      $44 = (invoke_iiii(3549,($L|0),($$lcssa70|0),(343824|0))|0); //@line 509 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $45 = __THREW__; __THREW__ = 0;
      $46 = $45&1;
      if ($46) {
       label = 23;
       break;
      } else {
       $$034 = $44;$$15 = 1;
       label = 37;
       break;
      }
     }
     else if ((label|0) == 21) {
      $47 = ___cxa_find_matching_catch_2()|0;
      $48 = tempRet0;
      $lpad$phi$index = $47;$lpad$phi$index7 = $48;
      break;
     }
     else if ((label|0) == 22) {
      $49 = ___cxa_find_matching_catch_2()|0;
      $50 = tempRet0;
      $lpad$phi$index = $49;$lpad$phi$index7 = $50;
      break;
     }
    } else {
     label = 36;
    }
   } while(0);
   if ((label|0) == 36) {
    __THREW__ = 0;
    invoke_vii(3603,($0|0),($colors|0)); //@line 524 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $94 = __THREW__; __THREW__ = 0;
    $95 = $94&1;
    if ($95) {
     label = 23;
    } else {
     $$034 = 0;$$15 = 0;
     label = 37;
    }
   }
   if ((label|0) == 23) {
    $51 = ___cxa_find_matching_catch_2()|0;
    $52 = tempRet0;
    $lpad$phi$index = $51;$lpad$phi$index7 = $52;
   }
   else if ((label|0) == 37) {
    $96 = HEAP32[$colors>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $97 = ($96|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $98 = $96; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($97)) {
     $99 = HEAP32[$7>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $100 = ($99|0)==($96|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($100)) {
      $scevgep$i$i$i20 = ((($99)) + -16|0);
      $101 = $scevgep$i$i$i20;
      $102 = (($101) - ($98))|0;
      $103 = $102 >>> 4;
      $104 = $103 ^ -1;
      $scevgep4$i$i$i21 = (($99) + ($104<<4)|0);
      HEAP32[$7>>2] = $scevgep4$i$i$i21; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __ZdlPv($96); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    }
    $cond2 = ($$15|0)==(0);
    if ($cond2) {
     break;
    } else {
     $$3 = $$034;
    }
    STACKTOP = sp;return ($$3|0); //@line 551 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
   }
   $53 = HEAP32[$colors>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $54 = ($53|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $55 = $53; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($54) {
    $$08 = $lpad$phi$index7;$$09 = $lpad$phi$index;
    ___resumeException($$09|0);
    // unreachable;
   }
   $56 = HEAP32[$7>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $57 = ($56|0)==($53|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($57)) {
    $scevgep$i$i$i = ((($56)) + -16|0);
    $58 = $scevgep$i$i$i;
    $59 = (($58) - ($55))|0;
    $60 = $59 >>> 4;
    $61 = $60 ^ -1;
    $scevgep4$i$i$i = (($56) + ($61<<4)|0);
    HEAP32[$7>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($53); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $$08 = $lpad$phi$index7;$$09 = $lpad$phi$index;
   ___resumeException($$09|0);
   // unreachable;
  } else {
   $105 = (($3) + 2)|0; //@line 529 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
   $106 = (($105|0) / 4)&-1; //@line 529 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
   $107 = ($4|0)==(3); //@line 531 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
   if (!($107)) {
    $108 = (($4|0) % 4)&-1; //@line 531 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $109 = ($108|0)!=(0); //@line 531 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $110 = ($4|0)==(0); //@line 531 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $or$cond = $110 | $109; //@line 531 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    if ($or$cond) {
     HEAP32[$vararg_buffer1>>2] = $108; //@line 532 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
     $111 = (_luaL_error($L,343852,$vararg_buffer1)|0); //@line 532 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
     $$3 = $111;
     STACKTOP = sp;return ($$3|0); //@line 551 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    }
   }
   $112 = ($105|0)>(35); //@line 534 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
   if ($112) {
    $113 = (_luaL_error($L,343786,$vararg_buffer3)|0); //@line 535 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $$3 = $113;
    STACKTOP = sp;return ($$3|0); //@line 551 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
   }
   HEAP32[$colors3>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $114 = ((($colors3)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$114>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $115 = ((($colors3)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$115>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $$off = (($3) + 5)|0; //@line 1080 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $116 = ($$off>>>0)<(7); //@line 1080 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   do {
    if (!($116)) {
     $117 = ($106>>>0)>(268435455); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($117) {
      __THREW__ = 0;
      invoke_vi(3528,($colors3|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $118 = __THREW__; __THREW__ = 0;
      $119 = $118&1;
      if (!($119)) {
       label = 50;
      }
     } else {
      label = 50;
     }
     if ((label|0) == 50) {
      $120 = $106 << 4; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      __THREW__ = 0;
      $121 = (invoke_ii(3348,($120|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $122 = __THREW__; __THREW__ = 0;
      $123 = $122&1;
      if (!($123)) {
       HEAP32[$114>>2] = $121; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$colors3>>2] = $121; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $124 = (($121) + ($106<<4)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $125 = ((($colors3)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       HEAP32[$125>>2] = $124; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $scevgep58 = (($121) + ($120)|0); //@line 985 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $$0$i$i26 = $106;$126 = $121;
       while(1) {
        ;HEAP32[$126>>2]=0|0;HEAP32[$126+4>>2]=0|0;HEAP32[$126+8>>2]=0|0;HEAP32[$126+12>>2]=0|0; //@line 37 "./modules/graphics/Color.h"
        $127 = ((($126)) + 16|0); //@line 989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $128 = (($$0$i$i26) + -1)|0; //@line 990 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $129 = ($128|0)==(0); //@line 992 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($129) {
         break;
        } else {
         $$0$i$i26 = $128;$126 = $127;
        }
       }
       HEAP32[$114>>2] = $scevgep58; //@line 989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       break;
      }
     }
     $131 = ___cxa_find_matching_catch_2()|0;
     $132 = tempRet0;
     $133 = HEAP32[$colors3>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $134 = ($133|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $135 = $133; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($134) {
      ___resumeException($131|0);
      // unreachable;
     }
     $136 = HEAP32[$114>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $137 = ($136|0)==($133|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($137)) {
      $scevgep$i$i$i$i27 = ((($136)) + -16|0);
      $138 = $scevgep$i$i$i$i27;
      $139 = (($138) - ($135))|0;
      $140 = $139 >>> 4;
      $141 = $140 ^ -1;
      $scevgep4$i$i$i$i28 = (($136) + ($141<<4)|0);
      HEAP32[$114>>2] = $scevgep4$i$i$i$i28; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __ZdlPv($133); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     ___resumeException($131|0);
     // unreachable;
    }
   } while(0);
   $130 = ($105|0)>(3); //@line 539 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
   L93: do {
    if ($130) {
     $i4$048 = 0;
     while(1) {
      $144 = $i4$048 << 2; //@line 541 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $145 = $144 | 1; //@line 541 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $146 = (($145) + 1)|0; //@line 541 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      __THREW__ = 0;
      $147 = (+invoke_dii(3361,($L|0),($146|0))); //@line 541 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $148 = __THREW__; __THREW__ = 0;
      $149 = $148&1;
      if ($149) {
       break;
      }
      $150 = $147; //@line 541 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $151 = HEAP32[$colors3>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $152 = (($151) + ($i4$048<<4)|0); //@line 541 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      HEAPF32[$152>>2] = $150; //@line 541 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $153 = $144 | 3; //@line 542 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      __THREW__ = 0;
      $154 = (+invoke_dii(3361,($L|0),($153|0))); //@line 542 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $155 = __THREW__; __THREW__ = 0;
      $156 = $155&1;
      if ($156) {
       break;
      }
      $157 = $154; //@line 542 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $158 = HEAP32[$colors3>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $159 = (((($158) + ($i4$048<<4)|0)) + 4|0); //@line 542 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      HEAPF32[$159>>2] = $157; //@line 542 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $160 = (($145) + 3)|0; //@line 543 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      __THREW__ = 0;
      $161 = (+invoke_dii(3361,($L|0),($160|0))); //@line 543 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $162 = __THREW__; __THREW__ = 0;
      $163 = $162&1;
      if ($163) {
       break;
      }
      $164 = $161; //@line 543 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $165 = HEAP32[$colors3>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $166 = (((($165) + ($i4$048<<4)|0)) + 8|0); //@line 543 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      HEAPF32[$166>>2] = $164; //@line 543 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $167 = (($145) + 4)|0; //@line 544 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      __THREW__ = 0;
      $168 = (+invoke_dii(3361,($L|0),($167|0))); //@line 544 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $169 = __THREW__; __THREW__ = 0;
      $170 = $169&1;
      if ($170) {
       break;
      }
      $171 = $168; //@line 544 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $172 = HEAP32[$colors3>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $173 = (((($172) + ($i4$048<<4)|0)) + 12|0); //@line 544 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      HEAPF32[$173>>2] = $171; //@line 544 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $174 = (($i4$048) + 1)|0; //@line 539 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $175 = ($174|0)<($106|0); //@line 539 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      if ($175) {
       $i4$048 = $174;
      } else {
       label = 60;
       break L93;
      }
     }
     $176 = ___cxa_find_matching_catch_2()|0;
     $177 = tempRet0;
     $lpad$phi44$index = $176;$lpad$phi44$index13 = $177;
    } else {
     label = 60;
    }
   } while(0);
   do {
    if ((label|0) == 60) {
     __THREW__ = 0;
     invoke_vii(3603,($0|0),($colors3|0)); //@line 547 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
     $142 = __THREW__; __THREW__ = 0;
     $143 = $142&1;
     if ($143) {
      $178 = ___cxa_find_matching_catch_2()|0;
      $179 = tempRet0;
      $lpad$phi44$index = $178;$lpad$phi44$index13 = $179;
      break;
     }
     $189 = HEAP32[$colors3>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $190 = ($189|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $191 = $189; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($190)) {
      $192 = HEAP32[$114>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $193 = ($192|0)==($189|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($193)) {
       $scevgep$i$i$i10 = ((($192)) + -16|0);
       $194 = $scevgep$i$i$i10;
       $195 = (($194) - ($191))|0;
       $196 = $195 >>> 4;
       $197 = $196 ^ -1;
       $scevgep4$i$i$i11 = (($192) + ($197<<4)|0);
       HEAP32[$114>>2] = $scevgep4$i$i$i11; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($189); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     break L1;
    }
   } while(0);
   $180 = HEAP32[$colors3>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $181 = ($180|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $182 = $180; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($181) {
    $$08 = $lpad$phi44$index13;$$09 = $lpad$phi44$index;
    ___resumeException($$09|0);
    // unreachable;
   }
   $183 = HEAP32[$114>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $184 = ($183|0)==($180|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($184)) {
    $scevgep$i$i$i15 = ((($183)) + -16|0);
    $185 = $scevgep$i$i$i15;
    $186 = (($185) - ($182))|0;
    $187 = $186 >>> 4;
    $188 = $187 ^ -1;
    $scevgep4$i$i$i16 = (($183) + ($188<<4)|0);
    HEAP32[$114>>2] = $scevgep4$i$i$i16; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($180); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $$08 = $lpad$phi44$index13;$$09 = $lpad$phi44$index;
   ___resumeException($$09|0);
   // unreachable;
  }
 } while(0);
 $$3 = 0;
 STACKTOP = sp;return ($$3|0); //@line 551 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl26w_ParticleSystem_getColorsEP9lua_State($L) {
 $L = $L|0;
 var $$in = 0, $$in9 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $7 = 0, $8 = 0, $9 = 0, $i$07 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i1 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZNK4love8graphics14ParticleSystem8getColorEv($0,$1); //@line 557 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = ((($0)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($3|0)==($4|0); //@line 559 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 L1: do {
  if ($5) {
   $$in = $3;$$in9 = $4;$$lcssa = 0;
  } else {
   $i$07 = 0;
   while(1) {
    __THREW__ = 0;
    invoke_viii(3393,($L|0),4,0); //@line 561 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $15 = __THREW__; __THREW__ = 0;
    $16 = $15&1;
    if ($16) {
     break;
    }
    $17 = HEAP32[$0>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $18 = (($17) + ($i$07<<4)|0); //@line 563 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $19 = +HEAPF32[$18>>2]; //@line 563 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $20 = $19; //@line 563 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    __THREW__ = 0;
    invoke_vid(3394,($L|0),(+$20)); //@line 563 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $21 = __THREW__; __THREW__ = 0;
    $22 = $21&1;
    if ($22) {
     break;
    }
    __THREW__ = 0;
    invoke_viii(3395,($L|0),-2,1); //@line 564 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $23 = __THREW__; __THREW__ = 0;
    $24 = $23&1;
    if ($24) {
     break;
    }
    $25 = HEAP32[$0>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $26 = (((($25) + ($i$07<<4)|0)) + 4|0); //@line 565 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $27 = +HEAPF32[$26>>2]; //@line 565 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $28 = $27; //@line 565 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    __THREW__ = 0;
    invoke_vid(3394,($L|0),(+$28)); //@line 565 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $29 = __THREW__; __THREW__ = 0;
    $30 = $29&1;
    if ($30) {
     break;
    }
    __THREW__ = 0;
    invoke_viii(3395,($L|0),-2,2); //@line 566 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $31 = __THREW__; __THREW__ = 0;
    $32 = $31&1;
    if ($32) {
     break;
    }
    $33 = HEAP32[$0>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $34 = (((($33) + ($i$07<<4)|0)) + 8|0); //@line 567 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $35 = +HEAPF32[$34>>2]; //@line 567 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $36 = $35; //@line 567 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    __THREW__ = 0;
    invoke_vid(3394,($L|0),(+$36)); //@line 567 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $37 = __THREW__; __THREW__ = 0;
    $38 = $37&1;
    if ($38) {
     break;
    }
    __THREW__ = 0;
    invoke_viii(3395,($L|0),-2,3); //@line 568 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $39 = __THREW__; __THREW__ = 0;
    $40 = $39&1;
    if ($40) {
     break;
    }
    $41 = HEAP32[$0>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $42 = (((($41) + ($i$07<<4)|0)) + 12|0); //@line 569 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $43 = +HEAPF32[$42>>2]; //@line 569 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $44 = $43; //@line 569 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    __THREW__ = 0;
    invoke_vid(3394,($L|0),(+$44)); //@line 569 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $45 = __THREW__; __THREW__ = 0;
    $46 = $45&1;
    if ($46) {
     break;
    }
    __THREW__ = 0;
    invoke_viii(3395,($L|0),-2,4); //@line 570 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $47 = __THREW__; __THREW__ = 0;
    $48 = $47&1;
    if ($48) {
     break;
    }
    $49 = (($i$07) + 1)|0; //@line 559 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $50 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $51 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $52 = (($50) - ($51))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $53 = $52 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $54 = ($49>>>0)<($53>>>0); //@line 559 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    if ($54) {
     $i$07 = $49;
    } else {
     $$in = $50;$$in9 = $51;$$lcssa = $53;
     break L1;
    }
   }
   $55 = ___cxa_find_matching_catch_2()|0;
   $56 = tempRet0;
   $57 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $58 = ($57|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $59 = $57; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($58) {
    ___resumeException($55|0);
    // unreachable;
   }
   $60 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $61 = ($60|0)==($57|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($61)) {
    $scevgep$i$i$i = ((($60)) + -16|0);
    $62 = $scevgep$i$i$i;
    $63 = (($62) - ($59))|0;
    $64 = $63 >>> 4;
    $65 = $64 ^ -1;
    $scevgep4$i$i$i = (($60) + ($65<<4)|0);
    HEAP32[$2>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($57); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($55|0);
   // unreachable;
  }
 } while(0);
 $6 = $$in; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ($$in9|0)==(0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = $$in9; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($7) {
  STACKTOP = sp;return ($$lcssa|0); //@line 574 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 }
 $9 = $$in9; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = ($6|0)==($9|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($10)) {
  $scevgep$i$i$i1 = ((($6)) + -16|0);
  $11 = $scevgep$i$i$i1;
  $12 = (($11) - ($$in9))|0;
  $13 = $12 >>> 4;
  $14 = $13 ^ -1;
  $scevgep4$i$i$i2 = (($6) + ($14<<4)|0);
  HEAP32[$2>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($8); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 STACKTOP = sp;return ($$lcssa|0); //@line 574 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl25w_ParticleSystem_setQuadsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $8 = 0, $9 = 0, $i$0 = 0, $i1$0 = 0, $lpad$phi$index = 0, $lpad$phi$index3 = 0, $q = 0, $q2 = 0, $quads = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i1 = 0;
 var $scevgep4$i$i$i = 0, $scevgep4$i$i$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $quads = sp + 8|0;
 $q = sp + 4|0;
 $q2 = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 HEAP32[$quads>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($quads)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($quads)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 $3 = (invoke_iii(3358,($L|0),2)|0); //@line 581 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 L1: do {
  if ($5) {
   label = 9;
  } else {
   $6 = ($3|0)==(5); //@line 581 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
   L3: do {
    if ($6) {
     $8 = ((($quads)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $i$0 = 1;
     while(1) {
      __THREW__ = 0;
      $9 = (invoke_iii(3381,($L|0),2)|0); //@line 583 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $10 = __THREW__; __THREW__ = 0;
      $11 = $10&1;
      if ($11) {
       break;
      }
      $12 = ($i$0|0)>($9|0); //@line 583 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      if ($12) {
       break L3;
      }
      __THREW__ = 0;
      invoke_viii(3382,($L|0),2,($i$0|0)); //@line 585 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $28 = __THREW__; __THREW__ = 0;
      $29 = $28&1;
      if ($29) {
       break;
      }
      __THREW__ = 0;
      $30 = (invoke_iiii(3604,($L|0),-1,13)|0); //@line 587 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $31 = __THREW__; __THREW__ = 0;
      $32 = $31&1;
      if ($32) {
       break;
      }
      HEAP32[$q>>2] = $30; //@line 587 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $33 = HEAP32[$1>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $34 = HEAP32[$8>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $35 = ($33|0)==($34|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($35) {
       __THREW__ = 0;
       invoke_vii(3605,($quads|0),($q|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $39 = __THREW__; __THREW__ = 0;
       $40 = $39&1;
       if ($40) {
        break;
       }
      } else {
       $36 = $30; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       HEAP32[$33>>2] = $36; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $37 = HEAP32[$1>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $38 = ((($37)) + 4|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$1>>2] = $38; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __THREW__ = 0;
      invoke_vii(3384,($L|0),-2); //@line 590 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $41 = __THREW__; __THREW__ = 0;
      $42 = $41&1;
      if ($42) {
       break;
      }
      $43 = (($i$0) + 1)|0; //@line 583 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $i$0 = $43;
     }
     $13 = ___cxa_find_matching_catch_2()|0;
     $14 = tempRet0;
     $lpad$phi$index = $13;$lpad$phi$index3 = $14;
     break L1;
    } else {
     $7 = ((($quads)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $i1$0 = 2;
     while(1) {
      __THREW__ = 0;
      $44 = (invoke_ii(3385,($L|0))|0); //@line 595 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $45 = __THREW__; __THREW__ = 0;
      $46 = $45&1;
      if ($46) {
       break;
      }
      $47 = ($i1$0|0)>($44|0); //@line 595 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      if ($47) {
       break L3;
      }
      __THREW__ = 0;
      $48 = (invoke_iiii(3604,($L|0),($i1$0|0),13)|0); //@line 597 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $49 = __THREW__; __THREW__ = 0;
      $50 = $49&1;
      if ($50) {
       break;
      }
      HEAP32[$q2>>2] = $48; //@line 597 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $51 = HEAP32[$1>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $52 = HEAP32[$7>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $53 = ($51|0)==($52|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($53) {
       __THREW__ = 0;
       invoke_vii(3605,($quads|0),($q2|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $57 = __THREW__; __THREW__ = 0;
       $58 = $57&1;
       if ($58) {
        break;
       }
      } else {
       $54 = $48; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       HEAP32[$51>>2] = $54; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $55 = HEAP32[$1>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $56 = ((($55)) + 4|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$1>>2] = $56; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      $59 = (($i1$0) + 1)|0; //@line 595 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
      $i1$0 = $59;
     }
     $15 = ___cxa_find_matching_catch_2()|0;
     $16 = tempRet0;
     $lpad$phi$index = $15;$lpad$phi$index3 = $16;
     break L1;
    }
   } while(0);
   __THREW__ = 0;
   invoke_vii(3606,($0|0),($quads|0)); //@line 602 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
   $60 = __THREW__; __THREW__ = 0;
   $61 = $60&1;
   if ($61) {
    label = 9;
   } else {
    $62 = HEAP32[$quads>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $63 = ($62|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $64 = $62; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($63) {
     STACKTOP = sp;return 0; //@line 604 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    }
    $65 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $66 = ($65|0)==($62|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($66)) {
     $scevgep$i$i$i = ((($65)) + -4|0);
     $67 = $scevgep$i$i$i;
     $68 = (($67) - ($64))|0;
     $69 = $68 >>> 2;
     $70 = $69 ^ -1;
     $scevgep4$i$i$i = (($65) + ($70<<2)|0);
     HEAP32[$1>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($62); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    STACKTOP = sp;return 0; //@line 604 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
   }
  }
 } while(0);
 if ((label|0) == 9) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  $lpad$phi$index = $17;$lpad$phi$index3 = $18;
 }
 $19 = HEAP32[$quads>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = ($19|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = $19; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($20) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $22 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $23 = ($22|0)==($19|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($23)) {
  $scevgep$i$i$i1 = ((($22)) + -4|0);
  $24 = $scevgep$i$i$i1;
  $25 = (($24) - ($21))|0;
  $26 = $25 >>> 2;
  $27 = $26 ^ -1;
  $scevgep4$i$i$i2 = (($22) + ($27<<2)|0);
  HEAP32[$1>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($19); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love14luax_checktypeINS_8graphics4QuadEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZNSt3__16vectorIPN4love8graphics4QuadENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl25w_ParticleSystem_getQuadsEP9lua_State($L) {
 $L = $L|0;
 var $$cast = 0, $$cast$lcssa = 0, $$lcssa = 0, $$lcssa23 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $quads = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i1 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $quads = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZNK4love8graphics14ParticleSystem8getQuadsEv($quads,$0); //@line 609 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = ((($quads)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$quads>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($2) - ($3))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = $4 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 __THREW__ = 0;
 invoke_viii(3393,($L|0),($5|0),0); //@line 611 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 do {
  if ($7) {
   $23 = ___cxa_find_matching_catch_2()|0;
   $24 = tempRet0;
   $lpad$phi$index = $23;$lpad$phi$index2 = $24;
  } else {
   $i$0 = 0;
   while(1) {
    $8 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $9 = HEAP32[$quads>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $10 = (($8) - ($9))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $11 = $10 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $12 = ($i$0|0)<($11|0); //@line 613 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $$cast = $9; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($12)) {
     $$cast$lcssa = $$cast;$$lcssa = $8;$$lcssa23 = $9;
     break;
    }
    $34 = (($$cast) + ($i$0<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $35 = HEAP32[$34>>2]|0; //@line 615 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    __THREW__ = 0;
    invoke_viii(3387,($L|0),13,($35|0)); //@line 615 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $36 = __THREW__; __THREW__ = 0;
    $37 = $36&1;
    if ($37) {
     label = 8;
     break;
    }
    $38 = (($i$0) + 1)|0; //@line 616 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    __THREW__ = 0;
    invoke_viii(3395,($L|0),-2,($38|0)); //@line 616 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
    $39 = __THREW__; __THREW__ = 0;
    $40 = $39&1;
    if ($40) {
     label = 8;
     break;
    } else {
     $i$0 = $38;
    }
   }
   if ((label|0) == 8) {
    $21 = ___cxa_find_matching_catch_2()|0;
    $22 = tempRet0;
    $lpad$phi$index = $21;$lpad$phi$index2 = $22;
    break;
   }
   $13 = $$lcssa; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $14 = ($$lcssa23|0)==(0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $15 = $$lcssa23; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($14) {
    STACKTOP = sp;return 1; //@line 620 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
   }
   $16 = ($13|0)==($$cast$lcssa|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($16)) {
    $scevgep$i$i$i1 = ((($13)) + -4|0);
    $17 = $scevgep$i$i$i1;
    $18 = (($17) - ($$lcssa23))|0;
    $19 = $18 >>> 2;
    $20 = $19 ^ -1;
    $scevgep4$i$i$i2 = (($13) + ($20<<2)|0);
    HEAP32[$1>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($15); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   STACKTOP = sp;return 1; //@line 620 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
  }
 } while(0);
 $25 = HEAP32[$quads>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $26 = ($25|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = $25; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($26) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $28 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $29 = ($28|0)==($25|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($29)) {
  $scevgep$i$i$i = ((($28)) + -4|0);
  $30 = $scevgep$i$i$i;
  $31 = (($30) - ($27))|0;
  $32 = $31 >>> 2;
  $33 = $32 ^ -1;
  $scevgep4$i$i$i = (($28) + ($33<<2)|0);
  HEAP32[$1>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($25); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8graphics6opengl36w_ParticleSystem_setRelativeRotationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,2)|0); //@line 625 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem19setRelativeRotationEb($0,$1); //@line 625 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 626 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl36w_ParticleSystem_hasRelativeRotationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (__ZNK4love8graphics14ParticleSystem19hasRelativeRotationEv($0)|0); //@line 632 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$1); //@line 632 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 1; //@line 633 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl25w_ParticleSystem_getCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (__ZNK4love8graphics14ParticleSystem8getCountEv($0)|0); //@line 639 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = (+($1>>>0)); //@line 639 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 _lua_pushnumber($L,$2); //@line 639 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 1; //@line 640 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl22w_ParticleSystem_startEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem5startEv($0); //@line 646 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 647 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl21w_ParticleSystem_stopEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem4stopEv($0); //@line 653 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 654 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl22w_ParticleSystem_pauseEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem5pauseEv($0); //@line 660 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 661 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl22w_ParticleSystem_resetEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem5resetEv($0); //@line 667 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 668 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl21w_ParticleSystem_emitEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 674 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = (~~(($1))); //@line 674 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem4emitEj($0,$2); //@line 675 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 676 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl25w_ParticleSystem_isActiveEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (__ZNK4love8graphics14ParticleSystem8isActiveEv($0)|0); //@line 682 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$1); //@line 682 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 1; //@line 683 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl25w_ParticleSystem_isPausedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (__ZNK4love8graphics14ParticleSystem8isPausedEv($0)|0); //@line 689 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$1); //@line 689 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 1; //@line 690 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl26w_ParticleSystem_isStoppedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (__ZNK4love8graphics14ParticleSystem9isStoppedEv($0)|0); //@line 696 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$1); //@line 696 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 1; //@line 697 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl23w_ParticleSystem_updateEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl14ParticleSystemEEEPT_P9lua_StateiNS_4TypeE($L,1,15)|0); //@line 44 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 703 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $2 = $1; //@line 703 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem6updateEf($0,$2); //@line 704 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 return 0; //@line 705 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function _luaopen_particlesystem($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 773 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 HEAP32[$vararg_buffer>>2] = 14328; //@line 775 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 775 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 775 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,15,343919,$vararg_buffer)|0); //@line 775 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
 STACKTOP = sp;return ($0|0); //@line 775 "modules/graphics/opengl/wrap_ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl16luax_checkshaderEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl6ShaderEEEPT_P9lua_StateiNS_4TypeE($L,$idx,18)|0); //@line 39 "modules/graphics/opengl/wrap_Shader.cpp"
 return ($0|0); //@line 39 "modules/graphics/opengl/wrap_Shader.cpp"
}
function __ZN4love14luax_checktypeINS_8graphics6opengl6ShaderEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love8graphics6opengl20w_Shader_getWarningsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__ZN4love14luax_checktypeINS_8graphics6opengl6ShaderEEEPT_P9lua_StateiNS_4TypeE($L,1,18)|0); //@line 39 "modules/graphics/opengl/wrap_Shader.cpp"
 __ZNK4love8graphics6opengl6Shader11getWarningsEv($0,$1); //@line 45 "modules/graphics/opengl/wrap_Shader.cpp"
 $2 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = $2 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ($3<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = HEAP32[$5>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $4 ? $7 : $6; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($8|0)); //@line 45 "modules/graphics/opengl/wrap_Shader.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $11 = ___cxa_find_matching_catch_2()|0;
  $12 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 45 "modules/graphics/opengl/wrap_Shader.cpp"
  ___resumeException($11|0);
  // unreachable;
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 45 "modules/graphics/opengl/wrap_Shader.cpp"
  STACKTOP = sp;return 1; //@line 46 "modules/graphics/opengl/wrap_Shader.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl16w_Shader_sendIntEP9lua_State($L) {
 $L = $L|0;
 var $$01 = 0, $$03 = 0, $$12 = 0, $$14 = 0, $$2 = 0, $$arith = 0, $$arith11 = 0, $$lcssa = 0, $$lcssa42 = 0, $$lcssa43 = 0, $$overflow = 0, $$overflow12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0.0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0.0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $9 = 0, $dimension$0$ph7 = 0, $i$02$i = 0, $i$08$i = 0, $k$07$i = 0;
 var $values$0$ph = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, $vararg_ptr3 = 0, $vararg_ptr4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 36|0;
 $1 = (__ZN4love14luax_checktypeINS_8graphics6opengl6ShaderEEEPT_P9lua_StateiNS_4TypeE($L,1,18)|0); //@line 39 "modules/graphics/opengl/wrap_Shader.cpp"
 $2 = (_luaL_checklstring($L,2,0)|0); //@line 117 "modules/graphics/opengl/wrap_Shader.cpp"
 $3 = (_lua_gettop($L)|0); //@line 118 "modules/graphics/opengl/wrap_Shader.cpp"
 $4 = (($3) + -2)|0; //@line 118 "modules/graphics/opengl/wrap_Shader.cpp"
 $5 = ($4|0)<(1); //@line 120 "modules/graphics/opengl/wrap_Shader.cpp"
 if ($5) {
  $6 = (_luaL_error($L,344578,$vararg_buffer)|0); //@line 121 "modules/graphics/opengl/wrap_Shader.cpp"
  $$2 = $6;
  STACKTOP = sp;return ($$2|0); //@line 153 "modules/graphics/opengl/wrap_Shader.cpp"
 }
 $7 = (_lua_isnumber($L,3)|0); //@line 126 "modules/graphics/opengl/wrap_Shader.cpp"
 $8 = ($7|0)==(0); //@line 126 "modules/graphics/opengl/wrap_Shader.cpp"
 L5: do {
  if ($8) {
   $9 = (_lua_type($L,3)|0); //@line 126 "modules/graphics/opengl/wrap_Shader.cpp"
   $10 = ($9|0)==(1); //@line 126 "modules/graphics/opengl/wrap_Shader.cpp"
   if ($10) {
    label = 5;
   } else {
    $25 = (_lua_type($L,3)|0); //@line 128 "modules/graphics/opengl/wrap_Shader.cpp"
    $26 = ($25|0)==(5); //@line 128 "modules/graphics/opengl/wrap_Shader.cpp"
    if (!($26)) {
     $56 = (_luax_typerror($L,3,344671)|0); //@line 131 "modules/graphics/opengl/wrap_Shader.cpp"
     $$2 = $56;
     STACKTOP = sp;return ($$2|0); //@line 153 "modules/graphics/opengl/wrap_Shader.cpp"
    }
    $27 = (__ZN4love11luax_objlenEP9lua_Statei($L,3)|0); //@line 75 "modules/graphics/opengl/wrap_Shader.cpp"
    $28 = Math_imul($27, $4)|0; //@line 76 "modules/graphics/opengl/wrap_Shader.cpp"
    $$arith = $28<<2; //@line 76 "modules/graphics/opengl/wrap_Shader.cpp"
    $$overflow = ($28>>>0)>(1073741823); //@line 76 "modules/graphics/opengl/wrap_Shader.cpp"
    $29 = $$overflow ? -1 : $$arith; //@line 76 "modules/graphics/opengl/wrap_Shader.cpp"
    $30 = (__Znaj($29)|0); //@line 76 "modules/graphics/opengl/wrap_Shader.cpp"
    $31 = ($27|0)<(1); //@line 94 "modules/graphics/opengl/wrap_Shader.cpp"
    $32 = $27 ^ -1; //@line 108 "modules/graphics/opengl/wrap_Shader.cpp"
    $i$08$i = 0;
    L12: while(1) {
     $33 = (($i$08$i) + 3)|0; //@line 80 "modules/graphics/opengl/wrap_Shader.cpp"
     $34 = (_lua_type($L,$33)|0); //@line 80 "modules/graphics/opengl/wrap_Shader.cpp"
     $35 = ($34|0)==(5); //@line 80 "modules/graphics/opengl/wrap_Shader.cpp"
     if (!($35)) {
      $$lcssa42 = $33;
      label = 15;
      break;
     }
     $36 = (__ZN4love11luax_objlenEP9lua_Statei($L,$33)|0); //@line 86 "modules/graphics/opengl/wrap_Shader.cpp"
     $37 = ($36|0)==($27|0); //@line 86 "modules/graphics/opengl/wrap_Shader.cpp"
     if (!($37)) {
      $$lcssa43 = $33;
      label = 19;
      break;
     }
     if (!($31)) {
      $38 = Math_imul($i$08$i, $27)|0; //@line 100 "modules/graphics/opengl/wrap_Shader.cpp"
      $39 = (($38) + -1)|0; //@line 100 "modules/graphics/opengl/wrap_Shader.cpp"
      $k$07$i = 1;
      while(1) {
       _lua_rawgeti($L,$33,$k$07$i); //@line 96 "modules/graphics/opengl/wrap_Shader.cpp"
       $41 = (_lua_isnumber($L,-1)|0); //@line 97 "modules/graphics/opengl/wrap_Shader.cpp"
       $42 = ($41|0)==(0); //@line 97 "modules/graphics/opengl/wrap_Shader.cpp"
       if ($42) {
        $47 = (_lua_type($L,-1)|0); //@line 99 "modules/graphics/opengl/wrap_Shader.cpp"
        $48 = ($47|0)==(1); //@line 99 "modules/graphics/opengl/wrap_Shader.cpp"
        if (!($48)) {
         label = 25;
         break L12;
        }
        $49 = (_lua_toboolean($L,-1)|0); //@line 100 "modules/graphics/opengl/wrap_Shader.cpp"
        $50 = (($39) + ($k$07$i))|0; //@line 100 "modules/graphics/opengl/wrap_Shader.cpp"
        $51 = (($30) + ($50<<2)|0); //@line 100 "modules/graphics/opengl/wrap_Shader.cpp"
        HEAP32[$51>>2] = $49; //@line 100 "modules/graphics/opengl/wrap_Shader.cpp"
       } else {
        $43 = (+_lua_tonumber($L,-1)); //@line 98 "modules/graphics/opengl/wrap_Shader.cpp"
        $44 = (~~(($43))); //@line 98 "modules/graphics/opengl/wrap_Shader.cpp"
        $45 = (($39) + ($k$07$i))|0; //@line 98 "modules/graphics/opengl/wrap_Shader.cpp"
        $46 = (($30) + ($45<<2)|0); //@line 98 "modules/graphics/opengl/wrap_Shader.cpp"
        HEAP32[$46>>2] = $44; //@line 98 "modules/graphics/opengl/wrap_Shader.cpp"
       }
       $52 = (($k$07$i) + 1)|0; //@line 94 "modules/graphics/opengl/wrap_Shader.cpp"
       $53 = ($k$07$i|0)<($27|0); //@line 94 "modules/graphics/opengl/wrap_Shader.cpp"
       if ($53) {
        $k$07$i = $52;
       } else {
        break;
       }
      }
     }
     _lua_settop($L,$32); //@line 108 "modules/graphics/opengl/wrap_Shader.cpp"
     $54 = (($i$08$i) + 1)|0; //@line 78 "modules/graphics/opengl/wrap_Shader.cpp"
     $55 = ($54|0)<($4|0); //@line 78 "modules/graphics/opengl/wrap_Shader.cpp"
     if ($55) {
      $i$08$i = $54;
     } else {
      $dimension$0$ph7 = $27;$values$0$ph = $30;
      break L5;
     }
    }
    if ((label|0) == 15) {
     __ZdaPv($30); //@line 82 "modules/graphics/opengl/wrap_Shader.cpp"
     (_luax_typerror($L,$$lcssa42,522280)|0); //@line 83 "modules/graphics/opengl/wrap_Shader.cpp"
     label = 28;
     break;
    }
    else if ((label|0) == 19) {
     __ZdaPv($30); //@line 88 "modules/graphics/opengl/wrap_Shader.cpp"
     $40 = (__ZN4love11luax_objlenEP9lua_Statei($L,$$lcssa43)|0); //@line 90 "modules/graphics/opengl/wrap_Shader.cpp"
     HEAP32[$vararg_buffer1>>2] = $$lcssa43; //@line 89 "modules/graphics/opengl/wrap_Shader.cpp"
     $vararg_ptr3 = ((($vararg_buffer1)) + 4|0); //@line 89 "modules/graphics/opengl/wrap_Shader.cpp"
     HEAP32[$vararg_ptr3>>2] = $27; //@line 89 "modules/graphics/opengl/wrap_Shader.cpp"
     $vararg_ptr4 = ((($vararg_buffer1)) + 8|0); //@line 89 "modules/graphics/opengl/wrap_Shader.cpp"
     HEAP32[$vararg_ptr4>>2] = $40; //@line 89 "modules/graphics/opengl/wrap_Shader.cpp"
     (_luaL_error($L,344617,$vararg_buffer1)|0); //@line 89 "modules/graphics/opengl/wrap_Shader.cpp"
     label = 28;
     break;
    }
    else if ((label|0) == 25) {
     __ZdaPv($30); //@line 103 "modules/graphics/opengl/wrap_Shader.cpp"
     (_luax_typerror($L,-1,344599)|0); //@line 104 "modules/graphics/opengl/wrap_Shader.cpp"
     label = 28;
     break;
    }
   }
  } else {
   label = 5;
  }
 } while(0);
 L30: do {
  if ((label|0) == 5) {
   $$arith11 = $4<<2; //@line 53 "modules/graphics/opengl/wrap_Shader.cpp"
   $$overflow12 = ($4>>>0)>(1073741823); //@line 53 "modules/graphics/opengl/wrap_Shader.cpp"
   $11 = $$overflow12 ? -1 : $$arith11; //@line 53 "modules/graphics/opengl/wrap_Shader.cpp"
   $12 = (__Znaj($11)|0); //@line 53 "modules/graphics/opengl/wrap_Shader.cpp"
   $i$02$i = 0;
   while(1) {
    $13 = (($i$02$i) + 3)|0; //@line 57 "modules/graphics/opengl/wrap_Shader.cpp"
    $14 = (_lua_isnumber($L,$13)|0); //@line 57 "modules/graphics/opengl/wrap_Shader.cpp"
    $15 = ($14|0)==(0); //@line 57 "modules/graphics/opengl/wrap_Shader.cpp"
    if ($15) {
     $19 = (_lua_type($L,$13)|0); //@line 59 "modules/graphics/opengl/wrap_Shader.cpp"
     $20 = ($19|0)==(1); //@line 59 "modules/graphics/opengl/wrap_Shader.cpp"
     if (!($20)) {
      $$lcssa = $13;
      break;
     }
     $21 = (_lua_toboolean($L,$13)|0); //@line 60 "modules/graphics/opengl/wrap_Shader.cpp"
     $22 = (($12) + ($i$02$i<<2)|0); //@line 60 "modules/graphics/opengl/wrap_Shader.cpp"
     HEAP32[$22>>2] = $21; //@line 60 "modules/graphics/opengl/wrap_Shader.cpp"
    } else {
     $16 = (+_lua_tonumber($L,$13)); //@line 58 "modules/graphics/opengl/wrap_Shader.cpp"
     $17 = (~~(($16))); //@line 58 "modules/graphics/opengl/wrap_Shader.cpp"
     $18 = (($12) + ($i$02$i<<2)|0); //@line 58 "modules/graphics/opengl/wrap_Shader.cpp"
     HEAP32[$18>>2] = $17; //@line 58 "modules/graphics/opengl/wrap_Shader.cpp"
    }
    $23 = (($i$02$i) + 1)|0; //@line 55 "modules/graphics/opengl/wrap_Shader.cpp"
    $24 = ($23|0)<($4|0); //@line 55 "modules/graphics/opengl/wrap_Shader.cpp"
    if ($24) {
     $i$02$i = $23;
    } else {
     $dimension$0$ph7 = 1;$values$0$ph = $12;
     break L30;
    }
   }
   __ZdaPv($12); //@line 63 "modules/graphics/opengl/wrap_Shader.cpp"
   (_luax_typerror($L,$$lcssa,344599)|0); //@line 64 "modules/graphics/opengl/wrap_Shader.cpp"
   label = 28;
  }
 } while(0);
 if ((label|0) == 28) {
  $57 = (_luaL_error($L,344697,$vararg_buffer5)|0); //@line 134 "modules/graphics/opengl/wrap_Shader.cpp"
  $$2 = $57;
  STACKTOP = sp;return ($$2|0); //@line 153 "modules/graphics/opengl/wrap_Shader.cpp"
 }
 $58 = (_strlen($2)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viii(3460,($0|0),($2|0),($58|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $59 = __THREW__; __THREW__ = 0;
 $60 = $59&1;
 do {
  if ($60) {
   $63 = ___cxa_find_matching_catch_3(96|0)|0;
   $64 = tempRet0;
   $$01 = $64;$$03 = $63;
  } else {
   __THREW__ = 0;
   invoke_viiiii(3607,($1|0),($0|0),($dimension$0$ph7|0),($values$0$ph|0),($4|0)); //@line 139 "modules/graphics/opengl/wrap_Shader.cpp"
   $61 = __THREW__; __THREW__ = 0;
   $62 = $61&1;
   if ($62) {
    $65 = ___cxa_find_matching_catch_3(96|0)|0;
    $66 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 139 "modules/graphics/opengl/wrap_Shader.cpp"
    $$01 = $66;$$03 = $65;
    break;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 139 "modules/graphics/opengl/wrap_Shader.cpp"
   __ZdaPv($values$0$ph); //@line 147 "modules/graphics/opengl/wrap_Shader.cpp"
   $$2 = 0;
   STACKTOP = sp;return ($$2|0); //@line 153 "modules/graphics/opengl/wrap_Shader.cpp"
  }
 } while(0);
 $67 = (_llvm_eh_typeid_for((96|0))|0); //@line 140 "modules/graphics/opengl/wrap_Shader.cpp"
 $68 = ($$01|0)==($67|0); //@line 140 "modules/graphics/opengl/wrap_Shader.cpp"
 if (!($68)) {
  $$12 = $$01;$$14 = $$03;
  ___resumeException($$14|0);
  // unreachable;
 }
 $69 = (___cxa_begin_catch(($$03|0))|0); //@line 140 "modules/graphics/opengl/wrap_Shader.cpp"
 $70 = HEAP32[$69>>2]|0; //@line 144 "modules/graphics/opengl/wrap_Shader.cpp"
 $71 = ((($70)) + 8|0); //@line 144 "modules/graphics/opengl/wrap_Shader.cpp"
 $72 = HEAP32[$71>>2]|0; //@line 144 "modules/graphics/opengl/wrap_Shader.cpp"
 $73 = (FUNCTION_TABLE_ii[$72 & 8191]($69)|0); //@line 144 "modules/graphics/opengl/wrap_Shader.cpp"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($73|0)); //@line 144 "modules/graphics/opengl/wrap_Shader.cpp"
 $74 = __THREW__; __THREW__ = 0;
 $75 = $74&1;
 if (!($75)) {
  ___cxa_end_catch(); //@line 145 "modules/graphics/opengl/wrap_Shader.cpp"
  __ZdaPv($values$0$ph); //@line 147 "modules/graphics/opengl/wrap_Shader.cpp"
  $76 = (_lua_tolstring($L,-1,0)|0); //@line 150 "modules/graphics/opengl/wrap_Shader.cpp"
  HEAP32[$vararg_buffer7>>2] = $76; //@line 150 "modules/graphics/opengl/wrap_Shader.cpp"
  $77 = (_luaL_error($L,595867,$vararg_buffer7)|0); //@line 150 "modules/graphics/opengl/wrap_Shader.cpp"
  $$2 = $77;
  STACKTOP = sp;return ($$2|0); //@line 153 "modules/graphics/opengl/wrap_Shader.cpp"
 }
 $78 = ___cxa_find_matching_catch_2()|0;
 $79 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 145 "modules/graphics/opengl/wrap_Shader.cpp"
 $80 = __THREW__; __THREW__ = 0;
 $81 = $80&1;
 if ($81) {
  $82 = ___cxa_find_matching_catch_3(0|0)|0;
  $83 = tempRet0;
  ___clang_call_terminate($82); //@line 145 "modules/graphics/opengl/wrap_Shader.cpp"
  // unreachable; //@line 145 "modules/graphics/opengl/wrap_Shader.cpp"
 } else {
  $$12 = $79;$$14 = $78;
  ___resumeException($$14|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl18w_Shader_sendFloatEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love8graphics6openglL19w__Shader_sendFloatEP9lua_Stateb($L,0)|0); //@line 216 "modules/graphics/opengl/wrap_Shader.cpp"
 return ($0|0); //@line 216 "modules/graphics/opengl/wrap_Shader.cpp"
}
function __ZN4love8graphics6openglL19w__Shader_sendFloatEP9lua_Stateb($L,$colors) {
 $L = $L|0;
 $colors = $colors|0;
 var $$02 = 0, $$04 = 0, $$13 = 0, $$15 = 0, $$2 = 0, $$arith = 0, $$arith11 = 0, $$lcssa = 0, $$lcssa77 = 0, $$lcssa78 = 0, $$overflow = 0, $$overflow12 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0.0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0.0, $45 = 0.0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0.0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0.0, $67 = 0.0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dimension$0$ph10 = 0, $exitcond = 0, $exitcond51 = 0, $exitcond52 = 0, $i$02$i = 0;
 var $i$022$us = 0, $i$08$i = 0, $j$021$us$us = 0, $j$021$us24 = 0, $k$07$i = 0, $or$cond = 0, $storemerge$us$us = 0.0, $values$0$ph = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, $vararg_ptr3 = 0, $vararg_ptr4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 36|0;
 $1 = (__ZN4love14luax_checktypeINS_8graphics6opengl6ShaderEEEPT_P9lua_StateiNS_4TypeE($L,1,18)|0); //@line 39 "modules/graphics/opengl/wrap_Shader.cpp"
 $2 = (_luaL_checklstring($L,2,0)|0); //@line 158 "modules/graphics/opengl/wrap_Shader.cpp"
 $3 = (_lua_gettop($L)|0); //@line 159 "modules/graphics/opengl/wrap_Shader.cpp"
 $4 = (($3) + -2)|0; //@line 159 "modules/graphics/opengl/wrap_Shader.cpp"
 $5 = ($4|0)<(1); //@line 161 "modules/graphics/opengl/wrap_Shader.cpp"
 if ($5) {
  $6 = (_luaL_error($L,344578,$vararg_buffer)|0); //@line 162 "modules/graphics/opengl/wrap_Shader.cpp"
  $$2 = $6;
  STACKTOP = sp;return ($$2|0); //@line 212 "modules/graphics/opengl/wrap_Shader.cpp"
 }
 $7 = (_lua_isnumber($L,3)|0); //@line 167 "modules/graphics/opengl/wrap_Shader.cpp"
 $8 = ($7|0)==(0); //@line 167 "modules/graphics/opengl/wrap_Shader.cpp"
 L5: do {
  if ($8) {
   $9 = (_lua_type($L,3)|0); //@line 167 "modules/graphics/opengl/wrap_Shader.cpp"
   $10 = ($9|0)==(1); //@line 167 "modules/graphics/opengl/wrap_Shader.cpp"
   if ($10) {
    label = 5;
   } else {
    $26 = (_lua_type($L,3)|0); //@line 169 "modules/graphics/opengl/wrap_Shader.cpp"
    $27 = ($26|0)==(5); //@line 169 "modules/graphics/opengl/wrap_Shader.cpp"
    if (!($27)) {
     $58 = (_luax_typerror($L,3,344671)|0); //@line 172 "modules/graphics/opengl/wrap_Shader.cpp"
     $$2 = $58;
     STACKTOP = sp;return ($$2|0); //@line 212 "modules/graphics/opengl/wrap_Shader.cpp"
    }
    $28 = (__ZN4love11luax_objlenEP9lua_Statei($L,3)|0); //@line 75 "modules/graphics/opengl/wrap_Shader.cpp"
    $29 = Math_imul($28, $4)|0; //@line 76 "modules/graphics/opengl/wrap_Shader.cpp"
    $$arith = $29<<2; //@line 76 "modules/graphics/opengl/wrap_Shader.cpp"
    $$overflow = ($29>>>0)>(1073741823); //@line 76 "modules/graphics/opengl/wrap_Shader.cpp"
    $30 = $$overflow ? -1 : $$arith; //@line 76 "modules/graphics/opengl/wrap_Shader.cpp"
    $31 = (__Znaj($30)|0); //@line 76 "modules/graphics/opengl/wrap_Shader.cpp"
    $32 = ($28|0)<(1); //@line 94 "modules/graphics/opengl/wrap_Shader.cpp"
    $33 = $28 ^ -1; //@line 108 "modules/graphics/opengl/wrap_Shader.cpp"
    $i$08$i = 0;
    L12: while(1) {
     $34 = (($i$08$i) + 3)|0; //@line 80 "modules/graphics/opengl/wrap_Shader.cpp"
     $35 = (_lua_type($L,$34)|0); //@line 80 "modules/graphics/opengl/wrap_Shader.cpp"
     $36 = ($35|0)==(5); //@line 80 "modules/graphics/opengl/wrap_Shader.cpp"
     if (!($36)) {
      $$lcssa77 = $34;
      label = 15;
      break;
     }
     $37 = (__ZN4love11luax_objlenEP9lua_Statei($L,$34)|0); //@line 86 "modules/graphics/opengl/wrap_Shader.cpp"
     $38 = ($37|0)==($28|0); //@line 86 "modules/graphics/opengl/wrap_Shader.cpp"
     if (!($38)) {
      $$lcssa78 = $34;
      label = 19;
      break;
     }
     if (!($32)) {
      $39 = Math_imul($i$08$i, $28)|0; //@line 100 "modules/graphics/opengl/wrap_Shader.cpp"
      $40 = (($39) + -1)|0; //@line 100 "modules/graphics/opengl/wrap_Shader.cpp"
      $k$07$i = 1;
      while(1) {
       _lua_rawgeti($L,$34,$k$07$i); //@line 96 "modules/graphics/opengl/wrap_Shader.cpp"
       $42 = (_lua_isnumber($L,-1)|0); //@line 97 "modules/graphics/opengl/wrap_Shader.cpp"
       $43 = ($42|0)==(0); //@line 97 "modules/graphics/opengl/wrap_Shader.cpp"
       if ($43) {
        $48 = (_lua_type($L,-1)|0); //@line 99 "modules/graphics/opengl/wrap_Shader.cpp"
        $49 = ($48|0)==(1); //@line 99 "modules/graphics/opengl/wrap_Shader.cpp"
        if (!($49)) {
         label = 25;
         break L12;
        }
        $50 = (_lua_toboolean($L,-1)|0); //@line 100 "modules/graphics/opengl/wrap_Shader.cpp"
        $51 = (+($50|0)); //@line 100 "modules/graphics/opengl/wrap_Shader.cpp"
        $52 = (($40) + ($k$07$i))|0; //@line 100 "modules/graphics/opengl/wrap_Shader.cpp"
        $53 = (($31) + ($52<<2)|0); //@line 100 "modules/graphics/opengl/wrap_Shader.cpp"
        HEAPF32[$53>>2] = $51; //@line 100 "modules/graphics/opengl/wrap_Shader.cpp"
       } else {
        $44 = (+_lua_tonumber($L,-1)); //@line 98 "modules/graphics/opengl/wrap_Shader.cpp"
        $45 = $44; //@line 98 "modules/graphics/opengl/wrap_Shader.cpp"
        $46 = (($40) + ($k$07$i))|0; //@line 98 "modules/graphics/opengl/wrap_Shader.cpp"
        $47 = (($31) + ($46<<2)|0); //@line 98 "modules/graphics/opengl/wrap_Shader.cpp"
        HEAPF32[$47>>2] = $45; //@line 98 "modules/graphics/opengl/wrap_Shader.cpp"
       }
       $54 = (($k$07$i) + 1)|0; //@line 94 "modules/graphics/opengl/wrap_Shader.cpp"
       $55 = ($k$07$i|0)<($28|0); //@line 94 "modules/graphics/opengl/wrap_Shader.cpp"
       if ($55) {
        $k$07$i = $54;
       } else {
        break;
       }
      }
     }
     _lua_settop($L,$33); //@line 108 "modules/graphics/opengl/wrap_Shader.cpp"
     $56 = (($i$08$i) + 1)|0; //@line 78 "modules/graphics/opengl/wrap_Shader.cpp"
     $57 = ($56|0)<($4|0); //@line 78 "modules/graphics/opengl/wrap_Shader.cpp"
     if ($57) {
      $i$08$i = $56;
     } else {
      $dimension$0$ph10 = $28;$values$0$ph = $31;
      break L5;
     }
    }
    if ((label|0) == 15) {
     __ZdaPv($31); //@line 82 "modules/graphics/opengl/wrap_Shader.cpp"
     (_luax_typerror($L,$$lcssa77,522280)|0); //@line 83 "modules/graphics/opengl/wrap_Shader.cpp"
     label = 28;
     break;
    }
    else if ((label|0) == 19) {
     __ZdaPv($31); //@line 88 "modules/graphics/opengl/wrap_Shader.cpp"
     $41 = (__ZN4love11luax_objlenEP9lua_Statei($L,$$lcssa78)|0); //@line 90 "modules/graphics/opengl/wrap_Shader.cpp"
     HEAP32[$vararg_buffer1>>2] = $$lcssa78; //@line 89 "modules/graphics/opengl/wrap_Shader.cpp"
     $vararg_ptr3 = ((($vararg_buffer1)) + 4|0); //@line 89 "modules/graphics/opengl/wrap_Shader.cpp"
     HEAP32[$vararg_ptr3>>2] = $28; //@line 89 "modules/graphics/opengl/wrap_Shader.cpp"
     $vararg_ptr4 = ((($vararg_buffer1)) + 8|0); //@line 89 "modules/graphics/opengl/wrap_Shader.cpp"
     HEAP32[$vararg_ptr4>>2] = $41; //@line 89 "modules/graphics/opengl/wrap_Shader.cpp"
     (_luaL_error($L,344617,$vararg_buffer1)|0); //@line 89 "modules/graphics/opengl/wrap_Shader.cpp"
     label = 28;
     break;
    }
    else if ((label|0) == 25) {
     __ZdaPv($31); //@line 103 "modules/graphics/opengl/wrap_Shader.cpp"
     (_luax_typerror($L,-1,344599)|0); //@line 104 "modules/graphics/opengl/wrap_Shader.cpp"
     label = 28;
     break;
    }
   }
  } else {
   label = 5;
  }
 } while(0);
 L30: do {
  if ((label|0) == 5) {
   $$arith11 = $4<<2; //@line 53 "modules/graphics/opengl/wrap_Shader.cpp"
   $$overflow12 = ($4>>>0)>(1073741823); //@line 53 "modules/graphics/opengl/wrap_Shader.cpp"
   $11 = $$overflow12 ? -1 : $$arith11; //@line 53 "modules/graphics/opengl/wrap_Shader.cpp"
   $12 = (__Znaj($11)|0); //@line 53 "modules/graphics/opengl/wrap_Shader.cpp"
   $i$02$i = 0;
   while(1) {
    $13 = (($i$02$i) + 3)|0; //@line 57 "modules/graphics/opengl/wrap_Shader.cpp"
    $14 = (_lua_isnumber($L,$13)|0); //@line 57 "modules/graphics/opengl/wrap_Shader.cpp"
    $15 = ($14|0)==(0); //@line 57 "modules/graphics/opengl/wrap_Shader.cpp"
    if ($15) {
     $19 = (_lua_type($L,$13)|0); //@line 59 "modules/graphics/opengl/wrap_Shader.cpp"
     $20 = ($19|0)==(1); //@line 59 "modules/graphics/opengl/wrap_Shader.cpp"
     if (!($20)) {
      $$lcssa = $13;
      break;
     }
     $21 = (_lua_toboolean($L,$13)|0); //@line 60 "modules/graphics/opengl/wrap_Shader.cpp"
     $22 = (+($21|0)); //@line 60 "modules/graphics/opengl/wrap_Shader.cpp"
     $23 = (($12) + ($i$02$i<<2)|0); //@line 60 "modules/graphics/opengl/wrap_Shader.cpp"
     HEAPF32[$23>>2] = $22; //@line 60 "modules/graphics/opengl/wrap_Shader.cpp"
    } else {
     $16 = (+_lua_tonumber($L,$13)); //@line 58 "modules/graphics/opengl/wrap_Shader.cpp"
     $17 = $16; //@line 58 "modules/graphics/opengl/wrap_Shader.cpp"
     $18 = (($12) + ($i$02$i<<2)|0); //@line 58 "modules/graphics/opengl/wrap_Shader.cpp"
     HEAPF32[$18>>2] = $17; //@line 58 "modules/graphics/opengl/wrap_Shader.cpp"
    }
    $24 = (($i$02$i) + 1)|0; //@line 55 "modules/graphics/opengl/wrap_Shader.cpp"
    $25 = ($24|0)<($4|0); //@line 55 "modules/graphics/opengl/wrap_Shader.cpp"
    if ($25) {
     $i$02$i = $24;
    } else {
     $dimension$0$ph10 = 1;$values$0$ph = $12;
     break L30;
    }
   }
   __ZdaPv($12); //@line 63 "modules/graphics/opengl/wrap_Shader.cpp"
   (_luax_typerror($L,$$lcssa,344599)|0); //@line 64 "modules/graphics/opengl/wrap_Shader.cpp"
   label = 28;
  }
 } while(0);
 if ((label|0) == 28) {
  $59 = (_luaL_error($L,344697,$vararg_buffer5)|0); //@line 175 "modules/graphics/opengl/wrap_Shader.cpp"
  $$2 = $59;
  STACKTOP = sp;return ($$2|0); //@line 212 "modules/graphics/opengl/wrap_Shader.cpp"
 }
 if ($colors) {
  $60 = (__ZN4love8graphics14isGammaCorrectEv()|0); //@line 179 "modules/graphics/opengl/wrap_Shader.cpp"
  $61 = ($3|0)>(2); //@line 182 "modules/graphics/opengl/wrap_Shader.cpp"
  $62 = ($dimension$0$ph10|0)>(0); //@line 184 "modules/graphics/opengl/wrap_Shader.cpp"
  $or$cond = $61 & $62; //@line 182 "modules/graphics/opengl/wrap_Shader.cpp"
  if ($or$cond) {
   $i$022$us = 0;
   while(1) {
    $64 = Math_imul($i$022$us, $dimension$0$ph10)|0; //@line 188 "modules/graphics/opengl/wrap_Shader.cpp"
    if ($60) {
     $j$021$us$us = 0;
     while(1) {
      $70 = ($j$021$us$us|0)<(3); //@line 187 "modules/graphics/opengl/wrap_Shader.cpp"
      $71 = (($j$021$us$us) + ($64))|0; //@line 188 "modules/graphics/opengl/wrap_Shader.cpp"
      $72 = (($values$0$ph) + ($71<<2)|0); //@line 188 "modules/graphics/opengl/wrap_Shader.cpp"
      $73 = +HEAPF32[$72>>2]; //@line 188 "modules/graphics/opengl/wrap_Shader.cpp"
      $74 = $73 / 255.0; //@line 188 "modules/graphics/opengl/wrap_Shader.cpp"
      if ($70) {
       $75 = (+__ZNK4love4math4Math13gammaToLinearEf(623984,$74)); //@line 188 "modules/graphics/opengl/wrap_Shader.cpp"
       $storemerge$us$us = $75;
      } else {
       $storemerge$us$us = $74;
      }
      HEAPF32[$72>>2] = $storemerge$us$us; //@line 188 "modules/graphics/opengl/wrap_Shader.cpp"
      $76 = (($j$021$us$us) + 1)|0; //@line 184 "modules/graphics/opengl/wrap_Shader.cpp"
      $exitcond51 = ($76|0)==($dimension$0$ph10|0); //@line 184 "modules/graphics/opengl/wrap_Shader.cpp"
      if ($exitcond51) {
       break;
      } else {
       $j$021$us$us = $76;
      }
     }
    } else {
     $j$021$us24 = 0;
     while(1) {
      $63 = (($j$021$us24) + ($64))|0; //@line 188 "modules/graphics/opengl/wrap_Shader.cpp"
      $65 = (($values$0$ph) + ($63<<2)|0); //@line 188 "modules/graphics/opengl/wrap_Shader.cpp"
      $66 = +HEAPF32[$65>>2]; //@line 188 "modules/graphics/opengl/wrap_Shader.cpp"
      $67 = $66 / 255.0; //@line 188 "modules/graphics/opengl/wrap_Shader.cpp"
      HEAPF32[$65>>2] = $67; //@line 188 "modules/graphics/opengl/wrap_Shader.cpp"
      $68 = (($j$021$us24) + 1)|0; //@line 184 "modules/graphics/opengl/wrap_Shader.cpp"
      $exitcond = ($68|0)==($dimension$0$ph10|0); //@line 184 "modules/graphics/opengl/wrap_Shader.cpp"
      if ($exitcond) {
       break;
      } else {
       $j$021$us24 = $68;
      }
     }
    }
    $69 = (($i$022$us) + 1)|0; //@line 182 "modules/graphics/opengl/wrap_Shader.cpp"
    $exitcond52 = ($69|0)==($4|0); //@line 182 "modules/graphics/opengl/wrap_Shader.cpp"
    if ($exitcond52) {
     break;
    } else {
     $i$022$us = $69;
    }
   }
  }
 }
 $77 = (_strlen($2)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viii(3460,($0|0),($2|0),($77|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $78 = __THREW__; __THREW__ = 0;
 $79 = $78&1;
 do {
  if ($79) {
   $82 = ___cxa_find_matching_catch_3(96|0)|0;
   $83 = tempRet0;
   $$02 = $83;$$04 = $82;
  } else {
   __THREW__ = 0;
   invoke_viiiii(3608,($1|0),($0|0),($dimension$0$ph10|0),($values$0$ph|0),($4|0)); //@line 198 "modules/graphics/opengl/wrap_Shader.cpp"
   $80 = __THREW__; __THREW__ = 0;
   $81 = $80&1;
   if ($81) {
    $84 = ___cxa_find_matching_catch_3(96|0)|0;
    $85 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 198 "modules/graphics/opengl/wrap_Shader.cpp"
    $$02 = $85;$$04 = $84;
    break;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 198 "modules/graphics/opengl/wrap_Shader.cpp"
   __ZdaPv($values$0$ph); //@line 206 "modules/graphics/opengl/wrap_Shader.cpp"
   $$2 = 0;
   STACKTOP = sp;return ($$2|0); //@line 212 "modules/graphics/opengl/wrap_Shader.cpp"
  }
 } while(0);
 $86 = (_llvm_eh_typeid_for((96|0))|0); //@line 199 "modules/graphics/opengl/wrap_Shader.cpp"
 $87 = ($$02|0)==($86|0); //@line 199 "modules/graphics/opengl/wrap_Shader.cpp"
 if (!($87)) {
  $$13 = $$02;$$15 = $$04;
  ___resumeException($$15|0);
  // unreachable;
 }
 $88 = (___cxa_begin_catch(($$04|0))|0); //@line 199 "modules/graphics/opengl/wrap_Shader.cpp"
 $89 = HEAP32[$88>>2]|0; //@line 203 "modules/graphics/opengl/wrap_Shader.cpp"
 $90 = ((($89)) + 8|0); //@line 203 "modules/graphics/opengl/wrap_Shader.cpp"
 $91 = HEAP32[$90>>2]|0; //@line 203 "modules/graphics/opengl/wrap_Shader.cpp"
 $92 = (FUNCTION_TABLE_ii[$91 & 8191]($88)|0); //@line 203 "modules/graphics/opengl/wrap_Shader.cpp"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($92|0)); //@line 203 "modules/graphics/opengl/wrap_Shader.cpp"
 $93 = __THREW__; __THREW__ = 0;
 $94 = $93&1;
 if (!($94)) {
  ___cxa_end_catch(); //@line 204 "modules/graphics/opengl/wrap_Shader.cpp"
  __ZdaPv($values$0$ph); //@line 206 "modules/graphics/opengl/wrap_Shader.cpp"
  $95 = (_lua_tolstring($L,-1,0)|0); //@line 209 "modules/graphics/opengl/wrap_Shader.cpp"
  HEAP32[$vararg_buffer7>>2] = $95; //@line 209 "modules/graphics/opengl/wrap_Shader.cpp"
  $96 = (_luaL_error($L,595867,$vararg_buffer7)|0); //@line 209 "modules/graphics/opengl/wrap_Shader.cpp"
  $$2 = $96;
  STACKTOP = sp;return ($$2|0); //@line 212 "modules/graphics/opengl/wrap_Shader.cpp"
 }
 $97 = ___cxa_find_matching_catch_2()|0;
 $98 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 204 "modules/graphics/opengl/wrap_Shader.cpp"
 $99 = __THREW__; __THREW__ = 0;
 $100 = $99&1;
 if ($100) {
  $101 = ___cxa_find_matching_catch_3(0|0)|0;
  $102 = tempRet0;
  ___clang_call_terminate($101); //@line 204 "modules/graphics/opengl/wrap_Shader.cpp"
  // unreachable; //@line 204 "modules/graphics/opengl/wrap_Shader.cpp"
 } else {
  $$13 = $98;$$15 = $97;
  ___resumeException($$15|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl18w_Shader_sendColorEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love8graphics6openglL19w__Shader_sendFloatEP9lua_Stateb($L,1)|0); //@line 221 "modules/graphics/opengl/wrap_Shader.cpp"
 return ($0|0); //@line 221 "modules/graphics/opengl/wrap_Shader.cpp"
}
function __ZN4love8graphics6opengl19w_Shader_sendMatrixEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $$5 = 0, $$arith = 0, $$lcssa = 0, $$off = 0, $$overflow = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $9 = 0, $dimension$1 = 0, $eh$lpad$body$i$index13Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $exitcond = 0, $exitcond33 = 0, $i$022 = 0, $j$020 = 0;
 var $k$015 = 0, $k1$013 = 0, $n$019 = 0, $n$114 = 0, $other_dimension$1 = 0, $other_dimension$1$lcssa = 0, $storemerge = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_buffer9 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, $vararg_ptr6 = 0, $vararg_ptr7 = 0, $vararg_ptr8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer9 = sp + 32|0;
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 36|0;
 $1 = (_lua_gettop($L)|0); //@line 226 "modules/graphics/opengl/wrap_Shader.cpp"
 $2 = (($1) + -2)|0; //@line 226 "modules/graphics/opengl/wrap_Shader.cpp"
 $3 = (__ZN4love14luax_checktypeINS_8graphics6opengl6ShaderEEEPT_P9lua_StateiNS_4TypeE($L,1,18)|0); //@line 39 "modules/graphics/opengl/wrap_Shader.cpp"
 $4 = (_luaL_checklstring($L,2,0)|0); //@line 228 "modules/graphics/opengl/wrap_Shader.cpp"
 $5 = (_lua_type($L,3)|0); //@line 230 "modules/graphics/opengl/wrap_Shader.cpp"
 $6 = ($5|0)==(5); //@line 230 "modules/graphics/opengl/wrap_Shader.cpp"
 if (!($6)) {
  $7 = (_luax_typerror($L,3,344717)|0); //@line 231 "modules/graphics/opengl/wrap_Shader.cpp"
  $$5 = $7;
  STACKTOP = sp;return ($$5|0); //@line 321 "modules/graphics/opengl/wrap_Shader.cpp"
 }
 _lua_rawgeti($L,3,1); //@line 235 "modules/graphics/opengl/wrap_Shader.cpp"
 $8 = (_lua_type($L,-1)|0); //@line 236 "modules/graphics/opengl/wrap_Shader.cpp"
 $9 = ($8|0)==(5); //@line 236 "modules/graphics/opengl/wrap_Shader.cpp"
 if ($9) {
  $10 = (__ZN4love11luax_objlenEP9lua_Statei($L,3)|0); //@line 237 "modules/graphics/opengl/wrap_Shader.cpp"
  _lua_settop($L,-2); //@line 238 "modules/graphics/opengl/wrap_Shader.cpp"
  $11 = ($10|0)==(0); //@line 240 "modules/graphics/opengl/wrap_Shader.cpp"
  if ($11) {
   label = 6;
  } else {
   $dimension$1 = $10;
  }
 } else {
  _lua_settop($L,-2); //@line 238 "modules/graphics/opengl/wrap_Shader.cpp"
  label = 6;
 }
 if ((label|0) == 6) {
  _lua_getfield($L,3,344730); //@line 242 "modules/graphics/opengl/wrap_Shader.cpp"
  $12 = (_lua_type($L,-1)|0); //@line 244 "modules/graphics/opengl/wrap_Shader.cpp"
  $13 = ($12|0)<(1); //@line 244 "modules/graphics/opengl/wrap_Shader.cpp"
  if ($13) {
   $15 = (__ZN4love11luax_objlenEP9lua_Statei($L,3)|0); //@line 247 "modules/graphics/opengl/wrap_Shader.cpp"
   $16 = (+($15>>>0)); //@line 247 "modules/graphics/opengl/wrap_Shader.cpp"
   $17 = (+Math_sqrt((+$16))); //@line 247 "modules/graphics/opengl/wrap_Shader.cpp"
   $18 = (~~(($17))); //@line 247 "modules/graphics/opengl/wrap_Shader.cpp"
   $storemerge = $18;
  } else {
   $14 = (_lua_tointeger($L,-1)|0); //@line 245 "modules/graphics/opengl/wrap_Shader.cpp"
   $storemerge = $14;
  }
  _lua_settop($L,-2); //@line 249 "modules/graphics/opengl/wrap_Shader.cpp"
  $dimension$1 = $storemerge;
 }
 $$off = (($dimension$1) + -2)|0; //@line 252 "modules/graphics/opengl/wrap_Shader.cpp"
 $19 = ($$off>>>0)>(2); //@line 252 "modules/graphics/opengl/wrap_Shader.cpp"
 if ($19) {
  HEAP32[$vararg_buffer>>2] = $dimension$1; //@line 253 "modules/graphics/opengl/wrap_Shader.cpp"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 253 "modules/graphics/opengl/wrap_Shader.cpp"
  HEAP32[$vararg_ptr1>>2] = $dimension$1; //@line 253 "modules/graphics/opengl/wrap_Shader.cpp"
  $20 = (_luaL_error($L,344740,$vararg_buffer)|0); //@line 253 "modules/graphics/opengl/wrap_Shader.cpp"
  $$5 = $20;
  STACKTOP = sp;return ($$5|0); //@line 321 "modules/graphics/opengl/wrap_Shader.cpp"
 }
 $21 = Math_imul($dimension$1, $dimension$1)|0; //@line 256 "modules/graphics/opengl/wrap_Shader.cpp"
 $22 = Math_imul($21, $2)|0; //@line 256 "modules/graphics/opengl/wrap_Shader.cpp"
 $$arith = $22<<2; //@line 256 "modules/graphics/opengl/wrap_Shader.cpp"
 $$overflow = ($22>>>0)>(1073741823); //@line 256 "modules/graphics/opengl/wrap_Shader.cpp"
 $23 = $$overflow ? -1 : $$arith; //@line 256 "modules/graphics/opengl/wrap_Shader.cpp"
 $24 = (__Znaj($23)|0); //@line 256 "modules/graphics/opengl/wrap_Shader.cpp"
 $25 = ($1|0)>(2); //@line 258 "modules/graphics/opengl/wrap_Shader.cpp"
 L19: do {
  if ($25) {
   $26 = ($dimension$1|0)<(1); //@line 289 "modules/graphics/opengl/wrap_Shader.cpp"
   $27 = (-2 - ($dimension$1))|0; //@line 300 "modules/graphics/opengl/wrap_Shader.cpp"
   $28 = ($21|0)==(0); //@line 305 "modules/graphics/opengl/wrap_Shader.cpp"
   $29 = $21 ^ -1; //@line 311 "modules/graphics/opengl/wrap_Shader.cpp"
   $30 = (($dimension$1) + 1)|0; //@line 258 "modules/graphics/opengl/wrap_Shader.cpp"
   $i$022 = 0;
   while(1) {
    $31 = (($i$022) + 3)|0; //@line 262 "modules/graphics/opengl/wrap_Shader.cpp"
    _lua_rawgeti($L,$31,1); //@line 262 "modules/graphics/opengl/wrap_Shader.cpp"
    $32 = (_lua_type($L,-1)|0); //@line 263 "modules/graphics/opengl/wrap_Shader.cpp"
    $33 = ($32|0)==(5); //@line 263 "modules/graphics/opengl/wrap_Shader.cpp"
    if ($33) {
     $34 = (__ZN4love11luax_objlenEP9lua_Statei($L,-1)|0); //@line 266 "modules/graphics/opengl/wrap_Shader.cpp"
     _lua_settop($L,-2); //@line 268 "modules/graphics/opengl/wrap_Shader.cpp"
     $other_dimension$1 = $34;
    } else {
     _lua_settop($L,-2); //@line 268 "modules/graphics/opengl/wrap_Shader.cpp"
     $35 = (__ZN4love11luax_objlenEP9lua_Statei($L,$31)|0); //@line 271 "modules/graphics/opengl/wrap_Shader.cpp"
     $36 = (+($35>>>0)); //@line 271 "modules/graphics/opengl/wrap_Shader.cpp"
     $37 = (+Math_sqrt((+$36))); //@line 271 "modules/graphics/opengl/wrap_Shader.cpp"
     $38 = (~~(($37))); //@line 271 "modules/graphics/opengl/wrap_Shader.cpp"
     $other_dimension$1 = $38;
    }
    $39 = ($other_dimension$1|0)==($dimension$1|0); //@line 273 "modules/graphics/opengl/wrap_Shader.cpp"
    if (!($39)) {
     $$lcssa = $31;$other_dimension$1$lcssa = $other_dimension$1;
     break;
    }
    if ($33) {
     if (!($26)) {
      $42 = Math_imul($21, $i$022)|0; //@line 296 "modules/graphics/opengl/wrap_Shader.cpp"
      $j$020 = 1;$n$019 = 0;
      while(1) {
       _lua_rawgeti($L,$31,$j$020); //@line 291 "modules/graphics/opengl/wrap_Shader.cpp"
       $k$015 = 1;$n$114 = $n$019;
       while(1) {
        $45 = (0 - ($k$015))|0; //@line 295 "modules/graphics/opengl/wrap_Shader.cpp"
        _lua_rawgeti($L,$45,$k$015); //@line 295 "modules/graphics/opengl/wrap_Shader.cpp"
        $46 = (+_lua_tonumber($L,-1)); //@line 296 "modules/graphics/opengl/wrap_Shader.cpp"
        $47 = $46; //@line 296 "modules/graphics/opengl/wrap_Shader.cpp"
        $48 = (($n$114) + ($42))|0; //@line 296 "modules/graphics/opengl/wrap_Shader.cpp"
        $49 = (($24) + ($48<<2)|0); //@line 296 "modules/graphics/opengl/wrap_Shader.cpp"
        HEAPF32[$49>>2] = $47; //@line 296 "modules/graphics/opengl/wrap_Shader.cpp"
        $50 = (($n$114) + 1)|0; //@line 297 "modules/graphics/opengl/wrap_Shader.cpp"
        $51 = (($k$015) + 1)|0; //@line 293 "modules/graphics/opengl/wrap_Shader.cpp"
        $exitcond = ($51|0)==($30|0); //@line 293 "modules/graphics/opengl/wrap_Shader.cpp"
        if ($exitcond) {
         break;
        } else {
         $k$015 = $51;$n$114 = $50;
        }
       }
       $43 = (($dimension$1) + ($n$019))|0; //@line 293 "modules/graphics/opengl/wrap_Shader.cpp"
       _lua_settop($L,$27); //@line 300 "modules/graphics/opengl/wrap_Shader.cpp"
       $44 = (($j$020) + 1)|0; //@line 289 "modules/graphics/opengl/wrap_Shader.cpp"
       $exitcond33 = ($44|0)==($30|0); //@line 289 "modules/graphics/opengl/wrap_Shader.cpp"
       if ($exitcond33) {
        break;
       } else {
        $j$020 = $44;$n$019 = $43;
       }
      }
     }
    } else {
     if (!($28)) {
      $40 = Math_imul($21, $i$022)|0; //@line 308 "modules/graphics/opengl/wrap_Shader.cpp"
      $41 = (($40) + -1)|0; //@line 308 "modules/graphics/opengl/wrap_Shader.cpp"
      $k1$013 = 1;
      while(1) {
       _lua_rawgeti($L,$31,$k1$013); //@line 307 "modules/graphics/opengl/wrap_Shader.cpp"
       $52 = (+_lua_tonumber($L,-1)); //@line 308 "modules/graphics/opengl/wrap_Shader.cpp"
       $53 = $52; //@line 308 "modules/graphics/opengl/wrap_Shader.cpp"
       $54 = (($41) + ($k1$013))|0; //@line 308 "modules/graphics/opengl/wrap_Shader.cpp"
       $55 = (($24) + ($54<<2)|0); //@line 308 "modules/graphics/opengl/wrap_Shader.cpp"
       HEAPF32[$55>>2] = $53; //@line 308 "modules/graphics/opengl/wrap_Shader.cpp"
       $56 = (($k1$013) + 1)|0; //@line 305 "modules/graphics/opengl/wrap_Shader.cpp"
       $57 = ($k1$013|0)<($21|0); //@line 305 "modules/graphics/opengl/wrap_Shader.cpp"
       if ($57) {
        $k1$013 = $56;
       } else {
        break;
       }
      }
     }
     _lua_settop($L,$29); //@line 311 "modules/graphics/opengl/wrap_Shader.cpp"
    }
    $58 = (($i$022) + 1)|0; //@line 258 "modules/graphics/opengl/wrap_Shader.cpp"
    $59 = ($58|0)<($2|0); //@line 258 "modules/graphics/opengl/wrap_Shader.cpp"
    if ($59) {
     $i$022 = $58;
    } else {
     break L19;
    }
   }
   __ZdaPv($24); //@line 280 "modules/graphics/opengl/wrap_Shader.cpp"
   HEAP32[$vararg_buffer2>>2] = $$lcssa; //@line 281 "modules/graphics/opengl/wrap_Shader.cpp"
   $vararg_ptr5 = ((($vararg_buffer2)) + 4|0); //@line 281 "modules/graphics/opengl/wrap_Shader.cpp"
   HEAP32[$vararg_ptr5>>2] = $dimension$1; //@line 281 "modules/graphics/opengl/wrap_Shader.cpp"
   $vararg_ptr6 = ((($vararg_buffer2)) + 8|0); //@line 281 "modules/graphics/opengl/wrap_Shader.cpp"
   HEAP32[$vararg_ptr6>>2] = $dimension$1; //@line 281 "modules/graphics/opengl/wrap_Shader.cpp"
   $vararg_ptr7 = ((($vararg_buffer2)) + 12|0); //@line 281 "modules/graphics/opengl/wrap_Shader.cpp"
   HEAP32[$vararg_ptr7>>2] = $other_dimension$1$lcssa; //@line 281 "modules/graphics/opengl/wrap_Shader.cpp"
   $vararg_ptr8 = ((($vararg_buffer2)) + 16|0); //@line 281 "modules/graphics/opengl/wrap_Shader.cpp"
   HEAP32[$vararg_ptr8>>2] = $other_dimension$1$lcssa; //@line 281 "modules/graphics/opengl/wrap_Shader.cpp"
   $60 = (_luaL_error($L,344815,$vararg_buffer2)|0); //@line 281 "modules/graphics/opengl/wrap_Shader.cpp"
   $$5 = $60;
   STACKTOP = sp;return ($$5|0); //@line 321 "modules/graphics/opengl/wrap_Shader.cpp"
  }
 } while(0);
 $61 = (_strlen($4)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viii(3460,($0|0),($4|0),($61|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $62 = __THREW__; __THREW__ = 0;
 $63 = $62&1;
 do {
  if ($63) {
   $68 = ___cxa_find_matching_catch_3(8992|0)|0;
   $69 = tempRet0;
   $eh$lpad$body$i$index13Z2D = $69;$eh$lpad$body$i$indexZ2D = $68;
  } else {
   __THREW__ = 0;
   invoke_viiiii(3609,($3|0),($0|0),($dimension$1|0),($24|0),($2|0)); //@line 316 "modules/graphics/opengl/wrap_Shader.cpp"
   $64 = __THREW__; __THREW__ = 0;
   $65 = $64&1;
   if ($65) {
    $66 = ___cxa_find_matching_catch_3(8992|0)|0;
    $67 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 316 "modules/graphics/opengl/wrap_Shader.cpp"
    $eh$lpad$body$i$index13Z2D = $67;$eh$lpad$body$i$indexZ2D = $66;
    break;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 316 "modules/graphics/opengl/wrap_Shader.cpp"
   __ZdaPv($24); //@line 317 "modules/graphics/opengl/wrap_Shader.cpp"
   $$5 = 0;
   STACKTOP = sp;return ($$5|0); //@line 321 "modules/graphics/opengl/wrap_Shader.cpp"
  }
 } while(0);
 $70 = (_llvm_eh_typeid_for((8992|0))|0); //@line 556 "./common/runtime.h"
 $71 = ($eh$lpad$body$i$index13Z2D|0)==($70|0); //@line 556 "./common/runtime.h"
 if (!($71)) {
  $$01$i = $eh$lpad$body$i$index13Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $72 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 556 "./common/runtime.h"
 $73 = HEAP32[$72>>2]|0; //@line 560 "./common/runtime.h"
 $74 = ((($73)) + 8|0); //@line 560 "./common/runtime.h"
 $75 = HEAP32[$74>>2]|0; //@line 560 "./common/runtime.h"
 $76 = (FUNCTION_TABLE_ii[$75 & 8191]($72)|0); //@line 560 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($76|0)); //@line 560 "./common/runtime.h"
 $77 = __THREW__; __THREW__ = 0;
 $78 = $77&1;
 if (!($78)) {
  ___cxa_end_catch(); //@line 561 "./common/runtime.h"
  __ZdaPv($24); //@line 317 "modules/graphics/opengl/wrap_Shader.cpp"
  $79 = (_lua_tolstring($L,-1,0)|0); //@line 566 "./common/runtime.h"
  HEAP32[$vararg_buffer9>>2] = $79; //@line 566 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer9)|0); //@line 566 "./common/runtime.h"
  $$5 = 0;
  STACKTOP = sp;return ($$5|0); //@line 321 "modules/graphics/opengl/wrap_Shader.cpp"
 }
 $80 = ___cxa_find_matching_catch_2()|0;
 $81 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 561 "./common/runtime.h"
 $82 = __THREW__; __THREW__ = 0;
 $83 = $82&1;
 if ($83) {
  $84 = ___cxa_find_matching_catch_3(0|0)|0;
  $85 = tempRet0;
  ___clang_call_terminate($84); //@line 561 "./common/runtime.h"
  // unreachable; //@line 561 "./common/runtime.h"
 } else {
  $$01$i = $81;$$02$i = $80;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl20w_Shader_sendTextureEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 $1 = (__ZN4love14luax_checktypeINS_8graphics6opengl6ShaderEEEPT_P9lua_StateiNS_4TypeE($L,1,18)|0); //@line 39 "modules/graphics/opengl/wrap_Shader.cpp"
 $2 = (_luaL_checklstring($L,2,0)|0); //@line 326 "modules/graphics/opengl/wrap_Shader.cpp"
 $3 = (__ZN4love8graphics17luax_checktextureEP9lua_Statei($L,3)|0); //@line 327 "modules/graphics/opengl/wrap_Shader.cpp"
 $4 = (_strlen($2)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viii(3460,($0|0),($2|0),($4|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 do {
  if ($6) {
   $11 = ___cxa_find_matching_catch_3(8992|0)|0;
   $12 = tempRet0;
   $eh$lpad$body$i$index2Z2D = $12;$eh$lpad$body$i$indexZ2D = $11;
  } else {
   __THREW__ = 0;
   invoke_viii(3610,($1|0),($0|0),($3|0)); //@line 329 "modules/graphics/opengl/wrap_Shader.cpp"
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if ($8) {
    $9 = ___cxa_find_matching_catch_3(8992|0)|0;
    $10 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 329 "modules/graphics/opengl/wrap_Shader.cpp"
    $eh$lpad$body$i$index2Z2D = $10;$eh$lpad$body$i$indexZ2D = $9;
    break;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 329 "modules/graphics/opengl/wrap_Shader.cpp"
   STACKTOP = sp;return 0; //@line 330 "modules/graphics/opengl/wrap_Shader.cpp"
  }
 } while(0);
 $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $14 = ($eh$lpad$body$i$index2Z2D|0)==($13|0); //@line 535 "./common/runtime.h"
 if (!($14)) {
  $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $15 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
 $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
 $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
 $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
 $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if (!($21)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 330 "modules/graphics/opengl/wrap_Shader.cpp"
 }
 $23 = ___cxa_find_matching_catch_2()|0;
 $24 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if ($26) {
  $27 = ___cxa_find_matching_catch_3(0|0)|0;
  $28 = tempRet0;
  ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $24;$$02$i = $23;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl13w_Shader_sendEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_type($L,3)|0); //@line 335 "modules/graphics/opengl/wrap_Shader.cpp"
 switch ($0|0) {
 case 1: case 3:  {
  $1 = (__ZN4love8graphics6openglL19w__Shader_sendFloatEP9lua_Stateb($L,0)|0); //@line 216 "modules/graphics/opengl/wrap_Shader.cpp"
  $$0 = $1;
  return ($$0|0); //@line 370 "modules/graphics/opengl/wrap_Shader.cpp"
  break;
 }
 case 7:  {
  $2 = (_lua_touserdata($L,3)|0); //@line 347 "modules/graphics/opengl/wrap_Shader.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 349 "modules/graphics/opengl/wrap_Shader.cpp"
  $4 = HEAP32[156504]|0; //@line 349 "modules/graphics/opengl/wrap_Shader.cpp"
  $5 = (($4) + ($3<<3)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
  $6 = HEAP32[$5>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $7 = $6 & 2048; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $8 = ($7|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  if (!($8)) {
   (__ZN4love8graphics6opengl20w_Shader_sendTextureEP9lua_State($L)|0); //@line 350 "modules/graphics/opengl/wrap_Shader.cpp"
   $$0 = 0;
   return ($$0|0); //@line 370 "modules/graphics/opengl/wrap_Shader.cpp"
  }
  break;
 }
 case 5:  {
  _lua_rawgeti($L,3,1); //@line 355 "modules/graphics/opengl/wrap_Shader.cpp"
  $9 = (_lua_type($L,-1)|0); //@line 356 "modules/graphics/opengl/wrap_Shader.cpp"
  _lua_settop($L,-2); //@line 357 "modules/graphics/opengl/wrap_Shader.cpp"
  $10 = $9 | 2; //@line 359 "modules/graphics/opengl/wrap_Shader.cpp"
  $11 = ($10|0)==(3); //@line 359 "modules/graphics/opengl/wrap_Shader.cpp"
  if ($11) {
   $12 = (__ZN4love8graphics6openglL19w__Shader_sendFloatEP9lua_Stateb($L,0)|0); //@line 216 "modules/graphics/opengl/wrap_Shader.cpp"
   $$0 = $12;
   return ($$0|0); //@line 370 "modules/graphics/opengl/wrap_Shader.cpp"
  }
  $13 = ($9|0)==(5); //@line 361 "modules/graphics/opengl/wrap_Shader.cpp"
  if ($13) {
   $14 = (__ZN4love8graphics6opengl19w_Shader_sendMatrixEP9lua_State($L)|0); //@line 362 "modules/graphics/opengl/wrap_Shader.cpp"
   $$0 = $14;
   return ($$0|0); //@line 370 "modules/graphics/opengl/wrap_Shader.cpp"
  }
  break;
 }
 default: {
 }
 }
 $15 = (_luaL_argerror($L,3,344883)|0); //@line 369 "modules/graphics/opengl/wrap_Shader.cpp"
 $$0 = $15;
 return ($$0|0); //@line 370 "modules/graphics/opengl/wrap_Shader.cpp"
}
function __ZN4love8graphics6opengl26w_Shader_getExternVariableEP9lua_State($L) {
 $L = $L|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $arrayelements = 0, $components = 0, $tname = 0;
 var $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $components = sp + 24|0;
 $arrayelements = sp + 20|0;
 $0 = sp + 8|0;
 $tname = sp + 4|0;
 $1 = (__ZN4love14luax_checktypeINS_8graphics6opengl6ShaderEEEPT_P9lua_StateiNS_4TypeE($L,1,18)|0); //@line 39 "modules/graphics/opengl/wrap_Shader.cpp"
 $2 = (_luaL_checklstring($L,2,0)|0); //@line 375 "modules/graphics/opengl/wrap_Shader.cpp"
 HEAP32[$components>>2] = 0; //@line 377 "modules/graphics/opengl/wrap_Shader.cpp"
 HEAP32[$arrayelements>>2] = 0; //@line 378 "modules/graphics/opengl/wrap_Shader.cpp"
 $3 = (_strlen($2)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($0,$2,$3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 $4 = (invoke_iiiii(3611,($1|0),($0|0),($components|0),($arrayelements|0))|0); //@line 381 "modules/graphics/opengl/wrap_Shader.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 381 "modules/graphics/opengl/wrap_Shader.cpp"
  ___resumeException($10|0);
  // unreachable;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 381 "modules/graphics/opengl/wrap_Shader.cpp"
 $7 = HEAP32[$components>>2]|0; //@line 384 "modules/graphics/opengl/wrap_Shader.cpp"
 $8 = ($7|0)>(0); //@line 384 "modules/graphics/opengl/wrap_Shader.cpp"
 if (!($8)) {
  _lua_pushnil($L); //@line 396 "modules/graphics/opengl/wrap_Shader.cpp"
  _lua_pushnil($L); //@line 397 "modules/graphics/opengl/wrap_Shader.cpp"
  _lua_pushnil($L); //@line 398 "modules/graphics/opengl/wrap_Shader.cpp"
  $$1 = 3;
  STACKTOP = sp;return ($$1|0); //@line 402 "modules/graphics/opengl/wrap_Shader.cpp"
 }
 HEAP32[$tname>>2] = 0; //@line 386 "modules/graphics/opengl/wrap_Shader.cpp"
 $9 = (__ZN4love8graphics6opengl6Shader11getConstantENS2_11UniformTypeERPKc($4,$tname)|0); //@line 387 "modules/graphics/opengl/wrap_Shader.cpp"
 if ($9) {
  $12 = HEAP32[$tname>>2]|0; //@line 390 "modules/graphics/opengl/wrap_Shader.cpp"
  _lua_pushstring($L,$12); //@line 390 "modules/graphics/opengl/wrap_Shader.cpp"
  $13 = HEAP32[$components>>2]|0; //@line 391 "modules/graphics/opengl/wrap_Shader.cpp"
  _lua_pushinteger($L,$13); //@line 391 "modules/graphics/opengl/wrap_Shader.cpp"
  $14 = HEAP32[$arrayelements>>2]|0; //@line 392 "modules/graphics/opengl/wrap_Shader.cpp"
  _lua_pushinteger($L,$14); //@line 392 "modules/graphics/opengl/wrap_Shader.cpp"
  $$1 = 3;
  STACKTOP = sp;return ($$1|0); //@line 402 "modules/graphics/opengl/wrap_Shader.cpp"
 } else {
  $15 = (_luaL_error($L,344933,$vararg_buffer)|0); //@line 388 "modules/graphics/opengl/wrap_Shader.cpp"
  $$1 = $15;
  STACKTOP = sp;return ($$1|0); //@line 402 "modules/graphics/opengl/wrap_Shader.cpp"
 }
 return (0)|0;
}
function _luaopen_shader($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 418 "modules/graphics/opengl/wrap_Shader.cpp"
 HEAP32[$vararg_buffer>>2] = 14816; //@line 420 "modules/graphics/opengl/wrap_Shader.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 420 "modules/graphics/opengl/wrap_Shader.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 420 "modules/graphics/opengl/wrap_Shader.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,18,344968,$vararg_buffer)|0); //@line 420 "modules/graphics/opengl/wrap_Shader.cpp"
 STACKTOP = sp;return ($0|0); //@line 420 "modules/graphics/opengl/wrap_Shader.cpp"
}
function __ZN4love14luax_checktypeINS_8graphics6opengl11SpriteBatchEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love8graphics6opengl17w_SpriteBatch_addEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl11SpriteBatchEEEPT_P9lua_StateiNS_4TypeE($L,1,16)|0); //@line 39 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $1 = (__ZN4love8graphics6openglL24w_SpriteBatch_add_or_setEP9lua_StatePNS1_11SpriteBatchEii($L,$0,2,-1)|0); //@line 78 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $2 = (($1) + 1)|0; //@line 79 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 _lua_pushinteger($L,$2); //@line 79 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 return 1; //@line 81 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
}
function __ZN4love8graphics6openglL24w_SpriteBatch_add_or_setEP9lua_StatePNS1_11SpriteBatchEii($L,$t,$startidx,$index) {
 $L = $L|0;
 $t = $t|0;
 $startidx = $startidx|0;
 $index = $index|0;
 var $$0 = 0, $$01 = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0;
 var $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0, $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $7 = 0, $8 = 0, $9 = 0, $quad$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 42 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $0 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,$startidx,13)|0); //@line 46 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 if ($0) {
  $1 = (_lua_touserdata($L,$startidx)|0); //@line 516 "./common/runtime.h"
  $2 = ((($1)) + 4|0); //@line 516 "./common/runtime.h"
  $3 = HEAP32[$2>>2]|0; //@line 516 "./common/runtime.h"
  $4 = (($startidx) + 1)|0; //@line 49 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $$0 = $4;$quad$0 = $3;
 } else {
  $5 = (_lua_type($L,$startidx)|0); //@line 51 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $6 = ($5|0)==(0); //@line 51 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  if ($6) {
   $7 = (($startidx) + 1)|0; //@line 51 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
   $8 = (_lua_type($L,$7)|0); //@line 51 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
   $9 = ($8|0)<(1); //@line 51 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
   if ($9) {
    $$0 = $startidx;$quad$0 = 0;
   } else {
    $10 = (_luax_typerror($L,$startidx,347523)|0); //@line 52 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
    $$01 = $10;
    STACKTOP = sp;return ($$01|0); //@line 72 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
   }
  } else {
   $$0 = $startidx;$quad$0 = 0;
  }
 }
 $11 = (+_luaL_optnumber($L,$$0,0.0)); //@line 54 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $12 = $11; //@line 54 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $13 = (($$0) + 1)|0; //@line 55 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $14 = (+_luaL_optnumber($L,$13,0.0)); //@line 55 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $15 = $14; //@line 55 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $16 = (($$0) + 2)|0; //@line 56 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $17 = (+_luaL_optnumber($L,$16,0.0)); //@line 56 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $18 = $17; //@line 56 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $19 = (($$0) + 3)|0; //@line 57 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $20 = (+_luaL_optnumber($L,$19,1.0)); //@line 57 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $21 = $20; //@line 57 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $22 = (($$0) + 4)|0; //@line 58 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $23 = $21; //@line 58 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $24 = (+_luaL_optnumber($L,$22,$23)); //@line 58 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $25 = $24; //@line 58 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $26 = (($$0) + 5)|0; //@line 59 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $27 = (+_luaL_optnumber($L,$26,0.0)); //@line 59 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $28 = $27; //@line 59 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $29 = (($$0) + 6)|0; //@line 60 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $30 = (+_luaL_optnumber($L,$29,0.0)); //@line 60 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $31 = $30; //@line 60 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $32 = (($$0) + 7)|0; //@line 61 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $33 = (+_luaL_optnumber($L,$32,0.0)); //@line 61 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $34 = $33; //@line 61 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $35 = (($$0) + 8)|0; //@line 62 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $36 = (+_luaL_optnumber($L,$35,0.0)); //@line 62 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $37 = $36; //@line 62 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $38 = ($quad$0|0)==(0|0); //@line 65 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 if ($38) {
  __THREW__ = 0;
  $42 = (invoke_iidddddddddi(3613,($t|0),(+$12),(+$15),(+$18),(+$21),(+$25),(+$28),(+$31),(+$34),(+$37),($index|0))|0); //@line 68 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $43 = __THREW__; __THREW__ = 0;
  $44 = $43&1;
  if (!($44)) {
   $$01 = $42;
   STACKTOP = sp;return ($$01|0); //@line 72 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  }
 } else {
  __THREW__ = 0;
  $39 = (invoke_iiidddddddddi(3612,($t|0),($quad$0|0),(+$12),(+$15),(+$18),(+$21),(+$25),(+$28),(+$31),(+$34),(+$37),($index|0))|0); //@line 66 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $40 = __THREW__; __THREW__ = 0;
  $41 = $40&1;
  if (!($41)) {
   $$01 = $39;
   STACKTOP = sp;return ($$01|0); //@line 72 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  }
 }
 $45 = ___cxa_find_matching_catch_3(8992|0)|0;
 $46 = tempRet0;
 $47 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $48 = ($46|0)==($47|0); //@line 535 "./common/runtime.h"
 if (!($48)) {
  $$01$i = $46;$$02$i = $45;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $49 = (___cxa_begin_catch(($45|0))|0); //@line 535 "./common/runtime.h"
 $50 = HEAP32[$49>>2]|0; //@line 539 "./common/runtime.h"
 $51 = ((($50)) + 8|0); //@line 539 "./common/runtime.h"
 $52 = HEAP32[$51>>2]|0; //@line 539 "./common/runtime.h"
 $53 = (FUNCTION_TABLE_ii[$52 & 8191]($49)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($53|0)); //@line 539 "./common/runtime.h"
 $54 = __THREW__; __THREW__ = 0;
 $55 = $54&1;
 if (!($55)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $56 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $56; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $$01 = $index;
  STACKTOP = sp;return ($$01|0); //@line 72 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 }
 $57 = ___cxa_find_matching_catch_2()|0;
 $58 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $59 = __THREW__; __THREW__ = 0;
 $60 = $59&1;
 if ($60) {
  $61 = ___cxa_find_matching_catch_3(0|0)|0;
  $62 = tempRet0;
  ___clang_call_terminate($61); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $58;$$02$i = $57;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl17w_SpriteBatch_setEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl11SpriteBatchEEEPT_P9lua_StateiNS_4TypeE($L,1,16)|0); //@line 39 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 87 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $2 = (~~(($1))); //@line 87 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $3 = (($2) + -1)|0; //@line 87 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 (__ZN4love8graphics6openglL24w_SpriteBatch_add_or_setEP9lua_StatePNS1_11SpriteBatchEii($L,$0,3,$3)|0); //@line 89 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 return 0; //@line 91 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
}
function __ZN4love8graphics6opengl19w_SpriteBatch_clearEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl11SpriteBatchEEEPT_P9lua_StateiNS_4TypeE($L,1,16)|0); //@line 39 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 __ZN4love8graphics6opengl11SpriteBatch5clearEv($0); //@line 97 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 return 0; //@line 98 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
}
function __ZN4love8graphics6opengl19w_SpriteBatch_flushEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl11SpriteBatchEEEPT_P9lua_StateiNS_4TypeE($L,1,16)|0); //@line 39 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 __ZN4love8graphics6opengl11SpriteBatch5flushEv($0); //@line 104 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 return 0; //@line 105 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
}
function __ZN4love8graphics6opengl24w_SpriteBatch_setTextureEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl11SpriteBatchEEEPT_P9lua_StateiNS_4TypeE($L,1,16)|0); //@line 39 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $1 = (__ZN4love8graphics17luax_checktextureEP9lua_Statei($L,2)|0); //@line 111 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 __ZN4love8graphics6opengl11SpriteBatch10setTextureEPNS0_7TextureE($0,$1); //@line 112 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 return 0; //@line 113 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
}
function __ZN4love8graphics6opengl24w_SpriteBatch_getTextureEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 116 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl11SpriteBatchEEEPT_P9lua_StateiNS_4TypeE($L,1,16)|0); //@line 39 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $1 = (__ZNK4love8graphics6opengl11SpriteBatch10getTextureEv($0)|0); //@line 119 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $2 = ($1|0)==(0|0); //@line 122 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 if ($2) {
  ___cxa_bad_typeid(); //@line 122 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  // unreachable; //@line 122 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 }
 $3 = HEAP32[$1>>2]|0; //@line 122 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $4 = ((($3)) + -4|0); //@line 122 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 122 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $6 = ((($5)) + 4|0); //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 $7 = HEAP32[$6>>2]|0; //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 $8 = HEAP32[(1972)>>2]|0; //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 $9 = ($7|0)==($8|0); //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 if ($9) {
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,12,$1); //@line 123 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 130 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 }
 $10 = HEAP32[(1860)>>2]|0; //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 $11 = ($7|0)==($10|0); //@line 129 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/typeinfo"
 if ($11) {
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,17,$1); //@line 125 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 130 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 } else {
  $12 = (_luaL_error($L,345068,$vararg_buffer)|0); //@line 127 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $$0 = $12;
  STACKTOP = sp;return ($$0|0); //@line 130 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl22w_SpriteBatch_setColorEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $23 = 0, $24 = 0.0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0, $c = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $c = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl11SpriteBatchEEEPT_P9lua_StateiNS_4TypeE($L,1,16)|0); //@line 39 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 HEAP32[$c>>2] = 0; //@line 37 "./modules/graphics/Color.h"
 $1 = (_lua_gettop($L)|0); //@line 137 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $2 = ($1|0)<(2); //@line 137 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 if ($2) {
  __ZN4love8graphics6opengl11SpriteBatch8setColorEv($0); //@line 139 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  STACKTOP = sp;return 0; //@line 165 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 }
 $3 = (_lua_type($L,2)|0); //@line 142 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $4 = ($3|0)==(5); //@line 142 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 if ($4) {
  _lua_rawgeti($L,2,1); //@line 145 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  _lua_rawgeti($L,2,2); //@line 145 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  _lua_rawgeti($L,2,3); //@line 145 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  _lua_rawgeti($L,2,4); //@line 145 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $5 = (+_luaL_checknumber($L,-4)); //@line 147 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $6 = (~~(($5))&255); //@line 147 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  HEAP8[$c>>0] = $6; //@line 147 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $7 = (+_luaL_checknumber($L,-3)); //@line 148 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $8 = (~~(($7))&255); //@line 148 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $9 = ((($c)) + 1|0); //@line 148 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  HEAP8[$9>>0] = $8; //@line 148 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $10 = (+_luaL_checknumber($L,-2)); //@line 149 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $11 = (~~(($10))&255); //@line 149 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $12 = ((($c)) + 2|0); //@line 149 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  HEAP8[$12>>0] = $11; //@line 149 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $13 = (+_luaL_optnumber($L,-1,255.0)); //@line 150 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $14 = (~~(($13))&255); //@line 150 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $15 = ((($c)) + 3|0); //@line 150 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  HEAP8[$15>>0] = $14; //@line 150 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  _lua_settop($L,-5); //@line 152 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 } else {
  $16 = (+_luaL_checknumber($L,2)); //@line 156 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $17 = (~~(($16))&255); //@line 156 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  HEAP8[$c>>0] = $17; //@line 156 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $18 = (+_luaL_checknumber($L,3)); //@line 157 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $19 = (~~(($18))&255); //@line 157 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $20 = ((($c)) + 1|0); //@line 157 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  HEAP8[$20>>0] = $19; //@line 157 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $21 = (+_luaL_checknumber($L,4)); //@line 158 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $22 = (~~(($21))&255); //@line 158 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $23 = ((($c)) + 2|0); //@line 158 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  HEAP8[$23>>0] = $22; //@line 158 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $24 = (+_luaL_optnumber($L,5,255.0)); //@line 159 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $25 = (~~(($24))&255); //@line 159 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  $26 = ((($c)) + 3|0); //@line 159 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  HEAP8[$26>>0] = $25; //@line 159 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 }
 __ZN4love8graphics6opengl11SpriteBatch8setColorERKNS0_6ColorTIhEE($0,$c); //@line 162 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 STACKTOP = sp;return 0; //@line 165 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
}
function __ZN4love8graphics6opengl22w_SpriteBatch_getColorEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl11SpriteBatchEEEPT_P9lua_StateiNS_4TypeE($L,1,16)|0); //@line 39 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $1 = (__ZNK4love8graphics6opengl11SpriteBatch8getColorEv($0)|0); //@line 170 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $2 = ($1|0)==(0|0); //@line 173 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 if ($2) {
  $$0 = 0;
  return ($$0|0); //@line 182 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 }
 $3 = HEAP8[$1>>0]|0; //@line 176 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $4 = (+($3&255)); //@line 176 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 _lua_pushnumber($L,$4); //@line 176 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $5 = ((($1)) + 1|0); //@line 177 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $6 = HEAP8[$5>>0]|0; //@line 177 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $7 = (+($6&255)); //@line 177 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 _lua_pushnumber($L,$7); //@line 177 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $8 = ((($1)) + 2|0); //@line 178 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 178 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $10 = (+($9&255)); //@line 178 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 _lua_pushnumber($L,$10); //@line 178 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $11 = ((($1)) + 3|0); //@line 179 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $12 = HEAP8[$11>>0]|0; //@line 179 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $13 = (+($12&255)); //@line 179 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 _lua_pushnumber($L,$13); //@line 179 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $$0 = 4;
 return ($$0|0); //@line 182 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
}
function __ZN4love8graphics6opengl22w_SpriteBatch_getCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl11SpriteBatchEEEPT_P9lua_StateiNS_4TypeE($L,1,16)|0); //@line 39 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $1 = (__ZNK4love8graphics6opengl11SpriteBatch8getCountEv($0)|0); //@line 187 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 _lua_pushinteger($L,$1); //@line 187 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 return 1; //@line 188 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
}
function __ZN4love8graphics6opengl27w_SpriteBatch_setBufferSizeEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 191 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl11SpriteBatchEEEPT_P9lua_StateiNS_4TypeE($L,1,16)|0); //@line 39 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 194 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $2 = (~~(($1))); //@line 194 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 __THREW__ = 0;
 invoke_vii(3614,($0|0),($2|0)); //@line 195 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  STACKTOP = sp;return 0; //@line 196 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 }
 $5 = ___cxa_find_matching_catch_3(8992|0)|0;
 $6 = tempRet0;
 $7 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $8 = ($6|0)==($7|0); //@line 535 "./common/runtime.h"
 if (!($8)) {
  $$01$i = $6;$$02$i = $5;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $9 = (___cxa_begin_catch(($5|0))|0); //@line 535 "./common/runtime.h"
 $10 = HEAP32[$9>>2]|0; //@line 539 "./common/runtime.h"
 $11 = ((($10)) + 8|0); //@line 539 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = (FUNCTION_TABLE_ii[$12 & 8191]($9)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($13|0)); //@line 539 "./common/runtime.h"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if (!($15)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $16 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $16; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 196 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 }
 $17 = ___cxa_find_matching_catch_2()|0;
 $18 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $21 = ___cxa_find_matching_catch_3(0|0)|0;
  $22 = tempRet0;
  ___clang_call_terminate($21); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $18;$$02$i = $17;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl27w_SpriteBatch_getBufferSizeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl11SpriteBatchEEEPT_P9lua_StateiNS_4TypeE($L,1,16)|0); //@line 39 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $1 = (__ZNK4love8graphics6opengl11SpriteBatch13getBufferSizeEv($0)|0); //@line 202 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 _lua_pushinteger($L,$1); //@line 202 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 return 1; //@line 203 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
}
function __ZN4love8graphics6opengl29w_SpriteBatch_attachAttributeEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 $1 = (__ZN4love14luax_checktypeINS_8graphics6opengl11SpriteBatchEEEPT_P9lua_StateiNS_4TypeE($L,1,16)|0); //@line 39 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $2 = (_luaL_checklstring($L,2,0)|0); //@line 209 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $3 = (__ZN4love14luax_checktypeINS_8graphics6opengl4MeshEEEPT_P9lua_StateiNS_4TypeE($L,3,19)|0); //@line 210 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $4 = (_strlen($2)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viii(3460,($0|0),($2|0),($4|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 do {
  if ($6) {
   $11 = ___cxa_find_matching_catch_3(8992|0)|0;
   $12 = tempRet0;
   $eh$lpad$body$i$index2Z2D = $12;$eh$lpad$body$i$indexZ2D = $11;
  } else {
   __THREW__ = 0;
   invoke_viii(3615,($1|0),($0|0),($3|0)); //@line 212 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if ($8) {
    $9 = ___cxa_find_matching_catch_3(8992|0)|0;
    $10 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 212 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
    $eh$lpad$body$i$index2Z2D = $10;$eh$lpad$body$i$indexZ2D = $9;
    break;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 212 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
   STACKTOP = sp;return 0; //@line 213 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
  }
 } while(0);
 $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $14 = ($eh$lpad$body$i$index2Z2D|0)==($13|0); //@line 535 "./common/runtime.h"
 if (!($14)) {
  $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $15 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
 $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
 $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
 $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
 $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if (!($21)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 213 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 }
 $23 = ___cxa_find_matching_catch_2()|0;
 $24 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if ($26) {
  $27 = ___cxa_find_matching_catch_3(0|0)|0;
  $28 = tempRet0;
  ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $24;$$02$i = $23;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function _luaopen_spritebatch($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 233 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 HEAP32[$vararg_buffer>>2] = 14896; //@line 235 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 235 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 235 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,16,345102,$vararg_buffer)|0); //@line 235 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
 STACKTOP = sp;return ($0|0); //@line 235 "modules/graphics/opengl/wrap_SpriteBatch.cpp"
}
function __ZN4love14luax_checktypeINS_8graphics6opengl4TextEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love8graphics6opengl23luax_checkcoloredstringEP9lua_StateiRNSt3__16vectorINS1_4Font13ColoredStringENS4_9allocatorIS7_EEEE($L,$idx,$strings) {
 $L = $L|0;
 $idx = $idx|0;
 $strings = $strings|0;
 var $$pre$phi18Z2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0, $36 = 0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $coloredstr = 0, $i$013 = 0, $lpad$phi$index = 0, $lpad$phi$index3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $coloredstr = sp;
 ;HEAP32[$coloredstr>>2]=0|0;HEAP32[$coloredstr+4>>2]=0|0;HEAP32[$coloredstr+8>>2]=0|0; //@line 37 "./modules/graphics/Color.h"
 $0 = ((($coloredstr)) + 12|0); //@line 38 "modules/graphics/opengl/wrap_Text.cpp"
 HEAP32[$0>>2] = -1; //@line 38 "modules/graphics/opengl/wrap_Text.cpp"
 __THREW__ = 0;
 $1 = (invoke_iii(3358,($L|0),($idx|0))|0); //@line 40 "modules/graphics/opengl/wrap_Text.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 do {
  if (!($3)) {
   $4 = ($1|0)==(5); //@line 40 "modules/graphics/opengl/wrap_Text.cpp"
   if (!($4)) {
    __THREW__ = 0;
    $67 = (invoke_iiii(3359,($L|0),($idx|0),(0|0))|0); //@line 71 "modules/graphics/opengl/wrap_Text.cpp"
    $68 = __THREW__; __THREW__ = 0;
    $69 = $68&1;
    if ($69) {
     break;
    }
    __THREW__ = 0;
    (invoke_iii(3360,($coloredstr|0),($67|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $70 = __THREW__; __THREW__ = 0;
    $71 = $70&1;
    if ($71) {
     break;
    }
    $72 = ((($strings)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $73 = HEAP32[$72>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $74 = ((($strings)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $75 = HEAP32[$74>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $76 = ($73|0)==($75|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($76) {
     __THREW__ = 0;
     invoke_vii(3616,($strings|0),($coloredstr|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $84 = __THREW__; __THREW__ = 0;
     $85 = $84&1;
     if ($85) {
      break;
     } else {
      $$pre$phi18Z2D = $coloredstr;
     }
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre$phi18Z2D); //@line 64 "modules/graphics/opengl/Font.h"
     STACKTOP = sp;return; //@line 74 "modules/graphics/opengl/wrap_Text.cpp"
    }
    __THREW__ = 0;
    invoke_vii(3351,($73|0),($coloredstr|0)); //@line 64 "modules/graphics/opengl/Font.h"
    $77 = __THREW__; __THREW__ = 0;
    $78 = $77&1;
    if ($78) {
     break;
    }
    $79 = ((($73)) + 12|0); //@line 64 "modules/graphics/opengl/Font.h"
    $80 = ((($coloredstr)) + 12|0); //@line 64 "modules/graphics/opengl/Font.h"
    $81 = HEAP32[$80>>2]|0; //@line 64 "modules/graphics/opengl/Font.h"
    HEAP32[$79>>2] = $81; //@line 64 "modules/graphics/opengl/Font.h"
    $82 = HEAP32[$72>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $83 = ((($82)) + 16|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$72>>2] = $83; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre$phi18Z2D = $coloredstr;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre$phi18Z2D); //@line 64 "modules/graphics/opengl/Font.h"
    STACKTOP = sp;return; //@line 74 "modules/graphics/opengl/wrap_Text.cpp"
   }
   __THREW__ = 0;
   $5 = (invoke_iii(3381,($L|0),($idx|0))|0); //@line 42 "modules/graphics/opengl/wrap_Text.cpp"
   $6 = __THREW__; __THREW__ = 0;
   $7 = $6&1;
   if (!($7)) {
    $8 = ($5|0)<(1); //@line 44 "modules/graphics/opengl/wrap_Text.cpp"
    if ($8) {
     $$pre$phi18Z2D = $coloredstr;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre$phi18Z2D); //@line 64 "modules/graphics/opengl/Font.h"
     STACKTOP = sp;return; //@line 74 "modules/graphics/opengl/wrap_Text.cpp"
    }
    $9 = ((($strings)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $10 = ((($strings)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $11 = ((($coloredstr)) + 12|0); //@line 64 "modules/graphics/opengl/Font.h"
    $12 = ((($coloredstr)) + 13|0); //@line 54 "modules/graphics/opengl/wrap_Text.cpp"
    $13 = ((($coloredstr)) + 14|0); //@line 55 "modules/graphics/opengl/wrap_Text.cpp"
    $14 = ((($coloredstr)) + 15|0); //@line 56 "modules/graphics/opengl/wrap_Text.cpp"
    $i$013 = 1;
    L18: while(1) {
     __THREW__ = 0;
     invoke_viii(3382,($L|0),($idx|0),($i$013|0)); //@line 46 "modules/graphics/opengl/wrap_Text.cpp"
     $21 = __THREW__; __THREW__ = 0;
     $22 = $21&1;
     if ($22) {
      label = 7;
      break;
     }
     __THREW__ = 0;
     $23 = (invoke_iii(3358,($L|0),-1)|0); //@line 48 "modules/graphics/opengl/wrap_Text.cpp"
     $24 = __THREW__; __THREW__ = 0;
     $25 = $24&1;
     if ($25) {
      label = 7;
      break;
     }
     $26 = ($23|0)==(5); //@line 48 "modules/graphics/opengl/wrap_Text.cpp"
     do {
      if ($26) {
       __THREW__ = 0;
       invoke_viii(3382,($L|0),-1,1); //@line 51 "modules/graphics/opengl/wrap_Text.cpp"
       $27 = __THREW__; __THREW__ = 0;
       $28 = $27&1;
       if ($28) {
        label = 6;
        break L18;
       }
       __THREW__ = 0;
       invoke_viii(3382,($L|0),-2,2); //@line 51 "modules/graphics/opengl/wrap_Text.cpp"
       $29 = __THREW__; __THREW__ = 0;
       $30 = $29&1;
       if ($30) {
        label = 6;
        break L18;
       }
       __THREW__ = 0;
       invoke_viii(3382,($L|0),-3,3); //@line 51 "modules/graphics/opengl/wrap_Text.cpp"
       $86 = __THREW__; __THREW__ = 0;
       $87 = $86&1;
       if ($87) {
        label = 6;
        break L18;
       }
       __THREW__ = 0;
       invoke_viii(3382,($L|0),-4,4); //@line 51 "modules/graphics/opengl/wrap_Text.cpp"
       $88 = __THREW__; __THREW__ = 0;
       $89 = $88&1;
       if ($89) {
        label = 6;
        break L18;
       }
       __THREW__ = 0;
       $32 = (+invoke_dii(3361,($L|0),-4)); //@line 53 "modules/graphics/opengl/wrap_Text.cpp"
       $90 = __THREW__; __THREW__ = 0;
       $91 = $90&1;
       if ($91) {
        label = 7;
        break L18;
       }
       $31 = (~~(($32))&255); //@line 53 "modules/graphics/opengl/wrap_Text.cpp"
       HEAP8[$0>>0] = $31; //@line 53 "modules/graphics/opengl/wrap_Text.cpp"
       __THREW__ = 0;
       $33 = (+invoke_dii(3361,($L|0),-3)); //@line 54 "modules/graphics/opengl/wrap_Text.cpp"
       $34 = __THREW__; __THREW__ = 0;
       $35 = $34&1;
       if ($35) {
        label = 7;
        break L18;
       }
       $36 = (~~(($33))&255); //@line 54 "modules/graphics/opengl/wrap_Text.cpp"
       HEAP8[$12>>0] = $36; //@line 54 "modules/graphics/opengl/wrap_Text.cpp"
       __THREW__ = 0;
       $37 = (+invoke_dii(3361,($L|0),-2)); //@line 55 "modules/graphics/opengl/wrap_Text.cpp"
       $38 = __THREW__; __THREW__ = 0;
       $39 = $38&1;
       if ($39) {
        label = 7;
        break L18;
       }
       $40 = (~~(($37))&255); //@line 55 "modules/graphics/opengl/wrap_Text.cpp"
       HEAP8[$13>>0] = $40; //@line 55 "modules/graphics/opengl/wrap_Text.cpp"
       __THREW__ = 0;
       $41 = (+invoke_diid(3525,($L|0),-1,255.0)); //@line 56 "modules/graphics/opengl/wrap_Text.cpp"
       $42 = __THREW__; __THREW__ = 0;
       $43 = $42&1;
       if ($43) {
        label = 7;
        break L18;
       }
       $44 = (~~(($41))&255); //@line 56 "modules/graphics/opengl/wrap_Text.cpp"
       HEAP8[$14>>0] = $44; //@line 56 "modules/graphics/opengl/wrap_Text.cpp"
       __THREW__ = 0;
       invoke_vii(3384,($L|0),-5); //@line 58 "modules/graphics/opengl/wrap_Text.cpp"
       $45 = __THREW__; __THREW__ = 0;
       $46 = $45&1;
       if ($46) {
        label = 7;
        break L18;
       }
      } else {
       __THREW__ = 0;
       $47 = (invoke_iiii(3359,($L|0),-1,(0|0))|0); //@line 62 "modules/graphics/opengl/wrap_Text.cpp"
       $48 = __THREW__; __THREW__ = 0;
       $49 = $48&1;
       if ($49) {
        label = 7;
        break L18;
       }
       __THREW__ = 0;
       (invoke_iii(3360,($coloredstr|0),($47|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $50 = __THREW__; __THREW__ = 0;
       $51 = $50&1;
       if ($51) {
        label = 7;
        break L18;
       }
       $52 = HEAP32[$9>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $53 = HEAP32[$10>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $54 = ($52|0)==($53|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($54) {
        __THREW__ = 0;
        invoke_vii(3616,($strings|0),($coloredstr|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $61 = __THREW__; __THREW__ = 0;
        $62 = $61&1;
        if ($62) {
         label = 7;
         break L18;
        } else {
         break;
        }
       }
       __THREW__ = 0;
       invoke_vii(3351,($52|0),($coloredstr|0)); //@line 64 "modules/graphics/opengl/Font.h"
       $55 = __THREW__; __THREW__ = 0;
       $56 = $55&1;
       if ($56) {
        label = 7;
        break L18;
       }
       $57 = ((($52)) + 12|0); //@line 64 "modules/graphics/opengl/Font.h"
       $58 = HEAP32[$11>>2]|0; //@line 64 "modules/graphics/opengl/Font.h"
       HEAP32[$57>>2] = $58; //@line 64 "modules/graphics/opengl/Font.h"
       $59 = HEAP32[$9>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $60 = ((($59)) + 16|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$9>>2] = $60; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } while(0);
     __THREW__ = 0;
     invoke_vii(3384,($L|0),-2); //@line 66 "modules/graphics/opengl/wrap_Text.cpp"
     $63 = __THREW__; __THREW__ = 0;
     $64 = $63&1;
     if ($64) {
      label = 7;
      break;
     }
     $65 = (($i$013) + 1)|0; //@line 44 "modules/graphics/opengl/wrap_Text.cpp"
     $66 = ($i$013|0)<($5|0); //@line 44 "modules/graphics/opengl/wrap_Text.cpp"
     if ($66) {
      $i$013 = $65;
     } else {
      $$pre$phi18Z2D = $coloredstr;
      label = 33;
      break;
     }
    }
    if ((label|0) == 6) {
     $15 = ___cxa_find_matching_catch_2()|0;
     $16 = tempRet0;
     $$pre$phiZ2D = $coloredstr;$lpad$phi$index = $15;$lpad$phi$index3 = $16;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre$phiZ2D); //@line 64 "modules/graphics/opengl/Font.h"
     ___resumeException($lpad$phi$index|0);
     // unreachable;
    }
    else if ((label|0) == 7) {
     $17 = ___cxa_find_matching_catch_2()|0;
     $18 = tempRet0;
     $$pre$phiZ2D = $coloredstr;$lpad$phi$index = $17;$lpad$phi$index3 = $18;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre$phiZ2D); //@line 64 "modules/graphics/opengl/Font.h"
     ___resumeException($lpad$phi$index|0);
     // unreachable;
    }
    else if ((label|0) == 33) {
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre$phi18Z2D); //@line 64 "modules/graphics/opengl/Font.h"
     STACKTOP = sp;return; //@line 74 "modules/graphics/opengl/wrap_Text.cpp"
    }
   }
  }
 } while(0);
 $19 = ___cxa_find_matching_catch_2()|0;
 $20 = tempRet0;
 $$pre$phiZ2D = $coloredstr;$lpad$phi$index = $19;$lpad$phi$index3 = $20;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre$phiZ2D); //@line 64 "modules/graphics/opengl/Font.h"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love8graphics6opengl4Font13ColoredStringENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i4 = 0, $$0$i5 = 0, $$01$i$i = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $7 = 0, $8 = 0, $9 = 0, $__v$sroa$6$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(268435455); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $10 = $$pre;
 } else {
  $10 = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($10))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $9 >> 4; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($11>>>0)<(134217727); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $16 = $9 >> 3; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = ($16>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $18 = $17 ? $5 : $16; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = (($19) - ($10))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = $20 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = ($18|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($22) {
   $$0$i5 = 0;$26 = 0;$27 = $21;
  } else {
   $$0$i4 = $18;$65 = $21;
   label = 6;
  }
 } else {
  $13 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) - ($10))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = $14 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i4 = 268435455;$65 = $15;
  label = 6;
 }
 if ((label|0) == 6) {
  $23 = $$0$i4 << 4; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = (__Znwj($23)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i5 = $$0$i4;$26 = $24;$27 = $65;
 }
 $25 = (($26) + ($27<<4)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $25; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = (($26) + ($$0$i5<<4)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = $29; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 __THREW__ = 0;
 invoke_vii(3351,($25|0),($__x|0)); //@line 64 "modules/graphics/opengl/Font.h"
 $31 = __THREW__; __THREW__ = 0;
 $32 = $31&1;
 if ($32) {
  $62 = ___cxa_find_matching_catch_2()|0;
  $63 = tempRet0;
  $64 = ($26|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($64) {
   ___resumeException($62|0);
   // unreachable;
  }
  __ZdlPv($26); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($62|0);
  // unreachable;
 }
 $33 = (((($26) + ($27<<4)|0)) + 12|0); //@line 64 "modules/graphics/opengl/Font.h"
 $34 = ((($__x)) + 12|0); //@line 64 "modules/graphics/opengl/Font.h"
 $35 = HEAP32[$34>>2]|0; //@line 64 "modules/graphics/opengl/Font.h"
 HEAP32[$33>>2] = $35; //@line 64 "modules/graphics/opengl/Font.h"
 $36 = ((($25)) + 16|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $37 = $36; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $38 = HEAP32[$this>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $39 = HEAP32[$0>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $40 = ($39|0)==($38|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($40) {
  $41 = $38; //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $52 = $28;$54 = $41;$56 = $38;
 } else {
  $$01$i$i = $39;$43 = $25;$__v$sroa$6$0 = $28;
  while(1) {
   $42 = ((($43)) + -16|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $44 = ((($$01$i$i)) + -16|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   ;HEAP32[$42>>2]=HEAP32[$44>>2]|0;HEAP32[$42+4>>2]=HEAP32[$44+4>>2]|0;HEAP32[$42+8>>2]=HEAP32[$44+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   ;HEAP32[$44>>2]=0|0;HEAP32[$44+4>>2]=0|0;HEAP32[$44+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $45 = ((($43)) + -4|0); //@line 64 "modules/graphics/opengl/Font.h"
   $46 = ((($$01$i$i)) + -4|0); //@line 64 "modules/graphics/opengl/Font.h"
   $47 = HEAP32[$46>>2]|0; //@line 64 "modules/graphics/opengl/Font.h"
   HEAP32[$45>>2] = $47; //@line 64 "modules/graphics/opengl/Font.h"
   $48 = $__v$sroa$6$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $49 = ((($48)) + -16|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $50 = $49; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $51 = ($44|0)==($38|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if ($51) {
    $$lcssa = $50;
    break;
   } else {
    $$01$i$i = $44;$43 = $49;$__v$sroa$6$0 = $50;
   }
  }
  $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $$pre6 = HEAP32[$0>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $52 = $$lcssa;$54 = $$pre$i;$56 = $$pre6;
 }
 HEAP32[$this>>2] = $52; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $53 = $54; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $55 = ($56|0)==($53|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if (!($55)) {
  $58 = $56;
  while(1) {
   $57 = ((($58)) + -16|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($57); //@line 64 "modules/graphics/opengl/Font.h"
   $59 = ($57|0)==($53|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($59) {
    break;
   } else {
    $58 = $57;
   }
  }
 }
 $60 = ($54|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($60) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $61 = $54; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($61); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl10w_Text_setEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$01$i17 = 0, $$01$i7 = 0, $$02 = 0, $$02$i = 0, $$02$i16 = 0, $$02$i6 = 0, $$03 = 0, $$1 = 0, $$pre$i = 0, $$pre$i12 = 0, $$pre$i25 = 0, $$pre$i31 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0.0, $76 = 0.0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $align = 0, $eh$lpad$body$index11Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body22$index16Z2D = 0, $eh$lpad$body22$indexZ2D = 0, $newtext = 0, $newtext1 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $newtext = sp + 44|0;
 $align = sp + 40|0;
 $newtext1 = sp + 28|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4TextEEEPT_P9lua_StateiNS_4TypeE($L,1,20)|0); //@line 32 "modules/graphics/opengl/wrap_Text.cpp"
 $1 = (_lua_type($L,2)|0); //@line 80 "modules/graphics/opengl/wrap_Text.cpp"
 $2 = ($1|0)<(1); //@line 80 "modules/graphics/opengl/wrap_Text.cpp"
 if ($2) {
  __THREW__ = 0;
  invoke_vi(3617,($0|0)); //@line 83 "modules/graphics/opengl/wrap_Text.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if (!($4)) {
   $$1 = 0;
   STACKTOP = sp;return ($$1|0); //@line 109 "modules/graphics/opengl/wrap_Text.cpp"
  }
  $5 = ___cxa_find_matching_catch_3(8992|0)|0;
  $6 = tempRet0;
  $7 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
  $8 = ($6|0)==($7|0); //@line 535 "./common/runtime.h"
  if (!($8)) {
   $$01$i = $6;$$02$i = $5;
   ___resumeException($$02$i|0);
   // unreachable;
  }
  $9 = (___cxa_begin_catch(($5|0))|0); //@line 535 "./common/runtime.h"
  $10 = HEAP32[$9>>2]|0; //@line 539 "./common/runtime.h"
  $11 = ((($10)) + 8|0); //@line 539 "./common/runtime.h"
  $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
  $13 = (FUNCTION_TABLE_ii[$12 & 8191]($9)|0); //@line 539 "./common/runtime.h"
  __THREW__ = 0;
  invoke_vii(3380,($L|0),($13|0)); //@line 539 "./common/runtime.h"
  $14 = __THREW__; __THREW__ = 0;
  $15 = $14&1;
  if (!($15)) {
   ___cxa_end_catch(); //@line 540 "./common/runtime.h"
   $16 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
   HEAP32[$vararg_buffer>>2] = $16; //@line 543 "./common/runtime.h"
   (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
   $$1 = 0;
   STACKTOP = sp;return ($$1|0); //@line 109 "modules/graphics/opengl/wrap_Text.cpp"
  }
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  __THREW__ = 0;
  invoke_v(3378); //@line 540 "./common/runtime.h"
  $19 = __THREW__; __THREW__ = 0;
  $20 = $19&1;
  if ($20) {
   $21 = ___cxa_find_matching_catch_3(0|0)|0;
   $22 = tempRet0;
   ___clang_call_terminate($21); //@line 540 "./common/runtime.h"
   // unreachable; //@line 540 "./common/runtime.h"
  } else {
   $$01$i = $18;$$02$i = $17;
   ___resumeException($$02$i|0);
   // unreachable;
  }
 }
 $23 = (_lua_type($L,3)|0); //@line 85 "modules/graphics/opengl/wrap_Text.cpp"
 $24 = ($23|0)<(1); //@line 85 "modules/graphics/opengl/wrap_Text.cpp"
 if ($24) {
  HEAP32[$newtext>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $25 = ((($newtext)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$25>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $26 = ((($newtext)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$26>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  __THREW__ = 0;
  invoke_viii(3568,($L|0),2,($newtext|0)); //@line 89 "modules/graphics/opengl/wrap_Text.cpp"
  $27 = __THREW__; __THREW__ = 0;
  $28 = $27&1;
  L19: do {
   if ($28) {
    label = 26;
   } else {
    __THREW__ = 0;
    invoke_vii(3618,($0|0),($newtext|0)); //@line 90 "modules/graphics/opengl/wrap_Text.cpp"
    $29 = __THREW__; __THREW__ = 0;
    $30 = $29&1;
    L21: do {
     if ($30) {
      $31 = ___cxa_find_matching_catch_3(8992|0)|0;
      $32 = tempRet0;
      $33 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
      $34 = ($32|0)==($33|0); //@line 535 "./common/runtime.h"
      do {
       if ($34) {
        $35 = (___cxa_begin_catch(($31|0))|0); //@line 535 "./common/runtime.h"
        $36 = HEAP32[$35>>2]|0; //@line 539 "./common/runtime.h"
        $37 = ((($36)) + 8|0); //@line 539 "./common/runtime.h"
        $38 = HEAP32[$37>>2]|0; //@line 539 "./common/runtime.h"
        $39 = (FUNCTION_TABLE_ii[$38 & 8191]($35)|0); //@line 539 "./common/runtime.h"
        __THREW__ = 0;
        invoke_vii(3380,($L|0),($39|0)); //@line 539 "./common/runtime.h"
        $40 = __THREW__; __THREW__ = 0;
        $41 = $40&1;
        if ($41) {
         $49 = ___cxa_find_matching_catch_2()|0;
         $50 = tempRet0;
         __THREW__ = 0;
         invoke_v(3378); //@line 540 "./common/runtime.h"
         $51 = __THREW__; __THREW__ = 0;
         $52 = $51&1;
         if (!($52)) {
          $$01$i7 = $50;$$02$i6 = $49;
          break;
         }
         $53 = ___cxa_find_matching_catch_3(0|0)|0;
         $54 = tempRet0;
         ___clang_call_terminate($53); //@line 540 "./common/runtime.h"
         // unreachable; //@line 540 "./common/runtime.h"
        }
        __THREW__ = 0;
        invoke_v(3378); //@line 540 "./common/runtime.h"
        $42 = __THREW__; __THREW__ = 0;
        $43 = $42&1;
        if ($43) {
         label = 26;
         break L19;
        }
        __THREW__ = 0;
        $44 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
        $45 = __THREW__; __THREW__ = 0;
        $46 = $45&1;
        if ($46) {
         label = 26;
         break L19;
        }
        __THREW__ = 0;
        HEAP32[$vararg_buffer1>>2] = $44; //@line 543 "./common/runtime.h"
        (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer1|0))|0); //@line 543 "./common/runtime.h"
        $47 = __THREW__; __THREW__ = 0;
        $48 = $47&1;
        if ($48) {
         label = 26;
         break L19;
        } else {
         break L21;
        }
       } else {
        $$01$i7 = $32;$$02$i6 = $31;
       }
      } while(0);
      $eh$lpad$body$index11Z2D = $$01$i7;$eh$lpad$body$indexZ2D = $$02$i6;
      break L19;
     }
    } while(0);
    $55 = HEAP32[$newtext>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $56 = ($55|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($56)) {
     $57 = HEAP32[$25>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $58 = ($57|0)==($55|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($58) {
      $63 = $55;
     } else {
      $60 = $57;
      while(1) {
       $59 = ((($60)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$25>>2] = $59; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($59); //@line 64 "modules/graphics/opengl/Font.h"
       $61 = HEAP32[$25>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $62 = ($61|0)==($55|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($62) {
        break;
       } else {
        $60 = $61;
       }
      }
      $$pre$i = HEAP32[$newtext>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $63 = $$pre$i;
     }
     __ZdlPv($63); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    }
    $$1 = 0;
    STACKTOP = sp;return ($$1|0); //@line 109 "modules/graphics/opengl/wrap_Text.cpp"
   }
  } while(0);
  if ((label|0) == 26) {
   $64 = ___cxa_find_matching_catch_2()|0;
   $65 = tempRet0;
   $eh$lpad$body$index11Z2D = $65;$eh$lpad$body$indexZ2D = $64;
  }
  $66 = HEAP32[$newtext>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $67 = ($66|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($67) {
   $$02 = $eh$lpad$body$index11Z2D;$$03 = $eh$lpad$body$indexZ2D;
   ___resumeException($$03|0);
   // unreachable;
  }
  $68 = HEAP32[$25>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $69 = ($68|0)==($66|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($69) {
   $74 = $66;
  } else {
   $71 = $68;
   while(1) {
    $70 = ((($71)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$25>>2] = $70; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($70); //@line 64 "modules/graphics/opengl/Font.h"
    $72 = HEAP32[$25>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $73 = ($72|0)==($66|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($73) {
     break;
    } else {
     $71 = $72;
    }
   }
   $$pre$i12 = HEAP32[$newtext>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $74 = $$pre$i12;
  }
  __ZdlPv($74); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$02 = $eh$lpad$body$index11Z2D;$$03 = $eh$lpad$body$indexZ2D;
  ___resumeException($$03|0);
  // unreachable;
 }
 $75 = (+_luaL_checknumber($L,3)); //@line 95 "modules/graphics/opengl/wrap_Text.cpp"
 $76 = $75; //@line 95 "modules/graphics/opengl/wrap_Text.cpp"
 $77 = (_luaL_checklstring($L,4,0)|0); //@line 98 "modules/graphics/opengl/wrap_Text.cpp"
 $78 = (__ZN4love8graphics6opengl4Font11getConstantEPKcRNS2_9AlignModeE($77,$align)|0); //@line 99 "modules/graphics/opengl/wrap_Text.cpp"
 if (!($78)) {
  HEAP32[$vararg_buffer7>>2] = $77; //@line 100 "modules/graphics/opengl/wrap_Text.cpp"
  $119 = (_luaL_error($L,345198,$vararg_buffer7)|0); //@line 100 "modules/graphics/opengl/wrap_Text.cpp"
  $$1 = $119;
  STACKTOP = sp;return ($$1|0); //@line 109 "modules/graphics/opengl/wrap_Text.cpp"
 }
 HEAP32[$newtext1>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $79 = ((($newtext1)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$79>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $80 = ((($newtext1)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$80>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 invoke_viii(3568,($L|0),2,($newtext1|0)); //@line 103 "modules/graphics/opengl/wrap_Text.cpp"
 $81 = __THREW__; __THREW__ = 0;
 $82 = $81&1;
 L59: do {
  if ($82) {
   label = 50;
  } else {
   $83 = HEAP32[$align>>2]|0; //@line 105 "modules/graphics/opengl/wrap_Text.cpp"
   __THREW__ = 0;
   invoke_viidi(3619,($0|0),($newtext1|0),(+$76),($83|0)); //@line 105 "modules/graphics/opengl/wrap_Text.cpp"
   $84 = __THREW__; __THREW__ = 0;
   $85 = $84&1;
   L61: do {
    if ($85) {
     $86 = ___cxa_find_matching_catch_3(8992|0)|0;
     $87 = tempRet0;
     $88 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
     $89 = ($87|0)==($88|0); //@line 535 "./common/runtime.h"
     do {
      if ($89) {
       $90 = (___cxa_begin_catch(($86|0))|0); //@line 535 "./common/runtime.h"
       $91 = HEAP32[$90>>2]|0; //@line 539 "./common/runtime.h"
       $92 = ((($91)) + 8|0); //@line 539 "./common/runtime.h"
       $93 = HEAP32[$92>>2]|0; //@line 539 "./common/runtime.h"
       $94 = (FUNCTION_TABLE_ii[$93 & 8191]($90)|0); //@line 539 "./common/runtime.h"
       __THREW__ = 0;
       invoke_vii(3380,($L|0),($94|0)); //@line 539 "./common/runtime.h"
       $95 = __THREW__; __THREW__ = 0;
       $96 = $95&1;
       if ($96) {
        $104 = ___cxa_find_matching_catch_2()|0;
        $105 = tempRet0;
        __THREW__ = 0;
        invoke_v(3378); //@line 540 "./common/runtime.h"
        $106 = __THREW__; __THREW__ = 0;
        $107 = $106&1;
        if (!($107)) {
         $$01$i17 = $105;$$02$i16 = $104;
         break;
        }
        $108 = ___cxa_find_matching_catch_3(0|0)|0;
        $109 = tempRet0;
        ___clang_call_terminate($108); //@line 540 "./common/runtime.h"
        // unreachable; //@line 540 "./common/runtime.h"
       }
       __THREW__ = 0;
       invoke_v(3378); //@line 540 "./common/runtime.h"
       $97 = __THREW__; __THREW__ = 0;
       $98 = $97&1;
       if ($98) {
        label = 50;
        break L59;
       }
       __THREW__ = 0;
       $99 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
       $100 = __THREW__; __THREW__ = 0;
       $101 = $100&1;
       if ($101) {
        label = 50;
        break L59;
       }
       __THREW__ = 0;
       HEAP32[$vararg_buffer4>>2] = $99; //@line 543 "./common/runtime.h"
       (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer4|0))|0); //@line 543 "./common/runtime.h"
       $102 = __THREW__; __THREW__ = 0;
       $103 = $102&1;
       if ($103) {
        label = 50;
        break L59;
       } else {
        break L61;
       }
      } else {
       $$01$i17 = $87;$$02$i16 = $86;
      }
     } while(0);
     $eh$lpad$body22$index16Z2D = $$01$i17;$eh$lpad$body22$indexZ2D = $$02$i16;
     break L59;
    }
   } while(0);
   $110 = HEAP32[$newtext1>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $111 = ($110|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($111)) {
    $112 = HEAP32[$79>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $113 = ($112|0)==($110|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($113) {
     $118 = $110;
    } else {
     $115 = $112;
     while(1) {
      $114 = ((($115)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$79>>2] = $114; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($114); //@line 64 "modules/graphics/opengl/Font.h"
      $116 = HEAP32[$79>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $117 = ($116|0)==($110|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($117) {
       break;
      } else {
       $115 = $116;
      }
     }
     $$pre$i25 = HEAP32[$newtext1>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $118 = $$pre$i25;
    }
    __ZdlPv($118); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $$1 = 0;
   STACKTOP = sp;return ($$1|0); //@line 109 "modules/graphics/opengl/wrap_Text.cpp"
  }
 } while(0);
 if ((label|0) == 50) {
  $120 = ___cxa_find_matching_catch_2()|0;
  $121 = tempRet0;
  $eh$lpad$body22$index16Z2D = $121;$eh$lpad$body22$indexZ2D = $120;
 }
 $122 = HEAP32[$newtext1>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $123 = ($122|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($123) {
  $$02 = $eh$lpad$body22$index16Z2D;$$03 = $eh$lpad$body22$indexZ2D;
  ___resumeException($$03|0);
  // unreachable;
 }
 $124 = HEAP32[$79>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $125 = ($124|0)==($122|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($125) {
  $130 = $122;
 } else {
  $127 = $124;
  while(1) {
   $126 = ((($127)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$79>>2] = $126; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($126); //@line 64 "modules/graphics/opengl/Font.h"
   $128 = HEAP32[$79>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $129 = ($128|0)==($122|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($129) {
    break;
   } else {
    $127 = $128;
   }
  }
  $$pre$i31 = HEAP32[$newtext1>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $130 = $$pre$i31;
 }
 __ZdlPv($130); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $$02 = $eh$lpad$body22$index16Z2D;$$03 = $eh$lpad$body22$indexZ2D;
 ___resumeException($$03|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8graphics6opengl11w_Text_setfEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$02$i = 0, $$pre$i = 0, $$pre$i5 = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $align = 0, $eh$lpad$body$index5Z2D = 0, $eh$lpad$body$indexZ2D = 0, $newtext = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $align = sp + 24|0;
 $newtext = sp + 12|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4TextEEEPT_P9lua_StateiNS_4TypeE($L,1,20)|0); //@line 32 "modules/graphics/opengl/wrap_Text.cpp"
 $1 = (+_luaL_checknumber($L,3)); //@line 115 "modules/graphics/opengl/wrap_Text.cpp"
 $2 = $1; //@line 115 "modules/graphics/opengl/wrap_Text.cpp"
 $3 = (_luaL_checklstring($L,4,0)|0); //@line 118 "modules/graphics/opengl/wrap_Text.cpp"
 $4 = (__ZN4love8graphics6opengl4Font11getConstantEPKcRNS2_9AlignModeE($3,$align)|0); //@line 119 "modules/graphics/opengl/wrap_Text.cpp"
 if (!($4)) {
  HEAP32[$vararg_buffer>>2] = $3; //@line 120 "modules/graphics/opengl/wrap_Text.cpp"
  $5 = (_luaL_error($L,345198,$vararg_buffer)|0); //@line 120 "modules/graphics/opengl/wrap_Text.cpp"
  $$0 = $5;
  STACKTOP = sp;return ($$0|0); //@line 128 "modules/graphics/opengl/wrap_Text.cpp"
 }
 HEAP32[$newtext>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ((($newtext)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$6>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ((($newtext)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$7>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 invoke_viii(3568,($L|0),2,($newtext|0)); //@line 123 "modules/graphics/opengl/wrap_Text.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 L5: do {
  if ($9) {
   label = 19;
  } else {
   $10 = HEAP32[$align>>2]|0; //@line 125 "modules/graphics/opengl/wrap_Text.cpp"
   __THREW__ = 0;
   invoke_viidi(3619,($0|0),($newtext|0),(+$2),($10|0)); //@line 125 "modules/graphics/opengl/wrap_Text.cpp"
   $11 = __THREW__; __THREW__ = 0;
   $12 = $11&1;
   L7: do {
    if ($12) {
     $13 = ___cxa_find_matching_catch_3(8992|0)|0;
     $14 = tempRet0;
     $15 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
     $16 = ($14|0)==($15|0); //@line 535 "./common/runtime.h"
     do {
      if ($16) {
       $17 = (___cxa_begin_catch(($13|0))|0); //@line 535 "./common/runtime.h"
       $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
       $19 = ((($18)) + 8|0); //@line 539 "./common/runtime.h"
       $20 = HEAP32[$19>>2]|0; //@line 539 "./common/runtime.h"
       $21 = (FUNCTION_TABLE_ii[$20 & 8191]($17)|0); //@line 539 "./common/runtime.h"
       __THREW__ = 0;
       invoke_vii(3380,($L|0),($21|0)); //@line 539 "./common/runtime.h"
       $22 = __THREW__; __THREW__ = 0;
       $23 = $22&1;
       if ($23) {
        $31 = ___cxa_find_matching_catch_2()|0;
        $32 = tempRet0;
        __THREW__ = 0;
        invoke_v(3378); //@line 540 "./common/runtime.h"
        $33 = __THREW__; __THREW__ = 0;
        $34 = $33&1;
        if (!($34)) {
         $$01$i = $32;$$02$i = $31;
         break;
        }
        $35 = ___cxa_find_matching_catch_3(0|0)|0;
        $36 = tempRet0;
        ___clang_call_terminate($35); //@line 540 "./common/runtime.h"
        // unreachable; //@line 540 "./common/runtime.h"
       }
       __THREW__ = 0;
       invoke_v(3378); //@line 540 "./common/runtime.h"
       $24 = __THREW__; __THREW__ = 0;
       $25 = $24&1;
       if ($25) {
        label = 19;
        break L5;
       }
       __THREW__ = 0;
       $26 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
       $27 = __THREW__; __THREW__ = 0;
       $28 = $27&1;
       if ($28) {
        label = 19;
        break L5;
       }
       __THREW__ = 0;
       HEAP32[$vararg_buffer1>>2] = $26; //@line 543 "./common/runtime.h"
       (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer1|0))|0); //@line 543 "./common/runtime.h"
       $29 = __THREW__; __THREW__ = 0;
       $30 = $29&1;
       if ($30) {
        label = 19;
        break L5;
       } else {
        break L7;
       }
      } else {
       $$01$i = $14;$$02$i = $13;
      }
     } while(0);
     $eh$lpad$body$index5Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
     break L5;
    }
   } while(0);
   $37 = HEAP32[$newtext>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $38 = ($37|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($38)) {
    $39 = HEAP32[$6>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $40 = ($39|0)==($37|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($40) {
     $45 = $37;
    } else {
     $42 = $39;
     while(1) {
      $41 = ((($42)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$6>>2] = $41; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($41); //@line 64 "modules/graphics/opengl/Font.h"
      $43 = HEAP32[$6>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $44 = ($43|0)==($37|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($44) {
       break;
      } else {
       $42 = $43;
      }
     }
     $$pre$i5 = HEAP32[$newtext>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $45 = $$pre$i5;
    }
    __ZdlPv($45); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $$0 = 0;
   STACKTOP = sp;return ($$0|0); //@line 128 "modules/graphics/opengl/wrap_Text.cpp"
  }
 } while(0);
 if ((label|0) == 19) {
  $46 = ___cxa_find_matching_catch_2()|0;
  $47 = tempRet0;
  $eh$lpad$body$index5Z2D = $47;$eh$lpad$body$indexZ2D = $46;
 }
 $48 = HEAP32[$newtext>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $49 = ($48|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($49) {
  ___resumeException($eh$lpad$body$indexZ2D|0);
  // unreachable;
 }
 $50 = HEAP32[$6>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $51 = ($50|0)==($48|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($51) {
  $56 = $48;
 } else {
  $53 = $50;
  while(1) {
   $52 = ((($53)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$6>>2] = $52; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($52); //@line 64 "modules/graphics/opengl/Font.h"
   $54 = HEAP32[$6>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $55 = ($54|0)==($48|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($55) {
    break;
   } else {
    $53 = $54;
   }
  }
  $$pre$i = HEAP32[$newtext>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $56 = $$pre$i;
 }
 __ZdlPv($56); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8graphics6opengl10w_Text_addEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $$pre$i = 0, $$pre$i5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0;
 var $23 = 0, $24 = 0, $25 = 0.0, $26 = 0.0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0.0;
 var $41 = 0.0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0.0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0.0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0.0, $90 = 0, $91 = 0, $92 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $index$0 = 0;
 var $text = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $text = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4TextEEEPT_P9lua_StateiNS_4TypeE($L,1,20)|0); //@line 32 "modules/graphics/opengl/wrap_Text.cpp"
 HEAP32[$text>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($text)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($text)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 invoke_viii(3568,($L|0),2,($text|0)); //@line 135 "modules/graphics/opengl/wrap_Text.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 L1: do {
  if ($4) {
   label = 27;
  } else {
   __THREW__ = 0;
   $5 = (+invoke_diid(3525,($L|0),3,0.0)); //@line 137 "modules/graphics/opengl/wrap_Text.cpp"
   $6 = __THREW__; __THREW__ = 0;
   $7 = $6&1;
   if ($7) {
    label = 27;
   } else {
    $8 = $5; //@line 137 "modules/graphics/opengl/wrap_Text.cpp"
    __THREW__ = 0;
    $9 = (+invoke_diid(3525,($L|0),4,0.0)); //@line 138 "modules/graphics/opengl/wrap_Text.cpp"
    $10 = __THREW__; __THREW__ = 0;
    $11 = $10&1;
    if ($11) {
     label = 27;
    } else {
     $12 = $9; //@line 138 "modules/graphics/opengl/wrap_Text.cpp"
     __THREW__ = 0;
     $13 = (+invoke_diid(3525,($L|0),5,0.0)); //@line 139 "modules/graphics/opengl/wrap_Text.cpp"
     $14 = __THREW__; __THREW__ = 0;
     $15 = $14&1;
     if ($15) {
      label = 27;
     } else {
      $16 = $13; //@line 139 "modules/graphics/opengl/wrap_Text.cpp"
      __THREW__ = 0;
      $17 = (+invoke_diid(3525,($L|0),6,1.0)); //@line 140 "modules/graphics/opengl/wrap_Text.cpp"
      $18 = __THREW__; __THREW__ = 0;
      $19 = $18&1;
      if ($19) {
       label = 27;
      } else {
       $20 = $17; //@line 140 "modules/graphics/opengl/wrap_Text.cpp"
       $21 = $20; //@line 141 "modules/graphics/opengl/wrap_Text.cpp"
       __THREW__ = 0;
       $22 = (+invoke_diid(3525,($L|0),7,(+$21))); //@line 141 "modules/graphics/opengl/wrap_Text.cpp"
       $23 = __THREW__; __THREW__ = 0;
       $24 = $23&1;
       if ($24) {
        label = 27;
       } else {
        $25 = $22; //@line 141 "modules/graphics/opengl/wrap_Text.cpp"
        __THREW__ = 0;
        $26 = (+invoke_diid(3525,($L|0),8,0.0)); //@line 142 "modules/graphics/opengl/wrap_Text.cpp"
        $27 = __THREW__; __THREW__ = 0;
        $28 = $27&1;
        if ($28) {
         label = 27;
        } else {
         $29 = $26; //@line 142 "modules/graphics/opengl/wrap_Text.cpp"
         __THREW__ = 0;
         $30 = (+invoke_diid(3525,($L|0),9,0.0)); //@line 143 "modules/graphics/opengl/wrap_Text.cpp"
         $31 = __THREW__; __THREW__ = 0;
         $32 = $31&1;
         if ($32) {
          label = 27;
         } else {
          $33 = $30; //@line 143 "modules/graphics/opengl/wrap_Text.cpp"
          __THREW__ = 0;
          $34 = (+invoke_diid(3525,($L|0),10,0.0)); //@line 144 "modules/graphics/opengl/wrap_Text.cpp"
          $35 = __THREW__; __THREW__ = 0;
          $36 = $35&1;
          if ($36) {
           label = 27;
          } else {
           __THREW__ = 0;
           $37 = (+invoke_diid(3525,($L|0),11,0.0)); //@line 145 "modules/graphics/opengl/wrap_Text.cpp"
           $38 = __THREW__; __THREW__ = 0;
           $39 = $38&1;
           if ($39) {
            label = 27;
           } else {
            $40 = $34; //@line 144 "modules/graphics/opengl/wrap_Text.cpp"
            $41 = $37; //@line 145 "modules/graphics/opengl/wrap_Text.cpp"
            __THREW__ = 0;
            $42 = (invoke_iiiddddddddd(3620,($0|0),($text|0),(+$8),(+$12),(+$16),(+$20),(+$25),(+$29),(+$33),(+$40),(+$41))|0); //@line 148 "modules/graphics/opengl/wrap_Text.cpp"
            $43 = __THREW__; __THREW__ = 0;
            $44 = $43&1;
            L12: do {
             if ($44) {
              $45 = ___cxa_find_matching_catch_3(8992|0)|0;
              $46 = tempRet0;
              $47 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
              $48 = ($46|0)==($47|0); //@line 535 "./common/runtime.h"
              do {
               if ($48) {
                $49 = (___cxa_begin_catch(($45|0))|0); //@line 535 "./common/runtime.h"
                $50 = HEAP32[$49>>2]|0; //@line 539 "./common/runtime.h"
                $51 = ((($50)) + 8|0); //@line 539 "./common/runtime.h"
                $52 = HEAP32[$51>>2]|0; //@line 539 "./common/runtime.h"
                $53 = (FUNCTION_TABLE_ii[$52 & 8191]($49)|0); //@line 539 "./common/runtime.h"
                __THREW__ = 0;
                invoke_vii(3380,($L|0),($53|0)); //@line 539 "./common/runtime.h"
                $54 = __THREW__; __THREW__ = 0;
                $55 = $54&1;
                if ($55) {
                 $63 = ___cxa_find_matching_catch_2()|0;
                 $64 = tempRet0;
                 __THREW__ = 0;
                 invoke_v(3378); //@line 540 "./common/runtime.h"
                 $65 = __THREW__; __THREW__ = 0;
                 $66 = $65&1;
                 if (!($66)) {
                  $$01$i = $64;$$02$i = $63;
                  break;
                 }
                 $67 = ___cxa_find_matching_catch_3(0|0)|0;
                 $68 = tempRet0;
                 ___clang_call_terminate($67); //@line 540 "./common/runtime.h"
                 // unreachable; //@line 540 "./common/runtime.h"
                }
                __THREW__ = 0;
                invoke_v(3378); //@line 540 "./common/runtime.h"
                $56 = __THREW__; __THREW__ = 0;
                $57 = $56&1;
                if ($57) {
                 label = 27;
                 break L1;
                }
                __THREW__ = 0;
                $58 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
                $59 = __THREW__; __THREW__ = 0;
                $60 = $59&1;
                if ($60) {
                 label = 27;
                 break L1;
                }
                __THREW__ = 0;
                HEAP32[$vararg_buffer>>2] = $58; //@line 543 "./common/runtime.h"
                (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 543 "./common/runtime.h"
                $61 = __THREW__; __THREW__ = 0;
                $62 = $61&1;
                if ($62) {
                 label = 27;
                 break L1;
                } else {
                 $index$0 = 0;
                 break L12;
                }
               } else {
                $$01$i = $46;$$02$i = $45;
               }
              } while(0);
              $eh$lpad$body$index2Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
              break L1;
             } else {
              $index$0 = $42;
             }
            } while(0);
            $69 = (($index$0) + 1)|0; //@line 149 "modules/graphics/opengl/wrap_Text.cpp"
            $70 = (+($69|0)); //@line 149 "modules/graphics/opengl/wrap_Text.cpp"
            __THREW__ = 0;
            invoke_vid(3394,($L|0),(+$70)); //@line 149 "modules/graphics/opengl/wrap_Text.cpp"
            $71 = __THREW__; __THREW__ = 0;
            $72 = $71&1;
            if ($72) {
             label = 27;
            } else {
             $73 = HEAP32[$text>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $74 = ($73|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             if ($74) {
              STACKTOP = sp;return 1; //@line 152 "modules/graphics/opengl/wrap_Text.cpp"
             }
             $75 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $76 = ($75|0)==($73|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             if ($76) {
              $81 = $73;
             } else {
              $78 = $75;
              while(1) {
               $77 = ((($78)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
               HEAP32[$1>>2] = $77; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
               __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($77); //@line 64 "modules/graphics/opengl/Font.h"
               $79 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
               $80 = ($79|0)==($73|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
               if ($80) {
                break;
               } else {
                $78 = $79;
               }
              }
              $$pre$i5 = HEAP32[$text>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $81 = $$pre$i5;
             }
             __ZdlPv($81); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
             STACKTOP = sp;return 1; //@line 152 "modules/graphics/opengl/wrap_Text.cpp"
            }
           }
          }
         }
        }
       }
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 27) {
  $82 = ___cxa_find_matching_catch_2()|0;
  $83 = tempRet0;
  $eh$lpad$body$index2Z2D = $83;$eh$lpad$body$indexZ2D = $82;
 }
 $84 = HEAP32[$text>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $85 = ($84|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($85) {
  ___resumeException($eh$lpad$body$indexZ2D|0);
  // unreachable;
 }
 $86 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $87 = ($86|0)==($84|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($87) {
  $92 = $84;
 } else {
  $89 = $86;
  while(1) {
   $88 = ((($89)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$1>>2] = $88; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($88); //@line 64 "modules/graphics/opengl/Font.h"
   $90 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $91 = ($90|0)==($84|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($91) {
    break;
   } else {
    $89 = $90;
   }
  }
  $$pre$i = HEAP32[$text>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $92 = $$pre$i;
 }
 __ZdlPv($92); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8graphics6opengl11w_Text_addfEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$02$i = 0, $$pre$i = 0, $$pre$i3 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0.0;
 var $34 = 0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0.0, $50 = 0.0, $51 = 0;
 var $52 = 0, $53 = 0.0, $54 = 0.0, $55 = 0, $56 = 0, $57 = 0.0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0, $63 = 0, $64 = 0.0, $65 = 0.0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0.0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0.0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $align = 0, $eh$lpad$body$index5Z2D = 0, $eh$lpad$body$indexZ2D = 0, $index$0 = 0, $text = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $text = sp + 16|0;
 $align = sp + 12|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4TextEEEPT_P9lua_StateiNS_4TypeE($L,1,20)|0); //@line 32 "modules/graphics/opengl/wrap_Text.cpp"
 HEAP32[$text>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($text)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($text)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 invoke_viii(3568,($L|0),2,($text|0)); //@line 159 "modules/graphics/opengl/wrap_Text.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 L1: do {
  if ($4) {
   label = 7;
  } else {
   __THREW__ = 0;
   $5 = (+invoke_dii(3361,($L|0),3)); //@line 161 "modules/graphics/opengl/wrap_Text.cpp"
   $6 = __THREW__; __THREW__ = 0;
   $7 = $6&1;
   if ($7) {
    label = 7;
   } else {
    $8 = $5; //@line 161 "modules/graphics/opengl/wrap_Text.cpp"
    HEAP32[$align>>2] = 4; //@line 163 "modules/graphics/opengl/wrap_Text.cpp"
    __THREW__ = 0;
    $9 = (invoke_iiii(3359,($L|0),4,(0|0))|0); //@line 164 "modules/graphics/opengl/wrap_Text.cpp"
    $10 = __THREW__; __THREW__ = 0;
    $11 = $10&1;
    if ($11) {
     label = 7;
    } else {
     __THREW__ = 0;
     $12 = (invoke_iii(3581,($9|0),($align|0))|0); //@line 166 "modules/graphics/opengl/wrap_Text.cpp"
     $13 = __THREW__; __THREW__ = 0;
     $14 = $13&1;
     if ($14) {
      label = 7;
     } else {
      if ($12) {
       __THREW__ = 0;
       $29 = (+invoke_diid(3525,($L|0),5,0.0)); //@line 169 "modules/graphics/opengl/wrap_Text.cpp"
       $30 = __THREW__; __THREW__ = 0;
       $31 = $30&1;
       if ($31) {
        label = 7;
        break;
       }
       $32 = $29; //@line 169 "modules/graphics/opengl/wrap_Text.cpp"
       __THREW__ = 0;
       $33 = (+invoke_diid(3525,($L|0),6,0.0)); //@line 170 "modules/graphics/opengl/wrap_Text.cpp"
       $34 = __THREW__; __THREW__ = 0;
       $35 = $34&1;
       if ($35) {
        label = 7;
        break;
       }
       $36 = $33; //@line 170 "modules/graphics/opengl/wrap_Text.cpp"
       __THREW__ = 0;
       $37 = (+invoke_diid(3525,($L|0),7,0.0)); //@line 171 "modules/graphics/opengl/wrap_Text.cpp"
       $38 = __THREW__; __THREW__ = 0;
       $39 = $38&1;
       if ($39) {
        label = 7;
        break;
       }
       $40 = $37; //@line 171 "modules/graphics/opengl/wrap_Text.cpp"
       __THREW__ = 0;
       $41 = (+invoke_diid(3525,($L|0),8,1.0)); //@line 172 "modules/graphics/opengl/wrap_Text.cpp"
       $42 = __THREW__; __THREW__ = 0;
       $43 = $42&1;
       if ($43) {
        label = 7;
        break;
       }
       $44 = $41; //@line 172 "modules/graphics/opengl/wrap_Text.cpp"
       $45 = $44; //@line 173 "modules/graphics/opengl/wrap_Text.cpp"
       __THREW__ = 0;
       $46 = (+invoke_diid(3525,($L|0),9,(+$45))); //@line 173 "modules/graphics/opengl/wrap_Text.cpp"
       $47 = __THREW__; __THREW__ = 0;
       $48 = $47&1;
       if ($48) {
        label = 7;
        break;
       }
       $49 = $46; //@line 173 "modules/graphics/opengl/wrap_Text.cpp"
       __THREW__ = 0;
       $50 = (+invoke_diid(3525,($L|0),10,0.0)); //@line 174 "modules/graphics/opengl/wrap_Text.cpp"
       $51 = __THREW__; __THREW__ = 0;
       $52 = $51&1;
       if ($52) {
        label = 7;
        break;
       }
       $53 = $50; //@line 174 "modules/graphics/opengl/wrap_Text.cpp"
       __THREW__ = 0;
       $54 = (+invoke_diid(3525,($L|0),11,0.0)); //@line 175 "modules/graphics/opengl/wrap_Text.cpp"
       $55 = __THREW__; __THREW__ = 0;
       $56 = $55&1;
       if ($56) {
        label = 7;
        break;
       }
       $57 = $54; //@line 175 "modules/graphics/opengl/wrap_Text.cpp"
       __THREW__ = 0;
       $58 = (+invoke_diid(3525,($L|0),12,0.0)); //@line 176 "modules/graphics/opengl/wrap_Text.cpp"
       $59 = __THREW__; __THREW__ = 0;
       $60 = $59&1;
       if ($60) {
        label = 7;
        break;
       }
       __THREW__ = 0;
       $61 = (+invoke_diid(3525,($L|0),13,0.0)); //@line 177 "modules/graphics/opengl/wrap_Text.cpp"
       $62 = __THREW__; __THREW__ = 0;
       $63 = $62&1;
       if ($63) {
        label = 7;
        break;
       }
       $64 = $58; //@line 176 "modules/graphics/opengl/wrap_Text.cpp"
       $65 = $61; //@line 177 "modules/graphics/opengl/wrap_Text.cpp"
       $66 = HEAP32[$align>>2]|0; //@line 180 "modules/graphics/opengl/wrap_Text.cpp"
       __THREW__ = 0;
       $67 = (invoke_iiididdddddddd(3621,($0|0),($text|0),(+$8),($66|0),(+$32),(+$36),(+$40),(+$44),(+$49),(+$53),(+$57),(+$64),(+$65))|0); //@line 180 "modules/graphics/opengl/wrap_Text.cpp"
       $68 = __THREW__; __THREW__ = 0;
       $69 = $68&1;
       L17: do {
        if ($69) {
         $70 = ___cxa_find_matching_catch_3(8992|0)|0;
         $71 = tempRet0;
         $72 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
         $73 = ($71|0)==($72|0); //@line 535 "./common/runtime.h"
         do {
          if ($73) {
           $74 = (___cxa_begin_catch(($70|0))|0); //@line 535 "./common/runtime.h"
           $75 = HEAP32[$74>>2]|0; //@line 539 "./common/runtime.h"
           $76 = ((($75)) + 8|0); //@line 539 "./common/runtime.h"
           $77 = HEAP32[$76>>2]|0; //@line 539 "./common/runtime.h"
           $78 = (FUNCTION_TABLE_ii[$77 & 8191]($74)|0); //@line 539 "./common/runtime.h"
           __THREW__ = 0;
           invoke_vii(3380,($L|0),($78|0)); //@line 539 "./common/runtime.h"
           $79 = __THREW__; __THREW__ = 0;
           $80 = $79&1;
           if ($80) {
            $88 = ___cxa_find_matching_catch_2()|0;
            $89 = tempRet0;
            __THREW__ = 0;
            invoke_v(3378); //@line 540 "./common/runtime.h"
            $90 = __THREW__; __THREW__ = 0;
            $91 = $90&1;
            if (!($91)) {
             $$01$i = $89;$$02$i = $88;
             break;
            }
            $92 = ___cxa_find_matching_catch_3(0|0)|0;
            $93 = tempRet0;
            ___clang_call_terminate($92); //@line 540 "./common/runtime.h"
            // unreachable; //@line 540 "./common/runtime.h"
           }
           __THREW__ = 0;
           invoke_v(3378); //@line 540 "./common/runtime.h"
           $81 = __THREW__; __THREW__ = 0;
           $82 = $81&1;
           if ($82) {
            label = 7;
            break L1;
           }
           __THREW__ = 0;
           $83 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
           $84 = __THREW__; __THREW__ = 0;
           $85 = $84&1;
           if ($85) {
            label = 7;
            break L1;
           }
           __THREW__ = 0;
           HEAP32[$vararg_buffer1>>2] = $83; //@line 543 "./common/runtime.h"
           (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer1|0))|0); //@line 543 "./common/runtime.h"
           $86 = __THREW__; __THREW__ = 0;
           $87 = $86&1;
           if ($87) {
            label = 7;
            break L1;
           } else {
            $index$0 = 0;
            break L17;
           }
          } else {
           $$01$i = $71;$$02$i = $70;
          }
         } while(0);
         $eh$lpad$body$index5Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
         break L1;
        } else {
         $index$0 = $67;
        }
       } while(0);
       $94 = (($index$0) + 1)|0; //@line 181 "modules/graphics/opengl/wrap_Text.cpp"
       $95 = (+($94|0)); //@line 181 "modules/graphics/opengl/wrap_Text.cpp"
       __THREW__ = 0;
       invoke_vid(3394,($L|0),(+$95)); //@line 181 "modules/graphics/opengl/wrap_Text.cpp"
       $96 = __THREW__; __THREW__ = 0;
       $97 = $96&1;
       if ($97) {
        label = 7;
        break;
       } else {
        $$0 = 1;
       }
      } else {
       __THREW__ = 0;
       HEAP32[$vararg_buffer>>2] = $9; //@line 167 "modules/graphics/opengl/wrap_Text.cpp"
       $15 = (invoke_iiii(3389,($L|0),(345198|0),($vararg_buffer|0))|0); //@line 167 "modules/graphics/opengl/wrap_Text.cpp"
       $16 = __THREW__; __THREW__ = 0;
       $17 = $16&1;
       if ($17) {
        label = 7;
        break;
       } else {
        $$0 = $15;
       }
      }
      $98 = HEAP32[$text>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $99 = ($98|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($99) {
       STACKTOP = sp;return ($$0|0); //@line 184 "modules/graphics/opengl/wrap_Text.cpp"
      }
      $100 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $101 = ($100|0)==($98|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($101) {
       $106 = $98;
      } else {
       $103 = $100;
       while(1) {
        $102 = ((($103)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$1>>2] = $102; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($102); //@line 64 "modules/graphics/opengl/Font.h"
        $104 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $105 = ($104|0)==($98|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($105) {
         break;
        } else {
         $103 = $104;
        }
       }
       $$pre$i = HEAP32[$text>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $106 = $$pre$i;
      }
      __ZdlPv($106); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      STACKTOP = sp;return ($$0|0); //@line 184 "modules/graphics/opengl/wrap_Text.cpp"
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 7) {
  $18 = ___cxa_find_matching_catch_2()|0;
  $19 = tempRet0;
  $eh$lpad$body$index5Z2D = $19;$eh$lpad$body$indexZ2D = $18;
 }
 $20 = HEAP32[$text>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = ($20|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($21) {
  ___resumeException($eh$lpad$body$indexZ2D|0);
  // unreachable;
 }
 $22 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $23 = ($22|0)==($20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($23) {
  $28 = $20;
 } else {
  $25 = $22;
  while(1) {
   $24 = ((($25)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$1>>2] = $24; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($24); //@line 64 "modules/graphics/opengl/Font.h"
   $26 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $27 = ($26|0)==($20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($27) {
    break;
   } else {
    $25 = $26;
   }
  }
  $$pre$i3 = HEAP32[$text>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $28 = $$pre$i3;
 }
 __ZdlPv($28); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8graphics6opengl12w_Text_clearEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 186 "modules/graphics/opengl/wrap_Text.cpp"
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4TextEEEPT_P9lua_StateiNS_4TypeE($L,1,20)|0); //@line 32 "modules/graphics/opengl/wrap_Text.cpp"
 __THREW__ = 0;
 invoke_vi(3622,($0|0)); //@line 189 "modules/graphics/opengl/wrap_Text.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if (!($2)) {
  STACKTOP = sp;return 0; //@line 190 "modules/graphics/opengl/wrap_Text.cpp"
 }
 $3 = ___cxa_find_matching_catch_3(8992|0)|0;
 $4 = tempRet0;
 $5 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $6 = ($4|0)==($5|0); //@line 535 "./common/runtime.h"
 if (!($6)) {
  $$01$i = $4;$$02$i = $3;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $7 = (___cxa_begin_catch(($3|0))|0); //@line 535 "./common/runtime.h"
 $8 = HEAP32[$7>>2]|0; //@line 539 "./common/runtime.h"
 $9 = ((($8)) + 8|0); //@line 539 "./common/runtime.h"
 $10 = HEAP32[$9>>2]|0; //@line 539 "./common/runtime.h"
 $11 = (FUNCTION_TABLE_ii[$10 & 8191]($7)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($11|0)); //@line 539 "./common/runtime.h"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if (!($13)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $14 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $14; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 190 "modules/graphics/opengl/wrap_Text.cpp"
 }
 $15 = ___cxa_find_matching_catch_2()|0;
 $16 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if ($18) {
  $19 = ___cxa_find_matching_catch_3(0|0)|0;
  $20 = tempRet0;
  ___clang_call_terminate($19); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $16;$$02$i = $15;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl14w_Text_setFontEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 193 "modules/graphics/opengl/wrap_Text.cpp"
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4TextEEEPT_P9lua_StateiNS_4TypeE($L,1,20)|0); //@line 32 "modules/graphics/opengl/wrap_Text.cpp"
 $1 = (__ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE($L,2,14)|0); //@line 196 "modules/graphics/opengl/wrap_Text.cpp"
 __THREW__ = 0;
 invoke_vii(3623,($0|0),($1|0)); //@line 197 "modules/graphics/opengl/wrap_Text.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  STACKTOP = sp;return 0; //@line 198 "modules/graphics/opengl/wrap_Text.cpp"
 }
 $4 = ___cxa_find_matching_catch_3(8992|0)|0;
 $5 = tempRet0;
 $6 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $7 = ($5|0)==($6|0); //@line 535 "./common/runtime.h"
 if (!($7)) {
  $$01$i = $5;$$02$i = $4;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $8 = (___cxa_begin_catch(($4|0))|0); //@line 535 "./common/runtime.h"
 $9 = HEAP32[$8>>2]|0; //@line 539 "./common/runtime.h"
 $10 = ((($9)) + 8|0); //@line 539 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($8)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($12|0)); //@line 539 "./common/runtime.h"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if (!($14)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $15 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $15; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 198 "modules/graphics/opengl/wrap_Text.cpp"
 }
 $16 = ___cxa_find_matching_catch_2()|0;
 $17 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $20 = ___cxa_find_matching_catch_3(0|0)|0;
  $21 = tempRet0;
  ___clang_call_terminate($20); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $17;$$02$i = $16;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl14w_Text_getFontEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4TextEEEPT_P9lua_StateiNS_4TypeE($L,1,20)|0); //@line 32 "modules/graphics/opengl/wrap_Text.cpp"
 $1 = (__ZNK4love8graphics6opengl4Text7getFontEv($0)|0); //@line 204 "modules/graphics/opengl/wrap_Text.cpp"
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,14,$1); //@line 205 "modules/graphics/opengl/wrap_Text.cpp"
 return 1; //@line 206 "modules/graphics/opengl/wrap_Text.cpp"
}
function __ZN4love8graphics6opengl15w_Text_getWidthEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4TextEEEPT_P9lua_StateiNS_4TypeE($L,1,20)|0); //@line 32 "modules/graphics/opengl/wrap_Text.cpp"
 $1 = (+_luaL_optnumber($L,2,0.0)); //@line 212 "modules/graphics/opengl/wrap_Text.cpp"
 $2 = (~~(($1))); //@line 212 "modules/graphics/opengl/wrap_Text.cpp"
 $3 = (($2) + -1)|0; //@line 212 "modules/graphics/opengl/wrap_Text.cpp"
 $4 = (__ZNK4love8graphics6opengl4Text8getWidthEi($0,$3)|0); //@line 213 "modules/graphics/opengl/wrap_Text.cpp"
 $5 = (+($4|0)); //@line 213 "modules/graphics/opengl/wrap_Text.cpp"
 _lua_pushnumber($L,$5); //@line 213 "modules/graphics/opengl/wrap_Text.cpp"
 return 1; //@line 214 "modules/graphics/opengl/wrap_Text.cpp"
}
function __ZN4love8graphics6opengl16w_Text_getHeightEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4TextEEEPT_P9lua_StateiNS_4TypeE($L,1,20)|0); //@line 32 "modules/graphics/opengl/wrap_Text.cpp"
 $1 = (+_luaL_optnumber($L,2,0.0)); //@line 220 "modules/graphics/opengl/wrap_Text.cpp"
 $2 = (~~(($1))); //@line 220 "modules/graphics/opengl/wrap_Text.cpp"
 $3 = (($2) + -1)|0; //@line 220 "modules/graphics/opengl/wrap_Text.cpp"
 $4 = (__ZNK4love8graphics6opengl4Text9getHeightEi($0,$3)|0); //@line 221 "modules/graphics/opengl/wrap_Text.cpp"
 $5 = (+($4|0)); //@line 221 "modules/graphics/opengl/wrap_Text.cpp"
 _lua_pushnumber($L,$5); //@line 221 "modules/graphics/opengl/wrap_Text.cpp"
 return 1; //@line 222 "modules/graphics/opengl/wrap_Text.cpp"
}
function __ZN4love8graphics6opengl20w_Text_getDimensionsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4TextEEEPT_P9lua_StateiNS_4TypeE($L,1,20)|0); //@line 32 "modules/graphics/opengl/wrap_Text.cpp"
 $1 = (+_luaL_optnumber($L,2,0.0)); //@line 228 "modules/graphics/opengl/wrap_Text.cpp"
 $2 = (~~(($1))); //@line 228 "modules/graphics/opengl/wrap_Text.cpp"
 $3 = (($2) + -1)|0; //@line 228 "modules/graphics/opengl/wrap_Text.cpp"
 $4 = (__ZNK4love8graphics6opengl4Text8getWidthEi($0,$3)|0); //@line 229 "modules/graphics/opengl/wrap_Text.cpp"
 $5 = (+($4|0)); //@line 229 "modules/graphics/opengl/wrap_Text.cpp"
 _lua_pushnumber($L,$5); //@line 229 "modules/graphics/opengl/wrap_Text.cpp"
 $6 = (__ZNK4love8graphics6opengl4Text9getHeightEi($0,$3)|0); //@line 230 "modules/graphics/opengl/wrap_Text.cpp"
 $7 = (+($6|0)); //@line 230 "modules/graphics/opengl/wrap_Text.cpp"
 _lua_pushnumber($L,$7); //@line 230 "modules/graphics/opengl/wrap_Text.cpp"
 return 2; //@line 231 "modules/graphics/opengl/wrap_Text.cpp"
}
function _luaopen_text($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 249 "modules/graphics/opengl/wrap_Text.cpp"
 HEAP32[$vararg_buffer>>2] = 15000; //@line 251 "modules/graphics/opengl/wrap_Text.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 251 "modules/graphics/opengl/wrap_Text.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 251 "modules/graphics/opengl/wrap_Text.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,20,345221,$vararg_buffer)|0); //@line 251 "modules/graphics/opengl/wrap_Text.cpp"
 STACKTOP = sp;return ($0|0); //@line 251 "modules/graphics/opengl/wrap_Text.cpp"
}
function __ZN4love14luax_checktypeINS_8graphics6opengl5VideoEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love8graphics6opengl17w_Video_getStreamEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl5VideoEEEPT_P9lua_StateiNS_4TypeE($L,1,21)|0); //@line 37 "modules/graphics/opengl/wrap_Video.cpp"
 $1 = (__ZN4love8graphics6opengl5Video9getStreamEv($0)|0); //@line 43 "modules/graphics/opengl/wrap_Video.cpp"
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,55,$1); //@line 43 "modules/graphics/opengl/wrap_Video.cpp"
 return 1; //@line 44 "modules/graphics/opengl/wrap_Video.cpp"
}
function __ZN4love8graphics6opengl17w_Video_getSourceEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl5VideoEEEPT_P9lua_StateiNS_4TypeE($L,1,21)|0); //@line 37 "modules/graphics/opengl/wrap_Video.cpp"
 $1 = (__ZN4love8graphics6opengl5Video9getSourceEv($0)|0); //@line 50 "modules/graphics/opengl/wrap_Video.cpp"
 $2 = ($1|0)==(0|0); //@line 51 "modules/graphics/opengl/wrap_Video.cpp"
 if ($2) {
  _lua_pushnil($L); //@line 54 "modules/graphics/opengl/wrap_Video.cpp"
  return 1; //@line 55 "modules/graphics/opengl/wrap_Video.cpp"
 } else {
  $3 = (__ZN4love8graphics6opengl5Video9getSourceEv($0)|0); //@line 52 "modules/graphics/opengl/wrap_Video.cpp"
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,28,$3); //@line 52 "modules/graphics/opengl/wrap_Video.cpp"
  return 1; //@line 55 "modules/graphics/opengl/wrap_Video.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl17w_Video_setSourceEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl5VideoEEEPT_P9lua_StateiNS_4TypeE($L,1,21)|0); //@line 37 "modules/graphics/opengl/wrap_Video.cpp"
 $1 = (_lua_type($L,2)|0); //@line 61 "modules/graphics/opengl/wrap_Video.cpp"
 $2 = ($1|0)<(1); //@line 61 "modules/graphics/opengl/wrap_Video.cpp"
 if ($2) {
  __ZN4love8graphics6opengl5Video9setSourceEPNS_5audio6SourceE($0,0); //@line 62 "modules/graphics/opengl/wrap_Video.cpp"
  return 0; //@line 68 "modules/graphics/opengl/wrap_Video.cpp"
 } else {
  $3 = (__ZN4love14luax_checktypeINS_5audio6SourceEEEPT_P9lua_StateiNS_4TypeE($L,2,28)|0); //@line 65 "modules/graphics/opengl/wrap_Video.cpp"
  __ZN4love8graphics6opengl5Video9setSourceEPNS_5audio6SourceE($0,$3); //@line 66 "modules/graphics/opengl/wrap_Video.cpp"
  return 0; //@line 68 "modules/graphics/opengl/wrap_Video.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl16w_Video_getWidthEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl5VideoEEEPT_P9lua_StateiNS_4TypeE($L,1,21)|0); //@line 37 "modules/graphics/opengl/wrap_Video.cpp"
 $1 = (__ZNK4love8graphics6opengl5Video8getWidthEv($0)|0); //@line 74 "modules/graphics/opengl/wrap_Video.cpp"
 $2 = (+($1|0)); //@line 74 "modules/graphics/opengl/wrap_Video.cpp"
 _lua_pushnumber($L,$2); //@line 74 "modules/graphics/opengl/wrap_Video.cpp"
 return 1; //@line 75 "modules/graphics/opengl/wrap_Video.cpp"
}
function __ZN4love8graphics6opengl17w_Video_getHeightEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl5VideoEEEPT_P9lua_StateiNS_4TypeE($L,1,21)|0); //@line 37 "modules/graphics/opengl/wrap_Video.cpp"
 $1 = (__ZNK4love8graphics6opengl5Video9getHeightEv($0)|0); //@line 81 "modules/graphics/opengl/wrap_Video.cpp"
 $2 = (+($1|0)); //@line 81 "modules/graphics/opengl/wrap_Video.cpp"
 _lua_pushnumber($L,$2); //@line 81 "modules/graphics/opengl/wrap_Video.cpp"
 return 1; //@line 82 "modules/graphics/opengl/wrap_Video.cpp"
}
function __ZN4love8graphics6opengl21w_Video_getDimensionsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl5VideoEEEPT_P9lua_StateiNS_4TypeE($L,1,21)|0); //@line 37 "modules/graphics/opengl/wrap_Video.cpp"
 $1 = (__ZNK4love8graphics6opengl5Video8getWidthEv($0)|0); //@line 88 "modules/graphics/opengl/wrap_Video.cpp"
 $2 = (+($1|0)); //@line 88 "modules/graphics/opengl/wrap_Video.cpp"
 _lua_pushnumber($L,$2); //@line 88 "modules/graphics/opengl/wrap_Video.cpp"
 $3 = (__ZNK4love8graphics6opengl5Video9getHeightEv($0)|0); //@line 89 "modules/graphics/opengl/wrap_Video.cpp"
 $4 = (+($3|0)); //@line 89 "modules/graphics/opengl/wrap_Video.cpp"
 _lua_pushnumber($L,$4); //@line 89 "modules/graphics/opengl/wrap_Video.cpp"
 return 2; //@line 90 "modules/graphics/opengl/wrap_Video.cpp"
}
function __ZN4love8graphics6opengl17w_Video_setFilterEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, $f = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $f = sp + 24|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl5VideoEEEPT_P9lua_StateiNS_4TypeE($L,1,21)|0); //@line 37 "modules/graphics/opengl/wrap_Video.cpp"
 $1 = (__ZNK4love8graphics6opengl5Video9getFilterEv($0)|0); //@line 96 "modules/graphics/opengl/wrap_Video.cpp"
 ;HEAP32[$f>>2]=HEAP32[$1>>2]|0;HEAP32[$f+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$f+8>>2]=HEAP32[$1+8>>2]|0;HEAP32[$f+12>>2]=HEAP32[$1+12>>2]|0; //@line 96 "modules/graphics/opengl/wrap_Video.cpp"
 $2 = (_luaL_checklstring($L,2,0)|0); //@line 98 "modules/graphics/opengl/wrap_Video.cpp"
 $3 = (_luaL_optlstring($L,3,$2,0)|0); //@line 99 "modules/graphics/opengl/wrap_Video.cpp"
 $4 = (__ZN4love8graphics7Texture11getConstantEPKcRNS1_10FilterModeE($2,$f)|0); //@line 101 "modules/graphics/opengl/wrap_Video.cpp"
 if (!($4)) {
  HEAP32[$vararg_buffer>>2] = $2; //@line 102 "modules/graphics/opengl/wrap_Video.cpp"
  $5 = (_luaL_error($L,439370,$vararg_buffer)|0); //@line 102 "modules/graphics/opengl/wrap_Video.cpp"
  $$0 = $5;
  STACKTOP = sp;return ($$0|0); //@line 110 "modules/graphics/opengl/wrap_Video.cpp"
 }
 $6 = ((($f)) + 4|0); //@line 103 "modules/graphics/opengl/wrap_Video.cpp"
 $7 = (__ZN4love8graphics7Texture11getConstantEPKcRNS1_10FilterModeE($3,$6)|0); //@line 103 "modules/graphics/opengl/wrap_Video.cpp"
 if (!($7)) {
  HEAP32[$vararg_buffer1>>2] = $3; //@line 104 "modules/graphics/opengl/wrap_Video.cpp"
  $8 = (_luaL_error($L,439370,$vararg_buffer1)|0); //@line 104 "modules/graphics/opengl/wrap_Video.cpp"
  $$0 = $8;
  STACKTOP = sp;return ($$0|0); //@line 110 "modules/graphics/opengl/wrap_Video.cpp"
 }
 $9 = (+_luaL_optnumber($L,4,1.0)); //@line 106 "modules/graphics/opengl/wrap_Video.cpp"
 $10 = $9; //@line 106 "modules/graphics/opengl/wrap_Video.cpp"
 $11 = ((($f)) + 12|0); //@line 106 "modules/graphics/opengl/wrap_Video.cpp"
 HEAPF32[$11>>2] = $10; //@line 106 "modules/graphics/opengl/wrap_Video.cpp"
 __THREW__ = 0;
 invoke_vii(3624,($0|0),($f|0)); //@line 108 "modules/graphics/opengl/wrap_Video.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if (!($13)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 110 "modules/graphics/opengl/wrap_Video.cpp"
 }
 $14 = ___cxa_find_matching_catch_3(8992|0)|0;
 $15 = tempRet0;
 $16 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $17 = ($15|0)==($16|0); //@line 535 "./common/runtime.h"
 if (!($17)) {
  $$01$i = $15;$$02$i = $14;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $18 = (___cxa_begin_catch(($14|0))|0); //@line 535 "./common/runtime.h"
 $19 = HEAP32[$18>>2]|0; //@line 539 "./common/runtime.h"
 $20 = ((($19)) + 8|0); //@line 539 "./common/runtime.h"
 $21 = HEAP32[$20>>2]|0; //@line 539 "./common/runtime.h"
 $22 = (FUNCTION_TABLE_ii[$21 & 8191]($18)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($22|0)); //@line 539 "./common/runtime.h"
 $23 = __THREW__; __THREW__ = 0;
 $24 = $23&1;
 if (!($24)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $25 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer4>>2] = $25; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer4)|0); //@line 543 "./common/runtime.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 110 "modules/graphics/opengl/wrap_Video.cpp"
 }
 $26 = ___cxa_find_matching_catch_2()|0;
 $27 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if ($29) {
  $30 = ___cxa_find_matching_catch_3(0|0)|0;
  $31 = tempRet0;
  ___clang_call_terminate($30); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $27;$$02$i = $26;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl17w_Video_getFilterEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, $magstr = 0, $minstr = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $minstr = sp + 16|0;
 $magstr = sp + 12|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl5VideoEEEPT_P9lua_StateiNS_4TypeE($L,1,21)|0); //@line 37 "modules/graphics/opengl/wrap_Video.cpp"
 $1 = (__ZNK4love8graphics6opengl5Video9getFilterEv($0)|0); //@line 115 "modules/graphics/opengl/wrap_Video.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 115 "modules/graphics/opengl/wrap_Video.cpp"
 $3 = ((($1)) + 4|0); //@line 115 "modules/graphics/opengl/wrap_Video.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 115 "modules/graphics/opengl/wrap_Video.cpp"
 $5 = ((($1)) + 12|0); //@line 115 "modules/graphics/opengl/wrap_Video.cpp"
 $6 = +HEAPF32[$5>>2]; //@line 115 "modules/graphics/opengl/wrap_Video.cpp"
 HEAP32[$minstr>>2] = 0; //@line 117 "modules/graphics/opengl/wrap_Video.cpp"
 HEAP32[$magstr>>2] = 0; //@line 118 "modules/graphics/opengl/wrap_Video.cpp"
 $7 = (__ZN4love8graphics7Texture11getConstantENS1_10FilterModeERPKc($2,$minstr)|0); //@line 120 "modules/graphics/opengl/wrap_Video.cpp"
 if (!($7)) {
  $8 = (_luaL_error($L,347552,$vararg_buffer)|0); //@line 121 "modules/graphics/opengl/wrap_Video.cpp"
  $$0 = $8;
  STACKTOP = sp;return ($$0|0); //@line 129 "modules/graphics/opengl/wrap_Video.cpp"
 }
 $9 = (__ZN4love8graphics7Texture11getConstantENS1_10FilterModeERPKc($4,$magstr)|0); //@line 122 "modules/graphics/opengl/wrap_Video.cpp"
 if ($9) {
  $11 = HEAP32[$minstr>>2]|0; //@line 125 "modules/graphics/opengl/wrap_Video.cpp"
  _lua_pushstring($L,$11); //@line 125 "modules/graphics/opengl/wrap_Video.cpp"
  $12 = HEAP32[$magstr>>2]|0; //@line 126 "modules/graphics/opengl/wrap_Video.cpp"
  _lua_pushstring($L,$12); //@line 126 "modules/graphics/opengl/wrap_Video.cpp"
  $13 = $6; //@line 127 "modules/graphics/opengl/wrap_Video.cpp"
  _lua_pushnumber($L,$13); //@line 127 "modules/graphics/opengl/wrap_Video.cpp"
  $$0 = 3;
  STACKTOP = sp;return ($$0|0); //@line 129 "modules/graphics/opengl/wrap_Video.cpp"
 } else {
  $10 = (_luaL_error($L,347552,$vararg_buffer1)|0); //@line 123 "modules/graphics/opengl/wrap_Video.cpp"
  $$0 = $10;
  STACKTOP = sp;return ($$0|0); //@line 129 "modules/graphics/opengl/wrap_Video.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl13luaopen_videoEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 144 "modules/graphics/opengl/wrap_Video.cpp"
 HEAP32[$vararg_buffer>>2] = 15088; //@line 146 "modules/graphics/opengl/wrap_Video.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 146 "modules/graphics/opengl/wrap_Video.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 146 "modules/graphics/opengl/wrap_Video.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,21,345252,$vararg_buffer)|0); //@line 146 "modules/graphics/opengl/wrap_Video.cpp"
 (_luaL_loadbuffer($L,345289,1619,346908)|0); //@line 148 "modules/graphics/opengl/wrap_Video.cpp"
 __ZN4love21luax_gettypemetatableEP9lua_StateNS_4TypeE($L,21); //@line 149 "modules/graphics/opengl/wrap_Video.cpp"
 _lua_call($L,1,0); //@line 150 "modules/graphics/opengl/wrap_Video.cpp"
 STACKTOP = sp;return ($0|0); //@line 152 "modules/graphics/opengl/wrap_Video.cpp"
}
function __ZN4love8graphics8GraphicsD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ModuleD2Ev($this); //@line 63 "modules/graphics/Graphics.cpp"
 return; //@line 63 "modules/graphics/Graphics.cpp"
}
function __ZNK4love8graphics8Graphics13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 4; //@line 236 "modules/graphics/Graphics.h"
}
function __GLOBAL__sub_I_Graphics_cpp() {
 var $$02$i$i$1$i = 0, $$02$i$i$1$i22 = 0, $$02$i$i$1$i36 = 0, $$02$i$i$1$i8 = 0, $$02$i$i$i = 0, $$02$i$i$i1 = 0, $$02$i$i$i15 = 0, $$02$i$i$i29 = 0, $$lcssa$i$1$i = 0, $$lcssa$i$1$i13 = 0, $$lcssa$i$1$i27 = 0, $$lcssa$i$1$i41 = 0, $$lcssa$i$i = 0, $$lcssa$i$i20 = 0, $$lcssa$i$i34 = 0, $$lcssa$i$i6 = 0, $$lcssa10$i$1$i = 0, $$lcssa10$i$1$i12 = 0, $$lcssa10$i$1$i26 = 0, $$lcssa10$i$1$i40 = 0;
 var $$lcssa10$i$i = 0, $$lcssa10$i$i19 = 0, $$lcssa10$i$i33 = 0, $$lcssa10$i$i5 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0;
 var $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0;
 var $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0;
 var $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0;
 var $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0;
 var $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0;
 var $365 = 0, $366 = 0, $367 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $hash$0$lcssa$i$i$1$i = 0, $hash$0$lcssa$i$i$1$i11 = 0, $hash$0$lcssa$i$i$1$i25 = 0, $hash$0$lcssa$i$i$1$i39 = 0, $hash$0$lcssa$i$i$i = 0, $hash$0$lcssa$i$i$i18 = 0, $hash$0$lcssa$i$i$i32 = 0, $hash$0$lcssa$i$i$i4 = 0;
 var $hash$01$i$i$1$i = 0, $hash$01$i$i$1$i23 = 0, $hash$01$i$i$1$i37 = 0, $hash$01$i$i$1$i9 = 0, $hash$01$i$i$i = 0, $hash$01$i$i$i16 = 0, $hash$01$i$i$i2 = 0, $hash$01$i$i$i30 = 0, $vararg_buffer = 0, $vararg_buffer10 = 0, $vararg_buffer14 = 0, $vararg_buffer18 = 0, $vararg_buffer2 = 0, $vararg_buffer22 = 0, $vararg_buffer26 = 0, $vararg_buffer6 = 0, $vararg_ptr1 = 0, $vararg_ptr13 = 0, $vararg_ptr17 = 0, $vararg_ptr21 = 0;
 var $vararg_ptr25 = 0, $vararg_ptr29 = 0, $vararg_ptr5 = 0, $vararg_ptr9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer26 = sp + 56|0;
 $vararg_buffer22 = sp + 48|0;
 $vararg_buffer18 = sp + 40|0;
 $vararg_buffer14 = sp + 32|0;
 $vararg_buffer10 = sp + 24|0;
 $vararg_buffer6 = sp + 16|0;
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 HEAP8[(627764)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627776)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627788)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627800)>>0] = 0; //@line 155 "./common/StringMap.h"
 $0 = (627804); //@line 44 "./common/StringMap.h"
 $1 = $0; //@line 44 "./common/StringMap.h"
 HEAP32[$1>>2] = 0; //@line 44 "./common/StringMap.h"
 $2 = (($0) + 4)|0; //@line 44 "./common/StringMap.h"
 $3 = $2; //@line 44 "./common/StringMap.h"
 HEAP32[$3>>2] = 0; //@line 44 "./common/StringMap.h"
 $4 = HEAP32[3790]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(15164)>>2]|0; //@line 49 "./common/StringMap.h"
 $6 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
 $7 = ($6<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($7) {
  $hash$0$lcssa$i$i$i = 5381;
 } else {
  $$02$i$i$i = $4;$9 = $6;$hash$01$i$i$i = 5381;
  while(1) {
   $8 = $9 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $10 = ((($$02$i$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $11 = ($hash$01$i$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $12 = (($11) + ($8))|0; //@line 143 "./common/StringMap.h"
   $13 = HEAP8[$10>>0]|0; //@line 142 "./common/StringMap.h"
   $14 = ($13<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($14) {
    $hash$0$lcssa$i$i$i = $12;
    break;
   } else {
    $$02$i$i$i = $10;$9 = $13;$hash$01$i$i$i = $12;
   }
  }
 }
 $15 = $hash$0$lcssa$i$i$i & 3; //@line 112 "./common/StringMap.h"
 $16 = (((627756 + (($15*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $17 = HEAP8[$16>>0]|0; //@line 114 "./common/StringMap.h"
 $18 = ($17<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($18) {
  $$lcssa$i$i = $15;$$lcssa10$i$i = $16;
  label = 5;
 } else {
  $19 = (($hash$0$lcssa$i$i$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $20 = $19 & 3; //@line 112 "./common/StringMap.h"
  $21 = (((627756 + (($20*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $22 = HEAP8[$21>>0]|0; //@line 114 "./common/StringMap.h"
  $23 = ($22<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($23) {
   $$lcssa$i$i = $20;$$lcssa10$i$i = $21;
   label = 5;
  } else {
   $28 = (($hash$0$lcssa$i$i$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $29 = $28 & 3; //@line 112 "./common/StringMap.h"
   $30 = (((627756 + (($29*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $31 = HEAP8[$30>>0]|0; //@line 114 "./common/StringMap.h"
   $32 = ($31<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($32) {
    $$lcssa$i$i = $29;$$lcssa10$i$i = $30;
    label = 5;
   } else {
    $33 = (($hash$0$lcssa$i$i$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $34 = $33 & 3; //@line 112 "./common/StringMap.h"
    $35 = (((627756 + (($34*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $36 = HEAP8[$35>>0]|0; //@line 114 "./common/StringMap.h"
    $37 = ($36<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($37) {
     $$lcssa$i$i = $34;$$lcssa10$i$i = $35;
     label = 5;
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10$i$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $24 = (627756 + (($$lcssa$i$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$24>>2] = $4; //@line 118 "./common/StringMap.h"
  $25 = (((627756 + (($$lcssa$i$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$25>>2] = $5; //@line 119 "./common/StringMap.h"
 }
 $26 = ($5>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($26) {
  HEAP32[$vararg_buffer>>2] = $4; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $5; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
 } else {
  $27 = (627804 + ($5<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$27>>2] = $4; //@line 132 "./common/StringMap.h"
 }
 $38 = HEAP32[(15168)>>2]|0; //@line 49 "./common/StringMap.h"
 $39 = HEAP32[(15172)>>2]|0; //@line 49 "./common/StringMap.h"
 $40 = HEAP8[$38>>0]|0; //@line 142 "./common/StringMap.h"
 $41 = ($40<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($41) {
  $hash$0$lcssa$i$i$1$i = 5381;
 } else {
  $$02$i$i$1$i = $38;$43 = $40;$hash$01$i$i$1$i = 5381;
  while(1) {
   $42 = $43 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $44 = ((($$02$i$i$1$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $45 = ($hash$01$i$i$1$i*33)|0; //@line 143 "./common/StringMap.h"
   $46 = (($45) + ($42))|0; //@line 143 "./common/StringMap.h"
   $47 = HEAP8[$44>>0]|0; //@line 142 "./common/StringMap.h"
   $48 = ($47<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($48) {
    $hash$0$lcssa$i$i$1$i = $46;
    break;
   } else {
    $$02$i$i$1$i = $44;$43 = $47;$hash$01$i$i$1$i = $46;
   }
  }
 }
 $49 = $hash$0$lcssa$i$i$1$i & 3; //@line 112 "./common/StringMap.h"
 $50 = (((627756 + (($49*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $51 = HEAP8[$50>>0]|0; //@line 114 "./common/StringMap.h"
 $52 = ($51<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($52) {
  $$lcssa$i$1$i = $49;$$lcssa10$i$1$i = $50;
  label = 17;
 } else {
  $53 = (($hash$0$lcssa$i$i$1$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $54 = $53 & 3; //@line 112 "./common/StringMap.h"
  $55 = (((627756 + (($54*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $56 = HEAP8[$55>>0]|0; //@line 114 "./common/StringMap.h"
  $57 = ($56<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($57) {
   $$lcssa$i$1$i = $54;$$lcssa10$i$1$i = $55;
   label = 17;
  } else {
   $58 = (($hash$0$lcssa$i$i$1$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $59 = $58 & 3; //@line 112 "./common/StringMap.h"
   $60 = (((627756 + (($59*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $61 = HEAP8[$60>>0]|0; //@line 114 "./common/StringMap.h"
   $62 = ($61<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($62) {
    $$lcssa$i$1$i = $59;$$lcssa10$i$1$i = $60;
    label = 17;
   } else {
    $63 = (($hash$0$lcssa$i$i$1$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $64 = $63 & 3; //@line 112 "./common/StringMap.h"
    $65 = (((627756 + (($64*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $66 = HEAP8[$65>>0]|0; //@line 114 "./common/StringMap.h"
    $67 = ($66<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($67) {
     $$lcssa$i$1$i = $64;$$lcssa10$i$1$i = $65;
     label = 17;
    }
   }
  }
 }
 if ((label|0) == 17) {
  HEAP8[$$lcssa10$i$1$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $68 = (627756 + (($$lcssa$i$1$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$68>>2] = $38; //@line 118 "./common/StringMap.h"
  $69 = (((627756 + (($$lcssa$i$1$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$69>>2] = $39; //@line 119 "./common/StringMap.h"
 }
 $70 = ($39>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($70) {
  HEAP32[$vararg_buffer2>>2] = $38; //@line 128 "./common/StringMap.h"
  $vararg_ptr5 = ((($vararg_buffer2)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr5>>2] = $39; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer2)|0); //@line 128 "./common/StringMap.h"
 } else {
  $71 = (627804 + ($39<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$71>>2] = $38; //@line 132 "./common/StringMap.h"
 }
 HEAP8[(627820)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627832)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627844)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627856)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627868)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627880)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(627884)>>2]=0|0;HEAP32[(627884)+4>>2]=0|0;HEAP32[(627884)+8>>2]=0|0; //@line 44 "./common/StringMap.h"
 $72 = HEAP32[3794]|0; //@line 49 "./common/StringMap.h"
 $73 = HEAP32[(15180)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics7ArcModeELj3EE3addEPKcS3_(627812,$72,$73)|0); //@line 49 "./common/StringMap.h"
 $74 = HEAP32[(15184)>>2]|0; //@line 49 "./common/StringMap.h"
 $75 = HEAP32[(15188)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics7ArcModeELj3EE3addEPKcS3_(627812,$74,$75)|0); //@line 49 "./common/StringMap.h"
 $76 = HEAP32[(15192)>>2]|0; //@line 49 "./common/StringMap.h"
 $77 = HEAP32[(15196)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics7ArcModeELj3EE3addEPKcS3_(627812,$76,$77)|0); //@line 49 "./common/StringMap.h"
 HEAP8[(627904)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627916)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627928)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627940)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627952)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627964)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627976)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(627988)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628000)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628012)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628024)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628036)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628048)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628060)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628072)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628084)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(628088)>>2]=0|0;HEAP32[(628088)+4>>2]=0|0;HEAP32[(628088)+8>>2]=0|0;HEAP32[(628088)+12>>2]=0|0;HEAP32[(628088)+16>>2]=0|0;HEAP32[(628088)+20>>2]=0|0;HEAP32[(628088)+24>>2]=0|0;HEAP32[(628088)+28>>2]=0|0; //@line 44 "./common/StringMap.h"
 $78 = HEAP32[3800]|0; //@line 49 "./common/StringMap.h"
 $79 = HEAP32[(15204)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics9BlendModeELj8EE3addEPKcS3_(627896,$78,$79)|0); //@line 49 "./common/StringMap.h"
 $80 = HEAP32[(15208)>>2]|0; //@line 49 "./common/StringMap.h"
 $81 = HEAP32[(15212)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics9BlendModeELj8EE3addEPKcS3_(627896,$80,$81)|0); //@line 49 "./common/StringMap.h"
 $82 = HEAP32[(15216)>>2]|0; //@line 49 "./common/StringMap.h"
 $83 = HEAP32[(15220)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics9BlendModeELj8EE3addEPKcS3_(627896,$82,$83)|0); //@line 49 "./common/StringMap.h"
 $84 = HEAP32[(15224)>>2]|0; //@line 49 "./common/StringMap.h"
 $85 = HEAP32[(15228)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics9BlendModeELj8EE3addEPKcS3_(627896,$84,$85)|0); //@line 49 "./common/StringMap.h"
 $86 = HEAP32[(15232)>>2]|0; //@line 49 "./common/StringMap.h"
 $87 = HEAP32[(15236)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics9BlendModeELj8EE3addEPKcS3_(627896,$86,$87)|0); //@line 49 "./common/StringMap.h"
 $88 = HEAP32[(15240)>>2]|0; //@line 49 "./common/StringMap.h"
 $89 = HEAP32[(15244)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics9BlendModeELj8EE3addEPKcS3_(627896,$88,$89)|0); //@line 49 "./common/StringMap.h"
 $90 = HEAP32[(15248)>>2]|0; //@line 49 "./common/StringMap.h"
 $91 = HEAP32[(15252)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics9BlendModeELj8EE3addEPKcS3_(627896,$90,$91)|0); //@line 49 "./common/StringMap.h"
 $92 = HEAP32[(15256)>>2]|0; //@line 49 "./common/StringMap.h"
 $93 = HEAP32[(15260)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics9BlendModeELj8EE3addEPKcS3_(627896,$92,$93)|0); //@line 49 "./common/StringMap.h"
 HEAP8[(628128)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628140)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628152)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628164)>>0] = 0; //@line 155 "./common/StringMap.h"
 $94 = (628168); //@line 44 "./common/StringMap.h"
 $95 = $94; //@line 44 "./common/StringMap.h"
 HEAP32[$95>>2] = 0; //@line 44 "./common/StringMap.h"
 $96 = (($94) + 4)|0; //@line 44 "./common/StringMap.h"
 $97 = $96; //@line 44 "./common/StringMap.h"
 HEAP32[$97>>2] = 0; //@line 44 "./common/StringMap.h"
 $98 = HEAP32[3816]|0; //@line 49 "./common/StringMap.h"
 $99 = HEAP32[(15268)>>2]|0; //@line 49 "./common/StringMap.h"
 $100 = HEAP8[$98>>0]|0; //@line 142 "./common/StringMap.h"
 $101 = ($100<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($101) {
  $hash$0$lcssa$i$i$i4 = 5381;
 } else {
  $$02$i$i$i1 = $98;$103 = $100;$hash$01$i$i$i2 = 5381;
  while(1) {
   $102 = $103 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $104 = ((($$02$i$i$i1)) + 1|0); //@line 142 "./common/StringMap.h"
   $105 = ($hash$01$i$i$i2*33)|0; //@line 143 "./common/StringMap.h"
   $106 = (($105) + ($102))|0; //@line 143 "./common/StringMap.h"
   $107 = HEAP8[$104>>0]|0; //@line 142 "./common/StringMap.h"
   $108 = ($107<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($108) {
    $hash$0$lcssa$i$i$i4 = $106;
    break;
   } else {
    $$02$i$i$i1 = $104;$103 = $107;$hash$01$i$i$i2 = $106;
   }
  }
 }
 $109 = $hash$0$lcssa$i$i$i4 & 3; //@line 112 "./common/StringMap.h"
 $110 = (((628120 + (($109*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $111 = HEAP8[$110>>0]|0; //@line 114 "./common/StringMap.h"
 $112 = ($111<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($112) {
  $$lcssa$i$i6 = $109;$$lcssa10$i$i5 = $110;
  label = 25;
 } else {
  $113 = (($hash$0$lcssa$i$i$i4) + 1)|0; //@line 112 "./common/StringMap.h"
  $114 = $113 & 3; //@line 112 "./common/StringMap.h"
  $115 = (((628120 + (($114*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $116 = HEAP8[$115>>0]|0; //@line 114 "./common/StringMap.h"
  $117 = ($116<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($117) {
   $$lcssa$i$i6 = $114;$$lcssa10$i$i5 = $115;
   label = 25;
  } else {
   $122 = (($hash$0$lcssa$i$i$i4) + 2)|0; //@line 112 "./common/StringMap.h"
   $123 = $122 & 3; //@line 112 "./common/StringMap.h"
   $124 = (((628120 + (($123*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $125 = HEAP8[$124>>0]|0; //@line 114 "./common/StringMap.h"
   $126 = ($125<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($126) {
    $$lcssa$i$i6 = $123;$$lcssa10$i$i5 = $124;
    label = 25;
   } else {
    $127 = (($hash$0$lcssa$i$i$i4) + 3)|0; //@line 112 "./common/StringMap.h"
    $128 = $127 & 3; //@line 112 "./common/StringMap.h"
    $129 = (((628120 + (($128*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $130 = HEAP8[$129>>0]|0; //@line 114 "./common/StringMap.h"
    $131 = ($130<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($131) {
     $$lcssa$i$i6 = $128;$$lcssa10$i$i5 = $129;
     label = 25;
    }
   }
  }
 }
 if ((label|0) == 25) {
  HEAP8[$$lcssa10$i$i5>>0] = 1; //@line 117 "./common/StringMap.h"
  $118 = (628120 + (($$lcssa$i$i6*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$118>>2] = $98; //@line 118 "./common/StringMap.h"
  $119 = (((628120 + (($$lcssa$i$i6*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$119>>2] = $99; //@line 119 "./common/StringMap.h"
 }
 $120 = ($99>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($120) {
  HEAP32[$vararg_buffer6>>2] = $98; //@line 128 "./common/StringMap.h"
  $vararg_ptr9 = ((($vararg_buffer6)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr9>>2] = $99; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer6)|0); //@line 128 "./common/StringMap.h"
 } else {
  $121 = (628168 + ($99<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$121>>2] = $98; //@line 132 "./common/StringMap.h"
 }
 $132 = HEAP32[(15272)>>2]|0; //@line 49 "./common/StringMap.h"
 $133 = HEAP32[(15276)>>2]|0; //@line 49 "./common/StringMap.h"
 $134 = HEAP8[$132>>0]|0; //@line 142 "./common/StringMap.h"
 $135 = ($134<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($135) {
  $hash$0$lcssa$i$i$1$i11 = 5381;
 } else {
  $$02$i$i$1$i8 = $132;$137 = $134;$hash$01$i$i$1$i9 = 5381;
  while(1) {
   $136 = $137 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $138 = ((($$02$i$i$1$i8)) + 1|0); //@line 142 "./common/StringMap.h"
   $139 = ($hash$01$i$i$1$i9*33)|0; //@line 143 "./common/StringMap.h"
   $140 = (($139) + ($136))|0; //@line 143 "./common/StringMap.h"
   $141 = HEAP8[$138>>0]|0; //@line 142 "./common/StringMap.h"
   $142 = ($141<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($142) {
    $hash$0$lcssa$i$i$1$i11 = $140;
    break;
   } else {
    $$02$i$i$1$i8 = $138;$137 = $141;$hash$01$i$i$1$i9 = $140;
   }
  }
 }
 $143 = $hash$0$lcssa$i$i$1$i11 & 3; //@line 112 "./common/StringMap.h"
 $144 = (((628120 + (($143*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $145 = HEAP8[$144>>0]|0; //@line 114 "./common/StringMap.h"
 $146 = ($145<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($146) {
  $$lcssa$i$1$i13 = $143;$$lcssa10$i$1$i12 = $144;
  label = 37;
 } else {
  $147 = (($hash$0$lcssa$i$i$1$i11) + 1)|0; //@line 112 "./common/StringMap.h"
  $148 = $147 & 3; //@line 112 "./common/StringMap.h"
  $149 = (((628120 + (($148*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $150 = HEAP8[$149>>0]|0; //@line 114 "./common/StringMap.h"
  $151 = ($150<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($151) {
   $$lcssa$i$1$i13 = $148;$$lcssa10$i$1$i12 = $149;
   label = 37;
  } else {
   $152 = (($hash$0$lcssa$i$i$1$i11) + 2)|0; //@line 112 "./common/StringMap.h"
   $153 = $152 & 3; //@line 112 "./common/StringMap.h"
   $154 = (((628120 + (($153*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $155 = HEAP8[$154>>0]|0; //@line 114 "./common/StringMap.h"
   $156 = ($155<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($156) {
    $$lcssa$i$1$i13 = $153;$$lcssa10$i$1$i12 = $154;
    label = 37;
   } else {
    $157 = (($hash$0$lcssa$i$i$1$i11) + 3)|0; //@line 112 "./common/StringMap.h"
    $158 = $157 & 3; //@line 112 "./common/StringMap.h"
    $159 = (((628120 + (($158*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $160 = HEAP8[$159>>0]|0; //@line 114 "./common/StringMap.h"
    $161 = ($160<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($161) {
     $$lcssa$i$1$i13 = $158;$$lcssa10$i$1$i12 = $159;
     label = 37;
    }
   }
  }
 }
 if ((label|0) == 37) {
  HEAP8[$$lcssa10$i$1$i12>>0] = 1; //@line 117 "./common/StringMap.h"
  $162 = (628120 + (($$lcssa$i$1$i13*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$162>>2] = $132; //@line 118 "./common/StringMap.h"
  $163 = (((628120 + (($$lcssa$i$1$i13*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$163>>2] = $133; //@line 119 "./common/StringMap.h"
 }
 $164 = ($133>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($164) {
  HEAP32[$vararg_buffer10>>2] = $132; //@line 128 "./common/StringMap.h"
  $vararg_ptr13 = ((($vararg_buffer10)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr13>>2] = $133; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer10)|0); //@line 128 "./common/StringMap.h"
 } else {
  $165 = (628168 + ($133<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$165>>2] = $132; //@line 132 "./common/StringMap.h"
 }
 HEAP8[(628184)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628196)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628208)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628220)>>0] = 0; //@line 155 "./common/StringMap.h"
 $166 = (628224); //@line 44 "./common/StringMap.h"
 $167 = $166; //@line 44 "./common/StringMap.h"
 HEAP32[$167>>2] = 0; //@line 44 "./common/StringMap.h"
 $168 = (($166) + 4)|0; //@line 44 "./common/StringMap.h"
 $169 = $168; //@line 44 "./common/StringMap.h"
 HEAP32[$169>>2] = 0; //@line 44 "./common/StringMap.h"
 $170 = HEAP32[3820]|0; //@line 49 "./common/StringMap.h"
 $171 = HEAP32[(15284)>>2]|0; //@line 49 "./common/StringMap.h"
 $172 = HEAP8[$170>>0]|0; //@line 142 "./common/StringMap.h"
 $173 = ($172<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($173) {
  $hash$0$lcssa$i$i$i18 = 5381;
 } else {
  $$02$i$i$i15 = $170;$175 = $172;$hash$01$i$i$i16 = 5381;
  while(1) {
   $174 = $175 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $176 = ((($$02$i$i$i15)) + 1|0); //@line 142 "./common/StringMap.h"
   $177 = ($hash$01$i$i$i16*33)|0; //@line 143 "./common/StringMap.h"
   $178 = (($177) + ($174))|0; //@line 143 "./common/StringMap.h"
   $179 = HEAP8[$176>>0]|0; //@line 142 "./common/StringMap.h"
   $180 = ($179<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($180) {
    $hash$0$lcssa$i$i$i18 = $178;
    break;
   } else {
    $$02$i$i$i15 = $176;$175 = $179;$hash$01$i$i$i16 = $178;
   }
  }
 }
 $181 = $hash$0$lcssa$i$i$i18 & 3; //@line 112 "./common/StringMap.h"
 $182 = (((628176 + (($181*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $183 = HEAP8[$182>>0]|0; //@line 114 "./common/StringMap.h"
 $184 = ($183<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($184) {
  $$lcssa$i$i20 = $181;$$lcssa10$i$i19 = $182;
  label = 45;
 } else {
  $185 = (($hash$0$lcssa$i$i$i18) + 1)|0; //@line 112 "./common/StringMap.h"
  $186 = $185 & 3; //@line 112 "./common/StringMap.h"
  $187 = (((628176 + (($186*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $188 = HEAP8[$187>>0]|0; //@line 114 "./common/StringMap.h"
  $189 = ($188<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($189) {
   $$lcssa$i$i20 = $186;$$lcssa10$i$i19 = $187;
   label = 45;
  } else {
   $194 = (($hash$0$lcssa$i$i$i18) + 2)|0; //@line 112 "./common/StringMap.h"
   $195 = $194 & 3; //@line 112 "./common/StringMap.h"
   $196 = (((628176 + (($195*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $197 = HEAP8[$196>>0]|0; //@line 114 "./common/StringMap.h"
   $198 = ($197<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($198) {
    $$lcssa$i$i20 = $195;$$lcssa10$i$i19 = $196;
    label = 45;
   } else {
    $199 = (($hash$0$lcssa$i$i$i18) + 3)|0; //@line 112 "./common/StringMap.h"
    $200 = $199 & 3; //@line 112 "./common/StringMap.h"
    $201 = (((628176 + (($200*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $202 = HEAP8[$201>>0]|0; //@line 114 "./common/StringMap.h"
    $203 = ($202<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($203) {
     $$lcssa$i$i20 = $200;$$lcssa10$i$i19 = $201;
     label = 45;
    }
   }
  }
 }
 if ((label|0) == 45) {
  HEAP8[$$lcssa10$i$i19>>0] = 1; //@line 117 "./common/StringMap.h"
  $190 = (628176 + (($$lcssa$i$i20*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$190>>2] = $170; //@line 118 "./common/StringMap.h"
  $191 = (((628176 + (($$lcssa$i$i20*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$191>>2] = $171; //@line 119 "./common/StringMap.h"
 }
 $192 = ($171>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($192) {
  HEAP32[$vararg_buffer14>>2] = $170; //@line 128 "./common/StringMap.h"
  $vararg_ptr17 = ((($vararg_buffer14)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr17>>2] = $171; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer14)|0); //@line 128 "./common/StringMap.h"
 } else {
  $193 = (628224 + ($171<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$193>>2] = $170; //@line 132 "./common/StringMap.h"
 }
 $204 = HEAP32[(15288)>>2]|0; //@line 49 "./common/StringMap.h"
 $205 = HEAP32[(15292)>>2]|0; //@line 49 "./common/StringMap.h"
 $206 = HEAP8[$204>>0]|0; //@line 142 "./common/StringMap.h"
 $207 = ($206<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($207) {
  $hash$0$lcssa$i$i$1$i25 = 5381;
 } else {
  $$02$i$i$1$i22 = $204;$209 = $206;$hash$01$i$i$1$i23 = 5381;
  while(1) {
   $208 = $209 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $210 = ((($$02$i$i$1$i22)) + 1|0); //@line 142 "./common/StringMap.h"
   $211 = ($hash$01$i$i$1$i23*33)|0; //@line 143 "./common/StringMap.h"
   $212 = (($211) + ($208))|0; //@line 143 "./common/StringMap.h"
   $213 = HEAP8[$210>>0]|0; //@line 142 "./common/StringMap.h"
   $214 = ($213<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($214) {
    $hash$0$lcssa$i$i$1$i25 = $212;
    break;
   } else {
    $$02$i$i$1$i22 = $210;$209 = $213;$hash$01$i$i$1$i23 = $212;
   }
  }
 }
 $215 = $hash$0$lcssa$i$i$1$i25 & 3; //@line 112 "./common/StringMap.h"
 $216 = (((628176 + (($215*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $217 = HEAP8[$216>>0]|0; //@line 114 "./common/StringMap.h"
 $218 = ($217<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($218) {
  $$lcssa$i$1$i27 = $215;$$lcssa10$i$1$i26 = $216;
  label = 57;
 } else {
  $219 = (($hash$0$lcssa$i$i$1$i25) + 1)|0; //@line 112 "./common/StringMap.h"
  $220 = $219 & 3; //@line 112 "./common/StringMap.h"
  $221 = (((628176 + (($220*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $222 = HEAP8[$221>>0]|0; //@line 114 "./common/StringMap.h"
  $223 = ($222<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($223) {
   $$lcssa$i$1$i27 = $220;$$lcssa10$i$1$i26 = $221;
   label = 57;
  } else {
   $224 = (($hash$0$lcssa$i$i$1$i25) + 2)|0; //@line 112 "./common/StringMap.h"
   $225 = $224 & 3; //@line 112 "./common/StringMap.h"
   $226 = (((628176 + (($225*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $227 = HEAP8[$226>>0]|0; //@line 114 "./common/StringMap.h"
   $228 = ($227<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($228) {
    $$lcssa$i$1$i27 = $225;$$lcssa10$i$1$i26 = $226;
    label = 57;
   } else {
    $229 = (($hash$0$lcssa$i$i$1$i25) + 3)|0; //@line 112 "./common/StringMap.h"
    $230 = $229 & 3; //@line 112 "./common/StringMap.h"
    $231 = (((628176 + (($230*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $232 = HEAP8[$231>>0]|0; //@line 114 "./common/StringMap.h"
    $233 = ($232<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($233) {
     $$lcssa$i$1$i27 = $230;$$lcssa10$i$1$i26 = $231;
     label = 57;
    }
   }
  }
 }
 if ((label|0) == 57) {
  HEAP8[$$lcssa10$i$1$i26>>0] = 1; //@line 117 "./common/StringMap.h"
  $234 = (628176 + (($$lcssa$i$1$i27*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$234>>2] = $204; //@line 118 "./common/StringMap.h"
  $235 = (((628176 + (($$lcssa$i$1$i27*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$235>>2] = $205; //@line 119 "./common/StringMap.h"
 }
 $236 = ($205>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($236) {
  HEAP32[$vararg_buffer18>>2] = $204; //@line 128 "./common/StringMap.h"
  $vararg_ptr21 = ((($vararg_buffer18)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr21>>2] = $205; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer18)|0); //@line 128 "./common/StringMap.h"
 } else {
  $237 = (628224 + ($205<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$237>>2] = $204; //@line 132 "./common/StringMap.h"
 }
 HEAP8[(628240)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628252)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628264)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628276)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628288)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628300)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(628304)>>2]=0|0;HEAP32[(628304)+4>>2]=0|0;HEAP32[(628304)+8>>2]=0|0; //@line 44 "./common/StringMap.h"
 $238 = HEAP32[3824]|0; //@line 49 "./common/StringMap.h"
 $239 = HEAP32[(15300)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics8LineJoinELj3EE3addEPKcS3_(628232,$238,$239)|0); //@line 49 "./common/StringMap.h"
 $240 = HEAP32[(15304)>>2]|0; //@line 49 "./common/StringMap.h"
 $241 = HEAP32[(15308)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics8LineJoinELj3EE3addEPKcS3_(628232,$240,$241)|0); //@line 49 "./common/StringMap.h"
 $242 = HEAP32[(15312)>>2]|0; //@line 49 "./common/StringMap.h"
 $243 = HEAP32[(15316)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics8LineJoinELj3EE3addEPKcS3_(628232,$242,$243)|0); //@line 49 "./common/StringMap.h"
 HEAP8[(628324)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628336)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628348)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628360)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628372)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628384)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628396)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628408)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628420)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628432)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628444)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628456)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(628460)>>2]=0|0;HEAP32[(628460)+4>>2]=0|0;HEAP32[(628460)+8>>2]=0|0;HEAP32[(628460)+12>>2]=0|0;HEAP32[(628460)+16>>2]=0|0;HEAP32[(628460)+20>>2]=0|0; //@line 44 "./common/StringMap.h"
 $244 = HEAP32[3830]|0; //@line 49 "./common/StringMap.h"
 $245 = HEAP32[(15324)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics13StencilActionELj6EE3addEPKcS3_(628316,$244,$245)|0); //@line 49 "./common/StringMap.h"
 $246 = HEAP32[(15328)>>2]|0; //@line 49 "./common/StringMap.h"
 $247 = HEAP32[(15332)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics13StencilActionELj6EE3addEPKcS3_(628316,$246,$247)|0); //@line 49 "./common/StringMap.h"
 $248 = HEAP32[(15336)>>2]|0; //@line 49 "./common/StringMap.h"
 $249 = HEAP32[(15340)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics13StencilActionELj6EE3addEPKcS3_(628316,$248,$249)|0); //@line 49 "./common/StringMap.h"
 $250 = HEAP32[(15344)>>2]|0; //@line 49 "./common/StringMap.h"
 $251 = HEAP32[(15348)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics13StencilActionELj6EE3addEPKcS3_(628316,$250,$251)|0); //@line 49 "./common/StringMap.h"
 $252 = HEAP32[(15352)>>2]|0; //@line 49 "./common/StringMap.h"
 $253 = HEAP32[(15356)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics13StencilActionELj6EE3addEPKcS3_(628316,$252,$253)|0); //@line 49 "./common/StringMap.h"
 $254 = HEAP32[(15360)>>2]|0; //@line 49 "./common/StringMap.h"
 $255 = HEAP32[(15364)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics13StencilActionELj6EE3addEPKcS3_(628316,$254,$255)|0); //@line 49 "./common/StringMap.h"
 HEAP8[(628492)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628504)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628516)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628528)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628540)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628552)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628564)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628576)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628588)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628600)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628612)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628624)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628636)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628648)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(628652)>>2]=0|0;HEAP32[(628652)+4>>2]=0|0;HEAP32[(628652)+8>>2]=0|0;HEAP32[(628652)+12>>2]=0|0;HEAP32[(628652)+16>>2]=0|0;HEAP32[(628652)+20>>2]=0|0;HEAP32[(628652)+24>>2]=0|0; //@line 44 "./common/StringMap.h"
 $256 = HEAP32[3842]|0; //@line 49 "./common/StringMap.h"
 $257 = HEAP32[(15372)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics11CompareModeELj7EE3addEPKcS3_(628484,$256,$257)|0); //@line 49 "./common/StringMap.h"
 $258 = HEAP32[(15376)>>2]|0; //@line 49 "./common/StringMap.h"
 $259 = HEAP32[(15380)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics11CompareModeELj7EE3addEPKcS3_(628484,$258,$259)|0); //@line 49 "./common/StringMap.h"
 $260 = HEAP32[(15384)>>2]|0; //@line 49 "./common/StringMap.h"
 $261 = HEAP32[(15388)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics11CompareModeELj7EE3addEPKcS3_(628484,$260,$261)|0); //@line 49 "./common/StringMap.h"
 $262 = HEAP32[(15392)>>2]|0; //@line 49 "./common/StringMap.h"
 $263 = HEAP32[(15396)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics11CompareModeELj7EE3addEPKcS3_(628484,$262,$263)|0); //@line 49 "./common/StringMap.h"
 $264 = HEAP32[(15400)>>2]|0; //@line 49 "./common/StringMap.h"
 $265 = HEAP32[(15404)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics11CompareModeELj7EE3addEPKcS3_(628484,$264,$265)|0); //@line 49 "./common/StringMap.h"
 $266 = HEAP32[(15408)>>2]|0; //@line 49 "./common/StringMap.h"
 $267 = HEAP32[(15412)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics11CompareModeELj7EE3addEPKcS3_(628484,$266,$267)|0); //@line 49 "./common/StringMap.h"
 $268 = HEAP32[(15416)>>2]|0; //@line 49 "./common/StringMap.h"
 $269 = HEAP32[(15420)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics11CompareModeELj7EE3addEPKcS3_(628484,$268,$269)|0); //@line 49 "./common/StringMap.h"
 HEAP8[(628688)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628700)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628712)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628724)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628736)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628748)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(628752)>>2]=0|0;HEAP32[(628752)+4>>2]=0|0;HEAP32[(628752)+8>>2]=0|0; //@line 44 "./common/StringMap.h"
 $270 = HEAP32[3856]|0; //@line 49 "./common/StringMap.h"
 $271 = HEAP32[(15428)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics7SupportELj3EE3addEPKcS3_(628680,$270,$271)|0); //@line 49 "./common/StringMap.h"
 $272 = HEAP32[(15432)>>2]|0; //@line 49 "./common/StringMap.h"
 $273 = HEAP32[(15436)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics7SupportELj3EE3addEPKcS3_(628680,$272,$273)|0); //@line 49 "./common/StringMap.h"
 $274 = HEAP32[(15440)>>2]|0; //@line 49 "./common/StringMap.h"
 $275 = HEAP32[(15444)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics7SupportELj3EE3addEPKcS3_(628680,$274,$275)|0); //@line 49 "./common/StringMap.h"
 HEAP8[(628772)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628784)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628796)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628808)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628820)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628832)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628844)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628856)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(628860)>>2]=0|0;HEAP32[(628860)+4>>2]=0|0;HEAP32[(628860)+8>>2]=0|0;HEAP32[(628860)+12>>2]=0|0; //@line 44 "./common/StringMap.h"
 $276 = HEAP32[3862]|0; //@line 49 "./common/StringMap.h"
 $277 = HEAP32[(15452)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics11SystemLimitELj4EE3addEPKcS3_(628764,$276,$277)|0); //@line 49 "./common/StringMap.h"
 $278 = HEAP32[(15456)>>2]|0; //@line 49 "./common/StringMap.h"
 $279 = HEAP32[(15460)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics11SystemLimitELj4EE3addEPKcS3_(628764,$278,$279)|0); //@line 49 "./common/StringMap.h"
 $280 = HEAP32[(15464)>>2]|0; //@line 49 "./common/StringMap.h"
 $281 = HEAP32[(15468)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics11SystemLimitELj4EE3addEPKcS3_(628764,$280,$281)|0); //@line 49 "./common/StringMap.h"
 $282 = HEAP32[(15472)>>2]|0; //@line 49 "./common/StringMap.h"
 $283 = HEAP32[(15476)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics8Graphics11SystemLimitELj4EE3addEPKcS3_(628764,$282,$283)|0); //@line 49 "./common/StringMap.h"
 HEAP8[(628884)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628896)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628908)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(628920)>>0] = 0; //@line 155 "./common/StringMap.h"
 $284 = (628924); //@line 44 "./common/StringMap.h"
 $285 = $284; //@line 44 "./common/StringMap.h"
 HEAP32[$285>>2] = 0; //@line 44 "./common/StringMap.h"
 $286 = (($284) + 4)|0; //@line 44 "./common/StringMap.h"
 $287 = $286; //@line 44 "./common/StringMap.h"
 HEAP32[$287>>2] = 0; //@line 44 "./common/StringMap.h"
 $288 = HEAP32[3870]|0; //@line 49 "./common/StringMap.h"
 $289 = HEAP32[(15484)>>2]|0; //@line 49 "./common/StringMap.h"
 $290 = HEAP8[$288>>0]|0; //@line 142 "./common/StringMap.h"
 $291 = ($290<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($291) {
  $hash$0$lcssa$i$i$i32 = 5381;
 } else {
  $$02$i$i$i29 = $288;$293 = $290;$hash$01$i$i$i30 = 5381;
  while(1) {
   $292 = $293 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $294 = ((($$02$i$i$i29)) + 1|0); //@line 142 "./common/StringMap.h"
   $295 = ($hash$01$i$i$i30*33)|0; //@line 143 "./common/StringMap.h"
   $296 = (($295) + ($292))|0; //@line 143 "./common/StringMap.h"
   $297 = HEAP8[$294>>0]|0; //@line 142 "./common/StringMap.h"
   $298 = ($297<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($298) {
    $hash$0$lcssa$i$i$i32 = $296;
    break;
   } else {
    $$02$i$i$i29 = $294;$293 = $297;$hash$01$i$i$i30 = $296;
   }
  }
 }
 $299 = $hash$0$lcssa$i$i$i32 & 3; //@line 112 "./common/StringMap.h"
 $300 = (((628876 + (($299*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $301 = HEAP8[$300>>0]|0; //@line 114 "./common/StringMap.h"
 $302 = ($301<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 do {
  if ($302) {
   $$lcssa$i$i34 = $299;$$lcssa10$i$i33 = $300;
   label = 65;
  } else {
   $303 = (($hash$0$lcssa$i$i$i32) + 1)|0; //@line 112 "./common/StringMap.h"
   $304 = $303 & 3; //@line 112 "./common/StringMap.h"
   $305 = (((628876 + (($304*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $306 = HEAP8[$305>>0]|0; //@line 114 "./common/StringMap.h"
   $307 = ($306<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($307) {
    $$lcssa$i$i34 = $304;$$lcssa10$i$i33 = $305;
    label = 65;
    break;
   }
   $312 = (($hash$0$lcssa$i$i$i32) + 2)|0; //@line 112 "./common/StringMap.h"
   $313 = $312 & 3; //@line 112 "./common/StringMap.h"
   $314 = (((628876 + (($313*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $315 = HEAP8[$314>>0]|0; //@line 114 "./common/StringMap.h"
   $316 = ($315<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($316) {
    $$lcssa$i$i34 = $313;$$lcssa10$i$i33 = $314;
    label = 65;
    break;
   }
   $317 = (($hash$0$lcssa$i$i$i32) + 3)|0; //@line 112 "./common/StringMap.h"
   $318 = $317 & 3; //@line 112 "./common/StringMap.h"
   $319 = (((628876 + (($318*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $320 = HEAP8[$319>>0]|0; //@line 114 "./common/StringMap.h"
   $321 = ($320<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($321) {
    $$lcssa$i$i34 = $318;$$lcssa10$i$i33 = $319;
    label = 65;
   }
  }
 } while(0);
 if ((label|0) == 65) {
  HEAP8[$$lcssa10$i$i33>>0] = 1; //@line 117 "./common/StringMap.h"
  $308 = (628876 + (($$lcssa$i$i34*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$308>>2] = $288; //@line 118 "./common/StringMap.h"
  $309 = (((628876 + (($$lcssa$i$i34*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$309>>2] = $289; //@line 119 "./common/StringMap.h"
 }
 $310 = ($289>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($310) {
  HEAP32[$vararg_buffer22>>2] = $288; //@line 128 "./common/StringMap.h"
  $vararg_ptr25 = ((($vararg_buffer22)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr25>>2] = $289; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer22)|0); //@line 128 "./common/StringMap.h"
 } else {
  $311 = (628924 + ($289<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$311>>2] = $288; //@line 132 "./common/StringMap.h"
 }
 $322 = HEAP32[(15488)>>2]|0; //@line 49 "./common/StringMap.h"
 $323 = HEAP32[(15492)>>2]|0; //@line 49 "./common/StringMap.h"
 $324 = HEAP8[$322>>0]|0; //@line 142 "./common/StringMap.h"
 $325 = ($324<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($325) {
  $hash$0$lcssa$i$i$1$i39 = 5381;
 } else {
  $$02$i$i$1$i36 = $322;$327 = $324;$hash$01$i$i$1$i37 = 5381;
  while(1) {
   $326 = $327 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $328 = ((($$02$i$i$1$i36)) + 1|0); //@line 142 "./common/StringMap.h"
   $329 = ($hash$01$i$i$1$i37*33)|0; //@line 143 "./common/StringMap.h"
   $330 = (($329) + ($326))|0; //@line 143 "./common/StringMap.h"
   $331 = HEAP8[$328>>0]|0; //@line 142 "./common/StringMap.h"
   $332 = ($331<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($332) {
    $hash$0$lcssa$i$i$1$i39 = $330;
    break;
   } else {
    $$02$i$i$1$i36 = $328;$327 = $331;$hash$01$i$i$1$i37 = $330;
   }
  }
 }
 $333 = $hash$0$lcssa$i$i$1$i39 & 3; //@line 112 "./common/StringMap.h"
 $334 = (((628876 + (($333*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $335 = HEAP8[$334>>0]|0; //@line 114 "./common/StringMap.h"
 $336 = ($335<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 do {
  if ($336) {
   $$lcssa$i$1$i41 = $333;$$lcssa10$i$1$i40 = $334;
   label = 77;
  } else {
   $337 = (($hash$0$lcssa$i$i$1$i39) + 1)|0; //@line 112 "./common/StringMap.h"
   $338 = $337 & 3; //@line 112 "./common/StringMap.h"
   $339 = (((628876 + (($338*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $340 = HEAP8[$339>>0]|0; //@line 114 "./common/StringMap.h"
   $341 = ($340<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($341) {
    $$lcssa$i$1$i41 = $338;$$lcssa10$i$1$i40 = $339;
    label = 77;
    break;
   }
   $342 = (($hash$0$lcssa$i$i$1$i39) + 2)|0; //@line 112 "./common/StringMap.h"
   $343 = $342 & 3; //@line 112 "./common/StringMap.h"
   $344 = (((628876 + (($343*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $345 = HEAP8[$344>>0]|0; //@line 114 "./common/StringMap.h"
   $346 = ($345<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($346) {
    $$lcssa$i$1$i41 = $343;$$lcssa10$i$1$i40 = $344;
    label = 77;
    break;
   }
   $347 = (($hash$0$lcssa$i$i$1$i39) + 3)|0; //@line 112 "./common/StringMap.h"
   $348 = $347 & 3; //@line 112 "./common/StringMap.h"
   $349 = (((628876 + (($348*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $350 = HEAP8[$349>>0]|0; //@line 114 "./common/StringMap.h"
   $351 = ($350<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($351) {
    $$lcssa$i$1$i41 = $348;$$lcssa10$i$1$i40 = $349;
    label = 77;
   }
  }
 } while(0);
 if ((label|0) == 77) {
  HEAP8[$$lcssa10$i$1$i40>>0] = 1; //@line 117 "./common/StringMap.h"
  $352 = (628876 + (($$lcssa$i$1$i41*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$352>>2] = $322; //@line 118 "./common/StringMap.h"
  $353 = (((628876 + (($$lcssa$i$1$i41*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$353>>2] = $323; //@line 119 "./common/StringMap.h"
 }
 $354 = ($323>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($354) {
  HEAP32[$vararg_buffer26>>2] = $322; //@line 128 "./common/StringMap.h"
  $vararg_ptr29 = ((($vararg_buffer26)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr29>>2] = $323; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer26)|0); //@line 128 "./common/StringMap.h"
  HEAP8[(628940)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(628952)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(628964)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(628976)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(628988)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(629000)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(629012)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(629024)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(629036)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(629048)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(629060)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(629072)>>0] = 0; //@line 155 "./common/StringMap.h"
  ;HEAP32[(629076)>>2]=0|0;HEAP32[(629076)+4>>2]=0|0;HEAP32[(629076)+8>>2]=0|0;HEAP32[(629076)+12>>2]=0|0;HEAP32[(629076)+16>>2]=0|0;HEAP32[(629076)+20>>2]=0|0; //@line 44 "./common/StringMap.h"
  $356 = HEAP32[3874]|0; //@line 49 "./common/StringMap.h"
  $357 = HEAP32[(15500)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_8graphics8Graphics8StatTypeELj6EE3addEPKcS3_(628932,$356,$357)|0); //@line 49 "./common/StringMap.h"
  $358 = HEAP32[(15504)>>2]|0; //@line 49 "./common/StringMap.h"
  $359 = HEAP32[(15508)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_8graphics8Graphics8StatTypeELj6EE3addEPKcS3_(628932,$358,$359)|0); //@line 49 "./common/StringMap.h"
  $360 = HEAP32[(15512)>>2]|0; //@line 49 "./common/StringMap.h"
  $361 = HEAP32[(15516)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_8graphics8Graphics8StatTypeELj6EE3addEPKcS3_(628932,$360,$361)|0); //@line 49 "./common/StringMap.h"
  $362 = HEAP32[(15520)>>2]|0; //@line 49 "./common/StringMap.h"
  $363 = HEAP32[(15524)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_8graphics8Graphics8StatTypeELj6EE3addEPKcS3_(628932,$362,$363)|0); //@line 49 "./common/StringMap.h"
  $364 = HEAP32[(15528)>>2]|0; //@line 49 "./common/StringMap.h"
  $365 = HEAP32[(15532)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_8graphics8Graphics8StatTypeELj6EE3addEPKcS3_(628932,$364,$365)|0); //@line 49 "./common/StringMap.h"
  $366 = HEAP32[(15536)>>2]|0; //@line 49 "./common/StringMap.h"
  $367 = HEAP32[(15540)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_8graphics8Graphics8StatTypeELj6EE3addEPKcS3_(628932,$366,$367)|0); //@line 49 "./common/StringMap.h"
  STACKTOP = sp;return;
 } else {
  $355 = (628924 + ($323<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$355>>2] = $322; //@line 132 "./common/StringMap.h"
  HEAP8[(628940)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(628952)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(628964)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(628976)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(628988)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(629000)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(629012)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(629024)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(629036)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(629048)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(629060)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(629072)>>0] = 0; //@line 155 "./common/StringMap.h"
  ;HEAP32[(629076)>>2]=0|0;HEAP32[(629076)+4>>2]=0|0;HEAP32[(629076)+8>>2]=0|0;HEAP32[(629076)+12>>2]=0|0;HEAP32[(629076)+16>>2]=0|0;HEAP32[(629076)+20>>2]=0|0; //@line 44 "./common/StringMap.h"
  $356 = HEAP32[3874]|0; //@line 49 "./common/StringMap.h"
  $357 = HEAP32[(15500)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_8graphics8Graphics8StatTypeELj6EE3addEPKcS3_(628932,$356,$357)|0); //@line 49 "./common/StringMap.h"
  $358 = HEAP32[(15504)>>2]|0; //@line 49 "./common/StringMap.h"
  $359 = HEAP32[(15508)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_8graphics8Graphics8StatTypeELj6EE3addEPKcS3_(628932,$358,$359)|0); //@line 49 "./common/StringMap.h"
  $360 = HEAP32[(15512)>>2]|0; //@line 49 "./common/StringMap.h"
  $361 = HEAP32[(15516)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_8graphics8Graphics8StatTypeELj6EE3addEPKcS3_(628932,$360,$361)|0); //@line 49 "./common/StringMap.h"
  $362 = HEAP32[(15520)>>2]|0; //@line 49 "./common/StringMap.h"
  $363 = HEAP32[(15524)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_8graphics8Graphics8StatTypeELj6EE3addEPKcS3_(628932,$362,$363)|0); //@line 49 "./common/StringMap.h"
  $364 = HEAP32[(15528)>>2]|0; //@line 49 "./common/StringMap.h"
  $365 = HEAP32[(15532)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_8graphics8Graphics8StatTypeELj6EE3addEPKcS3_(628932,$364,$365)|0); //@line 49 "./common/StringMap.h"
  $366 = HEAP32[(15536)>>2]|0; //@line 49 "./common/StringMap.h"
  $367 = HEAP32[(15540)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_8graphics8Graphics8StatTypeELj6EE3addEPKcS3_(628932,$366,$367)|0); //@line 49 "./common/StringMap.h"
  STACKTOP = sp;return;
 }
}
function __ZN4love9StringMapINS_8graphics8Graphics7ArcModeELj3EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $inserted$2$off0 = 0;
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(2); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 72|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love9StringMapINS_8graphics8Graphics9BlendModeELj8EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0;
 var $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 15; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = $13 & 15; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = $22 & 15; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = $27 & 15; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = $32 & 15; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = $37 & 15; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $42 = (($hash$0$lcssa$i) + 6)|0; //@line 112 "./common/StringMap.h"
       $43 = $42 & 15; //@line 112 "./common/StringMap.h"
       $44 = (((($this) + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
       $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
       $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
       if ($46) {
        $$lcssa = $43;$$lcssa10 = $44;
        label = 5;
       } else {
        $47 = (($hash$0$lcssa$i) + 7)|0; //@line 112 "./common/StringMap.h"
        $48 = $47 & 15; //@line 112 "./common/StringMap.h"
        $49 = (((($this) + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
        $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
        $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
        if ($51) {
         $$lcssa = $48;$$lcssa10 = $49;
         label = 5;
        } else {
         $52 = (($hash$0$lcssa$i) + 8)|0; //@line 112 "./common/StringMap.h"
         $53 = $52 & 15; //@line 112 "./common/StringMap.h"
         $54 = (((($this) + (($53*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
         $55 = HEAP8[$54>>0]|0; //@line 114 "./common/StringMap.h"
         $56 = ($55<<24>>24)==(0); //@line 114 "./common/StringMap.h"
         if ($56) {
          $$lcssa = $53;$$lcssa10 = $54;
          label = 5;
         } else {
          $57 = (($hash$0$lcssa$i) + 9)|0; //@line 112 "./common/StringMap.h"
          $58 = $57 & 15; //@line 112 "./common/StringMap.h"
          $59 = (((($this) + (($58*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
          $60 = HEAP8[$59>>0]|0; //@line 114 "./common/StringMap.h"
          $61 = ($60<<24>>24)==(0); //@line 114 "./common/StringMap.h"
          if ($61) {
           $$lcssa = $58;$$lcssa10 = $59;
           label = 5;
          } else {
           $62 = (($hash$0$lcssa$i) + 10)|0; //@line 112 "./common/StringMap.h"
           $63 = $62 & 15; //@line 112 "./common/StringMap.h"
           $64 = (((($this) + (($63*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
           $65 = HEAP8[$64>>0]|0; //@line 114 "./common/StringMap.h"
           $66 = ($65<<24>>24)==(0); //@line 114 "./common/StringMap.h"
           if ($66) {
            $$lcssa = $63;$$lcssa10 = $64;
            label = 5;
           } else {
            $67 = (($hash$0$lcssa$i) + 11)|0; //@line 112 "./common/StringMap.h"
            $68 = $67 & 15; //@line 112 "./common/StringMap.h"
            $69 = (((($this) + (($68*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
            $70 = HEAP8[$69>>0]|0; //@line 114 "./common/StringMap.h"
            $71 = ($70<<24>>24)==(0); //@line 114 "./common/StringMap.h"
            if ($71) {
             $$lcssa = $68;$$lcssa10 = $69;
             label = 5;
            } else {
             $72 = (($hash$0$lcssa$i) + 12)|0; //@line 112 "./common/StringMap.h"
             $73 = $72 & 15; //@line 112 "./common/StringMap.h"
             $74 = (((($this) + (($73*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
             $75 = HEAP8[$74>>0]|0; //@line 114 "./common/StringMap.h"
             $76 = ($75<<24>>24)==(0); //@line 114 "./common/StringMap.h"
             if ($76) {
              $$lcssa = $73;$$lcssa10 = $74;
              label = 5;
             } else {
              $77 = (($hash$0$lcssa$i) + 13)|0; //@line 112 "./common/StringMap.h"
              $78 = $77 & 15; //@line 112 "./common/StringMap.h"
              $79 = (((($this) + (($78*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
              $80 = HEAP8[$79>>0]|0; //@line 114 "./common/StringMap.h"
              $81 = ($80<<24>>24)==(0); //@line 114 "./common/StringMap.h"
              if ($81) {
               $$lcssa = $78;$$lcssa10 = $79;
               label = 5;
              } else {
               $82 = (($hash$0$lcssa$i) + 14)|0; //@line 112 "./common/StringMap.h"
               $83 = $82 & 15; //@line 112 "./common/StringMap.h"
               $84 = (((($this) + (($83*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
               $85 = HEAP8[$84>>0]|0; //@line 114 "./common/StringMap.h"
               $86 = ($85<<24>>24)==(0); //@line 114 "./common/StringMap.h"
               if ($86) {
                $$lcssa = $83;$$lcssa10 = $84;
                label = 5;
               } else {
                $87 = (($hash$0$lcssa$i) + 15)|0; //@line 112 "./common/StringMap.h"
                $88 = $87 & 15; //@line 112 "./common/StringMap.h"
                $89 = (((($this) + (($88*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
                $90 = HEAP8[$89>>0]|0; //@line 114 "./common/StringMap.h"
                $91 = ($90<<24>>24)==(0); //@line 114 "./common/StringMap.h"
                if ($91) {
                 $$lcssa = $88;$$lcssa10 = $89;
                 label = 5;
                } else {
                 $inserted$2$off0 = 0;
                }
               }
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(7); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 192|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love9StringMapINS_8graphics8Graphics8LineJoinELj3EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $inserted$2$off0 = 0;
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(2); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 72|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love9StringMapINS_8graphics8Graphics13StencilActionELj6EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0;
 var $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $42 = (($hash$0$lcssa$i) + 6)|0; //@line 112 "./common/StringMap.h"
       $43 = (($42>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
       $44 = (((($this) + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
       $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
       $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
       if ($46) {
        $$lcssa = $43;$$lcssa10 = $44;
        label = 5;
       } else {
        $47 = (($hash$0$lcssa$i) + 7)|0; //@line 112 "./common/StringMap.h"
        $48 = (($47>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
        $49 = (((($this) + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
        $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
        $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
        if ($51) {
         $$lcssa = $48;$$lcssa10 = $49;
         label = 5;
        } else {
         $52 = (($hash$0$lcssa$i) + 8)|0; //@line 112 "./common/StringMap.h"
         $53 = (($52>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
         $54 = (((($this) + (($53*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
         $55 = HEAP8[$54>>0]|0; //@line 114 "./common/StringMap.h"
         $56 = ($55<<24>>24)==(0); //@line 114 "./common/StringMap.h"
         if ($56) {
          $$lcssa = $53;$$lcssa10 = $54;
          label = 5;
         } else {
          $57 = (($hash$0$lcssa$i) + 9)|0; //@line 112 "./common/StringMap.h"
          $58 = (($57>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
          $59 = (((($this) + (($58*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
          $60 = HEAP8[$59>>0]|0; //@line 114 "./common/StringMap.h"
          $61 = ($60<<24>>24)==(0); //@line 114 "./common/StringMap.h"
          if ($61) {
           $$lcssa = $58;$$lcssa10 = $59;
           label = 5;
          } else {
           $62 = (($hash$0$lcssa$i) + 10)|0; //@line 112 "./common/StringMap.h"
           $63 = (($62>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
           $64 = (((($this) + (($63*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
           $65 = HEAP8[$64>>0]|0; //@line 114 "./common/StringMap.h"
           $66 = ($65<<24>>24)==(0); //@line 114 "./common/StringMap.h"
           if ($66) {
            $$lcssa = $63;$$lcssa10 = $64;
            label = 5;
           } else {
            $67 = (($hash$0$lcssa$i) + 11)|0; //@line 112 "./common/StringMap.h"
            $68 = (($67>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
            $69 = (((($this) + (($68*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
            $70 = HEAP8[$69>>0]|0; //@line 114 "./common/StringMap.h"
            $71 = ($70<<24>>24)==(0); //@line 114 "./common/StringMap.h"
            if ($71) {
             $$lcssa = $68;$$lcssa10 = $69;
             label = 5;
            } else {
             $inserted$2$off0 = 0;
            }
           }
          }
         }
        }
       }
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(5); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 144|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love9StringMapINS_8graphics8Graphics11CompareModeELj7EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $42 = (($hash$0$lcssa$i) + 6)|0; //@line 112 "./common/StringMap.h"
       $43 = (($42>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
       $44 = (((($this) + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
       $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
       $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
       if ($46) {
        $$lcssa = $43;$$lcssa10 = $44;
        label = 5;
       } else {
        $47 = (($hash$0$lcssa$i) + 7)|0; //@line 112 "./common/StringMap.h"
        $48 = (($47>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
        $49 = (((($this) + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
        $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
        $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
        if ($51) {
         $$lcssa = $48;$$lcssa10 = $49;
         label = 5;
        } else {
         $52 = (($hash$0$lcssa$i) + 8)|0; //@line 112 "./common/StringMap.h"
         $53 = (($52>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
         $54 = (((($this) + (($53*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
         $55 = HEAP8[$54>>0]|0; //@line 114 "./common/StringMap.h"
         $56 = ($55<<24>>24)==(0); //@line 114 "./common/StringMap.h"
         if ($56) {
          $$lcssa = $53;$$lcssa10 = $54;
          label = 5;
         } else {
          $57 = (($hash$0$lcssa$i) + 9)|0; //@line 112 "./common/StringMap.h"
          $58 = (($57>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
          $59 = (((($this) + (($58*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
          $60 = HEAP8[$59>>0]|0; //@line 114 "./common/StringMap.h"
          $61 = ($60<<24>>24)==(0); //@line 114 "./common/StringMap.h"
          if ($61) {
           $$lcssa = $58;$$lcssa10 = $59;
           label = 5;
          } else {
           $62 = (($hash$0$lcssa$i) + 10)|0; //@line 112 "./common/StringMap.h"
           $63 = (($62>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
           $64 = (((($this) + (($63*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
           $65 = HEAP8[$64>>0]|0; //@line 114 "./common/StringMap.h"
           $66 = ($65<<24>>24)==(0); //@line 114 "./common/StringMap.h"
           if ($66) {
            $$lcssa = $63;$$lcssa10 = $64;
            label = 5;
           } else {
            $67 = (($hash$0$lcssa$i) + 11)|0; //@line 112 "./common/StringMap.h"
            $68 = (($67>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
            $69 = (((($this) + (($68*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
            $70 = HEAP8[$69>>0]|0; //@line 114 "./common/StringMap.h"
            $71 = ($70<<24>>24)==(0); //@line 114 "./common/StringMap.h"
            if ($71) {
             $$lcssa = $68;$$lcssa10 = $69;
             label = 5;
            } else {
             $72 = (($hash$0$lcssa$i) + 12)|0; //@line 112 "./common/StringMap.h"
             $73 = (($72>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
             $74 = (((($this) + (($73*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
             $75 = HEAP8[$74>>0]|0; //@line 114 "./common/StringMap.h"
             $76 = ($75<<24>>24)==(0); //@line 114 "./common/StringMap.h"
             if ($76) {
              $$lcssa = $73;$$lcssa10 = $74;
              label = 5;
             } else {
              $77 = (($hash$0$lcssa$i) + 13)|0; //@line 112 "./common/StringMap.h"
              $78 = (($77>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
              $79 = (((($this) + (($78*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
              $80 = HEAP8[$79>>0]|0; //@line 114 "./common/StringMap.h"
              $81 = ($80<<24>>24)==(0); //@line 114 "./common/StringMap.h"
              if ($81) {
               $$lcssa = $78;$$lcssa10 = $79;
               label = 5;
              } else {
               $inserted$2$off0 = 0;
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(6); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 168|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love9StringMapINS_8graphics8Graphics7SupportELj3EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $inserted$2$off0 = 0;
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(2); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 72|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love9StringMapINS_8graphics8Graphics11SystemLimitELj4EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0;
 var $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 7; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = $13 & 7; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = $22 & 7; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = $27 & 7; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = $32 & 7; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = $37 & 7; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $42 = (($hash$0$lcssa$i) + 6)|0; //@line 112 "./common/StringMap.h"
       $43 = $42 & 7; //@line 112 "./common/StringMap.h"
       $44 = (((($this) + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
       $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
       $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
       if ($46) {
        $$lcssa = $43;$$lcssa10 = $44;
        label = 5;
       } else {
        $47 = (($hash$0$lcssa$i) + 7)|0; //@line 112 "./common/StringMap.h"
        $48 = $47 & 7; //@line 112 "./common/StringMap.h"
        $49 = (((($this) + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
        $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
        $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
        if ($51) {
         $$lcssa = $48;$$lcssa10 = $49;
         label = 5;
        } else {
         $inserted$2$off0 = 0;
        }
       }
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(3); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 96|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love9StringMapINS_8graphics8Graphics8StatTypeELj6EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0;
 var $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $42 = (($hash$0$lcssa$i) + 6)|0; //@line 112 "./common/StringMap.h"
       $43 = (($42>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
       $44 = (((($this) + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
       $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
       $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
       if ($46) {
        $$lcssa = $43;$$lcssa10 = $44;
        label = 5;
       } else {
        $47 = (($hash$0$lcssa$i) + 7)|0; //@line 112 "./common/StringMap.h"
        $48 = (($47>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
        $49 = (((($this) + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
        $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
        $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
        if ($51) {
         $$lcssa = $48;$$lcssa10 = $49;
         label = 5;
        } else {
         $52 = (($hash$0$lcssa$i) + 8)|0; //@line 112 "./common/StringMap.h"
         $53 = (($52>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
         $54 = (((($this) + (($53*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
         $55 = HEAP8[$54>>0]|0; //@line 114 "./common/StringMap.h"
         $56 = ($55<<24>>24)==(0); //@line 114 "./common/StringMap.h"
         if ($56) {
          $$lcssa = $53;$$lcssa10 = $54;
          label = 5;
         } else {
          $57 = (($hash$0$lcssa$i) + 9)|0; //@line 112 "./common/StringMap.h"
          $58 = (($57>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
          $59 = (((($this) + (($58*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
          $60 = HEAP8[$59>>0]|0; //@line 114 "./common/StringMap.h"
          $61 = ($60<<24>>24)==(0); //@line 114 "./common/StringMap.h"
          if ($61) {
           $$lcssa = $58;$$lcssa10 = $59;
           label = 5;
          } else {
           $62 = (($hash$0$lcssa$i) + 10)|0; //@line 112 "./common/StringMap.h"
           $63 = (($62>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
           $64 = (((($this) + (($63*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
           $65 = HEAP8[$64>>0]|0; //@line 114 "./common/StringMap.h"
           $66 = ($65<<24>>24)==(0); //@line 114 "./common/StringMap.h"
           if ($66) {
            $$lcssa = $63;$$lcssa10 = $64;
            label = 5;
           } else {
            $67 = (($hash$0$lcssa$i) + 11)|0; //@line 112 "./common/StringMap.h"
            $68 = (($67>>>0) % 12)&-1; //@line 112 "./common/StringMap.h"
            $69 = (((($this) + (($68*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
            $70 = HEAP8[$69>>0]|0; //@line 114 "./common/StringMap.h"
            $71 = ($70<<24>>24)==(0); //@line 114 "./common/StringMap.h"
            if ($71) {
             $$lcssa = $68;$$lcssa10 = $69;
             label = 5;
            } else {
             $inserted$2$off0 = 0;
            }
           }
          }
         }
        }
       }
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(5); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 144|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love8graphics15setGammaCorrectEb($gammacorrect) {
 $gammacorrect = $gammacorrect|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $gammacorrect&1; //@line 33 "modules/graphics/Graphics.cpp"
 HEAP8[1250482] = $0; //@line 33 "modules/graphics/Graphics.cpp"
 return; //@line 34 "modules/graphics/Graphics.cpp"
}
function __ZN4love8graphics14isGammaCorrectEv() {
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[1250482]|0; //@line 38 "modules/graphics/Graphics.cpp"
 $1 = ($0<<24>>24)!=(0); //@line 38 "modules/graphics/Graphics.cpp"
 return ($1|0); //@line 38 "modules/graphics/Graphics.cpp"
}
function __ZN4love8graphics17gammaCorrectColorERNS0_6ColorTIfEE($c) {
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0.0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[1250482]|0; //@line 38 "modules/graphics/Graphics.cpp"
 $1 = ($0<<24>>24)==(0); //@line 38 "modules/graphics/Graphics.cpp"
 if ($1) {
  return; //@line 49 "modules/graphics/Graphics.cpp"
 }
 $2 = +HEAPF32[$c>>2]; //@line 45 "modules/graphics/Graphics.cpp"
 $3 = (+__ZNK4love4math4Math13gammaToLinearEf(623984,$2)); //@line 45 "modules/graphics/Graphics.cpp"
 HEAPF32[$c>>2] = $3; //@line 45 "modules/graphics/Graphics.cpp"
 $4 = ((($c)) + 4|0); //@line 46 "modules/graphics/Graphics.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 46 "modules/graphics/Graphics.cpp"
 $6 = (+__ZNK4love4math4Math13gammaToLinearEf(623984,$5)); //@line 46 "modules/graphics/Graphics.cpp"
 HEAPF32[$4>>2] = $6; //@line 46 "modules/graphics/Graphics.cpp"
 $7 = ((($c)) + 8|0); //@line 47 "modules/graphics/Graphics.cpp"
 $8 = +HEAPF32[$7>>2]; //@line 47 "modules/graphics/Graphics.cpp"
 $9 = (+__ZNK4love4math4Math13gammaToLinearEf(623984,$8)); //@line 47 "modules/graphics/Graphics.cpp"
 HEAPF32[$7>>2] = $9; //@line 47 "modules/graphics/Graphics.cpp"
 return; //@line 49 "modules/graphics/Graphics.cpp"
}
function __ZN4love8graphics8Graphics11getConstantEPKcRNS1_8DrawModeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love9StringMapINS_8graphics8Graphics8DrawModeELj2EE4findEPKcRS3_(627756,$in,$out)|0); //@line 67 "modules/graphics/Graphics.cpp"
 return ($0|0); //@line 67 "modules/graphics/Graphics.cpp"
}
function __ZN4love9StringMapINS_8graphics8Graphics8DrawModeELj2EE4findEPKcRS3_($this,$key,$t) {
 $this = $this|0;
 $key = $key|0;
 $t = $t|0;
 var $$$2 = 0, $$011$i$137 = 0, $$011$i$233 = 0, $$011$i$330 = 0, $$011$i41 = 0, $$0110$i$138 = 0, $$0110$i$234 = 0, $$0110$i$331 = 0, $$0110$i42 = 0, $$02$i = 0, $$13 = 0, $$2 = 0, $$lcssa12 = 0, $$pre14 = 0, $$pre16 = 0, $$pre18 = 0, $$pre20 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $or$cond$i = 0, $or$cond$i$1 = 0, $or$cond$i$136 = 0, $or$cond$i$2 = 0, $or$cond$i$232 = 0, $or$cond$i$3 = 0, $or$cond$i$329 = 0, $or$cond$i40 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 3; //@line 72 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 74 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 74 "./common/StringMap.h"
 if ($12) {
  $$13 = 1;$$2 = 0;
  $$$2 = $$13 & $$2;
  return ($$$2|0); //@line 85 "./common/StringMap.h"
 }
 $13 = (($this) + (($9*12)|0)|0); //@line 77 "./common/StringMap.h"
 $14 = HEAP32[$13>>2]|0; //@line 77 "./common/StringMap.h"
 $15 = HEAP8[$14>>0]|0; //@line 54 "./common/StringMap.h"
 $16 = ($15<<24>>24)==(0); //@line 54 "./common/StringMap.h"
 L8: do {
  if ($16) {
   $26 = $0;
   label = 8;
  } else {
   $17 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
   $18 = ($15<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
   $or$cond$i40 = $17 & $18; //@line 54 "./common/StringMap.h"
   if ($or$cond$i40) {
    $$011$i41 = $key;$$0110$i42 = $14;
    while(1) {
     $19 = ((($$0110$i42)) + 1|0); //@line 59 "./common/StringMap.h"
     $20 = ((($$011$i41)) + 1|0); //@line 60 "./common/StringMap.h"
     $21 = HEAP8[$19>>0]|0; //@line 54 "./common/StringMap.h"
     $22 = ($21<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     $$pre14 = HEAP8[$20>>0]|0; //@line 63 "./common/StringMap.h"
     if ($22) {
      $26 = $$pre14;
      label = 8;
      break L8;
     }
     $23 = ($$pre14<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $24 = ($21<<24>>24)==($$pre14<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i = $23 & $24; //@line 54 "./common/StringMap.h"
     if ($or$cond$i) {
      $$011$i41 = $20;$$0110$i42 = $19;
     } else {
      label = 10;
      break;
     }
    }
   } else {
    label = 10;
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $25 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
  if ($25) {
   $$lcssa12 = $9;
  } else {
   label = 10;
  }
 }
 do {
  if ((label|0) == 10) {
   $29 = (($hash$0$lcssa$i) + 1)|0; //@line 72 "./common/StringMap.h"
   $30 = $29 & 3; //@line 72 "./common/StringMap.h"
   $31 = (((($this) + (($30*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $32 = HEAP8[$31>>0]|0; //@line 74 "./common/StringMap.h"
   $33 = ($32<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($33) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $34 = (($this) + (($30*12)|0)|0); //@line 77 "./common/StringMap.h"
   $35 = HEAP32[$34>>2]|0; //@line 77 "./common/StringMap.h"
   $36 = HEAP8[$35>>0]|0; //@line 54 "./common/StringMap.h"
   $37 = ($36<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L20: do {
    if ($37) {
     $47 = $0;
     label = 16;
    } else {
     $38 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $39 = ($36<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$136 = $38 & $39; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$136) {
      $$011$i$137 = $key;$$0110$i$138 = $35;
      while(1) {
       $40 = ((($$0110$i$138)) + 1|0); //@line 59 "./common/StringMap.h"
       $41 = ((($$011$i$137)) + 1|0); //@line 60 "./common/StringMap.h"
       $42 = HEAP8[$40>>0]|0; //@line 54 "./common/StringMap.h"
       $43 = ($42<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre16 = HEAP8[$41>>0]|0; //@line 63 "./common/StringMap.h"
       if ($43) {
        $47 = $$pre16;
        label = 16;
        break L20;
       }
       $44 = ($$pre16<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $45 = ($42<<24>>24)==($$pre16<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$1 = $44 & $45; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$1) {
        $$011$i$137 = $41;$$0110$i$138 = $40;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 16) {
    $46 = ($47<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($46) {
     $$lcssa12 = $30;
     break;
    }
   }
   $48 = (($hash$0$lcssa$i) + 2)|0; //@line 72 "./common/StringMap.h"
   $49 = $48 & 3; //@line 72 "./common/StringMap.h"
   $50 = (((($this) + (($49*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $51 = HEAP8[$50>>0]|0; //@line 74 "./common/StringMap.h"
   $52 = ($51<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($52) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $53 = (($this) + (($49*12)|0)|0); //@line 77 "./common/StringMap.h"
   $54 = HEAP32[$53>>2]|0; //@line 77 "./common/StringMap.h"
   $55 = HEAP8[$54>>0]|0; //@line 54 "./common/StringMap.h"
   $56 = ($55<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L31: do {
    if ($56) {
     $66 = $0;
     label = 22;
    } else {
     $57 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $58 = ($55<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$232 = $57 & $58; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$232) {
      $$011$i$233 = $key;$$0110$i$234 = $54;
      while(1) {
       $59 = ((($$0110$i$234)) + 1|0); //@line 59 "./common/StringMap.h"
       $60 = ((($$011$i$233)) + 1|0); //@line 60 "./common/StringMap.h"
       $61 = HEAP8[$59>>0]|0; //@line 54 "./common/StringMap.h"
       $62 = ($61<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre18 = HEAP8[$60>>0]|0; //@line 63 "./common/StringMap.h"
       if ($62) {
        $66 = $$pre18;
        label = 22;
        break L31;
       }
       $63 = ($$pre18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $64 = ($61<<24>>24)==($$pre18<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$2 = $63 & $64; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$2) {
        $$011$i$233 = $60;$$0110$i$234 = $59;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 22) {
    $65 = ($66<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($65) {
     $$lcssa12 = $49;
     break;
    }
   }
   $67 = (($hash$0$lcssa$i) + 3)|0; //@line 72 "./common/StringMap.h"
   $68 = $67 & 3; //@line 72 "./common/StringMap.h"
   $69 = (((($this) + (($68*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $70 = HEAP8[$69>>0]|0; //@line 74 "./common/StringMap.h"
   $71 = ($70<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($71) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $72 = (($this) + (($68*12)|0)|0); //@line 77 "./common/StringMap.h"
   $73 = HEAP32[$72>>2]|0; //@line 77 "./common/StringMap.h"
   $74 = HEAP8[$73>>0]|0; //@line 54 "./common/StringMap.h"
   $75 = ($74<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L42: do {
    if ($75) {
     $85 = $0;
    } else {
     $76 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $77 = ($74<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$329 = $76 & $77; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$329) {
      $$011$i$330 = $key;$$0110$i$331 = $73;
     } else {
      $$13 = 0;$$2 = 0;
      $$$2 = $$13 & $$2;
      return ($$$2|0); //@line 85 "./common/StringMap.h"
     }
     while(1) {
      $78 = ((($$0110$i$331)) + 1|0); //@line 59 "./common/StringMap.h"
      $79 = ((($$011$i$330)) + 1|0); //@line 60 "./common/StringMap.h"
      $80 = HEAP8[$78>>0]|0; //@line 54 "./common/StringMap.h"
      $81 = ($80<<24>>24)==(0); //@line 54 "./common/StringMap.h"
      $$pre20 = HEAP8[$79>>0]|0; //@line 63 "./common/StringMap.h"
      if ($81) {
       $85 = $$pre20;
       break L42;
      }
      $82 = ($$pre20<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
      $83 = ($80<<24>>24)==($$pre20<<24>>24); //@line 56 "./common/StringMap.h"
      $or$cond$i$3 = $82 & $83; //@line 54 "./common/StringMap.h"
      if ($or$cond$i$3) {
       $$011$i$330 = $79;$$0110$i$331 = $78;
      } else {
       $$13 = 0;$$2 = 0;
       break;
      }
     }
     $$$2 = $$13 & $$2;
     return ($$$2|0); //@line 85 "./common/StringMap.h"
    }
   } while(0);
   $84 = ($85<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($84) {
    $$lcssa12 = $68;
   } else {
    $$13 = 0;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
  }
 } while(0);
 $27 = (((($this) + (($$lcssa12*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
 $28 = HEAP32[$27>>2]|0; //@line 79 "./common/StringMap.h"
 HEAP32[$t>>2] = $28; //@line 79 "./common/StringMap.h"
 $$13 = 1;$$2 = 1;
 $$$2 = $$13 & $$2;
 return ($$$2|0); //@line 85 "./common/StringMap.h"
}
function __ZN4love8graphics8Graphics11getConstantEPKcRNS1_7ArcModeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 6)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((627812 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (627812 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(6); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((627812 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 77 "modules/graphics/Graphics.cpp"
  return ($$$2$i|0); //@line 77 "modules/graphics/Graphics.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 77 "modules/graphics/Graphics.cpp"
  return ($$$2$i|0); //@line 77 "modules/graphics/Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics8Graphics11getConstantEPKcRNS1_9BlendModeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = $9 & 15; //@line 72 "./common/StringMap.h"
  $11 = (((627896 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (627896 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(16); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((627896 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 87 "modules/graphics/Graphics.cpp"
  return ($$$2$i|0); //@line 87 "modules/graphics/Graphics.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 87 "modules/graphics/Graphics.cpp"
  return ($$$2$i|0); //@line 87 "modules/graphics/Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics8Graphics11getConstantENS1_9BlendModeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(7); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (628088 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 92 "modules/graphics/Graphics.cpp"
}
function __ZN4love8graphics8Graphics11getConstantEPKcRNS1_10BlendAlphaE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love9StringMapINS_8graphics8Graphics10BlendAlphaELj2EE4findEPKcRS3_(628120,$in,$out)|0); //@line 97 "modules/graphics/Graphics.cpp"
 return ($0|0); //@line 97 "modules/graphics/Graphics.cpp"
}
function __ZN4love9StringMapINS_8graphics8Graphics10BlendAlphaELj2EE4findEPKcRS3_($this,$key,$t) {
 $this = $this|0;
 $key = $key|0;
 $t = $t|0;
 var $$$2 = 0, $$011$i$137 = 0, $$011$i$233 = 0, $$011$i$330 = 0, $$011$i41 = 0, $$0110$i$138 = 0, $$0110$i$234 = 0, $$0110$i$331 = 0, $$0110$i42 = 0, $$02$i = 0, $$13 = 0, $$2 = 0, $$lcssa12 = 0, $$pre14 = 0, $$pre16 = 0, $$pre18 = 0, $$pre20 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $or$cond$i = 0, $or$cond$i$1 = 0, $or$cond$i$136 = 0, $or$cond$i$2 = 0, $or$cond$i$232 = 0, $or$cond$i$3 = 0, $or$cond$i$329 = 0, $or$cond$i40 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 3; //@line 72 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 74 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 74 "./common/StringMap.h"
 if ($12) {
  $$13 = 1;$$2 = 0;
  $$$2 = $$13 & $$2;
  return ($$$2|0); //@line 85 "./common/StringMap.h"
 }
 $13 = (($this) + (($9*12)|0)|0); //@line 77 "./common/StringMap.h"
 $14 = HEAP32[$13>>2]|0; //@line 77 "./common/StringMap.h"
 $15 = HEAP8[$14>>0]|0; //@line 54 "./common/StringMap.h"
 $16 = ($15<<24>>24)==(0); //@line 54 "./common/StringMap.h"
 L8: do {
  if ($16) {
   $26 = $0;
   label = 8;
  } else {
   $17 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
   $18 = ($15<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
   $or$cond$i40 = $17 & $18; //@line 54 "./common/StringMap.h"
   if ($or$cond$i40) {
    $$011$i41 = $key;$$0110$i42 = $14;
    while(1) {
     $19 = ((($$0110$i42)) + 1|0); //@line 59 "./common/StringMap.h"
     $20 = ((($$011$i41)) + 1|0); //@line 60 "./common/StringMap.h"
     $21 = HEAP8[$19>>0]|0; //@line 54 "./common/StringMap.h"
     $22 = ($21<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     $$pre14 = HEAP8[$20>>0]|0; //@line 63 "./common/StringMap.h"
     if ($22) {
      $26 = $$pre14;
      label = 8;
      break L8;
     }
     $23 = ($$pre14<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $24 = ($21<<24>>24)==($$pre14<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i = $23 & $24; //@line 54 "./common/StringMap.h"
     if ($or$cond$i) {
      $$011$i41 = $20;$$0110$i42 = $19;
     } else {
      label = 10;
      break;
     }
    }
   } else {
    label = 10;
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $25 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
  if ($25) {
   $$lcssa12 = $9;
  } else {
   label = 10;
  }
 }
 do {
  if ((label|0) == 10) {
   $29 = (($hash$0$lcssa$i) + 1)|0; //@line 72 "./common/StringMap.h"
   $30 = $29 & 3; //@line 72 "./common/StringMap.h"
   $31 = (((($this) + (($30*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $32 = HEAP8[$31>>0]|0; //@line 74 "./common/StringMap.h"
   $33 = ($32<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($33) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $34 = (($this) + (($30*12)|0)|0); //@line 77 "./common/StringMap.h"
   $35 = HEAP32[$34>>2]|0; //@line 77 "./common/StringMap.h"
   $36 = HEAP8[$35>>0]|0; //@line 54 "./common/StringMap.h"
   $37 = ($36<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L20: do {
    if ($37) {
     $47 = $0;
     label = 16;
    } else {
     $38 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $39 = ($36<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$136 = $38 & $39; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$136) {
      $$011$i$137 = $key;$$0110$i$138 = $35;
      while(1) {
       $40 = ((($$0110$i$138)) + 1|0); //@line 59 "./common/StringMap.h"
       $41 = ((($$011$i$137)) + 1|0); //@line 60 "./common/StringMap.h"
       $42 = HEAP8[$40>>0]|0; //@line 54 "./common/StringMap.h"
       $43 = ($42<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre16 = HEAP8[$41>>0]|0; //@line 63 "./common/StringMap.h"
       if ($43) {
        $47 = $$pre16;
        label = 16;
        break L20;
       }
       $44 = ($$pre16<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $45 = ($42<<24>>24)==($$pre16<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$1 = $44 & $45; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$1) {
        $$011$i$137 = $41;$$0110$i$138 = $40;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 16) {
    $46 = ($47<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($46) {
     $$lcssa12 = $30;
     break;
    }
   }
   $48 = (($hash$0$lcssa$i) + 2)|0; //@line 72 "./common/StringMap.h"
   $49 = $48 & 3; //@line 72 "./common/StringMap.h"
   $50 = (((($this) + (($49*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $51 = HEAP8[$50>>0]|0; //@line 74 "./common/StringMap.h"
   $52 = ($51<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($52) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $53 = (($this) + (($49*12)|0)|0); //@line 77 "./common/StringMap.h"
   $54 = HEAP32[$53>>2]|0; //@line 77 "./common/StringMap.h"
   $55 = HEAP8[$54>>0]|0; //@line 54 "./common/StringMap.h"
   $56 = ($55<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L31: do {
    if ($56) {
     $66 = $0;
     label = 22;
    } else {
     $57 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $58 = ($55<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$232 = $57 & $58; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$232) {
      $$011$i$233 = $key;$$0110$i$234 = $54;
      while(1) {
       $59 = ((($$0110$i$234)) + 1|0); //@line 59 "./common/StringMap.h"
       $60 = ((($$011$i$233)) + 1|0); //@line 60 "./common/StringMap.h"
       $61 = HEAP8[$59>>0]|0; //@line 54 "./common/StringMap.h"
       $62 = ($61<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre18 = HEAP8[$60>>0]|0; //@line 63 "./common/StringMap.h"
       if ($62) {
        $66 = $$pre18;
        label = 22;
        break L31;
       }
       $63 = ($$pre18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $64 = ($61<<24>>24)==($$pre18<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$2 = $63 & $64; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$2) {
        $$011$i$233 = $60;$$0110$i$234 = $59;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 22) {
    $65 = ($66<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($65) {
     $$lcssa12 = $49;
     break;
    }
   }
   $67 = (($hash$0$lcssa$i) + 3)|0; //@line 72 "./common/StringMap.h"
   $68 = $67 & 3; //@line 72 "./common/StringMap.h"
   $69 = (((($this) + (($68*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $70 = HEAP8[$69>>0]|0; //@line 74 "./common/StringMap.h"
   $71 = ($70<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($71) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $72 = (($this) + (($68*12)|0)|0); //@line 77 "./common/StringMap.h"
   $73 = HEAP32[$72>>2]|0; //@line 77 "./common/StringMap.h"
   $74 = HEAP8[$73>>0]|0; //@line 54 "./common/StringMap.h"
   $75 = ($74<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L42: do {
    if ($75) {
     $85 = $0;
    } else {
     $76 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $77 = ($74<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$329 = $76 & $77; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$329) {
      $$011$i$330 = $key;$$0110$i$331 = $73;
     } else {
      $$13 = 0;$$2 = 0;
      $$$2 = $$13 & $$2;
      return ($$$2|0); //@line 85 "./common/StringMap.h"
     }
     while(1) {
      $78 = ((($$0110$i$331)) + 1|0); //@line 59 "./common/StringMap.h"
      $79 = ((($$011$i$330)) + 1|0); //@line 60 "./common/StringMap.h"
      $80 = HEAP8[$78>>0]|0; //@line 54 "./common/StringMap.h"
      $81 = ($80<<24>>24)==(0); //@line 54 "./common/StringMap.h"
      $$pre20 = HEAP8[$79>>0]|0; //@line 63 "./common/StringMap.h"
      if ($81) {
       $85 = $$pre20;
       break L42;
      }
      $82 = ($$pre20<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
      $83 = ($80<<24>>24)==($$pre20<<24>>24); //@line 56 "./common/StringMap.h"
      $or$cond$i$3 = $82 & $83; //@line 54 "./common/StringMap.h"
      if ($or$cond$i$3) {
       $$011$i$330 = $79;$$0110$i$331 = $78;
      } else {
       $$13 = 0;$$2 = 0;
       break;
      }
     }
     $$$2 = $$13 & $$2;
     return ($$$2|0); //@line 85 "./common/StringMap.h"
    }
   } while(0);
   $84 = ($85<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($84) {
    $$lcssa12 = $68;
   } else {
    $$13 = 0;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
  }
 } while(0);
 $27 = (((($this) + (($$lcssa12*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
 $28 = HEAP32[$27>>2]|0; //@line 79 "./common/StringMap.h"
 HEAP32[$t>>2] = $28; //@line 79 "./common/StringMap.h"
 $$13 = 1;$$2 = 1;
 $$$2 = $$13 & $$2;
 return ($$$2|0); //@line 85 "./common/StringMap.h"
}
function __ZN4love8graphics8Graphics11getConstantENS1_10BlendAlphaERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(1); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (628168 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 102 "modules/graphics/Graphics.cpp"
}
function __ZN4love8graphics8Graphics11getConstantEPKcRNS1_9LineStyleE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love9StringMapINS_8graphics8Graphics9LineStyleELj2EE4findEPKcRS3_(628176,$in,$out)|0); //@line 107 "modules/graphics/Graphics.cpp"
 return ($0|0); //@line 107 "modules/graphics/Graphics.cpp"
}
function __ZN4love9StringMapINS_8graphics8Graphics9LineStyleELj2EE4findEPKcRS3_($this,$key,$t) {
 $this = $this|0;
 $key = $key|0;
 $t = $t|0;
 var $$$2 = 0, $$011$i$137 = 0, $$011$i$233 = 0, $$011$i$330 = 0, $$011$i41 = 0, $$0110$i$138 = 0, $$0110$i$234 = 0, $$0110$i$331 = 0, $$0110$i42 = 0, $$02$i = 0, $$13 = 0, $$2 = 0, $$lcssa12 = 0, $$pre14 = 0, $$pre16 = 0, $$pre18 = 0, $$pre20 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $or$cond$i = 0, $or$cond$i$1 = 0, $or$cond$i$136 = 0, $or$cond$i$2 = 0, $or$cond$i$232 = 0, $or$cond$i$3 = 0, $or$cond$i$329 = 0, $or$cond$i40 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 3; //@line 72 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 74 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 74 "./common/StringMap.h"
 if ($12) {
  $$13 = 1;$$2 = 0;
  $$$2 = $$13 & $$2;
  return ($$$2|0); //@line 85 "./common/StringMap.h"
 }
 $13 = (($this) + (($9*12)|0)|0); //@line 77 "./common/StringMap.h"
 $14 = HEAP32[$13>>2]|0; //@line 77 "./common/StringMap.h"
 $15 = HEAP8[$14>>0]|0; //@line 54 "./common/StringMap.h"
 $16 = ($15<<24>>24)==(0); //@line 54 "./common/StringMap.h"
 L8: do {
  if ($16) {
   $26 = $0;
   label = 8;
  } else {
   $17 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
   $18 = ($15<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
   $or$cond$i40 = $17 & $18; //@line 54 "./common/StringMap.h"
   if ($or$cond$i40) {
    $$011$i41 = $key;$$0110$i42 = $14;
    while(1) {
     $19 = ((($$0110$i42)) + 1|0); //@line 59 "./common/StringMap.h"
     $20 = ((($$011$i41)) + 1|0); //@line 60 "./common/StringMap.h"
     $21 = HEAP8[$19>>0]|0; //@line 54 "./common/StringMap.h"
     $22 = ($21<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     $$pre14 = HEAP8[$20>>0]|0; //@line 63 "./common/StringMap.h"
     if ($22) {
      $26 = $$pre14;
      label = 8;
      break L8;
     }
     $23 = ($$pre14<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $24 = ($21<<24>>24)==($$pre14<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i = $23 & $24; //@line 54 "./common/StringMap.h"
     if ($or$cond$i) {
      $$011$i41 = $20;$$0110$i42 = $19;
     } else {
      label = 10;
      break;
     }
    }
   } else {
    label = 10;
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $25 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
  if ($25) {
   $$lcssa12 = $9;
  } else {
   label = 10;
  }
 }
 do {
  if ((label|0) == 10) {
   $29 = (($hash$0$lcssa$i) + 1)|0; //@line 72 "./common/StringMap.h"
   $30 = $29 & 3; //@line 72 "./common/StringMap.h"
   $31 = (((($this) + (($30*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $32 = HEAP8[$31>>0]|0; //@line 74 "./common/StringMap.h"
   $33 = ($32<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($33) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $34 = (($this) + (($30*12)|0)|0); //@line 77 "./common/StringMap.h"
   $35 = HEAP32[$34>>2]|0; //@line 77 "./common/StringMap.h"
   $36 = HEAP8[$35>>0]|0; //@line 54 "./common/StringMap.h"
   $37 = ($36<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L20: do {
    if ($37) {
     $47 = $0;
     label = 16;
    } else {
     $38 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $39 = ($36<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$136 = $38 & $39; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$136) {
      $$011$i$137 = $key;$$0110$i$138 = $35;
      while(1) {
       $40 = ((($$0110$i$138)) + 1|0); //@line 59 "./common/StringMap.h"
       $41 = ((($$011$i$137)) + 1|0); //@line 60 "./common/StringMap.h"
       $42 = HEAP8[$40>>0]|0; //@line 54 "./common/StringMap.h"
       $43 = ($42<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre16 = HEAP8[$41>>0]|0; //@line 63 "./common/StringMap.h"
       if ($43) {
        $47 = $$pre16;
        label = 16;
        break L20;
       }
       $44 = ($$pre16<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $45 = ($42<<24>>24)==($$pre16<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$1 = $44 & $45; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$1) {
        $$011$i$137 = $41;$$0110$i$138 = $40;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 16) {
    $46 = ($47<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($46) {
     $$lcssa12 = $30;
     break;
    }
   }
   $48 = (($hash$0$lcssa$i) + 2)|0; //@line 72 "./common/StringMap.h"
   $49 = $48 & 3; //@line 72 "./common/StringMap.h"
   $50 = (((($this) + (($49*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $51 = HEAP8[$50>>0]|0; //@line 74 "./common/StringMap.h"
   $52 = ($51<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($52) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $53 = (($this) + (($49*12)|0)|0); //@line 77 "./common/StringMap.h"
   $54 = HEAP32[$53>>2]|0; //@line 77 "./common/StringMap.h"
   $55 = HEAP8[$54>>0]|0; //@line 54 "./common/StringMap.h"
   $56 = ($55<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L31: do {
    if ($56) {
     $66 = $0;
     label = 22;
    } else {
     $57 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $58 = ($55<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$232 = $57 & $58; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$232) {
      $$011$i$233 = $key;$$0110$i$234 = $54;
      while(1) {
       $59 = ((($$0110$i$234)) + 1|0); //@line 59 "./common/StringMap.h"
       $60 = ((($$011$i$233)) + 1|0); //@line 60 "./common/StringMap.h"
       $61 = HEAP8[$59>>0]|0; //@line 54 "./common/StringMap.h"
       $62 = ($61<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre18 = HEAP8[$60>>0]|0; //@line 63 "./common/StringMap.h"
       if ($62) {
        $66 = $$pre18;
        label = 22;
        break L31;
       }
       $63 = ($$pre18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $64 = ($61<<24>>24)==($$pre18<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$2 = $63 & $64; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$2) {
        $$011$i$233 = $60;$$0110$i$234 = $59;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 22) {
    $65 = ($66<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($65) {
     $$lcssa12 = $49;
     break;
    }
   }
   $67 = (($hash$0$lcssa$i) + 3)|0; //@line 72 "./common/StringMap.h"
   $68 = $67 & 3; //@line 72 "./common/StringMap.h"
   $69 = (((($this) + (($68*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $70 = HEAP8[$69>>0]|0; //@line 74 "./common/StringMap.h"
   $71 = ($70<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($71) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $72 = (($this) + (($68*12)|0)|0); //@line 77 "./common/StringMap.h"
   $73 = HEAP32[$72>>2]|0; //@line 77 "./common/StringMap.h"
   $74 = HEAP8[$73>>0]|0; //@line 54 "./common/StringMap.h"
   $75 = ($74<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L42: do {
    if ($75) {
     $85 = $0;
    } else {
     $76 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $77 = ($74<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$329 = $76 & $77; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$329) {
      $$011$i$330 = $key;$$0110$i$331 = $73;
     } else {
      $$13 = 0;$$2 = 0;
      $$$2 = $$13 & $$2;
      return ($$$2|0); //@line 85 "./common/StringMap.h"
     }
     while(1) {
      $78 = ((($$0110$i$331)) + 1|0); //@line 59 "./common/StringMap.h"
      $79 = ((($$011$i$330)) + 1|0); //@line 60 "./common/StringMap.h"
      $80 = HEAP8[$78>>0]|0; //@line 54 "./common/StringMap.h"
      $81 = ($80<<24>>24)==(0); //@line 54 "./common/StringMap.h"
      $$pre20 = HEAP8[$79>>0]|0; //@line 63 "./common/StringMap.h"
      if ($81) {
       $85 = $$pre20;
       break L42;
      }
      $82 = ($$pre20<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
      $83 = ($80<<24>>24)==($$pre20<<24>>24); //@line 56 "./common/StringMap.h"
      $or$cond$i$3 = $82 & $83; //@line 54 "./common/StringMap.h"
      if ($or$cond$i$3) {
       $$011$i$330 = $79;$$0110$i$331 = $78;
      } else {
       $$13 = 0;$$2 = 0;
       break;
      }
     }
     $$$2 = $$13 & $$2;
     return ($$$2|0); //@line 85 "./common/StringMap.h"
    }
   } while(0);
   $84 = ($85<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($84) {
    $$lcssa12 = $68;
   } else {
    $$13 = 0;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
  }
 } while(0);
 $27 = (((($this) + (($$lcssa12*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
 $28 = HEAP32[$27>>2]|0; //@line 79 "./common/StringMap.h"
 HEAP32[$t>>2] = $28; //@line 79 "./common/StringMap.h"
 $$13 = 1;$$2 = 1;
 $$$2 = $$13 & $$2;
 return ($$$2|0); //@line 85 "./common/StringMap.h"
}
function __ZN4love8graphics8Graphics11getConstantENS1_9LineStyleERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(1); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (628224 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 112 "modules/graphics/Graphics.cpp"
}
function __ZN4love8graphics8Graphics11getConstantEPKcRNS1_8LineJoinE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 6)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((628232 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (628232 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(6); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((628232 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 117 "modules/graphics/Graphics.cpp"
  return ($$$2$i|0); //@line 117 "modules/graphics/Graphics.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 117 "modules/graphics/Graphics.cpp"
  return ($$$2$i|0); //@line 117 "modules/graphics/Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics8Graphics11getConstantENS1_8LineJoinERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(2); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (628304 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 122 "modules/graphics/Graphics.cpp"
}
function __ZN4love8graphics8Graphics11getConstantEPKcRNS1_13StencilActionE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 12)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((628316 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (628316 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(12); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((628316 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 127 "modules/graphics/Graphics.cpp"
  return ($$$2$i|0); //@line 127 "modules/graphics/Graphics.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 127 "modules/graphics/Graphics.cpp"
  return ($$$2$i|0); //@line 127 "modules/graphics/Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics8Graphics11getConstantEPKcRNS1_11CompareModeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 14)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((628484 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (628484 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(14); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((628484 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 137 "modules/graphics/Graphics.cpp"
  return ($$$2$i|0); //@line 137 "modules/graphics/Graphics.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 137 "modules/graphics/Graphics.cpp"
  return ($$$2$i|0); //@line 137 "modules/graphics/Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics8Graphics11getConstantENS1_11CompareModeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(6); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (628652 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 142 "modules/graphics/Graphics.cpp"
}
function __ZN4love8graphics8Graphics11getConstantENS1_7SupportERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(2); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (628752 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 152 "modules/graphics/Graphics.cpp"
}
function __ZN4love8graphics8Graphics11getConstantENS1_11SystemLimitERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(3); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (628860 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 162 "modules/graphics/Graphics.cpp"
}
function __ZN4love8graphics8Graphics11getConstantEPKcRNS1_9StackTypeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love9StringMapINS_8graphics8Graphics9StackTypeELj2EE4findEPKcRS3_(628876,$in,$out)|0); //@line 167 "modules/graphics/Graphics.cpp"
 return ($0|0); //@line 167 "modules/graphics/Graphics.cpp"
}
function __ZN4love9StringMapINS_8graphics8Graphics9StackTypeELj2EE4findEPKcRS3_($this,$key,$t) {
 $this = $this|0;
 $key = $key|0;
 $t = $t|0;
 var $$$2 = 0, $$011$i$137 = 0, $$011$i$233 = 0, $$011$i$330 = 0, $$011$i41 = 0, $$0110$i$138 = 0, $$0110$i$234 = 0, $$0110$i$331 = 0, $$0110$i42 = 0, $$02$i = 0, $$13 = 0, $$2 = 0, $$lcssa12 = 0, $$pre14 = 0, $$pre16 = 0, $$pre18 = 0, $$pre20 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $or$cond$i = 0, $or$cond$i$1 = 0, $or$cond$i$136 = 0, $or$cond$i$2 = 0, $or$cond$i$232 = 0, $or$cond$i$3 = 0, $or$cond$i$329 = 0, $or$cond$i40 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 3; //@line 72 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 74 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 74 "./common/StringMap.h"
 if ($12) {
  $$13 = 1;$$2 = 0;
  $$$2 = $$13 & $$2;
  return ($$$2|0); //@line 85 "./common/StringMap.h"
 }
 $13 = (($this) + (($9*12)|0)|0); //@line 77 "./common/StringMap.h"
 $14 = HEAP32[$13>>2]|0; //@line 77 "./common/StringMap.h"
 $15 = HEAP8[$14>>0]|0; //@line 54 "./common/StringMap.h"
 $16 = ($15<<24>>24)==(0); //@line 54 "./common/StringMap.h"
 L8: do {
  if ($16) {
   $26 = $0;
   label = 8;
  } else {
   $17 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
   $18 = ($15<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
   $or$cond$i40 = $17 & $18; //@line 54 "./common/StringMap.h"
   if ($or$cond$i40) {
    $$011$i41 = $key;$$0110$i42 = $14;
    while(1) {
     $19 = ((($$0110$i42)) + 1|0); //@line 59 "./common/StringMap.h"
     $20 = ((($$011$i41)) + 1|0); //@line 60 "./common/StringMap.h"
     $21 = HEAP8[$19>>0]|0; //@line 54 "./common/StringMap.h"
     $22 = ($21<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     $$pre14 = HEAP8[$20>>0]|0; //@line 63 "./common/StringMap.h"
     if ($22) {
      $26 = $$pre14;
      label = 8;
      break L8;
     }
     $23 = ($$pre14<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $24 = ($21<<24>>24)==($$pre14<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i = $23 & $24; //@line 54 "./common/StringMap.h"
     if ($or$cond$i) {
      $$011$i41 = $20;$$0110$i42 = $19;
     } else {
      label = 10;
      break;
     }
    }
   } else {
    label = 10;
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $25 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
  if ($25) {
   $$lcssa12 = $9;
  } else {
   label = 10;
  }
 }
 do {
  if ((label|0) == 10) {
   $29 = (($hash$0$lcssa$i) + 1)|0; //@line 72 "./common/StringMap.h"
   $30 = $29 & 3; //@line 72 "./common/StringMap.h"
   $31 = (((($this) + (($30*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $32 = HEAP8[$31>>0]|0; //@line 74 "./common/StringMap.h"
   $33 = ($32<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($33) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $34 = (($this) + (($30*12)|0)|0); //@line 77 "./common/StringMap.h"
   $35 = HEAP32[$34>>2]|0; //@line 77 "./common/StringMap.h"
   $36 = HEAP8[$35>>0]|0; //@line 54 "./common/StringMap.h"
   $37 = ($36<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L20: do {
    if ($37) {
     $47 = $0;
     label = 16;
    } else {
     $38 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $39 = ($36<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$136 = $38 & $39; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$136) {
      $$011$i$137 = $key;$$0110$i$138 = $35;
      while(1) {
       $40 = ((($$0110$i$138)) + 1|0); //@line 59 "./common/StringMap.h"
       $41 = ((($$011$i$137)) + 1|0); //@line 60 "./common/StringMap.h"
       $42 = HEAP8[$40>>0]|0; //@line 54 "./common/StringMap.h"
       $43 = ($42<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre16 = HEAP8[$41>>0]|0; //@line 63 "./common/StringMap.h"
       if ($43) {
        $47 = $$pre16;
        label = 16;
        break L20;
       }
       $44 = ($$pre16<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $45 = ($42<<24>>24)==($$pre16<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$1 = $44 & $45; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$1) {
        $$011$i$137 = $41;$$0110$i$138 = $40;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 16) {
    $46 = ($47<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($46) {
     $$lcssa12 = $30;
     break;
    }
   }
   $48 = (($hash$0$lcssa$i) + 2)|0; //@line 72 "./common/StringMap.h"
   $49 = $48 & 3; //@line 72 "./common/StringMap.h"
   $50 = (((($this) + (($49*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $51 = HEAP8[$50>>0]|0; //@line 74 "./common/StringMap.h"
   $52 = ($51<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($52) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $53 = (($this) + (($49*12)|0)|0); //@line 77 "./common/StringMap.h"
   $54 = HEAP32[$53>>2]|0; //@line 77 "./common/StringMap.h"
   $55 = HEAP8[$54>>0]|0; //@line 54 "./common/StringMap.h"
   $56 = ($55<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L31: do {
    if ($56) {
     $66 = $0;
     label = 22;
    } else {
     $57 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $58 = ($55<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$232 = $57 & $58; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$232) {
      $$011$i$233 = $key;$$0110$i$234 = $54;
      while(1) {
       $59 = ((($$0110$i$234)) + 1|0); //@line 59 "./common/StringMap.h"
       $60 = ((($$011$i$233)) + 1|0); //@line 60 "./common/StringMap.h"
       $61 = HEAP8[$59>>0]|0; //@line 54 "./common/StringMap.h"
       $62 = ($61<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre18 = HEAP8[$60>>0]|0; //@line 63 "./common/StringMap.h"
       if ($62) {
        $66 = $$pre18;
        label = 22;
        break L31;
       }
       $63 = ($$pre18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $64 = ($61<<24>>24)==($$pre18<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$2 = $63 & $64; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$2) {
        $$011$i$233 = $60;$$0110$i$234 = $59;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 22) {
    $65 = ($66<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($65) {
     $$lcssa12 = $49;
     break;
    }
   }
   $67 = (($hash$0$lcssa$i) + 3)|0; //@line 72 "./common/StringMap.h"
   $68 = $67 & 3; //@line 72 "./common/StringMap.h"
   $69 = (((($this) + (($68*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $70 = HEAP8[$69>>0]|0; //@line 74 "./common/StringMap.h"
   $71 = ($70<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($71) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $72 = (($this) + (($68*12)|0)|0); //@line 77 "./common/StringMap.h"
   $73 = HEAP32[$72>>2]|0; //@line 77 "./common/StringMap.h"
   $74 = HEAP8[$73>>0]|0; //@line 54 "./common/StringMap.h"
   $75 = ($74<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L42: do {
    if ($75) {
     $85 = $0;
    } else {
     $76 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $77 = ($74<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$329 = $76 & $77; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$329) {
      $$011$i$330 = $key;$$0110$i$331 = $73;
     } else {
      $$13 = 0;$$2 = 0;
      $$$2 = $$13 & $$2;
      return ($$$2|0); //@line 85 "./common/StringMap.h"
     }
     while(1) {
      $78 = ((($$0110$i$331)) + 1|0); //@line 59 "./common/StringMap.h"
      $79 = ((($$011$i$330)) + 1|0); //@line 60 "./common/StringMap.h"
      $80 = HEAP8[$78>>0]|0; //@line 54 "./common/StringMap.h"
      $81 = ($80<<24>>24)==(0); //@line 54 "./common/StringMap.h"
      $$pre20 = HEAP8[$79>>0]|0; //@line 63 "./common/StringMap.h"
      if ($81) {
       $85 = $$pre20;
       break L42;
      }
      $82 = ($$pre20<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
      $83 = ($80<<24>>24)==($$pre20<<24>>24); //@line 56 "./common/StringMap.h"
      $or$cond$i$3 = $82 & $83; //@line 54 "./common/StringMap.h"
      if ($or$cond$i$3) {
       $$011$i$330 = $79;$$0110$i$331 = $78;
      } else {
       $$13 = 0;$$2 = 0;
       break;
      }
     }
     $$$2 = $$13 & $$2;
     return ($$$2|0); //@line 85 "./common/StringMap.h"
    }
   } while(0);
   $84 = ($85<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($84) {
    $$lcssa12 = $68;
   } else {
    $$13 = 0;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
  }
 } while(0);
 $27 = (((($this) + (($$lcssa12*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
 $28 = HEAP32[$27>>2]|0; //@line 79 "./common/StringMap.h"
 HEAP32[$t>>2] = $28; //@line 79 "./common/StringMap.h"
 $$13 = 1;$$2 = 1;
 $$$2 = $$13 & $$2;
 return ($$$2|0); //@line 85 "./common/StringMap.h"
}
function __ZN4love8graphics8Graphics11getConstantENS1_8StatTypeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(5); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (629076 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 182 "modules/graphics/Graphics.cpp"
}
function __ZN4love8graphics14ParticleSystemD2Ev($this) {
 $this = $this|0;
 var $$pre$i = 0, $$pre$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i = 0;
 var $scevgep$i$i$i1 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (15552); //@line 149 "modules/graphics/ParticleSystem.cpp"
 $0 = ((($this)) + 8|0); //@line 181 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 181 "modules/graphics/ParticleSystem.cpp"
 $2 = ($1|0)==(0|0); //@line 181 "modules/graphics/ParticleSystem.cpp"
 if (!($2)) {
  __ZdaPv($1); //@line 181 "modules/graphics/ParticleSystem.cpp"
 }
 HEAP32[$0>>2] = 0; //@line 183 "modules/graphics/ParticleSystem.cpp"
 $3 = ((($this)) + 36|0); //@line 184 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$3>>2] = 0; //@line 184 "modules/graphics/ParticleSystem.cpp"
 $4 = ((($this)) + 40|0); //@line 185 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$4>>2] = 0; //@line 185 "modules/graphics/ParticleSystem.cpp"
 $5 = ((($this)) + 212|0); //@line 151 "modules/graphics/ParticleSystem.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ($6|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($7)) {
  $8 = ((($this)) + 216|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $9 = HEAP32[$8>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $10 = ($9|0)==($6|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  do {
   if ($10) {
    $24 = $6;
   } else {
    $12 = $9;
    while(1) {
     $11 = ((($12)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$8>>2] = $11; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $13 = HEAP32[$11>>2]|0; //@line 106 "./common/Object.h"
     $14 = ($13|0)==(0|0); //@line 106 "./common/Object.h"
     if ($14) {
      $16 = $11;
     } else {
      $17 = HEAP32[$13>>2]|0; //@line 106 "./common/Object.h"
      $18 = ((($17)) + 12|0); //@line 106 "./common/Object.h"
      $19 = HEAP32[$18>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($19|0,($13|0)); //@line 106 "./common/Object.h"
      $20 = __THREW__; __THREW__ = 0;
      $21 = $20&1;
      if ($21) {
       label = 9;
       break;
      }
      $$pre$i$i$i = HEAP32[$8>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $16 = $$pre$i$i$i;
     }
     $15 = ($16|0)==($6|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($15) {
      label = 10;
      break;
     } else {
      $12 = $16;
     }
    }
    if ((label|0) == 9) {
     $22 = ___cxa_find_matching_catch_3(0|0)|0;
     $23 = tempRet0;
     ___clang_call_terminate($22); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
    else if ((label|0) == 10) {
     $$pre$i = HEAP32[$5>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $24 = $$pre$i;
     break;
    }
   }
  } while(0);
  __ZdlPv($24); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $25 = ((($this)) + 200|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $26 = HEAP32[$25>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = ($26|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $28 = $26; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($27)) {
  $29 = ((($this)) + 204|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $30 = HEAP32[$29>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $31 = ($30|0)==($26|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($31)) {
   $scevgep$i$i$i = ((($30)) + -16|0);
   $32 = $scevgep$i$i$i;
   $33 = (($32) - ($28))|0;
   $34 = $33 >>> 4;
   $35 = $34 ^ -1;
   $scevgep4$i$i$i = (($30) + ($35<<4)|0);
   HEAP32[$29>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($26); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $36 = ((($this)) + 152|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $37 = HEAP32[$36>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $38 = ($37|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $39 = $37; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($38)) {
  $40 = ((($this)) + 156|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $41 = HEAP32[$40>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $42 = ($41|0)==($37|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($42)) {
   $scevgep$i$i$i1 = ((($41)) + -4|0);
   $43 = $scevgep$i$i$i1;
   $44 = (($43) - ($39))|0;
   $45 = $44 >>> 2;
   $46 = $45 ^ -1;
   $scevgep4$i$i$i2 = (($41) + ($46<<2)|0);
   HEAP32[$40>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($37); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $47 = ((($this)) + 24|0); //@line 106 "./common/Object.h"
 $48 = HEAP32[$47>>2]|0; //@line 106 "./common/Object.h"
 $49 = ($48|0)==(0|0); //@line 106 "./common/Object.h"
 if ($49) {
  return; //@line 151 "modules/graphics/ParticleSystem.cpp"
 }
 $50 = HEAP32[$48>>2]|0; //@line 106 "./common/Object.h"
 $51 = ((($50)) + 12|0); //@line 106 "./common/Object.h"
 $52 = HEAP32[$51>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($52|0,($48|0)); //@line 106 "./common/Object.h"
 $53 = __THREW__; __THREW__ = 0;
 $54 = $53&1;
 if ($54) {
  $55 = ___cxa_find_matching_catch_3(0|0)|0;
  $56 = tempRet0;
  ___clang_call_terminate($55); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  return; //@line 151 "modules/graphics/ParticleSystem.cpp"
 }
}
function __ZN4love8graphics14ParticleSystemD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics14ParticleSystemD2Ev($this); //@line 149 "modules/graphics/ParticleSystem.cpp"
 __ZdlPv($this); //@line 149 "modules/graphics/ParticleSystem.cpp"
 return; //@line 151 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem13setBufferSizeEj($this,$size) {
 $this = $this|0;
 $size = $size|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (($size) + -1)|0; //@line 190 "modules/graphics/ParticleSystem.cpp"
 $1 = ($0>>>0)>(536870910); //@line 190 "modules/graphics/ParticleSystem.cpp"
 if ($1) {
  $2 = (___cxa_allocate_exception(16)|0); //@line 191 "modules/graphics/ParticleSystem.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($2|0),(347355|0),($vararg_buffer|0)); //@line 191 "modules/graphics/ParticleSystem.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if ($4) {
   $5 = ___cxa_find_matching_catch_2()|0;
   $6 = tempRet0;
   ___cxa_free_exception(($2|0)); //@line 191 "modules/graphics/ParticleSystem.cpp"
   ___resumeException($5|0);
   // unreachable;
  } else {
   ___cxa_throw(($2|0),(96|0),(344|0)); //@line 191 "modules/graphics/ParticleSystem.cpp"
   // unreachable; //@line 191 "modules/graphics/ParticleSystem.cpp"
  }
 }
 $7 = ((($this)) + 8|0); //@line 181 "modules/graphics/ParticleSystem.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 181 "modules/graphics/ParticleSystem.cpp"
 $9 = ($8|0)==(0|0); //@line 181 "modules/graphics/ParticleSystem.cpp"
 if (!($9)) {
  __ZdaPv($8); //@line 181 "modules/graphics/ParticleSystem.cpp"
 }
 HEAP32[$7>>2] = 0; //@line 183 "modules/graphics/ParticleSystem.cpp"
 $10 = ((($this)) + 36|0); //@line 184 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$10>>2] = 0; //@line 184 "modules/graphics/ParticleSystem.cpp"
 $11 = ((($this)) + 40|0); //@line 185 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$11>>2] = 0; //@line 185 "modules/graphics/ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem13createBuffersEj($this,$size); //@line 193 "modules/graphics/ParticleSystem.cpp"
 $12 = HEAP32[$7>>2]|0; //@line 781 "modules/graphics/ParticleSystem.cpp"
 $13 = ($12|0)==(0|0); //@line 781 "modules/graphics/ParticleSystem.cpp"
 if ($13) {
  STACKTOP = sp;return; //@line 195 "modules/graphics/ParticleSystem.cpp"
 }
 $14 = ((($this)) + 12|0); //@line 784 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$14>>2] = $12; //@line 784 "modules/graphics/ParticleSystem.cpp"
 $15 = ((($this)) + 16|0); //@line 785 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$15>>2] = 0; //@line 785 "modules/graphics/ParticleSystem.cpp"
 $16 = ((($this)) + 20|0); //@line 786 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$16>>2] = 0; //@line 786 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$11>>2] = 0; //@line 787 "modules/graphics/ParticleSystem.cpp"
 $17 = ((($this)) + 80|0); //@line 788 "modules/graphics/ParticleSystem.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 788 "modules/graphics/ParticleSystem.cpp"
 $19 = ((($this)) + 84|0); //@line 788 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$19>>2] = $18; //@line 788 "modules/graphics/ParticleSystem.cpp"
 $20 = ((($this)) + 48|0); //@line 789 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$20>>2] = 0.0; //@line 789 "modules/graphics/ParticleSystem.cpp"
 STACKTOP = sp;return; //@line 195 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem13createBuffersEj($this,$size) {
 $this = $this|0;
 $size = $size|0;
 var $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$arith = 0, $$overflow = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $$arith = ($size*108)|0; //@line 168 "modules/graphics/ParticleSystem.cpp"
 $$overflow = ($size>>>0)>(39768215); //@line 168 "modules/graphics/ParticleSystem.cpp"
 $0 = $$overflow ? -1 : $$arith; //@line 168 "modules/graphics/ParticleSystem.cpp"
 __THREW__ = 0;
 $1 = (invoke_ii(3492,($0|0))|0); //@line 168 "modules/graphics/ParticleSystem.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  $4 = ($size|0)==(0); //@line 168 "modules/graphics/ParticleSystem.cpp"
  if ($4) {
   $11 = ((($this)) + 8|0); //@line 168 "modules/graphics/ParticleSystem.cpp"
   HEAP32[$11>>2] = $1; //@line 168 "modules/graphics/ParticleSystem.cpp"
   $12 = ((($this)) + 12|0); //@line 168 "modules/graphics/ParticleSystem.cpp"
   HEAP32[$12>>2] = $1; //@line 168 "modules/graphics/ParticleSystem.cpp"
   $13 = ((($this)) + 36|0); //@line 169 "modules/graphics/ParticleSystem.cpp"
   HEAP32[$13>>2] = $size; //@line 169 "modules/graphics/ParticleSystem.cpp"
   STACKTOP = sp;return; //@line 176 "modules/graphics/ParticleSystem.cpp"
  }
  $5 = (($1) + (($size*108)|0)|0); //@line 168 "modules/graphics/ParticleSystem.cpp"
  $7 = $1;
  while(1) {
   $6 = ((($7)) + 16|0); //@line 219 "./common/Vector.h"
   $8 = ((($7)) + 88|0); //@line 513 "modules/graphics/ParticleSystem.h"
   ;HEAP32[$8>>2]=0|0;HEAP32[$8+4>>2]=0|0;HEAP32[$8+8>>2]=0|0;HEAP32[$8+12>>2]=0|0; //@line 37 "modules/graphics/Color.h"
   ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;HEAP32[$6+12>>2]=0|0;HEAP32[$6+16>>2]=0|0;HEAP32[$6+20>>2]=0|0;HEAP32[$6+24>>2]=0|0;HEAP32[$6+28>>2]=0|0; //@line 220 "./common/Vector.h"
   $9 = ((($7)) + 108|0); //@line 168 "modules/graphics/ParticleSystem.cpp"
   $10 = ($9|0)==($5|0); //@line 168 "modules/graphics/ParticleSystem.cpp"
   if ($10) {
    break;
   } else {
    $7 = $9;
   }
  }
  $11 = ((($this)) + 8|0); //@line 168 "modules/graphics/ParticleSystem.cpp"
  HEAP32[$11>>2] = $1; //@line 168 "modules/graphics/ParticleSystem.cpp"
  $12 = ((($this)) + 12|0); //@line 168 "modules/graphics/ParticleSystem.cpp"
  HEAP32[$12>>2] = $1; //@line 168 "modules/graphics/ParticleSystem.cpp"
  $13 = ((($this)) + 36|0); //@line 169 "modules/graphics/ParticleSystem.cpp"
  HEAP32[$13>>2] = $size; //@line 169 "modules/graphics/ParticleSystem.cpp"
  STACKTOP = sp;return; //@line 176 "modules/graphics/ParticleSystem.cpp"
 }
 $14 = ___cxa_find_matching_catch_3(9056|0)|0;
 $15 = tempRet0;
 $16 = (_llvm_eh_typeid_for((9056|0))|0); //@line 170 "modules/graphics/ParticleSystem.cpp"
 $17 = ($15|0)==($16|0); //@line 170 "modules/graphics/ParticleSystem.cpp"
 if (!($17)) {
  $$2 = $14;$$23 = $15;
  ___resumeException($$2|0);
  // unreachable;
 }
 (___cxa_begin_catch(($14|0))|0); //@line 170 "modules/graphics/ParticleSystem.cpp"
 $18 = ((($this)) + 8|0); //@line 181 "modules/graphics/ParticleSystem.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 181 "modules/graphics/ParticleSystem.cpp"
 $20 = ($19|0)==(0|0); //@line 181 "modules/graphics/ParticleSystem.cpp"
 if (!($20)) {
  __ZdaPv($19); //@line 181 "modules/graphics/ParticleSystem.cpp"
 }
 HEAP32[$18>>2] = 0; //@line 183 "modules/graphics/ParticleSystem.cpp"
 $21 = ((($this)) + 36|0); //@line 184 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$21>>2] = 0; //@line 184 "modules/graphics/ParticleSystem.cpp"
 $22 = ((($this)) + 40|0); //@line 185 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$22>>2] = 0; //@line 185 "modules/graphics/ParticleSystem.cpp"
 $23 = (___cxa_allocate_exception(16)|0); //@line 174 "modules/graphics/ParticleSystem.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($23|0),(596704|0),($vararg_buffer|0)); //@line 174 "modules/graphics/ParticleSystem.cpp"
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $29 = ___cxa_find_matching_catch_2()|0;
  $30 = tempRet0;
  ___cxa_free_exception(($23|0)); //@line 174 "modules/graphics/ParticleSystem.cpp"
  $$1 = $29;$$12 = $30;
 } else {
  __THREW__ = 0;
  invoke_viii(3350,($23|0),(96|0),(344|0)); //@line 174 "modules/graphics/ParticleSystem.cpp"
  $26 = __THREW__; __THREW__ = 0;
  $27 = ___cxa_find_matching_catch_2()|0;
  $28 = tempRet0;
  $$1 = $27;$$12 = $28;
 }
 __THREW__ = 0;
 invoke_v(3378); //@line 175 "modules/graphics/ParticleSystem.cpp"
 $31 = __THREW__; __THREW__ = 0;
 $32 = $31&1;
 if ($32) {
  $33 = ___cxa_find_matching_catch_3(0|0)|0;
  $34 = tempRet0;
  ___clang_call_terminate($33); //@line 175 "modules/graphics/ParticleSystem.cpp"
  // unreachable; //@line 175 "modules/graphics/ParticleSystem.cpp"
 } else {
  $$2 = $$1;$$23 = $$12;
  ___resumeException($$2|0);
  // unreachable;
 }
}
function __GLOBAL__sub_I_ParticleSystem_cpp() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love4math15RandomGeneratorC2Ev(623936); //@line 41 "modules/graphics/ParticleSystem.cpp"
 (___cxa_atexit((775|0),(623936|0),(___dso_handle|0))|0); //@line 41 "modules/graphics/ParticleSystem.cpp"
 HEAP8[(629108)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629120)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629132)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629144)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629156)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629168)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(629172)>>2]=0|0;HEAP32[(629172)+4>>2]=0|0;HEAP32[(629172)+8>>2]=0|0; //@line 44 "./common/StringMap.h"
 $0 = HEAP32[3895]|0; //@line 49 "./common/StringMap.h"
 $1 = HEAP32[(15584)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics14ParticleSystem22AreaSpreadDistributionELj3EE3addEPKcS3_(629100,$0,$1)|0); //@line 49 "./common/StringMap.h"
 $2 = HEAP32[(15588)>>2]|0; //@line 49 "./common/StringMap.h"
 $3 = HEAP32[(15592)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics14ParticleSystem22AreaSpreadDistributionELj3EE3addEPKcS3_(629100,$2,$3)|0); //@line 49 "./common/StringMap.h"
 $4 = HEAP32[(15596)>>2]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(15600)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics14ParticleSystem22AreaSpreadDistributionELj3EE3addEPKcS3_(629100,$4,$5)|0); //@line 49 "./common/StringMap.h"
 HEAP8[(629192)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629204)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629216)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629228)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629240)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629252)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(629256)>>2]=0|0;HEAP32[(629256)+4>>2]=0|0;HEAP32[(629256)+8>>2]=0|0; //@line 44 "./common/StringMap.h"
 $6 = HEAP32[3901]|0; //@line 49 "./common/StringMap.h"
 $7 = HEAP32[(15608)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics14ParticleSystem10InsertModeELj3EE3addEPKcS3_(629184,$6,$7)|0); //@line 49 "./common/StringMap.h"
 $8 = HEAP32[(15612)>>2]|0; //@line 49 "./common/StringMap.h"
 $9 = HEAP32[(15616)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics14ParticleSystem10InsertModeELj3EE3addEPKcS3_(629184,$8,$9)|0); //@line 49 "./common/StringMap.h"
 $10 = HEAP32[(15620)>>2]|0; //@line 49 "./common/StringMap.h"
 $11 = HEAP32[(15624)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics14ParticleSystem10InsertModeELj3EE3addEPKcS3_(629184,$10,$11)|0); //@line 49 "./common/StringMap.h"
 return;
}
function __ZN4love9StringMapINS_8graphics14ParticleSystem22AreaSpreadDistributionELj3EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $inserted$2$off0 = 0;
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(2); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 72|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love9StringMapINS_8graphics14ParticleSystem10InsertModeELj3EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $inserted$2$off0 = 0;
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(2); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 72|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love8graphics14ParticleSystemC2EPNS0_7TextureEj($this,$texture,$size) {
 $this = $this|0;
 $texture = $texture|0;
 $size = $size|0;
 var $$0 = 0, $$02 = 0, $$1 = 0, $$13 = 0, $$2 = 0, $$24 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i$i = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i5 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i6 = 0, $vararg_buffer = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 24|0;
 $1 = sp + 8|0;
 __ZN4love6ObjectC2Ev($this); //@line 36 "modules/graphics/Drawable.h"
 HEAP32[$this>>2] = (15552); //@line 91 "modules/graphics/ParticleSystem.cpp"
 $2 = ((($this)) + 8|0); //@line 54 "modules/graphics/ParticleSystem.cpp"
 $3 = ((($this)) + 24|0); //@line 93 "./common/Object.h"
 ;HEAP32[$2>>2]=0|0;HEAP32[$2+4>>2]=0|0;HEAP32[$2+8>>2]=0|0;HEAP32[$2+12>>2]=0|0; //@line 55 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$3>>2] = $texture; //@line 93 "./common/Object.h"
 $4 = ($texture|0)==(0|0); //@line 95 "./common/Object.h"
 if (!($4)) {
  $5 = HEAP32[$texture>>2]|0; //@line 95 "./common/Object.h"
  $6 = ((($5)) + 8|0); //@line 95 "./common/Object.h"
  $7 = HEAP32[$6>>2]|0; //@line 95 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($7|0,($texture|0)); //@line 95 "./common/Object.h"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if ($9) {
   $45 = ___cxa_find_matching_catch_2()|0;
   $46 = tempRet0;
   $$2 = $46;$$24 = $45;
   ___resumeException($$24|0);
   // unreachable;
  }
 }
 $10 = ((($this)) + 28|0); //@line 59 "modules/graphics/ParticleSystem.cpp"
 HEAP8[$10>>0] = 1; //@line 59 "modules/graphics/ParticleSystem.cpp"
 $11 = ((($this)) + 32|0); //@line 60 "modules/graphics/ParticleSystem.cpp"
 $12 = ((($this)) + 80|0); //@line 66 "modules/graphics/ParticleSystem.cpp"
 dest=$11; stop=dest+48|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 61 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$12>>2] = -1.0; //@line 66 "modules/graphics/ParticleSystem.cpp"
 $13 = ((($this)) + 84|0); //@line 67 "modules/graphics/ParticleSystem.cpp"
 $14 = ((($this)) + 152|0); //@line 53 "modules/graphics/ParticleSystem.cpp"
 $15 = ((($this)) + 156|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 dest=$13; stop=dest+104|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 68 "modules/graphics/ParticleSystem.cpp"
 $16 = HEAP32[$texture>>2]|0; //@line 88 "modules/graphics/ParticleSystem.cpp"
 $17 = ((($16)) + 24|0); //@line 88 "modules/graphics/ParticleSystem.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 88 "modules/graphics/ParticleSystem.cpp"
 __THREW__ = 0;
 $19 = (invoke_ii($18|0,($texture|0))|0); //@line 88 "modules/graphics/ParticleSystem.cpp"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  label = 9;
 } else {
  $22 = HEAP32[$texture>>2]|0; //@line 88 "modules/graphics/ParticleSystem.cpp"
  $23 = ((($22)) + 28|0); //@line 88 "modules/graphics/ParticleSystem.cpp"
  $24 = HEAP32[$23>>2]|0; //@line 88 "modules/graphics/ParticleSystem.cpp"
  __THREW__ = 0;
  $25 = (invoke_ii($24|0,($texture|0))|0); //@line 88 "modules/graphics/ParticleSystem.cpp"
  $26 = __THREW__; __THREW__ = 0;
  $27 = $26&1;
  if ($27) {
   label = 9;
  } else {
   $28 = (+($19|0)); //@line 88 "modules/graphics/ParticleSystem.cpp"
   $29 = $28 * 0.5; //@line 88 "modules/graphics/ParticleSystem.cpp"
   $30 = (+($25|0)); //@line 88 "modules/graphics/ParticleSystem.cpp"
   $31 = $30 * 0.5; //@line 88 "modules/graphics/ParticleSystem.cpp"
   $32 = ((($this)) + 188|0); //@line 225 "./common/Vector.h"
   HEAPF32[$32>>2] = $29; //@line 225 "./common/Vector.h"
   $33 = ((($this)) + 192|0); //@line 226 "./common/Vector.h"
   HEAPF32[$33>>2] = $31; //@line 226 "./common/Vector.h"
   $34 = ((($this)) + 196|0); //@line 89 "modules/graphics/ParticleSystem.cpp"
   HEAP8[$34>>0] = 1; //@line 89 "modules/graphics/ParticleSystem.cpp"
   $35 = ((($this)) + 200|0); //@line 53 "modules/graphics/ParticleSystem.cpp"
   $36 = ((($this)) + 204|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $37 = ((($this)) + 212|0); //@line 53 "modules/graphics/ParticleSystem.cpp"
   $38 = ((($this)) + 216|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $39 = (($size) + -1)|0; //@line 92 "modules/graphics/ParticleSystem.cpp"
   $40 = ($39>>>0)>(536870910); //@line 92 "modules/graphics/ParticleSystem.cpp"
   ;HEAP32[$35>>2]=0|0;HEAP32[$35+4>>2]=0|0;HEAP32[$35+8>>2]=0|0;HEAP32[$35+12>>2]=0|0;HEAP32[$35+16>>2]=0|0;HEAP32[$35+20>>2]=0|0;HEAP8[$35+24>>0]=0|0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   do {
    if ($40) {
     $41 = (___cxa_allocate_exception(16)|0); //@line 93 "modules/graphics/ParticleSystem.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($41|0),(347394|0),($vararg_buffer|0)); //@line 93 "modules/graphics/ParticleSystem.cpp"
     $42 = __THREW__; __THREW__ = 0;
     $43 = $42&1;
     if ($43) {
      $49 = ___cxa_find_matching_catch_2()|0;
      $50 = tempRet0;
      ___cxa_free_exception(($41|0)); //@line 93 "modules/graphics/ParticleSystem.cpp"
      $$0 = $50;$$02 = $49;
      break;
     } else {
      __THREW__ = 0;
      invoke_viii(3350,($41|0),(96|0),(344|0)); //@line 93 "modules/graphics/ParticleSystem.cpp"
      $44 = __THREW__; __THREW__ = 0;
      label = 11;
      break;
     }
    } else {
     HEAPF32[$0>>2] = 1.0; //@line 95 "modules/graphics/ParticleSystem.cpp"
     $53 = HEAP32[$15>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $54 = ((($this)) + 160|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $55 = HEAP32[$54>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $56 = ($53>>>0)<($55>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($56) {
      HEAP32[$53>>2] = 1065353216; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $57 = ((($53)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$15>>2] = $57; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $64 = 0;$65 = 0;
     } else {
      __THREW__ = 0;
      invoke_vii(3625,($14|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $58 = __THREW__; __THREW__ = 0;
      $59 = $58&1;
      if ($59) {
       label = 11;
       break;
      }
      $$pre = HEAP32[$36>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$phi$trans$insert = ((($this)) + 208|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $$pre10 = HEAP32[$$phi$trans$insert>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $64 = $$pre;$65 = $$pre10;
     }
     HEAPF32[$1>>2] = 1.0; //@line 38 "modules/graphics/Color.h"
     $60 = ((($1)) + 4|0); //@line 38 "modules/graphics/Color.h"
     HEAPF32[$60>>2] = 1.0; //@line 38 "modules/graphics/Color.h"
     $61 = ((($1)) + 8|0); //@line 38 "modules/graphics/Color.h"
     HEAPF32[$61>>2] = 1.0; //@line 38 "modules/graphics/Color.h"
     $62 = ((($1)) + 12|0); //@line 38 "modules/graphics/Color.h"
     HEAPF32[$62>>2] = 1.0; //@line 38 "modules/graphics/Color.h"
     $63 = ($64>>>0)<($65>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($63) {
      ;HEAP32[$64>>2]=HEAP32[$1>>2]|0;HEAP32[$64+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$64+8>>2]=HEAP32[$1+8>>2]|0;HEAP32[$64+12>>2]=HEAP32[$1+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $66 = HEAP32[$36>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $67 = ((($66)) + 16|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$36>>2] = $67; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     } else {
      __THREW__ = 0;
      invoke_vii(3626,($35|0),($1|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $68 = __THREW__; __THREW__ = 0;
      $69 = $68&1;
      if ($69) {
       label = 11;
       break;
      }
     }
     $70 = HEAP32[$this>>2]|0; //@line 97 "modules/graphics/ParticleSystem.cpp"
     $71 = ((($70)) + 24|0); //@line 97 "modules/graphics/ParticleSystem.cpp"
     $72 = HEAP32[$71>>2]|0; //@line 97 "modules/graphics/ParticleSystem.cpp"
     __THREW__ = 0;
     invoke_vii($72|0,($this|0),($size|0)); //@line 97 "modules/graphics/ParticleSystem.cpp"
     $73 = __THREW__; __THREW__ = 0;
     $74 = $73&1;
     if ($74) {
      label = 11;
     } else {
      STACKTOP = sp;return; //@line 98 "modules/graphics/ParticleSystem.cpp"
     }
    }
   } while(0);
   if ((label|0) == 11) {
    $51 = ___cxa_find_matching_catch_2()|0;
    $52 = tempRet0;
    $$0 = $52;$$02 = $51;
   }
   $75 = HEAP32[$37>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $76 = ($75|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($76)) {
    $77 = HEAP32[$38>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $78 = ($77|0)==($75|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    do {
     if ($78) {
      $92 = $75;
     } else {
      $80 = $77;
      while(1) {
       $79 = ((($80)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$38>>2] = $79; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $81 = HEAP32[$79>>2]|0; //@line 106 "./common/Object.h"
       $82 = ($81|0)==(0|0); //@line 106 "./common/Object.h"
       if ($82) {
        $84 = $79;
       } else {
        $85 = HEAP32[$81>>2]|0; //@line 106 "./common/Object.h"
        $86 = ((($85)) + 12|0); //@line 106 "./common/Object.h"
        $87 = HEAP32[$86>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($87|0,($81|0)); //@line 106 "./common/Object.h"
        $88 = __THREW__; __THREW__ = 0;
        $89 = $88&1;
        if ($89) {
         label = 27;
         break;
        }
        $$pre$i$i$i = HEAP32[$38>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $84 = $$pre$i$i$i;
       }
       $83 = ($84|0)==($75|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($83) {
        label = 28;
        break;
       } else {
        $80 = $84;
       }
      }
      if ((label|0) == 27) {
       $90 = ___cxa_find_matching_catch_3(0|0)|0;
       $91 = tempRet0;
       ___clang_call_terminate($90); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      else if ((label|0) == 28) {
       $$pre$i = HEAP32[$37>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $92 = $$pre$i;
       break;
      }
     }
    } while(0);
    __ZdlPv($92); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $93 = HEAP32[$35>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $94 = ($93|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $95 = $93; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($94) {
    $$1 = $$0;$$13 = $$02;
   } else {
    $96 = HEAP32[$36>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $97 = ($96|0)==($93|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($97)) {
     $scevgep$i$i$i5 = ((($96)) + -16|0);
     $98 = $scevgep$i$i$i5;
     $99 = (($98) - ($95))|0;
     $100 = $99 >>> 4;
     $101 = $100 ^ -1;
     $scevgep4$i$i$i6 = (($96) + ($101<<4)|0);
     HEAP32[$36>>2] = $scevgep4$i$i$i6; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($93); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $$1 = $$0;$$13 = $$02;
   }
  }
 }
 if ((label|0) == 9) {
  $47 = ___cxa_find_matching_catch_2()|0;
  $48 = tempRet0;
  $$1 = $48;$$13 = $47;
 }
 $102 = HEAP32[$14>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $103 = ($102|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $104 = $102; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($103)) {
  $105 = HEAP32[$15>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $106 = ($105|0)==($102|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($106)) {
   $scevgep$i$i$i = ((($105)) + -4|0);
   $107 = $scevgep$i$i$i;
   $108 = (($107) - ($104))|0;
   $109 = $108 >>> 2;
   $110 = $109 ^ -1;
   $scevgep4$i$i$i = (($105) + ($110<<2)|0);
   HEAP32[$15>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($102); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $111 = HEAP32[$3>>2]|0; //@line 106 "./common/Object.h"
 $112 = ($111|0)==(0|0); //@line 106 "./common/Object.h"
 if ($112) {
  $$2 = $$1;$$24 = $$13;
  ___resumeException($$24|0);
  // unreachable;
 }
 $113 = HEAP32[$111>>2]|0; //@line 106 "./common/Object.h"
 $114 = ((($113)) + 12|0); //@line 106 "./common/Object.h"
 $115 = HEAP32[$114>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($115|0,($111|0)); //@line 106 "./common/Object.h"
 $116 = __THREW__; __THREW__ = 0;
 $117 = $116&1;
 if ($117) {
  $118 = ___cxa_find_matching_catch_3(0|0)|0;
  $119 = tempRet0;
  ___clang_call_terminate($118); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  $$2 = $$1;$$24 = $$13;
  ___resumeException($$24|0);
  // unreachable;
 }
}
function __ZNSt3__16vectorIfNS_9allocatorIfEEE21__push_back_slow_pathIfEEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNSt3__16vectorIN4love8graphics6ColorTIfEENS_9allocatorIS4_EEE21__push_back_slow_pathIS4_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(268435455); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 4; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(134217727); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 3; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$32 = $18;
  } else {
   $$0$i1 = $17;$40 = $18;$41 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 268435455;$40 = $12;$41 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 4; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $41;$32 = $40;
 }
 $24 = (($25) + ($26<<4)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<4)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 ;HEAP32[$24>>2]=HEAP32[$__x>>2]|0;HEAP32[$24+4>>2]=HEAP32[$__x+4>>2]|0;HEAP32[$24+8>>2]=HEAP32[$__x+8>>2]|0;HEAP32[$24+12>>2]=HEAP32[$__x+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $29 = ((($24)) + 16|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = $29; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = (($32) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = $31 >> 4; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = (0 - ($33))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (($24) + ($34<<4)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = $35; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($35|0),($37|0),($31|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $36; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $38 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($38) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $39 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($39); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics14ParticleSystemC2ERKS1_($this,$p) {
 $this = $this|0;
 $p = $p|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$3 = 0, $$34 = 0, $$pre$i = 0, $$pre$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0;
 var $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0;
 var $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0;
 var $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i5 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 36 "modules/graphics/Drawable.h"
 HEAP32[$this>>2] = (15552); //@line 144 "modules/graphics/ParticleSystem.cpp"
 $0 = ((($this)) + 8|0); //@line 101 "modules/graphics/ParticleSystem.cpp"
 $1 = ((($this)) + 24|0); //@line 99 "./common/Object.h"
 $2 = ((($p)) + 24|0); //@line 139 "./common/Object.h"
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;HEAP32[$0+12>>2]=0|0; //@line 102 "modules/graphics/ParticleSystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 139 "./common/Object.h"
 HEAP32[$1>>2] = $3; //@line 99 "./common/Object.h"
 $4 = ($3|0)==(0|0); //@line 101 "./common/Object.h"
 if (!($4)) {
  $5 = HEAP32[$3>>2]|0; //@line 101 "./common/Object.h"
  $6 = ((($5)) + 8|0); //@line 101 "./common/Object.h"
  $7 = HEAP32[$6>>2]|0; //@line 101 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($7|0,($3|0)); //@line 101 "./common/Object.h"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if ($9) {
   $182 = ___cxa_find_matching_catch_2()|0;
   $183 = tempRet0;
   $$3 = $183;$$34 = $182;
   ___resumeException($$34|0);
   // unreachable;
  }
 }
 $10 = ((($this)) + 28|0); //@line 106 "modules/graphics/ParticleSystem.cpp"
 $11 = ((($p)) + 28|0); //@line 106 "modules/graphics/ParticleSystem.cpp"
 $12 = HEAP8[$11>>0]|0; //@line 106 "modules/graphics/ParticleSystem.cpp"
 HEAP8[$10>>0] = $12; //@line 106 "modules/graphics/ParticleSystem.cpp"
 $13 = ((($this)) + 32|0); //@line 107 "modules/graphics/ParticleSystem.cpp"
 $14 = ((($p)) + 32|0); //@line 107 "modules/graphics/ParticleSystem.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 107 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$13>>2] = $15; //@line 107 "modules/graphics/ParticleSystem.cpp"
 $16 = ((($this)) + 36|0); //@line 108 "modules/graphics/ParticleSystem.cpp"
 $17 = ((($p)) + 36|0); //@line 108 "modules/graphics/ParticleSystem.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 108 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$16>>2] = $18; //@line 108 "modules/graphics/ParticleSystem.cpp"
 $19 = ((($this)) + 40|0); //@line 109 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$19>>2] = 0; //@line 109 "modules/graphics/ParticleSystem.cpp"
 $20 = ((($this)) + 44|0); //@line 110 "modules/graphics/ParticleSystem.cpp"
 $21 = ((($p)) + 44|0); //@line 110 "modules/graphics/ParticleSystem.cpp"
 $22 = HEAP32[$21>>2]|0; //@line 110 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$20>>2] = $22; //@line 110 "modules/graphics/ParticleSystem.cpp"
 $23 = ((($this)) + 48|0); //@line 111 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$23>>2] = 0.0; //@line 111 "modules/graphics/ParticleSystem.cpp"
 $24 = ((($this)) + 52|0); //@line 112 "modules/graphics/ParticleSystem.cpp"
 $25 = ((($p)) + 52|0); //@line 112 "modules/graphics/ParticleSystem.cpp"
 $26 = $25; //@line 112 "modules/graphics/ParticleSystem.cpp"
 $27 = $26; //@line 112 "modules/graphics/ParticleSystem.cpp"
 $28 = HEAP32[$27>>2]|0; //@line 112 "modules/graphics/ParticleSystem.cpp"
 $29 = (($26) + 4)|0; //@line 112 "modules/graphics/ParticleSystem.cpp"
 $30 = $29; //@line 112 "modules/graphics/ParticleSystem.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 112 "modules/graphics/ParticleSystem.cpp"
 $32 = $24; //@line 112 "modules/graphics/ParticleSystem.cpp"
 $33 = $32; //@line 112 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$33>>2] = $28; //@line 112 "modules/graphics/ParticleSystem.cpp"
 $34 = (($32) + 4)|0; //@line 112 "modules/graphics/ParticleSystem.cpp"
 $35 = $34; //@line 112 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$35>>2] = $31; //@line 112 "modules/graphics/ParticleSystem.cpp"
 $36 = ((($this)) + 60|0); //@line 113 "modules/graphics/ParticleSystem.cpp"
 $37 = ((($p)) + 60|0); //@line 113 "modules/graphics/ParticleSystem.cpp"
 $38 = $37; //@line 113 "modules/graphics/ParticleSystem.cpp"
 $39 = $38; //@line 113 "modules/graphics/ParticleSystem.cpp"
 $40 = HEAP32[$39>>2]|0; //@line 113 "modules/graphics/ParticleSystem.cpp"
 $41 = (($38) + 4)|0; //@line 113 "modules/graphics/ParticleSystem.cpp"
 $42 = $41; //@line 113 "modules/graphics/ParticleSystem.cpp"
 $43 = HEAP32[$42>>2]|0; //@line 113 "modules/graphics/ParticleSystem.cpp"
 $44 = $36; //@line 113 "modules/graphics/ParticleSystem.cpp"
 $45 = $44; //@line 113 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$45>>2] = $40; //@line 113 "modules/graphics/ParticleSystem.cpp"
 $46 = (($44) + 4)|0; //@line 113 "modules/graphics/ParticleSystem.cpp"
 $47 = $46; //@line 113 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$47>>2] = $43; //@line 113 "modules/graphics/ParticleSystem.cpp"
 $48 = ((($this)) + 68|0); //@line 114 "modules/graphics/ParticleSystem.cpp"
 $49 = ((($p)) + 68|0); //@line 114 "modules/graphics/ParticleSystem.cpp"
 $50 = HEAP32[$49>>2]|0; //@line 114 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$48>>2] = $50; //@line 114 "modules/graphics/ParticleSystem.cpp"
 $51 = ((($this)) + 72|0); //@line 115 "modules/graphics/ParticleSystem.cpp"
 $52 = ((($p)) + 72|0); //@line 115 "modules/graphics/ParticleSystem.cpp"
 $53 = $52; //@line 115 "modules/graphics/ParticleSystem.cpp"
 $54 = $53; //@line 115 "modules/graphics/ParticleSystem.cpp"
 $55 = HEAP32[$54>>2]|0; //@line 115 "modules/graphics/ParticleSystem.cpp"
 $56 = (($53) + 4)|0; //@line 115 "modules/graphics/ParticleSystem.cpp"
 $57 = $56; //@line 115 "modules/graphics/ParticleSystem.cpp"
 $58 = HEAP32[$57>>2]|0; //@line 115 "modules/graphics/ParticleSystem.cpp"
 $59 = $51; //@line 115 "modules/graphics/ParticleSystem.cpp"
 $60 = $59; //@line 115 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$60>>2] = $55; //@line 115 "modules/graphics/ParticleSystem.cpp"
 $61 = (($59) + 4)|0; //@line 115 "modules/graphics/ParticleSystem.cpp"
 $62 = $61; //@line 115 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$62>>2] = $58; //@line 115 "modules/graphics/ParticleSystem.cpp"
 $63 = ((($this)) + 80|0); //@line 116 "modules/graphics/ParticleSystem.cpp"
 $64 = ((($p)) + 80|0); //@line 116 "modules/graphics/ParticleSystem.cpp"
 $65 = HEAP32[$64>>2]|0; //@line 116 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$63>>2] = $65; //@line 116 "modules/graphics/ParticleSystem.cpp"
 $66 = ((($this)) + 84|0); //@line 117 "modules/graphics/ParticleSystem.cpp"
 $67 = HEAP32[$64>>2]|0; //@line 117 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$66>>2] = $67; //@line 117 "modules/graphics/ParticleSystem.cpp"
 $68 = ((($this)) + 88|0); //@line 118 "modules/graphics/ParticleSystem.cpp"
 $69 = ((($p)) + 88|0); //@line 118 "modules/graphics/ParticleSystem.cpp"
 $70 = HEAP32[$69>>2]|0; //@line 118 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$68>>2] = $70; //@line 118 "modules/graphics/ParticleSystem.cpp"
 $71 = ((($this)) + 92|0); //@line 119 "modules/graphics/ParticleSystem.cpp"
 $72 = ((($p)) + 92|0); //@line 119 "modules/graphics/ParticleSystem.cpp"
 $73 = HEAP32[$72>>2]|0; //@line 119 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$71>>2] = $73; //@line 119 "modules/graphics/ParticleSystem.cpp"
 $74 = ((($this)) + 96|0); //@line 120 "modules/graphics/ParticleSystem.cpp"
 $75 = ((($p)) + 96|0); //@line 120 "modules/graphics/ParticleSystem.cpp"
 $76 = HEAP32[$75>>2]|0; //@line 120 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$74>>2] = $76; //@line 120 "modules/graphics/ParticleSystem.cpp"
 $77 = ((($this)) + 100|0); //@line 121 "modules/graphics/ParticleSystem.cpp"
 $78 = ((($p)) + 100|0); //@line 121 "modules/graphics/ParticleSystem.cpp"
 $79 = HEAP32[$78>>2]|0; //@line 121 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$77>>2] = $79; //@line 121 "modules/graphics/ParticleSystem.cpp"
 $80 = ((($this)) + 104|0); //@line 122 "modules/graphics/ParticleSystem.cpp"
 $81 = ((($p)) + 104|0); //@line 122 "modules/graphics/ParticleSystem.cpp"
 $82 = HEAP32[$81>>2]|0; //@line 122 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$80>>2] = $82; //@line 122 "modules/graphics/ParticleSystem.cpp"
 $83 = ((($this)) + 108|0); //@line 123 "modules/graphics/ParticleSystem.cpp"
 $84 = ((($p)) + 108|0); //@line 123 "modules/graphics/ParticleSystem.cpp"
 $85 = HEAP32[$84>>2]|0; //@line 123 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$83>>2] = $85; //@line 123 "modules/graphics/ParticleSystem.cpp"
 $86 = ((($this)) + 112|0); //@line 124 "modules/graphics/ParticleSystem.cpp"
 $87 = ((($p)) + 112|0); //@line 124 "modules/graphics/ParticleSystem.cpp"
 $88 = $87; //@line 124 "modules/graphics/ParticleSystem.cpp"
 $89 = $88; //@line 124 "modules/graphics/ParticleSystem.cpp"
 $90 = HEAP32[$89>>2]|0; //@line 124 "modules/graphics/ParticleSystem.cpp"
 $91 = (($88) + 4)|0; //@line 124 "modules/graphics/ParticleSystem.cpp"
 $92 = $91; //@line 124 "modules/graphics/ParticleSystem.cpp"
 $93 = HEAP32[$92>>2]|0; //@line 124 "modules/graphics/ParticleSystem.cpp"
 $94 = $86; //@line 124 "modules/graphics/ParticleSystem.cpp"
 $95 = $94; //@line 124 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$95>>2] = $90; //@line 124 "modules/graphics/ParticleSystem.cpp"
 $96 = (($94) + 4)|0; //@line 124 "modules/graphics/ParticleSystem.cpp"
 $97 = $96; //@line 124 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$97>>2] = $93; //@line 124 "modules/graphics/ParticleSystem.cpp"
 $98 = ((($this)) + 120|0); //@line 125 "modules/graphics/ParticleSystem.cpp"
 $99 = ((($p)) + 120|0); //@line 125 "modules/graphics/ParticleSystem.cpp"
 $100 = $99; //@line 125 "modules/graphics/ParticleSystem.cpp"
 $101 = $100; //@line 125 "modules/graphics/ParticleSystem.cpp"
 $102 = HEAP32[$101>>2]|0; //@line 125 "modules/graphics/ParticleSystem.cpp"
 $103 = (($100) + 4)|0; //@line 125 "modules/graphics/ParticleSystem.cpp"
 $104 = $103; //@line 125 "modules/graphics/ParticleSystem.cpp"
 $105 = HEAP32[$104>>2]|0; //@line 125 "modules/graphics/ParticleSystem.cpp"
 $106 = $98; //@line 125 "modules/graphics/ParticleSystem.cpp"
 $107 = $106; //@line 125 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$107>>2] = $102; //@line 125 "modules/graphics/ParticleSystem.cpp"
 $108 = (($106) + 4)|0; //@line 125 "modules/graphics/ParticleSystem.cpp"
 $109 = $108; //@line 125 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$109>>2] = $105; //@line 125 "modules/graphics/ParticleSystem.cpp"
 $110 = ((($this)) + 128|0); //@line 126 "modules/graphics/ParticleSystem.cpp"
 $111 = ((($p)) + 128|0); //@line 126 "modules/graphics/ParticleSystem.cpp"
 $112 = HEAP32[$111>>2]|0; //@line 126 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$110>>2] = $112; //@line 126 "modules/graphics/ParticleSystem.cpp"
 $113 = ((($this)) + 132|0); //@line 127 "modules/graphics/ParticleSystem.cpp"
 $114 = ((($p)) + 132|0); //@line 127 "modules/graphics/ParticleSystem.cpp"
 $115 = HEAP32[$114>>2]|0; //@line 127 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$113>>2] = $115; //@line 127 "modules/graphics/ParticleSystem.cpp"
 $116 = ((($this)) + 136|0); //@line 128 "modules/graphics/ParticleSystem.cpp"
 $117 = ((($p)) + 136|0); //@line 128 "modules/graphics/ParticleSystem.cpp"
 $118 = HEAP32[$117>>2]|0; //@line 128 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$116>>2] = $118; //@line 128 "modules/graphics/ParticleSystem.cpp"
 $119 = ((($this)) + 140|0); //@line 129 "modules/graphics/ParticleSystem.cpp"
 $120 = ((($p)) + 140|0); //@line 129 "modules/graphics/ParticleSystem.cpp"
 $121 = HEAP32[$120>>2]|0; //@line 129 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$119>>2] = $121; //@line 129 "modules/graphics/ParticleSystem.cpp"
 $122 = ((($this)) + 144|0); //@line 130 "modules/graphics/ParticleSystem.cpp"
 $123 = ((($p)) + 144|0); //@line 130 "modules/graphics/ParticleSystem.cpp"
 $124 = HEAP32[$123>>2]|0; //@line 130 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$122>>2] = $124; //@line 130 "modules/graphics/ParticleSystem.cpp"
 $125 = ((($this)) + 148|0); //@line 131 "modules/graphics/ParticleSystem.cpp"
 $126 = ((($p)) + 148|0); //@line 131 "modules/graphics/ParticleSystem.cpp"
 $127 = HEAP32[$126>>2]|0; //@line 131 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$125>>2] = $127; //@line 131 "modules/graphics/ParticleSystem.cpp"
 $128 = ((($this)) + 152|0); //@line 132 "modules/graphics/ParticleSystem.cpp"
 $129 = ((($p)) + 152|0); //@line 132 "modules/graphics/ParticleSystem.cpp"
 __THREW__ = 0;
 invoke_vii(3627,($128|0),($129|0)); //@line 132 "modules/graphics/ParticleSystem.cpp"
 $130 = __THREW__; __THREW__ = 0;
 $131 = $130&1;
 if ($131) {
  $184 = ___cxa_find_matching_catch_2()|0;
  $185 = tempRet0;
  $$2 = $185;$$23 = $184;
 } else {
  $132 = ((($this)) + 164|0); //@line 133 "modules/graphics/ParticleSystem.cpp"
  $133 = ((($p)) + 164|0); //@line 133 "modules/graphics/ParticleSystem.cpp"
  $134 = HEAP32[$133>>2]|0; //@line 133 "modules/graphics/ParticleSystem.cpp"
  HEAP32[$132>>2] = $134; //@line 133 "modules/graphics/ParticleSystem.cpp"
  $135 = ((($this)) + 168|0); //@line 134 "modules/graphics/ParticleSystem.cpp"
  $136 = ((($p)) + 168|0); //@line 134 "modules/graphics/ParticleSystem.cpp"
  $137 = HEAP32[$136>>2]|0; //@line 134 "modules/graphics/ParticleSystem.cpp"
  HEAP32[$135>>2] = $137; //@line 134 "modules/graphics/ParticleSystem.cpp"
  $138 = ((($this)) + 172|0); //@line 135 "modules/graphics/ParticleSystem.cpp"
  $139 = ((($p)) + 172|0); //@line 135 "modules/graphics/ParticleSystem.cpp"
  $140 = HEAP32[$139>>2]|0; //@line 135 "modules/graphics/ParticleSystem.cpp"
  HEAP32[$138>>2] = $140; //@line 135 "modules/graphics/ParticleSystem.cpp"
  $141 = ((($this)) + 176|0); //@line 136 "modules/graphics/ParticleSystem.cpp"
  $142 = ((($p)) + 176|0); //@line 136 "modules/graphics/ParticleSystem.cpp"
  $143 = HEAP32[$142>>2]|0; //@line 136 "modules/graphics/ParticleSystem.cpp"
  HEAP32[$141>>2] = $143; //@line 136 "modules/graphics/ParticleSystem.cpp"
  $144 = ((($this)) + 180|0); //@line 137 "modules/graphics/ParticleSystem.cpp"
  $145 = ((($p)) + 180|0); //@line 137 "modules/graphics/ParticleSystem.cpp"
  $146 = HEAP32[$145>>2]|0; //@line 137 "modules/graphics/ParticleSystem.cpp"
  HEAP32[$144>>2] = $146; //@line 137 "modules/graphics/ParticleSystem.cpp"
  $147 = ((($this)) + 184|0); //@line 138 "modules/graphics/ParticleSystem.cpp"
  $148 = ((($p)) + 184|0); //@line 138 "modules/graphics/ParticleSystem.cpp"
  $149 = HEAP32[$148>>2]|0; //@line 138 "modules/graphics/ParticleSystem.cpp"
  HEAP32[$147>>2] = $149; //@line 138 "modules/graphics/ParticleSystem.cpp"
  $150 = ((($this)) + 188|0); //@line 139 "modules/graphics/ParticleSystem.cpp"
  $151 = ((($p)) + 188|0); //@line 139 "modules/graphics/ParticleSystem.cpp"
  $152 = $151; //@line 139 "modules/graphics/ParticleSystem.cpp"
  $153 = $152; //@line 139 "modules/graphics/ParticleSystem.cpp"
  $154 = HEAP32[$153>>2]|0; //@line 139 "modules/graphics/ParticleSystem.cpp"
  $155 = (($152) + 4)|0; //@line 139 "modules/graphics/ParticleSystem.cpp"
  $156 = $155; //@line 139 "modules/graphics/ParticleSystem.cpp"
  $157 = HEAP32[$156>>2]|0; //@line 139 "modules/graphics/ParticleSystem.cpp"
  $158 = $150; //@line 139 "modules/graphics/ParticleSystem.cpp"
  $159 = $158; //@line 139 "modules/graphics/ParticleSystem.cpp"
  HEAP32[$159>>2] = $154; //@line 139 "modules/graphics/ParticleSystem.cpp"
  $160 = (($158) + 4)|0; //@line 139 "modules/graphics/ParticleSystem.cpp"
  $161 = $160; //@line 139 "modules/graphics/ParticleSystem.cpp"
  HEAP32[$161>>2] = $157; //@line 139 "modules/graphics/ParticleSystem.cpp"
  $162 = ((($this)) + 196|0); //@line 140 "modules/graphics/ParticleSystem.cpp"
  $163 = ((($p)) + 196|0); //@line 140 "modules/graphics/ParticleSystem.cpp"
  $164 = HEAP8[$163>>0]|0; //@line 140 "modules/graphics/ParticleSystem.cpp"
  HEAP8[$162>>0] = $164; //@line 140 "modules/graphics/ParticleSystem.cpp"
  $165 = ((($this)) + 200|0); //@line 141 "modules/graphics/ParticleSystem.cpp"
  $166 = ((($p)) + 200|0); //@line 141 "modules/graphics/ParticleSystem.cpp"
  __THREW__ = 0;
  invoke_vii(3628,($165|0),($166|0)); //@line 141 "modules/graphics/ParticleSystem.cpp"
  $167 = __THREW__; __THREW__ = 0;
  $168 = $167&1;
  if ($168) {
   $186 = ___cxa_find_matching_catch_2()|0;
   $187 = tempRet0;
   $$1 = $187;$$12 = $186;
  } else {
   $169 = ((($this)) + 212|0); //@line 142 "modules/graphics/ParticleSystem.cpp"
   $170 = ((($p)) + 212|0); //@line 142 "modules/graphics/ParticleSystem.cpp"
   __THREW__ = 0;
   invoke_vii(3629,($169|0),($170|0)); //@line 142 "modules/graphics/ParticleSystem.cpp"
   $171 = __THREW__; __THREW__ = 0;
   $172 = $171&1;
   if ($172) {
    $188 = ___cxa_find_matching_catch_2()|0;
    $189 = tempRet0;
    $$0 = $189;$$01 = $188;
   } else {
    $173 = ((($this)) + 224|0); //@line 143 "modules/graphics/ParticleSystem.cpp"
    $174 = ((($p)) + 224|0); //@line 143 "modules/graphics/ParticleSystem.cpp"
    $175 = HEAP8[$174>>0]|0; //@line 143 "modules/graphics/ParticleSystem.cpp"
    HEAP8[$173>>0] = $175; //@line 143 "modules/graphics/ParticleSystem.cpp"
    $176 = HEAP32[$this>>2]|0; //@line 145 "modules/graphics/ParticleSystem.cpp"
    $177 = ((($176)) + 24|0); //@line 145 "modules/graphics/ParticleSystem.cpp"
    $178 = HEAP32[$177>>2]|0; //@line 145 "modules/graphics/ParticleSystem.cpp"
    $179 = HEAP32[$16>>2]|0; //@line 145 "modules/graphics/ParticleSystem.cpp"
    __THREW__ = 0;
    invoke_vii($178|0,($this|0),($179|0)); //@line 145 "modules/graphics/ParticleSystem.cpp"
    $180 = __THREW__; __THREW__ = 0;
    $181 = $180&1;
    if (!($181)) {
     return; //@line 146 "modules/graphics/ParticleSystem.cpp"
    }
    $190 = ___cxa_find_matching_catch_2()|0;
    $191 = tempRet0;
    $192 = HEAP32[$169>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $193 = ($192|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($193) {
     $$0 = $191;$$01 = $190;
    } else {
     $194 = ((($this)) + 216|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $195 = HEAP32[$194>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $196 = ($195|0)==($192|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     do {
      if ($196) {
       $210 = $192;
      } else {
       $198 = $195;
       while(1) {
        $197 = ((($198)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$194>>2] = $197; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $199 = HEAP32[$197>>2]|0; //@line 106 "./common/Object.h"
        $200 = ($199|0)==(0|0); //@line 106 "./common/Object.h"
        if ($200) {
         $202 = $197;
        } else {
         $203 = HEAP32[$199>>2]|0; //@line 106 "./common/Object.h"
         $204 = ((($203)) + 12|0); //@line 106 "./common/Object.h"
         $205 = HEAP32[$204>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($205|0,($199|0)); //@line 106 "./common/Object.h"
         $206 = __THREW__; __THREW__ = 0;
         $207 = $206&1;
         if ($207) {
          label = 18;
          break;
         }
         $$pre$i$i$i = HEAP32[$194>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $202 = $$pre$i$i$i;
        }
        $201 = ($202|0)==($192|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($201) {
         label = 19;
         break;
        } else {
         $198 = $202;
        }
       }
       if ((label|0) == 18) {
        $208 = ___cxa_find_matching_catch_3(0|0)|0;
        $209 = tempRet0;
        ___clang_call_terminate($208); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       else if ((label|0) == 19) {
        $$pre$i = HEAP32[$169>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $210 = $$pre$i;
        break;
       }
      }
     } while(0);
     __ZdlPv($210); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $$0 = $191;$$01 = $190;
    }
   }
   $211 = HEAP32[$165>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $212 = ($211|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $213 = $211; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($212) {
    $$1 = $$0;$$12 = $$01;
   } else {
    $214 = ((($this)) + 204|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $215 = HEAP32[$214>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $216 = ($215|0)==($211|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($216)) {
     $scevgep$i$i$i = ((($215)) + -16|0);
     $217 = $scevgep$i$i$i;
     $218 = (($217) - ($213))|0;
     $219 = $218 >>> 4;
     $220 = $219 ^ -1;
     $scevgep4$i$i$i = (($215) + ($220<<4)|0);
     HEAP32[$214>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($211); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $$1 = $$0;$$12 = $$01;
   }
  }
  $221 = HEAP32[$128>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $222 = ($221|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $223 = $221; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($222) {
   $$2 = $$1;$$23 = $$12;
  } else {
   $224 = ((($this)) + 156|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $225 = HEAP32[$224>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $226 = ($225|0)==($221|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($226)) {
    $scevgep$i$i$i5 = ((($225)) + -4|0);
    $227 = $scevgep$i$i$i5;
    $228 = (($227) - ($223))|0;
    $229 = $228 >>> 2;
    $230 = $229 ^ -1;
    $scevgep4$i$i$i6 = (($225) + ($230<<2)|0);
    HEAP32[$224>>2] = $scevgep4$i$i$i6; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($221); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $$2 = $$1;$$23 = $$12;
  }
 }
 $231 = HEAP32[$1>>2]|0; //@line 106 "./common/Object.h"
 $232 = ($231|0)==(0|0); //@line 106 "./common/Object.h"
 if ($232) {
  $$3 = $$2;$$34 = $$23;
  ___resumeException($$34|0);
  // unreachable;
 }
 $233 = HEAP32[$231>>2]|0; //@line 106 "./common/Object.h"
 $234 = ((($233)) + 12|0); //@line 106 "./common/Object.h"
 $235 = HEAP32[$234>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($235|0,($231|0)); //@line 106 "./common/Object.h"
 $236 = __THREW__; __THREW__ = 0;
 $237 = $236&1;
 if ($237) {
  $238 = ___cxa_find_matching_catch_3(0|0)|0;
  $239 = tempRet0;
  ___clang_call_terminate($238); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  $$3 = $$2;$$34 = $$23;
  ___resumeException($$34|0);
  // unreachable;
 }
}
function __ZNSt3__16vectorIfNS_9allocatorIfEEEC2ERKS3_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$01$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $scevgep = 0, $scevgep$i$i$i = 0, $scevgep4 = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $0 = ((($this)) + 4|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($this)) + 8|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($__x)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$__x>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = $5 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ($6|0)==(0); //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($7) {
  return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $8 = ($6>>>0)>(1073741823); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($8) {
  __THREW__ = 0;
  invoke_vi(3528,($this|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if (!($10)) {
   label = 4;
  }
 } else {
  label = 4;
 }
 if ((label|0) == 4) {
  __THREW__ = 0;
  $11 = (invoke_ii(3348,($5|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   HEAP32[$0>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$this>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $14 = (($11) + ($6<<2)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $15 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$15>>2] = $14; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $16 = HEAP32[$__x>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $17 = HEAP32[$2>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $18 = ($16|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($18) {
    return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $19 = $16;
   $scevgep = ((($17)) + -4|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $20 = $scevgep; //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $21 = (($20) - ($19))|0; //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $22 = $21 >>> 2; //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $23 = (($22) + 1)|0; //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $scevgep4 = (($11) + ($23<<2)|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$01$i = $16;$25 = $11;
   while(1) {
    $24 = HEAP32[$$01$i>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$25>>2] = $24; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $26 = ((($25)) + 4|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $27 = ((($$01$i)) + 4|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $28 = ($27|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($28) {
     break;
    } else {
     $$01$i = $27;$25 = $26;
    }
   }
   HEAP32[$0>>2] = $scevgep4; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
 }
 $29 = ___cxa_find_matching_catch_2()|0;
 $30 = tempRet0;
 $31 = HEAP32[$this>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = ($31|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $33 = $31; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($32) {
  ___resumeException($29|0);
  // unreachable;
 }
 $34 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $35 = ($34|0)==($31|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($35)) {
  $scevgep$i$i$i = ((($34)) + -4|0);
  $36 = $scevgep$i$i$i;
  $37 = (($36) - ($33))|0;
  $38 = $37 >>> 2;
  $39 = $38 ^ -1;
  $scevgep4$i$i$i = (($34) + ($39<<2)|0);
  HEAP32[$0>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($31); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($29|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love8graphics6ColorTIfEENS_9allocatorIS4_EEEC2ERKS7_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$01$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $0 = ((($this)) + 4|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($this)) + 8|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($__x)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$__x>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = $5 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ($6|0)==(0); //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($7) {
  return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $8 = ($6>>>0)>(268435455); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($8) {
  __THREW__ = 0;
  invoke_vi(3528,($this|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if (!($10)) {
   label = 4;
  }
 } else {
  label = 4;
 }
 if ((label|0) == 4) {
  __THREW__ = 0;
  $11 = (invoke_ii(3348,($5|0))|0); //@line 1030 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   HEAP32[$0>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$this>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $14 = (($11) + ($6<<4)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $15 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$15>>2] = $14; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $16 = HEAP32[$__x>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $17 = HEAP32[$2>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $18 = ($16|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($18) {
    return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   } else {
    $$01$i = $16;$19 = $11;
   }
   while(1) {
    ;HEAP32[$19>>2]=HEAP32[$$01$i>>2]|0;HEAP32[$19+4>>2]=HEAP32[$$01$i+4>>2]|0;HEAP32[$19+8>>2]=HEAP32[$$01$i+8>>2]|0;HEAP32[$19+12>>2]=HEAP32[$$01$i+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $20 = HEAP32[$0>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $21 = ((($20)) + 16|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$0>>2] = $21; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $22 = ((($$01$i)) + 16|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $23 = ($22|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($23) {
     break;
    } else {
     $$01$i = $22;$19 = $21;
    }
   }
   return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
 }
 $24 = ___cxa_find_matching_catch_2()|0;
 $25 = tempRet0;
 $26 = HEAP32[$this>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = ($26|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $28 = $26; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($27) {
  ___resumeException($24|0);
  // unreachable;
 }
 $29 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = ($29|0)==($26|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($30)) {
  $scevgep$i$i$i = ((($29)) + -16|0);
  $31 = $scevgep$i$i$i;
  $32 = (($31) - ($28))|0;
  $33 = $32 >>> 4;
  $34 = $33 ^ -1;
  $scevgep4$i$i$i = (($29) + ($34<<4)|0);
  HEAP32[$0>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($26); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($24|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love9StrongRefINS1_8graphics4QuadEEENS_9allocatorIS5_EEEC2ERKS8_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$01$i = 0, $$pre$i = 0, $$pre$i$i$i = 0, $$pre2$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $0 = ((($this)) + 4|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($this)) + 8|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($__x)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$__x>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = $5 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ($6|0)==(0); //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($7) {
  return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $8 = ($6>>>0)>(1073741823); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($8) {
  __THREW__ = 0;
  invoke_vi(3528,($this|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if ($10) {
   label = 12;
  } else {
   label = 4;
  }
 } else {
  label = 4;
 }
 do {
  if ((label|0) == 4) {
   __THREW__ = 0;
   $11 = (invoke_ii(3348,($5|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $12 = __THREW__; __THREW__ = 0;
   $13 = $12&1;
   if ($13) {
    label = 12;
   } else {
    HEAP32[$0>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$this>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $14 = (($11) + ($6<<2)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $15 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$15>>2] = $14; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $16 = HEAP32[$__x>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $17 = HEAP32[$2>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $18 = ($16|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($18) {
     return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    $$01$i = $16;$20 = $11;
    while(1) {
     $19 = HEAP32[$$01$i>>2]|0; //@line 139 "./common/Object.h"
     HEAP32[$20>>2] = $19; //@line 99 "./common/Object.h"
     $21 = ($19|0)==(0|0); //@line 101 "./common/Object.h"
     if ($21) {
      $28 = $20;
     } else {
      $22 = HEAP32[$19>>2]|0; //@line 101 "./common/Object.h"
      $23 = ((($22)) + 8|0); //@line 101 "./common/Object.h"
      $24 = HEAP32[$23>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($24|0,($19|0)); //@line 101 "./common/Object.h"
      $25 = __THREW__; __THREW__ = 0;
      $26 = $25&1;
      if ($26) {
       label = 11;
       break;
      }
      $$pre2$i = HEAP32[$0>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $28 = $$pre2$i;
     }
     $27 = ((($28)) + 4|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$0>>2] = $27; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $29 = ((($$01$i)) + 4|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $30 = ($29|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($30) {
      label = 23;
      break;
     } else {
      $$01$i = $29;$20 = $27;
     }
    }
    if ((label|0) == 11) {
     $31 = ___cxa_find_matching_catch_2()|0;
     $32 = tempRet0;
     $lpad$phi$index = $31;$lpad$phi$index2 = $32;
     break;
    }
    else if ((label|0) == 23) {
     return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
   }
  }
 } while(0);
 if ((label|0) == 12) {
  $33 = ___cxa_find_matching_catch_2()|0;
  $34 = tempRet0;
  $lpad$phi$index = $33;$lpad$phi$index2 = $34;
 }
 $35 = HEAP32[$this>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $36 = ($35|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($36) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $37 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $38 = ($37|0)==($35|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($38) {
   $52 = $35;
  } else {
   $40 = $37;
   while(1) {
    $39 = ((($40)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$0>>2] = $39; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $41 = HEAP32[$39>>2]|0; //@line 106 "./common/Object.h"
    $42 = ($41|0)==(0|0); //@line 106 "./common/Object.h"
    if ($42) {
     $44 = $39;
    } else {
     $45 = HEAP32[$41>>2]|0; //@line 106 "./common/Object.h"
     $46 = ((($45)) + 12|0); //@line 106 "./common/Object.h"
     $47 = HEAP32[$46>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($47|0,($41|0)); //@line 106 "./common/Object.h"
     $48 = __THREW__; __THREW__ = 0;
     $49 = $48&1;
     if ($49) {
      label = 19;
      break;
     }
     $$pre$i$i$i = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $44 = $$pre$i$i$i;
    }
    $43 = ($44|0)==($35|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($43) {
     label = 20;
     break;
    } else {
     $40 = $44;
    }
   }
   if ((label|0) == 19) {
    $50 = ___cxa_find_matching_catch_3(0|0)|0;
    $51 = tempRet0;
    ___clang_call_terminate($50); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 20) {
    $$pre$i = HEAP32[$this>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $52 = $$pre$i;
    break;
   }
  }
 } while(0);
 __ZdlPv($52); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZN4love8graphics14ParticleSystem5resetEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 781 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 781 "modules/graphics/ParticleSystem.cpp"
 $2 = ($1|0)==(0|0); //@line 781 "modules/graphics/ParticleSystem.cpp"
 if ($2) {
  return; //@line 790 "modules/graphics/ParticleSystem.cpp"
 }
 $3 = ((($this)) + 12|0); //@line 784 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$3>>2] = $1; //@line 784 "modules/graphics/ParticleSystem.cpp"
 $4 = ((($this)) + 16|0); //@line 785 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$4>>2] = 0; //@line 785 "modules/graphics/ParticleSystem.cpp"
 $5 = ((($this)) + 20|0); //@line 786 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$5>>2] = 0; //@line 786 "modules/graphics/ParticleSystem.cpp"
 $6 = ((($this)) + 40|0); //@line 787 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$6>>2] = 0; //@line 787 "modules/graphics/ParticleSystem.cpp"
 $7 = ((($this)) + 80|0); //@line 788 "modules/graphics/ParticleSystem.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 788 "modules/graphics/ParticleSystem.cpp"
 $9 = ((($this)) + 84|0); //@line 788 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$9>>2] = $8; //@line 788 "modules/graphics/ParticleSystem.cpp"
 $10 = ((($this)) + 48|0); //@line 789 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$10>>2] = 0.0; //@line 789 "modules/graphics/ParticleSystem.cpp"
 return; //@line 790 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem13getBufferSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 36|0); //@line 199 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 199 "modules/graphics/ParticleSystem.cpp"
 return ($1|0); //@line 199 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem11addParticleEf($this,$t) {
 $this = $this|0;
 $t = +$t;
 var $$pre$i = 0, $$pre$i1 = 0, $$pre$phi$i2Z2D = 0, $$pre$phi$iZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0); //@line 825 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 825 "modules/graphics/ParticleSystem.cpp"
 $2 = ((($this)) + 36|0); //@line 825 "modules/graphics/ParticleSystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 825 "modules/graphics/ParticleSystem.cpp"
 $4 = ($1|0)==($3|0); //@line 825 "modules/graphics/ParticleSystem.cpp"
 if ($4) {
  return; //@line 226 "modules/graphics/ParticleSystem.cpp"
 }
 $5 = ((($this)) + 12|0); //@line 208 "modules/graphics/ParticleSystem.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 208 "modules/graphics/ParticleSystem.cpp"
 $7 = ((($6)) + 108|0); //@line 208 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$5>>2] = $7; //@line 208 "modules/graphics/ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystem12initParticleEPNS1_8ParticleEf($this,$6,$t); //@line 209 "modules/graphics/ParticleSystem.cpp"
 $8 = ((($this)) + 32|0); //@line 211 "modules/graphics/ParticleSystem.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 211 "modules/graphics/ParticleSystem.cpp"
 L4: do {
  switch ($9|0) {
  case 2:  {
   $26 = (__ZN4love4math15RandomGenerator4randEv(623936)|0); //@line 341 "modules/graphics/ParticleSystem.cpp"
   $27 = tempRet0; //@line 341 "modules/graphics/ParticleSystem.cpp"
   $28 = HEAP32[$0>>2]|0; //@line 341 "modules/graphics/ParticleSystem.cpp"
   $29 = (_i64Add(($28|0),0,1,0)|0); //@line 341 "modules/graphics/ParticleSystem.cpp"
   $30 = tempRet0; //@line 341 "modules/graphics/ParticleSystem.cpp"
   $31 = (___uremdi3(($26|0),($27|0),($29|0),($30|0))|0); //@line 341 "modules/graphics/ParticleSystem.cpp"
   $32 = tempRet0; //@line 341 "modules/graphics/ParticleSystem.cpp"
   $33 = ($31|0)==($28|0); //@line 344 "modules/graphics/ParticleSystem.cpp"
   $34 = ($32|0)==(0); //@line 344 "modules/graphics/ParticleSystem.cpp"
   $35 = $33 & $34; //@line 344 "modules/graphics/ParticleSystem.cpp"
   if ($35) {
    $36 = ((($this)) + 16|0); //@line 346 "modules/graphics/ParticleSystem.cpp"
    $37 = HEAP32[$36>>2]|0; //@line 346 "modules/graphics/ParticleSystem.cpp"
    $38 = ($37|0)==(0|0); //@line 347 "modules/graphics/ParticleSystem.cpp"
    if (!($38)) {
     HEAP32[$37>>2] = $6; //@line 348 "modules/graphics/ParticleSystem.cpp"
    }
    HEAP32[$6>>2] = 0; //@line 349 "modules/graphics/ParticleSystem.cpp"
    $39 = ((($6)) + 4|0); //@line 350 "modules/graphics/ParticleSystem.cpp"
    HEAP32[$39>>2] = $37; //@line 350 "modules/graphics/ParticleSystem.cpp"
    HEAP32[$36>>2] = $6; //@line 351 "modules/graphics/ParticleSystem.cpp"
    break L4;
   }
   $40 = ((($this)) + 8|0); //@line 356 "modules/graphics/ParticleSystem.cpp"
   $41 = HEAP32[$40>>2]|0; //@line 356 "modules/graphics/ParticleSystem.cpp"
   $42 = (($41) + (($31*108)|0)|0); //@line 356 "modules/graphics/ParticleSystem.cpp"
   $43 = (((($41) + (($31*108)|0)|0)) + 4|0); //@line 357 "modules/graphics/ParticleSystem.cpp"
   $44 = HEAP32[$43>>2]|0; //@line 357 "modules/graphics/ParticleSystem.cpp"
   HEAP32[$43>>2] = $6; //@line 358 "modules/graphics/ParticleSystem.cpp"
   $45 = ($44|0)==(0|0); //@line 359 "modules/graphics/ParticleSystem.cpp"
   if ($45) {
    $46 = ((($this)) + 20|0); //@line 362 "modules/graphics/ParticleSystem.cpp"
    HEAP32[$46>>2] = $6; //@line 362 "modules/graphics/ParticleSystem.cpp"
   } else {
    HEAP32[$44>>2] = $6; //@line 360 "modules/graphics/ParticleSystem.cpp"
   }
   HEAP32[$6>>2] = $42; //@line 363 "modules/graphics/ParticleSystem.cpp"
   $47 = ((($6)) + 4|0); //@line 364 "modules/graphics/ParticleSystem.cpp"
   HEAP32[$47>>2] = $44; //@line 364 "modules/graphics/ParticleSystem.cpp"
   break;
  }
  case 1:  {
   $18 = ((($this)) + 20|0); //@line 324 "modules/graphics/ParticleSystem.cpp"
   $19 = HEAP32[$18>>2]|0; //@line 324 "modules/graphics/ParticleSystem.cpp"
   $20 = ($19|0)==(0|0); //@line 324 "modules/graphics/ParticleSystem.cpp"
   if ($20) {
    HEAP32[$18>>2] = $6; //@line 326 "modules/graphics/ParticleSystem.cpp"
    $21 = ((($6)) + 4|0); //@line 327 "modules/graphics/ParticleSystem.cpp"
    HEAP32[$21>>2] = 0; //@line 327 "modules/graphics/ParticleSystem.cpp"
    $$pre$i1 = ((($this)) + 16|0); //@line 335 "modules/graphics/ParticleSystem.cpp"
    $$pre$phi$i2Z2D = $$pre$i1;
   } else {
    $22 = ((($this)) + 16|0); //@line 331 "modules/graphics/ParticleSystem.cpp"
    $23 = HEAP32[$22>>2]|0; //@line 331 "modules/graphics/ParticleSystem.cpp"
    HEAP32[$23>>2] = $6; //@line 331 "modules/graphics/ParticleSystem.cpp"
    $24 = $23; //@line 332 "modules/graphics/ParticleSystem.cpp"
    $25 = ((($6)) + 4|0); //@line 332 "modules/graphics/ParticleSystem.cpp"
    HEAP32[$25>>2] = $24; //@line 332 "modules/graphics/ParticleSystem.cpp"
    $$pre$phi$i2Z2D = $22;
   }
   HEAP32[$6>>2] = 0; //@line 334 "modules/graphics/ParticleSystem.cpp"
   HEAP32[$$pre$phi$i2Z2D>>2] = $6; //@line 335 "modules/graphics/ParticleSystem.cpp"
   break;
  }
  default: {
   $10 = ((($this)) + 16|0); //@line 308 "modules/graphics/ParticleSystem.cpp"
   $11 = HEAP32[$10>>2]|0; //@line 308 "modules/graphics/ParticleSystem.cpp"
   $12 = ($11|0)==(0|0); //@line 308 "modules/graphics/ParticleSystem.cpp"
   if ($12) {
    HEAP32[$10>>2] = $6; //@line 310 "modules/graphics/ParticleSystem.cpp"
    HEAP32[$6>>2] = 0; //@line 311 "modules/graphics/ParticleSystem.cpp"
    $$pre$i = ((($this)) + 20|0); //@line 319 "modules/graphics/ParticleSystem.cpp"
    $$pre$phi$iZ2D = $$pre$i;
   } else {
    $13 = ((($this)) + 20|0); //@line 315 "modules/graphics/ParticleSystem.cpp"
    $14 = HEAP32[$13>>2]|0; //@line 315 "modules/graphics/ParticleSystem.cpp"
    $15 = ((($14)) + 4|0); //@line 315 "modules/graphics/ParticleSystem.cpp"
    HEAP32[$15>>2] = $6; //@line 315 "modules/graphics/ParticleSystem.cpp"
    $16 = $14; //@line 316 "modules/graphics/ParticleSystem.cpp"
    HEAP32[$6>>2] = $16; //@line 316 "modules/graphics/ParticleSystem.cpp"
    $$pre$phi$iZ2D = $13;
   }
   $17 = ((($6)) + 4|0); //@line 318 "modules/graphics/ParticleSystem.cpp"
   HEAP32[$17>>2] = 0; //@line 318 "modules/graphics/ParticleSystem.cpp"
   HEAP32[$$pre$phi$iZ2D>>2] = $6; //@line 319 "modules/graphics/ParticleSystem.cpp"
  }
  }
 } while(0);
 $48 = HEAP32[$0>>2]|0; //@line 225 "modules/graphics/ParticleSystem.cpp"
 $49 = (($48) + 1)|0; //@line 225 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$0>>2] = $49; //@line 225 "modules/graphics/ParticleSystem.cpp"
 return; //@line 226 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem12initParticleEPNS1_8ParticleEf($this,$p,$t) {
 $this = $this|0;
 $p = $p|0;
 $t = +$t;
 var $$cast = 0, $$sink = 0.0, $0 = 0, $1 = 0.0, $10 = 0.0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0, $105 = 0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0.0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0;
 var $114 = 0.0, $115 = 0.0, $116 = 0, $117 = 0, $118 = 0, $119 = 0.0, $12 = 0.0, $120 = 0.0, $121 = 0, $122 = 0.0, $123 = 0.0, $124 = 0, $125 = 0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0.0, $130 = 0.0, $131 = 0.0;
 var $132 = 0, $133 = 0, $134 = 0.0, $135 = 0.0, $136 = 0, $137 = 0.0, $138 = 0.0, $139 = 0, $14 = 0, $140 = 0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0, $148 = 0, $149 = 0.0, $15 = 0.0;
 var $150 = 0, $151 = 0.0, $152 = 0.0, $153 = 0.0, $154 = 0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0, $163 = 0, $164 = 0.0, $165 = 0, $166 = 0.0, $167 = 0.0, $168 = 0.0;
 var $169 = 0, $17 = 0.0, $170 = 0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0, $178 = 0, $179 = 0.0, $18 = 0, $180 = 0, $181 = 0.0, $182 = 0.0, $183 = 0.0, $184 = 0, $185 = 0, $186 = 0.0;
 var $187 = 0.0, $188 = 0.0, $189 = 0.0, $19 = 0.0, $190 = 0.0, $191 = 0.0, $192 = 0, $193 = 0, $194 = 0.0, $195 = 0.0, $196 = 0, $197 = 0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0.0, $200 = 0.0, $201 = 0.0, $202 = 0, $203 = 0.0;
 var $204 = 0.0, $205 = 0, $206 = 0, $207 = 0.0, $208 = 0.0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0.0, $212 = 0.0, $213 = 0.0, $214 = 0, $215 = 0, $216 = 0.0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0.0, $23 = 0.0, $230 = 0, $231 = 0.0, $232 = 0, $233 = 0.0, $234 = 0, $235 = 0.0, $236 = 0, $237 = 0.0, $238 = 0.0, $239 = 0.0, $24 = 0.0;
 var $240 = 0.0, $241 = 0.0, $242 = 0, $243 = 0, $244 = 0.0, $245 = 0.0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0.0, $250 = 0.0, $251 = 0, $252 = 0.0, $253 = 0.0, $254 = 0.0, $255 = 0.0, $256 = 0.0, $257 = 0.0, $258 = 0.0;
 var $259 = 0, $26 = 0.0, $260 = 0, $261 = 0.0, $262 = 0.0, $263 = 0.0, $264 = 0.0, $265 = 0.0, $266 = 0.0, $267 = 0.0, $268 = 0, $269 = 0.0, $27 = 0.0, $270 = 0.0, $271 = 0, $272 = 0, $273 = 0.0, $274 = 0.0, $275 = 0.0, $276 = 0.0;
 var $277 = 0.0, $278 = 0.0, $279 = 0, $28 = 0.0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0.0, $285 = 0.0, $286 = 0.0, $287 = 0.0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $3 = 0.0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0;
 var $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0, $56 = 0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0.0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0, $66 = 0.0, $67 = 0.0, $68 = 0.0;
 var $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0, $8 = 0.0, $80 = 0, $81 = 0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0.0, $86 = 0.0;
 var $87 = 0, $88 = 0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0, $98 = 0.0, $99 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 237 "./common/Vector.h"
 $1 = +HEAPF32[$0>>2]; //@line 237 "./common/Vector.h"
 $2 = ((($this)) + 60|0); //@line 305 "./common/Vector.h"
 $3 = +HEAPF32[$2>>2]; //@line 305 "./common/Vector.h"
 $4 = $1 - $3; //@line 237 "./common/Vector.h"
 $5 = ((($this)) + 56|0); //@line 237 "./common/Vector.h"
 $6 = +HEAPF32[$5>>2]; //@line 237 "./common/Vector.h"
 $7 = ((($this)) + 64|0); //@line 310 "./common/Vector.h"
 $8 = +HEAPF32[$7>>2]; //@line 310 "./common/Vector.h"
 $9 = $6 - $8; //@line 237 "./common/Vector.h"
 $10 = $4 * $t; //@line 242 "./common/Vector.h"
 $11 = $9 * $t; //@line 242 "./common/Vector.h"
 $12 = $3 + $10; //@line 232 "./common/Vector.h"
 $13 = $8 + $11; //@line 232 "./common/Vector.h"
 $14 = ((($this)) + 88|0); //@line 235 "modules/graphics/ParticleSystem.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 235 "modules/graphics/ParticleSystem.cpp"
 $16 = ((($this)) + 92|0); //@line 236 "modules/graphics/ParticleSystem.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 236 "modules/graphics/ParticleSystem.cpp"
 $18 = $15 == $17; //@line 237 "modules/graphics/ParticleSystem.cpp"
 if ($18) {
  $$sink = $15;
 } else {
  $19 = $15; //@line 240 "modules/graphics/ParticleSystem.cpp"
  $20 = $17; //@line 240 "modules/graphics/ParticleSystem.cpp"
  $21 = (__ZN4love4math15RandomGenerator4randEv(623936)|0); //@line 76 "./modules/math/RandomGenerator.h"
  $22 = tempRet0; //@line 76 "./modules/math/RandomGenerator.h"
  $23 = (+($21>>>0)) + (4294967296.0*(+($22>>>0))); //@line 76 "./modules/math/RandomGenerator.h"
  $24 = $23 * 5.4210108624275222E-20; //@line 76 "./modules/math/RandomGenerator.h"
  $25 = $20 - $19; //@line 96 "./modules/math/RandomGenerator.h"
  $26 = $25 * $24; //@line 96 "./modules/math/RandomGenerator.h"
  $27 = $19 + $26; //@line 96 "./modules/math/RandomGenerator.h"
  $28 = $27; //@line 240 "modules/graphics/ParticleSystem.cpp"
  $$sink = $28;
 }
 $29 = ((($p)) + 12|0); //@line 240 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$29>>2] = $$sink; //@line 240 "modules/graphics/ParticleSystem.cpp"
 $30 = ((($p)) + 8|0); //@line 241 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$30>>2] = $$sink; //@line 241 "modules/graphics/ParticleSystem.cpp"
 $31 = ((($p)) + 16|0); //@line 243 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$31>>2] = $12; //@line 243 "modules/graphics/ParticleSystem.cpp"
 $32 = ((($p)) + 20|0); //@line 243 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$32>>2] = $13; //@line 243 "modules/graphics/ParticleSystem.cpp"
 $33 = ((($this)) + 68|0); //@line 245 "modules/graphics/ParticleSystem.cpp"
 $34 = HEAP32[$33>>2]|0; //@line 245 "modules/graphics/ParticleSystem.cpp"
 switch ($34|0) {
 case 1:  {
  $35 = ((($this)) + 72|0); //@line 305 "./common/Vector.h"
  $36 = +HEAPF32[$35>>2]; //@line 305 "./common/Vector.h"
  $37 = -$36; //@line 248 "modules/graphics/ParticleSystem.cpp"
  $38 = $37; //@line 248 "modules/graphics/ParticleSystem.cpp"
  $39 = $36; //@line 248 "modules/graphics/ParticleSystem.cpp"
  $40 = (__ZN4love4math15RandomGenerator4randEv(623936)|0); //@line 76 "./modules/math/RandomGenerator.h"
  $41 = tempRet0; //@line 76 "./modules/math/RandomGenerator.h"
  $42 = (+($40>>>0)) + (4294967296.0*(+($41>>>0))); //@line 76 "./modules/math/RandomGenerator.h"
  $43 = $42 * 5.4210108624275222E-20; //@line 76 "./modules/math/RandomGenerator.h"
  $44 = $39 + $39; //@line 96 "./modules/math/RandomGenerator.h"
  $45 = $44 * $43; //@line 96 "./modules/math/RandomGenerator.h"
  $46 = $38 + $45; //@line 96 "./modules/math/RandomGenerator.h"
  $47 = $46; //@line 248 "modules/graphics/ParticleSystem.cpp"
  $48 = +HEAPF32[$31>>2]; //@line 248 "modules/graphics/ParticleSystem.cpp"
  $49 = $48 + $47; //@line 248 "modules/graphics/ParticleSystem.cpp"
  HEAPF32[$31>>2] = $49; //@line 248 "modules/graphics/ParticleSystem.cpp"
  $50 = ((($this)) + 76|0); //@line 310 "./common/Vector.h"
  $51 = +HEAPF32[$50>>2]; //@line 310 "./common/Vector.h"
  $52 = -$51; //@line 249 "modules/graphics/ParticleSystem.cpp"
  $53 = $52; //@line 249 "modules/graphics/ParticleSystem.cpp"
  $54 = $51; //@line 249 "modules/graphics/ParticleSystem.cpp"
  $55 = (__ZN4love4math15RandomGenerator4randEv(623936)|0); //@line 76 "./modules/math/RandomGenerator.h"
  $56 = tempRet0; //@line 76 "./modules/math/RandomGenerator.h"
  $57 = (+($55>>>0)) + (4294967296.0*(+($56>>>0))); //@line 76 "./modules/math/RandomGenerator.h"
  $58 = $57 * 5.4210108624275222E-20; //@line 76 "./modules/math/RandomGenerator.h"
  $59 = $54 + $54; //@line 96 "./modules/math/RandomGenerator.h"
  $60 = $59 * $58; //@line 96 "./modules/math/RandomGenerator.h"
  $61 = $53 + $60; //@line 96 "./modules/math/RandomGenerator.h"
  $62 = $61; //@line 249 "modules/graphics/ParticleSystem.cpp"
  $63 = +HEAPF32[$32>>2]; //@line 249 "modules/graphics/ParticleSystem.cpp"
  $64 = $63 + $62; //@line 249 "modules/graphics/ParticleSystem.cpp"
  HEAPF32[$32>>2] = $64; //@line 249 "modules/graphics/ParticleSystem.cpp"
  break;
 }
 case 2:  {
  $65 = ((($this)) + 72|0); //@line 305 "./common/Vector.h"
  $66 = +HEAPF32[$65>>2]; //@line 305 "./common/Vector.h"
  $67 = $66; //@line 252 "modules/graphics/ParticleSystem.cpp"
  $68 = (+__ZN4love4math15RandomGenerator12randomNormalEd(623936,$67)); //@line 252 "modules/graphics/ParticleSystem.cpp"
  $69 = $68; //@line 252 "modules/graphics/ParticleSystem.cpp"
  $70 = +HEAPF32[$31>>2]; //@line 252 "modules/graphics/ParticleSystem.cpp"
  $71 = $70 + $69; //@line 252 "modules/graphics/ParticleSystem.cpp"
  HEAPF32[$31>>2] = $71; //@line 252 "modules/graphics/ParticleSystem.cpp"
  $72 = ((($this)) + 76|0); //@line 310 "./common/Vector.h"
  $73 = +HEAPF32[$72>>2]; //@line 310 "./common/Vector.h"
  $74 = $73; //@line 253 "modules/graphics/ParticleSystem.cpp"
  $75 = (+__ZN4love4math15RandomGenerator12randomNormalEd(623936,$74)); //@line 253 "modules/graphics/ParticleSystem.cpp"
  $76 = $75; //@line 253 "modules/graphics/ParticleSystem.cpp"
  $77 = +HEAPF32[$32>>2]; //@line 253 "modules/graphics/ParticleSystem.cpp"
  $78 = $77 + $76; //@line 253 "modules/graphics/ParticleSystem.cpp"
  HEAPF32[$32>>2] = $78; //@line 253 "modules/graphics/ParticleSystem.cpp"
  break;
 }
 default: {
 }
 }
 $79 = ((($p)) + 24|0); //@line 260 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$79>>2] = $12; //@line 260 "modules/graphics/ParticleSystem.cpp"
 $80 = ((($p)) + 28|0); //@line 260 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$80>>2] = $13; //@line 260 "modules/graphics/ParticleSystem.cpp"
 $81 = ((($this)) + 104|0); //@line 262 "modules/graphics/ParticleSystem.cpp"
 $82 = +HEAPF32[$81>>2]; //@line 262 "modules/graphics/ParticleSystem.cpp"
 $83 = ((($this)) + 108|0); //@line 263 "modules/graphics/ParticleSystem.cpp"
 $84 = +HEAPF32[$83>>2]; //@line 263 "modules/graphics/ParticleSystem.cpp"
 $85 = $82; //@line 264 "modules/graphics/ParticleSystem.cpp"
 $86 = $84; //@line 264 "modules/graphics/ParticleSystem.cpp"
 $87 = (__ZN4love4math15RandomGenerator4randEv(623936)|0); //@line 76 "./modules/math/RandomGenerator.h"
 $88 = tempRet0; //@line 76 "./modules/math/RandomGenerator.h"
 $89 = (+($87>>>0)) + (4294967296.0*(+($88>>>0))); //@line 76 "./modules/math/RandomGenerator.h"
 $90 = $89 * 5.4210108624275222E-20; //@line 76 "./modules/math/RandomGenerator.h"
 $91 = $86 - $85; //@line 96 "./modules/math/RandomGenerator.h"
 $92 = $91 * $90; //@line 96 "./modules/math/RandomGenerator.h"
 $93 = $85 + $92; //@line 96 "./modules/math/RandomGenerator.h"
 $94 = $93; //@line 264 "modules/graphics/ParticleSystem.cpp"
 $95 = ((($this)) + 96|0); //@line 266 "modules/graphics/ParticleSystem.cpp"
 $96 = +HEAPF32[$95>>2]; //@line 266 "modules/graphics/ParticleSystem.cpp"
 $97 = ((($this)) + 100|0); //@line 266 "modules/graphics/ParticleSystem.cpp"
 $98 = +HEAPF32[$97>>2]; //@line 266 "modules/graphics/ParticleSystem.cpp"
 $99 = $98 * 0.5; //@line 266 "modules/graphics/ParticleSystem.cpp"
 $100 = $96 - $99; //@line 266 "modules/graphics/ParticleSystem.cpp"
 $101 = $96 + $99; //@line 267 "modules/graphics/ParticleSystem.cpp"
 $102 = $100; //@line 268 "modules/graphics/ParticleSystem.cpp"
 $103 = $101; //@line 268 "modules/graphics/ParticleSystem.cpp"
 $104 = (__ZN4love4math15RandomGenerator4randEv(623936)|0); //@line 76 "./modules/math/RandomGenerator.h"
 $105 = tempRet0; //@line 76 "./modules/math/RandomGenerator.h"
 $106 = (+($104>>>0)) + (4294967296.0*(+($105>>>0))); //@line 76 "./modules/math/RandomGenerator.h"
 $107 = $106 * 5.4210108624275222E-20; //@line 76 "./modules/math/RandomGenerator.h"
 $108 = $103 - $102; //@line 96 "./modules/math/RandomGenerator.h"
 $109 = $107 * $108; //@line 96 "./modules/math/RandomGenerator.h"
 $110 = $102 + $109; //@line 96 "./modules/math/RandomGenerator.h"
 $111 = $110; //@line 268 "modules/graphics/ParticleSystem.cpp"
 $112 = (+Math_cos((+$111))); //@line 270 "modules/graphics/ParticleSystem.cpp"
 $113 = (+Math_sin((+$111))); //@line 270 "modules/graphics/ParticleSystem.cpp"
 $114 = $112 * $94; //@line 242 "./common/Vector.h"
 $115 = $113 * $94; //@line 242 "./common/Vector.h"
 $116 = ((($p)) + 32|0); //@line 270 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$116>>2] = $114; //@line 270 "modules/graphics/ParticleSystem.cpp"
 $117 = ((($p)) + 36|0); //@line 270 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$117>>2] = $115; //@line 270 "modules/graphics/ParticleSystem.cpp"
 $118 = ((($this)) + 112|0); //@line 272 "modules/graphics/ParticleSystem.cpp"
 $119 = +HEAPF32[$118>>2]; //@line 272 "modules/graphics/ParticleSystem.cpp"
 $120 = $119; //@line 272 "modules/graphics/ParticleSystem.cpp"
 $121 = ((($this)) + 120|0); //@line 272 "modules/graphics/ParticleSystem.cpp"
 $122 = +HEAPF32[$121>>2]; //@line 272 "modules/graphics/ParticleSystem.cpp"
 $123 = $122; //@line 272 "modules/graphics/ParticleSystem.cpp"
 $124 = (__ZN4love4math15RandomGenerator4randEv(623936)|0); //@line 76 "./modules/math/RandomGenerator.h"
 $125 = tempRet0; //@line 76 "./modules/math/RandomGenerator.h"
 $126 = (+($124>>>0)) + (4294967296.0*(+($125>>>0))); //@line 76 "./modules/math/RandomGenerator.h"
 $127 = $126 * 5.4210108624275222E-20; //@line 76 "./modules/math/RandomGenerator.h"
 $128 = $123 - $120; //@line 96 "./modules/math/RandomGenerator.h"
 $129 = $128 * $127; //@line 96 "./modules/math/RandomGenerator.h"
 $130 = $120 + $129; //@line 96 "./modules/math/RandomGenerator.h"
 $131 = $130; //@line 272 "modules/graphics/ParticleSystem.cpp"
 $132 = ((($p)) + 40|0); //@line 272 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$132>>2] = $131; //@line 272 "modules/graphics/ParticleSystem.cpp"
 $133 = ((($this)) + 116|0); //@line 273 "modules/graphics/ParticleSystem.cpp"
 $134 = +HEAPF32[$133>>2]; //@line 273 "modules/graphics/ParticleSystem.cpp"
 $135 = $134; //@line 273 "modules/graphics/ParticleSystem.cpp"
 $136 = ((($this)) + 124|0); //@line 273 "modules/graphics/ParticleSystem.cpp"
 $137 = +HEAPF32[$136>>2]; //@line 273 "modules/graphics/ParticleSystem.cpp"
 $138 = $137; //@line 273 "modules/graphics/ParticleSystem.cpp"
 $139 = (__ZN4love4math15RandomGenerator4randEv(623936)|0); //@line 76 "./modules/math/RandomGenerator.h"
 $140 = tempRet0; //@line 76 "./modules/math/RandomGenerator.h"
 $141 = (+($139>>>0)) + (4294967296.0*(+($140>>>0))); //@line 76 "./modules/math/RandomGenerator.h"
 $142 = $141 * 5.4210108624275222E-20; //@line 76 "./modules/math/RandomGenerator.h"
 $143 = $138 - $135; //@line 96 "./modules/math/RandomGenerator.h"
 $144 = $143 * $142; //@line 96 "./modules/math/RandomGenerator.h"
 $145 = $135 + $144; //@line 96 "./modules/math/RandomGenerator.h"
 $146 = $145; //@line 273 "modules/graphics/ParticleSystem.cpp"
 $147 = ((($p)) + 44|0); //@line 273 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$147>>2] = $146; //@line 273 "modules/graphics/ParticleSystem.cpp"
 $148 = ((($this)) + 128|0); //@line 275 "modules/graphics/ParticleSystem.cpp"
 $149 = +HEAPF32[$148>>2]; //@line 275 "modules/graphics/ParticleSystem.cpp"
 $150 = ((($this)) + 132|0); //@line 276 "modules/graphics/ParticleSystem.cpp"
 $151 = +HEAPF32[$150>>2]; //@line 276 "modules/graphics/ParticleSystem.cpp"
 $152 = $149; //@line 277 "modules/graphics/ParticleSystem.cpp"
 $153 = $151; //@line 277 "modules/graphics/ParticleSystem.cpp"
 $154 = (__ZN4love4math15RandomGenerator4randEv(623936)|0); //@line 76 "./modules/math/RandomGenerator.h"
 $155 = tempRet0; //@line 76 "./modules/math/RandomGenerator.h"
 $156 = (+($154>>>0)) + (4294967296.0*(+($155>>>0))); //@line 76 "./modules/math/RandomGenerator.h"
 $157 = $156 * 5.4210108624275222E-20; //@line 76 "./modules/math/RandomGenerator.h"
 $158 = $153 - $152; //@line 96 "./modules/math/RandomGenerator.h"
 $159 = $158 * $157; //@line 96 "./modules/math/RandomGenerator.h"
 $160 = $152 + $159; //@line 96 "./modules/math/RandomGenerator.h"
 $161 = $160; //@line 277 "modules/graphics/ParticleSystem.cpp"
 $162 = ((($p)) + 48|0); //@line 277 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$162>>2] = $161; //@line 277 "modules/graphics/ParticleSystem.cpp"
 $163 = ((($this)) + 136|0); //@line 279 "modules/graphics/ParticleSystem.cpp"
 $164 = +HEAPF32[$163>>2]; //@line 279 "modules/graphics/ParticleSystem.cpp"
 $165 = ((($this)) + 140|0); //@line 280 "modules/graphics/ParticleSystem.cpp"
 $166 = +HEAPF32[$165>>2]; //@line 280 "modules/graphics/ParticleSystem.cpp"
 $167 = $164; //@line 281 "modules/graphics/ParticleSystem.cpp"
 $168 = $166; //@line 281 "modules/graphics/ParticleSystem.cpp"
 $169 = (__ZN4love4math15RandomGenerator4randEv(623936)|0); //@line 76 "./modules/math/RandomGenerator.h"
 $170 = tempRet0; //@line 76 "./modules/math/RandomGenerator.h"
 $171 = (+($169>>>0)) + (4294967296.0*(+($170>>>0))); //@line 76 "./modules/math/RandomGenerator.h"
 $172 = $171 * 5.4210108624275222E-20; //@line 76 "./modules/math/RandomGenerator.h"
 $173 = $168 - $167; //@line 96 "./modules/math/RandomGenerator.h"
 $174 = $173 * $172; //@line 96 "./modules/math/RandomGenerator.h"
 $175 = $167 + $174; //@line 96 "./modules/math/RandomGenerator.h"
 $176 = $175; //@line 281 "modules/graphics/ParticleSystem.cpp"
 $177 = ((($p)) + 52|0); //@line 281 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$177>>2] = $176; //@line 281 "modules/graphics/ParticleSystem.cpp"
 $178 = ((($this)) + 144|0); //@line 283 "modules/graphics/ParticleSystem.cpp"
 $179 = +HEAPF32[$178>>2]; //@line 283 "modules/graphics/ParticleSystem.cpp"
 $180 = ((($this)) + 148|0); //@line 284 "modules/graphics/ParticleSystem.cpp"
 $181 = +HEAPF32[$180>>2]; //@line 284 "modules/graphics/ParticleSystem.cpp"
 $182 = $179; //@line 285 "modules/graphics/ParticleSystem.cpp"
 $183 = $181; //@line 285 "modules/graphics/ParticleSystem.cpp"
 $184 = (__ZN4love4math15RandomGenerator4randEv(623936)|0); //@line 76 "./modules/math/RandomGenerator.h"
 $185 = tempRet0; //@line 76 "./modules/math/RandomGenerator.h"
 $186 = (+($184>>>0)) + (4294967296.0*(+($185>>>0))); //@line 76 "./modules/math/RandomGenerator.h"
 $187 = $186 * 5.4210108624275222E-20; //@line 76 "./modules/math/RandomGenerator.h"
 $188 = $183 - $182; //@line 96 "./modules/math/RandomGenerator.h"
 $189 = $188 * $187; //@line 96 "./modules/math/RandomGenerator.h"
 $190 = $182 + $189; //@line 96 "./modules/math/RandomGenerator.h"
 $191 = $190; //@line 285 "modules/graphics/ParticleSystem.cpp"
 $192 = ((($p)) + 56|0); //@line 285 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$192>>2] = $191; //@line 285 "modules/graphics/ParticleSystem.cpp"
 $193 = ((($this)) + 164|0); //@line 287 "modules/graphics/ParticleSystem.cpp"
 $194 = +HEAPF32[$193>>2]; //@line 287 "modules/graphics/ParticleSystem.cpp"
 $195 = $194; //@line 287 "modules/graphics/ParticleSystem.cpp"
 $196 = (__ZN4love4math15RandomGenerator4randEv(623936)|0); //@line 76 "./modules/math/RandomGenerator.h"
 $197 = tempRet0; //@line 76 "./modules/math/RandomGenerator.h"
 $198 = (+($196>>>0)) + (4294967296.0*(+($197>>>0))); //@line 76 "./modules/math/RandomGenerator.h"
 $199 = $198 * 5.4210108624275222E-20; //@line 76 "./modules/math/RandomGenerator.h"
 $200 = $195 * $199; //@line 86 "./modules/math/RandomGenerator.h"
 $201 = $200; //@line 287 "modules/graphics/ParticleSystem.cpp"
 $202 = ((($p)) + 64|0); //@line 287 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$202>>2] = $201; //@line 287 "modules/graphics/ParticleSystem.cpp"
 $203 = +HEAPF32[$193>>2]; //@line 288 "modules/graphics/ParticleSystem.cpp"
 $204 = $203; //@line 288 "modules/graphics/ParticleSystem.cpp"
 $205 = (__ZN4love4math15RandomGenerator4randEv(623936)|0); //@line 76 "./modules/math/RandomGenerator.h"
 $206 = tempRet0; //@line 76 "./modules/math/RandomGenerator.h"
 $207 = (+($205>>>0)) + (4294967296.0*(+($206>>>0))); //@line 76 "./modules/math/RandomGenerator.h"
 $208 = $207 * 5.4210108624275222E-20; //@line 76 "./modules/math/RandomGenerator.h"
 $209 = $204 * $208; //@line 86 "./modules/math/RandomGenerator.h"
 $210 = $209; //@line 288 "modules/graphics/ParticleSystem.cpp"
 $211 = 1.0 - $210; //@line 288 "modules/graphics/ParticleSystem.cpp"
 $212 = +HEAPF32[$202>>2]; //@line 288 "modules/graphics/ParticleSystem.cpp"
 $213 = $211 - $212; //@line 288 "modules/graphics/ParticleSystem.cpp"
 $214 = ((($p)) + 68|0); //@line 288 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$214>>2] = $213; //@line 288 "modules/graphics/ParticleSystem.cpp"
 $215 = ((($this)) + 152|0); //@line 289 "modules/graphics/ParticleSystem.cpp"
 $216 = $212 + -0.5; //@line 289 "modules/graphics/ParticleSystem.cpp"
 $217 = (~~(($216))>>>0); //@line 289 "modules/graphics/ParticleSystem.cpp"
 $218 = ((($this)) + 156|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $219 = HEAP32[$218>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $220 = HEAP32[$215>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $221 = (($219) - ($220))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $222 = $221 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $223 = (($222) + -1)|0; //@line 289 "modules/graphics/ParticleSystem.cpp"
 $224 = Math_imul($223, $217)|0; //@line 289 "modules/graphics/ParticleSystem.cpp"
 $$cast = $220; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $225 = (($$cast) + ($224<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $226 = HEAP32[$225>>2]|0; //@line 289 "modules/graphics/ParticleSystem.cpp"
 $227 = ((($p)) + 60|0); //@line 289 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$227>>2] = $226; //@line 289 "modules/graphics/ParticleSystem.cpp"
 $228 = ((($this)) + 168|0); //@line 291 "modules/graphics/ParticleSystem.cpp"
 $229 = +HEAPF32[$228>>2]; //@line 291 "modules/graphics/ParticleSystem.cpp"
 $230 = ((($this)) + 172|0); //@line 292 "modules/graphics/ParticleSystem.cpp"
 $231 = +HEAPF32[$230>>2]; //@line 292 "modules/graphics/ParticleSystem.cpp"
 $232 = ((($this)) + 176|0); //@line 293 "modules/graphics/ParticleSystem.cpp"
 $233 = +HEAPF32[$232>>2]; //@line 293 "modules/graphics/ParticleSystem.cpp"
 $234 = ((($this)) + 180|0); //@line 293 "modules/graphics/ParticleSystem.cpp"
 $235 = +HEAPF32[$234>>2]; //@line 293 "modules/graphics/ParticleSystem.cpp"
 $236 = ((($this)) + 184|0); //@line 293 "modules/graphics/ParticleSystem.cpp"
 $237 = +HEAPF32[$236>>2]; //@line 293 "modules/graphics/ParticleSystem.cpp"
 $238 = $235 * 0.5; //@line 45 "modules/graphics/ParticleSystem.cpp"
 $239 = $238 * $237; //@line 45 "modules/graphics/ParticleSystem.cpp"
 $240 = $233 - $239; //@line 45 "modules/graphics/ParticleSystem.cpp"
 $241 = $233 + $239; //@line 46 "modules/graphics/ParticleSystem.cpp"
 $242 = (__ZN4love4math15RandomGenerator4randEv(623936)|0); //@line 76 "./modules/math/RandomGenerator.h"
 $243 = tempRet0; //@line 76 "./modules/math/RandomGenerator.h"
 $244 = (+($242>>>0)) + (4294967296.0*(+($243>>>0))); //@line 76 "./modules/math/RandomGenerator.h"
 $245 = $244 * 5.4210108624275222E-20; //@line 76 "./modules/math/RandomGenerator.h"
 $246 = $245; //@line 47 "modules/graphics/ParticleSystem.cpp"
 $247 = 1.0 - $246; //@line 48 "modules/graphics/ParticleSystem.cpp"
 $248 = $240 * $247; //@line 48 "modules/graphics/ParticleSystem.cpp"
 $249 = $241 * $246; //@line 48 "modules/graphics/ParticleSystem.cpp"
 $250 = $249 + $248; //@line 48 "modules/graphics/ParticleSystem.cpp"
 $251 = ((($p)) + 80|0); //@line 293 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$251>>2] = $250; //@line 293 "modules/graphics/ParticleSystem.cpp"
 $252 = +HEAPF32[$234>>2]; //@line 294 "modules/graphics/ParticleSystem.cpp"
 $253 = +HEAPF32[$232>>2]; //@line 294 "modules/graphics/ParticleSystem.cpp"
 $254 = +HEAPF32[$236>>2]; //@line 294 "modules/graphics/ParticleSystem.cpp"
 $255 = $253 * 0.5; //@line 45 "modules/graphics/ParticleSystem.cpp"
 $256 = $255 * $254; //@line 45 "modules/graphics/ParticleSystem.cpp"
 $257 = $252 - $256; //@line 45 "modules/graphics/ParticleSystem.cpp"
 $258 = $252 + $256; //@line 46 "modules/graphics/ParticleSystem.cpp"
 $259 = (__ZN4love4math15RandomGenerator4randEv(623936)|0); //@line 76 "./modules/math/RandomGenerator.h"
 $260 = tempRet0; //@line 76 "./modules/math/RandomGenerator.h"
 $261 = (+($259>>>0)) + (4294967296.0*(+($260>>>0))); //@line 76 "./modules/math/RandomGenerator.h"
 $262 = $261 * 5.4210108624275222E-20; //@line 76 "./modules/math/RandomGenerator.h"
 $263 = $262; //@line 47 "modules/graphics/ParticleSystem.cpp"
 $264 = 1.0 - $263; //@line 48 "modules/graphics/ParticleSystem.cpp"
 $265 = $257 * $264; //@line 48 "modules/graphics/ParticleSystem.cpp"
 $266 = $258 * $263; //@line 48 "modules/graphics/ParticleSystem.cpp"
 $267 = $266 + $265; //@line 48 "modules/graphics/ParticleSystem.cpp"
 $268 = ((($p)) + 84|0); //@line 294 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$268>>2] = $267; //@line 294 "modules/graphics/ParticleSystem.cpp"
 $269 = $229; //@line 295 "modules/graphics/ParticleSystem.cpp"
 $270 = $231; //@line 295 "modules/graphics/ParticleSystem.cpp"
 $271 = (__ZN4love4math15RandomGenerator4randEv(623936)|0); //@line 76 "./modules/math/RandomGenerator.h"
 $272 = tempRet0; //@line 76 "./modules/math/RandomGenerator.h"
 $273 = (+($271>>>0)) + (4294967296.0*(+($272>>>0))); //@line 76 "./modules/math/RandomGenerator.h"
 $274 = $273 * 5.4210108624275222E-20; //@line 76 "./modules/math/RandomGenerator.h"
 $275 = $270 - $269; //@line 96 "./modules/math/RandomGenerator.h"
 $276 = $275 * $274; //@line 96 "./modules/math/RandomGenerator.h"
 $277 = $269 + $276; //@line 96 "./modules/math/RandomGenerator.h"
 $278 = $277; //@line 295 "modules/graphics/ParticleSystem.cpp"
 $279 = ((($p)) + 72|0); //@line 295 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$279>>2] = $278; //@line 295 "modules/graphics/ParticleSystem.cpp"
 $280 = ((($p)) + 76|0); //@line 297 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$280>>2] = $278; //@line 297 "modules/graphics/ParticleSystem.cpp"
 $281 = ((($this)) + 224|0); //@line 298 "modules/graphics/ParticleSystem.cpp"
 $282 = HEAP8[$281>>0]|0; //@line 298 "modules/graphics/ParticleSystem.cpp"
 $283 = ($282<<24>>24)==(0); //@line 298 "modules/graphics/ParticleSystem.cpp"
 if ($283) {
  $288 = ((($p)) + 88|0); //@line 301 "modules/graphics/ParticleSystem.cpp"
  $289 = ((($this)) + 200|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $290 = HEAP32[$289>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  ;HEAP32[$288>>2]=HEAP32[$290>>2]|0;HEAP32[$288+4>>2]=HEAP32[$290+4>>2]|0;HEAP32[$288+8>>2]=HEAP32[$290+8>>2]|0;HEAP32[$288+12>>2]=HEAP32[$290+12>>2]|0; //@line 301 "modules/graphics/ParticleSystem.cpp"
  $291 = ((($p)) + 104|0); //@line 303 "modules/graphics/ParticleSystem.cpp"
  HEAP32[$291>>2] = 0; //@line 303 "modules/graphics/ParticleSystem.cpp"
  return; //@line 304 "modules/graphics/ParticleSystem.cpp"
 }
 $284 = +HEAPF32[$117>>2]; //@line 299 "modules/graphics/ParticleSystem.cpp"
 $285 = +HEAPF32[$116>>2]; //@line 299 "modules/graphics/ParticleSystem.cpp"
 $286 = (+Math_atan2((+$284),(+$285))); //@line 299 "modules/graphics/ParticleSystem.cpp"
 $287 = $278 + $286; //@line 299 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$280>>2] = $287; //@line 299 "modules/graphics/ParticleSystem.cpp"
 $288 = ((($p)) + 88|0); //@line 301 "modules/graphics/ParticleSystem.cpp"
 $289 = ((($this)) + 200|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $290 = HEAP32[$289>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 ;HEAP32[$288>>2]=HEAP32[$290>>2]|0;HEAP32[$288+4>>2]=HEAP32[$290+4>>2]|0;HEAP32[$288+8>>2]=HEAP32[$290+8>>2]|0;HEAP32[$288+12>>2]=HEAP32[$290+12>>2]|0; //@line 301 "modules/graphics/ParticleSystem.cpp"
 $291 = ((($p)) + 104|0); //@line 303 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$291>>2] = 0; //@line 303 "modules/graphics/ParticleSystem.cpp"
 return; //@line 304 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem10setTextureEPNS0_7TextureE($this,$tex) {
 $this = $this|0;
 $tex = $tex|0;
 var $$sink$i = 0.0, $$sink4$i = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $v$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v$i = sp;
 $0 = ($tex|0)==(0|0); //@line 132 "./common/Object.h"
 if (!($0)) {
  $1 = HEAP32[$tex>>2]|0; //@line 132 "./common/Object.h"
  $2 = ((($1)) + 8|0); //@line 132 "./common/Object.h"
  $3 = HEAP32[$2>>2]|0; //@line 132 "./common/Object.h"
  FUNCTION_TABLE_vi[$3 & 8191]($tex); //@line 132 "./common/Object.h"
 }
 $4 = ((($this)) + 24|0); //@line 133 "./common/Object.h"
 $5 = HEAP32[$4>>2]|0; //@line 133 "./common/Object.h"
 $6 = ($5|0)==(0|0); //@line 133 "./common/Object.h"
 if (!($6)) {
  $7 = HEAP32[$5>>2]|0; //@line 133 "./common/Object.h"
  $8 = ((($7)) + 12|0); //@line 133 "./common/Object.h"
  $9 = HEAP32[$8>>2]|0; //@line 133 "./common/Object.h"
  FUNCTION_TABLE_vi[$9 & 8191]($5); //@line 133 "./common/Object.h"
 }
 HEAP32[$4>>2] = $tex; //@line 134 "./common/Object.h"
 $10 = ((($this)) + 196|0); //@line 416 "modules/graphics/ParticleSystem.cpp"
 $11 = HEAP8[$10>>0]|0; //@line 416 "modules/graphics/ParticleSystem.cpp"
 $12 = ($11<<24>>24)==(0); //@line 416 "modules/graphics/ParticleSystem.cpp"
 if ($12) {
  STACKTOP = sp;return; //@line 418 "modules/graphics/ParticleSystem.cpp"
 }
 $13 = ((($this)) + 212|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = HEAP32[$13>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = ((($this)) + 216|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = HEAP32[$15>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ($14|0)==($16|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($17) {
  $18 = HEAP32[$tex>>2]|0; //@line 156 "modules/graphics/ParticleSystem.cpp"
  $19 = ((($18)) + 24|0); //@line 156 "modules/graphics/ParticleSystem.cpp"
  $20 = HEAP32[$19>>2]|0; //@line 156 "modules/graphics/ParticleSystem.cpp"
  $21 = (FUNCTION_TABLE_ii[$20 & 8191]($tex)|0); //@line 156 "modules/graphics/ParticleSystem.cpp"
  $22 = (+($21|0)); //@line 156 "modules/graphics/ParticleSystem.cpp"
  $23 = $22 * 0.5; //@line 156 "modules/graphics/ParticleSystem.cpp"
  $24 = HEAP32[$4>>2]|0; //@line 117 "./common/Object.h"
  $25 = HEAP32[$24>>2]|0; //@line 156 "modules/graphics/ParticleSystem.cpp"
  $26 = ((($25)) + 28|0); //@line 156 "modules/graphics/ParticleSystem.cpp"
  $27 = HEAP32[$26>>2]|0; //@line 156 "modules/graphics/ParticleSystem.cpp"
  $28 = (FUNCTION_TABLE_ii[$27 & 8191]($24)|0); //@line 156 "modules/graphics/ParticleSystem.cpp"
  $29 = (+($28|0)); //@line 156 "modules/graphics/ParticleSystem.cpp"
  $30 = $29 * 0.5; //@line 156 "modules/graphics/ParticleSystem.cpp"
  $$sink$i = $30;$$sink4$i = $23;
 } else {
  $31 = HEAP32[$14>>2]|0; //@line 117 "./common/Object.h"
  __ZNK4love8graphics4Quad11getViewportEv($v$i,$31); //@line 159 "modules/graphics/ParticleSystem.cpp"
  $32 = +HEAPF64[$v$i>>3]; //@line 160 "modules/graphics/ParticleSystem.cpp"
  $33 = $32 * 0.5; //@line 160 "modules/graphics/ParticleSystem.cpp"
  $34 = $33; //@line 160 "modules/graphics/ParticleSystem.cpp"
  $35 = ((($v$i)) + 8|0); //@line 160 "modules/graphics/ParticleSystem.cpp"
  $36 = +HEAPF64[$35>>3]; //@line 160 "modules/graphics/ParticleSystem.cpp"
  $37 = $36 * 0.5; //@line 160 "modules/graphics/ParticleSystem.cpp"
  $38 = $37; //@line 160 "modules/graphics/ParticleSystem.cpp"
  $$sink$i = $38;$$sink4$i = $34;
 }
 $39 = ((($this)) + 188|0); //@line 160 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$39>>2] = $$sink4$i; //@line 160 "modules/graphics/ParticleSystem.cpp"
 $40 = ((($this)) + 192|0); //@line 160 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$40>>2] = $$sink$i; //@line 160 "modules/graphics/ParticleSystem.cpp"
 STACKTOP = sp;return; //@line 418 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem10getTextureEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 139 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 139 "./common/Object.h"
 return ($1|0); //@line 422 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem13setInsertModeENS1_10InsertModeE($this,$mode) {
 $this = $this|0;
 $mode = $mode|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 427 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$0>>2] = $mode; //@line 427 "modules/graphics/ParticleSystem.cpp"
 return; //@line 428 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem13getInsertModeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 432 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 432 "modules/graphics/ParticleSystem.cpp"
 return ($1|0); //@line 432 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem15setEmissionRateEf($this,$rate) {
 $this = $this|0;
 $rate = +$rate;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = $rate < 0.0; //@line 437 "modules/graphics/ParticleSystem.cpp"
 if (!($0)) {
  $6 = ((($this)) + 44|0); //@line 439 "modules/graphics/ParticleSystem.cpp"
  HEAPF32[$6>>2] = $rate; //@line 439 "modules/graphics/ParticleSystem.cpp"
  STACKTOP = sp;return; //@line 440 "modules/graphics/ParticleSystem.cpp"
 }
 $1 = (___cxa_allocate_exception(16)|0); //@line 438 "modules/graphics/ParticleSystem.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($1|0),(347423|0),($vararg_buffer|0)); //@line 438 "modules/graphics/ParticleSystem.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  ___cxa_free_exception(($1|0)); //@line 438 "modules/graphics/ParticleSystem.cpp"
  ___resumeException($4|0);
  // unreachable;
 } else {
  ___cxa_throw(($1|0),(96|0),(344|0)); //@line 438 "modules/graphics/ParticleSystem.cpp"
  // unreachable; //@line 438 "modules/graphics/ParticleSystem.cpp"
 }
}
function __ZNK4love8graphics14ParticleSystem15getEmissionRateEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 44|0); //@line 444 "modules/graphics/ParticleSystem.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 444 "modules/graphics/ParticleSystem.cpp"
 return (+$1); //@line 444 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem18setEmitterLifetimeEf($this,$life) {
 $this = $this|0;
 $life = +$life;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 80|0); //@line 449 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$0>>2] = $life; //@line 449 "modules/graphics/ParticleSystem.cpp"
 $1 = ((($this)) + 84|0); //@line 449 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$1>>2] = $life; //@line 449 "modules/graphics/ParticleSystem.cpp"
 return; //@line 450 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem18getEmitterLifetimeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 80|0); //@line 454 "modules/graphics/ParticleSystem.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 454 "modules/graphics/ParticleSystem.cpp"
 return (+$1); //@line 454 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem19setParticleLifetimeEff($this,$min,$max) {
 $this = $this|0;
 $min = +$min;
 $max = +$max;
 var $0 = 0, $1 = 0, $2 = 0, $min$max = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 88|0); //@line 459 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$0>>2] = $min; //@line 459 "modules/graphics/ParticleSystem.cpp"
 $1 = $max == 0.0; //@line 460 "modules/graphics/ParticleSystem.cpp"
 $2 = ((($this)) + 92|0); //@line 461 "modules/graphics/ParticleSystem.cpp"
 $min$max = $1 ? $min : $max; //@line 461 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$2>>2] = $min$max; //@line 461 "modules/graphics/ParticleSystem.cpp"
 return; //@line 464 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem19getParticleLifetimeERfS2_($this,$min,$max) {
 $this = $this|0;
 $min = $min|0;
 $max = $max|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 88|0); //@line 468 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 468 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$min>>2] = $1; //@line 468 "modules/graphics/ParticleSystem.cpp"
 $2 = ((($this)) + 92|0); //@line 469 "modules/graphics/ParticleSystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 469 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$max>>2] = $3; //@line 469 "modules/graphics/ParticleSystem.cpp"
 return; //@line 470 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem11setPositionEff($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 474 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$0>>2] = $x; //@line 474 "modules/graphics/ParticleSystem.cpp"
 $1 = ((($this)) + 56|0); //@line 474 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$1>>2] = $y; //@line 474 "modules/graphics/ParticleSystem.cpp"
 $2 = ((($this)) + 60|0); //@line 475 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$2>>2] = $x; //@line 475 "modules/graphics/ParticleSystem.cpp"
 $3 = ((($this)) + 64|0); //@line 475 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$3>>2] = $y; //@line 475 "modules/graphics/ParticleSystem.cpp"
 return; //@line 476 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem11getPositionEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 480 "modules/graphics/ParticleSystem.cpp"
 return ($0|0); //@line 480 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem6moveToEff($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 485 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$0>>2] = $x; //@line 485 "modules/graphics/ParticleSystem.cpp"
 $1 = ((($this)) + 56|0); //@line 485 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$1>>2] = $y; //@line 485 "modules/graphics/ParticleSystem.cpp"
 return; //@line 486 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem13setAreaSpreadENS1_22AreaSpreadDistributionEff($this,$distribution,$x,$y) {
 $this = $this|0;
 $distribution = $distribution|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 72|0); //@line 490 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$0>>2] = $x; //@line 490 "modules/graphics/ParticleSystem.cpp"
 $1 = ((($this)) + 76|0); //@line 490 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$1>>2] = $y; //@line 490 "modules/graphics/ParticleSystem.cpp"
 $2 = ((($this)) + 68|0); //@line 491 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$2>>2] = $distribution; //@line 491 "modules/graphics/ParticleSystem.cpp"
 return; //@line 492 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem25getAreaSpreadDistributionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 68|0); //@line 496 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 496 "modules/graphics/ParticleSystem.cpp"
 return ($1|0); //@line 496 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem23getAreaSpreadParametersEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 72|0); //@line 501 "modules/graphics/ParticleSystem.cpp"
 return ($0|0); //@line 501 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem12setDirectionEf($this,$direction) {
 $this = $this|0;
 $direction = +$direction;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 96|0); //@line 506 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$0>>2] = $direction; //@line 506 "modules/graphics/ParticleSystem.cpp"
 return; //@line 507 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem12getDirectionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 96|0); //@line 511 "modules/graphics/ParticleSystem.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 511 "modules/graphics/ParticleSystem.cpp"
 return (+$1); //@line 511 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem9setSpreadEf($this,$spread) {
 $this = $this|0;
 $spread = +$spread;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 100|0); //@line 516 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$0>>2] = $spread; //@line 516 "modules/graphics/ParticleSystem.cpp"
 return; //@line 517 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem9getSpreadEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 100|0); //@line 521 "modules/graphics/ParticleSystem.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 521 "modules/graphics/ParticleSystem.cpp"
 return (+$1); //@line 521 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem8setSpeedEff($this,$min,$max) {
 $this = $this|0;
 $min = +$min;
 $max = +$max;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 104|0); //@line 531 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$0>>2] = $min; //@line 531 "modules/graphics/ParticleSystem.cpp"
 $1 = ((($this)) + 108|0); //@line 532 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$1>>2] = $max; //@line 532 "modules/graphics/ParticleSystem.cpp"
 return; //@line 533 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem8getSpeedERfS2_($this,$min,$max) {
 $this = $this|0;
 $min = $min|0;
 $max = $max|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 104|0); //@line 537 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 537 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$min>>2] = $1; //@line 537 "modules/graphics/ParticleSystem.cpp"
 $2 = ((($this)) + 108|0); //@line 538 "modules/graphics/ParticleSystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 538 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$max>>2] = $3; //@line 538 "modules/graphics/ParticleSystem.cpp"
 return; //@line 539 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem21setLinearAccelerationEffff($this,$xmin,$ymin,$xmax,$ymax) {
 $this = $this|0;
 $xmin = +$xmin;
 $ymin = +$ymin;
 $xmax = +$xmax;
 $ymax = +$ymax;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 549 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$0>>2] = $xmin; //@line 549 "modules/graphics/ParticleSystem.cpp"
 $1 = ((($this)) + 116|0); //@line 549 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$1>>2] = $ymin; //@line 549 "modules/graphics/ParticleSystem.cpp"
 $2 = ((($this)) + 120|0); //@line 550 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$2>>2] = $xmax; //@line 550 "modules/graphics/ParticleSystem.cpp"
 $3 = ((($this)) + 124|0); //@line 550 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$3>>2] = $ymax; //@line 550 "modules/graphics/ParticleSystem.cpp"
 return; //@line 551 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem21getLinearAccelerationERNS_6VectorES3_($this,$min,$max) {
 $this = $this|0;
 $min = $min|0;
 $max = $max|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 555 "modules/graphics/ParticleSystem.cpp"
 $1 = $0; //@line 555 "modules/graphics/ParticleSystem.cpp"
 $2 = $1; //@line 555 "modules/graphics/ParticleSystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 555 "modules/graphics/ParticleSystem.cpp"
 $4 = (($1) + 4)|0; //@line 555 "modules/graphics/ParticleSystem.cpp"
 $5 = $4; //@line 555 "modules/graphics/ParticleSystem.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 555 "modules/graphics/ParticleSystem.cpp"
 $7 = $min; //@line 555 "modules/graphics/ParticleSystem.cpp"
 $8 = $7; //@line 555 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$8>>2] = $3; //@line 555 "modules/graphics/ParticleSystem.cpp"
 $9 = (($7) + 4)|0; //@line 555 "modules/graphics/ParticleSystem.cpp"
 $10 = $9; //@line 555 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$10>>2] = $6; //@line 555 "modules/graphics/ParticleSystem.cpp"
 $11 = ((($this)) + 120|0); //@line 556 "modules/graphics/ParticleSystem.cpp"
 $12 = $11; //@line 556 "modules/graphics/ParticleSystem.cpp"
 $13 = $12; //@line 556 "modules/graphics/ParticleSystem.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 556 "modules/graphics/ParticleSystem.cpp"
 $15 = (($12) + 4)|0; //@line 556 "modules/graphics/ParticleSystem.cpp"
 $16 = $15; //@line 556 "modules/graphics/ParticleSystem.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 556 "modules/graphics/ParticleSystem.cpp"
 $18 = $max; //@line 556 "modules/graphics/ParticleSystem.cpp"
 $19 = $18; //@line 556 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$19>>2] = $14; //@line 556 "modules/graphics/ParticleSystem.cpp"
 $20 = (($18) + 4)|0; //@line 556 "modules/graphics/ParticleSystem.cpp"
 $21 = $20; //@line 556 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$21>>2] = $17; //@line 556 "modules/graphics/ParticleSystem.cpp"
 return; //@line 557 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem21setRadialAccelerationEff($this,$min,$max) {
 $this = $this|0;
 $min = +$min;
 $max = +$max;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 128|0); //@line 566 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$0>>2] = $min; //@line 566 "modules/graphics/ParticleSystem.cpp"
 $1 = ((($this)) + 132|0); //@line 567 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$1>>2] = $max; //@line 567 "modules/graphics/ParticleSystem.cpp"
 return; //@line 568 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem21getRadialAccelerationERfS2_($this,$min,$max) {
 $this = $this|0;
 $min = $min|0;
 $max = $max|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 128|0); //@line 572 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 572 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$min>>2] = $1; //@line 572 "modules/graphics/ParticleSystem.cpp"
 $2 = ((($this)) + 132|0); //@line 573 "modules/graphics/ParticleSystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 573 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$max>>2] = $3; //@line 573 "modules/graphics/ParticleSystem.cpp"
 return; //@line 574 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem25setTangentialAccelerationEff($this,$min,$max) {
 $this = $this|0;
 $min = +$min;
 $max = +$max;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 136|0); //@line 583 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$0>>2] = $min; //@line 583 "modules/graphics/ParticleSystem.cpp"
 $1 = ((($this)) + 140|0); //@line 584 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$1>>2] = $max; //@line 584 "modules/graphics/ParticleSystem.cpp"
 return; //@line 585 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem25getTangentialAccelerationERfS2_($this,$min,$max) {
 $this = $this|0;
 $min = $min|0;
 $max = $max|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 136|0); //@line 589 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 589 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$min>>2] = $1; //@line 589 "modules/graphics/ParticleSystem.cpp"
 $2 = ((($this)) + 140|0); //@line 590 "modules/graphics/ParticleSystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 590 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$max>>2] = $3; //@line 590 "modules/graphics/ParticleSystem.cpp"
 return; //@line 591 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem16setLinearDampingEff($this,$min,$max) {
 $this = $this|0;
 $min = +$min;
 $max = +$max;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 144|0); //@line 595 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$0>>2] = $min; //@line 595 "modules/graphics/ParticleSystem.cpp"
 $1 = ((($this)) + 148|0); //@line 596 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$1>>2] = $max; //@line 596 "modules/graphics/ParticleSystem.cpp"
 return; //@line 597 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem16getLinearDampingERfS2_($this,$min,$max) {
 $this = $this|0;
 $min = $min|0;
 $max = $max|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 144|0); //@line 601 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 601 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$min>>2] = $1; //@line 601 "modules/graphics/ParticleSystem.cpp"
 $2 = ((($this)) + 148|0); //@line 602 "modules/graphics/ParticleSystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 602 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$max>>2] = $3; //@line 602 "modules/graphics/ParticleSystem.cpp"
 return; //@line 603 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem7setSizeEf($this,$size) {
 $this = $this|0;
 $size = +$size;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__new_last2$i$i$i = 0, $scevgep$i$i$i = 0;
 var $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 152|0); //@line 607 "modules/graphics/ParticleSystem.cpp"
 $1 = ((($this)) + 156|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($2) - ($3))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = $4 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5|0)==(0); //@line 1998 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = $3; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = $2; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNSt3__16vectorIfNS_9allocatorIfEEE8__appendEj($0,1); //@line 1999 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$0>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = $$pre;
  HEAPF32[$16>>2] = $size; //@line 608 "modules/graphics/ParticleSystem.cpp"
  return; //@line 609 "modules/graphics/ParticleSystem.cpp"
 }
 $9 = ($5>>>0)>(1); //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($9)) {
  $16 = $7;
  HEAPF32[$16>>2] = $size; //@line 608 "modules/graphics/ParticleSystem.cpp"
  return; //@line 609 "modules/graphics/ParticleSystem.cpp"
 }
 $10 = ((($7)) + 4|0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($8|0)==($10|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $16 = $7;
  HEAPF32[$16>>2] = $size; //@line 608 "modules/graphics/ParticleSystem.cpp"
  return; //@line 609 "modules/graphics/ParticleSystem.cpp"
 }
 $__new_last2$i$i$i = $10; //@line 818 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $scevgep$i$i$i = ((($8)) + -4|0);
 $12 = $scevgep$i$i$i;
 $13 = (($12) - ($__new_last2$i$i$i))|0;
 $14 = $13 >>> 2;
 $15 = $14 ^ -1;
 $scevgep4$i$i$i = (($8) + ($15<<2)|0);
 HEAP32[$1>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = $7;
 HEAPF32[$16>>2] = $size; //@line 608 "modules/graphics/ParticleSystem.cpp"
 return; //@line 609 "modules/graphics/ParticleSystem.cpp"
}
function __ZNSt3__16vectorIfNS_9allocatorIfEEE8__appendEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$0$i4 = 0, $$0$i5 = 0, $$in = 0, $$pre = 0, $$pre6 = 0, $$promoted1$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i = 0, $scevgep$i3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($this)) + 4|0); //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($1) - ($3))|0; //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = $4 >> 2; //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)<($__n>>>0); //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($6)) {
  $7 = $3; //@line 988 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$promoted1$i = $3; //@line 1045 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $8 = $__n << 2; //@line 985 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  _memset(($$promoted1$i|0),0,($8|0))|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $scevgep$i = (($7) + ($__n<<2)|0); //@line 985 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$2>>2] = $scevgep$i; //@line 988 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  return; //@line 1053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $9 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($3) - ($9))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $10 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = (($11) + ($__n))|0; //@line 1049 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ($12>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($13) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$0>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre6 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre6;$15 = $$pre;
 } else {
  $$in = $9;$15 = $1;
 }
 $14 = (($15) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = $14 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ($16>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($17) {
  $21 = $14 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = ($21>>>0)<($12>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $23 = $22 ? $12 : $21; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $24 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $25 = (($24) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $26 = $25 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $27 = ($23|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($27) {
   $$0$i5 = 0;$31 = 0;$32 = $26;$38 = $24;
  } else {
   $$0$i4 = $23;$46 = $24;$47 = $26;
   label = 8;
  }
 } else {
  $18 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i4 = 1073741823;$46 = $18;$47 = $20;
  label = 8;
 }
 if ((label|0) == 8) {
  $28 = $$0$i4 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $29 = (__Znwj($28)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i5 = $$0$i4;$31 = $29;$32 = $47;$38 = $46;
 }
 $30 = (($31) + ($32<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $33 = (($31) + ($$0$i5<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $34 = $33; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $35 = $__n << 2; //@line 215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 _memset(($30|0),0,($35|0))|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $scevgep$i3 = (($30) + ($__n<<2)|0); //@line 215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $36 = $scevgep$i3; //@line 218 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $37 = (($38) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $39 = $37 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $40 = (0 - ($39))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $41 = (($30) + ($40<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $42 = $41; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $43 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($41|0),($43|0),($37|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $42; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$2>>2] = $36; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $34; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $44 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($44) {
  return; //@line 1053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $45 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($45); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics14ParticleSystem8setSizesERKNSt3__16vectorIfNS2_9allocatorIfEEEE($this,$newSizes) {
 $this = $this|0;
 $newSizes = $newSizes|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 152|0); //@line 613 "modules/graphics/ParticleSystem.cpp"
 $1 = ($0|0)==($newSizes|0); //@line 1359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($1) {
  return; //@line 614 "modules/graphics/ParticleSystem.cpp"
 }
 $2 = HEAP32[$newSizes>>2]|0; //@line 1362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($newSizes)) + 4|0); //@line 1362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$3>>2]|0; //@line 1362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 __ZNSt3__16vectorIfNS_9allocatorIfEEE6assignIPfEENS_9enable_ifIXaasr21__is_forward_iteratorIT_EE5valuesr16is_constructibleIfNS_15iterator_traitsIS7_E9referenceEEE5valueEvE4typeES7_S7_($0,$2,$4); //@line 1362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 return; //@line 614 "modules/graphics/ParticleSystem.cpp"
}
function __ZNSt3__16vectorIfNS_9allocatorIfEEE6assignIPfEENS_9enable_ifIXaasr21__is_forward_iteratorIT_EE5valuesr16is_constructibleIfNS_15iterator_traitsIS7_E9referenceEEE5valueEvE4typeES7_S7_($this,$__first,$__last) {
 $this = $this|0;
 $__first = $__first|0;
 $__last = $__last|0;
 var $$0$i6 = 0, $$01$i = 0, $$01$i3 = 0, $$__last = 0, $$pre = 0, $$pre$i1 = 0, $$pre23 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $__new_last2$i$i = 0;
 var $scevgep = 0, $scevgep$i$i = 0, $scevgep$i$i$i$i = 0, $scevgep11 = 0, $scevgep14 = 0, $scevgep18 = 0, $scevgep4$i$i = 0, $scevgep4$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $__first;
 $1 = $__last; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $2 = (($1) - ($0))|0; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $3 = $2 >> 2; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $4 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = HEAP32[$4>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = (($5) - ($6))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = $7 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ($3>>>0)>($8>>>0); //@line 1398 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $6; //@line 1408 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($9)) {
  $11 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $12 = HEAP32[$11>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($6))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = ($3>>>0)>($14>>>0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = (($__first) + ($14<<2)|0); //@line 467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $$__last = $15 ? $16 : $__last; //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = $$__last;
  $18 = (($17) - ($0))|0; //@line 1763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $19 = $18 >> 2; //@line 1763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $20 = $6; //@line 1764 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  _memmove(($20|0),($__first|0),($18|0))|0; //@line 1764 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $21 = (($10) + ($19<<2)|0); //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($15)) {
   $32 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $33 = ($32|0)==($21|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($33) {
    return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $__new_last2$i$i = $21; //@line 818 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $scevgep$i$i = ((($32)) + -4|0);
   $34 = $scevgep$i$i;
   $35 = (($34) - ($__new_last2$i$i))|0;
   $36 = $35 >>> 2;
   $37 = $36 ^ -1;
   $scevgep4$i$i = (($32) + ($37<<2)|0);
   HEAP32[$11>>2] = $scevgep4$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  $22 = ($$__last|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($22) {
   return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  $$pre$i1 = HEAP32[$11>>2]|0; //@line 1030 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $scevgep14 = ((($__last)) + -4|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = $scevgep14; //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $24 = (($23) - ($17))|0; //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $25 = $24 >>> 2; //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $26 = (($25) + 1)|0; //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $scevgep18 = (($$pre$i1) + ($26<<2)|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$01$i3 = $16;$28 = $$pre$i1;
  while(1) {
   $27 = HEAP32[$$01$i3>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$28>>2] = $27; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $29 = ((($28)) + 4|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $30 = ((($$01$i3)) + 4|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $31 = ($30|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($31) {
    break;
   } else {
    $$01$i3 = $30;$28 = $29;
   }
  }
  HEAP32[$11>>2] = $scevgep18; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $38 = ($6|0)==(0); //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $39 = $6; //@line 948 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($38) {
  $71 = $5;
 } else {
  $40 = ((($this)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $41 = HEAP32[$40>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $42 = ($41|0)==($10|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($42)) {
   $scevgep$i$i$i$i = ((($41)) + -4|0);
   $43 = $scevgep$i$i$i$i;
   $44 = (($43) - ($6))|0;
   $45 = $44 >>> 2;
   $46 = $45 ^ -1;
   $scevgep4$i$i$i$i = (($41) + ($46<<2)|0);
   HEAP32[$40>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($39); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  HEAP32[$4>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$40>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$this>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $71 = 0;
 }
 $47 = ($3>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($47) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$4>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre23 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $49 = $$pre;$50 = $$pre23;
 } else {
  $49 = $71;$50 = 0;
 }
 $48 = (($49) - ($50))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $51 = $48 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $52 = ($51>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($52) {
  $53 = $48 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $54 = ($53>>>0)<($3>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $55 = $54 ? $3 : $53; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $56 = ($55>>>0)>(1073741823); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($56) {
   __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$0$i6 = $55;
  } else {
   $$0$i6 = $55;
  }
 } else {
  $$0$i6 = 1073741823;
 }
 $57 = $$0$i6 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $58 = (__Znwj($57)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $59 = ((($this)) + 4|0); //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$59>>2] = $58; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$this>>2] = $58; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $60 = (($58) + ($$0$i6<<2)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$4>>2] = $60; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $61 = ($__first|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($61) {
  return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $scevgep = ((($__last)) + -4|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $62 = $scevgep; //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $63 = (($62) - ($0))|0; //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $64 = $63 >>> 2; //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $65 = (($64) + 1)|0; //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $scevgep11 = (($58) + ($65<<2)|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $$01$i = $__first;$67 = $58;
 while(1) {
  $66 = HEAP32[$$01$i>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$67>>2] = $66; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $68 = ((($67)) + 4|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $69 = ((($$01$i)) + 4|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $70 = ($69|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($70) {
   break;
  } else {
   $$01$i = $69;$67 = $68;
  }
 }
 HEAP32[$59>>2] = $scevgep11; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNK4love8graphics14ParticleSystem8getSizesEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 152|0); //@line 618 "modules/graphics/ParticleSystem.cpp"
 return ($0|0); //@line 618 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem16setSizeVariationEf($this,$variation) {
 $this = $this|0;
 $variation = +$variation;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 164|0); //@line 623 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$0>>2] = $variation; //@line 623 "modules/graphics/ParticleSystem.cpp"
 return; //@line 624 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem16getSizeVariationEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 164|0); //@line 628 "modules/graphics/ParticleSystem.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 628 "modules/graphics/ParticleSystem.cpp"
 return (+$1); //@line 628 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem11setRotationEff($this,$min,$max) {
 $this = $this|0;
 $min = +$min;
 $max = +$max;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 168|0); //@line 638 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$0>>2] = $min; //@line 638 "modules/graphics/ParticleSystem.cpp"
 $1 = ((($this)) + 172|0); //@line 639 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$1>>2] = $max; //@line 639 "modules/graphics/ParticleSystem.cpp"
 return; //@line 640 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem11getRotationERfS2_($this,$min,$max) {
 $this = $this|0;
 $min = $min|0;
 $max = $max|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 168|0); //@line 644 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 644 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$min>>2] = $1; //@line 644 "modules/graphics/ParticleSystem.cpp"
 $2 = ((($this)) + 172|0); //@line 645 "modules/graphics/ParticleSystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 645 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$max>>2] = $3; //@line 645 "modules/graphics/ParticleSystem.cpp"
 return; //@line 646 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem7setSpinEff($this,$start,$end) {
 $this = $this|0;
 $start = +$start;
 $end = +$end;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 176|0); //@line 656 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$0>>2] = $start; //@line 656 "modules/graphics/ParticleSystem.cpp"
 $1 = ((($this)) + 180|0); //@line 657 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$1>>2] = $end; //@line 657 "modules/graphics/ParticleSystem.cpp"
 return; //@line 658 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem7getSpinERfS2_($this,$start,$end) {
 $this = $this|0;
 $start = $start|0;
 $end = $end|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 176|0); //@line 662 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 662 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$start>>2] = $1; //@line 662 "modules/graphics/ParticleSystem.cpp"
 $2 = ((($this)) + 180|0); //@line 663 "modules/graphics/ParticleSystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 663 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$end>>2] = $3; //@line 663 "modules/graphics/ParticleSystem.cpp"
 return; //@line 664 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem16setSpinVariationEf($this,$variation) {
 $this = $this|0;
 $variation = +$variation;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 184|0); //@line 668 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$0>>2] = $variation; //@line 668 "modules/graphics/ParticleSystem.cpp"
 return; //@line 669 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem16getSpinVariationEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 184|0); //@line 673 "modules/graphics/ParticleSystem.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 673 "modules/graphics/ParticleSystem.cpp"
 return (+$1); //@line 673 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem9setOffsetEff($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 188|0); //@line 678 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$0>>2] = $x; //@line 678 "modules/graphics/ParticleSystem.cpp"
 $1 = ((($this)) + 192|0); //@line 678 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$1>>2] = $y; //@line 678 "modules/graphics/ParticleSystem.cpp"
 $2 = ((($this)) + 196|0); //@line 679 "modules/graphics/ParticleSystem.cpp"
 HEAP8[$2>>0] = 0; //@line 679 "modules/graphics/ParticleSystem.cpp"
 return; //@line 680 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem9getOffsetEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 188|0); //@line 684 "modules/graphics/ParticleSystem.cpp"
 $1 = $0; //@line 684 "modules/graphics/ParticleSystem.cpp"
 $2 = $1; //@line 684 "modules/graphics/ParticleSystem.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 684 "modules/graphics/ParticleSystem.cpp"
 $4 = (($1) + 4)|0; //@line 684 "modules/graphics/ParticleSystem.cpp"
 $5 = $4; //@line 684 "modules/graphics/ParticleSystem.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 684 "modules/graphics/ParticleSystem.cpp"
 $7 = $agg$result; //@line 684 "modules/graphics/ParticleSystem.cpp"
 $8 = $7; //@line 684 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$8>>2] = $3; //@line 684 "modules/graphics/ParticleSystem.cpp"
 $9 = (($7) + 4)|0; //@line 684 "modules/graphics/ParticleSystem.cpp"
 $10 = $9; //@line 684 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$10>>2] = $6; //@line 684 "modules/graphics/ParticleSystem.cpp"
 return; //@line 684 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem8setColorERKNSt3__16vectorINS0_6ColorTIfEENS2_9allocatorIS5_EEEE($this,$newColors) {
 $this = $this|0;
 $newColors = $newColors|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0.0, $__begin$sroa$0$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 200|0); //@line 689 "modules/graphics/ParticleSystem.cpp"
 $1 = ($0|0)==($newColors|0); //@line 1359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($1)) {
  $2 = HEAP32[$newColors>>2]|0; //@line 1362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $3 = ((($newColors)) + 4|0); //@line 1362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = HEAP32[$3>>2]|0; //@line 1362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  __ZNSt3__16vectorIN4love8graphics6ColorTIfEENS_9allocatorIS4_EEE6assignIPS4_EENS_9enable_ifIXaasr21__is_forward_iteratorIT_EE5valuesr16is_constructibleIS4_NS_15iterator_traitsISB_E9referenceEEE5valueEvE4typeESB_SB_($0,$2,$4); //@line 1362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $5 = HEAP32[$0>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ((($this)) + 204|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = HEAP32[$6>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ($5|0)==($7|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($8) {
  return; //@line 699 "modules/graphics/ParticleSystem.cpp"
 } else {
  $__begin$sroa$0$01 = $5;
 }
 while(1) {
  $9 = +HEAPF32[$__begin$sroa$0$01>>2]; //@line 694 "modules/graphics/ParticleSystem.cpp"
  $10 = $9 / 255.0; //@line 694 "modules/graphics/ParticleSystem.cpp"
  HEAPF32[$__begin$sroa$0$01>>2] = $10; //@line 694 "modules/graphics/ParticleSystem.cpp"
  $11 = ((($__begin$sroa$0$01)) + 4|0); //@line 695 "modules/graphics/ParticleSystem.cpp"
  $12 = +HEAPF32[$11>>2]; //@line 695 "modules/graphics/ParticleSystem.cpp"
  $13 = $12 / 255.0; //@line 695 "modules/graphics/ParticleSystem.cpp"
  HEAPF32[$11>>2] = $13; //@line 695 "modules/graphics/ParticleSystem.cpp"
  $14 = ((($__begin$sroa$0$01)) + 8|0); //@line 696 "modules/graphics/ParticleSystem.cpp"
  $15 = +HEAPF32[$14>>2]; //@line 696 "modules/graphics/ParticleSystem.cpp"
  $16 = $15 / 255.0; //@line 696 "modules/graphics/ParticleSystem.cpp"
  HEAPF32[$14>>2] = $16; //@line 696 "modules/graphics/ParticleSystem.cpp"
  $17 = ((($__begin$sroa$0$01)) + 12|0); //@line 697 "modules/graphics/ParticleSystem.cpp"
  $18 = +HEAPF32[$17>>2]; //@line 697 "modules/graphics/ParticleSystem.cpp"
  $19 = $18 / 255.0; //@line 697 "modules/graphics/ParticleSystem.cpp"
  HEAPF32[$17>>2] = $19; //@line 697 "modules/graphics/ParticleSystem.cpp"
  $20 = ((($__begin$sroa$0$01)) + 16|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $21 = ($20|0)==($7|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($21) {
   break;
  } else {
   $__begin$sroa$0$01 = $20;
  }
 }
 return; //@line 699 "modules/graphics/ParticleSystem.cpp"
}
function __ZNSt3__16vectorIN4love8graphics6ColorTIfEENS_9allocatorIS4_EEE6assignIPS4_EENS_9enable_ifIXaasr21__is_forward_iteratorIT_EE5valuesr16is_constructibleIS4_NS_15iterator_traitsISB_E9referenceEEE5valueEvE4typeESB_SB_($this,$__first,$__last) {
 $this = $this|0;
 $__first = $__first|0;
 $__last = $__last|0;
 var $$0$i5 = 0, $$01$i = 0, $$01$i3 = 0, $$__last = 0, $$pre = 0, $$pre$i1 = 0, $$pre8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, $__new_last2$i$i = 0, $scevgep$i$i = 0, $scevgep$i$i$i$i = 0, $scevgep4$i$i = 0, $scevgep4$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $__last; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $1 = $__first; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $2 = (($0) - ($1))|0; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $3 = $2 >> 4; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $4 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = HEAP32[$4>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = (($5) - ($6))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = $7 >> 4; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ($3>>>0)>($8>>>0); //@line 1398 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $6; //@line 1408 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($9)) {
  $11 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $12 = HEAP32[$11>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($6))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = ($3>>>0)>($14>>>0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = (($__first) + ($14<<4)|0); //@line 467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $$__last = $15 ? $16 : $__last; //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = $$__last; //@line 1763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $18 = (($17) - ($1))|0; //@line 1763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $19 = $18 >> 4; //@line 1763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $20 = $6; //@line 1764 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  _memmove(($20|0),($__first|0),($18|0))|0; //@line 1764 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $21 = (($10) + ($19<<4)|0); //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($15)) {
   $28 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $29 = ($28|0)==($21|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($29) {
    return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $__new_last2$i$i = $21; //@line 818 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $scevgep$i$i = ((($28)) + -16|0);
   $30 = $scevgep$i$i;
   $31 = (($30) - ($__new_last2$i$i))|0;
   $32 = $31 >>> 4;
   $33 = $32 ^ -1;
   $scevgep4$i$i = (($28) + ($33<<4)|0);
   HEAP32[$11>>2] = $scevgep4$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  $22 = ($$__last|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($22) {
   return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  $$pre$i1 = HEAP32[$11>>2]|0; //@line 1030 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$01$i3 = $16;$23 = $$pre$i1;
  while(1) {
   ;HEAP32[$23>>2]=HEAP32[$$01$i3>>2]|0;HEAP32[$23+4>>2]=HEAP32[$$01$i3+4>>2]|0;HEAP32[$23+8>>2]=HEAP32[$$01$i3+8>>2]|0;HEAP32[$23+12>>2]=HEAP32[$$01$i3+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $24 = HEAP32[$11>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $25 = ((($24)) + 16|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$11>>2] = $25; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $26 = ((($$01$i3)) + 16|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $27 = ($26|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($27) {
    break;
   } else {
    $$01$i3 = $26;$23 = $25;
   }
  }
  return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $34 = ($6|0)==(0); //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $35 = $6; //@line 948 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($34) {
  $63 = $5;
 } else {
  $36 = ((($this)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $37 = HEAP32[$36>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $38 = ($37|0)==($10|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($38)) {
   $scevgep$i$i$i$i = ((($37)) + -16|0);
   $39 = $scevgep$i$i$i$i;
   $40 = (($39) - ($6))|0;
   $41 = $40 >>> 4;
   $42 = $41 ^ -1;
   $scevgep4$i$i$i$i = (($37) + ($42<<4)|0);
   HEAP32[$36>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($35); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  HEAP32[$4>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$36>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$this>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $63 = 0;
 }
 $43 = ($3>>>0)>(268435455); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($43) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$4>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre8 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $45 = $$pre;$46 = $$pre8;
 } else {
  $45 = $63;$46 = 0;
 }
 $44 = (($45) - ($46))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $47 = $44 >> 4; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $48 = ($47>>>0)<(134217727); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($48) {
  $49 = $44 >> 3; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $50 = ($49>>>0)<($3>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $51 = $50 ? $3 : $49; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $52 = ($51>>>0)>(268435455); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($52) {
   __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$0$i5 = $51;
  } else {
   $$0$i5 = $51;
  }
 } else {
  $$0$i5 = 268435455;
 }
 $53 = $$0$i5 << 4; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $54 = (__Znwj($53)|0); //@line 1030 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $55 = ((($this)) + 4|0); //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$55>>2] = $54; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$this>>2] = $54; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $56 = (($54) + ($$0$i5<<4)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$4>>2] = $56; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $57 = ($__first|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($57) {
  return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 } else {
  $$01$i = $__first;$58 = $54;
 }
 while(1) {
  ;HEAP32[$58>>2]=HEAP32[$$01$i>>2]|0;HEAP32[$58+4>>2]=HEAP32[$$01$i+4>>2]|0;HEAP32[$58+8>>2]=HEAP32[$$01$i+8>>2]|0;HEAP32[$58+12>>2]=HEAP32[$$01$i+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $59 = HEAP32[$55>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $60 = ((($59)) + 16|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$55>>2] = $60; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $61 = ((($$01$i)) + 16|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $62 = ($61|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($62) {
   break;
  } else {
   $$01$i = $61;$58 = $60;
  }
 }
 return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNK4love8graphics14ParticleSystem8getColorEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0.0, $9 = 0.0, $__begin$sroa$0$01 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 200|0); //@line 704 "modules/graphics/ParticleSystem.cpp"
 __ZNSt3__16vectorIN4love8graphics6ColorTIfEENS_9allocatorIS4_EEEC2ERKS7_($agg$result,$0); //@line 704 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP32[$agg$result>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($agg$result)) + 4|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($1|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($4) {
  return; //@line 715 "modules/graphics/ParticleSystem.cpp"
 } else {
  $__begin$sroa$0$01 = $1;
 }
 while(1) {
  $5 = +HEAPF32[$__begin$sroa$0$01>>2]; //@line 708 "modules/graphics/ParticleSystem.cpp"
  $6 = $5 * 255.0; //@line 708 "modules/graphics/ParticleSystem.cpp"
  HEAPF32[$__begin$sroa$0$01>>2] = $6; //@line 708 "modules/graphics/ParticleSystem.cpp"
  $7 = ((($__begin$sroa$0$01)) + 4|0); //@line 709 "modules/graphics/ParticleSystem.cpp"
  $8 = +HEAPF32[$7>>2]; //@line 709 "modules/graphics/ParticleSystem.cpp"
  $9 = $8 * 255.0; //@line 709 "modules/graphics/ParticleSystem.cpp"
  HEAPF32[$7>>2] = $9; //@line 709 "modules/graphics/ParticleSystem.cpp"
  $10 = ((($__begin$sroa$0$01)) + 8|0); //@line 710 "modules/graphics/ParticleSystem.cpp"
  $11 = +HEAPF32[$10>>2]; //@line 710 "modules/graphics/ParticleSystem.cpp"
  $12 = $11 * 255.0; //@line 710 "modules/graphics/ParticleSystem.cpp"
  HEAPF32[$10>>2] = $12; //@line 710 "modules/graphics/ParticleSystem.cpp"
  $13 = ((($__begin$sroa$0$01)) + 12|0); //@line 711 "modules/graphics/ParticleSystem.cpp"
  $14 = +HEAPF32[$13>>2]; //@line 711 "modules/graphics/ParticleSystem.cpp"
  $15 = $14 * 255.0; //@line 711 "modules/graphics/ParticleSystem.cpp"
  HEAPF32[$13>>2] = $15; //@line 711 "modules/graphics/ParticleSystem.cpp"
  $16 = ((($__begin$sroa$0$01)) + 16|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $17 = ($16|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($17) {
   break;
  } else {
   $__begin$sroa$0$01 = $16;
  }
 }
 return; //@line 715 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem8setQuadsERKNSt3__16vectorIPNS0_4QuadENS2_9allocatorIS5_EEEE($this,$newQuads) {
 $this = $this|0;
 $newQuads = $newQuads|0;
 var $$0 = 0, $$01 = 0, $$lcssa47 = 0, $$pre$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i4 = 0, $$pre$i12 = 0, $$pre$i7 = 0, $$sink$i = 0.0, $$sink4$i = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0.0, $88 = 0.0, $89 = 0.0;
 var $9 = 0, $90 = 0.0, $91 = 0, $92 = 0, $93 = 0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0, $98 = 0.0, $99 = 0.0, $__begin$sroa$0$021 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $quadlist = 0, $v$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v$i = sp;
 $quadlist = sp + 36|0;
 $0 = sp + 32|0;
 HEAP32[$quadlist>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($quadlist)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($quadlist)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = ((($newQuads)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$newQuads>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = (($4) - ($5))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = $6 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 __THREW__ = 0;
 invoke_vii(3630,($quadlist|0),($7|0)); //@line 720 "modules/graphics/ParticleSystem.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 L1: do {
  if ($9) {
   label = 7;
  } else {
   $10 = HEAP32[$newQuads>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $11 = HEAP32[$3>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $12 = ($10|0)==($11|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   L3: do {
    if (!($12)) {
     $13 = ((($quadlist)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $__begin$sroa$0$021 = $10;
     while(1) {
      $24 = HEAP32[$__begin$sroa$0$021>>2]|0; //@line 106 "./common/Object.h"
      HEAP32[$0>>2] = $24; //@line 93 "./common/Object.h"
      $25 = ($24|0)==(0|0); //@line 95 "./common/Object.h"
      if (!($25)) {
       $26 = HEAP32[$24>>2]|0; //@line 95 "./common/Object.h"
       $27 = ((($26)) + 8|0); //@line 95 "./common/Object.h"
       $28 = HEAP32[$27>>2]|0; //@line 95 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($28|0,($24|0)); //@line 95 "./common/Object.h"
       $29 = __THREW__; __THREW__ = 0;
       $30 = $29&1;
       if ($30) {
        label = 6;
        break;
       }
      }
      $31 = HEAP32[$1>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $32 = HEAP32[$13>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $33 = ($31>>>0)<($32>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($33) {
       HEAP32[$31>>2] = $24; //@line 99 "./common/Object.h"
       if ($25) {
        $42 = $31;
       } else {
        $34 = HEAP32[$24>>2]|0; //@line 101 "./common/Object.h"
        $35 = ((($34)) + 8|0); //@line 101 "./common/Object.h"
        $36 = HEAP32[$35>>2]|0; //@line 101 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($36|0,($24|0)); //@line 101 "./common/Object.h"
        $37 = __THREW__; __THREW__ = 0;
        $38 = $37&1;
        if ($38) {
         $$lcssa47 = $24;
         label = 14;
         break;
        }
        $$pre$i12 = HEAP32[$1>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $42 = $$pre$i12;
       }
       $41 = ((($42)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$1>>2] = $41; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      } else {
       __THREW__ = 0;
       invoke_vii(3632,($quadlist|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $43 = __THREW__; __THREW__ = 0;
       $44 = $43&1;
       if ($44) {
        label = 22;
        break;
       }
      }
      if (!($25)) {
       $45 = HEAP32[$24>>2]|0; //@line 106 "./common/Object.h"
       $46 = ((($45)) + 12|0); //@line 106 "./common/Object.h"
       $47 = HEAP32[$46>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($47|0,($24|0)); //@line 106 "./common/Object.h"
       $48 = __THREW__; __THREW__ = 0;
       $49 = $48&1;
       if ($49) {
        label = 20;
        break;
       }
      }
      $52 = ((($__begin$sroa$0$021)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $53 = ($52|0)==($11|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($53) {
       break L3;
      } else {
       $__begin$sroa$0$021 = $52;
      }
     }
     if ((label|0) == 6) {
      $20 = ___cxa_find_matching_catch_2()|0;
      $21 = tempRet0;
      $lpad$phi$index = $20;$lpad$phi$index2 = $21;
      label = 8;
      break L1;
     }
     else if ((label|0) == 14) {
      $39 = ___cxa_find_matching_catch_2()|0;
      $40 = tempRet0;
      $139 = $40;$140 = $39;$57 = $$lcssa47;
     }
     else if ((label|0) == 20) {
      $50 = ___cxa_find_matching_catch_3(0|0)|0;
      $51 = tempRet0;
      ___clang_call_terminate($50); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     else if ((label|0) == 22) {
      $54 = ___cxa_find_matching_catch_2()|0;
      $55 = tempRet0;
      if ($25) {
       $$0 = $54;$$01 = $55;
       break L1;
      } else {
       $139 = $55;$140 = $54;$57 = $24;
      }
     }
     $56 = HEAP32[$57>>2]|0; //@line 106 "./common/Object.h"
     $58 = ((($56)) + 12|0); //@line 106 "./common/Object.h"
     $59 = HEAP32[$58>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($59|0,($57|0)); //@line 106 "./common/Object.h"
     $60 = __THREW__; __THREW__ = 0;
     $61 = $60&1;
     if (!($61)) {
      $$0 = $140;$$01 = $139;
      break L1;
     }
     $62 = ___cxa_find_matching_catch_3(0|0)|0;
     $63 = tempRet0;
     ___clang_call_terminate($62); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   } while(0);
   $14 = ((($this)) + 212|0); //@line 725 "modules/graphics/ParticleSystem.cpp"
   $15 = ($14|0)==($quadlist|0); //@line 1359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($15)) {
    $16 = HEAP32[$quadlist>>2]|0; //@line 1362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $17 = HEAP32[$1>>2]|0; //@line 1362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __THREW__ = 0;
    invoke_viii(3631,($14|0),($16|0),($17|0)); //@line 1362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $18 = __THREW__; __THREW__ = 0;
    $19 = $18&1;
    if ($19) {
     label = 7;
     break;
    }
   }
   $64 = ((($this)) + 196|0); //@line 727 "modules/graphics/ParticleSystem.cpp"
   $65 = HEAP8[$64>>0]|0; //@line 727 "modules/graphics/ParticleSystem.cpp"
   $66 = ($65<<24>>24)==(0); //@line 727 "modules/graphics/ParticleSystem.cpp"
   if (!($66)) {
    $67 = ((($this)) + 212|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $68 = HEAP32[$67>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $69 = ((($this)) + 216|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $70 = HEAP32[$69>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $71 = ($68|0)==($70|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($71) {
     $72 = ((($this)) + 24|0); //@line 117 "./common/Object.h"
     $73 = HEAP32[$72>>2]|0; //@line 117 "./common/Object.h"
     $74 = HEAP32[$73>>2]|0; //@line 156 "modules/graphics/ParticleSystem.cpp"
     $75 = ((($74)) + 24|0); //@line 156 "modules/graphics/ParticleSystem.cpp"
     $76 = HEAP32[$75>>2]|0; //@line 156 "modules/graphics/ParticleSystem.cpp"
     __THREW__ = 0;
     $77 = (invoke_ii($76|0,($73|0))|0); //@line 156 "modules/graphics/ParticleSystem.cpp"
     $78 = __THREW__; __THREW__ = 0;
     $79 = $78&1;
     if ($79) {
      label = 7;
      break;
     }
     $80 = HEAP32[$72>>2]|0; //@line 117 "./common/Object.h"
     $81 = HEAP32[$80>>2]|0; //@line 156 "modules/graphics/ParticleSystem.cpp"
     $82 = ((($81)) + 28|0); //@line 156 "modules/graphics/ParticleSystem.cpp"
     $83 = HEAP32[$82>>2]|0; //@line 156 "modules/graphics/ParticleSystem.cpp"
     __THREW__ = 0;
     $84 = (invoke_ii($83|0,($80|0))|0); //@line 156 "modules/graphics/ParticleSystem.cpp"
     $85 = __THREW__; __THREW__ = 0;
     $86 = $85&1;
     if ($86) {
      label = 7;
      break;
     }
     $87 = (+($77|0)); //@line 156 "modules/graphics/ParticleSystem.cpp"
     $88 = $87 * 0.5; //@line 156 "modules/graphics/ParticleSystem.cpp"
     $89 = (+($84|0)); //@line 156 "modules/graphics/ParticleSystem.cpp"
     $90 = $89 * 0.5; //@line 156 "modules/graphics/ParticleSystem.cpp"
     $$sink$i = $90;$$sink4$i = $88;
    } else {
     $91 = HEAP32[$68>>2]|0; //@line 117 "./common/Object.h"
     __THREW__ = 0;
     invoke_vii(3633,($v$i|0),($91|0)); //@line 159 "modules/graphics/ParticleSystem.cpp"
     $92 = __THREW__; __THREW__ = 0;
     $93 = $92&1;
     if ($93) {
      label = 7;
      break;
     }
     $94 = +HEAPF64[$v$i>>3]; //@line 160 "modules/graphics/ParticleSystem.cpp"
     $95 = $94 * 0.5; //@line 160 "modules/graphics/ParticleSystem.cpp"
     $96 = $95; //@line 160 "modules/graphics/ParticleSystem.cpp"
     $97 = ((($v$i)) + 8|0); //@line 160 "modules/graphics/ParticleSystem.cpp"
     $98 = +HEAPF64[$97>>3]; //@line 160 "modules/graphics/ParticleSystem.cpp"
     $99 = $98 * 0.5; //@line 160 "modules/graphics/ParticleSystem.cpp"
     $100 = $99; //@line 160 "modules/graphics/ParticleSystem.cpp"
     $$sink$i = $100;$$sink4$i = $96;
    }
    $101 = ((($this)) + 188|0); //@line 160 "modules/graphics/ParticleSystem.cpp"
    HEAPF32[$101>>2] = $$sink4$i; //@line 160 "modules/graphics/ParticleSystem.cpp"
    $102 = ((($this)) + 192|0); //@line 160 "modules/graphics/ParticleSystem.cpp"
    HEAPF32[$102>>2] = $$sink$i; //@line 160 "modules/graphics/ParticleSystem.cpp"
   }
   $103 = HEAP32[$quadlist>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $104 = ($103|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($104) {
    STACKTOP = sp;return; //@line 729 "modules/graphics/ParticleSystem.cpp"
   }
   $105 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $106 = ($105|0)==($103|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   do {
    if ($106) {
     $120 = $103;
    } else {
     $108 = $105;
     while(1) {
      $107 = ((($108)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$1>>2] = $107; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $109 = HEAP32[$107>>2]|0; //@line 106 "./common/Object.h"
      $110 = ($109|0)==(0|0); //@line 106 "./common/Object.h"
      if ($110) {
       $112 = $107;
      } else {
       $113 = HEAP32[$109>>2]|0; //@line 106 "./common/Object.h"
       $114 = ((($113)) + 12|0); //@line 106 "./common/Object.h"
       $115 = HEAP32[$114>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($115|0,($109|0)); //@line 106 "./common/Object.h"
       $116 = __THREW__; __THREW__ = 0;
       $117 = $116&1;
       if ($117) {
        label = 39;
        break;
       }
       $$pre$i$i$i4 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $112 = $$pre$i$i$i4;
      }
      $111 = ($112|0)==($103|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($111) {
       label = 40;
       break;
      } else {
       $108 = $112;
      }
     }
     if ((label|0) == 39) {
      $118 = ___cxa_find_matching_catch_3(0|0)|0;
      $119 = tempRet0;
      ___clang_call_terminate($118); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     else if ((label|0) == 40) {
      $$pre$i7 = HEAP32[$quadlist>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $120 = $$pre$i7;
      break;
     }
    }
   } while(0);
   __ZdlPv($120); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   STACKTOP = sp;return; //@line 729 "modules/graphics/ParticleSystem.cpp"
  }
 } while(0);
 if ((label|0) == 7) {
  $22 = ___cxa_find_matching_catch_2()|0;
  $23 = tempRet0;
  $lpad$phi$index = $22;$lpad$phi$index2 = $23;
  label = 8;
 }
 if ((label|0) == 8) {
  $$0 = $lpad$phi$index;$$01 = $lpad$phi$index2;
 }
 $121 = HEAP32[$quadlist>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $122 = ($121|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($122) {
  ___resumeException($$0|0);
  // unreachable;
 }
 $123 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $124 = ($123|0)==($121|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($124) {
   $138 = $121;
  } else {
   $126 = $123;
   while(1) {
    $125 = ((($126)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$1>>2] = $125; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $127 = HEAP32[$125>>2]|0; //@line 106 "./common/Object.h"
    $128 = ($127|0)==(0|0); //@line 106 "./common/Object.h"
    if ($128) {
     $130 = $125;
    } else {
     $131 = HEAP32[$127>>2]|0; //@line 106 "./common/Object.h"
     $132 = ((($131)) + 12|0); //@line 106 "./common/Object.h"
     $133 = HEAP32[$132>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($133|0,($127|0)); //@line 106 "./common/Object.h"
     $134 = __THREW__; __THREW__ = 0;
     $135 = $134&1;
     if ($135) {
      label = 49;
      break;
     }
     $$pre$i$i$i = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $130 = $$pre$i$i$i;
    }
    $129 = ($130|0)==($121|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($129) {
     label = 50;
     break;
    } else {
     $126 = $130;
    }
   }
   if ((label|0) == 49) {
    $136 = ___cxa_find_matching_catch_3(0|0)|0;
    $137 = tempRet0;
    ___clang_call_terminate($136); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 50) {
    $$pre$i = HEAP32[$quadlist>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $138 = $$pre$i;
    break;
   }
  }
 } while(0);
 __ZdlPv($138); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$0|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love9StrongRefINS1_8graphics4QuadEEENS_9allocatorIS5_EEE7reserveEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$01$i$i = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $8 = 0, $9 = 0, $__v$sroa$6$0 = 0, $__v$sroa$6$0$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($4>>>0)<($__n>>>0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = $2; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($5)) {
  return; //@line 1545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $7 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = HEAP32[$7>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($__n|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $12 = $8; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $16 = 0;
 } else {
  $13 = $__n << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $14 = (__Znwj($13)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $16 = $14;
 }
 $15 = (($16) + ($10<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $17 = $15; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $18 = (($16) + ($__n<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $19 = $18; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $20 = ($12|0)==($6|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 do {
  if ($20) {
   $35 = $17;$37 = $2;$39 = $6;
  } else {
   $$01$i$i = $12;$23 = $15;$__v$sroa$6$0 = $17;
   while(1) {
    $21 = ((($$01$i$i)) + -4|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $22 = ((($23)) + -4|0); //@line 99 "./common/Object.h"
    $24 = HEAP32[$21>>2]|0; //@line 139 "./common/Object.h"
    HEAP32[$22>>2] = $24; //@line 99 "./common/Object.h"
    $25 = ($24|0)==(0|0); //@line 101 "./common/Object.h"
    if (!($25)) {
     $26 = HEAP32[$24>>2]|0; //@line 101 "./common/Object.h"
     $27 = ((($26)) + 8|0); //@line 101 "./common/Object.h"
     $28 = HEAP32[$27>>2]|0; //@line 101 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($28|0,($24|0)); //@line 101 "./common/Object.h"
     $29 = __THREW__; __THREW__ = 0;
     $30 = $29&1;
     if ($30) {
      $__v$sroa$6$0$lcssa = $__v$sroa$6$0;
      break;
     }
    }
    $31 = $__v$sroa$6$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $32 = ((($31)) + -4|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $33 = $32; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $34 = ($21|0)==($6|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    if ($34) {
     $$lcssa = $33;
     label = 8;
     break;
    } else {
     $$01$i$i = $21;$23 = $32;$__v$sroa$6$0 = $33;
    }
   }
   if ((label|0) == 8) {
    $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $$pre = HEAP32[$7>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $35 = $$lcssa;$37 = $$pre$i;$39 = $$pre;
    break;
   }
   $54 = ___cxa_find_matching_catch_2()|0;
   $55 = tempRet0;
   $56 = $__v$sroa$6$0$lcssa; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $57 = ($15|0)==($56|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   L16: do {
    if (!($57)) {
     $59 = $15;
     while(1) {
      $58 = ((($59)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $60 = HEAP32[$58>>2]|0; //@line 106 "./common/Object.h"
      $61 = ($60|0)==(0|0); //@line 106 "./common/Object.h"
      if (!($61)) {
       $63 = HEAP32[$60>>2]|0; //@line 106 "./common/Object.h"
       $64 = ((($63)) + 12|0); //@line 106 "./common/Object.h"
       $65 = HEAP32[$64>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($65|0,($60|0)); //@line 106 "./common/Object.h"
       $66 = __THREW__; __THREW__ = 0;
       $67 = $66&1;
       if ($67) {
        break;
       }
      }
      $62 = ($58|0)==($56|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if ($62) {
       break L16;
      } else {
       $59 = $58;
      }
     }
     $68 = ___cxa_find_matching_catch_3(0|0)|0;
     $69 = tempRet0;
     ___clang_call_terminate($68); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   } while(0);
   $70 = ($16|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($70) {
    ___resumeException($54|0);
    // unreachable;
   }
   __ZdlPv($16); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($54|0);
   // unreachable;
  }
 } while(0);
 HEAP32[$this>>2] = $35; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $17; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $19; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $36 = $37; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $38 = ($39|0)==($36|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 L29: do {
  if (!($38)) {
   $41 = $39;
   while(1) {
    $40 = ((($41)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $42 = HEAP32[$40>>2]|0; //@line 106 "./common/Object.h"
    $43 = ($42|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($43)) {
     $45 = HEAP32[$42>>2]|0; //@line 106 "./common/Object.h"
     $46 = ((($45)) + 12|0); //@line 106 "./common/Object.h"
     $47 = HEAP32[$46>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($47|0,($42|0)); //@line 106 "./common/Object.h"
     $48 = __THREW__; __THREW__ = 0;
     $49 = $48&1;
     if ($49) {
      break;
     }
    }
    $44 = ($40|0)==($36|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($44) {
     break L29;
    } else {
     $41 = $40;
    }
   }
   $50 = ___cxa_find_matching_catch_3(0|0)|0;
   $51 = tempRet0;
   ___clang_call_terminate($50); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 } while(0);
 $52 = ($37|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($52) {
  return; //@line 1545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $53 = $37; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($53); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNSt3__16vectorIN4love9StrongRefINS1_8graphics4QuadEEENS_9allocatorIS5_EEE6assignIPS5_EENS_9enable_ifIXaasr21__is_forward_iteratorIT_EE5valuesr16is_constructibleIS5_NS_15iterator_traitsISC_E9referenceEEE5valueEvE4typeESC_SC_($this,$__first,$__last) {
 $this = $this|0;
 $__first = $__first|0;
 $__last = $__last|0;
 var $$0$i9 = 0, $$0$lcssa$i$i = 0, $$01$i = 0, $$01$i4 = 0, $$012$i$i = 0, $$03$i$i = 0, $$__last = 0, $$pre = 0, $$pre$i$i = 0, $$pre$i$i$i$i = 0, $$pre$i1 = 0, $$pre$i2 = 0, $$pre14 = 0, $$pre2$i = 0, $$pre2$i5 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $scevgep$i$i = 0, $scevgep7$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $__last; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $1 = $__first; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $2 = (($0) - ($1))|0; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $3 = $2 >> 2; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $4 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = HEAP32[$4>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = (($5) - ($6))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = $7 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ($3>>>0)>($8>>>0); //@line 1398 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $6; //@line 1408 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($9) {
  $61 = ($6|0)==(0); //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($61) {
   $105 = $5;
  } else {
   $62 = $6; //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $63 = ((($this)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $64 = HEAP32[$63>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $65 = ($64|0)==($10|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   do {
    if ($65) {
     $79 = $62;
    } else {
     $67 = $64;
     while(1) {
      $66 = ((($67)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$63>>2] = $66; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $68 = HEAP32[$66>>2]|0; //@line 106 "./common/Object.h"
      $69 = ($68|0)==(0|0); //@line 106 "./common/Object.h"
      if ($69) {
       $71 = $66;
      } else {
       $72 = HEAP32[$68>>2]|0; //@line 106 "./common/Object.h"
       $73 = ((($72)) + 12|0); //@line 106 "./common/Object.h"
       $74 = HEAP32[$73>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($74|0,($68|0)); //@line 106 "./common/Object.h"
       $75 = __THREW__; __THREW__ = 0;
       $76 = $75&1;
       if ($76) {
        label = 28;
        break;
       }
       $$pre$i$i$i$i = HEAP32[$63>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $71 = $$pre$i$i$i$i;
      }
      $70 = ($71|0)==($10|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($70) {
       label = 29;
       break;
      } else {
       $67 = $71;
      }
     }
     if ((label|0) == 28) {
      $77 = ___cxa_find_matching_catch_3(0|0)|0;
      $78 = tempRet0;
      ___clang_call_terminate($77); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     else if ((label|0) == 29) {
      $$pre$i1 = HEAP32[$this>>2]|0; //@line 948 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $79 = $$pre$i1;
      break;
     }
    }
   } while(0);
   __ZdlPv($79); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   HEAP32[$4>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$63>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$this>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $105 = 0;
  }
  $80 = ($3>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($80) {
   __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$pre = HEAP32[$4>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$pre14 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $82 = $$pre;$83 = $$pre14;
  } else {
   $82 = $105;$83 = 0;
  }
  $81 = (($82) - ($83))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $84 = $81 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $85 = ($84>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($85) {
   $86 = $81 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $87 = ($86>>>0)<($3>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $88 = $87 ? $3 : $86; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $89 = ($88>>>0)>(1073741823); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($89) {
    __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$0$i9 = $88;
   } else {
    $$0$i9 = $88;
   }
  } else {
   $$0$i9 = 1073741823;
  }
  $90 = $$0$i9 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $91 = (__Znwj($90)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $92 = ((($this)) + 4|0); //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$92>>2] = $91; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$this>>2] = $91; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $93 = (($91) + ($$0$i9<<2)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$4>>2] = $93; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $94 = ($__first|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($94) {
   return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  $$01$i = $__first;$96 = $91;
  while(1) {
   $95 = HEAP32[$$01$i>>2]|0; //@line 139 "./common/Object.h"
   HEAP32[$96>>2] = $95; //@line 99 "./common/Object.h"
   $97 = ($95|0)==(0|0); //@line 101 "./common/Object.h"
   if ($97) {
    $102 = $96;
   } else {
    $98 = HEAP32[$95>>2]|0; //@line 101 "./common/Object.h"
    $99 = ((($98)) + 8|0); //@line 101 "./common/Object.h"
    $100 = HEAP32[$99>>2]|0; //@line 101 "./common/Object.h"
    FUNCTION_TABLE_vi[$100 & 8191]($95); //@line 101 "./common/Object.h"
    $$pre2$i = HEAP32[$92>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $102 = $$pre2$i;
   }
   $101 = ((($102)) + 4|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$92>>2] = $101; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $103 = ((($$01$i)) + 4|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $104 = ($103|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($104) {
    break;
   } else {
    $$01$i = $103;$96 = $101;
   }
  }
  return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $11 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = HEAP32[$11>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = (($12) - ($6))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = ($3>>>0)>($14>>>0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = (($__first) + ($14<<2)|0); //@line 467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $$__last = $15 ? $16 : $__last; //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ($$__last|0)==($__first|0); //@line 1748 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 if ($17) {
  $$0$lcssa$i$i = $10;
 } else {
  $scevgep$i$i = ((($$__last)) + -4|0);
  $18 = $scevgep$i$i;
  $19 = (($18) - ($1))|0;
  $$012$i$i = $__first;$$03$i$i = $10;
  while(1) {
   $20 = HEAP32[$$012$i$i>>2]|0; //@line 139 "./common/Object.h"
   $21 = ($20|0)==(0|0); //@line 132 "./common/Object.h"
   if (!($21)) {
    $22 = HEAP32[$20>>2]|0; //@line 132 "./common/Object.h"
    $23 = ((($22)) + 8|0); //@line 132 "./common/Object.h"
    $24 = HEAP32[$23>>2]|0; //@line 132 "./common/Object.h"
    FUNCTION_TABLE_vi[$24 & 8191]($20); //@line 132 "./common/Object.h"
   }
   $25 = HEAP32[$$03$i$i>>2]|0; //@line 133 "./common/Object.h"
   $26 = ($25|0)==(0|0); //@line 133 "./common/Object.h"
   if (!($26)) {
    $27 = HEAP32[$25>>2]|0; //@line 133 "./common/Object.h"
    $28 = ((($27)) + 12|0); //@line 133 "./common/Object.h"
    $29 = HEAP32[$28>>2]|0; //@line 133 "./common/Object.h"
    FUNCTION_TABLE_vi[$29 & 8191]($25); //@line 133 "./common/Object.h"
   }
   HEAP32[$$03$i$i>>2] = $20; //@line 134 "./common/Object.h"
   $30 = ((($$012$i$i)) + 4|0); //@line 1748 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $31 = ((($$03$i$i)) + 4|0); //@line 1748 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $32 = ($30|0)==($$__last|0); //@line 1748 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   if ($32) {
    break;
   } else {
    $$012$i$i = $30;$$03$i$i = $31;
   }
  }
  $33 = $19 >>> 2;
  $34 = (($33) + 1)|0;
  $scevgep7$i$i = (($10) + ($34<<2)|0);
  $$0$lcssa$i$i = $scevgep7$i$i;
 }
 if ($15) {
  $35 = ($$__last|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($35) {
   return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  $$pre$i2 = HEAP32[$11>>2]|0; //@line 1030 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$01$i4 = $16;$37 = $$pre$i2;
  while(1) {
   $36 = HEAP32[$$01$i4>>2]|0; //@line 139 "./common/Object.h"
   HEAP32[$37>>2] = $36; //@line 99 "./common/Object.h"
   $38 = ($36|0)==(0|0); //@line 101 "./common/Object.h"
   if ($38) {
    $43 = $37;
   } else {
    $39 = HEAP32[$36>>2]|0; //@line 101 "./common/Object.h"
    $40 = ((($39)) + 8|0); //@line 101 "./common/Object.h"
    $41 = HEAP32[$40>>2]|0; //@line 101 "./common/Object.h"
    FUNCTION_TABLE_vi[$41 & 8191]($36); //@line 101 "./common/Object.h"
    $$pre2$i5 = HEAP32[$11>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $43 = $$pre2$i5;
   }
   $42 = ((($43)) + 4|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$11>>2] = $42; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $44 = ((($$01$i4)) + 4|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $45 = ($44|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($45) {
    break;
   } else {
    $$01$i4 = $44;$37 = $42;
   }
  }
  return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $46 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $47 = ($46|0)==($$0$lcssa$i$i|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($47) {
  return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 } else {
  $49 = $46;
 }
 while(1) {
  $48 = ((($49)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$11>>2] = $48; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $50 = HEAP32[$48>>2]|0; //@line 106 "./common/Object.h"
  $51 = ($50|0)==(0|0); //@line 106 "./common/Object.h"
  if ($51) {
   $53 = $48;
  } else {
   $54 = HEAP32[$50>>2]|0; //@line 106 "./common/Object.h"
   $55 = ((($54)) + 12|0); //@line 106 "./common/Object.h"
   $56 = HEAP32[$55>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($56|0,($50|0)); //@line 106 "./common/Object.h"
   $57 = __THREW__; __THREW__ = 0;
   $58 = $57&1;
   if ($58) {
    label = 21;
    break;
   }
   $$pre$i$i = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $53 = $$pre$i$i;
  }
  $52 = ($53|0)==($$0$lcssa$i$i|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($52) {
   label = 41;
   break;
  } else {
   $49 = $53;
  }
 }
 if ((label|0) == 21) {
  $59 = ___cxa_find_matching_catch_3(0|0)|0;
  $60 = tempRet0;
  ___clang_call_terminate($59); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 }
 else if ((label|0) == 41) {
  return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
}
function __ZNSt3__16vectorIN4love9StrongRefINS1_8graphics4QuadEEENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i8 = 0, $$0$i9 = 0, $$01$i$i = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre12 = 0, $$pre13 = 0, $$pre14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__v$sroa$15$0 = 0, $__v$sroa$6$0 = 0, $__v$sroa$6$0$lcssa = 0, $__v$sroa$6$1 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = $$in; //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = HEAP32[$8>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $10 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($11>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $17 = $10 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ($17>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $19 = $18 ? $5 : $17; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = (($20) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = $21 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = ($19|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $24 = $20; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($23) {
   $$0$i9 = 0;$101 = $24;$28 = 0;$29 = $22;
  } else {
   $$0$i8 = $19;$100 = $22;$99 = $24;
   label = 6;
  }
 } else {
  $13 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = $14 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = $13; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $$0$i8 = 1073741823;$100 = $15;$99 = $16;
  label = 6;
 }
 if ((label|0) == 6) {
  $25 = $$0$i8 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (__Znwj($25)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i9 = $$0$i8;$101 = $99;$28 = $26;$29 = $100;
 }
 $27 = (($28) + ($29<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = $27; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $31 = (($28) + ($$0$i9<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $32 = $31; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $33 = HEAP32[$__x>>2]|0; //@line 139 "./common/Object.h"
 HEAP32[$27>>2] = $33; //@line 99 "./common/Object.h"
 $34 = ($33|0)==(0|0); //@line 101 "./common/Object.h"
 do {
  if ($34) {
   $43 = $101;$44 = $7;
   label = 10;
  } else {
   $35 = HEAP32[$33>>2]|0; //@line 101 "./common/Object.h"
   $36 = ((($35)) + 8|0); //@line 101 "./common/Object.h"
   $37 = HEAP32[$36>>2]|0; //@line 101 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($37|0,($33|0)); //@line 101 "./common/Object.h"
   $38 = __THREW__; __THREW__ = 0;
   $39 = $38&1;
   if ($39) {
    $81 = ___cxa_find_matching_catch_2()|0;
    $82 = tempRet0;
    $__v$sroa$15$0 = $30;$__v$sroa$6$1 = $30;$lpad$phi$index = $81;$lpad$phi$index2 = $82;
    break;
   } else {
    $$pre12 = HEAP32[$this>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre13 = HEAP32[$0>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $43 = $$pre13;$44 = $$pre12;
    label = 10;
    break;
   }
  }
 } while(0);
 L15: do {
  if ((label|0) == 10) {
   $40 = ((($27)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $41 = $40; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $42 = ($43|0)==($44|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   do {
    if ($42) {
     $45 = $43; //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $60 = $30;$62 = $45;$64 = $43;
    } else {
     $$01$i$i = $43;$48 = $27;$__v$sroa$6$0 = $30;
     while(1) {
      $46 = ((($$01$i$i)) + -4|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $47 = ((($48)) + -4|0); //@line 99 "./common/Object.h"
      $49 = HEAP32[$46>>2]|0; //@line 139 "./common/Object.h"
      HEAP32[$47>>2] = $49; //@line 99 "./common/Object.h"
      $50 = ($49|0)==(0|0); //@line 101 "./common/Object.h"
      if (!($50)) {
       $51 = HEAP32[$49>>2]|0; //@line 101 "./common/Object.h"
       $52 = ((($51)) + 8|0); //@line 101 "./common/Object.h"
       $53 = HEAP32[$52>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($53|0,($49|0)); //@line 101 "./common/Object.h"
       $54 = __THREW__; __THREW__ = 0;
       $55 = $54&1;
       if ($55) {
        $__v$sroa$6$0$lcssa = $__v$sroa$6$0;
        label = 24;
        break;
       }
      }
      $56 = $__v$sroa$6$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $57 = ((($56)) + -4|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $58 = $57; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $59 = ($46|0)==($44|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      if ($59) {
       $$lcssa = $58;
       label = 15;
       break;
      } else {
       $$01$i$i = $46;$48 = $57;$__v$sroa$6$0 = $58;
      }
     }
     if ((label|0) == 15) {
      $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $$pre14 = HEAP32[$0>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $60 = $$lcssa;$62 = $$pre$i;$64 = $$pre14;
      break;
     }
     else if ((label|0) == 24) {
      $79 = ___cxa_find_matching_catch_2()|0;
      $80 = tempRet0;
      $__v$sroa$15$0 = $41;$__v$sroa$6$1 = $__v$sroa$6$0$lcssa;$lpad$phi$index = $79;$lpad$phi$index2 = $80;
      break L15;
     }
    }
   } while(0);
   HEAP32[$this>>2] = $60; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$0>>2] = $41; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$8>>2] = $32; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $61 = $62; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $63 = ($64|0)==($61|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   L28: do {
    if (!($63)) {
     $66 = $64;
     while(1) {
      $65 = ((($66)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $67 = HEAP32[$65>>2]|0; //@line 106 "./common/Object.h"
      $68 = ($67|0)==(0|0); //@line 106 "./common/Object.h"
      if (!($68)) {
       $70 = HEAP32[$67>>2]|0; //@line 106 "./common/Object.h"
       $71 = ((($70)) + 12|0); //@line 106 "./common/Object.h"
       $72 = HEAP32[$71>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($72|0,($67|0)); //@line 106 "./common/Object.h"
       $73 = __THREW__; __THREW__ = 0;
       $74 = $73&1;
       if ($74) {
        break;
       }
      }
      $69 = ($65|0)==($61|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if ($69) {
       break L28;
      } else {
       $66 = $65;
      }
     }
     $75 = ___cxa_find_matching_catch_3(0|0)|0;
     $76 = tempRet0;
     ___clang_call_terminate($75); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   } while(0);
   $77 = ($62|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($77) {
    return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $78 = $62; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __ZdlPv($78); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
 } while(0);
 $83 = $__v$sroa$6$1; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $84 = $__v$sroa$15$0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $85 = ($84|0)==($83|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 L41: do {
  if (!($85)) {
   $87 = $84;
   while(1) {
    $86 = ((($87)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $88 = HEAP32[$86>>2]|0; //@line 106 "./common/Object.h"
    $89 = ($88|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($89)) {
     $91 = HEAP32[$88>>2]|0; //@line 106 "./common/Object.h"
     $92 = ((($91)) + 12|0); //@line 106 "./common/Object.h"
     $93 = HEAP32[$92>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($93|0,($88|0)); //@line 106 "./common/Object.h"
     $94 = __THREW__; __THREW__ = 0;
     $95 = $94&1;
     if ($95) {
      break;
     }
    }
    $90 = ($86|0)==($83|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($90) {
     break L41;
    } else {
     $87 = $86;
    }
   }
   $96 = ___cxa_find_matching_catch_3(0|0)|0;
   $97 = tempRet0;
   ___clang_call_terminate($96); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 } while(0);
 $98 = ($28|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($98) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 __ZdlPv($28); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZNK4love8graphics14ParticleSystem8getQuadsEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$pre = 0, $$pre3 = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$02 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 HEAP32[$agg$result>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($agg$result)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($agg$result)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = ((($this)) + 212|0); //@line 739 "modules/graphics/ParticleSystem.cpp"
 $4 = ((($this)) + 216|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$4>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = (($5) - ($6))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = $7 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ((($agg$result)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $10 = ($8|0)==(0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $6; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = $5; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($10) {
  $38 = 0;$39 = 0;
 } else {
  $13 = (__Znwj($7)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $14 = $13; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $15 = (($13) + ($8<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $16 = $15; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  HEAP32[$agg$result>>2] = $14; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$1>>2] = $14; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$9>>2] = $16; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $38 = $15;$39 = $13;
 }
 $17 = ($11|0)==($12|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($17) {
  STACKTOP = sp;return; //@line 745 "modules/graphics/ParticleSystem.cpp"
 }
 $30 = $39;$31 = $38;$__begin$sroa$0$02 = $11;
 while(1) {
  $28 = HEAP32[$__begin$sroa$0$02>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$0>>2] = $28; //@line 742 "modules/graphics/ParticleSystem.cpp"
  $29 = ($30>>>0)<($31>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($29) {
   HEAP32[$30>>2] = $28; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $32 = HEAP32[$1>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $33 = ((($32)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$1>>2] = $33; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  } else {
   __THREW__ = 0;
   invoke_vii(3634,($agg$result|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $34 = __THREW__; __THREW__ = 0;
   $35 = $34&1;
   if ($35) {
    break;
   }
  }
  $36 = ((($__begin$sroa$0$02)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $37 = ($36|0)==($12|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($37) {
   label = 5;
   break;
  }
  $$pre = HEAP32[$1>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre3 = HEAP32[$9>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $30 = $$pre;$31 = $$pre3;$__begin$sroa$0$02 = $36;
 }
 if ((label|0) == 5) {
  STACKTOP = sp;return; //@line 745 "modules/graphics/ParticleSystem.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 $$pre4 = HEAP32[$agg$result>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = ($$pre4|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = $$pre4; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($20) {
  ___resumeException($18|0);
  // unreachable;
 }
 $22 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $23 = ($22|0)==($$pre4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($23)) {
  $scevgep$i$i$i = ((($22)) + -4|0);
  $24 = $scevgep$i$i$i;
  $25 = (($24) - ($21))|0;
  $26 = $25 >>> 2;
  $27 = $26 ^ -1;
  $scevgep4$i$i$i = (($22) + ($27<<2)|0);
  HEAP32[$1>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($$pre4); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($18|0);
 // unreachable;
}
function __ZNSt3__16vectorIPN4love8graphics4QuadENS_9allocatorIS4_EEE21__push_back_slow_pathIS4_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics14ParticleSystem19setRelativeRotationEb($this,$enable) {
 $this = $this|0;
 $enable = $enable|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 224|0); //@line 749 "modules/graphics/ParticleSystem.cpp"
 $1 = $enable&1; //@line 749 "modules/graphics/ParticleSystem.cpp"
 HEAP8[$0>>0] = $1; //@line 749 "modules/graphics/ParticleSystem.cpp"
 return; //@line 750 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem19hasRelativeRotationEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 224|0); //@line 754 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 754 "modules/graphics/ParticleSystem.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 754 "modules/graphics/ParticleSystem.cpp"
 return ($2|0); //@line 754 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem8getCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0); //@line 759 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 759 "modules/graphics/ParticleSystem.cpp"
 return ($1|0); //@line 759 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem5startEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 764 "modules/graphics/ParticleSystem.cpp"
 HEAP8[$0>>0] = 1; //@line 764 "modules/graphics/ParticleSystem.cpp"
 return; //@line 765 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem4stopEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 769 "modules/graphics/ParticleSystem.cpp"
 HEAP8[$0>>0] = 0; //@line 769 "modules/graphics/ParticleSystem.cpp"
 $1 = ((($this)) + 80|0); //@line 770 "modules/graphics/ParticleSystem.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 770 "modules/graphics/ParticleSystem.cpp"
 $3 = ((($this)) + 84|0); //@line 770 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$3>>2] = $2; //@line 770 "modules/graphics/ParticleSystem.cpp"
 $4 = ((($this)) + 48|0); //@line 771 "modules/graphics/ParticleSystem.cpp"
 HEAPF32[$4>>2] = 0.0; //@line 771 "modules/graphics/ParticleSystem.cpp"
 return; //@line 772 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem5pauseEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 776 "modules/graphics/ParticleSystem.cpp"
 HEAP8[$0>>0] = 0; //@line 776 "modules/graphics/ParticleSystem.cpp"
 return; //@line 777 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem4emitEj($this,$num) {
 $this = $this|0;
 $num = $num|0;
 var $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 794 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 794 "modules/graphics/ParticleSystem.cpp"
 $2 = ($1<<24>>24)==(0); //@line 794 "modules/graphics/ParticleSystem.cpp"
 if ($2) {
  return; //@line 801 "modules/graphics/ParticleSystem.cpp"
 }
 $3 = ((($this)) + 36|0); //@line 797 "modules/graphics/ParticleSystem.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 797 "modules/graphics/ParticleSystem.cpp"
 $5 = ((($this)) + 40|0); //@line 797 "modules/graphics/ParticleSystem.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 797 "modules/graphics/ParticleSystem.cpp"
 $7 = (($4) - ($6))|0; //@line 797 "modules/graphics/ParticleSystem.cpp"
 $8 = ($7>>>0)<($num>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $9 = $8 ? $7 : $num; //@line 797 "modules/graphics/ParticleSystem.cpp"
 $10 = ($9|0)==(0); //@line 799 "modules/graphics/ParticleSystem.cpp"
 if ($10) {
  return; //@line 801 "modules/graphics/ParticleSystem.cpp"
 } else {
  $$01 = $9;
 }
 while(1) {
  $11 = (($$01) + -1)|0; //@line 799 "modules/graphics/ParticleSystem.cpp"
  __ZN4love8graphics14ParticleSystem11addParticleEf($this,1.0); //@line 800 "modules/graphics/ParticleSystem.cpp"
  $12 = ($11|0)==(0); //@line 799 "modules/graphics/ParticleSystem.cpp"
  if ($12) {
   break;
  } else {
   $$01 = $11;
  }
 }
 return; //@line 801 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem8isActiveEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 805 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 805 "modules/graphics/ParticleSystem.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 805 "modules/graphics/ParticleSystem.cpp"
 return ($2|0); //@line 805 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem8isPausedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 810 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 810 "modules/graphics/ParticleSystem.cpp"
 $2 = ($1<<24>>24)==(0); //@line 810 "modules/graphics/ParticleSystem.cpp"
 if (!($2)) {
  $8 = 0;
  return ($8|0); //@line 810 "modules/graphics/ParticleSystem.cpp"
 }
 $3 = ((($this)) + 84|0); //@line 810 "modules/graphics/ParticleSystem.cpp"
 $4 = +HEAPF32[$3>>2]; //@line 810 "modules/graphics/ParticleSystem.cpp"
 $5 = ((($this)) + 80|0); //@line 810 "modules/graphics/ParticleSystem.cpp"
 $6 = +HEAPF32[$5>>2]; //@line 810 "modules/graphics/ParticleSystem.cpp"
 $7 = $4 < $6; //@line 810 "modules/graphics/ParticleSystem.cpp"
 $8 = $7;
 return ($8|0); //@line 810 "modules/graphics/ParticleSystem.cpp"
}
function __ZNK4love8graphics14ParticleSystem9isStoppedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 815 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 815 "modules/graphics/ParticleSystem.cpp"
 $2 = ($1<<24>>24)==(0); //@line 815 "modules/graphics/ParticleSystem.cpp"
 if (!($2)) {
  $8 = 0;
  return ($8|0); //@line 815 "modules/graphics/ParticleSystem.cpp"
 }
 $3 = ((($this)) + 84|0); //@line 815 "modules/graphics/ParticleSystem.cpp"
 $4 = +HEAPF32[$3>>2]; //@line 815 "modules/graphics/ParticleSystem.cpp"
 $5 = ((($this)) + 80|0); //@line 815 "modules/graphics/ParticleSystem.cpp"
 $6 = +HEAPF32[$5>>2]; //@line 815 "modules/graphics/ParticleSystem.cpp"
 $7 = $4 >= $6; //@line 815 "modules/graphics/ParticleSystem.cpp"
 $8 = $7;
 return ($8|0); //@line 815 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem6updateEf($this,$dt) {
 $this = $this|0;
 $dt = +$dt;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0, $112 = 0, $113 = 0, $114 = 0.0;
 var $115 = 0.0, $116 = 0, $117 = 0.0, $118 = 0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0.0, $128 = 0.0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0.0;
 var $133 = 0.0, $134 = 0, $135 = 0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0.0, $15 = 0, $150 = 0.0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0.0, $155 = 0.0, $156 = 0, $157 = 0.0, $158 = 0, $159 = 0.0, $16 = 0, $160 = 0, $161 = 0.0, $162 = 0, $163 = 0.0, $164 = 0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0.0;
 var $17 = 0, $170 = 0, $171 = 0.0, $172 = 0, $173 = 0.0, $174 = 0, $175 = 0.0, $176 = 0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0.0, $180 = 0.0, $181 = 0.0, $182 = 0.0, $183 = 0.0, $184 = 0.0, $185 = 0.0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0.0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0.0, $196 = 0.0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0.0, $212 = 0.0, $213 = 0, $214 = 0.0, $215 = 0.0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0.0, $22 = 0, $220 = 0.0, $221 = 0.0, $222 = 0.0;
 var $223 = 0.0, $224 = 0, $225 = 0, $226 = 0.0, $227 = 0.0, $228 = 0, $229 = 0.0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0;
 var $241 = 0, $242 = 0, $243 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0.0, $44 = 0, $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0.0;
 var $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0.0, $76 = 0.0;
 var $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0.0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0;
 var $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0, $not$ = 0, $not$5 = 0, $or$cond = 0, $or$cond6 = 0, $p$0$be = 0, $p$024 = 0, $p$pNext$0$i = 0, $pNext$0$i = 0, $pNext$2$i = 0, $radial$sroa$0$0$in = 0.0, $radial$sroa$10$0$in = 0.0, dest = 0, label = 0, sp = 0, src = 0;
 var stop = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 830 "modules/graphics/ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 830 "modules/graphics/ParticleSystem.cpp"
 $2 = ($1|0)==(0|0); //@line 830 "modules/graphics/ParticleSystem.cpp"
 $3 = $dt == 0.0; //@line 830 "modules/graphics/ParticleSystem.cpp"
 $or$cond = $3 | $2; //@line 830 "modules/graphics/ParticleSystem.cpp"
 if ($or$cond) {
  return; //@line 945 "modules/graphics/ParticleSystem.cpp"
 }
 $4 = ((($this)) + 16|0); //@line 834 "modules/graphics/ParticleSystem.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 834 "modules/graphics/ParticleSystem.cpp"
 $6 = ($5|0)==(0|0); //@line 836 "modules/graphics/ParticleSystem.cpp"
 if (!($6)) {
  $7 = ((($this)) + 224|0); //@line 885 "modules/graphics/ParticleSystem.cpp"
  $8 = ((($this)) + 152|0); //@line 897 "modules/graphics/ParticleSystem.cpp"
  $9 = ((($this)) + 156|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $10 = ((($this)) + 200|0); //@line 904 "modules/graphics/ParticleSystem.cpp"
  $11 = ((($this)) + 204|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $12 = ((($this)) + 212|0); //@line 911 "modules/graphics/ParticleSystem.cpp"
  $13 = ((($this)) + 216|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = ((($this)) + 20|0); //@line 386 "modules/graphics/ParticleSystem.cpp"
  $15 = ((($this)) + 12|0); //@line 390 "modules/graphics/ParticleSystem.cpp"
  $16 = ((($this)) + 40|0); //@line 408 "modules/graphics/ParticleSystem.cpp"
  $p$024 = $5;
  while(1) {
   $17 = ((($p$024)) + 12|0); //@line 839 "modules/graphics/ParticleSystem.cpp"
   $18 = +HEAPF32[$17>>2]; //@line 839 "modules/graphics/ParticleSystem.cpp"
   $19 = $18 - $dt; //@line 878 "modules/graphics/ParticleSystem.cpp"
   HEAPF32[$17>>2] = $19; //@line 839 "modules/graphics/ParticleSystem.cpp"
   $20 = !($19 <= 0.0); //@line 841 "modules/graphics/ParticleSystem.cpp"
   if ($20) {
    $42 = ((($p$024)) + 16|0); //@line 847 "modules/graphics/ParticleSystem.cpp"
    $43 = +HEAPF32[$42>>2]; //@line 847 "modules/graphics/ParticleSystem.cpp"
    $44 = ((($p$024)) + 20|0); //@line 847 "modules/graphics/ParticleSystem.cpp"
    $45 = +HEAPF32[$44>>2]; //@line 847 "modules/graphics/ParticleSystem.cpp"
    $46 = ((($p$024)) + 24|0); //@line 305 "./common/Vector.h"
    $47 = +HEAPF32[$46>>2]; //@line 305 "./common/Vector.h"
    $48 = $43 - $47; //@line 237 "./common/Vector.h"
    $49 = ((($p$024)) + 28|0); //@line 310 "./common/Vector.h"
    $50 = +HEAPF32[$49>>2]; //@line 310 "./common/Vector.h"
    $51 = $45 - $50; //@line 237 "./common/Vector.h"
    $52 = $48 * $48; //@line 190 "./common/Vector.h"
    $53 = $51 * $51; //@line 190 "./common/Vector.h"
    $54 = $52 + $53; //@line 190 "./common/Vector.h"
    $55 = (+Math_sqrt((+$54))); //@line 190 "./common/Vector.h"
    $56 = $55 > 0.0; //@line 208 "./common/Vector.h"
    if ($56) {
     $57 = 1.0 / $55; //@line 209 "./common/Vector.h"
     $58 = $48 * $57; //@line 269 "./common/Vector.h"
     $59 = $51 * $57; //@line 270 "./common/Vector.h"
     $radial$sroa$0$0$in = $58;$radial$sroa$10$0$in = $59;
    } else {
     $radial$sroa$0$0$in = $48;$radial$sroa$10$0$in = $51;
    }
    $60 = ((($p$024)) + 48|0); //@line 855 "modules/graphics/ParticleSystem.cpp"
    $61 = +HEAPF32[$60>>2]; //@line 855 "modules/graphics/ParticleSystem.cpp"
    $62 = $radial$sroa$0$0$in * $61; //@line 269 "./common/Vector.h"
    $63 = $radial$sroa$10$0$in * $61; //@line 270 "./common/Vector.h"
    $64 = ((($p$024)) + 52|0); //@line 865 "modules/graphics/ParticleSystem.cpp"
    $65 = +HEAPF32[$64>>2]; //@line 865 "modules/graphics/ParticleSystem.cpp"
    $66 = $radial$sroa$10$0$in * $65; //@line 269 "./common/Vector.h"
    $67 = $radial$sroa$0$0$in * $65; //@line 270 "./common/Vector.h"
    $68 = $62 - $66; //@line 232 "./common/Vector.h"
    $69 = $63 + $67; //@line 232 "./common/Vector.h"
    $70 = ((($p$024)) + 40|0); //@line 232 "./common/Vector.h"
    $71 = +HEAPF32[$70>>2]; //@line 232 "./common/Vector.h"
    $72 = $71 + $68; //@line 232 "./common/Vector.h"
    $73 = ((($p$024)) + 44|0); //@line 232 "./common/Vector.h"
    $74 = +HEAPF32[$73>>2]; //@line 232 "./common/Vector.h"
    $75 = $69 + $74; //@line 232 "./common/Vector.h"
    $76 = $72 * $dt; //@line 242 "./common/Vector.h"
    $77 = $75 * $dt; //@line 242 "./common/Vector.h"
    $78 = ((($p$024)) + 32|0); //@line 257 "./common/Vector.h"
    $79 = +HEAPF32[$78>>2]; //@line 257 "./common/Vector.h"
    $80 = $79 + $76; //@line 257 "./common/Vector.h"
    $81 = ((($p$024)) + 36|0); //@line 258 "./common/Vector.h"
    $82 = +HEAPF32[$81>>2]; //@line 258 "./common/Vector.h"
    $83 = $77 + $82; //@line 258 "./common/Vector.h"
    $84 = ((($p$024)) + 56|0); //@line 871 "modules/graphics/ParticleSystem.cpp"
    $85 = +HEAPF32[$84>>2]; //@line 871 "modules/graphics/ParticleSystem.cpp"
    $86 = $85 * $dt; //@line 871 "modules/graphics/ParticleSystem.cpp"
    $87 = $86 + 1.0; //@line 871 "modules/graphics/ParticleSystem.cpp"
    $88 = 1.0 / $87; //@line 871 "modules/graphics/ParticleSystem.cpp"
    $89 = $80 * $88; //@line 886 "modules/graphics/ParticleSystem.cpp"
    HEAPF32[$78>>2] = $89; //@line 269 "./common/Vector.h"
    $90 = $83 * $88; //@line 886 "modules/graphics/ParticleSystem.cpp"
    HEAPF32[$81>>2] = $90; //@line 270 "./common/Vector.h"
    $91 = $89 * $dt; //@line 242 "./common/Vector.h"
    $92 = $90 * $dt; //@line 242 "./common/Vector.h"
    $93 = $43 + $91; //@line 257 "./common/Vector.h"
    $94 = $45 + $92; //@line 258 "./common/Vector.h"
    HEAPF32[$42>>2] = $93; //@line 876 "modules/graphics/ParticleSystem.cpp"
    HEAPF32[$44>>2] = $94; //@line 876 "modules/graphics/ParticleSystem.cpp"
    $95 = ((($p$024)) + 8|0); //@line 878 "modules/graphics/ParticleSystem.cpp"
    $96 = +HEAPF32[$95>>2]; //@line 878 "modules/graphics/ParticleSystem.cpp"
    $97 = $19 / $96; //@line 878 "modules/graphics/ParticleSystem.cpp"
    $98 = 1.0 - $97; //@line 878 "modules/graphics/ParticleSystem.cpp"
    $99 = ((($p$024)) + 80|0); //@line 881 "modules/graphics/ParticleSystem.cpp"
    $100 = +HEAPF32[$99>>2]; //@line 881 "modules/graphics/ParticleSystem.cpp"
    $101 = 1.0 - $98; //@line 881 "modules/graphics/ParticleSystem.cpp"
    $102 = $100 * $101; //@line 881 "modules/graphics/ParticleSystem.cpp"
    $103 = ((($p$024)) + 84|0); //@line 881 "modules/graphics/ParticleSystem.cpp"
    $104 = +HEAPF32[$103>>2]; //@line 881 "modules/graphics/ParticleSystem.cpp"
    $105 = $104 * $98; //@line 881 "modules/graphics/ParticleSystem.cpp"
    $106 = $105 + $102; //@line 881 "modules/graphics/ParticleSystem.cpp"
    $107 = $106 * $dt; //@line 881 "modules/graphics/ParticleSystem.cpp"
    $108 = ((($p$024)) + 72|0); //@line 881 "modules/graphics/ParticleSystem.cpp"
    $109 = +HEAPF32[$108>>2]; //@line 881 "modules/graphics/ParticleSystem.cpp"
    $110 = $109 + $107; //@line 881 "modules/graphics/ParticleSystem.cpp"
    HEAPF32[$108>>2] = $110; //@line 881 "modules/graphics/ParticleSystem.cpp"
    $111 = ((($p$024)) + 76|0); //@line 883 "modules/graphics/ParticleSystem.cpp"
    HEAPF32[$111>>2] = $110; //@line 883 "modules/graphics/ParticleSystem.cpp"
    $112 = HEAP8[$7>>0]|0; //@line 885 "modules/graphics/ParticleSystem.cpp"
    $113 = ($112<<24>>24)==(0); //@line 885 "modules/graphics/ParticleSystem.cpp"
    if (!($113)) {
     $114 = (+Math_atan2((+$90),(+$89))); //@line 886 "modules/graphics/ParticleSystem.cpp"
     $115 = $110 + $114; //@line 886 "modules/graphics/ParticleSystem.cpp"
     HEAPF32[$111>>2] = $115; //@line 886 "modules/graphics/ParticleSystem.cpp"
    }
    $116 = ((($p$024)) + 64|0); //@line 896 "modules/graphics/ParticleSystem.cpp"
    $117 = +HEAPF32[$116>>2]; //@line 896 "modules/graphics/ParticleSystem.cpp"
    $118 = ((($p$024)) + 68|0); //@line 896 "modules/graphics/ParticleSystem.cpp"
    $119 = +HEAPF32[$118>>2]; //@line 896 "modules/graphics/ParticleSystem.cpp"
    $120 = $98 * $119; //@line 896 "modules/graphics/ParticleSystem.cpp"
    $121 = $117 + $120; //@line 896 "modules/graphics/ParticleSystem.cpp"
    $122 = HEAP32[$9>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $123 = HEAP32[$8>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $124 = (($122) - ($123))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $125 = $124 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $126 = (($125) + -1)|0; //@line 897 "modules/graphics/ParticleSystem.cpp"
    $127 = (+($126>>>0)); //@line 897 "modules/graphics/ParticleSystem.cpp"
    $128 = $121 * $127; //@line 897 "modules/graphics/ParticleSystem.cpp"
    $129 = (~~(($128))>>>0); //@line 898 "modules/graphics/ParticleSystem.cpp"
    $not$ = ($129|0)!=($126|0); //@line 899 "modules/graphics/ParticleSystem.cpp"
    $130 = $not$&1; //@line 899 "modules/graphics/ParticleSystem.cpp"
    $131 = (($130) + ($129))|0; //@line 899 "modules/graphics/ParticleSystem.cpp"
    $132 = (+($129>>>0)); //@line 900 "modules/graphics/ParticleSystem.cpp"
    $133 = $128 - $132; //@line 900 "modules/graphics/ParticleSystem.cpp"
    $134 = $123; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $135 = (($134) + ($129<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $136 = +HEAPF32[$135>>2]; //@line 901 "modules/graphics/ParticleSystem.cpp"
    $137 = 1.0 - $133; //@line 901 "modules/graphics/ParticleSystem.cpp"
    $138 = $136 * $137; //@line 901 "modules/graphics/ParticleSystem.cpp"
    $139 = (($134) + ($131<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $140 = +HEAPF32[$139>>2]; //@line 901 "modules/graphics/ParticleSystem.cpp"
    $141 = $140 * $133; //@line 901 "modules/graphics/ParticleSystem.cpp"
    $142 = $141 + $138; //@line 901 "modules/graphics/ParticleSystem.cpp"
    $143 = ((($p$024)) + 60|0); //@line 901 "modules/graphics/ParticleSystem.cpp"
    HEAPF32[$143>>2] = $142; //@line 901 "modules/graphics/ParticleSystem.cpp"
    $144 = HEAP32[$11>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $145 = HEAP32[$10>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $146 = (($144) - ($145))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $147 = $146 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $148 = (($147) + -1)|0; //@line 904 "modules/graphics/ParticleSystem.cpp"
    $149 = (+($148>>>0)); //@line 904 "modules/graphics/ParticleSystem.cpp"
    $150 = $98 * $149; //@line 904 "modules/graphics/ParticleSystem.cpp"
    $151 = (~~(($150))>>>0); //@line 905 "modules/graphics/ParticleSystem.cpp"
    $not$5 = ($151|0)!=($148|0); //@line 906 "modules/graphics/ParticleSystem.cpp"
    $152 = $not$5&1; //@line 906 "modules/graphics/ParticleSystem.cpp"
    $153 = (($152) + ($151))|0; //@line 906 "modules/graphics/ParticleSystem.cpp"
    $154 = (+($151>>>0)); //@line 907 "modules/graphics/ParticleSystem.cpp"
    $155 = $150 - $154; //@line 907 "modules/graphics/ParticleSystem.cpp"
    $156 = $145; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $157 = 1.0 - $155; //@line 908 "modules/graphics/ParticleSystem.cpp"
    $158 = (($156) + ($151<<4)|0); //@line 107 "modules/graphics/Color.h"
    $159 = +HEAPF32[$158>>2]; //@line 107 "modules/graphics/Color.h"
    $160 = (((($156) + ($151<<4)|0)) + 4|0); //@line 107 "modules/graphics/Color.h"
    $161 = +HEAPF32[$160>>2]; //@line 107 "modules/graphics/Color.h"
    $162 = (((($156) + ($151<<4)|0)) + 8|0); //@line 107 "modules/graphics/Color.h"
    $163 = +HEAPF32[$162>>2]; //@line 107 "modules/graphics/Color.h"
    $164 = (((($156) + ($151<<4)|0)) + 12|0); //@line 107 "modules/graphics/Color.h"
    $165 = +HEAPF32[$164>>2]; //@line 107 "modules/graphics/Color.h"
    $166 = $159 * $157; //@line 80 "modules/graphics/Color.h"
    $167 = $161 * $157; //@line 81 "modules/graphics/Color.h"
    $168 = $163 * $157; //@line 82 "modules/graphics/Color.h"
    $169 = $165 * $157; //@line 83 "modules/graphics/Color.h"
    $170 = (($156) + ($153<<4)|0); //@line 107 "modules/graphics/Color.h"
    $171 = +HEAPF32[$170>>2]; //@line 107 "modules/graphics/Color.h"
    $172 = (((($156) + ($153<<4)|0)) + 4|0); //@line 107 "modules/graphics/Color.h"
    $173 = +HEAPF32[$172>>2]; //@line 107 "modules/graphics/Color.h"
    $174 = (((($156) + ($153<<4)|0)) + 8|0); //@line 107 "modules/graphics/Color.h"
    $175 = +HEAPF32[$174>>2]; //@line 107 "modules/graphics/Color.h"
    $176 = (((($156) + ($153<<4)|0)) + 12|0); //@line 107 "modules/graphics/Color.h"
    $177 = +HEAPF32[$176>>2]; //@line 107 "modules/graphics/Color.h"
    $178 = $171 * $155; //@line 80 "modules/graphics/Color.h"
    $179 = $173 * $155; //@line 81 "modules/graphics/Color.h"
    $180 = $175 * $155; //@line 82 "modules/graphics/Color.h"
    $181 = $155 * $177; //@line 83 "modules/graphics/Color.h"
    $182 = $178 + $166; //@line 70 "modules/graphics/Color.h"
    $183 = $179 + $167; //@line 71 "modules/graphics/Color.h"
    $184 = $180 + $168; //@line 72 "modules/graphics/Color.h"
    $185 = $169 + $181; //@line 73 "modules/graphics/Color.h"
    $186 = ((($p$024)) + 88|0); //@line 908 "modules/graphics/ParticleSystem.cpp"
    HEAPF32[$186>>2] = $182; //@line 908 "modules/graphics/ParticleSystem.cpp"
    $187 = ((($p$024)) + 92|0); //@line 908 "modules/graphics/ParticleSystem.cpp"
    HEAPF32[$187>>2] = $183; //@line 908 "modules/graphics/ParticleSystem.cpp"
    $188 = ((($p$024)) + 96|0); //@line 908 "modules/graphics/ParticleSystem.cpp"
    HEAPF32[$188>>2] = $184; //@line 908 "modules/graphics/ParticleSystem.cpp"
    $189 = ((($p$024)) + 100|0); //@line 908 "modules/graphics/ParticleSystem.cpp"
    HEAPF32[$189>>2] = $185; //@line 908 "modules/graphics/ParticleSystem.cpp"
    $190 = HEAP32[$13>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $191 = HEAP32[$12>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $192 = (($190) - ($191))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $193 = $192 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $194 = ($193|0)==(0); //@line 912 "modules/graphics/ParticleSystem.cpp"
    if (!($194)) {
     $195 = (+($193>>>0)); //@line 914 "modules/graphics/ParticleSystem.cpp"
     $196 = $98 * $195; //@line 914 "modules/graphics/ParticleSystem.cpp"
     $197 = $196 > 0.0; //@line 915 "modules/graphics/ParticleSystem.cpp"
     $198 = (~~(($196))>>>0); //@line 915 "modules/graphics/ParticleSystem.cpp"
     $199 = $197 ? $198 : 0; //@line 915 "modules/graphics/ParticleSystem.cpp"
     $200 = ($199>>>0)<($193>>>0); //@line 916 "modules/graphics/ParticleSystem.cpp"
     $201 = (($193) + -1)|0; //@line 916 "modules/graphics/ParticleSystem.cpp"
     $202 = $200 ? $199 : $201; //@line 916 "modules/graphics/ParticleSystem.cpp"
     $203 = ((($p$024)) + 104|0); //@line 916 "modules/graphics/ParticleSystem.cpp"
     HEAP32[$203>>2] = $202; //@line 916 "modules/graphics/ParticleSystem.cpp"
    }
    $204 = ((($p$024)) + 4|0); //@line 920 "modules/graphics/ParticleSystem.cpp"
    $205 = HEAP32[$204>>2]|0; //@line 920 "modules/graphics/ParticleSystem.cpp"
    $p$0$be = $205;
   } else {
    $21 = HEAP32[$p$024>>2]|0; //@line 375 "modules/graphics/ParticleSystem.cpp"
    $22 = ($21|0)==(0|0); //@line 375 "modules/graphics/ParticleSystem.cpp"
    $23 = ((($p$024)) + 4|0); //@line 376 "modules/graphics/ParticleSystem.cpp"
    $24 = HEAP32[$23>>2]|0; //@line 376 "modules/graphics/ParticleSystem.cpp"
    $25 = $21; //@line 382 "modules/graphics/ParticleSystem.cpp"
    if ($22) {
     $27 = $24; //@line 375 "modules/graphics/ParticleSystem.cpp"
     HEAP32[$4>>2] = $24; //@line 378 "modules/graphics/ParticleSystem.cpp"
     $29 = $27;
    } else {
     $26 = ((($21)) + 4|0); //@line 376 "modules/graphics/ParticleSystem.cpp"
     HEAP32[$26>>2] = $24; //@line 376 "modules/graphics/ParticleSystem.cpp"
     $$pre$i = HEAP32[$23>>2]|0; //@line 380 "modules/graphics/ParticleSystem.cpp"
     $29 = $$pre$i;
    }
    $28 = ($29|0)==(0|0); //@line 380 "modules/graphics/ParticleSystem.cpp"
    if ($28) {
     HEAP32[$14>>2] = $25; //@line 386 "modules/graphics/ParticleSystem.cpp"
     $pNext$0$i = 0;
    } else {
     HEAP32[$29>>2] = $25; //@line 382 "modules/graphics/ParticleSystem.cpp"
     $pNext$0$i = $29;
    }
    $30 = HEAP32[$15>>2]|0; //@line 390 "modules/graphics/ParticleSystem.cpp"
    $31 = ((($30)) + -108|0); //@line 390 "modules/graphics/ParticleSystem.cpp"
    HEAP32[$15>>2] = $31; //@line 390 "modules/graphics/ParticleSystem.cpp"
    $32 = ($31|0)==($p$024|0); //@line 391 "modules/graphics/ParticleSystem.cpp"
    do {
     if ($32) {
      $pNext$2$i = $pNext$0$i;
     } else {
      dest=$p$024; src=$31; stop=dest+108|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 393 "modules/graphics/ParticleSystem.cpp"
      $33 = HEAP32[$15>>2]|0; //@line 394 "modules/graphics/ParticleSystem.cpp"
      $34 = ($pNext$0$i|0)==($33|0); //@line 394 "modules/graphics/ParticleSystem.cpp"
      $p$pNext$0$i = $34 ? $p$024 : $pNext$0$i; //@line 394 "modules/graphics/ParticleSystem.cpp"
      $35 = HEAP32[$p$024>>2]|0; //@line 397 "modules/graphics/ParticleSystem.cpp"
      $36 = ($35|0)==(0|0); //@line 397 "modules/graphics/ParticleSystem.cpp"
      if ($36) {
       HEAP32[$4>>2] = $p$024; //@line 400 "modules/graphics/ParticleSystem.cpp"
      } else {
       $37 = ((($35)) + 4|0); //@line 398 "modules/graphics/ParticleSystem.cpp"
       HEAP32[$37>>2] = $p$024; //@line 398 "modules/graphics/ParticleSystem.cpp"
      }
      $38 = HEAP32[$23>>2]|0; //@line 402 "modules/graphics/ParticleSystem.cpp"
      $39 = ($38|0)==(0|0); //@line 402 "modules/graphics/ParticleSystem.cpp"
      if ($39) {
       HEAP32[$14>>2] = $p$024; //@line 405 "modules/graphics/ParticleSystem.cpp"
       $pNext$2$i = $p$pNext$0$i;
       break;
      } else {
       HEAP32[$38>>2] = $p$024; //@line 403 "modules/graphics/ParticleSystem.cpp"
       $pNext$2$i = $p$pNext$0$i;
       break;
      }
     }
    } while(0);
    $40 = HEAP32[$16>>2]|0; //@line 408 "modules/graphics/ParticleSystem.cpp"
    $41 = (($40) + -1)|0; //@line 408 "modules/graphics/ParticleSystem.cpp"
    HEAP32[$16>>2] = $41; //@line 408 "modules/graphics/ParticleSystem.cpp"
    $p$0$be = $pNext$2$i;
   }
   $206 = ($p$0$be|0)==(0|0); //@line 836 "modules/graphics/ParticleSystem.cpp"
   if ($206) {
    break;
   } else {
    $p$024 = $p$0$be;
   }
  }
 }
 $207 = ((($this)) + 28|0); //@line 925 "modules/graphics/ParticleSystem.cpp"
 $208 = HEAP8[$207>>0]|0; //@line 925 "modules/graphics/ParticleSystem.cpp"
 $209 = ($208<<24>>24)==(0); //@line 925 "modules/graphics/ParticleSystem.cpp"
 if (!($209)) {
  $210 = ((($this)) + 44|0); //@line 927 "modules/graphics/ParticleSystem.cpp"
  $211 = +HEAPF32[$210>>2]; //@line 927 "modules/graphics/ParticleSystem.cpp"
  $212 = 1.0 / $211; //@line 927 "modules/graphics/ParticleSystem.cpp"
  $213 = ((($this)) + 48|0); //@line 928 "modules/graphics/ParticleSystem.cpp"
  $214 = +HEAPF32[$213>>2]; //@line 928 "modules/graphics/ParticleSystem.cpp"
  $215 = $214 + $dt; //@line 928 "modules/graphics/ParticleSystem.cpp"
  HEAPF32[$213>>2] = $215; //@line 928 "modules/graphics/ParticleSystem.cpp"
  $216 = $215 - $212; //@line 929 "modules/graphics/ParticleSystem.cpp"
  $217 = $215 > $212; //@line 930 "modules/graphics/ParticleSystem.cpp"
  if ($217) {
   $219 = $215;
   while(1) {
    $218 = $219 - $212; //@line 932 "modules/graphics/ParticleSystem.cpp"
    $220 = $218 / $216; //@line 932 "modules/graphics/ParticleSystem.cpp"
    $221 = 1.0 - $220; //@line 932 "modules/graphics/ParticleSystem.cpp"
    __ZN4love8graphics14ParticleSystem11addParticleEf($this,$221); //@line 932 "modules/graphics/ParticleSystem.cpp"
    $222 = +HEAPF32[$213>>2]; //@line 933 "modules/graphics/ParticleSystem.cpp"
    $223 = $222 - $212; //@line 933 "modules/graphics/ParticleSystem.cpp"
    HEAPF32[$213>>2] = $223; //@line 933 "modules/graphics/ParticleSystem.cpp"
    $224 = $223 > $212; //@line 930 "modules/graphics/ParticleSystem.cpp"
    if ($224) {
     $219 = $223;
    } else {
     break;
    }
   }
  }
  $225 = ((($this)) + 84|0); //@line 939 "modules/graphics/ParticleSystem.cpp"
  $226 = +HEAPF32[$225>>2]; //@line 939 "modules/graphics/ParticleSystem.cpp"
  $227 = $226 - $dt; //@line 939 "modules/graphics/ParticleSystem.cpp"
  HEAPF32[$225>>2] = $227; //@line 939 "modules/graphics/ParticleSystem.cpp"
  $228 = ((($this)) + 80|0); //@line 940 "modules/graphics/ParticleSystem.cpp"
  $229 = +HEAPF32[$228>>2]; //@line 940 "modules/graphics/ParticleSystem.cpp"
  $230 = $229 != -1.0; //@line 940 "modules/graphics/ParticleSystem.cpp"
  $231 = $227 < 0.0; //@line 940 "modules/graphics/ParticleSystem.cpp"
  $or$cond6 = $231 & $230; //@line 940 "modules/graphics/ParticleSystem.cpp"
  if ($or$cond6) {
   HEAP8[$207>>0] = 0; //@line 769 "modules/graphics/ParticleSystem.cpp"
   HEAPF32[$225>>2] = $229; //@line 770 "modules/graphics/ParticleSystem.cpp"
   HEAPF32[$213>>2] = 0.0; //@line 771 "modules/graphics/ParticleSystem.cpp"
  }
 }
 $232 = ((($this)) + 60|0); //@line 944 "modules/graphics/ParticleSystem.cpp"
 $233 = ((($this)) + 52|0); //@line 944 "modules/graphics/ParticleSystem.cpp"
 $234 = $233; //@line 944 "modules/graphics/ParticleSystem.cpp"
 $235 = $234; //@line 944 "modules/graphics/ParticleSystem.cpp"
 $236 = HEAP32[$235>>2]|0; //@line 944 "modules/graphics/ParticleSystem.cpp"
 $237 = (($234) + 4)|0; //@line 944 "modules/graphics/ParticleSystem.cpp"
 $238 = $237; //@line 944 "modules/graphics/ParticleSystem.cpp"
 $239 = HEAP32[$238>>2]|0; //@line 944 "modules/graphics/ParticleSystem.cpp"
 $240 = $232; //@line 944 "modules/graphics/ParticleSystem.cpp"
 $241 = $240; //@line 944 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$241>>2] = $236; //@line 944 "modules/graphics/ParticleSystem.cpp"
 $242 = (($240) + 4)|0; //@line 944 "modules/graphics/ParticleSystem.cpp"
 $243 = $242; //@line 944 "modules/graphics/ParticleSystem.cpp"
 HEAP32[$243>>2] = $239; //@line 944 "modules/graphics/ParticleSystem.cpp"
 return; //@line 945 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem11getConstantEPKcRNS1_22AreaSpreadDistributionE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 6)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((629100 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (629100 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(6); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((629100 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 949 "modules/graphics/ParticleSystem.cpp"
  return ($$$2$i|0); //@line 949 "modules/graphics/ParticleSystem.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 949 "modules/graphics/ParticleSystem.cpp"
  return ($$$2$i|0); //@line 949 "modules/graphics/ParticleSystem.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics14ParticleSystem11getConstantENS1_22AreaSpreadDistributionERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(2); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (629172 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 954 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics14ParticleSystem11getConstantEPKcRNS1_10InsertModeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 6)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((629184 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (629184 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(6); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((629184 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 959 "modules/graphics/ParticleSystem.cpp"
  return ($$$2$i|0); //@line 959 "modules/graphics/ParticleSystem.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 959 "modules/graphics/ParticleSystem.cpp"
  return ($$$2$i|0); //@line 959 "modules/graphics/ParticleSystem.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics14ParticleSystem11getConstantENS1_10InsertModeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(2); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (629256 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 964 "modules/graphics/ParticleSystem.cpp"
}
function __ZN4love8graphics4QuadD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 42 "modules/graphics/Quad.cpp"
}
function __ZN4love8graphics4QuadD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 41 "modules/graphics/Quad.cpp"
 return; //@line 42 "modules/graphics/Quad.cpp"
}
function __ZN4love8graphics4QuadC2ERKNS1_8ViewportEdd($this,$v,$sw,$sh) {
 $this = $this|0;
 $v = $v|0;
 $sw = +$sw;
 $sh = +$sh;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0.0, dest = 0;
 var label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 32 "modules/graphics/Quad.cpp"
 HEAP32[$this>>2] = (15636); //@line 35 "modules/graphics/Quad.cpp"
 $0 = ((($this)) + 120|0); //@line 33 "modules/graphics/Quad.cpp"
 HEAPF64[$0>>3] = $sw; //@line 33 "modules/graphics/Quad.cpp"
 $1 = ((($this)) + 128|0); //@line 34 "modules/graphics/Quad.cpp"
 HEAPF64[$1>>3] = $sh; //@line 34 "modules/graphics/Quad.cpp"
 $2 = ((($this)) + 8|0); //@line 36 "modules/graphics/Quad.cpp"
 dest=$2; stop=dest+80|0; do { HEAP32[dest>>2]=4294967295|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 36 "modules/graphics/Quad.cpp"
 $3 = ((($this)) + 88|0); //@line 46 "modules/graphics/Quad.cpp"
 ;HEAP32[$3>>2]=HEAP32[$v>>2]|0;HEAP32[$3+4>>2]=HEAP32[$v+4>>2]|0;HEAP32[$3+8>>2]=HEAP32[$v+8>>2]|0;HEAP32[$3+12>>2]=HEAP32[$v+12>>2]|0;HEAP32[$3+16>>2]=HEAP32[$v+16>>2]|0;HEAP32[$3+20>>2]=HEAP32[$v+20>>2]|0;HEAP32[$3+24>>2]=HEAP32[$v+24>>2]|0;HEAP32[$3+28>>2]=HEAP32[$v+28>>2]|0; //@line 46 "modules/graphics/Quad.cpp"
 $4 = ((($this)) + 8|0); //@line 53 "modules/graphics/Quad.cpp"
 HEAPF32[$4>>2] = 0.0; //@line 53 "modules/graphics/Quad.cpp"
 $5 = ((($this)) + 12|0); //@line 54 "modules/graphics/Quad.cpp"
 HEAPF32[$5>>2] = 0.0; //@line 54 "modules/graphics/Quad.cpp"
 $6 = ((($this)) + 28|0); //@line 55 "modules/graphics/Quad.cpp"
 HEAPF32[$6>>2] = 0.0; //@line 55 "modules/graphics/Quad.cpp"
 $7 = ((($v)) + 24|0); //@line 56 "modules/graphics/Quad.cpp"
 $8 = +HEAPF64[$7>>3]; //@line 56 "modules/graphics/Quad.cpp"
 $9 = $8; //@line 56 "modules/graphics/Quad.cpp"
 $10 = ((($this)) + 32|0); //@line 56 "modules/graphics/Quad.cpp"
 HEAPF32[$10>>2] = $9; //@line 56 "modules/graphics/Quad.cpp"
 $11 = ((($v)) + 16|0); //@line 57 "modules/graphics/Quad.cpp"
 $12 = +HEAPF64[$11>>3]; //@line 57 "modules/graphics/Quad.cpp"
 $13 = $12; //@line 57 "modules/graphics/Quad.cpp"
 $14 = ((($this)) + 48|0); //@line 57 "modules/graphics/Quad.cpp"
 HEAPF32[$14>>2] = $13; //@line 57 "modules/graphics/Quad.cpp"
 $15 = ((($this)) + 52|0); //@line 58 "modules/graphics/Quad.cpp"
 HEAPF32[$15>>2] = 0.0; //@line 58 "modules/graphics/Quad.cpp"
 $16 = ((($this)) + 68|0); //@line 59 "modules/graphics/Quad.cpp"
 HEAPF32[$16>>2] = $13; //@line 59 "modules/graphics/Quad.cpp"
 $17 = ((($this)) + 72|0); //@line 60 "modules/graphics/Quad.cpp"
 HEAPF32[$17>>2] = $9; //@line 60 "modules/graphics/Quad.cpp"
 $18 = +HEAPF64[$v>>3]; //@line 62 "modules/graphics/Quad.cpp"
 $19 = $18 / $sw; //@line 62 "modules/graphics/Quad.cpp"
 $20 = $19; //@line 62 "modules/graphics/Quad.cpp"
 $21 = ((($this)) + 16|0); //@line 62 "modules/graphics/Quad.cpp"
 HEAPF32[$21>>2] = $20; //@line 62 "modules/graphics/Quad.cpp"
 $22 = ((($v)) + 8|0); //@line 63 "modules/graphics/Quad.cpp"
 $23 = +HEAPF64[$22>>3]; //@line 63 "modules/graphics/Quad.cpp"
 $24 = $23 / $sh; //@line 63 "modules/graphics/Quad.cpp"
 $25 = $24; //@line 63 "modules/graphics/Quad.cpp"
 $26 = ((($this)) + 20|0); //@line 63 "modules/graphics/Quad.cpp"
 HEAPF32[$26>>2] = $25; //@line 63 "modules/graphics/Quad.cpp"
 $27 = ((($this)) + 36|0); //@line 64 "modules/graphics/Quad.cpp"
 HEAPF32[$27>>2] = $20; //@line 64 "modules/graphics/Quad.cpp"
 $28 = $8 + $23; //@line 65 "modules/graphics/Quad.cpp"
 $29 = $28 / $sh; //@line 65 "modules/graphics/Quad.cpp"
 $30 = $29; //@line 65 "modules/graphics/Quad.cpp"
 $31 = ((($this)) + 40|0); //@line 65 "modules/graphics/Quad.cpp"
 HEAPF32[$31>>2] = $30; //@line 65 "modules/graphics/Quad.cpp"
 $32 = $12 + $18; //@line 66 "modules/graphics/Quad.cpp"
 $33 = $32 / $sw; //@line 66 "modules/graphics/Quad.cpp"
 $34 = $33; //@line 66 "modules/graphics/Quad.cpp"
 $35 = ((($this)) + 56|0); //@line 66 "modules/graphics/Quad.cpp"
 HEAPF32[$35>>2] = $34; //@line 66 "modules/graphics/Quad.cpp"
 $36 = ((($this)) + 60|0); //@line 67 "modules/graphics/Quad.cpp"
 HEAPF32[$36>>2] = $25; //@line 67 "modules/graphics/Quad.cpp"
 $37 = ((($this)) + 76|0); //@line 68 "modules/graphics/Quad.cpp"
 HEAPF32[$37>>2] = $34; //@line 68 "modules/graphics/Quad.cpp"
 $38 = ((($this)) + 80|0); //@line 69 "modules/graphics/Quad.cpp"
 HEAPF32[$38>>2] = $30; //@line 69 "modules/graphics/Quad.cpp"
 return; //@line 38 "modules/graphics/Quad.cpp"
}
function __ZN4love8graphics4Quad7refreshERKNS1_8ViewportEdd($this,$v,$sw,$sh) {
 $this = $this|0;
 $v = $v|0;
 $sw = +$sw;
 $sh = +$sh;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0, $25 = 0.0, $26 = 0.0;
 var $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 88|0); //@line 46 "modules/graphics/Quad.cpp"
 ;HEAP32[$0>>2]=HEAP32[$v>>2]|0;HEAP32[$0+4>>2]=HEAP32[$v+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$v+8>>2]|0;HEAP32[$0+12>>2]=HEAP32[$v+12>>2]|0;HEAP32[$0+16>>2]=HEAP32[$v+16>>2]|0;HEAP32[$0+20>>2]=HEAP32[$v+20>>2]|0;HEAP32[$0+24>>2]=HEAP32[$v+24>>2]|0;HEAP32[$0+28>>2]=HEAP32[$v+28>>2]|0; //@line 46 "modules/graphics/Quad.cpp"
 $1 = ((($this)) + 8|0); //@line 53 "modules/graphics/Quad.cpp"
 HEAPF32[$1>>2] = 0.0; //@line 53 "modules/graphics/Quad.cpp"
 $2 = ((($this)) + 12|0); //@line 54 "modules/graphics/Quad.cpp"
 HEAPF32[$2>>2] = 0.0; //@line 54 "modules/graphics/Quad.cpp"
 $3 = ((($this)) + 28|0); //@line 55 "modules/graphics/Quad.cpp"
 HEAPF32[$3>>2] = 0.0; //@line 55 "modules/graphics/Quad.cpp"
 $4 = ((($v)) + 24|0); //@line 56 "modules/graphics/Quad.cpp"
 $5 = +HEAPF64[$4>>3]; //@line 56 "modules/graphics/Quad.cpp"
 $6 = $5; //@line 56 "modules/graphics/Quad.cpp"
 $7 = ((($this)) + 32|0); //@line 56 "modules/graphics/Quad.cpp"
 HEAPF32[$7>>2] = $6; //@line 56 "modules/graphics/Quad.cpp"
 $8 = ((($v)) + 16|0); //@line 57 "modules/graphics/Quad.cpp"
 $9 = +HEAPF64[$8>>3]; //@line 57 "modules/graphics/Quad.cpp"
 $10 = $9; //@line 57 "modules/graphics/Quad.cpp"
 $11 = ((($this)) + 48|0); //@line 57 "modules/graphics/Quad.cpp"
 HEAPF32[$11>>2] = $10; //@line 57 "modules/graphics/Quad.cpp"
 $12 = ((($this)) + 52|0); //@line 58 "modules/graphics/Quad.cpp"
 HEAPF32[$12>>2] = 0.0; //@line 58 "modules/graphics/Quad.cpp"
 $13 = ((($this)) + 68|0); //@line 59 "modules/graphics/Quad.cpp"
 HEAPF32[$13>>2] = $10; //@line 59 "modules/graphics/Quad.cpp"
 $14 = ((($this)) + 72|0); //@line 60 "modules/graphics/Quad.cpp"
 HEAPF32[$14>>2] = $6; //@line 60 "modules/graphics/Quad.cpp"
 $15 = +HEAPF64[$v>>3]; //@line 62 "modules/graphics/Quad.cpp"
 $16 = $15 / $sw; //@line 62 "modules/graphics/Quad.cpp"
 $17 = $16; //@line 62 "modules/graphics/Quad.cpp"
 $18 = ((($this)) + 16|0); //@line 62 "modules/graphics/Quad.cpp"
 HEAPF32[$18>>2] = $17; //@line 62 "modules/graphics/Quad.cpp"
 $19 = ((($v)) + 8|0); //@line 63 "modules/graphics/Quad.cpp"
 $20 = +HEAPF64[$19>>3]; //@line 63 "modules/graphics/Quad.cpp"
 $21 = $20 / $sh; //@line 63 "modules/graphics/Quad.cpp"
 $22 = $21; //@line 63 "modules/graphics/Quad.cpp"
 $23 = ((($this)) + 20|0); //@line 63 "modules/graphics/Quad.cpp"
 HEAPF32[$23>>2] = $22; //@line 63 "modules/graphics/Quad.cpp"
 $24 = ((($this)) + 36|0); //@line 64 "modules/graphics/Quad.cpp"
 HEAPF32[$24>>2] = $17; //@line 64 "modules/graphics/Quad.cpp"
 $25 = $20 + $5; //@line 65 "modules/graphics/Quad.cpp"
 $26 = $25 / $sh; //@line 65 "modules/graphics/Quad.cpp"
 $27 = $26; //@line 65 "modules/graphics/Quad.cpp"
 $28 = ((($this)) + 40|0); //@line 65 "modules/graphics/Quad.cpp"
 HEAPF32[$28>>2] = $27; //@line 65 "modules/graphics/Quad.cpp"
 $29 = $15 + $9; //@line 66 "modules/graphics/Quad.cpp"
 $30 = $29 / $sw; //@line 66 "modules/graphics/Quad.cpp"
 $31 = $30; //@line 66 "modules/graphics/Quad.cpp"
 $32 = ((($this)) + 56|0); //@line 66 "modules/graphics/Quad.cpp"
 HEAPF32[$32>>2] = $31; //@line 66 "modules/graphics/Quad.cpp"
 $33 = ((($this)) + 60|0); //@line 67 "modules/graphics/Quad.cpp"
 HEAPF32[$33>>2] = $22; //@line 67 "modules/graphics/Quad.cpp"
 $34 = ((($this)) + 76|0); //@line 68 "modules/graphics/Quad.cpp"
 HEAPF32[$34>>2] = $31; //@line 68 "modules/graphics/Quad.cpp"
 $35 = ((($this)) + 80|0); //@line 69 "modules/graphics/Quad.cpp"
 HEAPF32[$35>>2] = $27; //@line 69 "modules/graphics/Quad.cpp"
 return; //@line 70 "modules/graphics/Quad.cpp"
}
function __ZN4love8graphics4Quad11setViewportERKNS1_8ViewportE($this,$v) {
 $this = $this|0;
 $v = $v|0;
 var $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0.0;
 var $27 = 0, $28 = 0, $29 = 0.0, $3 = 0.0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 120|0); //@line 74 "modules/graphics/Quad.cpp"
 $1 = +HEAPF64[$0>>3]; //@line 74 "modules/graphics/Quad.cpp"
 $2 = ((($this)) + 128|0); //@line 74 "modules/graphics/Quad.cpp"
 $3 = +HEAPF64[$2>>3]; //@line 74 "modules/graphics/Quad.cpp"
 $4 = ((($this)) + 88|0); //@line 46 "modules/graphics/Quad.cpp"
 ;HEAP32[$4>>2]=HEAP32[$v>>2]|0;HEAP32[$4+4>>2]=HEAP32[$v+4>>2]|0;HEAP32[$4+8>>2]=HEAP32[$v+8>>2]|0;HEAP32[$4+12>>2]=HEAP32[$v+12>>2]|0;HEAP32[$4+16>>2]=HEAP32[$v+16>>2]|0;HEAP32[$4+20>>2]=HEAP32[$v+20>>2]|0;HEAP32[$4+24>>2]=HEAP32[$v+24>>2]|0;HEAP32[$4+28>>2]=HEAP32[$v+28>>2]|0; //@line 46 "modules/graphics/Quad.cpp"
 $5 = ((($this)) + 8|0); //@line 53 "modules/graphics/Quad.cpp"
 HEAPF32[$5>>2] = 0.0; //@line 53 "modules/graphics/Quad.cpp"
 $6 = ((($this)) + 12|0); //@line 54 "modules/graphics/Quad.cpp"
 HEAPF32[$6>>2] = 0.0; //@line 54 "modules/graphics/Quad.cpp"
 $7 = ((($this)) + 28|0); //@line 55 "modules/graphics/Quad.cpp"
 HEAPF32[$7>>2] = 0.0; //@line 55 "modules/graphics/Quad.cpp"
 $8 = ((($v)) + 24|0); //@line 56 "modules/graphics/Quad.cpp"
 $9 = +HEAPF64[$8>>3]; //@line 56 "modules/graphics/Quad.cpp"
 $10 = $9; //@line 56 "modules/graphics/Quad.cpp"
 $11 = ((($this)) + 32|0); //@line 56 "modules/graphics/Quad.cpp"
 HEAPF32[$11>>2] = $10; //@line 56 "modules/graphics/Quad.cpp"
 $12 = ((($v)) + 16|0); //@line 57 "modules/graphics/Quad.cpp"
 $13 = +HEAPF64[$12>>3]; //@line 57 "modules/graphics/Quad.cpp"
 $14 = $13; //@line 57 "modules/graphics/Quad.cpp"
 $15 = ((($this)) + 48|0); //@line 57 "modules/graphics/Quad.cpp"
 HEAPF32[$15>>2] = $14; //@line 57 "modules/graphics/Quad.cpp"
 $16 = ((($this)) + 52|0); //@line 58 "modules/graphics/Quad.cpp"
 HEAPF32[$16>>2] = 0.0; //@line 58 "modules/graphics/Quad.cpp"
 $17 = ((($this)) + 68|0); //@line 59 "modules/graphics/Quad.cpp"
 HEAPF32[$17>>2] = $14; //@line 59 "modules/graphics/Quad.cpp"
 $18 = ((($this)) + 72|0); //@line 60 "modules/graphics/Quad.cpp"
 HEAPF32[$18>>2] = $10; //@line 60 "modules/graphics/Quad.cpp"
 $19 = +HEAPF64[$v>>3]; //@line 62 "modules/graphics/Quad.cpp"
 $20 = $19 / $1; //@line 62 "modules/graphics/Quad.cpp"
 $21 = $20; //@line 62 "modules/graphics/Quad.cpp"
 $22 = ((($this)) + 16|0); //@line 62 "modules/graphics/Quad.cpp"
 HEAPF32[$22>>2] = $21; //@line 62 "modules/graphics/Quad.cpp"
 $23 = ((($v)) + 8|0); //@line 63 "modules/graphics/Quad.cpp"
 $24 = +HEAPF64[$23>>3]; //@line 63 "modules/graphics/Quad.cpp"
 $25 = $24 / $3; //@line 63 "modules/graphics/Quad.cpp"
 $26 = $25; //@line 63 "modules/graphics/Quad.cpp"
 $27 = ((($this)) + 20|0); //@line 63 "modules/graphics/Quad.cpp"
 HEAPF32[$27>>2] = $26; //@line 63 "modules/graphics/Quad.cpp"
 $28 = ((($this)) + 36|0); //@line 64 "modules/graphics/Quad.cpp"
 HEAPF32[$28>>2] = $21; //@line 64 "modules/graphics/Quad.cpp"
 $29 = $9 + $24; //@line 65 "modules/graphics/Quad.cpp"
 $30 = $29 / $3; //@line 65 "modules/graphics/Quad.cpp"
 $31 = $30; //@line 65 "modules/graphics/Quad.cpp"
 $32 = ((($this)) + 40|0); //@line 65 "modules/graphics/Quad.cpp"
 HEAPF32[$32>>2] = $31; //@line 65 "modules/graphics/Quad.cpp"
 $33 = $13 + $19; //@line 66 "modules/graphics/Quad.cpp"
 $34 = $33 / $1; //@line 66 "modules/graphics/Quad.cpp"
 $35 = $34; //@line 66 "modules/graphics/Quad.cpp"
 $36 = ((($this)) + 56|0); //@line 66 "modules/graphics/Quad.cpp"
 HEAPF32[$36>>2] = $35; //@line 66 "modules/graphics/Quad.cpp"
 $37 = ((($this)) + 60|0); //@line 67 "modules/graphics/Quad.cpp"
 HEAPF32[$37>>2] = $26; //@line 67 "modules/graphics/Quad.cpp"
 $38 = ((($this)) + 76|0); //@line 68 "modules/graphics/Quad.cpp"
 HEAPF32[$38>>2] = $35; //@line 68 "modules/graphics/Quad.cpp"
 $39 = ((($this)) + 80|0); //@line 69 "modules/graphics/Quad.cpp"
 HEAPF32[$39>>2] = $31; //@line 69 "modules/graphics/Quad.cpp"
 return; //@line 75 "modules/graphics/Quad.cpp"
}
function __ZNK4love8graphics4Quad11getViewportEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 88|0); //@line 79 "modules/graphics/Quad.cpp"
 ;HEAP32[$agg$result>>2]=HEAP32[$0>>2]|0;HEAP32[$agg$result+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$agg$result+8>>2]=HEAP32[$0+8>>2]|0;HEAP32[$agg$result+12>>2]=HEAP32[$0+12>>2]|0;HEAP32[$agg$result+16>>2]=HEAP32[$0+16>>2]|0;HEAP32[$agg$result+20>>2]=HEAP32[$0+20>>2]|0;HEAP32[$agg$result+24>>2]=HEAP32[$0+24>>2]|0;HEAP32[$agg$result+28>>2]=HEAP32[$0+28>>2]|0; //@line 79 "modules/graphics/Quad.cpp"
 return; //@line 79 "modules/graphics/Quad.cpp"
}
function __ZNK4love8graphics4Quad11getVerticesEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 84 "modules/graphics/Quad.cpp"
 return ($0|0); //@line 84 "modules/graphics/Quad.cpp"
}
function __ZN4love8graphics7TextureD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 41 "modules/graphics/Texture.cpp"
}
function __ZN4love8graphics7TextureD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 40 "modules/graphics/Texture.cpp"
 return; //@line 41 "modules/graphics/Texture.cpp"
}
function __ZNK4love8graphics7Texture8getWidthEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 45 "modules/graphics/Texture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 45 "modules/graphics/Texture.cpp"
 return ($1|0); //@line 45 "modules/graphics/Texture.cpp"
}
function __ZNK4love8graphics7Texture9getHeightEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 50 "modules/graphics/Texture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 50 "modules/graphics/Texture.cpp"
 return ($1|0); //@line 50 "modules/graphics/Texture.cpp"
}
function __ZNK4love8graphics7Texture9getFilterEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 55 "modules/graphics/Texture.cpp"
 return ($0|0); //@line 55 "modules/graphics/Texture.cpp"
}
function __ZNK4love8graphics7Texture7getWrapEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 60 "modules/graphics/Texture.cpp"
 return ($0|0); //@line 60 "modules/graphics/Texture.cpp"
}
function __ZNK4love8graphics7Texture11getVerticesEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0); //@line 65 "modules/graphics/Texture.cpp"
 return ($0|0); //@line 65 "modules/graphics/Texture.cpp"
}
function __GLOBAL__sub_I_Texture_cpp() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[(629276)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629288)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629300)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629312)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629324)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629336)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(629340)>>2]=0|0;HEAP32[(629340)+4>>2]=0|0;HEAP32[(629340)+8>>2]=0|0; //@line 44 "./common/StringMap.h"
 $0 = HEAP32[3933]|0; //@line 49 "./common/StringMap.h"
 $1 = HEAP32[(15736)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics7Texture10FilterModeELj3EE3addEPKcS3_(629268,$0,$1)|0); //@line 49 "./common/StringMap.h"
 $2 = HEAP32[(15740)>>2]|0; //@line 49 "./common/StringMap.h"
 $3 = HEAP32[(15744)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics7Texture10FilterModeELj3EE3addEPKcS3_(629268,$2,$3)|0); //@line 49 "./common/StringMap.h"
 $4 = HEAP32[(15748)>>2]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(15752)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics7Texture10FilterModeELj3EE3addEPKcS3_(629268,$4,$5)|0); //@line 49 "./common/StringMap.h"
 HEAP8[(629360)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629372)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629384)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629396)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629408)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629420)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629432)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(629444)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(629448)>>2]=0|0;HEAP32[(629448)+4>>2]=0|0;HEAP32[(629448)+8>>2]=0|0;HEAP32[(629448)+12>>2]=0|0; //@line 44 "./common/StringMap.h"
 $6 = HEAP32[3939]|0; //@line 49 "./common/StringMap.h"
 $7 = HEAP32[(15760)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics7Texture8WrapModeELj4EE3addEPKcS3_(629352,$6,$7)|0); //@line 49 "./common/StringMap.h"
 $8 = HEAP32[(15764)>>2]|0; //@line 49 "./common/StringMap.h"
 $9 = HEAP32[(15768)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics7Texture8WrapModeELj4EE3addEPKcS3_(629352,$8,$9)|0); //@line 49 "./common/StringMap.h"
 $10 = HEAP32[(15772)>>2]|0; //@line 49 "./common/StringMap.h"
 $11 = HEAP32[(15776)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics7Texture8WrapModeELj4EE3addEPKcS3_(629352,$10,$11)|0); //@line 49 "./common/StringMap.h"
 $12 = HEAP32[(15780)>>2]|0; //@line 49 "./common/StringMap.h"
 $13 = HEAP32[(15784)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics7Texture8WrapModeELj4EE3addEPKcS3_(629352,$12,$13)|0); //@line 49 "./common/StringMap.h"
 return;
}
function __ZN4love9StringMapINS_8graphics7Texture10FilterModeELj3EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $inserted$2$off0 = 0;
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(2); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 72|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love9StringMapINS_8graphics7Texture8WrapModeELj4EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0;
 var $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 7; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = $13 & 7; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = $22 & 7; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = $27 & 7; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = $32 & 7; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = $37 & 7; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $42 = (($hash$0$lcssa$i) + 6)|0; //@line 112 "./common/StringMap.h"
       $43 = $42 & 7; //@line 112 "./common/StringMap.h"
       $44 = (((($this) + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
       $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
       $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
       if ($46) {
        $$lcssa = $43;$$lcssa10 = $44;
        label = 5;
       } else {
        $47 = (($hash$0$lcssa$i) + 7)|0; //@line 112 "./common/StringMap.h"
        $48 = $47 & 7; //@line 112 "./common/StringMap.h"
        $49 = (((($this) + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
        $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
        $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
        if ($51) {
         $$lcssa = $48;$$lcssa10 = $49;
         label = 5;
        } else {
         $inserted$2$off0 = 0;
        }
       }
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(3); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 96|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love8graphics7TextureC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 36 "modules/graphics/Drawable.h"
 HEAP32[$this>>2] = (15676); //@line 36 "modules/graphics/Texture.cpp"
 $0 = ((($this)) + 8|0); //@line 31 "modules/graphics/Texture.cpp"
 HEAP32[$0>>2] = 0; //@line 31 "modules/graphics/Texture.cpp"
 $1 = ((($this)) + 12|0); //@line 32 "modules/graphics/Texture.cpp"
 HEAP32[$1>>2] = 0; //@line 32 "modules/graphics/Texture.cpp"
 $2 = ((($this)) + 16|0); //@line 33 "modules/graphics/Texture.cpp"
 ;HEAP32[$2>>2]=HEAP32[15652>>2]|0;HEAP32[$2+4>>2]=HEAP32[15652+4>>2]|0;HEAP32[$2+8>>2]=HEAP32[15652+8>>2]|0;HEAP32[$2+12>>2]=HEAP32[15652+12>>2]|0; //@line 33 "modules/graphics/Texture.cpp"
 $3 = ((($this)) + 32|0); //@line 70 "modules/graphics/Texture.h"
 dest=$3; stop=dest+88|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 70 "modules/graphics/Texture.h"
 return; //@line 37 "modules/graphics/Texture.cpp"
}
function __ZN4love8graphics7Texture16getDefaultFilterEv() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (15652|0); //@line 75 "modules/graphics/Texture.cpp"
}
function __ZN4love8graphics7Texture16setDefaultFilterERKNS1_6FilterE($f) {
 $f = $f|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[15652>>2]=HEAP32[$f>>2]|0;HEAP32[15652+4>>2]=HEAP32[$f+4>>2]|0;HEAP32[15652+8>>2]=HEAP32[$f+8>>2]|0;HEAP32[15652+12>>2]=HEAP32[$f+12>>2]|0; //@line 70 "modules/graphics/Texture.cpp"
 return; //@line 71 "modules/graphics/Texture.cpp"
}
function __ZN4love8graphics7Texture14validateFilterERKNS1_6FilterEb($f,$mipmapsAllowed) {
 $f = $f|0;
 $mipmapsAllowed = $mipmapsAllowed|0;
 var $$off = 0, $$off1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $switch = 0, $switch2 = 0, $switch3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 if (!($mipmapsAllowed)) {
  $0 = ((($f)) + 8|0); //@line 80 "modules/graphics/Texture.cpp"
  $1 = HEAP32[$0>>2]|0; //@line 80 "modules/graphics/Texture.cpp"
  $2 = ($1|0)==(0); //@line 80 "modules/graphics/Texture.cpp"
  if (!($2)) {
   return 0; //@line 93 "modules/graphics/Texture.cpp"
  }
 }
 $3 = ((($f)) + 4|0); //@line 83 "modules/graphics/Texture.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 83 "modules/graphics/Texture.cpp"
 $$off = (($4) + -1)|0; //@line 83 "modules/graphics/Texture.cpp"
 $switch = ($$off>>>0)<(2); //@line 83 "modules/graphics/Texture.cpp"
 if (!($switch)) {
  return 0; //@line 93 "modules/graphics/Texture.cpp"
 }
 $5 = HEAP32[$f>>2]|0; //@line 86 "modules/graphics/Texture.cpp"
 $$off1 = (($5) + -1)|0; //@line 86 "modules/graphics/Texture.cpp"
 $switch2 = ($$off1>>>0)<(2); //@line 86 "modules/graphics/Texture.cpp"
 if ($switch2) {
  $6 = ((($f)) + 8|0); //@line 89 "modules/graphics/Texture.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 89 "modules/graphics/Texture.cpp"
  $switch3 = ($7>>>0)<(3); //@line 89 "modules/graphics/Texture.cpp"
  return ($switch3|0); //@line 89 "modules/graphics/Texture.cpp"
 } else {
  return 0; //@line 93 "modules/graphics/Texture.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics7Texture11getConstantEPKcRNS1_10FilterModeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 6)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((629268 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (629268 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(6); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((629268 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 97 "modules/graphics/Texture.cpp"
  return ($$$2$i|0); //@line 97 "modules/graphics/Texture.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 97 "modules/graphics/Texture.cpp"
  return ($$$2$i|0); //@line 97 "modules/graphics/Texture.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics7Texture11getConstantENS1_10FilterModeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(2); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (629340 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 102 "modules/graphics/Texture.cpp"
}
function __ZN4love8graphics7Texture11getConstantEPKcRNS1_8WrapModeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = $9 & 7; //@line 72 "./common/StringMap.h"
  $11 = (((629352 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (629352 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(8); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((629352 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 107 "modules/graphics/Texture.cpp"
  return ($$$2$i|0); //@line 107 "modules/graphics/Texture.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 107 "modules/graphics/Texture.cpp"
  return ($$$2$i|0); //@line 107 "modules/graphics/Texture.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics7Texture11getConstantENS1_8WrapModeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(3); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (629448 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 112 "modules/graphics/Texture.cpp"
}
function __ZN4love8graphics18w_Quad_setViewportEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0.0, $2 = 0.0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics4QuadEEEPT_P9lua_StateiNS_4TypeE($L,1,13)|0); //@line 31 "modules/graphics/wrap_Quad.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 39 "modules/graphics/wrap_Quad.cpp"
 HEAPF64[$v>>3] = $1; //@line 39 "modules/graphics/wrap_Quad.cpp"
 $2 = (+_luaL_checknumber($L,3)); //@line 40 "modules/graphics/wrap_Quad.cpp"
 $3 = ((($v)) + 8|0); //@line 40 "modules/graphics/wrap_Quad.cpp"
 HEAPF64[$3>>3] = $2; //@line 40 "modules/graphics/wrap_Quad.cpp"
 $4 = (+_luaL_checknumber($L,4)); //@line 41 "modules/graphics/wrap_Quad.cpp"
 $5 = ((($v)) + 16|0); //@line 41 "modules/graphics/wrap_Quad.cpp"
 HEAPF64[$5>>3] = $4; //@line 41 "modules/graphics/wrap_Quad.cpp"
 $6 = (+_luaL_checknumber($L,5)); //@line 42 "modules/graphics/wrap_Quad.cpp"
 $7 = ((($v)) + 24|0); //@line 42 "modules/graphics/wrap_Quad.cpp"
 HEAPF64[$7>>3] = $6; //@line 42 "modules/graphics/wrap_Quad.cpp"
 $8 = (_lua_type($L,6)|0); //@line 44 "modules/graphics/wrap_Quad.cpp"
 $9 = ($8|0)<(1); //@line 44 "modules/graphics/wrap_Quad.cpp"
 if ($9) {
  __ZN4love8graphics4Quad11setViewportERKNS1_8ViewportE($0,$v); //@line 45 "modules/graphics/wrap_Quad.cpp"
  STACKTOP = sp;return 0; //@line 53 "modules/graphics/wrap_Quad.cpp"
 } else {
  $10 = (+_luaL_checknumber($L,6)); //@line 48 "modules/graphics/wrap_Quad.cpp"
  $11 = (+_luaL_checknumber($L,7)); //@line 49 "modules/graphics/wrap_Quad.cpp"
  __ZN4love8graphics4Quad7refreshERKNS1_8ViewportEdd($0,$v,$10,$11); //@line 50 "modules/graphics/wrap_Quad.cpp"
  STACKTOP = sp;return 0; //@line 53 "modules/graphics/wrap_Quad.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics18w_Quad_getViewportEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics4QuadEEEPT_P9lua_StateiNS_4TypeE($L,1,13)|0); //@line 31 "modules/graphics/wrap_Quad.cpp"
 __ZNK4love8graphics4Quad11getViewportEv($v,$0); //@line 59 "modules/graphics/wrap_Quad.cpp"
 $1 = +HEAPF64[$v>>3]; //@line 60 "modules/graphics/wrap_Quad.cpp"
 _lua_pushnumber($L,$1); //@line 60 "modules/graphics/wrap_Quad.cpp"
 $2 = ((($v)) + 8|0); //@line 61 "modules/graphics/wrap_Quad.cpp"
 $3 = +HEAPF64[$2>>3]; //@line 61 "modules/graphics/wrap_Quad.cpp"
 _lua_pushnumber($L,$3); //@line 61 "modules/graphics/wrap_Quad.cpp"
 $4 = ((($v)) + 16|0); //@line 62 "modules/graphics/wrap_Quad.cpp"
 $5 = +HEAPF64[$4>>3]; //@line 62 "modules/graphics/wrap_Quad.cpp"
 _lua_pushnumber($L,$5); //@line 62 "modules/graphics/wrap_Quad.cpp"
 $6 = ((($v)) + 24|0); //@line 63 "modules/graphics/wrap_Quad.cpp"
 $7 = +HEAPF64[$6>>3]; //@line 63 "modules/graphics/wrap_Quad.cpp"
 _lua_pushnumber($L,$7); //@line 63 "modules/graphics/wrap_Quad.cpp"
 STACKTOP = sp;return 4; //@line 64 "modules/graphics/wrap_Quad.cpp"
}
function _luaopen_quad($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 74 "modules/graphics/wrap_Quad.cpp"
 HEAP32[$vararg_buffer>>2] = 15788; //@line 76 "modules/graphics/wrap_Quad.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 76 "modules/graphics/wrap_Quad.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 76 "modules/graphics/wrap_Quad.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,13,347523,$vararg_buffer)|0); //@line 76 "modules/graphics/wrap_Quad.cpp"
 STACKTOP = sp;return ($0|0); //@line 76 "modules/graphics/wrap_Quad.cpp"
}
function __ZN4love8graphics18w_Texture_getWidthEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics7TextureEEEPT_P9lua_StateiNS_4TypeE($L,1,11)|0); //@line 30 "modules/graphics/wrap_Texture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 36 "modules/graphics/wrap_Texture.cpp"
 $2 = ((($1)) + 24|0); //@line 36 "modules/graphics/wrap_Texture.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 36 "modules/graphics/wrap_Texture.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 36 "modules/graphics/wrap_Texture.cpp"
 $5 = (+($4|0)); //@line 36 "modules/graphics/wrap_Texture.cpp"
 _lua_pushnumber($L,$5); //@line 36 "modules/graphics/wrap_Texture.cpp"
 return 1; //@line 37 "modules/graphics/wrap_Texture.cpp"
}
function __ZN4love14luax_checktypeINS_8graphics7TextureEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love8graphics19w_Texture_getHeightEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics7TextureEEEPT_P9lua_StateiNS_4TypeE($L,1,11)|0); //@line 30 "modules/graphics/wrap_Texture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 43 "modules/graphics/wrap_Texture.cpp"
 $2 = ((($1)) + 28|0); //@line 43 "modules/graphics/wrap_Texture.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 43 "modules/graphics/wrap_Texture.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 43 "modules/graphics/wrap_Texture.cpp"
 $5 = (+($4|0)); //@line 43 "modules/graphics/wrap_Texture.cpp"
 _lua_pushnumber($L,$5); //@line 43 "modules/graphics/wrap_Texture.cpp"
 return 1; //@line 44 "modules/graphics/wrap_Texture.cpp"
}
function __ZN4love8graphics23w_Texture_getDimensionsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics7TextureEEEPT_P9lua_StateiNS_4TypeE($L,1,11)|0); //@line 30 "modules/graphics/wrap_Texture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 50 "modules/graphics/wrap_Texture.cpp"
 $2 = ((($1)) + 24|0); //@line 50 "modules/graphics/wrap_Texture.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 50 "modules/graphics/wrap_Texture.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 50 "modules/graphics/wrap_Texture.cpp"
 $5 = (+($4|0)); //@line 50 "modules/graphics/wrap_Texture.cpp"
 _lua_pushnumber($L,$5); //@line 50 "modules/graphics/wrap_Texture.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 51 "modules/graphics/wrap_Texture.cpp"
 $7 = ((($6)) + 28|0); //@line 51 "modules/graphics/wrap_Texture.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 51 "modules/graphics/wrap_Texture.cpp"
 $9 = (FUNCTION_TABLE_ii[$8 & 8191]($0)|0); //@line 51 "modules/graphics/wrap_Texture.cpp"
 $10 = (+($9|0)); //@line 51 "modules/graphics/wrap_Texture.cpp"
 _lua_pushnumber($L,$10); //@line 51 "modules/graphics/wrap_Texture.cpp"
 return 2; //@line 52 "modules/graphics/wrap_Texture.cpp"
}
function __ZN4love8graphics19w_Texture_setFilterEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $f = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $f = sp + 24|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics7TextureEEEPT_P9lua_StateiNS_4TypeE($L,1,11)|0); //@line 30 "modules/graphics/wrap_Texture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 58 "modules/graphics/wrap_Texture.cpp"
 $2 = ((($1)) + 36|0); //@line 58 "modules/graphics/wrap_Texture.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 58 "modules/graphics/wrap_Texture.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 58 "modules/graphics/wrap_Texture.cpp"
 ;HEAP32[$f>>2]=HEAP32[$4>>2]|0;HEAP32[$f+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$f+8>>2]=HEAP32[$4+8>>2]|0;HEAP32[$f+12>>2]=HEAP32[$4+12>>2]|0; //@line 58 "modules/graphics/wrap_Texture.cpp"
 $5 = (_luaL_checklstring($L,2,0)|0); //@line 60 "modules/graphics/wrap_Texture.cpp"
 $6 = (_luaL_optlstring($L,3,$5,0)|0); //@line 61 "modules/graphics/wrap_Texture.cpp"
 $7 = (__ZN4love8graphics7Texture11getConstantEPKcRNS1_10FilterModeE($5,$f)|0); //@line 63 "modules/graphics/wrap_Texture.cpp"
 if (!($7)) {
  HEAP32[$vararg_buffer>>2] = $5; //@line 64 "modules/graphics/wrap_Texture.cpp"
  $8 = (_luaL_error($L,439370,$vararg_buffer)|0); //@line 64 "modules/graphics/wrap_Texture.cpp"
  $$0 = $8;
  STACKTOP = sp;return ($$0|0); //@line 72 "modules/graphics/wrap_Texture.cpp"
 }
 $9 = ((($f)) + 4|0); //@line 65 "modules/graphics/wrap_Texture.cpp"
 $10 = (__ZN4love8graphics7Texture11getConstantEPKcRNS1_10FilterModeE($6,$9)|0); //@line 65 "modules/graphics/wrap_Texture.cpp"
 if (!($10)) {
  HEAP32[$vararg_buffer1>>2] = $6; //@line 66 "modules/graphics/wrap_Texture.cpp"
  $11 = (_luaL_error($L,439370,$vararg_buffer1)|0); //@line 66 "modules/graphics/wrap_Texture.cpp"
  $$0 = $11;
  STACKTOP = sp;return ($$0|0); //@line 72 "modules/graphics/wrap_Texture.cpp"
 }
 $12 = (+_luaL_optnumber($L,4,1.0)); //@line 68 "modules/graphics/wrap_Texture.cpp"
 $13 = $12; //@line 68 "modules/graphics/wrap_Texture.cpp"
 $14 = ((($f)) + 12|0); //@line 68 "modules/graphics/wrap_Texture.cpp"
 HEAPF32[$14>>2] = $13; //@line 68 "modules/graphics/wrap_Texture.cpp"
 $15 = HEAP32[$0>>2]|0; //@line 70 "modules/graphics/wrap_Texture.cpp"
 $16 = ((($15)) + 32|0); //@line 70 "modules/graphics/wrap_Texture.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 70 "modules/graphics/wrap_Texture.cpp"
 __THREW__ = 0;
 invoke_vii($17|0,($0|0),($f|0)); //@line 70 "modules/graphics/wrap_Texture.cpp"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if (!($19)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 72 "modules/graphics/wrap_Texture.cpp"
 }
 $20 = ___cxa_find_matching_catch_3(8992|0)|0;
 $21 = tempRet0;
 $22 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $23 = ($21|0)==($22|0); //@line 535 "./common/runtime.h"
 if (!($23)) {
  $$01$i = $21;$$02$i = $20;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $24 = (___cxa_begin_catch(($20|0))|0); //@line 535 "./common/runtime.h"
 $25 = HEAP32[$24>>2]|0; //@line 539 "./common/runtime.h"
 $26 = ((($25)) + 8|0); //@line 539 "./common/runtime.h"
 $27 = HEAP32[$26>>2]|0; //@line 539 "./common/runtime.h"
 $28 = (FUNCTION_TABLE_ii[$27 & 8191]($24)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($28|0)); //@line 539 "./common/runtime.h"
 $29 = __THREW__; __THREW__ = 0;
 $30 = $29&1;
 if (!($30)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $31 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer4>>2] = $31; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer4)|0); //@line 543 "./common/runtime.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 72 "modules/graphics/wrap_Texture.cpp"
 }
 $32 = ___cxa_find_matching_catch_2()|0;
 $33 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $34 = __THREW__; __THREW__ = 0;
 $35 = $34&1;
 if ($35) {
  $36 = ___cxa_find_matching_catch_3(0|0)|0;
  $37 = tempRet0;
  ___clang_call_terminate($36); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $33;$$02$i = $32;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics19w_Texture_getFilterEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, $magstr = 0, $minstr = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $minstr = sp + 16|0;
 $magstr = sp + 12|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics7TextureEEEPT_P9lua_StateiNS_4TypeE($L,1,11)|0); //@line 30 "modules/graphics/wrap_Texture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 77 "modules/graphics/wrap_Texture.cpp"
 $2 = ((($1)) + 36|0); //@line 77 "modules/graphics/wrap_Texture.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 77 "modules/graphics/wrap_Texture.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 77 "modules/graphics/wrap_Texture.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 77 "modules/graphics/wrap_Texture.cpp"
 $6 = ((($4)) + 4|0); //@line 77 "modules/graphics/wrap_Texture.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 77 "modules/graphics/wrap_Texture.cpp"
 $8 = ((($4)) + 12|0); //@line 77 "modules/graphics/wrap_Texture.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 77 "modules/graphics/wrap_Texture.cpp"
 HEAP32[$minstr>>2] = 0; //@line 79 "modules/graphics/wrap_Texture.cpp"
 HEAP32[$magstr>>2] = 0; //@line 80 "modules/graphics/wrap_Texture.cpp"
 $10 = (__ZN4love8graphics7Texture11getConstantENS1_10FilterModeERPKc($5,$minstr)|0); //@line 82 "modules/graphics/wrap_Texture.cpp"
 if (!($10)) {
  $11 = (_luaL_error($L,347552,$vararg_buffer)|0); //@line 83 "modules/graphics/wrap_Texture.cpp"
  $$0 = $11;
  STACKTOP = sp;return ($$0|0); //@line 91 "modules/graphics/wrap_Texture.cpp"
 }
 $12 = (__ZN4love8graphics7Texture11getConstantENS1_10FilterModeERPKc($7,$magstr)|0); //@line 84 "modules/graphics/wrap_Texture.cpp"
 if ($12) {
  $14 = HEAP32[$minstr>>2]|0; //@line 87 "modules/graphics/wrap_Texture.cpp"
  _lua_pushstring($L,$14); //@line 87 "modules/graphics/wrap_Texture.cpp"
  $15 = HEAP32[$magstr>>2]|0; //@line 88 "modules/graphics/wrap_Texture.cpp"
  _lua_pushstring($L,$15); //@line 88 "modules/graphics/wrap_Texture.cpp"
  $16 = $9; //@line 89 "modules/graphics/wrap_Texture.cpp"
  _lua_pushnumber($L,$16); //@line 89 "modules/graphics/wrap_Texture.cpp"
  $$0 = 3;
  STACKTOP = sp;return ($$0|0); //@line 91 "modules/graphics/wrap_Texture.cpp"
 } else {
  $13 = (_luaL_error($L,347552,$vararg_buffer1)|0); //@line 85 "modules/graphics/wrap_Texture.cpp"
  $$0 = $13;
  STACKTOP = sp;return ($$0|0); //@line 91 "modules/graphics/wrap_Texture.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics17w_Texture_setWrapEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $w = sp + 16|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics7TextureEEEPT_P9lua_StateiNS_4TypeE($L,1,11)|0); //@line 30 "modules/graphics/wrap_Texture.cpp"
 HEAP32[$w>>2] = 0; //@line 70 "modules/graphics/Texture.h"
 $1 = ((($w)) + 4|0); //@line 71 "modules/graphics/Texture.h"
 HEAP32[$1>>2] = 0; //@line 71 "modules/graphics/Texture.h"
 $2 = (_luaL_checklstring($L,2,0)|0); //@line 98 "modules/graphics/wrap_Texture.cpp"
 $3 = (_luaL_optlstring($L,3,$2,0)|0); //@line 99 "modules/graphics/wrap_Texture.cpp"
 $4 = (__ZN4love8graphics7Texture11getConstantEPKcRNS1_8WrapModeE($2,$w)|0); //@line 101 "modules/graphics/wrap_Texture.cpp"
 if (!($4)) {
  HEAP32[$vararg_buffer>>2] = $2; //@line 102 "modules/graphics/wrap_Texture.cpp"
  $5 = (_luaL_error($L,347581,$vararg_buffer)|0); //@line 102 "modules/graphics/wrap_Texture.cpp"
  $$0 = $5;
  STACKTOP = sp;return ($$0|0); //@line 108 "modules/graphics/wrap_Texture.cpp"
 }
 $6 = (__ZN4love8graphics7Texture11getConstantEPKcRNS1_8WrapModeE($3,$1)|0); //@line 103 "modules/graphics/wrap_Texture.cpp"
 if ($6) {
  $8 = HEAP32[$0>>2]|0; //@line 106 "modules/graphics/wrap_Texture.cpp"
  $9 = ((($8)) + 40|0); //@line 106 "modules/graphics/wrap_Texture.cpp"
  $10 = HEAP32[$9>>2]|0; //@line 106 "modules/graphics/wrap_Texture.cpp"
  $11 = (FUNCTION_TABLE_iii[$10 & 8191]($0,$w)|0); //@line 106 "modules/graphics/wrap_Texture.cpp"
  __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$11); //@line 106 "modules/graphics/wrap_Texture.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 108 "modules/graphics/wrap_Texture.cpp"
 } else {
  HEAP32[$vararg_buffer1>>2] = $3; //@line 104 "modules/graphics/wrap_Texture.cpp"
  $7 = (_luaL_error($L,347603,$vararg_buffer1)|0); //@line 104 "modules/graphics/wrap_Texture.cpp"
  $$0 = $7;
  STACKTOP = sp;return ($$0|0); //@line 108 "modules/graphics/wrap_Texture.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics17w_Texture_getWrapEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $sstr = 0, $tstr = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $sstr = sp + 16|0;
 $tstr = sp + 12|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics7TextureEEEPT_P9lua_StateiNS_4TypeE($L,1,11)|0); //@line 30 "modules/graphics/wrap_Texture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 113 "modules/graphics/wrap_Texture.cpp"
 $2 = ((($1)) + 44|0); //@line 113 "modules/graphics/wrap_Texture.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 113 "modules/graphics/wrap_Texture.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 113 "modules/graphics/wrap_Texture.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 113 "modules/graphics/wrap_Texture.cpp"
 $6 = ((($4)) + 4|0); //@line 113 "modules/graphics/wrap_Texture.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 113 "modules/graphics/wrap_Texture.cpp"
 HEAP32[$sstr>>2] = 0; //@line 115 "modules/graphics/wrap_Texture.cpp"
 HEAP32[$tstr>>2] = 0; //@line 116 "modules/graphics/wrap_Texture.cpp"
 $8 = (__ZN4love8graphics7Texture11getConstantENS1_8WrapModeERPKc($5,$sstr)|0); //@line 118 "modules/graphics/wrap_Texture.cpp"
 if (!($8)) {
  $9 = (_luaL_error($L,347625,$vararg_buffer)|0); //@line 119 "modules/graphics/wrap_Texture.cpp"
  $$0 = $9;
  STACKTOP = sp;return ($$0|0); //@line 126 "modules/graphics/wrap_Texture.cpp"
 }
 $10 = (__ZN4love8graphics7Texture11getConstantENS1_8WrapModeERPKc($7,$tstr)|0); //@line 120 "modules/graphics/wrap_Texture.cpp"
 if ($10) {
  $12 = HEAP32[$sstr>>2]|0; //@line 123 "modules/graphics/wrap_Texture.cpp"
  _lua_pushstring($L,$12); //@line 123 "modules/graphics/wrap_Texture.cpp"
  $13 = HEAP32[$tstr>>2]|0; //@line 124 "modules/graphics/wrap_Texture.cpp"
  _lua_pushstring($L,$13); //@line 124 "modules/graphics/wrap_Texture.cpp"
  $$0 = 2;
  STACKTOP = sp;return ($$0|0); //@line 126 "modules/graphics/wrap_Texture.cpp"
 } else {
  $11 = (_luaL_error($L,347625,$vararg_buffer1)|0); //@line 121 "modules/graphics/wrap_Texture.cpp"
  $$0 = $11;
  STACKTOP = sp;return ($$0|0); //@line 126 "modules/graphics/wrap_Texture.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics17luax_checktextureEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics7TextureEEEPT_P9lua_StateiNS_4TypeE($L,$idx,11)|0); //@line 30 "modules/graphics/wrap_Texture.cpp"
 return ($0|0); //@line 30 "modules/graphics/wrap_Texture.cpp"
}
function _luaopen_texture($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 140 "modules/graphics/wrap_Texture.cpp"
 HEAP32[$vararg_buffer>>2] = 15812; //@line 142 "modules/graphics/wrap_Texture.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 142 "modules/graphics/wrap_Texture.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 142 "modules/graphics/wrap_Texture.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,11,347644,$vararg_buffer)|0); //@line 142 "modules/graphics/wrap_Texture.cpp"
 STACKTOP = sp;return ($0|0); //@line 142 "modules/graphics/wrap_Texture.cpp"
}
function __ZN4love5image19CompressedImageDataD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (15884); //@line 37 "modules/image/CompressedImageData.cpp"
 $0 = ((($this)) + 24|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ($1|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = $1; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($2) {
  return; //@line 38 "modules/image/CompressedImageData.cpp"
 }
 $4 = ((($this)) + 28|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5|0)==($1|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($6)) {
  $scevgep$i$i$i = ((($5)) + -16|0);
  $7 = $scevgep$i$i$i;
  $8 = (($7) - ($3))|0;
  $9 = $8 >>> 4;
  $10 = $9 ^ -1;
  $scevgep4$i$i$i = (($5) + ($10<<4)|0);
  HEAP32[$4>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($1); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 38 "modules/image/CompressedImageData.cpp"
}
function __ZN4love5image19CompressedImageDataD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i$i = 0, $scevgep4$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (15884); //@line 37 "modules/image/CompressedImageData.cpp"
 $0 = ((($this)) + 24|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ($1|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = $1; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($2) {
  __ZdlPv($this); //@line 37 "modules/image/CompressedImageData.cpp"
  return; //@line 38 "modules/image/CompressedImageData.cpp"
 }
 $4 = ((($this)) + 28|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5|0)==($1|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($6)) {
  $scevgep$i$i$i$i = ((($5)) + -16|0);
  $7 = $scevgep$i$i$i$i;
  $8 = (($7) - ($3))|0;
  $9 = $8 >>> 4;
  $10 = $9 ^ -1;
  $scevgep4$i$i$i$i = (($5) + ($10<<4)|0);
  HEAP32[$4>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($1); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 __ZdlPv($this); //@line 37 "modules/image/CompressedImageData.cpp"
 return; //@line 38 "modules/image/CompressedImageData.cpp"
}
function __ZNK4love5image19CompressedImageData7getDataEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 47 "modules/image/CompressedImageData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 47 "modules/image/CompressedImageData.cpp"
 return ($1|0); //@line 47 "modules/image/CompressedImageData.cpp"
}
function __ZNK4love5image19CompressedImageData7getSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 42 "modules/image/CompressedImageData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 42 "modules/image/CompressedImageData.cpp"
 return ($1|0); //@line 42 "modules/image/CompressedImageData.cpp"
}
function __GLOBAL__sub_I_CompressedImageData_cpp() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love9StringMapINS_5image19CompressedImageData6FormatELj37EEC2EPKNS4_5EntryEj(629464,15908,296); //@line 150 "modules/image/CompressedImageData.cpp"
 return;
}
function __ZN4love9StringMapINS_5image19CompressedImageData6FormatELj37EEC2EPKNS4_5EntryEj($this,$entries,$num) {
 $this = $this|0;
 $entries = $entries|0;
 $num = $num|0;
 var $$02$i$i = 0, $$lcssa26 = 0, $$lcssa28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$08$i = 0, $i1$05 = 0, $scevgep = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 8|0); //@line 155 "./common/StringMap.h"
 HEAP8[$0>>0] = 0; //@line 155 "./common/StringMap.h"
 $1 = ((($this)) + 20|0); //@line 155 "./common/StringMap.h"
 HEAP8[$1>>0] = 0; //@line 155 "./common/StringMap.h"
 $2 = ((($this)) + 32|0); //@line 155 "./common/StringMap.h"
 HEAP8[$2>>0] = 0; //@line 155 "./common/StringMap.h"
 $3 = ((($this)) + 44|0); //@line 155 "./common/StringMap.h"
 HEAP8[$3>>0] = 0; //@line 155 "./common/StringMap.h"
 $4 = ((($this)) + 56|0); //@line 155 "./common/StringMap.h"
 HEAP8[$4>>0] = 0; //@line 155 "./common/StringMap.h"
 $5 = ((($this)) + 68|0); //@line 155 "./common/StringMap.h"
 HEAP8[$5>>0] = 0; //@line 155 "./common/StringMap.h"
 $6 = ((($this)) + 80|0); //@line 155 "./common/StringMap.h"
 HEAP8[$6>>0] = 0; //@line 155 "./common/StringMap.h"
 $7 = ((($this)) + 92|0); //@line 155 "./common/StringMap.h"
 HEAP8[$7>>0] = 0; //@line 155 "./common/StringMap.h"
 $8 = ((($this)) + 104|0); //@line 155 "./common/StringMap.h"
 HEAP8[$8>>0] = 0; //@line 155 "./common/StringMap.h"
 $9 = ((($this)) + 116|0); //@line 155 "./common/StringMap.h"
 HEAP8[$9>>0] = 0; //@line 155 "./common/StringMap.h"
 $10 = ((($this)) + 128|0); //@line 155 "./common/StringMap.h"
 HEAP8[$10>>0] = 0; //@line 155 "./common/StringMap.h"
 $11 = ((($this)) + 140|0); //@line 155 "./common/StringMap.h"
 HEAP8[$11>>0] = 0; //@line 155 "./common/StringMap.h"
 $12 = ((($this)) + 152|0); //@line 155 "./common/StringMap.h"
 HEAP8[$12>>0] = 0; //@line 155 "./common/StringMap.h"
 $13 = ((($this)) + 164|0); //@line 155 "./common/StringMap.h"
 HEAP8[$13>>0] = 0; //@line 155 "./common/StringMap.h"
 $14 = ((($this)) + 176|0); //@line 155 "./common/StringMap.h"
 HEAP8[$14>>0] = 0; //@line 155 "./common/StringMap.h"
 $15 = ((($this)) + 188|0); //@line 155 "./common/StringMap.h"
 HEAP8[$15>>0] = 0; //@line 155 "./common/StringMap.h"
 $16 = ((($this)) + 200|0); //@line 155 "./common/StringMap.h"
 HEAP8[$16>>0] = 0; //@line 155 "./common/StringMap.h"
 $17 = ((($this)) + 212|0); //@line 155 "./common/StringMap.h"
 HEAP8[$17>>0] = 0; //@line 155 "./common/StringMap.h"
 $18 = ((($this)) + 224|0); //@line 155 "./common/StringMap.h"
 HEAP8[$18>>0] = 0; //@line 155 "./common/StringMap.h"
 $19 = ((($this)) + 236|0); //@line 155 "./common/StringMap.h"
 HEAP8[$19>>0] = 0; //@line 155 "./common/StringMap.h"
 $20 = ((($this)) + 248|0); //@line 155 "./common/StringMap.h"
 HEAP8[$20>>0] = 0; //@line 155 "./common/StringMap.h"
 $21 = ((($this)) + 260|0); //@line 155 "./common/StringMap.h"
 HEAP8[$21>>0] = 0; //@line 155 "./common/StringMap.h"
 $22 = ((($this)) + 272|0); //@line 155 "./common/StringMap.h"
 HEAP8[$22>>0] = 0; //@line 155 "./common/StringMap.h"
 $23 = ((($this)) + 284|0); //@line 155 "./common/StringMap.h"
 HEAP8[$23>>0] = 0; //@line 155 "./common/StringMap.h"
 $24 = ((($this)) + 296|0); //@line 155 "./common/StringMap.h"
 HEAP8[$24>>0] = 0; //@line 155 "./common/StringMap.h"
 $25 = ((($this)) + 308|0); //@line 155 "./common/StringMap.h"
 HEAP8[$25>>0] = 0; //@line 155 "./common/StringMap.h"
 $26 = ((($this)) + 320|0); //@line 155 "./common/StringMap.h"
 HEAP8[$26>>0] = 0; //@line 155 "./common/StringMap.h"
 $27 = ((($this)) + 332|0); //@line 155 "./common/StringMap.h"
 HEAP8[$27>>0] = 0; //@line 155 "./common/StringMap.h"
 $28 = ((($this)) + 344|0); //@line 155 "./common/StringMap.h"
 HEAP8[$28>>0] = 0; //@line 155 "./common/StringMap.h"
 $29 = ((($this)) + 356|0); //@line 155 "./common/StringMap.h"
 HEAP8[$29>>0] = 0; //@line 155 "./common/StringMap.h"
 $30 = ((($this)) + 368|0); //@line 155 "./common/StringMap.h"
 HEAP8[$30>>0] = 0; //@line 155 "./common/StringMap.h"
 $31 = ((($this)) + 380|0); //@line 155 "./common/StringMap.h"
 HEAP8[$31>>0] = 0; //@line 155 "./common/StringMap.h"
 $32 = ((($this)) + 392|0); //@line 155 "./common/StringMap.h"
 HEAP8[$32>>0] = 0; //@line 155 "./common/StringMap.h"
 $33 = ((($this)) + 404|0); //@line 155 "./common/StringMap.h"
 HEAP8[$33>>0] = 0; //@line 155 "./common/StringMap.h"
 $34 = ((($this)) + 416|0); //@line 155 "./common/StringMap.h"
 HEAP8[$34>>0] = 0; //@line 155 "./common/StringMap.h"
 $35 = ((($this)) + 428|0); //@line 155 "./common/StringMap.h"
 HEAP8[$35>>0] = 0; //@line 155 "./common/StringMap.h"
 $36 = ((($this)) + 440|0); //@line 155 "./common/StringMap.h"
 HEAP8[$36>>0] = 0; //@line 155 "./common/StringMap.h"
 $37 = ((($this)) + 452|0); //@line 155 "./common/StringMap.h"
 HEAP8[$37>>0] = 0; //@line 155 "./common/StringMap.h"
 $38 = ((($this)) + 464|0); //@line 155 "./common/StringMap.h"
 HEAP8[$38>>0] = 0; //@line 155 "./common/StringMap.h"
 $39 = ((($this)) + 476|0); //@line 155 "./common/StringMap.h"
 HEAP8[$39>>0] = 0; //@line 155 "./common/StringMap.h"
 $40 = ((($this)) + 488|0); //@line 155 "./common/StringMap.h"
 HEAP8[$40>>0] = 0; //@line 155 "./common/StringMap.h"
 $41 = ((($this)) + 500|0); //@line 155 "./common/StringMap.h"
 HEAP8[$41>>0] = 0; //@line 155 "./common/StringMap.h"
 $42 = ((($this)) + 512|0); //@line 155 "./common/StringMap.h"
 HEAP8[$42>>0] = 0; //@line 155 "./common/StringMap.h"
 $43 = ((($this)) + 524|0); //@line 155 "./common/StringMap.h"
 HEAP8[$43>>0] = 0; //@line 155 "./common/StringMap.h"
 $44 = ((($this)) + 536|0); //@line 155 "./common/StringMap.h"
 HEAP8[$44>>0] = 0; //@line 155 "./common/StringMap.h"
 $45 = ((($this)) + 548|0); //@line 155 "./common/StringMap.h"
 HEAP8[$45>>0] = 0; //@line 155 "./common/StringMap.h"
 $46 = ((($this)) + 560|0); //@line 155 "./common/StringMap.h"
 HEAP8[$46>>0] = 0; //@line 155 "./common/StringMap.h"
 $47 = ((($this)) + 572|0); //@line 155 "./common/StringMap.h"
 HEAP8[$47>>0] = 0; //@line 155 "./common/StringMap.h"
 $48 = ((($this)) + 584|0); //@line 155 "./common/StringMap.h"
 HEAP8[$48>>0] = 0; //@line 155 "./common/StringMap.h"
 $49 = ((($this)) + 596|0); //@line 155 "./common/StringMap.h"
 HEAP8[$49>>0] = 0; //@line 155 "./common/StringMap.h"
 $50 = ((($this)) + 608|0); //@line 155 "./common/StringMap.h"
 HEAP8[$50>>0] = 0; //@line 155 "./common/StringMap.h"
 $51 = ((($this)) + 620|0); //@line 155 "./common/StringMap.h"
 HEAP8[$51>>0] = 0; //@line 155 "./common/StringMap.h"
 $52 = ((($this)) + 632|0); //@line 155 "./common/StringMap.h"
 HEAP8[$52>>0] = 0; //@line 155 "./common/StringMap.h"
 $53 = ((($this)) + 644|0); //@line 155 "./common/StringMap.h"
 HEAP8[$53>>0] = 0; //@line 155 "./common/StringMap.h"
 $54 = ((($this)) + 656|0); //@line 155 "./common/StringMap.h"
 HEAP8[$54>>0] = 0; //@line 155 "./common/StringMap.h"
 $55 = ((($this)) + 668|0); //@line 155 "./common/StringMap.h"
 HEAP8[$55>>0] = 0; //@line 155 "./common/StringMap.h"
 $56 = ((($this)) + 680|0); //@line 155 "./common/StringMap.h"
 HEAP8[$56>>0] = 0; //@line 155 "./common/StringMap.h"
 $57 = ((($this)) + 692|0); //@line 155 "./common/StringMap.h"
 HEAP8[$57>>0] = 0; //@line 155 "./common/StringMap.h"
 $58 = ((($this)) + 704|0); //@line 155 "./common/StringMap.h"
 HEAP8[$58>>0] = 0; //@line 155 "./common/StringMap.h"
 $59 = ((($this)) + 716|0); //@line 155 "./common/StringMap.h"
 HEAP8[$59>>0] = 0; //@line 155 "./common/StringMap.h"
 $60 = ((($this)) + 728|0); //@line 155 "./common/StringMap.h"
 HEAP8[$60>>0] = 0; //@line 155 "./common/StringMap.h"
 $61 = ((($this)) + 740|0); //@line 155 "./common/StringMap.h"
 HEAP8[$61>>0] = 0; //@line 155 "./common/StringMap.h"
 $62 = ((($this)) + 752|0); //@line 155 "./common/StringMap.h"
 HEAP8[$62>>0] = 0; //@line 155 "./common/StringMap.h"
 $63 = ((($this)) + 764|0); //@line 155 "./common/StringMap.h"
 HEAP8[$63>>0] = 0; //@line 155 "./common/StringMap.h"
 $64 = ((($this)) + 776|0); //@line 155 "./common/StringMap.h"
 HEAP8[$64>>0] = 0; //@line 155 "./common/StringMap.h"
 $65 = ((($this)) + 788|0); //@line 155 "./common/StringMap.h"
 HEAP8[$65>>0] = 0; //@line 155 "./common/StringMap.h"
 $66 = ((($this)) + 800|0); //@line 155 "./common/StringMap.h"
 HEAP8[$66>>0] = 0; //@line 155 "./common/StringMap.h"
 $67 = ((($this)) + 812|0); //@line 155 "./common/StringMap.h"
 HEAP8[$67>>0] = 0; //@line 155 "./common/StringMap.h"
 $68 = ((($this)) + 824|0); //@line 155 "./common/StringMap.h"
 HEAP8[$68>>0] = 0; //@line 155 "./common/StringMap.h"
 $69 = ((($this)) + 836|0); //@line 155 "./common/StringMap.h"
 HEAP8[$69>>0] = 0; //@line 155 "./common/StringMap.h"
 $70 = ((($this)) + 848|0); //@line 155 "./common/StringMap.h"
 HEAP8[$70>>0] = 0; //@line 155 "./common/StringMap.h"
 $71 = ((($this)) + 860|0); //@line 155 "./common/StringMap.h"
 HEAP8[$71>>0] = 0; //@line 155 "./common/StringMap.h"
 $72 = ((($this)) + 872|0); //@line 155 "./common/StringMap.h"
 HEAP8[$72>>0] = 0; //@line 155 "./common/StringMap.h"
 $73 = ((($this)) + 884|0); //@line 155 "./common/StringMap.h"
 HEAP8[$73>>0] = 0; //@line 155 "./common/StringMap.h"
 $scevgep = ((($this)) + 888|0); //@line 44 "./common/StringMap.h"
 _memset(($scevgep|0),0,148)|0; //@line 44 "./common/StringMap.h"
 $74 = $num >>> 3; //@line 46 "./common/StringMap.h"
 $75 = ($74|0)==(0); //@line 48 "./common/StringMap.h"
 if ($75) {
  STACKTOP = sp;return; //@line 50 "./common/StringMap.h"
 } else {
  $i1$05 = 0;
 }
 while(1) {
  $76 = (($entries) + ($i1$05<<3)|0); //@line 49 "./common/StringMap.h"
  $77 = HEAP32[$76>>2]|0; //@line 49 "./common/StringMap.h"
  $78 = (((($entries) + ($i1$05<<3)|0)) + 4|0); //@line 49 "./common/StringMap.h"
  $79 = HEAP32[$78>>2]|0; //@line 49 "./common/StringMap.h"
  $80 = HEAP8[$77>>0]|0; //@line 142 "./common/StringMap.h"
  $81 = ($80<<24>>24)==(0); //@line 142 "./common/StringMap.h"
  if ($81) {
   $hash$0$lcssa$i$i = 5381;
  } else {
   $$02$i$i = $77;$83 = $80;$hash$01$i$i = 5381;
   while(1) {
    $82 = $83 << 24 >> 24; //@line 142 "./common/StringMap.h"
    $84 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
    $85 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
    $86 = (($85) + ($82))|0; //@line 143 "./common/StringMap.h"
    $87 = HEAP8[$84>>0]|0; //@line 142 "./common/StringMap.h"
    $88 = ($87<<24>>24)==(0); //@line 142 "./common/StringMap.h"
    if ($88) {
     $hash$0$lcssa$i$i = $86;
     break;
    } else {
     $$02$i$i = $84;$83 = $87;$hash$01$i$i = $86;
    }
   }
  }
  $i$08$i = 0;
  while(1) {
   $91 = (($i$08$i) + ($hash$0$lcssa$i$i))|0; //@line 112 "./common/StringMap.h"
   $92 = (($91>>>0) % 74)&-1; //@line 112 "./common/StringMap.h"
   $93 = (((($this) + (($92*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $94 = HEAP8[$93>>0]|0; //@line 114 "./common/StringMap.h"
   $95 = ($94<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   $90 = (($i$08$i) + 1)|0; //@line 110 "./common/StringMap.h"
   if ($95) {
    $$lcssa26 = $92;$$lcssa28 = $93;
    label = 8;
    break;
   }
   $89 = ($90>>>0)<(74); //@line 110 "./common/StringMap.h"
   if ($89) {
    $i$08$i = $90;
   } else {
    break;
   }
  }
  if ((label|0) == 8) {
   label = 0;
   HEAP8[$$lcssa28>>0] = 1; //@line 117 "./common/StringMap.h"
   $96 = (($this) + (($$lcssa26*12)|0)|0); //@line 118 "./common/StringMap.h"
   HEAP32[$96>>2] = $77; //@line 118 "./common/StringMap.h"
   $97 = (((($this) + (($$lcssa26*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
   HEAP32[$97>>2] = $79; //@line 119 "./common/StringMap.h"
  }
  $98 = ($79>>>0)>(36); //@line 126 "./common/StringMap.h"
  if ($98) {
   HEAP32[$vararg_buffer>>2] = $77; //@line 128 "./common/StringMap.h"
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
   HEAP32[$vararg_ptr1>>2] = $79; //@line 128 "./common/StringMap.h"
   (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  } else {
   $99 = (((($this)) + 888|0) + ($79<<2)|0); //@line 132 "./common/StringMap.h"
   HEAP32[$99>>2] = $77; //@line 132 "./common/StringMap.h"
  }
  $100 = (($i1$05) + 1)|0; //@line 48 "./common/StringMap.h"
  $exitcond = ($100|0)==($74|0); //@line 48 "./common/StringMap.h"
  if ($exitcond) {
   break;
  } else {
   $i1$05 = $100;
  }
 }
 STACKTOP = sp;return; //@line 50 "./common/StringMap.h"
}
function __ZN4love5image19CompressedImageDataC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 37 "./common/Data.h"
 HEAP32[$this>>2] = (15884); //@line 33 "modules/image/CompressedImageData.cpp"
 $0 = ((($this)) + 8|0); //@line 29 "modules/image/CompressedImageData.cpp"
 HEAP32[$0>>2] = 0; //@line 29 "modules/image/CompressedImageData.cpp"
 $1 = ((($this)) + 12|0); //@line 30 "modules/image/CompressedImageData.cpp"
 HEAP8[$1>>0] = 0; //@line 30 "modules/image/CompressedImageData.cpp"
 $2 = ((($this)) + 16|0); //@line 31 "modules/image/CompressedImageData.cpp"
 ;HEAP32[$2>>2]=0|0;HEAP32[$2+4>>2]=0|0;HEAP32[$2+8>>2]=0|0;HEAP32[$2+12>>2]=0|0;HEAP32[$2+16>>2]=0|0; //@line 31 "modules/image/CompressedImageData.cpp"
 return; //@line 34 "modules/image/CompressedImageData.cpp"
}
function __ZNK4love5image19CompressedImageData14getMipmapCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 52 "modules/image/CompressedImageData.cpp"
 $1 = ((($this)) + 28|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($2) - ($3))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = $4 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 return ($5|0); //@line 52 "modules/image/CompressedImageData.cpp"
}
function __ZNK4love5image19CompressedImageData7getSizeEi($this,$miplevel) {
 $this = $this|0;
 $miplevel = $miplevel|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($miplevel|0)<(0); //@line 95 "modules/image/CompressedImageData.cpp"
 if (!($0)) {
  $1 = ((($this)) + 24|0); //@line 95 "modules/image/CompressedImageData.cpp"
  $2 = ((($this)) + 28|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $6 = $5 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $7 = ($6|0)>($miplevel|0); //@line 95 "modules/image/CompressedImageData.cpp"
  if ($7) {
   $14 = $4; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $15 = (((($14) + ($miplevel<<4)|0)) + 8|0); //@line 59 "modules/image/CompressedImageData.cpp"
   $16 = HEAP32[$15>>2]|0; //@line 59 "modules/image/CompressedImageData.cpp"
   STACKTOP = sp;return ($16|0); //@line 59 "modules/image/CompressedImageData.cpp"
  }
 }
 $8 = (___cxa_allocate_exception(16)|0); //@line 96 "modules/image/CompressedImageData.cpp"
 $9 = (($miplevel) + 1)|0; //@line 96 "modules/image/CompressedImageData.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $9; //@line 96 "modules/image/CompressedImageData.cpp"
 invoke_viii(3349,($8|0),(347950|0),($vararg_buffer|0)); //@line 96 "modules/image/CompressedImageData.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $12 = ___cxa_find_matching_catch_2()|0;
  $13 = tempRet0;
  ___cxa_free_exception(($8|0)); //@line 96 "modules/image/CompressedImageData.cpp"
  ___resumeException($12|0);
  // unreachable;
 } else {
  ___cxa_throw(($8|0),(96|0),(344|0)); //@line 96 "modules/image/CompressedImageData.cpp"
  // unreachable; //@line 96 "modules/image/CompressedImageData.cpp"
 }
 return (0)|0;
}
function __ZNK4love5image19CompressedImageData7getDataEi($this,$miplevel) {
 $this = $this|0;
 $miplevel = $miplevel|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($miplevel|0)<(0); //@line 95 "modules/image/CompressedImageData.cpp"
 if (!($0)) {
  $1 = ((($this)) + 24|0); //@line 95 "modules/image/CompressedImageData.cpp"
  $2 = ((($this)) + 28|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $6 = $5 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $7 = ($6|0)>($miplevel|0); //@line 95 "modules/image/CompressedImageData.cpp"
  if ($7) {
   $14 = $4; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $15 = (((($14) + ($miplevel<<4)|0)) + 12|0); //@line 66 "modules/image/CompressedImageData.cpp"
   $16 = HEAP32[$15>>2]|0; //@line 66 "modules/image/CompressedImageData.cpp"
   STACKTOP = sp;return ($16|0); //@line 66 "modules/image/CompressedImageData.cpp"
  }
 }
 $8 = (___cxa_allocate_exception(16)|0); //@line 96 "modules/image/CompressedImageData.cpp"
 $9 = (($miplevel) + 1)|0; //@line 96 "modules/image/CompressedImageData.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $9; //@line 96 "modules/image/CompressedImageData.cpp"
 invoke_viii(3349,($8|0),(347950|0),($vararg_buffer|0)); //@line 96 "modules/image/CompressedImageData.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $12 = ___cxa_find_matching_catch_2()|0;
  $13 = tempRet0;
  ___cxa_free_exception(($8|0)); //@line 96 "modules/image/CompressedImageData.cpp"
  ___resumeException($12|0);
  // unreachable;
 } else {
  ___cxa_throw(($8|0),(96|0),(344|0)); //@line 96 "modules/image/CompressedImageData.cpp"
  // unreachable; //@line 96 "modules/image/CompressedImageData.cpp"
 }
 return (0)|0;
}
function __ZNK4love5image19CompressedImageData8getWidthEi($this,$miplevel) {
 $this = $this|0;
 $miplevel = $miplevel|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($miplevel|0)<(0); //@line 95 "modules/image/CompressedImageData.cpp"
 if (!($0)) {
  $1 = ((($this)) + 24|0); //@line 95 "modules/image/CompressedImageData.cpp"
  $2 = ((($this)) + 28|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $6 = $5 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $7 = ($6|0)>($miplevel|0); //@line 95 "modules/image/CompressedImageData.cpp"
  if ($7) {
   $14 = $4; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $15 = (($14) + ($miplevel<<4)|0); //@line 73 "modules/image/CompressedImageData.cpp"
   $16 = HEAP32[$15>>2]|0; //@line 73 "modules/image/CompressedImageData.cpp"
   STACKTOP = sp;return ($16|0); //@line 73 "modules/image/CompressedImageData.cpp"
  }
 }
 $8 = (___cxa_allocate_exception(16)|0); //@line 96 "modules/image/CompressedImageData.cpp"
 $9 = (($miplevel) + 1)|0; //@line 96 "modules/image/CompressedImageData.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $9; //@line 96 "modules/image/CompressedImageData.cpp"
 invoke_viii(3349,($8|0),(347950|0),($vararg_buffer|0)); //@line 96 "modules/image/CompressedImageData.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $12 = ___cxa_find_matching_catch_2()|0;
  $13 = tempRet0;
  ___cxa_free_exception(($8|0)); //@line 96 "modules/image/CompressedImageData.cpp"
  ___resumeException($12|0);
  // unreachable;
 } else {
  ___cxa_throw(($8|0),(96|0),(344|0)); //@line 96 "modules/image/CompressedImageData.cpp"
  // unreachable; //@line 96 "modules/image/CompressedImageData.cpp"
 }
 return (0)|0;
}
function __ZNK4love5image19CompressedImageData9getHeightEi($this,$miplevel) {
 $this = $this|0;
 $miplevel = $miplevel|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($miplevel|0)<(0); //@line 95 "modules/image/CompressedImageData.cpp"
 if (!($0)) {
  $1 = ((($this)) + 24|0); //@line 95 "modules/image/CompressedImageData.cpp"
  $2 = ((($this)) + 28|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $6 = $5 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $7 = ($6|0)>($miplevel|0); //@line 95 "modules/image/CompressedImageData.cpp"
  if ($7) {
   $14 = $4; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $15 = (((($14) + ($miplevel<<4)|0)) + 4|0); //@line 80 "modules/image/CompressedImageData.cpp"
   $16 = HEAP32[$15>>2]|0; //@line 80 "modules/image/CompressedImageData.cpp"
   STACKTOP = sp;return ($16|0); //@line 80 "modules/image/CompressedImageData.cpp"
  }
 }
 $8 = (___cxa_allocate_exception(16)|0); //@line 96 "modules/image/CompressedImageData.cpp"
 $9 = (($miplevel) + 1)|0; //@line 96 "modules/image/CompressedImageData.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $9; //@line 96 "modules/image/CompressedImageData.cpp"
 invoke_viii(3349,($8|0),(347950|0),($vararg_buffer|0)); //@line 96 "modules/image/CompressedImageData.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $12 = ___cxa_find_matching_catch_2()|0;
  $13 = tempRet0;
  ___cxa_free_exception(($8|0)); //@line 96 "modules/image/CompressedImageData.cpp"
  ___resumeException($12|0);
  // unreachable;
 } else {
  ___cxa_throw(($8|0),(96|0),(344|0)); //@line 96 "modules/image/CompressedImageData.cpp"
  // unreachable; //@line 96 "modules/image/CompressedImageData.cpp"
 }
 return (0)|0;
}
function __ZNK4love5image19CompressedImageData9getFormatEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 85 "modules/image/CompressedImageData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 85 "modules/image/CompressedImageData.cpp"
 return ($1|0); //@line 85 "modules/image/CompressedImageData.cpp"
}
function __ZN4love5image19CompressedImageData11getConstantENS1_6FormatERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(36); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (630352 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 106 "modules/image/CompressedImageData.cpp"
}
function __ZN4love5image29luax_checkcompressedimagedataEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5image19CompressedImageDataEEEPT_P9lua_StateiNS_4TypeE($L,$idx,23)|0); //@line 31 "modules/image/wrap_CompressedImageData.cpp"
 return ($0|0); //@line 31 "modules/image/wrap_CompressedImageData.cpp"
}
function __ZN4love14luax_checktypeINS_5image19CompressedImageDataEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love5image30w_CompressedImageData_getWidthEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $width$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 34 "modules/image/wrap_CompressedImageData.cpp"
 $0 = (__ZN4love14luax_checktypeINS_5image19CompressedImageDataEEEPT_P9lua_StateiNS_4TypeE($L,1,23)|0); //@line 31 "modules/image/wrap_CompressedImageData.cpp"
 $1 = (+_luaL_optnumber($L,2,1.0)); //@line 37 "modules/image/wrap_CompressedImageData.cpp"
 $2 = (~~(($1))); //@line 37 "modules/image/wrap_CompressedImageData.cpp"
 $3 = (($2) + -1)|0; //@line 40 "modules/image/wrap_CompressedImageData.cpp"
 __THREW__ = 0;
 $4 = (invoke_iii(3635,($0|0),($3|0))|0); //@line 40 "modules/image/wrap_CompressedImageData.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  $width$0 = $4;
  _lua_pushinteger($L,$width$0); //@line 42 "modules/image/wrap_CompressedImageData.cpp"
  STACKTOP = sp;return 1; //@line 43 "modules/image/wrap_CompressedImageData.cpp"
 }
 $7 = ___cxa_find_matching_catch_3(8992|0)|0;
 $8 = tempRet0;
 $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
 if (!($10)) {
  $$01$i = $8;$$02$i = $7;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
 $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
 $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if (!($17)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $18 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $18; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $width$0 = 0;
  _lua_pushinteger($L,$width$0); //@line 42 "modules/image/wrap_CompressedImageData.cpp"
  STACKTOP = sp;return 1; //@line 43 "modules/image/wrap_CompressedImageData.cpp"
 }
 $19 = ___cxa_find_matching_catch_2()|0;
 $20 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $23 = ___cxa_find_matching_catch_3(0|0)|0;
  $24 = tempRet0;
  ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $20;$$02$i = $19;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5image31w_CompressedImageData_getHeightEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $height$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 46 "modules/image/wrap_CompressedImageData.cpp"
 $0 = (__ZN4love14luax_checktypeINS_5image19CompressedImageDataEEEPT_P9lua_StateiNS_4TypeE($L,1,23)|0); //@line 31 "modules/image/wrap_CompressedImageData.cpp"
 $1 = (+_luaL_optnumber($L,2,1.0)); //@line 49 "modules/image/wrap_CompressedImageData.cpp"
 $2 = (~~(($1))); //@line 49 "modules/image/wrap_CompressedImageData.cpp"
 $3 = (($2) + -1)|0; //@line 52 "modules/image/wrap_CompressedImageData.cpp"
 __THREW__ = 0;
 $4 = (invoke_iii(3636,($0|0),($3|0))|0); //@line 52 "modules/image/wrap_CompressedImageData.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  $height$0 = $4;
  _lua_pushinteger($L,$height$0); //@line 54 "modules/image/wrap_CompressedImageData.cpp"
  STACKTOP = sp;return 1; //@line 55 "modules/image/wrap_CompressedImageData.cpp"
 }
 $7 = ___cxa_find_matching_catch_3(8992|0)|0;
 $8 = tempRet0;
 $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
 if (!($10)) {
  $$01$i = $8;$$02$i = $7;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
 $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
 $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if (!($17)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $18 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $18; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $height$0 = 0;
  _lua_pushinteger($L,$height$0); //@line 54 "modules/image/wrap_CompressedImageData.cpp"
  STACKTOP = sp;return 1; //@line 55 "modules/image/wrap_CompressedImageData.cpp"
 }
 $19 = ___cxa_find_matching_catch_2()|0;
 $20 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $23 = ___cxa_find_matching_catch_3(0|0)|0;
  $24 = tempRet0;
  ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $20;$$02$i = $19;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5image35w_CompressedImageData_getDimensionsEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $height$0 = 0, $vararg_buffer = 0, $width$0 = 0, $width$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 58 "modules/image/wrap_CompressedImageData.cpp"
 $0 = (__ZN4love14luax_checktypeINS_5image19CompressedImageDataEEEPT_P9lua_StateiNS_4TypeE($L,1,23)|0); //@line 31 "modules/image/wrap_CompressedImageData.cpp"
 $1 = (+_luaL_optnumber($L,2,1.0)); //@line 61 "modules/image/wrap_CompressedImageData.cpp"
 $2 = (~~(($1))); //@line 61 "modules/image/wrap_CompressedImageData.cpp"
 $3 = (($2) + -1)|0; //@line 66 "modules/image/wrap_CompressedImageData.cpp"
 __THREW__ = 0;
 $4 = (invoke_iii(3635,($0|0),($3|0))|0); //@line 66 "modules/image/wrap_CompressedImageData.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $width$0 = 0;
 } else {
  __THREW__ = 0;
  $7 = (invoke_iii(3636,($0|0),($3|0))|0); //@line 67 "modules/image/wrap_CompressedImageData.cpp"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if ($9) {
   $width$0 = $4;
  } else {
   $height$0 = $7;$width$1 = $4;
   _lua_pushinteger($L,$width$1); //@line 70 "modules/image/wrap_CompressedImageData.cpp"
   _lua_pushinteger($L,$height$0); //@line 71 "modules/image/wrap_CompressedImageData.cpp"
   STACKTOP = sp;return 2; //@line 72 "modules/image/wrap_CompressedImageData.cpp"
  }
 }
 $10 = ___cxa_find_matching_catch_3(8992|0)|0;
 $11 = tempRet0;
 $12 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $13 = ($11|0)==($12|0); //@line 535 "./common/runtime.h"
 if (!($13)) {
  $$01$i = $11;$$02$i = $10;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $14 = (___cxa_begin_catch(($10|0))|0); //@line 535 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = ((($15)) + 8|0); //@line 539 "./common/runtime.h"
 $17 = HEAP32[$16>>2]|0; //@line 539 "./common/runtime.h"
 $18 = (FUNCTION_TABLE_ii[$17 & 8191]($14)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($18|0)); //@line 539 "./common/runtime.h"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if (!($20)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $21 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $21; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $height$0 = 0;$width$1 = $width$0;
  _lua_pushinteger($L,$width$1); //@line 70 "modules/image/wrap_CompressedImageData.cpp"
  _lua_pushinteger($L,$height$0); //@line 71 "modules/image/wrap_CompressedImageData.cpp"
  STACKTOP = sp;return 2; //@line 72 "modules/image/wrap_CompressedImageData.cpp"
 }
 $22 = ___cxa_find_matching_catch_2()|0;
 $23 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $26 = ___cxa_find_matching_catch_3(0|0)|0;
  $27 = tempRet0;
  ___clang_call_terminate($26); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $23;$$02$i = $22;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5image36w_CompressedImageData_getMipmapCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5image19CompressedImageDataEEEPT_P9lua_StateiNS_4TypeE($L,1,23)|0); //@line 31 "modules/image/wrap_CompressedImageData.cpp"
 $1 = (__ZNK4love5image19CompressedImageData14getMipmapCountEv($0)|0); //@line 78 "modules/image/wrap_CompressedImageData.cpp"
 _lua_pushinteger($L,$1); //@line 78 "modules/image/wrap_CompressedImageData.cpp"
 return 1; //@line 79 "modules/image/wrap_CompressedImageData.cpp"
}
function __ZN4love5image31w_CompressedImageData_getFormatEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $str = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $str = sp;
 $0 = (__ZN4love14luax_checktypeINS_5image19CompressedImageDataEEEPT_P9lua_StateiNS_4TypeE($L,1,23)|0); //@line 31 "modules/image/wrap_CompressedImageData.cpp"
 $1 = (__ZNK4love5image19CompressedImageData9getFormatEv($0)|0); //@line 86 "modules/image/wrap_CompressedImageData.cpp"
 $2 = (__ZN4love5image19CompressedImageData11getConstantENS1_6FormatERPKc($1,$str)|0); //@line 89 "modules/image/wrap_CompressedImageData.cpp"
 if ($2) {
  $3 = HEAP32[$str>>2]|0; //@line 90 "modules/image/wrap_CompressedImageData.cpp"
  _lua_pushstring($L,$3); //@line 90 "modules/image/wrap_CompressedImageData.cpp"
  STACKTOP = sp;return 1; //@line 94 "modules/image/wrap_CompressedImageData.cpp"
 } else {
  _lua_pushstring($L,442424); //@line 92 "modules/image/wrap_CompressedImageData.cpp"
  STACKTOP = sp;return 1; //@line 94 "modules/image/wrap_CompressedImageData.cpp"
 }
 return (0)|0;
}
function _luaopen_compressedimagedata($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 107 "modules/image/wrap_CompressedImageData.cpp"
 HEAP32[$vararg_buffer>>2] = 12272; //@line 109 "modules/image/wrap_CompressedImageData.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 109 "modules/image/wrap_CompressedImageData.cpp"
 HEAP32[$vararg_ptr1>>2] = 16204; //@line 109 "modules/image/wrap_CompressedImageData.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 109 "modules/image/wrap_CompressedImageData.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 109 "modules/image/wrap_CompressedImageData.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,23,347981,$vararg_buffer)|0); //@line 109 "modules/image/wrap_CompressedImageData.cpp"
 STACKTOP = sp;return ($0|0); //@line 109 "modules/image/wrap_CompressedImageData.cpp"
}
function __ZN4love5image14w_newImageDataEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$01$i3 = 0, $$02$i = 0, $$02$i2 = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0.0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $bytes$0 = 0, $numbytes = 0, $or$cond = 0, $t$0 = 0, $t1$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $numbytes = sp + 28|0;
 $0 = (_lua_isnumber($L,1)|0); //@line 40 "modules/image/wrap_Image.cpp"
 $1 = ($0|0)==(0); //@line 40 "modules/image/wrap_Image.cpp"
 if ($1) {
  $59 = (__ZN4love10filesystem16luax_getfiledataEP9lua_Statei($L,1)|0); //@line 73 "modules/image/wrap_Image.cpp"
  $60 = HEAP32[(625628)>>2]|0; //@line 98 "./common/Module.h"
  $61 = HEAP32[$60>>2]|0; //@line 77 "modules/image/wrap_Image.cpp"
  $62 = ((($61)) + 24|0); //@line 77 "modules/image/wrap_Image.cpp"
  $63 = HEAP32[$62>>2]|0; //@line 77 "modules/image/wrap_Image.cpp"
  __THREW__ = 0;
  $64 = (invoke_iii($63|0,($60|0),($59|0))|0); //@line 77 "modules/image/wrap_Image.cpp"
  $65 = __THREW__; __THREW__ = 0;
  $66 = $65&1;
  do {
   if ($66) {
    $67 = ___cxa_find_matching_catch_3(8992|0)|0;
    $68 = tempRet0;
    $69 = (_llvm_eh_typeid_for((8992|0))|0); //@line 556 "./common/runtime.h"
    $70 = ($68|0)==($69|0); //@line 556 "./common/runtime.h"
    if (!($70)) {
     $$01$i3 = $68;$$02$i2 = $67;
     ___resumeException($$02$i2|0);
     // unreachable;
    }
    $71 = (___cxa_begin_catch(($67|0))|0); //@line 556 "./common/runtime.h"
    $72 = HEAP32[$71>>2]|0; //@line 560 "./common/runtime.h"
    $73 = ((($72)) + 8|0); //@line 560 "./common/runtime.h"
    $74 = HEAP32[$73>>2]|0; //@line 560 "./common/runtime.h"
    $75 = (FUNCTION_TABLE_ii[$74 & 8191]($71)|0); //@line 560 "./common/runtime.h"
    __THREW__ = 0;
    invoke_vii(3380,($L|0),($75|0)); //@line 560 "./common/runtime.h"
    $76 = __THREW__; __THREW__ = 0;
    $77 = $76&1;
    if (!($77)) {
     ___cxa_end_catch(); //@line 561 "./common/runtime.h"
     $78 = HEAP32[$59>>2]|0; //@line 78 "modules/image/wrap_Image.cpp"
     $79 = ((($78)) + 12|0); //@line 78 "modules/image/wrap_Image.cpp"
     $80 = HEAP32[$79>>2]|0; //@line 78 "modules/image/wrap_Image.cpp"
     FUNCTION_TABLE_vi[$80 & 8191]($59); //@line 78 "modules/image/wrap_Image.cpp"
     $81 = (_lua_tolstring($L,-1,0)|0); //@line 566 "./common/runtime.h"
     HEAP32[$vararg_buffer5>>2] = $81; //@line 566 "./common/runtime.h"
     (_luaL_error($L,595867,$vararg_buffer5)|0); //@line 566 "./common/runtime.h"
     $t1$0 = 0;
     break;
    }
    $82 = ___cxa_find_matching_catch_2()|0;
    $83 = tempRet0;
    __THREW__ = 0;
    invoke_v(3378); //@line 561 "./common/runtime.h"
    $84 = __THREW__; __THREW__ = 0;
    $85 = $84&1;
    if ($85) {
     $89 = ___cxa_find_matching_catch_3(0|0)|0;
     $90 = tempRet0;
     ___clang_call_terminate($89); //@line 561 "./common/runtime.h"
     // unreachable; //@line 561 "./common/runtime.h"
    } else {
     $$01$i3 = $83;$$02$i2 = $82;
     ___resumeException($$02$i2|0);
     // unreachable;
    }
   } else {
    $86 = HEAP32[$59>>2]|0; //@line 78 "modules/image/wrap_Image.cpp"
    $87 = ((($86)) + 12|0); //@line 78 "modules/image/wrap_Image.cpp"
    $88 = HEAP32[$87>>2]|0; //@line 78 "modules/image/wrap_Image.cpp"
    FUNCTION_TABLE_vi[$88 & 8191]($59); //@line 78 "modules/image/wrap_Image.cpp"
    $t1$0 = $64;
   }
  } while(0);
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,22,$t1$0); //@line 81 "modules/image/wrap_Image.cpp"
  $91 = HEAP32[$t1$0>>2]|0; //@line 82 "modules/image/wrap_Image.cpp"
  $92 = ((($91)) + 12|0); //@line 82 "modules/image/wrap_Image.cpp"
  $93 = HEAP32[$92>>2]|0; //@line 82 "modules/image/wrap_Image.cpp"
  FUNCTION_TABLE_vi[$93 & 8191]($t1$0); //@line 82 "modules/image/wrap_Image.cpp"
  $$2 = 1;
  STACKTOP = sp;return ($$2|0); //@line 84 "modules/image/wrap_Image.cpp"
 }
 $2 = (+_luaL_checknumber($L,1)); //@line 42 "modules/image/wrap_Image.cpp"
 $3 = (~~(($2))); //@line 42 "modules/image/wrap_Image.cpp"
 $4 = (+_luaL_checknumber($L,2)); //@line 43 "modules/image/wrap_Image.cpp"
 $5 = (~~(($4))); //@line 43 "modules/image/wrap_Image.cpp"
 $6 = ($3|0)<(1); //@line 44 "modules/image/wrap_Image.cpp"
 $7 = ($5|0)<(1); //@line 44 "modules/image/wrap_Image.cpp"
 $or$cond = $6 | $7; //@line 44 "modules/image/wrap_Image.cpp"
 if ($or$cond) {
  $8 = (_luaL_error($L,348016,$vararg_buffer)|0); //@line 45 "modules/image/wrap_Image.cpp"
  $$2 = $8;
  STACKTOP = sp;return ($$2|0); //@line 84 "modules/image/wrap_Image.cpp"
 }
 HEAP32[$numbytes>>2] = 0; //@line 47 "modules/image/wrap_Image.cpp"
 $9 = (_lua_type($L,3)|0); //@line 50 "modules/image/wrap_Image.cpp"
 $10 = ($9|0)<(1); //@line 50 "modules/image/wrap_Image.cpp"
 if ($10) {
  $bytes$0 = 0;
 } else {
  $11 = (_luaL_checklstring($L,3,$numbytes)|0); //@line 51 "modules/image/wrap_Image.cpp"
  $bytes$0 = $11;
 }
 $12 = HEAP32[(625628)>>2]|0; //@line 98 "./common/Module.h"
 $13 = HEAP32[$12>>2]|0; //@line 54 "modules/image/wrap_Image.cpp"
 $14 = ((($13)) + 28|0); //@line 54 "modules/image/wrap_Image.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 54 "modules/image/wrap_Image.cpp"
 __THREW__ = 0;
 $16 = (invoke_iiii($15|0,($12|0),($3|0),($5|0))|0); //@line 54 "modules/image/wrap_Image.cpp"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 do {
  if ($18) {
   $19 = ___cxa_find_matching_catch_3(8992|0)|0;
   $20 = tempRet0;
   $21 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $22 = ($20|0)==($21|0); //@line 535 "./common/runtime.h"
   if (!($22)) {
    $$01$i = $20;$$02$i = $19;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $23 = (___cxa_begin_catch(($19|0))|0); //@line 535 "./common/runtime.h"
   $24 = HEAP32[$23>>2]|0; //@line 539 "./common/runtime.h"
   $25 = ((($24)) + 8|0); //@line 539 "./common/runtime.h"
   $26 = HEAP32[$25>>2]|0; //@line 539 "./common/runtime.h"
   $27 = (FUNCTION_TABLE_ii[$26 & 8191]($23)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($27|0)); //@line 539 "./common/runtime.h"
   $28 = __THREW__; __THREW__ = 0;
   $29 = $28&1;
   if (!($29)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $30 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer1>>2] = $30; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
    $t$0 = 0;
    break;
   }
   $31 = ___cxa_find_matching_catch_2()|0;
   $32 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $33 = __THREW__; __THREW__ = 0;
   $34 = $33&1;
   if ($34) {
    $35 = ___cxa_find_matching_catch_3(0|0)|0;
    $36 = tempRet0;
    ___clang_call_terminate($35); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $32;$$02$i = $31;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $t$0 = $16;
  }
 } while(0);
 $37 = ($bytes$0|0)==(0|0); //@line 56 "modules/image/wrap_Image.cpp"
 do {
  if ($37) {
   label = 17;
  } else {
   $38 = HEAP32[$numbytes>>2]|0; //@line 58 "modules/image/wrap_Image.cpp"
   $39 = HEAP32[$t$0>>2]|0; //@line 58 "modules/image/wrap_Image.cpp"
   $40 = ((($39)) + 20|0); //@line 58 "modules/image/wrap_Image.cpp"
   $41 = HEAP32[$40>>2]|0; //@line 58 "modules/image/wrap_Image.cpp"
   $42 = (FUNCTION_TABLE_ii[$41 & 8191]($t$0)|0); //@line 58 "modules/image/wrap_Image.cpp"
   $43 = ($38|0)==($42|0); //@line 58 "modules/image/wrap_Image.cpp"
   if ($43) {
    $48 = HEAP32[$t$0>>2]|0; //@line 64 "modules/image/wrap_Image.cpp"
    $49 = ((($48)) + 16|0); //@line 64 "modules/image/wrap_Image.cpp"
    $50 = HEAP32[$49>>2]|0; //@line 64 "modules/image/wrap_Image.cpp"
    $51 = (FUNCTION_TABLE_ii[$50 & 8191]($t$0)|0); //@line 64 "modules/image/wrap_Image.cpp"
    $52 = HEAP32[$t$0>>2]|0; //@line 64 "modules/image/wrap_Image.cpp"
    $53 = ((($52)) + 20|0); //@line 64 "modules/image/wrap_Image.cpp"
    $54 = HEAP32[$53>>2]|0; //@line 64 "modules/image/wrap_Image.cpp"
    $55 = (FUNCTION_TABLE_ii[$54 & 8191]($t$0)|0); //@line 64 "modules/image/wrap_Image.cpp"
    _memcpy(($51|0),($bytes$0|0),($55|0))|0; //@line 64 "modules/image/wrap_Image.cpp"
    label = 17;
    break;
   } else {
    $44 = HEAP32[$t$0>>2]|0; //@line 60 "modules/image/wrap_Image.cpp"
    $45 = ((($44)) + 12|0); //@line 60 "modules/image/wrap_Image.cpp"
    $46 = HEAP32[$45>>2]|0; //@line 60 "modules/image/wrap_Image.cpp"
    FUNCTION_TABLE_vi[$46 & 8191]($t$0); //@line 60 "modules/image/wrap_Image.cpp"
    $47 = (_luaL_error($L,348036,$vararg_buffer3)|0); //@line 61 "modules/image/wrap_Image.cpp"
    $$0 = $47;
    break;
   }
  }
 } while(0);
 if ((label|0) == 17) {
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,22,$t$0); //@line 67 "modules/image/wrap_Image.cpp"
  $56 = HEAP32[$t$0>>2]|0; //@line 68 "modules/image/wrap_Image.cpp"
  $57 = ((($56)) + 12|0); //@line 68 "modules/image/wrap_Image.cpp"
  $58 = HEAP32[$57>>2]|0; //@line 68 "modules/image/wrap_Image.cpp"
  FUNCTION_TABLE_vi[$58 & 8191]($t$0); //@line 68 "modules/image/wrap_Image.cpp"
  $$0 = 1;
 }
 $$2 = $$0;
 STACKTOP = sp;return ($$2|0); //@line 84 "modules/image/wrap_Image.cpp"
}
function __ZN4love5image19w_newCompressedDataEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $t$0 = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 86 "modules/image/wrap_Image.cpp"
 $0 = (__ZN4love10filesystem16luax_getfiledataEP9lua_Statei($L,1)|0); //@line 88 "modules/image/wrap_Image.cpp"
 $1 = HEAP32[(625628)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 92 "modules/image/wrap_Image.cpp"
 $3 = ((($2)) + 36|0); //@line 92 "modules/image/wrap_Image.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 92 "modules/image/wrap_Image.cpp"
 __THREW__ = 0;
 $5 = (invoke_iii($4|0,($1|0),($0|0))|0); //@line 92 "modules/image/wrap_Image.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  $27 = HEAP32[$0>>2]|0; //@line 93 "modules/image/wrap_Image.cpp"
  $28 = ((($27)) + 12|0); //@line 93 "modules/image/wrap_Image.cpp"
  $29 = HEAP32[$28>>2]|0; //@line 93 "modules/image/wrap_Image.cpp"
  FUNCTION_TABLE_vi[$29 & 8191]($0); //@line 93 "modules/image/wrap_Image.cpp"
  $t$0 = $5;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,23,$t$0); //@line 96 "modules/image/wrap_Image.cpp"
  $32 = HEAP32[$t$0>>2]|0; //@line 97 "modules/image/wrap_Image.cpp"
  $33 = ((($32)) + 12|0); //@line 97 "modules/image/wrap_Image.cpp"
  $34 = HEAP32[$33>>2]|0; //@line 97 "modules/image/wrap_Image.cpp"
  FUNCTION_TABLE_vi[$34 & 8191]($t$0); //@line 97 "modules/image/wrap_Image.cpp"
  STACKTOP = sp;return 1; //@line 98 "modules/image/wrap_Image.cpp"
 }
 $8 = ___cxa_find_matching_catch_3(8992|0)|0;
 $9 = tempRet0;
 $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 556 "./common/runtime.h"
 $11 = ($9|0)==($10|0); //@line 556 "./common/runtime.h"
 if (!($11)) {
  $$01$i = $9;$$02$i = $8;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $12 = (___cxa_begin_catch(($8|0))|0); //@line 556 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 560 "./common/runtime.h"
 $14 = ((($13)) + 8|0); //@line 560 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 560 "./common/runtime.h"
 $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 560 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($16|0)); //@line 560 "./common/runtime.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if (!($18)) {
  ___cxa_end_catch(); //@line 561 "./common/runtime.h"
  $19 = HEAP32[$0>>2]|0; //@line 93 "modules/image/wrap_Image.cpp"
  $20 = ((($19)) + 12|0); //@line 93 "modules/image/wrap_Image.cpp"
  $21 = HEAP32[$20>>2]|0; //@line 93 "modules/image/wrap_Image.cpp"
  FUNCTION_TABLE_vi[$21 & 8191]($0); //@line 93 "modules/image/wrap_Image.cpp"
  $22 = (_lua_tolstring($L,-1,0)|0); //@line 566 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $22; //@line 566 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 566 "./common/runtime.h"
  $t$0 = 0;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,23,$t$0); //@line 96 "modules/image/wrap_Image.cpp"
  $32 = HEAP32[$t$0>>2]|0; //@line 97 "modules/image/wrap_Image.cpp"
  $33 = ((($32)) + 12|0); //@line 97 "modules/image/wrap_Image.cpp"
  $34 = HEAP32[$33>>2]|0; //@line 97 "modules/image/wrap_Image.cpp"
  FUNCTION_TABLE_vi[$34 & 8191]($t$0); //@line 97 "modules/image/wrap_Image.cpp"
  STACKTOP = sp;return 1; //@line 98 "modules/image/wrap_Image.cpp"
 }
 $23 = ___cxa_find_matching_catch_2()|0;
 $24 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 561 "./common/runtime.h"
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if ($26) {
  $30 = ___cxa_find_matching_catch_3(0|0)|0;
  $31 = tempRet0;
  ___clang_call_terminate($30); //@line 561 "./common/runtime.h"
  // unreachable; //@line 561 "./common/runtime.h"
 } else {
  $$01$i = $24;$$02$i = $23;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5image14w_isCompressedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love10filesystem16luax_getfiledataEP9lua_Statei($L,1)|0); //@line 103 "modules/image/wrap_Image.cpp"
 $1 = HEAP32[(625628)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 104 "modules/image/wrap_Image.cpp"
 $3 = ((($2)) + 40|0); //@line 104 "modules/image/wrap_Image.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 104 "modules/image/wrap_Image.cpp"
 $5 = (FUNCTION_TABLE_iii[$4 & 8191]($1,$0)|0); //@line 104 "modules/image/wrap_Image.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 105 "modules/image/wrap_Image.cpp"
 $7 = ((($6)) + 12|0); //@line 105 "modules/image/wrap_Image.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 105 "modules/image/wrap_Image.cpp"
 FUNCTION_TABLE_vi[$8 & 8191]($0); //@line 105 "modules/image/wrap_Image.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$5); //@line 107 "modules/image/wrap_Image.cpp"
 return 1; //@line 108 "modules/image/wrap_Image.cpp"
}
function _luaopen_love_image($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0;
 var $eh$lpad$body$i$indexZ2D = 0, $instance$sroa$0$1 = 0, $vararg_buffer = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $w = sp + 4|0;
 $0 = HEAP32[(625628)>>2]|0; //@line 98 "./common/Module.h"
 $1 = $0; //@line 129 "modules/image/wrap_Image.cpp"
 $2 = ($0|0)==(0|0); //@line 130 "modules/image/wrap_Image.cpp"
 L1: do {
  if ($2) {
   __THREW__ = 0;
   $3 = (invoke_ii(3348,32)|0); //@line 132 "modules/image/wrap_Image.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   do {
    if ($5) {
     $11 = ___cxa_find_matching_catch_3(8992|0)|0;
     $12 = tempRet0;
     $eh$lpad$body$i$index2Z2D = $12;$eh$lpad$body$i$indexZ2D = $11;
    } else {
     __THREW__ = 0;
     invoke_vi(3637,($3|0)); //@line 132 "modules/image/wrap_Image.cpp"
     $6 = __THREW__; __THREW__ = 0;
     $7 = $6&1;
     if ($7) {
      $8 = ___cxa_find_matching_catch_3(8992|0)|0;
      $9 = tempRet0;
      __ZdlPv($3); //@line 132 "modules/image/wrap_Image.cpp"
      $eh$lpad$body$i$index2Z2D = $9;$eh$lpad$body$i$indexZ2D = $8;
      break;
     } else {
      $10 = $3; //@line 132 "modules/image/wrap_Image.cpp"
      $instance$sroa$0$1 = $10;
      break L1;
     }
    }
   } while(0);
   $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $14 = ($eh$lpad$body$i$index2Z2D|0)==($13|0); //@line 535 "./common/runtime.h"
   if (!($14)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $15 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
   $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
   $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
   $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $instance$sroa$0$1 = $1;
    break;
   }
   $23 = ___cxa_find_matching_catch_2()|0;
   $24 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    $27 = ___cxa_find_matching_catch_3(0|0)|0;
    $28 = tempRet0;
    ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $24;$$02$i = $23;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $29 = HEAP32[$0>>2]|0; //@line 135 "modules/image/wrap_Image.cpp"
   $30 = ((($29)) + 8|0); //@line 135 "modules/image/wrap_Image.cpp"
   $31 = HEAP32[$30>>2]|0; //@line 135 "modules/image/wrap_Image.cpp"
   FUNCTION_TABLE_vi[$31 & 8191]($0); //@line 135 "modules/image/wrap_Image.cpp"
   $instance$sroa$0$1 = $1;
  }
 } while(0);
 HEAP32[$w>>2] = $instance$sroa$0$1; //@line 138 "modules/image/wrap_Image.cpp"
 $32 = ((($w)) + 4|0); //@line 139 "modules/image/wrap_Image.cpp"
 HEAP32[$32>>2] = 438231; //@line 139 "modules/image/wrap_Image.cpp"
 $33 = ((($w)) + 8|0); //@line 140 "modules/image/wrap_Image.cpp"
 HEAP32[$33>>2] = 58; //@line 140 "modules/image/wrap_Image.cpp"
 $34 = ((($w)) + 12|0); //@line 141 "modules/image/wrap_Image.cpp"
 HEAP32[$34>>2] = 16252; //@line 141 "modules/image/wrap_Image.cpp"
 $35 = ((($w)) + 16|0); //@line 142 "modules/image/wrap_Image.cpp"
 HEAP32[$35>>2] = 16284; //@line 142 "modules/image/wrap_Image.cpp"
 $36 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 144 "modules/image/wrap_Image.cpp"
 STACKTOP = sp;return ($36|0); //@line 144 "modules/image/wrap_Image.cpp"
}
function __ZN4love5image19luax_checkimagedataEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5image9ImageDataEEEPT_P9lua_StateiNS_4TypeE($L,$idx,22)|0); //@line 43 "modules/image/wrap_ImageData.cpp"
 return ($0|0); //@line 43 "modules/image/wrap_ImageData.cpp"
}
function __ZN4love5image20w_ImageData_getWidthEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5image9ImageDataEEEPT_P9lua_StateiNS_4TypeE($L,1,22)|0); //@line 43 "modules/image/wrap_ImageData.cpp"
 $1 = (__ZNK4love5image9ImageData8getWidthEv($0)|0); //@line 49 "modules/image/wrap_ImageData.cpp"
 _lua_pushinteger($L,$1); //@line 49 "modules/image/wrap_ImageData.cpp"
 return 1; //@line 50 "modules/image/wrap_ImageData.cpp"
}
function __ZN4love5image21w_ImageData_getHeightEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5image9ImageDataEEEPT_P9lua_StateiNS_4TypeE($L,1,22)|0); //@line 43 "modules/image/wrap_ImageData.cpp"
 $1 = (__ZNK4love5image9ImageData9getHeightEv($0)|0); //@line 56 "modules/image/wrap_ImageData.cpp"
 _lua_pushinteger($L,$1); //@line 56 "modules/image/wrap_ImageData.cpp"
 return 1; //@line 57 "modules/image/wrap_ImageData.cpp"
}
function __ZN4love5image25w_ImageData_getDimensionsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5image9ImageDataEEEPT_P9lua_StateiNS_4TypeE($L,1,22)|0); //@line 43 "modules/image/wrap_ImageData.cpp"
 $1 = (__ZNK4love5image9ImageData8getWidthEv($0)|0); //@line 63 "modules/image/wrap_ImageData.cpp"
 _lua_pushinteger($L,$1); //@line 63 "modules/image/wrap_ImageData.cpp"
 $2 = (__ZNK4love5image9ImageData9getHeightEv($0)|0); //@line 64 "modules/image/wrap_ImageData.cpp"
 _lua_pushinteger($L,$2); //@line 64 "modules/image/wrap_ImageData.cpp"
 return 2; //@line 65 "modules/image/wrap_ImageData.cpp"
}
function __ZN4love5image20w_ImageData_getPixelEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $c$sroa$0$0 = 0.0, $c$sroa$4$0 = 0.0, $c$sroa$5$0 = 0.0, $c$sroa$6$0 = 0.0;
 var $phitmp = 0.0, $phitmp1 = 0.0, $phitmp2 = 0.0, $phitmp3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 $1 = (__ZN4love14luax_checktypeINS_5image9ImageDataEEEPT_P9lua_StateiNS_4TypeE($L,1,22)|0); //@line 43 "modules/image/wrap_ImageData.cpp"
 $2 = (+_luaL_checknumber($L,2)); //@line 71 "modules/image/wrap_ImageData.cpp"
 $3 = (~~(($2))); //@line 71 "modules/image/wrap_ImageData.cpp"
 $4 = (+_luaL_checknumber($L,3)); //@line 72 "modules/image/wrap_ImageData.cpp"
 $5 = (~~(($4))); //@line 72 "modules/image/wrap_ImageData.cpp"
 __THREW__ = 0;
 invoke_viiii(3638,($0|0),($1|0),($3|0),($5|0)); //@line 75 "modules/image/wrap_ImageData.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  $8 = HEAP32[$0>>2]|0; //@line 75 "modules/image/wrap_ImageData.cpp"
  $9 = $8&255; //@line 75 "modules/image/wrap_ImageData.cpp"
  $10 = $8 >>> 8; //@line 75 "modules/image/wrap_ImageData.cpp"
  $11 = $10&255; //@line 75 "modules/image/wrap_ImageData.cpp"
  $12 = $8 >>> 16; //@line 75 "modules/image/wrap_ImageData.cpp"
  $13 = $12&255; //@line 75 "modules/image/wrap_ImageData.cpp"
  $14 = $8 >>> 24; //@line 75 "modules/image/wrap_ImageData.cpp"
  $15 = $14&255; //@line 75 "modules/image/wrap_ImageData.cpp"
  $phitmp = (+($9&255)); //@line 75 "modules/image/wrap_ImageData.cpp"
  $phitmp1 = (+($11&255)); //@line 75 "modules/image/wrap_ImageData.cpp"
  $phitmp2 = (+($13&255)); //@line 75 "modules/image/wrap_ImageData.cpp"
  $phitmp3 = (+($15&255)); //@line 75 "modules/image/wrap_ImageData.cpp"
  $c$sroa$0$0 = $phitmp;$c$sroa$4$0 = $phitmp1;$c$sroa$5$0 = $phitmp2;$c$sroa$6$0 = $phitmp3;
  _lua_pushnumber($L,$c$sroa$0$0); //@line 77 "modules/image/wrap_ImageData.cpp"
  _lua_pushnumber($L,$c$sroa$4$0); //@line 78 "modules/image/wrap_ImageData.cpp"
  _lua_pushnumber($L,$c$sroa$5$0); //@line 79 "modules/image/wrap_ImageData.cpp"
  _lua_pushnumber($L,$c$sroa$6$0); //@line 80 "modules/image/wrap_ImageData.cpp"
  STACKTOP = sp;return 4; //@line 81 "modules/image/wrap_ImageData.cpp"
 }
 $16 = ___cxa_find_matching_catch_3(8992|0)|0;
 $17 = tempRet0;
 $18 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $19 = ($17|0)==($18|0); //@line 535 "./common/runtime.h"
 if (!($19)) {
  $$01$i = $17;$$02$i = $16;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $20 = (___cxa_begin_catch(($16|0))|0); //@line 535 "./common/runtime.h"
 $21 = HEAP32[$20>>2]|0; //@line 539 "./common/runtime.h"
 $22 = ((($21)) + 8|0); //@line 539 "./common/runtime.h"
 $23 = HEAP32[$22>>2]|0; //@line 539 "./common/runtime.h"
 $24 = (FUNCTION_TABLE_ii[$23 & 8191]($20)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($24|0)); //@line 539 "./common/runtime.h"
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if (!($26)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $27 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $27; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $c$sroa$0$0 = 0.0;$c$sroa$4$0 = 0.0;$c$sroa$5$0 = 0.0;$c$sroa$6$0 = 0.0;
  _lua_pushnumber($L,$c$sroa$0$0); //@line 77 "modules/image/wrap_ImageData.cpp"
  _lua_pushnumber($L,$c$sroa$4$0); //@line 78 "modules/image/wrap_ImageData.cpp"
  _lua_pushnumber($L,$c$sroa$5$0); //@line 79 "modules/image/wrap_ImageData.cpp"
  _lua_pushnumber($L,$c$sroa$6$0); //@line 80 "modules/image/wrap_ImageData.cpp"
  STACKTOP = sp;return 4; //@line 81 "modules/image/wrap_ImageData.cpp"
 }
 $28 = ___cxa_find_matching_catch_2()|0;
 $29 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $30 = __THREW__; __THREW__ = 0;
 $31 = $30&1;
 if ($31) {
  $32 = ___cxa_find_matching_catch_3(0|0)|0;
  $33 = tempRet0;
  ___clang_call_terminate($32); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $29;$$02$i = $28;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5image20w_ImageData_setPixelEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $c$sroa$0$0$in = 0, $c$sroa$5$0$in = 0, $c$sroa$7$0$in = 0, $c$sroa$9$0$in = 0, $tmpcast$i$i$byval_copy = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tmpcast$i$i$byval_copy = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 $1 = (__ZN4love14luax_checktypeINS_5image9ImageDataEEEPT_P9lua_StateiNS_4TypeE($L,1,22)|0); //@line 43 "modules/image/wrap_ImageData.cpp"
 $2 = (+_luaL_checknumber($L,2)); //@line 87 "modules/image/wrap_ImageData.cpp"
 $3 = (~~(($2))); //@line 87 "modules/image/wrap_ImageData.cpp"
 $4 = (+_luaL_checknumber($L,3)); //@line 88 "modules/image/wrap_ImageData.cpp"
 $5 = (~~(($4))); //@line 88 "modules/image/wrap_ImageData.cpp"
 $6 = (_lua_type($L,4)|0); //@line 91 "modules/image/wrap_ImageData.cpp"
 $7 = ($6|0)==(5); //@line 91 "modules/image/wrap_ImageData.cpp"
 if ($7) {
  _lua_rawgeti($L,4,1); //@line 94 "modules/image/wrap_ImageData.cpp"
  _lua_rawgeti($L,4,2); //@line 94 "modules/image/wrap_ImageData.cpp"
  _lua_rawgeti($L,4,3); //@line 94 "modules/image/wrap_ImageData.cpp"
  _lua_rawgeti($L,4,4); //@line 94 "modules/image/wrap_ImageData.cpp"
  $8 = (_luaL_checkinteger($L,-4)|0); //@line 96 "modules/image/wrap_ImageData.cpp"
  $9 = (_luaL_checkinteger($L,-3)|0); //@line 97 "modules/image/wrap_ImageData.cpp"
  $10 = (_luaL_checkinteger($L,-2)|0); //@line 98 "modules/image/wrap_ImageData.cpp"
  $11 = (_luaL_optinteger($L,-1,255)|0); //@line 99 "modules/image/wrap_ImageData.cpp"
  _lua_settop($L,-5); //@line 101 "modules/image/wrap_ImageData.cpp"
  $c$sroa$0$0$in = $8;$c$sroa$5$0$in = $9;$c$sroa$7$0$in = $10;$c$sroa$9$0$in = $11;
 } else {
  $12 = (_luaL_checkinteger($L,4)|0); //@line 105 "modules/image/wrap_ImageData.cpp"
  $13 = (_luaL_checkinteger($L,5)|0); //@line 106 "modules/image/wrap_ImageData.cpp"
  $14 = (_luaL_checkinteger($L,6)|0); //@line 107 "modules/image/wrap_ImageData.cpp"
  $15 = (_luaL_optinteger($L,7,255)|0); //@line 108 "modules/image/wrap_ImageData.cpp"
  $c$sroa$0$0$in = $12;$c$sroa$5$0$in = $13;$c$sroa$7$0$in = $14;$c$sroa$9$0$in = $15;
 }
 $16 = $c$sroa$9$0$in << 24; //@line 111 "modules/image/wrap_ImageData.cpp"
 $17 = $c$sroa$7$0$in << 16; //@line 111 "modules/image/wrap_ImageData.cpp"
 $18 = $17 & 16711680; //@line 111 "modules/image/wrap_ImageData.cpp"
 $19 = $18 | $16; //@line 111 "modules/image/wrap_ImageData.cpp"
 $20 = $c$sroa$5$0$in << 8; //@line 111 "modules/image/wrap_ImageData.cpp"
 $21 = $20 & 65280; //@line 111 "modules/image/wrap_ImageData.cpp"
 $22 = $19 | $21; //@line 111 "modules/image/wrap_ImageData.cpp"
 $23 = $c$sroa$0$0$in & 255; //@line 111 "modules/image/wrap_ImageData.cpp"
 $24 = $22 | $23; //@line 111 "modules/image/wrap_ImageData.cpp"
 HEAP32[$0>>2] = $24; //@line 111 "modules/image/wrap_ImageData.cpp"
 __THREW__ = 0;
 ;HEAP8[$tmpcast$i$i$byval_copy>>0]=HEAP8[$0>>0]|0;HEAP8[$tmpcast$i$i$byval_copy+1>>0]=HEAP8[$0+1>>0]|0;HEAP8[$tmpcast$i$i$byval_copy+2>>0]=HEAP8[$0+2>>0]|0;HEAP8[$tmpcast$i$i$byval_copy+3>>0]=HEAP8[$0+3>>0]|0; //@line 111 "modules/image/wrap_ImageData.cpp"
 invoke_viiii(3639,($1|0),($3|0),($5|0),($tmpcast$i$i$byval_copy|0)); //@line 111 "modules/image/wrap_ImageData.cpp"
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if (!($26)) {
  STACKTOP = sp;return 0; //@line 112 "modules/image/wrap_ImageData.cpp"
 }
 $27 = ___cxa_find_matching_catch_3(8992|0)|0;
 $28 = tempRet0;
 $29 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $30 = ($28|0)==($29|0); //@line 535 "./common/runtime.h"
 if (!($30)) {
  $$01$i = $28;$$02$i = $27;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $31 = (___cxa_begin_catch(($27|0))|0); //@line 535 "./common/runtime.h"
 $32 = HEAP32[$31>>2]|0; //@line 539 "./common/runtime.h"
 $33 = ((($32)) + 8|0); //@line 539 "./common/runtime.h"
 $34 = HEAP32[$33>>2]|0; //@line 539 "./common/runtime.h"
 $35 = (FUNCTION_TABLE_ii[$34 & 8191]($31)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($35|0)); //@line 539 "./common/runtime.h"
 $36 = __THREW__; __THREW__ = 0;
 $37 = $36&1;
 if (!($37)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $38 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $38; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 112 "modules/image/wrap_ImageData.cpp"
 }
 $39 = ___cxa_find_matching_catch_2()|0;
 $40 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $41 = __THREW__; __THREW__ = 0;
 $42 = $41&1;
 if ($42) {
  $43 = ___cxa_find_matching_catch_3(0|0)|0;
  $44 = tempRet0;
  ___clang_call_terminate($43); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $40;$$02$i = $39;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5image27w_ImageData__mapPixelUnsafeEP9lua_State($L) {
 $L = $L|0;
 var $$410 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0.0, $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0.0, $30 = 0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0, $37 = 0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0.0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0.0, $63 = 0, $64 = 0, $65 = 0, $66 = 0.0, $67 = 0, $68 = 0, $69 = 0.0, $7 = 0.0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $8 = 0.0, $9 = 0, $c = 0, $i$024$us$lcssa = 0, $tmpcast$byval_copy = 0, $vararg_buffer = 0;
 var $vararg_buffer1 = 0, $vararg_ptr3 = 0, $vararg_ptr4 = 0, $x$026$us = 0, $y$027$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tmpcast$byval_copy = sp + 40|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 28|0;
 $c = sp + 24|0;
 $1 = sp + 20|0;
 $2 = (__ZN4love14luax_checktypeINS_5image9ImageDataEEEPT_P9lua_StateiNS_4TypeE($L,1,22)|0); //@line 43 "modules/image/wrap_ImageData.cpp"
 _luaL_checktype($L,2,6); //@line 148 "modules/image/wrap_ImageData.cpp"
 $3 = (+_lua_tonumber($L,3)); //@line 151 "modules/image/wrap_ImageData.cpp"
 $4 = (~~(($3))); //@line 151 "modules/image/wrap_ImageData.cpp"
 $5 = (+_lua_tonumber($L,4)); //@line 152 "modules/image/wrap_ImageData.cpp"
 $6 = (~~(($5))); //@line 152 "modules/image/wrap_ImageData.cpp"
 $7 = (+_lua_tonumber($L,5)); //@line 153 "modules/image/wrap_ImageData.cpp"
 $8 = (+_lua_tonumber($L,6)); //@line 154 "modules/image/wrap_ImageData.cpp"
 $9 = (__ZNK4love5image9ImageData6insideEii($2,$4,$6)|0); //@line 156 "modules/image/wrap_ImageData.cpp"
 if (!($9)) {
  $44 = (_luaL_error($L,436685,$vararg_buffer)|0); //@line 157 "modules/image/wrap_ImageData.cpp"
  STACKTOP = sp;return ($44|0); //@line 202 "modules/image/wrap_ImageData.cpp"
 }
 $10 = (~~(($8))); //@line 154 "modules/image/wrap_ImageData.cpp"
 $11 = (~~(($7))); //@line 153 "modules/image/wrap_ImageData.cpp"
 $12 = (($11) + ($4))|0; //@line 156 "modules/image/wrap_ImageData.cpp"
 $13 = (($12) + -1)|0; //@line 156 "modules/image/wrap_ImageData.cpp"
 $14 = (($10) + ($6))|0; //@line 156 "modules/image/wrap_ImageData.cpp"
 $15 = (($14) + -1)|0; //@line 156 "modules/image/wrap_ImageData.cpp"
 $16 = (__ZNK4love5image9ImageData6insideEii($2,$13,$15)|0); //@line 156 "modules/image/wrap_ImageData.cpp"
 if (!($16)) {
  $44 = (_luaL_error($L,436685,$vararg_buffer)|0); //@line 157 "modules/image/wrap_ImageData.cpp"
  STACKTOP = sp;return ($44|0); //@line 202 "modules/image/wrap_ImageData.cpp"
 }
 $17 = ($10|0)>(0); //@line 160 "modules/image/wrap_ImageData.cpp"
 if (!($17)) {
  $$410 = 0;
  STACKTOP = sp;return ($$410|0);
 }
 $18 = ($11|0)>(0); //@line 162 "modules/image/wrap_ImageData.cpp"
 $19 = ((($c)) + 1|0); //@line 169 "modules/image/wrap_ImageData.cpp"
 $20 = ((($c)) + 2|0); //@line 170 "modules/image/wrap_ImageData.cpp"
 $21 = ((($c)) + 3|0); //@line 171 "modules/image/wrap_ImageData.cpp"
 if (!($18)) {
  $$410 = 0;
  STACKTOP = sp;return ($$410|0);
 }
 $22 = ((($c)) + 1|0); //@line 185 "modules/image/wrap_ImageData.cpp"
 $23 = ((($c)) + 2|0); //@line 185 "modules/image/wrap_ImageData.cpp"
 $24 = ((($c)) + 3|0); //@line 185 "modules/image/wrap_ImageData.cpp"
 $25 = ((($c)) + 3|0); //@line 187 "modules/image/wrap_ImageData.cpp"
 $y$027$us = $6;
 L13: while(1) {
  $27 = (+($y$027$us|0)); //@line 166 "modules/image/wrap_ImageData.cpp"
  $x$026$us = $4;
  while(1) {
   _lua_pushvalue($L,2); //@line 164 "modules/image/wrap_ImageData.cpp"
   $26 = (+($x$026$us|0)); //@line 165 "modules/image/wrap_ImageData.cpp"
   _lua_pushnumber($L,$26); //@line 165 "modules/image/wrap_ImageData.cpp"
   _lua_pushnumber($L,$27); //@line 166 "modules/image/wrap_ImageData.cpp"
   __ZNK4love5image9ImageData8getPixelEii($c,$2,$x$026$us,$y$027$us); //@line 167 "modules/image/wrap_ImageData.cpp"
   $28 = HEAP8[$c>>0]|0; //@line 168 "modules/image/wrap_ImageData.cpp"
   $29 = (+($28&255)); //@line 168 "modules/image/wrap_ImageData.cpp"
   _lua_pushnumber($L,$29); //@line 168 "modules/image/wrap_ImageData.cpp"
   $30 = HEAP8[$19>>0]|0; //@line 169 "modules/image/wrap_ImageData.cpp"
   $31 = (+($30&255)); //@line 169 "modules/image/wrap_ImageData.cpp"
   _lua_pushnumber($L,$31); //@line 169 "modules/image/wrap_ImageData.cpp"
   $32 = HEAP8[$20>>0]|0; //@line 170 "modules/image/wrap_ImageData.cpp"
   $33 = (+($32&255)); //@line 170 "modules/image/wrap_ImageData.cpp"
   _lua_pushnumber($L,$33); //@line 170 "modules/image/wrap_ImageData.cpp"
   $34 = HEAP8[$21>>0]|0; //@line 171 "modules/image/wrap_ImageData.cpp"
   $35 = (+($34&255)); //@line 171 "modules/image/wrap_ImageData.cpp"
   _lua_pushnumber($L,$35); //@line 171 "modules/image/wrap_ImageData.cpp"
   _lua_call($L,6,4); //@line 172 "modules/image/wrap_ImageData.cpp"
   $36 = (_lua_type($L,-4)|0); //@line 182 "modules/image/wrap_ImageData.cpp"
   $37 = ($36|0)==(3); //@line 184 "modules/image/wrap_ImageData.cpp"
   if (!($37)) {
    $$lcssa = $36;$i$024$us$lcssa = 1;
    break L13;
   }
   $38 = (+_lua_tonumber($L,-4)); //@line 185 "modules/image/wrap_ImageData.cpp"
   $39 = (~~(($38))&255); //@line 185 "modules/image/wrap_ImageData.cpp"
   HEAP8[$c>>0] = $39; //@line 185 "modules/image/wrap_ImageData.cpp"
   $40 = (_lua_type($L,-3)|0); //@line 182 "modules/image/wrap_ImageData.cpp"
   $41 = ($40|0)==(3); //@line 184 "modules/image/wrap_ImageData.cpp"
   if (!($41)) {
    $$lcssa = $40;$i$024$us$lcssa = 2;
    break L13;
   }
   $62 = (+_lua_tonumber($L,-3)); //@line 185 "modules/image/wrap_ImageData.cpp"
   $63 = (~~(($62))&255); //@line 185 "modules/image/wrap_ImageData.cpp"
   HEAP8[$22>>0] = $63; //@line 185 "modules/image/wrap_ImageData.cpp"
   $64 = (_lua_type($L,-2)|0); //@line 182 "modules/image/wrap_ImageData.cpp"
   $65 = ($64|0)==(3); //@line 184 "modules/image/wrap_ImageData.cpp"
   if (!($65)) {
    $$lcssa = $64;$i$024$us$lcssa = 3;
    break L13;
   }
   $66 = (+_lua_tonumber($L,-2)); //@line 185 "modules/image/wrap_ImageData.cpp"
   $67 = (~~(($66))&255); //@line 185 "modules/image/wrap_ImageData.cpp"
   HEAP8[$23>>0] = $67; //@line 185 "modules/image/wrap_ImageData.cpp"
   $68 = (_lua_type($L,-1)|0); //@line 182 "modules/image/wrap_ImageData.cpp"
   switch ($68|0) {
   case 3:  {
    $69 = (+_lua_tonumber($L,-1)); //@line 185 "modules/image/wrap_ImageData.cpp"
    $70 = (~~(($69))&255); //@line 185 "modules/image/wrap_ImageData.cpp"
    HEAP8[$24>>0] = $70; //@line 185 "modules/image/wrap_ImageData.cpp"
    break;
   }
   case 0: case -1:  {
    HEAP8[$25>>0] = -1; //@line 187 "modules/image/wrap_ImageData.cpp"
    break;
   }
   default: {
    $$lcssa = $68;$i$024$us$lcssa = 4;
    break L13;
   }
   }
   _lua_settop($L,-5); //@line 194 "modules/image/wrap_ImageData.cpp"
   $71 = HEAP32[$c>>2]|0; //@line 197 "modules/image/wrap_ImageData.cpp"
   HEAP32[$1>>2] = $71; //@line 197 "modules/image/wrap_ImageData.cpp"
   ;HEAP8[$tmpcast$byval_copy>>0]=HEAP8[$1>>0]|0;HEAP8[$tmpcast$byval_copy+1>>0]=HEAP8[$1+1>>0]|0;HEAP8[$tmpcast$byval_copy+2>>0]=HEAP8[$1+2>>0]|0;HEAP8[$tmpcast$byval_copy+3>>0]=HEAP8[$1+3>>0]|0; //@line 197 "modules/image/wrap_ImageData.cpp"
   __ZN4love5image9ImageData14setPixelUnsafeEiiNS0_5pixelE($2,$x$026$us,$y$027$us,$tmpcast$byval_copy); //@line 197 "modules/image/wrap_ImageData.cpp"
   $72 = (($x$026$us) + 1)|0; //@line 162 "modules/image/wrap_ImageData.cpp"
   $73 = ($72|0)<($12|0); //@line 162 "modules/image/wrap_ImageData.cpp"
   if ($73) {
    $x$026$us = $72;
   } else {
    break;
   }
  }
  $42 = (($y$027$us) + 1)|0; //@line 160 "modules/image/wrap_ImageData.cpp"
  $43 = ($42|0)<($14|0); //@line 160 "modules/image/wrap_ImageData.cpp"
  if ($43) {
   $y$027$us = $42;
  } else {
   $$410 = 0;
   label = 16;
   break;
  }
 }
 if ((label|0) == 16) {
  STACKTOP = sp;return ($$410|0);
 }
 _luaL_where($L,2); //@line 118 "modules/image/wrap_ImageData.cpp"
 $45 = (_lua_tolstring($L,-1,0)|0); //@line 120 "modules/image/wrap_ImageData.cpp"
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $46 = ($45|0)==(0|0); //@line 122 "modules/image/wrap_ImageData.cpp"
 if (!($46)) {
  __THREW__ = 0;
  (invoke_iii(3360,($0|0),($45|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $47 = __THREW__; __THREW__ = 0;
  $48 = $47&1;
  if ($48) {
   $49 = ___cxa_find_matching_catch_2()|0;
   $50 = tempRet0;
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 127 "modules/image/wrap_ImageData.cpp"
   ___resumeException($49|0);
   // unreachable;
  }
 }
 __THREW__ = 0;
 invoke_vii(3384,($L|0),-2); //@line 125 "modules/image/wrap_ImageData.cpp"
 $51 = __THREW__; __THREW__ = 0;
 $52 = $51&1;
 if ($52) {
  $49 = ___cxa_find_matching_catch_2()|0;
  $50 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 127 "modules/image/wrap_ImageData.cpp"
  ___resumeException($49|0);
  // unreachable;
 }
 $53 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $54 = $53 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $55 = ($54<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $56 = ((($0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $57 = HEAP32[$56>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $58 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $59 = $55 ? $58 : $57; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 136 "modules/image/wrap_ImageData.cpp"
 $60 = (_lua_typename($L,$$lcssa)|0); //@line 137 "modules/image/wrap_ImageData.cpp"
 HEAP32[$vararg_buffer1>>2] = $59; //@line 139 "modules/image/wrap_ImageData.cpp"
 $vararg_ptr3 = ((($vararg_buffer1)) + 4|0); //@line 139 "modules/image/wrap_ImageData.cpp"
 HEAP32[$vararg_ptr3>>2] = $i$024$us$lcssa; //@line 139 "modules/image/wrap_ImageData.cpp"
 $vararg_ptr4 = ((($vararg_buffer1)) + 8|0); //@line 139 "modules/image/wrap_ImageData.cpp"
 HEAP32[$vararg_ptr4>>2] = $60; //@line 139 "modules/image/wrap_ImageData.cpp"
 $61 = (_luaL_error($L,348148,$vararg_buffer1)|0); //@line 139 "modules/image/wrap_ImageData.cpp"
 $$410 = $61;
 STACKTOP = sp;return ($$410|0);
}
function __ZN4love5image17w_ImageData_pasteEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $2 = 0.0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5image9ImageDataEEEPT_P9lua_StateiNS_4TypeE($L,1,22)|0); //@line 43 "modules/image/wrap_ImageData.cpp"
 $1 = (__ZN4love14luax_checktypeINS_5image9ImageDataEEEPT_P9lua_StateiNS_4TypeE($L,2,22)|0); //@line 43 "modules/image/wrap_ImageData.cpp"
 $2 = (+_luaL_checknumber($L,3)); //@line 208 "modules/image/wrap_ImageData.cpp"
 $3 = (~~(($2))); //@line 208 "modules/image/wrap_ImageData.cpp"
 $4 = (+_luaL_checknumber($L,4)); //@line 209 "modules/image/wrap_ImageData.cpp"
 $5 = (~~(($4))); //@line 209 "modules/image/wrap_ImageData.cpp"
 $6 = (+_luaL_optnumber($L,5,0.0)); //@line 210 "modules/image/wrap_ImageData.cpp"
 $7 = (~~(($6))); //@line 210 "modules/image/wrap_ImageData.cpp"
 $8 = (+_luaL_optnumber($L,6,0.0)); //@line 211 "modules/image/wrap_ImageData.cpp"
 $9 = (~~(($8))); //@line 211 "modules/image/wrap_ImageData.cpp"
 $10 = (__ZNK4love5image9ImageData8getWidthEv($1)|0); //@line 212 "modules/image/wrap_ImageData.cpp"
 $11 = (+($10|0)); //@line 212 "modules/image/wrap_ImageData.cpp"
 $12 = (+_luaL_optnumber($L,7,$11)); //@line 212 "modules/image/wrap_ImageData.cpp"
 $13 = (~~(($12))); //@line 212 "modules/image/wrap_ImageData.cpp"
 $14 = (__ZNK4love5image9ImageData9getHeightEv($1)|0); //@line 213 "modules/image/wrap_ImageData.cpp"
 $15 = (+($14|0)); //@line 213 "modules/image/wrap_ImageData.cpp"
 $16 = (+_luaL_optnumber($L,8,$15)); //@line 213 "modules/image/wrap_ImageData.cpp"
 $17 = (~~(($16))); //@line 213 "modules/image/wrap_ImageData.cpp"
 __ZN4love5image9ImageData5pasteEPS1_iiiiii($0,$1,$3,$5,$7,$9,$13,$17); //@line 214 "modules/image/wrap_ImageData.cpp"
 return 0; //@line 215 "modules/image/wrap_ImageData.cpp"
}
function __ZN4love5image18w_ImageData_encodeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $$01$i = 0, $$02 = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $9 = 0, $eh$lpad$body$index5Z2D = 0, $eh$lpad$body$indexZ2D = 0, $filedata$0 = 0, $filename = 0, $format = 0, $hasfilename$0$off0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $format = sp + 48|0;
 $filename = sp + 36|0;
 $0 = sp + 24|0;
 $1 = sp + 12|0;
 $2 = (__ZN4love14luax_checktypeINS_5image9ImageDataEEEPT_P9lua_StateiNS_4TypeE($L,1,22)|0); //@line 43 "modules/image/wrap_ImageData.cpp"
 $3 = (_luaL_checklstring($L,2,0)|0); //@line 223 "modules/image/wrap_ImageData.cpp"
 $4 = (__ZN4love5image9ImageData11getConstantEPKcRNS1_13EncodedFormatE($3,$format)|0); //@line 224 "modules/image/wrap_ImageData.cpp"
 if (!($4)) {
  HEAP32[$vararg_buffer>>2] = $3; //@line 225 "modules/image/wrap_ImageData.cpp"
  $5 = (_luaL_error($L,348197,$vararg_buffer)|0); //@line 225 "modules/image/wrap_ImageData.cpp"
  $$02 = $5;
  STACKTOP = sp;return ($$02|0); //@line 251 "modules/image/wrap_ImageData.cpp"
 }
 $6 = (_strlen($3)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($0,$3,$6); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 $7 = (invoke_iiii(3640,($0|0),0,(348232|0))|0); //@line 4027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $27 = ___cxa_find_matching_catch_2()|0;
  $28 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 229 "modules/image/wrap_ImageData.cpp"
  $$0 = $28;$$01 = $27;
  ___resumeException($$01|0);
  // unreachable;
 }
 ;HEAP32[$filename>>2]=HEAP32[$7>>2]|0;HEAP32[$filename+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$filename+8>>2]=HEAP32[$7+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 229 "modules/image/wrap_ImageData.cpp"
 __THREW__ = 0;
 $10 = (invoke_iii(3358,($L|0),3)|0); //@line 230 "modules/image/wrap_ImageData.cpp"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 L9: do {
  if ($12) {
   label = 14;
  } else {
   $13 = ($10|0)<(1); //@line 230 "modules/image/wrap_ImageData.cpp"
   do {
    if ($13) {
     $hasfilename$0$off0 = 0;
    } else {
     __THREW__ = 0;
     invoke_viii(3403,($1|0),($L|0),3); //@line 233 "modules/image/wrap_ImageData.cpp"
     $14 = __THREW__; __THREW__ = 0;
     $15 = $14&1;
     if ($15) {
      label = 14;
      break L9;
     }
     $16 = HEAP8[$filename>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $17 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $18 = ($17<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($18) {
      $22 = ((($filename)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$22>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$filename>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     } else {
      $19 = ((($filename)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $20 = HEAP32[$19>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$20>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $21 = ((($filename)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP32[$21>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     }
     __THREW__ = 0;
     invoke_vii(3368,($filename|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $23 = __THREW__; __THREW__ = 0;
     $24 = $23&1;
     if ($24) {
      $25 = ___cxa_find_matching_catch_3(0|0)|0;
      $26 = tempRet0;
      ___clang_call_terminate($25); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     } else {
      ;HEAP32[$filename>>2]=HEAP32[$1>>2]|0;HEAP32[$filename+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$filename+8>>2]=HEAP32[$1+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 233 "modules/image/wrap_ImageData.cpp"
      $hasfilename$0$off0 = 1;
      break;
     }
    }
   } while(0);
   $31 = HEAP32[$2>>2]|0; //@line 237 "modules/image/wrap_ImageData.cpp"
   $32 = ((($31)) + 24|0); //@line 237 "modules/image/wrap_ImageData.cpp"
   $33 = HEAP32[$32>>2]|0; //@line 237 "modules/image/wrap_ImageData.cpp"
   $34 = HEAP32[$format>>2]|0; //@line 237 "modules/image/wrap_ImageData.cpp"
   $35 = HEAP8[$filename>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $36 = $35 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $37 = ($36<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $38 = ((($filename)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $39 = HEAP32[$38>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $40 = ((($filename)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $41 = $37 ? $40 : $39; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   $42 = (invoke_iiii($33|0,($2|0),($34|0),($41|0))|0); //@line 237 "modules/image/wrap_ImageData.cpp"
   $43 = __THREW__; __THREW__ = 0;
   $44 = $43&1;
   L22: do {
    if ($44) {
     $45 = ___cxa_find_matching_catch_3(8992|0)|0;
     $46 = tempRet0;
     $47 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
     $48 = ($46|0)==($47|0); //@line 535 "./common/runtime.h"
     do {
      if ($48) {
       $49 = (___cxa_begin_catch(($45|0))|0); //@line 535 "./common/runtime.h"
       $50 = HEAP32[$49>>2]|0; //@line 539 "./common/runtime.h"
       $51 = ((($50)) + 8|0); //@line 539 "./common/runtime.h"
       $52 = HEAP32[$51>>2]|0; //@line 539 "./common/runtime.h"
       $53 = (FUNCTION_TABLE_ii[$52 & 8191]($49)|0); //@line 539 "./common/runtime.h"
       __THREW__ = 0;
       invoke_vii(3380,($L|0),($53|0)); //@line 539 "./common/runtime.h"
       $54 = __THREW__; __THREW__ = 0;
       $55 = $54&1;
       if ($55) {
        $63 = ___cxa_find_matching_catch_2()|0;
        $64 = tempRet0;
        __THREW__ = 0;
        invoke_v(3378); //@line 540 "./common/runtime.h"
        $65 = __THREW__; __THREW__ = 0;
        $66 = $65&1;
        if (!($66)) {
         $$01$i = $64;$$02$i = $63;
         break;
        }
        $67 = ___cxa_find_matching_catch_3(0|0)|0;
        $68 = tempRet0;
        ___clang_call_terminate($67); //@line 540 "./common/runtime.h"
        // unreachable; //@line 540 "./common/runtime.h"
       }
       __THREW__ = 0;
       invoke_v(3378); //@line 540 "./common/runtime.h"
       $56 = __THREW__; __THREW__ = 0;
       $57 = $56&1;
       if ($57) {
        label = 14;
        break L9;
       }
       __THREW__ = 0;
       $58 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
       $59 = __THREW__; __THREW__ = 0;
       $60 = $59&1;
       if ($60) {
        label = 14;
        break L9;
       }
       __THREW__ = 0;
       HEAP32[$vararg_buffer1>>2] = $58; //@line 543 "./common/runtime.h"
       (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer1|0))|0); //@line 543 "./common/runtime.h"
       $61 = __THREW__; __THREW__ = 0;
       $62 = $61&1;
       if ($62) {
        label = 14;
        break L9;
       } else {
        $filedata$0 = 0;
        break L22;
       }
      } else {
       $$01$i = $46;$$02$i = $45;
      }
     } while(0);
     $eh$lpad$body$index5Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
     break L9;
    } else {
     $filedata$0 = $42;
    }
   } while(0);
   __THREW__ = 0;
   invoke_viii(3387,($L|0),7,($filedata$0|0)); //@line 239 "modules/image/wrap_ImageData.cpp"
   $69 = __THREW__; __THREW__ = 0;
   $70 = $69&1;
   if ($70) {
    label = 14;
   } else {
    $71 = HEAP32[$filedata$0>>2]|0; //@line 240 "modules/image/wrap_ImageData.cpp"
    $72 = ((($71)) + 12|0); //@line 240 "modules/image/wrap_ImageData.cpp"
    $73 = HEAP32[$72>>2]|0; //@line 240 "modules/image/wrap_ImageData.cpp"
    __THREW__ = 0;
    invoke_vi($73|0,($filedata$0|0)); //@line 240 "modules/image/wrap_ImageData.cpp"
    $74 = __THREW__; __THREW__ = 0;
    $75 = $74&1;
    if ($75) {
     label = 14;
    } else {
     if ($hasfilename$0$off0) {
      __THREW__ = 0;
      (invoke_iiii(3551,($L|0),(354839|0),(521654|0))|0); //@line 244 "modules/image/wrap_ImageData.cpp"
      $76 = __THREW__; __THREW__ = 0;
      $77 = $76&1;
      if ($77) {
       label = 14;
       break;
      }
      __THREW__ = 0;
      invoke_vii(3454,($L|0),3); //@line 245 "modules/image/wrap_ImageData.cpp"
      $78 = __THREW__; __THREW__ = 0;
      $79 = $78&1;
      if ($79) {
       label = 14;
       break;
      }
      __THREW__ = 0;
      invoke_vii(3454,($L|0),-3); //@line 246 "modules/image/wrap_ImageData.cpp"
      $80 = __THREW__; __THREW__ = 0;
      $81 = $80&1;
      if ($81) {
       label = 14;
       break;
      }
      __THREW__ = 0;
      invoke_viii(3641,($L|0),2,0); //@line 247 "modules/image/wrap_ImageData.cpp"
      $82 = __THREW__; __THREW__ = 0;
      $83 = $82&1;
      if ($83) {
       label = 14;
       break;
      }
     }
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($filename); //@line 251 "modules/image/wrap_ImageData.cpp"
     $$02 = 1;
     STACKTOP = sp;return ($$02|0); //@line 251 "modules/image/wrap_ImageData.cpp"
    }
   }
  }
 } while(0);
 if ((label|0) == 14) {
  $29 = ___cxa_find_matching_catch_2()|0;
  $30 = tempRet0;
  $eh$lpad$body$index5Z2D = $30;$eh$lpad$body$indexZ2D = $29;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($filename); //@line 251 "modules/image/wrap_ImageData.cpp"
 $$0 = $eh$lpad$body$index5Z2D;$$01 = $eh$lpad$body$indexZ2D;
 ___resumeException($$01|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love5image26w_ImageData__performAtomicEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lock = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = (__ZN4love14luax_checktypeINS_5image9ImageDataEEEPT_P9lua_StateiNS_4TypeE($L,1,22)|0); //@line 43 "modules/image/wrap_ImageData.cpp"
 $1 = (__ZNK4love5image9ImageData8getMutexEv($0)|0); //@line 259 "modules/image/wrap_ImageData.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$1); //@line 259 "modules/image/wrap_ImageData.cpp"
 __THREW__ = 0;
 $2 = (invoke_ii(3385,($L|0))|0); //@line 261 "modules/image/wrap_ImageData.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $11 = ___cxa_find_matching_catch_2()|0;
  $12 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 262 "modules/image/wrap_ImageData.cpp"
  ___resumeException($11|0);
  // unreachable;
 }
 $5 = (($2) + -2)|0; //@line 261 "modules/image/wrap_ImageData.cpp"
 __THREW__ = 0;
 $6 = (invoke_iiiii(3552,($L|0),($5|0),-1,0)|0); //@line 261 "modules/image/wrap_ImageData.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $11 = ___cxa_find_matching_catch_2()|0;
  $12 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 262 "modules/image/wrap_ImageData.cpp"
  ___resumeException($11|0);
  // unreachable;
 }
 __ZN4love6thread4LockD2Ev($lock); //@line 262 "modules/image/wrap_ImageData.cpp"
 $9 = ($6|0)==(0); //@line 265 "modules/image/wrap_ImageData.cpp"
 if ($9) {
  $13 = (_lua_gettop($L)|0); //@line 270 "modules/image/wrap_ImageData.cpp"
  $14 = (($13) + -1)|0; //@line 270 "modules/image/wrap_ImageData.cpp"
  $$0 = $14;
  STACKTOP = sp;return ($$0|0); //@line 271 "modules/image/wrap_ImageData.cpp"
 } else {
  $10 = (_lua_error($L)|0); //@line 266 "modules/image/wrap_ImageData.cpp"
  $$0 = $10;
  STACKTOP = sp;return ($$0|0); //@line 271 "modules/image/wrap_ImageData.cpp"
 }
 return (0)|0;
}
function _luaopen_imagedata($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 314 "modules/image/wrap_ImageData.cpp"
 HEAP32[$vararg_buffer>>2] = 12272; //@line 316 "modules/image/wrap_ImageData.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 316 "modules/image/wrap_ImageData.cpp"
 HEAP32[$vararg_ptr1>>2] = 16296; //@line 316 "modules/image/wrap_ImageData.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 316 "modules/image/wrap_ImageData.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 316 "modules/image/wrap_ImageData.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,22,348239,$vararg_buffer)|0); //@line 316 "modules/image/wrap_ImageData.cpp"
 __ZN4love21luax_gettypemetatableEP9lua_StateNS_4TypeE($L,22); //@line 318 "modules/image/wrap_ImageData.cpp"
 $1 = (_lua_type($L,-1)|0); //@line 322 "modules/image/wrap_ImageData.cpp"
 $2 = ($1|0)==(5); //@line 322 "modules/image/wrap_ImageData.cpp"
 if (!($2)) {
  _lua_settop($L,-2); //@line 331 "modules/image/wrap_ImageData.cpp"
  STACKTOP = sp;return ($0|0); //@line 333 "modules/image/wrap_ImageData.cpp"
 }
 (_luaL_loadbuffer($L,348319,6366,354685)|0); //@line 324 "modules/image/wrap_ImageData.cpp"
 _lua_pushvalue($L,-2); //@line 325 "modules/image/wrap_ImageData.cpp"
 _lua_pushlightuserdata($L,16376); //@line 326 "modules/image/wrap_ImageData.cpp"
 _lua_call($L,2,0); //@line 327 "modules/image/wrap_ImageData.cpp"
 _lua_settop($L,-2); //@line 331 "modules/image/wrap_ImageData.cpp"
 STACKTOP = sp;return ($0|0); //@line 333 "modules/image/wrap_ImageData.cpp"
}
function __ZN4love5image3__38__invokeEPNS_5ProxyE($p) {
 $p = $p|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($p)) + 4|0); //@line 286 "modules/image/wrap_ImageData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 286 "modules/image/wrap_ImageData.cpp"
 $2 = (__ZNK4love5image9ImageData8getMutexEv($1)|0); //@line 287 "modules/image/wrap_ImageData.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 287 "modules/image/wrap_ImageData.cpp"
 $4 = ((($3)) + 8|0); //@line 287 "modules/image/wrap_ImageData.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 287 "modules/image/wrap_ImageData.cpp"
 FUNCTION_TABLE_vi[$5 & 8191]($2); //@line 287 "modules/image/wrap_ImageData.cpp"
 return; //@line 282 "modules/image/wrap_ImageData.cpp"
}
function __ZN4love5image3__48__invokeEPNS_5ProxyE($p) {
 $p = $p|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($p)) + 4|0); //@line 292 "modules/image/wrap_ImageData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 292 "modules/image/wrap_ImageData.cpp"
 $2 = (__ZNK4love5image9ImageData8getMutexEv($1)|0); //@line 293 "modules/image/wrap_ImageData.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 293 "modules/image/wrap_ImageData.cpp"
 $4 = ((($3)) + 12|0); //@line 293 "modules/image/wrap_ImageData.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 293 "modules/image/wrap_ImageData.cpp"
 FUNCTION_TABLE_vi[$5 & 8191]($2); //@line 293 "modules/image/wrap_ImageData.cpp"
 return; //@line 290 "modules/image/wrap_ImageData.cpp"
}
function __ZN4love8joystick14w_getJoysticksEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625632)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 37 "modules/joystick/wrap_JoystickModule.cpp"
 $2 = ((($1)) + 44|0); //@line 37 "modules/joystick/wrap_JoystickModule.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 37 "modules/joystick/wrap_JoystickModule.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 37 "modules/joystick/wrap_JoystickModule.cpp"
 _lua_createtable($L,$4,0); //@line 38 "modules/joystick/wrap_JoystickModule.cpp"
 $5 = ($4|0)>(0); //@line 40 "modules/joystick/wrap_JoystickModule.cpp"
 if ($5) {
  $i$01 = 0;
 } else {
  return 1; //@line 47 "modules/joystick/wrap_JoystickModule.cpp"
 }
 while(1) {
  $6 = HEAP32[(625632)>>2]|0; //@line 98 "./common/Module.h"
  $7 = HEAP32[$6>>2]|0; //@line 42 "modules/joystick/wrap_JoystickModule.cpp"
  $8 = ((($7)) + 36|0); //@line 42 "modules/joystick/wrap_JoystickModule.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 42 "modules/joystick/wrap_JoystickModule.cpp"
  $10 = (FUNCTION_TABLE_iii[$9 & 8191]($6,$i$01)|0); //@line 42 "modules/joystick/wrap_JoystickModule.cpp"
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,24,$10); //@line 43 "modules/joystick/wrap_JoystickModule.cpp"
  $11 = (($i$01) + 1)|0; //@line 44 "modules/joystick/wrap_JoystickModule.cpp"
  _lua_rawseti($L,-2,$11); //@line 44 "modules/joystick/wrap_JoystickModule.cpp"
  $exitcond = ($11|0)==($4|0); //@line 40 "modules/joystick/wrap_JoystickModule.cpp"
  if ($exitcond) {
   break;
  } else {
   $i$01 = $11;
  }
 }
 return 1; //@line 47 "modules/joystick/wrap_JoystickModule.cpp"
}
function __ZN4love8joystick10w_getIndexEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love8joystick18luax_checkjoystickEP9lua_Statei($L,1)|0); //@line 52 "modules/joystick/wrap_JoystickModule.cpp"
 $1 = HEAP32[(625632)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$1>>2]|0; //@line 53 "modules/joystick/wrap_JoystickModule.cpp"
 $3 = ((($2)) + 40|0); //@line 53 "modules/joystick/wrap_JoystickModule.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 53 "modules/joystick/wrap_JoystickModule.cpp"
 $5 = (FUNCTION_TABLE_iii[$4 & 8191]($1,$0)|0); //@line 53 "modules/joystick/wrap_JoystickModule.cpp"
 $6 = ($5|0)>(-1); //@line 54 "modules/joystick/wrap_JoystickModule.cpp"
 if ($6) {
  $7 = (($5) + 1)|0; //@line 55 "modules/joystick/wrap_JoystickModule.cpp"
  _lua_pushinteger($L,$7); //@line 55 "modules/joystick/wrap_JoystickModule.cpp"
  return 1; //@line 58 "modules/joystick/wrap_JoystickModule.cpp"
 } else {
  _lua_pushnil($L); //@line 57 "modules/joystick/wrap_JoystickModule.cpp"
  return 1; //@line 58 "modules/joystick/wrap_JoystickModule.cpp"
 }
 return (0)|0;
}
function __ZN4love8joystick18w_getJoystickCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625632)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 63 "modules/joystick/wrap_JoystickModule.cpp"
 $2 = ((($1)) + 44|0); //@line 63 "modules/joystick/wrap_JoystickModule.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 63 "modules/joystick/wrap_JoystickModule.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 63 "modules/joystick/wrap_JoystickModule.cpp"
 _lua_pushinteger($L,$4); //@line 63 "modules/joystick/wrap_JoystickModule.cpp"
 return 1; //@line 64 "modules/joystick/wrap_JoystickModule.cpp"
}
function __ZN4love8joystick19w_setGamepadMappingEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $$1 = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index14Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $gpinput = 0, $jinput = 0, $jinput$byval_copy = 0, $success$0 = 0, $tmpcast$i$i$byval_copy = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer10 = 0, $vararg_buffer4 = 0, $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $jinput$byval_copy = sp + 88|0;
 $tmpcast$i$i$byval_copy = sp + 80|0;
 $vararg_buffer10 = sp + 48|0;
 $vararg_buffer7 = sp + 40|0;
 $vararg_buffer4 = sp + 32|0;
 $vararg_buffer1 = sp + 24|0;
 $vararg_buffer = sp + 16|0;
 $0 = sp + 64|0;
 $1 = sp + 8|0;
 $gpinput = sp;
 $jinput = sp + 52|0;
 $2 = (_luaL_checklstring($L,1,0)|0); //@line 72 "modules/joystick/wrap_JoystickModule.cpp"
 $3 = (_luaL_checklstring($L,2,0)|0); //@line 74 "modules/joystick/wrap_JoystickModule.cpp"
 $4 = ((($gpinput)) + 4|0); //@line 77 "modules/joystick/wrap_JoystickModule.cpp"
 $5 = (__ZN4love8joystick8Joystick11getConstantEPKcRNS1_11GamepadAxisE($3,$4)|0); //@line 77 "modules/joystick/wrap_JoystickModule.cpp"
 do {
  if ($5) {
   HEAP32[$gpinput>>2] = 0; //@line 78 "modules/joystick/wrap_JoystickModule.cpp"
  } else {
   $6 = (__ZN4love8joystick8Joystick11getConstantEPKcRNS1_13GamepadButtonE($3,$4)|0); //@line 79 "modules/joystick/wrap_JoystickModule.cpp"
   if ($6) {
    HEAP32[$gpinput>>2] = 1; //@line 80 "modules/joystick/wrap_JoystickModule.cpp"
    break;
   }
   HEAP32[$vararg_buffer>>2] = $3; //@line 82 "modules/joystick/wrap_JoystickModule.cpp"
   $7 = (_luaL_error($L,354699,$vararg_buffer)|0); //@line 82 "modules/joystick/wrap_JoystickModule.cpp"
   $$2 = $7;
   STACKTOP = sp;return ($$2|0); //@line 115 "modules/joystick/wrap_JoystickModule.cpp"
  }
 } while(0);
 $8 = (_luaL_checklstring($L,3,0)|0); //@line 84 "modules/joystick/wrap_JoystickModule.cpp"
 $9 = (__ZN4love8joystick8Joystick11getConstantEPKcRNS1_9InputTypeE($8,$jinput)|0); //@line 87 "modules/joystick/wrap_JoystickModule.cpp"
 L9: do {
  if ($9) {
   $11 = HEAP32[$jinput>>2]|0; //@line 91 "modules/joystick/wrap_JoystickModule.cpp"
   switch ($11|0) {
   case 0:  {
    $12 = (+_luaL_checknumber($L,4)); //@line 94 "modules/joystick/wrap_JoystickModule.cpp"
    $13 = (~~(($12))); //@line 94 "modules/joystick/wrap_JoystickModule.cpp"
    $14 = (($13) + -1)|0; //@line 94 "modules/joystick/wrap_JoystickModule.cpp"
    $15 = ((($jinput)) + 4|0); //@line 94 "modules/joystick/wrap_JoystickModule.cpp"
    HEAP32[$15>>2] = $14; //@line 94 "modules/joystick/wrap_JoystickModule.cpp"
    break;
   }
   case 1:  {
    $16 = (+_luaL_checknumber($L,4)); //@line 97 "modules/joystick/wrap_JoystickModule.cpp"
    $17 = (~~(($16))); //@line 97 "modules/joystick/wrap_JoystickModule.cpp"
    $18 = (($17) + -1)|0; //@line 97 "modules/joystick/wrap_JoystickModule.cpp"
    $19 = ((($jinput)) + 4|0); //@line 97 "modules/joystick/wrap_JoystickModule.cpp"
    HEAP32[$19>>2] = $18; //@line 97 "modules/joystick/wrap_JoystickModule.cpp"
    break;
   }
   case 2:  {
    $20 = (+_luaL_checknumber($L,4)); //@line 101 "modules/joystick/wrap_JoystickModule.cpp"
    $21 = (~~(($20))); //@line 101 "modules/joystick/wrap_JoystickModule.cpp"
    $22 = (($21) + -1)|0; //@line 101 "modules/joystick/wrap_JoystickModule.cpp"
    $23 = ((($jinput)) + 4|0); //@line 101 "modules/joystick/wrap_JoystickModule.cpp"
    HEAP32[$23>>2] = $22; //@line 101 "modules/joystick/wrap_JoystickModule.cpp"
    $24 = (_luaL_checklstring($L,5,0)|0); //@line 102 "modules/joystick/wrap_JoystickModule.cpp"
    $25 = ((($jinput)) + 8|0); //@line 103 "modules/joystick/wrap_JoystickModule.cpp"
    $26 = (__ZN4love8joystick8Joystick11getConstantEPKcRNS1_3HatE($24,$25)|0); //@line 103 "modules/joystick/wrap_JoystickModule.cpp"
    if (!($26)) {
     HEAP32[$vararg_buffer4>>2] = $24; //@line 104 "modules/joystick/wrap_JoystickModule.cpp"
     $27 = (_luaL_error($L,354763,$vararg_buffer4)|0); //@line 104 "modules/joystick/wrap_JoystickModule.cpp"
     $$1 = $27;
     break L9;
    }
    break;
   }
   default: {
    HEAP32[$vararg_buffer7>>2] = $8; //@line 107 "modules/joystick/wrap_JoystickModule.cpp"
    $28 = (_luaL_error($L,354731,$vararg_buffer7)|0); //@line 107 "modules/joystick/wrap_JoystickModule.cpp"
    $$1 = $28;
    break L9;
   }
   }
   $29 = HEAP32[(625632)>>2]|0; //@line 98 "./common/Module.h"
   $30 = HEAP32[$29>>2]|0; //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
   $31 = ((($30)) + 48|0); //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
   $32 = HEAP32[$31>>2]|0; //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
   $33 = (_strlen($2)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   invoke_viii(3460,($0|0),($2|0),($33|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $34 = __THREW__; __THREW__ = 0;
   $35 = $34&1;
   do {
    if ($35) {
     $51 = ___cxa_find_matching_catch_3(8992|0)|0;
     $52 = tempRet0;
     $eh$lpad$body$i$index14Z2D = $52;$eh$lpad$body$i$indexZ2D = $51;
     label = 19;
    } else {
     $36 = $gpinput; //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
     $37 = $36; //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
     $38 = HEAP32[$37>>2]|0; //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
     $39 = (($36) + 4)|0; //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
     $40 = $39; //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
     $41 = HEAP32[$40>>2]|0; //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
     $42 = $1; //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
     $43 = $42; //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
     HEAP32[$43>>2] = $38; //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
     $44 = (($42) + 4)|0; //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
     $45 = $44; //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
     HEAP32[$45>>2] = $41; //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
     __THREW__ = 0;
     ;HEAP32[$tmpcast$i$i$byval_copy>>2]=HEAP32[$1>>2]|0;HEAP32[$tmpcast$i$i$byval_copy+4>>2]=HEAP32[$1+4>>2]|0; //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
     ;HEAP32[$jinput$byval_copy>>2]=HEAP32[$jinput>>2]|0;HEAP32[$jinput$byval_copy+4>>2]=HEAP32[$jinput+4>>2]|0;HEAP32[$jinput$byval_copy+8>>2]=HEAP32[$jinput+8>>2]|0; //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
     $46 = (invoke_iiiii($32|0,($29|0),($0|0),($tmpcast$i$i$byval_copy|0),($jinput$byval_copy|0))|0); //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
     $47 = __THREW__; __THREW__ = 0;
     $48 = $47&1;
     if ($48) {
      $49 = ___cxa_find_matching_catch_3(8992|0)|0;
      $50 = tempRet0;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
      $eh$lpad$body$i$index14Z2D = $50;$eh$lpad$body$i$indexZ2D = $49;
      label = 19;
      break;
     } else {
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 111 "modules/joystick/wrap_JoystickModule.cpp"
      $success$0 = $46;
      break;
     }
    }
   } while(0);
   do {
    if ((label|0) == 19) {
     $53 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
     $54 = ($eh$lpad$body$i$index14Z2D|0)==($53|0); //@line 535 "./common/runtime.h"
     if (!($54)) {
      $$01$i = $eh$lpad$body$i$index14Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
      ___resumeException($$02$i|0);
      // unreachable;
     }
     $55 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
     $56 = HEAP32[$55>>2]|0; //@line 539 "./common/runtime.h"
     $57 = ((($56)) + 8|0); //@line 539 "./common/runtime.h"
     $58 = HEAP32[$57>>2]|0; //@line 539 "./common/runtime.h"
     $59 = (FUNCTION_TABLE_ii[$58 & 8191]($55)|0); //@line 539 "./common/runtime.h"
     __THREW__ = 0;
     invoke_vii(3380,($L|0),($59|0)); //@line 539 "./common/runtime.h"
     $60 = __THREW__; __THREW__ = 0;
     $61 = $60&1;
     if (!($61)) {
      ___cxa_end_catch(); //@line 540 "./common/runtime.h"
      $62 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
      HEAP32[$vararg_buffer10>>2] = $62; //@line 543 "./common/runtime.h"
      (_luaL_error($L,595867,$vararg_buffer10)|0); //@line 543 "./common/runtime.h"
      $success$0 = 0;
      break;
     }
     $63 = ___cxa_find_matching_catch_2()|0;
     $64 = tempRet0;
     __THREW__ = 0;
     invoke_v(3378); //@line 540 "./common/runtime.h"
     $65 = __THREW__; __THREW__ = 0;
     $66 = $65&1;
     if ($66) {
      $67 = ___cxa_find_matching_catch_3(0|0)|0;
      $68 = tempRet0;
      ___clang_call_terminate($67); //@line 540 "./common/runtime.h"
      // unreachable; //@line 540 "./common/runtime.h"
     } else {
      $$01$i = $64;$$02$i = $63;
      ___resumeException($$02$i|0);
      // unreachable;
     }
    }
   } while(0);
   __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$success$0); //@line 113 "modules/joystick/wrap_JoystickModule.cpp"
   $$1 = 1;
  } else {
   HEAP32[$vararg_buffer1>>2] = $8; //@line 88 "modules/joystick/wrap_JoystickModule.cpp"
   $10 = (_luaL_error($L,354731,$vararg_buffer1)|0); //@line 88 "modules/joystick/wrap_JoystickModule.cpp"
   $$1 = $10;
  }
 } while(0);
 $$2 = $$1;
 STACKTOP = sp;return ($$2|0); //@line 115 "modules/joystick/wrap_JoystickModule.cpp"
}
function __ZN4love8joystick19w_getGamepadMappingEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$02$i = 0, $$1 = 0, $$3 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index9Z2D = 0, $eh$lpad$body$indexZ2D = 0, $gpinput = 0, $guid = 0, $hatstr = 0, $inputtypestr = 0, $tmpcast$i$i$byval_copy = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, $vararg_buffer6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tmpcast$i$i$byval_copy = sp + 104|0;
 $vararg_buffer6 = sp + 40|0;
 $vararg_buffer4 = sp + 32|0;
 $vararg_buffer1 = sp + 24|0;
 $vararg_buffer = sp + 16|0;
 $0 = sp + 88|0;
 $1 = sp + 8|0;
 $guid = sp + 76|0;
 $2 = sp + 64|0;
 $3 = sp + 52|0;
 $gpinput = sp;
 $inputtypestr = sp + 48|0;
 $hatstr = sp + 44|0;
 ;HEAP32[$guid>>2]=0|0;HEAP32[$guid+4>>2]=0|0;HEAP32[$guid+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 $4 = (invoke_iii(3358,($L|0),1)|0); //@line 123 "modules/joystick/wrap_JoystickModule.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 L1: do {
  if (!($6)) {
   $7 = ($4|0)==(4); //@line 123 "modules/joystick/wrap_JoystickModule.cpp"
   do {
    if ($7) {
     __THREW__ = 0;
     invoke_viii(3403,($2|0),($L|0),1); //@line 124 "modules/joystick/wrap_JoystickModule.cpp"
     $8 = __THREW__; __THREW__ = 0;
     $9 = $8&1;
     if ($9) {
      break L1;
     }
     $10 = HEAP8[$guid>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $11 = $10 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $12 = ($11<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($12) {
      $16 = ((($guid)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$16>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$guid>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     } else {
      $13 = ((($guid)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $14 = HEAP32[$13>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$14>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $15 = ((($guid)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP32[$15>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     }
     __THREW__ = 0;
     invoke_vii(3368,($guid|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $17 = __THREW__; __THREW__ = 0;
     $18 = $17&1;
     if ($18) {
      $19 = ___cxa_find_matching_catch_3(0|0)|0;
      $20 = tempRet0;
      ___clang_call_terminate($19); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     } else {
      ;HEAP32[$guid>>2]=HEAP32[$2>>2]|0;HEAP32[$guid+4>>2]=HEAP32[$2+4>>2]|0;HEAP32[$guid+8>>2]=HEAP32[$2+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      ;HEAP32[$2>>2]=0|0;HEAP32[$2+4>>2]=0|0;HEAP32[$2+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 124 "modules/joystick/wrap_JoystickModule.cpp"
      break;
     }
    } else {
     __THREW__ = 0;
     $23 = (invoke_iii(3642,($L|0),1)|0); //@line 127 "modules/joystick/wrap_JoystickModule.cpp"
     $24 = __THREW__; __THREW__ = 0;
     $25 = $24&1;
     if ($25) {
      break L1;
     }
     $26 = HEAP32[$23>>2]|0; //@line 128 "modules/joystick/wrap_JoystickModule.cpp"
     $27 = ((($26)) + 80|0); //@line 128 "modules/joystick/wrap_JoystickModule.cpp"
     $28 = HEAP32[$27>>2]|0; //@line 128 "modules/joystick/wrap_JoystickModule.cpp"
     __THREW__ = 0;
     invoke_vii($28|0,($3|0),($23|0)); //@line 128 "modules/joystick/wrap_JoystickModule.cpp"
     $29 = __THREW__; __THREW__ = 0;
     $30 = $29&1;
     if ($30) {
      break L1;
     }
     $31 = HEAP8[$guid>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $32 = $31 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $33 = ($32<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($33) {
      $37 = ((($guid)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$37>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$guid>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     } else {
      $34 = ((($guid)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $35 = HEAP32[$34>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$35>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $36 = ((($guid)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP32[$36>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     }
     __THREW__ = 0;
     invoke_vii(3368,($guid|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $38 = __THREW__; __THREW__ = 0;
     $39 = $38&1;
     if ($39) {
      $40 = ___cxa_find_matching_catch_3(0|0)|0;
      $41 = tempRet0;
      ___clang_call_terminate($40); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     } else {
      ;HEAP32[$guid>>2]=HEAP32[$3>>2]|0;HEAP32[$guid+4>>2]=HEAP32[$3+4>>2]|0;HEAP32[$guid+8>>2]=HEAP32[$3+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      ;HEAP32[$3>>2]=0|0;HEAP32[$3+4>>2]=0|0;HEAP32[$3+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 128 "modules/joystick/wrap_JoystickModule.cpp"
      break;
     }
    }
   } while(0);
   __THREW__ = 0;
   $42 = (invoke_iiii(3359,($L|0),2,(0|0))|0); //@line 131 "modules/joystick/wrap_JoystickModule.cpp"
   $43 = __THREW__; __THREW__ = 0;
   $44 = $43&1;
   if (!($44)) {
    $45 = ((($gpinput)) + 4|0); //@line 134 "modules/joystick/wrap_JoystickModule.cpp"
    __THREW__ = 0;
    $46 = (invoke_iii(3643,($42|0),($45|0))|0); //@line 134 "modules/joystick/wrap_JoystickModule.cpp"
    $47 = __THREW__; __THREW__ = 0;
    $48 = $47&1;
    if (!($48)) {
     do {
      if ($46) {
       HEAP32[$gpinput>>2] = 0; //@line 135 "modules/joystick/wrap_JoystickModule.cpp"
      } else {
       __THREW__ = 0;
       $49 = (invoke_iii(3644,($42|0),($45|0))|0); //@line 136 "modules/joystick/wrap_JoystickModule.cpp"
       $50 = __THREW__; __THREW__ = 0;
       $51 = $50&1;
       if ($51) {
        break L1;
       }
       if ($49) {
        HEAP32[$gpinput>>2] = 1; //@line 137 "modules/joystick/wrap_JoystickModule.cpp"
        break;
       }
       __THREW__ = 0;
       HEAP32[$vararg_buffer>>2] = $42; //@line 139 "modules/joystick/wrap_JoystickModule.cpp"
       $52 = (invoke_iiii(3389,($L|0),(354699|0),($vararg_buffer|0))|0); //@line 139 "modules/joystick/wrap_JoystickModule.cpp"
       $53 = __THREW__; __THREW__ = 0;
       $54 = $53&1;
       if ($54) {
        break L1;
       } else {
        $$3 = $52;
       }
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($guid); //@line 178 "modules/joystick/wrap_JoystickModule.cpp"
       STACKTOP = sp;return ($$3|0); //@line 178 "modules/joystick/wrap_JoystickModule.cpp"
      }
     } while(0);
     $55 = HEAP32[(625632)>>2]|0; //@line 98 "./common/Module.h"
     $56 = HEAP32[$55>>2]|0; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $57 = ((($56)) + 52|0); //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $58 = HEAP32[$57>>2]|0; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $59 = $gpinput; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $60 = $59; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $61 = HEAP32[$60>>2]|0; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $62 = (($59) + 4)|0; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $63 = $62; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $64 = HEAP32[$63>>2]|0; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $65 = $1; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $66 = $65; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     HEAP32[$66>>2] = $61; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $67 = (($65) + 4)|0; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $68 = $67; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     HEAP32[$68>>2] = $64; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     __THREW__ = 0;
     ;HEAP32[$tmpcast$i$i$byval_copy>>2]=HEAP32[$1>>2]|0;HEAP32[$tmpcast$i$i$byval_copy+4>>2]=HEAP32[$1+4>>2]|0; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     invoke_viiii($58|0,($0|0),($55|0),($guid|0),($tmpcast$i$i$byval_copy|0)); //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $69 = __THREW__; __THREW__ = 0;
     $70 = $69&1;
     if ($70) {
      $71 = ___cxa_find_matching_catch_3(8992|0)|0;
      $72 = tempRet0;
      $73 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
      $74 = ($72|0)==($73|0); //@line 535 "./common/runtime.h"
      do {
       if ($74) {
        $75 = (___cxa_begin_catch(($71|0))|0); //@line 535 "./common/runtime.h"
        $76 = HEAP32[$75>>2]|0; //@line 539 "./common/runtime.h"
        $77 = ((($76)) + 8|0); //@line 539 "./common/runtime.h"
        $78 = HEAP32[$77>>2]|0; //@line 539 "./common/runtime.h"
        $79 = (FUNCTION_TABLE_ii[$78 & 8191]($75)|0); //@line 539 "./common/runtime.h"
        __THREW__ = 0;
        invoke_vii(3380,($L|0),($79|0)); //@line 539 "./common/runtime.h"
        $80 = __THREW__; __THREW__ = 0;
        $81 = $80&1;
        if ($81) {
         $89 = ___cxa_find_matching_catch_2()|0;
         $90 = tempRet0;
         __THREW__ = 0;
         invoke_v(3378); //@line 540 "./common/runtime.h"
         $91 = __THREW__; __THREW__ = 0;
         $92 = $91&1;
         if (!($92)) {
          $$01$i = $90;$$02$i = $89;
          break;
         }
         $93 = ___cxa_find_matching_catch_3(0|0)|0;
         $94 = tempRet0;
         ___clang_call_terminate($93); //@line 540 "./common/runtime.h"
         // unreachable; //@line 540 "./common/runtime.h"
        }
        __THREW__ = 0;
        invoke_v(3378); //@line 540 "./common/runtime.h"
        $82 = __THREW__; __THREW__ = 0;
        $83 = $82&1;
        if ($83) {
         break L1;
        }
        __THREW__ = 0;
        $84 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
        $85 = __THREW__; __THREW__ = 0;
        $86 = $85&1;
        if ($86) {
         break L1;
        }
        __THREW__ = 0;
        HEAP32[$vararg_buffer1>>2] = $84; //@line 543 "./common/runtime.h"
        (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer1|0))|0); //@line 543 "./common/runtime.h"
        $87 = __THREW__; __THREW__ = 0;
        $88 = $87&1;
        if ($88) {
         break L1;
        } else {
         $$3 = 0;
        }
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($guid); //@line 178 "modules/joystick/wrap_JoystickModule.cpp"
        STACKTOP = sp;return ($$3|0); //@line 178 "modules/joystick/wrap_JoystickModule.cpp"
       } else {
        $$01$i = $72;$$02$i = $71;
       }
      } while(0);
      $eh$lpad$body$index9Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($guid); //@line 178 "modules/joystick/wrap_JoystickModule.cpp"
      ___resumeException($eh$lpad$body$indexZ2D|0);
      // unreachable;
     }
     $95 = HEAP32[$0>>2]|0; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $96 = ((($0)) + 4|0); //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $97 = HEAP32[$96>>2]|0; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $98 = ((($0)) + 8|0); //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $99 = HEAP32[$98>>2]|0; //@line 144 "modules/joystick/wrap_JoystickModule.cpp"
     $100 = ($95|0)==(3); //@line 146 "modules/joystick/wrap_JoystickModule.cpp"
     if ($100) {
      $$3 = 0;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($guid); //@line 178 "modules/joystick/wrap_JoystickModule.cpp"
      STACKTOP = sp;return ($$3|0); //@line 178 "modules/joystick/wrap_JoystickModule.cpp"
     }
     __THREW__ = 0;
     $101 = (invoke_iii(3645,($95|0),($inputtypestr|0))|0); //@line 150 "modules/joystick/wrap_JoystickModule.cpp"
     $102 = __THREW__; __THREW__ = 0;
     $103 = $102&1;
     if (!($103)) {
      if ($101) {
       $107 = HEAP32[$inputtypestr>>2]|0; //@line 153 "modules/joystick/wrap_JoystickModule.cpp"
       __THREW__ = 0;
       invoke_vii(3380,($L|0),($107|0)); //@line 153 "modules/joystick/wrap_JoystickModule.cpp"
       $108 = __THREW__; __THREW__ = 0;
       $109 = $108&1;
       if ($109) {
        break;
       }
       L56: do {
        switch ($95|0) {
        case 0:  {
         $110 = (($97) + 1)|0; //@line 159 "modules/joystick/wrap_JoystickModule.cpp"
         __THREW__ = 0;
         invoke_vii(3468,($L|0),($110|0)); //@line 159 "modules/joystick/wrap_JoystickModule.cpp"
         $111 = __THREW__; __THREW__ = 0;
         $112 = $111&1;
         if ($112) {
          break L1;
         } else {
          $$0 = 2;
         }
         break;
        }
        case 1:  {
         $113 = (($97) + 1)|0; //@line 162 "modules/joystick/wrap_JoystickModule.cpp"
         __THREW__ = 0;
         invoke_vii(3468,($L|0),($113|0)); //@line 162 "modules/joystick/wrap_JoystickModule.cpp"
         $114 = __THREW__; __THREW__ = 0;
         $115 = $114&1;
         if ($115) {
          break L1;
         } else {
          $$0 = 2;
         }
         break;
        }
        case 2:  {
         $116 = (($97) + 1)|0; //@line 165 "modules/joystick/wrap_JoystickModule.cpp"
         __THREW__ = 0;
         invoke_vii(3468,($L|0),($116|0)); //@line 165 "modules/joystick/wrap_JoystickModule.cpp"
         $117 = __THREW__; __THREW__ = 0;
         $118 = $117&1;
         if ($118) {
          break L1;
         }
         __THREW__ = 0;
         $119 = (invoke_iii(3646,($99|0),($hatstr|0))|0); //@line 166 "modules/joystick/wrap_JoystickModule.cpp"
         $120 = __THREW__; __THREW__ = 0;
         $121 = $120&1;
         if ($121) {
          break L1;
         }
         if ($119) {
          $122 = HEAP32[$hatstr>>2]|0; //@line 168 "modules/joystick/wrap_JoystickModule.cpp"
          __THREW__ = 0;
          invoke_vii(3380,($L|0),($122|0)); //@line 168 "modules/joystick/wrap_JoystickModule.cpp"
          $123 = __THREW__; __THREW__ = 0;
          $124 = $123&1;
          if ($124) {
           break L1;
          } else {
           $$0 = 3;
           break L56;
          }
         } else {
          __THREW__ = 0;
          $125 = (invoke_iiii(3389,($L|0),(354817|0),($vararg_buffer6|0))|0); //@line 172 "modules/joystick/wrap_JoystickModule.cpp"
          $126 = __THREW__; __THREW__ = 0;
          $127 = $126&1;
          if ($127) {
           break L1;
          } else {
           $$0 = $125;
           break L56;
          }
         }
         break;
        }
        default: {
         $$0 = 1;
        }
        }
       } while(0);
       $$1 = $$0;
      } else {
       __THREW__ = 0;
       $104 = (invoke_iiii(3389,($L|0),(354788|0),($vararg_buffer4|0))|0); //@line 151 "modules/joystick/wrap_JoystickModule.cpp"
       $105 = __THREW__; __THREW__ = 0;
       $106 = $105&1;
       if ($106) {
        break;
       } else {
        $$1 = $104;
       }
      }
      $$3 = $$1;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($guid); //@line 178 "modules/joystick/wrap_JoystickModule.cpp"
      STACKTOP = sp;return ($$3|0); //@line 178 "modules/joystick/wrap_JoystickModule.cpp"
     }
    }
   }
  }
 } while(0);
 $21 = ___cxa_find_matching_catch_2()|0;
 $22 = tempRet0;
 $eh$lpad$body$index9Z2D = $22;$eh$lpad$body$indexZ2D = $21;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($guid); //@line 178 "modules/joystick/wrap_JoystickModule.cpp"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8joystick21w_loadGamepadMappingsEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $mappings = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $mappings = sp + 28|0;
 $0 = sp + 16|0;
 $1 = sp + 4|0;
 ;HEAP32[$mappings>>2]=0|0;HEAP32[$mappings+4>>2]=0|0;HEAP32[$mappings+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 $2 = (invoke_iii(3453,($L|0),1)|0); //@line 185 "modules/joystick/wrap_JoystickModule.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 L1: do {
  if (!($4)) {
   $5 = ($2|0)==(0); //@line 185 "modules/joystick/wrap_JoystickModule.cpp"
   do {
    if ($5) {
     label = 10;
    } else {
     __THREW__ = 0;
     invoke_vii(3454,($L|0),1); //@line 187 "modules/joystick/wrap_JoystickModule.cpp"
     $6 = __THREW__; __THREW__ = 0;
     $7 = $6&1;
     if ($7) {
      break L1;
     }
     __THREW__ = 0;
     (invoke_iiiii(3457,($L|0),-1,(354839|0),(354850|0))|0); //@line 188 "modules/joystick/wrap_JoystickModule.cpp"
     $8 = __THREW__; __THREW__ = 0;
     $9 = $8&1;
     if ($9) {
      break L1;
     }
     __THREW__ = 0;
     $10 = (invoke_iii(3504,($L|0),-1)|0); //@line 189 "modules/joystick/wrap_JoystickModule.cpp"
     $11 = __THREW__; __THREW__ = 0;
     $12 = $11&1;
     if ($12) {
      break L1;
     }
     __THREW__ = 0;
     invoke_vii(3384,($L|0),-2); //@line 190 "modules/joystick/wrap_JoystickModule.cpp"
     $13 = __THREW__; __THREW__ = 0;
     $14 = $13&1;
     if ($14) {
      break L1;
     }
     if ($10) {
      label = 10;
     } else {
      __THREW__ = 0;
      invoke_viii(3403,($1|0),($L|0),1); //@line 201 "modules/joystick/wrap_JoystickModule.cpp"
      $50 = __THREW__; __THREW__ = 0;
      $51 = $50&1;
      if ($51) {
       break L1;
      }
      $52 = HEAP8[$mappings>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $53 = $52 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $54 = ($53<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($54) {
       $58 = ((($mappings)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       HEAP8[$58>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       HEAP8[$mappings>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      } else {
       $55 = ((($mappings)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $56 = HEAP32[$55>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       HEAP8[$56>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $57 = ((($mappings)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       HEAP32[$57>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      }
      __THREW__ = 0;
      invoke_vii(3368,($mappings|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if ($60) {
       $61 = ___cxa_find_matching_catch_3(0|0)|0;
       $62 = tempRet0;
       ___clang_call_terminate($61); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      } else {
       ;HEAP32[$mappings>>2]=HEAP32[$1>>2]|0;HEAP32[$mappings+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$mappings+8>>2]=HEAP32[$1+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 201 "modules/joystick/wrap_JoystickModule.cpp"
       break;
      }
     }
    }
   } while(0);
   do {
    if ((label|0) == 10) {
     __THREW__ = 0;
     $17 = (invoke_iii(3537,($L|0),1)|0); //@line 195 "modules/joystick/wrap_JoystickModule.cpp"
     $18 = __THREW__; __THREW__ = 0;
     $19 = $18&1;
     if ($19) {
      break L1;
     }
     $20 = HEAP32[$17>>2]|0; //@line 196 "modules/joystick/wrap_JoystickModule.cpp"
     $21 = ((($20)) + 16|0); //@line 196 "modules/joystick/wrap_JoystickModule.cpp"
     $22 = HEAP32[$21>>2]|0; //@line 196 "modules/joystick/wrap_JoystickModule.cpp"
     __THREW__ = 0;
     $23 = (invoke_ii($22|0,($17|0))|0); //@line 196 "modules/joystick/wrap_JoystickModule.cpp"
     $24 = __THREW__; __THREW__ = 0;
     $25 = $24&1;
     if ($25) {
      break L1;
     }
     $26 = HEAP32[$17>>2]|0; //@line 196 "modules/joystick/wrap_JoystickModule.cpp"
     $27 = ((($26)) + 20|0); //@line 196 "modules/joystick/wrap_JoystickModule.cpp"
     $28 = HEAP32[$27>>2]|0; //@line 196 "modules/joystick/wrap_JoystickModule.cpp"
     __THREW__ = 0;
     $29 = (invoke_ii($28|0,($17|0))|0); //@line 196 "modules/joystick/wrap_JoystickModule.cpp"
     $30 = __THREW__; __THREW__ = 0;
     $31 = $30&1;
     if ($31) {
      break L1;
     }
     __THREW__ = 0;
     invoke_viii(3460,($0|0),($23|0),($29|0)); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $32 = __THREW__; __THREW__ = 0;
     $33 = $32&1;
     if ($33) {
      break L1;
     }
     $34 = HEAP8[$mappings>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $35 = $34 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $36 = ($35<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($36) {
      $40 = ((($mappings)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$40>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$mappings>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     } else {
      $37 = ((($mappings)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $38 = HEAP32[$37>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$38>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $39 = ((($mappings)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP32[$39>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     }
     __THREW__ = 0;
     invoke_vii(3368,($mappings|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $41 = __THREW__; __THREW__ = 0;
     $42 = $41&1;
     if ($42) {
      $43 = ___cxa_find_matching_catch_3(0|0)|0;
      $44 = tempRet0;
      ___clang_call_terminate($43); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     } else {
      ;HEAP32[$mappings>>2]=HEAP32[$0>>2]|0;HEAP32[$mappings+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$mappings+8>>2]=HEAP32[$0+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 196 "modules/joystick/wrap_JoystickModule.cpp"
      $45 = HEAP32[$17>>2]|0; //@line 197 "modules/joystick/wrap_JoystickModule.cpp"
      $46 = ((($45)) + 12|0); //@line 197 "modules/joystick/wrap_JoystickModule.cpp"
      $47 = HEAP32[$46>>2]|0; //@line 197 "modules/joystick/wrap_JoystickModule.cpp"
      __THREW__ = 0;
      invoke_vi($47|0,($17|0)); //@line 197 "modules/joystick/wrap_JoystickModule.cpp"
      $48 = __THREW__; __THREW__ = 0;
      $49 = $48&1;
      if ($49) {
       break L1;
      } else {
       break;
      }
     }
    }
   } while(0);
   $63 = HEAP32[(625632)>>2]|0; //@line 98 "./common/Module.h"
   $64 = HEAP32[$63>>2]|0; //@line 203 "modules/joystick/wrap_JoystickModule.cpp"
   $65 = ((($64)) + 56|0); //@line 203 "modules/joystick/wrap_JoystickModule.cpp"
   $66 = HEAP32[$65>>2]|0; //@line 203 "modules/joystick/wrap_JoystickModule.cpp"
   __THREW__ = 0;
   invoke_vii($66|0,($63|0),($mappings|0)); //@line 203 "modules/joystick/wrap_JoystickModule.cpp"
   $67 = __THREW__; __THREW__ = 0;
   $68 = $67&1;
   if (!($68)) {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($mappings); //@line 205 "modules/joystick/wrap_JoystickModule.cpp"
    STACKTOP = sp;return 0; //@line 205 "modules/joystick/wrap_JoystickModule.cpp"
   }
   $69 = ___cxa_find_matching_catch_3(8992|0)|0;
   $70 = tempRet0;
   $71 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $72 = ($70|0)==($71|0); //@line 535 "./common/runtime.h"
   do {
    if ($72) {
     $73 = (___cxa_begin_catch(($69|0))|0); //@line 535 "./common/runtime.h"
     $74 = HEAP32[$73>>2]|0; //@line 539 "./common/runtime.h"
     $75 = ((($74)) + 8|0); //@line 539 "./common/runtime.h"
     $76 = HEAP32[$75>>2]|0; //@line 539 "./common/runtime.h"
     $77 = (FUNCTION_TABLE_ii[$76 & 8191]($73)|0); //@line 539 "./common/runtime.h"
     __THREW__ = 0;
     invoke_vii(3380,($L|0),($77|0)); //@line 539 "./common/runtime.h"
     $78 = __THREW__; __THREW__ = 0;
     $79 = $78&1;
     if ($79) {
      $87 = ___cxa_find_matching_catch_2()|0;
      $88 = tempRet0;
      __THREW__ = 0;
      invoke_v(3378); //@line 540 "./common/runtime.h"
      $89 = __THREW__; __THREW__ = 0;
      $90 = $89&1;
      if (!($90)) {
       $$01$i = $88;$$02$i = $87;
       break;
      }
      $91 = ___cxa_find_matching_catch_3(0|0)|0;
      $92 = tempRet0;
      ___clang_call_terminate($91); //@line 540 "./common/runtime.h"
      // unreachable; //@line 540 "./common/runtime.h"
     }
     __THREW__ = 0;
     invoke_v(3378); //@line 540 "./common/runtime.h"
     $80 = __THREW__; __THREW__ = 0;
     $81 = $80&1;
     if ($81) {
      break L1;
     }
     __THREW__ = 0;
     $82 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
     $83 = __THREW__; __THREW__ = 0;
     $84 = $83&1;
     if ($84) {
      break L1;
     }
     __THREW__ = 0;
     HEAP32[$vararg_buffer>>2] = $82; //@line 543 "./common/runtime.h"
     (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 543 "./common/runtime.h"
     $85 = __THREW__; __THREW__ = 0;
     $86 = $85&1;
     if ($86) {
      break L1;
     }
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($mappings); //@line 205 "modules/joystick/wrap_JoystickModule.cpp"
     STACKTOP = sp;return 0; //@line 205 "modules/joystick/wrap_JoystickModule.cpp"
    } else {
     $$01$i = $70;$$02$i = $69;
    }
   } while(0);
   $eh$lpad$body$index2Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($mappings); //@line 205 "modules/joystick/wrap_JoystickModule.cpp"
   ___resumeException($eh$lpad$body$indexZ2D|0);
   // unreachable;
  }
 } while(0);
 $15 = ___cxa_find_matching_catch_2()|0;
 $16 = tempRet0;
 $eh$lpad$body$index2Z2D = $16;$eh$lpad$body$indexZ2D = $15;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($mappings); //@line 205 "modules/joystick/wrap_JoystickModule.cpp"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8joystick21w_saveGamepadMappingsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $idxs = 0, $mappings = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $mappings = sp + 8|0;
 $idxs = sp;
 _lua_settop($L,1); //@line 209 "modules/joystick/wrap_JoystickModule.cpp"
 $0 = HEAP32[(625632)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 210 "modules/joystick/wrap_JoystickModule.cpp"
 $2 = ((($1)) + 60|0); //@line 210 "modules/joystick/wrap_JoystickModule.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 210 "modules/joystick/wrap_JoystickModule.cpp"
 FUNCTION_TABLE_vii[$3 & 8191]($mappings,$0); //@line 210 "modules/joystick/wrap_JoystickModule.cpp"
 __THREW__ = 0;
 $4 = (invoke_iii(3358,($L|0),1)|0); //@line 213 "modules/joystick/wrap_JoystickModule.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 do {
  if (!($6)) {
   $7 = ($4|0)<(1); //@line 213 "modules/joystick/wrap_JoystickModule.cpp"
   if (!($7)) {
    __THREW__ = 0;
    invoke_vii(3405,($L|0),($mappings|0)); //@line 215 "modules/joystick/wrap_JoystickModule.cpp"
    $8 = __THREW__; __THREW__ = 0;
    $9 = $8&1;
    if ($9) {
     break;
    }
    $10 = $idxs; //@line 216 "modules/joystick/wrap_JoystickModule.cpp"
    $11 = $10; //@line 216 "modules/joystick/wrap_JoystickModule.cpp"
    HEAP32[$11>>2] = 1; //@line 216 "modules/joystick/wrap_JoystickModule.cpp"
    $12 = (($10) + 4)|0; //@line 216 "modules/joystick/wrap_JoystickModule.cpp"
    $13 = $12; //@line 216 "modules/joystick/wrap_JoystickModule.cpp"
    HEAP32[$13>>2] = 2; //@line 216 "modules/joystick/wrap_JoystickModule.cpp"
    __THREW__ = 0;
    (invoke_iiiiii(3455,($L|0),($idxs|0),2,(354839|0),(521654|0))|0); //@line 217 "modules/joystick/wrap_JoystickModule.cpp"
    $14 = __THREW__; __THREW__ = 0;
    $15 = $14&1;
    if ($15) {
     break;
    }
    __THREW__ = 0;
    invoke_vii(3384,($L|0),-2); //@line 218 "modules/joystick/wrap_JoystickModule.cpp"
    $16 = __THREW__; __THREW__ = 0;
    $17 = $16&1;
    if ($17) {
     break;
    }
   }
   __THREW__ = 0;
   invoke_vii(3405,($L|0),($mappings|0)); //@line 222 "modules/joystick/wrap_JoystickModule.cpp"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($mappings); //@line 224 "modules/joystick/wrap_JoystickModule.cpp"
    STACKTOP = sp;return 1; //@line 224 "modules/joystick/wrap_JoystickModule.cpp"
   }
  }
 } while(0);
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($mappings); //@line 224 "modules/joystick/wrap_JoystickModule.cpp"
 ___resumeException($18|0);
 // unreachable;
 return (0)|0;
}
function _luaopen_love_joystick($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0;
 var $eh$lpad$body$i$indexZ2D = 0, $instance$sroa$0$1 = 0, $vararg_buffer = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $w = sp + 4|0;
 $0 = HEAP32[(625632)>>2]|0; //@line 98 "./common/Module.h"
 $1 = $0; //@line 246 "modules/joystick/wrap_JoystickModule.cpp"
 $2 = ($0|0)==(0|0); //@line 247 "modules/joystick/wrap_JoystickModule.cpp"
 L1: do {
  if ($2) {
   __THREW__ = 0;
   $3 = (invoke_ii(3348,44)|0); //@line 249 "modules/joystick/wrap_JoystickModule.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   do {
    if ($5) {
     $11 = ___cxa_find_matching_catch_3(8992|0)|0;
     $12 = tempRet0;
     $eh$lpad$body$i$index2Z2D = $12;$eh$lpad$body$i$indexZ2D = $11;
    } else {
     __THREW__ = 0;
     invoke_vi(3647,($3|0)); //@line 249 "modules/joystick/wrap_JoystickModule.cpp"
     $6 = __THREW__; __THREW__ = 0;
     $7 = $6&1;
     if ($7) {
      $8 = ___cxa_find_matching_catch_3(8992|0)|0;
      $9 = tempRet0;
      __ZdlPv($3); //@line 249 "modules/joystick/wrap_JoystickModule.cpp"
      $eh$lpad$body$i$index2Z2D = $9;$eh$lpad$body$i$indexZ2D = $8;
      break;
     } else {
      $10 = $3; //@line 249 "modules/joystick/wrap_JoystickModule.cpp"
      $instance$sroa$0$1 = $10;
      break L1;
     }
    }
   } while(0);
   $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $14 = ($eh$lpad$body$i$index2Z2D|0)==($13|0); //@line 535 "./common/runtime.h"
   if (!($14)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $15 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
   $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
   $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
   $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $instance$sroa$0$1 = $1;
    break;
   }
   $23 = ___cxa_find_matching_catch_2()|0;
   $24 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    $27 = ___cxa_find_matching_catch_3(0|0)|0;
    $28 = tempRet0;
    ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $24;$$02$i = $23;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $29 = HEAP32[$0>>2]|0; //@line 252 "modules/joystick/wrap_JoystickModule.cpp"
   $30 = ((($29)) + 8|0); //@line 252 "modules/joystick/wrap_JoystickModule.cpp"
   $31 = HEAP32[$30>>2]|0; //@line 252 "modules/joystick/wrap_JoystickModule.cpp"
   FUNCTION_TABLE_vi[$31 & 8191]($0); //@line 252 "modules/joystick/wrap_JoystickModule.cpp"
   $instance$sroa$0$1 = $1;
  }
 } while(0);
 HEAP32[$w>>2] = $instance$sroa$0$1; //@line 255 "modules/joystick/wrap_JoystickModule.cpp"
 $32 = ((($w)) + 4|0); //@line 256 "modules/joystick/wrap_JoystickModule.cpp"
 HEAP32[$32>>2] = 354857; //@line 256 "modules/joystick/wrap_JoystickModule.cpp"
 $33 = ((($w)) + 8|0); //@line 257 "modules/joystick/wrap_JoystickModule.cpp"
 HEAP32[$33>>2] = 3; //@line 257 "modules/joystick/wrap_JoystickModule.cpp"
 $34 = ((($w)) + 12|0); //@line 258 "modules/joystick/wrap_JoystickModule.cpp"
 HEAP32[$34>>2] = 16384; //@line 258 "modules/joystick/wrap_JoystickModule.cpp"
 $35 = ((($w)) + 16|0); //@line 259 "modules/joystick/wrap_JoystickModule.cpp"
 HEAP32[$35>>2] = 16440; //@line 259 "modules/joystick/wrap_JoystickModule.cpp"
 $36 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 261 "modules/joystick/wrap_JoystickModule.cpp"
 STACKTOP = sp;return ($36|0); //@line 261 "modules/joystick/wrap_JoystickModule.cpp"
}
function __ZN4love8keyboard14w_setKeyRepeatEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625636)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 36 "modules/keyboard/wrap_Keyboard.cpp"
 $2 = ((($1)) + 24|0); //@line 36 "modules/keyboard/wrap_Keyboard.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 36 "modules/keyboard/wrap_Keyboard.cpp"
 $4 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,1)|0); //@line 36 "modules/keyboard/wrap_Keyboard.cpp"
 FUNCTION_TABLE_vii[$3 & 8191]($0,$4); //@line 36 "modules/keyboard/wrap_Keyboard.cpp"
 return 0; //@line 37 "modules/keyboard/wrap_Keyboard.cpp"
}
function __ZN4love8keyboard14w_hasKeyRepeatEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625636)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 42 "modules/keyboard/wrap_Keyboard.cpp"
 $2 = ((($1)) + 28|0); //@line 42 "modules/keyboard/wrap_Keyboard.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 42 "modules/keyboard/wrap_Keyboard.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 42 "modules/keyboard/wrap_Keyboard.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 42 "modules/keyboard/wrap_Keyboard.cpp"
 return 1; //@line 43 "modules/keyboard/wrap_Keyboard.cpp"
}
function __ZN4love8keyboard8w_isDownEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$07 = 0;
 var $k = 0, $keylist = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i1 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $k = sp + 12|0;
 $keylist = sp;
 $0 = (_lua_gettop($L)|0); //@line 49 "modules/keyboard/wrap_Keyboard.cpp"
 HEAP32[$keylist>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($keylist)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($keylist)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = ((($keylist)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $4 = ($0|0)==(0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 L1: do {
  if ($4) {
   label = 4;
  } else {
   $5 = $0 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __THREW__ = 0;
   $6 = (invoke_ii(3348,($5|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if ($8) {
    label = 6;
   } else {
    $9 = $6; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $10 = (($6) + ($0<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $11 = $10; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    HEAP32[$keylist>>2] = $9; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$1>>2] = $9; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$3>>2] = $11; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $12 = ($0|0)>(0); //@line 53 "modules/keyboard/wrap_Keyboard.cpp"
    if ($12) {
     $i$07 = 0;
     L4: while(1) {
      $33 = (($i$07) + 1)|0; //@line 55 "modules/keyboard/wrap_Keyboard.cpp"
      __THREW__ = 0;
      $34 = (invoke_iiii(3359,($L|0),($33|0),(0|0))|0); //@line 55 "modules/keyboard/wrap_Keyboard.cpp"
      $35 = __THREW__; __THREW__ = 0;
      $36 = $35&1;
      if ($36) {
       break;
      }
      __THREW__ = 0;
      $37 = (invoke_iii(3648,($34|0),($k|0))|0); //@line 55 "modules/keyboard/wrap_Keyboard.cpp"
      $38 = __THREW__; __THREW__ = 0;
      $39 = $38&1;
      if ($39) {
       break;
      }
      do {
       if ($37) {
        $41 = HEAP32[$1>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $42 = HEAP32[$3>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $43 = ($41|0)==($42|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($43) {
         __THREW__ = 0;
         invoke_vii(3649,($keylist|0),($k|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $46 = __THREW__; __THREW__ = 0;
         $47 = $46&1;
         if ($47) {
          break L4;
         } else {
          break;
         }
        } else {
         $44 = HEAP32[$k>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         HEAP32[$41>>2] = $44; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $45 = ((($41)) + 4|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$1>>2] = $45; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         break;
        }
       }
      } while(0);
      $40 = ($33|0)<($0|0); //@line 53 "modules/keyboard/wrap_Keyboard.cpp"
      if ($40) {
       $i$07 = $33;
      } else {
       label = 4;
       break L1;
      }
     }
     $20 = ___cxa_find_matching_catch_2()|0;
     $21 = tempRet0;
     $lpad$phi$index = $20;$lpad$phi$index2 = $21;
    } else {
     label = 4;
    }
   }
  }
 } while(0);
 if ((label|0) == 4) {
  $13 = HEAP32[(625636)>>2]|0; //@line 98 "./common/Module.h"
  $14 = HEAP32[$13>>2]|0; //@line 59 "modules/keyboard/wrap_Keyboard.cpp"
  $15 = ((($14)) + 32|0); //@line 59 "modules/keyboard/wrap_Keyboard.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 59 "modules/keyboard/wrap_Keyboard.cpp"
  __THREW__ = 0;
  $17 = (invoke_iii($16|0,($13|0),($keylist|0))|0); //@line 59 "modules/keyboard/wrap_Keyboard.cpp"
  $18 = __THREW__; __THREW__ = 0;
  $19 = $18&1;
  if ($19) {
   label = 6;
  } else {
   __THREW__ = 0;
   invoke_vii(3397,($L|0),($17|0)); //@line 59 "modules/keyboard/wrap_Keyboard.cpp"
   $48 = __THREW__; __THREW__ = 0;
   $49 = $48&1;
   if ($49) {
    label = 6;
   } else {
    $50 = HEAP32[$keylist>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $51 = ($50|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $52 = $50; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($51) {
     STACKTOP = sp;return 1; //@line 61 "modules/keyboard/wrap_Keyboard.cpp"
    }
    $53 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $54 = ($53|0)==($50|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($54)) {
     $scevgep$i$i$i = ((($53)) + -4|0);
     $55 = $scevgep$i$i$i;
     $56 = (($55) - ($52))|0;
     $57 = $56 >>> 2;
     $58 = $57 ^ -1;
     $scevgep4$i$i$i = (($53) + ($58<<2)|0);
     HEAP32[$1>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($50); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    STACKTOP = sp;return 1; //@line 61 "modules/keyboard/wrap_Keyboard.cpp"
   }
  }
 }
 if ((label|0) == 6) {
  $22 = ___cxa_find_matching_catch_2()|0;
  $23 = tempRet0;
  $lpad$phi$index = $22;$lpad$phi$index2 = $23;
 }
 $24 = HEAP32[$keylist>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $25 = ($24|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $26 = $24; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($25) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $27 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $28 = ($27|0)==($24|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($28)) {
  $scevgep$i$i$i1 = ((($27)) + -4|0);
  $29 = $scevgep$i$i$i1;
  $30 = (($29) - ($26))|0;
  $31 = $30 >>> 2;
  $32 = $31 ^ -1;
  $scevgep4$i$i$i2 = (($27) + ($32<<2)|0);
  HEAP32[$1>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($24); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__16vectorIN4love8keyboard8Keyboard3KeyENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8keyboard16w_isScancodeDownEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$07 = 0;
 var $lpad$phi$index = 0, $lpad$phi$index2 = 0, $scancode = 0, $scancodelist = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i1 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $scancode = sp + 12|0;
 $scancodelist = sp;
 $0 = (_lua_gettop($L)|0); //@line 66 "modules/keyboard/wrap_Keyboard.cpp"
 HEAP32[$scancodelist>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($scancodelist)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($scancodelist)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = ((($scancodelist)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $4 = ($0|0)==(0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 L1: do {
  if ($4) {
   label = 4;
  } else {
   $5 = $0 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __THREW__ = 0;
   $6 = (invoke_ii(3348,($5|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if ($8) {
    label = 6;
   } else {
    $9 = $6; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $10 = (($6) + ($0<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $11 = $10; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    HEAP32[$scancodelist>>2] = $9; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$1>>2] = $9; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$3>>2] = $11; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $12 = ($0|0)>(0); //@line 70 "modules/keyboard/wrap_Keyboard.cpp"
    if ($12) {
     $i$07 = 0;
     L4: while(1) {
      $33 = (($i$07) + 1)|0; //@line 72 "modules/keyboard/wrap_Keyboard.cpp"
      __THREW__ = 0;
      $34 = (invoke_iiii(3359,($L|0),($33|0),(0|0))|0); //@line 72 "modules/keyboard/wrap_Keyboard.cpp"
      $35 = __THREW__; __THREW__ = 0;
      $36 = $35&1;
      if ($36) {
       break;
      }
      __THREW__ = 0;
      $37 = (invoke_iii(3650,($34|0),($scancode|0))|0); //@line 72 "modules/keyboard/wrap_Keyboard.cpp"
      $38 = __THREW__; __THREW__ = 0;
      $39 = $38&1;
      if ($39) {
       break;
      }
      do {
       if ($37) {
        $41 = HEAP32[$1>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $42 = HEAP32[$3>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $43 = ($41|0)==($42|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($43) {
         __THREW__ = 0;
         invoke_vii(3651,($scancodelist|0),($scancode|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $46 = __THREW__; __THREW__ = 0;
         $47 = $46&1;
         if ($47) {
          break L4;
         } else {
          break;
         }
        } else {
         $44 = HEAP32[$scancode>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         HEAP32[$41>>2] = $44; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $45 = ((($41)) + 4|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$1>>2] = $45; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         break;
        }
       }
      } while(0);
      $40 = ($33|0)<($0|0); //@line 70 "modules/keyboard/wrap_Keyboard.cpp"
      if ($40) {
       $i$07 = $33;
      } else {
       label = 4;
       break L1;
      }
     }
     $20 = ___cxa_find_matching_catch_2()|0;
     $21 = tempRet0;
     $lpad$phi$index = $20;$lpad$phi$index2 = $21;
    } else {
     label = 4;
    }
   }
  }
 } while(0);
 if ((label|0) == 4) {
  $13 = HEAP32[(625636)>>2]|0; //@line 98 "./common/Module.h"
  $14 = HEAP32[$13>>2]|0; //@line 76 "modules/keyboard/wrap_Keyboard.cpp"
  $15 = ((($14)) + 36|0); //@line 76 "modules/keyboard/wrap_Keyboard.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 76 "modules/keyboard/wrap_Keyboard.cpp"
  __THREW__ = 0;
  $17 = (invoke_iii($16|0,($13|0),($scancodelist|0))|0); //@line 76 "modules/keyboard/wrap_Keyboard.cpp"
  $18 = __THREW__; __THREW__ = 0;
  $19 = $18&1;
  if ($19) {
   label = 6;
  } else {
   __THREW__ = 0;
   invoke_vii(3397,($L|0),($17|0)); //@line 76 "modules/keyboard/wrap_Keyboard.cpp"
   $48 = __THREW__; __THREW__ = 0;
   $49 = $48&1;
   if ($49) {
    label = 6;
   } else {
    $50 = HEAP32[$scancodelist>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $51 = ($50|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $52 = $50; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($51) {
     STACKTOP = sp;return 1; //@line 78 "modules/keyboard/wrap_Keyboard.cpp"
    }
    $53 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $54 = ($53|0)==($50|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($54)) {
     $scevgep$i$i$i = ((($53)) + -4|0);
     $55 = $scevgep$i$i$i;
     $56 = (($55) - ($52))|0;
     $57 = $56 >>> 2;
     $58 = $57 ^ -1;
     $scevgep4$i$i$i = (($53) + ($58<<2)|0);
     HEAP32[$1>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($50); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    STACKTOP = sp;return 1; //@line 78 "modules/keyboard/wrap_Keyboard.cpp"
   }
  }
 }
 if ((label|0) == 6) {
  $22 = ___cxa_find_matching_catch_2()|0;
  $23 = tempRet0;
  $lpad$phi$index = $22;$lpad$phi$index2 = $23;
 }
 $24 = HEAP32[$scancodelist>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $25 = ($24|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $26 = $24; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($25) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $27 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $28 = ($27|0)==($24|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($28)) {
  $scevgep$i$i$i1 = ((($27)) + -4|0);
  $29 = $scevgep$i$i$i1;
  $30 = (($29) - ($26))|0;
  $31 = $30 >>> 2;
  $32 = $31 ^ -1;
  $scevgep4$i$i$i2 = (($27) + ($32<<2)|0);
  HEAP32[$1>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($24); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__16vectorIN4love8keyboard8Keyboard8ScancodeENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8keyboard20w_getScancodeFromKeyEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $key = 0, $scancodestr = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $key = sp + 16|0;
 $scancodestr = sp + 12|0;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 82 "modules/keyboard/wrap_Keyboard.cpp"
 $1 = (__ZN4love8keyboard8Keyboard11getConstantEPKcRNS1_3KeyE($0,$key)|0); //@line 84 "modules/keyboard/wrap_Keyboard.cpp"
 if (!($1)) {
  HEAP32[$vararg_buffer>>2] = $0; //@line 85 "modules/keyboard/wrap_Keyboard.cpp"
  $2 = (_luaL_error($L,354954,$vararg_buffer)|0); //@line 85 "modules/keyboard/wrap_Keyboard.cpp"
  $$1 = $2;
  STACKTOP = sp;return ($$1|0); //@line 95 "modules/keyboard/wrap_Keyboard.cpp"
 }
 $3 = HEAP32[(625636)>>2]|0; //@line 98 "./common/Module.h"
 $4 = HEAP32[$3>>2]|0; //@line 87 "modules/keyboard/wrap_Keyboard.cpp"
 $5 = ((($4)) + 44|0); //@line 87 "modules/keyboard/wrap_Keyboard.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 87 "modules/keyboard/wrap_Keyboard.cpp"
 $7 = HEAP32[$key>>2]|0; //@line 87 "modules/keyboard/wrap_Keyboard.cpp"
 $8 = (FUNCTION_TABLE_iii[$6 & 8191]($3,$7)|0); //@line 87 "modules/keyboard/wrap_Keyboard.cpp"
 $9 = (__ZN4love8keyboard8Keyboard11getConstantENS1_8ScancodeERPKc($8,$scancodestr)|0); //@line 90 "modules/keyboard/wrap_Keyboard.cpp"
 if ($9) {
  $11 = HEAP32[$scancodestr>>2]|0; //@line 93 "modules/keyboard/wrap_Keyboard.cpp"
  _lua_pushstring($L,$11); //@line 93 "modules/keyboard/wrap_Keyboard.cpp"
  $$0 = 1;
 } else {
  $10 = (_luaL_error($L,354979,$vararg_buffer1)|0); //@line 91 "modules/keyboard/wrap_Keyboard.cpp"
  $$0 = $10;
 }
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0); //@line 95 "modules/keyboard/wrap_Keyboard.cpp"
}
function __ZN4love8keyboard20w_getKeyFromScancodeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $keystr = 0, $scancode = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $scancode = sp + 16|0;
 $keystr = sp + 12|0;
 $0 = (_luaL_checklstring($L,1,0)|0); //@line 99 "modules/keyboard/wrap_Keyboard.cpp"
 $1 = (__ZN4love8keyboard8Keyboard11getConstantEPKcRNS1_8ScancodeE($0,$scancode)|0); //@line 101 "modules/keyboard/wrap_Keyboard.cpp"
 if (!($1)) {
  HEAP32[$vararg_buffer>>2] = $0; //@line 102 "modules/keyboard/wrap_Keyboard.cpp"
  $2 = (_luaL_error($L,354997,$vararg_buffer)|0); //@line 102 "modules/keyboard/wrap_Keyboard.cpp"
  $$1 = $2;
  STACKTOP = sp;return ($$1|0); //@line 112 "modules/keyboard/wrap_Keyboard.cpp"
 }
 $3 = HEAP32[(625636)>>2]|0; //@line 98 "./common/Module.h"
 $4 = HEAP32[$3>>2]|0; //@line 104 "modules/keyboard/wrap_Keyboard.cpp"
 $5 = ((($4)) + 40|0); //@line 104 "modules/keyboard/wrap_Keyboard.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 104 "modules/keyboard/wrap_Keyboard.cpp"
 $7 = HEAP32[$scancode>>2]|0; //@line 104 "modules/keyboard/wrap_Keyboard.cpp"
 $8 = (FUNCTION_TABLE_iii[$6 & 8191]($3,$7)|0); //@line 104 "modules/keyboard/wrap_Keyboard.cpp"
 $9 = (__ZN4love8keyboard8Keyboard11getConstantENS1_3KeyERPKc($8,$keystr)|0); //@line 107 "modules/keyboard/wrap_Keyboard.cpp"
 if ($9) {
  $11 = HEAP32[$keystr>>2]|0; //@line 110 "modules/keyboard/wrap_Keyboard.cpp"
  _lua_pushstring($L,$11); //@line 110 "modules/keyboard/wrap_Keyboard.cpp"
  $$0 = 1;
 } else {
  $10 = (_luaL_error($L,355018,$vararg_buffer1)|0); //@line 108 "modules/keyboard/wrap_Keyboard.cpp"
  $$0 = $10;
 }
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0); //@line 112 "modules/keyboard/wrap_Keyboard.cpp"
}
function __ZN4love8keyboard14w_setTextInputEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,1)|0); //@line 116 "modules/keyboard/wrap_Keyboard.cpp"
 $1 = (_lua_gettop($L)|0); //@line 118 "modules/keyboard/wrap_Keyboard.cpp"
 $2 = ($1|0)<(2); //@line 118 "modules/keyboard/wrap_Keyboard.cpp"
 if ($2) {
  $3 = HEAP32[(625636)>>2]|0; //@line 98 "./common/Module.h"
  $4 = HEAP32[$3>>2]|0; //@line 119 "modules/keyboard/wrap_Keyboard.cpp"
  $5 = ((($4)) + 48|0); //@line 119 "modules/keyboard/wrap_Keyboard.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 119 "modules/keyboard/wrap_Keyboard.cpp"
  FUNCTION_TABLE_vii[$6 & 8191]($3,$0); //@line 119 "modules/keyboard/wrap_Keyboard.cpp"
  return 0; //@line 129 "modules/keyboard/wrap_Keyboard.cpp"
 } else {
  $7 = (+_luaL_checknumber($L,2)); //@line 122 "modules/keyboard/wrap_Keyboard.cpp"
  $8 = (+_luaL_checknumber($L,3)); //@line 123 "modules/keyboard/wrap_Keyboard.cpp"
  $9 = (+_luaL_checknumber($L,4)); //@line 124 "modules/keyboard/wrap_Keyboard.cpp"
  $10 = (+_luaL_checknumber($L,5)); //@line 125 "modules/keyboard/wrap_Keyboard.cpp"
  $11 = HEAP32[(625636)>>2]|0; //@line 98 "./common/Module.h"
  $12 = HEAP32[$11>>2]|0; //@line 126 "modules/keyboard/wrap_Keyboard.cpp"
  $13 = ((($12)) + 52|0); //@line 126 "modules/keyboard/wrap_Keyboard.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 126 "modules/keyboard/wrap_Keyboard.cpp"
  FUNCTION_TABLE_viidddd[$14 & 2047]($11,$0,$7,$8,$9,$10); //@line 126 "modules/keyboard/wrap_Keyboard.cpp"
  return 0; //@line 129 "modules/keyboard/wrap_Keyboard.cpp"
 }
 return (0)|0;
}
function __ZN4love8keyboard14w_hasTextInputEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625636)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 134 "modules/keyboard/wrap_Keyboard.cpp"
 $2 = ((($1)) + 56|0); //@line 134 "modules/keyboard/wrap_Keyboard.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 134 "modules/keyboard/wrap_Keyboard.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 134 "modules/keyboard/wrap_Keyboard.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 134 "modules/keyboard/wrap_Keyboard.cpp"
 return 1; //@line 135 "modules/keyboard/wrap_Keyboard.cpp"
}
function __ZN4love8keyboard19w_hasScreenKeyboardEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625636)>>2]|0; //@line 98 "./common/Module.h"
 $1 = HEAP32[$0>>2]|0; //@line 140 "modules/keyboard/wrap_Keyboard.cpp"
 $2 = ((($1)) + 60|0); //@line 140 "modules/keyboard/wrap_Keyboard.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 140 "modules/keyboard/wrap_Keyboard.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 140 "modules/keyboard/wrap_Keyboard.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 140 "modules/keyboard/wrap_Keyboard.cpp"
 return 1; //@line 141 "modules/keyboard/wrap_Keyboard.cpp"
}
function _luaopen_love_keyboard($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0;
 var $eh$lpad$body$i$indexZ2D = 0, $instance$sroa$0$1 = 0, $vararg_buffer = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $w = sp + 4|0;
 $0 = HEAP32[(625636)>>2]|0; //@line 98 "./common/Module.h"
 $1 = $0; //@line 161 "modules/keyboard/wrap_Keyboard.cpp"
 $2 = ($0|0)==(0|0); //@line 162 "modules/keyboard/wrap_Keyboard.cpp"
 L1: do {
  if ($2) {
   __THREW__ = 0;
   $3 = (invoke_ii(3348,12)|0); //@line 164 "modules/keyboard/wrap_Keyboard.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   do {
    if ($5) {
     $11 = ___cxa_find_matching_catch_3(8992|0)|0;
     $12 = tempRet0;
     $eh$lpad$body$i$index2Z2D = $12;$eh$lpad$body$i$indexZ2D = $11;
    } else {
     __THREW__ = 0;
     invoke_vi(3652,($3|0)); //@line 164 "modules/keyboard/wrap_Keyboard.cpp"
     $6 = __THREW__; __THREW__ = 0;
     $7 = $6&1;
     if ($7) {
      $8 = ___cxa_find_matching_catch_3(8992|0)|0;
      $9 = tempRet0;
      __ZdlPv($3); //@line 164 "modules/keyboard/wrap_Keyboard.cpp"
      $eh$lpad$body$i$index2Z2D = $9;$eh$lpad$body$i$indexZ2D = $8;
      break;
     } else {
      $10 = $3; //@line 164 "modules/keyboard/wrap_Keyboard.cpp"
      $instance$sroa$0$1 = $10;
      break L1;
     }
    }
   } while(0);
   $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $14 = ($eh$lpad$body$i$index2Z2D|0)==($13|0); //@line 535 "./common/runtime.h"
   if (!($14)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $15 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
   $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
   $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
   $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $22 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $instance$sroa$0$1 = $1;
    break;
   }
   $23 = ___cxa_find_matching_catch_2()|0;
   $24 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    $27 = ___cxa_find_matching_catch_3(0|0)|0;
    $28 = tempRet0;
    ___clang_call_terminate($27); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $24;$$02$i = $23;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $29 = HEAP32[$0>>2]|0; //@line 167 "modules/keyboard/wrap_Keyboard.cpp"
   $30 = ((($29)) + 8|0); //@line 167 "modules/keyboard/wrap_Keyboard.cpp"
   $31 = HEAP32[$30>>2]|0; //@line 167 "modules/keyboard/wrap_Keyboard.cpp"
   FUNCTION_TABLE_vi[$31 & 8191]($0); //@line 167 "modules/keyboard/wrap_Keyboard.cpp"
   $instance$sroa$0$1 = $1;
  }
 } while(0);
 HEAP32[$w>>2] = $instance$sroa$0$1; //@line 170 "modules/keyboard/wrap_Keyboard.cpp"
 $32 = ((($w)) + 4|0); //@line 171 "modules/keyboard/wrap_Keyboard.cpp"
 HEAP32[$32>>2] = 355039; //@line 171 "modules/keyboard/wrap_Keyboard.cpp"
 $33 = ((($w)) + 8|0); //@line 172 "modules/keyboard/wrap_Keyboard.cpp"
 HEAP32[$33>>2] = 3; //@line 172 "modules/keyboard/wrap_Keyboard.cpp"
 $34 = ((($w)) + 12|0); //@line 173 "modules/keyboard/wrap_Keyboard.cpp"
 HEAP32[$34>>2] = 16448; //@line 173 "modules/keyboard/wrap_Keyboard.cpp"
 $35 = ((($w)) + 16|0); //@line 174 "modules/keyboard/wrap_Keyboard.cpp"
 HEAP32[$35>>2] = 0; //@line 174 "modules/keyboard/wrap_Keyboard.cpp"
 $36 = (__ZN4love20luax_register_moduleEP9lua_StateRKNS_13WrappedModuleE($L,$w)|0); //@line 176 "modules/keyboard/wrap_Keyboard.cpp"
 STACKTOP = sp;return ($36|0); //@line 176 "modules/keyboard/wrap_Keyboard.cpp"
}
function __ZN4love4math14CompressedDataD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (16536); //@line 53 "modules/math/CompressedData.cpp"
 $0 = ((($this)) + 12|0); //@line 54 "modules/math/CompressedData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 54 "modules/math/CompressedData.cpp"
 $2 = ($1|0)==(0|0); //@line 54 "modules/math/CompressedData.cpp"
 if ($2) {
  return; //@line 55 "modules/math/CompressedData.cpp"
 }
 __ZdaPv($1); //@line 54 "modules/math/CompressedData.cpp"
 return; //@line 55 "modules/math/CompressedData.cpp"
}
function __ZN4love4math14CompressedDataD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (16536); //@line 53 "modules/math/CompressedData.cpp"
 $0 = ((($this)) + 12|0); //@line 54 "modules/math/CompressedData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 54 "modules/math/CompressedData.cpp"
 $2 = ($1|0)==(0|0); //@line 54 "modules/math/CompressedData.cpp"
 if (!($2)) {
  __ZdaPv($1); //@line 54 "modules/math/CompressedData.cpp"
 }
 __ZdlPv($this); //@line 53 "modules/math/CompressedData.cpp"
 return; //@line 55 "modules/math/CompressedData.cpp"
}
function __ZNK4love4math14CompressedData7getDataEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 69 "modules/math/CompressedData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 69 "modules/math/CompressedData.cpp"
 return ($1|0); //@line 69 "modules/math/CompressedData.cpp"
}
function __ZNK4love4math14CompressedData7getSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 74 "modules/math/CompressedData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 74 "modules/math/CompressedData.cpp"
 return ($1|0); //@line 74 "modules/math/CompressedData.cpp"
}
function __ZN4love4math14CompressedDataC2ENS0_10Compressor6FormatEPcjjb($this,$format,$cdata,$compressedsize,$rawsize,$own) {
 $this = $this|0;
 $format = $format|0;
 $cdata = $cdata|0;
 $compressedsize = $compressedsize|0;
 $rawsize = $rawsize|0;
 $own = $own|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 __ZN4love6ObjectC2Ev($this); //@line 37 "./common/Data.h"
 HEAP32[$this>>2] = (16536); //@line 34 "modules/math/CompressedData.cpp"
 $0 = ((($this)) + 8|0); //@line 30 "modules/math/CompressedData.cpp"
 HEAP32[$0>>2] = $format; //@line 30 "modules/math/CompressedData.cpp"
 $1 = ((($this)) + 12|0); //@line 31 "modules/math/CompressedData.cpp"
 HEAP32[$1>>2] = 0; //@line 31 "modules/math/CompressedData.cpp"
 $2 = ((($this)) + 16|0); //@line 32 "modules/math/CompressedData.cpp"
 HEAP32[$2>>2] = $compressedsize; //@line 32 "modules/math/CompressedData.cpp"
 $3 = ((($this)) + 20|0); //@line 33 "modules/math/CompressedData.cpp"
 HEAP32[$3>>2] = $rawsize; //@line 33 "modules/math/CompressedData.cpp"
 if ($own) {
  HEAP32[$1>>2] = $cdata; //@line 36 "modules/math/CompressedData.cpp"
  STACKTOP = sp;return; //@line 50 "modules/math/CompressedData.cpp"
 }
 __THREW__ = 0;
 $4 = (invoke_ii(3492,($compressedsize|0))|0); //@line 41 "modules/math/CompressedData.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  HEAP32[$1>>2] = $4; //@line 41 "modules/math/CompressedData.cpp"
  _memcpy(($4|0),($cdata|0),($compressedsize|0))|0; //@line 48 "modules/math/CompressedData.cpp"
  STACKTOP = sp;return; //@line 50 "modules/math/CompressedData.cpp"
 }
 $7 = ___cxa_find_matching_catch_3(9056|0)|0;
 $8 = tempRet0;
 $9 = (_llvm_eh_typeid_for((9056|0))|0); //@line 42 "modules/math/CompressedData.cpp"
 $10 = ($8|0)==($9|0); //@line 42 "modules/math/CompressedData.cpp"
 if (!($10)) {
  $$1 = $7;$$12 = $8;
  ___resumeException($$1|0);
  // unreachable;
 }
 (___cxa_begin_catch(($7|0))|0); //@line 42 "modules/math/CompressedData.cpp"
 $11 = (___cxa_allocate_exception(16)|0); //@line 45 "modules/math/CompressedData.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($11|0),(507610|0),($vararg_buffer|0)); //@line 45 "modules/math/CompressedData.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if ($13) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  ___cxa_free_exception(($11|0)); //@line 45 "modules/math/CompressedData.cpp"
  $$0 = $17;$$01 = $18;
 } else {
  __THREW__ = 0;
  invoke_viii(3350,($11|0),(96|0),(344|0)); //@line 45 "modules/math/CompressedData.cpp"
  $14 = __THREW__; __THREW__ = 0;
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  $$0 = $15;$$01 = $16;
 }
 __THREW__ = 0;
 invoke_v(3378); //@line 46 "modules/math/CompressedData.cpp"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $21 = ___cxa_find_matching_catch_3(0|0)|0;
  $22 = tempRet0;
  ___clang_call_terminate($21); //@line 46 "modules/math/CompressedData.cpp"
  // unreachable; //@line 46 "modules/math/CompressedData.cpp"
 } else {
  $$1 = $$0;$$12 = $$01;
  ___resumeException($$1|0);
  // unreachable;
 }
}
function __ZNK4love4math14CompressedData9getFormatEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 59 "modules/math/CompressedData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 59 "modules/math/CompressedData.cpp"
 return ($1|0); //@line 59 "modules/math/CompressedData.cpp"
}
function __ZNK4love4math14CompressedData19getDecompressedSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 64 "modules/math/CompressedData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 64 "modules/math/CompressedData.cpp"
 return ($1|0); //@line 64 "modules/math/CompressedData.cpp"
}
function __GLOBAL__sub_I_Compressor_cpp() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[(630508)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(630520)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(630532)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(630544)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(630556)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(630568)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(630572)>>2]=0|0;HEAP32[(630572)+4>>2]=0|0;HEAP32[(630572)+8>>2]=0|0; //@line 44 "./common/StringMap.h"
 $0 = HEAP32[4140]|0; //@line 49 "./common/StringMap.h"
 $1 = HEAP32[(16564)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_4math10Compressor6FormatELj3EE3addEPKcS3_(630500,$0,$1)|0); //@line 49 "./common/StringMap.h"
 $2 = HEAP32[(16568)>>2]|0; //@line 49 "./common/StringMap.h"
 $3 = HEAP32[(16572)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_4math10Compressor6FormatELj3EE3addEPKcS3_(630500,$2,$3)|0); //@line 49 "./common/StringMap.h"
 $4 = HEAP32[(16576)>>2]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(16580)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_4math10Compressor6FormatELj3EE3addEPKcS3_(630500,$4,$5)|0); //@line 49 "./common/StringMap.h"
 return;
}
function __ZN4love9StringMapINS_4math10Compressor6FormatELj3EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $inserted$2$off0 = 0;
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(2); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 72|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love4math10Compressor13getCompressorENS1_6FormatE($format) {
 $format = $format|0;
 var $$ = 0, $$13 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $inttoptr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[623968]|0; //@line 360 "modules/math/Compressor.cpp"
 $1 = ($0<<24>>24)==(0); //@line 360 "modules/math/Compressor.cpp"
 if ($1) {
  $2 = (___cxa_guard_acquire((623968|0))|0); //@line 360 "modules/math/Compressor.cpp"
  $3 = ($2|0)==(0); //@line 360 "modules/math/Compressor.cpp"
  if (!($3)) {
   HEAP32[157646] = (16592); //@line 38 "modules/math/Compressor.cpp"
   ___cxa_guard_release((623968|0)); //@line 360 "modules/math/Compressor.cpp"
  }
 }
 $4 = HEAP8[623976]|0; //@line 361 "modules/math/Compressor.cpp"
 $5 = ($4<<24>>24)==(0); //@line 361 "modules/math/Compressor.cpp"
 if ($5) {
  $6 = (___cxa_guard_acquire((623976|0))|0); //@line 361 "modules/math/Compressor.cpp"
  $7 = ($6|0)==(0); //@line 361 "modules/math/Compressor.cpp"
  if (!($7)) {
   HEAP32[157647] = (16620); //@line 171 "modules/math/Compressor.cpp"
   ___cxa_guard_release((623976|0)); //@line 361 "modules/math/Compressor.cpp"
  }
 }
 $8 = 296; //@line 363 "modules/math/Compressor.cpp"
 $9 = $8; //@line 363 "modules/math/Compressor.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 363 "modules/math/Compressor.cpp"
 $11 = (($8) + 4)|0; //@line 363 "modules/math/Compressor.cpp"
 $12 = $11; //@line 363 "modules/math/Compressor.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 363 "modules/math/Compressor.cpp"
 $inttoptr = $10; //@line 365 "modules/math/Compressor.cpp"
 $14 = $10; //@line 367 "modules/math/Compressor.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 367 "modules/math/Compressor.cpp"
 $16 = ((($15)) + 16|0); //@line 367 "modules/math/Compressor.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 367 "modules/math/Compressor.cpp"
 $18 = (FUNCTION_TABLE_iii[$17 & 8191]($inttoptr,$format)|0); //@line 367 "modules/math/Compressor.cpp"
 $19 = $13; //@line 365 "modules/math/Compressor.cpp"
 if ($18) {
  $$13 = $inttoptr;
  return ($$13|0); //@line 372 "modules/math/Compressor.cpp"
 }
 $20 = $13; //@line 367 "modules/math/Compressor.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 367 "modules/math/Compressor.cpp"
 $22 = ((($21)) + 16|0); //@line 367 "modules/math/Compressor.cpp"
 $23 = HEAP32[$22>>2]|0; //@line 367 "modules/math/Compressor.cpp"
 $24 = (FUNCTION_TABLE_iii[$23 & 8191]($19,$format)|0); //@line 367 "modules/math/Compressor.cpp"
 $$ = $24 ? $19 : 0;
 $$13 = $$;
 return ($$13|0); //@line 372 "modules/math/Compressor.cpp"
}
function __ZN4love4math10CompressorD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 53 "modules/math/Compressor.h"
}
function __ZN4love4math13LZ4CompressorD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 38 "modules/math/Compressor.cpp"
 return; //@line 38 "modules/math/Compressor.cpp"
}
function __ZN4love4math13LZ4Compressor8compressENS0_10Compressor6FormatEPKcjiRj($this,$format,$data,$dataSize,$level,$compressedSize) {
 $this = $this|0;
 $format = $format|0;
 $data = $data|0;
 $dataSize = $dataSize|0;
 $level = $level|0;
 $compressedSize = $compressedSize|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0.0, $42 = 0, $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $compressedbytes$1 = 0, $csize$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($format|0)==(0); //@line 44 "modules/math/Compressor.cpp"
 if (!($0)) {
  $1 = (___cxa_allocate_exception(16)|0); //@line 45 "modules/math/Compressor.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($1|0),(355267|0),($vararg_buffer|0)); //@line 45 "modules/math/Compressor.cpp"
  $2 = __THREW__; __THREW__ = 0;
  $3 = $2&1;
  if (!($3)) {
   ___cxa_throw(($1|0),(96|0),(344|0)); //@line 45 "modules/math/Compressor.cpp"
   // unreachable; //@line 45 "modules/math/Compressor.cpp"
  }
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  ___cxa_free_exception(($1|0)); //@line 45 "modules/math/Compressor.cpp"
  $$1 = $4;$$12 = $5;
  ___resumeException($$1|0);
  // unreachable;
 }
 $6 = ($dataSize>>>0)>(2113929216); //@line 47 "modules/math/Compressor.cpp"
 if ($6) {
  $7 = (___cxa_allocate_exception(16)|0); //@line 48 "modules/math/Compressor.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($7|0),(355298|0),($vararg_buffer1|0)); //@line 48 "modules/math/Compressor.cpp"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if (!($9)) {
   ___cxa_throw(($7|0),(96|0),(344|0)); //@line 48 "modules/math/Compressor.cpp"
   // unreachable; //@line 48 "modules/math/Compressor.cpp"
  }
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  ___cxa_free_exception(($7|0)); //@line 48 "modules/math/Compressor.cpp"
  $$1 = $10;$$12 = $11;
  ___resumeException($$1|0);
  // unreachable;
 }
 $12 = (_LZ4_compressBound($dataSize)|0); //@line 53 "modules/math/Compressor.cpp"
 $13 = (($12) + 4)|0; //@line 54 "modules/math/Compressor.cpp"
 __THREW__ = 0;
 $14 = (invoke_ii(3492,($13|0))|0); //@line 59 "modules/math/Compressor.cpp"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if ($16) {
  $19 = ___cxa_find_matching_catch_3(9056|0)|0;
  $20 = tempRet0;
  $21 = (_llvm_eh_typeid_for((9056|0))|0); //@line 60 "modules/math/Compressor.cpp"
  $22 = ($20|0)==($21|0); //@line 60 "modules/math/Compressor.cpp"
  if (!($22)) {
   $$1 = $19;$$12 = $20;
   ___resumeException($$1|0);
   // unreachable;
  }
  (___cxa_begin_catch(($19|0))|0); //@line 60 "modules/math/Compressor.cpp"
  $23 = (___cxa_allocate_exception(16)|0); //@line 63 "modules/math/Compressor.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($23|0),(507610|0),($vararg_buffer3|0)); //@line 63 "modules/math/Compressor.cpp"
  $24 = __THREW__; __THREW__ = 0;
  $25 = $24&1;
  if ($25) {
   $29 = ___cxa_find_matching_catch_2()|0;
   $30 = tempRet0;
   ___cxa_free_exception(($23|0)); //@line 63 "modules/math/Compressor.cpp"
   $$0 = $29;$$01 = $30;
  } else {
   __THREW__ = 0;
   invoke_viii(3350,($23|0),(96|0),(344|0)); //@line 63 "modules/math/Compressor.cpp"
   $26 = __THREW__; __THREW__ = 0;
   $27 = ___cxa_find_matching_catch_2()|0;
   $28 = tempRet0;
   $$0 = $27;$$01 = $28;
  }
  __THREW__ = 0;
  invoke_v(3378); //@line 64 "modules/math/Compressor.cpp"
  $31 = __THREW__; __THREW__ = 0;
  $32 = $31&1;
  if ($32) {
   $48 = ___cxa_find_matching_catch_3(0|0)|0;
   $49 = tempRet0;
   ___clang_call_terminate($48); //@line 64 "modules/math/Compressor.cpp"
   // unreachable; //@line 64 "modules/math/Compressor.cpp"
  } else {
   $$1 = $$0;$$12 = $$01;
   ___resumeException($$1|0);
   // unreachable;
  }
 }
 HEAP32[$14>>2] = $dataSize; //@line 71 "modules/math/Compressor.cpp"
 $17 = ($level|0)>(8); //@line 76 "modules/math/Compressor.cpp"
 $18 = ((($14)) + 4|0); //@line 77 "modules/math/Compressor.cpp"
 if ($17) {
  $33 = (_LZ4_compress_HC($data,$18,$dataSize,$12,0)|0); //@line 77 "modules/math/Compressor.cpp"
  $csize$0 = $33;
 } else {
  $34 = (_LZ4_compress_default($data,$18,$dataSize,$12)|0); //@line 79 "modules/math/Compressor.cpp"
  $csize$0 = $34;
 }
 $35 = ($csize$0|0)<(1); //@line 81 "modules/math/Compressor.cpp"
 if ($35) {
  __ZdaPv($14); //@line 83 "modules/math/Compressor.cpp"
  $36 = (___cxa_allocate_exception(16)|0); //@line 84 "modules/math/Compressor.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($36|0),(355336|0),($vararg_buffer5|0)); //@line 84 "modules/math/Compressor.cpp"
  $37 = __THREW__; __THREW__ = 0;
  $38 = $37&1;
  if (!($38)) {
   ___cxa_throw(($36|0),(96|0),(344|0)); //@line 84 "modules/math/Compressor.cpp"
   // unreachable; //@line 84 "modules/math/Compressor.cpp"
  }
  $39 = ___cxa_find_matching_catch_2()|0;
  $40 = tempRet0;
  ___cxa_free_exception(($36|0)); //@line 84 "modules/math/Compressor.cpp"
  $$1 = $39;$$12 = $40;
  ___resumeException($$1|0);
  // unreachable;
 }
 $41 = (+($13>>>0)); //@line 90 "modules/math/Compressor.cpp"
 $42 = (($csize$0) + 4)|0; //@line 90 "modules/math/Compressor.cpp"
 $43 = (+($42>>>0)); //@line 90 "modules/math/Compressor.cpp"
 $44 = $41 / $43; //@line 90 "modules/math/Compressor.cpp"
 $45 = !($44 >= 1.2); //@line 90 "modules/math/Compressor.cpp"
 if ($45) {
  $compressedbytes$1 = $14;
  HEAP32[$compressedSize>>2] = $42; //@line 101 "modules/math/Compressor.cpp"
  STACKTOP = sp;return ($compressedbytes$1|0); //@line 102 "modules/math/Compressor.cpp"
 }
 $46 = (__ZnajRKSt9nothrow_t($42,1282467)|0); //@line 92 "modules/math/Compressor.cpp"
 $47 = ($46|0)==(0|0); //@line 93 "modules/math/Compressor.cpp"
 if ($47) {
  $compressedbytes$1 = $14;
  HEAP32[$compressedSize>>2] = $42; //@line 101 "modules/math/Compressor.cpp"
  STACKTOP = sp;return ($compressedbytes$1|0); //@line 102 "modules/math/Compressor.cpp"
 }
 _memcpy(($46|0),($14|0),($42|0))|0; //@line 95 "modules/math/Compressor.cpp"
 __ZdaPv($14); //@line 96 "modules/math/Compressor.cpp"
 $compressedbytes$1 = $46;
 HEAP32[$compressedSize>>2] = $42; //@line 101 "modules/math/Compressor.cpp"
 STACKTOP = sp;return ($compressedbytes$1|0); //@line 102 "modules/math/Compressor.cpp"
}
function __ZN4love4math13LZ4Compressor10decompressENS0_10Compressor6FormatEPKcjRj($this,$format,$data,$dataSize,$decompressedSize) {
 $this = $this|0;
 $format = $format|0;
 $data = $data|0;
 $dataSize = $dataSize|0;
 $decompressedSize = $decompressedSize|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0;
 var $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($format|0)==(0); //@line 107 "modules/math/Compressor.cpp"
 if (!($0)) {
  $1 = (___cxa_allocate_exception(16)|0); //@line 108 "modules/math/Compressor.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($1|0),(355267|0),($vararg_buffer|0)); //@line 108 "modules/math/Compressor.cpp"
  $2 = __THREW__; __THREW__ = 0;
  $3 = $2&1;
  if (!($3)) {
   ___cxa_throw(($1|0),(96|0),(344|0)); //@line 108 "modules/math/Compressor.cpp"
   // unreachable; //@line 108 "modules/math/Compressor.cpp"
  }
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  ___cxa_free_exception(($1|0)); //@line 108 "modules/math/Compressor.cpp"
  $$1 = $4;$$12 = $5;
  ___resumeException($$1|0);
  // unreachable;
 }
 $6 = ($dataSize>>>0)<(4); //@line 113 "modules/math/Compressor.cpp"
 if ($6) {
  $7 = (___cxa_allocate_exception(16)|0); //@line 114 "modules/math/Compressor.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($7|0),(355365|0),($vararg_buffer1|0)); //@line 114 "modules/math/Compressor.cpp"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if (!($9)) {
   ___cxa_throw(($7|0),(96|0),(344|0)); //@line 114 "modules/math/Compressor.cpp"
   // unreachable; //@line 114 "modules/math/Compressor.cpp"
  }
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  ___cxa_free_exception(($7|0)); //@line 114 "modules/math/Compressor.cpp"
  $$1 = $10;$$12 = $11;
  ___resumeException($$1|0);
  // unreachable;
 }
 $12 = HEAP32[$data>>2]|0; //@line 121 "modules/math/Compressor.cpp"
 __THREW__ = 0;
 $13 = (invoke_ii(3492,($12|0))|0); //@line 126 "modules/math/Compressor.cpp"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if ($15) {
  $20 = ___cxa_find_matching_catch_3(9056|0)|0;
  $21 = tempRet0;
  $22 = (_llvm_eh_typeid_for((9056|0))|0); //@line 127 "modules/math/Compressor.cpp"
  $23 = ($21|0)==($22|0); //@line 127 "modules/math/Compressor.cpp"
  if (!($23)) {
   $$1 = $20;$$12 = $21;
   ___resumeException($$1|0);
   // unreachable;
  }
  (___cxa_begin_catch(($20|0))|0); //@line 127 "modules/math/Compressor.cpp"
  $24 = (___cxa_allocate_exception(16)|0); //@line 130 "modules/math/Compressor.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($24|0),(507610|0),($vararg_buffer3|0)); //@line 130 "modules/math/Compressor.cpp"
  $25 = __THREW__; __THREW__ = 0;
  $26 = $25&1;
  if ($26) {
   $30 = ___cxa_find_matching_catch_2()|0;
   $31 = tempRet0;
   ___cxa_free_exception(($24|0)); //@line 130 "modules/math/Compressor.cpp"
   $$0 = $30;$$01 = $31;
  } else {
   __THREW__ = 0;
   invoke_viii(3350,($24|0),(96|0),(344|0)); //@line 130 "modules/math/Compressor.cpp"
   $27 = __THREW__; __THREW__ = 0;
   $28 = ___cxa_find_matching_catch_2()|0;
   $29 = tempRet0;
   $$0 = $28;$$01 = $29;
  }
  __THREW__ = 0;
  invoke_v(3378); //@line 131 "modules/math/Compressor.cpp"
  $32 = __THREW__; __THREW__ = 0;
  $33 = $32&1;
  if ($33) {
   $49 = ___cxa_find_matching_catch_3(0|0)|0;
   $50 = tempRet0;
   ___clang_call_terminate($49); //@line 131 "modules/math/Compressor.cpp"
   // unreachable; //@line 131 "modules/math/Compressor.cpp"
  } else {
   $$1 = $$0;$$12 = $$01;
   ___resumeException($$1|0);
   // unreachable;
  }
 }
 $16 = HEAP32[$decompressedSize>>2]|0; //@line 136 "modules/math/Compressor.cpp"
 $17 = ($16|0)!=(0); //@line 136 "modules/math/Compressor.cpp"
 $18 = ($16|0)==($12|0); //@line 136 "modules/math/Compressor.cpp"
 $or$cond = $17 & $18; //@line 136 "modules/math/Compressor.cpp"
 $19 = ((($data)) + 4|0); //@line 139 "modules/math/Compressor.cpp"
 if ($or$cond) {
  $34 = (_LZ4_decompress_fast($19,$13,$12)|0); //@line 139 "modules/math/Compressor.cpp"
  $35 = ($34|0)<(0); //@line 139 "modules/math/Compressor.cpp"
  if (!($35)) {
   STACKTOP = sp;return ($13|0); //@line 160 "modules/math/Compressor.cpp"
  }
  __ZdaPv($13); //@line 141 "modules/math/Compressor.cpp"
  $36 = (___cxa_allocate_exception(16)|0); //@line 142 "modules/math/Compressor.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($36|0),(355399|0),($vararg_buffer5|0)); //@line 142 "modules/math/Compressor.cpp"
  $37 = __THREW__; __THREW__ = 0;
  $38 = $37&1;
  if (!($38)) {
   ___cxa_throw(($36|0),(96|0),(344|0)); //@line 142 "modules/math/Compressor.cpp"
   // unreachable; //@line 142 "modules/math/Compressor.cpp"
  }
  $39 = ___cxa_find_matching_catch_2()|0;
  $40 = tempRet0;
  ___cxa_free_exception(($36|0)); //@line 142 "modules/math/Compressor.cpp"
  $$1 = $39;$$12 = $40;
  ___resumeException($$1|0);
  // unreachable;
 }
 $41 = (($dataSize) + -4)|0; //@line 149 "modules/math/Compressor.cpp"
 $42 = (_LZ4_decompress_safe($19,$13,$41,$12)|0); //@line 148 "modules/math/Compressor.cpp"
 $43 = ($42|0)<(0); //@line 151 "modules/math/Compressor.cpp"
 if (!($43)) {
  HEAP32[$decompressedSize>>2] = $42; //@line 157 "modules/math/Compressor.cpp"
  STACKTOP = sp;return ($13|0); //@line 160 "modules/math/Compressor.cpp"
 }
 __ZdaPv($13); //@line 153 "modules/math/Compressor.cpp"
 $44 = (___cxa_allocate_exception(16)|0); //@line 154 "modules/math/Compressor.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($44|0),(355399|0),($vararg_buffer7|0)); //@line 154 "modules/math/Compressor.cpp"
 $45 = __THREW__; __THREW__ = 0;
 $46 = $45&1;
 if (!($46)) {
  ___cxa_throw(($44|0),(96|0),(344|0)); //@line 154 "modules/math/Compressor.cpp"
  // unreachable; //@line 154 "modules/math/Compressor.cpp"
 }
 $47 = ___cxa_find_matching_catch_2()|0;
 $48 = tempRet0;
 ___cxa_free_exception(($44|0)); //@line 154 "modules/math/Compressor.cpp"
 $$1 = $47;$$12 = $48;
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNK4love4math13LZ4Compressor11isSupportedENS0_10Compressor6FormatE($this,$format) {
 $this = $this|0;
 $format = $format|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($format|0)==(0); //@line 165 "modules/math/Compressor.cpp"
 return ($0|0); //@line 165 "modules/math/Compressor.cpp"
}
function __ZN4love4math14zlibCompressorD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 171 "modules/math/Compressor.cpp"
 return; //@line 171 "modules/math/Compressor.cpp"
}
function __ZN4love4math14zlibCompressor8compressENS0_10Compressor6FormatEPKcjiRj($this,$format,$data,$dataSize,$level,$compressedSize) {
 $this = $this|0;
 $format = $format|0;
 $data = $data|0;
 $dataSize = $dataSize|0;
 $level = $level|0;
 $compressedSize = $compressedSize|0;
 var $$$i = 0, $$$i4 = 0, $$$v$i = 0, $$0 = 0, $$01 = 0, $$02 = 0, $$1 = 0, $$13 = 0, $$level = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $compressedbytes$1 = 0, $stream$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $stream$i = sp + 24|0;
 $0 = HEAP32[$this>>2]|0; //@line 259 "modules/math/Compressor.cpp"
 $1 = ((($0)) + 16|0); //@line 259 "modules/math/Compressor.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 259 "modules/math/Compressor.cpp"
 $3 = (FUNCTION_TABLE_iii[$2 & 8191]($this,$format)|0); //@line 259 "modules/math/Compressor.cpp"
 if (!($3)) {
  $4 = (___cxa_allocate_exception(16)|0); //@line 260 "modules/math/Compressor.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($4|0),(355470|0),($vararg_buffer|0)); //@line 260 "modules/math/Compressor.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   ___cxa_throw(($4|0),(96|0),(344|0)); //@line 260 "modules/math/Compressor.cpp"
   // unreachable; //@line 260 "modules/math/Compressor.cpp"
  }
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  ___cxa_free_exception(($4|0)); //@line 260 "modules/math/Compressor.cpp"
  $$1 = $7;$$13 = $8;
  ___resumeException($$1|0);
  // unreachable;
 }
 $9 = $dataSize >>> 12; //@line 181 "modules/math/Compressor.cpp"
 $10 = $dataSize >>> 14; //@line 181 "modules/math/Compressor.cpp"
 $11 = $dataSize >>> 25; //@line 181 "modules/math/Compressor.cpp"
 $12 = ($format|0)==(2); //@line 184 "modules/math/Compressor.cpp"
 $$$v$i = $12 ? 25 : 13; //@line 184 "modules/math/Compressor.cpp"
 $13 = (($$$v$i) + ($dataSize))|0; //@line 181 "modules/math/Compressor.cpp"
 $14 = (($13) + ($9))|0; //@line 181 "modules/math/Compressor.cpp"
 $15 = (($14) + ($10))|0; //@line 181 "modules/math/Compressor.cpp"
 $$$i = (($15) + ($11))|0; //@line 184 "modules/math/Compressor.cpp"
 __THREW__ = 0;
 $16 = (invoke_ii(3492,($$$i|0))|0); //@line 272 "modules/math/Compressor.cpp"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if ($18) {
  $32 = ___cxa_find_matching_catch_3(9056|0)|0;
  $33 = tempRet0;
  $34 = (_llvm_eh_typeid_for((9056|0))|0); //@line 273 "modules/math/Compressor.cpp"
  $35 = ($33|0)==($34|0); //@line 273 "modules/math/Compressor.cpp"
  if (!($35)) {
   $$1 = $32;$$13 = $33;
   ___resumeException($$1|0);
   // unreachable;
  }
  (___cxa_begin_catch(($32|0))|0); //@line 273 "modules/math/Compressor.cpp"
  $36 = (___cxa_allocate_exception(16)|0); //@line 276 "modules/math/Compressor.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($36|0),(507610|0),($vararg_buffer1|0)); //@line 276 "modules/math/Compressor.cpp"
  $37 = __THREW__; __THREW__ = 0;
  $38 = $37&1;
  if ($38) {
   $42 = ___cxa_find_matching_catch_2()|0;
   $43 = tempRet0;
   ___cxa_free_exception(($36|0)); //@line 276 "modules/math/Compressor.cpp"
   $$01 = $42;$$02 = $43;
  } else {
   __THREW__ = 0;
   invoke_viii(3350,($36|0),(96|0),(344|0)); //@line 276 "modules/math/Compressor.cpp"
   $39 = __THREW__; __THREW__ = 0;
   $40 = ___cxa_find_matching_catch_2()|0;
   $41 = tempRet0;
   $$01 = $40;$$02 = $41;
  }
  __THREW__ = 0;
  invoke_v(3378); //@line 277 "modules/math/Compressor.cpp"
  $44 = __THREW__; __THREW__ = 0;
  $45 = $44&1;
  if ($45) {
   $57 = ___cxa_find_matching_catch_3(0|0)|0;
   $58 = tempRet0;
   ___clang_call_terminate($57); //@line 277 "modules/math/Compressor.cpp"
   // unreachable; //@line 277 "modules/math/Compressor.cpp"
  } else {
   $$1 = $$01;$$13 = $$02;
   ___resumeException($$1|0);
   // unreachable;
  }
 }
 $19 = ($level|0)<(0); //@line 262 "modules/math/Compressor.cpp"
 $20 = ($level|0)>(9); //@line 264 "modules/math/Compressor.cpp"
 $$level = $20 ? 9 : $level; //@line 264 "modules/math/Compressor.cpp"
 $$0 = $19 ? -1 : $$level; //@line 262 "modules/math/Compressor.cpp"
 dest=$stream$i; stop=dest+56|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 192 "modules/math/Compressor.cpp"
 HEAP32[$stream$i>>2] = $data; //@line 194 "modules/math/Compressor.cpp"
 $21 = ((($stream$i)) + 4|0); //@line 195 "modules/math/Compressor.cpp"
 HEAP32[$21>>2] = $dataSize; //@line 195 "modules/math/Compressor.cpp"
 $22 = ((($stream$i)) + 12|0); //@line 197 "modules/math/Compressor.cpp"
 HEAP32[$22>>2] = $16; //@line 197 "modules/math/Compressor.cpp"
 $23 = ((($stream$i)) + 16|0); //@line 198 "modules/math/Compressor.cpp"
 HEAP32[$23>>2] = $$$i; //@line 198 "modules/math/Compressor.cpp"
 $$$i4 = $12 ? 31 : 15; //@line 201 "modules/math/Compressor.cpp"
 $24 = (_deflateInit2_($stream$i,$$0,8,$$$i4,8,0,613577,56)|0); //@line 204 "modules/math/Compressor.cpp"
 $25 = ($24|0)==(0); //@line 206 "modules/math/Compressor.cpp"
 do {
  if ($25) {
   $26 = (_deflate($stream$i,4)|0); //@line 209 "modules/math/Compressor.cpp"
   $27 = ($26|0)==(1); //@line 211 "modules/math/Compressor.cpp"
   if (!($27)) {
    (_deflateEnd($stream$i)|0); //@line 213 "modules/math/Compressor.cpp"
    label = 9;
    break;
   }
   $28 = ((($stream$i)) + 20|0); //@line 217 "modules/math/Compressor.cpp"
   $29 = HEAP32[$28>>2]|0; //@line 217 "modules/math/Compressor.cpp"
   $30 = (_deflateEnd($stream$i)|0); //@line 219 "modules/math/Compressor.cpp"
   $31 = ($30|0)==(0); //@line 282 "modules/math/Compressor.cpp"
   if ($31) {
    $51 = (+($$$i>>>0)); //@line 291 "modules/math/Compressor.cpp"
    $52 = (+($29>>>0)); //@line 291 "modules/math/Compressor.cpp"
    $53 = $51 / $52; //@line 291 "modules/math/Compressor.cpp"
    $54 = !($53 >= 1.3); //@line 291 "modules/math/Compressor.cpp"
    if ($54) {
     $compressedbytes$1 = $16;
     HEAP32[$compressedSize>>2] = $29; //@line 302 "modules/math/Compressor.cpp"
     STACKTOP = sp;return ($compressedbytes$1|0); //@line 303 "modules/math/Compressor.cpp"
    }
    $55 = (__ZnajRKSt9nothrow_t($29,1282467)|0); //@line 293 "modules/math/Compressor.cpp"
    $56 = ($55|0)==(0|0); //@line 294 "modules/math/Compressor.cpp"
    if ($56) {
     $compressedbytes$1 = $16;
     HEAP32[$compressedSize>>2] = $29; //@line 302 "modules/math/Compressor.cpp"
     STACKTOP = sp;return ($compressedbytes$1|0); //@line 303 "modules/math/Compressor.cpp"
    }
    _memcpy(($55|0),($16|0),($29|0))|0; //@line 296 "modules/math/Compressor.cpp"
    __ZdaPv($16); //@line 297 "modules/math/Compressor.cpp"
    $compressedbytes$1 = $55;
    HEAP32[$compressedSize>>2] = $29; //@line 302 "modules/math/Compressor.cpp"
    STACKTOP = sp;return ($compressedbytes$1|0); //@line 303 "modules/math/Compressor.cpp"
   }
  } else {
   label = 9;
  }
 } while(0);
 if ((label|0) == 9) {
 }
 __ZdaPv($16); //@line 284 "modules/math/Compressor.cpp"
 $46 = (___cxa_allocate_exception(16)|0); //@line 285 "modules/math/Compressor.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($46|0),(355510|0),($vararg_buffer3|0)); //@line 285 "modules/math/Compressor.cpp"
 $47 = __THREW__; __THREW__ = 0;
 $48 = $47&1;
 if (!($48)) {
  ___cxa_throw(($46|0),(96|0),(344|0)); //@line 285 "modules/math/Compressor.cpp"
  // unreachable; //@line 285 "modules/math/Compressor.cpp"
 }
 $49 = ___cxa_find_matching_catch_2()|0;
 $50 = tempRet0;
 ___cxa_free_exception(($46|0)); //@line 285 "modules/math/Compressor.cpp"
 $$1 = $49;$$13 = $50;
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love4math14zlibCompressor10decompressENS0_10Compressor6FormatEPKcjRj($this,$format,$data,$dataSize,$decompressedSize) {
 $this = $this|0;
 $format = $format|0;
 $data = $data|0;
 $dataSize = $dataSize|0;
 $decompressedSize = $decompressedSize|0;
 var $$0$i = 0, $$01 = 0, $$02 = 0, $$1 = 0, $$13 = 0, $$lcssa = 0, $$lcssa24 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $destLen$0 = 0, $destLen$0$lcssa25 = 0;
 var $rawsize$0 = 0, $stream$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $stream$i = sp + 24|0;
 $0 = HEAP32[$this>>2]|0; //@line 308 "modules/math/Compressor.cpp"
 $1 = ((($0)) + 16|0); //@line 308 "modules/math/Compressor.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 308 "modules/math/Compressor.cpp"
 $3 = (FUNCTION_TABLE_iii[$2 & 8191]($this,$format)|0); //@line 308 "modules/math/Compressor.cpp"
 if (!($3)) {
  $4 = (___cxa_allocate_exception(16)|0); //@line 309 "modules/math/Compressor.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($4|0),(355470|0),($vararg_buffer|0)); //@line 309 "modules/math/Compressor.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   ___cxa_throw(($4|0),(96|0),(344|0)); //@line 309 "modules/math/Compressor.cpp"
   // unreachable; //@line 309 "modules/math/Compressor.cpp"
  }
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  ___cxa_free_exception(($4|0)); //@line 309 "modules/math/Compressor.cpp"
  $$1 = $7;$$13 = $8;
  ___resumeException($$1|0);
  // unreachable;
 }
 $9 = HEAP32[$decompressedSize>>2]|0; //@line 314 "modules/math/Compressor.cpp"
 $10 = ($9|0)==(0); //@line 314 "modules/math/Compressor.cpp"
 $11 = $dataSize << 1; //@line 314 "modules/math/Compressor.cpp"
 $12 = $10 ? $11 : $9; //@line 314 "modules/math/Compressor.cpp"
 $13 = ((($stream$i)) + 4|0); //@line 227 "modules/math/Compressor.cpp"
 $14 = ((($stream$i)) + 12|0); //@line 229 "modules/math/Compressor.cpp"
 $15 = ((($stream$i)) + 16|0); //@line 230 "modules/math/Compressor.cpp"
 $16 = ((($stream$i)) + 20|0); //@line 250 "modules/math/Compressor.cpp"
 $rawsize$0 = $12;
 L8: while(1) {
  __THREW__ = 0;
  $17 = (invoke_ii(3492,($rawsize$0|0))|0); //@line 321 "modules/math/Compressor.cpp"
  $18 = __THREW__; __THREW__ = 0;
  $19 = $18&1;
  if ($19) {
   label = 15;
   break;
  }
  dest=$stream$i; stop=dest+56|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 224 "modules/math/Compressor.cpp"
  HEAP32[$stream$i>>2] = $data; //@line 226 "modules/math/Compressor.cpp"
  HEAP32[$13>>2] = $dataSize; //@line 227 "modules/math/Compressor.cpp"
  HEAP32[$14>>2] = $17; //@line 229 "modules/math/Compressor.cpp"
  HEAP32[$15>>2] = $rawsize$0; //@line 230 "modules/math/Compressor.cpp"
  $20 = (_inflateInit2_($stream$i,47,613577,56)|0); //@line 235 "modules/math/Compressor.cpp"
  $21 = ($20|0)==(0); //@line 237 "modules/math/Compressor.cpp"
  L11: do {
   if ($21) {
    $22 = (_inflate($stream$i,4)|0); //@line 240 "modules/math/Compressor.cpp"
    $23 = ($22|0)==(1); //@line 242 "modules/math/Compressor.cpp"
    if ($23) {
     $26 = HEAP32[$16>>2]|0; //@line 250 "modules/math/Compressor.cpp"
     $27 = (_inflateEnd($stream$i)|0); //@line 252 "modules/math/Compressor.cpp"
     $$0$i = $27;$destLen$0 = $26;
     label = 14;
     break;
    }
    (_inflateEnd($stream$i)|0); //@line 244 "modules/math/Compressor.cpp"
    switch ($22|0) {
    case 2:  {
     $$lcssa = $17;
     label = 13;
     break L8;
     break;
    }
    case -5:  {
     break;
    }
    default: {
     $$0$i = $22;$destLen$0 = $rawsize$0;
     label = 14;
     break L11;
    }
    }
    $24 = HEAP32[$13>>2]|0; //@line 245 "modules/math/Compressor.cpp"
    $25 = ($24|0)==(0); //@line 245 "modules/math/Compressor.cpp"
    if ($25) {
     $$lcssa = $17;
     label = 13;
     break L8;
    }
   } else {
    $$0$i = $20;$destLen$0 = $rawsize$0;
    label = 14;
   }
  } while(0);
  if ((label|0) == 14) {
   label = 0;
   switch ($$0$i|0) {
   case 0:  {
    $$lcssa24 = $17;$destLen$0$lcssa25 = $destLen$0;
    label = 24;
    break L8;
    break;
   }
   case -5:  {
    break;
   }
   default: {
    $42 = $17;
    break L8;
   }
   }
  }
  __ZdaPv($17); //@line 344 "modules/math/Compressor.cpp"
  $48 = $rawsize$0 << 1; //@line 345 "modules/math/Compressor.cpp"
  $rawsize$0 = $48;
 }
 if ((label|0) == 13) {
  $42 = $$lcssa;
 }
 else if ((label|0) == 15) {
  $28 = ___cxa_find_matching_catch_3(9056|0)|0;
  $29 = tempRet0;
  $30 = (_llvm_eh_typeid_for((9056|0))|0); //@line 322 "modules/math/Compressor.cpp"
  $31 = ($29|0)==($30|0); //@line 322 "modules/math/Compressor.cpp"
  if (!($31)) {
   $$1 = $28;$$13 = $29;
   ___resumeException($$1|0);
   // unreachable;
  }
  (___cxa_begin_catch(($28|0))|0); //@line 322 "modules/math/Compressor.cpp"
  $32 = (___cxa_allocate_exception(16)|0); //@line 325 "modules/math/Compressor.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($32|0),(507610|0),($vararg_buffer1|0)); //@line 325 "modules/math/Compressor.cpp"
  $33 = __THREW__; __THREW__ = 0;
  $34 = $33&1;
  if ($34) {
   $38 = ___cxa_find_matching_catch_2()|0;
   $39 = tempRet0;
   ___cxa_free_exception(($32|0)); //@line 325 "modules/math/Compressor.cpp"
   $$01 = $38;$$02 = $39;
  } else {
   __THREW__ = 0;
   invoke_viii(3350,($32|0),(96|0),(344|0)); //@line 325 "modules/math/Compressor.cpp"
   $35 = __THREW__; __THREW__ = 0;
   $36 = ___cxa_find_matching_catch_2()|0;
   $37 = tempRet0;
   $$01 = $36;$$02 = $37;
  }
  __THREW__ = 0;
  invoke_v(3378); //@line 326 "modules/math/Compressor.cpp"
  $40 = __THREW__; __THREW__ = 0;
  $41 = $40&1;
  if ($41) {
   $49 = ___cxa_find_matching_catch_3(0|0)|0;
   $50 = tempRet0;
   ___clang_call_terminate($49); //@line 326 "modules/math/Compressor.cpp"
   // unreachable; //@line 326 "modules/math/Compressor.cpp"
  } else {
   $$1 = $$01;$$13 = $$02;
   ___resumeException($$1|0);
   // unreachable;
  }
 }
 else if ((label|0) == 24) {
  HEAP32[$decompressedSize>>2] = $destLen$0$lcssa25; //@line 333 "modules/math/Compressor.cpp"
  STACKTOP = sp;return ($$lcssa24|0); //@line 348 "modules/math/Compressor.cpp"
 }
 __ZdaPv($42); //@line 339 "modules/math/Compressor.cpp"
 $43 = (___cxa_allocate_exception(16)|0); //@line 340 "modules/math/Compressor.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($43|0),(355545|0),($vararg_buffer3|0)); //@line 340 "modules/math/Compressor.cpp"
 $44 = __THREW__; __THREW__ = 0;
 $45 = $44&1;
 if (!($45)) {
  ___cxa_throw(($43|0),(96|0),(344|0)); //@line 340 "modules/math/Compressor.cpp"
  // unreachable; //@line 340 "modules/math/Compressor.cpp"
 }
 $46 = ___cxa_find_matching_catch_2()|0;
 $47 = tempRet0;
 ___cxa_free_exception(($43|0)); //@line 340 "modules/math/Compressor.cpp"
 $$1 = $46;$$13 = $47;
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNK4love4math14zlibCompressor11isSupportedENS0_10Compressor6FormatE($this,$format) {
 $this = $this|0;
 $format = $format|0;
 var $0 = 0, $format$off = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $format$off = (($format) + -1)|0; //@line 353 "modules/math/Compressor.cpp"
 $0 = ($format$off>>>0)<(2); //@line 353 "modules/math/Compressor.cpp"
 return ($0|0); //@line 353 "modules/math/Compressor.cpp"
}
function __ZN4love4math10Compressor11getConstantEPKcRNS1_6FormatE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 6)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((630500 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (630500 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(6); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((630500 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 376 "modules/math/Compressor.cpp"
  return ($$$2$i|0); //@line 376 "modules/math/Compressor.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 376 "modules/math/Compressor.cpp"
  return ($$$2$i|0); //@line 376 "modules/math/Compressor.cpp"
 }
 return (0)|0;
}
function __ZN4love4math10Compressor11getConstantENS1_6FormatERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(2); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (630572 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 381 "modules/math/Compressor.cpp"
}
function __ZN4love4math4MathD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (16648); //@line 95 "modules/math/MathModule.cpp"
 __ZN4love6ModuleD2Ev($this); //@line 96 "modules/math/MathModule.cpp"
 return; //@line 96 "modules/math/MathModule.cpp"
}
function __ZN4love4math4MathD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (16648); //@line 95 "modules/math/MathModule.cpp"
 __ZN4love6ModuleD2Ev($this); //@line 96 "modules/math/MathModule.cpp"
 __ZdlPv($this); //@line 95 "modules/math/MathModule.cpp"
 return; //@line 96 "modules/math/MathModule.cpp"
}
function __ZNK4love4math4Math13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 8; //@line 76 "modules/math/MathModule.h"
}
function __ZNK4love4math4Math7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (355611|0); //@line 81 "modules/math/MathModule.h"
}
function __GLOBAL__sub_I_MathModule_cpp() {
 var $$0$i$i = 0, $$01$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev(623984); //@line 33 "./common/Module.h"
 HEAP32[155996] = (16648); //@line 89 "modules/math/MathModule.cpp"
 __THREW__ = 0;
 invoke_vi(3653,((623992)|0)); //@line 88 "modules/math/MathModule.cpp"
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 if ($1) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  $$0$i$i = $8;$$01$i$i = $7;
  __ZN4love6ModuleD2Ev(623984); //@line 92 "modules/math/MathModule.cpp"
  ___resumeException($$01$i$i|0);
  // unreachable;
 }
 $2 = HEAP32[155996]|0; //@line 91 "modules/math/MathModule.cpp"
 $3 = ((($2)) + 8|0); //@line 91 "modules/math/MathModule.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 91 "modules/math/MathModule.cpp"
 __THREW__ = 0;
 invoke_vi($4|0,(623984|0)); //@line 91 "modules/math/MathModule.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  (___cxa_atexit((771|0),(623984|0),(___dso_handle|0))|0); //@line 85 "modules/math/MathModule.cpp"
  return;
 }
 $9 = ___cxa_find_matching_catch_2()|0;
 $10 = tempRet0;
 $$0$i$i = $10;$$01$i$i = $9;
 __ZN4love6ModuleD2Ev(623984); //@line 92 "modules/math/MathModule.cpp"
 ___resumeException($$01$i$i|0);
 // unreachable;
}
function __ZN4love4math4Math18newRandomGeneratorEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(32)|0); //@line 100 "modules/math/MathModule.cpp"
 __THREW__ = 0;
 invoke_vi(3653,($0|0)); //@line 100 "modules/math/MathModule.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 100 "modules/math/MathModule.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 100 "modules/math/MathModule.cpp"
 }
 return (0)|0;
}
function __ZN4love4math4Math14newBezierCurveERKNSt3__16vectorINS_6VectorENS2_9allocatorIS4_EEEE($this,$points) {
 $this = $this|0;
 $points = $points|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(20)|0); //@line 105 "modules/math/MathModule.cpp"
 __THREW__ = 0;
 invoke_vii(3654,($0|0),($points|0)); //@line 105 "modules/math/MathModule.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 105 "modules/math/MathModule.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 105 "modules/math/MathModule.cpp"
 }
 return (0)|0;
}
function __ZN4love4math4Math11triangulateERKNSt3__16vectorINS_6VertexENS2_9allocatorIS4_EEEE($agg$result,$this,$polygon) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $polygon = $polygon|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$3 = 0, $$34 = 0, $$cast = 0, $$cast56 = 0, $$idx = 0, $$idx$val = 0.0, $$idx$val$i$i$us = 0.0, $$idx10 = 0, $$idx10$val = 0.0, $$idx11 = 0, $$idx11$val = 0.0, $$idx12 = 0, $$idx12$val = 0.0, $$idx13 = 0;
 var $$idx13$val = 0.0, $$idx14 = 0, $$idx14$val = 0.0, $$idx15 = 0, $$idx15$val = 0.0, $$idx3$i$i$us = 0, $$idx3$val$i$i$us = 0.0, $$idx5 = 0, $$idx5$val = 0.0, $$idx6$val = 0.0, $$idx7 = 0, $$idx7$val = 0.0, $$idx8 = 0, $$idx8$val = 0.0, $$idx9 = 0, $$idx9$val = 0.0, $$lcssa184 = 0, $$lcssa187 = 0, $$lcssa190 = 0, $$pn = 0;
 var $$pn$in = 0, $$pre = 0, $$pre$phi139Z2D = 0, $$pre$phiZ2D = 0, $$pre111 = 0, $$pre112 = 0, $$pre113 = 0, $$pre114 = 0, $$pre116 = 0, $$pre119 = 0, $$pre121 = 0, $$pre126 = 0, $$pre127 = 0, $$pre136 = 0, $$pre138 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0, $114 = 0, $115 = 0.0, $116 = 0, $117 = 0.0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0.0, $122 = 0, $123 = 0.0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0.0, $168 = 0.0, $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0;
 var $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0.0, $182 = 0.0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0.0, $19 = 0, $190 = 0.0, $191 = 0.0, $192 = 0.0;
 var $193 = 0.0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0.0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0.0, $201 = 0.0, $202 = 0, $203 = 0.0, $204 = 0.0, $205 = 0.0, $206 = 0.0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0.0, $223 = 0.0, $224 = 0.0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0.0;
 var $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0;
 var $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0;
 var $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0;
 var $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0;
 var $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0;
 var $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0;
 var $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0;
 var $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0;
 var $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__f$01$i$i = 0, $__f$01$i$i44 = 0, $a$idx$val$i$us = 0.0, $a$idx1$i$us = 0, $a$idx1$val$i$us = 0.0, $b$idx$i$us$phi$trans$insert = 0, $b$idx$val$i$us = 0.0;
 var $b$idx$val$i$us$pre = 0.0, $b$idx2$i$us$phi$trans$insert = 0, $b$idx2$val$i$us = 0.0, $b$idx2$val$i$us$pre = 0.0, $c$idx$val$i$us = 0.0, $c$idx3$i$us = 0, $c$idx3$val$i$us = 0.0, $concave_vertices = 0, $current$0$ph$lcssa = 0, $current$0$ph196 = 0, $current$0$us = 0, $i$077 = 0, $i1$076 = 0, $idx_lm$0$lcssa = 0, $idx_lm$078 = 0, $idx_lm$1 = 0, $idx_lm$1$lcssa = 0, $it$sroa$0$0$i$i$us = 0, $it$sroa$0$08$i$i = 0, $it$sroa$0$08$i$i193 = 0;
 var $it$sroa$0$09$i$i$us = 0, $lpad$phi$index = 0, $lpad$phi$index4 = 0, $n_vertices$0$ph194 = 0, $next_idx = 0, $or$cond$i$i$us = 0, $or$cond2$i$i$us = 0, $prev_idx = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i$i34 = 0, $scevgep$i$i$i$i38 = 0, $scevgep$i$i$i16 = 0, $scevgep$i$i$i21 = 0, $scevgep$i$i$i24 = 0, $scevgep$i$i$i29 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i$i35 = 0, $scevgep4$i$i$i$i39 = 0, $scevgep4$i$i$i17 = 0, $scevgep4$i$i$i22 = 0;
 var $scevgep4$i$i$i25 = 0, $scevgep4$i$i$i30 = 0, $skipped$0$us = 0, $triangles = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 184|0;
 $next_idx = sp + 172|0;
 $prev_idx = sp + 160|0;
 $concave_vertices = sp + 148|0;
 $triangles = sp + 136|0;
 $1 = sp + 76|0;
 $2 = sp + 72|0;
 $3 = sp + 12|0;
 $4 = ((($polygon)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$4>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$polygon>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = (($5) - ($6))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = (($7|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ($8>>>0)<(3); //@line 110 "modules/math/MathModule.cpp"
 $10 = $6; //@line 110 "modules/math/MathModule.cpp"
 if ($9) {
  $11 = (___cxa_allocate_exception(16)|0); //@line 111 "modules/math/MathModule.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($11|0),(355621|0),($vararg_buffer|0)); //@line 111 "modules/math/MathModule.cpp"
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   ___cxa_throw(($11|0),(96|0),(344|0)); //@line 111 "modules/math/MathModule.cpp"
   // unreachable; //@line 111 "modules/math/MathModule.cpp"
  }
  $14 = ___cxa_find_matching_catch_2()|0;
  $15 = tempRet0;
  ___cxa_free_exception(($11|0)); //@line 111 "modules/math/MathModule.cpp"
  $$3 = $14;$$34 = $15;
  ___resumeException($$3|0);
  // unreachable;
 }
 $16 = ($7|0)==(60); //@line 112 "modules/math/MathModule.cpp"
 if ($16) {
  $17 = ((($10)) + 20|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ((($10)) + 40|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = $6; //@line 74 "./common/math.h"
  ;HEAP32[$0>>2]=HEAP32[$19>>2]|0;HEAP32[$0+4>>2]=HEAP32[$19+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$19+8>>2]|0;HEAP32[$0+12>>2]=HEAP32[$19+12>>2]|0;HEAP32[$0+16>>2]=HEAP32[$19+16>>2]|0; //@line 74 "./common/math.h"
  $20 = ((($0)) + 20|0); //@line 74 "./common/math.h"
  ;HEAP32[$20>>2]=HEAP32[$17>>2]|0;HEAP32[$20+4>>2]=HEAP32[$17+4>>2]|0;HEAP32[$20+8>>2]=HEAP32[$17+8>>2]|0;HEAP32[$20+12>>2]=HEAP32[$17+12>>2]|0;HEAP32[$20+16>>2]=HEAP32[$17+16>>2]|0; //@line 74 "./common/math.h"
  $21 = ((($0)) + 40|0); //@line 74 "./common/math.h"
  ;HEAP32[$21>>2]=HEAP32[$18>>2]|0;HEAP32[$21+4>>2]=HEAP32[$18+4>>2]|0;HEAP32[$21+8>>2]=HEAP32[$18+8>>2]|0;HEAP32[$21+12>>2]=HEAP32[$18+12>>2]|0;HEAP32[$21+16>>2]=HEAP32[$18+16>>2]|0; //@line 74 "./common/math.h"
  HEAP32[$agg$result>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = ((($agg$result)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$22>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = ((($agg$result)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$23>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = (__Znwj(60)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  HEAP32[$agg$result>>2] = $24; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $25 = ((($24)) + 60|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $26 = ((($agg$result)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$26>>2] = $25; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  dest=$24; src=$0; stop=dest+60|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $27 = ((($24)) + 60|0); //@line 1011 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$22>>2] = $27; //@line 1011 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  STACKTOP = sp;return; //@line 171 "modules/math/MathModule.cpp"
 }
 HEAP32[$next_idx>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $28 = ((($next_idx)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$28>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $29 = ((($next_idx)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$29>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ($5|0)==($6|0); //@line 1080 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($30) {
   $369 = $10;$53 = $5;$54 = $5;
  } else {
   $31 = ($8>>>0)>(1073741823); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($31) {
    __THREW__ = 0;
    invoke_vi(3528,($next_idx|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $32 = __THREW__; __THREW__ = 0;
    $33 = $32&1;
    if (!($33)) {
     label = 10;
    }
   } else {
    label = 10;
   }
   if ((label|0) == 10) {
    $34 = $8 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    __THREW__ = 0;
    $35 = (invoke_ii(3348,($34|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $36 = __THREW__; __THREW__ = 0;
    $37 = $36&1;
    if (!($37)) {
     HEAP32[$next_idx>>2] = $35; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $38 = (($35) + ($8<<2)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $39 = ((($next_idx)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$39>>2] = $38; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     _memset(($35|0),0,($34|0))|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$28>>2] = $38; //@line 988 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $$pre = HEAP32[$4>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $$pre111 = HEAP32[$polygon>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $40 = $$pre111; //@line 117 "modules/math/MathModule.cpp"
     $369 = $40;$53 = $$pre;$54 = $$pre111;
     break;
    }
   }
   $41 = ___cxa_find_matching_catch_2()|0;
   $42 = tempRet0;
   $43 = HEAP32[$next_idx>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $44 = ($43|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $45 = $43; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($44) {
    ___resumeException($41|0);
    // unreachable;
   }
   $46 = HEAP32[$28>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $47 = ($46|0)==($43|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($47)) {
    $scevgep$i$i$i$i34 = ((($46)) + -4|0);
    $48 = $scevgep$i$i$i$i34;
    $49 = (($48) - ($45))|0;
    $50 = $49 >>> 2;
    $51 = $50 ^ -1;
    $scevgep4$i$i$i$i35 = (($46) + ($51<<2)|0);
    HEAP32[$28>>2] = $scevgep4$i$i$i$i35; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($43); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($41|0);
   // unreachable;
  }
 } while(0);
 $52 = (($53) - ($54))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $55 = (($52|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$prev_idx>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $56 = ((($prev_idx)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$56>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $57 = ((($prev_idx)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$57>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $58 = ($53|0)==($54|0); //@line 1080 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($58) {
   $370 = $54;$371 = 0;$372 = 0;$373 = 0;$374 = $369;$375 = $54;
   label = 22;
  } else {
   $59 = ($55>>>0)>(1073741823); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($59) {
    __THREW__ = 0;
    invoke_vi(3528,($prev_idx|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $60 = __THREW__; __THREW__ = 0;
    $61 = $60&1;
    if (!($61)) {
     label = 20;
    }
   } else {
    label = 20;
   }
   if ((label|0) == 20) {
    $62 = $55 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    __THREW__ = 0;
    $63 = (invoke_ii(3348,($62|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $64 = __THREW__; __THREW__ = 0;
    $65 = $64&1;
    if (!($65)) {
     HEAP32[$prev_idx>>2] = $63; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $66 = (($63) + ($55<<2)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $67 = ((($prev_idx)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$67>>2] = $66; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     _memset(($63|0),0,($62|0))|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$56>>2] = $66; //@line 988 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $$pre112 = HEAP32[$4>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $$pre113 = HEAP32[$polygon>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $68 = $$pre113; //@line 117 "modules/math/MathModule.cpp"
     $69 = $63; //@line 117 "modules/math/MathModule.cpp"
     $70 = $66; //@line 117 "modules/math/MathModule.cpp"
     $71 = ($$pre112|0)==($$pre113|0); //@line 119 "modules/math/MathModule.cpp"
     if ($71) {
      $370 = $$pre113;$371 = $70;$372 = $69;$373 = $63;$374 = $68;$375 = $$pre112;
      label = 22;
      break;
     }
     $73 = HEAP32[$next_idx>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $74 = (($$pre112) - ($$pre113))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $75 = (($74|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $76 = $73; //@line 119 "modules/math/MathModule.cpp"
     $i$077 = 0;$idx_lm$078 = 0;
     while(1) {
      $114 = (($68) + (($i$077*20)|0)|0); //@line 122 "modules/math/MathModule.cpp"
      $115 = +HEAPF32[$114>>2]; //@line 122 "modules/math/MathModule.cpp"
      $116 = (($68) + (($idx_lm$078*20)|0)|0); //@line 122 "modules/math/MathModule.cpp"
      $117 = +HEAPF32[$116>>2]; //@line 122 "modules/math/MathModule.cpp"
      $118 = $115 < $117; //@line 122 "modules/math/MathModule.cpp"
      if ($118) {
       label = 34;
      } else {
       $119 = $115 == $117; //@line 122 "modules/math/MathModule.cpp"
       if ($119) {
        $120 = (((($68) + (($i$077*20)|0)|0)) + 4|0); //@line 122 "modules/math/MathModule.cpp"
        $121 = +HEAPF32[$120>>2]; //@line 122 "modules/math/MathModule.cpp"
        $122 = (((($68) + (($idx_lm$078*20)|0)|0)) + 4|0); //@line 122 "modules/math/MathModule.cpp"
        $123 = +HEAPF32[$122>>2]; //@line 122 "modules/math/MathModule.cpp"
        $124 = $121 < $123; //@line 122 "modules/math/MathModule.cpp"
        if ($124) {
         label = 34;
        } else {
         $idx_lm$1 = $idx_lm$078;
        }
       } else {
        $idx_lm$1 = $idx_lm$078;
       }
      }
      if ((label|0) == 34) {
       label = 0;
       $idx_lm$1 = $i$077;
      }
      $125 = (($i$077) + 1)|0; //@line 124 "modules/math/MathModule.cpp"
      $126 = (($73) + ($i$077<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$126>>2] = $125; //@line 124 "modules/math/MathModule.cpp"
      $127 = (($i$077) + -1)|0; //@line 125 "modules/math/MathModule.cpp"
      $128 = (($63) + ($i$077<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$128>>2] = $127; //@line 125 "modules/math/MathModule.cpp"
      $129 = ($125>>>0)<($75>>>0); //@line 119 "modules/math/MathModule.cpp"
      if ($129) {
       $i$077 = $125;$idx_lm$078 = $idx_lm$1;
      } else {
       $idx_lm$1$lcssa = $idx_lm$1;
       break;
      }
     }
     $$pre116 = HEAP32[$next_idx>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $$pre$phiZ2D = $next_idx;$100 = $63;$103 = $$pre116;$105 = $68;$142 = $$pre112;$143 = $$pre113;$90 = $76;$95 = $70;$96 = $69;$idx_lm$0$lcssa = $idx_lm$1$lcssa;
     label = 29;
     break;
    }
   }
   $77 = ___cxa_find_matching_catch_2()|0;
   $78 = tempRet0;
   $79 = HEAP32[$prev_idx>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $80 = ($79|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $81 = $79; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($80)) {
    $82 = HEAP32[$56>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $83 = ($82|0)==($79|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($83)) {
     $scevgep$i$i$i$i38 = ((($82)) + -4|0);
     $84 = $scevgep$i$i$i$i38;
     $85 = (($84) - ($81))|0;
     $86 = $85 >>> 2;
     $87 = $86 ^ -1;
     $scevgep4$i$i$i$i39 = (($82) + ($87<<2)|0);
     HEAP32[$56>>2] = $scevgep4$i$i$i$i39; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($79); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $$2 = $77;$$23 = $78;
  }
 } while(0);
 if ((label|0) == 22) {
  $$pre114 = HEAP32[$next_idx>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $72 = $$pre114; //@line 119 "modules/math/MathModule.cpp"
  $$pre$phiZ2D = $next_idx;$100 = $373;$103 = $72;$105 = $374;$142 = $375;$143 = $370;$90 = $$pre114;$95 = $371;$96 = $372;$idx_lm$0$lcssa = 0;
  label = 29;
 }
 if ((label|0) == 29) {
  $88 = HEAP32[$28>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $89 = (($88) - ($90))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $91 = $89 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $92 = (($91) + -1)|0; //@line 127 "modules/math/MathModule.cpp"
  $$cast = $90; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $93 = (($$cast) + ($92<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$93>>2] = 0; //@line 127 "modules/math/MathModule.cpp"
  $94 = (($95) - ($96))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $97 = $94 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $98 = (($97) + -1)|0; //@line 128 "modules/math/MathModule.cpp"
  $$cast56 = $96; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$$cast56>>2] = $98; //@line 128 "modules/math/MathModule.cpp"
  $99 = (($100) + ($idx_lm$0$lcssa<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $101 = HEAP32[$99>>2]|0; //@line 131 "modules/math/MathModule.cpp"
  $102 = (($103) + ($idx_lm$0$lcssa<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $104 = HEAP32[$102>>2]|0; //@line 131 "modules/math/MathModule.cpp"
  $$idx10 = (($105) + (($101*20)|0)|0);
  $$idx10$val = +HEAPF32[$$idx10>>2];
  $$idx11 = (((($105) + (($101*20)|0)|0)) + 4|0);
  $$idx11$val = +HEAPF32[$$idx11>>2];
  $$idx12 = (($105) + (($idx_lm$0$lcssa*20)|0)|0);
  $$idx12$val = +HEAPF32[$$idx12>>2];
  $$idx13 = (((($105) + (($idx_lm$0$lcssa*20)|0)|0)) + 4|0);
  $$idx13$val = +HEAPF32[$$idx13>>2];
  $$idx14 = (($105) + (($104*20)|0)|0);
  $$idx14$val = +HEAPF32[$$idx14>>2];
  $$idx15 = (((($105) + (($104*20)|0)|0)) + 4|0);
  $$idx15$val = +HEAPF32[$$idx15>>2];
  $106 = $$idx12$val - $$idx10$val; //@line 41 "modules/math/MathModule.cpp"
  $107 = $$idx15$val - $$idx11$val; //@line 41 "modules/math/MathModule.cpp"
  $108 = $106 * $107; //@line 41 "modules/math/MathModule.cpp"
  $109 = $$idx13$val - $$idx11$val; //@line 41 "modules/math/MathModule.cpp"
  $110 = $$idx14$val - $$idx10$val; //@line 41 "modules/math/MathModule.cpp"
  $111 = $109 * $110; //@line 41 "modules/math/MathModule.cpp"
  $112 = $108 - $111; //@line 41 "modules/math/MathModule.cpp"
  $113 = !($112 >= 0.0); //@line 41 "modules/math/MathModule.cpp"
  if ($113) {
   $130 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $131 = HEAP32[$prev_idx>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$$pre$phiZ2D>>2] = $131; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$prev_idx>>2] = $130; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $132 = HEAP32[$28>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $133 = HEAP32[$56>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$28>>2] = $133; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$56>>2] = $132; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $134 = ((($next_idx)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $135 = ((($prev_idx)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $136 = HEAP32[$134>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$134>>2] = $95; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$135>>2] = $136; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $137 = $130; //@line 132 "modules/math/MathModule.cpp"
   $138 = $131; //@line 132 "modules/math/MathModule.cpp"
   $215 = $137;$220 = $138;
  } else {
   $215 = $100;$220 = $103;
  }
  HEAP32[$concave_vertices>>2] = $concave_vertices; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $139 = ((($concave_vertices)) + 4|0); //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$139>>2] = $concave_vertices; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $140 = ((($concave_vertices)) + 8|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$140>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $141 = ($142|0)==($143|0); //@line 136 "modules/math/MathModule.cpp"
  L59: do {
   if ($141) {
    $$pre136 = (($142) - ($143))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre138 = (($$pre136|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre$phi139Z2D = $$pre138;
    label = 40;
   } else {
    $144 = (($142) - ($143))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $145 = (($144|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $146 = $143; //@line 136 "modules/math/MathModule.cpp"
    $218 = $105;$236 = $concave_vertices;$239 = 0;$i1$076 = 0;
    while(1) {
     $214 = (($215) + ($i1$076<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $216 = HEAP32[$214>>2]|0; //@line 138 "modules/math/MathModule.cpp"
     $217 = (($218) + (($i1$076*20)|0)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $219 = (($220) + ($i1$076<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $221 = HEAP32[$219>>2]|0; //@line 138 "modules/math/MathModule.cpp"
     $$idx = (($218) + (($216*20)|0)|0);
     $$idx$val = +HEAPF32[$$idx>>2];
     $$idx5 = (((($218) + (($216*20)|0)|0)) + 4|0);
     $$idx5$val = +HEAPF32[$$idx5>>2];
     $$idx6$val = +HEAPF32[$217>>2];
     $$idx7 = (((($218) + (($i1$076*20)|0)|0)) + 4|0);
     $$idx7$val = +HEAPF32[$$idx7>>2];
     $$idx8 = (($218) + (($221*20)|0)|0);
     $$idx8$val = +HEAPF32[$$idx8>>2];
     $$idx9 = (((($218) + (($221*20)|0)|0)) + 4|0);
     $$idx9$val = +HEAPF32[$$idx9>>2];
     $222 = $$idx6$val - $$idx$val; //@line 41 "modules/math/MathModule.cpp"
     $223 = $$idx9$val - $$idx5$val; //@line 41 "modules/math/MathModule.cpp"
     $224 = $222 * $223; //@line 41 "modules/math/MathModule.cpp"
     $225 = $$idx7$val - $$idx5$val; //@line 41 "modules/math/MathModule.cpp"
     $226 = $$idx8$val - $$idx$val; //@line 41 "modules/math/MathModule.cpp"
     $227 = $225 * $226; //@line 41 "modules/math/MathModule.cpp"
     $228 = $224 - $227; //@line 41 "modules/math/MathModule.cpp"
     $229 = !($228 >= 0.0); //@line 41 "modules/math/MathModule.cpp"
     if ($229) {
      __THREW__ = 0;
      $230 = (invoke_ii(3348,12)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $231 = __THREW__; __THREW__ = 0;
      $232 = $231&1;
      if ($232) {
       break;
      }
      $233 = $217; //@line 139 "modules/math/MathModule.cpp"
      $234 = ((($230)) + 8|0); //@line 1565 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      HEAP32[$234>>2] = $233; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $235 = ((($230)) + 4|0); //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      HEAP32[$235>>2] = $concave_vertices; //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      HEAP32[$230>>2] = $236; //@line 1098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $237 = ((($236)) + 4|0); //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      HEAP32[$237>>2] = $230; //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      HEAP32[$concave_vertices>>2] = $230; //@line 1100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $238 = (($239) + 1)|0; //@line 1567 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      HEAP32[$140>>2] = $238; //@line 1567 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $376 = $238;$377 = $230;
     } else {
      $376 = $239;$377 = $236;
     }
     $242 = (($i1$076) + 1)|0; //@line 136 "modules/math/MathModule.cpp"
     $243 = ($242>>>0)<($145>>>0); //@line 136 "modules/math/MathModule.cpp"
     if ($243) {
      $218 = $146;$236 = $377;$239 = $376;$i1$076 = $242;
     } else {
      $$pre$phi139Z2D = $145;
      label = 40;
      break L59;
     }
    }
    $240 = ___cxa_find_matching_catch_2()|0;
    $241 = tempRet0;
    $$1 = $240;$$12 = $241;
   }
  } while(0);
  if ((label|0) == 40) {
   HEAP32[$triangles>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $147 = ((($triangles)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$147>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $148 = ((($triangles)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$148>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $149 = ((($1)) + 20|0); //@line 74 "./common/math.h"
   $150 = ((($1)) + 40|0); //@line 74 "./common/math.h"
   $151 = ((($triangles)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $152 = ($$pre$phi139Z2D>>>0)>(3); //@line 146 "modules/math/MathModule.cpp"
   $153 = HEAP32[$next_idx>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $154 = HEAP32[$prev_idx>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $155 = HEAP32[$polygon>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $it$sroa$0$08$i$i193 = HEAP32[$139>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   L71: do {
    if ($152) {
     $$pn$in = $it$sroa$0$08$i$i193;$157 = $155;$159 = $153;$162 = $154;$current$0$ph196 = 1;$n_vertices$0$ph194 = $$pre$phi139Z2D;
     L72: while(1) {
      $$pn = $$pn$in; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $156 = ($$pn|0)==($concave_vertices|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $b$idx$i$us$phi$trans$insert = (($157) + (($current$0$ph196*20)|0)|0); //@line 151 "modules/math/MathModule.cpp"
      $b$idx$val$i$us$pre = +HEAPF32[$b$idx$i$us$phi$trans$insert>>2]; //@line 76 "modules/math/MathModule.cpp"
      $b$idx2$i$us$phi$trans$insert = (((($157) + (($current$0$ph196*20)|0)|0)) + 4|0); //@line 151 "modules/math/MathModule.cpp"
      $b$idx2$val$i$us$pre = +HEAPF32[$b$idx2$i$us$phi$trans$insert>>2]; //@line 76 "modules/math/MathModule.cpp"
      $b$idx$val$i$us = $b$idx$val$i$us$pre;$b$idx2$val$i$us = $b$idx2$val$i$us$pre;$current$0$us = $current$0$ph196;$skipped$0$us = 0;
      L74: while(1) {
       $158 = (($159) + ($current$0$us<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $160 = HEAP32[$158>>2]|0; //@line 148 "modules/math/MathModule.cpp"
       $161 = (($162) + ($current$0$us<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $163 = HEAP32[$161>>2]|0; //@line 149 "modules/math/MathModule.cpp"
       $164 = (($157) + (($163*20)|0)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $165 = (($157) + (($current$0$us*20)|0)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $166 = (($157) + (($160*20)|0)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $a$idx$val$i$us = +HEAPF32[$164>>2]; //@line 76 "modules/math/MathModule.cpp"
       $a$idx1$i$us = (((($157) + (($163*20)|0)|0)) + 4|0); //@line 151 "modules/math/MathModule.cpp"
       $a$idx1$val$i$us = +HEAPF32[$a$idx1$i$us>>2]; //@line 76 "modules/math/MathModule.cpp"
       $c$idx$val$i$us = +HEAPF32[$166>>2]; //@line 76 "modules/math/MathModule.cpp"
       $c$idx3$i$us = (((($157) + (($160*20)|0)|0)) + 4|0); //@line 151 "modules/math/MathModule.cpp"
       $c$idx3$val$i$us = +HEAPF32[$c$idx3$i$us>>2]; //@line 76 "modules/math/MathModule.cpp"
       $167 = $b$idx$val$i$us - $a$idx$val$i$us; //@line 41 "modules/math/MathModule.cpp"
       $168 = $c$idx3$val$i$us - $a$idx1$val$i$us; //@line 41 "modules/math/MathModule.cpp"
       $169 = $167 * $168; //@line 41 "modules/math/MathModule.cpp"
       $170 = $b$idx2$val$i$us - $a$idx1$val$i$us; //@line 41 "modules/math/MathModule.cpp"
       $171 = $c$idx$val$i$us - $a$idx$val$i$us; //@line 41 "modules/math/MathModule.cpp"
       $172 = $170 * $171; //@line 41 "modules/math/MathModule.cpp"
       $173 = $169 - $172; //@line 41 "modules/math/MathModule.cpp"
       $174 = !($173 >= 0.0); //@line 41 "modules/math/MathModule.cpp"
       L76: do {
        if (!($174)) {
         if ($156) {
          $256 = $164;$257 = $165;$258 = $166;$267 = $163;$268 = $160;
          break L74;
         }
         $175 = $c$idx$val$i$us - $b$idx$val$i$us; //@line 47 "modules/math/MathModule.cpp"
         $176 = $c$idx3$val$i$us - $b$idx2$val$i$us; //@line 47 "modules/math/MathModule.cpp"
         $177 = $a$idx1$val$i$us - $b$idx2$val$i$us; //@line 50 "modules/math/MathModule.cpp"
         $178 = $177 * $175; //@line 50 "modules/math/MathModule.cpp"
         $179 = $a$idx$val$i$us - $b$idx$val$i$us; //@line 50 "modules/math/MathModule.cpp"
         $180 = $179 * $176; //@line 50 "modules/math/MathModule.cpp"
         $181 = $178 - $180; //@line 50 "modules/math/MathModule.cpp"
         $182 = $172 - $169; //@line 50 "modules/math/MathModule.cpp"
         $184 = $$pn;$it$sroa$0$09$i$i$us = $$pn$in;
         while(1) {
          $183 = ((($184)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
          $185 = HEAP32[$183>>2]|0; //@line 66 "modules/math/MathModule.cpp"
          $186 = ($185|0)==($164|0); //@line 67 "modules/math/MathModule.cpp"
          $187 = ($185|0)==($165|0); //@line 67 "modules/math/MathModule.cpp"
          $or$cond$i$i$us = $186 | $187; //@line 67 "modules/math/MathModule.cpp"
          $188 = ($185|0)==($166|0); //@line 67 "modules/math/MathModule.cpp"
          $or$cond2$i$i$us = $188 | $or$cond$i$i$us; //@line 67 "modules/math/MathModule.cpp"
          if (!($or$cond2$i$i$us)) {
           $$idx$val$i$i$us = +HEAPF32[$185>>2]; //@line 76 "modules/math/MathModule.cpp"
           $$idx3$i$i$us = ((($185)) + 4|0); //@line 76 "modules/math/MathModule.cpp"
           $$idx3$val$i$i$us = +HEAPF32[$$idx3$i$i$us>>2]; //@line 76 "modules/math/MathModule.cpp"
           $189 = $$idx3$val$i$i$us - $b$idx2$val$i$us; //@line 49 "modules/math/MathModule.cpp"
           $190 = $175 * $189; //@line 49 "modules/math/MathModule.cpp"
           $191 = $$idx$val$i$i$us - $b$idx$val$i$us; //@line 49 "modules/math/MathModule.cpp"
           $192 = $176 * $191; //@line 49 "modules/math/MathModule.cpp"
           $193 = $190 - $192; //@line 49 "modules/math/MathModule.cpp"
           $194 = $181 * $193; //@line 51 "modules/math/MathModule.cpp"
           $195 = !($194 >= 0.0); //@line 51 "modules/math/MathModule.cpp"
           if (!($195)) {
            $196 = $$idx3$val$i$i$us - $a$idx1$val$i$us; //@line 49 "modules/math/MathModule.cpp"
            $197 = $171 * $196; //@line 49 "modules/math/MathModule.cpp"
            $198 = $$idx$val$i$i$us - $a$idx$val$i$us; //@line 49 "modules/math/MathModule.cpp"
            $199 = $168 * $198; //@line 49 "modules/math/MathModule.cpp"
            $200 = $197 - $199; //@line 49 "modules/math/MathModule.cpp"
            $201 = $182 * $200; //@line 51 "modules/math/MathModule.cpp"
            $202 = !($201 >= 0.0); //@line 51 "modules/math/MathModule.cpp"
            if (!($202)) {
             $203 = $167 * $196; //@line 49 "modules/math/MathModule.cpp"
             $204 = $170 * $198; //@line 49 "modules/math/MathModule.cpp"
             $205 = $203 - $204; //@line 49 "modules/math/MathModule.cpp"
             $206 = $173 * $205; //@line 51 "modules/math/MathModule.cpp"
             $207 = !($206 >= 0.0); //@line 51 "modules/math/MathModule.cpp"
             if (!($207)) {
              break L76;
             }
            }
           }
          }
          $208 = $it$sroa$0$09$i$i$us; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
          $209 = ((($208)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
          $it$sroa$0$0$i$i$us = HEAP32[$209>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
          $210 = $it$sroa$0$0$i$i$us; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
          $211 = ($210|0)==($concave_vertices|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
          if ($211) {
           $256 = $164;$257 = $165;$258 = $166;$267 = $163;$268 = $160;
           break L74;
          } else {
           $184 = $210;$it$sroa$0$09$i$i$us = $it$sroa$0$0$i$i$us;
          }
         }
        }
       } while(0);
       $212 = (($skipped$0$us) + 1)|0; //@line 160 "modules/math/MathModule.cpp"
       $213 = ($212>>>0)>($n_vertices$0$ph194>>>0); //@line 160 "modules/math/MathModule.cpp"
       if ($213) {
        break L72;
       } else {
        $b$idx$val$i$us = $c$idx$val$i$us;$b$idx2$val$i$us = $c$idx3$val$i$us;$current$0$us = $160;$skipped$0$us = $212;
       }
      }
      ;HEAP32[$1>>2]=HEAP32[$256>>2]|0;HEAP32[$1+4>>2]=HEAP32[$256+4>>2]|0;HEAP32[$1+8>>2]=HEAP32[$256+8>>2]|0;HEAP32[$1+12>>2]=HEAP32[$256+12>>2]|0;HEAP32[$1+16>>2]=HEAP32[$256+16>>2]|0; //@line 74 "./common/math.h"
      ;HEAP32[$149>>2]=HEAP32[$257>>2]|0;HEAP32[$149+4>>2]=HEAP32[$257+4>>2]|0;HEAP32[$149+8>>2]=HEAP32[$257+8>>2]|0;HEAP32[$149+12>>2]=HEAP32[$257+12>>2]|0;HEAP32[$149+16>>2]=HEAP32[$257+16>>2]|0; //@line 74 "./common/math.h"
      ;HEAP32[$150>>2]=HEAP32[$258>>2]|0;HEAP32[$150+4>>2]=HEAP32[$258+4>>2]|0;HEAP32[$150+8>>2]=HEAP32[$258+8>>2]|0;HEAP32[$150+12>>2]=HEAP32[$258+12>>2]|0;HEAP32[$150+16>>2]=HEAP32[$258+16>>2]|0; //@line 74 "./common/math.h"
      $259 = HEAP32[$147>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $260 = HEAP32[$151>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $261 = ($259>>>0)<($260>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($261) {
       dest=$259; src=$1; stop=dest+60|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $262 = ((($259)) + 60|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$147>>2] = $262; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $266 = $159;$270 = $162;
      } else {
       __THREW__ = 0;
       invoke_vii(3390,($triangles|0),($1|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $263 = __THREW__; __THREW__ = 0;
       $264 = $263&1;
       if ($264) {
        label = 62;
        break;
       }
       $$pre126 = HEAP32[$next_idx>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $$pre127 = HEAP32[$prev_idx>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $266 = $$pre126;$270 = $$pre127;
      }
      $265 = (($266) + ($267<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$265>>2] = $268; //@line 154 "modules/math/MathModule.cpp"
      $269 = (($270) + ($268<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$269>>2] = $267; //@line 155 "modules/math/MathModule.cpp"
      HEAP32[$2>>2] = $257; //@line 156 "modules/math/MathModule.cpp"
      __ZNSt3__14listIPKN4love6VertexENS_9allocatorIS4_EEE6removeERKS4_($concave_vertices,$2); //@line 156 "modules/math/MathModule.cpp"
      $271 = (($n_vertices$0$ph194) + -1)|0; //@line 157 "modules/math/MathModule.cpp"
      $272 = ($271>>>0)>(3); //@line 146 "modules/math/MathModule.cpp"
      $273 = HEAP32[$next_idx>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $274 = HEAP32[$prev_idx>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $275 = HEAP32[$polygon>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $it$sroa$0$08$i$i = HEAP32[$139>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if ($272) {
       $$pn$in = $it$sroa$0$08$i$i;$157 = $275;$159 = $273;$162 = $274;$current$0$ph196 = $268;$n_vertices$0$ph194 = $271;
      } else {
       $$lcssa184 = $273;$$lcssa187 = $274;$$lcssa190 = $275;$current$0$ph$lcssa = $268;
       label = 56;
       break L71;
      }
     }
     if ((label|0) == 62) {
      $276 = ___cxa_find_matching_catch_2()|0;
      $277 = tempRet0;
      $lpad$phi$index = $276;$lpad$phi$index4 = $277;
      label = 64;
      break;
     }
     $280 = (___cxa_allocate_exception(16)|0); //@line 162 "modules/math/MathModule.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($280|0),(355635|0),($vararg_buffer1|0)); //@line 162 "modules/math/MathModule.cpp"
     $281 = __THREW__; __THREW__ = 0;
     $282 = $281&1;
     if ($282) {
      $284 = ___cxa_find_matching_catch_2()|0;
      $285 = tempRet0;
      ___cxa_free_exception(($280|0)); //@line 162 "modules/math/MathModule.cpp"
      $$0 = $284;$$01 = $285;
      break;
     } else {
      __THREW__ = 0;
      invoke_viii(3350,($280|0),(96|0),(344|0)); //@line 162 "modules/math/MathModule.cpp"
      $283 = __THREW__; __THREW__ = 0;
      label = 63;
      break;
     }
    } else {
     $$lcssa184 = $153;$$lcssa187 = $154;$$lcssa190 = $155;$current$0$ph$lcssa = 1;
     label = 56;
    }
   } while(0);
   do {
    if ((label|0) == 56) {
     $244 = (($$lcssa184) + ($current$0$ph$lcssa<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $245 = HEAP32[$244>>2]|0; //@line 148 "modules/math/MathModule.cpp"
     $246 = (($$lcssa187) + ($current$0$ph$lcssa<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $247 = HEAP32[$246>>2]|0; //@line 149 "modules/math/MathModule.cpp"
     $248 = (($$lcssa190) + (($247*20)|0)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $249 = (($$lcssa190) + (($current$0$ph$lcssa*20)|0)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $250 = (($$lcssa190) + (($245*20)|0)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     ;HEAP32[$3>>2]=HEAP32[$248>>2]|0;HEAP32[$3+4>>2]=HEAP32[$248+4>>2]|0;HEAP32[$3+8>>2]=HEAP32[$248+8>>2]|0;HEAP32[$3+12>>2]=HEAP32[$248+12>>2]|0;HEAP32[$3+16>>2]=HEAP32[$248+16>>2]|0; //@line 74 "./common/math.h"
     $251 = ((($3)) + 20|0); //@line 74 "./common/math.h"
     ;HEAP32[$251>>2]=HEAP32[$249>>2]|0;HEAP32[$251+4>>2]=HEAP32[$249+4>>2]|0;HEAP32[$251+8>>2]=HEAP32[$249+8>>2]|0;HEAP32[$251+12>>2]=HEAP32[$249+12>>2]|0;HEAP32[$251+16>>2]=HEAP32[$249+16>>2]|0; //@line 74 "./common/math.h"
     $252 = ((($3)) + 40|0); //@line 74 "./common/math.h"
     ;HEAP32[$252>>2]=HEAP32[$250>>2]|0;HEAP32[$252+4>>2]=HEAP32[$250+4>>2]|0;HEAP32[$252+8>>2]=HEAP32[$250+8>>2]|0;HEAP32[$252+12>>2]=HEAP32[$250+12>>2]|0;HEAP32[$252+16>>2]=HEAP32[$250+16>>2]|0; //@line 74 "./common/math.h"
     $253 = HEAP32[$147>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $254 = HEAP32[$151>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $255 = ($253>>>0)<($254>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($255) {
      $286 = $254; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      dest=$253; src=$3; stop=dest+60|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $287 = ((($253)) + 60|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$147>>2] = $287; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $288 = $287; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $293 = $288;$295 = $286;
     } else {
      __THREW__ = 0;
      invoke_vii(3390,($triangles|0),($3|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $289 = __THREW__; __THREW__ = 0;
      $290 = $289&1;
      if ($290) {
       label = 63;
       break;
      }
      $$pre119 = HEAP32[$147>>2]|0; //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$pre121 = HEAP32[$151>>2]|0; //@line 1248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $293 = $$pre119;$295 = $$pre121;
     }
     $291 = ((($agg$result)) + 4|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $292 = HEAP32[$triangles>>2]|0; //@line 1246 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$agg$result>>2] = $292; //@line 1246 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$291>>2] = $293; //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $294 = ((($agg$result)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$294>>2] = $295; //@line 1248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $296 = HEAP32[$140>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $297 = ($296|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     if (!($297)) {
      $298 = HEAP32[$139>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $299 = HEAP32[$concave_vertices>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $300 = ((($299)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $301 = HEAP32[$300>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $302 = HEAP32[$298>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $303 = ((($302)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      HEAP32[$303>>2] = $301; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $304 = $302; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $305 = HEAP32[$300>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      HEAP32[$305>>2] = $304; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      HEAP32[$140>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $306 = ($298|0)==($concave_vertices|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if (!($306)) {
       $__f$01$i$i44 = $298;
       while(1) {
        $307 = ((($__f$01$i$i44)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
        $308 = HEAP32[$307>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
        __ZdlPv($__f$01$i$i44); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        $309 = ($308|0)==($concave_vertices|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
        if ($309) {
         break;
        } else {
         $__f$01$i$i44 = $308;
        }
       }
      }
     }
     $310 = HEAP32[$prev_idx>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $311 = ($310|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $312 = $310; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($311)) {
      $313 = HEAP32[$56>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $314 = ($313|0)==($310|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($314)) {
       $scevgep$i$i$i29 = ((($313)) + -4|0);
       $315 = $scevgep$i$i$i29;
       $316 = (($315) - ($312))|0;
       $317 = $316 >>> 2;
       $318 = $317 ^ -1;
       $scevgep4$i$i$i30 = (($313) + ($318<<2)|0);
       HEAP32[$56>>2] = $scevgep4$i$i$i30; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($310); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $319 = HEAP32[$next_idx>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $320 = ($319|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $321 = $319; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($320)) {
      $322 = HEAP32[$28>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $323 = ($322|0)==($319|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($323)) {
       $scevgep$i$i$i24 = ((($322)) + -4|0);
       $324 = $scevgep$i$i$i24;
       $325 = (($324) - ($321))|0;
       $326 = $325 >>> 2;
       $327 = $326 ^ -1;
       $scevgep4$i$i$i25 = (($322) + ($327<<2)|0);
       HEAP32[$28>>2] = $scevgep4$i$i$i25; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($319); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     STACKTOP = sp;return; //@line 171 "modules/math/MathModule.cpp"
    }
   } while(0);
   if ((label|0) == 63) {
    $278 = ___cxa_find_matching_catch_2()|0;
    $279 = tempRet0;
    $lpad$phi$index = $278;$lpad$phi$index4 = $279;
    label = 64;
   }
   if ((label|0) == 64) {
    $$0 = $lpad$phi$index;$$01 = $lpad$phi$index4;
   }
   $328 = HEAP32[$triangles>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $329 = ($328|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $330 = $328; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($329) {
    $$1 = $$0;$$12 = $$01;
   } else {
    $331 = HEAP32[$147>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $332 = ($331|0)==($328|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($332)) {
     $scevgep$i$i$i21 = ((($331)) + -60|0);
     $333 = $scevgep$i$i$i21;
     $334 = (($333) - ($330))|0;
     $335 = (($334>>>0) / 60)&-1;
     $336 = $335 ^ -1;
     $scevgep4$i$i$i22 = (($331) + (($336*60)|0)|0);
     HEAP32[$147>>2] = $scevgep4$i$i$i22; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($328); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $$1 = $$0;$$12 = $$01;
   }
  }
  $337 = HEAP32[$140>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $338 = ($337|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if (!($338)) {
   $339 = HEAP32[$139>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $340 = HEAP32[$concave_vertices>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $341 = ((($340)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $342 = HEAP32[$341>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $343 = HEAP32[$339>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $344 = ((($343)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   HEAP32[$344>>2] = $342; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $345 = $343; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $346 = HEAP32[$341>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   HEAP32[$346>>2] = $345; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   HEAP32[$140>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $347 = ($339|0)==($concave_vertices|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if (!($347)) {
    $__f$01$i$i = $339;
    while(1) {
     $348 = ((($__f$01$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $349 = HEAP32[$348>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     __ZdlPv($__f$01$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $350 = ($349|0)==($concave_vertices|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     if ($350) {
      break;
     } else {
      $__f$01$i$i = $349;
     }
    }
   }
  }
  $351 = HEAP32[$prev_idx>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $352 = ($351|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $353 = $351; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($352) {
   $$2 = $$1;$$23 = $$12;
  } else {
   $354 = HEAP32[$56>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $355 = ($354|0)==($351|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($355)) {
    $scevgep$i$i$i16 = ((($354)) + -4|0);
    $356 = $scevgep$i$i$i16;
    $357 = (($356) - ($353))|0;
    $358 = $357 >>> 2;
    $359 = $358 ^ -1;
    $scevgep4$i$i$i17 = (($354) + ($359<<2)|0);
    HEAP32[$56>>2] = $scevgep4$i$i$i17; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($351); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $$2 = $$1;$$23 = $$12;
  }
 }
 $360 = HEAP32[$next_idx>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $361 = ($360|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $362 = $360; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($361) {
  $$3 = $$2;$$34 = $$23;
  ___resumeException($$3|0);
  // unreachable;
 }
 $363 = HEAP32[$28>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $364 = ($363|0)==($360|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($364)) {
  $scevgep$i$i$i = ((($363)) + -4|0);
  $365 = $scevgep$i$i$i;
  $366 = (($365) - ($362))|0;
  $367 = $366 >>> 2;
  $368 = $367 ^ -1;
  $scevgep4$i$i$i = (($363) + ($368<<2)|0);
  HEAP32[$28>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($360); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $$3 = $$2;$$34 = $$23;
 ___resumeException($$3|0);
 // unreachable;
}
function __ZNSt3__14listIPKN4love6VertexENS_9allocatorIS4_EEE6removeERKS4_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$cast$i$i$i = 0, $$lcssa = 0, $$lcssa13 = 0, $$lcssa13$us = 0, $$lcssa14 = 0, $$lcssa14$us = 0, $$lcssa37 = 0, $$lcssa38 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__deleted_nodes = 0;
 var $__f$01$i$i = 0, $__i$sroa$0$0$be = 0, $__i$sroa$0$0$be$us = 0, $__i$sroa$0$023 = 0, $__i$sroa$0$023$us = 0, $__j$sroa$0$0$lcssa = 0, $__j$sroa$0$0$lcssa$us = 0, $__j$sroa$0$016 = 0, $__j$sroa$0$016$us = 0, $__r$01$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__deleted_nodes = sp;
 HEAP32[$__deleted_nodes>>2] = $__deleted_nodes; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $0 = ((($__deleted_nodes)) + 4|0); //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$0>>2] = $__deleted_nodes; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $1 = ((($__deleted_nodes)) + 8|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$1>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($this)) + 4|0); //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $3 = HEAP32[$2>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $4 = $3; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $5 = ($4|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($5) {
  STACKTOP = sp;return; //@line 2073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 }
 $6 = ($__deleted_nodes|0)==($this|0); //@line 2017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $7 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($6) {
  $9 = $4;$__i$sroa$0$023$us = $3;
  while(1) {
   $8 = ((($9)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $10 = HEAP32[$8>>2]|0; //@line 2060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $11 = HEAP32[$__x>>2]|0; //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $12 = ($10|0)==($11|0); //@line 2060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $13 = $__i$sroa$0$023$us; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $14 = ((($13)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $15 = HEAP32[$14>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($12) {
    $16 = $15; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $17 = ($16|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    L9: do {
     if ($17) {
      $$lcssa13$us = $this;$$lcssa14$us = 0;$__j$sroa$0$0$lcssa$us = $15;
     } else {
      $19 = $16;$__j$sroa$0$016$us = $15;
      while(1) {
       $18 = ((($19)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $20 = HEAP32[$18>>2]|0; //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $21 = ($20|0)==($10|0); //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       if (!($21)) {
        $$lcssa13$us = $19;$$lcssa14$us = 1;$__j$sroa$0$0$lcssa$us = $__j$sroa$0$016$us;
        break L9;
       }
       $39 = $__j$sroa$0$016$us; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $40 = ((($39)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $41 = HEAP32[$40>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $42 = $41; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $43 = ($42|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       if ($43) {
        $$lcssa13$us = $this;$$lcssa14$us = 0;$__j$sroa$0$0$lcssa$us = $41;
        break;
       } else {
        $19 = $42;$__j$sroa$0$016$us = $41;
       }
      }
     }
    } while(0);
    $22 = ($9|0)==($$lcssa13$us|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if (!($22)) {
     $23 = $__j$sroa$0$0$lcssa$us; //@line 487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $24 = HEAP32[$23>>2]|0; //@line 487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $25 = ((($24)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $26 = HEAP32[$25>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $27 = $__i$sroa$0$023$us; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $28 = HEAP32[$27>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $29 = ((($28)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$29>>2] = $26; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $30 = $28; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $31 = HEAP32[$25>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$31>>2] = $30; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $32 = HEAP32[$__deleted_nodes>>2]|0; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $33 = ((($32)) + 4|0); //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$33>>2] = $9; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $34 = $32; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $35 = $__i$sroa$0$023$us; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$35>>2] = $34; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$__deleted_nodes>>2] = $24; //@line 1075 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$25>>2] = $this; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    }
    if ($$lcssa14$us) {
     $36 = $__j$sroa$0$0$lcssa$us; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $37 = ((($36)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $38 = HEAP32[$37>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $__i$sroa$0$0$be$us = $38;
    } else {
     $__i$sroa$0$0$be$us = $__j$sroa$0$0$lcssa$us;
    }
   } else {
    $__i$sroa$0$0$be$us = $15;
   }
   $44 = $__i$sroa$0$0$be$us; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $45 = ($44|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($45) {
    break;
   } else {
    $9 = $44;$__i$sroa$0$023$us = $__i$sroa$0$0$be$us;
   }
  }
  STACKTOP = sp;return; //@line 2073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 }
 $106 = $__deleted_nodes;$59 = $4;$87 = 0;$__i$sroa$0$023 = $3;
 while(1) {
  $58 = ((($59)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $60 = HEAP32[$58>>2]|0; //@line 2060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $61 = HEAP32[$__x>>2]|0; //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $62 = ($60|0)==($61|0); //@line 2060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $63 = $__i$sroa$0$023; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $64 = ((($63)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $65 = HEAP32[$64>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($62) {
   $66 = $65; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $67 = ($66|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   L25: do {
    if ($67) {
     $$lcssa13 = $this;$$lcssa14 = 0;$__j$sroa$0$0$lcssa = $65;
    } else {
     $69 = $66;$__j$sroa$0$016 = $65;
     while(1) {
      $68 = ((($69)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $70 = HEAP32[$68>>2]|0; //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $71 = ($70|0)==($60|0); //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if (!($71)) {
       $$lcssa13 = $69;$$lcssa14 = 1;$__j$sroa$0$0$lcssa = $__j$sroa$0$016;
       break L25;
      }
      $72 = $__j$sroa$0$016; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $73 = ((($72)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $74 = HEAP32[$73>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $75 = $74; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $76 = ($75|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if ($76) {
       $$lcssa13 = $this;$$lcssa14 = 0;$__j$sroa$0$0$lcssa = $74;
       break;
      } else {
       $69 = $75;$__j$sroa$0$016 = $74;
      }
     }
    }
   } while(0);
   $77 = ($59|0)==($$lcssa13|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($77) {
    $109 = $106;$110 = $87;
   } else {
    $80 = $59;$__r$01$i$i$i = 0;
    while(1) {
     $78 = (($__r$01$i$i$i) + 1)|0; //@line 485 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $79 = ((($80)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $81 = HEAP32[$79>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $$cast$i$i$i = $81; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $82 = ($$cast$i$i$i|0)==($$lcssa13|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $83 = $81; //@line 484 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($82) {
      $$lcssa = $78;
      break;
     } else {
      $80 = $83;$__r$01$i$i$i = $78;
     }
    }
    $84 = HEAP32[$7>>2]|0; //@line 2020 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $85 = (($84) - ($$lcssa))|0; //@line 2020 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$7>>2] = $85; //@line 2020 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $86 = (($87) + ($$lcssa))|0; //@line 2021 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$1>>2] = $86; //@line 2021 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $88 = $__j$sroa$0$0$lcssa; //@line 487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $89 = HEAP32[$88>>2]|0; //@line 487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $90 = ((($89)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $91 = HEAP32[$90>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $92 = $__i$sroa$0$023; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $93 = HEAP32[$92>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $94 = ((($93)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$94>>2] = $91; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $95 = $93; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $96 = HEAP32[$90>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$96>>2] = $95; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $97 = HEAP32[$__deleted_nodes>>2]|0; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $98 = ((($97)) + 4|0); //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$98>>2] = $59; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $99 = $97; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $100 = $__i$sroa$0$023; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$100>>2] = $99; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$__deleted_nodes>>2] = $89; //@line 1075 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$90>>2] = $__deleted_nodes; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $109 = $89;$110 = $86;
   }
   if ($$lcssa14) {
    $101 = $__j$sroa$0$0$lcssa; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $102 = ((($101)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $103 = HEAP32[$102>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $107 = $109;$108 = $110;$__i$sroa$0$0$be = $103;
   } else {
    $107 = $109;$108 = $110;$__i$sroa$0$0$be = $__j$sroa$0$0$lcssa;
   }
  } else {
   $107 = $106;$108 = $87;$__i$sroa$0$0$be = $65;
  }
  $104 = $__i$sroa$0$0$be; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $105 = ($104|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($105) {
   $$lcssa37 = $107;$$lcssa38 = $108;
   break;
  } else {
   $106 = $107;$59 = $104;$87 = $108;$__i$sroa$0$023 = $__i$sroa$0$0$be;
  }
 }
 $46 = ($$lcssa38|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($46) {
  STACKTOP = sp;return; //@line 2073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 }
 $47 = HEAP32[$0>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $48 = ((($$lcssa37)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $49 = HEAP32[$48>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $50 = HEAP32[$47>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $51 = ((($50)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$51>>2] = $49; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $52 = $50; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $53 = HEAP32[$48>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$53>>2] = $52; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$1>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $54 = ($47|0)==($__deleted_nodes|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($54) {
  STACKTOP = sp;return; //@line 2073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 } else {
  $__f$01$i$i = $47;
 }
 while(1) {
  $55 = ((($__f$01$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $56 = HEAP32[$55>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  __ZdlPv($__f$01$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $57 = ($56|0)==($__deleted_nodes|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($57) {
   break;
  } else {
   $__f$01$i$i = $56;
  }
 }
 STACKTOP = sp;return; //@line 2073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
}
function __ZN4love4math4Math8isConvexERKNSt3__16vectorINS_6VertexENS2_9allocatorIS4_EEEE($this,$polygon) {
 $this = $this|0;
 $polygon = $polygon|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0.0;
 var $26 = 0.0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0.0;
 var $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $j$0 = 0, $j$0$phi = 0, $k$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($polygon)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$polygon>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($3|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($4>>>0)<(3); //@line 175 "modules/math/MathModule.cpp"
 $6 = $2; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($5) {
  $$1 = 0;
  return ($$1|0); //@line 198 "modules/math/MathModule.cpp"
 }
 $7 = (($4) + -2)|0; //@line 181 "modules/math/MathModule.cpp"
 $8 = (($4) + -1)|0; //@line 181 "modules/math/MathModule.cpp"
 $9 = (($6) + (($8*20)|0)|0); //@line 182 "modules/math/MathModule.cpp"
 $10 = +HEAPF32[$9>>2]; //@line 182 "modules/math/MathModule.cpp"
 $11 = (($6) + (($7*20)|0)|0); //@line 182 "modules/math/MathModule.cpp"
 $12 = +HEAPF32[$11>>2]; //@line 182 "modules/math/MathModule.cpp"
 $13 = $10 - $12; //@line 182 "modules/math/MathModule.cpp"
 $14 = (((($6) + (($8*20)|0)|0)) + 4|0); //@line 182 "modules/math/MathModule.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 182 "modules/math/MathModule.cpp"
 $16 = (((($6) + (($7*20)|0)|0)) + 4|0); //@line 182 "modules/math/MathModule.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 182 "modules/math/MathModule.cpp"
 $18 = $15 - $17; //@line 182 "modules/math/MathModule.cpp"
 $19 = +HEAPF32[$6>>2]; //@line 183 "modules/math/MathModule.cpp"
 $20 = $19 - $10; //@line 183 "modules/math/MathModule.cpp"
 $21 = ((($6)) + 4|0); //@line 183 "modules/math/MathModule.cpp"
 $22 = +HEAPF32[$21>>2]; //@line 183 "modules/math/MathModule.cpp"
 $23 = $22 - $15; //@line 183 "modules/math/MathModule.cpp"
 $24 = $13 * $23; //@line 286 "./common/Vector.h"
 $25 = $18 * $20; //@line 286 "./common/Vector.h"
 $26 = $24 - $25; //@line 286 "./common/Vector.h"
 $32 = $19;$36 = $22;$j$0 = $8;$k$0 = 0;
 while(1) {
  $27 = (($k$0) + 1)|0; //@line 186 "modules/math/MathModule.cpp"
  $28 = ($27>>>0)<($4>>>0); //@line 186 "modules/math/MathModule.cpp"
  if (!($28)) {
   $$1 = 1;
   label = 5;
   break;
  }
  $29 = (($6) + (($j$0*20)|0)|0); //@line 189 "modules/math/MathModule.cpp"
  $30 = +HEAPF32[$29>>2]; //@line 189 "modules/math/MathModule.cpp"
  $31 = $32 - $30; //@line 189 "modules/math/MathModule.cpp"
  $33 = (((($6) + (($j$0*20)|0)|0)) + 4|0); //@line 190 "modules/math/MathModule.cpp"
  $34 = +HEAPF32[$33>>2]; //@line 190 "modules/math/MathModule.cpp"
  $35 = $36 - $34; //@line 190 "modules/math/MathModule.cpp"
  $37 = (($6) + (($27*20)|0)|0); //@line 191 "modules/math/MathModule.cpp"
  $38 = +HEAPF32[$37>>2]; //@line 191 "modules/math/MathModule.cpp"
  $39 = $38 - $32; //@line 191 "modules/math/MathModule.cpp"
  $40 = (((($6) + (($27*20)|0)|0)) + 4|0); //@line 192 "modules/math/MathModule.cpp"
  $41 = +HEAPF32[$40>>2]; //@line 192 "modules/math/MathModule.cpp"
  $42 = $41 - $36; //@line 192 "modules/math/MathModule.cpp"
  $43 = $31 * $42; //@line 286 "./common/Vector.h"
  $44 = $35 * $39; //@line 286 "./common/Vector.h"
  $45 = $43 - $44; //@line 286 "./common/Vector.h"
  $46 = $26 * $45; //@line 194 "modules/math/MathModule.cpp"
  $47 = $46 < 0.0; //@line 194 "modules/math/MathModule.cpp"
  if ($47) {
   $$1 = 0;
   label = 5;
   break;
  } else {
   $j$0$phi = $k$0;$32 = $38;$36 = $41;$k$0 = $27;$j$0 = $j$0$phi;
  }
 }
 if ((label|0) == 5) {
  return ($$1|0); //@line 198 "modules/math/MathModule.cpp"
 }
 return (0)|0;
}
function __ZNK4love4math4Math13gammaToLinearEf($this,$c) {
 $this = $this|0;
 $c = +$c;
 var $$0 = 0.0, $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = !($c <= 0.040449999272823334); //@line 205 "modules/math/MathModule.cpp"
 if ($0) {
  $2 = $c + 0.054999999701976776; //@line 208 "modules/math/MathModule.cpp"
  $3 = $2 / 1.0549999475479126; //@line 208 "modules/math/MathModule.cpp"
  $4 = (+Math_pow((+$3),2.4000000953674316)); //@line 208 "modules/math/MathModule.cpp"
  $$0 = $4;
  return (+$$0); //@line 209 "modules/math/MathModule.cpp"
 } else {
  $1 = $c / 12.920000076293945; //@line 206 "modules/math/MathModule.cpp"
  $$0 = $1;
  return (+$$0); //@line 209 "modules/math/MathModule.cpp"
 }
 return +(0.0);
}
function __ZNK4love4math4Math13linearToGammaEf($this,$c) {
 $this = $this|0;
 $c = +$c;
 var $$0 = 0.0, $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = !($c <= 0.0031308000907301903); //@line 216 "modules/math/MathModule.cpp"
 if ($0) {
  $2 = (+Math_pow((+$c),0.4166666567325592)); //@line 219 "modules/math/MathModule.cpp"
  $3 = $2 * 1.0549999475479126; //@line 219 "modules/math/MathModule.cpp"
  $4 = $3 + -0.054999999701976776; //@line 219 "modules/math/MathModule.cpp"
  $$0 = $4;
  return (+$$0); //@line 220 "modules/math/MathModule.cpp"
 } else {
  $1 = $c * 12.920000076293945; //@line 217 "modules/math/MathModule.cpp"
  $$0 = $1;
  return (+$$0); //@line 220 "modules/math/MathModule.cpp"
 }
 return +(0.0);
}
function __ZN4love4math4Math8compressENS0_10Compressor6FormatEPNS_4DataEi($this,$format,$rawdata,$level) {
 $this = $this|0;
 $format = $format|0;
 $rawdata = $rawdata|0;
 $level = $level|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$rawdata>>2]|0; //@line 224 "modules/math/MathModule.cpp"
 $1 = ((($0)) + 16|0); //@line 224 "modules/math/MathModule.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 224 "modules/math/MathModule.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($rawdata)|0); //@line 224 "modules/math/MathModule.cpp"
 $4 = HEAP32[$rawdata>>2]|0; //@line 224 "modules/math/MathModule.cpp"
 $5 = ((($4)) + 20|0); //@line 224 "modules/math/MathModule.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 224 "modules/math/MathModule.cpp"
 $7 = (FUNCTION_TABLE_ii[$6 & 8191]($rawdata)|0); //@line 224 "modules/math/MathModule.cpp"
 $8 = (__ZN4love4math4Math8compressENS0_10Compressor6FormatEPKcji(0,$format,$3,$7,$level)|0); //@line 224 "modules/math/MathModule.cpp"
 return ($8|0); //@line 224 "modules/math/MathModule.cpp"
}
function __ZN4love4math4Math8compressENS0_10Compressor6FormatEPKcji($this,$format,$rawbytes,$rawsize,$level) {
 $this = $this|0;
 $format = $format|0;
 $rawbytes = $rawbytes|0;
 $rawsize = $rawsize|0;
 $level = $level|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $compressedsize = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $compressedsize = sp + 4|0;
 $0 = (__ZN4love4math10Compressor13getCompressorENS1_6FormatE($format)|0); //@line 229 "modules/math/MathModule.cpp"
 $1 = ($0|0)==(0|0); //@line 231 "modules/math/MathModule.cpp"
 if ($1) {
  $2 = (___cxa_allocate_exception(16)|0); //@line 232 "modules/math/MathModule.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($2|0),(355663|0),($vararg_buffer|0)); //@line 232 "modules/math/MathModule.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if (!($4)) {
   ___cxa_throw(($2|0),(96|0),(344|0)); //@line 232 "modules/math/MathModule.cpp"
   // unreachable; //@line 232 "modules/math/MathModule.cpp"
  }
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  ___cxa_free_exception(($2|0)); //@line 232 "modules/math/MathModule.cpp"
  $$1 = $5;$$12 = $6;
  ___resumeException($$1|0);
  // unreachable;
 }
 HEAP32[$compressedsize>>2] = 0; //@line 234 "modules/math/MathModule.cpp"
 $7 = HEAP32[$0>>2]|0; //@line 235 "modules/math/MathModule.cpp"
 $8 = ((($7)) + 8|0); //@line 235 "modules/math/MathModule.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 235 "modules/math/MathModule.cpp"
 $10 = (FUNCTION_TABLE_iiiiiii[$9 & 8191]($0,$format,$rawbytes,$rawsize,$level,$compressedsize)|0); //@line 235 "modules/math/MathModule.cpp"
 __THREW__ = 0;
 $11 = (invoke_ii(3348,24)|0); //@line 241 "modules/math/MathModule.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 do {
  if ($13) {
   $17 = ___cxa_find_matching_catch_3(96|0)|0;
   $18 = tempRet0;
   $$0 = $17;$$01 = $18;
  } else {
   $14 = HEAP32[$compressedsize>>2]|0; //@line 241 "modules/math/MathModule.cpp"
   __THREW__ = 0;
   invoke_viiiiii(3655,($11|0),($format|0),($10|0),($14|0),($rawsize|0),1); //@line 241 "modules/math/MathModule.cpp"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if ($16) {
    $19 = ___cxa_find_matching_catch_3(96|0)|0;
    $20 = tempRet0;
    __ZdlPv($11); //@line 241 "modules/math/MathModule.cpp"
    $$0 = $19;$$01 = $20;
    break;
   } else {
    STACKTOP = sp;return ($11|0); //@line 249 "modules/math/MathModule.cpp"
   }
  }
 } while(0);
 $21 = (_llvm_eh_typeid_for((96|0))|0); //@line 242 "modules/math/MathModule.cpp"
 $22 = ($$01|0)==($21|0); //@line 242 "modules/math/MathModule.cpp"
 if (!($22)) {
  $$1 = $$0;$$12 = $$01;
  ___resumeException($$1|0);
  // unreachable;
 }
 (___cxa_begin_catch(($$0|0))|0); //@line 242 "modules/math/MathModule.cpp"
 $23 = ($10|0)==(0|0); //@line 245 "modules/math/MathModule.cpp"
 if (!($23)) {
  __ZdaPv($10); //@line 245 "modules/math/MathModule.cpp"
 }
 __THREW__ = 0;
 invoke_v(3656); //@line 246 "modules/math/MathModule.cpp"
 $24 = __THREW__; __THREW__ = 0;
 $25 = ___cxa_find_matching_catch_2()|0;
 $26 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 247 "modules/math/MathModule.cpp"
 $27 = __THREW__; __THREW__ = 0;
 $28 = $27&1;
 if ($28) {
  $29 = ___cxa_find_matching_catch_3(0|0)|0;
  $30 = tempRet0;
  ___clang_call_terminate($29); //@line 247 "modules/math/MathModule.cpp"
  // unreachable; //@line 247 "modules/math/MathModule.cpp"
 } else {
  $$1 = $25;$$12 = $26;
  ___resumeException($$1|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love4math4Math10decompressEPNS0_14CompressedDataERj($this,$data,$decompressedsize) {
 $this = $this|0;
 $data = $data|0;
 $decompressedsize = $decompressedsize|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $rawsize = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $rawsize = sp + 4|0;
 $0 = (__ZNK4love4math14CompressedData19getDecompressedSizeEv($data)|0); //@line 254 "modules/math/MathModule.cpp"
 HEAP32[$rawsize>>2] = $0; //@line 254 "modules/math/MathModule.cpp"
 $1 = (__ZNK4love4math14CompressedData9getFormatEv($data)|0); //@line 256 "modules/math/MathModule.cpp"
 $2 = HEAP32[$data>>2]|0; //@line 256 "modules/math/MathModule.cpp"
 $3 = ((($2)) + 16|0); //@line 256 "modules/math/MathModule.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 256 "modules/math/MathModule.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($data)|0); //@line 256 "modules/math/MathModule.cpp"
 $6 = HEAP32[$data>>2]|0; //@line 257 "modules/math/MathModule.cpp"
 $7 = ((($6)) + 20|0); //@line 257 "modules/math/MathModule.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 257 "modules/math/MathModule.cpp"
 $9 = (FUNCTION_TABLE_ii[$8 & 8191]($data)|0); //@line 257 "modules/math/MathModule.cpp"
 $10 = (__ZN4love4math10Compressor13getCompressorENS1_6FormatE($1)|0); //@line 265 "modules/math/MathModule.cpp"
 $11 = ($10|0)==(0|0); //@line 267 "modules/math/MathModule.cpp"
 if (!($11)) {
  $17 = HEAP32[$10>>2]|0; //@line 270 "modules/math/MathModule.cpp"
  $18 = ((($17)) + 12|0); //@line 270 "modules/math/MathModule.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 270 "modules/math/MathModule.cpp"
  $20 = (FUNCTION_TABLE_iiiiii[$19 & 8191]($10,$1,$5,$9,$rawsize)|0); //@line 270 "modules/math/MathModule.cpp"
  $21 = HEAP32[$rawsize>>2]|0; //@line 259 "modules/math/MathModule.cpp"
  HEAP32[$decompressedsize>>2] = $21; //@line 259 "modules/math/MathModule.cpp"
  STACKTOP = sp;return ($20|0); //@line 260 "modules/math/MathModule.cpp"
 }
 $12 = (___cxa_allocate_exception(16)|0); //@line 268 "modules/math/MathModule.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($12|0),(355663|0),($vararg_buffer|0)); //@line 268 "modules/math/MathModule.cpp"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if ($14) {
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  ___cxa_free_exception(($12|0)); //@line 268 "modules/math/MathModule.cpp"
  ___resumeException($15|0);
  // unreachable;
 } else {
  ___cxa_throw(($12|0),(96|0),(344|0)); //@line 268 "modules/math/MathModule.cpp"
  // unreachable; //@line 268 "modules/math/MathModule.cpp"
 }
 return (0)|0;
}
function __ZN4love4math4Math10decompressENS0_10Compressor6FormatEPKcjRj($this,$format,$cbytes,$compressedsize,$rawsize) {
 $this = $this|0;
 $format = $format|0;
 $cbytes = $cbytes|0;
 $compressedsize = $compressedsize|0;
 $rawsize = $rawsize|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (__ZN4love4math10Compressor13getCompressorENS1_6FormatE($format)|0); //@line 265 "modules/math/MathModule.cpp"
 $1 = ($0|0)==(0|0); //@line 267 "modules/math/MathModule.cpp"
 if (!($1)) {
  $7 = HEAP32[$0>>2]|0; //@line 270 "modules/math/MathModule.cpp"
  $8 = ((($7)) + 12|0); //@line 270 "modules/math/MathModule.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 270 "modules/math/MathModule.cpp"
  $10 = (FUNCTION_TABLE_iiiiii[$9 & 8191]($0,$format,$cbytes,$compressedsize,$rawsize)|0); //@line 270 "modules/math/MathModule.cpp"
  STACKTOP = sp;return ($10|0); //@line 270 "modules/math/MathModule.cpp"
 }
 $2 = (___cxa_allocate_exception(16)|0); //@line 268 "modules/math/MathModule.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($2|0),(355663|0),($vararg_buffer|0)); //@line 268 "modules/math/MathModule.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  ___cxa_free_exception(($2|0)); //@line 268 "modules/math/MathModule.cpp"
  ___resumeException($5|0);
  // unreachable;
 } else {
  ___cxa_throw(($2|0),(96|0),(344|0)); //@line 268 "modules/math/MathModule.cpp"
  // unreachable; //@line 268 "modules/math/MathModule.cpp"
 }
 return (0)|0;
}
function __ZN4love4math15RandomGeneratorC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 54 "modules/math/RandomGenerator.cpp"
 HEAP32[$this>>2] = (16680); //@line 56 "modules/math/RandomGenerator.cpp"
 $0 = ((($this)) + 24|0); //@line 55 "modules/math/RandomGenerator.cpp"
 HEAPF64[$0>>3] = inf; //@line 55 "modules/math/RandomGenerator.cpp"
 $1 = ((($this)) + 8|0); //@line 93 "modules/math/RandomGenerator.cpp"
 $2 = $1; //@line 93 "modules/math/RandomGenerator.cpp"
 $3 = $2; //@line 93 "modules/math/RandomGenerator.cpp"
 HEAP32[$3>>2] = -876643772; //@line 93 "modules/math/RandomGenerator.cpp"
 $4 = (($2) + 4)|0; //@line 93 "modules/math/RandomGenerator.cpp"
 $5 = $4; //@line 93 "modules/math/RandomGenerator.cpp"
 HEAP32[$5>>2] = 20529293; //@line 93 "modules/math/RandomGenerator.cpp"
 $6 = ((($this)) + 16|0); //@line 104 "modules/math/RandomGenerator.cpp"
 $7 = $6; //@line 104 "modules/math/RandomGenerator.cpp"
 $8 = $7; //@line 104 "modules/math/RandomGenerator.cpp"
 HEAP32[$8>>2] = -1308989176; //@line 104 "modules/math/RandomGenerator.cpp"
 $9 = (($7) + 4)|0; //@line 104 "modules/math/RandomGenerator.cpp"
 $10 = $9; //@line 104 "modules/math/RandomGenerator.cpp"
 HEAP32[$10>>2] = 204383418; //@line 104 "modules/math/RandomGenerator.cpp"
 return; //@line 63 "modules/math/RandomGenerator.cpp"
}
function __ZN4love4math15RandomGeneratorD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 60 "modules/math/RandomGenerator.h"
 return; //@line 60 "modules/math/RandomGenerator.h"
}
function __ZN4love4math15RandomGenerator7setSeedENS1_4SeedE($this,$newseed) {
 $this = $this|0;
 $newseed = $newseed|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 93 "modules/math/RandomGenerator.cpp"
 $1 = $newseed; //@line 93 "modules/math/RandomGenerator.cpp"
 $2 = $1; //@line 93 "modules/math/RandomGenerator.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 93 "modules/math/RandomGenerator.cpp"
 $4 = (($1) + 4)|0; //@line 93 "modules/math/RandomGenerator.cpp"
 $5 = $4; //@line 93 "modules/math/RandomGenerator.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 93 "modules/math/RandomGenerator.cpp"
 $7 = $0; //@line 93 "modules/math/RandomGenerator.cpp"
 $8 = $7; //@line 93 "modules/math/RandomGenerator.cpp"
 HEAP32[$8>>2] = $3; //@line 93 "modules/math/RandomGenerator.cpp"
 $9 = (($7) + 4)|0; //@line 93 "modules/math/RandomGenerator.cpp"
 $10 = $9; //@line 93 "modules/math/RandomGenerator.cpp"
 HEAP32[$10>>2] = $6; //@line 93 "modules/math/RandomGenerator.cpp"
 $12 = $3;$14 = $6;
 while(1) {
  $11 = $12 ^ -1; //@line 40 "modules/math/RandomGenerator.cpp"
  $13 = $14 ^ -1; //@line 40 "modules/math/RandomGenerator.cpp"
  $15 = (_bitshift64Shl(($12|0),($14|0),21)|0); //@line 40 "modules/math/RandomGenerator.cpp"
  $16 = tempRet0; //@line 40 "modules/math/RandomGenerator.cpp"
  $17 = (_i64Add(($15|0),($16|0),($11|0),($13|0))|0); //@line 40 "modules/math/RandomGenerator.cpp"
  $18 = tempRet0; //@line 40 "modules/math/RandomGenerator.cpp"
  $19 = (_bitshift64Lshr(($17|0),($18|0),24)|0); //@line 41 "modules/math/RandomGenerator.cpp"
  $20 = tempRet0; //@line 41 "modules/math/RandomGenerator.cpp"
  $21 = $19 ^ $17; //@line 41 "modules/math/RandomGenerator.cpp"
  $22 = $20 ^ $18; //@line 41 "modules/math/RandomGenerator.cpp"
  $23 = (___muldi3(($21|0),($22|0),265,0)|0); //@line 42 "modules/math/RandomGenerator.cpp"
  $24 = tempRet0; //@line 42 "modules/math/RandomGenerator.cpp"
  $25 = (_bitshift64Lshr(($23|0),($24|0),14)|0); //@line 43 "modules/math/RandomGenerator.cpp"
  $26 = tempRet0; //@line 43 "modules/math/RandomGenerator.cpp"
  $27 = $25 ^ $23; //@line 43 "modules/math/RandomGenerator.cpp"
  $28 = $26 ^ $24; //@line 43 "modules/math/RandomGenerator.cpp"
  $29 = (___muldi3(($27|0),($28|0),21,0)|0); //@line 44 "modules/math/RandomGenerator.cpp"
  $30 = tempRet0; //@line 44 "modules/math/RandomGenerator.cpp"
  $31 = (_bitshift64Lshr(($29|0),($30|0),28)|0); //@line 45 "modules/math/RandomGenerator.cpp"
  $32 = tempRet0; //@line 45 "modules/math/RandomGenerator.cpp"
  $33 = $31 ^ $29; //@line 45 "modules/math/RandomGenerator.cpp"
  $34 = $32 ^ $30; //@line 45 "modules/math/RandomGenerator.cpp"
  $35 = (___muldi3(($33|0),($34|0),-2147483647,0)|0); //@line 46 "modules/math/RandomGenerator.cpp"
  $36 = tempRet0; //@line 46 "modules/math/RandomGenerator.cpp"
  $37 = ($35|0)==(0); //@line 102 "modules/math/RandomGenerator.cpp"
  $38 = ($36|0)==(0); //@line 102 "modules/math/RandomGenerator.cpp"
  $39 = $37 & $38; //@line 102 "modules/math/RandomGenerator.cpp"
  if ($39) {
   $12 = 0;$14 = 0;
  } else {
   $43 = $35;$46 = $36;
   break;
  }
 }
 $40 = ((($this)) + 16|0); //@line 104 "modules/math/RandomGenerator.cpp"
 $41 = $40; //@line 104 "modules/math/RandomGenerator.cpp"
 $42 = $41; //@line 104 "modules/math/RandomGenerator.cpp"
 HEAP32[$42>>2] = $43; //@line 104 "modules/math/RandomGenerator.cpp"
 $44 = (($41) + 4)|0; //@line 104 "modules/math/RandomGenerator.cpp"
 $45 = $44; //@line 104 "modules/math/RandomGenerator.cpp"
 HEAP32[$45>>2] = $46; //@line 104 "modules/math/RandomGenerator.cpp"
 return; //@line 105 "modules/math/RandomGenerator.cpp"
}
function __ZN4love4math15RandomGenerator4randEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 67 "modules/math/RandomGenerator.cpp"
 $1 = $0; //@line 67 "modules/math/RandomGenerator.cpp"
 $2 = $1; //@line 67 "modules/math/RandomGenerator.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 67 "modules/math/RandomGenerator.cpp"
 $4 = (($1) + 4)|0; //@line 67 "modules/math/RandomGenerator.cpp"
 $5 = $4; //@line 67 "modules/math/RandomGenerator.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 67 "modules/math/RandomGenerator.cpp"
 $7 = (_bitshift64Lshr(($3|0),($6|0),12)|0); //@line 67 "modules/math/RandomGenerator.cpp"
 $8 = tempRet0; //@line 67 "modules/math/RandomGenerator.cpp"
 $9 = $7 ^ $3; //@line 67 "modules/math/RandomGenerator.cpp"
 $10 = $8 ^ $6; //@line 67 "modules/math/RandomGenerator.cpp"
 $11 = (_bitshift64Shl(($9|0),($10|0),25)|0); //@line 68 "modules/math/RandomGenerator.cpp"
 $12 = tempRet0; //@line 68 "modules/math/RandomGenerator.cpp"
 $13 = $11 ^ $9; //@line 68 "modules/math/RandomGenerator.cpp"
 $14 = $12 ^ $10; //@line 68 "modules/math/RandomGenerator.cpp"
 $15 = (_bitshift64Lshr(($13|0),($14|0),27)|0); //@line 69 "modules/math/RandomGenerator.cpp"
 $16 = tempRet0; //@line 69 "modules/math/RandomGenerator.cpp"
 $17 = $15 ^ $13; //@line 69 "modules/math/RandomGenerator.cpp"
 $18 = $16 ^ $14; //@line 69 "modules/math/RandomGenerator.cpp"
 $19 = $0; //@line 69 "modules/math/RandomGenerator.cpp"
 $20 = $19; //@line 69 "modules/math/RandomGenerator.cpp"
 HEAP32[$20>>2] = $17; //@line 69 "modules/math/RandomGenerator.cpp"
 $21 = (($19) + 4)|0; //@line 69 "modules/math/RandomGenerator.cpp"
 $22 = $21; //@line 69 "modules/math/RandomGenerator.cpp"
 HEAP32[$22>>2] = $18; //@line 69 "modules/math/RandomGenerator.cpp"
 $23 = (___muldi3(($17|0),($18|0),1332534557,625341585)|0); //@line 70 "modules/math/RandomGenerator.cpp"
 $24 = tempRet0; //@line 70 "modules/math/RandomGenerator.cpp"
 tempRet0 = ($24); //@line 70 "modules/math/RandomGenerator.cpp"
 return ($23|0); //@line 70 "modules/math/RandomGenerator.cpp"
}
function __ZN4love4math15RandomGenerator12randomNormalEd($this,$stddev) {
 $this = $this|0;
 $stddev = +$stddev;
 var $$0 = 0.0, $$pn = 0.0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0.0;
 var $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 77 "modules/math/RandomGenerator.cpp"
 $1 = +HEAPF64[$0>>3]; //@line 77 "modules/math/RandomGenerator.cpp"
 $2 = $1 != inf; //@line 77 "modules/math/RandomGenerator.cpp"
 if ($2) {
  HEAPF64[$0>>3] = inf; //@line 80 "modules/math/RandomGenerator.cpp"
  $$pn = $1;
  $$0 = $$pn * $stddev; //@line 81 "modules/math/RandomGenerator.cpp"
  return (+$$0); //@line 89 "modules/math/RandomGenerator.cpp"
 } else {
  $3 = ((($this)) + 16|0); //@line 67 "modules/math/RandomGenerator.cpp"
  $4 = $3; //@line 67 "modules/math/RandomGenerator.cpp"
  $5 = $4; //@line 67 "modules/math/RandomGenerator.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 67 "modules/math/RandomGenerator.cpp"
  $7 = (($4) + 4)|0; //@line 67 "modules/math/RandomGenerator.cpp"
  $8 = $7; //@line 67 "modules/math/RandomGenerator.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 67 "modules/math/RandomGenerator.cpp"
  $10 = (_bitshift64Lshr(($6|0),($9|0),12)|0); //@line 67 "modules/math/RandomGenerator.cpp"
  $11 = tempRet0; //@line 67 "modules/math/RandomGenerator.cpp"
  $12 = $10 ^ $6; //@line 67 "modules/math/RandomGenerator.cpp"
  $13 = $11 ^ $9; //@line 67 "modules/math/RandomGenerator.cpp"
  $14 = (_bitshift64Shl(($12|0),($13|0),25)|0); //@line 68 "modules/math/RandomGenerator.cpp"
  $15 = tempRet0; //@line 68 "modules/math/RandomGenerator.cpp"
  $16 = $14 ^ $12; //@line 68 "modules/math/RandomGenerator.cpp"
  $17 = $15 ^ $13; //@line 68 "modules/math/RandomGenerator.cpp"
  $18 = (_bitshift64Lshr(($16|0),($17|0),27)|0); //@line 69 "modules/math/RandomGenerator.cpp"
  $19 = tempRet0; //@line 69 "modules/math/RandomGenerator.cpp"
  $20 = $18 ^ $16; //@line 69 "modules/math/RandomGenerator.cpp"
  $21 = $19 ^ $17; //@line 69 "modules/math/RandomGenerator.cpp"
  $22 = (___muldi3(($20|0),($21|0),1332534557,625341585)|0); //@line 70 "modules/math/RandomGenerator.cpp"
  $23 = tempRet0; //@line 70 "modules/math/RandomGenerator.cpp"
  $24 = (+($22>>>0)) + (4294967296.0*(+($23>>>0))); //@line 76 "modules/math/RandomGenerator.h"
  $25 = $24 * 5.4210108624275222E-20; //@line 76 "modules/math/RandomGenerator.h"
  $26 = 1.0 - $25; //@line 84 "modules/math/RandomGenerator.cpp"
  $27 = (+Math_log((+$26))); //@line 84 "modules/math/RandomGenerator.cpp"
  $28 = $27 * -2.0; //@line 84 "modules/math/RandomGenerator.cpp"
  $29 = (+Math_sqrt((+$28))); //@line 84 "modules/math/RandomGenerator.cpp"
  $30 = (_bitshift64Lshr(($20|0),($21|0),12)|0); //@line 67 "modules/math/RandomGenerator.cpp"
  $31 = tempRet0; //@line 67 "modules/math/RandomGenerator.cpp"
  $32 = $30 ^ $20; //@line 67 "modules/math/RandomGenerator.cpp"
  $33 = $31 ^ $21; //@line 67 "modules/math/RandomGenerator.cpp"
  $34 = (_bitshift64Shl(($32|0),($33|0),25)|0); //@line 68 "modules/math/RandomGenerator.cpp"
  $35 = tempRet0; //@line 68 "modules/math/RandomGenerator.cpp"
  $36 = $34 ^ $32; //@line 68 "modules/math/RandomGenerator.cpp"
  $37 = $35 ^ $33; //@line 68 "modules/math/RandomGenerator.cpp"
  $38 = (_bitshift64Lshr(($36|0),($37|0),27)|0); //@line 69 "modules/math/RandomGenerator.cpp"
  $39 = tempRet0; //@line 69 "modules/math/RandomGenerator.cpp"
  $40 = $38 ^ $36; //@line 69 "modules/math/RandomGenerator.cpp"
  $41 = $39 ^ $37; //@line 69 "modules/math/RandomGenerator.cpp"
  $42 = $3; //@line 69 "modules/math/RandomGenerator.cpp"
  $43 = $42; //@line 69 "modules/math/RandomGenerator.cpp"
  HEAP32[$43>>2] = $40; //@line 69 "modules/math/RandomGenerator.cpp"
  $44 = (($42) + 4)|0; //@line 69 "modules/math/RandomGenerator.cpp"
  $45 = $44; //@line 69 "modules/math/RandomGenerator.cpp"
  HEAP32[$45>>2] = $41; //@line 69 "modules/math/RandomGenerator.cpp"
  $46 = (___muldi3(($40|0),($41|0),1332534557,625341585)|0); //@line 70 "modules/math/RandomGenerator.cpp"
  $47 = tempRet0; //@line 70 "modules/math/RandomGenerator.cpp"
  $48 = (+($46>>>0)) + (4294967296.0*(+($47>>>0))); //@line 76 "modules/math/RandomGenerator.h"
  $49 = $48 * 5.4210108624275222E-20; //@line 76 "modules/math/RandomGenerator.h"
  $50 = 1.0 - $49; //@line 85 "modules/math/RandomGenerator.cpp"
  $51 = $50 * 6.2831853071795862; //@line 85 "modules/math/RandomGenerator.cpp"
  $52 = (+Math_cos((+$51))); //@line 87 "modules/math/RandomGenerator.cpp"
  $53 = $29 * $52; //@line 87 "modules/math/RandomGenerator.cpp"
  HEAPF64[$0>>3] = $53; //@line 87 "modules/math/RandomGenerator.cpp"
  $54 = (+Math_sin((+$51))); //@line 88 "modules/math/RandomGenerator.cpp"
  $55 = $29 * $54; //@line 88 "modules/math/RandomGenerator.cpp"
  $$pn = $55;
  $$0 = $$pn * $stddev; //@line 81 "modules/math/RandomGenerator.cpp"
  return (+$$0); //@line 89 "modules/math/RandomGenerator.cpp"
 }
 return +(0.0);
}
function __ZNK4love4math15RandomGenerator7getSeedEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 109 "modules/math/RandomGenerator.cpp"
 $1 = $0; //@line 109 "modules/math/RandomGenerator.cpp"
 $2 = $1; //@line 109 "modules/math/RandomGenerator.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 109 "modules/math/RandomGenerator.cpp"
 $4 = (($1) + 4)|0; //@line 109 "modules/math/RandomGenerator.cpp"
 $5 = $4; //@line 109 "modules/math/RandomGenerator.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 109 "modules/math/RandomGenerator.cpp"
 $7 = $agg$result; //@line 109 "modules/math/RandomGenerator.cpp"
 $8 = $7; //@line 109 "modules/math/RandomGenerator.cpp"
 HEAP32[$8>>2] = $3; //@line 109 "modules/math/RandomGenerator.cpp"
 $9 = (($7) + 4)|0; //@line 109 "modules/math/RandomGenerator.cpp"
 $10 = $9; //@line 109 "modules/math/RandomGenerator.cpp"
 HEAP32[$10>>2] = $6; //@line 109 "modules/math/RandomGenerator.cpp"
 return; //@line 109 "modules/math/RandomGenerator.cpp"
}
function __ZN4love4math15RandomGenerator8setStateERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE($this,$statestr) {
 $this = $this|0;
 $statestr = $statestr|0;
 var $$0 = 0, $$01 = 0, $$016$i$i$i = 0, $$016$i$i$i$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $__m1$0$i$i$i = 0, $__m2$0$i$i$i = 0, $end = 0, $notlhs = 0, $notrhs = 0, $or$cond$not = 0, $or$cond8 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $end = sp + 12|0;
 $0 = HEAP8[$statestr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $1 = $0 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = ($1<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ((($statestr)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = HEAP32[$3>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($statestr)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = HEAP32[$5>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($statestr)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $0&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $8 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $2 ? $7 : $4; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $2 ? $9 : $6; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = ($11>>>0)<(2); //@line 962 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 L1: do {
  if (!($12)) {
   $13 = (($10) + ($11)|0); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $14 = ($11|0)<(2); //@line 1467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   if (!($14)) {
    $15 = ((($13)) + -1|0); //@line 1469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $16 = ($15|0)==($10|0); //@line 1475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    if (!($16)) {
     $$016$i$i$i = $10;
     L4: while(1) {
      $17 = HEAP8[$$016$i$i$i>>0]|0; //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $18 = ($17<<24>>24)==(48); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($18) {
       $__m1$0$i$i$i = $$016$i$i$i;$__m2$0$i$i$i = 355721;
       while(1) {
        $21 = ((($__m2$0$i$i$i)) + 1|0); //@line 1517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $22 = ($21|0)==((355723)|0); //@line 1517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        if ($22) {
         $$016$i$i$i$lcssa = $$016$i$i$i;
         break L4;
        }
        $23 = ((($__m1$0$i$i$i)) + 1|0); //@line 1519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $24 = HEAP8[$23>>0]|0; //@line 1520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $25 = HEAP8[$21>>0]|0; //@line 1520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $26 = ($24<<24>>24)==($25<<24>>24); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if ($26) {
         $__m1$0$i$i$i = $23;$__m2$0$i$i$i = $21;
        } else {
         break;
        }
       }
      }
      $19 = ((($$016$i$i$i)) + 1|0); //@line 1479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $20 = ($19|0)==($15|0); //@line 1475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if ($20) {
       break L1;
      } else {
       $$016$i$i$i = $19;
      }
     }
     $notlhs = ($$016$i$i$i$lcssa|0)==($13|0);
     $notrhs = ($$016$i$i$i$lcssa|0)!=($10|0);
     $or$cond$not = $notrhs | $notlhs;
     $27 = ($11>>>0)<(3); //@line 118 "modules/math/RandomGenerator.cpp"
     $or$cond8 = $or$cond$not | $27;
     if (!($or$cond8)) {
      HEAP32[$end>>2] = 0; //@line 123 "modules/math/RandomGenerator.cpp"
      $38 = (_strtoull($10,$end,16)|0); //@line 124 "modules/math/RandomGenerator.cpp"
      $39 = tempRet0; //@line 124 "modules/math/RandomGenerator.cpp"
      $40 = HEAP32[$end>>2]|0; //@line 126 "modules/math/RandomGenerator.cpp"
      $41 = ($40|0)==(0|0); //@line 126 "modules/math/RandomGenerator.cpp"
      if ($41) {
       $54 = ((($this)) + 16|0); //@line 129 "modules/math/RandomGenerator.cpp"
       $55 = $54; //@line 129 "modules/math/RandomGenerator.cpp"
       $56 = $55; //@line 129 "modules/math/RandomGenerator.cpp"
       HEAP32[$56>>2] = $38; //@line 129 "modules/math/RandomGenerator.cpp"
       $57 = (($55) + 4)|0; //@line 129 "modules/math/RandomGenerator.cpp"
       $58 = $57; //@line 129 "modules/math/RandomGenerator.cpp"
       HEAP32[$58>>2] = $39; //@line 129 "modules/math/RandomGenerator.cpp"
       STACKTOP = sp;return; //@line 130 "modules/math/RandomGenerator.cpp"
      }
      $42 = HEAP8[$40>>0]|0; //@line 126 "modules/math/RandomGenerator.cpp"
      $43 = ($42<<24>>24)==(0); //@line 126 "modules/math/RandomGenerator.cpp"
      if ($43) {
       $54 = ((($this)) + 16|0); //@line 129 "modules/math/RandomGenerator.cpp"
       $55 = $54; //@line 129 "modules/math/RandomGenerator.cpp"
       $56 = $55; //@line 129 "modules/math/RandomGenerator.cpp"
       HEAP32[$56>>2] = $38; //@line 129 "modules/math/RandomGenerator.cpp"
       $57 = (($55) + 4)|0; //@line 129 "modules/math/RandomGenerator.cpp"
       $58 = $57; //@line 129 "modules/math/RandomGenerator.cpp"
       HEAP32[$58>>2] = $39; //@line 129 "modules/math/RandomGenerator.cpp"
       STACKTOP = sp;return; //@line 130 "modules/math/RandomGenerator.cpp"
      }
      $44 = (___cxa_allocate_exception(16)|0); //@line 127 "modules/math/RandomGenerator.cpp"
      $45 = HEAP8[$statestr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $46 = $45 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $47 = ($46<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $48 = HEAP32[$3>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $49 = $47 ? $7 : $48; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __THREW__ = 0;
      HEAP32[$vararg_buffer1>>2] = $49; //@line 127 "modules/math/RandomGenerator.cpp"
      invoke_viii(3349,($44|0),(355724|0),($vararg_buffer1|0)); //@line 127 "modules/math/RandomGenerator.cpp"
      $50 = __THREW__; __THREW__ = 0;
      $51 = $50&1;
      if (!($51)) {
       ___cxa_throw(($44|0),(96|0),(344|0)); //@line 127 "modules/math/RandomGenerator.cpp"
       // unreachable; //@line 127 "modules/math/RandomGenerator.cpp"
      }
      $52 = ___cxa_find_matching_catch_2()|0;
      $53 = tempRet0;
      ___cxa_free_exception(($44|0)); //@line 127 "modules/math/RandomGenerator.cpp"
      $$0 = $53;$$01 = $52;
      ___resumeException($$01|0);
      // unreachable;
     }
    }
   }
  }
 } while(0);
 $28 = (___cxa_allocate_exception(16)|0); //@line 119 "modules/math/RandomGenerator.cpp"
 $29 = HEAP8[$statestr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $30 = $29 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $31 = ($30<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $32 = HEAP32[$3>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $33 = $31 ? $7 : $32; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $33; //@line 119 "modules/math/RandomGenerator.cpp"
 invoke_viii(3349,($28|0),(355724|0),($vararg_buffer|0)); //@line 119 "modules/math/RandomGenerator.cpp"
 $34 = __THREW__; __THREW__ = 0;
 $35 = $34&1;
 if (!($35)) {
  ___cxa_throw(($28|0),(96|0),(344|0)); //@line 119 "modules/math/RandomGenerator.cpp"
  // unreachable; //@line 119 "modules/math/RandomGenerator.cpp"
 }
 $36 = ___cxa_find_matching_catch_2()|0;
 $37 = tempRet0;
 ___cxa_free_exception(($28|0)); //@line 119 "modules/math/RandomGenerator.cpp"
 $$0 = $37;$$01 = $36;
 ___resumeException($$01|0);
 // unreachable;
}
function __ZNK4love4math15RandomGenerator8getStateEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$0$i = 0, $$01$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0;
 var $eh$lpad$body$i$indexZ2D = 0, $ss = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 160|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 144|0;
 $ss = sp;
 $1 = ((($ss)) + 64|0); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $2 = ((($ss)) + 8|0); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$2>>2] = (9320); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $3 = ((($ss)) + 12|0); //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$ss>>2] = ((9524)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 HEAP32[$1>>2] = ((9544)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $4 = ((($ss)) + 4|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 HEAP32[$4>>2] = 0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $5 = ((($ss)) + 64|0); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 __THREW__ = 0;
 invoke_vii(3362,($5|0),($3|0)); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $18 = ___cxa_find_matching_catch_2()|0;
  $19 = tempRet0;
  $$0$i = $19;$$01$i = $18;
  __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($1); //@line 900 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  ___resumeException($$01$i|0);
  // unreachable;
 }
 $8 = ((($ss)) + 136|0); //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$8>>2] = 0; //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 $9 = ((($ss)) + 140|0); //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$9>>2] = -1; //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$ss>>2] = (9300); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$1>>2] = (9340); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$2>>2] = (9320); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 __THREW__ = 0;
 invoke_vi(3363,($3|0)); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 do {
  if ($11) {
   $20 = ___cxa_find_matching_catch_2()|0;
   $21 = tempRet0;
   $eh$lpad$body$i$index2Z2D = $21;$eh$lpad$body$i$indexZ2D = $20;
  } else {
   HEAP32[$3>>2] = (9356); //@line 243 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $12 = ((($ss)) + 44|0); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $13 = ((($ss)) + 60|0); //@line 242 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   ;HEAP32[$12>>2]=0|0;HEAP32[$12+4>>2]=0|0;HEAP32[$12+8>>2]=0|0;HEAP32[$12+12>>2]=0|0; //@line 241 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   HEAP32[$13>>2] = 24; //@line 242 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   invoke_vii(3364,($3|0),($0|0)); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $14 = __THREW__; __THREW__ = 0;
   $15 = $14&1;
   if ($15) {
    $16 = ___cxa_find_matching_catch_2()|0;
    $17 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12); //@line 245 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($3); //@line 245 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $eh$lpad$body$i$index2Z2D = $17;$eh$lpad$body$i$indexZ2D = $16;
    break;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $22 = ((($ss)) + 8|0); //@line 137 "modules/math/RandomGenerator.cpp"
   __THREW__ = 0;
   $23 = (invoke_iiii(3366,($22|0),(355721|0),2)|0); //@line 894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
   $24 = __THREW__; __THREW__ = 0;
   $25 = $24&1;
   if (!($25)) {
    $26 = HEAP32[$23>>2]|0; //@line 188 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iomanip"
    $27 = ((($26)) + -12|0); //@line 188 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iomanip"
    $28 = HEAP32[$27>>2]|0; //@line 188 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iomanip"
    $29 = (($23) + ($28)|0); //@line 188 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iomanip"
    $30 = ((($29)) + 76|0); //@line 752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    HEAP32[$30>>2] = 48; //@line 753 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    $31 = HEAP32[$23>>2]|0; //@line 263 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iomanip"
    $32 = ((($31)) + -12|0); //@line 263 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iomanip"
    $33 = HEAP32[$32>>2]|0; //@line 263 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iomanip"
    $34 = (($23) + ($33)|0); //@line 263 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iomanip"
    $35 = ((($34)) + 12|0); //@line 506 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    HEAP32[$35>>2] = 16; //@line 507 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    $36 = HEAP32[$23>>2]|0; //@line 340 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
    $37 = ((($36)) + -12|0); //@line 340 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
    $38 = HEAP32[$37>>2]|0; //@line 340 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
    $39 = (($23) + ($38)|0); //@line 340 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
    $40 = ((($39)) + 4|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    $41 = HEAP32[$40>>2]|0; //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    $42 = $41 & -75; //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    $43 = $42 | 8; //@line 471 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    HEAP32[$40>>2] = $43; //@line 471 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    $44 = ((($this)) + 16|0); //@line 137 "modules/math/RandomGenerator.cpp"
    $45 = $44; //@line 137 "modules/math/RandomGenerator.cpp"
    $46 = $45; //@line 137 "modules/math/RandomGenerator.cpp"
    $47 = HEAP32[$46>>2]|0; //@line 137 "modules/math/RandomGenerator.cpp"
    $48 = (($45) + 4)|0; //@line 137 "modules/math/RandomGenerator.cpp"
    $49 = $48; //@line 137 "modules/math/RandomGenerator.cpp"
    $50 = HEAP32[$49>>2]|0; //@line 137 "modules/math/RandomGenerator.cpp"
    __THREW__ = 0;
    (invoke_iiii(3657,($23|0),($47|0),($50|0))|0); //@line 137 "modules/math/RandomGenerator.cpp"
    $51 = __THREW__; __THREW__ = 0;
    $52 = $51&1;
    if (!($52)) {
     __THREW__ = 0;
     invoke_vii(3367,($agg$result|0),($3|0)); //@line 964 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
     $53 = __THREW__; __THREW__ = 0;
     $54 = $53&1;
     if (!($54)) {
      HEAP32[$ss>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      HEAP32[$1>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      HEAP32[$22>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      HEAP32[$3>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($3); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($1); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      STACKTOP = sp;return; //@line 140 "modules/math/RandomGenerator.cpp"
     }
    }
   }
   $55 = ___cxa_find_matching_catch_2()|0;
   $56 = tempRet0;
   HEAP32[$ss>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$1>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$22>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$3>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($3); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($1); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   ___resumeException($55|0);
   // unreachable;
  }
 } while(0);
 $$0$i = $eh$lpad$body$i$index2Z2D;$$01$i = $eh$lpad$body$i$indexZ2D;
 __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($1); //@line 900 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 ___resumeException($$01$i|0);
 // unreachable;
}
function __ZN4love14luax_checktypeINS_4math11BezierCurveEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love4math23w_BezierCurve_getDegreeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4math11BezierCurveEEEPT_P9lua_StateiNS_4TypeE($L,1,26)|0); //@line 33 "modules/math/wrap_BezierCurve.cpp"
 $1 = ((($0)) + 8|0); //@line 48 "modules/math/BezierCurve.h"
 $2 = ((($0)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = $5 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = (($6) + -1)|0; //@line 48 "modules/math/BezierCurve.h"
 $8 = (+($7>>>0)); //@line 39 "modules/math/wrap_BezierCurve.cpp"
 _lua_pushnumber($L,$8); //@line 39 "modules/math/wrap_BezierCurve.cpp"
 return 1; //@line 40 "modules/math/wrap_BezierCurve.cpp"
}
function __ZN4love4math27w_BezierCurve_getDerivativeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4math11BezierCurveEEEPT_P9lua_StateiNS_4TypeE($L,1,26)|0); //@line 33 "modules/math/wrap_BezierCurve.cpp"
 $1 = (__Znwj(20)|0); //@line 46 "modules/math/wrap_BezierCurve.cpp"
 __THREW__ = 0;
 invoke_vii(3658,($1|0),($0|0)); //@line 46 "modules/math/wrap_BezierCurve.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZdlPv($1); //@line 46 "modules/math/wrap_BezierCurve.cpp"
  ___resumeException($7|0);
  // unreachable;
 } else {
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,26,$1); //@line 47 "modules/math/wrap_BezierCurve.cpp"
  $4 = HEAP32[$1>>2]|0; //@line 48 "modules/math/wrap_BezierCurve.cpp"
  $5 = ((($4)) + 12|0); //@line 48 "modules/math/wrap_BezierCurve.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 48 "modules/math/wrap_BezierCurve.cpp"
  FUNCTION_TABLE_vi[$6 & 8191]($1); //@line 48 "modules/math/wrap_BezierCurve.cpp"
  return 1; //@line 49 "modules/math/wrap_BezierCurve.cpp"
 }
 return (0)|0;
}
function __ZN4love4math29w_BezierCurve_getControlPointEP9lua_State($L) {
 $L = $L|0;
 var $$ = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 52 "modules/math/wrap_BezierCurve.cpp"
 $0 = (__ZN4love14luax_checktypeINS_4math11BezierCurveEEEPT_P9lua_StateiNS_4TypeE($L,1,26)|0); //@line 33 "modules/math/wrap_BezierCurve.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 55 "modules/math/wrap_BezierCurve.cpp"
 $2 = (~~(($1))); //@line 55 "modules/math/wrap_BezierCurve.cpp"
 $3 = ($2|0)>(0); //@line 57 "modules/math/wrap_BezierCurve.cpp"
 $4 = $3 << 31 >> 31; //@line 57 "modules/math/wrap_BezierCurve.cpp"
 $$ = (($4) + ($2))|0; //@line 57 "modules/math/wrap_BezierCurve.cpp"
 __THREW__ = 0;
 $5 = (invoke_iii(3659,($0|0),($$|0))|0); //@line 61 "modules/math/wrap_BezierCurve.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  $8 = +HEAPF32[$5>>2]; //@line 61 "modules/math/wrap_BezierCurve.cpp"
  $9 = ((($5)) + 4|0); //@line 61 "modules/math/wrap_BezierCurve.cpp"
  $10 = +HEAPF32[$9>>2]; //@line 61 "modules/math/wrap_BezierCurve.cpp"
  $11 = $8; //@line 62 "modules/math/wrap_BezierCurve.cpp"
  __THREW__ = 0;
  invoke_vid(3394,($L|0),(+$11)); //@line 62 "modules/math/wrap_BezierCurve.cpp"
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   $14 = $10; //@line 63 "modules/math/wrap_BezierCurve.cpp"
   __THREW__ = 0;
   invoke_vid(3394,($L|0),(+$14)); //@line 63 "modules/math/wrap_BezierCurve.cpp"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if (!($16)) {
    STACKTOP = sp;return 2; //@line 66 "modules/math/wrap_BezierCurve.cpp"
   }
  }
 }
 $17 = ___cxa_find_matching_catch_3(8992|0)|0;
 $18 = tempRet0;
 $19 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $20 = ($18|0)==($19|0); //@line 535 "./common/runtime.h"
 if (!($20)) {
  $$01$i = $18;$$02$i = $17;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $21 = (___cxa_begin_catch(($17|0))|0); //@line 535 "./common/runtime.h"
 $22 = HEAP32[$21>>2]|0; //@line 539 "./common/runtime.h"
 $23 = ((($22)) + 8|0); //@line 539 "./common/runtime.h"
 $24 = HEAP32[$23>>2]|0; //@line 539 "./common/runtime.h"
 $25 = (FUNCTION_TABLE_ii[$24 & 8191]($21)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($25|0)); //@line 539 "./common/runtime.h"
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if (!($27)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $28 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $28; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 2; //@line 66 "modules/math/wrap_BezierCurve.cpp"
 }
 $29 = ___cxa_find_matching_catch_2()|0;
 $30 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $31 = __THREW__; __THREW__ = 0;
 $32 = $31&1;
 if ($32) {
  $33 = ___cxa_find_matching_catch_3(0|0)|0;
  $34 = tempRet0;
  ___clang_call_terminate($33); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $30;$$02$i = $29;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love4math29w_BezierCurve_setControlPointEP9lua_State($L) {
 $L = $L|0;
 var $$ = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = (__ZN4love14luax_checktypeINS_4math11BezierCurveEEEPT_P9lua_StateiNS_4TypeE($L,1,26)|0); //@line 33 "modules/math/wrap_BezierCurve.cpp"
 $2 = (+_luaL_checknumber($L,2)); //@line 72 "modules/math/wrap_BezierCurve.cpp"
 $3 = (~~(($2))); //@line 72 "modules/math/wrap_BezierCurve.cpp"
 $4 = (+_luaL_checknumber($L,3)); //@line 73 "modules/math/wrap_BezierCurve.cpp"
 $5 = $4; //@line 73 "modules/math/wrap_BezierCurve.cpp"
 $6 = (+_luaL_checknumber($L,4)); //@line 74 "modules/math/wrap_BezierCurve.cpp"
 $7 = $6; //@line 74 "modules/math/wrap_BezierCurve.cpp"
 $8 = ($3|0)>(0); //@line 76 "modules/math/wrap_BezierCurve.cpp"
 $9 = $8 << 31 >> 31; //@line 76 "modules/math/wrap_BezierCurve.cpp"
 $$ = (($9) + ($3))|0; //@line 76 "modules/math/wrap_BezierCurve.cpp"
 HEAPF32[$0>>2] = $5; //@line 225 "./common/Vector.h"
 $10 = ((($0)) + 4|0); //@line 226 "./common/Vector.h"
 HEAPF32[$10>>2] = $7; //@line 226 "./common/Vector.h"
 __THREW__ = 0;
 invoke_viii(3660,($1|0),($$|0),($0|0)); //@line 79 "modules/math/wrap_BezierCurve.cpp"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if (!($12)) {
  STACKTOP = sp;return 0; //@line 80 "modules/math/wrap_BezierCurve.cpp"
 }
 $13 = ___cxa_find_matching_catch_3(8992|0)|0;
 $14 = tempRet0;
 $15 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $16 = ($14|0)==($15|0); //@line 535 "./common/runtime.h"
 if (!($16)) {
  $$01$i = $14;$$02$i = $13;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $17 = (___cxa_begin_catch(($13|0))|0); //@line 535 "./common/runtime.h"
 $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
 $19 = ((($18)) + 8|0); //@line 539 "./common/runtime.h"
 $20 = HEAP32[$19>>2]|0; //@line 539 "./common/runtime.h"
 $21 = (FUNCTION_TABLE_ii[$20 & 8191]($17)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($21|0)); //@line 539 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if (!($23)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $24 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $24; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 80 "modules/math/wrap_BezierCurve.cpp"
 }
 $25 = ___cxa_find_matching_catch_2()|0;
 $26 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $27 = __THREW__; __THREW__ = 0;
 $28 = $27&1;
 if ($28) {
  $29 = ___cxa_find_matching_catch_3(0|0)|0;
  $30 = tempRet0;
  ___clang_call_terminate($29); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $26;$$02$i = $25;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love4math32w_BezierCurve_insertControlPointEP9lua_State($L) {
 $L = $L|0;
 var $$ = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = (__ZN4love14luax_checktypeINS_4math11BezierCurveEEEPT_P9lua_StateiNS_4TypeE($L,1,26)|0); //@line 33 "modules/math/wrap_BezierCurve.cpp"
 $2 = (+_luaL_checknumber($L,2)); //@line 86 "modules/math/wrap_BezierCurve.cpp"
 $3 = $2; //@line 86 "modules/math/wrap_BezierCurve.cpp"
 $4 = (+_luaL_checknumber($L,3)); //@line 87 "modules/math/wrap_BezierCurve.cpp"
 $5 = $4; //@line 87 "modules/math/wrap_BezierCurve.cpp"
 $6 = (+_luaL_optnumber($L,4,-1.0)); //@line 88 "modules/math/wrap_BezierCurve.cpp"
 $7 = (~~(($6))); //@line 88 "modules/math/wrap_BezierCurve.cpp"
 $8 = ($7|0)>(0); //@line 90 "modules/math/wrap_BezierCurve.cpp"
 $9 = $8 << 31 >> 31; //@line 90 "modules/math/wrap_BezierCurve.cpp"
 $$ = (($9) + ($7))|0; //@line 90 "modules/math/wrap_BezierCurve.cpp"
 HEAPF32[$0>>2] = $3; //@line 225 "./common/Vector.h"
 $10 = ((($0)) + 4|0); //@line 226 "./common/Vector.h"
 HEAPF32[$10>>2] = $5; //@line 226 "./common/Vector.h"
 __THREW__ = 0;
 invoke_viii(3661,($1|0),($0|0),($$|0)); //@line 93 "modules/math/wrap_BezierCurve.cpp"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if (!($12)) {
  STACKTOP = sp;return 0; //@line 94 "modules/math/wrap_BezierCurve.cpp"
 }
 $13 = ___cxa_find_matching_catch_3(8992|0)|0;
 $14 = tempRet0;
 $15 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $16 = ($14|0)==($15|0); //@line 535 "./common/runtime.h"
 if (!($16)) {
  $$01$i = $14;$$02$i = $13;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $17 = (___cxa_begin_catch(($13|0))|0); //@line 535 "./common/runtime.h"
 $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
 $19 = ((($18)) + 8|0); //@line 539 "./common/runtime.h"
 $20 = HEAP32[$19>>2]|0; //@line 539 "./common/runtime.h"
 $21 = (FUNCTION_TABLE_ii[$20 & 8191]($17)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($21|0)); //@line 539 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if (!($23)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $24 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $24; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 94 "modules/math/wrap_BezierCurve.cpp"
 }
 $25 = ___cxa_find_matching_catch_2()|0;
 $26 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $27 = __THREW__; __THREW__ = 0;
 $28 = $27&1;
 if ($28) {
  $29 = ___cxa_find_matching_catch_3(0|0)|0;
  $30 = tempRet0;
  ___clang_call_terminate($29); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $26;$$02$i = $25;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love4math32w_BezierCurve_removeControlPointEP9lua_State($L) {
 $L = $L|0;
 var $$ = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 97 "modules/math/wrap_BezierCurve.cpp"
 $0 = (__ZN4love14luax_checktypeINS_4math11BezierCurveEEEPT_P9lua_StateiNS_4TypeE($L,1,26)|0); //@line 33 "modules/math/wrap_BezierCurve.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 100 "modules/math/wrap_BezierCurve.cpp"
 $2 = (~~(($1))); //@line 100 "modules/math/wrap_BezierCurve.cpp"
 $3 = ($2|0)>(0); //@line 102 "modules/math/wrap_BezierCurve.cpp"
 $4 = $3 << 31 >> 31; //@line 102 "modules/math/wrap_BezierCurve.cpp"
 $$ = (($4) + ($2))|0; //@line 102 "modules/math/wrap_BezierCurve.cpp"
 __THREW__ = 0;
 invoke_vii(3662,($0|0),($$|0)); //@line 105 "modules/math/wrap_BezierCurve.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  STACKTOP = sp;return 0; //@line 106 "modules/math/wrap_BezierCurve.cpp"
 }
 $7 = ___cxa_find_matching_catch_3(8992|0)|0;
 $8 = tempRet0;
 $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
 if (!($10)) {
  $$01$i = $8;$$02$i = $7;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
 $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
 $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if (!($17)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $18 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $18; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 106 "modules/math/wrap_BezierCurve.cpp"
 }
 $19 = ___cxa_find_matching_catch_2()|0;
 $20 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $23 = ___cxa_find_matching_catch_3(0|0)|0;
  $24 = tempRet0;
  ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $20;$$02$i = $19;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love4math34w_BezierCurve_getControlPointCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4math11BezierCurveEEEPT_P9lua_StateiNS_4TypeE($L,1,26)|0); //@line 33 "modules/math/wrap_BezierCurve.cpp"
 $1 = ((($0)) + 8|0); //@line 87 "modules/math/BezierCurve.h"
 $2 = ((($0)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = $5 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 _lua_pushinteger($L,$6); //@line 112 "modules/math/wrap_BezierCurve.cpp"
 return 1; //@line 113 "modules/math/wrap_BezierCurve.cpp"
}
function __ZN4love4math23w_BezierCurve_translateEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__ZN4love14luax_checktypeINS_4math11BezierCurveEEEPT_P9lua_StateiNS_4TypeE($L,1,26)|0); //@line 33 "modules/math/wrap_BezierCurve.cpp"
 $2 = (+_luaL_checknumber($L,2)); //@line 119 "modules/math/wrap_BezierCurve.cpp"
 $3 = $2; //@line 119 "modules/math/wrap_BezierCurve.cpp"
 $4 = (+_luaL_checknumber($L,3)); //@line 120 "modules/math/wrap_BezierCurve.cpp"
 $5 = $4; //@line 120 "modules/math/wrap_BezierCurve.cpp"
 HEAPF32[$0>>2] = $3; //@line 225 "./common/Vector.h"
 $6 = ((($0)) + 4|0); //@line 226 "./common/Vector.h"
 HEAPF32[$6>>2] = $5; //@line 226 "./common/Vector.h"
 __ZN4love4math11BezierCurve9translateERKNS_6VectorE($1,$0); //@line 121 "modules/math/wrap_BezierCurve.cpp"
 STACKTOP = sp;return 0; //@line 122 "modules/math/wrap_BezierCurve.cpp"
}
function __ZN4love4math20w_BezierCurve_rotateEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__ZN4love14luax_checktypeINS_4math11BezierCurveEEEPT_P9lua_StateiNS_4TypeE($L,1,26)|0); //@line 33 "modules/math/wrap_BezierCurve.cpp"
 $2 = (+_luaL_checknumber($L,2)); //@line 128 "modules/math/wrap_BezierCurve.cpp"
 $3 = (+_luaL_optnumber($L,3,0.0)); //@line 129 "modules/math/wrap_BezierCurve.cpp"
 $4 = $3; //@line 129 "modules/math/wrap_BezierCurve.cpp"
 $5 = (+_luaL_optnumber($L,4,0.0)); //@line 130 "modules/math/wrap_BezierCurve.cpp"
 $6 = $5; //@line 130 "modules/math/wrap_BezierCurve.cpp"
 HEAPF32[$0>>2] = $4; //@line 225 "./common/Vector.h"
 $7 = ((($0)) + 4|0); //@line 226 "./common/Vector.h"
 HEAPF32[$7>>2] = $6; //@line 226 "./common/Vector.h"
 __ZN4love4math11BezierCurve6rotateEdRKNS_6VectorE($1,$2,$0); //@line 131 "modules/math/wrap_BezierCurve.cpp"
 STACKTOP = sp;return 0; //@line 132 "modules/math/wrap_BezierCurve.cpp"
}
function __ZN4love4math19w_BezierCurve_scaleEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__ZN4love14luax_checktypeINS_4math11BezierCurveEEEPT_P9lua_StateiNS_4TypeE($L,1,26)|0); //@line 33 "modules/math/wrap_BezierCurve.cpp"
 $2 = (+_luaL_checknumber($L,2)); //@line 138 "modules/math/wrap_BezierCurve.cpp"
 $3 = (+_luaL_optnumber($L,3,0.0)); //@line 139 "modules/math/wrap_BezierCurve.cpp"
 $4 = $3; //@line 139 "modules/math/wrap_BezierCurve.cpp"
 $5 = (+_luaL_optnumber($L,4,0.0)); //@line 140 "modules/math/wrap_BezierCurve.cpp"
 $6 = $5; //@line 140 "modules/math/wrap_BezierCurve.cpp"
 HEAPF32[$0>>2] = $4; //@line 225 "./common/Vector.h"
 $7 = ((($0)) + 4|0); //@line 226 "./common/Vector.h"
 HEAPF32[$7>>2] = $6; //@line 226 "./common/Vector.h"
 __ZN4love4math11BezierCurve5scaleEdRKNS_6VectorE($1,$2,$0); //@line 141 "modules/math/wrap_BezierCurve.cpp"
 STACKTOP = sp;return 0; //@line 142 "modules/math/wrap_BezierCurve.cpp"
}
function __ZN4love4math22w_BezierCurve_evaluateEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, $v$i$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $v$i$i = sp + 8|0;
 $0 = (__ZN4love14luax_checktypeINS_4math11BezierCurveEEEPT_P9lua_StateiNS_4TypeE($L,1,26)|0); //@line 33 "modules/math/wrap_BezierCurve.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 148 "modules/math/wrap_BezierCurve.cpp"
 __THREW__ = 0;
 invoke_viid(3663,($v$i$i|0),($0|0),(+$1)); //@line 151 "modules/math/wrap_BezierCurve.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  $4 = +HEAPF32[$v$i$i>>2]; //@line 152 "modules/math/wrap_BezierCurve.cpp"
  $5 = $4; //@line 152 "modules/math/wrap_BezierCurve.cpp"
  __THREW__ = 0;
  invoke_vid(3394,($L|0),(+$5)); //@line 152 "modules/math/wrap_BezierCurve.cpp"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if (!($7)) {
   $8 = ((($v$i$i)) + 4|0); //@line 153 "modules/math/wrap_BezierCurve.cpp"
   $9 = +HEAPF32[$8>>2]; //@line 153 "modules/math/wrap_BezierCurve.cpp"
   $10 = $9; //@line 153 "modules/math/wrap_BezierCurve.cpp"
   __THREW__ = 0;
   invoke_vid(3394,($L|0),(+$10)); //@line 153 "modules/math/wrap_BezierCurve.cpp"
   $11 = __THREW__; __THREW__ = 0;
   $12 = $11&1;
   if (!($12)) {
    STACKTOP = sp;return 2; //@line 156 "modules/math/wrap_BezierCurve.cpp"
   }
  }
 }
 $13 = ___cxa_find_matching_catch_3(8992|0)|0;
 $14 = tempRet0;
 $15 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $16 = ($14|0)==($15|0); //@line 535 "./common/runtime.h"
 if (!($16)) {
  $$01$i = $14;$$02$i = $13;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $17 = (___cxa_begin_catch(($13|0))|0); //@line 535 "./common/runtime.h"
 $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
 $19 = ((($18)) + 8|0); //@line 539 "./common/runtime.h"
 $20 = HEAP32[$19>>2]|0; //@line 539 "./common/runtime.h"
 $21 = (FUNCTION_TABLE_ii[$20 & 8191]($17)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($21|0)); //@line 539 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if (!($23)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $24 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $24; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 2; //@line 156 "modules/math/wrap_BezierCurve.cpp"
 }
 $25 = ___cxa_find_matching_catch_2()|0;
 $26 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $27 = __THREW__; __THREW__ = 0;
 $28 = $27&1;
 if ($28) {
  $29 = ___cxa_find_matching_catch_3(0|0)|0;
  $30 = tempRet0;
  ___clang_call_terminate($29); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $26;$$02$i = $25;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love4math24w_BezierCurve_getSegmentEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $segment$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 160 "modules/math/wrap_BezierCurve.cpp"
 $0 = (__ZN4love14luax_checktypeINS_4math11BezierCurveEEEPT_P9lua_StateiNS_4TypeE($L,1,26)|0); //@line 33 "modules/math/wrap_BezierCurve.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 163 "modules/math/wrap_BezierCurve.cpp"
 $2 = (+_luaL_checknumber($L,3)); //@line 164 "modules/math/wrap_BezierCurve.cpp"
 __THREW__ = 0;
 $3 = (invoke_iidd(3664,($0|0),(+$1),(+$2))|0); //@line 167 "modules/math/wrap_BezierCurve.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 do {
  if ($5) {
   $6 = ___cxa_find_matching_catch_3(8992|0)|0;
   $7 = tempRet0;
   $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
   if (!($9)) {
    $$01$i = $7;$$02$i = $6;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
   $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
   $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
   $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
   $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if (!($16)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $17; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $segment$0 = 0;
    break;
   }
   $18 = ___cxa_find_matching_catch_2()|0;
   $19 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if ($21) {
    $22 = ___cxa_find_matching_catch_3(0|0)|0;
    $23 = tempRet0;
    ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $19;$$02$i = $18;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $segment$0 = $3;
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,26,$segment$0); //@line 168 "modules/math/wrap_BezierCurve.cpp"
 $24 = HEAP32[$segment$0>>2]|0; //@line 169 "modules/math/wrap_BezierCurve.cpp"
 $25 = ((($24)) + 12|0); //@line 169 "modules/math/wrap_BezierCurve.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 169 "modules/math/wrap_BezierCurve.cpp"
 FUNCTION_TABLE_vi[$26 & 8191]($segment$0); //@line 169 "modules/math/wrap_BezierCurve.cpp"
 STACKTOP = sp;return 1; //@line 171 "modules/math/wrap_BezierCurve.cpp"
}
function __ZN4love4math20w_BezierCurve_renderEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body11$index8Z2D = 0, $eh$lpad$body11$indexZ2D = 0, $i$010 = 0, $points$sroa$0$0 = 0, $points$sroa$0$1$ph = 0, $points$sroa$0$2 = 0, $points$sroa$12$0 = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 $1 = (__ZN4love14luax_checktypeINS_4math11BezierCurveEEEPT_P9lua_StateiNS_4TypeE($L,1,26)|0); //@line 33 "modules/math/wrap_BezierCurve.cpp"
 $2 = (+_luaL_optnumber($L,2,5.0)); //@line 177 "modules/math/wrap_BezierCurve.cpp"
 $3 = (~~(($2))); //@line 177 "modules/math/wrap_BezierCurve.cpp"
 __THREW__ = 0;
 invoke_viii(3665,($0|0),($1|0),($3|0)); //@line 180 "modules/math/wrap_BezierCurve.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 L1: do {
  if ($5) {
   $9 = ___cxa_find_matching_catch_3(8992|0)|0;
   $10 = tempRet0;
   $11 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $12 = ($10|0)==($11|0); //@line 535 "./common/runtime.h"
   do {
    if ($12) {
     $13 = (___cxa_begin_catch(($9|0))|0); //@line 535 "./common/runtime.h"
     $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
     $15 = ((($14)) + 8|0); //@line 539 "./common/runtime.h"
     $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
     $17 = (FUNCTION_TABLE_ii[$16 & 8191]($13)|0); //@line 539 "./common/runtime.h"
     __THREW__ = 0;
     invoke_vii(3380,($L|0),($17|0)); //@line 539 "./common/runtime.h"
     $18 = __THREW__; __THREW__ = 0;
     $19 = $18&1;
     if ($19) {
      $27 = ___cxa_find_matching_catch_2()|0;
      $28 = tempRet0;
      __THREW__ = 0;
      invoke_v(3378); //@line 540 "./common/runtime.h"
      $29 = __THREW__; __THREW__ = 0;
      $30 = $29&1;
      if (!($30)) {
       $$01$i = $28;$$02$i = $27;
       break;
      }
      $31 = ___cxa_find_matching_catch_3(0|0)|0;
      $32 = tempRet0;
      ___clang_call_terminate($31); //@line 540 "./common/runtime.h"
      // unreachable; //@line 540 "./common/runtime.h"
     }
     __THREW__ = 0;
     invoke_v(3378); //@line 540 "./common/runtime.h"
     $20 = __THREW__; __THREW__ = 0;
     $21 = $20&1;
     if ($21) {
      $points$sroa$0$1$ph = 0;
      label = 17;
      break L1;
     }
     __THREW__ = 0;
     $22 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
     $23 = __THREW__; __THREW__ = 0;
     $24 = $23&1;
     if ($24) {
      $points$sroa$0$1$ph = 0;
      label = 17;
      break L1;
     }
     __THREW__ = 0;
     HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
     (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 543 "./common/runtime.h"
     $25 = __THREW__; __THREW__ = 0;
     $26 = $25&1;
     if ($26) {
      $points$sroa$0$1$ph = 0;
      label = 17;
      break L1;
     } else {
      $points$sroa$0$0 = 0;$points$sroa$12$0 = 0;
      label = 11;
      break L1;
     }
    } else {
     $$01$i = $10;$$02$i = $9;
    }
   } while(0);
   $eh$lpad$body11$index8Z2D = $$01$i;$eh$lpad$body11$indexZ2D = $$02$i;
   ___resumeException($eh$lpad$body11$indexZ2D|0);
   // unreachable;
  } else {
   $6 = HEAP32[$0>>2]|0; //@line 1343 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $7 = ((($0)) + 4|0); //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $8 = HEAP32[$7>>2]|0; //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $points$sroa$0$0 = $6;$points$sroa$12$0 = $8;
   label = 11;
  }
 } while(0);
 L14: do {
  if ((label|0) == 11) {
   $33 = (($points$sroa$12$0) - ($points$sroa$0$0))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $34 = $33 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $35 = $33 >> 2; //@line 182 "modules/math/wrap_BezierCurve.cpp"
   __THREW__ = 0;
   invoke_viii(3393,($L|0),($35|0),0); //@line 182 "modules/math/wrap_BezierCurve.cpp"
   $36 = __THREW__; __THREW__ = 0;
   $37 = $36&1;
   if ($37) {
    $points$sroa$0$1$ph = $points$sroa$0$0;
    label = 17;
   } else {
    $38 = ($33|0)>(0); //@line 183 "modules/math/wrap_BezierCurve.cpp"
    $39 = $points$sroa$0$0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    L17: do {
     if ($38) {
      $i$010 = 0;
      while(1) {
       $48 = (($39) + ($i$010<<3)|0); //@line 185 "modules/math/wrap_BezierCurve.cpp"
       $49 = +HEAPF32[$48>>2]; //@line 185 "modules/math/wrap_BezierCurve.cpp"
       $50 = $49; //@line 185 "modules/math/wrap_BezierCurve.cpp"
       __THREW__ = 0;
       invoke_vid(3394,($L|0),(+$50)); //@line 185 "modules/math/wrap_BezierCurve.cpp"
       $51 = __THREW__; __THREW__ = 0;
       $52 = $51&1;
       if ($52) {
        break;
       }
       $53 = $i$010 << 1; //@line 186 "modules/math/wrap_BezierCurve.cpp"
       $54 = $53 | 1; //@line 186 "modules/math/wrap_BezierCurve.cpp"
       __THREW__ = 0;
       invoke_viii(3395,($L|0),-2,($54|0)); //@line 186 "modules/math/wrap_BezierCurve.cpp"
       $55 = __THREW__; __THREW__ = 0;
       $56 = $55&1;
       if ($56) {
        break;
       }
       $57 = (((($39) + ($i$010<<3)|0)) + 4|0); //@line 187 "modules/math/wrap_BezierCurve.cpp"
       $58 = +HEAPF32[$57>>2]; //@line 187 "modules/math/wrap_BezierCurve.cpp"
       $59 = $58; //@line 187 "modules/math/wrap_BezierCurve.cpp"
       __THREW__ = 0;
       invoke_vid(3394,($L|0),(+$59)); //@line 187 "modules/math/wrap_BezierCurve.cpp"
       $60 = __THREW__; __THREW__ = 0;
       $61 = $60&1;
       if ($61) {
        break;
       }
       $62 = (($53) + 2)|0; //@line 188 "modules/math/wrap_BezierCurve.cpp"
       __THREW__ = 0;
       invoke_viii(3395,($L|0),-2,($62|0)); //@line 188 "modules/math/wrap_BezierCurve.cpp"
       $63 = __THREW__; __THREW__ = 0;
       $64 = $63&1;
       if ($64) {
        break;
       }
       $65 = (($i$010) + 1)|0; //@line 183 "modules/math/wrap_BezierCurve.cpp"
       $66 = ($65|0)<($34|0); //@line 183 "modules/math/wrap_BezierCurve.cpp"
       if ($66) {
        $i$010 = $65;
       } else {
        break L17;
       }
      }
      $42 = ___cxa_find_matching_catch_2()|0;
      $43 = tempRet0;
      $eh$lpad$body$index2Z2D = $43;$eh$lpad$body$indexZ2D = $42;$points$sroa$0$2 = $points$sroa$0$0;
      break L14;
     }
    } while(0);
    $40 = ($points$sroa$0$0|0)==(0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($40) {
     STACKTOP = sp;return 1; //@line 192 "modules/math/wrap_BezierCurve.cpp"
    }
    $41 = $points$sroa$0$0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __ZdlPv($41); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    STACKTOP = sp;return 1; //@line 192 "modules/math/wrap_BezierCurve.cpp"
   }
  }
 } while(0);
 if ((label|0) == 17) {
  $44 = ___cxa_find_matching_catch_2()|0;
  $45 = tempRet0;
  $eh$lpad$body$index2Z2D = $45;$eh$lpad$body$indexZ2D = $44;$points$sroa$0$2 = $points$sroa$0$1$ph;
 }
 $46 = ($points$sroa$0$2|0)==(0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($46) {
  $eh$lpad$body11$index8Z2D = $eh$lpad$body$index2Z2D;$eh$lpad$body11$indexZ2D = $eh$lpad$body$indexZ2D;
  ___resumeException($eh$lpad$body11$indexZ2D|0);
  // unreachable;
 }
 $47 = $points$sroa$0$2; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 __ZdlPv($47); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $eh$lpad$body11$index8Z2D = $eh$lpad$body$index2Z2D;$eh$lpad$body11$indexZ2D = $eh$lpad$body$indexZ2D;
 ___resumeException($eh$lpad$body11$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love4math27w_BezierCurve_renderSegmentEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0.0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body9$index8Z2D = 0, $eh$lpad$body9$indexZ2D = 0, $i$08 = 0, $points$sroa$0$0 = 0, $points$sroa$0$1$ph = 0, $points$sroa$0$2 = 0, $points$sroa$12$0 = 0, $vararg_buffer = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 $1 = (__ZN4love14luax_checktypeINS_4math11BezierCurveEEEPT_P9lua_StateiNS_4TypeE($L,1,26)|0); //@line 33 "modules/math/wrap_BezierCurve.cpp"
 $2 = (+_luaL_checknumber($L,2)); //@line 197 "modules/math/wrap_BezierCurve.cpp"
 $3 = (+_luaL_checknumber($L,3)); //@line 198 "modules/math/wrap_BezierCurve.cpp"
 $4 = (_luaL_optinteger($L,4,5)|0); //@line 199 "modules/math/wrap_BezierCurve.cpp"
 __THREW__ = 0;
 invoke_viiddi(3666,($0|0),($1|0),(+$2),(+$3),($4|0)); //@line 202 "modules/math/wrap_BezierCurve.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 L1: do {
  if ($6) {
   $10 = ___cxa_find_matching_catch_3(8992|0)|0;
   $11 = tempRet0;
   $12 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $13 = ($11|0)==($12|0); //@line 535 "./common/runtime.h"
   do {
    if ($13) {
     $14 = (___cxa_begin_catch(($10|0))|0); //@line 535 "./common/runtime.h"
     $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
     $16 = ((($15)) + 8|0); //@line 539 "./common/runtime.h"
     $17 = HEAP32[$16>>2]|0; //@line 539 "./common/runtime.h"
     $18 = (FUNCTION_TABLE_ii[$17 & 8191]($14)|0); //@line 539 "./common/runtime.h"
     __THREW__ = 0;
     invoke_vii(3380,($L|0),($18|0)); //@line 539 "./common/runtime.h"
     $19 = __THREW__; __THREW__ = 0;
     $20 = $19&1;
     if ($20) {
      $28 = ___cxa_find_matching_catch_2()|0;
      $29 = tempRet0;
      __THREW__ = 0;
      invoke_v(3378); //@line 540 "./common/runtime.h"
      $30 = __THREW__; __THREW__ = 0;
      $31 = $30&1;
      if (!($31)) {
       $$01$i = $29;$$02$i = $28;
       break;
      }
      $32 = ___cxa_find_matching_catch_3(0|0)|0;
      $33 = tempRet0;
      ___clang_call_terminate($32); //@line 540 "./common/runtime.h"
      // unreachable; //@line 540 "./common/runtime.h"
     }
     __THREW__ = 0;
     invoke_v(3378); //@line 540 "./common/runtime.h"
     $21 = __THREW__; __THREW__ = 0;
     $22 = $21&1;
     if ($22) {
      $points$sroa$0$1$ph = 0;
      label = 17;
      break L1;
     }
     __THREW__ = 0;
     $23 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
     $24 = __THREW__; __THREW__ = 0;
     $25 = $24&1;
     if ($25) {
      $points$sroa$0$1$ph = 0;
      label = 17;
      break L1;
     }
     __THREW__ = 0;
     HEAP32[$vararg_buffer>>2] = $23; //@line 543 "./common/runtime.h"
     (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 543 "./common/runtime.h"
     $26 = __THREW__; __THREW__ = 0;
     $27 = $26&1;
     if ($27) {
      $points$sroa$0$1$ph = 0;
      label = 17;
      break L1;
     } else {
      $points$sroa$0$0 = 0;$points$sroa$12$0 = 0;
      label = 11;
      break L1;
     }
    } else {
     $$01$i = $11;$$02$i = $10;
    }
   } while(0);
   $eh$lpad$body9$index8Z2D = $$01$i;$eh$lpad$body9$indexZ2D = $$02$i;
   ___resumeException($eh$lpad$body9$indexZ2D|0);
   // unreachable;
  } else {
   $7 = HEAP32[$0>>2]|0; //@line 1343 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $8 = ((($0)) + 4|0); //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $9 = HEAP32[$8>>2]|0; //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $points$sroa$0$0 = $7;$points$sroa$12$0 = $9;
   label = 11;
  }
 } while(0);
 L14: do {
  if ((label|0) == 11) {
   $34 = (($points$sroa$12$0) - ($points$sroa$0$0))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $35 = $34 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $36 = $34 >> 2; //@line 204 "modules/math/wrap_BezierCurve.cpp"
   __THREW__ = 0;
   invoke_viii(3393,($L|0),($36|0),0); //@line 204 "modules/math/wrap_BezierCurve.cpp"
   $37 = __THREW__; __THREW__ = 0;
   $38 = $37&1;
   if ($38) {
    $points$sroa$0$1$ph = $points$sroa$0$0;
    label = 17;
   } else {
    $39 = ($35|0)==(0); //@line 205 "modules/math/wrap_BezierCurve.cpp"
    $40 = $points$sroa$0$0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    L17: do {
     if (!($39)) {
      $i$08 = 0;
      while(1) {
       $49 = (($40) + ($i$08<<3)|0); //@line 207 "modules/math/wrap_BezierCurve.cpp"
       $50 = +HEAPF32[$49>>2]; //@line 207 "modules/math/wrap_BezierCurve.cpp"
       $51 = $50; //@line 207 "modules/math/wrap_BezierCurve.cpp"
       __THREW__ = 0;
       invoke_vid(3394,($L|0),(+$51)); //@line 207 "modules/math/wrap_BezierCurve.cpp"
       $52 = __THREW__; __THREW__ = 0;
       $53 = $52&1;
       if ($53) {
        break;
       }
       $54 = $i$08 << 1; //@line 208 "modules/math/wrap_BezierCurve.cpp"
       $55 = $54 | 1; //@line 208 "modules/math/wrap_BezierCurve.cpp"
       __THREW__ = 0;
       invoke_viii(3395,($L|0),-2,($55|0)); //@line 208 "modules/math/wrap_BezierCurve.cpp"
       $56 = __THREW__; __THREW__ = 0;
       $57 = $56&1;
       if ($57) {
        break;
       }
       $58 = (((($40) + ($i$08<<3)|0)) + 4|0); //@line 209 "modules/math/wrap_BezierCurve.cpp"
       $59 = +HEAPF32[$58>>2]; //@line 209 "modules/math/wrap_BezierCurve.cpp"
       $60 = $59; //@line 209 "modules/math/wrap_BezierCurve.cpp"
       __THREW__ = 0;
       invoke_vid(3394,($L|0),(+$60)); //@line 209 "modules/math/wrap_BezierCurve.cpp"
       $61 = __THREW__; __THREW__ = 0;
       $62 = $61&1;
       if ($62) {
        break;
       }
       $63 = (($54) + 2)|0; //@line 210 "modules/math/wrap_BezierCurve.cpp"
       __THREW__ = 0;
       invoke_viii(3395,($L|0),-2,($63|0)); //@line 210 "modules/math/wrap_BezierCurve.cpp"
       $64 = __THREW__; __THREW__ = 0;
       $65 = $64&1;
       if ($65) {
        break;
       }
       $66 = (($i$08) + 1)|0; //@line 205 "modules/math/wrap_BezierCurve.cpp"
       $67 = ($66>>>0)<($35>>>0); //@line 205 "modules/math/wrap_BezierCurve.cpp"
       if ($67) {
        $i$08 = $66;
       } else {
        break L17;
       }
      }
      $43 = ___cxa_find_matching_catch_2()|0;
      $44 = tempRet0;
      $eh$lpad$body$index2Z2D = $44;$eh$lpad$body$indexZ2D = $43;$points$sroa$0$2 = $points$sroa$0$0;
      break L14;
     }
    } while(0);
    $41 = ($points$sroa$0$0|0)==(0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($41) {
     STACKTOP = sp;return 1; //@line 214 "modules/math/wrap_BezierCurve.cpp"
    }
    $42 = $points$sroa$0$0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __ZdlPv($42); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    STACKTOP = sp;return 1; //@line 214 "modules/math/wrap_BezierCurve.cpp"
   }
  }
 } while(0);
 if ((label|0) == 17) {
  $45 = ___cxa_find_matching_catch_2()|0;
  $46 = tempRet0;
  $eh$lpad$body$index2Z2D = $46;$eh$lpad$body$indexZ2D = $45;$points$sroa$0$2 = $points$sroa$0$1$ph;
 }
 $47 = ($points$sroa$0$2|0)==(0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($47) {
  $eh$lpad$body9$index8Z2D = $eh$lpad$body$index2Z2D;$eh$lpad$body9$indexZ2D = $eh$lpad$body$indexZ2D;
  ___resumeException($eh$lpad$body9$indexZ2D|0);
  // unreachable;
 }
 $48 = $points$sroa$0$2; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 __ZdlPv($48); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $eh$lpad$body9$index8Z2D = $eh$lpad$body$index2Z2D;$eh$lpad$body9$indexZ2D = $eh$lpad$body$indexZ2D;
 ___resumeException($eh$lpad$body9$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function _luaopen_beziercurve($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 235 "modules/math/wrap_BezierCurve.cpp"
 HEAP32[$vararg_buffer>>2] = 16696; //@line 237 "modules/math/wrap_BezierCurve.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 237 "modules/math/wrap_BezierCurve.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 237 "modules/math/wrap_BezierCurve.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,26,355749,$vararg_buffer)|0); //@line 237 "modules/math/wrap_BezierCurve.cpp"
 STACKTOP = sp;return ($0|0); //@line 237 "modules/math/wrap_BezierCurve.cpp"
}
function __ZN4love4math24luax_checkcompresseddataEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_4math14CompressedDataEEEPT_P9lua_StateiNS_4TypeE($L,$idx,27)|0); //@line 32 "modules/math/wrap_CompressedData.cpp"
 return ($0|0); //@line 32 "modules/math/wrap_CompressedData.cpp"
}
function __ZN4love14luax_checktypeINS_4math14CompressedDataEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love4math26w_CompressedData_getFormatEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $fname = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $fname = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_4math14CompressedDataEEEPT_P9lua_StateiNS_4TypeE($L,1,27)|0); //@line 32 "modules/math/wrap_CompressedData.cpp"
 HEAP32[$fname>>2] = 0; //@line 39 "modules/math/wrap_CompressedData.cpp"
 $1 = (__ZNK4love4math14CompressedData9getFormatEv($0)|0); //@line 40 "modules/math/wrap_CompressedData.cpp"
 $2 = (__ZN4love4math10Compressor11getConstantENS1_6FormatERPKc($1,$fname)|0); //@line 40 "modules/math/wrap_CompressedData.cpp"
 if ($2) {
  $4 = HEAP32[$fname>>2]|0; //@line 43 "modules/math/wrap_CompressedData.cpp"
  _lua_pushstring($L,$4); //@line 43 "modules/math/wrap_CompressedData.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 45 "modules/math/wrap_CompressedData.cpp"
 } else {
  $3 = (_luaL_error($L,355940,$vararg_buffer)|0); //@line 41 "modules/math/wrap_CompressedData.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 45 "modules/math/wrap_CompressedData.cpp"
 }
 return (0)|0;
}
function _luaopen_compresseddata($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 54 "modules/math/wrap_CompressedData.cpp"
 HEAP32[$vararg_buffer>>2] = 12272; //@line 56 "modules/math/wrap_CompressedData.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 56 "modules/math/wrap_CompressedData.cpp"
 HEAP32[$vararg_ptr1>>2] = 16816; //@line 56 "modules/math/wrap_CompressedData.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 56 "modules/math/wrap_CompressedData.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 56 "modules/math/wrap_CompressedData.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,27,355972,$vararg_buffer)|0); //@line 56 "modules/math/wrap_CompressedData.cpp"
 STACKTOP = sp;return ($0|0); //@line 56 "modules/math/wrap_CompressedData.cpp"
}
function __ZN4love5mouse3sdl5MouseD2Ev($this) {
 $this = $this|0;
 var $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$01$i$i$i = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$03 = 0, $lpad$phi$index = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (16840); //@line 63 "modules/mouse/sdl/Mouse.cpp"
 $0 = ((($this)) + 8|0); //@line 139 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 133 "./common/Object.h"
 $2 = ($1|0)==(0|0); //@line 64 "modules/mouse/sdl/Mouse.cpp"
 do {
  if ($2) {
   label = 11;
  } else {
   $3 = HEAP32[$1>>2]|0; //@line 133 "./common/Object.h"
   $4 = ((($3)) + 12|0); //@line 133 "./common/Object.h"
   $5 = HEAP32[$4>>2]|0; //@line 133 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($5|0,($1|0)); //@line 133 "./common/Object.h"
   $6 = __THREW__; __THREW__ = 0;
   $7 = $6&1;
   if (!($7)) {
    HEAP32[$0>>2] = 0; //@line 134 "./common/Object.h"
    __THREW__ = 0;
    $8 = (invoke_i(3667)|0); //@line 101 "modules/mouse/sdl/Mouse.cpp"
    $9 = __THREW__; __THREW__ = 0;
    $10 = $9&1;
    if (!($10)) {
     __THREW__ = 0;
     invoke_vi(3668,($8|0)); //@line 101 "modules/mouse/sdl/Mouse.cpp"
     $11 = __THREW__; __THREW__ = 0;
     $12 = $11&1;
     if (!($12)) {
      label = 11;
      break;
     }
    }
   }
   $15 = ___cxa_find_matching_catch_3(0|0)|0;
   $16 = tempRet0;
   $$pre = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $$pre$phiZ2D = $$pre;$lpad$phi$index = $15;
  }
 } while(0);
 L7: do {
  if ((label|0) == 11) {
   $28 = ((($this)) + 12|0); //@line 67 "modules/mouse/sdl/Mouse.cpp"
   $29 = HEAP32[$28>>2]|0; //@line 911 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $30 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $31 = ($29|0)==($30|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   L9: do {
    if (!($31)) {
     $__begin$sroa$0$03 = $29;
     while(1) {
      $42 = ((($__begin$sroa$0$03)) + 20|0); //@line 68 "modules/mouse/sdl/Mouse.cpp"
      $43 = HEAP32[$42>>2]|0; //@line 68 "modules/mouse/sdl/Mouse.cpp"
      $44 = HEAP32[$43>>2]|0; //@line 68 "modules/mouse/sdl/Mouse.cpp"
      $45 = ((($44)) + 12|0); //@line 68 "modules/mouse/sdl/Mouse.cpp"
      $46 = HEAP32[$45>>2]|0; //@line 68 "modules/mouse/sdl/Mouse.cpp"
      __THREW__ = 0;
      invoke_vi($46|0,($43|0)); //@line 68 "modules/mouse/sdl/Mouse.cpp"
      $47 = __THREW__; __THREW__ = 0;
      $48 = $47&1;
      if ($48) {
       break;
      }
      $49 = ((($__begin$sroa$0$03)) + 4|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $50 = HEAP32[$49>>2]|0; //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $51 = ($50|0)==(0|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      if ($51) {
       $$0$i$i$i = $__begin$sroa$0$03;
       while(1) {
        $54 = ((($$0$i$i$i)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        $55 = HEAP32[$54>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        $56 = HEAP32[$55>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        $57 = ($56|0)==($$0$i$i$i|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        if ($57) {
         $$01$i$i$i = $55;
         break;
        } else {
         $$0$i$i$i = $55;
        }
       }
      } else {
       $$0$i$i$i$i = $50;
       while(1) {
        $52 = HEAP32[$$0$i$i$i$i>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        $53 = ($52|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        if ($53) {
         $$01$i$i$i = $$0$i$i$i$i;
         break;
        } else {
         $$0$i$i$i$i = $52;
        }
       }
      }
      $58 = ($$01$i$i$i|0)==($30|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      if ($58) {
       break L9;
      } else {
       $__begin$sroa$0$03 = $$01$i$i$i;
      }
     }
     $13 = ___cxa_find_matching_catch_3(0|0)|0;
     $14 = tempRet0;
     $$pre$phiZ2D = $30;$lpad$phi$index = $13;
     break L7;
    }
   } while(0);
   $32 = HEAP32[$30>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIN4love5mouse6Cursor12SystemCursorEPNS3_3sdl6CursorEEENS_19__map_value_compareIS5_S9_NS_4lessIS5_EELb1EEENS_9allocatorIS9_EEE7destroyEPNS_11__tree_nodeIS9_PvEE($28,$32); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $33 = HEAP32[$0>>2]|0; //@line 106 "./common/Object.h"
   $34 = ($33|0)==(0|0); //@line 106 "./common/Object.h"
   if ($34) {
    __ZN4love6ModuleD2Ev($this); //@line 69 "modules/mouse/sdl/Mouse.cpp"
    return; //@line 69 "modules/mouse/sdl/Mouse.cpp"
   }
   $35 = HEAP32[$33>>2]|0; //@line 106 "./common/Object.h"
   $36 = ((($35)) + 12|0); //@line 106 "./common/Object.h"
   $37 = HEAP32[$36>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($37|0,($33|0)); //@line 106 "./common/Object.h"
   $38 = __THREW__; __THREW__ = 0;
   $39 = $38&1;
   if ($39) {
    $40 = ___cxa_find_matching_catch_3(0|0)|0;
    $41 = tempRet0;
    ___clang_call_terminate($40); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   } else {
    __ZN4love6ModuleD2Ev($this); //@line 69 "modules/mouse/sdl/Mouse.cpp"
    return; //@line 69 "modules/mouse/sdl/Mouse.cpp"
   }
  }
 } while(0);
 $17 = ((($this)) + 12|0); //@line 783 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $18 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeIN4love5mouse6Cursor12SystemCursorEPNS3_3sdl6CursorEEENS_19__map_value_compareIS5_S9_NS_4lessIS5_EELb1EEENS_9allocatorIS9_EEE7destroyEPNS_11__tree_nodeIS9_PvEE($17,$18); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $19 = HEAP32[$0>>2]|0; //@line 106 "./common/Object.h"
 $20 = ($19|0)==(0|0); //@line 106 "./common/Object.h"
 if ($20) {
  __ZN4love6ModuleD2Ev($this); //@line 69 "modules/mouse/sdl/Mouse.cpp"
  ___clang_call_terminate($lpad$phi$index); //@line 69 "modules/mouse/sdl/Mouse.cpp"
  // unreachable; //@line 69 "modules/mouse/sdl/Mouse.cpp"
 }
 $21 = HEAP32[$19>>2]|0; //@line 106 "./common/Object.h"
 $22 = ((($21)) + 12|0); //@line 106 "./common/Object.h"
 $23 = HEAP32[$22>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($23|0,($19|0)); //@line 106 "./common/Object.h"
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $26 = ___cxa_find_matching_catch_3(0|0)|0;
  $27 = tempRet0;
  ___clang_call_terminate($26); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  __ZN4love6ModuleD2Ev($this); //@line 69 "modules/mouse/sdl/Mouse.cpp"
  ___clang_call_terminate($lpad$phi$index); //@line 69 "modules/mouse/sdl/Mouse.cpp"
  // unreachable; //@line 69 "modules/mouse/sdl/Mouse.cpp"
 }
}
function __ZNSt3__16__treeINS_12__value_typeIN4love5mouse6Cursor12SystemCursorEPNS3_3sdl6CursorEEENS_19__map_value_compareIS5_S9_NS_4lessIS5_EELb1EEENS_9allocatorIS9_EEE7destroyEPNS_11__tree_nodeIS9_PvEE($this,$__nd) {
 $this = $this|0;
 $__nd = $__nd|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__nd|0)==(0|0); //@line 1440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($0) {
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 } else {
  $1 = HEAP32[$__nd>>2]|0; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIN4love5mouse6Cursor12SystemCursorEPNS3_3sdl6CursorEEENS_19__map_value_compareIS5_S9_NS_4lessIS5_EELb1EEENS_9allocatorIS9_EEE7destroyEPNS_11__tree_nodeIS9_PvEE($this,$1); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2 = ((($__nd)) + 4|0); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3 = HEAP32[$2>>2]|0; //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIN4love5mouse6Cursor12SystemCursorEPNS3_3sdl6CursorEEENS_19__map_value_compareIS5_S9_NS_4lessIS5_EELb1EEENS_9allocatorIS9_EEE7destroyEPNS_11__tree_nodeIS9_PvEE($this,$3); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZdlPv($__nd); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
}
function __ZN4love5mouse3sdl5MouseD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5mouse3sdl5MouseD2Ev($this); //@line 63 "modules/mouse/sdl/Mouse.cpp"
 __ZdlPv($this); //@line 63 "modules/mouse/sdl/Mouse.cpp"
 return; //@line 69 "modules/mouse/sdl/Mouse.cpp"
}
function __ZNK4love5mouse5Mouse13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 9; //@line 44 "./modules/mouse/Mouse.h"
}
function __ZNK4love5mouse3sdl5Mouse7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (356031|0); //@line 54 "modules/mouse/sdl/Mouse.cpp"
}
function __ZN4love5mouse3sdl5Mouse9newCursorEPNS_5image9ImageDataEii($this,$data,$hotx,$hoty) {
 $this = $this|0;
 $data = $data|0;
 $hotx = $hotx|0;
 $hoty = $hoty|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(20)|0); //@line 73 "modules/mouse/sdl/Mouse.cpp"
 __THREW__ = 0;
 invoke_viiii(3669,($0|0),($data|0),($hotx|0),($hoty|0)); //@line 73 "modules/mouse/sdl/Mouse.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 73 "modules/mouse/sdl/Mouse.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 73 "modules/mouse/sdl/Mouse.cpp"
 }
 return (0)|0;
}
function __ZN4love5mouse3sdl5Mouse15getSystemCursorENS0_6Cursor12SystemCursorE($this,$cursortype) {
 $this = $this|0;
 $cursortype = $cursortype|0;
 var $$0$i7$i = 0, $$0$ph$lcssa$i$i$i = 0, $$0$ph6$i$i$i = 0, $$01$ph5$i$i$i = 0, $$013$i$i$i = 0, $$013$i$i$i$lcssa = 0, $$lcssa = 0, $$lcssa33 = 0, $$pre$i$i = 0, $$sink$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__nd$0$i$i = 0, $__nd$0$i$i$lcssa = 0, $__nd$0$i$i$lcssa31 = 0, $__nd$0$i$i$lcssa32 = 0, $__parent$i = 0, $__r$0$i = 0, $cursor$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__parent$i = sp;
 $0 = ((($this)) + 12|0); //@line 79 "modules/mouse/sdl/Mouse.cpp"
 $1 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = HEAP32[$1>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $3 = ($2|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if (!($3)) {
  $$0$ph6$i$i$i = $1;$$01$ph5$i$i$i = $2;
  L2: while(1) {
   $$013$i$i$i = $$01$ph5$i$i$i;
   while(1) {
    $4 = ((($$013$i$i$i)) + 16|0); //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5 = HEAP32[$4>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $6 = ($5|0)<($cursortype|0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if (!($6)) {
     $$013$i$i$i$lcssa = $$013$i$i$i;
     break;
    }
    $9 = ((($$013$i$i$i)) + 4|0); //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $10 = HEAP32[$9>>2]|0; //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $11 = ($10|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($11) {
     $$0$ph$lcssa$i$i$i = $$0$ph6$i$i$i;
     break L2;
    } else {
     $$013$i$i$i = $10;
    }
   }
   $7 = HEAP32[$$013$i$i$i$lcssa>>2]|0; //@line 2105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $8 = ($7|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($8) {
    $$0$ph$lcssa$i$i$i = $$013$i$i$i$lcssa;
    break;
   } else {
    $$0$ph6$i$i$i = $$013$i$i$i$lcssa;$$01$ph5$i$i$i = $7;
   }
  }
  $12 = ($$0$ph$lcssa$i$i$i|0)==($1|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if (!($12)) {
   $13 = ((($$0$ph$lcssa$i$i$i)) + 16|0); //@line 463 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $14 = HEAP32[$13>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   $15 = ($14|0)>($cursortype|0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   if (!($15)) {
    $16 = ((($$0$ph$lcssa$i$i$i)) + 20|0); //@line 82 "modules/mouse/sdl/Mouse.cpp"
    $17 = HEAP32[$16>>2]|0; //@line 82 "modules/mouse/sdl/Mouse.cpp"
    $cursor$0 = $17;
    STACKTOP = sp;return ($cursor$0|0); //@line 89 "modules/mouse/sdl/Mouse.cpp"
   }
  }
 }
 $18 = (__Znwj(20)|0); //@line 85 "modules/mouse/sdl/Mouse.cpp"
 __THREW__ = 0;
 invoke_vii(3670,($18|0),($cursortype|0)); //@line 85 "modules/mouse/sdl/Mouse.cpp"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $54 = ___cxa_find_matching_catch_2()|0;
  $55 = tempRet0;
  __ZdlPv($18); //@line 85 "modules/mouse/sdl/Mouse.cpp"
  ___resumeException($54|0);
  // unreachable;
 }
 $21 = HEAP32[$1>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $22 = ($21|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($22) {
   HEAP32[$__parent$i>>2] = $1; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $34 = $1; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i = $1;$42 = $34;
   label = 20;
  } else {
   $__nd$0$i$i = $21;
   while(1) {
    $23 = ((($__nd$0$i$i)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $24 = HEAP32[$23>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $25 = ($24|0)>($cursortype|0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($25) {
     $26 = HEAP32[$__nd$0$i$i>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $27 = ($26|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($27) {
      $$lcssa33 = $__nd$0$i$i;$__nd$0$i$i$lcssa32 = $__nd$0$i$i;
      label = 13;
      break;
     } else {
      $$sink$i$i = $26;
     }
    } else {
     $29 = ($24|0)<($cursortype|0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($29)) {
      $__nd$0$i$i$lcssa = $__nd$0$i$i;
      label = 19;
      break;
     }
     $30 = ((($__nd$0$i$i)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $31 = HEAP32[$30>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $32 = ($31|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($32) {
      $$lcssa = $30;$__nd$0$i$i$lcssa31 = $__nd$0$i$i;
      label = 17;
      break;
     } else {
      $$sink$i$i = $31;
     }
    }
    $__nd$0$i$i = $$sink$i$i;
   }
   if ((label|0) == 13) {
    HEAP32[$__parent$i>>2] = $__nd$0$i$i$lcssa32; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $28 = $__nd$0$i$i$lcssa32; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i = $$lcssa33;$42 = $28;
    label = 20;
    break;
   }
   else if ((label|0) == 17) {
    HEAP32[$__parent$i>>2] = $__nd$0$i$i$lcssa31; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $33 = $__nd$0$i$i$lcssa31; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i = $$lcssa;$42 = $33;
    label = 20;
    break;
   }
   else if ((label|0) == 19) {
    HEAP32[$__parent$i>>2] = $__nd$0$i$i$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $35 = $__nd$0$i$i$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $36 = ($__nd$0$i$i$lcssa|0)==(0|0); //@line 1378 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($36) {
     $$0$i7$i = $__parent$i;$42 = $35;
     label = 20;
     break;
    } else {
     $__r$0$i = $__nd$0$i$i$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 20) {
  $37 = (__Znwj(24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $38 = ((($37)) + 16|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$38>>2] = $cursortype; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $39 = ((($37)) + 20|0); //@line 1366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$39>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$37>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $40 = ((($37)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$40>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $41 = ((($37)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$41>>2] = $42; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i>>2] = $37; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $43 = HEAP32[$0>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $44 = HEAP32[$43>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $45 = ($44|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($45) {
   $49 = $37;
  } else {
   $46 = $44; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$0>>2] = $46; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i = HEAP32[$$0$i7$i>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $49 = $$pre$i$i;
  }
  $47 = ((($this)) + 16|0); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $48 = HEAP32[$47>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($48,$49); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $50 = ((($this)) + 20|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $51 = HEAP32[$50>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $52 = (($51) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$50>>2] = $52; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $__r$0$i = $37;
 }
 $53 = ((($__r$0$i)) + 20|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$53>>2] = $18; //@line 86 "modules/mouse/sdl/Mouse.cpp"
 $cursor$0 = $18;
 STACKTOP = sp;return ($cursor$0|0); //@line 89 "modules/mouse/sdl/Mouse.cpp"
}
function __ZN4love5mouse3sdl5Mouse9setCursorEPNS0_6CursorE($this,$cursor) {
 $this = $this|0;
 $cursor = $cursor|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($cursor|0)==(0|0); //@line 132 "./common/Object.h"
 if (!($0)) {
  $1 = HEAP32[$cursor>>2]|0; //@line 132 "./common/Object.h"
  $2 = ((($1)) + 8|0); //@line 132 "./common/Object.h"
  $3 = HEAP32[$2>>2]|0; //@line 132 "./common/Object.h"
  FUNCTION_TABLE_vi[$3 & 8191]($cursor); //@line 132 "./common/Object.h"
 }
 $4 = ((($this)) + 8|0); //@line 133 "./common/Object.h"
 $5 = HEAP32[$4>>2]|0; //@line 133 "./common/Object.h"
 $6 = ($5|0)==(0|0); //@line 133 "./common/Object.h"
 if (!($6)) {
  $7 = HEAP32[$5>>2]|0; //@line 133 "./common/Object.h"
  $8 = ((($7)) + 12|0); //@line 133 "./common/Object.h"
  $9 = HEAP32[$8>>2]|0; //@line 133 "./common/Object.h"
  FUNCTION_TABLE_vi[$9 & 8191]($5); //@line 133 "./common/Object.h"
 }
 HEAP32[$4>>2] = $cursor; //@line 134 "./common/Object.h"
 $10 = HEAP32[$cursor>>2]|0; //@line 95 "modules/mouse/sdl/Mouse.cpp"
 $11 = ((($10)) + 16|0); //@line 95 "modules/mouse/sdl/Mouse.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 95 "modules/mouse/sdl/Mouse.cpp"
 $13 = (FUNCTION_TABLE_ii[$12 & 8191]($cursor)|0); //@line 95 "modules/mouse/sdl/Mouse.cpp"
 _SDL_SetCursor($13); //@line 95 "modules/mouse/sdl/Mouse.cpp"
 return; //@line 96 "modules/mouse/sdl/Mouse.cpp"
}
function __ZN4love5mouse3sdl5Mouse9setCursorEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 133 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 133 "./common/Object.h"
 $2 = ($1|0)==(0|0); //@line 133 "./common/Object.h"
 if (!($2)) {
  $3 = HEAP32[$1>>2]|0; //@line 133 "./common/Object.h"
  $4 = ((($3)) + 12|0); //@line 133 "./common/Object.h"
  $5 = HEAP32[$4>>2]|0; //@line 133 "./common/Object.h"
  FUNCTION_TABLE_vi[$5 & 8191]($1); //@line 133 "./common/Object.h"
 }
 HEAP32[$0>>2] = 0; //@line 134 "./common/Object.h"
 $6 = (_SDL_GetDefaultCursor()|0); //@line 101 "modules/mouse/sdl/Mouse.cpp"
 _SDL_SetCursor($6); //@line 101 "modules/mouse/sdl/Mouse.cpp"
 return; //@line 102 "modules/mouse/sdl/Mouse.cpp"
}
function __ZNK4love5mouse3sdl5Mouse9getCursorEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 139 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 139 "./common/Object.h"
 return ($1|0); //@line 106 "modules/mouse/sdl/Mouse.cpp"
}
function __ZNK4love5mouse3sdl5Mouse9hasCursorEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GetDefaultCursor()|0); //@line 112 "modules/mouse/sdl/Mouse.cpp"
 $1 = ($0|0)!=(0|0); //@line 112 "modules/mouse/sdl/Mouse.cpp"
 return ($1|0); //@line 112 "modules/mouse/sdl/Mouse.cpp"
}
function __ZNK4love5mouse3sdl5Mouse4getXEv($this) {
 $this = $this|0;
 var $$pre = 0.0, $0 = 0, $1 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $dx = 0, $x = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $x = sp + 8|0;
 $dx = sp;
 (_SDL_GetMouseState($x,0)|0); //@line 118 "modules/mouse/sdl/Mouse.cpp"
 $0 = HEAP32[$x>>2]|0; //@line 120 "modules/mouse/sdl/Mouse.cpp"
 $1 = (+($0|0)); //@line 120 "modules/mouse/sdl/Mouse.cpp"
 HEAPF64[$dx>>3] = $1; //@line 120 "modules/mouse/sdl/Mouse.cpp"
 $2 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $3 = ($2|0)==(0|0); //@line 40 "modules/mouse/sdl/Mouse.cpp"
 if ($3) {
  $7 = $1;
  STACKTOP = sp;return (+$7); //@line 123 "modules/mouse/sdl/Mouse.cpp"
 }
 $4 = HEAP32[$2>>2]|0; //@line 41 "modules/mouse/sdl/Mouse.cpp"
 $5 = ((($4)) + 144|0); //@line 41 "modules/mouse/sdl/Mouse.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 41 "modules/mouse/sdl/Mouse.cpp"
 FUNCTION_TABLE_viii[$6 & 8191]($2,$dx,0); //@line 41 "modules/mouse/sdl/Mouse.cpp"
 $$pre = +HEAPF64[$dx>>3]; //@line 123 "modules/mouse/sdl/Mouse.cpp"
 $7 = $$pre;
 STACKTOP = sp;return (+$7); //@line 123 "modules/mouse/sdl/Mouse.cpp"
}
function __ZNK4love5mouse3sdl5Mouse4getYEv($this) {
 $this = $this|0;
 var $$pre = 0.0, $0 = 0, $1 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $dy = 0, $y = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $y = sp + 8|0;
 $dy = sp;
 (_SDL_GetMouseState(0,$y)|0); //@line 129 "modules/mouse/sdl/Mouse.cpp"
 $0 = HEAP32[$y>>2]|0; //@line 131 "modules/mouse/sdl/Mouse.cpp"
 $1 = (+($0|0)); //@line 131 "modules/mouse/sdl/Mouse.cpp"
 HEAPF64[$dy>>3] = $1; //@line 131 "modules/mouse/sdl/Mouse.cpp"
 $2 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $3 = ($2|0)==(0|0); //@line 40 "modules/mouse/sdl/Mouse.cpp"
 if ($3) {
  $7 = $1;
  STACKTOP = sp;return (+$7); //@line 134 "modules/mouse/sdl/Mouse.cpp"
 }
 $4 = HEAP32[$2>>2]|0; //@line 41 "modules/mouse/sdl/Mouse.cpp"
 $5 = ((($4)) + 144|0); //@line 41 "modules/mouse/sdl/Mouse.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 41 "modules/mouse/sdl/Mouse.cpp"
 FUNCTION_TABLE_viii[$6 & 8191]($2,0,$dy); //@line 41 "modules/mouse/sdl/Mouse.cpp"
 $$pre = +HEAPF64[$dy>>3]; //@line 134 "modules/mouse/sdl/Mouse.cpp"
 $7 = $$pre;
 STACKTOP = sp;return (+$7); //@line 134 "modules/mouse/sdl/Mouse.cpp"
}
function __ZNK4love5mouse3sdl5Mouse11getPositionERdS3_($this,$x,$y) {
 $this = $this|0;
 $x = $x|0;
 $y = $y|0;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $mx = 0, $my = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $mx = sp + 4|0;
 $my = sp;
 (_SDL_GetMouseState($mx,$my)|0); //@line 140 "modules/mouse/sdl/Mouse.cpp"
 $0 = HEAP32[$mx>>2]|0; //@line 142 "modules/mouse/sdl/Mouse.cpp"
 $1 = (+($0|0)); //@line 142 "modules/mouse/sdl/Mouse.cpp"
 HEAPF64[$x>>3] = $1; //@line 142 "modules/mouse/sdl/Mouse.cpp"
 $2 = HEAP32[$my>>2]|0; //@line 143 "modules/mouse/sdl/Mouse.cpp"
 $3 = (+($2|0)); //@line 143 "modules/mouse/sdl/Mouse.cpp"
 HEAPF64[$y>>3] = $3; //@line 143 "modules/mouse/sdl/Mouse.cpp"
 $4 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $5 = ($4|0)==(0|0); //@line 40 "modules/mouse/sdl/Mouse.cpp"
 if ($5) {
  STACKTOP = sp;return; //@line 145 "modules/mouse/sdl/Mouse.cpp"
 }
 $6 = HEAP32[$4>>2]|0; //@line 41 "modules/mouse/sdl/Mouse.cpp"
 $7 = ((($6)) + 144|0); //@line 41 "modules/mouse/sdl/Mouse.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 41 "modules/mouse/sdl/Mouse.cpp"
 FUNCTION_TABLE_viii[$8 & 8191]($4,$x,$y); //@line 41 "modules/mouse/sdl/Mouse.cpp"
 STACKTOP = sp;return; //@line 145 "modules/mouse/sdl/Mouse.cpp"
}
function __ZN4love5mouse3sdl5Mouse4setXEd($this,$x) {
 $this = $this|0;
 $x = +$x;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 166 "modules/mouse/sdl/Mouse.cpp"
 $1 = ((($0)) + 68|0); //@line 166 "modules/mouse/sdl/Mouse.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 166 "modules/mouse/sdl/Mouse.cpp"
 $3 = ((($0)) + 52|0); //@line 166 "modules/mouse/sdl/Mouse.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 166 "modules/mouse/sdl/Mouse.cpp"
 $5 = (+FUNCTION_TABLE_di[$4 & 4095]($this)); //@line 166 "modules/mouse/sdl/Mouse.cpp"
 FUNCTION_TABLE_vidd[$2 & 8191]($this,$x,$5); //@line 166 "modules/mouse/sdl/Mouse.cpp"
 return; //@line 167 "modules/mouse/sdl/Mouse.cpp"
}
function __ZN4love5mouse3sdl5Mouse4setYEd($this,$y) {
 $this = $this|0;
 $y = +$y;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 171 "modules/mouse/sdl/Mouse.cpp"
 $1 = ((($0)) + 68|0); //@line 171 "modules/mouse/sdl/Mouse.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 171 "modules/mouse/sdl/Mouse.cpp"
 $3 = ((($0)) + 48|0); //@line 171 "modules/mouse/sdl/Mouse.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 171 "modules/mouse/sdl/Mouse.cpp"
 $5 = (+FUNCTION_TABLE_di[$4 & 4095]($this)); //@line 171 "modules/mouse/sdl/Mouse.cpp"
 FUNCTION_TABLE_vidd[$2 & 8191]($this,$5,$y); //@line 171 "modules/mouse/sdl/Mouse.cpp"
 return; //@line 172 "modules/mouse/sdl/Mouse.cpp"
}
function __ZN4love5mouse3sdl5Mouse11setPositionEdd($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $$pr = 0, $$pre = 0.0, $$pre2 = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $handle$01 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 8|0;
 $1 = sp;
 HEAPF64[$0>>3] = $x;
 HEAPF64[$1>>3] = $y;
 $2 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $3 = ($2|0)==(0|0); //@line 152 "modules/mouse/sdl/Mouse.cpp"
 if ($3) {
  $13 = $x;$15 = $y;$handle$01 = 0;
 } else {
  $4 = HEAP32[$2>>2]|0; //@line 153 "modules/mouse/sdl/Mouse.cpp"
  $5 = ((($4)) + 172|0); //@line 153 "modules/mouse/sdl/Mouse.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 153 "modules/mouse/sdl/Mouse.cpp"
  $7 = (FUNCTION_TABLE_ii[$6 & 8191]($2)|0); //@line 153 "modules/mouse/sdl/Mouse.cpp"
  $$pr = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
  $8 = ($$pr|0)==(0|0); //@line 48 "modules/mouse/sdl/Mouse.cpp"
  if ($8) {
   $13 = $x;$15 = $y;$handle$01 = $7;
  } else {
   $9 = HEAP32[$$pr>>2]|0; //@line 49 "modules/mouse/sdl/Mouse.cpp"
   $10 = ((($9)) + 148|0); //@line 49 "modules/mouse/sdl/Mouse.cpp"
   $11 = HEAP32[$10>>2]|0; //@line 49 "modules/mouse/sdl/Mouse.cpp"
   FUNCTION_TABLE_viii[$11 & 8191]($$pr,$0,$1); //@line 49 "modules/mouse/sdl/Mouse.cpp"
   $$pre = +HEAPF64[$0>>3]; //@line 156 "modules/mouse/sdl/Mouse.cpp"
   $$pre2 = +HEAPF64[$1>>3]; //@line 156 "modules/mouse/sdl/Mouse.cpp"
   $13 = $$pre;$15 = $$pre2;$handle$01 = $7;
  }
 }
 $12 = (~~(($13))); //@line 156 "modules/mouse/sdl/Mouse.cpp"
 $14 = (~~(($15))); //@line 156 "modules/mouse/sdl/Mouse.cpp"
 _SDL_WarpMouseInWindow($handle$01,$12,$14); //@line 156 "modules/mouse/sdl/Mouse.cpp"
 _SDL_PumpEvents(); //@line 161 "modules/mouse/sdl/Mouse.cpp"
 STACKTOP = sp;return; //@line 162 "modules/mouse/sdl/Mouse.cpp"
}
function __ZN4love5mouse3sdl5Mouse10setVisibleEb($this,$visible) {
 $this = $this|0;
 $visible = $visible|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $visible&1; //@line 176 "modules/mouse/sdl/Mouse.cpp"
 (_SDL_ShowCursor($0)|0); //@line 176 "modules/mouse/sdl/Mouse.cpp"
 return; //@line 177 "modules/mouse/sdl/Mouse.cpp"
}
function __ZNK4love5mouse3sdl5Mouse6isDownERKNSt3__16vectorIiNS3_9allocatorIiEEEE($this,$buttons) {
 $this = $this|0;
 $buttons = $buttons|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$05 = 0, $button$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GetMouseState(0,0)|0); //@line 181 "modules/mouse/sdl/Mouse.cpp"
 $1 = HEAP32[$buttons>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($buttons)) + 4|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($1|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($4) {
  $$1 = 0;
  return ($$1|0); //@line 205 "modules/mouse/sdl/Mouse.cpp"
 } else {
  $__begin$sroa$0$05 = $1;
 }
 while(1) {
  $5 = HEAP32[$__begin$sroa$0$05>>2]|0; //@line 183 "modules/mouse/sdl/Mouse.cpp"
  $6 = ($5|0)<(1); //@line 185 "modules/mouse/sdl/Mouse.cpp"
  if (!($6)) {
   switch ($5|0) {
   case 2:  {
    $button$0 = 3;
    break;
   }
   case 3:  {
    $button$0 = 2;
    break;
   }
   default: {
    $button$0 = $5;
   }
   }
   $7 = (($button$0) + -1)|0; //@line 200 "modules/mouse/sdl/Mouse.cpp"
   $8 = 1 << $7; //@line 200 "modules/mouse/sdl/Mouse.cpp"
   $9 = $8 & $0; //@line 200 "modules/mouse/sdl/Mouse.cpp"
   $10 = ($9|0)==(0); //@line 200 "modules/mouse/sdl/Mouse.cpp"
   if (!($10)) {
    $$1 = 1;
    label = 8;
    break;
   }
  }
  $11 = ((($__begin$sroa$0$05)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $12 = ($11|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($12) {
   $$1 = 0;
   label = 8;
   break;
  } else {
   $__begin$sroa$0$05 = $11;
  }
 }
 if ((label|0) == 8) {
  return ($$1|0); //@line 205 "modules/mouse/sdl/Mouse.cpp"
 }
 return (0)|0;
}
function __ZNK4love5mouse3sdl5Mouse9isVisibleEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_ShowCursor(-1)|0); //@line 209 "modules/mouse/sdl/Mouse.cpp"
 $1 = ($0|0)==(1); //@line 209 "modules/mouse/sdl/Mouse.cpp"
 return ($1|0); //@line 209 "modules/mouse/sdl/Mouse.cpp"
}
function __ZN4love5mouse3sdl5Mouse10setGrabbedEb($this,$grab) {
 $this = $this|0;
 $grab = $grab|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $1 = ($0|0)==(0|0); //@line 215 "modules/mouse/sdl/Mouse.cpp"
 if ($1) {
  return; //@line 217 "modules/mouse/sdl/Mouse.cpp"
 }
 $2 = HEAP32[$0>>2]|0; //@line 216 "modules/mouse/sdl/Mouse.cpp"
 $3 = ((($2)) + 132|0); //@line 216 "modules/mouse/sdl/Mouse.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 216 "modules/mouse/sdl/Mouse.cpp"
 FUNCTION_TABLE_vii[$4 & 8191]($0,$grab); //@line 216 "modules/mouse/sdl/Mouse.cpp"
 return; //@line 217 "modules/mouse/sdl/Mouse.cpp"
}
function __ZNK4love5mouse3sdl5Mouse9isGrabbedEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $1 = ($0|0)==(0|0); //@line 222 "modules/mouse/sdl/Mouse.cpp"
 if ($1) {
  $$0 = 0;
  return ($$0|0); //@line 226 "modules/mouse/sdl/Mouse.cpp"
 }
 $2 = HEAP32[$0>>2]|0; //@line 223 "modules/mouse/sdl/Mouse.cpp"
 $3 = ((($2)) + 136|0); //@line 223 "modules/mouse/sdl/Mouse.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 223 "modules/mouse/sdl/Mouse.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($0)|0); //@line 223 "modules/mouse/sdl/Mouse.cpp"
 $$0 = $5;
 return ($$0|0); //@line 226 "modules/mouse/sdl/Mouse.cpp"
}
function __ZN4love5mouse3sdl5Mouse15setRelativeModeEb($this,$relative) {
 $this = $this|0;
 $relative = $relative|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $relative&1; //@line 230 "modules/mouse/sdl/Mouse.cpp"
 $1 = (_SDL_SetRelativeMouseMode($0)|0); //@line 230 "modules/mouse/sdl/Mouse.cpp"
 $2 = ($1|0)==(0); //@line 230 "modules/mouse/sdl/Mouse.cpp"
 return ($2|0); //@line 230 "modules/mouse/sdl/Mouse.cpp"
}
function __ZNK4love5mouse3sdl5Mouse15getRelativeModeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GetRelativeMouseMode()|0); //@line 235 "modules/mouse/sdl/Mouse.cpp"
 $1 = ($0|0)!=(0); //@line 235 "modules/mouse/sdl/Mouse.cpp"
 return ($1|0); //@line 235 "modules/mouse/sdl/Mouse.cpp"
}
function __ZN4love5mouse3sdl5MouseC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./common/Module.h"
 HEAP32[$this>>2] = (16840); //@line 59 "modules/mouse/sdl/Mouse.cpp"
 $0 = ((($this)) + 8|0); //@line 93 "./common/Object.h"
 HEAP32[$0>>2] = 0; //@line 93 "./common/Object.h"
 $1 = ((($this)) + 12|0); //@line 57 "modules/mouse/sdl/Mouse.cpp"
 $2 = ((($this)) + 16|0); //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$2>>2] = 0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $3 = ((($this)) + 20|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$3>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $4 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$1>>2] = $4; //@line 1130 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 return; //@line 60 "modules/mouse/sdl/Mouse.cpp"
}
function __ZN4love5mouse6CursorD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 30 "modules/mouse/Cursor.cpp"
}
function __GLOBAL__sub_I_Cursor_cpp() {
 var $$02$i$i$1$i = 0, $$02$i$i$i = 0, $$lcssa$i$1$i = 0, $$lcssa$i$i = 0, $$lcssa10$i$1$i = 0, $$lcssa10$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i$i$1$i = 0, $hash$0$lcssa$i$i$i = 0;
 var $hash$01$i$i$1$i = 0, $hash$01$i$i$i = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0; //@line 68 "modules/mouse/Cursor.cpp"
 $vararg_buffer = sp; //@line 68 "modules/mouse/Cursor.cpp"
 __ZN4love9StringMapINS_5mouse6Cursor12SystemCursorELj12EEC2EPKNS4_5EntryEj(630592,16940,96); //@line 68 "modules/mouse/Cursor.cpp"
 HEAP8[(630936)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(630948)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(630960)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(630972)>>0] = 0; //@line 155 "./common/StringMap.h"
 $0 = (630976); //@line 44 "./common/StringMap.h"
 $1 = $0; //@line 44 "./common/StringMap.h"
 HEAP32[$1>>2] = 0; //@line 44 "./common/StringMap.h"
 $2 = (($0) + 4)|0; //@line 44 "./common/StringMap.h"
 $3 = $2; //@line 44 "./common/StringMap.h"
 HEAP32[$3>>2] = 0; //@line 44 "./common/StringMap.h"
 $4 = HEAP32[4259]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(17040)>>2]|0; //@line 49 "./common/StringMap.h"
 $6 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
 $7 = ($6<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($7) {
  $hash$0$lcssa$i$i$i = 5381;
 } else {
  $$02$i$i$i = $4;$9 = $6;$hash$01$i$i$i = 5381;
  while(1) {
   $8 = $9 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $10 = ((($$02$i$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $11 = ($hash$01$i$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $12 = (($11) + ($8))|0; //@line 143 "./common/StringMap.h"
   $13 = HEAP8[$10>>0]|0; //@line 142 "./common/StringMap.h"
   $14 = ($13<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($14) {
    $hash$0$lcssa$i$i$i = $12;
    break;
   } else {
    $$02$i$i$i = $10;$9 = $13;$hash$01$i$i$i = $12;
   }
  }
 }
 $15 = $hash$0$lcssa$i$i$i & 3; //@line 112 "./common/StringMap.h"
 $16 = (((630928 + (($15*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $17 = HEAP8[$16>>0]|0; //@line 114 "./common/StringMap.h"
 $18 = ($17<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($18) {
  $$lcssa$i$i = $15;$$lcssa10$i$i = $16;
  label = 5;
 } else {
  $19 = (($hash$0$lcssa$i$i$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $20 = $19 & 3; //@line 112 "./common/StringMap.h"
  $21 = (((630928 + (($20*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $22 = HEAP8[$21>>0]|0; //@line 114 "./common/StringMap.h"
  $23 = ($22<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($23) {
   $$lcssa$i$i = $20;$$lcssa10$i$i = $21;
   label = 5;
  } else {
   $28 = (($hash$0$lcssa$i$i$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $29 = $28 & 3; //@line 112 "./common/StringMap.h"
   $30 = (((630928 + (($29*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $31 = HEAP8[$30>>0]|0; //@line 114 "./common/StringMap.h"
   $32 = ($31<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($32) {
    $$lcssa$i$i = $29;$$lcssa10$i$i = $30;
    label = 5;
   } else {
    $33 = (($hash$0$lcssa$i$i$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $34 = $33 & 3; //@line 112 "./common/StringMap.h"
    $35 = (((630928 + (($34*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $36 = HEAP8[$35>>0]|0; //@line 114 "./common/StringMap.h"
    $37 = ($36<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($37) {
     $$lcssa$i$i = $34;$$lcssa10$i$i = $35;
     label = 5;
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10$i$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $24 = (630928 + (($$lcssa$i$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$24>>2] = $4; //@line 118 "./common/StringMap.h"
  $25 = (((630928 + (($$lcssa$i$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$25>>2] = $5; //@line 119 "./common/StringMap.h"
 }
 $26 = ($5>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($26) {
  HEAP32[$vararg_buffer>>2] = $4; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $5; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
 } else {
  $27 = (630976 + ($5<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$27>>2] = $4; //@line 132 "./common/StringMap.h"
 }
 $38 = HEAP32[(17044)>>2]|0; //@line 49 "./common/StringMap.h"
 $39 = HEAP32[(17048)>>2]|0; //@line 49 "./common/StringMap.h"
 $40 = HEAP8[$38>>0]|0; //@line 142 "./common/StringMap.h"
 $41 = ($40<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($41) {
  $hash$0$lcssa$i$i$1$i = 5381;
 } else {
  $$02$i$i$1$i = $38;$43 = $40;$hash$01$i$i$1$i = 5381;
  while(1) {
   $42 = $43 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $44 = ((($$02$i$i$1$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $45 = ($hash$01$i$i$1$i*33)|0; //@line 143 "./common/StringMap.h"
   $46 = (($45) + ($42))|0; //@line 143 "./common/StringMap.h"
   $47 = HEAP8[$44>>0]|0; //@line 142 "./common/StringMap.h"
   $48 = ($47<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($48) {
    $hash$0$lcssa$i$i$1$i = $46;
    break;
   } else {
    $$02$i$i$1$i = $44;$43 = $47;$hash$01$i$i$1$i = $46;
   }
  }
 }
 $49 = $hash$0$lcssa$i$i$1$i & 3; //@line 112 "./common/StringMap.h"
 $50 = (((630928 + (($49*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $51 = HEAP8[$50>>0]|0; //@line 114 "./common/StringMap.h"
 $52 = ($51<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($52) {
  $$lcssa$i$1$i = $49;$$lcssa10$i$1$i = $50;
  label = 17;
 } else {
  $53 = (($hash$0$lcssa$i$i$1$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $54 = $53 & 3; //@line 112 "./common/StringMap.h"
  $55 = (((630928 + (($54*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $56 = HEAP8[$55>>0]|0; //@line 114 "./common/StringMap.h"
  $57 = ($56<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($57) {
   $$lcssa$i$1$i = $54;$$lcssa10$i$1$i = $55;
   label = 17;
  } else {
   $58 = (($hash$0$lcssa$i$i$1$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $59 = $58 & 3; //@line 112 "./common/StringMap.h"
   $60 = (((630928 + (($59*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $61 = HEAP8[$60>>0]|0; //@line 114 "./common/StringMap.h"
   $62 = ($61<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($62) {
    $$lcssa$i$1$i = $59;$$lcssa10$i$1$i = $60;
    label = 17;
   } else {
    $63 = (($hash$0$lcssa$i$i$1$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $64 = $63 & 3; //@line 112 "./common/StringMap.h"
    $65 = (((630928 + (($64*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $66 = HEAP8[$65>>0]|0; //@line 114 "./common/StringMap.h"
    $67 = ($66<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($67) {
     $$lcssa$i$1$i = $64;$$lcssa10$i$1$i = $65;
     label = 17;
    }
   }
  }
 }
 if ((label|0) == 17) {
  HEAP8[$$lcssa10$i$1$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $68 = (630928 + (($$lcssa$i$1$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$68>>2] = $38; //@line 118 "./common/StringMap.h"
  $69 = (((630928 + (($$lcssa$i$1$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$69>>2] = $39; //@line 119 "./common/StringMap.h"
 }
 $70 = ($39>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($70) {
  HEAP32[$vararg_buffer2>>2] = $38; //@line 128 "./common/StringMap.h"
  $vararg_ptr5 = ((($vararg_buffer2)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr5>>2] = $39; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer2)|0); //@line 128 "./common/StringMap.h"
  STACKTOP = sp;return;
 } else {
  $71 = (630976 + ($39<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$71>>2] = $38; //@line 132 "./common/StringMap.h"
  STACKTOP = sp;return;
 }
}
function __ZN4love9StringMapINS_5mouse6Cursor12SystemCursorELj12EEC2EPKNS4_5EntryEj($this,$entries,$num) {
 $this = $this|0;
 $entries = $entries|0;
 $num = $num|0;
 var $$02$i$i = 0, $$lcssa26 = 0, $$lcssa28 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$08$i = 0, $i1$05 = 0, $scevgep = 0;
 var $vararg_buffer = 0, $vararg_ptr1 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 8|0); //@line 155 "./common/StringMap.h"
 HEAP8[$0>>0] = 0; //@line 155 "./common/StringMap.h"
 $1 = ((($this)) + 20|0); //@line 155 "./common/StringMap.h"
 HEAP8[$1>>0] = 0; //@line 155 "./common/StringMap.h"
 $2 = ((($this)) + 32|0); //@line 155 "./common/StringMap.h"
 HEAP8[$2>>0] = 0; //@line 155 "./common/StringMap.h"
 $3 = ((($this)) + 44|0); //@line 155 "./common/StringMap.h"
 HEAP8[$3>>0] = 0; //@line 155 "./common/StringMap.h"
 $4 = ((($this)) + 56|0); //@line 155 "./common/StringMap.h"
 HEAP8[$4>>0] = 0; //@line 155 "./common/StringMap.h"
 $5 = ((($this)) + 68|0); //@line 155 "./common/StringMap.h"
 HEAP8[$5>>0] = 0; //@line 155 "./common/StringMap.h"
 $6 = ((($this)) + 80|0); //@line 155 "./common/StringMap.h"
 HEAP8[$6>>0] = 0; //@line 155 "./common/StringMap.h"
 $7 = ((($this)) + 92|0); //@line 155 "./common/StringMap.h"
 HEAP8[$7>>0] = 0; //@line 155 "./common/StringMap.h"
 $8 = ((($this)) + 104|0); //@line 155 "./common/StringMap.h"
 HEAP8[$8>>0] = 0; //@line 155 "./common/StringMap.h"
 $9 = ((($this)) + 116|0); //@line 155 "./common/StringMap.h"
 HEAP8[$9>>0] = 0; //@line 155 "./common/StringMap.h"
 $10 = ((($this)) + 128|0); //@line 155 "./common/StringMap.h"
 HEAP8[$10>>0] = 0; //@line 155 "./common/StringMap.h"
 $11 = ((($this)) + 140|0); //@line 155 "./common/StringMap.h"
 HEAP8[$11>>0] = 0; //@line 155 "./common/StringMap.h"
 $12 = ((($this)) + 152|0); //@line 155 "./common/StringMap.h"
 HEAP8[$12>>0] = 0; //@line 155 "./common/StringMap.h"
 $13 = ((($this)) + 164|0); //@line 155 "./common/StringMap.h"
 HEAP8[$13>>0] = 0; //@line 155 "./common/StringMap.h"
 $14 = ((($this)) + 176|0); //@line 155 "./common/StringMap.h"
 HEAP8[$14>>0] = 0; //@line 155 "./common/StringMap.h"
 $15 = ((($this)) + 188|0); //@line 155 "./common/StringMap.h"
 HEAP8[$15>>0] = 0; //@line 155 "./common/StringMap.h"
 $16 = ((($this)) + 200|0); //@line 155 "./common/StringMap.h"
 HEAP8[$16>>0] = 0; //@line 155 "./common/StringMap.h"
 $17 = ((($this)) + 212|0); //@line 155 "./common/StringMap.h"
 HEAP8[$17>>0] = 0; //@line 155 "./common/StringMap.h"
 $18 = ((($this)) + 224|0); //@line 155 "./common/StringMap.h"
 HEAP8[$18>>0] = 0; //@line 155 "./common/StringMap.h"
 $19 = ((($this)) + 236|0); //@line 155 "./common/StringMap.h"
 HEAP8[$19>>0] = 0; //@line 155 "./common/StringMap.h"
 $20 = ((($this)) + 248|0); //@line 155 "./common/StringMap.h"
 HEAP8[$20>>0] = 0; //@line 155 "./common/StringMap.h"
 $21 = ((($this)) + 260|0); //@line 155 "./common/StringMap.h"
 HEAP8[$21>>0] = 0; //@line 155 "./common/StringMap.h"
 $22 = ((($this)) + 272|0); //@line 155 "./common/StringMap.h"
 HEAP8[$22>>0] = 0; //@line 155 "./common/StringMap.h"
 $23 = ((($this)) + 284|0); //@line 155 "./common/StringMap.h"
 HEAP8[$23>>0] = 0; //@line 155 "./common/StringMap.h"
 $scevgep = ((($this)) + 288|0); //@line 44 "./common/StringMap.h"
 dest=$scevgep; stop=dest+48|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 44 "./common/StringMap.h"
 $24 = $num >>> 3; //@line 46 "./common/StringMap.h"
 $25 = ($24|0)==(0); //@line 48 "./common/StringMap.h"
 if ($25) {
  STACKTOP = sp;return; //@line 50 "./common/StringMap.h"
 } else {
  $i1$05 = 0;
 }
 while(1) {
  $26 = (($entries) + ($i1$05<<3)|0); //@line 49 "./common/StringMap.h"
  $27 = HEAP32[$26>>2]|0; //@line 49 "./common/StringMap.h"
  $28 = (((($entries) + ($i1$05<<3)|0)) + 4|0); //@line 49 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 49 "./common/StringMap.h"
  $30 = HEAP8[$27>>0]|0; //@line 142 "./common/StringMap.h"
  $31 = ($30<<24>>24)==(0); //@line 142 "./common/StringMap.h"
  if ($31) {
   $hash$0$lcssa$i$i = 5381;
  } else {
   $$02$i$i = $27;$33 = $30;$hash$01$i$i = 5381;
   while(1) {
    $32 = $33 << 24 >> 24; //@line 142 "./common/StringMap.h"
    $34 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
    $35 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
    $36 = (($35) + ($32))|0; //@line 143 "./common/StringMap.h"
    $37 = HEAP8[$34>>0]|0; //@line 142 "./common/StringMap.h"
    $38 = ($37<<24>>24)==(0); //@line 142 "./common/StringMap.h"
    if ($38) {
     $hash$0$lcssa$i$i = $36;
     break;
    } else {
     $$02$i$i = $34;$33 = $37;$hash$01$i$i = $36;
    }
   }
  }
  $i$08$i = 0;
  while(1) {
   $41 = (($i$08$i) + ($hash$0$lcssa$i$i))|0; //@line 112 "./common/StringMap.h"
   $42 = (($41>>>0) % 24)&-1; //@line 112 "./common/StringMap.h"
   $43 = (((($this) + (($42*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $44 = HEAP8[$43>>0]|0; //@line 114 "./common/StringMap.h"
   $45 = ($44<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   $40 = (($i$08$i) + 1)|0; //@line 110 "./common/StringMap.h"
   if ($45) {
    $$lcssa26 = $42;$$lcssa28 = $43;
    label = 8;
    break;
   }
   $39 = ($40>>>0)<(24); //@line 110 "./common/StringMap.h"
   if ($39) {
    $i$08$i = $40;
   } else {
    break;
   }
  }
  if ((label|0) == 8) {
   label = 0;
   HEAP8[$$lcssa28>>0] = 1; //@line 117 "./common/StringMap.h"
   $46 = (($this) + (($$lcssa26*12)|0)|0); //@line 118 "./common/StringMap.h"
   HEAP32[$46>>2] = $27; //@line 118 "./common/StringMap.h"
   $47 = (((($this) + (($$lcssa26*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
   HEAP32[$47>>2] = $29; //@line 119 "./common/StringMap.h"
  }
  $48 = ($29>>>0)>(11); //@line 126 "./common/StringMap.h"
  if ($48) {
   HEAP32[$vararg_buffer>>2] = $27; //@line 128 "./common/StringMap.h"
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
   HEAP32[$vararg_ptr1>>2] = $29; //@line 128 "./common/StringMap.h"
   (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  } else {
   $49 = (((($this)) + 288|0) + ($29<<2)|0); //@line 132 "./common/StringMap.h"
   HEAP32[$49>>2] = $27; //@line 132 "./common/StringMap.h"
  }
  $50 = (($i1$05) + 1)|0; //@line 48 "./common/StringMap.h"
  $exitcond = ($50|0)==($24|0); //@line 48 "./common/StringMap.h"
  if ($exitcond) {
   break;
  } else {
   $i1$05 = $50;
  }
 }
 STACKTOP = sp;return; //@line 50 "./common/StringMap.h"
}
function __ZN4love5mouse6Cursor11getConstantEPKcRNS1_12SystemCursorE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 24)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((630592 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (630592 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(24); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((630592 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 34 "modules/mouse/Cursor.cpp"
  return ($$$2$i|0); //@line 34 "modules/mouse/Cursor.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 34 "modules/mouse/Cursor.cpp"
  return ($$$2$i|0); //@line 34 "modules/mouse/Cursor.cpp"
 }
 return (0)|0;
}
function __ZN4love5mouse6Cursor11getConstantENS1_12SystemCursorERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(11); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (630880 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 39 "modules/mouse/Cursor.cpp"
}
function __ZN4love5mouse6Cursor11getConstantENS1_10CursorTypeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(1); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (630976 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 49 "modules/mouse/Cursor.cpp"
}
function __ZN4love5mouse16luax_checkcursorEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5mouse6CursorEEEPT_P9lua_StateiNS_4TypeE($L,$idx,31)|0); //@line 31 "modules/mouse/wrap_Cursor.cpp"
 return ($0|0); //@line 31 "modules/mouse/wrap_Cursor.cpp"
}
function __ZN4love14luax_checktypeINS_5mouse6CursorEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love5mouse16w_Cursor_getTypeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$pr = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $typestr = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $typestr = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_5mouse6CursorEEEPT_P9lua_StateiNS_4TypeE($L,1,31)|0); //@line 31 "modules/mouse/wrap_Cursor.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 38 "modules/mouse/wrap_Cursor.cpp"
 $2 = ((($1)) + 20|0); //@line 38 "modules/mouse/wrap_Cursor.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 38 "modules/mouse/wrap_Cursor.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 38 "modules/mouse/wrap_Cursor.cpp"
 HEAP32[$typestr>>2] = 0; //@line 39 "modules/mouse/wrap_Cursor.cpp"
 switch ($4|0) {
 case 1:  {
  (__ZN4love5mouse6Cursor11getConstantENS1_10CursorTypeERPKc(1,$typestr)|0); //@line 42 "modules/mouse/wrap_Cursor.cpp"
  label = 4;
  break;
 }
 case 0:  {
  $5 = HEAP32[$0>>2]|0; //@line 45 "modules/mouse/wrap_Cursor.cpp"
  $6 = ((($5)) + 24|0); //@line 45 "modules/mouse/wrap_Cursor.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 45 "modules/mouse/wrap_Cursor.cpp"
  $8 = (FUNCTION_TABLE_ii[$7 & 8191]($0)|0); //@line 45 "modules/mouse/wrap_Cursor.cpp"
  (__ZN4love5mouse6Cursor11getConstantENS1_12SystemCursorERPKc($8,$typestr)|0); //@line 46 "modules/mouse/wrap_Cursor.cpp"
  label = 4;
  break;
 }
 default: {
 }
 }
 if ((label|0) == 4) {
  $$pr = HEAP32[$typestr>>2]|0; //@line 49 "modules/mouse/wrap_Cursor.cpp"
  $9 = ($$pr|0)==(0|0); //@line 49 "modules/mouse/wrap_Cursor.cpp"
  if (!($9)) {
   _lua_pushstring($L,$$pr); //@line 52 "modules/mouse/wrap_Cursor.cpp"
   $$0 = 1;
   STACKTOP = sp;return ($$0|0); //@line 54 "modules/mouse/wrap_Cursor.cpp"
  }
 }
 $10 = (_luaL_error($L,356141,$vararg_buffer)|0); //@line 50 "modules/mouse/wrap_Cursor.cpp"
 $$0 = $10;
 STACKTOP = sp;return ($$0|0); //@line 54 "modules/mouse/wrap_Cursor.cpp"
}
function _luaopen_cursor($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 62 "modules/mouse/wrap_Cursor.cpp"
 HEAP32[$vararg_buffer>>2] = 17052; //@line 64 "modules/mouse/wrap_Cursor.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 64 "modules/mouse/wrap_Cursor.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 64 "modules/mouse/wrap_Cursor.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,31,356162,$vararg_buffer)|0); //@line 64 "modules/mouse/wrap_Cursor.cpp"
 STACKTOP = sp;return ($0|0); //@line 64 "modules/mouse/wrap_Cursor.cpp"
}
function __ZN4love7physics5box2d5JointD2Ev($this) {
 $this = $this|0;
 var $$pr$pre = 0, $$pr3 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (17076); //@line 63 "modules/physics/box2d/Joint.cpp"
 $0 = ((($this)) + 12|0); //@line 64 "modules/physics/box2d/Joint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 64 "modules/physics/box2d/Joint.cpp"
 $2 = ($1|0)==(0|0); //@line 64 "modules/physics/box2d/Joint.cpp"
 do {
  if (!($2)) {
   $3 = HEAP32[$1>>2]|0; //@line 65 "modules/physics/box2d/Joint.cpp"
   $4 = ($3|0)==(0|0); //@line 65 "modules/physics/box2d/Joint.cpp"
   if ($4) {
    $$pr3 = $1;
   } else {
    $5 = HEAP32[$3>>2]|0; //@line 65 "modules/physics/box2d/Joint.cpp"
    $6 = ((($5)) + 4|0); //@line 65 "modules/physics/box2d/Joint.cpp"
    $7 = HEAP32[$6>>2]|0; //@line 65 "modules/physics/box2d/Joint.cpp"
    FUNCTION_TABLE_vi[$7 & 8191]($3); //@line 65 "modules/physics/box2d/Joint.cpp"
    $$pr$pre = HEAP32[$0>>2]|0; //@line 66 "modules/physics/box2d/Joint.cpp"
    $8 = ($$pr$pre|0)==(0|0); //@line 66 "modules/physics/box2d/Joint.cpp"
    if ($8) {
     break;
    } else {
     $$pr3 = $$pr$pre;
    }
   }
   __ZdlPv($$pr3); //@line 66 "modules/physics/box2d/Joint.cpp"
  }
 } while(0);
 __ZN4love7physics5JointD2Ev($this); //@line 67 "modules/physics/box2d/Joint.cpp"
 return; //@line 67 "modules/physics/box2d/Joint.cpp"
}
function __ZN4love7physics5box2d5JointD0Ev($this) {
 $this = $this|0;
 var $$pr$pre$i = 0, $$pr3$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (17076); //@line 63 "modules/physics/box2d/Joint.cpp"
 $0 = ((($this)) + 12|0); //@line 64 "modules/physics/box2d/Joint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 64 "modules/physics/box2d/Joint.cpp"
 $2 = ($1|0)==(0|0); //@line 64 "modules/physics/box2d/Joint.cpp"
 do {
  if (!($2)) {
   $3 = HEAP32[$1>>2]|0; //@line 65 "modules/physics/box2d/Joint.cpp"
   $4 = ($3|0)==(0|0); //@line 65 "modules/physics/box2d/Joint.cpp"
   if ($4) {
    $$pr3$i = $1;
   } else {
    $5 = HEAP32[$3>>2]|0; //@line 65 "modules/physics/box2d/Joint.cpp"
    $6 = ((($5)) + 4|0); //@line 65 "modules/physics/box2d/Joint.cpp"
    $7 = HEAP32[$6>>2]|0; //@line 65 "modules/physics/box2d/Joint.cpp"
    FUNCTION_TABLE_vi[$7 & 8191]($3); //@line 65 "modules/physics/box2d/Joint.cpp"
    $$pr$pre$i = HEAP32[$0>>2]|0; //@line 66 "modules/physics/box2d/Joint.cpp"
    $8 = ($$pr$pre$i|0)==(0|0); //@line 66 "modules/physics/box2d/Joint.cpp"
    if ($8) {
     break;
    } else {
     $$pr3$i = $$pr$pre$i;
    }
   }
   __ZdlPv($$pr3$i); //@line 66 "modules/physics/box2d/Joint.cpp"
  }
 } while(0);
 __ZN4love7physics5JointD2Ev($this); //@line 67 "modules/physics/box2d/Joint.cpp"
 __ZdlPv($this); //@line 63 "modules/physics/box2d/Joint.cpp"
 return; //@line 67 "modules/physics/box2d/Joint.cpp"
}
function __ZN4love7physics5box2d5JointC2EPNS1_4BodyE($this,$body1) {
 $this = $this|0;
 $body1 = $body1|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./modules/physics/Joint.h"
 HEAP32[$this>>2] = (17076); //@line 47 "modules/physics/box2d/Joint.cpp"
 $0 = ((($this)) + 8|0); //@line 43 "modules/physics/box2d/Joint.cpp"
 $1 = ((($body1)) + 12|0); //@line 43 "modules/physics/box2d/Joint.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 43 "modules/physics/box2d/Joint.cpp"
 HEAP32[$0>>2] = $2; //@line 43 "modules/physics/box2d/Joint.cpp"
 $3 = ((($this)) + 12|0); //@line 44 "modules/physics/box2d/Joint.cpp"
 HEAP32[$3>>2] = 0; //@line 44 "modules/physics/box2d/Joint.cpp"
 $4 = ((($this)) + 16|0); //@line 45 "modules/physics/box2d/Joint.cpp"
 HEAP32[$4>>2] = $body1; //@line 45 "modules/physics/box2d/Joint.cpp"
 $5 = ((($this)) + 20|0); //@line 46 "modules/physics/box2d/Joint.cpp"
 HEAP32[$5>>2] = 0; //@line 46 "modules/physics/box2d/Joint.cpp"
 __THREW__ = 0;
 $6 = (invoke_ii(3348,4)|0); //@line 48 "modules/physics/box2d/Joint.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  __ZN4love7physics5JointD2Ev($this); //@line 50 "modules/physics/box2d/Joint.cpp"
  ___resumeException($9|0);
  // unreachable;
 } else {
  HEAP32[$3>>2] = $6; //@line 48 "modules/physics/box2d/Joint.cpp"
  HEAP32[$6>>2] = 0; //@line 49 "modules/physics/box2d/Joint.cpp"
  return; //@line 50 "modules/physics/box2d/Joint.cpp"
 }
}
function __ZN4love7physics5box2d5JointC2EPNS1_4BodyES4_($this,$body1,$body2) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./modules/physics/Joint.h"
 HEAP32[$this>>2] = (17076); //@line 57 "modules/physics/box2d/Joint.cpp"
 $0 = ((($this)) + 8|0); //@line 53 "modules/physics/box2d/Joint.cpp"
 $1 = ((($body1)) + 12|0); //@line 53 "modules/physics/box2d/Joint.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 53 "modules/physics/box2d/Joint.cpp"
 HEAP32[$0>>2] = $2; //@line 53 "modules/physics/box2d/Joint.cpp"
 $3 = ((($this)) + 12|0); //@line 54 "modules/physics/box2d/Joint.cpp"
 HEAP32[$3>>2] = 0; //@line 54 "modules/physics/box2d/Joint.cpp"
 $4 = ((($this)) + 16|0); //@line 55 "modules/physics/box2d/Joint.cpp"
 HEAP32[$4>>2] = $body1; //@line 55 "modules/physics/box2d/Joint.cpp"
 $5 = ((($this)) + 20|0); //@line 56 "modules/physics/box2d/Joint.cpp"
 HEAP32[$5>>2] = $body2; //@line 56 "modules/physics/box2d/Joint.cpp"
 __THREW__ = 0;
 $6 = (invoke_ii(3348,4)|0); //@line 58 "modules/physics/box2d/Joint.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  __ZN4love7physics5JointD2Ev($this); //@line 60 "modules/physics/box2d/Joint.cpp"
  ___resumeException($9|0);
  // unreachable;
 } else {
  HEAP32[$3>>2] = $6; //@line 58 "modules/physics/box2d/Joint.cpp"
  HEAP32[$6>>2] = 0; //@line 59 "modules/physics/box2d/Joint.cpp"
  return; //@line 60 "modules/physics/box2d/Joint.cpp"
 }
}
function __ZNK4love7physics5box2d5Joint7getTypeEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 71 "modules/physics/box2d/Joint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 71 "modules/physics/box2d/Joint.cpp"
 $2 = ((($1)) + 4|0); //@line 188 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $3 = HEAP32[$2>>2]|0; //@line 188 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 do {
  switch ($3|0) {
  case 1:  {
   $$0 = 2;
   break;
  }
  case 2:  {
   $$0 = 3;
   break;
  }
  case 3:  {
   $$0 = 1;
   break;
  }
  case 4:  {
   $$0 = 5;
   break;
  }
  case 5:  {
   $$0 = 4;
   break;
  }
  case 6:  {
   $$0 = $3;
   break;
  }
  case 9:  {
   $$0 = 7;
   break;
  }
  case 8:  {
   $$0 = $3;
   break;
  }
  case 7:  {
   $$0 = 9;
   break;
  }
  case 10:  {
   $$0 = $3;
   break;
  }
  case 11:  {
   $$0 = $3;
   break;
  }
  default: {
   $$0 = 0;
  }
  }
 } while(0);
 return ($$0|0); //@line 98 "modules/physics/box2d/Joint.cpp"
}
function __ZNK4love7physics5box2d5Joint8getBodyAEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 24|0); //@line 102 "modules/physics/box2d/Joint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 102 "modules/physics/box2d/Joint.cpp"
 $2 = ((($1)) + 48|0); //@line 193 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $3 = HEAP32[$2>>2]|0; //@line 193 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $4 = ($3|0)==(0|0); //@line 103 "modules/physics/box2d/Joint.cpp"
 if ($4) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 111 "modules/physics/box2d/Joint.cpp"
 }
 $5 = (__ZN4love8Memoizer4findEPv($3)|0); //@line 106 "modules/physics/box2d/Joint.cpp"
 $6 = ($5|0)==(0|0); //@line 107 "modules/physics/box2d/Joint.cpp"
 if (!($6)) {
  $$0 = $5;
  STACKTOP = sp;return ($$0|0); //@line 111 "modules/physics/box2d/Joint.cpp"
 }
 $7 = (___cxa_allocate_exception(16)|0); //@line 108 "modules/physics/box2d/Joint.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($7|0),(356926|0),($vararg_buffer|0)); //@line 108 "modules/physics/box2d/Joint.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  ___cxa_free_exception(($7|0)); //@line 108 "modules/physics/box2d/Joint.cpp"
  ___resumeException($10|0);
  // unreachable;
 } else {
  ___cxa_throw(($7|0),(96|0),(344|0)); //@line 108 "modules/physics/box2d/Joint.cpp"
  // unreachable; //@line 108 "modules/physics/box2d/Joint.cpp"
 }
 return (0)|0;
}
function __ZNK4love7physics5box2d5Joint8getBodyBEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 24|0); //@line 115 "modules/physics/box2d/Joint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 115 "modules/physics/box2d/Joint.cpp"
 $2 = ((($1)) + 52|0); //@line 198 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $3 = HEAP32[$2>>2]|0; //@line 198 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $4 = ($3|0)==(0|0); //@line 116 "modules/physics/box2d/Joint.cpp"
 if ($4) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 124 "modules/physics/box2d/Joint.cpp"
 }
 $5 = (__ZN4love8Memoizer4findEPv($3)|0); //@line 119 "modules/physics/box2d/Joint.cpp"
 $6 = ($5|0)==(0|0); //@line 120 "modules/physics/box2d/Joint.cpp"
 if (!($6)) {
  $$0 = $5;
  STACKTOP = sp;return ($$0|0); //@line 124 "modules/physics/box2d/Joint.cpp"
 }
 $7 = (___cxa_allocate_exception(16)|0); //@line 121 "modules/physics/box2d/Joint.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($7|0),(356926|0),($vararg_buffer|0)); //@line 121 "modules/physics/box2d/Joint.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  ___cxa_free_exception(($7|0)); //@line 121 "modules/physics/box2d/Joint.cpp"
  ___resumeException($10|0);
  // unreachable;
 } else {
  ___cxa_throw(($7|0),(96|0),(344|0)); //@line 121 "modules/physics/box2d/Joint.cpp"
  // unreachable; //@line 121 "modules/physics/box2d/Joint.cpp"
 }
 return (0)|0;
}
function __ZNK4love7physics5box2d5Joint7isValidEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 128 "modules/physics/box2d/Joint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 128 "modules/physics/box2d/Joint.cpp"
 $2 = ($1|0)!=(0|0); //@line 128 "modules/physics/box2d/Joint.cpp"
 return ($2|0); //@line 128 "modules/physics/box2d/Joint.cpp"
}
function __ZN4love7physics5box2d5Joint10getAnchorsEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 24|0;
 $1 = sp + 16|0;
 $2 = sp + 8|0;
 $3 = sp;
 $4 = ((($this)) + 24|0); //@line 133 "modules/physics/box2d/Joint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 133 "modules/physics/box2d/Joint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 133 "modules/physics/box2d/Joint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 133 "modules/physics/box2d/Joint.cpp"
 FUNCTION_TABLE_vii[$7 & 8191]($0,$5); //@line 133 "modules/physics/box2d/Joint.cpp"
 $8 = +HEAPF32[$0>>2]; //@line 133 "modules/physics/box2d/Joint.cpp"
 $9 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($8)); //@line 133 "modules/physics/box2d/Joint.cpp"
 $10 = $9; //@line 133 "modules/physics/box2d/Joint.cpp"
 _lua_pushnumber($L,$10); //@line 133 "modules/physics/box2d/Joint.cpp"
 $11 = HEAP32[$4>>2]|0; //@line 134 "modules/physics/box2d/Joint.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 134 "modules/physics/box2d/Joint.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 134 "modules/physics/box2d/Joint.cpp"
 FUNCTION_TABLE_vii[$13 & 8191]($1,$11); //@line 134 "modules/physics/box2d/Joint.cpp"
 $14 = ((($1)) + 4|0); //@line 134 "modules/physics/box2d/Joint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 134 "modules/physics/box2d/Joint.cpp"
 $16 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($15)); //@line 134 "modules/physics/box2d/Joint.cpp"
 $17 = $16; //@line 134 "modules/physics/box2d/Joint.cpp"
 _lua_pushnumber($L,$17); //@line 134 "modules/physics/box2d/Joint.cpp"
 $18 = HEAP32[$4>>2]|0; //@line 135 "modules/physics/box2d/Joint.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 135 "modules/physics/box2d/Joint.cpp"
 $20 = ((($19)) + 4|0); //@line 135 "modules/physics/box2d/Joint.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 135 "modules/physics/box2d/Joint.cpp"
 FUNCTION_TABLE_vii[$21 & 8191]($2,$18); //@line 135 "modules/physics/box2d/Joint.cpp"
 $22 = +HEAPF32[$2>>2]; //@line 135 "modules/physics/box2d/Joint.cpp"
 $23 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($22)); //@line 135 "modules/physics/box2d/Joint.cpp"
 $24 = $23; //@line 135 "modules/physics/box2d/Joint.cpp"
 _lua_pushnumber($L,$24); //@line 135 "modules/physics/box2d/Joint.cpp"
 $25 = HEAP32[$4>>2]|0; //@line 136 "modules/physics/box2d/Joint.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 136 "modules/physics/box2d/Joint.cpp"
 $27 = ((($26)) + 4|0); //@line 136 "modules/physics/box2d/Joint.cpp"
 $28 = HEAP32[$27>>2]|0; //@line 136 "modules/physics/box2d/Joint.cpp"
 FUNCTION_TABLE_vii[$28 & 8191]($3,$25); //@line 136 "modules/physics/box2d/Joint.cpp"
 $29 = ((($3)) + 4|0); //@line 136 "modules/physics/box2d/Joint.cpp"
 $30 = +HEAPF32[$29>>2]; //@line 136 "modules/physics/box2d/Joint.cpp"
 $31 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($30)); //@line 136 "modules/physics/box2d/Joint.cpp"
 $32 = $31; //@line 136 "modules/physics/box2d/Joint.cpp"
 _lua_pushnumber($L,$32); //@line 136 "modules/physics/box2d/Joint.cpp"
 STACKTOP = sp;return 4; //@line 137 "modules/physics/box2d/Joint.cpp"
}
function __ZN4love7physics5box2d5Joint16getReactionForceEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $10 = 0, $11 = 0.0, $12 = 0.0, $2 = 0.0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0.0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp + 8|0;
 $0 = sp;
 $1 = (+_luaL_checknumber($L,1)); //@line 142 "modules/physics/box2d/Joint.cpp"
 $2 = $1; //@line 142 "modules/physics/box2d/Joint.cpp"
 $3 = ((($this)) + 24|0); //@line 143 "modules/physics/box2d/Joint.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 143 "modules/physics/box2d/Joint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 143 "modules/physics/box2d/Joint.cpp"
 $6 = ((($5)) + 8|0); //@line 143 "modules/physics/box2d/Joint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 143 "modules/physics/box2d/Joint.cpp"
 FUNCTION_TABLE_viid[$7 & 8191]($0,$4,$2); //@line 143 "modules/physics/box2d/Joint.cpp"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($v,$0); //@line 143 "modules/physics/box2d/Joint.cpp"
 $8 = +HEAPF32[$v>>2]; //@line 144 "modules/physics/box2d/Joint.cpp"
 $9 = $8; //@line 144 "modules/physics/box2d/Joint.cpp"
 _lua_pushnumber($L,$9); //@line 144 "modules/physics/box2d/Joint.cpp"
 $10 = ((($v)) + 4|0); //@line 145 "modules/physics/box2d/Joint.cpp"
 $11 = +HEAPF32[$10>>2]; //@line 145 "modules/physics/box2d/Joint.cpp"
 $12 = $11; //@line 145 "modules/physics/box2d/Joint.cpp"
 _lua_pushnumber($L,$12); //@line 145 "modules/physics/box2d/Joint.cpp"
 STACKTOP = sp;return 2; //@line 146 "modules/physics/box2d/Joint.cpp"
}
function __ZN4love7physics5box2d5Joint17getReactionTorqueEf($this,$dt) {
 $this = $this|0;
 $dt = +$dt;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0.0, $7 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 151 "modules/physics/box2d/Joint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 151 "modules/physics/box2d/Joint.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 151 "modules/physics/box2d/Joint.cpp"
 $3 = ((($2)) + 12|0); //@line 151 "modules/physics/box2d/Joint.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 151 "modules/physics/box2d/Joint.cpp"
 $5 = (+FUNCTION_TABLE_did[$4 & 2047]($1,$dt)); //@line 151 "modules/physics/box2d/Joint.cpp"
 $6 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($5)); //@line 151 "modules/physics/box2d/Joint.cpp"
 $7 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($6)); //@line 151 "modules/physics/box2d/Joint.cpp"
 return (+$7); //@line 151 "modules/physics/box2d/Joint.cpp"
}
function __ZN4love7physics5box2d5Joint11createJointEP10b2JointDef($this,$def) {
 $this = $this|0;
 $def = $def|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 156 "modules/physics/box2d/Joint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 156 "modules/physics/box2d/Joint.cpp"
 $2 = ((($def)) + 4|0); //@line 156 "modules/physics/box2d/Joint.cpp"
 HEAP32[$2>>2] = $1; //@line 156 "modules/physics/box2d/Joint.cpp"
 $3 = ((($this)) + 8|0); //@line 157 "modules/physics/box2d/Joint.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 157 "modules/physics/box2d/Joint.cpp"
 $5 = ((($4)) + 20|0); //@line 157 "modules/physics/box2d/Joint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 157 "modules/physics/box2d/Joint.cpp"
 $7 = (__ZN7b2World11CreateJointEPK10b2JointDef($6,$def)|0); //@line 157 "modules/physics/box2d/Joint.cpp"
 $8 = ((($this)) + 24|0); //@line 157 "modules/physics/box2d/Joint.cpp"
 HEAP32[$8>>2] = $7; //@line 157 "modules/physics/box2d/Joint.cpp"
 __ZN4love8Memoizer3addEPvS1_($7,$this); //@line 158 "modules/physics/box2d/Joint.cpp"
 $9 = HEAP32[$this>>2]|0; //@line 160 "modules/physics/box2d/Joint.cpp"
 $10 = ((($9)) + 8|0); //@line 160 "modules/physics/box2d/Joint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 160 "modules/physics/box2d/Joint.cpp"
 FUNCTION_TABLE_vi[$11 & 8191]($this); //@line 160 "modules/physics/box2d/Joint.cpp"
 $12 = HEAP32[$8>>2]|0; //@line 161 "modules/physics/box2d/Joint.cpp"
 return ($12|0); //@line 161 "modules/physics/box2d/Joint.cpp"
}
function __ZN4love7physics5box2d5Joint12destroyJointEb($this,$implicit) {
 $this = $this|0;
 $implicit = $implicit|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = ((($this)) + 8|0); //@line 166 "modules/physics/box2d/Joint.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 166 "modules/physics/box2d/Joint.cpp"
 $3 = ((($2)) + 20|0); //@line 166 "modules/physics/box2d/Joint.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 166 "modules/physics/box2d/Joint.cpp"
 $5 = ((($4)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $6 = HEAP32[$5>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $7 = $6 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $8 = ($7|0)==(0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 if ($8) {
  $$pre = ((($this)) + 24|0); //@line 176 "modules/physics/box2d/Joint.cpp"
  if (!($implicit)) {
   $22 = HEAP32[$$pre>>2]|0; //@line 174 "modules/physics/box2d/Joint.cpp"
   $23 = ($22|0)==(0|0); //@line 174 "modules/physics/box2d/Joint.cpp"
   if (!($23)) {
    __ZN7b2World12DestroyJointEP7b2Joint($4,$22); //@line 175 "modules/physics/box2d/Joint.cpp"
   }
  }
  $24 = HEAP32[$$pre>>2]|0; //@line 176 "modules/physics/box2d/Joint.cpp"
  __ZN4love8Memoizer6removeEPv($24); //@line 176 "modules/physics/box2d/Joint.cpp"
  HEAP32[$$pre>>2] = 0; //@line 177 "modules/physics/box2d/Joint.cpp"
  $25 = HEAP32[$this>>2]|0; //@line 179 "modules/physics/box2d/Joint.cpp"
  $26 = ((($25)) + 12|0); //@line 179 "modules/physics/box2d/Joint.cpp"
  $27 = HEAP32[$26>>2]|0; //@line 179 "modules/physics/box2d/Joint.cpp"
  FUNCTION_TABLE_vi[$27 & 8191]($this); //@line 179 "modules/physics/box2d/Joint.cpp"
  STACKTOP = sp;return; //@line 180 "modules/physics/box2d/Joint.cpp"
 } else {
  $9 = HEAP32[$this>>2]|0; //@line 169 "modules/physics/box2d/Joint.cpp"
  $10 = ((($9)) + 8|0); //@line 169 "modules/physics/box2d/Joint.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 169 "modules/physics/box2d/Joint.cpp"
  FUNCTION_TABLE_vi[$11 & 8191]($this); //@line 169 "modules/physics/box2d/Joint.cpp"
  $12 = HEAP32[$1>>2]|0; //@line 170 "modules/physics/box2d/Joint.cpp"
  HEAP32[$0>>2] = $this; //@line 170 "modules/physics/box2d/Joint.cpp"
  $13 = ((($12)) + 56|0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = HEAP32[$13>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = ((($12)) + 60|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $16 = HEAP32[$15>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = ($14>>>0)<($16>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($17) {
   $18 = $this; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$14>>2] = $18; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $19 = HEAP32[$13>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $20 = ((($19)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$13>>2] = $20; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   STACKTOP = sp;return; //@line 180 "modules/physics/box2d/Joint.cpp"
  } else {
   $21 = ((($12)) + 52|0); //@line 170 "modules/physics/box2d/Joint.cpp"
   __ZNSt3__16vectorIPN4love7physics5box2d5JointENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_($21,$0); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   STACKTOP = sp;return; //@line 180 "modules/physics/box2d/Joint.cpp"
  }
 }
}
function __ZNSt3__16vectorIPN4love7physics5box2d5JointENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNK4love7physics5box2d5Joint19getCollideConnectedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 189 "modules/physics/box2d/Joint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 189 "modules/physics/box2d/Joint.cpp"
 $2 = ((($1)) + 61|0); //@line 223 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $3 = HEAP8[$2>>0]|0; //@line 223 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $4 = ($3<<24>>24)!=(0); //@line 223 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 return ($4|0); //@line 189 "modules/physics/box2d/Joint.cpp"
}
function __ZN4love7physics5box2d5Joint11setUserDataEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 (__ZN4love16luax_assert_argcEP9lua_Stateii($L,1,1)|0); //@line 194 "modules/physics/box2d/Joint.cpp"
 $0 = ((($this)) + 12|0); //@line 196 "modules/physics/box2d/Joint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 196 "modules/physics/box2d/Joint.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 196 "modules/physics/box2d/Joint.cpp"
 $3 = ($2|0)==(0|0); //@line 196 "modules/physics/box2d/Joint.cpp"
 if (!($3)) {
  $4 = HEAP32[$2>>2]|0; //@line 196 "modules/physics/box2d/Joint.cpp"
  $5 = ((($4)) + 4|0); //@line 196 "modules/physics/box2d/Joint.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 196 "modules/physics/box2d/Joint.cpp"
  FUNCTION_TABLE_vi[$6 & 8191]($2); //@line 196 "modules/physics/box2d/Joint.cpp"
 }
 $7 = (__Znwj(12)|0); //@line 197 "modules/physics/box2d/Joint.cpp"
 __THREW__ = 0;
 invoke_vii(3493,($7|0),($L|0)); //@line 197 "modules/physics/box2d/Joint.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $11 = ___cxa_find_matching_catch_2()|0;
  $12 = tempRet0;
  __ZdlPv($7); //@line 197 "modules/physics/box2d/Joint.cpp"
  ___resumeException($11|0);
  // unreachable;
 } else {
  $10 = HEAP32[$0>>2]|0; //@line 197 "modules/physics/box2d/Joint.cpp"
  HEAP32[$10>>2] = $7; //@line 197 "modules/physics/box2d/Joint.cpp"
  return 0; //@line 199 "modules/physics/box2d/Joint.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d5Joint11getUserDataEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 204 "modules/physics/box2d/Joint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 204 "modules/physics/box2d/Joint.cpp"
 $2 = ($1|0)==(0|0); //@line 204 "modules/physics/box2d/Joint.cpp"
 if (!($2)) {
  $3 = HEAP32[$1>>2]|0; //@line 204 "modules/physics/box2d/Joint.cpp"
  $4 = ($3|0)==(0|0); //@line 204 "modules/physics/box2d/Joint.cpp"
  if (!($4)) {
   __ZN4love9Reference4pushEP9lua_State($3,$L); //@line 205 "modules/physics/box2d/Joint.cpp"
   return 1; //@line 209 "modules/physics/box2d/Joint.cpp"
  }
 }
 _lua_pushnil($L); //@line 207 "modules/physics/box2d/Joint.cpp"
 return 1; //@line 209 "modules/physics/box2d/Joint.cpp"
}
function __ZN4love7physics5box2d7PhysicsD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ModuleD2Ev($this); //@line 55 "modules/physics/box2d/Physics.h"
 __ZdlPv($this); //@line 55 "modules/physics/box2d/Physics.h"
 return; //@line 55 "modules/physics/box2d/Physics.h"
}
function __ZNK4love7physics5box2d7Physics13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 10; //@line 66 "modules/physics/box2d/Physics.h"
}
function __ZNK4love7physics5box2d7Physics7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (356227|0); //@line 38 "modules/physics/box2d/Physics.cpp"
}
function __ZN4love7physics5box2d7Physics8newWorldEffb($this,$gx,$gy,$sleep) {
 $this = $this|0;
 $gx = +$gx;
 $gy = +$gy;
 $sleep = $sleep|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 8|0;
 $0 = sp;
 $1 = (__Znwj(108)|0); //@line 43 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$0>>2] = $gx; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $2 = ((($0)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$2>>2] = $gy; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;HEAP32[$$byval_copy+4>>2]=HEAP32[$0+4>>2]|0; //@line 43 "modules/physics/box2d/Physics.cpp"
 invoke_viii(3671,($1|0),($$byval_copy|0),($sleep|0)); //@line 43 "modules/physics/box2d/Physics.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  __ZdlPv($1); //@line 43 "modules/physics/box2d/Physics.cpp"
  ___resumeException($5|0);
  // unreachable;
 } else {
  STACKTOP = sp;return ($1|0); //@line 43 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics7newBodyEPNS1_5WorldEffNS0_4Body4TypeE($this,$world,$x,$y,$type) {
 $this = $this|0;
 $world = $world|0;
 $x = +$x;
 $y = +$y;
 $type = $type|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 8|0;
 $0 = sp;
 $1 = (__Znwj(20)|0); //@line 48 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$0>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $2 = ((($0)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$2>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;HEAP32[$$byval_copy+4>>2]=HEAP32[$0+4>>2]|0; //@line 48 "modules/physics/box2d/Physics.cpp"
 invoke_viiii(3672,($1|0),($world|0),($$byval_copy|0),($type|0)); //@line 48 "modules/physics/box2d/Physics.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  __ZdlPv($1); //@line 48 "modules/physics/box2d/Physics.cpp"
  ___resumeException($5|0);
  // unreachable;
 } else {
  STACKTOP = sp;return ($1|0); //@line 48 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics14newCircleShapeEf($this,$radius) {
 $this = $this|0;
 $radius = +$radius;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(20)|0); //@line 63 "modules/physics/box2d/Physics.cpp"
 HEAP32[$0>>2] = (20416); //@line 66 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 $1 = ((($0)) + 4|0); //@line 67 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 HEAP32[$1>>2] = 0; //@line 68 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 $2 = HEAP32[4273]|0; //@line 324 "modules/physics/box2d/Physics.cpp"
 $3 = (+($2|0)); //@line 324 "modules/physics/box2d/Physics.cpp"
 $4 = 0.0 / $3; //@line 324 "modules/physics/box2d/Physics.cpp"
 $5 = ((($0)) + 12|0); //@line 64 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$5>>2] = $4; //@line 64 "modules/physics/box2d/Physics.cpp"
 $6 = ((($0)) + 16|0); //@line 64 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$6>>2] = $4; //@line 64 "modules/physics/box2d/Physics.cpp"
 $7 = $radius / $3; //@line 336 "modules/physics/box2d/Physics.cpp"
 $8 = ((($0)) + 8|0); //@line 65 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$8>>2] = $7; //@line 65 "modules/physics/box2d/Physics.cpp"
 $9 = (__Znwj(16)|0); //@line 66 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viii(3673,($9|0),($0|0),1); //@line 66 "modules/physics/box2d/Physics.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $12 = ___cxa_find_matching_catch_2()|0;
  $13 = tempRet0;
  __ZdlPv($9); //@line 66 "modules/physics/box2d/Physics.cpp"
  ___resumeException($12|0);
  // unreachable;
 } else {
  return ($9|0); //@line 58 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics14newCircleShapeEfff($this,$x,$y,$radius) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $radius = +$radius;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(20)|0); //@line 63 "modules/physics/box2d/Physics.cpp"
 HEAP32[$0>>2] = (20416); //@line 66 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 $1 = ((($0)) + 4|0); //@line 67 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0; //@line 68 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 $2 = HEAP32[4273]|0; //@line 324 "modules/physics/box2d/Physics.cpp"
 $3 = (+($2|0)); //@line 324 "modules/physics/box2d/Physics.cpp"
 $4 = $x / $3; //@line 324 "modules/physics/box2d/Physics.cpp"
 $5 = $y / $3; //@line 325 "modules/physics/box2d/Physics.cpp"
 $6 = ((($0)) + 12|0); //@line 64 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$6>>2] = $4; //@line 64 "modules/physics/box2d/Physics.cpp"
 $7 = ((($0)) + 16|0); //@line 64 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$7>>2] = $5; //@line 64 "modules/physics/box2d/Physics.cpp"
 $8 = $radius / $3; //@line 336 "modules/physics/box2d/Physics.cpp"
 $9 = ((($0)) + 8|0); //@line 65 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$9>>2] = $8; //@line 65 "modules/physics/box2d/Physics.cpp"
 $10 = (__Znwj(16)|0); //@line 66 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viii(3673,($10|0),($0|0),1); //@line 66 "modules/physics/box2d/Physics.cpp"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  __ZdlPv($10); //@line 66 "modules/physics/box2d/Physics.cpp"
  ___resumeException($13|0);
  // unreachable;
 } else {
  return ($10|0); //@line 66 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($agg$result,$v) {
 $agg$result = $agg$result|0;
 $v = $v|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $bitcast = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $v; //@line 346 "modules/physics/box2d/Physics.cpp"
 $1 = $0; //@line 346 "modules/physics/box2d/Physics.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 346 "modules/physics/box2d/Physics.cpp"
 $3 = (($0) + 4)|0; //@line 346 "modules/physics/box2d/Physics.cpp"
 $4 = $3; //@line 346 "modules/physics/box2d/Physics.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 346 "modules/physics/box2d/Physics.cpp"
 $6 = $agg$result; //@line 346 "modules/physics/box2d/Physics.cpp"
 $7 = $6; //@line 346 "modules/physics/box2d/Physics.cpp"
 HEAP32[$7>>2] = $2; //@line 346 "modules/physics/box2d/Physics.cpp"
 $8 = (($6) + 4)|0; //@line 346 "modules/physics/box2d/Physics.cpp"
 $9 = $8; //@line 346 "modules/physics/box2d/Physics.cpp"
 HEAP32[$9>>2] = $5; //@line 346 "modules/physics/box2d/Physics.cpp"
 $10 = ((($agg$result)) + 4|0); //@line 347 "modules/physics/box2d/Physics.cpp"
 $11 = HEAP32[4273]|0; //@line 324 "modules/physics/box2d/Physics.cpp"
 $12 = (+($11|0)); //@line 324 "modules/physics/box2d/Physics.cpp"
 $bitcast = (HEAP32[tempDoublePtr>>2]=$2,+HEAPF32[tempDoublePtr>>2]); //@line 324 "modules/physics/box2d/Physics.cpp"
 $13 = $bitcast / $12; //@line 324 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$agg$result>>2] = $13; //@line 324 "modules/physics/box2d/Physics.cpp"
 $14 = (HEAP32[tempDoublePtr>>2]=$5,+HEAPF32[tempDoublePtr>>2]); //@line 325 "modules/physics/box2d/Physics.cpp"
 $15 = $14 / $12; //@line 325 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$10>>2] = $15; //@line 325 "modules/physics/box2d/Physics.cpp"
 return; //@line 348 "modules/physics/box2d/Physics.cpp"
}
function __ZN4love7physics5box2d7Physics9scaleDownEf($f) {
 $f = +$f;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[4273]|0; //@line 336 "modules/physics/box2d/Physics.cpp"
 $1 = (+($0|0)); //@line 336 "modules/physics/box2d/Physics.cpp"
 $2 = $f / $1; //@line 336 "modules/physics/box2d/Physics.cpp"
 return (+$2); //@line 336 "modules/physics/box2d/Physics.cpp"
}
function __ZN4love7physics5box2d7Physics17newRectangleShapeEff($this,$w,$h) {
 $this = $this|0;
 $w = +$w;
 $h = +$h;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0.0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__Znwj(152)|0); //@line 81 "modules/physics/box2d/Physics.cpp"
 HEAP32[$1>>2] = (20496); //@line 88 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $2 = ((($1)) + 4|0); //@line 89 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 HEAP32[$2>>2] = 2; //@line 89 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $3 = ((($1)) + 8|0); //@line 90 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 HEAPF32[$3>>2] = 0.0099999997764825821; //@line 90 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $4 = ((($1)) + 148|0); //@line 91 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 HEAP32[$4>>2] = 0; //@line 91 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $5 = ((($1)) + 12|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$5>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($1)) + 16|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$6>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $7 = $w * 0.5; //@line 82 "modules/physics/box2d/Physics.cpp"
 $8 = HEAP32[4273]|0; //@line 336 "modules/physics/box2d/Physics.cpp"
 $9 = (+($8|0)); //@line 336 "modules/physics/box2d/Physics.cpp"
 $10 = $7 / $9; //@line 336 "modules/physics/box2d/Physics.cpp"
 $11 = $h * 0.5; //@line 82 "modules/physics/box2d/Physics.cpp"
 $12 = $11 / $9; //@line 336 "modules/physics/box2d/Physics.cpp"
 $13 = $0; //@line 346 "modules/physics/box2d/Physics.cpp"
 $14 = $13; //@line 346 "modules/physics/box2d/Physics.cpp"
 HEAP32[$14>>2] = 0; //@line 346 "modules/physics/box2d/Physics.cpp"
 $15 = (($13) + 4)|0; //@line 346 "modules/physics/box2d/Physics.cpp"
 $16 = $15; //@line 346 "modules/physics/box2d/Physics.cpp"
 HEAP32[$16>>2] = 0; //@line 346 "modules/physics/box2d/Physics.cpp"
 $17 = ((($0)) + 4|0); //@line 347 "modules/physics/box2d/Physics.cpp"
 $18 = 0.0 / $9; //@line 324 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$0>>2] = $18; //@line 324 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$17>>2] = $18; //@line 325 "modules/physics/box2d/Physics.cpp"
 __ZN14b2PolygonShape8SetAsBoxEffRK6b2Vec2f($1,$10,$12,$0,0.0); //@line 82 "modules/physics/box2d/Physics.cpp"
 $19 = (__Znwj(16)|0); //@line 83 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viii(3674,($19|0),($1|0),1); //@line 83 "modules/physics/box2d/Physics.cpp"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_2()|0;
  $23 = tempRet0;
  __ZdlPv($19); //@line 83 "modules/physics/box2d/Physics.cpp"
  ___resumeException($22|0);
  // unreachable;
 } else {
  STACKTOP = sp;return ($19|0); //@line 71 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics17newRectangleShapeEfffff($this,$x,$y,$w,$h,$angle) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $w = +$w;
 $h = +$h;
 $angle = +$angle;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__Znwj(152)|0); //@line 81 "modules/physics/box2d/Physics.cpp"
 HEAP32[$1>>2] = (20496); //@line 88 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $2 = ((($1)) + 4|0); //@line 89 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 HEAP32[$2>>2] = 2; //@line 89 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $3 = ((($1)) + 8|0); //@line 90 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 HEAPF32[$3>>2] = 0.0099999997764825821; //@line 90 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $4 = ((($1)) + 148|0); //@line 91 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 HEAP32[$4>>2] = 0; //@line 91 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $5 = ((($1)) + 12|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$5>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($1)) + 16|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$6>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $7 = $w * 0.5; //@line 82 "modules/physics/box2d/Physics.cpp"
 $8 = HEAP32[4273]|0; //@line 336 "modules/physics/box2d/Physics.cpp"
 $9 = (+($8|0)); //@line 336 "modules/physics/box2d/Physics.cpp"
 $10 = $7 / $9; //@line 336 "modules/physics/box2d/Physics.cpp"
 $11 = $h * 0.5; //@line 82 "modules/physics/box2d/Physics.cpp"
 $12 = $11 / $9; //@line 336 "modules/physics/box2d/Physics.cpp"
 $13 = (HEAPF32[tempDoublePtr>>2]=$x,HEAP32[tempDoublePtr>>2]|0); //@line 346 "modules/physics/box2d/Physics.cpp"
 $14 = (HEAPF32[tempDoublePtr>>2]=$y,HEAP32[tempDoublePtr>>2]|0); //@line 346 "modules/physics/box2d/Physics.cpp"
 $15 = $0; //@line 346 "modules/physics/box2d/Physics.cpp"
 $16 = $15; //@line 346 "modules/physics/box2d/Physics.cpp"
 HEAP32[$16>>2] = $13; //@line 346 "modules/physics/box2d/Physics.cpp"
 $17 = (($15) + 4)|0; //@line 346 "modules/physics/box2d/Physics.cpp"
 $18 = $17; //@line 346 "modules/physics/box2d/Physics.cpp"
 HEAP32[$18>>2] = $14; //@line 346 "modules/physics/box2d/Physics.cpp"
 $19 = ((($0)) + 4|0); //@line 347 "modules/physics/box2d/Physics.cpp"
 $20 = $x / $9; //@line 324 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$0>>2] = $20; //@line 324 "modules/physics/box2d/Physics.cpp"
 $21 = $y / $9; //@line 325 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$19>>2] = $21; //@line 325 "modules/physics/box2d/Physics.cpp"
 __ZN14b2PolygonShape8SetAsBoxEffRK6b2Vec2f($1,$10,$12,$0,$angle); //@line 82 "modules/physics/box2d/Physics.cpp"
 $22 = (__Znwj(16)|0); //@line 83 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viii(3674,($22|0),($1|0),1); //@line 83 "modules/physics/box2d/Physics.cpp"
 $23 = __THREW__; __THREW__ = 0;
 $24 = $23&1;
 if ($24) {
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  __ZdlPv($22); //@line 83 "modules/physics/box2d/Physics.cpp"
  ___resumeException($25|0);
  // unreachable;
 } else {
  STACKTOP = sp;return ($22|0); //@line 83 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics12newEdgeShapeEffff($this,$x1,$y1,$x2,$y2) {
 $this = $this|0;
 $x1 = +$x1;
 $y1 = +$y1;
 $x2 = +$x2;
 $y2 = +$y2;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 8|0;
 $1 = sp;
 $2 = (__Znwj(48)|0); //@line 88 "modules/physics/box2d/Physics.cpp"
 HEAP32[$2>>2] = (20456); //@line 63 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $3 = ((($2)) + 4|0); //@line 64 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 HEAP32[$3>>2] = 1; //@line 64 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $4 = ((($2)) + 8|0); //@line 65 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 HEAPF32[$4>>2] = 0.0099999997764825821; //@line 65 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $5 = ((($2)) + 28|0); //@line 66 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;HEAP32[$5+12>>2]=0|0;HEAP16[$5+16>>1]=0|0; //@line 67 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $6 = (HEAPF32[tempDoublePtr>>2]=$x1,HEAP32[tempDoublePtr>>2]|0); //@line 346 "modules/physics/box2d/Physics.cpp"
 $7 = (HEAPF32[tempDoublePtr>>2]=$y1,HEAP32[tempDoublePtr>>2]|0); //@line 346 "modules/physics/box2d/Physics.cpp"
 $8 = $0; //@line 346 "modules/physics/box2d/Physics.cpp"
 $9 = $8; //@line 346 "modules/physics/box2d/Physics.cpp"
 HEAP32[$9>>2] = $6; //@line 346 "modules/physics/box2d/Physics.cpp"
 $10 = (($8) + 4)|0; //@line 346 "modules/physics/box2d/Physics.cpp"
 $11 = $10; //@line 346 "modules/physics/box2d/Physics.cpp"
 HEAP32[$11>>2] = $7; //@line 346 "modules/physics/box2d/Physics.cpp"
 $12 = ((($0)) + 4|0); //@line 347 "modules/physics/box2d/Physics.cpp"
 $13 = HEAP32[4273]|0; //@line 324 "modules/physics/box2d/Physics.cpp"
 $14 = (+($13|0)); //@line 324 "modules/physics/box2d/Physics.cpp"
 $15 = $x1 / $14; //@line 324 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$0>>2] = $15; //@line 324 "modules/physics/box2d/Physics.cpp"
 $16 = $y1 / $14; //@line 325 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$12>>2] = $16; //@line 325 "modules/physics/box2d/Physics.cpp"
 $17 = (HEAPF32[tempDoublePtr>>2]=$x2,HEAP32[tempDoublePtr>>2]|0); //@line 346 "modules/physics/box2d/Physics.cpp"
 $18 = (HEAPF32[tempDoublePtr>>2]=$y2,HEAP32[tempDoublePtr>>2]|0); //@line 346 "modules/physics/box2d/Physics.cpp"
 $19 = $1; //@line 346 "modules/physics/box2d/Physics.cpp"
 $20 = $19; //@line 346 "modules/physics/box2d/Physics.cpp"
 HEAP32[$20>>2] = $17; //@line 346 "modules/physics/box2d/Physics.cpp"
 $21 = (($19) + 4)|0; //@line 346 "modules/physics/box2d/Physics.cpp"
 $22 = $21; //@line 346 "modules/physics/box2d/Physics.cpp"
 HEAP32[$22>>2] = $18; //@line 346 "modules/physics/box2d/Physics.cpp"
 $23 = ((($1)) + 4|0); //@line 347 "modules/physics/box2d/Physics.cpp"
 $24 = $x2 / $14; //@line 324 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$1>>2] = $24; //@line 324 "modules/physics/box2d/Physics.cpp"
 $25 = $y2 / $14; //@line 325 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$23>>2] = $25; //@line 325 "modules/physics/box2d/Physics.cpp"
 __ZN11b2EdgeShape3SetERK6b2Vec2S2_($2,$0,$1); //@line 89 "modules/physics/box2d/Physics.cpp"
 $26 = (__Znwj(16)|0); //@line 90 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viii(3675,($26|0),($2|0),1); //@line 90 "modules/physics/box2d/Physics.cpp"
 $27 = __THREW__; __THREW__ = 0;
 $28 = $27&1;
 if ($28) {
  $29 = ___cxa_find_matching_catch_2()|0;
  $30 = tempRet0;
  __ZdlPv($26); //@line 90 "modules/physics/box2d/Physics.cpp"
  ___resumeException($29|0);
  // unreachable;
 } else {
  STACKTOP = sp;return ($26|0); //@line 90 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics15newPolygonShapeEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $$01 = 0, $$02 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $argc$0 = 0, $i$013 = 0, $i1$014 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0;
 var $vararg_buffer3 = 0, $vararg_ptr6 = 0, $vecs = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $vecs = sp + 24|0;
 $0 = (_lua_gettop($L)|0); //@line 95 "modules/physics/box2d/Physics.cpp"
 $1 = (_lua_type($L,1)|0); //@line 97 "modules/physics/box2d/Physics.cpp"
 $2 = ($1|0)==(5); //@line 97 "modules/physics/box2d/Physics.cpp"
 if ($2) {
  $3 = (__ZN4love11luax_objlenEP9lua_Statei($L,1)|0); //@line 100 "modules/physics/box2d/Physics.cpp"
  $argc$0 = $3;
 } else {
  $argc$0 = $0;
 }
 $4 = $argc$0 & 1; //@line 102 "modules/physics/box2d/Physics.cpp"
 $5 = ($4|0)==(0); //@line 102 "modules/physics/box2d/Physics.cpp"
 if (!($5)) {
  $6 = (_luaL_error($L,356246,$vararg_buffer)|0); //@line 103 "modules/physics/box2d/Physics.cpp"
  $$1 = $6;
  STACKTOP = sp;return ($$1|0); //@line 152 "modules/physics/box2d/Physics.cpp"
 }
 $7 = (($argc$0|0) / 2)&-1; //@line 106 "modules/physics/box2d/Physics.cpp"
 $8 = ($argc$0|0)<(6); //@line 107 "modules/physics/box2d/Physics.cpp"
 if ($8) {
  HEAP32[$vararg_buffer1>>2] = $7; //@line 108 "modules/physics/box2d/Physics.cpp"
  $9 = (_luaL_error($L,356301,$vararg_buffer1)|0); //@line 108 "modules/physics/box2d/Physics.cpp"
  $$1 = $9;
  STACKTOP = sp;return ($$1|0); //@line 152 "modules/physics/box2d/Physics.cpp"
 }
 $10 = ($argc$0|0)>(17); //@line 109 "modules/physics/box2d/Physics.cpp"
 if ($10) {
  HEAP32[$vararg_buffer3>>2] = 8; //@line 110 "modules/physics/box2d/Physics.cpp"
  $vararg_ptr6 = ((($vararg_buffer3)) + 4|0); //@line 110 "modules/physics/box2d/Physics.cpp"
  HEAP32[$vararg_ptr6>>2] = $7; //@line 110 "modules/physics/box2d/Physics.cpp"
  $11 = (_luaL_error($L,356343,$vararg_buffer3)|0); //@line 110 "modules/physics/box2d/Physics.cpp"
  $$1 = $11;
  STACKTOP = sp;return ($$1|0); //@line 152 "modules/physics/box2d/Physics.cpp"
 }
 if ($2) {
  $i$013 = 0;
  while(1) {
   $12 = $i$013 << 1; //@line 118 "modules/physics/box2d/Physics.cpp"
   $13 = $12 | 1; //@line 118 "modules/physics/box2d/Physics.cpp"
   _lua_rawgeti($L,1,$13); //@line 118 "modules/physics/box2d/Physics.cpp"
   $14 = (($12) + 2)|0; //@line 119 "modules/physics/box2d/Physics.cpp"
   _lua_rawgeti($L,1,$14); //@line 119 "modules/physics/box2d/Physics.cpp"
   $15 = (+_luaL_checknumber($L,-2)); //@line 120 "modules/physics/box2d/Physics.cpp"
   $16 = $15; //@line 120 "modules/physics/box2d/Physics.cpp"
   $17 = (+_luaL_checknumber($L,-1)); //@line 121 "modules/physics/box2d/Physics.cpp"
   $18 = $17; //@line 121 "modules/physics/box2d/Physics.cpp"
   $19 = HEAP32[4273]|0; //@line 324 "modules/physics/box2d/Physics.cpp"
   $20 = (+($19|0)); //@line 324 "modules/physics/box2d/Physics.cpp"
   $21 = $16 / $20; //@line 324 "modules/physics/box2d/Physics.cpp"
   $22 = $18 / $20; //@line 325 "modules/physics/box2d/Physics.cpp"
   $23 = (($vecs) + ($i$013<<3)|0); //@line 122 "modules/physics/box2d/Physics.cpp"
   HEAPF32[$23>>2] = $21; //@line 122 "modules/physics/box2d/Physics.cpp"
   $24 = (((($vecs) + ($i$013<<3)|0)) + 4|0); //@line 122 "modules/physics/box2d/Physics.cpp"
   HEAPF32[$24>>2] = $22; //@line 122 "modules/physics/box2d/Physics.cpp"
   _lua_settop($L,-3); //@line 123 "modules/physics/box2d/Physics.cpp"
   $25 = (($i$013) + 1)|0; //@line 116 "modules/physics/box2d/Physics.cpp"
   $26 = ($25|0)<($7|0); //@line 116 "modules/physics/box2d/Physics.cpp"
   if ($26) {
    $i$013 = $25;
   } else {
    break;
   }
  }
 } else {
  $i1$014 = 0;
  while(1) {
   $27 = $i1$014 << 1; //@line 130 "modules/physics/box2d/Physics.cpp"
   $28 = $27 | 1; //@line 130 "modules/physics/box2d/Physics.cpp"
   $29 = (+_luaL_checknumber($L,$28)); //@line 130 "modules/physics/box2d/Physics.cpp"
   $30 = $29; //@line 130 "modules/physics/box2d/Physics.cpp"
   $31 = (($27) + 2)|0; //@line 131 "modules/physics/box2d/Physics.cpp"
   $32 = (+_luaL_checknumber($L,$31)); //@line 131 "modules/physics/box2d/Physics.cpp"
   $33 = $32; //@line 131 "modules/physics/box2d/Physics.cpp"
   $34 = HEAP32[4273]|0; //@line 324 "modules/physics/box2d/Physics.cpp"
   $35 = (+($34|0)); //@line 324 "modules/physics/box2d/Physics.cpp"
   $36 = $30 / $35; //@line 324 "modules/physics/box2d/Physics.cpp"
   $37 = $33 / $35; //@line 325 "modules/physics/box2d/Physics.cpp"
   $38 = (($vecs) + ($i1$014<<3)|0); //@line 132 "modules/physics/box2d/Physics.cpp"
   HEAPF32[$38>>2] = $36; //@line 132 "modules/physics/box2d/Physics.cpp"
   $39 = (((($vecs) + ($i1$014<<3)|0)) + 4|0); //@line 132 "modules/physics/box2d/Physics.cpp"
   HEAPF32[$39>>2] = $37; //@line 132 "modules/physics/box2d/Physics.cpp"
   $40 = (($i1$014) + 1)|0; //@line 128 "modules/physics/box2d/Physics.cpp"
   $41 = ($40|0)<($7|0); //@line 128 "modules/physics/box2d/Physics.cpp"
   if ($41) {
    $i1$014 = $40;
   } else {
    break;
   }
  }
 }
 $42 = (__Znwj(152)|0); //@line 136 "modules/physics/box2d/Physics.cpp"
 HEAP32[$42>>2] = (20496); //@line 88 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $43 = ((($42)) + 4|0); //@line 89 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 HEAP32[$43>>2] = 2; //@line 89 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $44 = ((($42)) + 8|0); //@line 90 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 HEAPF32[$44>>2] = 0.0099999997764825821; //@line 90 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $45 = ((($42)) + 148|0); //@line 91 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 HEAP32[$45>>2] = 0; //@line 91 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $46 = ((($42)) + 12|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$46>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $47 = ((($42)) + 16|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$47>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 invoke_viii(3676,($42|0),($vecs|0),($7|0)); //@line 140 "modules/physics/box2d/Physics.cpp"
 $48 = __THREW__; __THREW__ = 0;
 $49 = $48&1;
 if ($49) {
  $50 = ___cxa_find_matching_catch_3(96|0)|0;
  $51 = tempRet0;
  $52 = (_llvm_eh_typeid_for((96|0))|0); //@line 141 "modules/physics/box2d/Physics.cpp"
  $53 = ($51|0)==($52|0); //@line 141 "modules/physics/box2d/Physics.cpp"
  if (!($53)) {
   $$01 = $51;$$02 = $50;
   ___resumeException($$02|0);
   // unreachable;
  }
  (___cxa_begin_catch(($50|0))|0); //@line 141 "modules/physics/box2d/Physics.cpp"
  $54 = HEAP32[$42>>2]|0; //@line 144 "modules/physics/box2d/Physics.cpp"
  $55 = ((($54)) + 4|0); //@line 144 "modules/physics/box2d/Physics.cpp"
  $56 = HEAP32[$55>>2]|0; //@line 144 "modules/physics/box2d/Physics.cpp"
  FUNCTION_TABLE_vi[$56 & 8191]($42); //@line 144 "modules/physics/box2d/Physics.cpp"
  __THREW__ = 0;
  invoke_v(3656); //@line 145 "modules/physics/box2d/Physics.cpp"
  $57 = __THREW__; __THREW__ = 0;
  $58 = ___cxa_find_matching_catch_2()|0;
  $59 = tempRet0;
  __THREW__ = 0;
  invoke_v(3378); //@line 146 "modules/physics/box2d/Physics.cpp"
  $60 = __THREW__; __THREW__ = 0;
  $61 = $60&1;
  if ($61) {
   $70 = ___cxa_find_matching_catch_3(0|0)|0;
   $71 = tempRet0;
   ___clang_call_terminate($70); //@line 146 "modules/physics/box2d/Physics.cpp"
   // unreachable; //@line 146 "modules/physics/box2d/Physics.cpp"
  } else {
   $$01 = $59;$$02 = $58;
   ___resumeException($$02|0);
   // unreachable;
  }
 } else {
  $62 = (__Znwj(16)|0); //@line 148 "modules/physics/box2d/Physics.cpp"
  __THREW__ = 0;
  invoke_viii(3674,($62|0),($42|0),1); //@line 148 "modules/physics/box2d/Physics.cpp"
  $63 = __THREW__; __THREW__ = 0;
  $64 = $63&1;
  if ($64) {
   $68 = ___cxa_find_matching_catch_2()|0;
   $69 = tempRet0;
   __ZdlPv($62); //@line 148 "modules/physics/box2d/Physics.cpp"
   $$01 = $69;$$02 = $68;
   ___resumeException($$02|0);
   // unreachable;
  } else {
   __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,38,$62); //@line 149 "modules/physics/box2d/Physics.cpp"
   $65 = HEAP32[$62>>2]|0; //@line 150 "modules/physics/box2d/Physics.cpp"
   $66 = ((($65)) + 12|0); //@line 150 "modules/physics/box2d/Physics.cpp"
   $67 = HEAP32[$66>>2]|0; //@line 150 "modules/physics/box2d/Physics.cpp"
   FUNCTION_TABLE_vi[$67 & 8191]($62); //@line 150 "modules/physics/box2d/Physics.cpp"
   $$1 = 1;
   STACKTOP = sp;return ($$1|0); //@line 152 "modules/physics/box2d/Physics.cpp"
  }
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics13newChainShapeEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $$arith = 0, $$overflow = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0;
 var $22 = 0.0, $23 = 0.0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $8 = 0;
 var $9 = 0, $argc$0 = 0, $i$014 = 0, $i1$015 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_lua_gettop($L)|0); //@line 156 "modules/physics/box2d/Physics.cpp"
 $1 = (($0) + -1)|0; //@line 156 "modules/physics/box2d/Physics.cpp"
 $2 = (_lua_type($L,2)|0); //@line 158 "modules/physics/box2d/Physics.cpp"
 $3 = ($2|0)==(5); //@line 158 "modules/physics/box2d/Physics.cpp"
 if ($3) {
  $4 = (__ZN4love11luax_objlenEP9lua_Statei($L,2)|0); //@line 161 "modules/physics/box2d/Physics.cpp"
  $argc$0 = $4;
 } else {
  $argc$0 = $1;
 }
 $5 = $argc$0 & 1; //@line 163 "modules/physics/box2d/Physics.cpp"
 $6 = ($5|0)==(0); //@line 163 "modules/physics/box2d/Physics.cpp"
 if (!($6)) {
  $7 = (_luaL_error($L,356246,$vararg_buffer)|0); //@line 164 "modules/physics/box2d/Physics.cpp"
  $$0 = $7;
  STACKTOP = sp;return ($$0|0); //@line 214 "modules/physics/box2d/Physics.cpp"
 }
 $8 = (($argc$0|0) / 2)&-1; //@line 166 "modules/physics/box2d/Physics.cpp"
 $9 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,1)|0); //@line 167 "modules/physics/box2d/Physics.cpp"
 $$arith = $8<<3; //@line 168 "modules/physics/box2d/Physics.cpp"
 $$overflow = ($8>>>0)>(536870911); //@line 168 "modules/physics/box2d/Physics.cpp"
 $10 = $$overflow ? -1 : $$arith; //@line 168 "modules/physics/box2d/Physics.cpp"
 $11 = (__Znaj($10)|0); //@line 168 "modules/physics/box2d/Physics.cpp"
 $12 = ($argc$0|0)>(1); //@line 172 "modules/physics/box2d/Physics.cpp"
 if ($3) {
  if ($12) {
   $i$014 = 0;
   while(1) {
    $13 = $i$014 << 1; //@line 174 "modules/physics/box2d/Physics.cpp"
    $14 = $13 | 1; //@line 174 "modules/physics/box2d/Physics.cpp"
    _lua_rawgeti($L,2,$14); //@line 174 "modules/physics/box2d/Physics.cpp"
    $15 = (($13) + 2)|0; //@line 175 "modules/physics/box2d/Physics.cpp"
    _lua_rawgeti($L,2,$15); //@line 175 "modules/physics/box2d/Physics.cpp"
    $16 = (+_lua_tonumber($L,-2)); //@line 176 "modules/physics/box2d/Physics.cpp"
    $17 = $16; //@line 176 "modules/physics/box2d/Physics.cpp"
    $18 = (+_lua_tonumber($L,-1)); //@line 177 "modules/physics/box2d/Physics.cpp"
    $19 = $18; //@line 177 "modules/physics/box2d/Physics.cpp"
    $20 = HEAP32[4273]|0; //@line 324 "modules/physics/box2d/Physics.cpp"
    $21 = (+($20|0)); //@line 324 "modules/physics/box2d/Physics.cpp"
    $22 = $17 / $21; //@line 324 "modules/physics/box2d/Physics.cpp"
    $23 = $19 / $21; //@line 325 "modules/physics/box2d/Physics.cpp"
    $24 = (($11) + ($i$014<<3)|0); //@line 178 "modules/physics/box2d/Physics.cpp"
    HEAPF32[$24>>2] = $22; //@line 178 "modules/physics/box2d/Physics.cpp"
    $25 = (((($11) + ($i$014<<3)|0)) + 4|0); //@line 178 "modules/physics/box2d/Physics.cpp"
    HEAPF32[$25>>2] = $23; //@line 178 "modules/physics/box2d/Physics.cpp"
    _lua_settop($L,-3); //@line 179 "modules/physics/box2d/Physics.cpp"
    $26 = (($i$014) + 1)|0; //@line 172 "modules/physics/box2d/Physics.cpp"
    $27 = ($26|0)<($8|0); //@line 172 "modules/physics/box2d/Physics.cpp"
    if ($27) {
     $i$014 = $26;
    } else {
     break;
    }
   }
  }
 } else {
  if ($12) {
   $i1$015 = 0;
   while(1) {
    $28 = $i1$015 << 1; //@line 186 "modules/physics/box2d/Physics.cpp"
    $29 = (($28) + 2)|0; //@line 186 "modules/physics/box2d/Physics.cpp"
    $30 = (+_luaL_checknumber($L,$29)); //@line 186 "modules/physics/box2d/Physics.cpp"
    $31 = $30; //@line 186 "modules/physics/box2d/Physics.cpp"
    $32 = (($28) + 3)|0; //@line 187 "modules/physics/box2d/Physics.cpp"
    $33 = (+_luaL_checknumber($L,$32)); //@line 187 "modules/physics/box2d/Physics.cpp"
    $34 = $33; //@line 187 "modules/physics/box2d/Physics.cpp"
    $35 = HEAP32[4273]|0; //@line 324 "modules/physics/box2d/Physics.cpp"
    $36 = (+($35|0)); //@line 324 "modules/physics/box2d/Physics.cpp"
    $37 = $31 / $36; //@line 324 "modules/physics/box2d/Physics.cpp"
    $38 = $34 / $36; //@line 325 "modules/physics/box2d/Physics.cpp"
    $39 = (($11) + ($i1$015<<3)|0); //@line 188 "modules/physics/box2d/Physics.cpp"
    HEAPF32[$39>>2] = $37; //@line 188 "modules/physics/box2d/Physics.cpp"
    $40 = (((($11) + ($i1$015<<3)|0)) + 4|0); //@line 188 "modules/physics/box2d/Physics.cpp"
    HEAPF32[$40>>2] = $38; //@line 188 "modules/physics/box2d/Physics.cpp"
    $41 = (($i1$015) + 1)|0; //@line 184 "modules/physics/box2d/Physics.cpp"
    $42 = ($41|0)<($8|0); //@line 184 "modules/physics/box2d/Physics.cpp"
    if ($42) {
     $i1$015 = $41;
    } else {
     break;
    }
   }
  }
 }
 $43 = (__Znwj(40)|0); //@line 192 "modules/physics/box2d/Physics.cpp"
 HEAP32[$43>>2] = (20376); //@line 96 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 $44 = ((($43)) + 4|0); //@line 97 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 HEAP32[$44>>2] = 3; //@line 97 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 $45 = ((($43)) + 8|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 HEAPF32[$45>>2] = 0.0099999997764825821; //@line 98 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 $46 = ((($43)) + 12|0); //@line 99 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 HEAP32[$46>>2] = 0; //@line 99 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 $47 = ((($43)) + 16|0); //@line 100 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 HEAP32[$47>>2] = 0; //@line 100 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 $48 = ((($43)) + 36|0); //@line 101 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 HEAP8[$48>>0] = 0; //@line 101 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 $49 = ((($43)) + 37|0); //@line 102 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 HEAP8[$49>>0] = 0; //@line 102 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 if ($9) {
  __THREW__ = 0;
  invoke_viii(3677,($43|0),($11|0),($8|0)); //@line 197 "modules/physics/box2d/Physics.cpp"
  $50 = __THREW__; __THREW__ = 0;
  $51 = $50&1;
  if ($51) {
   label = 12;
  } else {
   label = 15;
  }
 } else {
  __THREW__ = 0;
  invoke_viii(3678,($43|0),($11|0),($8|0)); //@line 199 "modules/physics/box2d/Physics.cpp"
  $64 = __THREW__; __THREW__ = 0;
  $65 = $64&1;
  if ($65) {
   label = 12;
  } else {
   label = 15;
  }
 }
 if ((label|0) == 12) {
  $52 = ___cxa_find_matching_catch_3(96|0)|0;
  $53 = tempRet0;
  $54 = (_llvm_eh_typeid_for((96|0))|0); //@line 200 "modules/physics/box2d/Physics.cpp"
  $55 = ($53|0)==($54|0); //@line 200 "modules/physics/box2d/Physics.cpp"
  if (!($55)) {
   $$01 = $52;$$02 = $53;
   ___resumeException($$01|0);
   // unreachable;
  }
  (___cxa_begin_catch(($52|0))|0); //@line 200 "modules/physics/box2d/Physics.cpp"
  __ZdaPv($11); //@line 203 "modules/physics/box2d/Physics.cpp"
  $56 = HEAP32[$43>>2]|0; //@line 204 "modules/physics/box2d/Physics.cpp"
  $57 = ((($56)) + 4|0); //@line 204 "modules/physics/box2d/Physics.cpp"
  $58 = HEAP32[$57>>2]|0; //@line 204 "modules/physics/box2d/Physics.cpp"
  FUNCTION_TABLE_vi[$58 & 8191]($43); //@line 204 "modules/physics/box2d/Physics.cpp"
  __THREW__ = 0;
  invoke_v(3656); //@line 205 "modules/physics/box2d/Physics.cpp"
  $59 = __THREW__; __THREW__ = 0;
  $60 = ___cxa_find_matching_catch_2()|0;
  $61 = tempRet0;
  __THREW__ = 0;
  invoke_v(3378); //@line 206 "modules/physics/box2d/Physics.cpp"
  $62 = __THREW__; __THREW__ = 0;
  $63 = $62&1;
  if ($63) {
   $74 = ___cxa_find_matching_catch_3(0|0)|0;
   $75 = tempRet0;
   ___clang_call_terminate($74); //@line 206 "modules/physics/box2d/Physics.cpp"
   // unreachable; //@line 206 "modules/physics/box2d/Physics.cpp"
  } else {
   $$01 = $60;$$02 = $61;
   ___resumeException($$01|0);
   // unreachable;
  }
 }
 else if ((label|0) == 15) {
  __ZdaPv($11); //@line 208 "modules/physics/box2d/Physics.cpp"
  $66 = (__Znwj(16)|0); //@line 210 "modules/physics/box2d/Physics.cpp"
  __THREW__ = 0;
  invoke_viiii(3679,($66|0),($43|0),0,1); //@line 210 "modules/physics/box2d/Physics.cpp"
  $67 = __THREW__; __THREW__ = 0;
  $68 = $67&1;
  if ($68) {
   $72 = ___cxa_find_matching_catch_2()|0;
   $73 = tempRet0;
   __ZdlPv($66); //@line 210 "modules/physics/box2d/Physics.cpp"
   $$01 = $72;$$02 = $73;
   ___resumeException($$01|0);
   // unreachable;
  } else {
   __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,40,$66); //@line 211 "modules/physics/box2d/Physics.cpp"
   $69 = HEAP32[$66>>2]|0; //@line 212 "modules/physics/box2d/Physics.cpp"
   $70 = ((($69)) + 12|0); //@line 212 "modules/physics/box2d/Physics.cpp"
   $71 = HEAP32[$70>>2]|0; //@line 212 "modules/physics/box2d/Physics.cpp"
   FUNCTION_TABLE_vi[$71 & 8191]($66); //@line 212 "modules/physics/box2d/Physics.cpp"
   $$0 = 1;
   STACKTOP = sp;return ($$0|0); //@line 214 "modules/physics/box2d/Physics.cpp"
  }
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics16newDistanceJointEPNS1_4BodyES4_ffffb($this,$body1,$body2,$x1,$y1,$x2,$y2,$collideConnected) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 $x1 = +$x1;
 $y1 = +$y1;
 $x2 = +$x2;
 $y2 = +$y2;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(32)|0); //@line 218 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viiiddddi(3680,($0|0),($body1|0),($body2|0),(+$x1),(+$y1),(+$x2),(+$y2),($collideConnected|0)); //@line 218 "modules/physics/box2d/Physics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 218 "modules/physics/box2d/Physics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 218 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics13newMouseJointEPNS1_4BodyEff($this,$body,$x,$y) {
 $this = $this|0;
 $body = $body|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(32)|0); //@line 223 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viidd(3681,($0|0),($body|0),(+$x),(+$y)); //@line 223 "modules/physics/box2d/Physics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 223 "modules/physics/box2d/Physics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 223 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics16newRevoluteJointEPNS1_4BodyES4_ffb($this,$body1,$body2,$x,$y,$collideConnected) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 $x = +$x;
 $y = +$y;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(32)|0); //@line 228 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viiiddi(3682,($0|0),($body1|0),($body2|0),(+$x),(+$y),($collideConnected|0)); //@line 228 "modules/physics/box2d/Physics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 228 "modules/physics/box2d/Physics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 228 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics17newPrismaticJointEPNS1_4BodyES4_ffffffb($this,$body1,$body2,$xA,$yA,$xB,$yB,$ax,$ay,$collideConnected) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 $xA = +$xA;
 $yA = +$yA;
 $xB = +$xB;
 $yB = +$yB;
 $ax = +$ax;
 $ay = +$ay;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(32)|0); //@line 233 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viiiddddddi(3683,($0|0),($body1|0),($body2|0),(+$xA),(+$yA),(+$xB),(+$yB),(+$ax),(+$ay),($collideConnected|0)); //@line 233 "modules/physics/box2d/Physics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 233 "modules/physics/box2d/Physics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 233 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics14newPulleyJointEPNS1_4BodyES4_6b2Vec2S5_S5_S5_fb($this,$body1,$body2,$groundAnchor1,$groundAnchor2,$anchor1,$anchor2,$ratio,$collideConnected) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 $groundAnchor1 = $groundAnchor1|0;
 $groundAnchor2 = $groundAnchor2|0;
 $anchor1 = $anchor1|0;
 $anchor2 = $anchor2|0;
 $ratio = +$ratio;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $tmpcast$byval_copy = 0, $tmpcast1$byval_copy = 0, $tmpcast2$byval_copy = 0, $tmpcast3$byval_copy = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tmpcast3$byval_copy = sp + 56|0;
 $tmpcast2$byval_copy = sp + 48|0;
 $tmpcast1$byval_copy = sp + 40|0;
 $tmpcast$byval_copy = sp + 32|0;
 $0 = sp + 24|0;
 $1 = sp + 16|0;
 $2 = sp + 8|0;
 $3 = sp;
 $4 = (__Znwj(32)|0); //@line 238 "modules/physics/box2d/Physics.cpp"
 $5 = $groundAnchor1; //@line 238 "modules/physics/box2d/Physics.cpp"
 $6 = $5; //@line 238 "modules/physics/box2d/Physics.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 $8 = (($5) + 4)|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 $9 = $8; //@line 238 "modules/physics/box2d/Physics.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 $11 = $0; //@line 238 "modules/physics/box2d/Physics.cpp"
 $12 = $11; //@line 238 "modules/physics/box2d/Physics.cpp"
 HEAP32[$12>>2] = $7; //@line 238 "modules/physics/box2d/Physics.cpp"
 $13 = (($11) + 4)|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 $14 = $13; //@line 238 "modules/physics/box2d/Physics.cpp"
 HEAP32[$14>>2] = $10; //@line 238 "modules/physics/box2d/Physics.cpp"
 $15 = $groundAnchor2; //@line 238 "modules/physics/box2d/Physics.cpp"
 $16 = $15; //@line 238 "modules/physics/box2d/Physics.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 $18 = (($15) + 4)|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 $19 = $18; //@line 238 "modules/physics/box2d/Physics.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 $21 = $1; //@line 238 "modules/physics/box2d/Physics.cpp"
 $22 = $21; //@line 238 "modules/physics/box2d/Physics.cpp"
 HEAP32[$22>>2] = $17; //@line 238 "modules/physics/box2d/Physics.cpp"
 $23 = (($21) + 4)|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 $24 = $23; //@line 238 "modules/physics/box2d/Physics.cpp"
 HEAP32[$24>>2] = $20; //@line 238 "modules/physics/box2d/Physics.cpp"
 $25 = $anchor1; //@line 238 "modules/physics/box2d/Physics.cpp"
 $26 = $25; //@line 238 "modules/physics/box2d/Physics.cpp"
 $27 = HEAP32[$26>>2]|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 $28 = (($25) + 4)|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 $29 = $28; //@line 238 "modules/physics/box2d/Physics.cpp"
 $30 = HEAP32[$29>>2]|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 $31 = $2; //@line 238 "modules/physics/box2d/Physics.cpp"
 $32 = $31; //@line 238 "modules/physics/box2d/Physics.cpp"
 HEAP32[$32>>2] = $27; //@line 238 "modules/physics/box2d/Physics.cpp"
 $33 = (($31) + 4)|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 $34 = $33; //@line 238 "modules/physics/box2d/Physics.cpp"
 HEAP32[$34>>2] = $30; //@line 238 "modules/physics/box2d/Physics.cpp"
 $35 = $anchor2; //@line 238 "modules/physics/box2d/Physics.cpp"
 $36 = $35; //@line 238 "modules/physics/box2d/Physics.cpp"
 $37 = HEAP32[$36>>2]|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 $38 = (($35) + 4)|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 $39 = $38; //@line 238 "modules/physics/box2d/Physics.cpp"
 $40 = HEAP32[$39>>2]|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 $41 = $3; //@line 238 "modules/physics/box2d/Physics.cpp"
 $42 = $41; //@line 238 "modules/physics/box2d/Physics.cpp"
 HEAP32[$42>>2] = $37; //@line 238 "modules/physics/box2d/Physics.cpp"
 $43 = (($41) + 4)|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 $44 = $43; //@line 238 "modules/physics/box2d/Physics.cpp"
 HEAP32[$44>>2] = $40; //@line 238 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 ;HEAP32[$tmpcast$byval_copy>>2]=HEAP32[$0>>2]|0;HEAP32[$tmpcast$byval_copy+4>>2]=HEAP32[$0+4>>2]|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 ;HEAP32[$tmpcast1$byval_copy>>2]=HEAP32[$1>>2]|0;HEAP32[$tmpcast1$byval_copy+4>>2]=HEAP32[$1+4>>2]|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 ;HEAP32[$tmpcast2$byval_copy>>2]=HEAP32[$2>>2]|0;HEAP32[$tmpcast2$byval_copy+4>>2]=HEAP32[$2+4>>2]|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 ;HEAP32[$tmpcast3$byval_copy>>2]=HEAP32[$3>>2]|0;HEAP32[$tmpcast3$byval_copy+4>>2]=HEAP32[$3+4>>2]|0; //@line 238 "modules/physics/box2d/Physics.cpp"
 invoke_viiiiiiidi(3684,($4|0),($body1|0),($body2|0),($tmpcast$byval_copy|0),($tmpcast1$byval_copy|0),($tmpcast2$byval_copy|0),($tmpcast3$byval_copy|0),(+$ratio),($collideConnected|0)); //@line 238 "modules/physics/box2d/Physics.cpp"
 $45 = __THREW__; __THREW__ = 0;
 $46 = $45&1;
 if ($46) {
  $47 = ___cxa_find_matching_catch_2()|0;
  $48 = tempRet0;
  __ZdlPv($4); //@line 238 "modules/physics/box2d/Physics.cpp"
  ___resumeException($47|0);
  // unreachable;
 } else {
  STACKTOP = sp;return ($4|0); //@line 238 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics12newGearJointEPNS1_5JointES4_fb($this,$joint1,$joint2,$ratio,$collideConnected) {
 $this = $this|0;
 $joint1 = $joint1|0;
 $joint2 = $joint2|0;
 $ratio = +$ratio;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(32)|0); //@line 243 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viiidi(3685,($0|0),($joint1|0),($joint2|0),(+$ratio),($collideConnected|0)); //@line 243 "modules/physics/box2d/Physics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 243 "modules/physics/box2d/Physics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 243 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics16newFrictionJointEPNS1_4BodyES4_ffffb($this,$body1,$body2,$xA,$yA,$xB,$yB,$collideConnected) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 $xA = +$xA;
 $yA = +$yA;
 $xB = +$xB;
 $yB = +$yB;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(32)|0); //@line 248 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viiiddddi(3686,($0|0),($body1|0),($body2|0),(+$xA),(+$yA),(+$xB),(+$yB),($collideConnected|0)); //@line 248 "modules/physics/box2d/Physics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 248 "modules/physics/box2d/Physics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 248 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics12newWeldJointEPNS1_4BodyES4_ffffb($this,$body1,$body2,$xA,$yA,$xB,$yB,$collideConnected) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 $xA = +$xA;
 $yA = +$yA;
 $xB = +$xB;
 $yB = +$yB;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(32)|0); //@line 253 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viiiddddi(3687,($0|0),($body1|0),($body2|0),(+$xA),(+$yA),(+$xB),(+$yB),($collideConnected|0)); //@line 253 "modules/physics/box2d/Physics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 253 "modules/physics/box2d/Physics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 253 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics13newWheelJointEPNS1_4BodyES4_ffffffb($this,$body1,$body2,$xA,$yA,$xB,$yB,$ax,$ay,$collideConnected) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 $xA = +$xA;
 $yA = +$yA;
 $xB = +$xB;
 $yB = +$yB;
 $ax = +$ax;
 $ay = +$ay;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(32)|0); //@line 258 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viiiddddddi(3688,($0|0),($body1|0),($body2|0),(+$xA),(+$yA),(+$xB),(+$yB),(+$ax),(+$ay),($collideConnected|0)); //@line 258 "modules/physics/box2d/Physics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 258 "modules/physics/box2d/Physics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 258 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics12newRopeJointEPNS1_4BodyES4_fffffb($this,$body1,$body2,$x1,$y1,$x2,$y2,$maxLength,$collideConnected) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 $x1 = +$x1;
 $y1 = +$y1;
 $x2 = +$x2;
 $y2 = +$y2;
 $maxLength = +$maxLength;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(32)|0); //@line 263 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viiidddddi(3689,($0|0),($body1|0),($body2|0),(+$x1),(+$y1),(+$x2),(+$y2),(+$maxLength),($collideConnected|0)); //@line 263 "modules/physics/box2d/Physics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 263 "modules/physics/box2d/Physics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 263 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics13newMotorJointEPNS1_4BodyES4_($this,$body1,$body2) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(32)|0); //@line 268 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viii(3690,($0|0),($body1|0),($body2|0)); //@line 268 "modules/physics/box2d/Physics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 268 "modules/physics/box2d/Physics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 268 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics13newMotorJointEPNS1_4BodyES4_fb($this,$body1,$body2,$correctionFactor,$collideConnected) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 $correctionFactor = +$correctionFactor;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(32)|0); //@line 273 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viiidi(3691,($0|0),($body1|0),($body2|0),(+$correctionFactor),($collideConnected|0)); //@line 273 "modules/physics/box2d/Physics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 273 "modules/physics/box2d/Physics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 273 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics10newFixtureEPNS1_4BodyEPNS1_5ShapeEf($this,$body,$shape,$density) {
 $this = $this|0;
 $body = $body|0;
 $shape = $shape|0;
 $density = +$density;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(20)|0); //@line 279 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viiid(3692,($0|0),($body|0),($shape|0),(+$density)); //@line 279 "modules/physics/box2d/Physics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 279 "modules/physics/box2d/Physics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 279 "modules/physics/box2d/Physics.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Physics11getDistanceEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0.0;
 var $61 = 0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0.0, $78 = 0, $79 = 0.0;
 var $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $9 = 0, $c = 0, $i = 0, $o = 0, $pA = 0, $pB = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $pA = sp + 160|0;
 $pB = sp + 132|0;
 $i = sp + 40|0;
 $o = sp + 16|0;
 $c = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 284 "modules/physics/box2d/Physics.cpp"
 $1 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,2,35)|0); //@line 285 "modules/physics/box2d/Physics.cpp"
 $2 = ((($pA)) + 16|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAP32[$2>>2] = 0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $3 = ((($pA)) + 20|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAP32[$3>>2] = 0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $4 = ((($pA)) + 24|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAPF32[$4>>2] = 0.0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $5 = ((($pB)) + 16|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAP32[$5>>2] = 0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $6 = ((($pB)) + 20|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAP32[$6>>2] = 0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $7 = ((($pB)) + 24|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAPF32[$7>>2] = 0.0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $8 = ((($i)) + 16|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAP32[$8>>2] = 0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $9 = ((($i)) + 20|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAP32[$9>>2] = 0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $10 = ((($i)) + 24|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAPF32[$10>>2] = 0.0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $11 = ((($i)) + 44|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAP32[$11>>2] = 0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $12 = ((($i)) + 48|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAP32[$12>>2] = 0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $13 = ((($i)) + 52|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAPF32[$13>>2] = 0.0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $14 = ((($c)) + 4|0); //@line 290 "modules/physics/box2d/Physics.cpp"
 HEAP16[$14>>1] = 0; //@line 290 "modules/physics/box2d/Physics.cpp"
 $15 = ((($0)) + 16|0); //@line 293 "modules/physics/box2d/Physics.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 293 "modules/physics/box2d/Physics.cpp"
 $17 = ((($16)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $18 = HEAP32[$17>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 __THREW__ = 0;
 invoke_viii(3693,($pA|0),($18|0),0); //@line 293 "modules/physics/box2d/Physics.cpp"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  label = 4;
 } else {
  $21 = ((($1)) + 16|0); //@line 294 "modules/physics/box2d/Physics.cpp"
  $22 = HEAP32[$21>>2]|0; //@line 294 "modules/physics/box2d/Physics.cpp"
  $23 = ((($22)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $24 = HEAP32[$23>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
  __THREW__ = 0;
  invoke_viii(3693,($pB|0),($24|0),0); //@line 294 "modules/physics/box2d/Physics.cpp"
  $25 = __THREW__; __THREW__ = 0;
  $26 = $25&1;
  if ($26) {
   label = 4;
  } else {
   ;HEAP32[$i>>2]=HEAP32[$pA>>2]|0;HEAP32[$i+4>>2]=HEAP32[$pA+4>>2]|0;HEAP32[$i+8>>2]=HEAP32[$pA+8>>2]|0;HEAP32[$i+12>>2]=HEAP32[$pA+12>>2]|0;HEAP32[$i+16>>2]=HEAP32[$pA+16>>2]|0;HEAP32[$i+20>>2]=HEAP32[$pA+20>>2]|0;HEAP32[$i+24>>2]=HEAP32[$pA+24>>2]|0; //@line 295 "modules/physics/box2d/Physics.cpp"
   $27 = ((($i)) + 28|0); //@line 296 "modules/physics/box2d/Physics.cpp"
   ;HEAP32[$27>>2]=HEAP32[$pB>>2]|0;HEAP32[$27+4>>2]=HEAP32[$pB+4>>2]|0;HEAP32[$27+8>>2]=HEAP32[$pB+8>>2]|0;HEAP32[$27+12>>2]=HEAP32[$pB+12>>2]|0;HEAP32[$27+16>>2]=HEAP32[$pB+16>>2]|0;HEAP32[$27+20>>2]=HEAP32[$pB+20>>2]|0;HEAP32[$27+24>>2]=HEAP32[$pB+24>>2]|0; //@line 296 "modules/physics/box2d/Physics.cpp"
   $28 = ((($i)) + 56|0); //@line 297 "modules/physics/box2d/Physics.cpp"
   $29 = HEAP32[$15>>2]|0; //@line 297 "modules/physics/box2d/Physics.cpp"
   $30 = ((($29)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
   $31 = HEAP32[$30>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
   $32 = ((($31)) + 12|0); //@line 476 "./libraries/Box2D/Dynamics/b2Body.h"
   ;HEAP32[$28>>2]=HEAP32[$32>>2]|0;HEAP32[$28+4>>2]=HEAP32[$32+4>>2]|0;HEAP32[$28+8>>2]=HEAP32[$32+8>>2]|0;HEAP32[$28+12>>2]=HEAP32[$32+12>>2]|0; //@line 297 "modules/physics/box2d/Physics.cpp"
   $33 = ((($i)) + 72|0); //@line 298 "modules/physics/box2d/Physics.cpp"
   $34 = HEAP32[$21>>2]|0; //@line 298 "modules/physics/box2d/Physics.cpp"
   $35 = ((($34)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
   $36 = HEAP32[$35>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
   $37 = ((($36)) + 12|0); //@line 476 "./libraries/Box2D/Dynamics/b2Body.h"
   ;HEAP32[$33>>2]=HEAP32[$37>>2]|0;HEAP32[$33+4>>2]=HEAP32[$37+4>>2]|0;HEAP32[$33+8>>2]=HEAP32[$37+8>>2]|0;HEAP32[$33+12>>2]=HEAP32[$37+12>>2]|0; //@line 298 "modules/physics/box2d/Physics.cpp"
   $38 = ((($i)) + 88|0); //@line 299 "modules/physics/box2d/Physics.cpp"
   HEAP8[$38>>0] = 1; //@line 299 "modules/physics/box2d/Physics.cpp"
   __THREW__ = 0;
   invoke_viii(3694,($o|0),($c|0),($i|0)); //@line 300 "modules/physics/box2d/Physics.cpp"
   $39 = __THREW__; __THREW__ = 0;
   $40 = $39&1;
   if ($40) {
    label = 4;
   }
  }
 }
 do {
  if ((label|0) == 4) {
   $41 = ___cxa_find_matching_catch_3(8992|0)|0;
   $42 = tempRet0;
   $43 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $44 = ($42|0)==($43|0); //@line 535 "./common/runtime.h"
   if (!($44)) {
    $$01$i = $42;$$02$i = $41;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $45 = (___cxa_begin_catch(($41|0))|0); //@line 535 "./common/runtime.h"
   $46 = HEAP32[$45>>2]|0; //@line 539 "./common/runtime.h"
   $47 = ((($46)) + 8|0); //@line 539 "./common/runtime.h"
   $48 = HEAP32[$47>>2]|0; //@line 539 "./common/runtime.h"
   $49 = (FUNCTION_TABLE_ii[$48 & 8191]($45)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($49|0)); //@line 539 "./common/runtime.h"
   $50 = __THREW__; __THREW__ = 0;
   $51 = $50&1;
   if (!($51)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $52 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $52; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    break;
   }
   $53 = ___cxa_find_matching_catch_2()|0;
   $54 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $55 = __THREW__; __THREW__ = 0;
   $56 = $55&1;
   if ($56) {
    $57 = ___cxa_find_matching_catch_3(0|0)|0;
    $58 = tempRet0;
    ___clang_call_terminate($57); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $54;$$02$i = $53;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  }
 } while(0);
 $59 = ((($o)) + 16|0); //@line 303 "modules/physics/box2d/Physics.cpp"
 $60 = +HEAPF32[$59>>2]; //@line 303 "modules/physics/box2d/Physics.cpp"
 $61 = HEAP32[4273]|0; //@line 341 "modules/physics/box2d/Physics.cpp"
 $62 = (+($61|0)); //@line 341 "modules/physics/box2d/Physics.cpp"
 $63 = $60 * $62; //@line 341 "modules/physics/box2d/Physics.cpp"
 $64 = $63; //@line 303 "modules/physics/box2d/Physics.cpp"
 _lua_pushnumber($L,$64); //@line 303 "modules/physics/box2d/Physics.cpp"
 $65 = +HEAPF32[$o>>2]; //@line 304 "modules/physics/box2d/Physics.cpp"
 $66 = HEAP32[4273]|0; //@line 341 "modules/physics/box2d/Physics.cpp"
 $67 = (+($66|0)); //@line 341 "modules/physics/box2d/Physics.cpp"
 $68 = $65 * $67; //@line 341 "modules/physics/box2d/Physics.cpp"
 $69 = $68; //@line 304 "modules/physics/box2d/Physics.cpp"
 _lua_pushnumber($L,$69); //@line 304 "modules/physics/box2d/Physics.cpp"
 $70 = ((($o)) + 4|0); //@line 305 "modules/physics/box2d/Physics.cpp"
 $71 = +HEAPF32[$70>>2]; //@line 305 "modules/physics/box2d/Physics.cpp"
 $72 = HEAP32[4273]|0; //@line 341 "modules/physics/box2d/Physics.cpp"
 $73 = (+($72|0)); //@line 341 "modules/physics/box2d/Physics.cpp"
 $74 = $71 * $73; //@line 341 "modules/physics/box2d/Physics.cpp"
 $75 = $74; //@line 305 "modules/physics/box2d/Physics.cpp"
 _lua_pushnumber($L,$75); //@line 305 "modules/physics/box2d/Physics.cpp"
 $76 = ((($o)) + 8|0); //@line 306 "modules/physics/box2d/Physics.cpp"
 $77 = +HEAPF32[$76>>2]; //@line 306 "modules/physics/box2d/Physics.cpp"
 $78 = HEAP32[4273]|0; //@line 341 "modules/physics/box2d/Physics.cpp"
 $79 = (+($78|0)); //@line 341 "modules/physics/box2d/Physics.cpp"
 $80 = $77 * $79; //@line 341 "modules/physics/box2d/Physics.cpp"
 $81 = $80; //@line 306 "modules/physics/box2d/Physics.cpp"
 _lua_pushnumber($L,$81); //@line 306 "modules/physics/box2d/Physics.cpp"
 $82 = ((($o)) + 12|0); //@line 307 "modules/physics/box2d/Physics.cpp"
 $83 = +HEAPF32[$82>>2]; //@line 307 "modules/physics/box2d/Physics.cpp"
 $84 = HEAP32[4273]|0; //@line 341 "modules/physics/box2d/Physics.cpp"
 $85 = (+($84|0)); //@line 341 "modules/physics/box2d/Physics.cpp"
 $86 = $83 * $85; //@line 341 "modules/physics/box2d/Physics.cpp"
 $87 = $86; //@line 307 "modules/physics/box2d/Physics.cpp"
 _lua_pushnumber($L,$87); //@line 307 "modules/physics/box2d/Physics.cpp"
 STACKTOP = sp;return 5; //@line 308 "modules/physics/box2d/Physics.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d7Physics7scaleUpEf($f) {
 $f = +$f;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[4273]|0; //@line 341 "modules/physics/box2d/Physics.cpp"
 $1 = (+($0|0)); //@line 341 "modules/physics/box2d/Physics.cpp"
 $2 = $1 * $f; //@line 341 "modules/physics/box2d/Physics.cpp"
 return (+$2); //@line 341 "modules/physics/box2d/Physics.cpp"
}
function __ZN4love7physics5box2d7Physics8setMeterEi($scale) {
 $scale = $scale|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 311 "modules/physics/box2d/Physics.cpp"
 $0 = ($scale|0)<(1); //@line 313 "modules/physics/box2d/Physics.cpp"
 if (!($0)) {
  HEAP32[4273] = $scale; //@line 314 "modules/physics/box2d/Physics.cpp"
  STACKTOP = sp;return; //@line 315 "modules/physics/box2d/Physics.cpp"
 }
 $1 = (___cxa_allocate_exception(16)|0); //@line 313 "modules/physics/box2d/Physics.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($1|0),(356386|0),($vararg_buffer|0)); //@line 313 "modules/physics/box2d/Physics.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  ___cxa_free_exception(($1|0)); //@line 313 "modules/physics/box2d/Physics.cpp"
  ___resumeException($4|0);
  // unreachable;
 } else {
  ___cxa_throw(($1|0),(96|0),(344|0)); //@line 313 "modules/physics/box2d/Physics.cpp"
  // unreachable; //@line 313 "modules/physics/box2d/Physics.cpp"
 }
}
function __ZN4love7physics5box2d7Physics8getMeterEv() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[4273]|0; //@line 319 "modules/physics/box2d/Physics.cpp"
 return ($0|0); //@line 319 "modules/physics/box2d/Physics.cpp"
}
function __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($agg$result,$v) {
 $agg$result = $agg$result|0;
 $v = $v|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $bitcast = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $v; //@line 353 "modules/physics/box2d/Physics.cpp"
 $1 = $0; //@line 353 "modules/physics/box2d/Physics.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 353 "modules/physics/box2d/Physics.cpp"
 $3 = (($0) + 4)|0; //@line 353 "modules/physics/box2d/Physics.cpp"
 $4 = $3; //@line 353 "modules/physics/box2d/Physics.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 353 "modules/physics/box2d/Physics.cpp"
 $6 = $agg$result; //@line 353 "modules/physics/box2d/Physics.cpp"
 $7 = $6; //@line 353 "modules/physics/box2d/Physics.cpp"
 HEAP32[$7>>2] = $2; //@line 353 "modules/physics/box2d/Physics.cpp"
 $8 = (($6) + 4)|0; //@line 353 "modules/physics/box2d/Physics.cpp"
 $9 = $8; //@line 353 "modules/physics/box2d/Physics.cpp"
 HEAP32[$9>>2] = $5; //@line 353 "modules/physics/box2d/Physics.cpp"
 $10 = ((($agg$result)) + 4|0); //@line 354 "modules/physics/box2d/Physics.cpp"
 $11 = HEAP32[4273]|0; //@line 330 "modules/physics/box2d/Physics.cpp"
 $12 = (+($11|0)); //@line 330 "modules/physics/box2d/Physics.cpp"
 $bitcast = (HEAP32[tempDoublePtr>>2]=$2,+HEAPF32[tempDoublePtr>>2]); //@line 330 "modules/physics/box2d/Physics.cpp"
 $13 = $bitcast * $12; //@line 330 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$agg$result>>2] = $13; //@line 330 "modules/physics/box2d/Physics.cpp"
 $14 = (HEAP32[tempDoublePtr>>2]=$5,+HEAPF32[tempDoublePtr>>2]); //@line 331 "modules/physics/box2d/Physics.cpp"
 $15 = $12 * $14; //@line 331 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$10>>2] = $15; //@line 331 "modules/physics/box2d/Physics.cpp"
 return; //@line 355 "modules/physics/box2d/Physics.cpp"
}
function __ZN4love7physics5box2d7Physics7scaleUpERK6b2AABB($agg$result,$aabb) {
 $agg$result = $agg$result|0;
 $aabb = $aabb|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $bitcast$i = 0.0, $bitcast$i3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $aabb; //@line 353 "modules/physics/box2d/Physics.cpp"
 $1 = $0; //@line 353 "modules/physics/box2d/Physics.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 353 "modules/physics/box2d/Physics.cpp"
 $3 = (($0) + 4)|0; //@line 353 "modules/physics/box2d/Physics.cpp"
 $4 = $3; //@line 353 "modules/physics/box2d/Physics.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 353 "modules/physics/box2d/Physics.cpp"
 $6 = HEAP32[4273]|0; //@line 330 "modules/physics/box2d/Physics.cpp"
 $7 = (+($6|0)); //@line 330 "modules/physics/box2d/Physics.cpp"
 $bitcast$i3 = (HEAP32[tempDoublePtr>>2]=$2,+HEAPF32[tempDoublePtr>>2]); //@line 330 "modules/physics/box2d/Physics.cpp"
 $8 = $bitcast$i3 * $7; //@line 330 "modules/physics/box2d/Physics.cpp"
 $9 = (HEAP32[tempDoublePtr>>2]=$5,+HEAPF32[tempDoublePtr>>2]); //@line 331 "modules/physics/box2d/Physics.cpp"
 $10 = $7 * $9; //@line 331 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$agg$result>>2] = $8; //@line 369 "modules/physics/box2d/Physics.cpp"
 $11 = ((($agg$result)) + 4|0); //@line 369 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$11>>2] = $10; //@line 369 "modules/physics/box2d/Physics.cpp"
 $12 = ((($aabb)) + 8|0); //@line 370 "modules/physics/box2d/Physics.cpp"
 $13 = $12; //@line 353 "modules/physics/box2d/Physics.cpp"
 $14 = $13; //@line 353 "modules/physics/box2d/Physics.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 353 "modules/physics/box2d/Physics.cpp"
 $16 = (($13) + 4)|0; //@line 353 "modules/physics/box2d/Physics.cpp"
 $17 = $16; //@line 353 "modules/physics/box2d/Physics.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 353 "modules/physics/box2d/Physics.cpp"
 $bitcast$i = (HEAP32[tempDoublePtr>>2]=$15,+HEAPF32[tempDoublePtr>>2]); //@line 330 "modules/physics/box2d/Physics.cpp"
 $19 = $bitcast$i * $7; //@line 330 "modules/physics/box2d/Physics.cpp"
 $20 = (HEAP32[tempDoublePtr>>2]=$18,+HEAPF32[tempDoublePtr>>2]); //@line 331 "modules/physics/box2d/Physics.cpp"
 $21 = $7 * $20; //@line 331 "modules/physics/box2d/Physics.cpp"
 $22 = ((($agg$result)) + 8|0); //@line 370 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$22>>2] = $19; //@line 370 "modules/physics/box2d/Physics.cpp"
 $23 = ((($agg$result)) + 12|0); //@line 370 "modules/physics/box2d/Physics.cpp"
 HEAPF32[$23>>2] = $21; //@line 370 "modules/physics/box2d/Physics.cpp"
 return; //@line 371 "modules/physics/box2d/Physics.cpp"
}
function __ZN4love7physics5box2d12PolygonShapeD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5ShapeD2Ev($this); //@line 44 "modules/physics/box2d/PolygonShape.cpp"
 return; //@line 44 "modules/physics/box2d/PolygonShape.cpp"
}
function __ZN4love7physics5box2d12PolygonShapeD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5ShapeD2Ev($this); //@line 44 "modules/physics/box2d/PolygonShape.cpp"
 __ZdlPv($this); //@line 43 "modules/physics/box2d/PolygonShape.cpp"
 return; //@line 44 "modules/physics/box2d/PolygonShape.cpp"
}
function __ZN4love7physics5box2d12PolygonShapeC2EP14b2PolygonShapeb($this,$p,$own) {
 $this = $this|0;
 $p = $p|0;
 $own = $own|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5ShapeC2EP7b2Shapeb($this,$p,$own); //@line 38 "modules/physics/box2d/PolygonShape.cpp"
 HEAP32[$this>>2] = (17136); //@line 39 "modules/physics/box2d/PolygonShape.cpp"
 return; //@line 40 "modules/physics/box2d/PolygonShape.cpp"
}
function __ZN4love7physics5box2d12PolygonShape9getPointsEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp;
 (__ZN4love16luax_assert_argcEP9lua_Statei($L,0)|0); //@line 48 "modules/physics/box2d/PolygonShape.cpp"
 $0 = ((($this)) + 8|0); //@line 49 "modules/physics/box2d/PolygonShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 49 "modules/physics/box2d/PolygonShape.cpp"
 $2 = ((($1)) + 148|0); //@line 72 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $3 = HEAP32[$2>>2]|0; //@line 72 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $4 = ($3|0)>(0); //@line 51 "modules/physics/box2d/PolygonShape.cpp"
 if (!($4)) {
  $6 = $3 << 1; //@line 57 "modules/physics/box2d/PolygonShape.cpp"
  STACKTOP = sp;return ($6|0); //@line 57 "modules/physics/box2d/PolygonShape.cpp"
 }
 $5 = ((($v)) + 4|0); //@line 55 "modules/physics/box2d/PolygonShape.cpp"
 $i$01 = 0;
 while(1) {
  $7 = HEAP32[$2>>2]|0; //@line 97 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
  $8 = ($7|0)>($i$01|0); //@line 97 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
  __Z10loveAssertbPKc($8,445986); //@line 97 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
  $9 = (((($1)) + 20|0) + ($i$01<<3)|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
  __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($v,$9); //@line 53 "modules/physics/box2d/PolygonShape.cpp"
  $10 = +HEAPF32[$v>>2]; //@line 54 "modules/physics/box2d/PolygonShape.cpp"
  $11 = $10; //@line 54 "modules/physics/box2d/PolygonShape.cpp"
  _lua_pushnumber($L,$11); //@line 54 "modules/physics/box2d/PolygonShape.cpp"
  $12 = +HEAPF32[$5>>2]; //@line 55 "modules/physics/box2d/PolygonShape.cpp"
  $13 = $12; //@line 55 "modules/physics/box2d/PolygonShape.cpp"
  _lua_pushnumber($L,$13); //@line 55 "modules/physics/box2d/PolygonShape.cpp"
  $14 = (($i$01) + 1)|0; //@line 51 "modules/physics/box2d/PolygonShape.cpp"
  $exitcond = ($14|0)==($3|0); //@line 51 "modules/physics/box2d/PolygonShape.cpp"
  if ($exitcond) {
   break;
  } else {
   $i$01 = $14;
  }
 }
 $6 = $3 << 1; //@line 57 "modules/physics/box2d/PolygonShape.cpp"
 STACKTOP = sp;return ($6|0); //@line 57 "modules/physics/box2d/PolygonShape.cpp"
}
function __ZNK4love7physics5box2d12PolygonShape8validateEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 62 "modules/physics/box2d/PolygonShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 62 "modules/physics/box2d/PolygonShape.cpp"
 $2 = (__ZNK14b2PolygonShape8ValidateEv($1)|0); //@line 63 "modules/physics/box2d/PolygonShape.cpp"
 return ($2|0); //@line 63 "modules/physics/box2d/PolygonShape.cpp"
}
function __ZN4love7physics5box2d14PrismaticJointD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 52 "modules/physics/box2d/PrismaticJoint.cpp"
 return; //@line 52 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d14PrismaticJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 52 "modules/physics/box2d/PrismaticJoint.cpp"
 __ZdlPv($this); //@line 51 "modules/physics/box2d/PrismaticJoint.cpp"
 return; //@line 52 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d14PrismaticJointC2EPNS1_4BodyES4_ffffffb($this,$body1,$body2,$xA,$yA,$xB,$yB,$ax,$ay,$collideConnected) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 $xA = +$xA;
 $yA = +$yA;
 $xB = +$xB;
 $yB = +$yB;
 $ax = +$ax;
 $ay = +$ay;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0.0, $44 = 0.0;
 var $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $def = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $def = sp + 40|0;
 $0 = sp + 32|0;
 $1 = sp + 24|0;
 $2 = sp + 16|0;
 $3 = sp + 8|0;
 $4 = sp;
 __ZN4love7physics5box2d5JointC2EPNS1_4BodyES4_($this,$body1,$body2); //@line 36 "modules/physics/box2d/PrismaticJoint.cpp"
 HEAP32[$this>>2] = (17160); //@line 38 "modules/physics/box2d/PrismaticJoint.cpp"
 $5 = ((($this)) + 28|0); //@line 37 "modules/physics/box2d/PrismaticJoint.cpp"
 HEAP32[$5>>2] = 0; //@line 37 "modules/physics/box2d/PrismaticJoint.cpp"
 ;HEAP32[$def>>2]=0|0;HEAP32[$def+4>>2]=0|0;HEAP32[$def+8>>2]=0|0;HEAP32[$def+12>>2]=0|0;HEAP8[$def+16>>0]=0|0; //@line 79 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 HEAP32[$def>>2] = 2; //@line 34 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 $6 = ((($def)) + 20|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($def)) + 36|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;HEAP32[$6+12>>2]=0|0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$7>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($def)) + 40|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$8>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($def)) + 44|0); //@line 38 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 HEAPF32[$9>>2] = 0.0; //@line 38 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 $10 = ((($def)) + 48|0); //@line 39 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 HEAP8[$10>>0] = 0; //@line 39 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 $11 = ((($def)) + 52|0); //@line 40 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 HEAPF32[$11>>2] = 0.0; //@line 40 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 $12 = ((($def)) + 56|0); //@line 41 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 HEAPF32[$12>>2] = 0.0; //@line 41 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 $13 = ((($def)) + 60|0); //@line 42 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 HEAP8[$13>>0] = 0; //@line 42 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 $14 = ((($def)) + 64|0); //@line 43 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 HEAPF32[$14>>2] = 0.0; //@line 43 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 $15 = ((($def)) + 68|0); //@line 44 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 HEAPF32[$15>>2] = 0.0; //@line 44 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 $16 = ((($body1)) + 8|0); //@line 41 "modules/physics/box2d/PrismaticJoint.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 41 "modules/physics/box2d/PrismaticJoint.cpp"
 $18 = ((($body2)) + 8|0); //@line 41 "modules/physics/box2d/PrismaticJoint.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 41 "modules/physics/box2d/PrismaticJoint.cpp"
 HEAPF32[$1>>2] = $xA; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $20 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$20>>2] = $yA; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 invoke_vii(3695,($0|0),($1|0)); //@line 41 "modules/physics/box2d/PrismaticJoint.cpp"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $56 = ___cxa_find_matching_catch_2()|0;
  $57 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 48 "modules/physics/box2d/PrismaticJoint.cpp"
  ___resumeException($56|0);
  // unreachable;
 }
 HEAPF32[$2>>2] = $ax; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $23 = ((($2)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$23>>2] = $ay; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 invoke_viiiii(3696,($def|0),($17|0),($19|0),($0|0),($2|0)); //@line 41 "modules/physics/box2d/PrismaticJoint.cpp"
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $56 = ___cxa_find_matching_catch_2()|0;
  $57 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 48 "modules/physics/box2d/PrismaticJoint.cpp"
  ___resumeException($56|0);
  // unreachable;
 }
 $26 = HEAP32[$18>>2]|0; //@line 42 "modules/physics/box2d/PrismaticJoint.cpp"
 HEAPF32[$4>>2] = $xB; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($4)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$27>>2] = $yB; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 invoke_vii(3695,($3|0),($4|0)); //@line 42 "modules/physics/box2d/PrismaticJoint.cpp"
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if ($29) {
  $56 = ___cxa_find_matching_catch_2()|0;
  $57 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 48 "modules/physics/box2d/PrismaticJoint.cpp"
  ___resumeException($56|0);
  // unreachable;
 }
 $30 = +HEAPF32[$3>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $31 = ((($26)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $32 = +HEAPF32[$31>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $33 = $30 - $32; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $34 = ((($3)) + 4|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $35 = +HEAPF32[$34>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $36 = ((($26)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $37 = +HEAPF32[$36>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $38 = $35 - $37; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $39 = ((($26)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $40 = +HEAPF32[$39>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $41 = $33 * $40; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $42 = ((($26)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $43 = +HEAPF32[$42>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $44 = $38 * $43; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $45 = $41 + $44; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $46 = $33 * $43; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $47 = $40 * $38; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $48 = $47 - $46; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $49 = ((($def)) + 28|0); //@line 42 "modules/physics/box2d/PrismaticJoint.cpp"
 HEAPF32[$49>>2] = $45; //@line 42 "modules/physics/box2d/PrismaticJoint.cpp"
 $50 = ((($def)) + 32|0); //@line 42 "modules/physics/box2d/PrismaticJoint.cpp"
 HEAPF32[$50>>2] = $48; //@line 42 "modules/physics/box2d/PrismaticJoint.cpp"
 HEAPF32[$11>>2] = 0.0; //@line 43 "modules/physics/box2d/PrismaticJoint.cpp"
 HEAPF32[$12>>2] = 100.0; //@line 44 "modules/physics/box2d/PrismaticJoint.cpp"
 HEAP8[$10>>0] = 1; //@line 45 "modules/physics/box2d/PrismaticJoint.cpp"
 $51 = ((($def)) + 16|0); //@line 46 "modules/physics/box2d/PrismaticJoint.cpp"
 $52 = $collideConnected&1; //@line 46 "modules/physics/box2d/PrismaticJoint.cpp"
 HEAP8[$51>>0] = $52; //@line 46 "modules/physics/box2d/PrismaticJoint.cpp"
 __THREW__ = 0;
 $53 = (invoke_iii(3697,($this|0),($def|0))|0); //@line 47 "modules/physics/box2d/PrismaticJoint.cpp"
 $54 = __THREW__; __THREW__ = 0;
 $55 = $54&1;
 if ($55) {
  $56 = ___cxa_find_matching_catch_2()|0;
  $57 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 48 "modules/physics/box2d/PrismaticJoint.cpp"
  ___resumeException($56|0);
  // unreachable;
 } else {
  HEAP32[$5>>2] = $53; //@line 47 "modules/physics/box2d/PrismaticJoint.cpp"
  STACKTOP = sp;return; //@line 48 "modules/physics/box2d/PrismaticJoint.cpp"
 }
}
function __ZNK4love7physics5box2d14PrismaticJoint19getJointTranslationEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 56 "modules/physics/box2d/PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 56 "modules/physics/box2d/PrismaticJoint.cpp"
 $2 = (+__ZNK16b2PrismaticJoint19GetJointTranslationEv($1)); //@line 56 "modules/physics/box2d/PrismaticJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 56 "modules/physics/box2d/PrismaticJoint.cpp"
 return (+$3); //@line 56 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZNK4love7physics5box2d14PrismaticJoint13getJointSpeedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 61 "modules/physics/box2d/PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 61 "modules/physics/box2d/PrismaticJoint.cpp"
 $2 = (+__ZNK16b2PrismaticJoint13GetJointSpeedEv($1)); //@line 61 "modules/physics/box2d/PrismaticJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 61 "modules/physics/box2d/PrismaticJoint.cpp"
 return (+$3); //@line 61 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d14PrismaticJoint15setMotorEnabledEb($this,$enable) {
 $this = $this|0;
 $enable = $enable|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 66 "modules/physics/box2d/PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 66 "modules/physics/box2d/PrismaticJoint.cpp"
 __ZN16b2PrismaticJoint11EnableMotorEb($1,$enable); //@line 66 "modules/physics/box2d/PrismaticJoint.cpp"
 return; //@line 66 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZNK4love7physics5box2d14PrismaticJoint14isMotorEnabledEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 71 "modules/physics/box2d/PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 71 "modules/physics/box2d/PrismaticJoint.cpp"
 $2 = (__ZNK16b2PrismaticJoint14IsMotorEnabledEv($1)|0); //@line 71 "modules/physics/box2d/PrismaticJoint.cpp"
 return ($2|0); //@line 71 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d14PrismaticJoint16setMaxMotorForceEf($this,$force) {
 $this = $this|0;
 $force = +$force;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 76 "modules/physics/box2d/PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 76 "modules/physics/box2d/PrismaticJoint.cpp"
 $2 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($force)); //@line 76 "modules/physics/box2d/PrismaticJoint.cpp"
 __ZN16b2PrismaticJoint16SetMaxMotorForceEf($1,$2); //@line 76 "modules/physics/box2d/PrismaticJoint.cpp"
 return; //@line 77 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d14PrismaticJoint13setMotorSpeedEf($this,$speed) {
 $this = $this|0;
 $speed = +$speed;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 81 "modules/physics/box2d/PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 81 "modules/physics/box2d/PrismaticJoint.cpp"
 $2 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($speed)); //@line 81 "modules/physics/box2d/PrismaticJoint.cpp"
 __ZN16b2PrismaticJoint13SetMotorSpeedEf($1,$2); //@line 81 "modules/physics/box2d/PrismaticJoint.cpp"
 return; //@line 82 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZNK4love7physics5box2d14PrismaticJoint13getMotorSpeedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 86 "modules/physics/box2d/PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 86 "modules/physics/box2d/PrismaticJoint.cpp"
 $2 = ((($1)) + 132|0); //@line 193 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 $3 = +HEAPF32[$2>>2]; //@line 193 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 $4 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($3)); //@line 86 "modules/physics/box2d/PrismaticJoint.cpp"
 return (+$4); //@line 86 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZNK4love7physics5box2d14PrismaticJoint13getMotorForceEf($this,$inv_dt) {
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 91 "modules/physics/box2d/PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 91 "modules/physics/box2d/PrismaticJoint.cpp"
 $2 = (+__ZNK16b2PrismaticJoint13GetMotorForceEf($1,$inv_dt)); //@line 91 "modules/physics/box2d/PrismaticJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 91 "modules/physics/box2d/PrismaticJoint.cpp"
 return (+$3); //@line 91 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZNK4love7physics5box2d14PrismaticJoint16getMaxMotorForceEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 96 "modules/physics/box2d/PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 96 "modules/physics/box2d/PrismaticJoint.cpp"
 $2 = ((($1)) + 128|0); //@line 142 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 $3 = +HEAPF32[$2>>2]; //@line 142 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 $4 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($3)); //@line 96 "modules/physics/box2d/PrismaticJoint.cpp"
 return (+$4); //@line 96 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d14PrismaticJoint16setLimitsEnabledEb($this,$enable) {
 $this = $this|0;
 $enable = $enable|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 101 "modules/physics/box2d/PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 101 "modules/physics/box2d/PrismaticJoint.cpp"
 __ZN16b2PrismaticJoint11EnableLimitEb($1,$enable); //@line 101 "modules/physics/box2d/PrismaticJoint.cpp"
 return; //@line 102 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZNK4love7physics5box2d14PrismaticJoint16hasLimitsEnabledEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 106 "modules/physics/box2d/PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 106 "modules/physics/box2d/PrismaticJoint.cpp"
 $2 = (__ZNK16b2PrismaticJoint14IsLimitEnabledEv($1)|0); //@line 106 "modules/physics/box2d/PrismaticJoint.cpp"
 return ($2|0); //@line 106 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d14PrismaticJoint13setUpperLimitEf($this,$limit) {
 $this = $this|0;
 $limit = +$limit;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 111 "modules/physics/box2d/PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 111 "modules/physics/box2d/PrismaticJoint.cpp"
 $2 = (+__ZNK16b2PrismaticJoint13GetLowerLimitEv($1)); //@line 111 "modules/physics/box2d/PrismaticJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($limit)); //@line 111 "modules/physics/box2d/PrismaticJoint.cpp"
 __ZN16b2PrismaticJoint9SetLimitsEff($1,$2,$3); //@line 111 "modules/physics/box2d/PrismaticJoint.cpp"
 return; //@line 112 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d14PrismaticJoint13setLowerLimitEf($this,$limit) {
 $this = $this|0;
 $limit = +$limit;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 116 "modules/physics/box2d/PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 116 "modules/physics/box2d/PrismaticJoint.cpp"
 $2 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($limit)); //@line 116 "modules/physics/box2d/PrismaticJoint.cpp"
 $3 = HEAP32[$0>>2]|0; //@line 116 "modules/physics/box2d/PrismaticJoint.cpp"
 $4 = (+__ZNK16b2PrismaticJoint13GetUpperLimitEv($3)); //@line 116 "modules/physics/box2d/PrismaticJoint.cpp"
 __ZN16b2PrismaticJoint9SetLimitsEff($1,$2,$4); //@line 116 "modules/physics/box2d/PrismaticJoint.cpp"
 return; //@line 117 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d14PrismaticJoint9setLimitsEff($this,$lower,$upper) {
 $this = $this|0;
 $lower = +$lower;
 $upper = +$upper;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 121 "modules/physics/box2d/PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 121 "modules/physics/box2d/PrismaticJoint.cpp"
 $2 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($lower)); //@line 121 "modules/physics/box2d/PrismaticJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($upper)); //@line 121 "modules/physics/box2d/PrismaticJoint.cpp"
 __ZN16b2PrismaticJoint9SetLimitsEff($1,$2,$3); //@line 121 "modules/physics/box2d/PrismaticJoint.cpp"
 return; //@line 122 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZNK4love7physics5box2d14PrismaticJoint13getLowerLimitEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 126 "modules/physics/box2d/PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 126 "modules/physics/box2d/PrismaticJoint.cpp"
 $2 = (+__ZNK16b2PrismaticJoint13GetLowerLimitEv($1)); //@line 126 "modules/physics/box2d/PrismaticJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 126 "modules/physics/box2d/PrismaticJoint.cpp"
 return (+$3); //@line 126 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZNK4love7physics5box2d14PrismaticJoint13getUpperLimitEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 131 "modules/physics/box2d/PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 131 "modules/physics/box2d/PrismaticJoint.cpp"
 $2 = (+__ZNK16b2PrismaticJoint13GetUpperLimitEv($1)); //@line 131 "modules/physics/box2d/PrismaticJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 131 "modules/physics/box2d/PrismaticJoint.cpp"
 return (+$3); //@line 131 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d14PrismaticJoint9getLimitsEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 136 "modules/physics/box2d/PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 136 "modules/physics/box2d/PrismaticJoint.cpp"
 $2 = (+__ZNK16b2PrismaticJoint13GetLowerLimitEv($1)); //@line 136 "modules/physics/box2d/PrismaticJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 136 "modules/physics/box2d/PrismaticJoint.cpp"
 $4 = $3; //@line 136 "modules/physics/box2d/PrismaticJoint.cpp"
 _lua_pushnumber($L,$4); //@line 136 "modules/physics/box2d/PrismaticJoint.cpp"
 $5 = HEAP32[$0>>2]|0; //@line 137 "modules/physics/box2d/PrismaticJoint.cpp"
 $6 = (+__ZNK16b2PrismaticJoint13GetUpperLimitEv($5)); //@line 137 "modules/physics/box2d/PrismaticJoint.cpp"
 $7 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($6)); //@line 137 "modules/physics/box2d/PrismaticJoint.cpp"
 $8 = $7; //@line 137 "modules/physics/box2d/PrismaticJoint.cpp"
 _lua_pushnumber($L,$8); //@line 137 "modules/physics/box2d/PrismaticJoint.cpp"
 return 2; //@line 138 "modules/physics/box2d/PrismaticJoint.cpp"
}
function __ZN4love7physics5box2d11PulleyJointD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 49 "modules/physics/box2d/PulleyJoint.cpp"
 return; //@line 49 "modules/physics/box2d/PulleyJoint.cpp"
}
function __ZN4love7physics5box2d11PulleyJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 49 "modules/physics/box2d/PulleyJoint.cpp"
 __ZdlPv($this); //@line 48 "modules/physics/box2d/PulleyJoint.cpp"
 return; //@line 49 "modules/physics/box2d/PulleyJoint.cpp"
}
function __ZN4love7physics5box2d11PulleyJointC2EPNS1_4BodyES4_6b2Vec2S5_S5_S5_fb($this,$bodyA,$bodyB,$groundAnchorA,$groundAnchorB,$anchorA,$anchorB,$ratio,$collideConnected) {
 $this = $this|0;
 $bodyA = $bodyA|0;
 $bodyB = $bodyB|0;
 $groundAnchorA = $groundAnchorA|0;
 $groundAnchorB = $groundAnchorB|0;
 $anchorA = $anchorA|0;
 $anchorB = $anchorB|0;
 $ratio = +$ratio;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $def = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $def = sp + 32|0;
 $0 = sp + 24|0;
 $1 = sp + 16|0;
 $2 = sp + 8|0;
 $3 = sp;
 __ZN4love7physics5box2d5JointC2EPNS1_4BodyES4_($this,$bodyA,$bodyB); //@line 36 "modules/physics/box2d/PulleyJoint.cpp"
 HEAP32[$this>>2] = (17184); //@line 38 "modules/physics/box2d/PulleyJoint.cpp"
 $4 = ((($this)) + 28|0); //@line 37 "modules/physics/box2d/PulleyJoint.cpp"
 HEAP32[$4>>2] = 0; //@line 37 "modules/physics/box2d/PulleyJoint.cpp"
 ;HEAP32[$def>>2]=0|0;HEAP32[$def+4>>2]=0|0;HEAP32[$def+8>>2]=0|0;HEAP32[$def+12>>2]=0|0; //@line 79 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 HEAP32[$def>>2] = 4; //@line 32 "./libraries/Box2D/Dynamics/Joints/b2PulleyJoint.h"
 $5 = ((($def)) + 20|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$5>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($def)) + 24|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$6>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($def)) + 28|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$7>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($def)) + 32|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$8>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($def)) + 36|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$9>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($def)) + 40|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$10>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($def)) + 44|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$11>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $12 = ((($def)) + 48|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$12>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($def)) + 52|0); //@line 37 "./libraries/Box2D/Dynamics/Joints/b2PulleyJoint.h"
 HEAPF32[$13>>2] = 0.0; //@line 37 "./libraries/Box2D/Dynamics/Joints/b2PulleyJoint.h"
 $14 = ((($def)) + 56|0); //@line 38 "./libraries/Box2D/Dynamics/Joints/b2PulleyJoint.h"
 HEAPF32[$14>>2] = 0.0; //@line 38 "./libraries/Box2D/Dynamics/Joints/b2PulleyJoint.h"
 $15 = ((($def)) + 60|0); //@line 39 "./libraries/Box2D/Dynamics/Joints/b2PulleyJoint.h"
 HEAPF32[$15>>2] = 1.0; //@line 39 "./libraries/Box2D/Dynamics/Joints/b2PulleyJoint.h"
 $16 = ((($def)) + 16|0); //@line 40 "./libraries/Box2D/Dynamics/Joints/b2PulleyJoint.h"
 HEAP8[$16>>0] = 1; //@line 40 "./libraries/Box2D/Dynamics/Joints/b2PulleyJoint.h"
 $17 = ((($bodyA)) + 8|0); //@line 40 "modules/physics/box2d/PulleyJoint.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 40 "modules/physics/box2d/PulleyJoint.cpp"
 $19 = ((($bodyB)) + 8|0); //@line 40 "modules/physics/box2d/PulleyJoint.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 40 "modules/physics/box2d/PulleyJoint.cpp"
 __THREW__ = 0;
 invoke_vii(3695,($0|0),($groundAnchorA|0)); //@line 40 "modules/physics/box2d/PulleyJoint.cpp"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if (!($22)) {
  __THREW__ = 0;
  invoke_vii(3695,($1|0),($groundAnchorB|0)); //@line 40 "modules/physics/box2d/PulleyJoint.cpp"
  $23 = __THREW__; __THREW__ = 0;
  $24 = $23&1;
  if (!($24)) {
   __THREW__ = 0;
   invoke_vii(3695,($2|0),($anchorA|0)); //@line 41 "modules/physics/box2d/PulleyJoint.cpp"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if (!($26)) {
    __THREW__ = 0;
    invoke_vii(3695,($3|0),($anchorB|0)); //@line 41 "modules/physics/box2d/PulleyJoint.cpp"
    $27 = __THREW__; __THREW__ = 0;
    $28 = $27&1;
    if (!($28)) {
     __THREW__ = 0;
     invoke_viiiiiiid(3698,($def|0),($18|0),($20|0),($0|0),($1|0),($2|0),($3|0),(+$ratio)); //@line 40 "modules/physics/box2d/PulleyJoint.cpp"
     $29 = __THREW__; __THREW__ = 0;
     $30 = $29&1;
     if (!($30)) {
      $31 = $collideConnected&1; //@line 42 "modules/physics/box2d/PulleyJoint.cpp"
      HEAP8[$16>>0] = $31; //@line 42 "modules/physics/box2d/PulleyJoint.cpp"
      __THREW__ = 0;
      $32 = (invoke_iii(3697,($this|0),($def|0))|0); //@line 44 "modules/physics/box2d/PulleyJoint.cpp"
      $33 = __THREW__; __THREW__ = 0;
      $34 = $33&1;
      if (!($34)) {
       HEAP32[$4>>2] = $32; //@line 44 "modules/physics/box2d/PulleyJoint.cpp"
       STACKTOP = sp;return; //@line 45 "modules/physics/box2d/PulleyJoint.cpp"
      }
     }
    }
   }
  }
 }
 $35 = ___cxa_find_matching_catch_2()|0;
 $36 = tempRet0;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 45 "modules/physics/box2d/PulleyJoint.cpp"
 ___resumeException($35|0);
 // unreachable;
}
function __ZN4love7physics5box2d11PulleyJoint16getGroundAnchorsEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0;
 var $7 = 0.0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 24|0;
 $1 = sp + 16|0;
 $2 = sp + 8|0;
 $3 = sp;
 $4 = ((($this)) + 28|0); //@line 53 "modules/physics/box2d/PulleyJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 53 "modules/physics/box2d/PulleyJoint.cpp"
 __ZNK13b2PulleyJoint16GetGroundAnchorAEv($0,$5); //@line 53 "modules/physics/box2d/PulleyJoint.cpp"
 $6 = +HEAPF32[$0>>2]; //@line 53 "modules/physics/box2d/PulleyJoint.cpp"
 $7 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($6)); //@line 53 "modules/physics/box2d/PulleyJoint.cpp"
 $8 = $7; //@line 53 "modules/physics/box2d/PulleyJoint.cpp"
 _lua_pushnumber($L,$8); //@line 53 "modules/physics/box2d/PulleyJoint.cpp"
 $9 = HEAP32[$4>>2]|0; //@line 54 "modules/physics/box2d/PulleyJoint.cpp"
 __ZNK13b2PulleyJoint16GetGroundAnchorAEv($1,$9); //@line 54 "modules/physics/box2d/PulleyJoint.cpp"
 $10 = ((($1)) + 4|0); //@line 54 "modules/physics/box2d/PulleyJoint.cpp"
 $11 = +HEAPF32[$10>>2]; //@line 54 "modules/physics/box2d/PulleyJoint.cpp"
 $12 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($11)); //@line 54 "modules/physics/box2d/PulleyJoint.cpp"
 $13 = $12; //@line 54 "modules/physics/box2d/PulleyJoint.cpp"
 _lua_pushnumber($L,$13); //@line 54 "modules/physics/box2d/PulleyJoint.cpp"
 $14 = HEAP32[$4>>2]|0; //@line 55 "modules/physics/box2d/PulleyJoint.cpp"
 __ZNK13b2PulleyJoint16GetGroundAnchorBEv($2,$14); //@line 55 "modules/physics/box2d/PulleyJoint.cpp"
 $15 = +HEAPF32[$2>>2]; //@line 55 "modules/physics/box2d/PulleyJoint.cpp"
 $16 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($15)); //@line 55 "modules/physics/box2d/PulleyJoint.cpp"
 $17 = $16; //@line 55 "modules/physics/box2d/PulleyJoint.cpp"
 _lua_pushnumber($L,$17); //@line 55 "modules/physics/box2d/PulleyJoint.cpp"
 $18 = HEAP32[$4>>2]|0; //@line 56 "modules/physics/box2d/PulleyJoint.cpp"
 __ZNK13b2PulleyJoint16GetGroundAnchorBEv($3,$18); //@line 56 "modules/physics/box2d/PulleyJoint.cpp"
 $19 = ((($3)) + 4|0); //@line 56 "modules/physics/box2d/PulleyJoint.cpp"
 $20 = +HEAPF32[$19>>2]; //@line 56 "modules/physics/box2d/PulleyJoint.cpp"
 $21 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($20)); //@line 56 "modules/physics/box2d/PulleyJoint.cpp"
 $22 = $21; //@line 56 "modules/physics/box2d/PulleyJoint.cpp"
 _lua_pushnumber($L,$22); //@line 56 "modules/physics/box2d/PulleyJoint.cpp"
 STACKTOP = sp;return 4; //@line 57 "modules/physics/box2d/PulleyJoint.cpp"
}
function __ZNK4love7physics5box2d11PulleyJoint10getLengthAEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 62 "modules/physics/box2d/PulleyJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 62 "modules/physics/box2d/PulleyJoint.cpp"
 $2 = (+__ZNK13b2PulleyJoint10GetLengthAEv($1)); //@line 62 "modules/physics/box2d/PulleyJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 62 "modules/physics/box2d/PulleyJoint.cpp"
 return (+$3); //@line 62 "modules/physics/box2d/PulleyJoint.cpp"
}
function __ZNK4love7physics5box2d11PulleyJoint10getLengthBEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 67 "modules/physics/box2d/PulleyJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 67 "modules/physics/box2d/PulleyJoint.cpp"
 $2 = (+__ZNK13b2PulleyJoint10GetLengthBEv($1)); //@line 67 "modules/physics/box2d/PulleyJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 67 "modules/physics/box2d/PulleyJoint.cpp"
 return (+$3); //@line 67 "modules/physics/box2d/PulleyJoint.cpp"
}
function __ZNK4love7physics5box2d11PulleyJoint8getRatioEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 72 "modules/physics/box2d/PulleyJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 72 "modules/physics/box2d/PulleyJoint.cpp"
 $2 = (+__ZNK13b2PulleyJoint8GetRatioEv($1)); //@line 72 "modules/physics/box2d/PulleyJoint.cpp"
 return (+$2); //@line 72 "modules/physics/box2d/PulleyJoint.cpp"
}
function __ZN4love7physics5box2d13RevoluteJointD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 49 "modules/physics/box2d/RevoluteJoint.cpp"
 return; //@line 49 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d13RevoluteJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 49 "modules/physics/box2d/RevoluteJoint.cpp"
 __ZdlPv($this); //@line 48 "modules/physics/box2d/RevoluteJoint.cpp"
 return; //@line 49 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d13RevoluteJointC2EPNS1_4BodyES4_ffb($this,$body1,$body2,$x,$y,$collideConnected) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 $x = +$x;
 $y = +$y;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $def = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $def = sp + 16|0;
 $0 = sp + 8|0;
 $1 = sp;
 __ZN4love7physics5box2d5JointC2EPNS1_4BodyES4_($this,$body1,$body2); //@line 38 "modules/physics/box2d/RevoluteJoint.cpp"
 HEAP32[$this>>2] = (17208); //@line 40 "modules/physics/box2d/RevoluteJoint.cpp"
 $2 = ((($this)) + 28|0); //@line 39 "modules/physics/box2d/RevoluteJoint.cpp"
 HEAP32[$2>>2] = 0; //@line 39 "modules/physics/box2d/RevoluteJoint.cpp"
 ;HEAP32[$def>>2]=0|0;HEAP32[$def+4>>2]=0|0;HEAP32[$def+8>>2]=0|0;HEAP32[$def+12>>2]=0|0;HEAP8[$def+16>>0]=0|0; //@line 79 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 HEAP32[$def>>2] = 1; //@line 39 "./libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.h"
 $3 = ((($def)) + 20|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($def)) + 44|0); //@line 43 "./libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.h"
 HEAPF32[$4>>2] = 0.0; //@line 43 "./libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.h"
 $5 = ((($def)) + 48|0); //@line 44 "./libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.h"
 HEAPF32[$5>>2] = 0.0; //@line 44 "./libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.h"
 $6 = ((($def)) + 60|0); //@line 45 "./libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.h"
 HEAPF32[$6>>2] = 0.0; //@line 45 "./libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.h"
 $7 = ((($def)) + 56|0); //@line 46 "./libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.h"
 HEAPF32[$7>>2] = 0.0; //@line 46 "./libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.h"
 $8 = ((($def)) + 52|0); //@line 48 "./libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.h"
 HEAP8[$8>>0] = 0; //@line 48 "./libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.h"
 ;HEAP32[$3>>2]=0|0;HEAP32[$3+4>>2]=0|0;HEAP32[$3+8>>2]=0|0;HEAP32[$3+12>>2]=0|0;HEAP32[$3+16>>2]=0|0;HEAP8[$3+20>>0]=0|0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($body1)) + 8|0); //@line 42 "modules/physics/box2d/RevoluteJoint.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 42 "modules/physics/box2d/RevoluteJoint.cpp"
 $11 = ((($body2)) + 8|0); //@line 42 "modules/physics/box2d/RevoluteJoint.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 42 "modules/physics/box2d/RevoluteJoint.cpp"
 HEAPF32[$1>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$13>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 invoke_vii(3695,($0|0),($1|0)); //@line 42 "modules/physics/box2d/RevoluteJoint.cpp"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if ($15) {
  $23 = ___cxa_find_matching_catch_2()|0;
  $24 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 45 "modules/physics/box2d/RevoluteJoint.cpp"
  ___resumeException($23|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_viiii(3699,($def|0),($10|0),($12|0),($0|0)); //@line 42 "modules/physics/box2d/RevoluteJoint.cpp"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if ($17) {
  $23 = ___cxa_find_matching_catch_2()|0;
  $24 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 45 "modules/physics/box2d/RevoluteJoint.cpp"
  ___resumeException($23|0);
  // unreachable;
 }
 $18 = ((($def)) + 16|0); //@line 43 "modules/physics/box2d/RevoluteJoint.cpp"
 $19 = $collideConnected&1; //@line 43 "modules/physics/box2d/RevoluteJoint.cpp"
 HEAP8[$18>>0] = $19; //@line 43 "modules/physics/box2d/RevoluteJoint.cpp"
 __THREW__ = 0;
 $20 = (invoke_iii(3697,($this|0),($def|0))|0); //@line 44 "modules/physics/box2d/RevoluteJoint.cpp"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $23 = ___cxa_find_matching_catch_2()|0;
  $24 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 45 "modules/physics/box2d/RevoluteJoint.cpp"
  ___resumeException($23|0);
  // unreachable;
 } else {
  HEAP32[$2>>2] = $20; //@line 44 "modules/physics/box2d/RevoluteJoint.cpp"
  STACKTOP = sp;return; //@line 45 "modules/physics/box2d/RevoluteJoint.cpp"
 }
}
function __ZNK4love7physics5box2d13RevoluteJoint13getJointAngleEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 53 "modules/physics/box2d/RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 53 "modules/physics/box2d/RevoluteJoint.cpp"
 $2 = (+__ZNK15b2RevoluteJoint13GetJointAngleEv($1)); //@line 53 "modules/physics/box2d/RevoluteJoint.cpp"
 return (+$2); //@line 53 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZNK4love7physics5box2d13RevoluteJoint13getJointSpeedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 58 "modules/physics/box2d/RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 58 "modules/physics/box2d/RevoluteJoint.cpp"
 $2 = (+__ZNK15b2RevoluteJoint13GetJointSpeedEv($1)); //@line 58 "modules/physics/box2d/RevoluteJoint.cpp"
 return (+$2); //@line 58 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d13RevoluteJoint15setMotorEnabledEb($this,$enable) {
 $this = $this|0;
 $enable = $enable|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 63 "modules/physics/box2d/RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 63 "modules/physics/box2d/RevoluteJoint.cpp"
 __ZN15b2RevoluteJoint11EnableMotorEb($1,$enable); //@line 63 "modules/physics/box2d/RevoluteJoint.cpp"
 return; //@line 63 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZNK4love7physics5box2d13RevoluteJoint14isMotorEnabledEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 68 "modules/physics/box2d/RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 68 "modules/physics/box2d/RevoluteJoint.cpp"
 $2 = (__ZNK15b2RevoluteJoint14IsMotorEnabledEv($1)|0); //@line 68 "modules/physics/box2d/RevoluteJoint.cpp"
 return ($2|0); //@line 68 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d13RevoluteJoint17setMaxMotorTorqueEf($this,$torque) {
 $this = $this|0;
 $torque = +$torque;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 73 "modules/physics/box2d/RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 73 "modules/physics/box2d/RevoluteJoint.cpp"
 $2 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($torque)); //@line 73 "modules/physics/box2d/RevoluteJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($2)); //@line 73 "modules/physics/box2d/RevoluteJoint.cpp"
 __ZN15b2RevoluteJoint17SetMaxMotorTorqueEf($1,$3); //@line 73 "modules/physics/box2d/RevoluteJoint.cpp"
 return; //@line 74 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d13RevoluteJoint13setMotorSpeedEf($this,$speed) {
 $this = $this|0;
 $speed = +$speed;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 78 "modules/physics/box2d/RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 78 "modules/physics/box2d/RevoluteJoint.cpp"
 __ZN15b2RevoluteJoint13SetMotorSpeedEf($1,$speed); //@line 78 "modules/physics/box2d/RevoluteJoint.cpp"
 return; //@line 79 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZNK4love7physics5box2d13RevoluteJoint13getMotorSpeedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 83 "modules/physics/box2d/RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 83 "modules/physics/box2d/RevoluteJoint.cpp"
 $2 = ((($1)) + 108|0); //@line 201 "./libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.h"
 $3 = +HEAPF32[$2>>2]; //@line 201 "./libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.h"
 return (+$3); //@line 83 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZNK4love7physics5box2d13RevoluteJoint14getMotorTorqueEf($this,$inv_dt) {
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 88 "modules/physics/box2d/RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 88 "modules/physics/box2d/RevoluteJoint.cpp"
 $2 = (+__ZNK15b2RevoluteJoint14GetMotorTorqueEf($1,$inv_dt)); //@line 88 "modules/physics/box2d/RevoluteJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 88 "modules/physics/box2d/RevoluteJoint.cpp"
 $4 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($3)); //@line 88 "modules/physics/box2d/RevoluteJoint.cpp"
 return (+$4); //@line 88 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZNK4love7physics5box2d13RevoluteJoint17getMaxMotorTorqueEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 93 "modules/physics/box2d/RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 93 "modules/physics/box2d/RevoluteJoint.cpp"
 $2 = ((($1)) + 104|0); //@line 140 "./libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.h"
 $3 = +HEAPF32[$2>>2]; //@line 140 "./libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.h"
 $4 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($3)); //@line 93 "modules/physics/box2d/RevoluteJoint.cpp"
 $5 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($4)); //@line 93 "modules/physics/box2d/RevoluteJoint.cpp"
 return (+$5); //@line 93 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d13RevoluteJoint16setLimitsEnabledEb($this,$enable) {
 $this = $this|0;
 $enable = $enable|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 98 "modules/physics/box2d/RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 98 "modules/physics/box2d/RevoluteJoint.cpp"
 __ZN15b2RevoluteJoint11EnableLimitEb($1,$enable); //@line 98 "modules/physics/box2d/RevoluteJoint.cpp"
 return; //@line 99 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZNK4love7physics5box2d13RevoluteJoint16hasLimitsEnabledEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 103 "modules/physics/box2d/RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 103 "modules/physics/box2d/RevoluteJoint.cpp"
 $2 = (__ZNK15b2RevoluteJoint14IsLimitEnabledEv($1)|0); //@line 103 "modules/physics/box2d/RevoluteJoint.cpp"
 return ($2|0); //@line 103 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d13RevoluteJoint13setUpperLimitEf($this,$limit) {
 $this = $this|0;
 $limit = +$limit;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 108 "modules/physics/box2d/RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 108 "modules/physics/box2d/RevoluteJoint.cpp"
 $2 = (+__ZNK15b2RevoluteJoint13GetLowerLimitEv($1)); //@line 108 "modules/physics/box2d/RevoluteJoint.cpp"
 __ZN15b2RevoluteJoint9SetLimitsEff($1,$2,$limit); //@line 108 "modules/physics/box2d/RevoluteJoint.cpp"
 return; //@line 109 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d13RevoluteJoint13setLowerLimitEf($this,$limit) {
 $this = $this|0;
 $limit = +$limit;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 113 "modules/physics/box2d/RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 113 "modules/physics/box2d/RevoluteJoint.cpp"
 $2 = (+__ZNK15b2RevoluteJoint13GetUpperLimitEv($1)); //@line 113 "modules/physics/box2d/RevoluteJoint.cpp"
 __ZN15b2RevoluteJoint9SetLimitsEff($1,$limit,$2); //@line 113 "modules/physics/box2d/RevoluteJoint.cpp"
 return; //@line 114 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d13RevoluteJoint9setLimitsEff($this,$lower,$upper) {
 $this = $this|0;
 $lower = +$lower;
 $upper = +$upper;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 118 "modules/physics/box2d/RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 118 "modules/physics/box2d/RevoluteJoint.cpp"
 __ZN15b2RevoluteJoint9SetLimitsEff($1,$lower,$upper); //@line 118 "modules/physics/box2d/RevoluteJoint.cpp"
 return; //@line 119 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZNK4love7physics5box2d13RevoluteJoint13getLowerLimitEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 123 "modules/physics/box2d/RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 123 "modules/physics/box2d/RevoluteJoint.cpp"
 $2 = (+__ZNK15b2RevoluteJoint13GetLowerLimitEv($1)); //@line 123 "modules/physics/box2d/RevoluteJoint.cpp"
 return (+$2); //@line 123 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZNK4love7physics5box2d13RevoluteJoint13getUpperLimitEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 128 "modules/physics/box2d/RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 128 "modules/physics/box2d/RevoluteJoint.cpp"
 $2 = (+__ZNK15b2RevoluteJoint13GetUpperLimitEv($1)); //@line 128 "modules/physics/box2d/RevoluteJoint.cpp"
 return (+$2); //@line 128 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d13RevoluteJoint9getLimitsEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 133 "modules/physics/box2d/RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 133 "modules/physics/box2d/RevoluteJoint.cpp"
 $2 = (+__ZNK15b2RevoluteJoint13GetLowerLimitEv($1)); //@line 133 "modules/physics/box2d/RevoluteJoint.cpp"
 $3 = $2; //@line 133 "modules/physics/box2d/RevoluteJoint.cpp"
 _lua_pushnumber($L,$3); //@line 133 "modules/physics/box2d/RevoluteJoint.cpp"
 $4 = HEAP32[$0>>2]|0; //@line 134 "modules/physics/box2d/RevoluteJoint.cpp"
 $5 = (+__ZNK15b2RevoluteJoint13GetUpperLimitEv($4)); //@line 134 "modules/physics/box2d/RevoluteJoint.cpp"
 $6 = $5; //@line 134 "modules/physics/box2d/RevoluteJoint.cpp"
 _lua_pushnumber($L,$6); //@line 134 "modules/physics/box2d/RevoluteJoint.cpp"
 return 2; //@line 135 "modules/physics/box2d/RevoluteJoint.cpp"
}
function __ZN4love7physics5box2d9RopeJointD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 55 "modules/physics/box2d/RopeJoint.cpp"
 return; //@line 55 "modules/physics/box2d/RopeJoint.cpp"
}
function __ZN4love7physics5box2d9RopeJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 55 "modules/physics/box2d/RopeJoint.cpp"
 __ZdlPv($this); //@line 54 "modules/physics/box2d/RopeJoint.cpp"
 return; //@line 55 "modules/physics/box2d/RopeJoint.cpp"
}
function __ZN4love7physics5box2d9RopeJointC2EPNS1_4BodyES4_fffffb($this,$body1,$body2,$x1,$y1,$x2,$y2,$maxLength,$collideConnected) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 $x1 = +$x1;
 $y1 = +$y1;
 $x2 = +$x2;
 $y2 = +$y2;
 $maxLength = +$maxLength;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $def = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 52|0;
 $1 = sp + 48|0;
 $2 = sp + 44|0;
 $3 = sp + 40|0;
 $def = sp;
 HEAPF32[$0>>2] = $x1;
 HEAPF32[$1>>2] = $y1;
 HEAPF32[$2>>2] = $x2;
 HEAPF32[$3>>2] = $y2;
 __ZN4love7physics5box2d5JointC2EPNS1_4BodyES4_($this,$body1,$body2); //@line 36 "modules/physics/box2d/RopeJoint.cpp"
 HEAP32[$this>>2] = (17232); //@line 38 "modules/physics/box2d/RopeJoint.cpp"
 $4 = ((($this)) + 28|0); //@line 37 "modules/physics/box2d/RopeJoint.cpp"
 HEAP32[$4>>2] = 0; //@line 37 "modules/physics/box2d/RopeJoint.cpp"
 ;HEAP32[$def>>2]=0|0;HEAP32[$def+4>>2]=0|0;HEAP32[$def+8>>2]=0|0;HEAP32[$def+12>>2]=0|0;HEAP8[$def+16>>0]=0|0; //@line 79 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 HEAP32[$def>>2] = 10; //@line 32 "./libraries/Box2D/Dynamics/Joints/b2RopeJoint.h"
 $5 = ((($def)) + 20|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$5>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($def)) + 24|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$6>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($def)) + 28|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$7>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($def)) + 32|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$8>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($def)) + 36|0); //@line 35 "./libraries/Box2D/Dynamics/Joints/b2RopeJoint.h"
 HEAPF32[$9>>2] = 0.0; //@line 35 "./libraries/Box2D/Dynamics/Joints/b2RopeJoint.h"
 $10 = ((($body1)) + 8|0); //@line 40 "modules/physics/box2d/RopeJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 40 "modules/physics/box2d/RopeJoint.cpp"
 $12 = ((($def)) + 8|0); //@line 40 "modules/physics/box2d/RopeJoint.cpp"
 HEAP32[$12>>2] = $11; //@line 40 "modules/physics/box2d/RopeJoint.cpp"
 $13 = ((($body2)) + 8|0); //@line 41 "modules/physics/box2d/RopeJoint.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 41 "modules/physics/box2d/RopeJoint.cpp"
 $15 = ((($def)) + 12|0); //@line 41 "modules/physics/box2d/RopeJoint.cpp"
 HEAP32[$15>>2] = $14; //@line 41 "modules/physics/box2d/RopeJoint.cpp"
 __THREW__ = 0;
 invoke_viddii(3700,($body1|0),(+$x1),(+$y1),($0|0),($1|0)); //@line 42 "modules/physics/box2d/RopeJoint.cpp"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if (!($17)) {
  __THREW__ = 0;
  invoke_viddii(3700,($body2|0),(+$x2),(+$y2),($2|0),($3|0)); //@line 43 "modules/physics/box2d/RopeJoint.cpp"
  $18 = __THREW__; __THREW__ = 0;
  $19 = $18&1;
  if (!($19)) {
   $20 = +HEAPF32[$0>>2]; //@line 44 "modules/physics/box2d/RopeJoint.cpp"
   __THREW__ = 0;
   $21 = (+invoke_dd(3701,(+$20))); //@line 44 "modules/physics/box2d/RopeJoint.cpp"
   $22 = __THREW__; __THREW__ = 0;
   $23 = $22&1;
   if (!($23)) {
    HEAPF32[$5>>2] = $21; //@line 44 "modules/physics/box2d/RopeJoint.cpp"
    $24 = +HEAPF32[$1>>2]; //@line 45 "modules/physics/box2d/RopeJoint.cpp"
    __THREW__ = 0;
    $25 = (+invoke_dd(3701,(+$24))); //@line 45 "modules/physics/box2d/RopeJoint.cpp"
    $26 = __THREW__; __THREW__ = 0;
    $27 = $26&1;
    if (!($27)) {
     HEAPF32[$6>>2] = $25; //@line 45 "modules/physics/box2d/RopeJoint.cpp"
     $28 = +HEAPF32[$2>>2]; //@line 46 "modules/physics/box2d/RopeJoint.cpp"
     __THREW__ = 0;
     $29 = (+invoke_dd(3701,(+$28))); //@line 46 "modules/physics/box2d/RopeJoint.cpp"
     $30 = __THREW__; __THREW__ = 0;
     $31 = $30&1;
     if (!($31)) {
      HEAPF32[$7>>2] = $29; //@line 46 "modules/physics/box2d/RopeJoint.cpp"
      $32 = +HEAPF32[$3>>2]; //@line 47 "modules/physics/box2d/RopeJoint.cpp"
      __THREW__ = 0;
      $33 = (+invoke_dd(3701,(+$32))); //@line 47 "modules/physics/box2d/RopeJoint.cpp"
      $34 = __THREW__; __THREW__ = 0;
      $35 = $34&1;
      if (!($35)) {
       HEAPF32[$8>>2] = $33; //@line 47 "modules/physics/box2d/RopeJoint.cpp"
       __THREW__ = 0;
       $36 = (+invoke_dd(3701,(+$maxLength))); //@line 48 "modules/physics/box2d/RopeJoint.cpp"
       $37 = __THREW__; __THREW__ = 0;
       $38 = $37&1;
       if (!($38)) {
        HEAPF32[$9>>2] = $36; //@line 48 "modules/physics/box2d/RopeJoint.cpp"
        $39 = ((($def)) + 16|0); //@line 49 "modules/physics/box2d/RopeJoint.cpp"
        $40 = $collideConnected&1; //@line 49 "modules/physics/box2d/RopeJoint.cpp"
        HEAP8[$39>>0] = $40; //@line 49 "modules/physics/box2d/RopeJoint.cpp"
        __THREW__ = 0;
        $41 = (invoke_iii(3697,($this|0),($def|0))|0); //@line 50 "modules/physics/box2d/RopeJoint.cpp"
        $42 = __THREW__; __THREW__ = 0;
        $43 = $42&1;
        if (!($43)) {
         HEAP32[$4>>2] = $41; //@line 50 "modules/physics/box2d/RopeJoint.cpp"
         STACKTOP = sp;return; //@line 51 "modules/physics/box2d/RopeJoint.cpp"
        }
       }
      }
     }
    }
   }
  }
 }
 $44 = ___cxa_find_matching_catch_2()|0;
 $45 = tempRet0;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 51 "modules/physics/box2d/RopeJoint.cpp"
 ___resumeException($44|0);
 // unreachable;
}
function __ZNK4love7physics5box2d9RopeJoint12getMaxLengthEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 59 "modules/physics/box2d/RopeJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 59 "modules/physics/box2d/RopeJoint.cpp"
 $2 = (+__ZNK11b2RopeJoint12GetMaxLengthEv($1)); //@line 59 "modules/physics/box2d/RopeJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 59 "modules/physics/box2d/RopeJoint.cpp"
 return (+$3); //@line 59 "modules/physics/box2d/RopeJoint.cpp"
}
function __ZN4love7physics5box2d5ShapeD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (17256); //@line 55 "modules/physics/box2d/Shape.cpp"
 $0 = ((($this)) + 8|0); //@line 56 "modules/physics/box2d/Shape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 56 "modules/physics/box2d/Shape.cpp"
 $2 = ($1|0)==(0|0); //@line 56 "modules/physics/box2d/Shape.cpp"
 if (!($2)) {
  $3 = ((($this)) + 12|0); //@line 56 "modules/physics/box2d/Shape.cpp"
  $4 = HEAP8[$3>>0]|0; //@line 56 "modules/physics/box2d/Shape.cpp"
  $5 = ($4<<24>>24)==(0); //@line 56 "modules/physics/box2d/Shape.cpp"
  if (!($5)) {
   __THREW__ = 0;
   invoke_vi(3702,($1|0)); //@line 58 "modules/physics/box2d/Shape.cpp"
   $6 = __THREW__; __THREW__ = 0;
   $7 = $6&1;
   if ($7) {
    $13 = ___cxa_find_matching_catch_3(0|0)|0;
    $14 = tempRet0;
    __ZN4love7physics5ShapeD2Ev($this); //@line 62 "modules/physics/box2d/Shape.cpp"
    ___clang_call_terminate($13); //@line 62 "modules/physics/box2d/Shape.cpp"
    // unreachable; //@line 62 "modules/physics/box2d/Shape.cpp"
   }
   $8 = HEAP32[$0>>2]|0; //@line 59 "modules/physics/box2d/Shape.cpp"
   $9 = ($8|0)==(0|0); //@line 59 "modules/physics/box2d/Shape.cpp"
   if (!($9)) {
    $10 = HEAP32[$8>>2]|0; //@line 59 "modules/physics/box2d/Shape.cpp"
    $11 = ((($10)) + 4|0); //@line 59 "modules/physics/box2d/Shape.cpp"
    $12 = HEAP32[$11>>2]|0; //@line 59 "modules/physics/box2d/Shape.cpp"
    FUNCTION_TABLE_vi[$12 & 8191]($8); //@line 59 "modules/physics/box2d/Shape.cpp"
   }
  }
 }
 HEAP32[$0>>2] = 0; //@line 61 "modules/physics/box2d/Shape.cpp"
 __ZN4love7physics5ShapeD2Ev($this); //@line 62 "modules/physics/box2d/Shape.cpp"
 return; //@line 62 "modules/physics/box2d/Shape.cpp"
}
function __ZN4love7physics5box2d5ShapeD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (17256); //@line 55 "modules/physics/box2d/Shape.cpp"
 $0 = ((($this)) + 8|0); //@line 56 "modules/physics/box2d/Shape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 56 "modules/physics/box2d/Shape.cpp"
 $2 = ($1|0)==(0|0); //@line 56 "modules/physics/box2d/Shape.cpp"
 if (!($2)) {
  $3 = ((($this)) + 12|0); //@line 56 "modules/physics/box2d/Shape.cpp"
  $4 = HEAP8[$3>>0]|0; //@line 56 "modules/physics/box2d/Shape.cpp"
  $5 = ($4<<24>>24)==(0); //@line 56 "modules/physics/box2d/Shape.cpp"
  if (!($5)) {
   __THREW__ = 0;
   invoke_vi(3702,($1|0)); //@line 58 "modules/physics/box2d/Shape.cpp"
   $6 = __THREW__; __THREW__ = 0;
   $7 = $6&1;
   if ($7) {
    $13 = ___cxa_find_matching_catch_3(0|0)|0;
    $14 = tempRet0;
    __ZN4love7physics5ShapeD2Ev($this); //@line 62 "modules/physics/box2d/Shape.cpp"
    ___clang_call_terminate($13); //@line 62 "modules/physics/box2d/Shape.cpp"
    // unreachable; //@line 62 "modules/physics/box2d/Shape.cpp"
   }
   $8 = HEAP32[$0>>2]|0; //@line 59 "modules/physics/box2d/Shape.cpp"
   $9 = ($8|0)==(0|0); //@line 59 "modules/physics/box2d/Shape.cpp"
   if (!($9)) {
    $10 = HEAP32[$8>>2]|0; //@line 59 "modules/physics/box2d/Shape.cpp"
    $11 = ((($10)) + 4|0); //@line 59 "modules/physics/box2d/Shape.cpp"
    $12 = HEAP32[$11>>2]|0; //@line 59 "modules/physics/box2d/Shape.cpp"
    FUNCTION_TABLE_vi[$12 & 8191]($8); //@line 59 "modules/physics/box2d/Shape.cpp"
   }
  }
 }
 HEAP32[$0>>2] = 0; //@line 61 "modules/physics/box2d/Shape.cpp"
 __ZN4love7physics5ShapeD2Ev($this); //@line 62 "modules/physics/box2d/Shape.cpp"
 __ZdlPv($this); //@line 55 "modules/physics/box2d/Shape.cpp"
 return; //@line 62 "modules/physics/box2d/Shape.cpp"
}
function __ZN4love7physics5box2d5ShapeC2EP7b2Shapeb($this,$shape,$own) {
 $this = $this|0;
 $shape = $shape|0;
 $own = $own|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./modules/physics/Shape.h"
 HEAP32[$this>>2] = (17256); //@line 49 "modules/physics/box2d/Shape.cpp"
 $0 = ((($this)) + 8|0); //@line 47 "modules/physics/box2d/Shape.cpp"
 HEAP32[$0>>2] = $shape; //@line 47 "modules/physics/box2d/Shape.cpp"
 $1 = ((($this)) + 12|0); //@line 48 "modules/physics/box2d/Shape.cpp"
 $2 = $own&1; //@line 48 "modules/physics/box2d/Shape.cpp"
 HEAP8[$1>>0] = $2; //@line 48 "modules/physics/box2d/Shape.cpp"
 if (!($own)) {
  return; //@line 52 "modules/physics/box2d/Shape.cpp"
 }
 __THREW__ = 0;
 invoke_vii(3703,($shape|0),($this|0)); //@line 51 "modules/physics/box2d/Shape.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  __ZN4love7physics5ShapeD2Ev($this); //@line 52 "modules/physics/box2d/Shape.cpp"
  ___resumeException($5|0);
  // unreachable;
 } else {
  return; //@line 52 "modules/physics/box2d/Shape.cpp"
 }
}
function __ZNK4love7physics5box2d5Shape7getTypeEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 66 "modules/physics/box2d/Shape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 66 "modules/physics/box2d/Shape.cpp"
 $2 = ((($1)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $3 = HEAP32[$2>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 switch ($3|0) {
 case 0:  {
  $$0 = 1;
  break;
 }
 case 2:  {
  $$0 = 2;
  break;
 }
 case 1:  {
  $$0 = 3;
  break;
 }
 case 3:  {
  $$0 = 4;
  break;
 }
 default: {
  $$0 = 0;
 }
 }
 return ($$0|0); //@line 79 "modules/physics/box2d/Shape.cpp"
}
function __ZNK4love7physics5box2d5Shape9getRadiusEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 83 "modules/physics/box2d/Shape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 83 "modules/physics/box2d/Shape.cpp"
 $2 = ((($1)) + 8|0); //@line 83 "modules/physics/box2d/Shape.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 83 "modules/physics/box2d/Shape.cpp"
 $4 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($3)); //@line 83 "modules/physics/box2d/Shape.cpp"
 return (+$4); //@line 83 "modules/physics/box2d/Shape.cpp"
}
function __ZNK4love7physics5box2d5Shape13getChildCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 88 "modules/physics/box2d/Shape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 88 "modules/physics/box2d/Shape.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 88 "modules/physics/box2d/Shape.cpp"
 $3 = ((($2)) + 12|0); //@line 88 "modules/physics/box2d/Shape.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 88 "modules/physics/box2d/Shape.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($1)|0); //@line 88 "modules/physics/box2d/Shape.cpp"
 return ($5|0); //@line 88 "modules/physics/box2d/Shape.cpp"
}
function __ZNK4love7physics5box2d5Shape9testPointEfffff($this,$x,$y,$r,$px,$py) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $r = +$r;
 $px = +$px;
 $py = +$py;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $point = 0, $transform = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $point = sp + 32|0;
 $transform = sp;
 $0 = sp + 24|0;
 $1 = sp + 16|0;
 HEAPF32[$point>>2] = $px; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $2 = ((($point)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$2>>2] = $py; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$0>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($0)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$3>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($transform,$0); //@line 94 "modules/physics/box2d/Shape.cpp"
 $4 = (+Math_sin((+$r))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $5 = (+Math_cos((+$r))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($transform)) + 8|0); //@line 356 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$6>>2] = $4; //@line 356 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($transform)) + 12|0); //@line 356 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$7>>2] = $5; //@line 356 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($this)) + 8|0); //@line 95 "modules/physics/box2d/Shape.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 95 "modules/physics/box2d/Shape.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 95 "modules/physics/box2d/Shape.cpp"
 $11 = ((($10)) + 16|0); //@line 95 "modules/physics/box2d/Shape.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 95 "modules/physics/box2d/Shape.cpp"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($1,$point); //@line 95 "modules/physics/box2d/Shape.cpp"
 $13 = (FUNCTION_TABLE_iiii[$12 & 8191]($9,$transform,$1)|0); //@line 95 "modules/physics/box2d/Shape.cpp"
 STACKTOP = sp;return ($13|0); //@line 95 "modules/physics/box2d/Shape.cpp"
}
function __ZNK4love7physics5box2d5Shape7rayCastEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $$0 = 0, $0 = 0.0, $1 = 0.0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0.0, $30 = 0.0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0.0;
 var $44 = 0.0, $45 = 0, $46 = 0.0, $47 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $input = 0, $output = 0, $transform = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $input = sp + 32|0;
 $transform = sp + 16|0;
 $output = sp;
 $0 = (+_luaL_checknumber($L,1)); //@line 100 "modules/physics/box2d/Shape.cpp"
 $1 = $0; //@line 100 "modules/physics/box2d/Shape.cpp"
 $2 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($1)); //@line 100 "modules/physics/box2d/Shape.cpp"
 $3 = (+_luaL_checknumber($L,2)); //@line 101 "modules/physics/box2d/Shape.cpp"
 $4 = $3; //@line 101 "modules/physics/box2d/Shape.cpp"
 $5 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($4)); //@line 101 "modules/physics/box2d/Shape.cpp"
 $6 = (+_luaL_checknumber($L,3)); //@line 102 "modules/physics/box2d/Shape.cpp"
 $7 = $6; //@line 102 "modules/physics/box2d/Shape.cpp"
 $8 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($7)); //@line 102 "modules/physics/box2d/Shape.cpp"
 $9 = (+_luaL_checknumber($L,4)); //@line 103 "modules/physics/box2d/Shape.cpp"
 $10 = $9; //@line 103 "modules/physics/box2d/Shape.cpp"
 $11 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($10)); //@line 103 "modules/physics/box2d/Shape.cpp"
 $12 = (+_luaL_checknumber($L,5)); //@line 104 "modules/physics/box2d/Shape.cpp"
 $13 = $12; //@line 104 "modules/physics/box2d/Shape.cpp"
 $14 = (+_luaL_checknumber($L,6)); //@line 105 "modules/physics/box2d/Shape.cpp"
 $15 = $14; //@line 105 "modules/physics/box2d/Shape.cpp"
 $16 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($15)); //@line 105 "modules/physics/box2d/Shape.cpp"
 $17 = (+_luaL_checknumber($L,7)); //@line 106 "modules/physics/box2d/Shape.cpp"
 $18 = $17; //@line 106 "modules/physics/box2d/Shape.cpp"
 $19 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($18)); //@line 106 "modules/physics/box2d/Shape.cpp"
 $20 = (+_luaL_checknumber($L,8)); //@line 107 "modules/physics/box2d/Shape.cpp"
 $21 = $20; //@line 107 "modules/physics/box2d/Shape.cpp"
 $22 = (+_luaL_optnumber($L,9,1.0)); //@line 108 "modules/physics/box2d/Shape.cpp"
 $23 = (~~(($22))); //@line 108 "modules/physics/box2d/Shape.cpp"
 $24 = (($23) + -1)|0; //@line 108 "modules/physics/box2d/Shape.cpp"
 HEAPF32[$input>>2] = $2; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $25 = ((($input)) + 4|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$25>>2] = $5; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $26 = ((($input)) + 8|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$26>>2] = $8; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($input)) + 12|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$27>>2] = $11; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $28 = ((($input)) + 16|0); //@line 112 "modules/physics/box2d/Shape.cpp"
 HEAPF32[$28>>2] = $13; //@line 112 "modules/physics/box2d/Shape.cpp"
 $29 = (+Math_sin((+$21))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $30 = (+Math_cos((+$21))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$transform>>2] = $16; //@line 356 "./libraries/Box2D/Common/b2Math.h"
 $31 = ((($transform)) + 4|0); //@line 356 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$31>>2] = $19; //@line 356 "./libraries/Box2D/Common/b2Math.h"
 $32 = ((($transform)) + 8|0); //@line 356 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$32>>2] = $29; //@line 356 "./libraries/Box2D/Common/b2Math.h"
 $33 = ((($transform)) + 12|0); //@line 356 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$33>>2] = $30; //@line 356 "./libraries/Box2D/Common/b2Math.h"
 $34 = ((($this)) + 8|0); //@line 115 "modules/physics/box2d/Shape.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 115 "modules/physics/box2d/Shape.cpp"
 $36 = HEAP32[$35>>2]|0; //@line 115 "modules/physics/box2d/Shape.cpp"
 $37 = ((($36)) + 20|0); //@line 115 "modules/physics/box2d/Shape.cpp"
 $38 = HEAP32[$37>>2]|0; //@line 115 "modules/physics/box2d/Shape.cpp"
 $39 = (FUNCTION_TABLE_iiiiii[$38 & 8191]($35,$output,$input,$transform,$24)|0); //@line 115 "modules/physics/box2d/Shape.cpp"
 if (!($39)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 121 "modules/physics/box2d/Shape.cpp"
 }
 $40 = +HEAPF32[$output>>2]; //@line 117 "modules/physics/box2d/Shape.cpp"
 $41 = $40; //@line 117 "modules/physics/box2d/Shape.cpp"
 _lua_pushnumber($L,$41); //@line 117 "modules/physics/box2d/Shape.cpp"
 $42 = ((($output)) + 4|0); //@line 118 "modules/physics/box2d/Shape.cpp"
 $43 = +HEAPF32[$42>>2]; //@line 118 "modules/physics/box2d/Shape.cpp"
 $44 = $43; //@line 118 "modules/physics/box2d/Shape.cpp"
 _lua_pushnumber($L,$44); //@line 118 "modules/physics/box2d/Shape.cpp"
 $45 = ((($output)) + 8|0); //@line 119 "modules/physics/box2d/Shape.cpp"
 $46 = +HEAPF32[$45>>2]; //@line 119 "modules/physics/box2d/Shape.cpp"
 $47 = $46; //@line 119 "modules/physics/box2d/Shape.cpp"
 _lua_pushnumber($L,$47); //@line 119 "modules/physics/box2d/Shape.cpp"
 $$0 = 3;
 STACKTOP = sp;return ($$0|0); //@line 121 "modules/physics/box2d/Shape.cpp"
}
function __ZNK4love7physics5box2d5Shape11computeAABBEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0;
 var $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0.0, $30 = 0, $31 = 0.0, $32 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $box = 0, $transform = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $transform = sp + 32|0;
 $box = sp + 16|0;
 $0 = sp;
 $1 = (+_luaL_checknumber($L,1)); //@line 125 "modules/physics/box2d/Shape.cpp"
 $2 = $1; //@line 125 "modules/physics/box2d/Shape.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($2)); //@line 125 "modules/physics/box2d/Shape.cpp"
 $4 = (+_luaL_checknumber($L,2)); //@line 126 "modules/physics/box2d/Shape.cpp"
 $5 = $4; //@line 126 "modules/physics/box2d/Shape.cpp"
 $6 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($5)); //@line 126 "modules/physics/box2d/Shape.cpp"
 $7 = (+_luaL_checknumber($L,3)); //@line 127 "modules/physics/box2d/Shape.cpp"
 $8 = $7; //@line 127 "modules/physics/box2d/Shape.cpp"
 $9 = (+_luaL_optnumber($L,4,1.0)); //@line 128 "modules/physics/box2d/Shape.cpp"
 $10 = (~~(($9))); //@line 128 "modules/physics/box2d/Shape.cpp"
 $11 = (($10) + -1)|0; //@line 128 "modules/physics/box2d/Shape.cpp"
 $12 = (+Math_sin((+$8))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $13 = (+Math_cos((+$8))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$transform>>2] = $3; //@line 356 "./libraries/Box2D/Common/b2Math.h"
 $14 = ((($transform)) + 4|0); //@line 356 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$14>>2] = $6; //@line 356 "./libraries/Box2D/Common/b2Math.h"
 $15 = ((($transform)) + 8|0); //@line 356 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$15>>2] = $12; //@line 356 "./libraries/Box2D/Common/b2Math.h"
 $16 = ((($transform)) + 12|0); //@line 356 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$16>>2] = $13; //@line 356 "./libraries/Box2D/Common/b2Math.h"
 $17 = ((($this)) + 8|0); //@line 131 "modules/physics/box2d/Shape.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 131 "modules/physics/box2d/Shape.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 131 "modules/physics/box2d/Shape.cpp"
 $20 = ((($19)) + 24|0); //@line 131 "modules/physics/box2d/Shape.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 131 "modules/physics/box2d/Shape.cpp"
 FUNCTION_TABLE_viiii[$21 & 8191]($18,$box,$transform,$11); //@line 131 "modules/physics/box2d/Shape.cpp"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2AABB($0,$box); //@line 132 "modules/physics/box2d/Shape.cpp"
 ;HEAP32[$box>>2]=HEAP32[$0>>2]|0;HEAP32[$box+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$box+8>>2]=HEAP32[$0+8>>2]|0;HEAP32[$box+12>>2]=HEAP32[$0+12>>2]|0; //@line 132 "modules/physics/box2d/Shape.cpp"
 $22 = +HEAPF32[$box>>2]; //@line 133 "modules/physics/box2d/Shape.cpp"
 $23 = $22; //@line 133 "modules/physics/box2d/Shape.cpp"
 _lua_pushnumber($L,$23); //@line 133 "modules/physics/box2d/Shape.cpp"
 $24 = ((($box)) + 4|0); //@line 134 "modules/physics/box2d/Shape.cpp"
 $25 = +HEAPF32[$24>>2]; //@line 134 "modules/physics/box2d/Shape.cpp"
 $26 = $25; //@line 134 "modules/physics/box2d/Shape.cpp"
 _lua_pushnumber($L,$26); //@line 134 "modules/physics/box2d/Shape.cpp"
 $27 = ((($box)) + 8|0); //@line 135 "modules/physics/box2d/Shape.cpp"
 $28 = +HEAPF32[$27>>2]; //@line 135 "modules/physics/box2d/Shape.cpp"
 $29 = $28; //@line 135 "modules/physics/box2d/Shape.cpp"
 _lua_pushnumber($L,$29); //@line 135 "modules/physics/box2d/Shape.cpp"
 $30 = ((($box)) + 12|0); //@line 136 "modules/physics/box2d/Shape.cpp"
 $31 = +HEAPF32[$30>>2]; //@line 136 "modules/physics/box2d/Shape.cpp"
 $32 = $31; //@line 136 "modules/physics/box2d/Shape.cpp"
 _lua_pushnumber($L,$32); //@line 136 "modules/physics/box2d/Shape.cpp"
 STACKTOP = sp;return 4; //@line 137 "modules/physics/box2d/Shape.cpp"
}
function __ZNK4love7physics5box2d5Shape11computeMassEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0.0;
 var $center = 0, $data = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $data = sp + 8|0;
 $center = sp;
 $0 = (+_luaL_checknumber($L,1)); //@line 142 "modules/physics/box2d/Shape.cpp"
 $1 = $0; //@line 142 "modules/physics/box2d/Shape.cpp"
 $2 = ((($this)) + 8|0); //@line 144 "modules/physics/box2d/Shape.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 144 "modules/physics/box2d/Shape.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 144 "modules/physics/box2d/Shape.cpp"
 $5 = ((($4)) + 28|0); //@line 144 "modules/physics/box2d/Shape.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 144 "modules/physics/box2d/Shape.cpp"
 FUNCTION_TABLE_viid[$6 & 8191]($3,$data,$1); //@line 144 "modules/physics/box2d/Shape.cpp"
 $7 = ((($data)) + 4|0); //@line 145 "modules/physics/box2d/Shape.cpp"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($center,$7); //@line 145 "modules/physics/box2d/Shape.cpp"
 $8 = +HEAPF32[$center>>2]; //@line 146 "modules/physics/box2d/Shape.cpp"
 $9 = $8; //@line 146 "modules/physics/box2d/Shape.cpp"
 _lua_pushnumber($L,$9); //@line 146 "modules/physics/box2d/Shape.cpp"
 $10 = ((($center)) + 4|0); //@line 147 "modules/physics/box2d/Shape.cpp"
 $11 = +HEAPF32[$10>>2]; //@line 147 "modules/physics/box2d/Shape.cpp"
 $12 = $11; //@line 147 "modules/physics/box2d/Shape.cpp"
 _lua_pushnumber($L,$12); //@line 147 "modules/physics/box2d/Shape.cpp"
 $13 = +HEAPF32[$data>>2]; //@line 148 "modules/physics/box2d/Shape.cpp"
 $14 = $13; //@line 148 "modules/physics/box2d/Shape.cpp"
 _lua_pushnumber($L,$14); //@line 148 "modules/physics/box2d/Shape.cpp"
 $15 = ((($data)) + 12|0); //@line 149 "modules/physics/box2d/Shape.cpp"
 $16 = +HEAPF32[$15>>2]; //@line 149 "modules/physics/box2d/Shape.cpp"
 $17 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($16)); //@line 149 "modules/physics/box2d/Shape.cpp"
 $18 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($17)); //@line 149 "modules/physics/box2d/Shape.cpp"
 $19 = $18; //@line 149 "modules/physics/box2d/Shape.cpp"
 _lua_pushnumber($L,$19); //@line 149 "modules/physics/box2d/Shape.cpp"
 STACKTOP = sp;return 4; //@line 150 "modules/physics/box2d/Shape.cpp"
}
function __ZN4love7physics5box2d9WeldJointD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 50 "modules/physics/box2d/WeldJoint.cpp"
 return; //@line 50 "modules/physics/box2d/WeldJoint.cpp"
}
function __ZN4love7physics5box2d9WeldJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 50 "modules/physics/box2d/WeldJoint.cpp"
 __ZdlPv($this); //@line 49 "modules/physics/box2d/WeldJoint.cpp"
 return; //@line 50 "modules/physics/box2d/WeldJoint.cpp"
}
function __ZN4love7physics5box2d9WeldJointC2EPNS1_4BodyES4_ffffb($this,$body1,$body2,$xA,$yA,$xB,$yB,$collideConnected) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 $xA = +$xA;
 $yA = +$yA;
 $xB = +$xB;
 $yB = +$yB;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0, $26 = 0.0;
 var $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $def = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $def = sp + 32|0;
 $0 = sp + 24|0;
 $1 = sp + 16|0;
 $2 = sp + 8|0;
 $3 = sp;
 __ZN4love7physics5box2d5JointC2EPNS1_4BodyES4_($this,$body1,$body2); //@line 38 "modules/physics/box2d/WeldJoint.cpp"
 HEAP32[$this>>2] = (17280); //@line 40 "modules/physics/box2d/WeldJoint.cpp"
 $4 = ((($this)) + 28|0); //@line 39 "modules/physics/box2d/WeldJoint.cpp"
 HEAP32[$4>>2] = 0; //@line 39 "modules/physics/box2d/WeldJoint.cpp"
 ;HEAP32[$def>>2]=0|0;HEAP32[$def+4>>2]=0|0;HEAP32[$def+8>>2]=0|0;HEAP32[$def+12>>2]=0|0;HEAP8[$def+16>>0]=0|0; //@line 79 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 HEAP32[$def>>2] = 8; //@line 31 "./libraries/Box2D/Dynamics/Joints/b2WeldJoint.h"
 $5 = ((($def)) + 20|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;HEAP32[$5+12>>2]=0|0;HEAP32[$5+16>>2]=0|0;HEAP32[$5+20>>2]=0|0;HEAP32[$5+24>>2]=0|0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($body1)) + 8|0); //@line 42 "modules/physics/box2d/WeldJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 42 "modules/physics/box2d/WeldJoint.cpp"
 $8 = ((($body2)) + 8|0); //@line 42 "modules/physics/box2d/WeldJoint.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 42 "modules/physics/box2d/WeldJoint.cpp"
 HEAPF32[$1>>2] = $xA; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$10>>2] = $yA; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 invoke_vii(3695,($0|0),($1|0)); //@line 42 "modules/physics/box2d/WeldJoint.cpp"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $45 = ___cxa_find_matching_catch_2()|0;
  $46 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 46 "modules/physics/box2d/WeldJoint.cpp"
  ___resumeException($45|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_viiii(3704,($def|0),($7|0),($9|0),($0|0)); //@line 42 "modules/physics/box2d/WeldJoint.cpp"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if ($14) {
  $45 = ___cxa_find_matching_catch_2()|0;
  $46 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 46 "modules/physics/box2d/WeldJoint.cpp"
  ___resumeException($45|0);
  // unreachable;
 }
 $15 = HEAP32[$8>>2]|0; //@line 43 "modules/physics/box2d/WeldJoint.cpp"
 HEAPF32[$3>>2] = $xB; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $16 = ((($3)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$16>>2] = $yB; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 invoke_vii(3695,($2|0),($3|0)); //@line 43 "modules/physics/box2d/WeldJoint.cpp"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if ($18) {
  $45 = ___cxa_find_matching_catch_2()|0;
  $46 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 46 "modules/physics/box2d/WeldJoint.cpp"
  ___resumeException($45|0);
  // unreachable;
 }
 $19 = +HEAPF32[$2>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $20 = ((($15)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $21 = +HEAPF32[$20>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $22 = $19 - $21; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $23 = ((($2)) + 4|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $24 = +HEAPF32[$23>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $25 = ((($15)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $26 = +HEAPF32[$25>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $27 = $24 - $26; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $28 = ((($15)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $29 = +HEAPF32[$28>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $30 = $22 * $29; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $31 = ((($15)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $32 = +HEAPF32[$31>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $33 = $27 * $32; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $34 = $30 + $33; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $35 = $22 * $32; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $36 = $29 * $27; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $37 = $36 - $35; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $38 = ((($def)) + 28|0); //@line 43 "modules/physics/box2d/WeldJoint.cpp"
 HEAPF32[$38>>2] = $34; //@line 43 "modules/physics/box2d/WeldJoint.cpp"
 $39 = ((($def)) + 32|0); //@line 43 "modules/physics/box2d/WeldJoint.cpp"
 HEAPF32[$39>>2] = $37; //@line 43 "modules/physics/box2d/WeldJoint.cpp"
 $40 = ((($def)) + 16|0); //@line 44 "modules/physics/box2d/WeldJoint.cpp"
 $41 = $collideConnected&1; //@line 44 "modules/physics/box2d/WeldJoint.cpp"
 HEAP8[$40>>0] = $41; //@line 44 "modules/physics/box2d/WeldJoint.cpp"
 __THREW__ = 0;
 $42 = (invoke_iii(3697,($this|0),($def|0))|0); //@line 45 "modules/physics/box2d/WeldJoint.cpp"
 $43 = __THREW__; __THREW__ = 0;
 $44 = $43&1;
 if ($44) {
  $45 = ___cxa_find_matching_catch_2()|0;
  $46 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 46 "modules/physics/box2d/WeldJoint.cpp"
  ___resumeException($45|0);
  // unreachable;
 } else {
  HEAP32[$4>>2] = $42; //@line 45 "modules/physics/box2d/WeldJoint.cpp"
  STACKTOP = sp;return; //@line 46 "modules/physics/box2d/WeldJoint.cpp"
 }
}
function __ZN4love7physics5box2d9WeldJoint12setFrequencyEf($this,$hz) {
 $this = $this|0;
 $hz = +$hz;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 54 "modules/physics/box2d/WeldJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 54 "modules/physics/box2d/WeldJoint.cpp"
 $2 = ((($1)) + 68|0); //@line 81 "./libraries/Box2D/Dynamics/Joints/b2WeldJoint.h"
 HEAPF32[$2>>2] = $hz; //@line 81 "./libraries/Box2D/Dynamics/Joints/b2WeldJoint.h"
 return; //@line 55 "modules/physics/box2d/WeldJoint.cpp"
}
function __ZNK4love7physics5box2d9WeldJoint12getFrequencyEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 59 "modules/physics/box2d/WeldJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 59 "modules/physics/box2d/WeldJoint.cpp"
 $2 = ((($1)) + 68|0); //@line 82 "./libraries/Box2D/Dynamics/Joints/b2WeldJoint.h"
 $3 = +HEAPF32[$2>>2]; //@line 82 "./libraries/Box2D/Dynamics/Joints/b2WeldJoint.h"
 return (+$3); //@line 59 "modules/physics/box2d/WeldJoint.cpp"
}
function __ZN4love7physics5box2d9WeldJoint15setDampingRatioEf($this,$d) {
 $this = $this|0;
 $d = +$d;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 64 "modules/physics/box2d/WeldJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 64 "modules/physics/box2d/WeldJoint.cpp"
 $2 = ((($1)) + 72|0); //@line 85 "./libraries/Box2D/Dynamics/Joints/b2WeldJoint.h"
 HEAPF32[$2>>2] = $d; //@line 85 "./libraries/Box2D/Dynamics/Joints/b2WeldJoint.h"
 return; //@line 65 "modules/physics/box2d/WeldJoint.cpp"
}
function __ZNK4love7physics5box2d9WeldJoint15getDampingRatioEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 69 "modules/physics/box2d/WeldJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 69 "modules/physics/box2d/WeldJoint.cpp"
 $2 = ((($1)) + 72|0); //@line 86 "./libraries/Box2D/Dynamics/Joints/b2WeldJoint.h"
 $3 = +HEAPF32[$2>>2]; //@line 86 "./libraries/Box2D/Dynamics/Joints/b2WeldJoint.h"
 return (+$3); //@line 69 "modules/physics/box2d/WeldJoint.cpp"
}
function __ZN4love7physics5box2d10WheelJointD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 49 "modules/physics/box2d/WheelJoint.cpp"
 return; //@line 49 "modules/physics/box2d/WheelJoint.cpp"
}
function __ZN4love7physics5box2d10WheelJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 49 "modules/physics/box2d/WheelJoint.cpp"
 __ZdlPv($this); //@line 48 "modules/physics/box2d/WheelJoint.cpp"
 return; //@line 49 "modules/physics/box2d/WheelJoint.cpp"
}
function __ZN4love7physics5box2d10WheelJointC2EPNS1_4BodyES4_ffffffb($this,$body1,$body2,$xA,$yA,$xB,$yB,$ax,$ay,$collideConnected) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 $xA = +$xA;
 $yA = +$yA;
 $xB = +$xB;
 $yB = +$yB;
 $ax = +$ax;
 $ay = +$ay;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0;
 var $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $def = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $def = sp + 40|0;
 $0 = sp + 32|0;
 $1 = sp + 24|0;
 $2 = sp + 16|0;
 $3 = sp + 8|0;
 $4 = sp;
 __ZN4love7physics5box2d5JointC2EPNS1_4BodyES4_($this,$body1,$body2); //@line 36 "modules/physics/box2d/WheelJoint.cpp"
 HEAP32[$this>>2] = (17304); //@line 38 "modules/physics/box2d/WheelJoint.cpp"
 $5 = ((($this)) + 28|0); //@line 37 "modules/physics/box2d/WheelJoint.cpp"
 HEAP32[$5>>2] = 0; //@line 37 "modules/physics/box2d/WheelJoint.cpp"
 ;HEAP32[$def>>2]=0|0;HEAP32[$def+4>>2]=0|0;HEAP32[$def+8>>2]=0|0;HEAP32[$def+12>>2]=0|0;HEAP8[$def+16>>0]=0|0; //@line 79 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 HEAP32[$def>>2] = 7; //@line 34 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 $6 = ((($def)) + 20|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($def)) + 36|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;HEAP32[$6+12>>2]=0|0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$7>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($def)) + 40|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$8>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($def)) + 44|0); //@line 38 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 HEAP8[$9>>0] = 0; //@line 38 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 $10 = ((($def)) + 48|0); //@line 39 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 HEAPF32[$10>>2] = 0.0; //@line 39 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 $11 = ((($def)) + 52|0); //@line 40 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 HEAPF32[$11>>2] = 0.0; //@line 40 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 $12 = ((($def)) + 56|0); //@line 41 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 HEAPF32[$12>>2] = 2.0; //@line 41 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 $13 = ((($def)) + 60|0); //@line 42 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 HEAPF32[$13>>2] = 0.69999998807907104; //@line 42 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 $14 = ((($body1)) + 8|0); //@line 41 "modules/physics/box2d/WheelJoint.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 41 "modules/physics/box2d/WheelJoint.cpp"
 $16 = ((($body2)) + 8|0); //@line 41 "modules/physics/box2d/WheelJoint.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 41 "modules/physics/box2d/WheelJoint.cpp"
 HEAPF32[$1>>2] = $xA; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $18 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$18>>2] = $yA; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 invoke_vii(3695,($0|0),($1|0)); //@line 41 "modules/physics/box2d/WheelJoint.cpp"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $54 = ___cxa_find_matching_catch_2()|0;
  $55 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 45 "modules/physics/box2d/WheelJoint.cpp"
  ___resumeException($54|0);
  // unreachable;
 }
 HEAPF32[$2>>2] = $ax; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($2)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$21>>2] = $ay; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 invoke_viiiii(3705,($def|0),($15|0),($17|0),($0|0),($2|0)); //@line 41 "modules/physics/box2d/WheelJoint.cpp"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if ($23) {
  $54 = ___cxa_find_matching_catch_2()|0;
  $55 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 45 "modules/physics/box2d/WheelJoint.cpp"
  ___resumeException($54|0);
  // unreachable;
 }
 $24 = HEAP32[$16>>2]|0; //@line 42 "modules/physics/box2d/WheelJoint.cpp"
 HEAPF32[$4>>2] = $xB; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $25 = ((($4)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$25>>2] = $yB; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 invoke_vii(3695,($3|0),($4|0)); //@line 42 "modules/physics/box2d/WheelJoint.cpp"
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  $54 = ___cxa_find_matching_catch_2()|0;
  $55 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 45 "modules/physics/box2d/WheelJoint.cpp"
  ___resumeException($54|0);
  // unreachable;
 }
 $28 = +HEAPF32[$3>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $29 = ((($24)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $30 = +HEAPF32[$29>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $31 = $28 - $30; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $32 = ((($3)) + 4|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $33 = +HEAPF32[$32>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $34 = ((($24)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $35 = +HEAPF32[$34>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $36 = $33 - $35; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $37 = ((($24)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $38 = +HEAPF32[$37>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $39 = $31 * $38; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $40 = ((($24)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $41 = +HEAPF32[$40>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $42 = $36 * $41; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $43 = $39 + $42; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $44 = $31 * $41; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $45 = $38 * $36; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $46 = $45 - $44; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $47 = ((($def)) + 28|0); //@line 42 "modules/physics/box2d/WheelJoint.cpp"
 HEAPF32[$47>>2] = $43; //@line 42 "modules/physics/box2d/WheelJoint.cpp"
 $48 = ((($def)) + 32|0); //@line 42 "modules/physics/box2d/WheelJoint.cpp"
 HEAPF32[$48>>2] = $46; //@line 42 "modules/physics/box2d/WheelJoint.cpp"
 $49 = ((($def)) + 16|0); //@line 43 "modules/physics/box2d/WheelJoint.cpp"
 $50 = $collideConnected&1; //@line 43 "modules/physics/box2d/WheelJoint.cpp"
 HEAP8[$49>>0] = $50; //@line 43 "modules/physics/box2d/WheelJoint.cpp"
 __THREW__ = 0;
 $51 = (invoke_iii(3697,($this|0),($def|0))|0); //@line 44 "modules/physics/box2d/WheelJoint.cpp"
 $52 = __THREW__; __THREW__ = 0;
 $53 = $52&1;
 if ($53) {
  $54 = ___cxa_find_matching_catch_2()|0;
  $55 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 45 "modules/physics/box2d/WheelJoint.cpp"
  ___resumeException($54|0);
  // unreachable;
 } else {
  HEAP32[$5>>2] = $51; //@line 44 "modules/physics/box2d/WheelJoint.cpp"
  STACKTOP = sp;return; //@line 45 "modules/physics/box2d/WheelJoint.cpp"
 }
}
function __ZNK4love7physics5box2d10WheelJoint19getJointTranslationEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 53 "modules/physics/box2d/WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 53 "modules/physics/box2d/WheelJoint.cpp"
 $2 = (+__ZNK12b2WheelJoint19GetJointTranslationEv($1)); //@line 53 "modules/physics/box2d/WheelJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 53 "modules/physics/box2d/WheelJoint.cpp"
 return (+$3); //@line 53 "modules/physics/box2d/WheelJoint.cpp"
}
function __ZNK4love7physics5box2d10WheelJoint13getJointSpeedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 58 "modules/physics/box2d/WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 58 "modules/physics/box2d/WheelJoint.cpp"
 $2 = (+__ZNK12b2WheelJoint13GetJointSpeedEv($1)); //@line 58 "modules/physics/box2d/WheelJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 58 "modules/physics/box2d/WheelJoint.cpp"
 return (+$3); //@line 58 "modules/physics/box2d/WheelJoint.cpp"
}
function __ZN4love7physics5box2d10WheelJoint15setMotorEnabledEb($this,$enable) {
 $this = $this|0;
 $enable = $enable|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 63 "modules/physics/box2d/WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 63 "modules/physics/box2d/WheelJoint.cpp"
 __ZN12b2WheelJoint11EnableMotorEb($1,$enable); //@line 63 "modules/physics/box2d/WheelJoint.cpp"
 return; //@line 63 "modules/physics/box2d/WheelJoint.cpp"
}
function __ZNK4love7physics5box2d10WheelJoint14isMotorEnabledEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 68 "modules/physics/box2d/WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 68 "modules/physics/box2d/WheelJoint.cpp"
 $2 = (__ZNK12b2WheelJoint14IsMotorEnabledEv($1)|0); //@line 68 "modules/physics/box2d/WheelJoint.cpp"
 return ($2|0); //@line 68 "modules/physics/box2d/WheelJoint.cpp"
}
function __ZN4love7physics5box2d10WheelJoint13setMotorSpeedEf($this,$speed) {
 $this = $this|0;
 $speed = +$speed;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 73 "modules/physics/box2d/WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 73 "modules/physics/box2d/WheelJoint.cpp"
 __ZN12b2WheelJoint13SetMotorSpeedEf($1,$speed); //@line 73 "modules/physics/box2d/WheelJoint.cpp"
 return; //@line 74 "modules/physics/box2d/WheelJoint.cpp"
}
function __ZNK4love7physics5box2d10WheelJoint13getMotorSpeedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 78 "modules/physics/box2d/WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 78 "modules/physics/box2d/WheelJoint.cpp"
 $2 = ((($1)) + 124|0); //@line 182 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 $3 = +HEAPF32[$2>>2]; //@line 182 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 return (+$3); //@line 78 "modules/physics/box2d/WheelJoint.cpp"
}
function __ZN4love7physics5box2d10WheelJoint17setMaxMotorTorqueEf($this,$torque) {
 $this = $this|0;
 $torque = +$torque;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 83 "modules/physics/box2d/WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 83 "modules/physics/box2d/WheelJoint.cpp"
 $2 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($torque)); //@line 83 "modules/physics/box2d/WheelJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($2)); //@line 83 "modules/physics/box2d/WheelJoint.cpp"
 __ZN12b2WheelJoint17SetMaxMotorTorqueEf($1,$3); //@line 83 "modules/physics/box2d/WheelJoint.cpp"
 return; //@line 84 "modules/physics/box2d/WheelJoint.cpp"
}
function __ZNK4love7physics5box2d10WheelJoint17getMaxMotorTorqueEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 88 "modules/physics/box2d/WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 88 "modules/physics/box2d/WheelJoint.cpp"
 $2 = ((($1)) + 120|0); //@line 187 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 $3 = +HEAPF32[$2>>2]; //@line 187 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 $4 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($3)); //@line 88 "modules/physics/box2d/WheelJoint.cpp"
 $5 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($4)); //@line 88 "modules/physics/box2d/WheelJoint.cpp"
 return (+$5); //@line 88 "modules/physics/box2d/WheelJoint.cpp"
}
function __ZNK4love7physics5box2d10WheelJoint14getMotorTorqueEf($this,$inv_dt) {
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 93 "modules/physics/box2d/WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 93 "modules/physics/box2d/WheelJoint.cpp"
 $2 = (+__ZNK12b2WheelJoint14GetMotorTorqueEf($1,$inv_dt)); //@line 93 "modules/physics/box2d/WheelJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 93 "modules/physics/box2d/WheelJoint.cpp"
 $4 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($3)); //@line 93 "modules/physics/box2d/WheelJoint.cpp"
 return (+$4); //@line 93 "modules/physics/box2d/WheelJoint.cpp"
}
function __ZN4love7physics5box2d10WheelJoint18setSpringFrequencyEf($this,$hz) {
 $this = $this|0;
 $hz = +$hz;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 98 "modules/physics/box2d/WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 98 "modules/physics/box2d/WheelJoint.cpp"
 $2 = ((($1)) + 68|0); //@line 192 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 HEAPF32[$2>>2] = $hz; //@line 192 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 return; //@line 99 "modules/physics/box2d/WheelJoint.cpp"
}
function __ZNK4love7physics5box2d10WheelJoint18getSpringFrequencyEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 103 "modules/physics/box2d/WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 103 "modules/physics/box2d/WheelJoint.cpp"
 $2 = ((($1)) + 68|0); //@line 197 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 $3 = +HEAPF32[$2>>2]; //@line 197 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 return (+$3); //@line 103 "modules/physics/box2d/WheelJoint.cpp"
}
function __ZN4love7physics5box2d10WheelJoint21setSpringDampingRatioEf($this,$ratio) {
 $this = $this|0;
 $ratio = +$ratio;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 108 "modules/physics/box2d/WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 108 "modules/physics/box2d/WheelJoint.cpp"
 $2 = ((($1)) + 72|0); //@line 202 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 HEAPF32[$2>>2] = $ratio; //@line 202 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 return; //@line 109 "modules/physics/box2d/WheelJoint.cpp"
}
function __ZNK4love7physics5box2d10WheelJoint21getSpringDampingRatioEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 113 "modules/physics/box2d/WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 113 "modules/physics/box2d/WheelJoint.cpp"
 $2 = ((($1)) + 72|0); //@line 207 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 $3 = +HEAPF32[$2>>2]; //@line 207 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 return (+$3); //@line 113 "modules/physics/box2d/WheelJoint.cpp"
}
function __ZN4love7physics5box2d5World13QueryCallbackD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 148 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World13QueryCallbackD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 147 "modules/physics/box2d/World.cpp"
 return; //@line 148 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World13QueryCallback13ReportFixtureEP9b2Fixture($this,$fixture) {
 $this = $this|0;
 $fixture = $fixture|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 4|0); //@line 152 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 152 "modules/physics/box2d/World.cpp"
 $2 = ($1|0)==(0|0); //@line 152 "modules/physics/box2d/World.cpp"
 if ($2) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 166 "modules/physics/box2d/World.cpp"
 }
 $3 = ((($this)) + 8|0); //@line 154 "modules/physics/box2d/World.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 154 "modules/physics/box2d/World.cpp"
 _lua_pushvalue($1,$4); //@line 154 "modules/physics/box2d/World.cpp"
 $5 = (__ZN4love8Memoizer4findEPv($fixture)|0); //@line 155 "modules/physics/box2d/World.cpp"
 $6 = ($5|0)==(0|0); //@line 156 "modules/physics/box2d/World.cpp"
 if (!($6)) {
  $12 = HEAP32[$0>>2]|0; //@line 158 "modules/physics/box2d/World.cpp"
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($12,35,$5); //@line 158 "modules/physics/box2d/World.cpp"
  $13 = HEAP32[$0>>2]|0; //@line 159 "modules/physics/box2d/World.cpp"
  _lua_call($13,1,1); //@line 159 "modules/physics/box2d/World.cpp"
  $14 = HEAP32[$0>>2]|0; //@line 160 "modules/physics/box2d/World.cpp"
  $15 = (__ZN4love14luax_tobooleanEP9lua_Statei($14,-1)|0); //@line 160 "modules/physics/box2d/World.cpp"
  $16 = HEAP32[$0>>2]|0; //@line 161 "modules/physics/box2d/World.cpp"
  _lua_settop($16,-2); //@line 161 "modules/physics/box2d/World.cpp"
  $$0 = $15;
  STACKTOP = sp;return ($$0|0); //@line 166 "modules/physics/box2d/World.cpp"
 }
 $7 = (___cxa_allocate_exception(16)|0); //@line 157 "modules/physics/box2d/World.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($7|0),(444616|0),($vararg_buffer|0)); //@line 157 "modules/physics/box2d/World.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  ___cxa_free_exception(($7|0)); //@line 157 "modules/physics/box2d/World.cpp"
  ___resumeException($10|0);
  // unreachable;
 } else {
  ___cxa_throw(($7|0),(96|0),(344|0)); //@line 157 "modules/physics/box2d/World.cpp"
  // unreachable; //@line 157 "modules/physics/box2d/World.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d5World15RayCastCallbackD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 177 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World15RayCastCallbackD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 176 "modules/physics/box2d/World.cpp"
 return; //@line 177 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World15RayCastCallback13ReportFixtureEP9b2FixtureRK6b2Vec2S8_f($this,$fixture,$point,$normal,$fraction) {
 $this = $this|0;
 $fixture = $fixture|0;
 $point = $point|0;
 $normal = $normal|0;
 $fraction = +$fraction;
 var $$0 = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0, $25 = 0.0;
 var $26 = 0.0, $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scaledPoint = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $scaledPoint = sp + 16|0;
 $0 = ((($this)) + 4|0); //@line 181 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 181 "modules/physics/box2d/World.cpp"
 $2 = ($1|0)==(0|0); //@line 181 "modules/physics/box2d/World.cpp"
 if ($2) {
  $$0 = 0.0;
  STACKTOP = sp;return (+$$0); //@line 203 "modules/physics/box2d/World.cpp"
 }
 $3 = ((($this)) + 8|0); //@line 183 "modules/physics/box2d/World.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 183 "modules/physics/box2d/World.cpp"
 _lua_pushvalue($1,$4); //@line 183 "modules/physics/box2d/World.cpp"
 $5 = (__ZN4love8Memoizer4findEPv($fixture)|0); //@line 184 "modules/physics/box2d/World.cpp"
 $6 = ($5|0)==(0|0); //@line 185 "modules/physics/box2d/World.cpp"
 if ($6) {
  $7 = (___cxa_allocate_exception(16)|0); //@line 186 "modules/physics/box2d/World.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($7|0),(444616|0),($vararg_buffer|0)); //@line 186 "modules/physics/box2d/World.cpp"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if ($9) {
   $10 = ___cxa_find_matching_catch_2()|0;
   $11 = tempRet0;
   ___cxa_free_exception(($7|0)); //@line 186 "modules/physics/box2d/World.cpp"
   ___resumeException($10|0);
   // unreachable;
  } else {
   ___cxa_throw(($7|0),(96|0),(344|0)); //@line 186 "modules/physics/box2d/World.cpp"
   // unreachable; //@line 186 "modules/physics/box2d/World.cpp"
  }
 }
 $12 = HEAP32[$0>>2]|0; //@line 187 "modules/physics/box2d/World.cpp"
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($12,35,$5); //@line 187 "modules/physics/box2d/World.cpp"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($scaledPoint,$point); //@line 188 "modules/physics/box2d/World.cpp"
 $13 = HEAP32[$0>>2]|0; //@line 189 "modules/physics/box2d/World.cpp"
 $14 = +HEAPF32[$scaledPoint>>2]; //@line 189 "modules/physics/box2d/World.cpp"
 $15 = $14; //@line 189 "modules/physics/box2d/World.cpp"
 _lua_pushnumber($13,$15); //@line 189 "modules/physics/box2d/World.cpp"
 $16 = HEAP32[$0>>2]|0; //@line 190 "modules/physics/box2d/World.cpp"
 $17 = ((($scaledPoint)) + 4|0); //@line 190 "modules/physics/box2d/World.cpp"
 $18 = +HEAPF32[$17>>2]; //@line 190 "modules/physics/box2d/World.cpp"
 $19 = $18; //@line 190 "modules/physics/box2d/World.cpp"
 _lua_pushnumber($16,$19); //@line 190 "modules/physics/box2d/World.cpp"
 $20 = HEAP32[$0>>2]|0; //@line 191 "modules/physics/box2d/World.cpp"
 $21 = +HEAPF32[$normal>>2]; //@line 191 "modules/physics/box2d/World.cpp"
 $22 = $21; //@line 191 "modules/physics/box2d/World.cpp"
 _lua_pushnumber($20,$22); //@line 191 "modules/physics/box2d/World.cpp"
 $23 = HEAP32[$0>>2]|0; //@line 192 "modules/physics/box2d/World.cpp"
 $24 = ((($normal)) + 4|0); //@line 192 "modules/physics/box2d/World.cpp"
 $25 = +HEAPF32[$24>>2]; //@line 192 "modules/physics/box2d/World.cpp"
 $26 = $25; //@line 192 "modules/physics/box2d/World.cpp"
 _lua_pushnumber($23,$26); //@line 192 "modules/physics/box2d/World.cpp"
 $27 = HEAP32[$0>>2]|0; //@line 193 "modules/physics/box2d/World.cpp"
 $28 = $fraction; //@line 193 "modules/physics/box2d/World.cpp"
 _lua_pushnumber($27,$28); //@line 193 "modules/physics/box2d/World.cpp"
 $29 = HEAP32[$0>>2]|0; //@line 194 "modules/physics/box2d/World.cpp"
 _lua_call($29,6,1); //@line 194 "modules/physics/box2d/World.cpp"
 $30 = HEAP32[$0>>2]|0; //@line 195 "modules/physics/box2d/World.cpp"
 $31 = (_lua_isnumber($30,-1)|0); //@line 195 "modules/physics/box2d/World.cpp"
 $32 = ($31|0)==(0); //@line 195 "modules/physics/box2d/World.cpp"
 if ($32) {
  $33 = HEAP32[$0>>2]|0; //@line 196 "modules/physics/box2d/World.cpp"
  (_luaL_error($33,356813,$vararg_buffer1)|0); //@line 196 "modules/physics/box2d/World.cpp"
 }
 $34 = HEAP32[$0>>2]|0; //@line 197 "modules/physics/box2d/World.cpp"
 $35 = (+_lua_tonumber($34,-1)); //@line 197 "modules/physics/box2d/World.cpp"
 $36 = $35; //@line 197 "modules/physics/box2d/World.cpp"
 $37 = HEAP32[$0>>2]|0; //@line 198 "modules/physics/box2d/World.cpp"
 _lua_settop($37,-2); //@line 198 "modules/physics/box2d/World.cpp"
 $$0 = $36;
 STACKTOP = sp;return (+$$0); //@line 203 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5WorldD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i10 = 0, $scevgep$i$i$i15 = 0, $scevgep$i$i$i25 = 0, $scevgep$i$i$i4 = 0, $scevgep$i$i$i7 = 0, $scevgep4$i$i$i = 0;
 var $scevgep4$i$i$i11 = 0, $scevgep4$i$i$i16 = 0, $scevgep4$i$i$i26 = 0, $scevgep4$i$i$i5 = 0, $scevgep4$i$i$i8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (17368); //@line 248 "modules/physics/box2d/World.cpp"
 $0 = ((($this)) + 8|0); //@line 248 "modules/physics/box2d/World.cpp"
 HEAP32[$0>>2] = (17420); //@line 248 "modules/physics/box2d/World.cpp"
 $1 = ((($this)) + 12|0); //@line 248 "modules/physics/box2d/World.cpp"
 HEAP32[$1>>2] = (17452); //@line 248 "modules/physics/box2d/World.cpp"
 $2 = ((($this)) + 16|0); //@line 248 "modules/physics/box2d/World.cpp"
 HEAP32[$2>>2] = (17472); //@line 248 "modules/physics/box2d/World.cpp"
 __THREW__ = 0;
 invoke_vi(3445,($this|0)); //@line 249 "modules/physics/box2d/World.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $68 = ___cxa_find_matching_catch_3(0|0)|0;
  $69 = tempRet0;
  $70 = ((($this)) + 100|0); //@line 106 "modules/physics/box2d/World.cpp"
  $71 = HEAP32[$70>>2]|0; //@line 106 "modules/physics/box2d/World.cpp"
  $72 = ($71|0)==(0|0); //@line 106 "modules/physics/box2d/World.cpp"
  if (!($72)) {
   $73 = HEAP32[$71>>2]|0; //@line 107 "modules/physics/box2d/World.cpp"
   $74 = ((($73)) + 4|0); //@line 107 "modules/physics/box2d/World.cpp"
   $75 = HEAP32[$74>>2]|0; //@line 107 "modules/physics/box2d/World.cpp"
   FUNCTION_TABLE_vi[$75 & 8191]($71); //@line 107 "modules/physics/box2d/World.cpp"
  }
  $76 = ((($this)) + 92|0); //@line 45 "modules/physics/box2d/World.cpp"
  $77 = HEAP32[$76>>2]|0; //@line 45 "modules/physics/box2d/World.cpp"
  $78 = ($77|0)==(0|0); //@line 45 "modules/physics/box2d/World.cpp"
  if (!($78)) {
   $79 = HEAP32[$77>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
   $80 = ((($79)) + 4|0); //@line 46 "modules/physics/box2d/World.cpp"
   $81 = HEAP32[$80>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
   FUNCTION_TABLE_vi[$81 & 8191]($77); //@line 46 "modules/physics/box2d/World.cpp"
  }
  $82 = ((($this)) + 84|0); //@line 45 "modules/physics/box2d/World.cpp"
  $83 = HEAP32[$82>>2]|0; //@line 45 "modules/physics/box2d/World.cpp"
  $84 = ($83|0)==(0|0); //@line 45 "modules/physics/box2d/World.cpp"
  if (!($84)) {
   $85 = HEAP32[$83>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
   $86 = ((($85)) + 4|0); //@line 46 "modules/physics/box2d/World.cpp"
   $87 = HEAP32[$86>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
   FUNCTION_TABLE_vi[$87 & 8191]($83); //@line 46 "modules/physics/box2d/World.cpp"
  }
  $88 = ((($this)) + 76|0); //@line 45 "modules/physics/box2d/World.cpp"
  $89 = HEAP32[$88>>2]|0; //@line 45 "modules/physics/box2d/World.cpp"
  $90 = ($89|0)==(0|0); //@line 45 "modules/physics/box2d/World.cpp"
  if (!($90)) {
   $91 = HEAP32[$89>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
   $92 = ((($91)) + 4|0); //@line 46 "modules/physics/box2d/World.cpp"
   $93 = HEAP32[$92>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
   FUNCTION_TABLE_vi[$93 & 8191]($89); //@line 46 "modules/physics/box2d/World.cpp"
  }
  $94 = ((($this)) + 68|0); //@line 45 "modules/physics/box2d/World.cpp"
  $95 = HEAP32[$94>>2]|0; //@line 45 "modules/physics/box2d/World.cpp"
  $96 = ($95|0)==(0|0); //@line 45 "modules/physics/box2d/World.cpp"
  if (!($96)) {
   $97 = HEAP32[$95>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
   $98 = ((($97)) + 4|0); //@line 46 "modules/physics/box2d/World.cpp"
   $99 = HEAP32[$98>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
   FUNCTION_TABLE_vi[$99 & 8191]($95); //@line 46 "modules/physics/box2d/World.cpp"
  }
  $100 = ((($this)) + 52|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $101 = HEAP32[$100>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $102 = ($101|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $103 = $101; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($102)) {
   $104 = ((($this)) + 56|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $105 = HEAP32[$104>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $106 = ($105|0)==($101|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($106)) {
    $scevgep$i$i$i25 = ((($105)) + -4|0);
    $107 = $scevgep$i$i$i25;
    $108 = (($107) - ($103))|0;
    $109 = $108 >>> 2;
    $110 = $109 ^ -1;
    $scevgep4$i$i$i26 = (($105) + ($110<<2)|0);
    HEAP32[$104>>2] = $scevgep4$i$i$i26; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($101); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  $111 = ((($this)) + 40|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $112 = HEAP32[$111>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $113 = ($112|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $114 = $112; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($113)) {
   $115 = ((($this)) + 44|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $116 = HEAP32[$115>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $117 = ($116|0)==($112|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($117)) {
    $scevgep$i$i$i15 = ((($116)) + -4|0);
    $118 = $scevgep$i$i$i15;
    $119 = (($118) - ($114))|0;
    $120 = $119 >>> 2;
    $121 = $120 ^ -1;
    $scevgep4$i$i$i16 = (($116) + ($121<<2)|0);
    HEAP32[$115>>2] = $scevgep4$i$i$i16; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($112); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  $122 = ((($this)) + 28|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $123 = HEAP32[$122>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $124 = ($123|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $125 = $123; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($124) {
   ___clang_call_terminate($68); //@line 250 "modules/physics/box2d/World.cpp"
   // unreachable; //@line 250 "modules/physics/box2d/World.cpp"
  }
  $126 = ((($this)) + 32|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $127 = HEAP32[$126>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $128 = ($127|0)==($123|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($128)) {
   $scevgep$i$i$i10 = ((($127)) + -4|0);
   $129 = $scevgep$i$i$i10;
   $130 = (($129) - ($125))|0;
   $131 = $130 >>> 2;
   $132 = $131 ^ -1;
   $scevgep4$i$i$i11 = (($127) + ($132<<2)|0);
   HEAP32[$126>>2] = $scevgep4$i$i$i11; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($123); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___clang_call_terminate($68); //@line 250 "modules/physics/box2d/World.cpp"
  // unreachable; //@line 250 "modules/physics/box2d/World.cpp"
 } else {
  $5 = ((($this)) + 100|0); //@line 106 "modules/physics/box2d/World.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 106 "modules/physics/box2d/World.cpp"
  $7 = ($6|0)==(0|0); //@line 106 "modules/physics/box2d/World.cpp"
  if (!($7)) {
   $8 = HEAP32[$6>>2]|0; //@line 107 "modules/physics/box2d/World.cpp"
   $9 = ((($8)) + 4|0); //@line 107 "modules/physics/box2d/World.cpp"
   $10 = HEAP32[$9>>2]|0; //@line 107 "modules/physics/box2d/World.cpp"
   FUNCTION_TABLE_vi[$10 & 8191]($6); //@line 107 "modules/physics/box2d/World.cpp"
  }
  $11 = ((($this)) + 92|0); //@line 45 "modules/physics/box2d/World.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 45 "modules/physics/box2d/World.cpp"
  $13 = ($12|0)==(0|0); //@line 45 "modules/physics/box2d/World.cpp"
  if (!($13)) {
   $14 = HEAP32[$12>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
   $15 = ((($14)) + 4|0); //@line 46 "modules/physics/box2d/World.cpp"
   $16 = HEAP32[$15>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
   FUNCTION_TABLE_vi[$16 & 8191]($12); //@line 46 "modules/physics/box2d/World.cpp"
  }
  $17 = ((($this)) + 84|0); //@line 45 "modules/physics/box2d/World.cpp"
  $18 = HEAP32[$17>>2]|0; //@line 45 "modules/physics/box2d/World.cpp"
  $19 = ($18|0)==(0|0); //@line 45 "modules/physics/box2d/World.cpp"
  if (!($19)) {
   $20 = HEAP32[$18>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
   $21 = ((($20)) + 4|0); //@line 46 "modules/physics/box2d/World.cpp"
   $22 = HEAP32[$21>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
   FUNCTION_TABLE_vi[$22 & 8191]($18); //@line 46 "modules/physics/box2d/World.cpp"
  }
  $23 = ((($this)) + 76|0); //@line 45 "modules/physics/box2d/World.cpp"
  $24 = HEAP32[$23>>2]|0; //@line 45 "modules/physics/box2d/World.cpp"
  $25 = ($24|0)==(0|0); //@line 45 "modules/physics/box2d/World.cpp"
  if (!($25)) {
   $26 = HEAP32[$24>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
   $27 = ((($26)) + 4|0); //@line 46 "modules/physics/box2d/World.cpp"
   $28 = HEAP32[$27>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
   FUNCTION_TABLE_vi[$28 & 8191]($24); //@line 46 "modules/physics/box2d/World.cpp"
  }
  $29 = ((($this)) + 68|0); //@line 45 "modules/physics/box2d/World.cpp"
  $30 = HEAP32[$29>>2]|0; //@line 45 "modules/physics/box2d/World.cpp"
  $31 = ($30|0)==(0|0); //@line 45 "modules/physics/box2d/World.cpp"
  if (!($31)) {
   $32 = HEAP32[$30>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
   $33 = ((($32)) + 4|0); //@line 46 "modules/physics/box2d/World.cpp"
   $34 = HEAP32[$33>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
   FUNCTION_TABLE_vi[$34 & 8191]($30); //@line 46 "modules/physics/box2d/World.cpp"
  }
  $35 = ((($this)) + 52|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $36 = HEAP32[$35>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $37 = ($36|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $38 = $36; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($37)) {
   $39 = ((($this)) + 56|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $40 = HEAP32[$39>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $41 = ($40|0)==($36|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($41)) {
    $scevgep$i$i$i = ((($40)) + -4|0);
    $42 = $scevgep$i$i$i;
    $43 = (($42) - ($38))|0;
    $44 = $43 >>> 2;
    $45 = $44 ^ -1;
    $scevgep4$i$i$i = (($40) + ($45<<2)|0);
    HEAP32[$39>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($36); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  $46 = ((($this)) + 40|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $47 = HEAP32[$46>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $48 = ($47|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $49 = $47; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($48)) {
   $50 = ((($this)) + 44|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $51 = HEAP32[$50>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $52 = ($51|0)==($47|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($52)) {
    $scevgep$i$i$i4 = ((($51)) + -4|0);
    $53 = $scevgep$i$i$i4;
    $54 = (($53) - ($49))|0;
    $55 = $54 >>> 2;
    $56 = $55 ^ -1;
    $scevgep4$i$i$i5 = (($51) + ($56<<2)|0);
    HEAP32[$50>>2] = $scevgep4$i$i$i5; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($47); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  $57 = ((($this)) + 28|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $58 = HEAP32[$57>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $59 = ($58|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $60 = $58; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($59) {
   return; //@line 250 "modules/physics/box2d/World.cpp"
  }
  $61 = ((($this)) + 32|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $62 = HEAP32[$61>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $63 = ($62|0)==($58|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($63)) {
   $scevgep$i$i$i7 = ((($62)) + -4|0);
   $64 = $scevgep$i$i$i7;
   $65 = (($64) - ($60))|0;
   $66 = $65 >>> 2;
   $67 = $66 ^ -1;
   $scevgep4$i$i$i8 = (($62) + ($67<<2)|0);
   HEAP32[$61>>2] = $scevgep4$i$i$i8; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($58); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  return; //@line 250 "modules/physics/box2d/World.cpp"
 }
}
function __ZN4love7physics5box2d5World7destroyEv($this) {
 $this = $this|0;
 var $$pre = 0, $$pre2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b$01 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 20|0); //@line 555 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 555 "modules/physics/box2d/World.cpp"
 $2 = ($1|0)==(0|0); //@line 555 "modules/physics/box2d/World.cpp"
 if ($2) {
  STACKTOP = sp;return; //@line 582 "modules/physics/box2d/World.cpp"
 }
 $3 = ((($1)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $4 = HEAP32[$3>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $5 = $4 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $6 = ($5|0)==(0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 if (!($6)) {
  $7 = ((($this)) + 64|0); //@line 560 "modules/physics/box2d/World.cpp"
  HEAP8[$7>>0] = 1; //@line 560 "modules/physics/box2d/World.cpp"
  STACKTOP = sp;return; //@line 582 "modules/physics/box2d/World.cpp"
 }
 $8 = ((($1)) + 102952|0); //@line 268 "./libraries/Box2D/Dynamics/b2World.h"
 $9 = HEAP32[$8>>2]|0; //@line 268 "./libraries/Box2D/Dynamics/b2World.h"
 $10 = ($9|0)==(0|0); //@line 566 "modules/physics/box2d/World.cpp"
 $$pre2 = ((($this)) + 24|0); //@line 578 "modules/physics/box2d/World.cpp"
 do {
  if ($10) {
   $24 = $1;
  } else {
   $b$01 = $9;
   while(1) {
    $11 = ((($b$01)) + 96|0); //@line 719 "./libraries/Box2D/Dynamics/b2Body.h"
    $12 = HEAP32[$11>>2]|0; //@line 719 "./libraries/Box2D/Dynamics/b2Body.h"
    $13 = HEAP32[$$pre2>>2]|0; //@line 570 "modules/physics/box2d/World.cpp"
    $14 = ($b$01|0)==($13|0); //@line 570 "modules/physics/box2d/World.cpp"
    if (!($14)) {
     $16 = (__ZN4love8Memoizer4findEPv($b$01)|0); //@line 572 "modules/physics/box2d/World.cpp"
     $17 = ($16|0)==(0|0); //@line 573 "modules/physics/box2d/World.cpp"
     if ($17) {
      break;
     }
     __ZN4love7physics5box2d4Body7destroyEv($16); //@line 575 "modules/physics/box2d/World.cpp"
    }
    $15 = ($12|0)==(0|0); //@line 566 "modules/physics/box2d/World.cpp"
    if ($15) {
     label = 12;
     break;
    } else {
     $b$01 = $12;
    }
   }
   if ((label|0) == 12) {
    $$pre = HEAP32[$0>>2]|0; //@line 578 "modules/physics/box2d/World.cpp"
    $24 = $$pre;
    break;
   }
   $18 = (___cxa_allocate_exception(16)|0); //@line 574 "modules/physics/box2d/World.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($18|0),(356926|0),($vararg_buffer|0)); //@line 574 "modules/physics/box2d/World.cpp"
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   if ($20) {
    $21 = ___cxa_find_matching_catch_2()|0;
    $22 = tempRet0;
    ___cxa_free_exception(($18|0)); //@line 574 "modules/physics/box2d/World.cpp"
    ___resumeException($21|0);
    // unreachable;
   } else {
    ___cxa_throw(($18|0),(96|0),(344|0)); //@line 574 "modules/physics/box2d/World.cpp"
    // unreachable; //@line 574 "modules/physics/box2d/World.cpp"
   }
  }
 } while(0);
 $23 = HEAP32[$$pre2>>2]|0; //@line 578 "modules/physics/box2d/World.cpp"
 __ZN7b2World11DestroyBodyEP6b2Body($24,$23); //@line 578 "modules/physics/box2d/World.cpp"
 $25 = HEAP32[$0>>2]|0; //@line 579 "modules/physics/box2d/World.cpp"
 __ZN4love8Memoizer6removeEPv($25); //@line 579 "modules/physics/box2d/World.cpp"
 $26 = HEAP32[$0>>2]|0; //@line 580 "modules/physics/box2d/World.cpp"
 $27 = ($26|0)==(0|0); //@line 580 "modules/physics/box2d/World.cpp"
 if (!($27)) {
  __ZN7b2WorldD2Ev($26); //@line 580 "modules/physics/box2d/World.cpp"
  __ZdlPv($26); //@line 580 "modules/physics/box2d/World.cpp"
 }
 HEAP32[$0>>2] = 0; //@line 581 "modules/physics/box2d/World.cpp"
 STACKTOP = sp;return; //@line 582 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5WorldD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5WorldD2Ev($this); //@line 248 "modules/physics/box2d/World.cpp"
 __ZdlPv($this); //@line 248 "modules/physics/box2d/World.cpp"
 return; //@line 250 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World12BeginContactEP9b2Contact($this,$contact) {
 $this = $this|0;
 $contact = $contact|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 68|0); //@line 285 "modules/physics/box2d/World.cpp"
 __ZN4love7physics5box2d5World15ContactCallback7processEP9b2ContactPK16b2ContactImpulse($0,$contact,0); //@line 285 "modules/physics/box2d/World.cpp"
 return; //@line 286 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World15ContactCallback7processEP9b2ContactPK16b2ContactImpulse($this,$contact,$impulse) {
 $this = $this|0;
 $contact = $contact|0;
 $impulse = $impulse|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $args$02 = 0, $args$1 = 0, $c$03 = 0, $cobj$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[$this>>2]|0; //@line 52 "modules/physics/box2d/World.cpp"
 $1 = ($0|0)==(0|0); //@line 52 "modules/physics/box2d/World.cpp"
 if ($1) {
  STACKTOP = sp;return; //@line 96 "modules/physics/box2d/World.cpp"
 }
 $2 = ((($this)) + 4|0); //@line 52 "modules/physics/box2d/World.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 52 "modules/physics/box2d/World.cpp"
 $4 = ($3|0)==(0|0); //@line 52 "modules/physics/box2d/World.cpp"
 if ($4) {
  STACKTOP = sp;return; //@line 96 "modules/physics/box2d/World.cpp"
 }
 __ZN4love9Reference4pushEP9lua_State($0,$3); //@line 54 "modules/physics/box2d/World.cpp"
 $5 = ((($contact)) + 48|0); //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $6 = HEAP32[$5>>2]|0; //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $7 = (__ZN4love8Memoizer4findEPv($6)|0); //@line 58 "modules/physics/box2d/World.cpp"
 $8 = ($7|0)==(0|0); //@line 59 "modules/physics/box2d/World.cpp"
 if ($8) {
  $14 = (___cxa_allocate_exception(16)|0); //@line 62 "modules/physics/box2d/World.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($14|0),(444616|0),($vararg_buffer|0)); //@line 62 "modules/physics/box2d/World.cpp"
  $15 = __THREW__; __THREW__ = 0;
  $16 = $15&1;
  if (!($16)) {
   ___cxa_throw(($14|0),(96|0),(344|0)); //@line 62 "modules/physics/box2d/World.cpp"
   // unreachable; //@line 62 "modules/physics/box2d/World.cpp"
  }
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  ___cxa_free_exception(($14|0)); //@line 62 "modules/physics/box2d/World.cpp"
  $$0 = $17;$$01 = $18;
  ___resumeException($$0|0);
  // unreachable;
 }
 $9 = HEAP32[$2>>2]|0; //@line 60 "modules/physics/box2d/World.cpp"
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($9,35,$7); //@line 60 "modules/physics/box2d/World.cpp"
 $10 = ((($contact)) + 52|0); //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $11 = HEAP32[$10>>2]|0; //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $12 = (__ZN4love8Memoizer4findEPv($11)|0); //@line 67 "modules/physics/box2d/World.cpp"
 $13 = ($12|0)==(0|0); //@line 68 "modules/physics/box2d/World.cpp"
 if ($13) {
  $22 = (___cxa_allocate_exception(16)|0); //@line 71 "modules/physics/box2d/World.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($22|0),(444616|0),($vararg_buffer1|0)); //@line 71 "modules/physics/box2d/World.cpp"
  $23 = __THREW__; __THREW__ = 0;
  $24 = $23&1;
  if (!($24)) {
   ___cxa_throw(($22|0),(96|0),(344|0)); //@line 71 "modules/physics/box2d/World.cpp"
   // unreachable; //@line 71 "modules/physics/box2d/World.cpp"
  }
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  ___cxa_free_exception(($22|0)); //@line 71 "modules/physics/box2d/World.cpp"
  $$0 = $25;$$01 = $26;
  ___resumeException($$0|0);
  // unreachable;
 }
 $19 = HEAP32[$2>>2]|0; //@line 69 "modules/physics/box2d/World.cpp"
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($19,35,$12); //@line 69 "modules/physics/box2d/World.cpp"
 $20 = (__ZN4love8Memoizer4findEPv($contact)|0); //@line 74 "modules/physics/box2d/World.cpp"
 $21 = ($20|0)==(0|0); //@line 75 "modules/physics/box2d/World.cpp"
 if ($21) {
  $27 = (__Znwj(12)|0); //@line 76 "modules/physics/box2d/World.cpp"
  __THREW__ = 0;
  invoke_vii(3706,($27|0),($contact|0)); //@line 76 "modules/physics/box2d/World.cpp"
  $28 = __THREW__; __THREW__ = 0;
  $29 = $28&1;
  if ($29) {
   $30 = ___cxa_find_matching_catch_2()|0;
   $31 = tempRet0;
   __ZdlPv($27); //@line 76 "modules/physics/box2d/World.cpp"
   $$0 = $30;$$01 = $31;
   ___resumeException($$0|0);
   // unreachable;
  } else {
   $cobj$0 = $27;
  }
 } else {
  $32 = HEAP32[$20>>2]|0; //@line 78 "modules/physics/box2d/World.cpp"
  $33 = ((($32)) + 8|0); //@line 78 "modules/physics/box2d/World.cpp"
  $34 = HEAP32[$33>>2]|0; //@line 78 "modules/physics/box2d/World.cpp"
  FUNCTION_TABLE_vi[$34 & 8191]($20); //@line 78 "modules/physics/box2d/World.cpp"
  $cobj$0 = $20;
 }
 $35 = HEAP32[$2>>2]|0; //@line 80 "modules/physics/box2d/World.cpp"
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($35,33,$cobj$0); //@line 80 "modules/physics/box2d/World.cpp"
 $36 = HEAP32[$cobj$0>>2]|0; //@line 81 "modules/physics/box2d/World.cpp"
 $37 = ((($36)) + 12|0); //@line 81 "modules/physics/box2d/World.cpp"
 $38 = HEAP32[$37>>2]|0; //@line 81 "modules/physics/box2d/World.cpp"
 FUNCTION_TABLE_vi[$38 & 8191]($cobj$0); //@line 81 "modules/physics/box2d/World.cpp"
 $39 = ($impulse|0)==(0|0); //@line 84 "modules/physics/box2d/World.cpp"
 if ($39) {
  $args$1 = 3;
 } else {
  $40 = ((($impulse)) + 16|0); //@line 86 "modules/physics/box2d/World.cpp"
  $41 = HEAP32[$40>>2]|0; //@line 86 "modules/physics/box2d/World.cpp"
  $42 = ($41|0)>(0); //@line 86 "modules/physics/box2d/World.cpp"
  if ($42) {
   $args$02 = 3;$c$03 = 0;
   while(1) {
    $43 = HEAP32[$2>>2]|0; //@line 88 "modules/physics/box2d/World.cpp"
    $44 = (($impulse) + ($c$03<<2)|0); //@line 88 "modules/physics/box2d/World.cpp"
    $45 = +HEAPF32[$44>>2]; //@line 88 "modules/physics/box2d/World.cpp"
    $46 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($45)); //@line 88 "modules/physics/box2d/World.cpp"
    $47 = $46; //@line 88 "modules/physics/box2d/World.cpp"
    _lua_pushnumber($43,$47); //@line 88 "modules/physics/box2d/World.cpp"
    $48 = HEAP32[$2>>2]|0; //@line 89 "modules/physics/box2d/World.cpp"
    $49 = (((($impulse)) + 8|0) + ($c$03<<2)|0); //@line 89 "modules/physics/box2d/World.cpp"
    $50 = +HEAPF32[$49>>2]; //@line 89 "modules/physics/box2d/World.cpp"
    $51 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($50)); //@line 89 "modules/physics/box2d/World.cpp"
    $52 = $51; //@line 89 "modules/physics/box2d/World.cpp"
    _lua_pushnumber($48,$52); //@line 89 "modules/physics/box2d/World.cpp"
    $53 = (($args$02) + 2)|0; //@line 90 "modules/physics/box2d/World.cpp"
    $54 = (($c$03) + 1)|0; //@line 86 "modules/physics/box2d/World.cpp"
    $55 = HEAP32[$40>>2]|0; //@line 86 "modules/physics/box2d/World.cpp"
    $56 = ($54|0)<($55|0); //@line 86 "modules/physics/box2d/World.cpp"
    if ($56) {
     $args$02 = $53;$c$03 = $54;
    } else {
     $args$1 = $53;
     break;
    }
   }
  } else {
   $args$1 = 3;
  }
 }
 $57 = HEAP32[$2>>2]|0; //@line 93 "modules/physics/box2d/World.cpp"
 _lua_call($57,$args$1,0); //@line 93 "modules/physics/box2d/World.cpp"
 STACKTOP = sp;return; //@line 96 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World10EndContactEP9b2Contact($this,$contact) {
 $this = $this|0;
 $contact = $contact|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 76|0); //@line 290 "modules/physics/box2d/World.cpp"
 __ZN4love7physics5box2d5World15ContactCallback7processEP9b2ContactPK16b2ContactImpulse($0,$contact,0); //@line 290 "modules/physics/box2d/World.cpp"
 $1 = (__ZN4love8Memoizer4findEPv($contact)|0); //@line 293 "modules/physics/box2d/World.cpp"
 $2 = ($1|0)==(0|0); //@line 294 "modules/physics/box2d/World.cpp"
 if ($2) {
  return; //@line 296 "modules/physics/box2d/World.cpp"
 }
 __ZN4love7physics5box2d7Contact10invalidateEv($1); //@line 295 "modules/physics/box2d/World.cpp"
 return; //@line 296 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World8PreSolveEP9b2ContactPK10b2Manifold($this,$contact,$oldManifold) {
 $this = $this|0;
 $contact = $contact|0;
 $oldManifold = $oldManifold|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 84|0); //@line 301 "modules/physics/box2d/World.cpp"
 __ZN4love7physics5box2d5World15ContactCallback7processEP9b2ContactPK16b2ContactImpulse($0,$contact,0); //@line 301 "modules/physics/box2d/World.cpp"
 return; //@line 302 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World9PostSolveEP9b2ContactPK16b2ContactImpulse($this,$contact,$impulse) {
 $this = $this|0;
 $contact = $contact|0;
 $impulse = $impulse|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 92|0); //@line 306 "modules/physics/box2d/World.cpp"
 __ZN4love7physics5box2d5World15ContactCallback7processEP9b2ContactPK16b2ContactImpulse($0,$contact,$impulse); //@line 306 "modules/physics/box2d/World.cpp"
 return; //@line 307 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World13ShouldCollideEP9b2FixtureS4_($this,$fixtureA,$fixtureB) {
 $this = $this|0;
 $fixtureA = $fixtureA|0;
 $fixtureB = $fixtureB|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (__ZN4love8Memoizer4findEPv($fixtureA)|0); //@line 312 "modules/physics/box2d/World.cpp"
 $1 = (__ZN4love8Memoizer4findEPv($fixtureB)|0); //@line 313 "modules/physics/box2d/World.cpp"
 $2 = ($0|0)!=(0|0); //@line 314 "modules/physics/box2d/World.cpp"
 $3 = ($1|0)!=(0|0); //@line 314 "modules/physics/box2d/World.cpp"
 $or$cond = $2 & $3; //@line 314 "modules/physics/box2d/World.cpp"
 if ($or$cond) {
  $9 = ((($this)) + 100|0); //@line 316 "modules/physics/box2d/World.cpp"
  $10 = (__ZN4love7physics5box2d5World13ContactFilter7processEPNS1_7FixtureES5_($9,$0,$1)|0); //@line 316 "modules/physics/box2d/World.cpp"
  STACKTOP = sp;return ($10|0); //@line 316 "modules/physics/box2d/World.cpp"
 }
 $4 = (___cxa_allocate_exception(16)|0); //@line 315 "modules/physics/box2d/World.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($4|0),(444616|0),($vararg_buffer|0)); //@line 315 "modules/physics/box2d/World.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  ___cxa_free_exception(($4|0)); //@line 315 "modules/physics/box2d/World.cpp"
  ___resumeException($7|0);
  // unreachable;
 } else {
  ___cxa_throw(($4|0),(96|0),(344|0)); //@line 315 "modules/physics/box2d/World.cpp"
  // unreachable; //@line 315 "modules/physics/box2d/World.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d5World13ContactFilter7processEPNS1_7FixtureES5_($this,$a,$b) {
 $this = $this|0;
 $a = $a|0;
 $b = $b|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $filterA = 0, $filterB = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $filterA = sp + 12|0;
 $filterB = sp;
 __ZN4love7physics5box2d7Fixture13getFilterDataEPi($a,$filterA); //@line 117 "modules/physics/box2d/World.cpp"
 __ZN4love7physics5box2d7Fixture13getFilterDataEPi($b,$filterB); //@line 118 "modules/physics/box2d/World.cpp"
 $0 = ((($filterA)) + 8|0); //@line 120 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 120 "modules/physics/box2d/World.cpp"
 $2 = ($1|0)==(0); //@line 120 "modules/physics/box2d/World.cpp"
 if (!($2)) {
  $3 = ((($filterB)) + 8|0); //@line 121 "modules/physics/box2d/World.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 121 "modules/physics/box2d/World.cpp"
  $5 = ($1|0)==($4|0); //@line 121 "modules/physics/box2d/World.cpp"
  if ($5) {
   $6 = ($1|0)>(0); //@line 122 "modules/physics/box2d/World.cpp"
   $$0 = $6;
   STACKTOP = sp;return ($$0|0); //@line 137 "modules/physics/box2d/World.cpp"
  }
 }
 $7 = ((($filterA)) + 4|0); //@line 124 "modules/physics/box2d/World.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 124 "modules/physics/box2d/World.cpp"
 $9 = HEAP32[$filterB>>2]|0; //@line 124 "modules/physics/box2d/World.cpp"
 $10 = $9 & $8; //@line 124 "modules/physics/box2d/World.cpp"
 $11 = ($10|0)==(0); //@line 124 "modules/physics/box2d/World.cpp"
 if ($11) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 137 "modules/physics/box2d/World.cpp"
 }
 $12 = ((($filterB)) + 4|0); //@line 125 "modules/physics/box2d/World.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 125 "modules/physics/box2d/World.cpp"
 $14 = HEAP32[$filterA>>2]|0; //@line 125 "modules/physics/box2d/World.cpp"
 $15 = $14 & $13; //@line 125 "modules/physics/box2d/World.cpp"
 $16 = ($15|0)==(0); //@line 125 "modules/physics/box2d/World.cpp"
 if ($16) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 137 "modules/physics/box2d/World.cpp"
 }
 $17 = HEAP32[$this>>2]|0; //@line 128 "modules/physics/box2d/World.cpp"
 $18 = ($17|0)==(0|0); //@line 128 "modules/physics/box2d/World.cpp"
 if ($18) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 137 "modules/physics/box2d/World.cpp"
 }
 $19 = ((($this)) + 4|0); //@line 128 "modules/physics/box2d/World.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 128 "modules/physics/box2d/World.cpp"
 $21 = ($20|0)==(0|0); //@line 128 "modules/physics/box2d/World.cpp"
 if ($21) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 137 "modules/physics/box2d/World.cpp"
 }
 __ZN4love9Reference4pushEP9lua_State($17,$20); //@line 130 "modules/physics/box2d/World.cpp"
 $22 = HEAP32[$19>>2]|0; //@line 131 "modules/physics/box2d/World.cpp"
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($22,35,$a); //@line 131 "modules/physics/box2d/World.cpp"
 $23 = HEAP32[$19>>2]|0; //@line 132 "modules/physics/box2d/World.cpp"
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($23,35,$b); //@line 132 "modules/physics/box2d/World.cpp"
 $24 = HEAP32[$19>>2]|0; //@line 133 "modules/physics/box2d/World.cpp"
 _lua_call($24,2,1); //@line 133 "modules/physics/box2d/World.cpp"
 $25 = HEAP32[$19>>2]|0; //@line 134 "modules/physics/box2d/World.cpp"
 $26 = (__ZN4love14luax_tobooleanEP9lua_Statei($25,-1)|0); //@line 134 "modules/physics/box2d/World.cpp"
 $$0 = $26;
 STACKTOP = sp;return ($$0|0); //@line 137 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World10SayGoodbyeEP9b2Fixture($this,$fixture) {
 $this = $this|0;
 $fixture = $fixture|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love8Memoizer4findEPv($fixture)|0); //@line 207 "modules/physics/box2d/World.cpp"
 $1 = ($0|0)==(0|0); //@line 209 "modules/physics/box2d/World.cpp"
 if ($1) {
  return; //@line 210 "modules/physics/box2d/World.cpp"
 }
 __ZN4love7physics5box2d7Fixture7destroyEb($0,1); //@line 209 "modules/physics/box2d/World.cpp"
 return; //@line 210 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World10SayGoodbyeEP7b2Joint($this,$joint) {
 $this = $this|0;
 $joint = $joint|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love8Memoizer4findEPv($joint)|0); //@line 214 "modules/physics/box2d/World.cpp"
 $1 = ($0|0)==(0|0); //@line 216 "modules/physics/box2d/World.cpp"
 if ($1) {
  return; //@line 217 "modules/physics/box2d/World.cpp"
 }
 __ZN4love7physics5box2d5Joint12destroyJointEb($0,1); //@line 216 "modules/physics/box2d/World.cpp"
 return; //@line 217 "modules/physics/box2d/World.cpp"
}
function __ZThn8_N4love7physics5box2d5WorldD1Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 247 "modules/physics/box2d/World.cpp"
 __ZN4love7physics5box2d5WorldD2Ev($0); //@line 247 "modules/physics/box2d/World.cpp"
 return; //@line 247 "modules/physics/box2d/World.cpp"
}
function __ZThn8_N4love7physics5box2d5WorldD0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 247 "modules/physics/box2d/World.cpp"
 __ZN4love7physics5box2d5WorldD2Ev($0); //@line 248 "modules/physics/box2d/World.cpp"
 __ZdlPv($0); //@line 248 "modules/physics/box2d/World.cpp"
 return; //@line 247 "modules/physics/box2d/World.cpp"
}
function __ZThn8_N4love7physics5box2d5World12BeginContactEP9b2Contact($this,$contact) {
 $this = $this|0;
 $contact = $contact|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 283 "modules/physics/box2d/World.cpp"
 $1 = ((($0)) + 68|0); //@line 285 "modules/physics/box2d/World.cpp"
 __ZN4love7physics5box2d5World15ContactCallback7processEP9b2ContactPK16b2ContactImpulse($1,$contact,0); //@line 285 "modules/physics/box2d/World.cpp"
 return; //@line 283 "modules/physics/box2d/World.cpp"
}
function __ZThn8_N4love7physics5box2d5World10EndContactEP9b2Contact($this,$contact) {
 $this = $this|0;
 $contact = $contact|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 288 "modules/physics/box2d/World.cpp"
 $1 = ((($0)) + 76|0); //@line 290 "modules/physics/box2d/World.cpp"
 __ZN4love7physics5box2d5World15ContactCallback7processEP9b2ContactPK16b2ContactImpulse($1,$contact,0); //@line 290 "modules/physics/box2d/World.cpp"
 $2 = (__ZN4love8Memoizer4findEPv($contact)|0); //@line 293 "modules/physics/box2d/World.cpp"
 $3 = ($2|0)==(0|0); //@line 294 "modules/physics/box2d/World.cpp"
 if ($3) {
  return; //@line 288 "modules/physics/box2d/World.cpp"
 }
 __ZN4love7physics5box2d7Contact10invalidateEv($2); //@line 295 "modules/physics/box2d/World.cpp"
 return; //@line 288 "modules/physics/box2d/World.cpp"
}
function __ZThn8_N4love7physics5box2d5World8PreSolveEP9b2ContactPK10b2Manifold($this,$contact,$oldManifold) {
 $this = $this|0;
 $contact = $contact|0;
 $oldManifold = $oldManifold|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 298 "modules/physics/box2d/World.cpp"
 $1 = ((($0)) + 84|0); //@line 301 "modules/physics/box2d/World.cpp"
 __ZN4love7physics5box2d5World15ContactCallback7processEP9b2ContactPK16b2ContactImpulse($1,$contact,0); //@line 301 "modules/physics/box2d/World.cpp"
 return; //@line 298 "modules/physics/box2d/World.cpp"
}
function __ZThn8_N4love7physics5box2d5World9PostSolveEP9b2ContactPK16b2ContactImpulse($this,$contact,$impulse) {
 $this = $this|0;
 $contact = $contact|0;
 $impulse = $impulse|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 304 "modules/physics/box2d/World.cpp"
 $1 = ((($0)) + 92|0); //@line 306 "modules/physics/box2d/World.cpp"
 __ZN4love7physics5box2d5World15ContactCallback7processEP9b2ContactPK16b2ContactImpulse($1,$contact,$impulse); //@line 306 "modules/physics/box2d/World.cpp"
 return; //@line 304 "modules/physics/box2d/World.cpp"
}
function __ZThn12_N4love7physics5box2d5WorldD1Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -12|0); //@line 247 "modules/physics/box2d/World.cpp"
 __ZN4love7physics5box2d5WorldD2Ev($0); //@line 247 "modules/physics/box2d/World.cpp"
 return; //@line 247 "modules/physics/box2d/World.cpp"
}
function __ZThn12_N4love7physics5box2d5WorldD0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -12|0); //@line 247 "modules/physics/box2d/World.cpp"
 __ZN4love7physics5box2d5WorldD2Ev($0); //@line 248 "modules/physics/box2d/World.cpp"
 __ZdlPv($0); //@line 248 "modules/physics/box2d/World.cpp"
 return; //@line 247 "modules/physics/box2d/World.cpp"
}
function __ZThn12_N4love7physics5box2d5World13ShouldCollideEP9b2FixtureS4_($this,$fixtureA,$fixtureB) {
 $this = $this|0;
 $fixtureA = $fixtureA|0;
 $fixtureB = $fixtureB|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -12|0); //@line 309 "modules/physics/box2d/World.cpp"
 $1 = (__ZN4love7physics5box2d5World13ShouldCollideEP9b2FixtureS4_($0,$fixtureA,$fixtureB)|0); //@line 309 "modules/physics/box2d/World.cpp"
 return ($1|0); //@line 309 "modules/physics/box2d/World.cpp"
}
function __ZThn16_N4love7physics5box2d5WorldD1Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -16|0); //@line 247 "modules/physics/box2d/World.cpp"
 __ZN4love7physics5box2d5WorldD2Ev($0); //@line 247 "modules/physics/box2d/World.cpp"
 return; //@line 247 "modules/physics/box2d/World.cpp"
}
function __ZThn16_N4love7physics5box2d5WorldD0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -16|0); //@line 247 "modules/physics/box2d/World.cpp"
 __ZN4love7physics5box2d5WorldD2Ev($0); //@line 248 "modules/physics/box2d/World.cpp"
 __ZdlPv($0); //@line 248 "modules/physics/box2d/World.cpp"
 return; //@line 247 "modules/physics/box2d/World.cpp"
}
function __ZThn16_N4love7physics5box2d5World10SayGoodbyeEP7b2Joint($this,$joint) {
 $this = $this|0;
 $joint = $joint|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love8Memoizer4findEPv($joint)|0); //@line 214 "modules/physics/box2d/World.cpp"
 $1 = ($0|0)==(0|0); //@line 216 "modules/physics/box2d/World.cpp"
 if ($1) {
  return; //@line 212 "modules/physics/box2d/World.cpp"
 }
 __ZN4love7physics5box2d5Joint12destroyJointEb($0,1); //@line 216 "modules/physics/box2d/World.cpp"
 return; //@line 212 "modules/physics/box2d/World.cpp"
}
function __ZThn16_N4love7physics5box2d5World10SayGoodbyeEP9b2Fixture($this,$fixture) {
 $this = $this|0;
 $fixture = $fixture|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love8Memoizer4findEPv($fixture)|0); //@line 207 "modules/physics/box2d/World.cpp"
 $1 = ($0|0)==(0|0); //@line 209 "modules/physics/box2d/World.cpp"
 if ($1) {
  return; //@line 205 "modules/physics/box2d/World.cpp"
 }
 __ZN4love7physics5box2d7Fixture7destroyEb($0,1); //@line 209 "modules/physics/box2d/World.cpp"
 return; //@line 205 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5WorldC2E6b2Vec2b($this,$gravity,$sleep) {
 $this = $this|0;
 $gravity = $gravity|0;
 $sleep = $sleep|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $def = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i10 = 0, $scevgep$i$i$i7 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i11 = 0, $scevgep4$i$i$i8 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 56|0;
 $def = sp;
 __ZN4love6ObjectC2Ev($this); //@line 233 "modules/physics/box2d/World.cpp"
 $1 = ((($this)) + 8|0); //@line 236 "modules/physics/box2d/World.cpp"
 $2 = ((($this)) + 12|0); //@line 236 "modules/physics/box2d/World.cpp"
 $3 = ((($this)) + 16|0); //@line 236 "modules/physics/box2d/World.cpp"
 HEAP32[$this>>2] = (17368); //@line 236 "modules/physics/box2d/World.cpp"
 HEAP32[$1>>2] = (17420); //@line 236 "modules/physics/box2d/World.cpp"
 HEAP32[$2>>2] = (17452); //@line 236 "modules/physics/box2d/World.cpp"
 HEAP32[$3>>2] = (17472); //@line 236 "modules/physics/box2d/World.cpp"
 $4 = ((($this)) + 20|0); //@line 234 "modules/physics/box2d/World.cpp"
 HEAP32[$4>>2] = 0; //@line 234 "modules/physics/box2d/World.cpp"
 $5 = ((($this)) + 28|0); //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ((($this)) + 32|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ((($this)) + 40|0); //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($this)) + 44|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ((($this)) + 52|0); //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = ((($this)) + 56|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ((($this)) + 68|0); //@line 38 "modules/physics/box2d/World.cpp"
 $12 = ((($this)) + 76|0); //@line 38 "modules/physics/box2d/World.cpp"
 $13 = ((($this)) + 84|0); //@line 38 "modules/physics/box2d/World.cpp"
 $14 = ((($this)) + 92|0); //@line 38 "modules/physics/box2d/World.cpp"
 $15 = ((($this)) + 100|0); //@line 99 "modules/physics/box2d/World.cpp"
 dest=$5; stop=dest+36|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));HEAP8[$5+36>>0]=0|0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 dest=$11; stop=dest+40|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 38 "modules/physics/box2d/World.cpp"
 __THREW__ = 0;
 $16 = (invoke_ii(3348,103028)|0); //@line 237 "modules/physics/box2d/World.cpp"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 do {
  if ($18) {
   label = 11;
  } else {
   __THREW__ = 0;
   invoke_vii(3695,($0|0),($gravity|0)); //@line 237 "modules/physics/box2d/World.cpp"
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   if (!($20)) {
    __THREW__ = 0;
    invoke_vii(3707,($16|0),($0|0)); //@line 237 "modules/physics/box2d/World.cpp"
    $21 = __THREW__; __THREW__ = 0;
    $22 = $21&1;
    if (!($22)) {
     HEAP32[$4>>2] = $16; //@line 237 "modules/physics/box2d/World.cpp"
     __THREW__ = 0;
     invoke_vii(3708,($16|0),($sleep|0)); //@line 238 "modules/physics/box2d/World.cpp"
     $23 = __THREW__; __THREW__ = 0;
     $24 = $23&1;
     if ($24) {
      label = 11;
      break;
     }
     $25 = HEAP32[$4>>2]|0; //@line 239 "modules/physics/box2d/World.cpp"
     __THREW__ = 0;
     invoke_vii(3709,($25|0),($1|0)); //@line 239 "modules/physics/box2d/World.cpp"
     $26 = __THREW__; __THREW__ = 0;
     $27 = $26&1;
     if ($27) {
      label = 11;
      break;
     }
     $28 = HEAP32[$4>>2]|0; //@line 240 "modules/physics/box2d/World.cpp"
     __THREW__ = 0;
     invoke_vii(3710,($28|0),($2|0)); //@line 240 "modules/physics/box2d/World.cpp"
     $29 = __THREW__; __THREW__ = 0;
     $30 = $29&1;
     if ($30) {
      label = 11;
      break;
     }
     $31 = HEAP32[$4>>2]|0; //@line 241 "modules/physics/box2d/World.cpp"
     __THREW__ = 0;
     invoke_vii(3711,($31|0),($3|0)); //@line 241 "modules/physics/box2d/World.cpp"
     $32 = __THREW__; __THREW__ = 0;
     $33 = $32&1;
     if ($33) {
      label = 11;
      break;
     }
     $34 = ((($def)) + 44|0); //@line 56 "./libraries/Box2D/Dynamics/b2Body.h"
     HEAP32[$34>>2] = 0; //@line 56 "./libraries/Box2D/Dynamics/b2Body.h"
     $35 = ((($def)) + 4|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
     $36 = ((($def)) + 36|0); //@line 63 "./libraries/Box2D/Dynamics/b2Body.h"
     ;HEAP32[$35>>2]=0|0;HEAP32[$35+4>>2]=0|0;HEAP32[$35+8>>2]=0|0;HEAP32[$35+12>>2]=0|0;HEAP32[$35+16>>2]=0|0;HEAP32[$35+20>>2]=0|0;HEAP32[$35+24>>2]=0|0;HEAP32[$35+28>>2]=0|0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
     HEAP8[$36>>0] = 1; //@line 63 "./libraries/Box2D/Dynamics/b2Body.h"
     $37 = ((($def)) + 37|0); //@line 64 "./libraries/Box2D/Dynamics/b2Body.h"
     HEAP8[$37>>0] = 1; //@line 64 "./libraries/Box2D/Dynamics/b2Body.h"
     $38 = ((($def)) + 38|0); //@line 65 "./libraries/Box2D/Dynamics/b2Body.h"
     HEAP8[$38>>0] = 0; //@line 65 "./libraries/Box2D/Dynamics/b2Body.h"
     $39 = ((($def)) + 39|0); //@line 66 "./libraries/Box2D/Dynamics/b2Body.h"
     HEAP8[$39>>0] = 0; //@line 66 "./libraries/Box2D/Dynamics/b2Body.h"
     HEAP32[$def>>2] = 0; //@line 67 "./libraries/Box2D/Dynamics/b2Body.h"
     $40 = ((($def)) + 40|0); //@line 68 "./libraries/Box2D/Dynamics/b2Body.h"
     HEAP8[$40>>0] = 1; //@line 68 "./libraries/Box2D/Dynamics/b2Body.h"
     $41 = ((($def)) + 48|0); //@line 69 "./libraries/Box2D/Dynamics/b2Body.h"
     HEAPF32[$41>>2] = 1.0; //@line 69 "./libraries/Box2D/Dynamics/b2Body.h"
     $42 = HEAP32[$4>>2]|0; //@line 243 "modules/physics/box2d/World.cpp"
     __THREW__ = 0;
     $43 = (invoke_iii(3712,($42|0),($def|0))|0); //@line 243 "modules/physics/box2d/World.cpp"
     $44 = __THREW__; __THREW__ = 0;
     $45 = $44&1;
     if ($45) {
      label = 11;
      break;
     }
     $46 = ((($this)) + 24|0); //@line 243 "modules/physics/box2d/World.cpp"
     HEAP32[$46>>2] = $43; //@line 243 "modules/physics/box2d/World.cpp"
     $47 = HEAP32[$4>>2]|0; //@line 244 "modules/physics/box2d/World.cpp"
     __THREW__ = 0;
     invoke_vii(3703,($47|0),($this|0)); //@line 244 "modules/physics/box2d/World.cpp"
     $48 = __THREW__; __THREW__ = 0;
     $49 = $48&1;
     if ($49) {
      label = 11;
      break;
     }
     STACKTOP = sp;return; //@line 245 "modules/physics/box2d/World.cpp"
    }
   }
   $52 = ___cxa_find_matching_catch_2()|0;
   $53 = tempRet0;
   __ZdlPv($16); //@line 237 "modules/physics/box2d/World.cpp"
   $$0 = $53;$$01 = $52;
  }
 } while(0);
 if ((label|0) == 11) {
  $50 = ___cxa_find_matching_catch_2()|0;
  $51 = tempRet0;
  $$0 = $51;$$01 = $50;
 }
 $54 = HEAP32[$15>>2]|0; //@line 106 "modules/physics/box2d/World.cpp"
 $55 = ($54|0)==(0|0); //@line 106 "modules/physics/box2d/World.cpp"
 if (!($55)) {
  $56 = HEAP32[$54>>2]|0; //@line 107 "modules/physics/box2d/World.cpp"
  $57 = ((($56)) + 4|0); //@line 107 "modules/physics/box2d/World.cpp"
  $58 = HEAP32[$57>>2]|0; //@line 107 "modules/physics/box2d/World.cpp"
  FUNCTION_TABLE_vi[$58 & 8191]($54); //@line 107 "modules/physics/box2d/World.cpp"
 }
 $59 = HEAP32[$14>>2]|0; //@line 45 "modules/physics/box2d/World.cpp"
 $60 = ($59|0)==(0|0); //@line 45 "modules/physics/box2d/World.cpp"
 if (!($60)) {
  $61 = HEAP32[$59>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
  $62 = ((($61)) + 4|0); //@line 46 "modules/physics/box2d/World.cpp"
  $63 = HEAP32[$62>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
  FUNCTION_TABLE_vi[$63 & 8191]($59); //@line 46 "modules/physics/box2d/World.cpp"
 }
 $64 = HEAP32[$13>>2]|0; //@line 45 "modules/physics/box2d/World.cpp"
 $65 = ($64|0)==(0|0); //@line 45 "modules/physics/box2d/World.cpp"
 if (!($65)) {
  $66 = HEAP32[$64>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
  $67 = ((($66)) + 4|0); //@line 46 "modules/physics/box2d/World.cpp"
  $68 = HEAP32[$67>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
  FUNCTION_TABLE_vi[$68 & 8191]($64); //@line 46 "modules/physics/box2d/World.cpp"
 }
 $69 = HEAP32[$12>>2]|0; //@line 45 "modules/physics/box2d/World.cpp"
 $70 = ($69|0)==(0|0); //@line 45 "modules/physics/box2d/World.cpp"
 if (!($70)) {
  $71 = HEAP32[$69>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
  $72 = ((($71)) + 4|0); //@line 46 "modules/physics/box2d/World.cpp"
  $73 = HEAP32[$72>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
  FUNCTION_TABLE_vi[$73 & 8191]($69); //@line 46 "modules/physics/box2d/World.cpp"
 }
 $74 = HEAP32[$11>>2]|0; //@line 45 "modules/physics/box2d/World.cpp"
 $75 = ($74|0)==(0|0); //@line 45 "modules/physics/box2d/World.cpp"
 if (!($75)) {
  $76 = HEAP32[$74>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
  $77 = ((($76)) + 4|0); //@line 46 "modules/physics/box2d/World.cpp"
  $78 = HEAP32[$77>>2]|0; //@line 46 "modules/physics/box2d/World.cpp"
  FUNCTION_TABLE_vi[$78 & 8191]($74); //@line 46 "modules/physics/box2d/World.cpp"
 }
 $79 = HEAP32[$9>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $80 = ($79|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $81 = $79; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($80)) {
  $82 = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $83 = ($82|0)==($79|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($83)) {
   $scevgep$i$i$i10 = ((($82)) + -4|0);
   $84 = $scevgep$i$i$i10;
   $85 = (($84) - ($81))|0;
   $86 = $85 >>> 2;
   $87 = $86 ^ -1;
   $scevgep4$i$i$i11 = (($82) + ($87<<2)|0);
   HEAP32[$10>>2] = $scevgep4$i$i$i11; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($79); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $88 = HEAP32[$7>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $89 = ($88|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $90 = $88; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($89)) {
  $91 = HEAP32[$8>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $92 = ($91|0)==($88|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($92)) {
   $scevgep$i$i$i7 = ((($91)) + -4|0);
   $93 = $scevgep$i$i$i7;
   $94 = (($93) - ($90))|0;
   $95 = $94 >>> 2;
   $96 = $95 ^ -1;
   $scevgep4$i$i$i8 = (($91) + ($96<<2)|0);
   HEAP32[$8>>2] = $scevgep4$i$i$i8; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($88); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $97 = HEAP32[$5>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $98 = ($97|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $99 = $97; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($98) {
  ___resumeException($$01|0);
  // unreachable;
 }
 $100 = HEAP32[$6>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $101 = ($100|0)==($97|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($101)) {
  $scevgep$i$i$i = ((($100)) + -4|0);
  $102 = $scevgep$i$i$i;
  $103 = (($102) - ($99))|0;
  $104 = $103 >>> 2;
  $105 = $104 ^ -1;
  $scevgep4$i$i$i = (($100) + ($105<<2)|0);
  HEAP32[$6>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($97); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN4love7physics5box2d5World6updateEf($this,$dt) {
 $this = $this|0;
 $dt = +$dt;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$014 = 0, $__begin2$sroa$0$010 = 0, $__begin5$sroa$0$09 = 0, $__new_last2$i$i$i = 0, $__new_last2$i$i$i1 = 0, $__new_last2$i$i$i5 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i2 = 0, $scevgep$i$i$i6 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i3 = 0, $scevgep4$i$i$i7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 254 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 254 "modules/physics/box2d/World.cpp"
 __ZN7b2World4StepEfii($1,$dt,8,6); //@line 254 "modules/physics/box2d/World.cpp"
 $2 = ((($this)) + 28|0); //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ((($this)) + 32|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$4>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($3|0)==($5|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if (!($6)) {
  $__begin$sroa$0$014 = $3;
  while(1) {
   $12 = HEAP32[$__begin$sroa$0$014>>2]|0; //@line 257 "modules/physics/box2d/World.cpp"
   $13 = ((($12)) + 8|0); //@line 259 "modules/physics/box2d/World.cpp"
   $14 = HEAP32[$13>>2]|0; //@line 259 "modules/physics/box2d/World.cpp"
   $15 = ($14|0)==(0|0); //@line 259 "modules/physics/box2d/World.cpp"
   if (!($15)) {
    __ZN4love7physics5box2d4Body7destroyEv($12); //@line 259 "modules/physics/box2d/World.cpp"
   }
   $16 = HEAP32[$12>>2]|0; //@line 261 "modules/physics/box2d/World.cpp"
   $17 = ((($16)) + 12|0); //@line 261 "modules/physics/box2d/World.cpp"
   $18 = HEAP32[$17>>2]|0; //@line 261 "modules/physics/box2d/World.cpp"
   FUNCTION_TABLE_vi[$18 & 8191]($12); //@line 261 "modules/physics/box2d/World.cpp"
   $19 = ((($__begin$sroa$0$014)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $20 = ($19|0)==($5|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($20) {
    break;
   } else {
    $__begin$sroa$0$014 = $19;
   }
  }
 }
 $7 = ((($this)) + 40|0); //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = HEAP32[$7>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ((($this)) + 44|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = HEAP32[$9>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($8|0)==($10|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if (!($11)) {
  $__begin2$sroa$0$010 = $8;
  while(1) {
   $26 = HEAP32[$__begin2$sroa$0$010>>2]|0; //@line 263 "modules/physics/box2d/World.cpp"
   $27 = (__ZNK4love7physics5box2d7Fixture7isValidEv($26)|0); //@line 265 "modules/physics/box2d/World.cpp"
   if ($27) {
    __ZN4love7physics5box2d7Fixture7destroyEb($26,0); //@line 265 "modules/physics/box2d/World.cpp"
   }
   $28 = HEAP32[$26>>2]|0; //@line 267 "modules/physics/box2d/World.cpp"
   $29 = ((($28)) + 12|0); //@line 267 "modules/physics/box2d/World.cpp"
   $30 = HEAP32[$29>>2]|0; //@line 267 "modules/physics/box2d/World.cpp"
   FUNCTION_TABLE_vi[$30 & 8191]($26); //@line 267 "modules/physics/box2d/World.cpp"
   $31 = ((($__begin2$sroa$0$010)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $32 = ($31|0)==($10|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($32) {
    break;
   } else {
    $__begin2$sroa$0$010 = $31;
   }
  }
 }
 $21 = ((($this)) + 52|0); //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $22 = HEAP32[$21>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $23 = ((($this)) + 56|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $24 = HEAP32[$23>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $25 = ($22|0)==($24|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if (!($25)) {
  $__begin5$sroa$0$09 = $22;
  while(1) {
   $57 = HEAP32[$__begin5$sroa$0$09>>2]|0; //@line 269 "modules/physics/box2d/World.cpp"
   $58 = (__ZNK4love7physics5box2d5Joint7isValidEv($57)|0); //@line 271 "modules/physics/box2d/World.cpp"
   if ($58) {
    __ZN4love7physics5box2d5Joint12destroyJointEb($57,0); //@line 271 "modules/physics/box2d/World.cpp"
   }
   $59 = HEAP32[$57>>2]|0; //@line 273 "modules/physics/box2d/World.cpp"
   $60 = ((($59)) + 12|0); //@line 273 "modules/physics/box2d/World.cpp"
   $61 = HEAP32[$60>>2]|0; //@line 273 "modules/physics/box2d/World.cpp"
   FUNCTION_TABLE_vi[$61 & 8191]($57); //@line 273 "modules/physics/box2d/World.cpp"
   $62 = ((($__begin5$sroa$0$09)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $63 = ($62|0)==($24|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($63) {
    break;
   } else {
    $__begin5$sroa$0$09 = $62;
   }
  }
 }
 $33 = HEAP32[$2>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $34 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $35 = ($34|0)==($33|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($35)) {
  $__new_last2$i$i$i5 = $33; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $scevgep$i$i$i6 = ((($34)) + -4|0);
  $36 = $scevgep$i$i$i6;
  $37 = (($36) - ($__new_last2$i$i$i5))|0;
  $38 = $37 >>> 2;
  $39 = $38 ^ -1;
  $scevgep4$i$i$i7 = (($34) + ($39<<2)|0);
  HEAP32[$4>>2] = $scevgep4$i$i$i7; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = HEAP32[$7>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $41 = HEAP32[$9>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $42 = ($41|0)==($40|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($42)) {
  $__new_last2$i$i$i1 = $40; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $scevgep$i$i$i2 = ((($41)) + -4|0);
  $43 = $scevgep$i$i$i2;
  $44 = (($43) - ($__new_last2$i$i$i1))|0;
  $45 = $44 >>> 2;
  $46 = $45 ^ -1;
  $scevgep4$i$i$i3 = (($41) + ($46<<2)|0);
  HEAP32[$9>>2] = $scevgep4$i$i$i3; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $47 = HEAP32[$21>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $48 = HEAP32[$23>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $49 = ($48|0)==($47|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($49)) {
  $__new_last2$i$i$i = $47; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $scevgep$i$i$i = ((($48)) + -4|0);
  $50 = $scevgep$i$i$i;
  $51 = (($50) - ($__new_last2$i$i$i))|0;
  $52 = $51 >>> 2;
  $53 = $52 ^ -1;
  $scevgep4$i$i$i = (($48) + ($53<<2)|0);
  HEAP32[$23>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $54 = ((($this)) + 64|0); //@line 279 "modules/physics/box2d/World.cpp"
 $55 = HEAP8[$54>>0]|0; //@line 279 "modules/physics/box2d/World.cpp"
 $56 = ($55<<24>>24)==(0); //@line 279 "modules/physics/box2d/World.cpp"
 if ($56) {
  return; //@line 281 "modules/physics/box2d/World.cpp"
 }
 __ZN4love7physics5box2d5World7destroyEv($this); //@line 280 "modules/physics/box2d/World.cpp"
 return; //@line 281 "modules/physics/box2d/World.cpp"
}
function __ZNK4love7physics5box2d5World7isValidEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 321 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 321 "modules/physics/box2d/World.cpp"
 $2 = ($1|0)!=(0|0); //@line 321 "modules/physics/box2d/World.cpp"
 return ($2|0); //@line 321 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World12setCallbacksEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_gettop($L)|0); //@line 326 "modules/physics/box2d/World.cpp"
 $1 = (_lua_type($L,1)|0); //@line 330 "modules/physics/box2d/World.cpp"
 $2 = ($1|0)<(1); //@line 330 "modules/physics/box2d/World.cpp"
 if (!($2)) {
  _luaL_checktype($L,1,6); //@line 331 "modules/physics/box2d/World.cpp"
 }
 $3 = (_lua_type($L,2)|0); //@line 330 "modules/physics/box2d/World.cpp"
 $4 = ($3|0)<(1); //@line 330 "modules/physics/box2d/World.cpp"
 if (!($4)) {
  _luaL_checktype($L,2,6); //@line 331 "modules/physics/box2d/World.cpp"
 }
 $40 = (_lua_type($L,3)|0); //@line 330 "modules/physics/box2d/World.cpp"
 $41 = ($40|0)<(1); //@line 330 "modules/physics/box2d/World.cpp"
 if (!($41)) {
  _luaL_checktype($L,3,6); //@line 331 "modules/physics/box2d/World.cpp"
 }
 $42 = (_lua_type($L,4)|0); //@line 330 "modules/physics/box2d/World.cpp"
 $43 = ($42|0)<(1); //@line 330 "modules/physics/box2d/World.cpp"
 if (!($43)) {
  _luaL_checktype($L,4,6); //@line 331 "modules/physics/box2d/World.cpp"
 }
 $9 = ((($this)) + 68|0); //@line 334 "modules/physics/box2d/World.cpp"
 $6 = HEAP32[$9>>2]|0; //@line 334 "modules/physics/box2d/World.cpp"
 $44 = ($6|0)==(0|0); //@line 334 "modules/physics/box2d/World.cpp"
 if (!($44)) {
  $5 = HEAP32[$6>>2]|0; //@line 334 "modules/physics/box2d/World.cpp"
  $7 = ((($5)) + 4|0); //@line 334 "modules/physics/box2d/World.cpp"
  $8 = HEAP32[$7>>2]|0; //@line 334 "modules/physics/box2d/World.cpp"
  FUNCTION_TABLE_vi[$8 & 8191]($6); //@line 334 "modules/physics/box2d/World.cpp"
 }
 HEAP32[$9>>2] = 0; //@line 335 "modules/physics/box2d/World.cpp"
 $10 = ((($this)) + 76|0); //@line 337 "modules/physics/box2d/World.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 337 "modules/physics/box2d/World.cpp"
 $12 = ($11|0)==(0|0); //@line 337 "modules/physics/box2d/World.cpp"
 if (!($12)) {
  $13 = HEAP32[$11>>2]|0; //@line 337 "modules/physics/box2d/World.cpp"
  $14 = ((($13)) + 4|0); //@line 337 "modules/physics/box2d/World.cpp"
  $15 = HEAP32[$14>>2]|0; //@line 337 "modules/physics/box2d/World.cpp"
  FUNCTION_TABLE_vi[$15 & 8191]($11); //@line 337 "modules/physics/box2d/World.cpp"
 }
 HEAP32[$10>>2] = 0; //@line 338 "modules/physics/box2d/World.cpp"
 $16 = ((($this)) + 84|0); //@line 340 "modules/physics/box2d/World.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 340 "modules/physics/box2d/World.cpp"
 $18 = ($17|0)==(0|0); //@line 340 "modules/physics/box2d/World.cpp"
 if (!($18)) {
  $19 = HEAP32[$17>>2]|0; //@line 340 "modules/physics/box2d/World.cpp"
  $20 = ((($19)) + 4|0); //@line 340 "modules/physics/box2d/World.cpp"
  $21 = HEAP32[$20>>2]|0; //@line 340 "modules/physics/box2d/World.cpp"
  FUNCTION_TABLE_vi[$21 & 8191]($17); //@line 340 "modules/physics/box2d/World.cpp"
 }
 HEAP32[$16>>2] = 0; //@line 341 "modules/physics/box2d/World.cpp"
 $22 = ((($this)) + 92|0); //@line 343 "modules/physics/box2d/World.cpp"
 $23 = HEAP32[$22>>2]|0; //@line 343 "modules/physics/box2d/World.cpp"
 $24 = ($23|0)==(0|0); //@line 343 "modules/physics/box2d/World.cpp"
 if (!($24)) {
  $25 = HEAP32[$23>>2]|0; //@line 343 "modules/physics/box2d/World.cpp"
  $26 = ((($25)) + 4|0); //@line 343 "modules/physics/box2d/World.cpp"
  $27 = HEAP32[$26>>2]|0; //@line 343 "modules/physics/box2d/World.cpp"
  FUNCTION_TABLE_vi[$27 & 8191]($23); //@line 343 "modules/physics/box2d/World.cpp"
 }
 HEAP32[$22>>2] = 0; //@line 344 "modules/physics/box2d/World.cpp"
 $28 = ($0|0)>(0); //@line 346 "modules/physics/box2d/World.cpp"
 if (!($28)) {
  return 0; //@line 374 "modules/physics/box2d/World.cpp"
 }
 _lua_pushvalue($L,1); //@line 348 "modules/physics/box2d/World.cpp"
 $29 = (__ZN4love10luax_refifEP9lua_Statei($L,6)|0); //@line 349 "modules/physics/box2d/World.cpp"
 HEAP32[$9>>2] = $29; //@line 349 "modules/physics/box2d/World.cpp"
 $30 = ((($this)) + 72|0); //@line 350 "modules/physics/box2d/World.cpp"
 HEAP32[$30>>2] = $L; //@line 350 "modules/physics/box2d/World.cpp"
 $31 = ($0|0)>(1); //@line 353 "modules/physics/box2d/World.cpp"
 if (!($31)) {
  return 0; //@line 374 "modules/physics/box2d/World.cpp"
 }
 _lua_pushvalue($L,2); //@line 355 "modules/physics/box2d/World.cpp"
 $32 = (__ZN4love10luax_refifEP9lua_Statei($L,6)|0); //@line 356 "modules/physics/box2d/World.cpp"
 HEAP32[$10>>2] = $32; //@line 356 "modules/physics/box2d/World.cpp"
 $33 = ((($this)) + 80|0); //@line 357 "modules/physics/box2d/World.cpp"
 HEAP32[$33>>2] = $L; //@line 357 "modules/physics/box2d/World.cpp"
 $34 = ($0|0)>(2); //@line 360 "modules/physics/box2d/World.cpp"
 if (!($34)) {
  return 0; //@line 374 "modules/physics/box2d/World.cpp"
 }
 _lua_pushvalue($L,3); //@line 362 "modules/physics/box2d/World.cpp"
 $35 = (__ZN4love10luax_refifEP9lua_Statei($L,6)|0); //@line 363 "modules/physics/box2d/World.cpp"
 HEAP32[$16>>2] = $35; //@line 363 "modules/physics/box2d/World.cpp"
 $36 = ((($this)) + 88|0); //@line 364 "modules/physics/box2d/World.cpp"
 HEAP32[$36>>2] = $L; //@line 364 "modules/physics/box2d/World.cpp"
 $37 = ($0|0)>(3); //@line 367 "modules/physics/box2d/World.cpp"
 if (!($37)) {
  return 0; //@line 374 "modules/physics/box2d/World.cpp"
 }
 _lua_pushvalue($L,4); //@line 369 "modules/physics/box2d/World.cpp"
 $38 = (__ZN4love10luax_refifEP9lua_Statei($L,6)|0); //@line 370 "modules/physics/box2d/World.cpp"
 HEAP32[$22>>2] = $38; //@line 370 "modules/physics/box2d/World.cpp"
 $39 = ((($this)) + 96|0); //@line 371 "modules/physics/box2d/World.cpp"
 HEAP32[$39>>2] = $L; //@line 371 "modules/physics/box2d/World.cpp"
 return 0; //@line 374 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World12getCallbacksEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 68|0); //@line 379 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 379 "modules/physics/box2d/World.cpp"
 $2 = ($1|0)==(0|0); //@line 379 "modules/physics/box2d/World.cpp"
 if ($2) {
  _lua_pushnil($L); //@line 379 "modules/physics/box2d/World.cpp"
 } else {
  __ZN4love9Reference4pushEP9lua_State($1,$L); //@line 379 "modules/physics/box2d/World.cpp"
 }
 $3 = ((($this)) + 76|0); //@line 380 "modules/physics/box2d/World.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 380 "modules/physics/box2d/World.cpp"
 $5 = ($4|0)==(0|0); //@line 380 "modules/physics/box2d/World.cpp"
 if ($5) {
  _lua_pushnil($L); //@line 380 "modules/physics/box2d/World.cpp"
 } else {
  __ZN4love9Reference4pushEP9lua_State($4,$L); //@line 380 "modules/physics/box2d/World.cpp"
 }
 $6 = ((($this)) + 84|0); //@line 381 "modules/physics/box2d/World.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 381 "modules/physics/box2d/World.cpp"
 $8 = ($7|0)==(0|0); //@line 381 "modules/physics/box2d/World.cpp"
 if ($8) {
  _lua_pushnil($L); //@line 381 "modules/physics/box2d/World.cpp"
 } else {
  __ZN4love9Reference4pushEP9lua_State($7,$L); //@line 381 "modules/physics/box2d/World.cpp"
 }
 $9 = ((($this)) + 92|0); //@line 382 "modules/physics/box2d/World.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 382 "modules/physics/box2d/World.cpp"
 $11 = ($10|0)==(0|0); //@line 382 "modules/physics/box2d/World.cpp"
 if ($11) {
  _lua_pushnil($L); //@line 382 "modules/physics/box2d/World.cpp"
  return 4; //@line 383 "modules/physics/box2d/World.cpp"
 } else {
  __ZN4love9Reference4pushEP9lua_State($10,$L); //@line 382 "modules/physics/box2d/World.cpp"
  return 4; //@line 383 "modules/physics/box2d/World.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d5World13setCallbacksLEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 104|0); //@line 388 "modules/physics/box2d/World.cpp"
 HEAP32[$0>>2] = $L; //@line 388 "modules/physics/box2d/World.cpp"
 $1 = ((($this)) + 96|0); //@line 388 "modules/physics/box2d/World.cpp"
 HEAP32[$1>>2] = $L; //@line 388 "modules/physics/box2d/World.cpp"
 $2 = ((($this)) + 88|0); //@line 388 "modules/physics/box2d/World.cpp"
 HEAP32[$2>>2] = $L; //@line 388 "modules/physics/box2d/World.cpp"
 $3 = ((($this)) + 80|0); //@line 388 "modules/physics/box2d/World.cpp"
 HEAP32[$3>>2] = $L; //@line 388 "modules/physics/box2d/World.cpp"
 $4 = ((($this)) + 72|0); //@line 388 "modules/physics/box2d/World.cpp"
 HEAP32[$4>>2] = $L; //@line 388 "modules/physics/box2d/World.cpp"
 return; //@line 389 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World16setContactFilterEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_type($L,1)|0); //@line 393 "modules/physics/box2d/World.cpp"
 $1 = ($0|0)<(1); //@line 393 "modules/physics/box2d/World.cpp"
 if (!($1)) {
  _luaL_checktype($L,1,6); //@line 394 "modules/physics/box2d/World.cpp"
 }
 $2 = ((($this)) + 100|0); //@line 396 "modules/physics/box2d/World.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 396 "modules/physics/box2d/World.cpp"
 $4 = ($3|0)==(0|0); //@line 396 "modules/physics/box2d/World.cpp"
 if (!($4)) {
  $5 = HEAP32[$3>>2]|0; //@line 397 "modules/physics/box2d/World.cpp"
  $6 = ((($5)) + 4|0); //@line 397 "modules/physics/box2d/World.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 397 "modules/physics/box2d/World.cpp"
  FUNCTION_TABLE_vi[$7 & 8191]($3); //@line 397 "modules/physics/box2d/World.cpp"
 }
 $8 = (__ZN4love10luax_refifEP9lua_Statei($L,6)|0); //@line 398 "modules/physics/box2d/World.cpp"
 HEAP32[$2>>2] = $8; //@line 398 "modules/physics/box2d/World.cpp"
 $9 = ((($this)) + 104|0); //@line 399 "modules/physics/box2d/World.cpp"
 HEAP32[$9>>2] = $L; //@line 399 "modules/physics/box2d/World.cpp"
 return 0; //@line 400 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World16getContactFilterEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 100|0); //@line 405 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 405 "modules/physics/box2d/World.cpp"
 $2 = ($1|0)==(0|0); //@line 405 "modules/physics/box2d/World.cpp"
 if ($2) {
  _lua_pushnil($L); //@line 405 "modules/physics/box2d/World.cpp"
  return 1; //@line 406 "modules/physics/box2d/World.cpp"
 } else {
  __ZN4love9Reference4pushEP9lua_State($1,$L); //@line 405 "modules/physics/box2d/World.cpp"
  return 1; //@line 406 "modules/physics/box2d/World.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d5World10setGravityEff($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = sp + 8|0;
 $2 = ((($this)) + 20|0); //@line 411 "modules/physics/box2d/World.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 411 "modules/physics/box2d/World.cpp"
 HEAPF32[$1>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$4>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($0,$1); //@line 411 "modules/physics/box2d/World.cpp"
 $5 = ((($3)) + 102968|0); //@line 313 "./libraries/Box2D/Dynamics/b2World.h"
 $6 = $0; //@line 313 "./libraries/Box2D/Dynamics/b2World.h"
 $7 = $6; //@line 313 "./libraries/Box2D/Dynamics/b2World.h"
 $8 = HEAP32[$7>>2]|0; //@line 313 "./libraries/Box2D/Dynamics/b2World.h"
 $9 = (($6) + 4)|0; //@line 313 "./libraries/Box2D/Dynamics/b2World.h"
 $10 = $9; //@line 313 "./libraries/Box2D/Dynamics/b2World.h"
 $11 = HEAP32[$10>>2]|0; //@line 313 "./libraries/Box2D/Dynamics/b2World.h"
 $12 = $5; //@line 313 "./libraries/Box2D/Dynamics/b2World.h"
 $13 = $12; //@line 313 "./libraries/Box2D/Dynamics/b2World.h"
 HEAP32[$13>>2] = $8; //@line 313 "./libraries/Box2D/Dynamics/b2World.h"
 $14 = (($12) + 4)|0; //@line 313 "./libraries/Box2D/Dynamics/b2World.h"
 $15 = $14; //@line 313 "./libraries/Box2D/Dynamics/b2World.h"
 HEAP32[$15>>2] = $11; //@line 313 "./libraries/Box2D/Dynamics/b2World.h"
 STACKTOP = sp;return; //@line 412 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World10getGravityEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $v = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp + 8|0;
 $0 = sp;
 $1 = ((($this)) + 20|0); //@line 416 "modules/physics/box2d/World.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 416 "modules/physics/box2d/World.cpp"
 $3 = ((($2)) + 102968|0); //@line 318 "./libraries/Box2D/Dynamics/b2World.h"
 $4 = $3; //@line 318 "./libraries/Box2D/Dynamics/b2World.h"
 $5 = $4; //@line 318 "./libraries/Box2D/Dynamics/b2World.h"
 $6 = HEAP32[$5>>2]|0; //@line 318 "./libraries/Box2D/Dynamics/b2World.h"
 $7 = (($4) + 4)|0; //@line 318 "./libraries/Box2D/Dynamics/b2World.h"
 $8 = $7; //@line 318 "./libraries/Box2D/Dynamics/b2World.h"
 $9 = HEAP32[$8>>2]|0; //@line 318 "./libraries/Box2D/Dynamics/b2World.h"
 $10 = $0; //@line 318 "./libraries/Box2D/Dynamics/b2World.h"
 $11 = $10; //@line 318 "./libraries/Box2D/Dynamics/b2World.h"
 HEAP32[$11>>2] = $6; //@line 318 "./libraries/Box2D/Dynamics/b2World.h"
 $12 = (($10) + 4)|0; //@line 318 "./libraries/Box2D/Dynamics/b2World.h"
 $13 = $12; //@line 318 "./libraries/Box2D/Dynamics/b2World.h"
 HEAP32[$13>>2] = $9; //@line 318 "./libraries/Box2D/Dynamics/b2World.h"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($v,$0); //@line 416 "modules/physics/box2d/World.cpp"
 $14 = +HEAPF32[$v>>2]; //@line 417 "modules/physics/box2d/World.cpp"
 $15 = $14; //@line 417 "modules/physics/box2d/World.cpp"
 _lua_pushnumber($L,$15); //@line 417 "modules/physics/box2d/World.cpp"
 $16 = ((($v)) + 4|0); //@line 418 "modules/physics/box2d/World.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 418 "modules/physics/box2d/World.cpp"
 $18 = $17; //@line 418 "modules/physics/box2d/World.cpp"
 _lua_pushnumber($L,$18); //@line 418 "modules/physics/box2d/World.cpp"
 STACKTOP = sp;return 2; //@line 419 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World15translateOriginEff($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 8|0;
 $1 = sp;
 $2 = ((($this)) + 20|0); //@line 424 "modules/physics/box2d/World.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 424 "modules/physics/box2d/World.cpp"
 HEAPF32[$1>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$4>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($0,$1); //@line 424 "modules/physics/box2d/World.cpp"
 __ZN7b2World11ShiftOriginERK6b2Vec2($3,$0); //@line 424 "modules/physics/box2d/World.cpp"
 STACKTOP = sp;return; //@line 425 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World18setSleepingAllowedEb($this,$allow) {
 $this = $this|0;
 $allow = $allow|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 429 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 429 "modules/physics/box2d/World.cpp"
 __ZN7b2World16SetAllowSleepingEb($1,$allow); //@line 429 "modules/physics/box2d/World.cpp"
 return; //@line 430 "modules/physics/box2d/World.cpp"
}
function __ZNK4love7physics5box2d5World17isSleepingAllowedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 434 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 434 "modules/physics/box2d/World.cpp"
 $2 = ((($1)) + 102976|0); //@line 146 "./libraries/Box2D/Dynamics/b2World.h"
 $3 = HEAP8[$2>>0]|0; //@line 146 "./libraries/Box2D/Dynamics/b2World.h"
 $4 = ($3<<24>>24)!=(0); //@line 146 "./libraries/Box2D/Dynamics/b2World.h"
 return ($4|0); //@line 434 "modules/physics/box2d/World.cpp"
}
function __ZNK4love7physics5box2d5World8isLockedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 439 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 439 "modules/physics/box2d/World.cpp"
 $2 = ((($1)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $3 = HEAP32[$2>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $4 = $3 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $5 = ($4|0)!=(0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 return ($5|0); //@line 439 "modules/physics/box2d/World.cpp"
}
function __ZNK4love7physics5box2d5World12getBodyCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 444 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 444 "modules/physics/box2d/World.cpp"
 $2 = ((($1)) + 102960|0); //@line 298 "./libraries/Box2D/Dynamics/b2World.h"
 $3 = HEAP32[$2>>2]|0; //@line 298 "./libraries/Box2D/Dynamics/b2World.h"
 $4 = (($3) + -1)|0; //@line 444 "modules/physics/box2d/World.cpp"
 return ($4|0); //@line 444 "modules/physics/box2d/World.cpp"
}
function __ZNK4love7physics5box2d5World13getJointCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 449 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 449 "modules/physics/box2d/World.cpp"
 $2 = ((($1)) + 102964|0); //@line 303 "./libraries/Box2D/Dynamics/b2World.h"
 $3 = HEAP32[$2>>2]|0; //@line 303 "./libraries/Box2D/Dynamics/b2World.h"
 return ($3|0); //@line 449 "modules/physics/box2d/World.cpp"
}
function __ZNK4love7physics5box2d5World15getContactCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 454 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 454 "modules/physics/box2d/World.cpp"
 $2 = ((($1)) + 102936|0); //@line 308 "./libraries/Box2D/Dynamics/b2World.h"
 $3 = HEAP32[$2>>2]|0; //@line 308 "./libraries/Box2D/Dynamics/b2World.h"
 return ($3|0); //@line 454 "modules/physics/box2d/World.cpp"
}
function __ZNK4love7physics5box2d5World11getBodyListEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $$old1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b$0 = 0;
 var $i$0 = 0, $i$1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _lua_createtable($L,0,0); //@line 459 "modules/physics/box2d/World.cpp"
 $0 = ((($this)) + 20|0); //@line 460 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 460 "modules/physics/box2d/World.cpp"
 $2 = ((($1)) + 102952|0); //@line 268 "./libraries/Box2D/Dynamics/b2World.h"
 $3 = HEAP32[$2>>2]|0; //@line 268 "./libraries/Box2D/Dynamics/b2World.h"
 $$old1 = ($3|0)==(0|0); //@line 464 "modules/physics/box2d/World.cpp"
 if ($$old1) {
  STACKTOP = sp;return 1; //@line 476 "modules/physics/box2d/World.cpp"
 }
 $4 = ((($this)) + 24|0); //@line 466 "modules/physics/box2d/World.cpp"
 $b$0 = $3;$i$0 = 1;
 while(1) {
  $5 = HEAP32[$4>>2]|0; //@line 466 "modules/physics/box2d/World.cpp"
  $6 = ($b$0|0)==($5|0); //@line 466 "modules/physics/box2d/World.cpp"
  if ($6) {
   $i$1 = $i$0;
  } else {
   $7 = (__ZN4love8Memoizer4findEPv($b$0)|0); //@line 468 "modules/physics/box2d/World.cpp"
   $8 = ($7|0)==(0|0); //@line 469 "modules/physics/box2d/World.cpp"
   if ($8) {
    break;
   }
   __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,34,$7); //@line 471 "modules/physics/box2d/World.cpp"
   _lua_rawseti($L,-2,$i$0); //@line 472 "modules/physics/box2d/World.cpp"
   $14 = (($i$0) + 1)|0; //@line 473 "modules/physics/box2d/World.cpp"
   $i$1 = $14;
  }
  $15 = ((($b$0)) + 96|0); //@line 719 "./libraries/Box2D/Dynamics/b2Body.h"
  $16 = HEAP32[$15>>2]|0; //@line 719 "./libraries/Box2D/Dynamics/b2Body.h"
  $17 = ($16|0)==(0|0); //@line 475 "modules/physics/box2d/World.cpp"
  if ($17) {
   label = 10;
   break;
  } else {
   $b$0 = $16;$i$0 = $i$1;
  }
 }
 if ((label|0) == 10) {
  STACKTOP = sp;return 1; //@line 476 "modules/physics/box2d/World.cpp"
 }
 $9 = (___cxa_allocate_exception(16)|0); //@line 470 "modules/physics/box2d/World.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($9|0),(356926|0),($vararg_buffer|0)); //@line 470 "modules/physics/box2d/World.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $12 = ___cxa_find_matching_catch_2()|0;
  $13 = tempRet0;
  ___cxa_free_exception(($9|0)); //@line 470 "modules/physics/box2d/World.cpp"
  ___resumeException($12|0);
  // unreachable;
 } else {
  ___cxa_throw(($9|0),(96|0),(344|0)); //@line 470 "modules/physics/box2d/World.cpp"
  // unreachable; //@line 470 "modules/physics/box2d/World.cpp"
 }
 return (0)|0;
}
function __ZNK4love7physics5box2d5World12getJointListEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $$old1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0, $j$0 = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _lua_createtable($L,0,0); //@line 481 "modules/physics/box2d/World.cpp"
 $0 = ((($this)) + 20|0); //@line 482 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 482 "modules/physics/box2d/World.cpp"
 $2 = ((($1)) + 102956|0); //@line 278 "./libraries/Box2D/Dynamics/b2World.h"
 $3 = HEAP32[$2>>2]|0; //@line 278 "./libraries/Box2D/Dynamics/b2World.h"
 $$old1 = ($3|0)==(0|0); //@line 486 "modules/physics/box2d/World.cpp"
 if ($$old1) {
  STACKTOP = sp;return 1; //@line 494 "modules/physics/box2d/World.cpp"
 } else {
  $i$0 = 1;$j$0 = $3;
 }
 while(1) {
  $4 = (__ZN4love8Memoizer4findEPv($j$0)|0); //@line 487 "modules/physics/box2d/World.cpp"
  $5 = ($4|0)==(0|0); //@line 488 "modules/physics/box2d/World.cpp"
  if ($5) {
   break;
  }
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,41,$4); //@line 489 "modules/physics/box2d/World.cpp"
  _lua_rawseti($L,-2,$i$0); //@line 490 "modules/physics/box2d/World.cpp"
  $11 = (($i$0) + 1)|0; //@line 491 "modules/physics/box2d/World.cpp"
  $12 = ((($j$0)) + 12|0); //@line 203 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
  $13 = HEAP32[$12>>2]|0; //@line 203 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
  $14 = ($13|0)==(0|0); //@line 493 "modules/physics/box2d/World.cpp"
  if ($14) {
   label = 7;
   break;
  } else {
   $i$0 = $11;$j$0 = $13;
  }
 }
 if ((label|0) == 7) {
  STACKTOP = sp;return 1; //@line 494 "modules/physics/box2d/World.cpp"
 }
 $6 = (___cxa_allocate_exception(16)|0); //@line 488 "modules/physics/box2d/World.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($6|0),(444962|0),($vararg_buffer|0)); //@line 488 "modules/physics/box2d/World.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  ___cxa_free_exception(($6|0)); //@line 488 "modules/physics/box2d/World.cpp"
  ___resumeException($9|0);
  // unreachable;
 } else {
  ___cxa_throw(($6|0),(96|0),(344|0)); //@line 488 "modules/physics/box2d/World.cpp"
  // unreachable; //@line 488 "modules/physics/box2d/World.cpp"
 }
 return (0)|0;
}
function __ZNK4love7physics5box2d5World14getContactListEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $$lcssa = 0, $$old1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $c$0 = 0, $contact$0 = 0, $i$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _lua_createtable($L,0,0); //@line 499 "modules/physics/box2d/World.cpp"
 $0 = ((($this)) + 20|0); //@line 500 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 500 "modules/physics/box2d/World.cpp"
 $2 = ((($1)) + 102932|0); //@line 288 "./libraries/Box2D/Dynamics/b2World.h"
 $3 = HEAP32[$2>>2]|0; //@line 288 "./libraries/Box2D/Dynamics/b2World.h"
 $$old1 = ($3|0)==(0|0); //@line 504 "modules/physics/box2d/World.cpp"
 if ($$old1) {
  return 1; //@line 516 "modules/physics/box2d/World.cpp"
 } else {
  $c$0 = $3;$i$0 = 1;
 }
 while(1) {
  $4 = (__ZN4love8Memoizer4findEPv($c$0)|0); //@line 505 "modules/physics/box2d/World.cpp"
  $5 = ($4|0)==(0|0); //@line 506 "modules/physics/box2d/World.cpp"
  if ($5) {
   $6 = (__Znwj(12)|0); //@line 507 "modules/physics/box2d/World.cpp"
   __THREW__ = 0;
   invoke_vii(3706,($6|0),($c$0|0)); //@line 507 "modules/physics/box2d/World.cpp"
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if ($8) {
    $$lcssa = $6;
    label = 4;
    break;
   } else {
    $contact$0 = $6;
   }
  } else {
   $11 = HEAP32[$4>>2]|0; //@line 509 "modules/physics/box2d/World.cpp"
   $12 = ((($11)) + 8|0); //@line 509 "modules/physics/box2d/World.cpp"
   $13 = HEAP32[$12>>2]|0; //@line 509 "modules/physics/box2d/World.cpp"
   FUNCTION_TABLE_vi[$13 & 8191]($4); //@line 509 "modules/physics/box2d/World.cpp"
   $contact$0 = $4;
  }
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,33,$contact$0); //@line 510 "modules/physics/box2d/World.cpp"
  $14 = HEAP32[$contact$0>>2]|0; //@line 511 "modules/physics/box2d/World.cpp"
  $15 = ((($14)) + 12|0); //@line 511 "modules/physics/box2d/World.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 511 "modules/physics/box2d/World.cpp"
  FUNCTION_TABLE_vi[$16 & 8191]($contact$0); //@line 511 "modules/physics/box2d/World.cpp"
  _lua_rawseti($L,-2,$i$0); //@line 512 "modules/physics/box2d/World.cpp"
  $17 = (($i$0) + 1)|0; //@line 513 "modules/physics/box2d/World.cpp"
  $18 = ((($c$0)) + 12|0); //@line 266 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
  $19 = HEAP32[$18>>2]|0; //@line 266 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
  $20 = ($19|0)==(0|0); //@line 515 "modules/physics/box2d/World.cpp"
  if ($20) {
   label = 7;
   break;
  } else {
   $c$0 = $19;$i$0 = $17;
  }
 }
 if ((label|0) == 4) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  __ZdlPv($$lcssa); //@line 507 "modules/physics/box2d/World.cpp"
  ___resumeException($9|0);
  // unreachable;
 }
 else if ((label|0) == 7) {
  return 1; //@line 516 "modules/physics/box2d/World.cpp"
 }
 return (0)|0;
}
function __ZNK4love7physics5box2d5World13getGroundBodyEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 521 "modules/physics/box2d/World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 521 "modules/physics/box2d/World.cpp"
 return ($1|0); //@line 521 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World16queryBoundingBoxEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $box = 0, $query = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $box = sp + 8|0;
 $0 = sp + 48|0;
 $1 = sp;
 $2 = sp + 40|0;
 $query = sp + 24|0;
 $3 = (+_luaL_checknumber($L,1)); //@line 527 "modules/physics/box2d/World.cpp"
 $4 = $3; //@line 527 "modules/physics/box2d/World.cpp"
 $5 = (+_luaL_checknumber($L,2)); //@line 528 "modules/physics/box2d/World.cpp"
 $6 = $5; //@line 528 "modules/physics/box2d/World.cpp"
 $7 = (+_luaL_checknumber($L,3)); //@line 529 "modules/physics/box2d/World.cpp"
 $8 = $7; //@line 529 "modules/physics/box2d/World.cpp"
 $9 = (+_luaL_checknumber($L,4)); //@line 530 "modules/physics/box2d/World.cpp"
 $10 = $9; //@line 530 "modules/physics/box2d/World.cpp"
 HEAPF32[$0>>2] = $4; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($0)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$11>>2] = $6; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($box,$0); //@line 531 "modules/physics/box2d/World.cpp"
 $12 = ((($box)) + 8|0); //@line 532 "modules/physics/box2d/World.cpp"
 HEAPF32[$2>>2] = $8; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($2)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$13>>2] = $10; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($1,$2); //@line 532 "modules/physics/box2d/World.cpp"
 $14 = $1; //@line 532 "modules/physics/box2d/World.cpp"
 $15 = $14; //@line 532 "modules/physics/box2d/World.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 532 "modules/physics/box2d/World.cpp"
 $17 = (($14) + 4)|0; //@line 532 "modules/physics/box2d/World.cpp"
 $18 = $17; //@line 532 "modules/physics/box2d/World.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 532 "modules/physics/box2d/World.cpp"
 $20 = $12; //@line 532 "modules/physics/box2d/World.cpp"
 $21 = $20; //@line 532 "modules/physics/box2d/World.cpp"
 HEAP32[$21>>2] = $16; //@line 532 "modules/physics/box2d/World.cpp"
 $22 = (($20) + 4)|0; //@line 532 "modules/physics/box2d/World.cpp"
 $23 = $22; //@line 532 "modules/physics/box2d/World.cpp"
 HEAP32[$23>>2] = $19; //@line 532 "modules/physics/box2d/World.cpp"
 _luaL_checktype($L,5,6); //@line 533 "modules/physics/box2d/World.cpp"
 HEAP32[$query>>2] = (17328); //@line 142 "modules/physics/box2d/World.cpp"
 $24 = ((($query)) + 4|0); //@line 140 "modules/physics/box2d/World.cpp"
 HEAP32[$24>>2] = $L; //@line 140 "modules/physics/box2d/World.cpp"
 $25 = ((($query)) + 8|0); //@line 141 "modules/physics/box2d/World.cpp"
 HEAP32[$25>>2] = 5; //@line 141 "modules/physics/box2d/World.cpp"
 _luaL_checktype($L,5,6); //@line 143 "modules/physics/box2d/World.cpp"
 $26 = ((($this)) + 20|0); //@line 535 "modules/physics/box2d/World.cpp"
 $27 = HEAP32[$26>>2]|0; //@line 535 "modules/physics/box2d/World.cpp"
 __ZNK7b2World9QueryAABBEP15b2QueryCallbackRK6b2AABB($27,$query,$box); //@line 535 "modules/physics/box2d/World.cpp"
 STACKTOP = sp;return 0; //@line 537 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d5World7rayCastEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $raycast = 0, $v1 = 0, $v2 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v1 = sp + 40|0;
 $0 = sp + 32|0;
 $v2 = sp + 24|0;
 $1 = sp + 16|0;
 $raycast = sp;
 $2 = (+_luaL_checknumber($L,1)); //@line 541 "modules/physics/box2d/World.cpp"
 $3 = $2; //@line 541 "modules/physics/box2d/World.cpp"
 $4 = (+_luaL_checknumber($L,2)); //@line 542 "modules/physics/box2d/World.cpp"
 $5 = $4; //@line 542 "modules/physics/box2d/World.cpp"
 $6 = (+_luaL_checknumber($L,3)); //@line 543 "modules/physics/box2d/World.cpp"
 $7 = $6; //@line 543 "modules/physics/box2d/World.cpp"
 $8 = (+_luaL_checknumber($L,4)); //@line 544 "modules/physics/box2d/World.cpp"
 $9 = $8; //@line 544 "modules/physics/box2d/World.cpp"
 HEAPF32[$0>>2] = $3; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($0)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$10>>2] = $5; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($v1,$0); //@line 545 "modules/physics/box2d/World.cpp"
 HEAPF32[$1>>2] = $7; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$11>>2] = $9; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($v2,$1); //@line 546 "modules/physics/box2d/World.cpp"
 _luaL_checktype($L,5,6); //@line 547 "modules/physics/box2d/World.cpp"
 HEAP32[$raycast>>2] = (17348); //@line 171 "modules/physics/box2d/World.cpp"
 $12 = ((($raycast)) + 4|0); //@line 169 "modules/physics/box2d/World.cpp"
 HEAP32[$12>>2] = $L; //@line 169 "modules/physics/box2d/World.cpp"
 $13 = ((($raycast)) + 8|0); //@line 170 "modules/physics/box2d/World.cpp"
 HEAP32[$13>>2] = 5; //@line 170 "modules/physics/box2d/World.cpp"
 _luaL_checktype($L,5,6); //@line 172 "modules/physics/box2d/World.cpp"
 $14 = ((($this)) + 20|0); //@line 549 "modules/physics/box2d/World.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 549 "modules/physics/box2d/World.cpp"
 __ZNK7b2World7RayCastEP17b2RayCastCallbackRK6b2Vec2S4_($15,$raycast,$v1,$v2); //@line 549 "modules/physics/box2d/World.cpp"
 STACKTOP = sp;return 0; //@line 551 "modules/physics/box2d/World.cpp"
}
function __ZN4love7physics5box2d14luax_checkbodyEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 31 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,$idx,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if (!($3)) {
  STACKTOP = sp;return ($0|0); //@line 36 "modules/physics/box2d/wrap_Body.cpp"
 }
 (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return ($0|0); //@line 36 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d11w_Body_getXEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 39 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+__ZN4love7physics5box2d4Body4getXEv($0)); //@line 42 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 42 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$5); //@line 42 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 1; //@line 43 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d11w_Body_getYEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 46 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+__ZN4love7physics5box2d4Body4getYEv($0)); //@line 49 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 49 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$5); //@line 49 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 1; //@line 50 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d15w_Body_getAngleEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 53 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+__ZN4love7physics5box2d4Body8getAngleEv($0)); //@line 56 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 56 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$5); //@line 56 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 1; //@line 57 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d18w_Body_getPositionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $vararg_buffer = 0, $x_o = 0, $y_o = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $x_o = sp + 8|0;
 $y_o = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 __ZN4love7physics5box2d4Body11getPositionERfS3_($0,$x_o,$y_o); //@line 65 "modules/physics/box2d/wrap_Body.cpp"
 $4 = +HEAPF32[$x_o>>2]; //@line 66 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 66 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$5); //@line 66 "modules/physics/box2d/wrap_Body.cpp"
 $6 = +HEAPF32[$y_o>>2]; //@line 67 "modules/physics/box2d/wrap_Body.cpp"
 $7 = $6; //@line 67 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$7); //@line 67 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 2; //@line 69 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d24w_Body_getLinearVelocityEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $vararg_buffer = 0, $x_o = 0, $y_o = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $x_o = sp + 8|0;
 $y_o = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 __ZN4love7physics5box2d4Body17getLinearVelocityERfS3_($0,$x_o,$y_o); //@line 77 "modules/physics/box2d/wrap_Body.cpp"
 $4 = +HEAPF32[$x_o>>2]; //@line 78 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 78 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$5); //@line 78 "modules/physics/box2d/wrap_Body.cpp"
 $6 = +HEAPF32[$y_o>>2]; //@line 79 "modules/physics/box2d/wrap_Body.cpp"
 $7 = $6; //@line 79 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$7); //@line 79 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 2; //@line 81 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d21w_Body_getWorldCenterEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $vararg_buffer = 0, $x_o = 0, $y_o = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $x_o = sp + 8|0;
 $y_o = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 __ZN4love7physics5box2d4Body14getWorldCenterERfS3_($0,$x_o,$y_o); //@line 89 "modules/physics/box2d/wrap_Body.cpp"
 $4 = +HEAPF32[$x_o>>2]; //@line 90 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 90 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$5); //@line 90 "modules/physics/box2d/wrap_Body.cpp"
 $6 = +HEAPF32[$y_o>>2]; //@line 91 "modules/physics/box2d/wrap_Body.cpp"
 $7 = $6; //@line 91 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$7); //@line 91 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 2; //@line 93 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d21w_Body_getLocalCenterEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $vararg_buffer = 0, $x_o = 0, $y_o = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $x_o = sp + 8|0;
 $y_o = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 __ZN4love7physics5box2d4Body14getLocalCenterERfS3_($0,$x_o,$y_o); //@line 101 "modules/physics/box2d/wrap_Body.cpp"
 $4 = +HEAPF32[$x_o>>2]; //@line 102 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 102 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$5); //@line 102 "modules/physics/box2d/wrap_Body.cpp"
 $6 = +HEAPF32[$y_o>>2]; //@line 103 "modules/physics/box2d/wrap_Body.cpp"
 $7 = $6; //@line 103 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$7); //@line 103 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 2; //@line 105 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d25w_Body_getAngularVelocityEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 108 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+__ZNK4love7physics5box2d4Body18getAngularVelocityEv($0)); //@line 111 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 111 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$5); //@line 111 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 1; //@line 112 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d14w_Body_getMassEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 115 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+__ZNK4love7physics5box2d4Body7getMassEv($0)); //@line 118 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 118 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$5); //@line 118 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 1; //@line 119 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d17w_Body_getInertiaEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 122 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+__ZNK4love7physics5box2d4Body10getInertiaEv($0)); //@line 125 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 125 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$5); //@line 125 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 1; //@line 126 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d18w_Body_getMassDataEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 129 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 _lua_remove($L,1); //@line 132 "modules/physics/box2d/wrap_Body.cpp"
 $4 = (__ZN4love7physics5box2d4Body11getMassDataEP9lua_State($0,$L)|0); //@line 133 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return ($4|0); //@line 133 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d24w_Body_getAngularDampingEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 136 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+__ZNK4love7physics5box2d4Body17getAngularDampingEv($0)); //@line 139 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 139 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$5); //@line 139 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 1; //@line 140 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d23w_Body_getLinearDampingEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 143 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+__ZNK4love7physics5box2d4Body16getLinearDampingEv($0)); //@line 146 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 146 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$5); //@line 146 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 1; //@line 147 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d22w_Body_getGravityScaleEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 150 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+__ZNK4love7physics5box2d4Body15getGravityScaleEv($0)); //@line 153 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 153 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$5); //@line 153 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 1; //@line 154 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d14w_Body_getTypeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $type = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $type = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 HEAP32[$type>>2] = 1282466; //@line 160 "modules/physics/box2d/wrap_Body.cpp"
 $4 = (__ZNK4love7physics5box2d4Body7getTypeEv($0)|0); //@line 161 "modules/physics/box2d/wrap_Body.cpp"
 (__ZN4love7physics4Body11getConstantENS1_4TypeERPKc($4,$type)|0); //@line 161 "modules/physics/box2d/wrap_Body.cpp"
 $5 = HEAP32[$type>>2]|0; //@line 162 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushstring($L,$5); //@line 162 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 1; //@line 163 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d25w_Body_applyLinearImpulseEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0;
 var $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 166 "modules/physics/box2d/wrap_Body.cpp"
 $vararg_buffer = sp; //@line 166 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 169 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 169 "modules/physics/box2d/wrap_Body.cpp"
 $6 = (+_luaL_checknumber($L,3)); //@line 170 "modules/physics/box2d/wrap_Body.cpp"
 $7 = $6; //@line 170 "modules/physics/box2d/wrap_Body.cpp"
 $8 = (_lua_gettop($L)|0); //@line 172 "modules/physics/box2d/wrap_Body.cpp"
 $9 = ($8|0)<(4); //@line 174 "modules/physics/box2d/wrap_Body.cpp"
 if (!($9)) {
  $10 = ($8|0)==(4); //@line 174 "modules/physics/box2d/wrap_Body.cpp"
  if (!($10)) {
   $14 = (+_luaL_checknumber($L,4)); //@line 181 "modules/physics/box2d/wrap_Body.cpp"
   $15 = $14; //@line 181 "modules/physics/box2d/wrap_Body.cpp"
   $16 = (+_luaL_checknumber($L,5)); //@line 182 "modules/physics/box2d/wrap_Body.cpp"
   $17 = $16; //@line 182 "modules/physics/box2d/wrap_Body.cpp"
   $18 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,6,1)|0); //@line 183 "modules/physics/box2d/wrap_Body.cpp"
   __ZN4love7physics5box2d4Body18applyLinearImpulseEffffb($0,$5,$7,$15,$17,$18); //@line 184 "modules/physics/box2d/wrap_Body.cpp"
   $$0 = 0;
   STACKTOP = sp;return ($$0|0); //@line 192 "modules/physics/box2d/wrap_Body.cpp"
  }
  $11 = (_lua_type($L,4)|0); //@line 174 "modules/physics/box2d/wrap_Body.cpp"
  $12 = ($11|0)==(1); //@line 174 "modules/physics/box2d/wrap_Body.cpp"
  if (!($12)) {
   $19 = (_luaL_error($L,356986,$vararg_buffer1)|0); //@line 188 "modules/physics/box2d/wrap_Body.cpp"
   $$0 = $19;
   STACKTOP = sp;return ($$0|0); //@line 192 "modules/physics/box2d/wrap_Body.cpp"
  }
 }
 $13 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,4,1)|0); //@line 176 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body18applyLinearImpulseEffb($0,$5,$7,$13); //@line 177 "modules/physics/box2d/wrap_Body.cpp"
 $$0 = 0;
 STACKTOP = sp;return ($$0|0); //@line 192 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d26w_Body_applyAngularImpulseEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 194 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 197 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 197 "modules/physics/box2d/wrap_Body.cpp"
 $6 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,3,1)|0); //@line 198 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body19applyAngularImpulseEfb($0,$5,$6); //@line 199 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 0; //@line 200 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d18w_Body_applyTorqueEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 203 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 206 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 206 "modules/physics/box2d/wrap_Body.cpp"
 $6 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,3,1)|0); //@line 207 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body11applyTorqueEfb($0,$5,$6); //@line 208 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 0; //@line 209 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d17w_Body_applyForceEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0;
 var $7 = 0.0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 212 "modules/physics/box2d/wrap_Body.cpp"
 $vararg_buffer = sp; //@line 212 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 215 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 215 "modules/physics/box2d/wrap_Body.cpp"
 $6 = (+_luaL_checknumber($L,3)); //@line 216 "modules/physics/box2d/wrap_Body.cpp"
 $7 = $6; //@line 216 "modules/physics/box2d/wrap_Body.cpp"
 $8 = (_lua_gettop($L)|0); //@line 218 "modules/physics/box2d/wrap_Body.cpp"
 $9 = ($8|0)<(4); //@line 220 "modules/physics/box2d/wrap_Body.cpp"
 do {
  if (!($9)) {
   $10 = ($8|0)==(4); //@line 220 "modules/physics/box2d/wrap_Body.cpp"
   if ($10) {
    $11 = (_lua_type($L,4)|0); //@line 220 "modules/physics/box2d/wrap_Body.cpp"
    $12 = ($11|0)==(1); //@line 220 "modules/physics/box2d/wrap_Body.cpp"
    if ($12) {
     break;
    }
   }
   $14 = (_lua_gettop($L)|0); //@line 225 "modules/physics/box2d/wrap_Body.cpp"
   $15 = ($14|0)>(4); //@line 225 "modules/physics/box2d/wrap_Body.cpp"
   if ($15) {
    $16 = (+_luaL_checknumber($L,4)); //@line 227 "modules/physics/box2d/wrap_Body.cpp"
    $17 = $16; //@line 227 "modules/physics/box2d/wrap_Body.cpp"
    $18 = (+_luaL_checknumber($L,5)); //@line 228 "modules/physics/box2d/wrap_Body.cpp"
    $19 = $18; //@line 228 "modules/physics/box2d/wrap_Body.cpp"
    $20 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,6,1)|0); //@line 229 "modules/physics/box2d/wrap_Body.cpp"
    __ZN4love7physics5box2d4Body10applyForceEffffb($0,$5,$7,$17,$19,$20); //@line 230 "modules/physics/box2d/wrap_Body.cpp"
    $$0 = 0;
    STACKTOP = sp;return ($$0|0); //@line 238 "modules/physics/box2d/wrap_Body.cpp"
   } else {
    $21 = (_luaL_error($L,356986,$vararg_buffer1)|0); //@line 234 "modules/physics/box2d/wrap_Body.cpp"
    $$0 = $21;
    STACKTOP = sp;return ($$0|0); //@line 238 "modules/physics/box2d/wrap_Body.cpp"
   }
  }
 } while(0);
 $13 = (__ZN4love15luax_optbooleanEP9lua_Stateib($L,4,1)|0); //@line 222 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body10applyForceEffb($0,$5,$7,$13); //@line 223 "modules/physics/box2d/wrap_Body.cpp"
 $$0 = 0;
 STACKTOP = sp;return ($$0|0); //@line 238 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d11w_Body_setXEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 240 "modules/physics/box2d/wrap_Body.cpp"
 $vararg_buffer = sp; //@line 240 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 243 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 243 "modules/physics/box2d/wrap_Body.cpp"
 __THREW__ = 0;
 invoke_vid(3713,($0|0),(+$5)); //@line 244 "modules/physics/box2d/wrap_Body.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  STACKTOP = sp;return 0; //@line 245 "modules/physics/box2d/wrap_Body.cpp"
 }
 $8 = ___cxa_find_matching_catch_3(8992|0)|0;
 $9 = tempRet0;
 $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $11 = ($9|0)==($10|0); //@line 535 "./common/runtime.h"
 if (!($11)) {
  $$01$i = $9;$$02$i = $8;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $12 = (___cxa_begin_catch(($8|0))|0); //@line 535 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = ((($13)) + 8|0); //@line 539 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($16|0)); //@line 539 "./common/runtime.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if (!($18)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $19 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $19; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 245 "modules/physics/box2d/wrap_Body.cpp"
 }
 $20 = ___cxa_find_matching_catch_2()|0;
 $21 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if ($23) {
  $24 = ___cxa_find_matching_catch_3(0|0)|0;
  $25 = tempRet0;
  ___clang_call_terminate($24); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $21;$$02$i = $20;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d11w_Body_setYEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 248 "modules/physics/box2d/wrap_Body.cpp"
 $vararg_buffer = sp; //@line 248 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 251 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 251 "modules/physics/box2d/wrap_Body.cpp"
 __THREW__ = 0;
 invoke_vid(3714,($0|0),(+$5)); //@line 252 "modules/physics/box2d/wrap_Body.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  STACKTOP = sp;return 0; //@line 253 "modules/physics/box2d/wrap_Body.cpp"
 }
 $8 = ___cxa_find_matching_catch_3(8992|0)|0;
 $9 = tempRet0;
 $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $11 = ($9|0)==($10|0); //@line 535 "./common/runtime.h"
 if (!($11)) {
  $$01$i = $9;$$02$i = $8;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $12 = (___cxa_begin_catch(($8|0))|0); //@line 535 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = ((($13)) + 8|0); //@line 539 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($16|0)); //@line 539 "./common/runtime.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if (!($18)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $19 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $19; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 253 "modules/physics/box2d/wrap_Body.cpp"
 }
 $20 = ___cxa_find_matching_catch_2()|0;
 $21 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if ($23) {
  $24 = ___cxa_find_matching_catch_3(0|0)|0;
  $25 = tempRet0;
  ___clang_call_terminate($24); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $21;$$02$i = $20;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d24w_Body_setLinearVelocityEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 256 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 259 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 259 "modules/physics/box2d/wrap_Body.cpp"
 $6 = (+_luaL_checknumber($L,3)); //@line 260 "modules/physics/box2d/wrap_Body.cpp"
 $7 = $6; //@line 260 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body17setLinearVelocityEff($0,$5,$7); //@line 261 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 0; //@line 262 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d15w_Body_setAngleEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 265 "modules/physics/box2d/wrap_Body.cpp"
 $vararg_buffer = sp; //@line 265 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 268 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 268 "modules/physics/box2d/wrap_Body.cpp"
 __THREW__ = 0;
 invoke_vid(3715,($0|0),(+$5)); //@line 269 "modules/physics/box2d/wrap_Body.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  STACKTOP = sp;return 0; //@line 270 "modules/physics/box2d/wrap_Body.cpp"
 }
 $8 = ___cxa_find_matching_catch_3(8992|0)|0;
 $9 = tempRet0;
 $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $11 = ($9|0)==($10|0); //@line 535 "./common/runtime.h"
 if (!($11)) {
  $$01$i = $9;$$02$i = $8;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $12 = (___cxa_begin_catch(($8|0))|0); //@line 535 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = ((($13)) + 8|0); //@line 539 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($16|0)); //@line 539 "./common/runtime.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if (!($18)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $19 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $19; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 270 "modules/physics/box2d/wrap_Body.cpp"
 }
 $20 = ___cxa_find_matching_catch_2()|0;
 $21 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if ($23) {
  $24 = ___cxa_find_matching_catch_3(0|0)|0;
  $25 = tempRet0;
  ___clang_call_terminate($24); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $21;$$02$i = $20;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d25w_Body_setAngularVelocityEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 273 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 276 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 276 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body18setAngularVelocityEf($0,$5); //@line 277 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 0; //@line 278 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d18w_Body_setPositionEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 281 "modules/physics/box2d/wrap_Body.cpp"
 $vararg_buffer = sp; //@line 281 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 284 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 284 "modules/physics/box2d/wrap_Body.cpp"
 $6 = (+_luaL_checknumber($L,3)); //@line 285 "modules/physics/box2d/wrap_Body.cpp"
 $7 = $6; //@line 285 "modules/physics/box2d/wrap_Body.cpp"
 __THREW__ = 0;
 invoke_vidd(3716,($0|0),(+$5),(+$7)); //@line 286 "modules/physics/box2d/wrap_Body.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if (!($9)) {
  STACKTOP = sp;return 0; //@line 287 "modules/physics/box2d/wrap_Body.cpp"
 }
 $10 = ___cxa_find_matching_catch_3(8992|0)|0;
 $11 = tempRet0;
 $12 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $13 = ($11|0)==($12|0); //@line 535 "./common/runtime.h"
 if (!($13)) {
  $$01$i = $11;$$02$i = $10;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $14 = (___cxa_begin_catch(($10|0))|0); //@line 535 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = ((($15)) + 8|0); //@line 539 "./common/runtime.h"
 $17 = HEAP32[$16>>2]|0; //@line 539 "./common/runtime.h"
 $18 = (FUNCTION_TABLE_ii[$17 & 8191]($14)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($18|0)); //@line 539 "./common/runtime.h"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if (!($20)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $21 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $21; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 287 "modules/physics/box2d/wrap_Body.cpp"
 }
 $22 = ___cxa_find_matching_catch_2()|0;
 $23 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $26 = ___cxa_find_matching_catch_3(0|0)|0;
  $27 = tempRet0;
  ___clang_call_terminate($26); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $23;$$02$i = $22;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d20w_Body_resetMassDataEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 290 "modules/physics/box2d/wrap_Body.cpp"
 $vararg_buffer = sp; //@line 290 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 __THREW__ = 0;
 invoke_vi(3717,($0|0)); //@line 293 "modules/physics/box2d/wrap_Body.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  STACKTOP = sp;return 0; //@line 294 "modules/physics/box2d/wrap_Body.cpp"
 }
 $6 = ___cxa_find_matching_catch_3(8992|0)|0;
 $7 = tempRet0;
 $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
 if (!($9)) {
  $$01$i = $7;$$02$i = $6;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $17; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 294 "modules/physics/box2d/wrap_Body.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d18w_Body_setMassDataEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 297 "modules/physics/box2d/wrap_Body.cpp"
 $vararg_buffer = sp; //@line 297 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 300 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 300 "modules/physics/box2d/wrap_Body.cpp"
 $6 = (+_luaL_checknumber($L,3)); //@line 301 "modules/physics/box2d/wrap_Body.cpp"
 $7 = $6; //@line 301 "modules/physics/box2d/wrap_Body.cpp"
 $8 = (+_luaL_checknumber($L,4)); //@line 302 "modules/physics/box2d/wrap_Body.cpp"
 $9 = $8; //@line 302 "modules/physics/box2d/wrap_Body.cpp"
 $10 = (+_luaL_checknumber($L,5)); //@line 303 "modules/physics/box2d/wrap_Body.cpp"
 $11 = $10; //@line 303 "modules/physics/box2d/wrap_Body.cpp"
 __THREW__ = 0;
 invoke_vidddd(3718,($0|0),(+$5),(+$7),(+$9),(+$11)); //@line 304 "modules/physics/box2d/wrap_Body.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if (!($13)) {
  STACKTOP = sp;return 0; //@line 305 "modules/physics/box2d/wrap_Body.cpp"
 }
 $14 = ___cxa_find_matching_catch_3(8992|0)|0;
 $15 = tempRet0;
 $16 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $17 = ($15|0)==($16|0); //@line 535 "./common/runtime.h"
 if (!($17)) {
  $$01$i = $15;$$02$i = $14;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $18 = (___cxa_begin_catch(($14|0))|0); //@line 535 "./common/runtime.h"
 $19 = HEAP32[$18>>2]|0; //@line 539 "./common/runtime.h"
 $20 = ((($19)) + 8|0); //@line 539 "./common/runtime.h"
 $21 = HEAP32[$20>>2]|0; //@line 539 "./common/runtime.h"
 $22 = (FUNCTION_TABLE_ii[$21 & 8191]($18)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($22|0)); //@line 539 "./common/runtime.h"
 $23 = __THREW__; __THREW__ = 0;
 $24 = $23&1;
 if (!($24)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $25 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $25; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 305 "modules/physics/box2d/wrap_Body.cpp"
 }
 $26 = ___cxa_find_matching_catch_2()|0;
 $27 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if ($29) {
  $30 = ___cxa_find_matching_catch_3(0|0)|0;
  $31 = tempRet0;
  ___clang_call_terminate($30); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $27;$$02$i = $26;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d14w_Body_setMassEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 308 "modules/physics/box2d/wrap_Body.cpp"
 $vararg_buffer = sp; //@line 308 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 311 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 311 "modules/physics/box2d/wrap_Body.cpp"
 __THREW__ = 0;
 invoke_vid(3719,($0|0),(+$5)); //@line 312 "modules/physics/box2d/wrap_Body.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  STACKTOP = sp;return 0; //@line 313 "modules/physics/box2d/wrap_Body.cpp"
 }
 $8 = ___cxa_find_matching_catch_3(8992|0)|0;
 $9 = tempRet0;
 $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $11 = ($9|0)==($10|0); //@line 535 "./common/runtime.h"
 if (!($11)) {
  $$01$i = $9;$$02$i = $8;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $12 = (___cxa_begin_catch(($8|0))|0); //@line 535 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = ((($13)) + 8|0); //@line 539 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($16|0)); //@line 539 "./common/runtime.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if (!($18)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $19 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $19; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 313 "modules/physics/box2d/wrap_Body.cpp"
 }
 $20 = ___cxa_find_matching_catch_2()|0;
 $21 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if ($23) {
  $24 = ___cxa_find_matching_catch_3(0|0)|0;
  $25 = tempRet0;
  ___clang_call_terminate($24); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $21;$$02$i = $20;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d17w_Body_setInertiaEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 316 "modules/physics/box2d/wrap_Body.cpp"
 $vararg_buffer = sp; //@line 316 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 319 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 319 "modules/physics/box2d/wrap_Body.cpp"
 __THREW__ = 0;
 invoke_vid(3720,($0|0),(+$5)); //@line 320 "modules/physics/box2d/wrap_Body.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  STACKTOP = sp;return 0; //@line 321 "modules/physics/box2d/wrap_Body.cpp"
 }
 $8 = ___cxa_find_matching_catch_3(8992|0)|0;
 $9 = tempRet0;
 $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $11 = ($9|0)==($10|0); //@line 535 "./common/runtime.h"
 if (!($11)) {
  $$01$i = $9;$$02$i = $8;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $12 = (___cxa_begin_catch(($8|0))|0); //@line 535 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = ((($13)) + 8|0); //@line 539 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($16|0)); //@line 539 "./common/runtime.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if (!($18)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $19 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $19; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 321 "modules/physics/box2d/wrap_Body.cpp"
 }
 $20 = ___cxa_find_matching_catch_2()|0;
 $21 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if ($23) {
  $24 = ___cxa_find_matching_catch_3(0|0)|0;
  $25 = tempRet0;
  ___clang_call_terminate($24); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $21;$$02$i = $20;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d24w_Body_setAngularDampingEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 324 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 327 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 327 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body17setAngularDampingEf($0,$5); //@line 328 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 0; //@line 329 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d23w_Body_setLinearDampingEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 332 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 335 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 335 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body16setLinearDampingEf($0,$5); //@line 336 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 0; //@line 337 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d22w_Body_setGravityScaleEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 340 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 343 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 343 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body15setGravityScaleEf($0,$5); //@line 344 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 0; //@line 345 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d14w_Body_setTypeEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $$idx1$val$val = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $type = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $type = sp + 12|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (_luaL_checklstring($L,2,0)|0); //@line 351 "modules/physics/box2d/wrap_Body.cpp"
 (__ZN4love7physics4Body11getConstantEPKcRNS1_4TypeE($4,$type)|0); //@line 353 "modules/physics/box2d/wrap_Body.cpp"
 $$idx1$val$val = HEAP32[$type>>2]|0;
 __THREW__ = 0;
 invoke_vii(3721,($0|0),($$idx1$val$val|0)); //@line 354 "modules/physics/box2d/wrap_Body.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  STACKTOP = sp;return 0; //@line 355 "modules/physics/box2d/wrap_Body.cpp"
 }
 $7 = ___cxa_find_matching_catch_3(8992|0)|0;
 $8 = tempRet0;
 $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
 if (!($10)) {
  $$01$i = $8;$$02$i = $7;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
 $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
 $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if (!($17)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $18 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $18; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 355 "modules/physics/box2d/wrap_Body.cpp"
 }
 $19 = ___cxa_find_matching_catch_2()|0;
 $20 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $23 = ___cxa_find_matching_catch_3(0|0)|0;
  $24 = tempRet0;
  ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $20;$$02$i = $19;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d20w_Body_getWorldPointEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $vararg_buffer = 0, $x_o = 0, $y_o = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $x_o = sp + 8|0;
 $y_o = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 362 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 362 "modules/physics/box2d/wrap_Body.cpp"
 $6 = (+_luaL_checknumber($L,3)); //@line 363 "modules/physics/box2d/wrap_Body.cpp"
 $7 = $6; //@line 363 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body13getWorldPointEffRfS3_($0,$5,$7,$x_o,$y_o); //@line 365 "modules/physics/box2d/wrap_Body.cpp"
 $8 = +HEAPF32[$x_o>>2]; //@line 366 "modules/physics/box2d/wrap_Body.cpp"
 $9 = $8; //@line 366 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$9); //@line 366 "modules/physics/box2d/wrap_Body.cpp"
 $10 = +HEAPF32[$y_o>>2]; //@line 367 "modules/physics/box2d/wrap_Body.cpp"
 $11 = $10; //@line 367 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$11); //@line 367 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 2; //@line 369 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d21w_Body_getWorldVectorEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $vararg_buffer = 0, $x_o = 0, $y_o = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $x_o = sp + 8|0;
 $y_o = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 376 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 376 "modules/physics/box2d/wrap_Body.cpp"
 $6 = (+_luaL_checknumber($L,3)); //@line 377 "modules/physics/box2d/wrap_Body.cpp"
 $7 = $6; //@line 377 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body14getWorldVectorEffRfS3_($0,$5,$7,$x_o,$y_o); //@line 379 "modules/physics/box2d/wrap_Body.cpp"
 $8 = +HEAPF32[$x_o>>2]; //@line 380 "modules/physics/box2d/wrap_Body.cpp"
 $9 = $8; //@line 380 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$9); //@line 380 "modules/physics/box2d/wrap_Body.cpp"
 $10 = +HEAPF32[$y_o>>2]; //@line 381 "modules/physics/box2d/wrap_Body.cpp"
 $11 = $10; //@line 381 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$11); //@line 381 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 2; //@line 383 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d21w_Body_getWorldPointsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 386 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 _lua_remove($L,1); //@line 389 "modules/physics/box2d/wrap_Body.cpp"
 $4 = (__ZN4love7physics5box2d4Body14getWorldPointsEP9lua_State($0,$L)|0); //@line 390 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return ($4|0); //@line 390 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d20w_Body_getLocalPointEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $vararg_buffer = 0, $x_o = 0, $y_o = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $x_o = sp + 8|0;
 $y_o = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 397 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 397 "modules/physics/box2d/wrap_Body.cpp"
 $6 = (+_luaL_checknumber($L,3)); //@line 398 "modules/physics/box2d/wrap_Body.cpp"
 $7 = $6; //@line 398 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body13getLocalPointEffRfS3_($0,$5,$7,$x_o,$y_o); //@line 400 "modules/physics/box2d/wrap_Body.cpp"
 $8 = +HEAPF32[$x_o>>2]; //@line 401 "modules/physics/box2d/wrap_Body.cpp"
 $9 = $8; //@line 401 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$9); //@line 401 "modules/physics/box2d/wrap_Body.cpp"
 $10 = +HEAPF32[$y_o>>2]; //@line 402 "modules/physics/box2d/wrap_Body.cpp"
 $11 = $10; //@line 402 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$11); //@line 402 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 2; //@line 404 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d21w_Body_getLocalVectorEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $vararg_buffer = 0, $x_o = 0, $y_o = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $x_o = sp + 8|0;
 $y_o = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 411 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 411 "modules/physics/box2d/wrap_Body.cpp"
 $6 = (+_luaL_checknumber($L,3)); //@line 412 "modules/physics/box2d/wrap_Body.cpp"
 $7 = $6; //@line 412 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body14getLocalVectorEffRfS3_($0,$5,$7,$x_o,$y_o); //@line 414 "modules/physics/box2d/wrap_Body.cpp"
 $8 = +HEAPF32[$x_o>>2]; //@line 415 "modules/physics/box2d/wrap_Body.cpp"
 $9 = $8; //@line 415 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$9); //@line 415 "modules/physics/box2d/wrap_Body.cpp"
 $10 = +HEAPF32[$y_o>>2]; //@line 416 "modules/physics/box2d/wrap_Body.cpp"
 $11 = $10; //@line 416 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$11); //@line 416 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 2; //@line 418 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d38w_Body_getLinearVelocityFromWorldPointEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $vararg_buffer = 0, $x_o = 0, $y_o = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $x_o = sp + 8|0;
 $y_o = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 425 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 425 "modules/physics/box2d/wrap_Body.cpp"
 $6 = (+_luaL_checknumber($L,3)); //@line 426 "modules/physics/box2d/wrap_Body.cpp"
 $7 = $6; //@line 426 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body31getLinearVelocityFromWorldPointEffRfS3_($0,$5,$7,$x_o,$y_o); //@line 428 "modules/physics/box2d/wrap_Body.cpp"
 $8 = +HEAPF32[$x_o>>2]; //@line 429 "modules/physics/box2d/wrap_Body.cpp"
 $9 = $8; //@line 429 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$9); //@line 429 "modules/physics/box2d/wrap_Body.cpp"
 $10 = +HEAPF32[$y_o>>2]; //@line 430 "modules/physics/box2d/wrap_Body.cpp"
 $11 = $10; //@line 430 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$11); //@line 430 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 2; //@line 432 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d38w_Body_getLinearVelocityFromLocalPointEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $vararg_buffer = 0, $x_o = 0, $y_o = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $x_o = sp + 8|0;
 $y_o = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (+_luaL_checknumber($L,2)); //@line 439 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4; //@line 439 "modules/physics/box2d/wrap_Body.cpp"
 $6 = (+_luaL_checknumber($L,3)); //@line 440 "modules/physics/box2d/wrap_Body.cpp"
 $7 = $6; //@line 440 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body31getLinearVelocityFromLocalPointEffRfS3_($0,$5,$7,$x_o,$y_o); //@line 442 "modules/physics/box2d/wrap_Body.cpp"
 $8 = +HEAPF32[$x_o>>2]; //@line 443 "modules/physics/box2d/wrap_Body.cpp"
 $9 = $8; //@line 443 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$9); //@line 443 "modules/physics/box2d/wrap_Body.cpp"
 $10 = +HEAPF32[$y_o>>2]; //@line 444 "modules/physics/box2d/wrap_Body.cpp"
 $11 = $10; //@line 444 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushnumber($L,$11); //@line 444 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 2; //@line 446 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d15w_Body_isBulletEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 449 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (__ZNK4love7physics5box2d4Body8isBulletEv($0)|0); //@line 452 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 452 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 1; //@line 453 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d16w_Body_setBulletEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 456 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,2)|0); //@line 459 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body9setBulletEb($0,$4); //@line 460 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 0; //@line 461 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d15w_Body_isActiveEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 464 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (__ZNK4love7physics5box2d4Body8isActiveEv($0)|0); //@line 467 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 467 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 1; //@line 468 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d14w_Body_isAwakeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 471 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (__ZNK4love7physics5box2d4Body7isAwakeEv($0)|0); //@line 474 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 474 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 1; //@line 475 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d25w_Body_setSleepingAllowedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 478 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,2)|0); //@line 481 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body18setSleepingAllowedEb($0,$4); //@line 482 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 0; //@line 483 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d24w_Body_isSleepingAllowedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 486 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (__ZNK4love7physics5box2d4Body17isSleepingAllowedEv($0)|0); //@line 489 "modules/physics/box2d/wrap_Body.cpp"
 $5 = $4&1; //@line 489 "modules/physics/box2d/wrap_Body.cpp"
 _lua_pushboolean($L,$5); //@line 489 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 1; //@line 490 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d16w_Body_setActiveEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 493 "modules/physics/box2d/wrap_Body.cpp"
 $vararg_buffer = sp; //@line 493 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,2)|0); //@line 496 "modules/physics/box2d/wrap_Body.cpp"
 __THREW__ = 0;
 invoke_vii(3722,($0|0),($4|0)); //@line 497 "modules/physics/box2d/wrap_Body.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  STACKTOP = sp;return 0; //@line 498 "modules/physics/box2d/wrap_Body.cpp"
 }
 $7 = ___cxa_find_matching_catch_3(8992|0)|0;
 $8 = tempRet0;
 $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
 if (!($10)) {
  $$01$i = $8;$$02$i = $7;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
 $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
 $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if (!($17)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $18 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $18; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 498 "modules/physics/box2d/wrap_Body.cpp"
 }
 $19 = ___cxa_find_matching_catch_2()|0;
 $20 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $23 = ___cxa_find_matching_catch_3(0|0)|0;
  $24 = tempRet0;
  ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $20;$$02$i = $19;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d15w_Body_setAwakeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 501 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,2)|0); //@line 504 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love7physics5box2d4Body8setAwakeEb($0,$4); //@line 505 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 0; //@line 506 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d23w_Body_setFixedRotationEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 509 "modules/physics/box2d/wrap_Body.cpp"
 $vararg_buffer = sp; //@line 509 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,2)|0); //@line 512 "modules/physics/box2d/wrap_Body.cpp"
 __THREW__ = 0;
 invoke_vii(3723,($0|0),($4|0)); //@line 513 "modules/physics/box2d/wrap_Body.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  STACKTOP = sp;return 0; //@line 514 "modules/physics/box2d/wrap_Body.cpp"
 }
 $7 = ___cxa_find_matching_catch_3(8992|0)|0;
 $8 = tempRet0;
 $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
 if (!($10)) {
  $$01$i = $8;$$02$i = $7;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
 $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
 $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if (!($17)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $18 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $18; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 514 "modules/physics/box2d/wrap_Body.cpp"
 }
 $19 = ___cxa_find_matching_catch_2()|0;
 $20 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $23 = ___cxa_find_matching_catch_3(0|0)|0;
  $24 = tempRet0;
  ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $20;$$02$i = $19;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d22w_Body_isFixedRotationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 517 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (__ZNK4love7physics5box2d4Body15isFixedRotationEv($0)|0); //@line 520 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 521 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 1; //@line 522 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d15w_Body_getWorldEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 525 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 $4 = (__ZNK4love7physics5box2d4Body8getWorldEv($0)|0); //@line 528 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,32,$4); //@line 529 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return 1; //@line 530 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d21w_Body_getFixtureListEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $n$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 533 "modules/physics/box2d/wrap_Body.cpp"
 $vararg_buffer = sp; //@line 533 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 _lua_remove($L,1); //@line 536 "modules/physics/box2d/wrap_Body.cpp"
 __THREW__ = 0;
 $4 = (invoke_iii(3724,($0|0),($L|0))|0); //@line 538 "modules/physics/box2d/wrap_Body.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  $n$0 = $4;
  STACKTOP = sp;return ($n$0|0); //@line 539 "modules/physics/box2d/wrap_Body.cpp"
 }
 $7 = ___cxa_find_matching_catch_3(8992|0)|0;
 $8 = tempRet0;
 $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
 if (!($10)) {
  $$01$i = $8;$$02$i = $7;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
 $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
 $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if (!($17)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $18 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $18; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  $n$0 = 0;
  STACKTOP = sp;return ($n$0|0); //@line 539 "modules/physics/box2d/wrap_Body.cpp"
 }
 $19 = ___cxa_find_matching_catch_2()|0;
 $20 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $23 = ___cxa_find_matching_catch_3(0|0)|0;
  $24 = tempRet0;
  ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $20;$$02$i = $19;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d19w_Body_getJointListEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $n$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 542 "modules/physics/box2d/wrap_Body.cpp"
 $vararg_buffer = sp; //@line 542 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 _lua_remove($L,1); //@line 545 "modules/physics/box2d/wrap_Body.cpp"
 __THREW__ = 0;
 $4 = (invoke_iii(3725,($0|0),($L|0))|0); //@line 547 "modules/physics/box2d/wrap_Body.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  $n$0 = $4;
  STACKTOP = sp;return ($n$0|0); //@line 548 "modules/physics/box2d/wrap_Body.cpp"
 }
 $7 = ___cxa_find_matching_catch_3(8992|0)|0;
 $8 = tempRet0;
 $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
 if (!($10)) {
  $$01$i = $8;$$02$i = $7;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
 $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
 $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if (!($17)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $18 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $18; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  $n$0 = 0;
  STACKTOP = sp;return ($n$0|0); //@line 548 "modules/physics/box2d/wrap_Body.cpp"
 }
 $19 = ___cxa_find_matching_catch_2()|0;
 $20 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $23 = ___cxa_find_matching_catch_3(0|0)|0;
  $24 = tempRet0;
  ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $20;$$02$i = $19;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d21w_Body_getContactListEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $n$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 551 "modules/physics/box2d/wrap_Body.cpp"
 $vararg_buffer = sp; //@line 551 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 _lua_remove($L,1); //@line 554 "modules/physics/box2d/wrap_Body.cpp"
 __THREW__ = 0;
 $4 = (invoke_iii(3726,($0|0),($L|0))|0); //@line 556 "modules/physics/box2d/wrap_Body.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  $n$0 = $4;
  STACKTOP = sp;return ($n$0|0); //@line 557 "modules/physics/box2d/wrap_Body.cpp"
 }
 $7 = ___cxa_find_matching_catch_3(8992|0)|0;
 $8 = tempRet0;
 $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
 if (!($10)) {
  $$01$i = $8;$$02$i = $7;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
 $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
 $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if (!($17)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $18 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $18; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  $n$0 = 0;
  STACKTOP = sp;return ($n$0|0); //@line 557 "modules/physics/box2d/wrap_Body.cpp"
 }
 $19 = ___cxa_find_matching_catch_2()|0;
 $20 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $23 = ___cxa_find_matching_catch_3(0|0)|0;
  $24 = tempRet0;
  ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $20;$$02$i = $19;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d14w_Body_destroyEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 560 "modules/physics/box2d/wrap_Body.cpp"
 $vararg_buffer = sp; //@line 560 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 __THREW__ = 0;
 invoke_vi(3727,($0|0)); //@line 563 "modules/physics/box2d/wrap_Body.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  STACKTOP = sp;return 0; //@line 564 "modules/physics/box2d/wrap_Body.cpp"
 }
 $6 = ___cxa_find_matching_catch_3(8992|0)|0;
 $7 = tempRet0;
 $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
 if (!($9)) {
  $$01$i = $7;$$02$i = $6;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $17; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 564 "modules/physics/box2d/wrap_Body.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d18w_Body_isDestroyedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 569 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 570 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 570 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 570 "modules/physics/box2d/wrap_Body.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$3); //@line 570 "modules/physics/box2d/wrap_Body.cpp"
 return 1; //@line 571 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d18w_Body_setUserDataEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 574 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 _lua_remove($L,1); //@line 577 "modules/physics/box2d/wrap_Body.cpp"
 $4 = (__ZN4love7physics5box2d4Body11setUserDataEP9lua_State($0,$L)|0); //@line 578 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return ($4|0); //@line 578 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love7physics5box2d18w_Body_getUserDataEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 581 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d4BodyEEEPT_P9lua_StateiNS_4TypeE($L,1,34)|0); //@line 33 "modules/physics/box2d/wrap_Body.cpp"
 $1 = ((($0)) + 8|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 $3 = ($2|0)==(0|0); //@line 34 "modules/physics/box2d/wrap_Body.cpp"
 if ($3) {
  (_luaL_error($L,356955,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Body.cpp"
 }
 _lua_remove($L,1); //@line 584 "modules/physics/box2d/wrap_Body.cpp"
 $4 = (__ZN4love7physics5box2d4Body11getUserDataEP9lua_State($0,$L)|0); //@line 585 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return ($4|0); //@line 585 "modules/physics/box2d/wrap_Body.cpp"
}
function _luaopen_body($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 651 "modules/physics/box2d/wrap_Body.cpp"
 HEAP32[$vararg_buffer>>2] = 17488; //@line 653 "modules/physics/box2d/wrap_Body.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 653 "modules/physics/box2d/wrap_Body.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 653 "modules/physics/box2d/wrap_Body.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,34,357014,$vararg_buffer)|0); //@line 653 "modules/physics/box2d/wrap_Body.cpp"
 STACKTOP = sp;return ($0|0); //@line 653 "modules/physics/box2d/wrap_Body.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d10ChainShapeEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d26w_ChainShape_setNextVertexEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 37 "modules/physics/box2d/wrap_ChainShape.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10ChainShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,40)|0); //@line 34 "modules/physics/box2d/wrap_ChainShape.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 40 "modules/physics/box2d/wrap_ChainShape.cpp"
 $2 = $1; //@line 40 "modules/physics/box2d/wrap_ChainShape.cpp"
 $3 = (+_luaL_checknumber($L,3)); //@line 41 "modules/physics/box2d/wrap_ChainShape.cpp"
 $4 = $3; //@line 41 "modules/physics/box2d/wrap_ChainShape.cpp"
 __THREW__ = 0;
 invoke_vidd(3728,($0|0),(+$2),(+$4)); //@line 42 "modules/physics/box2d/wrap_ChainShape.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  STACKTOP = sp;return 0; //@line 43 "modules/physics/box2d/wrap_ChainShape.cpp"
 }
 $7 = ___cxa_find_matching_catch_3(8992|0)|0;
 $8 = tempRet0;
 $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
 if (!($10)) {
  $$01$i = $8;$$02$i = $7;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
 $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
 $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if (!($17)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $18 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $18; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 43 "modules/physics/box2d/wrap_ChainShape.cpp"
 }
 $19 = ___cxa_find_matching_catch_2()|0;
 $20 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $23 = ___cxa_find_matching_catch_3(0|0)|0;
  $24 = tempRet0;
  ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $20;$$02$i = $19;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d30w_ChainShape_setPreviousVertexEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 46 "modules/physics/box2d/wrap_ChainShape.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10ChainShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,40)|0); //@line 34 "modules/physics/box2d/wrap_ChainShape.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 49 "modules/physics/box2d/wrap_ChainShape.cpp"
 $2 = $1; //@line 49 "modules/physics/box2d/wrap_ChainShape.cpp"
 $3 = (+_luaL_checknumber($L,3)); //@line 50 "modules/physics/box2d/wrap_ChainShape.cpp"
 $4 = $3; //@line 50 "modules/physics/box2d/wrap_ChainShape.cpp"
 __THREW__ = 0;
 invoke_vidd(3729,($0|0),(+$2),(+$4)); //@line 51 "modules/physics/box2d/wrap_ChainShape.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  STACKTOP = sp;return 0; //@line 52 "modules/physics/box2d/wrap_ChainShape.cpp"
 }
 $7 = ___cxa_find_matching_catch_3(8992|0)|0;
 $8 = tempRet0;
 $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
 if (!($10)) {
  $$01$i = $8;$$02$i = $7;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
 $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
 $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if (!($17)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $18 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $18; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 52 "modules/physics/box2d/wrap_ChainShape.cpp"
 }
 $19 = ___cxa_find_matching_catch_2()|0;
 $20 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $23 = ___cxa_find_matching_catch_3(0|0)|0;
  $24 = tempRet0;
  ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $20;$$02$i = $19;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d25w_ChainShape_getChildEdgeEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $e$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 55 "modules/physics/box2d/wrap_ChainShape.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10ChainShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,40)|0); //@line 34 "modules/physics/box2d/wrap_ChainShape.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 58 "modules/physics/box2d/wrap_ChainShape.cpp"
 $2 = (~~(($1))); //@line 58 "modules/physics/box2d/wrap_ChainShape.cpp"
 $3 = (($2) + -1)|0; //@line 58 "modules/physics/box2d/wrap_ChainShape.cpp"
 __THREW__ = 0;
 $4 = (invoke_iii(3730,($0|0),($3|0))|0); //@line 60 "modules/physics/box2d/wrap_ChainShape.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 do {
  if ($6) {
   $7 = ___cxa_find_matching_catch_3(8992|0)|0;
   $8 = tempRet0;
   $9 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $10 = ($8|0)==($9|0); //@line 535 "./common/runtime.h"
   if (!($10)) {
    $$01$i = $8;$$02$i = $7;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $11 = (___cxa_begin_catch(($7|0))|0); //@line 535 "./common/runtime.h"
   $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
   $13 = ((($12)) + 8|0); //@line 539 "./common/runtime.h"
   $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
   $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($15|0)); //@line 539 "./common/runtime.h"
   $16 = __THREW__; __THREW__ = 0;
   $17 = $16&1;
   if (!($17)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $18 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $18; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $e$0 = 0;
    break;
   }
   $19 = ___cxa_find_matching_catch_2()|0;
   $20 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $21 = __THREW__; __THREW__ = 0;
   $22 = $21&1;
   if ($22) {
    $23 = ___cxa_find_matching_catch_3(0|0)|0;
    $24 = tempRet0;
    ___clang_call_terminate($23); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $20;$$02$i = $19;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $e$0 = $4;
  }
 } while(0);
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,39,$e$0); //@line 61 "modules/physics/box2d/wrap_ChainShape.cpp"
 $25 = HEAP32[$e$0>>2]|0; //@line 62 "modules/physics/box2d/wrap_ChainShape.cpp"
 $26 = ((($25)) + 12|0); //@line 62 "modules/physics/box2d/wrap_ChainShape.cpp"
 $27 = HEAP32[$26>>2]|0; //@line 62 "modules/physics/box2d/wrap_ChainShape.cpp"
 FUNCTION_TABLE_vi[$27 & 8191]($e$0); //@line 62 "modules/physics/box2d/wrap_ChainShape.cpp"
 STACKTOP = sp;return 1; //@line 63 "modules/physics/box2d/wrap_ChainShape.cpp"
}
function __ZN4love7physics5box2d27w_ChainShape_getVertexCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10ChainShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,40)|0); //@line 34 "modules/physics/box2d/wrap_ChainShape.cpp"
 $1 = (__ZNK4love7physics5box2d10ChainShape14getVertexCountEv($0)|0); //@line 69 "modules/physics/box2d/wrap_ChainShape.cpp"
 _lua_pushinteger($L,$1); //@line 70 "modules/physics/box2d/wrap_ChainShape.cpp"
 return 1; //@line 71 "modules/physics/box2d/wrap_ChainShape.cpp"
}
function __ZN4love7physics5box2d21w_ChainShape_getPointEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0.0, $phitmp1 = 0.0, $phitmp3 = 0.0, $v$sroa$0$0 = 0.0, $v$sroa$4$0 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp + 8|0;
 $0 = sp;
 $1 = (__ZN4love14luax_checktypeINS_7physics5box2d10ChainShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,40)|0); //@line 34 "modules/physics/box2d/wrap_ChainShape.cpp"
 $2 = (+_luaL_checknumber($L,2)); //@line 77 "modules/physics/box2d/wrap_ChainShape.cpp"
 $3 = (~~(($2))); //@line 77 "modules/physics/box2d/wrap_ChainShape.cpp"
 $4 = (($3) + -1)|0; //@line 77 "modules/physics/box2d/wrap_ChainShape.cpp"
 __THREW__ = 0;
 invoke_viii(3731,($0|0),($1|0),($4|0)); //@line 79 "modules/physics/box2d/wrap_ChainShape.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  $7 = +HEAPF32[$0>>2]; //@line 79 "modules/physics/box2d/wrap_ChainShape.cpp"
  $8 = ((($0)) + 4|0); //@line 79 "modules/physics/box2d/wrap_ChainShape.cpp"
  $9 = +HEAPF32[$8>>2]; //@line 79 "modules/physics/box2d/wrap_ChainShape.cpp"
  $phitmp1 = $7; //@line 79 "modules/physics/box2d/wrap_ChainShape.cpp"
  $phitmp3 = $9; //@line 79 "modules/physics/box2d/wrap_ChainShape.cpp"
  $v$sroa$0$0 = $phitmp1;$v$sroa$4$0 = $phitmp3;
  _lua_pushnumber($L,$v$sroa$0$0); //@line 80 "modules/physics/box2d/wrap_ChainShape.cpp"
  _lua_pushnumber($L,$v$sroa$4$0); //@line 81 "modules/physics/box2d/wrap_ChainShape.cpp"
  STACKTOP = sp;return 2; //@line 82 "modules/physics/box2d/wrap_ChainShape.cpp"
 }
 $10 = ___cxa_find_matching_catch_3(8992|0)|0;
 $11 = tempRet0;
 $12 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $13 = ($11|0)==($12|0); //@line 535 "./common/runtime.h"
 if (!($13)) {
  $$01$i = $11;$$02$i = $10;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $14 = (___cxa_begin_catch(($10|0))|0); //@line 535 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = ((($15)) + 8|0); //@line 539 "./common/runtime.h"
 $17 = HEAP32[$16>>2]|0; //@line 539 "./common/runtime.h"
 $18 = (FUNCTION_TABLE_ii[$17 & 8191]($14)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($18|0)); //@line 539 "./common/runtime.h"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if (!($20)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $21 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $21; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $v$sroa$0$0 = 0.0;$v$sroa$4$0 = 0.0;
  _lua_pushnumber($L,$v$sroa$0$0); //@line 80 "modules/physics/box2d/wrap_ChainShape.cpp"
  _lua_pushnumber($L,$v$sroa$4$0); //@line 81 "modules/physics/box2d/wrap_ChainShape.cpp"
  STACKTOP = sp;return 2; //@line 82 "modules/physics/box2d/wrap_ChainShape.cpp"
 }
 $22 = ___cxa_find_matching_catch_2()|0;
 $23 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $26 = ___cxa_find_matching_catch_3(0|0)|0;
  $27 = tempRet0;
  ___clang_call_terminate($26); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $23;$$02$i = $22;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d22w_ChainShape_getPointsEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, $v = 0, $vararg_buffer = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $v = sp + 8|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10ChainShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,40)|0); //@line 34 "modules/physics/box2d/wrap_ChainShape.cpp"
 $1 = (__ZNK4love7physics5box2d10ChainShape9getPointsEv($0)|0); //@line 88 "modules/physics/box2d/wrap_ChainShape.cpp"
 $2 = (__ZNK4love7physics5box2d10ChainShape14getVertexCountEv($0)|0); //@line 89 "modules/physics/box2d/wrap_ChainShape.cpp"
 $3 = $2 << 1; //@line 90 "modules/physics/box2d/wrap_ChainShape.cpp"
 $4 = (_lua_checkstack($L,$3)|0); //@line 90 "modules/physics/box2d/wrap_ChainShape.cpp"
 $5 = ($4|0)==(0); //@line 90 "modules/physics/box2d/wrap_ChainShape.cpp"
 if ($5) {
  $8 = (_luaL_error($L,357735,$vararg_buffer)|0); //@line 91 "modules/physics/box2d/wrap_ChainShape.cpp"
  $$0 = $8;
  STACKTOP = sp;return ($$0|0); //@line 99 "modules/physics/box2d/wrap_ChainShape.cpp"
 }
 $6 = ($2|0)>(0); //@line 92 "modules/physics/box2d/wrap_ChainShape.cpp"
 if (!($6)) {
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 99 "modules/physics/box2d/wrap_ChainShape.cpp"
 }
 $7 = ((($v)) + 4|0); //@line 96 "modules/physics/box2d/wrap_ChainShape.cpp"
 $i$01 = 0;
 while(1) {
  $9 = (($1) + ($i$01<<3)|0); //@line 94 "modules/physics/box2d/wrap_ChainShape.cpp"
  __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($v,$9); //@line 94 "modules/physics/box2d/wrap_ChainShape.cpp"
  $10 = +HEAPF32[$v>>2]; //@line 95 "modules/physics/box2d/wrap_ChainShape.cpp"
  $11 = $10; //@line 95 "modules/physics/box2d/wrap_ChainShape.cpp"
  _lua_pushnumber($L,$11); //@line 95 "modules/physics/box2d/wrap_ChainShape.cpp"
  $12 = +HEAPF32[$7>>2]; //@line 96 "modules/physics/box2d/wrap_ChainShape.cpp"
  $13 = $12; //@line 96 "modules/physics/box2d/wrap_ChainShape.cpp"
  _lua_pushnumber($L,$13); //@line 96 "modules/physics/box2d/wrap_ChainShape.cpp"
  $14 = (($i$01) + 1)|0; //@line 92 "modules/physics/box2d/wrap_ChainShape.cpp"
  $exitcond = ($14|0)==($2|0); //@line 92 "modules/physics/box2d/wrap_ChainShape.cpp"
  if ($exitcond) {
   $$0 = $3;
   break;
  } else {
   $i$01 = $14;
  }
 }
 STACKTOP = sp;return ($$0|0); //@line 99 "modules/physics/box2d/wrap_ChainShape.cpp"
}
function _luaopen_chainshape($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 112 "modules/physics/box2d/wrap_ChainShape.cpp"
 HEAP32[$vararg_buffer>>2] = 10748; //@line 114 "modules/physics/box2d/wrap_ChainShape.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 114 "modules/physics/box2d/wrap_ChainShape.cpp"
 HEAP32[$vararg_ptr1>>2] = 17960; //@line 114 "modules/physics/box2d/wrap_ChainShape.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 114 "modules/physics/box2d/wrap_ChainShape.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 114 "modules/physics/box2d/wrap_ChainShape.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,40,357758,$vararg_buffer)|0); //@line 114 "modules/physics/box2d/wrap_ChainShape.cpp"
 STACKTOP = sp;return ($0|0); //@line 114 "modules/physics/box2d/wrap_ChainShape.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d11CircleShapeEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d23w_CircleShape_getRadiusEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d11CircleShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,37)|0); //@line 32 "modules/physics/box2d/wrap_CircleShape.cpp"
 $1 = (+__ZNK4love7physics5box2d11CircleShape9getRadiusEv($0)); //@line 38 "modules/physics/box2d/wrap_CircleShape.cpp"
 $2 = $1; //@line 38 "modules/physics/box2d/wrap_CircleShape.cpp"
 _lua_pushnumber($L,$2); //@line 38 "modules/physics/box2d/wrap_CircleShape.cpp"
 return 1; //@line 39 "modules/physics/box2d/wrap_CircleShape.cpp"
}
function __ZN4love7physics5box2d23w_CircleShape_setRadiusEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d11CircleShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,37)|0); //@line 32 "modules/physics/box2d/wrap_CircleShape.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 45 "modules/physics/box2d/wrap_CircleShape.cpp"
 $2 = $1; //@line 45 "modules/physics/box2d/wrap_CircleShape.cpp"
 __ZN4love7physics5box2d11CircleShape9setRadiusEf($0,$2); //@line 46 "modules/physics/box2d/wrap_CircleShape.cpp"
 return 0; //@line 47 "modules/physics/box2d/wrap_CircleShape.cpp"
}
function __ZN4love7physics5box2d22w_CircleShape_getPointEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $x = 0, $y = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $x = sp + 4|0;
 $y = sp;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d11CircleShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,37)|0); //@line 32 "modules/physics/box2d/wrap_CircleShape.cpp"
 __ZNK4love7physics5box2d11CircleShape8getPointERfS3_($0,$x,$y); //@line 54 "modules/physics/box2d/wrap_CircleShape.cpp"
 $1 = +HEAPF32[$x>>2]; //@line 55 "modules/physics/box2d/wrap_CircleShape.cpp"
 $2 = $1; //@line 55 "modules/physics/box2d/wrap_CircleShape.cpp"
 _lua_pushnumber($L,$2); //@line 55 "modules/physics/box2d/wrap_CircleShape.cpp"
 $3 = +HEAPF32[$y>>2]; //@line 56 "modules/physics/box2d/wrap_CircleShape.cpp"
 $4 = $3; //@line 56 "modules/physics/box2d/wrap_CircleShape.cpp"
 _lua_pushnumber($L,$4); //@line 56 "modules/physics/box2d/wrap_CircleShape.cpp"
 STACKTOP = sp;return 2; //@line 57 "modules/physics/box2d/wrap_CircleShape.cpp"
}
function __ZN4love7physics5box2d22w_CircleShape_setPointEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d11CircleShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,37)|0); //@line 32 "modules/physics/box2d/wrap_CircleShape.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 63 "modules/physics/box2d/wrap_CircleShape.cpp"
 $2 = $1; //@line 63 "modules/physics/box2d/wrap_CircleShape.cpp"
 $3 = (+_luaL_checknumber($L,3)); //@line 63 "modules/physics/box2d/wrap_CircleShape.cpp"
 $4 = $3; //@line 63 "modules/physics/box2d/wrap_CircleShape.cpp"
 __ZN4love7physics5box2d11CircleShape8setPointEff($0,$2,$4); //@line 63 "modules/physics/box2d/wrap_CircleShape.cpp"
 return 0; //@line 64 "modules/physics/box2d/wrap_CircleShape.cpp"
}
function _luaopen_circleshape($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 76 "modules/physics/box2d/wrap_CircleShape.cpp"
 HEAP32[$vararg_buffer>>2] = 10748; //@line 78 "modules/physics/box2d/wrap_CircleShape.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 78 "modules/physics/box2d/wrap_CircleShape.cpp"
 HEAP32[$vararg_ptr1>>2] = 18016; //@line 78 "modules/physics/box2d/wrap_CircleShape.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 78 "modules/physics/box2d/wrap_CircleShape.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 78 "modules/physics/box2d/wrap_CircleShape.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,37,357829,$vararg_buffer)|0); //@line 78 "modules/physics/box2d/wrap_CircleShape.cpp"
 STACKTOP = sp;return ($0|0); //@line 78 "modules/physics/box2d/wrap_CircleShape.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d7ContactEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d22w_Contact_getPositionsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 39 "modules/physics/box2d/wrap_Contact.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7ContactEEEPT_P9lua_StateiNS_4TypeE($L,1,33)|0); //@line 33 "modules/physics/box2d/wrap_Contact.cpp"
 $1 = (__ZN4love7physics5box2d7Contact7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Contact.cpp"
 if (!($1)) {
  (_luaL_error($L,357879,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Contact.cpp"
 }
 $2 = (__ZN4love7physics5box2d7Contact12getPositionsEP9lua_State($0,$L)|0); //@line 42 "modules/physics/box2d/wrap_Contact.cpp"
 STACKTOP = sp;return ($2|0); //@line 42 "modules/physics/box2d/wrap_Contact.cpp"
}
function __ZN4love7physics5box2d19w_Contact_getNormalEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 45 "modules/physics/box2d/wrap_Contact.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7ContactEEEPT_P9lua_StateiNS_4TypeE($L,1,33)|0); //@line 33 "modules/physics/box2d/wrap_Contact.cpp"
 $1 = (__ZN4love7physics5box2d7Contact7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Contact.cpp"
 if (!($1)) {
  (_luaL_error($L,357879,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Contact.cpp"
 }
 $2 = (__ZN4love7physics5box2d7Contact9getNormalEP9lua_State($0,$L)|0); //@line 48 "modules/physics/box2d/wrap_Contact.cpp"
 STACKTOP = sp;return ($2|0); //@line 48 "modules/physics/box2d/wrap_Contact.cpp"
}
function __ZN4love7physics5box2d21w_Contact_getFrictionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 51 "modules/physics/box2d/wrap_Contact.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7ContactEEEPT_P9lua_StateiNS_4TypeE($L,1,33)|0); //@line 33 "modules/physics/box2d/wrap_Contact.cpp"
 $1 = (__ZN4love7physics5box2d7Contact7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Contact.cpp"
 if (!($1)) {
  (_luaL_error($L,357879,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Contact.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d7Contact11getFrictionEv($0)); //@line 54 "modules/physics/box2d/wrap_Contact.cpp"
 $3 = $2; //@line 54 "modules/physics/box2d/wrap_Contact.cpp"
 _lua_pushnumber($L,$3); //@line 54 "modules/physics/box2d/wrap_Contact.cpp"
 STACKTOP = sp;return 1; //@line 55 "modules/physics/box2d/wrap_Contact.cpp"
}
function __ZN4love7physics5box2d24w_Contact_getRestitutionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 58 "modules/physics/box2d/wrap_Contact.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7ContactEEEPT_P9lua_StateiNS_4TypeE($L,1,33)|0); //@line 33 "modules/physics/box2d/wrap_Contact.cpp"
 $1 = (__ZN4love7physics5box2d7Contact7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Contact.cpp"
 if (!($1)) {
  (_luaL_error($L,357879,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Contact.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d7Contact14getRestitutionEv($0)); //@line 61 "modules/physics/box2d/wrap_Contact.cpp"
 $3 = $2; //@line 61 "modules/physics/box2d/wrap_Contact.cpp"
 _lua_pushnumber($L,$3); //@line 61 "modules/physics/box2d/wrap_Contact.cpp"
 STACKTOP = sp;return 1; //@line 62 "modules/physics/box2d/wrap_Contact.cpp"
}
function __ZN4love7physics5box2d19w_Contact_isEnabledEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 65 "modules/physics/box2d/wrap_Contact.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7ContactEEEPT_P9lua_StateiNS_4TypeE($L,1,33)|0); //@line 33 "modules/physics/box2d/wrap_Contact.cpp"
 $1 = (__ZN4love7physics5box2d7Contact7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Contact.cpp"
 if (!($1)) {
  (_luaL_error($L,357879,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Contact.cpp"
 }
 $2 = (__ZNK4love7physics5box2d7Contact9isEnabledEv($0)|0); //@line 68 "modules/physics/box2d/wrap_Contact.cpp"
 $3 = $2&1; //@line 68 "modules/physics/box2d/wrap_Contact.cpp"
 _lua_pushboolean($L,$3); //@line 68 "modules/physics/box2d/wrap_Contact.cpp"
 STACKTOP = sp;return 1; //@line 69 "modules/physics/box2d/wrap_Contact.cpp"
}
function __ZN4love7physics5box2d20w_Contact_isTouchingEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 72 "modules/physics/box2d/wrap_Contact.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7ContactEEEPT_P9lua_StateiNS_4TypeE($L,1,33)|0); //@line 33 "modules/physics/box2d/wrap_Contact.cpp"
 $1 = (__ZN4love7physics5box2d7Contact7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Contact.cpp"
 if (!($1)) {
  (_luaL_error($L,357879,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Contact.cpp"
 }
 $2 = (__ZNK4love7physics5box2d7Contact10isTouchingEv($0)|0); //@line 75 "modules/physics/box2d/wrap_Contact.cpp"
 $3 = $2&1; //@line 75 "modules/physics/box2d/wrap_Contact.cpp"
 _lua_pushboolean($L,$3); //@line 75 "modules/physics/box2d/wrap_Contact.cpp"
 STACKTOP = sp;return 1; //@line 76 "modules/physics/box2d/wrap_Contact.cpp"
}
function __ZN4love7physics5box2d21w_Contact_setFrictionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 79 "modules/physics/box2d/wrap_Contact.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7ContactEEEPT_P9lua_StateiNS_4TypeE($L,1,33)|0); //@line 33 "modules/physics/box2d/wrap_Contact.cpp"
 $1 = (__ZN4love7physics5box2d7Contact7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Contact.cpp"
 if (!($1)) {
  (_luaL_error($L,357879,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Contact.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 82 "modules/physics/box2d/wrap_Contact.cpp"
 $3 = $2; //@line 82 "modules/physics/box2d/wrap_Contact.cpp"
 __ZN4love7physics5box2d7Contact11setFrictionEf($0,$3); //@line 83 "modules/physics/box2d/wrap_Contact.cpp"
 STACKTOP = sp;return 0; //@line 84 "modules/physics/box2d/wrap_Contact.cpp"
}
function __ZN4love7physics5box2d24w_Contact_setRestitutionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 87 "modules/physics/box2d/wrap_Contact.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7ContactEEEPT_P9lua_StateiNS_4TypeE($L,1,33)|0); //@line 33 "modules/physics/box2d/wrap_Contact.cpp"
 $1 = (__ZN4love7physics5box2d7Contact7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Contact.cpp"
 if (!($1)) {
  (_luaL_error($L,357879,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Contact.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 90 "modules/physics/box2d/wrap_Contact.cpp"
 $3 = $2; //@line 90 "modules/physics/box2d/wrap_Contact.cpp"
 __ZN4love7physics5box2d7Contact14setRestitutionEf($0,$3); //@line 91 "modules/physics/box2d/wrap_Contact.cpp"
 STACKTOP = sp;return 0; //@line 92 "modules/physics/box2d/wrap_Contact.cpp"
}
function __ZN4love7physics5box2d20w_Contact_setEnabledEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 95 "modules/physics/box2d/wrap_Contact.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7ContactEEEPT_P9lua_StateiNS_4TypeE($L,1,33)|0); //@line 33 "modules/physics/box2d/wrap_Contact.cpp"
 $1 = (__ZN4love7physics5box2d7Contact7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Contact.cpp"
 if (!($1)) {
  (_luaL_error($L,357879,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Contact.cpp"
 }
 $2 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,2)|0); //@line 98 "modules/physics/box2d/wrap_Contact.cpp"
 __ZN4love7physics5box2d7Contact10setEnabledEb($0,$2); //@line 99 "modules/physics/box2d/wrap_Contact.cpp"
 STACKTOP = sp;return 0; //@line 100 "modules/physics/box2d/wrap_Contact.cpp"
}
function __ZN4love7physics5box2d23w_Contact_resetFrictionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 103 "modules/physics/box2d/wrap_Contact.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7ContactEEEPT_P9lua_StateiNS_4TypeE($L,1,33)|0); //@line 33 "modules/physics/box2d/wrap_Contact.cpp"
 $1 = (__ZN4love7physics5box2d7Contact7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Contact.cpp"
 if (!($1)) {
  (_luaL_error($L,357879,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Contact.cpp"
 }
 __ZN4love7physics5box2d7Contact13resetFrictionEv($0); //@line 106 "modules/physics/box2d/wrap_Contact.cpp"
 STACKTOP = sp;return 0; //@line 107 "modules/physics/box2d/wrap_Contact.cpp"
}
function __ZN4love7physics5box2d26w_Contact_resetRestitutionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 110 "modules/physics/box2d/wrap_Contact.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7ContactEEEPT_P9lua_StateiNS_4TypeE($L,1,33)|0); //@line 33 "modules/physics/box2d/wrap_Contact.cpp"
 $1 = (__ZN4love7physics5box2d7Contact7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Contact.cpp"
 if (!($1)) {
  (_luaL_error($L,357879,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Contact.cpp"
 }
 __ZN4love7physics5box2d7Contact16resetRestitutionEv($0); //@line 113 "modules/physics/box2d/wrap_Contact.cpp"
 STACKTOP = sp;return 0; //@line 114 "modules/physics/box2d/wrap_Contact.cpp"
}
function __ZN4love7physics5box2d25w_Contact_setTangentSpeedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 117 "modules/physics/box2d/wrap_Contact.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7ContactEEEPT_P9lua_StateiNS_4TypeE($L,1,33)|0); //@line 33 "modules/physics/box2d/wrap_Contact.cpp"
 $1 = (__ZN4love7physics5box2d7Contact7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Contact.cpp"
 if (!($1)) {
  (_luaL_error($L,357879,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Contact.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 120 "modules/physics/box2d/wrap_Contact.cpp"
 $3 = $2; //@line 120 "modules/physics/box2d/wrap_Contact.cpp"
 __ZN4love7physics5box2d7Contact15setTangentSpeedEf($0,$3); //@line 121 "modules/physics/box2d/wrap_Contact.cpp"
 STACKTOP = sp;return 0; //@line 122 "modules/physics/box2d/wrap_Contact.cpp"
}
function __ZN4love7physics5box2d25w_Contact_getTangentSpeedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 125 "modules/physics/box2d/wrap_Contact.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7ContactEEEPT_P9lua_StateiNS_4TypeE($L,1,33)|0); //@line 33 "modules/physics/box2d/wrap_Contact.cpp"
 $1 = (__ZN4love7physics5box2d7Contact7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Contact.cpp"
 if (!($1)) {
  (_luaL_error($L,357879,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Contact.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d7Contact15getTangentSpeedEv($0)); //@line 128 "modules/physics/box2d/wrap_Contact.cpp"
 $3 = $2; //@line 128 "modules/physics/box2d/wrap_Contact.cpp"
 _lua_pushnumber($L,$3); //@line 128 "modules/physics/box2d/wrap_Contact.cpp"
 STACKTOP = sp;return 1; //@line 129 "modules/physics/box2d/wrap_Contact.cpp"
}
function __ZN4love7physics5box2d21w_Contact_getChildrenEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0.0, $a = 0, $b = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $a = sp + 8|0;
 $b = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7ContactEEEPT_P9lua_StateiNS_4TypeE($L,1,33)|0); //@line 33 "modules/physics/box2d/wrap_Contact.cpp"
 $1 = (__ZN4love7physics5box2d7Contact7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Contact.cpp"
 if (!($1)) {
  (_luaL_error($L,357879,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Contact.cpp"
 }
 __ZN4love7physics5box2d7Contact11getChildrenERiS3_($0,$a,$b); //@line 136 "modules/physics/box2d/wrap_Contact.cpp"
 $2 = HEAP32[$a>>2]|0; //@line 137 "modules/physics/box2d/wrap_Contact.cpp"
 $3 = (($2) + 1)|0; //@line 137 "modules/physics/box2d/wrap_Contact.cpp"
 $4 = (+($3|0)); //@line 137 "modules/physics/box2d/wrap_Contact.cpp"
 _lua_pushnumber($L,$4); //@line 137 "modules/physics/box2d/wrap_Contact.cpp"
 $5 = HEAP32[$b>>2]|0; //@line 138 "modules/physics/box2d/wrap_Contact.cpp"
 $6 = (($5) + 1)|0; //@line 138 "modules/physics/box2d/wrap_Contact.cpp"
 $7 = (+($6|0)); //@line 138 "modules/physics/box2d/wrap_Contact.cpp"
 _lua_pushnumber($L,$7); //@line 138 "modules/physics/box2d/wrap_Contact.cpp"
 STACKTOP = sp;return 2; //@line 139 "modules/physics/box2d/wrap_Contact.cpp"
}
function __ZN4love7physics5box2d21w_Contact_getFixturesEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $a = 0, $b = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $a = sp + 16|0;
 $b = sp + 12|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7ContactEEEPT_P9lua_StateiNS_4TypeE($L,1,33)|0); //@line 33 "modules/physics/box2d/wrap_Contact.cpp"
 $1 = (__ZN4love7physics5box2d7Contact7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Contact.cpp"
 if (!($1)) {
  (_luaL_error($L,357879,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Contact.cpp"
 }
 HEAP32[$a>>2] = 0; //@line 145 "modules/physics/box2d/wrap_Contact.cpp"
 HEAP32[$b>>2] = 0; //@line 146 "modules/physics/box2d/wrap_Contact.cpp"
 __THREW__ = 0;
 invoke_viii(3732,($0|0),($a|0),($b|0)); //@line 147 "modules/physics/box2d/wrap_Contact.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  $22 = HEAP32[$a>>2]|0; //@line 149 "modules/physics/box2d/wrap_Contact.cpp"
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,35,$22); //@line 149 "modules/physics/box2d/wrap_Contact.cpp"
  $23 = HEAP32[$b>>2]|0; //@line 150 "modules/physics/box2d/wrap_Contact.cpp"
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,35,$23); //@line 150 "modules/physics/box2d/wrap_Contact.cpp"
  STACKTOP = sp;return 2; //@line 151 "modules/physics/box2d/wrap_Contact.cpp"
 }
 $4 = ___cxa_find_matching_catch_3(8992|0)|0;
 $5 = tempRet0;
 $6 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $7 = ($5|0)==($6|0); //@line 535 "./common/runtime.h"
 if (!($7)) {
  $$01$i = $5;$$02$i = $4;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $8 = (___cxa_begin_catch(($4|0))|0); //@line 535 "./common/runtime.h"
 $9 = HEAP32[$8>>2]|0; //@line 539 "./common/runtime.h"
 $10 = ((($9)) + 8|0); //@line 539 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($8)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($12|0)); //@line 539 "./common/runtime.h"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if (!($14)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $15 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $15; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  $22 = HEAP32[$a>>2]|0; //@line 149 "modules/physics/box2d/wrap_Contact.cpp"
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,35,$22); //@line 149 "modules/physics/box2d/wrap_Contact.cpp"
  $23 = HEAP32[$b>>2]|0; //@line 150 "modules/physics/box2d/wrap_Contact.cpp"
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,35,$23); //@line 150 "modules/physics/box2d/wrap_Contact.cpp"
  STACKTOP = sp;return 2; //@line 151 "modules/physics/box2d/wrap_Contact.cpp"
 }
 $16 = ___cxa_find_matching_catch_2()|0;
 $17 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $20 = ___cxa_find_matching_catch_3(0|0)|0;
  $21 = tempRet0;
  ___clang_call_terminate($20); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $17;$$02$i = $16;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d21w_Contact_isDestroyedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7ContactEEEPT_P9lua_StateiNS_4TypeE($L,1,33)|0); //@line 156 "modules/physics/box2d/wrap_Contact.cpp"
 $1 = (__ZN4love7physics5box2d7Contact7isValidEv($0)|0); //@line 157 "modules/physics/box2d/wrap_Contact.cpp"
 $2 = $1 ^ 1; //@line 157 "modules/physics/box2d/wrap_Contact.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$2); //@line 157 "modules/physics/box2d/wrap_Contact.cpp"
 return 1; //@line 158 "modules/physics/box2d/wrap_Contact.cpp"
}
function _luaopen_contact($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 182 "modules/physics/box2d/wrap_Contact.cpp"
 HEAP32[$vararg_buffer>>2] = 18056; //@line 184 "modules/physics/box2d/wrap_Contact.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 184 "modules/physics/box2d/wrap_Contact.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 184 "modules/physics/box2d/wrap_Contact.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,33,357913,$vararg_buffer)|0); //@line 184 "modules/physics/box2d/wrap_Contact.cpp"
 STACKTOP = sp;return ($0|0); //@line 184 "modules/physics/box2d/wrap_Contact.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d13DistanceJointEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d25w_DistanceJoint_setLengthEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 38 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13DistanceJointEEEPT_P9lua_StateiNS_4TypeE($L,1,43)|0); //@line 32 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 41 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $3 = $2; //@line 41 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 __ZN4love7physics5box2d13DistanceJoint9setLengthEf($0,$3); //@line 42 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 STACKTOP = sp;return 0; //@line 43 "modules/physics/box2d/wrap_DistanceJoint.cpp"
}
function __ZN4love7physics5box2d25w_DistanceJoint_getLengthEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 46 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13DistanceJointEEEPT_P9lua_StateiNS_4TypeE($L,1,43)|0); //@line 32 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d13DistanceJoint9getLengthEv($0)); //@line 49 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $3 = $2; //@line 49 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 _lua_pushnumber($L,$3); //@line 49 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 STACKTOP = sp;return 1; //@line 50 "modules/physics/box2d/wrap_DistanceJoint.cpp"
}
function __ZN4love7physics5box2d28w_DistanceJoint_setFrequencyEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 53 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13DistanceJointEEEPT_P9lua_StateiNS_4TypeE($L,1,43)|0); //@line 32 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 56 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $3 = $2; //@line 56 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 __ZN4love7physics5box2d13DistanceJoint12setFrequencyEf($0,$3); //@line 57 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 STACKTOP = sp;return 0; //@line 58 "modules/physics/box2d/wrap_DistanceJoint.cpp"
}
function __ZN4love7physics5box2d28w_DistanceJoint_getFrequencyEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 61 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13DistanceJointEEEPT_P9lua_StateiNS_4TypeE($L,1,43)|0); //@line 32 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d13DistanceJoint12getFrequencyEv($0)); //@line 64 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $3 = $2; //@line 64 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 _lua_pushnumber($L,$3); //@line 64 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 STACKTOP = sp;return 1; //@line 65 "modules/physics/box2d/wrap_DistanceJoint.cpp"
}
function __ZN4love7physics5box2d31w_DistanceJoint_setDampingRatioEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 68 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13DistanceJointEEEPT_P9lua_StateiNS_4TypeE($L,1,43)|0); //@line 32 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 71 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $3 = $2; //@line 71 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 __ZN4love7physics5box2d13DistanceJoint15setDampingRatioEf($0,$3); //@line 72 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 STACKTOP = sp;return 0; //@line 73 "modules/physics/box2d/wrap_DistanceJoint.cpp"
}
function __ZN4love7physics5box2d31w_DistanceJoint_getDampingRatioEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 76 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13DistanceJointEEEPT_P9lua_StateiNS_4TypeE($L,1,43)|0); //@line 32 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d13DistanceJoint15getDampingRatioEv($0)); //@line 79 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $3 = $2; //@line 79 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 _lua_pushnumber($L,$3); //@line 79 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 STACKTOP = sp;return 1; //@line 80 "modules/physics/box2d/wrap_DistanceJoint.cpp"
}
function _luaopen_distancejoint($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 94 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 HEAP32[$vararg_buffer>>2] = 18560; //@line 96 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 96 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 HEAP32[$vararg_ptr1>>2] = 18192; //@line 96 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 96 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 96 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,43,358063,$vararg_buffer)|0); //@line 96 "modules/physics/box2d/wrap_DistanceJoint.cpp"
 STACKTOP = sp;return ($0|0); //@line 96 "modules/physics/box2d/wrap_DistanceJoint.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d9EdgeShapeEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d21w_EdgeShape_getPointsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d9EdgeShapeEEEPT_P9lua_StateiNS_4TypeE($L,1,39)|0); //@line 32 "modules/physics/box2d/wrap_EdgeShape.cpp"
 _lua_remove($L,1); //@line 38 "modules/physics/box2d/wrap_EdgeShape.cpp"
 $1 = (__ZN4love7physics5box2d9EdgeShape9getPointsEP9lua_State($0,$L)|0); //@line 39 "modules/physics/box2d/wrap_EdgeShape.cpp"
 return ($1|0); //@line 39 "modules/physics/box2d/wrap_EdgeShape.cpp"
}
function _luaopen_edgeshape($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 48 "modules/physics/box2d/wrap_EdgeShape.cpp"
 HEAP32[$vararg_buffer>>2] = 10748; //@line 50 "modules/physics/box2d/wrap_EdgeShape.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 50 "modules/physics/box2d/wrap_EdgeShape.cpp"
 HEAP32[$vararg_ptr1>>2] = 18248; //@line 50 "modules/physics/box2d/wrap_EdgeShape.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 50 "modules/physics/box2d/wrap_EdgeShape.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 50 "modules/physics/box2d/wrap_EdgeShape.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,39,358097,$vararg_buffer)|0); //@line 50 "modules/physics/box2d/wrap_EdgeShape.cpp"
 STACKTOP = sp;return ($0|0); //@line 50 "modules/physics/box2d/wrap_EdgeShape.cpp"
}
function __ZN4love7physics5box2d17w_Fixture_getTypeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $type = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $type = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 HEAP32[$type>>2] = 1282466; //@line 42 "modules/physics/box2d/wrap_Fixture.cpp"
 $2 = (__ZNK4love7physics5box2d7Fixture7getTypeEv($0)|0); //@line 43 "modules/physics/box2d/wrap_Fixture.cpp"
 (__ZN4love7physics5Shape11getConstantENS1_4TypeERPKc($2,$type)|0); //@line 43 "modules/physics/box2d/wrap_Fixture.cpp"
 $3 = HEAP32[$type>>2]|0; //@line 44 "modules/physics/box2d/wrap_Fixture.cpp"
 _lua_pushstring($L,$3); //@line 44 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return 1; //@line 45 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d21w_Fixture_setFrictionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 48 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 51 "modules/physics/box2d/wrap_Fixture.cpp"
 $3 = $2; //@line 51 "modules/physics/box2d/wrap_Fixture.cpp"
 __ZN4love7physics5box2d7Fixture11setFrictionEf($0,$3); //@line 52 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return 0; //@line 53 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d24w_Fixture_setRestitutionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 56 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 59 "modules/physics/box2d/wrap_Fixture.cpp"
 $3 = $2; //@line 59 "modules/physics/box2d/wrap_Fixture.cpp"
 __ZN4love7physics5box2d7Fixture14setRestitutionEf($0,$3); //@line 60 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return 0; //@line 61 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d20w_Fixture_setDensityEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0.0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 64 "modules/physics/box2d/wrap_Fixture.cpp"
 $vararg_buffer = sp; //@line 64 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 67 "modules/physics/box2d/wrap_Fixture.cpp"
 $3 = $2; //@line 67 "modules/physics/box2d/wrap_Fixture.cpp"
 __THREW__ = 0;
 invoke_vid(3733,($0|0),(+$3)); //@line 68 "modules/physics/box2d/wrap_Fixture.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  STACKTOP = sp;return 0; //@line 69 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $6 = ___cxa_find_matching_catch_3(8992|0)|0;
 $7 = tempRet0;
 $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
 if (!($9)) {
  $$01$i = $7;$$02$i = $6;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $17; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 69 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d19w_Fixture_setSensorEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 72 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $2 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,2)|0); //@line 75 "modules/physics/box2d/wrap_Fixture.cpp"
 __ZN4love7physics5box2d7Fixture9setSensorEb($0,$2); //@line 76 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return 0; //@line 77 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d21w_Fixture_getFrictionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 80 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d7Fixture11getFrictionEv($0)); //@line 83 "modules/physics/box2d/wrap_Fixture.cpp"
 $3 = $2; //@line 83 "modules/physics/box2d/wrap_Fixture.cpp"
 _lua_pushnumber($L,$3); //@line 83 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return 1; //@line 84 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d24w_Fixture_getRestitutionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 87 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d7Fixture14getRestitutionEv($0)); //@line 90 "modules/physics/box2d/wrap_Fixture.cpp"
 $3 = $2; //@line 90 "modules/physics/box2d/wrap_Fixture.cpp"
 _lua_pushnumber($L,$3); //@line 90 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return 1; //@line 91 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d20w_Fixture_getDensityEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 94 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d7Fixture10getDensityEv($0)); //@line 97 "modules/physics/box2d/wrap_Fixture.cpp"
 $3 = $2; //@line 97 "modules/physics/box2d/wrap_Fixture.cpp"
 _lua_pushnumber($L,$3); //@line 97 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return 1; //@line 98 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d18w_Fixture_isSensorEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 101 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $2 = (__ZNK4love7physics5box2d7Fixture8isSensorEv($0)|0); //@line 104 "modules/physics/box2d/wrap_Fixture.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$2); //@line 104 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return 1; //@line 105 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d17w_Fixture_getBodyEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 108 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $2 = (__ZNK4love7physics5box2d7Fixture7getBodyEv($0)|0); //@line 111 "modules/physics/box2d/wrap_Fixture.cpp"
 $3 = ($2|0)==(0|0); //@line 112 "modules/physics/box2d/wrap_Fixture.cpp"
 if ($3) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 116 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,34,$2); //@line 114 "modules/physics/box2d/wrap_Fixture.cpp"
 $$0 = 1;
 STACKTOP = sp;return ($$0|0); //@line 116 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d18w_Fixture_getShapeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 118 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $2 = (__ZNK4love7physics5box2d7Fixture8getShapeEv($0)|0); //@line 121 "modules/physics/box2d/wrap_Fixture.cpp"
 $3 = ($2|0)==(0|0); //@line 122 "modules/physics/box2d/wrap_Fixture.cpp"
 if ($3) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 144 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $4 = (__ZNK4love7physics5box2d5Shape7getTypeEv($2)|0); //@line 124 "modules/physics/box2d/wrap_Fixture.cpp"
 switch ($4|0) {
 case 3:  {
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,39,$2); //@line 127 "modules/physics/box2d/wrap_Fixture.cpp"
  $$pre$phiZ2D = $2;
  break;
 }
 case 4:  {
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,40,$2); //@line 130 "modules/physics/box2d/wrap_Fixture.cpp"
  $$pre$phiZ2D = $2;
  break;
 }
 case 1:  {
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,37,$2); //@line 133 "modules/physics/box2d/wrap_Fixture.cpp"
  $$pre$phiZ2D = $2;
  break;
 }
 case 2:  {
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,38,$2); //@line 136 "modules/physics/box2d/wrap_Fixture.cpp"
  $$pre$phiZ2D = $2;
  break;
 }
 default: {
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,36,$2); //@line 139 "modules/physics/box2d/wrap_Fixture.cpp"
  $$pre$phiZ2D = $2;
 }
 }
 $5 = HEAP32[$2>>2]|0; //@line 142 "modules/physics/box2d/wrap_Fixture.cpp"
 $6 = ((($5)) + 12|0); //@line 142 "modules/physics/box2d/wrap_Fixture.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 142 "modules/physics/box2d/wrap_Fixture.cpp"
 FUNCTION_TABLE_vi[$7 & 8191]($$pre$phiZ2D); //@line 142 "modules/physics/box2d/wrap_Fixture.cpp"
 $$0 = 1;
 STACKTOP = sp;return ($$0|0); //@line 144 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d19w_Fixture_testPointEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 146 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 149 "modules/physics/box2d/wrap_Fixture.cpp"
 $3 = $2; //@line 149 "modules/physics/box2d/wrap_Fixture.cpp"
 $4 = (+_luaL_checknumber($L,3)); //@line 150 "modules/physics/box2d/wrap_Fixture.cpp"
 $5 = $4; //@line 150 "modules/physics/box2d/wrap_Fixture.cpp"
 $6 = (__ZNK4love7physics5box2d7Fixture9testPointEff($0,$3,$5)|0); //@line 151 "modules/physics/box2d/wrap_Fixture.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$6); //@line 151 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return 1; //@line 152 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d17w_Fixture_rayCastEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ret$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 155 "modules/physics/box2d/wrap_Fixture.cpp"
 $vararg_buffer = sp; //@line 155 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 _lua_remove($L,1); //@line 158 "modules/physics/box2d/wrap_Fixture.cpp"
 __THREW__ = 0;
 $2 = (invoke_iii(3734,($0|0),($L|0))|0); //@line 160 "modules/physics/box2d/wrap_Fixture.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  $ret$0 = $2;
  STACKTOP = sp;return ($ret$0|0); //@line 161 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $5 = ___cxa_find_matching_catch_3(8992|0)|0;
 $6 = tempRet0;
 $7 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $8 = ($6|0)==($7|0); //@line 535 "./common/runtime.h"
 if (!($8)) {
  $$01$i = $6;$$02$i = $5;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $9 = (___cxa_begin_catch(($5|0))|0); //@line 535 "./common/runtime.h"
 $10 = HEAP32[$9>>2]|0; //@line 539 "./common/runtime.h"
 $11 = ((($10)) + 8|0); //@line 539 "./common/runtime.h"
 $12 = HEAP32[$11>>2]|0; //@line 539 "./common/runtime.h"
 $13 = (FUNCTION_TABLE_ii[$12 & 8191]($9)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($13|0)); //@line 539 "./common/runtime.h"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if (!($15)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $16 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $16; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  $ret$0 = 0;
  STACKTOP = sp;return ($ret$0|0); //@line 161 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $17 = ___cxa_find_matching_catch_2()|0;
 $18 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $21 = ___cxa_find_matching_catch_3(0|0)|0;
  $22 = tempRet0;
  ___clang_call_terminate($21); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $18;$$02$i = $17;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d23w_Fixture_setFilterDataEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0, $v = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $v = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 168 "modules/physics/box2d/wrap_Fixture.cpp"
 $3 = (~~(($2))); //@line 168 "modules/physics/box2d/wrap_Fixture.cpp"
 HEAP32[$v>>2] = $3; //@line 168 "modules/physics/box2d/wrap_Fixture.cpp"
 $4 = (+_luaL_checknumber($L,3)); //@line 169 "modules/physics/box2d/wrap_Fixture.cpp"
 $5 = (~~(($4))); //@line 169 "modules/physics/box2d/wrap_Fixture.cpp"
 $6 = ((($v)) + 4|0); //@line 169 "modules/physics/box2d/wrap_Fixture.cpp"
 HEAP32[$6>>2] = $5; //@line 169 "modules/physics/box2d/wrap_Fixture.cpp"
 $7 = (+_luaL_checknumber($L,4)); //@line 170 "modules/physics/box2d/wrap_Fixture.cpp"
 $8 = (~~(($7))); //@line 170 "modules/physics/box2d/wrap_Fixture.cpp"
 $9 = ((($v)) + 8|0); //@line 170 "modules/physics/box2d/wrap_Fixture.cpp"
 HEAP32[$9>>2] = $8; //@line 170 "modules/physics/box2d/wrap_Fixture.cpp"
 __ZN4love7physics5box2d7Fixture13setFilterDataEPi($0,$v); //@line 171 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return 0; //@line 172 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d23w_Fixture_getFilterDataEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $v = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $v = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 __ZN4love7physics5box2d7Fixture13getFilterDataEPi($0,$v); //@line 179 "modules/physics/box2d/wrap_Fixture.cpp"
 $2 = HEAP32[$v>>2]|0; //@line 180 "modules/physics/box2d/wrap_Fixture.cpp"
 _lua_pushinteger($L,$2); //@line 180 "modules/physics/box2d/wrap_Fixture.cpp"
 $3 = ((($v)) + 4|0); //@line 181 "modules/physics/box2d/wrap_Fixture.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 181 "modules/physics/box2d/wrap_Fixture.cpp"
 _lua_pushinteger($L,$4); //@line 181 "modules/physics/box2d/wrap_Fixture.cpp"
 $5 = ((($v)) + 8|0); //@line 182 "modules/physics/box2d/wrap_Fixture.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 182 "modules/physics/box2d/wrap_Fixture.cpp"
 _lua_pushinteger($L,$6); //@line 182 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return 3; //@line 183 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d21w_Fixture_setCategoryEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 186 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 _lua_remove($L,1); //@line 189 "modules/physics/box2d/wrap_Fixture.cpp"
 $2 = (__ZN4love7physics5box2d7Fixture11setCategoryEP9lua_State($0,$L)|0); //@line 190 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return ($2|0); //@line 190 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d21w_Fixture_getCategoryEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 193 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 _lua_remove($L,1); //@line 196 "modules/physics/box2d/wrap_Fixture.cpp"
 $2 = (__ZN4love7physics5box2d7Fixture11getCategoryEP9lua_State($0,$L)|0); //@line 197 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return ($2|0); //@line 197 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d17w_Fixture_setMaskEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 200 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 _lua_remove($L,1); //@line 203 "modules/physics/box2d/wrap_Fixture.cpp"
 $2 = (__ZN4love7physics5box2d7Fixture7setMaskEP9lua_State($0,$L)|0); //@line 204 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return ($2|0); //@line 204 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d17w_Fixture_getMaskEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 207 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 _lua_remove($L,1); //@line 210 "modules/physics/box2d/wrap_Fixture.cpp"
 $2 = (__ZN4love7physics5box2d7Fixture7getMaskEP9lua_State($0,$L)|0); //@line 211 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return ($2|0); //@line 211 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d21w_Fixture_setUserDataEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 214 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 _lua_remove($L,1); //@line 217 "modules/physics/box2d/wrap_Fixture.cpp"
 $2 = (__ZN4love7physics5box2d7Fixture11setUserDataEP9lua_State($0,$L)|0); //@line 218 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return ($2|0); //@line 218 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d21w_Fixture_getUserDataEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 221 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 _lua_remove($L,1); //@line 224 "modules/physics/box2d/wrap_Fixture.cpp"
 $2 = (__ZN4love7physics5box2d7Fixture11getUserDataEP9lua_State($0,$L)|0); //@line 225 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return ($2|0); //@line 225 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d24w_Fixture_getBoundingBoxEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 228 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 _lua_remove($L,1); //@line 231 "modules/physics/box2d/wrap_Fixture.cpp"
 $2 = (__ZNK4love7physics5box2d7Fixture14getBoundingBoxEP9lua_State($0,$L)|0); //@line 232 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return ($2|0); //@line 232 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d21w_Fixture_getMassDataEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 235 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 _lua_remove($L,1); //@line 238 "modules/physics/box2d/wrap_Fixture.cpp"
 $2 = (__ZNK4love7physics5box2d7Fixture11getMassDataEP9lua_State($0,$L)|0); //@line 239 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return ($2|0); //@line 239 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d23w_Fixture_getGroupIndexEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 242 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $2 = (__ZNK4love7physics5box2d7Fixture13getGroupIndexEv($0)|0); //@line 245 "modules/physics/box2d/wrap_Fixture.cpp"
 _lua_pushinteger($L,$2); //@line 246 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return 1; //@line 247 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d23w_Fixture_setGroupIndexEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 250 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 253 "modules/physics/box2d/wrap_Fixture.cpp"
 $3 = (~~(($2))); //@line 253 "modules/physics/box2d/wrap_Fixture.cpp"
 __ZN4love7physics5box2d7Fixture13setGroupIndexEi($0,$3); //@line 254 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return 0; //@line 255 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love7physics5box2d17w_Fixture_destroyEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 258 "modules/physics/box2d/wrap_Fixture.cpp"
 $vararg_buffer = sp; //@line 258 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 33 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_Fixture.cpp"
 if (!($1)) {
  (_luaL_error($L,358117,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 __THREW__ = 0;
 invoke_vii(3735,($0|0),0); //@line 261 "modules/physics/box2d/wrap_Fixture.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  STACKTOP = sp;return 0; //@line 262 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $4 = ___cxa_find_matching_catch_3(8992|0)|0;
 $5 = tempRet0;
 $6 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $7 = ($5|0)==($6|0); //@line 535 "./common/runtime.h"
 if (!($7)) {
  $$01$i = $5;$$02$i = $4;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $8 = (___cxa_begin_catch(($4|0))|0); //@line 535 "./common/runtime.h"
 $9 = HEAP32[$8>>2]|0; //@line 539 "./common/runtime.h"
 $10 = ((($9)) + 8|0); //@line 539 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($8)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($12|0)); //@line 539 "./common/runtime.h"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if (!($14)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $15 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $15; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 262 "modules/physics/box2d/wrap_Fixture.cpp"
 }
 $16 = ___cxa_find_matching_catch_2()|0;
 $17 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $20 = ___cxa_find_matching_catch_3(0|0)|0;
  $21 = tempRet0;
  ___clang_call_terminate($20); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $17;$$02$i = $16;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d21w_Fixture_isDestroyedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d7FixtureEEEPT_P9lua_StateiNS_4TypeE($L,1,35)|0); //@line 267 "modules/physics/box2d/wrap_Fixture.cpp"
 $1 = (__ZNK4love7physics5box2d7Fixture7isValidEv($0)|0); //@line 268 "modules/physics/box2d/wrap_Fixture.cpp"
 $2 = $1 ^ 1; //@line 268 "modules/physics/box2d/wrap_Fixture.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$2); //@line 268 "modules/physics/box2d/wrap_Fixture.cpp"
 return 1; //@line 269 "modules/physics/box2d/wrap_Fixture.cpp"
}
function _luaopen_fixture($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 304 "modules/physics/box2d/wrap_Fixture.cpp"
 HEAP32[$vararg_buffer>>2] = 18264; //@line 306 "modules/physics/box2d/wrap_Fixture.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 306 "modules/physics/box2d/wrap_Fixture.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 306 "modules/physics/box2d/wrap_Fixture.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,35,358151,$vararg_buffer)|0); //@line 306 "modules/physics/box2d/wrap_Fixture.cpp"
 STACKTOP = sp;return ($0|0); //@line 306 "modules/physics/box2d/wrap_Fixture.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d13FrictionJointEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d27w_FrictionJoint_setMaxForceEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0.0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 39 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $vararg_buffer = sp; //@line 39 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13FrictionJointEEEPT_P9lua_StateiNS_4TypeE($L,1,48)|0); //@line 33 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 42 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $3 = $2; //@line 42 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 __THREW__ = 0;
 invoke_vid(3736,($0|0),(+$3)); //@line 43 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  STACKTOP = sp;return 0; //@line 44 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 }
 $6 = ___cxa_find_matching_catch_3(8992|0)|0;
 $7 = tempRet0;
 $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
 if (!($9)) {
  $$01$i = $7;$$02$i = $6;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $17; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 44 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d27w_FrictionJoint_getMaxForceEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 47 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13FrictionJointEEEPT_P9lua_StateiNS_4TypeE($L,1,48)|0); //@line 33 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d13FrictionJoint11getMaxForceEv($0)); //@line 50 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $3 = $2; //@line 50 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 _lua_pushnumber($L,$3); //@line 50 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 STACKTOP = sp;return 1; //@line 51 "modules/physics/box2d/wrap_FrictionJoint.cpp"
}
function __ZN4love7physics5box2d28w_FrictionJoint_setMaxTorqueEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0.0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 54 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $vararg_buffer = sp; //@line 54 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13FrictionJointEEEPT_P9lua_StateiNS_4TypeE($L,1,48)|0); //@line 33 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 57 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $3 = $2; //@line 57 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 __THREW__ = 0;
 invoke_vid(3737,($0|0),(+$3)); //@line 58 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  STACKTOP = sp;return 0; //@line 59 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 }
 $6 = ___cxa_find_matching_catch_3(8992|0)|0;
 $7 = tempRet0;
 $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
 if (!($9)) {
  $$01$i = $7;$$02$i = $6;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $17; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 59 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d28w_FrictionJoint_getMaxTorqueEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 62 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d13FrictionJointEEEPT_P9lua_StateiNS_4TypeE($L,1,48)|0); //@line 33 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d13FrictionJoint12getMaxTorqueEv($0)); //@line 65 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $3 = $2; //@line 65 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 _lua_pushnumber($L,$3); //@line 65 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 STACKTOP = sp;return 1; //@line 66 "modules/physics/box2d/wrap_FrictionJoint.cpp"
}
function _luaopen_frictionjoint($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 78 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 HEAP32[$vararg_buffer>>2] = 18560; //@line 80 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 80 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 HEAP32[$vararg_ptr1>>2] = 18488; //@line 80 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 80 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 80 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,48,358412,$vararg_buffer)|0); //@line 80 "modules/physics/box2d/wrap_FrictionJoint.cpp"
 STACKTOP = sp;return ($0|0); //@line 80 "modules/physics/box2d/wrap_FrictionJoint.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d9GearJointEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d20w_GearJoint_setRatioEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0.0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 39 "modules/physics/box2d/wrap_GearJoint.cpp"
 $vararg_buffer = sp; //@line 39 "modules/physics/box2d/wrap_GearJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d9GearJointEEEPT_P9lua_StateiNS_4TypeE($L,1,47)|0); //@line 33 "modules/physics/box2d/wrap_GearJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_GearJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_GearJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 42 "modules/physics/box2d/wrap_GearJoint.cpp"
 $3 = $2; //@line 42 "modules/physics/box2d/wrap_GearJoint.cpp"
 __THREW__ = 0;
 invoke_vid(3738,($0|0),(+$3)); //@line 43 "modules/physics/box2d/wrap_GearJoint.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  STACKTOP = sp;return 0; //@line 44 "modules/physics/box2d/wrap_GearJoint.cpp"
 }
 $6 = ___cxa_find_matching_catch_3(8992|0)|0;
 $7 = tempRet0;
 $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
 if (!($9)) {
  $$01$i = $7;$$02$i = $6;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $17; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 44 "modules/physics/box2d/wrap_GearJoint.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d20w_GearJoint_getRatioEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 47 "modules/physics/box2d/wrap_GearJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d9GearJointEEEPT_P9lua_StateiNS_4TypeE($L,1,47)|0); //@line 33 "modules/physics/box2d/wrap_GearJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_GearJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_GearJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d9GearJoint8getRatioEv($0)); //@line 50 "modules/physics/box2d/wrap_GearJoint.cpp"
 $3 = $2; //@line 50 "modules/physics/box2d/wrap_GearJoint.cpp"
 _lua_pushnumber($L,$3); //@line 50 "modules/physics/box2d/wrap_GearJoint.cpp"
 STACKTOP = sp;return 1; //@line 51 "modules/physics/box2d/wrap_GearJoint.cpp"
}
function __ZN4love7physics5box2d21w_GearJoint_getJointsEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $j1$0 = 0, $j1$1 = 0, $j2$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 54 "modules/physics/box2d/wrap_GearJoint.cpp"
 $vararg_buffer = sp; //@line 54 "modules/physics/box2d/wrap_GearJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d9GearJointEEEPT_P9lua_StateiNS_4TypeE($L,1,47)|0); //@line 33 "modules/physics/box2d/wrap_GearJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 34 "modules/physics/box2d/wrap_GearJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 35 "modules/physics/box2d/wrap_GearJoint.cpp"
 }
 __THREW__ = 0;
 $2 = (invoke_ii(3739,($0|0))|0); //@line 61 "modules/physics/box2d/wrap_GearJoint.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $j1$0 = 0;
 } else {
  __THREW__ = 0;
  $5 = (invoke_ii(3740,($0|0))|0); //@line 62 "modules/physics/box2d/wrap_GearJoint.cpp"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if ($7) {
   $j1$0 = $2;
  } else {
   $j1$1 = $2;$j2$0 = $5;
   __ZN4love7physics5box2d14luax_pushjointEP9lua_StatePNS1_5JointE($L,$j1$1); //@line 65 "modules/physics/box2d/wrap_GearJoint.cpp"
   __ZN4love7physics5box2d14luax_pushjointEP9lua_StatePNS1_5JointE($L,$j2$0); //@line 66 "modules/physics/box2d/wrap_GearJoint.cpp"
   STACKTOP = sp;return 2; //@line 67 "modules/physics/box2d/wrap_GearJoint.cpp"
  }
 }
 $8 = ___cxa_find_matching_catch_3(8992|0)|0;
 $9 = tempRet0;
 $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $11 = ($9|0)==($10|0); //@line 535 "./common/runtime.h"
 if (!($11)) {
  $$01$i = $9;$$02$i = $8;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $12 = (___cxa_begin_catch(($8|0))|0); //@line 535 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = ((($13)) + 8|0); //@line 539 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($16|0)); //@line 539 "./common/runtime.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if (!($18)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $19 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $19; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  $j1$1 = $j1$0;$j2$0 = 0;
  __ZN4love7physics5box2d14luax_pushjointEP9lua_StatePNS1_5JointE($L,$j1$1); //@line 65 "modules/physics/box2d/wrap_GearJoint.cpp"
  __ZN4love7physics5box2d14luax_pushjointEP9lua_StatePNS1_5JointE($L,$j2$0); //@line 66 "modules/physics/box2d/wrap_GearJoint.cpp"
  STACKTOP = sp;return 2; //@line 67 "modules/physics/box2d/wrap_GearJoint.cpp"
 }
 $20 = ___cxa_find_matching_catch_2()|0;
 $21 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if ($23) {
  $24 = ___cxa_find_matching_catch_3(0|0)|0;
  $25 = tempRet0;
  ___clang_call_terminate($24); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $21;$$02$i = $20;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function _luaopen_gearjoint($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 78 "modules/physics/box2d/wrap_GearJoint.cpp"
 HEAP32[$vararg_buffer>>2] = 18560; //@line 80 "modules/physics/box2d/wrap_GearJoint.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 80 "modules/physics/box2d/wrap_GearJoint.cpp"
 HEAP32[$vararg_ptr1>>2] = 18528; //@line 80 "modules/physics/box2d/wrap_GearJoint.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 80 "modules/physics/box2d/wrap_GearJoint.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 80 "modules/physics/box2d/wrap_GearJoint.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,47,358426,$vararg_buffer)|0); //@line 80 "modules/physics/box2d/wrap_GearJoint.cpp"
 STACKTOP = sp;return ($0|0); //@line 80 "modules/physics/box2d/wrap_GearJoint.cpp"
}
function __ZN4love7physics5box2d15w_Joint_getTypeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $type = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $type = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5JointEEEPT_P9lua_StateiNS_4TypeE($L,1,41)|0); //@line 69 "modules/physics/box2d/wrap_Joint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 70 "modules/physics/box2d/wrap_Joint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 71 "modules/physics/box2d/wrap_Joint.cpp"
 }
 HEAP32[$type>>2] = 1282466; //@line 78 "modules/physics/box2d/wrap_Joint.cpp"
 $2 = (__ZNK4love7physics5box2d5Joint7getTypeEv($0)|0); //@line 79 "modules/physics/box2d/wrap_Joint.cpp"
 (__ZN4love7physics5Joint11getConstantENS1_4TypeERPKc($2,$type)|0); //@line 79 "modules/physics/box2d/wrap_Joint.cpp"
 $3 = HEAP32[$type>>2]|0; //@line 80 "modules/physics/box2d/wrap_Joint.cpp"
 _lua_pushstring($L,$3); //@line 80 "modules/physics/box2d/wrap_Joint.cpp"
 STACKTOP = sp;return 1; //@line 81 "modules/physics/box2d/wrap_Joint.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d5JointEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d17w_Joint_getBodiesEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b1$0 = 0, $b1$1 = 0, $b2$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 84 "modules/physics/box2d/wrap_Joint.cpp"
 $vararg_buffer = sp; //@line 84 "modules/physics/box2d/wrap_Joint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5JointEEEPT_P9lua_StateiNS_4TypeE($L,1,41)|0); //@line 69 "modules/physics/box2d/wrap_Joint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 70 "modules/physics/box2d/wrap_Joint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 71 "modules/physics/box2d/wrap_Joint.cpp"
 }
 __THREW__ = 0;
 $2 = (invoke_ii(3741,($0|0))|0); //@line 91 "modules/physics/box2d/wrap_Joint.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $b1$0 = 0;
 } else {
  __THREW__ = 0;
  $5 = (invoke_ii(3742,($0|0))|0); //@line 92 "modules/physics/box2d/wrap_Joint.cpp"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if ($7) {
   $b1$0 = $2;
  } else {
   $b1$1 = $2;$b2$0 = $5;
   __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,34,$b1$1); //@line 95 "modules/physics/box2d/wrap_Joint.cpp"
   __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,34,$b2$0); //@line 96 "modules/physics/box2d/wrap_Joint.cpp"
   STACKTOP = sp;return 2; //@line 97 "modules/physics/box2d/wrap_Joint.cpp"
  }
 }
 $8 = ___cxa_find_matching_catch_3(8992|0)|0;
 $9 = tempRet0;
 $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $11 = ($9|0)==($10|0); //@line 535 "./common/runtime.h"
 if (!($11)) {
  $$01$i = $9;$$02$i = $8;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $12 = (___cxa_begin_catch(($8|0))|0); //@line 535 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = ((($13)) + 8|0); //@line 539 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 539 "./common/runtime.h"
 $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($16|0)); //@line 539 "./common/runtime.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if (!($18)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $19 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $19; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  $b1$1 = $b1$0;$b2$0 = 0;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,34,$b1$1); //@line 95 "modules/physics/box2d/wrap_Joint.cpp"
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,34,$b2$0); //@line 96 "modules/physics/box2d/wrap_Joint.cpp"
  STACKTOP = sp;return 2; //@line 97 "modules/physics/box2d/wrap_Joint.cpp"
 }
 $20 = ___cxa_find_matching_catch_2()|0;
 $21 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if ($23) {
  $24 = ___cxa_find_matching_catch_3(0|0)|0;
  $25 = tempRet0;
  ___clang_call_terminate($24); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $21;$$02$i = $20;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d18w_Joint_getAnchorsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 100 "modules/physics/box2d/wrap_Joint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5JointEEEPT_P9lua_StateiNS_4TypeE($L,1,41)|0); //@line 69 "modules/physics/box2d/wrap_Joint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 70 "modules/physics/box2d/wrap_Joint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 71 "modules/physics/box2d/wrap_Joint.cpp"
 }
 _lua_remove($L,1); //@line 103 "modules/physics/box2d/wrap_Joint.cpp"
 $2 = (__ZN4love7physics5box2d5Joint10getAnchorsEP9lua_State($0,$L)|0); //@line 104 "modules/physics/box2d/wrap_Joint.cpp"
 STACKTOP = sp;return ($2|0); //@line 104 "modules/physics/box2d/wrap_Joint.cpp"
}
function __ZN4love7physics5box2d24w_Joint_getReactionForceEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 107 "modules/physics/box2d/wrap_Joint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5JointEEEPT_P9lua_StateiNS_4TypeE($L,1,41)|0); //@line 69 "modules/physics/box2d/wrap_Joint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 70 "modules/physics/box2d/wrap_Joint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 71 "modules/physics/box2d/wrap_Joint.cpp"
 }
 _lua_remove($L,1); //@line 110 "modules/physics/box2d/wrap_Joint.cpp"
 $2 = (__ZN4love7physics5box2d5Joint16getReactionForceEP9lua_State($0,$L)|0); //@line 111 "modules/physics/box2d/wrap_Joint.cpp"
 STACKTOP = sp;return ($2|0); //@line 111 "modules/physics/box2d/wrap_Joint.cpp"
}
function __ZN4love7physics5box2d25w_Joint_getReactionTorqueEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 114 "modules/physics/box2d/wrap_Joint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5JointEEEPT_P9lua_StateiNS_4TypeE($L,1,41)|0); //@line 69 "modules/physics/box2d/wrap_Joint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 70 "modules/physics/box2d/wrap_Joint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 71 "modules/physics/box2d/wrap_Joint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 117 "modules/physics/box2d/wrap_Joint.cpp"
 $3 = $2; //@line 117 "modules/physics/box2d/wrap_Joint.cpp"
 $4 = (+__ZN4love7physics5box2d5Joint17getReactionTorqueEf($0,$3)); //@line 118 "modules/physics/box2d/wrap_Joint.cpp"
 $5 = $4; //@line 118 "modules/physics/box2d/wrap_Joint.cpp"
 _lua_pushnumber($L,$5); //@line 118 "modules/physics/box2d/wrap_Joint.cpp"
 STACKTOP = sp;return 1; //@line 119 "modules/physics/box2d/wrap_Joint.cpp"
}
function __ZN4love7physics5box2d27w_Joint_getCollideConnectedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 122 "modules/physics/box2d/wrap_Joint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5JointEEEPT_P9lua_StateiNS_4TypeE($L,1,41)|0); //@line 69 "modules/physics/box2d/wrap_Joint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 70 "modules/physics/box2d/wrap_Joint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 71 "modules/physics/box2d/wrap_Joint.cpp"
 }
 $2 = (__ZNK4love7physics5box2d5Joint19getCollideConnectedEv($0)|0); //@line 125 "modules/physics/box2d/wrap_Joint.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$2); //@line 125 "modules/physics/box2d/wrap_Joint.cpp"
 STACKTOP = sp;return 1; //@line 126 "modules/physics/box2d/wrap_Joint.cpp"
}
function __ZN4love7physics5box2d19w_Joint_setUserDataEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 129 "modules/physics/box2d/wrap_Joint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5JointEEEPT_P9lua_StateiNS_4TypeE($L,1,41)|0); //@line 69 "modules/physics/box2d/wrap_Joint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 70 "modules/physics/box2d/wrap_Joint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 71 "modules/physics/box2d/wrap_Joint.cpp"
 }
 _lua_remove($L,1); //@line 132 "modules/physics/box2d/wrap_Joint.cpp"
 $2 = (__ZN4love7physics5box2d5Joint11setUserDataEP9lua_State($0,$L)|0); //@line 133 "modules/physics/box2d/wrap_Joint.cpp"
 STACKTOP = sp;return ($2|0); //@line 133 "modules/physics/box2d/wrap_Joint.cpp"
}
function __ZN4love7physics5box2d19w_Joint_getUserDataEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 136 "modules/physics/box2d/wrap_Joint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5JointEEEPT_P9lua_StateiNS_4TypeE($L,1,41)|0); //@line 69 "modules/physics/box2d/wrap_Joint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 70 "modules/physics/box2d/wrap_Joint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 71 "modules/physics/box2d/wrap_Joint.cpp"
 }
 _lua_remove($L,1); //@line 139 "modules/physics/box2d/wrap_Joint.cpp"
 $2 = (__ZN4love7physics5box2d5Joint11getUserDataEP9lua_State($0,$L)|0); //@line 140 "modules/physics/box2d/wrap_Joint.cpp"
 STACKTOP = sp;return ($2|0); //@line 140 "modules/physics/box2d/wrap_Joint.cpp"
}
function __ZN4love7physics5box2d15w_Joint_destroyEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 143 "modules/physics/box2d/wrap_Joint.cpp"
 $vararg_buffer = sp; //@line 143 "modules/physics/box2d/wrap_Joint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5JointEEEPT_P9lua_StateiNS_4TypeE($L,1,41)|0); //@line 69 "modules/physics/box2d/wrap_Joint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 70 "modules/physics/box2d/wrap_Joint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 71 "modules/physics/box2d/wrap_Joint.cpp"
 }
 __THREW__ = 0;
 invoke_vii(3743,($0|0),0); //@line 146 "modules/physics/box2d/wrap_Joint.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  STACKTOP = sp;return 0; //@line 147 "modules/physics/box2d/wrap_Joint.cpp"
 }
 $4 = ___cxa_find_matching_catch_3(8992|0)|0;
 $5 = tempRet0;
 $6 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $7 = ($5|0)==($6|0); //@line 535 "./common/runtime.h"
 if (!($7)) {
  $$01$i = $5;$$02$i = $4;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $8 = (___cxa_begin_catch(($4|0))|0); //@line 535 "./common/runtime.h"
 $9 = HEAP32[$8>>2]|0; //@line 539 "./common/runtime.h"
 $10 = ((($9)) + 8|0); //@line 539 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($8)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($12|0)); //@line 539 "./common/runtime.h"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if (!($14)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $15 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $15; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 147 "modules/physics/box2d/wrap_Joint.cpp"
 }
 $16 = ___cxa_find_matching_catch_2()|0;
 $17 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $20 = ___cxa_find_matching_catch_3(0|0)|0;
  $21 = tempRet0;
  ___clang_call_terminate($20); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $17;$$02$i = $16;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d19w_Joint_isDestroyedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5JointEEEPT_P9lua_StateiNS_4TypeE($L,1,41)|0); //@line 152 "modules/physics/box2d/wrap_Joint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 153 "modules/physics/box2d/wrap_Joint.cpp"
 $2 = $1 ^ 1; //@line 153 "modules/physics/box2d/wrap_Joint.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$2); //@line 153 "modules/physics/box2d/wrap_Joint.cpp"
 return 1; //@line 154 "modules/physics/box2d/wrap_Joint.cpp"
}
function __ZN4love7physics5box2d14luax_pushjointEP9lua_StatePNS1_5JointE($L,$j) {
 $L = $L|0;
 $j = $j|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($j|0)==(0|0); //@line 35 "modules/physics/box2d/wrap_Joint.cpp"
 L1: do {
  if ($0) {
   _lua_pushnil($L); //@line 36 "modules/physics/box2d/wrap_Joint.cpp"
  } else {
   $1 = (__ZNK4love7physics5box2d5Joint7getTypeEv($j)|0); //@line 38 "modules/physics/box2d/wrap_Joint.cpp"
   do {
    switch ($1|0) {
    case 1:  {
     __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,43,$j); //@line 41 "modules/physics/box2d/wrap_Joint.cpp"
     break L1;
     break;
    }
    case 2:  {
     __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,45,$j); //@line 43 "modules/physics/box2d/wrap_Joint.cpp"
     break L1;
     break;
    }
    case 3:  {
     __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,44,$j); //@line 45 "modules/physics/box2d/wrap_Joint.cpp"
     break L1;
     break;
    }
    case 4:  {
     __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,42,$j); //@line 47 "modules/physics/box2d/wrap_Joint.cpp"
     break L1;
     break;
    }
    case 5:  {
     __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,46,$j); //@line 49 "modules/physics/box2d/wrap_Joint.cpp"
     break L1;
     break;
    }
    case 6:  {
     __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,47,$j); //@line 51 "modules/physics/box2d/wrap_Joint.cpp"
     break L1;
     break;
    }
    case 7:  {
     __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,48,$j); //@line 53 "modules/physics/box2d/wrap_Joint.cpp"
     break L1;
     break;
    }
    case 8:  {
     __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,49,$j); //@line 55 "modules/physics/box2d/wrap_Joint.cpp"
     break L1;
     break;
    }
    case 9:  {
     __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,51,$j); //@line 57 "modules/physics/box2d/wrap_Joint.cpp"
     break L1;
     break;
    }
    case 10:  {
     __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,50,$j); //@line 59 "modules/physics/box2d/wrap_Joint.cpp"
     break L1;
     break;
    }
    case 11:  {
     __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,52,$j); //@line 61 "modules/physics/box2d/wrap_Joint.cpp"
     break L1;
     break;
    }
    default: {
     _lua_pushnil($L); //@line 63 "modules/physics/box2d/wrap_Joint.cpp"
     break L1;
    }
    }
   } while(0);
  }
 } while(0);
 return; //@line 65 "modules/physics/box2d/wrap_Joint.cpp"
}
function __ZN4love7physics5box2d15luax_checkjointEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 67 "modules/physics/box2d/wrap_Joint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d5JointEEEPT_P9lua_StateiNS_4TypeE($L,$idx,41)|0); //@line 69 "modules/physics/box2d/wrap_Joint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 70 "modules/physics/box2d/wrap_Joint.cpp"
 if ($1) {
  STACKTOP = sp;return ($0|0); //@line 72 "modules/physics/box2d/wrap_Joint.cpp"
 }
 (_luaL_error($L,358741,$vararg_buffer)|0); //@line 71 "modules/physics/box2d/wrap_Joint.cpp"
 STACKTOP = sp;return ($0|0); //@line 72 "modules/physics/box2d/wrap_Joint.cpp"
}
function _luaopen_joint($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 172 "modules/physics/box2d/wrap_Joint.cpp"
 HEAP32[$vararg_buffer>>2] = 18560; //@line 174 "modules/physics/box2d/wrap_Joint.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 174 "modules/physics/box2d/wrap_Joint.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 174 "modules/physics/box2d/wrap_Joint.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,41,358592,$vararg_buffer)|0); //@line 174 "modules/physics/box2d/wrap_Joint.cpp"
 STACKTOP = sp;return ($0|0); //@line 174 "modules/physics/box2d/wrap_Joint.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d10MotorJointEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d28w_MotorJoint_setLinearOffsetEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 38 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MotorJointEEEPT_P9lua_StateiNS_4TypeE($L,1,52)|0); //@line 32 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MotorJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MotorJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 41 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $3 = $2; //@line 41 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $4 = (+_luaL_checknumber($L,3)); //@line 42 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $5 = $4; //@line 42 "modules/physics/box2d/wrap_MotorJoint.cpp"
 __ZN4love7physics5box2d10MotorJoint15setLinearOffsetEff($0,$3,$5); //@line 43 "modules/physics/box2d/wrap_MotorJoint.cpp"
 STACKTOP = sp;return 0; //@line 44 "modules/physics/box2d/wrap_MotorJoint.cpp"
}
function __ZN4love7physics5box2d28w_MotorJoint_getLinearOffsetEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 47 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MotorJointEEEPT_P9lua_StateiNS_4TypeE($L,1,52)|0); //@line 32 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MotorJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MotorJoint.cpp"
 }
 $2 = (__ZNK4love7physics5box2d10MotorJoint15getLinearOffsetEP9lua_State($0,$L)|0); //@line 50 "modules/physics/box2d/wrap_MotorJoint.cpp"
 STACKTOP = sp;return ($2|0); //@line 50 "modules/physics/box2d/wrap_MotorJoint.cpp"
}
function __ZN4love7physics5box2d29w_MotorJoint_setAngularOffsetEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 53 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MotorJointEEEPT_P9lua_StateiNS_4TypeE($L,1,52)|0); //@line 32 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MotorJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MotorJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 56 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $3 = $2; //@line 56 "modules/physics/box2d/wrap_MotorJoint.cpp"
 __ZN4love7physics5box2d10MotorJoint16setAngularOffsetEf($0,$3); //@line 57 "modules/physics/box2d/wrap_MotorJoint.cpp"
 STACKTOP = sp;return 0; //@line 58 "modules/physics/box2d/wrap_MotorJoint.cpp"
}
function __ZN4love7physics5box2d29w_MotorJoint_getAngularOffsetEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 61 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MotorJointEEEPT_P9lua_StateiNS_4TypeE($L,1,52)|0); //@line 32 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MotorJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MotorJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d10MotorJoint16getAngularOffsetEv($0)); //@line 64 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $3 = $2; //@line 64 "modules/physics/box2d/wrap_MotorJoint.cpp"
 _lua_pushnumber($L,$3); //@line 64 "modules/physics/box2d/wrap_MotorJoint.cpp"
 STACKTOP = sp;return 1; //@line 65 "modules/physics/box2d/wrap_MotorJoint.cpp"
}
function __ZN4love7physics5box2d24w_MotorJoint_setMaxForceEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0.0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 68 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $vararg_buffer = sp; //@line 68 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MotorJointEEEPT_P9lua_StateiNS_4TypeE($L,1,52)|0); //@line 32 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MotorJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MotorJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 71 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $3 = $2; //@line 71 "modules/physics/box2d/wrap_MotorJoint.cpp"
 __THREW__ = 0;
 invoke_vid(3744,($0|0),(+$3)); //@line 72 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  STACKTOP = sp;return 0; //@line 73 "modules/physics/box2d/wrap_MotorJoint.cpp"
 }
 $6 = ___cxa_find_matching_catch_3(8992|0)|0;
 $7 = tempRet0;
 $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
 if (!($9)) {
  $$01$i = $7;$$02$i = $6;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $17; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 73 "modules/physics/box2d/wrap_MotorJoint.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d24w_MotorJoint_getMaxForceEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 76 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MotorJointEEEPT_P9lua_StateiNS_4TypeE($L,1,52)|0); //@line 32 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MotorJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MotorJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d10MotorJoint11getMaxForceEv($0)); //@line 79 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $3 = $2; //@line 79 "modules/physics/box2d/wrap_MotorJoint.cpp"
 _lua_pushnumber($L,$3); //@line 79 "modules/physics/box2d/wrap_MotorJoint.cpp"
 STACKTOP = sp;return 1; //@line 80 "modules/physics/box2d/wrap_MotorJoint.cpp"
}
function __ZN4love7physics5box2d25w_MotorJoint_setMaxTorqueEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0.0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 83 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $vararg_buffer = sp; //@line 83 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MotorJointEEEPT_P9lua_StateiNS_4TypeE($L,1,52)|0); //@line 32 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MotorJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MotorJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 86 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $3 = $2; //@line 86 "modules/physics/box2d/wrap_MotorJoint.cpp"
 __THREW__ = 0;
 invoke_vid(3745,($0|0),(+$3)); //@line 87 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  STACKTOP = sp;return 0; //@line 88 "modules/physics/box2d/wrap_MotorJoint.cpp"
 }
 $6 = ___cxa_find_matching_catch_3(8992|0)|0;
 $7 = tempRet0;
 $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
 if (!($9)) {
  $$01$i = $7;$$02$i = $6;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $17; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 88 "modules/physics/box2d/wrap_MotorJoint.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d25w_MotorJoint_getMaxTorqueEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 91 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MotorJointEEEPT_P9lua_StateiNS_4TypeE($L,1,52)|0); //@line 32 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MotorJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MotorJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d10MotorJoint12getMaxTorqueEv($0)); //@line 94 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $3 = $2; //@line 94 "modules/physics/box2d/wrap_MotorJoint.cpp"
 _lua_pushnumber($L,$3); //@line 94 "modules/physics/box2d/wrap_MotorJoint.cpp"
 STACKTOP = sp;return 1; //@line 95 "modules/physics/box2d/wrap_MotorJoint.cpp"
}
function __ZN4love7physics5box2d32w_MotorJoint_setCorrectionFactorEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0.0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 98 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $vararg_buffer = sp; //@line 98 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MotorJointEEEPT_P9lua_StateiNS_4TypeE($L,1,52)|0); //@line 32 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MotorJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MotorJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 101 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $3 = $2; //@line 101 "modules/physics/box2d/wrap_MotorJoint.cpp"
 __THREW__ = 0;
 invoke_vid(3746,($0|0),(+$3)); //@line 102 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  STACKTOP = sp;return 0; //@line 103 "modules/physics/box2d/wrap_MotorJoint.cpp"
 }
 $6 = ___cxa_find_matching_catch_3(8992|0)|0;
 $7 = tempRet0;
 $8 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $9 = ($7|0)==($8|0); //@line 535 "./common/runtime.h"
 if (!($9)) {
  $$01$i = $7;$$02$i = $6;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $10 = (___cxa_begin_catch(($6|0))|0); //@line 535 "./common/runtime.h"
 $11 = HEAP32[$10>>2]|0; //@line 539 "./common/runtime.h"
 $12 = ((($11)) + 8|0); //@line 539 "./common/runtime.h"
 $13 = HEAP32[$12>>2]|0; //@line 539 "./common/runtime.h"
 $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($14|0)); //@line 539 "./common/runtime.h"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $17 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer1>>2] = $17; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer1)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 0; //@line 103 "modules/physics/box2d/wrap_MotorJoint.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $19;$$02$i = $18;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love7physics5box2d32w_MotorJoint_getCorrectionFactorEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 106 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MotorJointEEEPT_P9lua_StateiNS_4TypeE($L,1,52)|0); //@line 32 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MotorJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MotorJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d10MotorJoint19getCorrectionFactorEv($0)); //@line 109 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $3 = $2; //@line 109 "modules/physics/box2d/wrap_MotorJoint.cpp"
 _lua_pushnumber($L,$3); //@line 109 "modules/physics/box2d/wrap_MotorJoint.cpp"
 STACKTOP = sp;return 1; //@line 110 "modules/physics/box2d/wrap_MotorJoint.cpp"
}
function _luaopen_motorjoint($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 128 "modules/physics/box2d/wrap_MotorJoint.cpp"
 HEAP32[$vararg_buffer>>2] = 18560; //@line 130 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 130 "modules/physics/box2d/wrap_MotorJoint.cpp"
 HEAP32[$vararg_ptr1>>2] = 18648; //@line 130 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 130 "modules/physics/box2d/wrap_MotorJoint.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 130 "modules/physics/box2d/wrap_MotorJoint.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,52,358598,$vararg_buffer)|0); //@line 130 "modules/physics/box2d/wrap_MotorJoint.cpp"
 STACKTOP = sp;return ($0|0); //@line 130 "modules/physics/box2d/wrap_MotorJoint.cpp"
}
function __ZN4love14luax_checktypeINS_7physics5box2d10MouseJointEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love7physics5box2d22w_MouseJoint_setTargetEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 38 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MouseJointEEEPT_P9lua_StateiNS_4TypeE($L,1,42)|0); //@line 32 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MouseJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MouseJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 41 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $3 = $2; //@line 41 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $4 = (+_luaL_checknumber($L,3)); //@line 42 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $5 = $4; //@line 42 "modules/physics/box2d/wrap_MouseJoint.cpp"
 __ZN4love7physics5box2d10MouseJoint9setTargetEff($0,$3,$5); //@line 43 "modules/physics/box2d/wrap_MouseJoint.cpp"
 STACKTOP = sp;return 0; //@line 44 "modules/physics/box2d/wrap_MouseJoint.cpp"
}
function __ZN4love7physics5box2d22w_MouseJoint_getTargetEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 47 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MouseJointEEEPT_P9lua_StateiNS_4TypeE($L,1,42)|0); //@line 32 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MouseJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MouseJoint.cpp"
 }
 _lua_remove($L,1); //@line 50 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $2 = (__ZN4love7physics5box2d10MouseJoint9getTargetEP9lua_State($0,$L)|0); //@line 51 "modules/physics/box2d/wrap_MouseJoint.cpp"
 STACKTOP = sp;return ($2|0); //@line 51 "modules/physics/box2d/wrap_MouseJoint.cpp"
}
function __ZN4love7physics5box2d24w_MouseJoint_setMaxForceEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 54 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MouseJointEEEPT_P9lua_StateiNS_4TypeE($L,1,42)|0); //@line 32 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MouseJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MouseJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 57 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $3 = $2; //@line 57 "modules/physics/box2d/wrap_MouseJoint.cpp"
 __ZN4love7physics5box2d10MouseJoint11setMaxForceEf($0,$3); //@line 58 "modules/physics/box2d/wrap_MouseJoint.cpp"
 STACKTOP = sp;return 0; //@line 59 "modules/physics/box2d/wrap_MouseJoint.cpp"
}
function __ZN4love7physics5box2d24w_MouseJoint_getMaxForceEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 62 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MouseJointEEEPT_P9lua_StateiNS_4TypeE($L,1,42)|0); //@line 32 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MouseJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MouseJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d10MouseJoint11getMaxForceEv($0)); //@line 65 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $3 = $2; //@line 65 "modules/physics/box2d/wrap_MouseJoint.cpp"
 _lua_pushnumber($L,$3); //@line 65 "modules/physics/box2d/wrap_MouseJoint.cpp"
 STACKTOP = sp;return 1; //@line 66 "modules/physics/box2d/wrap_MouseJoint.cpp"
}
function __ZN4love7physics5box2d25w_MouseJoint_setFrequencyEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 69 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MouseJointEEEPT_P9lua_StateiNS_4TypeE($L,1,42)|0); //@line 32 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MouseJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MouseJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 72 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $3 = $2; //@line 72 "modules/physics/box2d/wrap_MouseJoint.cpp"
 __ZN4love7physics5box2d10MouseJoint12setFrequencyEf($0,$3); //@line 73 "modules/physics/box2d/wrap_MouseJoint.cpp"
 STACKTOP = sp;return 0; //@line 74 "modules/physics/box2d/wrap_MouseJoint.cpp"
}
function __ZN4love7physics5box2d25w_MouseJoint_getFrequencyEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 77 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MouseJointEEEPT_P9lua_StateiNS_4TypeE($L,1,42)|0); //@line 32 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MouseJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MouseJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d10MouseJoint12getFrequencyEv($0)); //@line 80 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $3 = $2; //@line 80 "modules/physics/box2d/wrap_MouseJoint.cpp"
 _lua_pushnumber($L,$3); //@line 80 "modules/physics/box2d/wrap_MouseJoint.cpp"
 STACKTOP = sp;return 1; //@line 81 "modules/physics/box2d/wrap_MouseJoint.cpp"
}
function __ZN4love7physics5box2d28w_MouseJoint_setDampingRatioEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 84 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MouseJointEEEPT_P9lua_StateiNS_4TypeE($L,1,42)|0); //@line 32 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MouseJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MouseJoint.cpp"
 }
 $2 = (+_luaL_checknumber($L,2)); //@line 87 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $3 = $2; //@line 87 "modules/physics/box2d/wrap_MouseJoint.cpp"
 __ZN4love7physics5box2d10MouseJoint15setDampingRatioEf($0,$3); //@line 88 "modules/physics/box2d/wrap_MouseJoint.cpp"
 STACKTOP = sp;return 0; //@line 89 "modules/physics/box2d/wrap_MouseJoint.cpp"
}
function __ZN4love7physics5box2d28w_MouseJoint_getDampingRatioEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 92 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $0 = (__ZN4love14luax_checktypeINS_7physics5box2d10MouseJointEEEPT_P9lua_StateiNS_4TypeE($L,1,42)|0); //@line 32 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $1 = (__ZNK4love7physics5box2d5Joint7isValidEv($0)|0); //@line 33 "modules/physics/box2d/wrap_MouseJoint.cpp"
 if (!($1)) {
  (_luaL_error($L,358741,$vararg_buffer)|0); //@line 34 "modules/physics/box2d/wrap_MouseJoint.cpp"
 }
 $2 = (+__ZNK4love7physics5box2d10MouseJoint15getDampingRatioEv($0)); //@line 95 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $3 = $2; //@line 95 "modules/physics/box2d/wrap_MouseJoint.cpp"
 _lua_pushnumber($L,$3); //@line 95 "modules/physics/box2d/wrap_MouseJoint.cpp"
 STACKTOP = sp;return 1; //@line 96 "modules/physics/box2d/wrap_MouseJoint.cpp"
}
function _luaopen_mousejoint($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 112 "modules/physics/box2d/wrap_MouseJoint.cpp"
 HEAP32[$vararg_buffer>>2] = 18560; //@line 114 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 114 "modules/physics/box2d/wrap_MouseJoint.cpp"
 HEAP32[$vararg_ptr1>>2] = 18736; //@line 114 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 114 "modules/physics/box2d/wrap_MouseJoint.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 114 "modules/physics/box2d/wrap_MouseJoint.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,42,358773,$vararg_buffer)|0); //@line 114 "modules/physics/box2d/wrap_MouseJoint.cpp"
 STACKTOP = sp;return ($0|0); //@line 114 "modules/physics/box2d/wrap_MouseJoint.cpp"
}
function __ZN4love7physics5JointD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 30 "modules/physics/Joint.cpp"
}
function __GLOBAL__sub_I_Joint_cpp() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love9StringMapINS_7physics5Joint4TypeELj12EEC2EPKNS4_5EntryEj(630984,18808,88); //@line 57 "modules/physics/Joint.cpp"
 return;
}
function __ZN4love9StringMapINS_7physics5Joint4TypeELj12EEC2EPKNS4_5EntryEj($this,$entries,$num) {
 $this = $this|0;
 $entries = $entries|0;
 $num = $num|0;
 var $$02$i$i = 0, $$lcssa26 = 0, $$lcssa28 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$08$i = 0, $i1$05 = 0, $scevgep = 0;
 var $vararg_buffer = 0, $vararg_ptr1 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 8|0); //@line 155 "./common/StringMap.h"
 HEAP8[$0>>0] = 0; //@line 155 "./common/StringMap.h"
 $1 = ((($this)) + 20|0); //@line 155 "./common/StringMap.h"
 HEAP8[$1>>0] = 0; //@line 155 "./common/StringMap.h"
 $2 = ((($this)) + 32|0); //@line 155 "./common/StringMap.h"
 HEAP8[$2>>0] = 0; //@line 155 "./common/StringMap.h"
 $3 = ((($this)) + 44|0); //@line 155 "./common/StringMap.h"
 HEAP8[$3>>0] = 0; //@line 155 "./common/StringMap.h"
 $4 = ((($this)) + 56|0); //@line 155 "./common/StringMap.h"
 HEAP8[$4>>0] = 0; //@line 155 "./common/StringMap.h"
 $5 = ((($this)) + 68|0); //@line 155 "./common/StringMap.h"
 HEAP8[$5>>0] = 0; //@line 155 "./common/StringMap.h"
 $6 = ((($this)) + 80|0); //@line 155 "./common/StringMap.h"
 HEAP8[$6>>0] = 0; //@line 155 "./common/StringMap.h"
 $7 = ((($this)) + 92|0); //@line 155 "./common/StringMap.h"
 HEAP8[$7>>0] = 0; //@line 155 "./common/StringMap.h"
 $8 = ((($this)) + 104|0); //@line 155 "./common/StringMap.h"
 HEAP8[$8>>0] = 0; //@line 155 "./common/StringMap.h"
 $9 = ((($this)) + 116|0); //@line 155 "./common/StringMap.h"
 HEAP8[$9>>0] = 0; //@line 155 "./common/StringMap.h"
 $10 = ((($this)) + 128|0); //@line 155 "./common/StringMap.h"
 HEAP8[$10>>0] = 0; //@line 155 "./common/StringMap.h"
 $11 = ((($this)) + 140|0); //@line 155 "./common/StringMap.h"
 HEAP8[$11>>0] = 0; //@line 155 "./common/StringMap.h"
 $12 = ((($this)) + 152|0); //@line 155 "./common/StringMap.h"
 HEAP8[$12>>0] = 0; //@line 155 "./common/StringMap.h"
 $13 = ((($this)) + 164|0); //@line 155 "./common/StringMap.h"
 HEAP8[$13>>0] = 0; //@line 155 "./common/StringMap.h"
 $14 = ((($this)) + 176|0); //@line 155 "./common/StringMap.h"
 HEAP8[$14>>0] = 0; //@line 155 "./common/StringMap.h"
 $15 = ((($this)) + 188|0); //@line 155 "./common/StringMap.h"
 HEAP8[$15>>0] = 0; //@line 155 "./common/StringMap.h"
 $16 = ((($this)) + 200|0); //@line 155 "./common/StringMap.h"
 HEAP8[$16>>0] = 0; //@line 155 "./common/StringMap.h"
 $17 = ((($this)) + 212|0); //@line 155 "./common/StringMap.h"
 HEAP8[$17>>0] = 0; //@line 155 "./common/StringMap.h"
 $18 = ((($this)) + 224|0); //@line 155 "./common/StringMap.h"
 HEAP8[$18>>0] = 0; //@line 155 "./common/StringMap.h"
 $19 = ((($this)) + 236|0); //@line 155 "./common/StringMap.h"
 HEAP8[$19>>0] = 0; //@line 155 "./common/StringMap.h"
 $20 = ((($this)) + 248|0); //@line 155 "./common/StringMap.h"
 HEAP8[$20>>0] = 0; //@line 155 "./common/StringMap.h"
 $21 = ((($this)) + 260|0); //@line 155 "./common/StringMap.h"
 HEAP8[$21>>0] = 0; //@line 155 "./common/StringMap.h"
 $22 = ((($this)) + 272|0); //@line 155 "./common/StringMap.h"
 HEAP8[$22>>0] = 0; //@line 155 "./common/StringMap.h"
 $23 = ((($this)) + 284|0); //@line 155 "./common/StringMap.h"
 HEAP8[$23>>0] = 0; //@line 155 "./common/StringMap.h"
 $scevgep = ((($this)) + 288|0); //@line 44 "./common/StringMap.h"
 dest=$scevgep; stop=dest+48|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 44 "./common/StringMap.h"
 $24 = $num >>> 3; //@line 46 "./common/StringMap.h"
 $25 = ($24|0)==(0); //@line 48 "./common/StringMap.h"
 if ($25) {
  STACKTOP = sp;return; //@line 50 "./common/StringMap.h"
 } else {
  $i1$05 = 0;
 }
 while(1) {
  $26 = (($entries) + ($i1$05<<3)|0); //@line 49 "./common/StringMap.h"
  $27 = HEAP32[$26>>2]|0; //@line 49 "./common/StringMap.h"
  $28 = (((($entries) + ($i1$05<<3)|0)) + 4|0); //@line 49 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 49 "./common/StringMap.h"
  $30 = HEAP8[$27>>0]|0; //@line 142 "./common/StringMap.h"
  $31 = ($30<<24>>24)==(0); //@line 142 "./common/StringMap.h"
  if ($31) {
   $hash$0$lcssa$i$i = 5381;
  } else {
   $$02$i$i = $27;$33 = $30;$hash$01$i$i = 5381;
   while(1) {
    $32 = $33 << 24 >> 24; //@line 142 "./common/StringMap.h"
    $34 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
    $35 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
    $36 = (($35) + ($32))|0; //@line 143 "./common/StringMap.h"
    $37 = HEAP8[$34>>0]|0; //@line 142 "./common/StringMap.h"
    $38 = ($37<<24>>24)==(0); //@line 142 "./common/StringMap.h"
    if ($38) {
     $hash$0$lcssa$i$i = $36;
     break;
    } else {
     $$02$i$i = $34;$33 = $37;$hash$01$i$i = $36;
    }
   }
  }
  $i$08$i = 0;
  while(1) {
   $41 = (($i$08$i) + ($hash$0$lcssa$i$i))|0; //@line 112 "./common/StringMap.h"
   $42 = (($41>>>0) % 24)&-1; //@line 112 "./common/StringMap.h"
   $43 = (((($this) + (($42*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $44 = HEAP8[$43>>0]|0; //@line 114 "./common/StringMap.h"
   $45 = ($44<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   $40 = (($i$08$i) + 1)|0; //@line 110 "./common/StringMap.h"
   if ($45) {
    $$lcssa26 = $42;$$lcssa28 = $43;
    label = 8;
    break;
   }
   $39 = ($40>>>0)<(24); //@line 110 "./common/StringMap.h"
   if ($39) {
    $i$08$i = $40;
   } else {
    break;
   }
  }
  if ((label|0) == 8) {
   label = 0;
   HEAP8[$$lcssa28>>0] = 1; //@line 117 "./common/StringMap.h"
   $46 = (($this) + (($$lcssa26*12)|0)|0); //@line 118 "./common/StringMap.h"
   HEAP32[$46>>2] = $27; //@line 118 "./common/StringMap.h"
   $47 = (((($this) + (($$lcssa26*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
   HEAP32[$47>>2] = $29; //@line 119 "./common/StringMap.h"
  }
  $48 = ($29>>>0)>(11); //@line 126 "./common/StringMap.h"
  if ($48) {
   HEAP32[$vararg_buffer>>2] = $27; //@line 128 "./common/StringMap.h"
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
   HEAP32[$vararg_ptr1>>2] = $29; //@line 128 "./common/StringMap.h"
   (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  } else {
   $49 = (((($this)) + 288|0) + ($29<<2)|0); //@line 132 "./common/StringMap.h"
   HEAP32[$49>>2] = $27; //@line 132 "./common/StringMap.h"
  }
  $50 = (($i1$05) + 1)|0; //@line 48 "./common/StringMap.h"
  $exitcond = ($50|0)==($24|0); //@line 48 "./common/StringMap.h"
  if ($exitcond) {
   break;
  } else {
   $i1$05 = $50;
  }
 }
 STACKTOP = sp;return; //@line 50 "./common/StringMap.h"
}
function __ZN4love7physics5Joint11getConstantENS1_4TypeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(11); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (631272 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 39 "modules/physics/Joint.cpp"
}
function __ZN4love5sound7lullaby5SoundD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (18904); //@line 53 "modules/sound/lullaby/Sound.cpp"
 __THREW__ = 0;
 invoke_v(3747); //@line 55 "modules/sound/lullaby/Sound.cpp"
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 if ($1) {
  $2 = ___cxa_find_matching_catch_3(0|0)|0;
  $3 = tempRet0;
  __ZN4love5sound5SoundD2Ev($this); //@line 57 "modules/sound/lullaby/Sound.cpp"
  ___clang_call_terminate($2); //@line 57 "modules/sound/lullaby/Sound.cpp"
  // unreachable; //@line 57 "modules/sound/lullaby/Sound.cpp"
 } else {
  __ZN4love5sound5SoundD2Ev($this); //@line 57 "modules/sound/lullaby/Sound.cpp"
  return; //@line 57 "modules/sound/lullaby/Sound.cpp"
 }
}
function __ZN4love5sound7lullaby5SoundD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (18904); //@line 53 "modules/sound/lullaby/Sound.cpp"
 __THREW__ = 0;
 invoke_v(3747); //@line 55 "modules/sound/lullaby/Sound.cpp"
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 if ($1) {
  $2 = ___cxa_find_matching_catch_3(0|0)|0;
  $3 = tempRet0;
  __ZN4love5sound5SoundD2Ev($this); //@line 57 "modules/sound/lullaby/Sound.cpp"
  ___clang_call_terminate($2); //@line 57 "modules/sound/lullaby/Sound.cpp"
  // unreachable; //@line 57 "modules/sound/lullaby/Sound.cpp"
 } else {
  __ZN4love5sound5SoundD2Ev($this); //@line 57 "modules/sound/lullaby/Sound.cpp"
  __ZdlPv($this); //@line 53 "modules/sound/lullaby/Sound.cpp"
  return; //@line 57 "modules/sound/lullaby/Sound.cpp"
 }
}
function __ZNK4love5sound5Sound13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 11; //@line 48 "./modules/sound/Sound.h"
}
function __ZNK4love5sound7lullaby5Sound7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (359013|0); //@line 61 "modules/sound/lullaby/Sound.cpp"
}
function __ZN4love5sound7lullaby5Sound10newDecoderEPNS_10filesystem8FileDataEi($this,$data,$bufferSize) {
 $this = $this|0;
 $data = $data|0;
 $bufferSize = $bufferSize|0;
 var $$0 = 0, $$01 = 0, $$sroa$05$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $decoder$0 = 0, $ext = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ext = sp;
 $0 = (__ZNK4love10filesystem8FileData12getExtensionEv($data)|0); //@line 66 "modules/sound/lullaby/Sound.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_($ext,$0); //@line 66 "modules/sound/lullaby/Sound.cpp"
 $1 = HEAP8[$ext>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = $1 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ($2<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ((($ext)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = HEAP32[$4>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = ((($ext)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = $3 ? $6 : $5; //@line 1755 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = ((($ext)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = HEAP32[$8>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $1&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $10 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = $3 ? $11 : $9; //@line 1755 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = (($7) + ($12)|0); //@line 1399 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = ($12|0)==(0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 L1: do {
  if ($14) {
   label = 4;
  } else {
   $$sroa$05$0 = $7;$21 = $7;
   while(1) {
    $15 = HEAP8[$$sroa$05$0>>0]|0; //@line 1954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $16 = $15 << 24 >> 24; //@line 1954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    __THREW__ = 0;
    $17 = (invoke_ii(3748,($16|0))|0); //@line 1954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $18 = __THREW__; __THREW__ = 0;
    $19 = $18&1;
    if ($19) {
     break;
    }
    $20 = $17&255; //@line 1954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    HEAP8[$21>>0] = $20; //@line 1954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $22 = ((($$sroa$05$0)) + 1|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $23 = ((($21)) + 1|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $24 = ($22|0)==($13|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($24) {
     label = 4;
     break L1;
    } else {
     $$sroa$05$0 = $22;$21 = $23;
    }
   }
   $33 = ___cxa_find_matching_catch_2()|0;
   $34 = tempRet0;
   $lpad$phi$index = $33;$lpad$phi$index2 = $34;
  }
 } while(0);
 if ((label|0) == 4) {
  __THREW__ = 0;
  $25 = (invoke_ii(3749,($ext|0))|0); //@line 75 "modules/sound/lullaby/Sound.cpp"
  $26 = __THREW__; __THREW__ = 0;
  $27 = $26&1;
  do {
   if (!($27)) {
    if ($25) {
     __THREW__ = 0;
     $28 = (invoke_ii(3348,112)|0); //@line 76 "modules/sound/lullaby/Sound.cpp"
     $29 = __THREW__; __THREW__ = 0;
     $30 = $29&1;
     if ($30) {
      break;
     }
     __THREW__ = 0;
     invoke_viiii(3750,($28|0),($data|0),($ext|0),($bufferSize|0)); //@line 76 "modules/sound/lullaby/Sound.cpp"
     $31 = __THREW__; __THREW__ = 0;
     $32 = $31&1;
     if ($32) {
      $37 = ___cxa_find_matching_catch_2()|0;
      $38 = tempRet0;
      __ZdlPv($28); //@line 76 "modules/sound/lullaby/Sound.cpp"
      $$0 = $37;$$01 = $38;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($ext); //@line 100 "modules/sound/lullaby/Sound.cpp"
      ___resumeException($$0|0);
      // unreachable;
     } else {
      $decoder$0 = $28;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($ext); //@line 100 "modules/sound/lullaby/Sound.cpp"
      STACKTOP = sp;return ($decoder$0|0); //@line 100 "modules/sound/lullaby/Sound.cpp"
     }
    }
    __THREW__ = 0;
    $39 = (invoke_ii(3751,($ext|0))|0); //@line 79 "modules/sound/lullaby/Sound.cpp"
    $40 = __THREW__; __THREW__ = 0;
    $41 = $40&1;
    if (!($41)) {
     if ($39) {
      __THREW__ = 0;
      $42 = (invoke_ii(3348,72)|0); //@line 80 "modules/sound/lullaby/Sound.cpp"
      $43 = __THREW__; __THREW__ = 0;
      $44 = $43&1;
      if ($44) {
       break;
      }
      __THREW__ = 0;
      invoke_viiii(3752,($42|0),($data|0),($ext|0),($bufferSize|0)); //@line 80 "modules/sound/lullaby/Sound.cpp"
      $45 = __THREW__; __THREW__ = 0;
      $46 = $45&1;
      if ($46) {
       $47 = ___cxa_find_matching_catch_2()|0;
       $48 = tempRet0;
       __ZdlPv($42); //@line 80 "modules/sound/lullaby/Sound.cpp"
       $$0 = $47;$$01 = $48;
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($ext); //@line 100 "modules/sound/lullaby/Sound.cpp"
       ___resumeException($$0|0);
       // unreachable;
      } else {
       $decoder$0 = $42;
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($ext); //@line 100 "modules/sound/lullaby/Sound.cpp"
       STACKTOP = sp;return ($decoder$0|0); //@line 100 "modules/sound/lullaby/Sound.cpp"
      }
     }
     __THREW__ = 0;
     $49 = (invoke_ii(3753,($ext|0))|0); //@line 82 "modules/sound/lullaby/Sound.cpp"
     $50 = __THREW__; __THREW__ = 0;
     $51 = $50&1;
     if (!($51)) {
      if ($49) {
       __THREW__ = 0;
       $52 = (invoke_ii(3348,816)|0); //@line 83 "modules/sound/lullaby/Sound.cpp"
       $53 = __THREW__; __THREW__ = 0;
       $54 = $53&1;
       if ($54) {
        break;
       }
       __THREW__ = 0;
       invoke_viiii(3754,($52|0),($data|0),($ext|0),($bufferSize|0)); //@line 83 "modules/sound/lullaby/Sound.cpp"
       $55 = __THREW__; __THREW__ = 0;
       $56 = $55&1;
       if ($56) {
        $57 = ___cxa_find_matching_catch_2()|0;
        $58 = tempRet0;
        __ZdlPv($52); //@line 83 "modules/sound/lullaby/Sound.cpp"
        $$0 = $57;$$01 = $58;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($ext); //@line 100 "modules/sound/lullaby/Sound.cpp"
        ___resumeException($$0|0);
        // unreachable;
       } else {
        $decoder$0 = $52;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($ext); //@line 100 "modules/sound/lullaby/Sound.cpp"
        STACKTOP = sp;return ($decoder$0|0); //@line 100 "modules/sound/lullaby/Sound.cpp"
       }
      }
      __THREW__ = 0;
      $59 = (invoke_ii(3755,($ext|0))|0); //@line 92 "modules/sound/lullaby/Sound.cpp"
      $60 = __THREW__; __THREW__ = 0;
      $61 = $60&1;
      if (!($61)) {
       if (!($59)) {
        $decoder$0 = 0;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($ext); //@line 100 "modules/sound/lullaby/Sound.cpp"
        STACKTOP = sp;return ($decoder$0|0); //@line 100 "modules/sound/lullaby/Sound.cpp"
       }
       __THREW__ = 0;
       $62 = (invoke_ii(3348,80)|0); //@line 93 "modules/sound/lullaby/Sound.cpp"
       $63 = __THREW__; __THREW__ = 0;
       $64 = $63&1;
       if (!($64)) {
        __THREW__ = 0;
        invoke_viiii(3756,($62|0),($data|0),($ext|0),($bufferSize|0)); //@line 93 "modules/sound/lullaby/Sound.cpp"
        $65 = __THREW__; __THREW__ = 0;
        $66 = $65&1;
        if ($66) {
         $67 = ___cxa_find_matching_catch_2()|0;
         $68 = tempRet0;
         __ZdlPv($62); //@line 93 "modules/sound/lullaby/Sound.cpp"
         $$0 = $67;$$01 = $68;
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($ext); //@line 100 "modules/sound/lullaby/Sound.cpp"
         ___resumeException($$0|0);
         // unreachable;
        } else {
         $decoder$0 = $62;
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($ext); //@line 100 "modules/sound/lullaby/Sound.cpp"
         STACKTOP = sp;return ($decoder$0|0); //@line 100 "modules/sound/lullaby/Sound.cpp"
        }
       }
      }
     }
    }
   }
  } while(0);
  $35 = ___cxa_find_matching_catch_2()|0;
  $36 = tempRet0;
  $lpad$phi$index = $35;$lpad$phi$index2 = $36;
 }
 $$0 = $lpad$phi$index;$$01 = $lpad$phi$index2;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($ext); //@line 100 "modules/sound/lullaby/Sound.cpp"
 ___resumeException($$0|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love5sound7lullaby5SoundC2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./common/Module.h"
 HEAP32[$this>>2] = (18904); //@line 49 "modules/sound/lullaby/Sound.cpp"
 return; //@line 50 "modules/sound/lullaby/Sound.cpp"
}
function __ZN4love5sound7lullaby13VorbisDecoderD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (18940); //@line 163 "modules/sound/lullaby/VorbisDecoder.cpp"
 $0 = ((($this)) + 72|0); //@line 164 "modules/sound/lullaby/VorbisDecoder.cpp"
 __THREW__ = 0;
 (invoke_ii(3757,($0|0))|0); //@line 164 "modules/sound/lullaby/VorbisDecoder.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_3(0|0)|0;
  $4 = tempRet0;
  __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 165 "modules/sound/lullaby/VorbisDecoder.cpp"
  ___clang_call_terminate($3); //@line 165 "modules/sound/lullaby/VorbisDecoder.cpp"
  // unreachable; //@line 165 "modules/sound/lullaby/VorbisDecoder.cpp"
 } else {
  __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 165 "modules/sound/lullaby/VorbisDecoder.cpp"
  return; //@line 165 "modules/sound/lullaby/VorbisDecoder.cpp"
 }
}
function __ZN4love5sound7lullaby13VorbisDecoderD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (18940); //@line 163 "modules/sound/lullaby/VorbisDecoder.cpp"
 $0 = ((($this)) + 72|0); //@line 164 "modules/sound/lullaby/VorbisDecoder.cpp"
 __THREW__ = 0;
 (invoke_ii(3757,($0|0))|0); //@line 164 "modules/sound/lullaby/VorbisDecoder.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_3(0|0)|0;
  $4 = tempRet0;
  __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 165 "modules/sound/lullaby/VorbisDecoder.cpp"
  ___clang_call_terminate($3); //@line 165 "modules/sound/lullaby/VorbisDecoder.cpp"
  // unreachable; //@line 165 "modules/sound/lullaby/VorbisDecoder.cpp"
 } else {
  __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 165 "modules/sound/lullaby/VorbisDecoder.cpp"
  __ZdlPv($this); //@line 163 "modules/sound/lullaby/VorbisDecoder.cpp"
  return; //@line 165 "modules/sound/lullaby/VorbisDecoder.cpp"
 }
}
function __ZN4love5sound7lullaby13VorbisDecoder5cloneEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(816)|0); //@line 185 "modules/sound/lullaby/VorbisDecoder.cpp"
 $1 = ((($this)) + 8|0); //@line 139 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 139 "./common/Object.h"
 $3 = ((($this)) + 12|0); //@line 185 "modules/sound/lullaby/VorbisDecoder.cpp"
 $4 = ((($this)) + 24|0); //@line 185 "modules/sound/lullaby/VorbisDecoder.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 185 "modules/sound/lullaby/VorbisDecoder.cpp"
 __THREW__ = 0;
 invoke_viiii(3754,($0|0),($2|0),($3|0),($5|0)); //@line 185 "modules/sound/lullaby/VorbisDecoder.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $8 = ___cxa_find_matching_catch_2()|0;
  $9 = tempRet0;
  __ZdlPv($0); //@line 185 "modules/sound/lullaby/VorbisDecoder.cpp"
  ___resumeException($8|0);
  // unreachable;
 } else {
  return ($0|0); //@line 185 "modules/sound/lullaby/VorbisDecoder.cpp"
 }
 return (0)|0;
}
function __ZN4love5sound7lullaby13VorbisDecoderC2EPNS_4DataERKNSt3__112basic_stringIcNS5_11char_traitsIcEENS5_9allocatorIcEEEEi($this,$data,$ext,$bufferSize) {
 $this = $this|0;
 $data = $data|0;
 $ext = $ext|0;
 $bufferSize = $bufferSize|0;
 var $$0 = 0, $$01 = 0, $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 8|0;
 $vararg_buffer = sp;
 __ZN4love5sound7lullaby7DecoderC2EPNS_4DataERKNSt3__112basic_stringIcNS5_11char_traitsIcEENS5_9allocatorIcEEEEi($this,$data,$ext,$bufferSize); //@line 132 "modules/sound/lullaby/VorbisDecoder.cpp"
 HEAP32[$this>>2] = (18940); //@line 134 "modules/sound/lullaby/VorbisDecoder.cpp"
 $0 = ((($this)) + 808|0); //@line 133 "modules/sound/lullaby/VorbisDecoder.cpp"
 HEAPF64[$0>>3] = -2.0; //@line 133 "modules/sound/lullaby/VorbisDecoder.cpp"
 $1 = ((($this)) + 52|0); //@line 136 "modules/sound/lullaby/VorbisDecoder.cpp"
 $2 = ((($this)) + 60|0); //@line 136 "modules/sound/lullaby/VorbisDecoder.cpp"
 HEAP32[$2>>2] = 3758; //@line 136 "modules/sound/lullaby/VorbisDecoder.cpp"
 $3 = ((($this)) + 56|0); //@line 137 "modules/sound/lullaby/VorbisDecoder.cpp"
 HEAP32[$3>>2] = 3759; //@line 137 "modules/sound/lullaby/VorbisDecoder.cpp"
 HEAP32[$1>>2] = 3760; //@line 138 "modules/sound/lullaby/VorbisDecoder.cpp"
 $4 = ((($this)) + 64|0); //@line 139 "modules/sound/lullaby/VorbisDecoder.cpp"
 HEAP32[$4>>2] = 3761; //@line 139 "modules/sound/lullaby/VorbisDecoder.cpp"
 $5 = ((($this)) + 800|0); //@line 145 "modules/sound/lullaby/VorbisDecoder.cpp"
 HEAP32[$5>>2] = 0; //@line 145 "modules/sound/lullaby/VorbisDecoder.cpp"
 $6 = HEAP32[$data>>2]|0; //@line 149 "modules/sound/lullaby/VorbisDecoder.cpp"
 $7 = ((($6)) + 16|0); //@line 149 "modules/sound/lullaby/VorbisDecoder.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 149 "modules/sound/lullaby/VorbisDecoder.cpp"
 __THREW__ = 0;
 $9 = (invoke_ii($8|0,($data|0))|0); //@line 149 "modules/sound/lullaby/VorbisDecoder.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 do {
  if (!($11)) {
   $12 = ((($this)) + 40|0); //@line 149 "modules/sound/lullaby/VorbisDecoder.cpp"
   HEAP32[$12>>2] = $9; //@line 149 "modules/sound/lullaby/VorbisDecoder.cpp"
   $13 = HEAP32[$data>>2]|0; //@line 150 "modules/sound/lullaby/VorbisDecoder.cpp"
   $14 = ((($13)) + 20|0); //@line 150 "modules/sound/lullaby/VorbisDecoder.cpp"
   $15 = HEAP32[$14>>2]|0; //@line 150 "modules/sound/lullaby/VorbisDecoder.cpp"
   __THREW__ = 0;
   $16 = (invoke_ii($15|0,($data|0))|0); //@line 150 "modules/sound/lullaby/VorbisDecoder.cpp"
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if (!($18)) {
    $19 = ((($this)) + 44|0); //@line 150 "modules/sound/lullaby/VorbisDecoder.cpp"
    HEAP32[$19>>2] = $16; //@line 150 "modules/sound/lullaby/VorbisDecoder.cpp"
    $20 = ((($this)) + 48|0); //@line 151 "modules/sound/lullaby/VorbisDecoder.cpp"
    HEAP32[$20>>2] = 0; //@line 151 "modules/sound/lullaby/VorbisDecoder.cpp"
    $21 = ((($this)) + 72|0); //@line 154 "modules/sound/lullaby/VorbisDecoder.cpp"
    __THREW__ = 0;
    ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;HEAP32[$$byval_copy+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$$byval_copy+8>>2]=HEAP32[$1+8>>2]|0;HEAP32[$$byval_copy+12>>2]=HEAP32[$1+12>>2]|0; //@line 154 "modules/sound/lullaby/VorbisDecoder.cpp"
    $22 = (invoke_iiiiii(3762,($12|0),($21|0),(0|0),0,($$byval_copy|0))|0); //@line 154 "modules/sound/lullaby/VorbisDecoder.cpp"
    $23 = __THREW__; __THREW__ = 0;
    $24 = $23&1;
    if (!($24)) {
     $25 = ($22|0)<(0); //@line 154 "modules/sound/lullaby/VorbisDecoder.cpp"
     if (!($25)) {
      __THREW__ = 0;
      $34 = (invoke_iii(3763,($21|0),-1)|0); //@line 158 "modules/sound/lullaby/VorbisDecoder.cpp"
      $35 = __THREW__; __THREW__ = 0;
      $36 = $35&1;
      if ($36) {
       break;
      }
      $37 = ((($this)) + 792|0); //@line 158 "modules/sound/lullaby/VorbisDecoder.cpp"
      HEAP32[$37>>2] = $34; //@line 158 "modules/sound/lullaby/VorbisDecoder.cpp"
      __THREW__ = 0;
      $38 = (invoke_iii(3764,($21|0),-1)|0); //@line 159 "modules/sound/lullaby/VorbisDecoder.cpp"
      $39 = __THREW__; __THREW__ = 0;
      $40 = $39&1;
      if ($40) {
       break;
      }
      $41 = ((($this)) + 796|0); //@line 159 "modules/sound/lullaby/VorbisDecoder.cpp"
      HEAP32[$41>>2] = $38; //@line 159 "modules/sound/lullaby/VorbisDecoder.cpp"
      STACKTOP = sp;return; //@line 160 "modules/sound/lullaby/VorbisDecoder.cpp"
     }
     $26 = (___cxa_allocate_exception(16)|0); //@line 155 "modules/sound/lullaby/VorbisDecoder.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($26|0),(359069|0),($vararg_buffer|0)); //@line 155 "modules/sound/lullaby/VorbisDecoder.cpp"
     $27 = __THREW__; __THREW__ = 0;
     $28 = $27&1;
     if (!($28)) {
      __THREW__ = 0;
      invoke_viii(3350,($26|0),(96|0),(344|0)); //@line 155 "modules/sound/lullaby/VorbisDecoder.cpp"
      $29 = __THREW__; __THREW__ = 0;
      break;
     }
     $32 = ___cxa_find_matching_catch_2()|0;
     $33 = tempRet0;
     ___cxa_free_exception(($26|0)); //@line 155 "modules/sound/lullaby/VorbisDecoder.cpp"
     $$0 = $33;$$01 = $32;
     __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 160 "modules/sound/lullaby/VorbisDecoder.cpp"
     ___resumeException($$01|0);
     // unreachable;
    }
   }
  }
 } while(0);
 $30 = ___cxa_find_matching_catch_2()|0;
 $31 = tempRet0;
 $$0 = $31;$$01 = $30;
 __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 160 "modules/sound/lullaby/VorbisDecoder.cpp"
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN4love5sound7lullabyL11vorbisCloseEPv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1; //@line 40 "modules/sound/lullaby/VorbisDecoder.cpp"
}
function __ZN4love5sound7lullabyL10vorbisSeekEPvxi($datasource,$0,$1,$whence) {
 $datasource = $datasource|0;
 $0 = $0|0;
 $1 = $1|0;
 $whence = $whence|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $actualOffset$0$off0 = 0, $actualOffset$1$off0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 switch ($whence|0) {
 case 0:  {
  $2 = ((($datasource)) + 4|0); //@line 91 "modules/sound/lullaby/VorbisDecoder.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 91 "modules/sound/lullaby/VorbisDecoder.cpp"
  $4 = ($3|0)<(0); //@line 91 "modules/sound/lullaby/VorbisDecoder.cpp"
  $5 = $4 << 31 >> 31; //@line 91 "modules/sound/lullaby/VorbisDecoder.cpp"
  $6 = ($5|0)<($1|0); //@line 91 "modules/sound/lullaby/VorbisDecoder.cpp"
  $7 = ($3>>>0)<($0>>>0); //@line 91 "modules/sound/lullaby/VorbisDecoder.cpp"
  $8 = ($5|0)==($1|0); //@line 91 "modules/sound/lullaby/VorbisDecoder.cpp"
  $9 = $8 & $7; //@line 91 "modules/sound/lullaby/VorbisDecoder.cpp"
  $10 = $6 | $9; //@line 91 "modules/sound/lullaby/VorbisDecoder.cpp"
  $actualOffset$0$off0 = $10 ? $3 : $0; //@line 91 "modules/sound/lullaby/VorbisDecoder.cpp"
  $11 = ((($datasource)) + 8|0); //@line 96 "modules/sound/lullaby/VorbisDecoder.cpp"
  HEAP32[$11>>2] = $actualOffset$0$off0; //@line 96 "modules/sound/lullaby/VorbisDecoder.cpp"
  return 0; //@line 118 "modules/sound/lullaby/VorbisDecoder.cpp"
  break;
 }
 case 1:  {
  $12 = ((($datasource)) + 4|0); //@line 100 "modules/sound/lullaby/VorbisDecoder.cpp"
  $13 = HEAP32[$12>>2]|0; //@line 100 "modules/sound/lullaby/VorbisDecoder.cpp"
  $14 = ((($datasource)) + 8|0); //@line 100 "modules/sound/lullaby/VorbisDecoder.cpp"
  $15 = HEAP32[$14>>2]|0; //@line 100 "modules/sound/lullaby/VorbisDecoder.cpp"
  $16 = (($13) - ($15))|0; //@line 100 "modules/sound/lullaby/VorbisDecoder.cpp"
  $17 = ($16|0)<(0); //@line 101 "modules/sound/lullaby/VorbisDecoder.cpp"
  $18 = $17 << 31 >> 31; //@line 101 "modules/sound/lullaby/VorbisDecoder.cpp"
  $19 = ($18|0)>($1|0); //@line 101 "modules/sound/lullaby/VorbisDecoder.cpp"
  $20 = ($16>>>0)>($0>>>0); //@line 101 "modules/sound/lullaby/VorbisDecoder.cpp"
  $21 = ($18|0)==($1|0); //@line 101 "modules/sound/lullaby/VorbisDecoder.cpp"
  $22 = $21 & $20; //@line 101 "modules/sound/lullaby/VorbisDecoder.cpp"
  $23 = $19 | $22; //@line 101 "modules/sound/lullaby/VorbisDecoder.cpp"
  $actualOffset$1$off0 = $23 ? $0 : $16; //@line 101 "modules/sound/lullaby/VorbisDecoder.cpp"
  $24 = (($actualOffset$1$off0) + ($15))|0; //@line 106 "modules/sound/lullaby/VorbisDecoder.cpp"
  HEAP32[$14>>2] = $24; //@line 106 "modules/sound/lullaby/VorbisDecoder.cpp"
  return 0; //@line 118 "modules/sound/lullaby/VorbisDecoder.cpp"
  break;
 }
 case 2:  {
  $25 = ($1|0)<(0); //@line 109 "modules/sound/lullaby/VorbisDecoder.cpp"
  $26 = ((($datasource)) + 4|0); //@line 110 "modules/sound/lullaby/VorbisDecoder.cpp"
  $27 = HEAP32[$26>>2]|0; //@line 110 "modules/sound/lullaby/VorbisDecoder.cpp"
  if ($25) {
   $28 = (_i64Add(($27|0),0,($0|0),($1|0))|0); //@line 110 "modules/sound/lullaby/VorbisDecoder.cpp"
   $29 = tempRet0; //@line 110 "modules/sound/lullaby/VorbisDecoder.cpp"
   $30 = ((($datasource)) + 8|0); //@line 110 "modules/sound/lullaby/VorbisDecoder.cpp"
   HEAP32[$30>>2] = $28; //@line 110 "modules/sound/lullaby/VorbisDecoder.cpp"
   return 0; //@line 118 "modules/sound/lullaby/VorbisDecoder.cpp"
  } else {
   $31 = ((($datasource)) + 8|0); //@line 112 "modules/sound/lullaby/VorbisDecoder.cpp"
   HEAP32[$31>>2] = $27; //@line 112 "modules/sound/lullaby/VorbisDecoder.cpp"
   return 0; //@line 118 "modules/sound/lullaby/VorbisDecoder.cpp"
  }
  break;
 }
 default: {
  return 0; //@line 118 "modules/sound/lullaby/VorbisDecoder.cpp"
 }
 }
 return (0)|0;
}
function __ZN4love5sound7lullabyL10vorbisReadEPvjjS2_($ptr,$byteSize,$sizeToRead,$datasource) {
 $ptr = $ptr|0;
 $byteSize = $byteSize|0;
 $sizeToRead = $sizeToRead|0;
 $datasource = $datasource|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($datasource)) + 4|0); //@line 56 "modules/sound/lullaby/VorbisDecoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 56 "modules/sound/lullaby/VorbisDecoder.cpp"
 $2 = ((($datasource)) + 8|0); //@line 56 "modules/sound/lullaby/VorbisDecoder.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 56 "modules/sound/lullaby/VorbisDecoder.cpp"
 $4 = (($1) - ($3))|0; //@line 56 "modules/sound/lullaby/VorbisDecoder.cpp"
 $5 = Math_imul($sizeToRead, $byteSize)|0; //@line 57 "modules/sound/lullaby/VorbisDecoder.cpp"
 $6 = ($5>>>0)<($4>>>0); //@line 57 "modules/sound/lullaby/VorbisDecoder.cpp"
 $$ = $6 ? $5 : $4; //@line 58 "modules/sound/lullaby/VorbisDecoder.cpp"
 $7 = ($$|0)==(0); //@line 63 "modules/sound/lullaby/VorbisDecoder.cpp"
 if ($7) {
  return ($$|0); //@line 72 "modules/sound/lullaby/VorbisDecoder.cpp"
 }
 $8 = HEAP32[$datasource>>2]|0; //@line 66 "modules/sound/lullaby/VorbisDecoder.cpp"
 $9 = (($8) + ($3)|0); //@line 66 "modules/sound/lullaby/VorbisDecoder.cpp"
 _memcpy(($ptr|0),($9|0),($$|0))|0; //@line 66 "modules/sound/lullaby/VorbisDecoder.cpp"
 $10 = HEAP32[$2>>2]|0; //@line 68 "modules/sound/lullaby/VorbisDecoder.cpp"
 $11 = (($10) + ($$))|0; //@line 68 "modules/sound/lullaby/VorbisDecoder.cpp"
 HEAP32[$2>>2] = $11; //@line 68 "modules/sound/lullaby/VorbisDecoder.cpp"
 return ($$|0); //@line 72 "modules/sound/lullaby/VorbisDecoder.cpp"
}
function __ZN4love5sound7lullabyL10vorbisTellEPv($datasource) {
 $datasource = $datasource|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($datasource)) + 8|0); //@line 125 "modules/sound/lullaby/VorbisDecoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 125 "modules/sound/lullaby/VorbisDecoder.cpp"
 return ($1|0); //@line 125 "modules/sound/lullaby/VorbisDecoder.cpp"
}
function __ZN4love5sound7lullaby13VorbisDecoder6decodeEv($this) {
 $this = $this|0;
 var $$2 = 0, $$lcssa = 0, $$size$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $size$0$ph = 0, $size$0$ph$lcssa19 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 192 "modules/sound/lullaby/VorbisDecoder.cpp"
 $1 = ((($this)) + 72|0); //@line 194 "modules/sound/lullaby/VorbisDecoder.cpp"
 $2 = ((($this)) + 32|0); //@line 194 "modules/sound/lullaby/VorbisDecoder.cpp"
 $3 = ((($this)) + 800|0); //@line 194 "modules/sound/lullaby/VorbisDecoder.cpp"
 $size$0$ph = 0;
 L1: while(1) {
  while(1) {
   $4 = HEAP32[$0>>2]|0; //@line 192 "modules/sound/lullaby/VorbisDecoder.cpp"
   $5 = ($4|0)>($size$0$ph|0); //@line 192 "modules/sound/lullaby/VorbisDecoder.cpp"
   if (!($5)) {
    $$2 = $size$0$ph;
    label = 9;
    break L1;
   }
   $6 = HEAP32[$2>>2]|0; //@line 194 "modules/sound/lullaby/VorbisDecoder.cpp"
   $7 = (($6) + ($size$0$ph)|0); //@line 194 "modules/sound/lullaby/VorbisDecoder.cpp"
   $8 = (($4) - ($size$0$ph))|0; //@line 194 "modules/sound/lullaby/VorbisDecoder.cpp"
   $9 = HEAP32[$3>>2]|0; //@line 194 "modules/sound/lullaby/VorbisDecoder.cpp"
   $10 = HEAP32[$this>>2]|0; //@line 194 "modules/sound/lullaby/VorbisDecoder.cpp"
   $11 = ((($10)) + 52|0); //@line 194 "modules/sound/lullaby/VorbisDecoder.cpp"
   $12 = HEAP32[$11>>2]|0; //@line 194 "modules/sound/lullaby/VorbisDecoder.cpp"
   $13 = (FUNCTION_TABLE_ii[$12 & 8191]($this)|0); //@line 194 "modules/sound/lullaby/VorbisDecoder.cpp"
   $14 = ($13|0)==(16); //@line 194 "modules/sound/lullaby/VorbisDecoder.cpp"
   $15 = $14 ? 2 : 1; //@line 194 "modules/sound/lullaby/VorbisDecoder.cpp"
   $16 = (_ov_read($1,$7,$8,$9,$15,1,0)|0); //@line 194 "modules/sound/lullaby/VorbisDecoder.cpp"
   $17 = ($16|0)==(-3); //@line 196 "modules/sound/lullaby/VorbisDecoder.cpp"
   if (!($17)) {
    $$lcssa = $16;
    break;
   }
  }
  $18 = ($$lcssa|0)<(-127); //@line 198 "modules/sound/lullaby/VorbisDecoder.cpp"
  if ($18) {
   $$2 = -1;
   label = 9;
   break;
  }
  $19 = ($$lcssa|0)==(0); //@line 200 "modules/sound/lullaby/VorbisDecoder.cpp"
  if ($19) {
   $size$0$ph$lcssa19 = $size$0$ph;
   label = 8;
   break;
  }
  $20 = ($$lcssa|0)>(0); //@line 205 "modules/sound/lullaby/VorbisDecoder.cpp"
  $21 = $20 ? $$lcssa : 0; //@line 205 "modules/sound/lullaby/VorbisDecoder.cpp"
  $$size$0 = (($21) + ($size$0$ph))|0; //@line 205 "modules/sound/lullaby/VorbisDecoder.cpp"
  $size$0$ph = $$size$0;
 }
 if ((label|0) == 8) {
  $22 = ((($this)) + 36|0); //@line 202 "modules/sound/lullaby/VorbisDecoder.cpp"
  HEAP8[$22>>0] = 1; //@line 202 "modules/sound/lullaby/VorbisDecoder.cpp"
  $$2 = $size$0$ph$lcssa19;
  return ($$2|0); //@line 210 "modules/sound/lullaby/VorbisDecoder.cpp"
 }
 else if ((label|0) == 9) {
  return ($$2|0); //@line 210 "modules/sound/lullaby/VorbisDecoder.cpp"
 }
 return (0)|0;
}
function __ZN4love5sound7lullaby13VorbisDecoder4seekEf($this,$s) {
 $this = $this|0;
 $s = +$s;
 var $$0 = 0, $0 = 0.0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $s; //@line 218 "modules/sound/lullaby/VorbisDecoder.cpp"
 $1 = !($0 <= 9.9999999999999995E-7); //@line 218 "modules/sound/lullaby/VorbisDecoder.cpp"
 $2 = ((($this)) + 72|0); //@line 219 "modules/sound/lullaby/VorbisDecoder.cpp"
 if ($1) {
  $4 = (_ov_time_seek($2,$0)|0); //@line 221 "modules/sound/lullaby/VorbisDecoder.cpp"
  $result$0 = $4;
 } else {
  $3 = (_ov_raw_seek($2,0,0)|0); //@line 219 "modules/sound/lullaby/VorbisDecoder.cpp"
  $result$0 = $3;
 }
 $5 = ($result$0|0)==(0); //@line 223 "modules/sound/lullaby/VorbisDecoder.cpp"
 if (!($5)) {
  $$0 = 0;
  return ($$0|0); //@line 230 "modules/sound/lullaby/VorbisDecoder.cpp"
 }
 $6 = ((($this)) + 36|0); //@line 225 "modules/sound/lullaby/VorbisDecoder.cpp"
 HEAP8[$6>>0] = 0; //@line 225 "modules/sound/lullaby/VorbisDecoder.cpp"
 $$0 = 1;
 return ($$0|0); //@line 230 "modules/sound/lullaby/VorbisDecoder.cpp"
}
function __ZN4love5sound7lullaby13VorbisDecoder6rewindEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 72|0); //@line 236 "modules/sound/lullaby/VorbisDecoder.cpp"
 $1 = (_ov_raw_seek($0,0,0)|0); //@line 236 "modules/sound/lullaby/VorbisDecoder.cpp"
 $2 = ($1|0)==(0); //@line 238 "modules/sound/lullaby/VorbisDecoder.cpp"
 if (!($2)) {
  $$0 = 0;
  return ($$0|0); //@line 245 "modules/sound/lullaby/VorbisDecoder.cpp"
 }
 $3 = ((($this)) + 36|0); //@line 240 "modules/sound/lullaby/VorbisDecoder.cpp"
 HEAP8[$3>>0] = 0; //@line 240 "modules/sound/lullaby/VorbisDecoder.cpp"
 $$0 = 1;
 return ($$0|0); //@line 245 "modules/sound/lullaby/VorbisDecoder.cpp"
}
function __ZN4love5sound7lullaby13VorbisDecoder10isSeekableEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 72|0); //@line 249 "modules/sound/lullaby/VorbisDecoder.cpp"
 $1 = (_ov_seekable($0)|0); //@line 249 "modules/sound/lullaby/VorbisDecoder.cpp"
 $2 = ($1|0)!=(0); //@line 250 "modules/sound/lullaby/VorbisDecoder.cpp"
 return ($2|0); //@line 250 "modules/sound/lullaby/VorbisDecoder.cpp"
}
function __ZNK4love5sound7lullaby13VorbisDecoder11getChannelsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 792|0); //@line 255 "modules/sound/lullaby/VorbisDecoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 255 "modules/sound/lullaby/VorbisDecoder.cpp"
 $2 = ((($1)) + 4|0); //@line 255 "modules/sound/lullaby/VorbisDecoder.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 255 "modules/sound/lullaby/VorbisDecoder.cpp"
 return ($3|0); //@line 255 "modules/sound/lullaby/VorbisDecoder.cpp"
}
function __ZNK4love5sound7lullaby13VorbisDecoder11getBitDepthEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 16; //@line 260 "modules/sound/lullaby/VorbisDecoder.cpp"
}
function __ZNK4love5sound7lullaby13VorbisDecoder13getSampleRateEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 792|0); //@line 265 "modules/sound/lullaby/VorbisDecoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 265 "modules/sound/lullaby/VorbisDecoder.cpp"
 $2 = ((($1)) + 8|0); //@line 265 "modules/sound/lullaby/VorbisDecoder.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 265 "modules/sound/lullaby/VorbisDecoder.cpp"
 return ($3|0); //@line 265 "modules/sound/lullaby/VorbisDecoder.cpp"
}
function __ZN4love5sound7lullaby13VorbisDecoder11getDurationEv($this) {
 $this = $this|0;
 var $$ = 0.0, $0 = 0, $1 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0.0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 808|0); //@line 271 "modules/sound/lullaby/VorbisDecoder.cpp"
 $1 = +HEAPF64[$0>>3]; //@line 271 "modules/sound/lullaby/VorbisDecoder.cpp"
 $2 = $1 == -2.0; //@line 271 "modules/sound/lullaby/VorbisDecoder.cpp"
 if (!($2)) {
  $7 = $1;
  return (+$7); //@line 279 "modules/sound/lullaby/VorbisDecoder.cpp"
 }
 $3 = ((($this)) + 72|0); //@line 273 "modules/sound/lullaby/VorbisDecoder.cpp"
 $4 = (+_ov_time_total($3,-1)); //@line 273 "modules/sound/lullaby/VorbisDecoder.cpp"
 $5 = $4 == -131.0; //@line 275 "modules/sound/lullaby/VorbisDecoder.cpp"
 $6 = $4 < 0.0; //@line 275 "modules/sound/lullaby/VorbisDecoder.cpp"
 $or$cond = $5 | $6; //@line 275 "modules/sound/lullaby/VorbisDecoder.cpp"
 $$ = $or$cond ? -1.0 : $4; //@line 275 "modules/sound/lullaby/VorbisDecoder.cpp"
 HEAPF64[$0>>3] = $$; //@line 276 "modules/sound/lullaby/VorbisDecoder.cpp"
 $7 = $$;
 return (+$7); //@line 279 "modules/sound/lullaby/VorbisDecoder.cpp"
}
function __ZN4love5sound7lullaby13VorbisDecoder7acceptsERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($ext) {
 $ext = $ext|0;
 var $$02$ph = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0;
 var $lpad$phi7$index = 0, $lpad$phi7$index2 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[624024]|0; //@line 169 "modules/sound/lullaby/VorbisDecoder.cpp"
 $1 = ($0<<24>>24)==(0); //@line 169 "modules/sound/lullaby/VorbisDecoder.cpp"
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((624024|0))|0); //@line 169 "modules/sound/lullaby/VorbisDecoder.cpp"
   $3 = ($2|0)==(0); //@line 169 "modules/sound/lullaby/VorbisDecoder.cpp"
   if (!($3)) {
    __THREW__ = 0;
    invoke_viii(3460,(631320|0),(359098|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $4 = __THREW__; __THREW__ = 0;
    $5 = $4&1;
    if ($5) {
     $36 = ___cxa_find_matching_catch_2()|0;
     $37 = tempRet0;
     $lpad$phi7$index = $36;$lpad$phi7$index2 = $37;
     ___cxa_guard_abort((624024|0)); //@line 169 "modules/sound/lullaby/VorbisDecoder.cpp"
     ___resumeException($lpad$phi7$index|0);
     // unreachable;
    }
    __THREW__ = 0;
    invoke_viii(3460,((631332)|0),(359102|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $6 = __THREW__; __THREW__ = 0;
    $7 = $6&1;
    if ($7) {
     $$02$ph = (631332);
    } else {
     __THREW__ = 0;
     invoke_viii(3460,((631344)|0),(359106|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $8 = __THREW__; __THREW__ = 0;
     $9 = $8&1;
     if ($9) {
      $$02$ph = (631344);
     } else {
      __THREW__ = 0;
      invoke_viii(3460,((631356)|0),(1282466|0),0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $10 = __THREW__; __THREW__ = 0;
      $11 = $10&1;
      if ($11) {
       $$02$ph = (631356);
      } else {
       (___cxa_atexit((3765|0),(0|0),(___dso_handle|0))|0); //@line 169 "modules/sound/lullaby/VorbisDecoder.cpp"
       ___cxa_guard_release((624024|0)); //@line 169 "modules/sound/lullaby/VorbisDecoder.cpp"
       break;
      }
     }
    }
    $34 = ___cxa_find_matching_catch_2()|0;
    $35 = tempRet0;
    $39 = $$02$ph;
    while(1) {
     $38 = ((($39)) + -12|0); //@line 170 "modules/sound/lullaby/VorbisDecoder.cpp"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($38); //@line 170 "modules/sound/lullaby/VorbisDecoder.cpp"
     $40 = ($38|0)==(631320|0); //@line 170 "modules/sound/lullaby/VorbisDecoder.cpp"
     if ($40) {
      $lpad$phi7$index = $34;$lpad$phi7$index2 = $35;
      break;
     } else {
      $39 = $38;
     }
    }
    ___cxa_guard_abort((624024|0)); //@line 169 "modules/sound/lullaby/VorbisDecoder.cpp"
    ___resumeException($lpad$phi7$index|0);
    // unreachable;
   }
  }
 } while(0);
 $12 = HEAP8[$ext>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = $12 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = ($13<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $15 = ((($ext)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $16 = HEAP32[$15>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $17 = $12&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $18 = $17 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $19 = $14 ? $18 : $16; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $20 = ((($ext)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $21 = HEAP32[$20>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $22 = ((($ext)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $23 = $14 ? $22 : $21; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $i$0 = 0;
 while(1) {
  $24 = (631320 + (($i$0*12)|0)|0); //@line 174 "modules/sound/lullaby/VorbisDecoder.cpp"
  $25 = HEAP8[$24>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $26 = $25 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $27 = ($26<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($27) {
   $30 = $25&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $31 = $30 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $33 = $31;
  } else {
   $28 = (((631320 + (($i$0*12)|0)|0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $29 = HEAP32[$28>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $33 = $29;
  }
  $32 = ($33|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($32) {
   $57 = 0;
   label = 23;
   break;
  }
  if ($27) {
   $45 = $25&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $46 = $45 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $47 = ((($24)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $49 = $46;$51 = $47;
  } else {
   $41 = (((631320 + (($i$0*12)|0)|0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $42 = HEAP32[$41>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $43 = (((631320 + (($i$0*12)|0)|0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $44 = HEAP32[$43>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $49 = $42;$51 = $44;
  }
  $48 = ($19>>>0)<($49>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $50 = $48 ? $19 : $49; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $52 = (_memcmp($51,$23,$50)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $53 = ($52|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($53) {
   $54 = ($49>>>0)>=($19>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $55 = $48 ^ 1; //@line 176 "modules/sound/lullaby/VorbisDecoder.cpp"
   $or$cond = $54 & $55;
   if ($or$cond) {
    $57 = 1;
    label = 23;
    break;
   }
  }
  $56 = (($i$0) + 1)|0; //@line 174 "modules/sound/lullaby/VorbisDecoder.cpp"
  $i$0 = $56;
 }
 if ((label|0) == 23) {
  return ($57|0); //@line 181 "modules/sound/lullaby/VorbisDecoder.cpp"
 }
 return (0)|0;
}
function ___cxx_global_array_dtor($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((631356)); //@line 171 "modules/sound/lullaby/VorbisDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((631344)); //@line 171 "modules/sound/lullaby/VorbisDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((631332)); //@line 171 "modules/sound/lullaby/VorbisDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev(631320); //@line 171 "modules/sound/lullaby/VorbisDecoder.cpp"
 return; //@line 171 "modules/sound/lullaby/VorbisDecoder.cpp"
}
function __ZN4love5sound7lullabyL13read_callbackEPvPhPj($userdata,$buffer,$size) {
 $userdata = $userdata|0;
 $buffer = $buffer|0;
 $size = $size|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($userdata)) + 4|0); //@line 38 "modules/sound/lullaby/WaveDecoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 38 "modules/sound/lullaby/WaveDecoder.cpp"
 $2 = ((($userdata)) + 8|0); //@line 38 "modules/sound/lullaby/WaveDecoder.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 38 "modules/sound/lullaby/WaveDecoder.cpp"
 $4 = (($1) - ($3))|0; //@line 38 "modules/sound/lullaby/WaveDecoder.cpp"
 $5 = HEAP32[$size>>2]|0; //@line 39 "modules/sound/lullaby/WaveDecoder.cpp"
 $6 = ($5>>>0)<($4>>>0); //@line 39 "modules/sound/lullaby/WaveDecoder.cpp"
 $$ = $6 ? $5 : $4; //@line 39 "modules/sound/lullaby/WaveDecoder.cpp"
 $7 = HEAP32[$userdata>>2]|0; //@line 40 "modules/sound/lullaby/WaveDecoder.cpp"
 $8 = (($7) + ($3)|0); //@line 40 "modules/sound/lullaby/WaveDecoder.cpp"
 _memcpy(($buffer|0),($8|0),($$|0))|0; //@line 40 "modules/sound/lullaby/WaveDecoder.cpp"
 $9 = HEAP32[$2>>2]|0; //@line 41 "modules/sound/lullaby/WaveDecoder.cpp"
 $10 = (($9) + ($$))|0; //@line 41 "modules/sound/lullaby/WaveDecoder.cpp"
 HEAP32[$2>>2] = $10; //@line 41 "modules/sound/lullaby/WaveDecoder.cpp"
 HEAP32[$size>>2] = $$; //@line 42 "modules/sound/lullaby/WaveDecoder.cpp"
 return 0; //@line 43 "modules/sound/lullaby/WaveDecoder.cpp"
}
function __ZN4love5sound7lullabyL13seek_callbackEPvy($userdata,$0,$1) {
 $userdata = $userdata|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$off0 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($userdata)) + 4|0); //@line 49 "modules/sound/lullaby/WaveDecoder.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 49 "modules/sound/lullaby/WaveDecoder.cpp"
 $4 = (0)>($1>>>0); //@line 49 "modules/sound/lullaby/WaveDecoder.cpp"
 $5 = ($3>>>0)>($0>>>0); //@line 49 "modules/sound/lullaby/WaveDecoder.cpp"
 $6 = (0)==($1|0); //@line 49 "modules/sound/lullaby/WaveDecoder.cpp"
 $7 = $6 & $5; //@line 49 "modules/sound/lullaby/WaveDecoder.cpp"
 $8 = $4 | $7; //@line 49 "modules/sound/lullaby/WaveDecoder.cpp"
 $$off0 = $8 ? $0 : $3; //@line 49 "modules/sound/lullaby/WaveDecoder.cpp"
 $9 = ((($userdata)) + 8|0); //@line 49 "modules/sound/lullaby/WaveDecoder.cpp"
 HEAP32[$9>>2] = $$off0; //@line 49 "modules/sound/lullaby/WaveDecoder.cpp"
 return 0; //@line 50 "modules/sound/lullaby/WaveDecoder.cpp"
}
function __ZN4love5sound7lullabyL13tell_callbackEPvPy($userdata,$offset) {
 $userdata = $userdata|0;
 $offset = $offset|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($userdata)) + 8|0); //@line 56 "modules/sound/lullaby/WaveDecoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 56 "modules/sound/lullaby/WaveDecoder.cpp"
 $2 = $offset; //@line 56 "modules/sound/lullaby/WaveDecoder.cpp"
 $3 = $2; //@line 56 "modules/sound/lullaby/WaveDecoder.cpp"
 HEAP32[$3>>2] = $1; //@line 56 "modules/sound/lullaby/WaveDecoder.cpp"
 $4 = (($2) + 4)|0; //@line 56 "modules/sound/lullaby/WaveDecoder.cpp"
 $5 = $4; //@line 56 "modules/sound/lullaby/WaveDecoder.cpp"
 HEAP32[$5>>2] = 0; //@line 56 "modules/sound/lullaby/WaveDecoder.cpp"
 return 0; //@line 57 "modules/sound/lullaby/WaveDecoder.cpp"
}
function __ZN4love5sound7lullaby11WaveDecoderD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19024); //@line 98 "modules/sound/lullaby/WaveDecoder.cpp"
 $0 = ((($this)) + 52|0); //@line 99 "modules/sound/lullaby/WaveDecoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 99 "modules/sound/lullaby/WaveDecoder.cpp"
 __THREW__ = 0;
 (invoke_ii(3766,($1|0))|0); //@line 99 "modules/sound/lullaby/WaveDecoder.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 100 "modules/sound/lullaby/WaveDecoder.cpp"
  ___clang_call_terminate($4); //@line 100 "modules/sound/lullaby/WaveDecoder.cpp"
  // unreachable; //@line 100 "modules/sound/lullaby/WaveDecoder.cpp"
 } else {
  __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 100 "modules/sound/lullaby/WaveDecoder.cpp"
  return; //@line 100 "modules/sound/lullaby/WaveDecoder.cpp"
 }
}
function __ZN4love5sound7lullaby11WaveDecoderD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19024); //@line 98 "modules/sound/lullaby/WaveDecoder.cpp"
 $0 = ((($this)) + 52|0); //@line 99 "modules/sound/lullaby/WaveDecoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 99 "modules/sound/lullaby/WaveDecoder.cpp"
 __THREW__ = 0;
 (invoke_ii(3766,($1|0))|0); //@line 99 "modules/sound/lullaby/WaveDecoder.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 100 "modules/sound/lullaby/WaveDecoder.cpp"
  ___clang_call_terminate($4); //@line 100 "modules/sound/lullaby/WaveDecoder.cpp"
  // unreachable; //@line 100 "modules/sound/lullaby/WaveDecoder.cpp"
 } else {
  __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 100 "modules/sound/lullaby/WaveDecoder.cpp"
  __ZdlPv($this); //@line 98 "modules/sound/lullaby/WaveDecoder.cpp"
  return; //@line 100 "modules/sound/lullaby/WaveDecoder.cpp"
 }
}
function __ZN4love5sound7lullaby11WaveDecoder5cloneEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(80)|0); //@line 120 "modules/sound/lullaby/WaveDecoder.cpp"
 $1 = ((($this)) + 8|0); //@line 139 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 139 "./common/Object.h"
 $3 = ((($this)) + 12|0); //@line 120 "modules/sound/lullaby/WaveDecoder.cpp"
 $4 = ((($this)) + 24|0); //@line 120 "modules/sound/lullaby/WaveDecoder.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 120 "modules/sound/lullaby/WaveDecoder.cpp"
 __THREW__ = 0;
 invoke_viiii(3756,($0|0),($2|0),($3|0),($5|0)); //@line 120 "modules/sound/lullaby/WaveDecoder.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $8 = ___cxa_find_matching_catch_2()|0;
  $9 = tempRet0;
  __ZdlPv($0); //@line 120 "modules/sound/lullaby/WaveDecoder.cpp"
  ___resumeException($8|0);
  // unreachable;
 } else {
  return ($0|0); //@line 120 "modules/sound/lullaby/WaveDecoder.cpp"
 }
 return (0)|0;
}
function __ZN4love5sound7lullaby11WaveDecoderC2EPNS_4DataERKNSt3__112basic_stringIcNS5_11char_traitsIcEENS5_9allocatorIcEEEEi($this,$data,$ext,$bufferSize) {
 $this = $this|0;
 $data = $data|0;
 $ext = $ext|0;
 $bufferSize = $bufferSize|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $8 = 0, $9 = 0, $switch = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 __ZN4love5sound7lullaby7DecoderC2EPNS_4DataERKNSt3__112basic_stringIcNS5_11char_traitsIcEENS5_9allocatorIcEEEEi($this,$data,$ext,$bufferSize); //@line 64 "modules/sound/lullaby/WaveDecoder.cpp"
 HEAP32[$this>>2] = (19024); //@line 65 "modules/sound/lullaby/WaveDecoder.cpp"
 $0 = HEAP32[$data>>2]|0; //@line 66 "modules/sound/lullaby/WaveDecoder.cpp"
 $1 = ((($0)) + 16|0); //@line 66 "modules/sound/lullaby/WaveDecoder.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 66 "modules/sound/lullaby/WaveDecoder.cpp"
 __THREW__ = 0;
 $3 = (invoke_ii($2|0,($data|0))|0); //@line 66 "modules/sound/lullaby/WaveDecoder.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 do {
  if (!($5)) {
   $6 = ((($this)) + 40|0); //@line 66 "modules/sound/lullaby/WaveDecoder.cpp"
   HEAP32[$6>>2] = $3; //@line 66 "modules/sound/lullaby/WaveDecoder.cpp"
   $7 = HEAP32[$data>>2]|0; //@line 67 "modules/sound/lullaby/WaveDecoder.cpp"
   $8 = ((($7)) + 20|0); //@line 67 "modules/sound/lullaby/WaveDecoder.cpp"
   $9 = HEAP32[$8>>2]|0; //@line 67 "modules/sound/lullaby/WaveDecoder.cpp"
   __THREW__ = 0;
   $10 = (invoke_ii($9|0,($data|0))|0); //@line 67 "modules/sound/lullaby/WaveDecoder.cpp"
   $11 = __THREW__; __THREW__ = 0;
   $12 = $11&1;
   if (!($12)) {
    $13 = ((($this)) + 44|0); //@line 67 "modules/sound/lullaby/WaveDecoder.cpp"
    HEAP32[$13>>2] = $10; //@line 67 "modules/sound/lullaby/WaveDecoder.cpp"
    $14 = ((($this)) + 48|0); //@line 68 "modules/sound/lullaby/WaveDecoder.cpp"
    HEAP32[$14>>2] = 0; //@line 68 "modules/sound/lullaby/WaveDecoder.cpp"
    $15 = ((($this)) + 52|0); //@line 70 "modules/sound/lullaby/WaveDecoder.cpp"
    __THREW__ = 0;
    $16 = (invoke_iiii(3767,($15|0),(19004|0),($6|0))|0); //@line 70 "modules/sound/lullaby/WaveDecoder.cpp"
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if (!($18)) {
     $19 = ($16|0)<(0); //@line 71 "modules/sound/lullaby/WaveDecoder.cpp"
     if ($19) {
      $20 = (___cxa_allocate_exception(16)|0); //@line 72 "modules/sound/lullaby/WaveDecoder.cpp"
      __THREW__ = 0;
      invoke_viii(3349,($20|0),(359145|0),($vararg_buffer|0)); //@line 72 "modules/sound/lullaby/WaveDecoder.cpp"
      $21 = __THREW__; __THREW__ = 0;
      $22 = $21&1;
      if (!($22)) {
       __THREW__ = 0;
       invoke_viii(3350,($20|0),(96|0),(344|0)); //@line 72 "modules/sound/lullaby/WaveDecoder.cpp"
       $23 = __THREW__; __THREW__ = 0;
       break;
      }
      $26 = ___cxa_find_matching_catch_2()|0;
      $27 = tempRet0;
      ___cxa_free_exception(($20|0)); //@line 72 "modules/sound/lullaby/WaveDecoder.cpp"
      $$1 = $26;$$12 = $27;
      __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 95 "modules/sound/lullaby/WaveDecoder.cpp"
      ___resumeException($$1|0);
      // unreachable;
     }
     $28 = HEAP32[$15>>2]|0; //@line 76 "modules/sound/lullaby/WaveDecoder.cpp"
     $29 = ((($this)) + 56|0); //@line 76 "modules/sound/lullaby/WaveDecoder.cpp"
     __THREW__ = 0;
     $30 = (invoke_iii(3768,($28|0),($29|0))|0); //@line 76 "modules/sound/lullaby/WaveDecoder.cpp"
     $31 = __THREW__; __THREW__ = 0;
     $32 = $31&1;
     do {
      if ($32) {
       label = 13;
      } else {
       $33 = ($30|0)<(0); //@line 77 "modules/sound/lullaby/WaveDecoder.cpp"
       if ($33) {
        $34 = (___cxa_allocate_exception(16)|0); //@line 78 "modules/sound/lullaby/WaveDecoder.cpp"
        __THREW__ = 0;
        invoke_viii(3349,($34|0),(359165|0),($vararg_buffer1|0)); //@line 78 "modules/sound/lullaby/WaveDecoder.cpp"
        $35 = __THREW__; __THREW__ = 0;
        $36 = $35&1;
        if ($36) {
         $40 = ___cxa_find_matching_catch_3(96|0)|0;
         $41 = tempRet0;
         ___cxa_free_exception(($34|0)); //@line 78 "modules/sound/lullaby/WaveDecoder.cpp"
         $$0 = $40;$$01 = $41;
         break;
        } else {
         __THREW__ = 0;
         invoke_viii(3350,($34|0),(96|0),(344|0)); //@line 78 "modules/sound/lullaby/WaveDecoder.cpp"
         $37 = __THREW__; __THREW__ = 0;
         label = 13;
         break;
        }
       }
       $42 = ((($this)) + 58|0); //@line 80 "modules/sound/lullaby/WaveDecoder.cpp"
       $43 = HEAP16[$42>>1]|0; //@line 80 "modules/sound/lullaby/WaveDecoder.cpp"
       $44 = ($43&65535)>(2); //@line 80 "modules/sound/lullaby/WaveDecoder.cpp"
       if ($44) {
        $45 = (___cxa_allocate_exception(16)|0); //@line 81 "modules/sound/lullaby/WaveDecoder.cpp"
        __THREW__ = 0;
        invoke_viii(3349,($45|0),(359201|0),($vararg_buffer3|0)); //@line 81 "modules/sound/lullaby/WaveDecoder.cpp"
        $46 = __THREW__; __THREW__ = 0;
        $47 = $46&1;
        if ($47) {
         $49 = ___cxa_find_matching_catch_3(96|0)|0;
         $50 = tempRet0;
         ___cxa_free_exception(($45|0)); //@line 81 "modules/sound/lullaby/WaveDecoder.cpp"
         $$0 = $49;$$01 = $50;
         break;
        } else {
         __THREW__ = 0;
         invoke_viii(3350,($45|0),(96|0),(344|0)); //@line 81 "modules/sound/lullaby/WaveDecoder.cpp"
         $48 = __THREW__; __THREW__ = 0;
         label = 13;
         break;
        }
       }
       $51 = HEAP16[$29>>1]|0; //@line 83 "modules/sound/lullaby/WaveDecoder.cpp"
       $switch = ($51&65535)<(2); //@line 83 "modules/sound/lullaby/WaveDecoder.cpp"
       if ($switch) {
        STACKTOP = sp;return; //@line 95 "modules/sound/lullaby/WaveDecoder.cpp"
       }
       $52 = HEAP32[$15>>2]|0; //@line 85 "modules/sound/lullaby/WaveDecoder.cpp"
       __THREW__ = 0;
       $53 = (invoke_iii(3769,($52|0),1)|0); //@line 85 "modules/sound/lullaby/WaveDecoder.cpp"
       $54 = __THREW__; __THREW__ = 0;
       $55 = $54&1;
       if ($55) {
        label = 13;
       } else {
        $56 = ($53|0)<(0); //@line 86 "modules/sound/lullaby/WaveDecoder.cpp"
        if (!($56)) {
         STACKTOP = sp;return; //@line 95 "modules/sound/lullaby/WaveDecoder.cpp"
        }
        $57 = (___cxa_allocate_exception(16)|0); //@line 87 "modules/sound/lullaby/WaveDecoder.cpp"
        __THREW__ = 0;
        invoke_viii(3349,($57|0),(359234|0),($vararg_buffer5|0)); //@line 87 "modules/sound/lullaby/WaveDecoder.cpp"
        $58 = __THREW__; __THREW__ = 0;
        $59 = $58&1;
        if ($59) {
         $61 = ___cxa_find_matching_catch_3(96|0)|0;
         $62 = tempRet0;
         ___cxa_free_exception(($57|0)); //@line 87 "modules/sound/lullaby/WaveDecoder.cpp"
         $$0 = $61;$$01 = $62;
         break;
        } else {
         __THREW__ = 0;
         invoke_viii(3350,($57|0),(96|0),(344|0)); //@line 87 "modules/sound/lullaby/WaveDecoder.cpp"
         $60 = __THREW__; __THREW__ = 0;
         label = 13;
         break;
        }
       }
      }
     } while(0);
     if ((label|0) == 13) {
      $38 = ___cxa_find_matching_catch_3(96|0)|0;
      $39 = tempRet0;
      $$0 = $38;$$01 = $39;
     }
     $63 = (_llvm_eh_typeid_for((96|0))|0); //@line 89 "modules/sound/lullaby/WaveDecoder.cpp"
     $64 = ($$01|0)==($63|0); //@line 89 "modules/sound/lullaby/WaveDecoder.cpp"
     if (!($64)) {
      $$1 = $$0;$$12 = $$01;
      __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 95 "modules/sound/lullaby/WaveDecoder.cpp"
      ___resumeException($$1|0);
      // unreachable;
     }
     (___cxa_begin_catch(($$0|0))|0); //@line 89 "modules/sound/lullaby/WaveDecoder.cpp"
     $65 = HEAP32[$15>>2]|0; //@line 92 "modules/sound/lullaby/WaveDecoder.cpp"
     __THREW__ = 0;
     (invoke_ii(3766,($65|0))|0); //@line 92 "modules/sound/lullaby/WaveDecoder.cpp"
     $66 = __THREW__; __THREW__ = 0;
     $67 = $66&1;
     if (!($67)) {
      __THREW__ = 0;
      invoke_v(3656); //@line 93 "modules/sound/lullaby/WaveDecoder.cpp"
      $68 = __THREW__; __THREW__ = 0;
     }
     $69 = ___cxa_find_matching_catch_2()|0;
     $70 = tempRet0;
     __THREW__ = 0;
     invoke_v(3378); //@line 94 "modules/sound/lullaby/WaveDecoder.cpp"
     $71 = __THREW__; __THREW__ = 0;
     $72 = $71&1;
     if ($72) {
      $73 = ___cxa_find_matching_catch_3(0|0)|0;
      $74 = tempRet0;
      ___clang_call_terminate($73); //@line 94 "modules/sound/lullaby/WaveDecoder.cpp"
      // unreachable; //@line 94 "modules/sound/lullaby/WaveDecoder.cpp"
     } else {
      $$1 = $69;$$12 = $70;
      __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 95 "modules/sound/lullaby/WaveDecoder.cpp"
      ___resumeException($$1|0);
      // unreachable;
     }
    }
   }
  }
 } while(0);
 $24 = ___cxa_find_matching_catch_2()|0;
 $25 = tempRet0;
 $$1 = $24;$$12 = $25;
 __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 95 "modules/sound/lullaby/WaveDecoder.cpp"
 ___resumeException($$1|0);
 // unreachable;
}
function __ZN4love5sound7lullaby11WaveDecoder6decodeEv($this) {
 $this = $this|0;
 var $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $bytes = 0;
 var $size$06 = 0, $size$06$lcssa15 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $bytes = sp;
 $0 = ((($this)) + 24|0); //@line 127 "modules/sound/lullaby/WaveDecoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 127 "modules/sound/lullaby/WaveDecoder.cpp"
 $2 = ($1|0)==(0); //@line 127 "modules/sound/lullaby/WaveDecoder.cpp"
 if ($2) {
  $$2 = 0;
  STACKTOP = sp;return ($$2|0); //@line 144 "modules/sound/lullaby/WaveDecoder.cpp"
 }
 $3 = ((($this)) + 52|0); //@line 130 "modules/sound/lullaby/WaveDecoder.cpp"
 $4 = ((($this)) + 32|0); //@line 130 "modules/sound/lullaby/WaveDecoder.cpp"
 $6 = $1;$size$06 = 0;
 while(1) {
  $5 = (($6) - ($size$06))|0; //@line 129 "modules/sound/lullaby/WaveDecoder.cpp"
  HEAP32[$bytes>>2] = $5; //@line 129 "modules/sound/lullaby/WaveDecoder.cpp"
  $7 = HEAP32[$3>>2]|0; //@line 130 "modules/sound/lullaby/WaveDecoder.cpp"
  $8 = HEAP32[$4>>2]|0; //@line 130 "modules/sound/lullaby/WaveDecoder.cpp"
  $9 = (($8) + ($size$06)|0); //@line 130 "modules/sound/lullaby/WaveDecoder.cpp"
  $10 = (_wuff_read($7,$9,$bytes)|0); //@line 130 "modules/sound/lullaby/WaveDecoder.cpp"
  $11 = ($10|0)<(0); //@line 132 "modules/sound/lullaby/WaveDecoder.cpp"
  if ($11) {
   label = 4;
   break;
  }
  $12 = HEAP32[$bytes>>2]|0; //@line 134 "modules/sound/lullaby/WaveDecoder.cpp"
  $13 = ($12|0)==(0); //@line 134 "modules/sound/lullaby/WaveDecoder.cpp"
  if ($13) {
   $size$06$lcssa15 = $size$06;
   label = 7;
   break;
  }
  $14 = (($12) + ($size$06))|0; //@line 140 "modules/sound/lullaby/WaveDecoder.cpp"
  $15 = HEAP32[$0>>2]|0; //@line 127 "modules/sound/lullaby/WaveDecoder.cpp"
  $16 = ($14>>>0)<($15>>>0); //@line 127 "modules/sound/lullaby/WaveDecoder.cpp"
  if ($16) {
   $6 = $15;$size$06 = $14;
  } else {
   $$2 = $14;
   label = 8;
   break;
  }
 }
 if ((label|0) == 4) {
  $$2 = 0;
  STACKTOP = sp;return ($$2|0); //@line 144 "modules/sound/lullaby/WaveDecoder.cpp"
 }
 else if ((label|0) == 7) {
  $17 = ((($this)) + 36|0); //@line 136 "modules/sound/lullaby/WaveDecoder.cpp"
  HEAP8[$17>>0] = 1; //@line 136 "modules/sound/lullaby/WaveDecoder.cpp"
  $$2 = $size$06$lcssa15;
  STACKTOP = sp;return ($$2|0); //@line 144 "modules/sound/lullaby/WaveDecoder.cpp"
 }
 else if ((label|0) == 8) {
  STACKTOP = sp;return ($$2|0); //@line 144 "modules/sound/lullaby/WaveDecoder.cpp"
 }
 return (0)|0;
}
function __ZN4love5sound7lullaby11WaveDecoder4seekEf($this,$s) {
 $this = $this|0;
 $s = +$s;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 148 "modules/sound/lullaby/WaveDecoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 148 "modules/sound/lullaby/WaveDecoder.cpp"
 $2 = ((($this)) + 60|0); //@line 148 "modules/sound/lullaby/WaveDecoder.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 148 "modules/sound/lullaby/WaveDecoder.cpp"
 $4 = (+($3>>>0)); //@line 148 "modules/sound/lullaby/WaveDecoder.cpp"
 $5 = $4 * $s; //@line 148 "modules/sound/lullaby/WaveDecoder.cpp"
 $6 = (~~$5)>>>0; //@line 148 "modules/sound/lullaby/WaveDecoder.cpp"
 $7 = +Math_abs($5) >= 1.0 ? $5 > 0.0 ? (~~+Math_min(+Math_floor($5 / 4294967296.0), 4294967295.0)) >>> 0 : ~~+Math_ceil(($5 - +(~~$5 >>> 0)) / 4294967296.0) >>> 0 : 0; //@line 148 "modules/sound/lullaby/WaveDecoder.cpp"
 $8 = (_wuff_seek($1,$6,$7)|0); //@line 148 "modules/sound/lullaby/WaveDecoder.cpp"
 $9 = ($8|0)>(-1); //@line 150 "modules/sound/lullaby/WaveDecoder.cpp"
 if (!($9)) {
  $$0 = 0;
  return ($$0|0); //@line 157 "modules/sound/lullaby/WaveDecoder.cpp"
 }
 $10 = ((($this)) + 36|0); //@line 152 "modules/sound/lullaby/WaveDecoder.cpp"
 HEAP8[$10>>0] = 0; //@line 152 "modules/sound/lullaby/WaveDecoder.cpp"
 $$0 = 1;
 return ($$0|0); //@line 157 "modules/sound/lullaby/WaveDecoder.cpp"
}
function __ZN4love5sound7lullaby11WaveDecoder6rewindEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 161 "modules/sound/lullaby/WaveDecoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 161 "modules/sound/lullaby/WaveDecoder.cpp"
 $2 = (_wuff_seek($1,0,0)|0); //@line 161 "modules/sound/lullaby/WaveDecoder.cpp"
 $3 = ($2|0)>(-1); //@line 163 "modules/sound/lullaby/WaveDecoder.cpp"
 if (!($3)) {
  $$0 = 0;
  return ($$0|0); //@line 170 "modules/sound/lullaby/WaveDecoder.cpp"
 }
 $4 = ((($this)) + 36|0); //@line 165 "modules/sound/lullaby/WaveDecoder.cpp"
 HEAP8[$4>>0] = 0; //@line 165 "modules/sound/lullaby/WaveDecoder.cpp"
 $$0 = 1;
 return ($$0|0); //@line 170 "modules/sound/lullaby/WaveDecoder.cpp"
}
function __ZN4love5sound7lullaby11WaveDecoder10isSeekableEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1; //@line 174 "modules/sound/lullaby/WaveDecoder.cpp"
}
function __ZNK4love5sound7lullaby11WaveDecoder11getChannelsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 58|0); //@line 179 "modules/sound/lullaby/WaveDecoder.cpp"
 $1 = HEAP16[$0>>1]|0; //@line 179 "modules/sound/lullaby/WaveDecoder.cpp"
 $2 = $1&65535; //@line 179 "modules/sound/lullaby/WaveDecoder.cpp"
 return ($2|0); //@line 179 "modules/sound/lullaby/WaveDecoder.cpp"
}
function __ZNK4love5sound7lullaby11WaveDecoder11getBitDepthEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 64|0); //@line 184 "modules/sound/lullaby/WaveDecoder.cpp"
 $1 = HEAP16[$0>>1]|0; //@line 184 "modules/sound/lullaby/WaveDecoder.cpp"
 $2 = ($1<<16>>16)==(8); //@line 184 "modules/sound/lullaby/WaveDecoder.cpp"
 $3 = $2 ? 8 : 16; //@line 184 "modules/sound/lullaby/WaveDecoder.cpp"
 return ($3|0); //@line 184 "modules/sound/lullaby/WaveDecoder.cpp"
}
function __ZNK4love5sound7lullaby11WaveDecoder13getSampleRateEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 60|0); //@line 189 "modules/sound/lullaby/WaveDecoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 189 "modules/sound/lullaby/WaveDecoder.cpp"
 return ($1|0); //@line 189 "modules/sound/lullaby/WaveDecoder.cpp"
}
function __ZN4love5sound7lullaby11WaveDecoder11getDurationEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 72|0); //@line 194 "modules/sound/lullaby/WaveDecoder.cpp"
 $1 = $0; //@line 194 "modules/sound/lullaby/WaveDecoder.cpp"
 $2 = $1; //@line 194 "modules/sound/lullaby/WaveDecoder.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 194 "modules/sound/lullaby/WaveDecoder.cpp"
 $4 = (($1) + 4)|0; //@line 194 "modules/sound/lullaby/WaveDecoder.cpp"
 $5 = $4; //@line 194 "modules/sound/lullaby/WaveDecoder.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 194 "modules/sound/lullaby/WaveDecoder.cpp"
 $7 = (+($3>>>0)) + (4294967296.0*(+($6>>>0))); //@line 194 "modules/sound/lullaby/WaveDecoder.cpp"
 $8 = ((($this)) + 60|0); //@line 194 "modules/sound/lullaby/WaveDecoder.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 194 "modules/sound/lullaby/WaveDecoder.cpp"
 $10 = (+($9>>>0)); //@line 194 "modules/sound/lullaby/WaveDecoder.cpp"
 $11 = $7 / $10; //@line 194 "modules/sound/lullaby/WaveDecoder.cpp"
 return (+$11); //@line 194 "modules/sound/lullaby/WaveDecoder.cpp"
}
function __ZN4love5sound7lullaby11WaveDecoder7acceptsERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($ext) {
 $ext = $ext|0;
 var $$index = 0, $$index2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[624032]|0; //@line 104 "modules/sound/lullaby/WaveDecoder.cpp"
 $1 = ($0<<24>>24)==(0); //@line 104 "modules/sound/lullaby/WaveDecoder.cpp"
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((624032|0))|0); //@line 104 "modules/sound/lullaby/WaveDecoder.cpp"
   $3 = ($2|0)==(0); //@line 104 "modules/sound/lullaby/WaveDecoder.cpp"
   if (!($3)) {
    __THREW__ = 0;
    invoke_viii(3460,(631368|0),(359262|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $4 = __THREW__; __THREW__ = 0;
    $5 = $4&1;
    if ($5) {
     $6 = ___cxa_find_matching_catch_2()|0;
     $7 = tempRet0;
     $$index = $6;$$index2 = $7;
     ___cxa_guard_abort((624032|0)); //@line 104 "modules/sound/lullaby/WaveDecoder.cpp"
     ___resumeException($$index|0);
     // unreachable;
    }
    __THREW__ = 0;
    invoke_viii(3460,((631380)|0),(1282466|0),0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $8 = __THREW__; __THREW__ = 0;
    $9 = $8&1;
    if (!($9)) {
     (___cxa_atexit((3770|0),(0|0),(___dso_handle|0))|0); //@line 104 "modules/sound/lullaby/WaveDecoder.cpp"
     ___cxa_guard_release((624032|0)); //@line 104 "modules/sound/lullaby/WaveDecoder.cpp"
     break;
    }
    $32 = ___cxa_find_matching_catch_2()|0;
    $33 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev(631368); //@line 105 "modules/sound/lullaby/WaveDecoder.cpp"
    $$index = $32;$$index2 = $33;
    ___cxa_guard_abort((624032|0)); //@line 104 "modules/sound/lullaby/WaveDecoder.cpp"
    ___resumeException($$index|0);
    // unreachable;
   }
  }
 } while(0);
 $10 = HEAP8[$ext>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $10 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = ($11<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ((($ext)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = HEAP32[$13>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $15 = $10&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $16 = $15 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $17 = $12 ? $16 : $14; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $18 = ((($ext)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $19 = HEAP32[$18>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $20 = ((($ext)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $21 = $12 ? $20 : $19; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $i$0 = 0;
 while(1) {
  $22 = (631368 + (($i$0*12)|0)|0); //@line 109 "modules/sound/lullaby/WaveDecoder.cpp"
  $23 = HEAP8[$22>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $24 = $23 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $25 = ($24<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($25) {
   $28 = $23&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $29 = $28 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $31 = $29;
  } else {
   $26 = (((631368 + (($i$0*12)|0)|0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $27 = HEAP32[$26>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $31 = $27;
  }
  $30 = ($31|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($30) {
   $50 = 0;
   label = 20;
   break;
  }
  if ($25) {
   $38 = $23&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $39 = $38 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $40 = ((($22)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $42 = $39;$44 = $40;
  } else {
   $34 = (((631368 + (($i$0*12)|0)|0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $35 = HEAP32[$34>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $36 = (((631368 + (($i$0*12)|0)|0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $37 = HEAP32[$36>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $42 = $35;$44 = $37;
  }
  $41 = ($17>>>0)<($42>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $43 = $41 ? $17 : $42; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $45 = (_memcmp($44,$21,$43)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $46 = ($45|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($46) {
   $47 = ($42>>>0)>=($17>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $48 = $41 ^ 1; //@line 111 "modules/sound/lullaby/WaveDecoder.cpp"
   $or$cond = $47 & $48;
   if ($or$cond) {
    $50 = 1;
    label = 20;
    break;
   }
  }
  $49 = (($i$0) + 1)|0; //@line 109 "modules/sound/lullaby/WaveDecoder.cpp"
  $i$0 = $49;
 }
 if ((label|0) == 20) {
  return ($50|0); //@line 116 "modules/sound/lullaby/WaveDecoder.cpp"
 }
 return (0)|0;
}
function ___cxx_global_array_dtor_2219($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((631380)); //@line 106 "modules/sound/lullaby/WaveDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev(631368); //@line 106 "modules/sound/lullaby/WaveDecoder.cpp"
 return; //@line 106 "modules/sound/lullaby/WaveDecoder.cpp"
}
function __ZN4love5sound5SoundD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ModuleD2Ev($this); //@line 30 "modules/sound/Sound.cpp"
 return; //@line 30 "modules/sound/Sound.cpp"
}
function __ZN4love5sound5Sound12newSoundDataEPNS0_7DecoderE($this,$decoder) {
 $this = $this|0;
 $decoder = $decoder|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(28)|0); //@line 34 "modules/sound/Sound.cpp"
 __THREW__ = 0;
 invoke_vii(3771,($0|0),($decoder|0)); //@line 34 "modules/sound/Sound.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 34 "modules/sound/Sound.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 34 "modules/sound/Sound.cpp"
 }
 return (0)|0;
}
function __ZN4love5sound5Sound12newSoundDataEiiii($this,$samples,$sampleRate,$bitDepth,$channels) {
 $this = $this|0;
 $samples = $samples|0;
 $sampleRate = $sampleRate|0;
 $bitDepth = $bitDepth|0;
 $channels = $channels|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(28)|0); //@line 39 "modules/sound/Sound.cpp"
 __THREW__ = 0;
 invoke_viiiii(3772,($0|0),($samples|0),($sampleRate|0),($bitDepth|0),($channels|0)); //@line 39 "modules/sound/Sound.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 39 "modules/sound/Sound.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 39 "modules/sound/Sound.cpp"
 }
 return (0)|0;
}
function __ZN4love5sound9SoundDataD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19096); //@line 110 "modules/sound/SoundData.cpp"
 $0 = ((($this)) + 8|0); //@line 111 "modules/sound/SoundData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 111 "modules/sound/SoundData.cpp"
 $2 = ($1|0)==(0|0); //@line 111 "modules/sound/SoundData.cpp"
 if ($2) {
  return; //@line 113 "modules/sound/SoundData.cpp"
 }
 _free($1); //@line 112 "modules/sound/SoundData.cpp"
 return; //@line 113 "modules/sound/SoundData.cpp"
}
function __ZN4love5sound9SoundDataD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19096); //@line 110 "modules/sound/SoundData.cpp"
 $0 = ((($this)) + 8|0); //@line 111 "modules/sound/SoundData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 111 "modules/sound/SoundData.cpp"
 $2 = ($1|0)==(0|0); //@line 111 "modules/sound/SoundData.cpp"
 if (!($2)) {
  _free($1); //@line 112 "modules/sound/SoundData.cpp"
 }
 __ZdlPv($this); //@line 110 "modules/sound/SoundData.cpp"
 return; //@line 113 "modules/sound/SoundData.cpp"
}
function __ZNK4love5sound9SoundData7getDataEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 157 "modules/sound/SoundData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 157 "modules/sound/SoundData.cpp"
 return ($1|0); //@line 157 "modules/sound/SoundData.cpp"
}
function __ZNK4love5sound9SoundData7getSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 162 "modules/sound/SoundData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 162 "modules/sound/SoundData.cpp"
 return ($1|0); //@line 162 "modules/sound/SoundData.cpp"
}
function __ZNK4love5sound9SoundData11getChannelsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 167 "modules/sound/SoundData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 167 "modules/sound/SoundData.cpp"
 return ($1|0); //@line 167 "modules/sound/SoundData.cpp"
}
function __ZNK4love5sound9SoundData11getBitDepthEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 172 "modules/sound/SoundData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 172 "modules/sound/SoundData.cpp"
 return ($1|0); //@line 172 "modules/sound/SoundData.cpp"
}
function __ZNK4love5sound9SoundData13getSampleRateEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 177 "modules/sound/SoundData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 177 "modules/sound/SoundData.cpp"
 return ($1|0); //@line 177 "modules/sound/SoundData.cpp"
}
function __ZNK4love5sound9SoundData14getSampleCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 182 "modules/sound/SoundData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 182 "modules/sound/SoundData.cpp"
 $2 = ((($this)) + 24|0); //@line 182 "modules/sound/SoundData.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 182 "modules/sound/SoundData.cpp"
 $4 = (($1>>>0) / ($3>>>0))&-1; //@line 182 "modules/sound/SoundData.cpp"
 $5 = ((($this)) + 20|0); //@line 182 "modules/sound/SoundData.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 182 "modules/sound/SoundData.cpp"
 $7 = (($6|0) / 8)&-1; //@line 182 "modules/sound/SoundData.cpp"
 $8 = (($4>>>0) / ($7>>>0))&-1; //@line 182 "modules/sound/SoundData.cpp"
 return ($8|0); //@line 182 "modules/sound/SoundData.cpp"
}
function __ZNK4love5sound9SoundData11getDurationEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $2 = 0.0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 187 "modules/sound/SoundData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 187 "modules/sound/SoundData.cpp"
 $2 = (+($1>>>0)); //@line 187 "modules/sound/SoundData.cpp"
 $3 = ((($this)) + 24|0); //@line 187 "modules/sound/SoundData.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 187 "modules/sound/SoundData.cpp"
 $5 = ((($this)) + 16|0); //@line 187 "modules/sound/SoundData.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 187 "modules/sound/SoundData.cpp"
 $7 = Math_imul($6, $4)|0; //@line 187 "modules/sound/SoundData.cpp"
 $8 = ((($this)) + 20|0); //@line 187 "modules/sound/SoundData.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 187 "modules/sound/SoundData.cpp"
 $10 = Math_imul($7, $9)|0; //@line 187 "modules/sound/SoundData.cpp"
 $11 = (($10|0) / 8)&-1; //@line 187 "modules/sound/SoundData.cpp"
 $12 = (+($11|0)); //@line 187 "modules/sound/SoundData.cpp"
 $13 = $2 / $12; //@line 187 "modules/sound/SoundData.cpp"
 return (+$13); //@line 187 "modules/sound/SoundData.cpp"
}
function __ZN4love5sound9SoundDataC2EPNS0_7DecoderE($this,$decoder) {
 $this = $this|0;
 $decoder = $decoder|0;
 var $$0 = 0, $$01 = 0, $$not = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bufferSize$0 = 0, $bufferSize$1 = 0, $bufferSize$1$lcssa = 0, $bufferSize$223 = 0, $decoded$0 = 0, $lpad$phi$index = 0, $lpad$phi$index6 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 __ZN4love6ObjectC2Ev($this); //@line 37 "./common/Data.h"
 HEAP32[$this>>2] = (19096); //@line 43 "modules/sound/SoundData.cpp"
 $0 = ((($this)) + 8|0); //@line 38 "modules/sound/SoundData.cpp"
 HEAP32[$0>>2] = 0; //@line 38 "modules/sound/SoundData.cpp"
 $1 = ((($this)) + 12|0); //@line 39 "modules/sound/SoundData.cpp"
 HEAP32[$1>>2] = 0; //@line 39 "modules/sound/SoundData.cpp"
 $2 = ((($this)) + 16|0); //@line 40 "modules/sound/SoundData.cpp"
 HEAP32[$2>>2] = 44100; //@line 40 "modules/sound/SoundData.cpp"
 $3 = ((($this)) + 20|0); //@line 41 "modules/sound/SoundData.cpp"
 HEAP32[$3>>2] = 0; //@line 41 "modules/sound/SoundData.cpp"
 $4 = ((($this)) + 24|0); //@line 42 "modules/sound/SoundData.cpp"
 HEAP32[$4>>2] = 0; //@line 42 "modules/sound/SoundData.cpp"
 $5 = HEAP32[$decoder>>2]|0; //@line 44 "modules/sound/SoundData.cpp"
 $6 = ((($5)) + 52|0); //@line 44 "modules/sound/SoundData.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 44 "modules/sound/SoundData.cpp"
 __THREW__ = 0;
 $8 = (invoke_ii($7|0,($decoder|0))|0); //@line 44 "modules/sound/SoundData.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 do {
  if ($10) {
   label = 9;
  } else {
   $11 = ($8|0)==(8); //@line 44 "modules/sound/SoundData.cpp"
   if (!($11)) {
    $12 = HEAP32[$decoder>>2]|0; //@line 44 "modules/sound/SoundData.cpp"
    $13 = ((($12)) + 52|0); //@line 44 "modules/sound/SoundData.cpp"
    $14 = HEAP32[$13>>2]|0; //@line 44 "modules/sound/SoundData.cpp"
    __THREW__ = 0;
    $15 = (invoke_ii($14|0,($decoder|0))|0); //@line 44 "modules/sound/SoundData.cpp"
    $16 = __THREW__; __THREW__ = 0;
    $17 = $16&1;
    if ($17) {
     label = 9;
     break;
    }
    $18 = ($15|0)==(16); //@line 44 "modules/sound/SoundData.cpp"
    if (!($18)) {
     $19 = (___cxa_allocate_exception(16)|0); //@line 45 "modules/sound/SoundData.cpp"
     $20 = HEAP32[$decoder>>2]|0; //@line 45 "modules/sound/SoundData.cpp"
     $21 = ((($20)) + 52|0); //@line 45 "modules/sound/SoundData.cpp"
     $22 = HEAP32[$21>>2]|0; //@line 45 "modules/sound/SoundData.cpp"
     __THREW__ = 0;
     $23 = (invoke_ii($22|0,($decoder|0))|0); //@line 45 "modules/sound/SoundData.cpp"
     $24 = __THREW__; __THREW__ = 0;
     $25 = $24&1;
     if (!($25)) {
      __THREW__ = 0;
      HEAP32[$vararg_buffer>>2] = $23; //@line 45 "modules/sound/SoundData.cpp"
      invoke_viii(3349,($19|0),(359310|0),($vararg_buffer|0)); //@line 45 "modules/sound/SoundData.cpp"
      $26 = __THREW__; __THREW__ = 0;
      $27 = $26&1;
      if (!($27)) {
       __THREW__ = 0;
       invoke_viii(3350,($19|0),(96|0),(344|0)); //@line 45 "modules/sound/SoundData.cpp"
       $28 = __THREW__; __THREW__ = 0;
       label = 9;
       break;
      }
     }
     $33 = ___cxa_find_matching_catch_2()|0;
     $34 = tempRet0;
     ___cxa_free_exception(($19|0)); //@line 45 "modules/sound/SoundData.cpp"
     $$0 = $34;$$01 = $33;
     ___resumeException($$01|0);
     // unreachable;
    }
   }
   $35 = HEAP32[$decoder>>2]|0; //@line 48 "modules/sound/SoundData.cpp"
   $36 = ((($35)) + 20|0); //@line 48 "modules/sound/SoundData.cpp"
   $37 = HEAP32[$36>>2]|0; //@line 48 "modules/sound/SoundData.cpp"
   __THREW__ = 0;
   $38 = (invoke_ii($37|0,($decoder|0))|0); //@line 48 "modules/sound/SoundData.cpp"
   $39 = __THREW__; __THREW__ = 0;
   $40 = $39&1;
   if ($40) {
    label = 9;
   } else {
    $bufferSize$0 = 524288;$decoded$0 = $38;
    while(1) {
     $41 = ($decoded$0|0)>(0); //@line 50 "modules/sound/SoundData.cpp"
     $42 = HEAP32[$0>>2]|0; //@line 58 "modules/sound/SoundData.cpp"
     $43 = ($42|0)!=(0|0); //@line 54 "modules/sound/SoundData.cpp"
     if (!($41)) {
      label = 28;
      break;
     }
     $44 = HEAP32[$1>>2]|0; //@line 54 "modules/sound/SoundData.cpp"
     $$not = $43 ^ 1; //@line 54 "modules/sound/SoundData.cpp"
     $45 = (($44) + ($decoded$0))|0; //@line 54 "modules/sound/SoundData.cpp"
     $46 = ($bufferSize$0>>>0)<($45>>>0); //@line 54 "modules/sound/SoundData.cpp"
     $or$cond = $46 | $$not; //@line 54 "modules/sound/SoundData.cpp"
     if ($or$cond) {
      $47 = (($44) + ($decoded$0))|0; //@line 56 "modules/sound/SoundData.cpp"
      $bufferSize$1 = $bufferSize$0;
      while(1) {
       $48 = ($bufferSize$1>>>0)<($47>>>0); //@line 56 "modules/sound/SoundData.cpp"
       $49 = $bufferSize$1 << 1; //@line 57 "modules/sound/SoundData.cpp"
       if ($48) {
        $bufferSize$1 = $49;
       } else {
        $bufferSize$1$lcssa = $bufferSize$1;
        break;
       }
      }
      $50 = (_realloc($42,$bufferSize$1$lcssa)|0); //@line 58 "modules/sound/SoundData.cpp"
      HEAP32[$0>>2] = $50; //@line 58 "modules/sound/SoundData.cpp"
      $51 = ($50|0)==(0|0); //@line 61 "modules/sound/SoundData.cpp"
      if ($51) {
       label = 19;
       break;
      } else {
       $66 = $50;$bufferSize$223 = $bufferSize$1$lcssa;
      }
     } else {
      $66 = $42;$bufferSize$223 = $bufferSize$0;
     }
     $58 = HEAP32[$1>>2]|0; //@line 65 "modules/sound/SoundData.cpp"
     $59 = HEAP32[$decoder>>2]|0; //@line 65 "modules/sound/SoundData.cpp"
     $60 = ((($59)) + 28|0); //@line 65 "modules/sound/SoundData.cpp"
     $61 = HEAP32[$60>>2]|0; //@line 65 "modules/sound/SoundData.cpp"
     __THREW__ = 0;
     $62 = (invoke_ii($61|0,($decoder|0))|0); //@line 65 "modules/sound/SoundData.cpp"
     $63 = __THREW__; __THREW__ = 0;
     $64 = $63&1;
     if ($64) {
      label = 8;
      break;
     }
     $65 = (($66) + ($58)|0); //@line 65 "modules/sound/SoundData.cpp"
     _memcpy(($65|0),($62|0),($decoded$0|0))|0; //@line 65 "modules/sound/SoundData.cpp"
     $67 = HEAP32[$1>>2]|0; //@line 68 "modules/sound/SoundData.cpp"
     $68 = $decoded$0 ^ -1; //@line 68 "modules/sound/SoundData.cpp"
     $69 = ($67>>>0)>($68>>>0); //@line 68 "modules/sound/SoundData.cpp"
     if ($69) {
      label = 24;
      break;
     }
     $77 = (($67) + ($decoded$0))|0; //@line 75 "modules/sound/SoundData.cpp"
     HEAP32[$1>>2] = $77; //@line 75 "modules/sound/SoundData.cpp"
     $78 = HEAP32[$decoder>>2]|0; //@line 77 "modules/sound/SoundData.cpp"
     $79 = ((($78)) + 20|0); //@line 77 "modules/sound/SoundData.cpp"
     $80 = HEAP32[$79>>2]|0; //@line 77 "modules/sound/SoundData.cpp"
     __THREW__ = 0;
     $81 = (invoke_ii($80|0,($decoder|0))|0); //@line 77 "modules/sound/SoundData.cpp"
     $82 = __THREW__; __THREW__ = 0;
     $83 = $82&1;
     if ($83) {
      label = 8;
      break;
     } else {
      $bufferSize$0 = $bufferSize$223;$decoded$0 = $81;
     }
    }
    if ((label|0) == 8) {
     $29 = ___cxa_find_matching_catch_2()|0;
     $30 = tempRet0;
     $lpad$phi$index = $29;$lpad$phi$index6 = $30;
     break;
    }
    else if ((label|0) == 19) {
     $52 = (___cxa_allocate_exception(16)|0); //@line 62 "modules/sound/SoundData.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($52|0),(359332|0),($vararg_buffer1|0)); //@line 62 "modules/sound/SoundData.cpp"
     $53 = __THREW__; __THREW__ = 0;
     $54 = $53&1;
     if (!($54)) {
      __THREW__ = 0;
      invoke_viii(3350,($52|0),(96|0),(344|0)); //@line 62 "modules/sound/SoundData.cpp"
      $55 = __THREW__; __THREW__ = 0;
      label = 9;
      break;
     }
     $56 = ___cxa_find_matching_catch_2()|0;
     $57 = tempRet0;
     ___cxa_free_exception(($52|0)); //@line 62 "modules/sound/SoundData.cpp"
     $$0 = $57;$$01 = $56;
     ___resumeException($$01|0);
     // unreachable;
    }
    else if ((label|0) == 24) {
     $70 = HEAP32[$0>>2]|0; //@line 70 "modules/sound/SoundData.cpp"
     _free($70); //@line 70 "modules/sound/SoundData.cpp"
     $71 = (___cxa_allocate_exception(16)|0); //@line 71 "modules/sound/SoundData.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($71|0),(359332|0),($vararg_buffer3|0)); //@line 71 "modules/sound/SoundData.cpp"
     $72 = __THREW__; __THREW__ = 0;
     $73 = $72&1;
     if (!($73)) {
      __THREW__ = 0;
      invoke_viii(3350,($71|0),(96|0),(344|0)); //@line 71 "modules/sound/SoundData.cpp"
      $74 = __THREW__; __THREW__ = 0;
      label = 9;
      break;
     }
     $75 = ___cxa_find_matching_catch_2()|0;
     $76 = tempRet0;
     ___cxa_free_exception(($71|0)); //@line 71 "modules/sound/SoundData.cpp"
     $$0 = $76;$$01 = $75;
     ___resumeException($$01|0);
     // unreachable;
    }
    else if ((label|0) == 28) {
     if ($43) {
      $84 = HEAP32[$1>>2]|0; //@line 81 "modules/sound/SoundData.cpp"
      $85 = ($bufferSize$0>>>0)>($84>>>0); //@line 81 "modules/sound/SoundData.cpp"
      if ($85) {
       $86 = (_realloc($42,$84)|0); //@line 82 "modules/sound/SoundData.cpp"
       HEAP32[$0>>2] = $86; //@line 82 "modules/sound/SoundData.cpp"
      }
     }
     $87 = HEAP32[$decoder>>2]|0; //@line 84 "modules/sound/SoundData.cpp"
     $88 = ((($87)) + 48|0); //@line 84 "modules/sound/SoundData.cpp"
     $89 = HEAP32[$88>>2]|0; //@line 84 "modules/sound/SoundData.cpp"
     __THREW__ = 0;
     $90 = (invoke_ii($89|0,($decoder|0))|0); //@line 84 "modules/sound/SoundData.cpp"
     $91 = __THREW__; __THREW__ = 0;
     $92 = $91&1;
     if ($92) {
      label = 9;
      break;
     }
     HEAP32[$4>>2] = $90; //@line 84 "modules/sound/SoundData.cpp"
     $93 = HEAP32[$decoder>>2]|0; //@line 85 "modules/sound/SoundData.cpp"
     $94 = ((($93)) + 52|0); //@line 85 "modules/sound/SoundData.cpp"
     $95 = HEAP32[$94>>2]|0; //@line 85 "modules/sound/SoundData.cpp"
     __THREW__ = 0;
     $96 = (invoke_ii($95|0,($decoder|0))|0); //@line 85 "modules/sound/SoundData.cpp"
     $97 = __THREW__; __THREW__ = 0;
     $98 = $97&1;
     if ($98) {
      label = 9;
      break;
     }
     HEAP32[$3>>2] = $96; //@line 85 "modules/sound/SoundData.cpp"
     $99 = HEAP32[$decoder>>2]|0; //@line 86 "modules/sound/SoundData.cpp"
     $100 = ((($99)) + 56|0); //@line 86 "modules/sound/SoundData.cpp"
     $101 = HEAP32[$100>>2]|0; //@line 86 "modules/sound/SoundData.cpp"
     __THREW__ = 0;
     $102 = (invoke_ii($101|0,($decoder|0))|0); //@line 86 "modules/sound/SoundData.cpp"
     $103 = __THREW__; __THREW__ = 0;
     $104 = $103&1;
     if ($104) {
      label = 9;
      break;
     }
     HEAP32[$2>>2] = $102; //@line 86 "modules/sound/SoundData.cpp"
     STACKTOP = sp;return; //@line 87 "modules/sound/SoundData.cpp"
    }
   }
  }
 } while(0);
 if ((label|0) == 9) {
  $31 = ___cxa_find_matching_catch_2()|0;
  $32 = tempRet0;
  $lpad$phi$index = $31;$lpad$phi$index6 = $32;
 }
 $$0 = $lpad$phi$index6;$$01 = $lpad$phi$index;
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN4love5sound9SoundDataC2Eiiii($this,$samples,$sampleRate,$bitDepth,$channels) {
 $this = $this|0;
 $samples = $samples|0;
 $sampleRate = $sampleRate|0;
 $bitDepth = $bitDepth|0;
 $channels = $channels|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 37 "./common/Data.h"
 HEAP32[$this>>2] = (19096); //@line 95 "modules/sound/SoundData.cpp"
 $0 = ((($this)) + 8|0); //@line 90 "modules/sound/SoundData.cpp"
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;HEAP32[$0+12>>2]=0|0;HEAP32[$0+16>>2]=0|0; //@line 91 "modules/sound/SoundData.cpp"
 __THREW__ = 0;
 invoke_viiiiii(3773,($this|0),($samples|0),($sampleRate|0),($bitDepth|0),($channels|0),(0|0)); //@line 96 "modules/sound/SoundData.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  ___resumeException($3|0);
  // unreachable;
 } else {
  return; //@line 97 "modules/sound/SoundData.cpp"
 }
}
function __ZN4love5sound9SoundData4loadEiiiiPv($this,$samples,$sampleRate,$bitDepth,$channels,$newData) {
 $this = $this|0;
 $samples = $samples|0;
 $sampleRate = $sampleRate|0;
 $bitDepth = $bitDepth|0;
 $channels = $channels|0;
 $newData = $newData|0;
 var $$0 = 0, $$02 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer10 = 0, $vararg_buffer12 = 0, $vararg_buffer4 = 0;
 var $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer12 = sp + 40|0;
 $vararg_buffer10 = sp + 32|0;
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($samples|0)<(1); //@line 117 "modules/sound/SoundData.cpp"
 if ($0) {
  $1 = (___cxa_allocate_exception(16)|0); //@line 118 "modules/sound/SoundData.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer>>2] = $samples; //@line 118 "modules/sound/SoundData.cpp"
  invoke_viii(3349,($1|0),(359351|0),($vararg_buffer|0)); //@line 118 "modules/sound/SoundData.cpp"
  $2 = __THREW__; __THREW__ = 0;
  $3 = $2&1;
  if (!($3)) {
   ___cxa_throw(($1|0),(96|0),(344|0)); //@line 118 "modules/sound/SoundData.cpp"
   // unreachable; //@line 118 "modules/sound/SoundData.cpp"
  }
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  ___cxa_free_exception(($1|0)); //@line 118 "modules/sound/SoundData.cpp"
  $$0 = $4;$$02 = $5;
  ___resumeException($$0|0);
  // unreachable;
 }
 $6 = ($sampleRate|0)<(1); //@line 120 "modules/sound/SoundData.cpp"
 if ($6) {
  $7 = (___cxa_allocate_exception(16)|0); //@line 121 "modules/sound/SoundData.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer1>>2] = $sampleRate; //@line 121 "modules/sound/SoundData.cpp"
  invoke_viii(3349,($7|0),(359376|0),($vararg_buffer1|0)); //@line 121 "modules/sound/SoundData.cpp"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if (!($9)) {
   ___cxa_throw(($7|0),(96|0),(344|0)); //@line 121 "modules/sound/SoundData.cpp"
   // unreachable; //@line 121 "modules/sound/SoundData.cpp"
  }
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  ___cxa_free_exception(($7|0)); //@line 121 "modules/sound/SoundData.cpp"
  $$0 = $10;$$02 = $11;
  ___resumeException($$0|0);
  // unreachable;
 }
 switch ($bitDepth|0) {
 case 8: case 16:  {
  break;
 }
 default: {
  $12 = (___cxa_allocate_exception(16)|0); //@line 124 "modules/sound/SoundData.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer4>>2] = $bitDepth; //@line 124 "modules/sound/SoundData.cpp"
  invoke_viii(3349,($12|0),(359310|0),($vararg_buffer4|0)); //@line 124 "modules/sound/SoundData.cpp"
  $13 = __THREW__; __THREW__ = 0;
  $14 = $13&1;
  if (!($14)) {
   ___cxa_throw(($12|0),(96|0),(344|0)); //@line 124 "modules/sound/SoundData.cpp"
   // unreachable; //@line 124 "modules/sound/SoundData.cpp"
  }
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  ___cxa_free_exception(($12|0)); //@line 124 "modules/sound/SoundData.cpp"
  $$0 = $15;$$02 = $16;
  ___resumeException($$0|0);
  // unreachable;
 }
 }
 $17 = ($channels|0)<(1); //@line 126 "modules/sound/SoundData.cpp"
 if ($17) {
  $18 = (___cxa_allocate_exception(16)|0); //@line 127 "modules/sound/SoundData.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer7>>2] = $channels; //@line 127 "modules/sound/SoundData.cpp"
  invoke_viii(3349,($18|0),(359400|0),($vararg_buffer7|0)); //@line 127 "modules/sound/SoundData.cpp"
  $19 = __THREW__; __THREW__ = 0;
  $20 = $19&1;
  if (!($20)) {
   ___cxa_throw(($18|0),(96|0),(344|0)); //@line 127 "modules/sound/SoundData.cpp"
   // unreachable; //@line 127 "modules/sound/SoundData.cpp"
  }
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  ___cxa_free_exception(($18|0)); //@line 127 "modules/sound/SoundData.cpp"
  $$0 = $21;$$02 = $22;
  ___resumeException($$0|0);
  // unreachable;
 }
 $23 = ((($this)) + 8|0); //@line 129 "modules/sound/SoundData.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 129 "modules/sound/SoundData.cpp"
 $25 = ($24|0)==(0|0); //@line 129 "modules/sound/SoundData.cpp"
 if (!($25)) {
  _free($24); //@line 131 "modules/sound/SoundData.cpp"
  HEAP32[$23>>2] = 0; //@line 132 "modules/sound/SoundData.cpp"
 }
 $26 = (($bitDepth|0) / 8)&-1; //@line 135 "modules/sound/SoundData.cpp"
 $27 = Math_imul($26, $samples)|0; //@line 135 "modules/sound/SoundData.cpp"
 $28 = Math_imul($27, $channels)|0; //@line 152 "modules/sound/SoundData.cpp"
 $29 = ((($this)) + 12|0); //@line 135 "modules/sound/SoundData.cpp"
 HEAP32[$29>>2] = $28; //@line 135 "modules/sound/SoundData.cpp"
 $30 = ((($this)) + 16|0); //@line 136 "modules/sound/SoundData.cpp"
 HEAP32[$30>>2] = $sampleRate; //@line 136 "modules/sound/SoundData.cpp"
 $31 = ((($this)) + 20|0); //@line 137 "modules/sound/SoundData.cpp"
 HEAP32[$31>>2] = $bitDepth; //@line 137 "modules/sound/SoundData.cpp"
 $32 = ((($this)) + 24|0); //@line 138 "modules/sound/SoundData.cpp"
 HEAP32[$32>>2] = $channels; //@line 138 "modules/sound/SoundData.cpp"
 $33 = (+($samples|0)); //@line 140 "modules/sound/SoundData.cpp"
 $34 = Math_imul($26, $channels)|0; //@line 141 "modules/sound/SoundData.cpp"
 $35 = (+($34|0)); //@line 141 "modules/sound/SoundData.cpp"
 $36 = $33 * $35; //@line 141 "modules/sound/SoundData.cpp"
 $37 = $36 > 4294967295.0; //@line 142 "modules/sound/SoundData.cpp"
 if ($37) {
  $38 = (___cxa_allocate_exception(16)|0); //@line 143 "modules/sound/SoundData.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($38|0),(359426|0),($vararg_buffer10|0)); //@line 143 "modules/sound/SoundData.cpp"
  $39 = __THREW__; __THREW__ = 0;
  $40 = $39&1;
  if (!($40)) {
   ___cxa_throw(($38|0),(96|0),(344|0)); //@line 143 "modules/sound/SoundData.cpp"
   // unreachable; //@line 143 "modules/sound/SoundData.cpp"
  }
  $41 = ___cxa_find_matching_catch_2()|0;
  $42 = tempRet0;
  ___cxa_free_exception(($38|0)); //@line 143 "modules/sound/SoundData.cpp"
  $$0 = $41;$$02 = $42;
  ___resumeException($$0|0);
  // unreachable;
 }
 $43 = (_malloc($28)|0); //@line 145 "modules/sound/SoundData.cpp"
 HEAP32[$23>>2] = $43; //@line 145 "modules/sound/SoundData.cpp"
 $44 = ($43|0)==(0|0); //@line 146 "modules/sound/SoundData.cpp"
 if ($44) {
  $45 = (___cxa_allocate_exception(16)|0); //@line 147 "modules/sound/SoundData.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($45|0),(359332|0),($vararg_buffer12|0)); //@line 147 "modules/sound/SoundData.cpp"
  $46 = __THREW__; __THREW__ = 0;
  $47 = $46&1;
  if (!($47)) {
   ___cxa_throw(($45|0),(96|0),(344|0)); //@line 147 "modules/sound/SoundData.cpp"
   // unreachable; //@line 147 "modules/sound/SoundData.cpp"
  }
  $48 = ___cxa_find_matching_catch_2()|0;
  $49 = tempRet0;
  ___cxa_free_exception(($45|0)); //@line 147 "modules/sound/SoundData.cpp"
  $$0 = $48;$$02 = $49;
  ___resumeException($$0|0);
  // unreachable;
 }
 $50 = ($newData|0)==(0|0); //@line 149 "modules/sound/SoundData.cpp"
 if ($50) {
  $51 = ($bitDepth|0)==(8); //@line 152 "modules/sound/SoundData.cpp"
  $52 = $51 ? -128 : 0; //@line 152 "modules/sound/SoundData.cpp"
  _memset(($43|0),($52|0),($28|0))|0; //@line 152 "modules/sound/SoundData.cpp"
  STACKTOP = sp;return; //@line 153 "modules/sound/SoundData.cpp"
 } else {
  _memcpy(($43|0),($newData|0),($28|0))|0; //@line 150 "modules/sound/SoundData.cpp"
  STACKTOP = sp;return; //@line 153 "modules/sound/SoundData.cpp"
 }
}
function __ZN4love5sound9SoundDataC2EPviiii($this,$d,$samples,$sampleRate,$bitDepth,$channels) {
 $this = $this|0;
 $d = $d|0;
 $samples = $samples|0;
 $sampleRate = $sampleRate|0;
 $bitDepth = $bitDepth|0;
 $channels = $channels|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 37 "./common/Data.h"
 HEAP32[$this>>2] = (19096); //@line 105 "modules/sound/SoundData.cpp"
 $0 = ((($this)) + 8|0); //@line 100 "modules/sound/SoundData.cpp"
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;HEAP32[$0+12>>2]=0|0;HEAP32[$0+16>>2]=0|0; //@line 101 "modules/sound/SoundData.cpp"
 __THREW__ = 0;
 invoke_viiiiii(3773,($this|0),($samples|0),($sampleRate|0),($bitDepth|0),($channels|0),($d|0)); //@line 106 "modules/sound/SoundData.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  ___resumeException($3|0);
  // unreachable;
 } else {
  return; //@line 107 "modules/sound/SoundData.cpp"
 }
}
function __ZN4love5sound9SoundData9setSampleEif($this,$i,$sample) {
 $this = $this|0;
 $i = $i|0;
 $sample = +$sample;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($i|0)<(0); //@line 193 "modules/sound/SoundData.cpp"
 if (!($0)) {
  $1 = ((($this)) + 12|0); //@line 193 "modules/sound/SoundData.cpp"
  $2 = HEAP32[$1>>2]|0; //@line 193 "modules/sound/SoundData.cpp"
  $3 = ((($this)) + 20|0); //@line 193 "modules/sound/SoundData.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 193 "modules/sound/SoundData.cpp"
  $5 = (($4|0) / 8)&-1; //@line 193 "modules/sound/SoundData.cpp"
  $6 = (($2>>>0) / ($5>>>0))&-1; //@line 193 "modules/sound/SoundData.cpp"
  $7 = ($6>>>0)>($i>>>0); //@line 193 "modules/sound/SoundData.cpp"
  if ($7) {
   $13 = ($4|0)==(16); //@line 196 "modules/sound/SoundData.cpp"
   if ($13) {
    $14 = ((($this)) + 8|0); //@line 199 "modules/sound/SoundData.cpp"
    $15 = HEAP32[$14>>2]|0; //@line 199 "modules/sound/SoundData.cpp"
    $16 = $sample * 32767.0; //@line 200 "modules/sound/SoundData.cpp"
    $17 = (~~(($16))); //@line 200 "modules/sound/SoundData.cpp"
    $18 = (($15) + ($i<<1)|0); //@line 200 "modules/sound/SoundData.cpp"
    HEAP16[$18>>1] = $17; //@line 200 "modules/sound/SoundData.cpp"
    STACKTOP = sp;return; //@line 207 "modules/sound/SoundData.cpp"
   } else {
    $19 = $sample * 127.0; //@line 205 "modules/sound/SoundData.cpp"
    $20 = $19 + 128.0; //@line 205 "modules/sound/SoundData.cpp"
    $21 = (~~(($20))&255); //@line 205 "modules/sound/SoundData.cpp"
    $22 = ((($this)) + 8|0); //@line 205 "modules/sound/SoundData.cpp"
    $23 = HEAP32[$22>>2]|0; //@line 205 "modules/sound/SoundData.cpp"
    $24 = (($23) + ($i)|0); //@line 205 "modules/sound/SoundData.cpp"
    HEAP8[$24>>0] = $21; //@line 205 "modules/sound/SoundData.cpp"
    STACKTOP = sp;return; //@line 207 "modules/sound/SoundData.cpp"
   }
  }
 }
 $8 = (___cxa_allocate_exception(16)|0); //@line 194 "modules/sound/SoundData.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($8|0),(359443|0),($vararg_buffer|0)); //@line 194 "modules/sound/SoundData.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $11 = ___cxa_find_matching_catch_2()|0;
  $12 = tempRet0;
  ___cxa_free_exception(($8|0)); //@line 194 "modules/sound/SoundData.cpp"
  ___resumeException($11|0);
  // unreachable;
 } else {
  ___cxa_throw(($8|0),(96|0),(344|0)); //@line 194 "modules/sound/SoundData.cpp"
  // unreachable; //@line 194 "modules/sound/SoundData.cpp"
 }
}
function __ZNK4love5sound9SoundData9getSampleEi($this,$i) {
 $this = $this|0;
 $i = $i|0;
 var $$0 = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0.0, $25 = 0.0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($i|0)<(0); //@line 212 "modules/sound/SoundData.cpp"
 if (!($0)) {
  $1 = ((($this)) + 12|0); //@line 212 "modules/sound/SoundData.cpp"
  $2 = HEAP32[$1>>2]|0; //@line 212 "modules/sound/SoundData.cpp"
  $3 = ((($this)) + 20|0); //@line 212 "modules/sound/SoundData.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 212 "modules/sound/SoundData.cpp"
  $5 = (($4|0) / 8)&-1; //@line 212 "modules/sound/SoundData.cpp"
  $6 = (($2>>>0) / ($5>>>0))&-1; //@line 212 "modules/sound/SoundData.cpp"
  $7 = ($6>>>0)>($i>>>0); //@line 212 "modules/sound/SoundData.cpp"
  if ($7) {
   $13 = ($4|0)==(16); //@line 215 "modules/sound/SoundData.cpp"
   $14 = ((($this)) + 8|0); //@line 218 "modules/sound/SoundData.cpp"
   if ($13) {
    $15 = HEAP32[$14>>2]|0; //@line 218 "modules/sound/SoundData.cpp"
    $16 = (($15) + ($i<<1)|0); //@line 219 "modules/sound/SoundData.cpp"
    $17 = HEAP16[$16>>1]|0; //@line 219 "modules/sound/SoundData.cpp"
    $18 = (+($17<<16>>16)); //@line 219 "modules/sound/SoundData.cpp"
    $19 = $18 / 32767.0; //@line 219 "modules/sound/SoundData.cpp"
    $$0 = $19;
    STACKTOP = sp;return (+$$0); //@line 226 "modules/sound/SoundData.cpp"
   } else {
    $20 = HEAP32[$14>>2]|0; //@line 224 "modules/sound/SoundData.cpp"
    $21 = (($20) + ($i)|0); //@line 224 "modules/sound/SoundData.cpp"
    $22 = HEAP8[$21>>0]|0; //@line 224 "modules/sound/SoundData.cpp"
    $23 = (+($22&255)); //@line 224 "modules/sound/SoundData.cpp"
    $24 = $23 + -128.0; //@line 224 "modules/sound/SoundData.cpp"
    $25 = $24 / 127.0; //@line 224 "modules/sound/SoundData.cpp"
    $$0 = $25;
    STACKTOP = sp;return (+$$0); //@line 226 "modules/sound/SoundData.cpp"
   }
  }
 }
 $8 = (___cxa_allocate_exception(16)|0); //@line 213 "modules/sound/SoundData.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($8|0),(359479|0),($vararg_buffer|0)); //@line 213 "modules/sound/SoundData.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $11 = ___cxa_find_matching_catch_2()|0;
  $12 = tempRet0;
  ___cxa_free_exception(($8|0)); //@line 213 "modules/sound/SoundData.cpp"
  ___resumeException($11|0);
  // unreachable;
 } else {
  ___cxa_throw(($8|0),(96|0),(344|0)); //@line 213 "modules/sound/SoundData.cpp"
  // unreachable; //@line 213 "modules/sound/SoundData.cpp"
 }
 return +(0.0);
}
function __ZN4love5sound17luax_checkdecoderEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5sound7DecoderEEEPT_P9lua_StateiNS_4TypeE($L,$idx,30)|0); //@line 30 "modules/sound/wrap_Decoder.cpp"
 return ($0|0); //@line 30 "modules/sound/wrap_Decoder.cpp"
}
function __ZN4love14luax_checktypeINS_5sound7DecoderEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love5sound21w_Decoder_getChannelsEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5sound7DecoderEEEPT_P9lua_StateiNS_4TypeE($L,1,30)|0); //@line 30 "modules/sound/wrap_Decoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 36 "modules/sound/wrap_Decoder.cpp"
 $2 = ((($1)) + 48|0); //@line 36 "modules/sound/wrap_Decoder.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 36 "modules/sound/wrap_Decoder.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 36 "modules/sound/wrap_Decoder.cpp"
 _lua_pushinteger($L,$4); //@line 36 "modules/sound/wrap_Decoder.cpp"
 return 1; //@line 37 "modules/sound/wrap_Decoder.cpp"
}
function __ZN4love5sound21w_Decoder_getBitDepthEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5sound7DecoderEEEPT_P9lua_StateiNS_4TypeE($L,1,30)|0); //@line 30 "modules/sound/wrap_Decoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 43 "modules/sound/wrap_Decoder.cpp"
 $2 = ((($1)) + 52|0); //@line 43 "modules/sound/wrap_Decoder.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 43 "modules/sound/wrap_Decoder.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 43 "modules/sound/wrap_Decoder.cpp"
 _lua_pushinteger($L,$4); //@line 43 "modules/sound/wrap_Decoder.cpp"
 return 1; //@line 44 "modules/sound/wrap_Decoder.cpp"
}
function __ZN4love5sound23w_Decoder_getSampleRateEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5sound7DecoderEEEPT_P9lua_StateiNS_4TypeE($L,1,30)|0); //@line 30 "modules/sound/wrap_Decoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 50 "modules/sound/wrap_Decoder.cpp"
 $2 = ((($1)) + 56|0); //@line 50 "modules/sound/wrap_Decoder.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 50 "modules/sound/wrap_Decoder.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 50 "modules/sound/wrap_Decoder.cpp"
 _lua_pushinteger($L,$4); //@line 50 "modules/sound/wrap_Decoder.cpp"
 return 1; //@line 51 "modules/sound/wrap_Decoder.cpp"
}
function __ZN4love5sound21w_Decoder_getDurationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_5sound7DecoderEEEPT_P9lua_StateiNS_4TypeE($L,1,30)|0); //@line 30 "modules/sound/wrap_Decoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 57 "modules/sound/wrap_Decoder.cpp"
 $2 = ((($1)) + 60|0); //@line 57 "modules/sound/wrap_Decoder.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 57 "modules/sound/wrap_Decoder.cpp"
 $4 = (+FUNCTION_TABLE_di[$3 & 4095]($0)); //@line 57 "modules/sound/wrap_Decoder.cpp"
 _lua_pushnumber($L,$4); //@line 57 "modules/sound/wrap_Decoder.cpp"
 return 1; //@line 58 "modules/sound/wrap_Decoder.cpp"
}
function _luaopen_decoder($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 70 "modules/sound/wrap_Decoder.cpp"
 HEAP32[$vararg_buffer>>2] = 19140; //@line 72 "modules/sound/wrap_Decoder.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 72 "modules/sound/wrap_Decoder.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 72 "modules/sound/wrap_Decoder.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,30,359515,$vararg_buffer)|0); //@line 72 "modules/sound/wrap_Decoder.cpp"
 STACKTOP = sp;return ($0|0); //@line 72 "modules/sound/wrap_Decoder.cpp"
}
function __ZN4love6system3sdl6SystemD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ModuleD2Ev($this); //@line 43 "modules/system/sdl/System.h"
 __ZdlPv($this); //@line 43 "modules/system/sdl/System.h"
 return; //@line 43 "modules/system/sdl/System.h"
}
function __ZNK4love6system6System13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 12; //@line 55 "./modules/system/System.h"
}
function __ZNK4love6system3sdl6System7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (359599|0); //@line 41 "modules/system/sdl/System.cpp"
}
function __ZNK4love6system3sdl6System17getProcessorCountEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GetCPUCount()|0); //@line 46 "modules/system/sdl/System.cpp"
 return ($0|0); //@line 46 "modules/system/sdl/System.cpp"
}
function __ZNK4love6system3sdl6System16setClipboardTextERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($this,$text) {
 $this = $this|0;
 $text = $text|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$text>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $1 = $0 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = ($1<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ((($text)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = HEAP32[$3>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($text)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = $2 ? $5 : $4; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 (_SDL_SetClipboardText($6)|0); //@line 51 "modules/system/sdl/System.cpp"
 return; //@line 52 "modules/system/sdl/System.cpp"
}
function __ZNK4love6system3sdl6System16getClipboardTextEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($agg$result,1282466,0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 $1 = (invoke_i(3774)|0); //@line 58 "modules/system/sdl/System.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($agg$result); //@line 66 "modules/system/sdl/System.cpp"
  ___resumeException($21|0);
  // unreachable;
 }
 $4 = ($1|0)==(0|0); //@line 59 "modules/system/sdl/System.cpp"
 if ($4) {
  STACKTOP = sp;return; //@line 66 "modules/system/sdl/System.cpp"
 }
 $5 = (_strlen($1)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viii(3460,($0|0),($1|0),($5|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($agg$result); //@line 66 "modules/system/sdl/System.cpp"
  ___resumeException($21|0);
  // unreachable;
 }
 $8 = HEAP8[$agg$result>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $8 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = ($9<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($10) {
  $14 = ((($agg$result)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$14>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$agg$result>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 } else {
  $11 = ((($agg$result)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $12 = HEAP32[$11>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$12>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $13 = ((($agg$result)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$13>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 __THREW__ = 0;
 invoke_vii(3368,($agg$result|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if ($16) {
  $17 = ___cxa_find_matching_catch_3(0|0)|0;
  $18 = tempRet0;
  ___clang_call_terminate($17); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 ;HEAP32[$agg$result>>2]=HEAP32[$0>>2]|0;HEAP32[$agg$result+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$agg$result+8>>2]=HEAP32[$0+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 61 "modules/system/sdl/System.cpp"
 __THREW__ = 0;
 invoke_vi(3775,($1|0)); //@line 62 "modules/system/sdl/System.cpp"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($agg$result); //@line 66 "modules/system/sdl/System.cpp"
  ___resumeException($21|0);
  // unreachable;
 } else {
  STACKTOP = sp;return; //@line 66 "modules/system/sdl/System.cpp"
 }
}
function __ZNK4love6system3sdl6System12getPowerInfoERiS3_($this,$seconds,$percent) {
 $this = $this|0;
 $seconds = $seconds|0;
 $percent = $percent|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $state$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GetPowerInfo($seconds,$percent)|0); //@line 70 "modules/system/sdl/System.cpp"
 $1 = ($0>>>0)<(5); //@line 75 "./common/EnumMap.h"
 if ($1) {
  $2 = (((631392 + ($0<<3)|0)) + 4|0); //@line 75 "./common/EnumMap.h"
  $3 = HEAP8[$2>>0]|0; //@line 75 "./common/EnumMap.h"
  $4 = ($3<<24>>24)==(0); //@line 75 "./common/EnumMap.h"
  if ($4) {
   $state$0 = 0;
  } else {
   $5 = (631392 + ($0<<3)|0); //@line 77 "./common/EnumMap.h"
   $6 = HEAP32[$5>>2]|0; //@line 77 "./common/EnumMap.h"
   $state$0 = $6;
  }
 } else {
  $state$0 = 0;
 }
 return ($state$0|0); //@line 75 "modules/system/sdl/System.cpp"
}
function __GLOBAL__sub_I_System_cpp() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[(631396)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(631404)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(631412)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(631420)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(631428)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(631436)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(631444)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(631452)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(631460)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(631468)>>0] = 0; //@line 90 "./common/EnumMap.h"
 $0 = HEAP32[4809]|0; //@line 46 "./common/EnumMap.h"
 $1 = HEAP32[(19240)>>2]|0; //@line 47 "./common/EnumMap.h"
 $2 = ($0>>>0)<(5); //@line 49 "./common/EnumMap.h"
 if ($2) {
  $3 = (631432 + ($0<<3)|0); //@line 51 "./common/EnumMap.h"
  HEAP32[$3>>2] = $1; //@line 51 "./common/EnumMap.h"
  $4 = (((631432 + ($0<<3)|0)) + 4|0); //@line 52 "./common/EnumMap.h"
  HEAP8[$4>>0] = 1; //@line 52 "./common/EnumMap.h"
 }
 $5 = ($1>>>0)<(5); //@line 54 "./common/EnumMap.h"
 if ($5) {
  $6 = (631392 + ($1<<3)|0); //@line 56 "./common/EnumMap.h"
  HEAP32[$6>>2] = $0; //@line 56 "./common/EnumMap.h"
  $7 = (((631392 + ($1<<3)|0)) + 4|0); //@line 57 "./common/EnumMap.h"
  HEAP8[$7>>0] = 1; //@line 57 "./common/EnumMap.h"
 }
 $8 = HEAP32[(19244)>>2]|0; //@line 46 "./common/EnumMap.h"
 $9 = HEAP32[(19248)>>2]|0; //@line 47 "./common/EnumMap.h"
 $10 = ($8>>>0)<(5); //@line 49 "./common/EnumMap.h"
 if ($10) {
  $11 = (631432 + ($8<<3)|0); //@line 51 "./common/EnumMap.h"
  HEAP32[$11>>2] = $9; //@line 51 "./common/EnumMap.h"
  $12 = (((631432 + ($8<<3)|0)) + 4|0); //@line 52 "./common/EnumMap.h"
  HEAP8[$12>>0] = 1; //@line 52 "./common/EnumMap.h"
 }
 $13 = ($9>>>0)<(5); //@line 54 "./common/EnumMap.h"
 if ($13) {
  $14 = (631392 + ($9<<3)|0); //@line 56 "./common/EnumMap.h"
  HEAP32[$14>>2] = $8; //@line 56 "./common/EnumMap.h"
  $15 = (((631392 + ($9<<3)|0)) + 4|0); //@line 57 "./common/EnumMap.h"
  HEAP8[$15>>0] = 1; //@line 57 "./common/EnumMap.h"
 }
 $16 = HEAP32[(19252)>>2]|0; //@line 46 "./common/EnumMap.h"
 $17 = HEAP32[(19256)>>2]|0; //@line 47 "./common/EnumMap.h"
 $18 = ($16>>>0)<(5); //@line 49 "./common/EnumMap.h"
 if ($18) {
  $19 = (631432 + ($16<<3)|0); //@line 51 "./common/EnumMap.h"
  HEAP32[$19>>2] = $17; //@line 51 "./common/EnumMap.h"
  $20 = (((631432 + ($16<<3)|0)) + 4|0); //@line 52 "./common/EnumMap.h"
  HEAP8[$20>>0] = 1; //@line 52 "./common/EnumMap.h"
 }
 $21 = ($17>>>0)<(5); //@line 54 "./common/EnumMap.h"
 if ($21) {
  $22 = (631392 + ($17<<3)|0); //@line 56 "./common/EnumMap.h"
  HEAP32[$22>>2] = $16; //@line 56 "./common/EnumMap.h"
  $23 = (((631392 + ($17<<3)|0)) + 4|0); //@line 57 "./common/EnumMap.h"
  HEAP8[$23>>0] = 1; //@line 57 "./common/EnumMap.h"
 }
 $24 = HEAP32[(19260)>>2]|0; //@line 46 "./common/EnumMap.h"
 $25 = HEAP32[(19264)>>2]|0; //@line 47 "./common/EnumMap.h"
 $26 = ($24>>>0)<(5); //@line 49 "./common/EnumMap.h"
 if ($26) {
  $27 = (631432 + ($24<<3)|0); //@line 51 "./common/EnumMap.h"
  HEAP32[$27>>2] = $25; //@line 51 "./common/EnumMap.h"
  $28 = (((631432 + ($24<<3)|0)) + 4|0); //@line 52 "./common/EnumMap.h"
  HEAP8[$28>>0] = 1; //@line 52 "./common/EnumMap.h"
 }
 $29 = ($25>>>0)<(5); //@line 54 "./common/EnumMap.h"
 if ($29) {
  $30 = (631392 + ($25<<3)|0); //@line 56 "./common/EnumMap.h"
  HEAP32[$30>>2] = $24; //@line 56 "./common/EnumMap.h"
  $31 = (((631392 + ($25<<3)|0)) + 4|0); //@line 57 "./common/EnumMap.h"
  HEAP8[$31>>0] = 1; //@line 57 "./common/EnumMap.h"
 }
 $32 = HEAP32[(19268)>>2]|0; //@line 46 "./common/EnumMap.h"
 $33 = HEAP32[(19272)>>2]|0; //@line 47 "./common/EnumMap.h"
 $34 = ($32>>>0)<(5); //@line 49 "./common/EnumMap.h"
 if ($34) {
  $35 = (631432 + ($32<<3)|0); //@line 51 "./common/EnumMap.h"
  HEAP32[$35>>2] = $33; //@line 51 "./common/EnumMap.h"
  $36 = (((631432 + ($32<<3)|0)) + 4|0); //@line 52 "./common/EnumMap.h"
  HEAP8[$36>>0] = 1; //@line 52 "./common/EnumMap.h"
 }
 $37 = ($33>>>0)<(5); //@line 54 "./common/EnumMap.h"
 if (!($37)) {
  return;
 }
 $38 = (631392 + ($33<<3)|0); //@line 56 "./common/EnumMap.h"
 HEAP32[$38>>2] = $32; //@line 56 "./common/EnumMap.h"
 $39 = (((631392 + ($33<<3)|0)) + 4|0); //@line 57 "./common/EnumMap.h"
 HEAP8[$39>>0] = 1; //@line 57 "./common/EnumMap.h"
 return;
}
function __ZN4love6system3sdl6SystemC2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6system6SystemC2Ev($this); //@line 35 "modules/system/sdl/System.cpp"
 HEAP32[$this>>2] = (19188); //@line 36 "modules/system/sdl/System.cpp"
 return; //@line 37 "modules/system/sdl/System.cpp"
}
function __ZN4love6system6SystemD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ModuleD2Ev($this); //@line 52 "modules/system/System.h"
 __ZdlPv($this); //@line 52 "modules/system/System.h"
 return; //@line 52 "modules/system/System.h"
}
function __ZNK4love6system6System7openURLERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE($this,$url) {
 $this = $this|0;
 $url = $url|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 // unreachable; //@line 88 "modules/system/System.cpp"
 return (0)|0;
}
function __ZNK4love6system6System7vibrateEd($this,$seconds) {
 $this = $this|0;
 $seconds = +$seconds;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 156 "modules/system/System.cpp"
}
function __GLOBAL__sub_I_System_cpp_2273() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[(631480)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(631492)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(631504)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(631516)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(631528)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(631540)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(631552)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(631564)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(631576)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(631588)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(631592)>>2]=0|0;HEAP32[(631592)+4>>2]=0|0;HEAP32[(631592)+8>>2]=0|0;HEAP32[(631592)+12>>2]=0|0;HEAP32[(631592)+16>>2]=0|0; //@line 44 "./common/StringMap.h"
 $0 = HEAP32[4833]|0; //@line 49 "./common/StringMap.h"
 $1 = HEAP32[(19336)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_6system6System10PowerStateELj5EE3addEPKcS3_(631472,$0,$1)|0); //@line 49 "./common/StringMap.h"
 $2 = HEAP32[(19340)>>2]|0; //@line 49 "./common/StringMap.h"
 $3 = HEAP32[(19344)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_6system6System10PowerStateELj5EE3addEPKcS3_(631472,$2,$3)|0); //@line 49 "./common/StringMap.h"
 $4 = HEAP32[(19348)>>2]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(19352)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_6system6System10PowerStateELj5EE3addEPKcS3_(631472,$4,$5)|0); //@line 49 "./common/StringMap.h"
 $6 = HEAP32[(19356)>>2]|0; //@line 49 "./common/StringMap.h"
 $7 = HEAP32[(19360)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_6system6System10PowerStateELj5EE3addEPKcS3_(631472,$6,$7)|0); //@line 49 "./common/StringMap.h"
 $8 = HEAP32[(19364)>>2]|0; //@line 49 "./common/StringMap.h"
 $9 = HEAP32[(19368)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_6system6System10PowerStateELj5EE3addEPKcS3_(631472,$8,$9)|0); //@line 49 "./common/StringMap.h"
 return;
}
function __ZN4love9StringMapINS_6system6System10PowerStateELj5EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $42 = (($hash$0$lcssa$i) + 6)|0; //@line 112 "./common/StringMap.h"
       $43 = (($42>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
       $44 = (((($this) + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
       $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
       $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
       if ($46) {
        $$lcssa = $43;$$lcssa10 = $44;
        label = 5;
       } else {
        $47 = (($hash$0$lcssa$i) + 7)|0; //@line 112 "./common/StringMap.h"
        $48 = (($47>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
        $49 = (((($this) + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
        $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
        $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
        if ($51) {
         $$lcssa = $48;$$lcssa10 = $49;
         label = 5;
        } else {
         $52 = (($hash$0$lcssa$i) + 8)|0; //@line 112 "./common/StringMap.h"
         $53 = (($52>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
         $54 = (((($this) + (($53*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
         $55 = HEAP8[$54>>0]|0; //@line 114 "./common/StringMap.h"
         $56 = ($55<<24>>24)==(0); //@line 114 "./common/StringMap.h"
         if ($56) {
          $$lcssa = $53;$$lcssa10 = $54;
          label = 5;
         } else {
          $57 = (($hash$0$lcssa$i) + 9)|0; //@line 112 "./common/StringMap.h"
          $58 = (($57>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
          $59 = (((($this) + (($58*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
          $60 = HEAP8[$59>>0]|0; //@line 114 "./common/StringMap.h"
          $61 = ($60<<24>>24)==(0); //@line 114 "./common/StringMap.h"
          if ($61) {
           $$lcssa = $58;$$lcssa10 = $59;
           label = 5;
          } else {
           $inserted$2$off0 = 0;
          }
         }
        }
       }
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(4); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 120|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love6system6SystemC2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./common/Module.h"
 HEAP32[$this>>2] = (19284); //@line 50 "modules/system/System.cpp"
 return; //@line 61 "modules/system/System.cpp"
}
function __ZNK4love6system6System5getOSEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($agg$result,359672,3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 return; //@line 66 "modules/system/System.cpp"
}
function __ZN4love6system6System11getConstantENS1_10PowerStateERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(4); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (631592 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 165 "modules/system/System.cpp"
}
function __ZN4love6thread12ThreadModuleD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ModuleD2Ev($this); //@line 45 "modules/thread/ThreadModule.h"
 __ZdlPv($this); //@line 45 "modules/thread/ThreadModule.h"
 return; //@line 45 "modules/thread/ThreadModule.h"
}
function __ZNK4love6thread12ThreadModule13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 13; //@line 52 "modules/thread/ThreadModule.h"
}
function __ZNK4love6thread12ThreadModule7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (359705|0); //@line 46 "modules/thread/ThreadModule.cpp"
}
function __ZN4love6thread12ThreadModule9newThreadERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEEPNS_4DataE($this,$name,$data) {
 $this = $this|0;
 $name = $name|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(64)|0); //@line 30 "modules/thread/ThreadModule.cpp"
 __THREW__ = 0;
 invoke_viii(3776,($0|0),($name|0),($data|0)); //@line 30 "modules/thread/ThreadModule.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 30 "modules/thread/ThreadModule.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 31 "modules/thread/ThreadModule.cpp"
 }
 return (0)|0;
}
function __ZN4love6thread12ThreadModule10newChannelEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(64)|0); //@line 36 "modules/thread/ThreadModule.cpp"
 __THREW__ = 0;
 invoke_vi(3777,($0|0)); //@line 36 "modules/thread/ThreadModule.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 36 "modules/thread/ThreadModule.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 36 "modules/thread/ThreadModule.cpp"
 }
 return (0)|0;
}
function __ZN4love6thread12ThreadModule10getChannelERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE($this,$name) {
 $this = $this|0;
 $name = $name|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love6thread7Channel10getChannelERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE($name)|0); //@line 41 "modules/thread/ThreadModule.cpp"
 return ($0|0); //@line 41 "modules/thread/ThreadModule.cpp"
}
function __ZN4love6thread10ThreadableD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19424); //@line 83 "modules/thread/threads.cpp"
 $0 = ((($this)) + 4|0); //@line 84 "modules/thread/threads.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 84 "modules/thread/threads.cpp"
 $2 = ($1|0)==(0|0); //@line 84 "modules/thread/threads.cpp"
 if (!($2)) {
  $3 = HEAP32[$1>>2]|0; //@line 84 "modules/thread/threads.cpp"
  $4 = ((($3)) + 4|0); //@line 84 "modules/thread/threads.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 84 "modules/thread/threads.cpp"
  FUNCTION_TABLE_vi[$5 & 8191]($1); //@line 84 "modules/thread/threads.cpp"
 }
 $6 = ((($this)) + 8|0); //@line 85 "modules/thread/threads.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 85 "modules/thread/threads.cpp"
 return; //@line 85 "modules/thread/threads.cpp"
}
function __ZN4love6thread10ThreadableD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19424); //@line 83 "modules/thread/threads.cpp"
 $0 = ((($this)) + 4|0); //@line 84 "modules/thread/threads.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 84 "modules/thread/threads.cpp"
 $2 = ($1|0)==(0|0); //@line 84 "modules/thread/threads.cpp"
 if (!($2)) {
  $3 = HEAP32[$1>>2]|0; //@line 84 "modules/thread/threads.cpp"
  $4 = ((($3)) + 4|0); //@line 84 "modules/thread/threads.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 84 "modules/thread/threads.cpp"
  FUNCTION_TABLE_vi[$5 & 8191]($1); //@line 84 "modules/thread/threads.cpp"
 }
 $6 = ((($this)) + 8|0); //@line 85 "modules/thread/threads.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 85 "modules/thread/threads.cpp"
 __ZdlPv($this); //@line 83 "modules/thread/threads.cpp"
 return; //@line 85 "modules/thread/threads.cpp"
}
function __ZN4love6thread4LockC2EPNS0_5MutexE($this,$m) {
 $this = $this|0;
 $m = $m|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = $m; //@line 29 "modules/thread/threads.cpp"
 $0 = HEAP32[$m>>2]|0; //@line 31 "modules/thread/threads.cpp"
 $1 = ((($0)) + 8|0); //@line 31 "modules/thread/threads.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 31 "modules/thread/threads.cpp"
 FUNCTION_TABLE_vi[$2 & 8191]($m); //@line 31 "modules/thread/threads.cpp"
 return; //@line 32 "modules/thread/threads.cpp"
}
function __ZN4love6thread4LockC2ERNS0_5MutexE($this,$m) {
 $this = $this|0;
 $m = $m|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = $m; //@line 35 "modules/thread/threads.cpp"
 $0 = HEAP32[$m>>2]|0; //@line 37 "modules/thread/threads.cpp"
 $1 = ((($0)) + 8|0); //@line 37 "modules/thread/threads.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 37 "modules/thread/threads.cpp"
 FUNCTION_TABLE_vi[$2 & 8191]($m); //@line 37 "modules/thread/threads.cpp"
 return; //@line 38 "modules/thread/threads.cpp"
}
function __ZN4love6thread4LockD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 42 "modules/thread/threads.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 42 "modules/thread/threads.cpp"
 $2 = ((($1)) + 12|0); //@line 42 "modules/thread/threads.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 42 "modules/thread/threads.cpp"
 __THREW__ = 0;
 invoke_vi($3|0,($0|0)); //@line 42 "modules/thread/threads.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_3(0|0)|0;
  $7 = tempRet0;
  ___clang_call_terminate($6); //@line 42 "modules/thread/threads.cpp"
  // unreachable; //@line 42 "modules/thread/threads.cpp"
 } else {
  return; //@line 43 "modules/thread/threads.cpp"
 }
}
function __ZN4love6thread9EmptyLockC2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = 0; //@line 46 "modules/thread/threads.cpp"
 return; //@line 48 "modules/thread/threads.cpp"
}
function __ZN4love6thread9EmptyLockD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 52 "modules/thread/threads.cpp"
 $1 = ($0|0)==(0|0); //@line 52 "modules/thread/threads.cpp"
 if ($1) {
  return; //@line 54 "modules/thread/threads.cpp"
 }
 $2 = HEAP32[$0>>2]|0; //@line 53 "modules/thread/threads.cpp"
 $3 = ((($2)) + 12|0); //@line 53 "modules/thread/threads.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 53 "modules/thread/threads.cpp"
 __THREW__ = 0;
 invoke_vi($4|0,($0|0)); //@line 53 "modules/thread/threads.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_3(0|0)|0;
  $8 = tempRet0;
  ___clang_call_terminate($7); //@line 53 "modules/thread/threads.cpp"
  // unreachable; //@line 53 "modules/thread/threads.cpp"
 } else {
  return; //@line 54 "modules/thread/threads.cpp"
 }
}
function __ZN4love6thread9EmptyLock7setLockEPNS0_5MutexE($this,$m) {
 $this = $this|0;
 $m = $m|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($m|0)==(0|0); //@line 58 "modules/thread/threads.cpp"
 if (!($0)) {
  $1 = HEAP32[$m>>2]|0; //@line 59 "modules/thread/threads.cpp"
  $2 = ((($1)) + 8|0); //@line 59 "modules/thread/threads.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 59 "modules/thread/threads.cpp"
  FUNCTION_TABLE_vi[$3 & 8191]($m); //@line 59 "modules/thread/threads.cpp"
 }
 $4 = HEAP32[$this>>2]|0; //@line 61 "modules/thread/threads.cpp"
 $5 = ($4|0)==(0|0); //@line 61 "modules/thread/threads.cpp"
 if ($5) {
  HEAP32[$this>>2] = $m; //@line 64 "modules/thread/threads.cpp"
  return; //@line 65 "modules/thread/threads.cpp"
 }
 $6 = HEAP32[$4>>2]|0; //@line 62 "modules/thread/threads.cpp"
 $7 = ((($6)) + 12|0); //@line 62 "modules/thread/threads.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 62 "modules/thread/threads.cpp"
 FUNCTION_TABLE_vi[$8 & 8191]($4); //@line 62 "modules/thread/threads.cpp"
 HEAP32[$this>>2] = $m; //@line 64 "modules/thread/threads.cpp"
 return; //@line 65 "modules/thread/threads.cpp"
}
function __ZN4love6thread10ThreadableC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19424); //@line 78 "modules/thread/threads.cpp"
 $0 = ((($this)) + 8|0); //@line 77 "modules/thread/threads.cpp"
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 $1 = (invoke_ii(3778,($this|0))|0); //@line 79 "modules/thread/threads.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 80 "modules/thread/threads.cpp"
  ___resumeException($5|0);
  // unreachable;
 } else {
  $4 = ((($this)) + 4|0); //@line 79 "modules/thread/threads.cpp"
  HEAP32[$4>>2] = $1; //@line 79 "modules/thread/threads.cpp"
  return; //@line 80 "modules/thread/threads.cpp"
 }
}
function __ZN4love6thread10Threadable5startEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 89 "modules/thread/threads.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 89 "modules/thread/threads.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 89 "modules/thread/threads.cpp"
 $3 = ((($2)) + 8|0); //@line 89 "modules/thread/threads.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 89 "modules/thread/threads.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($1)|0); //@line 89 "modules/thread/threads.cpp"
 return ($5|0); //@line 89 "modules/thread/threads.cpp"
}
function __ZN4love6thread10Threadable4waitEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 94 "modules/thread/threads.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 94 "modules/thread/threads.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 94 "modules/thread/threads.cpp"
 $3 = ((($2)) + 12|0); //@line 94 "modules/thread/threads.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 94 "modules/thread/threads.cpp"
 FUNCTION_TABLE_vi[$4 & 8191]($1); //@line 94 "modules/thread/threads.cpp"
 return; //@line 95 "modules/thread/threads.cpp"
}
function __ZNK4love6thread10Threadable9isRunningEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 99 "modules/thread/threads.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 99 "modules/thread/threads.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 99 "modules/thread/threads.cpp"
 $3 = ((($2)) + 16|0); //@line 99 "modules/thread/threads.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 99 "modules/thread/threads.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($1)|0); //@line 99 "modules/thread/threads.cpp"
 return ($5|0); //@line 99 "modules/thread/threads.cpp"
}
function __ZNK4love6thread10Threadable13getThreadNameEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 104 "modules/thread/threads.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = $1 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ($2<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($3) {
  $6 = $1&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = $6 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = $7;
 } else {
  $4 = ((($this)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $5 = HEAP32[$4>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = $5;
 }
 $8 = ($9|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($8) {
  $13 = 0;
  return ($13|0); //@line 104 "modules/thread/threads.cpp"
 }
 if ($3) {
  $12 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $13 = $12;
  return ($13|0); //@line 104 "modules/thread/threads.cpp"
 } else {
  $10 = ((($this)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $11 = HEAP32[$10>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $13 = $11;
  return ($13|0); //@line 104 "modules/thread/threads.cpp"
 }
 return (0)|0;
}
function __ZN4love6thread8MutexRefC2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love6thread8newMutexEv()|0); //@line 108 "modules/thread/threads.cpp"
 HEAP32[$this>>2] = $0; //@line 108 "modules/thread/threads.cpp"
 return; //@line 110 "modules/thread/threads.cpp"
}
function __ZN4love6thread8MutexRefD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 114 "modules/thread/threads.cpp"
 $1 = ($0|0)==(0|0); //@line 114 "modules/thread/threads.cpp"
 if ($1) {
  return; //@line 115 "modules/thread/threads.cpp"
 }
 $2 = HEAP32[$0>>2]|0; //@line 114 "modules/thread/threads.cpp"
 $3 = ((($2)) + 4|0); //@line 114 "modules/thread/threads.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 114 "modules/thread/threads.cpp"
 FUNCTION_TABLE_vi[$4 & 8191]($0); //@line 114 "modules/thread/threads.cpp"
 return; //@line 115 "modules/thread/threads.cpp"
}
function __ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 119 "modules/thread/threads.cpp"
 return ($0|0); //@line 119 "modules/thread/threads.cpp"
}
function __ZN4love14luax_checktypeINS_6thread7ChannelEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love6thread14w_Channel_pushEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_6thread7ChannelEEEPT_P9lua_StateiNS_4TypeE($L,1,54)|0); //@line 44 "modules/thread/wrap_Channel.cpp"
 $1 = (_lua_type($L,2)|0); //@line 50 "modules/thread/wrap_Channel.cpp"
 $2 = ($1|0)<(1); //@line 50 "modules/thread/wrap_Channel.cpp"
 if (!($2)) {
  $3 = (__ZN4love7Variant7fromLuaEP9lua_Stateib($L,2,1)|0); //@line 50 "modules/thread/wrap_Channel.cpp"
  $4 = ($3|0)==(0|0); //@line 51 "modules/thread/wrap_Channel.cpp"
  if (!($4)) {
   (__ZN4love6thread7Channel4pushEPNS_7VariantE($0,$3)|0); //@line 53 "modules/thread/wrap_Channel.cpp"
   __ZN4love6thread7Channel9lockMutexEv($0); //@line 37 "modules/thread/wrap_Channel.cpp"
   $6 = HEAP32[$3>>2]|0; //@line 38 "modules/thread/wrap_Channel.cpp"
   $7 = ((($6)) + 12|0); //@line 38 "modules/thread/wrap_Channel.cpp"
   $8 = HEAP32[$7>>2]|0; //@line 38 "modules/thread/wrap_Channel.cpp"
   FUNCTION_TABLE_vi[$8 & 8191]($3); //@line 38 "modules/thread/wrap_Channel.cpp"
   __ZN4love6thread7Channel11unlockMutexEv($0); //@line 39 "modules/thread/wrap_Channel.cpp"
   $$0 = 0;
   return ($$0|0); //@line 56 "modules/thread/wrap_Channel.cpp"
  }
 }
 $5 = (_luaL_argerror($L,2,359807)|0); //@line 52 "modules/thread/wrap_Channel.cpp"
 $$0 = $5;
 return ($$0|0); //@line 56 "modules/thread/wrap_Channel.cpp"
}
function __ZN4love6thread16w_Channel_supplyEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_6thread7ChannelEEEPT_P9lua_StateiNS_4TypeE($L,1,54)|0); //@line 44 "modules/thread/wrap_Channel.cpp"
 $1 = (_lua_type($L,2)|0); //@line 61 "modules/thread/wrap_Channel.cpp"
 $2 = ($1|0)<(1); //@line 61 "modules/thread/wrap_Channel.cpp"
 if (!($2)) {
  $3 = (__ZN4love7Variant7fromLuaEP9lua_Stateib($L,2,1)|0); //@line 61 "modules/thread/wrap_Channel.cpp"
  $4 = ($3|0)==(0|0); //@line 62 "modules/thread/wrap_Channel.cpp"
  if (!($4)) {
   __ZN4love6thread7Channel6supplyEPNS_7VariantE($0,$3); //@line 64 "modules/thread/wrap_Channel.cpp"
   __ZN4love6thread7Channel9lockMutexEv($0); //@line 37 "modules/thread/wrap_Channel.cpp"
   $6 = HEAP32[$3>>2]|0; //@line 38 "modules/thread/wrap_Channel.cpp"
   $7 = ((($6)) + 12|0); //@line 38 "modules/thread/wrap_Channel.cpp"
   $8 = HEAP32[$7>>2]|0; //@line 38 "modules/thread/wrap_Channel.cpp"
   FUNCTION_TABLE_vi[$8 & 8191]($3); //@line 38 "modules/thread/wrap_Channel.cpp"
   __ZN4love6thread7Channel11unlockMutexEv($0); //@line 39 "modules/thread/wrap_Channel.cpp"
   $$0 = 0;
   return ($$0|0); //@line 67 "modules/thread/wrap_Channel.cpp"
  }
 }
 $5 = (_luaL_argerror($L,2,359807)|0); //@line 63 "modules/thread/wrap_Channel.cpp"
 $$0 = $5;
 return ($$0|0); //@line 67 "modules/thread/wrap_Channel.cpp"
}
function __ZN4love6thread13w_Channel_popEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_6thread7ChannelEEEPT_P9lua_StateiNS_4TypeE($L,1,54)|0); //@line 44 "modules/thread/wrap_Channel.cpp"
 $1 = (__ZN4love6thread7Channel3popEv($0)|0); //@line 72 "modules/thread/wrap_Channel.cpp"
 $2 = ($1|0)==(0|0); //@line 73 "modules/thread/wrap_Channel.cpp"
 if ($2) {
  _lua_pushnil($L); //@line 79 "modules/thread/wrap_Channel.cpp"
  return 1; //@line 80 "modules/thread/wrap_Channel.cpp"
 } else {
  __ZN4love7Variant5toLuaEP9lua_State($1,$L); //@line 75 "modules/thread/wrap_Channel.cpp"
  __ZN4love6thread7Channel9lockMutexEv($0); //@line 37 "modules/thread/wrap_Channel.cpp"
  $3 = HEAP32[$1>>2]|0; //@line 38 "modules/thread/wrap_Channel.cpp"
  $4 = ((($3)) + 12|0); //@line 38 "modules/thread/wrap_Channel.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 38 "modules/thread/wrap_Channel.cpp"
  FUNCTION_TABLE_vi[$5 & 8191]($1); //@line 38 "modules/thread/wrap_Channel.cpp"
  __ZN4love6thread7Channel11unlockMutexEv($0); //@line 39 "modules/thread/wrap_Channel.cpp"
  return 1; //@line 80 "modules/thread/wrap_Channel.cpp"
 }
 return (0)|0;
}
function __ZN4love6thread16w_Channel_demandEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_6thread7ChannelEEEPT_P9lua_StateiNS_4TypeE($L,1,54)|0); //@line 44 "modules/thread/wrap_Channel.cpp"
 $1 = (__ZN4love6thread7Channel6demandEv($0)|0); //@line 86 "modules/thread/wrap_Channel.cpp"
 __ZN4love7Variant5toLuaEP9lua_State($1,$L); //@line 87 "modules/thread/wrap_Channel.cpp"
 __ZN4love6thread7Channel9lockMutexEv($0); //@line 37 "modules/thread/wrap_Channel.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 38 "modules/thread/wrap_Channel.cpp"
 $3 = ((($2)) + 12|0); //@line 38 "modules/thread/wrap_Channel.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 38 "modules/thread/wrap_Channel.cpp"
 FUNCTION_TABLE_vi[$4 & 8191]($1); //@line 38 "modules/thread/wrap_Channel.cpp"
 __ZN4love6thread7Channel11unlockMutexEv($0); //@line 39 "modules/thread/wrap_Channel.cpp"
 return 1; //@line 89 "modules/thread/wrap_Channel.cpp"
}
function __ZN4love6thread14w_Channel_peekEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_6thread7ChannelEEEPT_P9lua_StateiNS_4TypeE($L,1,54)|0); //@line 44 "modules/thread/wrap_Channel.cpp"
 $1 = (__ZN4love6thread7Channel4peekEv($0)|0); //@line 95 "modules/thread/wrap_Channel.cpp"
 $2 = ($1|0)==(0|0); //@line 96 "modules/thread/wrap_Channel.cpp"
 if ($2) {
  _lua_pushnil($L); //@line 102 "modules/thread/wrap_Channel.cpp"
  return 1; //@line 103 "modules/thread/wrap_Channel.cpp"
 } else {
  __ZN4love7Variant5toLuaEP9lua_State($1,$L); //@line 98 "modules/thread/wrap_Channel.cpp"
  __ZN4love6thread7Channel9lockMutexEv($0); //@line 37 "modules/thread/wrap_Channel.cpp"
  $3 = HEAP32[$1>>2]|0; //@line 38 "modules/thread/wrap_Channel.cpp"
  $4 = ((($3)) + 12|0); //@line 38 "modules/thread/wrap_Channel.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 38 "modules/thread/wrap_Channel.cpp"
  FUNCTION_TABLE_vi[$5 & 8191]($1); //@line 38 "modules/thread/wrap_Channel.cpp"
  __ZN4love6thread7Channel11unlockMutexEv($0); //@line 39 "modules/thread/wrap_Channel.cpp"
  return 1; //@line 103 "modules/thread/wrap_Channel.cpp"
 }
 return (0)|0;
}
function __ZN4love6thread18w_Channel_getCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_6thread7ChannelEEEPT_P9lua_StateiNS_4TypeE($L,1,54)|0); //@line 44 "modules/thread/wrap_Channel.cpp"
 $1 = (__ZN4love6thread7Channel8getCountEv($0)|0); //@line 109 "modules/thread/wrap_Channel.cpp"
 $2 = (+($1|0)); //@line 109 "modules/thread/wrap_Channel.cpp"
 _lua_pushnumber($L,$2); //@line 109 "modules/thread/wrap_Channel.cpp"
 return 1; //@line 110 "modules/thread/wrap_Channel.cpp"
}
function __ZN4love6thread15w_Channel_clearEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_6thread7ChannelEEEPT_P9lua_StateiNS_4TypeE($L,1,54)|0); //@line 44 "modules/thread/wrap_Channel.cpp"
 __ZN4love6thread7Channel5clearEv($0); //@line 116 "modules/thread/wrap_Channel.cpp"
 return 0; //@line 117 "modules/thread/wrap_Channel.cpp"
}
function __ZN4love6thread23w_Channel_performAtomicEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_6thread7ChannelEEEPT_P9lua_StateiNS_4TypeE($L,1,54)|0); //@line 44 "modules/thread/wrap_Channel.cpp"
 _luaL_checktype($L,2,6); //@line 123 "modules/thread/wrap_Channel.cpp"
 _lua_pushvalue($L,1); //@line 126 "modules/thread/wrap_Channel.cpp"
 _lua_insert($L,3); //@line 127 "modules/thread/wrap_Channel.cpp"
 __ZN4love6thread7Channel9lockMutexEv($0); //@line 129 "modules/thread/wrap_Channel.cpp"
 $1 = (_lua_gettop($L)|0); //@line 133 "modules/thread/wrap_Channel.cpp"
 $2 = (($1) + -2)|0; //@line 133 "modules/thread/wrap_Channel.cpp"
 $3 = (_lua_pcall($L,$2,-1,0)|0); //@line 134 "modules/thread/wrap_Channel.cpp"
 __ZN4love6thread7Channel11unlockMutexEv($0); //@line 136 "modules/thread/wrap_Channel.cpp"
 $4 = ($3|0)==(0); //@line 139 "modules/thread/wrap_Channel.cpp"
 if ($4) {
  $6 = (_lua_gettop($L)|0); //@line 144 "modules/thread/wrap_Channel.cpp"
  $7 = (($6) + -1)|0; //@line 144 "modules/thread/wrap_Channel.cpp"
  $$0 = $7;
  return ($$0|0); //@line 145 "modules/thread/wrap_Channel.cpp"
 } else {
  $5 = (_lua_error($L)|0); //@line 140 "modules/thread/wrap_Channel.cpp"
  $$0 = $5;
  return ($$0|0); //@line 145 "modules/thread/wrap_Channel.cpp"
 }
 return (0)|0;
}
function _luaopen_channel($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 160 "modules/thread/wrap_Channel.cpp"
 HEAP32[$vararg_buffer>>2] = 19436; //@line 162 "modules/thread/wrap_Channel.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 162 "modules/thread/wrap_Channel.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 162 "modules/thread/wrap_Channel.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,54,359748,$vararg_buffer)|0); //@line 162 "modules/thread/wrap_Channel.cpp"
 STACKTOP = sp;return ($0|0); //@line 162 "modules/thread/wrap_Channel.cpp"
}
function __ZN4love14luax_checktypeINS_6thread9LuaThreadEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love6thread14w_Thread_startEP9lua_State($L) {
 $L = $L|0;
 var $$1 = 0, $$arith = 0, $$lcssa = 0, $$overflow = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $args$0 = 0, $i$04 = 0, $i$04$lcssa = 0, $j$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_6thread9LuaThreadEEEPT_P9lua_StateiNS_4TypeE($L,1,53)|0); //@line 30 "modules/thread/wrap_LuaThread.cpp"
 $1 = (_lua_gettop($L)|0); //@line 36 "modules/thread/wrap_LuaThread.cpp"
 $2 = (($1) + -1)|0; //@line 36 "modules/thread/wrap_LuaThread.cpp"
 $3 = ($1|0)>(1); //@line 39 "modules/thread/wrap_LuaThread.cpp"
 L1: do {
  if ($3) {
   $$arith = $2<<2; //@line 41 "modules/thread/wrap_LuaThread.cpp"
   $$overflow = ($2>>>0)>(1073741823); //@line 41 "modules/thread/wrap_LuaThread.cpp"
   $4 = $$overflow ? -1 : $$arith; //@line 41 "modules/thread/wrap_LuaThread.cpp"
   $5 = (__Znaj($4)|0); //@line 41 "modules/thread/wrap_LuaThread.cpp"
   $i$04 = 0;
   while(1) {
    $6 = (($i$04) + 2)|0; //@line 44 "modules/thread/wrap_LuaThread.cpp"
    $7 = (__ZN4love7Variant7fromLuaEP9lua_Stateib($L,$6,1)|0); //@line 44 "modules/thread/wrap_LuaThread.cpp"
    $8 = (($5) + ($i$04<<2)|0); //@line 44 "modules/thread/wrap_LuaThread.cpp"
    HEAP32[$8>>2] = $7; //@line 44 "modules/thread/wrap_LuaThread.cpp"
    $9 = ($7|0)==(0|0); //@line 45 "modules/thread/wrap_LuaThread.cpp"
    if ($9) {
     $$lcssa = $6;$i$04$lcssa = $i$04;
     break;
    }
    $18 = (($i$04) + 1)|0; //@line 42 "modules/thread/wrap_LuaThread.cpp"
    $19 = ($18|0)<($2|0); //@line 42 "modules/thread/wrap_LuaThread.cpp"
    if ($19) {
     $i$04 = $18;
    } else {
     $args$0 = $5;
     break L1;
    }
   }
   $10 = ($i$04$lcssa|0)>(-1); //@line 47 "modules/thread/wrap_LuaThread.cpp"
   L7: do {
    if ($10) {
     $12 = 0;$j$03 = $i$04$lcssa;
     while(1) {
      $11 = ($12|0)==(0|0); //@line 48 "modules/thread/wrap_LuaThread.cpp"
      if (!($11)) {
       $13 = HEAP32[$12>>2]|0; //@line 48 "modules/thread/wrap_LuaThread.cpp"
       $14 = ((($13)) + 4|0); //@line 48 "modules/thread/wrap_LuaThread.cpp"
       $15 = HEAP32[$14>>2]|0; //@line 48 "modules/thread/wrap_LuaThread.cpp"
       FUNCTION_TABLE_vi[$15 & 8191]($12); //@line 48 "modules/thread/wrap_LuaThread.cpp"
      }
      $16 = (($j$03) + -1)|0; //@line 47 "modules/thread/wrap_LuaThread.cpp"
      $17 = ($j$03|0)>(0); //@line 47 "modules/thread/wrap_LuaThread.cpp"
      if (!($17)) {
       break L7;
      }
      $$phi$trans$insert = (($5) + ($16<<2)|0); //@line 48 "modules/thread/wrap_LuaThread.cpp"
      $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 48 "modules/thread/wrap_LuaThread.cpp"
      $12 = $$pre;$j$03 = $16;
     }
    }
   } while(0);
   __ZdaPv($5); //@line 49 "modules/thread/wrap_LuaThread.cpp"
   $20 = (_luaL_argerror($L,$$lcssa,359807)|0); //@line 50 "modules/thread/wrap_LuaThread.cpp"
   $$1 = $20;
   return ($$1|0); //@line 57 "modules/thread/wrap_LuaThread.cpp"
  } else {
   $args$0 = 0;
  }
 } while(0);
 $21 = (__ZN4love6thread9LuaThread5startEPPNS_7VariantEi($0,$args$0,$2)|0); //@line 55 "modules/thread/wrap_LuaThread.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$21); //@line 55 "modules/thread/wrap_LuaThread.cpp"
 $$1 = 1;
 return ($$1|0); //@line 57 "modules/thread/wrap_LuaThread.cpp"
}
function __ZN4love6thread13w_Thread_waitEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_6thread9LuaThreadEEEPT_P9lua_StateiNS_4TypeE($L,1,53)|0); //@line 30 "modules/thread/wrap_LuaThread.cpp"
 $1 = ((($0)) + 8|0); //@line 62 "modules/thread/wrap_LuaThread.cpp"
 __ZN4love6thread10Threadable4waitEv($1); //@line 62 "modules/thread/wrap_LuaThread.cpp"
 return 0; //@line 63 "modules/thread/wrap_LuaThread.cpp"
}
function __ZN4love6thread17w_Thread_getErrorEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $err = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $err = sp;
 $0 = (__ZN4love14luax_checktypeINS_6thread9LuaThreadEEEPT_P9lua_StateiNS_4TypeE($L,1,53)|0); //@line 30 "modules/thread/wrap_LuaThread.cpp"
 $1 = (__ZNK4love6thread9LuaThread8getErrorEv($0)|0); //@line 69 "modules/thread/wrap_LuaThread.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_($err,$1); //@line 69 "modules/thread/wrap_LuaThread.cpp"
 $2 = HEAP8[$err>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = $2 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ($3<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($err)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = HEAP32[$5>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = $2&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $7 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $4 ? $8 : $6; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = ($9|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($10) {
  __THREW__ = 0;
  invoke_vi(3577,($L|0)); //@line 71 "modules/thread/wrap_LuaThread.cpp"
  $11 = __THREW__; __THREW__ = 0;
  $12 = $11&1;
  if ($12) {
   $13 = ___cxa_find_matching_catch_2()|0;
   $14 = tempRet0;
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($err); //@line 75 "modules/thread/wrap_LuaThread.cpp"
   ___resumeException($13|0);
   // unreachable;
  } else {
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($err); //@line 75 "modules/thread/wrap_LuaThread.cpp"
   STACKTOP = sp;return 1; //@line 75 "modules/thread/wrap_LuaThread.cpp"
  }
 } else {
  __THREW__ = 0;
  invoke_vii(3405,($L|0),($err|0)); //@line 73 "modules/thread/wrap_LuaThread.cpp"
  $15 = __THREW__; __THREW__ = 0;
  $16 = $15&1;
  if ($16) {
   $13 = ___cxa_find_matching_catch_2()|0;
   $14 = tempRet0;
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($err); //@line 75 "modules/thread/wrap_LuaThread.cpp"
   ___resumeException($13|0);
   // unreachable;
  } else {
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($err); //@line 75 "modules/thread/wrap_LuaThread.cpp"
   STACKTOP = sp;return 1; //@line 75 "modules/thread/wrap_LuaThread.cpp"
  }
 }
 return (0)|0;
}
function __ZN4love6thread18w_Thread_isRunningEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_6thread9LuaThreadEEEPT_P9lua_StateiNS_4TypeE($L,1,53)|0); //@line 30 "modules/thread/wrap_LuaThread.cpp"
 $1 = ((($0)) + 8|0); //@line 80 "modules/thread/wrap_LuaThread.cpp"
 $2 = (__ZNK4love6thread10Threadable9isRunningEv($1)|0); //@line 80 "modules/thread/wrap_LuaThread.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$2); //@line 80 "modules/thread/wrap_LuaThread.cpp"
 return 1; //@line 81 "modules/thread/wrap_LuaThread.cpp"
}
function _luaopen_thread($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 93 "modules/thread/wrap_LuaThread.cpp"
 HEAP32[$vararg_buffer>>2] = 19508; //@line 95 "modules/thread/wrap_LuaThread.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 95 "modules/thread/wrap_LuaThread.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 95 "modules/thread/wrap_LuaThread.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,53,359866,$vararg_buffer)|0); //@line 95 "modules/thread/wrap_LuaThread.cpp"
 STACKTOP = sp;return ($0|0); //@line 95 "modules/thread/wrap_LuaThread.cpp"
}
function __ZN4love5timer3sdl5TimerD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ModuleD2Ev($this); //@line 51 "modules/timer/sdl/Timer.cpp"
 return; //@line 51 "modules/timer/sdl/Timer.cpp"
}
function __ZN4love5timer3sdl5TimerD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ModuleD2Ev($this); //@line 51 "modules/timer/sdl/Timer.cpp"
 __ZdlPv($this); //@line 46 "modules/timer/sdl/Timer.cpp"
 return; //@line 51 "modules/timer/sdl/Timer.cpp"
}
function __ZNK4love5timer5Timer13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 14; //@line 40 "./modules/timer/Timer.h"
}
function __ZNK4love5timer3sdl5Timer7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (359916|0); //@line 55 "modules/timer/sdl/Timer.cpp"
}
function __ZNK4love5timer3sdl5Timer5sleepEd($this,$seconds) {
 $this = $this|0;
 $seconds = +$seconds;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 64 "modules/timer/sdl/Timer.cpp"
}
function __ZN4love5timer3sdl5TimerC2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5timer5TimerC2Ev($this); //@line 36 "modules/timer/sdl/Timer.cpp"
 HEAP32[$this>>2] = (19556); //@line 37 "modules/timer/sdl/Timer.cpp"
 return; //@line 43 "modules/timer/sdl/Timer.cpp"
}
function __ZN4love5timer5TimerD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ModuleD2Ev($this); //@line 37 "modules/timer/Timer.h"
 __ZdlPv($this); //@line 37 "modules/timer/Timer.h"
 return; //@line 37 "modules/timer/Timer.h"
}
function __ZN4love5timer5Timer4stepEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0, $25 = 0.0, $26 = 0;
 var $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 56|0); //@line 68 "modules/timer/Timer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 68 "modules/timer/Timer.cpp"
 $2 = (($1) + 1)|0; //@line 68 "modules/timer/Timer.cpp"
 HEAP32[$0>>2] = $2; //@line 68 "modules/timer/Timer.cpp"
 $3 = ((($this)) + 8|0); //@line 71 "modules/timer/Timer.cpp"
 $4 = +HEAPF64[$3>>3]; //@line 71 "modules/timer/Timer.cpp"
 $5 = ((($this)) + 16|0); //@line 71 "modules/timer/Timer.cpp"
 HEAPF64[$5>>3] = $4; //@line 71 "modules/timer/Timer.cpp"
 $6 = HEAP32[$this>>2]|0; //@line 74 "modules/timer/Timer.cpp"
 $7 = ((($6)) + 44|0); //@line 74 "modules/timer/Timer.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 74 "modules/timer/Timer.cpp"
 $9 = (+FUNCTION_TABLE_di[$8 & 4095]($this)); //@line 85 "modules/timer/Timer.cpp"
 HEAPF64[$3>>3] = $9; //@line 74 "modules/timer/Timer.cpp"
 $10 = +HEAPF64[$5>>3]; //@line 77 "modules/timer/Timer.cpp"
 $11 = $9 - $10; //@line 77 "modules/timer/Timer.cpp"
 $12 = ((($this)) + 64|0); //@line 77 "modules/timer/Timer.cpp"
 HEAPF64[$12>>3] = $11; //@line 77 "modules/timer/Timer.cpp"
 $13 = ((($this)) + 24|0); //@line 79 "modules/timer/Timer.cpp"
 $14 = +HEAPF64[$13>>3]; //@line 79 "modules/timer/Timer.cpp"
 $15 = $9 - $14; //@line 79 "modules/timer/Timer.cpp"
 $16 = ((($this)) + 48|0); //@line 81 "modules/timer/Timer.cpp"
 $17 = +HEAPF64[$16>>3]; //@line 81 "modules/timer/Timer.cpp"
 $18 = $15 > $17; //@line 81 "modules/timer/Timer.cpp"
 if (!($18)) {
  return; //@line 88 "modules/timer/Timer.cpp"
 }
 $19 = HEAP32[$0>>2]|0; //@line 83 "modules/timer/Timer.cpp"
 $20 = (+($19|0)); //@line 83 "modules/timer/Timer.cpp"
 $21 = $20 / $15; //@line 83 "modules/timer/Timer.cpp"
 $22 = $21 + 0.5; //@line 83 "modules/timer/Timer.cpp"
 $23 = (~~(($22))); //@line 83 "modules/timer/Timer.cpp"
 $24 = ((($this)) + 32|0); //@line 83 "modules/timer/Timer.cpp"
 HEAP32[$24>>2] = $23; //@line 83 "modules/timer/Timer.cpp"
 $25 = $15 / $20; //@line 84 "modules/timer/Timer.cpp"
 $26 = ((($this)) + 40|0); //@line 84 "modules/timer/Timer.cpp"
 HEAPF64[$26>>3] = $25; //@line 84 "modules/timer/Timer.cpp"
 HEAPF64[$13>>3] = $9; //@line 85 "modules/timer/Timer.cpp"
 HEAP32[$0>>2] = 0; //@line 86 "modules/timer/Timer.cpp"
 return; //@line 88 "modules/timer/Timer.cpp"
}
function __ZNK4love5timer5Timer8getDeltaEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 64|0); //@line 92 "modules/timer/Timer.cpp"
 $1 = +HEAPF64[$0>>3]; //@line 92 "modules/timer/Timer.cpp"
 return (+$1); //@line 92 "modules/timer/Timer.cpp"
}
function __ZNK4love5timer5Timer6getFPSEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 97 "modules/timer/Timer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 97 "modules/timer/Timer.cpp"
 return ($1|0); //@line 97 "modules/timer/Timer.cpp"
}
function __ZNK4love5timer5Timer15getAverageDeltaEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0); //@line 102 "modules/timer/Timer.cpp"
 $1 = +HEAPF64[$0>>3]; //@line 102 "modules/timer/Timer.cpp"
 return (+$1); //@line 102 "modules/timer/Timer.cpp"
}
function __ZNK4love5timer5Timer7getTimeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[624040]|0; //@line 122 "modules/timer/Timer.cpp"
 $1 = ($0<<24>>24)==(0); //@line 122 "modules/timer/Timer.cpp"
 if ($1) {
  $2 = (___cxa_guard_acquire((624040|0))|0); //@line 122 "modules/timer/Timer.cpp"
  $3 = ($2|0)==(0); //@line 122 "modules/timer/Timer.cpp"
  if (!($3)) {
   HEAPF64[78006] = 0.0; //@line 122 "modules/timer/Timer.cpp"
   ___cxa_guard_release((624040|0)); //@line 122 "modules/timer/Timer.cpp"
  }
 }
 $4 = (+_emscripten_get_now()); //@line 148 "modules/timer/Timer.cpp"
 $5 = $4 / 1000.0; //@line 148 "modules/timer/Timer.cpp"
 return (+$5); //@line 154 "modules/timer/Timer.cpp"
}
function __ZN4love5timer5TimerC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./common/Module.h"
 HEAP32[$this>>2] = (19612); //@line 61 "modules/timer/Timer.cpp"
 $0 = ((($this)) + 8|0); //@line 54 "modules/timer/Timer.cpp"
 HEAPF64[$0>>3] = 0.0; //@line 54 "modules/timer/Timer.cpp"
 $1 = ((($this)) + 24|0); //@line 55 "modules/timer/Timer.cpp"
 HEAPF64[$1>>3] = 0.0; //@line 55 "modules/timer/Timer.cpp"
 $2 = ((($this)) + 32|0); //@line 56 "modules/timer/Timer.cpp"
 HEAP32[$2>>2] = 0; //@line 56 "modules/timer/Timer.cpp"
 $3 = ((($this)) + 40|0); //@line 57 "modules/timer/Timer.cpp"
 HEAPF64[$3>>3] = 0.0; //@line 57 "modules/timer/Timer.cpp"
 $4 = ((($this)) + 48|0); //@line 58 "modules/timer/Timer.cpp"
 HEAPF64[$4>>3] = 1.0; //@line 58 "modules/timer/Timer.cpp"
 $5 = ((($this)) + 56|0); //@line 59 "modules/timer/Timer.cpp"
 HEAP32[$5>>2] = 0; //@line 59 "modules/timer/Timer.cpp"
 $6 = ((($this)) + 64|0); //@line 60 "modules/timer/Timer.cpp"
 HEAPF64[$6>>3] = 0.0; //@line 60 "modules/timer/Timer.cpp"
 $7 = HEAP8[624040]|0; //@line 122 "modules/timer/Timer.cpp"
 $8 = ($7<<24>>24)==(0); //@line 122 "modules/timer/Timer.cpp"
 if ($8) {
  $9 = (___cxa_guard_acquire((624040|0))|0); //@line 122 "modules/timer/Timer.cpp"
  $10 = ($9|0)==(0); //@line 122 "modules/timer/Timer.cpp"
  if (!($10)) {
   HEAPF64[78006] = 0.0; //@line 122 "modules/timer/Timer.cpp"
   ___cxa_guard_release((624040|0)); //@line 122 "modules/timer/Timer.cpp"
  }
 }
 __THREW__ = 0;
 $11 = (+invoke_d(3779)); //@line 148 "modules/timer/Timer.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if ($13) {
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  __ZN4love6ModuleD2Ev($this); //@line 63 "modules/timer/Timer.cpp"
  ___resumeException($15|0);
  // unreachable;
 } else {
  $14 = $11 / 1000.0; //@line 148 "modules/timer/Timer.cpp"
  HEAPF64[$0>>3] = $14; //@line 62 "modules/timer/Timer.cpp"
  HEAPF64[$1>>3] = $14; //@line 62 "modules/timer/Timer.cpp"
  return; //@line 63 "modules/timer/Timer.cpp"
 }
}
function __ZN4love5timer5Timer17getTimeSinceEpochEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[624040]|0; //@line 122 "modules/timer/Timer.cpp"
 $1 = ($0<<24>>24)==(0); //@line 122 "modules/timer/Timer.cpp"
 if ($1) {
  $2 = (___cxa_guard_acquire((624040|0))|0); //@line 122 "modules/timer/Timer.cpp"
  $3 = ($2|0)==(0); //@line 122 "modules/timer/Timer.cpp"
  if (!($3)) {
   HEAPF64[78006] = 0.0; //@line 122 "modules/timer/Timer.cpp"
   ___cxa_guard_release((624040|0)); //@line 122 "modules/timer/Timer.cpp"
  }
 }
 $4 = (+_emscripten_get_now()); //@line 148 "modules/timer/Timer.cpp"
 $5 = $4 / 1000.0; //@line 148 "modules/timer/Timer.cpp"
 return (+$5); //@line 148 "modules/timer/Timer.cpp"
}
function __ZN4love5touch3sdl5TouchD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19668); //@line 41 "modules/touch/sdl/Touch.h"
 $0 = ((($this)) + 8|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ($1|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = $1; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($2) {
  __ZN4love6ModuleD2Ev($this); //@line 41 "modules/touch/sdl/Touch.h"
  return; //@line 41 "modules/touch/sdl/Touch.h"
 }
 $4 = ((($this)) + 12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5|0)==($1|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($6)) {
  $scevgep$i$i$i = ((($5)) + -48|0);
  $7 = $scevgep$i$i$i;
  $8 = (($7) - ($3))|0;
  $9 = (($8>>>0) / 48)&-1;
  $10 = $9 ^ -1;
  $scevgep4$i$i$i = (($5) + (($10*48)|0)|0);
  HEAP32[$4>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($1); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 __ZN4love6ModuleD2Ev($this); //@line 41 "modules/touch/sdl/Touch.h"
 return; //@line 41 "modules/touch/sdl/Touch.h"
}
function __ZN4love5touch3sdl5TouchD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i$i = 0, $scevgep4$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19668); //@line 41 "modules/touch/sdl/Touch.h"
 $0 = ((($this)) + 8|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ($1|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = $1; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($2) {
  __ZN4love6ModuleD2Ev($this); //@line 41 "modules/touch/sdl/Touch.h"
  __ZdlPv($this); //@line 41 "modules/touch/sdl/Touch.h"
  return; //@line 41 "modules/touch/sdl/Touch.h"
 }
 $4 = ((($this)) + 12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5|0)==($1|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($6)) {
  $scevgep$i$i$i$i = ((($5)) + -48|0);
  $7 = $scevgep$i$i$i$i;
  $8 = (($7) - ($3))|0;
  $9 = (($8>>>0) / 48)&-1;
  $10 = $9 ^ -1;
  $scevgep4$i$i$i$i = (($5) + (($10*48)|0)|0);
  HEAP32[$4>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($1); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 __ZN4love6ModuleD2Ev($this); //@line 41 "modules/touch/sdl/Touch.h"
 __ZdlPv($this); //@line 41 "modules/touch/sdl/Touch.h"
 return; //@line 41 "modules/touch/sdl/Touch.h"
}
function __ZNK4love5touch5Touch13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 15; //@line 55 "./modules/touch/Touch.h"
}
function __ZNK4love5touch3sdl5Touch7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (365610|0); //@line 54 "modules/touch/sdl/Touch.cpp"
}
function __ZNK4love5touch3sdl5Touch10getTouchesEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 38 "modules/touch/sdl/Touch.cpp"
 return ($0|0); //@line 38 "modules/touch/sdl/Touch.cpp"
}
function __ZNK4love5touch3sdl5Touch8getTouchEx($this,$0,$1) {
 $this = $this|0;
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$07 = 0, $__begin$sroa$0$07$lcssa = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $2 = ((($this)) + 8|0); //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ((($this)) + 12|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$4>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($3|0)==($5|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 L1: do {
  if (!($6)) {
   $__begin$sroa$0$07 = $3;
   while(1) {
    $9 = $__begin$sroa$0$07; //@line 45 "modules/touch/sdl/Touch.cpp"
    $10 = $9; //@line 45 "modules/touch/sdl/Touch.cpp"
    $11 = HEAP32[$10>>2]|0; //@line 45 "modules/touch/sdl/Touch.cpp"
    $12 = (($9) + 4)|0; //@line 45 "modules/touch/sdl/Touch.cpp"
    $13 = $12; //@line 45 "modules/touch/sdl/Touch.cpp"
    $14 = HEAP32[$13>>2]|0; //@line 45 "modules/touch/sdl/Touch.cpp"
    $15 = ($11|0)==($0|0); //@line 45 "modules/touch/sdl/Touch.cpp"
    $16 = ($14|0)==($1|0); //@line 45 "modules/touch/sdl/Touch.cpp"
    $17 = $15 & $16; //@line 45 "modules/touch/sdl/Touch.cpp"
    $8 = ((($__begin$sroa$0$07)) + 48|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($17) {
     $__begin$sroa$0$07$lcssa = $__begin$sroa$0$07;
     break;
    }
    $7 = ($8|0)==($5|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($7) {
     break L1;
    } else {
     $__begin$sroa$0$07 = $8;
    }
   }
   STACKTOP = sp;return ($__begin$sroa$0$07$lcssa|0); //@line 50 "modules/touch/sdl/Touch.cpp"
  }
 } while(0);
 $18 = (___cxa_allocate_exception(16)|0); //@line 49 "modules/touch/sdl/Touch.cpp"
 __THREW__ = 0;
 $19 = $vararg_buffer; //@line 49 "modules/touch/sdl/Touch.cpp"
 $20 = $19; //@line 49 "modules/touch/sdl/Touch.cpp"
 HEAP32[$20>>2] = $0; //@line 49 "modules/touch/sdl/Touch.cpp"
 $21 = (($19) + 4)|0; //@line 49 "modules/touch/sdl/Touch.cpp"
 $22 = $21; //@line 49 "modules/touch/sdl/Touch.cpp"
 HEAP32[$22>>2] = $1; //@line 49 "modules/touch/sdl/Touch.cpp"
 invoke_viii(3349,($18|0),(359995|0),($vararg_buffer|0)); //@line 49 "modules/touch/sdl/Touch.cpp"
 $23 = __THREW__; __THREW__ = 0;
 $24 = $23&1;
 if ($24) {
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  ___cxa_free_exception(($18|0)); //@line 49 "modules/touch/sdl/Touch.cpp"
  ___resumeException($25|0);
  // unreachable;
 } else {
  ___cxa_throw(($18|0),(96|0),(344|0)); //@line 49 "modules/touch/sdl/Touch.cpp"
  // unreachable; //@line 49 "modules/touch/sdl/Touch.cpp"
 }
 return (0)|0;
}
function __ZN4love5touch3sdl5Touch7onEventEjRKNS0_5Touch9TouchInfoE($this,$eventtype,$info) {
 $this = $this|0;
 $eventtype = $eventtype|0;
 $info = $info|0;
 var $$cast$i = 0, $$cast$i7 = 0, $$lcssa88 = 0, $$lcssa89 = 0, $$lcssa91 = 0, $$lcssa92 = 0, $$pre = 0, $$pre63 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $__begin$sroa$0$056 = 0, $__first$sroa$0$0$lcssa$i$i = 0, $__first$sroa$0$0$lcssa$i$i20 = 0, $__first$sroa$0$0$ph$i = 0, $__first$sroa$0$0$ph$i$lcssa = 0, $__first$sroa$0$0$ph$i24 = 0, $__first$sroa$0$0$ph$i24$lcssa = 0, $__first$sroa$0$02$i$i = 0, $__first$sroa$0$02$i$i16 = 0, $__first$sroa$0$1$i = 0, $__first$sroa$0$1$i31 = 0, $__i$sroa$0$0$i = 0, $__i$sroa$0$0$i27 = 0, $__i$sroa$0$0$ph$i = 0, $__i$sroa$0$0$ph$i25 = 0, $__new_last2$i$i$i = 0, $__new_last2$i$i$i8 = 0, $scevgep$i$i$i = 0;
 var $scevgep$i$i$i9 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i10 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 switch ($eventtype|0) {
 case 1792:  {
  $0 = ((($this)) + 8|0); //@line 67 "modules/touch/sdl/Touch.cpp"
  $1 = HEAP32[$0>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $2 = ((($this)) + 12|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $3 = HEAP32[$2>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = ($1|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $5 = $3; //@line 877 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  L3: do {
   if ($4) {
    $__first$sroa$0$0$lcssa$i$i20 = $1;
    label = 6;
   } else {
    $6 = $info; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $7 = $6; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $8 = HEAP32[$7>>2]|0; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $9 = (($6) + 4)|0; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $10 = $9; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $11 = HEAP32[$10>>2]|0; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $__first$sroa$0$02$i$i16 = $1;
    while(1) {
     $12 = $__first$sroa$0$02$i$i16; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $13 = $12; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $14 = HEAP32[$13>>2]|0; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $15 = (($12) + 4)|0; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $16 = $15; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $17 = HEAP32[$16>>2]|0; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $18 = ($14|0)==($8|0); //@line 61 "modules/touch/sdl/Touch.cpp"
     $19 = ($17|0)==($11|0); //@line 61 "modules/touch/sdl/Touch.cpp"
     $20 = $18 & $19; //@line 61 "modules/touch/sdl/Touch.cpp"
     if ($20) {
      $__first$sroa$0$0$lcssa$i$i20 = $__first$sroa$0$02$i$i16;
      label = 6;
      break L3;
     }
     $21 = ((($__first$sroa$0$02$i$i16)) + 48|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $22 = ($21|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($22) {
      $50 = $3;$53 = $5;$__first$sroa$0$1$i31 = $5;
      break;
     } else {
      $__first$sroa$0$02$i$i16 = $21;
     }
    }
   }
  } while(0);
  if ((label|0) == 6) {
   $23 = $__first$sroa$0$0$lcssa$i$i20; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $24 = ($__first$sroa$0$0$lcssa$i$i20|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($24) {
    $50 = $3;$53 = $5;$__first$sroa$0$1$i31 = $23;
   } else {
    $__first$sroa$0$0$ph$i24 = $23;$__i$sroa$0$0$ph$i25 = $23;
    L11: while(1) {
     $25 = $info; //@line 67 "modules/touch/sdl/Touch.cpp"
     $26 = $25; //@line 67 "modules/touch/sdl/Touch.cpp"
     $27 = HEAP32[$26>>2]|0; //@line 67 "modules/touch/sdl/Touch.cpp"
     $28 = (($25) + 4)|0; //@line 67 "modules/touch/sdl/Touch.cpp"
     $29 = $28; //@line 67 "modules/touch/sdl/Touch.cpp"
     $30 = HEAP32[$29>>2]|0; //@line 67 "modules/touch/sdl/Touch.cpp"
     $__i$sroa$0$0$i27 = $__i$sroa$0$0$ph$i25;
     while(1) {
      $31 = $__i$sroa$0$0$i27; //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $32 = ((($31)) + 48|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $33 = $32; //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $34 = ($32|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($34) {
       $__first$sroa$0$0$ph$i24$lcssa = $__first$sroa$0$0$ph$i24;
       break L11;
      }
      $35 = $32; //@line 67 "modules/touch/sdl/Touch.cpp"
      $36 = $35; //@line 67 "modules/touch/sdl/Touch.cpp"
      $37 = HEAP32[$36>>2]|0; //@line 67 "modules/touch/sdl/Touch.cpp"
      $38 = (($35) + 4)|0; //@line 67 "modules/touch/sdl/Touch.cpp"
      $39 = $38; //@line 67 "modules/touch/sdl/Touch.cpp"
      $40 = HEAP32[$39>>2]|0; //@line 67 "modules/touch/sdl/Touch.cpp"
      $41 = ($37|0)==($27|0); //@line 61 "modules/touch/sdl/Touch.cpp"
      $42 = ($40|0)==($30|0); //@line 61 "modules/touch/sdl/Touch.cpp"
      $43 = $41 & $42; //@line 61 "modules/touch/sdl/Touch.cpp"
      if ($43) {
       $__i$sroa$0$0$i27 = $33;
      } else {
       $$lcssa88 = $32;$$lcssa89 = $33;
       break;
      }
     }
     $44 = $__first$sroa$0$0$ph$i24; //@line 1163 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $45 = $__first$sroa$0$0$ph$i24; //@line 2148 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     dest=$45; src=$$lcssa88; stop=dest+48|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 2148 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $46 = ((($44)) + 48|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $47 = $46; //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $__first$sroa$0$0$ph$i24 = $47;$__i$sroa$0$0$ph$i25 = $$lcssa89;
    }
    $$pre63 = HEAP32[$2>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $48 = $$pre63;
    $50 = $$pre63;$53 = $48;$__first$sroa$0$1$i31 = $__first$sroa$0$0$ph$i24$lcssa;
   }
  }
  $$cast$i7 = $__first$sroa$0$1$i31; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $49 = ($$cast$i7|0)==($50|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($49) {
   $66 = $50;
  } else {
   $51 = $50; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $52 = (($53) - ($51))|0; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $54 = (($52|0) / 48)&-1; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $55 = $__first$sroa$0$1$i31; //@line 1893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   _memmove(($55|0),($50|0),($52|0))|0; //@line 1893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $56 = (($$cast$i7) + (($54*48)|0)|0); //@line 1894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $57 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $58 = ($57|0)==($56|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($58) {
    $66 = $56;
   } else {
    $__new_last2$i$i$i8 = $56; //@line 818 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $scevgep$i$i$i9 = ((($57)) + -48|0);
    $59 = $scevgep$i$i$i9;
    $60 = (($59) - ($__new_last2$i$i$i8))|0;
    $61 = (($60>>>0) / 48)&-1;
    $62 = $61 ^ -1;
    $scevgep4$i$i$i10 = (($57) + (($62*48)|0)|0);
    HEAP32[$2>>2] = $scevgep4$i$i$i10; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $66 = $scevgep4$i$i$i10;
   }
  }
  $63 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $64 = HEAP32[$63>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $65 = ($66|0)==($64|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($65) {
   __ZNSt3__16vectorIN4love5touch5Touch9TouchInfoENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_($0,$info); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   return; //@line 85 "modules/touch/sdl/Touch.cpp"
  } else {
   dest=$66; src=$info; stop=dest+48|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $67 = HEAP32[$2>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $68 = ((($67)) + 48|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$2>>2] = $68; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   return; //@line 85 "modules/touch/sdl/Touch.cpp"
  }
  break;
 }
 case 1794:  {
  $69 = ((($this)) + 8|0); //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $70 = HEAP32[$69>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $71 = ((($this)) + 12|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $72 = HEAP32[$71>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $73 = ($70|0)==($72|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($73) {
   return; //@line 85 "modules/touch/sdl/Touch.cpp"
  }
  $__begin$sroa$0$056 = $70;
  while(1) {
   $74 = $__begin$sroa$0$056; //@line 74 "modules/touch/sdl/Touch.cpp"
   $75 = $74; //@line 74 "modules/touch/sdl/Touch.cpp"
   $76 = HEAP32[$75>>2]|0; //@line 74 "modules/touch/sdl/Touch.cpp"
   $77 = (($74) + 4)|0; //@line 74 "modules/touch/sdl/Touch.cpp"
   $78 = $77; //@line 74 "modules/touch/sdl/Touch.cpp"
   $79 = HEAP32[$78>>2]|0; //@line 74 "modules/touch/sdl/Touch.cpp"
   $80 = $info; //@line 74 "modules/touch/sdl/Touch.cpp"
   $81 = $80; //@line 74 "modules/touch/sdl/Touch.cpp"
   $82 = HEAP32[$81>>2]|0; //@line 74 "modules/touch/sdl/Touch.cpp"
   $83 = (($80) + 4)|0; //@line 74 "modules/touch/sdl/Touch.cpp"
   $84 = $83; //@line 74 "modules/touch/sdl/Touch.cpp"
   $85 = HEAP32[$84>>2]|0; //@line 74 "modules/touch/sdl/Touch.cpp"
   $86 = ($76|0)==($82|0); //@line 74 "modules/touch/sdl/Touch.cpp"
   $87 = ($79|0)==($85|0); //@line 74 "modules/touch/sdl/Touch.cpp"
   $88 = $86 & $87; //@line 74 "modules/touch/sdl/Touch.cpp"
   if ($88) {
    dest=$__begin$sroa$0$056; src=$info; stop=dest+48|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 75 "modules/touch/sdl/Touch.cpp"
   }
   $89 = ((($__begin$sroa$0$056)) + 48|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $90 = ($89|0)==($72|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($90) {
    break;
   } else {
    $__begin$sroa$0$056 = $89;
   }
  }
  return; //@line 85 "modules/touch/sdl/Touch.cpp"
  break;
 }
 case 1793:  {
  $91 = ((($this)) + 8|0); //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $92 = HEAP32[$91>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $93 = ((($this)) + 12|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $94 = HEAP32[$93>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $95 = ($92|0)==($94|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $96 = $94; //@line 877 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  L39: do {
   if ($95) {
    $__first$sroa$0$0$lcssa$i$i = $92;
    label = 28;
   } else {
    $97 = $info; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $98 = $97; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $99 = HEAP32[$98>>2]|0; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $100 = (($97) + 4)|0; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $101 = $100; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $102 = HEAP32[$101>>2]|0; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $__first$sroa$0$02$i$i = $92;
    while(1) {
     $103 = $__first$sroa$0$02$i$i; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $104 = $103; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $105 = HEAP32[$104>>2]|0; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $106 = (($103) + 4)|0; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $107 = $106; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $108 = HEAP32[$107>>2]|0; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $109 = ($105|0)==($99|0); //@line 61 "modules/touch/sdl/Touch.cpp"
     $110 = ($108|0)==($102|0); //@line 61 "modules/touch/sdl/Touch.cpp"
     $111 = $109 & $110; //@line 61 "modules/touch/sdl/Touch.cpp"
     if ($111) {
      $__first$sroa$0$0$lcssa$i$i = $__first$sroa$0$02$i$i;
      label = 28;
      break L39;
     }
     $112 = ((($__first$sroa$0$02$i$i)) + 48|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $113 = ($112|0)==($94|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($113) {
      $141 = $94;$144 = $96;$__first$sroa$0$1$i = $96;
      break;
     } else {
      $__first$sroa$0$02$i$i = $112;
     }
    }
   }
  } while(0);
  if ((label|0) == 28) {
   $114 = $__first$sroa$0$0$lcssa$i$i; //@line 2139 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $115 = ($__first$sroa$0$0$lcssa$i$i|0)==($94|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($115) {
    $141 = $94;$144 = $96;$__first$sroa$0$1$i = $114;
   } else {
    $__first$sroa$0$0$ph$i = $114;$__i$sroa$0$0$ph$i = $114;
    L47: while(1) {
     $116 = $info; //@line 80 "modules/touch/sdl/Touch.cpp"
     $117 = $116; //@line 80 "modules/touch/sdl/Touch.cpp"
     $118 = HEAP32[$117>>2]|0; //@line 80 "modules/touch/sdl/Touch.cpp"
     $119 = (($116) + 4)|0; //@line 80 "modules/touch/sdl/Touch.cpp"
     $120 = $119; //@line 80 "modules/touch/sdl/Touch.cpp"
     $121 = HEAP32[$120>>2]|0; //@line 80 "modules/touch/sdl/Touch.cpp"
     $__i$sroa$0$0$i = $__i$sroa$0$0$ph$i;
     while(1) {
      $122 = $__i$sroa$0$0$i; //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $123 = ((($122)) + 48|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $124 = $123; //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $125 = ($123|0)==($94|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($125) {
       $__first$sroa$0$0$ph$i$lcssa = $__first$sroa$0$0$ph$i;
       break L47;
      }
      $126 = $123; //@line 80 "modules/touch/sdl/Touch.cpp"
      $127 = $126; //@line 80 "modules/touch/sdl/Touch.cpp"
      $128 = HEAP32[$127>>2]|0; //@line 80 "modules/touch/sdl/Touch.cpp"
      $129 = (($126) + 4)|0; //@line 80 "modules/touch/sdl/Touch.cpp"
      $130 = $129; //@line 80 "modules/touch/sdl/Touch.cpp"
      $131 = HEAP32[$130>>2]|0; //@line 80 "modules/touch/sdl/Touch.cpp"
      $132 = ($128|0)==($118|0); //@line 61 "modules/touch/sdl/Touch.cpp"
      $133 = ($131|0)==($121|0); //@line 61 "modules/touch/sdl/Touch.cpp"
      $134 = $132 & $133; //@line 61 "modules/touch/sdl/Touch.cpp"
      if ($134) {
       $__i$sroa$0$0$i = $124;
      } else {
       $$lcssa91 = $123;$$lcssa92 = $124;
       break;
      }
     }
     $135 = $__first$sroa$0$0$ph$i; //@line 1163 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $136 = $__first$sroa$0$0$ph$i; //@line 2148 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     dest=$136; src=$$lcssa91; stop=dest+48|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 2148 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $137 = ((($135)) + 48|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $138 = $137; //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $__first$sroa$0$0$ph$i = $138;$__i$sroa$0$0$ph$i = $$lcssa92;
    }
    $$pre = HEAP32[$93>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $139 = $$pre;
    $141 = $$pre;$144 = $139;$__first$sroa$0$1$i = $__first$sroa$0$0$ph$i$lcssa;
   }
  }
  $$cast$i = $__first$sroa$0$1$i; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $140 = ($$cast$i|0)==($141|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($140) {
   return; //@line 85 "modules/touch/sdl/Touch.cpp"
  }
  $142 = $141; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $143 = (($144) - ($142))|0; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $145 = (($143|0) / 48)&-1; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $146 = $__first$sroa$0$1$i; //@line 1893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  _memmove(($146|0),($141|0),($143|0))|0; //@line 1893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $147 = (($$cast$i) + (($145*48)|0)|0); //@line 1894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $148 = HEAP32[$93>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $149 = ($148|0)==($147|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($149) {
   return; //@line 85 "modules/touch/sdl/Touch.cpp"
  }
  $__new_last2$i$i$i = $147; //@line 818 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $scevgep$i$i$i = ((($148)) + -48|0);
  $150 = $scevgep$i$i$i;
  $151 = (($150) - ($__new_last2$i$i$i))|0;
  $152 = (($151>>>0) / 48)&-1;
  $153 = $152 ^ -1;
  $scevgep4$i$i$i = (($148) + (($153*48)|0)|0);
  HEAP32[$93>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  return; //@line 85 "modules/touch/sdl/Touch.cpp"
  break;
 }
 default: {
  return; //@line 85 "modules/touch/sdl/Touch.cpp"
 }
 }
}
function __ZNSt3__16vectorIN4love5touch5Touch9TouchInfoENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($3|0) / 48)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(89478485); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9|0) / 48)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(44739242); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $10 << 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = (($19|0) / 48)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$32 = $18;
  } else {
   $$0$i1 = $17;$39 = $18;$40 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13|0) / 48)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 89478485;$39 = $12;$40 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = ($$0$i1*48)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $40;$32 = $39;
 }
 $24 = (($25) + (($26*48)|0)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + (($$0$i2*48)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 dest=$24; src=$__x; stop=dest+48|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $29 = ((($24)) + 48|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = $29; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = (($32) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = (($31|0) / -48)&-1; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = (($24) + (($33*48)|0)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = $34; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($34|0),($36|0),($31|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $35; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $37 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($37) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $38 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($38); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love5video6theora5VideoD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19708); //@line 43 "modules/video/theora/Video.cpp"
 $0 = ((($this)) + 8|0); //@line 44 "modules/video/theora/Video.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 44 "modules/video/theora/Video.cpp"
 $2 = ($1|0)==(0|0); //@line 44 "modules/video/theora/Video.cpp"
 if ($2) {
  __ZN4love6ModuleD2Ev($this); //@line 45 "modules/video/theora/Video.cpp"
  return; //@line 45 "modules/video/theora/Video.cpp"
 }
 $3 = HEAP32[$1>>2]|0; //@line 44 "modules/video/theora/Video.cpp"
 $4 = ((($3)) + 4|0); //@line 44 "modules/video/theora/Video.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 44 "modules/video/theora/Video.cpp"
 FUNCTION_TABLE_vi[$5 & 8191]($1); //@line 44 "modules/video/theora/Video.cpp"
 __ZN4love6ModuleD2Ev($this); //@line 45 "modules/video/theora/Video.cpp"
 return; //@line 45 "modules/video/theora/Video.cpp"
}
function __ZN4love5video6theora5VideoD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19708); //@line 43 "modules/video/theora/Video.cpp"
 $0 = ((($this)) + 8|0); //@line 44 "modules/video/theora/Video.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 44 "modules/video/theora/Video.cpp"
 $2 = ($1|0)==(0|0); //@line 44 "modules/video/theora/Video.cpp"
 if (!($2)) {
  $3 = HEAP32[$1>>2]|0; //@line 44 "modules/video/theora/Video.cpp"
  $4 = ((($3)) + 4|0); //@line 44 "modules/video/theora/Video.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 44 "modules/video/theora/Video.cpp"
  FUNCTION_TABLE_vi[$5 & 8191]($1); //@line 44 "modules/video/theora/Video.cpp"
 }
 __ZN4love6ModuleD2Ev($this); //@line 45 "modules/video/theora/Video.cpp"
 __ZdlPv($this); //@line 43 "modules/video/theora/Video.cpp"
 return; //@line 45 "modules/video/theora/Video.cpp"
}
function __ZNK4love5video5Video13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 17; //@line 42 "./modules/video/Video.h"
}
function __ZNK4love5video6theora5Video7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (360070|0); //@line 56 "modules/video/theora/Video.cpp"
}
function __ZN4love5video6theora5Video14newVideoStreamEPNS_10filesystem4FileE($this,$file) {
 $this = $this|0;
 $file = $file|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(592)|0); //@line 49 "modules/video/theora/Video.cpp"
 __THREW__ = 0;
 invoke_vii(3780,($0|0),($file|0)); //@line 49 "modules/video/theora/Video.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  __ZdlPv($0); //@line 49 "modules/video/theora/Video.cpp"
  ___resumeException($5|0);
  // unreachable;
 } else {
  $3 = ((($this)) + 8|0); //@line 50 "modules/video/theora/Video.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 50 "modules/video/theora/Video.cpp"
  __ZN4love5video6theora6Worker9addStreamEPNS1_11VideoStreamE($4,$0); //@line 50 "modules/video/theora/Video.cpp"
  return ($0|0); //@line 51 "modules/video/theora/Video.cpp"
 }
 return (0)|0;
}
function __ZN4love5video6theora6Worker9addStreamEPNS1_11VideoStreamE($this,$stream) {
 $this = $this|0;
 $stream = $stream|0;
 var $$0 = 0, $$01 = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp + 4|0;
 $0 = sp;
 $1 = ((($this)) + 32|0); //@line 72 "modules/video/theora/Video.cpp"
 $2 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($1)|0); //@line 72 "modules/video/theora/Video.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($l,$2); //@line 72 "modules/video/theora/Video.cpp"
 HEAP32[$0>>2] = $stream; //@line 93 "./common/Object.h"
 $3 = ($stream|0)==(0|0); //@line 95 "./common/Object.h"
 if (!($3)) {
  $4 = HEAP32[$stream>>2]|0; //@line 95 "./common/Object.h"
  $5 = ((($4)) + 8|0); //@line 95 "./common/Object.h"
  $6 = HEAP32[$5>>2]|0; //@line 95 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($6|0,($stream|0)); //@line 95 "./common/Object.h"
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if ($8) {
   $33 = ___cxa_find_matching_catch_2()|0;
   $34 = tempRet0;
   $$0 = $34;$$01 = $33;
   __ZN4love6thread4LockD2Ev($l); //@line 74 "modules/video/theora/Video.cpp"
   ___resumeException($$01|0);
   // unreachable;
  }
 }
 $9 = ((($this)) + 24|0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = HEAP32[$9>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ((($this)) + 28|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $12 = HEAP32[$11>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ($10>>>0)<($12>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 L6: do {
  if ($13) {
   HEAP32[$10>>2] = $stream; //@line 99 "./common/Object.h"
   do {
    if ($3) {
     $22 = $10;
    } else {
     $14 = HEAP32[$stream>>2]|0; //@line 101 "./common/Object.h"
     $15 = ((($14)) + 8|0); //@line 101 "./common/Object.h"
     $16 = HEAP32[$15>>2]|0; //@line 101 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($16|0,($stream|0)); //@line 101 "./common/Object.h"
     $17 = __THREW__; __THREW__ = 0;
     $18 = $17&1;
     if ($18) {
      $19 = ___cxa_find_matching_catch_2()|0;
      $20 = tempRet0;
      $44 = $20;$45 = $19;
      label = 16;
      break L6;
     } else {
      $$pre$i = HEAP32[$9>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $22 = $$pre$i;
      break;
     }
    }
   } while(0);
   $21 = ((($22)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$9>>2] = $21; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  } else {
   $23 = ((($this)) + 20|0); //@line 73 "modules/video/theora/Video.cpp"
   __THREW__ = 0;
   invoke_vii(3781,($23|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $24 = __THREW__; __THREW__ = 0;
   $25 = $24&1;
   if ($25) {
    $35 = ___cxa_find_matching_catch_2()|0;
    $36 = tempRet0;
    if ($3) {
     $$0 = $36;$$01 = $35;
     __ZN4love6thread4LockD2Ev($l); //@line 74 "modules/video/theora/Video.cpp"
     ___resumeException($$01|0);
     // unreachable;
    } else {
     $44 = $36;$45 = $35;
     label = 16;
    }
   }
  }
 } while(0);
 if ((label|0) == 16) {
  $37 = HEAP32[$stream>>2]|0; //@line 106 "./common/Object.h"
  $38 = ((($37)) + 12|0); //@line 106 "./common/Object.h"
  $39 = HEAP32[$38>>2]|0; //@line 106 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($39|0,($stream|0)); //@line 106 "./common/Object.h"
  $40 = __THREW__; __THREW__ = 0;
  $41 = $40&1;
  if ($41) {
   $42 = ___cxa_find_matching_catch_3(0|0)|0;
   $43 = tempRet0;
   ___clang_call_terminate($42); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  } else {
   $$0 = $44;$$01 = $45;
   __ZN4love6thread4LockD2Ev($l); //@line 74 "modules/video/theora/Video.cpp"
   ___resumeException($$01|0);
   // unreachable;
  }
 }
 if ($3) {
  __ZN4love6thread4LockD2Ev($l); //@line 74 "modules/video/theora/Video.cpp"
  STACKTOP = sp;return; //@line 74 "modules/video/theora/Video.cpp"
 }
 $26 = HEAP32[$stream>>2]|0; //@line 106 "./common/Object.h"
 $27 = ((($26)) + 12|0); //@line 106 "./common/Object.h"
 $28 = HEAP32[$27>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($28|0,($stream|0)); //@line 106 "./common/Object.h"
 $29 = __THREW__; __THREW__ = 0;
 $30 = $29&1;
 if ($30) {
  $31 = ___cxa_find_matching_catch_3(0|0)|0;
  $32 = tempRet0;
  ___clang_call_terminate($31); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  __ZN4love6thread4LockD2Ev($l); //@line 74 "modules/video/theora/Video.cpp"
  STACKTOP = sp;return; //@line 74 "modules/video/theora/Video.cpp"
 }
}
function __ZNSt3__16vectorIN4love9StrongRefINS1_5video6theora11VideoStreamEEENS_9allocatorIS6_EEE21__push_back_slow_pathIS6_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i8 = 0, $$0$i9 = 0, $$01$i$i = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre12 = 0, $$pre13 = 0, $$pre14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__v$sroa$15$0 = 0, $__v$sroa$6$0 = 0, $__v$sroa$6$0$lcssa = 0, $__v$sroa$6$1 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = $$in; //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = HEAP32[$8>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $10 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($11>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $17 = $10 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ($17>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $19 = $18 ? $5 : $17; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = (($20) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = $21 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = ($19|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $24 = $20; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($23) {
   $$0$i9 = 0;$101 = $24;$28 = 0;$29 = $22;
  } else {
   $$0$i8 = $19;$100 = $22;$99 = $24;
   label = 6;
  }
 } else {
  $13 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = $14 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = $13; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $$0$i8 = 1073741823;$100 = $15;$99 = $16;
  label = 6;
 }
 if ((label|0) == 6) {
  $25 = $$0$i8 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (__Znwj($25)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i9 = $$0$i8;$101 = $99;$28 = $26;$29 = $100;
 }
 $27 = (($28) + ($29<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = $27; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $31 = (($28) + ($$0$i9<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $32 = $31; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $33 = HEAP32[$__x>>2]|0; //@line 139 "./common/Object.h"
 HEAP32[$27>>2] = $33; //@line 99 "./common/Object.h"
 $34 = ($33|0)==(0|0); //@line 101 "./common/Object.h"
 do {
  if ($34) {
   $43 = $101;$44 = $7;
   label = 10;
  } else {
   $35 = HEAP32[$33>>2]|0; //@line 101 "./common/Object.h"
   $36 = ((($35)) + 8|0); //@line 101 "./common/Object.h"
   $37 = HEAP32[$36>>2]|0; //@line 101 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($37|0,($33|0)); //@line 101 "./common/Object.h"
   $38 = __THREW__; __THREW__ = 0;
   $39 = $38&1;
   if ($39) {
    $81 = ___cxa_find_matching_catch_2()|0;
    $82 = tempRet0;
    $__v$sroa$15$0 = $30;$__v$sroa$6$1 = $30;$lpad$phi$index = $81;$lpad$phi$index2 = $82;
    break;
   } else {
    $$pre12 = HEAP32[$this>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre13 = HEAP32[$0>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $43 = $$pre13;$44 = $$pre12;
    label = 10;
    break;
   }
  }
 } while(0);
 L15: do {
  if ((label|0) == 10) {
   $40 = ((($27)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $41 = $40; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $42 = ($43|0)==($44|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   do {
    if ($42) {
     $45 = $43; //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $60 = $30;$62 = $45;$64 = $43;
    } else {
     $$01$i$i = $43;$48 = $27;$__v$sroa$6$0 = $30;
     while(1) {
      $46 = ((($$01$i$i)) + -4|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $47 = ((($48)) + -4|0); //@line 99 "./common/Object.h"
      $49 = HEAP32[$46>>2]|0; //@line 139 "./common/Object.h"
      HEAP32[$47>>2] = $49; //@line 99 "./common/Object.h"
      $50 = ($49|0)==(0|0); //@line 101 "./common/Object.h"
      if (!($50)) {
       $51 = HEAP32[$49>>2]|0; //@line 101 "./common/Object.h"
       $52 = ((($51)) + 8|0); //@line 101 "./common/Object.h"
       $53 = HEAP32[$52>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($53|0,($49|0)); //@line 101 "./common/Object.h"
       $54 = __THREW__; __THREW__ = 0;
       $55 = $54&1;
       if ($55) {
        $__v$sroa$6$0$lcssa = $__v$sroa$6$0;
        label = 24;
        break;
       }
      }
      $56 = $__v$sroa$6$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $57 = ((($56)) + -4|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $58 = $57; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $59 = ($46|0)==($44|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      if ($59) {
       $$lcssa = $58;
       label = 15;
       break;
      } else {
       $$01$i$i = $46;$48 = $57;$__v$sroa$6$0 = $58;
      }
     }
     if ((label|0) == 15) {
      $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $$pre14 = HEAP32[$0>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $60 = $$lcssa;$62 = $$pre$i;$64 = $$pre14;
      break;
     }
     else if ((label|0) == 24) {
      $79 = ___cxa_find_matching_catch_2()|0;
      $80 = tempRet0;
      $__v$sroa$15$0 = $41;$__v$sroa$6$1 = $__v$sroa$6$0$lcssa;$lpad$phi$index = $79;$lpad$phi$index2 = $80;
      break L15;
     }
    }
   } while(0);
   HEAP32[$this>>2] = $60; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$0>>2] = $41; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$8>>2] = $32; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $61 = $62; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $63 = ($64|0)==($61|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   L28: do {
    if (!($63)) {
     $66 = $64;
     while(1) {
      $65 = ((($66)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $67 = HEAP32[$65>>2]|0; //@line 106 "./common/Object.h"
      $68 = ($67|0)==(0|0); //@line 106 "./common/Object.h"
      if (!($68)) {
       $70 = HEAP32[$67>>2]|0; //@line 106 "./common/Object.h"
       $71 = ((($70)) + 12|0); //@line 106 "./common/Object.h"
       $72 = HEAP32[$71>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($72|0,($67|0)); //@line 106 "./common/Object.h"
       $73 = __THREW__; __THREW__ = 0;
       $74 = $73&1;
       if ($74) {
        break;
       }
      }
      $69 = ($65|0)==($61|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if ($69) {
       break L28;
      } else {
       $66 = $65;
      }
     }
     $75 = ___cxa_find_matching_catch_3(0|0)|0;
     $76 = tempRet0;
     ___clang_call_terminate($75); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   } while(0);
   $77 = ($62|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($77) {
    return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $78 = $62; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __ZdlPv($78); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
 } while(0);
 $83 = $__v$sroa$6$1; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $84 = $__v$sroa$15$0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $85 = ($84|0)==($83|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 L41: do {
  if (!($85)) {
   $87 = $84;
   while(1) {
    $86 = ((($87)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $88 = HEAP32[$86>>2]|0; //@line 106 "./common/Object.h"
    $89 = ($88|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($89)) {
     $91 = HEAP32[$88>>2]|0; //@line 106 "./common/Object.h"
     $92 = ((($91)) + 12|0); //@line 106 "./common/Object.h"
     $93 = HEAP32[$92>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($93|0,($88|0)); //@line 106 "./common/Object.h"
     $94 = __THREW__; __THREW__ = 0;
     $95 = $94&1;
     if ($95) {
      break;
     }
    }
    $90 = ($86|0)==($83|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($90) {
     break L41;
    } else {
     $87 = $86;
    }
   }
   $96 = ___cxa_find_matching_catch_3(0|0)|0;
   $97 = tempRet0;
   ___clang_call_terminate($96); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 } while(0);
 $98 = ($28|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($98) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 __ZdlPv($28); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZN4love5video6theora6WorkerD2Ev($this) {
 $this = $this|0;
 var $$pre$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i5 = 0, $$pre$i8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $l$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l$i = sp;
 HEAP32[$this>>2] = (19744); //@line 66 "modules/video/theora/Video.cpp"
 $0 = ((($this)) + 32|0); //@line 79 "modules/video/theora/Video.cpp"
 __THREW__ = 0;
 $1 = (invoke_ii(3782,($0|0))|0); //@line 79 "modules/video/theora/Video.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  __THREW__ = 0;
  invoke_vii(3783,($l$i|0),($1|0)); //@line 79 "modules/video/theora/Video.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if (!($5)) {
   $6 = ((($this)) + 36|0); //@line 80 "modules/video/theora/Video.cpp"
   HEAP8[$6>>0] = 1; //@line 80 "modules/video/theora/Video.cpp"
   __ZN4love6thread4LockD2Ev($l$i); //@line 81 "modules/video/theora/Video.cpp"
   $7 = ((($this)) + 4|0); //@line 83 "modules/video/theora/Video.cpp"
   $8 = HEAP32[$7>>2]|0; //@line 83 "modules/video/theora/Video.cpp"
   $9 = HEAP32[$8>>2]|0; //@line 83 "modules/video/theora/Video.cpp"
   $10 = ((($9)) + 12|0); //@line 83 "modules/video/theora/Video.cpp"
   $11 = HEAP32[$10>>2]|0; //@line 83 "modules/video/theora/Video.cpp"
   __THREW__ = 0;
   invoke_vi($11|0,($8|0)); //@line 83 "modules/video/theora/Video.cpp"
   $12 = __THREW__; __THREW__ = 0;
   $13 = $12&1;
   if (!($13)) {
    __ZN4love6thread8MutexRefD2Ev($0); //@line 68 "modules/video/theora/Video.cpp"
    $14 = ((($this)) + 20|0); //@line 68 "modules/video/theora/Video.cpp"
    $15 = HEAP32[$14>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $16 = ($15|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($16) {
     __ZN4love6thread10ThreadableD2Ev($this); //@line 68 "modules/video/theora/Video.cpp"
     STACKTOP = sp;return; //@line 68 "modules/video/theora/Video.cpp"
    }
    $17 = ((($this)) + 24|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $18 = HEAP32[$17>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $19 = ($18|0)==($15|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    do {
     if ($19) {
      $33 = $15;
     } else {
      $21 = $18;
      while(1) {
       $20 = ((($21)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$17>>2] = $20; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $22 = HEAP32[$20>>2]|0; //@line 106 "./common/Object.h"
       $23 = ($22|0)==(0|0); //@line 106 "./common/Object.h"
       if ($23) {
        $25 = $20;
       } else {
        $26 = HEAP32[$22>>2]|0; //@line 106 "./common/Object.h"
        $27 = ((($26)) + 12|0); //@line 106 "./common/Object.h"
        $28 = HEAP32[$27>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($28|0,($22|0)); //@line 106 "./common/Object.h"
        $29 = __THREW__; __THREW__ = 0;
        $30 = $29&1;
        if ($30) {
         label = 10;
         break;
        }
        $$pre$i$i$i = HEAP32[$17>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $25 = $$pre$i$i$i;
       }
       $24 = ($25|0)==($15|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($24) {
        label = 11;
        break;
       } else {
        $21 = $25;
       }
      }
      if ((label|0) == 10) {
       $31 = ___cxa_find_matching_catch_3(0|0)|0;
       $32 = tempRet0;
       ___clang_call_terminate($31); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      else if ((label|0) == 11) {
       $$pre$i = HEAP32[$14>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $33 = $$pre$i;
       break;
      }
     }
    } while(0);
    __ZdlPv($33); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    __ZN4love6thread10ThreadableD2Ev($this); //@line 68 "modules/video/theora/Video.cpp"
    STACKTOP = sp;return; //@line 68 "modules/video/theora/Video.cpp"
   }
  }
 }
 $34 = ___cxa_find_matching_catch_3(0|0)|0;
 $35 = tempRet0;
 __ZN4love6thread8MutexRefD2Ev($0); //@line 68 "modules/video/theora/Video.cpp"
 $36 = ((($this)) + 20|0); //@line 68 "modules/video/theora/Video.cpp"
 $37 = HEAP32[$36>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $38 = ($37|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($38) {
  __ZN4love6thread10ThreadableD2Ev($this); //@line 68 "modules/video/theora/Video.cpp"
  ___clang_call_terminate($34); //@line 68 "modules/video/theora/Video.cpp"
  // unreachable; //@line 68 "modules/video/theora/Video.cpp"
 }
 $39 = ((($this)) + 24|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $40 = HEAP32[$39>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $41 = ($40|0)==($37|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($41) {
   $55 = $37;
  } else {
   $43 = $40;
   while(1) {
    $42 = ((($43)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$39>>2] = $42; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $44 = HEAP32[$42>>2]|0; //@line 106 "./common/Object.h"
    $45 = ($44|0)==(0|0); //@line 106 "./common/Object.h"
    if ($45) {
     $47 = $42;
    } else {
     $48 = HEAP32[$44>>2]|0; //@line 106 "./common/Object.h"
     $49 = ((($48)) + 12|0); //@line 106 "./common/Object.h"
     $50 = HEAP32[$49>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($50|0,($44|0)); //@line 106 "./common/Object.h"
     $51 = __THREW__; __THREW__ = 0;
     $52 = $51&1;
     if ($52) {
      label = 20;
      break;
     }
     $$pre$i$i$i5 = HEAP32[$39>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $47 = $$pre$i$i$i5;
    }
    $46 = ($47|0)==($37|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($46) {
     label = 21;
     break;
    } else {
     $43 = $47;
    }
   }
   if ((label|0) == 20) {
    $53 = ___cxa_find_matching_catch_3(0|0)|0;
    $54 = tempRet0;
    ___clang_call_terminate($53); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 21) {
    $$pre$i8 = HEAP32[$36>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $55 = $$pre$i8;
    break;
   }
  }
 } while(0);
 __ZdlPv($55); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 __ZN4love6thread10ThreadableD2Ev($this); //@line 68 "modules/video/theora/Video.cpp"
 ___clang_call_terminate($34); //@line 68 "modules/video/theora/Video.cpp"
 // unreachable; //@line 68 "modules/video/theora/Video.cpp"
}
function __ZN4love5video6theora6WorkerD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5video6theora6WorkerD2Ev($this); //@line 66 "modules/video/theora/Video.cpp"
 __ZdlPv($this); //@line 66 "modules/video/theora/Video.cpp"
 return; //@line 68 "modules/video/theora/Video.cpp"
}
function __ZN4love5video6theora6Worker14threadFunctionEv($this) {
 $this = $this|0;
 var $$0$lcssa$i$i3$i = 0, $$012$i$i$i = 0, $$03$i$i$i = 0, $$pre$i = 0, $$pre$i$i$i = 0, $$sink16 = 0, $$sink16$lcssa30 = 0, $0 = 0.0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0, $7 = 0.0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0;
 var $__first5$i$i$i = 0, $l = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $scevgep$i$i$i = 0, $scevgep7$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = (+__ZN4love5timer5Timer17getTimeSinceEpochEv()); //@line 88 "modules/video/theora/Video.cpp"
 $1 = (+__ZN4love5timer5Timer17getTimeSinceEpochEv()); //@line 91 "modules/video/theora/Video.cpp"
 $2 = ((($this)) + 32|0); //@line 96 "modules/video/theora/Video.cpp"
 $3 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($2)|0); //@line 96 "modules/video/theora/Video.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($l,$3); //@line 96 "modules/video/theora/Video.cpp"
 $4 = ((($this)) + 36|0); //@line 98 "modules/video/theora/Video.cpp"
 $5 = HEAP8[$4>>0]|0; //@line 98 "modules/video/theora/Video.cpp"
 $6 = ($5<<24>>24)==(0); //@line 98 "modules/video/theora/Video.cpp"
 if (!($6)) {
  __ZN4love6thread4LockD2Ev($l); //@line 113 "modules/video/theora/Video.cpp"
  STACKTOP = sp;return; //@line 118 "modules/video/theora/Video.cpp"
 }
 $7 = $1 - $0; //@line 92 "modules/video/theora/Video.cpp"
 $8 = ((($this)) + 20|0); //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ((($this)) + 24|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $60 = $7;$68 = $1;
 L4: while(1) {
  $10 = HEAP32[$8>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $11 = HEAP32[$9>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $12 = ($10|0)==($11|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  L6: do {
   if (!($12)) {
    $$sink16 = $10;
    while(1) {
     $13 = HEAP32[$$sink16>>2]|0; //@line 127 "./common/Object.h"
     __THREW__ = 0;
     $14 = (invoke_ii(3784,($13|0))|0); //@line 104 "modules/video/theora/Video.cpp"
     $15 = __THREW__; __THREW__ = 0;
     $16 = $15&1;
     if ($16) {
      label = 21;
      break L4;
     }
     $17 = ($14|0)==(1); //@line 104 "modules/video/theora/Video.cpp"
     if ($17) {
      $$sink16$lcssa30 = $$sink16;
      break;
     }
     __THREW__ = 0;
     invoke_vid(3785,($13|0),(+$60)); //@line 111 "modules/video/theora/Video.cpp"
     $61 = __THREW__; __THREW__ = 0;
     $62 = $61&1;
     if ($62) {
      label = 21;
      break L4;
     }
     $63 = ((($$sink16)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $64 = HEAP32[$9>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $65 = ($63|0)==($64|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($65) {
      break L6;
     } else {
      $$sink16 = $63;
     }
    }
    $18 = ((($$sink16$lcssa30)) + 4|0); //@line 1684 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $19 = HEAP32[$9>>2]|0; //@line 1684 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $20 = ($18|0)==($19|0); //@line 1877 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    if ($20) {
     $$0$lcssa$i$i3$i = $$sink16$lcssa30;$72 = $18;
    } else {
     $__first5$i$i$i = $18; //@line 1902 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $scevgep$i$i$i = ((($19)) + -4|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $21 = $scevgep$i$i$i; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $22 = (($21) - ($__first5$i$i$i))|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $$012$i$i$i = $18;$$03$i$i$i = $$sink16$lcssa30;
     while(1) {
      $23 = HEAP32[$$012$i$i$i>>2]|0; //@line 139 "./common/Object.h"
      $24 = ($23|0)==(0|0); //@line 132 "./common/Object.h"
      if (!($24)) {
       $25 = HEAP32[$23>>2]|0; //@line 132 "./common/Object.h"
       $26 = ((($25)) + 8|0); //@line 132 "./common/Object.h"
       $27 = HEAP32[$26>>2]|0; //@line 132 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($27|0,($23|0)); //@line 132 "./common/Object.h"
       $28 = __THREW__; __THREW__ = 0;
       $29 = $28&1;
       if ($29) {
        label = 20;
        break L4;
       }
      }
      $30 = HEAP32[$$03$i$i$i>>2]|0; //@line 133 "./common/Object.h"
      $31 = ($30|0)==(0|0); //@line 133 "./common/Object.h"
      if (!($31)) {
       $32 = HEAP32[$30>>2]|0; //@line 133 "./common/Object.h"
       $33 = ((($32)) + 12|0); //@line 133 "./common/Object.h"
       $34 = HEAP32[$33>>2]|0; //@line 133 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($34|0,($30|0)); //@line 133 "./common/Object.h"
       $35 = __THREW__; __THREW__ = 0;
       $36 = $35&1;
       if ($36) {
        label = 20;
        break L4;
       }
      }
      HEAP32[$$03$i$i$i>>2] = $23; //@line 134 "./common/Object.h"
      $37 = ((($$012$i$i$i)) + 4|0); //@line 1877 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $38 = ((($$03$i$i$i)) + 4|0); //@line 1877 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $39 = ($37|0)==($19|0); //@line 1877 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if ($39) {
       break;
      } else {
       $$012$i$i$i = $37;$$03$i$i$i = $38;
      }
     }
     $40 = $22 >>> 2; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $41 = (($40) + 1)|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $scevgep7$i$i$i = (($$sink16$lcssa30) + ($41<<2)|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $$pre$i = HEAP32[$9>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $42 = ($$pre$i|0)==($scevgep7$i$i$i|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($42) {
      break;
     } else {
      $$0$lcssa$i$i3$i = $scevgep7$i$i$i;$72 = $$pre$i;
     }
    }
    $44 = $72;
    while(1) {
     $43 = ((($44)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$9>>2] = $43; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $45 = HEAP32[$43>>2]|0; //@line 106 "./common/Object.h"
     $46 = ($45|0)==(0|0); //@line 106 "./common/Object.h"
     if ($46) {
      $48 = $43;
     } else {
      $49 = HEAP32[$45>>2]|0; //@line 106 "./common/Object.h"
      $50 = ((($49)) + 12|0); //@line 106 "./common/Object.h"
      $51 = HEAP32[$50>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($51|0,($45|0)); //@line 106 "./common/Object.h"
      $52 = __THREW__; __THREW__ = 0;
      $53 = $52&1;
      if ($53) {
       label = 19;
       break L4;
      }
      $$pre$i$i$i = HEAP32[$9>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $48 = $$pre$i$i$i;
     }
     $47 = ($48|0)==($$0$lcssa$i$i3$i|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($47) {
      break;
     } else {
      $44 = $48;
     }
    }
   }
  } while(0);
  __ZN4love6thread4LockD2Ev($l); //@line 113 "modules/video/theora/Video.cpp"
  __ZN4love5delayEj(2); //@line 116 "modules/video/theora/Video.cpp"
  $66 = (+__ZN4love5timer5Timer17getTimeSinceEpochEv()); //@line 91 "modules/video/theora/Video.cpp"
  $67 = $66 - $68; //@line 92 "modules/video/theora/Video.cpp"
  $69 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($2)|0); //@line 96 "modules/video/theora/Video.cpp"
  __ZN4love6thread4LockC2EPNS0_5MutexE($l,$69); //@line 96 "modules/video/theora/Video.cpp"
  $70 = HEAP8[$4>>0]|0; //@line 98 "modules/video/theora/Video.cpp"
  $71 = ($70<<24>>24)==(0); //@line 98 "modules/video/theora/Video.cpp"
  if ($71) {
   $60 = $67;$68 = $66;
  } else {
   label = 26;
   break;
  }
 }
 if ((label|0) == 19) {
  $54 = ___cxa_find_matching_catch_3(0|0)|0;
  $55 = tempRet0;
  ___clang_call_terminate($54); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 }
 else if ((label|0) == 20) {
  $56 = ___cxa_find_matching_catch_2()|0;
  $57 = tempRet0;
  $lpad$phi$index = $56;$lpad$phi$index2 = $57;
  __ZN4love6thread4LockD2Ev($l); //@line 113 "modules/video/theora/Video.cpp"
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 else if ((label|0) == 21) {
  $58 = ___cxa_find_matching_catch_2()|0;
  $59 = tempRet0;
  $lpad$phi$index = $58;$lpad$phi$index2 = $59;
  __ZN4love6thread4LockD2Ev($l); //@line 113 "modules/video/theora/Video.cpp"
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 else if ((label|0) == 26) {
  __ZN4love6thread4LockD2Ev($l); //@line 113 "modules/video/theora/Video.cpp"
  STACKTOP = sp;return; //@line 118 "modules/video/theora/Video.cpp"
 }
}
function __ZN4love5video6theora5VideoC2Ev($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./common/Module.h"
 HEAP32[$this>>2] = (19708); //@line 37 "modules/video/theora/Video.cpp"
 __THREW__ = 0;
 $0 = (invoke_ii(3348,40)|0); //@line 38 "modules/video/theora/Video.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 do {
  if (!($2)) {
   __THREW__ = 0;
   invoke_vi(3786,($0|0)); //@line 38 "modules/video/theora/Video.cpp"
   $3 = __THREW__; __THREW__ = 0;
   $4 = $3&1;
   if ($4) {
    $10 = ___cxa_find_matching_catch_2()|0;
    $11 = tempRet0;
    __ZdlPv($0); //@line 38 "modules/video/theora/Video.cpp"
    $$0 = $11;$$01 = $10;
    __ZN4love6ModuleD2Ev($this); //@line 40 "modules/video/theora/Video.cpp"
    ___resumeException($$01|0);
    // unreachable;
   } else {
    $5 = ((($this)) + 8|0); //@line 38 "modules/video/theora/Video.cpp"
    HEAP32[$5>>2] = $0; //@line 38 "modules/video/theora/Video.cpp"
    __THREW__ = 0;
    (invoke_ii(3787,($0|0))|0); //@line 39 "modules/video/theora/Video.cpp"
    $6 = __THREW__; __THREW__ = 0;
    $7 = $6&1;
    if ($7) {
     break;
    }
    return; //@line 40 "modules/video/theora/Video.cpp"
   }
  }
 } while(0);
 $8 = ___cxa_find_matching_catch_2()|0;
 $9 = tempRet0;
 $$0 = $9;$$01 = $8;
 __ZN4love6ModuleD2Ev($this); //@line 40 "modules/video/theora/Video.cpp"
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN4love5video6theora6WorkerC2Ev($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$pre$i = 0, $$pre$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6thread10ThreadableC2Ev($this); //@line 59 "modules/video/theora/Video.cpp"
 HEAP32[$this>>2] = (19744); //@line 61 "modules/video/theora/Video.cpp"
 $0 = ((($this)) + 20|0); //@line 59 "modules/video/theora/Video.cpp"
 HEAP32[$0>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($this)) + 24|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($this)) + 28|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = ((($this)) + 32|0); //@line 59 "modules/video/theora/Video.cpp"
 __THREW__ = 0;
 invoke_vi(3788,($3|0)); //@line 59 "modules/video/theora/Video.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 do {
  if ($5) {
   $10 = ___cxa_find_matching_catch_2()|0;
   $11 = tempRet0;
   $$0 = $11;$$01 = $10;
  } else {
   $6 = ((($this)) + 36|0); //@line 60 "modules/video/theora/Video.cpp"
   HEAP8[$6>>0] = 0; //@line 60 "modules/video/theora/Video.cpp"
   $7 = ((($this)) + 8|0); //@line 62 "modules/video/theora/Video.cpp"
   __THREW__ = 0;
   (invoke_iii(3360,($7|0),(360116|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $8 = __THREW__; __THREW__ = 0;
   $9 = $8&1;
   if ($9) {
    $12 = ___cxa_find_matching_catch_2()|0;
    $13 = tempRet0;
    __ZN4love6thread8MutexRefD2Ev($3); //@line 63 "modules/video/theora/Video.cpp"
    $$0 = $13;$$01 = $12;
    break;
   } else {
    return; //@line 63 "modules/video/theora/Video.cpp"
   }
  }
 } while(0);
 $14 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = ($14|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($15) {
  __ZN4love6thread10ThreadableD2Ev($this); //@line 63 "modules/video/theora/Video.cpp"
  ___resumeException($$01|0);
  // unreachable;
 }
 $16 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ($16|0)==($14|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($17) {
   $31 = $14;
  } else {
   $19 = $16;
   while(1) {
    $18 = ((($19)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$1>>2] = $18; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $20 = HEAP32[$18>>2]|0; //@line 106 "./common/Object.h"
    $21 = ($20|0)==(0|0); //@line 106 "./common/Object.h"
    if ($21) {
     $23 = $18;
    } else {
     $24 = HEAP32[$20>>2]|0; //@line 106 "./common/Object.h"
     $25 = ((($24)) + 12|0); //@line 106 "./common/Object.h"
     $26 = HEAP32[$25>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($26|0,($20|0)); //@line 106 "./common/Object.h"
     $27 = __THREW__; __THREW__ = 0;
     $28 = $27&1;
     if ($28) {
      label = 12;
      break;
     }
     $$pre$i$i$i = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $23 = $$pre$i$i$i;
    }
    $22 = ($23|0)==($14|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($22) {
     label = 13;
     break;
    } else {
     $19 = $23;
    }
   }
   if ((label|0) == 12) {
    $29 = ___cxa_find_matching_catch_3(0|0)|0;
    $30 = tempRet0;
    ___clang_call_terminate($29); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 13) {
    $$pre$i = HEAP32[$0>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $31 = $$pre$i;
    break;
   }
  }
 } while(0);
 __ZdlPv($31); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 __ZN4love6thread10ThreadableD2Ev($this); //@line 63 "modules/video/theora/Video.cpp"
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN4love5video6theora11VideoStreamD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19764); //@line 72 "modules/video/theora/VideoStream.cpp"
 $0 = ((($this)) + 528|0); //@line 73 "modules/video/theora/VideoStream.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 73 "modules/video/theora/VideoStream.cpp"
 $2 = ($1|0)==(0|0); //@line 73 "modules/video/theora/VideoStream.cpp"
 if ($2) {
  label = 10;
 } else {
  __THREW__ = 0;
  invoke_vi(3789,($1|0)); //@line 74 "modules/video/theora/VideoStream.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if (!($4)) {
   label = 10;
  }
 }
 do {
  if ((label|0) == 10) {
   $28 = ((($this)) + 464|0); //@line 76 "modules/video/theora/VideoStream.cpp"
   __THREW__ = 0;
   invoke_vi(3790,($28|0)); //@line 76 "modules/video/theora/VideoStream.cpp"
   $29 = __THREW__; __THREW__ = 0;
   $30 = $29&1;
   if (!($30)) {
    $31 = ((($this)) + 16|0); //@line 77 "modules/video/theora/VideoStream.cpp"
    $32 = HEAP8[$31>>0]|0; //@line 77 "modules/video/theora/VideoStream.cpp"
    $33 = ($32<<24>>24)==(0); //@line 77 "modules/video/theora/VideoStream.cpp"
    if (!($33)) {
     $34 = ((($this)) + 56|0); //@line 78 "modules/video/theora/VideoStream.cpp"
     __THREW__ = 0;
     (invoke_ii(3791,($34|0))|0); //@line 78 "modules/video/theora/VideoStream.cpp"
     $35 = __THREW__; __THREW__ = 0;
     $36 = $35&1;
     if ($36) {
      break;
     }
    }
    $37 = ((($this)) + 24|0); //@line 80 "modules/video/theora/VideoStream.cpp"
    __THREW__ = 0;
    (invoke_ii(3792,($37|0))|0); //@line 80 "modules/video/theora/VideoStream.cpp"
    $38 = __THREW__; __THREW__ = 0;
    $39 = $38&1;
    if (!($39)) {
     $40 = ((($this)) + 532|0); //@line 82 "modules/video/theora/VideoStream.cpp"
     $41 = HEAP32[$40>>2]|0; //@line 82 "modules/video/theora/VideoStream.cpp"
     $42 = ($41|0)==(0|0); //@line 82 "modules/video/theora/VideoStream.cpp"
     if (!($42)) {
      __ZN4love5video11VideoStream5FrameD2Ev($41); //@line 82 "modules/video/theora/VideoStream.cpp"
      __ZdlPv($41); //@line 82 "modules/video/theora/VideoStream.cpp"
     }
     $43 = ((($this)) + 536|0); //@line 83 "modules/video/theora/VideoStream.cpp"
     $44 = HEAP32[$43>>2]|0; //@line 83 "modules/video/theora/VideoStream.cpp"
     $45 = ($44|0)==(0|0); //@line 83 "modules/video/theora/VideoStream.cpp"
     if (!($45)) {
      __ZN4love5video11VideoStream5FrameD2Ev($44); //@line 83 "modules/video/theora/VideoStream.cpp"
      __ZdlPv($44); //@line 83 "modules/video/theora/VideoStream.cpp"
     }
     $46 = ((($this)) + 556|0); //@line 84 "modules/video/theora/VideoStream.cpp"
     __ZN4love6thread8MutexRefD2Ev($46); //@line 84 "modules/video/theora/VideoStream.cpp"
     $47 = ((($this)) + 12|0); //@line 106 "./common/Object.h"
     $48 = HEAP32[$47>>2]|0; //@line 106 "./common/Object.h"
     $49 = ($48|0)==(0|0); //@line 106 "./common/Object.h"
     if (!($49)) {
      $50 = HEAP32[$48>>2]|0; //@line 106 "./common/Object.h"
      $51 = ((($50)) + 12|0); //@line 106 "./common/Object.h"
      $52 = HEAP32[$51>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($52|0,($48|0)); //@line 106 "./common/Object.h"
      $53 = __THREW__; __THREW__ = 0;
      $54 = $53&1;
      if ($54) {
       $55 = ___cxa_find_matching_catch_3(0|0)|0;
       $56 = tempRet0;
       ___clang_call_terminate($55); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
     }
     HEAP32[$this>>2] = (19948); //@line 37 "./modules/video/VideoStream.h"
     $57 = ((($this)) + 8|0); //@line 106 "./common/Object.h"
     $58 = HEAP32[$57>>2]|0; //@line 106 "./common/Object.h"
     $59 = ($58|0)==(0|0); //@line 106 "./common/Object.h"
     if ($59) {
      return; //@line 84 "modules/video/theora/VideoStream.cpp"
     }
     $60 = HEAP32[$58>>2]|0; //@line 106 "./common/Object.h"
     $61 = ((($60)) + 12|0); //@line 106 "./common/Object.h"
     $62 = HEAP32[$61>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($62|0,($58|0)); //@line 106 "./common/Object.h"
     $63 = __THREW__; __THREW__ = 0;
     $64 = $63&1;
     if ($64) {
      $65 = ___cxa_find_matching_catch_3(0|0)|0;
      $66 = tempRet0;
      ___clang_call_terminate($65); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     } else {
      return; //@line 84 "modules/video/theora/VideoStream.cpp"
     }
    }
   }
  }
 } while(0);
 $5 = ___cxa_find_matching_catch_3(0|0)|0;
 $6 = tempRet0;
 $7 = ((($this)) + 556|0); //@line 84 "modules/video/theora/VideoStream.cpp"
 __ZN4love6thread8MutexRefD2Ev($7); //@line 84 "modules/video/theora/VideoStream.cpp"
 $8 = ((($this)) + 12|0); //@line 106 "./common/Object.h"
 $9 = HEAP32[$8>>2]|0; //@line 106 "./common/Object.h"
 $10 = ($9|0)==(0|0); //@line 106 "./common/Object.h"
 if (!($10)) {
  $11 = HEAP32[$9>>2]|0; //@line 106 "./common/Object.h"
  $12 = ((($11)) + 12|0); //@line 106 "./common/Object.h"
  $13 = HEAP32[$12>>2]|0; //@line 106 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($13|0,($9|0)); //@line 106 "./common/Object.h"
  $14 = __THREW__; __THREW__ = 0;
  $15 = $14&1;
  if ($15) {
   $16 = ___cxa_find_matching_catch_3(0|0)|0;
   $17 = tempRet0;
   ___clang_call_terminate($16); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 }
 HEAP32[$this>>2] = (19948); //@line 37 "./modules/video/VideoStream.h"
 $18 = ((($this)) + 8|0); //@line 106 "./common/Object.h"
 $19 = HEAP32[$18>>2]|0; //@line 106 "./common/Object.h"
 $20 = ($19|0)==(0|0); //@line 106 "./common/Object.h"
 if ($20) {
  ___clang_call_terminate($5); //@line 84 "modules/video/theora/VideoStream.cpp"
  // unreachable; //@line 84 "modules/video/theora/VideoStream.cpp"
 }
 $21 = HEAP32[$19>>2]|0; //@line 106 "./common/Object.h"
 $22 = ((($21)) + 12|0); //@line 106 "./common/Object.h"
 $23 = HEAP32[$22>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($23|0,($19|0)); //@line 106 "./common/Object.h"
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $26 = ___cxa_find_matching_catch_3(0|0)|0;
  $27 = tempRet0;
  ___clang_call_terminate($26); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  ___clang_call_terminate($5); //@line 84 "modules/video/theora/VideoStream.cpp"
  // unreachable; //@line 84 "modules/video/theora/VideoStream.cpp"
 }
}
function __ZN4love5video6theora11VideoStreamD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5video6theora11VideoStreamD2Ev($this); //@line 72 "modules/video/theora/VideoStream.cpp"
 __ZdlPv($this); //@line 72 "modules/video/theora/VideoStream.cpp"
 return; //@line 84 "modules/video/theora/VideoStream.cpp"
}
function __ZN4love5video6theora11VideoStream14fillBackBufferEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 393 "modules/video/theora/VideoStream.cpp"
}
function __ZNK4love5video6theora11VideoStream14getFrontBufferEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 532|0); //@line 115 "modules/video/theora/VideoStream.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 115 "modules/video/theora/VideoStream.cpp"
 return ($1|0); //@line 115 "modules/video/theora/VideoStream.cpp"
}
function __ZNK4love5video6theora11VideoStream7getSizeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 28; //@line 120 "modules/video/theora/VideoStream.cpp"
}
function __ZN4love5video6theora11VideoStream11swapBuffersEv($this) {
 $this = $this|0;
 var $$0 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = ((($this)) + 584|0); //@line 397 "modules/video/theora/VideoStream.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 397 "modules/video/theora/VideoStream.cpp"
 $2 = ($1<<24>>24)==(0); //@line 397 "modules/video/theora/VideoStream.cpp"
 if (!($2)) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 410 "modules/video/theora/VideoStream.cpp"
 }
 $3 = ((($this)) + 556|0); //@line 400 "modules/video/theora/VideoStream.cpp"
 $4 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($3)|0); //@line 400 "modules/video/theora/VideoStream.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($l,$4); //@line 400 "modules/video/theora/VideoStream.cpp"
 $5 = ((($this)) + 560|0); //@line 401 "modules/video/theora/VideoStream.cpp"
 $6 = HEAP8[$5>>0]|0; //@line 401 "modules/video/theora/VideoStream.cpp"
 $7 = ($6<<24>>24)==(0); //@line 401 "modules/video/theora/VideoStream.cpp"
 if ($7) {
  $$0 = 0;
 } else {
  HEAP8[$5>>0] = 0; //@line 403 "modules/video/theora/VideoStream.cpp"
  $8 = ((($this)) + 532|0); //@line 405 "modules/video/theora/VideoStream.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 405 "modules/video/theora/VideoStream.cpp"
  $10 = ((($this)) + 536|0); //@line 406 "modules/video/theora/VideoStream.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 406 "modules/video/theora/VideoStream.cpp"
  HEAP32[$8>>2] = $11; //@line 406 "modules/video/theora/VideoStream.cpp"
  HEAP32[$10>>2] = $9; //@line 407 "modules/video/theora/VideoStream.cpp"
  $$0 = 1;
 }
 __ZN4love6thread4LockD2Ev($l); //@line 410 "modules/video/theora/VideoStream.cpp"
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0); //@line 410 "modules/video/theora/VideoStream.cpp"
}
function __ZNK4love5video6theora11VideoStream8getWidthEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 88 "modules/video/theora/VideoStream.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 88 "modules/video/theora/VideoStream.cpp"
 $2 = ($1<<24>>24)==(0); //@line 88 "modules/video/theora/VideoStream.cpp"
 if ($2) {
  $$0 = 0;
  return ($$0|0); //@line 92 "modules/video/theora/VideoStream.cpp"
 }
 $3 = ((($this)) + 476|0); //@line 89 "modules/video/theora/VideoStream.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 89 "modules/video/theora/VideoStream.cpp"
 $$0 = $4;
 return ($$0|0); //@line 92 "modules/video/theora/VideoStream.cpp"
}
function __ZNK4love5video6theora11VideoStream9getHeightEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 96 "modules/video/theora/VideoStream.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 96 "modules/video/theora/VideoStream.cpp"
 $2 = ($1<<24>>24)==(0); //@line 96 "modules/video/theora/VideoStream.cpp"
 if ($2) {
  $$0 = 0;
  return ($$0|0); //@line 100 "modules/video/theora/VideoStream.cpp"
 }
 $3 = ((($this)) + 480|0); //@line 97 "modules/video/theora/VideoStream.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 97 "modules/video/theora/VideoStream.cpp"
 $$0 = $4;
 return ($$0|0); //@line 100 "modules/video/theora/VideoStream.cpp"
}
function __ZNK4love5video6theora11VideoStream11getFilenameEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 104 "modules/video/theora/VideoStream.cpp"
 $3 = ((($2)) + 76|0); //@line 104 "modules/video/theora/VideoStream.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 104 "modules/video/theora/VideoStream.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($1)|0); //@line 104 "modules/video/theora/VideoStream.cpp"
 return ($5|0); //@line 104 "modules/video/theora/VideoStream.cpp"
}
function __ZNK4love5video6theora11VideoStream9isPlayingEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 125 "modules/video/theora/VideoStream.cpp"
 $3 = ((($2)) + 40|0); //@line 125 "modules/video/theora/VideoStream.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 125 "modules/video/theora/VideoStream.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($1)|0); //@line 125 "modules/video/theora/VideoStream.cpp"
 if (!($5)) {
  $9 = 0;
  return ($9|0); //@line 125 "modules/video/theora/VideoStream.cpp"
 }
 $6 = ((($this)) + 584|0); //@line 125 "modules/video/theora/VideoStream.cpp"
 $7 = HEAP8[$6>>0]|0; //@line 125 "modules/video/theora/VideoStream.cpp"
 $8 = ($7<<24>>24)==(0); //@line 125 "modules/video/theora/VideoStream.cpp"
 $9 = $8;
 return ($9|0); //@line 125 "modules/video/theora/VideoStream.cpp"
}
function __ZN4love5video6theora11VideoStream7setSyncEPNS0_11VideoStream9FrameSyncE($this,$frameSync) {
 $this = $this|0;
 $frameSync = $frameSync|0;
 var $$0 = 0, $$01 = 0, $$pre$phi6Z2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = ((($this)) + 556|0); //@line 109 "modules/video/theora/VideoStream.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 109 "modules/video/theora/VideoStream.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($l,$1); //@line 109 "modules/video/theora/VideoStream.cpp"
 $2 = ($frameSync|0)==(0|0); //@line 95 "./common/Object.h"
 do {
  if ($2) {
   label = 5;
  } else {
   $3 = HEAP32[$frameSync>>2]|0; //@line 95 "./common/Object.h"
   $4 = ((($3)) + 8|0); //@line 95 "./common/Object.h"
   $5 = HEAP32[$4>>2]|0; //@line 95 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($5|0,($frameSync|0)); //@line 95 "./common/Object.h"
   $6 = __THREW__; __THREW__ = 0;
   $7 = $6&1;
   if ($7) {
    $30 = ___cxa_find_matching_catch_2()|0;
    $31 = tempRet0;
    $$0 = $31;$$01 = $30;
    __ZN4love6thread4LockD2Ev($l); //@line 111 "modules/video/theora/VideoStream.cpp"
    ___resumeException($$01|0);
    // unreachable;
   } else {
    $8 = HEAP32[$frameSync>>2]|0; //@line 132 "./common/Object.h"
    $9 = ((($8)) + 8|0); //@line 132 "./common/Object.h"
    $10 = HEAP32[$9>>2]|0; //@line 132 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($10|0,($frameSync|0)); //@line 132 "./common/Object.h"
    $11 = __THREW__; __THREW__ = 0;
    $12 = $11&1;
    if (!($12)) {
     label = 5;
     break;
    }
    $13 = ___cxa_find_matching_catch_2()|0;
    $14 = tempRet0;
    $$pre$phi6Z2D = $frameSync;$$pre$phiZ2D = $frameSync;$41 = $14;$42 = $13;
    break;
   }
  }
 } while(0);
 do {
  if ((label|0) == 5) {
   $15 = ((($this)) + 8|0); //@line 133 "./common/Object.h"
   $16 = HEAP32[$15>>2]|0; //@line 133 "./common/Object.h"
   $17 = ($16|0)==(0|0); //@line 133 "./common/Object.h"
   if (!($17)) {
    $18 = HEAP32[$16>>2]|0; //@line 133 "./common/Object.h"
    $19 = ((($18)) + 12|0); //@line 133 "./common/Object.h"
    $20 = HEAP32[$19>>2]|0; //@line 133 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($20|0,($16|0)); //@line 133 "./common/Object.h"
    $21 = __THREW__; __THREW__ = 0;
    $22 = $21&1;
    if ($22) {
     $32 = ___cxa_find_matching_catch_2()|0;
     $33 = tempRet0;
     if ($2) {
      $$0 = $33;$$01 = $32;
      __ZN4love6thread4LockD2Ev($l); //@line 111 "modules/video/theora/VideoStream.cpp"
      ___resumeException($$01|0);
      // unreachable;
     } else {
      $$pre$phi6Z2D = $frameSync;$$pre$phiZ2D = $frameSync;$41 = $33;$42 = $32;
      break;
     }
    }
   }
   HEAP32[$15>>2] = $frameSync; //@line 134 "./common/Object.h"
   if ($2) {
    __ZN4love6thread4LockD2Ev($l); //@line 111 "modules/video/theora/VideoStream.cpp"
    STACKTOP = sp;return; //@line 111 "modules/video/theora/VideoStream.cpp"
   }
   $23 = HEAP32[$frameSync>>2]|0; //@line 106 "./common/Object.h"
   $24 = ((($23)) + 12|0); //@line 106 "./common/Object.h"
   $25 = HEAP32[$24>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($25|0,($frameSync|0)); //@line 106 "./common/Object.h"
   $26 = __THREW__; __THREW__ = 0;
   $27 = $26&1;
   if ($27) {
    $28 = ___cxa_find_matching_catch_3(0|0)|0;
    $29 = tempRet0;
    ___clang_call_terminate($28); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   } else {
    __ZN4love6thread4LockD2Ev($l); //@line 111 "modules/video/theora/VideoStream.cpp"
    STACKTOP = sp;return; //@line 111 "modules/video/theora/VideoStream.cpp"
   }
  }
 } while(0);
 $34 = HEAP32[$$pre$phi6Z2D>>2]|0; //@line 106 "./common/Object.h"
 $35 = ((($34)) + 12|0); //@line 106 "./common/Object.h"
 $36 = HEAP32[$35>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($36|0,($$pre$phiZ2D|0)); //@line 106 "./common/Object.h"
 $37 = __THREW__; __THREW__ = 0;
 $38 = $37&1;
 if ($38) {
  $39 = ___cxa_find_matching_catch_3(0|0)|0;
  $40 = tempRet0;
  ___clang_call_terminate($39); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  $$0 = $41;$$01 = $42;
  __ZN4love6thread4LockD2Ev($l); //@line 111 "modules/video/theora/VideoStream.cpp"
  ___resumeException($$01|0);
  // unreachable;
 }
}
function __ZN4love5video6theora11VideoStreamC2EPNS_10filesystem4FileE($this,$file) {
 $this = $this|0;
 $file = $file|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$3 = 0, $$34 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 31 "./common/Stream.h"
 $0 = ((($this)) + 8|0); //@line 88 "./common/Object.h"
 HEAP32[$0>>2] = 0; //@line 88 "./common/Object.h"
 HEAP32[$this>>2] = (19764); //@line 47 "modules/video/theora/VideoStream.cpp"
 $1 = ((($this)) + 12|0); //@line 93 "./common/Object.h"
 HEAP32[$1>>2] = $file; //@line 93 "./common/Object.h"
 $2 = ($file|0)==(0|0); //@line 95 "./common/Object.h"
 if ($2) {
  label = 3;
 } else {
  $3 = HEAP32[$file>>2]|0; //@line 95 "./common/Object.h"
  $4 = ((($3)) + 8|0); //@line 95 "./common/Object.h"
  $5 = HEAP32[$4>>2]|0; //@line 95 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($5|0,($file|0)); //@line 95 "./common/Object.h"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if ($7) {
   $38 = ___cxa_find_matching_catch_2()|0;
   $39 = tempRet0;
   $$3 = $38;$$34 = $39;
  } else {
   label = 3;
  }
 }
 if ((label|0) == 3) {
  $8 = ((($this)) + 16|0); //@line 38 "modules/video/theora/VideoStream.cpp"
  HEAP8[$8>>0] = 0; //@line 38 "modules/video/theora/VideoStream.cpp"
  $9 = ((($this)) + 17|0); //@line 39 "modules/video/theora/VideoStream.cpp"
  HEAP8[$9>>0] = 0; //@line 39 "modules/video/theora/VideoStream.cpp"
  $10 = ((($this)) + 20|0); //@line 40 "modules/video/theora/VideoStream.cpp"
  HEAP32[$10>>2] = 0; //@line 40 "modules/video/theora/VideoStream.cpp"
  $11 = ((($this)) + 528|0); //@line 41 "modules/video/theora/VideoStream.cpp"
  HEAP32[$11>>2] = 0; //@line 41 "modules/video/theora/VideoStream.cpp"
  $12 = ((($this)) + 556|0); //@line 36 "modules/video/theora/VideoStream.cpp"
  __THREW__ = 0;
  invoke_vi(3788,($12|0)); //@line 36 "modules/video/theora/VideoStream.cpp"
  $13 = __THREW__; __THREW__ = 0;
  $14 = $13&1;
  if ($14) {
   $40 = ___cxa_find_matching_catch_2()|0;
   $41 = tempRet0;
   $$2 = $40;$$23 = $41;
  } else {
   $15 = ((($this)) + 560|0); //@line 42 "modules/video/theora/VideoStream.cpp"
   HEAP8[$15>>0] = 0; //@line 42 "modules/video/theora/VideoStream.cpp"
   $16 = ((($this)) + 568|0); //@line 43 "modules/video/theora/VideoStream.cpp"
   $17 = ((($this)) + 588|0); //@line 46 "modules/video/theora/VideoStream.cpp"
   HEAP32[$17>>2] = 0; //@line 46 "modules/video/theora/VideoStream.cpp"
   $18 = ((($this)) + 24|0); //@line 48 "modules/video/theora/VideoStream.cpp"
   ;HEAP32[$16>>2]=0|0;HEAP32[$16+4>>2]=0|0;HEAP32[$16+8>>2]=0|0;HEAP32[$16+12>>2]=0|0;HEAP8[$16+16>>0]=0|0; //@line 43 "modules/video/theora/VideoStream.cpp"
   __THREW__ = 0;
   (invoke_ii(3793,($18|0))|0); //@line 48 "modules/video/theora/VideoStream.cpp"
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   L9: do {
    if ($20) {
     label = 13;
    } else {
     $21 = ((($this)) + 464|0); //@line 49 "modules/video/theora/VideoStream.cpp"
     __THREW__ = 0;
     invoke_vi(3794,($21|0)); //@line 49 "modules/video/theora/VideoStream.cpp"
     $22 = __THREW__; __THREW__ = 0;
     $23 = $22&1;
     if ($23) {
      label = 13;
     } else {
      __THREW__ = 0;
      $24 = (invoke_ii(3348,28)|0); //@line 51 "modules/video/theora/VideoStream.cpp"
      $25 = __THREW__; __THREW__ = 0;
      $26 = $25&1;
      if ($26) {
       label = 13;
      } else {
       __THREW__ = 0;
       invoke_vi(3795,($24|0)); //@line 51 "modules/video/theora/VideoStream.cpp"
       $27 = __THREW__; __THREW__ = 0;
       $28 = $27&1;
       if ($28) {
        $44 = ___cxa_find_matching_catch_2()|0;
        $45 = tempRet0;
        __ZdlPv($24); //@line 51 "modules/video/theora/VideoStream.cpp"
        $$1 = $44;$$12 = $45;
        break;
       }
       $29 = ((($this)) + 532|0); //@line 51 "modules/video/theora/VideoStream.cpp"
       HEAP32[$29>>2] = $24; //@line 51 "modules/video/theora/VideoStream.cpp"
       __THREW__ = 0;
       $30 = (invoke_ii(3348,28)|0); //@line 52 "modules/video/theora/VideoStream.cpp"
       $31 = __THREW__; __THREW__ = 0;
       $32 = $31&1;
       if ($32) {
        label = 13;
       } else {
        __THREW__ = 0;
        invoke_vi(3795,($30|0)); //@line 52 "modules/video/theora/VideoStream.cpp"
        $33 = __THREW__; __THREW__ = 0;
        $34 = $33&1;
        if ($34) {
         $46 = ___cxa_find_matching_catch_2()|0;
         $47 = tempRet0;
         __ZdlPv($30); //@line 52 "modules/video/theora/VideoStream.cpp"
         $$1 = $46;$$12 = $47;
         break;
        }
        $35 = ((($this)) + 536|0); //@line 52 "modules/video/theora/VideoStream.cpp"
        HEAP32[$35>>2] = $30; //@line 52 "modules/video/theora/VideoStream.cpp"
        __THREW__ = 0;
        invoke_vi(3796,($this|0)); //@line 56 "modules/video/theora/VideoStream.cpp"
        $36 = __THREW__; __THREW__ = 0;
        $37 = $36&1;
        if ($37) {
         $48 = ___cxa_find_matching_catch_3(96|0)|0;
         $49 = tempRet0;
         $50 = (_llvm_eh_typeid_for((96|0))|0); //@line 57 "modules/video/theora/VideoStream.cpp"
         $51 = ($49|0)==($50|0); //@line 57 "modules/video/theora/VideoStream.cpp"
         if (!($51)) {
          $$1 = $48;$$12 = $49;
          break;
         }
         $52 = (___cxa_begin_catch(($48|0))|0); //@line 57 "modules/video/theora/VideoStream.cpp"
         $53 = HEAP32[$35>>2]|0; //@line 60 "modules/video/theora/VideoStream.cpp"
         $54 = ($53|0)==(0|0); //@line 60 "modules/video/theora/VideoStream.cpp"
         if (!($54)) {
          __ZN4love5video11VideoStream5FrameD2Ev($53); //@line 60 "modules/video/theora/VideoStream.cpp"
          __ZdlPv($53); //@line 60 "modules/video/theora/VideoStream.cpp"
         }
         $55 = HEAP32[$29>>2]|0; //@line 61 "modules/video/theora/VideoStream.cpp"
         $56 = ($55|0)==(0|0); //@line 61 "modules/video/theora/VideoStream.cpp"
         if (!($56)) {
          __ZN4love5video11VideoStream5FrameD2Ev($55); //@line 61 "modules/video/theora/VideoStream.cpp"
          __ZdlPv($55); //@line 61 "modules/video/theora/VideoStream.cpp"
         }
         __THREW__ = 0;
         invoke_vi(3790,($21|0)); //@line 62 "modules/video/theora/VideoStream.cpp"
         $57 = __THREW__; __THREW__ = 0;
         $58 = $57&1;
         do {
          if ($58) {
           label = 25;
          } else {
           __THREW__ = 0;
           (invoke_ii(3792,($18|0))|0); //@line 63 "modules/video/theora/VideoStream.cpp"
           $59 = __THREW__; __THREW__ = 0;
           $60 = $59&1;
           if ($60) {
            label = 25;
           } else {
            $61 = (___cxa_allocate_exception(16)|0); //@line 64 "modules/video/theora/VideoStream.cpp"
            HEAP32[$61>>2] = (12260); //@line 36 "./common/Exception.h"
            $62 = ((($61)) + 4|0); //@line 36 "./common/Exception.h"
            $63 = ((($52)) + 4|0); //@line 36 "./common/Exception.h"
            __THREW__ = 0;
            invoke_vii(3351,($62|0),($63|0)); //@line 36 "./common/Exception.h"
            $64 = __THREW__; __THREW__ = 0;
            $65 = $64&1;
            if ($65) {
             $69 = ___cxa_find_matching_catch_2()|0;
             $70 = tempRet0;
             ___cxa_free_exception(($61|0)); //@line 64 "modules/video/theora/VideoStream.cpp"
             $$0 = $69;$$01 = $70;
             break;
            } else {
             __THREW__ = 0;
             invoke_viii(3350,($61|0),(96|0),(344|0)); //@line 64 "modules/video/theora/VideoStream.cpp"
             $66 = __THREW__; __THREW__ = 0;
             label = 25;
             break;
            }
           }
          }
         } while(0);
         if ((label|0) == 25) {
          $67 = ___cxa_find_matching_catch_2()|0;
          $68 = tempRet0;
          $$0 = $67;$$01 = $68;
         }
         __THREW__ = 0;
         invoke_v(3378); //@line 65 "modules/video/theora/VideoStream.cpp"
         $71 = __THREW__; __THREW__ = 0;
         $72 = $71&1;
         if (!($72)) {
          $$1 = $$0;$$12 = $$01;
          break;
         }
         $137 = ___cxa_find_matching_catch_3(0|0)|0;
         $138 = tempRet0;
         ___clang_call_terminate($137); //@line 65 "modules/video/theora/VideoStream.cpp"
         // unreachable; //@line 65 "modules/video/theora/VideoStream.cpp"
        } else {
         __THREW__ = 0;
         $73 = (invoke_ii(3348,40)|0); //@line 67 "modules/video/theora/VideoStream.cpp"
         $74 = __THREW__; __THREW__ = 0;
         $75 = $74&1;
         if ($75) {
          label = 13;
          break;
         }
         __THREW__ = 0;
         invoke_vi(3467,($73|0)); //@line 67 "modules/video/theora/VideoStream.cpp"
         $76 = __THREW__; __THREW__ = 0;
         $77 = $76&1;
         if ($77) {
          $108 = ___cxa_find_matching_catch_2()|0;
          $109 = tempRet0;
          __ZdlPv($73); //@line 67 "modules/video/theora/VideoStream.cpp"
          $$1 = $108;$$12 = $109;
          break;
         }
         $78 = HEAP32[$73>>2]|0; //@line 95 "./common/Object.h"
         $79 = ((($78)) + 8|0); //@line 95 "./common/Object.h"
         $80 = HEAP32[$79>>2]|0; //@line 95 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($80|0,($73|0)); //@line 95 "./common/Object.h"
         $81 = __THREW__; __THREW__ = 0;
         $82 = $81&1;
         if ($82) {
          label = 13;
          break;
         }
         $83 = HEAP32[$73>>2]|0; //@line 132 "./common/Object.h"
         $84 = ((($83)) + 8|0); //@line 132 "./common/Object.h"
         $85 = HEAP32[$84>>2]|0; //@line 132 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($85|0,($73|0)); //@line 132 "./common/Object.h"
         $86 = __THREW__; __THREW__ = 0;
         $87 = $86&1;
         do {
          if (!($87)) {
           $88 = HEAP32[$0>>2]|0; //@line 133 "./common/Object.h"
           $89 = ($88|0)==(0|0); //@line 133 "./common/Object.h"
           if (!($89)) {
            $90 = HEAP32[$88>>2]|0; //@line 133 "./common/Object.h"
            $91 = ((($90)) + 12|0); //@line 133 "./common/Object.h"
            $92 = HEAP32[$91>>2]|0; //@line 133 "./common/Object.h"
            __THREW__ = 0;
            invoke_vi($92|0,($88|0)); //@line 133 "./common/Object.h"
            $93 = __THREW__; __THREW__ = 0;
            $94 = $93&1;
            if ($94) {
             break;
            }
           }
           HEAP32[$0>>2] = $73; //@line 134 "./common/Object.h"
           $95 = HEAP32[$73>>2]|0; //@line 106 "./common/Object.h"
           $96 = ((($95)) + 12|0); //@line 106 "./common/Object.h"
           $97 = HEAP32[$96>>2]|0; //@line 106 "./common/Object.h"
           __THREW__ = 0;
           invoke_vi($97|0,($73|0)); //@line 106 "./common/Object.h"
           $98 = __THREW__; __THREW__ = 0;
           $99 = $98&1;
           if ($99) {
            $100 = ___cxa_find_matching_catch_3(0|0)|0;
            $101 = tempRet0;
            ___clang_call_terminate($100); //@line 106 "./common/Object.h"
            // unreachable; //@line 106 "./common/Object.h"
           }
           $102 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
           $103 = HEAP32[$102>>2]|0; //@line 68 "modules/video/theora/VideoStream.cpp"
           $104 = ((($103)) + 12|0); //@line 68 "modules/video/theora/VideoStream.cpp"
           $105 = HEAP32[$104>>2]|0; //@line 68 "modules/video/theora/VideoStream.cpp"
           __THREW__ = 0;
           invoke_vi($105|0,($102|0)); //@line 68 "modules/video/theora/VideoStream.cpp"
           $106 = __THREW__; __THREW__ = 0;
           $107 = $106&1;
           if ($107) {
            label = 13;
            break L9;
           }
           return; //@line 69 "modules/video/theora/VideoStream.cpp"
          }
         } while(0);
         $110 = ___cxa_find_matching_catch_2()|0;
         $111 = tempRet0;
         $112 = HEAP32[$73>>2]|0; //@line 106 "./common/Object.h"
         $113 = ((($112)) + 12|0); //@line 106 "./common/Object.h"
         $114 = HEAP32[$113>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($114|0,($73|0)); //@line 106 "./common/Object.h"
         $115 = __THREW__; __THREW__ = 0;
         $116 = $115&1;
         if (!($116)) {
          $$1 = $110;$$12 = $111;
          break;
         }
         $117 = ___cxa_find_matching_catch_3(0|0)|0;
         $118 = tempRet0;
         ___clang_call_terminate($117); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 13) {
    $42 = ___cxa_find_matching_catch_2()|0;
    $43 = tempRet0;
    $$1 = $42;$$12 = $43;
   }
   __ZN4love6thread8MutexRefD2Ev($12); //@line 69 "modules/video/theora/VideoStream.cpp"
   $$2 = $$1;$$23 = $$12;
  }
  $119 = HEAP32[$1>>2]|0; //@line 106 "./common/Object.h"
  $120 = ($119|0)==(0|0); //@line 106 "./common/Object.h"
  if ($120) {
   $$3 = $$2;$$34 = $$23;
  } else {
   $121 = HEAP32[$119>>2]|0; //@line 106 "./common/Object.h"
   $122 = ((($121)) + 12|0); //@line 106 "./common/Object.h"
   $123 = HEAP32[$122>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($123|0,($119|0)); //@line 106 "./common/Object.h"
   $124 = __THREW__; __THREW__ = 0;
   $125 = $124&1;
   if ($125) {
    $126 = ___cxa_find_matching_catch_3(0|0)|0;
    $127 = tempRet0;
    ___clang_call_terminate($126); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   } else {
    $$3 = $$2;$$34 = $$23;
   }
  }
 }
 HEAP32[$this>>2] = (19948); //@line 37 "./modules/video/VideoStream.h"
 $128 = HEAP32[$0>>2]|0; //@line 106 "./common/Object.h"
 $129 = ($128|0)==(0|0); //@line 106 "./common/Object.h"
 if ($129) {
  ___resumeException($$3|0);
  // unreachable;
 }
 $130 = HEAP32[$128>>2]|0; //@line 106 "./common/Object.h"
 $131 = ((($130)) + 12|0); //@line 106 "./common/Object.h"
 $132 = HEAP32[$131>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($132|0,($128|0)); //@line 106 "./common/Object.h"
 $133 = __THREW__; __THREW__ = 0;
 $134 = $133&1;
 if ($134) {
  $135 = ___cxa_find_matching_catch_3(0|0)|0;
  $136 = tempRet0;
  ___clang_call_terminate($135); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  ___resumeException($$3|0);
  // unreachable;
 }
}
function __ZN4love5video6theora11VideoStream11parseHeaderEv($this) {
 $this = $this|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $buffers = 0, $comment = 0, $setupInfo = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $comment = sp + 24|0;
 $setupInfo = sp + 16|0;
 $buffers = sp + 8|0;
 $0 = ((($this)) + 16|0); //@line 187 "modules/video/theora/VideoStream.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 187 "modules/video/theora/VideoStream.cpp"
 $2 = ($1<<24>>24)==(0); //@line 187 "modules/video/theora/VideoStream.cpp"
 if (!($2)) {
  STACKTOP = sp;return; //@line 248 "modules/video/theora/VideoStream.cpp"
 }
 HEAP32[$setupInfo>>2] = 0; //@line 191 "modules/video/theora/VideoStream.cpp"
 _th_comment_init($comment); //@line 192 "modules/video/theora/VideoStream.cpp"
 $3 = ((($this)) + 464|0); //@line 198 "modules/video/theora/VideoStream.cpp"
 $4 = ((($this)) + 432|0); //@line 198 "modules/video/theora/VideoStream.cpp"
 $5 = ((($this)) + 56|0); //@line 201 "modules/video/theora/VideoStream.cpp"
 $6 = ((($this)) + 17|0); //@line 202 "modules/video/theora/VideoStream.cpp"
 $7 = ((($this)) + 416|0); //@line 204 "modules/video/theora/VideoStream.cpp"
 while(1) {
  (__ZN4love5video6theora11VideoStream10readPacketEb($this,0)|0); //@line 197 "modules/video/theora/VideoStream.cpp"
  $8 = (_th_decode_headerin($3,$comment,$setupInfo,$4)|0); //@line 198 "modules/video/theora/VideoStream.cpp"
  $9 = ($8|0)==(-21); //@line 199 "modules/video/theora/VideoStream.cpp"
  if ($9) {
   (_ogg_stream_clear($5)|0); //@line 201 "modules/video/theora/VideoStream.cpp"
   HEAP8[$6>>0] = 0; //@line 202 "modules/video/theora/VideoStream.cpp"
  } else {
   $10 = ($8|0)<(0); //@line 204 "modules/video/theora/VideoStream.cpp"
   if (!($10)) {
    $$lcssa = $8;
    break;
   }
  }
  $12 = (_ogg_page_eos($7)|0); //@line 204 "modules/video/theora/VideoStream.cpp"
  $13 = ($12|0)==(0); //@line 204 "modules/video/theora/VideoStream.cpp"
  if (!($13)) {
   label = 8;
   break;
  }
 }
 if ((label|0) == 8) {
  _th_comment_clear($comment); //@line 208 "modules/video/theora/VideoStream.cpp"
  $14 = (___cxa_allocate_exception(16)|0); //@line 209 "modules/video/theora/VideoStream.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($14|0),(360177|0),($vararg_buffer|0)); //@line 209 "modules/video/theora/VideoStream.cpp"
  $15 = __THREW__; __THREW__ = 0;
  $16 = $15&1;
  if ($16) {
   $17 = ___cxa_find_matching_catch_2()|0;
   $18 = tempRet0;
   ___cxa_free_exception(($14|0)); //@line 209 "modules/video/theora/VideoStream.cpp"
   ___resumeException($17|0);
   // unreachable;
  } else {
   ___cxa_throw(($14|0),(96|0),(344|0)); //@line 209 "modules/video/theora/VideoStream.cpp"
   // unreachable; //@line 209 "modules/video/theora/VideoStream.cpp"
  }
 }
 $11 = ($$lcssa|0)>(0); //@line 212 "modules/video/theora/VideoStream.cpp"
 if ($11) {
  while(1) {
   (__ZN4love5video6theora11VideoStream10readPacketEb($this,0)|0); //@line 214 "modules/video/theora/VideoStream.cpp"
   $19 = (_th_decode_headerin($3,$comment,$setupInfo,$4)|0); //@line 215 "modules/video/theora/VideoStream.cpp"
   $20 = ($19|0)>(0); //@line 212 "modules/video/theora/VideoStream.cpp"
   if (!($20)) {
    break;
   }
  }
 }
 _th_comment_clear($comment); //@line 218 "modules/video/theora/VideoStream.cpp"
 $21 = HEAP32[$setupInfo>>2]|0; //@line 220 "modules/video/theora/VideoStream.cpp"
 $22 = (_th_decode_alloc($3,$21)|0); //@line 220 "modules/video/theora/VideoStream.cpp"
 $23 = ((($this)) + 528|0); //@line 220 "modules/video/theora/VideoStream.cpp"
 HEAP32[$23>>2] = $22; //@line 220 "modules/video/theora/VideoStream.cpp"
 $24 = HEAP32[$setupInfo>>2]|0; //@line 221 "modules/video/theora/VideoStream.cpp"
 _th_setup_free($24); //@line 221 "modules/video/theora/VideoStream.cpp"
 $25 = ((($this)) + 536|0); //@line 223 "modules/video/theora/VideoStream.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 223 "modules/video/theora/VideoStream.cpp"
 HEAP32[$buffers>>2] = $26; //@line 223 "modules/video/theora/VideoStream.cpp"
 $27 = ((($buffers)) + 4|0); //@line 223 "modules/video/theora/VideoStream.cpp"
 $28 = ((($this)) + 532|0); //@line 223 "modules/video/theora/VideoStream.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 223 "modules/video/theora/VideoStream.cpp"
 HEAP32[$27>>2] = $29; //@line 223 "modules/video/theora/VideoStream.cpp"
 $30 = ((($this)) + 484|0); //@line 225 "modules/video/theora/VideoStream.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 178 "modules/video/theora/VideoStream.cpp"
 $32 = ((($this)) + 544|0); //@line 225 "modules/video/theora/VideoStream.cpp"
 HEAP32[$32>>2] = $31; //@line 225 "modules/video/theora/VideoStream.cpp"
 $33 = ((($this)) + 540|0); //@line 225 "modules/video/theora/VideoStream.cpp"
 HEAP32[$33>>2] = $31; //@line 225 "modules/video/theora/VideoStream.cpp"
 $34 = ((($this)) + 488|0); //@line 226 "modules/video/theora/VideoStream.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 176 "modules/video/theora/VideoStream.cpp"
 $36 = ((($this)) + 552|0); //@line 226 "modules/video/theora/VideoStream.cpp"
 HEAP32[$36>>2] = $35; //@line 226 "modules/video/theora/VideoStream.cpp"
 $37 = ((($this)) + 548|0); //@line 226 "modules/video/theora/VideoStream.cpp"
 HEAP32[$37>>2] = $35; //@line 226 "modules/video/theora/VideoStream.cpp"
 $38 = ((($this)) + 512|0); //@line 228 "modules/video/theora/VideoStream.cpp"
 $39 = HEAP32[$38>>2]|0; //@line 235 "modules/video/theora/VideoStream.cpp"
 $40 = $26; //@line 232 "modules/video/theora/VideoStream.cpp"
 switch ($39|0) {
 case 0:  {
  $41 = $35 >>> 1; //@line 176 "modules/video/theora/VideoStream.cpp"
  HEAP32[$36>>2] = $41; //@line 176 "modules/video/theora/VideoStream.cpp"
  label = 14;
  break;
 }
 case 2:  {
  label = 14;
  break;
 }
 default: {
 }
 }
 if ((label|0) == 14) {
  $42 = $31 >>> 1; //@line 178 "modules/video/theora/VideoStream.cpp"
  HEAP32[$32>>2] = $42; //@line 178 "modules/video/theora/VideoStream.cpp"
 }
 $43 = ((($this)) + 476|0); //@line 232 "modules/video/theora/VideoStream.cpp"
 $44 = ((($this)) + 480|0); //@line 233 "modules/video/theora/VideoStream.cpp"
 $45 = HEAP32[$43>>2]|0; //@line 237 "modules/video/theora/VideoStream.cpp"
 HEAP32[$40>>2] = $45; //@line 232 "modules/video/theora/VideoStream.cpp"
 $46 = ((($40)) + 12|0); //@line 232 "modules/video/theora/VideoStream.cpp"
 HEAP32[$46>>2] = $45; //@line 232 "modules/video/theora/VideoStream.cpp"
 $47 = HEAP32[$44>>2]|0; //@line 237 "modules/video/theora/VideoStream.cpp"
 $48 = ((($40)) + 4|0); //@line 233 "modules/video/theora/VideoStream.cpp"
 HEAP32[$48>>2] = $47; //@line 233 "modules/video/theora/VideoStream.cpp"
 $49 = ((($40)) + 16|0); //@line 233 "modules/video/theora/VideoStream.cpp"
 HEAP32[$49>>2] = $47; //@line 233 "modules/video/theora/VideoStream.cpp"
 switch ($39|0) {
 case 0:  {
  $50 = (($47|0) / 2)&-1; //@line 176 "modules/video/theora/VideoStream.cpp"
  HEAP32[$49>>2] = $50; //@line 176 "modules/video/theora/VideoStream.cpp"
  label = 17;
  break;
 }
 case 2:  {
  label = 17;
  break;
 }
 default: {
 }
 }
 if ((label|0) == 17) {
  $51 = (($45|0) / 2)&-1; //@line 178 "modules/video/theora/VideoStream.cpp"
  HEAP32[$46>>2] = $51; //@line 178 "modules/video/theora/VideoStream.cpp"
 }
 $52 = Math_imul($47, $45)|0; //@line 237 "modules/video/theora/VideoStream.cpp"
 $53 = ($52|0)>(-1); //@line 237 "modules/video/theora/VideoStream.cpp"
 $54 = $53 ? $52 : -1; //@line 237 "modules/video/theora/VideoStream.cpp"
 $55 = (__Znaj($54)|0); //@line 237 "modules/video/theora/VideoStream.cpp"
 $56 = ((($40)) + 8|0); //@line 237 "modules/video/theora/VideoStream.cpp"
 HEAP32[$56>>2] = $55; //@line 237 "modules/video/theora/VideoStream.cpp"
 $57 = HEAP32[$46>>2]|0; //@line 238 "modules/video/theora/VideoStream.cpp"
 $58 = HEAP32[$49>>2]|0; //@line 238 "modules/video/theora/VideoStream.cpp"
 $59 = Math_imul($58, $57)|0; //@line 238 "modules/video/theora/VideoStream.cpp"
 $60 = ($59|0)>(-1); //@line 238 "modules/video/theora/VideoStream.cpp"
 $61 = $60 ? $59 : -1; //@line 238 "modules/video/theora/VideoStream.cpp"
 $62 = (__Znaj($61)|0); //@line 238 "modules/video/theora/VideoStream.cpp"
 $63 = HEAP32[$buffers>>2]|0; //@line 238 "modules/video/theora/VideoStream.cpp"
 $64 = ((($63)) + 20|0); //@line 238 "modules/video/theora/VideoStream.cpp"
 HEAP32[$64>>2] = $62; //@line 238 "modules/video/theora/VideoStream.cpp"
 $65 = ((($63)) + 12|0); //@line 239 "modules/video/theora/VideoStream.cpp"
 $66 = HEAP32[$65>>2]|0; //@line 239 "modules/video/theora/VideoStream.cpp"
 $67 = ((($63)) + 16|0); //@line 239 "modules/video/theora/VideoStream.cpp"
 $68 = HEAP32[$67>>2]|0; //@line 239 "modules/video/theora/VideoStream.cpp"
 $69 = Math_imul($68, $66)|0; //@line 239 "modules/video/theora/VideoStream.cpp"
 $70 = ($69|0)>(-1); //@line 239 "modules/video/theora/VideoStream.cpp"
 $71 = $70 ? $69 : -1; //@line 239 "modules/video/theora/VideoStream.cpp"
 $72 = (__Znaj($71)|0); //@line 239 "modules/video/theora/VideoStream.cpp"
 $73 = ((($63)) + 24|0); //@line 239 "modules/video/theora/VideoStream.cpp"
 HEAP32[$73>>2] = $72; //@line 239 "modules/video/theora/VideoStream.cpp"
 $74 = ((($63)) + 8|0); //@line 241 "modules/video/theora/VideoStream.cpp"
 $75 = HEAP32[$74>>2]|0; //@line 241 "modules/video/theora/VideoStream.cpp"
 $76 = HEAP32[$63>>2]|0; //@line 241 "modules/video/theora/VideoStream.cpp"
 $77 = ((($63)) + 4|0); //@line 241 "modules/video/theora/VideoStream.cpp"
 $78 = HEAP32[$77>>2]|0; //@line 241 "modules/video/theora/VideoStream.cpp"
 $79 = Math_imul($78, $76)|0; //@line 241 "modules/video/theora/VideoStream.cpp"
 _memset(($75|0),16,($79|0))|0; //@line 241 "modules/video/theora/VideoStream.cpp"
 $80 = HEAP32[$64>>2]|0; //@line 242 "modules/video/theora/VideoStream.cpp"
 $81 = HEAP32[$65>>2]|0; //@line 242 "modules/video/theora/VideoStream.cpp"
 $82 = HEAP32[$67>>2]|0; //@line 242 "modules/video/theora/VideoStream.cpp"
 $83 = Math_imul($82, $81)|0; //@line 242 "modules/video/theora/VideoStream.cpp"
 _memset(($80|0),-128,($83|0))|0; //@line 242 "modules/video/theora/VideoStream.cpp"
 $84 = HEAP32[$73>>2]|0; //@line 243 "modules/video/theora/VideoStream.cpp"
 $85 = HEAP32[$65>>2]|0; //@line 243 "modules/video/theora/VideoStream.cpp"
 $86 = HEAP32[$67>>2]|0; //@line 243 "modules/video/theora/VideoStream.cpp"
 $87 = Math_imul($86, $85)|0; //@line 243 "modules/video/theora/VideoStream.cpp"
 _memset(($84|0),-128,($87|0))|0; //@line 243 "modules/video/theora/VideoStream.cpp"
 $88 = HEAP32[$43>>2]|0; //@line 237 "modules/video/theora/VideoStream.cpp"
 $89 = HEAP32[$27>>2]|0; //@line 243 "modules/video/theora/VideoStream.cpp"
 HEAP32[$89>>2] = $88; //@line 232 "modules/video/theora/VideoStream.cpp"
 $90 = ((($89)) + 12|0); //@line 232 "modules/video/theora/VideoStream.cpp"
 HEAP32[$90>>2] = $88; //@line 232 "modules/video/theora/VideoStream.cpp"
 $91 = HEAP32[$44>>2]|0; //@line 237 "modules/video/theora/VideoStream.cpp"
 $92 = ((($89)) + 4|0); //@line 233 "modules/video/theora/VideoStream.cpp"
 HEAP32[$92>>2] = $91; //@line 233 "modules/video/theora/VideoStream.cpp"
 $93 = ((($89)) + 16|0); //@line 233 "modules/video/theora/VideoStream.cpp"
 HEAP32[$93>>2] = $91; //@line 233 "modules/video/theora/VideoStream.cpp"
 $94 = HEAP32[$38>>2]|0; //@line 235 "modules/video/theora/VideoStream.cpp"
 switch ($94|0) {
 case 0:  {
  $95 = (($91|0) / 2)&-1; //@line 176 "modules/video/theora/VideoStream.cpp"
  HEAP32[$93>>2] = $95; //@line 176 "modules/video/theora/VideoStream.cpp"
  label = 21;
  break;
 }
 case 2:  {
  label = 21;
  break;
 }
 default: {
 }
 }
 if ((label|0) == 21) {
  $96 = (($88|0) / 2)&-1; //@line 178 "modules/video/theora/VideoStream.cpp"
  HEAP32[$90>>2] = $96; //@line 178 "modules/video/theora/VideoStream.cpp"
 }
 $97 = Math_imul($91, $88)|0; //@line 237 "modules/video/theora/VideoStream.cpp"
 $98 = ($97|0)>(-1); //@line 237 "modules/video/theora/VideoStream.cpp"
 $99 = $98 ? $97 : -1; //@line 237 "modules/video/theora/VideoStream.cpp"
 $100 = (__Znaj($99)|0); //@line 237 "modules/video/theora/VideoStream.cpp"
 $101 = ((($89)) + 8|0); //@line 237 "modules/video/theora/VideoStream.cpp"
 HEAP32[$101>>2] = $100; //@line 237 "modules/video/theora/VideoStream.cpp"
 $102 = HEAP32[$90>>2]|0; //@line 238 "modules/video/theora/VideoStream.cpp"
 $103 = HEAP32[$93>>2]|0; //@line 238 "modules/video/theora/VideoStream.cpp"
 $104 = Math_imul($103, $102)|0; //@line 238 "modules/video/theora/VideoStream.cpp"
 $105 = ($104|0)>(-1); //@line 238 "modules/video/theora/VideoStream.cpp"
 $106 = $105 ? $104 : -1; //@line 238 "modules/video/theora/VideoStream.cpp"
 $107 = (__Znaj($106)|0); //@line 238 "modules/video/theora/VideoStream.cpp"
 $108 = ((($89)) + 20|0); //@line 238 "modules/video/theora/VideoStream.cpp"
 HEAP32[$108>>2] = $107; //@line 238 "modules/video/theora/VideoStream.cpp"
 $109 = HEAP32[$90>>2]|0; //@line 239 "modules/video/theora/VideoStream.cpp"
 $110 = HEAP32[$93>>2]|0; //@line 239 "modules/video/theora/VideoStream.cpp"
 $111 = Math_imul($110, $109)|0; //@line 239 "modules/video/theora/VideoStream.cpp"
 $112 = ($111|0)>(-1); //@line 239 "modules/video/theora/VideoStream.cpp"
 $113 = $112 ? $111 : -1; //@line 239 "modules/video/theora/VideoStream.cpp"
 $114 = (__Znaj($113)|0); //@line 239 "modules/video/theora/VideoStream.cpp"
 $115 = ((($89)) + 24|0); //@line 239 "modules/video/theora/VideoStream.cpp"
 HEAP32[$115>>2] = $114; //@line 239 "modules/video/theora/VideoStream.cpp"
 $116 = HEAP32[$101>>2]|0; //@line 241 "modules/video/theora/VideoStream.cpp"
 $117 = HEAP32[$89>>2]|0; //@line 241 "modules/video/theora/VideoStream.cpp"
 $118 = HEAP32[$92>>2]|0; //@line 241 "modules/video/theora/VideoStream.cpp"
 $119 = Math_imul($118, $117)|0; //@line 241 "modules/video/theora/VideoStream.cpp"
 _memset(($116|0),16,($119|0))|0; //@line 241 "modules/video/theora/VideoStream.cpp"
 $120 = HEAP32[$108>>2]|0; //@line 242 "modules/video/theora/VideoStream.cpp"
 $121 = HEAP32[$90>>2]|0; //@line 242 "modules/video/theora/VideoStream.cpp"
 $122 = HEAP32[$93>>2]|0; //@line 242 "modules/video/theora/VideoStream.cpp"
 $123 = Math_imul($122, $121)|0; //@line 242 "modules/video/theora/VideoStream.cpp"
 _memset(($120|0),-128,($123|0))|0; //@line 242 "modules/video/theora/VideoStream.cpp"
 $124 = HEAP32[$115>>2]|0; //@line 243 "modules/video/theora/VideoStream.cpp"
 $125 = HEAP32[$90>>2]|0; //@line 243 "modules/video/theora/VideoStream.cpp"
 $126 = HEAP32[$93>>2]|0; //@line 243 "modules/video/theora/VideoStream.cpp"
 $127 = Math_imul($126, $125)|0; //@line 243 "modules/video/theora/VideoStream.cpp"
 _memset(($124|0),-128,($127|0))|0; //@line 243 "modules/video/theora/VideoStream.cpp"
 HEAP8[$0>>0] = 1; //@line 246 "modules/video/theora/VideoStream.cpp"
 $128 = HEAP32[$23>>2]|0; //@line 247 "modules/video/theora/VideoStream.cpp"
 (_th_decode_packetin($128,$4,0)|0); //@line 247 "modules/video/theora/VideoStream.cpp"
 STACKTOP = sp;return; //@line 248 "modules/video/theora/VideoStream.cpp"
}
function __ZN4love5video6theora11VideoStream10readPacketEb($this,$mustSucceed) {
 $this = $this|0;
 $mustSucceed = $mustSucceed|0;
 var $$0 = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 17|0); //@line 144 "modules/video/theora/VideoStream.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 144 "modules/video/theora/VideoStream.cpp"
 $2 = ($1<<24>>24)==(0); //@line 144 "modules/video/theora/VideoStream.cpp"
 if ($2) {
  __ZN4love5video6theora11VideoStream8readPageEv($this); //@line 146 "modules/video/theora/VideoStream.cpp"
  $3 = ((($this)) + 416|0); //@line 147 "modules/video/theora/VideoStream.cpp"
  $4 = (_ogg_page_serialno($3)|0); //@line 147 "modules/video/theora/VideoStream.cpp"
  $5 = ((($this)) + 20|0); //@line 147 "modules/video/theora/VideoStream.cpp"
  HEAP32[$5>>2] = $4; //@line 147 "modules/video/theora/VideoStream.cpp"
  $6 = ((($this)) + 56|0); //@line 148 "modules/video/theora/VideoStream.cpp"
  (_ogg_stream_init($6,$4)|0); //@line 148 "modules/video/theora/VideoStream.cpp"
  HEAP8[$0>>0] = 1; //@line 149 "modules/video/theora/VideoStream.cpp"
  (_ogg_stream_pagein($6,$3)|0); //@line 150 "modules/video/theora/VideoStream.cpp"
  $$pre$phiZ2D = $6;
 } else {
  $$pre = ((($this)) + 56|0); //@line 153 "modules/video/theora/VideoStream.cpp"
  $$pre$phiZ2D = $$pre;
 }
 $7 = ((($this)) + 432|0); //@line 153 "modules/video/theora/VideoStream.cpp"
 $8 = (_ogg_stream_packetout($$pre$phiZ2D,$7)|0); //@line 153 "modules/video/theora/VideoStream.cpp"
 $9 = ($8|0)==(1); //@line 153 "modules/video/theora/VideoStream.cpp"
 if ($9) {
  $$0 = 0;
  return ($$0|0); //@line 168 "modules/video/theora/VideoStream.cpp"
 }
 $10 = ((($this)) + 416|0); //@line 156 "modules/video/theora/VideoStream.cpp"
 $11 = ((($this)) + 20|0); //@line 162 "modules/video/theora/VideoStream.cpp"
 if ($mustSucceed) {
  while(1) {
   while(1) {
    __ZN4love5video6theora11VideoStream8readPageEv($this); //@line 161 "modules/video/theora/VideoStream.cpp"
    $12 = (_ogg_page_serialno($10)|0); //@line 162 "modules/video/theora/VideoStream.cpp"
    $13 = HEAP32[$11>>2]|0; //@line 162 "modules/video/theora/VideoStream.cpp"
    $14 = ($12|0)==($13|0); //@line 162 "modules/video/theora/VideoStream.cpp"
    if ($14) {
     break;
    }
   }
   (_ogg_stream_pagein($$pre$phiZ2D,$10)|0); //@line 164 "modules/video/theora/VideoStream.cpp"
   $15 = (_ogg_stream_packetout($$pre$phiZ2D,$7)|0); //@line 153 "modules/video/theora/VideoStream.cpp"
   $16 = ($15|0)==(1); //@line 153 "modules/video/theora/VideoStream.cpp"
   if ($16) {
    $$0 = 0;
    break;
   }
  }
  return ($$0|0); //@line 168 "modules/video/theora/VideoStream.cpp"
 }
 while(1) {
  $17 = (_ogg_page_eos($10)|0); //@line 156 "modules/video/theora/VideoStream.cpp"
  $18 = ($17|0)==(0); //@line 156 "modules/video/theora/VideoStream.cpp"
  if (!($18)) {
   break;
  }
  while(1) {
   __ZN4love5video6theora11VideoStream8readPageEv($this); //@line 161 "modules/video/theora/VideoStream.cpp"
   $20 = (_ogg_page_serialno($10)|0); //@line 162 "modules/video/theora/VideoStream.cpp"
   $21 = HEAP32[$11>>2]|0; //@line 162 "modules/video/theora/VideoStream.cpp"
   $22 = ($20|0)==($21|0); //@line 162 "modules/video/theora/VideoStream.cpp"
   if ($22) {
    break;
   }
  }
  (_ogg_stream_pagein($$pre$phiZ2D,$10)|0); //@line 164 "modules/video/theora/VideoStream.cpp"
  $23 = (_ogg_stream_packetout($$pre$phiZ2D,$7)|0); //@line 153 "modules/video/theora/VideoStream.cpp"
  $24 = ($23|0)==(1); //@line 153 "modules/video/theora/VideoStream.cpp"
  if ($24) {
   $$0 = 0;
   label = 13;
   break;
  }
 }
 if ((label|0) == 13) {
  return ($$0|0); //@line 168 "modules/video/theora/VideoStream.cpp"
 }
 $19 = ((($this)) + 584|0); //@line 157 "modules/video/theora/VideoStream.cpp"
 HEAP8[$19>>0] = 1; //@line 157 "modules/video/theora/VideoStream.cpp"
 $$0 = 1;
 return ($$0|0); //@line 168 "modules/video/theora/VideoStream.cpp"
}
function __ZN4love5video6theora11VideoStream8readPageEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $syncBuffer$01 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 24|0); //@line 131 "modules/video/theora/VideoStream.cpp"
 $1 = ((($this)) + 416|0); //@line 131 "modules/video/theora/VideoStream.cpp"
 $2 = (_ogg_sync_pageout($0,$1)|0); //@line 131 "modules/video/theora/VideoStream.cpp"
 $3 = ($2|0)==(1); //@line 131 "modules/video/theora/VideoStream.cpp"
 if ($3) {
  STACKTOP = sp;return; //@line 140 "modules/video/theora/VideoStream.cpp"
 }
 $4 = ((($this)) + 12|0); //@line 117 "./common/Object.h"
 $5 = ((($this)) + 16|0); //@line 133 "modules/video/theora/VideoStream.cpp"
 $6 = ((($this)) + 56|0); //@line 133 "modules/video/theora/VideoStream.cpp"
 $syncBuffer$01 = 0;
 while(1) {
  $7 = ($syncBuffer$01|0)==(0|0); //@line 133 "modules/video/theora/VideoStream.cpp"
  if (!($7)) {
   $8 = HEAP8[$5>>0]|0; //@line 133 "modules/video/theora/VideoStream.cpp"
   $9 = ($8<<24>>24)==(0); //@line 133 "modules/video/theora/VideoStream.cpp"
   if ($9) {
    $10 = (_ogg_stream_check($6)|0); //@line 133 "modules/video/theora/VideoStream.cpp"
    $11 = ($10|0)==(0); //@line 133 "modules/video/theora/VideoStream.cpp"
    if (!($11)) {
     break;
    }
   }
  }
  $17 = (_ogg_sync_buffer($0,8192)|0); //@line 136 "modules/video/theora/VideoStream.cpp"
  $18 = HEAP32[$4>>2]|0; //@line 117 "./common/Object.h"
  $19 = HEAP32[$18>>2]|0; //@line 137 "modules/video/theora/VideoStream.cpp"
  $20 = ((($19)) + 36|0); //@line 137 "modules/video/theora/VideoStream.cpp"
  $21 = HEAP32[$20>>2]|0; //@line 137 "modules/video/theora/VideoStream.cpp"
  $22 = (FUNCTION_TABLE_iiiii[$21 & 8191]($18,$17,8192,0)|0); //@line 137 "modules/video/theora/VideoStream.cpp"
  $23 = tempRet0; //@line 137 "modules/video/theora/VideoStream.cpp"
  (_ogg_sync_wrote($0,$22)|0); //@line 138 "modules/video/theora/VideoStream.cpp"
  $24 = (_ogg_sync_pageout($0,$1)|0); //@line 131 "modules/video/theora/VideoStream.cpp"
  $25 = ($24|0)==(1); //@line 131 "modules/video/theora/VideoStream.cpp"
  if ($25) {
   label = 10;
   break;
  } else {
   $syncBuffer$01 = $17;
  }
 }
 if ((label|0) == 10) {
  STACKTOP = sp;return; //@line 140 "modules/video/theora/VideoStream.cpp"
 }
 $12 = (___cxa_allocate_exception(16)|0); //@line 134 "modules/video/theora/VideoStream.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($12|0),(360162|0),($vararg_buffer|0)); //@line 134 "modules/video/theora/VideoStream.cpp"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if ($14) {
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  ___cxa_free_exception(($12|0)); //@line 134 "modules/video/theora/VideoStream.cpp"
  ___resumeException($15|0);
  // unreachable;
 } else {
  ___cxa_throw(($12|0),(96|0),(344|0)); //@line 134 "modules/video/theora/VideoStream.cpp"
  // unreachable; //@line 134 "modules/video/theora/VideoStream.cpp"
 }
}
function __ZN4love5video6theora11VideoStream11seekDecoderEd($this,$target) {
 $this = $this|0;
 $target = +$target;
 var $$high$0 = 0.0, $$in = 0.0, $$in$in = 0.0, $$lcssa = 0.0, $$lcssa11 = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0.0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0.0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $granPos$i = 0, $high$0$ph8 = 0.0, $low$0$ = 0.0, $low$0$ph9 = 0.0, $or$cond = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $granPos$i = sp;
 $0 = $target < 0.01; //@line 274 "modules/video/theora/VideoStream.cpp"
 $1 = ((($this)) + 12|0); //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 117 "./common/Object.h"
 if ($0) {
  $3 = HEAP32[$2>>2]|0; //@line 254 "modules/video/theora/VideoStream.cpp"
  $4 = ((($3)) + 60|0); //@line 254 "modules/video/theora/VideoStream.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 254 "modules/video/theora/VideoStream.cpp"
  (FUNCTION_TABLE_iiii[$5 & 8191]($2,0,0)|0); //@line 254 "modules/video/theora/VideoStream.cpp"
  $6 = ((($this)) + 24|0); //@line 257 "modules/video/theora/VideoStream.cpp"
  (_ogg_sync_reset($6)|0); //@line 257 "modules/video/theora/VideoStream.cpp"
  $7 = ((($this)) + 416|0); //@line 258 "modules/video/theora/VideoStream.cpp"
  (_ogg_sync_pageseek($6,$7)|0); //@line 258 "modules/video/theora/VideoStream.cpp"
  (__ZN4love5video6theora11VideoStream10readPacketEb($this,1)|0); //@line 261 "modules/video/theora/VideoStream.cpp"
  $8 = $granPos$i; //@line 264 "modules/video/theora/VideoStream.cpp"
  $9 = $8; //@line 264 "modules/video/theora/VideoStream.cpp"
  HEAP32[$9>>2] = 1; //@line 264 "modules/video/theora/VideoStream.cpp"
  $10 = (($8) + 4)|0; //@line 264 "modules/video/theora/VideoStream.cpp"
  $11 = $10; //@line 264 "modules/video/theora/VideoStream.cpp"
  HEAP32[$11>>2] = 0; //@line 264 "modules/video/theora/VideoStream.cpp"
  $12 = ((($this)) + 528|0); //@line 265 "modules/video/theora/VideoStream.cpp"
  $13 = HEAP32[$12>>2]|0; //@line 265 "modules/video/theora/VideoStream.cpp"
  (_th_decode_ctl($13,5,$granPos$i,8)|0); //@line 265 "modules/video/theora/VideoStream.cpp"
  $14 = ((($this)) + 576|0); //@line 268 "modules/video/theora/VideoStream.cpp"
  HEAPF64[$14>>3] = -1.0; //@line 268 "modules/video/theora/VideoStream.cpp"
  $15 = ((($this)) + 568|0); //@line 268 "modules/video/theora/VideoStream.cpp"
  HEAPF64[$15>>3] = -1.0; //@line 268 "modules/video/theora/VideoStream.cpp"
  $16 = ((($this)) + 584|0); //@line 269 "modules/video/theora/VideoStream.cpp"
  HEAP8[$16>>0] = 0; //@line 269 "modules/video/theora/VideoStream.cpp"
  STACKTOP = sp;return; //@line 316 "modules/video/theora/VideoStream.cpp"
 }
 $17 = HEAP32[$2>>2]|0; //@line 281 "modules/video/theora/VideoStream.cpp"
 $18 = ((($17)) + 28|0); //@line 281 "modules/video/theora/VideoStream.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 281 "modules/video/theora/VideoStream.cpp"
 $20 = (FUNCTION_TABLE_ii[$19 & 8191]($2)|0); //@line 281 "modules/video/theora/VideoStream.cpp"
 $21 = tempRet0; //@line 281 "modules/video/theora/VideoStream.cpp"
 $22 = (+($20>>>0)) + (4294967296.0*(+($21|0))); //@line 281 "modules/video/theora/VideoStream.cpp"
 $23 = ((($this)) + 24|0); //@line 290 "modules/video/theora/VideoStream.cpp"
 $24 = ((($this)) + 416|0); //@line 291 "modules/video/theora/VideoStream.cpp"
 $25 = ((($this)) + 584|0); //@line 295 "modules/video/theora/VideoStream.cpp"
 $26 = ((($this)) + 528|0); //@line 299 "modules/video/theora/VideoStream.cpp"
 $27 = ((($this)) + 448|0); //@line 299 "modules/video/theora/VideoStream.cpp"
 $28 = ($21|0)>(0); //@line 283 "modules/video/theora/VideoStream.cpp"
 $29 = ($20>>>0)>(0); //@line 283 "modules/video/theora/VideoStream.cpp"
 $30 = ($21|0)==(0); //@line 283 "modules/video/theora/VideoStream.cpp"
 $31 = $30 & $29; //@line 283 "modules/video/theora/VideoStream.cpp"
 $32 = $28 | $31; //@line 283 "modules/video/theora/VideoStream.cpp"
 L5: do {
  if ($32) {
   $$in$in = $22;$high$0$ph8 = $22;$low$0$ph9 = 0.0;
   L6: while(1) {
    $$in = $$in$in * 0.5; //@line 286 "modules/video/theora/VideoStream.cpp"
    $33 = (~~$$in)>>>0; //@line 287 "modules/video/theora/VideoStream.cpp"
    $34 = +Math_abs($$in) >= 1.0 ? $$in > 0.0 ? (~~+Math_min(+Math_floor($$in / 4294967296.0), 4294967295.0)) >>> 0 : ~~+Math_ceil(($$in - +(~~$$in >>> 0)) / 4294967296.0) >>> 0 : 0; //@line 287 "modules/video/theora/VideoStream.cpp"
    while(1) {
     $35 = HEAP32[$1>>2]|0; //@line 117 "./common/Object.h"
     $36 = HEAP32[$35>>2]|0; //@line 287 "modules/video/theora/VideoStream.cpp"
     $37 = ((($36)) + 60|0); //@line 287 "modules/video/theora/VideoStream.cpp"
     $38 = HEAP32[$37>>2]|0; //@line 287 "modules/video/theora/VideoStream.cpp"
     (FUNCTION_TABLE_iiii[$38 & 8191]($35,$33,$34)|0); //@line 287 "modules/video/theora/VideoStream.cpp"
     (_ogg_sync_reset($23)|0); //@line 290 "modules/video/theora/VideoStream.cpp"
     (_ogg_sync_pageseek($23,$24)|0); //@line 291 "modules/video/theora/VideoStream.cpp"
     (__ZN4love5video6theora11VideoStream10readPacketEb($this,0)|0); //@line 294 "modules/video/theora/VideoStream.cpp"
     $39 = HEAP8[$25>>0]|0; //@line 295 "modules/video/theora/VideoStream.cpp"
     $40 = ($39<<24>>24)==(0); //@line 295 "modules/video/theora/VideoStream.cpp"
     if (!($40)) {
      break L6;
     }
     $41 = HEAP32[$26>>2]|0; //@line 299 "modules/video/theora/VideoStream.cpp"
     $42 = $27; //@line 299 "modules/video/theora/VideoStream.cpp"
     $43 = $42; //@line 299 "modules/video/theora/VideoStream.cpp"
     $44 = HEAP32[$43>>2]|0; //@line 299 "modules/video/theora/VideoStream.cpp"
     $45 = (($42) + 4)|0; //@line 299 "modules/video/theora/VideoStream.cpp"
     $46 = $45; //@line 299 "modules/video/theora/VideoStream.cpp"
     $47 = HEAP32[$46>>2]|0; //@line 299 "modules/video/theora/VideoStream.cpp"
     $48 = (+_th_granule_time($41,$44,$47)); //@line 299 "modules/video/theora/VideoStream.cpp"
     $49 = HEAP32[$26>>2]|0; //@line 300 "modules/video/theora/VideoStream.cpp"
     $50 = $27; //@line 300 "modules/video/theora/VideoStream.cpp"
     $51 = $50; //@line 300 "modules/video/theora/VideoStream.cpp"
     $52 = HEAP32[$51>>2]|0; //@line 300 "modules/video/theora/VideoStream.cpp"
     $53 = (($50) + 4)|0; //@line 300 "modules/video/theora/VideoStream.cpp"
     $54 = $53; //@line 300 "modules/video/theora/VideoStream.cpp"
     $55 = HEAP32[$54>>2]|0; //@line 300 "modules/video/theora/VideoStream.cpp"
     $56 = (_i64Add(($52|0),($55|0),1,0)|0); //@line 300 "modules/video/theora/VideoStream.cpp"
     $57 = tempRet0; //@line 300 "modules/video/theora/VideoStream.cpp"
     $58 = (+_th_granule_time($49,$56,$57)); //@line 300 "modules/video/theora/VideoStream.cpp"
     $59 = $48 == -1.0; //@line 302 "modules/video/theora/VideoStream.cpp"
     if (!($59)) {
      $$lcssa = $48;$$lcssa11 = $58;
      break;
     }
    }
    $60 = $$lcssa <= $target; //@line 304 "modules/video/theora/VideoStream.cpp"
    $61 = $$lcssa11 > $target; //@line 304 "modules/video/theora/VideoStream.cpp"
    $or$cond = $60 & $61; //@line 304 "modules/video/theora/VideoStream.cpp"
    if ($or$cond) {
     break L5;
    }
    $62 = $$lcssa > $target; //@line 306 "modules/video/theora/VideoStream.cpp"
    $$high$0 = $62 ? $$in : $high$0$ph8; //@line 307 "modules/video/theora/VideoStream.cpp"
    $low$0$ = $62 ? $low$0$ph9 : $$in; //@line 307 "modules/video/theora/VideoStream.cpp"
    $63 = $$high$0 - $low$0$; //@line 283 "modules/video/theora/VideoStream.cpp"
    $64 = $63 > 1.0E-4; //@line 283 "modules/video/theora/VideoStream.cpp"
    $65 = $$high$0 + $low$0$; //@line 286 "modules/video/theora/VideoStream.cpp"
    if ($64) {
     $$in$in = $65;$high$0$ph8 = $$high$0;$low$0$ph9 = $low$0$;
    } else {
     break L5;
    }
   }
   STACKTOP = sp;return; //@line 316 "modules/video/theora/VideoStream.cpp"
  }
 } while(0);
 $66 = ((($this)) + 576|0); //@line 313 "modules/video/theora/VideoStream.cpp"
 HEAPF64[$66>>3] = -1.0; //@line 313 "modules/video/theora/VideoStream.cpp"
 $67 = ((($this)) + 568|0); //@line 313 "modules/video/theora/VideoStream.cpp"
 HEAPF64[$67>>3] = -1.0; //@line 313 "modules/video/theora/VideoStream.cpp"
 HEAP8[$25>>0] = 0; //@line 314 "modules/video/theora/VideoStream.cpp"
 $68 = HEAP32[$26>>2]|0; //@line 315 "modules/video/theora/VideoStream.cpp"
 (_th_decode_ctl($68,5,$27,8)|0); //@line 315 "modules/video/theora/VideoStream.cpp"
 STACKTOP = sp;return; //@line 316 "modules/video/theora/VideoStream.cpp"
}
function __ZN4love5video6theora11VideoStream22threadedFillBackBufferEd($this,$dt) {
 $this = $this|0;
 $dt = +$dt;
 var $$lcssa = 0, $$lcssa14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0.0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0.0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0.0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0.0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0.0, $bufferinfo = 0, $granulePosition = 0, $l = 0, $l3 = 0, $y$05 = 0, $y1$03 = 0, $y2$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $bufferinfo = sp + 16|0;
 $granulePosition = sp;
 $l = sp + 12|0;
 $l3 = sp + 8|0;
 $0 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 321 "modules/video/theora/VideoStream.cpp"
 $3 = ((($2)) + 20|0); //@line 321 "modules/video/theora/VideoStream.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 321 "modules/video/theora/VideoStream.cpp"
 FUNCTION_TABLE_vid[$4 & 8191]($1,$dt); //@line 321 "modules/video/theora/VideoStream.cpp"
 $5 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $6 = HEAP32[$5>>2]|0; //@line 322 "modules/video/theora/VideoStream.cpp"
 $7 = ((($6)) + 16|0); //@line 322 "modules/video/theora/VideoStream.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 322 "modules/video/theora/VideoStream.cpp"
 $9 = (+FUNCTION_TABLE_di[$8 & 4095]($5)); //@line 322 "modules/video/theora/VideoStream.cpp"
 $10 = ((($this)) + 568|0); //@line 325 "modules/video/theora/VideoStream.cpp"
 $11 = +HEAPF64[$10>>3]; //@line 325 "modules/video/theora/VideoStream.cpp"
 $12 = $9 < $11; //@line 325 "modules/video/theora/VideoStream.cpp"
 if ($12) {
  __ZN4love5video6theora11VideoStream11seekDecoderEd($this,$9); //@line 326 "modules/video/theora/VideoStream.cpp"
 }
 $13 = ((($this)) + 584|0); //@line 330 "modules/video/theora/VideoStream.cpp"
 $14 = HEAP8[$13>>0]|0; //@line 330 "modules/video/theora/VideoStream.cpp"
 $15 = ($14<<24>>24)==(0); //@line 330 "modules/video/theora/VideoStream.cpp"
 if (!($15)) {
  STACKTOP = sp;return; //@line 388 "modules/video/theora/VideoStream.cpp"
 }
 $16 = ((($this)) + 576|0); //@line 330 "modules/video/theora/VideoStream.cpp"
 $17 = +HEAPF64[$16>>3]; //@line 330 "modules/video/theora/VideoStream.cpp"
 $18 = $9 < $17; //@line 330 "modules/video/theora/VideoStream.cpp"
 if ($18) {
  STACKTOP = sp;return; //@line 388 "modules/video/theora/VideoStream.cpp"
 }
 $19 = ((($this)) + 528|0); //@line 334 "modules/video/theora/VideoStream.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 334 "modules/video/theora/VideoStream.cpp"
 (_th_decode_ycbcr_out($20,$bufferinfo)|0); //@line 334 "modules/video/theora/VideoStream.cpp"
 $21 = ((($this)) + 432|0); //@line 341 "modules/video/theora/VideoStream.cpp"
 while(1) {
  $22 = (__ZN4love5video6theora11VideoStream10readPacketEb($this,0)|0); //@line 339 "modules/video/theora/VideoStream.cpp"
  if ($22) {
   break;
  }
  $23 = HEAP32[$19>>2]|0; //@line 341 "modules/video/theora/VideoStream.cpp"
  $24 = (_th_decode_packetin($23,$21,$granulePosition)|0); //@line 341 "modules/video/theora/VideoStream.cpp"
  $25 = ($24|0)==(0); //@line 341 "modules/video/theora/VideoStream.cpp"
  if ($25) {
   label = 8;
   break;
  }
 }
 if ((label|0) == 8) {
  $26 = +HEAPF64[$16>>3]; //@line 342 "modules/video/theora/VideoStream.cpp"
  HEAPF64[$10>>3] = $26; //@line 342 "modules/video/theora/VideoStream.cpp"
  $27 = HEAP32[$19>>2]|0; //@line 343 "modules/video/theora/VideoStream.cpp"
  $28 = $granulePosition; //@line 343 "modules/video/theora/VideoStream.cpp"
  $29 = $28; //@line 343 "modules/video/theora/VideoStream.cpp"
  $30 = HEAP32[$29>>2]|0; //@line 343 "modules/video/theora/VideoStream.cpp"
  $31 = (($28) + 4)|0; //@line 343 "modules/video/theora/VideoStream.cpp"
  $32 = $31; //@line 343 "modules/video/theora/VideoStream.cpp"
  $33 = HEAP32[$32>>2]|0; //@line 343 "modules/video/theora/VideoStream.cpp"
  $34 = (+_th_granule_time($27,$30,$33)); //@line 343 "modules/video/theora/VideoStream.cpp"
  HEAPF64[$16>>3] = $34; //@line 343 "modules/video/theora/VideoStream.cpp"
  $35 = ((($this)) + 556|0); //@line 347 "modules/video/theora/VideoStream.cpp"
  $36 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($35)|0); //@line 347 "modules/video/theora/VideoStream.cpp"
  __ZN4love6thread4LockC2EPNS0_5MutexE($l,$36); //@line 347 "modules/video/theora/VideoStream.cpp"
  $37 = ((($this)) + 560|0); //@line 348 "modules/video/theora/VideoStream.cpp"
  HEAP8[$37>>0] = 0; //@line 348 "modules/video/theora/VideoStream.cpp"
  __ZN4love6thread4LockD2Ev($l); //@line 349 "modules/video/theora/VideoStream.cpp"
  $38 = ((($this)) + 536|0); //@line 351 "modules/video/theora/VideoStream.cpp"
  $39 = HEAP32[$38>>2]|0; //@line 351 "modules/video/theora/VideoStream.cpp"
  $40 = ((($39)) + 4|0); //@line 351 "modules/video/theora/VideoStream.cpp"
  $41 = HEAP32[$40>>2]|0; //@line 351 "modules/video/theora/VideoStream.cpp"
  $42 = ($41|0)>(0); //@line 351 "modules/video/theora/VideoStream.cpp"
  if ($42) {
   $43 = ((($bufferinfo)) + 12|0); //@line 354 "modules/video/theora/VideoStream.cpp"
   $44 = ((($bufferinfo)) + 8|0); //@line 355 "modules/video/theora/VideoStream.cpp"
   $45 = ((($this)) + 548|0); //@line 355 "modules/video/theora/VideoStream.cpp"
   $46 = ((($this)) + 540|0); //@line 355 "modules/video/theora/VideoStream.cpp"
   $56 = $39;$y$05 = 0;
   while(1) {
    $55 = ((($56)) + 8|0); //@line 353 "modules/video/theora/VideoStream.cpp"
    $57 = HEAP32[$55>>2]|0; //@line 353 "modules/video/theora/VideoStream.cpp"
    $58 = HEAP32[$56>>2]|0; //@line 353 "modules/video/theora/VideoStream.cpp"
    $59 = Math_imul($58, $y$05)|0; //@line 353 "modules/video/theora/VideoStream.cpp"
    $60 = (($57) + ($59)|0); //@line 353 "modules/video/theora/VideoStream.cpp"
    $61 = HEAP32[$43>>2]|0; //@line 354 "modules/video/theora/VideoStream.cpp"
    $62 = HEAP32[$44>>2]|0; //@line 355 "modules/video/theora/VideoStream.cpp"
    $63 = HEAP32[$45>>2]|0; //@line 355 "modules/video/theora/VideoStream.cpp"
    $64 = (($63) + ($y$05))|0; //@line 355 "modules/video/theora/VideoStream.cpp"
    $65 = Math_imul($64, $62)|0; //@line 355 "modules/video/theora/VideoStream.cpp"
    $66 = (($61) + ($65)|0); //@line 354 "modules/video/theora/VideoStream.cpp"
    $67 = HEAP32[$46>>2]|0; //@line 355 "modules/video/theora/VideoStream.cpp"
    $68 = (($66) + ($67)|0); //@line 355 "modules/video/theora/VideoStream.cpp"
    _memcpy(($60|0),($68|0),($58|0))|0; //@line 353 "modules/video/theora/VideoStream.cpp"
    $69 = (($y$05) + 1)|0; //@line 351 "modules/video/theora/VideoStream.cpp"
    $70 = HEAP32[$38>>2]|0; //@line 351 "modules/video/theora/VideoStream.cpp"
    $71 = ((($70)) + 4|0); //@line 351 "modules/video/theora/VideoStream.cpp"
    $72 = HEAP32[$71>>2]|0; //@line 351 "modules/video/theora/VideoStream.cpp"
    $73 = ($69|0)<($72|0); //@line 351 "modules/video/theora/VideoStream.cpp"
    if ($73) {
     $56 = $70;$y$05 = $69;
    } else {
     $48 = $70;
     break;
    }
   }
  } else {
   $48 = $39;
  }
  $47 = ((($48)) + 16|0); //@line 359 "modules/video/theora/VideoStream.cpp"
  $49 = HEAP32[$47>>2]|0; //@line 359 "modules/video/theora/VideoStream.cpp"
  $50 = ($49|0)>(0); //@line 359 "modules/video/theora/VideoStream.cpp"
  if ($50) {
   $51 = ((($bufferinfo)) + 28|0); //@line 362 "modules/video/theora/VideoStream.cpp"
   $52 = ((($bufferinfo)) + 24|0); //@line 363 "modules/video/theora/VideoStream.cpp"
   $53 = ((($this)) + 552|0); //@line 363 "modules/video/theora/VideoStream.cpp"
   $54 = ((($this)) + 544|0); //@line 363 "modules/video/theora/VideoStream.cpp"
   $80 = $48;$y1$03 = 0;
   while(1) {
    $79 = ((($80)) + 20|0); //@line 361 "modules/video/theora/VideoStream.cpp"
    $81 = HEAP32[$79>>2]|0; //@line 361 "modules/video/theora/VideoStream.cpp"
    $82 = ((($80)) + 12|0); //@line 361 "modules/video/theora/VideoStream.cpp"
    $83 = HEAP32[$82>>2]|0; //@line 361 "modules/video/theora/VideoStream.cpp"
    $84 = Math_imul($83, $y1$03)|0; //@line 361 "modules/video/theora/VideoStream.cpp"
    $85 = (($81) + ($84)|0); //@line 361 "modules/video/theora/VideoStream.cpp"
    $86 = HEAP32[$51>>2]|0; //@line 362 "modules/video/theora/VideoStream.cpp"
    $87 = HEAP32[$52>>2]|0; //@line 363 "modules/video/theora/VideoStream.cpp"
    $88 = HEAP32[$53>>2]|0; //@line 363 "modules/video/theora/VideoStream.cpp"
    $89 = (($88) + ($y1$03))|0; //@line 363 "modules/video/theora/VideoStream.cpp"
    $90 = Math_imul($89, $87)|0; //@line 363 "modules/video/theora/VideoStream.cpp"
    $91 = (($86) + ($90)|0); //@line 362 "modules/video/theora/VideoStream.cpp"
    $92 = HEAP32[$54>>2]|0; //@line 363 "modules/video/theora/VideoStream.cpp"
    $93 = (($91) + ($92)|0); //@line 363 "modules/video/theora/VideoStream.cpp"
    _memcpy(($85|0),($93|0),($83|0))|0; //@line 361 "modules/video/theora/VideoStream.cpp"
    $94 = (($y1$03) + 1)|0; //@line 359 "modules/video/theora/VideoStream.cpp"
    $95 = HEAP32[$38>>2]|0; //@line 359 "modules/video/theora/VideoStream.cpp"
    $96 = ((($95)) + 16|0); //@line 359 "modules/video/theora/VideoStream.cpp"
    $97 = HEAP32[$96>>2]|0; //@line 359 "modules/video/theora/VideoStream.cpp"
    $98 = ($94|0)<($97|0); //@line 359 "modules/video/theora/VideoStream.cpp"
    if ($98) {
     $80 = $95;$y1$03 = $94;
    } else {
     $$lcssa = $95;$$lcssa14 = $97;
     break;
    }
   }
   $74 = ($$lcssa14|0)>(0); //@line 367 "modules/video/theora/VideoStream.cpp"
   if ($74) {
    $75 = ((($bufferinfo)) + 44|0); //@line 370 "modules/video/theora/VideoStream.cpp"
    $76 = ((($bufferinfo)) + 40|0); //@line 371 "modules/video/theora/VideoStream.cpp"
    $77 = ((($this)) + 552|0); //@line 371 "modules/video/theora/VideoStream.cpp"
    $78 = ((($this)) + 544|0); //@line 371 "modules/video/theora/VideoStream.cpp"
    $103 = $$lcssa;$y2$02 = 0;
    while(1) {
     $102 = ((($103)) + 24|0); //@line 369 "modules/video/theora/VideoStream.cpp"
     $104 = HEAP32[$102>>2]|0; //@line 369 "modules/video/theora/VideoStream.cpp"
     $105 = ((($103)) + 12|0); //@line 369 "modules/video/theora/VideoStream.cpp"
     $106 = HEAP32[$105>>2]|0; //@line 369 "modules/video/theora/VideoStream.cpp"
     $107 = Math_imul($106, $y2$02)|0; //@line 369 "modules/video/theora/VideoStream.cpp"
     $108 = (($104) + ($107)|0); //@line 369 "modules/video/theora/VideoStream.cpp"
     $109 = HEAP32[$75>>2]|0; //@line 370 "modules/video/theora/VideoStream.cpp"
     $110 = HEAP32[$76>>2]|0; //@line 371 "modules/video/theora/VideoStream.cpp"
     $111 = HEAP32[$77>>2]|0; //@line 371 "modules/video/theora/VideoStream.cpp"
     $112 = (($111) + ($y2$02))|0; //@line 371 "modules/video/theora/VideoStream.cpp"
     $113 = Math_imul($112, $110)|0; //@line 371 "modules/video/theora/VideoStream.cpp"
     $114 = (($109) + ($113)|0); //@line 370 "modules/video/theora/VideoStream.cpp"
     $115 = HEAP32[$78>>2]|0; //@line 371 "modules/video/theora/VideoStream.cpp"
     $116 = (($114) + ($115)|0); //@line 371 "modules/video/theora/VideoStream.cpp"
     _memcpy(($108|0),($116|0),($106|0))|0; //@line 369 "modules/video/theora/VideoStream.cpp"
     $117 = (($y2$02) + 1)|0; //@line 367 "modules/video/theora/VideoStream.cpp"
     $118 = HEAP32[$38>>2]|0; //@line 367 "modules/video/theora/VideoStream.cpp"
     $119 = ((($118)) + 16|0); //@line 367 "modules/video/theora/VideoStream.cpp"
     $120 = HEAP32[$119>>2]|0; //@line 367 "modules/video/theora/VideoStream.cpp"
     $121 = ($117|0)<($120|0); //@line 367 "modules/video/theora/VideoStream.cpp"
     if ($121) {
      $103 = $118;$y2$02 = $117;
     } else {
      break;
     }
    }
   }
  }
  $99 = +HEAPF64[$16>>3]; //@line 378 "modules/video/theora/VideoStream.cpp"
  $100 = $9 > $99; //@line 378 "modules/video/theora/VideoStream.cpp"
  $101 = ((($this)) + 588|0); //@line 380 "modules/video/theora/VideoStream.cpp"
  if ($100) {
   $122 = HEAP32[$101>>2]|0; //@line 380 "modules/video/theora/VideoStream.cpp"
   $123 = (($122) + 1)|0; //@line 380 "modules/video/theora/VideoStream.cpp"
   HEAP32[$101>>2] = $123; //@line 380 "modules/video/theora/VideoStream.cpp"
   $124 = ($123>>>0)>(5); //@line 380 "modules/video/theora/VideoStream.cpp"
   if ($124) {
    $125 = HEAP32[$this>>2]|0; //@line 381 "modules/video/theora/VideoStream.cpp"
    $126 = ((($125)) + 52|0); //@line 381 "modules/video/theora/VideoStream.cpp"
    $127 = HEAP32[$126>>2]|0; //@line 381 "modules/video/theora/VideoStream.cpp"
    FUNCTION_TABLE_vid[$127 & 8191]($this,$9); //@line 381 "modules/video/theora/VideoStream.cpp"
   }
  } else {
   HEAP32[$101>>2] = 0; //@line 384 "modules/video/theora/VideoStream.cpp"
  }
  $128 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($35)|0); //@line 386 "modules/video/theora/VideoStream.cpp"
  __ZN4love6thread4LockC2EPNS0_5MutexE($l3,$128); //@line 386 "modules/video/theora/VideoStream.cpp"
  HEAP8[$37>>0] = 1; //@line 387 "modules/video/theora/VideoStream.cpp"
  __ZN4love6thread4LockD2Ev($l3); //@line 388 "modules/video/theora/VideoStream.cpp"
 }
 STACKTOP = sp;return; //@line 388 "modules/video/theora/VideoStream.cpp"
}
function __ZN4love5video11VideoStream9DeltaSyncD2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19844); //@line 101 "modules/video/VideoStream.cpp"
 $0 = ((($this)) + 32|0); //@line 102 "modules/video/VideoStream.cpp"
 __ZN4love6thread8MutexRefD2Ev($0); //@line 102 "modules/video/VideoStream.cpp"
 return; //@line 102 "modules/video/VideoStream.cpp"
}
function __ZN4love5video11VideoStream9DeltaSyncD0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19844); //@line 101 "modules/video/VideoStream.cpp"
 $0 = ((($this)) + 32|0); //@line 102 "modules/video/VideoStream.cpp"
 __ZN4love6thread8MutexRefD2Ev($0); //@line 102 "modules/video/VideoStream.cpp"
 __ZdlPv($this); //@line 101 "modules/video/VideoStream.cpp"
 return; //@line 102 "modules/video/VideoStream.cpp"
}
function __ZNK4love5video11VideoStream9DeltaSync11getPositionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 106 "modules/video/VideoStream.cpp"
 $1 = +HEAPF64[$0>>3]; //@line 106 "modules/video/VideoStream.cpp"
 return (+$1); //@line 106 "modules/video/VideoStream.cpp"
}
function __ZN4love5video11VideoStream9DeltaSync6updateEd($this,$dt) {
 $this = $this|0;
 $dt = +$dt;
 var $0 = 0, $1 = 0, $10 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0, $9 = 0.0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = ((($this)) + 32|0); //@line 111 "modules/video/VideoStream.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 111 "modules/video/VideoStream.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($l,$1); //@line 111 "modules/video/VideoStream.cpp"
 $2 = ((($this)) + 8|0); //@line 112 "modules/video/VideoStream.cpp"
 $3 = HEAP8[$2>>0]|0; //@line 112 "modules/video/VideoStream.cpp"
 $4 = ($3<<24>>24)==(0); //@line 112 "modules/video/VideoStream.cpp"
 if ($4) {
  __ZN4love6thread4LockD2Ev($l); //@line 114 "modules/video/VideoStream.cpp"
  STACKTOP = sp;return; //@line 114 "modules/video/VideoStream.cpp"
 }
 $5 = ((($this)) + 24|0); //@line 113 "modules/video/VideoStream.cpp"
 $6 = +HEAPF64[$5>>3]; //@line 113 "modules/video/VideoStream.cpp"
 $7 = $6 * $dt; //@line 113 "modules/video/VideoStream.cpp"
 $8 = ((($this)) + 16|0); //@line 113 "modules/video/VideoStream.cpp"
 $9 = +HEAPF64[$8>>3]; //@line 113 "modules/video/VideoStream.cpp"
 $10 = $9 + $7; //@line 113 "modules/video/VideoStream.cpp"
 HEAPF64[$8>>3] = $10; //@line 113 "modules/video/VideoStream.cpp"
 __ZN4love6thread4LockD2Ev($l); //@line 114 "modules/video/VideoStream.cpp"
 STACKTOP = sp;return; //@line 114 "modules/video/VideoStream.cpp"
}
function __ZN4love5video11VideoStream9DeltaSync4playEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 118 "modules/video/VideoStream.cpp"
 HEAP8[$0>>0] = 1; //@line 118 "modules/video/VideoStream.cpp"
 return; //@line 119 "modules/video/VideoStream.cpp"
}
function __ZN4love5video11VideoStream9DeltaSync5pauseEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 123 "modules/video/VideoStream.cpp"
 HEAP8[$0>>0] = 0; //@line 123 "modules/video/VideoStream.cpp"
 return; //@line 124 "modules/video/VideoStream.cpp"
}
function __ZN4love5video11VideoStream9DeltaSync4seekEd($this,$time) {
 $this = $this|0;
 $time = +$time;
 var $0 = 0, $1 = 0, $2 = 0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = ((($this)) + 32|0); //@line 128 "modules/video/VideoStream.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 128 "modules/video/VideoStream.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($l,$1); //@line 128 "modules/video/VideoStream.cpp"
 $2 = ((($this)) + 16|0); //@line 129 "modules/video/VideoStream.cpp"
 HEAPF64[$2>>3] = $time; //@line 129 "modules/video/VideoStream.cpp"
 __ZN4love6thread4LockD2Ev($l); //@line 130 "modules/video/VideoStream.cpp"
 STACKTOP = sp;return; //@line 130 "modules/video/VideoStream.cpp"
}
function __ZNK4love5video11VideoStream9FrameSync4tellEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 90 "modules/video/VideoStream.cpp"
 $1 = ((($0)) + 16|0); //@line 90 "modules/video/VideoStream.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 90 "modules/video/VideoStream.cpp"
 $3 = (+FUNCTION_TABLE_di[$2 & 4095]($this)); //@line 90 "modules/video/VideoStream.cpp"
 return (+$3); //@line 90 "modules/video/VideoStream.cpp"
}
function __ZNK4love5video11VideoStream9DeltaSync9isPlayingEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 134 "modules/video/VideoStream.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 134 "modules/video/VideoStream.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 134 "modules/video/VideoStream.cpp"
 return ($2|0); //@line 134 "modules/video/VideoStream.cpp"
}
function __ZN4love5video11VideoStream10SourceSyncD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19896); //@line 109 "modules/video/VideoStream.h"
 $0 = ((($this)) + 8|0); //@line 106 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 106 "./common/Object.h"
 $2 = ($1|0)==(0|0); //@line 106 "./common/Object.h"
 if ($2) {
  return; //@line 109 "modules/video/VideoStream.h"
 }
 $3 = HEAP32[$1>>2]|0; //@line 106 "./common/Object.h"
 $4 = ((($3)) + 12|0); //@line 106 "./common/Object.h"
 $5 = HEAP32[$4>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($5|0,($1|0)); //@line 106 "./common/Object.h"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $8 = ___cxa_find_matching_catch_3(0|0)|0;
  $9 = tempRet0;
  ___clang_call_terminate($8); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  return; //@line 109 "modules/video/VideoStream.h"
 }
}
function __ZN4love5video11VideoStream10SourceSyncD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19896); //@line 109 "modules/video/VideoStream.h"
 $0 = ((($this)) + 8|0); //@line 106 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 106 "./common/Object.h"
 $2 = ($1|0)==(0|0); //@line 106 "./common/Object.h"
 if ($2) {
  __ZdlPv($this); //@line 109 "modules/video/VideoStream.h"
  return; //@line 109 "modules/video/VideoStream.h"
 }
 $3 = HEAP32[$1>>2]|0; //@line 106 "./common/Object.h"
 $4 = ((($3)) + 12|0); //@line 106 "./common/Object.h"
 $5 = HEAP32[$4>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($5|0,($1|0)); //@line 106 "./common/Object.h"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $8 = ___cxa_find_matching_catch_3(0|0)|0;
  $9 = tempRet0;
  ___clang_call_terminate($8); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  __ZdlPv($this); //@line 109 "modules/video/VideoStream.h"
  return; //@line 109 "modules/video/VideoStream.h"
 }
}
function __ZNK4love5video11VideoStream10SourceSync11getPositionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 144 "modules/video/VideoStream.cpp"
 $3 = ((($2)) + 76|0); //@line 144 "modules/video/VideoStream.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 144 "modules/video/VideoStream.cpp"
 $5 = (+FUNCTION_TABLE_dii[$4 & 8191]($1,0)); //@line 144 "modules/video/VideoStream.cpp"
 $6 = $5; //@line 144 "modules/video/VideoStream.cpp"
 return (+$6); //@line 144 "modules/video/VideoStream.cpp"
}
function __ZN4love5video11VideoStream9FrameSync6updateEd($this,$0) {
 $this = $this|0;
 $0 = +$0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 75 "modules/video/VideoStream.h"
}
function __ZN4love5video11VideoStream10SourceSync4playEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 149 "modules/video/VideoStream.cpp"
 $3 = ((($2)) + 20|0); //@line 149 "modules/video/VideoStream.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 149 "modules/video/VideoStream.cpp"
 (FUNCTION_TABLE_ii[$4 & 8191]($1)|0); //@line 149 "modules/video/VideoStream.cpp"
 return; //@line 150 "modules/video/VideoStream.cpp"
}
function __ZN4love5video11VideoStream10SourceSync5pauseEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 154 "modules/video/VideoStream.cpp"
 $3 = ((($2)) + 28|0); //@line 154 "modules/video/VideoStream.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 154 "modules/video/VideoStream.cpp"
 FUNCTION_TABLE_vi[$4 & 8191]($1); //@line 154 "modules/video/VideoStream.cpp"
 return; //@line 155 "modules/video/VideoStream.cpp"
}
function __ZN4love5video11VideoStream10SourceSync4seekEd($this,$time) {
 $this = $this|0;
 $time = +$time;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 159 "modules/video/VideoStream.cpp"
 $3 = ((($2)) + 72|0); //@line 159 "modules/video/VideoStream.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 159 "modules/video/VideoStream.cpp"
 $5 = $time; //@line 159 "modules/video/VideoStream.cpp"
 FUNCTION_TABLE_vidi[$4 & 2047]($1,$5,0); //@line 159 "modules/video/VideoStream.cpp"
 return; //@line 160 "modules/video/VideoStream.cpp"
}
function __ZNK4love5video11VideoStream10SourceSync9isPlayingEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 164 "modules/video/VideoStream.cpp"
 $3 = ((($2)) + 40|0); //@line 164 "modules/video/VideoStream.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 164 "modules/video/VideoStream.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($1)|0); //@line 164 "modules/video/VideoStream.cpp"
 if ($5) {
  $12 = 0;
  return ($12|0); //@line 164 "modules/video/VideoStream.cpp"
 }
 $6 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $7 = HEAP32[$6>>2]|0; //@line 164 "modules/video/VideoStream.cpp"
 $8 = ((($7)) + 44|0); //@line 164 "modules/video/VideoStream.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 164 "modules/video/VideoStream.cpp"
 $10 = (FUNCTION_TABLE_ii[$9 & 8191]($6)|0); //@line 164 "modules/video/VideoStream.cpp"
 $11 = $10 ^ 1; //@line 164 "modules/video/VideoStream.cpp"
 $12 = $11;
 return ($12|0); //@line 164 "modules/video/VideoStream.cpp"
}
function __ZN4love5video11VideoStreamD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19948); //@line 37 "modules/video/VideoStream.h"
 $0 = ((($this)) + 8|0); //@line 106 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 106 "./common/Object.h"
 $2 = ($1|0)==(0|0); //@line 106 "./common/Object.h"
 if ($2) {
  return; //@line 37 "modules/video/VideoStream.h"
 }
 $3 = HEAP32[$1>>2]|0; //@line 106 "./common/Object.h"
 $4 = ((($3)) + 12|0); //@line 106 "./common/Object.h"
 $5 = HEAP32[$4>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($5|0,($1|0)); //@line 106 "./common/Object.h"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $8 = ___cxa_find_matching_catch_3(0|0)|0;
  $9 = tempRet0;
  ___clang_call_terminate($8); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  return; //@line 37 "modules/video/VideoStream.h"
 }
}
function __ZN4love5video11VideoStreamD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (19948); //@line 37 "modules/video/VideoStream.h"
 $0 = ((($this)) + 8|0); //@line 106 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 106 "./common/Object.h"
 $2 = ($1|0)==(0|0); //@line 106 "./common/Object.h"
 if ($2) {
  __ZdlPv($this); //@line 37 "modules/video/VideoStream.h"
  return; //@line 37 "modules/video/VideoStream.h"
 }
 $3 = HEAP32[$1>>2]|0; //@line 106 "./common/Object.h"
 $4 = ((($3)) + 12|0); //@line 106 "./common/Object.h"
 $5 = HEAP32[$4>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($5|0,($1|0)); //@line 106 "./common/Object.h"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $8 = ___cxa_find_matching_catch_3(0|0)|0;
  $9 = tempRet0;
  ___clang_call_terminate($8); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  __ZdlPv($this); //@line 37 "modules/video/VideoStream.h"
  return; //@line 37 "modules/video/VideoStream.h"
 }
}
function __ZN4love6Stream14fillBackBufferEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 42 "./common/Stream.h"
}
function __ZN4love5video11VideoStream4playEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 42 "modules/video/VideoStream.cpp"
 $3 = ((($2)) + 24|0); //@line 42 "modules/video/VideoStream.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 42 "modules/video/VideoStream.cpp"
 FUNCTION_TABLE_vi[$4 & 8191]($1); //@line 42 "modules/video/VideoStream.cpp"
 return; //@line 43 "modules/video/VideoStream.cpp"
}
function __ZN4love5video11VideoStream5pauseEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 47 "modules/video/VideoStream.cpp"
 $3 = ((($2)) + 28|0); //@line 47 "modules/video/VideoStream.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 47 "modules/video/VideoStream.cpp"
 FUNCTION_TABLE_vi[$4 & 8191]($1); //@line 47 "modules/video/VideoStream.cpp"
 return; //@line 48 "modules/video/VideoStream.cpp"
}
function __ZN4love5video11VideoStream4seekEd($this,$offset) {
 $this = $this|0;
 $offset = +$offset;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 52 "modules/video/VideoStream.cpp"
 $3 = ((($2)) + 32|0); //@line 52 "modules/video/VideoStream.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 52 "modules/video/VideoStream.cpp"
 FUNCTION_TABLE_vid[$4 & 8191]($1,$offset); //@line 52 "modules/video/VideoStream.cpp"
 return; //@line 53 "modules/video/VideoStream.cpp"
}
function __ZNK4love5video11VideoStream4tellEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 57 "modules/video/VideoStream.cpp"
 $3 = ((($2)) + 36|0); //@line 57 "modules/video/VideoStream.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 57 "modules/video/VideoStream.cpp"
 $5 = (+FUNCTION_TABLE_di[$4 & 4095]($1)); //@line 57 "modules/video/VideoStream.cpp"
 return (+$5); //@line 57 "modules/video/VideoStream.cpp"
}
function __ZNK4love5video11VideoStream9isPlayingEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 62 "modules/video/VideoStream.cpp"
 $3 = ((($2)) + 40|0); //@line 62 "modules/video/VideoStream.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 62 "modules/video/VideoStream.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($1)|0); //@line 62 "modules/video/VideoStream.cpp"
 return ($5|0); //@line 62 "modules/video/VideoStream.cpp"
}
function __ZN4love5video11VideoStream7setSyncEPNS1_9FrameSyncE($this,$frameSync) {
 $this = $this|0;
 $frameSync = $frameSync|0;
 var $$index = 0, $$index10 = 0, $$index2 = 0, $$index6 = 0, $$pre$phi4Z2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($frameSync|0)==(0|0); //@line 95 "./common/Object.h"
 if ($0) {
  label = 4;
 } else {
  $1 = HEAP32[$frameSync>>2]|0; //@line 95 "./common/Object.h"
  $2 = ((($1)) + 8|0); //@line 95 "./common/Object.h"
  $3 = HEAP32[$2>>2]|0; //@line 95 "./common/Object.h"
  FUNCTION_TABLE_vi[$3 & 8191]($frameSync); //@line 95 "./common/Object.h"
  $4 = HEAP32[$frameSync>>2]|0; //@line 132 "./common/Object.h"
  $5 = ((($4)) + 8|0); //@line 132 "./common/Object.h"
  $6 = HEAP32[$5>>2]|0; //@line 132 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($6|0,($frameSync|0)); //@line 132 "./common/Object.h"
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if ($8) {
   $9 = ___cxa_find_matching_catch_2()|0;
   $10 = tempRet0;
   $$index = $9;$$index2 = $10;$$pre$phi4Z2D = $frameSync;$$pre$phiZ2D = $frameSync;
  } else {
   label = 4;
  }
 }
 do {
  if ((label|0) == 4) {
   $11 = ((($this)) + 8|0); //@line 133 "./common/Object.h"
   $12 = HEAP32[$11>>2]|0; //@line 133 "./common/Object.h"
   $13 = ($12|0)==(0|0); //@line 133 "./common/Object.h"
   if (!($13)) {
    $14 = HEAP32[$12>>2]|0; //@line 133 "./common/Object.h"
    $15 = ((($14)) + 12|0); //@line 133 "./common/Object.h"
    $16 = HEAP32[$15>>2]|0; //@line 133 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($16|0,($12|0)); //@line 133 "./common/Object.h"
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if ($18) {
     $26 = ___cxa_find_matching_catch_2()|0;
     $27 = tempRet0;
     if ($0) {
      $$index10 = $27;$$index6 = $26;
      ___resumeException($$index6|0);
      // unreachable;
     } else {
      $$index = $26;$$index2 = $27;$$pre$phi4Z2D = $frameSync;$$pre$phiZ2D = $frameSync;
      break;
     }
    }
   }
   HEAP32[$11>>2] = $frameSync; //@line 134 "./common/Object.h"
   if ($0) {
    return; //@line 33 "modules/video/VideoStream.cpp"
   }
   $19 = HEAP32[$frameSync>>2]|0; //@line 106 "./common/Object.h"
   $20 = ((($19)) + 12|0); //@line 106 "./common/Object.h"
   $21 = HEAP32[$20>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($21|0,($frameSync|0)); //@line 106 "./common/Object.h"
   $22 = __THREW__; __THREW__ = 0;
   $23 = $22&1;
   if ($23) {
    $24 = ___cxa_find_matching_catch_3(0|0)|0;
    $25 = tempRet0;
    ___clang_call_terminate($24); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   } else {
    return; //@line 33 "modules/video/VideoStream.cpp"
   }
  }
 } while(0);
 $28 = HEAP32[$$pre$phi4Z2D>>2]|0; //@line 106 "./common/Object.h"
 $29 = ((($28)) + 12|0); //@line 106 "./common/Object.h"
 $30 = HEAP32[$29>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($30|0,($$pre$phiZ2D|0)); //@line 106 "./common/Object.h"
 $31 = __THREW__; __THREW__ = 0;
 $32 = $31&1;
 if ($32) {
  $33 = ___cxa_find_matching_catch_3(0|0)|0;
  $34 = tempRet0;
  ___clang_call_terminate($33); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  $$index10 = $$index2;$$index6 = $$index;
  ___resumeException($$index6|0);
  // unreachable;
 }
}
function __ZNK4love5video11VideoStream7getSyncEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 127 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 127 "./common/Object.h"
 return ($1|0); //@line 37 "modules/video/VideoStream.cpp"
}
function __ZN4love5video11VideoStream5FrameC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 66 "modules/video/VideoStream.cpp"
 HEAP32[$0>>2] = 0; //@line 66 "modules/video/VideoStream.cpp"
 $1 = ((($this)) + 20|0); //@line 67 "modules/video/VideoStream.cpp"
 HEAP32[$1>>2] = 0; //@line 67 "modules/video/VideoStream.cpp"
 $2 = ((($this)) + 24|0); //@line 68 "modules/video/VideoStream.cpp"
 HEAP32[$2>>2] = 0; //@line 68 "modules/video/VideoStream.cpp"
 return; //@line 70 "modules/video/VideoStream.cpp"
}
function __ZN4love5video11VideoStream5FrameD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 74 "modules/video/VideoStream.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 74 "modules/video/VideoStream.cpp"
 $2 = ($1|0)==(0|0); //@line 74 "modules/video/VideoStream.cpp"
 if (!($2)) {
  __ZdaPv($1); //@line 74 "modules/video/VideoStream.cpp"
 }
 $3 = ((($this)) + 20|0); //@line 75 "modules/video/VideoStream.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 75 "modules/video/VideoStream.cpp"
 $5 = ($4|0)==(0|0); //@line 75 "modules/video/VideoStream.cpp"
 if (!($5)) {
  __ZdaPv($4); //@line 75 "modules/video/VideoStream.cpp"
 }
 $6 = ((($this)) + 24|0); //@line 76 "modules/video/VideoStream.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 76 "modules/video/VideoStream.cpp"
 $8 = ($7|0)==(0|0); //@line 76 "modules/video/VideoStream.cpp"
 if ($8) {
  return; //@line 77 "modules/video/VideoStream.cpp"
 }
 __ZdaPv($7); //@line 76 "modules/video/VideoStream.cpp"
 return; //@line 77 "modules/video/VideoStream.cpp"
}
function __ZN4love5video11VideoStream9FrameSync9copyStateEPKS2_($this,$other) {
 $this = $this|0;
 $other = $other|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 81 "modules/video/VideoStream.cpp"
 $1 = ((($0)) + 32|0); //@line 81 "modules/video/VideoStream.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 81 "modules/video/VideoStream.cpp"
 $3 = HEAP32[$other>>2]|0; //@line 81 "modules/video/VideoStream.cpp"
 $4 = ((($3)) + 36|0); //@line 81 "modules/video/VideoStream.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 81 "modules/video/VideoStream.cpp"
 $6 = (+FUNCTION_TABLE_di[$5 & 4095]($other)); //@line 81 "modules/video/VideoStream.cpp"
 FUNCTION_TABLE_vid[$2 & 8191]($this,$6); //@line 81 "modules/video/VideoStream.cpp"
 $7 = HEAP32[$other>>2]|0; //@line 82 "modules/video/VideoStream.cpp"
 $8 = ((($7)) + 40|0); //@line 82 "modules/video/VideoStream.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 82 "modules/video/VideoStream.cpp"
 $10 = (FUNCTION_TABLE_ii[$9 & 8191]($other)|0); //@line 82 "modules/video/VideoStream.cpp"
 $11 = HEAP32[$this>>2]|0; //@line 83 "modules/video/VideoStream.cpp"
 if ($10) {
  $12 = ((($11)) + 24|0); //@line 83 "modules/video/VideoStream.cpp"
  $13 = HEAP32[$12>>2]|0; //@line 83 "modules/video/VideoStream.cpp"
  FUNCTION_TABLE_vi[$13 & 8191]($this); //@line 83 "modules/video/VideoStream.cpp"
  return; //@line 86 "modules/video/VideoStream.cpp"
 } else {
  $14 = ((($11)) + 28|0); //@line 85 "modules/video/VideoStream.cpp"
  $15 = HEAP32[$14>>2]|0; //@line 85 "modules/video/VideoStream.cpp"
  FUNCTION_TABLE_vi[$15 & 8191]($this); //@line 85 "modules/video/VideoStream.cpp"
  return; //@line 86 "modules/video/VideoStream.cpp"
 }
}
function __ZN4love5video11VideoStream9DeltaSyncC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 71 "modules/video/VideoStream.h"
 HEAP32[$this>>2] = (19844); //@line 97 "modules/video/VideoStream.cpp"
 $0 = ((($this)) + 8|0); //@line 94 "modules/video/VideoStream.cpp"
 HEAP8[$0>>0] = 0; //@line 94 "modules/video/VideoStream.cpp"
 $1 = ((($this)) + 16|0); //@line 95 "modules/video/VideoStream.cpp"
 HEAPF64[$1>>3] = 0.0; //@line 95 "modules/video/VideoStream.cpp"
 $2 = ((($this)) + 24|0); //@line 96 "modules/video/VideoStream.cpp"
 HEAPF64[$2>>3] = 1.0; //@line 96 "modules/video/VideoStream.cpp"
 $3 = ((($this)) + 32|0); //@line 93 "modules/video/VideoStream.cpp"
 __THREW__ = 0;
 invoke_vi(3788,($3|0)); //@line 93 "modules/video/VideoStream.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  ___resumeException($6|0);
  // unreachable;
 } else {
  return; //@line 98 "modules/video/VideoStream.cpp"
 }
}
function __ZN4love5video11VideoStream10SourceSyncC2EPNS_5audio6SourceE($this,$source) {
 $this = $this|0;
 $source = $source|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 71 "modules/video/VideoStream.h"
 HEAP32[$this>>2] = (19896); //@line 139 "modules/video/VideoStream.cpp"
 $0 = ((($this)) + 8|0); //@line 93 "./common/Object.h"
 HEAP32[$0>>2] = $source; //@line 93 "./common/Object.h"
 $1 = ($source|0)==(0|0); //@line 95 "./common/Object.h"
 if ($1) {
  return; //@line 140 "modules/video/VideoStream.cpp"
 }
 $2 = HEAP32[$source>>2]|0; //@line 95 "./common/Object.h"
 $3 = ((($2)) + 8|0); //@line 95 "./common/Object.h"
 $4 = HEAP32[$3>>2]|0; //@line 95 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($4|0,($source|0)); //@line 95 "./common/Object.h"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  ___resumeException($7|0);
  // unreachable;
 } else {
  return; //@line 140 "modules/video/VideoStream.cpp"
 }
}
function __ZN4love6window3sdl6WindowD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (20028); //@line 78 "modules/window/sdl/Window.cpp"
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = ($0|0)==(0|0); //@line 593 "modules/window/sdl/Window.cpp"
 if ($1) {
  label = 3;
 } else {
  $2 = HEAP32[$0>>2]|0; //@line 594 "modules/window/sdl/Window.cpp"
  $3 = ((($2)) + 32|0); //@line 594 "modules/window/sdl/Window.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 594 "modules/window/sdl/Window.cpp"
  __THREW__ = 0;
  invoke_vi($4|0,($0|0)); //@line 594 "modules/window/sdl/Window.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   label = 3;
  }
 }
 do {
  if ((label|0) == 3) {
   $7 = ((($this)) + 116|0); //@line 596 "modules/window/sdl/Window.cpp"
   $8 = HEAP32[$7>>2]|0; //@line 596 "modules/window/sdl/Window.cpp"
   $9 = ($8|0)==(0|0); //@line 596 "modules/window/sdl/Window.cpp"
   if (!($9)) {
    __THREW__ = 0;
    invoke_vi(3797,($8|0)); //@line 598 "modules/window/sdl/Window.cpp"
    $10 = __THREW__; __THREW__ = 0;
    $11 = $10&1;
    if ($11) {
     break;
    }
    HEAP32[$7>>2] = 0; //@line 599 "modules/window/sdl/Window.cpp"
   }
   $12 = ((($this)) + 112|0); //@line 602 "modules/window/sdl/Window.cpp"
   $13 = HEAP32[$12>>2]|0; //@line 602 "modules/window/sdl/Window.cpp"
   $14 = ($13|0)==(0|0); //@line 602 "modules/window/sdl/Window.cpp"
   if (!($14)) {
    __THREW__ = 0;
    invoke_vi(3798,($13|0)); //@line 604 "modules/window/sdl/Window.cpp"
    $15 = __THREW__; __THREW__ = 0;
    $16 = $15&1;
    if ($16) {
     break;
    }
    HEAP32[$12>>2] = 0; //@line 605 "modules/window/sdl/Window.cpp"
    __THREW__ = 0;
    invoke_vi(3799,512); //@line 609 "modules/window/sdl/Window.cpp"
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if ($18) {
     break;
    }
   }
   $19 = ((($this)) + 108|0); //@line 612 "modules/window/sdl/Window.cpp"
   HEAP8[$19>>0] = 0; //@line 612 "modules/window/sdl/Window.cpp"
   __THREW__ = 0;
   invoke_vi(3800,32); //@line 81 "modules/window/sdl/Window.cpp"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    $22 = ((($this)) + 104|0); //@line 106 "./common/Object.h"
    $23 = HEAP32[$22>>2]|0; //@line 106 "./common/Object.h"
    $24 = ($23|0)==(0|0); //@line 106 "./common/Object.h"
    if ($24) {
     $32 = ((($this)) + 8|0); //@line 82 "modules/window/sdl/Window.cpp"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($32); //@line 82 "modules/window/sdl/Window.cpp"
     __ZN4love6window6WindowD2Ev($this); //@line 82 "modules/window/sdl/Window.cpp"
     return; //@line 82 "modules/window/sdl/Window.cpp"
    }
    $25 = HEAP32[$23>>2]|0; //@line 106 "./common/Object.h"
    $26 = ((($25)) + 12|0); //@line 106 "./common/Object.h"
    $27 = HEAP32[$26>>2]|0; //@line 106 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($27|0,($23|0)); //@line 106 "./common/Object.h"
    $28 = __THREW__; __THREW__ = 0;
    $29 = $28&1;
    if ($29) {
     $30 = ___cxa_find_matching_catch_3(0|0)|0;
     $31 = tempRet0;
     ___clang_call_terminate($30); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    } else {
     $32 = ((($this)) + 8|0); //@line 82 "modules/window/sdl/Window.cpp"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($32); //@line 82 "modules/window/sdl/Window.cpp"
     __ZN4love6window6WindowD2Ev($this); //@line 82 "modules/window/sdl/Window.cpp"
     return; //@line 82 "modules/window/sdl/Window.cpp"
    }
   }
  }
 } while(0);
 $33 = ___cxa_find_matching_catch_3(0|0)|0;
 $34 = tempRet0;
 $35 = ((($this)) + 104|0); //@line 106 "./common/Object.h"
 $36 = HEAP32[$35>>2]|0; //@line 106 "./common/Object.h"
 $37 = ($36|0)==(0|0); //@line 106 "./common/Object.h"
 if ($37) {
  $45 = ((($this)) + 8|0); //@line 82 "modules/window/sdl/Window.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($45); //@line 82 "modules/window/sdl/Window.cpp"
  __ZN4love6window6WindowD2Ev($this); //@line 82 "modules/window/sdl/Window.cpp"
  ___clang_call_terminate($33); //@line 82 "modules/window/sdl/Window.cpp"
  // unreachable; //@line 82 "modules/window/sdl/Window.cpp"
 }
 $38 = HEAP32[$36>>2]|0; //@line 106 "./common/Object.h"
 $39 = ((($38)) + 12|0); //@line 106 "./common/Object.h"
 $40 = HEAP32[$39>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($40|0,($36|0)); //@line 106 "./common/Object.h"
 $41 = __THREW__; __THREW__ = 0;
 $42 = $41&1;
 if ($42) {
  $43 = ___cxa_find_matching_catch_3(0|0)|0;
  $44 = tempRet0;
  ___clang_call_terminate($43); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  $45 = ((($this)) + 8|0); //@line 82 "modules/window/sdl/Window.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($45); //@line 82 "modules/window/sdl/Window.cpp"
  __ZN4love6window6WindowD2Ev($this); //@line 82 "modules/window/sdl/Window.cpp"
  ___clang_call_terminate($33); //@line 82 "modules/window/sdl/Window.cpp"
  // unreachable; //@line 82 "modules/window/sdl/Window.cpp"
 }
}
function __ZN4love6window3sdl6WindowD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6window3sdl6WindowD2Ev($this); //@line 78 "modules/window/sdl/Window.cpp"
 __ZdlPv($this); //@line 78 "modules/window/sdl/Window.cpp"
 return; //@line 82 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window6Window13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 16; //@line 109 "./modules/window/Window.h"
}
function __ZNK4love6window3sdl6Window7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (360380|0); //@line 1068 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window9setWindowEiiPNS0_14WindowSettingsE($this,$width,$height,$settings) {
 $this = $this|0;
 $width = $width|0;
 $height = $height|0;
 $settings = $settings|0;
 var $$03 = 0, $$06 = 0, $$2 = 0, $$25 = 0, $$28 = 0, $$pre = 0, $$pre13 = 0, $$pre14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $11 = 0;
 var $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $displaybounds = 0, $f = 0, $mode = 0, $mode1 = 0;
 var $or$cond = 0, $or$cond10 = 0, $or$cond12 = 0, $sdlflags$0 = 0, $sdlflags$0$ = 0, $sdlflags$2 = 0, $sdlflags$2$ = 0, $x$0 = 0, $y$0 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f = sp;
 $mode = sp + 100|0;
 $mode1 = sp + 80|0;
 $displaybounds = sp + 64|0;
 HEAP8[$f>>0] = 0; //@line 204 "./modules/window/Window.h"
 $0 = ((($f)) + 4|0); //@line 205 "./modules/window/Window.h"
 HEAP32[$0>>2] = 1; //@line 205 "./modules/window/Window.h"
 $1 = ((($f)) + 8|0); //@line 206 "./modules/window/Window.h"
 HEAP8[$1>>0] = 1; //@line 206 "./modules/window/Window.h"
 $2 = ((($f)) + 12|0); //@line 207 "./modules/window/Window.h"
 HEAP32[$2>>2] = 0; //@line 207 "./modules/window/Window.h"
 $3 = ((($f)) + 16|0); //@line 208 "./modules/window/Window.h"
 HEAP8[$3>>0] = 0; //@line 208 "./modules/window/Window.h"
 $4 = ((($f)) + 20|0); //@line 209 "./modules/window/Window.h"
 HEAP32[$4>>2] = 1; //@line 209 "./modules/window/Window.h"
 $5 = ((($f)) + 24|0); //@line 210 "./modules/window/Window.h"
 HEAP32[$5>>2] = 1; //@line 210 "./modules/window/Window.h"
 $6 = ((($f)) + 28|0); //@line 211 "./modules/window/Window.h"
 HEAP8[$6>>0] = 0; //@line 211 "./modules/window/Window.h"
 $7 = ((($f)) + 29|0); //@line 212 "./modules/window/Window.h"
 HEAP8[$7>>0] = 1; //@line 212 "./modules/window/Window.h"
 $8 = ((($f)) + 32|0); //@line 213 "./modules/window/Window.h"
 HEAP32[$8>>2] = 0; //@line 213 "./modules/window/Window.h"
 $9 = ((($f)) + 36|0); //@line 214 "./modules/window/Window.h"
 HEAP8[$9>>0] = 0; //@line 214 "./modules/window/Window.h"
 $10 = ((($f)) + 40|0); //@line 215 "./modules/window/Window.h"
 HEAPF64[$10>>3] = 0.0; //@line 215 "./modules/window/Window.h"
 $11 = ((($f)) + 48|0); //@line 216 "./modules/window/Window.h"
 HEAP8[$11>>0] = 0; //@line 216 "./modules/window/Window.h"
 $12 = ((($f)) + 52|0); //@line 217 "./modules/window/Window.h"
 HEAP32[$12>>2] = 0; //@line 217 "./modules/window/Window.h"
 $13 = ((($f)) + 56|0); //@line 218 "./modules/window/Window.h"
 HEAP32[$13>>2] = 0; //@line 218 "./modules/window/Window.h"
 $14 = ($settings|0)==(0|0); //@line 390 "modules/window/sdl/Window.cpp"
 if ($14) {
  $16 = 1;$19 = 1;$22 = 0;
 } else {
  dest=$f; src=$settings; stop=dest+60|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 391 "modules/window/sdl/Window.cpp"
  $$pre = HEAP32[$4>>2]|0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $$pre13 = HEAP32[$5>>2]|0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $$pre14 = HEAP32[$8>>2]|0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $16 = $$pre;$19 = $$pre13;$22 = $$pre14;
 }
 $15 = ($16|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $17 = $15 ? 1 : $16; //@line 478 "modules/window/sdl/Window.cpp"
 HEAP32[$4>>2] = $17; //@line 393 "modules/window/sdl/Window.cpp"
 $18 = ($19|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $20 = $18 ? 1 : $19; //@line 478 "modules/window/sdl/Window.cpp"
 HEAP32[$5>>2] = $20; //@line 394 "modules/window/sdl/Window.cpp"
 $21 = ($22|0)<(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $23 = HEAP32[$this>>2]|0; //@line 396 "modules/window/sdl/Window.cpp"
 $24 = ((($23)) + 48|0); //@line 396 "modules/window/sdl/Window.cpp"
 $25 = HEAP32[$24>>2]|0; //@line 396 "modules/window/sdl/Window.cpp"
 $26 = (FUNCTION_TABLE_ii[$25 & 8191]($this)|0); //@line 396 "modules/window/sdl/Window.cpp"
 $27 = (($26) + -1)|0; //@line 396 "modules/window/sdl/Window.cpp"
 $28 = $21 ? 0 : $22; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $29 = ($27|0)<($28|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $30 = $29 ? $27 : $28; //@line 454 "modules/window/sdl/Window.cpp"
 HEAP32[$8>>2] = $30; //@line 396 "modules/window/sdl/Window.cpp"
 $31 = ($width|0)==(0); //@line 399 "modules/window/sdl/Window.cpp"
 $32 = ($height|0)==(0); //@line 399 "modules/window/sdl/Window.cpp"
 $or$cond = $31 | $32; //@line 399 "modules/window/sdl/Window.cpp"
 if ($or$cond) {
  ;HEAP32[$mode>>2]=0|0;HEAP32[$mode+4>>2]=0|0;HEAP32[$mode+8>>2]=0|0;HEAP32[$mode+12>>2]=0|0;HEAP32[$mode+16>>2]=0|0; //@line 401 "modules/window/sdl/Window.cpp"
  (_SDL_GetDesktopDisplayMode($30,$mode)|0); //@line 402 "modules/window/sdl/Window.cpp"
  $33 = ((($mode)) + 4|0); //@line 403 "modules/window/sdl/Window.cpp"
  $34 = HEAP32[$33>>2]|0; //@line 403 "modules/window/sdl/Window.cpp"
  $35 = ((($mode)) + 8|0); //@line 404 "modules/window/sdl/Window.cpp"
  $36 = HEAP32[$35>>2]|0; //@line 404 "modules/window/sdl/Window.cpp"
  $$03 = $34;$$06 = $36;
 } else {
  $$03 = $width;$$06 = $height;
 }
 $37 = HEAP8[$f>>0]|0; //@line 480 "modules/window/sdl/Window.cpp"
 $38 = ($37<<24>>24)==(0); //@line 414 "modules/window/sdl/Window.cpp"
 if ($38) {
  $$25 = $$03;$$28 = $$06;$sdlflags$0 = 2;
 } else {
  $39 = HEAP32[$0>>2]|0; //@line 416 "modules/window/sdl/Window.cpp"
  $40 = ($39|0)==(1); //@line 416 "modules/window/sdl/Window.cpp"
  if ($40) {
   $$25 = $$03;$$28 = $$06;$sdlflags$0 = 4099;
  } else {
   HEAP32[$mode1>>2] = 0; //@line 421 "modules/window/sdl/Window.cpp"
   $41 = ((($mode1)) + 4|0); //@line 421 "modules/window/sdl/Window.cpp"
   HEAP32[$41>>2] = $$03; //@line 421 "modules/window/sdl/Window.cpp"
   $42 = ((($mode1)) + 8|0); //@line 421 "modules/window/sdl/Window.cpp"
   HEAP32[$42>>2] = $$06; //@line 421 "modules/window/sdl/Window.cpp"
   $43 = ((($mode1)) + 12|0); //@line 421 "modules/window/sdl/Window.cpp"
   HEAP32[$43>>2] = 0; //@line 421 "modules/window/sdl/Window.cpp"
   $44 = ((($mode1)) + 16|0); //@line 421 "modules/window/sdl/Window.cpp"
   HEAP32[$44>>2] = 0; //@line 421 "modules/window/sdl/Window.cpp"
   $45 = (_SDL_GetClosestDisplayMode($30,$mode1,$mode1)|0); //@line 424 "modules/window/sdl/Window.cpp"
   $46 = ($45|0)==(0|0); //@line 424 "modules/window/sdl/Window.cpp"
   if ($46) {
    $47 = (_SDL_GetDisplayMode($30,0,$mode1)|0); //@line 429 "modules/window/sdl/Window.cpp"
    $48 = ($47|0)<(0); //@line 429 "modules/window/sdl/Window.cpp"
    if ($48) {
     $$2 = 0;
     STACKTOP = sp;return ($$2|0); //@line 498 "modules/window/sdl/Window.cpp"
    }
   }
   $49 = HEAP32[$41>>2]|0; //@line 433 "modules/window/sdl/Window.cpp"
   $50 = HEAP32[$42>>2]|0; //@line 434 "modules/window/sdl/Window.cpp"
   $$25 = $49;$$28 = $50;$sdlflags$0 = 3;
  }
 }
 $51 = HEAP8[$3>>0]|0; //@line 438 "modules/window/sdl/Window.cpp"
 $52 = ($51<<24>>24)==(0); //@line 438 "modules/window/sdl/Window.cpp"
 $53 = $sdlflags$0 | 32; //@line 439 "modules/window/sdl/Window.cpp"
 $sdlflags$0$ = $52 ? $sdlflags$0 : $53; //@line 438 "modules/window/sdl/Window.cpp"
 $54 = HEAP16[$6>>1]|0; //@line 441 "modules/window/sdl/Window.cpp"
 $55 = $54&255; //@line 441 "modules/window/sdl/Window.cpp"
 $56 = ($55<<24>>24)==(0); //@line 441 "modules/window/sdl/Window.cpp"
 $57 = $sdlflags$0$ | 16; //@line 442 "modules/window/sdl/Window.cpp"
 $sdlflags$2 = $56 ? $sdlflags$0$ : $57; //@line 441 "modules/window/sdl/Window.cpp"
 $58 = HEAP8[$9>>0]|0; //@line 444 "modules/window/sdl/Window.cpp"
 $59 = ($58<<24>>24)==(0); //@line 444 "modules/window/sdl/Window.cpp"
 $60 = $sdlflags$2 | 8192; //@line 445 "modules/window/sdl/Window.cpp"
 $sdlflags$2$ = $59 ? $sdlflags$2 : $60; //@line 444 "modules/window/sdl/Window.cpp"
 $61 = HEAP8[$11>>0]|0; //@line 480 "modules/window/sdl/Window.cpp"
 $62 = ($61<<24>>24)!=(0); //@line 450 "modules/window/sdl/Window.cpp"
 $or$cond10 = $62 & $38; //@line 450 "modules/window/sdl/Window.cpp"
 $63 = ($54&65535) >>> 8; //@line 450 "modules/window/sdl/Window.cpp"
 $64 = $63&255; //@line 460 "modules/window/sdl/Window.cpp"
 do {
  if ($or$cond10) {
   $65 = HEAP32[$13>>2]|0; //@line 448 "modules/window/sdl/Window.cpp"
   $66 = HEAP32[$12>>2]|0; //@line 447 "modules/window/sdl/Window.cpp"
   ;HEAP32[$displaybounds>>2]=0|0;HEAP32[$displaybounds+4>>2]=0|0;HEAP32[$displaybounds+8>>2]=0|0;HEAP32[$displaybounds+12>>2]=0|0; //@line 453 "modules/window/sdl/Window.cpp"
   (_SDL_GetDisplayBounds($30,$displaybounds)|0); //@line 454 "modules/window/sdl/Window.cpp"
   $67 = HEAP32[$displaybounds>>2]|0; //@line 455 "modules/window/sdl/Window.cpp"
   $68 = (($67) + ($66))|0; //@line 455 "modules/window/sdl/Window.cpp"
   $69 = ((($displaybounds)) + 4|0); //@line 456 "modules/window/sdl/Window.cpp"
   $70 = HEAP32[$69>>2]|0; //@line 456 "modules/window/sdl/Window.cpp"
   $71 = (($70) + ($65))|0; //@line 456 "modules/window/sdl/Window.cpp"
   $x$0 = $68;$y$0 = $71;
  } else {
   $72 = ($64<<24>>24)==(0); //@line 460 "modules/window/sdl/Window.cpp"
   if ($72) {
    $74 = $30 | 536805376; //@line 463 "modules/window/sdl/Window.cpp"
    $x$0 = $74;$y$0 = $74;
    break;
   } else {
    $73 = $30 | 805240832; //@line 461 "modules/window/sdl/Window.cpp"
    $x$0 = $73;$y$0 = $73;
    break;
   }
  }
 } while(0);
 $75 = HEAP32[$this>>2]|0; //@line 466 "modules/window/sdl/Window.cpp"
 $76 = ((($75)) + 32|0); //@line 466 "modules/window/sdl/Window.cpp"
 $77 = HEAP32[$76>>2]|0; //@line 466 "modules/window/sdl/Window.cpp"
 FUNCTION_TABLE_vi[$77 & 8191]($this); //@line 466 "modules/window/sdl/Window.cpp"
 $78 = HEAP32[$2>>2]|0; //@line 468 "modules/window/sdl/Window.cpp"
 $79 = (__ZN4love6window3sdl6Window22createWindowAndContextEiiiiji($this,$x$0,$y$0,$$25,$$28,$sdlflags$2$,$78)|0); //@line 468 "modules/window/sdl/Window.cpp"
 if (!($79)) {
  $$2 = 0;
  STACKTOP = sp;return ($$2|0); //@line 498 "modules/window/sdl/Window.cpp"
 }
 $80 = HEAP32[$this>>2]|0; //@line 472 "modules/window/sdl/Window.cpp"
 $81 = ((($80)) + 84|0); //@line 472 "modules/window/sdl/Window.cpp"
 $82 = HEAP32[$81>>2]|0; //@line 472 "modules/window/sdl/Window.cpp"
 $83 = ((($this)) + 104|0); //@line 139 "./common/Object.h"
 $84 = HEAP32[$83>>2]|0; //@line 139 "./common/Object.h"
 (FUNCTION_TABLE_iii[$82 & 8191]($this,$84)|0); //@line 472 "modules/window/sdl/Window.cpp"
 $85 = HEAP32[$this>>2]|0; //@line 475 "modules/window/sdl/Window.cpp"
 $86 = ((($85)) + 132|0); //@line 475 "modules/window/sdl/Window.cpp"
 $87 = HEAP32[$86>>2]|0; //@line 475 "modules/window/sdl/Window.cpp"
 $88 = ((($this)) + 109|0); //@line 475 "modules/window/sdl/Window.cpp"
 $89 = HEAP8[$88>>0]|0; //@line 475 "modules/window/sdl/Window.cpp"
 $90 = ($89<<24>>24)!=(0); //@line 475 "modules/window/sdl/Window.cpp"
 FUNCTION_TABLE_vii[$87 & 8191]($this,$90); //@line 475 "modules/window/sdl/Window.cpp"
 $91 = ((($this)) + 112|0); //@line 478 "modules/window/sdl/Window.cpp"
 $92 = HEAP32[$91>>2]|0; //@line 478 "modules/window/sdl/Window.cpp"
 _SDL_SetWindowMinimumSize($92,$17,$20); //@line 478 "modules/window/sdl/Window.cpp"
 $93 = ($61<<24>>24)==(0); //@line 480 "modules/window/sdl/Window.cpp"
 if ($93) {
  $94 = ($64<<24>>24)!=(0); //@line 480 "modules/window/sdl/Window.cpp"
  $or$cond12 = $94 & $38; //@line 480 "modules/window/sdl/Window.cpp"
  if ($or$cond12) {
   label = 20;
  }
 } else {
  if ($38) {
   label = 20;
  }
 }
 if ((label|0) == 20) {
  $95 = HEAP32[$91>>2]|0; //@line 481 "modules/window/sdl/Window.cpp"
  _SDL_SetWindowPosition($95,$x$0,$y$0); //@line 481 "modules/window/sdl/Window.cpp"
 }
 $96 = HEAP32[$91>>2]|0; //@line 483 "modules/window/sdl/Window.cpp"
 _SDL_RaiseWindow($96); //@line 483 "modules/window/sdl/Window.cpp"
 $97 = HEAP8[$1>>0]|0; //@line 485 "modules/window/sdl/Window.cpp"
 $98 = $97&255; //@line 485 "modules/window/sdl/Window.cpp"
 (_SDL_GL_SetSwapInterval($98)|0); //@line 485 "modules/window/sdl/Window.cpp"
 __ZN4love6window3sdl6Window14updateSettingsERKNS0_14WindowSettingsE($this,$f); //@line 487 "modules/window/sdl/Window.cpp"
 $99 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $100 = ($99|0)==(0|0); //@line 490 "modules/window/sdl/Window.cpp"
 if ($100) {
  $$2 = 1;
  STACKTOP = sp;return ($$2|0); //@line 498 "modules/window/sdl/Window.cpp"
 }
 $101 = HEAP32[$99>>2]|0; //@line 491 "modules/window/sdl/Window.cpp"
 $102 = ((($101)) + 28|0); //@line 491 "modules/window/sdl/Window.cpp"
 $103 = HEAP32[$102>>2]|0; //@line 491 "modules/window/sdl/Window.cpp"
 $104 = ((($this)) + 28|0); //@line 491 "modules/window/sdl/Window.cpp"
 $105 = HEAP32[$104>>2]|0; //@line 491 "modules/window/sdl/Window.cpp"
 $106 = ((($this)) + 32|0); //@line 491 "modules/window/sdl/Window.cpp"
 $107 = HEAP32[$106>>2]|0; //@line 491 "modules/window/sdl/Window.cpp"
 (FUNCTION_TABLE_iiii[$103 & 8191]($99,$105,$107)|0); //@line 491 "modules/window/sdl/Window.cpp"
 $$2 = 1;
 STACKTOP = sp;return ($$2|0); //@line 498 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window22createWindowAndContextEiiiiji($this,$x,$y,$w,$h,$windowflags,$msaa) {
 $this = $this|0;
 $x = $x|0;
 $y = $y|0;
 $w = $w|0;
 $h = $h|0;
 $windowflags = $windowflags|0;
 $msaa = $msaa|0;
 var $$ = 0, $$$i = 0, $$0 = 0, $$07 = 0, $$08 = 0, $$1 = 0, $$1$i = 0, $$19 = 0, $$2 = 0, $$210 = 0, $$byval_copy = 0, $$byval_copy19 = 0, $$byval_copy20 = 0, $$byval_copy21 = 0, $$curSRGB$0$off0 = 0, $$msaa = 0, $$not = 0, $$not25 = 0, $$pre$phiZ2D = 0, $$pre113 = 0;
 var $$pre114 = 0, $$pre115 = 0, $$pre116 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0;
 var $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0;
 var $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0;
 var $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0;
 var $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0;
 var $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0;
 var $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0;
 var $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0;
 var $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $__begin2$sroa$0$0100 = 0, $__new_last2$i$i$i = 0, $__tmp$i1$i$i = 0, $attribslist = 0, $brmerge = 0, $brmerge26 = 0, $cmp = 0, $cmp$1 = 0, $cmp$2 = 0, $cmp$3 = 0, $cmp$4 = 0, $contexterror = 0, $create = 0, $curMSAA$0 = 0, $curSRGB$0$off0 = 0, $curSRGB$0$off0$not = 0, $curSRGB$1$off0 = 0;
 var $curSRGB$1$off0$lcssa = 0, $eh$lpad$body$index11Z2D = 0, $eh$lpad$body$indexZ2D = 0, $glversion = 0, $it$sroa$0$0101 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $message = 0, $not$ = 0, $not$96 = 0, $notlhs = 0, $notrhs = 0, $or$cond = 0, $or$cond5$not = 0, $preferGLES$3$off0 = 0, $profilemask$0$i = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i27 = 0, $scevgep$i$i$i32 = 0, $scevgep4$i$i$i = 0;
 var $scevgep4$i$i$i28 = 0, $scevgep4$i$i$i33 = 0, $strncmp = 0, $strncmp$1 = 0, $strncmp$2 = 0, $strncmp$3 = 0, $strncmp$4 = 0, $title = 0, $windowerror = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 320|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy21 = sp + 304|0;
 $$byval_copy20 = sp + 292|0;
 $$byval_copy19 = sp + 280|0;
 $$byval_copy = sp + 268|0;
 $0 = sp + 264|0;
 $__tmp$i1$i$i = sp + 252|0;
 $1 = sp + 248|0;
 $2 = sp + 244|0;
 $3 = sp + 240|0;
 $4 = sp + 236|0;
 $5 = sp + 232|0;
 $6 = sp + 228|0;
 $attribslist = sp + 216|0;
 $7 = sp + 180|0;
 $windowerror = sp + 168|0;
 $contexterror = sp + 156|0;
 $glversion = sp + 144|0;
 $create = sp + 108|0;
 $8 = sp + 96|0;
 $9 = sp + 84|0;
 $10 = sp + 72|0;
 $11 = sp + 60|0;
 $title = sp + 48|0;
 $message = sp + 36|0;
 $12 = sp + 24|0;
 $13 = sp + 12|0;
 $14 = sp;
 HEAP32[$2>>2] = $x;
 HEAP32[$3>>2] = $y;
 HEAP32[$4>>2] = $w;
 HEAP32[$5>>2] = $h;
 HEAP32[$6>>2] = $windowflags;
 $15 = (_SDL_GetCurrentVideoDriver()|0); //@line 199 "modules/window/sdl/Window.cpp"
 $16 = ($15|0)==(0|0); //@line 205 "modules/window/sdl/Window.cpp"
 do {
  if ($16) {
   label = 6;
  } else {
   $strncmp = (_strncmp($15,360396,3)|0); //@line 205 "modules/window/sdl/Window.cpp"
   $cmp = ($strncmp|0)==(0); //@line 205 "modules/window/sdl/Window.cpp"
   if (!($cmp)) {
    $strncmp$1 = (_strncmp($15,360400,7)|0); //@line 205 "modules/window/sdl/Window.cpp"
    $cmp$1 = ($strncmp$1|0)==(0); //@line 205 "modules/window/sdl/Window.cpp"
    if (!($cmp$1)) {
     $strncmp$2 = (_strncmp($15,360686,5)|0); //@line 205 "modules/window/sdl/Window.cpp"
     $cmp$2 = ($strncmp$2|0)==(0); //@line 205 "modules/window/sdl/Window.cpp"
     if (!($cmp$2)) {
      $strncmp$3 = (_strncmp($15,360692,5)|0); //@line 205 "modules/window/sdl/Window.cpp"
      $cmp$3 = ($strncmp$3|0)==(0); //@line 205 "modules/window/sdl/Window.cpp"
      if (!($cmp$3)) {
       $strncmp$4 = (_strncmp($15,610792,10)|0); //@line 205 "modules/window/sdl/Window.cpp"
       $cmp$4 = ($strncmp$4|0)==(0); //@line 205 "modules/window/sdl/Window.cpp"
       if (!($cmp$4)) {
        label = 6;
        break;
       }
      }
     }
    }
   }
   $17 = ((($this)) + 121|0); //@line 213 "modules/window/sdl/Window.cpp"
   $18 = HEAP8[$17>>0]|0; //@line 213 "modules/window/sdl/Window.cpp"
   $19 = ($18<<24>>24)==(0); //@line 213 "modules/window/sdl/Window.cpp"
   if ($19) {
    $preferGLES$3$off0 = 1;
   } else {
    __ZN4love8graphics15setGammaCorrectEb(0); //@line 214 "modules/window/sdl/Window.cpp"
    $preferGLES$3$off0 = 1;
   }
  }
 } while(0);
 if ((label|0) == 6) {
  $20 = (_SDL_GetHint(360408)|0); //@line 222 "modules/window/sdl/Window.cpp"
  $21 = ($20|0)==(0|0); //@line 223 "modules/window/sdl/Window.cpp"
  if ($21) {
   $preferGLES$3$off0 = 0;
  } else {
   $22 = HEAP8[$20>>0]|0; //@line 223 "modules/window/sdl/Window.cpp"
   $23 = ($22<<24>>24)!=(48); //@line 223 "modules/window/sdl/Window.cpp"
   $preferGLES$3$off0 = $23;
  }
 }
 $24 = (_SDL_GetHint(360435)|0); //@line 227 "modules/window/sdl/Window.cpp"
 $25 = ($24|0)==(0|0); //@line 228 "modules/window/sdl/Window.cpp"
 if ($25) {
  $32 = 0;
 } else {
  $26 = HEAP8[$24>>0]|0; //@line 228 "modules/window/sdl/Window.cpp"
  $27 = ($26<<24>>24)!=(48); //@line 228 "modules/window/sdl/Window.cpp"
  $32 = $27;
 }
 HEAP32[$7>>2] = 2; //@line 232 "modules/window/sdl/Window.cpp"
 $28 = ((($7)) + 4|0); //@line 232 "modules/window/sdl/Window.cpp"
 HEAP32[$28>>2] = 1; //@line 232 "modules/window/sdl/Window.cpp"
 $29 = ((($7)) + 8|0); //@line 232 "modules/window/sdl/Window.cpp"
 HEAP8[$29>>0] = 0; //@line 232 "modules/window/sdl/Window.cpp"
 $30 = ((($7)) + 9|0); //@line 232 "modules/window/sdl/Window.cpp"
 $31 = $32&1; //@line 232 "modules/window/sdl/Window.cpp"
 HEAP8[$30>>0] = $31; //@line 232 "modules/window/sdl/Window.cpp"
 $33 = ((($7)) + 12|0); //@line 233 "modules/window/sdl/Window.cpp"
 HEAP32[$33>>2] = 3; //@line 233 "modules/window/sdl/Window.cpp"
 $34 = ((($7)) + 16|0); //@line 233 "modules/window/sdl/Window.cpp"
 HEAP32[$34>>2] = 0; //@line 233 "modules/window/sdl/Window.cpp"
 $35 = ((($7)) + 20|0); //@line 233 "modules/window/sdl/Window.cpp"
 HEAP8[$35>>0] = 1; //@line 233 "modules/window/sdl/Window.cpp"
 $36 = ((($7)) + 21|0); //@line 233 "modules/window/sdl/Window.cpp"
 HEAP8[$36>>0] = $31; //@line 233 "modules/window/sdl/Window.cpp"
 $37 = ((($7)) + 24|0); //@line 234 "modules/window/sdl/Window.cpp"
 HEAP32[$37>>2] = 2; //@line 234 "modules/window/sdl/Window.cpp"
 $38 = ((($7)) + 28|0); //@line 234 "modules/window/sdl/Window.cpp"
 HEAP32[$38>>2] = 0; //@line 234 "modules/window/sdl/Window.cpp"
 $39 = ((($7)) + 32|0); //@line 234 "modules/window/sdl/Window.cpp"
 HEAP8[$39>>0] = 1; //@line 234 "modules/window/sdl/Window.cpp"
 $40 = ((($7)) + 33|0); //@line 234 "modules/window/sdl/Window.cpp"
 HEAP8[$40>>0] = $31; //@line 234 "modules/window/sdl/Window.cpp"
 HEAP32[$attribslist>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $41 = ((($attribslist)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$41>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $42 = ((($attribslist)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$42>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $43 = (__Znwj(36)|0); //@line 1030 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$attribslist>>2] = $43; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $44 = ((($43)) + 36|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $45 = ((($attribslist)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$45>>2] = $44; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 ;HEAP32[$43>>2]=HEAP32[$7>>2]|0;HEAP32[$43+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$43+8>>2]=HEAP32[$7+8>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $46 = ((($43)) + 12|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $47 = ((($7)) + 12|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 ;HEAP32[$46>>2]=HEAP32[$47>>2]|0;HEAP32[$46+4>>2]=HEAP32[$47+4>>2]|0;HEAP32[$46+8>>2]=HEAP32[$47+8>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $48 = ((($43)) + 24|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $49 = ((($7)) + 24|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 ;HEAP32[$48>>2]=HEAP32[$49>>2]|0;HEAP32[$48+4>>2]=HEAP32[$49+4>>2]|0;HEAP32[$48+8>>2]=HEAP32[$49+8>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $50 = ((($43)) + 36|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$41>>2] = $50; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $51 = ((($this)) + 121|0); //@line 238 "modules/window/sdl/Window.cpp"
 $52 = HEAP8[$51>>0]|0; //@line 238 "modules/window/sdl/Window.cpp"
 $53 = ($52<<24>>24)==(0); //@line 238 "modules/window/sdl/Window.cpp"
 if ($53) {
  $86 = $50;
 } else {
  $54 = $50; //@line 238 "modules/window/sdl/Window.cpp"
  $55 = $43; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $56 = $43; //@line 1681 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $418 = $50;$58 = $43;$69 = $54;$it$sroa$0$0101 = $55;
  while(1) {
   $57 = ((($58)) + 8|0); //@line 243 "modules/window/sdl/Window.cpp"
   $59 = HEAP8[$57>>0]|0; //@line 243 "modules/window/sdl/Window.cpp"
   $60 = ($59<<24>>24)==(0); //@line 243 "modules/window/sdl/Window.cpp"
   if ($60) {
    label = 16;
   } else {
    $61 = HEAP32[$58>>2]|0; //@line 243 "modules/window/sdl/Window.cpp"
    $62 = ($61|0)>(2); //@line 243 "modules/window/sdl/Window.cpp"
    if ($62) {
     $63 = (($it$sroa$0$0101) - ($56))|0; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $64 = (($63|0) / 12)&-1; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $65 = (($43) + (($64*12)|0)|0); //@line 1682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $66 = ((($65)) + 12|0); //@line 1684 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $67 = $66; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $68 = (($69) - ($67))|0; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $70 = (($68|0) / 12)&-1; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     _memmove(($65|0),($66|0),($68|0))|0; //@line 1893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $71 = (($65) + (($70*12)|0)|0); //@line 1894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $72 = HEAP32[$41>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $73 = ($72|0)==($71|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($73) {
      $$sink = $65;$81 = $71;
     } else {
      $__new_last2$i$i$i = $71; //@line 818 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $scevgep$i$i$i32 = ((($72)) + -12|0);
      $74 = $scevgep$i$i$i32;
      $75 = (($74) - ($__new_last2$i$i$i))|0;
      $76 = (($75>>>0) / 12)&-1;
      $77 = $76 ^ -1;
      $scevgep4$i$i$i33 = (($72) + (($77*12)|0)|0);
      HEAP32[$41>>2] = $scevgep4$i$i$i33; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$sink = $65;$81 = $scevgep4$i$i$i33;
     }
    } else {
     label = 16;
    }
   }
   if ((label|0) == 16) {
    label = 0;
    $78 = ((($58)) + 12|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $$sink = $78;$81 = $418;
   }
   $79 = $$sink; //@line 244 "modules/window/sdl/Window.cpp"
   $80 = ($$sink|0)==($81|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $82 = $81; //@line 241 "modules/window/sdl/Window.cpp"
   if ($80) {
    $86 = $81;
    break;
   } else {
    $418 = $81;$58 = $$sink;$69 = $82;$it$sroa$0$0101 = $79;
   }
  }
 }
 $83 = HEAP32[$attribslist>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($preferGLES$3$off0) {
  $84 = ((($83)) + 12|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $85 = ($84|0)==($86|0); //@line 1295 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if (!($85)) {
   $87 = $86; //@line 2520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $88 = $84; //@line 2504 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   ;HEAP32[$__tmp$i1$i$i>>2]=HEAP32[$83>>2]|0;HEAP32[$__tmp$i1$i$i+4>>2]=HEAP32[$83+4>>2]|0;HEAP32[$__tmp$i1$i$i+8>>2]=HEAP32[$83+8>>2]|0; //@line 2359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $89 = (($87) - ($88))|0; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $90 = (($89|0) / 12)&-1; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   _memmove(($83|0),($84|0),($89|0))|0; //@line 1893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $91 = (($83) + (($90*12)|0)|0); //@line 1894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   ;HEAP32[$91>>2]=HEAP32[$__tmp$i1$i$i>>2]|0;HEAP32[$91+4>>2]=HEAP32[$__tmp$i1$i$i+4>>2]|0;HEAP32[$91+8>>2]=HEAP32[$__tmp$i1$i$i+8>>2]|0; //@line 2361 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
 }
 ;HEAP32[$windowerror>>2]=0|0;HEAP32[$windowerror+4>>2]=0|0;HEAP32[$windowerror+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 ;HEAP32[$contexterror>>2]=0|0;HEAP32[$contexterror+4>>2]=0|0;HEAP32[$contexterror+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 ;HEAP32[$glversion>>2]=0|0;HEAP32[$glversion+4>>2]=0|0;HEAP32[$glversion+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 HEAP32[$create>>2] = $this; //@line 264 "modules/window/sdl/Window.cpp"
 $92 = ((($create)) + 4|0); //@line 264 "modules/window/sdl/Window.cpp"
 HEAP32[$92>>2] = $2; //@line 264 "modules/window/sdl/Window.cpp"
 $93 = ((($create)) + 8|0); //@line 264 "modules/window/sdl/Window.cpp"
 HEAP32[$93>>2] = $3; //@line 264 "modules/window/sdl/Window.cpp"
 $94 = ((($create)) + 12|0); //@line 264 "modules/window/sdl/Window.cpp"
 HEAP32[$94>>2] = $4; //@line 264 "modules/window/sdl/Window.cpp"
 $95 = ((($create)) + 16|0); //@line 264 "modules/window/sdl/Window.cpp"
 HEAP32[$95>>2] = $5; //@line 264 "modules/window/sdl/Window.cpp"
 $96 = ((($create)) + 20|0); //@line 264 "modules/window/sdl/Window.cpp"
 HEAP32[$96>>2] = $6; //@line 264 "modules/window/sdl/Window.cpp"
 $97 = ((($create)) + 24|0); //@line 264 "modules/window/sdl/Window.cpp"
 HEAP32[$97>>2] = $windowerror; //@line 264 "modules/window/sdl/Window.cpp"
 $98 = ((($create)) + 28|0); //@line 264 "modules/window/sdl/Window.cpp"
 HEAP32[$98>>2] = $contexterror; //@line 264 "modules/window/sdl/Window.cpp"
 $99 = ((($create)) + 32|0); //@line 264 "modules/window/sdl/Window.cpp"
 HEAP32[$99>>2] = $glversion; //@line 264 "modules/window/sdl/Window.cpp"
 $100 = ($83|0)==($86|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 L33: do {
  if ($100) {
   $$pre116 = ((($this)) + 116|0); //@line 357 "modules/window/sdl/Window.cpp"
   $$pre$phiZ2D = $$pre116;
   label = 56;
  } else {
   $101 = ((($windowerror)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $102 = ((($contexterror)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $103 = ((($8)) + 4|0); //@line 321 "modules/window/sdl/Window.cpp"
   $104 = ((($8)) + 8|0); //@line 321 "modules/window/sdl/Window.cpp"
   $105 = ((($8)) + 10|0); //@line 321 "modules/window/sdl/Window.cpp"
   $106 = ((($this)) + 112|0); //@line 323 "modules/window/sdl/Window.cpp"
   $107 = ($msaa|0)>(0); //@line 323 "modules/window/sdl/Window.cpp"
   $108 = ((($9)) + 4|0); //@line 327 "modules/window/sdl/Window.cpp"
   $109 = ((($9)) + 8|0); //@line 327 "modules/window/sdl/Window.cpp"
   $110 = ((($9)) + 10|0); //@line 327 "modules/window/sdl/Window.cpp"
   $111 = ((($this)) + 116|0); //@line 350 "modules/window/sdl/Window.cpp"
   $112 = ((($11)) + 4|0); //@line 343 "modules/window/sdl/Window.cpp"
   $113 = ((($11)) + 8|0); //@line 343 "modules/window/sdl/Window.cpp"
   $114 = ((($11)) + 10|0); //@line 343 "modules/window/sdl/Window.cpp"
   $115 = ((($10)) + 4|0); //@line 335 "modules/window/sdl/Window.cpp"
   $116 = ((($10)) + 8|0); //@line 335 "modules/window/sdl/Window.cpp"
   $117 = ((($10)) + 10|0); //@line 335 "modules/window/sdl/Window.cpp"
   $118 = ((($contexterror)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $119 = ((($contexterror)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $120 = ((($windowerror)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $121 = ((($windowerror)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $__begin2$sroa$0$0100 = $83;
   while(1) {
    $122 = HEAP32[$__begin2$sroa$0$0100>>2]|0; //@line 310 "modules/window/sdl/Window.cpp"
    $123 = ((($__begin2$sroa$0$0100)) + 4|0); //@line 310 "modules/window/sdl/Window.cpp"
    $124 = HEAP32[$123>>2]|0; //@line 310 "modules/window/sdl/Window.cpp"
    $125 = ((($__begin2$sroa$0$0100)) + 8|0); //@line 310 "modules/window/sdl/Window.cpp"
    $126 = HEAP32[$125>>2]|0; //@line 310 "modules/window/sdl/Window.cpp"
    $127 = $126&65535; //@line 310 "modules/window/sdl/Window.cpp"
    $128 = $126 >>> 16; //@line 310 "modules/window/sdl/Window.cpp"
    $129 = $128&65535; //@line 343 "modules/window/sdl/Window.cpp"
    __THREW__ = 0;
    $130 = (invoke_i(3801)|0); //@line 313 "modules/window/sdl/Window.cpp"
    $131 = __THREW__; __THREW__ = 0;
    $132 = $131&1;
    if ($132) {
     label = 41;
     break;
    }
    __THREW__ = 0;
    invoke_viii(3802,($this|0),($msaa|0),($130|0)); //@line 315 "modules/window/sdl/Window.cpp"
    $133 = __THREW__; __THREW__ = 0;
    $134 = $133&1;
    if ($134) {
     label = 41;
     break;
    }
    $135 = $126&255; //@line 123 "modules/window/sdl/Window.cpp"
    $136 = ($135<<24>>24)==(0); //@line 123 "modules/window/sdl/Window.cpp"
    $137 = ($127&65535) >>> 8; //@line 123 "modules/window/sdl/Window.cpp"
    $138 = ($137<<16>>16)==(0); //@line 125 "modules/window/sdl/Window.cpp"
    $$$i = $138 ? 0 : 2; //@line 125 "modules/window/sdl/Window.cpp"
    $profilemask$0$i = $136 ? $$$i : 4; //@line 123 "modules/window/sdl/Window.cpp"
    $$1$i = $137&65535; //@line 128 "modules/window/sdl/Window.cpp"
    __THREW__ = 0;
    (invoke_iii(3803,17,($122|0))|0); //@line 131 "modules/window/sdl/Window.cpp"
    $139 = __THREW__; __THREW__ = 0;
    $140 = $139&1;
    if ($140) {
     label = 41;
     break;
    }
    __THREW__ = 0;
    (invoke_iii(3803,18,($124|0))|0); //@line 132 "modules/window/sdl/Window.cpp"
    $141 = __THREW__; __THREW__ = 0;
    $142 = $141&1;
    if ($142) {
     label = 41;
     break;
    }
    __THREW__ = 0;
    (invoke_iii(3803,21,($profilemask$0$i|0))|0); //@line 133 "modules/window/sdl/Window.cpp"
    $143 = __THREW__; __THREW__ = 0;
    $144 = $143&1;
    if ($144) {
     label = 41;
     break;
    }
    __THREW__ = 0;
    (invoke_iii(3803,20,($$1$i|0))|0); //@line 134 "modules/window/sdl/Window.cpp"
    $145 = __THREW__; __THREW__ = 0;
    $146 = $145&1;
    if ($146) {
     label = 41;
     break;
    }
    $147 = HEAP8[$windowerror>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $148 = $147 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $149 = ($148<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if ($149) {
     HEAP8[$101>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP8[$windowerror>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    } else {
     $150 = HEAP32[$120>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP8[$150>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP32[$121>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    }
    $151 = HEAP8[$contexterror>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $152 = $151 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $153 = ($152<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if ($153) {
     HEAP8[$102>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP8[$contexterror>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    } else {
     $154 = HEAP32[$118>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP8[$154>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP32[$119>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    }
    HEAP32[$8>>2] = $122; //@line 321 "modules/window/sdl/Window.cpp"
    HEAP32[$103>>2] = $124; //@line 321 "modules/window/sdl/Window.cpp"
    HEAP16[$104>>1] = $127; //@line 321 "modules/window/sdl/Window.cpp"
    HEAP16[$105>>1] = $129; //@line 321 "modules/window/sdl/Window.cpp"
    __THREW__ = 0;
    ;HEAP32[$$byval_copy>>2]=HEAP32[$8>>2]|0;HEAP32[$$byval_copy+4>>2]=HEAP32[$8+4>>2]|0;HEAP32[$$byval_copy+8>>2]=HEAP32[$8+8>>2]|0; //@line 321 "modules/window/sdl/Window.cpp"
    (invoke_iii(3804,($create|0),($$byval_copy|0))|0); //@line 321 "modules/window/sdl/Window.cpp"
    $155 = __THREW__; __THREW__ = 0;
    $156 = $155&1;
    if ($156) {
     label = 41;
     break;
    }
    $157 = HEAP32[$106>>2]|0; //@line 323 "modules/window/sdl/Window.cpp"
    $158 = ($157|0)==(0|0); //@line 323 "modules/window/sdl/Window.cpp"
    $or$cond = $107 & $158; //@line 323 "modules/window/sdl/Window.cpp"
    if ($or$cond) {
     __THREW__ = 0;
     invoke_viii(3802,($this|0),0,($130|0)); //@line 326 "modules/window/sdl/Window.cpp"
     $159 = __THREW__; __THREW__ = 0;
     $160 = $159&1;
     if ($160) {
      label = 41;
      break;
     }
     HEAP32[$9>>2] = $122; //@line 327 "modules/window/sdl/Window.cpp"
     HEAP32[$108>>2] = $124; //@line 327 "modules/window/sdl/Window.cpp"
     HEAP16[$109>>1] = $127; //@line 327 "modules/window/sdl/Window.cpp"
     HEAP16[$110>>1] = $129; //@line 327 "modules/window/sdl/Window.cpp"
     __THREW__ = 0;
     ;HEAP32[$$byval_copy19>>2]=HEAP32[$9>>2]|0;HEAP32[$$byval_copy19+4>>2]=HEAP32[$9+4>>2]|0;HEAP32[$$byval_copy19+8>>2]=HEAP32[$9+8>>2]|0; //@line 327 "modules/window/sdl/Window.cpp"
     $161 = (invoke_iii(3804,($create|0),($$byval_copy19|0))|0); //@line 327 "modules/window/sdl/Window.cpp"
     $162 = __THREW__; __THREW__ = 0;
     $163 = $162&1;
     if ($163) {
      label = 41;
      break;
     }
     $$msaa = $161 ? 0 : $msaa; //@line 327 "modules/window/sdl/Window.cpp"
     $$pre113 = HEAP32[$106>>2]|0; //@line 331 "modules/window/sdl/Window.cpp"
     $168 = $$pre113;$curMSAA$0 = $$msaa;
    } else {
     $168 = $157;$curMSAA$0 = $msaa;
    }
    $$not = ($168|0)!=(0|0); //@line 331 "modules/window/sdl/Window.cpp"
    $$not25 = $130 ^ 1; //@line 331 "modules/window/sdl/Window.cpp"
    $brmerge = $$not | $$not25; //@line 331 "modules/window/sdl/Window.cpp"
    if ($brmerge) {
     $174 = $168;$curSRGB$0$off0 = $130;
    } else {
     __THREW__ = 0;
     invoke_viii(3802,($this|0),($curMSAA$0|0),0); //@line 334 "modules/window/sdl/Window.cpp"
     $169 = __THREW__; __THREW__ = 0;
     $170 = $169&1;
     if ($170) {
      label = 41;
      break;
     }
     HEAP32[$10>>2] = $122; //@line 335 "modules/window/sdl/Window.cpp"
     HEAP32[$115>>2] = $124; //@line 335 "modules/window/sdl/Window.cpp"
     HEAP16[$116>>1] = $127; //@line 335 "modules/window/sdl/Window.cpp"
     HEAP16[$117>>1] = $129; //@line 335 "modules/window/sdl/Window.cpp"
     __THREW__ = 0;
     ;HEAP32[$$byval_copy20>>2]=HEAP32[$10>>2]|0;HEAP32[$$byval_copy20+4>>2]=HEAP32[$10+4>>2]|0;HEAP32[$$byval_copy20+8>>2]=HEAP32[$10+8>>2]|0; //@line 335 "modules/window/sdl/Window.cpp"
     $171 = (invoke_iii(3804,($create|0),($$byval_copy20|0))|0); //@line 335 "modules/window/sdl/Window.cpp"
     $172 = __THREW__; __THREW__ = 0;
     $173 = $172&1;
     if ($173) {
      label = 41;
      break;
     }
     $not$ = $171 ^ 1; //@line 335 "modules/window/sdl/Window.cpp"
     $$ = $130 & $not$; //@line 335 "modules/window/sdl/Window.cpp"
     $$pre114 = HEAP32[$106>>2]|0; //@line 339 "modules/window/sdl/Window.cpp"
     $174 = $$pre114;$curSRGB$0$off0 = $$;
    }
    $notlhs = ($174|0)!=(0|0); //@line 339 "modules/window/sdl/Window.cpp"
    $notrhs = ($curMSAA$0|0)<(1); //@line 339 "modules/window/sdl/Window.cpp"
    $or$cond5$not = $notrhs | $notlhs; //@line 339 "modules/window/sdl/Window.cpp"
    $curSRGB$0$off0$not = $curSRGB$0$off0 ^ 1; //@line 339 "modules/window/sdl/Window.cpp"
    $brmerge26 = $or$cond5$not | $curSRGB$0$off0$not; //@line 339 "modules/window/sdl/Window.cpp"
    if ($brmerge26) {
     $181 = $174;$curSRGB$1$off0 = $curSRGB$0$off0;
    } else {
     __THREW__ = 0;
     invoke_viii(3802,($this|0),0,0); //@line 342 "modules/window/sdl/Window.cpp"
     $175 = __THREW__; __THREW__ = 0;
     $176 = $175&1;
     if ($176) {
      label = 41;
      break;
     }
     HEAP32[$11>>2] = $122; //@line 343 "modules/window/sdl/Window.cpp"
     HEAP32[$112>>2] = $124; //@line 343 "modules/window/sdl/Window.cpp"
     HEAP16[$113>>1] = $127; //@line 343 "modules/window/sdl/Window.cpp"
     HEAP16[$114>>1] = $129; //@line 343 "modules/window/sdl/Window.cpp"
     __THREW__ = 0;
     ;HEAP32[$$byval_copy21>>2]=HEAP32[$11>>2]|0;HEAP32[$$byval_copy21+4>>2]=HEAP32[$11+4>>2]|0;HEAP32[$$byval_copy21+8>>2]=HEAP32[$11+8>>2]|0; //@line 343 "modules/window/sdl/Window.cpp"
     $177 = (invoke_iii(3804,($create|0),($$byval_copy21|0))|0); //@line 343 "modules/window/sdl/Window.cpp"
     $178 = __THREW__; __THREW__ = 0;
     $179 = $178&1;
     if ($179) {
      label = 41;
      break;
     }
     $not$96 = $177 ^ 1; //@line 343 "modules/window/sdl/Window.cpp"
     $$curSRGB$0$off0 = $curSRGB$0$off0 & $not$96; //@line 343 "modules/window/sdl/Window.cpp"
     $$pre115 = HEAP32[$106>>2]|0; //@line 350 "modules/window/sdl/Window.cpp"
     $181 = $$pre115;$curSRGB$1$off0 = $$curSRGB$0$off0;
    }
    $180 = ($181|0)==(0|0); //@line 350 "modules/window/sdl/Window.cpp"
    if (!($180)) {
     $182 = HEAP32[$111>>2]|0; //@line 350 "modules/window/sdl/Window.cpp"
     $183 = ($182|0)==(0|0); //@line 350 "modules/window/sdl/Window.cpp"
     if (!($183)) {
      $curSRGB$1$off0$lcssa = $curSRGB$1$off0;
      label = 54;
      break;
     }
    }
    $186 = ((($__begin2$sroa$0$0100)) + 12|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $187 = ($186|0)==($86|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($187) {
     $$pre$phiZ2D = $111;
     label = 56;
     break L33;
    } else {
     $__begin2$sroa$0$0100 = $186;
    }
   }
   if ((label|0) == 41) {
    $164 = ___cxa_find_matching_catch_2()|0;
    $165 = tempRet0;
    $lpad$phi$index = $164;$lpad$phi$index2 = $165;
    label = 43;
    break;
   }
   else if ((label|0) == 54) {
    __THREW__ = 0;
    invoke_vi(3805,($curSRGB$1$off0$lcssa|0)); //@line 352 "modules/window/sdl/Window.cpp"
    $184 = __THREW__; __THREW__ = 0;
    $185 = $184&1;
    if ($185) {
     label = 42;
     break;
    } else {
     $$pre$phiZ2D = $111;
     label = 56;
     break;
    }
   }
  }
 } while(0);
 L74: do {
  if ((label|0) == 56) {
   $188 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 357 "modules/window/sdl/Window.cpp"
   $189 = ($188|0)==(0|0); //@line 357 "modules/window/sdl/Window.cpp"
   if ($189) {
    label = 58;
   } else {
    $190 = ((($this)) + 112|0); //@line 357 "modules/window/sdl/Window.cpp"
    $191 = HEAP32[$190>>2]|0; //@line 357 "modules/window/sdl/Window.cpp"
    $192 = ($191|0)==(0|0); //@line 357 "modules/window/sdl/Window.cpp"
    if ($192) {
     label = 58;
    } else {
     $405 = ((($this)) + 108|0); //@line 382 "modules/window/sdl/Window.cpp"
     HEAP8[$405>>0] = 1; //@line 382 "modules/window/sdl/Window.cpp"
     $$0 = 1;
    }
   }
   L79: do {
    if ((label|0) == 58) {
     __THREW__ = 0;
     invoke_viii(3460,($title|0),(360466|0),30); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $193 = __THREW__; __THREW__ = 0;
     $194 = $193&1;
     if ($194) {
      label = 42;
      break L74;
     }
     __THREW__ = 0;
     invoke_viii(3460,($message|0),(360497|0),96); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $195 = __THREW__; __THREW__ = 0;
     $196 = $195&1;
     if ($196) {
      $234 = ___cxa_find_matching_catch_2()|0;
      $235 = tempRet0;
      $$1 = $235;$$19 = $234;
     } else {
      $197 = HEAP8[$glversion>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $198 = $197 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $199 = ($198<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $200 = ((($glversion)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $201 = HEAP32[$200>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $202 = $197&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $203 = $202 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $204 = $199 ? $203 : $201; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $205 = ($204|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      do {
       if ($205) {
        $240 = HEAP8[$contexterror>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $241 = $240 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $242 = ($241<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $243 = ((($contexterror)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $244 = HEAP32[$243>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $245 = $240&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $246 = $245 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $247 = $242 ? $246 : $244; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $248 = ($247|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if (!($248)) {
         ;HEAP32[$13>>2]=0|0;HEAP32[$13+4>>2]=0|0;HEAP32[$13+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $249 = (($247) + 33)|0; //@line 3957 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         __THREW__ = 0;
         invoke_viiii(3514,($13|0),(360622|0),33,($249|0)); //@line 3957 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $250 = __THREW__; __THREW__ = 0;
         $251 = $250&1;
         if (!($251)) {
          $252 = HEAP8[$contexterror>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $253 = $252 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $254 = ($253<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $255 = ((($contexterror)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $256 = HEAP32[$255>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $257 = ((($contexterror)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $258 = $254 ? $257 : $256; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          __THREW__ = 0;
          (invoke_iiii(3462,($13|0),($258|0),($247|0))|0); //@line 3958 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $259 = __THREW__; __THREW__ = 0;
          $260 = $259&1;
          if (!($260)) {
           $263 = HEAP8[$13>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $264 = $263 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $265 = ($264<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $266 = ((($13)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $267 = HEAP32[$266>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $268 = ((($13)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $269 = $265 ? $268 : $267; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $270 = ((($13)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $271 = HEAP32[$270>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $272 = $263&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $273 = $272 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $274 = $265 ? $273 : $271; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           __THREW__ = 0;
           (invoke_iiii(3462,($message|0),($269|0),($274|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $275 = __THREW__; __THREW__ = 0;
           $276 = $275&1;
           if ($276) {
            $277 = ___cxa_find_matching_catch_2()|0;
            $278 = tempRet0;
            __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13); //@line 365 "modules/window/sdl/Window.cpp"
            $$07 = $278;$$08 = $277;
            break;
           } else {
            __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13); //@line 365 "modules/window/sdl/Window.cpp"
            label = 84;
            break;
           }
          }
         }
         $261 = ___cxa_find_matching_catch_2()|0;
         $262 = tempRet0;
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13); //@line 3960 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $eh$lpad$body$index11Z2D = $262;$eh$lpad$body$indexZ2D = $261;
         label = 68;
         break;
        }
        $279 = HEAP8[$windowerror>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $280 = $279 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $281 = ($280<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $282 = ((($windowerror)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $283 = HEAP32[$282>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $284 = $279&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $285 = $284 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $286 = $281 ? $285 : $283; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $287 = ($286|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if ($287) {
         label = 84;
        } else {
         ;HEAP32[$14>>2]=0|0;HEAP32[$14+4>>2]=0|0;HEAP32[$14+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $288 = (($286) + 29)|0; //@line 3957 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         __THREW__ = 0;
         invoke_viiii(3514,($14|0),(360656|0),29,($288|0)); //@line 3957 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $289 = __THREW__; __THREW__ = 0;
         $290 = $289&1;
         if (!($290)) {
          $291 = HEAP8[$windowerror>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $292 = $291 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $293 = ($292<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $294 = ((($windowerror)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $295 = HEAP32[$294>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $296 = ((($windowerror)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $297 = $293 ? $296 : $295; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          __THREW__ = 0;
          (invoke_iiii(3462,($14|0),($297|0),($286|0))|0); //@line 3958 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $298 = __THREW__; __THREW__ = 0;
          $299 = $298&1;
          if (!($299)) {
           $302 = HEAP8[$14>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $303 = $302 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $304 = ($303<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $305 = ((($14)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $306 = HEAP32[$305>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $307 = ((($14)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $308 = $304 ? $307 : $306; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $309 = ((($14)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $310 = HEAP32[$309>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $311 = $302&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $312 = $311 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $313 = $304 ? $312 : $310; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           __THREW__ = 0;
           (invoke_iiii(3462,($message|0),($308|0),($313|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $314 = __THREW__; __THREW__ = 0;
           $315 = $314&1;
           if ($315) {
            $316 = ___cxa_find_matching_catch_2()|0;
            $317 = tempRet0;
            __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($14); //@line 367 "modules/window/sdl/Window.cpp"
            $$07 = $317;$$08 = $316;
            break;
           } else {
            __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($14); //@line 367 "modules/window/sdl/Window.cpp"
            label = 84;
            break;
           }
          }
         }
         $300 = ___cxa_find_matching_catch_2()|0;
         $301 = tempRet0;
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($14); //@line 3960 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $eh$lpad$body$index11Z2D = $301;$eh$lpad$body$indexZ2D = $300;
         label = 68;
        }
       } else {
        ;HEAP32[$12>>2]=0|0;HEAP32[$12+4>>2]=0|0;HEAP32[$12+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $206 = (($204) + 27)|0; //@line 3957 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        __THREW__ = 0;
        invoke_viiii(3514,($12|0),(360594|0),27,($206|0)); //@line 3957 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $207 = __THREW__; __THREW__ = 0;
        $208 = $207&1;
        if (!($208)) {
         $209 = HEAP8[$glversion>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $210 = $209 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $211 = ($210<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $212 = ((($glversion)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $213 = HEAP32[$212>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $214 = ((($glversion)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $215 = $211 ? $214 : $213; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         __THREW__ = 0;
         (invoke_iiii(3462,($12|0),($215|0),($204|0))|0); //@line 3958 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $216 = __THREW__; __THREW__ = 0;
         $217 = $216&1;
         if (!($217)) {
          $220 = HEAP8[$12>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $221 = $220 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $222 = ($221<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $223 = ((($12)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $224 = HEAP32[$223>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $225 = ((($12)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $226 = $222 ? $225 : $224; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $227 = ((($12)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $228 = HEAP32[$227>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $229 = $220&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $230 = $229 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $231 = $222 ? $230 : $228; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          __THREW__ = 0;
          (invoke_iiii(3462,($message|0),($226|0),($231|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $232 = __THREW__; __THREW__ = 0;
          $233 = $232&1;
          if ($233) {
           $238 = ___cxa_find_matching_catch_2()|0;
           $239 = tempRet0;
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12); //@line 363 "modules/window/sdl/Window.cpp"
           $$07 = $239;$$08 = $238;
           break;
          } else {
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12); //@line 363 "modules/window/sdl/Window.cpp"
           label = 84;
           break;
          }
         }
        }
        $218 = ___cxa_find_matching_catch_2()|0;
        $219 = tempRet0;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12); //@line 3960 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $eh$lpad$body$index11Z2D = $219;$eh$lpad$body$indexZ2D = $218;
        label = 68;
       }
      } while(0);
      L113: do {
       if ((label|0) == 84) {
        $318 = HEAP8[$title>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $319 = $318 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $320 = ($319<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $321 = ((($title)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $322 = HEAP32[$321>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $323 = ((($title)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $324 = $320 ? $323 : $322; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $325 = ((($title)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $326 = HEAP32[$325>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $327 = $318&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $328 = $327 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $329 = $320 ? $328 : $326; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        __THREW__ = 0;
        $330 = (invoke_iiii(3366,(1206696|0),($324|0),($329|0))|0); //@line 1080 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
        $331 = __THREW__; __THREW__ = 0;
        $332 = $331&1;
        L115: do {
         if (!($332)) {
          $333 = HEAP32[$330>>2]|0; //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
          $334 = ((($333)) + -12|0); //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
          $335 = HEAP32[$334>>2]|0; //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
          $336 = (($330) + ($335)|0); //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
          __THREW__ = 0;
          $337 = (invoke_ii(3373,($336|0))|0); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
          $338 = __THREW__; __THREW__ = 0;
          $339 = $338&1;
          if (!($339)) {
           HEAP32[$0>>2] = $337; //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
           __THREW__ = 0;
           $340 = (invoke_iii(3374,($0|0),(1207576|0))|0); //@line 181 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
           $341 = __THREW__; __THREW__ = 0;
           $342 = $341&1;
           if (!($342)) {
            $343 = HEAP32[$340>>2]|0; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
            $344 = ((($343)) + 28|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
            $345 = HEAP32[$344>>2]|0; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
            __THREW__ = 0;
            $346 = (invoke_iii($345|0,($340|0),10)|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
            $347 = __THREW__; __THREW__ = 0;
            $348 = $347&1;
            if (!($348)) {
             __ZNSt3__16localeD2Ev($0); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
             __THREW__ = 0;
             (invoke_iii(3501,($330|0),($346|0))|0); //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
             $351 = __THREW__; __THREW__ = 0;
             $352 = $351&1;
             if ($352) {
              break;
             }
             __THREW__ = 0;
             (invoke_ii(3502,($330|0))|0); //@line 1035 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
             $353 = __THREW__; __THREW__ = 0;
             $354 = $353&1;
             if ($354) {
              break;
             }
             $355 = HEAP8[$message>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $356 = $355 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $357 = ($356<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $358 = ((($message)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $359 = HEAP32[$358>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $360 = ((($message)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $361 = $357 ? $360 : $359; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $362 = ((($message)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $363 = HEAP32[$362>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $364 = $355&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $365 = $364 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $366 = $357 ? $365 : $363; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             __THREW__ = 0;
             $367 = (invoke_iiii(3366,($330|0),($361|0),($366|0))|0); //@line 1080 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
             $368 = __THREW__; __THREW__ = 0;
             $369 = $368&1;
             if ($369) {
              break;
             }
             $370 = HEAP32[$367>>2]|0; //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
             $371 = ((($370)) + -12|0); //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
             $372 = HEAP32[$371>>2]|0; //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
             $373 = (($367) + ($372)|0); //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
             __THREW__ = 0;
             $374 = (invoke_ii(3373,($373|0))|0); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
             $375 = __THREW__; __THREW__ = 0;
             $376 = $375&1;
             if ($376) {
              break;
             }
             HEAP32[$1>>2] = $374; //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
             __THREW__ = 0;
             $377 = (invoke_iii(3374,($1|0),(1207576|0))|0); //@line 181 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
             $378 = __THREW__; __THREW__ = 0;
             $379 = $378&1;
             do {
              if (!($379)) {
               $380 = HEAP32[$377>>2]|0; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
               $381 = ((($380)) + 28|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
               $382 = HEAP32[$381>>2]|0; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
               __THREW__ = 0;
               $383 = (invoke_iii($382|0,($377|0),10)|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
               $384 = __THREW__; __THREW__ = 0;
               $385 = $384&1;
               if ($385) {
                break;
               }
               __ZNSt3__16localeD2Ev($1); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
               __THREW__ = 0;
               (invoke_iii(3501,($367|0),($383|0))|0); //@line 1034 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
               $388 = __THREW__; __THREW__ = 0;
               $389 = $388&1;
               if ($389) {
                break L115;
               }
               __THREW__ = 0;
               (invoke_ii(3502,($367|0))|0); //@line 1035 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
               $390 = __THREW__; __THREW__ = 0;
               $391 = $390&1;
               if ($391) {
                break L115;
               }
               $392 = ((($this)) + 120|0); //@line 372 "modules/window/sdl/Window.cpp"
               $393 = HEAP8[$392>>0]|0; //@line 372 "modules/window/sdl/Window.cpp"
               $394 = ($393<<24>>24)==(0); //@line 372 "modules/window/sdl/Window.cpp"
               if ($394) {
                $395 = HEAP32[$this>>2]|0; //@line 374 "modules/window/sdl/Window.cpp"
                $396 = ((($395)) + 176|0); //@line 374 "modules/window/sdl/Window.cpp"
                $397 = HEAP32[$396>>2]|0; //@line 374 "modules/window/sdl/Window.cpp"
                __THREW__ = 0;
                (invoke_iiiiii($397|0,($this|0),($title|0),($message|0),0,0)|0); //@line 374 "modules/window/sdl/Window.cpp"
                $398 = __THREW__; __THREW__ = 0;
                $399 = $398&1;
                if ($399) {
                 break L115;
                }
                HEAP8[$392>>0] = 1; //@line 375 "modules/window/sdl/Window.cpp"
               }
               $400 = HEAP32[$this>>2]|0; //@line 378 "modules/window/sdl/Window.cpp"
               $401 = ((($400)) + 32|0); //@line 378 "modules/window/sdl/Window.cpp"
               $402 = HEAP32[$401>>2]|0; //@line 378 "modules/window/sdl/Window.cpp"
               __THREW__ = 0;
               invoke_vi($402|0,($this|0)); //@line 378 "modules/window/sdl/Window.cpp"
               $403 = __THREW__; __THREW__ = 0;
               $404 = $403&1;
               if ($404) {
                break L115;
               }
               __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($message); //@line 380 "modules/window/sdl/Window.cpp"
               __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($title); //@line 380 "modules/window/sdl/Window.cpp"
               $$0 = 0;
               break L79;
              }
             } while(0);
             $386 = ___cxa_find_matching_catch_2()|0;
             $387 = tempRet0;
             __ZNSt3__16localeD2Ev($1); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
             $eh$lpad$body$index11Z2D = $387;$eh$lpad$body$indexZ2D = $386;
             label = 68;
             break L113;
            }
           }
           $349 = ___cxa_find_matching_catch_2()|0;
           $350 = tempRet0;
           __ZNSt3__16localeD2Ev($0); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
           $eh$lpad$body$index11Z2D = $350;$eh$lpad$body$indexZ2D = $349;
           label = 68;
           break L113;
          }
         }
        } while(0);
        $236 = ___cxa_find_matching_catch_2()|0;
        $237 = tempRet0;
        $eh$lpad$body$index11Z2D = $237;$eh$lpad$body$indexZ2D = $236;
        label = 68;
       }
      } while(0);
      if ((label|0) == 68) {
       $$07 = $eh$lpad$body$index11Z2D;$$08 = $eh$lpad$body$indexZ2D;
      }
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($message); //@line 380 "modules/window/sdl/Window.cpp"
      $$1 = $$07;$$19 = $$08;
     }
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($title); //@line 380 "modules/window/sdl/Window.cpp"
     $$2 = $$1;$$210 = $$19;
     break L74;
    }
   } while(0);
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($glversion); //@line 384 "modules/window/sdl/Window.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($contexterror); //@line 384 "modules/window/sdl/Window.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($windowerror); //@line 384 "modules/window/sdl/Window.cpp"
   $406 = ($83|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $407 = $83; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($406) {
    STACKTOP = sp;return ($$0|0); //@line 384 "modules/window/sdl/Window.cpp"
   }
   if (!($100)) {
    $scevgep$i$i$i27 = ((($86)) + -12|0);
    $408 = $scevgep$i$i$i27;
    $409 = (($408) - ($407))|0;
    $410 = (($409>>>0) / 12)&-1;
    $411 = $410 ^ -1;
    $scevgep4$i$i$i28 = (($86) + (($411*12)|0)|0);
    HEAP32[$41>>2] = $scevgep4$i$i$i28; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($83); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   STACKTOP = sp;return ($$0|0); //@line 384 "modules/window/sdl/Window.cpp"
  }
 } while(0);
 if ((label|0) == 42) {
  $166 = ___cxa_find_matching_catch_2()|0;
  $167 = tempRet0;
  $lpad$phi$index = $166;$lpad$phi$index2 = $167;
  label = 43;
 }
 if ((label|0) == 43) {
  $$2 = $lpad$phi$index2;$$210 = $lpad$phi$index;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($glversion); //@line 384 "modules/window/sdl/Window.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($contexterror); //@line 384 "modules/window/sdl/Window.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($windowerror); //@line 384 "modules/window/sdl/Window.cpp"
 $412 = ($83|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $413 = $83; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($412) {
  ___resumeException($$210|0);
  // unreachable;
 }
 if (!($100)) {
  $scevgep$i$i$i = ((($86)) + -12|0);
  $414 = $scevgep$i$i$i;
  $415 = (($414) - ($413))|0;
  $416 = (($415>>>0) / 12)&-1;
  $417 = $416 ^ -1;
  $scevgep4$i$i$i = (($86) + (($417*12)|0)|0);
  HEAP32[$41>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($83); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$210|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love6window3sdl6Window26setGLFramebufferAttributesEib($this,$msaa,$sRGB) {
 $this = $this|0;
 $msaa = $msaa|0;
 $sRGB = $sRGB|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $cmp = 0, $strncmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 (_SDL_GL_SetAttribute(0,8)|0); //@line 87 "modules/window/sdl/Window.cpp"
 (_SDL_GL_SetAttribute(1,8)|0); //@line 88 "modules/window/sdl/Window.cpp"
 (_SDL_GL_SetAttribute(2,8)|0); //@line 89 "modules/window/sdl/Window.cpp"
 (_SDL_GL_SetAttribute(3,8)|0); //@line 90 "modules/window/sdl/Window.cpp"
 (_SDL_GL_SetAttribute(5,1)|0); //@line 91 "modules/window/sdl/Window.cpp"
 (_SDL_GL_SetAttribute(7,8)|0); //@line 92 "modules/window/sdl/Window.cpp"
 (_SDL_GL_SetAttribute(16,0)|0); //@line 93 "modules/window/sdl/Window.cpp"
 $0 = ($msaa|0)>(0); //@line 95 "modules/window/sdl/Window.cpp"
 $1 = $0&1; //@line 95 "modules/window/sdl/Window.cpp"
 (_SDL_GL_SetAttribute(13,$1)|0); //@line 95 "modules/window/sdl/Window.cpp"
 $2 = $0 ? $msaa : 0; //@line 96 "modules/window/sdl/Window.cpp"
 (_SDL_GL_SetAttribute(14,$2)|0); //@line 96 "modules/window/sdl/Window.cpp"
 $3 = $sRGB&1; //@line 98 "modules/window/sdl/Window.cpp"
 (_SDL_GL_SetAttribute(23,$3)|0); //@line 98 "modules/window/sdl/Window.cpp"
 $4 = (_SDL_GetCurrentVideoDriver()|0); //@line 100 "modules/window/sdl/Window.cpp"
 $5 = ($4|0)==(0|0); //@line 101 "modules/window/sdl/Window.cpp"
 if ($5) {
  return; //@line 116 "modules/window/sdl/Window.cpp"
 }
 $strncmp = (_strncmp($4,360455,3)|0); //@line 101 "modules/window/sdl/Window.cpp"
 $cmp = ($strncmp|0)==(0); //@line 101 "modules/window/sdl/Window.cpp"
 if (!($cmp)) {
  return; //@line 116 "modules/window/sdl/Window.cpp"
 }
 $6 = ((($this)) + 121|0); //@line 106 "modules/window/sdl/Window.cpp"
 $7 = HEAP8[$6>>0]|0; //@line 106 "modules/window/sdl/Window.cpp"
 $8 = ($7<<24>>24)==(0); //@line 106 "modules/window/sdl/Window.cpp"
 if ($8) {
  return; //@line 116 "modules/window/sdl/Window.cpp"
 }
 (_SDL_GL_SetAttribute(23,0)|0); //@line 107 "modules/window/sdl/Window.cpp"
 return; //@line 116 "modules/window/sdl/Window.cpp"
}
function __ZZN4love6window3sdl6Window22createWindowAndContextEiiiijiENK3__0clENS2_14ContextAttribsE($this,$attribs) {
 $this = $this|0;
 $attribs = $attribs|0;
 var $$0 = 0, $$pr = 0, $$pr3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 12|0;
 $1 = sp;
 $2 = HEAP32[$this>>2]|0;
 $3 = ((($2)) + 116|0); //@line 266 "modules/window/sdl/Window.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 266 "modules/window/sdl/Window.cpp"
 $5 = ($4|0)==(0|0); //@line 266 "modules/window/sdl/Window.cpp"
 if (!($5)) {
  _SDL_GL_DeleteContext($4); //@line 268 "modules/window/sdl/Window.cpp"
  HEAP32[$3>>2] = 0; //@line 269 "modules/window/sdl/Window.cpp"
 }
 $6 = ((($2)) + 112|0); //@line 272 "modules/window/sdl/Window.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 272 "modules/window/sdl/Window.cpp"
 $8 = ($7|0)==(0|0); //@line 272 "modules/window/sdl/Window.cpp"
 if (!($8)) {
  _SDL_DestroyWindow($7); //@line 274 "modules/window/sdl/Window.cpp"
  _SDL_FlushEvent(512); //@line 275 "modules/window/sdl/Window.cpp"
  HEAP32[$6>>2] = 0; //@line 276 "modules/window/sdl/Window.cpp"
 }
 $9 = ((($2)) + 8|0); //@line 279 "modules/window/sdl/Window.cpp"
 $10 = HEAP8[$9>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $10 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = ($11<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($12) {
  $15 = ((($9)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $31 = $15;
 } else {
  $13 = ((($2)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $14 = HEAP32[$13>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $31 = $14;
 }
 $16 = ((($this)) + 4|0); //@line 279 "modules/window/sdl/Window.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 279 "modules/window/sdl/Window.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 279 "modules/window/sdl/Window.cpp"
 $19 = ((($this)) + 8|0); //@line 279 "modules/window/sdl/Window.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 279 "modules/window/sdl/Window.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 279 "modules/window/sdl/Window.cpp"
 $22 = ((($this)) + 12|0); //@line 279 "modules/window/sdl/Window.cpp"
 $23 = HEAP32[$22>>2]|0; //@line 279 "modules/window/sdl/Window.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 279 "modules/window/sdl/Window.cpp"
 $25 = ((($this)) + 16|0); //@line 279 "modules/window/sdl/Window.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 279 "modules/window/sdl/Window.cpp"
 $27 = HEAP32[$26>>2]|0; //@line 279 "modules/window/sdl/Window.cpp"
 $28 = ((($this)) + 20|0); //@line 279 "modules/window/sdl/Window.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 279 "modules/window/sdl/Window.cpp"
 $30 = HEAP32[$29>>2]|0; //@line 279 "modules/window/sdl/Window.cpp"
 $32 = (_SDL_CreateWindow($31,$18,$21,$24,$27,$30)|0); //@line 279 "modules/window/sdl/Window.cpp"
 HEAP32[$6>>2] = $32; //@line 279 "modules/window/sdl/Window.cpp"
 $33 = ($32|0)==(0|0); //@line 281 "modules/window/sdl/Window.cpp"
 if ($33) {
  $34 = ((($this)) + 24|0); //@line 283 "modules/window/sdl/Window.cpp"
  $35 = HEAP32[$34>>2]|0; //@line 283 "modules/window/sdl/Window.cpp"
  $36 = (_SDL_GetError()|0); //@line 283 "modules/window/sdl/Window.cpp"
  $37 = (_strlen($36)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($0,$36,$37); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $38 = HEAP8[$35>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $39 = $38 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $40 = ($39<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($40) {
   $44 = ((($35)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP8[$44>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP8[$35>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  } else {
   $41 = ((($35)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $42 = HEAP32[$41>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP8[$42>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $43 = ((($35)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP32[$43>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  }
  __THREW__ = 0;
  invoke_vii(3368,($35|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $45 = __THREW__; __THREW__ = 0;
  $46 = $45&1;
  if ($46) {
   $47 = ___cxa_find_matching_catch_3(0|0)|0;
   $48 = tempRet0;
   ___clang_call_terminate($47); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  }
  ;HEAP32[$35>>2]=HEAP32[$0>>2]|0;HEAP32[$35+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$35+8>>2]=HEAP32[$0+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 283 "modules/window/sdl/Window.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 307 "modules/window/sdl/Window.cpp"
 }
 $49 = (_SDL_GL_CreateContext($32)|0); //@line 287 "modules/window/sdl/Window.cpp"
 HEAP32[$3>>2] = $49; //@line 287 "modules/window/sdl/Window.cpp"
 $50 = ($49|0)==(0|0); //@line 289 "modules/window/sdl/Window.cpp"
 do {
  if ($50) {
   $51 = ((($this)) + 28|0); //@line 290 "modules/window/sdl/Window.cpp"
   $52 = HEAP32[$51>>2]|0; //@line 290 "modules/window/sdl/Window.cpp"
   $53 = (_SDL_GetError()|0); //@line 290 "modules/window/sdl/Window.cpp"
   $54 = (_strlen($53)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($1,$53,$54); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $55 = HEAP8[$52>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $56 = $55 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $57 = ($56<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if ($57) {
    $61 = ((($52)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$61>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$52>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   } else {
    $58 = ((($52)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $59 = HEAP32[$58>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$59>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $60 = ((($52)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP32[$60>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   }
   __THREW__ = 0;
   invoke_vii(3368,($52|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $62 = __THREW__; __THREW__ = 0;
   $63 = $62&1;
   if ($63) {
    $64 = ___cxa_find_matching_catch_3(0|0)|0;
    $65 = tempRet0;
    ___clang_call_terminate($64); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   } else {
    ;HEAP32[$52>>2]=HEAP32[$1>>2]|0;HEAP32[$52+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$52+8>>2]=HEAP32[$1+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 290 "modules/window/sdl/Window.cpp"
    $$pr = HEAP32[$3>>2]|0; //@line 293 "modules/window/sdl/Window.cpp"
    $66 = ($$pr|0)==(0|0); //@line 293 "modules/window/sdl/Window.cpp"
    if ($66) {
     break;
    } else {
     label = 22;
     break;
    }
   }
  } else {
   label = 22;
  }
 } while(0);
 do {
  if ((label|0) == 22) {
   $67 = ((($this)) + 32|0); //@line 293 "modules/window/sdl/Window.cpp"
   $68 = HEAP32[$67>>2]|0; //@line 293 "modules/window/sdl/Window.cpp"
   $69 = (__ZN4love6window3sdl6Window14checkGLVersionERKNS2_14ContextAttribsERNSt3__112basic_stringIcNS6_11char_traitsIcEENS6_9allocatorIcEEEE(0,$attribs,$68)|0); //@line 293 "modules/window/sdl/Window.cpp"
   $$pr3 = HEAP32[$3>>2]|0; //@line 299 "modules/window/sdl/Window.cpp"
   if (!($69)) {
    _SDL_GL_DeleteContext($$pr3); //@line 295 "modules/window/sdl/Window.cpp"
    HEAP32[$3>>2] = 0; //@line 296 "modules/window/sdl/Window.cpp"
    break;
   }
   $70 = ($$pr3|0)==(0|0); //@line 299 "modules/window/sdl/Window.cpp"
   if (!($70)) {
    $$0 = 1;
    STACKTOP = sp;return ($$0|0); //@line 307 "modules/window/sdl/Window.cpp"
   }
  }
 } while(0);
 $71 = HEAP32[$6>>2]|0; //@line 301 "modules/window/sdl/Window.cpp"
 _SDL_DestroyWindow($71); //@line 301 "modules/window/sdl/Window.cpp"
 HEAP32[$6>>2] = 0; //@line 302 "modules/window/sdl/Window.cpp"
 $$0 = 0;
 STACKTOP = sp;return ($$0|0); //@line 307 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window14checkGLVersionERKNS2_14ContextAttribsERNSt3__112basic_stringIcNS6_11char_traitsIcEENS6_9allocatorIcEEEE($this,$attribs,$outversion) {
 $this = $this|0;
 $attribs = $attribs|0;
 $outversion = $outversion|0;
 var $$01 = 0, $$04 = 0, $$12 = 0, $$15 = 0, $$2 = 0, $$23 = 0, $$26 = 0, $$3 = 0, $$37 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 48|0;
 $1 = sp + 36|0;
 $2 = sp + 24|0;
 $3 = sp + 12|0;
 $4 = sp;
 $5 = (_SDL_GL_GetProcAddress(613977)|0); //@line 148 "modules/window/sdl/Window.cpp"
 $6 = ($5|0)==(0|0); //@line 149 "modules/window/sdl/Window.cpp"
 if ($6) {
  $$2 = 0;
  STACKTOP = sp;return ($$2|0); //@line 189 "modules/window/sdl/Window.cpp"
 }
 $7 = (FUNCTION_TABLE_ii[$5 & 8191](7938)|0); //@line 152 "modules/window/sdl/Window.cpp"
 $8 = ($7|0)==(0|0); //@line 153 "modules/window/sdl/Window.cpp"
 if ($8) {
  $$2 = 0;
  STACKTOP = sp;return ($$2|0); //@line 189 "modules/window/sdl/Window.cpp"
 }
 (__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc($outversion,$7)|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = (FUNCTION_TABLE_ii[$5 & 8191](7937)|0); //@line 158 "modules/window/sdl/Window.cpp"
 $10 = ($9|0)==(0|0); //@line 159 "modules/window/sdl/Window.cpp"
 L7: do {
  if (!($10)) {
   $11 = (_strlen($9)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($1,$9,$11); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   $12 = (invoke_iiii(3640,($1|0),0,(360459|0))|0); //@line 4027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $13 = __THREW__; __THREW__ = 0;
   $14 = $13&1;
   do {
    if ($14) {
     $29 = ___cxa_find_matching_catch_2()|0;
     $30 = tempRet0;
     $$01 = $29;$$04 = $30;
    } else {
     ;HEAP32[$0>>2]=HEAP32[$12>>2]|0;HEAP32[$0+4>>2]=HEAP32[$12+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$12+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     ;HEAP32[$12>>2]=0|0;HEAP32[$12+4>>2]=0|0;HEAP32[$12+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $15 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $16 = $15 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $17 = ($16<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $18 = ((($0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $19 = HEAP32[$18>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $20 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $21 = $17 ? $20 : $19; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $22 = ((($0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $23 = HEAP32[$22>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $24 = $15&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $25 = $24 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $26 = $17 ? $25 : $23; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     __THREW__ = 0;
     (invoke_iiii(3462,($outversion|0),($21|0),($26|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $27 = __THREW__; __THREW__ = 0;
     $28 = $27&1;
     if ($28) {
      $31 = ___cxa_find_matching_catch_2()|0;
      $32 = tempRet0;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 160 "modules/window/sdl/Window.cpp"
      $$01 = $31;$$04 = $32;
      break;
     } else {
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 160 "modules/window/sdl/Window.cpp"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 160 "modules/window/sdl/Window.cpp"
      break L7;
     }
    }
   } while(0);
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 160 "modules/window/sdl/Window.cpp"
   $$3 = $$01;$$37 = $$04;
   ___resumeException($$3|0);
   // unreachable;
  }
 } while(0);
 $33 = (FUNCTION_TABLE_ii[$5 & 8191](7936)|0); //@line 162 "modules/window/sdl/Window.cpp"
 $34 = ($33|0)==(0|0); //@line 163 "modules/window/sdl/Window.cpp"
 L18: do {
  if (!($34)) {
   $35 = (_strlen($33)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($4,$33,$35); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   $36 = (invoke_iiii(3640,($4|0),0,(360463|0))|0); //@line 4027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $37 = __THREW__; __THREW__ = 0;
   $38 = $37&1;
   if ($38) {
    $56 = ___cxa_find_matching_catch_2()|0;
    $57 = tempRet0;
    $$23 = $56;$$26 = $57;
   } else {
    ;HEAP32[$3>>2]=HEAP32[$36>>2]|0;HEAP32[$3+4>>2]=HEAP32[$36+4>>2]|0;HEAP32[$3+8>>2]=HEAP32[$36+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    ;HEAP32[$36>>2]=0|0;HEAP32[$36+4>>2]=0|0;HEAP32[$36+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    __THREW__ = 0;
    $39 = (invoke_iii(3806,($3|0),(442473|0))|0); //@line 4044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $40 = __THREW__; __THREW__ = 0;
    $41 = $40&1;
    do {
     if ($41) {
      $58 = ___cxa_find_matching_catch_2()|0;
      $59 = tempRet0;
      $$12 = $58;$$15 = $59;
     } else {
      ;HEAP32[$2>>2]=HEAP32[$39>>2]|0;HEAP32[$2+4>>2]=HEAP32[$39+4>>2]|0;HEAP32[$2+8>>2]=HEAP32[$39+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      ;HEAP32[$39>>2]=0|0;HEAP32[$39+4>>2]=0|0;HEAP32[$39+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $42 = HEAP8[$2>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $43 = $42 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $44 = ($43<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $45 = ((($2)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $46 = HEAP32[$45>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $47 = ((($2)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $48 = $44 ? $47 : $46; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $49 = ((($2)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $50 = HEAP32[$49>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $51 = $42&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $52 = $51 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $53 = $44 ? $52 : $50; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __THREW__ = 0;
      (invoke_iiii(3462,($outversion|0),($48|0),($53|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $54 = __THREW__; __THREW__ = 0;
      $55 = $54&1;
      if ($55) {
       $60 = ___cxa_find_matching_catch_2()|0;
       $61 = tempRet0;
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 164 "modules/window/sdl/Window.cpp"
       $$12 = $60;$$15 = $61;
       break;
      } else {
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 164 "modules/window/sdl/Window.cpp"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 164 "modules/window/sdl/Window.cpp"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 164 "modules/window/sdl/Window.cpp"
       break L18;
      }
     }
    } while(0);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 164 "modules/window/sdl/Window.cpp"
    $$23 = $$12;$$26 = $$15;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 164 "modules/window/sdl/Window.cpp"
   $$3 = $$23;$$37 = $$26;
   ___resumeException($$3|0);
   // unreachable;
  }
 } while(0);
 $62 = HEAP32[$attribs>>2]|0; //@line 184 "modules/window/sdl/Window.cpp"
 $63 = ($62|0)>(2); //@line 184 "modules/window/sdl/Window.cpp"
 if ($63) {
  $$2 = 0;
  STACKTOP = sp;return ($$2|0); //@line 189 "modules/window/sdl/Window.cpp"
 }
 $64 = ($62|0)==(2); //@line 185 "modules/window/sdl/Window.cpp"
 if ($64) {
  $65 = ((($attribs)) + 4|0); //@line 185 "modules/window/sdl/Window.cpp"
  $66 = HEAP32[$65>>2]|0; //@line 185 "modules/window/sdl/Window.cpp"
  $67 = ($66|0)>(0); //@line 185 "modules/window/sdl/Window.cpp"
  if ($67) {
   $$2 = 0;
   STACKTOP = sp;return ($$2|0); //@line 189 "modules/window/sdl/Window.cpp"
  }
 }
 $$2 = 1;
 STACKTOP = sp;return ($$2|0); //@line 189 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window14updateSettingsERKNS0_14WindowSettingsE($this,$newsettings) {
 $this = $this|0;
 $newsettings = $newsettings|0;
 var $$lobit = 0, $$lobit1 = 0, $$lobit2 = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0.0, $62 = 0, $7 = 0, $8 = 0, $9 = 0, $buffers = 0, $dmode = 0, $samples = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $buffers = sp + 24|0;
 $samples = sp + 20|0;
 $dmode = sp;
 $0 = ((($this)) + 112|0); //@line 519 "modules/window/sdl/Window.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 519 "modules/window/sdl/Window.cpp"
 $2 = (_SDL_GetWindowFlags($1)|0); //@line 519 "modules/window/sdl/Window.cpp"
 $3 = HEAP32[$0>>2]|0; //@line 522 "modules/window/sdl/Window.cpp"
 $4 = ((($this)) + 20|0); //@line 522 "modules/window/sdl/Window.cpp"
 $5 = ((($this)) + 24|0); //@line 522 "modules/window/sdl/Window.cpp"
 _SDL_GetWindowSize($3,$4,$5); //@line 522 "modules/window/sdl/Window.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 523 "modules/window/sdl/Window.cpp"
 $7 = ((($this)) + 28|0); //@line 523 "modules/window/sdl/Window.cpp"
 $8 = ((($this)) + 32|0); //@line 523 "modules/window/sdl/Window.cpp"
 _SDL_GL_GetDrawableSize($6,$7,$8); //@line 523 "modules/window/sdl/Window.cpp"
 $9 = $2 & 4097; //@line 525 "modules/window/sdl/Window.cpp"
 $10 = ($9|0)==(4097); //@line 525 "modules/window/sdl/Window.cpp"
 do {
  if ($10) {
   $11 = ((($this)) + 40|0); //@line 527 "modules/window/sdl/Window.cpp"
   HEAP8[$11>>0] = 1; //@line 527 "modules/window/sdl/Window.cpp"
   $12 = ((($this)) + 44|0); //@line 528 "modules/window/sdl/Window.cpp"
   HEAP32[$12>>2] = 1; //@line 528 "modules/window/sdl/Window.cpp"
   $$pre$phiZ2D = $11;
  } else {
   $13 = $2 & 1; //@line 530 "modules/window/sdl/Window.cpp"
   $14 = ($13|0)==(0); //@line 530 "modules/window/sdl/Window.cpp"
   $15 = ((($this)) + 40|0); //@line 532 "modules/window/sdl/Window.cpp"
   if ($14) {
    HEAP8[$15>>0] = 0; //@line 537 "modules/window/sdl/Window.cpp"
    $17 = ((($newsettings)) + 4|0); //@line 538 "modules/window/sdl/Window.cpp"
    $18 = HEAP32[$17>>2]|0; //@line 538 "modules/window/sdl/Window.cpp"
    $19 = ((($this)) + 44|0); //@line 538 "modules/window/sdl/Window.cpp"
    HEAP32[$19>>2] = $18; //@line 538 "modules/window/sdl/Window.cpp"
    $$pre$phiZ2D = $15;
    break;
   } else {
    HEAP8[$15>>0] = 1; //@line 532 "modules/window/sdl/Window.cpp"
    $16 = ((($this)) + 44|0); //@line 533 "modules/window/sdl/Window.cpp"
    HEAP32[$16>>2] = 0; //@line 533 "modules/window/sdl/Window.cpp"
    $$pre$phiZ2D = $15;
    break;
   }
  }
 } while(0);
 $20 = ((($newsettings)) + 20|0); //@line 546 "modules/window/sdl/Window.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 546 "modules/window/sdl/Window.cpp"
 $22 = ((($this)) + 60|0); //@line 546 "modules/window/sdl/Window.cpp"
 HEAP32[$22>>2] = $21; //@line 546 "modules/window/sdl/Window.cpp"
 $23 = ((($newsettings)) + 24|0); //@line 547 "modules/window/sdl/Window.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 547 "modules/window/sdl/Window.cpp"
 $25 = ((($this)) + 64|0); //@line 547 "modules/window/sdl/Window.cpp"
 HEAP32[$25>>2] = $24; //@line 547 "modules/window/sdl/Window.cpp"
 $26 = ((($this)) + 56|0); //@line 549 "modules/window/sdl/Window.cpp"
 $27 = $2 >>> 5; //@line 549 "modules/window/sdl/Window.cpp"
 $$lobit = $27 & 1; //@line 549 "modules/window/sdl/Window.cpp"
 $28 = $$lobit&255; //@line 549 "modules/window/sdl/Window.cpp"
 HEAP8[$26>>0] = $28; //@line 549 "modules/window/sdl/Window.cpp"
 $29 = ((($this)) + 68|0); //@line 550 "modules/window/sdl/Window.cpp"
 $30 = $2 >>> 4; //@line 550 "modules/window/sdl/Window.cpp"
 $$lobit1 = $30 & 1; //@line 550 "modules/window/sdl/Window.cpp"
 $31 = $$lobit1&255; //@line 550 "modules/window/sdl/Window.cpp"
 HEAP8[$29>>0] = $31; //@line 550 "modules/window/sdl/Window.cpp"
 $32 = ((($newsettings)) + 29|0); //@line 551 "modules/window/sdl/Window.cpp"
 $33 = HEAP8[$32>>0]|0; //@line 551 "modules/window/sdl/Window.cpp"
 $34 = ((($this)) + 69|0); //@line 551 "modules/window/sdl/Window.cpp"
 HEAP8[$34>>0] = $33; //@line 551 "modules/window/sdl/Window.cpp"
 $35 = HEAP32[$this>>2]|0; //@line 553 "modules/window/sdl/Window.cpp"
 $36 = ((($35)) + 68|0); //@line 553 "modules/window/sdl/Window.cpp"
 $37 = HEAP32[$36>>2]|0; //@line 553 "modules/window/sdl/Window.cpp"
 $38 = ((($this)) + 92|0); //@line 553 "modules/window/sdl/Window.cpp"
 $39 = ((($this)) + 96|0); //@line 553 "modules/window/sdl/Window.cpp"
 $40 = ((($this)) + 72|0); //@line 553 "modules/window/sdl/Window.cpp"
 FUNCTION_TABLE_viiii[$37 & 8191]($this,$38,$39,$40); //@line 553 "modules/window/sdl/Window.cpp"
 $41 = ((($this)) + 76|0); //@line 555 "modules/window/sdl/Window.cpp"
 $42 = $2 >>> 13; //@line 555 "modules/window/sdl/Window.cpp"
 $$lobit2 = $42 & 1; //@line 555 "modules/window/sdl/Window.cpp"
 $43 = $$lobit2&255; //@line 555 "modules/window/sdl/Window.cpp"
 HEAP8[$41>>0] = $43; //@line 555 "modules/window/sdl/Window.cpp"
 $44 = HEAP8[$$pre$phiZ2D>>0]|0; //@line 558 "modules/window/sdl/Window.cpp"
 $45 = ($44<<24>>24)==(0); //@line 558 "modules/window/sdl/Window.cpp"
 if ($45) {
  label = 9;
 } else {
  $46 = ((($this)) + 44|0); //@line 558 "modules/window/sdl/Window.cpp"
  $47 = HEAP32[$46>>2]|0; //@line 558 "modules/window/sdl/Window.cpp"
  $48 = ($47|0)==(0); //@line 558 "modules/window/sdl/Window.cpp"
  if ($48) {
   (_SDL_SetHint(604513,532555)|0); //@line 559 "modules/window/sdl/Window.cpp"
  } else {
   label = 9;
  }
 }
 if ((label|0) == 9) {
  (_SDL_SetHint(604513,604546)|0); //@line 561 "modules/window/sdl/Window.cpp"
 }
 HEAP32[$buffers>>2] = 0; //@line 564 "modules/window/sdl/Window.cpp"
 HEAP32[$samples>>2] = 0; //@line 565 "modules/window/sdl/Window.cpp"
 (_SDL_GL_GetAttribute(13,$buffers)|0); //@line 566 "modules/window/sdl/Window.cpp"
 (_SDL_GL_GetAttribute(14,$samples)|0); //@line 567 "modules/window/sdl/Window.cpp"
 $49 = HEAP32[$buffers>>2]|0; //@line 569 "modules/window/sdl/Window.cpp"
 $50 = ($49|0)>(0); //@line 569 "modules/window/sdl/Window.cpp"
 $51 = HEAP32[$samples>>2]|0; //@line 569 "modules/window/sdl/Window.cpp"
 $52 = $50 ? $51 : 0; //@line 569 "modules/window/sdl/Window.cpp"
 $53 = ((($this)) + 52|0); //@line 569 "modules/window/sdl/Window.cpp"
 HEAP32[$53>>2] = $52; //@line 569 "modules/window/sdl/Window.cpp"
 $54 = (_SDL_GL_GetSwapInterval()|0); //@line 570 "modules/window/sdl/Window.cpp"
 $55 = ($54|0)!=(0); //@line 570 "modules/window/sdl/Window.cpp"
 $56 = ((($this)) + 48|0); //@line 570 "modules/window/sdl/Window.cpp"
 $57 = $55&1; //@line 570 "modules/window/sdl/Window.cpp"
 HEAP8[$56>>0] = $57; //@line 570 "modules/window/sdl/Window.cpp"
 ;HEAP32[$dmode>>2]=0|0;HEAP32[$dmode+4>>2]=0|0;HEAP32[$dmode+8>>2]=0|0;HEAP32[$dmode+12>>2]=0|0;HEAP32[$dmode+16>>2]=0|0; //@line 572 "modules/window/sdl/Window.cpp"
 $58 = HEAP32[$40>>2]|0; //@line 573 "modules/window/sdl/Window.cpp"
 (_SDL_GetCurrentDisplayMode($58,$dmode)|0); //@line 573 "modules/window/sdl/Window.cpp"
 $59 = ((($dmode)) + 12|0); //@line 576 "modules/window/sdl/Window.cpp"
 $60 = HEAP32[$59>>2]|0; //@line 576 "modules/window/sdl/Window.cpp"
 $61 = (+($60|0)); //@line 576 "modules/window/sdl/Window.cpp"
 $62 = ((($this)) + 80|0); //@line 576 "modules/window/sdl/Window.cpp"
 HEAPF64[$62>>3] = $61; //@line 576 "modules/window/sdl/Window.cpp"
 STACKTOP = sp;return; //@line 577 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window9getWindowERiS3_RNS0_14WindowSettingsE($this,$width,$height,$newsettings) {
 $this = $this|0;
 $width = $width|0;
 $height = $height|0;
 $newsettings = $newsettings|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 582 "modules/window/sdl/Window.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 582 "modules/window/sdl/Window.cpp"
 $2 = ($1|0)==(0|0); //@line 582 "modules/window/sdl/Window.cpp"
 if (!($2)) {
  $3 = ((($this)) + 40|0); //@line 583 "modules/window/sdl/Window.cpp"
  __ZN4love6window3sdl6Window14updateSettingsERKNS0_14WindowSettingsE($this,$3); //@line 583 "modules/window/sdl/Window.cpp"
 }
 $4 = ((($this)) + 20|0); //@line 585 "modules/window/sdl/Window.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 585 "modules/window/sdl/Window.cpp"
 HEAP32[$width>>2] = $5; //@line 585 "modules/window/sdl/Window.cpp"
 $6 = ((($this)) + 24|0); //@line 586 "modules/window/sdl/Window.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 586 "modules/window/sdl/Window.cpp"
 HEAP32[$height>>2] = $7; //@line 586 "modules/window/sdl/Window.cpp"
 $8 = ((($this)) + 40|0); //@line 587 "modules/window/sdl/Window.cpp"
 dest=$newsettings; src=$8; stop=dest+60|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 587 "modules/window/sdl/Window.cpp"
 return; //@line 588 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window5closeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $1 = ($0|0)==(0|0); //@line 593 "modules/window/sdl/Window.cpp"
 if (!($1)) {
  $2 = HEAP32[$0>>2]|0; //@line 594 "modules/window/sdl/Window.cpp"
  $3 = ((($2)) + 32|0); //@line 594 "modules/window/sdl/Window.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 594 "modules/window/sdl/Window.cpp"
  FUNCTION_TABLE_vi[$4 & 8191]($0); //@line 594 "modules/window/sdl/Window.cpp"
 }
 $5 = ((($this)) + 116|0); //@line 596 "modules/window/sdl/Window.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 596 "modules/window/sdl/Window.cpp"
 $7 = ($6|0)==(0|0); //@line 596 "modules/window/sdl/Window.cpp"
 if (!($7)) {
  _SDL_GL_DeleteContext($6); //@line 598 "modules/window/sdl/Window.cpp"
  HEAP32[$5>>2] = 0; //@line 599 "modules/window/sdl/Window.cpp"
 }
 $8 = ((($this)) + 112|0); //@line 602 "modules/window/sdl/Window.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 602 "modules/window/sdl/Window.cpp"
 $10 = ($9|0)==(0|0); //@line 602 "modules/window/sdl/Window.cpp"
 if ($10) {
  $11 = ((($this)) + 108|0); //@line 612 "modules/window/sdl/Window.cpp"
  HEAP8[$11>>0] = 0; //@line 612 "modules/window/sdl/Window.cpp"
  return; //@line 613 "modules/window/sdl/Window.cpp"
 }
 _SDL_DestroyWindow($9); //@line 604 "modules/window/sdl/Window.cpp"
 HEAP32[$8>>2] = 0; //@line 605 "modules/window/sdl/Window.cpp"
 _SDL_FlushEvent(512); //@line 609 "modules/window/sdl/Window.cpp"
 $11 = ((($this)) + 108|0); //@line 612 "modules/window/sdl/Window.cpp"
 HEAP8[$11>>0] = 0; //@line 612 "modules/window/sdl/Window.cpp"
 return; //@line 613 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window13setFullscreenEbNS0_6Window14FullscreenTypeE($this,$fullscreen,$fstype) {
 $this = $this|0;
 $fullscreen = $fullscreen|0;
 $fstype = $fstype|0;
 var $$0 = 0, $$1 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $mode = 0, $newsettings = 0;
 var $sdlflags$0 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $newsettings = sp;
 $mode = sp + 64|0;
 $0 = ((($this)) + 112|0); //@line 617 "modules/window/sdl/Window.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 638 "modules/window/sdl/Window.cpp"
 $2 = ($1|0)==(0|0); //@line 617 "modules/window/sdl/Window.cpp"
 if ($2) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 665 "modules/window/sdl/Window.cpp"
 }
 $3 = ((($this)) + 40|0); //@line 620 "modules/window/sdl/Window.cpp"
 dest=$newsettings; src=$3; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 620 "modules/window/sdl/Window.cpp"
 $4 = $fullscreen&1; //@line 621 "modules/window/sdl/Window.cpp"
 HEAP8[$newsettings>>0] = $4; //@line 621 "modules/window/sdl/Window.cpp"
 $5 = ((($newsettings)) + 4|0); //@line 622 "modules/window/sdl/Window.cpp"
 HEAP32[$5>>2] = $fstype; //@line 622 "modules/window/sdl/Window.cpp"
 if ($fullscreen) {
  $6 = ($fstype|0)==(1); //@line 628 "modules/window/sdl/Window.cpp"
  if ($6) {
   $15 = $1;$sdlflags$0 = 4097;
  } else {
   ;HEAP32[$mode>>2]=0|0;HEAP32[$mode+4>>2]=0|0;HEAP32[$mode+8>>2]=0|0;HEAP32[$mode+12>>2]=0|0;HEAP32[$mode+16>>2]=0|0; //@line 634 "modules/window/sdl/Window.cpp"
   $7 = ((($this)) + 20|0); //@line 635 "modules/window/sdl/Window.cpp"
   $8 = HEAP32[$7>>2]|0; //@line 635 "modules/window/sdl/Window.cpp"
   $9 = ((($mode)) + 4|0); //@line 635 "modules/window/sdl/Window.cpp"
   HEAP32[$9>>2] = $8; //@line 635 "modules/window/sdl/Window.cpp"
   $10 = ((($this)) + 24|0); //@line 636 "modules/window/sdl/Window.cpp"
   $11 = HEAP32[$10>>2]|0; //@line 636 "modules/window/sdl/Window.cpp"
   $12 = ((($mode)) + 8|0); //@line 636 "modules/window/sdl/Window.cpp"
   HEAP32[$12>>2] = $11; //@line 636 "modules/window/sdl/Window.cpp"
   $13 = (_SDL_GetWindowDisplayIndex($1)|0); //@line 638 "modules/window/sdl/Window.cpp"
   (_SDL_GetClosestDisplayMode($13,$mode,$mode)|0); //@line 638 "modules/window/sdl/Window.cpp"
   $14 = HEAP32[$0>>2]|0; //@line 639 "modules/window/sdl/Window.cpp"
   (_SDL_SetWindowDisplayMode($14,$mode)|0); //@line 639 "modules/window/sdl/Window.cpp"
   $$pre = HEAP32[$0>>2]|0; //@line 647 "modules/window/sdl/Window.cpp"
   $15 = $$pre;$sdlflags$0 = 1;
  }
 } else {
  $15 = $1;$sdlflags$0 = 0;
 }
 $16 = (_SDL_SetWindowFullscreen($15,$sdlflags$0)|0); //@line 647 "modules/window/sdl/Window.cpp"
 $17 = ($16|0)==(0); //@line 647 "modules/window/sdl/Window.cpp"
 if ($17) {
  $18 = HEAP32[$0>>2]|0; //@line 649 "modules/window/sdl/Window.cpp"
  $19 = ((($this)) + 116|0); //@line 649 "modules/window/sdl/Window.cpp"
  $20 = HEAP32[$19>>2]|0; //@line 649 "modules/window/sdl/Window.cpp"
  (_SDL_GL_MakeCurrent($18,$20)|0); //@line 649 "modules/window/sdl/Window.cpp"
  __ZN4love6window3sdl6Window14updateSettingsERKNS0_14WindowSettingsE($this,$newsettings); //@line 650 "modules/window/sdl/Window.cpp"
  if (!($fullscreen)) {
   $21 = HEAP32[$0>>2]|0; //@line 654 "modules/window/sdl/Window.cpp"
   $22 = ((($this)) + 60|0); //@line 654 "modules/window/sdl/Window.cpp"
   $23 = HEAP32[$22>>2]|0; //@line 654 "modules/window/sdl/Window.cpp"
   $24 = ((($this)) + 64|0); //@line 654 "modules/window/sdl/Window.cpp"
   $25 = HEAP32[$24>>2]|0; //@line 654 "modules/window/sdl/Window.cpp"
   _SDL_SetWindowMinimumSize($21,$23,$25); //@line 654 "modules/window/sdl/Window.cpp"
  }
  $26 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
  $27 = ($26|0)==(0|0); //@line 658 "modules/window/sdl/Window.cpp"
  if ($27) {
   $$0 = 1;
  } else {
   $28 = HEAP32[$26>>2]|0; //@line 659 "modules/window/sdl/Window.cpp"
   $29 = ((($28)) + 24|0); //@line 659 "modules/window/sdl/Window.cpp"
   $30 = HEAP32[$29>>2]|0; //@line 659 "modules/window/sdl/Window.cpp"
   $31 = ((($this)) + 28|0); //@line 659 "modules/window/sdl/Window.cpp"
   $32 = HEAP32[$31>>2]|0; //@line 659 "modules/window/sdl/Window.cpp"
   $33 = ((($this)) + 32|0); //@line 659 "modules/window/sdl/Window.cpp"
   $34 = HEAP32[$33>>2]|0; //@line 659 "modules/window/sdl/Window.cpp"
   FUNCTION_TABLE_viii[$30 & 8191]($26,$32,$34); //@line 659 "modules/window/sdl/Window.cpp"
   $$0 = 1;
  }
 } else {
  $$0 = 0;
 }
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0); //@line 665 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window13setFullscreenEb($this,$fullscreen) {
 $this = $this|0;
 $fullscreen = $fullscreen|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 669 "modules/window/sdl/Window.cpp"
 $1 = ((($0)) + 36|0); //@line 669 "modules/window/sdl/Window.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 669 "modules/window/sdl/Window.cpp"
 $3 = ((($this)) + 44|0); //@line 669 "modules/window/sdl/Window.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 669 "modules/window/sdl/Window.cpp"
 $5 = (FUNCTION_TABLE_iiii[$2 & 8191]($this,$fullscreen,$4)|0); //@line 669 "modules/window/sdl/Window.cpp"
 return ($5|0); //@line 669 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window13onSizeChangedEii($this,$width,$height) {
 $this = $this|0;
 $width = $width|0;
 $height = $height|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 502 "modules/window/sdl/Window.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 508 "modules/window/sdl/Window.cpp"
 $2 = ($1|0)==(0|0); //@line 502 "modules/window/sdl/Window.cpp"
 if ($2) {
  $$0 = 0;
  return ($$0|0); //@line 515 "modules/window/sdl/Window.cpp"
 }
 $3 = ((($this)) + 20|0); //@line 505 "modules/window/sdl/Window.cpp"
 HEAP32[$3>>2] = $width; //@line 505 "modules/window/sdl/Window.cpp"
 $4 = ((($this)) + 24|0); //@line 506 "modules/window/sdl/Window.cpp"
 HEAP32[$4>>2] = $height; //@line 506 "modules/window/sdl/Window.cpp"
 $5 = ((($this)) + 28|0); //@line 508 "modules/window/sdl/Window.cpp"
 $6 = ((($this)) + 32|0); //@line 508 "modules/window/sdl/Window.cpp"
 _SDL_GL_GetDrawableSize($1,$5,$6); //@line 508 "modules/window/sdl/Window.cpp"
 $7 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $8 = ($7|0)==(0|0); //@line 511 "modules/window/sdl/Window.cpp"
 if ($8) {
  $$0 = 1;
  return ($$0|0); //@line 515 "modules/window/sdl/Window.cpp"
 }
 $9 = HEAP32[$7>>2]|0; //@line 512 "modules/window/sdl/Window.cpp"
 $10 = ((($9)) + 24|0); //@line 512 "modules/window/sdl/Window.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 512 "modules/window/sdl/Window.cpp"
 $12 = HEAP32[$5>>2]|0; //@line 512 "modules/window/sdl/Window.cpp"
 $13 = HEAP32[$6>>2]|0; //@line 512 "modules/window/sdl/Window.cpp"
 FUNCTION_TABLE_viii[$11 & 8191]($7,$12,$13); //@line 512 "modules/window/sdl/Window.cpp"
 $$0 = 1;
 return ($$0|0); //@line 515 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window15getDisplayCountEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GetNumVideoDisplays()|0); //@line 674 "modules/window/sdl/Window.cpp"
 return ($0|0); //@line 674 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window14getDisplayNameEi($this,$displayindex) {
 $this = $this|0;
 $displayindex = $displayindex|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_SDL_GetDisplayName($displayindex)|0); //@line 679 "modules/window/sdl/Window.cpp"
 $1 = ($0|0)==(0|0); //@line 681 "modules/window/sdl/Window.cpp"
 if (!($1)) {
  STACKTOP = sp;return ($0|0); //@line 684 "modules/window/sdl/Window.cpp"
 }
 $2 = (___cxa_allocate_exception(16)|0); //@line 682 "modules/window/sdl/Window.cpp"
 $3 = (($displayindex) + 1)|0; //@line 682 "modules/window/sdl/Window.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $3; //@line 682 "modules/window/sdl/Window.cpp"
 invoke_viii(3349,($2|0),(360698|0),($vararg_buffer|0)); //@line 682 "modules/window/sdl/Window.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  ___cxa_free_exception(($2|0)); //@line 682 "modules/window/sdl/Window.cpp"
  ___resumeException($6|0);
  // unreachable;
 } else {
  ___cxa_throw(($2|0),(96|0),(344|0)); //@line 682 "modules/window/sdl/Window.cpp"
  // unreachable; //@line 682 "modules/window/sdl/Window.cpp"
 }
 return (0)|0;
}
function __ZNK4love6window3sdl6Window18getFullscreenSizesEi($agg$result,$this,$displayindex) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $displayindex = $displayindex|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0, $mode = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, $w = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $mode = sp + 8|0;
 $w = sp;
 HEAP32[$agg$result>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $0 = ((($agg$result)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($agg$result)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($mode)) + 4|0); //@line 698 "modules/window/sdl/Window.cpp"
 $3 = ((($w)) + 4|0); //@line 698 "modules/window/sdl/Window.cpp"
 $4 = ((($mode)) + 8|0); //@line 698 "modules/window/sdl/Window.cpp"
 $5 = ((($agg$result)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $i$0 = 0;
 L1: while(1) {
  __THREW__ = 0;
  $6 = (invoke_ii(3807,($displayindex|0))|0); //@line 693 "modules/window/sdl/Window.cpp"
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if ($8) {
   break;
  }
  $9 = ($i$0|0)<($6|0); //@line 693 "modules/window/sdl/Window.cpp"
  if (!($9)) {
   label = 19;
   break;
  }
  ;HEAP32[$mode>>2]=0|0;HEAP32[$mode+4>>2]=0|0;HEAP32[$mode+8>>2]=0|0;HEAP32[$mode+12>>2]=0|0;HEAP32[$mode+16>>2]=0|0; //@line 695 "modules/window/sdl/Window.cpp"
  __THREW__ = 0;
  (invoke_iiii(3808,($displayindex|0),($i$0|0),($mode|0))|0); //@line 696 "modules/window/sdl/Window.cpp"
  $21 = __THREW__; __THREW__ = 0;
  $22 = $21&1;
  if ($22) {
   break;
  }
  $23 = HEAP32[$2>>2]|0; //@line 88 "./modules/window/Window.h"
  HEAP32[$w>>2] = $23; //@line 698 "modules/window/sdl/Window.cpp"
  $24 = HEAP32[$4>>2]|0; //@line 698 "modules/window/sdl/Window.cpp"
  HEAP32[$3>>2] = $24; //@line 698 "modules/window/sdl/Window.cpp"
  $25 = HEAP32[$agg$result>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $26 = HEAP32[$0>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $27 = ($25|0)==($26|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  L6: do {
   if ($27) {
    $37 = $25;
    label = 14;
   } else {
    $29 = $25;
    while(1) {
     $28 = HEAP32[$29>>2]|0; //@line 88 "./modules/window/Window.h"
     $30 = ($23|0)==($28|0); //@line 88 "./modules/window/Window.h"
     if ($30) {
      $31 = ((($29)) + 4|0); //@line 88 "./modules/window/Window.h"
      $32 = HEAP32[$31>>2]|0; //@line 88 "./modules/window/Window.h"
      $33 = ($24|0)==($32|0); //@line 88 "./modules/window/Window.h"
      if ($33) {
       $37 = $29;
       label = 14;
       break L6;
      }
     }
     $34 = ((($29)) + 8|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $35 = ($34|0)==($26|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($35) {
      label = 15;
      break;
     } else {
      $29 = $34;
     }
    }
   }
  } while(0);
  if ((label|0) == 14) {
   label = 0;
   $36 = ($37|0)==($26|0); //@line 1295 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($36) {
    label = 15;
   }
  }
  do {
   if ((label|0) == 15) {
    label = 0;
    $38 = HEAP32[$5>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $39 = ($26|0)==($38|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($39) {
     __THREW__ = 0;
     invoke_vii(3809,($agg$result|0),($w|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $52 = __THREW__; __THREW__ = 0;
     $53 = $52&1;
     if ($53) {
      break L1;
     } else {
      break;
     }
    } else {
     $40 = $w; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $41 = $40; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $42 = HEAP32[$41>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $43 = (($40) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $44 = $43; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $45 = HEAP32[$44>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $46 = $26; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $47 = $46; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$47>>2] = $42; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $48 = (($46) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $49 = $48; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$49>>2] = $45; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $50 = HEAP32[$0>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $51 = ((($50)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$0>>2] = $51; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     break;
    }
   }
  } while(0);
  $54 = (($i$0) + 1)|0; //@line 693 "modules/window/sdl/Window.cpp"
  $i$0 = $54;
 }
 if ((label|0) == 19) {
  STACKTOP = sp;return; //@line 707 "modules/window/sdl/Window.cpp"
 }
 $10 = ___cxa_find_matching_catch_2()|0;
 $11 = tempRet0;
 $12 = HEAP32[$agg$result>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ($12|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = $12; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($13) {
  ___resumeException($10|0);
  // unreachable;
 }
 $15 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = ($15|0)==($12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($16)) {
  $scevgep$i$i$i = ((($15)) + -8|0);
  $17 = $scevgep$i$i$i;
  $18 = (($17) - ($14))|0;
  $19 = $18 >>> 3;
  $20 = $19 ^ -1;
  $scevgep4$i$i$i = (($15) + ($20<<3)|0);
  HEAP32[$0>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($12); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($10|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love6window6Window10WindowSizeENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(536870911); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 3; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(268435455); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 2; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$42 = $18;
  } else {
   $$0$i1 = $17;$50 = $18;$51 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 536870911;$50 = $12;$51 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 3; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $51;$42 = $50;
 }
 $24 = (($25) + ($26<<3)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = $__x; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $31 = HEAP32[$30>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $32 = (($29) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = $32; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = HEAP32[$33>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = $24; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = $35; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$36>>2] = $31; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = (($35) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $37; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$38>>2] = $34; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $39 = ((($24)) + 8|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $40 = $39; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $41 = (($42) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $43 = $41 >> 3; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $44 = (0 - ($43))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $45 = (($24) + ($44<<3)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $46 = $45; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($45|0),($47|0),($41|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $46; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $40; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $48 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($48) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $49 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($49); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNK4love6window3sdl6Window20getDesktopDimensionsEiRiS3_($this,$displayindex,$width,$height) {
 $this = $this|0;
 $displayindex = $displayindex|0;
 $width = $width|0;
 $height = $height|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $mode = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $mode = sp;
 $0 = ($displayindex|0)>(-1); //@line 711 "modules/window/sdl/Window.cpp"
 if ($0) {
  $1 = HEAP32[$this>>2]|0; //@line 711 "modules/window/sdl/Window.cpp"
  $2 = ((($1)) + 48|0); //@line 711 "modules/window/sdl/Window.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 711 "modules/window/sdl/Window.cpp"
  $4 = (FUNCTION_TABLE_ii[$3 & 8191]($this)|0); //@line 711 "modules/window/sdl/Window.cpp"
  $5 = ($4|0)>($displayindex|0); //@line 711 "modules/window/sdl/Window.cpp"
  if ($5) {
   ;HEAP32[$mode>>2]=0|0;HEAP32[$mode+4>>2]=0|0;HEAP32[$mode+8>>2]=0|0;HEAP32[$mode+12>>2]=0|0;HEAP32[$mode+16>>2]=0|0; //@line 713 "modules/window/sdl/Window.cpp"
   (_SDL_GetDesktopDisplayMode($displayindex,$mode)|0); //@line 714 "modules/window/sdl/Window.cpp"
   $6 = ((($mode)) + 4|0); //@line 715 "modules/window/sdl/Window.cpp"
   $7 = HEAP32[$6>>2]|0; //@line 715 "modules/window/sdl/Window.cpp"
   HEAP32[$width>>2] = $7; //@line 715 "modules/window/sdl/Window.cpp"
   $8 = ((($mode)) + 8|0); //@line 716 "modules/window/sdl/Window.cpp"
   $9 = HEAP32[$8>>2]|0; //@line 716 "modules/window/sdl/Window.cpp"
   HEAP32[$height>>2] = $9; //@line 716 "modules/window/sdl/Window.cpp"
   STACKTOP = sp;return; //@line 723 "modules/window/sdl/Window.cpp"
  }
 }
 HEAP32[$width>>2] = 0; //@line 720 "modules/window/sdl/Window.cpp"
 HEAP32[$height>>2] = 0; //@line 721 "modules/window/sdl/Window.cpp"
 STACKTOP = sp;return; //@line 723 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window11setPositionEiii($this,$x,$y,$displayindex) {
 $this = $this|0;
 $x = $x|0;
 $y = $y|0;
 $displayindex = $displayindex|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $displaybounds = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $displaybounds = sp;
 $0 = ((($this)) + 112|0); //@line 727 "modules/window/sdl/Window.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 727 "modules/window/sdl/Window.cpp"
 $2 = ($1|0)==(0|0); //@line 727 "modules/window/sdl/Window.cpp"
 if ($2) {
  STACKTOP = sp;return; //@line 742 "modules/window/sdl/Window.cpp"
 }
 $3 = ($displayindex|0)<(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $4 = HEAP32[$this>>2]|0; //@line 730 "modules/window/sdl/Window.cpp"
 $5 = ((($4)) + 48|0); //@line 730 "modules/window/sdl/Window.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 730 "modules/window/sdl/Window.cpp"
 $7 = (FUNCTION_TABLE_ii[$6 & 8191]($this)|0); //@line 730 "modules/window/sdl/Window.cpp"
 $8 = (($7) + -1)|0; //@line 730 "modules/window/sdl/Window.cpp"
 $9 = $3 ? 0 : $displayindex; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $10 = ($8|0)<($9|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $11 = $10 ? $8 : $9; //@line 730 "modules/window/sdl/Window.cpp"
 ;HEAP32[$displaybounds>>2]=0|0;HEAP32[$displaybounds+4>>2]=0|0;HEAP32[$displaybounds+8>>2]=0|0;HEAP32[$displaybounds+12>>2]=0|0; //@line 732 "modules/window/sdl/Window.cpp"
 (_SDL_GetDisplayBounds($11,$displaybounds)|0); //@line 733 "modules/window/sdl/Window.cpp"
 $12 = HEAP32[$displaybounds>>2]|0; //@line 736 "modules/window/sdl/Window.cpp"
 $13 = (($12) + ($x))|0; //@line 736 "modules/window/sdl/Window.cpp"
 $14 = ((($displaybounds)) + 4|0); //@line 737 "modules/window/sdl/Window.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 737 "modules/window/sdl/Window.cpp"
 $16 = (($15) + ($y))|0; //@line 737 "modules/window/sdl/Window.cpp"
 $17 = HEAP32[$0>>2]|0; //@line 739 "modules/window/sdl/Window.cpp"
 _SDL_SetWindowPosition($17,$13,$16); //@line 739 "modules/window/sdl/Window.cpp"
 $18 = ((($this)) + 88|0); //@line 741 "modules/window/sdl/Window.cpp"
 HEAP8[$18>>0] = 1; //@line 741 "modules/window/sdl/Window.cpp"
 STACKTOP = sp;return; //@line 742 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window11getPositionERiS3_S3_($this,$x,$y,$displayindex) {
 $this = $this|0;
 $x = $x|0;
 $y = $y|0;
 $displayindex = $displayindex|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $displaybounds = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $displaybounds = sp;
 $0 = ((($this)) + 112|0); //@line 746 "modules/window/sdl/Window.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 746 "modules/window/sdl/Window.cpp"
 $2 = ($1|0)==(0|0); //@line 746 "modules/window/sdl/Window.cpp"
 if ($2) {
  HEAP32[$y>>2] = 0; //@line 748 "modules/window/sdl/Window.cpp"
  HEAP32[$x>>2] = 0; //@line 748 "modules/window/sdl/Window.cpp"
  HEAP32[$displayindex>>2] = 0; //@line 749 "modules/window/sdl/Window.cpp"
  STACKTOP = sp;return; //@line 768 "modules/window/sdl/Window.cpp"
 }
 $3 = (_SDL_GetWindowDisplayIndex($1)|0); //@line 753 "modules/window/sdl/Window.cpp"
 $4 = ($3|0)<(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $5 = $4 ? 0 : $3; //@line 753 "modules/window/sdl/Window.cpp"
 HEAP32[$displayindex>>2] = $5; //@line 753 "modules/window/sdl/Window.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 755 "modules/window/sdl/Window.cpp"
 _SDL_GetWindowPosition($6,$x,$y); //@line 755 "modules/window/sdl/Window.cpp"
 $7 = HEAP32[$x>>2]|0; //@line 760 "modules/window/sdl/Window.cpp"
 $8 = HEAP32[$y>>2]|0; //@line 760 "modules/window/sdl/Window.cpp"
 $9 = $8 | $7; //@line 760 "modules/window/sdl/Window.cpp"
 $10 = ($9|0)==(0); //@line 760 "modules/window/sdl/Window.cpp"
 if ($10) {
  STACKTOP = sp;return; //@line 768 "modules/window/sdl/Window.cpp"
 }
 ;HEAP32[$displaybounds>>2]=0|0;HEAP32[$displaybounds+4>>2]=0|0;HEAP32[$displaybounds+8>>2]=0|0;HEAP32[$displaybounds+12>>2]=0|0; //@line 762 "modules/window/sdl/Window.cpp"
 $11 = HEAP32[$displayindex>>2]|0; //@line 763 "modules/window/sdl/Window.cpp"
 (_SDL_GetDisplayBounds($11,$displaybounds)|0); //@line 763 "modules/window/sdl/Window.cpp"
 $12 = HEAP32[$displaybounds>>2]|0; //@line 765 "modules/window/sdl/Window.cpp"
 $13 = HEAP32[$x>>2]|0; //@line 765 "modules/window/sdl/Window.cpp"
 $14 = (($13) - ($12))|0; //@line 765 "modules/window/sdl/Window.cpp"
 HEAP32[$x>>2] = $14; //@line 765 "modules/window/sdl/Window.cpp"
 $15 = ((($displaybounds)) + 4|0); //@line 766 "modules/window/sdl/Window.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 766 "modules/window/sdl/Window.cpp"
 $17 = HEAP32[$y>>2]|0; //@line 766 "modules/window/sdl/Window.cpp"
 $18 = (($17) - ($16))|0; //@line 766 "modules/window/sdl/Window.cpp"
 HEAP32[$y>>2] = $18; //@line 766 "modules/window/sdl/Window.cpp"
 STACKTOP = sp;return; //@line 768 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window6isOpenEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 108|0); //@line 772 "modules/window/sdl/Window.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 772 "modules/window/sdl/Window.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 772 "modules/window/sdl/Window.cpp"
 return ($2|0); //@line 772 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window14setWindowTitleERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($this,$title) {
 $this = $this|0;
 $title = $title|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 777 "modules/window/sdl/Window.cpp"
 (__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEaSERKS5_($0,$title)|0); //@line 777 "modules/window/sdl/Window.cpp"
 $1 = ((($this)) + 112|0); //@line 779 "modules/window/sdl/Window.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 779 "modules/window/sdl/Window.cpp"
 $3 = ($2|0)==(0|0); //@line 779 "modules/window/sdl/Window.cpp"
 if ($3) {
  return; //@line 781 "modules/window/sdl/Window.cpp"
 }
 $4 = HEAP8[$title>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = $4 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = ($5<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($title)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = HEAP32[$7>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = ((($title)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $6 ? $9 : $8; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 _SDL_SetWindowTitle($2,$10); //@line 780 "modules/window/sdl/Window.cpp"
 return; //@line 781 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window14getWindowTitleEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 785 "modules/window/sdl/Window.cpp"
 return ($0|0); //@line 785 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window7setIconEPNS_5image9ImageDataE($this,$imgd) {
 $this = $this|0;
 $imgd = $imgd|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lock = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ($imgd|0)==(0|0); //@line 790 "modules/window/sdl/Window.cpp"
 if ($0) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 830 "modules/window/sdl/Window.cpp"
 }
 $1 = HEAP32[$imgd>>2]|0; //@line 132 "./common/Object.h"
 $2 = ((($1)) + 8|0); //@line 132 "./common/Object.h"
 $3 = HEAP32[$2>>2]|0; //@line 132 "./common/Object.h"
 FUNCTION_TABLE_vi[$3 & 8191]($imgd); //@line 132 "./common/Object.h"
 $4 = ((($this)) + 104|0); //@line 133 "./common/Object.h"
 $5 = HEAP32[$4>>2]|0; //@line 133 "./common/Object.h"
 $6 = ($5|0)==(0|0); //@line 133 "./common/Object.h"
 if (!($6)) {
  $7 = HEAP32[$5>>2]|0; //@line 133 "./common/Object.h"
  $8 = ((($7)) + 12|0); //@line 133 "./common/Object.h"
  $9 = HEAP32[$8>>2]|0; //@line 133 "./common/Object.h"
  FUNCTION_TABLE_vi[$9 & 8191]($5); //@line 133 "./common/Object.h"
 }
 HEAP32[$4>>2] = $imgd; //@line 134 "./common/Object.h"
 $10 = ((($this)) + 112|0); //@line 795 "modules/window/sdl/Window.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 795 "modules/window/sdl/Window.cpp"
 $12 = ($11|0)==(0|0); //@line 795 "modules/window/sdl/Window.cpp"
 if ($12) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 830 "modules/window/sdl/Window.cpp"
 }
 $13 = (__ZNK4love5image9ImageData8getWidthEv($imgd)|0); //@line 811 "modules/window/sdl/Window.cpp"
 $14 = (__ZNK4love5image9ImageData9getHeightEv($imgd)|0); //@line 812 "modules/window/sdl/Window.cpp"
 $15 = (__ZNK4love5image9ImageData8getWidthEv($imgd)|0); //@line 813 "modules/window/sdl/Window.cpp"
 $16 = (__ZNK4love5image9ImageData8getMutexEv($imgd)|0); //@line 819 "modules/window/sdl/Window.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$16); //@line 819 "modules/window/sdl/Window.cpp"
 $17 = HEAP32[$imgd>>2]|0; //@line 820 "modules/window/sdl/Window.cpp"
 $18 = ((($17)) + 16|0); //@line 820 "modules/window/sdl/Window.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 820 "modules/window/sdl/Window.cpp"
 __THREW__ = 0;
 $20 = (invoke_ii($19|0,($imgd|0))|0); //@line 820 "modules/window/sdl/Window.cpp"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $28 = ___cxa_find_matching_catch_2()|0;
  $29 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 821 "modules/window/sdl/Window.cpp"
  ___resumeException($28|0);
  // unreachable;
 }
 $23 = $15 << 2; //@line 813 "modules/window/sdl/Window.cpp"
 __THREW__ = 0;
 $24 = (invoke_iiiiiiiiii(3810,($20|0),($13|0),($14|0),32,($23|0),255,65280,16711680,-16777216)|0); //@line 820 "modules/window/sdl/Window.cpp"
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if ($26) {
  $28 = ___cxa_find_matching_catch_2()|0;
  $29 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 821 "modules/window/sdl/Window.cpp"
  ___resumeException($28|0);
  // unreachable;
 }
 __ZN4love6thread4LockD2Ev($lock); //@line 821 "modules/window/sdl/Window.cpp"
 $27 = ($24|0)==(0|0); //@line 823 "modules/window/sdl/Window.cpp"
 if ($27) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 830 "modules/window/sdl/Window.cpp"
 }
 $30 = HEAP32[$10>>2]|0; //@line 826 "modules/window/sdl/Window.cpp"
 _SDL_SetWindowIcon($30,$24); //@line 826 "modules/window/sdl/Window.cpp"
 _SDL_FreeSurface($24); //@line 827 "modules/window/sdl/Window.cpp"
 $$1 = 1;
 STACKTOP = sp;return ($$1|0); //@line 830 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window7getIconEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 104|0); //@line 139 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 139 "./common/Object.h"
 return ($1|0); //@line 834 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window22setDisplaySleepEnabledEb($this,$enable) {
 $this = $this|0;
 $enable = $enable|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 if ($enable) {
  _SDL_EnableScreenSaver(); //@line 840 "modules/window/sdl/Window.cpp"
 } else {
  _SDL_DisableScreenSaver(); //@line 842 "modules/window/sdl/Window.cpp"
 }
 return; //@line 843 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window21isDisplaySleepEnabledEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_IsScreenSaverEnabled()|0); //@line 847 "modules/window/sdl/Window.cpp"
 $1 = ($0|0)!=(0); //@line 847 "modules/window/sdl/Window.cpp"
 return ($1|0); //@line 847 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window8minimizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 852 "modules/window/sdl/Window.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 852 "modules/window/sdl/Window.cpp"
 $2 = ($1|0)==(0|0); //@line 852 "modules/window/sdl/Window.cpp"
 if ($2) {
  return; //@line 854 "modules/window/sdl/Window.cpp"
 }
 _SDL_MinimizeWindow($1); //@line 853 "modules/window/sdl/Window.cpp"
 return; //@line 854 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window8maximizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 858 "modules/window/sdl/Window.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 858 "modules/window/sdl/Window.cpp"
 $2 = ($1|0)==(0|0); //@line 858 "modules/window/sdl/Window.cpp"
 if ($2) {
  return; //@line 860 "modules/window/sdl/Window.cpp"
 }
 _SDL_MaximizeWindow($1); //@line 859 "modules/window/sdl/Window.cpp"
 return; //@line 860 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window11swapBuffersEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 864 "modules/window/sdl/Window.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 864 "modules/window/sdl/Window.cpp"
 _SDL_GL_SwapWindow($1); //@line 864 "modules/window/sdl/Window.cpp"
 return; //@line 865 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window8hasFocusEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 869 "modules/window/sdl/Window.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 869 "modules/window/sdl/Window.cpp"
 $2 = ($1|0)==(0|0); //@line 869 "modules/window/sdl/Window.cpp"
 if ($2) {
  $6 = 0;
  return ($6|0); //@line 869 "modules/window/sdl/Window.cpp"
 }
 $3 = (_SDL_GetKeyboardFocus()|0); //@line 869 "modules/window/sdl/Window.cpp"
 $4 = HEAP32[$0>>2]|0; //@line 869 "modules/window/sdl/Window.cpp"
 $5 = ($3|0)==($4|0); //@line 869 "modules/window/sdl/Window.cpp"
 $6 = $5;
 return ($6|0); //@line 869 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window13hasMouseFocusEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 874 "modules/window/sdl/Window.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 874 "modules/window/sdl/Window.cpp"
 $2 = ($1|0)==(0|0); //@line 874 "modules/window/sdl/Window.cpp"
 if ($2) {
  $6 = 0;
  return ($6|0); //@line 874 "modules/window/sdl/Window.cpp"
 }
 $3 = (_SDL_GetMouseFocus()|0); //@line 874 "modules/window/sdl/Window.cpp"
 $4 = HEAP32[$0>>2]|0; //@line 874 "modules/window/sdl/Window.cpp"
 $5 = ($3|0)==($4|0); //@line 874 "modules/window/sdl/Window.cpp"
 $6 = $5;
 return ($6|0); //@line 874 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window9isVisibleEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 879 "modules/window/sdl/Window.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 879 "modules/window/sdl/Window.cpp"
 $2 = ($1|0)==(0|0); //@line 879 "modules/window/sdl/Window.cpp"
 if ($2) {
  $6 = 0;
  return ($6|0); //@line 879 "modules/window/sdl/Window.cpp"
 }
 $3 = (_SDL_GetWindowFlags($1)|0); //@line 879 "modules/window/sdl/Window.cpp"
 $4 = $3 & 4; //@line 879 "modules/window/sdl/Window.cpp"
 $5 = ($4|0)!=(0); //@line 879 "modules/window/sdl/Window.cpp"
 $6 = $5;
 return ($6|0); //@line 879 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window15setMouseVisibleEb($this,$visible) {
 $this = $this|0;
 $visible = $visible|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $visible&1; //@line 884 "modules/window/sdl/Window.cpp"
 (_SDL_ShowCursor($0)|0); //@line 884 "modules/window/sdl/Window.cpp"
 return; //@line 885 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window15getMouseVisibleEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_ShowCursor(-1)|0); //@line 889 "modules/window/sdl/Window.cpp"
 $1 = ($0|0)==(1); //@line 889 "modules/window/sdl/Window.cpp"
 return ($1|0); //@line 889 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window12setMouseGrabEb($this,$grab) {
 $this = $this|0;
 $grab = $grab|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 109|0); //@line 894 "modules/window/sdl/Window.cpp"
 $1 = $grab&1; //@line 894 "modules/window/sdl/Window.cpp"
 HEAP8[$0>>0] = $1; //@line 894 "modules/window/sdl/Window.cpp"
 $2 = ((($this)) + 112|0); //@line 895 "modules/window/sdl/Window.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 895 "modules/window/sdl/Window.cpp"
 $4 = ($3|0)==(0|0); //@line 895 "modules/window/sdl/Window.cpp"
 if ($4) {
  return; //@line 897 "modules/window/sdl/Window.cpp"
 }
 $5 = $grab&1; //@line 896 "modules/window/sdl/Window.cpp"
 _SDL_SetWindowGrab($3,$5); //@line 896 "modules/window/sdl/Window.cpp"
 return; //@line 897 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window14isMouseGrabbedEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 901 "modules/window/sdl/Window.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 901 "modules/window/sdl/Window.cpp"
 $2 = ($1|0)==(0|0); //@line 901 "modules/window/sdl/Window.cpp"
 if ($2) {
  $5 = ((($this)) + 109|0); //@line 904 "modules/window/sdl/Window.cpp"
  $6 = HEAP8[$5>>0]|0; //@line 904 "modules/window/sdl/Window.cpp"
  $7 = ($6<<24>>24)!=(0); //@line 904 "modules/window/sdl/Window.cpp"
  $$0 = $7;
  return ($$0|0); //@line 905 "modules/window/sdl/Window.cpp"
 } else {
  $3 = (_SDL_GetWindowGrab($1)|0); //@line 902 "modules/window/sdl/Window.cpp"
  $4 = ($3|0)!=(0); //@line 902 "modules/window/sdl/Window.cpp"
  $$0 = $4;
  return ($$0|0); //@line 905 "modules/window/sdl/Window.cpp"
 }
 return (0)|0;
}
function __ZNK4love6window3sdl6Window18getPixelDimensionsERiS3_($this,$w,$h) {
 $this = $this|0;
 $w = $w|0;
 $h = $h|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 909 "modules/window/sdl/Window.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 909 "modules/window/sdl/Window.cpp"
 HEAP32[$w>>2] = $1; //@line 909 "modules/window/sdl/Window.cpp"
 $2 = ((($this)) + 32|0); //@line 910 "modules/window/sdl/Window.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 910 "modules/window/sdl/Window.cpp"
 HEAP32[$h>>2] = $3; //@line 910 "modules/window/sdl/Window.cpp"
 return; //@line 911 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window19windowToPixelCoordsEPdS3_($this,$x,$y) {
 $this = $this|0;
 $x = $x|0;
 $y = $y|0;
 var $0 = 0, $1 = 0.0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0.0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($x|0)==(0|0); //@line 915 "modules/window/sdl/Window.cpp"
 if (!($0)) {
  $1 = +HEAPF64[$x>>3]; //@line 916 "modules/window/sdl/Window.cpp"
  $2 = ((($this)) + 28|0); //@line 916 "modules/window/sdl/Window.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 916 "modules/window/sdl/Window.cpp"
  $4 = (+($3|0)); //@line 916 "modules/window/sdl/Window.cpp"
  $5 = ((($this)) + 20|0); //@line 916 "modules/window/sdl/Window.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 916 "modules/window/sdl/Window.cpp"
  $7 = (+($6|0)); //@line 916 "modules/window/sdl/Window.cpp"
  $8 = $4 / $7; //@line 916 "modules/window/sdl/Window.cpp"
  $9 = $1 * $8; //@line 916 "modules/window/sdl/Window.cpp"
  HEAPF64[$x>>3] = $9; //@line 916 "modules/window/sdl/Window.cpp"
 }
 $10 = ($y|0)==(0|0); //@line 917 "modules/window/sdl/Window.cpp"
 if ($10) {
  return; //@line 919 "modules/window/sdl/Window.cpp"
 }
 $11 = +HEAPF64[$y>>3]; //@line 918 "modules/window/sdl/Window.cpp"
 $12 = ((($this)) + 32|0); //@line 918 "modules/window/sdl/Window.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 918 "modules/window/sdl/Window.cpp"
 $14 = (+($13|0)); //@line 918 "modules/window/sdl/Window.cpp"
 $15 = ((($this)) + 24|0); //@line 918 "modules/window/sdl/Window.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 918 "modules/window/sdl/Window.cpp"
 $17 = (+($16|0)); //@line 918 "modules/window/sdl/Window.cpp"
 $18 = $14 / $17; //@line 918 "modules/window/sdl/Window.cpp"
 $19 = $11 * $18; //@line 918 "modules/window/sdl/Window.cpp"
 HEAPF64[$y>>3] = $19; //@line 918 "modules/window/sdl/Window.cpp"
 return; //@line 919 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window19pixelToWindowCoordsEPdS3_($this,$x,$y) {
 $this = $this|0;
 $x = $x|0;
 $y = $y|0;
 var $0 = 0, $1 = 0.0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0.0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($x|0)==(0|0); //@line 923 "modules/window/sdl/Window.cpp"
 if (!($0)) {
  $1 = +HEAPF64[$x>>3]; //@line 924 "modules/window/sdl/Window.cpp"
  $2 = ((($this)) + 20|0); //@line 924 "modules/window/sdl/Window.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 924 "modules/window/sdl/Window.cpp"
  $4 = (+($3|0)); //@line 924 "modules/window/sdl/Window.cpp"
  $5 = ((($this)) + 28|0); //@line 924 "modules/window/sdl/Window.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 924 "modules/window/sdl/Window.cpp"
  $7 = (+($6|0)); //@line 924 "modules/window/sdl/Window.cpp"
  $8 = $4 / $7; //@line 924 "modules/window/sdl/Window.cpp"
  $9 = $1 * $8; //@line 924 "modules/window/sdl/Window.cpp"
  HEAPF64[$x>>3] = $9; //@line 924 "modules/window/sdl/Window.cpp"
 }
 $10 = ($y|0)==(0|0); //@line 925 "modules/window/sdl/Window.cpp"
 if ($10) {
  return; //@line 927 "modules/window/sdl/Window.cpp"
 }
 $11 = +HEAPF64[$y>>3]; //@line 926 "modules/window/sdl/Window.cpp"
 $12 = ((($this)) + 24|0); //@line 926 "modules/window/sdl/Window.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 926 "modules/window/sdl/Window.cpp"
 $14 = (+($13|0)); //@line 926 "modules/window/sdl/Window.cpp"
 $15 = ((($this)) + 32|0); //@line 926 "modules/window/sdl/Window.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 926 "modules/window/sdl/Window.cpp"
 $17 = (+($16|0)); //@line 926 "modules/window/sdl/Window.cpp"
 $18 = $14 / $17; //@line 926 "modules/window/sdl/Window.cpp"
 $19 = $11 * $18; //@line 926 "modules/window/sdl/Window.cpp"
 HEAPF64[$y>>3] = $19; //@line 926 "modules/window/sdl/Window.cpp"
 return; //@line 927 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window13getPixelScaleEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 934 "modules/window/sdl/Window.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 934 "modules/window/sdl/Window.cpp"
 $2 = (+($1|0)); //@line 934 "modules/window/sdl/Window.cpp"
 $3 = ((($this)) + 24|0); //@line 934 "modules/window/sdl/Window.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 934 "modules/window/sdl/Window.cpp"
 $5 = (+($4|0)); //@line 934 "modules/window/sdl/Window.cpp"
 $6 = $2 / $5; //@line 934 "modules/window/sdl/Window.cpp"
 return (+$6); //@line 934 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window8toPixelsEd($this,$x) {
 $this = $this|0;
 $x = +$x;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 940 "modules/window/sdl/Window.cpp"
 $1 = ((($0)) + 152|0); //@line 940 "modules/window/sdl/Window.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 940 "modules/window/sdl/Window.cpp"
 $3 = (+FUNCTION_TABLE_di[$2 & 4095]($this)); //@line 940 "modules/window/sdl/Window.cpp"
 $4 = $3 * $x; //@line 940 "modules/window/sdl/Window.cpp"
 return (+$4); //@line 940 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window8toPixelsEddRdS3_($this,$wx,$wy,$px,$py) {
 $this = $this|0;
 $wx = +$wx;
 $wy = +$wy;
 $px = $px|0;
 $py = $py|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 945 "modules/window/sdl/Window.cpp"
 $1 = ((($0)) + 152|0); //@line 945 "modules/window/sdl/Window.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 945 "modules/window/sdl/Window.cpp"
 $3 = (+FUNCTION_TABLE_di[$2 & 4095]($this)); //@line 945 "modules/window/sdl/Window.cpp"
 $4 = $3 * $wx; //@line 946 "modules/window/sdl/Window.cpp"
 HEAPF64[$px>>3] = $4; //@line 946 "modules/window/sdl/Window.cpp"
 $5 = $3 * $wy; //@line 947 "modules/window/sdl/Window.cpp"
 HEAPF64[$py>>3] = $5; //@line 947 "modules/window/sdl/Window.cpp"
 return; //@line 948 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window10fromPixelsEd($this,$x) {
 $this = $this|0;
 $x = +$x;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 952 "modules/window/sdl/Window.cpp"
 $1 = ((($0)) + 152|0); //@line 952 "modules/window/sdl/Window.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 952 "modules/window/sdl/Window.cpp"
 $3 = (+FUNCTION_TABLE_di[$2 & 4095]($this)); //@line 952 "modules/window/sdl/Window.cpp"
 $4 = $x / $3; //@line 952 "modules/window/sdl/Window.cpp"
 return (+$4); //@line 952 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window10fromPixelsEddRdS3_($this,$px,$py,$wx,$wy) {
 $this = $this|0;
 $px = +$px;
 $py = +$py;
 $wx = $wx|0;
 $wy = $wy|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 957 "modules/window/sdl/Window.cpp"
 $1 = ((($0)) + 152|0); //@line 957 "modules/window/sdl/Window.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 957 "modules/window/sdl/Window.cpp"
 $3 = (+FUNCTION_TABLE_di[$2 & 4095]($this)); //@line 957 "modules/window/sdl/Window.cpp"
 $4 = $px / $3; //@line 958 "modules/window/sdl/Window.cpp"
 HEAPF64[$wx>>3] = $4; //@line 958 "modules/window/sdl/Window.cpp"
 $5 = $py / $3; //@line 959 "modules/window/sdl/Window.cpp"
 HEAPF64[$wy>>3] = $5; //@line 959 "modules/window/sdl/Window.cpp"
 return; //@line 960 "modules/window/sdl/Window.cpp"
}
function __ZNK4love6window3sdl6Window9getHandleEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 964 "modules/window/sdl/Window.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 964 "modules/window/sdl/Window.cpp"
 return ($1|0); //@line 964 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window14showMessageBoxERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEESB_NS0_6Window14MessageBoxTypeEb($this,$title,$message,$type,$attachtowindow) {
 $this = $this|0;
 $title = $title|0;
 $message = $message|0;
 $type = $type|0;
 $attachtowindow = $attachtowindow|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $switch$select$i = 0;
 var $switch$select2$i = 0, $switch$selectcmp$i = 0, $switch$selectcmp1$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $switch$selectcmp$i = ($type|0)==(1); //@line 969 "modules/window/sdl/Window.cpp"
 $switch$select$i = $switch$selectcmp$i ? 32 : 64; //@line 969 "modules/window/sdl/Window.cpp"
 $switch$selectcmp1$i = ($type|0)==(0); //@line 969 "modules/window/sdl/Window.cpp"
 $switch$select2$i = $switch$selectcmp1$i ? 16 : $switch$select$i; //@line 969 "modules/window/sdl/Window.cpp"
 if ($attachtowindow) {
  $0 = ((($this)) + 112|0); //@line 984 "modules/window/sdl/Window.cpp"
  $1 = HEAP32[$0>>2]|0; //@line 984 "modules/window/sdl/Window.cpp"
  $16 = $1;
 } else {
  $16 = 0;
 }
 $2 = HEAP8[$title>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = $2 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ($3<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($title)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = HEAP32[$5>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($title)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $4 ? $7 : $6; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = HEAP8[$message>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $9 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = ($10<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = ((($message)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = HEAP32[$12>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = ((($message)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $15 = $11 ? $14 : $13; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $17 = (_SDL_ShowSimpleMessageBox($switch$select2$i,$8,$15,$16)|0); //@line 986 "modules/window/sdl/Window.cpp"
 $18 = ($17|0)>(-1); //@line 986 "modules/window/sdl/Window.cpp"
 return ($18|0); //@line 986 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6Window14showMessageBoxERKNS0_6Window14MessageBoxDataE($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $i$08 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0;
 var $pressedbutton = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i2 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i3 = 0, $sdlbutton = 0, $sdlbuttons = 0, $sdldata = 0, $switch$select$i = 0, $switch$select2$i = 0, $switch$selectcmp$i = 0, $switch$selectcmp1$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $sdldata = sp + 28|0;
 $sdlbuttons = sp + 16|0;
 $sdlbutton = sp + 4|0;
 $pressedbutton = sp;
 ;HEAP32[$sdldata>>2]=0|0;HEAP32[$sdldata+4>>2]=0|0;HEAP32[$sdldata+8>>2]=0|0;HEAP32[$sdldata+12>>2]=0|0;HEAP32[$sdldata+16>>2]=0|0;HEAP32[$sdldata+20>>2]=0|0;HEAP32[$sdldata+24>>2]=0|0; //@line 991 "modules/window/sdl/Window.cpp"
 $0 = HEAP32[$data>>2]|0; //@line 993 "modules/window/sdl/Window.cpp"
 $switch$selectcmp$i = ($0|0)==(1); //@line 969 "modules/window/sdl/Window.cpp"
 $switch$select$i = $switch$selectcmp$i ? 32 : 64; //@line 969 "modules/window/sdl/Window.cpp"
 $switch$selectcmp1$i = ($0|0)==(0); //@line 969 "modules/window/sdl/Window.cpp"
 $switch$select2$i = $switch$selectcmp1$i ? 16 : $switch$select$i; //@line 969 "modules/window/sdl/Window.cpp"
 HEAP32[$sdldata>>2] = $switch$select2$i; //@line 993 "modules/window/sdl/Window.cpp"
 $1 = ((($data)) + 4|0); //@line 994 "modules/window/sdl/Window.cpp"
 $2 = HEAP8[$1>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = $2 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ($3<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($4) {
  $7 = ((($1)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = $7;
 } else {
  $5 = ((($data)) + 12|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $6 = HEAP32[$5>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = $6;
 }
 $8 = ((($sdldata)) + 8|0); //@line 994 "modules/window/sdl/Window.cpp"
 HEAP32[$8>>2] = $9; //@line 994 "modules/window/sdl/Window.cpp"
 $10 = ((($data)) + 16|0); //@line 995 "modules/window/sdl/Window.cpp"
 $11 = HEAP8[$10>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = $11 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ($12<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($13) {
  $16 = ((($10)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $18 = $16;
 } else {
  $14 = ((($data)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $15 = HEAP32[$14>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $18 = $15;
 }
 $17 = ((($sdldata)) + 12|0); //@line 995 "modules/window/sdl/Window.cpp"
 HEAP32[$17>>2] = $18; //@line 995 "modules/window/sdl/Window.cpp"
 $19 = ((($data)) + 48|0); //@line 996 "modules/window/sdl/Window.cpp"
 $20 = HEAP8[$19>>0]|0; //@line 996 "modules/window/sdl/Window.cpp"
 $21 = ($20<<24>>24)==(0); //@line 996 "modules/window/sdl/Window.cpp"
 if ($21) {
  $25 = 0;
 } else {
  $22 = ((($this)) + 112|0); //@line 996 "modules/window/sdl/Window.cpp"
  $23 = HEAP32[$22>>2]|0; //@line 996 "modules/window/sdl/Window.cpp"
  $25 = $23;
 }
 $24 = ((($sdldata)) + 4|0); //@line 996 "modules/window/sdl/Window.cpp"
 HEAP32[$24>>2] = $25; //@line 996 "modules/window/sdl/Window.cpp"
 $26 = ((($data)) + 28|0); //@line 998 "modules/window/sdl/Window.cpp"
 $27 = ((($data)) + 32|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $28 = HEAP32[$27>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $29 = HEAP32[$26>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = (($28) - ($29))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = (($30|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = ((($sdldata)) + 16|0); //@line 998 "modules/window/sdl/Window.cpp"
 HEAP32[$32>>2] = $31; //@line 998 "modules/window/sdl/Window.cpp"
 HEAP32[$sdlbuttons>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $33 = ((($sdlbuttons)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$33>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $34 = ((($sdlbuttons)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$34>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = ($30|0)>(0); //@line 1002 "modules/window/sdl/Window.cpp"
 do {
  if ($35) {
   $36 = ((($sdlbutton)) + 4|0); //@line 1006 "modules/window/sdl/Window.cpp"
   $37 = ((($sdlbutton)) + 8|0); //@line 1007 "modules/window/sdl/Window.cpp"
   $38 = ((($data)) + 40|0); //@line 1009 "modules/window/sdl/Window.cpp"
   $39 = ((($data)) + 44|0); //@line 1012 "modules/window/sdl/Window.cpp"
   $40 = ((($sdlbuttons)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $$in = $29;$i$08 = 0;
   while(1) {
    $45 = $$in; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    ;HEAP32[$sdlbutton>>2]=0|0;HEAP32[$sdlbutton+4>>2]=0|0;HEAP32[$sdlbutton+8>>2]=0|0; //@line 1004 "modules/window/sdl/Window.cpp"
    HEAP32[$36>>2] = $i$08; //@line 1006 "modules/window/sdl/Window.cpp"
    $46 = (($45) + (($i$08*12)|0)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $47 = HEAP8[$46>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $48 = $47 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $49 = ($48<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if ($49) {
     $52 = ((($46)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $53 = $52;
    } else {
     $50 = (((($45) + (($i$08*12)|0)|0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $51 = HEAP32[$50>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $53 = $51;
    }
    HEAP32[$37>>2] = $53; //@line 1007 "modules/window/sdl/Window.cpp"
    $54 = HEAP32[$38>>2]|0; //@line 1009 "modules/window/sdl/Window.cpp"
    $55 = ($i$08|0)==($54|0); //@line 1009 "modules/window/sdl/Window.cpp"
    if ($55) {
     HEAP32[$sdlbutton>>2] = 1; //@line 1010 "modules/window/sdl/Window.cpp"
     $71 = 3;
    } else {
     $71 = 2;
    }
    $69 = HEAP32[$39>>2]|0; //@line 1012 "modules/window/sdl/Window.cpp"
    $70 = ($i$08|0)==($69|0); //@line 1012 "modules/window/sdl/Window.cpp"
    if ($70) {
     HEAP32[$sdlbutton>>2] = $71; //@line 1013 "modules/window/sdl/Window.cpp"
    }
    $72 = HEAP32[$33>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $73 = HEAP32[$40>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $74 = ($72|0)==($73|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($74) {
     __THREW__ = 0;
     invoke_vii(3812,($sdlbuttons|0),($sdlbutton|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $77 = __THREW__; __THREW__ = 0;
     $78 = $77&1;
     if ($78) {
      label = 18;
      break;
     }
    } else {
     ;HEAP32[$72>>2]=HEAP32[$sdlbutton>>2]|0;HEAP32[$72+4>>2]=HEAP32[$sdlbutton+4>>2]|0;HEAP32[$72+8>>2]=HEAP32[$sdlbutton+8>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $75 = HEAP32[$33>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $76 = ((($75)) + 12|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$33>>2] = $76; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    $79 = (($i$08) + 1)|0; //@line 1002 "modules/window/sdl/Window.cpp"
    $80 = HEAP32[$27>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $81 = HEAP32[$26>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $82 = (($80) - ($81))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $83 = (($82|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $84 = ($79|0)<($83|0); //@line 1002 "modules/window/sdl/Window.cpp"
    if ($84) {
     $$in = $81;$i$08 = $79;
    } else {
     label = 11;
     break;
    }
   }
   if ((label|0) == 11) {
    $$pre = HEAP32[$sdlbuttons>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $42 = $$pre;
    label = 12;
    break;
   }
   else if ((label|0) == 18) {
    $56 = ___cxa_find_matching_catch_2()|0;
    $57 = tempRet0;
    $lpad$phi$index = $56;$lpad$phi$index2 = $57;
    break;
   }
  } else {
   $42 = 0;
   label = 12;
  }
 } while(0);
 do {
  if ((label|0) == 12) {
   $41 = ((($sdldata)) + 20|0); //@line 1018 "modules/window/sdl/Window.cpp"
   HEAP32[$41>>2] = $42; //@line 1018 "modules/window/sdl/Window.cpp"
   HEAP32[$pressedbutton>>2] = -2; //@line 1020 "modules/window/sdl/Window.cpp"
   __THREW__ = 0;
   (invoke_iii(3811,($sdldata|0),($pressedbutton|0))|0); //@line 1021 "modules/window/sdl/Window.cpp"
   $43 = __THREW__; __THREW__ = 0;
   $44 = $43&1;
   if ($44) {
    $58 = ___cxa_find_matching_catch_2()|0;
    $59 = tempRet0;
    $lpad$phi$index = $58;$lpad$phi$index2 = $59;
    break;
   }
   $85 = HEAP32[$pressedbutton>>2]|0; //@line 1023 "modules/window/sdl/Window.cpp"
   $86 = HEAP32[$sdlbuttons>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $87 = ($86|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $88 = $86; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($87) {
    STACKTOP = sp;return ($85|0); //@line 1024 "modules/window/sdl/Window.cpp"
   }
   $89 = HEAP32[$33>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $90 = ($89|0)==($86|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($90)) {
    $scevgep$i$i$i = ((($89)) + -12|0);
    $91 = $scevgep$i$i$i;
    $92 = (($91) - ($88))|0;
    $93 = (($92>>>0) / 12)&-1;
    $94 = $93 ^ -1;
    $scevgep4$i$i$i = (($89) + (($94*12)|0)|0);
    HEAP32[$33>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($86); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   STACKTOP = sp;return ($85|0); //@line 1024 "modules/window/sdl/Window.cpp"
  }
 } while(0);
 $60 = HEAP32[$sdlbuttons>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $61 = ($60|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $62 = $60; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($61) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $63 = HEAP32[$33>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $64 = ($63|0)==($60|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($64)) {
  $scevgep$i$i$i2 = ((($63)) + -12|0);
  $65 = $scevgep$i$i$i2;
  $66 = (($65) - ($62))|0;
  $67 = (($66>>>0) / 12)&-1;
  $68 = $67 ^ -1;
  $scevgep4$i$i$i3 = (($63) + (($68*12)|0)|0);
  HEAP32[$33>>2] = $scevgep4$i$i$i3; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($60); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__16vectorI24SDL_MessageBoxButtonDataNS_9allocatorIS1_EEE21__push_back_slow_pathIRKS1_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($3|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(357913941); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9|0) / 12)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(178956970); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $10 << 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = (($19|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$32 = $18;
  } else {
   $$0$i1 = $17;$39 = $18;$40 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 357913941;$39 = $12;$40 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = ($$0$i1*12)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $40;$32 = $39;
 }
 $24 = (($25) + (($26*12)|0)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + (($$0$i2*12)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 ;HEAP32[$24>>2]=HEAP32[$__x>>2]|0;HEAP32[$24+4>>2]=HEAP32[$__x+4>>2]|0;HEAP32[$24+8>>2]=HEAP32[$__x+8>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $29 = ((($24)) + 12|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = $29; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = (($32) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = (($31|0) / -12)&-1; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = (($24) + (($33*12)|0)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = $34; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($34|0),($36|0),($31|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $35; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $37 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($37) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $38 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($38); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love6window3sdl6Window16requestAttentionEb($this,$continuous) {
 $this = $this|0;
 $continuous = $continuous|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 1064 "modules/window/sdl/Window.cpp"
}
function __ZN4love6window3sdl6WindowC2Ev($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, $vararg_buffer = 0, $version = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $version = sp + 4|0;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./common/Module.h"
 HEAP32[$this>>2] = (20028); //@line 65 "modules/window/sdl/Window.cpp"
 $0 = ((($this)) + 8|0); //@line 58 "modules/window/sdl/Window.cpp"
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $1 = ((($this)) + 20|0); //@line 134 "modules/window/sdl/Window.h"
 HEAP32[$1>>2] = 800; //@line 134 "modules/window/sdl/Window.h"
 $2 = ((($this)) + 24|0); //@line 135 "modules/window/sdl/Window.h"
 HEAP32[$2>>2] = 600; //@line 135 "modules/window/sdl/Window.h"
 $3 = ((($this)) + 28|0); //@line 136 "modules/window/sdl/Window.h"
 HEAP32[$3>>2] = 800; //@line 136 "modules/window/sdl/Window.h"
 $4 = ((($this)) + 32|0); //@line 137 "modules/window/sdl/Window.h"
 HEAP32[$4>>2] = 600; //@line 137 "modules/window/sdl/Window.h"
 $5 = ((($this)) + 40|0); //@line 204 "./modules/window/Window.h"
 HEAP8[$5>>0] = 0; //@line 204 "./modules/window/Window.h"
 $6 = ((($this)) + 44|0); //@line 205 "./modules/window/Window.h"
 HEAP32[$6>>2] = 1; //@line 205 "./modules/window/Window.h"
 $7 = ((($this)) + 48|0); //@line 206 "./modules/window/Window.h"
 HEAP8[$7>>0] = 1; //@line 206 "./modules/window/Window.h"
 $8 = ((($this)) + 52|0); //@line 207 "./modules/window/Window.h"
 HEAP32[$8>>2] = 0; //@line 207 "./modules/window/Window.h"
 $9 = ((($this)) + 56|0); //@line 208 "./modules/window/Window.h"
 HEAP8[$9>>0] = 0; //@line 208 "./modules/window/Window.h"
 $10 = ((($this)) + 60|0); //@line 209 "./modules/window/Window.h"
 HEAP32[$10>>2] = 1; //@line 209 "./modules/window/Window.h"
 $11 = ((($this)) + 64|0); //@line 210 "./modules/window/Window.h"
 HEAP32[$11>>2] = 1; //@line 210 "./modules/window/Window.h"
 $12 = ((($this)) + 68|0); //@line 211 "./modules/window/Window.h"
 HEAP8[$12>>0] = 0; //@line 211 "./modules/window/Window.h"
 $13 = ((($this)) + 69|0); //@line 212 "./modules/window/Window.h"
 HEAP8[$13>>0] = 1; //@line 212 "./modules/window/Window.h"
 $14 = ((($this)) + 72|0); //@line 213 "./modules/window/Window.h"
 HEAP32[$14>>2] = 0; //@line 213 "./modules/window/Window.h"
 $15 = ((($this)) + 76|0); //@line 214 "./modules/window/Window.h"
 HEAP8[$15>>0] = 0; //@line 214 "./modules/window/Window.h"
 $16 = ((($this)) + 80|0); //@line 215 "./modules/window/Window.h"
 HEAPF64[$16>>3] = 0.0; //@line 215 "./modules/window/Window.h"
 $17 = ((($this)) + 88|0); //@line 216 "./modules/window/Window.h"
 HEAP8[$17>>0] = 0; //@line 216 "./modules/window/Window.h"
 $18 = ((($this)) + 92|0); //@line 217 "./modules/window/Window.h"
 HEAP32[$18>>2] = 0; //@line 217 "./modules/window/Window.h"
 $19 = ((($this)) + 96|0); //@line 218 "./modules/window/Window.h"
 HEAP32[$19>>2] = 0; //@line 218 "./modules/window/Window.h"
 $20 = ((($this)) + 104|0); //@line 88 "./common/Object.h"
 HEAP32[$20>>2] = 0; //@line 88 "./common/Object.h"
 $21 = ((($this)) + 108|0); //@line 59 "modules/window/sdl/Window.cpp"
 HEAP8[$21>>0] = 0; //@line 59 "modules/window/sdl/Window.cpp"
 $22 = ((($this)) + 109|0); //@line 60 "modules/window/sdl/Window.cpp"
 HEAP8[$22>>0] = 0; //@line 60 "modules/window/sdl/Window.cpp"
 $23 = ((($this)) + 112|0); //@line 61 "modules/window/sdl/Window.cpp"
 $24 = ((($this)) + 121|0); //@line 64 "modules/window/sdl/Window.cpp"
 ;HEAP32[$23>>2]=0|0;HEAP32[$23+4>>2]=0|0;HEAP16[$23+8>>1]=0|0; //@line 61 "modules/window/sdl/Window.cpp"
 __THREW__ = 0;
 $25 = (invoke_ii(3813,32)|0); //@line 66 "modules/window/sdl/Window.cpp"
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 do {
  if ($27) {
   label = 6;
  } else {
   $28 = ($25|0)<(0); //@line 66 "modules/window/sdl/Window.cpp"
   if ($28) {
    $29 = (___cxa_allocate_exception(16)|0); //@line 67 "modules/window/sdl/Window.cpp"
    __THREW__ = 0;
    $30 = (invoke_i(3814)|0); //@line 67 "modules/window/sdl/Window.cpp"
    $31 = __THREW__; __THREW__ = 0;
    $32 = $31&1;
    if (!($32)) {
     __THREW__ = 0;
     HEAP32[$vararg_buffer>>2] = $30; //@line 67 "modules/window/sdl/Window.cpp"
     invoke_viii(3349,($29|0),(360724|0),($vararg_buffer|0)); //@line 67 "modules/window/sdl/Window.cpp"
     $33 = __THREW__; __THREW__ = 0;
     $34 = $33&1;
     if (!($34)) {
      __THREW__ = 0;
      invoke_viii(3350,($29|0),(96|0),(344|0)); //@line 67 "modules/window/sdl/Window.cpp"
      $35 = __THREW__; __THREW__ = 0;
      label = 6;
      break;
     }
    }
    $38 = ___cxa_find_matching_catch_2()|0;
    $39 = tempRet0;
    ___cxa_free_exception(($29|0)); //@line 67 "modules/window/sdl/Window.cpp"
    $$0 = $39;$$01 = $38;
    break;
   }
   $40 = HEAP32[$this>>2]|0; //@line 70 "modules/window/sdl/Window.cpp"
   $41 = ((($40)) + 92|0); //@line 70 "modules/window/sdl/Window.cpp"
   $42 = HEAP32[$41>>2]|0; //@line 70 "modules/window/sdl/Window.cpp"
   __THREW__ = 0;
   invoke_vii($42|0,($this|0),0); //@line 70 "modules/window/sdl/Window.cpp"
   $43 = __THREW__; __THREW__ = 0;
   $44 = $43&1;
   if ($44) {
    label = 6;
   } else {
    ;HEAP8[$version>>0]=0|0;HEAP8[$version+1>>0]=0|0;HEAP8[$version+2>>0]=0|0; //@line 72 "modules/window/sdl/Window.cpp"
    __THREW__ = 0;
    invoke_vi(3815,($version|0)); //@line 73 "modules/window/sdl/Window.cpp"
    $45 = __THREW__; __THREW__ = 0;
    $46 = $45&1;
    if ($46) {
     label = 6;
    } else {
     $47 = HEAP8[$version>>0]|0; //@line 74 "modules/window/sdl/Window.cpp"
     $48 = ($47<<24>>24)==(2); //@line 74 "modules/window/sdl/Window.cpp"
     if (!($48)) {
      $55 = 0;
      HEAP8[$24>>0] = $55; //@line 74 "modules/window/sdl/Window.cpp"
      STACKTOP = sp;return; //@line 75 "modules/window/sdl/Window.cpp"
     }
     $49 = ((($version)) + 1|0); //@line 74 "modules/window/sdl/Window.cpp"
     $50 = HEAP8[$49>>0]|0; //@line 74 "modules/window/sdl/Window.cpp"
     $51 = ($50<<24>>24)==(0); //@line 74 "modules/window/sdl/Window.cpp"
     if (!($51)) {
      $55 = 0;
      HEAP8[$24>>0] = $55; //@line 74 "modules/window/sdl/Window.cpp"
      STACKTOP = sp;return; //@line 75 "modules/window/sdl/Window.cpp"
     }
     $52 = ((($version)) + 2|0); //@line 74 "modules/window/sdl/Window.cpp"
     $53 = HEAP8[$52>>0]|0; //@line 74 "modules/window/sdl/Window.cpp"
     $54 = ($53&255)<(4); //@line 74 "modules/window/sdl/Window.cpp"
     $phitmp = $54&1;
     $55 = $phitmp;
     HEAP8[$24>>0] = $55; //@line 74 "modules/window/sdl/Window.cpp"
     STACKTOP = sp;return; //@line 75 "modules/window/sdl/Window.cpp"
    }
   }
  }
 } while(0);
 if ((label|0) == 6) {
  $36 = ___cxa_find_matching_catch_2()|0;
  $37 = tempRet0;
  $$0 = $37;$$01 = $36;
 }
 $56 = HEAP32[$20>>2]|0; //@line 106 "./common/Object.h"
 $57 = ($56|0)==(0|0); //@line 106 "./common/Object.h"
 if ($57) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 75 "modules/window/sdl/Window.cpp"
  __ZN4love6window6WindowD2Ev($this); //@line 75 "modules/window/sdl/Window.cpp"
  ___resumeException($$01|0);
  // unreachable;
 }
 $58 = HEAP32[$56>>2]|0; //@line 106 "./common/Object.h"
 $59 = ((($58)) + 12|0); //@line 106 "./common/Object.h"
 $60 = HEAP32[$59>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($60|0,($56|0)); //@line 106 "./common/Object.h"
 $61 = __THREW__; __THREW__ = 0;
 $62 = $61&1;
 if ($62) {
  $63 = ___cxa_find_matching_catch_3(0|0)|0;
  $64 = tempRet0;
  ___clang_call_terminate($63); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 75 "modules/window/sdl/Window.cpp"
  __ZN4love6window6WindowD2Ev($this); //@line 75 "modules/window/sdl/Window.cpp"
  ___resumeException($$01|0);
  // unreachable;
 }
}
function __ZN4love6window6WindowD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ModuleD2Ev($this); //@line 31 "modules/window/Window.cpp"
 return; //@line 31 "modules/window/Window.cpp"
}
function __GLOBAL__sub_I_Window_cpp() {
 var $$02$i$i$1$i = 0, $$02$i$i$i = 0, $$lcssa$i$1$i = 0, $$lcssa$i$i = 0, $$lcssa10$i$1$i = 0, $$lcssa10$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i$i$1$i = 0, $hash$0$lcssa$i$i$i = 0, $hash$01$i$i$1$i = 0, $hash$01$i$i$i = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0; //@line 85 "modules/window/Window.cpp"
 $vararg_buffer = sp; //@line 85 "modules/window/Window.cpp"
 __ZN4love9StringMapINS_6window6Window7SettingELj14EEC2EPKNS4_5EntryEj(631668,20216,112); //@line 85 "modules/window/Window.cpp"
 HEAP8[(631620)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(631632)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(631644)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(631656)>>0] = 0; //@line 155 "./common/StringMap.h"
 $0 = (631660); //@line 44 "./common/StringMap.h"
 $1 = $0; //@line 44 "./common/StringMap.h"
 HEAP32[$1>>2] = 0; //@line 44 "./common/StringMap.h"
 $2 = (($0) + 4)|0; //@line 44 "./common/StringMap.h"
 $3 = $2; //@line 44 "./common/StringMap.h"
 HEAP32[$3>>2] = 0; //@line 44 "./common/StringMap.h"
 $4 = HEAP32[5082]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(20332)>>2]|0; //@line 49 "./common/StringMap.h"
 $6 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
 $7 = ($6<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($7) {
  $hash$0$lcssa$i$i$i = 5381;
 } else {
  $$02$i$i$i = $4;$9 = $6;$hash$01$i$i$i = 5381;
  while(1) {
   $8 = $9 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $10 = ((($$02$i$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $11 = ($hash$01$i$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $12 = (($11) + ($8))|0; //@line 143 "./common/StringMap.h"
   $13 = HEAP8[$10>>0]|0; //@line 142 "./common/StringMap.h"
   $14 = ($13<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($14) {
    $hash$0$lcssa$i$i$i = $12;
    break;
   } else {
    $$02$i$i$i = $10;$9 = $13;$hash$01$i$i$i = $12;
   }
  }
 }
 $15 = $hash$0$lcssa$i$i$i & 3; //@line 112 "./common/StringMap.h"
 $16 = (((631612 + (($15*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $17 = HEAP8[$16>>0]|0; //@line 114 "./common/StringMap.h"
 $18 = ($17<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($18) {
  $$lcssa$i$i = $15;$$lcssa10$i$i = $16;
  label = 5;
 } else {
  $19 = (($hash$0$lcssa$i$i$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $20 = $19 & 3; //@line 112 "./common/StringMap.h"
  $21 = (((631612 + (($20*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $22 = HEAP8[$21>>0]|0; //@line 114 "./common/StringMap.h"
  $23 = ($22<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($23) {
   $$lcssa$i$i = $20;$$lcssa10$i$i = $21;
   label = 5;
  } else {
   $28 = (($hash$0$lcssa$i$i$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $29 = $28 & 3; //@line 112 "./common/StringMap.h"
   $30 = (((631612 + (($29*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $31 = HEAP8[$30>>0]|0; //@line 114 "./common/StringMap.h"
   $32 = ($31<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($32) {
    $$lcssa$i$i = $29;$$lcssa10$i$i = $30;
    label = 5;
   } else {
    $33 = (($hash$0$lcssa$i$i$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $34 = $33 & 3; //@line 112 "./common/StringMap.h"
    $35 = (((631612 + (($34*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $36 = HEAP8[$35>>0]|0; //@line 114 "./common/StringMap.h"
    $37 = ($36<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($37) {
     $$lcssa$i$i = $34;$$lcssa10$i$i = $35;
     label = 5;
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10$i$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $24 = (631612 + (($$lcssa$i$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$24>>2] = $4; //@line 118 "./common/StringMap.h"
  $25 = (((631612 + (($$lcssa$i$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$25>>2] = $5; //@line 119 "./common/StringMap.h"
 }
 $26 = ($5>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($26) {
  HEAP32[$vararg_buffer>>2] = $4; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $5; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
 } else {
  $27 = (631660 + ($5<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$27>>2] = $4; //@line 132 "./common/StringMap.h"
 }
 $38 = HEAP32[(20336)>>2]|0; //@line 49 "./common/StringMap.h"
 $39 = HEAP32[(20340)>>2]|0; //@line 49 "./common/StringMap.h"
 $40 = HEAP8[$38>>0]|0; //@line 142 "./common/StringMap.h"
 $41 = ($40<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($41) {
  $hash$0$lcssa$i$i$1$i = 5381;
 } else {
  $$02$i$i$1$i = $38;$43 = $40;$hash$01$i$i$1$i = 5381;
  while(1) {
   $42 = $43 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $44 = ((($$02$i$i$1$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $45 = ($hash$01$i$i$1$i*33)|0; //@line 143 "./common/StringMap.h"
   $46 = (($45) + ($42))|0; //@line 143 "./common/StringMap.h"
   $47 = HEAP8[$44>>0]|0; //@line 142 "./common/StringMap.h"
   $48 = ($47<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($48) {
    $hash$0$lcssa$i$i$1$i = $46;
    break;
   } else {
    $$02$i$i$1$i = $44;$43 = $47;$hash$01$i$i$1$i = $46;
   }
  }
 }
 $49 = $hash$0$lcssa$i$i$1$i & 3; //@line 112 "./common/StringMap.h"
 $50 = (((631612 + (($49*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $51 = HEAP8[$50>>0]|0; //@line 114 "./common/StringMap.h"
 $52 = ($51<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($52) {
  $$lcssa$i$1$i = $49;$$lcssa10$i$1$i = $50;
  label = 17;
 } else {
  $53 = (($hash$0$lcssa$i$i$1$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $54 = $53 & 3; //@line 112 "./common/StringMap.h"
  $55 = (((631612 + (($54*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $56 = HEAP8[$55>>0]|0; //@line 114 "./common/StringMap.h"
  $57 = ($56<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($57) {
   $$lcssa$i$1$i = $54;$$lcssa10$i$1$i = $55;
   label = 17;
  } else {
   $58 = (($hash$0$lcssa$i$i$1$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $59 = $58 & 3; //@line 112 "./common/StringMap.h"
   $60 = (((631612 + (($59*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $61 = HEAP8[$60>>0]|0; //@line 114 "./common/StringMap.h"
   $62 = ($61<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($62) {
    $$lcssa$i$1$i = $59;$$lcssa10$i$1$i = $60;
    label = 17;
   } else {
    $63 = (($hash$0$lcssa$i$i$1$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $64 = $63 & 3; //@line 112 "./common/StringMap.h"
    $65 = (((631612 + (($64*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $66 = HEAP8[$65>>0]|0; //@line 114 "./common/StringMap.h"
    $67 = ($66<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($67) {
     $$lcssa$i$1$i = $64;$$lcssa10$i$1$i = $65;
     label = 17;
    }
   }
  }
 }
 if ((label|0) == 17) {
  HEAP8[$$lcssa10$i$1$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $68 = (631612 + (($$lcssa$i$1$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$68>>2] = $38; //@line 118 "./common/StringMap.h"
  $69 = (((631612 + (($$lcssa$i$1$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$69>>2] = $39; //@line 119 "./common/StringMap.h"
 }
 $70 = ($39>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($70) {
  HEAP32[$vararg_buffer2>>2] = $38; //@line 128 "./common/StringMap.h"
  $vararg_ptr5 = ((($vararg_buffer2)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr5>>2] = $39; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer2)|0); //@line 128 "./common/StringMap.h"
  HEAP8[(632068)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(632080)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(632092)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(632104)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(632116)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(632128)>>0] = 0; //@line 155 "./common/StringMap.h"
  ;HEAP32[(632132)>>2]=0|0;HEAP32[(632132)+4>>2]=0|0;HEAP32[(632132)+8>>2]=0|0; //@line 44 "./common/StringMap.h"
  $72 = HEAP32[5086]|0; //@line 49 "./common/StringMap.h"
  $73 = HEAP32[(20348)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_6window6Window14MessageBoxTypeELj3EE3addEPKcS3_(632060,$72,$73)|0); //@line 49 "./common/StringMap.h"
  $74 = HEAP32[(20352)>>2]|0; //@line 49 "./common/StringMap.h"
  $75 = HEAP32[(20356)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_6window6Window14MessageBoxTypeELj3EE3addEPKcS3_(632060,$74,$75)|0); //@line 49 "./common/StringMap.h"
  $76 = HEAP32[(20360)>>2]|0; //@line 49 "./common/StringMap.h"
  $77 = HEAP32[(20364)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_6window6Window14MessageBoxTypeELj3EE3addEPKcS3_(632060,$76,$77)|0); //@line 49 "./common/StringMap.h"
  STACKTOP = sp;return;
 } else {
  $71 = (631660 + ($39<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$71>>2] = $38; //@line 132 "./common/StringMap.h"
  HEAP8[(632068)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(632080)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(632092)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(632104)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(632116)>>0] = 0; //@line 155 "./common/StringMap.h"
  HEAP8[(632128)>>0] = 0; //@line 155 "./common/StringMap.h"
  ;HEAP32[(632132)>>2]=0|0;HEAP32[(632132)+4>>2]=0|0;HEAP32[(632132)+8>>2]=0|0; //@line 44 "./common/StringMap.h"
  $72 = HEAP32[5086]|0; //@line 49 "./common/StringMap.h"
  $73 = HEAP32[(20348)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_6window6Window14MessageBoxTypeELj3EE3addEPKcS3_(632060,$72,$73)|0); //@line 49 "./common/StringMap.h"
  $74 = HEAP32[(20352)>>2]|0; //@line 49 "./common/StringMap.h"
  $75 = HEAP32[(20356)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_6window6Window14MessageBoxTypeELj3EE3addEPKcS3_(632060,$74,$75)|0); //@line 49 "./common/StringMap.h"
  $76 = HEAP32[(20360)>>2]|0; //@line 49 "./common/StringMap.h"
  $77 = HEAP32[(20364)>>2]|0; //@line 49 "./common/StringMap.h"
  (__ZN4love9StringMapINS_6window6Window14MessageBoxTypeELj3EE3addEPKcS3_(632060,$76,$77)|0); //@line 49 "./common/StringMap.h"
  STACKTOP = sp;return;
 }
}
function __ZN4love9StringMapINS_6window6Window7SettingELj14EEC2EPKNS4_5EntryEj($this,$entries,$num) {
 $this = $this|0;
 $entries = $entries|0;
 $num = $num|0;
 var $$02$i$i = 0, $$lcssa26 = 0, $$lcssa28 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $hash$0$lcssa$i$i = 0;
 var $hash$01$i$i = 0, $i$08$i = 0, $i1$05 = 0, $scevgep = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 8|0); //@line 155 "./common/StringMap.h"
 HEAP8[$0>>0] = 0; //@line 155 "./common/StringMap.h"
 $1 = ((($this)) + 20|0); //@line 155 "./common/StringMap.h"
 HEAP8[$1>>0] = 0; //@line 155 "./common/StringMap.h"
 $2 = ((($this)) + 32|0); //@line 155 "./common/StringMap.h"
 HEAP8[$2>>0] = 0; //@line 155 "./common/StringMap.h"
 $3 = ((($this)) + 44|0); //@line 155 "./common/StringMap.h"
 HEAP8[$3>>0] = 0; //@line 155 "./common/StringMap.h"
 $4 = ((($this)) + 56|0); //@line 155 "./common/StringMap.h"
 HEAP8[$4>>0] = 0; //@line 155 "./common/StringMap.h"
 $5 = ((($this)) + 68|0); //@line 155 "./common/StringMap.h"
 HEAP8[$5>>0] = 0; //@line 155 "./common/StringMap.h"
 $6 = ((($this)) + 80|0); //@line 155 "./common/StringMap.h"
 HEAP8[$6>>0] = 0; //@line 155 "./common/StringMap.h"
 $7 = ((($this)) + 92|0); //@line 155 "./common/StringMap.h"
 HEAP8[$7>>0] = 0; //@line 155 "./common/StringMap.h"
 $8 = ((($this)) + 104|0); //@line 155 "./common/StringMap.h"
 HEAP8[$8>>0] = 0; //@line 155 "./common/StringMap.h"
 $9 = ((($this)) + 116|0); //@line 155 "./common/StringMap.h"
 HEAP8[$9>>0] = 0; //@line 155 "./common/StringMap.h"
 $10 = ((($this)) + 128|0); //@line 155 "./common/StringMap.h"
 HEAP8[$10>>0] = 0; //@line 155 "./common/StringMap.h"
 $11 = ((($this)) + 140|0); //@line 155 "./common/StringMap.h"
 HEAP8[$11>>0] = 0; //@line 155 "./common/StringMap.h"
 $12 = ((($this)) + 152|0); //@line 155 "./common/StringMap.h"
 HEAP8[$12>>0] = 0; //@line 155 "./common/StringMap.h"
 $13 = ((($this)) + 164|0); //@line 155 "./common/StringMap.h"
 HEAP8[$13>>0] = 0; //@line 155 "./common/StringMap.h"
 $14 = ((($this)) + 176|0); //@line 155 "./common/StringMap.h"
 HEAP8[$14>>0] = 0; //@line 155 "./common/StringMap.h"
 $15 = ((($this)) + 188|0); //@line 155 "./common/StringMap.h"
 HEAP8[$15>>0] = 0; //@line 155 "./common/StringMap.h"
 $16 = ((($this)) + 200|0); //@line 155 "./common/StringMap.h"
 HEAP8[$16>>0] = 0; //@line 155 "./common/StringMap.h"
 $17 = ((($this)) + 212|0); //@line 155 "./common/StringMap.h"
 HEAP8[$17>>0] = 0; //@line 155 "./common/StringMap.h"
 $18 = ((($this)) + 224|0); //@line 155 "./common/StringMap.h"
 HEAP8[$18>>0] = 0; //@line 155 "./common/StringMap.h"
 $19 = ((($this)) + 236|0); //@line 155 "./common/StringMap.h"
 HEAP8[$19>>0] = 0; //@line 155 "./common/StringMap.h"
 $20 = ((($this)) + 248|0); //@line 155 "./common/StringMap.h"
 HEAP8[$20>>0] = 0; //@line 155 "./common/StringMap.h"
 $21 = ((($this)) + 260|0); //@line 155 "./common/StringMap.h"
 HEAP8[$21>>0] = 0; //@line 155 "./common/StringMap.h"
 $22 = ((($this)) + 272|0); //@line 155 "./common/StringMap.h"
 HEAP8[$22>>0] = 0; //@line 155 "./common/StringMap.h"
 $23 = ((($this)) + 284|0); //@line 155 "./common/StringMap.h"
 HEAP8[$23>>0] = 0; //@line 155 "./common/StringMap.h"
 $24 = ((($this)) + 296|0); //@line 155 "./common/StringMap.h"
 HEAP8[$24>>0] = 0; //@line 155 "./common/StringMap.h"
 $25 = ((($this)) + 308|0); //@line 155 "./common/StringMap.h"
 HEAP8[$25>>0] = 0; //@line 155 "./common/StringMap.h"
 $26 = ((($this)) + 320|0); //@line 155 "./common/StringMap.h"
 HEAP8[$26>>0] = 0; //@line 155 "./common/StringMap.h"
 $27 = ((($this)) + 332|0); //@line 155 "./common/StringMap.h"
 HEAP8[$27>>0] = 0; //@line 155 "./common/StringMap.h"
 $scevgep = ((($this)) + 336|0); //@line 44 "./common/StringMap.h"
 dest=$scevgep; stop=dest+56|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 44 "./common/StringMap.h"
 $28 = $num >>> 3; //@line 46 "./common/StringMap.h"
 $29 = ($28|0)==(0); //@line 48 "./common/StringMap.h"
 if ($29) {
  STACKTOP = sp;return; //@line 50 "./common/StringMap.h"
 } else {
  $i1$05 = 0;
 }
 while(1) {
  $30 = (($entries) + ($i1$05<<3)|0); //@line 49 "./common/StringMap.h"
  $31 = HEAP32[$30>>2]|0; //@line 49 "./common/StringMap.h"
  $32 = (((($entries) + ($i1$05<<3)|0)) + 4|0); //@line 49 "./common/StringMap.h"
  $33 = HEAP32[$32>>2]|0; //@line 49 "./common/StringMap.h"
  $34 = HEAP8[$31>>0]|0; //@line 142 "./common/StringMap.h"
  $35 = ($34<<24>>24)==(0); //@line 142 "./common/StringMap.h"
  if ($35) {
   $hash$0$lcssa$i$i = 5381;
  } else {
   $$02$i$i = $31;$37 = $34;$hash$01$i$i = 5381;
   while(1) {
    $36 = $37 << 24 >> 24; //@line 142 "./common/StringMap.h"
    $38 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
    $39 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
    $40 = (($39) + ($36))|0; //@line 143 "./common/StringMap.h"
    $41 = HEAP8[$38>>0]|0; //@line 142 "./common/StringMap.h"
    $42 = ($41<<24>>24)==(0); //@line 142 "./common/StringMap.h"
    if ($42) {
     $hash$0$lcssa$i$i = $40;
     break;
    } else {
     $$02$i$i = $38;$37 = $41;$hash$01$i$i = $40;
    }
   }
  }
  $i$08$i = 0;
  while(1) {
   $45 = (($i$08$i) + ($hash$0$lcssa$i$i))|0; //@line 112 "./common/StringMap.h"
   $46 = (($45>>>0) % 28)&-1; //@line 112 "./common/StringMap.h"
   $47 = (((($this) + (($46*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $48 = HEAP8[$47>>0]|0; //@line 114 "./common/StringMap.h"
   $49 = ($48<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   $44 = (($i$08$i) + 1)|0; //@line 110 "./common/StringMap.h"
   if ($49) {
    $$lcssa26 = $46;$$lcssa28 = $47;
    label = 8;
    break;
   }
   $43 = ($44>>>0)<(28); //@line 110 "./common/StringMap.h"
   if ($43) {
    $i$08$i = $44;
   } else {
    break;
   }
  }
  if ((label|0) == 8) {
   label = 0;
   HEAP8[$$lcssa28>>0] = 1; //@line 117 "./common/StringMap.h"
   $50 = (($this) + (($$lcssa26*12)|0)|0); //@line 118 "./common/StringMap.h"
   HEAP32[$50>>2] = $31; //@line 118 "./common/StringMap.h"
   $51 = (((($this) + (($$lcssa26*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
   HEAP32[$51>>2] = $33; //@line 119 "./common/StringMap.h"
  }
  $52 = ($33>>>0)>(13); //@line 126 "./common/StringMap.h"
  if ($52) {
   HEAP32[$vararg_buffer>>2] = $31; //@line 128 "./common/StringMap.h"
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
   HEAP32[$vararg_ptr1>>2] = $33; //@line 128 "./common/StringMap.h"
   (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  } else {
   $53 = (((($this)) + 336|0) + ($33<<2)|0); //@line 132 "./common/StringMap.h"
   HEAP32[$53>>2] = $31; //@line 132 "./common/StringMap.h"
  }
  $54 = (($i1$05) + 1)|0; //@line 48 "./common/StringMap.h"
  $exitcond = ($54|0)==($28|0); //@line 48 "./common/StringMap.h"
  if ($exitcond) {
   break;
  } else {
   $i1$05 = $54;
  }
 }
 STACKTOP = sp;return; //@line 50 "./common/StringMap.h"
}
function __ZN4love9StringMapINS_6window6Window14MessageBoxTypeELj3EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $inserted$2$off0 = 0;
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(2); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 72|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love6window6Window11getConstantEPKcRNS1_14FullscreenTypeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love9StringMapINS_6window6Window14FullscreenTypeELj2EE4findEPKcRS3_(631612,$in,$out)|0); //@line 39 "modules/window/Window.cpp"
 return ($0|0); //@line 39 "modules/window/Window.cpp"
}
function __ZN4love9StringMapINS_6window6Window14FullscreenTypeELj2EE4findEPKcRS3_($this,$key,$t) {
 $this = $this|0;
 $key = $key|0;
 $t = $t|0;
 var $$$2 = 0, $$011$i$137 = 0, $$011$i$233 = 0, $$011$i$330 = 0, $$011$i41 = 0, $$0110$i$138 = 0, $$0110$i$234 = 0, $$0110$i$331 = 0, $$0110$i42 = 0, $$02$i = 0, $$13 = 0, $$2 = 0, $$lcssa12 = 0, $$pre14 = 0, $$pre16 = 0, $$pre18 = 0, $$pre20 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $or$cond$i = 0, $or$cond$i$1 = 0, $or$cond$i$136 = 0, $or$cond$i$2 = 0, $or$cond$i$232 = 0, $or$cond$i$3 = 0, $or$cond$i$329 = 0, $or$cond$i40 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 3; //@line 72 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 74 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 74 "./common/StringMap.h"
 if ($12) {
  $$13 = 1;$$2 = 0;
  $$$2 = $$13 & $$2;
  return ($$$2|0); //@line 85 "./common/StringMap.h"
 }
 $13 = (($this) + (($9*12)|0)|0); //@line 77 "./common/StringMap.h"
 $14 = HEAP32[$13>>2]|0; //@line 77 "./common/StringMap.h"
 $15 = HEAP8[$14>>0]|0; //@line 54 "./common/StringMap.h"
 $16 = ($15<<24>>24)==(0); //@line 54 "./common/StringMap.h"
 L8: do {
  if ($16) {
   $26 = $0;
   label = 8;
  } else {
   $17 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
   $18 = ($15<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
   $or$cond$i40 = $17 & $18; //@line 54 "./common/StringMap.h"
   if ($or$cond$i40) {
    $$011$i41 = $key;$$0110$i42 = $14;
    while(1) {
     $19 = ((($$0110$i42)) + 1|0); //@line 59 "./common/StringMap.h"
     $20 = ((($$011$i41)) + 1|0); //@line 60 "./common/StringMap.h"
     $21 = HEAP8[$19>>0]|0; //@line 54 "./common/StringMap.h"
     $22 = ($21<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     $$pre14 = HEAP8[$20>>0]|0; //@line 63 "./common/StringMap.h"
     if ($22) {
      $26 = $$pre14;
      label = 8;
      break L8;
     }
     $23 = ($$pre14<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $24 = ($21<<24>>24)==($$pre14<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i = $23 & $24; //@line 54 "./common/StringMap.h"
     if ($or$cond$i) {
      $$011$i41 = $20;$$0110$i42 = $19;
     } else {
      label = 10;
      break;
     }
    }
   } else {
    label = 10;
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $25 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
  if ($25) {
   $$lcssa12 = $9;
  } else {
   label = 10;
  }
 }
 do {
  if ((label|0) == 10) {
   $29 = (($hash$0$lcssa$i) + 1)|0; //@line 72 "./common/StringMap.h"
   $30 = $29 & 3; //@line 72 "./common/StringMap.h"
   $31 = (((($this) + (($30*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $32 = HEAP8[$31>>0]|0; //@line 74 "./common/StringMap.h"
   $33 = ($32<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($33) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $34 = (($this) + (($30*12)|0)|0); //@line 77 "./common/StringMap.h"
   $35 = HEAP32[$34>>2]|0; //@line 77 "./common/StringMap.h"
   $36 = HEAP8[$35>>0]|0; //@line 54 "./common/StringMap.h"
   $37 = ($36<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L20: do {
    if ($37) {
     $47 = $0;
     label = 16;
    } else {
     $38 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $39 = ($36<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$136 = $38 & $39; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$136) {
      $$011$i$137 = $key;$$0110$i$138 = $35;
      while(1) {
       $40 = ((($$0110$i$138)) + 1|0); //@line 59 "./common/StringMap.h"
       $41 = ((($$011$i$137)) + 1|0); //@line 60 "./common/StringMap.h"
       $42 = HEAP8[$40>>0]|0; //@line 54 "./common/StringMap.h"
       $43 = ($42<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre16 = HEAP8[$41>>0]|0; //@line 63 "./common/StringMap.h"
       if ($43) {
        $47 = $$pre16;
        label = 16;
        break L20;
       }
       $44 = ($$pre16<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $45 = ($42<<24>>24)==($$pre16<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$1 = $44 & $45; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$1) {
        $$011$i$137 = $41;$$0110$i$138 = $40;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 16) {
    $46 = ($47<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($46) {
     $$lcssa12 = $30;
     break;
    }
   }
   $48 = (($hash$0$lcssa$i) + 2)|0; //@line 72 "./common/StringMap.h"
   $49 = $48 & 3; //@line 72 "./common/StringMap.h"
   $50 = (((($this) + (($49*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $51 = HEAP8[$50>>0]|0; //@line 74 "./common/StringMap.h"
   $52 = ($51<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($52) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $53 = (($this) + (($49*12)|0)|0); //@line 77 "./common/StringMap.h"
   $54 = HEAP32[$53>>2]|0; //@line 77 "./common/StringMap.h"
   $55 = HEAP8[$54>>0]|0; //@line 54 "./common/StringMap.h"
   $56 = ($55<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L31: do {
    if ($56) {
     $66 = $0;
     label = 22;
    } else {
     $57 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $58 = ($55<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$232 = $57 & $58; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$232) {
      $$011$i$233 = $key;$$0110$i$234 = $54;
      while(1) {
       $59 = ((($$0110$i$234)) + 1|0); //@line 59 "./common/StringMap.h"
       $60 = ((($$011$i$233)) + 1|0); //@line 60 "./common/StringMap.h"
       $61 = HEAP8[$59>>0]|0; //@line 54 "./common/StringMap.h"
       $62 = ($61<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre18 = HEAP8[$60>>0]|0; //@line 63 "./common/StringMap.h"
       if ($62) {
        $66 = $$pre18;
        label = 22;
        break L31;
       }
       $63 = ($$pre18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $64 = ($61<<24>>24)==($$pre18<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$2 = $63 & $64; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$2) {
        $$011$i$233 = $60;$$0110$i$234 = $59;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 22) {
    $65 = ($66<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($65) {
     $$lcssa12 = $49;
     break;
    }
   }
   $67 = (($hash$0$lcssa$i) + 3)|0; //@line 72 "./common/StringMap.h"
   $68 = $67 & 3; //@line 72 "./common/StringMap.h"
   $69 = (((($this) + (($68*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $70 = HEAP8[$69>>0]|0; //@line 74 "./common/StringMap.h"
   $71 = ($70<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($71) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $72 = (($this) + (($68*12)|0)|0); //@line 77 "./common/StringMap.h"
   $73 = HEAP32[$72>>2]|0; //@line 77 "./common/StringMap.h"
   $74 = HEAP8[$73>>0]|0; //@line 54 "./common/StringMap.h"
   $75 = ($74<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L42: do {
    if ($75) {
     $85 = $0;
    } else {
     $76 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $77 = ($74<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$329 = $76 & $77; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$329) {
      $$011$i$330 = $key;$$0110$i$331 = $73;
     } else {
      $$13 = 0;$$2 = 0;
      $$$2 = $$13 & $$2;
      return ($$$2|0); //@line 85 "./common/StringMap.h"
     }
     while(1) {
      $78 = ((($$0110$i$331)) + 1|0); //@line 59 "./common/StringMap.h"
      $79 = ((($$011$i$330)) + 1|0); //@line 60 "./common/StringMap.h"
      $80 = HEAP8[$78>>0]|0; //@line 54 "./common/StringMap.h"
      $81 = ($80<<24>>24)==(0); //@line 54 "./common/StringMap.h"
      $$pre20 = HEAP8[$79>>0]|0; //@line 63 "./common/StringMap.h"
      if ($81) {
       $85 = $$pre20;
       break L42;
      }
      $82 = ($$pre20<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
      $83 = ($80<<24>>24)==($$pre20<<24>>24); //@line 56 "./common/StringMap.h"
      $or$cond$i$3 = $82 & $83; //@line 54 "./common/StringMap.h"
      if ($or$cond$i$3) {
       $$011$i$330 = $79;$$0110$i$331 = $78;
      } else {
       $$13 = 0;$$2 = 0;
       break;
      }
     }
     $$$2 = $$13 & $$2;
     return ($$$2|0); //@line 85 "./common/StringMap.h"
    }
   } while(0);
   $84 = ($85<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($84) {
    $$lcssa12 = $68;
   } else {
    $$13 = 0;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
  }
 } while(0);
 $27 = (((($this) + (($$lcssa12*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
 $28 = HEAP32[$27>>2]|0; //@line 79 "./common/StringMap.h"
 HEAP32[$t>>2] = $28; //@line 79 "./common/StringMap.h"
 $$13 = 1;$$2 = 1;
 $$$2 = $$13 & $$2;
 return ($$$2|0); //@line 85 "./common/StringMap.h"
}
function __ZN4love6window6Window11getConstantENS1_14FullscreenTypeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(1); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (631660 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 44 "modules/window/Window.cpp"
}
function __ZN4love6window6Window11getConstantEPKcRNS1_7SettingE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 28)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((631668 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (631668 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(28); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((631668 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 49 "modules/window/Window.cpp"
  return ($$$2$i|0); //@line 49 "modules/window/Window.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 49 "modules/window/Window.cpp"
  return ($$$2$i|0); //@line 49 "modules/window/Window.cpp"
 }
 return (0)|0;
}
function __ZN4love6window6Window11getConstantENS1_7SettingERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(13); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (632004 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 54 "modules/window/Window.cpp"
}
function __ZN4love6window6Window11getConstantEPKcRNS1_14MessageBoxTypeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 6)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((632060 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (632060 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(6); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((632060 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 59 "modules/window/Window.cpp"
  return ($$$2$i|0); //@line 59 "modules/window/Window.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 59 "modules/window/Window.cpp"
  return ($$$2$i|0); //@line 59 "modules/window/Window.cpp"
 }
 return (0)|0;
}
function __ZN15b2DistanceProxy3SetEPK7b2Shapei($this,$shape,$index) {
 $this = $this|0;
 $shape = $shape|0;
 $index = $index|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($shape)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $1 = HEAP32[$0>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 switch ($1|0) {
 case 0:  {
  $2 = ((($shape)) + 12|0); //@line 35 "libraries/Box2D/Collision/b2Distance.cpp"
  $3 = ((($this)) + 16|0); //@line 35 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$3>>2] = $2; //@line 35 "libraries/Box2D/Collision/b2Distance.cpp"
  $4 = ((($this)) + 20|0); //@line 36 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$4>>2] = 1; //@line 36 "libraries/Box2D/Collision/b2Distance.cpp"
  $5 = ((($shape)) + 8|0); //@line 37 "libraries/Box2D/Collision/b2Distance.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 37 "libraries/Box2D/Collision/b2Distance.cpp"
  $7 = ((($this)) + 24|0); //@line 37 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$7>>2] = $6; //@line 37 "libraries/Box2D/Collision/b2Distance.cpp"
  return; //@line 83 "libraries/Box2D/Collision/b2Distance.cpp"
  break;
 }
 case 2:  {
  $8 = ((($shape)) + 20|0); //@line 44 "libraries/Box2D/Collision/b2Distance.cpp"
  $9 = ((($this)) + 16|0); //@line 44 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$9>>2] = $8; //@line 44 "libraries/Box2D/Collision/b2Distance.cpp"
  $10 = ((($shape)) + 148|0); //@line 45 "libraries/Box2D/Collision/b2Distance.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 45 "libraries/Box2D/Collision/b2Distance.cpp"
  $12 = ((($this)) + 20|0); //@line 45 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$12>>2] = $11; //@line 45 "libraries/Box2D/Collision/b2Distance.cpp"
  $13 = ((($shape)) + 8|0); //@line 46 "libraries/Box2D/Collision/b2Distance.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 46 "libraries/Box2D/Collision/b2Distance.cpp"
  $15 = ((($this)) + 24|0); //@line 46 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$15>>2] = $14; //@line 46 "libraries/Box2D/Collision/b2Distance.cpp"
  return; //@line 83 "libraries/Box2D/Collision/b2Distance.cpp"
  break;
 }
 case 3:  {
  $16 = ($index|0)>(-1); //@line 53 "libraries/Box2D/Collision/b2Distance.cpp"
  $17 = ((($shape)) + 16|0); //@line 53 "libraries/Box2D/Collision/b2Distance.cpp"
  if ($16) {
   $18 = HEAP32[$17>>2]|0; //@line 53 "libraries/Box2D/Collision/b2Distance.cpp"
   $19 = ($18|0)>($index|0); //@line 53 "libraries/Box2D/Collision/b2Distance.cpp"
   $20 = $19;
  } else {
   $20 = 0;
  }
  __Z10loveAssertbPKc($20,360932); //@line 53 "libraries/Box2D/Collision/b2Distance.cpp"
  $21 = ((($shape)) + 12|0); //@line 55 "libraries/Box2D/Collision/b2Distance.cpp"
  $22 = HEAP32[$21>>2]|0; //@line 55 "libraries/Box2D/Collision/b2Distance.cpp"
  $23 = (($22) + ($index<<3)|0); //@line 55 "libraries/Box2D/Collision/b2Distance.cpp"
  $24 = $23; //@line 55 "libraries/Box2D/Collision/b2Distance.cpp"
  $25 = $24; //@line 55 "libraries/Box2D/Collision/b2Distance.cpp"
  $26 = HEAP32[$25>>2]|0; //@line 55 "libraries/Box2D/Collision/b2Distance.cpp"
  $27 = (($24) + 4)|0; //@line 55 "libraries/Box2D/Collision/b2Distance.cpp"
  $28 = $27; //@line 55 "libraries/Box2D/Collision/b2Distance.cpp"
  $29 = HEAP32[$28>>2]|0; //@line 55 "libraries/Box2D/Collision/b2Distance.cpp"
  $30 = $this; //@line 55 "libraries/Box2D/Collision/b2Distance.cpp"
  $31 = $30; //@line 55 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$31>>2] = $26; //@line 55 "libraries/Box2D/Collision/b2Distance.cpp"
  $32 = (($30) + 4)|0; //@line 55 "libraries/Box2D/Collision/b2Distance.cpp"
  $33 = $32; //@line 55 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$33>>2] = $29; //@line 55 "libraries/Box2D/Collision/b2Distance.cpp"
  $34 = (($index) + 1)|0; //@line 56 "libraries/Box2D/Collision/b2Distance.cpp"
  $35 = HEAP32[$17>>2]|0; //@line 56 "libraries/Box2D/Collision/b2Distance.cpp"
  $36 = ($34|0)<($35|0); //@line 56 "libraries/Box2D/Collision/b2Distance.cpp"
  $37 = ((($this)) + 8|0); //@line 58 "libraries/Box2D/Collision/b2Distance.cpp"
  if ($36) {
   $38 = HEAP32[$21>>2]|0; //@line 58 "libraries/Box2D/Collision/b2Distance.cpp"
   $39 = (($38) + ($34<<3)|0); //@line 58 "libraries/Box2D/Collision/b2Distance.cpp"
   $40 = $39; //@line 58 "libraries/Box2D/Collision/b2Distance.cpp"
   $41 = $40; //@line 58 "libraries/Box2D/Collision/b2Distance.cpp"
   $42 = HEAP32[$41>>2]|0; //@line 58 "libraries/Box2D/Collision/b2Distance.cpp"
   $43 = (($40) + 4)|0; //@line 58 "libraries/Box2D/Collision/b2Distance.cpp"
   $44 = $43; //@line 58 "libraries/Box2D/Collision/b2Distance.cpp"
   $45 = HEAP32[$44>>2]|0; //@line 58 "libraries/Box2D/Collision/b2Distance.cpp"
   $46 = $37; //@line 58 "libraries/Box2D/Collision/b2Distance.cpp"
   $47 = $46; //@line 58 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAP32[$47>>2] = $42; //@line 58 "libraries/Box2D/Collision/b2Distance.cpp"
   $48 = (($46) + 4)|0; //@line 58 "libraries/Box2D/Collision/b2Distance.cpp"
   $49 = $48; //@line 58 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAP32[$49>>2] = $45; //@line 58 "libraries/Box2D/Collision/b2Distance.cpp"
  } else {
   $50 = HEAP32[$21>>2]|0; //@line 62 "libraries/Box2D/Collision/b2Distance.cpp"
   $51 = $50; //@line 62 "libraries/Box2D/Collision/b2Distance.cpp"
   $52 = $51; //@line 62 "libraries/Box2D/Collision/b2Distance.cpp"
   $53 = HEAP32[$52>>2]|0; //@line 62 "libraries/Box2D/Collision/b2Distance.cpp"
   $54 = (($51) + 4)|0; //@line 62 "libraries/Box2D/Collision/b2Distance.cpp"
   $55 = $54; //@line 62 "libraries/Box2D/Collision/b2Distance.cpp"
   $56 = HEAP32[$55>>2]|0; //@line 62 "libraries/Box2D/Collision/b2Distance.cpp"
   $57 = $37; //@line 62 "libraries/Box2D/Collision/b2Distance.cpp"
   $58 = $57; //@line 62 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAP32[$58>>2] = $53; //@line 62 "libraries/Box2D/Collision/b2Distance.cpp"
   $59 = (($57) + 4)|0; //@line 62 "libraries/Box2D/Collision/b2Distance.cpp"
   $60 = $59; //@line 62 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAP32[$60>>2] = $56; //@line 62 "libraries/Box2D/Collision/b2Distance.cpp"
  }
  $61 = ((($this)) + 16|0); //@line 65 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$61>>2] = $this; //@line 65 "libraries/Box2D/Collision/b2Distance.cpp"
  $62 = ((($this)) + 20|0); //@line 66 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$62>>2] = 2; //@line 66 "libraries/Box2D/Collision/b2Distance.cpp"
  $63 = ((($shape)) + 8|0); //@line 67 "libraries/Box2D/Collision/b2Distance.cpp"
  $64 = HEAP32[$63>>2]|0; //@line 67 "libraries/Box2D/Collision/b2Distance.cpp"
  $65 = ((($this)) + 24|0); //@line 67 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$65>>2] = $64; //@line 67 "libraries/Box2D/Collision/b2Distance.cpp"
  return; //@line 83 "libraries/Box2D/Collision/b2Distance.cpp"
  break;
 }
 case 1:  {
  $66 = ((($shape)) + 12|0); //@line 74 "libraries/Box2D/Collision/b2Distance.cpp"
  $67 = ((($this)) + 16|0); //@line 74 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$67>>2] = $66; //@line 74 "libraries/Box2D/Collision/b2Distance.cpp"
  $68 = ((($this)) + 20|0); //@line 75 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$68>>2] = 2; //@line 75 "libraries/Box2D/Collision/b2Distance.cpp"
  $69 = ((($shape)) + 8|0); //@line 76 "libraries/Box2D/Collision/b2Distance.cpp"
  $70 = HEAP32[$69>>2]|0; //@line 76 "libraries/Box2D/Collision/b2Distance.cpp"
  $71 = ((($this)) + 24|0); //@line 76 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$71>>2] = $70; //@line 76 "libraries/Box2D/Collision/b2Distance.cpp"
  return; //@line 83 "libraries/Box2D/Collision/b2Distance.cpp"
  break;
 }
 default: {
  __Z10loveAssertbPKc(0,622525); //@line 81 "libraries/Box2D/Collision/b2Distance.cpp"
  return; //@line 83 "libraries/Box2D/Collision/b2Distance.cpp"
 }
 }
}
function __ZN9b2Simplex6Solve3Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0, $107 = 0, $11 = 0.0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0;
 var $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0.0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0;
 var $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0.0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0, $64 = 0.0, $65 = 0.0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0.0;
 var $74 = 0, $75 = 0.0, $76 = 0.0, $77 = 0, $78 = 0, $79 = 0, $8 = 0.0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0.0, $91 = 0.0;
 var $92 = 0.0, $93 = 0, $94 = 0.0, $95 = 0.0, $96 = 0, $97 = 0, $98 = 0.0, $99 = 0.0, $or$cond = 0, $or$cond11 = 0, $or$cond13 = 0, $or$cond15 = 0, $or$cond17 = 0, $or$cond3 = 0, $or$cond5 = 0, $or$cond7 = 0, $or$cond9 = 0, dest = 0, label = 0, sp = 0;
 var src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 337 "libraries/Box2D/Collision/b2Distance.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 337 "libraries/Box2D/Collision/b2Distance.cpp"
 $2 = ((($this)) + 20|0); //@line 337 "libraries/Box2D/Collision/b2Distance.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 337 "libraries/Box2D/Collision/b2Distance.cpp"
 $4 = ((($this)) + 36|0); //@line 338 "libraries/Box2D/Collision/b2Distance.cpp"
 $5 = ((($this)) + 52|0); //@line 338 "libraries/Box2D/Collision/b2Distance.cpp"
 $6 = +HEAPF32[$5>>2]; //@line 338 "libraries/Box2D/Collision/b2Distance.cpp"
 $7 = ((($this)) + 56|0); //@line 338 "libraries/Box2D/Collision/b2Distance.cpp"
 $8 = +HEAPF32[$7>>2]; //@line 338 "libraries/Box2D/Collision/b2Distance.cpp"
 $9 = ((($this)) + 72|0); //@line 339 "libraries/Box2D/Collision/b2Distance.cpp"
 $10 = ((($this)) + 88|0); //@line 339 "libraries/Box2D/Collision/b2Distance.cpp"
 $11 = +HEAPF32[$10>>2]; //@line 339 "libraries/Box2D/Collision/b2Distance.cpp"
 $12 = ((($this)) + 92|0); //@line 339 "libraries/Box2D/Collision/b2Distance.cpp"
 $13 = +HEAPF32[$12>>2]; //@line 339 "libraries/Box2D/Collision/b2Distance.cpp"
 $14 = $6 - $1; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $15 = $8 - $3; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $16 = $1 * $14; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $15; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $19 = $6 * $14; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $20 = $8 * $15; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $21 = $19 + $20; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $22 = $11 - $1; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $23 = $13 - $3; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $24 = $1 * $22; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $25 = $3 * $23; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $26 = $24 + $25; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $27 = $11 * $22; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $28 = $13 * $23; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $29 = $27 + $28; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $30 = $11 - $6; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $31 = $13 - $8; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $32 = $6 * $30; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $33 = $8 * $31; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $34 = $32 + $33; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $35 = $11 * $30; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $36 = $13 * $31; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $37 = $35 + $36; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $38 = $14 * $23; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $39 = $15 * $22; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $40 = $38 - $39; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $41 = $6 * $13; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $42 = $8 * $11; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $43 = $41 - $42; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $44 = $43 * $40; //@line 374 "libraries/Box2D/Collision/b2Distance.cpp"
 $45 = $3 * $11; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $46 = $1 * $13; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $47 = $45 - $46; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $48 = $47 * $40; //@line 375 "libraries/Box2D/Collision/b2Distance.cpp"
 $49 = $1 * $8; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $50 = $3 * $6; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $51 = $49 - $50; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $52 = $51 * $40; //@line 376 "libraries/Box2D/Collision/b2Distance.cpp"
 $53 = $18 >= -0.0; //@line 379 "libraries/Box2D/Collision/b2Distance.cpp"
 $54 = $26 >= -0.0; //@line 379 "libraries/Box2D/Collision/b2Distance.cpp"
 $or$cond = $53 & $54; //@line 379 "libraries/Box2D/Collision/b2Distance.cpp"
 if ($or$cond) {
  $55 = ((($this)) + 24|0); //@line 381 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$55>>2] = 1.0; //@line 381 "libraries/Box2D/Collision/b2Distance.cpp"
  $56 = ((($this)) + 108|0); //@line 382 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$56>>2] = 1; //@line 382 "libraries/Box2D/Collision/b2Distance.cpp"
  return; //@line 442 "libraries/Box2D/Collision/b2Distance.cpp"
 }
 $57 = $21 > 0.0; //@line 387 "libraries/Box2D/Collision/b2Distance.cpp"
 $58 = $18 < -0.0; //@line 387 "libraries/Box2D/Collision/b2Distance.cpp"
 $or$cond3 = $57 & $58; //@line 387 "libraries/Box2D/Collision/b2Distance.cpp"
 $59 = $52 <= 0.0; //@line 387 "libraries/Box2D/Collision/b2Distance.cpp"
 $or$cond5 = $or$cond3 & $59; //@line 387 "libraries/Box2D/Collision/b2Distance.cpp"
 if ($or$cond5) {
  $60 = $21 - $18; //@line 389 "libraries/Box2D/Collision/b2Distance.cpp"
  $61 = 1.0 / $60; //@line 389 "libraries/Box2D/Collision/b2Distance.cpp"
  $62 = $21 * $61; //@line 390 "libraries/Box2D/Collision/b2Distance.cpp"
  $63 = ((($this)) + 24|0); //@line 390 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$63>>2] = $62; //@line 390 "libraries/Box2D/Collision/b2Distance.cpp"
  $64 = $18 * $61; //@line 391 "libraries/Box2D/Collision/b2Distance.cpp"
  $65 = -$64; //@line 391 "libraries/Box2D/Collision/b2Distance.cpp"
  $66 = ((($this)) + 60|0); //@line 391 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$66>>2] = $65; //@line 391 "libraries/Box2D/Collision/b2Distance.cpp"
  $67 = ((($this)) + 108|0); //@line 392 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$67>>2] = 2; //@line 392 "libraries/Box2D/Collision/b2Distance.cpp"
  return; //@line 442 "libraries/Box2D/Collision/b2Distance.cpp"
 }
 $68 = $29 > 0.0; //@line 397 "libraries/Box2D/Collision/b2Distance.cpp"
 $69 = $26 < -0.0; //@line 397 "libraries/Box2D/Collision/b2Distance.cpp"
 $or$cond7 = $68 & $69; //@line 397 "libraries/Box2D/Collision/b2Distance.cpp"
 $70 = $48 <= 0.0; //@line 397 "libraries/Box2D/Collision/b2Distance.cpp"
 $or$cond9 = $or$cond7 & $70; //@line 397 "libraries/Box2D/Collision/b2Distance.cpp"
 if ($or$cond9) {
  $71 = $29 - $26; //@line 399 "libraries/Box2D/Collision/b2Distance.cpp"
  $72 = 1.0 / $71; //@line 399 "libraries/Box2D/Collision/b2Distance.cpp"
  $73 = $29 * $72; //@line 400 "libraries/Box2D/Collision/b2Distance.cpp"
  $74 = ((($this)) + 24|0); //@line 400 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$74>>2] = $73; //@line 400 "libraries/Box2D/Collision/b2Distance.cpp"
  $75 = $26 * $72; //@line 401 "libraries/Box2D/Collision/b2Distance.cpp"
  $76 = -$75; //@line 401 "libraries/Box2D/Collision/b2Distance.cpp"
  $77 = ((($this)) + 96|0); //@line 401 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$77>>2] = $76; //@line 401 "libraries/Box2D/Collision/b2Distance.cpp"
  $78 = ((($this)) + 108|0); //@line 402 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$78>>2] = 2; //@line 402 "libraries/Box2D/Collision/b2Distance.cpp"
  dest=$4; src=$9; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 403 "libraries/Box2D/Collision/b2Distance.cpp"
  return; //@line 442 "libraries/Box2D/Collision/b2Distance.cpp"
 }
 $79 = $21 <= 0.0; //@line 408 "libraries/Box2D/Collision/b2Distance.cpp"
 $80 = $34 >= -0.0; //@line 408 "libraries/Box2D/Collision/b2Distance.cpp"
 $or$cond11 = $79 & $80; //@line 408 "libraries/Box2D/Collision/b2Distance.cpp"
 if ($or$cond11) {
  $81 = ((($this)) + 60|0); //@line 410 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$81>>2] = 1.0; //@line 410 "libraries/Box2D/Collision/b2Distance.cpp"
  $82 = ((($this)) + 108|0); //@line 411 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$82>>2] = 1; //@line 411 "libraries/Box2D/Collision/b2Distance.cpp"
  dest=$this; src=$4; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 412 "libraries/Box2D/Collision/b2Distance.cpp"
  return; //@line 442 "libraries/Box2D/Collision/b2Distance.cpp"
 }
 $83 = $29 <= 0.0; //@line 417 "libraries/Box2D/Collision/b2Distance.cpp"
 $84 = $37 <= 0.0; //@line 417 "libraries/Box2D/Collision/b2Distance.cpp"
 $or$cond13 = $83 & $84; //@line 417 "libraries/Box2D/Collision/b2Distance.cpp"
 if ($or$cond13) {
  $85 = ((($this)) + 96|0); //@line 419 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$85>>2] = 1.0; //@line 419 "libraries/Box2D/Collision/b2Distance.cpp"
  $86 = ((($this)) + 108|0); //@line 420 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$86>>2] = 1; //@line 420 "libraries/Box2D/Collision/b2Distance.cpp"
  dest=$this; src=$9; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 421 "libraries/Box2D/Collision/b2Distance.cpp"
  return; //@line 442 "libraries/Box2D/Collision/b2Distance.cpp"
 }
 $87 = $37 > 0.0; //@line 426 "libraries/Box2D/Collision/b2Distance.cpp"
 $88 = $34 < -0.0; //@line 426 "libraries/Box2D/Collision/b2Distance.cpp"
 $or$cond15 = $87 & $88; //@line 426 "libraries/Box2D/Collision/b2Distance.cpp"
 $89 = $44 <= 0.0; //@line 426 "libraries/Box2D/Collision/b2Distance.cpp"
 $or$cond17 = $or$cond15 & $89; //@line 426 "libraries/Box2D/Collision/b2Distance.cpp"
 if ($or$cond17) {
  $90 = $37 - $34; //@line 428 "libraries/Box2D/Collision/b2Distance.cpp"
  $91 = 1.0 / $90; //@line 428 "libraries/Box2D/Collision/b2Distance.cpp"
  $92 = $37 * $91; //@line 429 "libraries/Box2D/Collision/b2Distance.cpp"
  $93 = ((($this)) + 60|0); //@line 429 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$93>>2] = $92; //@line 429 "libraries/Box2D/Collision/b2Distance.cpp"
  $94 = $34 * $91; //@line 430 "libraries/Box2D/Collision/b2Distance.cpp"
  $95 = -$94; //@line 430 "libraries/Box2D/Collision/b2Distance.cpp"
  $96 = ((($this)) + 96|0); //@line 430 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$96>>2] = $95; //@line 430 "libraries/Box2D/Collision/b2Distance.cpp"
  $97 = ((($this)) + 108|0); //@line 431 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$97>>2] = 2; //@line 431 "libraries/Box2D/Collision/b2Distance.cpp"
  dest=$this; src=$9; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 432 "libraries/Box2D/Collision/b2Distance.cpp"
  return; //@line 442 "libraries/Box2D/Collision/b2Distance.cpp"
 } else {
  $98 = $44 + $48; //@line 437 "libraries/Box2D/Collision/b2Distance.cpp"
  $99 = $52 + $98; //@line 437 "libraries/Box2D/Collision/b2Distance.cpp"
  $100 = 1.0 / $99; //@line 437 "libraries/Box2D/Collision/b2Distance.cpp"
  $101 = $44 * $100; //@line 438 "libraries/Box2D/Collision/b2Distance.cpp"
  $102 = ((($this)) + 24|0); //@line 438 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$102>>2] = $101; //@line 438 "libraries/Box2D/Collision/b2Distance.cpp"
  $103 = $48 * $100; //@line 439 "libraries/Box2D/Collision/b2Distance.cpp"
  $104 = ((($this)) + 60|0); //@line 439 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$104>>2] = $103; //@line 439 "libraries/Box2D/Collision/b2Distance.cpp"
  $105 = $52 * $100; //@line 440 "libraries/Box2D/Collision/b2Distance.cpp"
  $106 = ((($this)) + 96|0); //@line 440 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$106>>2] = $105; //@line 440 "libraries/Box2D/Collision/b2Distance.cpp"
  $107 = ((($this)) + 108|0); //@line 441 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$107>>2] = 3; //@line 441 "libraries/Box2D/Collision/b2Distance.cpp"
  return; //@line 442 "libraries/Box2D/Collision/b2Distance.cpp"
 }
}
function __Z10b2DistanceP16b2DistanceOutputP14b2SimplexCachePK15b2DistanceInput($output,$cache,$input) {
 $output = $output|0;
 $cache = $cache|0;
 $input = $input|0;
 var $$0$i6 = 0.0, $$pr = 0, $$pr50 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0;
 var $130 = 0.0, $131 = 0.0, $132 = 0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0.0, $147 = 0.0, $148 = 0;
 var $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0.0, $152 = 0, $153 = 0.0, $154 = 0.0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0.0, $169 = 0.0, $17 = 0, $170 = 0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0, $182 = 0, $183 = 0.0, $184 = 0;
 var $185 = 0.0, $186 = 0, $187 = 0.0, $188 = 0.0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0.0, $236 = 0.0, $237 = 0.0, $238 = 0;
 var $239 = 0.0, $24 = 0, $240 = 0.0, $241 = 0.0, $242 = 0, $243 = 0.0, $244 = 0.0, $245 = 0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0, $251 = 0, $252 = 0.0, $253 = 0.0, $254 = 0, $255 = 0.0, $256 = 0.0;
 var $257 = 0, $258 = 0.0, $259 = 0.0, $26 = 0, $260 = 0, $261 = 0.0, $262 = 0.0, $263 = 0.0, $264 = 0.0, $265 = 0, $266 = 0.0, $267 = 0.0, $268 = 0.0, $269 = 0, $27 = 0, $270 = 0.0, $271 = 0.0, $272 = 0.0, $273 = 0, $274 = 0.0;
 var $275 = 0.0, $276 = 0, $277 = 0.0, $278 = 0.0, $279 = 0.0, $28 = 0, $280 = 0.0, $281 = 0, $282 = 0.0, $283 = 0, $284 = 0.0, $285 = 0.0, $286 = 0, $287 = 0.0, $288 = 0.0, $289 = 0.0, $29 = 0, $290 = 0.0, $291 = 0, $292 = 0;
 var $293 = 0.0, $294 = 0.0, $295 = 0.0, $296 = 0, $297 = 0.0, $298 = 0, $299 = 0.0, $3 = 0, $30 = 0, $300 = 0.0, $301 = 0.0, $302 = 0.0, $303 = 0.0, $304 = 0.0, $305 = 0, $306 = 0, $307 = 0, $308 = 0.0, $309 = 0.0, $31 = 0;
 var $310 = 0.0, $311 = 0.0, $312 = 0.0, $313 = 0.0, $314 = 0.0, $315 = 0.0, $316 = 0.0, $317 = 0.0, $318 = 0.0, $319 = 0.0, $32 = 0.0, $320 = 0.0, $321 = 0.0, $322 = 0.0, $323 = 0.0, $324 = 0, $325 = 0.0, $326 = 0.0, $327 = 0, $328 = 0.0;
 var $329 = 0.0, $33 = 0.0, $330 = 0.0, $331 = 0.0, $332 = 0.0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0.0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0;
 var $347 = 0, $348 = 0, $349 = 0, $35 = 0.0, $350 = 0, $351 = 0.0, $352 = 0, $353 = 0.0, $354 = 0.0, $355 = 0.0, $356 = 0, $357 = 0, $358 = 0.0, $359 = 0.0, $36 = 0.0, $360 = 0.0, $361 = 0.0, $362 = 0.0, $363 = 0.0, $364 = 0.0;
 var $365 = 0.0, $366 = 0.0, $367 = 0.0, $368 = 0.0, $369 = 0, $37 = 0.0, $370 = 0.0, $371 = 0.0, $372 = 0.0, $373 = 0.0, $374 = 0.0, $375 = 0.0, $376 = 0.0, $377 = 0.0, $378 = 0.0, $379 = 0.0, $38 = 0.0, $380 = 0.0, $381 = 0.0, $382 = 0.0;
 var $383 = 0.0, $384 = 0.0, $385 = 0.0, $386 = 0.0, $387 = 0.0, $388 = 0.0, $389 = 0, $39 = 0.0, $390 = 0, $391 = 0, $392 = 0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0.0, $47 = 0.0;
 var $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0;
 var $66 = 0.0, $67 = 0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0.0, $99 = 0.0, $bestIndex$0$lcssa$i1035 = 0, $bestIndex$0$lcssa$i1036 = 0, $bestIndex$0$lcssa$i33 = 0;
 var $bestIndex$0$lcssa$i34 = 0, $bestIndex$03$i = 0, $bestIndex$03$i11 = 0, $bestIndex$1$i = 0, $bestIndex$1$i$lcssa = 0, $bestIndex$1$i15 = 0, $bestIndex$1$i15$lcssa = 0, $bestValue$01$i = 0.0, $bestValue$01$i13 = 0.0, $bestValue$1$i = 0.0, $bestValue$1$i14 = 0.0, $d$sroa$0$0 = 0, $d$sroa$9$0 = 0, $exitcond = 0, $exitcond48 = 0, $exitcond49 = 0, $i$01$i = 0, $i$02$i = 0, $i$02$i12 = 0, $i$043 = 0;
 var $i1$044 = 0, $iter$0 = 0, $iter$2 = 0, $normal$sroa$0$0 = 0.0, $normal$sroa$7$0 = 0.0, $or$cond5 = 0, $saveA = 0, $saveB = 0, $simplex = 0, $transformA = 0, $transformB = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $transformA = sp + 152|0;
 $transformB = sp + 136|0;
 $simplex = sp;
 $saveA = sp + 124|0;
 $saveB = sp + 112|0;
 $0 = HEAP32[158036]|0; //@line 448 "libraries/Box2D/Collision/b2Distance.cpp"
 $1 = (($0) + 1)|0; //@line 448 "libraries/Box2D/Collision/b2Distance.cpp"
 HEAP32[158036] = $1; //@line 448 "libraries/Box2D/Collision/b2Distance.cpp"
 $2 = ((($input)) + 28|0); //@line 451 "libraries/Box2D/Collision/b2Distance.cpp"
 $3 = ((($input)) + 56|0); //@line 453 "libraries/Box2D/Collision/b2Distance.cpp"
 ;HEAP32[$transformA>>2]=HEAP32[$3>>2]|0;HEAP32[$transformA+4>>2]=HEAP32[$3+4>>2]|0;HEAP32[$transformA+8>>2]=HEAP32[$3+8>>2]|0;HEAP32[$transformA+12>>2]=HEAP32[$3+12>>2]|0; //@line 453 "libraries/Box2D/Collision/b2Distance.cpp"
 $4 = ((($input)) + 72|0); //@line 454 "libraries/Box2D/Collision/b2Distance.cpp"
 ;HEAP32[$transformB>>2]=HEAP32[$4>>2]|0;HEAP32[$transformB+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$transformB+8>>2]=HEAP32[$4+8>>2]|0;HEAP32[$transformB+12>>2]=HEAP32[$4+12>>2]|0; //@line 454 "libraries/Box2D/Collision/b2Distance.cpp"
 __ZN9b2Simplex9ReadCacheEPK14b2SimplexCachePK15b2DistanceProxyRK11b2TransformS5_S8_($simplex,$cache,$input,$transformA,$2,$transformB); //@line 458 "libraries/Box2D/Collision/b2Distance.cpp"
 $5 = ((($simplex)) + 108|0); //@line 477 "libraries/Box2D/Collision/b2Distance.cpp"
 $6 = ((($transformA)) + 12|0); //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($transformA)) + 8|0); //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($input)) + 16|0); //@line 110 "./libraries/Box2D/Collision/b2Distance.h"
 $9 = ((($input)) + 20|0); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
 $10 = ((($transformA)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($transformB)) + 12|0); //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $12 = ((($transformB)) + 8|0); //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($input)) + 44|0); //@line 110 "./libraries/Box2D/Collision/b2Distance.h"
 $14 = ((($input)) + 48|0); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
 $15 = ((($transformB)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $16 = ((($simplex)) + 16|0); //@line 68 "./libraries/Box2D/Common/b2Math.h"
 $17 = ((($simplex)) + 20|0); //@line 68 "./libraries/Box2D/Common/b2Math.h"
 $18 = ((($simplex)) + 52|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($simplex)) + 56|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $20 = ((($simplex)) + 36|0); //@line 299 "libraries/Box2D/Collision/b2Distance.cpp"
 $21 = ((($simplex)) + 24|0); //@line 325 "libraries/Box2D/Collision/b2Distance.cpp"
 $22 = ((($simplex)) + 60|0); //@line 326 "libraries/Box2D/Collision/b2Distance.cpp"
 $$pre = HEAP32[$5>>2]|0; //@line 477 "libraries/Box2D/Collision/b2Distance.cpp"
 $24 = $$pre;$iter$0 = 0;
 L1: while(1) {
  $23 = ($24|0)>(0); //@line 478 "libraries/Box2D/Collision/b2Distance.cpp"
  if ($23) {
   $i$043 = 0;
   while(1) {
    $25 = (((($simplex) + (($i$043*36)|0)|0)) + 28|0); //@line 480 "libraries/Box2D/Collision/b2Distance.cpp"
    $26 = HEAP32[$25>>2]|0; //@line 480 "libraries/Box2D/Collision/b2Distance.cpp"
    $27 = (($saveA) + ($i$043<<2)|0); //@line 480 "libraries/Box2D/Collision/b2Distance.cpp"
    HEAP32[$27>>2] = $26; //@line 480 "libraries/Box2D/Collision/b2Distance.cpp"
    $28 = (((($simplex) + (($i$043*36)|0)|0)) + 32|0); //@line 481 "libraries/Box2D/Collision/b2Distance.cpp"
    $29 = HEAP32[$28>>2]|0; //@line 481 "libraries/Box2D/Collision/b2Distance.cpp"
    $30 = (($saveB) + ($i$043<<2)|0); //@line 481 "libraries/Box2D/Collision/b2Distance.cpp"
    HEAP32[$30>>2] = $29; //@line 481 "libraries/Box2D/Collision/b2Distance.cpp"
    $31 = (($i$043) + 1)|0; //@line 478 "libraries/Box2D/Collision/b2Distance.cpp"
    $exitcond = ($31|0)==($24|0); //@line 478 "libraries/Box2D/Collision/b2Distance.cpp"
    if ($exitcond) {
     break;
    } else {
     $i$043 = $31;
    }
   }
  }
  L7: do {
   switch ($24|0) {
   case 1:  {
    label = 15;
    break;
   }
   case 2:  {
    $32 = +HEAPF32[$16>>2]; //@line 298 "libraries/Box2D/Collision/b2Distance.cpp"
    $33 = +HEAPF32[$17>>2]; //@line 298 "libraries/Box2D/Collision/b2Distance.cpp"
    $34 = +HEAPF32[$18>>2]; //@line 299 "libraries/Box2D/Collision/b2Distance.cpp"
    $35 = +HEAPF32[$19>>2]; //@line 299 "libraries/Box2D/Collision/b2Distance.cpp"
    $36 = $34 - $32; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $37 = $35 - $33; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $38 = $32 * $36; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $39 = $33 * $37; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $40 = $38 + $39; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $41 = !($40 >= -0.0); //@line 304 "libraries/Box2D/Collision/b2Distance.cpp"
    if (!($41)) {
     HEAPF32[$21>>2] = 1.0; //@line 307 "libraries/Box2D/Collision/b2Distance.cpp"
     HEAP32[$5>>2] = 1; //@line 308 "libraries/Box2D/Collision/b2Distance.cpp"
     label = 17;
     break L7;
    }
    $42 = $34 * $36; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $43 = $35 * $37; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $44 = $42 + $43; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $45 = !($44 <= 0.0); //@line 314 "libraries/Box2D/Collision/b2Distance.cpp"
    if ($45) {
     $46 = $44 - $40; //@line 324 "libraries/Box2D/Collision/b2Distance.cpp"
     $47 = 1.0 / $46; //@line 324 "libraries/Box2D/Collision/b2Distance.cpp"
     $48 = $44 * $47; //@line 325 "libraries/Box2D/Collision/b2Distance.cpp"
     HEAPF32[$21>>2] = $48; //@line 325 "libraries/Box2D/Collision/b2Distance.cpp"
     $49 = $40 * $47; //@line 326 "libraries/Box2D/Collision/b2Distance.cpp"
     $50 = -$49; //@line 326 "libraries/Box2D/Collision/b2Distance.cpp"
     HEAPF32[$22>>2] = $50; //@line 326 "libraries/Box2D/Collision/b2Distance.cpp"
     HEAP32[$5>>2] = 2; //@line 327 "libraries/Box2D/Collision/b2Distance.cpp"
     label = 18;
     break L7;
    } else {
     HEAPF32[$22>>2] = 1.0; //@line 317 "libraries/Box2D/Collision/b2Distance.cpp"
     HEAP32[$5>>2] = 1; //@line 318 "libraries/Box2D/Collision/b2Distance.cpp"
     dest=$simplex; src=$20; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 319 "libraries/Box2D/Collision/b2Distance.cpp"
     label = 12;
     break L7;
    }
    break;
   }
   case 3:  {
    __ZN9b2Simplex6Solve3Ev($simplex); //@line 494 "libraries/Box2D/Collision/b2Distance.cpp"
    label = 12;
    break;
   }
   default: {
    __Z10loveAssertbPKc(0,622525); //@line 498 "libraries/Box2D/Collision/b2Distance.cpp"
    label = 12;
   }
   }
  } while(0);
  L17: do {
   if ((label|0) == 12) {
    label = 0;
    $$pr = HEAP32[$5>>2]|0; //@line 502 "libraries/Box2D/Collision/b2Distance.cpp"
    switch ($$pr|0) {
    case 3:  {
     $iter$2 = $iter$0;
     break L1;
     break;
    }
    case 2: case 1:  {
     $51 = $$pr;
     label = 16;
     break L17;
     break;
    }
    case 0:  {
     __Z10loveAssertbPKc(0,622525); //@line 195 "libraries/Box2D/Collision/b2Distance.cpp"
     label = 15;
     break L17;
     break;
    }
    default: {
     __Z10loveAssertbPKc(0,622525); //@line 208 "libraries/Box2D/Collision/b2Distance.cpp"
     label = 15;
     break L17;
    }
    }
   }
  } while(0);
  if ((label|0) == 15) {
   label = 0;
   $$pr50 = HEAP32[$5>>2]|0; //@line 163 "libraries/Box2D/Collision/b2Distance.cpp"
   $51 = $$pr50;
   label = 16;
  }
  L24: do {
   if ((label|0) == 16) {
    label = 0;
    switch ($51|0) {
    case 1:  {
     label = 17;
     break L24;
     break;
    }
    case 2:  {
     label = 18;
     break L24;
     break;
    }
    default: {
    }
    }
    __Z10loveAssertbPKc(0,622525); //@line 185 "libraries/Box2D/Collision/b2Distance.cpp"
    $74 = HEAP32[158039]|0; //@line 186 "libraries/Box2D/Collision/b2Distance.cpp"
    $75 = HEAP32[(632160)>>2]|0; //@line 186 "libraries/Box2D/Collision/b2Distance.cpp"
    $d$sroa$0$0 = $74;$d$sroa$9$0 = $75;
   }
  } while(0);
  do {
   if ((label|0) == 17) {
    label = 0;
    $52 = +HEAPF32[$16>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
    $53 = -$52; //@line 68 "./libraries/Box2D/Common/b2Math.h"
    $54 = +HEAPF32[$17>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
    $55 = -$54; //@line 68 "./libraries/Box2D/Common/b2Math.h"
    $56 = (HEAPF32[tempDoublePtr>>2]=$53,HEAP32[tempDoublePtr>>2]|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $57 = (HEAPF32[tempDoublePtr>>2]=$55,HEAP32[tempDoublePtr>>2]|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $d$sroa$0$0 = $56;$d$sroa$9$0 = $57;
   }
   else if ((label|0) == 18) {
    label = 0;
    $58 = +HEAPF32[$18>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $59 = +HEAPF32[$16>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $60 = $58 - $59; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $61 = +HEAPF32[$19>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $62 = +HEAPF32[$17>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $63 = $61 - $62; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $64 = $60 * $62; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $65 = $59 * $63; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $66 = $65 - $64; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $67 = $66 > 0.0; //@line 172 "libraries/Box2D/Collision/b2Distance.cpp"
    if ($67) {
     $68 = -$63; //@line 428 "./libraries/Box2D/Common/b2Math.h"
     $69 = (HEAPF32[tempDoublePtr>>2]=$68,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
     $70 = (HEAPF32[tempDoublePtr>>2]=$60,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
     $d$sroa$0$0 = $69;$d$sroa$9$0 = $70;
     break;
    } else {
     $71 = -$60; //@line 421 "./libraries/Box2D/Common/b2Math.h"
     $72 = (HEAPF32[tempDoublePtr>>2]=$63,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
     $73 = (HEAPF32[tempDoublePtr>>2]=$71,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
     $d$sroa$0$0 = $72;$d$sroa$9$0 = $73;
     break;
    }
   }
  } while(0);
  $76 = (HEAP32[tempDoublePtr>>2]=$d$sroa$0$0,+HEAPF32[tempDoublePtr>>2]); //@line 110 "./libraries/Box2D/Common/b2Math.h"
  $77 = $76 * $76; //@line 110 "./libraries/Box2D/Common/b2Math.h"
  $78 = (HEAP32[tempDoublePtr>>2]=$d$sroa$9$0,+HEAPF32[tempDoublePtr>>2]); //@line 110 "./libraries/Box2D/Common/b2Math.h"
  $79 = $78 * $78; //@line 110 "./libraries/Box2D/Common/b2Math.h"
  $80 = $77 + $79; //@line 110 "./libraries/Box2D/Common/b2Math.h"
  $81 = $80 < 1.4210854715202004E-14; //@line 522 "libraries/Box2D/Collision/b2Distance.cpp"
  if ($81) {
   $iter$2 = $iter$0;
   break;
  }
  $82 = HEAP32[$5>>2]|0; //@line 534 "libraries/Box2D/Collision/b2Distance.cpp"
  $83 = (($simplex) + (($82*36)|0)|0); //@line 534 "libraries/Box2D/Collision/b2Distance.cpp"
  $84 = -$76; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $85 = -$78; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $86 = +HEAPF32[$6>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $87 = $86 * $84; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $88 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $89 = $88 * $85; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $90 = $87 + $89; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $91 = $88 * $84; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $92 = $86 * $85; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $93 = $92 - $91; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $94 = HEAP32[$8>>2]|0; //@line 113 "./libraries/Box2D/Collision/b2Distance.h"
  $95 = HEAP32[$9>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $96 = ($95|0)>(1); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
  if ($96) {
   $98 = +HEAPF32[$94>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $99 = $90 * $98; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $100 = ((($94)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $101 = +HEAPF32[$100>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $102 = $93 * $101; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $103 = $99 + $102; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $bestIndex$03$i = 0;$bestValue$01$i = $103;$i$02$i = 1;
   while(1) {
    $104 = (($94) + ($i$02$i<<3)|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $105 = +HEAPF32[$104>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $106 = $90 * $105; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $107 = (((($94) + ($i$02$i<<3)|0)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $108 = +HEAPF32[$107>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $109 = $93 * $108; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $110 = $106 + $109; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $111 = $110 > $bestValue$01$i; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $bestValue$1$i = $111 ? $110 : $bestValue$01$i; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $bestIndex$1$i = $111 ? $i$02$i : $bestIndex$03$i; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $112 = (($i$02$i) + 1)|0; //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
    $exitcond48 = ($112|0)==($95|0); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
    if ($exitcond48) {
     $bestIndex$1$i$lcssa = $bestIndex$1$i;
     break;
    } else {
     $bestIndex$03$i = $bestIndex$1$i;$bestValue$01$i = $bestValue$1$i;$i$02$i = $112;
    }
   }
   $113 = (((($simplex) + (($82*36)|0)|0)) + 28|0); //@line 535 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAP32[$113>>2] = $bestIndex$1$i$lcssa; //@line 535 "libraries/Box2D/Collision/b2Distance.cpp"
   $114 = ($bestIndex$1$i$lcssa|0)>(-1); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   if ($114) {
    $389 = $83;$390 = $113;$bestIndex$0$lcssa$i34 = $bestIndex$1$i$lcssa;
    label = 28;
   } else {
    $116 = 0;$184 = $83;$195 = $113;$bestIndex$0$lcssa$i33 = $bestIndex$1$i$lcssa;
   }
  } else {
   $97 = (((($simplex) + (($82*36)|0)|0)) + 28|0); //@line 535 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAP32[$97>>2] = 0; //@line 535 "libraries/Box2D/Collision/b2Distance.cpp"
   $389 = $83;$390 = $97;$bestIndex$0$lcssa$i34 = 0;
   label = 28;
  }
  if ((label|0) == 28) {
   label = 0;
   $115 = ($95|0)>($bestIndex$0$lcssa$i34|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $116 = $115;$184 = $389;$195 = $390;$bestIndex$0$lcssa$i33 = $bestIndex$0$lcssa$i34;
  }
  __Z10loveAssertbPKc($116,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $117 = HEAP32[$8>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $118 = (($117) + ($bestIndex$0$lcssa$i33<<3)|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $119 = +HEAPF32[$118>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $120 = $86 * $119; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $121 = (((($117) + ($bestIndex$0$lcssa$i33<<3)|0)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $122 = +HEAPF32[$121>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $123 = $88 * $122; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $124 = $120 - $123; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $125 = +HEAPF32[$transformA>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $126 = $125 + $124; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $127 = $119 * $88; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $128 = $86 * $122; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $129 = $127 + $128; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $130 = +HEAPF32[$10>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $131 = $129 + $130; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$83>>2] = $126; //@line 536 "libraries/Box2D/Collision/b2Distance.cpp"
  $132 = (((($simplex) + (($82*36)|0)|0)) + 4|0); //@line 536 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$132>>2] = $131; //@line 536 "libraries/Box2D/Collision/b2Distance.cpp"
  $133 = +HEAPF32[$11>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $134 = $76 * $133; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $135 = +HEAPF32[$12>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $136 = $78 * $135; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $137 = $134 + $136; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $138 = $76 * $135; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $139 = $78 * $133; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $140 = $139 - $138; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $141 = HEAP32[$13>>2]|0; //@line 113 "./libraries/Box2D/Collision/b2Distance.h"
  $142 = HEAP32[$14>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $143 = ($142|0)>(1); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
  if ($143) {
   $146 = +HEAPF32[$141>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $147 = $137 * $146; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $148 = ((($141)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $149 = +HEAPF32[$148>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $150 = $140 * $149; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $151 = $147 + $150; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $bestIndex$03$i11 = 0;$bestValue$01$i13 = $151;$i$02$i12 = 1;
   while(1) {
    $152 = (($141) + ($i$02$i12<<3)|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $153 = +HEAPF32[$152>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $154 = $137 * $153; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $155 = (((($141) + ($i$02$i12<<3)|0)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $156 = +HEAPF32[$155>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $157 = $140 * $156; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $158 = $154 + $157; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $159 = $158 > $bestValue$01$i13; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $bestValue$1$i14 = $159 ? $158 : $bestValue$01$i13; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $bestIndex$1$i15 = $159 ? $i$02$i12 : $bestIndex$03$i11; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $160 = (($i$02$i12) + 1)|0; //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
    $exitcond49 = ($160|0)==($142|0); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
    if ($exitcond49) {
     $bestIndex$1$i15$lcssa = $bestIndex$1$i15;
     break;
    } else {
     $bestIndex$03$i11 = $bestIndex$1$i15;$bestValue$01$i13 = $bestValue$1$i14;$i$02$i12 = $160;
    }
   }
   $161 = (((($simplex) + (($82*36)|0)|0)) + 32|0); //@line 538 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAP32[$161>>2] = $bestIndex$1$i15$lcssa; //@line 538 "libraries/Box2D/Collision/b2Distance.cpp"
   $162 = (((($simplex) + (($82*36)|0)|0)) + 8|0); //@line 539 "libraries/Box2D/Collision/b2Distance.cpp"
   $163 = ($bestIndex$1$i15$lcssa|0)>(-1); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   if ($163) {
    $391 = $162;$392 = $161;$bestIndex$0$lcssa$i1035 = $bestIndex$1$i15$lcssa;
    label = 34;
   } else {
    $165 = 0;$181 = $162;$200 = $161;$bestIndex$0$lcssa$i1036 = $bestIndex$1$i15$lcssa;
   }
  } else {
   $144 = (((($simplex) + (($82*36)|0)|0)) + 32|0); //@line 538 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAP32[$144>>2] = 0; //@line 538 "libraries/Box2D/Collision/b2Distance.cpp"
   $145 = (((($simplex) + (($82*36)|0)|0)) + 8|0); //@line 539 "libraries/Box2D/Collision/b2Distance.cpp"
   $391 = $145;$392 = $144;$bestIndex$0$lcssa$i1035 = 0;
   label = 34;
  }
  if ((label|0) == 34) {
   label = 0;
   $164 = ($142|0)>($bestIndex$0$lcssa$i1035|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $165 = $164;$181 = $391;$200 = $392;$bestIndex$0$lcssa$i1036 = $bestIndex$0$lcssa$i1035;
  }
  __Z10loveAssertbPKc($165,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $166 = HEAP32[$13>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $167 = (($166) + ($bestIndex$0$lcssa$i1036<<3)|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $168 = +HEAPF32[$167>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $169 = $133 * $168; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $170 = (((($166) + ($bestIndex$0$lcssa$i1036<<3)|0)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $171 = +HEAPF32[$170>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $172 = $135 * $171; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $173 = $169 - $172; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $174 = +HEAPF32[$transformB>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $175 = $174 + $173; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $176 = $168 * $135; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $177 = $133 * $171; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $178 = $176 + $177; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $179 = +HEAPF32[$15>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $180 = $178 + $179; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$181>>2] = $175; //@line 539 "libraries/Box2D/Collision/b2Distance.cpp"
  $182 = ((($181)) + 4|0); //@line 539 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$182>>2] = $180; //@line 539 "libraries/Box2D/Collision/b2Distance.cpp"
  $183 = +HEAPF32[$184>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $185 = $175 - $183; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $186 = ((($184)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $187 = +HEAPF32[$186>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $188 = $180 - $187; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $189 = (((($simplex) + (($82*36)|0)|0)) + 16|0); //@line 540 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$189>>2] = $185; //@line 540 "libraries/Box2D/Collision/b2Distance.cpp"
  $190 = (((($simplex) + (($82*36)|0)|0)) + 20|0); //@line 540 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$190>>2] = $188; //@line 540 "libraries/Box2D/Collision/b2Distance.cpp"
  $191 = (($iter$0) + 1)|0; //@line 543 "libraries/Box2D/Collision/b2Distance.cpp"
  $192 = HEAP32[158037]|0; //@line 544 "libraries/Box2D/Collision/b2Distance.cpp"
  $193 = (($192) + 1)|0; //@line 544 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[158037] = $193; //@line 544 "libraries/Box2D/Collision/b2Distance.cpp"
  if ($23) {
   $194 = HEAP32[$195>>2]|0; //@line 550 "libraries/Box2D/Collision/b2Distance.cpp"
   $i1$044 = 0;
   while(1) {
    $196 = (($saveA) + ($i1$044<<2)|0); //@line 550 "libraries/Box2D/Collision/b2Distance.cpp"
    $197 = HEAP32[$196>>2]|0; //@line 550 "libraries/Box2D/Collision/b2Distance.cpp"
    $198 = ($194|0)==($197|0); //@line 550 "libraries/Box2D/Collision/b2Distance.cpp"
    if ($198) {
     $199 = HEAP32[$200>>2]|0; //@line 550 "libraries/Box2D/Collision/b2Distance.cpp"
     $201 = (($saveB) + ($i1$044<<2)|0); //@line 550 "libraries/Box2D/Collision/b2Distance.cpp"
     $202 = HEAP32[$201>>2]|0; //@line 550 "libraries/Box2D/Collision/b2Distance.cpp"
     $203 = ($199|0)==($202|0); //@line 550 "libraries/Box2D/Collision/b2Distance.cpp"
     if ($203) {
      $iter$2 = $191;
      break L1;
     }
    }
    $204 = (($i1$044) + 1)|0; //@line 548 "libraries/Box2D/Collision/b2Distance.cpp"
    $205 = ($204|0)<($24|0); //@line 548 "libraries/Box2D/Collision/b2Distance.cpp"
    if ($205) {
     $i1$044 = $204;
    } else {
     break;
    }
   }
  }
  $206 = HEAP32[$5>>2]|0; //@line 564 "libraries/Box2D/Collision/b2Distance.cpp"
  $207 = (($206) + 1)|0; //@line 564 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$5>>2] = $207; //@line 564 "libraries/Box2D/Collision/b2Distance.cpp"
  $208 = ($191|0)<(20); //@line 474 "libraries/Box2D/Collision/b2Distance.cpp"
  if ($208) {
   $24 = $207;$iter$0 = $191;
  } else {
   $iter$2 = $191;
   break;
  }
 }
 $209 = HEAP32[158038]|0; //@line 567 "libraries/Box2D/Collision/b2Distance.cpp"
 $210 = ($209|0)>($iter$2|0); //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $211 = $210 ? $209 : $iter$2; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 HEAP32[158038] = $211; //@line 567 "libraries/Box2D/Collision/b2Distance.cpp"
 $212 = ((($output)) + 8|0); //@line 570 "libraries/Box2D/Collision/b2Distance.cpp"
 $213 = HEAP32[$5>>2]|0; //@line 215 "libraries/Box2D/Collision/b2Distance.cpp"
 switch ($213|0) {
 case 0:  {
  __Z10loveAssertbPKc(0,622525); //@line 218 "libraries/Box2D/Collision/b2Distance.cpp"
  break;
 }
 case 1:  {
  $214 = $simplex; //@line 222 "libraries/Box2D/Collision/b2Distance.cpp"
  $215 = $214; //@line 222 "libraries/Box2D/Collision/b2Distance.cpp"
  $216 = HEAP32[$215>>2]|0; //@line 222 "libraries/Box2D/Collision/b2Distance.cpp"
  $217 = (($214) + 4)|0; //@line 222 "libraries/Box2D/Collision/b2Distance.cpp"
  $218 = $217; //@line 222 "libraries/Box2D/Collision/b2Distance.cpp"
  $219 = HEAP32[$218>>2]|0; //@line 222 "libraries/Box2D/Collision/b2Distance.cpp"
  $220 = $output; //@line 222 "libraries/Box2D/Collision/b2Distance.cpp"
  $221 = $220; //@line 222 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$221>>2] = $216; //@line 222 "libraries/Box2D/Collision/b2Distance.cpp"
  $222 = (($220) + 4)|0; //@line 222 "libraries/Box2D/Collision/b2Distance.cpp"
  $223 = $222; //@line 222 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$223>>2] = $219; //@line 222 "libraries/Box2D/Collision/b2Distance.cpp"
  $224 = ((($simplex)) + 8|0); //@line 223 "libraries/Box2D/Collision/b2Distance.cpp"
  $225 = $224; //@line 223 "libraries/Box2D/Collision/b2Distance.cpp"
  $226 = $225; //@line 223 "libraries/Box2D/Collision/b2Distance.cpp"
  $227 = HEAP32[$226>>2]|0; //@line 223 "libraries/Box2D/Collision/b2Distance.cpp"
  $228 = (($225) + 4)|0; //@line 223 "libraries/Box2D/Collision/b2Distance.cpp"
  $229 = $228; //@line 223 "libraries/Box2D/Collision/b2Distance.cpp"
  $230 = HEAP32[$229>>2]|0; //@line 223 "libraries/Box2D/Collision/b2Distance.cpp"
  $231 = $212; //@line 223 "libraries/Box2D/Collision/b2Distance.cpp"
  $232 = $231; //@line 223 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$232>>2] = $227; //@line 223 "libraries/Box2D/Collision/b2Distance.cpp"
  $233 = (($231) + 4)|0; //@line 223 "libraries/Box2D/Collision/b2Distance.cpp"
  $234 = $233; //@line 223 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAP32[$234>>2] = $230; //@line 223 "libraries/Box2D/Collision/b2Distance.cpp"
  break;
 }
 case 2:  {
  $235 = +HEAPF32[$21>>2]; //@line 227 "libraries/Box2D/Collision/b2Distance.cpp"
  $236 = +HEAPF32[$simplex>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $237 = $235 * $236; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $238 = ((($simplex)) + 4|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $239 = +HEAPF32[$238>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $240 = $235 * $239; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $241 = +HEAPF32[$22>>2]; //@line 227 "libraries/Box2D/Collision/b2Distance.cpp"
  $242 = ((($simplex)) + 36|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $243 = +HEAPF32[$242>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $244 = $241 * $243; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $245 = ((($simplex)) + 40|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $246 = +HEAPF32[$245>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $247 = $241 * $246; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $248 = $237 + $244; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $249 = $240 + $247; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$output>>2] = $248; //@line 227 "libraries/Box2D/Collision/b2Distance.cpp"
  $250 = ((($output)) + 4|0); //@line 227 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$250>>2] = $249; //@line 227 "libraries/Box2D/Collision/b2Distance.cpp"
  $251 = ((($simplex)) + 8|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $252 = +HEAPF32[$251>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $253 = $235 * $252; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $254 = ((($simplex)) + 12|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $255 = +HEAPF32[$254>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $256 = $235 * $255; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $257 = ((($simplex)) + 44|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $258 = +HEAPF32[$257>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $259 = $241 * $258; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $260 = ((($simplex)) + 48|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $261 = +HEAPF32[$260>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $262 = $241 * $261; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $263 = $253 + $259; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $264 = $256 + $262; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$212>>2] = $263; //@line 228 "libraries/Box2D/Collision/b2Distance.cpp"
  $265 = ((($output)) + 12|0); //@line 228 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$265>>2] = $264; //@line 228 "libraries/Box2D/Collision/b2Distance.cpp"
  break;
 }
 case 3:  {
  $266 = +HEAPF32[$21>>2]; //@line 232 "libraries/Box2D/Collision/b2Distance.cpp"
  $267 = +HEAPF32[$simplex>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $268 = $266 * $267; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $269 = ((($simplex)) + 4|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $270 = +HEAPF32[$269>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $271 = $266 * $270; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $272 = +HEAPF32[$22>>2]; //@line 232 "libraries/Box2D/Collision/b2Distance.cpp"
  $273 = ((($simplex)) + 36|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $274 = +HEAPF32[$273>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $275 = $272 * $274; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $276 = ((($simplex)) + 40|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $277 = +HEAPF32[$276>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $278 = $272 * $277; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $279 = $268 + $275; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $280 = $271 + $278; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $281 = ((($simplex)) + 96|0); //@line 232 "libraries/Box2D/Collision/b2Distance.cpp"
  $282 = +HEAPF32[$281>>2]; //@line 232 "libraries/Box2D/Collision/b2Distance.cpp"
  $283 = ((($simplex)) + 72|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $284 = +HEAPF32[$283>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $285 = $282 * $284; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $286 = ((($simplex)) + 76|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $287 = +HEAPF32[$286>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $288 = $282 * $287; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $289 = $279 + $285; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $290 = $280 + $288; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$output>>2] = $289; //@line 232 "libraries/Box2D/Collision/b2Distance.cpp"
  $291 = ((($output)) + 4|0); //@line 232 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$291>>2] = $290; //@line 232 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$212>>2] = $289; //@line 233 "libraries/Box2D/Collision/b2Distance.cpp"
  $292 = ((($output)) + 12|0); //@line 233 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$292>>2] = $290; //@line 233 "libraries/Box2D/Collision/b2Distance.cpp"
  break;
 }
 default: {
  __Z10loveAssertbPKc(0,622525); //@line 237 "libraries/Box2D/Collision/b2Distance.cpp"
 }
 }
 $293 = +HEAPF32[$output>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $294 = +HEAPF32[$212>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $295 = $293 - $294; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $296 = ((($output)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $297 = +HEAPF32[$296>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $298 = ((($output)) + 12|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $299 = +HEAPF32[$298>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $300 = $297 - $299; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $301 = $295 * $295; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $302 = $300 * $300; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $303 = $301 + $302; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $304 = (+Math_sqrt((+$303))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $305 = ((($output)) + 16|0); //@line 571 "libraries/Box2D/Collision/b2Distance.cpp"
 HEAPF32[$305>>2] = $304; //@line 571 "libraries/Box2D/Collision/b2Distance.cpp"
 $306 = ((($output)) + 20|0); //@line 572 "libraries/Box2D/Collision/b2Distance.cpp"
 HEAP32[$306>>2] = $iter$2; //@line 572 "libraries/Box2D/Collision/b2Distance.cpp"
 $307 = HEAP32[$5>>2]|0; //@line 244 "libraries/Box2D/Collision/b2Distance.cpp"
 switch ($307|0) {
 case 0:  {
  __Z10loveAssertbPKc(0,622525); //@line 247 "libraries/Box2D/Collision/b2Distance.cpp"
  $$0$i6 = 0.0;
  break;
 }
 case 1:  {
  $$0$i6 = 0.0;
  break;
 }
 case 2:  {
  $308 = +HEAPF32[$16>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $309 = +HEAPF32[$18>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $310 = $308 - $309; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $311 = +HEAPF32[$17>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $312 = +HEAPF32[$19>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $313 = $311 - $312; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $314 = $310 * $310; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $315 = $313 * $313; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $316 = $314 + $315; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $317 = (+Math_sqrt((+$316))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $$0$i6 = $317;
  break;
 }
 case 3:  {
  $318 = +HEAPF32[$18>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $319 = +HEAPF32[$16>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $320 = $318 - $319; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $321 = +HEAPF32[$19>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $322 = +HEAPF32[$17>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $323 = $321 - $322; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $324 = ((($simplex)) + 88|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $325 = +HEAPF32[$324>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $326 = $325 - $319; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $327 = ((($simplex)) + 92|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $328 = +HEAPF32[$327>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $329 = $328 - $322; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $330 = $320 * $329; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $331 = $323 * $326; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $332 = $330 - $331; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $$0$i6 = $332;
  break;
 }
 default: {
  __Z10loveAssertbPKc(0,622525); //@line 260 "libraries/Box2D/Collision/b2Distance.cpp"
  $$0$i6 = 0.0;
 }
 }
 HEAPF32[$cache>>2] = $$0$i6; //@line 151 "libraries/Box2D/Collision/b2Distance.cpp"
 $333 = HEAP32[$5>>2]|0; //@line 152 "libraries/Box2D/Collision/b2Distance.cpp"
 $334 = $333&65535; //@line 152 "libraries/Box2D/Collision/b2Distance.cpp"
 $335 = ((($cache)) + 4|0); //@line 152 "libraries/Box2D/Collision/b2Distance.cpp"
 HEAP16[$335>>1] = $334; //@line 152 "libraries/Box2D/Collision/b2Distance.cpp"
 $336 = ($333|0)>(0); //@line 154 "libraries/Box2D/Collision/b2Distance.cpp"
 if ($336) {
  $i$01$i = 0;
  while(1) {
   $337 = (((($simplex) + (($i$01$i*36)|0)|0)) + 28|0); //@line 156 "libraries/Box2D/Collision/b2Distance.cpp"
   $338 = HEAP32[$337>>2]|0; //@line 156 "libraries/Box2D/Collision/b2Distance.cpp"
   $339 = $338&255; //@line 156 "libraries/Box2D/Collision/b2Distance.cpp"
   $340 = (((($cache)) + 6|0) + ($i$01$i)|0); //@line 156 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAP8[$340>>0] = $339; //@line 156 "libraries/Box2D/Collision/b2Distance.cpp"
   $341 = (((($simplex) + (($i$01$i*36)|0)|0)) + 32|0); //@line 157 "libraries/Box2D/Collision/b2Distance.cpp"
   $342 = HEAP32[$341>>2]|0; //@line 157 "libraries/Box2D/Collision/b2Distance.cpp"
   $343 = $342&255; //@line 157 "libraries/Box2D/Collision/b2Distance.cpp"
   $344 = (((($cache)) + 9|0) + ($i$01$i)|0); //@line 157 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAP8[$344>>0] = $343; //@line 157 "libraries/Box2D/Collision/b2Distance.cpp"
   $345 = (($i$01$i) + 1)|0; //@line 154 "libraries/Box2D/Collision/b2Distance.cpp"
   $346 = ($345|0)<($333|0); //@line 154 "libraries/Box2D/Collision/b2Distance.cpp"
   if ($346) {
    $i$01$i = $345;
   } else {
    break;
   }
  }
 }
 $347 = ((($input)) + 88|0); //@line 578 "libraries/Box2D/Collision/b2Distance.cpp"
 $348 = HEAP8[$347>>0]|0; //@line 578 "libraries/Box2D/Collision/b2Distance.cpp"
 $349 = ($348<<24>>24)==(0); //@line 578 "libraries/Box2D/Collision/b2Distance.cpp"
 if ($349) {
  STACKTOP = sp;return; //@line 603 "libraries/Box2D/Collision/b2Distance.cpp"
 }
 $350 = ((($input)) + 24|0); //@line 580 "libraries/Box2D/Collision/b2Distance.cpp"
 $351 = +HEAPF32[$350>>2]; //@line 580 "libraries/Box2D/Collision/b2Distance.cpp"
 $352 = ((($input)) + 52|0); //@line 581 "libraries/Box2D/Collision/b2Distance.cpp"
 $353 = +HEAPF32[$352>>2]; //@line 581 "libraries/Box2D/Collision/b2Distance.cpp"
 $354 = +HEAPF32[$305>>2]; //@line 583 "libraries/Box2D/Collision/b2Distance.cpp"
 $355 = $351 + $353; //@line 583 "libraries/Box2D/Collision/b2Distance.cpp"
 $356 = $354 > $355; //@line 583 "libraries/Box2D/Collision/b2Distance.cpp"
 $357 = $354 > 1.1920928955078125E-7; //@line 583 "libraries/Box2D/Collision/b2Distance.cpp"
 $or$cond5 = $356 & $357; //@line 583 "libraries/Box2D/Collision/b2Distance.cpp"
 if (!($or$cond5)) {
  $381 = +HEAPF32[$output>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $382 = +HEAPF32[$212>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $383 = $381 + $382; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $384 = +HEAPF32[$296>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $385 = +HEAPF32[$298>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $386 = $384 + $385; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $387 = $383 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $388 = $386 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$output>>2] = $387; //@line 598 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$296>>2] = $388; //@line 598 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$212>>2] = $387; //@line 599 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$298>>2] = $388; //@line 599 "libraries/Box2D/Collision/b2Distance.cpp"
  HEAPF32[$305>>2] = 0.0; //@line 600 "libraries/Box2D/Collision/b2Distance.cpp"
  STACKTOP = sp;return; //@line 603 "libraries/Box2D/Collision/b2Distance.cpp"
 }
 $358 = $354 - $355; //@line 587 "libraries/Box2D/Collision/b2Distance.cpp"
 HEAPF32[$305>>2] = $358; //@line 587 "libraries/Box2D/Collision/b2Distance.cpp"
 $359 = +HEAPF32[$212>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $360 = +HEAPF32[$output>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $361 = $359 - $360; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $362 = +HEAPF32[$298>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $363 = +HEAPF32[$296>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $364 = $362 - $363; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $365 = $361 * $361; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $366 = $364 * $364; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $367 = $365 + $366; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $368 = (+Math_sqrt((+$367))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $369 = $368 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
 if ($369) {
  $normal$sroa$0$0 = $361;$normal$sroa$7$0 = $364;
 } else {
  $370 = 1.0 / $368; //@line 121 "./libraries/Box2D/Common/b2Math.h"
  $371 = $361 * $370; //@line 122 "./libraries/Box2D/Common/b2Math.h"
  $372 = $364 * $370; //@line 123 "./libraries/Box2D/Common/b2Math.h"
  $normal$sroa$0$0 = $371;$normal$sroa$7$0 = $372;
 }
 $373 = $351 * $normal$sroa$0$0; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $374 = $351 * $normal$sroa$7$0; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $375 = $373 + $360; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$output>>2] = $375; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $376 = $374 + $363; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$296>>2] = $376; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $377 = $353 * $normal$sroa$0$0; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $378 = $353 * $normal$sroa$7$0; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $379 = $359 - $377; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$212>>2] = $379; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $380 = $362 - $378; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$298>>2] = $380; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 STACKTOP = sp;return; //@line 603 "libraries/Box2D/Collision/b2Distance.cpp"
}
function __ZN9b2Simplex9ReadCacheEPK14b2SimplexCachePK15b2DistanceProxyRK11b2TransformS5_S8_($this,$cache,$proxyA,$transformA,$proxyB,$transformB) {
 $this = $this|0;
 $cache = $cache|0;
 $proxyA = $proxyA|0;
 $transformA = $transformA|0;
 $proxyB = $proxyB|0;
 $transformB = $transformB|0;
 var $$0$i = 0.0, $$lcssa = 0, $$pr = 0, $$pr22 = 0, $$pr22$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0.0, $103 = 0, $104 = 0.0, $105 = 0.0, $106 = 0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0;
 var $111 = 0.0, $112 = 0.0, $113 = 0, $114 = 0.0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0.0, $126 = 0, $127 = 0.0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0.0, $134 = 0, $135 = 0.0, $136 = 0, $137 = 0.0, $138 = 0.0, $139 = 0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0.0;
 var $148 = 0, $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0, $152 = 0, $153 = 0.0, $154 = 0.0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0, $165 = 0.0;
 var $166 = 0.0, $167 = 0, $168 = 0, $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0, $172 = 0, $173 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0.0;
 var $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0, $56 = 0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0;
 var $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0.0, $79 = 0, $8 = 0, $80 = 0.0, $81 = 0;
 var $82 = 0.0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0, $99 = 0.0, $i$023 = 0;
 var $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cache)) + 4|0); //@line 102 "libraries/Box2D/Collision/b2Distance.cpp"
 $1 = HEAP16[$0>>1]|0; //@line 102 "libraries/Box2D/Collision/b2Distance.cpp"
 $2 = ($1&65535)<(4); //@line 102 "libraries/Box2D/Collision/b2Distance.cpp"
 __Z10loveAssertbPKc($2,360969); //@line 102 "libraries/Box2D/Collision/b2Distance.cpp"
 $3 = HEAP16[$0>>1]|0; //@line 105 "libraries/Box2D/Collision/b2Distance.cpp"
 $4 = $3&65535; //@line 105 "libraries/Box2D/Collision/b2Distance.cpp"
 $5 = ((($this)) + 108|0); //@line 105 "libraries/Box2D/Collision/b2Distance.cpp"
 HEAP32[$5>>2] = $4; //@line 105 "libraries/Box2D/Collision/b2Distance.cpp"
 $6 = ($3<<16>>16)==(0); //@line 107 "libraries/Box2D/Collision/b2Distance.cpp"
 if ($6) {
  $$lcssa = $4;
 } else {
  $7 = ((($proxyA)) + 20|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $8 = ((($proxyA)) + 16|0); //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $9 = ((($proxyB)) + 20|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $10 = ((($proxyB)) + 16|0); //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $11 = ((($transformA)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $12 = ((($transformA)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $13 = ((($transformA)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $14 = ((($transformB)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $15 = ((($transformB)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $16 = ((($transformB)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $i$023 = 0;
  while(1) {
   $18 = (((($cache)) + 6|0) + ($i$023)|0); //@line 110 "libraries/Box2D/Collision/b2Distance.cpp"
   $19 = HEAP8[$18>>0]|0; //@line 110 "libraries/Box2D/Collision/b2Distance.cpp"
   $20 = $19&255; //@line 110 "libraries/Box2D/Collision/b2Distance.cpp"
   $21 = (((($this) + (($i$023*36)|0)|0)) + 28|0); //@line 110 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAP32[$21>>2] = $20; //@line 110 "libraries/Box2D/Collision/b2Distance.cpp"
   $22 = (((($cache)) + 9|0) + ($i$023)|0); //@line 111 "libraries/Box2D/Collision/b2Distance.cpp"
   $23 = HEAP8[$22>>0]|0; //@line 111 "libraries/Box2D/Collision/b2Distance.cpp"
   $24 = $23&255; //@line 111 "libraries/Box2D/Collision/b2Distance.cpp"
   $25 = (((($this) + (($i$023*36)|0)|0)) + 32|0); //@line 111 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAP32[$25>>2] = $24; //@line 111 "libraries/Box2D/Collision/b2Distance.cpp"
   $26 = HEAP32[$7>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $27 = ($26|0)>($20|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   __Z10loveAssertbPKc($27,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $28 = HEAP32[$8>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
   $29 = (($28) + ($20<<3)|0); //@line 112 "libraries/Box2D/Collision/b2Distance.cpp"
   $30 = +HEAPF32[$29>>2]; //@line 112 "libraries/Box2D/Collision/b2Distance.cpp"
   $31 = (((($28) + ($20<<3)|0)) + 4|0); //@line 112 "libraries/Box2D/Collision/b2Distance.cpp"
   $32 = +HEAPF32[$31>>2]; //@line 112 "libraries/Box2D/Collision/b2Distance.cpp"
   $33 = HEAP32[$25>>2]|0; //@line 113 "libraries/Box2D/Collision/b2Distance.cpp"
   $34 = ($33|0)>(-1); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   if ($34) {
    $35 = HEAP32[$9>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
    $36 = ($35|0)>($33|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
    $37 = $36;
   } else {
    $37 = 0;
   }
   __Z10loveAssertbPKc($37,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $38 = HEAP32[$10>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
   $39 = (($38) + ($33<<3)|0); //@line 113 "libraries/Box2D/Collision/b2Distance.cpp"
   $40 = +HEAPF32[$39>>2]; //@line 113 "libraries/Box2D/Collision/b2Distance.cpp"
   $41 = (((($38) + ($33<<3)|0)) + 4|0); //@line 113 "libraries/Box2D/Collision/b2Distance.cpp"
   $42 = +HEAPF32[$41>>2]; //@line 113 "libraries/Box2D/Collision/b2Distance.cpp"
   $43 = +HEAPF32[$11>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $44 = $30 * $43; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $45 = +HEAPF32[$12>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $46 = $32 * $45; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $47 = $44 - $46; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $48 = +HEAPF32[$transformA>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $49 = $48 + $47; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $50 = $30 * $45; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $51 = $32 * $43; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $52 = $51 + $50; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $53 = +HEAPF32[$13>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $54 = $52 + $53; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $55 = (($this) + (($i$023*36)|0)|0); //@line 114 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAPF32[$55>>2] = $49; //@line 114 "libraries/Box2D/Collision/b2Distance.cpp"
   $56 = (((($this) + (($i$023*36)|0)|0)) + 4|0); //@line 114 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAPF32[$56>>2] = $54; //@line 114 "libraries/Box2D/Collision/b2Distance.cpp"
   $57 = +HEAPF32[$14>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $58 = $40 * $57; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $59 = +HEAPF32[$15>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $60 = $42 * $59; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $61 = $58 - $60; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $62 = +HEAPF32[$transformB>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $63 = $62 + $61; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $64 = $40 * $59; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $65 = $42 * $57; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $66 = $65 + $64; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $67 = +HEAPF32[$16>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $68 = $66 + $67; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $69 = (((($this) + (($i$023*36)|0)|0)) + 8|0); //@line 115 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAPF32[$69>>2] = $63; //@line 115 "libraries/Box2D/Collision/b2Distance.cpp"
   $70 = (((($this) + (($i$023*36)|0)|0)) + 12|0); //@line 115 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAPF32[$70>>2] = $68; //@line 115 "libraries/Box2D/Collision/b2Distance.cpp"
   $71 = $63 - $49; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $72 = $68 - $54; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $73 = (((($this) + (($i$023*36)|0)|0)) + 16|0); //@line 116 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAPF32[$73>>2] = $71; //@line 116 "libraries/Box2D/Collision/b2Distance.cpp"
   $74 = (((($this) + (($i$023*36)|0)|0)) + 20|0); //@line 116 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAPF32[$74>>2] = $72; //@line 116 "libraries/Box2D/Collision/b2Distance.cpp"
   $75 = (((($this) + (($i$023*36)|0)|0)) + 24|0); //@line 117 "libraries/Box2D/Collision/b2Distance.cpp"
   HEAPF32[$75>>2] = 0.0; //@line 117 "libraries/Box2D/Collision/b2Distance.cpp"
   $76 = (($i$023) + 1)|0; //@line 107 "libraries/Box2D/Collision/b2Distance.cpp"
   $$pr = HEAP32[$5>>2]|0; //@line 107 "libraries/Box2D/Collision/b2Distance.cpp"
   $77 = ($76|0)<($$pr|0); //@line 107 "libraries/Box2D/Collision/b2Distance.cpp"
   if ($77) {
    $i$023 = $76;
   } else {
    $$lcssa = $$pr;
    break;
   }
  }
 }
 $17 = ($$lcssa|0)>(1); //@line 122 "libraries/Box2D/Collision/b2Distance.cpp"
 do {
  if ($17) {
   $78 = +HEAPF32[$cache>>2]; //@line 124 "libraries/Box2D/Collision/b2Distance.cpp"
   switch ($$lcssa|0) {
   case 2:  {
    $79 = ((($this)) + 16|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $80 = +HEAPF32[$79>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $81 = ((($this)) + 52|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $82 = +HEAPF32[$81>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $83 = $80 - $82; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $84 = ((($this)) + 20|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $85 = +HEAPF32[$84>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $86 = ((($this)) + 56|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $87 = +HEAPF32[$86>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $88 = $85 - $87; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $89 = $83 * $83; //@line 103 "./libraries/Box2D/Common/b2Math.h"
    $90 = $88 * $88; //@line 103 "./libraries/Box2D/Common/b2Math.h"
    $91 = $89 + $90; //@line 103 "./libraries/Box2D/Common/b2Math.h"
    $92 = (+Math_sqrt((+$91))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
    $$0$i = $92;
    break;
   }
   case 3:  {
    $93 = ((($this)) + 52|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $94 = +HEAPF32[$93>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $95 = ((($this)) + 16|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $96 = +HEAPF32[$95>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $97 = $94 - $96; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $98 = ((($this)) + 56|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $99 = +HEAPF32[$98>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $100 = ((($this)) + 20|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $101 = +HEAPF32[$100>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $102 = $99 - $101; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $103 = ((($this)) + 88|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $104 = +HEAPF32[$103>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $105 = $104 - $96; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $106 = ((($this)) + 92|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $107 = +HEAPF32[$106>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $108 = $107 - $101; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $109 = $97 * $108; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $110 = $102 * $105; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $111 = $109 - $110; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $$0$i = $111;
    break;
   }
   default: {
    __Z10loveAssertbPKc(0,622525); //@line 260 "libraries/Box2D/Collision/b2Distance.cpp"
    $$0$i = 0.0;
   }
   }
   $112 = $78 * 0.5; //@line 126 "libraries/Box2D/Collision/b2Distance.cpp"
   $113 = $$0$i < $112; //@line 126 "libraries/Box2D/Collision/b2Distance.cpp"
   if (!($113)) {
    $114 = $78 * 2.0; //@line 126 "libraries/Box2D/Collision/b2Distance.cpp"
    $115 = $114 < $$0$i; //@line 126 "libraries/Box2D/Collision/b2Distance.cpp"
    $116 = $$0$i < 1.1920928955078125E-7; //@line 126 "libraries/Box2D/Collision/b2Distance.cpp"
    $or$cond = $115 | $116; //@line 126 "libraries/Box2D/Collision/b2Distance.cpp"
    if (!($or$cond)) {
     $$pr22$pre = HEAP32[$5>>2]|0; //@line 134 "libraries/Box2D/Collision/b2Distance.cpp"
     $$pr22 = $$pr22$pre;
     label = 15;
     break;
    }
   }
   HEAP32[$5>>2] = 0; //@line 129 "libraries/Box2D/Collision/b2Distance.cpp"
  } else {
   $$pr22 = $$lcssa;
   label = 15;
  }
 } while(0);
 if ((label|0) == 15) {
  $117 = ($$pr22|0)==(0); //@line 134 "libraries/Box2D/Collision/b2Distance.cpp"
  if (!($117)) {
   return; //@line 147 "libraries/Box2D/Collision/b2Distance.cpp"
  }
 }
 $118 = ((($this)) + 28|0); //@line 137 "libraries/Box2D/Collision/b2Distance.cpp"
 HEAP32[$118>>2] = 0; //@line 137 "libraries/Box2D/Collision/b2Distance.cpp"
 $119 = ((($this)) + 32|0); //@line 138 "libraries/Box2D/Collision/b2Distance.cpp"
 HEAP32[$119>>2] = 0; //@line 138 "libraries/Box2D/Collision/b2Distance.cpp"
 $120 = ((($proxyA)) + 20|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
 $121 = HEAP32[$120>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
 $122 = ($121|0)>(0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
 __Z10loveAssertbPKc($122,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
 $123 = ((($proxyA)) + 16|0); //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
 $124 = HEAP32[$123>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
 $125 = +HEAPF32[$124>>2]; //@line 139 "libraries/Box2D/Collision/b2Distance.cpp"
 $126 = ((($124)) + 4|0); //@line 139 "libraries/Box2D/Collision/b2Distance.cpp"
 $127 = +HEAPF32[$126>>2]; //@line 139 "libraries/Box2D/Collision/b2Distance.cpp"
 $128 = ((($proxyB)) + 20|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
 $129 = HEAP32[$128>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
 $130 = ($129|0)>(0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
 __Z10loveAssertbPKc($130,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
 $131 = ((($proxyB)) + 16|0); //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
 $132 = HEAP32[$131>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
 $133 = +HEAPF32[$132>>2]; //@line 140 "libraries/Box2D/Collision/b2Distance.cpp"
 $134 = ((($132)) + 4|0); //@line 140 "libraries/Box2D/Collision/b2Distance.cpp"
 $135 = +HEAPF32[$134>>2]; //@line 140 "libraries/Box2D/Collision/b2Distance.cpp"
 $136 = ((($transformA)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $137 = +HEAPF32[$136>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $138 = $125 * $137; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $139 = ((($transformA)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $140 = +HEAPF32[$139>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $141 = $127 * $140; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $142 = $138 - $141; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $143 = +HEAPF32[$transformA>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $144 = $143 + $142; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $145 = $125 * $140; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $146 = $127 * $137; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $147 = $146 + $145; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $148 = ((($transformA)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $149 = +HEAPF32[$148>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $150 = $147 + $149; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$this>>2] = $144; //@line 141 "libraries/Box2D/Collision/b2Distance.cpp"
 $151 = ((($this)) + 4|0); //@line 141 "libraries/Box2D/Collision/b2Distance.cpp"
 HEAPF32[$151>>2] = $150; //@line 141 "libraries/Box2D/Collision/b2Distance.cpp"
 $152 = ((($transformB)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $153 = +HEAPF32[$152>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $154 = $133 * $153; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $155 = ((($transformB)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $156 = +HEAPF32[$155>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $157 = $135 * $156; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $158 = $154 - $157; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $159 = +HEAPF32[$transformB>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $160 = $159 + $158; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $161 = $133 * $156; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $162 = $135 * $153; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $163 = $162 + $161; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $164 = ((($transformB)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $165 = +HEAPF32[$164>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $166 = $163 + $165; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $167 = ((($this)) + 8|0); //@line 142 "libraries/Box2D/Collision/b2Distance.cpp"
 HEAPF32[$167>>2] = $160; //@line 142 "libraries/Box2D/Collision/b2Distance.cpp"
 $168 = ((($this)) + 12|0); //@line 142 "libraries/Box2D/Collision/b2Distance.cpp"
 HEAPF32[$168>>2] = $166; //@line 142 "libraries/Box2D/Collision/b2Distance.cpp"
 $169 = $160 - $144; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $170 = $166 - $150; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $171 = ((($this)) + 16|0); //@line 143 "libraries/Box2D/Collision/b2Distance.cpp"
 HEAPF32[$171>>2] = $169; //@line 143 "libraries/Box2D/Collision/b2Distance.cpp"
 $172 = ((($this)) + 20|0); //@line 143 "libraries/Box2D/Collision/b2Distance.cpp"
 HEAPF32[$172>>2] = $170; //@line 143 "libraries/Box2D/Collision/b2Distance.cpp"
 $173 = ((($this)) + 24|0); //@line 144 "libraries/Box2D/Collision/b2Distance.cpp"
 HEAPF32[$173>>2] = 1.0; //@line 144 "libraries/Box2D/Collision/b2Distance.cpp"
 HEAP32[$5>>2] = 1; //@line 145 "libraries/Box2D/Collision/b2Distance.cpp"
 return; //@line 147 "libraries/Box2D/Collision/b2Distance.cpp"
}
function __ZN12b2ChainShapeD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (20376); //@line 25 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $0 = ((($this)) + 12|0); //@line 31 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 31 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 __THREW__ = 0;
 invoke_vi(3816,($1|0)); //@line 31 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $5 = ___cxa_find_matching_catch_3(0|0)|0;
  $6 = tempRet0;
  ___clang_call_terminate($5); //@line 27 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  // unreachable; //@line 27 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 } else {
  HEAP32[$0>>2] = 0; //@line 32 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $4 = ((($this)) + 16|0); //@line 33 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  HEAP32[$4>>2] = 0; //@line 33 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  return; //@line 27 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 }
}
function __ZN12b2ChainShapeD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (20376); //@line 25 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $0 = ((($this)) + 12|0); //@line 31 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 31 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 __THREW__ = 0;
 invoke_vi(3816,($1|0)); //@line 31 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  ___clang_call_terminate($4); //@line 27 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  // unreachable; //@line 27 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 } else {
  __ZdlPv($this); //@line 25 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  return; //@line 27 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 }
}
function __ZNK12b2ChainShape5CloneEP16b2BlockAllocator($this,$allocator) {
 $this = $this|0;
 $allocator = $allocator|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN16b2BlockAllocator8AllocateEi($allocator,40)|0); //@line 93 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$0>>2] = (20376); //@line 96 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 $1 = ((($0)) + 4|0); //@line 97 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 HEAP32[$1>>2] = 3; //@line 97 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 $2 = ((($0)) + 8|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 HEAPF32[$2>>2] = 0.0099999997764825821; //@line 98 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 $3 = ((($0)) + 12|0); //@line 99 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 HEAP32[$3>>2] = 0; //@line 99 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 $4 = ((($0)) + 16|0); //@line 100 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 HEAP32[$4>>2] = 0; //@line 100 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 $5 = ((($0)) + 36|0); //@line 101 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 HEAP8[$5>>0] = 0; //@line 101 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 $6 = ((($0)) + 37|0); //@line 102 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 HEAP8[$6>>0] = 0; //@line 102 "./libraries/Box2D/Collision/Shapes/b2ChainShape.h"
 $7 = ((($this)) + 12|0); //@line 95 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 95 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $9 = ((($this)) + 16|0); //@line 95 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 95 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 __ZN12b2ChainShape11CreateChainEPK6b2Vec2i($0,$8,$10); //@line 95 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $11 = ((($0)) + 20|0); //@line 96 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $12 = ((($this)) + 20|0); //@line 96 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $13 = $12; //@line 96 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $14 = $13; //@line 96 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 96 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $16 = (($13) + 4)|0; //@line 96 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $17 = $16; //@line 96 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 96 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $19 = $11; //@line 96 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $20 = $19; //@line 96 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$20>>2] = $15; //@line 96 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $21 = (($19) + 4)|0; //@line 96 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $22 = $21; //@line 96 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$22>>2] = $18; //@line 96 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $23 = ((($0)) + 28|0); //@line 97 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $24 = ((($this)) + 28|0); //@line 97 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $25 = $24; //@line 97 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $26 = $25; //@line 97 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $27 = HEAP32[$26>>2]|0; //@line 97 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $28 = (($25) + 4)|0; //@line 97 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $29 = $28; //@line 97 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $30 = HEAP32[$29>>2]|0; //@line 97 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $31 = $23; //@line 97 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $32 = $31; //@line 97 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$32>>2] = $27; //@line 97 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $33 = (($31) + 4)|0; //@line 97 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $34 = $33; //@line 97 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$34>>2] = $30; //@line 97 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $35 = ((($this)) + 36|0); //@line 98 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $36 = HEAP16[$35>>1]|0; //@line 98 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $37 = $36&255; //@line 98 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP8[$5>>0] = $37; //@line 98 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $38 = ($36&65535) >>> 8; //@line 99 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $39 = $38&255; //@line 99 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP8[$6>>0] = $39; //@line 99 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 return ($0|0); //@line 100 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
}
function __ZN12b2ChainShape11CreateChainEPK6b2Vec2i($this,$vertices,$count) {
 $this = $this|0;
 $vertices = $vertices|0;
 $count = $count|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0.0;
 var $27 = 0.0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 60 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 60 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $2 = ($1|0)==(0|0); //@line 60 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 if ($2) {
  $3 = ((($this)) + 16|0); //@line 60 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 60 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $5 = ($4|0)==(0); //@line 60 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $6 = $5;
 } else {
  $6 = 0;
 }
 __Z10loveAssertbPKc($6,361011); //@line 60 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $7 = ($count|0)>(1); //@line 61 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 __Z10loveAssertbPKc($7,361046); //@line 61 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 if ($7) {
  $i$01 = 1;
  while(1) {
   $14 = (($i$01) + -1)|0; //@line 65 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   $15 = (($vertices) + ($14<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $16 = +HEAPF32[$15>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $17 = (($vertices) + ($i$01<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $18 = +HEAPF32[$17>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $19 = $16 - $18; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $20 = (((($vertices) + ($14<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $21 = +HEAPF32[$20>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $22 = (((($vertices) + ($i$01<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $23 = +HEAPF32[$22>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $24 = $21 - $23; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $25 = $19 * $19; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $26 = $24 * $24; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $27 = $25 + $26; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $28 = $27 > 2.4999999368446879E-5; //@line 65 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   __Z10loveAssertbPKc($28,361057); //@line 65 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   $29 = (($i$01) + 1)|0; //@line 62 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   $exitcond = ($29|0)==($count|0); //@line 62 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   if ($exitcond) {
    break;
   } else {
    $i$01 = $29;
   }
  }
 }
 $8 = ((($this)) + 16|0); //@line 68 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$8>>2] = $count; //@line 68 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $9 = $count << 3; //@line 69 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $10 = (__Z7b2Alloci($9)|0); //@line 69 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$0>>2] = $10; //@line 69 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $11 = HEAP32[$8>>2]|0; //@line 70 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $12 = $11 << 3; //@line 70 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 _memcpy(($10|0),($vertices|0),($12|0))|0; //@line 70 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $13 = ((($this)) + 20|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 ;HEAP32[$13>>2]=0|0;HEAP32[$13+4>>2]=0|0;HEAP32[$13+8>>2]=0|0;HEAP32[$13+12>>2]=0|0;HEAP16[$13+16>>1]=0|0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 77 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
}
function __ZNK12b2ChainShape13GetChildCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 106 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 106 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $2 = (($1) + -1)|0; //@line 106 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 return ($2|0); //@line 106 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
}
function __ZNK12b2ChainShape9TestPointERK11b2TransformRK6b2Vec2($this,$xf,$p) {
 $this = $this|0;
 $xf = $xf|0;
 $p = $p|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0; //@line 145 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
}
function __ZNK12b2ChainShape7RayCastEP15b2RayCastOutputRK14b2RayCastInputRK11b2Transformi($this,$output,$input,$xf,$childIndex) {
 $this = $this|0;
 $output = $output|0;
 $input = $input|0;
 $xf = $xf|0;
 $childIndex = $childIndex|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $edgeShape = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $edgeShape = sp;
 $0 = ((($this)) + 16|0); //@line 151 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 151 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $2 = ($1|0)>($childIndex|0); //@line 151 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 __Z10loveAssertbPKc($2,361135); //@line 151 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$edgeShape>>2] = (20456); //@line 63 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $3 = ((($edgeShape)) + 4|0); //@line 64 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 HEAP32[$3>>2] = 1; //@line 64 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $4 = ((($edgeShape)) + 8|0); //@line 65 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 HEAPF32[$4>>2] = 0.0099999997764825821; //@line 65 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $5 = ((($edgeShape)) + 28|0); //@line 66 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;HEAP32[$5+12>>2]=0|0;HEAP16[$5+16>>1]=0|0; //@line 67 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $6 = (($childIndex) + 1)|0; //@line 156 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $7 = HEAP32[$0>>2]|0; //@line 157 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $8 = ($6|0)==($7|0); //@line 157 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $$ = $8 ? 0 : $6; //@line 157 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $9 = ((($edgeShape)) + 12|0); //@line 162 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $10 = ((($this)) + 12|0); //@line 162 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 162 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $12 = (($11) + ($childIndex<<3)|0); //@line 162 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $13 = $12; //@line 162 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $14 = $13; //@line 162 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 162 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $16 = (($13) + 4)|0; //@line 162 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $17 = $16; //@line 162 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 162 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $19 = $9; //@line 162 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $20 = $19; //@line 162 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$20>>2] = $15; //@line 162 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $21 = (($19) + 4)|0; //@line 162 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $22 = $21; //@line 162 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$22>>2] = $18; //@line 162 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $23 = ((($edgeShape)) + 20|0); //@line 163 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $24 = (($11) + ($$<<3)|0); //@line 163 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $25 = $24; //@line 163 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $26 = $25; //@line 163 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $27 = HEAP32[$26>>2]|0; //@line 163 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $28 = (($25) + 4)|0; //@line 163 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $29 = $28; //@line 163 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $30 = HEAP32[$29>>2]|0; //@line 163 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $31 = $23; //@line 163 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $32 = $31; //@line 163 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$32>>2] = $27; //@line 163 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $33 = (($31) + 4)|0; //@line 163 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $34 = $33; //@line 163 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$34>>2] = $30; //@line 163 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $35 = (__ZNK11b2EdgeShape7RayCastEP15b2RayCastOutputRK14b2RayCastInputRK11b2Transformi($edgeShape,$output,$input,$xf,0)|0); //@line 165 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 STACKTOP = sp;return ($35|0); //@line 166 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
}
function __ZNK12b2ChainShape11ComputeAABBEP6b2AABBRK11b2Transformi($this,$aabb,$xf,$childIndex) {
 $this = $this|0;
 $aabb = $aabb|0;
 $xf = $xf|0;
 $childIndex = $childIndex|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0, $25 = 0.0;
 var $26 = 0.0, $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0;
 var $44 = 0, $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 170 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 170 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $2 = ($1|0)>($childIndex|0); //@line 170 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 __Z10loveAssertbPKc($2,361135); //@line 170 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $3 = (($childIndex) + 1)|0; //@line 173 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $4 = HEAP32[$0>>2]|0; //@line 174 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $5 = ($3|0)==($4|0); //@line 174 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $$ = $5 ? 0 : $3; //@line 174 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $6 = ((($this)) + 12|0); //@line 179 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 179 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $8 = ((($xf)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = +HEAPF32[$8>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = (($7) + ($childIndex<<3)|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = +HEAPF32[$10>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $9 * $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($xf)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = (((($7) + ($childIndex<<3)|0)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = +HEAPF32[$15>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $17 = $14 * $16; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $18 = $12 - $17; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $19 = +HEAPF32[$xf>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $20 = $19 + $18; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $21 = $11 * $14; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $22 = $9 * $16; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $23 = $21 + $22; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($xf)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $26 = $23 + $25; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $27 = (($7) + ($$<<3)|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $28 = +HEAPF32[$27>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $29 = $9 * $28; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $30 = (((($7) + ($$<<3)|0)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $31 = +HEAPF32[$30>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $32 = $14 * $31; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $33 = $29 - $32; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $34 = $19 + $33; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $35 = $14 * $28; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $36 = $9 * $31; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $37 = $35 + $36; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $38 = $25 + $37; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $39 = $20 < $34; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $40 = $39 ? $20 : $34; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $41 = $26 < $38; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $42 = $41 ? $26 : $38; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$aabb>>2] = $40; //@line 182 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $43 = ((($aabb)) + 4|0); //@line 182 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAPF32[$43>>2] = $42; //@line 182 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $44 = $20 > $34; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $45 = $44 ? $20 : $34; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $46 = $26 > $38; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $47 = $46 ? $26 : $38; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $48 = ((($aabb)) + 8|0); //@line 183 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAPF32[$48>>2] = $45; //@line 183 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $49 = ((($aabb)) + 12|0); //@line 183 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAPF32[$49>>2] = $47; //@line 183 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 return; //@line 184 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
}
function __ZNK12b2ChainShape11ComputeMassEP10b2MassDataf($this,$massData,$density) {
 $this = $this|0;
 $massData = $massData|0;
 $density = +$density;
 var label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$massData>>2]=0|0;HEAP32[$massData+4>>2]=0|0;HEAP32[$massData+8>>2]=0|0;HEAP32[$massData+12>>2]=0|0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 193 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
}
function __ZN12b2ChainShape10CreateLoopEPK6b2Vec2i($this,$vertices,$count) {
 $this = $this|0;
 $vertices = $vertices|0;
 $count = $count|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0;
 var $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 38 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 38 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $2 = ($1|0)==(0|0); //@line 38 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 if ($2) {
  $3 = ((($this)) + 16|0); //@line 38 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 38 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $5 = ($4|0)==(0); //@line 38 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $6 = $5;
 } else {
  $6 = 0;
 }
 __Z10loveAssertbPKc($6,361011); //@line 38 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $7 = ($count|0)>(2); //@line 39 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 __Z10loveAssertbPKc($7,361415); //@line 39 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $8 = ($count|0)>(1); //@line 40 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 if ($8) {
  $i$02 = 1;
  while(1) {
   $55 = (($i$02) + -1)|0; //@line 42 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   $56 = (($vertices) + ($55<<3)|0); //@line 42 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   $57 = +HEAPF32[$56>>2]; //@line 42 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   $58 = (((($vertices) + ($55<<3)|0)) + 4|0); //@line 42 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   $59 = +HEAPF32[$58>>2]; //@line 42 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   $60 = (($vertices) + ($i$02<<3)|0); //@line 43 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   $61 = +HEAPF32[$60>>2]; //@line 43 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   $62 = (((($vertices) + ($i$02<<3)|0)) + 4|0); //@line 43 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   $63 = +HEAPF32[$62>>2]; //@line 43 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   $64 = $57 - $61; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $65 = $59 - $63; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $66 = $64 * $64; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $67 = $65 * $65; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $68 = $66 + $67; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $69 = $68 > 2.4999999368446879E-5; //@line 45 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   __Z10loveAssertbPKc($69,361156); //@line 45 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   $70 = (($i$02) + 1)|0; //@line 40 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   $exitcond = ($70|0)==($count|0); //@line 40 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
   if ($exitcond) {
    break;
   } else {
    $i$02 = $70;
   }
  }
 }
 $9 = (($count) + 1)|0; //@line 48 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $10 = ((($this)) + 16|0); //@line 48 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$10>>2] = $9; //@line 48 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $11 = $9 << 3; //@line 49 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $12 = (__Z7b2Alloci($11)|0); //@line 49 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$0>>2] = $12; //@line 49 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $13 = $count << 3; //@line 50 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 _memcpy(($12|0),($vertices|0),($13|0))|0; //@line 50 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $14 = HEAP32[$0>>2]|0; //@line 51 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $15 = (($14) + ($count<<3)|0); //@line 51 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $16 = $14; //@line 51 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $17 = $16; //@line 51 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 51 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $19 = (($16) + 4)|0; //@line 51 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $20 = $19; //@line 51 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 51 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $22 = $15; //@line 51 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $23 = $22; //@line 51 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$23>>2] = $18; //@line 51 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $24 = (($22) + 4)|0; //@line 51 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $25 = $24; //@line 51 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$25>>2] = $21; //@line 51 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $26 = ((($this)) + 20|0); //@line 52 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $27 = HEAP32[$10>>2]|0; //@line 52 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $28 = (($27) + -2)|0; //@line 52 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $29 = HEAP32[$0>>2]|0; //@line 52 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $30 = (($29) + ($28<<3)|0); //@line 52 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $31 = $30; //@line 52 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $32 = $31; //@line 52 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $33 = HEAP32[$32>>2]|0; //@line 52 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $34 = (($31) + 4)|0; //@line 52 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $35 = $34; //@line 52 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $36 = HEAP32[$35>>2]|0; //@line 52 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $37 = $26; //@line 52 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $38 = $37; //@line 52 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$38>>2] = $33; //@line 52 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $39 = (($37) + 4)|0; //@line 52 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $40 = $39; //@line 52 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$40>>2] = $36; //@line 52 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $41 = ((($this)) + 28|0); //@line 53 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $42 = ((($29)) + 8|0); //@line 53 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $43 = $42; //@line 53 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $44 = $43; //@line 53 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $45 = HEAP32[$44>>2]|0; //@line 53 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $46 = (($43) + 4)|0; //@line 53 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $47 = $46; //@line 53 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $48 = HEAP32[$47>>2]|0; //@line 53 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $49 = $41; //@line 53 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $50 = $49; //@line 53 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$50>>2] = $45; //@line 53 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $51 = (($49) + 4)|0; //@line 53 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $52 = $51; //@line 53 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$52>>2] = $48; //@line 53 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $53 = ((($this)) + 36|0); //@line 54 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP8[$53>>0] = 1; //@line 54 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $54 = ((($this)) + 37|0); //@line 55 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP8[$54>>0] = 1; //@line 55 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 return; //@line 56 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
}
function __ZN12b2ChainShape13SetNextVertexERK6b2Vec2($this,$nextVertex) {
 $this = $this|0;
 $nextVertex = $nextVertex|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 87 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $1 = $nextVertex; //@line 87 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $2 = $1; //@line 87 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 87 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $4 = (($1) + 4)|0; //@line 87 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $5 = $4; //@line 87 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 87 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $7 = $0; //@line 87 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $8 = $7; //@line 87 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$8>>2] = $3; //@line 87 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $9 = (($7) + 4)|0; //@line 87 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $10 = $9; //@line 87 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$10>>2] = $6; //@line 87 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $11 = ((($this)) + 37|0); //@line 88 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP8[$11>>0] = 1; //@line 88 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 return; //@line 89 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
}
function __ZNK12b2ChainShape12GetChildEdgeEP11b2EdgeShapei($this,$edge,$index) {
 $this = $this|0;
 $edge = $edge|0;
 $index = $index|0;
 var $$sink = 0, $$sink1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($index|0)>(-1); //@line 111 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 if ($0) {
  $1 = ((($this)) + 16|0); //@line 111 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $2 = HEAP32[$1>>2]|0; //@line 111 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $3 = (($2) + -1)|0; //@line 111 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $4 = ($3|0)>($index|0); //@line 111 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $5 = $4;
 } else {
  $5 = 0;
 }
 __Z10loveAssertbPKc($5,361214); //@line 111 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $6 = ((($edge)) + 4|0); //@line 112 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$6>>2] = 1; //@line 112 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $7 = ((($this)) + 8|0); //@line 113 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 113 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $9 = ((($edge)) + 8|0); //@line 113 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$9>>2] = $8; //@line 113 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $10 = ((($edge)) + 12|0); //@line 115 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $11 = ((($this)) + 12|0); //@line 115 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 115 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $13 = (($12) + ($index<<3)|0); //@line 115 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $14 = $13; //@line 115 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $15 = $14; //@line 115 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 115 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $17 = (($14) + 4)|0; //@line 115 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $18 = $17; //@line 115 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 115 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $20 = $10; //@line 115 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $21 = $20; //@line 115 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$21>>2] = $16; //@line 115 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $22 = (($20) + 4)|0; //@line 115 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $23 = $22; //@line 115 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$23>>2] = $19; //@line 115 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $24 = ((($edge)) + 20|0); //@line 116 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $25 = (($index) + 1)|0; //@line 116 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $26 = HEAP32[$11>>2]|0; //@line 116 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $27 = (($26) + ($25<<3)|0); //@line 116 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $28 = $27; //@line 116 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $29 = $28; //@line 116 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $30 = HEAP32[$29>>2]|0; //@line 116 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $31 = (($28) + 4)|0; //@line 116 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $32 = $31; //@line 116 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $33 = HEAP32[$32>>2]|0; //@line 116 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $34 = $24; //@line 116 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $35 = $34; //@line 116 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$35>>2] = $30; //@line 116 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $36 = (($34) + 4)|0; //@line 116 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $37 = $36; //@line 116 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP32[$37>>2] = $33; //@line 116 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $38 = ($index|0)>(0); //@line 118 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $39 = ((($edge)) + 28|0); //@line 120 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 if ($38) {
  $40 = (($index) + -1)|0; //@line 120 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $41 = HEAP32[$11>>2]|0; //@line 120 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $42 = (($41) + ($40<<3)|0); //@line 120 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $43 = $42; //@line 120 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $44 = $43; //@line 120 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $45 = HEAP32[$44>>2]|0; //@line 120 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $46 = (($43) + 4)|0; //@line 120 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $47 = $46; //@line 120 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $48 = HEAP32[$47>>2]|0; //@line 120 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $49 = $39; //@line 120 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $50 = $49; //@line 120 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  HEAP32[$50>>2] = $45; //@line 120 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $51 = (($49) + 4)|0; //@line 120 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $52 = $51; //@line 120 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  HEAP32[$52>>2] = $48; //@line 120 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $$sink = 1;
 } else {
  $53 = ((($this)) + 20|0); //@line 125 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $54 = $53; //@line 125 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $55 = $54; //@line 125 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $56 = HEAP32[$55>>2]|0; //@line 125 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $57 = (($54) + 4)|0; //@line 125 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $58 = $57; //@line 125 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $59 = HEAP32[$58>>2]|0; //@line 125 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $60 = $39; //@line 125 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $61 = $60; //@line 125 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  HEAP32[$61>>2] = $56; //@line 125 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $62 = (($60) + 4)|0; //@line 125 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $63 = $62; //@line 125 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  HEAP32[$63>>2] = $59; //@line 125 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $64 = ((($this)) + 36|0); //@line 126 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $65 = HEAP8[$64>>0]|0; //@line 126 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $$sink = $65;
 }
 $66 = ((($edge)) + 44|0); //@line 126 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 HEAP8[$66>>0] = $$sink; //@line 126 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $67 = ((($this)) + 16|0); //@line 129 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $68 = HEAP32[$67>>2]|0; //@line 129 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $69 = (($68) + -2)|0; //@line 129 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $70 = ($69|0)>($index|0); //@line 129 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 $71 = ((($edge)) + 36|0); //@line 131 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 if ($70) {
  $72 = (($index) + 2)|0; //@line 131 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $73 = HEAP32[$11>>2]|0; //@line 131 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $74 = (($73) + ($72<<3)|0); //@line 131 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $75 = $74; //@line 131 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $76 = $75; //@line 131 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $77 = HEAP32[$76>>2]|0; //@line 131 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $78 = (($75) + 4)|0; //@line 131 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $79 = $78; //@line 131 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $80 = HEAP32[$79>>2]|0; //@line 131 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $81 = $71; //@line 131 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $82 = $81; //@line 131 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  HEAP32[$82>>2] = $77; //@line 131 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $83 = (($81) + 4)|0; //@line 131 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $84 = $83; //@line 131 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  HEAP32[$84>>2] = $80; //@line 131 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $$sink1 = 1;
  $98 = ((($edge)) + 45|0); //@line 137 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  HEAP8[$98>>0] = $$sink1; //@line 137 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  return; //@line 139 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 } else {
  $85 = ((($this)) + 28|0); //@line 136 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $86 = $85; //@line 136 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $87 = $86; //@line 136 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $88 = HEAP32[$87>>2]|0; //@line 136 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $89 = (($86) + 4)|0; //@line 136 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $90 = $89; //@line 136 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $91 = HEAP32[$90>>2]|0; //@line 136 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $92 = $71; //@line 136 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $93 = $92; //@line 136 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  HEAP32[$93>>2] = $88; //@line 136 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $94 = (($92) + 4)|0; //@line 136 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $95 = $94; //@line 136 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  HEAP32[$95>>2] = $91; //@line 136 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $96 = ((($this)) + 37|0); //@line 137 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $97 = HEAP8[$96>>0]|0; //@line 137 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  $$sink1 = $97;
  $98 = ((($edge)) + 45|0); //@line 137 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  HEAP8[$98>>0] = $$sink1; //@line 137 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
  return; //@line 139 "libraries/Box2D/Collision/Shapes/b2ChainShape.cpp"
 }
}
function __ZN7b2ShapeD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 55 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
}
function __ZN13b2CircleShapeD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 25 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 return; //@line 25 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
}
function __ZNK13b2CircleShape5CloneEP16b2BlockAllocator($this,$allocator) {
 $this = $this|0;
 $allocator = $allocator|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN16b2BlockAllocator8AllocateEi($allocator,20)|0); //@line 24 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 HEAP32[$0>>2] = (20416); //@line 66 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 $1 = ((($0)) + 4|0); //@line 67 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0;HEAP32[$1+12>>2]=0|0; //@line 68 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 $2 = ((($this)) + 4|0); //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $3 = $2; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $4 = $3; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $5 = HEAP32[$4>>2]|0; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $6 = (($3) + 4)|0; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $7 = $6; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $8 = HEAP32[$7>>2]|0; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $9 = $1; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $10 = $9; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 HEAP32[$10>>2] = $5; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $11 = (($9) + 4)|0; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $12 = $11; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 HEAP32[$12>>2] = $8; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $13 = ((($0)) + 12|0); //@line 25 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 $14 = ((($this)) + 12|0); //@line 25 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 $15 = $14; //@line 25 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 $16 = $15; //@line 25 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 $17 = HEAP32[$16>>2]|0; //@line 25 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 $18 = (($15) + 4)|0; //@line 25 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 $19 = $18; //@line 25 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 $20 = HEAP32[$19>>2]|0; //@line 25 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 $21 = $13; //@line 25 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 $22 = $21; //@line 25 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 HEAP32[$22>>2] = $17; //@line 25 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 $23 = (($21) + 4)|0; //@line 25 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 $24 = $23; //@line 25 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 HEAP32[$24>>2] = $20; //@line 25 "./libraries/Box2D/Collision/Shapes/b2CircleShape.h"
 return ($0|0); //@line 27 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
}
function __ZNK13b2CircleShape13GetChildCountEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1; //@line 32 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
}
function __ZNK13b2CircleShape9TestPointERK11b2TransformRK6b2Vec2($this,$transform,$p) {
 $this = $this|0;
 $transform = $transform|0;
 $p = $p|0;
 var $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0.0;
 var $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0.0, $30 = 0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0.0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($transform)) + 12|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $1 = +HEAPF32[$0>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $2 = ((($this)) + 12|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $4 = $1 * $3; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($transform)) + 8|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($this)) + 16|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $9 = $6 * $8; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $10 = $4 - $9; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $11 = $3 * $6; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $12 = $1 * $8; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $13 = $11 + $12; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$transform>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $10; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $16 = ((($transform)) + 4|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $17 = +HEAPF32[$16>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $18 = $17 + $13; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $19 = +HEAPF32[$p>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $20 = $19 - $15; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($p)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $22 = +HEAPF32[$21>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $23 = $22 - $18; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $24 = $20 * $20; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $25 = $23 * $23; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $26 = $24 + $25; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($this)) + 8|0); //@line 39 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $28 = +HEAPF32[$27>>2]; //@line 39 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $29 = $28 * $28; //@line 39 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $30 = $26 <= $29; //@line 39 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 return ($30|0); //@line 39 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
}
function __ZNK13b2CircleShape7RayCastEP15b2RayCastOutputRK14b2RayCastInputRK11b2Transformi($this,$output,$input,$transform,$childIndex) {
 $this = $this|0;
 $output = $output|0;
 $input = $input|0;
 $transform = $transform|0;
 $childIndex = $childIndex|0;
 var $$1 = 0, $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0.0;
 var $26 = 0.0, $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0.0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0.0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0;
 var $44 = 0.0, $45 = 0.0, $46 = 0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0, $53 = 0.0, $54 = 0.0, $55 = 0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0.0, $6 = 0.0, $60 = 0.0, $61 = 0.0;
 var $62 = 0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0.0, $8 = 0.0, $9 = 0.0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($transform)) + 12|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $1 = +HEAPF32[$0>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $2 = ((($this)) + 12|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $4 = $1 * $3; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($transform)) + 8|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($this)) + 16|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $9 = $6 * $8; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $10 = $4 - $9; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $11 = $3 * $6; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $12 = $1 * $8; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $13 = $11 + $12; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$transform>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $10; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $16 = ((($transform)) + 4|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $17 = +HEAPF32[$16>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $18 = $17 + $13; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $19 = +HEAPF32[$input>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $20 = $19 - $15; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($input)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $22 = +HEAPF32[$21>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $23 = $22 - $18; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $24 = $20 * $20; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $25 = $23 * $23; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $26 = $24 + $25; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($this)) + 8|0); //@line 53 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $28 = +HEAPF32[$27>>2]; //@line 53 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $29 = $28 * $28; //@line 53 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $30 = $26 - $29; //@line 53 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $31 = ((($input)) + 8|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $32 = +HEAPF32[$31>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $33 = $32 - $19; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $34 = ((($input)) + 12|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $35 = +HEAPF32[$34>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $36 = $35 - $22; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $37 = $20 * $33; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $38 = $23 * $36; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $39 = $37 + $38; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $40 = $33 * $33; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $41 = $36 * $36; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $42 = $40 + $41; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $43 = $39 * $39; //@line 59 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $44 = $30 * $42; //@line 59 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $45 = $43 - $44; //@line 59 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $46 = $45 < 0.0; //@line 62 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $47 = $42 < 1.1920928955078125E-7; //@line 62 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $or$cond = $47 | $46; //@line 62 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 if ($or$cond) {
  $$1 = 0;
  return ($$1|0); //@line 81 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 }
 $48 = (+Math_sqrt((+$45))); //@line 68 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $49 = $39 + $48; //@line 68 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $50 = -$49; //@line 68 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $51 = !($49 <= -0.0); //@line 71 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 if ($51) {
  $$1 = 0;
  return ($$1|0); //@line 81 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 }
 $52 = ((($input)) + 16|0); //@line 71 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $53 = +HEAPF32[$52>>2]; //@line 71 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $54 = $42 * $53; //@line 71 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $55 = !($54 >= $50); //@line 71 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 if ($55) {
  $$1 = 0;
  return ($$1|0); //@line 81 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 }
 $56 = $50 / $42; //@line 73 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $57 = ((($output)) + 8|0); //@line 74 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 HEAPF32[$57>>2] = $56; //@line 74 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $58 = $33 * $56; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $59 = $36 * $56; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $60 = $20 + $58; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $61 = $23 + $59; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$output>>2] = $60; //@line 75 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $62 = ((($output)) + 4|0); //@line 75 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 HEAPF32[$62>>2] = $61; //@line 75 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $63 = $60 * $60; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $64 = $61 * $61; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $65 = $63 + $64; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $66 = (+Math_sqrt((+$65))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $67 = $66 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
 if ($67) {
  $$1 = 1;
  return ($$1|0); //@line 81 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 }
 $68 = 1.0 / $66; //@line 121 "./libraries/Box2D/Common/b2Math.h"
 $69 = $60 * $68; //@line 122 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$output>>2] = $69; //@line 122 "./libraries/Box2D/Common/b2Math.h"
 $70 = $61 * $68; //@line 123 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$62>>2] = $70; //@line 123 "./libraries/Box2D/Common/b2Math.h"
 $$1 = 1;
 return ($$1|0); //@line 81 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
}
function __ZNK13b2CircleShape11ComputeAABBEP6b2AABBRK11b2Transformi($this,$aabb,$transform,$childIndex) {
 $this = $this|0;
 $aabb = $aabb|0;
 $transform = $transform|0;
 $childIndex = $childIndex|0;
 var $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0, $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0.0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($transform)) + 12|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $1 = +HEAPF32[$0>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $2 = ((($this)) + 12|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $4 = $1 * $3; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($transform)) + 8|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($this)) + 16|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $9 = $6 * $8; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $10 = $4 - $9; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $11 = $3 * $6; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $12 = $1 * $8; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $13 = $11 + $12; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$transform>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $10; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $16 = ((($transform)) + 4|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $17 = +HEAPF32[$16>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $18 = $17 + $13; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($this)) + 8|0); //@line 88 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $20 = +HEAPF32[$19>>2]; //@line 88 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $21 = $15 - $20; //@line 88 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $22 = $18 - $20; //@line 88 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 HEAPF32[$aabb>>2] = $21; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $23 = ((($aabb)) + 4|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$23>>2] = $22; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $24 = $15 + $20; //@line 89 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $25 = $18 + $20; //@line 89 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $26 = ((($aabb)) + 8|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$26>>2] = $24; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($aabb)) + 12|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$27>>2] = $25; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 90 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
}
function __ZNK13b2CircleShape11ComputeMassEP10b2MassDataf($this,$massData,$density) {
 $this = $this|0;
 $massData = $massData|0;
 $density = +$density;
 var $0 = 0.0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0.0;
 var $27 = 0.0, $28 = 0, $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $density * 3.1415927410125732; //@line 94 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $1 = ((($this)) + 8|0); //@line 94 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $2 = +HEAPF32[$1>>2]; //@line 94 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $3 = $0 * $2; //@line 94 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $4 = $2 * $3; //@line 94 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 HEAPF32[$massData>>2] = $4; //@line 94 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $5 = ((($massData)) + 4|0); //@line 95 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $6 = ((($this)) + 12|0); //@line 95 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $7 = $6; //@line 95 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $8 = $7; //@line 95 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 95 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $10 = (($7) + 4)|0; //@line 95 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $11 = $10; //@line 95 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 95 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $13 = $5; //@line 95 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $14 = $13; //@line 95 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 HEAP32[$14>>2] = $9; //@line 95 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $15 = (($13) + 4)|0; //@line 95 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $16 = $15; //@line 95 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 HEAP32[$16>>2] = $12; //@line 95 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $17 = +HEAPF32[$1>>2]; //@line 98 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $18 = $17 * 0.5; //@line 98 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $19 = $17 * $18; //@line 98 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $20 = +HEAPF32[$6>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $21 = $20 * $20; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($this)) + 16|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $23 = +HEAPF32[$22>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $24 = $23 * $23; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $25 = $21 + $24; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $26 = $19 + $25; //@line 98 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $27 = $4 * $26; //@line 98 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 $28 = ((($massData)) + 12|0); //@line 98 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 HEAPF32[$28>>2] = $27; //@line 98 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
 return; //@line 99 "libraries/Box2D/Collision/Shapes/b2CircleShape.cpp"
}
function __ZN11b2EdgeShapeD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 27 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 return; //@line 27 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
}
function __ZNK11b2EdgeShape5CloneEP16b2BlockAllocator($this,$allocator) {
 $this = $this|0;
 $allocator = $allocator|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN16b2BlockAllocator8AllocateEi($allocator,48)|0); //@line 32 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 HEAP32[$0>>2] = (20456); //@line 63 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $1 = ((($0)) + 4|0); //@line 64 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 HEAP32[$1>>2] = 1; //@line 64 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $2 = ((($0)) + 8|0); //@line 65 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 HEAPF32[$2>>2] = 0.0099999997764825821; //@line 65 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $3 = ((($0)) + 28|0); //@line 66 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 ;HEAP32[$3>>2]=0|0;HEAP32[$3+4>>2]=0|0;HEAP32[$3+8>>2]=0|0;HEAP32[$3+12>>2]=0|0;HEAP16[$3+16>>1]=0|0; //@line 67 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $4 = ((($this)) + 4|0); //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $5 = $4; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $6 = $5; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $7 = HEAP32[$6>>2]|0; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $8 = (($5) + 4)|0; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $9 = $8; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $10 = HEAP32[$9>>2]|0; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $11 = $1; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $12 = $11; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 HEAP32[$12>>2] = $7; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $13 = (($11) + 4)|0; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $14 = $13; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 HEAP32[$14>>2] = $10; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $15 = ((($0)) + 12|0); //@line 27 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $16 = ((($this)) + 12|0); //@line 27 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 ;HEAP32[$15>>2]=HEAP32[$16>>2]|0;HEAP32[$15+4>>2]=HEAP32[$16+4>>2]|0;HEAP32[$15+8>>2]=HEAP32[$16+8>>2]|0;HEAP32[$15+12>>2]=HEAP32[$16+12>>2]|0;HEAP32[$15+16>>2]=HEAP32[$16+16>>2]|0;HEAP32[$15+20>>2]=HEAP32[$16+20>>2]|0;HEAP32[$15+24>>2]=HEAP32[$16+24>>2]|0;HEAP32[$15+28>>2]=HEAP32[$16+28>>2]|0;HEAP16[$15+32>>1]=HEAP16[$16+32>>1]|0; //@line 27 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 return ($0|0); //@line 35 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
}
function __ZNK11b2EdgeShape13GetChildCountEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1; //@line 40 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
}
function __ZNK11b2EdgeShape9TestPointERK11b2TransformRK6b2Vec2($this,$xf,$p) {
 $this = $this|0;
 $xf = $xf|0;
 $p = $p|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0; //@line 47 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
}
function __ZNK11b2EdgeShape7RayCastEP15b2RayCastOutputRK14b2RayCastInputRK11b2Transformi($this,$output,$input,$xf,$childIndex) {
 $this = $this|0;
 $output = $output|0;
 $input = $input|0;
 $xf = $xf|0;
 $childIndex = $childIndex|0;
 var $$3 = 0, $0 = 0.0, $1 = 0.0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0.0, $2 = 0.0, $20 = 0.0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0.0;
 var $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0.0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0;
 var $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0, $6 = 0.0, $60 = 0.0, $61 = 0;
 var $62 = 0, $63 = 0.0, $64 = 0, $65 = 0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0.0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0, $89 = 0, $9 = 0.0, $normal$sroa$0$0 = 0.0, $normal$sroa$9$0 = 0.0, $or$cond = 0, $or$cond19 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = +HEAPF32[$input>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $1 = +HEAPF32[$xf>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $2 = $0 - $1; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($input)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$3>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($xf)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $7 = $4 - $6; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($xf)) + 12|0); //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $9 = +HEAPF32[$8>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $10 = $2 * $9; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($xf)) + 8|0); //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$11>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $13 = $7 * $12; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $14 = $10 + $13; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $15 = $2 * $12; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $16 = $9 * $7; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $17 = $16 - $15; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $18 = ((($input)) + 8|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $19 = +HEAPF32[$18>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $20 = $19 - $1; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($input)) + 12|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $22 = +HEAPF32[$21>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $23 = $22 - $6; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $24 = $9 * $20; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $25 = $12 * $23; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $26 = $24 + $25; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $27 = $12 * $20; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $28 = $9 * $23; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $29 = $28 - $27; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $30 = $26 - $14; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $31 = $29 - $17; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $32 = ((($this)) + 12|0); //@line 64 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $33 = +HEAPF32[$32>>2]; //@line 64 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $34 = ((($this)) + 16|0); //@line 64 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $35 = +HEAPF32[$34>>2]; //@line 64 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $36 = ((($this)) + 20|0); //@line 65 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $37 = +HEAPF32[$36>>2]; //@line 65 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $38 = ((($this)) + 24|0); //@line 65 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $39 = +HEAPF32[$38>>2]; //@line 65 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $40 = $37 - $33; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $41 = $39 - $35; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $42 = -$40; //@line 67 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $43 = $41 * $41; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $44 = $40 * $40; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $45 = $44 + $43; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $46 = (+Math_sqrt((+$45))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $47 = $46 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
 if ($47) {
  $normal$sroa$0$0 = $41;$normal$sroa$9$0 = $42;
 } else {
  $48 = 1.0 / $46; //@line 121 "./libraries/Box2D/Common/b2Math.h"
  $49 = $41 * $48; //@line 122 "./libraries/Box2D/Common/b2Math.h"
  $50 = $48 * $42; //@line 123 "./libraries/Box2D/Common/b2Math.h"
  $normal$sroa$0$0 = $49;$normal$sroa$9$0 = $50;
 }
 $51 = $33 - $14; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $52 = $35 - $17; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $53 = $51 * $normal$sroa$0$0; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $54 = $52 * $normal$sroa$9$0; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $55 = $53 + $54; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $56 = $30 * $normal$sroa$0$0; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $57 = $31 * $normal$sroa$9$0; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $58 = $56 + $57; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $59 = $58 == 0.0; //@line 76 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 if ($59) {
  $$3 = 0;
  return ($$3|0); //@line 114 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 }
 $60 = $55 / $58; //@line 81 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $61 = $60 < 0.0; //@line 82 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 if ($61) {
  $$3 = 0;
  return ($$3|0); //@line 114 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 }
 $62 = ((($input)) + 16|0); //@line 82 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $63 = +HEAPF32[$62>>2]; //@line 82 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $64 = $63 < $60; //@line 82 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $65 = $45 == 0.0; //@line 93 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $or$cond19 = $64 | $65; //@line 82 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 if ($or$cond19) {
  $$3 = 0;
  return ($$3|0); //@line 114 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 }
 $66 = $31 * $60; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $67 = $17 + $66; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $68 = $30 * $60; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $69 = $14 + $68; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $70 = $69 - $33; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $71 = $67 - $35; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $72 = $40 * $70; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $73 = $41 * $71; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $74 = $72 + $73; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $75 = $74 / $45; //@line 98 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $76 = $75 < 0.0; //@line 99 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $77 = $75 > 1.0; //@line 99 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $or$cond = $76 | $77; //@line 99 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 if ($or$cond) {
  $$3 = 0;
  return ($$3|0); //@line 114 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 }
 $78 = ((($output)) + 8|0); //@line 104 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 HEAPF32[$78>>2] = $60; //@line 104 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $79 = $55 > 0.0; //@line 105 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $80 = $normal$sroa$0$0 * $9; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $81 = $normal$sroa$9$0 * $12; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $82 = $80 - $81; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $83 = $normal$sroa$0$0 * $12; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $84 = $normal$sroa$9$0 * $9; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $85 = $84 + $83; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 if ($79) {
  $86 = -$82; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $87 = -$85; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$output>>2] = $86; //@line 107 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
  $88 = ((($output)) + 4|0); //@line 107 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
  HEAPF32[$88>>2] = $87; //@line 107 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
  $$3 = 1;
  return ($$3|0); //@line 114 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 } else {
  HEAPF32[$output>>2] = $82; //@line 111 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
  $89 = ((($output)) + 4|0); //@line 111 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
  HEAPF32[$89>>2] = $85; //@line 111 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
  $$3 = 1;
  return ($$3|0); //@line 114 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 }
 return (0)|0;
}
function __ZNK11b2EdgeShape11ComputeAABBEP6b2AABBRK11b2Transformi($this,$aabb,$xf,$childIndex) {
 $this = $this|0;
 $aabb = $aabb|0;
 $xf = $xf|0;
 $childIndex = $childIndex|0;
 var $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0.0;
 var $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0.0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0, $4 = 0.0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0.0;
 var $45 = 0.0, $46 = 0, $47 = 0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0.0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($xf)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $1 = +HEAPF32[$0>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $2 = ((($this)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = $1 * $3; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($xf)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($this)) + 16|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = $6 * $8; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = $4 - $9; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = +HEAPF32[$xf>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $11 + $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = $3 * $6; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $14 = $1 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $15 = $13 + $14; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $16 = ((($xf)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = +HEAPF32[$16>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $15 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($this)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $21 = $1 * $20; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($this)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $23 = +HEAPF32[$22>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $24 = $6 * $23; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $25 = $21 - $24; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $26 = $11 + $25; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $27 = $6 * $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $28 = $1 * $23; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $29 = $27 + $28; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $30 = $17 + $29; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $31 = $12 < $26; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $32 = $31 ? $12 : $26; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $33 = $18 < $30; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $34 = $33 ? $18 : $30; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $35 = $12 > $26; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $36 = $35 ? $12 : $26; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $37 = $18 > $30; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $38 = $37 ? $18 : $30; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $39 = ((($this)) + 8|0); //@line 126 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $40 = +HEAPF32[$39>>2]; //@line 126 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $41 = $32 - $40; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $42 = $34 - $40; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$aabb>>2] = $41; //@line 127 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $43 = ((($aabb)) + 4|0); //@line 127 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 HEAPF32[$43>>2] = $42; //@line 127 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $44 = $40 + $36; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $45 = $40 + $38; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $46 = ((($aabb)) + 8|0); //@line 128 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 HEAPF32[$46>>2] = $44; //@line 128 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $47 = ((($aabb)) + 12|0); //@line 128 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 HEAPF32[$47>>2] = $45; //@line 128 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 return; //@line 129 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
}
function __ZNK11b2EdgeShape11ComputeMassEP10b2MassDataf($this,$massData,$density) {
 $this = $this|0;
 $massData = $massData|0;
 $density = +$density;
 var $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0.0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAPF32[$massData>>2] = 0.0; //@line 135 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $0 = ((($this)) + 12|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $1 = +HEAPF32[$0>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $2 = ((($this)) + 20|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $4 = $1 + $3; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($this)) + 16|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($this)) + 24|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $9 = $6 + $8; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $10 = $4 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $11 = $9 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $12 = ((($massData)) + 4|0); //@line 136 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 HEAPF32[$12>>2] = $10; //@line 136 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $13 = ((($massData)) + 8|0); //@line 136 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 HEAPF32[$13>>2] = $11; //@line 136 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $14 = ((($massData)) + 12|0); //@line 137 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 HEAPF32[$14>>2] = 0.0; //@line 137 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 return; //@line 138 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
}
function __ZN11b2EdgeShape3SetERK6b2Vec2S2_($this,$v1,$v2) {
 $this = $this|0;
 $v1 = $v1|0;
 $v2 = $v2|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 24 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $1 = $v1; //@line 24 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $2 = $1; //@line 24 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 24 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $4 = (($1) + 4)|0; //@line 24 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $5 = $4; //@line 24 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 24 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $7 = $0; //@line 24 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $8 = $7; //@line 24 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 HEAP32[$8>>2] = $3; //@line 24 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $9 = (($7) + 4)|0; //@line 24 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $10 = $9; //@line 24 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 HEAP32[$10>>2] = $6; //@line 24 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $11 = ((($this)) + 20|0); //@line 25 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $12 = $v2; //@line 25 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $13 = $12; //@line 25 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 25 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $15 = (($12) + 4)|0; //@line 25 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $16 = $15; //@line 25 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 25 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $18 = $11; //@line 25 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $19 = $18; //@line 25 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 HEAP32[$19>>2] = $14; //@line 25 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $20 = (($18) + 4)|0; //@line 25 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $21 = $20; //@line 25 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 HEAP32[$21>>2] = $17; //@line 25 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $22 = ((($this)) + 44|0); //@line 26 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 HEAP8[$22>>0] = 0; //@line 26 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 $23 = ((($this)) + 45|0); //@line 27 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 HEAP8[$23>>0] = 0; //@line 27 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
 return; //@line 28 "libraries/Box2D/Collision/Shapes/b2EdgeShape.cpp"
}
function __ZN14b2PolygonShapeD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 28 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 return; //@line 28 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
}
function __ZNK14b2PolygonShape5CloneEP16b2BlockAllocator($this,$allocator) {
 $this = $this|0;
 $allocator = $allocator|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN16b2BlockAllocator8AllocateEi($allocator,152)|0); //@line 24 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 HEAP32[$0>>2] = (20496); //@line 88 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $1 = ((($0)) + 4|0); //@line 89 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 HEAP32[$1>>2] = 2; //@line 89 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $2 = ((($0)) + 8|0); //@line 90 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 HEAPF32[$2>>2] = 0.0099999997764825821; //@line 90 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $3 = ((($0)) + 148|0); //@line 91 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 HEAP32[$3>>2] = 0; //@line 91 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 $4 = ((($0)) + 12|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$4>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($0)) + 16|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$5>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($this)) + 4|0); //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $7 = $6; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $8 = $7; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $9 = HEAP32[$8>>2]|0; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $10 = (($7) + 4)|0; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $11 = $10; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $12 = HEAP32[$11>>2]|0; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $13 = $1; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $14 = $13; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 HEAP32[$14>>2] = $9; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $15 = (($13) + 4)|0; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $16 = $15; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 HEAP32[$16>>2] = $12; //@line 42 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $17 = ((($this)) + 12|0); //@line 28 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 _memcpy(($4|0),($17|0),140)|0; //@line 28 "./libraries/Box2D/Collision/Shapes/b2PolygonShape.h"
 return ($0|0); //@line 27 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
}
function __ZNK14b2PolygonShape13GetChildCountEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1; //@line 71 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
}
function __ZNK14b2PolygonShape9TestPointERK11b2TransformRK6b2Vec2($this,$xf,$p) {
 $this = $this|0;
 $xf = $xf|0;
 $p = $p|0;
 var $$13 = 0, $0 = 0.0, $1 = 0.0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0.0, $25 = 0.0;
 var $26 = 0, $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0, $9 = 0.0, $i$07 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = +HEAPF32[$p>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $1 = +HEAPF32[$xf>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $2 = $0 - $1; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($p)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$3>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($xf)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $7 = $4 - $6; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($xf)) + 12|0); //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $9 = +HEAPF32[$8>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $10 = $2 * $9; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($xf)) + 8|0); //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$11>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $13 = $7 * $12; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $14 = $10 + $13; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $15 = $2 * $12; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $16 = $9 * $7; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $17 = $16 - $15; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $18 = ((($this)) + 148|0); //@line 255 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 255 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $20 = ($19|0)>(0); //@line 255 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 if ($20) {
  $i$07 = 0;
 } else {
  $$13 = 1;
  return ($$13|0); //@line 265 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 }
 while(1) {
  $23 = (((($this)) + 20|0) + ($i$07<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $24 = +HEAPF32[$23>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $25 = $14 - $24; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $26 = (((((($this)) + 20|0) + ($i$07<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $27 = +HEAPF32[$26>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $28 = $17 - $27; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $29 = (((($this)) + 84|0) + ($i$07<<3)|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $30 = +HEAPF32[$29>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $31 = $25 * $30; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $32 = (((((($this)) + 84|0) + ($i$07<<3)|0)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $33 = +HEAPF32[$32>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $34 = $28 * $33; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $35 = $31 + $34; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $36 = $35 > 0.0; //@line 258 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $22 = (($i$07) + 1)|0; //@line 255 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  if ($36) {
   $$13 = 0;
   label = 4;
   break;
  }
  $21 = ($22|0)<($19|0); //@line 255 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  if ($21) {
   $i$07 = $22;
  } else {
   $$13 = 1;
   label = 4;
   break;
  }
 }
 if ((label|0) == 4) {
  return ($$13|0); //@line 265 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 }
 return (0)|0;
}
function __ZNK14b2PolygonShape7RayCastEP15b2RayCastOutputRK14b2RayCastInputRK11b2Transformi($this,$output,$input,$xf,$childIndex) {
 $this = $this|0;
 $output = $output|0;
 $input = $input|0;
 $xf = $xf|0;
 $childIndex = $childIndex|0;
 var $$3 = 0, $0 = 0.0, $1 = 0.0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0.0, $2 = 0.0, $20 = 0.0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0.0;
 var $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0.0;
 var $44 = 0.0, $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0, $56 = 0, $57 = 0, $58 = 0.0, $59 = 0, $6 = 0.0, $60 = 0.0, $61 = 0;
 var $62 = 0.0, $63 = 0, $64 = 0.0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0.0, $70 = 0.0, $71 = 0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0;
 var $80 = 0.0, $81 = 0.0, $82 = 0, $9 = 0.0, $i$022 = 0, $index$0$lcssa = 0, $index$021 = 0, $index$1 = 0, $lower$0$lcssa = 0.0, $lower$023 = 0.0, $lower$1 = 0.0, $not$ = 0, $or$cond = 0, $or$cond6 = 0, $upper$020 = 0.0, $upper$1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = +HEAPF32[$input>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $1 = +HEAPF32[$xf>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $2 = $0 - $1; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($input)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$3>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($xf)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $7 = $4 - $6; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($xf)) + 12|0); //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $9 = +HEAPF32[$8>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $10 = $2 * $9; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($xf)) + 8|0); //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$11>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $13 = $7 * $12; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $14 = $10 + $13; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $15 = $2 * $12; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $16 = $9 * $7; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $17 = $16 - $15; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $18 = ((($input)) + 8|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $19 = +HEAPF32[$18>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $20 = $19 - $1; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($input)) + 12|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $22 = +HEAPF32[$21>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $23 = $22 - $6; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $24 = $9 * $20; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $25 = $12 * $23; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $26 = $24 + $25; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $27 = $12 * $20; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $28 = $9 * $23; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $29 = $28 - $27; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $30 = $26 - $14; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $31 = $29 - $17; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $32 = ((($input)) + 16|0); //@line 277 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $33 = +HEAPF32[$32>>2]; //@line 327 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $34 = ((($this)) + 148|0); //@line 281 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 281 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $36 = ($35|0)>(0); //@line 281 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 L1: do {
  if ($36) {
   $i$022 = 0;$index$021 = -1;$lower$023 = 0.0;$upper$020 = $33;
   L2: while(1) {
    $39 = (((($this)) + 20|0) + ($i$022<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $40 = +HEAPF32[$39>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $41 = $40 - $14; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $42 = (((((($this)) + 20|0) + ($i$022<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $43 = +HEAPF32[$42>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $44 = $43 - $17; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $45 = (((($this)) + 84|0) + ($i$022<<3)|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $46 = +HEAPF32[$45>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $47 = $41 * $46; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $48 = (((((($this)) + 84|0) + ($i$022<<3)|0)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $49 = +HEAPF32[$48>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $50 = $44 * $49; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $51 = $47 + $50; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $52 = $30 * $46; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $53 = $31 * $49; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $54 = $52 + $53; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $55 = $54 == 0.0; //@line 289 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    do {
     if ($55) {
      $56 = $51 < 0.0; //@line 291 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      if ($56) {
       $$3 = 0;
       label = 12;
       break L2;
      } else {
       $index$1 = $index$021;$lower$1 = $lower$023;$upper$1 = $upper$020;
      }
     } else {
      $57 = $54 < 0.0; //@line 302 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $58 = $lower$023 * $54; //@line 302 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $59 = $51 < $58; //@line 302 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $or$cond = $57 & $59; //@line 302 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      if ($or$cond) {
       $60 = $51 / $54; //@line 306 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
       $index$1 = $i$022;$lower$1 = $60;$upper$1 = $upper$020;
       break;
      }
      $61 = $54 > 0.0; //@line 309 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $62 = $upper$020 * $54; //@line 309 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $63 = $51 < $62; //@line 309 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $or$cond6 = $61 & $63; //@line 309 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      if ($or$cond6) {
       $64 = $51 / $54; //@line 313 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
       $index$1 = $index$021;$lower$1 = $lower$023;$upper$1 = $64;
      } else {
       $index$1 = $index$021;$lower$1 = $lower$023;$upper$1 = $upper$020;
      }
     }
    } while(0);
    $65 = $upper$1 < $lower$1; //@line 321 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $38 = (($i$022) + 1)|0; //@line 281 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    if ($65) {
     $$3 = 0;
     label = 12;
     break;
    }
    $37 = ($38|0)<($35|0); //@line 281 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    if ($37) {
     $i$022 = $38;$index$021 = $index$1;$lower$023 = $lower$1;$upper$020 = $upper$1;
    } else {
     $index$0$lcssa = $index$1;$lower$0$lcssa = $lower$1;
     break L1;
    }
   }
   if ((label|0) == 12) {
    return ($$3|0); //@line 337 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   }
  } else {
   $index$0$lcssa = -1;$lower$0$lcssa = 0.0;
  }
 } while(0);
 $66 = $lower$0$lcssa <= $33; //@line 327 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $not$ = $lower$0$lcssa >= 0.0; //@line 327 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $67 = $not$ & $66; //@line 327 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 __Z10loveAssertbPKc($67,361295); //@line 327 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $68 = ($index$0$lcssa|0)>(-1); //@line 329 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 if (!($68)) {
  $$3 = 0;
  return ($$3|0); //@line 337 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 }
 $69 = ((($output)) + 8|0); //@line 331 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 HEAPF32[$69>>2] = $lower$0$lcssa; //@line 331 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $70 = +HEAPF32[$8>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $71 = (((($this)) + 84|0) + ($index$0$lcssa<<3)|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $72 = +HEAPF32[$71>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $73 = $70 * $72; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $74 = +HEAPF32[$11>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $75 = (((((($this)) + 84|0) + ($index$0$lcssa<<3)|0)) + 4|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $76 = +HEAPF32[$75>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $77 = $74 * $76; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $78 = $73 - $77; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $79 = $72 * $74; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $80 = $70 * $76; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $81 = $79 + $80; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$output>>2] = $78; //@line 332 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $82 = ((($output)) + 4|0); //@line 332 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 HEAPF32[$82>>2] = $81; //@line 332 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $$3 = 1;
 return ($$3|0); //@line 337 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
}
function __ZNK14b2PolygonShape11ComputeAABBEP6b2AABBRK11b2Transformi($this,$aabb,$xf,$childIndex) {
 $this = $this|0;
 $aabb = $aabb|0;
 $xf = $xf|0;
 $childIndex = $childIndex|0;
 var $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0.0, $26 = 0.0;
 var $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0.0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0.0, $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0;
 var $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0, $56 = 0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0.0, $60 = 0.0, $61 = 0, $62 = 0.0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0.0, $9 = 0.0, $i$016 = 0, $lower$sroa$0$0$lcssa = 0, $lower$sroa$0$014 = 0, $lower$sroa$7$0$lcssa = 0, $lower$sroa$7$015 = 0, $upper$sroa$0$0$lcssa = 0, $upper$sroa$0$012 = 0, $upper$sroa$6$0$lcssa = 0, $upper$sroa$6$013 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($xf)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $1 = +HEAPF32[$0>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $2 = ((($this)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = $1 * $3; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($xf)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($this)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = $6 * $8; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = $4 - $9; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = +HEAPF32[$xf>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $11 + $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = $3 * $6; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $14 = $1 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $15 = $13 + $14; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $16 = ((($xf)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = +HEAPF32[$16>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $15 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = (HEAPF32[tempDoublePtr>>2]=$12,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $20 = (HEAPF32[tempDoublePtr>>2]=$18,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($this)) + 148|0); //@line 346 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $22 = HEAP32[$21>>2]|0; //@line 346 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $23 = ($22|0)>(1); //@line 346 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 if ($23) {
  $i$016 = 1;$lower$sroa$0$014 = $19;$lower$sroa$7$015 = $20;$upper$sroa$0$012 = $19;$upper$sroa$6$013 = $20;
  while(1) {
   $37 = (((($this)) + 20|0) + ($i$016<<3)|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $38 = +HEAPF32[$37>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $39 = $1 * $38; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $40 = (((((($this)) + 20|0) + ($i$016<<3)|0)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $41 = +HEAPF32[$40>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $42 = $6 * $41; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $43 = $39 - $42; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $44 = $11 + $43; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $45 = $38 * $6; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $46 = $1 * $41; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $47 = $45 + $46; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $48 = $47 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $49 = (HEAP32[tempDoublePtr>>2]=$lower$sroa$0$014,+HEAPF32[tempDoublePtr>>2]); //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $50 = $49 < $44; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $51 = $50 ? $49 : $44; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $52 = (HEAP32[tempDoublePtr>>2]=$lower$sroa$7$015,+HEAPF32[tempDoublePtr>>2]); //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $53 = $52 < $48; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $54 = $53 ? $52 : $48; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $55 = (HEAPF32[tempDoublePtr>>2]=$51,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
   $56 = (HEAPF32[tempDoublePtr>>2]=$54,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
   $57 = (HEAP32[tempDoublePtr>>2]=$upper$sroa$0$012,+HEAPF32[tempDoublePtr>>2]); //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $58 = $57 > $44; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $59 = $58 ? $57 : $44; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $60 = (HEAP32[tempDoublePtr>>2]=$upper$sroa$6$013,+HEAPF32[tempDoublePtr>>2]); //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $61 = $60 > $48; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $62 = $61 ? $60 : $48; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $63 = (HEAPF32[tempDoublePtr>>2]=$59,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
   $64 = (HEAPF32[tempDoublePtr>>2]=$62,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
   $65 = (($i$016) + 1)|0; //@line 346 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   $66 = ($65|0)<($22|0); //@line 346 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   if ($66) {
    $i$016 = $65;$lower$sroa$0$014 = $55;$lower$sroa$7$015 = $56;$upper$sroa$0$012 = $63;$upper$sroa$6$013 = $64;
   } else {
    $lower$sroa$0$0$lcssa = $55;$lower$sroa$7$0$lcssa = $56;$upper$sroa$0$0$lcssa = $63;$upper$sroa$6$0$lcssa = $64;
    break;
   }
  }
 } else {
  $lower$sroa$0$0$lcssa = $19;$lower$sroa$7$0$lcssa = $20;$upper$sroa$0$0$lcssa = $19;$upper$sroa$6$0$lcssa = $20;
 }
 $24 = ((($this)) + 8|0); //@line 353 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $25 = +HEAPF32[$24>>2]; //@line 353 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $26 = (HEAP32[tempDoublePtr>>2]=$lower$sroa$0$0$lcssa,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $27 = $26 - $25; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $28 = (HEAP32[tempDoublePtr>>2]=$lower$sroa$7$0$lcssa,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $29 = $28 - $25; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$aabb>>2] = $27; //@line 354 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $30 = ((($aabb)) + 4|0); //@line 354 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 HEAPF32[$30>>2] = $29; //@line 354 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $31 = (HEAP32[tempDoublePtr>>2]=$upper$sroa$0$0$lcssa,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $32 = $31 + $25; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $33 = (HEAP32[tempDoublePtr>>2]=$upper$sroa$6$0$lcssa,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $34 = $33 + $25; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $35 = ((($aabb)) + 8|0); //@line 355 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 HEAPF32[$35>>2] = $32; //@line 355 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $36 = ((($aabb)) + 12|0); //@line 355 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 HEAPF32[$36>>2] = $34; //@line 355 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 return; //@line 356 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
}
function __ZNK14b2PolygonShape11ComputeMassEP10b2MassDataf($this,$massData,$density) {
 $this = $this|0;
 $massData = $massData|0;
 $density = +$density;
 var $$lcssa46 = 0.0, $$lcssa47 = 0.0, $$pn = 0.0, $$pn$in = 0, $$sink = 0.0, $$sink$in = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0;
 var $21 = 0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0;
 var $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0.0, $57 = 0.0;
 var $58 = 0.0, $59 = 0.0, $6 = 0.0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0.0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0;
 var $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0.0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $9 = 0.0, $I$0$lcssa = 0.0, $I$08 = 0.0, $area$0$lcssa = 0.0, $area$09 = 0.0, $center$sroa$0$0$lcssa = 0.0, $center$sroa$0$05 = 0.0, $center$sroa$10$0$lcssa = 0.0, $center$sroa$10$06 = 0.0, $e2$sroa$0$0 = 0.0, $i$015 = 0, $i1$07 = 0;
 var $s$sroa$0$013 = 0.0, $s$sroa$11$014 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 148|0); //@line 384 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 384 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $2 = ($1|0)>(2); //@line 384 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 __Z10loveAssertbPKc($2,361339); //@line 384 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $3 = HEAP32[$0>>2]|0; //@line 403 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $4 = ($3|0)>(0); //@line 395 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 if ($4) {
  $i$015 = 0;$s$sroa$0$013 = 0.0;$s$sroa$11$014 = 0.0;
  while(1) {
   $15 = (((($this)) + 20|0) + ($i$015<<3)|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
   $16 = +HEAPF32[$15>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
   $17 = $s$sroa$0$013 + $16; //@line 85 "./libraries/Box2D/Common/b2Math.h"
   $18 = (((((($this)) + 20|0) + ($i$015<<3)|0)) + 4|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
   $19 = +HEAPF32[$18>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
   $20 = $s$sroa$11$014 + $19; //@line 85 "./libraries/Box2D/Common/b2Math.h"
   $21 = (($i$015) + 1)|0; //@line 395 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   $22 = ($21|0)<($3|0); //@line 395 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   if ($22) {
    $i$015 = $21;$s$sroa$0$013 = $17;$s$sroa$11$014 = $20;
   } else {
    $$lcssa46 = $17;$$lcssa47 = $20;
    break;
   }
  }
  $9 = (+($3|0)); //@line 399 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $10 = 1.0 / $9; //@line 399 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $11 = $$lcssa46 * $10; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $12 = $$lcssa47 * $10; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  if ($4) {
   $13 = ((($this)) + 20|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $14 = ((($this)) + 24|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $I$08 = 0.0;$area$09 = 0.0;$center$sroa$0$05 = 0.0;$center$sroa$10$06 = 0.0;$i1$07 = 0;
   while(1) {
    $46 = (((($this)) + 20|0) + ($i1$07<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $47 = +HEAPF32[$46>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $48 = $47 - $11; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $49 = (((((($this)) + 20|0) + ($i1$07<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $50 = +HEAPF32[$49>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $51 = $50 - $12; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $52 = (($i1$07) + 1)|0; //@line 407 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $53 = ($52|0)<($3|0); //@line 407 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $54 = (((($this)) + 20|0) + ($52<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $55 = (((((($this)) + 20|0) + ($52<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $$sink$in = $53 ? $55 : $14; //@line 407 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $$pn$in = $53 ? $54 : $13; //@line 407 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $$pn = +HEAPF32[$$pn$in>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $e2$sroa$0$0 = $$pn - $11; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $$sink = +HEAPF32[$$sink$in>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $56 = $$sink - $12; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $57 = $48 * $56; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $58 = $51 * $e2$sroa$0$0; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $59 = $57 - $58; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $60 = $59 * 0.5; //@line 411 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $61 = $area$09 + $60; //@line 412 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $62 = $60 * 0.3333333432674408; //@line 415 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $63 = $48 + $e2$sroa$0$0; //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $64 = $51 + $56; //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $65 = $63 * $62; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $66 = $64 * $62; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $67 = $center$sroa$0$05 + $65; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $68 = $center$sroa$10$06 + $66; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $69 = $48 * $48; //@line 420 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $70 = $48 * $e2$sroa$0$0; //@line 420 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $71 = $69 + $70; //@line 420 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $72 = $e2$sroa$0$0 * $e2$sroa$0$0; //@line 420 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $73 = $72 + $71; //@line 420 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $74 = $51 * $51; //@line 421 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $75 = $51 * $56; //@line 421 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $76 = $74 + $75; //@line 421 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $77 = $56 * $56; //@line 421 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $78 = $77 + $76; //@line 421 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $79 = $59 * 0.083333335816860198; //@line 423 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $80 = $73 + $78; //@line 423 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $81 = $79 * $80; //@line 423 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $82 = $I$08 + $81; //@line 423 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    if ($53) {
     $I$08 = $82;$area$09 = $61;$center$sroa$0$05 = $67;$center$sroa$10$06 = $68;$i1$07 = $52;
    } else {
     $29 = $11;$31 = $12;$I$0$lcssa = $82;$area$0$lcssa = $61;$center$sroa$0$0$lcssa = $67;$center$sroa$10$0$lcssa = $68;
     break;
    }
   }
  } else {
   $29 = $11;$31 = $12;$I$0$lcssa = 0.0;$area$0$lcssa = 0.0;$center$sroa$0$0$lcssa = 0.0;$center$sroa$10$0$lcssa = 0.0;
  }
 } else {
  $5 = (+($3|0)); //@line 399 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $6 = 1.0 / $5; //@line 399 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $7 = $6 * 0.0; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $8 = $6 * 0.0; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $29 = $7;$31 = $8;$I$0$lcssa = 0.0;$area$0$lcssa = 0.0;$center$sroa$0$0$lcssa = 0.0;$center$sroa$10$0$lcssa = 0.0;
 }
 $23 = $area$0$lcssa * $density; //@line 427 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 HEAPF32[$massData>>2] = $23; //@line 427 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $24 = $area$0$lcssa > 1.1920928955078125E-7; //@line 430 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 __Z10loveAssertbPKc($24,361352); //@line 430 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $25 = 1.0 / $area$0$lcssa; //@line 431 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $26 = $center$sroa$0$0$lcssa * $25; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 $27 = $center$sroa$10$0$lcssa * $25; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 $28 = $29 + $26; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $30 = $31 + $27; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $32 = ((($massData)) + 4|0); //@line 432 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 HEAPF32[$32>>2] = $28; //@line 432 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $33 = ((($massData)) + 8|0); //@line 432 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 HEAPF32[$33>>2] = $30; //@line 432 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $34 = $I$0$lcssa * $density; //@line 435 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $35 = ((($massData)) + 12|0); //@line 435 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $36 = +HEAPF32[$massData>>2]; //@line 438 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $37 = $28 * $28; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $38 = $30 * $30; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $39 = $37 + $38; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $40 = $26 * $26; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $41 = $27 * $27; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $42 = $40 + $41; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $43 = $39 - $42; //@line 438 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $44 = $36 * $43; //@line 438 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $45 = $34 + $44; //@line 438 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 HEAPF32[$35>>2] = $45; //@line 438 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 return; //@line 439 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
}
function __ZN14b2PolygonShape8SetAsBoxEffRK6b2Vec2f($this,$hx,$hy,$center,$angle) {
 $this = $this|0;
 $hx = +$hx;
 $hy = +$hy;
 $center = $center|0;
 $angle = +$angle;
 var $$phi$trans$insert = 0, $$phi$trans$insert5 = 0, $$pre = 0.0, $$pre6 = 0.0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0;
 var $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$04 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 148|0); //@line 46 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 HEAP32[$0>>2] = 4; //@line 46 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $1 = -$hx; //@line 47 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $2 = -$hy; //@line 47 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $3 = ((($this)) + 20|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$3>>2] = $1; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 24|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$4>>2] = $2; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($this)) + 28|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$5>>2] = $hx; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($this)) + 32|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$6>>2] = $2; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($this)) + 36|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$7>>2] = $hx; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($this)) + 40|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$8>>2] = $hy; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 44|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$9>>2] = $1; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($this)) + 48|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$10>>2] = $hy; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($this)) + 84|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$11>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $12 = ((($this)) + 88|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$12>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($this)) + 92|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$13>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $14 = ((($this)) + 96|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$14>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $15 = ((($this)) + 100|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$15>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $16 = ((($this)) + 104|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$16>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $17 = ((($this)) + 108|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$17>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $18 = ((($this)) + 112|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$18>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($this)) + 12|0); //@line 55 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $20 = $center; //@line 55 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $21 = $20; //@line 55 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $22 = HEAP32[$21>>2]|0; //@line 55 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $23 = (($20) + 4)|0; //@line 55 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $24 = $23; //@line 55 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $25 = HEAP32[$24>>2]|0; //@line 55 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $26 = $19; //@line 55 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $27 = $26; //@line 55 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 HEAP32[$27>>2] = $22; //@line 55 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $28 = (($26) + 4)|0; //@line 55 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $29 = $28; //@line 55 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 HEAP32[$29>>2] = $25; //@line 55 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $30 = +HEAPF32[$center>>2]; //@line 58 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $31 = ((($center)) + 4|0); //@line 58 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $32 = +HEAPF32[$31>>2]; //@line 58 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $33 = (+Math_sin((+$angle))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $34 = (+Math_cos((+$angle))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $40 = $2;$52 = -1.0;$i$04 = 0;
 while(1) {
  $35 = (((($this)) + 20|0) + ($i$04<<3)|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $36 = +HEAPF32[$35>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $37 = $34 * $36; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $38 = (((((($this)) + 20|0) + ($i$04<<3)|0)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $39 = $33 * $40; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $41 = $37 - $39; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $42 = $30 + $41; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $43 = $33 * $36; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $44 = $34 * $40; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $45 = $43 + $44; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $46 = $32 + $45; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$35>>2] = $42; //@line 64 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  HEAPF32[$38>>2] = $46; //@line 64 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $47 = (((($this)) + 84|0) + ($i$04<<3)|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $48 = +HEAPF32[$47>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $49 = $34 * $48; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $50 = (((((($this)) + 84|0) + ($i$04<<3)|0)) + 4|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $51 = $33 * $52; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $53 = $49 - $51; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $54 = $33 * $48; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $55 = $34 * $52; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $56 = $54 + $55; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$47>>2] = $53; //@line 65 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  HEAPF32[$50>>2] = $56; //@line 65 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $57 = (($i$04) + 1)|0; //@line 62 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $58 = HEAP32[$0>>2]|0; //@line 62 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $59 = ($57|0)<($58|0); //@line 62 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  if (!($59)) {
   break;
  }
  $$phi$trans$insert = (((((($this)) + 20|0) + ($57<<3)|0)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $$pre = +HEAPF32[$$phi$trans$insert>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $$phi$trans$insert5 = (((((($this)) + 84|0) + ($57<<3)|0)) + 4|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $$pre6 = +HEAPF32[$$phi$trans$insert5>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $40 = $$pre;$52 = $$pre6;$i$04 = $57;
 }
 return; //@line 67 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
}
function __ZN14b2PolygonShape3SetEPK6b2Vec2i($this,$vertices,$count) {
 $this = $this|0;
 $vertices = $vertices|0;
 $count = $count|0;
 var $$lcssa49 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0.0, $103 = 0, $104 = 0, $105 = 0, $106 = 0.0, $107 = 0, $108 = 0.0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0.0, $16 = 0, $160 = 0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0.0;
 var $17 = 0, $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0.0, $182 = 0.0, $183 = 0.0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0.0, $194 = 0, $195 = 0.0, $196 = 0, $197 = 0.0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0.0, $201 = 0.0, $202 = 0, $203 = 0.0, $204 = 0;
 var $205 = 0, $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0.0, $21 = 0, $210 = 0.0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0.0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0.0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0;
 var $89 = 0.0, $9 = 0, $90 = 0, $91 = 0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0, $99 = 0, $area$0$lcssa$i = 0.0, $area$09$i = 0.0, $count$off = 0, $exitcond = 0, $exitcond$i = 0, $exitcond35 = 0, $exitcond36 = 0, $exitcond37 = 0;
 var $hull = 0, $i$023 = 0, $i$08$i = 0, $i0$013 = 0, $i0$1 = 0, $i0$1$lcssa = 0, $i1$015 = 0, $i4$08 = 0, $i5$07 = 0, $ie$010$us = 0, $ie$1$us = 0, $ie$3$us = 0, $ie$3$us$lcssa = 0, $ih$0$us = 0, $j$020 = 0, $j2$011$us = 0, $m$0$us = 0, $m$0$us$lcssa = 0, $not$$us = 0, $ps = 0;
 var $tempCount$024 = 0, $tempCount$1 = 0, $tempCount$1$lcssa = 0, $x0$014 = 0.0, $x0$1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ps = sp + 32|0;
 $hull = sp;
 $count$off = (($count) + -3)|0; //@line 122 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $0 = ($count$off>>>0)<(6); //@line 122 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 __Z10loveAssertbPKc($0,361370); //@line 122 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $1 = ($count|0)<(3); //@line 123 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 if ($1) {
  $2 = ((($this)) + 148|0); //@line 32 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  HEAP32[$2>>2] = 4; //@line 32 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $3 = ((($this)) + 20|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$3>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $4 = ((($this)) + 24|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$4>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $5 = ((($this)) + 28|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$5>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $6 = ((($this)) + 32|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$6>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $7 = ((($this)) + 36|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$7>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $8 = ((($this)) + 40|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$8>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $9 = ((($this)) + 44|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$9>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $10 = ((($this)) + 48|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$10>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $11 = ((($this)) + 84|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$11>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $12 = ((($this)) + 88|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$12>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $13 = ((($this)) + 92|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$13>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $14 = ((($this)) + 96|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$14>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $15 = ((($this)) + 100|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$15>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $16 = ((($this)) + 104|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$16>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $17 = ((($this)) + 108|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$17>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $18 = ((($this)) + 112|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$18>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $19 = ((($this)) + 12|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$19>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $20 = ((($this)) + 16|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$20>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  STACKTOP = sp;return; //@line 249 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 }
 $21 = ($count|0)<(8); //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $22 = $21 ? $count : 8; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $23 = ($22|0)>(0); //@line 134 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 if ($23) {
  $i$023 = 0;$tempCount$024 = 0;
  while(1) {
   $25 = (($vertices) + ($i$023<<3)|0); //@line 136 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   $26 = HEAP32[$25>>2]|0; //@line 136 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   $27 = (((($vertices) + ($i$023<<3)|0)) + 4|0); //@line 136 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   $28 = HEAP32[$27>>2]|0; //@line 136 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   $29 = ($tempCount$024|0)>(0); //@line 139 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   L8: do {
    if ($29) {
     $30 = (HEAP32[tempDoublePtr>>2]=$26,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $31 = (HEAP32[tempDoublePtr>>2]=$28,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $j$020 = 0;
     while(1) {
      $34 = (($ps) + ($j$020<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $35 = +HEAPF32[$34>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $36 = $30 - $35; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $37 = (((($ps) + ($j$020<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $38 = +HEAPF32[$37>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $39 = $31 - $38; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $40 = $36 * $36; //@line 408 "./libraries/Box2D/Common/b2Math.h"
      $41 = $39 * $39; //@line 408 "./libraries/Box2D/Common/b2Math.h"
      $42 = $40 + $41; //@line 408 "./libraries/Box2D/Common/b2Math.h"
      $43 = $42 < 6.2499998421117198E-6; //@line 141 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $33 = (($j$020) + 1)|0; //@line 139 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      if ($43) {
       $tempCount$1 = $tempCount$024;
       break L8;
      }
      $32 = ($33|0)<($tempCount$024|0); //@line 139 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      if ($32) {
       $j$020 = $33;
      } else {
       label = 9;
       break;
      }
     }
    } else {
     label = 9;
    }
   } while(0);
   if ((label|0) == 9) {
    label = 0;
    $44 = (($tempCount$024) + 1)|0; //@line 150 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $45 = (($ps) + ($tempCount$024<<3)|0); //@line 150 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    HEAP32[$45>>2] = $26; //@line 150 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $46 = (((($ps) + ($tempCount$024<<3)|0)) + 4|0); //@line 150 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    HEAP32[$46>>2] = $28; //@line 150 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $tempCount$1 = $44;
   }
   $47 = (($i$023) + 1)|0; //@line 134 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   $48 = ($47|0)<($22|0); //@line 134 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   if ($48) {
    $i$023 = $47;$tempCount$024 = $tempCount$1;
   } else {
    $tempCount$1$lcssa = $tempCount$1;
    break;
   }
  }
  $24 = ($tempCount$1$lcssa|0)<(3); //@line 155 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  if ($24) {
   label = 11;
  } else {
   $68 = +HEAPF32[$ps>>2]; //@line 168 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   $i0$013 = 0;$i1$015 = 1;$x0$014 = $68;
   while(1) {
    $101 = (($ps) + ($i1$015<<3)|0); //@line 171 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $102 = +HEAPF32[$101>>2]; //@line 171 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $103 = $102 > $x0$014; //@line 172 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    if ($103) {
     label = 24;
    } else {
     $104 = $102 == $x0$014; //@line 172 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
     if ($104) {
      $105 = (((($ps) + ($i1$015<<3)|0)) + 4|0); //@line 172 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $106 = +HEAPF32[$105>>2]; //@line 172 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $107 = (((($ps) + ($i0$013<<3)|0)) + 4|0); //@line 172 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $108 = +HEAPF32[$107>>2]; //@line 172 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $109 = $106 < $108; //@line 172 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      if ($109) {
       label = 24;
      } else {
       $i0$1 = $i0$013;$x0$1 = $x0$014;
      }
     } else {
      $i0$1 = $i0$013;$x0$1 = $x0$014;
     }
    }
    if ((label|0) == 24) {
     label = 0;
     $i0$1 = $i1$015;$x0$1 = $102;
    }
    $110 = (($i1$015) + 1)|0; //@line 169 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $exitcond37 = ($110|0)==($tempCount$1$lcssa|0); //@line 169 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    if ($exitcond37) {
     $i0$1$lcssa = $i0$1;
     break;
    } else {
     $i0$013 = $i0$1;$i1$015 = $110;$x0$014 = $x0$1;
    }
   }
   $ih$0$us = $i0$1$lcssa;$m$0$us = 0;
   while(1) {
    $69 = (($hull) + ($m$0$us<<2)|0); //@line 185 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    HEAP32[$69>>2] = $ih$0$us; //@line 185 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $70 = (($ps) + ($ih$0$us<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $71 = (((($ps) + ($ih$0$us<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $ie$010$us = 0;$j2$011$us = 1;
    while(1) {
     $72 = ($ie$010$us|0)==($ih$0$us|0); //@line 190 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
     if ($72) {
      $ie$3$us = $j2$011$us;
     } else {
      $73 = (($ps) + ($ie$010$us<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $74 = +HEAPF32[$73>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $75 = +HEAPF32[$70>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $76 = $74 - $75; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $77 = (((($ps) + ($ie$010$us<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $78 = +HEAPF32[$77>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $79 = +HEAPF32[$71>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $80 = $78 - $79; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $81 = (($ps) + ($j2$011$us<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $82 = +HEAPF32[$81>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $83 = $82 - $75; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $84 = (((($ps) + ($j2$011$us<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $85 = +HEAPF32[$84>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $86 = $85 - $79; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $87 = $76 * $86; //@line 414 "./libraries/Box2D/Common/b2Math.h"
      $88 = $80 * $83; //@line 414 "./libraries/Box2D/Common/b2Math.h"
      $89 = $87 - $88; //@line 414 "./libraries/Box2D/Common/b2Math.h"
      $90 = $89 < 0.0; //@line 199 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $ie$1$us = $90 ? $j2$011$us : $ie$010$us; //@line 199 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $91 = $89 == 0.0; //@line 205 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      if ($91) {
       $92 = $83 * $83; //@line 110 "./libraries/Box2D/Common/b2Math.h"
       $93 = $86 * $86; //@line 110 "./libraries/Box2D/Common/b2Math.h"
       $94 = $92 + $93; //@line 110 "./libraries/Box2D/Common/b2Math.h"
       $95 = $76 * $76; //@line 110 "./libraries/Box2D/Common/b2Math.h"
       $96 = $80 * $80; //@line 110 "./libraries/Box2D/Common/b2Math.h"
       $97 = $95 + $96; //@line 110 "./libraries/Box2D/Common/b2Math.h"
       $98 = $94 > $97; //@line 205 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
       if ($98) {
        $ie$3$us = $j2$011$us;
       } else {
        $ie$3$us = $ie$1$us;
       }
      } else {
       $ie$3$us = $ie$1$us;
      }
     }
     $99 = (($j2$011$us) + 1)|0; //@line 188 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
     $exitcond36 = ($99|0)==($tempCount$1$lcssa|0); //@line 188 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
     if ($exitcond36) {
      $ie$3$us$lcssa = $ie$3$us;
      break;
     } else {
      $ie$010$us = $ie$3$us;$j2$011$us = $99;
     }
    }
    $100 = (($m$0$us) + 1)|0; //@line 211 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $not$$us = ($ie$3$us$lcssa|0)==($i0$1$lcssa|0); //@line 216 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    if ($not$$us) {
     $$lcssa49 = $100;$m$0$us$lcssa = $m$0$us;
     break;
    } else {
     $ih$0$us = $ie$3$us$lcssa;$m$0$us = $100;
    }
   }
   $111 = ($$lcssa49|0)<(3); //@line 220 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   if ($111) {
    __Z10loveAssertbPKc(0,622525); //@line 223 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $112 = ((($this)) + 148|0); //@line 32 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    HEAP32[$112>>2] = 4; //@line 32 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $113 = ((($this)) + 20|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$113>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $114 = ((($this)) + 24|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$114>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $115 = ((($this)) + 28|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$115>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $116 = ((($this)) + 32|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$116>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $117 = ((($this)) + 36|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$117>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $118 = ((($this)) + 40|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$118>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $119 = ((($this)) + 44|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$119>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $120 = ((($this)) + 48|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$120>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $121 = ((($this)) + 84|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$121>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $122 = ((($this)) + 88|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$122>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $123 = ((($this)) + 92|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$123>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $124 = ((($this)) + 96|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$124>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $125 = ((($this)) + 100|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$125>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $126 = ((($this)) + 104|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$126>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $127 = ((($this)) + 108|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$127>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $128 = ((($this)) + 112|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$128>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
    $129 = ((($this)) + 12|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$129>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
    $130 = ((($this)) + 16|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$130>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
   } else {
    $131 = ((($this)) + 148|0); //@line 228 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    HEAP32[$131>>2] = $$lcssa49; //@line 228 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $132 = ($m$0$us$lcssa|0)<(0); //@line 231 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    if (!($132)) {
     $i4$08 = 0;
     while(1) {
      $133 = (((($this)) + 20|0) + ($i4$08<<3)|0); //@line 233 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $134 = (($hull) + ($i4$08<<2)|0); //@line 233 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $135 = HEAP32[$134>>2]|0; //@line 233 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $136 = (($ps) + ($135<<3)|0); //@line 233 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $137 = $136; //@line 233 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $138 = $137; //@line 233 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $139 = HEAP32[$138>>2]|0; //@line 233 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $140 = (($137) + 4)|0; //@line 233 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $141 = $140; //@line 233 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $142 = HEAP32[$141>>2]|0; //@line 233 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $143 = $133; //@line 233 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $144 = $143; //@line 233 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      HEAP32[$144>>2] = $139; //@line 233 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $145 = (($143) + 4)|0; //@line 233 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $146 = $145; //@line 233 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      HEAP32[$146>>2] = $142; //@line 233 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $147 = (($i4$08) + 1)|0; //@line 231 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $exitcond35 = ($i4$08|0)==($m$0$us$lcssa|0); //@line 231 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      if ($exitcond35) {
       break;
      } else {
       $i4$08 = $147;
      }
     }
     if (!($132)) {
      $i5$07 = 0;
      while(1) {
       $186 = (($i5$07) + 1)|0; //@line 240 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
       $187 = ($i5$07|0)<($m$0$us$lcssa|0); //@line 240 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
       $188 = $187 ? $186 : 0; //@line 240 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
       $189 = (((($this)) + 20|0) + ($188<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
       $190 = +HEAPF32[$189>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
       $191 = (((($this)) + 20|0) + ($i5$07<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
       $192 = +HEAPF32[$191>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
       $193 = $190 - $192; //@line 454 "./libraries/Box2D/Common/b2Math.h"
       $194 = (((((($this)) + 20|0) + ($188<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
       $195 = +HEAPF32[$194>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
       $196 = (((((($this)) + 20|0) + ($i5$07<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
       $197 = +HEAPF32[$196>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
       $198 = $195 - $197; //@line 454 "./libraries/Box2D/Common/b2Math.h"
       $199 = $193 * $193; //@line 110 "./libraries/Box2D/Common/b2Math.h"
       $200 = $198 * $198; //@line 110 "./libraries/Box2D/Common/b2Math.h"
       $201 = $199 + $200; //@line 110 "./libraries/Box2D/Common/b2Math.h"
       $202 = $201 > 1.4210854715202004E-14; //@line 242 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
       __Z10loveAssertbPKc($202,361426); //@line 242 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
       $203 = -$193; //@line 421 "./libraries/Box2D/Common/b2Math.h"
       $204 = (((($this)) + 84|0) + ($i5$07<<3)|0); //@line 243 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
       HEAPF32[$204>>2] = $198; //@line 243 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
       $205 = (((((($this)) + 84|0) + ($i5$07<<3)|0)) + 4|0); //@line 243 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
       HEAPF32[$205>>2] = $203; //@line 243 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
       $206 = (+Math_sqrt((+$201))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
       $207 = $206 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
       if (!($207)) {
        $208 = 1.0 / $206; //@line 121 "./libraries/Box2D/Common/b2Math.h"
        $209 = $198 * $208; //@line 122 "./libraries/Box2D/Common/b2Math.h"
        HEAPF32[$204>>2] = $209; //@line 122 "./libraries/Box2D/Common/b2Math.h"
        $210 = $208 * $203; //@line 123 "./libraries/Box2D/Common/b2Math.h"
        HEAPF32[$205>>2] = $210; //@line 123 "./libraries/Box2D/Common/b2Math.h"
       }
       $exitcond = ($i5$07|0)==($m$0$us$lcssa|0); //@line 237 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
       if ($exitcond) {
        break;
       } else {
        $i5$07 = $186;
       }
      }
     }
    }
    $148 = ((($this)) + 20|0); //@line 248 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $149 = ($m$0$us$lcssa|0)>(1); //@line 76 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    __Z10loveAssertbPKc($149,361415); //@line 76 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $150 = ($m$0$us$lcssa|0)>(-1); //@line 95 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    if ($150) {
     $175 = 0.0;$177 = 0.0;$area$09$i = 0.0;$i$08$i = 0;
     while(1) {
      $151 = (((($this)) + 20|0) + ($i$08$i<<3)|0); //@line 99 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $152 = +HEAPF32[$151>>2]; //@line 99 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $153 = (((((($this)) + 20|0) + ($i$08$i<<3)|0)) + 4|0); //@line 99 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $154 = +HEAPF32[$153>>2]; //@line 99 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $155 = (($i$08$i) + 1)|0; //@line 100 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $156 = ($i$08$i|0)<($m$0$us$lcssa|0); //@line 100 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $157 = (((($this)) + 20|0) + ($155<<3)|0); //@line 100 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $158 = $156 ? $157 : $148; //@line 100 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $159 = +HEAPF32[$158>>2]; //@line 100 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $160 = ((($158)) + 4|0); //@line 100 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $161 = +HEAPF32[$160>>2]; //@line 100 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $162 = $152 * $161; //@line 414 "./libraries/Box2D/Common/b2Math.h"
      $163 = $154 * $159; //@line 414 "./libraries/Box2D/Common/b2Math.h"
      $164 = $162 - $163; //@line 414 "./libraries/Box2D/Common/b2Math.h"
      $165 = $164 * 0.5; //@line 107 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $166 = $area$09$i + $165; //@line 108 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $167 = $165 * 0.3333333432674408; //@line 111 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      $168 = $152 + 0.0; //@line 448 "./libraries/Box2D/Common/b2Math.h"
      $169 = $154 + 0.0; //@line 448 "./libraries/Box2D/Common/b2Math.h"
      $170 = $168 + $159; //@line 448 "./libraries/Box2D/Common/b2Math.h"
      $171 = $169 + $161; //@line 448 "./libraries/Box2D/Common/b2Math.h"
      $172 = $170 * $167; //@line 459 "./libraries/Box2D/Common/b2Math.h"
      $173 = $171 * $167; //@line 459 "./libraries/Box2D/Common/b2Math.h"
      $174 = $175 + $172; //@line 85 "./libraries/Box2D/Common/b2Math.h"
      $176 = $177 + $173; //@line 85 "./libraries/Box2D/Common/b2Math.h"
      $exitcond$i = ($155|0)==($$lcssa49|0); //@line 95 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
      if ($exitcond$i) {
       $181 = $174;$183 = $176;$area$0$lcssa$i = $166;
       break;
      } else {
       $175 = $174;$177 = $176;$area$09$i = $166;$i$08$i = $155;
      }
     }
    } else {
     $181 = 0.0;$183 = 0.0;$area$0$lcssa$i = 0.0;
    }
    $178 = $area$0$lcssa$i > 1.1920928955078125E-7; //@line 115 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    __Z10loveAssertbPKc($178,361352); //@line 115 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $179 = 1.0 / $area$0$lcssa$i; //@line 116 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $180 = $181 * $179; //@line 97 "./libraries/Box2D/Common/b2Math.h"
    $182 = $183 * $179; //@line 97 "./libraries/Box2D/Common/b2Math.h"
    $184 = ((($this)) + 12|0); //@line 248 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    HEAPF32[$184>>2] = $180; //@line 248 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    $185 = ((($this)) + 16|0); //@line 248 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    HEAPF32[$185>>2] = $182; //@line 248 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   }
  }
 } else {
  label = 11;
 }
 if ((label|0) == 11) {
  __Z10loveAssertbPKc(0,622525); //@line 158 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $49 = ((($this)) + 148|0); //@line 32 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  HEAP32[$49>>2] = 4; //@line 32 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $50 = ((($this)) + 20|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$50>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $51 = ((($this)) + 24|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$51>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $52 = ((($this)) + 28|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$52>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $53 = ((($this)) + 32|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$53>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $54 = ((($this)) + 36|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$54>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $55 = ((($this)) + 40|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$55>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $56 = ((($this)) + 44|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$56>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $57 = ((($this)) + 48|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$57>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $58 = ((($this)) + 84|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$58>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $59 = ((($this)) + 88|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$59>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $60 = ((($this)) + 92|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$60>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $61 = ((($this)) + 96|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$61>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $62 = ((($this)) + 100|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$62>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $63 = ((($this)) + 104|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$63>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $64 = ((($this)) + 108|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$64>>2] = -1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $65 = ((($this)) + 112|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$65>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $66 = ((($this)) + 12|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$66>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $67 = ((($this)) + 16|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$67>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 }
 STACKTOP = sp;return; //@line 249 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
}
function __ZNK14b2PolygonShape8ValidateEv($this) {
 $this = $this|0;
 var $$36 = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0.0, $25 = 0.0;
 var $26 = 0.0, $27 = 0.0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0, $i$013$us = 0, $j$012$us = 0, $or$cond$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 148|0); //@line 443 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 450 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $2 = ($1|0)>(0); //@line 443 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 if (!($2)) {
  $$36 = 1;
  return ($$36|0); //@line 467 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 }
 $3 = (($1) + -1)|0; //@line 446 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 $i$013$us = 0;
 L4: while(1) {
  $4 = ($i$013$us|0)<($3|0); //@line 446 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $5 = (($i$013$us) + 1)|0; //@line 446 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $6 = $4 ? $5 : 0; //@line 446 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $7 = (((($this)) + 20|0) + ($i$013$us<<3)|0); //@line 447 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $8 = +HEAPF32[$7>>2]; //@line 447 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $9 = (((((($this)) + 20|0) + ($i$013$us<<3)|0)) + 4|0); //@line 447 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $10 = +HEAPF32[$9>>2]; //@line 447 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  $11 = (((($this)) + 20|0) + ($6<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $12 = +HEAPF32[$11>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $13 = $12 - $8; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $14 = (((((($this)) + 20|0) + ($6<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $15 = +HEAPF32[$14>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $16 = $15 - $10; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $j$012$us = 0;
  while(1) {
   $17 = ($j$012$us|0)==($i$013$us|0); //@line 452 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   $18 = ($j$012$us|0)==($6|0); //@line 452 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   $or$cond$us = $17 | $18; //@line 452 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   if (!($or$cond$us)) {
    $19 = (((($this)) + 20|0) + ($j$012$us<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $20 = +HEAPF32[$19>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $21 = $20 - $8; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $22 = (((((($this)) + 20|0) + ($j$012$us<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $23 = +HEAPF32[$22>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $24 = $23 - $10; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $25 = $13 * $24; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $26 = $16 * $21; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $27 = $25 - $26; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $28 = $27 < 0.0; //@line 459 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
    if ($28) {
     $$36 = 0;
     label = 8;
     break L4;
    }
   }
   $29 = (($j$012$us) + 1)|0; //@line 450 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   $30 = ($29|0)<($1|0); //@line 450 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
   if ($30) {
    $j$012$us = $29;
   } else {
    break;
   }
  }
  $31 = ($5|0)<($1|0); //@line 443 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
  if ($31) {
   $i$013$us = $5;
  } else {
   $$36 = 1;
   label = 8;
   break;
  }
 }
 if ((label|0) == 8) {
  return ($$36|0); //@line 467 "libraries/Box2D/Collision/Shapes/b2PolygonShape.cpp"
 }
 return (0)|0;
}
function __ZN16b2BlockAllocatorC2Ev($this) {
 $this = $this|0;
 var $$j$01 = 0, $$j$013 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$02 = 0;
 var $j$01 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 __Z10loveAssertbPKc(1,361473); //@line 58 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $0 = ((($this)) + 8|0); //@line 60 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 HEAP32[$0>>2] = 128; //@line 60 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $1 = ((($this)) + 4|0); //@line 61 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 HEAP32[$1>>2] = 0; //@line 61 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $2 = (__Z7b2Alloci(1024)|0); //@line 62 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 HEAP32[$this>>2] = $2; //@line 62 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $3 = HEAP32[$0>>2]|0; //@line 64 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $4 = $3 << 3; //@line 64 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 _memset(($2|0),0,($4|0))|0; //@line 64 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $5 = ((($this)) + 12|0); //@line 65 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 dest=$5; stop=dest+56|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 65 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $6 = HEAP8[1251124]|0; //@line 67 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $7 = ($6<<24>>24)==(0); //@line 67 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 if ($7) {
  $i$02 = 1;$j$01 = 0;
 } else {
  return; //@line 86 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 }
 while(1) {
  $8 = ($j$01|0)<(14); //@line 72 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  __Z10loveAssertbPKc($8,361499); //@line 72 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $9 = (20528 + ($j$01<<2)|0); //@line 73 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $10 = HEAP32[$9>>2]|0; //@line 73 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $11 = ($i$02|0)>($10|0); //@line 73 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $12 = (($j$01) + 1)|0; //@line 79 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $$j$01 = $11 ? $12 : $j$01; //@line 73 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $$j$013 = $11 ? $12 : $j$01; //@line 73 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $$sink = $$j$01&255; //@line 75 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $13 = (1250483 + ($i$02)|0); //@line 75 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  HEAP8[$13>>0] = $$sink; //@line 75 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $14 = (($i$02) + 1)|0; //@line 70 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $exitcond = ($14|0)==(641); //@line 70 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  if ($exitcond) {
   break;
  } else {
   $i$02 = $14;$j$01 = $$j$013;
  }
 }
 HEAP8[1251124] = 1; //@line 84 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 return; //@line 86 "libraries/Box2D/Common/b2BlockAllocator.cpp"
}
function __ZN16b2BlockAllocatorD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $lpad$phi$index = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 90 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 90 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $2 = ($1|0)>(0); //@line 90 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 L1: do {
  if ($2) {
   $i$01 = 0;
   while(1) {
    $6 = HEAP32[$this>>2]|0; //@line 92 "libraries/Box2D/Common/b2BlockAllocator.cpp"
    $7 = (((($6) + ($i$01<<3)|0)) + 4|0); //@line 92 "libraries/Box2D/Common/b2BlockAllocator.cpp"
    $8 = HEAP32[$7>>2]|0; //@line 92 "libraries/Box2D/Common/b2BlockAllocator.cpp"
    __THREW__ = 0;
    invoke_vi(3816,($8|0)); //@line 92 "libraries/Box2D/Common/b2BlockAllocator.cpp"
    $9 = __THREW__; __THREW__ = 0;
    $10 = $9&1;
    if ($10) {
     break;
    }
    $11 = (($i$01) + 1)|0; //@line 90 "libraries/Box2D/Common/b2BlockAllocator.cpp"
    $12 = HEAP32[$0>>2]|0; //@line 90 "libraries/Box2D/Common/b2BlockAllocator.cpp"
    $13 = ($11|0)<($12|0); //@line 90 "libraries/Box2D/Common/b2BlockAllocator.cpp"
    if ($13) {
     $i$01 = $11;
    } else {
     break L1;
    }
   }
   $14 = ___cxa_find_matching_catch_3(0|0)|0;
   $15 = tempRet0;
   $lpad$phi$index = $14;
   ___clang_call_terminate($lpad$phi$index); //@line 92 "libraries/Box2D/Common/b2BlockAllocator.cpp"
   // unreachable; //@line 92 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  }
 } while(0);
 $3 = HEAP32[$this>>2]|0; //@line 95 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 __THREW__ = 0;
 invoke_vi(3816,($3|0)); //@line 95 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  return; //@line 96 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 }
 $16 = ___cxa_find_matching_catch_3(0|0)|0;
 $17 = tempRet0;
 $lpad$phi$index = $16;
 ___clang_call_terminate($lpad$phi$index); //@line 92 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 // unreachable; //@line 92 "libraries/Box2D/Common/b2BlockAllocator.cpp"
}
function __ZN16b2BlockAllocator8AllocateEi($this,$size) {
 $this = $this|0;
 $size = $size|0;
 var $$1 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($size|0)==(0); //@line 100 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 if ($0) {
  $$1 = 0;
  return ($$1|0); //@line 154 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 }
 $1 = ($size|0)>(0); //@line 103 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 __Z10loveAssertbPKc($1,361517); //@line 103 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $2 = ($size|0)>(640); //@line 105 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 if ($2) {
  $3 = (__Z7b2Alloci($size)|0); //@line 107 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $$1 = $3;
  return ($$1|0); //@line 154 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 }
 $4 = (1250483 + ($size)|0); //@line 110 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $5 = HEAP8[$4>>0]|0; //@line 110 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $6 = $5&255; //@line 110 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $7 = ($5&255)<(14); //@line 111 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 __Z10loveAssertbPKc($7,361526); //@line 111 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $8 = (((($this)) + 12|0) + ($6<<2)|0); //@line 113 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 113 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $10 = ($9|0)==(0|0); //@line 113 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 if (!($10)) {
  $11 = HEAP32[$9>>2]|0; //@line 116 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  HEAP32[$8>>2] = $11; //@line 116 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $$1 = $9;
  return ($$1|0); //@line 154 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 }
 $12 = ((($this)) + 4|0); //@line 121 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 121 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $14 = ((($this)) + 8|0); //@line 121 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 121 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $16 = ($13|0)==($15|0); //@line 121 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 if ($16) {
  $17 = HEAP32[$this>>2]|0; //@line 123 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $18 = (($13) + 128)|0; //@line 124 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  HEAP32[$14>>2] = $18; //@line 124 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $19 = $18 << 3; //@line 125 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $20 = (__Z7b2Alloci($19)|0); //@line 125 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  HEAP32[$this>>2] = $20; //@line 125 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $21 = HEAP32[$12>>2]|0; //@line 126 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $22 = $21 << 3; //@line 126 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  _memcpy(($20|0),($17|0),($22|0))|0; //@line 126 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $23 = HEAP32[$this>>2]|0; //@line 127 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $24 = HEAP32[$12>>2]|0; //@line 127 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $25 = (($23) + ($24<<3)|0); //@line 127 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  _memset(($25|0),0,1024)|0; //@line 127 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  __Z6b2FreePv($17); //@line 128 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $$pre = HEAP32[$12>>2]|0; //@line 131 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $29 = $$pre;
 } else {
  $29 = $13;
 }
 $26 = HEAP32[$this>>2]|0; //@line 131 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $27 = (__Z7b2Alloci(16384)|0); //@line 132 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $28 = (((($26) + ($29<<3)|0)) + 4|0); //@line 132 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 HEAP32[$28>>2] = $27; //@line 132 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $30 = (20528 + ($6<<2)|0); //@line 136 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 136 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $32 = (($26) + ($29<<3)|0); //@line 137 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 HEAP32[$32>>2] = $31; //@line 137 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $33 = (16384 / ($31|0))&-1; //@line 138 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $34 = Math_imul($33, $31)|0; //@line 139 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $35 = ($34|0)<(16385); //@line 139 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 __Z10loveAssertbPKc($35,361562); //@line 139 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $36 = (($33) + -1)|0; //@line 140 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $37 = ($33|0)>(1); //@line 140 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $38 = HEAP32[$28>>2]|0; //@line 142 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 if ($37) {
  $i$02 = 0;
  while(1) {
   $45 = Math_imul($i$02, $31)|0; //@line 142 "libraries/Box2D/Common/b2BlockAllocator.cpp"
   $46 = (($38) + ($45)|0); //@line 142 "libraries/Box2D/Common/b2BlockAllocator.cpp"
   $47 = (($i$02) + 1)|0; //@line 143 "libraries/Box2D/Common/b2BlockAllocator.cpp"
   $48 = Math_imul($47, $31)|0; //@line 143 "libraries/Box2D/Common/b2BlockAllocator.cpp"
   $49 = (($38) + ($48)|0); //@line 143 "libraries/Box2D/Common/b2BlockAllocator.cpp"
   HEAP32[$46>>2] = $49; //@line 144 "libraries/Box2D/Common/b2BlockAllocator.cpp"
   $exitcond = ($47|0)==($36|0); //@line 140 "libraries/Box2D/Common/b2BlockAllocator.cpp"
   if ($exitcond) {
    break;
   } else {
    $i$02 = $47;
   }
  }
 }
 $39 = Math_imul($36, $31)|0; //@line 146 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $40 = (($38) + ($39)|0); //@line 146 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 HEAP32[$40>>2] = 0; //@line 147 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $41 = HEAP32[$38>>2]|0; //@line 149 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 HEAP32[$8>>2] = $41; //@line 149 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $42 = HEAP32[$12>>2]|0; //@line 150 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $43 = (($42) + 1)|0; //@line 150 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 HEAP32[$12>>2] = $43; //@line 150 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $44 = HEAP32[$28>>2]|0; //@line 152 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $$1 = $44;
 return ($$1|0); //@line 154 "libraries/Box2D/Common/b2BlockAllocator.cpp"
}
function __ZN16b2BlockAllocator4FreeEPvi($this,$p,$size) {
 $this = $this|0;
 $p = $p|0;
 $size = $size|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($size|0)==(0); //@line 158 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 if ($0) {
  return; //@line 203 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 }
 $1 = ($size|0)>(0); //@line 163 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 __Z10loveAssertbPKc($1,361517); //@line 163 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 $2 = ($size|0)>(640); //@line 165 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 if ($2) {
  __Z6b2FreePv($p); //@line 167 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  return; //@line 203 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 } else {
  $3 = (1250483 + ($size)|0); //@line 171 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $4 = HEAP8[$3>>0]|0; //@line 171 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $5 = $4&255; //@line 171 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $6 = ($4&255)<(14); //@line 172 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  __Z10loveAssertbPKc($6,361526); //@line 172 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $7 = (((($this)) + 12|0) + ($5<<2)|0); //@line 201 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  $8 = HEAP32[$7>>2]|0; //@line 201 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  HEAP32[$p>>2] = $8; //@line 201 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  HEAP32[$7>>2] = $p; //@line 202 "libraries/Box2D/Common/b2BlockAllocator.cpp"
  return; //@line 203 "libraries/Box2D/Common/b2BlockAllocator.cpp"
 }
}
function __ZNK7b2Mat337Solve33ERK6b2Vec3($agg$result,$this,$b) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $b = $b|0;
 var $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0.0;
 var $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0.0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0.0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0;
 var $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0, $6 = 0.0, $60 = 0.0, $61 = 0.0, $62 = 0.0;
 var $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0, $8 = 0.0, $9 = 0.0, $det$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $1 = +HEAPF32[$0>>2]; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $2 = ((($this)) + 32|0); //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $4 = $1 * $3; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($this)) + 20|0); //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($this)) + 28|0); //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $9 = $6 * $8; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $10 = $4 - $9; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($this)) + 24|0); //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$11>>2]; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $13 = $6 * $12; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $14 = ((($this)) + 12|0); //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $15 = +HEAPF32[$14>>2]; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $16 = $3 * $15; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $17 = $13 - $16; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $18 = $8 * $15; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $19 = $1 * $12; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $20 = $18 - $19; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $21 = +HEAPF32[$this>>2]; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $22 = $10 * $21; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $23 = ((($this)) + 4|0); //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $24 = +HEAPF32[$23>>2]; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $25 = $24 * $17; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $26 = $22 + $25; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($this)) + 8|0); //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $28 = +HEAPF32[$27>>2]; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $29 = $20 * $28; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $30 = $26 + $29; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $31 = $30 != 0.0; //@line 28 "libraries/Box2D/Common/b2Math.cpp"
 $32 = 1.0 / $30; //@line 30 "libraries/Box2D/Common/b2Math.cpp"
 $det$0 = $31 ? $32 : $30; //@line 28 "libraries/Box2D/Common/b2Math.cpp"
 $33 = +HEAPF32[$b>>2]; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $34 = $10 * $33; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $35 = ((($b)) + 4|0); //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $36 = +HEAPF32[$35>>2]; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $37 = $17 * $36; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $38 = $34 + $37; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $39 = ((($b)) + 8|0); //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $40 = +HEAPF32[$39>>2]; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $41 = $20 * $40; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $42 = $38 + $41; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $43 = $det$0 * $42; //@line 33 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$agg$result>>2] = $43; //@line 33 "libraries/Box2D/Common/b2Math.cpp"
 $44 = $36 * $3; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $45 = $40 * $8; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $46 = $44 - $45; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $47 = $40 * $12; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $48 = $3 * $33; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $49 = $47 - $48; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $50 = $8 * $33; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $51 = $36 * $12; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $52 = $50 - $51; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $53 = $46 * $21; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $54 = $24 * $49; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $55 = $53 + $54; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $56 = $52 * $28; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $57 = $55 + $56; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $58 = $det$0 * $57; //@line 34 "libraries/Box2D/Common/b2Math.cpp"
 $59 = ((($agg$result)) + 4|0); //@line 34 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$59>>2] = $58; //@line 34 "libraries/Box2D/Common/b2Math.cpp"
 $60 = $1 * $40; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $61 = $6 * $36; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $62 = $60 - $61; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $63 = $6 * $33; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $64 = $40 * $15; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $65 = $63 - $64; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $66 = $36 * $15; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $67 = $1 * $33; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $68 = $66 - $67; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $69 = $62 * $21; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $70 = $24 * $65; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $71 = $69 + $70; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $72 = $68 * $28; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $73 = $71 + $72; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $74 = $det$0 * $73; //@line 35 "libraries/Box2D/Common/b2Math.cpp"
 $75 = ((($agg$result)) + 8|0); //@line 35 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$75>>2] = $74; //@line 35 "libraries/Box2D/Common/b2Math.cpp"
 return; //@line 37 "libraries/Box2D/Common/b2Math.cpp"
}
function __ZNK7b2Mat337Solve22ERK6b2Vec2($agg$result,$this,$b) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $b = $b|0;
 var $0 = 0.0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0, $3 = 0, $4 = 0.0, $5 = 0;
 var $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $det$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = +HEAPF32[$this>>2]; //@line 43 "libraries/Box2D/Common/b2Math.cpp"
 $1 = ((($this)) + 12|0); //@line 43 "libraries/Box2D/Common/b2Math.cpp"
 $2 = +HEAPF32[$1>>2]; //@line 43 "libraries/Box2D/Common/b2Math.cpp"
 $3 = ((($this)) + 4|0); //@line 43 "libraries/Box2D/Common/b2Math.cpp"
 $4 = +HEAPF32[$3>>2]; //@line 43 "libraries/Box2D/Common/b2Math.cpp"
 $5 = ((($this)) + 16|0); //@line 43 "libraries/Box2D/Common/b2Math.cpp"
 $6 = +HEAPF32[$5>>2]; //@line 43 "libraries/Box2D/Common/b2Math.cpp"
 $7 = $0 * $6; //@line 44 "libraries/Box2D/Common/b2Math.cpp"
 $8 = $2 * $4; //@line 44 "libraries/Box2D/Common/b2Math.cpp"
 $9 = $7 - $8; //@line 44 "libraries/Box2D/Common/b2Math.cpp"
 $10 = $9 != 0.0; //@line 45 "libraries/Box2D/Common/b2Math.cpp"
 $11 = 1.0 / $9; //@line 47 "libraries/Box2D/Common/b2Math.cpp"
 $det$0 = $10 ? $11 : $9; //@line 45 "libraries/Box2D/Common/b2Math.cpp"
 $12 = +HEAPF32[$b>>2]; //@line 50 "libraries/Box2D/Common/b2Math.cpp"
 $13 = $6 * $12; //@line 50 "libraries/Box2D/Common/b2Math.cpp"
 $14 = ((($b)) + 4|0); //@line 50 "libraries/Box2D/Common/b2Math.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 50 "libraries/Box2D/Common/b2Math.cpp"
 $16 = $2 * $15; //@line 50 "libraries/Box2D/Common/b2Math.cpp"
 $17 = $13 - $16; //@line 50 "libraries/Box2D/Common/b2Math.cpp"
 $18 = $det$0 * $17; //@line 50 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$agg$result>>2] = $18; //@line 50 "libraries/Box2D/Common/b2Math.cpp"
 $19 = $0 * $15; //@line 51 "libraries/Box2D/Common/b2Math.cpp"
 $20 = $4 * $12; //@line 51 "libraries/Box2D/Common/b2Math.cpp"
 $21 = $19 - $20; //@line 51 "libraries/Box2D/Common/b2Math.cpp"
 $22 = $det$0 * $21; //@line 51 "libraries/Box2D/Common/b2Math.cpp"
 $23 = ((($agg$result)) + 4|0); //@line 51 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$23>>2] = $22; //@line 51 "libraries/Box2D/Common/b2Math.cpp"
 return; //@line 53 "libraries/Box2D/Common/b2Math.cpp"
}
function __ZNK7b2Mat3312GetInverse22EPS_($this,$M) {
 $this = $this|0;
 $M = $M|0;
 var $0 = 0.0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0, $19 = 0.0, $2 = 0.0, $20 = 0, $21 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0.0;
 var $8 = 0.0, $9 = 0.0, $det$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = +HEAPF32[$this>>2]; //@line 58 "libraries/Box2D/Common/b2Math.cpp"
 $1 = ((($this)) + 12|0); //@line 58 "libraries/Box2D/Common/b2Math.cpp"
 $2 = +HEAPF32[$1>>2]; //@line 58 "libraries/Box2D/Common/b2Math.cpp"
 $3 = ((($this)) + 4|0); //@line 58 "libraries/Box2D/Common/b2Math.cpp"
 $4 = +HEAPF32[$3>>2]; //@line 58 "libraries/Box2D/Common/b2Math.cpp"
 $5 = ((($this)) + 16|0); //@line 58 "libraries/Box2D/Common/b2Math.cpp"
 $6 = +HEAPF32[$5>>2]; //@line 58 "libraries/Box2D/Common/b2Math.cpp"
 $7 = $0 * $6; //@line 59 "libraries/Box2D/Common/b2Math.cpp"
 $8 = $2 * $4; //@line 59 "libraries/Box2D/Common/b2Math.cpp"
 $9 = $7 - $8; //@line 59 "libraries/Box2D/Common/b2Math.cpp"
 $10 = $9 != 0.0; //@line 60 "libraries/Box2D/Common/b2Math.cpp"
 $11 = 1.0 / $9; //@line 62 "libraries/Box2D/Common/b2Math.cpp"
 $det$0 = $10 ? $11 : $9; //@line 60 "libraries/Box2D/Common/b2Math.cpp"
 $12 = $6 * $det$0; //@line 65 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$M>>2] = $12; //@line 65 "libraries/Box2D/Common/b2Math.cpp"
 $13 = -$det$0; //@line 65 "libraries/Box2D/Common/b2Math.cpp"
 $14 = $2 * $13; //@line 65 "libraries/Box2D/Common/b2Math.cpp"
 $15 = ((($M)) + 12|0); //@line 65 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$15>>2] = $14; //@line 65 "libraries/Box2D/Common/b2Math.cpp"
 $16 = ((($M)) + 8|0); //@line 65 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$16>>2] = 0.0; //@line 65 "libraries/Box2D/Common/b2Math.cpp"
 $17 = $4 * $13; //@line 66 "libraries/Box2D/Common/b2Math.cpp"
 $18 = ((($M)) + 4|0); //@line 66 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$18>>2] = $17; //@line 66 "libraries/Box2D/Common/b2Math.cpp"
 $19 = $0 * $det$0; //@line 66 "libraries/Box2D/Common/b2Math.cpp"
 $20 = ((($M)) + 16|0); //@line 66 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$20>>2] = $19; //@line 66 "libraries/Box2D/Common/b2Math.cpp"
 $21 = ((($M)) + 20|0); //@line 66 "libraries/Box2D/Common/b2Math.cpp"
 ;HEAP32[$21>>2]=0|0;HEAP32[$21+4>>2]=0|0;HEAP32[$21+8>>2]=0|0;HEAP32[$21+12>>2]=0|0; //@line 67 "libraries/Box2D/Common/b2Math.cpp"
 return; //@line 68 "libraries/Box2D/Common/b2Math.cpp"
}
function __ZNK7b2Mat3315GetSymInverse33EPS_($this,$M) {
 $this = $this|0;
 $M = $M|0;
 var $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0.0;
 var $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0.0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0.0, $40 = 0.0, $41 = 0, $42 = 0, $43 = 0.0, $44 = 0.0;
 var $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0, $54 = 0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0, $6 = 0.0, $7 = 0, $8 = 0.0, $9 = 0.0;
 var $det$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $1 = +HEAPF32[$0>>2]; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $2 = ((($this)) + 32|0); //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $4 = $1 * $3; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($this)) + 20|0); //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($this)) + 28|0); //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $9 = $6 * $8; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $10 = $4 - $9; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($this)) + 24|0); //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$11>>2]; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $13 = $6 * $12; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $14 = ((($this)) + 12|0); //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $15 = +HEAPF32[$14>>2]; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $16 = $3 * $15; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $17 = $13 - $16; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $18 = $8 * $15; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $19 = $1 * $12; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $20 = $18 - $19; //@line 505 "./libraries/Box2D/Common/b2Math.h"
 $21 = +HEAPF32[$this>>2]; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $22 = $10 * $21; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $23 = ((($this)) + 4|0); //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $24 = +HEAPF32[$23>>2]; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $25 = $24 * $17; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $26 = $22 + $25; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($this)) + 8|0); //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $28 = +HEAPF32[$27>>2]; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $29 = $20 * $28; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $30 = $26 + $29; //@line 499 "./libraries/Box2D/Common/b2Math.h"
 $31 = $30 != 0.0; //@line 74 "libraries/Box2D/Common/b2Math.cpp"
 $32 = 1.0 / $30; //@line 76 "libraries/Box2D/Common/b2Math.cpp"
 $det$0 = $31 ? $32 : $30; //@line 74 "libraries/Box2D/Common/b2Math.cpp"
 $33 = $8 * $8; //@line 83 "libraries/Box2D/Common/b2Math.cpp"
 $34 = $4 - $33; //@line 83 "libraries/Box2D/Common/b2Math.cpp"
 $35 = $det$0 * $34; //@line 83 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$M>>2] = $35; //@line 83 "libraries/Box2D/Common/b2Math.cpp"
 $36 = $12 * $8; //@line 84 "libraries/Box2D/Common/b2Math.cpp"
 $37 = $36 - $16; //@line 84 "libraries/Box2D/Common/b2Math.cpp"
 $38 = $det$0 * $37; //@line 84 "libraries/Box2D/Common/b2Math.cpp"
 $39 = ((($M)) + 4|0); //@line 84 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$39>>2] = $38; //@line 84 "libraries/Box2D/Common/b2Math.cpp"
 $40 = $det$0 * $20; //@line 85 "libraries/Box2D/Common/b2Math.cpp"
 $41 = ((($M)) + 8|0); //@line 85 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$41>>2] = $40; //@line 85 "libraries/Box2D/Common/b2Math.cpp"
 $42 = ((($M)) + 12|0); //@line 87 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$42>>2] = $38; //@line 87 "libraries/Box2D/Common/b2Math.cpp"
 $43 = $21 * $3; //@line 88 "libraries/Box2D/Common/b2Math.cpp"
 $44 = $12 * $12; //@line 88 "libraries/Box2D/Common/b2Math.cpp"
 $45 = $43 - $44; //@line 88 "libraries/Box2D/Common/b2Math.cpp"
 $46 = $det$0 * $45; //@line 88 "libraries/Box2D/Common/b2Math.cpp"
 $47 = ((($M)) + 16|0); //@line 88 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$47>>2] = $46; //@line 88 "libraries/Box2D/Common/b2Math.cpp"
 $48 = $12 * $15; //@line 89 "libraries/Box2D/Common/b2Math.cpp"
 $49 = $21 * $8; //@line 89 "libraries/Box2D/Common/b2Math.cpp"
 $50 = $48 - $49; //@line 89 "libraries/Box2D/Common/b2Math.cpp"
 $51 = $det$0 * $50; //@line 89 "libraries/Box2D/Common/b2Math.cpp"
 $52 = ((($M)) + 20|0); //@line 89 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$52>>2] = $51; //@line 89 "libraries/Box2D/Common/b2Math.cpp"
 $53 = ((($M)) + 24|0); //@line 91 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$53>>2] = $40; //@line 91 "libraries/Box2D/Common/b2Math.cpp"
 $54 = ((($M)) + 28|0); //@line 92 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$54>>2] = $51; //@line 92 "libraries/Box2D/Common/b2Math.cpp"
 $55 = $21 * $1; //@line 93 "libraries/Box2D/Common/b2Math.cpp"
 $56 = $15 * $15; //@line 93 "libraries/Box2D/Common/b2Math.cpp"
 $57 = $55 - $56; //@line 93 "libraries/Box2D/Common/b2Math.cpp"
 $58 = $det$0 * $57; //@line 93 "libraries/Box2D/Common/b2Math.cpp"
 $59 = ((($M)) + 32|0); //@line 93 "libraries/Box2D/Common/b2Math.cpp"
 HEAPF32[$59>>2] = $58; //@line 93 "libraries/Box2D/Common/b2Math.cpp"
 return; //@line 94 "libraries/Box2D/Common/b2Math.cpp"
}
function __Z7b2Alloci($size) {
 $size = $size|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_malloc($size)|0); //@line 31 "libraries/Box2D/Common/b2Settings.cpp"
 return ($0|0); //@line 31 "libraries/Box2D/Common/b2Settings.cpp"
}
function __Z6b2FreePv($mem) {
 $mem = $mem|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _free($mem); //@line 36 "libraries/Box2D/Common/b2Settings.cpp"
 return; //@line 37 "libraries/Box2D/Common/b2Settings.cpp"
}
function __Z5b2LogPKcz($string,$varargs) {
 $string = $string|0;
 $varargs = $varargs|0;
 var $args = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $args = sp;
 HEAP32[$args>>2] = $varargs; //@line 43 "libraries/Box2D/Common/b2Settings.cpp"
 (_vprintf($string,$args)|0); //@line 44 "libraries/Box2D/Common/b2Settings.cpp"
 STACKTOP = sp;return; //@line 46 "libraries/Box2D/Common/b2Settings.cpp"
}
function __Z10loveAssertbPKc($test,$teststr) {
 $test = $test|0;
 $teststr = $teststr|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 48 "libraries/Box2D/Common/b2Settings.cpp"
 if ($test) {
  STACKTOP = sp;return; //@line 52 "libraries/Box2D/Common/b2Settings.cpp"
 }
 $0 = (___cxa_allocate_exception(16)|0); //@line 51 "libraries/Box2D/Common/b2Settings.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $teststr; //@line 51 "libraries/Box2D/Common/b2Settings.cpp"
 invoke_viii(3349,($0|0),(361601|0),($vararg_buffer|0)); //@line 51 "libraries/Box2D/Common/b2Settings.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  ___cxa_free_exception(($0|0)); //@line 51 "libraries/Box2D/Common/b2Settings.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  ___cxa_throw(($0|0),(96|0),(344|0)); //@line 51 "libraries/Box2D/Common/b2Settings.cpp"
  // unreachable; //@line 51 "libraries/Box2D/Common/b2Settings.cpp"
 }
}
function __ZN7b2WorldC2ERK6b2Vec2($this,$gravity) {
 $this = $this|0;
 $gravity = $gravity|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN16b2BlockAllocatorC2Ev($this); //@line 37 "libraries/Box2D/Dynamics/b2World.cpp"
 $0 = ((($this)) + 68|0); //@line 37 "libraries/Box2D/Dynamics/b2World.cpp"
 __THREW__ = 0;
 invoke_vi(3817,($0|0)); //@line 37 "libraries/Box2D/Dynamics/b2World.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $29 = ___cxa_find_matching_catch_2()|0;
  $30 = tempRet0;
  $$0 = $30;$$01 = $29;
  __ZN16b2BlockAllocatorD2Ev($this); //@line 64 "libraries/Box2D/Dynamics/b2World.cpp"
  ___resumeException($$01|0);
  // unreachable;
 }
 $3 = ((($this)) + 102872|0); //@line 37 "libraries/Box2D/Dynamics/b2World.cpp"
 __THREW__ = 0;
 invoke_vi(3818,($3|0)); //@line 37 "libraries/Box2D/Dynamics/b2World.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  $6 = ((($this)) + 102968|0); //@line 37 "libraries/Box2D/Dynamics/b2World.cpp"
  $7 = ((($this)) + 102980|0); //@line 39 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$7>>2] = 0; //@line 39 "libraries/Box2D/Dynamics/b2World.cpp"
  $8 = ((($this)) + 102984|0); //@line 40 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$8>>2] = 0; //@line 40 "libraries/Box2D/Dynamics/b2World.cpp"
  $9 = ((($this)) + 102952|0); //@line 42 "libraries/Box2D/Dynamics/b2World.cpp"
  $10 = ((($this)) + 102992|0); //@line 48 "libraries/Box2D/Dynamics/b2World.cpp"
  ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;HEAP32[$9+12>>2]=0|0; //@line 42 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP8[$10>>0] = 1; //@line 48 "libraries/Box2D/Dynamics/b2World.cpp"
  $11 = ((($this)) + 102993|0); //@line 49 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP8[$11>>0] = 1; //@line 49 "libraries/Box2D/Dynamics/b2World.cpp"
  $12 = ((($this)) + 102994|0); //@line 50 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP8[$12>>0] = 0; //@line 50 "libraries/Box2D/Dynamics/b2World.cpp"
  $13 = ((($this)) + 102995|0); //@line 52 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP8[$13>>0] = 1; //@line 52 "libraries/Box2D/Dynamics/b2World.cpp"
  $14 = ((($this)) + 102976|0); //@line 54 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP8[$14>>0] = 1; //@line 54 "libraries/Box2D/Dynamics/b2World.cpp"
  $15 = $gravity; //@line 55 "libraries/Box2D/Dynamics/b2World.cpp"
  $16 = $15; //@line 55 "libraries/Box2D/Dynamics/b2World.cpp"
  $17 = HEAP32[$16>>2]|0; //@line 55 "libraries/Box2D/Dynamics/b2World.cpp"
  $18 = (($15) + 4)|0; //@line 55 "libraries/Box2D/Dynamics/b2World.cpp"
  $19 = $18; //@line 55 "libraries/Box2D/Dynamics/b2World.cpp"
  $20 = HEAP32[$19>>2]|0; //@line 55 "libraries/Box2D/Dynamics/b2World.cpp"
  $21 = $6; //@line 55 "libraries/Box2D/Dynamics/b2World.cpp"
  $22 = $21; //@line 55 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$22>>2] = $17; //@line 55 "libraries/Box2D/Dynamics/b2World.cpp"
  $23 = (($21) + 4)|0; //@line 55 "libraries/Box2D/Dynamics/b2World.cpp"
  $24 = $23; //@line 55 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$24>>2] = $20; //@line 55 "libraries/Box2D/Dynamics/b2World.cpp"
  $25 = ((($this)) + 102868|0); //@line 57 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$25>>2] = 4; //@line 57 "libraries/Box2D/Dynamics/b2World.cpp"
  $26 = ((($this)) + 102988|0); //@line 59 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAPF32[$26>>2] = 0.0; //@line 59 "libraries/Box2D/Dynamics/b2World.cpp"
  $27 = ((($this)) + 102948|0); //@line 61 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$27>>2] = $this; //@line 61 "libraries/Box2D/Dynamics/b2World.cpp"
  $28 = ((($this)) + 102996|0); //@line 63 "libraries/Box2D/Dynamics/b2World.cpp"
  ;HEAP32[$28>>2]=0|0;HEAP32[$28+4>>2]=0|0;HEAP32[$28+8>>2]=0|0;HEAP32[$28+12>>2]=0|0;HEAP32[$28+16>>2]=0|0;HEAP32[$28+20>>2]=0|0;HEAP32[$28+24>>2]=0|0;HEAP32[$28+28>>2]=0|0; //@line 63 "libraries/Box2D/Dynamics/b2World.cpp"
  return; //@line 64 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $31 = ___cxa_find_matching_catch_2()|0;
 $32 = tempRet0;
 __ZN16b2StackAllocatorD2Ev($0); //@line 64 "libraries/Box2D/Dynamics/b2World.cpp"
 $$0 = $32;$$01 = $31;
 __ZN16b2BlockAllocatorD2Ev($this); //@line 64 "libraries/Box2D/Dynamics/b2World.cpp"
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN7b2WorldD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $b$01 = 0, $f$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 102952|0); //@line 69 "libraries/Box2D/Dynamics/b2World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 69 "libraries/Box2D/Dynamics/b2World.cpp"
 $2 = ($1|0)==(0|0); //@line 70 "libraries/Box2D/Dynamics/b2World.cpp"
 L1: do {
  if (!($2)) {
   $b$01 = $1;
   L2: while(1) {
    $5 = ((($b$01)) + 96|0); //@line 72 "libraries/Box2D/Dynamics/b2World.cpp"
    $4 = HEAP32[$5>>2]|0; //@line 72 "libraries/Box2D/Dynamics/b2World.cpp"
    $6 = ((($b$01)) + 100|0); //@line 74 "libraries/Box2D/Dynamics/b2World.cpp"
    $7 = HEAP32[$6>>2]|0; //@line 74 "libraries/Box2D/Dynamics/b2World.cpp"
    $f$0 = $7;
    while(1) {
     $8 = ($f$0|0)==(0|0); //@line 75 "libraries/Box2D/Dynamics/b2World.cpp"
     if ($8) {
      break;
     }
     $9 = ((($f$0)) + 4|0); //@line 77 "libraries/Box2D/Dynamics/b2World.cpp"
     $10 = HEAP32[$9>>2]|0; //@line 77 "libraries/Box2D/Dynamics/b2World.cpp"
     $11 = ((($f$0)) + 28|0); //@line 78 "libraries/Box2D/Dynamics/b2World.cpp"
     HEAP32[$11>>2] = 0; //@line 78 "libraries/Box2D/Dynamics/b2World.cpp"
     __THREW__ = 0;
     invoke_vii(3819,($f$0|0),($this|0)); //@line 79 "libraries/Box2D/Dynamics/b2World.cpp"
     $12 = __THREW__; __THREW__ = 0;
     $13 = $12&1;
     if ($13) {
      break L2;
     } else {
      $f$0 = $10;
     }
    }
    $3 = ($4|0)==(0|0); //@line 70 "libraries/Box2D/Dynamics/b2World.cpp"
    if ($3) {
     break L1;
    } else {
     $b$01 = $4;
    }
   }
   $14 = ___cxa_find_matching_catch_3(0|0)|0;
   $15 = tempRet0;
   $16 = ((($this)) + 102872|0); //@line 30 "./libraries/Box2D/Dynamics/b2ContactManager.h"
   __ZN12b2BroadPhaseD2Ev($16); //@line 30 "./libraries/Box2D/Dynamics/b2ContactManager.h"
   $17 = ((($this)) + 68|0); //@line 85 "libraries/Box2D/Dynamics/b2World.cpp"
   __ZN16b2StackAllocatorD2Ev($17); //@line 85 "libraries/Box2D/Dynamics/b2World.cpp"
   __ZN16b2BlockAllocatorD2Ev($this); //@line 85 "libraries/Box2D/Dynamics/b2World.cpp"
   ___clang_call_terminate($14); //@line 85 "libraries/Box2D/Dynamics/b2World.cpp"
   // unreachable; //@line 85 "libraries/Box2D/Dynamics/b2World.cpp"
  }
 } while(0);
 $18 = ((($this)) + 102872|0); //@line 30 "./libraries/Box2D/Dynamics/b2ContactManager.h"
 __ZN12b2BroadPhaseD2Ev($18); //@line 30 "./libraries/Box2D/Dynamics/b2ContactManager.h"
 $19 = ((($this)) + 68|0); //@line 85 "libraries/Box2D/Dynamics/b2World.cpp"
 __ZN16b2StackAllocatorD2Ev($19); //@line 85 "libraries/Box2D/Dynamics/b2World.cpp"
 __ZN16b2BlockAllocatorD2Ev($this); //@line 85 "libraries/Box2D/Dynamics/b2World.cpp"
 return; //@line 85 "libraries/Box2D/Dynamics/b2World.cpp"
}
function __ZN7b2World22SetDestructionListenerEP21b2DestructionListener($this,$listener) {
 $this = $this|0;
 $listener = $listener|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 102980|0); //@line 89 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$0>>2] = $listener; //@line 89 "libraries/Box2D/Dynamics/b2World.cpp"
 return; //@line 90 "libraries/Box2D/Dynamics/b2World.cpp"
}
function __ZN7b2World16SetContactFilterEP15b2ContactFilter($this,$filter) {
 $this = $this|0;
 $filter = $filter|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 102940|0); //@line 94 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$0>>2] = $filter; //@line 94 "libraries/Box2D/Dynamics/b2World.cpp"
 return; //@line 95 "libraries/Box2D/Dynamics/b2World.cpp"
}
function __ZN7b2World18SetContactListenerEP17b2ContactListener($this,$listener) {
 $this = $this|0;
 $listener = $listener|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 102944|0); //@line 99 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$0>>2] = $listener; //@line 99 "libraries/Box2D/Dynamics/b2World.cpp"
 return; //@line 100 "libraries/Box2D/Dynamics/b2World.cpp"
}
function __ZN7b2World10CreateBodyEPK9b2BodyDef($this,$def) {
 $this = $this|0;
 $def = $def|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $1 = HEAP32[$0>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $2 = $1 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $3 = ($2|0)==(0); //@line 109 "libraries/Box2D/Dynamics/b2World.cpp"
 __Z10loveAssertbPKc($3,361628); //@line 109 "libraries/Box2D/Dynamics/b2World.cpp"
 $4 = HEAP32[$0>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $5 = $4 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $6 = ($5|0)==(0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 if (!($6)) {
  $$0 = 0;
  return ($$0|0); //@line 129 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $7 = (__ZN16b2BlockAllocator8AllocateEi($this,152)|0); //@line 115 "libraries/Box2D/Dynamics/b2World.cpp"
 __ZN6b2BodyC2EPK9b2BodyDefP7b2World($7,$def,$this); //@line 116 "libraries/Box2D/Dynamics/b2World.cpp"
 $8 = ((($7)) + 92|0); //@line 119 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$8>>2] = 0; //@line 119 "libraries/Box2D/Dynamics/b2World.cpp"
 $9 = ((($this)) + 102952|0); //@line 120 "libraries/Box2D/Dynamics/b2World.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 120 "libraries/Box2D/Dynamics/b2World.cpp"
 $11 = ((($7)) + 96|0); //@line 120 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$11>>2] = $10; //@line 120 "libraries/Box2D/Dynamics/b2World.cpp"
 $12 = ($10|0)==(0); //@line 121 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($12)) {
  $13 = $10; //@line 121 "libraries/Box2D/Dynamics/b2World.cpp"
  $14 = ((($13)) + 92|0); //@line 123 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$14>>2] = $7; //@line 123 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 HEAP32[$9>>2] = $7; //@line 125 "libraries/Box2D/Dynamics/b2World.cpp"
 $15 = ((($this)) + 102960|0); //@line 126 "libraries/Box2D/Dynamics/b2World.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 126 "libraries/Box2D/Dynamics/b2World.cpp"
 $17 = (($16) + 1)|0; //@line 126 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$15>>2] = $17; //@line 126 "libraries/Box2D/Dynamics/b2World.cpp"
 $$0 = $7;
 return ($$0|0); //@line 129 "libraries/Box2D/Dynamics/b2World.cpp"
}
function __ZN7b2World11DestroyBodyEP6b2Body($this,$b) {
 $this = $this|0;
 $b = $b|0;
 var $$pre = 0, $$pre$phi14Z2D = 0, $$pre$phiZ2D = 0, $$pre11 = 0, $$pre13 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $ce$02 = 0, $f$01 = 0, $je$06 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 102960|0); //@line 133 "libraries/Box2D/Dynamics/b2World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 133 "libraries/Box2D/Dynamics/b2World.cpp"
 $2 = ($1|0)>(0); //@line 133 "libraries/Box2D/Dynamics/b2World.cpp"
 __Z10loveAssertbPKc($2,361648); //@line 133 "libraries/Box2D/Dynamics/b2World.cpp"
 $3 = ((($this)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $4 = HEAP32[$3>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $5 = $4 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $6 = ($5|0)==(0); //@line 134 "libraries/Box2D/Dynamics/b2World.cpp"
 __Z10loveAssertbPKc($6,361628); //@line 134 "libraries/Box2D/Dynamics/b2World.cpp"
 $7 = HEAP32[$3>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $8 = $7 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $9 = ($8|0)==(0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 if (!($9)) {
  return; //@line 210 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $10 = ((($b)) + 108|0); //@line 141 "libraries/Box2D/Dynamics/b2World.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 141 "libraries/Box2D/Dynamics/b2World.cpp"
 $12 = ($11|0)==(0|0); //@line 142 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($12)) {
  $13 = ((($this)) + 102980|0); //@line 147 "libraries/Box2D/Dynamics/b2World.cpp"
  $je$06 = $11;
  while(1) {
   $14 = ((($je$06)) + 12|0); //@line 145 "libraries/Box2D/Dynamics/b2World.cpp"
   $15 = HEAP32[$14>>2]|0; //@line 145 "libraries/Box2D/Dynamics/b2World.cpp"
   $16 = HEAP32[$13>>2]|0; //@line 147 "libraries/Box2D/Dynamics/b2World.cpp"
   $17 = ($16|0)==(0|0); //@line 147 "libraries/Box2D/Dynamics/b2World.cpp"
   if ($17) {
    $$pre13 = ((($je$06)) + 4|0); //@line 152 "libraries/Box2D/Dynamics/b2World.cpp"
    $$pre$phi14Z2D = $$pre13;
   } else {
    $18 = HEAP32[$16>>2]|0; //@line 149 "libraries/Box2D/Dynamics/b2World.cpp"
    $19 = ((($18)) + 8|0); //@line 149 "libraries/Box2D/Dynamics/b2World.cpp"
    $20 = HEAP32[$19>>2]|0; //@line 149 "libraries/Box2D/Dynamics/b2World.cpp"
    $21 = ((($je$06)) + 4|0); //@line 149 "libraries/Box2D/Dynamics/b2World.cpp"
    $22 = HEAP32[$21>>2]|0; //@line 149 "libraries/Box2D/Dynamics/b2World.cpp"
    FUNCTION_TABLE_vii[$20 & 8191]($16,$22); //@line 149 "libraries/Box2D/Dynamics/b2World.cpp"
    $$pre$phi14Z2D = $21;
   }
   $23 = HEAP32[$$pre$phi14Z2D>>2]|0; //@line 152 "libraries/Box2D/Dynamics/b2World.cpp"
   __ZN7b2World12DestroyJointEP7b2Joint($this,$23); //@line 152 "libraries/Box2D/Dynamics/b2World.cpp"
   HEAP32[$10>>2] = $15; //@line 154 "libraries/Box2D/Dynamics/b2World.cpp"
   $24 = ($15|0)==(0|0); //@line 142 "libraries/Box2D/Dynamics/b2World.cpp"
   if ($24) {
    break;
   } else {
    $je$06 = $15;
   }
  }
 }
 HEAP32[$10>>2] = 0; //@line 156 "libraries/Box2D/Dynamics/b2World.cpp"
 $25 = ((($b)) + 112|0); //@line 159 "libraries/Box2D/Dynamics/b2World.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 159 "libraries/Box2D/Dynamics/b2World.cpp"
 $27 = ($26|0)==(0|0); //@line 160 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($27)) {
  $28 = ((($this)) + 102872|0); //@line 164 "libraries/Box2D/Dynamics/b2World.cpp"
  $ce$02 = $26;
  while(1) {
   $29 = ((($ce$02)) + 12|0); //@line 163 "libraries/Box2D/Dynamics/b2World.cpp"
   $30 = HEAP32[$29>>2]|0; //@line 163 "libraries/Box2D/Dynamics/b2World.cpp"
   $31 = ((($ce$02)) + 4|0); //@line 164 "libraries/Box2D/Dynamics/b2World.cpp"
   $32 = HEAP32[$31>>2]|0; //@line 164 "libraries/Box2D/Dynamics/b2World.cpp"
   __ZN16b2ContactManager7DestroyEP9b2Contact($28,$32); //@line 164 "libraries/Box2D/Dynamics/b2World.cpp"
   $33 = ($30|0)==(0|0); //@line 160 "libraries/Box2D/Dynamics/b2World.cpp"
   if ($33) {
    break;
   } else {
    $ce$02 = $30;
   }
  }
 }
 HEAP32[$25>>2] = 0; //@line 166 "libraries/Box2D/Dynamics/b2World.cpp"
 $34 = ((($b)) + 100|0); //@line 169 "libraries/Box2D/Dynamics/b2World.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 169 "libraries/Box2D/Dynamics/b2World.cpp"
 $36 = ($35|0)==(0|0); //@line 170 "libraries/Box2D/Dynamics/b2World.cpp"
 if ($36) {
  $$pre = ((($b)) + 104|0); //@line 189 "libraries/Box2D/Dynamics/b2World.cpp"
  $$pre$phiZ2D = $$pre;
 } else {
  $37 = ((($this)) + 102980|0); //@line 175 "libraries/Box2D/Dynamics/b2World.cpp"
  $38 = ((($this)) + 102872|0); //@line 180 "libraries/Box2D/Dynamics/b2World.cpp"
  $39 = ((($b)) + 104|0); //@line 186 "libraries/Box2D/Dynamics/b2World.cpp"
  $f$01 = $35;
  while(1) {
   $40 = ((($f$01)) + 4|0); //@line 173 "libraries/Box2D/Dynamics/b2World.cpp"
   $41 = HEAP32[$40>>2]|0; //@line 173 "libraries/Box2D/Dynamics/b2World.cpp"
   $42 = HEAP32[$37>>2]|0; //@line 175 "libraries/Box2D/Dynamics/b2World.cpp"
   $43 = ($42|0)==(0|0); //@line 175 "libraries/Box2D/Dynamics/b2World.cpp"
   if (!($43)) {
    $44 = HEAP32[$42>>2]|0; //@line 177 "libraries/Box2D/Dynamics/b2World.cpp"
    $45 = ((($44)) + 12|0); //@line 177 "libraries/Box2D/Dynamics/b2World.cpp"
    $46 = HEAP32[$45>>2]|0; //@line 177 "libraries/Box2D/Dynamics/b2World.cpp"
    FUNCTION_TABLE_vii[$46 & 8191]($42,$f$01); //@line 177 "libraries/Box2D/Dynamics/b2World.cpp"
   }
   __ZN9b2Fixture14DestroyProxiesEP12b2BroadPhase($f$01,$38); //@line 180 "libraries/Box2D/Dynamics/b2World.cpp"
   __ZN9b2Fixture7DestroyEP16b2BlockAllocator($f$01,$this); //@line 181 "libraries/Box2D/Dynamics/b2World.cpp"
   __ZN16b2BlockAllocator4FreeEPvi($this,$f$01,44); //@line 183 "libraries/Box2D/Dynamics/b2World.cpp"
   HEAP32[$34>>2] = $41; //@line 185 "libraries/Box2D/Dynamics/b2World.cpp"
   $47 = HEAP32[$39>>2]|0; //@line 186 "libraries/Box2D/Dynamics/b2World.cpp"
   $48 = (($47) + -1)|0; //@line 186 "libraries/Box2D/Dynamics/b2World.cpp"
   HEAP32[$39>>2] = $48; //@line 186 "libraries/Box2D/Dynamics/b2World.cpp"
   $49 = ($41|0)==(0|0); //@line 170 "libraries/Box2D/Dynamics/b2World.cpp"
   if ($49) {
    $$pre$phiZ2D = $39;
    break;
   } else {
    $f$01 = $41;
   }
  }
 }
 HEAP32[$34>>2] = 0; //@line 188 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$$pre$phiZ2D>>2] = 0; //@line 189 "libraries/Box2D/Dynamics/b2World.cpp"
 $50 = ((($b)) + 92|0); //@line 192 "libraries/Box2D/Dynamics/b2World.cpp"
 $51 = HEAP32[$50>>2]|0; //@line 192 "libraries/Box2D/Dynamics/b2World.cpp"
 $52 = ($51|0)==(0|0); //@line 192 "libraries/Box2D/Dynamics/b2World.cpp"
 $53 = $51; //@line 199 "libraries/Box2D/Dynamics/b2World.cpp"
 $$pre11 = ((($b)) + 96|0); //@line 197 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($52)) {
  $54 = HEAP32[$$pre11>>2]|0; //@line 194 "libraries/Box2D/Dynamics/b2World.cpp"
  $55 = ((($51)) + 96|0); //@line 194 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$55>>2] = $54; //@line 194 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $56 = HEAP32[$$pre11>>2]|0; //@line 197 "libraries/Box2D/Dynamics/b2World.cpp"
 $57 = ($56|0)==(0|0); //@line 197 "libraries/Box2D/Dynamics/b2World.cpp"
 $58 = $56; //@line 204 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($57)) {
  $59 = ((($56)) + 92|0); //@line 199 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$59>>2] = $53; //@line 199 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $60 = ((($this)) + 102952|0); //@line 202 "libraries/Box2D/Dynamics/b2World.cpp"
 $61 = HEAP32[$60>>2]|0; //@line 202 "libraries/Box2D/Dynamics/b2World.cpp"
 $62 = ($61|0)==($b|0); //@line 202 "libraries/Box2D/Dynamics/b2World.cpp"
 if ($62) {
  HEAP32[$60>>2] = $58; //@line 204 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $63 = HEAP32[$0>>2]|0; //@line 207 "libraries/Box2D/Dynamics/b2World.cpp"
 $64 = (($63) + -1)|0; //@line 207 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$0>>2] = $64; //@line 207 "libraries/Box2D/Dynamics/b2World.cpp"
 __ZN6b2BodyD2Ev($b); //@line 208 "libraries/Box2D/Dynamics/b2World.cpp"
 __ZN16b2BlockAllocator4FreeEPvi($this,$b,152); //@line 209 "libraries/Box2D/Dynamics/b2World.cpp"
 return; //@line 210 "libraries/Box2D/Dynamics/b2World.cpp"
}
function __ZN7b2World12DestroyJointEP7b2Joint($this,$j) {
 $this = $this|0;
 $j = $j|0;
 var $$pre = 0, $$pre5 = 0, $$pre7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $edge$0 = 0, $edge$02 = 0, $edge$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $1 = HEAP32[$0>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $2 = $1 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $3 = ($2|0)==(0); //@line 274 "libraries/Box2D/Dynamics/b2World.cpp"
 __Z10loveAssertbPKc($3,361628); //@line 274 "libraries/Box2D/Dynamics/b2World.cpp"
 $4 = HEAP32[$0>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $5 = $4 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $6 = ($5|0)==(0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 if (!($6)) {
  return; //@line 365 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $7 = ((($j)) + 61|0); //@line 280 "libraries/Box2D/Dynamics/b2World.cpp"
 $8 = HEAP8[$7>>0]|0; //@line 280 "libraries/Box2D/Dynamics/b2World.cpp"
 $9 = ((($j)) + 8|0); //@line 283 "libraries/Box2D/Dynamics/b2World.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 283 "libraries/Box2D/Dynamics/b2World.cpp"
 $11 = ($10|0)==(0|0); //@line 283 "libraries/Box2D/Dynamics/b2World.cpp"
 $12 = $10; //@line 290 "libraries/Box2D/Dynamics/b2World.cpp"
 $$pre = ((($j)) + 12|0); //@line 288 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($11)) {
  $13 = HEAP32[$$pre>>2]|0; //@line 285 "libraries/Box2D/Dynamics/b2World.cpp"
  $14 = ((($10)) + 12|0); //@line 285 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$14>>2] = $13; //@line 285 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $15 = HEAP32[$$pre>>2]|0; //@line 288 "libraries/Box2D/Dynamics/b2World.cpp"
 $16 = ($15|0)==(0|0); //@line 288 "libraries/Box2D/Dynamics/b2World.cpp"
 $17 = $15; //@line 295 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($16)) {
  $18 = ((($15)) + 8|0); //@line 290 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$18>>2] = $12; //@line 290 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $19 = ((($this)) + 102956|0); //@line 293 "libraries/Box2D/Dynamics/b2World.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 293 "libraries/Box2D/Dynamics/b2World.cpp"
 $21 = ($20|0)==($j|0); //@line 293 "libraries/Box2D/Dynamics/b2World.cpp"
 if ($21) {
  HEAP32[$19>>2] = $17; //@line 295 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $22 = ((($j)) + 48|0); //@line 299 "libraries/Box2D/Dynamics/b2World.cpp"
 $23 = HEAP32[$22>>2]|0; //@line 299 "libraries/Box2D/Dynamics/b2World.cpp"
 $24 = ((($j)) + 52|0); //@line 300 "libraries/Box2D/Dynamics/b2World.cpp"
 $25 = HEAP32[$24>>2]|0; //@line 300 "libraries/Box2D/Dynamics/b2World.cpp"
 $26 = ((($23)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $27 = HEAP16[$26>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $28 = $27&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $29 = $28 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $30 = ($29|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($30) {
  $31 = $28 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $32 = $31&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$26>>1] = $32; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $33 = ((($23)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$33>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $34 = ((($25)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $35 = HEAP16[$34>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $36 = $35&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $37 = $36 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $38 = ($37|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($38) {
  $39 = $36 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $40 = $39&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$34>>1] = $40; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $41 = ((($25)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$41>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $42 = ((($j)) + 16|0); //@line 307 "libraries/Box2D/Dynamics/b2World.cpp"
 $43 = ((($j)) + 24|0); //@line 307 "libraries/Box2D/Dynamics/b2World.cpp"
 $44 = HEAP32[$43>>2]|0; //@line 307 "libraries/Box2D/Dynamics/b2World.cpp"
 $45 = ($44|0)==(0|0); //@line 307 "libraries/Box2D/Dynamics/b2World.cpp"
 $46 = $44; //@line 314 "libraries/Box2D/Dynamics/b2World.cpp"
 $$pre5 = ((($j)) + 28|0); //@line 312 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($45)) {
  $47 = HEAP32[$$pre5>>2]|0; //@line 309 "libraries/Box2D/Dynamics/b2World.cpp"
  $48 = ((($44)) + 12|0); //@line 309 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$48>>2] = $47; //@line 309 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $49 = HEAP32[$$pre5>>2]|0; //@line 312 "libraries/Box2D/Dynamics/b2World.cpp"
 $50 = ($49|0)==(0|0); //@line 312 "libraries/Box2D/Dynamics/b2World.cpp"
 $51 = $49; //@line 319 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($50)) {
  $52 = ((($49)) + 8|0); //@line 314 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$52>>2] = $46; //@line 314 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $53 = ((($23)) + 108|0); //@line 317 "libraries/Box2D/Dynamics/b2World.cpp"
 $54 = HEAP32[$53>>2]|0; //@line 317 "libraries/Box2D/Dynamics/b2World.cpp"
 $55 = ($42|0)==($54|0); //@line 317 "libraries/Box2D/Dynamics/b2World.cpp"
 if ($55) {
  HEAP32[$53>>2] = $51; //@line 319 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 HEAP32[$43>>2] = 0; //@line 322 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$$pre5>>2] = 0; //@line 323 "libraries/Box2D/Dynamics/b2World.cpp"
 $56 = ((($j)) + 32|0); //@line 326 "libraries/Box2D/Dynamics/b2World.cpp"
 $57 = ((($j)) + 40|0); //@line 326 "libraries/Box2D/Dynamics/b2World.cpp"
 $58 = HEAP32[$57>>2]|0; //@line 326 "libraries/Box2D/Dynamics/b2World.cpp"
 $59 = ($58|0)==(0|0); //@line 326 "libraries/Box2D/Dynamics/b2World.cpp"
 $60 = $58; //@line 333 "libraries/Box2D/Dynamics/b2World.cpp"
 $$pre7 = ((($j)) + 44|0); //@line 331 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($59)) {
  $61 = HEAP32[$$pre7>>2]|0; //@line 328 "libraries/Box2D/Dynamics/b2World.cpp"
  $62 = ((($58)) + 12|0); //@line 328 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$62>>2] = $61; //@line 328 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $63 = HEAP32[$$pre7>>2]|0; //@line 331 "libraries/Box2D/Dynamics/b2World.cpp"
 $64 = ($63|0)==(0|0); //@line 331 "libraries/Box2D/Dynamics/b2World.cpp"
 $65 = $63; //@line 338 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($64)) {
  $66 = ((($63)) + 8|0); //@line 333 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$66>>2] = $60; //@line 333 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $67 = ((($25)) + 108|0); //@line 336 "libraries/Box2D/Dynamics/b2World.cpp"
 $68 = HEAP32[$67>>2]|0; //@line 336 "libraries/Box2D/Dynamics/b2World.cpp"
 $69 = ($56|0)==($68|0); //@line 336 "libraries/Box2D/Dynamics/b2World.cpp"
 if ($69) {
  HEAP32[$67>>2] = $65; //@line 338 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 HEAP32[$57>>2] = 0; //@line 341 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$$pre7>>2] = 0; //@line 342 "libraries/Box2D/Dynamics/b2World.cpp"
 __ZN7b2Joint7DestroyEPS_P16b2BlockAllocator($j,$this); //@line 344 "libraries/Box2D/Dynamics/b2World.cpp"
 $70 = ((($this)) + 102964|0); //@line 346 "libraries/Box2D/Dynamics/b2World.cpp"
 $71 = HEAP32[$70>>2]|0; //@line 346 "libraries/Box2D/Dynamics/b2World.cpp"
 $72 = ($71|0)>(0); //@line 346 "libraries/Box2D/Dynamics/b2World.cpp"
 __Z10loveAssertbPKc($72,361664); //@line 346 "libraries/Box2D/Dynamics/b2World.cpp"
 $73 = HEAP32[$70>>2]|0; //@line 347 "libraries/Box2D/Dynamics/b2World.cpp"
 $74 = (($73) + -1)|0; //@line 347 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$70>>2] = $74; //@line 347 "libraries/Box2D/Dynamics/b2World.cpp"
 $75 = ($8<<24>>24)==(0); //@line 350 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($75)) {
  return; //@line 365 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $76 = ((($25)) + 112|0); //@line 709 "./libraries/Box2D/Dynamics/b2Body.h"
 $edge$02 = HEAP32[$76>>2]|0; //@line 709 "./libraries/Box2D/Dynamics/b2Body.h"
 $77 = ($edge$02|0)==(0|0); //@line 353 "libraries/Box2D/Dynamics/b2World.cpp"
 if ($77) {
  return; //@line 365 "libraries/Box2D/Dynamics/b2World.cpp"
 } else {
  $edge$03 = $edge$02;
 }
 while(1) {
  $78 = HEAP32[$edge$03>>2]|0; //@line 355 "libraries/Box2D/Dynamics/b2World.cpp"
  $79 = ($78|0)==($23|0); //@line 355 "libraries/Box2D/Dynamics/b2World.cpp"
  if ($79) {
   $80 = ((($edge$03)) + 4|0); //@line 359 "libraries/Box2D/Dynamics/b2World.cpp"
   $81 = HEAP32[$80>>2]|0; //@line 359 "libraries/Box2D/Dynamics/b2World.cpp"
   $82 = ((($81)) + 4|0); //@line 306 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
   $83 = HEAP32[$82>>2]|0; //@line 306 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
   $84 = $83 | 8; //@line 306 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
   HEAP32[$82>>2] = $84; //@line 306 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
  }
  $85 = ((($edge$03)) + 12|0); //@line 362 "libraries/Box2D/Dynamics/b2World.cpp"
  $edge$0 = HEAP32[$85>>2]|0; //@line 709 "./libraries/Box2D/Dynamics/b2Body.h"
  $86 = ($edge$0|0)==(0|0); //@line 353 "libraries/Box2D/Dynamics/b2World.cpp"
  if ($86) {
   break;
  } else {
   $edge$03 = $edge$0;
  }
 }
 return; //@line 365 "libraries/Box2D/Dynamics/b2World.cpp"
}
function __ZN7b2World11CreateJointEPK10b2JointDef($this,$def) {
 $this = $this|0;
 $def = $def|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $edge$0 = 0, $edge$01 = 0, $edge$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $1 = HEAP32[$0>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $2 = $1 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $3 = ($2|0)==(0); //@line 214 "libraries/Box2D/Dynamics/b2World.cpp"
 __Z10loveAssertbPKc($3,361628); //@line 214 "libraries/Box2D/Dynamics/b2World.cpp"
 $4 = HEAP32[$0>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $5 = $4 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $6 = ($5|0)==(0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 if (!($6)) {
  $$0 = 0;
  return ($$0|0); //@line 270 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $7 = (__ZN7b2Joint6CreateEPK10b2JointDefP16b2BlockAllocator($def,$this)|0); //@line 220 "libraries/Box2D/Dynamics/b2World.cpp"
 $8 = ((($7)) + 8|0); //@line 223 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$8>>2] = 0; //@line 223 "libraries/Box2D/Dynamics/b2World.cpp"
 $9 = ((($this)) + 102956|0); //@line 224 "libraries/Box2D/Dynamics/b2World.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 224 "libraries/Box2D/Dynamics/b2World.cpp"
 $11 = ((($7)) + 12|0); //@line 224 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$11>>2] = $10; //@line 224 "libraries/Box2D/Dynamics/b2World.cpp"
 $12 = ($10|0)==(0); //@line 225 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($12)) {
  $13 = $10; //@line 225 "libraries/Box2D/Dynamics/b2World.cpp"
  $14 = ((($13)) + 8|0); //@line 227 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$14>>2] = $7; //@line 227 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 HEAP32[$9>>2] = $7; //@line 229 "libraries/Box2D/Dynamics/b2World.cpp"
 $15 = ((($this)) + 102964|0); //@line 230 "libraries/Box2D/Dynamics/b2World.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 230 "libraries/Box2D/Dynamics/b2World.cpp"
 $17 = (($16) + 1)|0; //@line 230 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$15>>2] = $17; //@line 230 "libraries/Box2D/Dynamics/b2World.cpp"
 $18 = ((($7)) + 16|0); //@line 233 "libraries/Box2D/Dynamics/b2World.cpp"
 $19 = ((($7)) + 20|0); //@line 233 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$19>>2] = $7; //@line 233 "libraries/Box2D/Dynamics/b2World.cpp"
 $20 = ((($7)) + 52|0); //@line 234 "libraries/Box2D/Dynamics/b2World.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 234 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$18>>2] = $21; //@line 234 "libraries/Box2D/Dynamics/b2World.cpp"
 $22 = ((($7)) + 24|0); //@line 235 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$22>>2] = 0; //@line 235 "libraries/Box2D/Dynamics/b2World.cpp"
 $23 = ((($7)) + 48|0); //@line 236 "libraries/Box2D/Dynamics/b2World.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 238 "libraries/Box2D/Dynamics/b2World.cpp"
 $25 = ((($24)) + 108|0); //@line 236 "libraries/Box2D/Dynamics/b2World.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 236 "libraries/Box2D/Dynamics/b2World.cpp"
 $27 = ((($7)) + 28|0); //@line 236 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$27>>2] = $26; //@line 236 "libraries/Box2D/Dynamics/b2World.cpp"
 $28 = ($26|0)==(0); //@line 237 "libraries/Box2D/Dynamics/b2World.cpp"
 $29 = $24; //@line 241 "libraries/Box2D/Dynamics/b2World.cpp"
 $30 = $21; //@line 243 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($28)) {
  $31 = $26; //@line 237 "libraries/Box2D/Dynamics/b2World.cpp"
  $32 = ((($31)) + 8|0); //@line 237 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$32>>2] = $18; //@line 237 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 HEAP32[$25>>2] = $18; //@line 238 "libraries/Box2D/Dynamics/b2World.cpp"
 $33 = ((($7)) + 32|0); //@line 240 "libraries/Box2D/Dynamics/b2World.cpp"
 $34 = ((($7)) + 36|0); //@line 240 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$34>>2] = $7; //@line 240 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$33>>2] = $29; //@line 241 "libraries/Box2D/Dynamics/b2World.cpp"
 $35 = ((($7)) + 40|0); //@line 242 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$35>>2] = 0; //@line 242 "libraries/Box2D/Dynamics/b2World.cpp"
 $36 = ((($30)) + 108|0); //@line 243 "libraries/Box2D/Dynamics/b2World.cpp"
 $37 = HEAP32[$36>>2]|0; //@line 243 "libraries/Box2D/Dynamics/b2World.cpp"
 $38 = ((($7)) + 44|0); //@line 243 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$38>>2] = $37; //@line 243 "libraries/Box2D/Dynamics/b2World.cpp"
 $39 = ($37|0)==(0); //@line 244 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($39)) {
  $40 = $37; //@line 244 "libraries/Box2D/Dynamics/b2World.cpp"
  $41 = ((($40)) + 8|0); //@line 244 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$41>>2] = $33; //@line 244 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 HEAP32[$36>>2] = $33; //@line 245 "libraries/Box2D/Dynamics/b2World.cpp"
 $42 = ((($def)) + 8|0); //@line 247 "libraries/Box2D/Dynamics/b2World.cpp"
 $43 = HEAP32[$42>>2]|0; //@line 247 "libraries/Box2D/Dynamics/b2World.cpp"
 $44 = ((($def)) + 16|0); //@line 251 "libraries/Box2D/Dynamics/b2World.cpp"
 $45 = HEAP8[$44>>0]|0; //@line 251 "libraries/Box2D/Dynamics/b2World.cpp"
 $46 = ($45<<24>>24)==(0); //@line 251 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($46)) {
  $$0 = $7;
  return ($$0|0); //@line 270 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $47 = ((($def)) + 12|0); //@line 248 "libraries/Box2D/Dynamics/b2World.cpp"
 $48 = HEAP32[$47>>2]|0; //@line 248 "libraries/Box2D/Dynamics/b2World.cpp"
 $49 = ((($48)) + 112|0); //@line 709 "./libraries/Box2D/Dynamics/b2Body.h"
 $edge$01 = HEAP32[$49>>2]|0; //@line 709 "./libraries/Box2D/Dynamics/b2Body.h"
 $50 = ($edge$01|0)==(0|0); //@line 254 "libraries/Box2D/Dynamics/b2World.cpp"
 if ($50) {
  $$0 = $7;
  return ($$0|0); //@line 270 "libraries/Box2D/Dynamics/b2World.cpp"
 } else {
  $edge$02 = $edge$01;
 }
 while(1) {
  $51 = HEAP32[$edge$02>>2]|0; //@line 256 "libraries/Box2D/Dynamics/b2World.cpp"
  $52 = ($51|0)==($43|0); //@line 256 "libraries/Box2D/Dynamics/b2World.cpp"
  if ($52) {
   $53 = ((($edge$02)) + 4|0); //@line 260 "libraries/Box2D/Dynamics/b2World.cpp"
   $54 = HEAP32[$53>>2]|0; //@line 260 "libraries/Box2D/Dynamics/b2World.cpp"
   $55 = ((($54)) + 4|0); //@line 306 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
   $56 = HEAP32[$55>>2]|0; //@line 306 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
   $57 = $56 | 8; //@line 306 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
   HEAP32[$55>>2] = $57; //@line 306 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
  }
  $58 = ((($edge$02)) + 12|0); //@line 263 "libraries/Box2D/Dynamics/b2World.cpp"
  $edge$0 = HEAP32[$58>>2]|0; //@line 709 "./libraries/Box2D/Dynamics/b2Body.h"
  $59 = ($edge$0|0)==(0|0); //@line 254 "libraries/Box2D/Dynamics/b2World.cpp"
  if ($59) {
   $$0 = $7;
   break;
  } else {
   $edge$02 = $edge$0;
  }
 }
 return ($$0|0); //@line 270 "libraries/Box2D/Dynamics/b2World.cpp"
}
function __ZN7b2World16SetAllowSleepingEb($this,$flag) {
 $this = $this|0;
 $flag = $flag|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b$0 = 0, $b$01 = 0;
 var $b$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $flag&1; //@line 370 "libraries/Box2D/Dynamics/b2World.cpp"
 $1 = ((($this)) + 102976|0); //@line 370 "libraries/Box2D/Dynamics/b2World.cpp"
 $2 = HEAP8[$1>>0]|0; //@line 370 "libraries/Box2D/Dynamics/b2World.cpp"
 $3 = $2&255; //@line 370 "libraries/Box2D/Dynamics/b2World.cpp"
 $4 = ($0|0)==($3|0); //@line 370 "libraries/Box2D/Dynamics/b2World.cpp"
 if ($4) {
  return; //@line 383 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $5 = $flag&1; //@line 375 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP8[$1>>0] = $5; //@line 375 "libraries/Box2D/Dynamics/b2World.cpp"
 if ($flag) {
  return; //@line 383 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $6 = ((($this)) + 102952|0); //@line 378 "libraries/Box2D/Dynamics/b2World.cpp"
 $b$01 = HEAP32[$6>>2]|0; //@line 378 "libraries/Box2D/Dynamics/b2World.cpp"
 $7 = ($b$01|0)==(0|0); //@line 378 "libraries/Box2D/Dynamics/b2World.cpp"
 if ($7) {
  return; //@line 383 "libraries/Box2D/Dynamics/b2World.cpp"
 } else {
  $b$02 = $b$01;
 }
 while(1) {
  $8 = ((($b$02)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $9 = HEAP16[$8>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $10 = $9&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $11 = $10 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $12 = ($11|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  if ($12) {
   $13 = $10 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $14 = $13&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAP16[$8>>1] = $14; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $15 = ((($b$02)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAPF32[$15>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  }
  $16 = ((($b$02)) + 96|0); //@line 378 "libraries/Box2D/Dynamics/b2World.cpp"
  $b$0 = HEAP32[$16>>2]|0; //@line 378 "libraries/Box2D/Dynamics/b2World.cpp"
  $17 = ($b$0|0)==(0|0); //@line 378 "libraries/Box2D/Dynamics/b2World.cpp"
  if ($17) {
   break;
  } else {
   $b$02 = $b$0;
  }
 }
 return; //@line 383 "libraries/Box2D/Dynamics/b2World.cpp"
}
function __ZN7b2World5SolveERK10b2TimeStep($this,$step) {
 $this = $this|0;
 $step = $step|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0.0, $193 = 0.0, $194 = 0.0, $195 = 0.0, $196 = 0.0, $197 = 0.0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0.0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0.0, $233 = 0, $234 = 0, $235 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $b$0 = 0, $b$057 = 0, $b$058 = 0, $b6$0 = 0;
 var $b6$023 = 0, $b6$024 = 0, $c$0 = 0, $c$052 = 0, $c$053 = 0, $ce$0 = 0, $ce$025 = 0, $ce$027 = 0, $i$038 = 0, $island = 0, $j$0 = 0, $j$047 = 0, $j$048 = 0, $je$0 = 0, $je$031 = 0, $je$033 = 0, $lpad$phi$index = 0, $lpad$phi$index6 = 0, $profile = 0, $seed$0 = 0;
 var $seed$042 = 0, $seed$043 = 0, $stackCount$0$be = 0, $stackCount$037 = 0, $stackCount$1$lcssa = 0, $stackCount$126 = 0, $stackCount$4 = 0, $stackCount$532 = 0, $stackCount$7 = 0, $timer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $island = sp + 32|0;
 $profile = sp;
 $timer = sp + 84|0;
 $0 = ((($this)) + 103008|0); //@line 388 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAPF32[$0>>2] = 0.0; //@line 388 "libraries/Box2D/Dynamics/b2World.cpp"
 $1 = ((($this)) + 103012|0); //@line 389 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAPF32[$1>>2] = 0.0; //@line 389 "libraries/Box2D/Dynamics/b2World.cpp"
 $2 = ((($this)) + 103016|0); //@line 390 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAPF32[$2>>2] = 0.0; //@line 390 "libraries/Box2D/Dynamics/b2World.cpp"
 $3 = ((($this)) + 102960|0); //@line 393 "libraries/Box2D/Dynamics/b2World.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 393 "libraries/Box2D/Dynamics/b2World.cpp"
 $5 = ((($this)) + 102872|0); //@line 394 "libraries/Box2D/Dynamics/b2World.cpp"
 $6 = ((($this)) + 102936|0); //@line 394 "libraries/Box2D/Dynamics/b2World.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 394 "libraries/Box2D/Dynamics/b2World.cpp"
 $8 = ((($this)) + 102964|0); //@line 395 "libraries/Box2D/Dynamics/b2World.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 395 "libraries/Box2D/Dynamics/b2World.cpp"
 $10 = ((($this)) + 68|0); //@line 396 "libraries/Box2D/Dynamics/b2World.cpp"
 $11 = ((($this)) + 102944|0); //@line 397 "libraries/Box2D/Dynamics/b2World.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 397 "libraries/Box2D/Dynamics/b2World.cpp"
 __ZN8b2IslandC2EiiiP16b2StackAllocatorP17b2ContactListener($island,$4,$7,$9,$10,$12); //@line 393 "libraries/Box2D/Dynamics/b2World.cpp"
 $13 = ((($this)) + 102952|0); //@line 400 "libraries/Box2D/Dynamics/b2World.cpp"
 $b$057 = HEAP32[$13>>2]|0; //@line 400 "libraries/Box2D/Dynamics/b2World.cpp"
 $14 = ($b$057|0)==(0|0); //@line 400 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($14)) {
  $b$058 = $b$057;
  while(1) {
   $17 = ((($b$058)) + 4|0); //@line 402 "libraries/Box2D/Dynamics/b2World.cpp"
   $18 = HEAP16[$17>>1]|0; //@line 402 "libraries/Box2D/Dynamics/b2World.cpp"
   $19 = $18&65535; //@line 402 "libraries/Box2D/Dynamics/b2World.cpp"
   $20 = $19 & 65534; //@line 402 "libraries/Box2D/Dynamics/b2World.cpp"
   $21 = $20&65535; //@line 402 "libraries/Box2D/Dynamics/b2World.cpp"
   HEAP16[$17>>1] = $21; //@line 402 "libraries/Box2D/Dynamics/b2World.cpp"
   $22 = ((($b$058)) + 96|0); //@line 400 "libraries/Box2D/Dynamics/b2World.cpp"
   $b$0 = HEAP32[$22>>2]|0; //@line 400 "libraries/Box2D/Dynamics/b2World.cpp"
   $23 = ($b$0|0)==(0|0); //@line 400 "libraries/Box2D/Dynamics/b2World.cpp"
   if ($23) {
    break;
   } else {
    $b$058 = $b$0;
   }
  }
 }
 $15 = ((($this)) + 102932|0); //@line 404 "libraries/Box2D/Dynamics/b2World.cpp"
 $c$052 = HEAP32[$15>>2]|0; //@line 404 "libraries/Box2D/Dynamics/b2World.cpp"
 $16 = ($c$052|0)==(0|0); //@line 404 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($16)) {
  $c$053 = $c$052;
  while(1) {
   $26 = ((($c$053)) + 4|0); //@line 406 "libraries/Box2D/Dynamics/b2World.cpp"
   $27 = HEAP32[$26>>2]|0; //@line 406 "libraries/Box2D/Dynamics/b2World.cpp"
   $28 = $27 & -2; //@line 406 "libraries/Box2D/Dynamics/b2World.cpp"
   HEAP32[$26>>2] = $28; //@line 406 "libraries/Box2D/Dynamics/b2World.cpp"
   $29 = ((($c$053)) + 12|0); //@line 404 "libraries/Box2D/Dynamics/b2World.cpp"
   $c$0 = HEAP32[$29>>2]|0; //@line 404 "libraries/Box2D/Dynamics/b2World.cpp"
   $30 = ($c$0|0)==(0|0); //@line 404 "libraries/Box2D/Dynamics/b2World.cpp"
   if ($30) {
    break;
   } else {
    $c$053 = $c$0;
   }
  }
 }
 $24 = ((($this)) + 102956|0); //@line 408 "libraries/Box2D/Dynamics/b2World.cpp"
 $j$047 = HEAP32[$24>>2]|0; //@line 408 "libraries/Box2D/Dynamics/b2World.cpp"
 $25 = ($j$047|0)==(0|0); //@line 408 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($25)) {
  $j$048 = $j$047;
  while(1) {
   $36 = ((($j$048)) + 60|0); //@line 410 "libraries/Box2D/Dynamics/b2World.cpp"
   HEAP8[$36>>0] = 0; //@line 410 "libraries/Box2D/Dynamics/b2World.cpp"
   $37 = ((($j$048)) + 12|0); //@line 408 "libraries/Box2D/Dynamics/b2World.cpp"
   $j$0 = HEAP32[$37>>2]|0; //@line 408 "libraries/Box2D/Dynamics/b2World.cpp"
   $38 = ($j$0|0)==(0|0); //@line 408 "libraries/Box2D/Dynamics/b2World.cpp"
   if ($38) {
    break;
   } else {
    $j$048 = $j$0;
   }
  }
 }
 $31 = HEAP32[$3>>2]|0; //@line 414 "libraries/Box2D/Dynamics/b2World.cpp"
 $32 = $31 << 2; //@line 415 "libraries/Box2D/Dynamics/b2World.cpp"
 __THREW__ = 0;
 $33 = (invoke_iii(3820,($10|0),($32|0))|0); //@line 415 "libraries/Box2D/Dynamics/b2World.cpp"
 $34 = __THREW__; __THREW__ = 0;
 $35 = $34&1;
 if (!($35)) {
  $seed$042 = HEAP32[$13>>2]|0; //@line 416 "libraries/Box2D/Dynamics/b2World.cpp"
  $39 = ($seed$042|0)==(0|0); //@line 416 "libraries/Box2D/Dynamics/b2World.cpp"
  L15: do {
   if (!($39)) {
    $40 = ((($island)) + 28|0); //@line 43 "./libraries/Box2D/Dynamics/b2Island.h"
    $41 = ((($island)) + 36|0); //@line 44 "./libraries/Box2D/Dynamics/b2Island.h"
    $42 = ((($island)) + 32|0); //@line 45 "./libraries/Box2D/Dynamics/b2Island.h"
    $43 = ((($island)) + 40|0); //@line 54 "./libraries/Box2D/Dynamics/b2Island.h"
    $44 = ((($island)) + 8|0); //@line 56 "./libraries/Box2D/Dynamics/b2Island.h"
    $45 = ((($island)) + 48|0); //@line 68 "./libraries/Box2D/Dynamics/b2Island.h"
    $46 = ((($island)) + 16|0); //@line 69 "./libraries/Box2D/Dynamics/b2Island.h"
    $47 = ((($island)) + 44|0); //@line 62 "./libraries/Box2D/Dynamics/b2Island.h"
    $48 = ((($island)) + 12|0); //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
    $49 = ((($this)) + 102968|0); //@line 531 "libraries/Box2D/Dynamics/b2World.cpp"
    $50 = ((($this)) + 102976|0); //@line 531 "libraries/Box2D/Dynamics/b2World.cpp"
    $51 = ((($profile)) + 12|0); //@line 532 "libraries/Box2D/Dynamics/b2World.cpp"
    $52 = ((($profile)) + 16|0); //@line 533 "libraries/Box2D/Dynamics/b2World.cpp"
    $53 = ((($profile)) + 20|0); //@line 534 "libraries/Box2D/Dynamics/b2World.cpp"
    $seed$043 = $seed$042;
    L17: while(1) {
     $68 = ((($seed$043)) + 4|0); //@line 418 "libraries/Box2D/Dynamics/b2World.cpp"
     $69 = HEAP16[$68>>1]|0; //@line 438 "libraries/Box2D/Dynamics/b2World.cpp"
     $70 = $69 & 35; //@line 418 "libraries/Box2D/Dynamics/b2World.cpp"
     $71 = ($70<<16>>16)==(34); //@line 418 "libraries/Box2D/Dynamics/b2World.cpp"
     if ($71) {
      $72 = HEAP32[$seed$043>>2]|0; //@line 471 "./libraries/Box2D/Dynamics/b2Body.h"
      $73 = ($72|0)==(0); //@line 429 "libraries/Box2D/Dynamics/b2World.cpp"
      if (!($73)) {
       HEAP32[$40>>2] = 0; //@line 43 "./libraries/Box2D/Dynamics/b2Island.h"
       HEAP32[$41>>2] = 0; //@line 44 "./libraries/Box2D/Dynamics/b2Island.h"
       HEAP32[$42>>2] = 0; //@line 45 "./libraries/Box2D/Dynamics/b2Island.h"
       HEAP32[$33>>2] = $seed$043; //@line 437 "libraries/Box2D/Dynamics/b2World.cpp"
       $74 = $69&65535; //@line 438 "libraries/Box2D/Dynamics/b2World.cpp"
       $75 = $74 | 1; //@line 438 "libraries/Box2D/Dynamics/b2World.cpp"
       $76 = $75&65535; //@line 438 "libraries/Box2D/Dynamics/b2World.cpp"
       HEAP16[$68>>1] = $76; //@line 438 "libraries/Box2D/Dynamics/b2World.cpp"
       $stackCount$037 = 1;
       while(1) {
        $77 = (($stackCount$037) + -1)|0; //@line 444 "libraries/Box2D/Dynamics/b2World.cpp"
        $78 = (($33) + ($77<<2)|0); //@line 444 "libraries/Box2D/Dynamics/b2World.cpp"
        $79 = HEAP32[$78>>2]|0; //@line 444 "libraries/Box2D/Dynamics/b2World.cpp"
        $80 = ((($79)) + 4|0); //@line 661 "./libraries/Box2D/Dynamics/b2Body.h"
        $81 = HEAP16[$80>>1]|0; //@line 661 "./libraries/Box2D/Dynamics/b2Body.h"
        $82 = $81 & 32; //@line 661 "./libraries/Box2D/Dynamics/b2Body.h"
        $83 = ($82<<16>>16)!=(0); //@line 661 "./libraries/Box2D/Dynamics/b2Body.h"
        __THREW__ = 0;
        invoke_vii(3822,($83|0),(361681|0)); //@line 445 "libraries/Box2D/Dynamics/b2World.cpp"
        $84 = __THREW__; __THREW__ = 0;
        $85 = $84&1;
        if ($85) {
         label = 14;
         break L17;
        }
        $86 = HEAP32[$40>>2]|0; //@line 54 "./libraries/Box2D/Dynamics/b2Island.h"
        $87 = HEAP32[$43>>2]|0; //@line 54 "./libraries/Box2D/Dynamics/b2Island.h"
        $88 = ($86|0)<($87|0); //@line 54 "./libraries/Box2D/Dynamics/b2Island.h"
        __THREW__ = 0;
        invoke_vii(3822,($88|0),(361703|0)); //@line 54 "./libraries/Box2D/Dynamics/b2Island.h"
        $89 = __THREW__; __THREW__ = 0;
        $90 = $89&1;
        if ($90) {
         label = 14;
         break L17;
        }
        $91 = HEAP32[$40>>2]|0; //@line 55 "./libraries/Box2D/Dynamics/b2Island.h"
        $92 = ((($79)) + 8|0); //@line 55 "./libraries/Box2D/Dynamics/b2Island.h"
        HEAP32[$92>>2] = $91; //@line 55 "./libraries/Box2D/Dynamics/b2Island.h"
        $93 = HEAP32[$44>>2]|0; //@line 56 "./libraries/Box2D/Dynamics/b2Island.h"
        $94 = (($93) + ($91<<2)|0); //@line 56 "./libraries/Box2D/Dynamics/b2Island.h"
        HEAP32[$94>>2] = $79; //@line 56 "./libraries/Box2D/Dynamics/b2Island.h"
        $95 = (($91) + 1)|0; //@line 57 "./libraries/Box2D/Dynamics/b2Island.h"
        HEAP32[$40>>2] = $95; //@line 57 "./libraries/Box2D/Dynamics/b2Island.h"
        $96 = HEAP16[$80>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
        $97 = $96&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
        $98 = $97 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
        $99 = ($98|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
        if ($99) {
         $100 = $97 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
         $101 = $100&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
         HEAP16[$80>>1] = $101; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
         $102 = ((($79)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
         HEAPF32[$102>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
        }
        $103 = HEAP32[$79>>2]|0; //@line 471 "./libraries/Box2D/Dynamics/b2Body.h"
        $104 = ($103|0)==(0); //@line 453 "libraries/Box2D/Dynamics/b2World.cpp"
        if ($104) {
         $stackCount$0$be = $77;
        } else {
         $106 = ((($79)) + 112|0); //@line 459 "libraries/Box2D/Dynamics/b2World.cpp"
         $ce$025 = HEAP32[$106>>2]|0; //@line 459 "libraries/Box2D/Dynamics/b2World.cpp"
         $107 = ($ce$025|0)==(0|0); //@line 459 "libraries/Box2D/Dynamics/b2World.cpp"
         if ($107) {
          $stackCount$1$lcssa = $77;
         } else {
          $ce$027 = $ce$025;$stackCount$126 = $77;
          while(1) {
           $110 = ((($ce$027)) + 4|0); //@line 461 "libraries/Box2D/Dynamics/b2World.cpp"
           $111 = HEAP32[$110>>2]|0; //@line 461 "libraries/Box2D/Dynamics/b2World.cpp"
           $112 = ((($111)) + 4|0); //@line 464 "libraries/Box2D/Dynamics/b2World.cpp"
           $113 = HEAP32[$112>>2]|0; //@line 464 "libraries/Box2D/Dynamics/b2World.cpp"
           $114 = $113 & 7; //@line 464 "libraries/Box2D/Dynamics/b2World.cpp"
           $115 = ($114|0)==(6); //@line 464 "libraries/Box2D/Dynamics/b2World.cpp"
           do {
            if ($115) {
             $116 = ((($111)) + 48|0); //@line 477 "libraries/Box2D/Dynamics/b2World.cpp"
             $117 = HEAP32[$116>>2]|0; //@line 477 "libraries/Box2D/Dynamics/b2World.cpp"
             $118 = ((($117)) + 38|0); //@line 477 "libraries/Box2D/Dynamics/b2World.cpp"
             $119 = HEAP8[$118>>0]|0; //@line 477 "libraries/Box2D/Dynamics/b2World.cpp"
             $120 = ($119<<24>>24)==(0); //@line 477 "libraries/Box2D/Dynamics/b2World.cpp"
             if (!($120)) {
              $stackCount$4 = $stackCount$126;
              break;
             }
             $121 = ((($111)) + 52|0); //@line 478 "libraries/Box2D/Dynamics/b2World.cpp"
             $122 = HEAP32[$121>>2]|0; //@line 478 "libraries/Box2D/Dynamics/b2World.cpp"
             $123 = ((($122)) + 38|0); //@line 478 "libraries/Box2D/Dynamics/b2World.cpp"
             $124 = HEAP8[$123>>0]|0; //@line 478 "libraries/Box2D/Dynamics/b2World.cpp"
             $125 = ($124<<24>>24)==(0); //@line 478 "libraries/Box2D/Dynamics/b2World.cpp"
             if (!($125)) {
              $stackCount$4 = $stackCount$126;
              break;
             }
             $126 = HEAP32[$41>>2]|0; //@line 62 "./libraries/Box2D/Dynamics/b2Island.h"
             $127 = HEAP32[$47>>2]|0; //@line 62 "./libraries/Box2D/Dynamics/b2Island.h"
             $128 = ($126|0)<($127|0); //@line 62 "./libraries/Box2D/Dynamics/b2Island.h"
             __THREW__ = 0;
             invoke_vii(3822,($128|0),(361732|0)); //@line 62 "./libraries/Box2D/Dynamics/b2Island.h"
             $129 = __THREW__; __THREW__ = 0;
             $130 = $129&1;
             if ($130) {
              label = 13;
              break L17;
             }
             $131 = HEAP32[$41>>2]|0; //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
             $132 = (($131) + 1)|0; //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
             HEAP32[$41>>2] = $132; //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
             $133 = HEAP32[$48>>2]|0; //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
             $134 = (($133) + ($131<<2)|0); //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
             HEAP32[$134>>2] = $111; //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
             $135 = HEAP32[$112>>2]|0; //@line 485 "libraries/Box2D/Dynamics/b2World.cpp"
             $136 = $135 | 1; //@line 485 "libraries/Box2D/Dynamics/b2World.cpp"
             HEAP32[$112>>2] = $136; //@line 485 "libraries/Box2D/Dynamics/b2World.cpp"
             $137 = HEAP32[$ce$027>>2]|0; //@line 487 "libraries/Box2D/Dynamics/b2World.cpp"
             $138 = ((($137)) + 4|0); //@line 490 "libraries/Box2D/Dynamics/b2World.cpp"
             $139 = HEAP16[$138>>1]|0; //@line 490 "libraries/Box2D/Dynamics/b2World.cpp"
             $140 = $139 & 1; //@line 490 "libraries/Box2D/Dynamics/b2World.cpp"
             $141 = ($140<<16>>16)==(0); //@line 490 "libraries/Box2D/Dynamics/b2World.cpp"
             if (!($141)) {
              $stackCount$4 = $stackCount$126;
              break;
             }
             $142 = ($stackCount$126|0)<($31|0); //@line 495 "libraries/Box2D/Dynamics/b2World.cpp"
             __THREW__ = 0;
             invoke_vii(3822,($142|0),(361767|0)); //@line 495 "libraries/Box2D/Dynamics/b2World.cpp"
             $143 = __THREW__; __THREW__ = 0;
             $144 = $143&1;
             if ($144) {
              label = 13;
              break L17;
             }
             $145 = (($stackCount$126) + 1)|0; //@line 496 "libraries/Box2D/Dynamics/b2World.cpp"
             $146 = (($33) + ($stackCount$126<<2)|0); //@line 496 "libraries/Box2D/Dynamics/b2World.cpp"
             HEAP32[$146>>2] = $137; //@line 496 "libraries/Box2D/Dynamics/b2World.cpp"
             $147 = HEAP16[$138>>1]|0; //@line 497 "libraries/Box2D/Dynamics/b2World.cpp"
             $148 = $147&65535; //@line 497 "libraries/Box2D/Dynamics/b2World.cpp"
             $149 = $148 | 1; //@line 497 "libraries/Box2D/Dynamics/b2World.cpp"
             $150 = $149&65535; //@line 497 "libraries/Box2D/Dynamics/b2World.cpp"
             HEAP16[$138>>1] = $150; //@line 497 "libraries/Box2D/Dynamics/b2World.cpp"
             $stackCount$4 = $145;
            } else {
             $stackCount$4 = $stackCount$126;
            }
           } while(0);
           $151 = ((($ce$027)) + 12|0); //@line 459 "libraries/Box2D/Dynamics/b2World.cpp"
           $ce$0 = HEAP32[$151>>2]|0; //@line 459 "libraries/Box2D/Dynamics/b2World.cpp"
           $152 = ($ce$0|0)==(0|0); //@line 459 "libraries/Box2D/Dynamics/b2World.cpp"
           if ($152) {
            $stackCount$1$lcssa = $stackCount$4;
            break;
           } else {
            $ce$027 = $ce$0;$stackCount$126 = $stackCount$4;
           }
          }
         }
         $108 = ((($79)) + 108|0); //@line 501 "libraries/Box2D/Dynamics/b2World.cpp"
         $je$031 = HEAP32[$108>>2]|0; //@line 501 "libraries/Box2D/Dynamics/b2World.cpp"
         $109 = ($je$031|0)==(0|0); //@line 501 "libraries/Box2D/Dynamics/b2World.cpp"
         if ($109) {
          $stackCount$0$be = $stackCount$1$lcssa;
         } else {
          $je$033 = $je$031;$stackCount$532 = $stackCount$1$lcssa;
          while(1) {
           $153 = ((($je$033)) + 4|0); //@line 503 "libraries/Box2D/Dynamics/b2World.cpp"
           $154 = HEAP32[$153>>2]|0; //@line 503 "libraries/Box2D/Dynamics/b2World.cpp"
           $155 = ((($154)) + 60|0); //@line 503 "libraries/Box2D/Dynamics/b2World.cpp"
           $156 = HEAP8[$155>>0]|0; //@line 503 "libraries/Box2D/Dynamics/b2World.cpp"
           $157 = ($156<<24>>24)==(0); //@line 503 "libraries/Box2D/Dynamics/b2World.cpp"
           do {
            if ($157) {
             $158 = HEAP32[$je$033>>2]|0; //@line 508 "libraries/Box2D/Dynamics/b2World.cpp"
             $159 = ((($158)) + 4|0); //@line 661 "./libraries/Box2D/Dynamics/b2Body.h"
             $160 = HEAP16[$159>>1]|0; //@line 661 "./libraries/Box2D/Dynamics/b2Body.h"
             $161 = $160 & 32; //@line 661 "./libraries/Box2D/Dynamics/b2Body.h"
             $162 = ($161<<16>>16)==(0); //@line 661 "./libraries/Box2D/Dynamics/b2Body.h"
             if ($162) {
              $stackCount$7 = $stackCount$532;
              break;
             }
             $163 = HEAP32[$42>>2]|0; //@line 68 "./libraries/Box2D/Dynamics/b2Island.h"
             $164 = HEAP32[$45>>2]|0; //@line 68 "./libraries/Box2D/Dynamics/b2Island.h"
             $165 = ($163|0)<($164|0); //@line 68 "./libraries/Box2D/Dynamics/b2Island.h"
             __THREW__ = 0;
             invoke_vii(3822,($165|0),(361790|0)); //@line 68 "./libraries/Box2D/Dynamics/b2Island.h"
             $166 = __THREW__; __THREW__ = 0;
             $167 = $166&1;
             if ($167) {
              label = 12;
              break L17;
             }
             $168 = HEAP32[$42>>2]|0; //@line 69 "./libraries/Box2D/Dynamics/b2Island.h"
             $169 = (($168) + 1)|0; //@line 69 "./libraries/Box2D/Dynamics/b2Island.h"
             HEAP32[$42>>2] = $169; //@line 69 "./libraries/Box2D/Dynamics/b2Island.h"
             $170 = HEAP32[$46>>2]|0; //@line 69 "./libraries/Box2D/Dynamics/b2Island.h"
             $171 = (($170) + ($168<<2)|0); //@line 69 "./libraries/Box2D/Dynamics/b2Island.h"
             HEAP32[$171>>2] = $154; //@line 69 "./libraries/Box2D/Dynamics/b2Island.h"
             $172 = HEAP32[$153>>2]|0; //@line 517 "libraries/Box2D/Dynamics/b2World.cpp"
             $173 = ((($172)) + 60|0); //@line 517 "libraries/Box2D/Dynamics/b2World.cpp"
             HEAP8[$173>>0] = 1; //@line 517 "libraries/Box2D/Dynamics/b2World.cpp"
             $174 = HEAP16[$159>>1]|0; //@line 519 "libraries/Box2D/Dynamics/b2World.cpp"
             $175 = $174 & 1; //@line 519 "libraries/Box2D/Dynamics/b2World.cpp"
             $176 = ($175<<16>>16)==(0); //@line 519 "libraries/Box2D/Dynamics/b2World.cpp"
             if (!($176)) {
              $stackCount$7 = $stackCount$532;
              break;
             }
             $177 = ($stackCount$532|0)<($31|0); //@line 524 "libraries/Box2D/Dynamics/b2World.cpp"
             __THREW__ = 0;
             invoke_vii(3822,($177|0),(361767|0)); //@line 524 "libraries/Box2D/Dynamics/b2World.cpp"
             $178 = __THREW__; __THREW__ = 0;
             $179 = $178&1;
             if ($179) {
              label = 12;
              break L17;
             }
             $180 = (($stackCount$532) + 1)|0; //@line 525 "libraries/Box2D/Dynamics/b2World.cpp"
             $181 = (($33) + ($stackCount$532<<2)|0); //@line 525 "libraries/Box2D/Dynamics/b2World.cpp"
             HEAP32[$181>>2] = $158; //@line 525 "libraries/Box2D/Dynamics/b2World.cpp"
             $182 = HEAP16[$159>>1]|0; //@line 526 "libraries/Box2D/Dynamics/b2World.cpp"
             $183 = $182&65535; //@line 526 "libraries/Box2D/Dynamics/b2World.cpp"
             $184 = $183 | 1; //@line 526 "libraries/Box2D/Dynamics/b2World.cpp"
             $185 = $184&65535; //@line 526 "libraries/Box2D/Dynamics/b2World.cpp"
             HEAP16[$159>>1] = $185; //@line 526 "libraries/Box2D/Dynamics/b2World.cpp"
             $stackCount$7 = $180;
            } else {
             $stackCount$7 = $stackCount$532;
            }
           } while(0);
           $186 = ((($je$033)) + 12|0); //@line 501 "libraries/Box2D/Dynamics/b2World.cpp"
           $je$0 = HEAP32[$186>>2]|0; //@line 501 "libraries/Box2D/Dynamics/b2World.cpp"
           $187 = ($je$0|0)==(0|0); //@line 501 "libraries/Box2D/Dynamics/b2World.cpp"
           if ($187) {
            $stackCount$0$be = $stackCount$7;
            break;
           } else {
            $je$033 = $je$0;$stackCount$532 = $stackCount$7;
           }
          }
         }
        }
        $105 = ($stackCount$0$be|0)>(0); //@line 441 "libraries/Box2D/Dynamics/b2World.cpp"
        if ($105) {
         $stackCount$037 = $stackCount$0$be;
        } else {
         break;
        }
       }
       $188 = HEAP8[$50>>0]|0; //@line 531 "libraries/Box2D/Dynamics/b2World.cpp"
       $189 = ($188<<24>>24)!=(0); //@line 531 "libraries/Box2D/Dynamics/b2World.cpp"
       __THREW__ = 0;
       invoke_viiiii(3823,($island|0),($profile|0),($step|0),($49|0),($189|0)); //@line 531 "libraries/Box2D/Dynamics/b2World.cpp"
       $190 = __THREW__; __THREW__ = 0;
       $191 = $190&1;
       if ($191) {
        label = 15;
        break;
       }
       $192 = +HEAPF32[$51>>2]; //@line 532 "libraries/Box2D/Dynamics/b2World.cpp"
       $193 = +HEAPF32[$0>>2]; //@line 532 "libraries/Box2D/Dynamics/b2World.cpp"
       $194 = $192 + $193; //@line 532 "libraries/Box2D/Dynamics/b2World.cpp"
       HEAPF32[$0>>2] = $194; //@line 532 "libraries/Box2D/Dynamics/b2World.cpp"
       $195 = +HEAPF32[$52>>2]; //@line 533 "libraries/Box2D/Dynamics/b2World.cpp"
       $196 = +HEAPF32[$1>>2]; //@line 533 "libraries/Box2D/Dynamics/b2World.cpp"
       $197 = $195 + $196; //@line 533 "libraries/Box2D/Dynamics/b2World.cpp"
       HEAPF32[$1>>2] = $197; //@line 533 "libraries/Box2D/Dynamics/b2World.cpp"
       $198 = +HEAPF32[$53>>2]; //@line 534 "libraries/Box2D/Dynamics/b2World.cpp"
       $199 = +HEAPF32[$2>>2]; //@line 534 "libraries/Box2D/Dynamics/b2World.cpp"
       $200 = $198 + $199; //@line 534 "libraries/Box2D/Dynamics/b2World.cpp"
       HEAPF32[$2>>2] = $200; //@line 534 "libraries/Box2D/Dynamics/b2World.cpp"
       $201 = HEAP32[$40>>2]|0; //@line 537 "libraries/Box2D/Dynamics/b2World.cpp"
       $202 = ($201|0)>(0); //@line 537 "libraries/Box2D/Dynamics/b2World.cpp"
       if ($202) {
        $203 = HEAP32[$44>>2]|0; //@line 540 "libraries/Box2D/Dynamics/b2World.cpp"
        $i$038 = 0;
        while(1) {
         $204 = (($203) + ($i$038<<2)|0); //@line 540 "libraries/Box2D/Dynamics/b2World.cpp"
         $205 = HEAP32[$204>>2]|0; //@line 540 "libraries/Box2D/Dynamics/b2World.cpp"
         $206 = HEAP32[$205>>2]|0; //@line 471 "./libraries/Box2D/Dynamics/b2Body.h"
         $207 = ($206|0)==(0); //@line 541 "libraries/Box2D/Dynamics/b2World.cpp"
         if ($207) {
          $208 = ((($205)) + 4|0); //@line 543 "libraries/Box2D/Dynamics/b2World.cpp"
          $209 = HEAP16[$208>>1]|0; //@line 543 "libraries/Box2D/Dynamics/b2World.cpp"
          $210 = $209&65535; //@line 543 "libraries/Box2D/Dynamics/b2World.cpp"
          $211 = $210 & 65534; //@line 543 "libraries/Box2D/Dynamics/b2World.cpp"
          $212 = $211&65535; //@line 543 "libraries/Box2D/Dynamics/b2World.cpp"
          HEAP16[$208>>1] = $212; //@line 543 "libraries/Box2D/Dynamics/b2World.cpp"
         }
         $213 = (($i$038) + 1)|0; //@line 537 "libraries/Box2D/Dynamics/b2World.cpp"
         $214 = ($213|0)<($201|0); //@line 537 "libraries/Box2D/Dynamics/b2World.cpp"
         if ($214) {
          $i$038 = $213;
         } else {
          break;
         }
        }
       }
      }
     }
     $215 = ((($seed$043)) + 96|0); //@line 416 "libraries/Box2D/Dynamics/b2World.cpp"
     $seed$0 = HEAP32[$215>>2]|0; //@line 416 "libraries/Box2D/Dynamics/b2World.cpp"
     $216 = ($seed$0|0)==(0|0); //@line 416 "libraries/Box2D/Dynamics/b2World.cpp"
     if ($216) {
      break L15;
     } else {
      $seed$043 = $seed$0;
     }
    }
    if ((label|0) == 12) {
     $58 = ___cxa_find_matching_catch_2()|0;
     $59 = tempRet0;
     $lpad$phi$index = $58;$lpad$phi$index6 = $59;
     __ZN8b2IslandD2Ev($island); //@line 574 "libraries/Box2D/Dynamics/b2World.cpp"
     ___resumeException($lpad$phi$index|0);
     // unreachable;
    }
    else if ((label|0) == 13) {
     $60 = ___cxa_find_matching_catch_2()|0;
     $61 = tempRet0;
     $lpad$phi$index = $60;$lpad$phi$index6 = $61;
     __ZN8b2IslandD2Ev($island); //@line 574 "libraries/Box2D/Dynamics/b2World.cpp"
     ___resumeException($lpad$phi$index|0);
     // unreachable;
    }
    else if ((label|0) == 14) {
     $62 = ___cxa_find_matching_catch_2()|0;
     $63 = tempRet0;
     $lpad$phi$index = $62;$lpad$phi$index6 = $63;
     __ZN8b2IslandD2Ev($island); //@line 574 "libraries/Box2D/Dynamics/b2World.cpp"
     ___resumeException($lpad$phi$index|0);
     // unreachable;
    }
    else if ((label|0) == 15) {
     $64 = ___cxa_find_matching_catch_2()|0;
     $65 = tempRet0;
     $lpad$phi$index = $64;$lpad$phi$index6 = $65;
     __ZN8b2IslandD2Ev($island); //@line 574 "libraries/Box2D/Dynamics/b2World.cpp"
     ___resumeException($lpad$phi$index|0);
     // unreachable;
    }
   }
  } while(0);
  __THREW__ = 0;
  invoke_vii(3821,($10|0),($33|0)); //@line 548 "libraries/Box2D/Dynamics/b2World.cpp"
  $54 = __THREW__; __THREW__ = 0;
  $55 = $54&1;
  if (!($55)) {
   __THREW__ = 0;
   invoke_vi(3824,($timer|0)); //@line 551 "libraries/Box2D/Dynamics/b2World.cpp"
   $217 = __THREW__; __THREW__ = 0;
   $218 = $217&1;
   if (!($218)) {
    $b6$023 = HEAP32[$13>>2]|0; //@line 553 "libraries/Box2D/Dynamics/b2World.cpp"
    $219 = ($b6$023|0)==(0|0); //@line 553 "libraries/Box2D/Dynamics/b2World.cpp"
    L76: do {
     if (!($219)) {
      $b6$024 = $b6$023;
      while(1) {
       $222 = ((($b6$024)) + 4|0); //@line 556 "libraries/Box2D/Dynamics/b2World.cpp"
       $223 = HEAP16[$222>>1]|0; //@line 556 "libraries/Box2D/Dynamics/b2World.cpp"
       $224 = $223 & 1; //@line 556 "libraries/Box2D/Dynamics/b2World.cpp"
       $225 = ($224<<16>>16)==(0); //@line 556 "libraries/Box2D/Dynamics/b2World.cpp"
       if (!($225)) {
        $226 = HEAP32[$b6$024>>2]|0; //@line 471 "./libraries/Box2D/Dynamics/b2Body.h"
        $227 = ($226|0)==(0); //@line 561 "libraries/Box2D/Dynamics/b2World.cpp"
        if (!($227)) {
         __THREW__ = 0;
         invoke_vi(3826,($b6$024|0)); //@line 567 "libraries/Box2D/Dynamics/b2World.cpp"
         $228 = __THREW__; __THREW__ = 0;
         $229 = $228&1;
         if ($229) {
          break;
         }
        }
       }
       $230 = ((($b6$024)) + 96|0); //@line 719 "./libraries/Box2D/Dynamics/b2Body.h"
       $b6$0 = HEAP32[$230>>2]|0; //@line 553 "libraries/Box2D/Dynamics/b2World.cpp"
       $231 = ($b6$0|0)==(0|0); //@line 553 "libraries/Box2D/Dynamics/b2World.cpp"
       if ($231) {
        break L76;
       } else {
        $b6$024 = $b6$0;
       }
      }
      $56 = ___cxa_find_matching_catch_2()|0;
      $57 = tempRet0;
      $lpad$phi$index = $56;$lpad$phi$index6 = $57;
      __ZN8b2IslandD2Ev($island); //@line 574 "libraries/Box2D/Dynamics/b2World.cpp"
      ___resumeException($lpad$phi$index|0);
      // unreachable;
     }
    } while(0);
    __THREW__ = 0;
    invoke_vi(3825,($5|0)); //@line 571 "libraries/Box2D/Dynamics/b2World.cpp"
    $220 = __THREW__; __THREW__ = 0;
    $221 = $220&1;
    if (!($221)) {
     __THREW__ = 0;
     $232 = (+invoke_di(3827,($timer|0))); //@line 572 "libraries/Box2D/Dynamics/b2World.cpp"
     $233 = __THREW__; __THREW__ = 0;
     $234 = $233&1;
     if (!($234)) {
      $235 = ((($this)) + 103020|0); //@line 572 "libraries/Box2D/Dynamics/b2World.cpp"
      HEAPF32[$235>>2] = $232; //@line 572 "libraries/Box2D/Dynamics/b2World.cpp"
      __ZN8b2IslandD2Ev($island); //@line 574 "libraries/Box2D/Dynamics/b2World.cpp"
      STACKTOP = sp;return; //@line 574 "libraries/Box2D/Dynamics/b2World.cpp"
     }
    }
   }
  }
 }
 $66 = ___cxa_find_matching_catch_2()|0;
 $67 = tempRet0;
 $lpad$phi$index = $66;$lpad$phi$index6 = $67;
 __ZN8b2IslandD2Ev($island); //@line 574 "libraries/Box2D/Dynamics/b2World.cpp"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZN7b2World8SolveTOIERK10b2TimeStep($this,$step) {
 $this = $this|0;
 $step = $step|0;
 var $$phi$trans$insert = 0, $$pre = 0, $$pre$phiZ2D = 0, $$pre79 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0.0, $118 = 0, $119 = 0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0, $13 = 0;
 var $130 = 0.0, $131 = 0, $132 = 0.0, $133 = 0.0, $134 = 0, $135 = 0.0, $136 = 0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0, $144 = 0.0, $145 = 0, $146 = 0.0, $147 = 0.0, $148 = 0.0;
 var $149 = 0.0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0, $159 = 0.0, $16 = 0, $160 = 0, $161 = 0.0, $162 = 0.0, $163 = 0, $164 = 0.0, $165 = 0, $166 = 0.0;
 var $167 = 0.0, $168 = 0.0, $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0.0, $172 = 0, $173 = 0.0, $174 = 0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0.0, $201 = 0.0;
 var $202 = 0, $203 = 0.0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0.0, $222 = 0, $223 = 0, $224 = 0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0.0, $229 = 0, $23 = 0, $230 = 0.0, $231 = 0, $232 = 0.0, $233 = 0.0, $234 = 0, $235 = 0.0, $236 = 0, $237 = 0.0, $238 = 0.0;
 var $239 = 0.0, $24 = 0, $240 = 0.0, $241 = 0.0, $242 = 0.0, $243 = 0, $244 = 0.0, $245 = 0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0.0, $263 = 0, $264 = 0.0, $265 = 0, $266 = 0, $267 = 0.0, $268 = 0.0, $269 = 0, $27 = 0, $270 = 0.0, $271 = 0.0, $272 = 0.0, $273 = 0.0, $274 = 0.0;
 var $275 = 0.0, $276 = 0.0, $277 = 0.0, $278 = 0.0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0.0, $283 = 0, $284 = 0, $285 = 0, $286 = 0.0, $287 = 0.0, $288 = 0.0, $289 = 0.0, $29 = 0, $290 = 0, $291 = 0.0, $292 = 0;
 var $293 = 0.0, $294 = 0.0, $295 = 0, $296 = 0.0, $297 = 0, $298 = 0.0, $299 = 0.0, $3 = 0, $30 = 0, $300 = 0.0, $301 = 0.0, $302 = 0.0, $303 = 0.0, $304 = 0, $305 = 0.0, $306 = 0, $307 = 0.0, $308 = 0.0, $309 = 0.0, $31 = 0;
 var $310 = 0.0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0.0, $324 = 0, $325 = 0.0, $326 = 0, $327 = 0, $328 = 0.0;
 var $329 = 0.0, $33 = 0, $330 = 0, $331 = 0.0, $332 = 0.0, $333 = 0.0, $334 = 0.0, $335 = 0.0, $336 = 0.0, $337 = 0.0, $338 = 0.0, $339 = 0.0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0;
 var $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0.0, $355 = 0.0, $356 = 0.0, $357 = 0.0, $358 = 0.0, $359 = 0.0, $36 = 0, $360 = 0.0, $361 = 0.0, $362 = 0.0, $363 = 0.0, $364 = 0.0;
 var $365 = 0.0, $366 = 0.0, $367 = 0.0, $368 = 0.0, $369 = 0.0, $37 = 0, $370 = 0.0, $371 = 0.0, $372 = 0.0, $373 = 0.0, $374 = 0.0, $375 = 0.0, $376 = 0.0, $377 = 0.0, $378 = 0.0, $379 = 0.0, $38 = 0, $380 = 0.0, $381 = 0.0, $382 = 0.0;
 var $383 = 0.0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0;
 var $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0;
 var $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0;
 var $437 = 0, $438 = 0.0, $439 = 0.0, $44 = 0, $440 = 0.0, $441 = 0.0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0;
 var $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0;
 var $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0;
 var $491 = 0, $492 = 0, $493 = 0.0, $494 = 0, $495 = 0, $496 = 0, $497 = 0.0, $498 = 0.0, $499 = 0.0, $5 = 0, $50 = 0, $500 = 0.0, $501 = 0, $502 = 0.0, $503 = 0, $504 = 0.0, $505 = 0.0, $506 = 0, $507 = 0.0, $508 = 0;
 var $509 = 0.0, $51 = 0, $510 = 0.0, $511 = 0.0, $512 = 0.0, $513 = 0.0, $514 = 0.0, $515 = 0, $516 = 0.0, $517 = 0, $518 = 0.0, $519 = 0.0, $52 = 0, $520 = 0.0, $521 = 0.0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0;
 var $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0.0, $535 = 0, $536 = 0.0, $537 = 0, $538 = 0, $539 = 0.0, $54 = 0, $540 = 0.0, $541 = 0, $542 = 0.0, $543 = 0.0, $544 = 0.0;
 var $545 = 0.0, $546 = 0.0, $547 = 0.0, $548 = 0.0, $549 = 0.0, $55 = 0, $550 = 0.0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0.0, $561 = 0.0, $562 = 0;
 var $563 = 0.0, $564 = 0, $565 = 0, $566 = 0.0, $567 = 0.0, $568 = 0, $569 = 0.0, $57 = 0, $570 = 0.0, $571 = 0.0, $572 = 0.0, $573 = 0.0, $574 = 0.0, $575 = 0, $576 = 0.0, $577 = 0.0, $578 = 0, $579 = 0.0, $58 = 0, $580 = 0.0;
 var $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0.0, $587 = 0.0, $588 = 0, $589 = 0.0, $59 = 0, $590 = 0, $591 = 0, $592 = 0.0, $593 = 0.0, $594 = 0, $595 = 0.0, $596 = 0.0, $597 = 0.0, $598 = 0.0, $599 = 0.0;
 var $6 = 0, $60 = 0, $600 = 0.0, $601 = 0, $602 = 0.0, $603 = 0.0, $604 = 0, $605 = 0.0, $606 = 0.0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0;
 var $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0;
 var $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0;
 var $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0;
 var $671 = 0, $672 = 0, $673 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0.0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $alpha$0 = 0.0, $alpha$4 = 0.0;
 var $alpha0$0 = 0.0, $b$0 = 0, $b$074 = 0, $b$075 = 0, $backup = 0, $backup1 = 0, $backup2 = 0, $bitcast$i = 0.0, $bitcast$i15 = 0.0, $bitcast$i18 = 0.0, $bodies = 0, $c$0 = 0, $c$070 = 0, $c$071 = 0, $c1$0 = 0, $c1$049 = 0, $c1$04981 = 0, $c1$052 = 0, $ce$0 = 0, $ce$056 = 0;
 var $ce$057 = 0, $ce10$0 = 0, $ce10$063 = 0, $ce10$064 = 0, $i$062 = 0, $i8$067 = 0, $input = 0, $island = 0, $lpad$phi$index = 0, $lpad$phi$index4 = 0, $minAlpha$050 = 0.0, $minAlpha$1 = 0.0, $minAlpha$3 = 0.0, $minContact$051 = 0, $minContact$1 = 0, $minContact$3 = 0, $or$cond = 0, $or$cond28 = 0, $or$cond5 = 0, $or$cond7 = 0;
 var $output = 0, $subStep = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 336|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $island = sp + 284|0;
 $input = sp + 152|0;
 $output = sp + 144|0;
 $backup1 = sp + 108|0;
 $backup2 = sp + 72|0;
 $bodies = sp + 64|0;
 $backup = sp + 24|0;
 $subStep = sp;
 $0 = ((($this)) + 68|0); //@line 579 "libraries/Box2D/Dynamics/b2World.cpp"
 $1 = ((($this)) + 102872|0); //@line 579 "libraries/Box2D/Dynamics/b2World.cpp"
 $2 = ((($this)) + 102944|0); //@line 579 "libraries/Box2D/Dynamics/b2World.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 579 "libraries/Box2D/Dynamics/b2World.cpp"
 __ZN8b2IslandC2EiiiP16b2StackAllocatorP17b2ContactListener($island,64,32,0,$0,$3); //@line 579 "libraries/Box2D/Dynamics/b2World.cpp"
 $4 = ((($this)) + 102995|0); //@line 581 "libraries/Box2D/Dynamics/b2World.cpp"
 $5 = HEAP8[$4>>0]|0; //@line 581 "libraries/Box2D/Dynamics/b2World.cpp"
 $6 = ($5<<24>>24)==(0); //@line 581 "libraries/Box2D/Dynamics/b2World.cpp"
 if ($6) {
  $$pre79 = ((($this)) + 102932|0); //@line 605 "libraries/Box2D/Dynamics/b2World.cpp"
  $$pre$phiZ2D = $$pre79;
 } else {
  $7 = ((($this)) + 102952|0); //@line 583 "libraries/Box2D/Dynamics/b2World.cpp"
  $b$074 = HEAP32[$7>>2]|0; //@line 583 "libraries/Box2D/Dynamics/b2World.cpp"
  $8 = ($b$074|0)==(0|0); //@line 583 "libraries/Box2D/Dynamics/b2World.cpp"
  if (!($8)) {
   $b$075 = $b$074;
   while(1) {
    $11 = ((($b$075)) + 4|0); //@line 585 "libraries/Box2D/Dynamics/b2World.cpp"
    $12 = HEAP16[$11>>1]|0; //@line 585 "libraries/Box2D/Dynamics/b2World.cpp"
    $13 = $12&65535; //@line 585 "libraries/Box2D/Dynamics/b2World.cpp"
    $14 = $13 & 65534; //@line 585 "libraries/Box2D/Dynamics/b2World.cpp"
    $15 = $14&65535; //@line 585 "libraries/Box2D/Dynamics/b2World.cpp"
    HEAP16[$11>>1] = $15; //@line 585 "libraries/Box2D/Dynamics/b2World.cpp"
    $16 = ((($b$075)) + 60|0); //@line 586 "libraries/Box2D/Dynamics/b2World.cpp"
    HEAPF32[$16>>2] = 0.0; //@line 586 "libraries/Box2D/Dynamics/b2World.cpp"
    $17 = ((($b$075)) + 96|0); //@line 583 "libraries/Box2D/Dynamics/b2World.cpp"
    $b$0 = HEAP32[$17>>2]|0; //@line 583 "libraries/Box2D/Dynamics/b2World.cpp"
    $18 = ($b$0|0)==(0|0); //@line 583 "libraries/Box2D/Dynamics/b2World.cpp"
    if ($18) {
     break;
    } else {
     $b$075 = $b$0;
    }
   }
  }
  $9 = ((($this)) + 102932|0); //@line 589 "libraries/Box2D/Dynamics/b2World.cpp"
  $c$070 = HEAP32[$9>>2]|0; //@line 589 "libraries/Box2D/Dynamics/b2World.cpp"
  $10 = ($c$070|0)==(0|0); //@line 589 "libraries/Box2D/Dynamics/b2World.cpp"
  if ($10) {
   $$pre$phiZ2D = $9;
  } else {
   $c$071 = $c$070;
   while(1) {
    $46 = ((($c$071)) + 4|0); //@line 592 "libraries/Box2D/Dynamics/b2World.cpp"
    $47 = HEAP32[$46>>2]|0; //@line 592 "libraries/Box2D/Dynamics/b2World.cpp"
    $48 = $47 & -34; //@line 592 "libraries/Box2D/Dynamics/b2World.cpp"
    HEAP32[$46>>2] = $48; //@line 592 "libraries/Box2D/Dynamics/b2World.cpp"
    $49 = ((($c$071)) + 128|0); //@line 593 "libraries/Box2D/Dynamics/b2World.cpp"
    HEAP32[$49>>2] = 0; //@line 593 "libraries/Box2D/Dynamics/b2World.cpp"
    $50 = ((($c$071)) + 132|0); //@line 594 "libraries/Box2D/Dynamics/b2World.cpp"
    HEAPF32[$50>>2] = 1.0; //@line 594 "libraries/Box2D/Dynamics/b2World.cpp"
    $51 = ((($c$071)) + 12|0); //@line 589 "libraries/Box2D/Dynamics/b2World.cpp"
    $c$0 = HEAP32[$51>>2]|0; //@line 589 "libraries/Box2D/Dynamics/b2World.cpp"
    $52 = ($c$0|0)==(0|0); //@line 589 "libraries/Box2D/Dynamics/b2World.cpp"
    if ($52) {
     $$pre$phiZ2D = $9;
     break;
    } else {
     $c$071 = $c$0;
    }
   }
  }
 }
 $19 = ((($island)) + 28|0); //@line 43 "./libraries/Box2D/Dynamics/b2Island.h"
 $20 = ((($island)) + 36|0); //@line 44 "./libraries/Box2D/Dynamics/b2Island.h"
 $21 = ((($island)) + 32|0); //@line 45 "./libraries/Box2D/Dynamics/b2Island.h"
 $22 = ((($island)) + 40|0); //@line 54 "./libraries/Box2D/Dynamics/b2Island.h"
 $23 = ((($island)) + 8|0); //@line 56 "./libraries/Box2D/Dynamics/b2Island.h"
 $24 = ((($island)) + 44|0); //@line 62 "./libraries/Box2D/Dynamics/b2Island.h"
 $25 = ((($island)) + 12|0); //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
 $26 = ((($bodies)) + 4|0); //@line 765 "libraries/Box2D/Dynamics/b2World.cpp"
 $27 = ((($subStep)) + 4|0); //@line 858 "libraries/Box2D/Dynamics/b2World.cpp"
 $28 = ((($subStep)) + 8|0); //@line 859 "libraries/Box2D/Dynamics/b2World.cpp"
 $29 = ((($subStep)) + 16|0); //@line 860 "libraries/Box2D/Dynamics/b2World.cpp"
 $30 = ((($step)) + 12|0); //@line 861 "libraries/Box2D/Dynamics/b2World.cpp"
 $31 = ((($subStep)) + 12|0); //@line 861 "libraries/Box2D/Dynamics/b2World.cpp"
 $32 = ((($subStep)) + 20|0); //@line 862 "libraries/Box2D/Dynamics/b2World.cpp"
 $33 = ((($this)) + 102994|0); //@line 889 "libraries/Box2D/Dynamics/b2World.cpp"
 $34 = ((($input)) + 16|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $35 = ((($input)) + 20|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $36 = ((($input)) + 24|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $37 = ((($input)) + 44|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $38 = ((($input)) + 48|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $39 = ((($input)) + 52|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $40 = ((($input)) + 28|0); //@line 684 "libraries/Box2D/Dynamics/b2World.cpp"
 $41 = ((($input)) + 56|0); //@line 685 "libraries/Box2D/Dynamics/b2World.cpp"
 $42 = ((($input)) + 92|0); //@line 686 "libraries/Box2D/Dynamics/b2World.cpp"
 $43 = ((($input)) + 128|0); //@line 687 "libraries/Box2D/Dynamics/b2World.cpp"
 $44 = ((($output)) + 4|0); //@line 693 "libraries/Box2D/Dynamics/b2World.cpp"
 $c1$04981 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 605 "libraries/Box2D/Dynamics/b2World.cpp"
 $45 = ($c1$04981|0)==(0|0); //@line 605 "libraries/Box2D/Dynamics/b2World.cpp"
 L11: do {
  if (!($45)) {
   $c1$052 = $c1$04981;$minAlpha$050 = 1.0;$minContact$051 = 0;
   L12: while(1) {
    $55 = ((($c1$052)) + 4|0); //@line 256 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
    $56 = HEAP32[$55>>2]|0; //@line 256 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
    $57 = $56 & 4; //@line 256 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
    $58 = ($57|0)==(0); //@line 256 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
    do {
     if ($58) {
      $minAlpha$3 = $minAlpha$050;$minContact$3 = $minContact$051;
     } else {
      $67 = ((($c1$052)) + 128|0); //@line 614 "libraries/Box2D/Dynamics/b2World.cpp"
      $68 = HEAP32[$67>>2]|0; //@line 614 "libraries/Box2D/Dynamics/b2World.cpp"
      $69 = ($68|0)>(8); //@line 614 "libraries/Box2D/Dynamics/b2World.cpp"
      if ($69) {
       $minAlpha$3 = $minAlpha$050;$minContact$3 = $minContact$051;
      } else {
       $70 = $56 & 32; //@line 620 "libraries/Box2D/Dynamics/b2World.cpp"
       $71 = ($70|0)==(0); //@line 620 "libraries/Box2D/Dynamics/b2World.cpp"
       if ($71) {
        $74 = ((($c1$052)) + 48|0); //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
        $75 = HEAP32[$74>>2]|0; //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
        $76 = ((($c1$052)) + 52|0); //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
        $77 = HEAP32[$76>>2]|0; //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
        $78 = ((($75)) + 38|0); //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
        $79 = HEAP8[$78>>0]|0; //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
        $80 = ($79<<24>>24)==(0); //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
        if (!($80)) {
         $minAlpha$3 = $minAlpha$050;$minContact$3 = $minContact$051;
         break;
        }
        $81 = ((($77)) + 38|0); //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
        $82 = HEAP8[$81>>0]|0; //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
        $83 = ($82<<24>>24)==(0); //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
        if (!($83)) {
         $minAlpha$3 = $minAlpha$050;$minContact$3 = $minContact$051;
         break;
        }
        $84 = ((($75)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
        $85 = HEAP32[$84>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
        $86 = ((($77)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
        $87 = HEAP32[$86>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
        $88 = HEAP32[$85>>2]|0; //@line 639 "libraries/Box2D/Dynamics/b2World.cpp"
        $89 = HEAP32[$87>>2]|0; //@line 640 "libraries/Box2D/Dynamics/b2World.cpp"
        $90 = ($88|0)==(2); //@line 641 "libraries/Box2D/Dynamics/b2World.cpp"
        $91 = ($89|0)==(2); //@line 641 "libraries/Box2D/Dynamics/b2World.cpp"
        $92 = $90 | $91; //@line 641 "libraries/Box2D/Dynamics/b2World.cpp"
        __THREW__ = 0;
        invoke_vii(3822,($92|0),(361821|0)); //@line 641 "libraries/Box2D/Dynamics/b2World.cpp"
        $93 = __THREW__; __THREW__ = 0;
        $94 = $93&1;
        if ($94) {
         label = 12;
         break L12;
        }
        $95 = ((($85)) + 4|0); //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
        $96 = HEAP16[$95>>1]|0; //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
        $97 = $96 & 2; //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
        $98 = ($97<<16>>16)!=(0); //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
        $99 = ($88|0)!=(0); //@line 643 "libraries/Box2D/Dynamics/b2World.cpp"
        $100 = $99 & $98; //@line 643 "libraries/Box2D/Dynamics/b2World.cpp"
        $101 = ((($87)) + 4|0); //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
        $102 = HEAP16[$101>>1]|0; //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
        $103 = $102 & 2; //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
        $104 = ($103<<16>>16)!=(0); //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
        $105 = ($89|0)!=(0); //@line 644 "libraries/Box2D/Dynamics/b2World.cpp"
        $106 = $105 & $104; //@line 644 "libraries/Box2D/Dynamics/b2World.cpp"
        $or$cond5 = $100 | $106; //@line 647 "libraries/Box2D/Dynamics/b2World.cpp"
        if (!($or$cond5)) {
         $minAlpha$3 = $minAlpha$050;$minContact$3 = $minContact$051;
         break;
        }
        $107 = $96 & 8; //@line 630 "./libraries/Box2D/Dynamics/b2Body.h"
        $108 = ($107<<16>>16)!=(0); //@line 630 "./libraries/Box2D/Dynamics/b2Body.h"
        $109 = ($88|0)!=(2); //@line 652 "libraries/Box2D/Dynamics/b2World.cpp"
        $110 = $109 | $108; //@line 652 "libraries/Box2D/Dynamics/b2World.cpp"
        $111 = $102 & 8; //@line 630 "./libraries/Box2D/Dynamics/b2Body.h"
        $112 = ($111<<16>>16)!=(0); //@line 630 "./libraries/Box2D/Dynamics/b2Body.h"
        $113 = ($89|0)!=(2); //@line 653 "libraries/Box2D/Dynamics/b2World.cpp"
        $114 = $113 | $112; //@line 653 "libraries/Box2D/Dynamics/b2World.cpp"
        $or$cond7 = $110 | $114; //@line 656 "libraries/Box2D/Dynamics/b2World.cpp"
        if (!($or$cond7)) {
         $minAlpha$3 = $minAlpha$050;$minContact$3 = $minContact$051;
         break;
        }
        $115 = ((($85)) + 28|0); //@line 663 "libraries/Box2D/Dynamics/b2World.cpp"
        $116 = ((($85)) + 60|0); //@line 663 "libraries/Box2D/Dynamics/b2World.cpp"
        $117 = +HEAPF32[$116>>2]; //@line 704 "./libraries/Box2D/Common/b2Math.h"
        $118 = ((($87)) + 28|0); //@line 665 "libraries/Box2D/Dynamics/b2World.cpp"
        $119 = ((($87)) + 60|0); //@line 665 "libraries/Box2D/Dynamics/b2World.cpp"
        $120 = +HEAPF32[$119>>2]; //@line 704 "./libraries/Box2D/Common/b2Math.h"
        $121 = $117 < $120; //@line 665 "libraries/Box2D/Dynamics/b2World.cpp"
        if ($121) {
         $122 = $117 < 1.0; //@line 704 "./libraries/Box2D/Common/b2Math.h"
         __THREW__ = 0;
         invoke_vii(3822,($122|0),(361872|0)); //@line 704 "./libraries/Box2D/Common/b2Math.h"
         $123 = __THREW__; __THREW__ = 0;
         $124 = $123&1;
         if ($124) {
          label = 12;
          break L12;
         }
         $125 = +HEAPF32[$116>>2]; //@line 705 "./libraries/Box2D/Common/b2Math.h"
         $126 = $120 - $125; //@line 705 "./libraries/Box2D/Common/b2Math.h"
         $127 = 1.0 - $125; //@line 705 "./libraries/Box2D/Common/b2Math.h"
         $128 = $126 / $127; //@line 705 "./libraries/Box2D/Common/b2Math.h"
         $129 = ((($85)) + 44|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
         $130 = +HEAPF32[$129>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
         $131 = ((($85)) + 36|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
         $132 = +HEAPF32[$131>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
         $133 = $130 - $132; //@line 454 "./libraries/Box2D/Common/b2Math.h"
         $134 = ((($85)) + 48|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
         $135 = +HEAPF32[$134>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
         $136 = ((($85)) + 40|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
         $137 = +HEAPF32[$136>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
         $138 = $135 - $137; //@line 454 "./libraries/Box2D/Common/b2Math.h"
         $139 = $128 * $133; //@line 459 "./libraries/Box2D/Common/b2Math.h"
         $140 = $128 * $138; //@line 459 "./libraries/Box2D/Common/b2Math.h"
         $141 = $132 + $139; //@line 85 "./libraries/Box2D/Common/b2Math.h"
         HEAPF32[$131>>2] = $141; //@line 85 "./libraries/Box2D/Common/b2Math.h"
         $142 = $137 + $140; //@line 85 "./libraries/Box2D/Common/b2Math.h"
         HEAPF32[$136>>2] = $142; //@line 85 "./libraries/Box2D/Common/b2Math.h"
         $143 = ((($85)) + 56|0); //@line 707 "./libraries/Box2D/Common/b2Math.h"
         $144 = +HEAPF32[$143>>2]; //@line 707 "./libraries/Box2D/Common/b2Math.h"
         $145 = ((($85)) + 52|0); //@line 707 "./libraries/Box2D/Common/b2Math.h"
         $146 = +HEAPF32[$145>>2]; //@line 707 "./libraries/Box2D/Common/b2Math.h"
         $147 = $144 - $146; //@line 707 "./libraries/Box2D/Common/b2Math.h"
         $148 = $128 * $147; //@line 707 "./libraries/Box2D/Common/b2Math.h"
         $149 = $146 + $148; //@line 707 "./libraries/Box2D/Common/b2Math.h"
         HEAPF32[$145>>2] = $149; //@line 707 "./libraries/Box2D/Common/b2Math.h"
         HEAPF32[$116>>2] = $120; //@line 708 "./libraries/Box2D/Common/b2Math.h"
         $alpha0$0 = $120;
        } else {
         $150 = $120 < $117; //@line 670 "libraries/Box2D/Dynamics/b2World.cpp"
         if ($150) {
          $151 = $120 < 1.0; //@line 704 "./libraries/Box2D/Common/b2Math.h"
          __THREW__ = 0;
          invoke_vii(3822,($151|0),(361872|0)); //@line 704 "./libraries/Box2D/Common/b2Math.h"
          $152 = __THREW__; __THREW__ = 0;
          $153 = $152&1;
          if ($153) {
           label = 12;
           break L12;
          }
          $154 = +HEAPF32[$119>>2]; //@line 705 "./libraries/Box2D/Common/b2Math.h"
          $155 = $117 - $154; //@line 705 "./libraries/Box2D/Common/b2Math.h"
          $156 = 1.0 - $154; //@line 705 "./libraries/Box2D/Common/b2Math.h"
          $157 = $155 / $156; //@line 705 "./libraries/Box2D/Common/b2Math.h"
          $158 = ((($87)) + 44|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
          $159 = +HEAPF32[$158>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
          $160 = ((($87)) + 36|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
          $161 = +HEAPF32[$160>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
          $162 = $159 - $161; //@line 454 "./libraries/Box2D/Common/b2Math.h"
          $163 = ((($87)) + 48|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
          $164 = +HEAPF32[$163>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
          $165 = ((($87)) + 40|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
          $166 = +HEAPF32[$165>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
          $167 = $164 - $166; //@line 454 "./libraries/Box2D/Common/b2Math.h"
          $168 = $157 * $162; //@line 459 "./libraries/Box2D/Common/b2Math.h"
          $169 = $157 * $167; //@line 459 "./libraries/Box2D/Common/b2Math.h"
          $170 = $161 + $168; //@line 85 "./libraries/Box2D/Common/b2Math.h"
          HEAPF32[$160>>2] = $170; //@line 85 "./libraries/Box2D/Common/b2Math.h"
          $171 = $166 + $169; //@line 85 "./libraries/Box2D/Common/b2Math.h"
          HEAPF32[$165>>2] = $171; //@line 85 "./libraries/Box2D/Common/b2Math.h"
          $172 = ((($87)) + 56|0); //@line 707 "./libraries/Box2D/Common/b2Math.h"
          $173 = +HEAPF32[$172>>2]; //@line 707 "./libraries/Box2D/Common/b2Math.h"
          $174 = ((($87)) + 52|0); //@line 707 "./libraries/Box2D/Common/b2Math.h"
          $175 = +HEAPF32[$174>>2]; //@line 707 "./libraries/Box2D/Common/b2Math.h"
          $176 = $173 - $175; //@line 707 "./libraries/Box2D/Common/b2Math.h"
          $177 = $157 * $176; //@line 707 "./libraries/Box2D/Common/b2Math.h"
          $178 = $175 + $177; //@line 707 "./libraries/Box2D/Common/b2Math.h"
          HEAPF32[$174>>2] = $178; //@line 707 "./libraries/Box2D/Common/b2Math.h"
          HEAPF32[$119>>2] = $117; //@line 708 "./libraries/Box2D/Common/b2Math.h"
          $alpha0$0 = $117;
         } else {
          $alpha0$0 = $117;
         }
        }
        $179 = $alpha0$0 < 1.0; //@line 676 "libraries/Box2D/Dynamics/b2World.cpp"
        __THREW__ = 0;
        invoke_vii(3822,($179|0),(361872|0)); //@line 676 "libraries/Box2D/Dynamics/b2World.cpp"
        $180 = __THREW__; __THREW__ = 0;
        $181 = $180&1;
        if ($181) {
         label = 12;
         break L12;
        }
        $182 = ((($c1$052)) + 56|0); //@line 291 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
        $183 = HEAP32[$182>>2]|0; //@line 291 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
        $184 = ((($c1$052)) + 60|0); //@line 301 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
        $185 = HEAP32[$184>>2]|0; //@line 301 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
        HEAP32[$34>>2] = 0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
        HEAP32[$35>>2] = 0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
        HEAPF32[$36>>2] = 0.0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
        HEAP32[$37>>2] = 0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
        HEAP32[$38>>2] = 0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
        HEAPF32[$39>>2] = 0.0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
        $186 = ((($75)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
        $187 = HEAP32[$186>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
        __THREW__ = 0;
        invoke_viii(3693,($input|0),($187|0),($183|0)); //@line 683 "libraries/Box2D/Dynamics/b2World.cpp"
        $188 = __THREW__; __THREW__ = 0;
        $189 = $188&1;
        if ($189) {
         label = 12;
         break L12;
        }
        $190 = ((($77)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
        $191 = HEAP32[$190>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
        __THREW__ = 0;
        invoke_viii(3693,($40|0),($191|0),($185|0)); //@line 684 "libraries/Box2D/Dynamics/b2World.cpp"
        $192 = __THREW__; __THREW__ = 0;
        $193 = $192&1;
        if ($193) {
         label = 12;
         break L12;
        }
        dest=$41; src=$115; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 685 "libraries/Box2D/Dynamics/b2World.cpp"
        dest=$42; src=$118; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 686 "libraries/Box2D/Dynamics/b2World.cpp"
        HEAPF32[$43>>2] = 1.0; //@line 687 "libraries/Box2D/Dynamics/b2World.cpp"
        __THREW__ = 0;
        invoke_vii(3828,($output|0),($input|0)); //@line 690 "libraries/Box2D/Dynamics/b2World.cpp"
        $194 = __THREW__; __THREW__ = 0;
        $195 = $194&1;
        if ($195) {
         label = 12;
         break L12;
        }
        $196 = HEAP32[$output>>2]|0; //@line 694 "libraries/Box2D/Dynamics/b2World.cpp"
        $197 = ($196|0)==(3); //@line 694 "libraries/Box2D/Dynamics/b2World.cpp"
        if ($197) {
         $198 = +HEAPF32[$44>>2]; //@line 693 "libraries/Box2D/Dynamics/b2World.cpp"
         $199 = 1.0 - $alpha0$0; //@line 696 "libraries/Box2D/Dynamics/b2World.cpp"
         $200 = $199 * $198; //@line 696 "libraries/Box2D/Dynamics/b2World.cpp"
         $201 = $alpha0$0 + $200; //@line 696 "libraries/Box2D/Dynamics/b2World.cpp"
         $202 = $201 < 1.0; //@line 634 "./libraries/Box2D/Common/b2Math.h"
         $203 = $202 ? $201 : 1.0; //@line 634 "./libraries/Box2D/Common/b2Math.h"
         $alpha$0 = $203;
        } else {
         $alpha$0 = 1.0;
        }
        $204 = ((($c1$052)) + 132|0); //@line 703 "libraries/Box2D/Dynamics/b2World.cpp"
        HEAPF32[$204>>2] = $alpha$0; //@line 703 "libraries/Box2D/Dynamics/b2World.cpp"
        $205 = HEAP32[$55>>2]|0; //@line 704 "libraries/Box2D/Dynamics/b2World.cpp"
        $206 = $205 | 32; //@line 704 "libraries/Box2D/Dynamics/b2World.cpp"
        HEAP32[$55>>2] = $206; //@line 704 "libraries/Box2D/Dynamics/b2World.cpp"
        $alpha$4 = $alpha$0;
       } else {
        $72 = ((($c1$052)) + 132|0); //@line 623 "libraries/Box2D/Dynamics/b2World.cpp"
        $73 = +HEAPF32[$72>>2]; //@line 623 "libraries/Box2D/Dynamics/b2World.cpp"
        $alpha$4 = $73;
       }
       $207 = $alpha$4 < $minAlpha$050; //@line 707 "libraries/Box2D/Dynamics/b2World.cpp"
       $minAlpha$1 = $207 ? $alpha$4 : $minAlpha$050; //@line 707 "libraries/Box2D/Dynamics/b2World.cpp"
       $minContact$1 = $207 ? $c1$052 : $minContact$051; //@line 707 "libraries/Box2D/Dynamics/b2World.cpp"
       $minAlpha$3 = $minAlpha$1;$minContact$3 = $minContact$1;
      }
     }
    } while(0);
    $208 = ((($c1$052)) + 12|0); //@line 605 "libraries/Box2D/Dynamics/b2World.cpp"
    $c1$0 = HEAP32[$208>>2]|0; //@line 605 "libraries/Box2D/Dynamics/b2World.cpp"
    $209 = ($c1$0|0)==(0|0); //@line 605 "libraries/Box2D/Dynamics/b2World.cpp"
    if (!($209)) {
     $c1$052 = $c1$0;$minAlpha$050 = $minAlpha$3;$minContact$051 = $minContact$3;
     continue;
    }
    $53 = ($minContact$3|0)==(0|0); //@line 715 "libraries/Box2D/Dynamics/b2World.cpp"
    $54 = $minAlpha$3 > 0.99999880790710449; //@line 715 "libraries/Box2D/Dynamics/b2World.cpp"
    $or$cond = $54 | $53; //@line 715 "libraries/Box2D/Dynamics/b2World.cpp"
    if ($or$cond) {
     break L11;
    }
    $210 = ((($minContact$3)) + 48|0); //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
    $211 = HEAP32[$210>>2]|0; //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
    $212 = ((($minContact$3)) + 52|0); //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
    $213 = HEAP32[$212>>2]|0; //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
    $214 = ((($211)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
    $215 = HEAP32[$214>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
    $216 = ((($213)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
    $217 = HEAP32[$216>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
    $218 = ((($215)) + 28|0); //@line 728 "libraries/Box2D/Dynamics/b2World.cpp"
    dest=$backup1; src=$218; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 728 "libraries/Box2D/Dynamics/b2World.cpp"
    $219 = ((($217)) + 28|0); //@line 729 "libraries/Box2D/Dynamics/b2World.cpp"
    dest=$backup2; src=$219; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 729 "libraries/Box2D/Dynamics/b2World.cpp"
    $220 = ((($215)) + 60|0); //@line 704 "./libraries/Box2D/Common/b2Math.h"
    $221 = +HEAPF32[$220>>2]; //@line 704 "./libraries/Box2D/Common/b2Math.h"
    $222 = $221 < 1.0; //@line 704 "./libraries/Box2D/Common/b2Math.h"
    __THREW__ = 0;
    invoke_vii(3822,($222|0),(361872|0)); //@line 704 "./libraries/Box2D/Common/b2Math.h"
    $223 = __THREW__; __THREW__ = 0;
    $224 = $223&1;
    if ($224) {
     label = 13;
     break;
    }
    $225 = +HEAPF32[$220>>2]; //@line 705 "./libraries/Box2D/Common/b2Math.h"
    $226 = $minAlpha$3 - $225; //@line 705 "./libraries/Box2D/Common/b2Math.h"
    $227 = 1.0 - $225; //@line 705 "./libraries/Box2D/Common/b2Math.h"
    $228 = $226 / $227; //@line 705 "./libraries/Box2D/Common/b2Math.h"
    $229 = ((($215)) + 44|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $230 = +HEAPF32[$229>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $231 = ((($215)) + 36|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $232 = +HEAPF32[$231>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $233 = $230 - $232; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $234 = ((($215)) + 48|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $235 = +HEAPF32[$234>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $236 = ((($215)) + 40|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $237 = +HEAPF32[$236>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $238 = $235 - $237; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $239 = $228 * $233; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $240 = $228 * $238; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $241 = $232 + $239; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$231>>2] = $241; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $242 = $237 + $240; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$236>>2] = $242; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $243 = ((($215)) + 56|0); //@line 707 "./libraries/Box2D/Common/b2Math.h"
    $244 = +HEAPF32[$243>>2]; //@line 707 "./libraries/Box2D/Common/b2Math.h"
    $245 = ((($215)) + 52|0); //@line 707 "./libraries/Box2D/Common/b2Math.h"
    $246 = +HEAPF32[$245>>2]; //@line 707 "./libraries/Box2D/Common/b2Math.h"
    $247 = $244 - $246; //@line 707 "./libraries/Box2D/Common/b2Math.h"
    $248 = $228 * $247; //@line 707 "./libraries/Box2D/Common/b2Math.h"
    $249 = $246 + $248; //@line 707 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$245>>2] = $249; //@line 707 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$220>>2] = $minAlpha$3; //@line 708 "./libraries/Box2D/Common/b2Math.h"
    $250 = ((($215)) + 44|0); //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $251 = ((($215)) + 36|0); //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $252 = $251; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $253 = $252; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $254 = HEAP32[$253>>2]|0; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $255 = (($252) + 4)|0; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $256 = $255; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $257 = HEAP32[$256>>2]|0; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $258 = $250; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $259 = $258; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAP32[$259>>2] = $254; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $260 = (($258) + 4)|0; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $261 = $260; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAP32[$261>>2] = $257; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAPF32[$243>>2] = $249; //@line 845 "./libraries/Box2D/Dynamics/b2Body.h"
    $262 = (+Math_sin((+$249))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
    $263 = ((($215)) + 20|0); //@line 315 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$263>>2] = $262; //@line 315 "./libraries/Box2D/Common/b2Math.h"
    $264 = (+Math_cos((+$249))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
    $265 = ((($215)) + 24|0); //@line 316 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$265>>2] = $264; //@line 316 "./libraries/Box2D/Common/b2Math.h"
    $266 = ((($215)) + 28|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $267 = +HEAPF32[$266>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $268 = $264 * $267; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $269 = ((($215)) + 32|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $270 = +HEAPF32[$269>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $271 = $262 * $270; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $272 = $268 - $271; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $273 = $262 * $267; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $274 = $264 * $270; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $275 = $273 + $274; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $bitcast$i18 = (HEAP32[tempDoublePtr>>2]=$254,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $276 = $bitcast$i18 - $272; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $277 = (HEAP32[tempDoublePtr>>2]=$257,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $278 = $277 - $275; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $279 = ((($215)) + 12|0); //@line 847 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAPF32[$279>>2] = $276; //@line 847 "./libraries/Box2D/Dynamics/b2Body.h"
    $280 = ((($215)) + 16|0); //@line 847 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAPF32[$280>>2] = $278; //@line 847 "./libraries/Box2D/Dynamics/b2Body.h"
    $281 = ((($217)) + 60|0); //@line 704 "./libraries/Box2D/Common/b2Math.h"
    $282 = +HEAPF32[$281>>2]; //@line 704 "./libraries/Box2D/Common/b2Math.h"
    $283 = $282 < 1.0; //@line 704 "./libraries/Box2D/Common/b2Math.h"
    __THREW__ = 0;
    invoke_vii(3822,($283|0),(361872|0)); //@line 704 "./libraries/Box2D/Common/b2Math.h"
    $284 = __THREW__; __THREW__ = 0;
    $285 = $284&1;
    if ($285) {
     label = 13;
     break;
    }
    $286 = +HEAPF32[$281>>2]; //@line 705 "./libraries/Box2D/Common/b2Math.h"
    $287 = $minAlpha$3 - $286; //@line 705 "./libraries/Box2D/Common/b2Math.h"
    $288 = 1.0 - $286; //@line 705 "./libraries/Box2D/Common/b2Math.h"
    $289 = $287 / $288; //@line 705 "./libraries/Box2D/Common/b2Math.h"
    $290 = ((($217)) + 44|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $291 = +HEAPF32[$290>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $292 = ((($217)) + 36|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $293 = +HEAPF32[$292>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $294 = $291 - $293; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $295 = ((($217)) + 48|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $296 = +HEAPF32[$295>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $297 = ((($217)) + 40|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $298 = +HEAPF32[$297>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $299 = $296 - $298; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $300 = $289 * $294; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $301 = $289 * $299; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $302 = $293 + $300; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$292>>2] = $302; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $303 = $298 + $301; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$297>>2] = $303; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $304 = ((($217)) + 56|0); //@line 707 "./libraries/Box2D/Common/b2Math.h"
    $305 = +HEAPF32[$304>>2]; //@line 707 "./libraries/Box2D/Common/b2Math.h"
    $306 = ((($217)) + 52|0); //@line 707 "./libraries/Box2D/Common/b2Math.h"
    $307 = +HEAPF32[$306>>2]; //@line 707 "./libraries/Box2D/Common/b2Math.h"
    $308 = $305 - $307; //@line 707 "./libraries/Box2D/Common/b2Math.h"
    $309 = $289 * $308; //@line 707 "./libraries/Box2D/Common/b2Math.h"
    $310 = $307 + $309; //@line 707 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$306>>2] = $310; //@line 707 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$281>>2] = $minAlpha$3; //@line 708 "./libraries/Box2D/Common/b2Math.h"
    $311 = ((($217)) + 44|0); //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $312 = ((($217)) + 36|0); //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $313 = $312; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $314 = $313; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $315 = HEAP32[$314>>2]|0; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $316 = (($313) + 4)|0; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $317 = $316; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $318 = HEAP32[$317>>2]|0; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $319 = $311; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $320 = $319; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAP32[$320>>2] = $315; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $321 = (($319) + 4)|0; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    $322 = $321; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAP32[$322>>2] = $318; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAPF32[$304>>2] = $310; //@line 845 "./libraries/Box2D/Dynamics/b2Body.h"
    $323 = (+Math_sin((+$310))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
    $324 = ((($217)) + 20|0); //@line 315 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$324>>2] = $323; //@line 315 "./libraries/Box2D/Common/b2Math.h"
    $325 = (+Math_cos((+$310))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
    $326 = ((($217)) + 24|0); //@line 316 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$326>>2] = $325; //@line 316 "./libraries/Box2D/Common/b2Math.h"
    $327 = ((($217)) + 28|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $328 = +HEAPF32[$327>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $329 = $325 * $328; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $330 = ((($217)) + 32|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $331 = +HEAPF32[$330>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $332 = $323 * $331; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $333 = $329 - $332; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $334 = $323 * $328; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $335 = $325 * $331; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $336 = $334 + $335; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $bitcast$i15 = (HEAP32[tempDoublePtr>>2]=$315,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $337 = $bitcast$i15 - $333; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $338 = (HEAP32[tempDoublePtr>>2]=$318,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $339 = $338 - $336; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $340 = ((($217)) + 12|0); //@line 847 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAPF32[$340>>2] = $337; //@line 847 "./libraries/Box2D/Dynamics/b2Body.h"
    $341 = ((($217)) + 16|0); //@line 847 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAPF32[$341>>2] = $339; //@line 847 "./libraries/Box2D/Dynamics/b2Body.h"
    $342 = HEAP32[$2>>2]|0; //@line 735 "libraries/Box2D/Dynamics/b2World.cpp"
    __THREW__ = 0;
    invoke_vii(3829,($minContact$3|0),($342|0)); //@line 735 "libraries/Box2D/Dynamics/b2World.cpp"
    $343 = __THREW__; __THREW__ = 0;
    $344 = $343&1;
    if ($344) {
     label = 13;
     break;
    }
    $345 = ((($minContact$3)) + 4|0); //@line 736 "libraries/Box2D/Dynamics/b2World.cpp"
    $346 = HEAP32[$345>>2]|0; //@line 736 "libraries/Box2D/Dynamics/b2World.cpp"
    $347 = $346 & -33; //@line 246 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
    HEAP32[$345>>2] = $347; //@line 736 "libraries/Box2D/Dynamics/b2World.cpp"
    $348 = ((($minContact$3)) + 128|0); //@line 737 "libraries/Box2D/Dynamics/b2World.cpp"
    $349 = HEAP32[$348>>2]|0; //@line 737 "libraries/Box2D/Dynamics/b2World.cpp"
    $350 = (($349) + 1)|0; //@line 737 "libraries/Box2D/Dynamics/b2World.cpp"
    HEAP32[$348>>2] = $350; //@line 737 "libraries/Box2D/Dynamics/b2World.cpp"
    $351 = $346 & 6; //@line 740 "libraries/Box2D/Dynamics/b2World.cpp"
    $352 = ($351|0)==(6); //@line 740 "libraries/Box2D/Dynamics/b2World.cpp"
    if ($352) {
     $385 = ((($215)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
     $386 = HEAP16[$385>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
     $387 = $386&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
     $388 = $387 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
     $389 = ($388|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
     if ($389) {
      $390 = $387 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
      $391 = $390&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
      HEAP16[$385>>1] = $391; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
      $392 = ((($215)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
      HEAPF32[$392>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
     }
     $393 = ((($217)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
     $394 = HEAP16[$393>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
     $395 = $394&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
     $396 = $395 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
     $397 = ($396|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
     if ($397) {
      $398 = $395 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
      $399 = $398&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
      HEAP16[$393>>1] = $399; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
      $400 = ((($217)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
      HEAPF32[$400>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
     }
     HEAP32[$19>>2] = 0; //@line 43 "./libraries/Box2D/Dynamics/b2Island.h"
     HEAP32[$20>>2] = 0; //@line 44 "./libraries/Box2D/Dynamics/b2Island.h"
     HEAP32[$21>>2] = 0; //@line 45 "./libraries/Box2D/Dynamics/b2Island.h"
     $401 = HEAP32[$22>>2]|0; //@line 54 "./libraries/Box2D/Dynamics/b2Island.h"
     $402 = ($401|0)>(0); //@line 54 "./libraries/Box2D/Dynamics/b2Island.h"
     __THREW__ = 0;
     invoke_vii(3822,($402|0),(361703|0)); //@line 54 "./libraries/Box2D/Dynamics/b2Island.h"
     $403 = __THREW__; __THREW__ = 0;
     $404 = $403&1;
     if ($404) {
      label = 13;
      break;
     }
     $405 = HEAP32[$19>>2]|0; //@line 55 "./libraries/Box2D/Dynamics/b2Island.h"
     $406 = ((($215)) + 8|0); //@line 55 "./libraries/Box2D/Dynamics/b2Island.h"
     HEAP32[$406>>2] = $405; //@line 55 "./libraries/Box2D/Dynamics/b2Island.h"
     $407 = HEAP32[$23>>2]|0; //@line 56 "./libraries/Box2D/Dynamics/b2Island.h"
     $408 = (($407) + ($405<<2)|0); //@line 56 "./libraries/Box2D/Dynamics/b2Island.h"
     HEAP32[$408>>2] = $215; //@line 56 "./libraries/Box2D/Dynamics/b2Island.h"
     $409 = (($405) + 1)|0; //@line 57 "./libraries/Box2D/Dynamics/b2Island.h"
     HEAP32[$19>>2] = $409; //@line 57 "./libraries/Box2D/Dynamics/b2Island.h"
     $410 = HEAP32[$22>>2]|0; //@line 54 "./libraries/Box2D/Dynamics/b2Island.h"
     $411 = ($409|0)<($410|0); //@line 54 "./libraries/Box2D/Dynamics/b2Island.h"
     __THREW__ = 0;
     invoke_vii(3822,($411|0),(361703|0)); //@line 54 "./libraries/Box2D/Dynamics/b2Island.h"
     $412 = __THREW__; __THREW__ = 0;
     $413 = $412&1;
     if ($413) {
      label = 13;
      break;
     }
     $414 = HEAP32[$19>>2]|0; //@line 55 "./libraries/Box2D/Dynamics/b2Island.h"
     $415 = ((($217)) + 8|0); //@line 55 "./libraries/Box2D/Dynamics/b2Island.h"
     HEAP32[$415>>2] = $414; //@line 55 "./libraries/Box2D/Dynamics/b2Island.h"
     $416 = HEAP32[$23>>2]|0; //@line 56 "./libraries/Box2D/Dynamics/b2Island.h"
     $417 = (($416) + ($414<<2)|0); //@line 56 "./libraries/Box2D/Dynamics/b2Island.h"
     HEAP32[$417>>2] = $217; //@line 56 "./libraries/Box2D/Dynamics/b2Island.h"
     $418 = (($414) + 1)|0; //@line 57 "./libraries/Box2D/Dynamics/b2Island.h"
     HEAP32[$19>>2] = $418; //@line 57 "./libraries/Box2D/Dynamics/b2Island.h"
     $419 = HEAP32[$20>>2]|0; //@line 62 "./libraries/Box2D/Dynamics/b2Island.h"
     $420 = HEAP32[$24>>2]|0; //@line 62 "./libraries/Box2D/Dynamics/b2Island.h"
     $421 = ($419|0)<($420|0); //@line 62 "./libraries/Box2D/Dynamics/b2Island.h"
     __THREW__ = 0;
     invoke_vii(3822,($421|0),(361732|0)); //@line 62 "./libraries/Box2D/Dynamics/b2Island.h"
     $422 = __THREW__; __THREW__ = 0;
     $423 = $422&1;
     if ($423) {
      label = 13;
      break;
     }
     $424 = HEAP32[$20>>2]|0; //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
     $425 = (($424) + 1)|0; //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
     HEAP32[$20>>2] = $425; //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
     $426 = HEAP32[$25>>2]|0; //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
     $427 = (($426) + ($424<<2)|0); //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
     HEAP32[$427>>2] = $minContact$3; //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
     $428 = HEAP16[$385>>1]|0; //@line 760 "libraries/Box2D/Dynamics/b2World.cpp"
     $429 = $428&65535; //@line 760 "libraries/Box2D/Dynamics/b2World.cpp"
     $430 = $429 | 1; //@line 760 "libraries/Box2D/Dynamics/b2World.cpp"
     $431 = $430&65535; //@line 760 "libraries/Box2D/Dynamics/b2World.cpp"
     HEAP16[$385>>1] = $431; //@line 760 "libraries/Box2D/Dynamics/b2World.cpp"
     $432 = HEAP16[$393>>1]|0; //@line 761 "libraries/Box2D/Dynamics/b2World.cpp"
     $433 = $432&65535; //@line 761 "libraries/Box2D/Dynamics/b2World.cpp"
     $434 = $433 | 1; //@line 761 "libraries/Box2D/Dynamics/b2World.cpp"
     $435 = $434&65535; //@line 761 "libraries/Box2D/Dynamics/b2World.cpp"
     HEAP16[$393>>1] = $435; //@line 761 "libraries/Box2D/Dynamics/b2World.cpp"
     $436 = HEAP32[$345>>2]|0; //@line 762 "libraries/Box2D/Dynamics/b2World.cpp"
     $437 = $436 | 1; //@line 762 "libraries/Box2D/Dynamics/b2World.cpp"
     HEAP32[$345>>2] = $437; //@line 762 "libraries/Box2D/Dynamics/b2World.cpp"
     HEAP32[$bodies>>2] = $215; //@line 765 "libraries/Box2D/Dynamics/b2World.cpp"
     HEAP32[$26>>2] = $217; //@line 765 "libraries/Box2D/Dynamics/b2World.cpp"
     $450 = $215;$i$062 = 0;
     while(1) {
      $449 = HEAP32[$450>>2]|0; //@line 769 "libraries/Box2D/Dynamics/b2World.cpp"
      $451 = ($449|0)==(2); //@line 769 "libraries/Box2D/Dynamics/b2World.cpp"
      L59: do {
       if ($451) {
        $452 = ((($450)) + 112|0); //@line 771 "libraries/Box2D/Dynamics/b2World.cpp"
        $ce$056 = HEAP32[$452>>2]|0; //@line 771 "libraries/Box2D/Dynamics/b2World.cpp"
        $453 = ($ce$056|0)==(0|0); //@line 771 "libraries/Box2D/Dynamics/b2World.cpp"
        if ($453) {
         break;
        }
        $454 = ((($450)) + 4|0); //@line 630 "./libraries/Box2D/Dynamics/b2Body.h"
        $ce$057 = $ce$056;
        while(1) {
         $455 = HEAP32[$19>>2]|0; //@line 773 "libraries/Box2D/Dynamics/b2World.cpp"
         $456 = HEAP32[$22>>2]|0; //@line 773 "libraries/Box2D/Dynamics/b2World.cpp"
         $457 = ($455|0)==($456|0); //@line 773 "libraries/Box2D/Dynamics/b2World.cpp"
         if ($457) {
          break L59;
         }
         $458 = HEAP32[$20>>2]|0; //@line 778 "libraries/Box2D/Dynamics/b2World.cpp"
         $459 = HEAP32[$24>>2]|0; //@line 778 "libraries/Box2D/Dynamics/b2World.cpp"
         $460 = ($458|0)==($459|0); //@line 778 "libraries/Box2D/Dynamics/b2World.cpp"
         if ($460) {
          break L59;
         }
         $461 = ((($ce$057)) + 4|0); //@line 783 "libraries/Box2D/Dynamics/b2World.cpp"
         $462 = HEAP32[$461>>2]|0; //@line 783 "libraries/Box2D/Dynamics/b2World.cpp"
         $463 = ((($462)) + 4|0); //@line 786 "libraries/Box2D/Dynamics/b2World.cpp"
         $464 = HEAP32[$463>>2]|0; //@line 786 "libraries/Box2D/Dynamics/b2World.cpp"
         $465 = $464 & 1; //@line 786 "libraries/Box2D/Dynamics/b2World.cpp"
         $466 = ($465|0)==(0); //@line 786 "libraries/Box2D/Dynamics/b2World.cpp"
         L66: do {
          if ($466) {
           $467 = HEAP32[$ce$057>>2]|0; //@line 792 "libraries/Box2D/Dynamics/b2World.cpp"
           $468 = HEAP32[$467>>2]|0; //@line 793 "libraries/Box2D/Dynamics/b2World.cpp"
           $469 = ($468|0)==(2); //@line 793 "libraries/Box2D/Dynamics/b2World.cpp"
           do {
            if ($469) {
             $470 = HEAP16[$454>>1]|0; //@line 630 "./libraries/Box2D/Dynamics/b2Body.h"
             $471 = $470 & 8; //@line 630 "./libraries/Box2D/Dynamics/b2Body.h"
             $472 = ($471<<16>>16)==(0); //@line 630 "./libraries/Box2D/Dynamics/b2Body.h"
             if (!($472)) {
              break;
             }
             $473 = ((($467)) + 4|0); //@line 630 "./libraries/Box2D/Dynamics/b2Body.h"
             $474 = HEAP16[$473>>1]|0; //@line 630 "./libraries/Box2D/Dynamics/b2Body.h"
             $475 = $474 & 8; //@line 630 "./libraries/Box2D/Dynamics/b2Body.h"
             $476 = ($475<<16>>16)==(0); //@line 630 "./libraries/Box2D/Dynamics/b2Body.h"
             if ($476) {
              break L66;
             }
            }
           } while(0);
           $477 = ((($462)) + 48|0); //@line 800 "libraries/Box2D/Dynamics/b2World.cpp"
           $478 = HEAP32[$477>>2]|0; //@line 800 "libraries/Box2D/Dynamics/b2World.cpp"
           $479 = ((($478)) + 38|0); //@line 800 "libraries/Box2D/Dynamics/b2World.cpp"
           $480 = HEAP8[$479>>0]|0; //@line 800 "libraries/Box2D/Dynamics/b2World.cpp"
           $481 = ($480<<24>>24)==(0); //@line 800 "libraries/Box2D/Dynamics/b2World.cpp"
           if (!($481)) {
            break;
           }
           $482 = ((($462)) + 52|0); //@line 801 "libraries/Box2D/Dynamics/b2World.cpp"
           $483 = HEAP32[$482>>2]|0; //@line 801 "libraries/Box2D/Dynamics/b2World.cpp"
           $484 = ((($483)) + 38|0); //@line 801 "libraries/Box2D/Dynamics/b2World.cpp"
           $485 = HEAP8[$484>>0]|0; //@line 801 "libraries/Box2D/Dynamics/b2World.cpp"
           $486 = ($485<<24>>24)==(0); //@line 801 "libraries/Box2D/Dynamics/b2World.cpp"
           if (!($486)) {
            break;
           }
           $487 = ((($467)) + 28|0); //@line 808 "libraries/Box2D/Dynamics/b2World.cpp"
           dest=$backup; src=$487; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 808 "libraries/Box2D/Dynamics/b2World.cpp"
           $488 = ((($467)) + 4|0); //@line 809 "libraries/Box2D/Dynamics/b2World.cpp"
           $489 = HEAP16[$488>>1]|0; //@line 809 "libraries/Box2D/Dynamics/b2World.cpp"
           $490 = $489 & 1; //@line 809 "libraries/Box2D/Dynamics/b2World.cpp"
           $491 = ($490<<16>>16)==(0); //@line 809 "libraries/Box2D/Dynamics/b2World.cpp"
           if ($491) {
            $492 = ((($467)) + 60|0); //@line 704 "./libraries/Box2D/Common/b2Math.h"
            $493 = +HEAPF32[$492>>2]; //@line 704 "./libraries/Box2D/Common/b2Math.h"
            $494 = $493 < 1.0; //@line 704 "./libraries/Box2D/Common/b2Math.h"
            __THREW__ = 0;
            invoke_vii(3822,($494|0),(361872|0)); //@line 704 "./libraries/Box2D/Common/b2Math.h"
            $495 = __THREW__; __THREW__ = 0;
            $496 = $495&1;
            if ($496) {
             label = 10;
             break L12;
            }
            $497 = +HEAPF32[$492>>2]; //@line 705 "./libraries/Box2D/Common/b2Math.h"
            $498 = $minAlpha$3 - $497; //@line 705 "./libraries/Box2D/Common/b2Math.h"
            $499 = 1.0 - $497; //@line 705 "./libraries/Box2D/Common/b2Math.h"
            $500 = $498 / $499; //@line 705 "./libraries/Box2D/Common/b2Math.h"
            $501 = ((($467)) + 44|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
            $502 = +HEAPF32[$501>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
            $503 = ((($467)) + 36|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
            $504 = +HEAPF32[$503>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
            $505 = $502 - $504; //@line 454 "./libraries/Box2D/Common/b2Math.h"
            $506 = ((($467)) + 48|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
            $507 = +HEAPF32[$506>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
            $508 = ((($467)) + 40|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
            $509 = +HEAPF32[$508>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
            $510 = $507 - $509; //@line 454 "./libraries/Box2D/Common/b2Math.h"
            $511 = $500 * $505; //@line 459 "./libraries/Box2D/Common/b2Math.h"
            $512 = $500 * $510; //@line 459 "./libraries/Box2D/Common/b2Math.h"
            $513 = $504 + $511; //@line 85 "./libraries/Box2D/Common/b2Math.h"
            HEAPF32[$503>>2] = $513; //@line 85 "./libraries/Box2D/Common/b2Math.h"
            $514 = $509 + $512; //@line 85 "./libraries/Box2D/Common/b2Math.h"
            HEAPF32[$508>>2] = $514; //@line 85 "./libraries/Box2D/Common/b2Math.h"
            $515 = ((($467)) + 56|0); //@line 707 "./libraries/Box2D/Common/b2Math.h"
            $516 = +HEAPF32[$515>>2]; //@line 707 "./libraries/Box2D/Common/b2Math.h"
            $517 = ((($467)) + 52|0); //@line 707 "./libraries/Box2D/Common/b2Math.h"
            $518 = +HEAPF32[$517>>2]; //@line 707 "./libraries/Box2D/Common/b2Math.h"
            $519 = $516 - $518; //@line 707 "./libraries/Box2D/Common/b2Math.h"
            $520 = $500 * $519; //@line 707 "./libraries/Box2D/Common/b2Math.h"
            $521 = $518 + $520; //@line 707 "./libraries/Box2D/Common/b2Math.h"
            HEAPF32[$517>>2] = $521; //@line 707 "./libraries/Box2D/Common/b2Math.h"
            HEAPF32[$492>>2] = $minAlpha$3; //@line 708 "./libraries/Box2D/Common/b2Math.h"
            $522 = ((($467)) + 44|0); //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
            $523 = ((($467)) + 36|0); //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
            $524 = $523; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
            $525 = $524; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
            $526 = HEAP32[$525>>2]|0; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
            $527 = (($524) + 4)|0; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
            $528 = $527; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
            $529 = HEAP32[$528>>2]|0; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
            $530 = $522; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
            $531 = $530; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
            HEAP32[$531>>2] = $526; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
            $532 = (($530) + 4)|0; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
            $533 = $532; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
            HEAP32[$533>>2] = $529; //@line 844 "./libraries/Box2D/Dynamics/b2Body.h"
            HEAPF32[$515>>2] = $521; //@line 845 "./libraries/Box2D/Dynamics/b2Body.h"
            $534 = (+Math_sin((+$521))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
            $535 = ((($467)) + 20|0); //@line 315 "./libraries/Box2D/Common/b2Math.h"
            HEAPF32[$535>>2] = $534; //@line 315 "./libraries/Box2D/Common/b2Math.h"
            $536 = (+Math_cos((+$521))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
            $537 = ((($467)) + 24|0); //@line 316 "./libraries/Box2D/Common/b2Math.h"
            HEAPF32[$537>>2] = $536; //@line 316 "./libraries/Box2D/Common/b2Math.h"
            $538 = ((($467)) + 28|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
            $539 = +HEAPF32[$538>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
            $540 = $536 * $539; //@line 568 "./libraries/Box2D/Common/b2Math.h"
            $541 = ((($467)) + 32|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
            $542 = +HEAPF32[$541>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
            $543 = $534 * $542; //@line 568 "./libraries/Box2D/Common/b2Math.h"
            $544 = $540 - $543; //@line 568 "./libraries/Box2D/Common/b2Math.h"
            $545 = $534 * $539; //@line 568 "./libraries/Box2D/Common/b2Math.h"
            $546 = $536 * $542; //@line 568 "./libraries/Box2D/Common/b2Math.h"
            $547 = $545 + $546; //@line 568 "./libraries/Box2D/Common/b2Math.h"
            $bitcast$i = (HEAP32[tempDoublePtr>>2]=$526,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
            $548 = $bitcast$i - $544; //@line 454 "./libraries/Box2D/Common/b2Math.h"
            $549 = (HEAP32[tempDoublePtr>>2]=$529,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
            $550 = $549 - $547; //@line 454 "./libraries/Box2D/Common/b2Math.h"
            $551 = ((($467)) + 12|0); //@line 847 "./libraries/Box2D/Dynamics/b2Body.h"
            HEAPF32[$551>>2] = $548; //@line 847 "./libraries/Box2D/Dynamics/b2Body.h"
            $552 = ((($467)) + 16|0); //@line 847 "./libraries/Box2D/Dynamics/b2Body.h"
            HEAPF32[$552>>2] = $550; //@line 847 "./libraries/Box2D/Dynamics/b2Body.h"
           }
           $553 = HEAP32[$2>>2]|0; //@line 815 "libraries/Box2D/Dynamics/b2World.cpp"
           __THREW__ = 0;
           invoke_vii(3829,($462|0),($553|0)); //@line 815 "libraries/Box2D/Dynamics/b2World.cpp"
           $554 = __THREW__; __THREW__ = 0;
           $555 = $554&1;
           if ($555) {
            label = 10;
            break L12;
           }
           $556 = HEAP32[$463>>2]|0; //@line 256 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
           $557 = $556 & 4; //@line 256 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
           $558 = ($557|0)==(0); //@line 256 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
           do {
            if ($558) {
             dest=$487; src=$backup; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 820 "libraries/Box2D/Dynamics/b2World.cpp"
             $559 = ((($467)) + 56|0); //@line 836 "./libraries/Box2D/Dynamics/b2Body.h"
             $560 = +HEAPF32[$559>>2]; //@line 836 "./libraries/Box2D/Dynamics/b2Body.h"
             $561 = (+Math_sin((+$560))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
             $562 = ((($467)) + 20|0); //@line 315 "./libraries/Box2D/Common/b2Math.h"
             HEAPF32[$562>>2] = $561; //@line 315 "./libraries/Box2D/Common/b2Math.h"
             $563 = (+Math_cos((+$560))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
             $564 = ((($467)) + 24|0); //@line 316 "./libraries/Box2D/Common/b2Math.h"
             HEAPF32[$564>>2] = $563; //@line 316 "./libraries/Box2D/Common/b2Math.h"
             $565 = ((($467)) + 28|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
             $566 = +HEAPF32[$565>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
             $567 = $563 * $566; //@line 568 "./libraries/Box2D/Common/b2Math.h"
             $568 = ((($467)) + 32|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
             $569 = +HEAPF32[$568>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
             $570 = $561 * $569; //@line 568 "./libraries/Box2D/Common/b2Math.h"
             $571 = $567 - $570; //@line 568 "./libraries/Box2D/Common/b2Math.h"
             $572 = $561 * $566; //@line 568 "./libraries/Box2D/Common/b2Math.h"
             $573 = $563 * $569; //@line 568 "./libraries/Box2D/Common/b2Math.h"
             $574 = $572 + $573; //@line 568 "./libraries/Box2D/Common/b2Math.h"
             $575 = ((($467)) + 44|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
             $576 = +HEAPF32[$575>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
             $577 = $576 - $571; //@line 454 "./libraries/Box2D/Common/b2Math.h"
             $578 = ((($467)) + 48|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
             $579 = +HEAPF32[$578>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
             $580 = $579 - $574; //@line 454 "./libraries/Box2D/Common/b2Math.h"
             $581 = ((($467)) + 12|0); //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
             HEAPF32[$581>>2] = $577; //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
             $582 = ((($467)) + 16|0); //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
             HEAPF32[$582>>2] = $580; //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
            } else {
             $583 = $556 & 2; //@line 261 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
             $584 = ($583|0)==(0); //@line 261 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
             if ($584) {
              dest=$487; src=$backup; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 828 "libraries/Box2D/Dynamics/b2World.cpp"
              $585 = ((($467)) + 56|0); //@line 836 "./libraries/Box2D/Dynamics/b2Body.h"
              $586 = +HEAPF32[$585>>2]; //@line 836 "./libraries/Box2D/Dynamics/b2Body.h"
              $587 = (+Math_sin((+$586))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
              $588 = ((($467)) + 20|0); //@line 315 "./libraries/Box2D/Common/b2Math.h"
              HEAPF32[$588>>2] = $587; //@line 315 "./libraries/Box2D/Common/b2Math.h"
              $589 = (+Math_cos((+$586))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
              $590 = ((($467)) + 24|0); //@line 316 "./libraries/Box2D/Common/b2Math.h"
              HEAPF32[$590>>2] = $589; //@line 316 "./libraries/Box2D/Common/b2Math.h"
              $591 = ((($467)) + 28|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
              $592 = +HEAPF32[$591>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
              $593 = $589 * $592; //@line 568 "./libraries/Box2D/Common/b2Math.h"
              $594 = ((($467)) + 32|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
              $595 = +HEAPF32[$594>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
              $596 = $587 * $595; //@line 568 "./libraries/Box2D/Common/b2Math.h"
              $597 = $593 - $596; //@line 568 "./libraries/Box2D/Common/b2Math.h"
              $598 = $587 * $592; //@line 568 "./libraries/Box2D/Common/b2Math.h"
              $599 = $589 * $595; //@line 568 "./libraries/Box2D/Common/b2Math.h"
              $600 = $598 + $599; //@line 568 "./libraries/Box2D/Common/b2Math.h"
              $601 = ((($467)) + 44|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
              $602 = +HEAPF32[$601>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
              $603 = $602 - $597; //@line 454 "./libraries/Box2D/Common/b2Math.h"
              $604 = ((($467)) + 48|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
              $605 = +HEAPF32[$604>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
              $606 = $605 - $600; //@line 454 "./libraries/Box2D/Common/b2Math.h"
              $607 = ((($467)) + 12|0); //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
              HEAPF32[$607>>2] = $603; //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
              $608 = ((($467)) + 16|0); //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
              HEAPF32[$608>>2] = $606; //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
              break;
             }
             $609 = $556 | 1; //@line 834 "libraries/Box2D/Dynamics/b2World.cpp"
             HEAP32[$463>>2] = $609; //@line 834 "libraries/Box2D/Dynamics/b2World.cpp"
             $610 = HEAP32[$20>>2]|0; //@line 62 "./libraries/Box2D/Dynamics/b2Island.h"
             $611 = HEAP32[$24>>2]|0; //@line 62 "./libraries/Box2D/Dynamics/b2Island.h"
             $612 = ($610|0)<($611|0); //@line 62 "./libraries/Box2D/Dynamics/b2Island.h"
             __THREW__ = 0;
             invoke_vii(3822,($612|0),(361732|0)); //@line 62 "./libraries/Box2D/Dynamics/b2Island.h"
             $613 = __THREW__; __THREW__ = 0;
             $614 = $613&1;
             if ($614) {
              label = 10;
              break L12;
             }
             $615 = HEAP32[$20>>2]|0; //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
             $616 = (($615) + 1)|0; //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
             HEAP32[$20>>2] = $616; //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
             $617 = HEAP32[$25>>2]|0; //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
             $618 = (($617) + ($615<<2)|0); //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
             HEAP32[$618>>2] = $462; //@line 63 "./libraries/Box2D/Dynamics/b2Island.h"
             $619 = HEAP16[$488>>1]|0; //@line 838 "libraries/Box2D/Dynamics/b2World.cpp"
             $620 = $619&65535; //@line 838 "libraries/Box2D/Dynamics/b2World.cpp"
             $621 = $620 & 1; //@line 838 "libraries/Box2D/Dynamics/b2World.cpp"
             $622 = ($621|0)==(0); //@line 838 "libraries/Box2D/Dynamics/b2World.cpp"
             if (!($622)) {
              break;
             }
             $623 = $620 | 1; //@line 844 "libraries/Box2D/Dynamics/b2World.cpp"
             $624 = $623&65535; //@line 844 "libraries/Box2D/Dynamics/b2World.cpp"
             HEAP16[$488>>1] = $624; //@line 844 "libraries/Box2D/Dynamics/b2World.cpp"
             $625 = HEAP32[$467>>2]|0; //@line 846 "libraries/Box2D/Dynamics/b2World.cpp"
             $626 = ($625|0)!=(0); //@line 846 "libraries/Box2D/Dynamics/b2World.cpp"
             $627 = $620 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
             $628 = ($627|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
             $or$cond28 = $628 & $626; //@line 846 "libraries/Box2D/Dynamics/b2World.cpp"
             if ($or$cond28) {
              $629 = $620 | 3; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
              $630 = $629&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
              HEAP16[$488>>1] = $630; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
              $631 = ((($467)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
              HEAPF32[$631>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
             }
             $632 = HEAP32[$19>>2]|0; //@line 54 "./libraries/Box2D/Dynamics/b2Island.h"
             $633 = HEAP32[$22>>2]|0; //@line 54 "./libraries/Box2D/Dynamics/b2Island.h"
             $634 = ($632|0)<($633|0); //@line 54 "./libraries/Box2D/Dynamics/b2Island.h"
             __THREW__ = 0;
             invoke_vii(3822,($634|0),(361703|0)); //@line 54 "./libraries/Box2D/Dynamics/b2Island.h"
             $635 = __THREW__; __THREW__ = 0;
             $636 = $635&1;
             if ($636) {
              label = 10;
              break L12;
             }
             $637 = HEAP32[$19>>2]|0; //@line 55 "./libraries/Box2D/Dynamics/b2Island.h"
             $638 = ((($467)) + 8|0); //@line 55 "./libraries/Box2D/Dynamics/b2Island.h"
             HEAP32[$638>>2] = $637; //@line 55 "./libraries/Box2D/Dynamics/b2Island.h"
             $639 = HEAP32[$23>>2]|0; //@line 56 "./libraries/Box2D/Dynamics/b2Island.h"
             $640 = (($639) + ($637<<2)|0); //@line 56 "./libraries/Box2D/Dynamics/b2Island.h"
             HEAP32[$640>>2] = $467; //@line 56 "./libraries/Box2D/Dynamics/b2Island.h"
             $641 = (($637) + 1)|0; //@line 57 "./libraries/Box2D/Dynamics/b2Island.h"
             HEAP32[$19>>2] = $641; //@line 57 "./libraries/Box2D/Dynamics/b2Island.h"
            }
           } while(0);
          }
         } while(0);
         $642 = ((($ce$057)) + 12|0); //@line 771 "libraries/Box2D/Dynamics/b2World.cpp"
         $ce$0 = HEAP32[$642>>2]|0; //@line 771 "libraries/Box2D/Dynamics/b2World.cpp"
         $643 = ($ce$0|0)==(0|0); //@line 771 "libraries/Box2D/Dynamics/b2World.cpp"
         if ($643) {
          break;
         } else {
          $ce$057 = $ce$0;
         }
        }
       }
      } while(0);
      $644 = (($i$062) + 1)|0; //@line 766 "libraries/Box2D/Dynamics/b2World.cpp"
      $645 = ($644|0)<(2); //@line 766 "libraries/Box2D/Dynamics/b2World.cpp"
      if (!($645)) {
       break;
      }
      $$phi$trans$insert = (($bodies) + ($644<<2)|0); //@line 768 "libraries/Box2D/Dynamics/b2World.cpp"
      $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 768 "libraries/Box2D/Dynamics/b2World.cpp"
      $450 = $$pre;$i$062 = $644;
     }
     $438 = 1.0 - $minAlpha$3; //@line 857 "libraries/Box2D/Dynamics/b2World.cpp"
     $439 = +HEAPF32[$step>>2]; //@line 857 "libraries/Box2D/Dynamics/b2World.cpp"
     $440 = $438 * $439; //@line 857 "libraries/Box2D/Dynamics/b2World.cpp"
     HEAPF32[$subStep>>2] = $440; //@line 857 "libraries/Box2D/Dynamics/b2World.cpp"
     $441 = 1.0 / $440; //@line 858 "libraries/Box2D/Dynamics/b2World.cpp"
     HEAPF32[$27>>2] = $441; //@line 858 "libraries/Box2D/Dynamics/b2World.cpp"
     HEAPF32[$28>>2] = 1.0; //@line 859 "libraries/Box2D/Dynamics/b2World.cpp"
     HEAP32[$29>>2] = 20; //@line 860 "libraries/Box2D/Dynamics/b2World.cpp"
     $442 = HEAP32[$30>>2]|0; //@line 861 "libraries/Box2D/Dynamics/b2World.cpp"
     HEAP32[$31>>2] = $442; //@line 861 "libraries/Box2D/Dynamics/b2World.cpp"
     HEAP8[$32>>0] = 0; //@line 862 "libraries/Box2D/Dynamics/b2World.cpp"
     $443 = HEAP32[$406>>2]|0; //@line 863 "libraries/Box2D/Dynamics/b2World.cpp"
     $444 = HEAP32[$415>>2]|0; //@line 863 "libraries/Box2D/Dynamics/b2World.cpp"
     __THREW__ = 0;
     invoke_viiii(3830,($island|0),($subStep|0),($443|0),($444|0)); //@line 863 "libraries/Box2D/Dynamics/b2World.cpp"
     $445 = __THREW__; __THREW__ = 0;
     $446 = $445&1;
     if ($446) {
      label = 13;
      break;
     }
     $447 = HEAP32[$19>>2]|0; //@line 866 "libraries/Box2D/Dynamics/b2World.cpp"
     $448 = ($447|0)>(0); //@line 866 "libraries/Box2D/Dynamics/b2World.cpp"
     if ($448) {
      $i8$067 = 0;
      while(1) {
       $648 = HEAP32[$23>>2]|0; //@line 868 "libraries/Box2D/Dynamics/b2World.cpp"
       $649 = (($648) + ($i8$067<<2)|0); //@line 868 "libraries/Box2D/Dynamics/b2World.cpp"
       $650 = HEAP32[$649>>2]|0; //@line 868 "libraries/Box2D/Dynamics/b2World.cpp"
       $651 = ((($650)) + 4|0); //@line 869 "libraries/Box2D/Dynamics/b2World.cpp"
       $652 = HEAP16[$651>>1]|0; //@line 869 "libraries/Box2D/Dynamics/b2World.cpp"
       $653 = $652&65535; //@line 869 "libraries/Box2D/Dynamics/b2World.cpp"
       $654 = $653 & 65534; //@line 869 "libraries/Box2D/Dynamics/b2World.cpp"
       $655 = $654&65535; //@line 869 "libraries/Box2D/Dynamics/b2World.cpp"
       HEAP16[$651>>1] = $655; //@line 869 "libraries/Box2D/Dynamics/b2World.cpp"
       $656 = HEAP32[$650>>2]|0; //@line 871 "libraries/Box2D/Dynamics/b2World.cpp"
       $657 = ($656|0)==(2); //@line 871 "libraries/Box2D/Dynamics/b2World.cpp"
       do {
        if ($657) {
         __THREW__ = 0;
         invoke_vi(3826,($650|0)); //@line 876 "libraries/Box2D/Dynamics/b2World.cpp"
         $658 = __THREW__; __THREW__ = 0;
         $659 = $658&1;
         if ($659) {
          label = 11;
          break L12;
         }
         $660 = ((($650)) + 112|0); //@line 879 "libraries/Box2D/Dynamics/b2World.cpp"
         $ce10$063 = HEAP32[$660>>2]|0; //@line 879 "libraries/Box2D/Dynamics/b2World.cpp"
         $661 = ($ce10$063|0)==(0|0); //@line 879 "libraries/Box2D/Dynamics/b2World.cpp"
         if ($661) {
          break;
         } else {
          $ce10$064 = $ce10$063;
         }
         while(1) {
          $662 = ((($ce10$064)) + 4|0); //@line 881 "libraries/Box2D/Dynamics/b2World.cpp"
          $663 = HEAP32[$662>>2]|0; //@line 881 "libraries/Box2D/Dynamics/b2World.cpp"
          $664 = ((($663)) + 4|0); //@line 881 "libraries/Box2D/Dynamics/b2World.cpp"
          $665 = HEAP32[$664>>2]|0; //@line 881 "libraries/Box2D/Dynamics/b2World.cpp"
          $666 = $665 & -34; //@line 881 "libraries/Box2D/Dynamics/b2World.cpp"
          HEAP32[$664>>2] = $666; //@line 881 "libraries/Box2D/Dynamics/b2World.cpp"
          $667 = ((($ce10$064)) + 12|0); //@line 879 "libraries/Box2D/Dynamics/b2World.cpp"
          $ce10$0 = HEAP32[$667>>2]|0; //@line 879 "libraries/Box2D/Dynamics/b2World.cpp"
          $668 = ($ce10$0|0)==(0|0); //@line 879 "libraries/Box2D/Dynamics/b2World.cpp"
          if ($668) {
           break;
          } else {
           $ce10$064 = $ce10$0;
          }
         }
        }
       } while(0);
       $669 = (($i8$067) + 1)|0; //@line 866 "libraries/Box2D/Dynamics/b2World.cpp"
       $670 = HEAP32[$19>>2]|0; //@line 866 "libraries/Box2D/Dynamics/b2World.cpp"
       $671 = ($669|0)<($670|0); //@line 866 "libraries/Box2D/Dynamics/b2World.cpp"
       if ($671) {
        $i8$067 = $669;
       } else {
        break;
       }
      }
     }
     __THREW__ = 0;
     invoke_vi(3825,($1|0)); //@line 887 "libraries/Box2D/Dynamics/b2World.cpp"
     $646 = __THREW__; __THREW__ = 0;
     $647 = $646&1;
     if ($647) {
      label = 13;
      break;
     }
     $672 = HEAP8[$33>>0]|0; //@line 889 "libraries/Box2D/Dynamics/b2World.cpp"
     $673 = ($672<<24>>24)==(0); //@line 889 "libraries/Box2D/Dynamics/b2World.cpp"
     if (!($673)) {
      label = 91;
      break;
     }
    } else {
     $353 = $346 & -37; //@line 250 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     HEAP32[$345>>2] = $353; //@line 246 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     dest=$218; src=$backup1; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 744 "libraries/Box2D/Dynamics/b2World.cpp"
     dest=$219; src=$backup2; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 745 "libraries/Box2D/Dynamics/b2World.cpp"
     $354 = +HEAPF32[$243>>2]; //@line 836 "./libraries/Box2D/Dynamics/b2Body.h"
     $355 = (+Math_sin((+$354))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
     HEAPF32[$263>>2] = $355; //@line 315 "./libraries/Box2D/Common/b2Math.h"
     $356 = (+Math_cos((+$354))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
     HEAPF32[$265>>2] = $356; //@line 316 "./libraries/Box2D/Common/b2Math.h"
     $357 = +HEAPF32[$266>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
     $358 = $356 * $357; //@line 568 "./libraries/Box2D/Common/b2Math.h"
     $359 = +HEAPF32[$269>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
     $360 = $355 * $359; //@line 568 "./libraries/Box2D/Common/b2Math.h"
     $361 = $358 - $360; //@line 568 "./libraries/Box2D/Common/b2Math.h"
     $362 = $355 * $357; //@line 568 "./libraries/Box2D/Common/b2Math.h"
     $363 = $356 * $359; //@line 568 "./libraries/Box2D/Common/b2Math.h"
     $364 = $362 + $363; //@line 568 "./libraries/Box2D/Common/b2Math.h"
     $365 = +HEAPF32[$229>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $366 = $365 - $361; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $367 = +HEAPF32[$234>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $368 = $367 - $364; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     HEAPF32[$279>>2] = $366; //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
     HEAPF32[$280>>2] = $368; //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
     $369 = +HEAPF32[$304>>2]; //@line 836 "./libraries/Box2D/Dynamics/b2Body.h"
     $370 = (+Math_sin((+$369))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
     HEAPF32[$324>>2] = $370; //@line 315 "./libraries/Box2D/Common/b2Math.h"
     $371 = (+Math_cos((+$369))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
     HEAPF32[$326>>2] = $371; //@line 316 "./libraries/Box2D/Common/b2Math.h"
     $372 = +HEAPF32[$327>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
     $373 = $371 * $372; //@line 568 "./libraries/Box2D/Common/b2Math.h"
     $374 = +HEAPF32[$330>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
     $375 = $370 * $374; //@line 568 "./libraries/Box2D/Common/b2Math.h"
     $376 = $373 - $375; //@line 568 "./libraries/Box2D/Common/b2Math.h"
     $377 = $370 * $372; //@line 568 "./libraries/Box2D/Common/b2Math.h"
     $378 = $371 * $374; //@line 568 "./libraries/Box2D/Common/b2Math.h"
     $379 = $377 + $378; //@line 568 "./libraries/Box2D/Common/b2Math.h"
     $380 = +HEAPF32[$290>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $381 = $380 - $376; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $382 = +HEAPF32[$295>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $383 = $382 - $379; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     HEAPF32[$340>>2] = $381; //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
     HEAPF32[$341>>2] = $383; //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
    }
    $c1$049 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 605 "libraries/Box2D/Dynamics/b2World.cpp"
    $384 = ($c1$049|0)==(0|0); //@line 605 "libraries/Box2D/Dynamics/b2World.cpp"
    if ($384) {
     break L11;
    } else {
     $c1$052 = $c1$049;$minAlpha$050 = 1.0;$minContact$051 = 0;
    }
   }
   if ((label|0) == 10) {
    $59 = ___cxa_find_matching_catch_2()|0;
    $60 = tempRet0;
    $lpad$phi$index = $59;$lpad$phi$index4 = $60;
    __ZN8b2IslandD2Ev($island); //@line 895 "libraries/Box2D/Dynamics/b2World.cpp"
    ___resumeException($lpad$phi$index|0);
    // unreachable;
   }
   else if ((label|0) == 11) {
    $61 = ___cxa_find_matching_catch_2()|0;
    $62 = tempRet0;
    $lpad$phi$index = $61;$lpad$phi$index4 = $62;
    __ZN8b2IslandD2Ev($island); //@line 895 "libraries/Box2D/Dynamics/b2World.cpp"
    ___resumeException($lpad$phi$index|0);
    // unreachable;
   }
   else if ((label|0) == 12) {
    $63 = ___cxa_find_matching_catch_2()|0;
    $64 = tempRet0;
    $lpad$phi$index = $63;$lpad$phi$index4 = $64;
    __ZN8b2IslandD2Ev($island); //@line 895 "libraries/Box2D/Dynamics/b2World.cpp"
    ___resumeException($lpad$phi$index|0);
    // unreachable;
   }
   else if ((label|0) == 13) {
    $65 = ___cxa_find_matching_catch_2()|0;
    $66 = tempRet0;
    $lpad$phi$index = $65;$lpad$phi$index4 = $66;
    __ZN8b2IslandD2Ev($island); //@line 895 "libraries/Box2D/Dynamics/b2World.cpp"
    ___resumeException($lpad$phi$index|0);
    // unreachable;
   }
   else if ((label|0) == 91) {
    HEAP8[$4>>0] = 0; //@line 891 "libraries/Box2D/Dynamics/b2World.cpp"
    __ZN8b2IslandD2Ev($island); //@line 895 "libraries/Box2D/Dynamics/b2World.cpp"
    STACKTOP = sp;return; //@line 895 "libraries/Box2D/Dynamics/b2World.cpp"
   }
  }
 } while(0);
 HEAP8[$4>>0] = 1; //@line 718 "libraries/Box2D/Dynamics/b2World.cpp"
 __ZN8b2IslandD2Ev($island); //@line 895 "libraries/Box2D/Dynamics/b2World.cpp"
 STACKTOP = sp;return; //@line 895 "libraries/Box2D/Dynamics/b2World.cpp"
}
function __ZN7b2World4StepEfii($this,$dt,$velocityIterations,$positionIterations) {
 $this = $this|0;
 $dt = +$dt;
 $velocityIterations = $velocityIterations|0;
 $positionIterations = $positionIterations|0;
 var $$pr = 0.0, $$pre = 0.0, $$sink = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0.0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0.0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $body$0$i = 0, $body$01$i = 0, $body$02$i = 0, $or$cond = 0, $or$cond3 = 0;
 var $step = 0, $stepTimer = 0, $timer = 0, $timer1 = 0, $timer2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $stepTimer = sp + 27|0;
 $step = sp;
 $timer = sp + 26|0;
 $timer1 = sp + 25|0;
 $timer2 = sp + 24|0;
 __ZN7b2TimerC2Ev($stepTimer); //@line 899 "libraries/Box2D/Dynamics/b2World.cpp"
 $0 = ((($this)) + 102868|0); //@line 902 "libraries/Box2D/Dynamics/b2World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 902 "libraries/Box2D/Dynamics/b2World.cpp"
 $2 = $1 & 1; //@line 902 "libraries/Box2D/Dynamics/b2World.cpp"
 $3 = ($2|0)==(0); //@line 902 "libraries/Box2D/Dynamics/b2World.cpp"
 if ($3) {
  $8 = $1;
 } else {
  $4 = ((($this)) + 102872|0); //@line 904 "libraries/Box2D/Dynamics/b2World.cpp"
  __ZN16b2ContactManager15FindNewContactsEv($4); //@line 904 "libraries/Box2D/Dynamics/b2World.cpp"
  $5 = HEAP32[$0>>2]|0; //@line 905 "libraries/Box2D/Dynamics/b2World.cpp"
  $6 = $5 & -2; //@line 905 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$0>>2] = $6; //@line 905 "libraries/Box2D/Dynamics/b2World.cpp"
  $8 = $6;
 }
 $7 = $8 | 2; //@line 908 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$0>>2] = $7; //@line 908 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAPF32[$step>>2] = $dt; //@line 911 "libraries/Box2D/Dynamics/b2World.cpp"
 $9 = ((($step)) + 12|0); //@line 912 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$9>>2] = $velocityIterations; //@line 912 "libraries/Box2D/Dynamics/b2World.cpp"
 $10 = ((($step)) + 16|0); //@line 913 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$10>>2] = $positionIterations; //@line 913 "libraries/Box2D/Dynamics/b2World.cpp"
 $11 = $dt > 0.0; //@line 914 "libraries/Box2D/Dynamics/b2World.cpp"
 $12 = 1.0 / $dt; //@line 916 "libraries/Box2D/Dynamics/b2World.cpp"
 $$sink = $11 ? $12 : 0.0; //@line 914 "libraries/Box2D/Dynamics/b2World.cpp"
 $13 = ((($step)) + 4|0); //@line 920 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAPF32[$13>>2] = $$sink; //@line 920 "libraries/Box2D/Dynamics/b2World.cpp"
 $14 = ((($this)) + 102988|0); //@line 923 "libraries/Box2D/Dynamics/b2World.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 923 "libraries/Box2D/Dynamics/b2World.cpp"
 $16 = $15 * $dt; //@line 923 "libraries/Box2D/Dynamics/b2World.cpp"
 $17 = ((($step)) + 8|0); //@line 923 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAPF32[$17>>2] = $16; //@line 923 "libraries/Box2D/Dynamics/b2World.cpp"
 $18 = ((($this)) + 102992|0); //@line 925 "libraries/Box2D/Dynamics/b2World.cpp"
 $19 = HEAP8[$18>>0]|0; //@line 925 "libraries/Box2D/Dynamics/b2World.cpp"
 $20 = ((($step)) + 20|0); //@line 925 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP8[$20>>0] = $19; //@line 925 "libraries/Box2D/Dynamics/b2World.cpp"
 __ZN7b2TimerC2Ev($timer); //@line 929 "libraries/Box2D/Dynamics/b2World.cpp"
 $21 = ((($this)) + 102872|0); //@line 930 "libraries/Box2D/Dynamics/b2World.cpp"
 __ZN16b2ContactManager7CollideEv($21); //@line 930 "libraries/Box2D/Dynamics/b2World.cpp"
 $22 = (+__ZNK7b2Timer15GetMillisecondsEv($timer)); //@line 931 "libraries/Box2D/Dynamics/b2World.cpp"
 $23 = ((($this)) + 103000|0); //@line 931 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAPF32[$23>>2] = $22; //@line 931 "libraries/Box2D/Dynamics/b2World.cpp"
 $24 = ((($this)) + 102995|0); //@line 935 "libraries/Box2D/Dynamics/b2World.cpp"
 $25 = HEAP8[$24>>0]|0; //@line 935 "libraries/Box2D/Dynamics/b2World.cpp"
 $26 = ($25<<24>>24)!=(0); //@line 935 "libraries/Box2D/Dynamics/b2World.cpp"
 $or$cond = $26 & $11; //@line 935 "libraries/Box2D/Dynamics/b2World.cpp"
 if ($or$cond) {
  __ZN7b2TimerC2Ev($timer1); //@line 937 "libraries/Box2D/Dynamics/b2World.cpp"
  __ZN7b2World5SolveERK10b2TimeStep($this,$step); //@line 938 "libraries/Box2D/Dynamics/b2World.cpp"
  $27 = (+__ZNK7b2Timer15GetMillisecondsEv($timer1)); //@line 939 "libraries/Box2D/Dynamics/b2World.cpp"
  $28 = ((($this)) + 103004|0); //@line 939 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAPF32[$28>>2] = $27; //@line 939 "libraries/Box2D/Dynamics/b2World.cpp"
  $$pre = +HEAPF32[$step>>2]; //@line 943 "libraries/Box2D/Dynamics/b2World.cpp"
  $33 = $$pre;
 } else {
  $33 = $dt;
 }
 $29 = ((($this)) + 102993|0); //@line 943 "libraries/Box2D/Dynamics/b2World.cpp"
 $30 = HEAP8[$29>>0]|0; //@line 943 "libraries/Box2D/Dynamics/b2World.cpp"
 $31 = ($30<<24>>24)!=(0); //@line 943 "libraries/Box2D/Dynamics/b2World.cpp"
 $32 = $33 > 0.0; //@line 943 "libraries/Box2D/Dynamics/b2World.cpp"
 $or$cond3 = $31 & $32; //@line 943 "libraries/Box2D/Dynamics/b2World.cpp"
 if ($or$cond3) {
  __ZN7b2TimerC2Ev($timer2); //@line 945 "libraries/Box2D/Dynamics/b2World.cpp"
  __ZN7b2World8SolveTOIERK10b2TimeStep($this,$step); //@line 946 "libraries/Box2D/Dynamics/b2World.cpp"
  $34 = (+__ZNK7b2Timer15GetMillisecondsEv($timer2)); //@line 947 "libraries/Box2D/Dynamics/b2World.cpp"
  $35 = ((($this)) + 103024|0); //@line 947 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAPF32[$35>>2] = $34; //@line 947 "libraries/Box2D/Dynamics/b2World.cpp"
  $$pr = +HEAPF32[$step>>2]; //@line 950 "libraries/Box2D/Dynamics/b2World.cpp"
  $37 = $$pr;
 } else {
  $37 = $33;
 }
 $36 = $37 > 0.0; //@line 950 "libraries/Box2D/Dynamics/b2World.cpp"
 if ($36) {
  $38 = HEAP32[$13>>2]|0; //@line 952 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$14>>2] = $38; //@line 952 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $39 = HEAP32[$0>>2]|0; //@line 960 "libraries/Box2D/Dynamics/b2World.cpp"
 $40 = $39 & 4; //@line 955 "libraries/Box2D/Dynamics/b2World.cpp"
 $41 = ($40|0)==(0); //@line 955 "libraries/Box2D/Dynamics/b2World.cpp"
 if ($41) {
  $49 = $39 & -3; //@line 960 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$0>>2] = $49; //@line 960 "libraries/Box2D/Dynamics/b2World.cpp"
  $50 = (+__ZNK7b2Timer15GetMillisecondsEv($stepTimer)); //@line 962 "libraries/Box2D/Dynamics/b2World.cpp"
  $51 = ((($this)) + 102996|0); //@line 962 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAPF32[$51>>2] = $50; //@line 962 "libraries/Box2D/Dynamics/b2World.cpp"
  STACKTOP = sp;return; //@line 963 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $42 = ((($this)) + 102952|0); //@line 967 "libraries/Box2D/Dynamics/b2World.cpp"
 $body$01$i = HEAP32[$42>>2]|0; //@line 967 "libraries/Box2D/Dynamics/b2World.cpp"
 $43 = ($body$01$i|0)==(0|0); //@line 967 "libraries/Box2D/Dynamics/b2World.cpp"
 if ($43) {
  $49 = $39 & -3; //@line 960 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAP32[$0>>2] = $49; //@line 960 "libraries/Box2D/Dynamics/b2World.cpp"
  $50 = (+__ZNK7b2Timer15GetMillisecondsEv($stepTimer)); //@line 962 "libraries/Box2D/Dynamics/b2World.cpp"
  $51 = ((($this)) + 102996|0); //@line 962 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAPF32[$51>>2] = $50; //@line 962 "libraries/Box2D/Dynamics/b2World.cpp"
  STACKTOP = sp;return; //@line 963 "libraries/Box2D/Dynamics/b2World.cpp"
 } else {
  $body$02$i = $body$01$i;
 }
 while(1) {
  $44 = ((($body$02$i)) + 76|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$44>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $45 = ((($body$02$i)) + 80|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$45>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $46 = ((($body$02$i)) + 84|0); //@line 970 "libraries/Box2D/Dynamics/b2World.cpp"
  HEAPF32[$46>>2] = 0.0; //@line 970 "libraries/Box2D/Dynamics/b2World.cpp"
  $47 = ((($body$02$i)) + 96|0); //@line 719 "./libraries/Box2D/Dynamics/b2Body.h"
  $body$0$i = HEAP32[$47>>2]|0; //@line 967 "libraries/Box2D/Dynamics/b2World.cpp"
  $48 = ($body$0$i|0)==(0|0); //@line 967 "libraries/Box2D/Dynamics/b2World.cpp"
  if ($48) {
   break;
  } else {
   $body$02$i = $body$0$i;
  }
 }
 $49 = $39 & -3; //@line 960 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$0>>2] = $49; //@line 960 "libraries/Box2D/Dynamics/b2World.cpp"
 $50 = (+__ZNK7b2Timer15GetMillisecondsEv($stepTimer)); //@line 962 "libraries/Box2D/Dynamics/b2World.cpp"
 $51 = ((($this)) + 102996|0); //@line 962 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAPF32[$51>>2] = $50; //@line 962 "libraries/Box2D/Dynamics/b2World.cpp"
 STACKTOP = sp;return; //@line 963 "libraries/Box2D/Dynamics/b2World.cpp"
}
function __ZNK7b2World9QueryAABBEP15b2QueryCallbackRK6b2AABB($this,$callback,$aabb) {
 $this = $this|0;
 $callback = $callback|0;
 $aabb = $aabb|0;
 var $0 = 0, $1 = 0, $wrapper = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $wrapper = sp;
 $0 = ((($this)) + 102872|0); //@line 989 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$wrapper>>2] = $0; //@line 989 "libraries/Box2D/Dynamics/b2World.cpp"
 $1 = ((($wrapper)) + 4|0); //@line 990 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$1>>2] = $callback; //@line 990 "libraries/Box2D/Dynamics/b2World.cpp"
 __ZNK13b2DynamicTree5QueryI19b2WorldQueryWrapperEEvPT_RK6b2AABB($0,$wrapper,$aabb); //@line 243 "./libraries/Box2D/Collision/b2BroadPhase.h"
 STACKTOP = sp;return; //@line 992 "libraries/Box2D/Dynamics/b2World.cpp"
}
function __ZNK13b2DynamicTree5QueryI19b2WorldQueryWrapperEEvPT_RK6b2AABB($this,$callback,$aabb) {
 $this = $this|0;
 $callback = $callback|0;
 $aabb = $aabb|0;
 var $$old = 0, $$old12 = 0, $$pr = 0, $$pre = 0, $$pre$pre = 0, $$pre15 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0.0, $27 = 0;
 var $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0;
 var $or$cond$i = 0, $or$cond10$i = 0, $or$cond8$i = 0, $stack = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1040|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $stack = sp;
 $0 = ((($stack)) + 4|0); //@line 33 "./libraries/Box2D/Common/b2GrowableStack.h"
 HEAP32[$stack>>2] = $0; //@line 33 "./libraries/Box2D/Common/b2GrowableStack.h"
 $1 = ((($stack)) + 1028|0); //@line 34 "./libraries/Box2D/Common/b2GrowableStack.h"
 $2 = ((($stack)) + 1032|0); //@line 35 "./libraries/Box2D/Common/b2GrowableStack.h"
 HEAP32[$2>>2] = 256; //@line 35 "./libraries/Box2D/Common/b2GrowableStack.h"
 $3 = HEAP32[$this>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
 HEAP32[$0>>2] = $3; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
 HEAP32[$1>>2] = 1; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
 $4 = ((($this)) + 4|0); //@line 182 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $5 = ((($aabb)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($aabb)) + 8|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($aabb)) + 12|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($callback)) + 4|0); //@line 979 "libraries/Box2D/Dynamics/b2World.cpp"
 while(1) {
  __THREW__ = 0;
  invoke_vii(3822,1,(446492|0)); //@line 67 "./libraries/Box2D/Common/b2GrowableStack.h"
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if ($10) {
   label = 5;
   break;
  }
  $11 = HEAP32[$1>>2]|0; //@line 68 "./libraries/Box2D/Common/b2GrowableStack.h"
  $12 = (($11) + -1)|0; //@line 68 "./libraries/Box2D/Common/b2GrowableStack.h"
  HEAP32[$1>>2] = $12; //@line 68 "./libraries/Box2D/Common/b2GrowableStack.h"
  $13 = HEAP32[$stack>>2]|0; //@line 69 "./libraries/Box2D/Common/b2GrowableStack.h"
  $14 = (($13) + ($12<<2)|0); //@line 69 "./libraries/Box2D/Common/b2GrowableStack.h"
  $15 = HEAP32[$14>>2]|0; //@line 69 "./libraries/Box2D/Common/b2GrowableStack.h"
  $16 = ($15|0)==(-1); //@line 177 "./libraries/Box2D/Collision/b2DynamicTree.h"
  if ($16) {
   $$old = $12;$$pre15 = $13;
  } else {
   $25 = HEAP32[$4>>2]|0; //@line 182 "./libraries/Box2D/Collision/b2DynamicTree.h"
   $26 = +HEAPF32[$aabb>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $27 = (((($25) + (($15*36)|0)|0)) + 8|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $28 = +HEAPF32[$27>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $29 = $26 - $28; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $30 = +HEAPF32[$5>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $31 = (((($25) + (($15*36)|0)|0)) + 12|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $32 = +HEAPF32[$31>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $33 = $30 - $32; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $34 = (($25) + (($15*36)|0)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $35 = +HEAPF32[$34>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $36 = +HEAPF32[$6>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $37 = $35 - $36; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $38 = (((($25) + (($15*36)|0)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $39 = +HEAPF32[$38>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $40 = +HEAPF32[$7>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $41 = $39 - $40; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $42 = $29 > 0.0; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
   $43 = $33 > 0.0; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
   $or$cond$i = $42 | $43; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
   $44 = $37 > 0.0; //@line 271 "./libraries/Box2D/Collision/b2Collision.h"
   $or$cond8$i = $or$cond$i | $44; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
   $45 = $41 > 0.0; //@line 271 "./libraries/Box2D/Collision/b2Collision.h"
   $or$cond10$i = $or$cond8$i | $45; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
   if ($or$cond10$i) {
    $$old = $12;$$pre15 = $13;
   } else {
    $46 = (((($25) + (($15*36)|0)|0)) + 24|0); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $47 = HEAP32[$46>>2]|0; //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $48 = ($47|0)==(-1); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
    if ($48) {
     $49 = HEAP32[$callback>>2]|0; //@line 978 "libraries/Box2D/Dynamics/b2World.cpp"
     $50 = ($15|0)>(-1); //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
     if ($50) {
      $51 = ((($49)) + 12|0); //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $52 = HEAP32[$51>>2]|0; //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $53 = ($52|0)>($15|0); //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $54 = $53;
     } else {
      $54 = 0;
     }
     __THREW__ = 0;
     invoke_vii(3822,($54|0),(446451|0)); //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
     $55 = __THREW__; __THREW__ = 0;
     $56 = $55&1;
     if ($56) {
      label = 5;
      break;
     }
     $57 = ((($49)) + 4|0); //@line 159 "./libraries/Box2D/Collision/b2DynamicTree.h"
     $58 = HEAP32[$57>>2]|0; //@line 159 "./libraries/Box2D/Collision/b2DynamicTree.h"
     $59 = (((($58) + (($15*36)|0)|0)) + 16|0); //@line 159 "./libraries/Box2D/Collision/b2DynamicTree.h"
     $60 = HEAP32[$59>>2]|0; //@line 159 "./libraries/Box2D/Collision/b2DynamicTree.h"
     $61 = HEAP32[$8>>2]|0; //@line 979 "libraries/Box2D/Dynamics/b2World.cpp"
     $62 = HEAP32[$61>>2]|0; //@line 979 "libraries/Box2D/Dynamics/b2World.cpp"
     $63 = ((($62)) + 8|0); //@line 979 "libraries/Box2D/Dynamics/b2World.cpp"
     $64 = HEAP32[$63>>2]|0; //@line 979 "libraries/Box2D/Dynamics/b2World.cpp"
     $65 = ((($60)) + 16|0); //@line 979 "libraries/Box2D/Dynamics/b2World.cpp"
     $66 = HEAP32[$65>>2]|0; //@line 979 "libraries/Box2D/Dynamics/b2World.cpp"
     __THREW__ = 0;
     $67 = (invoke_iii($64|0,($61|0),($66|0))|0); //@line 979 "libraries/Box2D/Dynamics/b2World.cpp"
     $68 = __THREW__; __THREW__ = 0;
     $69 = $68&1;
     if ($69) {
      label = 5;
      break;
     }
     $$pr = HEAP32[$1>>2]|0; //@line 74 "./libraries/Box2D/Common/b2GrowableStack.h"
     $70 = ($$pr|0)>(0); //@line 174 "./libraries/Box2D/Collision/b2DynamicTree.h"
     $or$cond = $67 & $70;
     if ($or$cond) {
      continue;
     } else {
      label = 17;
      break;
     }
    }
    $71 = HEAP32[$2>>2]|0; //@line 49 "./libraries/Box2D/Common/b2GrowableStack.h"
    $72 = ($12|0)==($71|0); //@line 49 "./libraries/Box2D/Common/b2GrowableStack.h"
    if ($72) {
     $73 = $12 << 1; //@line 52 "./libraries/Box2D/Common/b2GrowableStack.h"
     HEAP32[$2>>2] = $73; //@line 52 "./libraries/Box2D/Common/b2GrowableStack.h"
     $74 = $12 << 3; //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
     __THREW__ = 0;
     $75 = (invoke_ii(3831,($74|0))|0); //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
     $76 = __THREW__; __THREW__ = 0;
     $77 = $76&1;
     if ($77) {
      label = 5;
      break;
     }
     HEAP32[$stack>>2] = $75; //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
     $78 = HEAP32[$1>>2]|0; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
     $79 = $78 << 2; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
     _memcpy(($75|0),($13|0),($79|0))|0; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
     $80 = ($13|0)==($0|0); //@line 55 "./libraries/Box2D/Common/b2GrowableStack.h"
     if (!($80)) {
      __THREW__ = 0;
      invoke_vi(3816,($13|0)); //@line 57 "./libraries/Box2D/Common/b2GrowableStack.h"
      $81 = __THREW__; __THREW__ = 0;
      $82 = $81&1;
      if ($82) {
       label = 5;
       break;
      }
     }
    }
    $83 = HEAP32[$46>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    $84 = HEAP32[$1>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    $85 = HEAP32[$stack>>2]|0; //@line 51 "./libraries/Box2D/Common/b2GrowableStack.h"
    $86 = (($85) + ($84<<2)|0); //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    HEAP32[$86>>2] = $83; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    $87 = HEAP32[$1>>2]|0; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
    $88 = (($87) + 1)|0; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
    HEAP32[$1>>2] = $88; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
    $89 = (((($25) + (($15*36)|0)|0)) + 28|0); //@line 197 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $90 = HEAP32[$2>>2]|0; //@line 49 "./libraries/Box2D/Common/b2GrowableStack.h"
    $91 = ($88|0)==($90|0); //@line 49 "./libraries/Box2D/Common/b2GrowableStack.h"
    if ($91) {
     $92 = $88 << 1; //@line 52 "./libraries/Box2D/Common/b2GrowableStack.h"
     HEAP32[$2>>2] = $92; //@line 52 "./libraries/Box2D/Common/b2GrowableStack.h"
     $93 = $88 << 3; //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
     __THREW__ = 0;
     $94 = (invoke_ii(3831,($93|0))|0); //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
     $95 = __THREW__; __THREW__ = 0;
     $96 = $95&1;
     if ($96) {
      label = 5;
      break;
     }
     HEAP32[$stack>>2] = $94; //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
     $97 = HEAP32[$1>>2]|0; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
     $98 = $97 << 2; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
     _memcpy(($94|0),($85|0),($98|0))|0; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
     $99 = ($85|0)==($0|0); //@line 55 "./libraries/Box2D/Common/b2GrowableStack.h"
     if (!($99)) {
      __THREW__ = 0;
      invoke_vi(3816,($85|0)); //@line 57 "./libraries/Box2D/Common/b2GrowableStack.h"
      $100 = __THREW__; __THREW__ = 0;
      $101 = $100&1;
      if ($101) {
       label = 5;
       break;
      }
     }
    }
    $102 = HEAP32[$89>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    $103 = HEAP32[$1>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    $104 = HEAP32[$stack>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    $105 = (($104) + ($103<<2)|0); //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    HEAP32[$105>>2] = $102; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    $106 = HEAP32[$1>>2]|0; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
    $107 = (($106) + 1)|0; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
    HEAP32[$1>>2] = $107; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
    $$old = $107;$$pre15 = $104;
   }
  }
  $$old12 = ($$old|0)>(0); //@line 174 "./libraries/Box2D/Collision/b2DynamicTree.h"
  if (!($$old12)) {
   $$pre = $$pre15;
   break;
  }
 }
 if ((label|0) == 5) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  $19 = HEAP32[$stack>>2]|0; //@line 40 "./libraries/Box2D/Common/b2GrowableStack.h"
  $20 = ($19|0)==($0|0); //@line 40 "./libraries/Box2D/Common/b2GrowableStack.h"
  if ($20) {
   ___resumeException($17|0);
   // unreachable;
  }
  __THREW__ = 0;
  invoke_vi(3816,($19|0)); //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
  $21 = __THREW__; __THREW__ = 0;
  $22 = $21&1;
  if ($22) {
   $23 = ___cxa_find_matching_catch_3(0|0)|0;
   $24 = tempRet0;
   ___clang_call_terminate($23); //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
   // unreachable; //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
  }
  HEAP32[$stack>>2] = 0; //@line 43 "./libraries/Box2D/Common/b2GrowableStack.h"
  ___resumeException($17|0);
  // unreachable;
 }
 else if ((label|0) == 17) {
  $$pre$pre = HEAP32[$stack>>2]|0; //@line 40 "./libraries/Box2D/Common/b2GrowableStack.h"
  $$pre = $$pre$pre;
 }
 $108 = ($$pre|0)==($0|0); //@line 40 "./libraries/Box2D/Common/b2GrowableStack.h"
 if ($108) {
  STACKTOP = sp;return; //@line 201 "./libraries/Box2D/Collision/b2DynamicTree.h"
 }
 __THREW__ = 0;
 invoke_vi(3816,($$pre|0)); //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
 $109 = __THREW__; __THREW__ = 0;
 $110 = $109&1;
 if ($110) {
  $111 = ___cxa_find_matching_catch_3(0|0)|0;
  $112 = tempRet0;
  ___clang_call_terminate($111); //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
  // unreachable; //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
 }
 HEAP32[$stack>>2] = 0; //@line 43 "./libraries/Box2D/Common/b2GrowableStack.h"
 STACKTOP = sp;return; //@line 201 "./libraries/Box2D/Collision/b2DynamicTree.h"
}
function __ZNK7b2World7RayCastEP17b2RayCastCallbackRK6b2Vec2S4_($this,$callback,$point1,$point2) {
 $this = $this|0;
 $callback = $callback|0;
 $point1 = $point1|0;
 $point2 = $point2|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $input = 0, $wrapper = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $wrapper = sp + 24|0;
 $input = sp;
 $0 = ((($this)) + 102872|0); //@line 1022 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$wrapper>>2] = $0; //@line 1022 "libraries/Box2D/Dynamics/b2World.cpp"
 $1 = ((($wrapper)) + 4|0); //@line 1023 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$1>>2] = $callback; //@line 1023 "libraries/Box2D/Dynamics/b2World.cpp"
 $2 = ((($input)) + 16|0); //@line 1025 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAPF32[$2>>2] = 1.0; //@line 1025 "libraries/Box2D/Dynamics/b2World.cpp"
 $3 = $point1; //@line 1026 "libraries/Box2D/Dynamics/b2World.cpp"
 $4 = $3; //@line 1026 "libraries/Box2D/Dynamics/b2World.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1026 "libraries/Box2D/Dynamics/b2World.cpp"
 $6 = (($3) + 4)|0; //@line 1026 "libraries/Box2D/Dynamics/b2World.cpp"
 $7 = $6; //@line 1026 "libraries/Box2D/Dynamics/b2World.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 1026 "libraries/Box2D/Dynamics/b2World.cpp"
 $9 = $input; //@line 1026 "libraries/Box2D/Dynamics/b2World.cpp"
 $10 = $9; //@line 1026 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$10>>2] = $5; //@line 1026 "libraries/Box2D/Dynamics/b2World.cpp"
 $11 = (($9) + 4)|0; //@line 1026 "libraries/Box2D/Dynamics/b2World.cpp"
 $12 = $11; //@line 1026 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$12>>2] = $8; //@line 1026 "libraries/Box2D/Dynamics/b2World.cpp"
 $13 = ((($input)) + 8|0); //@line 1027 "libraries/Box2D/Dynamics/b2World.cpp"
 $14 = $point2; //@line 1027 "libraries/Box2D/Dynamics/b2World.cpp"
 $15 = $14; //@line 1027 "libraries/Box2D/Dynamics/b2World.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1027 "libraries/Box2D/Dynamics/b2World.cpp"
 $17 = (($14) + 4)|0; //@line 1027 "libraries/Box2D/Dynamics/b2World.cpp"
 $18 = $17; //@line 1027 "libraries/Box2D/Dynamics/b2World.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 1027 "libraries/Box2D/Dynamics/b2World.cpp"
 $20 = $13; //@line 1027 "libraries/Box2D/Dynamics/b2World.cpp"
 $21 = $20; //@line 1027 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$21>>2] = $16; //@line 1027 "libraries/Box2D/Dynamics/b2World.cpp"
 $22 = (($20) + 4)|0; //@line 1027 "libraries/Box2D/Dynamics/b2World.cpp"
 $23 = $22; //@line 1027 "libraries/Box2D/Dynamics/b2World.cpp"
 HEAP32[$23>>2] = $19; //@line 1027 "libraries/Box2D/Dynamics/b2World.cpp"
 __ZNK13b2DynamicTree7RayCastI21b2WorldRayCastWrapperEEvPT_RK14b2RayCastInput($0,$wrapper,$input); //@line 249 "./libraries/Box2D/Collision/b2BroadPhase.h"
 STACKTOP = sp;return; //@line 1029 "libraries/Box2D/Dynamics/b2World.cpp"
}
function __ZNK13b2DynamicTree7RayCastI21b2WorldRayCastWrapperEEvPT_RK14b2RayCastInput($this,$callback,$input) {
 $this = $this|0;
 $callback = $callback|0;
 $input = $input|0;
 var $$0$i10 = 0.0, $$pr = 0, $$pr$pre = 0, $0 = 0.0, $1 = 0, $10 = 0.0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0.0, $110 = 0.0, $111 = 0.0, $112 = 0.0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0.0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0.0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0.0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0.0, $165 = 0.0, $166 = 0.0, $167 = 0.0;
 var $168 = 0.0, $169 = 0.0, $17 = 0.0, $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0.0, $180 = 0.0, $181 = 0, $182 = 0, $183 = 0.0, $184 = 0, $185 = 0;
 var $186 = 0.0, $187 = 0.0, $188 = 0.0, $189 = 0.0, $19 = 0, $190 = 0, $191 = 0.0, $192 = 0, $193 = 0.0, $194 = 0, $195 = 0, $196 = 0, $197 = 0.0, $198 = 0, $199 = 0.0, $2 = 0.0, $20 = 0.0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0.0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0.0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0, $37 = 0;
 var $38 = 0.0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0.0, $42 = 0, $43 = 0, $44 = 0.0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0.0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0.0, $70 = 0, $71 = 0, $72 = 0, $73 = 0.0;
 var $74 = 0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0.0, $80 = 0.0, $81 = 0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0.0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $maxFraction$047 = 0.0, $maxFraction$2$ph = 0.0, $maxFraction$6$ph = 0.0, $or$cond$i = 0, $or$cond10$i = 0, $or$cond8$i = 0, $output$i = 0, $point$i = 0, $r$sroa$0$0 = 0.0, $r$sroa$7$0 = 0.0, $segmentAABB$sroa$0$043 = 0, $segmentAABB$sroa$0$1$ph = 0;
 var $segmentAABB$sroa$0$4$ph = 0, $segmentAABB$sroa$11$046 = 0, $segmentAABB$sroa$11$1$ph = 0, $segmentAABB$sroa$11$4$ph = 0, $segmentAABB$sroa$5$044 = 0, $segmentAABB$sroa$5$1$ph = 0, $segmentAABB$sroa$5$4$ph = 0, $segmentAABB$sroa$8$045 = 0, $segmentAABB$sroa$8$1$ph = 0, $segmentAABB$sroa$8$4$ph = 0, $stack = 0, $subInput = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1088|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $output$i = sp + 1064|0;
 $point$i = sp + 1056|0;
 $stack = sp + 20|0;
 $subInput = sp;
 $0 = +HEAPF32[$input>>2]; //@line 206 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $1 = ((($input)) + 4|0); //@line 206 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $2 = +HEAPF32[$1>>2]; //@line 206 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $3 = ((($input)) + 8|0); //@line 207 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $4 = +HEAPF32[$3>>2]; //@line 207 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $5 = ((($input)) + 12|0); //@line 207 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $6 = +HEAPF32[$5>>2]; //@line 207 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $7 = $4 - $0; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $8 = $6 - $2; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $9 = $7 * $7; //@line 110 "./libraries/Box2D/Common/b2Math.h"
 $10 = $8 * $8; //@line 110 "./libraries/Box2D/Common/b2Math.h"
 $11 = $9 + $10; //@line 110 "./libraries/Box2D/Common/b2Math.h"
 $12 = $11 > 0.0; //@line 209 "./libraries/Box2D/Collision/b2DynamicTree.h"
 __Z10loveAssertbPKc($12,361886); //@line 209 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $13 = (+Math_sqrt((+$11))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $14 = $13 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
 if ($14) {
  $r$sroa$0$0 = $7;$r$sroa$7$0 = $8;
 } else {
  $15 = 1.0 / $13; //@line 121 "./libraries/Box2D/Common/b2Math.h"
  $16 = $7 * $15; //@line 122 "./libraries/Box2D/Common/b2Math.h"
  $17 = $8 * $15; //@line 123 "./libraries/Box2D/Common/b2Math.h"
  $r$sroa$0$0 = $16;$r$sroa$7$0 = $17;
 }
 $18 = -$r$sroa$7$0; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $19 = $r$sroa$7$0 < -0.0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $20 = $19 ? $18 : $r$sroa$7$0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $21 = $r$sroa$0$0 > 0.0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $22 = -$r$sroa$0$0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $23 = $21 ? $r$sroa$0$0 : $22; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($input)) + 16|0); //@line 219 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $25 = +HEAPF32[$24>>2]; //@line 219 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $26 = $7 * $25; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $27 = $8 * $25; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $28 = $0 + $26; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $29 = $2 + $27; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $30 = ((($stack)) + 4|0); //@line 33 "./libraries/Box2D/Common/b2GrowableStack.h"
 HEAP32[$stack>>2] = $30; //@line 33 "./libraries/Box2D/Common/b2GrowableStack.h"
 $31 = ((($stack)) + 1028|0); //@line 34 "./libraries/Box2D/Common/b2GrowableStack.h"
 $32 = ((($stack)) + 1032|0); //@line 35 "./libraries/Box2D/Common/b2GrowableStack.h"
 HEAP32[$32>>2] = 256; //@line 35 "./libraries/Box2D/Common/b2GrowableStack.h"
 $33 = HEAP32[$this>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
 HEAP32[$30>>2] = $33; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
 HEAP32[$31>>2] = 1; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
 $34 = $2 > $29; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $35 = $34 ? $2 : $29; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $36 = (HEAPF32[tempDoublePtr>>2]=$35,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $37 = $0 > $28; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $38 = $37 ? $0 : $28; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $39 = (HEAPF32[tempDoublePtr>>2]=$38,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $40 = $2 < $29; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $41 = $40 ? $2 : $29; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $42 = (HEAPF32[tempDoublePtr>>2]=$41,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $43 = $0 < $28; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $44 = $43 ? $0 : $28; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $45 = (HEAPF32[tempDoublePtr>>2]=$44,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $46 = ((($this)) + 4|0); //@line 240 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $47 = ((($subInput)) + 8|0); //@line 261 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $48 = ((($subInput)) + 16|0); //@line 262 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $49 = ((($output$i)) + 8|0); //@line 1007 "libraries/Box2D/Dynamics/b2World.cpp"
 $50 = ((($subInput)) + 4|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $51 = ((($subInput)) + 8|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $52 = ((($subInput)) + 12|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $53 = ((($point$i)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $54 = ((($callback)) + 4|0); //@line 1009 "libraries/Box2D/Dynamics/b2World.cpp"
 $maxFraction$047 = $25;$segmentAABB$sroa$0$043 = $45;$segmentAABB$sroa$11$046 = $36;$segmentAABB$sroa$5$044 = $42;$segmentAABB$sroa$8$045 = $39;
 L4: while(1) {
  __THREW__ = 0;
  invoke_vii(3822,1,(446492|0)); //@line 67 "./libraries/Box2D/Common/b2GrowableStack.h"
  $56 = __THREW__; __THREW__ = 0;
  $57 = $56&1;
  if ($57) {
   label = 7;
   break;
  }
  $58 = HEAP32[$31>>2]|0; //@line 68 "./libraries/Box2D/Common/b2GrowableStack.h"
  $59 = (($58) + -1)|0; //@line 68 "./libraries/Box2D/Common/b2GrowableStack.h"
  HEAP32[$31>>2] = $59; //@line 68 "./libraries/Box2D/Common/b2GrowableStack.h"
  $60 = HEAP32[$stack>>2]|0; //@line 69 "./libraries/Box2D/Common/b2GrowableStack.h"
  $61 = (($60) + ($59<<2)|0); //@line 69 "./libraries/Box2D/Common/b2GrowableStack.h"
  $62 = HEAP32[$61>>2]|0; //@line 69 "./libraries/Box2D/Common/b2GrowableStack.h"
  $63 = ($62|0)==(-1); //@line 235 "./libraries/Box2D/Collision/b2DynamicTree.h"
  do {
   if ($63) {
    $$pr = $59;$maxFraction$6$ph = $maxFraction$047;$segmentAABB$sroa$0$4$ph = $segmentAABB$sroa$0$043;$segmentAABB$sroa$11$4$ph = $segmentAABB$sroa$11$046;$segmentAABB$sroa$5$4$ph = $segmentAABB$sroa$5$044;$segmentAABB$sroa$8$4$ph = $segmentAABB$sroa$8$045;
   } else {
    $72 = HEAP32[$46>>2]|0; //@line 240 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $73 = (HEAP32[tempDoublePtr>>2]=$segmentAABB$sroa$0$043,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $74 = (((($72) + (($62*36)|0)|0)) + 8|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $75 = +HEAPF32[$74>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $76 = $73 - $75; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $77 = (HEAP32[tempDoublePtr>>2]=$segmentAABB$sroa$5$044,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $78 = (((($72) + (($62*36)|0)|0)) + 12|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $79 = +HEAPF32[$78>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $80 = $77 - $79; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $81 = (($72) + (($62*36)|0)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $82 = +HEAPF32[$81>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $83 = (HEAP32[tempDoublePtr>>2]=$segmentAABB$sroa$8$045,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $84 = $82 - $83; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $85 = (((($72) + (($62*36)|0)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $86 = +HEAPF32[$85>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $87 = (HEAP32[tempDoublePtr>>2]=$segmentAABB$sroa$11$046,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $88 = $86 - $87; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $89 = $76 > 0.0; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
    $90 = $80 > 0.0; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
    $or$cond$i = $89 | $90; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
    $91 = $84 > 0.0; //@line 271 "./libraries/Box2D/Collision/b2Collision.h"
    $or$cond8$i = $or$cond$i | $91; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
    $92 = $88 > 0.0; //@line 271 "./libraries/Box2D/Collision/b2Collision.h"
    $or$cond10$i = $or$cond8$i | $92; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
    if ($or$cond10$i) {
     $$pr = $59;$maxFraction$6$ph = $maxFraction$047;$segmentAABB$sroa$0$4$ph = $segmentAABB$sroa$0$043;$segmentAABB$sroa$11$4$ph = $segmentAABB$sroa$11$046;$segmentAABB$sroa$5$4$ph = $segmentAABB$sroa$5$044;$segmentAABB$sroa$8$4$ph = $segmentAABB$sroa$8$045;
    } else {
     $93 = $75 + $82; //@line 448 "./libraries/Box2D/Common/b2Math.h"
     $94 = $79 + $86; //@line 448 "./libraries/Box2D/Common/b2Math.h"
     $95 = $93 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $96 = $94 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $97 = $75 - $82; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $98 = $79 - $86; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $99 = $97 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $100 = $98 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $101 = $0 - $95; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $102 = $2 - $96; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $103 = $101 * $18; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $104 = $r$sroa$0$0 * $102; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $105 = $103 + $104; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $106 = $105 > 0.0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
     $107 = -$105; //@line 618 "./libraries/Box2D/Common/b2Math.h"
     $108 = $106 ? $105 : $107; //@line 618 "./libraries/Box2D/Common/b2Math.h"
     $109 = $20 * $99; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $110 = $23 * $100; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $111 = $109 + $110; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $112 = $108 - $111; //@line 251 "./libraries/Box2D/Collision/b2DynamicTree.h"
     $113 = $112 > 0.0; //@line 252 "./libraries/Box2D/Collision/b2DynamicTree.h"
     if ($113) {
      $$pr = $59;$maxFraction$6$ph = $maxFraction$047;$segmentAABB$sroa$0$4$ph = $segmentAABB$sroa$0$043;$segmentAABB$sroa$11$4$ph = $segmentAABB$sroa$11$046;$segmentAABB$sroa$5$4$ph = $segmentAABB$sroa$5$044;$segmentAABB$sroa$8$4$ph = $segmentAABB$sroa$8$045;
     } else {
      $114 = (((($72) + (($62*36)|0)|0)) + 24|0); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $115 = HEAP32[$114>>2]|0; //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $116 = ($115|0)==(-1); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
      if (!($116)) {
       $202 = HEAP32[$32>>2]|0; //@line 49 "./libraries/Box2D/Common/b2GrowableStack.h"
       $203 = ($59|0)==($202|0); //@line 49 "./libraries/Box2D/Common/b2GrowableStack.h"
       if ($203) {
        $204 = $59 << 1; //@line 52 "./libraries/Box2D/Common/b2GrowableStack.h"
        HEAP32[$32>>2] = $204; //@line 52 "./libraries/Box2D/Common/b2GrowableStack.h"
        $205 = $59 << 3; //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
        __THREW__ = 0;
        $206 = (invoke_ii(3831,($205|0))|0); //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
        $207 = __THREW__; __THREW__ = 0;
        $208 = $207&1;
        if ($208) {
         label = 7;
         break L4;
        }
        HEAP32[$stack>>2] = $206; //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
        $209 = HEAP32[$31>>2]|0; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
        $210 = $209 << 2; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
        _memcpy(($206|0),($60|0),($210|0))|0; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
        $211 = ($60|0)==($30|0); //@line 55 "./libraries/Box2D/Common/b2GrowableStack.h"
        if (!($211)) {
         __THREW__ = 0;
         invoke_vi(3816,($60|0)); //@line 57 "./libraries/Box2D/Common/b2GrowableStack.h"
         $212 = __THREW__; __THREW__ = 0;
         $213 = $212&1;
         if ($213) {
          label = 7;
          break L4;
         }
        }
       }
       $214 = HEAP32[$114>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
       $215 = HEAP32[$31>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
       $216 = HEAP32[$stack>>2]|0; //@line 51 "./libraries/Box2D/Common/b2GrowableStack.h"
       $217 = (($216) + ($215<<2)|0); //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
       HEAP32[$217>>2] = $214; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
       $218 = HEAP32[$31>>2]|0; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
       $219 = (($218) + 1)|0; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
       HEAP32[$31>>2] = $219; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
       $220 = (((($72) + (($62*36)|0)|0)) + 28|0); //@line 284 "./libraries/Box2D/Collision/b2DynamicTree.h"
       $221 = HEAP32[$32>>2]|0; //@line 49 "./libraries/Box2D/Common/b2GrowableStack.h"
       $222 = ($219|0)==($221|0); //@line 49 "./libraries/Box2D/Common/b2GrowableStack.h"
       if ($222) {
        $223 = $219 << 1; //@line 52 "./libraries/Box2D/Common/b2GrowableStack.h"
        HEAP32[$32>>2] = $223; //@line 52 "./libraries/Box2D/Common/b2GrowableStack.h"
        $224 = $219 << 3; //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
        __THREW__ = 0;
        $225 = (invoke_ii(3831,($224|0))|0); //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
        $226 = __THREW__; __THREW__ = 0;
        $227 = $226&1;
        if ($227) {
         label = 7;
         break L4;
        }
        HEAP32[$stack>>2] = $225; //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
        $228 = HEAP32[$31>>2]|0; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
        $229 = $228 << 2; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
        _memcpy(($225|0),($216|0),($229|0))|0; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
        $230 = ($216|0)==($30|0); //@line 55 "./libraries/Box2D/Common/b2GrowableStack.h"
        if (!($230)) {
         __THREW__ = 0;
         invoke_vi(3816,($216|0)); //@line 57 "./libraries/Box2D/Common/b2GrowableStack.h"
         $231 = __THREW__; __THREW__ = 0;
         $232 = $231&1;
         if ($232) {
          label = 7;
          break L4;
         }
        }
       }
       $233 = HEAP32[$220>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
       $234 = HEAP32[$31>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
       $235 = HEAP32[$stack>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
       $236 = (($235) + ($234<<2)|0); //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
       HEAP32[$236>>2] = $233; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
       $237 = HEAP32[$31>>2]|0; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
       $238 = (($237) + 1)|0; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
       HEAP32[$31>>2] = $238; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
       $$pr = $238;$maxFraction$6$ph = $maxFraction$047;$segmentAABB$sroa$0$4$ph = $segmentAABB$sroa$0$043;$segmentAABB$sroa$11$4$ph = $segmentAABB$sroa$11$046;$segmentAABB$sroa$5$4$ph = $segmentAABB$sroa$5$044;$segmentAABB$sroa$8$4$ph = $segmentAABB$sroa$8$045;
       break;
      }
      $117 = $input; //@line 260 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $118 = $117; //@line 260 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $119 = HEAP32[$118>>2]|0; //@line 260 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $120 = (($117) + 4)|0; //@line 260 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $121 = $120; //@line 260 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $122 = HEAP32[$121>>2]|0; //@line 260 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $123 = $subInput; //@line 260 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $124 = $123; //@line 260 "./libraries/Box2D/Collision/b2DynamicTree.h"
      HEAP32[$124>>2] = $119; //@line 260 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $125 = (($123) + 4)|0; //@line 260 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $126 = $125; //@line 260 "./libraries/Box2D/Collision/b2DynamicTree.h"
      HEAP32[$126>>2] = $122; //@line 260 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $127 = $3; //@line 261 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $128 = $127; //@line 261 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $129 = HEAP32[$128>>2]|0; //@line 261 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $130 = (($127) + 4)|0; //@line 261 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $131 = $130; //@line 261 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $132 = HEAP32[$131>>2]|0; //@line 261 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $133 = $47; //@line 261 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $134 = $133; //@line 261 "./libraries/Box2D/Collision/b2DynamicTree.h"
      HEAP32[$134>>2] = $129; //@line 261 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $135 = (($133) + 4)|0; //@line 261 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $136 = $135; //@line 261 "./libraries/Box2D/Collision/b2DynamicTree.h"
      HEAP32[$136>>2] = $132; //@line 261 "./libraries/Box2D/Collision/b2DynamicTree.h"
      HEAPF32[$48>>2] = $maxFraction$047; //@line 262 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $137 = HEAP32[$callback>>2]|0; //@line 998 "libraries/Box2D/Dynamics/b2World.cpp"
      $138 = ($62|0)>(-1); //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
      if ($138) {
       $139 = ((($137)) + 12|0); //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
       $140 = HEAP32[$139>>2]|0; //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
       $141 = ($140|0)>($62|0); //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
       $142 = $141;
      } else {
       $142 = 0;
      }
      __THREW__ = 0;
      invoke_vii(3822,($142|0),(446451|0)); //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $143 = __THREW__; __THREW__ = 0;
      $144 = $143&1;
      if ($144) {
       label = 7;
       break L4;
      }
      $145 = ((($137)) + 4|0); //@line 159 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $146 = HEAP32[$145>>2]|0; //@line 159 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $147 = (((($146) + (($62*36)|0)|0)) + 16|0); //@line 159 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $148 = HEAP32[$147>>2]|0; //@line 159 "./libraries/Box2D/Collision/b2DynamicTree.h"
      $149 = ((($148)) + 16|0); //@line 1000 "libraries/Box2D/Dynamics/b2World.cpp"
      $150 = HEAP32[$149>>2]|0; //@line 1000 "libraries/Box2D/Dynamics/b2World.cpp"
      $151 = ((($148)) + 20|0); //@line 1001 "libraries/Box2D/Dynamics/b2World.cpp"
      $152 = HEAP32[$151>>2]|0; //@line 1001 "libraries/Box2D/Dynamics/b2World.cpp"
      $153 = ((($150)) + 12|0); //@line 331 "./libraries/Box2D/Dynamics/b2Fixture.h"
      $154 = HEAP32[$153>>2]|0; //@line 331 "./libraries/Box2D/Dynamics/b2Fixture.h"
      $155 = HEAP32[$154>>2]|0; //@line 331 "./libraries/Box2D/Dynamics/b2Fixture.h"
      $156 = ((($155)) + 20|0); //@line 331 "./libraries/Box2D/Dynamics/b2Fixture.h"
      $157 = HEAP32[$156>>2]|0; //@line 331 "./libraries/Box2D/Dynamics/b2Fixture.h"
      $158 = ((($150)) + 8|0); //@line 331 "./libraries/Box2D/Dynamics/b2Fixture.h"
      $159 = HEAP32[$158>>2]|0; //@line 331 "./libraries/Box2D/Dynamics/b2Fixture.h"
      $160 = ((($159)) + 12|0); //@line 476 "./libraries/Box2D/Dynamics/b2Body.h"
      __THREW__ = 0;
      $161 = (invoke_iiiiii($157|0,($154|0),($output$i|0),($subInput|0),($160|0),($152|0))|0); //@line 331 "./libraries/Box2D/Dynamics/b2Fixture.h"
      $162 = __THREW__; __THREW__ = 0;
      $163 = $162&1;
      if ($163) {
       label = 7;
       break L4;
      }
      if ($161) {
       $164 = +HEAPF32[$49>>2]; //@line 1007 "libraries/Box2D/Dynamics/b2World.cpp"
       $165 = 1.0 - $164; //@line 1008 "libraries/Box2D/Dynamics/b2World.cpp"
       $166 = +HEAPF32[$subInput>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
       $167 = $165 * $166; //@line 459 "./libraries/Box2D/Common/b2Math.h"
       $168 = +HEAPF32[$50>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
       $169 = $165 * $168; //@line 459 "./libraries/Box2D/Common/b2Math.h"
       $170 = +HEAPF32[$51>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
       $171 = $164 * $170; //@line 459 "./libraries/Box2D/Common/b2Math.h"
       $172 = +HEAPF32[$52>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
       $173 = $164 * $172; //@line 459 "./libraries/Box2D/Common/b2Math.h"
       $174 = $167 + $171; //@line 448 "./libraries/Box2D/Common/b2Math.h"
       $175 = $169 + $173; //@line 448 "./libraries/Box2D/Common/b2Math.h"
       HEAPF32[$point$i>>2] = $174; //@line 59 "./libraries/Box2D/Common/b2Math.h"
       HEAPF32[$53>>2] = $175; //@line 59 "./libraries/Box2D/Common/b2Math.h"
       $176 = HEAP32[$54>>2]|0; //@line 1009 "libraries/Box2D/Dynamics/b2World.cpp"
       $177 = HEAP32[$176>>2]|0; //@line 1009 "libraries/Box2D/Dynamics/b2World.cpp"
       $178 = ((($177)) + 8|0); //@line 1009 "libraries/Box2D/Dynamics/b2World.cpp"
       $179 = HEAP32[$178>>2]|0; //@line 1009 "libraries/Box2D/Dynamics/b2World.cpp"
       __THREW__ = 0;
       $180 = (+invoke_diiiid($179|0,($176|0),($150|0),($point$i|0),($output$i|0),(+$164))); //@line 1009 "libraries/Box2D/Dynamics/b2World.cpp"
       $181 = __THREW__; __THREW__ = 0;
       $182 = $181&1;
       if ($182) {
        label = 7;
        break L4;
       }
       $$0$i10 = $180;
      } else {
       $183 = +HEAPF32[$48>>2]; //@line 1012 "libraries/Box2D/Dynamics/b2World.cpp"
       $$0$i10 = $183;
      }
      $184 = $$0$i10 == 0.0; //@line 266 "./libraries/Box2D/Collision/b2DynamicTree.h"
      if ($184) {
       label = 36;
       break L4;
      }
      $185 = $$0$i10 > 0.0; //@line 272 "./libraries/Box2D/Collision/b2DynamicTree.h"
      if ($185) {
       $186 = $7 * $$0$i10; //@line 459 "./libraries/Box2D/Common/b2Math.h"
       $187 = $8 * $$0$i10; //@line 459 "./libraries/Box2D/Common/b2Math.h"
       $188 = $0 + $186; //@line 448 "./libraries/Box2D/Common/b2Math.h"
       $189 = $2 + $187; //@line 448 "./libraries/Box2D/Common/b2Math.h"
       $190 = $0 < $188; //@line 634 "./libraries/Box2D/Common/b2Math.h"
       $191 = $190 ? $0 : $188; //@line 634 "./libraries/Box2D/Common/b2Math.h"
       $192 = $2 < $189; //@line 634 "./libraries/Box2D/Common/b2Math.h"
       $193 = $192 ? $2 : $189; //@line 634 "./libraries/Box2D/Common/b2Math.h"
       $194 = (HEAPF32[tempDoublePtr>>2]=$191,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
       $195 = (HEAPF32[tempDoublePtr>>2]=$193,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
       $196 = $0 > $188; //@line 645 "./libraries/Box2D/Common/b2Math.h"
       $197 = $196 ? $0 : $188; //@line 645 "./libraries/Box2D/Common/b2Math.h"
       $198 = $2 > $189; //@line 645 "./libraries/Box2D/Common/b2Math.h"
       $199 = $198 ? $2 : $189; //@line 645 "./libraries/Box2D/Common/b2Math.h"
       $200 = (HEAPF32[tempDoublePtr>>2]=$197,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
       $201 = (HEAPF32[tempDoublePtr>>2]=$199,HEAP32[tempDoublePtr>>2]|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
       $maxFraction$2$ph = $$0$i10;$segmentAABB$sroa$0$1$ph = $194;$segmentAABB$sroa$11$1$ph = $201;$segmentAABB$sroa$5$1$ph = $195;$segmentAABB$sroa$8$1$ph = $200;
      } else {
       $maxFraction$2$ph = $maxFraction$047;$segmentAABB$sroa$0$1$ph = $segmentAABB$sroa$0$043;$segmentAABB$sroa$11$1$ph = $segmentAABB$sroa$11$046;$segmentAABB$sroa$5$1$ph = $segmentAABB$sroa$5$044;$segmentAABB$sroa$8$1$ph = $segmentAABB$sroa$8$045;
      }
      $$pr$pre = HEAP32[$31>>2]|0; //@line 74 "./libraries/Box2D/Common/b2GrowableStack.h"
      $$pr = $$pr$pre;$maxFraction$6$ph = $maxFraction$2$ph;$segmentAABB$sroa$0$4$ph = $segmentAABB$sroa$0$1$ph;$segmentAABB$sroa$11$4$ph = $segmentAABB$sroa$11$1$ph;$segmentAABB$sroa$5$4$ph = $segmentAABB$sroa$5$1$ph;$segmentAABB$sroa$8$4$ph = $segmentAABB$sroa$8$1$ph;
     }
    }
   }
  } while(0);
  $55 = ($$pr|0)>(0); //@line 232 "./libraries/Box2D/Collision/b2DynamicTree.h"
  if ($55) {
   $maxFraction$047 = $maxFraction$6$ph;$segmentAABB$sroa$0$043 = $segmentAABB$sroa$0$4$ph;$segmentAABB$sroa$11$046 = $segmentAABB$sroa$11$4$ph;$segmentAABB$sroa$5$044 = $segmentAABB$sroa$5$4$ph;$segmentAABB$sroa$8$045 = $segmentAABB$sroa$8$4$ph;
  } else {
   break;
  }
 }
 if ((label|0) == 7) {
  $64 = ___cxa_find_matching_catch_2()|0;
  $65 = tempRet0;
  $66 = HEAP32[$stack>>2]|0; //@line 40 "./libraries/Box2D/Common/b2GrowableStack.h"
  $67 = ($66|0)==($30|0); //@line 40 "./libraries/Box2D/Common/b2GrowableStack.h"
  if ($67) {
   ___resumeException($64|0);
   // unreachable;
  }
  __THREW__ = 0;
  invoke_vi(3816,($66|0)); //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
  $68 = __THREW__; __THREW__ = 0;
  $69 = $68&1;
  if ($69) {
   $70 = ___cxa_find_matching_catch_3(0|0)|0;
   $71 = tempRet0;
   ___clang_call_terminate($70); //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
   // unreachable; //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
  }
  HEAP32[$stack>>2] = 0; //@line 43 "./libraries/Box2D/Common/b2GrowableStack.h"
  ___resumeException($64|0);
  // unreachable;
 }
 else if ((label|0) == 36) {
 }
 $239 = HEAP32[$stack>>2]|0; //@line 40 "./libraries/Box2D/Common/b2GrowableStack.h"
 $240 = ($239|0)==($30|0); //@line 40 "./libraries/Box2D/Common/b2GrowableStack.h"
 if ($240) {
  STACKTOP = sp;return; //@line 287 "./libraries/Box2D/Collision/b2DynamicTree.h"
 }
 __THREW__ = 0;
 invoke_vi(3816,($239|0)); //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
 $241 = __THREW__; __THREW__ = 0;
 $242 = $241&1;
 if ($242) {
  $243 = ___cxa_find_matching_catch_3(0|0)|0;
  $244 = tempRet0;
  ___clang_call_terminate($243); //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
  // unreachable; //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
 }
 HEAP32[$stack>>2] = 0; //@line 43 "./libraries/Box2D/Common/b2GrowableStack.h"
 STACKTOP = sp;return; //@line 287 "./libraries/Box2D/Collision/b2DynamicTree.h"
}
function __ZN7b2World11ShiftOriginERK6b2Vec2($this,$newOrigin) {
 $this = $this|0;
 $newOrigin = $newOrigin|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0.0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b$0 = 0, $b$03 = 0, $b$04 = 0, $j$0 = 0, $j$01 = 0, $j$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 102868|0); //@line 1261 "libraries/Box2D/Dynamics/b2World.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1261 "libraries/Box2D/Dynamics/b2World.cpp"
 $2 = $1 & 2; //@line 1261 "libraries/Box2D/Dynamics/b2World.cpp"
 $3 = ($2|0)==(0); //@line 1261 "libraries/Box2D/Dynamics/b2World.cpp"
 __Z10loveAssertbPKc($3,361911); //@line 1261 "libraries/Box2D/Dynamics/b2World.cpp"
 $4 = HEAP32[$0>>2]|0; //@line 1262 "libraries/Box2D/Dynamics/b2World.cpp"
 $5 = $4 & 2; //@line 1262 "libraries/Box2D/Dynamics/b2World.cpp"
 $6 = ($5|0)==(0); //@line 1262 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($6)) {
  return; //@line 1280 "libraries/Box2D/Dynamics/b2World.cpp"
 }
 $7 = ((($this)) + 102952|0); //@line 1267 "libraries/Box2D/Dynamics/b2World.cpp"
 $b$03 = HEAP32[$7>>2]|0; //@line 1267 "libraries/Box2D/Dynamics/b2World.cpp"
 $8 = ($b$03|0)==(0|0); //@line 1267 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($8)) {
  $9 = ((($newOrigin)) + 4|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $b$04 = $b$03;
  while(1) {
   $12 = +HEAPF32[$newOrigin>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $13 = ((($b$04)) + 12|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $14 = +HEAPF32[$13>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $15 = $14 - $12; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$13>>2] = $15; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $16 = +HEAPF32[$9>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $17 = ((($b$04)) + 16|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $18 = +HEAPF32[$17>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $19 = $18 - $16; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$17>>2] = $19; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $20 = +HEAPF32[$newOrigin>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $21 = ((($b$04)) + 36|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $22 = +HEAPF32[$21>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $23 = $22 - $20; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$21>>2] = $23; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $24 = +HEAPF32[$9>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $25 = ((($b$04)) + 40|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $26 = +HEAPF32[$25>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $27 = $26 - $24; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$25>>2] = $27; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $28 = +HEAPF32[$newOrigin>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $29 = ((($b$04)) + 44|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $30 = +HEAPF32[$29>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $31 = $30 - $28; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$29>>2] = $31; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $32 = +HEAPF32[$9>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $33 = ((($b$04)) + 48|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $34 = +HEAPF32[$33>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $35 = $34 - $32; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$33>>2] = $35; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $36 = ((($b$04)) + 96|0); //@line 1267 "libraries/Box2D/Dynamics/b2World.cpp"
   $b$0 = HEAP32[$36>>2]|0; //@line 1267 "libraries/Box2D/Dynamics/b2World.cpp"
   $37 = ($b$0|0)==(0|0); //@line 1267 "libraries/Box2D/Dynamics/b2World.cpp"
   if ($37) {
    break;
   } else {
    $b$04 = $b$0;
   }
  }
 }
 $10 = ((($this)) + 102956|0); //@line 1274 "libraries/Box2D/Dynamics/b2World.cpp"
 $j$01 = HEAP32[$10>>2]|0; //@line 1274 "libraries/Box2D/Dynamics/b2World.cpp"
 $11 = ($j$01|0)==(0|0); //@line 1274 "libraries/Box2D/Dynamics/b2World.cpp"
 if (!($11)) {
  $j$02 = $j$01;
  while(1) {
   $39 = HEAP32[$j$02>>2]|0; //@line 1276 "libraries/Box2D/Dynamics/b2World.cpp"
   $40 = ((($39)) + 20|0); //@line 1276 "libraries/Box2D/Dynamics/b2World.cpp"
   $41 = HEAP32[$40>>2]|0; //@line 1276 "libraries/Box2D/Dynamics/b2World.cpp"
   FUNCTION_TABLE_vii[$41 & 8191]($j$02,$newOrigin); //@line 1276 "libraries/Box2D/Dynamics/b2World.cpp"
   $42 = ((($j$02)) + 12|0); //@line 1274 "libraries/Box2D/Dynamics/b2World.cpp"
   $j$0 = HEAP32[$42>>2]|0; //@line 1274 "libraries/Box2D/Dynamics/b2World.cpp"
   $43 = ($j$0|0)==(0|0); //@line 1274 "libraries/Box2D/Dynamics/b2World.cpp"
   if ($43) {
    break;
   } else {
    $j$02 = $j$0;
   }
  }
 }
 $38 = ((($this)) + 102872|0); //@line 254 "./libraries/Box2D/Collision/b2BroadPhase.h"
 __ZN13b2DynamicTree11ShiftOriginERK6b2Vec2($38,$newOrigin); //@line 254 "./libraries/Box2D/Collision/b2BroadPhase.h"
 return; //@line 1280 "libraries/Box2D/Dynamics/b2World.cpp"
}
function __ZN15b2ContactFilterD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 55 "./libraries/Box2D/Dynamics/b2WorldCallbacks.h"
}
function __ZN15b2ContactFilterD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 55 "./libraries/Box2D/Dynamics/b2WorldCallbacks.h"
 return; //@line 55 "./libraries/Box2D/Dynamics/b2WorldCallbacks.h"
}
function __ZN15b2ContactFilter13ShouldCollideEP9b2FixtureS1_($this,$fixtureA,$fixtureB) {
 $this = $this|0;
 $fixtureA = $fixtureA|0;
 $fixtureB = $fixtureB|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fixtureA)) + 36|0); //@line 29 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 $1 = HEAP16[$0>>1]|0; //@line 29 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 $2 = ((($fixtureB)) + 36|0); //@line 29 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 $3 = HEAP16[$2>>1]|0; //@line 29 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 $4 = ($1<<16>>16)!=($3<<16>>16); //@line 29 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 $5 = ($1<<16>>16)==(0); //@line 29 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 $or$cond = $5 | $4; //@line 29 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 if (!($or$cond)) {
  $6 = ($1<<16>>16)>(0); //@line 31 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
  $$0 = $6;
  return ($$0|0); //@line 36 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 }
 $7 = ((($fixtureA)) + 34|0); //@line 34 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 $8 = HEAP16[$7>>1]|0; //@line 34 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 $9 = ((($fixtureB)) + 32|0); //@line 34 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 $10 = HEAP16[$9>>1]|0; //@line 34 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 $11 = $10 & $8; //@line 34 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 $12 = ($11<<16>>16)==(0); //@line 34 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 if ($12) {
  $$0 = 0;
  return ($$0|0); //@line 36 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 }
 $13 = ((($fixtureA)) + 32|0); //@line 34 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 $14 = HEAP16[$13>>1]|0; //@line 34 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 $15 = ((($fixtureB)) + 34|0); //@line 34 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 $16 = HEAP16[$15>>1]|0; //@line 34 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 $17 = $16 & $14; //@line 34 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 $18 = ($17<<16>>16)!=(0); //@line 34 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
 $$0 = $18;
 return ($$0|0); //@line 36 "libraries/Box2D/Dynamics/b2WorldCallbacks.cpp"
}
function __ZN9b2Contact19InitializeRegistersEv() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __Z10loveAssertbPKc(1,361955); //@line 54 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc(1,361998); //@line 55 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[158041] = 3832; //@line 57 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632168)>>2] = 3833; //@line 58 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP8[(632172)>>0] = 1; //@line 59 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc(1,361955); //@line 54 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc(1,361998); //@line 55 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632260)>>2] = 3834; //@line 57 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632264)>>2] = 3835; //@line 58 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP8[(632268)>>0] = 1; //@line 59 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632188)>>2] = 3834; //@line 63 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632192)>>2] = 3835; //@line 64 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP8[(632196)>>0] = 0; //@line 65 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc(1,361955); //@line 54 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc(1,361998); //@line 55 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632284)>>2] = 3836; //@line 57 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632288)>>2] = 3837; //@line 58 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP8[(632292)>>0] = 1; //@line 59 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc(1,361955); //@line 54 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc(1,361998); //@line 55 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632212)>>2] = 3838; //@line 57 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632216)>>2] = 3839; //@line 58 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP8[(632220)>>0] = 1; //@line 59 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632176)>>2] = 3838; //@line 63 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632180)>>2] = 3839; //@line 64 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP8[(632184)>>0] = 0; //@line 65 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc(1,361955); //@line 54 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc(1,361998); //@line 55 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632236)>>2] = 3840; //@line 57 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632240)>>2] = 3841; //@line 58 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP8[(632244)>>0] = 1; //@line 59 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632272)>>2] = 3840; //@line 63 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632276)>>2] = 3841; //@line 64 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP8[(632280)>>0] = 0; //@line 65 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc(1,361955); //@line 54 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc(1,361998); //@line 55 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632308)>>2] = 3842; //@line 57 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632312)>>2] = 3843; //@line 58 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP8[(632316)>>0] = 1; //@line 59 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632200)>>2] = 3842; //@line 63 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632204)>>2] = 3843; //@line 64 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP8[(632208)>>0] = 0; //@line 65 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc(1,361955); //@line 54 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc(1,361998); //@line 55 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632332)>>2] = 3844; //@line 57 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632336)>>2] = 3845; //@line 58 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP8[(632340)>>0] = 1; //@line 59 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632296)>>2] = 3844; //@line 63 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[(632300)>>2] = 3845; //@line 64 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP8[(632304)>>0] = 0; //@line 65 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 return; //@line 49 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
}
function __ZN9b2Contact6CreateEP9b2FixtureiS1_iP16b2BlockAllocator($fixtureA,$indexA,$fixtureB,$indexB,$allocator) {
 $fixtureA = $fixtureA|0;
 $indexA = $indexA|0;
 $fixtureB = $fixtureB|0;
 $indexB = $indexB|0;
 $allocator = $allocator|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[1251125]|0; //@line 71 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $1 = ($0<<24>>24)==(0); //@line 71 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 if ($1) {
  __ZN9b2Contact19InitializeRegistersEv(); //@line 73 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  HEAP8[1251125] = 1; //@line 74 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 }
 $2 = ((($fixtureA)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $3 = HEAP32[$2>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $4 = ((($3)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $5 = HEAP32[$4>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $6 = ((($fixtureB)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $7 = HEAP32[$6>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $8 = ((($7)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $9 = HEAP32[$8>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $10 = ($5>>>0)<(4); //@line 80 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc($10,361955); //@line 80 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $11 = ($9>>>0)<(4); //@line 81 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc($11,361998); //@line 81 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $12 = ((632164 + (($5*48)|0)|0) + (($9*12)|0)|0); //@line 83 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 83 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $14 = ($13|0)==(0|0); //@line 84 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 if ($14) {
  $$0 = 0;
  return ($$0|0); //@line 99 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 }
 $15 = ((((632164 + (($5*48)|0)|0) + (($9*12)|0)|0)) + 8|0); //@line 86 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $16 = HEAP8[$15>>0]|0; //@line 86 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $17 = ($16<<24>>24)==(0); //@line 86 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 if ($17) {
  $19 = (FUNCTION_TABLE_iiiiii[$13 & 8191]($fixtureB,$indexB,$fixtureA,$indexA,$allocator)|0); //@line 92 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $$0 = $19;
  return ($$0|0); //@line 99 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 } else {
  $18 = (FUNCTION_TABLE_iiiiii[$13 & 8191]($fixtureA,$indexA,$fixtureB,$indexB,$allocator)|0); //@line 88 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $$0 = $18;
  return ($$0|0); //@line 99 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 }
 return (0)|0;
}
function __ZN9b2Contact7DestroyEPS_P16b2BlockAllocator($contact,$allocator) {
 $contact = $contact|0;
 $allocator = $allocator|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[1251125]|0; //@line 103 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $1 = ($0<<24>>24)!=(0); //@line 103 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc($1,362041); //@line 103 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $2 = ((($contact)) + 48|0); //@line 105 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 105 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $4 = ((($contact)) + 52|0); //@line 106 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 106 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $6 = ((($contact)) + 124|0); //@line 108 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 108 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $8 = ($7|0)>(0); //@line 108 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 if ($8) {
  $9 = ((($3)) + 38|0); //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $10 = HEAP8[$9>>0]|0; //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $11 = ($10<<24>>24)==(0); //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
  if ($11) {
   $12 = ((($5)) + 38|0); //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
   $13 = HEAP8[$12>>0]|0; //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
   $14 = ($13<<24>>24)==(0); //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
   if ($14) {
    $15 = ((($3)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
    $16 = HEAP32[$15>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
    $17 = ((($16)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
    $18 = HEAP16[$17>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
    $19 = $18&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
    $20 = $19 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
    $21 = ($20|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
    if ($21) {
     $22 = $19 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
     $23 = $22&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
     HEAP16[$17>>1] = $23; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
     $24 = ((($16)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
     HEAPF32[$24>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
    }
    $25 = ((($5)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
    $26 = HEAP32[$25>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
    $27 = ((($26)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
    $28 = HEAP16[$27>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
    $29 = $28&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
    $30 = $29 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
    $31 = ($30|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
    if ($31) {
     $32 = $29 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
     $33 = $32&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
     HEAP16[$27>>1] = $33; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
     $34 = ((($26)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
     HEAPF32[$34>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
    }
   }
  }
 }
 $35 = ((($3)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $36 = HEAP32[$35>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $37 = ((($36)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $38 = HEAP32[$37>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $39 = ((($5)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $40 = HEAP32[$39>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $41 = ((($40)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $42 = HEAP32[$41>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $43 = ($38|0)>(-1); //@line 119 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $44 = ($42|0)<(4); //@line 119 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $45 = $43 & $44; //@line 119 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc($45,362063); //@line 119 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 __Z10loveAssertbPKc($45,362063); //@line 120 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $46 = ((((632164 + (($38*48)|0)|0) + (($42*12)|0)|0)) + 4|0); //@line 122 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $47 = HEAP32[$46>>2]|0; //@line 122 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 FUNCTION_TABLE_vii[$47 & 8191]($contact,$allocator); //@line 123 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 return; //@line 124 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
}
function __ZN9b2ContactC2EP9b2FixtureiS1_i($this,$fA,$indexA,$fB,$indexB) {
 $this = $this|0;
 $fA = $fA|0;
 $indexA = $indexA|0;
 $fB = $fB|0;
 $indexB = $indexB|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0.0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (20612); //@line 127 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $0 = ((($this)) + 4|0); //@line 128 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[$0>>2] = 4; //@line 128 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $1 = ((($this)) + 48|0); //@line 130 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[$1>>2] = $fA; //@line 130 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $2 = ((($this)) + 52|0); //@line 131 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[$2>>2] = $fB; //@line 131 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $3 = ((($this)) + 56|0); //@line 133 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[$3>>2] = $indexA; //@line 133 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $4 = ((($this)) + 60|0); //@line 134 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[$4>>2] = $indexB; //@line 134 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $5 = ((($this)) + 124|0); //@line 136 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[$5>>2] = 0; //@line 136 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $6 = ((($this)) + 8|0); //@line 138 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $7 = ((($this)) + 128|0); //@line 151 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[$7>>2] = 0; //@line 151 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $8 = ((($fA)) + 16|0); //@line 153 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 dest=$6; stop=dest+40|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 138 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 153 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $10 = ((($fB)) + 16|0); //@line 153 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $11 = +HEAPF32[$10>>2]; //@line 153 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $12 = $9 * $11; //@line 39 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $13 = (+Math_sqrt((+$12))); //@line 39 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $14 = ((($this)) + 136|0); //@line 153 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAPF32[$14>>2] = $13; //@line 153 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $15 = ((($fA)) + 20|0); //@line 154 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $16 = +HEAPF32[$15>>2]; //@line 154 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $17 = ((($fB)) + 20|0); //@line 154 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $18 = +HEAPF32[$17>>2]; //@line 154 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $19 = $16 > $18; //@line 46 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $20 = $19 ? $16 : $18; //@line 46 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $21 = ((($this)) + 140|0); //@line 154 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAPF32[$21>>2] = $20; //@line 154 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $22 = ((($this)) + 144|0); //@line 156 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAPF32[$22>>2] = 0.0; //@line 156 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 return; //@line 157 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
}
function __ZN9b2ContactD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 188 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
}
function __ZN9b2ContactD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 188 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 return; //@line 188 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
}
function __ZN9b2Contact6UpdateEP17b2ContactListener($this,$listener) {
 $this = $this|0;
 $listener = $listener|0;
 var $$not = 0, $$not11 = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $i$015 = 0, $i$015$us = 0;
 var $j$014$us = 0, $j$014$us$lcssa = 0, $oldManifold = 0, $or$cond = 0, $or$cond12 = 0, $or$cond3 = 0, $or$cond8 = 0, $or$cond9 = 0, $storemerge = 0, $touching$0$off0 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $oldManifold = sp;
 $0 = ((($this)) + 64|0); //@line 163 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 dest=$oldManifold; src=$0; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 163 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $1 = ((($this)) + 4|0); //@line 166 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 166 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $3 = $2 | 4; //@line 166 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[$1>>2] = $3; //@line 166 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $4 = $2 >>> 1; //@line 169 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $5 = ((($this)) + 48|0); //@line 171 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 171 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $7 = ((($6)) + 38|0); //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $8 = HEAP8[$7>>0]|0; //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $9 = ((($this)) + 52|0); //@line 172 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 172 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $11 = ((($10)) + 38|0); //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $12 = HEAP8[$11>>0]|0; //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $13 = $12 | $8; //@line 173 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $14 = ($13<<24>>24)!=(0); //@line 173 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $15 = ((($6)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $16 = HEAP32[$15>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $17 = ((($10)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $18 = HEAP32[$17>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $19 = ((($16)) + 12|0); //@line 476 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = ((($18)) + 12|0); //@line 476 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($14) {
  $21 = ((($6)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $22 = HEAP32[$21>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $23 = ((($10)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $24 = HEAP32[$23>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $25 = ((($this)) + 56|0); //@line 185 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $26 = HEAP32[$25>>2]|0; //@line 185 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $27 = ((($this)) + 60|0); //@line 185 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $28 = HEAP32[$27>>2]|0; //@line 185 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $29 = (__Z13b2TestOverlapPK7b2ShapeiS1_iRK11b2TransformS4_($22,$26,$24,$28,$19,$20)|0); //@line 185 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $30 = ((($this)) + 124|0); //@line 188 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  HEAP32[$30>>2] = 0; //@line 188 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $$pre = $4 & 1; //@line 233 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $$pre$phiZ2D = $$pre;$touching$0$off0 = $29;
 } else {
  $31 = HEAP32[$this>>2]|0; //@line 192 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $32 = HEAP32[$31>>2]|0; //@line 192 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  FUNCTION_TABLE_viiii[$32 & 8191]($this,$0,$19,$20); //@line 192 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $33 = ((($this)) + 124|0); //@line 193 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $34 = HEAP32[$33>>2]|0; //@line 193 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $35 = ($34|0)>(0); //@line 193 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  L4: do {
   if ($35) {
    $36 = ((($oldManifold)) + 60|0); //@line 204 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
    $37 = HEAP32[$36>>2]|0; //@line 204 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
    $38 = ($37|0)>(0); //@line 204 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
    if ($38) {
     $i$015$us = 0;
    } else {
     $i$015 = 0;
     while(1) {
      $58 = (((((($this)) + 64|0) + (($i$015*20)|0)|0)) + 8|0); //@line 200 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
      HEAPF32[$58>>2] = 0.0; //@line 200 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
      $59 = (((((($this)) + 64|0) + (($i$015*20)|0)|0)) + 12|0); //@line 201 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
      HEAPF32[$59>>2] = 0.0; //@line 201 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
      $60 = (($i$015) + 1)|0; //@line 197 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
      $61 = HEAP32[$33>>2]|0; //@line 197 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
      $62 = ($60|0)<($61|0); //@line 197 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
      if ($62) {
       $i$015 = $60;
      } else {
       break L4;
      }
     }
    }
    while(1) {
     $39 = (((((($this)) + 64|0) + (($i$015$us*20)|0)|0)) + 8|0); //@line 200 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
     HEAPF32[$39>>2] = 0.0; //@line 200 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
     $40 = (((((($this)) + 64|0) + (($i$015$us*20)|0)|0)) + 12|0); //@line 201 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
     HEAPF32[$40>>2] = 0.0; //@line 201 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
     $41 = (((((($this)) + 64|0) + (($i$015$us*20)|0)|0)) + 16|0); //@line 202 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
     $42 = HEAP32[$41>>2]|0; //@line 202 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
     $j$014$us = 0;
     while(1) {
      $43 = (((($oldManifold) + (($j$014$us*20)|0)|0)) + 16|0); //@line 208 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
      $44 = HEAP32[$43>>2]|0; //@line 208 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
      $45 = ($44|0)==($42|0); //@line 208 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
      if ($45) {
       $j$014$us$lcssa = $j$014$us;
       label = 8;
       break;
      }
      $46 = (($j$014$us) + 1)|0; //@line 204 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
      $47 = ($46|0)<($37|0); //@line 204 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
      if ($47) {
       $j$014$us = $46;
      } else {
       break;
      }
     }
     if ((label|0) == 8) {
      label = 0;
      $48 = (((($oldManifold) + (($j$014$us$lcssa*20)|0)|0)) + 8|0); //@line 210 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
      $49 = HEAP32[$48>>2]|0; //@line 210 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
      HEAP32[$39>>2] = $49; //@line 210 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
      $50 = (((($oldManifold) + (($j$014$us$lcssa*20)|0)|0)) + 12|0); //@line 211 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
      $51 = HEAP32[$50>>2]|0; //@line 211 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
      HEAP32[$40>>2] = $51; //@line 211 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
     }
     $52 = (($i$015$us) + 1)|0; //@line 197 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
     $53 = HEAP32[$33>>2]|0; //@line 197 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
     $54 = ($52|0)<($53|0); //@line 197 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
     if ($54) {
      $i$015$us = $52;
     } else {
      break;
     }
    }
   }
  } while(0);
  $55 = $4 & 1; //@line 217 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $56 = ($55|0)!=(0); //@line 217 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $57 = $56 ^ $35; //@line 217 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  if ($57) {
   $63 = ((($16)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   $64 = HEAP16[$63>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   $65 = $64&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   $66 = $65 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   $67 = ($66|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   if ($67) {
    $68 = $65 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
    $69 = $68&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAP16[$63>>1] = $69; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
    $70 = ((($16)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAPF32[$70>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   }
   $71 = ((($18)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   $72 = HEAP16[$71>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   $73 = $72&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   $74 = $73 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   $75 = ($74|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   if ($75) {
    $76 = $73 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
    $77 = $76&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAP16[$71>>1] = $77; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
    $78 = ((($18)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAPF32[$78>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
    $$pre$phiZ2D = $55;$touching$0$off0 = $35;
   } else {
    $$pre$phiZ2D = $55;$touching$0$off0 = $35;
   }
  } else {
   $$pre$phiZ2D = $55;$touching$0$off0 = $35;
  }
 }
 $79 = HEAP32[$1>>2]|0; //@line 226 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $80 = $79 & -3; //@line 230 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $81 = $79 | 2; //@line 226 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $storemerge = $touching$0$off0 ? $81 : $80; //@line 224 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 HEAP32[$1>>2] = $storemerge; //@line 226 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $82 = ($$pre$phiZ2D|0)==(0); //@line 233 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $83 = ($listener|0)!=(0|0); //@line 233 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $or$cond = $83 & $touching$0$off0; //@line 233 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $or$cond8 = $82 & $or$cond; //@line 233 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 if ($or$cond8) {
  $84 = HEAP32[$listener>>2]|0; //@line 235 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $85 = ((($84)) + 8|0); //@line 235 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $86 = HEAP32[$85>>2]|0; //@line 235 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  FUNCTION_TABLE_vii[$86 & 8191]($listener,$this); //@line 235 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 }
 $$not11 = $82 ^ 1; //@line 238 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $87 = $touching$0$off0 ^ 1; //@line 238 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $or$cond3 = $83 & $87; //@line 238 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $or$cond12 = $or$cond3 & $$not11; //@line 238 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 if ($or$cond12) {
  $88 = HEAP32[$listener>>2]|0; //@line 240 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $89 = ((($88)) + 12|0); //@line 240 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  $90 = HEAP32[$89>>2]|0; //@line 240 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
  FUNCTION_TABLE_vii[$90 & 8191]($listener,$this); //@line 240 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 }
 $$not = $14 ^ 1; //@line 243 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $or$cond9 = $or$cond & $$not; //@line 243 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 if (!($or$cond9)) {
  STACKTOP = sp;return; //@line 247 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 }
 $91 = HEAP32[$listener>>2]|0; //@line 245 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $92 = ((($91)) + 16|0); //@line 245 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 $93 = HEAP32[$92>>2]|0; //@line 245 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 FUNCTION_TABLE_viii[$93 & 8191]($listener,$this,$oldManifold); //@line 245 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
 STACKTOP = sp;return; //@line 247 "libraries/Box2D/Dynamics/Contacts/b2Contact.cpp"
}
function __ZN22b2EdgeAndCircleContact8EvaluateEP10b2ManifoldRK11b2TransformS4_($this,$manifold,$xfA,$xfB) {
 $this = $this|0;
 $manifold = $manifold|0;
 $xfA = $xfA|0;
 $xfB = $xfB|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 47 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 47 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 $2 = ((($1)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $3 = HEAP32[$2>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $4 = ((($this)) + 52|0); //@line 48 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 48 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 $6 = ((($5)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $7 = HEAP32[$6>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 __Z22b2CollideEdgeAndCircleP10b2ManifoldPK11b2EdgeShapeRK11b2TransformPK13b2CircleShapeS6_($manifold,$3,$xfA,$7,$xfB); //@line 46 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 return; //@line 49 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
}
function __ZN22b2EdgeAndCircleContactD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 34 "./libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.h"
 return; //@line 34 "./libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.h"
}
function __ZN22b2EdgeAndCircleContact6CreateEP9b2FixtureiS1_iP16b2BlockAllocator($fixtureA,$0,$fixtureB,$1,$allocator) {
 $fixtureA = $fixtureA|0;
 $0 = $0|0;
 $fixtureB = $fixtureB|0;
 $1 = $1|0;
 $allocator = $allocator|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZN16b2BlockAllocator8AllocateEi($allocator,148)|0); //@line 27 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 __ZN9b2ContactC2EP9b2FixtureiS1_i($2,$fixtureA,0,$fixtureB,0); //@line 38 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 HEAP32[$2>>2] = (20632); //@line 39 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 $3 = ((($2)) + 48|0); //@line 40 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 40 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 $5 = ((($4)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $6 = HEAP32[$5>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $7 = ((($6)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $8 = HEAP32[$7>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $9 = ($8|0)==(1); //@line 40 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 __Z10loveAssertbPKc($9,362168); //@line 40 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 $10 = ((($2)) + 52|0); //@line 41 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 41 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 $12 = ((($11)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $13 = HEAP32[$12>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $14 = ((($13)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $15 = HEAP32[$14>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $16 = ($15|0)==(0); //@line 41 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 __Z10loveAssertbPKc($16,446770); //@line 41 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 return ($2|0); //@line 28 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
}
function __ZN22b2EdgeAndCircleContact7DestroyEP9b2ContactP16b2BlockAllocator($contact,$allocator) {
 $contact = $contact|0;
 $allocator = $allocator|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$contact>>2]|0; //@line 33 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 $1 = ((($0)) + 4|0); //@line 33 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 33 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 FUNCTION_TABLE_vi[$2 & 8191]($contact); //@line 33 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 __ZN16b2BlockAllocator4FreeEPvi($allocator,$contact,148); //@line 34 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
 return; //@line 35 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndCircleContact.cpp"
}
function __ZN23b2EdgeAndPolygonContact8EvaluateEP10b2ManifoldRK11b2TransformS4_($this,$manifold,$xfA,$xfB) {
 $this = $this|0;
 $manifold = $manifold|0;
 $xfA = $xfA|0;
 $xfB = $xfB|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 47 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 47 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 $2 = ((($1)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $3 = HEAP32[$2>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $4 = ((($this)) + 52|0); //@line 48 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 48 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 $6 = ((($5)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $7 = HEAP32[$6>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 __Z23b2CollideEdgeAndPolygonP10b2ManifoldPK11b2EdgeShapeRK11b2TransformPK14b2PolygonShapeS6_($manifold,$3,$xfA,$7,$xfB); //@line 46 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 return; //@line 49 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
}
function __ZN23b2EdgeAndPolygonContactD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 34 "./libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.h"
 return; //@line 34 "./libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.h"
}
function __ZN23b2EdgeAndPolygonContact6CreateEP9b2FixtureiS1_iP16b2BlockAllocator($fixtureA,$0,$fixtureB,$1,$allocator) {
 $fixtureA = $fixtureA|0;
 $0 = $0|0;
 $fixtureB = $fixtureB|0;
 $1 = $1|0;
 $allocator = $allocator|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZN16b2BlockAllocator8AllocateEi($allocator,148)|0); //@line 27 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 __ZN9b2ContactC2EP9b2FixtureiS1_i($2,$fixtureA,0,$fixtureB,0); //@line 38 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 HEAP32[$2>>2] = (20652); //@line 39 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 $3 = ((($2)) + 48|0); //@line 40 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 40 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 $5 = ((($4)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $6 = HEAP32[$5>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $7 = ((($6)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $8 = HEAP32[$7>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $9 = ($8|0)==(1); //@line 40 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 __Z10loveAssertbPKc($9,362168); //@line 40 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 $10 = ((($2)) + 52|0); //@line 41 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 41 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 $12 = ((($11)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $13 = HEAP32[$12>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $14 = ((($13)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $15 = HEAP32[$14>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $16 = ($15|0)==(2); //@line 41 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 __Z10loveAssertbPKc($16,446665); //@line 41 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 return ($2|0); //@line 28 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
}
function __ZN23b2EdgeAndPolygonContact7DestroyEP9b2ContactP16b2BlockAllocator($contact,$allocator) {
 $contact = $contact|0;
 $allocator = $allocator|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$contact>>2]|0; //@line 33 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 $1 = ((($0)) + 4|0); //@line 33 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 33 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 FUNCTION_TABLE_vi[$2 & 8191]($contact); //@line 33 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 __ZN16b2BlockAllocator4FreeEPvi($allocator,$contact,148); //@line 34 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
 return; //@line 35 "libraries/Box2D/Dynamics/Contacts/b2EdgeAndPolygonContact.cpp"
}
function __ZN25b2PolygonAndCircleContact8EvaluateEP10b2ManifoldRK11b2TransformS4_($this,$manifold,$xfA,$xfB) {
 $this = $this|0;
 $manifold = $manifold|0;
 $xfA = $xfA|0;
 $xfB = $xfB|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 47 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 47 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 $2 = ((($1)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $3 = HEAP32[$2>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $4 = ((($this)) + 52|0); //@line 48 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 48 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 $6 = ((($5)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $7 = HEAP32[$6>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 __Z25b2CollidePolygonAndCircleP10b2ManifoldPK14b2PolygonShapeRK11b2TransformPK13b2CircleShapeS6_($manifold,$3,$xfA,$7,$xfB); //@line 46 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 return; //@line 49 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
}
function __ZN25b2PolygonAndCircleContactD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 33 "./libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.h"
 return; //@line 33 "./libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.h"
}
function __ZN25b2PolygonAndCircleContact6CreateEP9b2FixtureiS1_iP16b2BlockAllocator($fixtureA,$0,$fixtureB,$1,$allocator) {
 $fixtureA = $fixtureA|0;
 $0 = $0|0;
 $fixtureB = $fixtureB|0;
 $1 = $1|0;
 $allocator = $allocator|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZN16b2BlockAllocator8AllocateEi($allocator,148)|0); //@line 27 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 __ZN9b2ContactC2EP9b2FixtureiS1_i($2,$fixtureA,0,$fixtureB,0); //@line 38 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 HEAP32[$2>>2] = (20672); //@line 39 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 $3 = ((($2)) + 48|0); //@line 40 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 40 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 $5 = ((($4)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $6 = HEAP32[$5>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $7 = ((($6)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $8 = HEAP32[$7>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $9 = ($8|0)==(2); //@line 40 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 __Z10loveAssertbPKc($9,362256); //@line 40 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 $10 = ((($2)) + 52|0); //@line 41 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 41 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 $12 = ((($11)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $13 = HEAP32[$12>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $14 = ((($13)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $15 = HEAP32[$14>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $16 = ($15|0)==(0); //@line 41 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 __Z10loveAssertbPKc($16,446770); //@line 41 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 return ($2|0); //@line 28 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
}
function __ZN25b2PolygonAndCircleContact7DestroyEP9b2ContactP16b2BlockAllocator($contact,$allocator) {
 $contact = $contact|0;
 $allocator = $allocator|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$contact>>2]|0; //@line 33 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 $1 = ((($0)) + 4|0); //@line 33 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 33 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 FUNCTION_TABLE_vi[$2 & 8191]($contact); //@line 33 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 __ZN16b2BlockAllocator4FreeEPvi($allocator,$contact,148); //@line 34 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
 return; //@line 35 "libraries/Box2D/Dynamics/Contacts/b2PolygonAndCircleContact.cpp"
}
function __ZN16b2PolygonContact8EvaluateEP10b2ManifoldRK11b2TransformS4_($this,$manifold,$xfA,$xfB) {
 $this = $this|0;
 $manifold = $manifold|0;
 $xfA = $xfA|0;
 $xfB = $xfB|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 50 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 50 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 $2 = ((($1)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $3 = HEAP32[$2>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $4 = ((($this)) + 52|0); //@line 51 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 51 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 $6 = ((($5)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $7 = HEAP32[$6>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 __Z17b2CollidePolygonsP10b2ManifoldPK14b2PolygonShapeRK11b2TransformS3_S6_($manifold,$3,$xfA,$7,$xfB); //@line 49 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 return; //@line 52 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
}
function __ZN16b2PolygonContactD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 34 "./libraries/Box2D/Dynamics/Contacts/b2PolygonContact.h"
 return; //@line 34 "./libraries/Box2D/Dynamics/Contacts/b2PolygonContact.h"
}
function __ZN16b2PolygonContact6CreateEP9b2FixtureiS1_iP16b2BlockAllocator($fixtureA,$0,$fixtureB,$1,$allocator) {
 $fixtureA = $fixtureA|0;
 $0 = $0|0;
 $fixtureB = $fixtureB|0;
 $1 = $1|0;
 $allocator = $allocator|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZN16b2BlockAllocator8AllocateEi($allocator,148)|0); //@line 30 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 __ZN9b2ContactC2EP9b2FixtureiS1_i($2,$fixtureA,0,$fixtureB,0); //@line 41 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 HEAP32[$2>>2] = (20692); //@line 42 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 $3 = ((($2)) + 48|0); //@line 43 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 43 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 $5 = ((($4)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $6 = HEAP32[$5>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $7 = ((($6)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $8 = HEAP32[$7>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $9 = ($8|0)==(2); //@line 43 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 __Z10loveAssertbPKc($9,362256); //@line 43 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 $10 = ((($2)) + 52|0); //@line 44 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 44 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 $12 = ((($11)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $13 = HEAP32[$12>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $14 = ((($13)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $15 = HEAP32[$14>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $16 = ($15|0)==(2); //@line 44 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 __Z10loveAssertbPKc($16,446665); //@line 44 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 return ($2|0); //@line 31 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
}
function __ZN16b2PolygonContact7DestroyEP9b2ContactP16b2BlockAllocator($contact,$allocator) {
 $contact = $contact|0;
 $allocator = $allocator|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$contact>>2]|0; //@line 36 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 $1 = ((($0)) + 4|0); //@line 36 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 36 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 FUNCTION_TABLE_vi[$2 & 8191]($contact); //@line 36 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 __ZN16b2BlockAllocator4FreeEPvi($allocator,$contact,148); //@line 37 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
 return; //@line 38 "libraries/Box2D/Dynamics/Contacts/b2PolygonContact.cpp"
}
function __ZN7b2Joint6CreateEPK10b2JointDefP16b2BlockAllocator($def,$allocator) {
 $def = $def|0;
 $allocator = $allocator|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $joint$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$def>>2]|0; //@line 41 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 do {
  switch ($0|0) {
  case 3:  {
   $1 = (__ZN16b2BlockAllocator8AllocateEi($allocator,176)|0); //@line 45 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   __ZN15b2DistanceJointC2EPK18b2DistanceJointDef($1,$def); //@line 46 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   $joint$0 = $1;
   break;
  }
  case 5:  {
   $2 = (__ZN16b2BlockAllocator8AllocateEi($allocator,168)|0); //@line 52 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   __ZN12b2MouseJointC2EPK15b2MouseJointDef($2,$def); //@line 53 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   $joint$0 = $2;
   break;
  }
  case 2:  {
   $3 = (__ZN16b2BlockAllocator8AllocateEi($allocator,256)|0); //@line 59 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   __ZN16b2PrismaticJointC2EPK19b2PrismaticJointDef($3,$def); //@line 60 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   $joint$0 = $3;
   break;
  }
  case 1:  {
   $4 = (__ZN16b2BlockAllocator8AllocateEi($allocator,228)|0); //@line 66 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   __ZN15b2RevoluteJointC2EPK18b2RevoluteJointDef($4,$def); //@line 67 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   $joint$0 = $4;
   break;
  }
  case 4:  {
   $5 = (__ZN16b2BlockAllocator8AllocateEi($allocator,196)|0); //@line 73 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   __ZN13b2PulleyJointC2EPK16b2PulleyJointDef($5,$def); //@line 74 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   $joint$0 = $5;
   break;
  }
  case 6:  {
   $6 = (__ZN16b2BlockAllocator8AllocateEi($allocator,276)|0); //@line 80 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   __ZN11b2GearJointC2EPK14b2GearJointDef($6,$def); //@line 81 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   $joint$0 = $6;
   break;
  }
  case 7:  {
   $7 = (__ZN16b2BlockAllocator8AllocateEi($allocator,224)|0); //@line 87 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   __ZN12b2WheelJointC2EPK15b2WheelJointDef($7,$def); //@line 88 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   $joint$0 = $7;
   break;
  }
  case 8:  {
   $8 = (__ZN16b2BlockAllocator8AllocateEi($allocator,208)|0); //@line 94 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   __ZN11b2WeldJointC2EPK14b2WeldJointDef($8,$def); //@line 95 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   $joint$0 = $8;
   break;
  }
  case 9:  {
   $9 = (__ZN16b2BlockAllocator8AllocateEi($allocator,180)|0); //@line 101 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   __ZN15b2FrictionJointC2EPK18b2FrictionJointDef($9,$def); //@line 102 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   $joint$0 = $9;
   break;
  }
  case 10:  {
   $10 = (__ZN16b2BlockAllocator8AllocateEi($allocator,168)|0); //@line 108 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   __ZN11b2RopeJointC2EPK14b2RopeJointDef($10,$def); //@line 109 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   $joint$0 = $10;
   break;
  }
  case 11:  {
   $11 = (__ZN16b2BlockAllocator8AllocateEi($allocator,192)|0); //@line 115 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   __ZN12b2MotorJointC2EPK15b2MotorJointDef($11,$def); //@line 116 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   $joint$0 = $11;
   break;
  }
  default: {
   __Z10loveAssertbPKc(0,622525); //@line 121 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   $joint$0 = 0;
  }
  }
 } while(0);
 return ($joint$0|0); //@line 125 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
}
function __ZN7b2Joint7DestroyEPS_P16b2BlockAllocator($joint,$allocator) {
 $joint = $joint|0;
 $allocator = $allocator|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$joint>>2]|0; //@line 130 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $1 = ((($0)) + 24|0); //@line 130 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 130 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 FUNCTION_TABLE_vi[$2 & 8191]($joint); //@line 130 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $3 = ((($joint)) + 4|0); //@line 131 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 131 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 do {
  switch ($4|0) {
  case 3:  {
   __ZN16b2BlockAllocator4FreeEPvi($allocator,$joint,176); //@line 134 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   break;
  }
  case 5:  {
   __ZN16b2BlockAllocator4FreeEPvi($allocator,$joint,168); //@line 138 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   break;
  }
  case 2:  {
   __ZN16b2BlockAllocator4FreeEPvi($allocator,$joint,256); //@line 142 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   break;
  }
  case 1:  {
   __ZN16b2BlockAllocator4FreeEPvi($allocator,$joint,228); //@line 146 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   break;
  }
  case 4:  {
   __ZN16b2BlockAllocator4FreeEPvi($allocator,$joint,196); //@line 150 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   break;
  }
  case 6:  {
   __ZN16b2BlockAllocator4FreeEPvi($allocator,$joint,276); //@line 154 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   break;
  }
  case 7:  {
   __ZN16b2BlockAllocator4FreeEPvi($allocator,$joint,224); //@line 158 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   break;
  }
  case 8:  {
   __ZN16b2BlockAllocator4FreeEPvi($allocator,$joint,208); //@line 162 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   break;
  }
  case 9:  {
   __ZN16b2BlockAllocator4FreeEPvi($allocator,$joint,180); //@line 166 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   break;
  }
  case 10:  {
   __ZN16b2BlockAllocator4FreeEPvi($allocator,$joint,168); //@line 170 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   break;
  }
  case 11:  {
   __ZN16b2BlockAllocator4FreeEPvi($allocator,$joint,192); //@line 174 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
   break;
  }
  default: {
   __Z10loveAssertbPKc(0,622525); //@line 178 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
  }
  }
 } while(0);
 return; //@line 181 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
}
function __ZN7b2JointC2EPK10b2JointDef($this,$def) {
 $this = $this|0;
 $def = $def|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (20712); //@line 184 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $0 = ((($def)) + 8|0); //@line 185 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 185 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $2 = ((($def)) + 12|0); //@line 185 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 185 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $4 = ($1|0)!=($3|0); //@line 185 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 __Z10loveAssertbPKc($4,362356); //@line 185 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $5 = HEAP32[$def>>2]|0; //@line 187 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $6 = ((($this)) + 4|0); //@line 187 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 HEAP32[$6>>2] = $5; //@line 187 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $7 = ((($this)) + 8|0); //@line 188 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 HEAP32[$7>>2] = 0; //@line 188 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $8 = ((($this)) + 12|0); //@line 189 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 HEAP32[$8>>2] = 0; //@line 189 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $9 = HEAP32[$0>>2]|0; //@line 190 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $10 = ((($this)) + 48|0); //@line 190 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 HEAP32[$10>>2] = $9; //@line 190 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $11 = HEAP32[$2>>2]|0; //@line 191 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $12 = ((($this)) + 52|0); //@line 191 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 HEAP32[$12>>2] = $11; //@line 191 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $13 = ((($this)) + 56|0); //@line 192 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 HEAP32[$13>>2] = 0; //@line 192 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $14 = ((($def)) + 16|0); //@line 193 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $15 = HEAP8[$14>>0]|0; //@line 193 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $16 = ((($this)) + 61|0); //@line 193 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 HEAP8[$16>>0] = $15; //@line 193 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $17 = ((($this)) + 60|0); //@line 194 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 HEAP8[$17>>0] = 0; //@line 194 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $18 = ((($def)) + 4|0); //@line 195 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 195 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $20 = ((($this)) + 64|0); //@line 195 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 HEAP32[$20>>2] = $19; //@line 195 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 $21 = ((($this)) + 16|0); //@line 198 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 ;HEAP32[$21>>2]=0|0;HEAP32[$21+4>>2]=0|0;HEAP32[$21+8>>2]=0|0;HEAP32[$21+12>>2]=0|0;HEAP32[$21+16>>2]=0|0;HEAP32[$21+20>>2]=0|0;HEAP32[$21+24>>2]=0|0;HEAP32[$21+28>>2]=0|0; //@line 198 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
 return; //@line 206 "libraries/Box2D/Dynamics/Joints/b2Joint.cpp"
}
function __ZN7b2Joint4DumpEv($this) {
 $this = $this|0;
 var $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 __Z5b2LogPKcz(362309,$vararg_buffer); //@line 147 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 STACKTOP = sp;return; //@line 147 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
}
function __ZN7b2Joint11ShiftOriginERK6b2Vec2($this,$newOrigin) {
 $this = $this|0;
 $newOrigin = $newOrigin|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 150 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
}
function __ZN7b2JointD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 162 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
}
function __ZN7b2JointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 162 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 return; //@line 162 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
}
function __ZNK12b2MotorJoint10GetAnchorAEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 208 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 208 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $2 = ((($1)) + 12|0); //@line 481 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = $2; //@line 208 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $4 = $3; //@line 208 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 208 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $6 = (($3) + 4)|0; //@line 208 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $7 = $6; //@line 208 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 208 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $9 = $agg$result; //@line 208 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $10 = $9; //@line 208 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$10>>2] = $5; //@line 208 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $11 = (($9) + 4)|0; //@line 208 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $12 = $11; //@line 208 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$12>>2] = $8; //@line 208 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 return; //@line 208 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZNK12b2MotorJoint10GetAnchorBEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 213 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 213 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $2 = ((($1)) + 12|0); //@line 481 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = $2; //@line 213 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $4 = $3; //@line 213 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 213 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $6 = (($3) + 4)|0; //@line 213 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $7 = $6; //@line 213 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 213 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $9 = $agg$result; //@line 213 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $10 = $9; //@line 213 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$10>>2] = $5; //@line 213 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $11 = (($9) + 4)|0; //@line 213 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $12 = $11; //@line 213 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$12>>2] = $8; //@line 213 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 return; //@line 213 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZNK12b2MotorJoint16GetReactionForceEf($agg$result,$this,$inv_dt) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 80|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $1 = +HEAPF32[$0>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $2 = $1 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($this)) + 84|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$3>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $5 = $4 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $2; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$6>>2] = $5; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 218 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZNK12b2MotorJoint17GetReactionTorqueEf($this,$inv_dt) {
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 88|0); //@line 223 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 223 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $2 = $1 * $inv_dt; //@line 223 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 return (+$2); //@line 223 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZN12b2MotorJoint4DumpEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer13 = 0, $vararg_buffer16 = 0, $vararg_buffer19 = 0, $vararg_buffer22 = 0, $vararg_buffer25 = 0, $vararg_buffer3 = 0, $vararg_buffer6 = 0;
 var $vararg_buffer9 = 0, $vararg_ptr12 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer25 = sp + 80|0;
 $vararg_buffer22 = sp + 72|0;
 $vararg_buffer19 = sp + 64|0;
 $vararg_buffer16 = sp + 56|0;
 $vararg_buffer13 = sp + 48|0;
 $vararg_buffer9 = sp + 32|0;
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($this)) + 48|0); //@line 291 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 291 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 291 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 291 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $4 = ((($this)) + 52|0); //@line 292 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 292 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $6 = ((($5)) + 8|0); //@line 292 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 292 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 __Z5b2LogPKcz(362396,$vararg_buffer); //@line 294 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$vararg_buffer1>>2] = $3; //@line 295 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 __Z5b2LogPKcz(447349,$vararg_buffer1); //@line 295 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$vararg_buffer3>>2] = $7; //@line 296 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 __Z5b2LogPKcz(447375,$vararg_buffer3); //@line 296 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $8 = ((($this)) + 61|0); //@line 297 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 297 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $10 = $9&255; //@line 297 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$vararg_buffer6>>2] = $10; //@line 297 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 __Z5b2LogPKcz(447401,$vararg_buffer6); //@line 297 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $11 = ((($this)) + 68|0); //@line 298 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $12 = +HEAPF32[$11>>2]; //@line 298 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $13 = $12; //@line 298 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $14 = ((($this)) + 72|0); //@line 298 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 298 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $16 = $15; //@line 298 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF64[$vararg_buffer9>>3] = $13; //@line 298 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $vararg_ptr12 = ((($vararg_buffer9)) + 8|0); //@line 298 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF64[$vararg_ptr12>>3] = $16; //@line 298 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 __Z5b2LogPKcz(362419,$vararg_buffer9); //@line 298 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $17 = ((($this)) + 76|0); //@line 299 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $18 = +HEAPF32[$17>>2]; //@line 299 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $19 = $18; //@line 299 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF64[$vararg_buffer13>>3] = $19; //@line 299 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 __Z5b2LogPKcz(362461,$vararg_buffer13); //@line 299 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $20 = ((($this)) + 92|0); //@line 300 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 300 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $22 = $21; //@line 300 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF64[$vararg_buffer16>>3] = $22; //@line 300 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 __Z5b2LogPKcz(447190,$vararg_buffer16); //@line 300 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $23 = ((($this)) + 96|0); //@line 301 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $24 = +HEAPF32[$23>>2]; //@line 301 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $25 = $24; //@line 301 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF64[$vararg_buffer19>>3] = $25; //@line 301 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 __Z5b2LogPKcz(447216,$vararg_buffer19); //@line 301 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $26 = ((($this)) + 100|0); //@line 302 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $27 = +HEAPF32[$26>>2]; //@line 302 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $28 = $27; //@line 302 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF64[$vararg_buffer22>>3] = $28; //@line 302 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 __Z5b2LogPKcz(362492,$vararg_buffer22); //@line 302 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $29 = ((($this)) + 56|0); //@line 303 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $30 = HEAP32[$29>>2]|0; //@line 303 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$vararg_buffer25>>2] = $30; //@line 303 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 __Z5b2LogPKcz(447513,$vararg_buffer25); //@line 303 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 STACKTOP = sp;return; //@line 304 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZN12b2MotorJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 59 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 return; //@line 59 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
}
function __ZN12b2MotorJoint23InitVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0;
 var $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0.0, $135 = 0, $136 = 0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0, $145 = 0.0, $146 = 0.0, $147 = 0, $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0.0;
 var $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0, $157 = 0, $158 = 0.0, $159 = 0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0.0, $169 = 0.0, $17 = 0;
 var $170 = 0.0, $171 = 0, $172 = 0.0, $173 = 0.0, $174 = 0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0, $182 = 0.0, $183 = 0.0, $184 = 0, $185 = 0.0, $186 = 0.0, $187 = 0.0, $188 = 0.0;
 var $189 = 0.0, $19 = 0, $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0.0, $194 = 0.0, $195 = 0.0, $196 = 0, $197 = 0.0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0, $200 = 0.0, $201 = 0.0, $202 = 0.0, $203 = 0.0, $204 = 0.0, $205 = 0.0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0, $53 = 0.0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0.0;
 var $62 = 0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0;
 var $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0, $88 = 0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0;
 var $99 = 0, $bitcast = 0.0, $bitcast30 = 0.0, $det$0$i = 0.0, $storemerge = 0.0, $vA$sroa$0$0 = 0, $vA$sroa$6$0 = 0, $vB$sroa$0$0 = 0, $vB$sroa$6$0 = 0, $wA$0 = 0.0, $wB$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 63 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 63 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 63 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 63 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $4 = ((($this)) + 104|0); //@line 63 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$4>>2] = $3; //@line 63 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $5 = ((($this)) + 52|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 64 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $7 = ((($6)) + 8|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 64 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $9 = ((($this)) + 108|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$9>>2] = $8; //@line 64 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $10 = ((($this)) + 128|0); //@line 65 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $11 = ((($1)) + 28|0); //@line 65 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $12 = $11; //@line 65 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $13 = $12; //@line 65 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 65 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $15 = (($12) + 4)|0; //@line 65 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $16 = $15; //@line 65 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 65 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $18 = $10; //@line 65 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $19 = $18; //@line 65 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$19>>2] = $14; //@line 65 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $20 = (($18) + 4)|0; //@line 65 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $21 = $20; //@line 65 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$21>>2] = $17; //@line 65 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $22 = ((($this)) + 136|0); //@line 66 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $23 = ((($6)) + 28|0); //@line 66 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $24 = $23; //@line 66 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $25 = $24; //@line 66 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 66 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $27 = (($24) + 4)|0; //@line 66 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $28 = $27; //@line 66 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 66 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $30 = $22; //@line 66 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $31 = $30; //@line 66 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$31>>2] = $26; //@line 66 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $32 = (($30) + 4)|0; //@line 66 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $33 = $32; //@line 66 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$33>>2] = $29; //@line 66 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $34 = ((($1)) + 120|0); //@line 67 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 67 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $36 = ((($this)) + 156|0); //@line 67 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$36>>2] = $35; //@line 67 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $37 = ((($6)) + 120|0); //@line 68 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $38 = HEAP32[$37>>2]|0; //@line 68 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $39 = ((($this)) + 160|0); //@line 68 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$39>>2] = $38; //@line 68 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $40 = ((($1)) + 128|0); //@line 69 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $41 = HEAP32[$40>>2]|0; //@line 69 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $42 = ((($this)) + 164|0); //@line 69 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$42>>2] = $41; //@line 69 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $43 = ((($6)) + 128|0); //@line 70 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $44 = HEAP32[$43>>2]|0; //@line 70 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $45 = ((($this)) + 168|0); //@line 70 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$45>>2] = $44; //@line 70 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $46 = ((($data)) + 24|0); //@line 72 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $47 = HEAP32[$46>>2]|0; //@line 72 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $48 = (($47) + (($3*12)|0)|0); //@line 72 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $49 = +HEAPF32[$48>>2]; //@line 72 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $50 = (((($47) + (($3*12)|0)|0)) + 4|0); //@line 72 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $51 = +HEAPF32[$50>>2]; //@line 72 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $52 = (((($47) + (($3*12)|0)|0)) + 8|0); //@line 73 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $53 = +HEAPF32[$52>>2]; //@line 73 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $54 = ((($data)) + 28|0); //@line 74 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $55 = HEAP32[$54>>2]|0; //@line 74 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $56 = (($55) + (($3*12)|0)|0); //@line 74 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $57 = HEAP32[$56>>2]|0; //@line 74 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $58 = (((($55) + (($3*12)|0)|0)) + 4|0); //@line 74 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $59 = HEAP32[$58>>2]|0; //@line 74 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $60 = (((($55) + (($3*12)|0)|0)) + 8|0); //@line 75 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $61 = +HEAPF32[$60>>2]; //@line 75 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $62 = (($47) + (($8*12)|0)|0); //@line 77 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $63 = +HEAPF32[$62>>2]; //@line 77 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $64 = (((($47) + (($8*12)|0)|0)) + 4|0); //@line 77 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $65 = +HEAPF32[$64>>2]; //@line 77 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $66 = (((($47) + (($8*12)|0)|0)) + 8|0); //@line 78 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $67 = +HEAPF32[$66>>2]; //@line 78 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $68 = (($55) + (($8*12)|0)|0); //@line 79 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $69 = HEAP32[$68>>2]|0; //@line 79 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $70 = (((($55) + (($8*12)|0)|0)) + 4|0); //@line 79 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $71 = HEAP32[$70>>2]|0; //@line 79 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $72 = (((($55) + (($8*12)|0)|0)) + 8|0); //@line 80 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $73 = +HEAPF32[$72>>2]; //@line 80 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $74 = (+Math_sin((+$53))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $75 = (+Math_cos((+$53))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $76 = (+Math_sin((+$67))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $77 = (+Math_cos((+$67))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $bitcast = (HEAP32[tempDoublePtr>>2]=$14,+HEAPF32[tempDoublePtr>>2]); //@line 68 "./libraries/Box2D/Common/b2Math.h"
 $78 = -$bitcast; //@line 68 "./libraries/Box2D/Common/b2Math.h"
 $79 = (HEAP32[tempDoublePtr>>2]=$17,+HEAPF32[tempDoublePtr>>2]); //@line 68 "./libraries/Box2D/Common/b2Math.h"
 $80 = -$79; //@line 68 "./libraries/Box2D/Common/b2Math.h"
 $81 = $75 * $78; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $82 = $74 * $80; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $83 = $81 - $82; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $84 = $74 * $78; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $85 = $75 * $80; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $86 = $84 + $85; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $87 = ((($this)) + 112|0); //@line 85 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$87>>2] = $83; //@line 85 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $88 = ((($this)) + 116|0); //@line 85 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$88>>2] = $86; //@line 85 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $bitcast30 = (HEAP32[tempDoublePtr>>2]=$26,+HEAPF32[tempDoublePtr>>2]); //@line 68 "./libraries/Box2D/Common/b2Math.h"
 $89 = -$bitcast30; //@line 68 "./libraries/Box2D/Common/b2Math.h"
 $90 = (HEAP32[tempDoublePtr>>2]=$29,+HEAPF32[tempDoublePtr>>2]); //@line 68 "./libraries/Box2D/Common/b2Math.h"
 $91 = -$90; //@line 68 "./libraries/Box2D/Common/b2Math.h"
 $92 = $77 * $89; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $93 = $76 * $91; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $94 = $92 - $93; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $95 = $76 * $89; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $96 = $77 * $91; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $97 = $95 + $96; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $98 = ((($this)) + 120|0); //@line 86 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$98>>2] = $94; //@line 86 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $99 = ((($this)) + 124|0); //@line 86 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$99>>2] = $97; //@line 86 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $100 = (HEAP32[tempDoublePtr>>2]=$35,+HEAPF32[tempDoublePtr>>2]); //@line 97 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $101 = (HEAP32[tempDoublePtr>>2]=$38,+HEAPF32[tempDoublePtr>>2]); //@line 97 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $102 = (HEAP32[tempDoublePtr>>2]=$41,+HEAPF32[tempDoublePtr>>2]); //@line 98 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $103 = (HEAP32[tempDoublePtr>>2]=$44,+HEAPF32[tempDoublePtr>>2]); //@line 98 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $104 = $100 + $101; //@line 101 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $105 = $102 * $86; //@line 101 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $106 = $86 * $105; //@line 101 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $107 = $104 + $106; //@line 101 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $108 = $103 * $97; //@line 101 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $109 = $97 * $108; //@line 101 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $110 = $107 + $109; //@line 101 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $111 = $102 * $83; //@line 102 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $112 = $86 * $111; //@line 102 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $113 = -$112; //@line 102 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $114 = $103 * $94; //@line 102 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $115 = $97 * $114; //@line 102 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $116 = $113 - $115; //@line 102 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $117 = $83 * $111; //@line 104 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $118 = $104 + $117; //@line 104 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $119 = $94 * $114; //@line 104 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $120 = $118 + $119; //@line 104 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $121 = $110 * $120; //@line 227 "./libraries/Box2D/Common/b2Math.h"
 $122 = $116 * $116; //@line 227 "./libraries/Box2D/Common/b2Math.h"
 $123 = $121 - $122; //@line 227 "./libraries/Box2D/Common/b2Math.h"
 $124 = $123 != 0.0; //@line 228 "./libraries/Box2D/Common/b2Math.h"
 $125 = 1.0 / $123; //@line 230 "./libraries/Box2D/Common/b2Math.h"
 $det$0$i = $124 ? $125 : $123; //@line 228 "./libraries/Box2D/Common/b2Math.h"
 $126 = $120 * $det$0$i; //@line 232 "./libraries/Box2D/Common/b2Math.h"
 $127 = $116 * $det$0$i; //@line 232 "./libraries/Box2D/Common/b2Math.h"
 $128 = -$127; //@line 232 "./libraries/Box2D/Common/b2Math.h"
 $129 = $110 * $det$0$i; //@line 233 "./libraries/Box2D/Common/b2Math.h"
 $130 = ((($this)) + 172|0); //@line 106 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$130>>2] = $126; //@line 106 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $131 = ((($this)) + 176|0); //@line 106 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$131>>2] = $128; //@line 106 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $132 = ((($this)) + 180|0); //@line 106 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$132>>2] = $128; //@line 106 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $133 = ((($this)) + 184|0); //@line 106 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$133>>2] = $129; //@line 106 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $134 = $102 + $103; //@line 108 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $135 = ((($this)) + 188|0); //@line 108 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $136 = $134 > 0.0; //@line 109 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $137 = 1.0 / $134; //@line 111 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $storemerge = $136 ? $137 : $134; //@line 109 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$135>>2] = $storemerge; //@line 108 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $138 = $63 + $94; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $139 = $65 + $97; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $140 = $138 - $49; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $141 = $139 - $51; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $142 = $140 - $83; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $143 = $141 - $86; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $144 = ((($this)) + 68|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $145 = +HEAPF32[$144>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $146 = $75 * $145; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $147 = ((($this)) + 72|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $148 = +HEAPF32[$147>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $149 = $74 * $148; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $150 = $146 - $149; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $151 = $74 * $145; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $152 = $75 * $148; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $153 = $151 + $152; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $154 = $142 - $150; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $155 = $143 - $153; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $156 = ((($this)) + 144|0); //@line 114 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$156>>2] = $154; //@line 114 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $157 = ((($this)) + 148|0); //@line 114 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$157>>2] = $155; //@line 114 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $158 = $67 - $53; //@line 115 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $159 = ((($this)) + 76|0); //@line 115 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $160 = +HEAPF32[$159>>2]; //@line 115 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $161 = $158 - $160; //@line 115 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $162 = ((($this)) + 152|0); //@line 115 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$162>>2] = $161; //@line 115 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $163 = ((($data)) + 20|0); //@line 117 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $164 = HEAP8[$163>>0]|0; //@line 117 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $165 = ($164<<24>>24)==(0); //@line 117 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $166 = ((($this)) + 80|0); //@line 120 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 if ($165) {
  HEAPF32[$166>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $206 = ((($this)) + 84|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$206>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $207 = ((($this)) + 88|0); //@line 132 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  HEAPF32[$207>>2] = 0.0; //@line 132 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $vA$sroa$0$0 = $57;$vA$sroa$6$0 = $59;$vB$sroa$0$0 = $69;$vB$sroa$6$0 = $71;$wA$0 = $61;$wB$0 = $73;
  $208 = HEAP32[$4>>2]|0; //@line 135 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $209 = HEAP32[$54>>2]|0; //@line 135 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $210 = (($209) + (($208*12)|0)|0); //@line 135 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  HEAP32[$210>>2] = $vA$sroa$0$0; //@line 135 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $211 = (((($209) + (($208*12)|0)|0)) + 4|0); //@line 135 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  HEAP32[$211>>2] = $vA$sroa$6$0; //@line 135 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $212 = HEAP32[$4>>2]|0; //@line 136 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $213 = HEAP32[$54>>2]|0; //@line 136 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $214 = (((($213) + (($212*12)|0)|0)) + 8|0); //@line 136 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  HEAPF32[$214>>2] = $wA$0; //@line 136 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $215 = HEAP32[$9>>2]|0; //@line 137 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $216 = (($213) + (($215*12)|0)|0); //@line 137 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  HEAP32[$216>>2] = $vB$sroa$0$0; //@line 137 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $217 = (((($213) + (($215*12)|0)|0)) + 4|0); //@line 137 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  HEAP32[$217>>2] = $vB$sroa$6$0; //@line 137 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $218 = HEAP32[$9>>2]|0; //@line 138 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $219 = HEAP32[$54>>2]|0; //@line 138 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $220 = (((($219) + (($218*12)|0)|0)) + 8|0); //@line 138 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  HEAPF32[$220>>2] = $wB$0; //@line 138 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  return; //@line 139 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 } else {
  $167 = ((($data)) + 8|0); //@line 120 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $168 = +HEAPF32[$167>>2]; //@line 120 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $169 = +HEAPF32[$166>>2]; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $170 = $168 * $169; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$166>>2] = $170; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $171 = ((($this)) + 84|0); //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $172 = +HEAPF32[$171>>2]; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $173 = $168 * $172; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$171>>2] = $173; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $174 = ((($this)) + 88|0); //@line 121 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $175 = +HEAPF32[$174>>2]; //@line 121 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $176 = $168 * $175; //@line 121 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  HEAPF32[$174>>2] = $176; //@line 121 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $177 = $100 * $170; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $178 = $100 * $173; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $179 = (HEAP32[tempDoublePtr>>2]=$57,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $180 = $179 - $177; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $181 = (HEAPF32[tempDoublePtr>>2]=$180,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $182 = (HEAP32[tempDoublePtr>>2]=$59,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $183 = $182 - $178; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $184 = (HEAPF32[tempDoublePtr>>2]=$183,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $185 = +HEAPF32[$87>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $186 = $173 * $185; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $187 = $170 * $86; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $188 = $186 - $187; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $189 = $176 + $188; //@line 125 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $190 = $102 * $189; //@line 125 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $191 = $61 - $190; //@line 125 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $192 = $101 * $170; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $193 = $101 * $173; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $194 = (HEAP32[tempDoublePtr>>2]=$69,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $195 = $194 + $192; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $196 = (HEAPF32[tempDoublePtr>>2]=$195,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $197 = (HEAP32[tempDoublePtr>>2]=$71,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $198 = $197 + $193; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $199 = (HEAPF32[tempDoublePtr>>2]=$198,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $200 = $173 * $94; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $201 = $170 * $97; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $202 = $200 - $201; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $203 = $176 + $202; //@line 127 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $204 = $103 * $203; //@line 127 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $205 = $73 + $204; //@line 127 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $vA$sroa$0$0 = $181;$vA$sroa$6$0 = $184;$vB$sroa$0$0 = $196;$vB$sroa$6$0 = $199;$wA$0 = $191;$wB$0 = $205;
  $208 = HEAP32[$4>>2]|0; //@line 135 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $209 = HEAP32[$54>>2]|0; //@line 135 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $210 = (($209) + (($208*12)|0)|0); //@line 135 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  HEAP32[$210>>2] = $vA$sroa$0$0; //@line 135 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $211 = (((($209) + (($208*12)|0)|0)) + 4|0); //@line 135 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  HEAP32[$211>>2] = $vA$sroa$6$0; //@line 135 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $212 = HEAP32[$4>>2]|0; //@line 136 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $213 = HEAP32[$54>>2]|0; //@line 136 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $214 = (((($213) + (($212*12)|0)|0)) + 8|0); //@line 136 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  HEAPF32[$214>>2] = $wA$0; //@line 136 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $215 = HEAP32[$9>>2]|0; //@line 137 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $216 = (($213) + (($215*12)|0)|0); //@line 137 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  HEAP32[$216>>2] = $vB$sroa$0$0; //@line 137 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $217 = (((($213) + (($215*12)|0)|0)) + 4|0); //@line 137 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  HEAP32[$217>>2] = $vB$sroa$6$0; //@line 137 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $218 = HEAP32[$9>>2]|0; //@line 138 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $219 = HEAP32[$54>>2]|0; //@line 138 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $220 = (((($219) + (($218*12)|0)|0)) + 8|0); //@line 138 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  HEAPF32[$220>>2] = $wB$0; //@line 138 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  return; //@line 139 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 }
}
function __ZN12b2MotorJoint24SolveVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0, $11 = 0, $110 = 0.0, $111 = 0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0;
 var $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0.0, $130 = 0.0, $131 = 0.0, $132 = 0.0, $133 = 0.0;
 var $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0.0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $16 = 0, $17 = 0.0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0, $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0;
 var $32 = 0.0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0, $5 = 0.0;
 var $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0.0;
 var $69 = 0, $7 = 0.0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0, $75 = 0.0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0.0, $85 = 0, $86 = 0.0;
 var $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0.0, $90 = 0.0, $91 = 0.0, $92 = 0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0, $97 = 0.0, $98 = 0, $99 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 104|0); //@line 143 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 143 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $2 = ((($data)) + 28|0); //@line 143 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 143 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $4 = (($3) + (($1*12)|0)|0); //@line 143 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 143 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $6 = (((($3) + (($1*12)|0)|0)) + 4|0); //@line 143 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 143 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $8 = (((($3) + (($1*12)|0)|0)) + 8|0); //@line 144 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 144 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $10 = ((($this)) + 108|0); //@line 145 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 145 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $12 = (($3) + (($11*12)|0)|0); //@line 145 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $13 = +HEAPF32[$12>>2]; //@line 145 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $14 = (((($3) + (($11*12)|0)|0)) + 4|0); //@line 145 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 145 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $16 = (((($3) + (($11*12)|0)|0)) + 8|0); //@line 146 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 146 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $18 = ((($this)) + 156|0); //@line 148 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $19 = +HEAPF32[$18>>2]; //@line 148 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $20 = ((($this)) + 160|0); //@line 148 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 148 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $22 = ((($this)) + 164|0); //@line 149 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $23 = +HEAPF32[$22>>2]; //@line 149 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $24 = ((($this)) + 168|0); //@line 149 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $25 = +HEAPF32[$24>>2]; //@line 149 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $26 = +HEAPF32[$data>>2]; //@line 151 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $27 = ((($data)) + 4|0); //@line 152 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $28 = +HEAPF32[$27>>2]; //@line 152 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $29 = $17 - $9; //@line 156 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $30 = ((($this)) + 100|0); //@line 156 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $31 = +HEAPF32[$30>>2]; //@line 156 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $32 = $28 * $31; //@line 156 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $33 = ((($this)) + 152|0); //@line 156 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $34 = +HEAPF32[$33>>2]; //@line 156 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $35 = $32 * $34; //@line 156 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $36 = $29 + $35; //@line 156 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $37 = ((($this)) + 188|0); //@line 157 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $38 = +HEAPF32[$37>>2]; //@line 157 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $39 = $38 * $36; //@line 157 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $40 = ((($this)) + 88|0); //@line 159 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $41 = +HEAPF32[$40>>2]; //@line 159 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $42 = ((($this)) + 96|0); //@line 160 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $43 = +HEAPF32[$42>>2]; //@line 160 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $44 = $26 * $43; //@line 160 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $45 = $41 - $39; //@line 161 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $46 = -$44; //@line 161 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $47 = $45 < $44; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $48 = $47 ? $45 : $44; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $49 = $48 < $46; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $50 = $49 ? $46 : $48; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$40>>2] = $50; //@line 161 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $51 = $50 - $41; //@line 162 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $52 = $23 * $51; //@line 164 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $53 = $9 - $52; //@line 164 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $54 = $25 * $51; //@line 165 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $55 = $17 + $54; //@line 165 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $56 = ((($this)) + 124|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $57 = +HEAPF32[$56>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $58 = $57 * $55; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $59 = ((($this)) + 120|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $60 = +HEAPF32[$59>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $61 = $60 * $55; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $62 = $13 - $58; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $63 = $15 + $61; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $64 = $62 - $5; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $65 = $63 - $7; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $66 = ((($this)) + 116|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $67 = +HEAPF32[$66>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $68 = $67 * $53; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $69 = ((($this)) + 112|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $70 = +HEAPF32[$69>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $71 = $70 * $53; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $72 = $68 + $64; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $73 = $65 - $71; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $74 = ((($this)) + 144|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $75 = +HEAPF32[$74>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $76 = $75 * $32; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $77 = ((($this)) + 148|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $78 = +HEAPF32[$77>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $79 = $32 * $78; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $80 = $76 + $72; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $81 = $79 + $73; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $82 = ((($this)) + 172|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $83 = +HEAPF32[$82>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $84 = $83 * $80; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $85 = ((($this)) + 180|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $86 = +HEAPF32[$85>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $87 = $86 * $81; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $88 = $84 + $87; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $89 = ((($this)) + 176|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $90 = +HEAPF32[$89>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $91 = $90 * $80; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $92 = ((($this)) + 184|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $93 = +HEAPF32[$92>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $94 = $93 * $81; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $95 = $91 + $94; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $96 = ((($this)) + 80|0); //@line 173 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $97 = +HEAPF32[$96>>2]; //@line 173 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $98 = ((($this)) + 84|0); //@line 173 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $99 = +HEAPF32[$98>>2]; //@line 173 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $100 = $97 - $88; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$96>>2] = $100; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $101 = $99 - $95; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$98>>2] = $101; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $102 = ((($this)) + 92|0); //@line 176 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $103 = +HEAPF32[$102>>2]; //@line 176 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $104 = $26 * $103; //@line 176 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $105 = $100 * $100; //@line 110 "./libraries/Box2D/Common/b2Math.h"
 $106 = $101 * $101; //@line 110 "./libraries/Box2D/Common/b2Math.h"
 $107 = $106 + $105; //@line 110 "./libraries/Box2D/Common/b2Math.h"
 $108 = $104 * $104; //@line 178 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $109 = $107 > $108; //@line 178 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 if ($109) {
  $110 = (+Math_sqrt((+$107))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $111 = $110 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
  if ($111) {
   $116 = $100;$118 = $101;
  } else {
   $112 = 1.0 / $110; //@line 121 "./libraries/Box2D/Common/b2Math.h"
   $113 = $100 * $112; //@line 122 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$96>>2] = $113; //@line 122 "./libraries/Box2D/Common/b2Math.h"
   $114 = $101 * $112; //@line 123 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$98>>2] = $114; //@line 123 "./libraries/Box2D/Common/b2Math.h"
   $116 = $113;$118 = $114;
  }
  $115 = $104 * $116; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$96>>2] = $115; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $117 = $104 * $118; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$98>>2] = $117; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $120 = $115;$122 = $117;
 } else {
  $120 = $100;$122 = $101;
 }
 $119 = $120 - $97; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $121 = $122 - $99; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $123 = $19 * $119; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $124 = $19 * $121; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $125 = $5 - $123; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $126 = $7 - $124; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $127 = $70 * $121; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $128 = $119 * $67; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $129 = $127 - $128; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $130 = $23 * $129; //@line 187 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $131 = $53 - $130; //@line 187 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $132 = $21 * $119; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $133 = $21 * $121; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $134 = $13 + $132; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $135 = $15 + $133; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $136 = $121 * $60; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $137 = $119 * $57; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $138 = $136 - $137; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $139 = $25 * $138; //@line 190 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $140 = $55 + $139; //@line 190 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $141 = HEAP32[$0>>2]|0; //@line 193 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $142 = HEAP32[$2>>2]|0; //@line 193 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $143 = (($142) + (($141*12)|0)|0); //@line 193 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$143>>2] = $125; //@line 193 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $144 = (((($142) + (($141*12)|0)|0)) + 4|0); //@line 193 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$144>>2] = $126; //@line 193 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $145 = HEAP32[$0>>2]|0; //@line 194 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $146 = HEAP32[$2>>2]|0; //@line 194 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $147 = (((($146) + (($145*12)|0)|0)) + 8|0); //@line 194 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$147>>2] = $131; //@line 194 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $148 = HEAP32[$10>>2]|0; //@line 195 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $149 = (($146) + (($148*12)|0)|0); //@line 195 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$149>>2] = $134; //@line 195 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $150 = (((($146) + (($148*12)|0)|0)) + 4|0); //@line 195 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$150>>2] = $135; //@line 195 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $151 = HEAP32[$10>>2]|0; //@line 196 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $152 = HEAP32[$2>>2]|0; //@line 196 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $153 = (((($152) + (($151*12)|0)|0)) + 8|0); //@line 196 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$153>>2] = $140; //@line 196 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 return; //@line 197 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZN12b2MotorJoint24SolvePositionConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1; //@line 203 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZN15b2MotorJointDef10InitializeEP6b2BodyS1_($this,$bA,$bB) {
 $this = $this|0;
 $bA = $bA|0;
 $bB = $bB|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0, $24 = 0, $25 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 37 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$0>>2] = $bA; //@line 37 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $1 = ((($this)) + 12|0); //@line 38 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$1>>2] = $bB; //@line 38 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $2 = ((($bB)) + 12|0); //@line 39 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 39 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $4 = ((($bB)) + 16|0); //@line 39 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 39 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $6 = ((($bA)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $8 = $3 - $7; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($bA)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $11 = $5 - $10; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $12 = ((($bA)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $13 = +HEAPF32[$12>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $14 = $8 * $13; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $15 = ((($bA)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $16 = +HEAPF32[$15>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $17 = $11 * $16; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $18 = $14 + $17; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $19 = $8 * $16; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $20 = $13 * $11; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $21 = $20 - $19; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($this)) + 20|0); //@line 40 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$22>>2] = $18; //@line 40 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $23 = ((($this)) + 24|0); //@line 40 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$23>>2] = $21; //@line 40 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $24 = ((($bA)) + 56|0); //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $25 = +HEAPF32[$24>>2]; //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $26 = ((($bB)) + 56|0); //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $27 = +HEAPF32[$26>>2]; //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $28 = $27 - $25; //@line 44 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $29 = ((($this)) + 28|0); //@line 44 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$29>>2] = $28; //@line 44 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 return; //@line 45 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZN12b2MotorJointC2EPK15b2MotorJointDef($this,$def) {
 $this = $this|0;
 $def = $def|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN7b2JointC2EPK10b2JointDef($this,$def); //@line 48 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$this>>2] = (20764); //@line 49 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $0 = ((($this)) + 68|0); //@line 47 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $1 = ((($def)) + 20|0); //@line 50 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $2 = $1; //@line 50 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $3 = $2; //@line 50 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 50 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $5 = (($2) + 4)|0; //@line 50 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $6 = $5; //@line 50 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 50 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $8 = $0; //@line 50 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $9 = $8; //@line 50 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$9>>2] = $4; //@line 50 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $10 = (($8) + 4)|0; //@line 50 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $11 = $10; //@line 50 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$11>>2] = $7; //@line 50 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $12 = ((($def)) + 28|0); //@line 51 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 51 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $14 = ((($this)) + 76|0); //@line 51 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$14>>2] = $13; //@line 51 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $15 = ((($this)) + 80|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$15>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $16 = ((($this)) + 84|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$16>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $17 = ((($this)) + 88|0); //@line 54 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$17>>2] = 0.0; //@line 54 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $18 = ((($def)) + 32|0); //@line 56 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 56 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $20 = ((($this)) + 92|0); //@line 56 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$20>>2] = $19; //@line 56 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $21 = ((($def)) + 36|0); //@line 57 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $22 = HEAP32[$21>>2]|0; //@line 57 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $23 = ((($this)) + 96|0); //@line 57 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$23>>2] = $22; //@line 57 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $24 = ((($def)) + 40|0); //@line 58 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $25 = HEAP32[$24>>2]|0; //@line 58 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $26 = ((($this)) + 100|0); //@line 58 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$26>>2] = $25; //@line 58 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 return; //@line 59 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZN12b2MotorJoint11SetMaxForceEf($this,$force) {
 $this = $this|0;
 $force = +$force;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (HEAPF32[tempDoublePtr>>2]=$force,HEAP32[tempDoublePtr>>2]|0); //@line 228 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $1 = $0 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $2 = ($1|0)!=(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $3 = $force >= 0.0; //@line 228 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $4 = $3 & $2; //@line 228 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 __Z10loveAssertbPKc($4,447243); //@line 228 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $5 = ((($this)) + 92|0); //@line 229 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$5>>2] = $force; //@line 229 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 return; //@line 230 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZNK12b2MotorJoint11GetMaxForceEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 92|0); //@line 234 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 234 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 return (+$1); //@line 234 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZN12b2MotorJoint12SetMaxTorqueEf($this,$torque) {
 $this = $this|0;
 $torque = +$torque;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (HEAPF32[tempDoublePtr>>2]=$torque,HEAP32[tempDoublePtr>>2]|0); //@line 239 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $1 = $0 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $2 = ($1|0)!=(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $3 = $torque >= 0.0; //@line 239 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $4 = $3 & $2; //@line 239 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 __Z10loveAssertbPKc($4,447277); //@line 239 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $5 = ((($this)) + 96|0); //@line 240 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$5>>2] = $torque; //@line 240 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 return; //@line 241 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZNK12b2MotorJoint12GetMaxTorqueEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 96|0); //@line 245 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 245 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 return (+$1); //@line 245 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZN12b2MotorJoint19SetCorrectionFactorEf($this,$factor) {
 $this = $this|0;
 $factor = +$factor;
 var $$ = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (HEAPF32[tempDoublePtr>>2]=$factor,HEAP32[tempDoublePtr>>2]|0); //@line 250 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $1 = $0 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $2 = ($1|0)!=(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $3 = $factor >= 0.0; //@line 250 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $or$cond = $3 & $2; //@line 250 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $4 = $factor <= 1.0; //@line 250 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $$ = $4 & $or$cond; //@line 250 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 __Z10loveAssertbPKc($$,362526); //@line 250 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $5 = ((($this)) + 100|0); //@line 251 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAPF32[$5>>2] = $factor; //@line 251 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 return; //@line 252 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZNK12b2MotorJoint19GetCorrectionFactorEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 100|0); //@line 256 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 256 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 return (+$1); //@line 256 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZN12b2MotorJoint15SetLinearOffsetERK6b2Vec2($this,$linearOffset) {
 $this = $this|0;
 $linearOffset = $linearOffset|0;
 var $0 = 0.0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = +HEAPF32[$linearOffset>>2]; //@line 261 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $1 = ((($this)) + 68|0); //@line 261 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $2 = +HEAPF32[$1>>2]; //@line 261 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $3 = $0 != $2; //@line 261 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 if (!($3)) {
  $4 = ((($linearOffset)) + 4|0); //@line 261 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $5 = +HEAPF32[$4>>2]; //@line 261 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $6 = ((($this)) + 72|0); //@line 261 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $7 = +HEAPF32[$6>>2]; //@line 261 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  $8 = $5 != $7; //@line 261 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  if (!($8)) {
   return; //@line 267 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
  }
 }
 $9 = ((($this)) + 48|0); //@line 263 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 263 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $11 = ((($10)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $12 = HEAP16[$11>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $13 = $12&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $14 = $13 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $15 = ($14|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($15) {
  $16 = $13 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $17 = $16&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$11>>1] = $17; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $18 = ((($10)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$18>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $19 = ((($this)) + 52|0); //@line 264 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 264 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $21 = ((($20)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $22 = HEAP16[$21>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $23 = $22&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $24 = $23 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $25 = ($24|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($25) {
  $26 = $23 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $27 = $26&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$21>>1] = $27; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $28 = ((($20)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$28>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $29 = $linearOffset; //@line 265 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $30 = $29; //@line 265 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 265 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $32 = (($29) + 4)|0; //@line 265 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $33 = $32; //@line 265 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $34 = HEAP32[$33>>2]|0; //@line 265 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $35 = $1; //@line 265 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $36 = $35; //@line 265 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$36>>2] = $31; //@line 265 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $37 = (($35) + 4)|0; //@line 265 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $38 = $37; //@line 265 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 HEAP32[$38>>2] = $34; //@line 265 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 return; //@line 267 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZNK12b2MotorJoint15GetLinearOffsetEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 68|0); //@line 271 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 return ($0|0); //@line 271 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZN12b2MotorJoint16SetAngularOffsetEf($this,$angularOffset) {
 $this = $this|0;
 $angularOffset = +$angularOffset;
 var $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 76|0); //@line 276 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 276 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $2 = $1 != $angularOffset; //@line 276 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 if (!($2)) {
  return; //@line 282 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 }
 $3 = ((($this)) + 48|0); //@line 278 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 278 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $5 = ((($4)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $6 = HEAP16[$5>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $7 = $6&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $8 = $7 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $9 = ($8|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($9) {
  $10 = $7 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $11 = $10&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$5>>1] = $11; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $12 = ((($4)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$12>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $13 = ((($this)) + 52|0); //@line 279 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 279 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $15 = ((($14)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $16 = HEAP16[$15>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $17 = $16&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $18 = $17 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = ($18|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($19) {
  $20 = $17 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $21 = $20&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$15>>1] = $21; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $22 = ((($14)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$22>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 HEAPF32[$0>>2] = $angularOffset; //@line 280 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 return; //@line 282 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZNK12b2MotorJoint16GetAngularOffsetEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 76|0); //@line 286 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 286 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
 return (+$1); //@line 286 "libraries/Box2D/Dynamics/Joints/b2MotorJoint.cpp"
}
function __ZNK12b2MouseJoint10GetAnchorAEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 76|0); //@line 201 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $1 = $0; //@line 201 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $2 = $1; //@line 201 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 201 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $4 = (($1) + 4)|0; //@line 201 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $5 = $4; //@line 201 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 201 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $7 = $agg$result; //@line 201 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $8 = $7; //@line 201 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAP32[$8>>2] = $3; //@line 201 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $9 = (($7) + 4)|0; //@line 201 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $10 = $9; //@line 201 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAP32[$10>>2] = $6; //@line 201 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 return; //@line 201 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
}
function __ZNK12b2MouseJoint10GetAnchorBEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 206 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 206 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 68|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 72|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 206 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
}
function __ZNK12b2MouseJoint16GetReactionForceEf($agg$result,$this,$inv_dt) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 96|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $1 = +HEAPF32[$0>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $2 = $1 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($this)) + 100|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$3>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $5 = $4 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $2; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$6>>2] = $5; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 211 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
}
function __ZNK12b2MouseJoint17GetReactionTorqueEf($this,$inv_dt) {
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $inv_dt * 0.0; //@line 216 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 return (+$0); //@line 216 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
}
function __ZN12b2MouseJoint4DumpEv($this) {
 $this = $this|0;
 var $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 __Z5b2LogPKcz(362595,$vararg_buffer); //@line 93 "./libraries/Box2D/Dynamics/Joints/b2MouseJoint.h"
 STACKTOP = sp;return; //@line 93 "./libraries/Box2D/Dynamics/Joints/b2MouseJoint.h"
}
function __ZN12b2MouseJoint11ShiftOriginERK6b2Vec2($this,$newOrigin) {
 $this = $this|0;
 $newOrigin = $newOrigin|0;
 var $0 = 0.0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = +HEAPF32[$newOrigin>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $1 = ((($this)) + 76|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $2 = +HEAPF32[$1>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $3 = $2 - $0; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$1>>2] = $3; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($newOrigin)) + 4|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($this)) + 80|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $8 = $7 - $5; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$6>>2] = $8; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 222 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
}
function __ZN12b2MouseJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 60 "./libraries/Box2D/Dynamics/Joints/b2MouseJoint.h"
 return; //@line 60 "./libraries/Box2D/Dynamics/Joints/b2MouseJoint.h"
}
function __ZN12b2MouseJoint23InitVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0.0, $104 = 0.0, $105 = 0, $106 = 0.0, $107 = 0.0, $108 = 0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0, $112 = 0, $113 = 0.0, $114 = 0.0, $115 = 0.0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0, $132 = 0.0, $133 = 0.0;
 var $134 = 0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0.0, $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0;
 var $56 = 0, $57 = 0.0, $58 = 0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0;
 var $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0, $79 = 0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0;
 var $92 = 0.0, $93 = 0.0, $94 = 0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0, $det$0$i = 0.0, $storemerge = 0.0, $vB$sroa$0$0 = 0, $vB$sroa$6$0 = 0, $wB$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 98 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 98 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 98 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 98 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $4 = ((($this)) + 116|0); //@line 98 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAP32[$4>>2] = $3; //@line 98 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $5 = ((($this)) + 128|0); //@line 99 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $6 = ((($1)) + 28|0); //@line 99 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $7 = $6; //@line 99 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $8 = $7; //@line 99 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 99 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $10 = (($7) + 4)|0; //@line 99 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $11 = $10; //@line 99 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 99 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $13 = $5; //@line 99 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $14 = $13; //@line 99 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAP32[$14>>2] = $9; //@line 99 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $15 = (($13) + 4)|0; //@line 99 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $16 = $15; //@line 99 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAP32[$16>>2] = $12; //@line 99 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $17 = ((($1)) + 120|0); //@line 100 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 100 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $19 = ((($this)) + 136|0); //@line 100 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAP32[$19>>2] = $18; //@line 100 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $20 = ((($1)) + 128|0); //@line 101 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 101 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $22 = ((($this)) + 140|0); //@line 101 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAP32[$22>>2] = $21; //@line 101 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $23 = ((($data)) + 24|0); //@line 103 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 103 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $25 = (($24) + (($3*12)|0)|0); //@line 103 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $26 = +HEAPF32[$25>>2]; //@line 103 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $27 = (((($24) + (($3*12)|0)|0)) + 4|0); //@line 103 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $28 = +HEAPF32[$27>>2]; //@line 103 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $29 = (((($24) + (($3*12)|0)|0)) + 8|0); //@line 104 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $30 = +HEAPF32[$29>>2]; //@line 104 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $31 = ((($data)) + 28|0); //@line 105 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $32 = HEAP32[$31>>2]|0; //@line 105 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $33 = (($32) + (($3*12)|0)|0); //@line 105 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $34 = HEAP32[$33>>2]|0; //@line 105 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $35 = (((($32) + (($3*12)|0)|0)) + 4|0); //@line 105 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $36 = HEAP32[$35>>2]|0; //@line 105 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $37 = (((($32) + (($3*12)|0)|0)) + 8|0); //@line 106 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $38 = +HEAPF32[$37>>2]; //@line 106 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $39 = (+Math_sin((+$30))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $40 = (+Math_cos((+$30))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $41 = ((($1)) + 116|0); //@line 541 "./libraries/Box2D/Dynamics/b2Body.h"
 $42 = +HEAPF32[$41>>2]; //@line 541 "./libraries/Box2D/Dynamics/b2Body.h"
 $43 = ((($this)) + 84|0); //@line 113 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $44 = +HEAPF32[$43>>2]; //@line 113 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $45 = $44 * 6.2831854820251465; //@line 113 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $46 = $42 * 2.0; //@line 116 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $47 = ((($this)) + 88|0); //@line 116 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $48 = +HEAPF32[$47>>2]; //@line 116 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $49 = $46 * $48; //@line 116 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $50 = $45 * $49; //@line 116 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $51 = $45 * $45; //@line 119 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $52 = $42 * $51; //@line 119 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $53 = +HEAPF32[$data>>2]; //@line 124 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $54 = $53 * $52; //@line 125 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $55 = $50 + $54; //@line 125 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $56 = $55 > 1.1920928955078125E-7; //@line 125 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 __Z10loveAssertbPKc($56,362634); //@line 125 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $57 = $53 * $55; //@line 126 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $58 = ((($this)) + 108|0); //@line 126 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $59 = $57 != 0.0; //@line 127 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $60 = 1.0 / $57; //@line 129 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $storemerge = $59 ? $60 : $57; //@line 127 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$58>>2] = $storemerge; //@line 126 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $61 = $54 * $storemerge; //@line 131 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $62 = ((($this)) + 92|0); //@line 131 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$62>>2] = $61; //@line 131 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $63 = ((($this)) + 68|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $64 = +HEAPF32[$63>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $65 = +HEAPF32[$5>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $66 = $64 - $65; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $67 = ((($this)) + 72|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $68 = +HEAPF32[$67>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $69 = ((($this)) + 132|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $70 = +HEAPF32[$69>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $71 = $68 - $70; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $72 = $40 * $66; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $73 = $39 * $71; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $74 = $72 - $73; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $75 = $39 * $66; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $76 = $40 * $71; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $77 = $75 + $76; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $78 = ((($this)) + 120|0); //@line 134 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$78>>2] = $74; //@line 134 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $79 = ((($this)) + 124|0); //@line 134 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$79>>2] = $77; //@line 134 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $80 = +HEAPF32[$19>>2]; //@line 156 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $81 = +HEAPF32[$22>>2]; //@line 157 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $82 = $77 * $81; //@line 140 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $83 = $77 * $82; //@line 140 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $84 = $80 + $83; //@line 140 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $85 = $storemerge + $84; //@line 140 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $86 = $81 * $74; //@line 141 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $87 = $77 * $86; //@line 141 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $88 = $74 * $86; //@line 143 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $89 = $80 + $88; //@line 143 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $90 = $storemerge + $89; //@line 143 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $91 = $85 * $90; //@line 227 "./libraries/Box2D/Common/b2Math.h"
 $92 = $87 * $87; //@line 227 "./libraries/Box2D/Common/b2Math.h"
 $93 = $91 - $92; //@line 227 "./libraries/Box2D/Common/b2Math.h"
 $94 = $93 != 0.0; //@line 228 "./libraries/Box2D/Common/b2Math.h"
 $95 = 1.0 / $93; //@line 230 "./libraries/Box2D/Common/b2Math.h"
 $det$0$i = $94 ? $95 : $93; //@line 228 "./libraries/Box2D/Common/b2Math.h"
 $96 = $90 * $det$0$i; //@line 232 "./libraries/Box2D/Common/b2Math.h"
 $97 = $87 * $det$0$i; //@line 232 "./libraries/Box2D/Common/b2Math.h"
 $98 = $85 * $det$0$i; //@line 233 "./libraries/Box2D/Common/b2Math.h"
 $99 = ((($this)) + 144|0); //@line 145 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$99>>2] = $96; //@line 145 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $100 = ((($this)) + 148|0); //@line 145 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$100>>2] = $97; //@line 145 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $101 = ((($this)) + 152|0); //@line 145 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$101>>2] = $97; //@line 145 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $102 = ((($this)) + 156|0); //@line 145 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$102>>2] = $98; //@line 145 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $103 = $26 + $74; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $104 = $28 + $77; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $105 = ((($this)) + 76|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $106 = +HEAPF32[$105>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $107 = $103 - $106; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $108 = ((($this)) + 80|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $109 = +HEAPF32[$108>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $110 = $104 - $109; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $111 = ((($this)) + 160|0); //@line 147 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $112 = ((($this)) + 164|0); //@line 147 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $113 = $107 * $61; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$111>>2] = $113; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 $114 = $110 * $61; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$112>>2] = $114; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 $115 = $38 * 0.98000001907348633; //@line 151 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $116 = ((($data)) + 20|0); //@line 153 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $117 = HEAP8[$116>>0]|0; //@line 153 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $118 = ($117<<24>>24)==(0); //@line 153 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $119 = ((($this)) + 96|0); //@line 155 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 if ($118) {
  HEAPF32[$119>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $140 = ((($this)) + 100|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$140>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $vB$sroa$0$0 = $34;$vB$sroa$6$0 = $36;$wB$0 = $115;
  $141 = HEAP32[$4>>2]|0; //@line 164 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  $142 = HEAP32[$31>>2]|0; //@line 164 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  $143 = (($142) + (($141*12)|0)|0); //@line 164 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  HEAP32[$143>>2] = $vB$sroa$0$0; //@line 164 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  $144 = (((($142) + (($141*12)|0)|0)) + 4|0); //@line 164 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  HEAP32[$144>>2] = $vB$sroa$6$0; //@line 164 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  $145 = HEAP32[$4>>2]|0; //@line 165 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  $146 = HEAP32[$31>>2]|0; //@line 165 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  $147 = (((($146) + (($145*12)|0)|0)) + 8|0); //@line 165 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  HEAPF32[$147>>2] = $wB$0; //@line 165 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  return; //@line 166 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 } else {
  $120 = ((($data)) + 8|0); //@line 155 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  $121 = +HEAPF32[$120>>2]; //@line 155 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  $122 = +HEAPF32[$119>>2]; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $123 = $121 * $122; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$119>>2] = $123; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $124 = ((($this)) + 100|0); //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $125 = +HEAPF32[$124>>2]; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $126 = $121 * $125; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$124>>2] = $126; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $127 = $123 * $80; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $128 = $126 * $80; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $129 = (HEAP32[tempDoublePtr>>2]=$34,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $130 = $129 + $127; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $131 = (HEAPF32[tempDoublePtr>>2]=$130,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $132 = (HEAP32[tempDoublePtr>>2]=$36,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $133 = $132 + $128; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $134 = (HEAPF32[tempDoublePtr>>2]=$133,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $135 = $126 * $74; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $136 = $123 * $77; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $137 = $135 - $136; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $138 = $81 * $137; //@line 157 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  $139 = $115 + $138; //@line 157 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  $vB$sroa$0$0 = $131;$vB$sroa$6$0 = $134;$wB$0 = $139;
  $141 = HEAP32[$4>>2]|0; //@line 164 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  $142 = HEAP32[$31>>2]|0; //@line 164 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  $143 = (($142) + (($141*12)|0)|0); //@line 164 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  HEAP32[$143>>2] = $vB$sroa$0$0; //@line 164 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  $144 = (((($142) + (($141*12)|0)|0)) + 4|0); //@line 164 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  HEAP32[$144>>2] = $vB$sroa$6$0; //@line 164 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  $145 = HEAP32[$4>>2]|0; //@line 165 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  $146 = HEAP32[$31>>2]|0; //@line 165 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  $147 = (((($146) + (($145*12)|0)|0)) + 8|0); //@line 165 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  HEAPF32[$147>>2] = $wB$0; //@line 165 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  return; //@line 166 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 }
}
function __ZN12b2MouseJoint24SolveVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0.0;
 var $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0.0;
 var $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0, $7 = 0.0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0;
 var $81 = 0.0, $82 = 0, $83 = 0, $84 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 116|0); //@line 170 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 189 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $2 = ((($data)) + 28|0); //@line 170 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 189 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $4 = (($3) + (($1*12)|0)|0); //@line 170 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 170 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $6 = (((($3) + (($1*12)|0)|0)) + 4|0); //@line 170 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 170 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $8 = (((($3) + (($1*12)|0)|0)) + 8|0); //@line 171 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 171 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $10 = ((($this)) + 124|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $11 = +HEAPF32[$10>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $12 = $9 * $11; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($this)) + 120|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $15 = $9 * $14; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 - $12; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $17 = $7 + $15; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $18 = ((($this)) + 160|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $19 = +HEAPF32[$18>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $20 = $16 + $19; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($this)) + 164|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $22 = +HEAPF32[$21>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $23 = $17 + $22; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($this)) + 108|0); //@line 175 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $25 = +HEAPF32[$24>>2]; //@line 175 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $26 = ((($this)) + 96|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $27 = +HEAPF32[$26>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $28 = $25 * $27; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $29 = ((($this)) + 100|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $30 = +HEAPF32[$29>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $31 = $25 * $30; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $32 = $20 + $28; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $33 = $23 + $31; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $34 = -$32; //@line 68 "./libraries/Box2D/Common/b2Math.h"
 $35 = -$33; //@line 68 "./libraries/Box2D/Common/b2Math.h"
 $36 = ((($this)) + 144|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $37 = +HEAPF32[$36>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $38 = $37 * $34; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $39 = ((($this)) + 152|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $40 = +HEAPF32[$39>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $41 = $40 * $35; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $42 = $38 + $41; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $43 = ((($this)) + 148|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $44 = +HEAPF32[$43>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $45 = $44 * $34; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $46 = ((($this)) + 156|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $47 = +HEAPF32[$46>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $48 = $47 * $35; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $49 = $45 + $48; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $50 = $27 + $42; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$26>>2] = $50; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $51 = $49 + $30; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$29>>2] = $51; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $52 = +HEAPF32[$data>>2]; //@line 179 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $53 = ((($this)) + 104|0); //@line 179 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $54 = +HEAPF32[$53>>2]; //@line 179 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $55 = $52 * $54; //@line 179 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $56 = $50 * $50; //@line 110 "./libraries/Box2D/Common/b2Math.h"
 $57 = $51 * $51; //@line 110 "./libraries/Box2D/Common/b2Math.h"
 $58 = $57 + $56; //@line 110 "./libraries/Box2D/Common/b2Math.h"
 $59 = $55 * $55; //@line 180 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $60 = $58 > $59; //@line 180 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 if ($60) {
  $61 = (+Math_sqrt((+$58))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $62 = $55 / $61; //@line 182 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
  $63 = $50 * $62; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$26>>2] = $63; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $64 = $62 * $51; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$29>>2] = $64; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $66 = $63;$68 = $64;
 } else {
  $66 = $50;$68 = $51;
 }
 $65 = $66 - $27; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $67 = $68 - $30; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $69 = ((($this)) + 136|0); //@line 186 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $70 = +HEAPF32[$69>>2]; //@line 186 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $71 = $65 * $70; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $72 = $67 * $70; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $73 = $5 + $71; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $74 = $7 + $72; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $75 = ((($this)) + 140|0); //@line 187 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $76 = +HEAPF32[$75>>2]; //@line 187 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $77 = $67 * $14; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $78 = $65 * $11; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $79 = $77 - $78; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $80 = $76 * $79; //@line 187 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $81 = $9 + $80; //@line 187 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$4>>2] = $73; //@line 189 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$6>>2] = $74; //@line 189 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $82 = HEAP32[$0>>2]|0; //@line 190 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $83 = HEAP32[$2>>2]|0; //@line 190 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $84 = (((($83) + (($82*12)|0)|0)) + 8|0); //@line 190 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$84>>2] = $81; //@line 190 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 return; //@line 191 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
}
function __ZN12b2MouseJoint24SolvePositionConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1; //@line 196 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
}
function __ZN12b2MouseJointC2EPK15b2MouseJointDef($this,$def) {
 $this = $this|0;
 $def = $def|0;
 var $$ = 0, $$1 = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0.0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0.0;
 var $42 = 0.0, $43 = 0.0, $44 = 0, $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $bitcast = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN7b2JointC2EPK10b2JointDef($this,$def); //@line 32 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAP32[$this>>2] = (20816); //@line 33 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $0 = ((($this)) + 76|0); //@line 31 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $1 = ((($def)) + 20|0); //@line 34 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 131 "./libraries/Box2D/Common/b2Math.h"
 $3 = $2 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $4 = ($3|0)==(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
 if ($4) {
  $9 = 0;
 } else {
  $5 = ((($def)) + 24|0); //@line 131 "./libraries/Box2D/Common/b2Math.h"
  $6 = HEAP32[$5>>2]|0; //@line 131 "./libraries/Box2D/Common/b2Math.h"
  $7 = $6 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
  $8 = ($7|0)!=(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
  $9 = $8;
 }
 __Z10loveAssertbPKc($9,362657); //@line 34 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $10 = ((($def)) + 28|0); //@line 35 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $11 = +HEAPF32[$10>>2]; //@line 35 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $12 = (HEAPF32[tempDoublePtr>>2]=$11,HEAP32[tempDoublePtr>>2]|0); //@line 35 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $13 = $12 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $14 = ($13|0)!=(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $15 = $11 >= 0.0; //@line 35 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $$ = $15 & $14; //@line 35 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 __Z10loveAssertbPKc($$,362679); //@line 35 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $16 = ((($def)) + 32|0); //@line 36 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 36 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $18 = (HEAPF32[tempDoublePtr>>2]=$17,HEAP32[tempDoublePtr>>2]|0); //@line 36 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $19 = $18 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $20 = ($19|0)!=(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $21 = $17 >= 0.0; //@line 36 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $$1 = $21 & $20; //@line 36 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 __Z10loveAssertbPKc($$1,362729); //@line 36 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $22 = ((($def)) + 36|0); //@line 37 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $23 = +HEAPF32[$22>>2]; //@line 37 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $24 = (HEAPF32[tempDoublePtr>>2]=$23,HEAP32[tempDoublePtr>>2]|0); //@line 37 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $25 = $24 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $26 = ($25|0)!=(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $27 = $23 >= 0.0; //@line 37 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $$2 = $27 & $26; //@line 37 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 __Z10loveAssertbPKc($$2,362785); //@line 37 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $28 = $1; //@line 39 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $29 = $28; //@line 39 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $30 = HEAP32[$29>>2]|0; //@line 39 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $31 = (($28) + 4)|0; //@line 39 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $32 = $31; //@line 39 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $33 = HEAP32[$32>>2]|0; //@line 39 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $34 = $0; //@line 39 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $35 = $34; //@line 39 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAP32[$35>>2] = $30; //@line 39 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $36 = (($34) + 4)|0; //@line 39 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $37 = $36; //@line 39 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAP32[$37>>2] = $33; //@line 39 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $38 = ((($this)) + 52|0); //@line 40 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $39 = HEAP32[$38>>2]|0; //@line 40 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $bitcast = (HEAP32[tempDoublePtr>>2]=$30,+HEAPF32[tempDoublePtr>>2]); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $40 = ((($39)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $41 = +HEAPF32[$40>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $42 = $bitcast - $41; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $43 = (HEAP32[tempDoublePtr>>2]=$33,+HEAPF32[tempDoublePtr>>2]); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $44 = ((($39)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $45 = +HEAPF32[$44>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $46 = $43 - $45; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $47 = ((($39)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $48 = +HEAPF32[$47>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $49 = $42 * $48; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $50 = ((($39)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $51 = +HEAPF32[$50>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $52 = $46 * $51; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $53 = $49 + $52; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $54 = $42 * $51; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $55 = $48 * $46; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $56 = $55 - $54; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $57 = ((($this)) + 68|0); //@line 40 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$57>>2] = $53; //@line 40 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $58 = ((($this)) + 72|0); //@line 40 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$58>>2] = $56; //@line 40 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $59 = HEAP32[$10>>2]|0; //@line 42 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $60 = ((($this)) + 104|0); //@line 42 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAP32[$60>>2] = $59; //@line 42 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $61 = ((($this)) + 96|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$61>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $62 = ((($this)) + 100|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$62>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $63 = HEAP32[$16>>2]|0; //@line 45 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $64 = ((($this)) + 84|0); //@line 45 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAP32[$64>>2] = $63; //@line 45 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $65 = HEAP32[$22>>2]|0; //@line 46 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $66 = ((($this)) + 88|0); //@line 46 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAP32[$66>>2] = $65; //@line 46 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $67 = ((($this)) + 92|0); //@line 48 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$67>>2] = 0.0; //@line 48 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $68 = ((($this)) + 108|0); //@line 49 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$68>>2] = 0.0; //@line 49 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 return; //@line 50 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
}
function __ZN12b2MouseJoint9SetTargetERK6b2Vec2($this,$target) {
 $this = $this|0;
 $target = $target|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 54 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 54 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $2 = ((($1)) + 4|0); //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3 & 2; //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = ($4<<16>>16)==(0); //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($5) {
  $6 = $3&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $7 = $6 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $8 = ($7|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  if ($8) {
   $9 = $6 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $10 = $9&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAP16[$2>>1] = $10; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $11 = ((($1)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAPF32[$11>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  }
 }
 $12 = ((($this)) + 76|0); //@line 58 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $13 = $target; //@line 58 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $14 = $13; //@line 58 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 58 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $16 = (($13) + 4)|0; //@line 58 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $17 = $16; //@line 58 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 58 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $19 = $12; //@line 58 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $20 = $19; //@line 58 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAP32[$20>>2] = $15; //@line 58 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $21 = (($19) + 4)|0; //@line 58 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $22 = $21; //@line 58 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAP32[$22>>2] = $18; //@line 58 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 return; //@line 59 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
}
function __ZNK12b2MouseJoint9GetTargetEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 76|0); //@line 63 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 return ($0|0); //@line 63 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
}
function __ZN12b2MouseJoint11SetMaxForceEf($this,$force) {
 $this = $this|0;
 $force = +$force;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 104|0); //@line 68 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$0>>2] = $force; //@line 68 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 return; //@line 69 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
}
function __ZNK12b2MouseJoint11GetMaxForceEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 104|0); //@line 73 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 73 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 return (+$1); //@line 73 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
}
function __ZN12b2MouseJoint12SetFrequencyEf($this,$hz) {
 $this = $this|0;
 $hz = +$hz;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 84|0); //@line 78 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$0>>2] = $hz; //@line 78 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 return; //@line 79 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
}
function __ZNK12b2MouseJoint12GetFrequencyEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 84|0); //@line 83 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 83 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 return (+$1); //@line 83 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
}
function __ZN12b2MouseJoint15SetDampingRatioEf($this,$ratio) {
 $this = $this|0;
 $ratio = +$ratio;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 88|0); //@line 88 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 HEAPF32[$0>>2] = $ratio; //@line 88 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 return; //@line 89 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
}
function __ZNK12b2MouseJoint15GetDampingRatioEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 88|0); //@line 93 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 93 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
 return (+$1); //@line 93 "libraries/Box2D/Dynamics/Joints/b2MouseJoint.cpp"
}
function __ZNK16b2PrismaticJoint10GetAnchorAEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 489 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 489 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 68|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 72|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 489 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZNK16b2PrismaticJoint10GetAnchorBEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 494 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 494 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 76|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 80|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 494 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZNK16b2PrismaticJoint16GetReactionForceEf($agg$result,$this,$inv_dt) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0, $3 = 0.0, $4 = 0.0, $5 = 0;
 var $6 = 0.0, $7 = 0.0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 104|0); //@line 499 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 499 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = ((($this)) + 192|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $4 = $1 * $3; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($this)) + 196|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $7 = $1 * $6; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($this)) + 116|0); //@line 499 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 499 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $10 = ((($this)) + 112|0); //@line 499 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $11 = +HEAPF32[$10>>2]; //@line 499 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $12 = $9 + $11; //@line 499 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $13 = ((($this)) + 184|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 * $12; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $16 = ((($this)) + 188|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $17 = +HEAPF32[$16>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $18 = $12 * $17; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $19 = $4 + $15; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $20 = $7 + $18; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $21 = $19 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $22 = $20 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $23 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$23>>2] = $22; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 499 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZNK16b2PrismaticJoint17GetReactionTorqueEf($this,$inv_dt) {
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 108|0); //@line 504 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 504 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = $1 * $inv_dt; //@line 504 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 return (+$2); //@line 504 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZN16b2PrismaticJoint4DumpEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer13 = 0, $vararg_buffer17 = 0, $vararg_buffer21 = 0, $vararg_buffer24 = 0, $vararg_buffer27 = 0, $vararg_buffer3 = 0;
 var $vararg_buffer30 = 0, $vararg_buffer33 = 0, $vararg_buffer36 = 0, $vararg_buffer39 = 0, $vararg_buffer42 = 0, $vararg_buffer6 = 0, $vararg_buffer9 = 0, $vararg_ptr12 = 0, $vararg_ptr16 = 0, $vararg_ptr20 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer42 = sp + 136|0;
 $vararg_buffer39 = sp + 128|0;
 $vararg_buffer36 = sp + 120|0;
 $vararg_buffer33 = sp + 112|0;
 $vararg_buffer30 = sp + 104|0;
 $vararg_buffer27 = sp + 96|0;
 $vararg_buffer24 = sp + 88|0;
 $vararg_buffer21 = sp + 80|0;
 $vararg_buffer17 = sp + 64|0;
 $vararg_buffer13 = sp + 48|0;
 $vararg_buffer9 = sp + 32|0;
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($this)) + 48|0); //@line 611 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 611 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 611 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 611 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $4 = ((($this)) + 52|0); //@line 612 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 612 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $6 = ((($5)) + 8|0); //@line 612 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 612 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 __Z5b2LogPKcz(362862,$vararg_buffer); //@line 614 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$vararg_buffer1>>2] = $3; //@line 615 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 __Z5b2LogPKcz(447349,$vararg_buffer1); //@line 615 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$vararg_buffer3>>2] = $7; //@line 616 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 __Z5b2LogPKcz(447375,$vararg_buffer3); //@line 616 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $8 = ((($this)) + 61|0); //@line 617 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 617 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $10 = $9&255; //@line 617 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$vararg_buffer6>>2] = $10; //@line 617 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 __Z5b2LogPKcz(447401,$vararg_buffer6); //@line 617 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $11 = ((($this)) + 68|0); //@line 618 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $12 = +HEAPF32[$11>>2]; //@line 618 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $13 = $12; //@line 618 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $14 = ((($this)) + 72|0); //@line 618 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 618 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $16 = $15; //@line 618 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF64[$vararg_buffer9>>3] = $13; //@line 618 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $vararg_ptr12 = ((($vararg_buffer9)) + 8|0); //@line 618 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF64[$vararg_ptr12>>3] = $16; //@line 618 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 __Z5b2LogPKcz(447106,$vararg_buffer9); //@line 618 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $17 = ((($this)) + 76|0); //@line 619 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $18 = +HEAPF32[$17>>2]; //@line 619 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $19 = $18; //@line 619 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $20 = ((($this)) + 80|0); //@line 619 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 619 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $22 = $21; //@line 619 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF64[$vararg_buffer13>>3] = $19; //@line 619 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $vararg_ptr16 = ((($vararg_buffer13)) + 8|0); //@line 619 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF64[$vararg_ptr16>>3] = $22; //@line 619 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 __Z5b2LogPKcz(447148,$vararg_buffer13); //@line 619 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $23 = ((($this)) + 84|0); //@line 620 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $24 = +HEAPF32[$23>>2]; //@line 620 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $25 = $24; //@line 620 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $26 = ((($this)) + 88|0); //@line 620 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $27 = +HEAPF32[$26>>2]; //@line 620 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $28 = $27; //@line 620 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF64[$vararg_buffer17>>3] = $25; //@line 620 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $vararg_ptr20 = ((($vararg_buffer17)) + 8|0); //@line 620 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF64[$vararg_ptr20>>3] = $28; //@line 620 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 __Z5b2LogPKcz(363516,$vararg_buffer17); //@line 620 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $29 = ((($this)) + 100|0); //@line 621 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $30 = +HEAPF32[$29>>2]; //@line 621 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $31 = $30; //@line 621 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF64[$vararg_buffer21>>3] = $31; //@line 621 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 __Z5b2LogPKcz(363446,$vararg_buffer21); //@line 621 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $32 = ((($this)) + 136|0); //@line 622 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $33 = HEAP8[$32>>0]|0; //@line 622 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $34 = $33&255; //@line 622 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$vararg_buffer24>>2] = $34; //@line 622 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 __Z5b2LogPKcz(363246,$vararg_buffer24); //@line 622 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $35 = ((($this)) + 120|0); //@line 623 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $36 = +HEAPF32[$35>>2]; //@line 623 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $37 = $36; //@line 623 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF64[$vararg_buffer27>>3] = $37; //@line 623 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 __Z5b2LogPKcz(362889,$vararg_buffer27); //@line 623 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $38 = ((($this)) + 124|0); //@line 624 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $39 = +HEAPF32[$38>>2]; //@line 624 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $40 = $39; //@line 624 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF64[$vararg_buffer30>>3] = $40; //@line 624 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 __Z5b2LogPKcz(362923,$vararg_buffer30); //@line 624 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $41 = ((($this)) + 137|0); //@line 625 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $42 = HEAP8[$41>>0]|0; //@line 625 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $43 = $42&255; //@line 625 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$vararg_buffer33>>2] = $43; //@line 625 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 __Z5b2LogPKcz(363556,$vararg_buffer33); //@line 625 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $44 = ((($this)) + 132|0); //@line 626 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $45 = +HEAPF32[$44>>2]; //@line 626 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $46 = $45; //@line 626 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF64[$vararg_buffer36>>3] = $46; //@line 626 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 __Z5b2LogPKcz(363586,$vararg_buffer36); //@line 626 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $47 = ((($this)) + 128|0); //@line 627 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $48 = +HEAPF32[$47>>2]; //@line 627 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $49 = $48; //@line 627 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF64[$vararg_buffer39>>3] = $49; //@line 627 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 __Z5b2LogPKcz(362957,$vararg_buffer39); //@line 627 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $50 = ((($this)) + 56|0); //@line 628 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $51 = HEAP32[$50>>2]|0; //@line 628 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$vararg_buffer42>>2] = $51; //@line 628 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 __Z5b2LogPKcz(447513,$vararg_buffer42); //@line 628 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 STACKTOP = sp;return; //@line 629 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZN16b2PrismaticJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 86 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
 return; //@line 86 "./libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.h"
}
function __ZN16b2PrismaticJoint23InitVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0, $115 = 0.0;
 var $116 = 0.0, $117 = 0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0, $125 = 0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0, $132 = 0.0, $133 = 0.0;
 var $134 = 0.0, $135 = 0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0, $144 = 0, $145 = 0.0, $146 = 0, $147 = 0.0, $148 = 0.0, $149 = 0, $15 = 0, $150 = 0.0, $151 = 0.0;
 var $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0, $157 = 0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0.0, $17 = 0;
 var $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0, $178 = 0.0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0.0, $193 = 0.0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0, $201 = 0.0, $202 = 0.0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0, $229 = 0.0, $23 = 0, $230 = 0.0, $231 = 0, $232 = 0.0, $233 = 0.0, $234 = 0, $235 = 0.0, $236 = 0.0, $237 = 0.0, $238 = 0.0, $239 = 0.0, $24 = 0, $240 = 0.0, $241 = 0.0;
 var $242 = 0.0, $243 = 0.0, $244 = 0.0, $245 = 0.0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0.0, $254 = 0.0, $255 = 0.0, $256 = 0, $257 = 0.0, $258 = 0.0, $259 = 0, $26 = 0;
 var $260 = 0.0, $261 = 0.0, $262 = 0.0, $263 = 0.0, $264 = 0.0, $265 = 0.0, $266 = 0, $267 = 0.0, $268 = 0.0, $269 = 0, $27 = 0, $270 = 0.0, $271 = 0.0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0, $53 = 0.0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0;
 var $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0, $92 = 0.0, $93 = 0.0, $94 = 0;
 var $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $bitcast = 0.0, $bitcast29 = 0.0, $k22$0 = 0.0, $storemerge = 0.0, $vA$sroa$0$0 = 0, $vA$sroa$6$0 = 0, $vB$sroa$0$0 = 0, $vB$sroa$6$0 = 0, $wA$0 = 0.0, $wB$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 128 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 128 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 128 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 128 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $4 = ((($this)) + 144|0); //@line 128 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$4>>2] = $3; //@line 128 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $5 = ((($this)) + 52|0); //@line 129 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 129 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $7 = ((($6)) + 8|0); //@line 129 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 129 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $9 = ((($this)) + 148|0); //@line 129 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$9>>2] = $8; //@line 129 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $10 = ((($this)) + 152|0); //@line 130 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $11 = ((($1)) + 28|0); //@line 130 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $12 = $11; //@line 130 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $13 = $12; //@line 130 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 130 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $15 = (($12) + 4)|0; //@line 130 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $16 = $15; //@line 130 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 130 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $18 = $10; //@line 130 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $19 = $18; //@line 130 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$19>>2] = $14; //@line 130 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $20 = (($18) + 4)|0; //@line 130 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $21 = $20; //@line 130 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$21>>2] = $17; //@line 130 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $22 = ((($this)) + 160|0); //@line 131 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $23 = ((($6)) + 28|0); //@line 131 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $24 = $23; //@line 131 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $25 = $24; //@line 131 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 131 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $27 = (($24) + 4)|0; //@line 131 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $28 = $27; //@line 131 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 131 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $30 = $22; //@line 131 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $31 = $30; //@line 131 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$31>>2] = $26; //@line 131 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $32 = (($30) + 4)|0; //@line 131 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $33 = $32; //@line 131 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$33>>2] = $29; //@line 131 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $34 = ((($1)) + 120|0); //@line 132 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 132 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $36 = ((($this)) + 168|0); //@line 132 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$36>>2] = $35; //@line 132 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $37 = ((($6)) + 120|0); //@line 133 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $38 = HEAP32[$37>>2]|0; //@line 133 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $39 = ((($this)) + 172|0); //@line 133 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$39>>2] = $38; //@line 133 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $40 = ((($1)) + 128|0); //@line 134 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $41 = HEAP32[$40>>2]|0; //@line 134 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $42 = ((($this)) + 176|0); //@line 134 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$42>>2] = $41; //@line 134 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $43 = ((($6)) + 128|0); //@line 135 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $44 = HEAP32[$43>>2]|0; //@line 135 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $45 = ((($this)) + 180|0); //@line 135 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$45>>2] = $44; //@line 135 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $46 = ((($data)) + 24|0); //@line 137 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $47 = HEAP32[$46>>2]|0; //@line 137 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $48 = (($47) + (($3*12)|0)|0); //@line 137 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $49 = +HEAPF32[$48>>2]; //@line 137 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $50 = (((($47) + (($3*12)|0)|0)) + 4|0); //@line 137 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $51 = +HEAPF32[$50>>2]; //@line 137 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $52 = (((($47) + (($3*12)|0)|0)) + 8|0); //@line 138 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $53 = +HEAPF32[$52>>2]; //@line 138 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $54 = ((($data)) + 28|0); //@line 139 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $55 = HEAP32[$54>>2]|0; //@line 139 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $56 = (($55) + (($3*12)|0)|0); //@line 139 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $57 = HEAP32[$56>>2]|0; //@line 139 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $58 = (((($55) + (($3*12)|0)|0)) + 4|0); //@line 139 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $59 = HEAP32[$58>>2]|0; //@line 139 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $60 = (((($55) + (($3*12)|0)|0)) + 8|0); //@line 140 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $61 = +HEAPF32[$60>>2]; //@line 140 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $62 = (($47) + (($8*12)|0)|0); //@line 142 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $63 = +HEAPF32[$62>>2]; //@line 142 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $64 = (((($47) + (($8*12)|0)|0)) + 4|0); //@line 142 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $65 = +HEAPF32[$64>>2]; //@line 142 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $66 = (((($47) + (($8*12)|0)|0)) + 8|0); //@line 143 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $67 = +HEAPF32[$66>>2]; //@line 143 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $68 = (($55) + (($8*12)|0)|0); //@line 144 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $69 = HEAP32[$68>>2]|0; //@line 144 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $70 = (((($55) + (($8*12)|0)|0)) + 4|0); //@line 144 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $71 = HEAP32[$70>>2]|0; //@line 144 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $72 = (((($55) + (($8*12)|0)|0)) + 8|0); //@line 145 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $73 = +HEAPF32[$72>>2]; //@line 145 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $74 = (+Math_sin((+$53))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $75 = (+Math_cos((+$53))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $76 = (+Math_sin((+$67))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $77 = (+Math_cos((+$67))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $78 = ((($this)) + 68|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $79 = +HEAPF32[$78>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $bitcast = (HEAP32[tempDoublePtr>>2]=$14,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $80 = $79 - $bitcast; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $81 = ((($this)) + 72|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $82 = +HEAPF32[$81>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $83 = (HEAP32[tempDoublePtr>>2]=$17,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $84 = $82 - $83; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $85 = $75 * $80; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $86 = $74 * $84; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $87 = $85 - $86; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $88 = $74 * $80; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $89 = $75 * $84; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $90 = $88 + $89; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $91 = ((($this)) + 76|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $92 = +HEAPF32[$91>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $bitcast29 = (HEAP32[tempDoublePtr>>2]=$26,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $93 = $92 - $bitcast29; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $94 = ((($this)) + 80|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $95 = +HEAPF32[$94>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $96 = (HEAP32[tempDoublePtr>>2]=$29,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $97 = $95 - $96; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $98 = $77 * $93; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $99 = $76 * $97; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $100 = $98 - $99; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $101 = $76 * $93; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $102 = $77 * $97; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $103 = $101 + $102; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $104 = $63 - $49; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $105 = $65 - $51; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $106 = $104 + $100; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $107 = $105 + $103; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $108 = $106 - $87; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $109 = $107 - $90; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $110 = (HEAP32[tempDoublePtr>>2]=$35,+HEAPF32[tempDoublePtr>>2]); //@line 154 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $111 = (HEAP32[tempDoublePtr>>2]=$38,+HEAPF32[tempDoublePtr>>2]); //@line 154 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $112 = (HEAP32[tempDoublePtr>>2]=$41,+HEAPF32[tempDoublePtr>>2]); //@line 155 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $113 = (HEAP32[tempDoublePtr>>2]=$44,+HEAPF32[tempDoublePtr>>2]); //@line 155 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $114 = ((($this)) + 84|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $115 = +HEAPF32[$114>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $116 = $75 * $115; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $117 = ((($this)) + 88|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $118 = +HEAPF32[$117>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $119 = $74 * $118; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $120 = $116 - $119; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $121 = $74 * $115; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $122 = $75 * $118; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $123 = $121 + $122; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $124 = ((($this)) + 184|0); //@line 159 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$124>>2] = $120; //@line 159 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $125 = ((($this)) + 188|0); //@line 159 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$125>>2] = $123; //@line 159 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $126 = $87 + $108; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $127 = $90 + $109; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $128 = $126 * $123; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $129 = $127 * $120; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $130 = $128 - $129; //@line 245 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $131 = ((($this)) + 208|0); //@line 160 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$131>>2] = $130; //@line 160 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $132 = $100 * $123; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $133 = $103 * $120; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $134 = $132 - $133; //@line 246 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $135 = ((($this)) + 212|0); //@line 161 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$135>>2] = $134; //@line 161 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $136 = $110 + $111; //@line 163 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $137 = $112 * $130; //@line 163 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $138 = $130 * $137; //@line 163 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $139 = $136 + $138; //@line 163 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $140 = $113 * $134; //@line 163 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $141 = $134 * $140; //@line 163 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $142 = $141 + $139; //@line 163 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $143 = ((($this)) + 252|0); //@line 163 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $144 = $142 > 0.0; //@line 164 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $145 = 1.0 / $142; //@line 166 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $storemerge = $144 ? $145 : $142; //@line 164 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$143>>2] = $storemerge; //@line 163 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $146 = ((($this)) + 92|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $147 = +HEAPF32[$146>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $148 = $75 * $147; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $149 = ((($this)) + 96|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $150 = +HEAPF32[$149>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $151 = $74 * $150; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $152 = $148 - $151; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $153 = $74 * $147; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $154 = $75 * $150; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $155 = $153 + $154; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $156 = ((($this)) + 192|0); //@line 172 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$156>>2] = $152; //@line 172 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $157 = ((($this)) + 196|0); //@line 172 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$157>>2] = $155; //@line 172 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $158 = $126 * $155; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $159 = $127 * $152; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $160 = $158 - $159; //@line 245 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $161 = ((($this)) + 200|0); //@line 174 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$161>>2] = $160; //@line 174 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $162 = $100 * $155; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $163 = $103 * $152; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $164 = $162 - $163; //@line 246 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $165 = ((($this)) + 204|0); //@line 175 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$165>>2] = $164; //@line 175 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $166 = $112 * $160; //@line 180 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $167 = $160 * $166; //@line 180 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $168 = $136 + $167; //@line 180 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $169 = $113 * $164; //@line 180 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $170 = $164 * $169; //@line 180 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $171 = $170 + $168; //@line 180 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $172 = $166 + $169; //@line 181 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $173 = $166 * $130; //@line 182 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $174 = $169 * $134; //@line 182 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $175 = $173 + $174; //@line 182 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $176 = $112 + $113; //@line 183 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $177 = $176 == 0.0; //@line 184 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $k22$0 = $177 ? 1.0 : $176; //@line 184 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $178 = $137 + $140; //@line 189 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $179 = ((($this)) + 216|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$179>>2] = $171; //@line 156 "./libraries/Box2D/Common/b2Math.h"
 $180 = ((($this)) + 220|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$180>>2] = $172; //@line 156 "./libraries/Box2D/Common/b2Math.h"
 $181 = ((($this)) + 224|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$181>>2] = $175; //@line 156 "./libraries/Box2D/Common/b2Math.h"
 $182 = ((($this)) + 228|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$182>>2] = $172; //@line 156 "./libraries/Box2D/Common/b2Math.h"
 $183 = ((($this)) + 232|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$183>>2] = $k22$0; //@line 156 "./libraries/Box2D/Common/b2Math.h"
 $184 = ((($this)) + 236|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$184>>2] = $178; //@line 156 "./libraries/Box2D/Common/b2Math.h"
 $185 = ((($this)) + 240|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$185>>2] = $175; //@line 156 "./libraries/Box2D/Common/b2Math.h"
 $186 = ((($this)) + 244|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$186>>2] = $178; //@line 156 "./libraries/Box2D/Common/b2Math.h"
 $187 = ((($this)) + 248|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$187>>2] = $142; //@line 156 "./libraries/Box2D/Common/b2Math.h"
 $188 = ((($this)) + 136|0); //@line 198 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $189 = HEAP16[$188>>1]|0; //@line 198 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $190 = $189&255; //@line 198 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $191 = ($190<<24>>24)==(0); //@line 198 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 do {
  if ($191) {
   $216 = ((($this)) + 140|0); //@line 229 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   HEAP32[$216>>2] = 0; //@line 229 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $217 = ((($this)) + 112|0); //@line 230 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   HEAPF32[$217>>2] = 0.0; //@line 230 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  } else {
   $192 = $108 * $120; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $193 = $109 * $123; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $194 = $192 + $193; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $195 = ((($this)) + 124|0); //@line 201 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $196 = +HEAPF32[$195>>2]; //@line 201 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $197 = ((($this)) + 120|0); //@line 201 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $198 = +HEAPF32[$197>>2]; //@line 201 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $199 = $196 - $198; //@line 201 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $200 = $199 > 0.0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
   $201 = -$199; //@line 618 "./libraries/Box2D/Common/b2Math.h"
   $202 = $200 ? $199 : $201; //@line 618 "./libraries/Box2D/Common/b2Math.h"
   $203 = $202 < 0.0099999997764825821; //@line 201 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   if ($203) {
    $204 = ((($this)) + 140|0); //@line 203 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    HEAP32[$204>>2] = 3; //@line 203 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    break;
   }
   $205 = !($194 <= $198); //@line 205 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   if (!($205)) {
    $206 = ((($this)) + 140|0); //@line 207 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    $207 = HEAP32[$206>>2]|0; //@line 207 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    $208 = ($207|0)==(1); //@line 207 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    if ($208) {
     break;
    }
    HEAP32[$206>>2] = 1; //@line 209 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    $209 = ((($this)) + 112|0); //@line 210 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    HEAPF32[$209>>2] = 0.0; //@line 210 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    break;
   }
   $210 = !($194 >= $196); //@line 213 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $211 = ((($this)) + 140|0); //@line 215 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   if ($210) {
    HEAP32[$211>>2] = 0; //@line 223 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    $215 = ((($this)) + 112|0); //@line 224 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    HEAPF32[$215>>2] = 0.0; //@line 224 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    break;
   }
   $212 = HEAP32[$211>>2]|0; //@line 215 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $213 = ($212|0)==(2); //@line 215 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   if (!($213)) {
    HEAP32[$211>>2] = 2; //@line 217 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    $214 = ((($this)) + 112|0); //@line 218 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    HEAPF32[$214>>2] = 0.0; //@line 218 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   }
  }
 } while(0);
 $218 = ($189&65535)<(256); //@line 233 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 if ($218) {
  $219 = ((($this)) + 116|0); //@line 235 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAPF32[$219>>2] = 0.0; //@line 235 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 }
 $220 = ((($data)) + 20|0); //@line 238 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $221 = HEAP8[$220>>0]|0; //@line 238 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $222 = ($221<<24>>24)==(0); //@line 238 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $223 = ((($this)) + 104|0); //@line 241 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 if ($222) {
  ;HEAP32[$223>>2]=0|0;HEAP32[$223+4>>2]=0|0;HEAP32[$223+8>>2]=0|0;HEAP32[$223+12>>2]=0|0; //@line 153 "./libraries/Box2D/Common/b2Math.h"
  $vA$sroa$0$0 = $57;$vA$sroa$6$0 = $59;$vB$sroa$0$0 = $69;$vB$sroa$6$0 = $71;$wA$0 = $61;$wB$0 = $73;
  $272 = HEAP32[$4>>2]|0; //@line 260 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $273 = HEAP32[$54>>2]|0; //@line 260 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $274 = (($273) + (($272*12)|0)|0); //@line 260 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAP32[$274>>2] = $vA$sroa$0$0; //@line 260 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $275 = (((($273) + (($272*12)|0)|0)) + 4|0); //@line 260 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAP32[$275>>2] = $vA$sroa$6$0; //@line 260 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $276 = HEAP32[$4>>2]|0; //@line 261 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $277 = HEAP32[$54>>2]|0; //@line 261 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $278 = (((($277) + (($276*12)|0)|0)) + 8|0); //@line 261 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAPF32[$278>>2] = $wA$0; //@line 261 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $279 = HEAP32[$9>>2]|0; //@line 262 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $280 = (($277) + (($279*12)|0)|0); //@line 262 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAP32[$280>>2] = $vB$sroa$0$0; //@line 262 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $281 = (((($277) + (($279*12)|0)|0)) + 4|0); //@line 262 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAP32[$281>>2] = $vB$sroa$6$0; //@line 262 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $282 = HEAP32[$9>>2]|0; //@line 263 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $283 = HEAP32[$54>>2]|0; //@line 263 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $284 = (((($283) + (($282*12)|0)|0)) + 8|0); //@line 263 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAPF32[$284>>2] = $wB$0; //@line 263 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  return; //@line 264 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 } else {
  $224 = ((($data)) + 8|0); //@line 241 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $225 = +HEAPF32[$224>>2]; //@line 241 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $226 = +HEAPF32[$223>>2]; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $227 = $225 * $226; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$223>>2] = $227; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $228 = ((($this)) + 108|0); //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $229 = +HEAPF32[$228>>2]; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $230 = $225 * $229; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$228>>2] = $230; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $231 = ((($this)) + 112|0); //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $232 = +HEAPF32[$231>>2]; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $233 = $225 * $232; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$231>>2] = $233; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $234 = ((($this)) + 116|0); //@line 242 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $235 = +HEAPF32[$234>>2]; //@line 242 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $236 = $225 * $235; //@line 242 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAPF32[$234>>2] = $236; //@line 242 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $237 = $227 * $152; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $238 = $227 * $155; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $239 = $236 + $233; //@line 244 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $240 = $120 * $239; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $241 = $239 * $123; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $242 = $237 + $240; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $243 = $238 + $241; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $244 = $227 * $160; //@line 245 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $245 = $244 + $230; //@line 245 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $246 = $239 * $130; //@line 245 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $247 = $245 + $246; //@line 245 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $248 = $227 * $164; //@line 246 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $249 = $230 + $248; //@line 246 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $250 = $239 * $134; //@line 246 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $251 = $249 + $250; //@line 246 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $252 = $110 * $242; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $253 = $110 * $243; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $254 = (HEAP32[tempDoublePtr>>2]=$57,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $255 = $254 - $252; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $256 = (HEAPF32[tempDoublePtr>>2]=$255,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $257 = (HEAP32[tempDoublePtr>>2]=$59,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $258 = $257 - $253; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $259 = (HEAPF32[tempDoublePtr>>2]=$258,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $260 = $112 * $247; //@line 249 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $261 = $61 - $260; //@line 249 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $262 = $111 * $242; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $263 = $111 * $243; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $264 = (HEAP32[tempDoublePtr>>2]=$69,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $265 = $264 + $262; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $266 = (HEAPF32[tempDoublePtr>>2]=$265,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $267 = (HEAP32[tempDoublePtr>>2]=$71,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $268 = $267 + $263; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $269 = (HEAPF32[tempDoublePtr>>2]=$268,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $270 = $113 * $251; //@line 252 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $271 = $73 + $270; //@line 252 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $vA$sroa$0$0 = $256;$vA$sroa$6$0 = $259;$vB$sroa$0$0 = $266;$vB$sroa$6$0 = $269;$wA$0 = $261;$wB$0 = $271;
  $272 = HEAP32[$4>>2]|0; //@line 260 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $273 = HEAP32[$54>>2]|0; //@line 260 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $274 = (($273) + (($272*12)|0)|0); //@line 260 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAP32[$274>>2] = $vA$sroa$0$0; //@line 260 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $275 = (((($273) + (($272*12)|0)|0)) + 4|0); //@line 260 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAP32[$275>>2] = $vA$sroa$6$0; //@line 260 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $276 = HEAP32[$4>>2]|0; //@line 261 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $277 = HEAP32[$54>>2]|0; //@line 261 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $278 = (((($277) + (($276*12)|0)|0)) + 8|0); //@line 261 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAPF32[$278>>2] = $wA$0; //@line 261 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $279 = HEAP32[$9>>2]|0; //@line 262 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $280 = (($277) + (($279*12)|0)|0); //@line 262 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAP32[$280>>2] = $vB$sroa$0$0; //@line 262 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $281 = (((($277) + (($279*12)|0)|0)) + 4|0); //@line 262 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAP32[$281>>2] = $vB$sroa$6$0; //@line 262 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $282 = HEAP32[$9>>2]|0; //@line 263 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $283 = HEAP32[$54>>2]|0; //@line 263 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $284 = (((($283) + (($282*12)|0)|0)) + 8|0); //@line 263 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAPF32[$284>>2] = $wB$0; //@line 263 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  return; //@line 264 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 }
}
function __ZN16b2PrismaticJoint24SolveVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$pn = 0.0, $$pn2 = 0.0, $$pn3 = 0.0, $$pn35 = 0.0, $$pn36 = 0.0, $$pn37 = 0.0, $$pn38 = 0.0, $$pn39 = 0.0, $$pn4 = 0.0, $$pn40 = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0, $102 = 0.0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0;
 var $107 = 0.0, $108 = 0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0, $118 = 0, $119 = 0, $12 = 0.0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0.0;
 var $125 = 0.0, $126 = 0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0, $131 = 0.0, $132 = 0.0, $133 = 0.0, $134 = 0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0, $139 = 0.0, $14 = 0, $140 = 0, $141 = 0.0, $142 = 0;
 var $143 = 0.0, $144 = 0, $145 = 0.0, $146 = 0.0, $147 = 0.0, $148 = 0, $149 = 0, $15 = 0, $150 = 0.0, $151 = 0, $152 = 0.0, $153 = 0.0, $154 = 0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0, $159 = 0.0, $16 = 0, $160 = 0.0;
 var $161 = 0.0, $162 = 0, $163 = 0, $164 = 0.0, $165 = 0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0, $17 = 0, $170 = 0.0, $171 = 0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0, $178 = 0.0, $179 = 0.0;
 var $18 = 0, $180 = 0, $181 = 0.0, $182 = 0.0, $183 = 0.0, $184 = 0.0, $185 = 0.0, $186 = 0.0, $187 = 0.0, $188 = 0.0, $189 = 0.0, $19 = 0, $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0.0, $194 = 0.0, $195 = 0.0, $196 = 0.0, $197 = 0.0;
 var $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0.0, $200 = 0.0, $201 = 0.0, $202 = 0.0, $203 = 0.0, $204 = 0.0, $205 = 0.0, $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0, $21 = 0, $210 = 0.0, $211 = 0.0, $212 = 0, $213 = 0.0, $214 = 0;
 var $215 = 0.0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0, $22 = 0.0, $220 = 0.0, $221 = 0.0, $222 = 0.0, $223 = 0.0, $224 = 0.0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0.0, $229 = 0.0, $23 = 0, $230 = 0.0, $231 = 0.0, $232 = 0;
 var $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0.0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $25 = 0, $26 = 0.0, $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0, $49 = 0.0;
 var $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0.0, $65 = 0, $66 = 0.0, $67 = 0.0;
 var $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0, $85 = 0.0;
 var $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0, $91 = 0.0, $92 = 0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $b = 0, $df = 0, $df7 = 0, $vA$sroa$0$0 = 0, $vA$sroa$0$1$in = 0.0;
 var $vA$sroa$13$0 = 0, $vA$sroa$13$1$in = 0.0, $vB$sroa$0$0 = 0, $vB$sroa$0$1$in = 0.0, $vB$sroa$13$0 = 0, $vB$sroa$13$1$in = 0.0, $wA$0 = 0.0, $wA$1 = 0.0, $wB$0 = 0.0, $wB$1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $df = sp + 44|0;
 $0 = sp + 32|0;
 $b = sp + 24|0;
 $1 = sp + 16|0;
 $df7 = sp + 8|0;
 $2 = sp;
 $3 = ((($this)) + 144|0); //@line 268 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 268 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $5 = ((($data)) + 28|0); //@line 268 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 268 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $7 = (($6) + (($4*12)|0)|0); //@line 268 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 268 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $9 = (((($6) + (($4*12)|0)|0)) + 4|0); //@line 268 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 268 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $11 = (((($6) + (($4*12)|0)|0)) + 8|0); //@line 269 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $12 = +HEAPF32[$11>>2]; //@line 269 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $13 = ((($this)) + 148|0); //@line 270 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 270 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $15 = (($6) + (($14*12)|0)|0); //@line 270 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 270 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $17 = (((($6) + (($14*12)|0)|0)) + 4|0); //@line 270 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 270 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $19 = (((($6) + (($14*12)|0)|0)) + 8|0); //@line 271 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $20 = +HEAPF32[$19>>2]; //@line 271 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $21 = ((($this)) + 168|0); //@line 273 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $22 = +HEAPF32[$21>>2]; //@line 273 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $23 = ((($this)) + 172|0); //@line 273 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $24 = +HEAPF32[$23>>2]; //@line 273 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $25 = ((($this)) + 176|0); //@line 274 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $26 = +HEAPF32[$25>>2]; //@line 274 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $27 = ((($this)) + 180|0); //@line 274 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $28 = +HEAPF32[$27>>2]; //@line 274 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $29 = ((($this)) + 137|0); //@line 277 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $30 = HEAP8[$29>>0]|0; //@line 277 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $31 = ($30<<24>>24)==(0); //@line 277 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 if ($31) {
  $vA$sroa$0$0 = $8;$vA$sroa$13$0 = $10;$vB$sroa$0$0 = $16;$vB$sroa$13$0 = $18;$wA$0 = $12;$wB$0 = $20;
 } else {
  $32 = ((($this)) + 140|0); //@line 277 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $33 = HEAP32[$32>>2]|0; //@line 277 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $34 = ($33|0)==(3); //@line 277 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  if ($34) {
   $vA$sroa$0$0 = $8;$vA$sroa$13$0 = $10;$vB$sroa$0$0 = $16;$vB$sroa$13$0 = $18;$wA$0 = $12;$wB$0 = $20;
  } else {
   $35 = (HEAP32[tempDoublePtr>>2]=$16,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $36 = (HEAP32[tempDoublePtr>>2]=$8,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $37 = $35 - $36; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $38 = (HEAP32[tempDoublePtr>>2]=$18,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $39 = (HEAP32[tempDoublePtr>>2]=$10,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $40 = $38 - $39; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $41 = ((($this)) + 184|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $42 = +HEAPF32[$41>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $43 = $37 * $42; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $44 = ((($this)) + 188|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $45 = +HEAPF32[$44>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $46 = $40 * $45; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $47 = $43 + $46; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $48 = ((($this)) + 212|0); //@line 279 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $49 = +HEAPF32[$48>>2]; //@line 279 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $50 = $20 * $49; //@line 279 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $51 = $47 + $50; //@line 279 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $52 = ((($this)) + 208|0); //@line 279 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $53 = +HEAPF32[$52>>2]; //@line 279 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $54 = $12 * $53; //@line 279 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $55 = $51 - $54; //@line 279 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $56 = ((($this)) + 252|0); //@line 280 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $57 = +HEAPF32[$56>>2]; //@line 280 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $58 = ((($this)) + 132|0); //@line 280 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $59 = +HEAPF32[$58>>2]; //@line 280 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $60 = $59 - $55; //@line 280 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $61 = $57 * $60; //@line 280 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $62 = ((($this)) + 116|0); //@line 281 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $63 = +HEAPF32[$62>>2]; //@line 281 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $64 = +HEAPF32[$data>>2]; //@line 282 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $65 = ((($this)) + 128|0); //@line 282 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $66 = +HEAPF32[$65>>2]; //@line 282 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $67 = $64 * $66; //@line 282 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $68 = $63 + $61; //@line 283 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $69 = -$67; //@line 283 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $70 = $68 < $67; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $71 = $70 ? $68 : $67; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $72 = $71 < $69; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $73 = $72 ? $69 : $71; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$62>>2] = $73; //@line 283 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $74 = $73 - $63; //@line 284 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $75 = $42 * $74; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $76 = $45 * $74; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $77 = $53 * $74; //@line 287 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $78 = $49 * $74; //@line 288 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $79 = $22 * $75; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $80 = $22 * $76; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $81 = $36 - $79; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $82 = (HEAPF32[tempDoublePtr>>2]=$81,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $83 = $39 - $80; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $84 = (HEAPF32[tempDoublePtr>>2]=$83,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $85 = $26 * $77; //@line 291 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $86 = $12 - $85; //@line 291 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $87 = $24 * $75; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $88 = $24 * $76; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $89 = $35 + $87; //@line 85 "./libraries/Box2D/Common/b2Math.h"
   $90 = (HEAPF32[tempDoublePtr>>2]=$89,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
   $91 = $38 + $88; //@line 85 "./libraries/Box2D/Common/b2Math.h"
   $92 = (HEAPF32[tempDoublePtr>>2]=$91,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
   $93 = $28 * $78; //@line 294 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $94 = $20 + $93; //@line 294 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $vA$sroa$0$0 = $82;$vA$sroa$13$0 = $84;$vB$sroa$0$0 = $90;$vB$sroa$13$0 = $92;$wA$0 = $86;$wB$0 = $94;
  }
 }
 $95 = (HEAP32[tempDoublePtr>>2]=$vB$sroa$0$0,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $96 = (HEAP32[tempDoublePtr>>2]=$vA$sroa$0$0,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $97 = $95 - $96; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $98 = (HEAP32[tempDoublePtr>>2]=$vB$sroa$13$0,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $99 = (HEAP32[tempDoublePtr>>2]=$vA$sroa$13$0,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $100 = $98 - $99; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $101 = ((($this)) + 192|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $102 = +HEAPF32[$101>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $103 = $97 * $102; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $104 = ((($this)) + 196|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $105 = +HEAPF32[$104>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $106 = $100 * $105; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $107 = $103 + $106; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $108 = ((($this)) + 204|0); //@line 298 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $109 = +HEAPF32[$108>>2]; //@line 298 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $110 = $wB$0 * $109; //@line 298 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $111 = $107 + $110; //@line 298 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $112 = ((($this)) + 200|0); //@line 298 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $113 = +HEAPF32[$112>>2]; //@line 298 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $114 = $wA$0 * $113; //@line 298 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $115 = $111 - $114; //@line 298 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $116 = $wB$0 - $wA$0; //@line 299 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $117 = ((($this)) + 136|0); //@line 301 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $118 = HEAP8[$117>>0]|0; //@line 301 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $119 = ($118<<24>>24)==(0); //@line 301 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 if (!($119)) {
  $120 = ((($this)) + 140|0); //@line 301 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $121 = HEAP32[$120>>2]|0; //@line 301 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $122 = ($121|0)==(0); //@line 301 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  if (!($122)) {
   $123 = ((($this)) + 184|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $124 = +HEAPF32[$123>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $125 = $97 * $124; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $126 = ((($this)) + 188|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $127 = +HEAPF32[$126>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $128 = $100 * $127; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $129 = $125 + $128; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $130 = ((($this)) + 212|0); //@line 305 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $131 = +HEAPF32[$130>>2]; //@line 305 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $132 = $wB$0 * $131; //@line 305 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $133 = $129 + $132; //@line 305 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $134 = ((($this)) + 208|0); //@line 305 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $135 = +HEAPF32[$134>>2]; //@line 305 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $136 = $wA$0 * $135; //@line 305 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $137 = $133 - $136; //@line 305 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $138 = ((($this)) + 104|0); //@line 308 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $139 = +HEAPF32[$138>>2]; //@line 308 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $140 = ((($this)) + 108|0); //@line 308 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $141 = +HEAPF32[$140>>2]; //@line 308 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $142 = ((($this)) + 112|0); //@line 308 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $143 = +HEAPF32[$142>>2]; //@line 308 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $144 = ((($this)) + 216|0); //@line 309 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $145 = -$115; //@line 159 "./libraries/Box2D/Common/b2Math.h"
   $146 = -$116; //@line 159 "./libraries/Box2D/Common/b2Math.h"
   $147 = -$137; //@line 159 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$0>>2] = $145; //@line 156 "./libraries/Box2D/Common/b2Math.h"
   $148 = ((($0)) + 4|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$148>>2] = $146; //@line 156 "./libraries/Box2D/Common/b2Math.h"
   $149 = ((($0)) + 8|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$149>>2] = $147; //@line 156 "./libraries/Box2D/Common/b2Math.h"
   __ZNK7b2Mat337Solve33ERK6b2Vec3($df,$144,$0); //@line 309 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $150 = +HEAPF32[$df>>2]; //@line 164 "./libraries/Box2D/Common/b2Math.h"
   $151 = ((($this)) + 104|0); //@line 164 "./libraries/Box2D/Common/b2Math.h"
   $152 = +HEAPF32[$151>>2]; //@line 164 "./libraries/Box2D/Common/b2Math.h"
   $153 = $150 + $152; //@line 164 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$151>>2] = $153; //@line 164 "./libraries/Box2D/Common/b2Math.h"
   $154 = ((($df)) + 4|0); //@line 164 "./libraries/Box2D/Common/b2Math.h"
   $155 = +HEAPF32[$154>>2]; //@line 164 "./libraries/Box2D/Common/b2Math.h"
   $156 = +HEAPF32[$140>>2]; //@line 164 "./libraries/Box2D/Common/b2Math.h"
   $157 = $155 + $156; //@line 164 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$140>>2] = $157; //@line 164 "./libraries/Box2D/Common/b2Math.h"
   $158 = ((($df)) + 8|0); //@line 164 "./libraries/Box2D/Common/b2Math.h"
   $159 = +HEAPF32[$158>>2]; //@line 164 "./libraries/Box2D/Common/b2Math.h"
   $160 = +HEAPF32[$142>>2]; //@line 164 "./libraries/Box2D/Common/b2Math.h"
   $161 = $159 + $160; //@line 314 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   HEAPF32[$142>>2] = $161; //@line 164 "./libraries/Box2D/Common/b2Math.h"
   $162 = HEAP32[$120>>2]|0; //@line 312 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   switch ($162|0) {
   case 1:  {
    $163 = $161 > 0.0; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $164 = $163 ? $161 : 0.0; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$142>>2] = $164; //@line 314 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    $168 = $164;
    break;
   }
   case 2:  {
    $165 = $161 < 0.0; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $166 = $165 ? $161 : 0.0; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$142>>2] = $166; //@line 318 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    $168 = $166;
    break;
   }
   default: {
    $168 = $161;
   }
   }
   $167 = $168 - $143; //@line 322 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $169 = ((($this)) + 240|0); //@line 322 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $170 = +HEAPF32[$169>>2]; //@line 322 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $171 = ((($this)) + 244|0); //@line 322 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $172 = +HEAPF32[$171>>2]; //@line 322 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $173 = $170 * $167; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $174 = $167 * $172; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $175 = $145 - $173; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $176 = $146 - $174; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$b>>2] = $175; //@line 59 "./libraries/Box2D/Common/b2Math.h"
   $177 = ((($b)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$177>>2] = $176; //@line 59 "./libraries/Box2D/Common/b2Math.h"
   __ZNK7b2Mat337Solve22ERK6b2Vec2($1,$144,$b); //@line 323 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $178 = +HEAPF32[$1>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
   $179 = $139 + $178; //@line 448 "./libraries/Box2D/Common/b2Math.h"
   $180 = ((($1)) + 4|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
   $181 = +HEAPF32[$180>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
   $182 = $141 + $181; //@line 448 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$151>>2] = $179; //@line 324 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   HEAPF32[$140>>2] = $182; //@line 325 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $183 = $179 - $139; //@line 493 "./libraries/Box2D/Common/b2Math.h"
   $184 = $182 - $141; //@line 493 "./libraries/Box2D/Common/b2Math.h"
   $185 = +HEAPF32[$142>>2]; //@line 493 "./libraries/Box2D/Common/b2Math.h"
   $186 = $185 - $143; //@line 493 "./libraries/Box2D/Common/b2Math.h"
   $187 = +HEAPF32[$101>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $188 = $183 * $187; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $189 = +HEAPF32[$104>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $190 = $183 * $189; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $191 = +HEAPF32[$123>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $192 = $186 * $191; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $193 = +HEAPF32[$126>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $194 = $186 * $193; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $195 = $188 + $192; //@line 448 "./libraries/Box2D/Common/b2Math.h"
   $196 = $190 + $194; //@line 448 "./libraries/Box2D/Common/b2Math.h"
   $197 = +HEAPF32[$112>>2]; //@line 330 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $198 = $183 * $197; //@line 330 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $199 = $198 + $184; //@line 330 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $200 = +HEAPF32[$134>>2]; //@line 330 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $201 = $186 * $200; //@line 330 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $202 = $199 + $201; //@line 330 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $203 = +HEAPF32[$108>>2]; //@line 331 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $204 = $183 * $203; //@line 331 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $205 = $184 + $204; //@line 331 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $206 = +HEAPF32[$130>>2]; //@line 331 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $207 = $186 * $206; //@line 331 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $208 = $205 + $207; //@line 331 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $$pn3 = $208;$$pn39 = $195;$$pn4 = $202;$$pn40 = $196;
   $$pn38 = $22 * $$pn40; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $$pn37 = $22 * $$pn39; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $$pn36 = $24 * $$pn40; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $$pn35 = $24 * $$pn39; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $vA$sroa$13$1$in = $99 - $$pn38; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $vA$sroa$0$1$in = $96 - $$pn37; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $vB$sroa$13$1$in = $98 + $$pn36; //@line 85 "./libraries/Box2D/Common/b2Math.h"
   $vB$sroa$0$1$in = $95 + $$pn35; //@line 85 "./libraries/Box2D/Common/b2Math.h"
   $$pn2 = $26 * $$pn4; //@line 334 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $$pn = $28 * $$pn3; //@line 337 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $wA$1 = $wA$0 - $$pn2; //@line 334 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $wB$1 = $wB$0 + $$pn; //@line 337 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $232 = HEAP32[$3>>2]|0; //@line 357 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $233 = HEAP32[$5>>2]|0; //@line 357 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $234 = (($233) + (($232*12)|0)|0); //@line 357 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   HEAPF32[$234>>2] = $vA$sroa$0$1$in; //@line 357 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $235 = (((($233) + (($232*12)|0)|0)) + 4|0); //@line 357 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   HEAPF32[$235>>2] = $vA$sroa$13$1$in; //@line 357 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $236 = HEAP32[$3>>2]|0; //@line 358 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $237 = HEAP32[$5>>2]|0; //@line 358 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $238 = (((($237) + (($236*12)|0)|0)) + 8|0); //@line 358 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   HEAPF32[$238>>2] = $wA$1; //@line 358 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $239 = HEAP32[$13>>2]|0; //@line 359 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $240 = (($237) + (($239*12)|0)|0); //@line 359 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   HEAPF32[$240>>2] = $vB$sroa$0$1$in; //@line 359 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $241 = (((($237) + (($239*12)|0)|0)) + 4|0); //@line 359 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   HEAPF32[$241>>2] = $vB$sroa$13$1$in; //@line 359 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $242 = HEAP32[$13>>2]|0; //@line 360 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $243 = HEAP32[$5>>2]|0; //@line 360 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $244 = (((($243) + (($242*12)|0)|0)) + 8|0); //@line 360 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   HEAPF32[$244>>2] = $wB$1; //@line 360 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   STACKTOP = sp;return; //@line 361 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  }
 }
 $209 = ((($this)) + 216|0); //@line 342 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $210 = -$115; //@line 68 "./libraries/Box2D/Common/b2Math.h"
 $211 = -$116; //@line 68 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$2>>2] = $210; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $212 = ((($2)) + 4|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$212>>2] = $211; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 __ZNK7b2Mat337Solve22ERK6b2Vec2($df7,$209,$2); //@line 342 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $213 = +HEAPF32[$df7>>2]; //@line 343 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $214 = ((($this)) + 104|0); //@line 343 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $215 = +HEAPF32[$214>>2]; //@line 343 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $216 = $213 + $215; //@line 343 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$214>>2] = $216; //@line 343 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $217 = ((($df7)) + 4|0); //@line 344 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $218 = +HEAPF32[$217>>2]; //@line 344 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $219 = ((($this)) + 108|0); //@line 344 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $220 = +HEAPF32[$219>>2]; //@line 344 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $221 = $218 + $220; //@line 344 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$219>>2] = $221; //@line 344 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $222 = +HEAPF32[$101>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $223 = $213 * $222; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $224 = +HEAPF32[$104>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $225 = $213 * $224; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $226 = +HEAPF32[$112>>2]; //@line 347 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $227 = $213 * $226; //@line 347 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $228 = $227 + $218; //@line 347 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $229 = +HEAPF32[$108>>2]; //@line 348 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $230 = $213 * $229; //@line 348 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $231 = $218 + $230; //@line 348 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $$pn3 = $231;$$pn39 = $223;$$pn4 = $228;$$pn40 = $225;
 $$pn38 = $22 * $$pn40; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $$pn37 = $22 * $$pn39; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $$pn36 = $24 * $$pn40; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $$pn35 = $24 * $$pn39; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $vA$sroa$13$1$in = $99 - $$pn38; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $vA$sroa$0$1$in = $96 - $$pn37; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $vB$sroa$13$1$in = $98 + $$pn36; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $vB$sroa$0$1$in = $95 + $$pn35; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $$pn2 = $26 * $$pn4; //@line 334 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $$pn = $28 * $$pn3; //@line 337 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $wA$1 = $wA$0 - $$pn2; //@line 334 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $wB$1 = $wB$0 + $$pn; //@line 337 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $232 = HEAP32[$3>>2]|0; //@line 357 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $233 = HEAP32[$5>>2]|0; //@line 357 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $234 = (($233) + (($232*12)|0)|0); //@line 357 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$234>>2] = $vA$sroa$0$1$in; //@line 357 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $235 = (((($233) + (($232*12)|0)|0)) + 4|0); //@line 357 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$235>>2] = $vA$sroa$13$1$in; //@line 357 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $236 = HEAP32[$3>>2]|0; //@line 358 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $237 = HEAP32[$5>>2]|0; //@line 358 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $238 = (((($237) + (($236*12)|0)|0)) + 8|0); //@line 358 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$238>>2] = $wA$1; //@line 358 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $239 = HEAP32[$13>>2]|0; //@line 359 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $240 = (($237) + (($239*12)|0)|0); //@line 359 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$240>>2] = $vB$sroa$0$1$in; //@line 359 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $241 = (((($237) + (($239*12)|0)|0)) + 4|0); //@line 359 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$241>>2] = $vB$sroa$13$1$in; //@line 359 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $242 = HEAP32[$13>>2]|0; //@line 360 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $243 = HEAP32[$5>>2]|0; //@line 360 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $244 = (((($243) + (($242*12)|0)|0)) + 8|0); //@line 360 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$244>>2] = $wB$1; //@line 360 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 STACKTOP = sp;return; //@line 361 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZN16b2PrismaticJoint24SolvePositionConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0, $109 = 0.0, $11 = 0.0, $110 = 0.0, $111 = 0, $112 = 0.0, $113 = 0.0, $114 = 0, $115 = 0.0;
 var $116 = 0.0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0, $124 = 0.0, $125 = 0, $126 = 0.0, $127 = 0.0, $128 = 0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0, $132 = 0, $133 = 0.0;
 var $134 = 0, $135 = 0.0, $136 = 0, $137 = 0.0, $138 = 0.0, $139 = 0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0.0, $144 = 0, $145 = 0.0, $146 = 0, $147 = 0.0, $148 = 0.0, $149 = 0, $15 = 0.0, $150 = 0.0, $151 = 0;
 var $152 = 0.0, $153 = 0.0, $154 = 0, $155 = 0.0, $156 = 0, $157 = 0.0, $158 = 0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0.0, $17 = 0.0;
 var $170 = 0.0, $171 = 0.0, $172 = 0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0.0;
 var $189 = 0, $19 = 0.0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0.0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0.0, $200 = 0.0, $201 = 0, $202 = 0.0, $203 = 0.0, $204 = 0.0, $205 = 0.0;
 var $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0.0, $21 = 0.0, $210 = 0.0, $211 = 0.0, $212 = 0, $213 = 0.0, $214 = 0.0, $215 = 0.0, $216 = 0.0, $217 = 0.0, $218 = 0.0, $219 = 0.0, $22 = 0.0, $220 = 0.0, $221 = 0.0, $222 = 0.0, $223 = 0.0;
 var $224 = 0.0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0.0, $229 = 0.0, $23 = 0.0, $230 = 0.0, $231 = 0.0, $232 = 0.0, $233 = 0.0, $234 = 0.0, $235 = 0.0, $236 = 0.0, $237 = 0.0, $238 = 0.0, $239 = 0.0, $24 = 0, $240 = 0.0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0.0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $26 = 0, $27 = 0.0, $28 = 0, $29 = 0.0;
 var $3 = 0, $30 = 0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0.0;
 var $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0;
 var $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0.0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0;
 var $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $C2$1 = 0.0, $K = 0, $active$1$off0 = 0;
 var $det$0$i = 0.0, $impulse$sroa$0$0 = 0, $impulse$sroa$6$0 = 0, $impulse$sroa$8$0 = 0.0, $k22$0 = 0.0, $k223$0 = 0.0, $linearError$1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $K = sp + 24|0;
 $0 = sp + 12|0;
 $1 = sp;
 $2 = ((($this)) + 144|0); //@line 365 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 365 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $4 = ((($data)) + 24|0); //@line 365 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 365 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $6 = (($5) + (($3*12)|0)|0); //@line 365 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 365 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $8 = (((($5) + (($3*12)|0)|0)) + 4|0); //@line 365 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 365 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $10 = (((($5) + (($3*12)|0)|0)) + 8|0); //@line 366 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $11 = +HEAPF32[$10>>2]; //@line 366 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $12 = ((($this)) + 148|0); //@line 367 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 367 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $14 = (($5) + (($13*12)|0)|0); //@line 367 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 367 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $16 = (((($5) + (($13*12)|0)|0)) + 4|0); //@line 367 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 367 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $18 = (((($5) + (($13*12)|0)|0)) + 8|0); //@line 368 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $19 = +HEAPF32[$18>>2]; //@line 368 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $20 = (+Math_sin((+$11))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $21 = (+Math_cos((+$11))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $22 = (+Math_sin((+$19))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $23 = (+Math_cos((+$19))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($this)) + 168|0); //@line 372 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $25 = +HEAPF32[$24>>2]; //@line 372 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $26 = ((($this)) + 172|0); //@line 372 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $27 = +HEAPF32[$26>>2]; //@line 372 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $28 = ((($this)) + 176|0); //@line 373 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $29 = +HEAPF32[$28>>2]; //@line 373 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $30 = ((($this)) + 180|0); //@line 373 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $31 = +HEAPF32[$30>>2]; //@line 373 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $32 = ((($this)) + 68|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $33 = +HEAPF32[$32>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $34 = ((($this)) + 152|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $35 = +HEAPF32[$34>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $36 = $33 - $35; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $37 = ((($this)) + 72|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $38 = +HEAPF32[$37>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $39 = ((($this)) + 156|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $40 = +HEAPF32[$39>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $41 = $38 - $40; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $42 = $21 * $36; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $43 = $20 * $41; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $44 = $42 - $43; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $45 = $20 * $36; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $46 = $21 * $41; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $47 = $45 + $46; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $48 = ((($this)) + 76|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $49 = +HEAPF32[$48>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $50 = ((($this)) + 160|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $51 = +HEAPF32[$50>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $52 = $49 - $51; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $53 = ((($this)) + 80|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $54 = +HEAPF32[$53>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $55 = ((($this)) + 164|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $56 = +HEAPF32[$55>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $57 = $54 - $56; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $58 = $23 * $52; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $59 = $22 * $57; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $60 = $58 - $59; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $61 = $22 * $52; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $62 = $23 * $57; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $63 = $61 + $62; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $64 = $15 + $60; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $65 = $17 + $63; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $66 = $64 - $7; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $67 = $65 - $9; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $68 = $66 - $44; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $69 = $67 - $47; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $70 = ((($this)) + 84|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $71 = +HEAPF32[$70>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $72 = $21 * $71; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $73 = ((($this)) + 88|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $74 = +HEAPF32[$73>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $75 = $20 * $74; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $76 = $72 - $75; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $77 = $20 * $71; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $78 = $21 * $74; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $79 = $77 + $78; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $80 = $44 + $68; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $81 = $47 + $69; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $82 = $79 * $80; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $83 = $76 * $81; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $84 = $82 - $83; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $85 = $60 * $79; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $86 = $63 * $76; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $87 = $85 - $86; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $88 = ((($this)) + 92|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $89 = +HEAPF32[$88>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $90 = $21 * $89; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $91 = ((($this)) + 96|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $92 = +HEAPF32[$91>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $93 = $20 * $92; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $94 = $90 - $93; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $95 = $20 * $89; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $96 = $21 * $92; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $97 = $95 + $96; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $98 = $97 * $80; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $99 = $94 * $81; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $100 = $98 - $99; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $101 = $60 * $97; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $102 = $63 * $94; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $103 = $101 - $102; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $104 = $94 * $68; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $105 = $97 * $69; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $106 = $104 + $105; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $107 = $19 - $11; //@line 391 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $108 = ((($this)) + 100|0); //@line 391 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $109 = +HEAPF32[$108>>2]; //@line 391 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $110 = $107 - $109; //@line 391 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $111 = $106 > 0.0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $112 = -$106; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $113 = $111 ? $106 : $112; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $114 = $110 > 0.0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $115 = -$110; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $116 = $114 ? $110 : $115; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $117 = ((($this)) + 136|0); //@line 398 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $118 = HEAP8[$117>>0]|0; //@line 398 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $119 = ($118<<24>>24)==(0); //@line 398 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 do {
  if ($119) {
   $C2$1 = 0.0;$active$1$off0 = 0;$linearError$1 = $113;
  } else {
   $120 = $76 * $68; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $121 = $79 * $69; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $122 = $120 + $121; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $123 = ((($this)) + 124|0); //@line 401 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $124 = +HEAPF32[$123>>2]; //@line 401 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $125 = ((($this)) + 120|0); //@line 401 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $126 = +HEAPF32[$125>>2]; //@line 401 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $127 = $124 - $126; //@line 401 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   $128 = $127 > 0.0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
   $129 = -$127; //@line 618 "./libraries/Box2D/Common/b2Math.h"
   $130 = $128 ? $127 : $129; //@line 618 "./libraries/Box2D/Common/b2Math.h"
   $131 = $130 < 0.0099999997764825821; //@line 401 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   if ($131) {
    $132 = $122 < 0.20000000298023224; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $133 = $132 ? $122 : 0.20000000298023224; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $134 = $133 < -0.20000000298023224; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $135 = $134 ? -0.20000000298023224 : $133; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $136 = $122 > 0.0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
    $137 = -$122; //@line 618 "./libraries/Box2D/Common/b2Math.h"
    $138 = $136 ? $122 : $137; //@line 618 "./libraries/Box2D/Common/b2Math.h"
    $139 = $113 > $138; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $140 = $139 ? $113 : $138; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $C2$1 = $135;$active$1$off0 = 1;$linearError$1 = $140;
    break;
   }
   $141 = !($122 <= $126); //@line 408 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   if (!($141)) {
    $142 = $122 - $126; //@line 411 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    $143 = $142 + 0.004999999888241291; //@line 411 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    $144 = $143 < 0.0; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $145 = $144 ? $143 : 0.0; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $146 = $145 < -0.20000000298023224; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $147 = $146 ? -0.20000000298023224 : $145; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $148 = $126 - $122; //@line 412 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    $149 = $113 > $148; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $150 = $149 ? $113 : $148; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $C2$1 = $147;$active$1$off0 = 1;$linearError$1 = $150;
    break;
   }
   $151 = !($122 >= $124); //@line 415 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
   if ($151) {
    $C2$1 = 0.0;$active$1$off0 = 0;$linearError$1 = $113;
   } else {
    $152 = $122 - $124; //@line 418 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    $153 = $152 + -0.004999999888241291; //@line 418 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
    $154 = $153 < 0.20000000298023224; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $155 = $154 ? $153 : 0.20000000298023224; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $156 = $155 < 0.0; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $157 = $156 ? 0.0 : $155; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $158 = $113 > $152; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $159 = $158 ? $113 : $152; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $C2$1 = $157;$active$1$off0 = 1;$linearError$1 = $159;
   }
  }
 } while(0);
 $160 = $25 + $27; //@line 426 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $161 = $29 * $100; //@line 426 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $162 = $100 * $161; //@line 426 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $163 = $160 + $162; //@line 426 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $164 = $31 * $103; //@line 426 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $165 = $103 * $164; //@line 426 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $166 = $165 + $163; //@line 426 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $167 = $164 + $161; //@line 427 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 if ($active$1$off0) {
  $168 = $84 * $161; //@line 428 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $169 = $87 * $164; //@line 428 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $170 = $169 + $168; //@line 428 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $171 = $29 + $31; //@line 429 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $172 = $171 == 0.0; //@line 430 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $k22$0 = $172 ? 1.0 : $171; //@line 430 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $173 = $29 * $84; //@line 435 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $174 = $31 * $87; //@line 435 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $175 = $174 + $173; //@line 435 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $176 = $84 * $173; //@line 436 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $177 = $160 + $176; //@line 436 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $178 = $87 * $174; //@line 436 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $179 = $178 + $177; //@line 436 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAPF32[$K>>2] = $166; //@line 156 "./libraries/Box2D/Common/b2Math.h"
  $180 = ((($K)) + 4|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$180>>2] = $167; //@line 156 "./libraries/Box2D/Common/b2Math.h"
  $181 = ((($K)) + 8|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$181>>2] = $170; //@line 156 "./libraries/Box2D/Common/b2Math.h"
  $182 = ((($K)) + 12|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$182>>2] = $167; //@line 156 "./libraries/Box2D/Common/b2Math.h"
  $183 = ((($K)) + 16|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$183>>2] = $k22$0; //@line 156 "./libraries/Box2D/Common/b2Math.h"
  $184 = ((($K)) + 20|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$184>>2] = $175; //@line 156 "./libraries/Box2D/Common/b2Math.h"
  $185 = ((($K)) + 24|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$185>>2] = $170; //@line 156 "./libraries/Box2D/Common/b2Math.h"
  $186 = ((($K)) + 28|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$186>>2] = $175; //@line 156 "./libraries/Box2D/Common/b2Math.h"
  $187 = ((($K)) + 32|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$187>>2] = $179; //@line 156 "./libraries/Box2D/Common/b2Math.h"
  $188 = -$C2$1; //@line 159 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$1>>2] = $112; //@line 156 "./libraries/Box2D/Common/b2Math.h"
  $189 = ((($1)) + 4|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$189>>2] = $115; //@line 156 "./libraries/Box2D/Common/b2Math.h"
  $190 = ((($1)) + 8|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$190>>2] = $188; //@line 156 "./libraries/Box2D/Common/b2Math.h"
  __ZNK7b2Mat337Solve33ERK6b2Vec3($0,$K,$1); //@line 448 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $191 = HEAP32[$0>>2]|0; //@line 448 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $192 = ((($0)) + 4|0); //@line 448 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $193 = HEAP32[$192>>2]|0; //@line 448 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $194 = ((($0)) + 8|0); //@line 448 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $195 = +HEAPF32[$194>>2]; //@line 448 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $impulse$sroa$0$0 = $191;$impulse$sroa$6$0 = $193;$impulse$sroa$8$0 = $195;
 } else {
  $196 = $29 + $31; //@line 454 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $197 = $196 == 0.0; //@line 455 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $k223$0 = $197 ? 1.0 : $196; //@line 455 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $198 = $k223$0 * $166; //@line 242 "./libraries/Box2D/Common/b2Math.h"
  $199 = $167 * $167; //@line 242 "./libraries/Box2D/Common/b2Math.h"
  $200 = $198 - $199; //@line 242 "./libraries/Box2D/Common/b2Math.h"
  $201 = $200 != 0.0; //@line 243 "./libraries/Box2D/Common/b2Math.h"
  $202 = 1.0 / $200; //@line 245 "./libraries/Box2D/Common/b2Math.h"
  $det$0$i = $201 ? $202 : $200; //@line 243 "./libraries/Box2D/Common/b2Math.h"
  $203 = $k223$0 * $112; //@line 248 "./libraries/Box2D/Common/b2Math.h"
  $204 = $167 * $115; //@line 248 "./libraries/Box2D/Common/b2Math.h"
  $205 = $203 - $204; //@line 248 "./libraries/Box2D/Common/b2Math.h"
  $206 = $205 * $det$0$i; //@line 248 "./libraries/Box2D/Common/b2Math.h"
  $207 = (HEAPF32[tempDoublePtr>>2]=$206,HEAP32[tempDoublePtr>>2]|0); //@line 248 "./libraries/Box2D/Common/b2Math.h"
  $208 = $166 * $115; //@line 249 "./libraries/Box2D/Common/b2Math.h"
  $209 = $167 * $112; //@line 249 "./libraries/Box2D/Common/b2Math.h"
  $210 = $208 - $209; //@line 249 "./libraries/Box2D/Common/b2Math.h"
  $211 = $210 * $det$0$i; //@line 249 "./libraries/Box2D/Common/b2Math.h"
  $212 = (HEAPF32[tempDoublePtr>>2]=$211,HEAP32[tempDoublePtr>>2]|0); //@line 249 "./libraries/Box2D/Common/b2Math.h"
  $impulse$sroa$0$0 = $207;$impulse$sroa$6$0 = $212;$impulse$sroa$8$0 = 0.0;
 }
 $213 = (HEAP32[tempDoublePtr>>2]=$impulse$sroa$0$0,+HEAPF32[tempDoublePtr>>2]); //@line 470 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $214 = $94 * $213; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $215 = $97 * $213; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $216 = $76 * $impulse$sroa$8$0; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $217 = $79 * $impulse$sroa$8$0; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $218 = $214 + $216; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $219 = $215 + $217; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $220 = $100 * $213; //@line 471 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $221 = (HEAP32[tempDoublePtr>>2]=$impulse$sroa$6$0,+HEAPF32[tempDoublePtr>>2]); //@line 471 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $222 = $220 + $221; //@line 471 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $223 = $84 * $impulse$sroa$8$0; //@line 471 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $224 = $222 + $223; //@line 471 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $225 = $103 * $213; //@line 472 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $226 = $225 + $221; //@line 472 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $227 = $87 * $impulse$sroa$8$0; //@line 472 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $228 = $226 + $227; //@line 472 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $229 = $25 * $218; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $230 = $25 * $219; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $231 = $7 - $229; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $232 = $9 - $230; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $233 = $29 * $224; //@line 475 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $234 = $11 - $233; //@line 475 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $235 = $27 * $218; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $236 = $27 * $219; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $237 = $15 + $235; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $238 = $17 + $236; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $239 = $31 * $228; //@line 477 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $240 = $19 + $239; //@line 477 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $241 = HEAP32[$2>>2]|0; //@line 479 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $242 = HEAP32[$4>>2]|0; //@line 479 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $243 = (($242) + (($241*12)|0)|0); //@line 479 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$243>>2] = $231; //@line 479 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $244 = (((($242) + (($241*12)|0)|0)) + 4|0); //@line 479 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$244>>2] = $232; //@line 479 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $245 = HEAP32[$2>>2]|0; //@line 480 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $246 = HEAP32[$4>>2]|0; //@line 480 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $247 = (((($246) + (($245*12)|0)|0)) + 8|0); //@line 480 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$247>>2] = $234; //@line 480 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $248 = HEAP32[$12>>2]|0; //@line 481 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $249 = (($246) + (($248*12)|0)|0); //@line 481 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$249>>2] = $237; //@line 481 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $250 = (((($246) + (($248*12)|0)|0)) + 4|0); //@line 481 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$250>>2] = $238; //@line 481 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $251 = HEAP32[$12>>2]|0; //@line 482 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $252 = HEAP32[$4>>2]|0; //@line 482 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $253 = (((($252) + (($251*12)|0)|0)) + 8|0); //@line 482 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$253>>2] = $240; //@line 482 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $254 = $linearError$1 <= 0.004999999888241291; //@line 484 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $255 = $116 <= 0.034906588494777679; //@line 484 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $256 = $255 & $254; //@line 484 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 STACKTOP = sp;return ($256|0); //@line 484 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZN19b2PrismaticJointDef10InitializeEP6b2BodyS1_RK6b2Vec2S4_($this,$bA,$bB,$anchor,$axis) {
 $this = $this|0;
 $bA = $bA|0;
 $bB = $bB|0;
 $anchor = $anchor|0;
 $axis = $axis|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0;
 var $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0.0, $40 = 0.0, $41 = 0, $42 = 0, $43 = 0.0, $44 = 0.0;
 var $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0, $55 = 0, $56 = 0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0, $7 = 0.0;
 var $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 92 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$0>>2] = $bA; //@line 92 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = ((($this)) + 12|0); //@line 93 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$1>>2] = $bB; //@line 93 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = +HEAPF32[$anchor>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($bA)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$3>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $5 = $2 - $4; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($anchor)) + 4|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($bA)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $9 = +HEAPF32[$8>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $10 = $7 - $9; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($bA)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$11>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $13 = $5 * $12; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $14 = ((($bA)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $15 = +HEAPF32[$14>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $16 = $10 * $15; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $17 = $13 + $16; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $18 = $5 * $15; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $19 = $12 * $10; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $20 = $19 - $18; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($this)) + 20|0); //@line 94 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$21>>2] = $17; //@line 94 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $22 = ((($this)) + 24|0); //@line 94 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$22>>2] = $20; //@line 94 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $23 = +HEAPF32[$anchor>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($bB)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $26 = $23 - $25; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $27 = +HEAPF32[$6>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $28 = ((($bB)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $29 = +HEAPF32[$28>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $30 = $27 - $29; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $31 = ((($bB)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $32 = +HEAPF32[$31>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $33 = $26 * $32; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $34 = ((($bB)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $35 = +HEAPF32[$34>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $36 = $30 * $35; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $37 = $33 + $36; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $38 = $26 * $35; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $39 = $32 * $30; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $40 = $39 - $38; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $41 = ((($this)) + 28|0); //@line 95 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$41>>2] = $37; //@line 95 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $42 = ((($this)) + 32|0); //@line 95 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$42>>2] = $40; //@line 95 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $43 = +HEAPF32[$11>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $44 = +HEAPF32[$axis>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $45 = $43 * $44; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $46 = +HEAPF32[$14>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $47 = ((($axis)) + 4|0); //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $48 = +HEAPF32[$47>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $49 = $46 * $48; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $50 = $45 + $49; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $51 = $44 * $46; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $52 = $43 * $48; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $53 = $52 - $51; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $54 = ((($this)) + 36|0); //@line 96 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$54>>2] = $50; //@line 96 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $55 = ((($this)) + 40|0); //@line 96 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$55>>2] = $53; //@line 96 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $56 = ((($bB)) + 56|0); //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $57 = +HEAPF32[$56>>2]; //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $58 = ((($bA)) + 56|0); //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $59 = +HEAPF32[$58>>2]; //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $60 = $57 - $59; //@line 97 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $61 = ((($this)) + 44|0); //@line 97 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$61>>2] = $60; //@line 97 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 return; //@line 98 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZN16b2PrismaticJointC2EPK19b2PrismaticJointDef($this,$def) {
 $this = $this|0;
 $def = $def|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0, $43 = 0.0, $44 = 0.0;
 var $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $8 = 0, $9 = 0, $bitcast = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN7b2JointC2EPK10b2JointDef($this,$def); //@line 101 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$this>>2] = (20868); //@line 102 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $0 = ((($this)) + 68|0); //@line 100 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = ((($this)) + 76|0); //@line 100 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = ((($this)) + 84|0); //@line 100 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $3 = ((($def)) + 20|0); //@line 103 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $4 = $3; //@line 103 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $5 = $4; //@line 103 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 103 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $7 = (($4) + 4)|0; //@line 103 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $8 = $7; //@line 103 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 103 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $10 = $0; //@line 103 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $11 = $10; //@line 103 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$11>>2] = $6; //@line 103 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $12 = (($10) + 4)|0; //@line 103 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $13 = $12; //@line 103 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$13>>2] = $9; //@line 103 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $14 = ((($def)) + 28|0); //@line 104 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $15 = $14; //@line 104 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $16 = $15; //@line 104 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 104 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $18 = (($15) + 4)|0; //@line 104 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $19 = $18; //@line 104 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 104 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $21 = $1; //@line 104 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $22 = $21; //@line 104 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$22>>2] = $17; //@line 104 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $23 = (($21) + 4)|0; //@line 104 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $24 = $23; //@line 104 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$24>>2] = $20; //@line 104 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $25 = ((($def)) + 36|0); //@line 105 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $26 = $25; //@line 105 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $27 = $26; //@line 105 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $28 = HEAP32[$27>>2]|0; //@line 105 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $29 = (($26) + 4)|0; //@line 105 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $30 = $29; //@line 105 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 105 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $32 = $2; //@line 105 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $33 = $32; //@line 105 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$33>>2] = $28; //@line 105 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $34 = (($32) + 4)|0; //@line 105 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $35 = $34; //@line 105 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$35>>2] = $31; //@line 105 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $bitcast = (HEAP32[tempDoublePtr>>2]=$28,+HEAPF32[tempDoublePtr>>2]); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $36 = $bitcast * $bitcast; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $37 = (HEAP32[tempDoublePtr>>2]=$31,+HEAPF32[tempDoublePtr>>2]); //@line 123 "./libraries/Box2D/Common/b2Math.h"
 $38 = $37 * $37; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $39 = $36 + $38; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $40 = (+Math_sqrt((+$39))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $41 = $40 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
 if ($41) {
  $48 = $37;$51 = $28;
 } else {
  $42 = ((($this)) + 88|0); //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $43 = 1.0 / $40; //@line 121 "./libraries/Box2D/Common/b2Math.h"
  $44 = $bitcast * $43; //@line 122 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$2>>2] = $44; //@line 122 "./libraries/Box2D/Common/b2Math.h"
  $45 = $37 * $43; //@line 123 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$42>>2] = $45; //@line 123 "./libraries/Box2D/Common/b2Math.h"
  $46 = (HEAPF32[tempDoublePtr>>2]=$44,HEAP32[tempDoublePtr>>2]|0); //@line 106 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $48 = $45;$51 = $46;
 }
 $47 = -$48; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $49 = ((($this)) + 92|0); //@line 107 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$49>>2] = $47; //@line 107 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $50 = ((($this)) + 96|0); //@line 107 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$50>>2] = $51; //@line 107 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $52 = ((($def)) + 44|0); //@line 108 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $53 = HEAP32[$52>>2]|0; //@line 108 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $54 = ((($this)) + 100|0); //@line 108 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$54>>2] = $53; //@line 108 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $55 = ((($this)) + 104|0); //@line 153 "./libraries/Box2D/Common/b2Math.h"
 $56 = ((($this)) + 252|0); //@line 111 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$56>>2] = 0.0; //@line 111 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $57 = ((($def)) + 52|0); //@line 114 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 ;HEAP32[$55>>2]=0|0;HEAP32[$55+4>>2]=0|0;HEAP32[$55+8>>2]=0|0;HEAP32[$55+12>>2]=0|0; //@line 153 "./libraries/Box2D/Common/b2Math.h"
 $58 = HEAP32[$57>>2]|0; //@line 114 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $59 = ((($this)) + 120|0); //@line 114 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$59>>2] = $58; //@line 114 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $60 = ((($def)) + 56|0); //@line 115 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $61 = HEAP32[$60>>2]|0; //@line 115 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $62 = ((($this)) + 124|0); //@line 115 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$62>>2] = $61; //@line 115 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $63 = ((($def)) + 64|0); //@line 116 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $64 = HEAP32[$63>>2]|0; //@line 116 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $65 = ((($this)) + 128|0); //@line 116 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$65>>2] = $64; //@line 116 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $66 = ((($def)) + 68|0); //@line 117 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $67 = HEAP32[$66>>2]|0; //@line 117 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $68 = ((($this)) + 132|0); //@line 117 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$68>>2] = $67; //@line 117 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $69 = ((($def)) + 48|0); //@line 118 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $70 = HEAP8[$69>>0]|0; //@line 118 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $71 = ((($this)) + 136|0); //@line 118 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP8[$71>>0] = $70; //@line 118 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $72 = ((($def)) + 60|0); //@line 119 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $73 = HEAP8[$72>>0]|0; //@line 119 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $74 = ((($this)) + 137|0); //@line 119 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP8[$74>>0] = $73; //@line 119 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $75 = ((($this)) + 140|0); //@line 120 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP32[$75>>2] = 0; //@line 120 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $76 = ((($this)) + 184|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 ;HEAP32[$76>>2]=0|0;HEAP32[$76+4>>2]=0|0;HEAP32[$76+8>>2]=0|0;HEAP32[$76+12>>2]=0|0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 124 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZNK16b2PrismaticJoint19GetJointTranslationEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0, $24 = 0, $25 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0.0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0.0;
 var $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $6 = 0.0, $7 = 0, $8 = 0.0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 509 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 509 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 68|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 72|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($this)) + 52|0); //@line 510 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $23 = HEAP32[$22>>2]|0; //@line 510 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $24 = ((($23)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $26 = ((($this)) + 76|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $27 = +HEAPF32[$26>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $28 = $25 * $27; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $29 = ((($23)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $30 = +HEAPF32[$29>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $31 = ((($this)) + 80|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $32 = +HEAPF32[$31>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $33 = $30 * $32; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $34 = $28 - $33; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $35 = ((($23)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $36 = +HEAPF32[$35>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $37 = $36 + $34; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $38 = $27 * $30; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $39 = $25 * $32; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $40 = $38 + $39; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $41 = ((($23)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $42 = +HEAPF32[$41>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $43 = $40 + $42; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $44 = $37 - $15; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $45 = $43 - $21; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $46 = ((($this)) + 84|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $47 = +HEAPF32[$46>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $48 = $3 * $47; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $49 = ((($this)) + 88|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $50 = +HEAPF32[$49>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $51 = $8 * $50; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $52 = $48 - $51; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $53 = $8 * $47; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $54 = $3 * $50; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $55 = $53 + $54; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $56 = $44 * $52; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $57 = $45 * $55; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $58 = $56 + $57; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 return (+$58); //@line 515 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZNK16b2PrismaticJoint13GetJointSpeedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0;
 var $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0.0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0, $62 = 0.0;
 var $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0, $69 = 0.0, $7 = 0.0, $70 = 0, $71 = 0.0, $72 = 0, $73 = 0.0, $74 = 0, $75 = 0.0, $76 = 0, $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0.0, $80 = 0.0;
 var $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 520 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 520 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = ((($this)) + 52|0); //@line 521 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 521 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $4 = ((($this)) + 68|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($1)) + 28|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $8 = $5 - $7; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 72|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($1)) + 32|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$11>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $13 = $10 - $12; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $14 = ((($1)) + 24|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $15 = +HEAPF32[$14>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $16 = $8 * $15; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $17 = ((($1)) + 20|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $18 = +HEAPF32[$17>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $19 = $13 * $18; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $20 = $16 - $19; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $21 = $8 * $18; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $22 = $15 * $13; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $23 = $22 + $21; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($this)) + 76|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $26 = ((($3)) + 28|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $27 = +HEAPF32[$26>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $28 = $25 - $27; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $29 = ((($this)) + 80|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $30 = +HEAPF32[$29>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $31 = ((($3)) + 32|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $32 = +HEAPF32[$31>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $33 = $30 - $32; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $34 = ((($3)) + 24|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $35 = +HEAPF32[$34>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $36 = $28 * $35; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $37 = ((($3)) + 20|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $38 = +HEAPF32[$37>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $39 = $33 * $38; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $40 = $36 - $39; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $41 = $28 * $38; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $42 = $35 * $33; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $43 = $42 + $41; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $44 = ((($1)) + 44|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $45 = +HEAPF32[$44>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $46 = $20 + $45; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $47 = ((($1)) + 48|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $48 = +HEAPF32[$47>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $49 = $23 + $48; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $50 = ((($3)) + 44|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $51 = +HEAPF32[$50>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $52 = $40 + $51; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $53 = ((($3)) + 48|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $54 = +HEAPF32[$53>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $55 = $43 + $54; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $56 = $52 - $46; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $57 = $55 - $49; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $58 = ((($this)) + 84|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $59 = +HEAPF32[$58>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $60 = $15 * $59; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $61 = ((($this)) + 88|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $62 = +HEAPF32[$61>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $63 = $18 * $62; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $64 = $60 - $63; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $65 = $18 * $59; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $66 = $15 * $62; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $67 = $65 + $66; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $68 = ((($1)) + 64|0); //@line 530 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $69 = +HEAPF32[$68>>2]; //@line 530 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $70 = ((($1)) + 68|0); //@line 530 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $71 = +HEAPF32[$70>>2]; //@line 530 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $72 = ((($3)) + 64|0); //@line 531 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $73 = +HEAPF32[$72>>2]; //@line 531 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $74 = ((($3)) + 68|0); //@line 531 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $75 = +HEAPF32[$74>>2]; //@line 531 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $76 = ((($1)) + 72|0); //@line 532 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $77 = +HEAPF32[$76>>2]; //@line 532 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $78 = ((($3)) + 72|0); //@line 533 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $79 = +HEAPF32[$78>>2]; //@line 533 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $80 = $67 * $77; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $81 = $64 * $77; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $82 = $56 * $80; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $83 = $57 * $81; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $84 = $83 - $82; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $85 = $43 * $79; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $86 = $40 * $79; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $87 = $73 - $85; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $88 = $75 + $86; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $89 = $87 - $69; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $90 = $88 - $71; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $91 = $23 * $77; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $92 = $20 * $77; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $93 = $91 + $89; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $94 = $90 - $92; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $95 = $64 * $93; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $96 = $67 * $94; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $97 = $95 + $96; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $98 = $84 + $97; //@line 535 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 return (+$98); //@line 536 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZNK16b2PrismaticJoint14IsLimitEnabledEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 136|0); //@line 541 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 541 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 541 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 return ($2|0); //@line 541 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZN16b2PrismaticJoint11EnableLimitEb($this,$flag) {
 $this = $this|0;
 $flag = $flag|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $flag&1; //@line 546 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = ((($this)) + 136|0); //@line 546 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = HEAP8[$1>>0]|0; //@line 546 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $3 = $2&255; //@line 546 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $4 = ($0|0)==($3|0); //@line 546 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 if ($4) {
  return; //@line 553 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 }
 $5 = ((($this)) + 48|0); //@line 548 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 548 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $7 = ((($6)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $8 = HEAP16[$7>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $9 = $8&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $10 = $9 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $11 = ($10|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($11) {
  $12 = $9 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $13 = $12&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$7>>1] = $13; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $14 = ((($6)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$14>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $15 = ((($this)) + 52|0); //@line 549 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 549 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $17 = ((($16)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $18 = HEAP16[$17>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = $18&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = $19 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $21 = ($20|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($21) {
  $22 = $19 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $23 = $22&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$17>>1] = $23; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $24 = ((($16)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$24>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $25 = $flag&1; //@line 550 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP8[$1>>0] = $25; //@line 550 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $26 = ((($this)) + 112|0); //@line 551 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$26>>2] = 0.0; //@line 551 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 return; //@line 553 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZNK16b2PrismaticJoint13GetLowerLimitEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 120|0); //@line 557 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 557 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 return (+$1); //@line 557 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZNK16b2PrismaticJoint13GetUpperLimitEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 124|0); //@line 562 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 562 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 return (+$1); //@line 562 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZN16b2PrismaticJoint9SetLimitsEff($this,$lower,$upper) {
 $this = $this|0;
 $lower = +$lower;
 $upper = +$upper;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $lower <= $upper; //@line 567 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 __Z10loveAssertbPKc($0,363332); //@line 567 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = ((($this)) + 120|0); //@line 568 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = +HEAPF32[$1>>2]; //@line 568 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $3 = $2 != $lower; //@line 568 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 if (!($3)) {
  $4 = ((($this)) + 124|0); //@line 568 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $5 = +HEAPF32[$4>>2]; //@line 568 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $6 = $5 != $upper; //@line 568 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  if (!($6)) {
   return; //@line 576 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  }
 }
 $7 = ((($this)) + 48|0); //@line 570 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 570 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $9 = ((($8)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $10 = HEAP16[$9>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $11 = $10&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $12 = $11 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $13 = ($12|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($13) {
  $14 = $11 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $15 = $14&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$9>>1] = $15; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $16 = ((($8)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$16>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $17 = ((($this)) + 52|0); //@line 571 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 571 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $19 = ((($18)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = HEAP16[$19>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $21 = $20&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $22 = $21 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $23 = ($22|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($23) {
  $24 = $21 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $25 = $24&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$19>>1] = $25; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $26 = ((($18)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$26>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 HEAPF32[$1>>2] = $lower; //@line 572 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $27 = ((($this)) + 124|0); //@line 573 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$27>>2] = $upper; //@line 573 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $28 = ((($this)) + 112|0); //@line 574 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$28>>2] = 0.0; //@line 574 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 return; //@line 576 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZNK16b2PrismaticJoint14IsMotorEnabledEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 137|0); //@line 580 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 580 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 580 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 return ($2|0); //@line 580 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZN16b2PrismaticJoint11EnableMotorEb($this,$flag) {
 $this = $this|0;
 $flag = $flag|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 585 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 585 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = ((($1)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = $4 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $6 = ($5|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($6) {
  $7 = $4 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $8 = $7&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$2>>1] = $8; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $9 = ((($1)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$9>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $10 = ((($this)) + 52|0); //@line 586 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 586 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $12 = ((($11)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $13 = HEAP16[$12>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $14 = $13&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $15 = $14 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $16 = ($15|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($16)) {
  $20 = ((($this)) + 137|0); //@line 587 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  $21 = $flag&1; //@line 587 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAP8[$20>>0] = $21; //@line 587 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  return; //@line 588 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 }
 $17 = $14 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $18 = $17&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP16[$12>>1] = $18; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = ((($11)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$19>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = ((($this)) + 137|0); //@line 587 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $21 = $flag&1; //@line 587 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAP8[$20>>0] = $21; //@line 587 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 return; //@line 588 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZN16b2PrismaticJoint13SetMotorSpeedEf($this,$speed) {
 $this = $this|0;
 $speed = +$speed;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 592 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 592 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = ((($1)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = $4 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $6 = ($5|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($6) {
  $7 = $4 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $8 = $7&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$2>>1] = $8; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $9 = ((($1)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$9>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $10 = ((($this)) + 52|0); //@line 593 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 593 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $12 = ((($11)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $13 = HEAP16[$12>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $14 = $13&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $15 = $14 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $16 = ($15|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($16)) {
  $20 = ((($this)) + 132|0); //@line 594 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAPF32[$20>>2] = $speed; //@line 594 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  return; //@line 595 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 }
 $17 = $14 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $18 = $17&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP16[$12>>1] = $18; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = ((($11)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$19>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = ((($this)) + 132|0); //@line 594 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$20>>2] = $speed; //@line 594 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 return; //@line 595 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZN16b2PrismaticJoint16SetMaxMotorForceEf($this,$force) {
 $this = $this|0;
 $force = +$force;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 599 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 599 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = ((($1)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = $4 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $6 = ($5|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($6) {
  $7 = $4 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $8 = $7&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$2>>1] = $8; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $9 = ((($1)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$9>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $10 = ((($this)) + 52|0); //@line 600 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 600 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $12 = ((($11)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $13 = HEAP16[$12>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $14 = $13&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $15 = $14 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $16 = ($15|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($16)) {
  $20 = ((($this)) + 128|0); //@line 601 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  HEAPF32[$20>>2] = $force; //@line 601 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
  return; //@line 602 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 }
 $17 = $14 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $18 = $17&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP16[$12>>1] = $18; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = ((($11)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$19>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = ((($this)) + 128|0); //@line 601 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 HEAPF32[$20>>2] = $force; //@line 601 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 return; //@line 602 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZNK16b2PrismaticJoint13GetMotorForceEf($this,$inv_dt) {
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 116|0); //@line 606 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 606 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 $2 = $1 * $inv_dt; //@line 606 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
 return (+$2); //@line 606 "libraries/Box2D/Dynamics/Joints/b2PrismaticJoint.cpp"
}
function __ZNK13b2PulleyJoint10GetAnchorAEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 264 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 264 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 92|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 96|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 264 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
}
function __ZNK13b2PulleyJoint10GetAnchorBEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 269 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 269 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 100|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 104|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 269 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
}
function __ZNK13b2PulleyJoint16GetReactionForceEf($agg$result,$this,$inv_dt) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $10 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 116|0); //@line 274 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 274 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $2 = ((($this)) + 136|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $4 = $1 * $3; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($this)) + 140|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $7 = $1 * $6; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $8 = $4 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $9 = $7 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $8; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$10>>2] = $9; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 276 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
}
function __ZNK13b2PulleyJoint17GetReactionTorqueEf($this,$inv_dt) {
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return +0; //@line 281 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
}
function __ZN13b2PulleyJoint4DumpEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0;
 var $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer13 = 0, $vararg_buffer17 = 0, $vararg_buffer21 = 0, $vararg_buffer25 = 0, $vararg_buffer28 = 0, $vararg_buffer3 = 0, $vararg_buffer31 = 0, $vararg_buffer34 = 0, $vararg_buffer6 = 0, $vararg_buffer9 = 0, $vararg_ptr12 = 0, $vararg_ptr16 = 0;
 var $vararg_ptr20 = 0, $vararg_ptr24 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer34 = sp + 120|0;
 $vararg_buffer31 = sp + 112|0;
 $vararg_buffer28 = sp + 104|0;
 $vararg_buffer25 = sp + 96|0;
 $vararg_buffer21 = sp + 80|0;
 $vararg_buffer17 = sp + 64|0;
 $vararg_buffer13 = sp + 48|0;
 $vararg_buffer9 = sp + 32|0;
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($this)) + 48|0); //@line 327 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 327 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 327 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 327 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $4 = ((($this)) + 52|0); //@line 328 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 328 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $6 = ((($5)) + 8|0); //@line 328 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 328 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 __Z5b2LogPKcz(363004,$vararg_buffer); //@line 330 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$vararg_buffer1>>2] = $3; //@line 331 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 __Z5b2LogPKcz(447349,$vararg_buffer1); //@line 331 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$vararg_buffer3>>2] = $7; //@line 332 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 __Z5b2LogPKcz(447375,$vararg_buffer3); //@line 332 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $8 = ((($this)) + 61|0); //@line 333 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 333 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $10 = $9&255; //@line 333 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$vararg_buffer6>>2] = $10; //@line 333 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 __Z5b2LogPKcz(447401,$vararg_buffer6); //@line 333 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $11 = ((($this)) + 68|0); //@line 334 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $12 = +HEAPF32[$11>>2]; //@line 334 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $13 = $12; //@line 334 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $14 = ((($this)) + 72|0); //@line 334 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 334 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $16 = $15; //@line 334 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF64[$vararg_buffer9>>3] = $13; //@line 334 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $vararg_ptr12 = ((($vararg_buffer9)) + 8|0); //@line 334 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF64[$vararg_ptr12>>3] = $16; //@line 334 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 __Z5b2LogPKcz(363028,$vararg_buffer9); //@line 334 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $17 = ((($this)) + 76|0); //@line 335 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $18 = +HEAPF32[$17>>2]; //@line 335 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $19 = $18; //@line 335 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $20 = ((($this)) + 80|0); //@line 335 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 335 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $22 = $21; //@line 335 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF64[$vararg_buffer13>>3] = $19; //@line 335 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $vararg_ptr16 = ((($vararg_buffer13)) + 8|0); //@line 335 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF64[$vararg_ptr16>>3] = $22; //@line 335 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 __Z5b2LogPKcz(363071,$vararg_buffer13); //@line 335 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $23 = ((($this)) + 92|0); //@line 336 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $24 = +HEAPF32[$23>>2]; //@line 336 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $25 = $24; //@line 336 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $26 = ((($this)) + 96|0); //@line 336 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $27 = +HEAPF32[$26>>2]; //@line 336 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $28 = $27; //@line 336 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF64[$vararg_buffer17>>3] = $25; //@line 336 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $vararg_ptr20 = ((($vararg_buffer17)) + 8|0); //@line 336 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF64[$vararg_ptr20>>3] = $28; //@line 336 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 __Z5b2LogPKcz(447106,$vararg_buffer17); //@line 336 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $29 = ((($this)) + 100|0); //@line 337 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $30 = +HEAPF32[$29>>2]; //@line 337 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $31 = $30; //@line 337 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $32 = ((($this)) + 104|0); //@line 337 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $33 = +HEAPF32[$32>>2]; //@line 337 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $34 = $33; //@line 337 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF64[$vararg_buffer21>>3] = $31; //@line 337 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $vararg_ptr24 = ((($vararg_buffer21)) + 8|0); //@line 337 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF64[$vararg_ptr24>>3] = $34; //@line 337 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 __Z5b2LogPKcz(447148,$vararg_buffer21); //@line 337 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $35 = ((($this)) + 84|0); //@line 338 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $36 = +HEAPF32[$35>>2]; //@line 338 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $37 = $36; //@line 338 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF64[$vararg_buffer25>>3] = $37; //@line 338 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 __Z5b2LogPKcz(363114,$vararg_buffer25); //@line 338 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $38 = ((($this)) + 88|0); //@line 339 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $39 = +HEAPF32[$38>>2]; //@line 339 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $40 = $39; //@line 339 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF64[$vararg_buffer28>>3] = $40; //@line 339 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 __Z5b2LogPKcz(363139,$vararg_buffer28); //@line 339 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $41 = ((($this)) + 112|0); //@line 340 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $42 = +HEAPF32[$41>>2]; //@line 340 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $43 = $42; //@line 340 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF64[$vararg_buffer31>>3] = $43; //@line 340 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 __Z5b2LogPKcz(447490,$vararg_buffer31); //@line 340 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $44 = ((($this)) + 56|0); //@line 341 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $45 = HEAP32[$44>>2]|0; //@line 341 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$vararg_buffer34>>2] = $45; //@line 341 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 __Z5b2LogPKcz(447513,$vararg_buffer34); //@line 341 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 STACKTOP = sp;return; //@line 342 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
}
function __ZN13b2PulleyJoint11ShiftOriginERK6b2Vec2($this,$newOrigin) {
 $this = $this|0;
 $newOrigin = $newOrigin|0;
 var $0 = 0.0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = +HEAPF32[$newOrigin>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $1 = ((($this)) + 68|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $2 = +HEAPF32[$1>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $3 = $2 - $0; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$1>>2] = $3; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($newOrigin)) + 4|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($this)) + 72|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $8 = $7 - $5; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$6>>2] = $8; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $9 = +HEAPF32[$newOrigin>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($this)) + 76|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $11 = +HEAPF32[$10>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $12 = $11 - $9; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$10>>2] = $12; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $13 = +HEAPF32[$4>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $14 = ((($this)) + 80|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $15 = +HEAPF32[$14>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $16 = $15 - $13; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$14>>2] = $16; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 348 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
}
function __ZN13b2PulleyJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 79 "./libraries/Box2D/Dynamics/Joints/b2PulleyJoint.h"
 return; //@line 79 "./libraries/Box2D/Dynamics/Joints/b2PulleyJoint.h"
}
function __ZN13b2PulleyJoint23InitVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0, $107 = 0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0, $111 = 0.0, $112 = 0.0, $113 = 0, $114 = 0.0, $115 = 0.0;
 var $116 = 0, $117 = 0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0, $121 = 0.0, $122 = 0.0, $123 = 0, $124 = 0.0, $125 = 0.0, $126 = 0, $127 = 0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0.0, $132 = 0.0, $133 = 0.0;
 var $134 = 0.0, $135 = 0.0, $136 = 0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0.0, $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0.0;
 var $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0.0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0.0, $177 = 0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0.0, $182 = 0.0, $183 = 0.0, $184 = 0.0, $185 = 0.0, $186 = 0.0, $187 = 0.0, $188 = 0.0;
 var $189 = 0.0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0.0, $194 = 0, $195 = 0.0, $196 = 0.0, $197 = 0.0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0.0, $201 = 0.0, $202 = 0.0, $203 = 0.0, $204 = 0, $205 = 0.0;
 var $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0.0, $212 = 0.0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0, $53 = 0.0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0.0, $74 = 0.0;
 var $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0.0, $95 = 0.0, $96 = 0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $bitcast = 0.0, $bitcast29 = 0.0, $storemerge = 0.0, $vA$sroa$0$0 = 0, $vA$sroa$6$0 = 0, $vB$sroa$0$0 = 0, $vB$sroa$6$0 = 0, $wA$0 = 0.0, $wB$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 75 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 75 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 75 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 75 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $4 = ((($this)) + 120|0); //@line 75 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$4>>2] = $3; //@line 75 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $5 = ((($this)) + 52|0); //@line 76 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 76 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $7 = ((($6)) + 8|0); //@line 76 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 76 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $9 = ((($this)) + 124|0); //@line 76 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$9>>2] = $8; //@line 76 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $10 = ((($this)) + 160|0); //@line 77 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $11 = ((($1)) + 28|0); //@line 77 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $12 = $11; //@line 77 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $13 = $12; //@line 77 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 77 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $15 = (($12) + 4)|0; //@line 77 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $16 = $15; //@line 77 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 77 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $18 = $10; //@line 77 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $19 = $18; //@line 77 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$19>>2] = $14; //@line 77 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $20 = (($18) + 4)|0; //@line 77 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $21 = $20; //@line 77 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$21>>2] = $17; //@line 77 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $22 = ((($this)) + 168|0); //@line 78 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $23 = ((($6)) + 28|0); //@line 78 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $24 = $23; //@line 78 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $25 = $24; //@line 78 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 78 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $27 = (($24) + 4)|0; //@line 78 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $28 = $27; //@line 78 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 78 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $30 = $22; //@line 78 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $31 = $30; //@line 78 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$31>>2] = $26; //@line 78 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $32 = (($30) + 4)|0; //@line 78 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $33 = $32; //@line 78 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$33>>2] = $29; //@line 78 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $34 = ((($1)) + 120|0); //@line 79 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 79 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $36 = ((($this)) + 176|0); //@line 79 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$36>>2] = $35; //@line 79 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $37 = ((($6)) + 120|0); //@line 80 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $38 = HEAP32[$37>>2]|0; //@line 80 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $39 = ((($this)) + 180|0); //@line 80 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$39>>2] = $38; //@line 80 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $40 = ((($1)) + 128|0); //@line 81 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $41 = HEAP32[$40>>2]|0; //@line 81 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $42 = ((($this)) + 184|0); //@line 81 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$42>>2] = $41; //@line 81 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $43 = ((($6)) + 128|0); //@line 82 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $44 = HEAP32[$43>>2]|0; //@line 82 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $45 = ((($this)) + 188|0); //@line 82 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$45>>2] = $44; //@line 82 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $46 = ((($data)) + 24|0); //@line 84 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $47 = HEAP32[$46>>2]|0; //@line 84 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $48 = (($47) + (($3*12)|0)|0); //@line 84 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $49 = +HEAPF32[$48>>2]; //@line 84 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $50 = (((($47) + (($3*12)|0)|0)) + 4|0); //@line 84 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $51 = +HEAPF32[$50>>2]; //@line 84 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $52 = (((($47) + (($3*12)|0)|0)) + 8|0); //@line 85 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $53 = +HEAPF32[$52>>2]; //@line 85 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $54 = ((($data)) + 28|0); //@line 86 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $55 = HEAP32[$54>>2]|0; //@line 86 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $56 = (($55) + (($3*12)|0)|0); //@line 86 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $57 = HEAP32[$56>>2]|0; //@line 86 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $58 = (((($55) + (($3*12)|0)|0)) + 4|0); //@line 86 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $59 = HEAP32[$58>>2]|0; //@line 86 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $60 = (((($55) + (($3*12)|0)|0)) + 8|0); //@line 87 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $61 = +HEAPF32[$60>>2]; //@line 87 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $62 = (($47) + (($8*12)|0)|0); //@line 89 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $63 = +HEAPF32[$62>>2]; //@line 89 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $64 = (((($47) + (($8*12)|0)|0)) + 4|0); //@line 89 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $65 = +HEAPF32[$64>>2]; //@line 89 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $66 = (((($47) + (($8*12)|0)|0)) + 8|0); //@line 90 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $67 = +HEAPF32[$66>>2]; //@line 90 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $68 = (($55) + (($8*12)|0)|0); //@line 91 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $69 = HEAP32[$68>>2]|0; //@line 91 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $70 = (((($55) + (($8*12)|0)|0)) + 4|0); //@line 91 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $71 = HEAP32[$70>>2]|0; //@line 91 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $72 = (((($55) + (($8*12)|0)|0)) + 8|0); //@line 92 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $73 = +HEAPF32[$72>>2]; //@line 92 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $74 = (+Math_sin((+$53))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $75 = (+Math_cos((+$53))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $76 = (+Math_sin((+$67))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $77 = (+Math_cos((+$67))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $78 = ((($this)) + 92|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $79 = +HEAPF32[$78>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $bitcast = (HEAP32[tempDoublePtr>>2]=$14,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $80 = $79 - $bitcast; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $81 = ((($this)) + 96|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $82 = +HEAPF32[$81>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $83 = (HEAP32[tempDoublePtr>>2]=$17,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $84 = $82 - $83; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $85 = $75 * $80; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $86 = $74 * $84; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $87 = $85 - $86; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $88 = $74 * $80; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $89 = $75 * $84; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $90 = $88 + $89; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $91 = ((($this)) + 144|0); //@line 96 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$91>>2] = $87; //@line 96 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $92 = ((($this)) + 148|0); //@line 96 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$92>>2] = $90; //@line 96 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $93 = ((($this)) + 100|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $94 = +HEAPF32[$93>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $bitcast29 = (HEAP32[tempDoublePtr>>2]=$26,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $95 = $94 - $bitcast29; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $96 = ((($this)) + 104|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $97 = +HEAPF32[$96>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $98 = (HEAP32[tempDoublePtr>>2]=$29,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $99 = $97 - $98; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $100 = $77 * $95; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $101 = $76 * $99; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $102 = $100 - $101; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $103 = $76 * $95; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $104 = $77 * $99; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $105 = $103 + $104; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $106 = ((($this)) + 152|0); //@line 97 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$106>>2] = $102; //@line 97 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $107 = ((($this)) + 156|0); //@line 97 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$107>>2] = $105; //@line 97 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $108 = $49 + $87; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $109 = $51 + $90; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $110 = ((($this)) + 68|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $111 = +HEAPF32[$110>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $112 = $108 - $111; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $113 = ((($this)) + 72|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $114 = +HEAPF32[$113>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $115 = $109 - $114; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 $116 = ((($this)) + 128|0); //@line 100 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$116>>2] = $112; //@line 100 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $117 = ((($this)) + 132|0); //@line 100 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$117>>2] = $115; //@line 100 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $118 = $63 + $102; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $119 = $65 + $105; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $120 = ((($this)) + 76|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $121 = +HEAPF32[$120>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $122 = $118 - $121; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 $123 = ((($this)) + 80|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $124 = +HEAPF32[$123>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $125 = $119 - $124; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 $126 = ((($this)) + 136|0); //@line 101 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$126>>2] = $122; //@line 101 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $127 = ((($this)) + 140|0); //@line 101 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$127>>2] = $125; //@line 101 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $128 = $112 * $112; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $129 = $115 * $115; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $130 = $128 + $129; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $131 = (+Math_sqrt((+$130))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $132 = $122 * $122; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $133 = $125 * $125; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $134 = $132 + $133; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $135 = (+Math_sqrt((+$134))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $136 = $131 > 0.049999997019767761; //@line 106 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 if ($136) {
  $137 = 1.0 / $131; //@line 108 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $138 = $112 * $137; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$116>>2] = $138; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $139 = $137 * $115; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $140 = $139;$148 = $138;
 } else {
  HEAPF32[$116>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $140 = 0.0;$148 = 0.0;
 }
 HEAPF32[$117>>2] = $140; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 $141 = $135 > 0.049999997019767761; //@line 115 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 if ($141) {
  $142 = 1.0 / $135; //@line 117 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $143 = $142 * $122; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$126>>2] = $143; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $144 = $142 * $125; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $145 = $144;$152 = $143;
 } else {
  HEAPF32[$126>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $145 = 0.0;$152 = 0.0;
 }
 HEAPF32[$127>>2] = $145; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 $146 = $87 * $140; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $147 = $90 * $148; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $149 = $146 - $147; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $150 = $102 * $145; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $151 = $105 * $152; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $153 = $150 - $151; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $154 = +HEAPF32[$36>>2]; //@line 147 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $155 = +HEAPF32[$42>>2]; //@line 148 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $156 = $149 * $155; //@line 128 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $157 = $149 * $156; //@line 128 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $158 = $154 + $157; //@line 128 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $159 = +HEAPF32[$39>>2]; //@line 149 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $160 = +HEAPF32[$45>>2]; //@line 150 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $161 = $153 * $160; //@line 129 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $162 = $153 * $161; //@line 129 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $163 = $159 + $162; //@line 129 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $164 = ((($this)) + 112|0); //@line 131 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $165 = +HEAPF32[$164>>2]; //@line 145 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $166 = $165 * $165; //@line 131 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $167 = $166 * $163; //@line 131 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $168 = $158 + $167; //@line 131 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $169 = ((($this)) + 192|0); //@line 131 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $170 = $168 > 0.0; //@line 133 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $171 = 1.0 / $168; //@line 135 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $storemerge = $170 ? $171 : $168; //@line 133 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$169>>2] = $storemerge; //@line 131 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $172 = ((($data)) + 20|0); //@line 138 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $173 = HEAP8[$172>>0]|0; //@line 138 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $174 = ($173<<24>>24)==(0); //@line 138 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 if ($174) {
  $213 = ((($this)) + 116|0); //@line 154 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  HEAPF32[$213>>2] = 0.0; //@line 154 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $vA$sroa$0$0 = $57;$vA$sroa$6$0 = $59;$vB$sroa$0$0 = $69;$vB$sroa$6$0 = $71;$wA$0 = $61;$wB$0 = $73;
  $214 = HEAP32[$4>>2]|0; //@line 157 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $215 = HEAP32[$54>>2]|0; //@line 157 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $216 = (($215) + (($214*12)|0)|0); //@line 157 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  HEAP32[$216>>2] = $vA$sroa$0$0; //@line 157 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $217 = (((($215) + (($214*12)|0)|0)) + 4|0); //@line 157 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  HEAP32[$217>>2] = $vA$sroa$6$0; //@line 157 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $218 = HEAP32[$4>>2]|0; //@line 158 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $219 = HEAP32[$54>>2]|0; //@line 158 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $220 = (((($219) + (($218*12)|0)|0)) + 8|0); //@line 158 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  HEAPF32[$220>>2] = $wA$0; //@line 158 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $221 = HEAP32[$9>>2]|0; //@line 159 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $222 = (($219) + (($221*12)|0)|0); //@line 159 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  HEAP32[$222>>2] = $vB$sroa$0$0; //@line 159 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $223 = (((($219) + (($221*12)|0)|0)) + 4|0); //@line 159 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  HEAP32[$223>>2] = $vB$sroa$6$0; //@line 159 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $224 = HEAP32[$9>>2]|0; //@line 160 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $225 = HEAP32[$54>>2]|0; //@line 160 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $226 = (((($225) + (($224*12)|0)|0)) + 8|0); //@line 160 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  HEAPF32[$226>>2] = $wB$0; //@line 160 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  return; //@line 161 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 } else {
  $175 = ((($data)) + 8|0); //@line 141 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $176 = +HEAPF32[$175>>2]; //@line 141 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $177 = ((($this)) + 116|0); //@line 141 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $178 = +HEAPF32[$177>>2]; //@line 141 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $179 = $176 * $178; //@line 141 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  HEAPF32[$177>>2] = $179; //@line 141 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $180 = -$179; //@line 144 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $181 = $148 * $180; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $182 = $140 * $180; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $183 = $179 * $165; //@line 145 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $184 = -$183; //@line 145 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $185 = $152 * $184; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $186 = $145 * $184; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $187 = $181 * $154; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $188 = $182 * $154; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $189 = (HEAP32[tempDoublePtr>>2]=$57,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $190 = $189 + $187; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $191 = (HEAPF32[tempDoublePtr>>2]=$190,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $192 = (HEAP32[tempDoublePtr>>2]=$59,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $193 = $192 + $188; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $194 = (HEAPF32[tempDoublePtr>>2]=$193,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $195 = $182 * $87; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $196 = $181 * $90; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $197 = $195 - $196; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $198 = $155 * $197; //@line 148 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $199 = $61 + $198; //@line 148 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $200 = $185 * $159; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $201 = $186 * $159; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $202 = (HEAP32[tempDoublePtr>>2]=$69,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $203 = $202 + $200; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $204 = (HEAPF32[tempDoublePtr>>2]=$203,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $205 = (HEAP32[tempDoublePtr>>2]=$71,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $206 = $205 + $201; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $207 = (HEAPF32[tempDoublePtr>>2]=$206,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $208 = $186 * $102; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $209 = $185 * $105; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $210 = $208 - $209; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $211 = $160 * $210; //@line 150 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $212 = $73 + $211; //@line 150 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $vA$sroa$0$0 = $191;$vA$sroa$6$0 = $194;$vB$sroa$0$0 = $204;$vB$sroa$6$0 = $207;$wA$0 = $199;$wB$0 = $212;
  $214 = HEAP32[$4>>2]|0; //@line 157 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $215 = HEAP32[$54>>2]|0; //@line 157 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $216 = (($215) + (($214*12)|0)|0); //@line 157 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  HEAP32[$216>>2] = $vA$sroa$0$0; //@line 157 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $217 = (((($215) + (($214*12)|0)|0)) + 4|0); //@line 157 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  HEAP32[$217>>2] = $vA$sroa$6$0; //@line 157 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $218 = HEAP32[$4>>2]|0; //@line 158 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $219 = HEAP32[$54>>2]|0; //@line 158 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $220 = (((($219) + (($218*12)|0)|0)) + 8|0); //@line 158 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  HEAPF32[$220>>2] = $wA$0; //@line 158 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $221 = HEAP32[$9>>2]|0; //@line 159 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $222 = (($219) + (($221*12)|0)|0); //@line 159 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  HEAP32[$222>>2] = $vB$sroa$0$0; //@line 159 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $223 = (((($219) + (($221*12)|0)|0)) + 4|0); //@line 159 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  HEAP32[$223>>2] = $vB$sroa$6$0; //@line 159 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $224 = HEAP32[$9>>2]|0; //@line 160 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $225 = HEAP32[$54>>2]|0; //@line 160 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $226 = (((($225) + (($224*12)|0)|0)) + 8|0); //@line 160 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  HEAPF32[$226>>2] = $wB$0; //@line 160 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  return; //@line 161 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 }
}
function __ZN13b2PulleyJoint24SolveVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0.0, $44 = 0;
 var $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0;
 var $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0.0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0;
 var $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0.0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 120|0); //@line 165 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 165 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $2 = ((($data)) + 28|0); //@line 165 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 165 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $4 = (($3) + (($1*12)|0)|0); //@line 165 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 165 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $6 = (((($3) + (($1*12)|0)|0)) + 4|0); //@line 165 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 165 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $8 = (((($3) + (($1*12)|0)|0)) + 8|0); //@line 166 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 166 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $10 = ((($this)) + 124|0); //@line 167 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 167 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $12 = (($3) + (($11*12)|0)|0); //@line 167 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $13 = +HEAPF32[$12>>2]; //@line 167 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $14 = (((($3) + (($11*12)|0)|0)) + 4|0); //@line 167 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 167 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $16 = (((($3) + (($11*12)|0)|0)) + 8|0); //@line 168 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 168 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $18 = ((($this)) + 148|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $19 = +HEAPF32[$18>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $20 = $9 * $19; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($this)) + 144|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $22 = +HEAPF32[$21>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $23 = $9 * $22; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $24 = $5 - $20; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $25 = $7 + $23; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $26 = ((($this)) + 156|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $27 = +HEAPF32[$26>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $28 = $17 * $27; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $29 = ((($this)) + 152|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $30 = +HEAPF32[$29>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $31 = $17 * $30; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $32 = $13 - $28; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $33 = $15 + $31; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $34 = ((($this)) + 128|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $35 = +HEAPF32[$34>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $36 = $24 * $35; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $37 = ((($this)) + 132|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $38 = +HEAPF32[$37>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $39 = $25 * $38; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $40 = $36 + $39; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $41 = -$40; //@line 173 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $42 = ((($this)) + 112|0); //@line 173 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $43 = +HEAPF32[$42>>2]; //@line 173 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $44 = ((($this)) + 136|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $45 = +HEAPF32[$44>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $46 = $32 * $45; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $47 = ((($this)) + 140|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $48 = +HEAPF32[$47>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $49 = $33 * $48; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $50 = $46 + $49; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $51 = $43 * $50; //@line 173 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $52 = $41 - $51; //@line 173 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $53 = ((($this)) + 192|0); //@line 174 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $54 = +HEAPF32[$53>>2]; //@line 174 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $55 = $54 * $52; //@line 174 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $56 = ((($this)) + 116|0); //@line 175 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $57 = +HEAPF32[$56>>2]; //@line 175 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $58 = $57 - $55; //@line 175 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$56>>2] = $58; //@line 175 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $59 = $35 * $55; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $60 = $38 * $55; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $61 = $55 * $43; //@line 178 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $62 = $45 * $61; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $63 = $61 * $48; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $64 = ((($this)) + 176|0); //@line 179 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $65 = +HEAPF32[$64>>2]; //@line 179 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $66 = $59 * $65; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $67 = $60 * $65; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $68 = $5 + $66; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $69 = $7 + $67; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $70 = ((($this)) + 184|0); //@line 180 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $71 = +HEAPF32[$70>>2]; //@line 180 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $72 = $60 * $22; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $73 = $59 * $19; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $74 = $72 - $73; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $75 = $71 * $74; //@line 180 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $76 = $9 + $75; //@line 180 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $77 = ((($this)) + 180|0); //@line 181 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $78 = +HEAPF32[$77>>2]; //@line 181 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $79 = $62 * $78; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $80 = $63 * $78; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $81 = $13 + $79; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $82 = $15 + $80; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $83 = ((($this)) + 188|0); //@line 182 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $84 = +HEAPF32[$83>>2]; //@line 182 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $85 = $63 * $30; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $86 = $62 * $27; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $87 = $85 - $86; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $88 = $84 * $87; //@line 182 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $89 = $17 + $88; //@line 182 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$4>>2] = $68; //@line 184 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$6>>2] = $69; //@line 184 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $90 = HEAP32[$0>>2]|0; //@line 185 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $91 = HEAP32[$2>>2]|0; //@line 185 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $92 = (((($91) + (($90*12)|0)|0)) + 8|0); //@line 185 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$92>>2] = $76; //@line 185 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $93 = HEAP32[$10>>2]|0; //@line 186 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $94 = (($91) + (($93*12)|0)|0); //@line 186 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$94>>2] = $81; //@line 186 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $95 = (((($91) + (($93*12)|0)|0)) + 4|0); //@line 186 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$95>>2] = $82; //@line 186 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $96 = HEAP32[$10>>2]|0; //@line 187 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $97 = HEAP32[$2>>2]|0; //@line 187 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $98 = (((($97) + (($96*12)|0)|0)) + 8|0); //@line 187 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$98>>2] = $89; //@line 187 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 return; //@line 188 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
}
function __ZN13b2PulleyJoint24SolvePositionConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0, $112 = 0.0, $113 = 0, $114 = 0.0, $115 = 0.0;
 var $116 = 0.0, $117 = 0.0, $118 = 0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0.0, $130 = 0.0, $131 = 0.0, $132 = 0.0, $133 = 0.0;
 var $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0.0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0, $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0;
 var $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0.0, $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0;
 var $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0.0, $67 = 0;
 var $68 = 0.0, $69 = 0.0, $7 = 0.0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0.0, $85 = 0.0;
 var $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0.0, $92 = 0, $93 = 0.0, $94 = 0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0, $mass$0 = 0.0, $uA$sroa$0$0 = 0.0, $uA$sroa$9$0 = 0.0, $uB$sroa$0$0 = 0.0, $uB$sroa$9$0 = 0.0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 120|0); //@line 192 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 254 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $2 = ((($data)) + 24|0); //@line 192 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 254 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $4 = (($3) + (($1*12)|0)|0); //@line 192 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 192 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $6 = (((($3) + (($1*12)|0)|0)) + 4|0); //@line 192 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 192 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $8 = (((($3) + (($1*12)|0)|0)) + 8|0); //@line 193 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 193 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $10 = ((($this)) + 124|0); //@line 194 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 194 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $12 = (($3) + (($11*12)|0)|0); //@line 194 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $13 = +HEAPF32[$12>>2]; //@line 194 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $14 = (((($3) + (($11*12)|0)|0)) + 4|0); //@line 194 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 194 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $16 = (((($3) + (($11*12)|0)|0)) + 8|0); //@line 195 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 195 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $18 = (+Math_sin((+$9))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $19 = (+Math_cos((+$9))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $20 = (+Math_sin((+$17))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $21 = (+Math_cos((+$17))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($this)) + 92|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $23 = +HEAPF32[$22>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($this)) + 160|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $26 = $23 - $25; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($this)) + 96|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $28 = +HEAPF32[$27>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $29 = ((($this)) + 164|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $30 = +HEAPF32[$29>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $31 = $28 - $30; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $32 = $19 * $26; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $33 = $18 * $31; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $34 = $32 - $33; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $35 = $18 * $26; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $36 = $19 * $31; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $37 = $35 + $36; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $38 = ((($this)) + 100|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $39 = +HEAPF32[$38>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $40 = ((($this)) + 168|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $41 = +HEAPF32[$40>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $42 = $39 - $41; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $43 = ((($this)) + 104|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $44 = +HEAPF32[$43>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $45 = ((($this)) + 172|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $46 = +HEAPF32[$45>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $47 = $44 - $46; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $48 = $21 * $42; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $49 = $20 * $47; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $50 = $48 - $49; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $51 = $20 * $42; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $52 = $21 * $47; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $53 = $51 + $52; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $54 = $5 + $34; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $55 = $7 + $37; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $56 = ((($this)) + 68|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $57 = +HEAPF32[$56>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $58 = $54 - $57; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $59 = ((($this)) + 72|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $60 = +HEAPF32[$59>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $61 = $55 - $60; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $62 = $13 + $50; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $63 = $15 + $53; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $64 = ((($this)) + 76|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $65 = +HEAPF32[$64>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $66 = $62 - $65; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $67 = ((($this)) + 80|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $68 = +HEAPF32[$67>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $69 = $63 - $68; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $70 = $58 * $58; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $71 = $61 * $61; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $72 = $70 + $71; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $73 = (+Math_sqrt((+$72))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $74 = $66 * $66; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $75 = $69 * $69; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $76 = $74 + $75; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $77 = (+Math_sqrt((+$76))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $78 = $73 > 0.049999997019767761; //@line 209 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 if ($78) {
  $79 = 1.0 / $73; //@line 211 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $80 = $58 * $79; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $81 = $61 * $79; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $uA$sroa$0$0 = $80;$uA$sroa$9$0 = $81;
 } else {
  $uA$sroa$0$0 = 0.0;$uA$sroa$9$0 = 0.0;
 }
 $82 = $77 > 0.049999997019767761; //@line 218 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 if ($82) {
  $83 = 1.0 / $77; //@line 220 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
  $84 = $66 * $83; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $85 = $69 * $83; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $uB$sroa$0$0 = $84;$uB$sroa$9$0 = $85;
 } else {
  $uB$sroa$0$0 = 0.0;$uB$sroa$9$0 = 0.0;
 }
 $86 = $34 * $uA$sroa$9$0; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $87 = $37 * $uA$sroa$0$0; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $88 = $86 - $87; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $89 = $50 * $uB$sroa$9$0; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $90 = $53 * $uB$sroa$0$0; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $91 = $89 - $90; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $92 = ((($this)) + 176|0); //@line 231 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $93 = +HEAPF32[$92>>2]; //@line 231 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $94 = ((($this)) + 184|0); //@line 231 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $95 = +HEAPF32[$94>>2]; //@line 231 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $96 = $88 * $95; //@line 231 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $97 = $88 * $96; //@line 231 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $98 = $93 + $97; //@line 231 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $99 = ((($this)) + 180|0); //@line 232 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $100 = +HEAPF32[$99>>2]; //@line 232 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $101 = ((($this)) + 188|0); //@line 232 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $102 = +HEAPF32[$101>>2]; //@line 232 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $103 = $91 * $102; //@line 232 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $104 = $91 * $103; //@line 232 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $105 = $100 + $104; //@line 232 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $106 = ((($this)) + 112|0); //@line 234 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $107 = +HEAPF32[$106>>2]; //@line 234 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $108 = $107 * $107; //@line 234 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $109 = $108 * $105; //@line 234 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $110 = $98 + $109; //@line 234 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $111 = $110 > 0.0; //@line 236 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $112 = 1.0 / $110; //@line 238 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $mass$0 = $111 ? $112 : $110; //@line 236 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $113 = ((($this)) + 108|0); //@line 241 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $114 = +HEAPF32[$113>>2]; //@line 241 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $115 = $114 - $73; //@line 241 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $116 = $77 * $107; //@line 241 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $117 = $115 - $116; //@line 241 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $118 = $117 > 0.0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $119 = -$117; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $120 = $118 ? $117 : $119; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $121 = $117 * $mass$0; //@line 244 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $122 = $uA$sroa$0$0 * $121; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $123 = $uA$sroa$9$0 * $121; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $124 = $107 * $121; //@line 247 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $125 = $uB$sroa$0$0 * $124; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $126 = $uB$sroa$9$0 * $124; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $127 = $93 * $122; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $128 = $93 * $123; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $129 = $5 + $127; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $130 = $7 + $128; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $131 = $34 * $123; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $132 = $37 * $122; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $133 = $131 - $132; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $134 = $95 * $133; //@line 250 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $135 = $9 + $134; //@line 250 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $136 = $100 * $125; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $137 = $100 * $126; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $138 = $13 + $136; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $139 = $15 + $137; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $140 = $50 * $126; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $141 = $53 * $125; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $142 = $140 - $141; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $143 = $102 * $142; //@line 252 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $144 = $17 + $143; //@line 252 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$4>>2] = $129; //@line 254 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$6>>2] = $130; //@line 254 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $145 = HEAP32[$0>>2]|0; //@line 255 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $146 = HEAP32[$2>>2]|0; //@line 255 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $147 = (((($146) + (($145*12)|0)|0)) + 8|0); //@line 255 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$147>>2] = $135; //@line 255 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $148 = HEAP32[$10>>2]|0; //@line 256 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $149 = (($146) + (($148*12)|0)|0); //@line 256 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$149>>2] = $138; //@line 256 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $150 = (((($146) + (($148*12)|0)|0)) + 4|0); //@line 256 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$150>>2] = $139; //@line 256 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $151 = HEAP32[$10>>2]|0; //@line 257 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $152 = HEAP32[$2>>2]|0; //@line 257 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $153 = (((($152) + (($151*12)|0)|0)) + 8|0); //@line 257 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$153>>2] = $144; //@line 257 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $154 = $120 < 0.004999999888241291; //@line 259 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 return ($154|0); //@line 259 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
}
function __ZN16b2PulleyJointDef10InitializeEP6b2BodyS1_RK6b2Vec2S4_S4_S4_f($this,$bA,$bB,$groundA,$groundB,$anchorA,$anchorB,$r) {
 $this = $this|0;
 $bA = $bA|0;
 $bB = $bB|0;
 $groundA = $groundA|0;
 $groundB = $groundB|0;
 $anchorA = $anchorA|0;
 $anchorB = $anchorB|0;
 $r = +$r;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0.0, $25 = 0, $26 = 0.0;
 var $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0;
 var $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0, $55 = 0.0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0;
 var $63 = 0.0, $64 = 0, $65 = 0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0;
 var $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 40 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$0>>2] = $bA; //@line 40 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $1 = ((($this)) + 12|0); //@line 41 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$1>>2] = $bB; //@line 41 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $2 = ((($this)) + 20|0); //@line 42 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $3 = $groundA; //@line 42 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $4 = $3; //@line 42 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 42 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $6 = (($3) + 4)|0; //@line 42 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $7 = $6; //@line 42 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 42 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $9 = $2; //@line 42 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $10 = $9; //@line 42 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$10>>2] = $5; //@line 42 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $11 = (($9) + 4)|0; //@line 42 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $12 = $11; //@line 42 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$12>>2] = $8; //@line 42 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $13 = ((($this)) + 28|0); //@line 43 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $14 = $groundB; //@line 43 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $15 = $14; //@line 43 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 43 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $17 = (($14) + 4)|0; //@line 43 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $18 = $17; //@line 43 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 43 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $20 = $13; //@line 43 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $21 = $20; //@line 43 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$21>>2] = $16; //@line 43 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $22 = (($20) + 4)|0; //@line 43 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $23 = $22; //@line 43 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$23>>2] = $19; //@line 43 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $24 = +HEAPF32[$anchorA>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $25 = ((($bA)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $26 = +HEAPF32[$25>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $27 = $24 - $26; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $28 = ((($anchorA)) + 4|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $29 = +HEAPF32[$28>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $30 = ((($bA)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $31 = +HEAPF32[$30>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $32 = $29 - $31; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $33 = ((($bA)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $34 = +HEAPF32[$33>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $35 = $27 * $34; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $36 = ((($bA)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $37 = +HEAPF32[$36>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $38 = $32 * $37; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $39 = $35 + $38; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $40 = $27 * $37; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $41 = $34 * $32; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $42 = $41 - $40; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $43 = ((($this)) + 36|0); //@line 44 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$43>>2] = $39; //@line 44 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $44 = ((($this)) + 40|0); //@line 44 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$44>>2] = $42; //@line 44 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $45 = +HEAPF32[$anchorB>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $46 = ((($bB)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $47 = +HEAPF32[$46>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $48 = $45 - $47; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $49 = ((($anchorB)) + 4|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $50 = +HEAPF32[$49>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $51 = ((($bB)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $52 = +HEAPF32[$51>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $53 = $50 - $52; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $54 = ((($bB)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $55 = +HEAPF32[$54>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $56 = $48 * $55; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $57 = ((($bB)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $58 = +HEAPF32[$57>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $59 = $53 * $58; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $60 = $56 + $59; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $61 = $48 * $58; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $62 = $55 * $53; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $63 = $62 - $61; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $64 = ((($this)) + 44|0); //@line 45 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$64>>2] = $60; //@line 45 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $65 = ((($this)) + 48|0); //@line 45 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$65>>2] = $63; //@line 45 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $66 = +HEAPF32[$anchorA>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $67 = +HEAPF32[$groundA>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $68 = $66 - $67; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $69 = +HEAPF32[$28>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $70 = ((($groundA)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $71 = +HEAPF32[$70>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $72 = $69 - $71; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $73 = $68 * $68; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $74 = $72 * $72; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $75 = $73 + $74; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $76 = (+Math_sqrt((+$75))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $77 = ((($this)) + 52|0); //@line 47 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$77>>2] = $76; //@line 47 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $78 = +HEAPF32[$anchorB>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $79 = +HEAPF32[$groundB>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $80 = $78 - $79; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $81 = +HEAPF32[$49>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $82 = ((($groundB)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $83 = +HEAPF32[$82>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $84 = $81 - $83; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $85 = $80 * $80; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $86 = $84 * $84; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $87 = $85 + $86; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $88 = (+Math_sqrt((+$87))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $89 = ((($this)) + 56|0); //@line 49 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$89>>2] = $88; //@line 49 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $90 = ((($this)) + 60|0); //@line 50 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$90>>2] = $r; //@line 50 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $91 = $r > 1.1920928955078125E-7; //@line 51 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 __Z10loveAssertbPKc($91,363164); //@line 51 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 return; //@line 52 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
}
function __ZN13b2PulleyJointC2EPK16b2PulleyJointDef($this,$def) {
 $this = $this|0;
 $def = $def|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0;
 var $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN7b2JointC2EPK10b2JointDef($this,$def); //@line 55 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$this>>2] = (20920); //@line 56 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $0 = ((($this)) + 68|0); //@line 54 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $1 = ((($this)) + 76|0); //@line 54 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $2 = ((($this)) + 92|0); //@line 54 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $3 = ((($this)) + 100|0); //@line 54 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $4 = ((($def)) + 20|0); //@line 57 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $5 = $4; //@line 57 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $6 = $5; //@line 57 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 57 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $8 = (($5) + 4)|0; //@line 57 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $9 = $8; //@line 57 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 57 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $11 = $0; //@line 57 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $12 = $11; //@line 57 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$12>>2] = $7; //@line 57 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $13 = (($11) + 4)|0; //@line 57 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $14 = $13; //@line 57 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$14>>2] = $10; //@line 57 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $15 = ((($def)) + 28|0); //@line 58 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $16 = $15; //@line 58 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $17 = $16; //@line 58 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 58 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $19 = (($16) + 4)|0; //@line 58 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $20 = $19; //@line 58 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 58 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $22 = $1; //@line 58 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $23 = $22; //@line 58 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$23>>2] = $18; //@line 58 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $24 = (($22) + 4)|0; //@line 58 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $25 = $24; //@line 58 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$25>>2] = $21; //@line 58 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $26 = ((($def)) + 36|0); //@line 59 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $27 = $26; //@line 59 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $28 = $27; //@line 59 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 59 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $30 = (($27) + 4)|0; //@line 59 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $31 = $30; //@line 59 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $32 = HEAP32[$31>>2]|0; //@line 59 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $33 = $2; //@line 59 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $34 = $33; //@line 59 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$34>>2] = $29; //@line 59 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $35 = (($33) + 4)|0; //@line 59 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $36 = $35; //@line 59 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$36>>2] = $32; //@line 59 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $37 = ((($def)) + 44|0); //@line 60 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $38 = $37; //@line 60 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $39 = $38; //@line 60 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $40 = HEAP32[$39>>2]|0; //@line 60 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $41 = (($38) + 4)|0; //@line 60 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $42 = $41; //@line 60 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $43 = HEAP32[$42>>2]|0; //@line 60 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $44 = $3; //@line 60 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $45 = $44; //@line 60 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$45>>2] = $40; //@line 60 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $46 = (($44) + 4)|0; //@line 60 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $47 = $46; //@line 60 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$47>>2] = $43; //@line 60 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $48 = ((($def)) + 52|0); //@line 62 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $49 = HEAP32[$48>>2]|0; //@line 62 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $50 = ((($this)) + 84|0); //@line 62 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$50>>2] = $49; //@line 62 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $51 = ((($def)) + 56|0); //@line 63 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $52 = HEAP32[$51>>2]|0; //@line 63 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $53 = ((($this)) + 88|0); //@line 63 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$53>>2] = $52; //@line 63 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $54 = ((($def)) + 60|0); //@line 65 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $55 = +HEAPF32[$54>>2]; //@line 65 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $56 = $55 != 0.0; //@line 65 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 __Z10loveAssertbPKc($56,363183); //@line 65 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $57 = +HEAPF32[$54>>2]; //@line 66 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $58 = ((($this)) + 112|0); //@line 66 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$58>>2] = $57; //@line 66 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $59 = +HEAPF32[$48>>2]; //@line 68 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $60 = +HEAPF32[$51>>2]; //@line 68 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $61 = $57 * $60; //@line 68 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $62 = $59 + $61; //@line 68 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $63 = ((($this)) + 108|0); //@line 68 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$63>>2] = $62; //@line 68 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $64 = ((($this)) + 116|0); //@line 70 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAPF32[$64>>2] = 0.0; //@line 70 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 return; //@line 71 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
}
function __ZNK13b2PulleyJoint16GetGroundAnchorAEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 68|0); //@line 286 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $1 = $0; //@line 286 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $2 = $1; //@line 286 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 286 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $4 = (($1) + 4)|0; //@line 286 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $5 = $4; //@line 286 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 286 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $7 = $agg$result; //@line 286 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $8 = $7; //@line 286 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$8>>2] = $3; //@line 286 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $9 = (($7) + 4)|0; //@line 286 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $10 = $9; //@line 286 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$10>>2] = $6; //@line 286 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 return; //@line 286 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
}
function __ZNK13b2PulleyJoint16GetGroundAnchorBEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 76|0); //@line 291 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $1 = $0; //@line 291 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $2 = $1; //@line 291 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 291 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $4 = (($1) + 4)|0; //@line 291 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $5 = $4; //@line 291 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 291 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $7 = $agg$result; //@line 291 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $8 = $7; //@line 291 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$8>>2] = $3; //@line 291 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $9 = (($7) + 4)|0; //@line 291 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $10 = $9; //@line 291 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 HEAP32[$10>>2] = $6; //@line 291 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 return; //@line 291 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
}
function __ZNK13b2PulleyJoint10GetLengthAEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 84|0); //@line 296 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 296 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 return (+$1); //@line 296 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
}
function __ZNK13b2PulleyJoint10GetLengthBEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 88|0); //@line 301 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 301 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 return (+$1); //@line 301 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
}
function __ZNK13b2PulleyJoint8GetRatioEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 306 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 306 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
 return (+$1); //@line 306 "libraries/Box2D/Dynamics/Joints/b2PulleyJoint.cpp"
}
function __ZNK15b2RevoluteJoint10GetAnchorAEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 379 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 379 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 68|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 72|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 379 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZNK15b2RevoluteJoint10GetAnchorBEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 384 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 384 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 76|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 80|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 384 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZNK15b2RevoluteJoint16GetReactionForceEf($agg$result,$this,$inv_dt) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 84|0); //@line 389 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 389 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = ((($this)) + 88|0); //@line 389 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 389 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $4 = $1 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $5 = $3 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $4; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$6>>2] = $5; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 391 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZNK15b2RevoluteJoint17GetReactionTorqueEf($this,$inv_dt) {
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 92|0); //@line 395 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 395 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = $1 * $inv_dt; //@line 395 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 return (+$2); //@line 395 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZN15b2RevoluteJoint4DumpEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0;
 var $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer13 = 0, $vararg_buffer17 = 0, $vararg_buffer20 = 0, $vararg_buffer23 = 0, $vararg_buffer26 = 0, $vararg_buffer29 = 0, $vararg_buffer3 = 0, $vararg_buffer32 = 0, $vararg_buffer35 = 0, $vararg_buffer38 = 0, $vararg_buffer6 = 0, $vararg_buffer9 = 0;
 var $vararg_ptr12 = 0, $vararg_ptr16 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer38 = sp + 120|0;
 $vararg_buffer35 = sp + 112|0;
 $vararg_buffer32 = sp + 104|0;
 $vararg_buffer29 = sp + 96|0;
 $vararg_buffer26 = sp + 88|0;
 $vararg_buffer23 = sp + 80|0;
 $vararg_buffer20 = sp + 72|0;
 $vararg_buffer17 = sp + 64|0;
 $vararg_buffer13 = sp + 48|0;
 $vararg_buffer9 = sp + 32|0;
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($this)) + 48|0); //@line 485 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 485 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 485 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 485 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $4 = ((($this)) + 52|0); //@line 486 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 486 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $6 = ((($5)) + 8|0); //@line 486 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 486 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 __Z5b2LogPKcz(363220,$vararg_buffer); //@line 488 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$vararg_buffer1>>2] = $3; //@line 489 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 __Z5b2LogPKcz(447349,$vararg_buffer1); //@line 489 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$vararg_buffer3>>2] = $7; //@line 490 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 __Z5b2LogPKcz(447375,$vararg_buffer3); //@line 490 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $8 = ((($this)) + 61|0); //@line 491 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 491 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $10 = $9&255; //@line 491 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$vararg_buffer6>>2] = $10; //@line 491 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 __Z5b2LogPKcz(447401,$vararg_buffer6); //@line 491 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $11 = ((($this)) + 68|0); //@line 492 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $12 = +HEAPF32[$11>>2]; //@line 492 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $13 = $12; //@line 492 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $14 = ((($this)) + 72|0); //@line 492 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 492 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $16 = $15; //@line 492 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF64[$vararg_buffer9>>3] = $13; //@line 492 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $vararg_ptr12 = ((($vararg_buffer9)) + 8|0); //@line 492 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF64[$vararg_ptr12>>3] = $16; //@line 492 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 __Z5b2LogPKcz(447106,$vararg_buffer9); //@line 492 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $17 = ((($this)) + 76|0); //@line 493 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $18 = +HEAPF32[$17>>2]; //@line 493 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $19 = $18; //@line 493 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $20 = ((($this)) + 80|0); //@line 493 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 493 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $22 = $21; //@line 493 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF64[$vararg_buffer13>>3] = $19; //@line 493 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $vararg_ptr16 = ((($vararg_buffer13)) + 8|0); //@line 493 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF64[$vararg_ptr16>>3] = $22; //@line 493 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 __Z5b2LogPKcz(447148,$vararg_buffer13); //@line 493 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $23 = ((($this)) + 116|0); //@line 494 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $24 = +HEAPF32[$23>>2]; //@line 494 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $25 = $24; //@line 494 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF64[$vararg_buffer17>>3] = $25; //@line 494 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 __Z5b2LogPKcz(363446,$vararg_buffer17); //@line 494 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $26 = ((($this)) + 112|0); //@line 495 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $27 = HEAP8[$26>>0]|0; //@line 495 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $28 = $27&255; //@line 495 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$vararg_buffer20>>2] = $28; //@line 495 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 __Z5b2LogPKcz(363246,$vararg_buffer20); //@line 495 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $29 = ((($this)) + 120|0); //@line 496 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $30 = +HEAPF32[$29>>2]; //@line 496 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $31 = $30; //@line 496 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF64[$vararg_buffer23>>3] = $31; //@line 496 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 __Z5b2LogPKcz(363276,$vararg_buffer23); //@line 496 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $32 = ((($this)) + 124|0); //@line 497 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $33 = +HEAPF32[$32>>2]; //@line 497 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $34 = $33; //@line 497 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF64[$vararg_buffer26>>3] = $34; //@line 497 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 __Z5b2LogPKcz(363304,$vararg_buffer26); //@line 497 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $35 = ((($this)) + 100|0); //@line 498 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $36 = HEAP8[$35>>0]|0; //@line 498 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $37 = $36&255; //@line 498 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$vararg_buffer29>>2] = $37; //@line 498 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 __Z5b2LogPKcz(363556,$vararg_buffer29); //@line 498 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $38 = ((($this)) + 108|0); //@line 499 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $39 = +HEAPF32[$38>>2]; //@line 499 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $40 = $39; //@line 499 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF64[$vararg_buffer32>>3] = $40; //@line 499 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 __Z5b2LogPKcz(363586,$vararg_buffer32); //@line 499 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $41 = ((($this)) + 104|0); //@line 500 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $42 = +HEAPF32[$41>>2]; //@line 500 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $43 = $42; //@line 500 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF64[$vararg_buffer35>>3] = $43; //@line 500 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 __Z5b2LogPKcz(363614,$vararg_buffer35); //@line 500 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $44 = ((($this)) + 56|0); //@line 501 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $45 = HEAP32[$44>>2]|0; //@line 501 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$vararg_buffer38>>2] = $45; //@line 501 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 __Z5b2LogPKcz(447513,$vararg_buffer38); //@line 501 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 STACKTOP = sp;return; //@line 502 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZN15b2RevoluteJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 90 "./libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.h"
 return; //@line 90 "./libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.h"
}
function __ZN15b2RevoluteJoint23InitVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0, $114 = 0.0, $115 = 0.0;
 var $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0, $126 = 0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0.0, $132 = 0.0, $133 = 0;
 var $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0.0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0.0;
 var $152 = 0, $153 = 0.0, $154 = 0.0, $155 = 0, $156 = 0.0, $157 = 0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0, $161 = 0.0, $162 = 0.0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0.0, $183 = 0.0, $184 = 0.0, $185 = 0, $186 = 0.0, $187 = 0.0, $188 = 0;
 var $189 = 0.0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0.0, $194 = 0.0, $195 = 0.0, $196 = 0.0, $197 = 0.0, $198 = 0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0.0, $201 = 0, $202 = 0.0, $203 = 0.0, $204 = 0.0, $205 = 0.0;
 var $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0.0, $212 = 0.0, $213 = 0, $214 = 0.0, $215 = 0.0, $216 = 0, $217 = 0.0, $218 = 0.0, $219 = 0.0, $22 = 0, $220 = 0.0, $221 = 0.0, $222 = 0.0, $223 = 0.0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0.0;
 var $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0.0, $87 = 0.0, $88 = 0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0, $99 = 0, $bitcast = 0.0, $bitcast16 = 0.0, $brmerge = 0;
 var $brmerge4 = 0, $storemerge = 0.0, $vA$sroa$0$0 = 0, $vA$sroa$6$0 = 0, $vB$sroa$0$0 = 0, $vB$sroa$6$0 = 0, $wA$0 = 0.0, $wB$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 66 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 66 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 66 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 66 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $4 = ((($this)) + 128|0); //@line 66 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$4>>2] = $3; //@line 66 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $5 = ((($this)) + 52|0); //@line 67 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 67 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $7 = ((($6)) + 8|0); //@line 67 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 67 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $9 = ((($this)) + 132|0); //@line 67 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$9>>2] = $8; //@line 67 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $10 = ((($this)) + 152|0); //@line 68 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $11 = ((($1)) + 28|0); //@line 68 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $12 = $11; //@line 68 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $13 = $12; //@line 68 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 68 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $15 = (($12) + 4)|0; //@line 68 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $16 = $15; //@line 68 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 68 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $18 = $10; //@line 68 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $19 = $18; //@line 68 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$19>>2] = $14; //@line 68 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $20 = (($18) + 4)|0; //@line 68 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $21 = $20; //@line 68 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$21>>2] = $17; //@line 68 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $22 = ((($this)) + 160|0); //@line 69 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $23 = ((($6)) + 28|0); //@line 69 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $24 = $23; //@line 69 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $25 = $24; //@line 69 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 69 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $27 = (($24) + 4)|0; //@line 69 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $28 = $27; //@line 69 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 69 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $30 = $22; //@line 69 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $31 = $30; //@line 69 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$31>>2] = $26; //@line 69 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $32 = (($30) + 4)|0; //@line 69 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $33 = $32; //@line 69 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$33>>2] = $29; //@line 69 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $34 = ((($1)) + 120|0); //@line 70 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 70 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $36 = ((($this)) + 168|0); //@line 70 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$36>>2] = $35; //@line 70 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $37 = ((($6)) + 120|0); //@line 71 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $38 = HEAP32[$37>>2]|0; //@line 71 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $39 = ((($this)) + 172|0); //@line 71 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$39>>2] = $38; //@line 71 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $40 = ((($1)) + 128|0); //@line 72 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $41 = HEAP32[$40>>2]|0; //@line 72 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $42 = ((($this)) + 176|0); //@line 72 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$42>>2] = $41; //@line 72 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $43 = ((($6)) + 128|0); //@line 73 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $44 = HEAP32[$43>>2]|0; //@line 73 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $45 = ((($this)) + 180|0); //@line 73 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$45>>2] = $44; //@line 73 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $46 = ((($data)) + 24|0); //@line 75 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $47 = HEAP32[$46>>2]|0; //@line 75 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $48 = (((($47) + (($3*12)|0)|0)) + 8|0); //@line 75 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $49 = +HEAPF32[$48>>2]; //@line 75 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $50 = ((($data)) + 28|0); //@line 76 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $51 = HEAP32[$50>>2]|0; //@line 76 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $52 = (($51) + (($3*12)|0)|0); //@line 76 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $53 = HEAP32[$52>>2]|0; //@line 76 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $54 = (((($51) + (($3*12)|0)|0)) + 4|0); //@line 76 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $55 = HEAP32[$54>>2]|0; //@line 76 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $56 = (((($51) + (($3*12)|0)|0)) + 8|0); //@line 77 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $57 = +HEAPF32[$56>>2]; //@line 77 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $58 = (((($47) + (($8*12)|0)|0)) + 8|0); //@line 79 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $59 = +HEAPF32[$58>>2]; //@line 79 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $60 = (($51) + (($8*12)|0)|0); //@line 80 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $61 = HEAP32[$60>>2]|0; //@line 80 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $62 = (((($51) + (($8*12)|0)|0)) + 4|0); //@line 80 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $63 = HEAP32[$62>>2]|0; //@line 80 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $64 = (((($51) + (($8*12)|0)|0)) + 8|0); //@line 81 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $65 = +HEAPF32[$64>>2]; //@line 81 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $66 = (+Math_sin((+$49))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $67 = (+Math_cos((+$49))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $68 = (+Math_sin((+$59))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $69 = (+Math_cos((+$59))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $70 = ((($this)) + 68|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $71 = +HEAPF32[$70>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $bitcast = (HEAP32[tempDoublePtr>>2]=$14,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $72 = $71 - $bitcast; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $73 = ((($this)) + 72|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $74 = +HEAPF32[$73>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $75 = (HEAP32[tempDoublePtr>>2]=$17,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $76 = $74 - $75; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $77 = $67 * $72; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $78 = $66 * $76; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $79 = $77 - $78; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $80 = $66 * $72; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $81 = $67 * $76; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $82 = $80 + $81; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $83 = ((($this)) + 136|0); //@line 85 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$83>>2] = $79; //@line 85 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $84 = ((($this)) + 140|0); //@line 85 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$84>>2] = $82; //@line 85 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $85 = ((($this)) + 76|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $86 = +HEAPF32[$85>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $bitcast16 = (HEAP32[tempDoublePtr>>2]=$26,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $87 = $86 - $bitcast16; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $88 = ((($this)) + 80|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $89 = +HEAPF32[$88>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $90 = (HEAP32[tempDoublePtr>>2]=$29,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $91 = $89 - $90; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $92 = $69 * $87; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $93 = $68 * $91; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $94 = $92 - $93; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $95 = $68 * $87; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $96 = $69 * $91; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $97 = $95 + $96; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $98 = ((($this)) + 144|0); //@line 86 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$98>>2] = $94; //@line 86 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $99 = ((($this)) + 148|0); //@line 86 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$99>>2] = $97; //@line 86 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $100 = (HEAP32[tempDoublePtr>>2]=$35,+HEAPF32[tempDoublePtr>>2]); //@line 97 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $101 = (HEAP32[tempDoublePtr>>2]=$38,+HEAPF32[tempDoublePtr>>2]); //@line 97 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $102 = (HEAP32[tempDoublePtr>>2]=$41,+HEAPF32[tempDoublePtr>>2]); //@line 98 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $103 = (HEAP32[tempDoublePtr>>2]=$44,+HEAPF32[tempDoublePtr>>2]); //@line 98 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $104 = $102 + $103; //@line 100 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $105 = $104 == 0.0; //@line 100 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $106 = $100 + $101; //@line 102 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $107 = $82 * $82; //@line 102 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $108 = $102 * $107; //@line 102 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $109 = $106 + $108; //@line 102 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $110 = $97 * $97; //@line 102 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $111 = $103 * $110; //@line 102 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $112 = $109 + $111; //@line 102 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $113 = ((($this)) + 184|0); //@line 102 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$113>>2] = $112; //@line 102 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $114 = $82 * $79; //@line 103 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $115 = $102 * $114; //@line 103 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $116 = -$115; //@line 103 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $117 = $97 * $94; //@line 103 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $118 = $103 * $117; //@line 103 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $119 = $116 - $118; //@line 103 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $120 = ((($this)) + 196|0); //@line 103 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$120>>2] = $119; //@line 103 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $121 = $102 * $82; //@line 104 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $122 = -$121; //@line 104 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $123 = $103 * $97; //@line 104 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $124 = $122 - $123; //@line 104 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $125 = ((($this)) + 208|0); //@line 104 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$125>>2] = $124; //@line 104 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $126 = ((($this)) + 188|0); //@line 105 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$126>>2] = $119; //@line 105 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $127 = $79 * $79; //@line 106 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $128 = $102 * $127; //@line 106 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $129 = $106 + $128; //@line 106 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $130 = $94 * $94; //@line 106 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $131 = $103 * $130; //@line 106 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $132 = $129 + $131; //@line 106 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $133 = ((($this)) + 200|0); //@line 106 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$133>>2] = $132; //@line 106 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $134 = $102 * $79; //@line 107 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $135 = $103 * $94; //@line 107 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $136 = $134 + $135; //@line 107 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $137 = ((($this)) + 212|0); //@line 107 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$137>>2] = $136; //@line 107 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $138 = ((($this)) + 192|0); //@line 108 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$138>>2] = $124; //@line 108 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $139 = ((($this)) + 204|0); //@line 109 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$139>>2] = $136; //@line 109 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $140 = ((($this)) + 216|0); //@line 110 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$140>>2] = $104; //@line 110 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $141 = ((($this)) + 220|0); //@line 112 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $142 = $104 > 0.0; //@line 113 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $143 = 1.0 / $104; //@line 115 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $storemerge = $142 ? $143 : $104; //@line 113 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$141>>2] = $storemerge; //@line 112 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $144 = ((($this)) + 100|0); //@line 118 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $145 = HEAP8[$144>>0]|0; //@line 118 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $146 = ($145<<24>>24)==(0); //@line 118 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $brmerge = $105 | $146; //@line 118 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 if ($brmerge) {
  $147 = ((($this)) + 96|0); //@line 120 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  HEAPF32[$147>>2] = 0.0; //@line 120 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 }
 $148 = ((($this)) + 112|0); //@line 123 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $149 = HEAP8[$148>>0]|0; //@line 123 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $150 = ($149<<24>>24)==(0); //@line 123 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $brmerge4 = $105 | $150; //@line 123 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 do {
  if ($brmerge4) {
   $176 = ((($this)) + 224|0); //@line 154 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   HEAP32[$176>>2] = 0; //@line 154 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  } else {
   $151 = $59 - $49; //@line 125 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $152 = ((($this)) + 116|0); //@line 125 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $153 = +HEAPF32[$152>>2]; //@line 125 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $154 = $151 - $153; //@line 125 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $155 = ((($this)) + 124|0); //@line 126 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $156 = +HEAPF32[$155>>2]; //@line 126 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $157 = ((($this)) + 120|0); //@line 126 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $158 = +HEAPF32[$157>>2]; //@line 126 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $159 = $156 - $158; //@line 126 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $160 = $159 > 0.0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
   $161 = -$159; //@line 618 "./libraries/Box2D/Common/b2Math.h"
   $162 = $160 ? $159 : $161; //@line 618 "./libraries/Box2D/Common/b2Math.h"
   $163 = $162 < 0.069813176989555359; //@line 126 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   if ($163) {
    $164 = ((($this)) + 224|0); //@line 128 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    HEAP32[$164>>2] = 3; //@line 128 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    break;
   }
   $165 = !($154 <= $158); //@line 130 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   if (!($165)) {
    $166 = ((($this)) + 224|0); //@line 132 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $167 = HEAP32[$166>>2]|0; //@line 132 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $168 = ($167|0)==(1); //@line 132 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    if (!($168)) {
     $169 = ((($this)) + 92|0); //@line 134 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
     HEAPF32[$169>>2] = 0.0; //@line 134 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    }
    HEAP32[$166>>2] = 1; //@line 136 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    break;
   }
   $170 = !($154 >= $156); //@line 138 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $171 = ((($this)) + 224|0); //@line 140 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   if ($170) {
    HEAP32[$171>>2] = 0; //@line 148 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $175 = ((($this)) + 92|0); //@line 149 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    HEAPF32[$175>>2] = 0.0; //@line 149 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    break;
   }
   $172 = HEAP32[$171>>2]|0; //@line 140 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $173 = ($172|0)==(2); //@line 140 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   if (!($173)) {
    $174 = ((($this)) + 92|0); //@line 142 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    HEAPF32[$174>>2] = 0.0; //@line 142 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   }
   HEAP32[$171>>2] = 2; //@line 144 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  }
 } while(0);
 $177 = ((($data)) + 20|0); //@line 157 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $178 = HEAP8[$177>>0]|0; //@line 157 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $179 = ($178<<24>>24)==(0); //@line 157 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $180 = ((($this)) + 84|0); //@line 160 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 if ($179) {
  ;HEAP32[$180>>2]=0|0;HEAP32[$180+4>>2]=0|0;HEAP32[$180+8>>2]=0|0;HEAP32[$180+12>>2]=0|0; //@line 153 "./libraries/Box2D/Common/b2Math.h"
  $vA$sroa$0$0 = $53;$vA$sroa$6$0 = $55;$vB$sroa$0$0 = $61;$vB$sroa$6$0 = $63;$wA$0 = $57;$wB$0 = $65;
  $224 = HEAP32[$4>>2]|0; //@line 177 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $225 = HEAP32[$50>>2]|0; //@line 177 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $226 = (($225) + (($224*12)|0)|0); //@line 177 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  HEAP32[$226>>2] = $vA$sroa$0$0; //@line 177 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $227 = (((($225) + (($224*12)|0)|0)) + 4|0); //@line 177 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  HEAP32[$227>>2] = $vA$sroa$6$0; //@line 177 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $228 = HEAP32[$4>>2]|0; //@line 178 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $229 = HEAP32[$50>>2]|0; //@line 178 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $230 = (((($229) + (($228*12)|0)|0)) + 8|0); //@line 178 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  HEAPF32[$230>>2] = $wA$0; //@line 178 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $231 = HEAP32[$9>>2]|0; //@line 179 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $232 = (($229) + (($231*12)|0)|0); //@line 179 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  HEAP32[$232>>2] = $vB$sroa$0$0; //@line 179 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $233 = (((($229) + (($231*12)|0)|0)) + 4|0); //@line 179 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  HEAP32[$233>>2] = $vB$sroa$6$0; //@line 179 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $234 = HEAP32[$9>>2]|0; //@line 180 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $235 = HEAP32[$50>>2]|0; //@line 180 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $236 = (((($235) + (($234*12)|0)|0)) + 8|0); //@line 180 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  HEAPF32[$236>>2] = $wB$0; //@line 180 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  return; //@line 181 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 } else {
  $181 = ((($data)) + 8|0); //@line 160 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $182 = +HEAPF32[$181>>2]; //@line 160 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $183 = +HEAPF32[$180>>2]; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $184 = $182 * $183; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$180>>2] = $184; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $185 = ((($this)) + 88|0); //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $186 = +HEAPF32[$185>>2]; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $187 = $182 * $186; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$185>>2] = $187; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $188 = ((($this)) + 92|0); //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $189 = +HEAPF32[$188>>2]; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $190 = $182 * $189; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$188>>2] = $190; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $191 = ((($this)) + 96|0); //@line 161 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $192 = +HEAPF32[$191>>2]; //@line 161 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $193 = $182 * $192; //@line 161 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  HEAPF32[$191>>2] = $193; //@line 161 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $194 = $100 * $184; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $195 = $100 * $187; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $196 = (HEAP32[tempDoublePtr>>2]=$53,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $197 = $196 - $194; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $198 = (HEAPF32[tempDoublePtr>>2]=$197,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $199 = (HEAP32[tempDoublePtr>>2]=$55,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $200 = $199 - $195; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $201 = (HEAPF32[tempDoublePtr>>2]=$200,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $202 = $187 * $79; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $203 = $184 * $82; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $204 = $202 - $203; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $205 = $193 + $204; //@line 166 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $206 = $190 + $205; //@line 166 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $207 = $102 * $206; //@line 166 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $208 = $57 - $207; //@line 166 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $209 = $101 * $184; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $210 = $101 * $187; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $211 = (HEAP32[tempDoublePtr>>2]=$61,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $212 = $211 + $209; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $213 = (HEAPF32[tempDoublePtr>>2]=$212,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $214 = (HEAP32[tempDoublePtr>>2]=$63,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $215 = $214 + $210; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $216 = (HEAPF32[tempDoublePtr>>2]=$215,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $217 = $187 * $94; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $218 = $184 * $97; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $219 = $217 - $218; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $220 = $193 + $219; //@line 169 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $221 = $190 + $220; //@line 169 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $222 = $103 * $221; //@line 169 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $223 = $65 + $222; //@line 169 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $vA$sroa$0$0 = $198;$vA$sroa$6$0 = $201;$vB$sroa$0$0 = $213;$vB$sroa$6$0 = $216;$wA$0 = $208;$wB$0 = $223;
  $224 = HEAP32[$4>>2]|0; //@line 177 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $225 = HEAP32[$50>>2]|0; //@line 177 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $226 = (($225) + (($224*12)|0)|0); //@line 177 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  HEAP32[$226>>2] = $vA$sroa$0$0; //@line 177 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $227 = (((($225) + (($224*12)|0)|0)) + 4|0); //@line 177 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  HEAP32[$227>>2] = $vA$sroa$6$0; //@line 177 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $228 = HEAP32[$4>>2]|0; //@line 178 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $229 = HEAP32[$50>>2]|0; //@line 178 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $230 = (((($229) + (($228*12)|0)|0)) + 8|0); //@line 178 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  HEAPF32[$230>>2] = $wA$0; //@line 178 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $231 = HEAP32[$9>>2]|0; //@line 179 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $232 = (($229) + (($231*12)|0)|0); //@line 179 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  HEAP32[$232>>2] = $vB$sroa$0$0; //@line 179 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $233 = (((($229) + (($231*12)|0)|0)) + 4|0); //@line 179 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  HEAP32[$233>>2] = $vB$sroa$6$0; //@line 179 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $234 = HEAP32[$9>>2]|0; //@line 180 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $235 = HEAP32[$50>>2]|0; //@line 180 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $236 = (((($235) + (($234*12)|0)|0)) + 8|0); //@line 180 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  HEAPF32[$236>>2] = $wB$0; //@line 180 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  return; //@line 181 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 }
}
function __ZN15b2RevoluteJoint24SolveVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$cast = 0.0, $$cast40 = 0.0, $$pn = 0.0, $$pn2 = 0.0, $$pn3 = 0.0, $$pn4 = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0.0, $105 = 0.0, $106 = 0, $107 = 0.0, $108 = 0.0, $109 = 0, $11 = 0.0;
 var $110 = 0.0, $111 = 0.0, $112 = 0, $113 = 0, $114 = 0.0, $115 = 0.0, $116 = 0, $117 = 0, $118 = 0.0, $119 = 0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0.0, $132 = 0.0, $133 = 0.0, $134 = 0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0, $143 = 0, $144 = 0.0, $145 = 0.0, $146 = 0;
 var $147 = 0, $148 = 0.0, $149 = 0, $15 = 0, $150 = 0.0, $151 = 0.0, $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0;
 var $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0, $17 = 0, $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0.0, $182 = 0.0;
 var $183 = 0.0, $184 = 0.0, $185 = 0.0, $186 = 0.0, $187 = 0.0, $188 = 0.0, $189 = 0.0, $19 = 0.0, $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0.0, $194 = 0, $195 = 0.0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0;
 var $200 = 0.0, $201 = 0.0, $202 = 0.0, $203 = 0.0, $204 = 0.0, $205 = 0.0, $206 = 0.0, $207 = 0.0, $208 = 0, $209 = 0.0, $21 = 0.0, $210 = 0.0, $211 = 0, $212 = 0.0, $213 = 0.0, $214 = 0.0, $215 = 0.0, $216 = 0, $217 = 0.0, $218 = 0.0;
 var $219 = 0, $22 = 0, $220 = 0.0, $221 = 0, $222 = 0.0, $223 = 0.0, $224 = 0, $225 = 0.0, $226 = 0, $227 = 0.0, $228 = 0.0, $229 = 0.0, $23 = 0.0, $230 = 0.0, $231 = 0.0, $232 = 0.0, $233 = 0.0, $234 = 0.0, $235 = 0.0, $236 = 0.0;
 var $237 = 0.0, $238 = 0.0, $239 = 0.0, $24 = 0, $240 = 0.0, $241 = 0.0, $242 = 0.0, $243 = 0.0, $244 = 0.0, $245 = 0.0, $246 = 0.0, $247 = 0, $248 = 0, $249 = 0, $25 = 0.0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0;
 var $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0;
 var $4 = 0, $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0.0, $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0;
 var $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0.0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0;
 var $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0.0, $93 = 0.0;
 var $94 = 0, $95 = 0.0, $96 = 0.0, $97 = 0, $98 = 0.0, $99 = 0.0, $Cdot3 = 0, $brmerge = 0, $brmerge5 = 0, $impulse4$sroa$0$0 = 0, $impulse4$sroa$16$0 = 0.0, $impulse4$sroa$9$0 = 0, $impulse9 = 0, $reduced = 0, $reduced7 = 0, $rhs = 0, $rhs6 = 0, $vA$sroa$0$0$in = 0.0, $vA$sroa$10$0$in = 0.0, $vB$sroa$0$0$in = 0.0;
 var $vB$sroa$10$0$in = 0.0, $wA$0 = 0.0, $wA$1 = 0.0, $wB$0 = 0.0, $wB$1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $Cdot3 = sp + 60|0;
 $0 = sp + 48|0;
 $rhs = sp + 40|0;
 $reduced = sp + 32|0;
 $rhs6 = sp + 24|0;
 $reduced7 = sp + 16|0;
 $impulse9 = sp + 8|0;
 $1 = sp;
 $2 = ((($this)) + 128|0); //@line 185 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 185 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $4 = ((($data)) + 28|0); //@line 185 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 185 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $6 = (($5) + (($3*12)|0)|0); //@line 185 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 185 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $8 = (((($5) + (($3*12)|0)|0)) + 4|0); //@line 185 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 185 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $10 = (((($5) + (($3*12)|0)|0)) + 8|0); //@line 186 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $11 = +HEAPF32[$10>>2]; //@line 186 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $12 = ((($this)) + 132|0); //@line 187 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 187 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $14 = (($5) + (($13*12)|0)|0); //@line 187 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 187 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $16 = (((($5) + (($13*12)|0)|0)) + 4|0); //@line 187 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 187 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $18 = (((($5) + (($13*12)|0)|0)) + 8|0); //@line 188 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $19 = +HEAPF32[$18>>2]; //@line 188 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $20 = ((($this)) + 168|0); //@line 190 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 190 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $22 = ((($this)) + 172|0); //@line 190 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $23 = +HEAPF32[$22>>2]; //@line 190 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $24 = ((($this)) + 176|0); //@line 191 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $25 = +HEAPF32[$24>>2]; //@line 191 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $26 = ((($this)) + 180|0); //@line 191 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $27 = +HEAPF32[$26>>2]; //@line 191 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $28 = $25 + $27; //@line 193 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $29 = $28 == 0.0; //@line 193 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $30 = ((($this)) + 100|0); //@line 196 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $31 = HEAP8[$30>>0]|0; //@line 196 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $32 = ($31<<24>>24)==(0); //@line 196 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 if ($32) {
  $wA$0 = $11;$wB$0 = $19;
 } else {
  $33 = ((($this)) + 224|0); //@line 196 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $34 = HEAP32[$33>>2]|0; //@line 196 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $35 = ($34|0)==(3); //@line 196 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $brmerge = $29 | $35; //@line 196 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  if ($brmerge) {
   $wA$0 = $11;$wB$0 = $19;
  } else {
   $36 = $19 - $11; //@line 198 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $37 = ((($this)) + 108|0); //@line 198 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $38 = +HEAPF32[$37>>2]; //@line 198 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $39 = $36 - $38; //@line 198 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $40 = ((($this)) + 220|0); //@line 199 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $41 = +HEAPF32[$40>>2]; //@line 199 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $42 = $41 * $39; //@line 199 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $43 = ((($this)) + 96|0); //@line 200 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $44 = +HEAPF32[$43>>2]; //@line 200 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $45 = +HEAPF32[$data>>2]; //@line 201 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $46 = ((($this)) + 104|0); //@line 201 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $47 = +HEAPF32[$46>>2]; //@line 201 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $48 = $45 * $47; //@line 201 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $49 = $44 - $42; //@line 202 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $50 = -$48; //@line 202 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $51 = $49 < $48; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $52 = $51 ? $49 : $48; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $53 = $52 < $50; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $54 = $53 ? $50 : $52; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$43>>2] = $54; //@line 202 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $55 = $54 - $44; //@line 203 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $56 = $25 * $55; //@line 205 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $57 = $11 - $56; //@line 205 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $58 = $27 * $55; //@line 206 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $59 = $19 + $58; //@line 206 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $wA$0 = $57;$wB$0 = $59;
  }
 }
 $60 = ((($this)) + 112|0); //@line 210 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $61 = HEAP8[$60>>0]|0; //@line 210 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $62 = ($61<<24>>24)==(0); //@line 210 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 if (!($62)) {
  $63 = ((($this)) + 224|0); //@line 210 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $64 = HEAP32[$63>>2]|0; //@line 210 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $65 = ($64|0)==(0); //@line 210 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $brmerge5 = $29 | $65; //@line 210 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  if (!($brmerge5)) {
   $66 = ((($this)) + 148|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
   $67 = +HEAPF32[$66>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
   $68 = $wB$0 * $67; //@line 428 "./libraries/Box2D/Common/b2Math.h"
   $69 = ((($this)) + 144|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
   $70 = +HEAPF32[$69>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
   $71 = $wB$0 * $70; //@line 428 "./libraries/Box2D/Common/b2Math.h"
   $72 = (HEAP32[tempDoublePtr>>2]=$15,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
   $73 = $72 - $68; //@line 448 "./libraries/Box2D/Common/b2Math.h"
   $74 = (HEAP32[tempDoublePtr>>2]=$17,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
   $75 = $74 + $71; //@line 448 "./libraries/Box2D/Common/b2Math.h"
   $76 = (HEAP32[tempDoublePtr>>2]=$7,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $77 = $73 - $76; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $78 = (HEAP32[tempDoublePtr>>2]=$9,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $79 = $75 - $78; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $80 = ((($this)) + 140|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
   $81 = +HEAPF32[$80>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
   $82 = $wA$0 * $81; //@line 428 "./libraries/Box2D/Common/b2Math.h"
   $83 = ((($this)) + 136|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
   $84 = +HEAPF32[$83>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
   $85 = $wA$0 * $84; //@line 428 "./libraries/Box2D/Common/b2Math.h"
   $86 = $77 + $82; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $87 = $79 - $85; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $88 = $wB$0 - $wA$0; //@line 213 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   HEAPF32[$Cdot3>>2] = $86; //@line 150 "./libraries/Box2D/Common/b2Math.h"
   $89 = ((($Cdot3)) + 4|0); //@line 150 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$89>>2] = $87; //@line 150 "./libraries/Box2D/Common/b2Math.h"
   $90 = ((($Cdot3)) + 8|0); //@line 150 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$90>>2] = $88; //@line 150 "./libraries/Box2D/Common/b2Math.h"
   $91 = ((($this)) + 184|0); //@line 216 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   __ZNK7b2Mat337Solve33ERK6b2Vec3($0,$91,$Cdot3); //@line 216 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $92 = +HEAPF32[$0>>2]; //@line 159 "./libraries/Box2D/Common/b2Math.h"
   $93 = -$92; //@line 159 "./libraries/Box2D/Common/b2Math.h"
   $94 = ((($0)) + 4|0); //@line 159 "./libraries/Box2D/Common/b2Math.h"
   $95 = +HEAPF32[$94>>2]; //@line 159 "./libraries/Box2D/Common/b2Math.h"
   $96 = -$95; //@line 159 "./libraries/Box2D/Common/b2Math.h"
   $97 = ((($0)) + 8|0); //@line 159 "./libraries/Box2D/Common/b2Math.h"
   $98 = +HEAPF32[$97>>2]; //@line 159 "./libraries/Box2D/Common/b2Math.h"
   $99 = -$98; //@line 159 "./libraries/Box2D/Common/b2Math.h"
   $100 = (HEAPF32[tempDoublePtr>>2]=$93,HEAP32[tempDoublePtr>>2]|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
   $101 = (HEAPF32[tempDoublePtr>>2]=$96,HEAP32[tempDoublePtr>>2]|0); //@line 156 "./libraries/Box2D/Common/b2Math.h"
   $102 = HEAP32[$63>>2]|0; //@line 218 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   L8: do {
    switch ($102|0) {
    case 3:  {
     $103 = ((($this)) + 84|0); //@line 164 "./libraries/Box2D/Common/b2Math.h"
     $104 = +HEAPF32[$103>>2]; //@line 164 "./libraries/Box2D/Common/b2Math.h"
     $105 = $104 - $92; //@line 164 "./libraries/Box2D/Common/b2Math.h"
     HEAPF32[$103>>2] = $105; //@line 164 "./libraries/Box2D/Common/b2Math.h"
     $106 = ((($this)) + 88|0); //@line 164 "./libraries/Box2D/Common/b2Math.h"
     $107 = +HEAPF32[$106>>2]; //@line 164 "./libraries/Box2D/Common/b2Math.h"
     $108 = $107 - $95; //@line 164 "./libraries/Box2D/Common/b2Math.h"
     HEAPF32[$106>>2] = $108; //@line 164 "./libraries/Box2D/Common/b2Math.h"
     $109 = ((($this)) + 92|0); //@line 164 "./libraries/Box2D/Common/b2Math.h"
     $110 = +HEAPF32[$109>>2]; //@line 164 "./libraries/Box2D/Common/b2Math.h"
     $111 = $110 - $98; //@line 164 "./libraries/Box2D/Common/b2Math.h"
     HEAPF32[$109>>2] = $111; //@line 164 "./libraries/Box2D/Common/b2Math.h"
     $impulse4$sroa$0$0 = $100;$impulse4$sroa$16$0 = $99;$impulse4$sroa$9$0 = $101;
     break;
    }
    case 1:  {
     $112 = ((($this)) + 84|0); //@line 224 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
     $113 = ((($this)) + 92|0); //@line 224 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
     $114 = +HEAPF32[$113>>2]; //@line 227 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
     $115 = $114 - $98; //@line 224 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
     $116 = $115 < 0.0; //@line 225 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
     if ($116) {
      $117 = ((($this)) + 208|0); //@line 227 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $118 = +HEAPF32[$117>>2]; //@line 227 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $119 = ((($this)) + 212|0); //@line 227 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $120 = +HEAPF32[$119>>2]; //@line 227 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $121 = $114 * $118; //@line 459 "./libraries/Box2D/Common/b2Math.h"
      $122 = $114 * $120; //@line 459 "./libraries/Box2D/Common/b2Math.h"
      $123 = $121 - $86; //@line 448 "./libraries/Box2D/Common/b2Math.h"
      $124 = $122 - $87; //@line 448 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$rhs>>2] = $123; //@line 59 "./libraries/Box2D/Common/b2Math.h"
      $125 = ((($rhs)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$125>>2] = $124; //@line 59 "./libraries/Box2D/Common/b2Math.h"
      __ZNK7b2Mat337Solve22ERK6b2Vec2($reduced,$91,$rhs); //@line 228 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $126 = HEAP32[$reduced>>2]|0; //@line 229 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $127 = ((($reduced)) + 4|0); //@line 230 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $128 = HEAP32[$127>>2]|0; //@line 230 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $129 = +HEAPF32[$113>>2]; //@line 231 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $130 = -$129; //@line 231 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $$cast40 = (HEAP32[tempDoublePtr>>2]=$126,+HEAPF32[tempDoublePtr>>2]); //@line 232 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $131 = +HEAPF32[$112>>2]; //@line 232 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $132 = $$cast40 + $131; //@line 232 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      HEAPF32[$112>>2] = $132; //@line 232 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $133 = (HEAP32[tempDoublePtr>>2]=$128,+HEAPF32[tempDoublePtr>>2]); //@line 233 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $134 = ((($this)) + 88|0); //@line 233 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $135 = +HEAPF32[$134>>2]; //@line 233 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $136 = $133 + $135; //@line 233 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      HEAPF32[$134>>2] = $136; //@line 233 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      HEAPF32[$113>>2] = 0.0; //@line 234 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $impulse4$sroa$0$0 = $126;$impulse4$sroa$16$0 = $130;$impulse4$sroa$9$0 = $128;
      break L8;
     } else {
      $137 = +HEAPF32[$112>>2]; //@line 164 "./libraries/Box2D/Common/b2Math.h"
      $138 = $137 - $92; //@line 164 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$112>>2] = $138; //@line 164 "./libraries/Box2D/Common/b2Math.h"
      $139 = ((($this)) + 88|0); //@line 164 "./libraries/Box2D/Common/b2Math.h"
      $140 = +HEAPF32[$139>>2]; //@line 164 "./libraries/Box2D/Common/b2Math.h"
      $141 = $140 - $95; //@line 164 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$139>>2] = $141; //@line 164 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$113>>2] = $115; //@line 164 "./libraries/Box2D/Common/b2Math.h"
      $impulse4$sroa$0$0 = $100;$impulse4$sroa$16$0 = $99;$impulse4$sroa$9$0 = $101;
      break L8;
     }
     break;
    }
    case 2:  {
     $142 = ((($this)) + 84|0); //@line 243 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
     $143 = ((($this)) + 92|0); //@line 243 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
     $144 = +HEAPF32[$143>>2]; //@line 246 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
     $145 = $144 - $98; //@line 243 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
     $146 = $145 > 0.0; //@line 244 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
     if ($146) {
      $147 = ((($this)) + 208|0); //@line 246 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $148 = +HEAPF32[$147>>2]; //@line 246 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $149 = ((($this)) + 212|0); //@line 246 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $150 = +HEAPF32[$149>>2]; //@line 246 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $151 = $144 * $148; //@line 459 "./libraries/Box2D/Common/b2Math.h"
      $152 = $144 * $150; //@line 459 "./libraries/Box2D/Common/b2Math.h"
      $153 = $151 - $86; //@line 448 "./libraries/Box2D/Common/b2Math.h"
      $154 = $152 - $87; //@line 448 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$rhs6>>2] = $153; //@line 59 "./libraries/Box2D/Common/b2Math.h"
      $155 = ((($rhs6)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$155>>2] = $154; //@line 59 "./libraries/Box2D/Common/b2Math.h"
      __ZNK7b2Mat337Solve22ERK6b2Vec2($reduced7,$91,$rhs6); //@line 247 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $156 = HEAP32[$reduced7>>2]|0; //@line 248 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $157 = ((($reduced7)) + 4|0); //@line 249 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $158 = HEAP32[$157>>2]|0; //@line 249 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $159 = +HEAPF32[$143>>2]; //@line 250 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $160 = -$159; //@line 250 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $$cast = (HEAP32[tempDoublePtr>>2]=$156,+HEAPF32[tempDoublePtr>>2]); //@line 251 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $161 = +HEAPF32[$142>>2]; //@line 251 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $162 = $$cast + $161; //@line 251 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      HEAPF32[$142>>2] = $162; //@line 251 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $163 = (HEAP32[tempDoublePtr>>2]=$158,+HEAPF32[tempDoublePtr>>2]); //@line 252 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $164 = ((($this)) + 88|0); //@line 252 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $165 = +HEAPF32[$164>>2]; //@line 252 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $166 = $163 + $165; //@line 252 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      HEAPF32[$164>>2] = $166; //@line 252 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      HEAPF32[$143>>2] = 0.0; //@line 253 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
      $impulse4$sroa$0$0 = $156;$impulse4$sroa$16$0 = $160;$impulse4$sroa$9$0 = $158;
      break L8;
     } else {
      $167 = +HEAPF32[$142>>2]; //@line 164 "./libraries/Box2D/Common/b2Math.h"
      $168 = $167 - $92; //@line 164 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$142>>2] = $168; //@line 164 "./libraries/Box2D/Common/b2Math.h"
      $169 = ((($this)) + 88|0); //@line 164 "./libraries/Box2D/Common/b2Math.h"
      $170 = +HEAPF32[$169>>2]; //@line 164 "./libraries/Box2D/Common/b2Math.h"
      $171 = $170 - $95; //@line 164 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$169>>2] = $171; //@line 164 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$143>>2] = $145; //@line 164 "./libraries/Box2D/Common/b2Math.h"
      $impulse4$sroa$0$0 = $100;$impulse4$sroa$16$0 = $99;$impulse4$sroa$9$0 = $101;
      break L8;
     }
     break;
    }
    default: {
     $impulse4$sroa$0$0 = $100;$impulse4$sroa$16$0 = $99;$impulse4$sroa$9$0 = $101;
    }
    }
   } while(0);
   $172 = (HEAP32[tempDoublePtr>>2]=$impulse4$sroa$0$0,+HEAPF32[tempDoublePtr>>2]); //@line 261 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $173 = (HEAP32[tempDoublePtr>>2]=$impulse4$sroa$9$0,+HEAPF32[tempDoublePtr>>2]); //@line 261 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $174 = $21 * $172; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $175 = $21 * $173; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $176 = $76 - $174; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $177 = $78 - $175; //@line 91 "./libraries/Box2D/Common/b2Math.h"
   $178 = +HEAPF32[$83>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
   $179 = $173 * $178; //@line 414 "./libraries/Box2D/Common/b2Math.h"
   $180 = +HEAPF32[$80>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
   $181 = $172 * $180; //@line 414 "./libraries/Box2D/Common/b2Math.h"
   $182 = $179 - $181; //@line 414 "./libraries/Box2D/Common/b2Math.h"
   $183 = $impulse4$sroa$16$0 + $182; //@line 264 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $184 = $23 * $172; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $185 = $23 * $173; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $186 = $72 + $184; //@line 85 "./libraries/Box2D/Common/b2Math.h"
   $187 = $74 + $185; //@line 85 "./libraries/Box2D/Common/b2Math.h"
   $188 = +HEAPF32[$69>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
   $189 = $173 * $188; //@line 414 "./libraries/Box2D/Common/b2Math.h"
   $190 = +HEAPF32[$66>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
   $191 = $172 * $190; //@line 414 "./libraries/Box2D/Common/b2Math.h"
   $192 = $189 - $191; //@line 414 "./libraries/Box2D/Common/b2Math.h"
   $193 = $impulse4$sroa$16$0 + $192; //@line 267 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $$pn3 = $193;$$pn4 = $183;$vA$sroa$0$0$in = $176;$vA$sroa$10$0$in = $177;$vB$sroa$0$0$in = $186;$vB$sroa$10$0$in = $187;
   $$pn2 = $25 * $$pn4; //@line 264 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $$pn = $27 * $$pn3; //@line 267 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $wA$1 = $wA$0 - $$pn2; //@line 264 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $wB$1 = $wB$0 + $$pn; //@line 267 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $247 = HEAP32[$2>>2]|0; //@line 285 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $248 = HEAP32[$4>>2]|0; //@line 285 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $249 = (($248) + (($247*12)|0)|0); //@line 285 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   HEAPF32[$249>>2] = $vA$sroa$0$0$in; //@line 285 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $250 = (((($248) + (($247*12)|0)|0)) + 4|0); //@line 285 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   HEAPF32[$250>>2] = $vA$sroa$10$0$in; //@line 285 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $251 = HEAP32[$2>>2]|0; //@line 286 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $252 = HEAP32[$4>>2]|0; //@line 286 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $253 = (((($252) + (($251*12)|0)|0)) + 8|0); //@line 286 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   HEAPF32[$253>>2] = $wA$1; //@line 286 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $254 = HEAP32[$12>>2]|0; //@line 287 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $255 = (($252) + (($254*12)|0)|0); //@line 287 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   HEAPF32[$255>>2] = $vB$sroa$0$0$in; //@line 287 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $256 = (((($252) + (($254*12)|0)|0)) + 4|0); //@line 287 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   HEAPF32[$256>>2] = $vB$sroa$10$0$in; //@line 287 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $257 = HEAP32[$12>>2]|0; //@line 288 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $258 = HEAP32[$4>>2]|0; //@line 288 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $259 = (((($258) + (($257*12)|0)|0)) + 8|0); //@line 288 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   HEAPF32[$259>>2] = $wB$1; //@line 288 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   STACKTOP = sp;return; //@line 289 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  }
 }
 $194 = ((($this)) + 148|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $195 = +HEAPF32[$194>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $196 = $wB$0 * $195; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $197 = ((($this)) + 144|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $198 = +HEAPF32[$197>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $199 = $wB$0 * $198; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $200 = (HEAP32[tempDoublePtr>>2]=$15,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $201 = $200 - $196; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $202 = (HEAP32[tempDoublePtr>>2]=$17,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $203 = $202 + $199; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $204 = (HEAP32[tempDoublePtr>>2]=$7,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $205 = $201 - $204; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $206 = (HEAP32[tempDoublePtr>>2]=$9,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $207 = $203 - $206; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $208 = ((($this)) + 140|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $209 = +HEAPF32[$208>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $210 = $wA$0 * $209; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $211 = ((($this)) + 136|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $212 = +HEAPF32[$211>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $213 = $wA$0 * $212; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $214 = $205 + $210; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $215 = $207 - $213; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $216 = ((($this)) + 184|0); //@line 273 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $217 = -$214; //@line 68 "./libraries/Box2D/Common/b2Math.h"
 $218 = -$215; //@line 68 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$1>>2] = $217; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $219 = ((($1)) + 4|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$219>>2] = $218; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 __ZNK7b2Mat337Solve22ERK6b2Vec2($impulse9,$216,$1); //@line 273 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $220 = +HEAPF32[$impulse9>>2]; //@line 275 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $221 = ((($this)) + 84|0); //@line 275 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $222 = +HEAPF32[$221>>2]; //@line 275 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $223 = $220 + $222; //@line 275 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$221>>2] = $223; //@line 275 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $224 = ((($impulse9)) + 4|0); //@line 276 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $225 = +HEAPF32[$224>>2]; //@line 276 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $226 = ((($this)) + 88|0); //@line 276 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $227 = +HEAPF32[$226>>2]; //@line 276 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $228 = $225 + $227; //@line 276 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$226>>2] = $228; //@line 276 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $229 = $21 * $220; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $230 = $21 * $225; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $231 = $204 - $229; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $232 = $206 - $230; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $233 = +HEAPF32[$211>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $234 = $225 * $233; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $235 = +HEAPF32[$208>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $236 = $220 * $235; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $237 = $234 - $236; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $238 = $23 * $220; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $239 = $23 * $225; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $240 = $200 + $238; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $241 = $202 + $239; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $242 = +HEAPF32[$197>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $243 = $225 * $242; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $244 = +HEAPF32[$194>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $245 = $220 * $244; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $246 = $243 - $245; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $$pn3 = $246;$$pn4 = $237;$vA$sroa$0$0$in = $231;$vA$sroa$10$0$in = $232;$vB$sroa$0$0$in = $240;$vB$sroa$10$0$in = $241;
 $$pn2 = $25 * $$pn4; //@line 264 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $$pn = $27 * $$pn3; //@line 267 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $wA$1 = $wA$0 - $$pn2; //@line 264 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $wB$1 = $wB$0 + $$pn; //@line 267 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $247 = HEAP32[$2>>2]|0; //@line 285 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $248 = HEAP32[$4>>2]|0; //@line 285 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $249 = (($248) + (($247*12)|0)|0); //@line 285 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$249>>2] = $vA$sroa$0$0$in; //@line 285 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $250 = (((($248) + (($247*12)|0)|0)) + 4|0); //@line 285 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$250>>2] = $vA$sroa$10$0$in; //@line 285 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $251 = HEAP32[$2>>2]|0; //@line 286 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $252 = HEAP32[$4>>2]|0; //@line 286 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $253 = (((($252) + (($251*12)|0)|0)) + 8|0); //@line 286 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$253>>2] = $wA$1; //@line 286 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $254 = HEAP32[$12>>2]|0; //@line 287 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $255 = (($252) + (($254*12)|0)|0); //@line 287 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$255>>2] = $vB$sroa$0$0$in; //@line 287 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $256 = (((($252) + (($254*12)|0)|0)) + 4|0); //@line 287 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$256>>2] = $vB$sroa$10$0$in; //@line 287 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $257 = HEAP32[$12>>2]|0; //@line 288 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $258 = HEAP32[$4>>2]|0; //@line 288 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $259 = (((($258) + (($257*12)|0)|0)) + 8|0); //@line 288 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$259>>2] = $wB$1; //@line 288 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 STACKTOP = sp;return; //@line 289 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZN15b2RevoluteJoint24SolvePositionConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$pre = 0.0, $$pre14 = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0;
 var $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0.0, $123 = 0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0.0, $130 = 0.0, $131 = 0.0;
 var $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0, $146 = 0.0, $147 = 0.0, $148 = 0.0, $149 = 0.0, $15 = 0.0;
 var $150 = 0.0, $151 = 0.0, $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0;
 var $169 = 0.0, $17 = 0.0, $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0;
 var $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0.0, $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0.0, $50 = 0.0;
 var $51 = 0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0, $64 = 0.0, $65 = 0, $66 = 0.0, $67 = 0, $68 = 0.0, $69 = 0.0;
 var $7 = 0.0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0.0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0, $87 = 0.0;
 var $88 = 0.0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0, $98 = 0.0, $99 = 0.0, $aA$0 = 0.0, $aB$0 = 0.0, $angularError$0 = 0.0, $angularError$1 = 0.0, $brmerge = 0, $det$0$i = 0.0, $limitImpulse$0 = 0.0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 128|0); //@line 293 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 293 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = ((($data)) + 24|0); //@line 293 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 293 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $4 = (($3) + (($1*12)|0)|0); //@line 293 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 293 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $6 = (((($3) + (($1*12)|0)|0)) + 4|0); //@line 293 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 293 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $8 = (((($3) + (($1*12)|0)|0)) + 8|0); //@line 294 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 294 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $10 = ((($this)) + 132|0); //@line 295 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 295 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $12 = (($3) + (($11*12)|0)|0); //@line 295 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $13 = +HEAPF32[$12>>2]; //@line 295 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $14 = (((($3) + (($11*12)|0)|0)) + 4|0); //@line 295 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 295 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $16 = (((($3) + (($11*12)|0)|0)) + 8|0); //@line 296 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 296 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $18 = ((($this)) + 176|0); //@line 303 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $19 = ((($this)) + 180|0); //@line 303 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $20 = ((($this)) + 112|0); //@line 306 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $21 = HEAP8[$20>>0]|0; //@line 306 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $22 = ($21<<24>>24)==(0); //@line 306 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $$pre = +HEAPF32[$18>>2]; //@line 352 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $$pre14 = +HEAPF32[$19>>2]; //@line 352 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 if ($22) {
  $aA$0 = $9;$aB$0 = $17;$angularError$1 = 0.0;
 } else {
  $23 = $$pre + $$pre14; //@line 303 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $24 = $23 == 0.0; //@line 303 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $25 = ((($this)) + 224|0); //@line 306 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $26 = HEAP32[$25>>2]|0; //@line 306 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $27 = ($26|0)==(0); //@line 306 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $brmerge = $27 | $24; //@line 306 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  if ($brmerge) {
   $aA$0 = $9;$aB$0 = $17;$angularError$1 = 0.0;
  } else {
   $28 = $17 - $9; //@line 308 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $29 = ((($this)) + 116|0); //@line 308 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $30 = +HEAPF32[$29>>2]; //@line 308 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $31 = $28 - $30; //@line 308 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   switch ($26|0) {
   case 3:  {
    $32 = ((($this)) + 120|0); //@line 314 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $33 = +HEAPF32[$32>>2]; //@line 314 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $34 = $31 - $33; //@line 314 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $35 = $34 < 0.13962635397911072; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $36 = $35 ? $34 : 0.13962635397911072; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $37 = $36 < -0.13962635397911072; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $38 = $37 ? -0.13962635397911072 : $36; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $39 = ((($this)) + 220|0); //@line 315 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $40 = +HEAPF32[$39>>2]; //@line 315 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $41 = $40 * $38; //@line 315 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $42 = -$41; //@line 315 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $43 = $38 > 0.0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
    $44 = -$38; //@line 618 "./libraries/Box2D/Common/b2Math.h"
    $45 = $43 ? $38 : $44; //@line 618 "./libraries/Box2D/Common/b2Math.h"
    $angularError$0 = $45;$limitImpulse$0 = $42;
    break;
   }
   case 1:  {
    $46 = ((($this)) + 120|0); //@line 320 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $47 = +HEAPF32[$46>>2]; //@line 320 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $48 = $31 - $47; //@line 320 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $49 = -$48; //@line 321 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $50 = $48 + 0.034906588494777679; //@line 324 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $51 = $50 < 0.0; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $52 = $51 ? $50 : 0.0; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $53 = $52 < -0.13962635397911072; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $54 = $53 ? -0.13962635397911072 : $52; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $55 = ((($this)) + 220|0); //@line 325 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $56 = +HEAPF32[$55>>2]; //@line 325 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $57 = $56 * $54; //@line 325 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $58 = -$57; //@line 325 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $angularError$0 = $49;$limitImpulse$0 = $58;
    break;
   }
   case 2:  {
    $59 = ((($this)) + 124|0); //@line 329 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $60 = +HEAPF32[$59>>2]; //@line 329 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $61 = $31 - $60; //@line 329 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $62 = $61 + -0.034906588494777679; //@line 333 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $63 = $62 < 0.13962635397911072; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $64 = $63 ? $62 : 0.13962635397911072; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $65 = $64 < 0.0; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $66 = $65 ? 0.0 : $64; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $67 = ((($this)) + 220|0); //@line 334 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $68 = +HEAPF32[$67>>2]; //@line 334 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $69 = $68 * $66; //@line 334 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $70 = -$69; //@line 334 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
    $angularError$0 = $61;$limitImpulse$0 = $70;
    break;
   }
   default: {
    $angularError$0 = 0.0;$limitImpulse$0 = 0.0;
   }
   }
   $71 = $limitImpulse$0 * $$pre; //@line 337 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $72 = $9 - $71; //@line 337 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $73 = $limitImpulse$0 * $$pre14; //@line 338 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $74 = $17 + $73; //@line 338 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
   $aA$0 = $72;$aB$0 = $74;$angularError$1 = $angularError$0;
  }
 }
 $75 = (+Math_sin((+$aA$0))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $76 = (+Math_cos((+$aA$0))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $77 = (+Math_sin((+$aB$0))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $78 = (+Math_cos((+$aB$0))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $79 = ((($this)) + 68|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $80 = +HEAPF32[$79>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $81 = ((($this)) + 152|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $82 = +HEAPF32[$81>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $83 = $80 - $82; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $84 = ((($this)) + 72|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $85 = +HEAPF32[$84>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $86 = ((($this)) + 156|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $87 = +HEAPF32[$86>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $88 = $85 - $87; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $89 = $76 * $83; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $90 = $75 * $88; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $91 = $89 - $90; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $92 = $75 * $83; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $93 = $76 * $88; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $94 = $92 + $93; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $95 = ((($this)) + 76|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $96 = +HEAPF32[$95>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $97 = ((($this)) + 160|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $98 = +HEAPF32[$97>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $99 = $96 - $98; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $100 = ((($this)) + 80|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $101 = +HEAPF32[$100>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $102 = ((($this)) + 164|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $103 = +HEAPF32[$102>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $104 = $101 - $103; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $105 = $78 * $99; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $106 = $77 * $104; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $107 = $105 - $106; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $108 = $77 * $99; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $109 = $78 * $104; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $110 = $108 + $109; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $111 = $13 + $107; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $112 = $15 + $110; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $113 = $111 - $5; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $114 = $112 - $7; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $115 = $113 - $91; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $116 = $114 - $94; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $117 = $115 * $115; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $118 = $116 * $116; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $119 = $117 + $118; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $120 = (+Math_sqrt((+$119))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $121 = ((($this)) + 168|0); //@line 351 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $122 = +HEAPF32[$121>>2]; //@line 351 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $123 = ((($this)) + 172|0); //@line 351 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $124 = +HEAPF32[$123>>2]; //@line 351 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $125 = $122 + $124; //@line 355 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $126 = $94 * $$pre; //@line 355 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $127 = $94 * $126; //@line 355 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $128 = $125 + $127; //@line 355 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $129 = $110 * $$pre14; //@line 355 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $130 = $110 * $129; //@line 355 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $131 = $128 + $130; //@line 355 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $132 = $91 * $$pre; //@line 356 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $133 = $94 * $132; //@line 356 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $134 = -$133; //@line 356 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $135 = $107 * $$pre14; //@line 356 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $136 = $110 * $135; //@line 356 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $137 = $134 - $136; //@line 356 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $138 = $91 * $132; //@line 358 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $139 = $125 + $138; //@line 358 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $140 = $107 * $135; //@line 358 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $141 = $139 + $140; //@line 358 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $142 = $131 * $141; //@line 242 "./libraries/Box2D/Common/b2Math.h"
 $143 = $137 * $137; //@line 242 "./libraries/Box2D/Common/b2Math.h"
 $144 = $142 - $143; //@line 242 "./libraries/Box2D/Common/b2Math.h"
 $145 = $144 != 0.0; //@line 243 "./libraries/Box2D/Common/b2Math.h"
 $146 = 1.0 / $144; //@line 245 "./libraries/Box2D/Common/b2Math.h"
 $det$0$i = $145 ? $146 : $144; //@line 243 "./libraries/Box2D/Common/b2Math.h"
 $147 = $115 * $141; //@line 248 "./libraries/Box2D/Common/b2Math.h"
 $148 = $116 * $137; //@line 248 "./libraries/Box2D/Common/b2Math.h"
 $149 = $147 - $148; //@line 248 "./libraries/Box2D/Common/b2Math.h"
 $150 = $149 * $det$0$i; //@line 248 "./libraries/Box2D/Common/b2Math.h"
 $151 = $116 * $131; //@line 249 "./libraries/Box2D/Common/b2Math.h"
 $152 = $115 * $137; //@line 249 "./libraries/Box2D/Common/b2Math.h"
 $153 = $151 - $152; //@line 249 "./libraries/Box2D/Common/b2Math.h"
 $154 = $153 * $det$0$i; //@line 249 "./libraries/Box2D/Common/b2Math.h"
 $155 = -$150; //@line 68 "./libraries/Box2D/Common/b2Math.h"
 $156 = -$154; //@line 68 "./libraries/Box2D/Common/b2Math.h"
 $157 = $122 * $155; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $158 = $122 * $156; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $159 = $5 - $157; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $160 = $7 - $158; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $161 = $91 * $156; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $162 = $94 * $155; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $163 = $161 - $162; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $164 = $$pre * $163; //@line 363 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $165 = $aA$0 - $164; //@line 363 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $166 = $124 * $155; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $167 = $124 * $156; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $168 = $13 + $166; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $169 = $15 + $167; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $170 = $107 * $156; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $171 = $110 * $155; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $172 = $170 - $171; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $173 = $$pre14 * $172; //@line 366 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $174 = $aB$0 + $173; //@line 366 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $175 = HEAP32[$0>>2]|0; //@line 369 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $176 = HEAP32[$2>>2]|0; //@line 369 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $177 = (($176) + (($175*12)|0)|0); //@line 369 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$177>>2] = $159; //@line 369 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $178 = (((($176) + (($175*12)|0)|0)) + 4|0); //@line 369 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$178>>2] = $160; //@line 369 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $179 = HEAP32[$0>>2]|0; //@line 370 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $180 = HEAP32[$2>>2]|0; //@line 370 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $181 = (((($180) + (($179*12)|0)|0)) + 8|0); //@line 370 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$181>>2] = $165; //@line 370 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $182 = HEAP32[$10>>2]|0; //@line 371 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $183 = (($180) + (($182*12)|0)|0); //@line 371 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$183>>2] = $168; //@line 371 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $184 = (((($180) + (($182*12)|0)|0)) + 4|0); //@line 371 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$184>>2] = $169; //@line 371 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $185 = HEAP32[$10>>2]|0; //@line 372 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $186 = HEAP32[$2>>2]|0; //@line 372 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $187 = (((($186) + (($185*12)|0)|0)) + 8|0); //@line 372 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$187>>2] = $174; //@line 372 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $188 = $120 <= 0.004999999888241291; //@line 374 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $189 = $angularError$1 <= 0.034906588494777679; //@line 374 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $190 = $189 & $188; //@line 374 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 return ($190|0); //@line 374 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZN18b2RevoluteJointDef10InitializeEP6b2BodyS1_RK6b2Vec2($this,$bA,$bB,$anchor) {
 $this = $this|0;
 $bA = $bA|0;
 $bB = $bB|0;
 $anchor = $anchor|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0;
 var $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0.0, $40 = 0.0, $41 = 0, $42 = 0, $43 = 0, $44 = 0.0;
 var $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 38 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$0>>2] = $bA; //@line 38 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = ((($this)) + 12|0); //@line 39 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$1>>2] = $bB; //@line 39 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = +HEAPF32[$anchor>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($bA)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$3>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $5 = $2 - $4; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($anchor)) + 4|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($bA)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $9 = +HEAPF32[$8>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $10 = $7 - $9; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($bA)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$11>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $13 = $5 * $12; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $14 = ((($bA)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $15 = +HEAPF32[$14>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $16 = $10 * $15; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $17 = $13 + $16; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $18 = $5 * $15; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $19 = $12 * $10; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $20 = $19 - $18; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($this)) + 20|0); //@line 40 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$21>>2] = $17; //@line 40 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $22 = ((($this)) + 24|0); //@line 40 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$22>>2] = $20; //@line 40 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $23 = +HEAPF32[$anchor>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($bB)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $26 = $23 - $25; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $27 = +HEAPF32[$6>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $28 = ((($bB)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $29 = +HEAPF32[$28>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $30 = $27 - $29; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $31 = ((($bB)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $32 = +HEAPF32[$31>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $33 = $26 * $32; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $34 = ((($bB)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $35 = +HEAPF32[$34>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $36 = $30 * $35; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $37 = $33 + $36; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $38 = $26 * $35; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $39 = $32 * $30; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $40 = $39 - $38; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $41 = ((($this)) + 28|0); //@line 41 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$41>>2] = $37; //@line 41 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $42 = ((($this)) + 32|0); //@line 41 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$42>>2] = $40; //@line 41 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $43 = ((($bB)) + 56|0); //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $44 = +HEAPF32[$43>>2]; //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $45 = ((($bA)) + 56|0); //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $46 = +HEAPF32[$45>>2]; //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $47 = $44 - $46; //@line 42 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $48 = ((($this)) + 36|0); //@line 42 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$48>>2] = $47; //@line 42 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 return; //@line 43 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZN15b2RevoluteJointC2EPK18b2RevoluteJointDef($this,$def) {
 $this = $this|0;
 $def = $def|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN7b2JointC2EPK10b2JointDef($this,$def); //@line 46 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$this>>2] = (20972); //@line 47 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $0 = ((($this)) + 68|0); //@line 45 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = ((($this)) + 76|0); //@line 45 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = ((($def)) + 20|0); //@line 48 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $3 = $2; //@line 48 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $4 = $3; //@line 48 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 48 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $6 = (($3) + 4)|0; //@line 48 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $7 = $6; //@line 48 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 48 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $9 = $0; //@line 48 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $10 = $9; //@line 48 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$10>>2] = $5; //@line 48 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $11 = (($9) + 4)|0; //@line 48 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $12 = $11; //@line 48 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$12>>2] = $8; //@line 48 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $13 = ((($def)) + 28|0); //@line 49 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $14 = $13; //@line 49 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $15 = $14; //@line 49 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 49 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $17 = (($14) + 4)|0; //@line 49 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $18 = $17; //@line 49 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 49 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $20 = $1; //@line 49 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $21 = $20; //@line 49 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$21>>2] = $16; //@line 49 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $22 = (($20) + 4)|0; //@line 49 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $23 = $22; //@line 49 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$23>>2] = $19; //@line 49 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $24 = ((($def)) + 36|0); //@line 50 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $25 = HEAP32[$24>>2]|0; //@line 50 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $26 = ((($this)) + 116|0); //@line 50 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$26>>2] = $25; //@line 50 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $27 = ((($this)) + 84|0); //@line 153 "./libraries/Box2D/Common/b2Math.h"
 $28 = ((($def)) + 44|0); //@line 55 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 ;HEAP32[$27>>2]=0|0;HEAP32[$27+4>>2]=0|0;HEAP32[$27+8>>2]=0|0;HEAP32[$27+12>>2]=0|0; //@line 153 "./libraries/Box2D/Common/b2Math.h"
 $29 = HEAP32[$28>>2]|0; //@line 55 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $30 = ((($this)) + 120|0); //@line 55 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$30>>2] = $29; //@line 55 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $31 = ((($def)) + 48|0); //@line 56 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $32 = HEAP32[$31>>2]|0; //@line 56 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $33 = ((($this)) + 124|0); //@line 56 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$33>>2] = $32; //@line 56 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $34 = ((($def)) + 60|0); //@line 57 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 57 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $36 = ((($this)) + 104|0); //@line 57 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$36>>2] = $35; //@line 57 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $37 = ((($def)) + 56|0); //@line 58 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $38 = HEAP32[$37>>2]|0; //@line 58 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $39 = ((($this)) + 108|0); //@line 58 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$39>>2] = $38; //@line 58 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $40 = ((($def)) + 40|0); //@line 59 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $41 = HEAP8[$40>>0]|0; //@line 59 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $42 = ((($this)) + 112|0); //@line 59 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP8[$42>>0] = $41; //@line 59 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $43 = ((($def)) + 52|0); //@line 60 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $44 = HEAP8[$43>>0]|0; //@line 60 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $45 = ((($this)) + 100|0); //@line 60 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP8[$45>>0] = $44; //@line 60 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $46 = ((($this)) + 224|0); //@line 61 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP32[$46>>2] = 0; //@line 61 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 return; //@line 62 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZNK15b2RevoluteJoint13GetJointAngleEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 400 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 400 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = ((($this)) + 52|0); //@line 401 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 401 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $4 = ((($3)) + 56|0); //@line 402 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 402 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $6 = ((($1)) + 56|0); //@line 402 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 402 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $8 = $5 - $7; //@line 402 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $9 = ((($this)) + 116|0); //@line 402 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $10 = +HEAPF32[$9>>2]; //@line 402 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $11 = $8 - $10; //@line 402 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 return (+$11); //@line 402 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZNK15b2RevoluteJoint13GetJointSpeedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 407 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 407 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = ((($this)) + 52|0); //@line 408 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 408 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $4 = ((($3)) + 72|0); //@line 409 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 409 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $6 = ((($1)) + 72|0); //@line 409 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 409 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $8 = $5 - $7; //@line 409 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 return (+$8); //@line 409 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZNK15b2RevoluteJoint14IsMotorEnabledEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 100|0); //@line 414 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 414 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 414 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 return ($2|0); //@line 414 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZN15b2RevoluteJoint11EnableMotorEb($this,$flag) {
 $this = $this|0;
 $flag = $flag|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 419 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 419 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = ((($1)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = $4 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $6 = ($5|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($6) {
  $7 = $4 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $8 = $7&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$2>>1] = $8; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $9 = ((($1)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$9>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $10 = ((($this)) + 52|0); //@line 420 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 420 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $12 = ((($11)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $13 = HEAP16[$12>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $14 = $13&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $15 = $14 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $16 = ($15|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($16)) {
  $20 = ((($this)) + 100|0); //@line 421 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $21 = $flag&1; //@line 421 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  HEAP8[$20>>0] = $21; //@line 421 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  return; //@line 422 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 }
 $17 = $14 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $18 = $17&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP16[$12>>1] = $18; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = ((($11)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$19>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = ((($this)) + 100|0); //@line 421 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $21 = $flag&1; //@line 421 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP8[$20>>0] = $21; //@line 421 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 return; //@line 422 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZNK15b2RevoluteJoint14GetMotorTorqueEf($this,$inv_dt) {
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 96|0); //@line 426 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 426 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = $1 * $inv_dt; //@line 426 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 return (+$2); //@line 426 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZN15b2RevoluteJoint13SetMotorSpeedEf($this,$speed) {
 $this = $this|0;
 $speed = +$speed;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 431 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 431 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = ((($1)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = $4 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $6 = ($5|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($6) {
  $7 = $4 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $8 = $7&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$2>>1] = $8; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $9 = ((($1)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$9>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $10 = ((($this)) + 52|0); //@line 432 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 432 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $12 = ((($11)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $13 = HEAP16[$12>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $14 = $13&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $15 = $14 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $16 = ($15|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($16)) {
  $20 = ((($this)) + 108|0); //@line 433 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  HEAPF32[$20>>2] = $speed; //@line 433 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  return; //@line 434 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 }
 $17 = $14 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $18 = $17&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP16[$12>>1] = $18; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = ((($11)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$19>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = ((($this)) + 108|0); //@line 433 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$20>>2] = $speed; //@line 433 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 return; //@line 434 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZN15b2RevoluteJoint17SetMaxMotorTorqueEf($this,$torque) {
 $this = $this|0;
 $torque = +$torque;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 438 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 438 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = ((($1)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = $4 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $6 = ($5|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($6) {
  $7 = $4 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $8 = $7&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$2>>1] = $8; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $9 = ((($1)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$9>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $10 = ((($this)) + 52|0); //@line 439 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 439 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $12 = ((($11)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $13 = HEAP16[$12>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $14 = $13&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $15 = $14 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $16 = ($15|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($16)) {
  $20 = ((($this)) + 104|0); //@line 440 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  HEAPF32[$20>>2] = $torque; //@line 440 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  return; //@line 441 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 }
 $17 = $14 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $18 = $17&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP16[$12>>1] = $18; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = ((($11)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$19>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = ((($this)) + 104|0); //@line 440 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$20>>2] = $torque; //@line 440 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 return; //@line 441 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZNK15b2RevoluteJoint14IsLimitEnabledEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 445 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 445 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 445 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 return ($2|0); //@line 445 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZN15b2RevoluteJoint11EnableLimitEb($this,$flag) {
 $this = $this|0;
 $flag = $flag|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $flag&1; //@line 450 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = ((($this)) + 112|0); //@line 450 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = HEAP8[$1>>0]|0; //@line 450 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $3 = $2&255; //@line 450 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $4 = ($0|0)==($3|0); //@line 450 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 if ($4) {
  return; //@line 457 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 }
 $5 = ((($this)) + 48|0); //@line 452 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 452 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $7 = ((($6)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $8 = HEAP16[$7>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $9 = $8&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $10 = $9 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $11 = ($10|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($11) {
  $12 = $9 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $13 = $12&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$7>>1] = $13; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $14 = ((($6)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$14>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $15 = ((($this)) + 52|0); //@line 453 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 453 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $17 = ((($16)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $18 = HEAP16[$17>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = $18&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = $19 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $21 = ($20|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($21) {
  $22 = $19 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $23 = $22&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$17>>1] = $23; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $24 = ((($16)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$24>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $25 = $flag&1; //@line 454 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAP8[$1>>0] = $25; //@line 454 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $26 = ((($this)) + 92|0); //@line 455 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$26>>2] = 0.0; //@line 455 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 return; //@line 457 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZNK15b2RevoluteJoint13GetLowerLimitEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 120|0); //@line 461 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 461 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 return (+$1); //@line 461 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZNK15b2RevoluteJoint13GetUpperLimitEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 124|0); //@line 466 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 466 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 return (+$1); //@line 466 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZN15b2RevoluteJoint9SetLimitsEff($this,$lower,$upper) {
 $this = $this|0;
 $lower = +$lower;
 $upper = +$upper;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $lower <= $upper; //@line 471 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 __Z10loveAssertbPKc($0,363332); //@line 471 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $1 = ((($this)) + 120|0); //@line 473 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $2 = +HEAPF32[$1>>2]; //@line 473 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $3 = $2 != $lower; //@line 473 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 if (!($3)) {
  $4 = ((($this)) + 124|0); //@line 473 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $5 = +HEAPF32[$4>>2]; //@line 473 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  $6 = $5 != $upper; //@line 473 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  if (!($6)) {
   return; //@line 481 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
  }
 }
 $7 = ((($this)) + 48|0); //@line 475 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 475 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $9 = ((($8)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $10 = HEAP16[$9>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $11 = $10&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $12 = $11 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $13 = ($12|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($13) {
  $14 = $11 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $15 = $14&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$9>>1] = $15; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $16 = ((($8)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$16>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $17 = ((($this)) + 52|0); //@line 476 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 476 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $19 = ((($18)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = HEAP16[$19>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $21 = $20&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $22 = $21 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $23 = ($22|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($23) {
  $24 = $21 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $25 = $24&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$19>>1] = $25; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $26 = ((($18)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$26>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $27 = ((($this)) + 92|0); //@line 477 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$27>>2] = 0.0; //@line 477 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$1>>2] = $lower; //@line 478 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 $28 = ((($this)) + 124|0); //@line 479 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 HEAPF32[$28>>2] = $upper; //@line 479 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
 return; //@line 481 "libraries/Box2D/Dynamics/Joints/b2RevoluteJoint.cpp"
}
function __ZNK11b2RopeJoint10GetAnchorAEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 198 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 198 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 68|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 72|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 198 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
}
function __ZNK11b2RopeJoint10GetAnchorBEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 203 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 203 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 76|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 80|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 203 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
}
function __ZNK11b2RopeJoint16GetReactionForceEf($agg$result,$this,$inv_dt) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 92|0); //@line 208 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 208 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $2 = $1 * $inv_dt; //@line 208 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $3 = ((($this)) + 104|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$3>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $5 = $4 * $2; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($this)) + 108|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $8 = $2 * $7; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $5; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$9>>2] = $8; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 209 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
}
function __ZNK11b2RopeJoint17GetReactionTorqueEf($this,$inv_dt) {
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return +0; //@line 215 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
}
function __ZN11b2RopeJoint4DumpEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer13 = 0, $vararg_buffer17 = 0, $vararg_buffer20 = 0, $vararg_buffer3 = 0, $vararg_buffer6 = 0, $vararg_buffer9 = 0, $vararg_ptr12 = 0, $vararg_ptr16 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer20 = sp + 72|0;
 $vararg_buffer17 = sp + 64|0;
 $vararg_buffer13 = sp + 48|0;
 $vararg_buffer9 = sp + 32|0;
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($this)) + 48|0); //@line 230 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 230 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 230 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 230 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $4 = ((($this)) + 52|0); //@line 231 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 231 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $6 = ((($5)) + 8|0); //@line 231 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 231 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 __Z5b2LogPKcz(363361,$vararg_buffer); //@line 233 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$vararg_buffer1>>2] = $3; //@line 234 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 __Z5b2LogPKcz(447349,$vararg_buffer1); //@line 234 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$vararg_buffer3>>2] = $7; //@line 235 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 __Z5b2LogPKcz(447375,$vararg_buffer3); //@line 235 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $8 = ((($this)) + 61|0); //@line 236 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 236 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $10 = $9&255; //@line 236 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$vararg_buffer6>>2] = $10; //@line 236 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 __Z5b2LogPKcz(447401,$vararg_buffer6); //@line 236 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $11 = ((($this)) + 68|0); //@line 237 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $12 = +HEAPF32[$11>>2]; //@line 237 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $13 = $12; //@line 237 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $14 = ((($this)) + 72|0); //@line 237 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 237 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $16 = $15; //@line 237 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF64[$vararg_buffer9>>3] = $13; //@line 237 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $vararg_ptr12 = ((($vararg_buffer9)) + 8|0); //@line 237 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF64[$vararg_ptr12>>3] = $16; //@line 237 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 __Z5b2LogPKcz(447106,$vararg_buffer9); //@line 237 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $17 = ((($this)) + 76|0); //@line 238 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $18 = +HEAPF32[$17>>2]; //@line 238 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $19 = $18; //@line 238 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $20 = ((($this)) + 80|0); //@line 238 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 238 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $22 = $21; //@line 238 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF64[$vararg_buffer13>>3] = $19; //@line 238 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $vararg_ptr16 = ((($vararg_buffer13)) + 8|0); //@line 238 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF64[$vararg_ptr16>>3] = $22; //@line 238 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 __Z5b2LogPKcz(447148,$vararg_buffer13); //@line 238 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $23 = ((($this)) + 84|0); //@line 239 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $24 = +HEAPF32[$23>>2]; //@line 239 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $25 = $24; //@line 239 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF64[$vararg_buffer17>>3] = $25; //@line 239 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 __Z5b2LogPKcz(363383,$vararg_buffer17); //@line 239 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $26 = ((($this)) + 56|0); //@line 240 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $27 = HEAP32[$26>>2]|0; //@line 240 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$vararg_buffer20>>2] = $27; //@line 240 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 __Z5b2LogPKcz(447513,$vararg_buffer20); //@line 240 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 STACKTOP = sp;return; //@line 241 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
}
function __ZN11b2RopeJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 58 "./libraries/Box2D/Dynamics/Joints/b2RopeJoint.h"
 return; //@line 58 "./libraries/Box2D/Dynamics/Joints/b2RopeJoint.h"
}
function __ZN11b2RopeJoint23InitVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0, $107 = 0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0;
 var $115 = 0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0, $121 = 0, $122 = 0.0, $123 = 0.0, $124 = 0, $125 = 0, $126 = 0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0.0, $132 = 0.0;
 var $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0, $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0.0, $158 = 0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0.0, $167 = 0, $168 = 0.0, $169 = 0.0;
 var $17 = 0, $170 = 0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0, $181 = 0.0, $182 = 0.0, $183 = 0, $184 = 0.0, $185 = 0.0, $186 = 0.0, $187 = 0.0;
 var $188 = 0.0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0, $53 = 0.0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0;
 var $78 = 0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0, $92 = 0, $93 = 0, $94 = 0.0, $95 = 0.0;
 var $96 = 0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $bitcast = 0.0, $bitcast24 = 0.0, $vA$sroa$0$0 = 0, $vA$sroa$6$0 = 0, $vB$sroa$0$0 = 0, $vB$sroa$6$0 = 0, $wA$0 = 0.0, $wB$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 48 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 48 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 48 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 48 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $4 = ((($this)) + 96|0); //@line 48 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$4>>2] = $3; //@line 48 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $5 = ((($this)) + 52|0); //@line 49 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 49 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $7 = ((($6)) + 8|0); //@line 49 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 49 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $9 = ((($this)) + 100|0); //@line 49 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$9>>2] = $8; //@line 49 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $10 = ((($this)) + 128|0); //@line 50 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $11 = ((($1)) + 28|0); //@line 50 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $12 = $11; //@line 50 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $13 = $12; //@line 50 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 50 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $15 = (($12) + 4)|0; //@line 50 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $16 = $15; //@line 50 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 50 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $18 = $10; //@line 50 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $19 = $18; //@line 50 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$19>>2] = $14; //@line 50 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $20 = (($18) + 4)|0; //@line 50 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $21 = $20; //@line 50 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$21>>2] = $17; //@line 50 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $22 = ((($this)) + 136|0); //@line 51 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $23 = ((($6)) + 28|0); //@line 51 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $24 = $23; //@line 51 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $25 = $24; //@line 51 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 51 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $27 = (($24) + 4)|0; //@line 51 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $28 = $27; //@line 51 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 51 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $30 = $22; //@line 51 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $31 = $30; //@line 51 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$31>>2] = $26; //@line 51 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $32 = (($30) + 4)|0; //@line 51 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $33 = $32; //@line 51 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$33>>2] = $29; //@line 51 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $34 = ((($1)) + 120|0); //@line 52 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 52 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $36 = ((($this)) + 144|0); //@line 52 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$36>>2] = $35; //@line 52 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $37 = ((($6)) + 120|0); //@line 53 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $38 = HEAP32[$37>>2]|0; //@line 53 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $39 = ((($this)) + 148|0); //@line 53 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$39>>2] = $38; //@line 53 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $40 = ((($1)) + 128|0); //@line 54 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $41 = HEAP32[$40>>2]|0; //@line 54 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $42 = ((($this)) + 152|0); //@line 54 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$42>>2] = $41; //@line 54 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $43 = ((($6)) + 128|0); //@line 55 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $44 = HEAP32[$43>>2]|0; //@line 55 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $45 = ((($this)) + 156|0); //@line 55 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$45>>2] = $44; //@line 55 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $46 = ((($data)) + 24|0); //@line 57 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $47 = HEAP32[$46>>2]|0; //@line 57 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $48 = (($47) + (($3*12)|0)|0); //@line 57 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $49 = +HEAPF32[$48>>2]; //@line 57 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $50 = (((($47) + (($3*12)|0)|0)) + 4|0); //@line 57 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $51 = +HEAPF32[$50>>2]; //@line 57 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $52 = (((($47) + (($3*12)|0)|0)) + 8|0); //@line 58 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $53 = +HEAPF32[$52>>2]; //@line 58 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $54 = ((($data)) + 28|0); //@line 59 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $55 = HEAP32[$54>>2]|0; //@line 59 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $56 = (($55) + (($3*12)|0)|0); //@line 59 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $57 = HEAP32[$56>>2]|0; //@line 59 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $58 = (((($55) + (($3*12)|0)|0)) + 4|0); //@line 59 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $59 = HEAP32[$58>>2]|0; //@line 59 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $60 = (((($55) + (($3*12)|0)|0)) + 8|0); //@line 60 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $61 = +HEAPF32[$60>>2]; //@line 60 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $62 = (($47) + (($8*12)|0)|0); //@line 62 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $63 = +HEAPF32[$62>>2]; //@line 62 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $64 = (((($47) + (($8*12)|0)|0)) + 4|0); //@line 62 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $65 = +HEAPF32[$64>>2]; //@line 62 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $66 = (((($47) + (($8*12)|0)|0)) + 8|0); //@line 63 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $67 = +HEAPF32[$66>>2]; //@line 63 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $68 = (($55) + (($8*12)|0)|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $69 = HEAP32[$68>>2]|0; //@line 64 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $70 = (((($55) + (($8*12)|0)|0)) + 4|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $71 = HEAP32[$70>>2]|0; //@line 64 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $72 = (((($55) + (($8*12)|0)|0)) + 8|0); //@line 65 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $73 = +HEAPF32[$72>>2]; //@line 65 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $74 = (+Math_sin((+$53))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $75 = (+Math_cos((+$53))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $76 = (+Math_sin((+$67))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $77 = (+Math_cos((+$67))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $78 = ((($this)) + 68|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $79 = +HEAPF32[$78>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $bitcast = (HEAP32[tempDoublePtr>>2]=$14,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $80 = $79 - $bitcast; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $81 = ((($this)) + 72|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $82 = +HEAPF32[$81>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $83 = (HEAP32[tempDoublePtr>>2]=$17,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $84 = $82 - $83; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $85 = $75 * $80; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $86 = $74 * $84; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $87 = $85 - $86; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $88 = $74 * $80; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $89 = $75 * $84; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $90 = $88 + $89; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $91 = ((($this)) + 112|0); //@line 69 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$91>>2] = $87; //@line 69 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $92 = ((($this)) + 116|0); //@line 69 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$92>>2] = $90; //@line 69 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $93 = ((($this)) + 76|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $94 = +HEAPF32[$93>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $bitcast24 = (HEAP32[tempDoublePtr>>2]=$26,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $95 = $94 - $bitcast24; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $96 = ((($this)) + 80|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $97 = +HEAPF32[$96>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $98 = (HEAP32[tempDoublePtr>>2]=$29,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $99 = $97 - $98; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $100 = $77 * $95; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $101 = $76 * $99; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $102 = $100 - $101; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $103 = $76 * $95; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $104 = $77 * $99; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $105 = $103 + $104; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $106 = ((($this)) + 120|0); //@line 70 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$106>>2] = $102; //@line 70 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $107 = ((($this)) + 124|0); //@line 70 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$107>>2] = $105; //@line 70 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $108 = $63 + $102; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $109 = $65 + $105; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $110 = $108 - $49; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $111 = $109 - $51; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $112 = $110 - $87; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 $113 = $111 - $90; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 $114 = ((($this)) + 104|0); //@line 71 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$114>>2] = $112; //@line 71 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $115 = ((($this)) + 108|0); //@line 71 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$115>>2] = $113; //@line 71 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $116 = $112 * $112; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $117 = $113 * $113; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $118 = $116 + $117; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $119 = (+Math_sqrt((+$118))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $120 = ((($this)) + 88|0); //@line 73 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$120>>2] = $119; //@line 73 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $121 = ((($this)) + 84|0); //@line 75 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $122 = +HEAPF32[$121>>2]; //@line 75 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $123 = $119 - $122; //@line 75 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $124 = $123 > 0.0; //@line 76 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $125 = ((($this)) + 164|0); //@line 78 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $$ = $124 ? 2 : 0; //@line 79 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$125>>2] = $$; //@line 78 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $126 = $119 > 0.004999999888241291; //@line 85 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 if (!($126)) {
  HEAPF32[$114>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$115>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $154 = ((($this)) + 160|0); //@line 92 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  HEAPF32[$154>>2] = 0.0; //@line 92 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  $155 = ((($this)) + 92|0); //@line 93 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  HEAPF32[$155>>2] = 0.0; //@line 93 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  return; //@line 124 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 }
 $127 = 1.0 / $119; //@line 87 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $128 = $127 * $112; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$114>>2] = $128; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 $129 = $127 * $113; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$115>>2] = $129; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 $130 = $129 * $87; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $131 = $90 * $128; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $132 = $130 - $131; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $133 = $129 * $102; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $134 = $128 * $105; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $135 = $133 - $134; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $136 = +HEAPF32[$36>>2]; //@line 110 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $137 = +HEAPF32[$42>>2]; //@line 111 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $138 = $132 * $137; //@line 100 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $139 = $132 * $138; //@line 100 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $140 = $136 + $139; //@line 100 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $141 = +HEAPF32[$39>>2]; //@line 112 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $142 = $141 + $140; //@line 100 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $143 = +HEAPF32[$45>>2]; //@line 113 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $144 = $135 * $143; //@line 100 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $145 = $135 * $144; //@line 100 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $146 = $142 + $145; //@line 100 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $147 = $146 != 0.0; //@line 102 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $148 = 1.0 / $146; //@line 102 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $149 = $147 ? $148 : 0.0; //@line 102 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $150 = ((($this)) + 160|0); //@line 102 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$150>>2] = $149; //@line 102 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $151 = ((($data)) + 20|0); //@line 104 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $152 = HEAP8[$151>>0]|0; //@line 104 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $153 = ($152<<24>>24)==(0); //@line 104 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 if ($153) {
  $189 = ((($this)) + 92|0); //@line 117 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  HEAPF32[$189>>2] = 0.0; //@line 117 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  $vA$sroa$0$0 = $57;$vA$sroa$6$0 = $59;$vB$sroa$0$0 = $69;$vB$sroa$6$0 = $71;$wA$0 = $61;$wB$0 = $73;
 } else {
  $156 = ((($data)) + 8|0); //@line 107 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  $157 = +HEAPF32[$156>>2]; //@line 107 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  $158 = ((($this)) + 92|0); //@line 107 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  $159 = +HEAPF32[$158>>2]; //@line 107 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  $160 = $157 * $159; //@line 107 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  HEAPF32[$158>>2] = $160; //@line 107 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  $161 = $160 * $128; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $162 = $160 * $129; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $163 = $161 * $136; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $164 = $162 * $136; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $165 = (HEAP32[tempDoublePtr>>2]=$57,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $166 = $165 - $163; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $167 = (HEAPF32[tempDoublePtr>>2]=$166,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $168 = (HEAP32[tempDoublePtr>>2]=$59,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $169 = $168 - $164; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $170 = (HEAPF32[tempDoublePtr>>2]=$169,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $171 = $162 * $87; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $172 = $161 * $90; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $173 = $171 - $172; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $174 = $137 * $173; //@line 111 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  $175 = $61 - $174; //@line 111 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  $176 = $161 * $141; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $177 = $162 * $141; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $178 = (HEAP32[tempDoublePtr>>2]=$69,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $179 = $178 + $176; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $180 = (HEAPF32[tempDoublePtr>>2]=$179,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $181 = (HEAP32[tempDoublePtr>>2]=$71,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $182 = $181 + $177; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $183 = (HEAPF32[tempDoublePtr>>2]=$182,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $184 = $162 * $102; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $185 = $161 * $105; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $186 = $184 - $185; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $187 = $143 * $186; //@line 113 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  $188 = $73 + $187; //@line 113 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  $vA$sroa$0$0 = $167;$vA$sroa$6$0 = $170;$vB$sroa$0$0 = $180;$vB$sroa$6$0 = $183;$wA$0 = $175;$wB$0 = $188;
 }
 $190 = HEAP32[$4>>2]|0; //@line 120 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $191 = HEAP32[$54>>2]|0; //@line 120 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $192 = (($191) + (($190*12)|0)|0); //@line 120 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$192>>2] = $vA$sroa$0$0; //@line 120 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $193 = (((($191) + (($190*12)|0)|0)) + 4|0); //@line 120 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$193>>2] = $vA$sroa$6$0; //@line 120 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $194 = HEAP32[$4>>2]|0; //@line 121 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $195 = HEAP32[$54>>2]|0; //@line 121 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $196 = (((($195) + (($194*12)|0)|0)) + 8|0); //@line 121 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$196>>2] = $wA$0; //@line 121 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $197 = HEAP32[$9>>2]|0; //@line 122 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $198 = (($195) + (($197*12)|0)|0); //@line 122 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$198>>2] = $vB$sroa$0$0; //@line 122 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $199 = (((($195) + (($197*12)|0)|0)) + 4|0); //@line 122 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$199>>2] = $vB$sroa$6$0; //@line 122 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $200 = HEAP32[$9>>2]|0; //@line 123 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $201 = HEAP32[$54>>2]|0; //@line 123 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $202 = (((($201) + (($200*12)|0)|0)) + 8|0); //@line 123 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$202>>2] = $wB$0; //@line 123 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 return; //@line 124 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
}
function __ZN11b2RopeJoint24SolveVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0;
 var $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0, $49 = 0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0;
 var $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0.0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0;
 var $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0.0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $Cdot$0 = 0.0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 96|0); //@line 128 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 156 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $2 = ((($data)) + 28|0); //@line 128 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 156 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $4 = (($3) + (($1*12)|0)|0); //@line 128 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 128 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $6 = (((($3) + (($1*12)|0)|0)) + 4|0); //@line 128 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 128 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $8 = (((($3) + (($1*12)|0)|0)) + 8|0); //@line 129 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 129 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $10 = ((($this)) + 100|0); //@line 130 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 130 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $12 = (($3) + (($11*12)|0)|0); //@line 130 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $13 = +HEAPF32[$12>>2]; //@line 130 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $14 = (((($3) + (($11*12)|0)|0)) + 4|0); //@line 130 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 130 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $16 = (((($3) + (($11*12)|0)|0)) + 8|0); //@line 131 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 131 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $18 = ((($this)) + 116|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $19 = +HEAPF32[$18>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $20 = $9 * $19; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($this)) + 112|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $22 = +HEAPF32[$21>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $23 = $9 * $22; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $24 = $5 - $20; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $25 = $7 + $23; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $26 = ((($this)) + 124|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $27 = +HEAPF32[$26>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $28 = $17 * $27; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $29 = ((($this)) + 120|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $30 = +HEAPF32[$29>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $31 = $17 * $30; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $32 = $13 - $28; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $33 = $15 + $31; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $34 = ((($this)) + 88|0); //@line 136 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $35 = +HEAPF32[$34>>2]; //@line 136 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $36 = ((($this)) + 84|0); //@line 136 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $37 = +HEAPF32[$36>>2]; //@line 136 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $38 = $35 - $37; //@line 136 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $39 = $32 - $24; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $40 = $33 - $25; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $41 = ((($this)) + 104|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $42 = +HEAPF32[$41>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $43 = $39 * $42; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $44 = ((($this)) + 108|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $45 = +HEAPF32[$44>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $46 = $40 * $45; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $47 = $43 + $46; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $48 = $38 < 0.0; //@line 140 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 if ($48) {
  $49 = ((($data)) + 4|0); //@line 142 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  $50 = +HEAPF32[$49>>2]; //@line 142 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  $51 = $38 * $50; //@line 142 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  $52 = $47 + $51; //@line 142 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
  $Cdot$0 = $52;
 } else {
  $Cdot$0 = $47;
 }
 $53 = ((($this)) + 160|0); //@line 145 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $54 = +HEAPF32[$53>>2]; //@line 145 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $55 = $Cdot$0 * $54; //@line 145 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $56 = ((($this)) + 92|0); //@line 146 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $57 = +HEAPF32[$56>>2]; //@line 146 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $58 = $57 - $55; //@line 147 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $59 = $58 > 0.0; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $60 = $59 ? 0.0 : $58; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$56>>2] = $60; //@line 147 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $61 = $60 - $57; //@line 148 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $62 = $42 * $61; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $63 = $45 * $61; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $64 = ((($this)) + 144|0); //@line 151 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $65 = +HEAPF32[$64>>2]; //@line 151 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $66 = $65 * $62; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $67 = $65 * $63; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $68 = $5 - $66; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $69 = $7 - $67; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $70 = ((($this)) + 152|0); //@line 152 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $71 = +HEAPF32[$70>>2]; //@line 152 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $72 = $63 * $22; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $73 = $62 * $19; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $74 = $72 - $73; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $75 = $71 * $74; //@line 152 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $76 = $9 - $75; //@line 152 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $77 = ((($this)) + 148|0); //@line 153 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $78 = +HEAPF32[$77>>2]; //@line 153 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $79 = $62 * $78; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $80 = $63 * $78; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $81 = $13 + $79; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $82 = $15 + $80; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $83 = ((($this)) + 156|0); //@line 154 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $84 = +HEAPF32[$83>>2]; //@line 154 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $85 = $63 * $30; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $86 = $62 * $27; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $87 = $85 - $86; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $88 = $84 * $87; //@line 154 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $89 = $17 + $88; //@line 154 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$4>>2] = $68; //@line 156 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$6>>2] = $69; //@line 156 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $90 = HEAP32[$0>>2]|0; //@line 157 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $91 = HEAP32[$2>>2]|0; //@line 157 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $92 = (((($91) + (($90*12)|0)|0)) + 8|0); //@line 157 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$92>>2] = $76; //@line 157 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $93 = HEAP32[$10>>2]|0; //@line 158 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $94 = (($91) + (($93*12)|0)|0); //@line 158 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$94>>2] = $81; //@line 158 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $95 = (((($91) + (($93*12)|0)|0)) + 4|0); //@line 158 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$95>>2] = $82; //@line 158 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $96 = HEAP32[$10>>2]|0; //@line 159 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $97 = HEAP32[$2>>2]|0; //@line 159 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $98 = (((($97) + (($96*12)|0)|0)) + 8|0); //@line 159 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$98>>2] = $89; //@line 159 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 return; //@line 160 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
}
function __ZN11b2RopeJoint24SolvePositionConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$0$i = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0.0, $117 = 0.0, $118 = 0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0;
 var $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0.0;
 var $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0;
 var $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0, $69 = 0.0, $7 = 0.0, $70 = 0.0, $71 = 0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0;
 var $81 = 0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0, $88 = 0.0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0;
 var $u$sroa$0$0 = 0.0, $u$sroa$6$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 96|0); //@line 164 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 188 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $2 = ((($data)) + 24|0); //@line 164 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 188 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $4 = (($3) + (($1*12)|0)|0); //@line 164 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 164 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $6 = (((($3) + (($1*12)|0)|0)) + 4|0); //@line 164 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 164 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $8 = (((($3) + (($1*12)|0)|0)) + 8|0); //@line 165 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 165 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $10 = ((($this)) + 100|0); //@line 166 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 166 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $12 = (($3) + (($11*12)|0)|0); //@line 166 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $13 = +HEAPF32[$12>>2]; //@line 166 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $14 = (((($3) + (($11*12)|0)|0)) + 4|0); //@line 166 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 166 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $16 = (((($3) + (($11*12)|0)|0)) + 8|0); //@line 167 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 167 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $18 = (+Math_sin((+$9))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $19 = (+Math_cos((+$9))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $20 = (+Math_sin((+$17))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $21 = (+Math_cos((+$17))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($this)) + 68|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $23 = +HEAPF32[$22>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($this)) + 128|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $26 = $23 - $25; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($this)) + 72|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $28 = +HEAPF32[$27>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $29 = ((($this)) + 132|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $30 = +HEAPF32[$29>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $31 = $28 - $30; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $32 = $19 * $26; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $33 = $18 * $31; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $34 = $32 - $33; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $35 = $18 * $26; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $36 = $19 * $31; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $37 = $35 + $36; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $38 = ((($this)) + 76|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $39 = +HEAPF32[$38>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $40 = ((($this)) + 136|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $41 = +HEAPF32[$40>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $42 = $39 - $41; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $43 = ((($this)) + 80|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $44 = +HEAPF32[$43>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $45 = ((($this)) + 140|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $46 = +HEAPF32[$45>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $47 = $44 - $46; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $48 = $21 * $42; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $49 = $20 * $47; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $50 = $48 - $49; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $51 = $20 * $42; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $52 = $21 * $47; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $53 = $51 + $52; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $54 = $13 + $50; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $55 = $15 + $53; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $56 = $54 - $5; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $57 = $55 - $7; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $58 = $56 - $34; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $59 = $57 - $37; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $60 = $58 * $58; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $61 = $59 * $59; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $62 = $60 + $61; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $63 = (+Math_sqrt((+$62))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $64 = $63 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
 if ($64) {
  $$0$i = 0.0;$u$sroa$0$0 = $58;$u$sroa$6$0 = $59;
 } else {
  $65 = 1.0 / $63; //@line 121 "./libraries/Box2D/Common/b2Math.h"
  $66 = $65 * $58; //@line 122 "./libraries/Box2D/Common/b2Math.h"
  $67 = $65 * $59; //@line 123 "./libraries/Box2D/Common/b2Math.h"
  $$0$i = $63;$u$sroa$0$0 = $66;$u$sroa$6$0 = $67;
 }
 $68 = ((($this)) + 84|0); //@line 176 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $69 = +HEAPF32[$68>>2]; //@line 176 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $70 = $$0$i - $69; //@line 176 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $71 = $70 < 0.20000000298023224; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $72 = $71 ? $70 : 0.20000000298023224; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $73 = $72 < 0.0; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $74 = $73 ? 0.0 : $72; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $75 = ((($this)) + 160|0); //@line 180 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $76 = +HEAPF32[$75>>2]; //@line 180 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $77 = $76 * $74; //@line 180 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $78 = -$77; //@line 180 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $79 = $u$sroa$0$0 * $78; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $80 = $u$sroa$6$0 * $78; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $81 = ((($this)) + 144|0); //@line 183 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $82 = +HEAPF32[$81>>2]; //@line 183 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $83 = $82 * $79; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $84 = $82 * $80; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $85 = $5 - $83; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $86 = $7 - $84; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $87 = ((($this)) + 152|0); //@line 184 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $88 = +HEAPF32[$87>>2]; //@line 184 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $89 = $34 * $80; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $90 = $37 * $79; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $91 = $89 - $90; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $92 = $88 * $91; //@line 184 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $93 = $9 - $92; //@line 184 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $94 = ((($this)) + 148|0); //@line 185 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $95 = +HEAPF32[$94>>2]; //@line 185 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $96 = $95 * $79; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $97 = $95 * $80; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $98 = $13 + $96; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $99 = $15 + $97; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $100 = ((($this)) + 156|0); //@line 186 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $101 = +HEAPF32[$100>>2]; //@line 186 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $102 = $50 * $80; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $103 = $53 * $79; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $104 = $102 - $103; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $105 = $101 * $104; //@line 186 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $106 = $17 + $105; //@line 186 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$4>>2] = $85; //@line 188 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$6>>2] = $86; //@line 188 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $107 = HEAP32[$0>>2]|0; //@line 189 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $108 = HEAP32[$2>>2]|0; //@line 189 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $109 = (((($108) + (($107*12)|0)|0)) + 8|0); //@line 189 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$109>>2] = $93; //@line 189 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $110 = HEAP32[$10>>2]|0; //@line 190 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $111 = (($108) + (($110*12)|0)|0); //@line 190 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$111>>2] = $98; //@line 190 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $112 = (((($108) + (($110*12)|0)|0)) + 4|0); //@line 190 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$112>>2] = $99; //@line 190 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $113 = HEAP32[$10>>2]|0; //@line 191 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $114 = HEAP32[$2>>2]|0; //@line 191 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $115 = (((($114) + (($113*12)|0)|0)) + 8|0); //@line 191 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$115>>2] = $106; //@line 191 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $116 = +HEAPF32[$68>>2]; //@line 193 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $117 = $$0$i - $116; //@line 193 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $118 = $117 < 0.004999999888241291; //@line 193 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 return ($118|0); //@line 193 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
}
function __ZN11b2RopeJointC2EPK14b2RopeJointDef($this,$def) {
 $this = $this|0;
 $def = $def|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN7b2JointC2EPK10b2JointDef($this,$def); //@line 33 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$this>>2] = (21024); //@line 34 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $0 = ((($this)) + 68|0); //@line 32 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $1 = ((($this)) + 76|0); //@line 32 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $2 = ((($def)) + 20|0); //@line 35 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $3 = $2; //@line 35 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $4 = $3; //@line 35 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 35 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $6 = (($3) + 4)|0; //@line 35 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $7 = $6; //@line 35 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 35 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $9 = $0; //@line 35 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $10 = $9; //@line 35 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$10>>2] = $5; //@line 35 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $11 = (($9) + 4)|0; //@line 35 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $12 = $11; //@line 35 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$12>>2] = $8; //@line 35 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $13 = ((($def)) + 28|0); //@line 36 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $14 = $13; //@line 36 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $15 = $14; //@line 36 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 36 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $17 = (($14) + 4)|0; //@line 36 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $18 = $17; //@line 36 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 36 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $20 = $1; //@line 36 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $21 = $20; //@line 36 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$21>>2] = $16; //@line 36 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $22 = (($20) + 4)|0; //@line 36 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $23 = $22; //@line 36 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$23>>2] = $19; //@line 36 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $24 = ((($def)) + 36|0); //@line 38 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $25 = HEAP32[$24>>2]|0; //@line 38 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $26 = ((($this)) + 84|0); //@line 38 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$26>>2] = $25; //@line 38 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $27 = ((($this)) + 160|0); //@line 40 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$27>>2] = 0.0; //@line 40 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $28 = ((($this)) + 92|0); //@line 41 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$28>>2] = 0.0; //@line 41 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $29 = ((($this)) + 164|0); //@line 42 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAP32[$29>>2] = 0; //@line 42 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $30 = ((($this)) + 88|0); //@line 43 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 HEAPF32[$30>>2] = 0.0; //@line 43 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 return; //@line 44 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
}
function __ZNK11b2RopeJoint12GetMaxLengthEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 84|0); //@line 220 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 220 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
 return (+$1); //@line 220 "libraries/Box2D/Dynamics/Joints/b2RopeJoint.cpp"
}
function __ZNK11b2WeldJoint10GetAnchorAEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 310 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 310 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 80|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 84|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 310 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
}
function __ZNK11b2WeldJoint10GetAnchorBEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 315 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 315 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 88|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 92|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 315 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
}
function __ZNK11b2WeldJoint16GetReactionForceEf($agg$result,$this,$inv_dt) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 104|0); //@line 320 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 320 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $2 = ((($this)) + 108|0); //@line 320 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 320 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $4 = $1 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $5 = $3 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $4; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$6>>2] = $5; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 322 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
}
function __ZNK11b2WeldJoint17GetReactionTorqueEf($this,$inv_dt) {
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 326 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 326 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $2 = $1 * $inv_dt; //@line 326 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 return (+$2); //@line 326 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
}
function __ZN11b2WeldJoint4DumpEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer13 = 0, $vararg_buffer17 = 0, $vararg_buffer20 = 0, $vararg_buffer23 = 0;
 var $vararg_buffer26 = 0, $vararg_buffer3 = 0, $vararg_buffer6 = 0, $vararg_buffer9 = 0, $vararg_ptr12 = 0, $vararg_ptr16 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer26 = sp + 88|0;
 $vararg_buffer23 = sp + 80|0;
 $vararg_buffer20 = sp + 72|0;
 $vararg_buffer17 = sp + 64|0;
 $vararg_buffer13 = sp + 48|0;
 $vararg_buffer9 = sp + 32|0;
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($this)) + 48|0); //@line 331 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 331 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 331 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 331 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $4 = ((($this)) + 52|0); //@line 332 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 332 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $6 = ((($5)) + 8|0); //@line 332 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 332 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 __Z5b2LogPKcz(363424,$vararg_buffer); //@line 334 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$vararg_buffer1>>2] = $3; //@line 335 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 __Z5b2LogPKcz(447349,$vararg_buffer1); //@line 335 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$vararg_buffer3>>2] = $7; //@line 336 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 __Z5b2LogPKcz(447375,$vararg_buffer3); //@line 336 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $8 = ((($this)) + 61|0); //@line 337 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 337 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $10 = $9&255; //@line 337 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$vararg_buffer6>>2] = $10; //@line 337 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 __Z5b2LogPKcz(447401,$vararg_buffer6); //@line 337 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $11 = ((($this)) + 80|0); //@line 338 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $12 = +HEAPF32[$11>>2]; //@line 338 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $13 = $12; //@line 338 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $14 = ((($this)) + 84|0); //@line 338 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 338 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $16 = $15; //@line 338 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF64[$vararg_buffer9>>3] = $13; //@line 338 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $vararg_ptr12 = ((($vararg_buffer9)) + 8|0); //@line 338 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF64[$vararg_ptr12>>3] = $16; //@line 338 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 __Z5b2LogPKcz(447106,$vararg_buffer9); //@line 338 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $17 = ((($this)) + 88|0); //@line 339 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $18 = +HEAPF32[$17>>2]; //@line 339 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $19 = $18; //@line 339 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $20 = ((($this)) + 92|0); //@line 339 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 339 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $22 = $21; //@line 339 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF64[$vararg_buffer13>>3] = $19; //@line 339 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $vararg_ptr16 = ((($vararg_buffer13)) + 8|0); //@line 339 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF64[$vararg_ptr16>>3] = $22; //@line 339 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 __Z5b2LogPKcz(447148,$vararg_buffer13); //@line 339 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $23 = ((($this)) + 96|0); //@line 340 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $24 = +HEAPF32[$23>>2]; //@line 340 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $25 = $24; //@line 340 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF64[$vararg_buffer17>>3] = $25; //@line 340 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 __Z5b2LogPKcz(363446,$vararg_buffer17); //@line 340 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $26 = ((($this)) + 68|0); //@line 341 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $27 = +HEAPF32[$26>>2]; //@line 341 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $28 = $27; //@line 341 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF64[$vararg_buffer20>>3] = $28; //@line 341 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 __Z5b2LogPKcz(447003,$vararg_buffer20); //@line 341 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $29 = ((($this)) + 72|0); //@line 342 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $30 = +HEAPF32[$29>>2]; //@line 342 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $31 = $30; //@line 342 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF64[$vararg_buffer23>>3] = $31; //@line 342 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 __Z5b2LogPKcz(447032,$vararg_buffer23); //@line 342 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $32 = ((($this)) + 56|0); //@line 343 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $33 = HEAP32[$32>>2]|0; //@line 343 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$vararg_buffer26>>2] = $33; //@line 343 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 __Z5b2LogPKcz(447513,$vararg_buffer26); //@line 343 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 STACKTOP = sp;return; //@line 344 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
}
function __ZN11b2WeldJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 62 "./libraries/Box2D/Dynamics/Joints/b2WeldJoint.h"
 return; //@line 62 "./libraries/Box2D/Dynamics/Joints/b2WeldJoint.h"
}
function __ZN11b2WeldJoint23InitVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0;
 var $116 = 0.0, $117 = 0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0, $123 = 0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0, $131 = 0.0, $132 = 0.0, $133 = 0.0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0.0, $138 = 0, $139 = 0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0, $143 = 0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0, $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0.0;
 var $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0, $164 = 0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0.0, $17 = 0;
 var $170 = 0, $171 = 0.0, $172 = 0, $173 = 0.0, $174 = 0.0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0.0, $188 = 0.0;
 var $189 = 0.0, $19 = 0, $190 = 0, $191 = 0.0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0.0, $196 = 0.0, $197 = 0.0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0, $201 = 0.0, $202 = 0.0, $203 = 0, $204 = 0.0, $205 = 0.0;
 var $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0.0, $212 = 0.0, $213 = 0.0, $214 = 0.0, $215 = 0.0, $216 = 0, $217 = 0.0, $218 = 0.0, $219 = 0, $22 = 0, $220 = 0.0, $221 = 0.0, $222 = 0.0, $223 = 0.0;
 var $224 = 0.0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0;
 var $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0, $85 = 0, $86 = 0.0, $87 = 0.0, $88 = 0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0;
 var $97 = 0.0, $98 = 0, $99 = 0, $K = 0, $bitcast = 0.0, $bitcast15 = 0.0, $vA$sroa$0$0 = 0, $vA$sroa$6$0 = 0, $vB$sroa$0$0 = 0, $vB$sroa$6$0 = 0, $wA$0 = 0.0, $wB$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $K = sp;
 $0 = ((($this)) + 48|0); //@line 60 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 60 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 60 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 60 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $4 = ((($this)) + 116|0); //@line 60 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$4>>2] = $3; //@line 60 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $5 = ((($this)) + 52|0); //@line 61 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 61 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $7 = ((($6)) + 8|0); //@line 61 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 61 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $9 = ((($this)) + 120|0); //@line 61 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$9>>2] = $8; //@line 61 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $10 = ((($this)) + 140|0); //@line 62 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $11 = ((($1)) + 28|0); //@line 62 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $12 = $11; //@line 62 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $13 = $12; //@line 62 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 62 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $15 = (($12) + 4)|0; //@line 62 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $16 = $15; //@line 62 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 62 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $18 = $10; //@line 62 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $19 = $18; //@line 62 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$19>>2] = $14; //@line 62 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $20 = (($18) + 4)|0; //@line 62 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $21 = $20; //@line 62 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$21>>2] = $17; //@line 62 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $22 = ((($this)) + 148|0); //@line 63 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $23 = ((($6)) + 28|0); //@line 63 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $24 = $23; //@line 63 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $25 = $24; //@line 63 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 63 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $27 = (($24) + 4)|0; //@line 63 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $28 = $27; //@line 63 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 63 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $30 = $22; //@line 63 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $31 = $30; //@line 63 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$31>>2] = $26; //@line 63 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $32 = (($30) + 4)|0; //@line 63 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $33 = $32; //@line 63 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$33>>2] = $29; //@line 63 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $34 = ((($1)) + 120|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 64 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $36 = ((($this)) + 156|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$36>>2] = $35; //@line 64 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $37 = ((($6)) + 120|0); //@line 65 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $38 = HEAP32[$37>>2]|0; //@line 65 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $39 = ((($this)) + 160|0); //@line 65 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$39>>2] = $38; //@line 65 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $40 = ((($1)) + 128|0); //@line 66 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $41 = HEAP32[$40>>2]|0; //@line 66 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $42 = ((($this)) + 164|0); //@line 66 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$42>>2] = $41; //@line 66 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $43 = ((($6)) + 128|0); //@line 67 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $44 = HEAP32[$43>>2]|0; //@line 67 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $45 = ((($this)) + 168|0); //@line 67 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$45>>2] = $44; //@line 67 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $46 = ((($data)) + 24|0); //@line 69 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $47 = HEAP32[$46>>2]|0; //@line 69 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $48 = (((($47) + (($3*12)|0)|0)) + 8|0); //@line 69 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $49 = +HEAPF32[$48>>2]; //@line 69 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $50 = ((($data)) + 28|0); //@line 70 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $51 = HEAP32[$50>>2]|0; //@line 70 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $52 = (($51) + (($3*12)|0)|0); //@line 70 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $53 = HEAP32[$52>>2]|0; //@line 70 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $54 = (((($51) + (($3*12)|0)|0)) + 4|0); //@line 70 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $55 = HEAP32[$54>>2]|0; //@line 70 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $56 = (((($51) + (($3*12)|0)|0)) + 8|0); //@line 71 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $57 = +HEAPF32[$56>>2]; //@line 71 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $58 = (((($47) + (($8*12)|0)|0)) + 8|0); //@line 73 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $59 = +HEAPF32[$58>>2]; //@line 73 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $60 = (($51) + (($8*12)|0)|0); //@line 74 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $61 = HEAP32[$60>>2]|0; //@line 74 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $62 = (((($51) + (($8*12)|0)|0)) + 4|0); //@line 74 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $63 = HEAP32[$62>>2]|0; //@line 74 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $64 = (((($51) + (($8*12)|0)|0)) + 8|0); //@line 75 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $65 = +HEAPF32[$64>>2]; //@line 75 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $66 = (+Math_sin((+$49))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $67 = (+Math_cos((+$49))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $68 = (+Math_sin((+$59))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $69 = (+Math_cos((+$59))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $70 = ((($this)) + 80|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $71 = +HEAPF32[$70>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $bitcast = (HEAP32[tempDoublePtr>>2]=$14,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $72 = $71 - $bitcast; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $73 = ((($this)) + 84|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $74 = +HEAPF32[$73>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $75 = (HEAP32[tempDoublePtr>>2]=$17,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $76 = $74 - $75; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $77 = $67 * $72; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $78 = $66 * $76; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $79 = $77 - $78; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $80 = $66 * $72; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $81 = $67 * $76; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $82 = $80 + $81; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $83 = ((($this)) + 124|0); //@line 79 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$83>>2] = $79; //@line 79 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $84 = ((($this)) + 128|0); //@line 79 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$84>>2] = $82; //@line 79 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $85 = ((($this)) + 88|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $86 = +HEAPF32[$85>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $bitcast15 = (HEAP32[tempDoublePtr>>2]=$26,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $87 = $86 - $bitcast15; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $88 = ((($this)) + 92|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $89 = +HEAPF32[$88>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $90 = (HEAP32[tempDoublePtr>>2]=$29,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $91 = $89 - $90; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $92 = $69 * $87; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $93 = $68 * $91; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $94 = $92 - $93; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $95 = $68 * $87; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $96 = $69 * $91; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $97 = $95 + $96; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $98 = ((($this)) + 132|0); //@line 80 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$98>>2] = $94; //@line 80 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $99 = ((($this)) + 136|0); //@line 80 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$99>>2] = $97; //@line 80 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $100 = (HEAP32[tempDoublePtr>>2]=$35,+HEAPF32[tempDoublePtr>>2]); //@line 91 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $101 = (HEAP32[tempDoublePtr>>2]=$38,+HEAPF32[tempDoublePtr>>2]); //@line 91 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $102 = (HEAP32[tempDoublePtr>>2]=$41,+HEAPF32[tempDoublePtr>>2]); //@line 92 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $103 = (HEAP32[tempDoublePtr>>2]=$44,+HEAPF32[tempDoublePtr>>2]); //@line 92 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $104 = $100 + $101; //@line 95 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $105 = $82 * $82; //@line 95 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $106 = $102 * $105; //@line 95 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $107 = $104 + $106; //@line 95 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $108 = $97 * $97; //@line 95 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $109 = $103 * $108; //@line 95 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $110 = $107 + $109; //@line 95 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$K>>2] = $110; //@line 95 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $111 = $82 * $79; //@line 96 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $112 = $102 * $111; //@line 96 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $113 = -$112; //@line 96 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $114 = $97 * $94; //@line 96 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $115 = $103 * $114; //@line 96 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $116 = $113 - $115; //@line 96 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $117 = ((($K)) + 12|0); //@line 96 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$117>>2] = $116; //@line 96 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $118 = $102 * $82; //@line 97 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $119 = -$118; //@line 97 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $120 = $103 * $97; //@line 97 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $121 = $119 - $120; //@line 97 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $122 = ((($K)) + 24|0); //@line 97 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$122>>2] = $121; //@line 97 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $123 = ((($K)) + 4|0); //@line 98 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$123>>2] = $116; //@line 98 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $124 = $79 * $79; //@line 99 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $125 = $102 * $124; //@line 99 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $126 = $104 + $125; //@line 99 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $127 = $94 * $94; //@line 99 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $128 = $103 * $127; //@line 99 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $129 = $126 + $128; //@line 99 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $130 = ((($K)) + 16|0); //@line 99 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$130>>2] = $129; //@line 99 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $131 = $102 * $79; //@line 100 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $132 = $103 * $94; //@line 100 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $133 = $131 + $132; //@line 100 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $134 = ((($K)) + 28|0); //@line 100 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$134>>2] = $133; //@line 100 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $135 = ((($K)) + 8|0); //@line 101 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$135>>2] = $121; //@line 101 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $136 = ((($K)) + 20|0); //@line 102 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$136>>2] = $133; //@line 102 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $137 = $102 + $103; //@line 103 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $138 = ((($K)) + 32|0); //@line 103 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$138>>2] = $137; //@line 103 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $139 = ((($this)) + 68|0); //@line 105 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $140 = +HEAPF32[$139>>2]; //@line 105 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $141 = $140 > 0.0; //@line 105 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 do {
  if ($141) {
   $142 = ((($this)) + 172|0); //@line 107 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   __ZNK7b2Mat3312GetInverse22EPS_($K,$142); //@line 107 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $143 = $137 > 0.0; //@line 110 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $144 = 1.0 / $137; //@line 110 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $145 = $143 ? $144 : 0.0; //@line 110 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $146 = $59 - $49; //@line 112 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $147 = ((($this)) + 96|0); //@line 112 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $148 = +HEAPF32[$147>>2]; //@line 112 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $149 = $146 - $148; //@line 112 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $150 = +HEAPF32[$139>>2]; //@line 115 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $151 = $150 * 6.2831854820251465; //@line 115 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $152 = $145 * 2.0; //@line 118 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $153 = ((($this)) + 72|0); //@line 118 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $154 = +HEAPF32[$153>>2]; //@line 118 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $155 = $152 * $154; //@line 118 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $156 = $151 * $155; //@line 118 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $157 = $145 * $151; //@line 121 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $158 = $151 * $157; //@line 121 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $159 = +HEAPF32[$data>>2]; //@line 124 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $160 = $159 * $158; //@line 125 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $161 = $156 + $160; //@line 125 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $162 = $159 * $161; //@line 125 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $163 = ((($this)) + 100|0); //@line 125 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $164 = $162 != 0.0; //@line 126 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $165 = 1.0 / $162; //@line 126 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $166 = $164 ? $165 : 0.0; //@line 126 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   HEAPF32[$163>>2] = $166; //@line 126 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $167 = $149 * $159; //@line 127 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $168 = $167 * $158; //@line 127 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $169 = $168 * $166; //@line 127 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $170 = ((($this)) + 76|0); //@line 127 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   HEAPF32[$170>>2] = $169; //@line 127 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $171 = $137 + $166; //@line 129 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $172 = $171 != 0.0; //@line 130 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $173 = 1.0 / $171; //@line 130 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $174 = $172 ? $173 : 0.0; //@line 130 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $175 = ((($this)) + 204|0); //@line 130 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   HEAPF32[$175>>2] = $174; //@line 130 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  } else {
   $176 = $137 == 0.0; //@line 132 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   $177 = ((($this)) + 172|0); //@line 134 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
   if ($176) {
    __ZNK7b2Mat3312GetInverse22EPS_($K,$177); //@line 134 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
    $178 = ((($this)) + 100|0); //@line 135 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
    HEAPF32[$178>>2] = 0.0; //@line 135 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
    $179 = ((($this)) + 76|0); //@line 136 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
    HEAPF32[$179>>2] = 0.0; //@line 136 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
    break;
   } else {
    __ZNK7b2Mat3315GetSymInverse33EPS_($K,$177); //@line 140 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
    $180 = ((($this)) + 100|0); //@line 141 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
    HEAPF32[$180>>2] = 0.0; //@line 141 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
    $181 = ((($this)) + 76|0); //@line 142 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
    HEAPF32[$181>>2] = 0.0; //@line 142 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
    break;
   }
  }
 } while(0);
 $182 = ((($data)) + 20|0); //@line 145 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $183 = HEAP8[$182>>0]|0; //@line 145 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $184 = ($183<<24>>24)==(0); //@line 145 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $185 = ((($this)) + 104|0); //@line 148 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 if ($184) {
  HEAPF32[$185>>2] = 0.0; //@line 153 "./libraries/Box2D/Common/b2Math.h"
  $228 = ((($this)) + 108|0); //@line 153 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$228>>2] = 0.0; //@line 153 "./libraries/Box2D/Common/b2Math.h"
  $229 = ((($this)) + 112|0); //@line 153 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$229>>2] = 0.0; //@line 153 "./libraries/Box2D/Common/b2Math.h"
  $vA$sroa$0$0 = $53;$vA$sroa$6$0 = $55;$vB$sroa$0$0 = $61;$vB$sroa$6$0 = $63;$wA$0 = $57;$wB$0 = $65;
  $230 = HEAP32[$4>>2]|0; //@line 163 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $231 = HEAP32[$50>>2]|0; //@line 163 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $232 = (($231) + (($230*12)|0)|0); //@line 163 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAP32[$232>>2] = $vA$sroa$0$0; //@line 163 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $233 = (((($231) + (($230*12)|0)|0)) + 4|0); //@line 163 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAP32[$233>>2] = $vA$sroa$6$0; //@line 163 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $234 = HEAP32[$4>>2]|0; //@line 164 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $235 = HEAP32[$50>>2]|0; //@line 164 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $236 = (((($235) + (($234*12)|0)|0)) + 8|0); //@line 164 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$236>>2] = $wA$0; //@line 164 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $237 = HEAP32[$9>>2]|0; //@line 165 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $238 = (($235) + (($237*12)|0)|0); //@line 165 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAP32[$238>>2] = $vB$sroa$0$0; //@line 165 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $239 = (((($235) + (($237*12)|0)|0)) + 4|0); //@line 165 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAP32[$239>>2] = $vB$sroa$6$0; //@line 165 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $240 = HEAP32[$9>>2]|0; //@line 166 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $241 = HEAP32[$50>>2]|0; //@line 166 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $242 = (((($241) + (($240*12)|0)|0)) + 8|0); //@line 166 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$242>>2] = $wB$0; //@line 166 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  STACKTOP = sp;return; //@line 167 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 } else {
  $186 = ((($data)) + 8|0); //@line 148 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $187 = +HEAPF32[$186>>2]; //@line 148 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $188 = +HEAPF32[$185>>2]; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $189 = $187 * $188; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$185>>2] = $189; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $190 = ((($this)) + 108|0); //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $191 = +HEAPF32[$190>>2]; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $192 = $187 * $191; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$190>>2] = $192; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $193 = ((($this)) + 112|0); //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $194 = +HEAPF32[$193>>2]; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $195 = $187 * $194; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$193>>2] = $195; //@line 176 "./libraries/Box2D/Common/b2Math.h"
  $196 = $100 * $189; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $197 = $100 * $192; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $198 = (HEAP32[tempDoublePtr>>2]=$53,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $199 = $198 - $196; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $200 = (HEAPF32[tempDoublePtr>>2]=$199,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $201 = (HEAP32[tempDoublePtr>>2]=$55,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $202 = $201 - $197; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $203 = (HEAPF32[tempDoublePtr>>2]=$202,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $204 = +HEAPF32[$83>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $205 = $192 * $204; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $206 = +HEAPF32[$84>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $207 = $189 * $206; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $208 = $205 - $207; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $209 = $195 + $208; //@line 153 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $210 = $102 * $209; //@line 153 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $211 = $57 - $210; //@line 153 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $212 = $101 * $189; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $213 = $101 * $192; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $214 = (HEAP32[tempDoublePtr>>2]=$61,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $215 = $214 + $212; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $216 = (HEAPF32[tempDoublePtr>>2]=$215,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $217 = (HEAP32[tempDoublePtr>>2]=$63,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $218 = $217 + $213; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $219 = (HEAPF32[tempDoublePtr>>2]=$218,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $220 = +HEAPF32[$98>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $221 = $192 * $220; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $222 = +HEAPF32[$99>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $223 = $189 * $222; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $224 = $221 - $223; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $225 = $195 + $224; //@line 156 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $226 = $103 * $225; //@line 156 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $227 = $65 + $226; //@line 156 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $vA$sroa$0$0 = $200;$vA$sroa$6$0 = $203;$vB$sroa$0$0 = $216;$vB$sroa$6$0 = $219;$wA$0 = $211;$wB$0 = $227;
  $230 = HEAP32[$4>>2]|0; //@line 163 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $231 = HEAP32[$50>>2]|0; //@line 163 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $232 = (($231) + (($230*12)|0)|0); //@line 163 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAP32[$232>>2] = $vA$sroa$0$0; //@line 163 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $233 = (((($231) + (($230*12)|0)|0)) + 4|0); //@line 163 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAP32[$233>>2] = $vA$sroa$6$0; //@line 163 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $234 = HEAP32[$4>>2]|0; //@line 164 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $235 = HEAP32[$50>>2]|0; //@line 164 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $236 = (((($235) + (($234*12)|0)|0)) + 8|0); //@line 164 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$236>>2] = $wA$0; //@line 164 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $237 = HEAP32[$9>>2]|0; //@line 165 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $238 = (($235) + (($237*12)|0)|0); //@line 165 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAP32[$238>>2] = $vB$sroa$0$0; //@line 165 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $239 = (((($235) + (($237*12)|0)|0)) + 4|0); //@line 165 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAP32[$239>>2] = $vB$sroa$6$0; //@line 165 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $240 = HEAP32[$9>>2]|0; //@line 166 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $241 = HEAP32[$50>>2]|0; //@line 166 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $242 = (((($241) + (($240*12)|0)|0)) + 8|0); //@line 166 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$242>>2] = $wB$0; //@line 166 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  STACKTOP = sp;return; //@line 167 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 }
}
function __ZN11b2WeldJoint24SolveVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0, $111 = 0.0, $112 = 0.0, $113 = 0, $114 = 0.0, $115 = 0.0;
 var $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0, $125 = 0.0, $126 = 0.0, $127 = 0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0.0, $132 = 0.0, $133 = 0;
 var $134 = 0.0, $135 = 0.0, $136 = 0, $137 = 0.0, $138 = 0.0, $139 = 0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0, $143 = 0.0, $144 = 0.0, $145 = 0, $146 = 0.0, $147 = 0.0, $148 = 0, $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0.0;
 var $152 = 0.0, $153 = 0.0, $154 = 0, $155 = 0.0, $156 = 0.0, $157 = 0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0, $167 = 0.0, $168 = 0.0, $169 = 0, $17 = 0.0;
 var $170 = 0.0, $171 = 0.0, $172 = 0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0.0, $182 = 0.0, $183 = 0.0, $184 = 0.0, $185 = 0.0, $186 = 0.0, $187 = 0.0, $188 = 0.0;
 var $189 = 0.0, $19 = 0.0, $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0.0, $194 = 0.0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $21 = 0.0;
 var $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0, $27 = 0.0, $28 = 0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0;
 var $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0;
 var $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0.0, $65 = 0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0.0, $76 = 0.0;
 var $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0, $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0.0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0;
 var $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $vA$sroa$0$0$in = 0.0, $vA$sroa$10$0$in = 0.0, $vB$sroa$0$0$in = 0.0, $vB$sroa$10$0$in = 0.0, $wA$0 = 0.0, $wB$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 116|0); //@line 171 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 221 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $2 = ((($data)) + 28|0); //@line 171 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 221 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $4 = (($3) + (($1*12)|0)|0); //@line 171 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 171 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $6 = (((($3) + (($1*12)|0)|0)) + 4|0); //@line 171 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 171 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $8 = (((($3) + (($1*12)|0)|0)) + 8|0); //@line 172 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 172 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $10 = ((($this)) + 120|0); //@line 173 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 173 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $12 = (($3) + (($11*12)|0)|0); //@line 173 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 173 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $14 = (((($3) + (($11*12)|0)|0)) + 4|0); //@line 173 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 173 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $16 = (((($3) + (($11*12)|0)|0)) + 8|0); //@line 174 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 174 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $18 = ((($this)) + 156|0); //@line 176 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $19 = +HEAPF32[$18>>2]; //@line 176 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $20 = ((($this)) + 160|0); //@line 176 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 176 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $22 = ((($this)) + 164|0); //@line 177 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $23 = +HEAPF32[$22>>2]; //@line 177 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $24 = ((($this)) + 168|0); //@line 177 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $25 = +HEAPF32[$24>>2]; //@line 177 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $26 = ((($this)) + 68|0); //@line 179 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $27 = +HEAPF32[$26>>2]; //@line 179 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $28 = $27 > 0.0; //@line 179 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $29 = ((($this)) + 204|0); //@line 183 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $30 = +HEAPF32[$29>>2]; //@line 183 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 if ($28) {
  $31 = $17 - $9; //@line 181 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $32 = ((($this)) + 76|0); //@line 183 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $33 = +HEAPF32[$32>>2]; //@line 183 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $34 = $31 + $33; //@line 183 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $35 = ((($this)) + 100|0); //@line 183 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $36 = +HEAPF32[$35>>2]; //@line 183 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $37 = ((($this)) + 112|0); //@line 183 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $38 = +HEAPF32[$37>>2]; //@line 183 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $39 = $36 * $38; //@line 183 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $40 = $34 + $39; //@line 183 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $41 = $30 * $40; //@line 183 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $42 = -$41; //@line 183 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $43 = $38 - $41; //@line 184 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$37>>2] = $43; //@line 184 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $44 = $23 * $42; //@line 186 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $45 = $9 - $44; //@line 186 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $46 = $25 * $42; //@line 187 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $47 = $17 + $46; //@line 187 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $48 = ((($this)) + 136|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $49 = +HEAPF32[$48>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $50 = $49 * $47; //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $51 = ((($this)) + 132|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $52 = +HEAPF32[$51>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $53 = $52 * $47; //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $54 = (HEAP32[tempDoublePtr>>2]=$13,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $55 = $54 - $50; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $56 = (HEAP32[tempDoublePtr>>2]=$15,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $57 = $56 + $53; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $58 = (HEAP32[tempDoublePtr>>2]=$5,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $59 = $55 - $58; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $60 = (HEAP32[tempDoublePtr>>2]=$7,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $61 = $57 - $60; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $62 = ((($this)) + 128|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $63 = +HEAPF32[$62>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $64 = $63 * $45; //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $65 = ((($this)) + 124|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $66 = +HEAPF32[$65>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $67 = $66 * $45; //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $68 = $64 + $59; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $69 = $61 - $67; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $70 = ((($this)) + 172|0); //@line 536 "./libraries/Box2D/Common/b2Math.h"
  $71 = +HEAPF32[$70>>2]; //@line 536 "./libraries/Box2D/Common/b2Math.h"
  $72 = $71 * $68; //@line 536 "./libraries/Box2D/Common/b2Math.h"
  $73 = ((($this)) + 184|0); //@line 536 "./libraries/Box2D/Common/b2Math.h"
  $74 = +HEAPF32[$73>>2]; //@line 536 "./libraries/Box2D/Common/b2Math.h"
  $75 = $74 * $69; //@line 536 "./libraries/Box2D/Common/b2Math.h"
  $76 = $72 + $75; //@line 536 "./libraries/Box2D/Common/b2Math.h"
  $77 = ((($this)) + 176|0); //@line 536 "./libraries/Box2D/Common/b2Math.h"
  $78 = +HEAPF32[$77>>2]; //@line 536 "./libraries/Box2D/Common/b2Math.h"
  $79 = $78 * $68; //@line 536 "./libraries/Box2D/Common/b2Math.h"
  $80 = ((($this)) + 188|0); //@line 536 "./libraries/Box2D/Common/b2Math.h"
  $81 = +HEAPF32[$80>>2]; //@line 536 "./libraries/Box2D/Common/b2Math.h"
  $82 = $69 * $81; //@line 536 "./libraries/Box2D/Common/b2Math.h"
  $83 = $79 + $82; //@line 536 "./libraries/Box2D/Common/b2Math.h"
  $84 = -$76; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $85 = -$83; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $86 = ((($this)) + 104|0); //@line 192 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $87 = +HEAPF32[$86>>2]; //@line 192 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $88 = $87 - $76; //@line 192 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$86>>2] = $88; //@line 192 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $89 = ((($this)) + 108|0); //@line 193 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $90 = +HEAPF32[$89>>2]; //@line 193 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $91 = $90 - $83; //@line 193 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$89>>2] = $91; //@line 193 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $92 = $19 * $84; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $93 = $19 * $85; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $94 = $58 - $92; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $95 = $60 - $93; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $96 = $66 * $85; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $97 = $63 * $84; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $98 = $96 - $97; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $99 = $23 * $98; //@line 198 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $100 = $45 - $99; //@line 198 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $101 = $21 * $84; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $102 = $21 * $85; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $103 = $54 + $101; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $104 = $56 + $102; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $105 = $52 * $85; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $106 = $49 * $84; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $107 = $105 - $106; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $108 = $25 * $107; //@line 201 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $109 = $47 + $108; //@line 201 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $vA$sroa$0$0$in = $94;$vA$sroa$10$0$in = $95;$vB$sroa$0$0$in = $103;$vB$sroa$10$0$in = $104;$wA$0 = $100;$wB$0 = $109;
  $195 = (($3) + (($1*12)|0)|0); //@line 221 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$195>>2] = $vA$sroa$0$0$in; //@line 221 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$6>>2] = $vA$sroa$10$0$in; //@line 221 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $196 = HEAP32[$0>>2]|0; //@line 222 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $197 = HEAP32[$2>>2]|0; //@line 222 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $198 = (((($197) + (($196*12)|0)|0)) + 8|0); //@line 222 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$198>>2] = $wA$0; //@line 222 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $199 = HEAP32[$10>>2]|0; //@line 223 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $200 = (($197) + (($199*12)|0)|0); //@line 223 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$200>>2] = $vB$sroa$0$0$in; //@line 223 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $201 = (((($197) + (($199*12)|0)|0)) + 4|0); //@line 223 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$201>>2] = $vB$sroa$10$0$in; //@line 223 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $202 = HEAP32[$10>>2]|0; //@line 224 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $203 = HEAP32[$2>>2]|0; //@line 224 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $204 = (((($203) + (($202*12)|0)|0)) + 8|0); //@line 224 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$204>>2] = $wB$0; //@line 224 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  return; //@line 225 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 } else {
  $110 = ((($this)) + 136|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $111 = +HEAPF32[$110>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $112 = $17 * $111; //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $113 = ((($this)) + 132|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $114 = +HEAPF32[$113>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $115 = $17 * $114; //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $116 = (HEAP32[tempDoublePtr>>2]=$13,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $117 = $116 - $112; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $118 = (HEAP32[tempDoublePtr>>2]=$15,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $119 = $118 + $115; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $120 = (HEAP32[tempDoublePtr>>2]=$5,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $121 = $117 - $120; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $122 = (HEAP32[tempDoublePtr>>2]=$7,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $123 = $119 - $122; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $124 = ((($this)) + 128|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $125 = +HEAPF32[$124>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $126 = $9 * $125; //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $127 = ((($this)) + 124|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $128 = +HEAPF32[$127>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $129 = $9 * $128; //@line 428 "./libraries/Box2D/Common/b2Math.h"
  $130 = $121 + $126; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $131 = $123 - $129; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $132 = $17 - $9; //@line 206 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $133 = ((($this)) + 172|0); //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $134 = +HEAPF32[$133>>2]; //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $135 = $130 * $134; //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $136 = ((($this)) + 176|0); //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $137 = +HEAPF32[$136>>2]; //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $138 = $130 * $137; //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $139 = ((($this)) + 180|0); //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $140 = +HEAPF32[$139>>2]; //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $141 = $130 * $140; //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $142 = ((($this)) + 184|0); //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $143 = +HEAPF32[$142>>2]; //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $144 = $131 * $143; //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $145 = ((($this)) + 188|0); //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $146 = +HEAPF32[$145>>2]; //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $147 = $131 * $146; //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $148 = ((($this)) + 192|0); //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $149 = +HEAPF32[$148>>2]; //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $150 = $131 * $149; //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $151 = $135 + $144; //@line 487 "./libraries/Box2D/Common/b2Math.h"
  $152 = $138 + $147; //@line 487 "./libraries/Box2D/Common/b2Math.h"
  $153 = $141 + $150; //@line 487 "./libraries/Box2D/Common/b2Math.h"
  $154 = ((($this)) + 196|0); //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $155 = +HEAPF32[$154>>2]; //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $156 = $132 * $155; //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $157 = ((($this)) + 200|0); //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $158 = +HEAPF32[$157>>2]; //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $159 = $132 * $158; //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $160 = $132 * $30; //@line 481 "./libraries/Box2D/Common/b2Math.h"
  $161 = $151 + $156; //@line 487 "./libraries/Box2D/Common/b2Math.h"
  $162 = $152 + $159; //@line 487 "./libraries/Box2D/Common/b2Math.h"
  $163 = $153 + $160; //@line 487 "./libraries/Box2D/Common/b2Math.h"
  $164 = -$161; //@line 159 "./libraries/Box2D/Common/b2Math.h"
  $165 = -$162; //@line 159 "./libraries/Box2D/Common/b2Math.h"
  $166 = ((($this)) + 104|0); //@line 164 "./libraries/Box2D/Common/b2Math.h"
  $167 = +HEAPF32[$166>>2]; //@line 164 "./libraries/Box2D/Common/b2Math.h"
  $168 = $167 - $161; //@line 164 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$166>>2] = $168; //@line 164 "./libraries/Box2D/Common/b2Math.h"
  $169 = ((($this)) + 108|0); //@line 164 "./libraries/Box2D/Common/b2Math.h"
  $170 = +HEAPF32[$169>>2]; //@line 164 "./libraries/Box2D/Common/b2Math.h"
  $171 = $170 - $162; //@line 164 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$169>>2] = $171; //@line 164 "./libraries/Box2D/Common/b2Math.h"
  $172 = ((($this)) + 112|0); //@line 164 "./libraries/Box2D/Common/b2Math.h"
  $173 = +HEAPF32[$172>>2]; //@line 164 "./libraries/Box2D/Common/b2Math.h"
  $174 = $173 - $163; //@line 164 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$172>>2] = $174; //@line 164 "./libraries/Box2D/Common/b2Math.h"
  $175 = $19 * $164; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $176 = $19 * $165; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $177 = $120 - $175; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $178 = $122 - $176; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $179 = $128 * $165; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $180 = $125 * $164; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $181 = $179 - $180; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $182 = $181 - $163; //@line 215 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $183 = $23 * $182; //@line 215 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $184 = $9 - $183; //@line 215 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $185 = $21 * $164; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $186 = $21 * $165; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $187 = $116 + $185; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $188 = $118 + $186; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $189 = $114 * $165; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $190 = $111 * $164; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $191 = $189 - $190; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $192 = $191 - $163; //@line 218 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $193 = $25 * $192; //@line 218 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $194 = $17 + $193; //@line 218 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $vA$sroa$0$0$in = $177;$vA$sroa$10$0$in = $178;$vB$sroa$0$0$in = $187;$vB$sroa$10$0$in = $188;$wA$0 = $184;$wB$0 = $194;
  $195 = (($3) + (($1*12)|0)|0); //@line 221 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$195>>2] = $vA$sroa$0$0$in; //@line 221 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$6>>2] = $vA$sroa$10$0$in; //@line 221 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $196 = HEAP32[$0>>2]|0; //@line 222 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $197 = HEAP32[$2>>2]|0; //@line 222 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $198 = (((($197) + (($196*12)|0)|0)) + 8|0); //@line 222 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$198>>2] = $wA$0; //@line 222 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $199 = HEAP32[$10>>2]|0; //@line 223 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $200 = (($197) + (($199*12)|0)|0); //@line 223 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$200>>2] = $vB$sroa$0$0$in; //@line 223 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $201 = (((($197) + (($199*12)|0)|0)) + 4|0); //@line 223 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$201>>2] = $vB$sroa$10$0$in; //@line 223 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $202 = HEAP32[$10>>2]|0; //@line 224 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $203 = HEAP32[$2>>2]|0; //@line 224 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $204 = (((($203) + (($202*12)|0)|0)) + 8|0); //@line 224 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$204>>2] = $wB$0; //@line 224 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  return; //@line 225 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 }
}
function __ZN11b2WeldJoint24SolvePositionConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$pn = 0.0, $$pn2 = 0.0, $$pn25 = 0.0, $$pn26 = 0.0, $$pn3 = 0.0, $$pn4 = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0;
 var $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0.0, $120 = 0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0;
 var $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0.0, $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0.0;
 var $147 = 0, $148 = 0.0, $149 = 0, $15 = 0, $150 = 0.0, $151 = 0.0, $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0, $157 = 0.0, $158 = 0.0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0.0, $163 = 0, $164 = 0.0;
 var $165 = 0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0, $17 = 0, $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0.0, $182 = 0.0;
 var $183 = 0.0, $184 = 0.0, $185 = 0.0, $186 = 0.0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0.0;
 var $200 = 0, $201 = 0, $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0.0, $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0;
 var $38 = 0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0, $55 = 0.0;
 var $56 = 0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0;
 var $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0;
 var $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0, $96 = 0, $97 = 0, $98 = 0.0, $99 = 0, $C = 0, $C1 = 0, $C11 = 0, $K = 0, $aA$0 = 0.0, $aB$0 = 0.0, $angularError$0 = 0, $cA$sroa$0$0$in = 0.0, $cA$sroa$10$0$in = 0.0, $cB$sroa$0$0$in = 0.0, $cB$sroa$10$0$in = 0.0, $impulse$sroa$0$0 = 0.0;
 var $impulse$sroa$5$0 = 0.0, $impulse$sroa$8$0 = 0.0, $phitmp = 0, $positionError$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $K = sp + 56|0;
 $C1 = sp + 48|0;
 $0 = sp + 40|0;
 $C11 = sp + 32|0;
 $C = sp + 20|0;
 $1 = sp + 8|0;
 $2 = sp;
 $3 = ((($this)) + 116|0); //@line 229 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 229 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $5 = ((($data)) + 24|0); //@line 229 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 229 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $7 = (($6) + (($4*12)|0)|0); //@line 229 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 229 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $9 = (((($6) + (($4*12)|0)|0)) + 4|0); //@line 229 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 229 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $11 = (((($6) + (($4*12)|0)|0)) + 8|0); //@line 230 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $12 = +HEAPF32[$11>>2]; //@line 230 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $13 = ((($this)) + 120|0); //@line 231 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 231 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $15 = (($6) + (($14*12)|0)|0); //@line 231 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 231 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $17 = (((($6) + (($14*12)|0)|0)) + 4|0); //@line 231 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 231 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $19 = (((($6) + (($14*12)|0)|0)) + 8|0); //@line 232 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $20 = +HEAPF32[$19>>2]; //@line 232 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $21 = (+Math_sin((+$12))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $22 = (+Math_cos((+$12))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $23 = (+Math_sin((+$20))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $24 = (+Math_cos((+$20))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $25 = ((($this)) + 156|0); //@line 236 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $26 = +HEAPF32[$25>>2]; //@line 236 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $27 = ((($this)) + 160|0); //@line 236 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $28 = +HEAPF32[$27>>2]; //@line 236 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $29 = ((($this)) + 164|0); //@line 237 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $30 = +HEAPF32[$29>>2]; //@line 237 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $31 = ((($this)) + 168|0); //@line 237 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $32 = +HEAPF32[$31>>2]; //@line 237 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $33 = ((($this)) + 80|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $34 = +HEAPF32[$33>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $35 = ((($this)) + 140|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $36 = +HEAPF32[$35>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $37 = $34 - $36; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $38 = ((($this)) + 84|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $39 = +HEAPF32[$38>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $40 = ((($this)) + 144|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $41 = +HEAPF32[$40>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $42 = $39 - $41; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $43 = $22 * $37; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $44 = $21 * $42; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $45 = $43 - $44; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $46 = $21 * $37; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $47 = $22 * $42; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $48 = $46 + $47; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $49 = ((($this)) + 88|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $50 = +HEAPF32[$49>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $51 = ((($this)) + 148|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $52 = +HEAPF32[$51>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $53 = $50 - $52; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $54 = ((($this)) + 92|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $55 = +HEAPF32[$54>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $56 = ((($this)) + 152|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $57 = +HEAPF32[$56>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $58 = $55 - $57; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $59 = $24 * $53; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $60 = $23 * $58; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $61 = $59 - $60; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $62 = $23 * $53; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $63 = $24 * $58; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $64 = $62 + $63; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $65 = $26 + $28; //@line 245 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $66 = $48 * $48; //@line 245 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $67 = $30 * $66; //@line 245 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $68 = $65 + $67; //@line 245 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $69 = $64 * $64; //@line 245 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $70 = $32 * $69; //@line 245 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $71 = $68 + $70; //@line 245 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$K>>2] = $71; //@line 245 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $72 = $48 * $45; //@line 246 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $73 = $30 * $72; //@line 246 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $74 = -$73; //@line 246 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $75 = $64 * $61; //@line 246 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $76 = $32 * $75; //@line 246 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $77 = $74 - $76; //@line 246 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $78 = ((($K)) + 12|0); //@line 246 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$78>>2] = $77; //@line 246 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $79 = $30 * $48; //@line 247 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $80 = -$79; //@line 247 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $81 = $32 * $64; //@line 247 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $82 = $80 - $81; //@line 247 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $83 = ((($K)) + 24|0); //@line 247 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$83>>2] = $82; //@line 247 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $84 = ((($K)) + 4|0); //@line 248 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$84>>2] = $77; //@line 248 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $85 = $45 * $45; //@line 249 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $86 = $30 * $85; //@line 249 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $87 = $65 + $86; //@line 249 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $88 = $61 * $61; //@line 249 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $89 = $32 * $88; //@line 249 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $90 = $87 + $89; //@line 249 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $91 = ((($K)) + 16|0); //@line 249 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$91>>2] = $90; //@line 249 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $92 = $30 * $45; //@line 250 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $93 = $32 * $61; //@line 250 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $94 = $92 + $93; //@line 250 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $95 = ((($K)) + 28|0); //@line 250 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$95>>2] = $94; //@line 250 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $96 = ((($K)) + 8|0); //@line 251 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$96>>2] = $82; //@line 251 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $97 = ((($K)) + 20|0); //@line 252 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$97>>2] = $94; //@line 252 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $98 = $30 + $32; //@line 281 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $99 = ((($K)) + 32|0); //@line 253 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$99>>2] = $98; //@line 253 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $100 = ((($this)) + 68|0); //@line 255 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $101 = +HEAPF32[$100>>2]; //@line 255 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $102 = $101 > 0.0; //@line 255 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 if ($102) {
  $103 = (HEAP32[tempDoublePtr>>2]=$16,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $104 = $103 + $61; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $105 = (HEAP32[tempDoublePtr>>2]=$18,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $106 = $105 + $64; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $107 = (HEAP32[tempDoublePtr>>2]=$8,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $108 = $104 - $107; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $109 = (HEAP32[tempDoublePtr>>2]=$10,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $110 = $106 - $109; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $111 = $108 - $45; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $112 = $110 - $48; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$C1>>2] = $111; //@line 59 "./libraries/Box2D/Common/b2Math.h"
  $113 = ((($C1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$113>>2] = $112; //@line 59 "./libraries/Box2D/Common/b2Math.h"
  $114 = $111 * $111; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $115 = $112 * $112; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $116 = $114 + $115; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $117 = (+Math_sqrt((+$116))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
  __ZNK7b2Mat337Solve22ERK6b2Vec2($0,$K,$C1); //@line 262 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $118 = +HEAPF32[$0>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $119 = -$118; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $120 = ((($0)) + 4|0); //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $121 = +HEAPF32[$120>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $122 = -$121; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $123 = $26 * $119; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $124 = $26 * $122; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $125 = $107 - $123; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $126 = $109 - $124; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $127 = $45 * $122; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $128 = $48 * $119; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $129 = $127 - $128; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $130 = $28 * $119; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $131 = $28 * $122; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $132 = $103 + $130; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $133 = $105 + $131; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $134 = $61 * $122; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $135 = $64 * $119; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $136 = $134 - $135; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $$pn3 = $136;$$pn4 = $129;$angularError$0 = 1;$cA$sroa$0$0$in = $125;$cA$sroa$10$0$in = $126;$cB$sroa$0$0$in = $132;$cB$sroa$10$0$in = $133;$positionError$0 = $117;
  $$pn2 = $30 * $$pn4; //@line 265 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $$pn = $32 * $$pn3; //@line 268 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $aA$0 = $12 - $$pn2; //@line 265 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $aB$0 = $20 + $$pn; //@line 268 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $187 = HEAP32[$3>>2]|0; //@line 300 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $188 = HEAP32[$5>>2]|0; //@line 300 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $189 = (($188) + (($187*12)|0)|0); //@line 300 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$189>>2] = $cA$sroa$0$0$in; //@line 300 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $190 = (((($188) + (($187*12)|0)|0)) + 4|0); //@line 300 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$190>>2] = $cA$sroa$10$0$in; //@line 300 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $191 = HEAP32[$3>>2]|0; //@line 301 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $192 = HEAP32[$5>>2]|0; //@line 301 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $193 = (((($192) + (($191*12)|0)|0)) + 8|0); //@line 301 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$193>>2] = $aA$0; //@line 301 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $194 = HEAP32[$13>>2]|0; //@line 302 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $195 = (($192) + (($194*12)|0)|0); //@line 302 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$195>>2] = $cB$sroa$0$0$in; //@line 302 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $196 = (((($192) + (($194*12)|0)|0)) + 4|0); //@line 302 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$196>>2] = $cB$sroa$10$0$in; //@line 302 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $197 = HEAP32[$13>>2]|0; //@line 303 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $198 = HEAP32[$5>>2]|0; //@line 303 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $199 = (((($198) + (($197*12)|0)|0)) + 8|0); //@line 303 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  HEAPF32[$199>>2] = $aB$0; //@line 303 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $200 = $positionError$0 <= 0.004999999888241291; //@line 305 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $201 = $200 & $angularError$0; //@line 305 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  STACKTOP = sp;return ($201|0); //@line 305 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 }
 $137 = (HEAP32[tempDoublePtr>>2]=$16,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $138 = $137 + $61; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $139 = (HEAP32[tempDoublePtr>>2]=$18,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $140 = $139 + $64; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $141 = (HEAP32[tempDoublePtr>>2]=$8,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $142 = $138 - $141; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $143 = (HEAP32[tempDoublePtr>>2]=$10,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $144 = $140 - $143; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $145 = $142 - $45; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $146 = $144 - $48; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$C11>>2] = $145; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $147 = ((($C11)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$147>>2] = $146; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $148 = $20 - $12; //@line 273 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $149 = ((($this)) + 96|0); //@line 273 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $150 = +HEAPF32[$149>>2]; //@line 273 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $151 = $148 - $150; //@line 273 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $152 = $145 * $145; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $153 = $146 * $146; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $154 = $153 + $152; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $155 = (+Math_sqrt((+$154))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $156 = $151 > 0.0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $157 = -$151; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $158 = $156 ? $151 : $157; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$C>>2] = $145; //@line 150 "./libraries/Box2D/Common/b2Math.h"
 $159 = ((($C)) + 4|0); //@line 150 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$159>>2] = $146; //@line 150 "./libraries/Box2D/Common/b2Math.h"
 $160 = ((($C)) + 8|0); //@line 150 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$160>>2] = $151; //@line 150 "./libraries/Box2D/Common/b2Math.h"
 $161 = $98 > 0.0; //@line 281 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 if ($161) {
  __ZNK7b2Mat337Solve33ERK6b2Vec3($1,$K,$C); //@line 283 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $162 = +HEAPF32[$1>>2]; //@line 159 "./libraries/Box2D/Common/b2Math.h"
  $163 = ((($1)) + 4|0); //@line 159 "./libraries/Box2D/Common/b2Math.h"
  $164 = +HEAPF32[$163>>2]; //@line 159 "./libraries/Box2D/Common/b2Math.h"
  $165 = ((($1)) + 8|0); //@line 159 "./libraries/Box2D/Common/b2Math.h"
  $166 = +HEAPF32[$165>>2]; //@line 159 "./libraries/Box2D/Common/b2Math.h"
  $167 = -$166; //@line 159 "./libraries/Box2D/Common/b2Math.h"
  $$pn25 = $162;$$pn26 = $164;$impulse$sroa$8$0 = $167;
 } else {
  __ZNK7b2Mat337Solve22ERK6b2Vec2($2,$K,$C11); //@line 287 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
  $168 = +HEAPF32[$2>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $169 = ((($2)) + 4|0); //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $170 = +HEAPF32[$169>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $$pn25 = $168;$$pn26 = $170;$impulse$sroa$8$0 = 0.0;
 }
 $impulse$sroa$5$0 = -$$pn26; //@line 159 "./libraries/Box2D/Common/b2Math.h"
 $impulse$sroa$0$0 = -$$pn25; //@line 159 "./libraries/Box2D/Common/b2Math.h"
 $171 = $26 * $impulse$sroa$0$0; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $172 = $26 * $impulse$sroa$5$0; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $173 = $141 - $171; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $174 = $143 - $172; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $175 = $45 * $impulse$sroa$5$0; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $176 = $48 * $impulse$sroa$0$0; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $177 = $175 - $176; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $178 = $impulse$sroa$8$0 + $177; //@line 294 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $179 = $28 * $impulse$sroa$0$0; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $180 = $28 * $impulse$sroa$5$0; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $181 = $137 + $179; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $182 = $139 + $180; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $183 = $61 * $impulse$sroa$5$0; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $184 = $64 * $impulse$sroa$0$0; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $185 = $183 - $184; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $186 = $impulse$sroa$8$0 + $185; //@line 297 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $phitmp = $158 <= 0.034906588494777679;
 $$pn3 = $186;$$pn4 = $178;$angularError$0 = $phitmp;$cA$sroa$0$0$in = $173;$cA$sroa$10$0$in = $174;$cB$sroa$0$0$in = $181;$cB$sroa$10$0$in = $182;$positionError$0 = $155;
 $$pn2 = $30 * $$pn4; //@line 265 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $$pn = $32 * $$pn3; //@line 268 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $aA$0 = $12 - $$pn2; //@line 265 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $aB$0 = $20 + $$pn; //@line 268 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $187 = HEAP32[$3>>2]|0; //@line 300 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $188 = HEAP32[$5>>2]|0; //@line 300 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $189 = (($188) + (($187*12)|0)|0); //@line 300 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$189>>2] = $cA$sroa$0$0$in; //@line 300 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $190 = (((($188) + (($187*12)|0)|0)) + 4|0); //@line 300 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$190>>2] = $cA$sroa$10$0$in; //@line 300 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $191 = HEAP32[$3>>2]|0; //@line 301 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $192 = HEAP32[$5>>2]|0; //@line 301 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $193 = (((($192) + (($191*12)|0)|0)) + 8|0); //@line 301 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$193>>2] = $aA$0; //@line 301 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $194 = HEAP32[$13>>2]|0; //@line 302 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $195 = (($192) + (($194*12)|0)|0); //@line 302 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$195>>2] = $cB$sroa$0$0$in; //@line 302 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $196 = (((($192) + (($194*12)|0)|0)) + 4|0); //@line 302 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$196>>2] = $cB$sroa$10$0$in; //@line 302 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $197 = HEAP32[$13>>2]|0; //@line 303 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $198 = HEAP32[$5>>2]|0; //@line 303 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $199 = (((($198) + (($197*12)|0)|0)) + 8|0); //@line 303 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$199>>2] = $aB$0; //@line 303 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $200 = $positionError$0 <= 0.004999999888241291; //@line 305 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $201 = $200 & $angularError$0; //@line 305 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 STACKTOP = sp;return ($201|0); //@line 305 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
}
function __ZN14b2WeldJointDef10InitializeEP6b2BodyS1_RK6b2Vec2($this,$bA,$bB,$anchor) {
 $this = $this|0;
 $bA = $bA|0;
 $bB = $bB|0;
 $anchor = $anchor|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0;
 var $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0.0, $40 = 0.0, $41 = 0, $42 = 0, $43 = 0, $44 = 0.0;
 var $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 39 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$0>>2] = $bA; //@line 39 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $1 = ((($this)) + 12|0); //@line 40 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$1>>2] = $bB; //@line 40 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $2 = +HEAPF32[$anchor>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($bA)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$3>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $5 = $2 - $4; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($anchor)) + 4|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($bA)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $9 = +HEAPF32[$8>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $10 = $7 - $9; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($bA)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$11>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $13 = $5 * $12; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $14 = ((($bA)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $15 = +HEAPF32[$14>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $16 = $10 * $15; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $17 = $13 + $16; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $18 = $5 * $15; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $19 = $12 * $10; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $20 = $19 - $18; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($this)) + 20|0); //@line 41 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$21>>2] = $17; //@line 41 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $22 = ((($this)) + 24|0); //@line 41 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$22>>2] = $20; //@line 41 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $23 = +HEAPF32[$anchor>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($bB)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $26 = $23 - $25; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $27 = +HEAPF32[$6>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $28 = ((($bB)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $29 = +HEAPF32[$28>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $30 = $27 - $29; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $31 = ((($bB)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $32 = +HEAPF32[$31>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $33 = $26 * $32; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $34 = ((($bB)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $35 = +HEAPF32[$34>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $36 = $30 * $35; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $37 = $33 + $36; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $38 = $26 * $35; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $39 = $32 * $30; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $40 = $39 - $38; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $41 = ((($this)) + 28|0); //@line 42 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$41>>2] = $37; //@line 42 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $42 = ((($this)) + 32|0); //@line 42 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$42>>2] = $40; //@line 42 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $43 = ((($bB)) + 56|0); //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $44 = +HEAPF32[$43>>2]; //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $45 = ((($bA)) + 56|0); //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $46 = +HEAPF32[$45>>2]; //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $47 = $44 - $46; //@line 43 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $48 = ((($this)) + 36|0); //@line 43 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAPF32[$48>>2] = $47; //@line 43 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 return; //@line 44 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
}
function __ZN11b2WeldJointC2EPK14b2WeldJointDef($this,$def) {
 $this = $this|0;
 $def = $def|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN7b2JointC2EPK10b2JointDef($this,$def); //@line 47 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$this>>2] = (21076); //@line 48 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $0 = ((($this)) + 80|0); //@line 46 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $1 = ((($this)) + 88|0); //@line 46 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $2 = ((($def)) + 20|0); //@line 49 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $3 = $2; //@line 49 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $4 = $3; //@line 49 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 49 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $6 = (($3) + 4)|0; //@line 49 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $7 = $6; //@line 49 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 49 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $9 = $0; //@line 49 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $10 = $9; //@line 49 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$10>>2] = $5; //@line 49 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $11 = (($9) + 4)|0; //@line 49 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $12 = $11; //@line 49 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$12>>2] = $8; //@line 49 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $13 = ((($def)) + 28|0); //@line 50 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $14 = $13; //@line 50 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $15 = $14; //@line 50 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 50 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $17 = (($14) + 4)|0; //@line 50 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $18 = $17; //@line 50 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 50 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $20 = $1; //@line 50 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $21 = $20; //@line 50 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$21>>2] = $16; //@line 50 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $22 = (($20) + 4)|0; //@line 50 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $23 = $22; //@line 50 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$23>>2] = $19; //@line 50 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $24 = ((($def)) + 36|0); //@line 51 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $25 = HEAP32[$24>>2]|0; //@line 51 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $26 = ((($this)) + 96|0); //@line 51 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$26>>2] = $25; //@line 51 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $27 = ((($def)) + 40|0); //@line 52 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $28 = HEAP32[$27>>2]|0; //@line 52 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $29 = ((($this)) + 68|0); //@line 52 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$29>>2] = $28; //@line 52 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $30 = ((($def)) + 44|0); //@line 53 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 53 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $32 = ((($this)) + 72|0); //@line 53 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 HEAP32[$32>>2] = $31; //@line 53 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
 $33 = ((($this)) + 104|0); //@line 153 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$33>>2] = 0.0; //@line 153 "./libraries/Box2D/Common/b2Math.h"
 $34 = ((($this)) + 108|0); //@line 153 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$34>>2] = 0.0; //@line 153 "./libraries/Box2D/Common/b2Math.h"
 $35 = ((($this)) + 112|0); //@line 153 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$35>>2] = 0.0; //@line 153 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 56 "libraries/Box2D/Dynamics/Joints/b2WeldJoint.cpp"
}
function __ZNK12b2WheelJoint10GetAnchorAEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 331 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 331 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 76|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 80|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 331 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
}
function __ZNK12b2WheelJoint10GetAnchorBEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 336 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 336 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 84|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 88|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 336 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
}
function __ZNK12b2WheelJoint16GetReactionForceEf($agg$result,$this,$inv_dt) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0;
 var $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 108|0); //@line 341 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 341 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $2 = ((($this)) + 180|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $4 = $1 * $3; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($this)) + 184|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $7 = $1 * $6; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($this)) + 116|0); //@line 341 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 341 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $10 = ((($this)) + 172|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $11 = +HEAPF32[$10>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $12 = $9 * $11; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($this)) + 176|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $15 = $9 * $14; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $16 = $4 + $12; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $17 = $7 + $15; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $19 = $17 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $18; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $20 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$20>>2] = $19; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 341 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
}
function __ZNK12b2WheelJoint17GetReactionTorqueEf($this,$inv_dt) {
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 346 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 346 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $2 = $1 * $inv_dt; //@line 346 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 return (+$2); //@line 346 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
}
function __ZN12b2WheelJoint4DumpEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0;
 var $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer13 = 0, $vararg_buffer17 = 0, $vararg_buffer21 = 0, $vararg_buffer24 = 0, $vararg_buffer27 = 0, $vararg_buffer3 = 0, $vararg_buffer30 = 0, $vararg_buffer33 = 0, $vararg_buffer36 = 0, $vararg_buffer6 = 0, $vararg_buffer9 = 0, $vararg_ptr12 = 0;
 var $vararg_ptr16 = 0, $vararg_ptr20 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer36 = sp + 120|0;
 $vararg_buffer33 = sp + 112|0;
 $vararg_buffer30 = sp + 104|0;
 $vararg_buffer27 = sp + 96|0;
 $vararg_buffer24 = sp + 88|0;
 $vararg_buffer21 = sp + 80|0;
 $vararg_buffer17 = sp + 64|0;
 $vararg_buffer13 = sp + 48|0;
 $vararg_buffer9 = sp + 32|0;
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($this)) + 48|0); //@line 403 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 403 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 403 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 403 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $4 = ((($this)) + 52|0); //@line 404 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 404 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $6 = ((($5)) + 8|0); //@line 404 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 404 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 __Z5b2LogPKcz(363493,$vararg_buffer); //@line 406 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$vararg_buffer1>>2] = $3; //@line 407 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 __Z5b2LogPKcz(447349,$vararg_buffer1); //@line 407 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$vararg_buffer3>>2] = $7; //@line 408 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 __Z5b2LogPKcz(447375,$vararg_buffer3); //@line 408 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $8 = ((($this)) + 61|0); //@line 409 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 409 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $10 = $9&255; //@line 409 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$vararg_buffer6>>2] = $10; //@line 409 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 __Z5b2LogPKcz(447401,$vararg_buffer6); //@line 409 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $11 = ((($this)) + 76|0); //@line 410 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $12 = +HEAPF32[$11>>2]; //@line 410 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $13 = $12; //@line 410 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $14 = ((($this)) + 80|0); //@line 410 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 410 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $16 = $15; //@line 410 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF64[$vararg_buffer9>>3] = $13; //@line 410 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $vararg_ptr12 = ((($vararg_buffer9)) + 8|0); //@line 410 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF64[$vararg_ptr12>>3] = $16; //@line 410 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 __Z5b2LogPKcz(447106,$vararg_buffer9); //@line 410 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $17 = ((($this)) + 84|0); //@line 411 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $18 = +HEAPF32[$17>>2]; //@line 411 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $19 = $18; //@line 411 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $20 = ((($this)) + 88|0); //@line 411 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 411 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $22 = $21; //@line 411 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF64[$vararg_buffer13>>3] = $19; //@line 411 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $vararg_ptr16 = ((($vararg_buffer13)) + 8|0); //@line 411 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF64[$vararg_ptr16>>3] = $22; //@line 411 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 __Z5b2LogPKcz(447148,$vararg_buffer13); //@line 411 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $23 = ((($this)) + 92|0); //@line 412 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $24 = +HEAPF32[$23>>2]; //@line 412 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $25 = $24; //@line 412 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $26 = ((($this)) + 96|0); //@line 412 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $27 = +HEAPF32[$26>>2]; //@line 412 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $28 = $27; //@line 412 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF64[$vararg_buffer17>>3] = $25; //@line 412 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $vararg_ptr20 = ((($vararg_buffer17)) + 8|0); //@line 412 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF64[$vararg_ptr20>>3] = $28; //@line 412 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 __Z5b2LogPKcz(363516,$vararg_buffer17); //@line 412 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $29 = ((($this)) + 128|0); //@line 413 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $30 = HEAP8[$29>>0]|0; //@line 413 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $31 = $30&255; //@line 413 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$vararg_buffer21>>2] = $31; //@line 413 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 __Z5b2LogPKcz(363556,$vararg_buffer21); //@line 413 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $32 = ((($this)) + 124|0); //@line 414 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $33 = +HEAPF32[$32>>2]; //@line 414 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $34 = $33; //@line 414 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF64[$vararg_buffer24>>3] = $34; //@line 414 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 __Z5b2LogPKcz(363586,$vararg_buffer24); //@line 414 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $35 = ((($this)) + 120|0); //@line 415 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $36 = +HEAPF32[$35>>2]; //@line 415 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $37 = $36; //@line 415 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF64[$vararg_buffer27>>3] = $37; //@line 415 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 __Z5b2LogPKcz(363614,$vararg_buffer27); //@line 415 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $38 = ((($this)) + 68|0); //@line 416 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $39 = +HEAPF32[$38>>2]; //@line 416 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $40 = $39; //@line 416 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF64[$vararg_buffer30>>3] = $40; //@line 416 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 __Z5b2LogPKcz(447003,$vararg_buffer30); //@line 416 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $41 = ((($this)) + 72|0); //@line 417 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $42 = +HEAPF32[$41>>2]; //@line 417 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $43 = $42; //@line 417 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF64[$vararg_buffer33>>3] = $43; //@line 417 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 __Z5b2LogPKcz(447032,$vararg_buffer33); //@line 417 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $44 = ((($this)) + 56|0); //@line 418 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $45 = HEAP32[$44>>2]|0; //@line 418 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$vararg_buffer36>>2] = $45; //@line 418 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 __Z5b2LogPKcz(447513,$vararg_buffer36); //@line 418 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 STACKTOP = sp;return; //@line 419 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
}
function __ZN12b2WheelJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 78 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
 return; //@line 78 "./libraries/Box2D/Dynamics/Joints/b2WheelJoint.h"
}
function __ZN12b2WheelJoint23InitVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0, $114 = 0.0, $115 = 0.0;
 var $116 = 0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0, $124 = 0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0, $131 = 0.0, $132 = 0.0, $133 = 0.0;
 var $134 = 0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0, $143 = 0, $144 = 0.0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0.0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0.0, $153 = 0.0, $154 = 0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0, $162 = 0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0, $167 = 0.0, $168 = 0.0, $169 = 0.0, $17 = 0;
 var $170 = 0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0.0, $182 = 0.0, $183 = 0.0, $184 = 0, $185 = 0.0, $186 = 0.0, $187 = 0.0, $188 = 0.0;
 var $189 = 0.0, $19 = 0, $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0.0, $194 = 0, $195 = 0.0, $196 = 0.0, $197 = 0.0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0, $201 = 0.0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0.0, $207 = 0, $208 = 0, $209 = 0.0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0.0, $22 = 0, $220 = 0, $221 = 0.0, $222 = 0.0, $223 = 0;
 var $224 = 0.0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0, $229 = 0.0, $23 = 0, $230 = 0.0, $231 = 0, $232 = 0.0, $233 = 0.0, $234 = 0.0, $235 = 0.0, $236 = 0.0, $237 = 0, $238 = 0.0, $239 = 0.0, $24 = 0, $240 = 0.0, $241 = 0.0;
 var $242 = 0.0, $243 = 0, $244 = 0.0, $245 = 0.0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0, $254 = 0.0, $255 = 0.0, $256 = 0, $257 = 0.0, $258 = 0.0, $259 = 0.0, $26 = 0;
 var $260 = 0.0, $261 = 0.0, $262 = 0.0, $263 = 0.0, $264 = 0.0, $265 = 0, $266 = 0.0, $267 = 0.0, $268 = 0, $269 = 0.0, $27 = 0, $270 = 0.0, $271 = 0.0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0.0, $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0;
 var $56 = 0.0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0.0, $65 = 0, $66 = 0.0, $67 = 0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0.0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0;
 var $92 = 0.0, $93 = 0.0, $94 = 0, $95 = 0.0, $96 = 0.0, $97 = 0, $98 = 0.0, $99 = 0.0, $bitcast = 0.0, $bitcast31 = 0.0, $storemerge = 0.0, $storemerge6 = 0.0, $storemerge7 = 0.0, $vA$sroa$0$0 = 0, $vA$sroa$6$0 = 0, $vB$sroa$0$0 = 0, $vB$sroa$6$0 = 0, $wA$0 = 0.0, $wB$0 = 0.0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 79 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 79 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 79 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 79 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $4 = ((($this)) + 132|0); //@line 79 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$4>>2] = $3; //@line 79 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $5 = ((($this)) + 52|0); //@line 80 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 80 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $7 = ((($6)) + 8|0); //@line 80 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 80 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $9 = ((($this)) + 136|0); //@line 80 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$9>>2] = $8; //@line 80 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $10 = ((($this)) + 140|0); //@line 81 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $11 = ((($1)) + 28|0); //@line 81 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $12 = $11; //@line 81 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $13 = $12; //@line 81 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 81 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $15 = (($12) + 4)|0; //@line 81 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $16 = $15; //@line 81 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 81 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $18 = $10; //@line 81 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $19 = $18; //@line 81 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$19>>2] = $14; //@line 81 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $20 = (($18) + 4)|0; //@line 81 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $21 = $20; //@line 81 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$21>>2] = $17; //@line 81 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $22 = ((($this)) + 148|0); //@line 82 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $23 = ((($6)) + 28|0); //@line 82 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $24 = $23; //@line 82 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $25 = $24; //@line 82 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 82 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $27 = (($24) + 4)|0; //@line 82 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $28 = $27; //@line 82 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 82 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $30 = $22; //@line 82 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $31 = $30; //@line 82 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$31>>2] = $26; //@line 82 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $32 = (($30) + 4)|0; //@line 82 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $33 = $32; //@line 82 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$33>>2] = $29; //@line 82 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $34 = ((($1)) + 120|0); //@line 83 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 83 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $36 = ((($this)) + 156|0); //@line 83 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$36>>2] = $35; //@line 83 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $37 = ((($6)) + 120|0); //@line 84 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $38 = HEAP32[$37>>2]|0; //@line 84 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $39 = ((($this)) + 160|0); //@line 84 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$39>>2] = $38; //@line 84 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $40 = ((($1)) + 128|0); //@line 85 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $41 = HEAP32[$40>>2]|0; //@line 85 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $42 = ((($this)) + 164|0); //@line 85 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$42>>2] = $41; //@line 85 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $43 = ((($6)) + 128|0); //@line 86 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $44 = +HEAPF32[$43>>2]; //@line 86 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $45 = ((($this)) + 168|0); //@line 86 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$45>>2] = $44; //@line 86 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $46 = (HEAP32[tempDoublePtr>>2]=$35,+HEAPF32[tempDoublePtr>>2]); //@line 88 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $47 = (HEAP32[tempDoublePtr>>2]=$38,+HEAPF32[tempDoublePtr>>2]); //@line 88 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $48 = (HEAP32[tempDoublePtr>>2]=$41,+HEAPF32[tempDoublePtr>>2]); //@line 89 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $49 = ((($data)) + 24|0); //@line 91 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $50 = HEAP32[$49>>2]|0; //@line 91 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $51 = (($50) + (($3*12)|0)|0); //@line 91 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $52 = +HEAPF32[$51>>2]; //@line 91 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $53 = (((($50) + (($3*12)|0)|0)) + 4|0); //@line 91 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $54 = +HEAPF32[$53>>2]; //@line 91 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $55 = (((($50) + (($3*12)|0)|0)) + 8|0); //@line 92 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $56 = +HEAPF32[$55>>2]; //@line 92 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $57 = ((($data)) + 28|0); //@line 93 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $58 = HEAP32[$57>>2]|0; //@line 93 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $59 = (($58) + (($3*12)|0)|0); //@line 93 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $60 = HEAP32[$59>>2]|0; //@line 93 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $61 = (((($58) + (($3*12)|0)|0)) + 4|0); //@line 93 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $62 = HEAP32[$61>>2]|0; //@line 93 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $63 = (((($58) + (($3*12)|0)|0)) + 8|0); //@line 94 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $64 = +HEAPF32[$63>>2]; //@line 94 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $65 = (($50) + (($8*12)|0)|0); //@line 96 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $66 = +HEAPF32[$65>>2]; //@line 96 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $67 = (((($50) + (($8*12)|0)|0)) + 4|0); //@line 96 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $68 = +HEAPF32[$67>>2]; //@line 96 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $69 = (((($50) + (($8*12)|0)|0)) + 8|0); //@line 97 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $70 = +HEAPF32[$69>>2]; //@line 97 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $71 = (($58) + (($8*12)|0)|0); //@line 98 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $72 = HEAP32[$71>>2]|0; //@line 98 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $73 = (((($58) + (($8*12)|0)|0)) + 4|0); //@line 98 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $74 = HEAP32[$73>>2]|0; //@line 98 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $75 = (((($58) + (($8*12)|0)|0)) + 8|0); //@line 99 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $76 = +HEAPF32[$75>>2]; //@line 99 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $77 = (+Math_sin((+$56))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $78 = (+Math_cos((+$56))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $79 = (+Math_sin((+$70))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $80 = (+Math_cos((+$70))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $81 = ((($this)) + 76|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $82 = +HEAPF32[$81>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $bitcast = (HEAP32[tempDoublePtr>>2]=$14,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $83 = $82 - $bitcast; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $84 = ((($this)) + 80|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $85 = +HEAPF32[$84>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $86 = (HEAP32[tempDoublePtr>>2]=$17,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $87 = $85 - $86; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $88 = $78 * $83; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $89 = $77 * $87; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $90 = $88 - $89; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $91 = $77 * $83; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $92 = $78 * $87; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $93 = $91 + $92; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $94 = ((($this)) + 84|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $95 = +HEAPF32[$94>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $bitcast31 = (HEAP32[tempDoublePtr>>2]=$26,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $96 = $95 - $bitcast31; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $97 = ((($this)) + 88|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $98 = +HEAPF32[$97>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $99 = (HEAP32[tempDoublePtr>>2]=$29,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $100 = $98 - $99; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $101 = $80 * $96; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $102 = $79 * $100; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $103 = $101 - $102; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $104 = $79 * $96; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $105 = $80 * $100; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $106 = $104 + $105; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $107 = $66 + $103; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $108 = $68 + $106; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $109 = $107 - $52; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $110 = $108 - $54; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $111 = $109 - $90; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $112 = $110 - $93; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $113 = ((($this)) + 100|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $114 = +HEAPF32[$113>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $115 = $78 * $114; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $116 = ((($this)) + 104|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $117 = +HEAPF32[$116>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $118 = $77 * $117; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $119 = $115 - $118; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $120 = $77 * $114; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $121 = $78 * $117; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $122 = $120 + $121; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $123 = ((($this)) + 180|0); //@line 110 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$123>>2] = $119; //@line 110 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $124 = ((($this)) + 184|0); //@line 110 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$124>>2] = $122; //@line 110 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $125 = $90 + $111; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $126 = $93 + $112; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $127 = $122 * $125; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $128 = $119 * $126; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $129 = $127 - $128; //@line 194 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $130 = ((($this)) + 196|0); //@line 111 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$130>>2] = $129; //@line 111 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $131 = $103 * $122; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $132 = $106 * $119; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $133 = $131 - $132; //@line 195 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $134 = ((($this)) + 200|0); //@line 112 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$134>>2] = $133; //@line 112 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $135 = $46 + $47; //@line 114 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $136 = $48 * $129; //@line 114 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $137 = $129 * $136; //@line 114 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $138 = $135 + $137; //@line 114 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $139 = $44 * $133; //@line 114 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $140 = $133 * $139; //@line 114 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $141 = $140 + $138; //@line 114 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $142 = ((($this)) + 204|0); //@line 114 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $143 = $141 > 0.0; //@line 116 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $144 = 1.0 / $141; //@line 118 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $storemerge7 = $143 ? $144 : $141; //@line 116 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$142>>2] = $storemerge7; //@line 114 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $145 = ((($this)) + 212|0); //@line 123 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$145>>2] = 0.0; //@line 123 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $146 = ((($this)) + 216|0); //@line 124 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$146>>2] = 0.0; //@line 124 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $147 = ((($this)) + 220|0); //@line 125 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$147>>2] = 0.0; //@line 125 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $148 = ((($this)) + 68|0); //@line 126 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $149 = +HEAPF32[$148>>2]; //@line 141 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $150 = $149 > 0.0; //@line 126 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 if ($150) {
  $151 = ((($this)) + 92|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $152 = +HEAPF32[$151>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $153 = $78 * $152; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $154 = ((($this)) + 96|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $155 = +HEAPF32[$154>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $156 = $77 * $155; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $157 = $153 - $156; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $158 = $77 * $152; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $159 = $78 * $155; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $160 = $158 + $159; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $161 = ((($this)) + 172|0); //@line 128 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$161>>2] = $157; //@line 128 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $162 = ((($this)) + 176|0); //@line 128 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$162>>2] = $160; //@line 128 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $163 = $125 * $160; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $164 = $126 * $157; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $165 = $163 - $164; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $166 = ((($this)) + 188|0); //@line 129 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$166>>2] = $165; //@line 129 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $167 = $103 * $160; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $168 = $106 * $157; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $169 = $167 - $168; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $170 = ((($this)) + 192|0); //@line 130 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$170>>2] = $169; //@line 130 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $171 = $48 * $165; //@line 132 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $172 = $165 * $171; //@line 132 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $173 = $135 + $172; //@line 132 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $174 = $44 * $169; //@line 132 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $175 = $169 * $174; //@line 132 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $176 = $175 + $173; //@line 132 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $177 = $176 > 0.0; //@line 134 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  if ($177) {
   $178 = 1.0 / $176; //@line 136 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $179 = $111 * $157; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $180 = $112 * $160; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $181 = $179 + $180; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $182 = $149 * 6.2831854820251465; //@line 141 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $183 = $178 * 2.0; //@line 144 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $184 = ((($this)) + 72|0); //@line 144 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $185 = +HEAPF32[$184>>2]; //@line 144 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $186 = $183 * $185; //@line 144 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $187 = $182 * $186; //@line 144 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $188 = $178 * $182; //@line 147 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $189 = $182 * $188; //@line 147 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $190 = +HEAPF32[$data>>2]; //@line 150 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $191 = $190 * $189; //@line 151 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $192 = $187 + $191; //@line 151 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $193 = $190 * $192; //@line 151 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $194 = $193 > 0.0; //@line 152 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $195 = 1.0 / $193; //@line 154 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $storemerge6 = $194 ? $195 : $193; //@line 152 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   HEAPF32[$147>>2] = $storemerge6; //@line 151 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $196 = $181 * $190; //@line 157 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $197 = $196 * $189; //@line 157 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $198 = $197 * $storemerge6; //@line 157 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   HEAPF32[$146>>2] = $198; //@line 157 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $199 = $176 + $storemerge6; //@line 159 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $200 = $199 > 0.0; //@line 160 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $201 = 1.0 / $199; //@line 162 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   $storemerge = $200 ? $201 : $199; //@line 160 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   HEAPF32[$145>>2] = $storemerge; //@line 159 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  }
 } else {
  $202 = ((($this)) + 116|0); //@line 168 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$202>>2] = 0.0; //@line 168 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 }
 $203 = ((($this)) + 128|0); //@line 172 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $204 = HEAP8[$203>>0]|0; //@line 172 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $205 = ($204<<24>>24)==(0); //@line 172 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 if ($205) {
  $210 = ((($this)) + 208|0); //@line 182 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$210>>2] = 0.0; //@line 182 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $211 = ((($this)) + 112|0); //@line 183 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$211>>2] = 0.0; //@line 183 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 } else {
  $206 = $44 + $48; //@line 174 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $207 = ((($this)) + 208|0); //@line 174 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$207>>2] = $206; //@line 174 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $208 = $206 > 0.0; //@line 175 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  if ($208) {
   $209 = 1.0 / $206; //@line 177 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
   HEAPF32[$207>>2] = $209; //@line 177 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  }
 }
 $212 = ((($data)) + 20|0); //@line 186 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $213 = HEAP8[$212>>0]|0; //@line 186 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $214 = ($213<<24>>24)==(0); //@line 186 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 if ($214) {
  $272 = ((($this)) + 108|0); //@line 205 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$272>>2] = 0.0; //@line 205 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $273 = ((($this)) + 116|0); //@line 206 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$273>>2] = 0.0; //@line 206 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $274 = ((($this)) + 112|0); //@line 207 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$274>>2] = 0.0; //@line 207 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $vA$sroa$0$0 = $60;$vA$sroa$6$0 = $62;$vB$sroa$0$0 = $72;$vB$sroa$6$0 = $74;$wA$0 = $64;$wB$0 = $76;
  $275 = HEAP32[$4>>2]|0; //@line 210 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $276 = HEAP32[$57>>2]|0; //@line 210 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $277 = (($276) + (($275*12)|0)|0); //@line 210 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAP32[$277>>2] = $vA$sroa$0$0; //@line 210 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $278 = (((($276) + (($275*12)|0)|0)) + 4|0); //@line 210 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAP32[$278>>2] = $vA$sroa$6$0; //@line 210 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $279 = HEAP32[$4>>2]|0; //@line 211 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $280 = HEAP32[$57>>2]|0; //@line 211 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $281 = (((($280) + (($279*12)|0)|0)) + 8|0); //@line 211 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$281>>2] = $wA$0; //@line 211 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $282 = HEAP32[$9>>2]|0; //@line 212 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $283 = (($280) + (($282*12)|0)|0); //@line 212 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAP32[$283>>2] = $vB$sroa$0$0; //@line 212 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $284 = (((($280) + (($282*12)|0)|0)) + 4|0); //@line 212 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAP32[$284>>2] = $vB$sroa$6$0; //@line 212 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $285 = HEAP32[$9>>2]|0; //@line 213 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $286 = HEAP32[$57>>2]|0; //@line 213 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $287 = (((($286) + (($285*12)|0)|0)) + 8|0); //@line 213 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$287>>2] = $wB$0; //@line 213 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  return; //@line 214 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 } else {
  $215 = ((($data)) + 8|0); //@line 189 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $216 = +HEAPF32[$215>>2]; //@line 189 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $217 = ((($this)) + 108|0); //@line 189 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $218 = +HEAPF32[$217>>2]; //@line 189 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $219 = $216 * $218; //@line 189 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$217>>2] = $219; //@line 189 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $220 = ((($this)) + 116|0); //@line 190 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $221 = +HEAPF32[$220>>2]; //@line 190 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $222 = $216 * $221; //@line 190 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$220>>2] = $222; //@line 190 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $223 = ((($this)) + 112|0); //@line 191 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $224 = +HEAPF32[$223>>2]; //@line 191 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $225 = $216 * $224; //@line 191 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$223>>2] = $225; //@line 191 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $226 = $219 * $119; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $227 = $219 * $122; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $228 = ((($this)) + 172|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $229 = +HEAPF32[$228>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $230 = $222 * $229; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $231 = ((($this)) + 176|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $232 = +HEAPF32[$231>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $233 = $222 * $232; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $234 = $226 + $230; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $235 = $227 + $233; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $236 = $219 * $129; //@line 194 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $237 = ((($this)) + 188|0); //@line 194 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $238 = +HEAPF32[$237>>2]; //@line 194 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $239 = $222 * $238; //@line 194 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $240 = $236 + $239; //@line 194 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $241 = $225 + $240; //@line 194 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $242 = $219 * $133; //@line 195 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $243 = ((($this)) + 192|0); //@line 195 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $244 = +HEAPF32[$243>>2]; //@line 195 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $245 = $222 * $244; //@line 195 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $246 = $242 + $245; //@line 195 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $247 = $225 + $246; //@line 195 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $248 = +HEAPF32[$36>>2]; //@line 197 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $249 = $234 * $248; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $250 = $235 * $248; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $251 = (HEAP32[tempDoublePtr>>2]=$60,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $252 = $251 - $249; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $253 = (HEAPF32[tempDoublePtr>>2]=$252,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $254 = (HEAP32[tempDoublePtr>>2]=$62,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $255 = $254 - $250; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $256 = (HEAPF32[tempDoublePtr>>2]=$255,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $257 = +HEAPF32[$42>>2]; //@line 198 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $258 = $241 * $257; //@line 198 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $259 = $64 - $258; //@line 198 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $260 = +HEAPF32[$39>>2]; //@line 200 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $261 = $234 * $260; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $262 = $235 * $260; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $263 = (HEAP32[tempDoublePtr>>2]=$72,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $264 = $263 + $261; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $265 = (HEAPF32[tempDoublePtr>>2]=$264,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $266 = (HEAP32[tempDoublePtr>>2]=$74,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $267 = $266 + $262; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $268 = (HEAPF32[tempDoublePtr>>2]=$267,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $269 = +HEAPF32[$45>>2]; //@line 201 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $270 = $247 * $269; //@line 201 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $271 = $76 + $270; //@line 201 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $vA$sroa$0$0 = $253;$vA$sroa$6$0 = $256;$vB$sroa$0$0 = $265;$vB$sroa$6$0 = $268;$wA$0 = $259;$wB$0 = $271;
  $275 = HEAP32[$4>>2]|0; //@line 210 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $276 = HEAP32[$57>>2]|0; //@line 210 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $277 = (($276) + (($275*12)|0)|0); //@line 210 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAP32[$277>>2] = $vA$sroa$0$0; //@line 210 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $278 = (((($276) + (($275*12)|0)|0)) + 4|0); //@line 210 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAP32[$278>>2] = $vA$sroa$6$0; //@line 210 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $279 = HEAP32[$4>>2]|0; //@line 211 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $280 = HEAP32[$57>>2]|0; //@line 211 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $281 = (((($280) + (($279*12)|0)|0)) + 8|0); //@line 211 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$281>>2] = $wA$0; //@line 211 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $282 = HEAP32[$9>>2]|0; //@line 212 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $283 = (($280) + (($282*12)|0)|0); //@line 212 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAP32[$283>>2] = $vB$sroa$0$0; //@line 212 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $284 = (((($280) + (($282*12)|0)|0)) + 4|0); //@line 212 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAP32[$284>>2] = $vB$sroa$6$0; //@line 212 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $285 = HEAP32[$9>>2]|0; //@line 213 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $286 = HEAP32[$57>>2]|0; //@line 213 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $287 = (((($286) + (($285*12)|0)|0)) + 8|0); //@line 213 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$287>>2] = $wB$0; //@line 213 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  return; //@line 214 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 }
}
function __ZN12b2WheelJoint24SolveVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0.0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0, $115 = 0.0;
 var $116 = 0.0, $117 = 0.0, $118 = 0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0.0, $130 = 0.0, $131 = 0.0, $132 = 0.0, $133 = 0.0;
 var $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0.0, $16 = 0, $17 = 0.0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0.0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0;
 var $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0.0, $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0, $49 = 0.0, $5 = 0.0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0.0;
 var $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0.0, $70 = 0.0, $71 = 0.0;
 var $72 = 0.0, $73 = 0.0, $74 = 0, $75 = 0.0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0, $89 = 0.0, $9 = 0;
 var $90 = 0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 156|0); //@line 218 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 218 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $2 = ((($this)) + 160|0); //@line 218 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 218 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $4 = ((($this)) + 164|0); //@line 219 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 219 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $6 = ((($this)) + 168|0); //@line 219 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 219 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $8 = ((($this)) + 132|0); //@line 221 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 221 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $10 = ((($data)) + 28|0); //@line 221 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 221 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $12 = (($11) + (($9*12)|0)|0); //@line 221 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $13 = +HEAPF32[$12>>2]; //@line 221 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $14 = (((($11) + (($9*12)|0)|0)) + 4|0); //@line 221 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 221 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $16 = (((($11) + (($9*12)|0)|0)) + 8|0); //@line 222 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 222 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $18 = ((($this)) + 136|0); //@line 223 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 223 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $20 = (($11) + (($19*12)|0)|0); //@line 223 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 223 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $22 = (((($11) + (($19*12)|0)|0)) + 4|0); //@line 223 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $23 = +HEAPF32[$22>>2]; //@line 223 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $24 = (((($11) + (($19*12)|0)|0)) + 8|0); //@line 224 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $25 = +HEAPF32[$24>>2]; //@line 224 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $26 = $21 - $13; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $27 = $23 - $15; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $28 = ((($this)) + 172|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $29 = +HEAPF32[$28>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $30 = $26 * $29; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $31 = ((($this)) + 176|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $32 = +HEAPF32[$31>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $33 = $27 * $32; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $34 = $30 + $33; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $35 = ((($this)) + 192|0); //@line 228 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $36 = +HEAPF32[$35>>2]; //@line 228 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $37 = $25 * $36; //@line 228 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $38 = $34 + $37; //@line 228 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $39 = ((($this)) + 188|0); //@line 228 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $40 = +HEAPF32[$39>>2]; //@line 228 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $41 = $17 * $40; //@line 228 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $42 = $38 - $41; //@line 228 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $43 = ((($this)) + 212|0); //@line 229 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $44 = +HEAPF32[$43>>2]; //@line 229 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $45 = ((($this)) + 216|0); //@line 229 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $46 = +HEAPF32[$45>>2]; //@line 229 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $47 = $42 + $46; //@line 229 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $48 = ((($this)) + 220|0); //@line 229 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $49 = +HEAPF32[$48>>2]; //@line 229 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $50 = ((($this)) + 116|0); //@line 229 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $51 = +HEAPF32[$50>>2]; //@line 229 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $52 = $49 * $51; //@line 229 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $53 = $47 + $52; //@line 229 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $54 = $44 * $53; //@line 229 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $55 = -$54; //@line 229 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $56 = $51 - $54; //@line 230 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$50>>2] = $56; //@line 230 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $57 = $29 * $55; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $58 = $32 * $55; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $59 = $40 * $55; //@line 233 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $60 = $36 * $55; //@line 234 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $61 = $1 * $57; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $62 = $1 * $58; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $63 = $13 - $61; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $64 = $15 - $62; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $65 = $5 * $59; //@line 237 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $66 = $17 - $65; //@line 237 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $67 = $3 * $57; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $68 = $3 * $58; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $69 = $21 + $67; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $70 = $23 + $68; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $71 = $7 * $60; //@line 240 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $72 = $25 + $71; //@line 240 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $73 = $72 - $66; //@line 245 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $74 = ((($this)) + 124|0); //@line 245 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $75 = +HEAPF32[$74>>2]; //@line 245 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $76 = $73 - $75; //@line 245 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $77 = ((($this)) + 208|0); //@line 246 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $78 = +HEAPF32[$77>>2]; //@line 246 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $79 = $78 * $76; //@line 246 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $80 = ((($this)) + 112|0); //@line 248 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $81 = +HEAPF32[$80>>2]; //@line 248 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $82 = +HEAPF32[$data>>2]; //@line 249 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $83 = ((($this)) + 120|0); //@line 249 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $84 = +HEAPF32[$83>>2]; //@line 249 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $85 = $82 * $84; //@line 249 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $86 = $81 - $79; //@line 250 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $87 = -$85; //@line 250 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $88 = $86 < $85; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $89 = $88 ? $86 : $85; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $90 = $89 < $87; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $91 = $90 ? $87 : $89; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$80>>2] = $91; //@line 250 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $92 = $91 - $81; //@line 251 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $93 = $5 * $92; //@line 253 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $94 = $66 - $93; //@line 253 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $95 = $7 * $92; //@line 254 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $96 = $72 + $95; //@line 254 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $97 = $69 - $63; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $98 = $70 - $64; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $99 = ((($this)) + 180|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $100 = +HEAPF32[$99>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $101 = $97 * $100; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $102 = ((($this)) + 184|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $103 = +HEAPF32[$102>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $104 = $98 * $103; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $105 = $101 + $104; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $106 = ((($this)) + 200|0); //@line 259 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $107 = +HEAPF32[$106>>2]; //@line 259 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $108 = $107 * $96; //@line 259 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $109 = $105 + $108; //@line 259 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $110 = ((($this)) + 196|0); //@line 259 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $111 = +HEAPF32[$110>>2]; //@line 259 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $112 = $111 * $94; //@line 259 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $113 = $109 - $112; //@line 259 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $114 = ((($this)) + 204|0); //@line 260 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $115 = +HEAPF32[$114>>2]; //@line 260 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $116 = $115 * $113; //@line 260 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $117 = -$116; //@line 260 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $118 = ((($this)) + 108|0); //@line 261 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $119 = +HEAPF32[$118>>2]; //@line 261 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $120 = $119 - $116; //@line 261 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$118>>2] = $120; //@line 261 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $121 = $100 * $117; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $122 = $103 * $117; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $123 = $111 * $117; //@line 264 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $124 = $107 * $117; //@line 265 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $125 = $1 * $121; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $126 = $1 * $122; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $127 = $63 - $125; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $128 = $64 - $126; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $129 = $5 * $123; //@line 268 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $130 = $94 - $129; //@line 268 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $131 = $3 * $121; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $132 = $3 * $122; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $133 = $69 + $131; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $134 = $70 + $132; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $135 = $7 * $124; //@line 271 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $136 = $96 + $135; //@line 271 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $137 = HEAP32[$8>>2]|0; //@line 274 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $138 = HEAP32[$10>>2]|0; //@line 274 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $139 = (($138) + (($137*12)|0)|0); //@line 274 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$139>>2] = $127; //@line 274 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $140 = (((($138) + (($137*12)|0)|0)) + 4|0); //@line 274 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$140>>2] = $128; //@line 274 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $141 = HEAP32[$8>>2]|0; //@line 275 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $142 = HEAP32[$10>>2]|0; //@line 275 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $143 = (((($142) + (($141*12)|0)|0)) + 8|0); //@line 275 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$143>>2] = $130; //@line 275 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $144 = HEAP32[$18>>2]|0; //@line 276 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $145 = (($142) + (($144*12)|0)|0); //@line 276 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$145>>2] = $133; //@line 276 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $146 = (((($142) + (($144*12)|0)|0)) + 4|0); //@line 276 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$146>>2] = $134; //@line 276 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $147 = HEAP32[$18>>2]|0; //@line 277 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $148 = HEAP32[$10>>2]|0; //@line 277 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $149 = (((($148) + (($147*12)|0)|0)) + 8|0); //@line 277 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$149>>2] = $136; //@line 277 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 return; //@line 278 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
}
function __ZN12b2WheelJoint24SolvePositionConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0;
 var $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0.0, $130 = 0, $131 = 0, $132 = 0.0, $133 = 0;
 var $14 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0, $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0;
 var $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0.0, $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0.0;
 var $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0.0, $63 = 0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0;
 var $69 = 0.0, $7 = 0.0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0.0, $86 = 0;
 var $87 = 0.0, $88 = 0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $impulse$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 132|0); //@line 282 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 282 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $2 = ((($data)) + 24|0); //@line 282 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 321 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $4 = (($3) + (($1*12)|0)|0); //@line 282 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 282 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $6 = (((($3) + (($1*12)|0)|0)) + 4|0); //@line 282 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 282 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $8 = (((($3) + (($1*12)|0)|0)) + 8|0); //@line 283 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 283 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $10 = ((($this)) + 136|0); //@line 284 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 284 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $12 = (($3) + (($11*12)|0)|0); //@line 284 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $13 = +HEAPF32[$12>>2]; //@line 284 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $14 = (((($3) + (($11*12)|0)|0)) + 4|0); //@line 284 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 284 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $16 = (((($3) + (($11*12)|0)|0)) + 8|0); //@line 285 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 285 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $18 = (+Math_sin((+$9))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $19 = (+Math_cos((+$9))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $20 = (+Math_sin((+$17))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $21 = (+Math_cos((+$17))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($this)) + 76|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $23 = +HEAPF32[$22>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($this)) + 140|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $26 = $23 - $25; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($this)) + 80|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $28 = +HEAPF32[$27>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $29 = ((($this)) + 144|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $30 = +HEAPF32[$29>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $31 = $28 - $30; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $32 = $19 * $26; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $33 = $18 * $31; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $34 = $32 - $33; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $35 = $18 * $26; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $36 = $19 * $31; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $37 = $35 + $36; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $38 = ((($this)) + 84|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $39 = +HEAPF32[$38>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $40 = ((($this)) + 148|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $41 = +HEAPF32[$40>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $42 = $39 - $41; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $43 = ((($this)) + 88|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $44 = +HEAPF32[$43>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $45 = ((($this)) + 152|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $46 = +HEAPF32[$45>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $47 = $44 - $46; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $48 = $21 * $42; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $49 = $20 * $47; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $50 = $48 - $49; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $51 = $20 * $42; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $52 = $21 * $47; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $53 = $51 + $52; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $54 = $13 - $5; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $55 = $15 - $7; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $56 = $54 + $50; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $57 = $55 + $53; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $58 = $56 - $34; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $59 = $57 - $37; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $60 = ((($this)) + 100|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $61 = +HEAPF32[$60>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $62 = $19 * $61; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $63 = ((($this)) + 104|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $64 = +HEAPF32[$63>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $65 = $18 * $64; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $66 = $62 - $65; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $67 = $18 * $61; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $68 = $19 * $64; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $69 = $67 + $68; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $70 = $34 + $58; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $71 = $37 + $59; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $72 = $69 * $70; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $73 = $66 * $71; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $74 = $72 - $73; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $75 = $50 * $69; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $76 = $53 * $66; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $77 = $75 - $76; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $78 = $66 * $58; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $79 = $69 * $59; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $80 = $78 + $79; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $81 = ((($this)) + 156|0); //@line 300 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $82 = +HEAPF32[$81>>2]; //@line 316 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $83 = ((($this)) + 160|0); //@line 300 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $84 = +HEAPF32[$83>>2]; //@line 318 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $85 = $82 + $84; //@line 300 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $86 = ((($this)) + 164|0); //@line 300 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $87 = +HEAPF32[$86>>2]; //@line 317 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $88 = ((($this)) + 196|0); //@line 300 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $89 = +HEAPF32[$88>>2]; //@line 300 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $90 = $87 * $89; //@line 300 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $91 = $89 * $90; //@line 300 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $92 = $85 + $91; //@line 300 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $93 = ((($this)) + 168|0); //@line 300 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $94 = +HEAPF32[$93>>2]; //@line 319 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $95 = ((($this)) + 200|0); //@line 300 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $96 = +HEAPF32[$95>>2]; //@line 300 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $97 = $94 * $96; //@line 300 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $98 = $96 * $97; //@line 300 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $99 = $92 + $98; //@line 300 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $100 = $99 != 0.0; //@line 303 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $101 = -$80; //@line 305 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $102 = $101 / $99; //@line 305 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $impulse$0 = $100 ? $102 : 0.0; //@line 303 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $103 = $66 * $impulse$0; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $104 = $69 * $impulse$0; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $105 = $74 * $impulse$0; //@line 313 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $106 = $77 * $impulse$0; //@line 314 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $107 = $103 * $82; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $108 = $104 * $82; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $109 = $5 - $107; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $110 = $7 - $108; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $111 = $105 * $87; //@line 317 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $112 = $9 - $111; //@line 317 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $113 = $103 * $84; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $114 = $104 * $84; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $115 = $13 + $113; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $116 = $15 + $114; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $117 = $106 * $94; //@line 319 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $118 = $17 + $117; //@line 319 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $119 = HEAP32[$0>>2]|0; //@line 321 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $120 = (($3) + (($119*12)|0)|0); //@line 321 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$120>>2] = $109; //@line 321 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $121 = (((($3) + (($119*12)|0)|0)) + 4|0); //@line 321 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$121>>2] = $110; //@line 321 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $122 = HEAP32[$0>>2]|0; //@line 322 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $123 = HEAP32[$2>>2]|0; //@line 322 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $124 = (((($123) + (($122*12)|0)|0)) + 8|0); //@line 322 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$124>>2] = $112; //@line 322 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $125 = HEAP32[$10>>2]|0; //@line 323 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $126 = (($123) + (($125*12)|0)|0); //@line 323 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$126>>2] = $115; //@line 323 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $127 = (((($123) + (($125*12)|0)|0)) + 4|0); //@line 323 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$127>>2] = $116; //@line 323 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $128 = HEAP32[$10>>2]|0; //@line 324 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $129 = HEAP32[$2>>2]|0; //@line 324 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $130 = (((($129) + (($128*12)|0)|0)) + 8|0); //@line 324 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$130>>2] = $118; //@line 324 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $131 = $80 > 0.0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $132 = $131 ? $80 : $101; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $133 = $132 <= 0.004999999888241291; //@line 326 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 return ($133|0); //@line 326 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
}
function __ZN15b2WheelJointDef10InitializeEP6b2BodyS1_RK6b2Vec2S4_($this,$bA,$bB,$anchor,$axis) {
 $this = $this|0;
 $bA = $bA|0;
 $bB = $bB|0;
 $anchor = $anchor|0;
 $axis = $axis|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0;
 var $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0.0, $40 = 0.0, $41 = 0, $42 = 0, $43 = 0.0, $44 = 0.0;
 var $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0, $55 = 0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 41 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$0>>2] = $bA; //@line 41 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $1 = ((($this)) + 12|0); //@line 42 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$1>>2] = $bB; //@line 42 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $2 = +HEAPF32[$anchor>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($bA)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$3>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $5 = $2 - $4; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($anchor)) + 4|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($bA)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $9 = +HEAPF32[$8>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $10 = $7 - $9; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($bA)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$11>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $13 = $5 * $12; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $14 = ((($bA)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $15 = +HEAPF32[$14>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $16 = $10 * $15; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $17 = $13 + $16; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $18 = $5 * $15; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $19 = $12 * $10; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $20 = $19 - $18; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($this)) + 20|0); //@line 43 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$21>>2] = $17; //@line 43 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $22 = ((($this)) + 24|0); //@line 43 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$22>>2] = $20; //@line 43 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $23 = +HEAPF32[$anchor>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($bB)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $26 = $23 - $25; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $27 = +HEAPF32[$6>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $28 = ((($bB)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $29 = +HEAPF32[$28>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $30 = $27 - $29; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $31 = ((($bB)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $32 = +HEAPF32[$31>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $33 = $26 * $32; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $34 = ((($bB)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $35 = +HEAPF32[$34>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $36 = $30 * $35; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $37 = $33 + $36; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $38 = $26 * $35; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $39 = $32 * $30; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $40 = $39 - $38; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $41 = ((($this)) + 28|0); //@line 44 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$41>>2] = $37; //@line 44 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $42 = ((($this)) + 32|0); //@line 44 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$42>>2] = $40; //@line 44 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $43 = +HEAPF32[$11>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $44 = +HEAPF32[$axis>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $45 = $43 * $44; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $46 = +HEAPF32[$14>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $47 = ((($axis)) + 4|0); //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $48 = +HEAPF32[$47>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $49 = $46 * $48; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $50 = $45 + $49; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $51 = $44 * $46; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $52 = $43 * $48; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $53 = $52 - $51; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $54 = ((($this)) + 36|0); //@line 45 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$54>>2] = $50; //@line 45 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $55 = ((($this)) + 40|0); //@line 45 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$55>>2] = $53; //@line 45 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 return; //@line 46 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
}
function __ZN12b2WheelJointC2EPK15b2WheelJointDef($this,$def) {
 $this = $this|0;
 $def = $def|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN7b2JointC2EPK10b2JointDef($this,$def); //@line 49 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$this>>2] = (21128); //@line 50 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $0 = ((($this)) + 76|0); //@line 48 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $1 = ((($this)) + 84|0); //@line 48 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $2 = ((($this)) + 92|0); //@line 48 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $3 = ((($def)) + 20|0); //@line 51 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $4 = $3; //@line 51 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $5 = $4; //@line 51 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 51 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $7 = (($4) + 4)|0; //@line 51 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $8 = $7; //@line 51 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 51 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $10 = $0; //@line 51 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $11 = $10; //@line 51 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$11>>2] = $6; //@line 51 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $12 = (($10) + 4)|0; //@line 51 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $13 = $12; //@line 51 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$13>>2] = $9; //@line 51 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $14 = ((($def)) + 28|0); //@line 52 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $15 = $14; //@line 52 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $16 = $15; //@line 52 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 52 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $18 = (($15) + 4)|0; //@line 52 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $19 = $18; //@line 52 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 52 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $21 = $1; //@line 52 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $22 = $21; //@line 52 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$22>>2] = $17; //@line 52 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $23 = (($21) + 4)|0; //@line 52 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $24 = $23; //@line 52 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$24>>2] = $20; //@line 52 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $25 = ((($def)) + 36|0); //@line 53 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $26 = $25; //@line 53 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $27 = $26; //@line 53 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $28 = HEAP32[$27>>2]|0; //@line 53 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $29 = (($26) + 4)|0; //@line 53 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $30 = $29; //@line 53 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 53 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $32 = $2; //@line 53 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $33 = $32; //@line 53 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$33>>2] = $28; //@line 53 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $34 = (($32) + 4)|0; //@line 53 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $35 = $34; //@line 53 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$35>>2] = $31; //@line 53 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $36 = (HEAP32[tempDoublePtr>>2]=$31,+HEAPF32[tempDoublePtr>>2]); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $37 = -$36; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $38 = ((($this)) + 100|0); //@line 54 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$38>>2] = $37; //@line 54 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $39 = ((($this)) + 104|0); //@line 54 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$39>>2] = $28; //@line 54 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $40 = ((($this)) + 204|0); //@line 56 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$40>>2] = 0.0; //@line 56 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $41 = ((($this)) + 108|0); //@line 57 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$41>>2] = 0.0; //@line 57 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $42 = ((($this)) + 208|0); //@line 58 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$42>>2] = 0.0; //@line 58 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $43 = ((($this)) + 112|0); //@line 59 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$43>>2] = 0.0; //@line 59 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $44 = ((($this)) + 212|0); //@line 60 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$44>>2] = 0.0; //@line 60 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $45 = ((($this)) + 116|0); //@line 61 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$45>>2] = 0.0; //@line 61 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $46 = ((($def)) + 48|0); //@line 63 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $47 = HEAP32[$46>>2]|0; //@line 63 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $48 = ((($this)) + 120|0); //@line 63 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$48>>2] = $47; //@line 63 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $49 = ((($def)) + 52|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $50 = HEAP32[$49>>2]|0; //@line 64 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $51 = ((($this)) + 124|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$51>>2] = $50; //@line 64 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $52 = ((($def)) + 44|0); //@line 65 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $53 = HEAP8[$52>>0]|0; //@line 65 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $54 = ((($this)) + 128|0); //@line 65 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP8[$54>>0] = $53; //@line 65 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $55 = ((($def)) + 56|0); //@line 67 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $56 = HEAP32[$55>>2]|0; //@line 67 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $57 = ((($this)) + 68|0); //@line 67 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$57>>2] = $56; //@line 67 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $58 = ((($def)) + 60|0); //@line 68 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $59 = HEAP32[$58>>2]|0; //@line 68 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $60 = ((($this)) + 72|0); //@line 68 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP32[$60>>2] = $59; //@line 68 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $61 = ((($this)) + 216|0); //@line 70 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$61>>2] = 0.0; //@line 70 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $62 = ((($this)) + 220|0); //@line 71 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$62>>2] = 0.0; //@line 71 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $63 = ((($this)) + 172|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 ;HEAP32[$63>>2]=0|0;HEAP32[$63+4>>2]=0|0;HEAP32[$63+8>>2]=0|0;HEAP32[$63+12>>2]=0|0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 75 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
}
function __ZNK12b2WheelJoint19GetJointTranslationEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0.0;
 var $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 351 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 351 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $2 = ((($this)) + 52|0); //@line 352 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 352 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $4 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($this)) + 76|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = $5 * $7; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($this)) + 80|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$11>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = $10 * $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = $8 - $13; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = +HEAPF32[$15>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $17 = $16 + $14; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $18 = $7 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = $5 * $12; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = $18 + $19; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $22 = +HEAPF32[$21>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $23 = $20 + $22; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($3)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $26 = ((($this)) + 84|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $27 = +HEAPF32[$26>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $28 = $25 * $27; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $29 = ((($3)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $30 = +HEAPF32[$29>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $31 = ((($this)) + 88|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $32 = +HEAPF32[$31>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $33 = $30 * $32; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $34 = $28 - $33; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $35 = ((($3)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $36 = +HEAPF32[$35>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $37 = $36 + $34; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $38 = $27 * $30; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $39 = $25 * $32; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $40 = $38 + $39; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $41 = ((($3)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $42 = +HEAPF32[$41>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $43 = $40 + $42; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $44 = $37 - $17; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $45 = $43 - $23; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $46 = ((($this)) + 92|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $47 = +HEAPF32[$46>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $48 = $5 * $47; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $49 = ((($this)) + 96|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $50 = +HEAPF32[$49>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $51 = $10 * $50; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $52 = $48 - $51; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $53 = $10 * $47; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $54 = $5 * $50; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $55 = $53 + $54; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $56 = $44 * $52; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $57 = $45 * $55; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $58 = $56 + $57; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 return (+$58); //@line 360 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
}
function __ZNK12b2WheelJoint13GetJointSpeedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 365 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 365 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $2 = ((($1)) + 72|0); //@line 365 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 365 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $4 = ((($this)) + 52|0); //@line 366 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 366 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $6 = ((($5)) + 72|0); //@line 366 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 366 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $8 = $7 - $3; //@line 367 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 return (+$8); //@line 367 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
}
function __ZNK12b2WheelJoint14IsMotorEnabledEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 128|0); //@line 372 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 372 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 372 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 return ($2|0); //@line 372 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
}
function __ZN12b2WheelJoint11EnableMotorEb($this,$flag) {
 $this = $this|0;
 $flag = $flag|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 377 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 377 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $2 = ((($1)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = $4 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $6 = ($5|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($6) {
  $7 = $4 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $8 = $7&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$2>>1] = $8; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $9 = ((($1)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$9>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $10 = ((($this)) + 52|0); //@line 378 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 378 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $12 = ((($11)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $13 = HEAP16[$12>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $14 = $13&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $15 = $14 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $16 = ($15|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($16)) {
  $20 = ((($this)) + 128|0); //@line 379 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  $21 = $flag&1; //@line 379 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAP8[$20>>0] = $21; //@line 379 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  return; //@line 380 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 }
 $17 = $14 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $18 = $17&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP16[$12>>1] = $18; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = ((($11)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$19>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = ((($this)) + 128|0); //@line 379 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $21 = $flag&1; //@line 379 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAP8[$20>>0] = $21; //@line 379 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 return; //@line 380 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
}
function __ZN12b2WheelJoint13SetMotorSpeedEf($this,$speed) {
 $this = $this|0;
 $speed = +$speed;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 384 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 384 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $2 = ((($1)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = $4 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $6 = ($5|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($6) {
  $7 = $4 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $8 = $7&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$2>>1] = $8; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $9 = ((($1)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$9>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $10 = ((($this)) + 52|0); //@line 385 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 385 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $12 = ((($11)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $13 = HEAP16[$12>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $14 = $13&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $15 = $14 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $16 = ($15|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($16)) {
  $20 = ((($this)) + 124|0); //@line 386 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$20>>2] = $speed; //@line 386 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  return; //@line 387 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 }
 $17 = $14 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $18 = $17&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP16[$12>>1] = $18; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = ((($11)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$19>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = ((($this)) + 124|0); //@line 386 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$20>>2] = $speed; //@line 386 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 return; //@line 387 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
}
function __ZN12b2WheelJoint17SetMaxMotorTorqueEf($this,$torque) {
 $this = $this|0;
 $torque = +$torque;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 391 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 391 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $2 = ((($1)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = $4 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $6 = ($5|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($6) {
  $7 = $4 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $8 = $7&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$2>>1] = $8; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $9 = ((($1)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$9>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $10 = ((($this)) + 52|0); //@line 392 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 392 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $12 = ((($11)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $13 = HEAP16[$12>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $14 = $13&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $15 = $14 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $16 = ($15|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($16)) {
  $20 = ((($this)) + 120|0); //@line 393 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  HEAPF32[$20>>2] = $torque; //@line 393 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
  return; //@line 394 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 }
 $17 = $14 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $18 = $17&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP16[$12>>1] = $18; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = ((($11)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$19>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = ((($this)) + 120|0); //@line 393 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 HEAPF32[$20>>2] = $torque; //@line 393 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 return; //@line 394 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
}
function __ZNK12b2WheelJoint14GetMotorTorqueEf($this,$inv_dt) {
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 398 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 398 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 $2 = $1 * $inv_dt; //@line 398 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
 return (+$2); //@line 398 "libraries/Box2D/Dynamics/Joints/b2WheelJoint.cpp"
}
function _wuff_open($handle_pointer,$callback,$userdata) {
 $handle_pointer = $handle_pointer|0;
 $callback = $callback|0;
 $userdata = $userdata|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $handle = 0, $or$cond = 0, $wuff_status = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $handle_pointer;
 $2 = $callback;
 $3 = $userdata;
 $4 = $1;
 $5 = ($4|0)==(0|0);
 $6 = $2;
 $7 = ($6|0)==(0|0);
 $or$cond = $5 | $7;
 if ($or$cond) {
  $0 = -2;
  $27 = $0;
  STACKTOP = sp;return ($27|0);
 }
 $8 = (_wuff_alloc(144)|0);
 $handle = $8;
 $9 = $handle;
 $10 = ($9|0)==(0|0);
 if ($10) {
  $0 = -3;
  $27 = $0;
  STACKTOP = sp;return ($27|0);
 }
 $11 = $handle;
 _memset(($11|0),0,144)|0;
 $12 = $handle;
 $13 = ((($12)) + 80|0);
 HEAP32[$13>>2] = 0;
 $14 = $2;
 $15 = $handle;
 $16 = ((($15)) + 132|0);
 HEAP32[$16>>2] = $14;
 $17 = $3;
 $18 = $handle;
 $19 = ((($18)) + 136|0);
 HEAP32[$19>>2] = $17;
 $20 = $handle;
 $21 = (_wuff_setup($20)|0);
 $wuff_status = $21;
 $22 = $wuff_status;
 $23 = ($22|0)<(0);
 $24 = $handle;
 if ($23) {
  (_wuff_cleanup($24)|0);
  $25 = $wuff_status;
  $0 = $25;
  $27 = $0;
  STACKTOP = sp;return ($27|0);
 } else {
  $26 = $1;
  HEAP32[$26>>2] = $24;
  $0 = 0;
  $27 = $0;
  STACKTOP = sp;return ($27|0);
 }
 return (0)|0;
}
function _wuff_close($handle) {
 $handle = $handle|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $wuff_status = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $handle;
 $2 = $1;
 $3 = ($2|0)==(0|0);
 do {
  if ($3) {
   $0 = -2;
  } else {
   $4 = $1;
   $5 = (_wuff_cleanup($4)|0);
   $wuff_status = $5;
   $6 = $wuff_status;
   $7 = ($6|0)<(0);
   if ($7) {
    $8 = $wuff_status;
    $0 = $8;
    break;
   } else {
    $0 = 0;
    break;
   }
  }
 } while(0);
 $9 = $0;
 STACKTOP = sp;return ($9|0);
}
function _wuff_seek($handle,$0,$1) {
 $handle = $handle|0;
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $seek_offset = 0, $wuff_status = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $4 = sp + 8|0;
 $seek_offset = sp;
 $3 = $handle;
 $5 = $4;
 $6 = $5;
 HEAP32[$6>>2] = $0;
 $7 = (($5) + 4)|0;
 $8 = $7;
 HEAP32[$8>>2] = $1;
 $9 = $3;
 $10 = ($9|0)==(0|0);
 if ($10) {
  $2 = -2;
  $115 = $2;
  STACKTOP = sp;return ($115|0);
 }
 $11 = $4;
 $12 = $11;
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11) + 4)|0;
 $15 = $14;
 $16 = HEAP32[$15>>2]|0;
 $17 = $3;
 $18 = ((($17)) + 8|0);
 $19 = $18;
 $20 = $19;
 $21 = HEAP32[$20>>2]|0;
 $22 = (($19) + 4)|0;
 $23 = $22;
 $24 = HEAP32[$23>>2]|0;
 $25 = ($16>>>0)<($24>>>0);
 $26 = ($13>>>0)<=($21>>>0);
 $27 = ($16|0)==($24|0);
 $28 = $27 & $26;
 $29 = $25 | $28;
 if ($29) {
  $30 = $4;
  $31 = $30;
  $32 = HEAP32[$31>>2]|0;
  $33 = (($30) + 4)|0;
  $34 = $33;
  $35 = HEAP32[$34>>2]|0;
  $46 = $32;$49 = $35;
 } else {
  $36 = $3;
  $37 = ((($36)) + 8|0);
  $38 = $37;
  $39 = $38;
  $40 = HEAP32[$39>>2]|0;
  $41 = (($38) + 4)|0;
  $42 = $41;
  $43 = HEAP32[$42>>2]|0;
  $46 = $40;$49 = $43;
 }
 $44 = $4;
 $45 = $44;
 HEAP32[$45>>2] = $46;
 $47 = (($44) + 4)|0;
 $48 = $47;
 HEAP32[$48>>2] = $49;
 $50 = $4;
 $51 = $50;
 $52 = HEAP32[$51>>2]|0;
 $53 = (($50) + 4)|0;
 $54 = $53;
 $55 = HEAP32[$54>>2]|0;
 $56 = $3;
 $57 = ((($56)) + 32|0);
 $58 = ((($57)) + 28|0);
 $59 = HEAP32[$58>>2]|0;
 $60 = (___muldi3(($52|0),($55|0),($59|0),0)|0);
 $61 = tempRet0;
 $62 = $seek_offset;
 $63 = $62;
 HEAP32[$63>>2] = $60;
 $64 = (($62) + 4)|0;
 $65 = $64;
 HEAP32[$65>>2] = $61;
 $66 = $3;
 $67 = ((($66)) + 132|0);
 $68 = HEAP32[$67>>2]|0;
 $69 = ((($68)) + 4|0);
 $70 = HEAP32[$69>>2]|0;
 $71 = $3;
 $72 = ((($71)) + 136|0);
 $73 = HEAP32[$72>>2]|0;
 $74 = $3;
 $75 = ((($74)) + 64|0);
 $76 = ((($75)) + 8|0);
 $77 = $76;
 $78 = $77;
 $79 = HEAP32[$78>>2]|0;
 $80 = (($77) + 4)|0;
 $81 = $80;
 $82 = HEAP32[$81>>2]|0;
 $83 = $seek_offset;
 $84 = $83;
 $85 = HEAP32[$84>>2]|0;
 $86 = (($83) + 4)|0;
 $87 = $86;
 $88 = HEAP32[$87>>2]|0;
 $89 = (_i64Add(($79|0),($82|0),($85|0),($88|0))|0);
 $90 = tempRet0;
 $91 = (FUNCTION_TABLE_iiii[$70 & 8191]($73,$89,$90)|0);
 $wuff_status = $91;
 $92 = $wuff_status;
 $93 = ($92|0)<(0);
 if ($93) {
  $94 = $wuff_status;
  $2 = $94;
  $115 = $2;
  STACKTOP = sp;return ($115|0);
 }
 $95 = $4;
 $96 = $95;
 $97 = HEAP32[$96>>2]|0;
 $98 = (($95) + 4)|0;
 $99 = $98;
 $100 = HEAP32[$99>>2]|0;
 $101 = $3;
 $102 = ((($101)) + 24|0);
 $103 = $102;
 $104 = $103;
 HEAP32[$104>>2] = $97;
 $105 = (($103) + 4)|0;
 $106 = $105;
 HEAP32[$106>>2] = $100;
 $107 = $3;
 $108 = ((($107)) + 112|0);
 $109 = ((($108)) + 12|0);
 HEAP32[$109>>2] = 0;
 $110 = $3;
 $111 = (_wuff_buffer_clear($110)|0);
 $wuff_status = $111;
 $112 = $wuff_status;
 $113 = ($112|0)<(0);
 if ($113) {
  $114 = $wuff_status;
  $2 = $114;
  $115 = $2;
  STACKTOP = sp;return ($115|0);
 } else {
  $2 = 0;
  $115 = $2;
  STACKTOP = sp;return ($115|0);
 }
 return (0)|0;
}
function _wuff_stream_info($handle,$info) {
 $handle = $handle|0;
 $info = $info|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $handle;
 $2 = $info;
 $3 = $1;
 $4 = ($3|0)==(0|0);
 $5 = $2;
 $6 = ($5|0)==(0|0);
 $or$cond = $4 | $6;
 if ($or$cond) {
  $0 = -2;
  $43 = $0;
  STACKTOP = sp;return ($43|0);
 } else {
  $7 = $1;
  $8 = ((($7)) + 16|0);
  $9 = HEAP16[$8>>1]|0;
  $10 = $2;
  HEAP16[$10>>1] = $9;
  $11 = $1;
  $12 = ((($11)) + 32|0);
  $13 = ((($12)) + 18|0);
  $14 = HEAP16[$13>>1]|0;
  $15 = $2;
  $16 = ((($15)) + 2|0);
  HEAP16[$16>>1] = $14;
  $17 = $1;
  $18 = ((($17)) + 32|0);
  $19 = ((($18)) + 20|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = $2;
  $22 = ((($21)) + 4|0);
  HEAP32[$22>>2] = $20;
  $23 = $1;
  $24 = ((($23)) + 32|0);
  $25 = ((($24)) + 24|0);
  $26 = HEAP16[$25>>1]|0;
  $27 = $2;
  $28 = ((($27)) + 8|0);
  HEAP16[$28>>1] = $26;
  $29 = $1;
  $30 = ((($29)) + 8|0);
  $31 = $30;
  $32 = $31;
  $33 = HEAP32[$32>>2]|0;
  $34 = (($31) + 4)|0;
  $35 = $34;
  $36 = HEAP32[$35>>2]|0;
  $37 = $2;
  $38 = ((($37)) + 16|0);
  $39 = $38;
  $40 = $39;
  HEAP32[$40>>2] = $33;
  $41 = (($39) + 4)|0;
  $42 = $41;
  HEAP32[$42>>2] = $36;
  $0 = 0;
  $43 = $0;
  STACKTOP = sp;return ($43|0);
 }
 return (0)|0;
}
function _wuff_format($handle,$format) {
 $handle = $handle|0;
 $format = $format|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $wuff_status = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $handle;
 $2 = $format;
 $3 = $1;
 $4 = ($3|0)==(0|0);
 do {
  if ($4) {
   $0 = -2;
  } else {
   $5 = $2;
   $6 = $5&65535;
   $7 = ($6|0)>=(6);
   if ($7) {
    $0 = -200;
    break;
   }
   $8 = $1;
   $9 = $1;
   $10 = ((($9)) + 24|0);
   $11 = $10;
   $12 = $11;
   $13 = HEAP32[$12>>2]|0;
   $14 = (($11) + 4)|0;
   $15 = $14;
   $16 = HEAP32[$15>>2]|0;
   $17 = (_wuff_seek($8,$13,$16)|0);
   $wuff_status = $17;
   $18 = $wuff_status;
   $19 = ($18|0)<(0);
   if ($19) {
    $20 = $wuff_status;
    $0 = $20;
    break;
   }
   $21 = $1;
   $22 = $2;
   $23 = (_wuff_set_output_format($21,$22)|0);
   $wuff_status = $23;
   $24 = $wuff_status;
   $25 = ($24|0)<(0);
   if ($25) {
    $26 = $wuff_status;
    $0 = $26;
    break;
   } else {
    $0 = 0;
    break;
   }
  }
 } while(0);
 $27 = $0;
 STACKTOP = sp;return ($27|0);
}
function _wuff_read($handle,$out_buffer,$out_size) {
 $handle = $handle|0;
 $out_buffer = $out_buffer|0;
 $out_size = $out_size|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $current_offset = 0, $head = 0, $head_offset = 0, $in_buffer = 0, $num_samples = 0, $or$cond = 0, $or$cond3 = 0, $r_samples = 0, $sample_size = 0, $tail = 0, $wuff_status = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $r_samples = sp + 12|0;
 $head = sp + 38|0;
 $tail = sp + 37|0;
 $in_buffer = sp + 4|0;
 $1 = $handle;
 $2 = $out_buffer;
 $3 = $out_size;
 $4 = $1;
 $5 = ($4|0)==(0|0);
 $6 = $2;
 $7 = ($6|0)==(0|0);
 $or$cond = $5 | $7;
 $8 = $3;
 $9 = ($8|0)==(0|0);
 $or$cond3 = $or$cond | $9;
 if ($or$cond3) {
  $0 = -2;
  $148 = $0;
  STACKTOP = sp;return ($148|0);
 }
 $10 = $3;
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0);
 if ($12) {
  $0 = 0;
  $148 = $0;
  STACKTOP = sp;return ($148|0);
 }
 $13 = $1;
 $14 = ((($13)) + 112|0);
 $15 = ((($14)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16&255;
 $sample_size = $17;
 $18 = $1;
 $19 = ((($18)) + 112|0);
 $20 = ((($19)) + 12|0);
 $21 = HEAP32[$20>>2]|0;
 $current_offset = $21;
 $22 = $current_offset;
 $23 = $sample_size;
 $24 = $23&255;
 $25 = (($22>>>0) % ($24>>>0))&-1;
 $26 = $25&255;
 $head_offset = $26;
 $27 = $head_offset;
 $28 = $27&255;
 $29 = ($28|0)==(0);
 if ($29) {
  $36 = 0;
 } else {
  $30 = $sample_size;
  $31 = $30&255;
  $32 = $head_offset;
  $33 = $32&255;
  $34 = (($31) - ($33))|0;
  $36 = $34;
 }
 $35 = $36&255;
 HEAP8[$head>>0] = $35;
 $37 = $3;
 $38 = HEAP32[$37>>2]|0;
 $39 = $sample_size;
 $40 = (_wuff_calculate_samples($38,$39,$head,$tail)|0);
 $num_samples = $40;
 $41 = $num_samples;
 HEAP32[$r_samples>>2] = $41;
 $42 = $1;
 $43 = (_wuff_buffer_request($42,$in_buffer,$r_samples)|0);
 $wuff_status = $43;
 $44 = $wuff_status;
 $45 = ($44|0)<(0);
 if ($45) {
  $46 = $wuff_status;
  $0 = $46;
  $148 = $0;
  STACKTOP = sp;return ($148|0);
 }
 $47 = HEAP32[$r_samples>>2]|0;
 $48 = ($47|0)==(0);
 if ($48) {
  $49 = $3;
  HEAP32[$49>>2] = 0;
 } else {
  $50 = HEAP32[$r_samples>>2]|0;
  $51 = ($50|0)==(1);
  if ($51) {
   $52 = HEAP8[$head>>0]|0;
   $53 = $52&255;
   $54 = ($53|0)!=(0);
   if ($54) {
    $num_samples = 0;
    HEAP8[$tail>>0] = 0;
   } else {
    label = 14;
   }
  } else {
   label = 14;
  }
  if ((label|0) == 14) {
   $55 = HEAP32[$r_samples>>2]|0;
   $56 = $num_samples;
   $57 = ($55>>>0)<($56>>>0);
   if ($57) {
    HEAP8[$tail>>0] = 0;
   }
   $58 = HEAP32[$r_samples>>2]|0;
   $59 = HEAP8[$head>>0]|0;
   $60 = ($59<<24>>24)!=(0);
   $61 = $60 ^ 1;
   $62 = $61 ^ 1;
   $63 = $62&1;
   $64 = (($58) - ($63))|0;
   $65 = HEAP8[$tail>>0]|0;
   $66 = ($65<<24>>24)!=(0);
   $67 = $66 ^ 1;
   $68 = $67 ^ 1;
   $69 = $68&1;
   $70 = (($64) - ($69))|0;
   $num_samples = $70;
  }
  $71 = $1;
  $72 = ((($71)) + 112|0);
  $73 = ((($72)) + 16|0);
  $74 = HEAP32[$73>>2]|0;
  $75 = $2;
  $76 = HEAP32[$in_buffer>>2]|0;
  $77 = $num_samples;
  $78 = $head_offset;
  $79 = HEAP8[$head>>0]|0;
  $80 = HEAP8[$tail>>0]|0;
  FUNCTION_TABLE_viiiiii[$74 & 8191]($75,$76,$77,$78,$79,$80);
  $81 = $num_samples;
  $82 = $sample_size;
  $83 = $82&255;
  $84 = Math_imul($81, $83)|0;
  $85 = HEAP8[$head>>0]|0;
  $86 = $85&255;
  $87 = (($84) + ($86))|0;
  $88 = HEAP8[$tail>>0]|0;
  $89 = $88&255;
  $90 = (($87) + ($89))|0;
  $91 = $3;
  HEAP32[$91>>2] = $90;
  $92 = $3;
  $93 = HEAP32[$92>>2]|0;
  $94 = $current_offset;
  $95 = (($94) + ($93))|0;
  $current_offset = $95;
  $96 = $current_offset;
  $97 = $1;
  $98 = ((($97)) + 112|0);
  $99 = ((($98)) + 8|0);
  $100 = HEAP32[$99>>2]|0;
  $101 = ($96>>>0)>=($100>>>0);
  $102 = $current_offset;
  $103 = $1;
  $104 = ((($103)) + 112|0);
  if ($101) {
   $105 = ((($104)) + 8|0);
   $106 = HEAP32[$105>>2]|0;
   $107 = (($102>>>0) / ($106>>>0))&-1;
   $108 = $1;
   $109 = ((($108)) + 24|0);
   $110 = $109;
   $111 = $110;
   $112 = HEAP32[$111>>2]|0;
   $113 = (($110) + 4)|0;
   $114 = $113;
   $115 = HEAP32[$114>>2]|0;
   $116 = (_i64Add(($112|0),($115|0),($107|0),0)|0);
   $117 = tempRet0;
   $118 = $109;
   $119 = $118;
   HEAP32[$119>>2] = $116;
   $120 = (($118) + 4)|0;
   $121 = $120;
   HEAP32[$121>>2] = $117;
   $122 = $current_offset;
   $123 = $1;
   $124 = ((($123)) + 112|0);
   $125 = ((($124)) + 8|0);
   $126 = HEAP32[$125>>2]|0;
   $127 = (($122>>>0) % ($126>>>0))&-1;
   $128 = $1;
   $129 = ((($128)) + 112|0);
   $130 = ((($129)) + 12|0);
   HEAP32[$130>>2] = $127;
  } else {
   $131 = ((($104)) + 12|0);
   HEAP32[$131>>2] = $102;
  }
  $132 = $1;
  $133 = $head_offset;
  $134 = $133&255;
  $135 = HEAP8[$head>>0]|0;
  $136 = $135&255;
  $137 = (($134) + ($136))|0;
  $138 = $sample_size;
  $139 = $138&255;
  $140 = ($137|0)==($139|0);
  $141 = $num_samples;
  $142 = (($141) + 1)|0;
  $143 = $140 ? $142 : $141;
  $144 = (_wuff_buffer_release($132,$143)|0);
  $wuff_status = $144;
  $145 = $wuff_status;
  $146 = ($145|0)<(0);
  if ($146) {
   $147 = $wuff_status;
   $0 = $147;
   $148 = $0;
   STACKTOP = sp;return ($148|0);
  }
 }
 $0 = 0;
 $148 = $0;
 STACKTOP = sp;return ($148|0);
}
function _wuff_setup($handle) {
 $handle = $handle|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $wuff_status = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $handle;
 $2 = $1;
 $3 = ($2|0)==(0|0);
 do {
  if ($3) {
   $0 = -2;
  } else {
   $4 = $1;
   $5 = (_wuff_init_stream($4)|0);
   $wuff_status = $5;
   $6 = $wuff_status;
   $7 = ($6|0)<(0);
   if ($7) {
    $8 = $wuff_status;
    $0 = $8;
    break;
   }
   $9 = $1;
   $10 = (_wuff_buffer_alloc($9)|0);
   $wuff_status = $10;
   $11 = $wuff_status;
   $12 = ($11|0)<(0);
   if ($12) {
    $13 = $wuff_status;
    $0 = $13;
    break;
   }
   $14 = $1;
   $15 = $1;
   $16 = ((($15)) + 16|0);
   $17 = HEAP16[$16>>1]|0;
   $18 = (_wuff_format($14,$17)|0);
   $wuff_status = $18;
   $19 = $wuff_status;
   $20 = ($19|0)<(0);
   if ($20) {
    $21 = $wuff_status;
    $0 = $21;
    break;
   } else {
    $0 = 0;
    break;
   }
  }
 } while(0);
 $22 = $0;
 STACKTOP = sp;return ($22|0);
}
function _wuff_init_stream($handle) {
 $handle = $handle|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $buffer = 0, $buffer_size = 0, $chunk = 0, $search_offset = 0, $wuff_status = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $buffer = sp + 40|0;
 $buffer_size = sp + 24|0;
 $search_offset = sp;
 $chunk = sp + 16|0;
 $1 = $handle;
 HEAP32[$buffer_size>>2] = 80;
 $2 = $1;
 $3 = ((($2)) + 132|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = HEAP32[$4>>2]|0;
 $6 = $1;
 $7 = ((($6)) + 136|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (FUNCTION_TABLE_iiii[$5 & 8191]($8,$buffer,$buffer_size)|0);
 $wuff_status = $9;
 $10 = $wuff_status;
 $11 = ($10|0)<(0);
 if ($11) {
  $12 = $wuff_status;
  $0 = $12;
  $286 = $0;
  STACKTOP = sp;return ($286|0);
 }
 $13 = HEAP32[$buffer_size>>2]|0;
 $14 = ($13>>>0)<(36);
 if ($14) {
  $0 = -100;
  $286 = $0;
  STACKTOP = sp;return ($286|0);
 }
 _wuff_copy_chunk_header_data($chunk,$buffer);
 $15 = HEAP32[$chunk>>2]|0;
 $16 = (_wuff_get_chunk_id(363646)|0);
 $17 = ($15|0)!=($16|0);
 if ($17) {
  $0 = -100;
  $286 = $0;
  STACKTOP = sp;return ($286|0);
 }
 $18 = ((($chunk)) + 4|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = $1;
 $21 = $20;
 $22 = $21;
 HEAP32[$22>>2] = $19;
 $23 = (($21) + 4)|0;
 $24 = $23;
 HEAP32[$24>>2] = 0;
 $25 = ((($buffer)) + 8|0);
 _wuff_copy_chunk_header_data($chunk,$25);
 $26 = HEAP32[$chunk>>2]|0;
 $27 = (_wuff_get_chunk_id(363651)|0);
 $28 = ($26|0)!=($27|0);
 if ($28) {
  $0 = -101;
  $286 = $0;
  STACKTOP = sp;return ($286|0);
 }
 $29 = ((($buffer)) + 12|0);
 _wuff_copy_chunk_header_data($chunk,$29);
 $30 = $search_offset;
 $31 = $30;
 HEAP32[$31>>2] = 12;
 $32 = (($30) + 4)|0;
 $33 = $32;
 HEAP32[$33>>2] = 0;
 $34 = HEAP32[$chunk>>2]|0;
 $35 = (_wuff_get_chunk_id(363656)|0);
 $36 = ($34|0)!=($35|0);
 if ($36) {
  HEAP32[$chunk>>2] = 0;
  $37 = $1;
  $38 = (_wuff_get_chunk_id(363656)|0);
  $39 = (_wuff_get_chunk_id(363661)|0);
  $40 = (_wuff_search_chunk($37,$chunk,$search_offset,$38,$39)|0);
  $wuff_status = $40;
  $41 = $wuff_status;
  $42 = ($41|0)==(-108);
  if ($42) {
   $0 = -106;
   $286 = $0;
   STACKTOP = sp;return ($286|0);
  }
  $43 = $wuff_status;
  $44 = ($43|0)<(0);
  if ($44) {
   $45 = $wuff_status;
   $0 = $45;
   $286 = $0;
   STACKTOP = sp;return ($286|0);
  }
  HEAP32[$buffer_size>>2] = 60;
  $46 = $1;
  $47 = ((($46)) + 132|0);
  $48 = HEAP32[$47>>2]|0;
  $49 = HEAP32[$48>>2]|0;
  $50 = $1;
  $51 = ((($50)) + 136|0);
  $52 = HEAP32[$51>>2]|0;
  $53 = ((($buffer)) + 20|0);
  $54 = (FUNCTION_TABLE_iiii[$49 & 8191]($52,$53,$buffer_size)|0);
  $wuff_status = $54;
  $55 = $wuff_status;
  $56 = ($55|0)<(0);
  if ($56) {
   $57 = $wuff_status;
   $0 = $57;
   $286 = $0;
   STACKTOP = sp;return ($286|0);
  }
  $58 = HEAP32[$buffer_size>>2]|0;
  $59 = ($58>>>0)<(16);
  if ($59) {
   $0 = -102;
   $286 = $0;
   STACKTOP = sp;return ($286|0);
  }
 }
 $60 = ((($chunk)) + 4|0);
 $61 = HEAP32[$60>>2]|0;
 $62 = $1;
 $63 = ((($62)) + 32|0);
 $64 = $63;
 $65 = $64;
 HEAP32[$65>>2] = $61;
 $66 = (($64) + 4)|0;
 $67 = $66;
 HEAP32[$67>>2] = 0;
 $68 = $search_offset;
 $69 = $68;
 $70 = HEAP32[$69>>2]|0;
 $71 = (($68) + 4)|0;
 $72 = $71;
 $73 = HEAP32[$72>>2]|0;
 $74 = (_i64Add(($70|0),($73|0),8,0)|0);
 $75 = tempRet0;
 $76 = $1;
 $77 = ((($76)) + 32|0);
 $78 = ((($77)) + 8|0);
 $79 = $78;
 $80 = $79;
 HEAP32[$80>>2] = $74;
 $81 = (($79) + 4)|0;
 $82 = $81;
 HEAP32[$82>>2] = $75;
 $83 = ((($buffer)) + 20|0);
 $84 = (_wuff_get_uint16($83)|0);
 $85 = $1;
 $86 = ((($85)) + 32|0);
 $87 = ((($86)) + 16|0);
 HEAP16[$87>>1] = $84;
 $88 = ((($buffer)) + 22|0);
 $89 = (_wuff_get_uint16($88)|0);
 $90 = $1;
 $91 = ((($90)) + 32|0);
 $92 = ((($91)) + 18|0);
 HEAP16[$92>>1] = $89;
 $93 = ((($buffer)) + 24|0);
 $94 = (_wuff_get_uint32($93)|0);
 $95 = $1;
 $96 = ((($95)) + 32|0);
 $97 = ((($96)) + 20|0);
 HEAP32[$97>>2] = $94;
 $98 = ((($buffer)) + 34|0);
 $99 = (_wuff_get_uint16($98)|0);
 $100 = $1;
 $101 = ((($100)) + 32|0);
 $102 = ((($101)) + 24|0);
 HEAP16[$102>>1] = $99;
 $103 = $1;
 $104 = ((($103)) + 32|0);
 $105 = ((($104)) + 24|0);
 $106 = HEAP16[$105>>1]|0;
 $107 = $106&65535;
 $108 = (($107|0) / 8)&-1;
 $109 = $108&65535;
 $110 = $1;
 $111 = ((($110)) + 32|0);
 $112 = ((($111)) + 26|0);
 HEAP16[$112>>1] = $109;
 $113 = $1;
 $114 = ((($113)) + 32|0);
 $115 = ((($114)) + 18|0);
 $116 = HEAP16[$115>>1]|0;
 $117 = $116&65535;
 $118 = $1;
 $119 = ((($118)) + 32|0);
 $120 = ((($119)) + 26|0);
 $121 = HEAP16[$120>>1]|0;
 $122 = $121&65535;
 $123 = Math_imul($117, $122)|0;
 $124 = $1;
 $125 = ((($124)) + 32|0);
 $126 = ((($125)) + 28|0);
 HEAP32[$126>>2] = $123;
 $127 = $1;
 $128 = ((($127)) + 32|0);
 $129 = ((($128)) + 18|0);
 $130 = HEAP16[$129>>1]|0;
 $131 = $130&65535;
 $132 = ($131|0)==(0);
 if ($132) {
  $0 = -103;
  $286 = $0;
  STACKTOP = sp;return ($286|0);
 }
 $133 = $1;
 $134 = ((($133)) + 32|0);
 $135 = ((($134)) + 20|0);
 $136 = HEAP32[$135>>2]|0;
 $137 = ($136|0)==(0);
 if ($137) {
  $0 = -104;
  $286 = $0;
  STACKTOP = sp;return ($286|0);
 }
 $138 = $1;
 $139 = ((($138)) + 32|0);
 $140 = ((($139)) + 24|0);
 $141 = HEAP16[$140>>1]|0;
 $142 = $141&65535;
 $143 = ($142|0)==(0);
 if ($143) {
  $0 = -105;
  $286 = $0;
  STACKTOP = sp;return ($286|0);
 }
 $144 = $1;
 $145 = ((($144)) + 32|0);
 $146 = $145;
 $147 = $146;
 $148 = HEAP32[$147>>2]|0;
 $149 = (($146) + 4)|0;
 $150 = $149;
 $151 = HEAP32[$150>>2]|0;
 $152 = ($151>>>0)>(0);
 $153 = ($148>>>0)>(16);
 $154 = ($151|0)==(0);
 $155 = $154 & $153;
 $156 = $152 | $155;
 if ($156) {
  $157 = ((($buffer)) + 36|0);
  $158 = (_wuff_get_uint16($157)|0);
  $159 = $158&65535;
  $160 = ($159|0)==(22);
  if ($160) {
   $161 = $1;
   $162 = ((($161)) + 32|0);
   $163 = ((($162)) + 16|0);
   $164 = HEAP16[$163>>1]|0;
   $165 = $164&65535;
   $166 = ($165|0)==(65534);
   if ($166) {
    $167 = ((($buffer)) + 44|0);
    $168 = (_wuff_get_uint16($167)|0);
    $169 = $1;
    $170 = ((($169)) + 32|0);
    $171 = ((($170)) + 16|0);
    HEAP16[$171>>1] = $168;
   }
  }
 }
 $172 = $1;
 $173 = ((($172)) + 32|0);
 $174 = ((($173)) + 16|0);
 $175 = HEAP16[$174>>1]|0;
 $176 = $1;
 $177 = ((($176)) + 16|0);
 HEAP16[$177>>1] = $175;
 $178 = $1;
 $179 = ((($178)) + 32|0);
 $180 = ((($179)) + 24|0);
 $181 = HEAP16[$180>>1]|0;
 $182 = $1;
 $183 = ((($182)) + 16|0);
 $184 = (_wuff_check_bits($181,$183)|0);
 $wuff_status = $184;
 $185 = $wuff_status;
 $186 = ($185|0)<(0);
 if ($186) {
  $187 = $wuff_status;
  $0 = $187;
  $286 = $0;
  STACKTOP = sp;return ($286|0);
 }
 $188 = ((($buffer)) + 20|0);
 $189 = $1;
 $190 = ((($189)) + 32|0);
 $191 = $190;
 $192 = $191;
 $193 = HEAP32[$192>>2]|0;
 $194 = (($191) + 4)|0;
 $195 = $194;
 $196 = HEAP32[$195>>2]|0;
 $197 = (($188) + ($193)|0);
 _wuff_copy_chunk_header_data($chunk,$197);
 $198 = $1;
 $199 = ((($198)) + 32|0);
 $200 = ((($199)) + 8|0);
 $201 = $200;
 $202 = $201;
 $203 = HEAP32[$202>>2]|0;
 $204 = (($201) + 4)|0;
 $205 = $204;
 $206 = HEAP32[$205>>2]|0;
 $207 = $1;
 $208 = ((($207)) + 32|0);
 $209 = $208;
 $210 = $209;
 $211 = HEAP32[$210>>2]|0;
 $212 = (($209) + 4)|0;
 $213 = $212;
 $214 = HEAP32[$213>>2]|0;
 $215 = (_i64Add(($203|0),($206|0),($211|0),($214|0))|0);
 $216 = tempRet0;
 $217 = $search_offset;
 $218 = $217;
 HEAP32[$218>>2] = $215;
 $219 = (($217) + 4)|0;
 $220 = $219;
 HEAP32[$220>>2] = $216;
 $221 = $1;
 $222 = (_wuff_get_chunk_id(363661)|0);
 $223 = (_wuff_search_chunk($221,$chunk,$search_offset,$222,0)|0);
 $wuff_status = $223;
 $224 = $wuff_status;
 $225 = ($224|0)==(-108);
 if ($225) {
  $0 = -107;
  $286 = $0;
  STACKTOP = sp;return ($286|0);
 }
 $226 = $wuff_status;
 $227 = ($226|0)<(0);
 if ($227) {
  $228 = $wuff_status;
  $0 = $228;
  $286 = $0;
  STACKTOP = sp;return ($286|0);
 } else {
  $229 = ((($chunk)) + 4|0);
  $230 = HEAP32[$229>>2]|0;
  $231 = $1;
  $232 = ((($231)) + 64|0);
  $233 = $232;
  $234 = $233;
  HEAP32[$234>>2] = $230;
  $235 = (($233) + 4)|0;
  $236 = $235;
  HEAP32[$236>>2] = 0;
  $237 = $search_offset;
  $238 = $237;
  $239 = HEAP32[$238>>2]|0;
  $240 = (($237) + 4)|0;
  $241 = $240;
  $242 = HEAP32[$241>>2]|0;
  $243 = (_i64Add(($239|0),($242|0),8,0)|0);
  $244 = tempRet0;
  $245 = $1;
  $246 = ((($245)) + 64|0);
  $247 = ((($246)) + 8|0);
  $248 = $247;
  $249 = $248;
  HEAP32[$249>>2] = $243;
  $250 = (($248) + 4)|0;
  $251 = $250;
  HEAP32[$251>>2] = $244;
  $252 = $1;
  $253 = ((($252)) + 64|0);
  $254 = $253;
  $255 = $254;
  $256 = HEAP32[$255>>2]|0;
  $257 = (($254) + 4)|0;
  $258 = $257;
  $259 = HEAP32[$258>>2]|0;
  $260 = $1;
  $261 = ((($260)) + 32|0);
  $262 = ((($261)) + 18|0);
  $263 = HEAP16[$262>>1]|0;
  $264 = $263&65535;
  $265 = (___udivdi3(($256|0),($259|0),($264|0),0)|0);
  $266 = tempRet0;
  $267 = $1;
  $268 = ((($267)) + 32|0);
  $269 = ((($268)) + 26|0);
  $270 = HEAP16[$269>>1]|0;
  $271 = $270&65535;
  $272 = (___udivdi3(($265|0),($266|0),($271|0),0)|0);
  $273 = tempRet0;
  $274 = $1;
  $275 = ((($274)) + 8|0);
  $276 = $275;
  $277 = $276;
  HEAP32[$277>>2] = $272;
  $278 = (($276) + 4)|0;
  $279 = $278;
  HEAP32[$279>>2] = $273;
  $280 = $1;
  $281 = ((($280)) + 24|0);
  $282 = $281;
  $283 = $282;
  HEAP32[$283>>2] = 0;
  $284 = (($282) + 4)|0;
  $285 = $284;
  HEAP32[$285>>2] = 0;
  $0 = 0;
  $286 = $0;
  STACKTOP = sp;return ($286|0);
 }
 return (0)|0;
}
function _wuff_copy_chunk_header_data($chunk,$data) {
 $chunk = $chunk|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $id = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $chunk;
 $1 = $data;
 $2 = $0;
 $id = $2;
 $3 = $1;
 $4 = HEAP8[$3>>0]|0;
 $5 = $id;
 HEAP8[$5>>0] = $4;
 $6 = $1;
 $7 = ((($6)) + 1|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = $id;
 $10 = ((($9)) + 1|0);
 HEAP8[$10>>0] = $8;
 $11 = $1;
 $12 = ((($11)) + 2|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = $id;
 $15 = ((($14)) + 2|0);
 HEAP8[$15>>0] = $13;
 $16 = $1;
 $17 = ((($16)) + 3|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = $id;
 $20 = ((($19)) + 3|0);
 HEAP8[$20>>0] = $18;
 $21 = $1;
 $22 = ((($21)) + 4|0);
 $23 = (_wuff_get_uint32($22)|0);
 $24 = $0;
 $25 = ((($24)) + 4|0);
 HEAP32[$25>>2] = $23;
 STACKTOP = sp;return;
}
function _wuff_get_uint32($data) {
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $data;
 $1 = $0;
 $2 = HEAP8[$1>>0]|0;
 $3 = $2&255;
 $4 = $0;
 $5 = ((($4)) + 1|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = $6&255;
 $8 = $7 << 8;
 $9 = (($3) + ($8))|0;
 $10 = $0;
 $11 = ((($10)) + 2|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = $12&255;
 $14 = $13 << 16;
 $15 = (($9) + ($14))|0;
 $16 = $0;
 $17 = ((($16)) + 3|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = $18&255;
 $20 = $19 << 24;
 $21 = (($15) + ($20))|0;
 STACKTOP = sp;return ($21|0);
}
function _wuff_get_chunk_id($txt) {
 $txt = $txt|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $id = 0, $id_bytes = 0, $int_id = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $int_id = sp + 4|0;
 $0 = $txt;
 $1 = $0;
 $id = $1;
 $id_bytes = $int_id;
 $2 = $id;
 $3 = HEAP8[$2>>0]|0;
 $4 = $id_bytes;
 HEAP8[$4>>0] = $3;
 $5 = $id;
 $6 = ((($5)) + 1|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $id_bytes;
 $9 = ((($8)) + 1|0);
 HEAP8[$9>>0] = $7;
 $10 = $id;
 $11 = ((($10)) + 2|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = $id_bytes;
 $14 = ((($13)) + 2|0);
 HEAP8[$14>>0] = $12;
 $15 = $id;
 $16 = ((($15)) + 3|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = $id_bytes;
 $19 = ((($18)) + 3|0);
 HEAP8[$19>>0] = $17;
 $20 = HEAP32[$int_id>>2]|0;
 STACKTOP = sp;return ($20|0);
}
function _wuff_search_chunk($handle,$chunk,$offset,$id,$stop_id) {
 $handle = $handle|0;
 $chunk = $chunk|0;
 $offset = $offset|0;
 $id = $id|0;
 $stop_id = $stop_id|0;
 var $$old = 0, $$old1 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $buffer = 0, $buffer_size = 0, $or$cond = 0, $search_offset = 0, $wuff_status = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $buffer = sp + 40|0;
 $search_offset = sp;
 $buffer_size = sp + 12|0;
 $1 = $handle;
 $2 = $chunk;
 $3 = $offset;
 $4 = $id;
 $5 = $stop_id;
 $wuff_status = 0;
 $6 = $2;
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $2;
  $10 = HEAP32[$9>>2]|0;
  $11 = $4;
  $12 = ($10|0)==($11|0);
  if ($12) {
   $0 = 0;
   $117 = $0;
   STACKTOP = sp;return ($117|0);
  }
 }
 $13 = $3;
 $14 = $13;
 $15 = $14;
 $16 = HEAP32[$15>>2]|0;
 $17 = (($14) + 4)|0;
 $18 = $17;
 $19 = HEAP32[$18>>2]|0;
 $20 = $search_offset;
 $21 = $20;
 HEAP32[$21>>2] = $16;
 $22 = (($20) + 4)|0;
 $23 = $22;
 HEAP32[$23>>2] = $19;
 $$old = $wuff_status;
 $$old1 = ($$old|0)>=(0);
 L6: do {
  if ($$old1) {
   while(1) {
    $24 = $2;
    $25 = ((($24)) + 4|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = (8 + ($26))|0;
    $28 = $search_offset;
    $29 = $28;
    $30 = HEAP32[$29>>2]|0;
    $31 = (($28) + 4)|0;
    $32 = $31;
    $33 = HEAP32[$32>>2]|0;
    $34 = (_i64Add(($30|0),($33|0),($27|0),0)|0);
    $35 = tempRet0;
    $36 = $search_offset;
    $37 = $36;
    HEAP32[$37>>2] = $34;
    $38 = (($36) + 4)|0;
    $39 = $38;
    HEAP32[$39>>2] = $35;
    $40 = $search_offset;
    $41 = $40;
    $42 = HEAP32[$41>>2]|0;
    $43 = (($40) + 4)|0;
    $44 = $43;
    $45 = HEAP32[$44>>2]|0;
    $46 = $42 & 1;
    $47 = ($46|0)!=(0);
    $48 = (0)!=(0);
    $49 = $47 | $48;
    if ($49) {
     $50 = $search_offset;
     $51 = $50;
     $52 = HEAP32[$51>>2]|0;
     $53 = (($50) + 4)|0;
     $54 = $53;
     $55 = HEAP32[$54>>2]|0;
     $56 = (_i64Add(($52|0),($55|0),1,0)|0);
     $57 = tempRet0;
     $58 = $search_offset;
     $59 = $58;
     HEAP32[$59>>2] = $56;
     $60 = (($58) + 4)|0;
     $61 = $60;
     HEAP32[$61>>2] = $57;
    }
    $62 = $1;
    $63 = ((($62)) + 132|0);
    $64 = HEAP32[$63>>2]|0;
    $65 = ((($64)) + 4|0);
    $66 = HEAP32[$65>>2]|0;
    $67 = $1;
    $68 = ((($67)) + 136|0);
    $69 = HEAP32[$68>>2]|0;
    $70 = $search_offset;
    $71 = $70;
    $72 = HEAP32[$71>>2]|0;
    $73 = (($70) + 4)|0;
    $74 = $73;
    $75 = HEAP32[$74>>2]|0;
    $76 = (FUNCTION_TABLE_iiii[$66 & 8191]($69,$72,$75)|0);
    $wuff_status = $76;
    $77 = $wuff_status;
    $78 = ($77|0)<(0);
    if ($78) {
     label = 8;
     break;
    }
    HEAP32[$buffer_size>>2] = 8;
    $80 = $1;
    $81 = ((($80)) + 132|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = HEAP32[$82>>2]|0;
    $84 = $1;
    $85 = ((($84)) + 136|0);
    $86 = HEAP32[$85>>2]|0;
    $87 = (FUNCTION_TABLE_iiii[$83 & 8191]($86,$buffer,$buffer_size)|0);
    $wuff_status = $87;
    $88 = $wuff_status;
    $89 = ($88|0)<(0);
    if ($89) {
     label = 10;
     break;
    }
    $91 = $2;
    _wuff_copy_chunk_header_data($91,$buffer);
    $92 = HEAP32[$buffer_size>>2]|0;
    $93 = ($92>>>0)<(8);
    if ($93) {
     label = 14;
     break;
    }
    $94 = $5;
    $95 = ($94|0)!=(0);
    if ($95) {
     $96 = $2;
     $97 = HEAP32[$96>>2]|0;
     $98 = $5;
     $99 = ($97|0)==($98|0);
     if ($99) {
      label = 14;
      break;
     }
    }
    $100 = $2;
    $101 = HEAP32[$100>>2]|0;
    $102 = $4;
    $103 = ($101|0)!=($102|0);
    $104 = $wuff_status;
    $105 = ($104|0)>=(0);
    $or$cond = $103 & $105;
    if (!($or$cond)) {
     break L6;
    }
   }
   if ((label|0) == 8) {
    $79 = $wuff_status;
    $0 = $79;
    $117 = $0;
    STACKTOP = sp;return ($117|0);
   }
   else if ((label|0) == 10) {
    $90 = $wuff_status;
    $0 = $90;
    $117 = $0;
    STACKTOP = sp;return ($117|0);
   }
   else if ((label|0) == 14) {
    $0 = -108;
    $117 = $0;
    STACKTOP = sp;return ($117|0);
   }
  }
 } while(0);
 $106 = $search_offset;
 $107 = $106;
 $108 = HEAP32[$107>>2]|0;
 $109 = (($106) + 4)|0;
 $110 = $109;
 $111 = HEAP32[$110>>2]|0;
 $112 = $3;
 $113 = $112;
 $114 = $113;
 HEAP32[$114>>2] = $108;
 $115 = (($113) + 4)|0;
 $116 = $115;
 HEAP32[$116>>2] = $111;
 $0 = 0;
 $117 = $0;
 STACKTOP = sp;return ($117|0);
}
function _wuff_get_uint16($data) {
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $data;
 $1 = $0;
 $2 = HEAP8[$1>>0]|0;
 $3 = $2&255;
 $4 = $0;
 $5 = ((($4)) + 1|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = $6&255;
 $8 = $7 << 8;
 $9 = (($3) + ($8))|0;
 $10 = $9&65535;
 STACKTOP = sp;return ($10|0);
}
function _wuff_check_bits($bits,$format) {
 $bits = $bits|0;
 $format = $format|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $bits;
 $2 = $format;
 $3 = $2;
 $4 = HEAP16[$3>>1]|0;
 $5 = $4&65535;
 $6 = ($5|0)==(1);
 L1: do {
  if ($6) {
   $7 = $1;
   $8 = $7&65535;
   switch ($8|0) {
   case 8:  {
    $9 = $2;
    HEAP16[$9>>1] = 0;
    label = 14;
    break L1;
    break;
   }
   case 16:  {
    $10 = $2;
    HEAP16[$10>>1] = 1;
    label = 14;
    break L1;
    break;
   }
   case 24:  {
    $11 = $2;
    HEAP16[$11>>1] = 2;
    label = 14;
    break L1;
    break;
   }
   case 32:  {
    $12 = $2;
    HEAP16[$12>>1] = 3;
    label = 14;
    break L1;
    break;
   }
   default: {
    $0 = -200;
    break L1;
   }
   }
  } else {
   $13 = $2;
   $14 = HEAP16[$13>>1]|0;
   $15 = $14&65535;
   $16 = ($15|0)==(3);
   if (!($16)) {
    $0 = -200;
    break;
   }
   $17 = $1;
   $18 = $17&65535;
   switch ($18|0) {
   case 32:  {
    $19 = $2;
    HEAP16[$19>>1] = 4;
    label = 14;
    break L1;
    break;
   }
   case 64:  {
    $20 = $2;
    HEAP16[$20>>1] = 5;
    label = 14;
    break L1;
    break;
   }
   default: {
    $0 = -200;
    break L1;
   }
   }
  }
 } while(0);
 if ((label|0) == 14) {
  $0 = 0;
 }
 $21 = $0;
 STACKTOP = sp;return ($21|0);
}
function _wuff_buffer_alloc($handle) {
 $handle = $handle|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $wuff_status = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $handle;
 $2 = $1;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $0 = -2;
  $46 = $0;
  STACKTOP = sp;return ($46|0);
 }
 $4 = $1;
 $5 = ((($4)) + 32|0);
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $1;
 $9 = ((($8)) + 32|0);
 $10 = ((($9)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = Math_imul($7, $11)|0;
 $13 = (($12>>>0) / 4)&-1;
 $14 = $1;
 $15 = ((($14)) + 80|0);
 $16 = ((($15)) + 16|0);
 HEAP32[$16>>2] = $13;
 $17 = $1;
 $18 = ((($17)) + 80|0);
 $19 = ((($18)) + 16|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ($20>>>0)<(4096);
 $22 = $1;
 $23 = ((($22)) + 80|0);
 $24 = ((($23)) + 16|0);
 if ($21) {
  HEAP32[$24>>2] = 4096;
 } else {
  $25 = HEAP32[$24>>2]|0;
  $26 = ($25>>>0)>(2097152);
  if ($26) {
   $27 = $1;
   $28 = ((($27)) + 80|0);
   $29 = ((($28)) + 16|0);
   HEAP32[$29>>2] = 2097152;
  }
 }
 $30 = $1;
 $31 = ((($30)) + 80|0);
 $32 = ((($31)) + 16|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = (_wuff_alloc($33)|0);
 $35 = $1;
 $36 = ((($35)) + 80|0);
 HEAP32[$36>>2] = $34;
 $37 = $1;
 $38 = ((($37)) + 80|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ($39|0)==(0|0);
 if ($40) {
  $0 = -3;
  $46 = $0;
  STACKTOP = sp;return ($46|0);
 }
 $41 = $1;
 $42 = (_wuff_buffer_clear($41)|0);
 $wuff_status = $42;
 $43 = $wuff_status;
 $44 = ($43|0)<(0);
 if ($44) {
  $45 = $wuff_status;
  $0 = $45;
  $46 = $0;
  STACKTOP = sp;return ($46|0);
 } else {
  $0 = 0;
  $46 = $0;
  STACKTOP = sp;return ($46|0);
 }
 return (0)|0;
}
function _wuff_buffer_clear($handle) {
 $handle = $handle|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $position = 0, $wuff_status = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $position = sp;
 $1 = $handle;
 $2 = $1;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $0 = -2;
  $106 = $0;
  STACKTOP = sp;return ($106|0);
 }
 $4 = $1;
 $5 = ((($4)) + 132|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($6)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $1;
 $10 = ((($9)) + 136|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (FUNCTION_TABLE_iii[$8 & 8191]($11,$position)|0);
 $wuff_status = $12;
 $13 = $wuff_status;
 $14 = ($13|0)<(0);
 if ($14) {
  $15 = $wuff_status;
  $0 = $15;
  $106 = $0;
  STACKTOP = sp;return ($106|0);
 }
 $16 = $position;
 $17 = $16;
 $18 = HEAP32[$17>>2]|0;
 $19 = (($16) + 4)|0;
 $20 = $19;
 $21 = HEAP32[$20>>2]|0;
 $22 = $1;
 $23 = ((($22)) + 64|0);
 $24 = ((($23)) + 8|0);
 $25 = $24;
 $26 = $25;
 $27 = HEAP32[$26>>2]|0;
 $28 = (($25) + 4)|0;
 $29 = $28;
 $30 = HEAP32[$29>>2]|0;
 $31 = ($21>>>0)<($30>>>0);
 $32 = ($18>>>0)<($27>>>0);
 $33 = ($21|0)==($30|0);
 $34 = $33 & $32;
 $35 = $31 | $34;
 if (!($35)) {
  $36 = $position;
  $37 = $36;
  $38 = HEAP32[$37>>2]|0;
  $39 = (($36) + 4)|0;
  $40 = $39;
  $41 = HEAP32[$40>>2]|0;
  $42 = $1;
  $43 = ((($42)) + 64|0);
  $44 = ((($43)) + 8|0);
  $45 = $44;
  $46 = $45;
  $47 = HEAP32[$46>>2]|0;
  $48 = (($45) + 4)|0;
  $49 = $48;
  $50 = HEAP32[$49>>2]|0;
  $51 = $1;
  $52 = ((($51)) + 64|0);
  $53 = $52;
  $54 = $53;
  $55 = HEAP32[$54>>2]|0;
  $56 = (($53) + 4)|0;
  $57 = $56;
  $58 = HEAP32[$57>>2]|0;
  $59 = (_i64Add(($47|0),($50|0),($55|0),($58|0))|0);
  $60 = tempRet0;
  $61 = ($41>>>0)>($60>>>0);
  $62 = ($38>>>0)>($59>>>0);
  $63 = ($41|0)==($60|0);
  $64 = $63 & $62;
  $65 = $61 | $64;
  if (!($65)) {
   $66 = $1;
   $67 = ((($66)) + 64|0);
   $68 = $67;
   $69 = $68;
   $70 = HEAP32[$69>>2]|0;
   $71 = (($68) + 4)|0;
   $72 = $71;
   $73 = HEAP32[$72>>2]|0;
   $74 = $position;
   $75 = $74;
   $76 = HEAP32[$75>>2]|0;
   $77 = (($74) + 4)|0;
   $78 = $77;
   $79 = HEAP32[$78>>2]|0;
   $80 = $1;
   $81 = ((($80)) + 64|0);
   $82 = ((($81)) + 8|0);
   $83 = $82;
   $84 = $83;
   $85 = HEAP32[$84>>2]|0;
   $86 = (($83) + 4)|0;
   $87 = $86;
   $88 = HEAP32[$87>>2]|0;
   $89 = (_i64Subtract(($76|0),($79|0),($85|0),($88|0))|0);
   $90 = tempRet0;
   $91 = (_i64Subtract(($70|0),($73|0),($89|0),($90|0))|0);
   $92 = tempRet0;
   $93 = $1;
   $94 = ((($93)) + 80|0);
   $95 = ((($94)) + 8|0);
   $96 = $95;
   $97 = $96;
   HEAP32[$97>>2] = $91;
   $98 = (($96) + 4)|0;
   $99 = $98;
   HEAP32[$99>>2] = $92;
   $100 = $1;
   $101 = ((($100)) + 80|0);
   $102 = ((($101)) + 20|0);
   HEAP32[$102>>2] = 0;
   $103 = $1;
   $104 = ((($103)) + 80|0);
   $105 = ((($104)) + 24|0);
   HEAP32[$105>>2] = 0;
   $0 = 0;
   $106 = $0;
   STACKTOP = sp;return ($106|0);
  }
 }
 $0 = -301;
 $106 = $0;
 STACKTOP = sp;return ($106|0);
}
function _wuff_cleanup($handle) {
 $handle = $handle|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $handle;
 $2 = $1;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $0 = -2;
  $12 = $0;
  STACKTOP = sp;return ($12|0);
 }
 $4 = $1;
 $5 = ((($4)) + 80|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)!=(0|0);
 if ($7) {
  $8 = $1;
  $9 = ((($8)) + 80|0);
  $10 = HEAP32[$9>>2]|0;
  _wuff_free($10);
 }
 $11 = $1;
 _wuff_free($11);
 $0 = 0;
 $12 = $0;
 STACKTOP = sp;return ($12|0);
}
function _wuff_set_output_format($handle,$format) {
 $handle = $handle|0;
 $format = $format|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bits = 0, $stream_format = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $handle;
 $2 = $format;
 $3 = $1;
 $4 = ($3|0)==(0|0);
 if ($4) {
  $0 = -2;
  $154 = $0;
  STACKTOP = sp;return ($154|0);
 }
 $5 = $2;
 $6 = $5&65535;
 $7 = ($6|0)>=(6);
 if ($7) {
  $0 = -200;
  $154 = $0;
  STACKTOP = sp;return ($154|0);
 }
 $8 = $1;
 $9 = ((($8)) + 16|0);
 $10 = HEAP16[$9>>1]|0;
 $stream_format = $10;
 $11 = $2;
 $12 = $11&65535;
 L9: do {
  switch ($12|0) {
  case 0:  {
   $bits = 8;
   $13 = $stream_format;
   $14 = $13&65535;
   switch ($14|0) {
   case 0:  {
    $15 = $1;
    $16 = ((($15)) + 112|0);
    $17 = ((($16)) + 16|0);
    HEAP32[$17>>2] = 3846;
    break L9;
    break;
   }
   case 1:  {
    $18 = $1;
    $19 = ((($18)) + 112|0);
    $20 = ((($19)) + 16|0);
    HEAP32[$20>>2] = 3847;
    break L9;
    break;
   }
   case 2:  {
    $21 = $1;
    $22 = ((($21)) + 112|0);
    $23 = ((($22)) + 16|0);
    HEAP32[$23>>2] = 3848;
    break L9;
    break;
   }
   case 3:  {
    $24 = $1;
    $25 = ((($24)) + 112|0);
    $26 = ((($25)) + 16|0);
    HEAP32[$26>>2] = 3849;
    break L9;
    break;
   }
   case 4:  {
    $27 = $1;
    $28 = ((($27)) + 112|0);
    $29 = ((($28)) + 16|0);
    HEAP32[$29>>2] = 3850;
    break L9;
    break;
   }
   case 5:  {
    $30 = $1;
    $31 = ((($30)) + 112|0);
    $32 = ((($31)) + 16|0);
    HEAP32[$32>>2] = 3851;
    break L9;
    break;
   }
   default: {
    break L9;
   }
   }
   break;
  }
  case 1:  {
   $bits = 16;
   $33 = $stream_format;
   $34 = $33&65535;
   switch ($34|0) {
   case 0:  {
    $35 = $1;
    $36 = ((($35)) + 112|0);
    $37 = ((($36)) + 16|0);
    HEAP32[$37>>2] = 3852;
    break L9;
    break;
   }
   case 1:  {
    $38 = $1;
    $39 = ((($38)) + 112|0);
    $40 = ((($39)) + 16|0);
    HEAP32[$40>>2] = 3853;
    break L9;
    break;
   }
   case 2:  {
    $41 = $1;
    $42 = ((($41)) + 112|0);
    $43 = ((($42)) + 16|0);
    HEAP32[$43>>2] = 3854;
    break L9;
    break;
   }
   case 3:  {
    $44 = $1;
    $45 = ((($44)) + 112|0);
    $46 = ((($45)) + 16|0);
    HEAP32[$46>>2] = 3855;
    break L9;
    break;
   }
   case 4:  {
    $47 = $1;
    $48 = ((($47)) + 112|0);
    $49 = ((($48)) + 16|0);
    HEAP32[$49>>2] = 3856;
    break L9;
    break;
   }
   case 5:  {
    $50 = $1;
    $51 = ((($50)) + 112|0);
    $52 = ((($51)) + 16|0);
    HEAP32[$52>>2] = 3857;
    break L9;
    break;
   }
   default: {
    break L9;
   }
   }
   break;
  }
  case 2:  {
   $bits = 24;
   $53 = $stream_format;
   $54 = $53&65535;
   switch ($54|0) {
   case 0:  {
    $55 = $1;
    $56 = ((($55)) + 112|0);
    $57 = ((($56)) + 16|0);
    HEAP32[$57>>2] = 3858;
    break L9;
    break;
   }
   case 1:  {
    $58 = $1;
    $59 = ((($58)) + 112|0);
    $60 = ((($59)) + 16|0);
    HEAP32[$60>>2] = 3859;
    break L9;
    break;
   }
   case 2:  {
    $61 = $1;
    $62 = ((($61)) + 112|0);
    $63 = ((($62)) + 16|0);
    HEAP32[$63>>2] = 3860;
    break L9;
    break;
   }
   case 3:  {
    $64 = $1;
    $65 = ((($64)) + 112|0);
    $66 = ((($65)) + 16|0);
    HEAP32[$66>>2] = 3861;
    break L9;
    break;
   }
   case 4:  {
    $67 = $1;
    $68 = ((($67)) + 112|0);
    $69 = ((($68)) + 16|0);
    HEAP32[$69>>2] = 3862;
    break L9;
    break;
   }
   case 5:  {
    $70 = $1;
    $71 = ((($70)) + 112|0);
    $72 = ((($71)) + 16|0);
    HEAP32[$72>>2] = 3863;
    break L9;
    break;
   }
   default: {
    break L9;
   }
   }
   break;
  }
  case 3:  {
   $bits = 32;
   $73 = $stream_format;
   $74 = $73&65535;
   switch ($74|0) {
   case 0:  {
    $75 = $1;
    $76 = ((($75)) + 112|0);
    $77 = ((($76)) + 16|0);
    HEAP32[$77>>2] = 3864;
    break L9;
    break;
   }
   case 1:  {
    $78 = $1;
    $79 = ((($78)) + 112|0);
    $80 = ((($79)) + 16|0);
    HEAP32[$80>>2] = 3865;
    break L9;
    break;
   }
   case 2:  {
    $81 = $1;
    $82 = ((($81)) + 112|0);
    $83 = ((($82)) + 16|0);
    HEAP32[$83>>2] = 3866;
    break L9;
    break;
   }
   case 3:  {
    $84 = $1;
    $85 = ((($84)) + 112|0);
    $86 = ((($85)) + 16|0);
    HEAP32[$86>>2] = 3867;
    break L9;
    break;
   }
   case 4:  {
    $87 = $1;
    $88 = ((($87)) + 112|0);
    $89 = ((($88)) + 16|0);
    HEAP32[$89>>2] = 3868;
    break L9;
    break;
   }
   case 5:  {
    $90 = $1;
    $91 = ((($90)) + 112|0);
    $92 = ((($91)) + 16|0);
    HEAP32[$92>>2] = 3869;
    break L9;
    break;
   }
   default: {
    break L9;
   }
   }
   break;
  }
  case 4:  {
   $bits = 32;
   $93 = $stream_format;
   $94 = $93&65535;
   switch ($94|0) {
   case 0:  {
    $95 = $1;
    $96 = ((($95)) + 112|0);
    $97 = ((($96)) + 16|0);
    HEAP32[$97>>2] = 3870;
    break L9;
    break;
   }
   case 1:  {
    $98 = $1;
    $99 = ((($98)) + 112|0);
    $100 = ((($99)) + 16|0);
    HEAP32[$100>>2] = 3871;
    break L9;
    break;
   }
   case 2:  {
    $101 = $1;
    $102 = ((($101)) + 112|0);
    $103 = ((($102)) + 16|0);
    HEAP32[$103>>2] = 3872;
    break L9;
    break;
   }
   case 3:  {
    $104 = $1;
    $105 = ((($104)) + 112|0);
    $106 = ((($105)) + 16|0);
    HEAP32[$106>>2] = 3873;
    break L9;
    break;
   }
   case 4:  {
    $107 = $1;
    $108 = ((($107)) + 112|0);
    $109 = ((($108)) + 16|0);
    HEAP32[$109>>2] = 3874;
    break L9;
    break;
   }
   case 5:  {
    $110 = $1;
    $111 = ((($110)) + 112|0);
    $112 = ((($111)) + 16|0);
    HEAP32[$112>>2] = 3875;
    break L9;
    break;
   }
   default: {
    break L9;
   }
   }
   break;
  }
  case 5:  {
   $bits = 64;
   $113 = $stream_format;
   $114 = $113&65535;
   switch ($114|0) {
   case 0:  {
    $115 = $1;
    $116 = ((($115)) + 112|0);
    $117 = ((($116)) + 16|0);
    HEAP32[$117>>2] = 3876;
    break L9;
    break;
   }
   case 1:  {
    $118 = $1;
    $119 = ((($118)) + 112|0);
    $120 = ((($119)) + 16|0);
    HEAP32[$120>>2] = 3877;
    break L9;
    break;
   }
   case 2:  {
    $121 = $1;
    $122 = ((($121)) + 112|0);
    $123 = ((($122)) + 16|0);
    HEAP32[$123>>2] = 3878;
    break L9;
    break;
   }
   case 3:  {
    $124 = $1;
    $125 = ((($124)) + 112|0);
    $126 = ((($125)) + 16|0);
    HEAP32[$126>>2] = 3879;
    break L9;
    break;
   }
   case 4:  {
    $127 = $1;
    $128 = ((($127)) + 112|0);
    $129 = ((($128)) + 16|0);
    HEAP32[$129>>2] = 3880;
    break L9;
    break;
   }
   case 5:  {
    $130 = $1;
    $131 = ((($130)) + 112|0);
    $132 = ((($131)) + 16|0);
    HEAP32[$132>>2] = 3881;
    break L9;
    break;
   }
   default: {
    break L9;
   }
   }
   break;
  }
  default: {
   $0 = -200;
   $154 = $0;
   STACKTOP = sp;return ($154|0);
  }
  }
 } while(0);
 $133 = $2;
 $134 = $1;
 $135 = ((($134)) + 112|0);
 HEAP16[$135>>1] = $133;
 $136 = $bits;
 $137 = $136&65535;
 $138 = (($137|0) / 8)&-1;
 $139 = $1;
 $140 = ((($139)) + 112|0);
 $141 = ((($140)) + 4|0);
 HEAP32[$141>>2] = $138;
 $142 = $1;
 $143 = ((($142)) + 32|0);
 $144 = ((($143)) + 18|0);
 $145 = HEAP16[$144>>1]|0;
 $146 = $145&65535;
 $147 = $bits;
 $148 = $147&65535;
 $149 = (($148|0) / 8)&-1;
 $150 = Math_imul($146, $149)|0;
 $151 = $1;
 $152 = ((($151)) + 112|0);
 $153 = ((($152)) + 8|0);
 HEAP32[$153>>2] = $150;
 $0 = 0;
 $154 = $0;
 STACKTOP = sp;return ($154|0);
}
function _wuff_calculate_samples($target_size,$sample_size,$head,$tail) {
 $target_size = $target_size|0;
 $sample_size = $sample_size|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $samples = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $target_size;
 $2 = $sample_size;
 $3 = $head;
 $4 = $tail;
 $samples = 0;
 $5 = $3;
 $6 = HEAP8[$5>>0]|0;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 do {
  if ($8) {
   $9 = $1;
   $10 = $3;
   $11 = HEAP8[$10>>0]|0;
   $12 = $11&255;
   $13 = ($9>>>0)<=($12>>>0);
   if (!($13)) {
    $18 = $3;
    $19 = HEAP8[$18>>0]|0;
    $20 = $19&255;
    $21 = $1;
    $22 = (($21) - ($20))|0;
    $1 = $22;
    $23 = $samples;
    $24 = (($23) + 1)|0;
    $samples = $24;
    break;
   }
   $14 = $1;
   $15 = $14&255;
   $16 = $3;
   HEAP8[$16>>0] = $15;
   $17 = $4;
   HEAP8[$17>>0] = 0;
   $0 = 1;
   $42 = $0;
   STACKTOP = sp;return ($42|0);
  }
 } while(0);
 $25 = $1;
 $26 = $2;
 $27 = $26&255;
 $28 = (($25>>>0) / ($27>>>0))&-1;
 $samples = $28;
 $29 = $1;
 $30 = $2;
 $31 = $30&255;
 $32 = (($29>>>0) % ($31>>>0))&-1;
 $33 = $32&255;
 $34 = $4;
 HEAP8[$34>>0] = $33;
 $35 = $4;
 $36 = HEAP8[$35>>0]|0;
 $37 = $36&255;
 $38 = ($37|0)!=(0);
 if ($38) {
  $39 = $samples;
  $40 = (($39) + 1)|0;
  $samples = $40;
 }
 $41 = $samples;
 $0 = $41;
 $42 = $0;
 STACKTOP = sp;return ($42|0);
}
function _wuff_buffer_fill($handle) {
 $handle = $handle|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bytes_in_buffer = 0, $bytes_to_read = 0, $wuff_status = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $bytes_to_read = sp + 4|0;
 $1 = $handle;
 $2 = $1;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $0 = -2;
  $104 = $0;
  STACKTOP = sp;return ($104|0);
 }
 $4 = $1;
 $5 = ((($4)) + 80|0);
 $6 = ((($5)) + 24|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $1;
 $9 = ((($8)) + 80|0);
 $10 = ((($9)) + 20|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($7) - ($11))|0;
 $bytes_in_buffer = $12;
 $13 = $bytes_in_buffer;
 $14 = $1;
 $15 = ((($14)) + 80|0);
 $16 = ((($15)) + 16|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ($13|0)==($17|0);
 if ($18) {
  $0 = 0;
  $104 = $0;
  STACKTOP = sp;return ($104|0);
 }
 $19 = $bytes_in_buffer;
 $20 = ($19>>>0)>(0);
 if ($20) {
  $21 = $1;
  $22 = ((($21)) + 80|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = $1;
  $25 = ((($24)) + 80|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = $1;
  $28 = ((($27)) + 80|0);
  $29 = ((($28)) + 20|0);
  $30 = HEAP32[$29>>2]|0;
  $31 = (($26) + ($30)|0);
  $32 = $bytes_in_buffer;
  _memmove(($23|0),($31|0),($32|0))|0;
 }
 $33 = $1;
 $34 = ((($33)) + 80|0);
 $35 = ((($34)) + 16|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = $bytes_in_buffer;
 $38 = (($36) - ($37))|0;
 HEAP32[$bytes_to_read>>2] = $38;
 $39 = HEAP32[$bytes_to_read>>2]|0;
 $40 = $1;
 $41 = ((($40)) + 80|0);
 $42 = ((($41)) + 8|0);
 $43 = $42;
 $44 = $43;
 $45 = HEAP32[$44>>2]|0;
 $46 = (($43) + 4)|0;
 $47 = $46;
 $48 = HEAP32[$47>>2]|0;
 $49 = (0)>($48>>>0);
 $50 = ($39>>>0)>($45>>>0);
 $51 = (0)==($48|0);
 $52 = $51 & $50;
 $53 = $49 | $52;
 if ($53) {
  $54 = $1;
  $55 = ((($54)) + 80|0);
  $56 = ((($55)) + 8|0);
  $57 = $56;
  $58 = $57;
  $59 = HEAP32[$58>>2]|0;
  $60 = (($57) + 4)|0;
  $61 = $60;
  $62 = HEAP32[$61>>2]|0;
  HEAP32[$bytes_to_read>>2] = $59;
 }
 $63 = $1;
 $64 = ((($63)) + 132|0);
 $65 = HEAP32[$64>>2]|0;
 $66 = HEAP32[$65>>2]|0;
 $67 = $1;
 $68 = ((($67)) + 136|0);
 $69 = HEAP32[$68>>2]|0;
 $70 = $1;
 $71 = ((($70)) + 80|0);
 $72 = HEAP32[$71>>2]|0;
 $73 = $bytes_in_buffer;
 $74 = (($72) + ($73)|0);
 $75 = (FUNCTION_TABLE_iiii[$66 & 8191]($69,$74,$bytes_to_read)|0);
 $wuff_status = $75;
 $76 = $wuff_status;
 $77 = ($76|0)<(0);
 if ($77) {
  $78 = $wuff_status;
  $0 = $78;
  $104 = $0;
  STACKTOP = sp;return ($104|0);
 } else {
  $79 = $1;
  $80 = ((($79)) + 80|0);
  $81 = ((($80)) + 20|0);
  HEAP32[$81>>2] = 0;
  $82 = $bytes_in_buffer;
  $83 = HEAP32[$bytes_to_read>>2]|0;
  $84 = (($82) + ($83))|0;
  $85 = $1;
  $86 = ((($85)) + 80|0);
  $87 = ((($86)) + 24|0);
  HEAP32[$87>>2] = $84;
  $88 = HEAP32[$bytes_to_read>>2]|0;
  $89 = $1;
  $90 = ((($89)) + 80|0);
  $91 = ((($90)) + 8|0);
  $92 = $91;
  $93 = $92;
  $94 = HEAP32[$93>>2]|0;
  $95 = (($92) + 4)|0;
  $96 = $95;
  $97 = HEAP32[$96>>2]|0;
  $98 = (_i64Subtract(($94|0),($97|0),($88|0),0)|0);
  $99 = tempRet0;
  $100 = $91;
  $101 = $100;
  HEAP32[$101>>2] = $98;
  $102 = (($100) + 4)|0;
  $103 = $102;
  HEAP32[$103>>2] = $99;
  $0 = 0;
  $104 = $0;
  STACKTOP = sp;return ($104|0);
 }
 return (0)|0;
}
function _wuff_buffer_release($handle,$samples) {
 $handle = $handle|0;
 $samples = $samples|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $size = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $handle;
 $2 = $samples;
 $3 = $1;
 $4 = ($3|0)==(0|0);
 if ($4) {
  $0 = -2;
  $29 = $0;
  STACKTOP = sp;return ($29|0);
 }
 $5 = $2;
 $6 = $1;
 $7 = ((($6)) + 32|0);
 $8 = ((($7)) + 26|0);
 $9 = HEAP16[$8>>1]|0;
 $10 = $9&65535;
 $11 = Math_imul($5, $10)|0;
 $size = $11;
 $12 = $size;
 $13 = $1;
 $14 = ((($13)) + 80|0);
 $15 = ((($14)) + 24|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $1;
 $18 = ((($17)) + 80|0);
 $19 = ((($18)) + 20|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($16) - ($20))|0;
 $22 = ($12>>>0)>($21>>>0);
 if ($22) {
  $0 = -300;
  $29 = $0;
  STACKTOP = sp;return ($29|0);
 } else {
  $23 = $size;
  $24 = $1;
  $25 = ((($24)) + 80|0);
  $26 = ((($25)) + 20|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = (($27) + ($23))|0;
  HEAP32[$26>>2] = $28;
  $0 = 0;
  $29 = $0;
  STACKTOP = sp;return ($29|0);
 }
 return (0)|0;
}
function _wuff_buffer_request($handle,$buffer,$samples) {
 $handle = $handle|0;
 $buffer = $buffer|0;
 $samples = $samples|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $bps = 0, $buffer_samples = 0, $or$cond = 0, $or$cond3 = 0, $request_samples = 0, $size = 0, $wuff_status = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $handle;
 $2 = $buffer;
 $3 = $samples;
 $4 = $3;
 $5 = HEAP32[$4>>2]|0;
 $request_samples = $5;
 $6 = $1;
 $7 = ((($6)) + 32|0);
 $8 = ((($7)) + 26|0);
 $9 = HEAP16[$8>>1]|0;
 $10 = $9&65535;
 $bps = $10;
 $11 = $1;
 $12 = ($11|0)==(0|0);
 $13 = $2;
 $14 = ($13|0)==(0|0);
 $or$cond = $12 | $14;
 $15 = $3;
 $16 = ($15|0)==(0|0);
 $or$cond3 = $or$cond | $16;
 if ($or$cond3) {
  $0 = -2;
  $61 = $0;
  STACKTOP = sp;return ($61|0);
 }
 $17 = $request_samples;
 $18 = $bps;
 $19 = Math_imul($17, $18)|0;
 $size = $19;
 $20 = $size;
 $21 = $1;
 $22 = ((($21)) + 80|0);
 $23 = ((($22)) + 24|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = $1;
 $26 = ((($25)) + 80|0);
 $27 = ((($26)) + 20|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = (($24) - ($28))|0;
 $30 = ($20>>>0)>($29>>>0);
 if ($30) {
  $31 = $1;
  $32 = (_wuff_buffer_fill($31)|0);
  $wuff_status = $32;
  $33 = $wuff_status;
  $34 = ($33|0)<(0);
  if ($34) {
   $35 = $wuff_status;
   $0 = $35;
   $61 = $0;
   STACKTOP = sp;return ($61|0);
  }
 }
 $36 = $1;
 $37 = ((($36)) + 80|0);
 $38 = ((($37)) + 24|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = $1;
 $41 = ((($40)) + 80|0);
 $42 = ((($41)) + 20|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = (($39) - ($43))|0;
 $45 = $bps;
 $46 = (($44>>>0) / ($45>>>0))&-1;
 $buffer_samples = $46;
 $47 = $buffer_samples;
 $48 = $request_samples;
 $49 = ($47>>>0)<($48>>>0);
 if ($49) {
  $50 = $buffer_samples;
  $51 = $3;
  HEAP32[$51>>2] = $50;
 }
 $52 = $1;
 $53 = ((($52)) + 80|0);
 $54 = HEAP32[$53>>2]|0;
 $55 = $1;
 $56 = ((($55)) + 80|0);
 $57 = ((($56)) + 20|0);
 $58 = HEAP32[$57>>2]|0;
 $59 = (($54) + ($58)|0);
 $60 = $2;
 HEAP32[$60>>2] = $59;
 $0 = 0;
 $61 = $0;
 STACKTOP = sp;return ($61|0);
}
function _wuff_alloc($size) {
 $size = $size|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $size;
 $1 = $0;
 $2 = (_malloc($1)|0);
 STACKTOP = sp;return ($2|0);
}
function _wuff_free($mem) {
 $mem = $mem|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $mem;
 $1 = $0;
 _free($1);
 STACKTOP = sp;return;
}
function _luaopen_socket_core($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 (_base_open($1)|0);
 $i = 0;
 while(1) {
  $2 = $i;
  $3 = (21196 + ($2<<3)|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ($4|0)!=(0|0);
  if (!($5)) {
   break;
  }
  $6 = $i;
  $7 = (21196 + ($6<<3)|0);
  $8 = ((($7)) + 4|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = $0;
  (FUNCTION_TABLE_ii[$9 & 8191]($10)|0);
  $11 = $i;
  $12 = (($11) + 1)|0;
  $i = $12;
 }
 STACKTOP = sp;return 1;
}
function _base_open($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $L;
 $2 = (_socket_open()|0);
 $3 = ($2|0)!=(0);
 $4 = $1;
 if ($3) {
  _luax_register($4,363666,21172);
  $5 = $1;
  _lua_pushstring($5,520693);
  $6 = $1;
  _lua_pushstring($6,363687);
  $7 = $1;
  _lua_rawset($7,-3);
  $0 = 1;
  $9 = $0;
  STACKTOP = sp;return ($9|0);
 } else {
  _lua_pushstring($4,363703);
  $8 = $1;
  (_lua_error($8)|0);
  $0 = 0;
  $9 = $0;
  STACKTOP = sp;return ($9|0);
 }
 return (0)|0;
}
function _global_skip($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0.0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $amount = 0, $ret = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (+_luaL_checknumber($1,1));
 $3 = (~~(($2)));
 $amount = $3;
 $4 = $0;
 $5 = (_lua_gettop($4)|0);
 $6 = $amount;
 $7 = (($5) - ($6))|0;
 $8 = (($7) - 1)|0;
 $ret = $8;
 $9 = $ret;
 $10 = ($9|0)>=(0);
 $11 = $ret;
 $12 = $10 ? $11 : 0;
 STACKTOP = sp;return ($12|0);
}
function _global_unload($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 (_socket_close()|0);
 STACKTOP = sp;return 0;
}
function _luaopen_mime_core($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 _luax_register($1,363760,21268);
 $2 = $0;
 _lua_pushstring($2,520693);
 $3 = $0;
 _lua_pushstring($3,363894);
 $4 = $0;
 _lua_rawset($4,-3);
 _qpsetup(1251126,1251638);
 _b64setup(1251382);
 STACKTOP = sp;return 1;
}
function _mime_global_dot($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0.0, $26 = 0;
 var $3 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buffer = 0, $input = 0, $isize = 0, $last = 0, $state = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1072|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $isize = sp + 1048|0;
 $buffer = sp;
 $1 = $L;
 HEAP32[$isize>>2] = 0;
 $2 = $1;
 $3 = (+_luaL_checknumber($2,1));
 $4 = (~~(($3))>>>0);
 $state = $4;
 $5 = $1;
 $6 = (_luaL_optlstring($5,2,0,$isize)|0);
 $input = $6;
 $7 = $input;
 $8 = HEAP32[$isize>>2]|0;
 $9 = (($7) + ($8)|0);
 $last = $9;
 $10 = $input;
 $11 = ($10|0)!=(0|0);
 $12 = $1;
 if (!($11)) {
  _lua_pushnil($12);
  $13 = $1;
  _lua_pushnumber($13,2.0);
  $0 = 2;
  $26 = $0;
  STACKTOP = sp;return ($26|0);
 }
 _luaL_buffinit($12,$buffer);
 while(1) {
  $14 = $input;
  $15 = $last;
  $16 = ($14>>>0)<($15>>>0);
  if (!($16)) {
   break;
  }
  $17 = $input;
  $18 = ((($17)) + 1|0);
  $input = $18;
  $19 = HEAP8[$17>>0]|0;
  $20 = $19 << 24 >> 24;
  $21 = $state;
  $22 = (_dot($20,$21,$buffer)|0);
  $state = $22;
 }
 _luaL_pushresult($buffer);
 $23 = $1;
 $24 = $state;
 $25 = (+($24>>>0));
 _lua_pushnumber($23,$25);
 $0 = 2;
 $26 = $0;
 STACKTOP = sp;return ($26|0);
}
function _dot($c,$state,$buffer) {
 $c = $c|0;
 $state = $state|0;
 $buffer = $buffer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $c;
 $2 = $state;
 $3 = $buffer;
 $4 = $3;
 $5 = HEAP32[$4>>2]|0;
 $6 = $3;
 $7 = ((($6)) + 12|0);
 $8 = ((($7)) + 1024|0);
 $9 = ($5>>>0)<($8>>>0);
 if (!($9)) {
  $10 = $3;
  (_luaL_prepbuffer($10)|0);
 }
 $11 = $1;
 $12 = $11&255;
 $13 = $3;
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($14)) + 1|0);
 HEAP32[$13>>2] = $15;
 HEAP8[$14>>0] = $12;
 $16 = $1;
 switch ($16|0) {
 case 13:  {
  $0 = 1;
  $32 = $0;
  STACKTOP = sp;return ($32|0);
  break;
 }
 case 10:  {
  $17 = $2;
  $18 = ($17|0)==(1);
  $19 = $18 ? 2 : 0;
  $0 = $19;
  $32 = $0;
  STACKTOP = sp;return ($32|0);
  break;
 }
 case 46:  {
  $20 = $2;
  $21 = ($20|0)==(2);
  if ($21) {
   $22 = $3;
   $23 = HEAP32[$22>>2]|0;
   $24 = $3;
   $25 = ((($24)) + 12|0);
   $26 = ((($25)) + 1024|0);
   $27 = ($23>>>0)<($26>>>0);
   if (!($27)) {
    $28 = $3;
    (_luaL_prepbuffer($28)|0);
   }
   $29 = $3;
   $30 = HEAP32[$29>>2]|0;
   $31 = ((($30)) + 1|0);
   HEAP32[$29>>2] = $31;
   HEAP8[$30>>0] = 46;
  }
  break;
 }
 default: {
 }
 }
 $0 = 0;
 $32 = $0;
 STACKTOP = sp;return ($32|0);
}
function _mime_global_b64($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $asize = 0, $atom = 0, $buffer = 0, $input = 0, $isize = 0, $last = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1072|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $atom = sp + 1060|0;
 $isize = sp + 1048|0;
 $buffer = sp;
 $1 = $L;
 HEAP32[$isize>>2] = 0;
 $asize = 0;
 $2 = $1;
 $3 = (_luaL_optlstring($2,1,0,$isize)|0);
 $input = $3;
 $4 = $input;
 $5 = HEAP32[$isize>>2]|0;
 $6 = (($4) + ($5)|0);
 $last = $6;
 $7 = $input;
 $8 = ($7|0)!=(0|0);
 $9 = $1;
 if (!($8)) {
  _lua_pushnil($9);
  $10 = $1;
  _lua_pushnil($10);
  $0 = 2;
  $44 = $0;
  STACKTOP = sp;return ($44|0);
 }
 _luaL_buffinit($9,$buffer);
 while(1) {
  $11 = $input;
  $12 = $last;
  $13 = ($11>>>0)<($12>>>0);
  if (!($13)) {
   break;
  }
  $14 = $input;
  $15 = ((($14)) + 1|0);
  $input = $15;
  $16 = HEAP8[$14>>0]|0;
  $17 = $asize;
  $18 = (_b64encode($16,$atom,$17,$buffer)|0);
  $asize = $18;
 }
 $19 = $1;
 $20 = (_luaL_optlstring($19,2,0,$isize)|0);
 $input = $20;
 $21 = $input;
 $22 = ($21|0)!=(0|0);
 if (!($22)) {
  $23 = $asize;
  $24 = (_b64pad($atom,$23,$buffer)|0);
  $asize = $24;
  _luaL_pushresult($buffer);
  $25 = $1;
  $26 = (_lua_tolstring($25,-1,0)|0);
  $27 = HEAP8[$26>>0]|0;
  $28 = ($27<<24>>24)!=(0);
  if (!($28)) {
   $29 = $1;
   _lua_pushnil($29);
  }
  $30 = $1;
  _lua_pushnil($30);
  $0 = 2;
  $44 = $0;
  STACKTOP = sp;return ($44|0);
 }
 $31 = $input;
 $32 = HEAP32[$isize>>2]|0;
 $33 = (($31) + ($32)|0);
 $last = $33;
 while(1) {
  $34 = $input;
  $35 = $last;
  $36 = ($34>>>0)<($35>>>0);
  if (!($36)) {
   break;
  }
  $37 = $input;
  $38 = ((($37)) + 1|0);
  $input = $38;
  $39 = HEAP8[$37>>0]|0;
  $40 = $asize;
  $41 = (_b64encode($39,$atom,$40,$buffer)|0);
  $asize = $41;
 }
 _luaL_pushresult($buffer);
 $42 = $1;
 $43 = $asize;
 _lua_pushlstring($42,$atom,$43);
 $0 = 2;
 $44 = $0;
 STACKTOP = sp;return ($44|0);
}
function _b64encode($c,$input,$size,$buffer) {
 $c = $c|0;
 $input = $input|0;
 $size = $size|0;
 $buffer = $buffer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $code = 0, $value = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $code = sp + 16|0;
 $0 = $c;
 $1 = $input;
 $2 = $size;
 $3 = $buffer;
 $4 = $0;
 $5 = $2;
 $6 = (($5) + 1)|0;
 $2 = $6;
 $7 = $1;
 $8 = (($7) + ($5)|0);
 HEAP8[$8>>0] = $4;
 $9 = $2;
 $10 = ($9|0)==(3);
 if (!($10)) {
  $57 = $2;
  STACKTOP = sp;return ($57|0);
 }
 $value = 0;
 $11 = $1;
 $12 = HEAP8[$11>>0]|0;
 $13 = $12&255;
 $14 = $value;
 $15 = (($14) + ($13))|0;
 $value = $15;
 $16 = $value;
 $17 = $16 << 8;
 $value = $17;
 $18 = $1;
 $19 = ((($18)) + 1|0);
 $20 = HEAP8[$19>>0]|0;
 $21 = $20&255;
 $22 = $value;
 $23 = (($22) + ($21))|0;
 $value = $23;
 $24 = $value;
 $25 = $24 << 8;
 $value = $25;
 $26 = $1;
 $27 = ((($26)) + 2|0);
 $28 = HEAP8[$27>>0]|0;
 $29 = $28&255;
 $30 = $value;
 $31 = (($30) + ($29))|0;
 $value = $31;
 $32 = $value;
 $33 = $32 & 63;
 $34 = (363773 + ($33)|0);
 $35 = HEAP8[$34>>0]|0;
 $36 = ((($code)) + 3|0);
 HEAP8[$36>>0] = $35;
 $37 = $value;
 $38 = $37 >>> 6;
 $value = $38;
 $39 = $value;
 $40 = $39 & 63;
 $41 = (363773 + ($40)|0);
 $42 = HEAP8[$41>>0]|0;
 $43 = ((($code)) + 2|0);
 HEAP8[$43>>0] = $42;
 $44 = $value;
 $45 = $44 >>> 6;
 $value = $45;
 $46 = $value;
 $47 = $46 & 63;
 $48 = (363773 + ($47)|0);
 $49 = HEAP8[$48>>0]|0;
 $50 = ((($code)) + 1|0);
 HEAP8[$50>>0] = $49;
 $51 = $value;
 $52 = $51 >>> 6;
 $value = $52;
 $53 = $value;
 $54 = (363773 + ($53)|0);
 $55 = HEAP8[$54>>0]|0;
 HEAP8[$code>>0] = $55;
 $56 = $3;
 _luaL_addlstring($56,$code,4);
 $2 = 0;
 $57 = $2;
 STACKTOP = sp;return ($57|0);
}
function _b64pad($input,$size,$buffer) {
 $input = $input|0;
 $size = $size|0;
 $buffer = $buffer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $code = 0, $value = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $code = sp + 16|0;
 $0 = $input;
 $1 = $size;
 $2 = $buffer;
 $value = 0;
 ;HEAP8[$code>>0]=HEAP8[363838>>0]|0;HEAP8[$code+1>>0]=HEAP8[363838+1>>0]|0;HEAP8[$code+2>>0]=HEAP8[363838+2>>0]|0;HEAP8[$code+3>>0]=HEAP8[363838+3>>0]|0;
 $3 = $1;
 switch ($3|0) {
 case 1:  {
  $4 = $0;
  $5 = HEAP8[$4>>0]|0;
  $6 = $5&255;
  $7 = $6 << 4;
  $value = $7;
  $8 = $value;
  $9 = $8 & 63;
  $10 = (363773 + ($9)|0);
  $11 = HEAP8[$10>>0]|0;
  $12 = ((($code)) + 1|0);
  HEAP8[$12>>0] = $11;
  $13 = $value;
  $14 = $13 >>> 6;
  $value = $14;
  $15 = $value;
  $16 = (363773 + ($15)|0);
  $17 = HEAP8[$16>>0]|0;
  HEAP8[$code>>0] = $17;
  $18 = $2;
  _luaL_addlstring($18,$code,4);
  STACKTOP = sp;return 0;
  break;
 }
 case 2:  {
  $19 = $0;
  $20 = HEAP8[$19>>0]|0;
  $21 = $20&255;
  $value = $21;
  $22 = $value;
  $23 = $22 << 8;
  $value = $23;
  $24 = $0;
  $25 = ((($24)) + 1|0);
  $26 = HEAP8[$25>>0]|0;
  $27 = $26&255;
  $28 = $value;
  $29 = $28 | $27;
  $value = $29;
  $30 = $value;
  $31 = $30 << 2;
  $value = $31;
  $32 = $value;
  $33 = $32 & 63;
  $34 = (363773 + ($33)|0);
  $35 = HEAP8[$34>>0]|0;
  $36 = ((($code)) + 2|0);
  HEAP8[$36>>0] = $35;
  $37 = $value;
  $38 = $37 >>> 6;
  $value = $38;
  $39 = $value;
  $40 = $39 & 63;
  $41 = (363773 + ($40)|0);
  $42 = HEAP8[$41>>0]|0;
  $43 = ((($code)) + 1|0);
  HEAP8[$43>>0] = $42;
  $44 = $value;
  $45 = $44 >>> 6;
  $value = $45;
  $46 = $value;
  $47 = (363773 + ($46)|0);
  $48 = HEAP8[$47>>0]|0;
  HEAP8[$code>>0] = $48;
  $49 = $2;
  _luaL_addlstring($49,$code,4);
  STACKTOP = sp;return 0;
  break;
 }
 default: {
  STACKTOP = sp;return 0;
 }
 }
 return (0)|0;
}
function _mime_global_eol($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0.0, $3 = 0.0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buffer = 0, $ctx = 0, $input = 0, $isize = 0, $last = 0, $marker = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1072|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $isize = sp + 1048|0;
 $buffer = sp;
 $1 = $L;
 $2 = $1;
 $3 = (+_luaL_checknumber($2,1));
 $4 = (~~(($3)));
 $ctx = $4;
 HEAP32[$isize>>2] = 0;
 $5 = $1;
 $6 = (_luaL_optlstring($5,2,0,$isize)|0);
 $input = $6;
 $7 = $input;
 $8 = HEAP32[$isize>>2]|0;
 $9 = (($7) + ($8)|0);
 $last = $9;
 $10 = $1;
 $11 = (_luaL_optlstring($10,3,363846,0)|0);
 $marker = $11;
 $12 = $1;
 _luaL_buffinit($12,$buffer);
 $13 = $input;
 $14 = ($13|0)!=(0|0);
 if (!($14)) {
  $15 = $1;
  _lua_pushnil($15);
  $16 = $1;
  _lua_pushnumber($16,0.0);
  $0 = 2;
  $30 = $0;
  STACKTOP = sp;return ($30|0);
 }
 while(1) {
  $17 = $input;
  $18 = $last;
  $19 = ($17>>>0)<($18>>>0);
  if (!($19)) {
   break;
  }
  $20 = $input;
  $21 = ((($20)) + 1|0);
  $input = $21;
  $22 = HEAP8[$20>>0]|0;
  $23 = $22 << 24 >> 24;
  $24 = $ctx;
  $25 = $marker;
  $26 = (_eolprocess($23,$24,$25,$buffer)|0);
  $ctx = $26;
 }
 _luaL_pushresult($buffer);
 $27 = $1;
 $28 = $ctx;
 $29 = (+($28|0));
 _lua_pushnumber($27,$29);
 $0 = 2;
 $30 = $0;
 STACKTOP = sp;return ($30|0);
}
function _eolprocess($c,$last,$marker,$buffer) {
 $c = $c|0;
 $last = $last|0;
 $marker = $marker|0;
 $buffer = $buffer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $c;
 $2 = $last;
 $3 = $marker;
 $4 = $buffer;
 $5 = $1;
 $6 = ($5|0)==(13);
 $7 = $1;
 $8 = ($7|0)==(10);
 $or$cond = $6 | $8;
 if (!($or$cond)) {
  $21 = $4;
  $22 = HEAP32[$21>>2]|0;
  $23 = $4;
  $24 = ((($23)) + 12|0);
  $25 = ((($24)) + 1024|0);
  $26 = ($22>>>0)<($25>>>0);
  if (!($26)) {
   $27 = $4;
   (_luaL_prepbuffer($27)|0);
  }
  $28 = $1;
  $29 = $28&255;
  $30 = $4;
  $31 = HEAP32[$30>>2]|0;
  $32 = ((($31)) + 1|0);
  HEAP32[$30>>2] = $32;
  HEAP8[$31>>0] = $29;
  $0 = 0;
  $33 = $0;
  STACKTOP = sp;return ($33|0);
 }
 $9 = $2;
 $10 = ($9|0)==(13);
 $11 = $2;
 $12 = ($11|0)==(10);
 $or$cond3 = $10 | $12;
 if (!($or$cond3)) {
  $18 = $4;
  $19 = $3;
  _luaL_addstring($18,$19);
  $20 = $1;
  $0 = $20;
  $33 = $0;
  STACKTOP = sp;return ($33|0);
 }
 $13 = $1;
 $14 = $2;
 $15 = ($13|0)==($14|0);
 if ($15) {
  $16 = $4;
  $17 = $3;
  _luaL_addstring($16,$17);
 }
 $0 = 0;
 $33 = $0;
 STACKTOP = sp;return ($33|0);
}
function _mime_global_qp($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $asize = 0, $atom = 0, $buffer = 0, $input = 0, $isize = 0, $last = 0, $marker = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1072|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $isize = sp + 1048|0;
 $atom = sp + 1064|0;
 $buffer = sp;
 $1 = $L;
 $asize = 0;
 HEAP32[$isize>>2] = 0;
 $2 = $1;
 $3 = (_luaL_optlstring($2,1,0,$isize)|0);
 $input = $3;
 $4 = $input;
 $5 = HEAP32[$isize>>2]|0;
 $6 = (($4) + ($5)|0);
 $last = $6;
 $7 = $1;
 $8 = (_luaL_optlstring($7,3,363846,0)|0);
 $marker = $8;
 $9 = $input;
 $10 = ($9|0)!=(0|0);
 $11 = $1;
 if (!($10)) {
  _lua_pushnil($11);
  $12 = $1;
  _lua_pushnil($12);
  $0 = 2;
  $48 = $0;
  STACKTOP = sp;return ($48|0);
 }
 _luaL_buffinit($11,$buffer);
 while(1) {
  $13 = $input;
  $14 = $last;
  $15 = ($13>>>0)<($14>>>0);
  if (!($15)) {
   break;
  }
  $16 = $input;
  $17 = ((($16)) + 1|0);
  $input = $17;
  $18 = HEAP8[$16>>0]|0;
  $19 = $asize;
  $20 = $marker;
  $21 = (_qpencode($18,$atom,$19,$20,$buffer)|0);
  $asize = $21;
 }
 $22 = $1;
 $23 = (_luaL_optlstring($22,2,0,$isize)|0);
 $input = $23;
 $24 = $input;
 $25 = ($24|0)!=(0|0);
 if (!($25)) {
  $26 = $asize;
  $27 = (_qppad($atom,$26,$buffer)|0);
  $asize = $27;
  _luaL_pushresult($buffer);
  $28 = $1;
  $29 = (_lua_tolstring($28,-1,0)|0);
  $30 = HEAP8[$29>>0]|0;
  $31 = ($30<<24>>24)!=(0);
  if (!($31)) {
   $32 = $1;
   _lua_pushnil($32);
  }
  $33 = $1;
  _lua_pushnil($33);
  $0 = 2;
  $48 = $0;
  STACKTOP = sp;return ($48|0);
 }
 $34 = $input;
 $35 = HEAP32[$isize>>2]|0;
 $36 = (($34) + ($35)|0);
 $last = $36;
 while(1) {
  $37 = $input;
  $38 = $last;
  $39 = ($37>>>0)<($38>>>0);
  if (!($39)) {
   break;
  }
  $40 = $input;
  $41 = ((($40)) + 1|0);
  $input = $41;
  $42 = HEAP8[$40>>0]|0;
  $43 = $asize;
  $44 = $marker;
  $45 = (_qpencode($42,$atom,$43,$44,$buffer)|0);
  $asize = $45;
 }
 _luaL_pushresult($buffer);
 $46 = $1;
 $47 = $asize;
 _lua_pushlstring($46,$atom,$47);
 $0 = 2;
 $48 = $0;
 STACKTOP = sp;return ($48|0);
}
function _qpencode($c,$input,$size,$marker,$buffer) {
 $c = $c|0;
 $input = $input|0;
 $size = $size|0;
 $marker = $marker|0;
 $buffer = $buffer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $c;
 $2 = $input;
 $3 = $size;
 $4 = $marker;
 $5 = $buffer;
 $6 = $1;
 $7 = $3;
 $8 = (($7) + 1)|0;
 $3 = $8;
 $9 = $2;
 $10 = (($9) + ($7)|0);
 HEAP8[$10>>0] = $6;
 L1: while(1) {
  $11 = $3;
  $12 = ($11>>>0)>(0);
  if (!($12)) {
   label = 22;
   break;
  }
  $13 = $2;
  $14 = HEAP8[$13>>0]|0;
  $15 = $14&255;
  $16 = (1251126 + ($15)|0);
  $17 = HEAP8[$16>>0]|0;
  $18 = $17&255;
  switch ($18|0) {
  case 2:  {
   $19 = $3;
   $20 = ($19>>>0)<(2);
   if ($20) {
    label = 5;
    break L1;
   }
   $22 = $2;
   $23 = ((($22)) + 1|0);
   $24 = HEAP8[$23>>0]|0;
   $25 = $24&255;
   $26 = ($25|0)==(10);
   if ($26) {
    label = 7;
    break L1;
   }
   $29 = $2;
   $30 = HEAP8[$29>>0]|0;
   $31 = $5;
   _qpquote($30,$31);
   break;
  }
  case 3:  {
   $32 = $3;
   $33 = ($32>>>0)<(3);
   if ($33) {
    label = 10;
    break L1;
   }
   $35 = $2;
   $36 = ((($35)) + 1|0);
   $37 = HEAP8[$36>>0]|0;
   $38 = $37&255;
   $39 = ($38|0)==(13);
   if ($39) {
    $40 = $2;
    $41 = ((($40)) + 2|0);
    $42 = HEAP8[$41>>0]|0;
    $43 = $42&255;
    $44 = ($43|0)==(10);
    if ($44) {
     label = 13;
     break L1;
    }
   }
   $50 = $5;
   $51 = HEAP32[$50>>2]|0;
   $52 = $5;
   $53 = ((($52)) + 12|0);
   $54 = ((($53)) + 1024|0);
   $55 = ($51>>>0)<($54>>>0);
   if (!($55)) {
    $56 = $5;
    (_luaL_prepbuffer($56)|0);
   }
   $57 = $2;
   $58 = HEAP8[$57>>0]|0;
   $59 = $5;
   $60 = HEAP32[$59>>2]|0;
   $61 = ((($60)) + 1|0);
   HEAP32[$59>>2] = $61;
   HEAP8[$60>>0] = $58;
   break;
  }
  case 1:  {
   $62 = $2;
   $63 = HEAP8[$62>>0]|0;
   $64 = $5;
   _qpquote($63,$64);
   break;
  }
  default: {
   $65 = $5;
   $66 = HEAP32[$65>>2]|0;
   $67 = $5;
   $68 = ((($67)) + 12|0);
   $69 = ((($68)) + 1024|0);
   $70 = ($66>>>0)<($69>>>0);
   if (!($70)) {
    $71 = $5;
    (_luaL_prepbuffer($71)|0);
   }
   $72 = $2;
   $73 = HEAP8[$72>>0]|0;
   $74 = $5;
   $75 = HEAP32[$74>>2]|0;
   $76 = ((($75)) + 1|0);
   HEAP32[$74>>2] = $76;
   HEAP8[$75>>0] = $73;
  }
  }
  $77 = $2;
  $78 = ((($77)) + 1|0);
  $79 = HEAP8[$78>>0]|0;
  $80 = $2;
  HEAP8[$80>>0] = $79;
  $81 = $2;
  $82 = ((($81)) + 2|0);
  $83 = HEAP8[$82>>0]|0;
  $84 = $2;
  $85 = ((($84)) + 1|0);
  HEAP8[$85>>0] = $83;
  $86 = $3;
  $87 = (($86) + -1)|0;
  $3 = $87;
 }
 if ((label|0) == 5) {
  $21 = $3;
  $0 = $21;
  $88 = $0;
  STACKTOP = sp;return ($88|0);
 }
 else if ((label|0) == 7) {
  $27 = $5;
  $28 = $4;
  _luaL_addstring($27,$28);
  $0 = 0;
  $88 = $0;
  STACKTOP = sp;return ($88|0);
 }
 else if ((label|0) == 10) {
  $34 = $3;
  $0 = $34;
  $88 = $0;
  STACKTOP = sp;return ($88|0);
 }
 else if ((label|0) == 13) {
  $45 = $2;
  $46 = HEAP8[$45>>0]|0;
  $47 = $5;
  _qpquote($46,$47);
  $48 = $5;
  $49 = $4;
  _luaL_addstring($48,$49);
  $0 = 0;
  $88 = $0;
  STACKTOP = sp;return ($88|0);
 }
 else if ((label|0) == 22) {
  $0 = 0;
  $88 = $0;
  STACKTOP = sp;return ($88|0);
 }
 return (0)|0;
}
function _qpquote($c,$buffer) {
 $c = $c|0;
 $buffer = $buffer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $c;
 $1 = $buffer;
 $2 = $1;
 $3 = HEAP32[$2>>2]|0;
 $4 = $1;
 $5 = ((($4)) + 12|0);
 $6 = ((($5)) + 1024|0);
 $7 = ($3>>>0)<($6>>>0);
 if (!($7)) {
  $8 = $1;
  (_luaL_prepbuffer($8)|0);
 }
 $9 = $1;
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + 1|0);
 HEAP32[$9>>2] = $11;
 HEAP8[$10>>0] = 61;
 $12 = $1;
 $13 = HEAP32[$12>>2]|0;
 $14 = $1;
 $15 = ((($14)) + 12|0);
 $16 = ((($15)) + 1024|0);
 $17 = ($13>>>0)<($16>>>0);
 if (!($17)) {
  $18 = $1;
  (_luaL_prepbuffer($18)|0);
 }
 $19 = $0;
 $20 = $19&255;
 $21 = $20 >> 4;
 $22 = (363852 + ($21)|0);
 $23 = HEAP8[$22>>0]|0;
 $24 = $1;
 $25 = HEAP32[$24>>2]|0;
 $26 = ((($25)) + 1|0);
 HEAP32[$24>>2] = $26;
 HEAP8[$25>>0] = $23;
 $27 = $1;
 $28 = HEAP32[$27>>2]|0;
 $29 = $1;
 $30 = ((($29)) + 12|0);
 $31 = ((($30)) + 1024|0);
 $32 = ($28>>>0)<($31>>>0);
 if (!($32)) {
  $33 = $1;
  (_luaL_prepbuffer($33)|0);
 }
 $34 = $0;
 $35 = $34&255;
 $36 = $35 & 15;
 $37 = (363852 + ($36)|0);
 $38 = HEAP8[$37>>0]|0;
 $39 = $1;
 $40 = HEAP32[$39>>2]|0;
 $41 = ((($40)) + 1|0);
 HEAP32[$39>>2] = $41;
 HEAP8[$40>>0] = $38;
 STACKTOP = sp;return;
}
function _qppad($input,$size,$buffer) {
 $input = $input|0;
 $size = $size|0;
 $buffer = $buffer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $input;
 $1 = $size;
 $2 = $buffer;
 $i = 0;
 while(1) {
  $3 = $i;
  $4 = $1;
  $5 = ($3>>>0)<($4>>>0);
  if (!($5)) {
   break;
  }
  $6 = $i;
  $7 = $0;
  $8 = (($7) + ($6)|0);
  $9 = HEAP8[$8>>0]|0;
  $10 = $9&255;
  $11 = (1251126 + ($10)|0);
  $12 = HEAP8[$11>>0]|0;
  $13 = $12&255;
  $14 = ($13|0)==(0);
  if ($14) {
   $15 = $2;
   $16 = HEAP32[$15>>2]|0;
   $17 = $2;
   $18 = ((($17)) + 12|0);
   $19 = ((($18)) + 1024|0);
   $20 = ($16>>>0)<($19>>>0);
   if (!($20)) {
    $21 = $2;
    (_luaL_prepbuffer($21)|0);
   }
   $22 = $i;
   $23 = $0;
   $24 = (($23) + ($22)|0);
   $25 = HEAP8[$24>>0]|0;
   $26 = $2;
   $27 = HEAP32[$26>>2]|0;
   $28 = ((($27)) + 1|0);
   HEAP32[$26>>2] = $28;
   HEAP8[$27>>0] = $25;
  } else {
   $29 = $i;
   $30 = $0;
   $31 = (($30) + ($29)|0);
   $32 = HEAP8[$31>>0]|0;
   $33 = $2;
   _qpquote($32,$33);
  }
  $34 = $i;
  $35 = (($34) + 1)|0;
  $i = $35;
 }
 $36 = $1;
 $37 = ($36>>>0)>(0);
 if (!($37)) {
  STACKTOP = sp;return 0;
 }
 $38 = $2;
 _luaL_addstring($38,363869);
 STACKTOP = sp;return 0;
}
function _mime_global_qpwrp($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $buffer = 0, $input = 0, $last = 0, $left = 0, $length = 0, $size = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1072|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $size = sp + 1052|0;
 $buffer = sp;
 $1 = $L;
 HEAP32[$size>>2] = 0;
 $2 = $1;
 $3 = (+_luaL_checknumber($2,1));
 $4 = (~~(($3)));
 $left = $4;
 $5 = $1;
 $6 = (_luaL_optlstring($5,2,0,$size)|0);
 $input = $6;
 $7 = $input;
 $8 = HEAP32[$size>>2]|0;
 $9 = (($7) + ($8)|0);
 $last = $9;
 $10 = $1;
 $11 = (+_luaL_optnumber($10,3,76.0));
 $12 = (~~(($11)));
 $length = $12;
 $13 = $input;
 $14 = ($13|0)!=(0|0);
 if (!($14)) {
  $15 = $left;
  $16 = $length;
  $17 = ($15|0)<($16|0);
  $18 = $1;
  if ($17) {
   _lua_pushstring($18,363869);
  } else {
   _lua_pushnil($18);
  }
  $19 = $1;
  $20 = $length;
  $21 = (+($20|0));
  _lua_pushnumber($19,$21);
  $0 = 2;
  $61 = $0;
  STACKTOP = sp;return ($61|0);
 }
 $22 = $1;
 _luaL_buffinit($22,$buffer);
 while(1) {
  $23 = $input;
  $24 = $last;
  $25 = ($23>>>0)<($24>>>0);
  if (!($25)) {
   break;
  }
  $26 = $input;
  $27 = HEAP8[$26>>0]|0;
  $28 = $27&255;
  switch ($28|0) {
  case 13:  {
   break;
  }
  case 10:  {
   $29 = $length;
   $left = $29;
   _luaL_addstring($buffer,363846);
   break;
  }
  case 61:  {
   $30 = $left;
   $31 = ($30|0)<=(3);
   if ($31) {
    $32 = $length;
    $left = $32;
    _luaL_addstring($buffer,363869);
   }
   $33 = HEAP32[$buffer>>2]|0;
   $34 = ((($buffer)) + 12|0);
   $35 = ((($34)) + 1024|0);
   $36 = ($33>>>0)<($35>>>0);
   if (!($36)) {
    (_luaL_prepbuffer($buffer)|0);
   }
   $37 = $input;
   $38 = HEAP8[$37>>0]|0;
   $39 = HEAP32[$buffer>>2]|0;
   $40 = ((($39)) + 1|0);
   HEAP32[$buffer>>2] = $40;
   HEAP8[$39>>0] = $38;
   $41 = $left;
   $42 = (($41) + -1)|0;
   $left = $42;
   break;
  }
  default: {
   $43 = $left;
   $44 = ($43|0)<=(1);
   if ($44) {
    $45 = $length;
    $left = $45;
    _luaL_addstring($buffer,363869);
   }
   $46 = HEAP32[$buffer>>2]|0;
   $47 = ((($buffer)) + 12|0);
   $48 = ((($47)) + 1024|0);
   $49 = ($46>>>0)<($48>>>0);
   if (!($49)) {
    (_luaL_prepbuffer($buffer)|0);
   }
   $50 = $input;
   $51 = HEAP8[$50>>0]|0;
   $52 = HEAP32[$buffer>>2]|0;
   $53 = ((($52)) + 1|0);
   HEAP32[$buffer>>2] = $53;
   HEAP8[$52>>0] = $51;
   $54 = $left;
   $55 = (($54) + -1)|0;
   $left = $55;
  }
  }
  $56 = $input;
  $57 = ((($56)) + 1|0);
  $input = $57;
 }
 _luaL_pushresult($buffer);
 $58 = $1;
 $59 = $left;
 $60 = (+($59|0));
 _lua_pushnumber($58,$60);
 $0 = 2;
 $61 = $0;
 STACKTOP = sp;return ($61|0);
}
function _mime_global_unb64($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $asize = 0, $atom = 0, $buffer = 0, $input = 0, $isize = 0, $last = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1072|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $atom = sp + 1060|0;
 $isize = sp + 1048|0;
 $buffer = sp;
 $1 = $L;
 HEAP32[$isize>>2] = 0;
 $asize = 0;
 $2 = $1;
 $3 = (_luaL_optlstring($2,1,0,$isize)|0);
 $input = $3;
 $4 = $input;
 $5 = HEAP32[$isize>>2]|0;
 $6 = (($4) + ($5)|0);
 $last = $6;
 $7 = $input;
 $8 = ($7|0)!=(0|0);
 $9 = $1;
 if (!($8)) {
  _lua_pushnil($9);
  $10 = $1;
  _lua_pushnil($10);
  $0 = 2;
  $42 = $0;
  STACKTOP = sp;return ($42|0);
 }
 _luaL_buffinit($9,$buffer);
 while(1) {
  $11 = $input;
  $12 = $last;
  $13 = ($11>>>0)<($12>>>0);
  if (!($13)) {
   break;
  }
  $14 = $input;
  $15 = ((($14)) + 1|0);
  $input = $15;
  $16 = HEAP8[$14>>0]|0;
  $17 = $asize;
  $18 = (_b64decode($16,$atom,$17,$buffer)|0);
  $asize = $18;
 }
 $19 = $1;
 $20 = (_luaL_optlstring($19,2,0,$isize)|0);
 $input = $20;
 $21 = $input;
 $22 = ($21|0)!=(0|0);
 if (!($22)) {
  _luaL_pushresult($buffer);
  $23 = $1;
  $24 = (_lua_tolstring($23,-1,0)|0);
  $25 = HEAP8[$24>>0]|0;
  $26 = ($25<<24>>24)!=(0);
  if (!($26)) {
   $27 = $1;
   _lua_pushnil($27);
  }
  $28 = $1;
  _lua_pushnil($28);
  $0 = 2;
  $42 = $0;
  STACKTOP = sp;return ($42|0);
 }
 $29 = $input;
 $30 = HEAP32[$isize>>2]|0;
 $31 = (($29) + ($30)|0);
 $last = $31;
 while(1) {
  $32 = $input;
  $33 = $last;
  $34 = ($32>>>0)<($33>>>0);
  if (!($34)) {
   break;
  }
  $35 = $input;
  $36 = ((($35)) + 1|0);
  $input = $36;
  $37 = HEAP8[$35>>0]|0;
  $38 = $asize;
  $39 = (_b64decode($37,$atom,$38,$buffer)|0);
  $asize = $39;
 }
 _luaL_pushresult($buffer);
 $40 = $1;
 $41 = $asize;
 _lua_pushlstring($40,$atom,$41);
 $0 = 2;
 $42 = $0;
 STACKTOP = sp;return ($42|0);
}
function _b64decode($c,$input,$size,$buffer) {
 $c = $c|0;
 $input = $input|0;
 $size = $size|0;
 $buffer = $buffer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $9 = 0, $decoded = 0, $valid = 0, $value = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $decoded = sp + 24|0;
 $1 = $c;
 $2 = $input;
 $3 = $size;
 $4 = $buffer;
 $5 = $1;
 $6 = $5&255;
 $7 = (1251382 + ($6)|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = $8&255;
 $10 = ($9|0)>(64);
 if ($10) {
  $11 = $3;
  $0 = $11;
  $87 = $0;
  STACKTOP = sp;return ($87|0);
 }
 $12 = $1;
 $13 = $3;
 $14 = (($13) + 1)|0;
 $3 = $14;
 $15 = $2;
 $16 = (($15) + ($13)|0);
 HEAP8[$16>>0] = $12;
 $17 = $3;
 $18 = ($17|0)==(4);
 if (!($18)) {
  $86 = $3;
  $0 = $86;
  $87 = $0;
  STACKTOP = sp;return ($87|0);
 }
 $value = 0;
 $19 = $2;
 $20 = HEAP8[$19>>0]|0;
 $21 = $20&255;
 $22 = (1251382 + ($21)|0);
 $23 = HEAP8[$22>>0]|0;
 $24 = $23&255;
 $value = $24;
 $25 = $value;
 $26 = $25 << 6;
 $value = $26;
 $27 = $2;
 $28 = ((($27)) + 1|0);
 $29 = HEAP8[$28>>0]|0;
 $30 = $29&255;
 $31 = (1251382 + ($30)|0);
 $32 = HEAP8[$31>>0]|0;
 $33 = $32&255;
 $34 = $value;
 $35 = $34 | $33;
 $value = $35;
 $36 = $value;
 $37 = $36 << 6;
 $value = $37;
 $38 = $2;
 $39 = ((($38)) + 2|0);
 $40 = HEAP8[$39>>0]|0;
 $41 = $40&255;
 $42 = (1251382 + ($41)|0);
 $43 = HEAP8[$42>>0]|0;
 $44 = $43&255;
 $45 = $value;
 $46 = $45 | $44;
 $value = $46;
 $47 = $value;
 $48 = $47 << 6;
 $value = $48;
 $49 = $2;
 $50 = ((($49)) + 3|0);
 $51 = HEAP8[$50>>0]|0;
 $52 = $51&255;
 $53 = (1251382 + ($52)|0);
 $54 = HEAP8[$53>>0]|0;
 $55 = $54&255;
 $56 = $value;
 $57 = $56 | $55;
 $value = $57;
 $58 = $value;
 $59 = $58 & 255;
 $60 = $59&255;
 $61 = ((($decoded)) + 2|0);
 HEAP8[$61>>0] = $60;
 $62 = $value;
 $63 = $62 >> 8;
 $value = $63;
 $64 = $value;
 $65 = $64 & 255;
 $66 = $65&255;
 $67 = ((($decoded)) + 1|0);
 HEAP8[$67>>0] = $66;
 $68 = $value;
 $69 = $68 >> 8;
 $value = $69;
 $70 = $value;
 $71 = $70&255;
 HEAP8[$decoded>>0] = $71;
 $72 = $2;
 $73 = ((($72)) + 2|0);
 $74 = HEAP8[$73>>0]|0;
 $75 = $74&255;
 $76 = ($75|0)==(61);
 if ($76) {
  $83 = 1;
 } else {
  $77 = $2;
  $78 = ((($77)) + 3|0);
  $79 = HEAP8[$78>>0]|0;
  $80 = $79&255;
  $81 = ($80|0)==(61);
  $82 = $81 ? 2 : 3;
  $83 = $82;
 }
 $valid = $83;
 $84 = $4;
 $85 = $valid;
 _luaL_addlstring($84,$decoded,$85);
 $0 = 0;
 $87 = $0;
 STACKTOP = sp;return ($87|0);
}
function _mime_global_unqp($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $asize = 0, $atom = 0, $buffer = 0, $input = 0, $isize = 0, $last = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1072|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $isize = sp + 1044|0;
 $atom = sp + 1060|0;
 $buffer = sp;
 $1 = $L;
 $asize = 0;
 HEAP32[$isize>>2] = 0;
 $2 = $1;
 $3 = (_luaL_optlstring($2,1,0,$isize)|0);
 $input = $3;
 $4 = $input;
 $5 = HEAP32[$isize>>2]|0;
 $6 = (($4) + ($5)|0);
 $last = $6;
 $7 = $input;
 $8 = ($7|0)!=(0|0);
 $9 = $1;
 if (!($8)) {
  _lua_pushnil($9);
  $10 = $1;
  _lua_pushnil($10);
  $0 = 2;
  $42 = $0;
  STACKTOP = sp;return ($42|0);
 }
 _luaL_buffinit($9,$buffer);
 while(1) {
  $11 = $input;
  $12 = $last;
  $13 = ($11>>>0)<($12>>>0);
  if (!($13)) {
   break;
  }
  $14 = $input;
  $15 = ((($14)) + 1|0);
  $input = $15;
  $16 = HEAP8[$14>>0]|0;
  $17 = $asize;
  $18 = (_qpdecode($16,$atom,$17,$buffer)|0);
  $asize = $18;
 }
 $19 = $1;
 $20 = (_luaL_optlstring($19,2,0,$isize)|0);
 $input = $20;
 $21 = $input;
 $22 = ($21|0)!=(0|0);
 if (!($22)) {
  _luaL_pushresult($buffer);
  $23 = $1;
  $24 = (_lua_tolstring($23,-1,0)|0);
  $25 = HEAP8[$24>>0]|0;
  $26 = ($25<<24>>24)!=(0);
  if (!($26)) {
   $27 = $1;
   _lua_pushnil($27);
  }
  $28 = $1;
  _lua_pushnil($28);
  $0 = 2;
  $42 = $0;
  STACKTOP = sp;return ($42|0);
 }
 $29 = $input;
 $30 = HEAP32[$isize>>2]|0;
 $31 = (($29) + ($30)|0);
 $last = $31;
 while(1) {
  $32 = $input;
  $33 = $last;
  $34 = ($32>>>0)<($33>>>0);
  if (!($34)) {
   break;
  }
  $35 = $input;
  $36 = ((($35)) + 1|0);
  $input = $36;
  $37 = HEAP8[$35>>0]|0;
  $38 = $asize;
  $39 = (_qpdecode($37,$atom,$38,$buffer)|0);
  $asize = $39;
 }
 _luaL_pushresult($buffer);
 $40 = $1;
 $41 = $asize;
 _lua_pushlstring($40,$atom,$41);
 $0 = 2;
 $42 = $0;
 STACKTOP = sp;return ($42|0);
}
function _qpdecode($c,$input,$size,$buffer) {
 $c = $c|0;
 $input = $input|0;
 $size = $size|0;
 $buffer = $buffer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $d = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $c;
 $2 = $input;
 $3 = $size;
 $4 = $buffer;
 $5 = $1;
 $6 = $3;
 $7 = (($6) + 1)|0;
 $3 = $7;
 $8 = $2;
 $9 = (($8) + ($6)|0);
 HEAP8[$9>>0] = $5;
 $10 = $2;
 $11 = HEAP8[$10>>0]|0;
 $12 = $11&255;
 switch ($12|0) {
 case 61:  {
  $13 = $3;
  $14 = ($13>>>0)<(3);
  if ($14) {
   $15 = $3;
   $0 = $15;
   $95 = $0;
   STACKTOP = sp;return ($95|0);
  }
  $16 = $2;
  $17 = ((($16)) + 1|0);
  $18 = HEAP8[$17>>0]|0;
  $19 = $18&255;
  $20 = ($19|0)==(13);
  if ($20) {
   $21 = $2;
   $22 = ((($21)) + 2|0);
   $23 = HEAP8[$22>>0]|0;
   $24 = $23&255;
   $25 = ($24|0)==(10);
   if ($25) {
    $0 = 0;
    $95 = $0;
    STACKTOP = sp;return ($95|0);
   }
  }
  $26 = $2;
  $27 = ((($26)) + 1|0);
  $28 = HEAP8[$27>>0]|0;
  $29 = $28&255;
  $30 = (1251638 + ($29)|0);
  $31 = HEAP8[$30>>0]|0;
  $1 = $31;
  $32 = $2;
  $33 = ((($32)) + 2|0);
  $34 = HEAP8[$33>>0]|0;
  $35 = $34&255;
  $36 = (1251638 + ($35)|0);
  $37 = HEAP8[$36>>0]|0;
  $38 = $37&255;
  $d = $38;
  $39 = $1;
  $40 = $39&255;
  $41 = ($40|0)>(15);
  $42 = $d;
  $43 = ($42|0)>(15);
  $or$cond = $41 | $43;
  $44 = $4;
  if ($or$cond) {
   $45 = $2;
   _luaL_addlstring($44,$45,3);
  } else {
   $46 = HEAP32[$44>>2]|0;
   $47 = $4;
   $48 = ((($47)) + 12|0);
   $49 = ((($48)) + 1024|0);
   $50 = ($46>>>0)<($49>>>0);
   if (!($50)) {
    $51 = $4;
    (_luaL_prepbuffer($51)|0);
   }
   $52 = $1;
   $53 = $52&255;
   $54 = $53 << 4;
   $55 = $d;
   $56 = (($54) + ($55))|0;
   $57 = $56&255;
   $58 = $4;
   $59 = HEAP32[$58>>2]|0;
   $60 = ((($59)) + 1|0);
   HEAP32[$58>>2] = $60;
   HEAP8[$59>>0] = $57;
  }
  $0 = 0;
  $95 = $0;
  STACKTOP = sp;return ($95|0);
  break;
 }
 case 13:  {
  $61 = $3;
  $62 = ($61>>>0)<(2);
  if ($62) {
   $63 = $3;
   $0 = $63;
   $95 = $0;
   STACKTOP = sp;return ($95|0);
  }
  $64 = $2;
  $65 = ((($64)) + 1|0);
  $66 = HEAP8[$65>>0]|0;
  $67 = $66&255;
  $68 = ($67|0)==(10);
  if ($68) {
   $69 = $4;
   $70 = $2;
   _luaL_addlstring($69,$70,2);
  }
  $0 = 0;
  $95 = $0;
  STACKTOP = sp;return ($95|0);
  break;
 }
 default: {
  $71 = $2;
  $72 = HEAP8[$71>>0]|0;
  $73 = $72&255;
  $74 = ($73|0)==(9);
  if ($74) {
   label = 21;
  } else {
   $75 = $2;
   $76 = HEAP8[$75>>0]|0;
   $77 = $76&255;
   $78 = ($77|0)>(31);
   if ($78) {
    $79 = $2;
    $80 = HEAP8[$79>>0]|0;
    $81 = $80&255;
    $82 = ($81|0)<(127);
    if ($82) {
     label = 21;
    }
   }
  }
  if ((label|0) == 21) {
   $83 = $4;
   $84 = HEAP32[$83>>2]|0;
   $85 = $4;
   $86 = ((($85)) + 12|0);
   $87 = ((($86)) + 1024|0);
   $88 = ($84>>>0)<($87>>>0);
   if (!($88)) {
    $89 = $4;
    (_luaL_prepbuffer($89)|0);
   }
   $90 = $2;
   $91 = HEAP8[$90>>0]|0;
   $92 = $4;
   $93 = HEAP32[$92>>2]|0;
   $94 = ((($93)) + 1|0);
   HEAP32[$92>>2] = $94;
   HEAP8[$93>>0] = $91;
  }
  $0 = 0;
  $95 = $0;
  STACKTOP = sp;return ($95|0);
 }
 }
 return (0)|0;
}
function _mime_global_wrp($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0.0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buffer = 0, $input = 0, $last = 0, $left = 0, $length = 0, $size = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1072|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $size = sp + 1052|0;
 $buffer = sp;
 $1 = $L;
 HEAP32[$size>>2] = 0;
 $2 = $1;
 $3 = (+_luaL_checknumber($2,1));
 $4 = (~~(($3)));
 $left = $4;
 $5 = $1;
 $6 = (_luaL_optlstring($5,2,0,$size)|0);
 $input = $6;
 $7 = $input;
 $8 = HEAP32[$size>>2]|0;
 $9 = (($7) + ($8)|0);
 $last = $9;
 $10 = $1;
 $11 = (+_luaL_optnumber($10,3,76.0));
 $12 = (~~(($11)));
 $length = $12;
 $13 = $input;
 $14 = ($13|0)!=(0|0);
 if (!($14)) {
  $15 = $left;
  $16 = $length;
  $17 = ($15|0)<($16|0);
  $18 = $1;
  if ($17) {
   _lua_pushstring($18,363846);
  } else {
   _lua_pushnil($18);
  }
  $19 = $1;
  $20 = $length;
  $21 = (+($20|0));
  _lua_pushnumber($19,$21);
  $0 = 2;
  $48 = $0;
  STACKTOP = sp;return ($48|0);
 }
 $22 = $1;
 _luaL_buffinit($22,$buffer);
 while(1) {
  $23 = $input;
  $24 = $last;
  $25 = ($23>>>0)<($24>>>0);
  if (!($25)) {
   break;
  }
  $26 = $input;
  $27 = HEAP8[$26>>0]|0;
  $28 = $27&255;
  switch ($28|0) {
  case 13:  {
   break;
  }
  case 10:  {
   _luaL_addstring($buffer,363846);
   $29 = $length;
   $left = $29;
   break;
  }
  default: {
   $30 = $left;
   $31 = ($30|0)<=(0);
   if ($31) {
    $32 = $length;
    $left = $32;
    _luaL_addstring($buffer,363846);
   }
   $33 = HEAP32[$buffer>>2]|0;
   $34 = ((($buffer)) + 12|0);
   $35 = ((($34)) + 1024|0);
   $36 = ($33>>>0)<($35>>>0);
   if (!($36)) {
    (_luaL_prepbuffer($buffer)|0);
   }
   $37 = $input;
   $38 = HEAP8[$37>>0]|0;
   $39 = HEAP32[$buffer>>2]|0;
   $40 = ((($39)) + 1|0);
   HEAP32[$buffer>>2] = $40;
   HEAP8[$39>>0] = $38;
   $41 = $left;
   $42 = (($41) + -1)|0;
   $left = $42;
  }
  }
  $43 = $input;
  $44 = ((($43)) + 1|0);
  $input = $44;
 }
 _luaL_pushresult($buffer);
 $45 = $1;
 $46 = $left;
 $47 = (+($46|0));
 _lua_pushnumber($45,$47);
 $0 = 2;
 $48 = $0;
 STACKTOP = sp;return ($48|0);
}
function _qpsetup($qpclass,$qpunbase) {
 $qpclass = $qpclass|0;
 $qpunbase = $qpunbase|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $9 = 0;
 var $i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $qpclass;
 $1 = $qpunbase;
 $i = 0;
 while(1) {
  $2 = $i;
  $3 = ($2|0)<(256);
  if (!($3)) {
   break;
  }
  $4 = $i;
  $5 = $0;
  $6 = (($5) + ($4)|0);
  HEAP8[$6>>0] = 1;
  $7 = $i;
  $8 = (($7) + 1)|0;
  $i = $8;
 }
 $i = 33;
 while(1) {
  $9 = $i;
  $10 = ($9|0)<=(60);
  if (!($10)) {
   break;
  }
  $11 = $i;
  $12 = $0;
  $13 = (($12) + ($11)|0);
  HEAP8[$13>>0] = 0;
  $14 = $i;
  $15 = (($14) + 1)|0;
  $i = $15;
 }
 $i = 62;
 while(1) {
  $16 = $i;
  $17 = ($16|0)<=(126);
  if (!($17)) {
   break;
  }
  $18 = $i;
  $19 = $0;
  $20 = (($19) + ($18)|0);
  HEAP8[$20>>0] = 0;
  $21 = $i;
  $22 = (($21) + 1)|0;
  $i = $22;
 }
 $23 = $0;
 $24 = ((($23)) + 9|0);
 HEAP8[$24>>0] = 3;
 $25 = $0;
 $26 = ((($25)) + 32|0);
 HEAP8[$26>>0] = 3;
 $27 = $0;
 $28 = ((($27)) + 13|0);
 HEAP8[$28>>0] = 2;
 $i = 0;
 while(1) {
  $29 = $i;
  $30 = ($29|0)<(256);
  if (!($30)) {
   break;
  }
  $31 = $i;
  $32 = $1;
  $33 = (($32) + ($31)|0);
  HEAP8[$33>>0] = -1;
  $34 = $i;
  $35 = (($34) + 1)|0;
  $i = $35;
 }
 $36 = $1;
 $37 = ((($36)) + 48|0);
 HEAP8[$37>>0] = 0;
 $38 = $1;
 $39 = ((($38)) + 49|0);
 HEAP8[$39>>0] = 1;
 $40 = $1;
 $41 = ((($40)) + 50|0);
 HEAP8[$41>>0] = 2;
 $42 = $1;
 $43 = ((($42)) + 51|0);
 HEAP8[$43>>0] = 3;
 $44 = $1;
 $45 = ((($44)) + 52|0);
 HEAP8[$45>>0] = 4;
 $46 = $1;
 $47 = ((($46)) + 53|0);
 HEAP8[$47>>0] = 5;
 $48 = $1;
 $49 = ((($48)) + 54|0);
 HEAP8[$49>>0] = 6;
 $50 = $1;
 $51 = ((($50)) + 55|0);
 HEAP8[$51>>0] = 7;
 $52 = $1;
 $53 = ((($52)) + 56|0);
 HEAP8[$53>>0] = 8;
 $54 = $1;
 $55 = ((($54)) + 57|0);
 HEAP8[$55>>0] = 9;
 $56 = $1;
 $57 = ((($56)) + 65|0);
 HEAP8[$57>>0] = 10;
 $58 = $1;
 $59 = ((($58)) + 97|0);
 HEAP8[$59>>0] = 10;
 $60 = $1;
 $61 = ((($60)) + 66|0);
 HEAP8[$61>>0] = 11;
 $62 = $1;
 $63 = ((($62)) + 98|0);
 HEAP8[$63>>0] = 11;
 $64 = $1;
 $65 = ((($64)) + 67|0);
 HEAP8[$65>>0] = 12;
 $66 = $1;
 $67 = ((($66)) + 99|0);
 HEAP8[$67>>0] = 12;
 $68 = $1;
 $69 = ((($68)) + 68|0);
 HEAP8[$69>>0] = 13;
 $70 = $1;
 $71 = ((($70)) + 100|0);
 HEAP8[$71>>0] = 13;
 $72 = $1;
 $73 = ((($72)) + 69|0);
 HEAP8[$73>>0] = 14;
 $74 = $1;
 $75 = ((($74)) + 101|0);
 HEAP8[$75>>0] = 14;
 $76 = $1;
 $77 = ((($76)) + 70|0);
 HEAP8[$77>>0] = 15;
 $78 = $1;
 $79 = ((($78)) + 102|0);
 HEAP8[$79>>0] = 15;
 STACKTOP = sp;return;
}
function _b64setup($b64unbase) {
 $b64unbase = $b64unbase|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $b64unbase;
 $i = 0;
 while(1) {
  $1 = $i;
  $2 = ($1|0)<=(255);
  if (!($2)) {
   break;
  }
  $3 = $i;
  $4 = $0;
  $5 = (($4) + ($3)|0);
  HEAP8[$5>>0] = -1;
  $6 = $i;
  $7 = (($6) + 1)|0;
  $i = $7;
 }
 $i = 0;
 while(1) {
  $8 = $i;
  $9 = ($8|0)<(64);
  if (!($9)) {
   break;
  }
  $10 = $i;
  $11 = $10&255;
  $12 = $i;
  $13 = (363773 + ($12)|0);
  $14 = HEAP8[$13>>0]|0;
  $15 = $14&255;
  $16 = $0;
  $17 = (($16) + ($15)|0);
  HEAP8[$17>>0] = $11;
  $18 = $i;
  $19 = (($18) + 1)|0;
  $i = $19;
 }
 $20 = $0;
 $21 = ((($20)) + 61|0);
 HEAP8[$21>>0] = 0;
 STACKTOP = sp;return;
}
function _select_open($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 _luax_register($1,0,21340);
 STACKTOP = sp;return 0;
}
function _global_select($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, $__b = 0, $__b2 = 0, $__i = 0, $__i1 = 0, $itab = 0, $max_fd = 0, $ndirty = 0, $or$cond = 0, $ret = 0, $rset = 0, $rtab = 0, $t = 0.0, $tm = 0;
 var $wset = 0, $wtab = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 336|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $rset = sp + 176|0;
 $wset = sp + 48|0;
 $tm = sp + 8|0;
 $1 = $L;
 $2 = $1;
 $3 = (+_luaL_optnumber($2,3,-1.0));
 $t = $3;
 $__b = $rset;
 $__i = 32;
 while(1) {
  $4 = $__i;
  $5 = ($4|0)!=(0);
  if (!($5)) {
   break;
  }
  $6 = $__b;
  $7 = ((($6)) + 4|0);
  $__b = $7;
  HEAP32[$6>>2] = 0;
  $8 = $__i;
  $9 = (($8) + -1)|0;
  $__i = $9;
 }
 $__b2 = $wset;
 $__i1 = 32;
 while(1) {
  $10 = $__i1;
  $11 = ($10|0)!=(0);
  if (!($11)) {
   break;
  }
  $12 = $__b2;
  $13 = ((($12)) + 4|0);
  $__b2 = $13;
  HEAP32[$12>>2] = 0;
  $14 = $__i1;
  $15 = (($14) + -1)|0;
  $__i1 = $15;
 }
 $16 = $1;
 _lua_settop($16,3);
 $17 = $1;
 _lua_createtable($17,0,0);
 $18 = $1;
 $19 = (_lua_gettop($18)|0);
 $itab = $19;
 $20 = $1;
 _lua_createtable($20,0,0);
 $21 = $1;
 $22 = (_lua_gettop($21)|0);
 $rtab = $22;
 $23 = $1;
 _lua_createtable($23,0,0);
 $24 = $1;
 $25 = (_lua_gettop($24)|0);
 $wtab = $25;
 $26 = $1;
 $27 = $itab;
 $28 = (_collect_fd($26,1,-1,$27,$rset)|0);
 $max_fd = $28;
 $29 = $1;
 $30 = $rtab;
 $31 = (_check_dirty($29,1,$30,$rset)|0);
 $ndirty = $31;
 $32 = $ndirty;
 $33 = ($32|0)>(0);
 $34 = $t;
 $35 = $33 ? 0.0 : $34;
 $t = $35;
 $36 = $t;
 _timeout_init($tm,$36,-1.0);
 (_timeout_markstart($tm)|0);
 $37 = $1;
 $38 = $max_fd;
 $39 = $itab;
 $40 = (_collect_fd($37,2,$38,$39,$wset)|0);
 $max_fd = $40;
 $41 = $max_fd;
 $42 = (($41) + 1)|0;
 $43 = (_socket_select($42,$rset,$wset,0,$tm)|0);
 $ret = $43;
 $44 = $ret;
 $45 = ($44|0)>(0);
 $46 = $ndirty;
 $47 = ($46|0)>(0);
 $or$cond = $45 | $47;
 if ($or$cond) {
  $48 = $1;
  $49 = $max_fd;
  $50 = (($49) + 1)|0;
  $51 = $itab;
  $52 = $rtab;
  $53 = $ndirty;
  _return_fd($48,$rset,$50,$51,$52,$53);
  $54 = $1;
  $55 = $max_fd;
  $56 = (($55) + 1)|0;
  $57 = $itab;
  $58 = $wtab;
  _return_fd($54,$wset,$56,$57,$58,0);
  $59 = $1;
  $60 = $rtab;
  _make_assoc($59,$60);
  $61 = $1;
  $62 = $wtab;
  _make_assoc($61,$62);
  $0 = 2;
  $66 = $0;
  STACKTOP = sp;return ($66|0);
 }
 $63 = $ret;
 $64 = ($63|0)==(0);
 $65 = $1;
 if ($64) {
  _lua_pushstring($65,506810);
  $0 = 3;
  $66 = $0;
  STACKTOP = sp;return ($66|0);
 } else {
  _lua_pushstring($65,520240);
  $0 = 3;
  $66 = $0;
  STACKTOP = sp;return ($66|0);
 }
 return (0)|0;
}
function _collect_fd($L,$tab,$max_fd,$itab,$set) {
 $L = $L|0;
 $tab = $tab|0;
 $max_fd = $max_fd|0;
 $itab = $itab|0;
 $set = $set|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fd = 0, $i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $L;
 $2 = $tab;
 $3 = $max_fd;
 $4 = $itab;
 $5 = $set;
 $i = 1;
 $6 = $1;
 $7 = $2;
 $8 = (_lua_type($6,$7)|0);
 $9 = ($8|0)==(0);
 if ($9) {
  $10 = $3;
  $0 = $10;
  $48 = $0;
  STACKTOP = sp;return ($48|0);
 }
 while(1) {
  $11 = $1;
  $12 = $i;
  $13 = (+($12|0));
  _lua_pushnumber($11,$13);
  $14 = $1;
  $15 = $2;
  _lua_gettable($14,$15);
  $16 = $1;
  $17 = (_lua_type($16,-1)|0);
  $18 = ($17|0)==(0);
  $19 = $1;
  if ($18) {
   break;
  }
  $21 = (_getfd($19)|0);
  $fd = $21;
  $22 = $fd;
  $23 = ($22|0)!=(-1);
  if ($23) {
   $24 = $fd;
   $25 = (($24>>>0) % 32)&-1;
   $26 = 1 << $25;
   $27 = $fd;
   $28 = (($27>>>0) / 32)&-1;
   $29 = $5;
   $30 = (($29) + ($28<<2)|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = $31 | $26;
   HEAP32[$30>>2] = $32;
   $33 = $3;
   $34 = ($33|0)==(-1);
   if ($34) {
    label = 8;
   } else {
    $35 = $3;
    $36 = $fd;
    $37 = ($35|0)<($36|0);
    if ($37) {
     label = 8;
    }
   }
   if ((label|0) == 8) {
    label = 0;
    $38 = $fd;
    $3 = $38;
   }
   $39 = $1;
   $40 = $fd;
   $41 = (+($40|0));
   _lua_pushnumber($39,$41);
   $42 = $1;
   _lua_pushvalue($42,-2);
   $43 = $1;
   $44 = $4;
   _lua_settable($43,$44);
  }
  $45 = $1;
  _lua_settop($45,-2);
  $46 = $i;
  $47 = (($46) + 1)|0;
  $i = $47;
 }
 _lua_settop($19,-2);
 $20 = $3;
 $0 = $20;
 $48 = $0;
 STACKTOP = sp;return ($48|0);
}
function _getfd($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fd = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $fd = -1;
 $1 = $0;
 _lua_pushstring($1,364125);
 $2 = $0;
 _lua_gettable($2,-2);
 $3 = $0;
 $4 = (_lua_type($3,-1)|0);
 $5 = ($4|0)==(0);
 if (!($5)) {
  $6 = $0;
  _lua_pushvalue($6,-2);
  $7 = $0;
  _lua_call($7,1,1);
  $8 = $0;
  $9 = (_lua_isnumber($8,-1)|0);
  $10 = ($9|0)!=(0);
  if ($10) {
   $11 = $0;
   $12 = (+_lua_tonumber($11,-1));
   $13 = (~~(($12)));
   $fd = $13;
  }
 }
 $14 = $0;
 _lua_settop($14,-2);
 $15 = $fd;
 STACKTOP = sp;return ($15|0);
}
function _check_dirty($L,$tab,$dtab,$set) {
 $L = $L|0;
 $tab = $tab|0;
 $dtab = $dtab|0;
 $set = $set|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fd = 0, $i = 0, $ndirty = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $L;
 $2 = $tab;
 $3 = $dtab;
 $4 = $set;
 $ndirty = 0;
 $i = 1;
 $5 = $1;
 $6 = $2;
 $7 = (_lua_type($5,$6)|0);
 $8 = ($7|0)==(0);
 if ($8) {
  $0 = 0;
  $45 = $0;
  STACKTOP = sp;return ($45|0);
 }
 while(1) {
  $9 = $1;
  $10 = $i;
  $11 = (+($10|0));
  _lua_pushnumber($9,$11);
  $12 = $1;
  $13 = $2;
  _lua_gettable($12,$13);
  $14 = $1;
  $15 = (_lua_type($14,-1)|0);
  $16 = ($15|0)==(0);
  $17 = $1;
  if ($16) {
   break;
  }
  $19 = (_getfd($17)|0);
  $fd = $19;
  $20 = $fd;
  $21 = ($20|0)!=(-1);
  if ($21) {
   $22 = $1;
   $23 = (_dirty($22)|0);
   $24 = ($23|0)!=(0);
   if ($24) {
    $25 = $1;
    $26 = $ndirty;
    $27 = (($26) + 1)|0;
    $ndirty = $27;
    $28 = (+($27|0));
    _lua_pushnumber($25,$28);
    $29 = $1;
    _lua_pushvalue($29,-2);
    $30 = $1;
    $31 = $3;
    _lua_settable($30,$31);
    $32 = $fd;
    $33 = (($32>>>0) % 32)&-1;
    $34 = 1 << $33;
    $35 = $34 ^ -1;
    $36 = $fd;
    $37 = (($36>>>0) / 32)&-1;
    $38 = $4;
    $39 = (($38) + ($37<<2)|0);
    $40 = HEAP32[$39>>2]|0;
    $41 = $40 & $35;
    HEAP32[$39>>2] = $41;
   }
  }
  $42 = $1;
  _lua_settop($42,-2);
  $43 = $i;
  $44 = (($43) + 1)|0;
  $i = $44;
 }
 _lua_settop($17,-2);
 $18 = $ndirty;
 $0 = $18;
 $45 = $0;
 STACKTOP = sp;return ($45|0);
}
function _dirty($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $is = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $is = 0;
 $1 = $0;
 _lua_pushstring($1,364119);
 $2 = $0;
 _lua_gettable($2,-2);
 $3 = $0;
 $4 = (_lua_type($3,-1)|0);
 $5 = ($4|0)==(0);
 if (!($5)) {
  $6 = $0;
  _lua_pushvalue($6,-2);
  $7 = $0;
  _lua_call($7,1,1);
  $8 = $0;
  $9 = (_lua_toboolean($8,-1)|0);
  $is = $9;
 }
 $10 = $0;
 _lua_settop($10,-2);
 $11 = $is;
 STACKTOP = sp;return ($11|0);
}
function _return_fd($L,$set,$max_fd,$itab,$tab,$start) {
 $L = $L|0;
 $set = $set|0;
 $max_fd = $max_fd|0;
 $itab = $itab|0;
 $tab = $tab|0;
 $start = $start|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0.0, $23 = 0, $24 = 0, $25 = 0.0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fd = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $set;
 $2 = $max_fd;
 $3 = $itab;
 $4 = $tab;
 $5 = $start;
 $fd = 0;
 while(1) {
  $6 = $fd;
  $7 = $2;
  $8 = ($6|0)<($7|0);
  if (!($8)) {
   break;
  }
  $9 = $fd;
  $10 = (($9>>>0) / 32)&-1;
  $11 = $1;
  $12 = (($11) + ($10<<2)|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = $fd;
  $15 = (($14>>>0) % 32)&-1;
  $16 = 1 << $15;
  $17 = $13 & $16;
  $18 = ($17|0)!=(0);
  if ($18) {
   $19 = $0;
   $20 = $5;
   $21 = (($20) + 1)|0;
   $5 = $21;
   $22 = (+($21|0));
   _lua_pushnumber($19,$22);
   $23 = $0;
   $24 = $fd;
   $25 = (+($24|0));
   _lua_pushnumber($23,$25);
   $26 = $0;
   $27 = $3;
   _lua_gettable($26,$27);
   $28 = $0;
   $29 = $4;
   _lua_settable($28,$29);
  }
  $30 = $fd;
  $31 = (($30) + 1)|0;
  $fd = $31;
 }
 STACKTOP = sp;return;
}
function _make_assoc($L,$tab) {
 $L = $L|0;
 $tab = $tab|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0, $atab = 0, $i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $tab;
 $i = 1;
 $2 = $0;
 _lua_createtable($2,0,0);
 $3 = $0;
 $4 = (_lua_gettop($3)|0);
 $atab = $4;
 while(1) {
  $5 = $0;
  $6 = $i;
  $7 = (+($6|0));
  _lua_pushnumber($5,$7);
  $8 = $0;
  $9 = $1;
  _lua_gettable($8,$9);
  $10 = $0;
  $11 = (_lua_type($10,-1)|0);
  $12 = ($11|0)==(0);
  $13 = $0;
  if ($12) {
   break;
  }
  $14 = $i;
  $15 = (+($14|0));
  _lua_pushnumber($13,$15);
  $16 = $0;
  _lua_pushvalue($16,-2);
  $17 = $0;
  $18 = $atab;
  _lua_settable($17,$18);
  $19 = $0;
  $20 = $i;
  $21 = (+($20|0));
  _lua_pushnumber($19,$21);
  $22 = $0;
  $23 = $atab;
  _lua_settable($22,$23);
  $24 = $i;
  $25 = (($24) + 1)|0;
  $i = $25;
 }
 _lua_settop($13,-2);
 STACKTOP = sp;return;
}
function _tcp_open($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 _auxiliar_newclass($1,363905,21356);
 $2 = $0;
 _auxiliar_newclass($2,363945,21356);
 $3 = $0;
 _auxiliar_newclass($3,363933,21356);
 $4 = $0;
 _auxiliar_add2group($4,363905,363917);
 $5 = $0;
 _auxiliar_add2group($5,363945,363917);
 $6 = $0;
 _auxiliar_add2group($6,363933,363917);
 $7 = $0;
 _luax_register($7,0,21572);
 STACKTOP = sp;return 0;
}
function _meth_close($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $tcp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkgroup($1,363917,1)|0);
 $tcp = $2;
 $3 = $tcp;
 _socket_destroy($3);
 $4 = $0;
 _lua_pushnumber($4,1.0);
 STACKTOP = sp;return 1;
}
function _meth_accept($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $clnt = 0, $err = 0, $server = 0, $sock = 0, $tm = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $sock = sp + 8|0;
 $1 = $L;
 $2 = $1;
 $3 = (_auxiliar_checkclass($2,363933,1)|0);
 $server = $3;
 $4 = $server;
 $5 = ((($4)) + 8248|0);
 $6 = (_timeout_markstart($5)|0);
 $tm = $6;
 $7 = $server;
 $8 = $tm;
 $9 = (_socket_accept($7,$sock,0,0,$8)|0);
 $err = $9;
 $10 = $err;
 $11 = ($10|0)==(0);
 $12 = $1;
 if ($11) {
  $13 = (_lua_newuserdata($12,8272)|0);
  $clnt = $13;
  $14 = $1;
  _auxiliar_setclass($14,363945,-1);
  _socket_setnonblocking($sock);
  $15 = HEAP32[$sock>>2]|0;
  $16 = $clnt;
  HEAP32[$16>>2] = $15;
  $17 = $clnt;
  $18 = ((($17)) + 4|0);
  $19 = $clnt;
  _io_init($18,3882,3883,3884,$19);
  $20 = $clnt;
  $21 = ((($20)) + 8248|0);
  _timeout_init($21,-1.0,-1.0);
  $22 = $clnt;
  $23 = ((($22)) + 24|0);
  $24 = $clnt;
  $25 = ((($24)) + 4|0);
  $26 = $clnt;
  $27 = ((($26)) + 8248|0);
  _buffer_init($23,$25,$27);
  $0 = 1;
  $31 = $0;
  STACKTOP = sp;return ($31|0);
 } else {
  _lua_pushnil($12);
  $28 = $1;
  $29 = $err;
  $30 = (_socket_strerror($29)|0);
  _lua_pushstring($28,$30);
  $0 = 2;
  $31 = $0;
  STACKTOP = sp;return ($31|0);
 }
 return (0)|0;
}
function _meth_bind($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0, $address = 0;
 var $err = 0, $port = 0, $tcp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $L;
 $2 = $1;
 $3 = (_auxiliar_checkclass($2,363905,1)|0);
 $tcp = $3;
 $4 = $1;
 $5 = (_luaL_checklstring($4,2,0)|0);
 $address = $5;
 $6 = $1;
 $7 = (+_luaL_checknumber($6,3));
 $8 = (~~(($7))&65535);
 $port = $8;
 $9 = $tcp;
 $10 = $address;
 $11 = $port;
 $12 = (_inet_trybind($9,$10,$11)|0);
 $err = $12;
 $13 = $err;
 $14 = ($13|0)!=(0|0);
 $15 = $1;
 if ($14) {
  _lua_pushnil($15);
  $16 = $1;
  $17 = $err;
  _lua_pushstring($16,$17);
  $0 = 2;
  $18 = $0;
  STACKTOP = sp;return ($18|0);
 } else {
  _lua_pushnumber($15,1.0);
  $0 = 1;
  $18 = $0;
  STACKTOP = sp;return ($18|0);
 }
 return (0)|0;
}
function _meth_connect($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0.0, $8 = 0, $9 = 0, $address = 0, $err = 0, $port = 0, $tcp = 0, $tm = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $L;
 $2 = $1;
 $3 = (_auxiliar_checkgroup($2,363917,1)|0);
 $tcp = $3;
 $4 = $1;
 $5 = (_luaL_checklstring($4,2,0)|0);
 $address = $5;
 $6 = $1;
 $7 = (+_luaL_checknumber($6,3));
 $8 = (~~(($7))&65535);
 $port = $8;
 $9 = $tcp;
 $10 = ((($9)) + 8248|0);
 $11 = (_timeout_markstart($10)|0);
 $tm = $11;
 $12 = $tcp;
 $13 = $address;
 $14 = $port;
 $15 = $tm;
 $16 = (_inet_tryconnect($12,$13,$14,$15)|0);
 $err = $16;
 $17 = $1;
 _auxiliar_setclass($17,363945,1);
 $18 = $err;
 $19 = ($18|0)!=(0|0);
 $20 = $1;
 if ($19) {
  _lua_pushnil($20);
  $21 = $1;
  $22 = $err;
  _lua_pushstring($21,$22);
  $0 = 2;
  $23 = $0;
  STACKTOP = sp;return ($23|0);
 } else {
  _lua_pushnumber($20,1.0);
  $0 = 1;
  $23 = $0;
  STACKTOP = sp;return ($23|0);
 }
 return (0)|0;
}
function _meth_dirty($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $tcp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkgroup($1,363917,1)|0);
 $tcp = $2;
 $3 = $0;
 $4 = $tcp;
 $5 = ((($4)) + 24|0);
 $6 = (_buffer_isempty($5)|0);
 $7 = ($6|0)!=(0);
 $8 = $7 ^ 1;
 $9 = $8&1;
 _lua_pushboolean($3,$9);
 STACKTOP = sp;return 1;
}
function _meth_getfd($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $tcp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkgroup($1,363917,1)|0);
 $tcp = $2;
 $3 = $0;
 $4 = $tcp;
 $5 = HEAP32[$4>>2]|0;
 $6 = (+($5|0));
 _lua_pushnumber($3,$6);
 STACKTOP = sp;return 1;
}
function _meth_getpeername($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $tcp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkgroup($1,363917,1)|0);
 $tcp = $2;
 $3 = $0;
 $4 = $tcp;
 $5 = (_inet_meth_getpeername($3,$4)|0);
 STACKTOP = sp;return ($5|0);
}
function _meth_getsockname($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $tcp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkgroup($1,363917,1)|0);
 $tcp = $2;
 $3 = $0;
 $4 = $tcp;
 $5 = (_inet_meth_getsockname($3,$4)|0);
 STACKTOP = sp;return ($5|0);
}
function _meth_getstats($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $tcp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkclass($1,363945,1)|0);
 $tcp = $2;
 $3 = $0;
 $4 = $tcp;
 $5 = ((($4)) + 24|0);
 $6 = (_buffer_meth_getstats($3,$5)|0);
 STACKTOP = sp;return ($6|0);
}
function _meth_setstats($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $tcp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkclass($1,363945,1)|0);
 $tcp = $2;
 $3 = $0;
 $4 = $tcp;
 $5 = ((($4)) + 24|0);
 $6 = (_buffer_meth_setstats($3,$5)|0);
 STACKTOP = sp;return ($6|0);
}
function _meth_listen($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $backlog = 0, $err = 0;
 var $tcp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $L;
 $2 = $1;
 $3 = (_auxiliar_checkclass($2,363905,1)|0);
 $tcp = $3;
 $4 = $1;
 $5 = (+_luaL_optnumber($4,2,32.0));
 $6 = (~~(($5)));
 $backlog = $6;
 $7 = $tcp;
 $8 = $backlog;
 $9 = (_socket_listen($7,$8)|0);
 $err = $9;
 $10 = $err;
 $11 = ($10|0)!=(0);
 $12 = $1;
 if ($11) {
  _lua_pushnil($12);
  $13 = $1;
  $14 = $err;
  $15 = (_socket_strerror($14)|0);
  _lua_pushstring($13,$15);
  $0 = 2;
  $17 = $0;
  STACKTOP = sp;return ($17|0);
 } else {
  _auxiliar_setclass($12,363933,1);
  $16 = $1;
  _lua_pushnumber($16,1.0);
  $0 = 1;
  $17 = $0;
  STACKTOP = sp;return ($17|0);
 }
 return (0)|0;
}
function _meth_receive($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $tcp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkclass($1,363945,1)|0);
 $tcp = $2;
 $3 = $0;
 $4 = $tcp;
 $5 = ((($4)) + 24|0);
 $6 = (_buffer_meth_receive($3,$5)|0);
 STACKTOP = sp;return ($6|0);
}
function _meth_send($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $tcp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkclass($1,363945,1)|0);
 $tcp = $2;
 $3 = $0;
 $4 = $tcp;
 $5 = ((($4)) + 24|0);
 $6 = (_buffer_meth_send($3,$5)|0);
 STACKTOP = sp;return ($6|0);
}
function _meth_setfd($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $tcp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkgroup($1,363917,1)|0);
 $tcp = $2;
 $3 = $0;
 $4 = (+_luaL_checknumber($3,2));
 $5 = (~~(($4)));
 $6 = $tcp;
 HEAP32[$6>>2] = $5;
 STACKTOP = sp;return 0;
}
function _meth_setoption($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $tcp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkgroup($1,363917,1)|0);
 $tcp = $2;
 $3 = $0;
 $4 = $tcp;
 $5 = (_opt_meth_setoption($3,21532,$4)|0);
 STACKTOP = sp;return ($5|0);
}
function _meth_settimeout($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $tcp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkgroup($1,363917,1)|0);
 $tcp = $2;
 $3 = $0;
 $4 = $tcp;
 $5 = ((($4)) + 8248|0);
 $6 = (_timeout_meth_settimeout($3,$5)|0);
 STACKTOP = sp;return ($6|0);
}
function _meth_shutdown($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $how = 0, $tcp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $L;
 $2 = $1;
 $3 = (_auxiliar_checkclass($2,363945,1)|0);
 $tcp = $3;
 $4 = $1;
 $5 = (_luaL_optlstring($4,2,364033,0)|0);
 $how = $5;
 $6 = $how;
 $7 = HEAP8[$6>>0]|0;
 $8 = $7 << 24 >> 24;
 switch ($8|0) {
 case 98:  {
  $9 = $how;
  $10 = (_strcmp($9,364033)|0);
  $11 = ($10|0)!=(0);
  if ($11) {
   label = 9;
  } else {
   $12 = $tcp;
   _socket_shutdown($12,2);
   label = 8;
  }
  break;
 }
 case 115:  {
  $13 = $how;
  $14 = (_strcmp($13,364200)|0);
  $15 = ($14|0)!=(0);
  if ($15) {
   label = 9;
  } else {
   $16 = $tcp;
   _socket_shutdown($16,1);
   label = 8;
  }
  break;
 }
 case 114:  {
  $17 = $how;
  $18 = (_strcmp($17,364155)|0);
  $19 = ($18|0)!=(0);
  if ($19) {
   label = 9;
  } else {
   $20 = $tcp;
   _socket_shutdown($20,0);
   label = 8;
  }
  break;
 }
 default: {
  label = 8;
 }
 }
 if ((label|0) == 8) {
  $21 = $1;
  _lua_pushnumber($21,1.0);
  $0 = 1;
  $23 = $0;
  STACKTOP = sp;return ($23|0);
 }
 else if ((label|0) == 9) {
  $22 = $1;
  (_luaL_argerror($22,2,364038)|0);
  $0 = 0;
  $23 = $0;
  STACKTOP = sp;return ($23|0);
 }
 return (0)|0;
}
function _global_create($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $err = 0, $sock = 0, $tcp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $sock = sp + 8|0;
 $1 = $L;
 $2 = (_inet_trycreate($sock,1)|0);
 $err = $2;
 $3 = $err;
 $4 = ($3|0)!=(0|0);
 $5 = $1;
 if ($4) {
  _lua_pushnil($5);
  $21 = $1;
  $22 = $err;
  _lua_pushstring($21,$22);
  $0 = 2;
  $23 = $0;
  STACKTOP = sp;return ($23|0);
 } else {
  $6 = (_lua_newuserdata($5,8272)|0);
  $tcp = $6;
  $7 = $1;
  _auxiliar_setclass($7,363905,-1);
  _socket_setnonblocking($sock);
  $8 = HEAP32[$sock>>2]|0;
  $9 = $tcp;
  HEAP32[$9>>2] = $8;
  $10 = $tcp;
  $11 = ((($10)) + 4|0);
  $12 = $tcp;
  _io_init($11,3882,3883,3884,$12);
  $13 = $tcp;
  $14 = ((($13)) + 8248|0);
  _timeout_init($14,-1.0,-1.0);
  $15 = $tcp;
  $16 = ((($15)) + 24|0);
  $17 = $tcp;
  $18 = ((($17)) + 4|0);
  $19 = $tcp;
  $20 = ((($19)) + 8248|0);
  _buffer_init($16,$18,$20);
  $0 = 1;
  $23 = $0;
  STACKTOP = sp;return ($23|0);
 }
 return (0)|0;
}
function _timeout_init($tm,$block,$total) {
 $tm = $tm|0;
 $block = +$block;
 $total = +$total;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $tm;
 $1 = $block;
 $2 = $total;
 $3 = $1;
 $4 = $0;
 HEAPF64[$4>>3] = $3;
 $5 = $2;
 $6 = $0;
 $7 = ((($6)) + 8|0);
 HEAPF64[$7>>3] = $5;
 STACKTOP = sp;return;
}
function _timeout_gettime() {
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp;
 (_gettimeofday(($v|0),(0|0))|0);
 $0 = HEAP32[$v>>2]|0;
 $1 = (+($0|0));
 $2 = ((($v)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (+($3|0));
 $5 = $4 / 1.0E+6;
 $6 = $1 + $5;
 STACKTOP = sp;return (+$6);
}
function _timeout_getretry($tm) {
 $tm = $tm|0;
 var $0 = 0.0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0, $44 = 0.0;
 var $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $7 = 0.0, $8 = 0;
 var $9 = 0, $t = 0.0, $t1 = 0.0, $t2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $tm;
 $2 = $1;
 $3 = +HEAPF64[$2>>3];
 $4 = $3 < 0.0;
 if ($4) {
  $5 = $1;
  $6 = ((($5)) + 8|0);
  $7 = +HEAPF64[$6>>3];
  $8 = $7 < 0.0;
  if ($8) {
   $0 = -1.0;
   $60 = $0;
   STACKTOP = sp;return (+$60);
  }
 }
 $9 = $1;
 $10 = +HEAPF64[$9>>3];
 $11 = $10 < 0.0;
 $12 = $1;
 $13 = ((($12)) + 8|0);
 $14 = +HEAPF64[$13>>3];
 if ($11) {
  $15 = (+_timeout_gettime());
  $16 = $14 - $15;
  $17 = $1;
  $18 = ((($17)) + 16|0);
  $19 = +HEAPF64[$18>>3];
  $20 = $16 + $19;
  $t = $20;
  $21 = $t;
  $22 = $21 > 0.0;
  $23 = $t;
  $24 = $22 ? $23 : 0.0;
  $0 = $24;
  $60 = $0;
  STACKTOP = sp;return (+$60);
 }
 $25 = $14 < 0.0;
 $26 = $1;
 if ($25) {
  $27 = +HEAPF64[$26>>3];
  $28 = (+_timeout_gettime());
  $29 = $27 - $28;
  $30 = $1;
  $31 = ((($30)) + 16|0);
  $32 = +HEAPF64[$31>>3];
  $33 = $29 + $32;
  $t1 = $33;
  $34 = $t1;
  $35 = $34 > 0.0;
  $36 = $t1;
  $37 = $35 ? $36 : 0.0;
  $0 = $37;
  $60 = $0;
  STACKTOP = sp;return (+$60);
 }
 $38 = ((($26)) + 8|0);
 $39 = +HEAPF64[$38>>3];
 $40 = (+_timeout_gettime());
 $41 = $39 - $40;
 $42 = $1;
 $43 = ((($42)) + 16|0);
 $44 = +HEAPF64[$43>>3];
 $45 = $41 + $44;
 $t2 = $45;
 $46 = $1;
 $47 = +HEAPF64[$46>>3];
 $48 = $t2;
 $49 = $48 > 0.0;
 $50 = $t2;
 $51 = $49 ? $50 : 0.0;
 $52 = $47 < $51;
 if ($52) {
  $53 = $1;
  $54 = +HEAPF64[$53>>3];
  $59 = $54;
 } else {
  $55 = $t2;
  $56 = $55 > 0.0;
  $57 = $t2;
  $58 = $56 ? $57 : 0.0;
  $59 = $58;
 }
 $0 = $59;
 $60 = $0;
 STACKTOP = sp;return (+$60);
}
function _timeout_markstart($tm) {
 $tm = $tm|0;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $tm;
 $1 = (+_timeout_gettime());
 $2 = $0;
 $3 = ((($2)) + 16|0);
 HEAPF64[$3>>3] = $1;
 $4 = $0;
 STACKTOP = sp;return ($4|0);
}
function _timeout_open($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 _luax_register($1,0,21588);
 STACKTOP = sp;return 0;
}
function _timeout_lua_gettime($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (+_timeout_gettime());
 _lua_pushnumber($1,$2);
 STACKTOP = sp;return 1;
}
function _timeout_lua_sleep($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0.0;
 var $7 = 0.0, $8 = 0.0, $9 = 0.0, $n = 0.0, $r = 0, $t = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $t = sp + 16|0;
 $r = sp + 8|0;
 $0 = $L;
 $1 = $0;
 $2 = (+_luaL_checknumber($1,1));
 $n = $2;
 $3 = $n;
 $4 = (~~(($3)));
 HEAP32[$t>>2] = $4;
 $5 = HEAP32[$t>>2]|0;
 $6 = (+($5|0));
 $7 = $n;
 $8 = $7 - $6;
 $n = $8;
 $9 = $n;
 $10 = $9 * 1.0E+9;
 $11 = (~~(($10)));
 $12 = ((($t)) + 4|0);
 HEAP32[$12>>2] = $11;
 $13 = ((($t)) + 4|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ($14|0)>=(1000000000);
 if ($15) {
  $16 = ((($t)) + 4|0);
  HEAP32[$16>>2] = 999999999;
 }
 while(1) {
  $17 = (_nanosleep(($t|0),($r|0))|0);
  $18 = ($17|0)!=(0);
  if (!($18)) {
   break;
  }
  $19 = HEAP32[$r>>2]|0;
  HEAP32[$t>>2] = $19;
  $20 = ((($r)) + 4|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = ((($t)) + 4|0);
  HEAP32[$22>>2] = $21;
 }
 STACKTOP = sp;return 0;
}
function _timeout_meth_settimeout($L,$tm) {
 $L = $L|0;
 $tm = $tm|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, $mode = 0, $t = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $tm;
 $2 = $0;
 $3 = (+_luaL_optnumber($2,2,-1.0));
 $t = $3;
 $4 = $0;
 $5 = (_luaL_optlstring($4,3,610851,0)|0);
 $mode = $5;
 $6 = $mode;
 $7 = HEAP8[$6>>0]|0;
 $8 = $7 << 24 >> 24;
 switch ($8|0) {
 case 98:  {
  $9 = $t;
  $10 = $1;
  HEAPF64[$10>>3] = $9;
  break;
 }
 case 116: case 114:  {
  $11 = $t;
  $12 = $1;
  $13 = ((($12)) + 8|0);
  HEAPF64[$13>>3] = $11;
  break;
 }
 default: {
  $14 = $0;
  (_luaL_argerror($14,3,364074)|0);
 }
 }
 $15 = $0;
 _lua_pushnumber($15,1.0);
 STACKTOP = sp;return 1;
}
function _udp_open($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 _auxiliar_newclass($1,364095,21612);
 $2 = $0;
 _auxiliar_newclass($2,364183,21612);
 $3 = $0;
 _auxiliar_add2group($3,364095,364110);
 $4 = $0;
 _auxiliar_add2group($4,364183,364110);
 $5 = $0;
 _auxiliar_add2group($5,364095,364384);
 $6 = $0;
 _auxiliar_add2group($6,364183,364384);
 $7 = $0;
 _luax_register($7,0,21812);
 STACKTOP = sp;return 0;
}
function _meth_close_2900($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $udp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkgroup($1,364110,1)|0);
 $udp = $2;
 $3 = $udp;
 _socket_destroy($3);
 $4 = $0;
 _lua_pushnumber($4,1.0);
 STACKTOP = sp;return 1;
}
function _meth_dirty_2905($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $udp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkgroup($1,364110,1)|0);
 $udp = $2;
 $3 = $0;
 _lua_pushboolean($3,0);
 STACKTOP = sp;return 1;
}
function _meth_getfd_2907($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $udp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkgroup($1,364110,1)|0);
 $udp = $2;
 $3 = $0;
 $4 = $udp;
 $5 = HEAP32[$4>>2]|0;
 $6 = (+($5|0));
 _lua_pushnumber($3,$6);
 STACKTOP = sp;return 1;
}
function _meth_getpeername_2909($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $udp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkclass($1,364095,1)|0);
 $udp = $2;
 $3 = $0;
 $4 = $udp;
 $5 = (_inet_meth_getpeername($3,$4)|0);
 STACKTOP = sp;return ($5|0);
}
function _meth_getsockname_2911($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $udp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkgroup($1,364110,1)|0);
 $udp = $2;
 $3 = $0;
 $4 = $udp;
 $5 = (_inet_meth_getsockname($3,$4)|0);
 STACKTOP = sp;return ($5|0);
}
function _meth_receive_2913($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buffer = 0, $count = 0, $err = 0, $got = 0, $tm = 0, $udp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 8224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $buffer = sp + 32|0;
 $got = sp + 12|0;
 $1 = $L;
 $2 = $1;
 $3 = (_auxiliar_checkgroup($2,364110,1)|0);
 $udp = $3;
 $4 = $1;
 $5 = (+_luaL_optnumber($4,2,8192.0));
 $6 = (~~(($5))>>>0);
 $count = $6;
 $7 = $udp;
 $8 = ((($7)) + 8|0);
 $tm = $8;
 $9 = $count;
 $10 = ($9>>>0)<(8192);
 $11 = $count;
 $12 = $10 ? $11 : 8192;
 $count = $12;
 $13 = $tm;
 (_timeout_markstart($13)|0);
 $14 = $udp;
 $15 = $count;
 $16 = $tm;
 $17 = (_socket_recv($14,$buffer,$15,$got,$16)|0);
 $err = $17;
 $18 = $err;
 $19 = ($18|0)!=(0);
 $20 = $1;
 if ($19) {
  _lua_pushnil($20);
  $21 = $1;
  $22 = $err;
  $23 = (_udp_strerror($22)|0);
  _lua_pushstring($21,$23);
  $0 = 2;
  $25 = $0;
  STACKTOP = sp;return ($25|0);
 } else {
  $24 = HEAP32[$got>>2]|0;
  _lua_pushlstring($20,$buffer,$24);
  $0 = 1;
  $25 = $0;
  STACKTOP = sp;return ($25|0);
 }
 return (0)|0;
}
function _udp_strerror($err) {
 $err = $err|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $err;
 $2 = $1;
 $3 = ($2|0)==(-2);
 if ($3) {
  $0 = 364163;
 } else {
  $4 = $1;
  $5 = (_socket_strerror($4)|0);
  $0 = $5;
 }
 $6 = $0;
 STACKTOP = sp;return ($6|0);
}
function _meth_receivefrom($L) {
 $L = $L|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $addr = 0, $addr_len = 0, $buffer = 0, $count = 0, $err = 0;
 var $got = 0, $tm = 0, $udp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 8256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 52|0;
 $addr = sp + 24|0;
 $addr_len = sp + 16|0;
 $buffer = sp + 56|0;
 $got = sp + 12|0;
 $1 = $L;
 $2 = $1;
 $3 = (_auxiliar_checkclass($2,364183,1)|0);
 $udp = $3;
 HEAP32[$addr_len>>2] = 16;
 $4 = $1;
 $5 = (+_luaL_optnumber($4,2,8192.0));
 $6 = (~~(($5))>>>0);
 $count = $6;
 $7 = $udp;
 $8 = ((($7)) + 8|0);
 $tm = $8;
 $9 = $tm;
 (_timeout_markstart($9)|0);
 $10 = $count;
 $11 = ($10>>>0)<(8192);
 $12 = $count;
 $13 = $11 ? $12 : 8192;
 $count = $13;
 $14 = $udp;
 $15 = $count;
 $16 = $tm;
 $17 = (_socket_recvfrom($14,$buffer,$15,$got,$addr,$addr_len,$16)|0);
 $err = $17;
 $18 = $err;
 $19 = ($18|0)==(0);
 $20 = $1;
 if ($19) {
  $21 = HEAP32[$got>>2]|0;
  _lua_pushlstring($20,$buffer,$21);
  $22 = $1;
  $23 = ((($addr)) + 4|0);
  ;HEAP32[$$byval_copy>>2]=HEAP32[$23>>2]|0;
  $24 = (_inet_ntoa($$byval_copy)|0);
  _lua_pushstring($22,$24);
  $25 = $1;
  $26 = ((($addr)) + 2|0);
  $27 = HEAP16[$26>>1]|0;
  $28 = (_ntohs($27)|0);
  $29 = (+($28&65535));
  _lua_pushnumber($25,$29);
  $0 = 3;
  $33 = $0;
  STACKTOP = sp;return ($33|0);
 } else {
  _lua_pushnil($20);
  $30 = $1;
  $31 = $err;
  $32 = (_udp_strerror($31)|0);
  _lua_pushstring($30,$32);
  $0 = 2;
  $33 = $0;
  STACKTOP = sp;return ($33|0);
 }
 return (0)|0;
}
function _meth_send_2918($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $count = 0, $data = 0, $err = 0, $sent = 0, $tm = 0, $udp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $count = sp + 12|0;
 $sent = sp + 8|0;
 $1 = $L;
 $2 = $1;
 $3 = (_auxiliar_checkclass($2,364095,1)|0);
 $udp = $3;
 $4 = $udp;
 $5 = ((($4)) + 8|0);
 $tm = $5;
 HEAP32[$sent>>2] = 0;
 $6 = $1;
 $7 = (_luaL_checklstring($6,2,$count)|0);
 $data = $7;
 $8 = $tm;
 (_timeout_markstart($8)|0);
 $9 = $udp;
 $10 = $data;
 $11 = HEAP32[$count>>2]|0;
 $12 = $tm;
 $13 = (_socket_send($9,$10,$11,$sent,$12)|0);
 $err = $13;
 $14 = $err;
 $15 = ($14|0)!=(0);
 $16 = $1;
 if ($15) {
  _lua_pushnil($16);
  $17 = $1;
  $18 = $err;
  $19 = (_udp_strerror($18)|0);
  _lua_pushstring($17,$19);
  $0 = 2;
  $22 = $0;
  STACKTOP = sp;return ($22|0);
 } else {
  $20 = HEAP32[$sent>>2]|0;
  $21 = (+($20>>>0));
  _lua_pushnumber($16,$21);
  $0 = 1;
  $22 = $0;
  STACKTOP = sp;return ($22|0);
 }
 return (0)|0;
}
function _meth_sendto($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0.0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, $addr = 0, $count = 0, $data = 0, $err = 0;
 var $ip = 0, $port = 0, $sent = 0, $tm = 0, $udp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $count = sp + 40|0;
 $sent = sp + 36|0;
 $addr = sp + 8|0;
 $1 = $L;
 $2 = $1;
 $3 = (_auxiliar_checkclass($2,364183,1)|0);
 $udp = $3;
 HEAP32[$sent>>2] = 0;
 $4 = $1;
 $5 = (_luaL_checklstring($4,2,$count)|0);
 $data = $5;
 $6 = $1;
 $7 = (_luaL_checklstring($6,3,0)|0);
 $ip = $7;
 $8 = $1;
 $9 = (+_luaL_checknumber($8,4));
 $10 = (~~(($9))&65535);
 $port = $10;
 $11 = $udp;
 $12 = ((($11)) + 8|0);
 $tm = $12;
 ;HEAP32[$addr>>2]=0|0;HEAP32[$addr+4>>2]=0|0;HEAP32[$addr+8>>2]=0|0;HEAP32[$addr+12>>2]=0|0;
 $13 = $ip;
 $14 = ((($addr)) + 4|0);
 $15 = (_inet_aton($13,$14)|0);
 $16 = ($15|0)!=(0);
 if (!($16)) {
  $17 = $1;
  (_luaL_argerror($17,3,364212)|0);
 }
 HEAP16[$addr>>1] = 2;
 $18 = $port;
 $19 = (_htons($18)|0);
 $20 = ((($addr)) + 2|0);
 HEAP16[$20>>1] = $19;
 $21 = $tm;
 (_timeout_markstart($21)|0);
 $22 = $udp;
 $23 = $data;
 $24 = HEAP32[$count>>2]|0;
 $25 = $tm;
 $26 = (_socket_sendto($22,$23,$24,$sent,$addr,16,$25)|0);
 $err = $26;
 $27 = $err;
 $28 = ($27|0)!=(0);
 $29 = $1;
 if ($28) {
  _lua_pushnil($29);
  $30 = $1;
  $31 = $err;
  $32 = (_udp_strerror($31)|0);
  _lua_pushstring($30,$32);
  $0 = 2;
  $35 = $0;
  STACKTOP = sp;return ($35|0);
 } else {
  $33 = HEAP32[$sent>>2]|0;
  $34 = (+($33>>>0));
  _lua_pushnumber($29,$34);
  $0 = 1;
  $35 = $0;
  STACKTOP = sp;return ($35|0);
 }
 return (0)|0;
}
function _meth_setfd_2922($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $udp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkgroup($1,364110,1)|0);
 $udp = $2;
 $3 = $0;
 $4 = (+_luaL_checknumber($3,2));
 $5 = (~~(($4)));
 $6 = $udp;
 HEAP32[$6>>2] = $5;
 STACKTOP = sp;return 0;
}
function _meth_setoption_2924($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $udp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkgroup($1,364110,1)|0);
 $udp = $2;
 $3 = $0;
 $4 = $udp;
 $5 = (_opt_meth_setoption($3,21748,$4)|0);
 STACKTOP = sp;return ($5|0);
}
function _meth_setpeername($L) {
 $L = $L|0;
 var $$sink1 = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $address = 0, $connecting = 0, $err = 0, $port = 0, $tm = 0, $udp = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $L;
 $2 = $1;
 $3 = (_auxiliar_checkgroup($2,364110,1)|0);
 $udp = $3;
 $4 = $udp;
 $5 = ((($4)) + 8|0);
 $tm = $5;
 $6 = $1;
 $7 = (_luaL_checklstring($6,2,0)|0);
 $address = $7;
 $8 = $address;
 $9 = (_strcmp($8,506969)|0);
 $connecting = $9;
 $10 = $connecting;
 $11 = ($10|0)!=(0);
 $12 = $1;
 if ($11) {
  $13 = (+_luaL_checknumber($12,3));
  $$sink1 = $13;
 } else {
  $14 = (+_luaL_optnumber($12,3,0.0));
  $$sink1 = $14;
 }
 $15 = (~~(($$sink1))&65535);
 $16 = $15&65535;
 $17 = $16&65535;
 $port = $17;
 $18 = $udp;
 $19 = $address;
 $20 = $port;
 $21 = $tm;
 $22 = (_inet_tryconnect($18,$19,$20,$21)|0);
 $err = $22;
 $23 = $err;
 $24 = ($23|0)!=(0|0);
 if ($24) {
  $25 = $1;
  _lua_pushnil($25);
  $26 = $1;
  $27 = $err;
  _lua_pushstring($26,$27);
  $0 = 2;
  $32 = $0;
  STACKTOP = sp;return ($32|0);
 }
 $28 = $connecting;
 $29 = ($28|0)!=(0);
 $30 = $1;
 if ($29) {
  _auxiliar_setclass($30,364095,1);
 } else {
  _auxiliar_setclass($30,364183,1);
 }
 $31 = $1;
 _lua_pushnumber($31,1.0);
 $0 = 1;
 $32 = $0;
 STACKTOP = sp;return ($32|0);
}
function _meth_setsockname($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0, $address = 0;
 var $err = 0, $port = 0, $udp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $L;
 $2 = $1;
 $3 = (_auxiliar_checkclass($2,364183,1)|0);
 $udp = $3;
 $4 = $1;
 $5 = (_luaL_checklstring($4,2,0)|0);
 $address = $5;
 $6 = $1;
 $7 = (+_luaL_checknumber($6,3));
 $8 = (~~(($7))&65535);
 $port = $8;
 $9 = $udp;
 $10 = $address;
 $11 = $port;
 $12 = (_inet_trybind($9,$10,$11)|0);
 $err = $12;
 $13 = $err;
 $14 = ($13|0)!=(0|0);
 $15 = $1;
 if ($14) {
  _lua_pushnil($15);
  $16 = $1;
  $17 = $err;
  _lua_pushstring($16,$17);
  $0 = 2;
  $18 = $0;
  STACKTOP = sp;return ($18|0);
 } else {
  _lua_pushnumber($15,1.0);
  $0 = 1;
  $18 = $0;
  STACKTOP = sp;return ($18|0);
 }
 return (0)|0;
}
function _meth_settimeout_2937($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $udp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 $2 = (_auxiliar_checkgroup($1,364110,1)|0);
 $udp = $2;
 $3 = $0;
 $4 = $udp;
 $5 = ((($4)) + 8|0);
 $6 = (_timeout_meth_settimeout($3,$5)|0);
 STACKTOP = sp;return ($6|0);
}
function _global_create_2941($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $err = 0, $sock = 0, $udp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $sock = sp + 8|0;
 $1 = $L;
 $2 = (_inet_trycreate($sock,2)|0);
 $err = $2;
 $3 = $err;
 $4 = ($3|0)!=(0|0);
 $5 = $1;
 if ($4) {
  _lua_pushnil($5);
  $12 = $1;
  $13 = $err;
  _lua_pushstring($12,$13);
  $0 = 2;
  $14 = $0;
  STACKTOP = sp;return ($14|0);
 } else {
  $6 = (_lua_newuserdata($5,32)|0);
  $udp = $6;
  $7 = $1;
  _auxiliar_setclass($7,364183,-1);
  _socket_setnonblocking($sock);
  $8 = HEAP32[$sock>>2]|0;
  $9 = $udp;
  HEAP32[$9>>2] = $8;
  $10 = $udp;
  $11 = ((($10)) + 8|0);
  _timeout_init($11,-1.0,-1.0);
  $0 = 1;
  $14 = $0;
  STACKTOP = sp;return ($14|0);
 }
 return (0)|0;
}
function _socket_waitfd($ps,$sw,$tm) {
 $ps = $ps|0;
 $sw = $sw|0;
 $tm = $tm|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0, $49 = 0.0, $5 = 0.0, $50 = 0, $51 = 0.0, $52 = 0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $__b = 0, $__b2 = 0, $__i = 0, $__i1 = 0, $ret = 0, $rfds = 0, $rp = 0, $t = 0.0, $tp = 0, $tv = 0;
 var $wfds = 0, $wp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 336|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $rfds = sp + 176|0;
 $wfds = sp + 48|0;
 $tv = sp + 32|0;
 $1 = $ps;
 $2 = $sw;
 $3 = $tm;
 $4 = $3;
 $5 = +HEAPF64[$4>>3];
 $6 = $5 == 0.0;
 if ($6) {
  $0 = -1;
  $89 = $0;
  STACKTOP = sp;return ($89|0);
 }
 while(1) {
  $wp = 0;
  $rp = 0;
  $7 = $2;
  $8 = $7 & 1;
  $9 = ($8|0)!=(0);
  if ($9) {
   $__b = $rfds;
   $__i = 32;
   while(1) {
    $10 = $__i;
    $11 = ($10|0)!=(0);
    if (!($11)) {
     break;
    }
    $12 = $__b;
    $13 = ((($12)) + 4|0);
    $__b = $13;
    HEAP32[$12>>2] = 0;
    $14 = $__i;
    $15 = (($14) + -1)|0;
    $__i = $15;
   }
   $16 = $1;
   $17 = HEAP32[$16>>2]|0;
   $18 = (($17>>>0) % 32)&-1;
   $19 = 1 << $18;
   $20 = $1;
   $21 = HEAP32[$20>>2]|0;
   $22 = (($21>>>0) / 32)&-1;
   $23 = (($rfds) + ($22<<2)|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = $24 | $19;
   HEAP32[$23>>2] = $25;
   $rp = $rfds;
  }
  $26 = $2;
  $27 = $26 & 2;
  $28 = ($27|0)!=(0);
  if ($28) {
   $__b2 = $wfds;
   $__i1 = 32;
   while(1) {
    $29 = $__i1;
    $30 = ($29|0)!=(0);
    if (!($30)) {
     break;
    }
    $31 = $__b2;
    $32 = ((($31)) + 4|0);
    $__b2 = $32;
    HEAP32[$31>>2] = 0;
    $33 = $__i1;
    $34 = (($33) + -1)|0;
    $__i1 = $34;
   }
   $35 = $1;
   $36 = HEAP32[$35>>2]|0;
   $37 = (($36>>>0) % 32)&-1;
   $38 = 1 << $37;
   $39 = $1;
   $40 = HEAP32[$39>>2]|0;
   $41 = (($40>>>0) / 32)&-1;
   $42 = (($wfds) + ($41<<2)|0);
   $43 = HEAP32[$42>>2]|0;
   $44 = $43 | $38;
   HEAP32[$42>>2] = $44;
   $wp = $wfds;
  }
  $45 = $3;
  $46 = (+_timeout_getretry($45));
  $t = $46;
  $tp = 0;
  $47 = $t;
  $48 = $47 >= 0.0;
  if ($48) {
   $49 = $t;
   $50 = (~~(($49)));
   HEAP32[$tv>>2] = $50;
   $51 = $t;
   $52 = HEAP32[$tv>>2]|0;
   $53 = (+($52|0));
   $54 = $51 - $53;
   $55 = $54 * 1.0E+6;
   $56 = (~~(($55)));
   $57 = ((($tv)) + 4|0);
   HEAP32[$57>>2] = $56;
   $tp = $tv;
  }
  $58 = $1;
  $59 = HEAP32[$58>>2]|0;
  $60 = (($59) + 1)|0;
  $61 = $rp;
  $62 = $wp;
  $63 = $tp;
  $64 = (_select($60,$61,$62,0,$63)|0);
  $ret = $64;
  $65 = $ret;
  $66 = ($65|0)==(-1);
  if (!($66)) {
   break;
  }
  $67 = (___errno_location()|0);
  $68 = HEAP32[$67>>2]|0;
  $69 = ($68|0)==(4);
  if (!($69)) {
   break;
  }
 }
 $70 = $ret;
 $71 = ($70|0)==(-1);
 if ($71) {
  $72 = (___errno_location()|0);
  $73 = HEAP32[$72>>2]|0;
  $0 = $73;
  $89 = $0;
  STACKTOP = sp;return ($89|0);
 }
 $74 = $ret;
 $75 = ($74|0)==(0);
 if ($75) {
  $0 = -1;
  $89 = $0;
  STACKTOP = sp;return ($89|0);
 }
 $76 = $2;
 $77 = ($76|0)==(3);
 if ($77) {
  $78 = $1;
  $79 = HEAP32[$78>>2]|0;
  $80 = (($79>>>0) / 32)&-1;
  $81 = (($rfds) + ($80<<2)|0);
  $82 = HEAP32[$81>>2]|0;
  $83 = $1;
  $84 = HEAP32[$83>>2]|0;
  $85 = (($84>>>0) % 32)&-1;
  $86 = 1 << $85;
  $87 = $82 & $86;
  $88 = ($87|0)!=(0);
  if ($88) {
   $0 = -2;
   $89 = $0;
   STACKTOP = sp;return ($89|0);
  }
 }
 $0 = 0;
 $89 = $0;
 STACKTOP = sp;return ($89|0);
}
function _socket_open() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 (_signal(13,((1)|0))|0);
 return 1;
}
function _socket_close() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1;
}
function _socket_destroy($ps) {
 $ps = $ps|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $ps;
 $1 = $0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)!=(-1);
 if (!($3)) {
  STACKTOP = sp;return;
 }
 $4 = $0;
 _socket_setblocking($4);
 $5 = $0;
 $6 = HEAP32[$5>>2]|0;
 (_close($6)|0);
 $7 = $0;
 HEAP32[$7>>2] = -1;
 STACKTOP = sp;return;
}
function _socket_setblocking($ps) {
 $ps = $ps|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $flags = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = $ps;
 $1 = $0;
 $2 = HEAP32[$1>>2]|0;
 HEAP32[$vararg_buffer>>2] = 0;
 $3 = (_fcntl($2,3,$vararg_buffer)|0);
 $flags = $3;
 $4 = $flags;
 $5 = $4 & -2049;
 $flags = $5;
 $6 = $0;
 $7 = HEAP32[$6>>2]|0;
 $8 = $flags;
 HEAP32[$vararg_buffer1>>2] = $8;
 (_fcntl($7,4,$vararg_buffer1)|0);
 STACKTOP = sp;return;
}
function _socket_select($n,$rfds,$wfds,$efds,$tm) {
 $n = $n|0;
 $rfds = $rfds|0;
 $wfds = $wfds|0;
 $efds = $efds|0;
 $tm = $tm|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0, $9 = 0.0, $ret = 0, $t = 0.0, $tv = 0, $tv$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tv = sp + 8|0;
 $0 = $n;
 $1 = $rfds;
 $2 = $wfds;
 $3 = $efds;
 $4 = $tm;
 while(1) {
  $5 = $4;
  $6 = (+_timeout_getretry($5));
  $t = $6;
  $7 = $t;
  $8 = (~~(($7)));
  HEAP32[$tv>>2] = $8;
  $9 = $t;
  $10 = HEAP32[$tv>>2]|0;
  $11 = (+($10|0));
  $12 = $9 - $11;
  $13 = $12 * 1.0E+6;
  $14 = (~~(($13)));
  $15 = ((($tv)) + 4|0);
  HEAP32[$15>>2] = $14;
  $16 = $0;
  $17 = $1;
  $18 = $2;
  $19 = $3;
  $20 = $t;
  $21 = $20 >= 0.0;
  $tv$ = $21 ? $tv : 0;
  $22 = (_select($16,$17,$18,$19,$tv$)|0);
  $ret = $22;
  $23 = $ret;
  $24 = ($23|0)<(0);
  if (!($24)) {
   label = 4;
   break;
  }
  $25 = (___errno_location()|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = ($26|0)==(4);
  if (!($27)) {
   label = 4;
   break;
  }
 }
 if ((label|0) == 4) {
  $28 = $ret;
  STACKTOP = sp;return ($28|0);
 }
 return (0)|0;
}
function _socket_create($ps,$domain,$type,$protocol) {
 $ps = $ps|0;
 $domain = $domain|0;
 $type = $type|0;
 $protocol = $protocol|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $ps;
 $2 = $domain;
 $3 = $type;
 $4 = $protocol;
 $5 = $2;
 $6 = $3;
 $7 = $4;
 $8 = (_socket($5,$6,$7)|0);
 $9 = $1;
 HEAP32[$9>>2] = $8;
 $10 = $1;
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)!=(-1);
 if ($12) {
  $0 = 0;
  $15 = $0;
  STACKTOP = sp;return ($15|0);
 } else {
  $13 = (___errno_location()|0);
  $14 = HEAP32[$13>>2]|0;
  $0 = $14;
  $15 = $0;
  STACKTOP = sp;return ($15|0);
 }
 return (0)|0;
}
function _socket_bind($ps,$addr,$len) {
 $ps = $ps|0;
 $addr = $addr|0;
 $len = $len|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $err = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $ps;
 $1 = $addr;
 $2 = $len;
 $err = 0;
 $3 = $0;
 _socket_setblocking($3);
 $4 = $0;
 $5 = HEAP32[$4>>2]|0;
 $6 = $1;
 $7 = $2;
 $8 = (_bind($5,$6,$7)|0);
 $9 = ($8|0)<(0);
 if ($9) {
  $10 = (___errno_location()|0);
  $11 = HEAP32[$10>>2]|0;
  $err = $11;
 }
 $12 = $0;
 _socket_setnonblocking($12);
 $13 = $err;
 STACKTOP = sp;return ($13|0);
}
function _socket_setnonblocking($ps) {
 $ps = $ps|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $flags = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = $ps;
 $1 = $0;
 $2 = HEAP32[$1>>2]|0;
 HEAP32[$vararg_buffer>>2] = 0;
 $3 = (_fcntl($2,3,$vararg_buffer)|0);
 $flags = $3;
 $4 = $flags;
 $5 = $4 | 2048;
 $flags = $5;
 $6 = $0;
 $7 = HEAP32[$6>>2]|0;
 $8 = $flags;
 HEAP32[$vararg_buffer1>>2] = $8;
 (_fcntl($7,4,$vararg_buffer1)|0);
 STACKTOP = sp;return;
}
function _socket_listen($ps,$backlog) {
 $ps = $ps|0;
 $backlog = $backlog|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $err = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $ps;
 $1 = $backlog;
 $err = 0;
 $2 = $0;
 _socket_setblocking($2);
 $3 = $0;
 $4 = HEAP32[$3>>2]|0;
 $5 = $1;
 $6 = (_listen($4,$5)|0);
 $7 = ($6|0)!=(0);
 if ($7) {
  $8 = (___errno_location()|0);
  $9 = HEAP32[$8>>2]|0;
  $err = $9;
 }
 $10 = $0;
 _socket_setnonblocking($10);
 $11 = $err;
 STACKTOP = sp;return ($11|0);
}
function _socket_shutdown($ps,$how) {
 $ps = $ps|0;
 $how = $how|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $ps;
 $1 = $how;
 $2 = $0;
 _socket_setblocking($2);
 $3 = $0;
 $4 = HEAP32[$3>>2]|0;
 $5 = $1;
 (_shutdown($4,$5)|0);
 $6 = $0;
 _socket_setnonblocking($6);
 STACKTOP = sp;return;
}
function _socket_connect($ps,$addr,$len,$tm) {
 $ps = $ps|0;
 $addr = $addr|0;
 $len = $len|0;
 $tm = $tm|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $err = 0, $or$cond = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $err = sp;
 $1 = $ps;
 $2 = $addr;
 $3 = $len;
 $4 = $tm;
 $5 = $1;
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(-1);
 do {
  if ($7) {
   $0 = -2;
  } else {
   while(1) {
    $8 = $1;
    $9 = HEAP32[$8>>2]|0;
    $10 = $2;
    $11 = $3;
    $12 = (_connect($9,$10,$11)|0);
    $13 = ($12|0)==(0);
    if ($13) {
     label = 4;
     break;
    }
    $14 = (___errno_location()|0);
    $15 = HEAP32[$14>>2]|0;
    HEAP32[$err>>2] = $15;
    $16 = ($15|0)==(4);
    if (!($16)) {
     break;
    }
   }
   if ((label|0) == 4) {
    $0 = 0;
    break;
   }
   $17 = HEAP32[$err>>2]|0;
   $18 = ($17|0)!=(115);
   $19 = HEAP32[$err>>2]|0;
   $20 = ($19|0)!=(11);
   $or$cond = $18 & $20;
   if ($or$cond) {
    $21 = HEAP32[$err>>2]|0;
    $0 = $21;
    break;
   }
   $22 = $4;
   $23 = +HEAPF64[$22>>3];
   $24 = $23 == 0.0;
   if ($24) {
    $0 = -1;
    break;
   }
   $25 = $1;
   $26 = $4;
   $27 = (_socket_waitfd($25,3,$26)|0);
   HEAP32[$err>>2] = $27;
   $28 = HEAP32[$err>>2]|0;
   $29 = ($28|0)==(-2);
   if (!($29)) {
    $36 = HEAP32[$err>>2]|0;
    $0 = $36;
    break;
   }
   $30 = $1;
   $31 = HEAP32[$30>>2]|0;
   $32 = (_recv($31,$err,0,0)|0);
   $33 = ($32|0)==(0);
   if ($33) {
    $0 = 0;
    break;
   } else {
    $34 = (___errno_location()|0);
    $35 = HEAP32[$34>>2]|0;
    $0 = $35;
    break;
   }
  }
 } while(0);
 $37 = $0;
 STACKTOP = sp;return ($37|0);
}
function _socket_accept($ps,$pa,$addr,$len,$tm) {
 $ps = $ps|0;
 $pa = $pa|0;
 $addr = $addr|0;
 $len = $len|0;
 $tm = $tm|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $daddr = 0, $dlen = 0, $err = 0, $or$cond = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $daddr = sp + 32|0;
 $dlen = sp + 4|0;
 $1 = $ps;
 $2 = $pa;
 $3 = $addr;
 $4 = $len;
 $5 = $tm;
 HEAP32[$dlen>>2] = 16;
 $6 = $1;
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(-1);
 if ($8) {
  $0 = -2;
  $34 = $0;
  STACKTOP = sp;return ($34|0);
 }
 $9 = $3;
 $10 = ($9|0)!=(0|0);
 if (!($10)) {
  $3 = $daddr;
 }
 $11 = $4;
 $12 = ($11|0)!=(0|0);
 if (!($12)) {
  $4 = $dlen;
 }
 while(1) {
  $13 = $1;
  $14 = HEAP32[$13>>2]|0;
  $15 = $3;
  $16 = $4;
  $17 = (_accept($14,$15,$16)|0);
  $18 = $2;
  HEAP32[$18>>2] = $17;
  $19 = ($17|0)!=(-1);
  if ($19) {
   label = 8;
   break;
  }
  $20 = (___errno_location()|0);
  $21 = HEAP32[$20>>2]|0;
  $err = $21;
  $22 = $err;
  $23 = ($22|0)==(4);
  if ($23) {
   continue;
  }
  $24 = $err;
  $25 = ($24|0)!=(11);
  $26 = $err;
  $27 = ($26|0)!=(103);
  $or$cond = $25 & $27;
  if ($or$cond) {
   label = 11;
   break;
  }
  $29 = $1;
  $30 = $5;
  $31 = (_socket_waitfd($29,1,$30)|0);
  $err = $31;
  $32 = ($31|0)!=(0);
  if ($32) {
   label = 13;
   break;
  }
 }
 if ((label|0) == 8) {
  $0 = 0;
  $34 = $0;
  STACKTOP = sp;return ($34|0);
 }
 else if ((label|0) == 11) {
  $28 = $err;
  $0 = $28;
  $34 = $0;
  STACKTOP = sp;return ($34|0);
 }
 else if ((label|0) == 13) {
  $33 = $err;
  $0 = $33;
  $34 = $0;
  STACKTOP = sp;return ($34|0);
 }
 return (0)|0;
}
function _socket_send($ps,$data,$count,$sent,$tm) {
 $ps = $ps|0;
 $data = $data|0;
 $count = $count|0;
 $sent = $sent|0;
 $tm = $tm|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $err = 0, $or$cond = 0, $put = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $ps;
 $2 = $data;
 $3 = $count;
 $4 = $sent;
 $5 = $tm;
 $6 = $4;
 HEAP32[$6>>2] = 0;
 $7 = $1;
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(-1);
 do {
  if ($9) {
   $0 = -2;
  } else {
   while(1) {
    $10 = $1;
    $11 = HEAP32[$10>>2]|0;
    $12 = $2;
    $13 = $3;
    $14 = (_send($11,$12,$13,0)|0);
    $put = $14;
    $15 = $put;
    $16 = ($15|0)>(0);
    if ($16) {
     label = 4;
     break;
    }
    $19 = (___errno_location()|0);
    $20 = HEAP32[$19>>2]|0;
    $err = $20;
    $21 = $put;
    $22 = ($21|0)==(0);
    $23 = $err;
    $24 = ($23|0)==(32);
    $or$cond = $22 | $24;
    if ($or$cond) {
     label = 6;
     break;
    }
    $25 = $err;
    $26 = ($25|0)==(4);
    if ($26) {
     continue;
    }
    $27 = $err;
    $28 = ($27|0)!=(11);
    if ($28) {
     label = 9;
     break;
    }
    $30 = $1;
    $31 = $5;
    $32 = (_socket_waitfd($30,2,$31)|0);
    $err = $32;
    $33 = ($32|0)!=(0);
    if ($33) {
     label = 11;
     break;
    }
   }
   if ((label|0) == 4) {
    $17 = $put;
    $18 = $4;
    HEAP32[$18>>2] = $17;
    $0 = 0;
    break;
   }
   else if ((label|0) == 6) {
    $0 = -2;
    break;
   }
   else if ((label|0) == 9) {
    $29 = $err;
    $0 = $29;
    break;
   }
   else if ((label|0) == 11) {
    $34 = $err;
    $0 = $34;
    break;
   }
  }
 } while(0);
 $35 = $0;
 STACKTOP = sp;return ($35|0);
}
function _socket_sendto($ps,$data,$count,$sent,$addr,$len,$tm) {
 $ps = $ps|0;
 $data = $data|0;
 $count = $count|0;
 $sent = $sent|0;
 $addr = $addr|0;
 $len = $len|0;
 $tm = $tm|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $err = 0, $or$cond = 0, $put = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $ps;
 $2 = $data;
 $3 = $count;
 $4 = $sent;
 $5 = $addr;
 $6 = $len;
 $7 = $tm;
 $8 = $4;
 HEAP32[$8>>2] = 0;
 $9 = $1;
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10|0)==(-1);
 do {
  if ($11) {
   $0 = -2;
  } else {
   while(1) {
    $12 = $1;
    $13 = HEAP32[$12>>2]|0;
    $14 = $2;
    $15 = $3;
    $16 = $5;
    $17 = $6;
    $18 = (_sendto($13,$14,$15,0,$16,$17)|0);
    $put = $18;
    $19 = $put;
    $20 = ($19|0)>(0);
    if ($20) {
     label = 4;
     break;
    }
    $23 = (___errno_location()|0);
    $24 = HEAP32[$23>>2]|0;
    $err = $24;
    $25 = $put;
    $26 = ($25|0)==(0);
    $27 = $err;
    $28 = ($27|0)==(32);
    $or$cond = $26 | $28;
    if ($or$cond) {
     label = 6;
     break;
    }
    $29 = $err;
    $30 = ($29|0)==(4);
    if ($30) {
     continue;
    }
    $31 = $err;
    $32 = ($31|0)!=(11);
    if ($32) {
     label = 9;
     break;
    }
    $34 = $1;
    $35 = $7;
    $36 = (_socket_waitfd($34,2,$35)|0);
    $err = $36;
    $37 = ($36|0)!=(0);
    if ($37) {
     label = 11;
     break;
    }
   }
   if ((label|0) == 4) {
    $21 = $put;
    $22 = $4;
    HEAP32[$22>>2] = $21;
    $0 = 0;
    break;
   }
   else if ((label|0) == 6) {
    $0 = -2;
    break;
   }
   else if ((label|0) == 9) {
    $33 = $err;
    $0 = $33;
    break;
   }
   else if ((label|0) == 11) {
    $38 = $err;
    $0 = $38;
    break;
   }
  }
 } while(0);
 $39 = $0;
 STACKTOP = sp;return ($39|0);
}
function _socket_recv($ps,$data,$count,$got,$tm) {
 $ps = $ps|0;
 $data = $data|0;
 $count = $count|0;
 $got = $got|0;
 $tm = $tm|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $err = 0, $taken = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $ps;
 $2 = $data;
 $3 = $count;
 $4 = $got;
 $5 = $tm;
 $6 = $4;
 HEAP32[$6>>2] = 0;
 $7 = $1;
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(-1);
 do {
  if ($9) {
   $0 = -2;
  } else {
   while(1) {
    $10 = $1;
    $11 = HEAP32[$10>>2]|0;
    $12 = $2;
    $13 = $3;
    $14 = (_recv($11,$12,$13,0)|0);
    $taken = $14;
    $15 = $taken;
    $16 = ($15|0)>(0);
    if ($16) {
     label = 4;
     break;
    }
    $19 = (___errno_location()|0);
    $20 = HEAP32[$19>>2]|0;
    $err = $20;
    $21 = $taken;
    $22 = ($21|0)==(0);
    if ($22) {
     label = 6;
     break;
    }
    $23 = $err;
    $24 = ($23|0)==(4);
    if ($24) {
     continue;
    }
    $25 = $err;
    $26 = ($25|0)!=(11);
    if ($26) {
     label = 9;
     break;
    }
    $28 = $1;
    $29 = $5;
    $30 = (_socket_waitfd($28,1,$29)|0);
    $err = $30;
    $31 = ($30|0)!=(0);
    if ($31) {
     label = 11;
     break;
    }
   }
   if ((label|0) == 4) {
    $17 = $taken;
    $18 = $4;
    HEAP32[$18>>2] = $17;
    $0 = 0;
    break;
   }
   else if ((label|0) == 6) {
    $0 = -2;
    break;
   }
   else if ((label|0) == 9) {
    $27 = $err;
    $0 = $27;
    break;
   }
   else if ((label|0) == 11) {
    $32 = $err;
    $0 = $32;
    break;
   }
  }
 } while(0);
 $33 = $0;
 STACKTOP = sp;return ($33|0);
}
function _socket_recvfrom($ps,$data,$count,$got,$addr,$len,$tm) {
 $ps = $ps|0;
 $data = $data|0;
 $count = $count|0;
 $got = $got|0;
 $addr = $addr|0;
 $len = $len|0;
 $tm = $tm|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $err = 0, $taken = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $ps;
 $2 = $data;
 $3 = $count;
 $4 = $got;
 $5 = $addr;
 $6 = $len;
 $7 = $tm;
 $8 = $4;
 HEAP32[$8>>2] = 0;
 $9 = $1;
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10|0)==(-1);
 do {
  if ($11) {
   $0 = -2;
  } else {
   while(1) {
    $12 = $1;
    $13 = HEAP32[$12>>2]|0;
    $14 = $2;
    $15 = $3;
    $16 = $5;
    $17 = $6;
    $18 = (_recvfrom($13,$14,$15,0,$16,$17)|0);
    $taken = $18;
    $19 = $taken;
    $20 = ($19|0)>(0);
    if ($20) {
     label = 4;
     break;
    }
    $23 = (___errno_location()|0);
    $24 = HEAP32[$23>>2]|0;
    $err = $24;
    $25 = $taken;
    $26 = ($25|0)==(0);
    if ($26) {
     label = 6;
     break;
    }
    $27 = $err;
    $28 = ($27|0)==(4);
    if ($28) {
     continue;
    }
    $29 = $err;
    $30 = ($29|0)!=(11);
    if ($30) {
     label = 9;
     break;
    }
    $32 = $1;
    $33 = $7;
    $34 = (_socket_waitfd($32,1,$33)|0);
    $err = $34;
    $35 = ($34|0)!=(0);
    if ($35) {
     label = 11;
     break;
    }
   }
   if ((label|0) == 4) {
    $21 = $taken;
    $22 = $4;
    HEAP32[$22>>2] = $21;
    $0 = 0;
    break;
   }
   else if ((label|0) == 6) {
    $0 = -2;
    break;
   }
   else if ((label|0) == 9) {
    $31 = $err;
    $0 = $31;
    break;
   }
   else if ((label|0) == 11) {
    $36 = $err;
    $0 = $36;
    break;
   }
  }
 } while(0);
 $37 = $0;
 STACKTOP = sp;return ($37|0);
}
function _socket_gethostbyaddr($addr,$len,$hp) {
 $addr = $addr|0;
 $len = $len|0;
 $hp = $hp|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $addr;
 $2 = $len;
 $3 = $hp;
 $4 = $1;
 $5 = $2;
 $6 = (_gethostbyaddr(($4|0),($5|0),2)|0);
 $7 = $3;
 HEAP32[$7>>2] = $6;
 $8 = $3;
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)!=(0|0);
 do {
  if ($10) {
   $0 = 0;
  } else {
   $11 = (___h_errno_location()|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($12|0)!=(0);
   if ($13) {
    $14 = (___h_errno_location()|0);
    $15 = HEAP32[$14>>2]|0;
    $0 = $15;
    break;
   }
   $16 = (___errno_location()|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ($17|0)!=(0);
   if ($18) {
    $19 = (___errno_location()|0);
    $20 = HEAP32[$19>>2]|0;
    $0 = $20;
    break;
   } else {
    $0 = -3;
    break;
   }
  }
 } while(0);
 $21 = $0;
 STACKTOP = sp;return ($21|0);
}
function _socket_gethostbyname($addr,$hp) {
 $addr = $addr|0;
 $hp = $hp|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $addr;
 $2 = $hp;
 $3 = $1;
 $4 = (_gethostbyname(($3|0))|0);
 $5 = $2;
 HEAP32[$5>>2] = $4;
 $6 = $2;
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)!=(0|0);
 do {
  if ($8) {
   $0 = 0;
  } else {
   $9 = (___h_errno_location()|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($10|0)!=(0);
   if ($11) {
    $12 = (___h_errno_location()|0);
    $13 = HEAP32[$12>>2]|0;
    $0 = $13;
    break;
   }
   $14 = (___errno_location()|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)!=(0);
   if ($16) {
    $17 = (___errno_location()|0);
    $18 = HEAP32[$17>>2]|0;
    $0 = $18;
    break;
   } else {
    $0 = -3;
    break;
   }
  }
 } while(0);
 $19 = $0;
 STACKTOP = sp;return ($19|0);
}
function _socket_hoststrerror($err) {
 $err = $err|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $err;
 $2 = $1;
 $3 = ($2|0)<=(0);
 $4 = $1;
 do {
  if ($3) {
   $5 = (_io_strerror($4)|0);
   $0 = $5;
  } else {
   $cond = ($4|0)==(1);
   if ($cond) {
    $0 = 364401;
    break;
   } else {
    $6 = $1;
    $7 = (_hstrerror($6)|0);
    $0 = $7;
    break;
   }
  }
 } while(0);
 $8 = $0;
 STACKTOP = sp;return ($8|0);
}
function _socket_strerror($err) {
 $err = $err|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $err;
 $2 = $1;
 $3 = ($2|0)<=(0);
 $4 = $1;
 L1: do {
  if ($3) {
   $5 = (_io_strerror($4)|0);
   $0 = $5;
  } else {
   switch ($4|0) {
   case 98:  {
    $0 = 364416;
    break L1;
    break;
   }
   case 106:  {
    $0 = 364439;
    break L1;
    break;
   }
   case 13:  {
    $0 = 364457;
    break L1;
    break;
   }
   case 111:  {
    $0 = 364475;
    break L1;
    break;
   }
   case 103:  {
    $0 = 506730;
    break L1;
    break;
   }
   case 104:  {
    $0 = 506730;
    break L1;
    break;
   }
   case 110:  {
    $0 = 506810;
    break L1;
    break;
   }
   default: {
    $6 = (___errno_location()|0);
    $7 = HEAP32[$6>>2]|0;
    $8 = (_strerror($7)|0);
    $0 = $8;
    break L1;
   }
   }
  }
 } while(0);
 $9 = $0;
 STACKTOP = sp;return ($9|0);
}
function _socket_ioerror($ps,$err) {
 $ps = $ps|0;
 $err = $err|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $ps;
 $1 = $err;
 $2 = $1;
 $3 = (_socket_strerror($2)|0);
 STACKTOP = sp;return ($3|0);
}
function _LZ4_compressBound($isize) {
 $isize = $isize|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $isize;
 $1 = $0;
 $2 = ($1>>>0)>(2113929216);
 if ($2) {
  $8 = 0;
  STACKTOP = sp;return ($8|0);
 }
 $3 = $0;
 $4 = $0;
 $5 = (($4|0) / 255)&-1;
 $6 = (($3) + ($5))|0;
 $7 = (($6) + 16)|0;
 $8 = $7;
 STACKTOP = sp;return ($8|0);
}
function _LZ4_compress_fast_extState($state,$source,$dest,$inputSize,$maxOutputSize,$acceleration) {
 $state = $state|0;
 $source = $source|0;
 $dest = $dest|0;
 $inputSize = $inputSize|0;
 $maxOutputSize = $maxOutputSize|0;
 $acceleration = $acceleration|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0;
 var $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0;
 var $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0;
 var $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0;
 var $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0;
 var $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0;
 var $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0;
 var $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0;
 var $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0;
 var $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0;
 var $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0;
 var $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0;
 var $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0;
 var $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0;
 var $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0, $1267 = 0;
 var $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0, $1285 = 0;
 var $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0, $1302 = 0;
 var $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0, $1320 = 0;
 var $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0, $1339 = 0;
 var $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0, $1357 = 0;
 var $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0, $1375 = 0;
 var $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0, $1393 = 0;
 var $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0, $1410 = 0;
 var $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0, $1429 = 0;
 var $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0, $1447 = 0;
 var $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0, $1465 = 0;
 var $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0, $1483 = 0;
 var $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0, $1500 = 0;
 var $1501 = 0, $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0, $1511 = 0, $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0, $1518 = 0, $1519 = 0;
 var $152 = 0, $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0, $153 = 0, $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0, $1536 = 0, $1537 = 0;
 var $1538 = 0, $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0, $1548 = 0, $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0, $1554 = 0, $1555 = 0;
 var $1556 = 0, $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $1565 = 0, $1566 = 0, $1567 = 0, $1568 = 0, $1569 = 0, $157 = 0, $1570 = 0, $1571 = 0, $1572 = 0, $1573 = 0;
 var $1574 = 0, $1575 = 0, $1576 = 0, $1577 = 0, $1578 = 0, $1579 = 0, $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0, $1583 = 0, $1584 = 0, $1585 = 0, $1586 = 0, $1587 = 0, $1588 = 0, $1589 = 0, $159 = 0, $1590 = 0, $1591 = 0;
 var $1592 = 0, $1593 = 0, $1594 = 0, $1595 = 0, $1596 = 0, $1597 = 0, $1598 = 0, $1599 = 0, $16 = 0, $160 = 0, $1600 = 0, $1601 = 0, $1602 = 0, $1603 = 0, $1604 = 0, $1605 = 0, $1606 = 0, $1607 = 0, $1608 = 0, $1609 = 0;
 var $161 = 0, $1610 = 0, $1611 = 0, $1612 = 0, $1613 = 0, $1614 = 0, $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0, $1619 = 0, $162 = 0, $1620 = 0, $1621 = 0, $1622 = 0, $1623 = 0, $1624 = 0, $1625 = 0, $1626 = 0, $1627 = 0;
 var $1628 = 0, $1629 = 0, $163 = 0, $1630 = 0, $1631 = 0, $1632 = 0, $1633 = 0, $1634 = 0, $1635 = 0, $1636 = 0, $1637 = 0, $1638 = 0, $1639 = 0, $164 = 0, $1640 = 0, $1641 = 0, $1642 = 0, $1643 = 0, $1644 = 0, $1645 = 0;
 var $1646 = 0, $1647 = 0, $1648 = 0, $1649 = 0, $165 = 0, $1650 = 0, $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0, $1655 = 0, $1656 = 0, $1657 = 0, $1658 = 0, $1659 = 0, $166 = 0, $1660 = 0, $1661 = 0, $1662 = 0, $1663 = 0;
 var $1664 = 0, $1665 = 0, $1666 = 0, $1667 = 0, $1668 = 0, $1669 = 0, $167 = 0, $1670 = 0, $1671 = 0, $1672 = 0, $1673 = 0, $1674 = 0, $1675 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0;
 var $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0;
 var $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0;
 var $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0;
 var $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0;
 var $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0;
 var $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0;
 var $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0;
 var $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0;
 var $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0;
 var $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0;
 var $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0;
 var $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0;
 var $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0;
 var $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0;
 var $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0;
 var $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0;
 var $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0;
 var $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0;
 var $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0;
 var $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0;
 var $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0;
 var $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0;
 var $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0;
 var $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0;
 var $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0;
 var $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0;
 var $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0;
 var $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0;
 var $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0;
 var $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0;
 var $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0;
 var $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0;
 var $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0;
 var $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0;
 var $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0;
 var $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0;
 var $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0;
 var $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0;
 var $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0;
 var $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0;
 var $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0;
 var $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $accumulator$i = 0, $accumulator$i29 = 0, $accumulator$i59 = 0, $accumulator$i89 = 0;
 var $anchor$i = 0, $anchor$i39 = 0, $anchor$i69 = 0, $anchor$i9 = 0, $base$i = 0, $base$i3 = 0, $base$i33 = 0, $base$i63 = 0, $dictDelta$i = 0, $dictDelta$i38 = 0, $dictDelta$i68 = 0, $dictDelta$i8 = 0, $dictEnd$i = 0, $dictEnd$i37 = 0, $dictEnd$i67 = 0, $dictEnd$i7 = 0, $dictPtr$i = 0, $dictPtr$i1 = 0, $dictPtr$i31 = 0, $dictPtr$i61 = 0;
 var $dictionary$i = 0, $dictionary$i36 = 0, $dictionary$i6 = 0, $dictionary$i66 = 0, $forwardH$i = 0, $forwardH$i15 = 0, $forwardH$i45 = 0, $forwardH$i75 = 0, $forwardIp$i = 0, $forwardIp$i19 = 0, $forwardIp$i49 = 0, $forwardIp$i79 = 0, $h$i = 0, $h$i22 = 0, $h$i52 = 0, $h$i82 = 0, $iend$i = 0, $iend$i10 = 0, $iend$i40 = 0, $iend$i70 = 0;
 var $ip$i = 0, $ip$i2 = 0, $ip$i32 = 0, $ip$i62 = 0, $lastRun$i = 0, $lastRun$i28 = 0, $lastRun$i58 = 0, $lastRun$i88 = 0, $len$i = 0, $len$i24 = 0, $len$i54 = 0, $len$i84 = 0, $limit$i = 0, $limit$i26 = 0, $limit$i56 = 0, $limit$i86 = 0, $litLength$i = 0, $litLength$i23 = 0, $litLength$i53 = 0, $litLength$i83 = 0;
 var $lowLimit$i = 0, $lowLimit$i34 = 0, $lowLimit$i4 = 0, $lowLimit$i64 = 0, $lowRefLimit$i = 0, $lowRefLimit$i35 = 0, $lowRefLimit$i5 = 0, $lowRefLimit$i65 = 0, $match$i = 0, $match$i17 = 0, $match$i47 = 0, $match$i77 = 0, $matchLength$i = 0, $matchLength$i25 = 0, $matchLength$i55 = 0, $matchLength$i85 = 0, $matchlimit$i = 0, $matchlimit$i12 = 0, $matchlimit$i42 = 0, $matchlimit$i72 = 0;
 var $mflimit$i = 0, $mflimit$i11 = 0, $mflimit$i41 = 0, $mflimit$i71 = 0, $more$i = 0, $more$i27 = 0, $more$i57 = 0, $more$i87 = 0, $olimit$i = 0, $olimit$i14 = 0, $olimit$i44 = 0, $olimit$i74 = 0, $op$i = 0, $op$i13 = 0, $op$i43 = 0, $op$i73 = 0, $or$cond = 0, $or$cond12 = 0, $or$cond4 = 0, $or$cond8 = 0;
 var $refDelta$i = 0, $refDelta$i16 = 0, $refDelta$i46 = 0, $refDelta$i76 = 0, $searchMatchNb$i = 0, $searchMatchNb$i21 = 0, $searchMatchNb$i51 = 0, $searchMatchNb$i81 = 0, $step$i = 0, $step$i20 = 0, $step$i50 = 0, $step$i80 = 0, $token$i = 0, $token$i18 = 0, $token$i48 = 0, $token$i78 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 672|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $45 = $state;
 $46 = $source;
 $47 = $dest;
 $48 = $inputSize;
 $49 = $maxOutputSize;
 $50 = $acceleration;
 $51 = $45;
 _LZ4_resetStream($51);
 $52 = $50;
 $53 = ($52|0)<(1);
 if ($53) {
  $50 = 1;
 }
 $54 = $49;
 $55 = $48;
 $56 = (_LZ4_compressBound($55)|0);
 $57 = ($54|0)>=($56|0);
 $58 = $48;
 $59 = ($58|0)<(65547);
 $60 = $45;
 $61 = $46;
 $62 = $47;
 $63 = $48;
 if ($57) {
  if ($59) {
   $64 = $50;
   $34 = $60;
   $35 = $61;
   $36 = $62;
   $37 = $63;
   $38 = 0;
   $39 = 0;
   $40 = 2;
   $41 = 0;
   $42 = 0;
   $43 = $64;
   $65 = $34;
   $dictPtr$i = $65;
   $66 = $35;
   $ip$i = $66;
   $67 = $ip$i;
   $68 = $dictPtr$i;
   $69 = ((($68)) + 16400|0);
   $70 = HEAP32[$69>>2]|0;
   $71 = (0 - ($70))|0;
   $72 = (($67) + ($71)|0);
   $lowRefLimit$i = $72;
   $73 = $dictPtr$i;
   $74 = ((($73)) + 16392|0);
   $75 = HEAP32[$74>>2]|0;
   $dictionary$i = $75;
   $76 = $dictionary$i;
   $77 = $dictPtr$i;
   $78 = ((($77)) + 16400|0);
   $79 = HEAP32[$78>>2]|0;
   $80 = (($76) + ($79)|0);
   $dictEnd$i = $80;
   $81 = $dictEnd$i;
   $82 = $35;
   $83 = $81;
   $84 = $82;
   $85 = (($83) - ($84))|0;
   $dictDelta$i = $85;
   $86 = $35;
   $anchor$i = $86;
   $87 = $ip$i;
   $88 = $37;
   $89 = (($87) + ($88)|0);
   $iend$i = $89;
   $90 = $iend$i;
   $91 = ((($90)) + -12|0);
   $mflimit$i = $91;
   $92 = $iend$i;
   $93 = ((($92)) + -5|0);
   $matchlimit$i = $93;
   $94 = $36;
   $op$i = $94;
   $95 = $op$i;
   $96 = $38;
   $97 = (($95) + ($96)|0);
   $olimit$i = $97;
   $refDelta$i = 0;
   $98 = $37;
   $99 = ($98>>>0)>(2113929216);
   L8: do {
    if ($99) {
     $33 = 0;
    } else {
     $100 = $41;
     switch ($100|0) {
     case 2:  {
      $115 = $35;
      $116 = $dictPtr$i;
      $117 = ((($116)) + 16384|0);
      $118 = HEAP32[$117>>2]|0;
      $119 = (0 - ($118))|0;
      $120 = (($115) + ($119)|0);
      $base$i = $120;
      $121 = $35;
      $lowLimit$i = $121;
      break;
     }
     case 1:  {
      $103 = $35;
      $104 = $dictPtr$i;
      $105 = ((($104)) + 16384|0);
      $106 = HEAP32[$105>>2]|0;
      $107 = (0 - ($106))|0;
      $108 = (($103) + ($107)|0);
      $base$i = $108;
      $109 = $35;
      $110 = $dictPtr$i;
      $111 = ((($110)) + 16400|0);
      $112 = HEAP32[$111>>2]|0;
      $113 = (0 - ($112))|0;
      $114 = (($109) + ($113)|0);
      $lowLimit$i = $114;
      break;
     }
     default: {
      $101 = $35;
      $base$i = $101;
      $102 = $35;
      $lowLimit$i = $102;
     }
     }
     $122 = $40;
     $123 = ($122|0)==(2);
     $124 = $37;
     $125 = ($124|0)>=(65547);
     $or$cond = $123 & $125;
     if ($or$cond) {
      $33 = 0;
      break;
     }
     $126 = $37;
     $127 = ($126|0)<(13);
     L19: do {
      if (!($127)) {
       $128 = $ip$i;
       $129 = $34;
       $130 = $40;
       $131 = $base$i;
       _LZ4_putPosition($128,$129,$130,$131);
       $132 = $ip$i;
       $133 = ((($132)) + 1|0);
       $ip$i = $133;
       $134 = $ip$i;
       $135 = $40;
       $136 = (_LZ4_hashPosition($134,$135)|0);
       $forwardH$i = $136;
       L21: while(1) {
        $137 = $ip$i;
        $forwardIp$i = $137;
        $step$i = 1;
        $138 = $43;
        $139 = $138 << 6;
        $searchMatchNb$i = $139;
        while(1) {
         $140 = $forwardH$i;
         $h$i = $140;
         $141 = $forwardIp$i;
         $ip$i = $141;
         $142 = $step$i;
         $143 = $forwardIp$i;
         $144 = (($143) + ($142)|0);
         $forwardIp$i = $144;
         $145 = $searchMatchNb$i;
         $146 = (($145) + 1)|0;
         $searchMatchNb$i = $146;
         $147 = $145 >>> 6;
         $step$i = $147;
         $148 = $forwardIp$i;
         $149 = $mflimit$i;
         $150 = ($148>>>0)>($149>>>0);
         if ($150) {
          break L19;
         }
         $151 = $h$i;
         $152 = $34;
         $153 = $40;
         $154 = $base$i;
         $155 = (_LZ4_getPositionOnHash($151,$152,$153,$154)|0);
         $match$i = $155;
         $156 = $41;
         $157 = ($156|0)==(2);
         do {
          if ($157) {
           $158 = $match$i;
           $159 = $35;
           $160 = ($158>>>0)<($159>>>0);
           if ($160) {
            $161 = $dictDelta$i;
            $refDelta$i = $161;
            $162 = $dictionary$i;
            $lowLimit$i = $162;
            break;
           } else {
            $refDelta$i = 0;
            $163 = $35;
            $lowLimit$i = $163;
            break;
           }
          }
         } while(0);
         $164 = $forwardIp$i;
         $165 = $40;
         $166 = (_LZ4_hashPosition($164,$165)|0);
         $forwardH$i = $166;
         $167 = $ip$i;
         $168 = $h$i;
         $169 = $34;
         $170 = $40;
         $171 = $base$i;
         _LZ4_putPositionOnHash($167,$168,$169,$170,$171);
         $172 = $42;
         $173 = ($172|0)==(1);
         if ($173) {
          $174 = $match$i;
          $175 = $lowRefLimit$i;
          $176 = ($174>>>0)<($175>>>0);
          if ($176) {
           continue;
          }
         }
         $177 = $40;
         $178 = ($177|0)==(2);
         if (!($178)) {
          $179 = $match$i;
          $180 = ((($179)) + 65535|0);
          $181 = $ip$i;
          $182 = ($180>>>0)<($181>>>0);
          if ($182) {
           continue;
          }
         }
         $183 = $match$i;
         $184 = $refDelta$i;
         $185 = (($183) + ($184)|0);
         $186 = (_LZ4_read32($185)|0);
         $187 = $ip$i;
         $188 = (_LZ4_read32($187)|0);
         $189 = ($186|0)!=($188|0);
         if (!($189)) {
          break;
         }
        }
        while(1) {
         $190 = $ip$i;
         $191 = $anchor$i;
         $192 = ($190>>>0)>($191>>>0);
         if ($192) {
          $193 = $match$i;
          $194 = $refDelta$i;
          $195 = (($193) + ($194)|0);
          $196 = $lowLimit$i;
          $197 = ($195>>>0)>($196>>>0);
          if ($197) {
           $198 = $ip$i;
           $199 = ((($198)) + -1|0);
           $200 = HEAP8[$199>>0]|0;
           $201 = $200&255;
           $202 = $refDelta$i;
           $203 = (($202) - 1)|0;
           $204 = $match$i;
           $205 = (($204) + ($203)|0);
           $206 = HEAP8[$205>>0]|0;
           $207 = $206&255;
           $208 = ($201|0)==($207|0);
           $1672 = $208;
          } else {
           $1672 = 0;
          }
         } else {
          $1672 = 0;
         }
         $209 = $ip$i;
         if (!($1672)) {
          break;
         }
         $210 = ((($209)) + -1|0);
         $ip$i = $210;
         $211 = $match$i;
         $212 = ((($211)) + -1|0);
         $match$i = $212;
        }
        $213 = $anchor$i;
        $214 = $209;
        $215 = $213;
        $216 = (($214) - ($215))|0;
        $litLength$i = $216;
        $217 = $op$i;
        $218 = ((($217)) + 1|0);
        $op$i = $218;
        $token$i = $217;
        $219 = $39;
        $220 = ($219|0)!=(0);
        if ($220) {
         $221 = $op$i;
         $222 = $litLength$i;
         $223 = (($221) + ($222)|0);
         $224 = ((($223)) + 8|0);
         $225 = $litLength$i;
         $226 = (($225>>>0) / 255)&-1;
         $227 = (($224) + ($226)|0);
         $228 = $olimit$i;
         $229 = ($227>>>0)>($228>>>0);
         if ($229) {
          label = 33;
          break;
         }
        }
        $230 = $litLength$i;
        $231 = ($230>>>0)>=(15);
        $232 = $litLength$i;
        if ($231) {
         $233 = (($232) - 15)|0;
         $len$i = $233;
         $234 = $token$i;
         HEAP8[$234>>0] = -16;
         while(1) {
          $235 = $len$i;
          $236 = ($235|0)>=(255);
          if (!($236)) {
           break;
          }
          $237 = $op$i;
          $238 = ((($237)) + 1|0);
          $op$i = $238;
          HEAP8[$237>>0] = -1;
          $239 = $len$i;
          $240 = (($239) - 255)|0;
          $len$i = $240;
         }
         $241 = $len$i;
         $242 = $241&255;
         $243 = $op$i;
         $244 = ((($243)) + 1|0);
         $op$i = $244;
         HEAP8[$243>>0] = $242;
        } else {
         $245 = $232 << 4;
         $246 = $245&255;
         $247 = $token$i;
         HEAP8[$247>>0] = $246;
        }
        $248 = $op$i;
        $249 = $anchor$i;
        $250 = $op$i;
        $251 = $litLength$i;
        $252 = (($250) + ($251)|0);
        _LZ4_wildCopy($248,$249,$252);
        $253 = $litLength$i;
        $254 = $op$i;
        $255 = (($254) + ($253)|0);
        $op$i = $255;
        while(1) {
         $256 = $op$i;
         $257 = $ip$i;
         $258 = $match$i;
         $259 = $257;
         $260 = $258;
         $261 = (($259) - ($260))|0;
         $262 = $261&65535;
         _LZ4_writeLE16($256,$262);
         $263 = $op$i;
         $264 = ((($263)) + 2|0);
         $op$i = $264;
         $265 = $41;
         $266 = ($265|0)==(2);
         if ($266) {
          $267 = $lowLimit$i;
          $268 = $dictionary$i;
          $269 = ($267|0)==($268|0);
          if ($269) {
           $270 = $refDelta$i;
           $271 = $match$i;
           $272 = (($271) + ($270)|0);
           $match$i = $272;
           $273 = $ip$i;
           $274 = $dictEnd$i;
           $275 = $match$i;
           $276 = $274;
           $277 = $275;
           $278 = (($276) - ($277))|0;
           $279 = (($273) + ($278)|0);
           $limit$i = $279;
           $280 = $limit$i;
           $281 = $matchlimit$i;
           $282 = ($280>>>0)>($281>>>0);
           if ($282) {
            $283 = $matchlimit$i;
            $limit$i = $283;
           }
           $284 = $ip$i;
           $285 = ((($284)) + 4|0);
           $286 = $match$i;
           $287 = ((($286)) + 4|0);
           $288 = $limit$i;
           $289 = (_LZ4_count($285,$287,$288)|0);
           $matchLength$i = $289;
           $290 = $matchLength$i;
           $291 = (4 + ($290))|0;
           $292 = $ip$i;
           $293 = (($292) + ($291)|0);
           $ip$i = $293;
           $294 = $ip$i;
           $295 = $limit$i;
           $296 = ($294|0)==($295|0);
           if ($296) {
            $297 = $ip$i;
            $298 = $35;
            $299 = $matchlimit$i;
            $300 = (_LZ4_count($297,$298,$299)|0);
            $more$i = $300;
            $301 = $more$i;
            $302 = $matchLength$i;
            $303 = (($302) + ($301))|0;
            $matchLength$i = $303;
            $304 = $more$i;
            $305 = $ip$i;
            $306 = (($305) + ($304)|0);
            $ip$i = $306;
           }
          } else {
           label = 47;
          }
         } else {
          label = 47;
         }
         if ((label|0) == 47) {
          label = 0;
          $307 = $ip$i;
          $308 = ((($307)) + 4|0);
          $309 = $match$i;
          $310 = ((($309)) + 4|0);
          $311 = $matchlimit$i;
          $312 = (_LZ4_count($308,$310,$311)|0);
          $matchLength$i = $312;
          $313 = $matchLength$i;
          $314 = (4 + ($313))|0;
          $315 = $ip$i;
          $316 = (($315) + ($314)|0);
          $ip$i = $316;
         }
         $317 = $39;
         $318 = ($317|0)!=(0);
         if ($318) {
          $319 = $op$i;
          $320 = ((($319)) + 6|0);
          $321 = $matchLength$i;
          $322 = $321 >>> 8;
          $323 = (($320) + ($322)|0);
          $324 = $olimit$i;
          $325 = ($323>>>0)>($324>>>0);
          if ($325) {
           label = 50;
           break L21;
          }
         }
         $326 = $matchLength$i;
         $327 = ($326>>>0)>=(15);
         if ($327) {
          $328 = $token$i;
          $329 = HEAP8[$328>>0]|0;
          $330 = $329&255;
          $331 = (($330) + 15)|0;
          $332 = $331&255;
          HEAP8[$328>>0] = $332;
          $333 = $matchLength$i;
          $334 = (($333) - 15)|0;
          $matchLength$i = $334;
          while(1) {
           $335 = $matchLength$i;
           $336 = ($335>>>0)>=(510);
           if (!($336)) {
            break;
           }
           $337 = $op$i;
           $338 = ((($337)) + 1|0);
           $op$i = $338;
           HEAP8[$337>>0] = -1;
           $339 = $op$i;
           $340 = ((($339)) + 1|0);
           $op$i = $340;
           HEAP8[$339>>0] = -1;
           $341 = $matchLength$i;
           $342 = (($341) - 510)|0;
           $matchLength$i = $342;
          }
          $343 = $matchLength$i;
          $344 = ($343>>>0)>=(255);
          if ($344) {
           $345 = $matchLength$i;
           $346 = (($345) - 255)|0;
           $matchLength$i = $346;
           $347 = $op$i;
           $348 = ((($347)) + 1|0);
           $op$i = $348;
           HEAP8[$347>>0] = -1;
          }
          $349 = $matchLength$i;
          $350 = $349&255;
          $351 = $op$i;
          $352 = ((($351)) + 1|0);
          $op$i = $352;
          HEAP8[$351>>0] = $350;
         } else {
          $353 = $matchLength$i;
          $354 = $353&255;
          $355 = $354&255;
          $356 = $token$i;
          $357 = HEAP8[$356>>0]|0;
          $358 = $357&255;
          $359 = (($358) + ($355))|0;
          $360 = $359&255;
          HEAP8[$356>>0] = $360;
         }
         $361 = $ip$i;
         $anchor$i = $361;
         $362 = $ip$i;
         $363 = $mflimit$i;
         $364 = ($362>>>0)>($363>>>0);
         if ($364) {
          break L19;
         }
         $365 = $ip$i;
         $366 = ((($365)) + -2|0);
         $367 = $34;
         $368 = $40;
         $369 = $base$i;
         _LZ4_putPosition($366,$367,$368,$369);
         $370 = $ip$i;
         $371 = $34;
         $372 = $40;
         $373 = $base$i;
         $374 = (_LZ4_getPosition($370,$371,$372,$373)|0);
         $match$i = $374;
         $375 = $41;
         $376 = ($375|0)==(2);
         do {
          if ($376) {
           $377 = $match$i;
           $378 = $35;
           $379 = ($377>>>0)<($378>>>0);
           if ($379) {
            $380 = $dictDelta$i;
            $refDelta$i = $380;
            $381 = $dictionary$i;
            $lowLimit$i = $381;
            break;
           } else {
            $refDelta$i = 0;
            $382 = $35;
            $lowLimit$i = $382;
            break;
           }
          }
         } while(0);
         $383 = $ip$i;
         $384 = $34;
         $385 = $40;
         $386 = $base$i;
         _LZ4_putPosition($383,$384,$385,$386);
         $387 = $42;
         $388 = ($387|0)==(1);
         if ($388) {
          $389 = $match$i;
          $390 = $lowRefLimit$i;
          $391 = ($389>>>0)>=($390>>>0);
          if (!($391)) {
           break;
          }
         }
         $392 = $match$i;
         $393 = ((($392)) + 65535|0);
         $394 = $ip$i;
         $395 = ($393>>>0)>=($394>>>0);
         if (!($395)) {
          break;
         }
         $396 = $match$i;
         $397 = $refDelta$i;
         $398 = (($396) + ($397)|0);
         $399 = (_LZ4_read32($398)|0);
         $400 = $ip$i;
         $401 = (_LZ4_read32($400)|0);
         $402 = ($399|0)==($401|0);
         if (!($402)) {
          break;
         }
         $403 = $op$i;
         $404 = ((($403)) + 1|0);
         $op$i = $404;
         $token$i = $403;
         $405 = $token$i;
         HEAP8[$405>>0] = 0;
        }
        $406 = $ip$i;
        $407 = ((($406)) + 1|0);
        $ip$i = $407;
        $408 = $40;
        $409 = (_LZ4_hashPosition($407,$408)|0);
        $forwardH$i = $409;
       }
       if ((label|0) == 33) {
        $33 = 0;
        break L8;
       }
       else if ((label|0) == 50) {
        $33 = 0;
        break L8;
       }
      }
     } while(0);
     $410 = $iend$i;
     $411 = $anchor$i;
     $412 = $410;
     $413 = $411;
     $414 = (($412) - ($413))|0;
     $lastRun$i = $414;
     $415 = $39;
     $416 = ($415|0)!=(0);
     if ($416) {
      $417 = $op$i;
      $418 = $36;
      $419 = $417;
      $420 = $418;
      $421 = (($419) - ($420))|0;
      $422 = $lastRun$i;
      $423 = (($421) + ($422))|0;
      $424 = (($423) + 1)|0;
      $425 = $lastRun$i;
      $426 = (($425) + 255)|0;
      $427 = (($426) - 15)|0;
      $428 = (($427>>>0) / 255)&-1;
      $429 = (($424) + ($428))|0;
      $430 = $38;
      $431 = ($429>>>0)>($430>>>0);
      if ($431) {
       $33 = 0;
       break;
      }
     }
     $432 = $lastRun$i;
     $433 = ($432>>>0)>=(15);
     $434 = $lastRun$i;
     if ($433) {
      $435 = (($434) - 15)|0;
      $accumulator$i = $435;
      $436 = $op$i;
      $437 = ((($436)) + 1|0);
      $op$i = $437;
      HEAP8[$436>>0] = -16;
      while(1) {
       $438 = $accumulator$i;
       $439 = ($438>>>0)>=(255);
       if (!($439)) {
        break;
       }
       $440 = $op$i;
       $441 = ((($440)) + 1|0);
       $op$i = $441;
       HEAP8[$440>>0] = -1;
       $442 = $accumulator$i;
       $443 = (($442) - 255)|0;
       $accumulator$i = $443;
      }
      $444 = $accumulator$i;
      $445 = $444&255;
      $446 = $op$i;
      $447 = ((($446)) + 1|0);
      $op$i = $447;
      HEAP8[$446>>0] = $445;
     } else {
      $448 = $434 << 4;
      $449 = $448&255;
      $450 = $op$i;
      $451 = ((($450)) + 1|0);
      $op$i = $451;
      HEAP8[$450>>0] = $449;
     }
     $452 = $op$i;
     $453 = $anchor$i;
     $454 = $lastRun$i;
     _memcpy(($452|0),($453|0),($454|0))|0;
     $455 = $lastRun$i;
     $456 = $op$i;
     $457 = (($456) + ($455)|0);
     $op$i = $457;
     $458 = $op$i;
     $459 = $36;
     $460 = $458;
     $461 = $459;
     $462 = (($460) - ($461))|0;
     $33 = $462;
    }
   } while(0);
   $463 = $33;
   $44 = $463;
   $1671 = $44;
   STACKTOP = sp;return ($1671|0);
  } else {
   $464 = (_LZ4_64bits()|0);
   $465 = ($464|0)!=(0);
   $466 = $465 ? 1 : 0;
   $467 = $50;
   $23 = $60;
   $24 = $61;
   $25 = $62;
   $26 = $63;
   $27 = 0;
   $28 = 0;
   $29 = $466;
   $30 = 0;
   $31 = 0;
   $32 = $467;
   $468 = $23;
   $dictPtr$i1 = $468;
   $469 = $24;
   $ip$i2 = $469;
   $470 = $ip$i2;
   $471 = $dictPtr$i1;
   $472 = ((($471)) + 16400|0);
   $473 = HEAP32[$472>>2]|0;
   $474 = (0 - ($473))|0;
   $475 = (($470) + ($474)|0);
   $lowRefLimit$i5 = $475;
   $476 = $dictPtr$i1;
   $477 = ((($476)) + 16392|0);
   $478 = HEAP32[$477>>2]|0;
   $dictionary$i6 = $478;
   $479 = $dictionary$i6;
   $480 = $dictPtr$i1;
   $481 = ((($480)) + 16400|0);
   $482 = HEAP32[$481>>2]|0;
   $483 = (($479) + ($482)|0);
   $dictEnd$i7 = $483;
   $484 = $dictEnd$i7;
   $485 = $24;
   $486 = $484;
   $487 = $485;
   $488 = (($486) - ($487))|0;
   $dictDelta$i8 = $488;
   $489 = $24;
   $anchor$i9 = $489;
   $490 = $ip$i2;
   $491 = $26;
   $492 = (($490) + ($491)|0);
   $iend$i10 = $492;
   $493 = $iend$i10;
   $494 = ((($493)) + -12|0);
   $mflimit$i11 = $494;
   $495 = $iend$i10;
   $496 = ((($495)) + -5|0);
   $matchlimit$i12 = $496;
   $497 = $25;
   $op$i13 = $497;
   $498 = $op$i13;
   $499 = $27;
   $500 = (($498) + ($499)|0);
   $olimit$i14 = $500;
   $refDelta$i16 = 0;
   $501 = $26;
   $502 = ($501>>>0)>(2113929216);
   L115: do {
    if ($502) {
     $22 = 0;
    } else {
     $503 = $30;
     switch ($503|0) {
     case 2:  {
      $518 = $24;
      $519 = $dictPtr$i1;
      $520 = ((($519)) + 16384|0);
      $521 = HEAP32[$520>>2]|0;
      $522 = (0 - ($521))|0;
      $523 = (($518) + ($522)|0);
      $base$i3 = $523;
      $524 = $24;
      $lowLimit$i4 = $524;
      break;
     }
     case 1:  {
      $506 = $24;
      $507 = $dictPtr$i1;
      $508 = ((($507)) + 16384|0);
      $509 = HEAP32[$508>>2]|0;
      $510 = (0 - ($509))|0;
      $511 = (($506) + ($510)|0);
      $base$i3 = $511;
      $512 = $24;
      $513 = $dictPtr$i1;
      $514 = ((($513)) + 16400|0);
      $515 = HEAP32[$514>>2]|0;
      $516 = (0 - ($515))|0;
      $517 = (($512) + ($516)|0);
      $lowLimit$i4 = $517;
      break;
     }
     default: {
      $504 = $24;
      $base$i3 = $504;
      $505 = $24;
      $lowLimit$i4 = $505;
     }
     }
     $525 = $29;
     $526 = ($525|0)==(2);
     $527 = $26;
     $528 = ($527|0)>=(65547);
     $or$cond4 = $526 & $528;
     if ($or$cond4) {
      $22 = 0;
      break;
     }
     $529 = $26;
     $530 = ($529|0)<(13);
     L126: do {
      if (!($530)) {
       $531 = $ip$i2;
       $532 = $23;
       $533 = $29;
       $534 = $base$i3;
       _LZ4_putPosition($531,$532,$533,$534);
       $535 = $ip$i2;
       $536 = ((($535)) + 1|0);
       $ip$i2 = $536;
       $537 = $ip$i2;
       $538 = $29;
       $539 = (_LZ4_hashPosition($537,$538)|0);
       $forwardH$i15 = $539;
       L128: while(1) {
        $540 = $ip$i2;
        $forwardIp$i19 = $540;
        $step$i20 = 1;
        $541 = $32;
        $542 = $541 << 6;
        $searchMatchNb$i21 = $542;
        while(1) {
         $543 = $forwardH$i15;
         $h$i22 = $543;
         $544 = $forwardIp$i19;
         $ip$i2 = $544;
         $545 = $step$i20;
         $546 = $forwardIp$i19;
         $547 = (($546) + ($545)|0);
         $forwardIp$i19 = $547;
         $548 = $searchMatchNb$i21;
         $549 = (($548) + 1)|0;
         $searchMatchNb$i21 = $549;
         $550 = $548 >>> 6;
         $step$i20 = $550;
         $551 = $forwardIp$i19;
         $552 = $mflimit$i11;
         $553 = ($551>>>0)>($552>>>0);
         if ($553) {
          break L126;
         }
         $554 = $h$i22;
         $555 = $23;
         $556 = $29;
         $557 = $base$i3;
         $558 = (_LZ4_getPositionOnHash($554,$555,$556,$557)|0);
         $match$i17 = $558;
         $559 = $30;
         $560 = ($559|0)==(2);
         do {
          if ($560) {
           $561 = $match$i17;
           $562 = $24;
           $563 = ($561>>>0)<($562>>>0);
           if ($563) {
            $564 = $dictDelta$i8;
            $refDelta$i16 = $564;
            $565 = $dictionary$i6;
            $lowLimit$i4 = $565;
            break;
           } else {
            $refDelta$i16 = 0;
            $566 = $24;
            $lowLimit$i4 = $566;
            break;
           }
          }
         } while(0);
         $567 = $forwardIp$i19;
         $568 = $29;
         $569 = (_LZ4_hashPosition($567,$568)|0);
         $forwardH$i15 = $569;
         $570 = $ip$i2;
         $571 = $h$i22;
         $572 = $23;
         $573 = $29;
         $574 = $base$i3;
         _LZ4_putPositionOnHash($570,$571,$572,$573,$574);
         $575 = $31;
         $576 = ($575|0)==(1);
         if ($576) {
          $577 = $match$i17;
          $578 = $lowRefLimit$i5;
          $579 = ($577>>>0)<($578>>>0);
          if ($579) {
           continue;
          }
         }
         $580 = $29;
         $581 = ($580|0)==(2);
         if (!($581)) {
          $582 = $match$i17;
          $583 = ((($582)) + 65535|0);
          $584 = $ip$i2;
          $585 = ($583>>>0)<($584>>>0);
          if ($585) {
           continue;
          }
         }
         $586 = $match$i17;
         $587 = $refDelta$i16;
         $588 = (($586) + ($587)|0);
         $589 = (_LZ4_read32($588)|0);
         $590 = $ip$i2;
         $591 = (_LZ4_read32($590)|0);
         $592 = ($589|0)!=($591|0);
         if (!($592)) {
          break;
         }
        }
        while(1) {
         $593 = $ip$i2;
         $594 = $anchor$i9;
         $595 = ($593>>>0)>($594>>>0);
         if ($595) {
          $596 = $match$i17;
          $597 = $refDelta$i16;
          $598 = (($596) + ($597)|0);
          $599 = $lowLimit$i4;
          $600 = ($598>>>0)>($599>>>0);
          if ($600) {
           $601 = $ip$i2;
           $602 = ((($601)) + -1|0);
           $603 = HEAP8[$602>>0]|0;
           $604 = $603&255;
           $605 = $refDelta$i16;
           $606 = (($605) - 1)|0;
           $607 = $match$i17;
           $608 = (($607) + ($606)|0);
           $609 = HEAP8[$608>>0]|0;
           $610 = $609&255;
           $611 = ($604|0)==($610|0);
           $1673 = $611;
          } else {
           $1673 = 0;
          }
         } else {
          $1673 = 0;
         }
         $612 = $ip$i2;
         if (!($1673)) {
          break;
         }
         $613 = ((($612)) + -1|0);
         $ip$i2 = $613;
         $614 = $match$i17;
         $615 = ((($614)) + -1|0);
         $match$i17 = $615;
        }
        $616 = $anchor$i9;
        $617 = $612;
        $618 = $616;
        $619 = (($617) - ($618))|0;
        $litLength$i23 = $619;
        $620 = $op$i13;
        $621 = ((($620)) + 1|0);
        $op$i13 = $621;
        $token$i18 = $620;
        $622 = $28;
        $623 = ($622|0)!=(0);
        if ($623) {
         $624 = $op$i13;
         $625 = $litLength$i23;
         $626 = (($624) + ($625)|0);
         $627 = ((($626)) + 8|0);
         $628 = $litLength$i23;
         $629 = (($628>>>0) / 255)&-1;
         $630 = (($627) + ($629)|0);
         $631 = $olimit$i14;
         $632 = ($630>>>0)>($631>>>0);
         if ($632) {
          label = 109;
          break;
         }
        }
        $633 = $litLength$i23;
        $634 = ($633>>>0)>=(15);
        $635 = $litLength$i23;
        if ($634) {
         $636 = (($635) - 15)|0;
         $len$i24 = $636;
         $637 = $token$i18;
         HEAP8[$637>>0] = -16;
         while(1) {
          $638 = $len$i24;
          $639 = ($638|0)>=(255);
          if (!($639)) {
           break;
          }
          $640 = $op$i13;
          $641 = ((($640)) + 1|0);
          $op$i13 = $641;
          HEAP8[$640>>0] = -1;
          $642 = $len$i24;
          $643 = (($642) - 255)|0;
          $len$i24 = $643;
         }
         $644 = $len$i24;
         $645 = $644&255;
         $646 = $op$i13;
         $647 = ((($646)) + 1|0);
         $op$i13 = $647;
         HEAP8[$646>>0] = $645;
        } else {
         $648 = $635 << 4;
         $649 = $648&255;
         $650 = $token$i18;
         HEAP8[$650>>0] = $649;
        }
        $651 = $op$i13;
        $652 = $anchor$i9;
        $653 = $op$i13;
        $654 = $litLength$i23;
        $655 = (($653) + ($654)|0);
        _LZ4_wildCopy($651,$652,$655);
        $656 = $litLength$i23;
        $657 = $op$i13;
        $658 = (($657) + ($656)|0);
        $op$i13 = $658;
        while(1) {
         $659 = $op$i13;
         $660 = $ip$i2;
         $661 = $match$i17;
         $662 = $660;
         $663 = $661;
         $664 = (($662) - ($663))|0;
         $665 = $664&65535;
         _LZ4_writeLE16($659,$665);
         $666 = $op$i13;
         $667 = ((($666)) + 2|0);
         $op$i13 = $667;
         $668 = $30;
         $669 = ($668|0)==(2);
         if ($669) {
          $670 = $lowLimit$i4;
          $671 = $dictionary$i6;
          $672 = ($670|0)==($671|0);
          if ($672) {
           $673 = $refDelta$i16;
           $674 = $match$i17;
           $675 = (($674) + ($673)|0);
           $match$i17 = $675;
           $676 = $ip$i2;
           $677 = $dictEnd$i7;
           $678 = $match$i17;
           $679 = $677;
           $680 = $678;
           $681 = (($679) - ($680))|0;
           $682 = (($676) + ($681)|0);
           $limit$i26 = $682;
           $683 = $limit$i26;
           $684 = $matchlimit$i12;
           $685 = ($683>>>0)>($684>>>0);
           if ($685) {
            $686 = $matchlimit$i12;
            $limit$i26 = $686;
           }
           $687 = $ip$i2;
           $688 = ((($687)) + 4|0);
           $689 = $match$i17;
           $690 = ((($689)) + 4|0);
           $691 = $limit$i26;
           $692 = (_LZ4_count($688,$690,$691)|0);
           $matchLength$i25 = $692;
           $693 = $matchLength$i25;
           $694 = (4 + ($693))|0;
           $695 = $ip$i2;
           $696 = (($695) + ($694)|0);
           $ip$i2 = $696;
           $697 = $ip$i2;
           $698 = $limit$i26;
           $699 = ($697|0)==($698|0);
           if ($699) {
            $700 = $ip$i2;
            $701 = $24;
            $702 = $matchlimit$i12;
            $703 = (_LZ4_count($700,$701,$702)|0);
            $more$i27 = $703;
            $704 = $more$i27;
            $705 = $matchLength$i25;
            $706 = (($705) + ($704))|0;
            $matchLength$i25 = $706;
            $707 = $more$i27;
            $708 = $ip$i2;
            $709 = (($708) + ($707)|0);
            $ip$i2 = $709;
           }
          } else {
           label = 123;
          }
         } else {
          label = 123;
         }
         if ((label|0) == 123) {
          label = 0;
          $710 = $ip$i2;
          $711 = ((($710)) + 4|0);
          $712 = $match$i17;
          $713 = ((($712)) + 4|0);
          $714 = $matchlimit$i12;
          $715 = (_LZ4_count($711,$713,$714)|0);
          $matchLength$i25 = $715;
          $716 = $matchLength$i25;
          $717 = (4 + ($716))|0;
          $718 = $ip$i2;
          $719 = (($718) + ($717)|0);
          $ip$i2 = $719;
         }
         $720 = $28;
         $721 = ($720|0)!=(0);
         if ($721) {
          $722 = $op$i13;
          $723 = ((($722)) + 6|0);
          $724 = $matchLength$i25;
          $725 = $724 >>> 8;
          $726 = (($723) + ($725)|0);
          $727 = $olimit$i14;
          $728 = ($726>>>0)>($727>>>0);
          if ($728) {
           label = 126;
           break L128;
          }
         }
         $729 = $matchLength$i25;
         $730 = ($729>>>0)>=(15);
         if ($730) {
          $731 = $token$i18;
          $732 = HEAP8[$731>>0]|0;
          $733 = $732&255;
          $734 = (($733) + 15)|0;
          $735 = $734&255;
          HEAP8[$731>>0] = $735;
          $736 = $matchLength$i25;
          $737 = (($736) - 15)|0;
          $matchLength$i25 = $737;
          while(1) {
           $738 = $matchLength$i25;
           $739 = ($738>>>0)>=(510);
           if (!($739)) {
            break;
           }
           $740 = $op$i13;
           $741 = ((($740)) + 1|0);
           $op$i13 = $741;
           HEAP8[$740>>0] = -1;
           $742 = $op$i13;
           $743 = ((($742)) + 1|0);
           $op$i13 = $743;
           HEAP8[$742>>0] = -1;
           $744 = $matchLength$i25;
           $745 = (($744) - 510)|0;
           $matchLength$i25 = $745;
          }
          $746 = $matchLength$i25;
          $747 = ($746>>>0)>=(255);
          if ($747) {
           $748 = $matchLength$i25;
           $749 = (($748) - 255)|0;
           $matchLength$i25 = $749;
           $750 = $op$i13;
           $751 = ((($750)) + 1|0);
           $op$i13 = $751;
           HEAP8[$750>>0] = -1;
          }
          $752 = $matchLength$i25;
          $753 = $752&255;
          $754 = $op$i13;
          $755 = ((($754)) + 1|0);
          $op$i13 = $755;
          HEAP8[$754>>0] = $753;
         } else {
          $756 = $matchLength$i25;
          $757 = $756&255;
          $758 = $757&255;
          $759 = $token$i18;
          $760 = HEAP8[$759>>0]|0;
          $761 = $760&255;
          $762 = (($761) + ($758))|0;
          $763 = $762&255;
          HEAP8[$759>>0] = $763;
         }
         $764 = $ip$i2;
         $anchor$i9 = $764;
         $765 = $ip$i2;
         $766 = $mflimit$i11;
         $767 = ($765>>>0)>($766>>>0);
         if ($767) {
          break L126;
         }
         $768 = $ip$i2;
         $769 = ((($768)) + -2|0);
         $770 = $23;
         $771 = $29;
         $772 = $base$i3;
         _LZ4_putPosition($769,$770,$771,$772);
         $773 = $ip$i2;
         $774 = $23;
         $775 = $29;
         $776 = $base$i3;
         $777 = (_LZ4_getPosition($773,$774,$775,$776)|0);
         $match$i17 = $777;
         $778 = $30;
         $779 = ($778|0)==(2);
         do {
          if ($779) {
           $780 = $match$i17;
           $781 = $24;
           $782 = ($780>>>0)<($781>>>0);
           if ($782) {
            $783 = $dictDelta$i8;
            $refDelta$i16 = $783;
            $784 = $dictionary$i6;
            $lowLimit$i4 = $784;
            break;
           } else {
            $refDelta$i16 = 0;
            $785 = $24;
            $lowLimit$i4 = $785;
            break;
           }
          }
         } while(0);
         $786 = $ip$i2;
         $787 = $23;
         $788 = $29;
         $789 = $base$i3;
         _LZ4_putPosition($786,$787,$788,$789);
         $790 = $31;
         $791 = ($790|0)==(1);
         if ($791) {
          $792 = $match$i17;
          $793 = $lowRefLimit$i5;
          $794 = ($792>>>0)>=($793>>>0);
          if (!($794)) {
           break;
          }
         }
         $795 = $match$i17;
         $796 = ((($795)) + 65535|0);
         $797 = $ip$i2;
         $798 = ($796>>>0)>=($797>>>0);
         if (!($798)) {
          break;
         }
         $799 = $match$i17;
         $800 = $refDelta$i16;
         $801 = (($799) + ($800)|0);
         $802 = (_LZ4_read32($801)|0);
         $803 = $ip$i2;
         $804 = (_LZ4_read32($803)|0);
         $805 = ($802|0)==($804|0);
         if (!($805)) {
          break;
         }
         $806 = $op$i13;
         $807 = ((($806)) + 1|0);
         $op$i13 = $807;
         $token$i18 = $806;
         $808 = $token$i18;
         HEAP8[$808>>0] = 0;
        }
        $809 = $ip$i2;
        $810 = ((($809)) + 1|0);
        $ip$i2 = $810;
        $811 = $29;
        $812 = (_LZ4_hashPosition($810,$811)|0);
        $forwardH$i15 = $812;
       }
       if ((label|0) == 109) {
        $22 = 0;
        break L115;
       }
       else if ((label|0) == 126) {
        $22 = 0;
        break L115;
       }
      }
     } while(0);
     $813 = $iend$i10;
     $814 = $anchor$i9;
     $815 = $813;
     $816 = $814;
     $817 = (($815) - ($816))|0;
     $lastRun$i28 = $817;
     $818 = $28;
     $819 = ($818|0)!=(0);
     if ($819) {
      $820 = $op$i13;
      $821 = $25;
      $822 = $820;
      $823 = $821;
      $824 = (($822) - ($823))|0;
      $825 = $lastRun$i28;
      $826 = (($824) + ($825))|0;
      $827 = (($826) + 1)|0;
      $828 = $lastRun$i28;
      $829 = (($828) + 255)|0;
      $830 = (($829) - 15)|0;
      $831 = (($830>>>0) / 255)&-1;
      $832 = (($827) + ($831))|0;
      $833 = $27;
      $834 = ($832>>>0)>($833>>>0);
      if ($834) {
       $22 = 0;
       break;
      }
     }
     $835 = $lastRun$i28;
     $836 = ($835>>>0)>=(15);
     $837 = $lastRun$i28;
     if ($836) {
      $838 = (($837) - 15)|0;
      $accumulator$i29 = $838;
      $839 = $op$i13;
      $840 = ((($839)) + 1|0);
      $op$i13 = $840;
      HEAP8[$839>>0] = -16;
      while(1) {
       $841 = $accumulator$i29;
       $842 = ($841>>>0)>=(255);
       if (!($842)) {
        break;
       }
       $843 = $op$i13;
       $844 = ((($843)) + 1|0);
       $op$i13 = $844;
       HEAP8[$843>>0] = -1;
       $845 = $accumulator$i29;
       $846 = (($845) - 255)|0;
       $accumulator$i29 = $846;
      }
      $847 = $accumulator$i29;
      $848 = $847&255;
      $849 = $op$i13;
      $850 = ((($849)) + 1|0);
      $op$i13 = $850;
      HEAP8[$849>>0] = $848;
     } else {
      $851 = $837 << 4;
      $852 = $851&255;
      $853 = $op$i13;
      $854 = ((($853)) + 1|0);
      $op$i13 = $854;
      HEAP8[$853>>0] = $852;
     }
     $855 = $op$i13;
     $856 = $anchor$i9;
     $857 = $lastRun$i28;
     _memcpy(($855|0),($856|0),($857|0))|0;
     $858 = $lastRun$i28;
     $859 = $op$i13;
     $860 = (($859) + ($858)|0);
     $op$i13 = $860;
     $861 = $op$i13;
     $862 = $25;
     $863 = $861;
     $864 = $862;
     $865 = (($863) - ($864))|0;
     $22 = $865;
    }
   } while(0);
   $866 = $22;
   $44 = $866;
   $1671 = $44;
   STACKTOP = sp;return ($1671|0);
  }
 } else {
  $867 = $49;
  if ($59) {
   $868 = $50;
   $12 = $60;
   $13 = $61;
   $14 = $62;
   $15 = $63;
   $16 = $867;
   $17 = 1;
   $18 = 2;
   $19 = 0;
   $20 = 0;
   $21 = $868;
   $869 = $12;
   $dictPtr$i31 = $869;
   $870 = $13;
   $ip$i32 = $870;
   $871 = $ip$i32;
   $872 = $dictPtr$i31;
   $873 = ((($872)) + 16400|0);
   $874 = HEAP32[$873>>2]|0;
   $875 = (0 - ($874))|0;
   $876 = (($871) + ($875)|0);
   $lowRefLimit$i35 = $876;
   $877 = $dictPtr$i31;
   $878 = ((($877)) + 16392|0);
   $879 = HEAP32[$878>>2]|0;
   $dictionary$i36 = $879;
   $880 = $dictionary$i36;
   $881 = $dictPtr$i31;
   $882 = ((($881)) + 16400|0);
   $883 = HEAP32[$882>>2]|0;
   $884 = (($880) + ($883)|0);
   $dictEnd$i37 = $884;
   $885 = $dictEnd$i37;
   $886 = $13;
   $887 = $885;
   $888 = $886;
   $889 = (($887) - ($888))|0;
   $dictDelta$i38 = $889;
   $890 = $13;
   $anchor$i39 = $890;
   $891 = $ip$i32;
   $892 = $15;
   $893 = (($891) + ($892)|0);
   $iend$i40 = $893;
   $894 = $iend$i40;
   $895 = ((($894)) + -12|0);
   $mflimit$i41 = $895;
   $896 = $iend$i40;
   $897 = ((($896)) + -5|0);
   $matchlimit$i42 = $897;
   $898 = $14;
   $op$i43 = $898;
   $899 = $op$i43;
   $900 = $16;
   $901 = (($899) + ($900)|0);
   $olimit$i44 = $901;
   $refDelta$i46 = 0;
   $902 = $15;
   $903 = ($902>>>0)>(2113929216);
   L224: do {
    if ($903) {
     $11 = 0;
    } else {
     $904 = $19;
     switch ($904|0) {
     case 2:  {
      $919 = $13;
      $920 = $dictPtr$i31;
      $921 = ((($920)) + 16384|0);
      $922 = HEAP32[$921>>2]|0;
      $923 = (0 - ($922))|0;
      $924 = (($919) + ($923)|0);
      $base$i33 = $924;
      $925 = $13;
      $lowLimit$i34 = $925;
      break;
     }
     case 1:  {
      $907 = $13;
      $908 = $dictPtr$i31;
      $909 = ((($908)) + 16384|0);
      $910 = HEAP32[$909>>2]|0;
      $911 = (0 - ($910))|0;
      $912 = (($907) + ($911)|0);
      $base$i33 = $912;
      $913 = $13;
      $914 = $dictPtr$i31;
      $915 = ((($914)) + 16400|0);
      $916 = HEAP32[$915>>2]|0;
      $917 = (0 - ($916))|0;
      $918 = (($913) + ($917)|0);
      $lowLimit$i34 = $918;
      break;
     }
     default: {
      $905 = $13;
      $base$i33 = $905;
      $906 = $13;
      $lowLimit$i34 = $906;
     }
     }
     $926 = $18;
     $927 = ($926|0)==(2);
     $928 = $15;
     $929 = ($928|0)>=(65547);
     $or$cond8 = $927 & $929;
     if ($or$cond8) {
      $11 = 0;
      break;
     }
     $930 = $15;
     $931 = ($930|0)<(13);
     L235: do {
      if (!($931)) {
       $932 = $ip$i32;
       $933 = $12;
       $934 = $18;
       $935 = $base$i33;
       _LZ4_putPosition($932,$933,$934,$935);
       $936 = $ip$i32;
       $937 = ((($936)) + 1|0);
       $ip$i32 = $937;
       $938 = $ip$i32;
       $939 = $18;
       $940 = (_LZ4_hashPosition($938,$939)|0);
       $forwardH$i45 = $940;
       L237: while(1) {
        $941 = $ip$i32;
        $forwardIp$i49 = $941;
        $step$i50 = 1;
        $942 = $21;
        $943 = $942 << 6;
        $searchMatchNb$i51 = $943;
        while(1) {
         $944 = $forwardH$i45;
         $h$i52 = $944;
         $945 = $forwardIp$i49;
         $ip$i32 = $945;
         $946 = $step$i50;
         $947 = $forwardIp$i49;
         $948 = (($947) + ($946)|0);
         $forwardIp$i49 = $948;
         $949 = $searchMatchNb$i51;
         $950 = (($949) + 1)|0;
         $searchMatchNb$i51 = $950;
         $951 = $949 >>> 6;
         $step$i50 = $951;
         $952 = $forwardIp$i49;
         $953 = $mflimit$i41;
         $954 = ($952>>>0)>($953>>>0);
         if ($954) {
          break L235;
         }
         $955 = $h$i52;
         $956 = $12;
         $957 = $18;
         $958 = $base$i33;
         $959 = (_LZ4_getPositionOnHash($955,$956,$957,$958)|0);
         $match$i47 = $959;
         $960 = $19;
         $961 = ($960|0)==(2);
         do {
          if ($961) {
           $962 = $match$i47;
           $963 = $13;
           $964 = ($962>>>0)<($963>>>0);
           if ($964) {
            $965 = $dictDelta$i38;
            $refDelta$i46 = $965;
            $966 = $dictionary$i36;
            $lowLimit$i34 = $966;
            break;
           } else {
            $refDelta$i46 = 0;
            $967 = $13;
            $lowLimit$i34 = $967;
            break;
           }
          }
         } while(0);
         $968 = $forwardIp$i49;
         $969 = $18;
         $970 = (_LZ4_hashPosition($968,$969)|0);
         $forwardH$i45 = $970;
         $971 = $ip$i32;
         $972 = $h$i52;
         $973 = $12;
         $974 = $18;
         $975 = $base$i33;
         _LZ4_putPositionOnHash($971,$972,$973,$974,$975);
         $976 = $20;
         $977 = ($976|0)==(1);
         if ($977) {
          $978 = $match$i47;
          $979 = $lowRefLimit$i35;
          $980 = ($978>>>0)<($979>>>0);
          if ($980) {
           continue;
          }
         }
         $981 = $18;
         $982 = ($981|0)==(2);
         if (!($982)) {
          $983 = $match$i47;
          $984 = ((($983)) + 65535|0);
          $985 = $ip$i32;
          $986 = ($984>>>0)<($985>>>0);
          if ($986) {
           continue;
          }
         }
         $987 = $match$i47;
         $988 = $refDelta$i46;
         $989 = (($987) + ($988)|0);
         $990 = (_LZ4_read32($989)|0);
         $991 = $ip$i32;
         $992 = (_LZ4_read32($991)|0);
         $993 = ($990|0)!=($992|0);
         if (!($993)) {
          break;
         }
        }
        while(1) {
         $994 = $ip$i32;
         $995 = $anchor$i39;
         $996 = ($994>>>0)>($995>>>0);
         if ($996) {
          $997 = $match$i47;
          $998 = $refDelta$i46;
          $999 = (($997) + ($998)|0);
          $1000 = $lowLimit$i34;
          $1001 = ($999>>>0)>($1000>>>0);
          if ($1001) {
           $1002 = $ip$i32;
           $1003 = ((($1002)) + -1|0);
           $1004 = HEAP8[$1003>>0]|0;
           $1005 = $1004&255;
           $1006 = $refDelta$i46;
           $1007 = (($1006) - 1)|0;
           $1008 = $match$i47;
           $1009 = (($1008) + ($1007)|0);
           $1010 = HEAP8[$1009>>0]|0;
           $1011 = $1010&255;
           $1012 = ($1005|0)==($1011|0);
           $1674 = $1012;
          } else {
           $1674 = 0;
          }
         } else {
          $1674 = 0;
         }
         $1013 = $ip$i32;
         if (!($1674)) {
          break;
         }
         $1014 = ((($1013)) + -1|0);
         $ip$i32 = $1014;
         $1015 = $match$i47;
         $1016 = ((($1015)) + -1|0);
         $match$i47 = $1016;
        }
        $1017 = $anchor$i39;
        $1018 = $1013;
        $1019 = $1017;
        $1020 = (($1018) - ($1019))|0;
        $litLength$i53 = $1020;
        $1021 = $op$i43;
        $1022 = ((($1021)) + 1|0);
        $op$i43 = $1022;
        $token$i48 = $1021;
        $1023 = $17;
        $1024 = ($1023|0)!=(0);
        if ($1024) {
         $1025 = $op$i43;
         $1026 = $litLength$i53;
         $1027 = (($1025) + ($1026)|0);
         $1028 = ((($1027)) + 8|0);
         $1029 = $litLength$i53;
         $1030 = (($1029>>>0) / 255)&-1;
         $1031 = (($1028) + ($1030)|0);
         $1032 = $olimit$i44;
         $1033 = ($1031>>>0)>($1032>>>0);
         if ($1033) {
          label = 186;
          break;
         }
        }
        $1034 = $litLength$i53;
        $1035 = ($1034>>>0)>=(15);
        $1036 = $litLength$i53;
        if ($1035) {
         $1037 = (($1036) - 15)|0;
         $len$i54 = $1037;
         $1038 = $token$i48;
         HEAP8[$1038>>0] = -16;
         while(1) {
          $1039 = $len$i54;
          $1040 = ($1039|0)>=(255);
          if (!($1040)) {
           break;
          }
          $1041 = $op$i43;
          $1042 = ((($1041)) + 1|0);
          $op$i43 = $1042;
          HEAP8[$1041>>0] = -1;
          $1043 = $len$i54;
          $1044 = (($1043) - 255)|0;
          $len$i54 = $1044;
         }
         $1045 = $len$i54;
         $1046 = $1045&255;
         $1047 = $op$i43;
         $1048 = ((($1047)) + 1|0);
         $op$i43 = $1048;
         HEAP8[$1047>>0] = $1046;
        } else {
         $1049 = $1036 << 4;
         $1050 = $1049&255;
         $1051 = $token$i48;
         HEAP8[$1051>>0] = $1050;
        }
        $1052 = $op$i43;
        $1053 = $anchor$i39;
        $1054 = $op$i43;
        $1055 = $litLength$i53;
        $1056 = (($1054) + ($1055)|0);
        _LZ4_wildCopy($1052,$1053,$1056);
        $1057 = $litLength$i53;
        $1058 = $op$i43;
        $1059 = (($1058) + ($1057)|0);
        $op$i43 = $1059;
        while(1) {
         $1060 = $op$i43;
         $1061 = $ip$i32;
         $1062 = $match$i47;
         $1063 = $1061;
         $1064 = $1062;
         $1065 = (($1063) - ($1064))|0;
         $1066 = $1065&65535;
         _LZ4_writeLE16($1060,$1066);
         $1067 = $op$i43;
         $1068 = ((($1067)) + 2|0);
         $op$i43 = $1068;
         $1069 = $19;
         $1070 = ($1069|0)==(2);
         if ($1070) {
          $1071 = $lowLimit$i34;
          $1072 = $dictionary$i36;
          $1073 = ($1071|0)==($1072|0);
          if ($1073) {
           $1074 = $refDelta$i46;
           $1075 = $match$i47;
           $1076 = (($1075) + ($1074)|0);
           $match$i47 = $1076;
           $1077 = $ip$i32;
           $1078 = $dictEnd$i37;
           $1079 = $match$i47;
           $1080 = $1078;
           $1081 = $1079;
           $1082 = (($1080) - ($1081))|0;
           $1083 = (($1077) + ($1082)|0);
           $limit$i56 = $1083;
           $1084 = $limit$i56;
           $1085 = $matchlimit$i42;
           $1086 = ($1084>>>0)>($1085>>>0);
           if ($1086) {
            $1087 = $matchlimit$i42;
            $limit$i56 = $1087;
           }
           $1088 = $ip$i32;
           $1089 = ((($1088)) + 4|0);
           $1090 = $match$i47;
           $1091 = ((($1090)) + 4|0);
           $1092 = $limit$i56;
           $1093 = (_LZ4_count($1089,$1091,$1092)|0);
           $matchLength$i55 = $1093;
           $1094 = $matchLength$i55;
           $1095 = (4 + ($1094))|0;
           $1096 = $ip$i32;
           $1097 = (($1096) + ($1095)|0);
           $ip$i32 = $1097;
           $1098 = $ip$i32;
           $1099 = $limit$i56;
           $1100 = ($1098|0)==($1099|0);
           if ($1100) {
            $1101 = $ip$i32;
            $1102 = $13;
            $1103 = $matchlimit$i42;
            $1104 = (_LZ4_count($1101,$1102,$1103)|0);
            $more$i57 = $1104;
            $1105 = $more$i57;
            $1106 = $matchLength$i55;
            $1107 = (($1106) + ($1105))|0;
            $matchLength$i55 = $1107;
            $1108 = $more$i57;
            $1109 = $ip$i32;
            $1110 = (($1109) + ($1108)|0);
            $ip$i32 = $1110;
           }
          } else {
           label = 200;
          }
         } else {
          label = 200;
         }
         if ((label|0) == 200) {
          label = 0;
          $1111 = $ip$i32;
          $1112 = ((($1111)) + 4|0);
          $1113 = $match$i47;
          $1114 = ((($1113)) + 4|0);
          $1115 = $matchlimit$i42;
          $1116 = (_LZ4_count($1112,$1114,$1115)|0);
          $matchLength$i55 = $1116;
          $1117 = $matchLength$i55;
          $1118 = (4 + ($1117))|0;
          $1119 = $ip$i32;
          $1120 = (($1119) + ($1118)|0);
          $ip$i32 = $1120;
         }
         $1121 = $17;
         $1122 = ($1121|0)!=(0);
         if ($1122) {
          $1123 = $op$i43;
          $1124 = ((($1123)) + 6|0);
          $1125 = $matchLength$i55;
          $1126 = $1125 >>> 8;
          $1127 = (($1124) + ($1126)|0);
          $1128 = $olimit$i44;
          $1129 = ($1127>>>0)>($1128>>>0);
          if ($1129) {
           label = 203;
           break L237;
          }
         }
         $1130 = $matchLength$i55;
         $1131 = ($1130>>>0)>=(15);
         if ($1131) {
          $1132 = $token$i48;
          $1133 = HEAP8[$1132>>0]|0;
          $1134 = $1133&255;
          $1135 = (($1134) + 15)|0;
          $1136 = $1135&255;
          HEAP8[$1132>>0] = $1136;
          $1137 = $matchLength$i55;
          $1138 = (($1137) - 15)|0;
          $matchLength$i55 = $1138;
          while(1) {
           $1139 = $matchLength$i55;
           $1140 = ($1139>>>0)>=(510);
           if (!($1140)) {
            break;
           }
           $1141 = $op$i43;
           $1142 = ((($1141)) + 1|0);
           $op$i43 = $1142;
           HEAP8[$1141>>0] = -1;
           $1143 = $op$i43;
           $1144 = ((($1143)) + 1|0);
           $op$i43 = $1144;
           HEAP8[$1143>>0] = -1;
           $1145 = $matchLength$i55;
           $1146 = (($1145) - 510)|0;
           $matchLength$i55 = $1146;
          }
          $1147 = $matchLength$i55;
          $1148 = ($1147>>>0)>=(255);
          if ($1148) {
           $1149 = $matchLength$i55;
           $1150 = (($1149) - 255)|0;
           $matchLength$i55 = $1150;
           $1151 = $op$i43;
           $1152 = ((($1151)) + 1|0);
           $op$i43 = $1152;
           HEAP8[$1151>>0] = -1;
          }
          $1153 = $matchLength$i55;
          $1154 = $1153&255;
          $1155 = $op$i43;
          $1156 = ((($1155)) + 1|0);
          $op$i43 = $1156;
          HEAP8[$1155>>0] = $1154;
         } else {
          $1157 = $matchLength$i55;
          $1158 = $1157&255;
          $1159 = $1158&255;
          $1160 = $token$i48;
          $1161 = HEAP8[$1160>>0]|0;
          $1162 = $1161&255;
          $1163 = (($1162) + ($1159))|0;
          $1164 = $1163&255;
          HEAP8[$1160>>0] = $1164;
         }
         $1165 = $ip$i32;
         $anchor$i39 = $1165;
         $1166 = $ip$i32;
         $1167 = $mflimit$i41;
         $1168 = ($1166>>>0)>($1167>>>0);
         if ($1168) {
          break L235;
         }
         $1169 = $ip$i32;
         $1170 = ((($1169)) + -2|0);
         $1171 = $12;
         $1172 = $18;
         $1173 = $base$i33;
         _LZ4_putPosition($1170,$1171,$1172,$1173);
         $1174 = $ip$i32;
         $1175 = $12;
         $1176 = $18;
         $1177 = $base$i33;
         $1178 = (_LZ4_getPosition($1174,$1175,$1176,$1177)|0);
         $match$i47 = $1178;
         $1179 = $19;
         $1180 = ($1179|0)==(2);
         do {
          if ($1180) {
           $1181 = $match$i47;
           $1182 = $13;
           $1183 = ($1181>>>0)<($1182>>>0);
           if ($1183) {
            $1184 = $dictDelta$i38;
            $refDelta$i46 = $1184;
            $1185 = $dictionary$i36;
            $lowLimit$i34 = $1185;
            break;
           } else {
            $refDelta$i46 = 0;
            $1186 = $13;
            $lowLimit$i34 = $1186;
            break;
           }
          }
         } while(0);
         $1187 = $ip$i32;
         $1188 = $12;
         $1189 = $18;
         $1190 = $base$i33;
         _LZ4_putPosition($1187,$1188,$1189,$1190);
         $1191 = $20;
         $1192 = ($1191|0)==(1);
         if ($1192) {
          $1193 = $match$i47;
          $1194 = $lowRefLimit$i35;
          $1195 = ($1193>>>0)>=($1194>>>0);
          if (!($1195)) {
           break;
          }
         }
         $1196 = $match$i47;
         $1197 = ((($1196)) + 65535|0);
         $1198 = $ip$i32;
         $1199 = ($1197>>>0)>=($1198>>>0);
         if (!($1199)) {
          break;
         }
         $1200 = $match$i47;
         $1201 = $refDelta$i46;
         $1202 = (($1200) + ($1201)|0);
         $1203 = (_LZ4_read32($1202)|0);
         $1204 = $ip$i32;
         $1205 = (_LZ4_read32($1204)|0);
         $1206 = ($1203|0)==($1205|0);
         if (!($1206)) {
          break;
         }
         $1207 = $op$i43;
         $1208 = ((($1207)) + 1|0);
         $op$i43 = $1208;
         $token$i48 = $1207;
         $1209 = $token$i48;
         HEAP8[$1209>>0] = 0;
        }
        $1210 = $ip$i32;
        $1211 = ((($1210)) + 1|0);
        $ip$i32 = $1211;
        $1212 = $18;
        $1213 = (_LZ4_hashPosition($1211,$1212)|0);
        $forwardH$i45 = $1213;
       }
       if ((label|0) == 186) {
        $11 = 0;
        break L224;
       }
       else if ((label|0) == 203) {
        $11 = 0;
        break L224;
       }
      }
     } while(0);
     $1214 = $iend$i40;
     $1215 = $anchor$i39;
     $1216 = $1214;
     $1217 = $1215;
     $1218 = (($1216) - ($1217))|0;
     $lastRun$i58 = $1218;
     $1219 = $17;
     $1220 = ($1219|0)!=(0);
     if ($1220) {
      $1221 = $op$i43;
      $1222 = $14;
      $1223 = $1221;
      $1224 = $1222;
      $1225 = (($1223) - ($1224))|0;
      $1226 = $lastRun$i58;
      $1227 = (($1225) + ($1226))|0;
      $1228 = (($1227) + 1)|0;
      $1229 = $lastRun$i58;
      $1230 = (($1229) + 255)|0;
      $1231 = (($1230) - 15)|0;
      $1232 = (($1231>>>0) / 255)&-1;
      $1233 = (($1228) + ($1232))|0;
      $1234 = $16;
      $1235 = ($1233>>>0)>($1234>>>0);
      if ($1235) {
       $11 = 0;
       break;
      }
     }
     $1236 = $lastRun$i58;
     $1237 = ($1236>>>0)>=(15);
     $1238 = $lastRun$i58;
     if ($1237) {
      $1239 = (($1238) - 15)|0;
      $accumulator$i59 = $1239;
      $1240 = $op$i43;
      $1241 = ((($1240)) + 1|0);
      $op$i43 = $1241;
      HEAP8[$1240>>0] = -16;
      while(1) {
       $1242 = $accumulator$i59;
       $1243 = ($1242>>>0)>=(255);
       if (!($1243)) {
        break;
       }
       $1244 = $op$i43;
       $1245 = ((($1244)) + 1|0);
       $op$i43 = $1245;
       HEAP8[$1244>>0] = -1;
       $1246 = $accumulator$i59;
       $1247 = (($1246) - 255)|0;
       $accumulator$i59 = $1247;
      }
      $1248 = $accumulator$i59;
      $1249 = $1248&255;
      $1250 = $op$i43;
      $1251 = ((($1250)) + 1|0);
      $op$i43 = $1251;
      HEAP8[$1250>>0] = $1249;
     } else {
      $1252 = $1238 << 4;
      $1253 = $1252&255;
      $1254 = $op$i43;
      $1255 = ((($1254)) + 1|0);
      $op$i43 = $1255;
      HEAP8[$1254>>0] = $1253;
     }
     $1256 = $op$i43;
     $1257 = $anchor$i39;
     $1258 = $lastRun$i58;
     _memcpy(($1256|0),($1257|0),($1258|0))|0;
     $1259 = $lastRun$i58;
     $1260 = $op$i43;
     $1261 = (($1260) + ($1259)|0);
     $op$i43 = $1261;
     $1262 = $op$i43;
     $1263 = $14;
     $1264 = $1262;
     $1265 = $1263;
     $1266 = (($1264) - ($1265))|0;
     $11 = $1266;
    }
   } while(0);
   $1267 = $11;
   $44 = $1267;
   $1671 = $44;
   STACKTOP = sp;return ($1671|0);
  } else {
   $1268 = (_LZ4_64bits()|0);
   $1269 = ($1268|0)!=(0);
   $1270 = $1269 ? 1 : 0;
   $1271 = $50;
   $1 = $60;
   $2 = $61;
   $3 = $62;
   $4 = $63;
   $5 = $867;
   $6 = 1;
   $7 = $1270;
   $8 = 0;
   $9 = 0;
   $10 = $1271;
   $1272 = $1;
   $dictPtr$i61 = $1272;
   $1273 = $2;
   $ip$i62 = $1273;
   $1274 = $ip$i62;
   $1275 = $dictPtr$i61;
   $1276 = ((($1275)) + 16400|0);
   $1277 = HEAP32[$1276>>2]|0;
   $1278 = (0 - ($1277))|0;
   $1279 = (($1274) + ($1278)|0);
   $lowRefLimit$i65 = $1279;
   $1280 = $dictPtr$i61;
   $1281 = ((($1280)) + 16392|0);
   $1282 = HEAP32[$1281>>2]|0;
   $dictionary$i66 = $1282;
   $1283 = $dictionary$i66;
   $1284 = $dictPtr$i61;
   $1285 = ((($1284)) + 16400|0);
   $1286 = HEAP32[$1285>>2]|0;
   $1287 = (($1283) + ($1286)|0);
   $dictEnd$i67 = $1287;
   $1288 = $dictEnd$i67;
   $1289 = $2;
   $1290 = $1288;
   $1291 = $1289;
   $1292 = (($1290) - ($1291))|0;
   $dictDelta$i68 = $1292;
   $1293 = $2;
   $anchor$i69 = $1293;
   $1294 = $ip$i62;
   $1295 = $4;
   $1296 = (($1294) + ($1295)|0);
   $iend$i70 = $1296;
   $1297 = $iend$i70;
   $1298 = ((($1297)) + -12|0);
   $mflimit$i71 = $1298;
   $1299 = $iend$i70;
   $1300 = ((($1299)) + -5|0);
   $matchlimit$i72 = $1300;
   $1301 = $3;
   $op$i73 = $1301;
   $1302 = $op$i73;
   $1303 = $5;
   $1304 = (($1302) + ($1303)|0);
   $olimit$i74 = $1304;
   $refDelta$i76 = 0;
   $1305 = $4;
   $1306 = ($1305>>>0)>(2113929216);
   L331: do {
    if ($1306) {
     $0 = 0;
    } else {
     $1307 = $8;
     switch ($1307|0) {
     case 2:  {
      $1322 = $2;
      $1323 = $dictPtr$i61;
      $1324 = ((($1323)) + 16384|0);
      $1325 = HEAP32[$1324>>2]|0;
      $1326 = (0 - ($1325))|0;
      $1327 = (($1322) + ($1326)|0);
      $base$i63 = $1327;
      $1328 = $2;
      $lowLimit$i64 = $1328;
      break;
     }
     case 1:  {
      $1310 = $2;
      $1311 = $dictPtr$i61;
      $1312 = ((($1311)) + 16384|0);
      $1313 = HEAP32[$1312>>2]|0;
      $1314 = (0 - ($1313))|0;
      $1315 = (($1310) + ($1314)|0);
      $base$i63 = $1315;
      $1316 = $2;
      $1317 = $dictPtr$i61;
      $1318 = ((($1317)) + 16400|0);
      $1319 = HEAP32[$1318>>2]|0;
      $1320 = (0 - ($1319))|0;
      $1321 = (($1316) + ($1320)|0);
      $lowLimit$i64 = $1321;
      break;
     }
     default: {
      $1308 = $2;
      $base$i63 = $1308;
      $1309 = $2;
      $lowLimit$i64 = $1309;
     }
     }
     $1329 = $7;
     $1330 = ($1329|0)==(2);
     $1331 = $4;
     $1332 = ($1331|0)>=(65547);
     $or$cond12 = $1330 & $1332;
     if ($or$cond12) {
      $0 = 0;
      break;
     }
     $1333 = $4;
     $1334 = ($1333|0)<(13);
     L342: do {
      if (!($1334)) {
       $1335 = $ip$i62;
       $1336 = $1;
       $1337 = $7;
       $1338 = $base$i63;
       _LZ4_putPosition($1335,$1336,$1337,$1338);
       $1339 = $ip$i62;
       $1340 = ((($1339)) + 1|0);
       $ip$i62 = $1340;
       $1341 = $ip$i62;
       $1342 = $7;
       $1343 = (_LZ4_hashPosition($1341,$1342)|0);
       $forwardH$i75 = $1343;
       L344: while(1) {
        $1344 = $ip$i62;
        $forwardIp$i79 = $1344;
        $step$i80 = 1;
        $1345 = $10;
        $1346 = $1345 << 6;
        $searchMatchNb$i81 = $1346;
        while(1) {
         $1347 = $forwardH$i75;
         $h$i82 = $1347;
         $1348 = $forwardIp$i79;
         $ip$i62 = $1348;
         $1349 = $step$i80;
         $1350 = $forwardIp$i79;
         $1351 = (($1350) + ($1349)|0);
         $forwardIp$i79 = $1351;
         $1352 = $searchMatchNb$i81;
         $1353 = (($1352) + 1)|0;
         $searchMatchNb$i81 = $1353;
         $1354 = $1352 >>> 6;
         $step$i80 = $1354;
         $1355 = $forwardIp$i79;
         $1356 = $mflimit$i71;
         $1357 = ($1355>>>0)>($1356>>>0);
         if ($1357) {
          break L342;
         }
         $1358 = $h$i82;
         $1359 = $1;
         $1360 = $7;
         $1361 = $base$i63;
         $1362 = (_LZ4_getPositionOnHash($1358,$1359,$1360,$1361)|0);
         $match$i77 = $1362;
         $1363 = $8;
         $1364 = ($1363|0)==(2);
         do {
          if ($1364) {
           $1365 = $match$i77;
           $1366 = $2;
           $1367 = ($1365>>>0)<($1366>>>0);
           if ($1367) {
            $1368 = $dictDelta$i68;
            $refDelta$i76 = $1368;
            $1369 = $dictionary$i66;
            $lowLimit$i64 = $1369;
            break;
           } else {
            $refDelta$i76 = 0;
            $1370 = $2;
            $lowLimit$i64 = $1370;
            break;
           }
          }
         } while(0);
         $1371 = $forwardIp$i79;
         $1372 = $7;
         $1373 = (_LZ4_hashPosition($1371,$1372)|0);
         $forwardH$i75 = $1373;
         $1374 = $ip$i62;
         $1375 = $h$i82;
         $1376 = $1;
         $1377 = $7;
         $1378 = $base$i63;
         _LZ4_putPositionOnHash($1374,$1375,$1376,$1377,$1378);
         $1379 = $9;
         $1380 = ($1379|0)==(1);
         if ($1380) {
          $1381 = $match$i77;
          $1382 = $lowRefLimit$i65;
          $1383 = ($1381>>>0)<($1382>>>0);
          if ($1383) {
           continue;
          }
         }
         $1384 = $7;
         $1385 = ($1384|0)==(2);
         if (!($1385)) {
          $1386 = $match$i77;
          $1387 = ((($1386)) + 65535|0);
          $1388 = $ip$i62;
          $1389 = ($1387>>>0)<($1388>>>0);
          if ($1389) {
           continue;
          }
         }
         $1390 = $match$i77;
         $1391 = $refDelta$i76;
         $1392 = (($1390) + ($1391)|0);
         $1393 = (_LZ4_read32($1392)|0);
         $1394 = $ip$i62;
         $1395 = (_LZ4_read32($1394)|0);
         $1396 = ($1393|0)!=($1395|0);
         if (!($1396)) {
          break;
         }
        }
        while(1) {
         $1397 = $ip$i62;
         $1398 = $anchor$i69;
         $1399 = ($1397>>>0)>($1398>>>0);
         if ($1399) {
          $1400 = $match$i77;
          $1401 = $refDelta$i76;
          $1402 = (($1400) + ($1401)|0);
          $1403 = $lowLimit$i64;
          $1404 = ($1402>>>0)>($1403>>>0);
          if ($1404) {
           $1405 = $ip$i62;
           $1406 = ((($1405)) + -1|0);
           $1407 = HEAP8[$1406>>0]|0;
           $1408 = $1407&255;
           $1409 = $refDelta$i76;
           $1410 = (($1409) - 1)|0;
           $1411 = $match$i77;
           $1412 = (($1411) + ($1410)|0);
           $1413 = HEAP8[$1412>>0]|0;
           $1414 = $1413&255;
           $1415 = ($1408|0)==($1414|0);
           $1675 = $1415;
          } else {
           $1675 = 0;
          }
         } else {
          $1675 = 0;
         }
         $1416 = $ip$i62;
         if (!($1675)) {
          break;
         }
         $1417 = ((($1416)) + -1|0);
         $ip$i62 = $1417;
         $1418 = $match$i77;
         $1419 = ((($1418)) + -1|0);
         $match$i77 = $1419;
        }
        $1420 = $anchor$i69;
        $1421 = $1416;
        $1422 = $1420;
        $1423 = (($1421) - ($1422))|0;
        $litLength$i83 = $1423;
        $1424 = $op$i73;
        $1425 = ((($1424)) + 1|0);
        $op$i73 = $1425;
        $token$i78 = $1424;
        $1426 = $6;
        $1427 = ($1426|0)!=(0);
        if ($1427) {
         $1428 = $op$i73;
         $1429 = $litLength$i83;
         $1430 = (($1428) + ($1429)|0);
         $1431 = ((($1430)) + 8|0);
         $1432 = $litLength$i83;
         $1433 = (($1432>>>0) / 255)&-1;
         $1434 = (($1431) + ($1433)|0);
         $1435 = $olimit$i74;
         $1436 = ($1434>>>0)>($1435>>>0);
         if ($1436) {
          label = 262;
          break;
         }
        }
        $1437 = $litLength$i83;
        $1438 = ($1437>>>0)>=(15);
        $1439 = $litLength$i83;
        if ($1438) {
         $1440 = (($1439) - 15)|0;
         $len$i84 = $1440;
         $1441 = $token$i78;
         HEAP8[$1441>>0] = -16;
         while(1) {
          $1442 = $len$i84;
          $1443 = ($1442|0)>=(255);
          if (!($1443)) {
           break;
          }
          $1444 = $op$i73;
          $1445 = ((($1444)) + 1|0);
          $op$i73 = $1445;
          HEAP8[$1444>>0] = -1;
          $1446 = $len$i84;
          $1447 = (($1446) - 255)|0;
          $len$i84 = $1447;
         }
         $1448 = $len$i84;
         $1449 = $1448&255;
         $1450 = $op$i73;
         $1451 = ((($1450)) + 1|0);
         $op$i73 = $1451;
         HEAP8[$1450>>0] = $1449;
        } else {
         $1452 = $1439 << 4;
         $1453 = $1452&255;
         $1454 = $token$i78;
         HEAP8[$1454>>0] = $1453;
        }
        $1455 = $op$i73;
        $1456 = $anchor$i69;
        $1457 = $op$i73;
        $1458 = $litLength$i83;
        $1459 = (($1457) + ($1458)|0);
        _LZ4_wildCopy($1455,$1456,$1459);
        $1460 = $litLength$i83;
        $1461 = $op$i73;
        $1462 = (($1461) + ($1460)|0);
        $op$i73 = $1462;
        while(1) {
         $1463 = $op$i73;
         $1464 = $ip$i62;
         $1465 = $match$i77;
         $1466 = $1464;
         $1467 = $1465;
         $1468 = (($1466) - ($1467))|0;
         $1469 = $1468&65535;
         _LZ4_writeLE16($1463,$1469);
         $1470 = $op$i73;
         $1471 = ((($1470)) + 2|0);
         $op$i73 = $1471;
         $1472 = $8;
         $1473 = ($1472|0)==(2);
         if ($1473) {
          $1474 = $lowLimit$i64;
          $1475 = $dictionary$i66;
          $1476 = ($1474|0)==($1475|0);
          if ($1476) {
           $1477 = $refDelta$i76;
           $1478 = $match$i77;
           $1479 = (($1478) + ($1477)|0);
           $match$i77 = $1479;
           $1480 = $ip$i62;
           $1481 = $dictEnd$i67;
           $1482 = $match$i77;
           $1483 = $1481;
           $1484 = $1482;
           $1485 = (($1483) - ($1484))|0;
           $1486 = (($1480) + ($1485)|0);
           $limit$i86 = $1486;
           $1487 = $limit$i86;
           $1488 = $matchlimit$i72;
           $1489 = ($1487>>>0)>($1488>>>0);
           if ($1489) {
            $1490 = $matchlimit$i72;
            $limit$i86 = $1490;
           }
           $1491 = $ip$i62;
           $1492 = ((($1491)) + 4|0);
           $1493 = $match$i77;
           $1494 = ((($1493)) + 4|0);
           $1495 = $limit$i86;
           $1496 = (_LZ4_count($1492,$1494,$1495)|0);
           $matchLength$i85 = $1496;
           $1497 = $matchLength$i85;
           $1498 = (4 + ($1497))|0;
           $1499 = $ip$i62;
           $1500 = (($1499) + ($1498)|0);
           $ip$i62 = $1500;
           $1501 = $ip$i62;
           $1502 = $limit$i86;
           $1503 = ($1501|0)==($1502|0);
           if ($1503) {
            $1504 = $ip$i62;
            $1505 = $2;
            $1506 = $matchlimit$i72;
            $1507 = (_LZ4_count($1504,$1505,$1506)|0);
            $more$i87 = $1507;
            $1508 = $more$i87;
            $1509 = $matchLength$i85;
            $1510 = (($1509) + ($1508))|0;
            $matchLength$i85 = $1510;
            $1511 = $more$i87;
            $1512 = $ip$i62;
            $1513 = (($1512) + ($1511)|0);
            $ip$i62 = $1513;
           }
          } else {
           label = 276;
          }
         } else {
          label = 276;
         }
         if ((label|0) == 276) {
          label = 0;
          $1514 = $ip$i62;
          $1515 = ((($1514)) + 4|0);
          $1516 = $match$i77;
          $1517 = ((($1516)) + 4|0);
          $1518 = $matchlimit$i72;
          $1519 = (_LZ4_count($1515,$1517,$1518)|0);
          $matchLength$i85 = $1519;
          $1520 = $matchLength$i85;
          $1521 = (4 + ($1520))|0;
          $1522 = $ip$i62;
          $1523 = (($1522) + ($1521)|0);
          $ip$i62 = $1523;
         }
         $1524 = $6;
         $1525 = ($1524|0)!=(0);
         if ($1525) {
          $1526 = $op$i73;
          $1527 = ((($1526)) + 6|0);
          $1528 = $matchLength$i85;
          $1529 = $1528 >>> 8;
          $1530 = (($1527) + ($1529)|0);
          $1531 = $olimit$i74;
          $1532 = ($1530>>>0)>($1531>>>0);
          if ($1532) {
           label = 279;
           break L344;
          }
         }
         $1533 = $matchLength$i85;
         $1534 = ($1533>>>0)>=(15);
         if ($1534) {
          $1535 = $token$i78;
          $1536 = HEAP8[$1535>>0]|0;
          $1537 = $1536&255;
          $1538 = (($1537) + 15)|0;
          $1539 = $1538&255;
          HEAP8[$1535>>0] = $1539;
          $1540 = $matchLength$i85;
          $1541 = (($1540) - 15)|0;
          $matchLength$i85 = $1541;
          while(1) {
           $1542 = $matchLength$i85;
           $1543 = ($1542>>>0)>=(510);
           if (!($1543)) {
            break;
           }
           $1544 = $op$i73;
           $1545 = ((($1544)) + 1|0);
           $op$i73 = $1545;
           HEAP8[$1544>>0] = -1;
           $1546 = $op$i73;
           $1547 = ((($1546)) + 1|0);
           $op$i73 = $1547;
           HEAP8[$1546>>0] = -1;
           $1548 = $matchLength$i85;
           $1549 = (($1548) - 510)|0;
           $matchLength$i85 = $1549;
          }
          $1550 = $matchLength$i85;
          $1551 = ($1550>>>0)>=(255);
          if ($1551) {
           $1552 = $matchLength$i85;
           $1553 = (($1552) - 255)|0;
           $matchLength$i85 = $1553;
           $1554 = $op$i73;
           $1555 = ((($1554)) + 1|0);
           $op$i73 = $1555;
           HEAP8[$1554>>0] = -1;
          }
          $1556 = $matchLength$i85;
          $1557 = $1556&255;
          $1558 = $op$i73;
          $1559 = ((($1558)) + 1|0);
          $op$i73 = $1559;
          HEAP8[$1558>>0] = $1557;
         } else {
          $1560 = $matchLength$i85;
          $1561 = $1560&255;
          $1562 = $1561&255;
          $1563 = $token$i78;
          $1564 = HEAP8[$1563>>0]|0;
          $1565 = $1564&255;
          $1566 = (($1565) + ($1562))|0;
          $1567 = $1566&255;
          HEAP8[$1563>>0] = $1567;
         }
         $1568 = $ip$i62;
         $anchor$i69 = $1568;
         $1569 = $ip$i62;
         $1570 = $mflimit$i71;
         $1571 = ($1569>>>0)>($1570>>>0);
         if ($1571) {
          break L342;
         }
         $1572 = $ip$i62;
         $1573 = ((($1572)) + -2|0);
         $1574 = $1;
         $1575 = $7;
         $1576 = $base$i63;
         _LZ4_putPosition($1573,$1574,$1575,$1576);
         $1577 = $ip$i62;
         $1578 = $1;
         $1579 = $7;
         $1580 = $base$i63;
         $1581 = (_LZ4_getPosition($1577,$1578,$1579,$1580)|0);
         $match$i77 = $1581;
         $1582 = $8;
         $1583 = ($1582|0)==(2);
         do {
          if ($1583) {
           $1584 = $match$i77;
           $1585 = $2;
           $1586 = ($1584>>>0)<($1585>>>0);
           if ($1586) {
            $1587 = $dictDelta$i68;
            $refDelta$i76 = $1587;
            $1588 = $dictionary$i66;
            $lowLimit$i64 = $1588;
            break;
           } else {
            $refDelta$i76 = 0;
            $1589 = $2;
            $lowLimit$i64 = $1589;
            break;
           }
          }
         } while(0);
         $1590 = $ip$i62;
         $1591 = $1;
         $1592 = $7;
         $1593 = $base$i63;
         _LZ4_putPosition($1590,$1591,$1592,$1593);
         $1594 = $9;
         $1595 = ($1594|0)==(1);
         if ($1595) {
          $1596 = $match$i77;
          $1597 = $lowRefLimit$i65;
          $1598 = ($1596>>>0)>=($1597>>>0);
          if (!($1598)) {
           break;
          }
         }
         $1599 = $match$i77;
         $1600 = ((($1599)) + 65535|0);
         $1601 = $ip$i62;
         $1602 = ($1600>>>0)>=($1601>>>0);
         if (!($1602)) {
          break;
         }
         $1603 = $match$i77;
         $1604 = $refDelta$i76;
         $1605 = (($1603) + ($1604)|0);
         $1606 = (_LZ4_read32($1605)|0);
         $1607 = $ip$i62;
         $1608 = (_LZ4_read32($1607)|0);
         $1609 = ($1606|0)==($1608|0);
         if (!($1609)) {
          break;
         }
         $1610 = $op$i73;
         $1611 = ((($1610)) + 1|0);
         $op$i73 = $1611;
         $token$i78 = $1610;
         $1612 = $token$i78;
         HEAP8[$1612>>0] = 0;
        }
        $1613 = $ip$i62;
        $1614 = ((($1613)) + 1|0);
        $ip$i62 = $1614;
        $1615 = $7;
        $1616 = (_LZ4_hashPosition($1614,$1615)|0);
        $forwardH$i75 = $1616;
       }
       if ((label|0) == 262) {
        $0 = 0;
        break L331;
       }
       else if ((label|0) == 279) {
        $0 = 0;
        break L331;
       }
      }
     } while(0);
     $1617 = $iend$i70;
     $1618 = $anchor$i69;
     $1619 = $1617;
     $1620 = $1618;
     $1621 = (($1619) - ($1620))|0;
     $lastRun$i88 = $1621;
     $1622 = $6;
     $1623 = ($1622|0)!=(0);
     if ($1623) {
      $1624 = $op$i73;
      $1625 = $3;
      $1626 = $1624;
      $1627 = $1625;
      $1628 = (($1626) - ($1627))|0;
      $1629 = $lastRun$i88;
      $1630 = (($1628) + ($1629))|0;
      $1631 = (($1630) + 1)|0;
      $1632 = $lastRun$i88;
      $1633 = (($1632) + 255)|0;
      $1634 = (($1633) - 15)|0;
      $1635 = (($1634>>>0) / 255)&-1;
      $1636 = (($1631) + ($1635))|0;
      $1637 = $5;
      $1638 = ($1636>>>0)>($1637>>>0);
      if ($1638) {
       $0 = 0;
       break;
      }
     }
     $1639 = $lastRun$i88;
     $1640 = ($1639>>>0)>=(15);
     $1641 = $lastRun$i88;
     if ($1640) {
      $1642 = (($1641) - 15)|0;
      $accumulator$i89 = $1642;
      $1643 = $op$i73;
      $1644 = ((($1643)) + 1|0);
      $op$i73 = $1644;
      HEAP8[$1643>>0] = -16;
      while(1) {
       $1645 = $accumulator$i89;
       $1646 = ($1645>>>0)>=(255);
       if (!($1646)) {
        break;
       }
       $1647 = $op$i73;
       $1648 = ((($1647)) + 1|0);
       $op$i73 = $1648;
       HEAP8[$1647>>0] = -1;
       $1649 = $accumulator$i89;
       $1650 = (($1649) - 255)|0;
       $accumulator$i89 = $1650;
      }
      $1651 = $accumulator$i89;
      $1652 = $1651&255;
      $1653 = $op$i73;
      $1654 = ((($1653)) + 1|0);
      $op$i73 = $1654;
      HEAP8[$1653>>0] = $1652;
     } else {
      $1655 = $1641 << 4;
      $1656 = $1655&255;
      $1657 = $op$i73;
      $1658 = ((($1657)) + 1|0);
      $op$i73 = $1658;
      HEAP8[$1657>>0] = $1656;
     }
     $1659 = $op$i73;
     $1660 = $anchor$i69;
     $1661 = $lastRun$i88;
     _memcpy(($1659|0),($1660|0),($1661|0))|0;
     $1662 = $lastRun$i88;
     $1663 = $op$i73;
     $1664 = (($1663) + ($1662)|0);
     $op$i73 = $1664;
     $1665 = $op$i73;
     $1666 = $3;
     $1667 = $1665;
     $1668 = $1666;
     $1669 = (($1667) - ($1668))|0;
     $0 = $1669;
    }
   } while(0);
   $1670 = $0;
   $44 = $1670;
   $1671 = $44;
   STACKTOP = sp;return ($1671|0);
  }
 }
 return (0)|0;
}
function _LZ4_resetStream($LZ4_stream) {
 $LZ4_stream = $LZ4_stream|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $LZ4_stream;
 $1 = $0;
 _memset(($1|0),0,16416)|0;
 STACKTOP = sp;return;
}
function _LZ4_putPosition($p,$tableBase,$tableType,$srcBase) {
 $p = $p|0;
 $tableBase = $tableBase|0;
 $tableType = $tableType|0;
 $srcBase = $srcBase|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $h = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $p;
 $1 = $tableBase;
 $2 = $tableType;
 $3 = $srcBase;
 $4 = $0;
 $5 = $2;
 $6 = (_LZ4_hashPosition($4,$5)|0);
 $h = $6;
 $7 = $0;
 $8 = $h;
 $9 = $1;
 $10 = $2;
 $11 = $3;
 _LZ4_putPositionOnHash($7,$8,$9,$10,$11);
 STACKTOP = sp;return;
}
function _LZ4_hashPosition($p,$tableType) {
 $p = $p|0;
 $tableType = $tableType|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $p;
 $1 = $tableType;
 $2 = $0;
 $3 = (_LZ4_read_ARCH($2)|0);
 $4 = $1;
 $5 = (_LZ4_hashSequenceT($3,$4)|0);
 STACKTOP = sp;return ($5|0);
}
function _LZ4_read_ARCH($p) {
 $p = $p|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $p;
 $2 = (_LZ4_64bits()|0);
 $3 = ($2|0)!=(0);
 $4 = $1;
 if ($3) {
  $5 = (_LZ4_read64($4)|0);
  $6 = tempRet0;
  $0 = $5;
 } else {
  $7 = (_LZ4_read32($4)|0);
  $0 = $7;
 }
 $8 = $0;
 STACKTOP = sp;return ($8|0);
}
function _LZ4_64bits() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _LZ4_read64($memPtr) {
 $memPtr = $memPtr|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $val64 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $val64 = sp;
 $0 = $memPtr;
 $1 = $0;
 ;HEAP8[$val64>>0]=HEAP8[$1>>0]|0;HEAP8[$val64+1>>0]=HEAP8[$1+1>>0]|0;HEAP8[$val64+2>>0]=HEAP8[$1+2>>0]|0;HEAP8[$val64+3>>0]=HEAP8[$1+3>>0]|0;HEAP8[$val64+4>>0]=HEAP8[$1+4>>0]|0;HEAP8[$val64+5>>0]=HEAP8[$1+5>>0]|0;HEAP8[$val64+6>>0]=HEAP8[$1+6>>0]|0;HEAP8[$val64+7>>0]=HEAP8[$1+7>>0]|0;
 $2 = $val64;
 $3 = $2;
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2) + 4)|0;
 $6 = $5;
 $7 = HEAP32[$6>>2]|0;
 tempRet0 = ($7);
 STACKTOP = sp;return ($4|0);
}
function _LZ4_read32($memPtr) {
 $memPtr = $memPtr|0;
 var $0 = 0, $1 = 0, $2 = 0, $val32 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $val32 = sp;
 $0 = $memPtr;
 $1 = $0;
 ;HEAP8[$val32>>0]=HEAP8[$1>>0]|0;HEAP8[$val32+1>>0]=HEAP8[$1+1>>0]|0;HEAP8[$val32+2>>0]=HEAP8[$1+2>>0]|0;HEAP8[$val32+3>>0]=HEAP8[$1+3>>0]|0;
 $2 = HEAP32[$val32>>2]|0;
 STACKTOP = sp;return ($2|0);
}
function _LZ4_hashSequenceT($sequence,$tableType) {
 $sequence = $sequence|0;
 $tableType = $tableType|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $sequence;
 $2 = $tableType;
 $3 = (_LZ4_64bits()|0);
 $4 = ($3|0)!=(0);
 $5 = $1;
 $6 = $2;
 if ($4) {
  $7 = (_LZ4_hashSequence64($5,$6)|0);
  $0 = $7;
  $9 = $0;
  STACKTOP = sp;return ($9|0);
 } else {
  $8 = (_LZ4_hashSequence($5,$6)|0);
  $0 = $8;
  $9 = $0;
  STACKTOP = sp;return ($9|0);
 }
 return (0)|0;
}
function _LZ4_hashSequence64($sequence,$tableType) {
 $sequence = $sequence|0;
 $tableType = $tableType|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hashLog = 0, $hashMask = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $sequence;
 $1 = $tableType;
 $2 = $1;
 $3 = ($2|0)==(2);
 $4 = $3 ? 13 : 12;
 $hashLog = $4;
 $5 = $hashLog;
 $6 = 1 << $5;
 $7 = (($6) - 1)|0;
 $hashMask = $7;
 $8 = $0;
 $9 = (___muldi3(($8|0),0,465362107,207)|0);
 $10 = tempRet0;
 $11 = $hashLog;
 $12 = (40 - ($11))|0;
 $13 = (_bitshift64Lshr(($9|0),($10|0),($12|0))|0);
 $14 = tempRet0;
 $15 = $hashMask;
 $16 = $13 & $15;
 STACKTOP = sp;return ($16|0);
}
function _LZ4_hashSequence($sequence,$tableType) {
 $sequence = $sequence|0;
 $tableType = $tableType|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $sequence;
 $2 = $tableType;
 $3 = $2;
 $4 = ($3|0)==(2);
 $5 = $1;
 $6 = Math_imul($5, -1640531535)|0;
 if ($4) {
  $7 = $6 >>> 19;
  $0 = $7;
  $9 = $0;
  STACKTOP = sp;return ($9|0);
 } else {
  $8 = $6 >>> 20;
  $0 = $8;
  $9 = $0;
  STACKTOP = sp;return ($9|0);
 }
 return (0)|0;
}
function _LZ4_putPositionOnHash($p,$h,$tableBase,$tableType,$srcBase) {
 $p = $p|0;
 $h = $h|0;
 $tableBase = $tableBase|0;
 $tableType = $tableType|0;
 $srcBase = $srcBase|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hashTable = 0, $hashTable1 = 0, $hashTable2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $p;
 $1 = $h;
 $2 = $tableBase;
 $3 = $tableType;
 $4 = $srcBase;
 $5 = $3;
 switch ($5|0) {
 case 0:  {
  $6 = $2;
  $hashTable = $6;
  $7 = $0;
  $8 = $1;
  $9 = $hashTable;
  $10 = (($9) + ($8<<2)|0);
  HEAP32[$10>>2] = $7;
  STACKTOP = sp;return;
  break;
 }
 case 1:  {
  $11 = $2;
  $hashTable1 = $11;
  $12 = $0;
  $13 = $4;
  $14 = $12;
  $15 = $13;
  $16 = (($14) - ($15))|0;
  $17 = $1;
  $18 = $hashTable1;
  $19 = (($18) + ($17<<2)|0);
  HEAP32[$19>>2] = $16;
  STACKTOP = sp;return;
  break;
 }
 case 2:  {
  $20 = $2;
  $hashTable2 = $20;
  $21 = $0;
  $22 = $4;
  $23 = $21;
  $24 = $22;
  $25 = (($23) - ($24))|0;
  $26 = $25&65535;
  $27 = $1;
  $28 = $hashTable2;
  $29 = (($28) + ($27<<1)|0);
  HEAP16[$29>>1] = $26;
  STACKTOP = sp;return;
  break;
 }
 default: {
  STACKTOP = sp;return;
 }
 }
}
function _LZ4_getPositionOnHash($h,$tableBase,$tableType,$srcBase) {
 $h = $h|0;
 $tableBase = $tableBase|0;
 $tableType = $tableType|0;
 $srcBase = $srcBase|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hashTable = 0, $hashTable1 = 0, $hashTable2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $h;
 $2 = $tableBase;
 $3 = $tableType;
 $4 = $srcBase;
 $5 = $3;
 $6 = ($5|0)==(0);
 if ($6) {
  $7 = $2;
  $hashTable = $7;
  $8 = $1;
  $9 = $hashTable;
  $10 = (($9) + ($8<<2)|0);
  $11 = HEAP32[$10>>2]|0;
  $0 = $11;
  $28 = $0;
  STACKTOP = sp;return ($28|0);
 }
 $12 = $3;
 $13 = ($12|0)==(1);
 $14 = $2;
 if ($13) {
  $hashTable1 = $14;
  $15 = $1;
  $16 = $hashTable1;
  $17 = (($16) + ($15<<2)|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = $4;
  $20 = (($19) + ($18)|0);
  $0 = $20;
  $28 = $0;
  STACKTOP = sp;return ($28|0);
 } else {
  $hashTable2 = $14;
  $21 = $1;
  $22 = $hashTable2;
  $23 = (($22) + ($21<<1)|0);
  $24 = HEAP16[$23>>1]|0;
  $25 = $24&65535;
  $26 = $4;
  $27 = (($26) + ($25)|0);
  $0 = $27;
  $28 = $0;
  STACKTOP = sp;return ($28|0);
 }
 return (0)|0;
}
function _LZ4_wildCopy($dstPtr,$srcPtr,$dstEnd) {
 $dstPtr = $dstPtr|0;
 $srcPtr = $srcPtr|0;
 $dstEnd = $dstEnd|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $d = 0, $e = 0, $s = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $dstPtr;
 $1 = $srcPtr;
 $2 = $dstEnd;
 $3 = $0;
 $d = $3;
 $4 = $1;
 $s = $4;
 $5 = $2;
 $e = $5;
 while(1) {
  $6 = $d;
  $7 = $s;
  _LZ4_copy8($6,$7);
  $8 = $d;
  $9 = ((($8)) + 8|0);
  $d = $9;
  $10 = $s;
  $11 = ((($10)) + 8|0);
  $s = $11;
  $12 = $d;
  $13 = $e;
  $14 = ($12>>>0)<($13>>>0);
  if (!($14)) {
   break;
  }
 }
 STACKTOP = sp;return;
}
function _LZ4_copy8($dstPtr,$srcPtr) {
 $dstPtr = $dstPtr|0;
 $srcPtr = $srcPtr|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $dstPtr;
 $1 = $srcPtr;
 $2 = $0;
 $3 = $1;
 ;HEAP8[$2>>0]=HEAP8[$3>>0]|0;HEAP8[$2+1>>0]=HEAP8[$3+1>>0]|0;HEAP8[$2+2>>0]=HEAP8[$3+2>>0]|0;HEAP8[$2+3>>0]=HEAP8[$3+3>>0]|0;HEAP8[$2+4>>0]=HEAP8[$3+4>>0]|0;HEAP8[$2+5>>0]=HEAP8[$3+5>>0]|0;HEAP8[$2+6>>0]=HEAP8[$3+6>>0]|0;HEAP8[$2+7>>0]=HEAP8[$3+7>>0]|0;
 STACKTOP = sp;return;
}
function _LZ4_writeLE16($memPtr,$value) {
 $memPtr = $memPtr|0;
 $value = $value|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = sp + 8|0;
 $0 = $memPtr;
 HEAP16[$1>>1] = $value;
 $2 = (_LZ4_isLittleEndian()|0);
 $3 = ($2|0)!=(0);
 $4 = $0;
 if ($3) {
  ;HEAP8[$4>>0]=HEAP8[$1>>0]|0;HEAP8[$4+1>>0]=HEAP8[$1+1>>0]|0;
  STACKTOP = sp;return;
 } else {
  $p = $4;
  $5 = HEAP16[$1>>1]|0;
  $6 = $5&255;
  $7 = $p;
  HEAP8[$7>>0] = $6;
  $8 = HEAP16[$1>>1]|0;
  $9 = $8&65535;
  $10 = $9 >> 8;
  $11 = $10&255;
  $12 = $p;
  $13 = ((($12)) + 1|0);
  HEAP8[$13>>0] = $11;
  STACKTOP = sp;return;
 }
}
function _LZ4_isLittleEndian() {
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[21828]|0;
 $1 = $0&255;
 return ($1|0);
}
function _LZ4_count($pIn,$pMatch,$pInLimit) {
 $pIn = $pIn|0;
 $pMatch = $pMatch|0;
 $pInLimit = $pInLimit|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $diff = 0;
 var $pStart = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $pIn;
 $2 = $pMatch;
 $3 = $pInLimit;
 $4 = $1;
 $pStart = $4;
 while(1) {
  $5 = $1;
  $6 = $3;
  $7 = ((($6)) + -3|0);
  $8 = ($5>>>0)<($7>>>0);
  $9 = $8&1;
  $10 = ($9|0)!=(0);
  if (!($10)) {
   break;
  }
  $11 = $2;
  $12 = (_LZ4_read_ARCH($11)|0);
  $13 = $1;
  $14 = (_LZ4_read_ARCH($13)|0);
  $15 = $12 ^ $14;
  $diff = $15;
  $16 = $diff;
  $17 = ($16|0)!=(0);
  if ($17) {
   label = 5;
   break;
  }
  $18 = $1;
  $19 = ((($18)) + 4|0);
  $1 = $19;
  $20 = $2;
  $21 = ((($20)) + 4|0);
  $2 = $21;
 }
 if ((label|0) == 5) {
  $22 = $diff;
  $23 = (_LZ4_NbCommonBytes($22)|0);
  $24 = $1;
  $25 = (($24) + ($23)|0);
  $1 = $25;
  $26 = $1;
  $27 = $pStart;
  $28 = $26;
  $29 = $27;
  $30 = (($28) - ($29))|0;
  $0 = $30;
  $78 = $0;
  STACKTOP = sp;return ($78|0);
 }
 $31 = (_LZ4_64bits()|0);
 $32 = ($31|0)!=(0);
 if ($32) {
  $33 = $1;
  $34 = $3;
  $35 = ((($34)) + -3|0);
  $36 = ($33>>>0)<($35>>>0);
  if ($36) {
   $37 = $2;
   $38 = (_LZ4_read32($37)|0);
   $39 = $1;
   $40 = (_LZ4_read32($39)|0);
   $41 = ($38|0)==($40|0);
   if ($41) {
    $42 = $1;
    $43 = ((($42)) + 4|0);
    $1 = $43;
    $44 = $2;
    $45 = ((($44)) + 4|0);
    $2 = $45;
   }
  }
 }
 $46 = $1;
 $47 = $3;
 $48 = ((($47)) + -1|0);
 $49 = ($46>>>0)<($48>>>0);
 if ($49) {
  $50 = $2;
  $51 = (_LZ4_read16($50)|0);
  $52 = $51&65535;
  $53 = $1;
  $54 = (_LZ4_read16($53)|0);
  $55 = $54&65535;
  $56 = ($52|0)==($55|0);
  if ($56) {
   $57 = $1;
   $58 = ((($57)) + 2|0);
   $1 = $58;
   $59 = $2;
   $60 = ((($59)) + 2|0);
   $2 = $60;
  }
 }
 $61 = $1;
 $62 = $3;
 $63 = ($61>>>0)<($62>>>0);
 if ($63) {
  $64 = $2;
  $65 = HEAP8[$64>>0]|0;
  $66 = $65&255;
  $67 = $1;
  $68 = HEAP8[$67>>0]|0;
  $69 = $68&255;
  $70 = ($66|0)==($69|0);
  if ($70) {
   $71 = $1;
   $72 = ((($71)) + 1|0);
   $1 = $72;
  }
 }
 $73 = $1;
 $74 = $pStart;
 $75 = $73;
 $76 = $74;
 $77 = (($75) - ($76))|0;
 $0 = $77;
 $78 = $0;
 STACKTOP = sp;return ($78|0);
}
function _LZ4_NbCommonBytes($val) {
 $val = $val|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $val;
 $2 = (_LZ4_isLittleEndian()|0);
 $3 = ($2|0)!=(0);
 $4 = (_LZ4_64bits()|0);
 $5 = ($4|0)!=(0);
 $6 = $1;
 do {
  if ($3) {
   if ($5) {
    $7 = (_llvm_cttz_i64(($6|0),0,0)|0);
    $8 = tempRet0;
    $9 = $7 >> 3;
    $0 = $9;
    break;
   } else {
    $10 = (_llvm_cttz_i32(($6|0))|0);
    $11 = $10 >> 3;
    $0 = $11;
    break;
   }
  } else {
   if ($5) {
    $12 = (_llvm_ctlz_i64(($6|0),0,0)|0);
    $13 = tempRet0;
    $14 = $12 >> 3;
    $0 = $14;
    break;
   } else {
    $15 = (Math_clz32(($6|0))|0);
    $16 = $15 >> 3;
    $0 = $16;
    break;
   }
  }
 } while(0);
 $17 = $0;
 STACKTOP = sp;return ($17|0);
}
function _LZ4_read16($memPtr) {
 $memPtr = $memPtr|0;
 var $0 = 0, $1 = 0, $2 = 0, $val16 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $val16 = sp + 4|0;
 $0 = $memPtr;
 $1 = $0;
 ;HEAP8[$val16>>0]=HEAP8[$1>>0]|0;HEAP8[$val16+1>>0]=HEAP8[$1+1>>0]|0;
 $2 = HEAP16[$val16>>1]|0;
 STACKTOP = sp;return ($2|0);
}
function _LZ4_getPosition($p,$tableBase,$tableType,$srcBase) {
 $p = $p|0;
 $tableBase = $tableBase|0;
 $tableType = $tableType|0;
 $srcBase = $srcBase|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $h = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $p;
 $1 = $tableBase;
 $2 = $tableType;
 $3 = $srcBase;
 $4 = $0;
 $5 = $2;
 $6 = (_LZ4_hashPosition($4,$5)|0);
 $h = $6;
 $7 = $h;
 $8 = $1;
 $9 = $2;
 $10 = $3;
 $11 = (_LZ4_getPositionOnHash($7,$8,$9,$10)|0);
 STACKTOP = sp;return ($11|0);
}
function _LZ4_compress_fast($source,$dest,$inputSize,$maxOutputSize,$acceleration) {
 $source = $source|0;
 $dest = $dest|0;
 $inputSize = $inputSize|0;
 $maxOutputSize = $maxOutputSize|0;
 $acceleration = $acceleration|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ctx = 0, $ctxPtr = 0, $result = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16448|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ctx = sp;
 $0 = $source;
 $1 = $dest;
 $2 = $inputSize;
 $3 = $maxOutputSize;
 $4 = $acceleration;
 $ctxPtr = $ctx;
 $5 = $ctxPtr;
 $6 = $0;
 $7 = $1;
 $8 = $2;
 $9 = $3;
 $10 = $4;
 $11 = (_LZ4_compress_fast_extState($5,$6,$7,$8,$9,$10)|0);
 $result = $11;
 $12 = $result;
 STACKTOP = sp;return ($12|0);
}
function _LZ4_compress_default($source,$dest,$inputSize,$maxOutputSize) {
 $source = $source|0;
 $dest = $dest|0;
 $inputSize = $inputSize|0;
 $maxOutputSize = $maxOutputSize|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $source;
 $1 = $dest;
 $2 = $inputSize;
 $3 = $maxOutputSize;
 $4 = $0;
 $5 = $1;
 $6 = $2;
 $7 = $3;
 $8 = (_LZ4_compress_fast($4,$5,$6,$7,1)|0);
 STACKTOP = sp;return ($8|0);
}
function _LZ4_decompress_safe($source,$dest,$compressedSize,$maxDecompressedSize) {
 $source = $source|0;
 $dest = $dest|0;
 $compressedSize = $compressedSize|0;
 $maxDecompressedSize = $maxDecompressedSize|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0;
 var $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0;
 var $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0;
 var $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0;
 var $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0;
 var $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $checkOffset$i = 0;
 var $copyFrom$i = 0, $copySize$i = 0, $cpy$i = 0, $dec64$i = 0, $dictEnd$i = 0, $endOfMatch$i = 0, $iend$i = 0, $ip$i = 0, $length$i = 0, $lowLimit$i = 0, $match$i = 0, $oend$i = 0, $oexit$i = 0, $op$i = 0, $s$i = 0, $s1$i = 0, $safeDecode$i = 0, $token$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $12 = $source;
 $13 = $dest;
 $14 = $compressedSize;
 $15 = $maxDecompressedSize;
 $16 = $12;
 $17 = $13;
 $18 = $14;
 $19 = $15;
 $20 = $13;
 $1 = $16;
 $2 = $17;
 $3 = $18;
 $4 = $19;
 $5 = 1;
 $6 = 0;
 $7 = 0;
 $8 = 0;
 $9 = $20;
 $10 = 0;
 $11 = 0;
 $21 = $1;
 $ip$i = $21;
 $22 = $ip$i;
 $23 = $3;
 $24 = (($22) + ($23)|0);
 $iend$i = $24;
 $25 = $2;
 $op$i = $25;
 $26 = $op$i;
 $27 = $4;
 $28 = (($26) + ($27)|0);
 $oend$i = $28;
 $29 = $op$i;
 $30 = $7;
 $31 = (($29) + ($30)|0);
 $oexit$i = $31;
 $32 = $9;
 $33 = $11;
 $34 = (0 - ($33))|0;
 $35 = (($32) + ($34)|0);
 $lowLimit$i = $35;
 $36 = $10;
 $37 = $11;
 $38 = (($36) + ($37)|0);
 $dictEnd$i = $38;
 $39 = $5;
 $40 = ($39|0)==(1);
 $41 = $40&1;
 $safeDecode$i = $41;
 $42 = $safeDecode$i;
 $43 = ($42|0)!=(0);
 $44 = $11;
 $45 = ($44>>>0)<(65536);
 $46 = $43 ? $45 : 0;
 $47 = $46&1;
 $checkOffset$i = $47;
 $48 = $6;
 $49 = ($48|0)!=(0);
 if ($49) {
  $50 = $oexit$i;
  $51 = $oend$i;
  $52 = ((($51)) + -12|0);
  $53 = ($50>>>0)>($52>>>0);
  if ($53) {
   $54 = $oend$i;
   $55 = ((($54)) + -12|0);
   $oexit$i = $55;
  }
 }
 $56 = $5;
 $57 = ($56|0)!=(0);
 if ($57) {
  $58 = $4;
  $59 = ($58|0)==(0);
  if ($59) {
   $60 = $3;
   $61 = ($60|0)==(1);
   if ($61) {
    $62 = $ip$i;
    $63 = HEAP8[$62>>0]|0;
    $64 = $63&255;
    $65 = ($64|0)==(0);
    $67 = $65;
   } else {
    $67 = 0;
   }
   $66 = $67 ? 0 : -1;
   $0 = $66;
   $418 = $0;
   STACKTOP = sp;return ($418|0);
  }
 }
 $68 = $5;
 $69 = ($68|0)!=(0);
 if (!($69)) {
  $70 = $4;
  $71 = ($70|0)==(0);
  if ($71) {
   $72 = $ip$i;
   $73 = HEAP8[$72>>0]|0;
   $74 = $73&255;
   $75 = ($74|0)==(0);
   $76 = $75 ? 1 : -1;
   $0 = $76;
   $418 = $0;
   STACKTOP = sp;return ($418|0);
  }
 }
 L17: while(1) {
  $77 = $ip$i;
  $78 = ((($77)) + 1|0);
  $ip$i = $78;
  $79 = HEAP8[$77>>0]|0;
  $80 = $79&255;
  $token$i = $80;
  $81 = $token$i;
  $82 = $81 >>> 4;
  $length$i = $82;
  $83 = ($82|0)==(15);
  if ($83) {
   while(1) {
    $84 = $ip$i;
    $85 = ((($84)) + 1|0);
    $ip$i = $85;
    $86 = HEAP8[$84>>0]|0;
    $87 = $86&255;
    $s$i = $87;
    $88 = $s$i;
    $89 = $length$i;
    $90 = (($89) + ($88))|0;
    $length$i = $90;
    $91 = $5;
    $92 = ($91|0)!=(0);
    if ($92) {
     $93 = $ip$i;
     $94 = $iend$i;
     $95 = ((($94)) + -15|0);
     $96 = ($93>>>0)<($95>>>0);
     $97 = $96&1;
     $99 = $97;
    } else {
     $99 = 1;
    }
    $98 = ($99|0)!=(0);
    $100 = $s$i;
    $101 = ($100|0)==(255);
    $102 = $98 ? $101 : 0;
    if (!($102)) {
     break;
    }
   }
   $103 = $safeDecode$i;
   $104 = ($103|0)!=(0);
   if ($104) {
    $105 = $op$i;
    $106 = $length$i;
    $107 = (($105) + ($106)|0);
    $108 = $107;
    $109 = $op$i;
    $110 = $109;
    $111 = ($108>>>0)<($110>>>0);
    if ($111) {
     break;
    }
   }
   $112 = $safeDecode$i;
   $113 = ($112|0)!=(0);
   if ($113) {
    $114 = $ip$i;
    $115 = $length$i;
    $116 = (($114) + ($115)|0);
    $117 = $116;
    $118 = $ip$i;
    $119 = $118;
    $120 = ($117>>>0)<($119>>>0);
    if ($120) {
     break;
    }
   }
  }
  $121 = $op$i;
  $122 = $length$i;
  $123 = (($121) + ($122)|0);
  $cpy$i = $123;
  $124 = $5;
  $125 = ($124|0)!=(0);
  if ($125) {
   $126 = $cpy$i;
   $127 = $6;
   $128 = ($127|0)!=(0);
   $129 = $oexit$i;
   $130 = $oend$i;
   $131 = ((($130)) + -12|0);
   $132 = $128 ? $129 : $131;
   $133 = ($126>>>0)>($132>>>0);
   if ($133) {
    label = 25;
    break;
   }
   $134 = $ip$i;
   $135 = $length$i;
   $136 = (($134) + ($135)|0);
   $137 = $iend$i;
   $138 = ((($137)) + -8|0);
   $139 = ($136>>>0)>($138>>>0);
   if ($139) {
    label = 25;
    break;
   }
  }
  $140 = $5;
  $141 = ($140|0)!=(0);
  if (!($141)) {
   $142 = $cpy$i;
   $143 = $oend$i;
   $144 = ((($143)) + -8|0);
   $145 = ($142>>>0)>($144>>>0);
   if ($145) {
    label = 25;
    break;
   }
  }
  $184 = $op$i;
  $185 = $ip$i;
  $186 = $cpy$i;
  _LZ4_wildCopy($184,$185,$186);
  $187 = $length$i;
  $188 = $ip$i;
  $189 = (($188) + ($187)|0);
  $ip$i = $189;
  $190 = $cpy$i;
  $op$i = $190;
  $191 = $cpy$i;
  $192 = $ip$i;
  $193 = (_LZ4_readLE16($192)|0);
  $194 = $193&65535;
  $195 = (0 - ($194))|0;
  $196 = (($191) + ($195)|0);
  $match$i = $196;
  $197 = $ip$i;
  $198 = ((($197)) + 2|0);
  $ip$i = $198;
  $199 = $checkOffset$i;
  $200 = ($199|0)!=(0);
  if ($200) {
   $201 = $match$i;
   $202 = $lowLimit$i;
   $203 = ($201>>>0)<($202>>>0);
   if ($203) {
    break;
   }
  }
  $204 = $token$i;
  $205 = $204 & 15;
  $length$i = $205;
  $206 = $length$i;
  $207 = ($206|0)==(15);
  if ($207) {
   while(1) {
    $208 = $5;
    $209 = ($208|0)!=(0);
    if ($209) {
     $210 = $ip$i;
     $211 = $iend$i;
     $212 = ((($211)) + -5|0);
     $213 = ($210>>>0)>($212>>>0);
     if ($213) {
      break L17;
     }
    }
    $214 = $ip$i;
    $215 = ((($214)) + 1|0);
    $ip$i = $215;
    $216 = HEAP8[$214>>0]|0;
    $217 = $216&255;
    $s1$i = $217;
    $218 = $s1$i;
    $219 = $length$i;
    $220 = (($219) + ($218))|0;
    $length$i = $220;
    $221 = $s1$i;
    $222 = ($221|0)==(255);
    if (!($222)) {
     break;
    }
   }
   $223 = $safeDecode$i;
   $224 = ($223|0)!=(0);
   if ($224) {
    $225 = $op$i;
    $226 = $length$i;
    $227 = (($225) + ($226)|0);
    $228 = $227;
    $229 = $op$i;
    $230 = $229;
    $231 = ($228>>>0)<($230>>>0);
    if ($231) {
     break;
    }
   }
  }
  $232 = $length$i;
  $233 = (($232) + 4)|0;
  $length$i = $233;
  $234 = $8;
  $235 = ($234|0)==(2);
  if ($235) {
   $236 = $match$i;
   $237 = $9;
   $238 = ($236>>>0)<($237>>>0);
   if ($238) {
    $239 = $op$i;
    $240 = $length$i;
    $241 = (($239) + ($240)|0);
    $242 = $oend$i;
    $243 = ((($242)) + -5|0);
    $244 = ($241>>>0)>($243>>>0);
    if ($244) {
     break;
    }
    $245 = $length$i;
    $246 = $9;
    $247 = $match$i;
    $248 = $246;
    $249 = $247;
    $250 = (($248) - ($249))|0;
    $251 = ($245>>>0)<=($250>>>0);
    if ($251) {
     $252 = $dictEnd$i;
     $253 = $9;
     $254 = $match$i;
     $255 = $253;
     $256 = $254;
     $257 = (($255) - ($256))|0;
     $258 = (0 - ($257))|0;
     $259 = (($252) + ($258)|0);
     $match$i = $259;
     $260 = $op$i;
     $261 = $match$i;
     $262 = $length$i;
     _memmove(($260|0),($261|0),($262|0))|0;
     $263 = $length$i;
     $264 = $op$i;
     $265 = (($264) + ($263)|0);
     $op$i = $265;
     continue;
    }
    $266 = $9;
    $267 = $match$i;
    $268 = $266;
    $269 = $267;
    $270 = (($268) - ($269))|0;
    $copySize$i = $270;
    $271 = $op$i;
    $272 = $dictEnd$i;
    $273 = $copySize$i;
    $274 = (0 - ($273))|0;
    $275 = (($272) + ($274)|0);
    $276 = $copySize$i;
    _memcpy(($271|0),($275|0),($276|0))|0;
    $277 = $copySize$i;
    $278 = $op$i;
    $279 = (($278) + ($277)|0);
    $op$i = $279;
    $280 = $length$i;
    $281 = $copySize$i;
    $282 = (($280) - ($281))|0;
    $copySize$i = $282;
    $283 = $copySize$i;
    $284 = $op$i;
    $285 = $9;
    $286 = $284;
    $287 = $285;
    $288 = (($286) - ($287))|0;
    $289 = ($283>>>0)>($288>>>0);
    $290 = $op$i;
    if (!($289)) {
     $302 = $9;
     $303 = $copySize$i;
     _memcpy(($290|0),($302|0),($303|0))|0;
     $304 = $copySize$i;
     $305 = $op$i;
     $306 = (($305) + ($304)|0);
     $op$i = $306;
     continue;
    }
    $291 = $copySize$i;
    $292 = (($290) + ($291)|0);
    $endOfMatch$i = $292;
    $293 = $9;
    $copyFrom$i = $293;
    while(1) {
     $294 = $op$i;
     $295 = $endOfMatch$i;
     $296 = ($294>>>0)<($295>>>0);
     if (!($296)) {
      continue L17;
     }
     $297 = $copyFrom$i;
     $298 = ((($297)) + 1|0);
     $copyFrom$i = $298;
     $299 = HEAP8[$297>>0]|0;
     $300 = $op$i;
     $301 = ((($300)) + 1|0);
     $op$i = $301;
     HEAP8[$300>>0] = $299;
    }
   }
  }
  $307 = $op$i;
  $308 = $length$i;
  $309 = (($307) + ($308)|0);
  $cpy$i = $309;
  $310 = $op$i;
  $311 = $match$i;
  $312 = $310;
  $313 = $311;
  $314 = (($312) - ($313))|0;
  $315 = ($314|0)<(8);
  $316 = $op$i;
  $317 = $match$i;
  if ($315) {
   $318 = $316;
   $319 = $317;
   $320 = (($318) - ($319))|0;
   $321 = (21832 + ($320<<2)|0);
   $322 = HEAP32[$321>>2]|0;
   $dec64$i = $322;
   $323 = $match$i;
   $324 = HEAP8[$323>>0]|0;
   $325 = $op$i;
   HEAP8[$325>>0] = $324;
   $326 = $match$i;
   $327 = ((($326)) + 1|0);
   $328 = HEAP8[$327>>0]|0;
   $329 = $op$i;
   $330 = ((($329)) + 1|0);
   HEAP8[$330>>0] = $328;
   $331 = $match$i;
   $332 = ((($331)) + 2|0);
   $333 = HEAP8[$332>>0]|0;
   $334 = $op$i;
   $335 = ((($334)) + 2|0);
   HEAP8[$335>>0] = $333;
   $336 = $match$i;
   $337 = ((($336)) + 3|0);
   $338 = HEAP8[$337>>0]|0;
   $339 = $op$i;
   $340 = ((($339)) + 3|0);
   HEAP8[$340>>0] = $338;
   $341 = $op$i;
   $342 = $match$i;
   $343 = $341;
   $344 = $342;
   $345 = (($343) - ($344))|0;
   $346 = (21864 + ($345<<2)|0);
   $347 = HEAP32[$346>>2]|0;
   $348 = $match$i;
   $349 = (($348) + ($347)|0);
   $match$i = $349;
   $350 = $op$i;
   $351 = ((($350)) + 4|0);
   $352 = $match$i;
   _LZ4_copy4($351,$352);
   $353 = $op$i;
   $354 = ((($353)) + 8|0);
   $op$i = $354;
   $355 = $dec64$i;
   $356 = $match$i;
   $357 = (0 - ($355))|0;
   $358 = (($356) + ($357)|0);
   $match$i = $358;
  } else {
   _LZ4_copy8($316,$317);
   $359 = $op$i;
   $360 = ((($359)) + 8|0);
   $op$i = $360;
   $361 = $match$i;
   $362 = ((($361)) + 8|0);
   $match$i = $362;
  }
  $363 = $cpy$i;
  $364 = $oend$i;
  $365 = ((($364)) + -12|0);
  $366 = ($363>>>0)>($365>>>0);
  L68: do {
   if ($366) {
    $367 = $cpy$i;
    $368 = $oend$i;
    $369 = ((($368)) + -5|0);
    $370 = ($367>>>0)>($369>>>0);
    if ($370) {
     break L17;
    }
    $371 = $op$i;
    $372 = $oend$i;
    $373 = ((($372)) + -8|0);
    $374 = ($371>>>0)<($373>>>0);
    if ($374) {
     $375 = $op$i;
     $376 = $match$i;
     $377 = $oend$i;
     $378 = ((($377)) + -8|0);
     _LZ4_wildCopy($375,$376,$378);
     $379 = $oend$i;
     $380 = ((($379)) + -8|0);
     $381 = $op$i;
     $382 = $380;
     $383 = $381;
     $384 = (($382) - ($383))|0;
     $385 = $match$i;
     $386 = (($385) + ($384)|0);
     $match$i = $386;
     $387 = $oend$i;
     $388 = ((($387)) + -8|0);
     $op$i = $388;
    }
    while(1) {
     $389 = $op$i;
     $390 = $cpy$i;
     $391 = ($389>>>0)<($390>>>0);
     if (!($391)) {
      break L68;
     }
     $392 = $match$i;
     $393 = ((($392)) + 1|0);
     $match$i = $393;
     $394 = HEAP8[$392>>0]|0;
     $395 = $op$i;
     $396 = ((($395)) + 1|0);
     $op$i = $396;
     HEAP8[$395>>0] = $394;
    }
   } else {
    $397 = $op$i;
    $398 = $match$i;
    $399 = $cpy$i;
    _LZ4_wildCopy($397,$398,$399);
   }
  } while(0);
  $400 = $cpy$i;
  $op$i = $400;
 }
 do {
  if ((label|0) == 25) {
   $146 = $6;
   $147 = ($146|0)!=(0);
   if ($147) {
    $148 = $cpy$i;
    $149 = $oend$i;
    $150 = ($148>>>0)>($149>>>0);
    if ($150) {
     break;
    }
    $151 = $5;
    $152 = ($151|0)!=(0);
    if ($152) {
     $153 = $ip$i;
     $154 = $length$i;
     $155 = (($153) + ($154)|0);
     $156 = $iend$i;
     $157 = ($155>>>0)>($156>>>0);
     if ($157) {
      break;
     }
    }
   } else {
    $158 = $5;
    $159 = ($158|0)!=(0);
    if (!($159)) {
     $160 = $cpy$i;
     $161 = $oend$i;
     $162 = ($160|0)!=($161|0);
     if ($162) {
      break;
     }
    }
    $163 = $5;
    $164 = ($163|0)!=(0);
    if ($164) {
     $165 = $ip$i;
     $166 = $length$i;
     $167 = (($165) + ($166)|0);
     $168 = $iend$i;
     $169 = ($167|0)!=($168|0);
     if ($169) {
      break;
     }
     $170 = $cpy$i;
     $171 = $oend$i;
     $172 = ($170>>>0)>($171>>>0);
     if ($172) {
      break;
     }
    }
   }
   $173 = $op$i;
   $174 = $ip$i;
   $175 = $length$i;
   _memcpy(($173|0),($174|0),($175|0))|0;
   $176 = $length$i;
   $177 = $ip$i;
   $178 = (($177) + ($176)|0);
   $ip$i = $178;
   $179 = $length$i;
   $180 = $op$i;
   $181 = (($180) + ($179)|0);
   $op$i = $181;
   $182 = $5;
   $183 = ($182|0)!=(0);
   if ($183) {
    $401 = $op$i;
    $402 = $2;
    $403 = $401;
    $404 = $402;
    $405 = (($403) - ($404))|0;
    $0 = $405;
    $418 = $0;
    STACKTOP = sp;return ($418|0);
   } else {
    $406 = $ip$i;
    $407 = $1;
    $408 = $406;
    $409 = $407;
    $410 = (($408) - ($409))|0;
    $0 = $410;
    $418 = $0;
    STACKTOP = sp;return ($418|0);
   }
  }
 } while(0);
 $411 = $ip$i;
 $412 = $1;
 $413 = $411;
 $414 = $412;
 $415 = (($413) - ($414))|0;
 $416 = (0 - ($415))|0;
 $417 = (($416) - 1)|0;
 $0 = $417;
 $418 = $0;
 STACKTOP = sp;return ($418|0);
}
function _LZ4_readLE16($memPtr) {
 $memPtr = $memPtr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $memPtr;
 $2 = (_LZ4_isLittleEndian()|0);
 $3 = ($2|0)!=(0);
 $4 = $1;
 if ($3) {
  $5 = (_LZ4_read16($4)|0);
  $0 = $5;
  $17 = $0;
  STACKTOP = sp;return ($17|0);
 } else {
  $p = $4;
  $6 = $p;
  $7 = HEAP8[$6>>0]|0;
  $8 = $7&255;
  $9 = $8&65535;
  $10 = $p;
  $11 = ((($10)) + 1|0);
  $12 = HEAP8[$11>>0]|0;
  $13 = $12&255;
  $14 = $13 << 8;
  $15 = (($9) + ($14))|0;
  $16 = $15&65535;
  $0 = $16;
  $17 = $0;
  STACKTOP = sp;return ($17|0);
 }
 return (0)|0;
}
function _LZ4_copy4($dstPtr,$srcPtr) {
 $dstPtr = $dstPtr|0;
 $srcPtr = $srcPtr|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $dstPtr;
 $1 = $srcPtr;
 $2 = $0;
 $3 = $1;
 ;HEAP8[$2>>0]=HEAP8[$3>>0]|0;HEAP8[$2+1>>0]=HEAP8[$3+1>>0]|0;HEAP8[$2+2>>0]=HEAP8[$3+2>>0]|0;HEAP8[$2+3>>0]=HEAP8[$3+3>>0]|0;
 STACKTOP = sp;return;
}
function _LZ4_decompress_fast($source,$dest,$originalSize) {
 $source = $source|0;
 $dest = $dest|0;
 $originalSize = $originalSize|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0;
 var $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0;
 var $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0;
 var $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0;
 var $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0;
 var $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $checkOffset$i = 0, $copyFrom$i = 0;
 var $copySize$i = 0, $cpy$i = 0, $dec64$i = 0, $dictEnd$i = 0, $endOfMatch$i = 0, $iend$i = 0, $ip$i = 0, $length$i = 0, $lowLimit$i = 0, $match$i = 0, $oend$i = 0, $oexit$i = 0, $op$i = 0, $s$i = 0, $s1$i = 0, $safeDecode$i = 0, $token$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $12 = $source;
 $13 = $dest;
 $14 = $originalSize;
 $15 = $12;
 $16 = $13;
 $17 = $14;
 $18 = $13;
 $19 = ((($18)) + -65536|0);
 $1 = $15;
 $2 = $16;
 $3 = 0;
 $4 = $17;
 $5 = 0;
 $6 = 0;
 $7 = 0;
 $8 = 1;
 $9 = $19;
 $10 = 0;
 $11 = 65536;
 $20 = $1;
 $ip$i = $20;
 $21 = $ip$i;
 $22 = $3;
 $23 = (($21) + ($22)|0);
 $iend$i = $23;
 $24 = $2;
 $op$i = $24;
 $25 = $op$i;
 $26 = $4;
 $27 = (($25) + ($26)|0);
 $oend$i = $27;
 $28 = $op$i;
 $29 = $7;
 $30 = (($28) + ($29)|0);
 $oexit$i = $30;
 $31 = $9;
 $32 = $11;
 $33 = (0 - ($32))|0;
 $34 = (($31) + ($33)|0);
 $lowLimit$i = $34;
 $35 = $10;
 $36 = $11;
 $37 = (($35) + ($36)|0);
 $dictEnd$i = $37;
 $38 = $5;
 $39 = ($38|0)==(1);
 $40 = $39&1;
 $safeDecode$i = $40;
 $41 = $safeDecode$i;
 $42 = ($41|0)!=(0);
 $43 = $11;
 $44 = ($43>>>0)<(65536);
 $45 = $42 ? $44 : 0;
 $46 = $45&1;
 $checkOffset$i = $46;
 $47 = $6;
 $48 = ($47|0)!=(0);
 if ($48) {
  $49 = $oexit$i;
  $50 = $oend$i;
  $51 = ((($50)) + -12|0);
  $52 = ($49>>>0)>($51>>>0);
  if ($52) {
   $53 = $oend$i;
   $54 = ((($53)) + -12|0);
   $oexit$i = $54;
  }
 }
 $55 = $5;
 $56 = ($55|0)!=(0);
 if ($56) {
  $57 = $4;
  $58 = ($57|0)==(0);
  if ($58) {
   $59 = $3;
   $60 = ($59|0)==(1);
   if ($60) {
    $61 = $ip$i;
    $62 = HEAP8[$61>>0]|0;
    $63 = $62&255;
    $64 = ($63|0)==(0);
    $66 = $64;
   } else {
    $66 = 0;
   }
   $65 = $66 ? 0 : -1;
   $0 = $65;
   $417 = $0;
   STACKTOP = sp;return ($417|0);
  }
 }
 $67 = $5;
 $68 = ($67|0)!=(0);
 if (!($68)) {
  $69 = $4;
  $70 = ($69|0)==(0);
  if ($70) {
   $71 = $ip$i;
   $72 = HEAP8[$71>>0]|0;
   $73 = $72&255;
   $74 = ($73|0)==(0);
   $75 = $74 ? 1 : -1;
   $0 = $75;
   $417 = $0;
   STACKTOP = sp;return ($417|0);
  }
 }
 L17: while(1) {
  $76 = $ip$i;
  $77 = ((($76)) + 1|0);
  $ip$i = $77;
  $78 = HEAP8[$76>>0]|0;
  $79 = $78&255;
  $token$i = $79;
  $80 = $token$i;
  $81 = $80 >>> 4;
  $length$i = $81;
  $82 = ($81|0)==(15);
  if ($82) {
   while(1) {
    $83 = $ip$i;
    $84 = ((($83)) + 1|0);
    $ip$i = $84;
    $85 = HEAP8[$83>>0]|0;
    $86 = $85&255;
    $s$i = $86;
    $87 = $s$i;
    $88 = $length$i;
    $89 = (($88) + ($87))|0;
    $length$i = $89;
    $90 = $5;
    $91 = ($90|0)!=(0);
    if ($91) {
     $92 = $ip$i;
     $93 = $iend$i;
     $94 = ((($93)) + -15|0);
     $95 = ($92>>>0)<($94>>>0);
     $96 = $95&1;
     $98 = $96;
    } else {
     $98 = 1;
    }
    $97 = ($98|0)!=(0);
    $99 = $s$i;
    $100 = ($99|0)==(255);
    $101 = $97 ? $100 : 0;
    if (!($101)) {
     break;
    }
   }
   $102 = $safeDecode$i;
   $103 = ($102|0)!=(0);
   if ($103) {
    $104 = $op$i;
    $105 = $length$i;
    $106 = (($104) + ($105)|0);
    $107 = $106;
    $108 = $op$i;
    $109 = $108;
    $110 = ($107>>>0)<($109>>>0);
    if ($110) {
     break;
    }
   }
   $111 = $safeDecode$i;
   $112 = ($111|0)!=(0);
   if ($112) {
    $113 = $ip$i;
    $114 = $length$i;
    $115 = (($113) + ($114)|0);
    $116 = $115;
    $117 = $ip$i;
    $118 = $117;
    $119 = ($116>>>0)<($118>>>0);
    if ($119) {
     break;
    }
   }
  }
  $120 = $op$i;
  $121 = $length$i;
  $122 = (($120) + ($121)|0);
  $cpy$i = $122;
  $123 = $5;
  $124 = ($123|0)!=(0);
  if ($124) {
   $125 = $cpy$i;
   $126 = $6;
   $127 = ($126|0)!=(0);
   $128 = $oexit$i;
   $129 = $oend$i;
   $130 = ((($129)) + -12|0);
   $131 = $127 ? $128 : $130;
   $132 = ($125>>>0)>($131>>>0);
   if ($132) {
    label = 25;
    break;
   }
   $133 = $ip$i;
   $134 = $length$i;
   $135 = (($133) + ($134)|0);
   $136 = $iend$i;
   $137 = ((($136)) + -8|0);
   $138 = ($135>>>0)>($137>>>0);
   if ($138) {
    label = 25;
    break;
   }
  }
  $139 = $5;
  $140 = ($139|0)!=(0);
  if (!($140)) {
   $141 = $cpy$i;
   $142 = $oend$i;
   $143 = ((($142)) + -8|0);
   $144 = ($141>>>0)>($143>>>0);
   if ($144) {
    label = 25;
    break;
   }
  }
  $183 = $op$i;
  $184 = $ip$i;
  $185 = $cpy$i;
  _LZ4_wildCopy($183,$184,$185);
  $186 = $length$i;
  $187 = $ip$i;
  $188 = (($187) + ($186)|0);
  $ip$i = $188;
  $189 = $cpy$i;
  $op$i = $189;
  $190 = $cpy$i;
  $191 = $ip$i;
  $192 = (_LZ4_readLE16($191)|0);
  $193 = $192&65535;
  $194 = (0 - ($193))|0;
  $195 = (($190) + ($194)|0);
  $match$i = $195;
  $196 = $ip$i;
  $197 = ((($196)) + 2|0);
  $ip$i = $197;
  $198 = $checkOffset$i;
  $199 = ($198|0)!=(0);
  if ($199) {
   $200 = $match$i;
   $201 = $lowLimit$i;
   $202 = ($200>>>0)<($201>>>0);
   if ($202) {
    break;
   }
  }
  $203 = $token$i;
  $204 = $203 & 15;
  $length$i = $204;
  $205 = $length$i;
  $206 = ($205|0)==(15);
  if ($206) {
   while(1) {
    $207 = $5;
    $208 = ($207|0)!=(0);
    if ($208) {
     $209 = $ip$i;
     $210 = $iend$i;
     $211 = ((($210)) + -5|0);
     $212 = ($209>>>0)>($211>>>0);
     if ($212) {
      break L17;
     }
    }
    $213 = $ip$i;
    $214 = ((($213)) + 1|0);
    $ip$i = $214;
    $215 = HEAP8[$213>>0]|0;
    $216 = $215&255;
    $s1$i = $216;
    $217 = $s1$i;
    $218 = $length$i;
    $219 = (($218) + ($217))|0;
    $length$i = $219;
    $220 = $s1$i;
    $221 = ($220|0)==(255);
    if (!($221)) {
     break;
    }
   }
   $222 = $safeDecode$i;
   $223 = ($222|0)!=(0);
   if ($223) {
    $224 = $op$i;
    $225 = $length$i;
    $226 = (($224) + ($225)|0);
    $227 = $226;
    $228 = $op$i;
    $229 = $228;
    $230 = ($227>>>0)<($229>>>0);
    if ($230) {
     break;
    }
   }
  }
  $231 = $length$i;
  $232 = (($231) + 4)|0;
  $length$i = $232;
  $233 = $8;
  $234 = ($233|0)==(2);
  if ($234) {
   $235 = $match$i;
   $236 = $9;
   $237 = ($235>>>0)<($236>>>0);
   if ($237) {
    $238 = $op$i;
    $239 = $length$i;
    $240 = (($238) + ($239)|0);
    $241 = $oend$i;
    $242 = ((($241)) + -5|0);
    $243 = ($240>>>0)>($242>>>0);
    if ($243) {
     break;
    }
    $244 = $length$i;
    $245 = $9;
    $246 = $match$i;
    $247 = $245;
    $248 = $246;
    $249 = (($247) - ($248))|0;
    $250 = ($244>>>0)<=($249>>>0);
    if ($250) {
     $251 = $dictEnd$i;
     $252 = $9;
     $253 = $match$i;
     $254 = $252;
     $255 = $253;
     $256 = (($254) - ($255))|0;
     $257 = (0 - ($256))|0;
     $258 = (($251) + ($257)|0);
     $match$i = $258;
     $259 = $op$i;
     $260 = $match$i;
     $261 = $length$i;
     _memmove(($259|0),($260|0),($261|0))|0;
     $262 = $length$i;
     $263 = $op$i;
     $264 = (($263) + ($262)|0);
     $op$i = $264;
     continue;
    }
    $265 = $9;
    $266 = $match$i;
    $267 = $265;
    $268 = $266;
    $269 = (($267) - ($268))|0;
    $copySize$i = $269;
    $270 = $op$i;
    $271 = $dictEnd$i;
    $272 = $copySize$i;
    $273 = (0 - ($272))|0;
    $274 = (($271) + ($273)|0);
    $275 = $copySize$i;
    _memcpy(($270|0),($274|0),($275|0))|0;
    $276 = $copySize$i;
    $277 = $op$i;
    $278 = (($277) + ($276)|0);
    $op$i = $278;
    $279 = $length$i;
    $280 = $copySize$i;
    $281 = (($279) - ($280))|0;
    $copySize$i = $281;
    $282 = $copySize$i;
    $283 = $op$i;
    $284 = $9;
    $285 = $283;
    $286 = $284;
    $287 = (($285) - ($286))|0;
    $288 = ($282>>>0)>($287>>>0);
    $289 = $op$i;
    if (!($288)) {
     $301 = $9;
     $302 = $copySize$i;
     _memcpy(($289|0),($301|0),($302|0))|0;
     $303 = $copySize$i;
     $304 = $op$i;
     $305 = (($304) + ($303)|0);
     $op$i = $305;
     continue;
    }
    $290 = $copySize$i;
    $291 = (($289) + ($290)|0);
    $endOfMatch$i = $291;
    $292 = $9;
    $copyFrom$i = $292;
    while(1) {
     $293 = $op$i;
     $294 = $endOfMatch$i;
     $295 = ($293>>>0)<($294>>>0);
     if (!($295)) {
      continue L17;
     }
     $296 = $copyFrom$i;
     $297 = ((($296)) + 1|0);
     $copyFrom$i = $297;
     $298 = HEAP8[$296>>0]|0;
     $299 = $op$i;
     $300 = ((($299)) + 1|0);
     $op$i = $300;
     HEAP8[$299>>0] = $298;
    }
   }
  }
  $306 = $op$i;
  $307 = $length$i;
  $308 = (($306) + ($307)|0);
  $cpy$i = $308;
  $309 = $op$i;
  $310 = $match$i;
  $311 = $309;
  $312 = $310;
  $313 = (($311) - ($312))|0;
  $314 = ($313|0)<(8);
  $315 = $op$i;
  $316 = $match$i;
  if ($314) {
   $317 = $315;
   $318 = $316;
   $319 = (($317) - ($318))|0;
   $320 = (21832 + ($319<<2)|0);
   $321 = HEAP32[$320>>2]|0;
   $dec64$i = $321;
   $322 = $match$i;
   $323 = HEAP8[$322>>0]|0;
   $324 = $op$i;
   HEAP8[$324>>0] = $323;
   $325 = $match$i;
   $326 = ((($325)) + 1|0);
   $327 = HEAP8[$326>>0]|0;
   $328 = $op$i;
   $329 = ((($328)) + 1|0);
   HEAP8[$329>>0] = $327;
   $330 = $match$i;
   $331 = ((($330)) + 2|0);
   $332 = HEAP8[$331>>0]|0;
   $333 = $op$i;
   $334 = ((($333)) + 2|0);
   HEAP8[$334>>0] = $332;
   $335 = $match$i;
   $336 = ((($335)) + 3|0);
   $337 = HEAP8[$336>>0]|0;
   $338 = $op$i;
   $339 = ((($338)) + 3|0);
   HEAP8[$339>>0] = $337;
   $340 = $op$i;
   $341 = $match$i;
   $342 = $340;
   $343 = $341;
   $344 = (($342) - ($343))|0;
   $345 = (21864 + ($344<<2)|0);
   $346 = HEAP32[$345>>2]|0;
   $347 = $match$i;
   $348 = (($347) + ($346)|0);
   $match$i = $348;
   $349 = $op$i;
   $350 = ((($349)) + 4|0);
   $351 = $match$i;
   _LZ4_copy4($350,$351);
   $352 = $op$i;
   $353 = ((($352)) + 8|0);
   $op$i = $353;
   $354 = $dec64$i;
   $355 = $match$i;
   $356 = (0 - ($354))|0;
   $357 = (($355) + ($356)|0);
   $match$i = $357;
  } else {
   _LZ4_copy8($315,$316);
   $358 = $op$i;
   $359 = ((($358)) + 8|0);
   $op$i = $359;
   $360 = $match$i;
   $361 = ((($360)) + 8|0);
   $match$i = $361;
  }
  $362 = $cpy$i;
  $363 = $oend$i;
  $364 = ((($363)) + -12|0);
  $365 = ($362>>>0)>($364>>>0);
  L68: do {
   if ($365) {
    $366 = $cpy$i;
    $367 = $oend$i;
    $368 = ((($367)) + -5|0);
    $369 = ($366>>>0)>($368>>>0);
    if ($369) {
     break L17;
    }
    $370 = $op$i;
    $371 = $oend$i;
    $372 = ((($371)) + -8|0);
    $373 = ($370>>>0)<($372>>>0);
    if ($373) {
     $374 = $op$i;
     $375 = $match$i;
     $376 = $oend$i;
     $377 = ((($376)) + -8|0);
     _LZ4_wildCopy($374,$375,$377);
     $378 = $oend$i;
     $379 = ((($378)) + -8|0);
     $380 = $op$i;
     $381 = $379;
     $382 = $380;
     $383 = (($381) - ($382))|0;
     $384 = $match$i;
     $385 = (($384) + ($383)|0);
     $match$i = $385;
     $386 = $oend$i;
     $387 = ((($386)) + -8|0);
     $op$i = $387;
    }
    while(1) {
     $388 = $op$i;
     $389 = $cpy$i;
     $390 = ($388>>>0)<($389>>>0);
     if (!($390)) {
      break L68;
     }
     $391 = $match$i;
     $392 = ((($391)) + 1|0);
     $match$i = $392;
     $393 = HEAP8[$391>>0]|0;
     $394 = $op$i;
     $395 = ((($394)) + 1|0);
     $op$i = $395;
     HEAP8[$394>>0] = $393;
    }
   } else {
    $396 = $op$i;
    $397 = $match$i;
    $398 = $cpy$i;
    _LZ4_wildCopy($396,$397,$398);
   }
  } while(0);
  $399 = $cpy$i;
  $op$i = $399;
 }
 do {
  if ((label|0) == 25) {
   $145 = $6;
   $146 = ($145|0)!=(0);
   if ($146) {
    $147 = $cpy$i;
    $148 = $oend$i;
    $149 = ($147>>>0)>($148>>>0);
    if ($149) {
     break;
    }
    $150 = $5;
    $151 = ($150|0)!=(0);
    if ($151) {
     $152 = $ip$i;
     $153 = $length$i;
     $154 = (($152) + ($153)|0);
     $155 = $iend$i;
     $156 = ($154>>>0)>($155>>>0);
     if ($156) {
      break;
     }
    }
   } else {
    $157 = $5;
    $158 = ($157|0)!=(0);
    if (!($158)) {
     $159 = $cpy$i;
     $160 = $oend$i;
     $161 = ($159|0)!=($160|0);
     if ($161) {
      break;
     }
    }
    $162 = $5;
    $163 = ($162|0)!=(0);
    if ($163) {
     $164 = $ip$i;
     $165 = $length$i;
     $166 = (($164) + ($165)|0);
     $167 = $iend$i;
     $168 = ($166|0)!=($167|0);
     if ($168) {
      break;
     }
     $169 = $cpy$i;
     $170 = $oend$i;
     $171 = ($169>>>0)>($170>>>0);
     if ($171) {
      break;
     }
    }
   }
   $172 = $op$i;
   $173 = $ip$i;
   $174 = $length$i;
   _memcpy(($172|0),($173|0),($174|0))|0;
   $175 = $length$i;
   $176 = $ip$i;
   $177 = (($176) + ($175)|0);
   $ip$i = $177;
   $178 = $length$i;
   $179 = $op$i;
   $180 = (($179) + ($178)|0);
   $op$i = $180;
   $181 = $5;
   $182 = ($181|0)!=(0);
   if ($182) {
    $400 = $op$i;
    $401 = $2;
    $402 = $400;
    $403 = $401;
    $404 = (($402) - ($403))|0;
    $0 = $404;
    $417 = $0;
    STACKTOP = sp;return ($417|0);
   } else {
    $405 = $ip$i;
    $406 = $1;
    $407 = $405;
    $408 = $406;
    $409 = (($407) - ($408))|0;
    $0 = $409;
    $417 = $0;
    STACKTOP = sp;return ($417|0);
   }
  }
 } while(0);
 $410 = $ip$i;
 $411 = $1;
 $412 = $410;
 $413 = $411;
 $414 = (($412) - ($413))|0;
 $415 = (0 - ($414))|0;
 $416 = (($415) - 1)|0;
 $0 = $416;
 $417 = $0;
 STACKTOP = sp;return ($417|0);
}
function _LZ4_compress_HC_extStateHC($state,$src,$dst,$srcSize,$maxDstSize,$compressionLevel) {
 $state = $state|0;
 $src = $src|0;
 $dst = $dst|0;
 $srcSize = $srcSize|0;
 $maxDstSize = $maxDstSize|0;
 $compressionLevel = $compressionLevel|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $state;
 $2 = $src;
 $3 = $dst;
 $4 = $srcSize;
 $5 = $maxDstSize;
 $6 = $compressionLevel;
 $7 = $1;
 $8 = $7;
 $9 = $8 & 3;
 $10 = ($9|0)!=(0);
 if ($10) {
  $0 = 0;
  $25 = $0;
  STACKTOP = sp;return ($25|0);
 }
 $11 = $1;
 $12 = $2;
 _LZ4HC_init($11,$12);
 $13 = $5;
 $14 = $4;
 $15 = (_LZ4_compressBound($14)|0);
 $16 = ($13|0)<($15|0);
 $17 = $1;
 $18 = $2;
 $19 = $3;
 $20 = $4;
 $21 = $5;
 $22 = $6;
 if ($16) {
  $23 = (_LZ4HC_compress_generic($17,$18,$19,$20,$21,$22,1)|0);
  $0 = $23;
  $25 = $0;
  STACKTOP = sp;return ($25|0);
 } else {
  $24 = (_LZ4HC_compress_generic($17,$18,$19,$20,$21,$22,0)|0);
  $0 = $24;
  $25 = $0;
  STACKTOP = sp;return ($25|0);
 }
 return (0)|0;
}
function _LZ4HC_init($hc4,$start) {
 $hc4 = $hc4|0;
 $start = $start|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $hc4;
 $1 = $start;
 $2 = $0;
 _memset(($2|0),0,131072)|0;
 $3 = $0;
 $4 = ((($3)) + 131072|0);
 _memset(($4|0),-1,131072)|0;
 $5 = $0;
 $6 = ((($5)) + 262168|0);
 HEAP32[$6>>2] = 65536;
 $7 = $1;
 $8 = ((($7)) + -65536|0);
 $9 = $0;
 $10 = ((($9)) + 262148|0);
 HEAP32[$10>>2] = $8;
 $11 = $1;
 $12 = $0;
 $13 = ((($12)) + 262144|0);
 HEAP32[$13>>2] = $11;
 $14 = $1;
 $15 = ((($14)) + -65536|0);
 $16 = $0;
 $17 = ((($16)) + 262152|0);
 HEAP32[$17>>2] = $15;
 $18 = $0;
 $19 = ((($18)) + 262160|0);
 HEAP32[$19>>2] = 65536;
 $20 = $0;
 $21 = ((($20)) + 262164|0);
 HEAP32[$21>>2] = 65536;
 STACKTOP = sp;return;
}
function _LZ4HC_compress_generic($ctxvoid,$source,$dest,$inputSize,$maxOutputSize,$compressionLevel,$limit) {
 $ctxvoid = $ctxvoid|0;
 $source = $source|0;
 $dest = $dest|0;
 $inputSize = $inputSize|0;
 $maxOutputSize = $maxOutputSize|0;
 $compressionLevel = $compressionLevel|0;
 $limit = $limit|0;
 var $$ = 0, $$1 = 0, $$2 = 0, $$3 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0;
 var $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0;
 var $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0;
 var $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0;
 var $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0;
 var $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0;
 var $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0;
 var $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0;
 var $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0;
 var $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0;
 var $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0;
 var $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0;
 var $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0;
 var $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0;
 var $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0;
 var $1264 = 0, $1265 = 0, $1266 = 0, $1267 = 0, $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0;
 var $1282 = 0, $1283 = 0, $1284 = 0, $1285 = 0, $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0;
 var $130 = 0, $1300 = 0, $1301 = 0, $1302 = 0, $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0;
 var $1318 = 0, $1319 = 0, $132 = 0, $1320 = 0, $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0;
 var $1336 = 0, $1337 = 0, $1338 = 0, $1339 = 0, $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0;
 var $1354 = 0, $1355 = 0, $1356 = 0, $1357 = 0, $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0;
 var $1372 = 0, $1373 = 0, $1374 = 0, $1375 = 0, $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0;
 var $1390 = 0, $1391 = 0, $1392 = 0, $1393 = 0, $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0;
 var $1408 = 0, $1409 = 0, $141 = 0, $1410 = 0, $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0;
 var $1426 = 0, $1427 = 0, $1428 = 0, $1429 = 0, $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0;
 var $1444 = 0, $1445 = 0, $1446 = 0, $1447 = 0, $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0;
 var $1462 = 0, $1463 = 0, $1464 = 0, $1465 = 0, $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0;
 var $1480 = 0, $1481 = 0, $1482 = 0, $1483 = 0, $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0;
 var $1499 = 0, $15 = 0, $150 = 0, $1500 = 0, $1501 = 0, $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0, $1511 = 0, $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0;
 var $1516 = 0, $1517 = 0, $1518 = 0, $1519 = 0, $152 = 0, $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0, $153 = 0, $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0;
 var $1534 = 0, $1535 = 0, $1536 = 0, $1537 = 0, $1538 = 0, $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0, $1548 = 0, $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0;
 var $1552 = 0, $1553 = 0, $1554 = 0, $1555 = 0, $1556 = 0, $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $1565 = 0, $1566 = 0, $1567 = 0, $1568 = 0, $1569 = 0, $157 = 0;
 var $1570 = 0, $1571 = 0, $1572 = 0, $1573 = 0, $1574 = 0, $1575 = 0, $1576 = 0, $1577 = 0, $1578 = 0, $1579 = 0, $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0, $1583 = 0, $1584 = 0, $1585 = 0, $1586 = 0, $1587 = 0, $1588 = 0;
 var $1589 = 0, $159 = 0, $1590 = 0, $1591 = 0, $1592 = 0, $1593 = 0, $1594 = 0, $1595 = 0, $1596 = 0, $1597 = 0, $1598 = 0, $1599 = 0, $16 = 0, $160 = 0, $1600 = 0, $1601 = 0, $1602 = 0, $1603 = 0, $1604 = 0, $1605 = 0;
 var $1606 = 0, $1607 = 0, $1608 = 0, $1609 = 0, $161 = 0, $1610 = 0, $1611 = 0, $1612 = 0, $1613 = 0, $1614 = 0, $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0, $1619 = 0, $162 = 0, $1620 = 0, $1621 = 0, $1622 = 0, $1623 = 0;
 var $1624 = 0, $1625 = 0, $1626 = 0, $1627 = 0, $1628 = 0, $1629 = 0, $163 = 0, $1630 = 0, $1631 = 0, $1632 = 0, $1633 = 0, $1634 = 0, $1635 = 0, $1636 = 0, $1637 = 0, $1638 = 0, $1639 = 0, $164 = 0, $1640 = 0, $1641 = 0;
 var $1642 = 0, $1643 = 0, $1644 = 0, $1645 = 0, $1646 = 0, $1647 = 0, $1648 = 0, $1649 = 0, $165 = 0, $1650 = 0, $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0, $1655 = 0, $1656 = 0, $1657 = 0, $1658 = 0, $1659 = 0, $166 = 0;
 var $1660 = 0, $1661 = 0, $1662 = 0, $1663 = 0, $1664 = 0, $1665 = 0, $1666 = 0, $1667 = 0, $1668 = 0, $1669 = 0, $167 = 0, $1670 = 0, $1671 = 0, $1672 = 0, $1673 = 0, $1674 = 0, $1675 = 0, $1676 = 0, $1677 = 0, $1678 = 0;
 var $1679 = 0, $168 = 0, $1680 = 0, $1681 = 0, $1682 = 0, $1683 = 0, $1684 = 0, $1685 = 0, $1686 = 0, $1687 = 0, $1688 = 0, $1689 = 0, $169 = 0, $1690 = 0, $1691 = 0, $1692 = 0, $1693 = 0, $1694 = 0, $1695 = 0, $1696 = 0;
 var $1697 = 0, $1698 = 0, $1699 = 0, $17 = 0, $170 = 0, $1700 = 0, $1701 = 0, $1702 = 0, $1703 = 0, $1704 = 0, $1705 = 0, $1706 = 0, $1707 = 0, $1708 = 0, $1709 = 0, $171 = 0, $1710 = 0, $1711 = 0, $1712 = 0, $1713 = 0;
 var $1714 = 0, $1715 = 0, $1716 = 0, $1717 = 0, $1718 = 0, $1719 = 0, $172 = 0, $1720 = 0, $1721 = 0, $1722 = 0, $1723 = 0, $1724 = 0, $1725 = 0, $1726 = 0, $1727 = 0, $1728 = 0, $1729 = 0, $173 = 0, $1730 = 0, $1731 = 0;
 var $1732 = 0, $1733 = 0, $1734 = 0, $1735 = 0, $1736 = 0, $1737 = 0, $1738 = 0, $1739 = 0, $174 = 0, $1740 = 0, $1741 = 0, $1742 = 0, $1743 = 0, $1744 = 0, $1745 = 0, $1746 = 0, $1747 = 0, $1748 = 0, $1749 = 0, $175 = 0;
 var $1750 = 0, $1751 = 0, $1752 = 0, $1753 = 0, $1754 = 0, $1755 = 0, $1756 = 0, $1757 = 0, $1758 = 0, $1759 = 0, $176 = 0, $1760 = 0, $1761 = 0, $1762 = 0, $1763 = 0, $1764 = 0, $1765 = 0, $1766 = 0, $1767 = 0, $1768 = 0;
 var $1769 = 0, $177 = 0, $1770 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0;
 var $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0;
 var $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0;
 var $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0;
 var $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0;
 var $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0;
 var $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0;
 var $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0;
 var $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0;
 var $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0;
 var $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0;
 var $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0;
 var $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0;
 var $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0;
 var $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0;
 var $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0;
 var $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0;
 var $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0;
 var $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0;
 var $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0;
 var $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0;
 var $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0;
 var $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0;
 var $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0;
 var $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0;
 var $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0;
 var $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0;
 var $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0;
 var $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0;
 var $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0;
 var $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0;
 var $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0;
 var $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0;
 var $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0;
 var $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0;
 var $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0;
 var $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0;
 var $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0;
 var $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0;
 var $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0;
 var $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0;
 var $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0;
 var $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0;
 var $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $HashTable$i = 0, $HashTable$i$i = 0, $HashTable$i$i22 = 0, $HashTable$i$i4 = 0, $HashTable$i11 = 0;
 var $HashTable$i29 = 0, $anchor = 0, $back$i = 0, $back$i40 = 0, $back3$i = 0, $back3$i43 = 0, $base$i = 0, $base$i$i = 0, $base$i$i23 = 0, $base$i$i5 = 0, $base$i12 = 0, $base$i30 = 0, $chainTable$i = 0, $chainTable$i$i = 0, $chainTable$i$i21 = 0, $chainTable$i$i3 = 0, $chainTable$i10 = 0, $chainTable$i28 = 0, $correction = 0, $correction1 = 0;
 var $correction2 = 0, $ctx = 0, $delta$i = 0, $delta$i$i = 0, $delta$i$i27 = 0, $delta$i$i9 = 0, $delta$i37 = 0, $dictBase$i = 0, $dictBase$i15 = 0, $dictBase$i34 = 0, $dictLimit$i = 0, $dictLimit$i13 = 0, $dictLimit$i31 = 0, $h$i$i = 0, $h$i$i26 = 0, $h$i$i8 = 0, $idx$i$i = 0, $idx$i$i25 = 0, $idx$i$i7 = 0, $iend = 0;
 var $ip = 0, $lastRun = 0, $len$i = 0, $len$i49 = 0, $len$i53 = 0, $len$i57 = 0, $len$i61 = 0, $length$i = 0, $length$i47 = 0, $length$i51 = 0, $length$i55 = 0, $length$i59 = 0, $lowLimit$i = 0, $lowLimit$i14 = 0, $lowLimit$i33 = 0, $lowPrefixPtr$i = 0, $lowPrefixPtr$i32 = 0, $match$i = 0, $matchIndex$i = 0, $matchIndex$i16 = 0;
 var $matchIndex$i35 = 0, $matchPtr$i = 0, $matchPtr$i38 = 0, $matchPtr1$i = 0, $matchPtr1$i41 = 0, $matchlimit = 0, $maxNbAttempts = 0, $mflimit = 0, $ml = 0, $ml$i = 0, $ml0 = 0, $ml2 = 0, $ml3 = 0, $mlt$i = 0, $mlt$i18 = 0, $mlt$i39 = 0, $mlt1$i = 0, $mlt2$i = 0, $mlt2$i42 = 0, $nbAttempts$i = 0;
 var $nbAttempts$i17 = 0, $nbAttempts$i36 = 0, $new_ml = 0, $oend = 0, $op = 0, $ref = 0, $ref0 = 0, $ref2 = 0, $ref3 = 0, $start0 = 0, $start2 = 0, $start3 = 0, $target$i$i = 0, $target$i$i24 = 0, $target$i$i6 = 0, $token$i = 0, $token$i48 = 0, $token$i52 = 0, $token$i56 = 0, $token$i60 = 0;
 var $vLimit$i = 0, $vLimit$i19 = 0, $vLimit$i44 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 736|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ip = sp + 92|0;
 $anchor = sp + 88|0;
 $op = sp + 72|0;
 $ref = sp + 44|0;
 $start2 = sp + 40|0;
 $ref2 = sp + 36|0;
 $start3 = sp + 32|0;
 $ref3 = sp + 28|0;
 $68 = $ctxvoid;
 $69 = $source;
 $70 = $dest;
 $71 = $inputSize;
 $72 = $maxOutputSize;
 $73 = $compressionLevel;
 $74 = $limit;
 $75 = $68;
 $ctx = $75;
 $76 = $69;
 HEAP32[$ip>>2] = $76;
 $77 = HEAP32[$ip>>2]|0;
 HEAP32[$anchor>>2] = $77;
 $78 = HEAP32[$ip>>2]|0;
 $79 = $71;
 $80 = (($78) + ($79)|0);
 $iend = $80;
 $81 = $iend;
 $82 = ((($81)) + -12|0);
 $mflimit = $82;
 $83 = $iend;
 $84 = ((($83)) + -5|0);
 $matchlimit = $84;
 $85 = $70;
 HEAP32[$op>>2] = $85;
 $86 = HEAP32[$op>>2]|0;
 $87 = $72;
 $88 = (($86) + ($87)|0);
 $oend = $88;
 HEAP32[$ref>>2] = 0;
 HEAP32[$start2>>2] = 0;
 HEAP32[$ref2>>2] = 0;
 HEAP32[$start3>>2] = 0;
 HEAP32[$ref3>>2] = 0;
 $89 = $73;
 $90 = ($89|0)>(16);
 if ($90) {
  $73 = 16;
 }
 $91 = $73;
 $92 = ($91|0)<(1);
 if ($92) {
  $73 = 9;
 }
 $93 = $73;
 $94 = (($93) - 1)|0;
 $95 = 1 << $94;
 $maxNbAttempts = $95;
 $96 = $71;
 $97 = $ctx;
 $98 = ((($97)) + 262144|0);
 $99 = HEAP32[$98>>2]|0;
 $100 = (($99) + ($96)|0);
 HEAP32[$98>>2] = $100;
 $101 = HEAP32[$ip>>2]|0;
 $102 = ((($101)) + 1|0);
 HEAP32[$ip>>2] = $102;
 L7: while(1) {
  $103 = HEAP32[$ip>>2]|0;
  $104 = $mflimit;
  $105 = ($103>>>0)<($104>>>0);
  if (!($105)) {
   label = 250;
   break;
  }
  $106 = $ctx;
  $107 = HEAP32[$ip>>2]|0;
  $108 = $matchlimit;
  $109 = $maxNbAttempts;
  $62 = $106;
  $63 = $107;
  $64 = $108;
  $65 = $ref;
  $66 = $109;
  $110 = $62;
  $111 = ((($110)) + 131072|0);
  $chainTable$i = $111;
  $112 = $62;
  $HashTable$i = $112;
  $113 = $62;
  $114 = ((($113)) + 262148|0);
  $115 = HEAP32[$114>>2]|0;
  $base$i = $115;
  $116 = $62;
  $117 = ((($116)) + 262152|0);
  $118 = HEAP32[$117>>2]|0;
  $dictBase$i = $118;
  $119 = $62;
  $120 = ((($119)) + 262160|0);
  $121 = HEAP32[$120>>2]|0;
  $dictLimit$i = $121;
  $122 = $62;
  $123 = ((($122)) + 262164|0);
  $124 = HEAP32[$123>>2]|0;
  $125 = (($124) + 65536)|0;
  $126 = $63;
  $127 = $base$i;
  $128 = $126;
  $129 = $127;
  $130 = (($128) - ($129))|0;
  $131 = ($125>>>0)>($130>>>0);
  if ($131) {
   $132 = $62;
   $133 = ((($132)) + 262164|0);
   $134 = HEAP32[$133>>2]|0;
   $141 = $134;
  } else {
   $135 = $63;
   $136 = $base$i;
   $137 = $135;
   $138 = $136;
   $139 = (($137) - ($138))|0;
   $140 = (($139) - 65535)|0;
   $141 = $140;
  }
  $lowLimit$i = $141;
  $142 = $66;
  $nbAttempts$i = $142;
  $ml$i = 0;
  $143 = $62;
  $144 = $63;
  $60 = $143;
  $61 = $144;
  $145 = $60;
  $146 = ((($145)) + 131072|0);
  $chainTable$i$i = $146;
  $147 = $60;
  $HashTable$i$i = $147;
  $148 = $60;
  $149 = ((($148)) + 262148|0);
  $150 = HEAP32[$149>>2]|0;
  $base$i$i = $150;
  $151 = $61;
  $152 = $base$i$i;
  $153 = $151;
  $154 = $152;
  $155 = (($153) - ($154))|0;
  $target$i$i = $155;
  $156 = $60;
  $157 = ((($156)) + 262168|0);
  $158 = HEAP32[$157>>2]|0;
  $idx$i$i = $158;
  while(1) {
   $159 = $idx$i$i;
   $160 = $target$i$i;
   $161 = ($159>>>0)<($160>>>0);
   if (!($161)) {
    break;
   }
   $162 = $base$i$i;
   $163 = $idx$i$i;
   $164 = (($162) + ($163)|0);
   $165 = (_LZ4HC_hashPtr($164)|0);
   $h$i$i = $165;
   $166 = $idx$i$i;
   $167 = $h$i$i;
   $168 = $HashTable$i$i;
   $169 = (($168) + ($167<<2)|0);
   $170 = HEAP32[$169>>2]|0;
   $171 = (($166) - ($170))|0;
   $delta$i$i = $171;
   $172 = $delta$i$i;
   $173 = ($172>>>0)>(65535);
   $$ = $173 ? 65535 : $171;
   $delta$i$i = $$;
   $174 = $delta$i$i;
   $175 = $174&65535;
   $176 = $idx$i$i;
   $177 = $176&65535;
   $178 = $177&65535;
   $179 = $chainTable$i$i;
   $180 = (($179) + ($178<<1)|0);
   HEAP16[$180>>1] = $175;
   $181 = $idx$i$i;
   $182 = $h$i$i;
   $183 = $HashTable$i$i;
   $184 = (($183) + ($182<<2)|0);
   HEAP32[$184>>2] = $181;
   $185 = $idx$i$i;
   $186 = (($185) + 1)|0;
   $idx$i$i = $186;
  }
  $187 = $target$i$i;
  $188 = $60;
  $189 = ((($188)) + 262168|0);
  HEAP32[$189>>2] = $187;
  $190 = $63;
  $191 = (_LZ4HC_hashPtr($190)|0);
  $192 = $HashTable$i;
  $193 = (($192) + ($191<<2)|0);
  $194 = HEAP32[$193>>2]|0;
  $matchIndex$i = $194;
  while(1) {
   $195 = $matchIndex$i;
   $196 = $lowLimit$i;
   $197 = ($195>>>0)>=($196>>>0);
   $198 = $nbAttempts$i;
   $199 = ($198|0)!=(0);
   $200 = $197 ? $199 : 0;
   if (!($200)) {
    break;
   }
   $201 = $nbAttempts$i;
   $202 = (($201) + -1)|0;
   $nbAttempts$i = $202;
   $203 = $matchIndex$i;
   $204 = $dictLimit$i;
   $205 = ($203>>>0)>=($204>>>0);
   if ($205) {
    $206 = $base$i;
    $207 = $matchIndex$i;
    $208 = (($206) + ($207)|0);
    $match$i = $208;
    $209 = $match$i;
    $210 = $ml$i;
    $211 = (($209) + ($210)|0);
    $212 = HEAP8[$211>>0]|0;
    $213 = $212&255;
    $214 = $63;
    $215 = $ml$i;
    $216 = (($214) + ($215)|0);
    $217 = HEAP8[$216>>0]|0;
    $218 = $217&255;
    $219 = ($213|0)==($218|0);
    if ($219) {
     $220 = $match$i;
     $221 = (_LZ4_read32_2985($220)|0);
     $222 = $63;
     $223 = (_LZ4_read32_2985($222)|0);
     $224 = ($221|0)==($223|0);
     if ($224) {
      $225 = $63;
      $226 = ((($225)) + 4|0);
      $227 = $match$i;
      $228 = ((($227)) + 4|0);
      $229 = $64;
      $230 = (_LZ4_count_2986($226,$228,$229)|0);
      $231 = (($230) + 4)|0;
      $mlt$i = $231;
      $232 = $mlt$i;
      $233 = $ml$i;
      $234 = ($232>>>0)>($233>>>0);
      if ($234) {
       $235 = $mlt$i;
       $ml$i = $235;
       $236 = $match$i;
       $237 = $65;
       HEAP32[$237>>2] = $236;
      }
     }
    }
   } else {
    $238 = $dictBase$i;
    $239 = $matchIndex$i;
    $240 = (($238) + ($239)|0);
    $match$i = $240;
    $241 = $match$i;
    $242 = (_LZ4_read32_2985($241)|0);
    $243 = $63;
    $244 = (_LZ4_read32_2985($243)|0);
    $245 = ($242|0)==($244|0);
    if ($245) {
     $246 = $63;
     $247 = $dictLimit$i;
     $248 = $matchIndex$i;
     $249 = (($247) - ($248))|0;
     $250 = (($246) + ($249)|0);
     $vLimit$i = $250;
     $251 = $vLimit$i;
     $252 = $64;
     $253 = ($251>>>0)>($252>>>0);
     if ($253) {
      $254 = $64;
      $vLimit$i = $254;
     }
     $255 = $63;
     $256 = ((($255)) + 4|0);
     $257 = $match$i;
     $258 = ((($257)) + 4|0);
     $259 = $vLimit$i;
     $260 = (_LZ4_count_2986($256,$258,$259)|0);
     $261 = (($260) + 4)|0;
     $mlt1$i = $261;
     $262 = $63;
     $263 = $mlt1$i;
     $264 = (($262) + ($263)|0);
     $265 = $vLimit$i;
     $266 = ($264|0)==($265|0);
     if ($266) {
      $267 = $vLimit$i;
      $268 = $64;
      $269 = ($267>>>0)<($268>>>0);
      if ($269) {
       $270 = $63;
       $271 = $mlt1$i;
       $272 = (($270) + ($271)|0);
       $273 = $base$i;
       $274 = $dictLimit$i;
       $275 = (($273) + ($274)|0);
       $276 = $64;
       $277 = (_LZ4_count_2986($272,$275,$276)|0);
       $278 = $mlt1$i;
       $279 = (($278) + ($277))|0;
       $mlt1$i = $279;
      }
     }
     $280 = $mlt1$i;
     $281 = $ml$i;
     $282 = ($280>>>0)>($281>>>0);
     if ($282) {
      $283 = $mlt1$i;
      $ml$i = $283;
      $284 = $base$i;
      $285 = $matchIndex$i;
      $286 = (($284) + ($285)|0);
      $287 = $65;
      HEAP32[$287>>2] = $286;
     }
    }
   }
   $288 = $matchIndex$i;
   $289 = $288&65535;
   $290 = $289&65535;
   $291 = $chainTable$i;
   $292 = (($291) + ($290<<1)|0);
   $293 = HEAP16[$292>>1]|0;
   $294 = $293&65535;
   $295 = $matchIndex$i;
   $296 = (($295) - ($294))|0;
   $matchIndex$i = $296;
  }
  $297 = $ml$i;
  $ml = $297;
  $298 = $ml;
  $299 = ($298|0)!=(0);
  $300 = HEAP32[$ip>>2]|0;
  if (!($299)) {
   $301 = ((($300)) + 1|0);
   HEAP32[$ip>>2] = $301;
   continue;
  }
  $start0 = $300;
  $302 = HEAP32[$ref>>2]|0;
  $ref0 = $302;
  $303 = $ml;
  $ml0 = $303;
  L41: while(1) {
   $304 = HEAP32[$ip>>2]|0;
   $305 = $ml;
   $306 = (($304) + ($305)|0);
   $307 = $mflimit;
   $308 = ($306>>>0)<($307>>>0);
   if ($308) {
    $309 = $ctx;
    $310 = HEAP32[$ip>>2]|0;
    $311 = $ml;
    $312 = (($310) + ($311)|0);
    $313 = ((($312)) + -2|0);
    $314 = HEAP32[$ip>>2]|0;
    $315 = ((($314)) + 1|0);
    $316 = $matchlimit;
    $317 = $ml;
    $318 = $maxNbAttempts;
    $52 = $309;
    $53 = $313;
    $54 = $315;
    $55 = $316;
    $56 = $317;
    $57 = $ref2;
    $58 = $start2;
    $59 = $318;
    $319 = $52;
    $320 = ((($319)) + 131072|0);
    $chainTable$i10 = $320;
    $321 = $52;
    $HashTable$i11 = $321;
    $322 = $52;
    $323 = ((($322)) + 262148|0);
    $324 = HEAP32[$323>>2]|0;
    $base$i12 = $324;
    $325 = $52;
    $326 = ((($325)) + 262160|0);
    $327 = HEAP32[$326>>2]|0;
    $dictLimit$i13 = $327;
    $328 = $base$i12;
    $329 = $dictLimit$i13;
    $330 = (($328) + ($329)|0);
    $lowPrefixPtr$i = $330;
    $331 = $52;
    $332 = ((($331)) + 262164|0);
    $333 = HEAP32[$332>>2]|0;
    $334 = (($333) + 65536)|0;
    $335 = $53;
    $336 = $base$i12;
    $337 = $335;
    $338 = $336;
    $339 = (($337) - ($338))|0;
    $340 = ($334>>>0)>($339>>>0);
    if ($340) {
     $341 = $52;
     $342 = ((($341)) + 262164|0);
     $343 = HEAP32[$342>>2]|0;
     $350 = $343;
    } else {
     $344 = $53;
     $345 = $base$i12;
     $346 = $344;
     $347 = $345;
     $348 = (($346) - ($347))|0;
     $349 = (($348) - 65535)|0;
     $350 = $349;
    }
    $lowLimit$i14 = $350;
    $351 = $52;
    $352 = ((($351)) + 262152|0);
    $353 = HEAP32[$352>>2]|0;
    $dictBase$i15 = $353;
    $354 = $59;
    $nbAttempts$i17 = $354;
    $355 = $53;
    $356 = $54;
    $357 = $355;
    $358 = $356;
    $359 = (($357) - ($358))|0;
    $delta$i = $359;
    $360 = $52;
    $361 = $53;
    $50 = $360;
    $51 = $361;
    $362 = $50;
    $363 = ((($362)) + 131072|0);
    $chainTable$i$i3 = $363;
    $364 = $50;
    $HashTable$i$i4 = $364;
    $365 = $50;
    $366 = ((($365)) + 262148|0);
    $367 = HEAP32[$366>>2]|0;
    $base$i$i5 = $367;
    $368 = $51;
    $369 = $base$i$i5;
    $370 = $368;
    $371 = $369;
    $372 = (($370) - ($371))|0;
    $target$i$i6 = $372;
    $373 = $50;
    $374 = ((($373)) + 262168|0);
    $375 = HEAP32[$374>>2]|0;
    $idx$i$i7 = $375;
    while(1) {
     $376 = $idx$i$i7;
     $377 = $target$i$i6;
     $378 = ($376>>>0)<($377>>>0);
     if (!($378)) {
      break;
     }
     $379 = $base$i$i5;
     $380 = $idx$i$i7;
     $381 = (($379) + ($380)|0);
     $382 = (_LZ4HC_hashPtr($381)|0);
     $h$i$i8 = $382;
     $383 = $idx$i$i7;
     $384 = $h$i$i8;
     $385 = $HashTable$i$i4;
     $386 = (($385) + ($384<<2)|0);
     $387 = HEAP32[$386>>2]|0;
     $388 = (($383) - ($387))|0;
     $delta$i$i9 = $388;
     $389 = $delta$i$i9;
     $390 = ($389>>>0)>(65535);
     $$1 = $390 ? 65535 : $388;
     $delta$i$i9 = $$1;
     $391 = $delta$i$i9;
     $392 = $391&65535;
     $393 = $idx$i$i7;
     $394 = $393&65535;
     $395 = $394&65535;
     $396 = $chainTable$i$i3;
     $397 = (($396) + ($395<<1)|0);
     HEAP16[$397>>1] = $392;
     $398 = $idx$i$i7;
     $399 = $h$i$i8;
     $400 = $HashTable$i$i4;
     $401 = (($400) + ($399<<2)|0);
     HEAP32[$401>>2] = $398;
     $402 = $idx$i$i7;
     $403 = (($402) + 1)|0;
     $idx$i$i7 = $403;
    }
    $404 = $target$i$i6;
    $405 = $50;
    $406 = ((($405)) + 262168|0);
    HEAP32[$406>>2] = $404;
    $407 = $53;
    $408 = (_LZ4HC_hashPtr($407)|0);
    $409 = $HashTable$i11;
    $410 = (($409) + ($408<<2)|0);
    $411 = HEAP32[$410>>2]|0;
    $matchIndex$i16 = $411;
    while(1) {
     $412 = $matchIndex$i16;
     $413 = $lowLimit$i14;
     $414 = ($412>>>0)>=($413>>>0);
     $415 = $nbAttempts$i17;
     $416 = ($415|0)!=(0);
     $417 = $414 ? $416 : 0;
     if (!($417)) {
      break;
     }
     $418 = $nbAttempts$i17;
     $419 = (($418) + -1)|0;
     $nbAttempts$i17 = $419;
     $420 = $matchIndex$i16;
     $421 = $dictLimit$i13;
     $422 = ($420>>>0)>=($421>>>0);
     if ($422) {
      $423 = $base$i12;
      $424 = $matchIndex$i16;
      $425 = (($423) + ($424)|0);
      $matchPtr$i = $425;
      $426 = $54;
      $427 = $56;
      $428 = (($426) + ($427)|0);
      $429 = HEAP8[$428>>0]|0;
      $430 = $429&255;
      $431 = $matchPtr$i;
      $432 = $delta$i;
      $433 = (0 - ($432))|0;
      $434 = (($431) + ($433)|0);
      $435 = $56;
      $436 = (($434) + ($435)|0);
      $437 = HEAP8[$436>>0]|0;
      $438 = $437&255;
      $439 = ($430|0)==($438|0);
      if ($439) {
       $440 = $matchPtr$i;
       $441 = (_LZ4_read32_2985($440)|0);
       $442 = $53;
       $443 = (_LZ4_read32_2985($442)|0);
       $444 = ($441|0)==($443|0);
       if ($444) {
        $445 = $53;
        $446 = ((($445)) + 4|0);
        $447 = $matchPtr$i;
        $448 = ((($447)) + 4|0);
        $449 = $55;
        $450 = (_LZ4_count_2986($446,$448,$449)|0);
        $451 = (4 + ($450))|0;
        $mlt$i18 = $451;
        $back$i = 0;
        while(1) {
         $452 = $53;
         $453 = $back$i;
         $454 = (($452) + ($453)|0);
         $455 = $54;
         $456 = ($454>>>0)>($455>>>0);
         if ($456) {
          $457 = $matchPtr$i;
          $458 = $back$i;
          $459 = (($457) + ($458)|0);
          $460 = $lowPrefixPtr$i;
          $461 = ($459>>>0)>($460>>>0);
          if ($461) {
           $462 = $back$i;
           $463 = (($462) - 1)|0;
           $464 = $53;
           $465 = (($464) + ($463)|0);
           $466 = HEAP8[$465>>0]|0;
           $467 = $466&255;
           $468 = $back$i;
           $469 = (($468) - 1)|0;
           $470 = $matchPtr$i;
           $471 = (($470) + ($469)|0);
           $472 = HEAP8[$471>>0]|0;
           $473 = $472&255;
           $474 = ($467|0)==($473|0);
           $1767 = $474;
          } else {
           $1767 = 0;
          }
         } else {
          $1767 = 0;
         }
         $475 = $back$i;
         if (!($1767)) {
          break;
         }
         $476 = (($475) + -1)|0;
         $back$i = $476;
        }
        $477 = $mlt$i18;
        $478 = (($477) - ($475))|0;
        $mlt$i18 = $478;
        $479 = $mlt$i18;
        $480 = $56;
        $481 = ($479|0)>($480|0);
        if ($481) {
         $482 = $mlt$i18;
         $56 = $482;
         $483 = $matchPtr$i;
         $484 = $back$i;
         $485 = (($483) + ($484)|0);
         $486 = $57;
         HEAP32[$486>>2] = $485;
         $487 = $53;
         $488 = $back$i;
         $489 = (($487) + ($488)|0);
         $490 = $58;
         HEAP32[$490>>2] = $489;
        }
       }
      }
     } else {
      $491 = $dictBase$i15;
      $492 = $matchIndex$i16;
      $493 = (($491) + ($492)|0);
      $matchPtr1$i = $493;
      $494 = $matchPtr1$i;
      $495 = (_LZ4_read32_2985($494)|0);
      $496 = $53;
      $497 = (_LZ4_read32_2985($496)|0);
      $498 = ($495|0)==($497|0);
      if ($498) {
       $back3$i = 0;
       $499 = $53;
       $500 = $dictLimit$i13;
       $501 = $matchIndex$i16;
       $502 = (($500) - ($501))|0;
       $503 = (($499) + ($502)|0);
       $vLimit$i19 = $503;
       $504 = $vLimit$i19;
       $505 = $55;
       $506 = ($504>>>0)>($505>>>0);
       if ($506) {
        $507 = $55;
        $vLimit$i19 = $507;
       }
       $508 = $53;
       $509 = ((($508)) + 4|0);
       $510 = $matchPtr1$i;
       $511 = ((($510)) + 4|0);
       $512 = $vLimit$i19;
       $513 = (_LZ4_count_2986($509,$511,$512)|0);
       $514 = (($513) + 4)|0;
       $mlt2$i = $514;
       $515 = $53;
       $516 = $mlt2$i;
       $517 = (($515) + ($516)|0);
       $518 = $vLimit$i19;
       $519 = ($517|0)==($518|0);
       if ($519) {
        $520 = $vLimit$i19;
        $521 = $55;
        $522 = ($520>>>0)<($521>>>0);
        if ($522) {
         $523 = $53;
         $524 = $mlt2$i;
         $525 = (($523) + ($524)|0);
         $526 = $base$i12;
         $527 = $dictLimit$i13;
         $528 = (($526) + ($527)|0);
         $529 = $55;
         $530 = (_LZ4_count_2986($525,$528,$529)|0);
         $531 = $mlt2$i;
         $532 = (($531) + ($530))|0;
         $mlt2$i = $532;
        }
       }
       while(1) {
        $533 = $53;
        $534 = $back3$i;
        $535 = (($533) + ($534)|0);
        $536 = $54;
        $537 = ($535>>>0)>($536>>>0);
        do {
         if ($537) {
          $538 = $matchIndex$i16;
          $539 = $back3$i;
          $540 = (($538) + ($539))|0;
          $541 = $lowLimit$i14;
          $542 = ($540>>>0)>($541>>>0);
          if (!($542)) {
           $1768 = 0;
           break;
          }
          $543 = $back3$i;
          $544 = (($543) - 1)|0;
          $545 = $53;
          $546 = (($545) + ($544)|0);
          $547 = HEAP8[$546>>0]|0;
          $548 = $547&255;
          $549 = $back3$i;
          $550 = (($549) - 1)|0;
          $551 = $matchPtr1$i;
          $552 = (($551) + ($550)|0);
          $553 = HEAP8[$552>>0]|0;
          $554 = $553&255;
          $555 = ($548|0)==($554|0);
          $1768 = $555;
         } else {
          $1768 = 0;
         }
        } while(0);
        $556 = $back3$i;
        if (!($1768)) {
         break;
        }
        $557 = (($556) + -1)|0;
        $back3$i = $557;
       }
       $558 = $mlt2$i;
       $559 = (($558) - ($556))|0;
       $mlt2$i = $559;
       $560 = $mlt2$i;
       $561 = $56;
       $562 = ($560|0)>($561|0);
       if ($562) {
        $563 = $mlt2$i;
        $56 = $563;
        $564 = $base$i12;
        $565 = $matchIndex$i16;
        $566 = (($564) + ($565)|0);
        $567 = $back3$i;
        $568 = (($566) + ($567)|0);
        $569 = $57;
        HEAP32[$569>>2] = $568;
        $570 = $53;
        $571 = $back3$i;
        $572 = (($570) + ($571)|0);
        $573 = $58;
        HEAP32[$573>>2] = $572;
       }
      }
     }
     $574 = $matchIndex$i16;
     $575 = $574&65535;
     $576 = $575&65535;
     $577 = $chainTable$i10;
     $578 = (($577) + ($576<<1)|0);
     $579 = HEAP16[$578>>1]|0;
     $580 = $579&65535;
     $581 = $matchIndex$i16;
     $582 = (($581) - ($580))|0;
     $matchIndex$i16 = $582;
    }
    $583 = $56;
    $ml2 = $583;
   } else {
    $584 = $ml;
    $ml2 = $584;
   }
   $585 = $ml2;
   $586 = $ml;
   $587 = ($585|0)==($586|0);
   if ($587) {
    label = 69;
    break;
   }
   $719 = $start0;
   $720 = HEAP32[$ip>>2]|0;
   $721 = ($719>>>0)<($720>>>0);
   if ($721) {
    $722 = HEAP32[$start2>>2]|0;
    $723 = HEAP32[$ip>>2]|0;
    $724 = $ml0;
    $725 = (($723) + ($724)|0);
    $726 = ($722>>>0)<($725>>>0);
    if ($726) {
     $727 = $start0;
     HEAP32[$ip>>2] = $727;
     $728 = $ref0;
     HEAP32[$ref>>2] = $728;
     $729 = $ml0;
     $ml = $729;
    }
   }
   $730 = HEAP32[$start2>>2]|0;
   $731 = HEAP32[$ip>>2]|0;
   $732 = $730;
   $733 = $731;
   $734 = (($732) - ($733))|0;
   $735 = ($734|0)<(3);
   if ($735) {
    $736 = $ml2;
    $ml = $736;
    $737 = HEAP32[$start2>>2]|0;
    HEAP32[$ip>>2] = $737;
    $738 = HEAP32[$ref2>>2]|0;
    HEAP32[$ref>>2] = $738;
    continue;
   }
   while(1) {
    $739 = HEAP32[$start2>>2]|0;
    $740 = HEAP32[$ip>>2]|0;
    $741 = $739;
    $742 = $740;
    $743 = (($741) - ($742))|0;
    $744 = ($743|0)<(18);
    if ($744) {
     $745 = $ml;
     $new_ml = $745;
     $746 = $new_ml;
     $747 = ($746|0)>(18);
     $$2 = $747 ? 18 : $745;
     $new_ml = $$2;
     $748 = HEAP32[$ip>>2]|0;
     $749 = $new_ml;
     $750 = (($748) + ($749)|0);
     $751 = HEAP32[$start2>>2]|0;
     $752 = $ml2;
     $753 = (($751) + ($752)|0);
     $754 = ((($753)) + -4|0);
     $755 = ($750>>>0)>($754>>>0);
     if ($755) {
      $756 = HEAP32[$start2>>2]|0;
      $757 = HEAP32[$ip>>2]|0;
      $758 = $756;
      $759 = $757;
      $760 = (($758) - ($759))|0;
      $761 = $ml2;
      $762 = (($760) + ($761))|0;
      $763 = (($762) - 4)|0;
      $new_ml = $763;
     }
     $764 = $new_ml;
     $765 = HEAP32[$start2>>2]|0;
     $766 = HEAP32[$ip>>2]|0;
     $767 = $765;
     $768 = $766;
     $769 = (($767) - ($768))|0;
     $770 = (($764) - ($769))|0;
     $correction = $770;
     $771 = $correction;
     $772 = ($771|0)>(0);
     if ($772) {
      $773 = $correction;
      $774 = HEAP32[$start2>>2]|0;
      $775 = (($774) + ($773)|0);
      HEAP32[$start2>>2] = $775;
      $776 = $correction;
      $777 = HEAP32[$ref2>>2]|0;
      $778 = (($777) + ($776)|0);
      HEAP32[$ref2>>2] = $778;
      $779 = $correction;
      $780 = $ml2;
      $781 = (($780) - ($779))|0;
      $ml2 = $781;
     }
    }
    $782 = HEAP32[$start2>>2]|0;
    $783 = $ml2;
    $784 = (($782) + ($783)|0);
    $785 = $mflimit;
    $786 = ($784>>>0)<($785>>>0);
    if ($786) {
     $787 = $ctx;
     $788 = HEAP32[$start2>>2]|0;
     $789 = $ml2;
     $790 = (($788) + ($789)|0);
     $791 = ((($790)) + -3|0);
     $792 = HEAP32[$start2>>2]|0;
     $793 = $matchlimit;
     $794 = $ml2;
     $795 = $maxNbAttempts;
     $34 = $787;
     $35 = $791;
     $36 = $792;
     $37 = $793;
     $38 = $794;
     $39 = $ref3;
     $40 = $start3;
     $41 = $795;
     $796 = $34;
     $797 = ((($796)) + 131072|0);
     $chainTable$i28 = $797;
     $798 = $34;
     $HashTable$i29 = $798;
     $799 = $34;
     $800 = ((($799)) + 262148|0);
     $801 = HEAP32[$800>>2]|0;
     $base$i30 = $801;
     $802 = $34;
     $803 = ((($802)) + 262160|0);
     $804 = HEAP32[$803>>2]|0;
     $dictLimit$i31 = $804;
     $805 = $base$i30;
     $806 = $dictLimit$i31;
     $807 = (($805) + ($806)|0);
     $lowPrefixPtr$i32 = $807;
     $808 = $34;
     $809 = ((($808)) + 262164|0);
     $810 = HEAP32[$809>>2]|0;
     $811 = (($810) + 65536)|0;
     $812 = $35;
     $813 = $base$i30;
     $814 = $812;
     $815 = $813;
     $816 = (($814) - ($815))|0;
     $817 = ($811>>>0)>($816>>>0);
     if ($817) {
      $818 = $34;
      $819 = ((($818)) + 262164|0);
      $820 = HEAP32[$819>>2]|0;
      $827 = $820;
     } else {
      $821 = $35;
      $822 = $base$i30;
      $823 = $821;
      $824 = $822;
      $825 = (($823) - ($824))|0;
      $826 = (($825) - 65535)|0;
      $827 = $826;
     }
     $lowLimit$i33 = $827;
     $828 = $34;
     $829 = ((($828)) + 262152|0);
     $830 = HEAP32[$829>>2]|0;
     $dictBase$i34 = $830;
     $831 = $41;
     $nbAttempts$i36 = $831;
     $832 = $35;
     $833 = $36;
     $834 = $832;
     $835 = $833;
     $836 = (($834) - ($835))|0;
     $delta$i37 = $836;
     $837 = $34;
     $838 = $35;
     $32 = $837;
     $33 = $838;
     $839 = $32;
     $840 = ((($839)) + 131072|0);
     $chainTable$i$i21 = $840;
     $841 = $32;
     $HashTable$i$i22 = $841;
     $842 = $32;
     $843 = ((($842)) + 262148|0);
     $844 = HEAP32[$843>>2]|0;
     $base$i$i23 = $844;
     $845 = $33;
     $846 = $base$i$i23;
     $847 = $845;
     $848 = $846;
     $849 = (($847) - ($848))|0;
     $target$i$i24 = $849;
     $850 = $32;
     $851 = ((($850)) + 262168|0);
     $852 = HEAP32[$851>>2]|0;
     $idx$i$i25 = $852;
     while(1) {
      $853 = $idx$i$i25;
      $854 = $target$i$i24;
      $855 = ($853>>>0)<($854>>>0);
      if (!($855)) {
       break;
      }
      $856 = $base$i$i23;
      $857 = $idx$i$i25;
      $858 = (($856) + ($857)|0);
      $859 = (_LZ4HC_hashPtr($858)|0);
      $h$i$i26 = $859;
      $860 = $idx$i$i25;
      $861 = $h$i$i26;
      $862 = $HashTable$i$i22;
      $863 = (($862) + ($861<<2)|0);
      $864 = HEAP32[$863>>2]|0;
      $865 = (($860) - ($864))|0;
      $delta$i$i27 = $865;
      $866 = $delta$i$i27;
      $867 = ($866>>>0)>(65535);
      $$3 = $867 ? 65535 : $865;
      $delta$i$i27 = $$3;
      $868 = $delta$i$i27;
      $869 = $868&65535;
      $870 = $idx$i$i25;
      $871 = $870&65535;
      $872 = $871&65535;
      $873 = $chainTable$i$i21;
      $874 = (($873) + ($872<<1)|0);
      HEAP16[$874>>1] = $869;
      $875 = $idx$i$i25;
      $876 = $h$i$i26;
      $877 = $HashTable$i$i22;
      $878 = (($877) + ($876<<2)|0);
      HEAP32[$878>>2] = $875;
      $879 = $idx$i$i25;
      $880 = (($879) + 1)|0;
      $idx$i$i25 = $880;
     }
     $881 = $target$i$i24;
     $882 = $32;
     $883 = ((($882)) + 262168|0);
     HEAP32[$883>>2] = $881;
     $884 = $35;
     $885 = (_LZ4HC_hashPtr($884)|0);
     $886 = $HashTable$i29;
     $887 = (($886) + ($885<<2)|0);
     $888 = HEAP32[$887>>2]|0;
     $matchIndex$i35 = $888;
     while(1) {
      $889 = $matchIndex$i35;
      $890 = $lowLimit$i33;
      $891 = ($889>>>0)>=($890>>>0);
      $892 = $nbAttempts$i36;
      $893 = ($892|0)!=(0);
      $894 = $891 ? $893 : 0;
      if (!($894)) {
       break;
      }
      $895 = $nbAttempts$i36;
      $896 = (($895) + -1)|0;
      $nbAttempts$i36 = $896;
      $897 = $matchIndex$i35;
      $898 = $dictLimit$i31;
      $899 = ($897>>>0)>=($898>>>0);
      do {
       if ($899) {
        $900 = $base$i30;
        $901 = $matchIndex$i35;
        $902 = (($900) + ($901)|0);
        $matchPtr$i38 = $902;
        $903 = $36;
        $904 = $38;
        $905 = (($903) + ($904)|0);
        $906 = HEAP8[$905>>0]|0;
        $907 = $906&255;
        $908 = $matchPtr$i38;
        $909 = $delta$i37;
        $910 = (0 - ($909))|0;
        $911 = (($908) + ($910)|0);
        $912 = $38;
        $913 = (($911) + ($912)|0);
        $914 = HEAP8[$913>>0]|0;
        $915 = $914&255;
        $916 = ($907|0)==($915|0);
        if (!($916)) {
         break;
        }
        $917 = $matchPtr$i38;
        $918 = (_LZ4_read32_2985($917)|0);
        $919 = $35;
        $920 = (_LZ4_read32_2985($919)|0);
        $921 = ($918|0)==($920|0);
        if (!($921)) {
         break;
        }
        $922 = $35;
        $923 = ((($922)) + 4|0);
        $924 = $matchPtr$i38;
        $925 = ((($924)) + 4|0);
        $926 = $37;
        $927 = (_LZ4_count_2986($923,$925,$926)|0);
        $928 = (4 + ($927))|0;
        $mlt$i39 = $928;
        $back$i40 = 0;
        while(1) {
         $929 = $35;
         $930 = $back$i40;
         $931 = (($929) + ($930)|0);
         $932 = $36;
         $933 = ($931>>>0)>($932>>>0);
         do {
          if ($933) {
           $934 = $matchPtr$i38;
           $935 = $back$i40;
           $936 = (($934) + ($935)|0);
           $937 = $lowPrefixPtr$i32;
           $938 = ($936>>>0)>($937>>>0);
           if (!($938)) {
            $1769 = 0;
            break;
           }
           $939 = $back$i40;
           $940 = (($939) - 1)|0;
           $941 = $35;
           $942 = (($941) + ($940)|0);
           $943 = HEAP8[$942>>0]|0;
           $944 = $943&255;
           $945 = $back$i40;
           $946 = (($945) - 1)|0;
           $947 = $matchPtr$i38;
           $948 = (($947) + ($946)|0);
           $949 = HEAP8[$948>>0]|0;
           $950 = $949&255;
           $951 = ($944|0)==($950|0);
           $1769 = $951;
          } else {
           $1769 = 0;
          }
         } while(0);
         $952 = $back$i40;
         if (!($1769)) {
          break;
         }
         $953 = (($952) + -1)|0;
         $back$i40 = $953;
        }
        $954 = $mlt$i39;
        $955 = (($954) - ($952))|0;
        $mlt$i39 = $955;
        $956 = $mlt$i39;
        $957 = $38;
        $958 = ($956|0)>($957|0);
        if (!($958)) {
         break;
        }
        $959 = $mlt$i39;
        $38 = $959;
        $960 = $matchPtr$i38;
        $961 = $back$i40;
        $962 = (($960) + ($961)|0);
        $963 = $39;
        HEAP32[$963>>2] = $962;
        $964 = $35;
        $965 = $back$i40;
        $966 = (($964) + ($965)|0);
        $967 = $40;
        HEAP32[$967>>2] = $966;
       } else {
        $968 = $dictBase$i34;
        $969 = $matchIndex$i35;
        $970 = (($968) + ($969)|0);
        $matchPtr1$i41 = $970;
        $971 = $matchPtr1$i41;
        $972 = (_LZ4_read32_2985($971)|0);
        $973 = $35;
        $974 = (_LZ4_read32_2985($973)|0);
        $975 = ($972|0)==($974|0);
        if (!($975)) {
         break;
        }
        $back3$i43 = 0;
        $976 = $35;
        $977 = $dictLimit$i31;
        $978 = $matchIndex$i35;
        $979 = (($977) - ($978))|0;
        $980 = (($976) + ($979)|0);
        $vLimit$i44 = $980;
        $981 = $vLimit$i44;
        $982 = $37;
        $983 = ($981>>>0)>($982>>>0);
        if ($983) {
         $984 = $37;
         $vLimit$i44 = $984;
        }
        $985 = $35;
        $986 = ((($985)) + 4|0);
        $987 = $matchPtr1$i41;
        $988 = ((($987)) + 4|0);
        $989 = $vLimit$i44;
        $990 = (_LZ4_count_2986($986,$988,$989)|0);
        $991 = (($990) + 4)|0;
        $mlt2$i42 = $991;
        $992 = $35;
        $993 = $mlt2$i42;
        $994 = (($992) + ($993)|0);
        $995 = $vLimit$i44;
        $996 = ($994|0)==($995|0);
        do {
         if ($996) {
          $997 = $vLimit$i44;
          $998 = $37;
          $999 = ($997>>>0)<($998>>>0);
          if (!($999)) {
           break;
          }
          $1000 = $35;
          $1001 = $mlt2$i42;
          $1002 = (($1000) + ($1001)|0);
          $1003 = $base$i30;
          $1004 = $dictLimit$i31;
          $1005 = (($1003) + ($1004)|0);
          $1006 = $37;
          $1007 = (_LZ4_count_2986($1002,$1005,$1006)|0);
          $1008 = $mlt2$i42;
          $1009 = (($1008) + ($1007))|0;
          $mlt2$i42 = $1009;
         }
        } while(0);
        while(1) {
         $1010 = $35;
         $1011 = $back3$i43;
         $1012 = (($1010) + ($1011)|0);
         $1013 = $36;
         $1014 = ($1012>>>0)>($1013>>>0);
         do {
          if ($1014) {
           $1015 = $matchIndex$i35;
           $1016 = $back3$i43;
           $1017 = (($1015) + ($1016))|0;
           $1018 = $lowLimit$i33;
           $1019 = ($1017>>>0)>($1018>>>0);
           if (!($1019)) {
            $1770 = 0;
            break;
           }
           $1020 = $back3$i43;
           $1021 = (($1020) - 1)|0;
           $1022 = $35;
           $1023 = (($1022) + ($1021)|0);
           $1024 = HEAP8[$1023>>0]|0;
           $1025 = $1024&255;
           $1026 = $back3$i43;
           $1027 = (($1026) - 1)|0;
           $1028 = $matchPtr1$i41;
           $1029 = (($1028) + ($1027)|0);
           $1030 = HEAP8[$1029>>0]|0;
           $1031 = $1030&255;
           $1032 = ($1025|0)==($1031|0);
           $1770 = $1032;
          } else {
           $1770 = 0;
          }
         } while(0);
         $1033 = $back3$i43;
         if (!($1770)) {
          break;
         }
         $1034 = (($1033) + -1)|0;
         $back3$i43 = $1034;
        }
        $1035 = $mlt2$i42;
        $1036 = (($1035) - ($1033))|0;
        $mlt2$i42 = $1036;
        $1037 = $mlt2$i42;
        $1038 = $38;
        $1039 = ($1037|0)>($1038|0);
        if (!($1039)) {
         break;
        }
        $1040 = $mlt2$i42;
        $38 = $1040;
        $1041 = $base$i30;
        $1042 = $matchIndex$i35;
        $1043 = (($1041) + ($1042)|0);
        $1044 = $back3$i43;
        $1045 = (($1043) + ($1044)|0);
        $1046 = $39;
        HEAP32[$1046>>2] = $1045;
        $1047 = $35;
        $1048 = $back3$i43;
        $1049 = (($1047) + ($1048)|0);
        $1050 = $40;
        HEAP32[$1050>>2] = $1049;
       }
      } while(0);
      $1051 = $matchIndex$i35;
      $1052 = $1051&65535;
      $1053 = $1052&65535;
      $1054 = $chainTable$i28;
      $1055 = (($1054) + ($1053<<1)|0);
      $1056 = HEAP16[$1055>>1]|0;
      $1057 = $1056&65535;
      $1058 = $matchIndex$i35;
      $1059 = (($1058) - ($1057))|0;
      $matchIndex$i35 = $1059;
     }
     $1060 = $38;
     $ml3 = $1060;
    } else {
     $1061 = $ml2;
     $ml3 = $1061;
    }
    $1062 = $ml3;
    $1063 = $ml2;
    $1064 = ($1062|0)==($1063|0);
    if ($1064) {
     break L41;
    }
    $1338 = HEAP32[$start3>>2]|0;
    $1339 = HEAP32[$ip>>2]|0;
    $1340 = $ml;
    $1341 = (($1339) + ($1340)|0);
    $1342 = ((($1341)) + 3|0);
    $1343 = ($1338>>>0)<($1342>>>0);
    if ($1343) {
     $1344 = HEAP32[$start3>>2]|0;
     $1345 = HEAP32[$ip>>2]|0;
     $1346 = $ml;
     $1347 = (($1345) + ($1346)|0);
     $1348 = ($1344>>>0)>=($1347>>>0);
     if ($1348) {
      break;
     }
     $1512 = HEAP32[$start3>>2]|0;
     HEAP32[$start2>>2] = $1512;
     $1513 = HEAP32[$ref3>>2]|0;
     HEAP32[$ref2>>2] = $1513;
     $1514 = $ml3;
     $ml2 = $1514;
     continue;
    }
    $1515 = HEAP32[$start2>>2]|0;
    $1516 = HEAP32[$ip>>2]|0;
    $1517 = $ml;
    $1518 = (($1516) + ($1517)|0);
    $1519 = ($1515>>>0)<($1518>>>0);
    do {
     if ($1519) {
      $1520 = HEAP32[$start2>>2]|0;
      $1521 = HEAP32[$ip>>2]|0;
      $1522 = $1520;
      $1523 = $1521;
      $1524 = (($1522) - ($1523))|0;
      $1525 = ($1524|0)<(15);
      if (!($1525)) {
       $1562 = HEAP32[$start2>>2]|0;
       $1563 = HEAP32[$ip>>2]|0;
       $1564 = $1562;
       $1565 = $1563;
       $1566 = (($1564) - ($1565))|0;
       $ml = $1566;
       break;
      }
      $1526 = $ml;
      $1527 = ($1526|0)>(18);
      if ($1527) {
       $ml = 18;
      }
      $1528 = HEAP32[$ip>>2]|0;
      $1529 = $ml;
      $1530 = (($1528) + ($1529)|0);
      $1531 = HEAP32[$start2>>2]|0;
      $1532 = $ml2;
      $1533 = (($1531) + ($1532)|0);
      $1534 = ((($1533)) + -4|0);
      $1535 = ($1530>>>0)>($1534>>>0);
      if ($1535) {
       $1536 = HEAP32[$start2>>2]|0;
       $1537 = HEAP32[$ip>>2]|0;
       $1538 = $1536;
       $1539 = $1537;
       $1540 = (($1538) - ($1539))|0;
       $1541 = $ml2;
       $1542 = (($1540) + ($1541))|0;
       $1543 = (($1542) - 4)|0;
       $ml = $1543;
      }
      $1544 = $ml;
      $1545 = HEAP32[$start2>>2]|0;
      $1546 = HEAP32[$ip>>2]|0;
      $1547 = $1545;
      $1548 = $1546;
      $1549 = (($1547) - ($1548))|0;
      $1550 = (($1544) - ($1549))|0;
      $correction2 = $1550;
      $1551 = $correction2;
      $1552 = ($1551|0)>(0);
      if (!($1552)) {
       break;
      }
      $1553 = $correction2;
      $1554 = HEAP32[$start2>>2]|0;
      $1555 = (($1554) + ($1553)|0);
      HEAP32[$start2>>2] = $1555;
      $1556 = $correction2;
      $1557 = HEAP32[$ref2>>2]|0;
      $1558 = (($1557) + ($1556)|0);
      HEAP32[$ref2>>2] = $1558;
      $1559 = $correction2;
      $1560 = $ml2;
      $1561 = (($1560) - ($1559))|0;
      $ml2 = $1561;
     }
    } while(0);
    $1567 = $ml;
    $1568 = HEAP32[$ref>>2]|0;
    $1569 = $74;
    $1570 = $oend;
    $1 = $ip;
    $2 = $op;
    $3 = $anchor;
    $4 = $1567;
    $5 = $1568;
    $6 = $1569;
    $7 = $1570;
    $1571 = $1;
    $1572 = HEAP32[$1571>>2]|0;
    $1573 = $3;
    $1574 = HEAP32[$1573>>2]|0;
    $1575 = $1572;
    $1576 = $1574;
    $1577 = (($1575) - ($1576))|0;
    $length$i59 = $1577;
    $1578 = $2;
    $1579 = HEAP32[$1578>>2]|0;
    $1580 = ((($1579)) + 1|0);
    HEAP32[$1578>>2] = $1580;
    $token$i60 = $1579;
    $1581 = $6;
    $1582 = ($1581|0)!=(0);
    if ($1582) {
     $1583 = $2;
     $1584 = HEAP32[$1583>>2]|0;
     $1585 = $length$i59;
     $1586 = $1585 >> 8;
     $1587 = (($1584) + ($1586)|0);
     $1588 = $length$i59;
     $1589 = (($1587) + ($1588)|0);
     $1590 = ((($1589)) + 8|0);
     $1591 = $7;
     $1592 = ($1590>>>0)>($1591>>>0);
     if ($1592) {
      $0 = 1;
     } else {
      label = 229;
     }
    } else {
     label = 229;
    }
    L174: do {
     if ((label|0) == 229) {
      label = 0;
      $1593 = $length$i59;
      $1594 = ($1593|0)>=(15);
      if ($1594) {
       $1595 = $token$i60;
       HEAP8[$1595>>0] = -16;
       $1596 = $length$i59;
       $1597 = (($1596) - 15)|0;
       $len$i61 = $1597;
       while(1) {
        $1598 = $len$i61;
        $1599 = ($1598|0)>(254);
        if (!($1599)) {
         break;
        }
        $1600 = $2;
        $1601 = HEAP32[$1600>>2]|0;
        $1602 = ((($1601)) + 1|0);
        HEAP32[$1600>>2] = $1602;
        HEAP8[$1601>>0] = -1;
        $1603 = $len$i61;
        $1604 = (($1603) - 255)|0;
        $len$i61 = $1604;
       }
       $1605 = $len$i61;
       $1606 = $1605&255;
       $1607 = $2;
       $1608 = HEAP32[$1607>>2]|0;
       $1609 = ((($1608)) + 1|0);
       HEAP32[$1607>>2] = $1609;
       HEAP8[$1608>>0] = $1606;
      } else {
       $1610 = $length$i59;
       $1611 = $1610 << 4;
       $1612 = $1611&255;
       $1613 = $token$i60;
       HEAP8[$1613>>0] = $1612;
      }
      $1614 = $2;
      $1615 = HEAP32[$1614>>2]|0;
      $1616 = $3;
      $1617 = HEAP32[$1616>>2]|0;
      $1618 = $2;
      $1619 = HEAP32[$1618>>2]|0;
      $1620 = $length$i59;
      $1621 = (($1619) + ($1620)|0);
      _LZ4_wildCopy_2994($1615,$1617,$1621);
      $1622 = $length$i59;
      $1623 = $2;
      $1624 = HEAP32[$1623>>2]|0;
      $1625 = (($1624) + ($1622)|0);
      HEAP32[$1623>>2] = $1625;
      $1626 = $2;
      $1627 = HEAP32[$1626>>2]|0;
      $1628 = $1;
      $1629 = HEAP32[$1628>>2]|0;
      $1630 = $5;
      $1631 = $1629;
      $1632 = $1630;
      $1633 = (($1631) - ($1632))|0;
      $1634 = $1633&65535;
      _LZ4_writeLE16_2996($1627,$1634);
      $1635 = $2;
      $1636 = HEAP32[$1635>>2]|0;
      $1637 = ((($1636)) + 2|0);
      HEAP32[$1635>>2] = $1637;
      $1638 = $4;
      $1639 = (($1638) - 4)|0;
      $length$i59 = $1639;
      $1640 = $6;
      $1641 = ($1640|0)!=(0);
      do {
       if ($1641) {
        $1642 = $2;
        $1643 = HEAP32[$1642>>2]|0;
        $1644 = $length$i59;
        $1645 = $1644 >> 8;
        $1646 = (($1643) + ($1645)|0);
        $1647 = ((($1646)) + 6|0);
        $1648 = $7;
        $1649 = ($1647>>>0)>($1648>>>0);
        if (!($1649)) {
         break;
        }
        $0 = 1;
        break L174;
       }
      } while(0);
      $1650 = $length$i59;
      $1651 = ($1650|0)>=(15);
      if ($1651) {
       $1652 = $token$i60;
       $1653 = HEAP8[$1652>>0]|0;
       $1654 = $1653&255;
       $1655 = (($1654) + 15)|0;
       $1656 = $1655&255;
       HEAP8[$1652>>0] = $1656;
       $1657 = $length$i59;
       $1658 = (($1657) - 15)|0;
       $length$i59 = $1658;
       while(1) {
        $1659 = $length$i59;
        $1660 = ($1659|0)>(509);
        if (!($1660)) {
         break;
        }
        $1661 = $2;
        $1662 = HEAP32[$1661>>2]|0;
        $1663 = ((($1662)) + 1|0);
        HEAP32[$1661>>2] = $1663;
        HEAP8[$1662>>0] = -1;
        $1664 = $2;
        $1665 = HEAP32[$1664>>2]|0;
        $1666 = ((($1665)) + 1|0);
        HEAP32[$1664>>2] = $1666;
        HEAP8[$1665>>0] = -1;
        $1667 = $length$i59;
        $1668 = (($1667) - 510)|0;
        $length$i59 = $1668;
       }
       $1669 = $length$i59;
       $1670 = ($1669|0)>(254);
       if ($1670) {
        $1671 = $length$i59;
        $1672 = (($1671) - 255)|0;
        $length$i59 = $1672;
        $1673 = $2;
        $1674 = HEAP32[$1673>>2]|0;
        $1675 = ((($1674)) + 1|0);
        HEAP32[$1673>>2] = $1675;
        HEAP8[$1674>>0] = -1;
       }
       $1676 = $length$i59;
       $1677 = $1676&255;
       $1678 = $2;
       $1679 = HEAP32[$1678>>2]|0;
       $1680 = ((($1679)) + 1|0);
       HEAP32[$1678>>2] = $1680;
       HEAP8[$1679>>0] = $1677;
      } else {
       $1681 = $length$i59;
       $1682 = $1681&255;
       $1683 = $1682&255;
       $1684 = $token$i60;
       $1685 = HEAP8[$1684>>0]|0;
       $1686 = $1685&255;
       $1687 = (($1686) + ($1683))|0;
       $1688 = $1687&255;
       HEAP8[$1684>>0] = $1688;
      }
      $1689 = $4;
      $1690 = $1;
      $1691 = HEAP32[$1690>>2]|0;
      $1692 = (($1691) + ($1689)|0);
      HEAP32[$1690>>2] = $1692;
      $1693 = $1;
      $1694 = HEAP32[$1693>>2]|0;
      $1695 = $3;
      HEAP32[$1695>>2] = $1694;
      $0 = 0;
     }
    } while(0);
    $1696 = $0;
    $1697 = ($1696|0)!=(0);
    if ($1697) {
     label = 248;
     break L7;
    }
    $1698 = HEAP32[$start2>>2]|0;
    HEAP32[$ip>>2] = $1698;
    $1699 = HEAP32[$ref2>>2]|0;
    HEAP32[$ref>>2] = $1699;
    $1700 = $ml2;
    $ml = $1700;
    $1701 = HEAP32[$start3>>2]|0;
    HEAP32[$start2>>2] = $1701;
    $1702 = HEAP32[$ref3>>2]|0;
    HEAP32[$ref2>>2] = $1702;
    $1703 = $ml3;
    $ml2 = $1703;
   }
   $1349 = HEAP32[$start2>>2]|0;
   $1350 = HEAP32[$ip>>2]|0;
   $1351 = $ml;
   $1352 = (($1350) + ($1351)|0);
   $1353 = ($1349>>>0)<($1352>>>0);
   if ($1353) {
    $1354 = HEAP32[$ip>>2]|0;
    $1355 = $ml;
    $1356 = (($1354) + ($1355)|0);
    $1357 = HEAP32[$start2>>2]|0;
    $1358 = $1356;
    $1359 = $1357;
    $1360 = (($1358) - ($1359))|0;
    $correction1 = $1360;
    $1361 = $correction1;
    $1362 = HEAP32[$start2>>2]|0;
    $1363 = (($1362) + ($1361)|0);
    HEAP32[$start2>>2] = $1363;
    $1364 = $correction1;
    $1365 = HEAP32[$ref2>>2]|0;
    $1366 = (($1365) + ($1364)|0);
    HEAP32[$ref2>>2] = $1366;
    $1367 = $correction1;
    $1368 = $ml2;
    $1369 = (($1368) - ($1367))|0;
    $ml2 = $1369;
    $1370 = $ml2;
    $1371 = ($1370|0)<(4);
    if ($1371) {
     $1372 = HEAP32[$start3>>2]|0;
     HEAP32[$start2>>2] = $1372;
     $1373 = HEAP32[$ref3>>2]|0;
     HEAP32[$ref2>>2] = $1373;
     $1374 = $ml3;
     $ml2 = $1374;
    }
   }
   $1375 = $ml;
   $1376 = HEAP32[$ref>>2]|0;
   $1377 = $74;
   $1378 = $oend;
   $9 = $ip;
   $10 = $op;
   $11 = $anchor;
   $12 = $1375;
   $13 = $1376;
   $14 = $1377;
   $15 = $1378;
   $1379 = $9;
   $1380 = HEAP32[$1379>>2]|0;
   $1381 = $11;
   $1382 = HEAP32[$1381>>2]|0;
   $1383 = $1380;
   $1384 = $1382;
   $1385 = (($1383) - ($1384))|0;
   $length$i55 = $1385;
   $1386 = $10;
   $1387 = HEAP32[$1386>>2]|0;
   $1388 = ((($1387)) + 1|0);
   HEAP32[$1386>>2] = $1388;
   $token$i56 = $1387;
   $1389 = $14;
   $1390 = ($1389|0)!=(0);
   if ($1390) {
    $1391 = $10;
    $1392 = HEAP32[$1391>>2]|0;
    $1393 = $length$i55;
    $1394 = $1393 >> 8;
    $1395 = (($1392) + ($1394)|0);
    $1396 = $length$i55;
    $1397 = (($1395) + ($1396)|0);
    $1398 = ((($1397)) + 8|0);
    $1399 = $15;
    $1400 = ($1398>>>0)>($1399>>>0);
    if ($1400) {
     $8 = 1;
    } else {
     label = 195;
    }
   } else {
    label = 195;
   }
   do {
    if ((label|0) == 195) {
     label = 0;
     $1401 = $length$i55;
     $1402 = ($1401|0)>=(15);
     if ($1402) {
      $1403 = $token$i56;
      HEAP8[$1403>>0] = -16;
      $1404 = $length$i55;
      $1405 = (($1404) - 15)|0;
      $len$i57 = $1405;
      while(1) {
       $1406 = $len$i57;
       $1407 = ($1406|0)>(254);
       if (!($1407)) {
        break;
       }
       $1408 = $10;
       $1409 = HEAP32[$1408>>2]|0;
       $1410 = ((($1409)) + 1|0);
       HEAP32[$1408>>2] = $1410;
       HEAP8[$1409>>0] = -1;
       $1411 = $len$i57;
       $1412 = (($1411) - 255)|0;
       $len$i57 = $1412;
      }
      $1413 = $len$i57;
      $1414 = $1413&255;
      $1415 = $10;
      $1416 = HEAP32[$1415>>2]|0;
      $1417 = ((($1416)) + 1|0);
      HEAP32[$1415>>2] = $1417;
      HEAP8[$1416>>0] = $1414;
     } else {
      $1418 = $length$i55;
      $1419 = $1418 << 4;
      $1420 = $1419&255;
      $1421 = $token$i56;
      HEAP8[$1421>>0] = $1420;
     }
     $1422 = $10;
     $1423 = HEAP32[$1422>>2]|0;
     $1424 = $11;
     $1425 = HEAP32[$1424>>2]|0;
     $1426 = $10;
     $1427 = HEAP32[$1426>>2]|0;
     $1428 = $length$i55;
     $1429 = (($1427) + ($1428)|0);
     _LZ4_wildCopy_2994($1423,$1425,$1429);
     $1430 = $length$i55;
     $1431 = $10;
     $1432 = HEAP32[$1431>>2]|0;
     $1433 = (($1432) + ($1430)|0);
     HEAP32[$1431>>2] = $1433;
     $1434 = $10;
     $1435 = HEAP32[$1434>>2]|0;
     $1436 = $9;
     $1437 = HEAP32[$1436>>2]|0;
     $1438 = $13;
     $1439 = $1437;
     $1440 = $1438;
     $1441 = (($1439) - ($1440))|0;
     $1442 = $1441&65535;
     _LZ4_writeLE16_2996($1435,$1442);
     $1443 = $10;
     $1444 = HEAP32[$1443>>2]|0;
     $1445 = ((($1444)) + 2|0);
     HEAP32[$1443>>2] = $1445;
     $1446 = $12;
     $1447 = (($1446) - 4)|0;
     $length$i55 = $1447;
     $1448 = $14;
     $1449 = ($1448|0)!=(0);
     if ($1449) {
      $1450 = $10;
      $1451 = HEAP32[$1450>>2]|0;
      $1452 = $length$i55;
      $1453 = $1452 >> 8;
      $1454 = (($1451) + ($1453)|0);
      $1455 = ((($1454)) + 6|0);
      $1456 = $15;
      $1457 = ($1455>>>0)>($1456>>>0);
      if ($1457) {
       $8 = 1;
       break;
      }
     }
     $1458 = $length$i55;
     $1459 = ($1458|0)>=(15);
     if ($1459) {
      $1460 = $token$i56;
      $1461 = HEAP8[$1460>>0]|0;
      $1462 = $1461&255;
      $1463 = (($1462) + 15)|0;
      $1464 = $1463&255;
      HEAP8[$1460>>0] = $1464;
      $1465 = $length$i55;
      $1466 = (($1465) - 15)|0;
      $length$i55 = $1466;
      while(1) {
       $1467 = $length$i55;
       $1468 = ($1467|0)>(509);
       if (!($1468)) {
        break;
       }
       $1469 = $10;
       $1470 = HEAP32[$1469>>2]|0;
       $1471 = ((($1470)) + 1|0);
       HEAP32[$1469>>2] = $1471;
       HEAP8[$1470>>0] = -1;
       $1472 = $10;
       $1473 = HEAP32[$1472>>2]|0;
       $1474 = ((($1473)) + 1|0);
       HEAP32[$1472>>2] = $1474;
       HEAP8[$1473>>0] = -1;
       $1475 = $length$i55;
       $1476 = (($1475) - 510)|0;
       $length$i55 = $1476;
      }
      $1477 = $length$i55;
      $1478 = ($1477|0)>(254);
      if ($1478) {
       $1479 = $length$i55;
       $1480 = (($1479) - 255)|0;
       $length$i55 = $1480;
       $1481 = $10;
       $1482 = HEAP32[$1481>>2]|0;
       $1483 = ((($1482)) + 1|0);
       HEAP32[$1481>>2] = $1483;
       HEAP8[$1482>>0] = -1;
      }
      $1484 = $length$i55;
      $1485 = $1484&255;
      $1486 = $10;
      $1487 = HEAP32[$1486>>2]|0;
      $1488 = ((($1487)) + 1|0);
      HEAP32[$1486>>2] = $1488;
      HEAP8[$1487>>0] = $1485;
     } else {
      $1489 = $length$i55;
      $1490 = $1489&255;
      $1491 = $1490&255;
      $1492 = $token$i56;
      $1493 = HEAP8[$1492>>0]|0;
      $1494 = $1493&255;
      $1495 = (($1494) + ($1491))|0;
      $1496 = $1495&255;
      HEAP8[$1492>>0] = $1496;
     }
     $1497 = $12;
     $1498 = $9;
     $1499 = HEAP32[$1498>>2]|0;
     $1500 = (($1499) + ($1497)|0);
     HEAP32[$1498>>2] = $1500;
     $1501 = $9;
     $1502 = HEAP32[$1501>>2]|0;
     $1503 = $11;
     HEAP32[$1503>>2] = $1502;
     $8 = 0;
    }
   } while(0);
   $1504 = $8;
   $1505 = ($1504|0)!=(0);
   if ($1505) {
    label = 214;
    break L7;
   }
   $1506 = HEAP32[$start3>>2]|0;
   HEAP32[$ip>>2] = $1506;
   $1507 = HEAP32[$ref3>>2]|0;
   HEAP32[$ref>>2] = $1507;
   $1508 = $ml3;
   $ml = $1508;
   $1509 = HEAP32[$start2>>2]|0;
   $start0 = $1509;
   $1510 = HEAP32[$ref2>>2]|0;
   $ref0 = $1510;
   $1511 = $ml2;
   $ml0 = $1511;
  }
  if ((label|0) == 69) {
   label = 0;
   $588 = $ml;
   $589 = HEAP32[$ref>>2]|0;
   $590 = $74;
   $591 = $oend;
   $43 = $ip;
   $44 = $op;
   $45 = $anchor;
   $46 = $588;
   $47 = $589;
   $48 = $590;
   $49 = $591;
   $592 = $43;
   $593 = HEAP32[$592>>2]|0;
   $594 = $45;
   $595 = HEAP32[$594>>2]|0;
   $596 = $593;
   $597 = $595;
   $598 = (($596) - ($597))|0;
   $length$i = $598;
   $599 = $44;
   $600 = HEAP32[$599>>2]|0;
   $601 = ((($600)) + 1|0);
   HEAP32[$599>>2] = $601;
   $token$i = $600;
   $602 = $48;
   $603 = ($602|0)!=(0);
   if ($603) {
    $604 = $44;
    $605 = HEAP32[$604>>2]|0;
    $606 = $length$i;
    $607 = $606 >> 8;
    $608 = (($605) + ($607)|0);
    $609 = $length$i;
    $610 = (($608) + ($609)|0);
    $611 = ((($610)) + 8|0);
    $612 = $49;
    $613 = ($611>>>0)>($612>>>0);
    if ($613) {
     $42 = 1;
    } else {
     label = 72;
    }
   } else {
    label = 72;
   }
   do {
    if ((label|0) == 72) {
     label = 0;
     $614 = $length$i;
     $615 = ($614|0)>=(15);
     if ($615) {
      $616 = $token$i;
      HEAP8[$616>>0] = -16;
      $617 = $length$i;
      $618 = (($617) - 15)|0;
      $len$i = $618;
      while(1) {
       $619 = $len$i;
       $620 = ($619|0)>(254);
       if (!($620)) {
        break;
       }
       $621 = $44;
       $622 = HEAP32[$621>>2]|0;
       $623 = ((($622)) + 1|0);
       HEAP32[$621>>2] = $623;
       HEAP8[$622>>0] = -1;
       $624 = $len$i;
       $625 = (($624) - 255)|0;
       $len$i = $625;
      }
      $626 = $len$i;
      $627 = $626&255;
      $628 = $44;
      $629 = HEAP32[$628>>2]|0;
      $630 = ((($629)) + 1|0);
      HEAP32[$628>>2] = $630;
      HEAP8[$629>>0] = $627;
     } else {
      $631 = $length$i;
      $632 = $631 << 4;
      $633 = $632&255;
      $634 = $token$i;
      HEAP8[$634>>0] = $633;
     }
     $635 = $44;
     $636 = HEAP32[$635>>2]|0;
     $637 = $45;
     $638 = HEAP32[$637>>2]|0;
     $639 = $44;
     $640 = HEAP32[$639>>2]|0;
     $641 = $length$i;
     $642 = (($640) + ($641)|0);
     _LZ4_wildCopy_2994($636,$638,$642);
     $643 = $length$i;
     $644 = $44;
     $645 = HEAP32[$644>>2]|0;
     $646 = (($645) + ($643)|0);
     HEAP32[$644>>2] = $646;
     $647 = $44;
     $648 = HEAP32[$647>>2]|0;
     $649 = $43;
     $650 = HEAP32[$649>>2]|0;
     $651 = $47;
     $652 = $650;
     $653 = $651;
     $654 = (($652) - ($653))|0;
     $655 = $654&65535;
     _LZ4_writeLE16_2996($648,$655);
     $656 = $44;
     $657 = HEAP32[$656>>2]|0;
     $658 = ((($657)) + 2|0);
     HEAP32[$656>>2] = $658;
     $659 = $46;
     $660 = (($659) - 4)|0;
     $length$i = $660;
     $661 = $48;
     $662 = ($661|0)!=(0);
     if ($662) {
      $663 = $44;
      $664 = HEAP32[$663>>2]|0;
      $665 = $length$i;
      $666 = $665 >> 8;
      $667 = (($664) + ($666)|0);
      $668 = ((($667)) + 6|0);
      $669 = $49;
      $670 = ($668>>>0)>($669>>>0);
      if ($670) {
       $42 = 1;
       break;
      }
     }
     $671 = $length$i;
     $672 = ($671|0)>=(15);
     if ($672) {
      $673 = $token$i;
      $674 = HEAP8[$673>>0]|0;
      $675 = $674&255;
      $676 = (($675) + 15)|0;
      $677 = $676&255;
      HEAP8[$673>>0] = $677;
      $678 = $length$i;
      $679 = (($678) - 15)|0;
      $length$i = $679;
      while(1) {
       $680 = $length$i;
       $681 = ($680|0)>(509);
       if (!($681)) {
        break;
       }
       $682 = $44;
       $683 = HEAP32[$682>>2]|0;
       $684 = ((($683)) + 1|0);
       HEAP32[$682>>2] = $684;
       HEAP8[$683>>0] = -1;
       $685 = $44;
       $686 = HEAP32[$685>>2]|0;
       $687 = ((($686)) + 1|0);
       HEAP32[$685>>2] = $687;
       HEAP8[$686>>0] = -1;
       $688 = $length$i;
       $689 = (($688) - 510)|0;
       $length$i = $689;
      }
      $690 = $length$i;
      $691 = ($690|0)>(254);
      if ($691) {
       $692 = $length$i;
       $693 = (($692) - 255)|0;
       $length$i = $693;
       $694 = $44;
       $695 = HEAP32[$694>>2]|0;
       $696 = ((($695)) + 1|0);
       HEAP32[$694>>2] = $696;
       HEAP8[$695>>0] = -1;
      }
      $697 = $length$i;
      $698 = $697&255;
      $699 = $44;
      $700 = HEAP32[$699>>2]|0;
      $701 = ((($700)) + 1|0);
      HEAP32[$699>>2] = $701;
      HEAP8[$700>>0] = $698;
     } else {
      $702 = $length$i;
      $703 = $702&255;
      $704 = $703&255;
      $705 = $token$i;
      $706 = HEAP8[$705>>0]|0;
      $707 = $706&255;
      $708 = (($707) + ($704))|0;
      $709 = $708&255;
      HEAP8[$705>>0] = $709;
     }
     $710 = $46;
     $711 = $43;
     $712 = HEAP32[$711>>2]|0;
     $713 = (($712) + ($710)|0);
     HEAP32[$711>>2] = $713;
     $714 = $43;
     $715 = HEAP32[$714>>2]|0;
     $716 = $45;
     HEAP32[$716>>2] = $715;
     $42 = 0;
    }
   } while(0);
   $717 = $42;
   $718 = ($717|0)!=(0);
   if ($718) {
    label = 91;
    break;
   } else {
    continue;
   }
  }
  $1065 = HEAP32[$start2>>2]|0;
  $1066 = HEAP32[$ip>>2]|0;
  $1067 = $ml;
  $1068 = (($1066) + ($1067)|0);
  $1069 = ($1065>>>0)<($1068>>>0);
  if ($1069) {
   $1070 = HEAP32[$start2>>2]|0;
   $1071 = HEAP32[$ip>>2]|0;
   $1072 = $1070;
   $1073 = $1071;
   $1074 = (($1072) - ($1073))|0;
   $ml = $1074;
  }
  $1075 = $ml;
  $1076 = HEAP32[$ref>>2]|0;
  $1077 = $74;
  $1078 = $oend;
  $25 = $ip;
  $26 = $op;
  $27 = $anchor;
  $28 = $1075;
  $29 = $1076;
  $30 = $1077;
  $31 = $1078;
  $1079 = $25;
  $1080 = HEAP32[$1079>>2]|0;
  $1081 = $27;
  $1082 = HEAP32[$1081>>2]|0;
  $1083 = $1080;
  $1084 = $1082;
  $1085 = (($1083) - ($1084))|0;
  $length$i47 = $1085;
  $1086 = $26;
  $1087 = HEAP32[$1086>>2]|0;
  $1088 = ((($1087)) + 1|0);
  HEAP32[$1086>>2] = $1088;
  $token$i48 = $1087;
  $1089 = $30;
  $1090 = ($1089|0)!=(0);
  if ($1090) {
   $1091 = $26;
   $1092 = HEAP32[$1091>>2]|0;
   $1093 = $length$i47;
   $1094 = $1093 >> 8;
   $1095 = (($1092) + ($1094)|0);
   $1096 = $length$i47;
   $1097 = (($1095) + ($1096)|0);
   $1098 = ((($1097)) + 8|0);
   $1099 = $31;
   $1100 = ($1098>>>0)>($1099>>>0);
   if ($1100) {
    $24 = 1;
   } else {
    label = 144;
   }
  } else {
   label = 144;
  }
  do {
   if ((label|0) == 144) {
    label = 0;
    $1101 = $length$i47;
    $1102 = ($1101|0)>=(15);
    if ($1102) {
     $1103 = $token$i48;
     HEAP8[$1103>>0] = -16;
     $1104 = $length$i47;
     $1105 = (($1104) - 15)|0;
     $len$i49 = $1105;
     while(1) {
      $1106 = $len$i49;
      $1107 = ($1106|0)>(254);
      if (!($1107)) {
       break;
      }
      $1108 = $26;
      $1109 = HEAP32[$1108>>2]|0;
      $1110 = ((($1109)) + 1|0);
      HEAP32[$1108>>2] = $1110;
      HEAP8[$1109>>0] = -1;
      $1111 = $len$i49;
      $1112 = (($1111) - 255)|0;
      $len$i49 = $1112;
     }
     $1113 = $len$i49;
     $1114 = $1113&255;
     $1115 = $26;
     $1116 = HEAP32[$1115>>2]|0;
     $1117 = ((($1116)) + 1|0);
     HEAP32[$1115>>2] = $1117;
     HEAP8[$1116>>0] = $1114;
    } else {
     $1118 = $length$i47;
     $1119 = $1118 << 4;
     $1120 = $1119&255;
     $1121 = $token$i48;
     HEAP8[$1121>>0] = $1120;
    }
    $1122 = $26;
    $1123 = HEAP32[$1122>>2]|0;
    $1124 = $27;
    $1125 = HEAP32[$1124>>2]|0;
    $1126 = $26;
    $1127 = HEAP32[$1126>>2]|0;
    $1128 = $length$i47;
    $1129 = (($1127) + ($1128)|0);
    _LZ4_wildCopy_2994($1123,$1125,$1129);
    $1130 = $length$i47;
    $1131 = $26;
    $1132 = HEAP32[$1131>>2]|0;
    $1133 = (($1132) + ($1130)|0);
    HEAP32[$1131>>2] = $1133;
    $1134 = $26;
    $1135 = HEAP32[$1134>>2]|0;
    $1136 = $25;
    $1137 = HEAP32[$1136>>2]|0;
    $1138 = $29;
    $1139 = $1137;
    $1140 = $1138;
    $1141 = (($1139) - ($1140))|0;
    $1142 = $1141&65535;
    _LZ4_writeLE16_2996($1135,$1142);
    $1143 = $26;
    $1144 = HEAP32[$1143>>2]|0;
    $1145 = ((($1144)) + 2|0);
    HEAP32[$1143>>2] = $1145;
    $1146 = $28;
    $1147 = (($1146) - 4)|0;
    $length$i47 = $1147;
    $1148 = $30;
    $1149 = ($1148|0)!=(0);
    if ($1149) {
     $1150 = $26;
     $1151 = HEAP32[$1150>>2]|0;
     $1152 = $length$i47;
     $1153 = $1152 >> 8;
     $1154 = (($1151) + ($1153)|0);
     $1155 = ((($1154)) + 6|0);
     $1156 = $31;
     $1157 = ($1155>>>0)>($1156>>>0);
     if ($1157) {
      $24 = 1;
      break;
     }
    }
    $1158 = $length$i47;
    $1159 = ($1158|0)>=(15);
    if ($1159) {
     $1160 = $token$i48;
     $1161 = HEAP8[$1160>>0]|0;
     $1162 = $1161&255;
     $1163 = (($1162) + 15)|0;
     $1164 = $1163&255;
     HEAP8[$1160>>0] = $1164;
     $1165 = $length$i47;
     $1166 = (($1165) - 15)|0;
     $length$i47 = $1166;
     while(1) {
      $1167 = $length$i47;
      $1168 = ($1167|0)>(509);
      if (!($1168)) {
       break;
      }
      $1169 = $26;
      $1170 = HEAP32[$1169>>2]|0;
      $1171 = ((($1170)) + 1|0);
      HEAP32[$1169>>2] = $1171;
      HEAP8[$1170>>0] = -1;
      $1172 = $26;
      $1173 = HEAP32[$1172>>2]|0;
      $1174 = ((($1173)) + 1|0);
      HEAP32[$1172>>2] = $1174;
      HEAP8[$1173>>0] = -1;
      $1175 = $length$i47;
      $1176 = (($1175) - 510)|0;
      $length$i47 = $1176;
     }
     $1177 = $length$i47;
     $1178 = ($1177|0)>(254);
     if ($1178) {
      $1179 = $length$i47;
      $1180 = (($1179) - 255)|0;
      $length$i47 = $1180;
      $1181 = $26;
      $1182 = HEAP32[$1181>>2]|0;
      $1183 = ((($1182)) + 1|0);
      HEAP32[$1181>>2] = $1183;
      HEAP8[$1182>>0] = -1;
     }
     $1184 = $length$i47;
     $1185 = $1184&255;
     $1186 = $26;
     $1187 = HEAP32[$1186>>2]|0;
     $1188 = ((($1187)) + 1|0);
     HEAP32[$1186>>2] = $1188;
     HEAP8[$1187>>0] = $1185;
    } else {
     $1189 = $length$i47;
     $1190 = $1189&255;
     $1191 = $1190&255;
     $1192 = $token$i48;
     $1193 = HEAP8[$1192>>0]|0;
     $1194 = $1193&255;
     $1195 = (($1194) + ($1191))|0;
     $1196 = $1195&255;
     HEAP8[$1192>>0] = $1196;
    }
    $1197 = $28;
    $1198 = $25;
    $1199 = HEAP32[$1198>>2]|0;
    $1200 = (($1199) + ($1197)|0);
    HEAP32[$1198>>2] = $1200;
    $1201 = $25;
    $1202 = HEAP32[$1201>>2]|0;
    $1203 = $27;
    HEAP32[$1203>>2] = $1202;
    $24 = 0;
   }
  } while(0);
  $1204 = $24;
  $1205 = ($1204|0)!=(0);
  if ($1205) {
   label = 163;
   break;
  }
  $1206 = HEAP32[$start2>>2]|0;
  HEAP32[$ip>>2] = $1206;
  $1207 = $ml2;
  $1208 = HEAP32[$ref2>>2]|0;
  $1209 = $74;
  $1210 = $oend;
  $17 = $ip;
  $18 = $op;
  $19 = $anchor;
  $20 = $1207;
  $21 = $1208;
  $22 = $1209;
  $23 = $1210;
  $1211 = $17;
  $1212 = HEAP32[$1211>>2]|0;
  $1213 = $19;
  $1214 = HEAP32[$1213>>2]|0;
  $1215 = $1212;
  $1216 = $1214;
  $1217 = (($1215) - ($1216))|0;
  $length$i51 = $1217;
  $1218 = $18;
  $1219 = HEAP32[$1218>>2]|0;
  $1220 = ((($1219)) + 1|0);
  HEAP32[$1218>>2] = $1220;
  $token$i52 = $1219;
  $1221 = $22;
  $1222 = ($1221|0)!=(0);
  if ($1222) {
   $1223 = $18;
   $1224 = HEAP32[$1223>>2]|0;
   $1225 = $length$i51;
   $1226 = $1225 >> 8;
   $1227 = (($1224) + ($1226)|0);
   $1228 = $length$i51;
   $1229 = (($1227) + ($1228)|0);
   $1230 = ((($1229)) + 8|0);
   $1231 = $23;
   $1232 = ($1230>>>0)>($1231>>>0);
   if ($1232) {
    $16 = 1;
   } else {
    label = 167;
   }
  } else {
   label = 167;
  }
  do {
   if ((label|0) == 167) {
    label = 0;
    $1233 = $length$i51;
    $1234 = ($1233|0)>=(15);
    if ($1234) {
     $1235 = $token$i52;
     HEAP8[$1235>>0] = -16;
     $1236 = $length$i51;
     $1237 = (($1236) - 15)|0;
     $len$i53 = $1237;
     while(1) {
      $1238 = $len$i53;
      $1239 = ($1238|0)>(254);
      if (!($1239)) {
       break;
      }
      $1240 = $18;
      $1241 = HEAP32[$1240>>2]|0;
      $1242 = ((($1241)) + 1|0);
      HEAP32[$1240>>2] = $1242;
      HEAP8[$1241>>0] = -1;
      $1243 = $len$i53;
      $1244 = (($1243) - 255)|0;
      $len$i53 = $1244;
     }
     $1245 = $len$i53;
     $1246 = $1245&255;
     $1247 = $18;
     $1248 = HEAP32[$1247>>2]|0;
     $1249 = ((($1248)) + 1|0);
     HEAP32[$1247>>2] = $1249;
     HEAP8[$1248>>0] = $1246;
    } else {
     $1250 = $length$i51;
     $1251 = $1250 << 4;
     $1252 = $1251&255;
     $1253 = $token$i52;
     HEAP8[$1253>>0] = $1252;
    }
    $1254 = $18;
    $1255 = HEAP32[$1254>>2]|0;
    $1256 = $19;
    $1257 = HEAP32[$1256>>2]|0;
    $1258 = $18;
    $1259 = HEAP32[$1258>>2]|0;
    $1260 = $length$i51;
    $1261 = (($1259) + ($1260)|0);
    _LZ4_wildCopy_2994($1255,$1257,$1261);
    $1262 = $length$i51;
    $1263 = $18;
    $1264 = HEAP32[$1263>>2]|0;
    $1265 = (($1264) + ($1262)|0);
    HEAP32[$1263>>2] = $1265;
    $1266 = $18;
    $1267 = HEAP32[$1266>>2]|0;
    $1268 = $17;
    $1269 = HEAP32[$1268>>2]|0;
    $1270 = $21;
    $1271 = $1269;
    $1272 = $1270;
    $1273 = (($1271) - ($1272))|0;
    $1274 = $1273&65535;
    _LZ4_writeLE16_2996($1267,$1274);
    $1275 = $18;
    $1276 = HEAP32[$1275>>2]|0;
    $1277 = ((($1276)) + 2|0);
    HEAP32[$1275>>2] = $1277;
    $1278 = $20;
    $1279 = (($1278) - 4)|0;
    $length$i51 = $1279;
    $1280 = $22;
    $1281 = ($1280|0)!=(0);
    if ($1281) {
     $1282 = $18;
     $1283 = HEAP32[$1282>>2]|0;
     $1284 = $length$i51;
     $1285 = $1284 >> 8;
     $1286 = (($1283) + ($1285)|0);
     $1287 = ((($1286)) + 6|0);
     $1288 = $23;
     $1289 = ($1287>>>0)>($1288>>>0);
     if ($1289) {
      $16 = 1;
      break;
     }
    }
    $1290 = $length$i51;
    $1291 = ($1290|0)>=(15);
    if ($1291) {
     $1292 = $token$i52;
     $1293 = HEAP8[$1292>>0]|0;
     $1294 = $1293&255;
     $1295 = (($1294) + 15)|0;
     $1296 = $1295&255;
     HEAP8[$1292>>0] = $1296;
     $1297 = $length$i51;
     $1298 = (($1297) - 15)|0;
     $length$i51 = $1298;
     while(1) {
      $1299 = $length$i51;
      $1300 = ($1299|0)>(509);
      if (!($1300)) {
       break;
      }
      $1301 = $18;
      $1302 = HEAP32[$1301>>2]|0;
      $1303 = ((($1302)) + 1|0);
      HEAP32[$1301>>2] = $1303;
      HEAP8[$1302>>0] = -1;
      $1304 = $18;
      $1305 = HEAP32[$1304>>2]|0;
      $1306 = ((($1305)) + 1|0);
      HEAP32[$1304>>2] = $1306;
      HEAP8[$1305>>0] = -1;
      $1307 = $length$i51;
      $1308 = (($1307) - 510)|0;
      $length$i51 = $1308;
     }
     $1309 = $length$i51;
     $1310 = ($1309|0)>(254);
     if ($1310) {
      $1311 = $length$i51;
      $1312 = (($1311) - 255)|0;
      $length$i51 = $1312;
      $1313 = $18;
      $1314 = HEAP32[$1313>>2]|0;
      $1315 = ((($1314)) + 1|0);
      HEAP32[$1313>>2] = $1315;
      HEAP8[$1314>>0] = -1;
     }
     $1316 = $length$i51;
     $1317 = $1316&255;
     $1318 = $18;
     $1319 = HEAP32[$1318>>2]|0;
     $1320 = ((($1319)) + 1|0);
     HEAP32[$1318>>2] = $1320;
     HEAP8[$1319>>0] = $1317;
    } else {
     $1321 = $length$i51;
     $1322 = $1321&255;
     $1323 = $1322&255;
     $1324 = $token$i52;
     $1325 = HEAP8[$1324>>0]|0;
     $1326 = $1325&255;
     $1327 = (($1326) + ($1323))|0;
     $1328 = $1327&255;
     HEAP8[$1324>>0] = $1328;
    }
    $1329 = $20;
    $1330 = $17;
    $1331 = HEAP32[$1330>>2]|0;
    $1332 = (($1331) + ($1329)|0);
    HEAP32[$1330>>2] = $1332;
    $1333 = $17;
    $1334 = HEAP32[$1333>>2]|0;
    $1335 = $19;
    HEAP32[$1335>>2] = $1334;
    $16 = 0;
   }
  } while(0);
  $1336 = $16;
  $1337 = ($1336|0)!=(0);
  if ($1337) {
   label = 186;
   break;
  }
 }
 if ((label|0) == 91) {
  $67 = 0;
  $1766 = $67;
  STACKTOP = sp;return ($1766|0);
 }
 else if ((label|0) == 163) {
  $67 = 0;
  $1766 = $67;
  STACKTOP = sp;return ($1766|0);
 }
 else if ((label|0) == 186) {
  $67 = 0;
  $1766 = $67;
  STACKTOP = sp;return ($1766|0);
 }
 else if ((label|0) == 214) {
  $67 = 0;
  $1766 = $67;
  STACKTOP = sp;return ($1766|0);
 }
 else if ((label|0) == 248) {
  $67 = 0;
  $1766 = $67;
  STACKTOP = sp;return ($1766|0);
 }
 else if ((label|0) == 250) {
  $1704 = $iend;
  $1705 = HEAP32[$anchor>>2]|0;
  $1706 = $1704;
  $1707 = $1705;
  $1708 = (($1706) - ($1707))|0;
  $lastRun = $1708;
  $1709 = $74;
  $1710 = ($1709|0)!=(0);
  if ($1710) {
   $1711 = HEAP32[$op>>2]|0;
   $1712 = $70;
   $1713 = $1711;
   $1714 = $1712;
   $1715 = (($1713) - ($1714))|0;
   $1716 = $lastRun;
   $1717 = (($1715) + ($1716))|0;
   $1718 = (($1717) + 1)|0;
   $1719 = $lastRun;
   $1720 = (($1719) + 255)|0;
   $1721 = (($1720) - 15)|0;
   $1722 = (($1721>>>0) / 255)&-1;
   $1723 = (($1718) + ($1722))|0;
   $1724 = $72;
   $1725 = ($1723>>>0)>($1724>>>0);
   if ($1725) {
    $67 = 0;
    $1766 = $67;
    STACKTOP = sp;return ($1766|0);
   }
  }
  $1726 = $lastRun;
  $1727 = ($1726|0)>=(15);
  if ($1727) {
   $1728 = HEAP32[$op>>2]|0;
   $1729 = ((($1728)) + 1|0);
   HEAP32[$op>>2] = $1729;
   HEAP8[$1728>>0] = -16;
   $1730 = $lastRun;
   $1731 = (($1730) - 15)|0;
   $lastRun = $1731;
   while(1) {
    $1732 = $lastRun;
    $1733 = ($1732|0)>(254);
    if (!($1733)) {
     break;
    }
    $1734 = HEAP32[$op>>2]|0;
    $1735 = ((($1734)) + 1|0);
    HEAP32[$op>>2] = $1735;
    HEAP8[$1734>>0] = -1;
    $1736 = $lastRun;
    $1737 = (($1736) - 255)|0;
    $lastRun = $1737;
   }
   $1738 = $lastRun;
   $1739 = $1738&255;
   $1740 = HEAP32[$op>>2]|0;
   $1741 = ((($1740)) + 1|0);
   HEAP32[$op>>2] = $1741;
   HEAP8[$1740>>0] = $1739;
  } else {
   $1742 = $lastRun;
   $1743 = $1742 << 4;
   $1744 = $1743&255;
   $1745 = HEAP32[$op>>2]|0;
   $1746 = ((($1745)) + 1|0);
   HEAP32[$op>>2] = $1746;
   HEAP8[$1745>>0] = $1744;
  }
  $1747 = HEAP32[$op>>2]|0;
  $1748 = HEAP32[$anchor>>2]|0;
  $1749 = $iend;
  $1750 = HEAP32[$anchor>>2]|0;
  $1751 = $1749;
  $1752 = $1750;
  $1753 = (($1751) - ($1752))|0;
  _memcpy(($1747|0),($1748|0),($1753|0))|0;
  $1754 = $iend;
  $1755 = HEAP32[$anchor>>2]|0;
  $1756 = $1754;
  $1757 = $1755;
  $1758 = (($1756) - ($1757))|0;
  $1759 = HEAP32[$op>>2]|0;
  $1760 = (($1759) + ($1758)|0);
  HEAP32[$op>>2] = $1760;
  $1761 = HEAP32[$op>>2]|0;
  $1762 = $70;
  $1763 = $1761;
  $1764 = $1762;
  $1765 = (($1763) - ($1764))|0;
  $67 = $1765;
  $1766 = $67;
  STACKTOP = sp;return ($1766|0);
 }
 return (0)|0;
}
function _LZ4HC_hashPtr($ptr) {
 $ptr = $ptr|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $ptr;
 $1 = $0;
 $2 = (_LZ4_read32_2985($1)|0);
 $3 = Math_imul($2, -1640531535)|0;
 $4 = $3 >>> 17;
 STACKTOP = sp;return ($4|0);
}
function _LZ4_read32_2985($memPtr) {
 $memPtr = $memPtr|0;
 var $0 = 0, $1 = 0, $2 = 0, $val32 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $val32 = sp;
 $0 = $memPtr;
 $1 = $0;
 ;HEAP8[$val32>>0]=HEAP8[$1>>0]|0;HEAP8[$val32+1>>0]=HEAP8[$1+1>>0]|0;HEAP8[$val32+2>>0]=HEAP8[$1+2>>0]|0;HEAP8[$val32+3>>0]=HEAP8[$1+3>>0]|0;
 $2 = HEAP32[$val32>>2]|0;
 STACKTOP = sp;return ($2|0);
}
function _LZ4_count_2986($pIn,$pMatch,$pInLimit) {
 $pIn = $pIn|0;
 $pMatch = $pMatch|0;
 $pInLimit = $pInLimit|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $diff = 0;
 var $pStart = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $pIn;
 $2 = $pMatch;
 $3 = $pInLimit;
 $4 = $1;
 $pStart = $4;
 while(1) {
  $5 = $1;
  $6 = $3;
  $7 = ((($6)) + -3|0);
  $8 = ($5>>>0)<($7>>>0);
  $9 = $8&1;
  $10 = ($9|0)!=(0);
  if (!($10)) {
   break;
  }
  $11 = $2;
  $12 = (_LZ4_read_ARCH_2987($11)|0);
  $13 = $1;
  $14 = (_LZ4_read_ARCH_2987($13)|0);
  $15 = $12 ^ $14;
  $diff = $15;
  $16 = $diff;
  $17 = ($16|0)!=(0);
  if ($17) {
   label = 5;
   break;
  }
  $18 = $1;
  $19 = ((($18)) + 4|0);
  $1 = $19;
  $20 = $2;
  $21 = ((($20)) + 4|0);
  $2 = $21;
 }
 if ((label|0) == 5) {
  $22 = $diff;
  $23 = (_LZ4_NbCommonBytes_2990($22)|0);
  $24 = $1;
  $25 = (($24) + ($23)|0);
  $1 = $25;
  $26 = $1;
  $27 = $pStart;
  $28 = $26;
  $29 = $27;
  $30 = (($28) - ($29))|0;
  $0 = $30;
  $78 = $0;
  STACKTOP = sp;return ($78|0);
 }
 $31 = (_LZ4_64bits_2988()|0);
 $32 = ($31|0)!=(0);
 if ($32) {
  $33 = $1;
  $34 = $3;
  $35 = ((($34)) + -3|0);
  $36 = ($33>>>0)<($35>>>0);
  if ($36) {
   $37 = $2;
   $38 = (_LZ4_read32_2985($37)|0);
   $39 = $1;
   $40 = (_LZ4_read32_2985($39)|0);
   $41 = ($38|0)==($40|0);
   if ($41) {
    $42 = $1;
    $43 = ((($42)) + 4|0);
    $1 = $43;
    $44 = $2;
    $45 = ((($44)) + 4|0);
    $2 = $45;
   }
  }
 }
 $46 = $1;
 $47 = $3;
 $48 = ((($47)) + -1|0);
 $49 = ($46>>>0)<($48>>>0);
 if ($49) {
  $50 = $2;
  $51 = (_LZ4_read16_2993($50)|0);
  $52 = $51&65535;
  $53 = $1;
  $54 = (_LZ4_read16_2993($53)|0);
  $55 = $54&65535;
  $56 = ($52|0)==($55|0);
  if ($56) {
   $57 = $1;
   $58 = ((($57)) + 2|0);
   $1 = $58;
   $59 = $2;
   $60 = ((($59)) + 2|0);
   $2 = $60;
  }
 }
 $61 = $1;
 $62 = $3;
 $63 = ($61>>>0)<($62>>>0);
 if ($63) {
  $64 = $2;
  $65 = HEAP8[$64>>0]|0;
  $66 = $65&255;
  $67 = $1;
  $68 = HEAP8[$67>>0]|0;
  $69 = $68&255;
  $70 = ($66|0)==($69|0);
  if ($70) {
   $71 = $1;
   $72 = ((($71)) + 1|0);
   $1 = $72;
  }
 }
 $73 = $1;
 $74 = $pStart;
 $75 = $73;
 $76 = $74;
 $77 = (($75) - ($76))|0;
 $0 = $77;
 $78 = $0;
 STACKTOP = sp;return ($78|0);
}
function _LZ4_read_ARCH_2987($p) {
 $p = $p|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $p;
 $2 = (_LZ4_64bits_2988()|0);
 $3 = ($2|0)!=(0);
 $4 = $1;
 if ($3) {
  $5 = (_LZ4_read64_2989($4)|0);
  $6 = tempRet0;
  $0 = $5;
  $8 = $0;
  STACKTOP = sp;return ($8|0);
 } else {
  $7 = (_LZ4_read32_2985($4)|0);
  $0 = $7;
  $8 = $0;
  STACKTOP = sp;return ($8|0);
 }
 return (0)|0;
}
function _LZ4_64bits_2988() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _LZ4_read64_2989($memPtr) {
 $memPtr = $memPtr|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $val64 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $val64 = sp;
 $0 = $memPtr;
 $1 = $0;
 ;HEAP8[$val64>>0]=HEAP8[$1>>0]|0;HEAP8[$val64+1>>0]=HEAP8[$1+1>>0]|0;HEAP8[$val64+2>>0]=HEAP8[$1+2>>0]|0;HEAP8[$val64+3>>0]=HEAP8[$1+3>>0]|0;HEAP8[$val64+4>>0]=HEAP8[$1+4>>0]|0;HEAP8[$val64+5>>0]=HEAP8[$1+5>>0]|0;HEAP8[$val64+6>>0]=HEAP8[$1+6>>0]|0;HEAP8[$val64+7>>0]=HEAP8[$1+7>>0]|0;
 $2 = $val64;
 $3 = $2;
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2) + 4)|0;
 $6 = $5;
 $7 = HEAP32[$6>>2]|0;
 tempRet0 = ($7);
 STACKTOP = sp;return ($4|0);
}
function _LZ4_NbCommonBytes_2990($val) {
 $val = $val|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $val;
 $2 = (_LZ4_isLittleEndian_2991()|0);
 $3 = ($2|0)!=(0);
 $4 = (_LZ4_64bits_2988()|0);
 $5 = ($4|0)!=(0);
 $6 = $1;
 do {
  if ($3) {
   if ($5) {
    $7 = (_llvm_cttz_i64(($6|0),0,0)|0);
    $8 = tempRet0;
    $9 = $7 >> 3;
    $0 = $9;
    break;
   } else {
    $10 = (_llvm_cttz_i32(($6|0))|0);
    $11 = $10 >> 3;
    $0 = $11;
    break;
   }
  } else {
   if ($5) {
    $12 = (_llvm_ctlz_i64(($6|0),0,0)|0);
    $13 = tempRet0;
    $14 = $12 >> 3;
    $0 = $14;
    break;
   } else {
    $15 = (Math_clz32(($6|0))|0);
    $16 = $15 >> 3;
    $0 = $16;
    break;
   }
  }
 } while(0);
 $17 = $0;
 STACKTOP = sp;return ($17|0);
}
function _LZ4_isLittleEndian_2991() {
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[21896]|0;
 $1 = $0&255;
 return ($1|0);
}
function _LZ4_read16_2993($memPtr) {
 $memPtr = $memPtr|0;
 var $0 = 0, $1 = 0, $2 = 0, $val16 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $val16 = sp + 4|0;
 $0 = $memPtr;
 $1 = $0;
 ;HEAP8[$val16>>0]=HEAP8[$1>>0]|0;HEAP8[$val16+1>>0]=HEAP8[$1+1>>0]|0;
 $2 = HEAP16[$val16>>1]|0;
 STACKTOP = sp;return ($2|0);
}
function _LZ4_wildCopy_2994($dstPtr,$srcPtr,$dstEnd) {
 $dstPtr = $dstPtr|0;
 $srcPtr = $srcPtr|0;
 $dstEnd = $dstEnd|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $d = 0, $e = 0, $s = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $dstPtr;
 $1 = $srcPtr;
 $2 = $dstEnd;
 $3 = $0;
 $d = $3;
 $4 = $1;
 $s = $4;
 $5 = $2;
 $e = $5;
 while(1) {
  $6 = $d;
  $7 = $s;
  _LZ4_copy8_2995($6,$7);
  $8 = $d;
  $9 = ((($8)) + 8|0);
  $d = $9;
  $10 = $s;
  $11 = ((($10)) + 8|0);
  $s = $11;
  $12 = $d;
  $13 = $e;
  $14 = ($12>>>0)<($13>>>0);
  if (!($14)) {
   break;
  }
 }
 STACKTOP = sp;return;
}
function _LZ4_copy8_2995($dstPtr,$srcPtr) {
 $dstPtr = $dstPtr|0;
 $srcPtr = $srcPtr|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $dstPtr;
 $1 = $srcPtr;
 $2 = $0;
 $3 = $1;
 ;HEAP8[$2>>0]=HEAP8[$3>>0]|0;HEAP8[$2+1>>0]=HEAP8[$3+1>>0]|0;HEAP8[$2+2>>0]=HEAP8[$3+2>>0]|0;HEAP8[$2+3>>0]=HEAP8[$3+3>>0]|0;HEAP8[$2+4>>0]=HEAP8[$3+4>>0]|0;HEAP8[$2+5>>0]=HEAP8[$3+5>>0]|0;HEAP8[$2+6>>0]=HEAP8[$3+6>>0]|0;HEAP8[$2+7>>0]=HEAP8[$3+7>>0]|0;
 STACKTOP = sp;return;
}
function _LZ4_writeLE16_2996($memPtr,$value) {
 $memPtr = $memPtr|0;
 $value = $value|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = sp + 8|0;
 $0 = $memPtr;
 HEAP16[$1>>1] = $value;
 $2 = (_LZ4_isLittleEndian_2991()|0);
 $3 = ($2|0)!=(0);
 $4 = $0;
 if ($3) {
  ;HEAP8[$4>>0]=HEAP8[$1>>0]|0;HEAP8[$4+1>>0]=HEAP8[$1+1>>0]|0;
  STACKTOP = sp;return;
 } else {
  $p = $4;
  $5 = HEAP16[$1>>1]|0;
  $6 = $5&255;
  $7 = $p;
  HEAP8[$7>>0] = $6;
  $8 = HEAP16[$1>>1]|0;
  $9 = $8&65535;
  $10 = $9 >> 8;
  $11 = $10&255;
  $12 = $p;
  $13 = ((($12)) + 1|0);
  HEAP8[$13>>0] = $11;
  STACKTOP = sp;return;
 }
}
function _LZ4_compress_HC($src,$dst,$srcSize,$maxDstSize,$compressionLevel) {
 $src = $src|0;
 $dst = $dst|0;
 $srcSize = $srcSize|0;
 $maxDstSize = $maxDstSize|0;
 $compressionLevel = $compressionLevel|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $state = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 262208|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $state = sp;
 $0 = $src;
 $1 = $dst;
 $2 = $srcSize;
 $3 = $maxDstSize;
 $4 = $compressionLevel;
 $5 = $0;
 $6 = $1;
 $7 = $2;
 $8 = $3;
 $9 = $4;
 $10 = (_LZ4_compress_HC_extStateHC($state,$5,$6,$7,$8,$9)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZN4love5sound7lullaby13Mpg123DecoderD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (21908); //@line 168 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $0 = ((($this)) + 52|0); //@line 169 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 169 "modules/sound/lullaby/Mpg123Decoder.cpp"
 __THREW__ = 0;
 invoke_vi(3885,($1|0)); //@line 169 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 171 "modules/sound/lullaby/Mpg123Decoder.cpp"
  ___clang_call_terminate($4); //@line 171 "modules/sound/lullaby/Mpg123Decoder.cpp"
  // unreachable; //@line 171 "modules/sound/lullaby/Mpg123Decoder.cpp"
 } else {
  __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 171 "modules/sound/lullaby/Mpg123Decoder.cpp"
  return; //@line 171 "modules/sound/lullaby/Mpg123Decoder.cpp"
 }
}
function __ZN4love5sound7lullaby13Mpg123DecoderD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (21908); //@line 168 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $0 = ((($this)) + 52|0); //@line 169 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 169 "modules/sound/lullaby/Mpg123Decoder.cpp"
 __THREW__ = 0;
 invoke_vi(3885,($1|0)); //@line 169 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 171 "modules/sound/lullaby/Mpg123Decoder.cpp"
  ___clang_call_terminate($4); //@line 171 "modules/sound/lullaby/Mpg123Decoder.cpp"
  // unreachable; //@line 171 "modules/sound/lullaby/Mpg123Decoder.cpp"
 } else {
  __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 171 "modules/sound/lullaby/Mpg123Decoder.cpp"
  __ZdlPv($this); //@line 168 "modules/sound/lullaby/Mpg123Decoder.cpp"
  return; //@line 171 "modules/sound/lullaby/Mpg123Decoder.cpp"
 }
}
function __ZN4love5sound7lullaby13Mpg123Decoder5cloneEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(72)|0); //@line 197 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $1 = ((($this)) + 8|0); //@line 139 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 139 "./common/Object.h"
 $3 = ((($this)) + 12|0); //@line 197 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $4 = ((($this)) + 24|0); //@line 197 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 197 "modules/sound/lullaby/Mpg123Decoder.cpp"
 __THREW__ = 0;
 invoke_viiii(3752,($0|0),($2|0),($3|0),($5|0)); //@line 197 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $8 = ___cxa_find_matching_catch_2()|0;
  $9 = tempRet0;
  __ZdlPv($0); //@line 197 "modules/sound/lullaby/Mpg123Decoder.cpp"
  ___resumeException($8|0);
  // unreachable;
 } else {
  return ($0|0); //@line 197 "modules/sound/lullaby/Mpg123Decoder.cpp"
 }
 return (0)|0;
}
function __ZN4love5sound7lullaby13Mpg123DecoderC2EPNS_4DataERKNSt3__112basic_stringIcNS5_11char_traitsIcEENS5_9allocatorIcEEEEi($this,$data,$ext,$bufferSize) {
 $this = $this|0;
 $data = $data|0;
 $ext = $ext|0;
 $bufferSize = $bufferSize|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $rate = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $rate = sp + 36|0;
 __ZN4love5sound7lullaby7DecoderC2EPNS_4DataERKNSt3__112basic_stringIcNS5_11char_traitsIcEENS5_9allocatorIcEEEEi($this,$data,$ext,$bufferSize); //@line 109 "modules/sound/lullaby/Mpg123Decoder.cpp"
 HEAP32[$this>>2] = (21908); //@line 114 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $0 = ((($this)) + 40|0); //@line 110 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $1 = HEAP32[$data>>2]|0; //@line 49 "modules/sound/lullaby/Mpg123Decoder.h"
 $2 = ((($1)) + 16|0); //@line 49 "modules/sound/lullaby/Mpg123Decoder.h"
 $3 = HEAP32[$2>>2]|0; //@line 49 "modules/sound/lullaby/Mpg123Decoder.h"
 __THREW__ = 0;
 $4 = (invoke_ii($3|0,($data|0))|0); //@line 49 "modules/sound/lullaby/Mpg123Decoder.h"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 L1: do {
  if (!($6)) {
   HEAP32[$0>>2] = $4; //@line 49 "modules/sound/lullaby/Mpg123Decoder.h"
   $7 = HEAP32[$data>>2]|0; //@line 50 "modules/sound/lullaby/Mpg123Decoder.h"
   $8 = ((($7)) + 20|0); //@line 50 "modules/sound/lullaby/Mpg123Decoder.h"
   $9 = HEAP32[$8>>2]|0; //@line 50 "modules/sound/lullaby/Mpg123Decoder.h"
   __THREW__ = 0;
   $10 = (invoke_ii($9|0,($data|0))|0); //@line 50 "modules/sound/lullaby/Mpg123Decoder.h"
   $11 = __THREW__; __THREW__ = 0;
   $12 = $11&1;
   if (!($12)) {
    $13 = ((($this)) + 44|0); //@line 50 "modules/sound/lullaby/Mpg123Decoder.h"
    HEAP32[$13>>2] = $10; //@line 50 "modules/sound/lullaby/Mpg123Decoder.h"
    $14 = ((($this)) + 48|0); //@line 51 "modules/sound/lullaby/Mpg123Decoder.h"
    HEAP32[$14>>2] = 0; //@line 51 "modules/sound/lullaby/Mpg123Decoder.h"
    $15 = ((($this)) + 52|0); //@line 111 "modules/sound/lullaby/Mpg123Decoder.cpp"
    HEAP32[$15>>2] = 0; //@line 111 "modules/sound/lullaby/Mpg123Decoder.cpp"
    $16 = ((($this)) + 56|0); //@line 112 "modules/sound/lullaby/Mpg123Decoder.cpp"
    HEAP32[$16>>2] = 2; //@line 112 "modules/sound/lullaby/Mpg123Decoder.cpp"
    $17 = ((($this)) + 64|0); //@line 113 "modules/sound/lullaby/Mpg123Decoder.cpp"
    HEAPF64[$17>>3] = -2.0; //@line 113 "modules/sound/lullaby/Mpg123Decoder.cpp"
    $18 = HEAP8[1251894]|0; //@line 117 "modules/sound/lullaby/Mpg123Decoder.cpp"
    $19 = ($18<<24>>24)==(0); //@line 117 "modules/sound/lullaby/Mpg123Decoder.cpp"
    do {
     if ($19) {
      __THREW__ = 0;
      $20 = (invoke_i(3886)|0); //@line 119 "modules/sound/lullaby/Mpg123Decoder.cpp"
      $21 = __THREW__; __THREW__ = 0;
      $22 = $21&1;
      if ($22) {
       break L1;
      }
      $23 = ($20|0)==(0); //@line 120 "modules/sound/lullaby/Mpg123Decoder.cpp"
      if ($23) {
       HEAP8[1251894] = 1; //@line 122 "modules/sound/lullaby/Mpg123Decoder.cpp"
       break;
      }
      $24 = (___cxa_allocate_exception(16)|0); //@line 121 "modules/sound/lullaby/Mpg123Decoder.cpp"
      __THREW__ = 0;
      invoke_viii(3349,($24|0),(364531|0),($vararg_buffer|0)); //@line 121 "modules/sound/lullaby/Mpg123Decoder.cpp"
      $25 = __THREW__; __THREW__ = 0;
      $26 = $25&1;
      if (!($26)) {
       __THREW__ = 0;
       invoke_viii(3350,($24|0),(96|0),(344|0)); //@line 121 "modules/sound/lullaby/Mpg123Decoder.cpp"
       $27 = __THREW__; __THREW__ = 0;
       break L1;
      }
      $30 = ___cxa_find_matching_catch_2()|0;
      $31 = tempRet0;
      ___cxa_free_exception(($24|0)); //@line 121 "modules/sound/lullaby/Mpg123Decoder.cpp"
      $$1 = $30;$$12 = $31;
      __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 165 "modules/sound/lullaby/Mpg123Decoder.cpp"
      ___resumeException($$1|0);
      // unreachable;
     }
    } while(0);
    __THREW__ = 0;
    $32 = (invoke_iii(3887,(0|0),(0|0))|0); //@line 126 "modules/sound/lullaby/Mpg123Decoder.cpp"
    $33 = __THREW__; __THREW__ = 0;
    $34 = $33&1;
    if (!($34)) {
     HEAP32[$15>>2] = $32; //@line 126 "modules/sound/lullaby/Mpg123Decoder.cpp"
     $35 = ($32|0)==(0|0); //@line 127 "modules/sound/lullaby/Mpg123Decoder.cpp"
     if ($35) {
      $36 = (___cxa_allocate_exception(16)|0); //@line 128 "modules/sound/lullaby/Mpg123Decoder.cpp"
      __THREW__ = 0;
      invoke_viii(3349,($36|0),(364560|0),($vararg_buffer1|0)); //@line 128 "modules/sound/lullaby/Mpg123Decoder.cpp"
      $37 = __THREW__; __THREW__ = 0;
      $38 = $37&1;
      if (!($38)) {
       __THREW__ = 0;
       invoke_viii(3350,($36|0),(96|0),(344|0)); //@line 128 "modules/sound/lullaby/Mpg123Decoder.cpp"
       $39 = __THREW__; __THREW__ = 0;
       break;
      }
      $40 = ___cxa_find_matching_catch_2()|0;
      $41 = tempRet0;
      ___cxa_free_exception(($36|0)); //@line 128 "modules/sound/lullaby/Mpg123Decoder.cpp"
      $$1 = $40;$$12 = $41;
      __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 165 "modules/sound/lullaby/Mpg123Decoder.cpp"
      ___resumeException($$1|0);
      // unreachable;
     }
     __THREW__ = 0;
     (invoke_iiiid(3888,($32|0),2,32,0.0)|0); //@line 131 "modules/sound/lullaby/Mpg123Decoder.cpp"
     $42 = __THREW__; __THREW__ = 0;
     $43 = $42&1;
     if (!($43)) {
      $44 = HEAP32[$15>>2]|0; //@line 135 "modules/sound/lullaby/Mpg123Decoder.cpp"
      __THREW__ = 0;
      $45 = (invoke_iiiii(3889,($44|0),(3890|0),(3891|0),(3892|0))|0); //@line 135 "modules/sound/lullaby/Mpg123Decoder.cpp"
      $46 = __THREW__; __THREW__ = 0;
      $47 = $46&1;
      do {
       if ($47) {
        label = 21;
       } else {
        $48 = ($45|0)==(0); //@line 136 "modules/sound/lullaby/Mpg123Decoder.cpp"
        if (!($48)) {
         $49 = (___cxa_allocate_exception(16)|0); //@line 137 "modules/sound/lullaby/Mpg123Decoder.cpp"
         __THREW__ = 0;
         invoke_viii(3349,($49|0),(364586|0),($vararg_buffer3|0)); //@line 137 "modules/sound/lullaby/Mpg123Decoder.cpp"
         $50 = __THREW__; __THREW__ = 0;
         $51 = $50&1;
         if ($51) {
          $55 = ___cxa_find_matching_catch_3(96|0)|0;
          $56 = tempRet0;
          ___cxa_free_exception(($49|0)); //@line 137 "modules/sound/lullaby/Mpg123Decoder.cpp"
          $$0 = $55;$$01 = $56;
          break;
         } else {
          __THREW__ = 0;
          invoke_viii(3350,($49|0),(96|0),(344|0)); //@line 137 "modules/sound/lullaby/Mpg123Decoder.cpp"
          $52 = __THREW__; __THREW__ = 0;
          label = 21;
          break;
         }
        }
        $57 = HEAP32[$15>>2]|0; //@line 139 "modules/sound/lullaby/Mpg123Decoder.cpp"
        __THREW__ = 0;
        $58 = (invoke_iii(3893,($57|0),($0|0))|0); //@line 139 "modules/sound/lullaby/Mpg123Decoder.cpp"
        $59 = __THREW__; __THREW__ = 0;
        $60 = $59&1;
        if ($60) {
         label = 21;
        } else {
         $61 = ($58|0)==(0); //@line 140 "modules/sound/lullaby/Mpg123Decoder.cpp"
         if (!($61)) {
          $62 = (___cxa_allocate_exception(16)|0); //@line 141 "modules/sound/lullaby/Mpg123Decoder.cpp"
          __THREW__ = 0;
          invoke_viii(3349,($62|0),(364619|0),($vararg_buffer5|0)); //@line 141 "modules/sound/lullaby/Mpg123Decoder.cpp"
          $63 = __THREW__; __THREW__ = 0;
          $64 = $63&1;
          if ($64) {
           $66 = ___cxa_find_matching_catch_3(96|0)|0;
           $67 = tempRet0;
           ___cxa_free_exception(($62|0)); //@line 141 "modules/sound/lullaby/Mpg123Decoder.cpp"
           $$0 = $66;$$01 = $67;
           break;
          } else {
           __THREW__ = 0;
           invoke_viii(3350,($62|0),(96|0),(344|0)); //@line 141 "modules/sound/lullaby/Mpg123Decoder.cpp"
           $65 = __THREW__; __THREW__ = 0;
           label = 21;
           break;
          }
         }
         HEAP32[$rate>>2] = 0; //@line 144 "modules/sound/lullaby/Mpg123Decoder.cpp"
         $68 = HEAP32[$15>>2]|0; //@line 145 "modules/sound/lullaby/Mpg123Decoder.cpp"
         __THREW__ = 0;
         $69 = (invoke_iiiii(3894,($68|0),($rate|0),($16|0),(0|0))|0); //@line 145 "modules/sound/lullaby/Mpg123Decoder.cpp"
         $70 = __THREW__; __THREW__ = 0;
         $71 = $70&1;
         if ($71) {
          label = 21;
         } else {
          $72 = ($69|0)==(-1); //@line 146 "modules/sound/lullaby/Mpg123Decoder.cpp"
          if ($72) {
           $73 = (___cxa_allocate_exception(16)|0); //@line 147 "modules/sound/lullaby/Mpg123Decoder.cpp"
           __THREW__ = 0;
           invoke_viii(3349,($73|0),(364643|0),($vararg_buffer7|0)); //@line 147 "modules/sound/lullaby/Mpg123Decoder.cpp"
           $74 = __THREW__; __THREW__ = 0;
           $75 = $74&1;
           if ($75) {
            $77 = ___cxa_find_matching_catch_3(96|0)|0;
            $78 = tempRet0;
            ___cxa_free_exception(($73|0)); //@line 147 "modules/sound/lullaby/Mpg123Decoder.cpp"
            $$0 = $77;$$01 = $78;
            break;
           } else {
            __THREW__ = 0;
            invoke_viii(3350,($73|0),(96|0),(344|0)); //@line 147 "modules/sound/lullaby/Mpg123Decoder.cpp"
            $76 = __THREW__; __THREW__ = 0;
            label = 21;
            break;
           }
          }
          $85 = HEAP32[$16>>2]|0; //@line 150 "modules/sound/lullaby/Mpg123Decoder.cpp"
          $86 = ($85|0)==(0); //@line 150 "modules/sound/lullaby/Mpg123Decoder.cpp"
          if ($86) {
           HEAP32[$16>>2] = 2; //@line 151 "modules/sound/lullaby/Mpg123Decoder.cpp"
           $89 = 2;
          } else {
           $89 = $85;
          }
          $87 = HEAP32[$15>>2]|0; //@line 154 "modules/sound/lullaby/Mpg123Decoder.cpp"
          $88 = ($89|0)==(2); //@line 154 "modules/sound/lullaby/Mpg123Decoder.cpp"
          $90 = $88 ? 8 : 4; //@line 154 "modules/sound/lullaby/Mpg123Decoder.cpp"
          __THREW__ = 0;
          (invoke_iiiid(3888,($87|0),1,($90|0),0.0)|0); //@line 154 "modules/sound/lullaby/Mpg123Decoder.cpp"
          $91 = __THREW__; __THREW__ = 0;
          $92 = $91&1;
          if ($92) {
           label = 21;
          } else {
           $93 = HEAP32[$15>>2]|0; //@line 155 "modules/sound/lullaby/Mpg123Decoder.cpp"
           __THREW__ = 0;
           (invoke_ii(3895,($93|0))|0); //@line 155 "modules/sound/lullaby/Mpg123Decoder.cpp"
           $94 = __THREW__; __THREW__ = 0;
           $95 = $94&1;
           if ($95) {
            label = 21;
           } else {
            $96 = HEAP32[$15>>2]|0; //@line 156 "modules/sound/lullaby/Mpg123Decoder.cpp"
            $97 = HEAP32[$rate>>2]|0; //@line 156 "modules/sound/lullaby/Mpg123Decoder.cpp"
            $98 = HEAP32[$16>>2]|0; //@line 156 "modules/sound/lullaby/Mpg123Decoder.cpp"
            __THREW__ = 0;
            (invoke_iiiii(3896,($96|0),($97|0),($98|0),208)|0); //@line 156 "modules/sound/lullaby/Mpg123Decoder.cpp"
            $99 = __THREW__; __THREW__ = 0;
            $100 = $99&1;
            if ($100) {
             label = 21;
            } else {
             $101 = HEAP32[$rate>>2]|0; //@line 158 "modules/sound/lullaby/Mpg123Decoder.cpp"
             $102 = ((($this)) + 28|0); //@line 158 "modules/sound/lullaby/Mpg123Decoder.cpp"
             HEAP32[$102>>2] = $101; //@line 158 "modules/sound/lullaby/Mpg123Decoder.cpp"
             STACKTOP = sp;return; //@line 165 "modules/sound/lullaby/Mpg123Decoder.cpp"
            }
           }
          }
         }
        }
       }
      } while(0);
      if ((label|0) == 21) {
       $53 = ___cxa_find_matching_catch_3(96|0)|0;
       $54 = tempRet0;
       $$0 = $53;$$01 = $54;
      }
      $79 = (_llvm_eh_typeid_for((96|0))|0); //@line 159 "modules/sound/lullaby/Mpg123Decoder.cpp"
      $80 = ($$01|0)==($79|0); //@line 159 "modules/sound/lullaby/Mpg123Decoder.cpp"
      if (!($80)) {
       $$1 = $$0;$$12 = $$01;
       __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 165 "modules/sound/lullaby/Mpg123Decoder.cpp"
       ___resumeException($$1|0);
       // unreachable;
      }
      (___cxa_begin_catch(($$0|0))|0); //@line 159 "modules/sound/lullaby/Mpg123Decoder.cpp"
      $81 = HEAP32[$15>>2]|0; //@line 162 "modules/sound/lullaby/Mpg123Decoder.cpp"
      __THREW__ = 0;
      invoke_vi(3885,($81|0)); //@line 162 "modules/sound/lullaby/Mpg123Decoder.cpp"
      $82 = __THREW__; __THREW__ = 0;
      $83 = $82&1;
      if (!($83)) {
       __THREW__ = 0;
       invoke_v(3656); //@line 163 "modules/sound/lullaby/Mpg123Decoder.cpp"
       $84 = __THREW__; __THREW__ = 0;
      }
      $103 = ___cxa_find_matching_catch_2()|0;
      $104 = tempRet0;
      __THREW__ = 0;
      invoke_v(3378); //@line 164 "modules/sound/lullaby/Mpg123Decoder.cpp"
      $105 = __THREW__; __THREW__ = 0;
      $106 = $105&1;
      if ($106) {
       $107 = ___cxa_find_matching_catch_3(0|0)|0;
       $108 = tempRet0;
       ___clang_call_terminate($107); //@line 164 "modules/sound/lullaby/Mpg123Decoder.cpp"
       // unreachable; //@line 164 "modules/sound/lullaby/Mpg123Decoder.cpp"
      } else {
       $$1 = $103;$$12 = $104;
       __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 165 "modules/sound/lullaby/Mpg123Decoder.cpp"
       ___resumeException($$1|0);
       // unreachable;
      }
     }
    }
   }
  }
 } while(0);
 $28 = ___cxa_find_matching_catch_2()|0;
 $29 = tempRet0;
 $$1 = $28;$$12 = $29;
 __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 165 "modules/sound/lullaby/Mpg123Decoder.cpp"
 ___resumeException($$1|0);
 // unreachable;
}
function __ZN4love5sound7lullabyL13read_callbackEPvS2_j($udata,$buffer,$count) {
 $udata = $udata|0;
 $buffer = $buffer|0;
 $count = $count|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($udata)) + 4|0); //@line 44 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 44 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $2 = ((($udata)) + 8|0); //@line 44 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 44 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $4 = (($1) - ($3))|0; //@line 44 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $5 = ($4>>>0)<($count>>>0); //@line 45 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $6 = $5 ? $4 : $count; //@line 45 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $7 = ($6|0)==(0); //@line 47 "modules/sound/lullaby/Mpg123Decoder.cpp"
 if ($7) {
  return ($6|0); //@line 54 "modules/sound/lullaby/Mpg123Decoder.cpp"
 }
 $8 = HEAP32[$udata>>2]|0; //@line 49 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $9 = (($8) + ($3)|0); //@line 49 "modules/sound/lullaby/Mpg123Decoder.cpp"
 _memcpy(($buffer|0),($9|0),($6|0))|0; //@line 49 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $10 = HEAP32[$2>>2]|0; //@line 50 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $11 = (($10) + ($6))|0; //@line 50 "modules/sound/lullaby/Mpg123Decoder.cpp"
 HEAP32[$2>>2] = $11; //@line 50 "modules/sound/lullaby/Mpg123Decoder.cpp"
 return ($6|0); //@line 54 "modules/sound/lullaby/Mpg123Decoder.cpp"
}
function __ZN4love5sound7lullabyL13seek_callbackEPvii($udata,$offset,$whence) {
 $udata = $udata|0;
 $offset = $offset|0;
 $whence = $whence|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 L1: do {
  switch ($whence|0) {
  case 0:  {
   $0 = ($offset|0)<(0); //@line 65 "modules/sound/lullaby/Mpg123Decoder.cpp"
   if ($0) {
    $$0 = -1;
   } else {
    $1 = ((($udata)) + 4|0); //@line 69 "modules/sound/lullaby/Mpg123Decoder.cpp"
    $2 = HEAP32[$1>>2]|0; //@line 69 "modules/sound/lullaby/Mpg123Decoder.cpp"
    $3 = ($2>>>0)>($offset>>>0); //@line 69 "modules/sound/lullaby/Mpg123Decoder.cpp"
    $4 = ((($udata)) + 8|0); //@line 70 "modules/sound/lullaby/Mpg123Decoder.cpp"
    if ($3) {
     HEAP32[$4>>2] = $offset; //@line 70 "modules/sound/lullaby/Mpg123Decoder.cpp"
     $$0 = $offset;
     break L1;
    } else {
     HEAP32[$4>>2] = $2; //@line 72 "modules/sound/lullaby/Mpg123Decoder.cpp"
     $$0 = $2;
     break L1;
    }
   }
   break;
  }
  case 2:  {
   $5 = ((($udata)) + 4|0); //@line 76 "modules/sound/lullaby/Mpg123Decoder.cpp"
   $6 = HEAP32[$5>>2]|0; //@line 76 "modules/sound/lullaby/Mpg123Decoder.cpp"
   $7 = ((($udata)) + 8|0); //@line 76 "modules/sound/lullaby/Mpg123Decoder.cpp"
   HEAP32[$7>>2] = $6; //@line 76 "modules/sound/lullaby/Mpg123Decoder.cpp"
   label = 7;
   break;
  }
  case 1:  {
   label = 7;
   break;
  }
  default: {
   $$0 = -1;
  }
  }
 } while(0);
 do {
  if ((label|0) == 7) {
   $8 = ($offset|0)>(0); //@line 79 "modules/sound/lullaby/Mpg123Decoder.cpp"
   if ($8) {
    $9 = ((($udata)) + 4|0); //@line 81 "modules/sound/lullaby/Mpg123Decoder.cpp"
    $10 = HEAP32[$9>>2]|0; //@line 81 "modules/sound/lullaby/Mpg123Decoder.cpp"
    $11 = ((($udata)) + 8|0); //@line 81 "modules/sound/lullaby/Mpg123Decoder.cpp"
    $12 = HEAP32[$11>>2]|0; //@line 81 "modules/sound/lullaby/Mpg123Decoder.cpp"
    $13 = (($12) + ($offset))|0; //@line 81 "modules/sound/lullaby/Mpg123Decoder.cpp"
    $14 = ($10>>>0)>($13>>>0); //@line 81 "modules/sound/lullaby/Mpg123Decoder.cpp"
    if ($14) {
     HEAP32[$11>>2] = $13; //@line 82 "modules/sound/lullaby/Mpg123Decoder.cpp"
     $$0 = $13;
     break;
    } else {
     HEAP32[$11>>2] = $10; //@line 84 "modules/sound/lullaby/Mpg123Decoder.cpp"
     $$0 = $10;
     break;
    }
   }
   $15 = ($offset|0)<(0); //@line 86 "modules/sound/lullaby/Mpg123Decoder.cpp"
   $16 = ((($udata)) + 8|0); //@line 88 "modules/sound/lullaby/Mpg123Decoder.cpp"
   $17 = HEAP32[$16>>2]|0; //@line 88 "modules/sound/lullaby/Mpg123Decoder.cpp"
   if ($15) {
    $18 = (0 - ($offset))|0; //@line 88 "modules/sound/lullaby/Mpg123Decoder.cpp"
    $19 = ($17>>>0)<($18>>>0); //@line 88 "modules/sound/lullaby/Mpg123Decoder.cpp"
    if ($19) {
     HEAP32[$16>>2] = 0; //@line 91 "modules/sound/lullaby/Mpg123Decoder.cpp"
     $$0 = 0;
     break;
    } else {
     $20 = (($17) + ($offset))|0; //@line 89 "modules/sound/lullaby/Mpg123Decoder.cpp"
     HEAP32[$16>>2] = $20; //@line 89 "modules/sound/lullaby/Mpg123Decoder.cpp"
     $$0 = $20;
     break;
    }
   } else {
    $$0 = $17;
   }
  }
 } while(0);
 return ($$0|0); //@line 99 "modules/sound/lullaby/Mpg123Decoder.cpp"
}
function __ZN4love5sound7lullabyL16cleanup_callbackEPv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 104 "modules/sound/lullaby/Mpg123Decoder.cpp"
}
function __ZN4love5sound7lullaby13Mpg123Decoder6decodeEv($this) {
 $this = $this|0;
 var $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $numbytes = 0, $size$06 = 0, $size$06$lcssa18 = 0, $size$2$ph = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $numbytes = sp;
 $0 = ((($this)) + 24|0); //@line 204 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 204 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $2 = ($1|0)>(0); //@line 204 "modules/sound/lullaby/Mpg123Decoder.cpp"
 if (!($2)) {
  $$2 = 0;
  STACKTOP = sp;return ($$2|0); //@line 225 "modules/sound/lullaby/Mpg123Decoder.cpp"
 }
 $3 = ((($this)) + 36|0); //@line 204 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $4 = ((($this)) + 52|0); //@line 207 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $5 = ((($this)) + 32|0); //@line 207 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $12 = $1;$size$06 = 0;
 L4: while(1) {
  $6 = HEAP8[$3>>0]|0; //@line 204 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $7 = ($6<<24>>24)==(0); //@line 204 "modules/sound/lullaby/Mpg123Decoder.cpp"
  if (!($7)) {
   $$2 = $size$06;
   label = 8;
   break;
  }
  HEAP32[$numbytes>>2] = 0; //@line 206 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $8 = HEAP32[$4>>2]|0; //@line 207 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $9 = HEAP32[$5>>2]|0; //@line 207 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $10 = (($9) + ($size$06)|0); //@line 207 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $11 = (($12) - ($size$06))|0; //@line 207 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $13 = (_mpg123_read($8,$10,$11,$numbytes)|0); //@line 207 "modules/sound/lullaby/Mpg123Decoder.cpp"
  switch ($13|0) {
  case -12:  {
   $size$06$lcssa18 = $size$06;
   label = 5;
   break L4;
   break;
  }
  case 0: case -11: case -10:  {
   break;
  }
  default: {
   $size$2$ph = $size$06;
   break L4;
  }
  }
  $16 = HEAP32[$numbytes>>2]|0; //@line 214 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $17 = (($16) + ($size$06))|0; //@line 214 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $18 = HEAP32[$0>>2]|0; //@line 204 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $19 = ($17|0)<($18|0); //@line 204 "modules/sound/lullaby/Mpg123Decoder.cpp"
  if ($19) {
   $12 = $18;$size$06 = $17;
  } else {
   $$2 = $17;
   label = 8;
   break;
  }
 }
 if ((label|0) == 5) {
  $14 = HEAP32[$numbytes>>2]|0; //@line 217 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $15 = (($14) + ($size$06$lcssa18))|0; //@line 217 "modules/sound/lullaby/Mpg123Decoder.cpp"
  HEAP8[$3>>0] = 1; //@line 218 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $size$2$ph = $15;
 }
 else if ((label|0) == 8) {
  STACKTOP = sp;return ($$2|0); //@line 225 "modules/sound/lullaby/Mpg123Decoder.cpp"
 }
 $$2 = $size$2$ph;
 STACKTOP = sp;return ($$2|0); //@line 225 "modules/sound/lullaby/Mpg123Decoder.cpp"
}
function __ZN4love5sound7lullaby13Mpg123Decoder4seekEf($this,$s) {
 $this = $this|0;
 $s = +$s;
 var $$0 = 0, $0 = 0.0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $s; //@line 229 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $1 = ((($this)) + 28|0); //@line 229 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 229 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $3 = (+($2|0)); //@line 229 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $4 = $0 * $3; //@line 229 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $5 = (~~(($4))); //@line 229 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $6 = ($5|0)<(0); //@line 231 "modules/sound/lullaby/Mpg123Decoder.cpp"
 if ($6) {
  $$0 = 0;
  return ($$0|0); //@line 241 "modules/sound/lullaby/Mpg123Decoder.cpp"
 }
 $7 = ((($this)) + 52|0); //@line 234 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 234 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $9 = (_mpg123_seek($8,$5,0)|0); //@line 234 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $10 = ($9|0)>(-1); //@line 234 "modules/sound/lullaby/Mpg123Decoder.cpp"
 if (!($10)) {
  $$0 = 0;
  return ($$0|0); //@line 241 "modules/sound/lullaby/Mpg123Decoder.cpp"
 }
 $11 = ((($this)) + 36|0); //@line 236 "modules/sound/lullaby/Mpg123Decoder.cpp"
 HEAP8[$11>>0] = 0; //@line 236 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $$0 = 1;
 return ($$0|0); //@line 241 "modules/sound/lullaby/Mpg123Decoder.cpp"
}
function __ZN4love5sound7lullaby13Mpg123Decoder6rewindEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 36|0); //@line 245 "modules/sound/lullaby/Mpg123Decoder.cpp"
 HEAP8[$0>>0] = 0; //@line 245 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $1 = ((($this)) + 52|0); //@line 247 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 247 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $3 = (_mpg123_seek($2,0,0)|0); //@line 247 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $4 = ($3|0)>(-1); //@line 247 "modules/sound/lullaby/Mpg123Decoder.cpp"
 return ($4|0); //@line 251 "modules/sound/lullaby/Mpg123Decoder.cpp"
}
function __ZN4love5sound7lullaby13Mpg123Decoder10isSeekableEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1; //@line 255 "modules/sound/lullaby/Mpg123Decoder.cpp"
}
function __ZNK4love5sound7lullaby13Mpg123Decoder11getChannelsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 56|0); //@line 260 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 260 "modules/sound/lullaby/Mpg123Decoder.cpp"
 return ($1|0); //@line 260 "modules/sound/lullaby/Mpg123Decoder.cpp"
}
function __ZNK4love5sound7lullaby13Mpg123Decoder11getBitDepthEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 16; //@line 265 "modules/sound/lullaby/Mpg123Decoder.cpp"
}
function __ZN4love5sound7lullaby13Mpg123Decoder11getDurationEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0, $storemerge = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 64|0); //@line 271 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $1 = +HEAPF64[$0>>3]; //@line 271 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $2 = $1 == -2.0; //@line 271 "modules/sound/lullaby/Mpg123Decoder.cpp"
 if (!($2)) {
  $13 = $1;
  return (+$13); //@line 283 "modules/sound/lullaby/Mpg123Decoder.cpp"
 }
 $3 = ((($this)) + 52|0); //@line 273 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 273 "modules/sound/lullaby/Mpg123Decoder.cpp"
 (_mpg123_scan($4)|0); //@line 273 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $5 = HEAP32[$3>>2]|0; //@line 275 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $6 = (_mpg123_length($5)|0); //@line 275 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $7 = ($6|0)<(0); //@line 277 "modules/sound/lullaby/Mpg123Decoder.cpp"
 if ($7) {
  $storemerge = -1.0;
 } else {
  $8 = (+($6|0)); //@line 280 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $9 = ((($this)) + 28|0); //@line 280 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $10 = HEAP32[$9>>2]|0; //@line 280 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $11 = (+($10|0)); //@line 280 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $12 = $8 / $11; //@line 280 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $storemerge = $12;
 }
 HEAPF64[$0>>3] = $storemerge; //@line 278 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $13 = $storemerge;
 return (+$13); //@line 283 "modules/sound/lullaby/Mpg123Decoder.cpp"
}
function __ZN4love5sound7lullaby13Mpg123Decoder7acceptsERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($ext) {
 $ext = $ext|0;
 var $$index = 0, $$index2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[624056]|0; //@line 175 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $1 = ($0<<24>>24)==(0); //@line 175 "modules/sound/lullaby/Mpg123Decoder.cpp"
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((624056|0))|0); //@line 175 "modules/sound/lullaby/Mpg123Decoder.cpp"
   $3 = ($2|0)==(0); //@line 175 "modules/sound/lullaby/Mpg123Decoder.cpp"
   if (!($3)) {
    __THREW__ = 0;
    invoke_viii(3460,(632356|0),(364677|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $4 = __THREW__; __THREW__ = 0;
    $5 = $4&1;
    if ($5) {
     $6 = ___cxa_find_matching_catch_2()|0;
     $7 = tempRet0;
     $$index = $6;$$index2 = $7;
     ___cxa_guard_abort((624056|0)); //@line 175 "modules/sound/lullaby/Mpg123Decoder.cpp"
     ___resumeException($$index|0);
     // unreachable;
    }
    __THREW__ = 0;
    invoke_viii(3460,((632368)|0),(1282466|0),0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $8 = __THREW__; __THREW__ = 0;
    $9 = $8&1;
    if (!($9)) {
     (___cxa_atexit((3897|0),(0|0),(___dso_handle|0))|0); //@line 175 "modules/sound/lullaby/Mpg123Decoder.cpp"
     ___cxa_guard_release((624056|0)); //@line 175 "modules/sound/lullaby/Mpg123Decoder.cpp"
     break;
    }
    $32 = ___cxa_find_matching_catch_2()|0;
    $33 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev(632356); //@line 176 "modules/sound/lullaby/Mpg123Decoder.cpp"
    $$index = $32;$$index2 = $33;
    ___cxa_guard_abort((624056|0)); //@line 175 "modules/sound/lullaby/Mpg123Decoder.cpp"
    ___resumeException($$index|0);
    // unreachable;
   }
  }
 } while(0);
 $10 = HEAP8[$ext>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $10 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = ($11<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ((($ext)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = HEAP32[$13>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $15 = $10&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $16 = $15 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $17 = $12 ? $16 : $14; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $18 = ((($ext)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $19 = HEAP32[$18>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $20 = ((($ext)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $21 = $12 ? $20 : $19; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $i$0 = 0;
 while(1) {
  $22 = (632356 + (($i$0*12)|0)|0); //@line 180 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $23 = HEAP8[$22>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $24 = $23 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $25 = ($24<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($25) {
   $28 = $23&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $29 = $28 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $31 = $29;
  } else {
   $26 = (((632356 + (($i$0*12)|0)|0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $27 = HEAP32[$26>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $31 = $27;
  }
  $30 = ($31|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($30) {
   $50 = 0;
   label = 20;
   break;
  }
  if ($25) {
   $38 = $23&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $39 = $38 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $40 = ((($22)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $42 = $39;$44 = $40;
  } else {
   $34 = (((632356 + (($i$0*12)|0)|0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $35 = HEAP32[$34>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $36 = (((632356 + (($i$0*12)|0)|0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $37 = HEAP32[$36>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $42 = $35;$44 = $37;
  }
  $41 = ($17>>>0)<($42>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $43 = $41 ? $17 : $42; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $45 = (_memcmp($44,$21,$43)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $46 = ($45|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($46) {
   $47 = ($42>>>0)>=($17>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $48 = $41 ^ 1; //@line 182 "modules/sound/lullaby/Mpg123Decoder.cpp"
   $or$cond = $47 & $48;
   if ($or$cond) {
    $50 = 1;
    label = 20;
    break;
   }
  }
  $49 = (($i$0) + 1)|0; //@line 180 "modules/sound/lullaby/Mpg123Decoder.cpp"
  $i$0 = $49;
 }
 if ((label|0) == 20) {
  return ($50|0); //@line 187 "modules/sound/lullaby/Mpg123Decoder.cpp"
 }
 return (0)|0;
}
function ___cxx_global_array_dtor_3008($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((632368)); //@line 177 "modules/sound/lullaby/Mpg123Decoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev(632356); //@line 177 "modules/sound/lullaby/Mpg123Decoder.cpp"
 return; //@line 177 "modules/sound/lullaby/Mpg123Decoder.cpp"
}
function __ZN4love5sound7lullaby13Mpg123Decoder4quitEv() {
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[1251894]|0; //@line 191 "modules/sound/lullaby/Mpg123Decoder.cpp"
 $1 = ($0<<24>>24)==(0); //@line 191 "modules/sound/lullaby/Mpg123Decoder.cpp"
 if (!($1)) {
 }
 return; //@line 193 "modules/sound/lullaby/Mpg123Decoder.cpp"
}
function __ZN4love5delayEj($ms) {
 $ms = $ms|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _SDL_Delay($ms); //@line 30 "common/delay.cpp"
 return; //@line 31 "common/delay.cpp"
}
function __GLOBAL__sub_I_Memoizer_cpp() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(632384)>>2] = 0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[(632388)>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[158095] = (632384); //@line 1130 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 (___cxa_atexit((3898|0),(632380|0),(___dso_handle|0))|0); //@line 27 "common/Memoizer.cpp"
 return;
}
function __ZNSt3__13mapIPvS1_NS_4lessIS1_EENS_9allocatorINS_4pairIKS1_S1_EEEEED2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeIPvS2_EENS_19__map_value_compareIS2_S3_NS_4lessIS2_EELb1EEENS_9allocatorIS3_EEE7destroyEPNS_11__tree_nodeIS3_S2_EE($this,$1); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 return; //@line 783 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
}
function __ZNSt3__16__treeINS_12__value_typeIPvS2_EENS_19__map_value_compareIS2_S3_NS_4lessIS2_EELb1EEENS_9allocatorIS3_EEE7destroyEPNS_11__tree_nodeIS3_S2_EE($this,$__nd) {
 $this = $this|0;
 $__nd = $__nd|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__nd|0)==(0|0); //@line 1440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($0) {
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 } else {
  $1 = HEAP32[$__nd>>2]|0; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIPvS2_EENS_19__map_value_compareIS2_S3_NS_4lessIS2_EELb1EEENS_9allocatorIS3_EEE7destroyEPNS_11__tree_nodeIS3_S2_EE($this,$1); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2 = ((($__nd)) + 4|0); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3 = HEAP32[$2>>2]|0; //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIPvS2_EENS_19__map_value_compareIS2_S3_NS_4lessIS2_EELb1EEENS_9allocatorIS3_EEE7destroyEPNS_11__tree_nodeIS3_S2_EE($this,$3); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZdlPv($__nd); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
}
function __ZN4love8Memoizer3addEPvS1_($key,$val) {
 $key = $key|0;
 $val = $val|0;
 var $$0$i7$i = 0, $$lcssa = 0, $$lcssa18 = 0, $$pre$i$i = 0, $$sink$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__nd$0$i$i = 0, $__nd$0$i$i$lcssa = 0, $__nd$0$i$i$lcssa16 = 0, $__nd$0$i$i$lcssa17 = 0;
 var $__parent$i = 0, $__r$0$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__parent$i = sp;
 $0 = $key;
 $1 = HEAP32[(632384)>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $2 = ($1|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2) {
   HEAP32[$__parent$i>>2] = (632384); //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i = (632384);$21 = ((632384));
  } else {
   $__nd$0$i$i = $1;
   while(1) {
    $3 = ((($__nd$0$i$i)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4 = HEAP32[$3>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5 = ($4>>>0)>($key>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5) {
     $6 = HEAP32[$__nd$0$i$i>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $7 = ($6|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($7) {
      $$lcssa18 = $__nd$0$i$i;$__nd$0$i$i$lcssa17 = $__nd$0$i$i;
      label = 4;
      break;
     } else {
      $$sink$i$i = $6;
     }
    } else {
     $9 = ($4>>>0)<($key>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($9)) {
      $__nd$0$i$i$lcssa = $__nd$0$i$i;
      label = 10;
      break;
     }
     $10 = ((($__nd$0$i$i)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $11 = HEAP32[$10>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $12 = ($11|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($12) {
      $$lcssa = $10;$__nd$0$i$i$lcssa16 = $__nd$0$i$i;
      label = 8;
      break;
     } else {
      $$sink$i$i = $11;
     }
    }
    $__nd$0$i$i = $$sink$i$i;
   }
   if ((label|0) == 4) {
    HEAP32[$__parent$i>>2] = $__nd$0$i$i$lcssa17; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $8 = $__nd$0$i$i$lcssa17; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i = $$lcssa18;$21 = $8;
    break;
   }
   else if ((label|0) == 8) {
    HEAP32[$__parent$i>>2] = $__nd$0$i$i$lcssa16; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $13 = $__nd$0$i$i$lcssa16; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i = $$lcssa;$21 = $13;
    break;
   }
   else if ((label|0) == 10) {
    HEAP32[$__parent$i>>2] = $__nd$0$i$i$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $14 = $__nd$0$i$i$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $15 = ($__nd$0$i$i$lcssa|0)==(0|0); //@line 1378 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($15) {
     $$0$i7$i = $__parent$i;$21 = $14;
     break;
    } else {
     $__r$0$i = $__nd$0$i$i$lcssa;
    }
    $30 = ((($__r$0$i)) + 20|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    HEAP32[$30>>2] = $val; //@line 31 "common/Memoizer.cpp"
    STACKTOP = sp;return; //@line 32 "common/Memoizer.cpp"
   }
  }
 } while(0);
 $16 = (__Znwj(24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $17 = ((($16)) + 16|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$17>>2] = $0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $18 = ((($16)) + 20|0); //@line 1366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$18>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$16>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $19 = ((($16)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$19>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $20 = ((($16)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$20>>2] = $21; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$$0$i7$i>>2] = $16; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $22 = HEAP32[158095]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $23 = HEAP32[$22>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $24 = ($23|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($24) {
  $27 = $16;
 } else {
  $25 = $23; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[158095] = $25; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre$i$i = HEAP32[$$0$i7$i>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $27 = $$pre$i$i;
 }
 $26 = HEAP32[(632384)>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($26,$27); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $28 = HEAP32[(632388)>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $29 = (($28) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[(632388)>>2] = $29; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $__r$0$i = $16;
 $30 = ((($__r$0$i)) + 20|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$30>>2] = $val; //@line 31 "common/Memoizer.cpp"
 STACKTOP = sp;return; //@line 32 "common/Memoizer.cpp"
}
function __ZN4love8Memoizer6removeEPv($key) {
 $key = $key|0;
 var $$0$i$i$i$i$i = 0, $$0$i$i$i$i$i$i = 0, $$0$ph$lcssa$i$i$i$i = 0, $$0$ph6$i$i$i$i = 0, $$01$i$i$i$i$i = 0, $$01$ph5$i$i$i$i = 0, $$013$i$i$i$i = 0, $$013$i$i$i$i$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(632384)>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $1 = ($0|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($1) {
  return; //@line 37 "common/Memoizer.cpp"
 } else {
  $$0$ph6$i$i$i$i = (632384);$$01$ph5$i$i$i$i = $0;
 }
 L3: while(1) {
  $$013$i$i$i$i = $$01$ph5$i$i$i$i;
  while(1) {
   $2 = ((($$013$i$i$i$i)) + 16|0); //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3 = HEAP32[$2>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   $4 = ($3>>>0)<($key>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   if (!($4)) {
    $$013$i$i$i$i$lcssa = $$013$i$i$i$i;
    break;
   }
   $7 = ((($$013$i$i$i$i)) + 4|0); //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $8 = HEAP32[$7>>2]|0; //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $9 = ($8|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($9) {
    $$0$ph$lcssa$i$i$i$i = $$0$ph6$i$i$i$i;
    break L3;
   } else {
    $$013$i$i$i$i = $8;
   }
  }
  $5 = HEAP32[$$013$i$i$i$i$lcssa>>2]|0; //@line 2105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6 = ($5|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($6) {
   $$0$ph$lcssa$i$i$i$i = $$013$i$i$i$i$lcssa;
   break;
  } else {
   $$0$ph6$i$i$i$i = $$013$i$i$i$i$lcssa;$$01$ph5$i$i$i$i = $5;
  }
 }
 $10 = ($$0$ph$lcssa$i$i$i$i|0)==((632384)|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($10) {
  return; //@line 37 "common/Memoizer.cpp"
 }
 $11 = ((($$0$ph$lcssa$i$i$i$i)) + 16|0); //@line 463 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $12 = HEAP32[$11>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
 $13 = ($12>>>0)>($key>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
 if ($13) {
  return; //@line 37 "common/Memoizer.cpp"
 }
 $14 = ((($$0$ph$lcssa$i$i$i$i)) + 4|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $15 = HEAP32[$14>>2]|0; //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $16 = ($15|0)==(0|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($16) {
  $$0$i$i$i$i$i = $$0$ph$lcssa$i$i$i$i;
  while(1) {
   $19 = ((($$0$i$i$i$i$i)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $20 = HEAP32[$19>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $21 = HEAP32[$20>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $22 = ($21|0)==($$0$i$i$i$i$i|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($22) {
    $$01$i$i$i$i$i = $20;
    break;
   } else {
    $$0$i$i$i$i$i = $20;
   }
  }
 } else {
  $$0$i$i$i$i$i$i = $15;
  while(1) {
   $17 = HEAP32[$$0$i$i$i$i$i$i>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $18 = ($17|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($18) {
    $$01$i$i$i$i$i = $$0$i$i$i$i$i$i;
    break;
   } else {
    $$0$i$i$i$i$i$i = $17;
   }
  }
 }
 $23 = HEAP32[158095]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $24 = ($23|0)==($$0$ph$lcssa$i$i$i$i|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($24) {
  $25 = $$01$i$i$i$i$i; //@line 656 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[158095] = $25; //@line 1980 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
 $26 = HEAP32[(632388)>>2]|0; //@line 1981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $27 = (($26) + -1)|0; //@line 1981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[(632388)>>2] = $27; //@line 1981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__113__tree_removeIPNS_16__tree_node_baseIPvEEEEvT_S5_($0,$$0$ph$lcssa$i$i$i$i); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZdlPv($$0$ph$lcssa$i$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 37 "common/Memoizer.cpp"
}
function __ZN4love8Memoizer4findEPv($key) {
 $key = $key|0;
 var $$0 = 0, $$0$ph$lcssa$i$i$i = 0, $$0$ph6$i$i$i = 0, $$01$ph5$i$i$i = 0, $$013$i$i$i = 0, $$013$i$i$i$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(632384)>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $1 = ($0|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($1) {
  $$0 = 0;
  return ($$0|0); //@line 47 "common/Memoizer.cpp"
 } else {
  $$0$ph6$i$i$i = (632384);$$01$ph5$i$i$i = $0;
 }
 L3: while(1) {
  $$013$i$i$i = $$01$ph5$i$i$i;
  while(1) {
   $2 = ((($$013$i$i$i)) + 16|0); //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3 = HEAP32[$2>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   $4 = ($3>>>0)<($key>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   if (!($4)) {
    $$013$i$i$i$lcssa = $$013$i$i$i;
    break;
   }
   $7 = ((($$013$i$i$i)) + 4|0); //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $8 = HEAP32[$7>>2]|0; //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $9 = ($8|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($9) {
    $$0$ph$lcssa$i$i$i = $$0$ph6$i$i$i;
    break L3;
   } else {
    $$013$i$i$i = $8;
   }
  }
  $5 = HEAP32[$$013$i$i$i$lcssa>>2]|0; //@line 2105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6 = ($5|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($6) {
   $$0$ph$lcssa$i$i$i = $$013$i$i$i$lcssa;
   break;
  } else {
   $$0$ph6$i$i$i = $$013$i$i$i$lcssa;$$01$ph5$i$i$i = $5;
  }
 }
 $10 = ($$0$ph$lcssa$i$i$i|0)==((632384)|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($10) {
  $$0 = 0;
  return ($$0|0); //@line 47 "common/Memoizer.cpp"
 }
 $11 = ((($$0$ph$lcssa$i$i$i)) + 16|0); //@line 463 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $12 = HEAP32[$11>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
 $13 = ($12>>>0)>($key>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
 if ($13) {
  $$0 = 0;
  return ($$0|0); //@line 47 "common/Memoizer.cpp"
 }
 $14 = ((($$0$ph$lcssa$i$i$i)) + 20|0); //@line 44 "common/Memoizer.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 44 "common/Memoizer.cpp"
 $$0 = $15;
 return ($$0|0); //@line 47 "common/Memoizer.cpp"
}
function __ZN4love9ReferenceD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (21980); //@line 42 "common/Reference.cpp"
 $0 = ((($this)) + 8|0); //@line 58 "common/Reference.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 58 "common/Reference.cpp"
 $2 = ($1|0)==(-1); //@line 58 "common/Reference.cpp"
 if ($2) {
  return; //@line 44 "common/Reference.cpp"
 }
 $3 = ((($this)) + 4|0); //@line 63 "common/Reference.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 63 "common/Reference.cpp"
 __THREW__ = 0;
 (invoke_iiii(3899,($4|0),-10000,(364699|0))|0); //@line 63 "common/Reference.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  $7 = HEAP32[$3>>2]|0; //@line 64 "common/Reference.cpp"
  $8 = HEAP32[$0>>2]|0; //@line 64 "common/Reference.cpp"
  __THREW__ = 0;
  invoke_viii(3900,($7|0),-1,($8|0)); //@line 64 "common/Reference.cpp"
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if (!($10)) {
   $11 = HEAP32[$3>>2]|0; //@line 65 "common/Reference.cpp"
   __THREW__ = 0;
   invoke_vii(3384,($11|0),-2); //@line 65 "common/Reference.cpp"
   $12 = __THREW__; __THREW__ = 0;
   $13 = $12&1;
   if (!($13)) {
    HEAP32[$0>>2] = -1; //@line 66 "common/Reference.cpp"
    return; //@line 44 "common/Reference.cpp"
   }
  }
 }
 $14 = ___cxa_find_matching_catch_3(0|0)|0;
 $15 = tempRet0;
 ___clang_call_terminate($14); //@line 43 "common/Reference.cpp"
 // unreachable; //@line 43 "common/Reference.cpp"
}
function __ZN4love9ReferenceD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (21980); //@line 42 "common/Reference.cpp"
 $0 = ((($this)) + 8|0); //@line 58 "common/Reference.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 58 "common/Reference.cpp"
 $2 = ($1|0)==(-1); //@line 58 "common/Reference.cpp"
 if ($2) {
  __ZdlPv($this); //@line 42 "common/Reference.cpp"
  return; //@line 44 "common/Reference.cpp"
 }
 $3 = ((($this)) + 4|0); //@line 63 "common/Reference.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 63 "common/Reference.cpp"
 __THREW__ = 0;
 (invoke_iiii(3899,($4|0),-10000,(364699|0))|0); //@line 63 "common/Reference.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  $7 = HEAP32[$3>>2]|0; //@line 64 "common/Reference.cpp"
  $8 = HEAP32[$0>>2]|0; //@line 64 "common/Reference.cpp"
  __THREW__ = 0;
  invoke_viii(3900,($7|0),-1,($8|0)); //@line 64 "common/Reference.cpp"
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if (!($10)) {
   $11 = HEAP32[$3>>2]|0; //@line 65 "common/Reference.cpp"
   __THREW__ = 0;
   invoke_vii(3384,($11|0),-2); //@line 65 "common/Reference.cpp"
   $12 = __THREW__; __THREW__ = 0;
   $13 = $12&1;
   if (!($13)) {
    __ZdlPv($this); //@line 42 "common/Reference.cpp"
    return; //@line 44 "common/Reference.cpp"
   }
  }
 }
 $14 = ___cxa_find_matching_catch_3(0|0)|0;
 $15 = tempRet0;
 ___clang_call_terminate($14); //@line 43 "common/Reference.cpp"
 // unreachable; //@line 43 "common/Reference.cpp"
}
function __ZN4love9ReferenceC2EP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (21980); //@line 37 "common/Reference.cpp"
 $0 = ((($this)) + 4|0); //@line 35 "common/Reference.cpp"
 HEAP32[$0>>2] = 0; //@line 35 "common/Reference.cpp"
 $1 = ((($this)) + 8|0); //@line 36 "common/Reference.cpp"
 HEAP32[$1>>2] = -1; //@line 36 "common/Reference.cpp"
 __ZN4love9Reference3refEP9lua_State($this,$L); //@line 38 "common/Reference.cpp"
 return; //@line 39 "common/Reference.cpp"
}
function __ZN4love9Reference3refEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 58 "common/Reference.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 58 "common/Reference.cpp"
 $2 = ($1|0)==(-1); //@line 58 "common/Reference.cpp"
 $$pre = ((($this)) + 4|0); //@line 49 "common/Reference.cpp"
 if (!($2)) {
  $3 = HEAP32[$$pre>>2]|0; //@line 63 "common/Reference.cpp"
  (__ZN4love11luax_insistEP9lua_StateiPKc($3,-10000,364699)|0); //@line 63 "common/Reference.cpp"
  $4 = HEAP32[$$pre>>2]|0; //@line 64 "common/Reference.cpp"
  $5 = HEAP32[$0>>2]|0; //@line 64 "common/Reference.cpp"
  _luaL_unref($4,-1,$5); //@line 64 "common/Reference.cpp"
  $6 = HEAP32[$$pre>>2]|0; //@line 65 "common/Reference.cpp"
  _lua_settop($6,-2); //@line 65 "common/Reference.cpp"
  HEAP32[$0>>2] = -1; //@line 66 "common/Reference.cpp"
 }
 $7 = (__ZN4love20luax_getpinnedthreadEP9lua_State($L)|0); //@line 49 "common/Reference.cpp"
 HEAP32[$$pre>>2] = $7; //@line 49 "common/Reference.cpp"
 (__ZN4love11luax_insistEP9lua_StateiPKc($L,-10000,364699)|0); //@line 50 "common/Reference.cpp"
 _lua_insert($L,-2); //@line 51 "common/Reference.cpp"
 $8 = (_luaL_ref($L,-2)|0); //@line 52 "common/Reference.cpp"
 HEAP32[$0>>2] = $8; //@line 52 "common/Reference.cpp"
 _lua_settop($L,-2); //@line 53 "common/Reference.cpp"
 return; //@line 54 "common/Reference.cpp"
}
function __ZN4love9Reference4pushEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 72 "common/Reference.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 72 "common/Reference.cpp"
 $2 = ($1|0)==(-1); //@line 72 "common/Reference.cpp"
 if ($2) {
  _lua_pushnil($L); //@line 79 "common/Reference.cpp"
  return; //@line 80 "common/Reference.cpp"
 } else {
  (__ZN4love11luax_insistEP9lua_StateiPKc($L,-10000,364699)|0); //@line 74 "common/Reference.cpp"
  $3 = HEAP32[$0>>2]|0; //@line 75 "common/Reference.cpp"
  _lua_rawgeti($L,-1,$3); //@line 75 "common/Reference.cpp"
  _lua_remove($L,-2); //@line 76 "common/Reference.cpp"
  return; //@line 80 "common/Reference.cpp"
 }
}
function __ZN4love7VariantD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $lpad$phi$index = 0, $scevgep4$i$i$i1$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (21996); //@line 117 "common/Variant.cpp"
 $0 = ((($this)) + 8|0); //@line 118 "common/Variant.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 118 "common/Variant.cpp"
 switch ($1|0) {
 case 4:  {
  $2 = ((($this)) + 16|0); //@line 121 "common/Variant.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 121 "common/Variant.cpp"
  $4 = ($3|0)==(0|0); //@line 121 "common/Variant.cpp"
  if ($4) {
   return; //@line 131 "common/Variant.cpp"
  }
  __ZdaPv($3); //@line 121 "common/Variant.cpp"
  return; //@line 131 "common/Variant.cpp"
  break;
 }
 case 6:  {
  $5 = ((($this)) + 16|0); //@line 124 "common/Variant.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 124 "common/Variant.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 124 "common/Variant.cpp"
  $8 = ((($7)) + 12|0); //@line 124 "common/Variant.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 124 "common/Variant.cpp"
  __THREW__ = 0;
  invoke_vi($9|0,($6|0)); //@line 124 "common/Variant.cpp"
  $10 = __THREW__; __THREW__ = 0;
  $11 = $10&1;
  if (!($11)) {
   return; //@line 131 "common/Variant.cpp"
  }
  $14 = ___cxa_find_matching_catch_3(0|0)|0;
  $15 = tempRet0;
  $lpad$phi$index = $14;
  ___clang_call_terminate($lpad$phi$index); //@line 131 "common/Variant.cpp"
  // unreachable; //@line 131 "common/Variant.cpp"
  break;
 }
 case 8:  {
  $16 = ((($this)) + 16|0); //@line 127 "common/Variant.cpp"
  $17 = HEAP32[$16>>2]|0; //@line 127 "common/Variant.cpp"
  $18 = HEAP32[$17>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = ((($17)) + 4|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = HEAP32[$19>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($18|0)==($20|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  L13: do {
   if ($21) {
    $41 = $18;
   } else {
    $23 = $20;
    while(1) {
     $22 = ((($23)) + -8|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $24 = HEAP32[$22>>2]|0; //@line 48 "common/Variant.cpp"
     $25 = HEAP32[$24>>2]|0; //@line 48 "common/Variant.cpp"
     $26 = ((($25)) + 12|0); //@line 48 "common/Variant.cpp"
     $27 = HEAP32[$26>>2]|0; //@line 48 "common/Variant.cpp"
     __THREW__ = 0;
     invoke_vi($27|0,($24|0)); //@line 48 "common/Variant.cpp"
     $28 = __THREW__; __THREW__ = 0;
     $29 = $28&1;
     if ($29) {
      break;
     }
     $30 = ((($23)) + -4|0); //@line 49 "common/Variant.cpp"
     $31 = HEAP32[$30>>2]|0; //@line 49 "common/Variant.cpp"
     $32 = HEAP32[$31>>2]|0; //@line 49 "common/Variant.cpp"
     $33 = ((($32)) + 12|0); //@line 49 "common/Variant.cpp"
     $34 = HEAP32[$33>>2]|0; //@line 49 "common/Variant.cpp"
     __THREW__ = 0;
     invoke_vi($34|0,($31|0)); //@line 49 "common/Variant.cpp"
     $35 = __THREW__; __THREW__ = 0;
     $36 = $35&1;
     if ($36) {
      break;
     }
     $37 = HEAP32[$19>>2]|0; //@line 1666 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $scevgep4$i$i$i1$i = ((($37)) + -8|0);
     HEAP32[$19>>2] = $scevgep4$i$i$i1$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $38 = HEAP32[$17>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $39 = ($38|0)==($scevgep4$i$i$i1$i|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($39) {
      $41 = $scevgep4$i$i$i1$i;
      break L13;
     } else {
      $23 = $scevgep4$i$i$i1$i;
     }
    }
    $12 = ___cxa_find_matching_catch_3(0|0)|0;
    $13 = tempRet0;
    $lpad$phi$index = $12;
    ___clang_call_terminate($lpad$phi$index); //@line 131 "common/Variant.cpp"
    // unreachable; //@line 131 "common/Variant.cpp"
   }
  } while(0);
  $40 = ($41|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($40)) {
   __ZdlPv($41); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  __ZdlPv($17); //@line 52 "common/Variant.cpp"
  return; //@line 131 "common/Variant.cpp"
  break;
 }
 default: {
  return; //@line 131 "common/Variant.cpp"
 }
 }
}
function __ZN4love7VariantD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7VariantD2Ev($this); //@line 117 "common/Variant.cpp"
 __ZdlPv($this); //@line 117 "common/Variant.cpp"
 return; //@line 131 "common/Variant.cpp"
}
function __ZN4love7VariantC2Eb($this,$boolean) {
 $this = $this|0;
 $boolean = $boolean|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 61 "common/Variant.cpp"
 HEAP32[$this>>2] = (21996); //@line 63 "common/Variant.cpp"
 $0 = ((($this)) + 8|0); //@line 62 "common/Variant.cpp"
 HEAP32[$0>>2] = 1; //@line 62 "common/Variant.cpp"
 $1 = ((($this)) + 16|0); //@line 64 "common/Variant.cpp"
 $2 = $boolean&1; //@line 64 "common/Variant.cpp"
 HEAP8[$1>>0] = $2; //@line 64 "common/Variant.cpp"
 return; //@line 65 "common/Variant.cpp"
}
function __ZN4love7VariantC2Ed($this,$number) {
 $this = $this|0;
 $number = +$number;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 67 "common/Variant.cpp"
 HEAP32[$this>>2] = (21996); //@line 69 "common/Variant.cpp"
 $0 = ((($this)) + 8|0); //@line 68 "common/Variant.cpp"
 HEAP32[$0>>2] = 2; //@line 68 "common/Variant.cpp"
 $1 = ((($this)) + 16|0); //@line 70 "common/Variant.cpp"
 HEAPF64[$1>>3] = $number; //@line 70 "common/Variant.cpp"
 return; //@line 71 "common/Variant.cpp"
}
function __ZN4love7VariantC2EPKcj($this,$string,$len) {
 $this = $this|0;
 $string = $string|0;
 $len = $len|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 73 "common/Variant.cpp"
 HEAP32[$this>>2] = (21996); //@line 75 "common/Variant.cpp"
 $0 = ((($this)) + 8|0); //@line 74 "common/Variant.cpp"
 HEAP32[$0>>2] = 4; //@line 74 "common/Variant.cpp"
 $1 = (($len) + 1)|0; //@line 76 "common/Variant.cpp"
 __THREW__ = 0;
 $2 = (invoke_ii(3492,($1|0))|0); //@line 76 "common/Variant.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  ___resumeException($9|0);
  // unreachable;
 } else {
  $not$ = ($len|0)!=(-1); //@line 78 "common/Variant.cpp"
  $5 = $not$&1; //@line 78 "common/Variant.cpp"
  $6 = (($2) + ($len)|0); //@line 78 "common/Variant.cpp"
  _memset(($6|0),0,($5|0))|0; //@line 78 "common/Variant.cpp"
  _memcpy(($2|0),($string|0),($len|0))|0; //@line 78 "common/Variant.cpp"
  $7 = ((($this)) + 16|0); //@line 79 "common/Variant.cpp"
  HEAP32[$7>>2] = $2; //@line 79 "common/Variant.cpp"
  $8 = ((($7)) + 4|0); //@line 80 "common/Variant.cpp"
  HEAP32[$8>>2] = $len; //@line 80 "common/Variant.cpp"
  return; //@line 81 "common/Variant.cpp"
 }
}
function __ZN4love7VariantC2EPv($this,$userdata) {
 $this = $this|0;
 $userdata = $userdata|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 89 "common/Variant.cpp"
 HEAP32[$this>>2] = (21996); //@line 91 "common/Variant.cpp"
 $0 = ((($this)) + 8|0); //@line 90 "common/Variant.cpp"
 HEAP32[$0>>2] = 5; //@line 90 "common/Variant.cpp"
 $1 = ((($this)) + 16|0); //@line 92 "common/Variant.cpp"
 HEAP32[$1>>2] = $userdata; //@line 92 "common/Variant.cpp"
 return; //@line 93 "common/Variant.cpp"
}
function __ZN4love7VariantC2ENS_4TypeEPv($this,$udatatype,$userdata) {
 $this = $this|0;
 $udatatype = $udatatype|0;
 $userdata = $userdata|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 95 "common/Variant.cpp"
 HEAP32[$this>>2] = (21996); //@line 98 "common/Variant.cpp"
 $0 = ((($this)) + 8|0); //@line 96 "common/Variant.cpp"
 HEAP32[$0>>2] = 6; //@line 96 "common/Variant.cpp"
 $1 = ((($this)) + 24|0); //@line 97 "common/Variant.cpp"
 HEAP32[$1>>2] = $udatatype; //@line 97 "common/Variant.cpp"
 $2 = ($udatatype|0)==(0); //@line 99 "common/Variant.cpp"
 if ($2) {
  $14 = ((($this)) + 16|0); //@line 106 "common/Variant.cpp"
  HEAP32[$14>>2] = $userdata; //@line 106 "common/Variant.cpp"
  return; //@line 107 "common/Variant.cpp"
 }
 $3 = ((($userdata)) + 4|0); //@line 102 "common/Variant.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 102 "common/Variant.cpp"
 $5 = ((($this)) + 16|0); //@line 102 "common/Variant.cpp"
 HEAP32[$5>>2] = $4; //@line 102 "common/Variant.cpp"
 $6 = HEAP32[$3>>2]|0; //@line 103 "common/Variant.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 103 "common/Variant.cpp"
 $8 = ((($7)) + 8|0); //@line 103 "common/Variant.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 103 "common/Variant.cpp"
 __THREW__ = 0;
 invoke_vi($9|0,($6|0)); //@line 103 "common/Variant.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $12 = ___cxa_find_matching_catch_2()|0;
  $13 = tempRet0;
  ___resumeException($12|0);
  // unreachable;
 } else {
  return; //@line 107 "common/Variant.cpp"
 }
}
function __ZN4love7Variant7fromLuaEP9lua_Stateib($L,$n,$allowTables) {
 $L = $L|0;
 $n = $n|0;
 $allowTables = $allowTables|0;
 var $$0$i = 0, $$0$pre$i = 0, $$01 = 0, $$02 = 0, $$03 = 0, $$lcssa33 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body18$index7Z2D = 0, $eh$lpad$body18$indexZ2D = 0;
 var $len = 0, $not$$i = 0, $scevgep4$i$i$i1$i = 0, $t$i = 0, $v$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $t$i = sp + 12|0;
 $len = sp + 8|0;
 $0 = sp;
 $1 = ($n|0)<(0); //@line 138 "common/Variant.cpp"
 if ($1) {
  $2 = (_lua_gettop($L)|0); //@line 139 "common/Variant.cpp"
  $3 = (($n) + 1)|0; //@line 139 "common/Variant.cpp"
  $4 = (($3) + ($2))|0; //@line 139 "common/Variant.cpp"
  $$01 = $4;
 } else {
  $$01 = $n;
 }
 $5 = (_lua_type($L,$$01)|0); //@line 141 "common/Variant.cpp"
 switch ($5|0) {
 case 1:  {
  $6 = (__Znwj(32)|0); //@line 144 "common/Variant.cpp"
  __THREW__ = 0;
  $7 = (invoke_iii(3504,($L|0),($$01|0))|0); //@line 144 "common/Variant.cpp"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if (!($9)) {
   __THREW__ = 0;
   invoke_vi(3431,($6|0)); //@line 61 "common/Variant.cpp"
   $10 = __THREW__; __THREW__ = 0;
   $11 = $10&1;
   if (!($11)) {
    HEAP32[$6>>2] = (21996); //@line 63 "common/Variant.cpp"
    $12 = ((($6)) + 8|0); //@line 62 "common/Variant.cpp"
    HEAP32[$12>>2] = 1; //@line 62 "common/Variant.cpp"
    $13 = ((($6)) + 16|0); //@line 64 "common/Variant.cpp"
    $14 = $7&1; //@line 64 "common/Variant.cpp"
    HEAP8[$13>>0] = $14; //@line 64 "common/Variant.cpp"
    $v$1 = $6;
    STACKTOP = sp;return ($v$1|0); //@line 199 "common/Variant.cpp"
   }
  }
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  __ZdlPv($6); //@line 144 "common/Variant.cpp"
  $$02 = $16;$$03 = $15;
  ___resumeException($$03|0);
  // unreachable;
  break;
 }
 case 3:  {
  $17 = (__Znwj(32)|0); //@line 147 "common/Variant.cpp"
  __THREW__ = 0;
  $18 = (+invoke_dii(3901,($L|0),($$01|0))); //@line 147 "common/Variant.cpp"
  $19 = __THREW__; __THREW__ = 0;
  $20 = $19&1;
  if (!($20)) {
   __THREW__ = 0;
   invoke_vi(3431,($17|0)); //@line 67 "common/Variant.cpp"
   $21 = __THREW__; __THREW__ = 0;
   $22 = $21&1;
   if (!($22)) {
    HEAP32[$17>>2] = (21996); //@line 69 "common/Variant.cpp"
    $23 = ((($17)) + 8|0); //@line 68 "common/Variant.cpp"
    HEAP32[$23>>2] = 2; //@line 68 "common/Variant.cpp"
    $24 = ((($17)) + 16|0); //@line 70 "common/Variant.cpp"
    HEAPF64[$24>>3] = $18; //@line 70 "common/Variant.cpp"
    $v$1 = $17;
    STACKTOP = sp;return ($v$1|0); //@line 199 "common/Variant.cpp"
   }
  }
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  __ZdlPv($17); //@line 147 "common/Variant.cpp"
  $$02 = $26;$$03 = $25;
  ___resumeException($$03|0);
  // unreachable;
  break;
 }
 case 4:  {
  $27 = (_lua_tolstring($L,$$01,$len)|0); //@line 150 "common/Variant.cpp"
  $28 = (__Znwj(32)|0); //@line 151 "common/Variant.cpp"
  $29 = HEAP32[$len>>2]|0; //@line 151 "common/Variant.cpp"
  __THREW__ = 0;
  invoke_vi(3431,($28|0)); //@line 73 "common/Variant.cpp"
  $30 = __THREW__; __THREW__ = 0;
  $31 = $30&1;
  do {
   if ($31) {
    $43 = ___cxa_find_matching_catch_2()|0;
    $44 = tempRet0;
    $eh$lpad$body$index2Z2D = $44;$eh$lpad$body$indexZ2D = $43;
   } else {
    HEAP32[$28>>2] = (21996); //@line 75 "common/Variant.cpp"
    $32 = ((($28)) + 8|0); //@line 74 "common/Variant.cpp"
    HEAP32[$32>>2] = 4; //@line 74 "common/Variant.cpp"
    $33 = (($29) + 1)|0; //@line 76 "common/Variant.cpp"
    __THREW__ = 0;
    $34 = (invoke_ii(3492,($33|0))|0); //@line 76 "common/Variant.cpp"
    $35 = __THREW__; __THREW__ = 0;
    $36 = $35&1;
    if ($36) {
     $37 = ___cxa_find_matching_catch_2()|0;
     $38 = tempRet0;
     $eh$lpad$body$index2Z2D = $38;$eh$lpad$body$indexZ2D = $37;
     break;
    }
    $not$$i = ($29|0)!=(-1); //@line 78 "common/Variant.cpp"
    $39 = $not$$i&1; //@line 78 "common/Variant.cpp"
    $40 = (($34) + ($29)|0); //@line 78 "common/Variant.cpp"
    _memset(($40|0),0,($39|0))|0; //@line 78 "common/Variant.cpp"
    _memcpy(($34|0),($27|0),($29|0))|0; //@line 78 "common/Variant.cpp"
    $41 = ((($28)) + 16|0); //@line 79 "common/Variant.cpp"
    HEAP32[$41>>2] = $34; //@line 79 "common/Variant.cpp"
    $42 = ((($28)) + 20|0); //@line 80 "common/Variant.cpp"
    HEAP32[$42>>2] = $29; //@line 80 "common/Variant.cpp"
    $v$1 = $28;
    STACKTOP = sp;return ($v$1|0); //@line 199 "common/Variant.cpp"
   }
  } while(0);
  __ZdlPv($28); //@line 151 "common/Variant.cpp"
  $$02 = $eh$lpad$body$index2Z2D;$$03 = $eh$lpad$body$indexZ2D;
  ___resumeException($$03|0);
  // unreachable;
  break;
 }
 case 2:  {
  $45 = (__Znwj(32)|0); //@line 154 "common/Variant.cpp"
  __THREW__ = 0;
  $46 = (invoke_iii(3499,($L|0),($$01|0))|0); //@line 154 "common/Variant.cpp"
  $47 = __THREW__; __THREW__ = 0;
  $48 = $47&1;
  if (!($48)) {
   __THREW__ = 0;
   invoke_vi(3431,($45|0)); //@line 89 "common/Variant.cpp"
   $49 = __THREW__; __THREW__ = 0;
   $50 = $49&1;
   if (!($50)) {
    HEAP32[$45>>2] = (21996); //@line 91 "common/Variant.cpp"
    $51 = ((($45)) + 8|0); //@line 90 "common/Variant.cpp"
    HEAP32[$51>>2] = 5; //@line 90 "common/Variant.cpp"
    $52 = ((($45)) + 16|0); //@line 92 "common/Variant.cpp"
    HEAP32[$52>>2] = $46; //@line 92 "common/Variant.cpp"
    $v$1 = $45;
    STACKTOP = sp;return ($v$1|0); //@line 199 "common/Variant.cpp"
   }
  }
  $53 = ___cxa_find_matching_catch_2()|0;
  $54 = tempRet0;
  __ZdlPv($45); //@line 154 "common/Variant.cpp"
  $$02 = $54;$$03 = $53;
  ___resumeException($$03|0);
  // unreachable;
  break;
 }
 case 7:  {
  $55 = (__Znwj(32)|0); //@line 157 "common/Variant.cpp"
  HEAP32[$t$i>>2] = 0; //@line 29 "common/Variant.cpp"
  __THREW__ = 0;
  $56 = (invoke_iii(3902,($L|0),($$01|0))|0); //@line 30 "common/Variant.cpp"
  $57 = __THREW__; __THREW__ = 0;
  $58 = $57&1;
  L37: do {
   if ($58) {
    label = 40;
   } else {
    $59 = ($56|0)==(0); //@line 30 "common/Variant.cpp"
    do {
     if ($59) {
      $$0$i = 0;
     } else {
      __THREW__ = 0;
      $60 = (invoke_iiii(3903,($L|0),($$01|0),(521785|0))|0); //@line 32 "common/Variant.cpp"
      $61 = __THREW__; __THREW__ = 0;
      $62 = $61&1;
      if ($62) {
       label = 40;
       break L37;
      }
      $63 = ($60|0)==(0); //@line 32 "common/Variant.cpp"
      if ($63) {
       $$0$i = 0;
      } else {
       __THREW__ = 0;
       invoke_vii(3454,($L|0),($$01|0)); //@line 34 "common/Variant.cpp"
       $64 = __THREW__; __THREW__ = 0;
       $65 = $64&1;
       if ($65) {
        label = 40;
        break L37;
       }
       __THREW__ = 0;
       $66 = (invoke_iiiii(3552,($L|0),1,1,0)|0); //@line 35 "common/Variant.cpp"
       $67 = __THREW__; __THREW__ = 0;
       $68 = $67&1;
       if ($68) {
        label = 40;
        break L37;
       }
       $69 = ($66|0)==(0); //@line 36 "common/Variant.cpp"
       if ($69) {
        __THREW__ = 0;
        $70 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 37 "common/Variant.cpp"
        $71 = __THREW__; __THREW__ = 0;
        $72 = $71&1;
        if ($72) {
         label = 40;
         break L37;
        }
        __THREW__ = 0;
        (invoke_iii(3904,($70|0),($t$i|0))|0); //@line 37 "common/Variant.cpp"
        $73 = __THREW__; __THREW__ = 0;
        $74 = $73&1;
        if ($74) {
         label = 40;
         break L37;
        }
       } else {
        $75 = $66 | 2; //@line 38 "common/Variant.cpp"
        $76 = ($75|0)==(2); //@line 38 "common/Variant.cpp"
        if (!($76)) {
         $$0$i = 0;
         break;
        }
       }
       __THREW__ = 0;
       invoke_vii(3384,($L|0),-2); //@line 39 "common/Variant.cpp"
       $77 = __THREW__; __THREW__ = 0;
       $78 = $77&1;
       if ($78) {
        label = 40;
        break L37;
       }
       $$0$pre$i = HEAP32[$t$i>>2]|0; //@line 33 "common/Variant.cpp"
       $$0$i = $$0$pre$i;
      }
     }
    } while(0);
    __THREW__ = 0;
    $79 = (invoke_iii(3499,($L|0),($$01|0))|0); //@line 157 "common/Variant.cpp"
    $80 = __THREW__; __THREW__ = 0;
    $81 = $80&1;
    if ($81) {
     label = 40;
    } else {
     __THREW__ = 0;
     invoke_vi(3431,($55|0)); //@line 95 "common/Variant.cpp"
     $82 = __THREW__; __THREW__ = 0;
     $83 = $82&1;
     if ($83) {
      label = 40;
     } else {
      HEAP32[$55>>2] = (21996); //@line 98 "common/Variant.cpp"
      $84 = ((($55)) + 8|0); //@line 96 "common/Variant.cpp"
      HEAP32[$84>>2] = 6; //@line 96 "common/Variant.cpp"
      $85 = ((($55)) + 24|0); //@line 97 "common/Variant.cpp"
      HEAP32[$85>>2] = $$0$i; //@line 97 "common/Variant.cpp"
      $86 = ($$0$i|0)==(0); //@line 99 "common/Variant.cpp"
      if ($86) {
       $98 = ((($55)) + 16|0); //@line 106 "common/Variant.cpp"
       HEAP32[$98>>2] = $79; //@line 106 "common/Variant.cpp"
       $v$1 = $55;
       STACKTOP = sp;return ($v$1|0); //@line 199 "common/Variant.cpp"
      }
      $87 = ((($79)) + 4|0); //@line 102 "common/Variant.cpp"
      $88 = HEAP32[$87>>2]|0; //@line 102 "common/Variant.cpp"
      $89 = ((($55)) + 16|0); //@line 102 "common/Variant.cpp"
      HEAP32[$89>>2] = $88; //@line 102 "common/Variant.cpp"
      $90 = HEAP32[$87>>2]|0; //@line 103 "common/Variant.cpp"
      $91 = HEAP32[$90>>2]|0; //@line 103 "common/Variant.cpp"
      $92 = ((($91)) + 8|0); //@line 103 "common/Variant.cpp"
      $93 = HEAP32[$92>>2]|0; //@line 103 "common/Variant.cpp"
      __THREW__ = 0;
      invoke_vi($93|0,($90|0)); //@line 103 "common/Variant.cpp"
      $94 = __THREW__; __THREW__ = 0;
      $95 = $94&1;
      if ($95) {
       $96 = ___cxa_find_matching_catch_2()|0;
       $97 = tempRet0;
       $eh$lpad$body18$index7Z2D = $97;$eh$lpad$body18$indexZ2D = $96;
       break;
      } else {
       $v$1 = $55;
       STACKTOP = sp;return ($v$1|0); //@line 199 "common/Variant.cpp"
      }
     }
    }
   }
  } while(0);
  if ((label|0) == 40) {
   $99 = ___cxa_find_matching_catch_2()|0;
   $100 = tempRet0;
   $eh$lpad$body18$index7Z2D = $100;$eh$lpad$body18$indexZ2D = $99;
  }
  __ZdlPv($55); //@line 157 "common/Variant.cpp"
  $$02 = $eh$lpad$body18$index7Z2D;$$03 = $eh$lpad$body18$indexZ2D;
  ___resumeException($$03|0);
  // unreachable;
  break;
 }
 case 0:  {
  $101 = (__Znwj(32)|0); //@line 160 "common/Variant.cpp"
  __THREW__ = 0;
  invoke_vi(3431,($101|0)); //@line 55 "common/Variant.cpp"
  $102 = __THREW__; __THREW__ = 0;
  $103 = $102&1;
  if ($103) {
   $110 = ___cxa_find_matching_catch_2()|0;
   $111 = tempRet0;
   __ZdlPv($101); //@line 160 "common/Variant.cpp"
   $$02 = $111;$$03 = $110;
   ___resumeException($$03|0);
   // unreachable;
  } else {
   HEAP32[$101>>2] = (21996); //@line 58 "common/Variant.cpp"
   $104 = ((($101)) + 8|0); //@line 56 "common/Variant.cpp"
   HEAP32[$104>>2] = 7; //@line 56 "common/Variant.cpp"
   $105 = ((($101)) + 16|0); //@line 57 "common/Variant.cpp"
   $106 = $105; //@line 57 "common/Variant.cpp"
   $107 = $106; //@line 57 "common/Variant.cpp"
   HEAP32[$107>>2] = 0; //@line 57 "common/Variant.cpp"
   $108 = (($106) + 4)|0; //@line 57 "common/Variant.cpp"
   $109 = $108; //@line 57 "common/Variant.cpp"
   HEAP32[$109>>2] = 0; //@line 57 "common/Variant.cpp"
   $v$1 = $101;
   STACKTOP = sp;return ($v$1|0); //@line 199 "common/Variant.cpp"
  }
  break;
 }
 case 5:  {
  if (!($allowTables)) {
   $v$1 = 0;
   STACKTOP = sp;return ($v$1|0); //@line 199 "common/Variant.cpp"
  }
  $112 = (__Znwj(12)|0); //@line 166 "common/Variant.cpp"
  HEAP32[$112>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $113 = ((($112)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$113>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $114 = ((($112)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$114>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  _lua_pushnil($L); //@line 167 "common/Variant.cpp"
  $115 = (_lua_next($L,$$01)|0); //@line 168 "common/Variant.cpp"
  $116 = ($115|0)==(0); //@line 168 "common/Variant.cpp"
  L75: do {
   if (!($116)) {
    $117 = ((($0)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
    while(1) {
     $118 = (__ZN4love7Variant7fromLuaEP9lua_Stateib($L,-2,0)|0); //@line 170 "common/Variant.cpp"
     $119 = $118; //@line 170 "common/Variant.cpp"
     $120 = ($118|0)==(0|0); //@line 171 "common/Variant.cpp"
     if ($120) {
      label = 49;
      break;
     }
     $121 = (__ZN4love7Variant7fromLuaEP9lua_Stateib($L,-1,0)|0); //@line 178 "common/Variant.cpp"
     $122 = ($121|0)==(0|0); //@line 179 "common/Variant.cpp"
     if ($122) {
      $$lcssa33 = $118;
      label = 51;
      break;
     }
     $126 = $121; //@line 178 "common/Variant.cpp"
     HEAP32[$0>>2] = $119; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     HEAP32[$117>>2] = $126; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     $127 = HEAP32[$113>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $128 = HEAP32[$114>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $129 = ($127>>>0)<($128>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($129) {
      $130 = $0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $131 = $130; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $132 = HEAP32[$131>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $133 = (($130) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $134 = $133; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $135 = HEAP32[$134>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $136 = $127; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $137 = $136; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$137>>2] = $132; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $138 = (($136) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $139 = $138; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$139>>2] = $135; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $140 = HEAP32[$113>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $141 = ((($140)) + 8|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$113>>2] = $141; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     } else {
      __ZNSt3__16vectorINS_4pairIPN4love7VariantES4_EENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_($112,$0); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     _lua_settop($L,-2); //@line 189 "common/Variant.cpp"
     $142 = (_lua_next($L,$$01)|0); //@line 168 "common/Variant.cpp"
     $143 = ($142|0)==(0); //@line 168 "common/Variant.cpp"
     if ($143) {
      break L75;
     }
    }
    if ((label|0) == 49) {
     _lua_settop($L,-3); //@line 174 "common/Variant.cpp"
    }
    else if ((label|0) == 51) {
     $123 = HEAP32[$$lcssa33>>2]|0; //@line 181 "common/Variant.cpp"
     $124 = ((($123)) + 4|0); //@line 181 "common/Variant.cpp"
     $125 = HEAP32[$124>>2]|0; //@line 181 "common/Variant.cpp"
     FUNCTION_TABLE_vi[$125 & 8191]($$lcssa33); //@line 181 "common/Variant.cpp"
     _lua_settop($L,-3); //@line 183 "common/Variant.cpp"
    }
    $151 = HEAP32[$112>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $152 = HEAP32[$113>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $153 = ($151|0)==($152|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($153) {
     $169 = $151;
    } else {
     $155 = $152;
     while(1) {
      $154 = ((($155)) + -8|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $156 = HEAP32[$154>>2]|0; //@line 48 "common/Variant.cpp"
      $157 = HEAP32[$156>>2]|0; //@line 48 "common/Variant.cpp"
      $158 = ((($157)) + 12|0); //@line 48 "common/Variant.cpp"
      $159 = HEAP32[$158>>2]|0; //@line 48 "common/Variant.cpp"
      FUNCTION_TABLE_vi[$159 & 8191]($156); //@line 48 "common/Variant.cpp"
      $160 = ((($155)) + -4|0); //@line 49 "common/Variant.cpp"
      $161 = HEAP32[$160>>2]|0; //@line 49 "common/Variant.cpp"
      $162 = HEAP32[$161>>2]|0; //@line 49 "common/Variant.cpp"
      $163 = ((($162)) + 12|0); //@line 49 "common/Variant.cpp"
      $164 = HEAP32[$163>>2]|0; //@line 49 "common/Variant.cpp"
      FUNCTION_TABLE_vi[$164 & 8191]($161); //@line 49 "common/Variant.cpp"
      $165 = HEAP32[$113>>2]|0; //@line 1666 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $scevgep4$i$i$i1$i = ((($165)) + -8|0);
      HEAP32[$113>>2] = $scevgep4$i$i$i1$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $166 = HEAP32[$112>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $167 = ($166|0)==($scevgep4$i$i$i1$i|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($167) {
       $169 = $scevgep4$i$i$i1$i;
       break;
      } else {
       $155 = $scevgep4$i$i$i1$i;
      }
     }
    }
    $168 = ($169|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($168)) {
     __ZdlPv($169); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    }
    __ZdlPv($112); //@line 52 "common/Variant.cpp"
    $v$1 = 0;
    STACKTOP = sp;return ($v$1|0); //@line 199 "common/Variant.cpp"
   }
  } while(0);
  $144 = (__Znwj(32)|0); //@line 193 "common/Variant.cpp"
  __THREW__ = 0;
  invoke_vi(3431,($144|0)); //@line 110 "common/Variant.cpp"
  $145 = __THREW__; __THREW__ = 0;
  $146 = $145&1;
  if ($146) {
   $149 = ___cxa_find_matching_catch_2()|0;
   $150 = tempRet0;
   __ZdlPv($144); //@line 193 "common/Variant.cpp"
   $$02 = $150;$$03 = $149;
   ___resumeException($$03|0);
   // unreachable;
  } else {
   HEAP32[$144>>2] = (21996); //@line 112 "common/Variant.cpp"
   $147 = ((($144)) + 8|0); //@line 111 "common/Variant.cpp"
   HEAP32[$147>>2] = 8; //@line 111 "common/Variant.cpp"
   $148 = ((($144)) + 16|0); //@line 113 "common/Variant.cpp"
   HEAP32[$148>>2] = $112; //@line 113 "common/Variant.cpp"
   $v$1 = $144;
   STACKTOP = sp;return ($v$1|0); //@line 199 "common/Variant.cpp"
  }
  break;
 }
 default: {
  $v$1 = 0;
  STACKTOP = sp;return ($v$1|0); //@line 199 "common/Variant.cpp"
 }
 }
 return (0)|0;
}
function __ZNSt3__16vectorINS_4pairIPN4love7VariantES4_EENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(536870911); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 3; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(268435455); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 2; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$42 = $18;
  } else {
   $$0$i1 = $17;$50 = $18;$51 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 536870911;$50 = $12;$51 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 3; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $51;$42 = $50;
 }
 $24 = (($25) + ($26<<3)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = $__x; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $31 = HEAP32[$30>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $32 = (($29) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = $32; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = HEAP32[$33>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = $24; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = $35; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$36>>2] = $31; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = (($35) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $37; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$38>>2] = $34; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $39 = ((($24)) + 8|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $40 = $39; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $41 = (($42) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $43 = $41 >> 3; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $44 = (0 - ($43))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $45 = (($24) + ($44<<3)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $46 = $45; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($45|0),($47|0),($41|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $46; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $40; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $48 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($48) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $49 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($49); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love7Variant5toLuaEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 204 "common/Variant.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 204 "common/Variant.cpp"
 switch ($1|0) {
 case 1:  {
  $2 = ((($this)) + 16|0); //@line 207 "common/Variant.cpp"
  $3 = HEAP8[$2>>0]|0; //@line 207 "common/Variant.cpp"
  $4 = $3&255; //@line 207 "common/Variant.cpp"
  _lua_pushboolean($L,$4); //@line 207 "common/Variant.cpp"
  return; //@line 245 "common/Variant.cpp"
  break;
 }
 case 3:  {
  $5 = ((($this)) + 16|0); //@line 210 "common/Variant.cpp"
  _lua_pushlstring($L,$5,1); //@line 210 "common/Variant.cpp"
  return; //@line 245 "common/Variant.cpp"
  break;
 }
 case 2:  {
  $6 = ((($this)) + 16|0); //@line 213 "common/Variant.cpp"
  $7 = +HEAPF64[$6>>3]; //@line 213 "common/Variant.cpp"
  _lua_pushnumber($L,$7); //@line 213 "common/Variant.cpp"
  return; //@line 245 "common/Variant.cpp"
  break;
 }
 case 4:  {
  $8 = ((($this)) + 16|0); //@line 216 "common/Variant.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 216 "common/Variant.cpp"
  $10 = ((($8)) + 4|0); //@line 216 "common/Variant.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 216 "common/Variant.cpp"
  _lua_pushlstring($L,$9,$11); //@line 216 "common/Variant.cpp"
  return; //@line 245 "common/Variant.cpp"
  break;
 }
 case 5:  {
  $12 = ((($this)) + 16|0); //@line 219 "common/Variant.cpp"
  $13 = HEAP32[$12>>2]|0; //@line 219 "common/Variant.cpp"
  _lua_pushlightuserdata($L,$13); //@line 219 "common/Variant.cpp"
  return; //@line 245 "common/Variant.cpp"
  break;
 }
 case 6:  {
  $14 = ((($this)) + 24|0); //@line 222 "common/Variant.cpp"
  $15 = HEAP32[$14>>2]|0; //@line 222 "common/Variant.cpp"
  $16 = ($15|0)==(0); //@line 222 "common/Variant.cpp"
  $17 = ((($this)) + 16|0); //@line 225 "common/Variant.cpp"
  if ($16) {
   $19 = HEAP32[$17>>2]|0; //@line 225 "common/Variant.cpp"
   _lua_pushlightuserdata($L,$19); //@line 225 "common/Variant.cpp"
   return; //@line 245 "common/Variant.cpp"
  } else {
   $18 = HEAP32[$17>>2]|0; //@line 223 "common/Variant.cpp"
   __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,$15,$18); //@line 223 "common/Variant.cpp"
   return; //@line 245 "common/Variant.cpp"
  }
  break;
 }
 case 8:  {
  $20 = ((($this)) + 16|0); //@line 231 "common/Variant.cpp"
  $21 = HEAP32[$20>>2]|0; //@line 231 "common/Variant.cpp"
  $22 = ((($21)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = HEAP32[$22>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $24 = HEAP32[$21>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $25 = (($23) - ($24))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $26 = $25 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  _lua_createtable($L,0,$26); //@line 231 "common/Variant.cpp"
  $27 = HEAP32[$20>>2]|0; //@line 232 "common/Variant.cpp"
  $28 = ((($27)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $29 = HEAP32[$28>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $30 = HEAP32[$27>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $31 = ($29|0)==($30|0); //@line 232 "common/Variant.cpp"
  if ($31) {
   return; //@line 245 "common/Variant.cpp"
  } else {
   $33 = $30;$i$01 = 0;
  }
  while(1) {
   $32 = $33; //@line 1521 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $34 = (($32) + ($i$01<<3)|0); //@line 235 "common/Variant.cpp"
   $35 = HEAP32[$34>>2]|0; //@line 235 "common/Variant.cpp"
   __ZN4love7Variant5toLuaEP9lua_State($35,$L); //@line 235 "common/Variant.cpp"
   $36 = (((($32) + ($i$01<<3)|0)) + 4|0); //@line 236 "common/Variant.cpp"
   $37 = HEAP32[$36>>2]|0; //@line 236 "common/Variant.cpp"
   __ZN4love7Variant5toLuaEP9lua_State($37,$L); //@line 236 "common/Variant.cpp"
   _lua_settable($L,-3); //@line 237 "common/Variant.cpp"
   $38 = (($i$01) + 1)|0; //@line 232 "common/Variant.cpp"
   $39 = HEAP32[$20>>2]|0; //@line 232 "common/Variant.cpp"
   $40 = ((($39)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $41 = HEAP32[$40>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $42 = HEAP32[$39>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $43 = (($41) - ($42))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $44 = $43 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $45 = ($38>>>0)<($44>>>0); //@line 232 "common/Variant.cpp"
   if ($45) {
    $33 = $42;$i$01 = $38;
   } else {
    break;
   }
  }
  return; //@line 245 "common/Variant.cpp"
  break;
 }
 default: {
  _lua_pushnil($L); //@line 242 "common/Variant.cpp"
  return; //@line 245 "common/Variant.cpp"
 }
 }
}
function __ZN4love5audio4null5AudioD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ModuleD2Ev($this); //@line 37 "modules/audio/null/Audio.cpp"
 return; //@line 37 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5AudioD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ModuleD2Ev($this); //@line 37 "modules/audio/null/Audio.cpp"
 __ZdlPv($this); //@line 36 "modules/audio/null/Audio.cpp"
 return; //@line 37 "modules/audio/null/Audio.cpp"
}
function __ZNK4love5audio5Audio13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0; //@line 70 "./modules/audio/Audio.h"
}
function __ZNK4love5audio4null5Audio7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (364776|0); //@line 41 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio9newSourceEPNS_5sound7DecoderE($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (__Znwj(56)|0); //@line 46 "modules/audio/null/Audio.cpp"
 __THREW__ = 0;
 invoke_vi(3905,($1|0)); //@line 46 "modules/audio/null/Audio.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  __ZdlPv($1); //@line 46 "modules/audio/null/Audio.cpp"
  ___resumeException($4|0);
  // unreachable;
 } else {
  return ($1|0); //@line 46 "modules/audio/null/Audio.cpp"
 }
 return (0)|0;
}
function __ZN4love5audio4null5Audio9newSourceEPNS_5sound9SoundDataE($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (__Znwj(56)|0); //@line 51 "modules/audio/null/Audio.cpp"
 __THREW__ = 0;
 invoke_vi(3905,($1|0)); //@line 51 "modules/audio/null/Audio.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  __ZdlPv($1); //@line 51 "modules/audio/null/Audio.cpp"
  ___resumeException($4|0);
  // unreachable;
 } else {
  return ($1|0); //@line 51 "modules/audio/null/Audio.cpp"
 }
 return (0)|0;
}
function __ZNK4love5audio4null5Audio14getSourceCountEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0; //@line 56 "modules/audio/null/Audio.cpp"
}
function __ZNK4love5audio4null5Audio13getMaxSourcesEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0; //@line 61 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio6updateEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 67 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio4playEPNS0_6SourceE($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0; //@line 71 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio4stopEPNS0_6SourceE($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 76 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio4stopEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 80 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio5pauseEPNS0_6SourceE($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 84 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio5pauseEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 88 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio6resumeEPNS0_6SourceE($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 92 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio6resumeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 96 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio6rewindEPNS0_6SourceE($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 100 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio6rewindEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 104 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio9setVolumeEf($this,$volume) {
 $this = $this|0;
 $volume = +$volume;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 108 "modules/audio/null/Audio.cpp"
 HEAPF32[$0>>2] = $volume; //@line 108 "modules/audio/null/Audio.cpp"
 return; //@line 109 "modules/audio/null/Audio.cpp"
}
function __ZNK4love5audio4null5Audio9getVolumeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 113 "modules/audio/null/Audio.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 113 "modules/audio/null/Audio.cpp"
 return (+$1); //@line 113 "modules/audio/null/Audio.cpp"
}
function __ZNK4love5audio4null5Audio11getPositionEPf($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 118 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio11setPositionEPf($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 122 "modules/audio/null/Audio.cpp"
}
function __ZNK4love5audio4null5Audio14getOrientationEPf($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 126 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio14setOrientationEPf($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 130 "modules/audio/null/Audio.cpp"
}
function __ZNK4love5audio4null5Audio11getVelocityEPf($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 134 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio11setVelocityEPf($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 138 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio15setDopplerScaleEf($this,$0) {
 $this = $this|0;
 $0 = +$0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 142 "modules/audio/null/Audio.cpp"
}
function __ZNK4love5audio4null5Audio15getDopplerScaleEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return +1; //@line 146 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio6recordEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 151 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio15getRecordedDataEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (0|0); //@line 155 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio13stopRecordingEb($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (0|0); //@line 160 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio9canRecordEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0; //@line 165 "modules/audio/null/Audio.cpp"
}
function __ZNK4love5audio4null5Audio16getDistanceModelEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 170 "modules/audio/null/Audio.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 170 "modules/audio/null/Audio.cpp"
 return ($1|0); //@line 170 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5Audio16setDistanceModelENS0_5Audio13DistanceModelE($this,$distanceModel) {
 $this = $this|0;
 $distanceModel = $distanceModel|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 175 "modules/audio/null/Audio.cpp"
 HEAP32[$0>>2] = $distanceModel; //@line 175 "modules/audio/null/Audio.cpp"
 return; //@line 176 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null5AudioC2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./common/Module.h"
 HEAP32[$this>>2] = (22020); //@line 32 "modules/audio/null/Audio.cpp"
 $0 = ((($this)) + 12|0); //@line 31 "modules/audio/null/Audio.cpp"
 HEAP32[$0>>2] = 0; //@line 31 "modules/audio/null/Audio.cpp"
 return; //@line 33 "modules/audio/null/Audio.cpp"
}
function __ZN4love5audio4null6SourceD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5audio6SourceD2Ev($this); //@line 37 "modules/audio/null/Source.cpp"
 return; //@line 37 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6SourceD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5audio6SourceD2Ev($this); //@line 37 "modules/audio/null/Source.cpp"
 __ZdlPv($this); //@line 36 "modules/audio/null/Source.cpp"
 return; //@line 37 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source5cloneEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 41 "modules/audio/null/Source.cpp"
 $1 = ((($0)) + 8|0); //@line 41 "modules/audio/null/Source.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 41 "modules/audio/null/Source.cpp"
 FUNCTION_TABLE_vi[$2 & 8191]($this); //@line 41 "modules/audio/null/Source.cpp"
 return ($this|0); //@line 42 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source4playEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0; //@line 47 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source4stopEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 52 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source5pauseEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 56 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source6resumeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 60 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source6rewindEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 64 "modules/audio/null/Source.cpp"
}
function __ZNK4love5audio4null6Source9isStoppedEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1; //@line 68 "modules/audio/null/Source.cpp"
}
function __ZNK4love5audio4null6Source8isPausedEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0; //@line 73 "modules/audio/null/Source.cpp"
}
function __ZNK4love5audio4null6Source10isFinishedEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1; //@line 78 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source6updateEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0; //@line 83 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source8setPitchEf($this,$pitch) {
 $this = $this|0;
 $pitch = +$pitch;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 88 "modules/audio/null/Source.cpp"
 HEAPF32[$0>>2] = $pitch; //@line 88 "modules/audio/null/Source.cpp"
 return; //@line 89 "modules/audio/null/Source.cpp"
}
function __ZNK4love5audio4null6Source8getPitchEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 93 "modules/audio/null/Source.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 93 "modules/audio/null/Source.cpp"
 return (+$1); //@line 93 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source9setVolumeEf($this,$volume) {
 $this = $this|0;
 $volume = +$volume;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 98 "modules/audio/null/Source.cpp"
 HEAPF32[$0>>2] = $volume; //@line 98 "modules/audio/null/Source.cpp"
 return; //@line 99 "modules/audio/null/Source.cpp"
}
function __ZNK4love5audio4null6Source9getVolumeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 103 "modules/audio/null/Source.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 103 "modules/audio/null/Source.cpp"
 return (+$1); //@line 103 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source4seekEfNS0_6Source4UnitE($this,$0,$1) {
 $this = $this|0;
 $0 = +$0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 108 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source4tellENS0_6Source4UnitE($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return +0; //@line 112 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source11getDurationENS0_6Source4UnitE($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return +-1; //@line 117 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source11setPositionEPf($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 122 "modules/audio/null/Source.cpp"
}
function __ZNK4love5audio4null6Source11getPositionEPf($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 126 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source11setVelocityEPf($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 130 "modules/audio/null/Source.cpp"
}
function __ZNK4love5audio4null6Source11getVelocityEPf($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 134 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source12setDirectionEPf($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 138 "modules/audio/null/Source.cpp"
}
function __ZNK4love5audio4null6Source12getDirectionEPf($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 142 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source7setConeEfff($this,$innerAngle,$outerAngle,$outerVolume) {
 $this = $this|0;
 $innerAngle = +$innerAngle;
 $outerAngle = +$outerAngle;
 $outerVolume = +$outerVolume;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 146 "modules/audio/null/Source.cpp"
 HEAPF32[$0>>2] = $innerAngle; //@line 146 "modules/audio/null/Source.cpp"
 $1 = ((($this)) + 24|0); //@line 147 "modules/audio/null/Source.cpp"
 HEAPF32[$1>>2] = $outerAngle; //@line 147 "modules/audio/null/Source.cpp"
 $2 = ((($this)) + 28|0); //@line 148 "modules/audio/null/Source.cpp"
 HEAPF32[$2>>2] = $outerVolume; //@line 148 "modules/audio/null/Source.cpp"
 return; //@line 149 "modules/audio/null/Source.cpp"
}
function __ZNK4love5audio4null6Source7getConeERfS3_S3_($this,$innerAngle,$outerAngle,$outerVolume) {
 $this = $this|0;
 $innerAngle = $innerAngle|0;
 $outerAngle = $outerAngle|0;
 $outerVolume = $outerVolume|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 153 "modules/audio/null/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 153 "modules/audio/null/Source.cpp"
 HEAP32[$innerAngle>>2] = $1; //@line 153 "modules/audio/null/Source.cpp"
 $2 = ((($this)) + 24|0); //@line 154 "modules/audio/null/Source.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 154 "modules/audio/null/Source.cpp"
 HEAP32[$outerAngle>>2] = $3; //@line 154 "modules/audio/null/Source.cpp"
 $4 = ((($this)) + 28|0); //@line 155 "modules/audio/null/Source.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 155 "modules/audio/null/Source.cpp"
 HEAP32[$outerVolume>>2] = $5; //@line 155 "modules/audio/null/Source.cpp"
 return; //@line 156 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source11setRelativeEb($this,$enable) {
 $this = $this|0;
 $enable = $enable|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 160 "modules/audio/null/Source.cpp"
 $1 = $enable&1; //@line 160 "modules/audio/null/Source.cpp"
 HEAP8[$0>>0] = $1; //@line 160 "modules/audio/null/Source.cpp"
 return; //@line 161 "modules/audio/null/Source.cpp"
}
function __ZNK4love5audio4null6Source10isRelativeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 165 "modules/audio/null/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 165 "modules/audio/null/Source.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 165 "modules/audio/null/Source.cpp"
 return ($2|0); //@line 165 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source10setLoopingEb($this,$looping) {
 $this = $this|0;
 $looping = $looping|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 33|0); //@line 170 "modules/audio/null/Source.cpp"
 $1 = $looping&1; //@line 170 "modules/audio/null/Source.cpp"
 HEAP8[$0>>0] = $1; //@line 170 "modules/audio/null/Source.cpp"
 return; //@line 171 "modules/audio/null/Source.cpp"
}
function __ZNK4love5audio4null6Source9isLoopingEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 33|0); //@line 175 "modules/audio/null/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 175 "modules/audio/null/Source.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 175 "modules/audio/null/Source.cpp"
 return ($2|0); //@line 175 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source12setMinVolumeEf($this,$volume) {
 $this = $this|0;
 $volume = +$volume;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 36|0); //@line 180 "modules/audio/null/Source.cpp"
 HEAPF32[$0>>2] = $volume; //@line 180 "modules/audio/null/Source.cpp"
 return; //@line 181 "modules/audio/null/Source.cpp"
}
function __ZNK4love5audio4null6Source12getMinVolumeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 36|0); //@line 185 "modules/audio/null/Source.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 185 "modules/audio/null/Source.cpp"
 return (+$1); //@line 185 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source12setMaxVolumeEf($this,$volume) {
 $this = $this|0;
 $volume = +$volume;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0); //@line 190 "modules/audio/null/Source.cpp"
 HEAPF32[$0>>2] = $volume; //@line 190 "modules/audio/null/Source.cpp"
 return; //@line 191 "modules/audio/null/Source.cpp"
}
function __ZNK4love5audio4null6Source12getMaxVolumeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0); //@line 195 "modules/audio/null/Source.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 195 "modules/audio/null/Source.cpp"
 return (+$1); //@line 195 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source20setReferenceDistanceEf($this,$distance) {
 $this = $this|0;
 $distance = +$distance;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 44|0); //@line 200 "modules/audio/null/Source.cpp"
 HEAPF32[$0>>2] = $distance; //@line 200 "modules/audio/null/Source.cpp"
 return; //@line 201 "modules/audio/null/Source.cpp"
}
function __ZNK4love5audio4null6Source20getReferenceDistanceEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 44|0); //@line 205 "modules/audio/null/Source.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 205 "modules/audio/null/Source.cpp"
 return (+$1); //@line 205 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source16setRolloffFactorEf($this,$factor) {
 $this = $this|0;
 $factor = +$factor;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 210 "modules/audio/null/Source.cpp"
 HEAPF32[$0>>2] = $factor; //@line 210 "modules/audio/null/Source.cpp"
 return; //@line 211 "modules/audio/null/Source.cpp"
}
function __ZNK4love5audio4null6Source16getRolloffFactorEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 215 "modules/audio/null/Source.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 215 "modules/audio/null/Source.cpp"
 return (+$1); //@line 215 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6Source14setMaxDistanceEf($this,$distance) {
 $this = $this|0;
 $distance = +$distance;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 220 "modules/audio/null/Source.cpp"
 HEAPF32[$0>>2] = $distance; //@line 220 "modules/audio/null/Source.cpp"
 return; //@line 221 "modules/audio/null/Source.cpp"
}
function __ZNK4love5audio4null6Source14getMaxDistanceEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 225 "modules/audio/null/Source.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 225 "modules/audio/null/Source.cpp"
 return (+$1); //@line 225 "modules/audio/null/Source.cpp"
}
function __ZNK4love5audio4null6Source11getChannelsEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 2; //@line 230 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio4null6SourceC2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5audio6SourceC2ENS1_4TypeE($this,0); //@line 31 "modules/audio/null/Source.cpp"
 HEAP32[$this>>2] = (22172); //@line 32 "modules/audio/null/Source.cpp"
 return; //@line 33 "modules/audio/null/Source.cpp"
}
function __ZN4love5audio6openal5Audio10PoolThreadD2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (22360); //@line 43 "modules/audio/openal/Audio.cpp"
 $0 = ((($this)) + 28|0); //@line 44 "modules/audio/openal/Audio.cpp"
 __ZN4love6thread8MutexRefD2Ev($0); //@line 44 "modules/audio/openal/Audio.cpp"
 __ZN4love6thread10ThreadableD2Ev($this); //@line 44 "modules/audio/openal/Audio.cpp"
 return; //@line 44 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio10PoolThreadD0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (22360); //@line 43 "modules/audio/openal/Audio.cpp"
 $0 = ((($this)) + 28|0); //@line 44 "modules/audio/openal/Audio.cpp"
 __ZN4love6thread8MutexRefD2Ev($0); //@line 44 "modules/audio/openal/Audio.cpp"
 __ZN4love6thread10ThreadableD2Ev($this); //@line 44 "modules/audio/openal/Audio.cpp"
 __ZdlPv($this); //@line 43 "modules/audio/openal/Audio.cpp"
 return; //@line 44 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio10PoolThread14threadFunctionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lock = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ((($this)) + 28|0); //@line 52 "modules/audio/openal/Audio.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 52 "modules/audio/openal/Audio.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$1); //@line 52 "modules/audio/openal/Audio.cpp"
 $2 = ((($this)) + 24|0); //@line 53 "modules/audio/openal/Audio.cpp"
 $3 = HEAP8[$2>>0]|0; //@line 53 "modules/audio/openal/Audio.cpp"
 $4 = ($3<<24>>24)==(0); //@line 53 "modules/audio/openal/Audio.cpp"
 __ZN4love6thread4LockD2Ev($lock); //@line 57 "modules/audio/openal/Audio.cpp"
 if (!($4)) {
  STACKTOP = sp;return; //@line 62 "modules/audio/openal/Audio.cpp"
 }
 $5 = ((($this)) + 20|0); //@line 59 "modules/audio/openal/Audio.cpp"
 while(1) {
  $6 = HEAP32[$5>>2]|0; //@line 59 "modules/audio/openal/Audio.cpp"
  __ZN4love5audio6openal4Pool6updateEv($6); //@line 59 "modules/audio/openal/Audio.cpp"
  __ZN4love5delayEj(5); //@line 60 "modules/audio/openal/Audio.cpp"
  $7 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 52 "modules/audio/openal/Audio.cpp"
  __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$7); //@line 52 "modules/audio/openal/Audio.cpp"
  $8 = HEAP8[$2>>0]|0; //@line 53 "modules/audio/openal/Audio.cpp"
  $9 = ($8<<24>>24)==(0); //@line 53 "modules/audio/openal/Audio.cpp"
  __ZN4love6thread4LockD2Ev($lock); //@line 57 "modules/audio/openal/Audio.cpp"
  if (!($9)) {
   break;
  }
 }
 STACKTOP = sp;return; //@line 62 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5AudioD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lock$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock$i = sp;
 HEAP32[$this>>2] = (22380); //@line 131 "modules/audio/openal/Audio.cpp"
 $0 = ((($this)) + 24|0); //@line 132 "modules/audio/openal/Audio.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 132 "modules/audio/openal/Audio.cpp"
 $2 = ((($1)) + 28|0); //@line 66 "modules/audio/openal/Audio.cpp"
 __THREW__ = 0;
 $3 = (invoke_ii(3782,($2|0))|0); //@line 66 "modules/audio/openal/Audio.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  __THREW__ = 0;
  invoke_vii(3783,($lock$i|0),($3|0)); //@line 66 "modules/audio/openal/Audio.cpp"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if (!($7)) {
   $8 = ((($1)) + 24|0); //@line 67 "modules/audio/openal/Audio.cpp"
   HEAP8[$8>>0] = 1; //@line 67 "modules/audio/openal/Audio.cpp"
   __ZN4love6thread4LockD2Ev($lock$i); //@line 68 "modules/audio/openal/Audio.cpp"
   $9 = HEAP32[$0>>2]|0; //@line 133 "modules/audio/openal/Audio.cpp"
   __THREW__ = 0;
   invoke_vi(3906,($9|0)); //@line 133 "modules/audio/openal/Audio.cpp"
   $10 = __THREW__; __THREW__ = 0;
   $11 = $10&1;
   if (!($11)) {
    $12 = HEAP32[$0>>2]|0; //@line 135 "modules/audio/openal/Audio.cpp"
    $13 = ($12|0)==(0|0); //@line 135 "modules/audio/openal/Audio.cpp"
    if (!($13)) {
     $14 = HEAP32[$12>>2]|0; //@line 135 "modules/audio/openal/Audio.cpp"
     $15 = ((($14)) + 4|0); //@line 135 "modules/audio/openal/Audio.cpp"
     $16 = HEAP32[$15>>2]|0; //@line 135 "modules/audio/openal/Audio.cpp"
     FUNCTION_TABLE_vi[$16 & 8191]($12); //@line 135 "modules/audio/openal/Audio.cpp"
    }
    $17 = ((($this)) + 20|0); //@line 136 "modules/audio/openal/Audio.cpp"
    $18 = HEAP32[$17>>2]|0; //@line 136 "modules/audio/openal/Audio.cpp"
    $19 = ($18|0)==(0|0); //@line 136 "modules/audio/openal/Audio.cpp"
    if (!($19)) {
     __ZN4love5audio6openal4PoolD2Ev($18); //@line 136 "modules/audio/openal/Audio.cpp"
     __ZdlPv($18); //@line 136 "modules/audio/openal/Audio.cpp"
    }
    __THREW__ = 0;
    (invoke_ii(3907,(0|0))|0); //@line 138 "modules/audio/openal/Audio.cpp"
    $20 = __THREW__; __THREW__ = 0;
    $21 = $20&1;
    if (!($21)) {
     $22 = ((($this)) + 16|0); //@line 139 "modules/audio/openal/Audio.cpp"
     $23 = HEAP32[$22>>2]|0; //@line 139 "modules/audio/openal/Audio.cpp"
     __THREW__ = 0;
     invoke_vi(3908,($23|0)); //@line 139 "modules/audio/openal/Audio.cpp"
     $24 = __THREW__; __THREW__ = 0;
     $25 = $24&1;
     if (!($25)) {
      $26 = ((($this)) + 8|0); //@line 141 "modules/audio/openal/Audio.cpp"
      $27 = HEAP32[$26>>2]|0; //@line 141 "modules/audio/openal/Audio.cpp"
      __THREW__ = 0;
      (invoke_ii(3909,($27|0))|0); //@line 141 "modules/audio/openal/Audio.cpp"
      $28 = __THREW__; __THREW__ = 0;
      $29 = $28&1;
      if (!($29)) {
       __ZN4love6ModuleD2Ev($this); //@line 142 "modules/audio/openal/Audio.cpp"
       STACKTOP = sp;return; //@line 142 "modules/audio/openal/Audio.cpp"
      }
     }
    }
   }
  }
 }
 $30 = ___cxa_find_matching_catch_3(0|0)|0;
 $31 = tempRet0;
 __ZN4love6ModuleD2Ev($this); //@line 142 "modules/audio/openal/Audio.cpp"
 ___clang_call_terminate($30); //@line 142 "modules/audio/openal/Audio.cpp"
 // unreachable; //@line 142 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5AudioD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5audio6openal5AudioD2Ev($this); //@line 131 "modules/audio/openal/Audio.cpp"
 __ZdlPv($this); //@line 131 "modules/audio/openal/Audio.cpp"
 return; //@line 142 "modules/audio/openal/Audio.cpp"
}
function __ZNK4love5audio6openal5Audio7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (364884|0); //@line 147 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio9newSourceEPNS_5sound7DecoderE($this,$decoder) {
 $this = $this|0;
 $decoder = $decoder|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(168)|0); //@line 152 "modules/audio/openal/Audio.cpp"
 $1 = ((($this)) + 20|0); //@line 152 "modules/audio/openal/Audio.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 152 "modules/audio/openal/Audio.cpp"
 __THREW__ = 0;
 invoke_viii(3910,($0|0),($2|0),($decoder|0)); //@line 152 "modules/audio/openal/Audio.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  __ZdlPv($0); //@line 152 "modules/audio/openal/Audio.cpp"
  ___resumeException($5|0);
  // unreachable;
 } else {
  return ($0|0); //@line 152 "modules/audio/openal/Audio.cpp"
 }
 return (0)|0;
}
function __ZN4love5audio6openal5Audio9newSourceEPNS_5sound9SoundDataE($this,$soundData) {
 $this = $this|0;
 $soundData = $soundData|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(168)|0); //@line 157 "modules/audio/openal/Audio.cpp"
 $1 = ((($this)) + 20|0); //@line 157 "modules/audio/openal/Audio.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 157 "modules/audio/openal/Audio.cpp"
 __THREW__ = 0;
 invoke_viii(3911,($0|0),($2|0),($soundData|0)); //@line 157 "modules/audio/openal/Audio.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  __ZdlPv($0); //@line 157 "modules/audio/openal/Audio.cpp"
  ___resumeException($5|0);
  // unreachable;
 } else {
  return ($0|0); //@line 157 "modules/audio/openal/Audio.cpp"
 }
 return (0)|0;
}
function __ZNK4love5audio6openal5Audio14getSourceCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 162 "modules/audio/openal/Audio.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 162 "modules/audio/openal/Audio.cpp"
 $2 = (__ZNK4love5audio6openal4Pool14getSourceCountEv($1)|0); //@line 162 "modules/audio/openal/Audio.cpp"
 return ($2|0); //@line 162 "modules/audio/openal/Audio.cpp"
}
function __ZNK4love5audio6openal5Audio13getMaxSourcesEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 167 "modules/audio/openal/Audio.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 167 "modules/audio/openal/Audio.cpp"
 $2 = (__ZNK4love5audio6openal4Pool13getMaxSourcesEv($1)|0); //@line 167 "modules/audio/openal/Audio.cpp"
 return ($2|0); //@line 167 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio6updateEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 172 "modules/audio/openal/Audio.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 172 "modules/audio/openal/Audio.cpp"
 __ZN4love5audio6openal4Pool6updateEv($1); //@line 172 "modules/audio/openal/Audio.cpp"
 return; //@line 173 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio4playEPNS0_6SourceE($this,$source) {
 $this = $this|0;
 $source = $source|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$source>>2]|0; //@line 177 "modules/audio/openal/Audio.cpp"
 $1 = ((($0)) + 20|0); //@line 177 "modules/audio/openal/Audio.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 177 "modules/audio/openal/Audio.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($source)|0); //@line 177 "modules/audio/openal/Audio.cpp"
 return ($3|0); //@line 177 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio4stopEPNS0_6SourceE($this,$source) {
 $this = $this|0;
 $source = $source|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$source>>2]|0; //@line 182 "modules/audio/openal/Audio.cpp"
 $1 = ((($0)) + 24|0); //@line 182 "modules/audio/openal/Audio.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 182 "modules/audio/openal/Audio.cpp"
 FUNCTION_TABLE_vi[$2 & 8191]($source); //@line 182 "modules/audio/openal/Audio.cpp"
 return; //@line 183 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio4stopEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 187 "modules/audio/openal/Audio.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 187 "modules/audio/openal/Audio.cpp"
 __ZN4love5audio6openal4Pool4stopEv($1); //@line 187 "modules/audio/openal/Audio.cpp"
 return; //@line 188 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio5pauseEPNS0_6SourceE($this,$source) {
 $this = $this|0;
 $source = $source|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$source>>2]|0; //@line 192 "modules/audio/openal/Audio.cpp"
 $1 = ((($0)) + 28|0); //@line 192 "modules/audio/openal/Audio.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 192 "modules/audio/openal/Audio.cpp"
 FUNCTION_TABLE_vi[$2 & 8191]($source); //@line 192 "modules/audio/openal/Audio.cpp"
 return; //@line 193 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio5pauseEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 197 "modules/audio/openal/Audio.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 197 "modules/audio/openal/Audio.cpp"
 __ZN4love5audio6openal4Pool5pauseEv($1); //@line 197 "modules/audio/openal/Audio.cpp"
 return; //@line 201 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio6resumeEPNS0_6SourceE($this,$source) {
 $this = $this|0;
 $source = $source|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$source>>2]|0; //@line 205 "modules/audio/openal/Audio.cpp"
 $1 = ((($0)) + 32|0); //@line 205 "modules/audio/openal/Audio.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 205 "modules/audio/openal/Audio.cpp"
 FUNCTION_TABLE_vi[$2 & 8191]($source); //@line 205 "modules/audio/openal/Audio.cpp"
 return; //@line 206 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio6resumeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 213 "modules/audio/openal/Audio.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 213 "modules/audio/openal/Audio.cpp"
 __ZN4love5audio6openal4Pool6resumeEv($1); //@line 213 "modules/audio/openal/Audio.cpp"
 return; //@line 214 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio6rewindEPNS0_6SourceE($this,$source) {
 $this = $this|0;
 $source = $source|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$source>>2]|0; //@line 218 "modules/audio/openal/Audio.cpp"
 $1 = ((($0)) + 36|0); //@line 218 "modules/audio/openal/Audio.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 218 "modules/audio/openal/Audio.cpp"
 FUNCTION_TABLE_vi[$2 & 8191]($source); //@line 218 "modules/audio/openal/Audio.cpp"
 return; //@line 219 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio6rewindEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 223 "modules/audio/openal/Audio.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 223 "modules/audio/openal/Audio.cpp"
 __ZN4love5audio6openal4Pool6rewindEv($1); //@line 223 "modules/audio/openal/Audio.cpp"
 return; //@line 224 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio9setVolumeEf($this,$volume) {
 $this = $this|0;
 $volume = +$volume;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _alListenerf(4106,(+$volume)); //@line 228 "modules/audio/openal/Audio.cpp"
 return; //@line 229 "modules/audio/openal/Audio.cpp"
}
function __ZNK4love5audio6openal5Audio9getVolumeEv($this) {
 $this = $this|0;
 var $0 = 0.0, $volume = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $volume = sp;
 _alGetListenerf(4106,($volume|0)); //@line 234 "modules/audio/openal/Audio.cpp"
 $0 = +HEAPF32[$volume>>2]; //@line 235 "modules/audio/openal/Audio.cpp"
 STACKTOP = sp;return (+$0); //@line 235 "modules/audio/openal/Audio.cpp"
}
function __ZNK4love5audio6openal5Audio11getPositionEPf($this,$v) {
 $this = $this|0;
 $v = $v|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _alGetListenerfv(4100,($v|0)); //@line 240 "modules/audio/openal/Audio.cpp"
 return; //@line 241 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio11setPositionEPf($this,$v) {
 $this = $this|0;
 $v = $v|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _alListenerfv(4100,($v|0)); //@line 245 "modules/audio/openal/Audio.cpp"
 return; //@line 246 "modules/audio/openal/Audio.cpp"
}
function __ZNK4love5audio6openal5Audio14getOrientationEPf($this,$v) {
 $this = $this|0;
 $v = $v|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _alGetListenerfv(4111,($v|0)); //@line 250 "modules/audio/openal/Audio.cpp"
 return; //@line 251 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio14setOrientationEPf($this,$v) {
 $this = $this|0;
 $v = $v|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _alListenerfv(4111,($v|0)); //@line 255 "modules/audio/openal/Audio.cpp"
 return; //@line 256 "modules/audio/openal/Audio.cpp"
}
function __ZNK4love5audio6openal5Audio11getVelocityEPf($this,$v) {
 $this = $this|0;
 $v = $v|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _alGetListenerfv(4102,($v|0)); //@line 260 "modules/audio/openal/Audio.cpp"
 return; //@line 261 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio11setVelocityEPf($this,$v) {
 $this = $this|0;
 $v = $v|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _alListenerfv(4102,($v|0)); //@line 265 "modules/audio/openal/Audio.cpp"
 return; //@line 266 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio15setDopplerScaleEf($this,$scale) {
 $this = $this|0;
 $scale = +$scale;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = !($scale >= 0.0); //@line 270 "modules/audio/openal/Audio.cpp"
 if (!($0)) {
  _alDopplerFactor((+$scale)); //@line 271 "modules/audio/openal/Audio.cpp"
 }
 return; //@line 272 "modules/audio/openal/Audio.cpp"
}
function __ZNK4love5audio6openal5Audio15getDopplerScaleEv($this) {
 $this = $this|0;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_alGetFloat(49152)); //@line 276 "modules/audio/openal/Audio.cpp"
 return (+$0); //@line 276 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio6recordEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 281 "modules/audio/openal/Audio.cpp"
 $1 = ((($0)) + 132|0); //@line 281 "modules/audio/openal/Audio.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 281 "modules/audio/openal/Audio.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 281 "modules/audio/openal/Audio.cpp"
 if (!($3)) {
  return; //@line 283 "modules/audio/openal/Audio.cpp"
 }
 $4 = ((($this)) + 12|0); //@line 282 "modules/audio/openal/Audio.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 282 "modules/audio/openal/Audio.cpp"
 _alcCaptureStart(($5|0)); //@line 282 "modules/audio/openal/Audio.cpp"
 return; //@line 283 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio15getRecordedDataEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $samples = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples = sp;
 $0 = HEAP32[$this>>2]|0; //@line 287 "modules/audio/openal/Audio.cpp"
 $1 = ((($0)) + 132|0); //@line 287 "modules/audio/openal/Audio.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 287 "modules/audio/openal/Audio.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 287 "modules/audio/openal/Audio.cpp"
 if (!($3)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 297 "modules/audio/openal/Audio.cpp"
 }
 $4 = ((($this)) + 12|0); //@line 291 "modules/audio/openal/Audio.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 291 "modules/audio/openal/Audio.cpp"
 _alcGetIntegerv(($5|0),786,4,($samples|0)); //@line 291 "modules/audio/openal/Audio.cpp"
 $6 = HEAP32[$samples>>2]|0; //@line 292 "modules/audio/openal/Audio.cpp"
 $7 = $6 << 1; //@line 292 "modules/audio/openal/Audio.cpp"
 $8 = (_malloc($7)|0); //@line 292 "modules/audio/openal/Audio.cpp"
 $9 = HEAP32[$4>>2]|0; //@line 293 "modules/audio/openal/Audio.cpp"
 _alcCaptureSamples(($9|0),($8|0),($6|0)); //@line 293 "modules/audio/openal/Audio.cpp"
 $10 = (__Znwj(28)|0); //@line 294 "modules/audio/openal/Audio.cpp"
 $11 = HEAP32[$samples>>2]|0; //@line 294 "modules/audio/openal/Audio.cpp"
 __THREW__ = 0;
 invoke_viiiiii(3912,($10|0),($8|0),($11|0),8000,16,1); //@line 294 "modules/audio/openal/Audio.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if ($13) {
  $14 = ___cxa_find_matching_catch_2()|0;
  $15 = tempRet0;
  __ZdlPv($10); //@line 294 "modules/audio/openal/Audio.cpp"
  ___resumeException($14|0);
  // unreachable;
 }
 _free($8); //@line 295 "modules/audio/openal/Audio.cpp"
 $$0 = $10;
 STACKTOP = sp;return ($$0|0); //@line 297 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio13stopRecordingEb($this,$returnData) {
 $this = $this|0;
 $returnData = $returnData|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $sd$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 301 "modules/audio/openal/Audio.cpp"
 $1 = ((($0)) + 132|0); //@line 301 "modules/audio/openal/Audio.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 301 "modules/audio/openal/Audio.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 301 "modules/audio/openal/Audio.cpp"
 if (!($3)) {
  $$0 = 0;
  return ($$0|0); //@line 310 "modules/audio/openal/Audio.cpp"
 }
 if ($returnData) {
  $4 = HEAP32[$this>>2]|0; //@line 306 "modules/audio/openal/Audio.cpp"
  $5 = ((($4)) + 124|0); //@line 306 "modules/audio/openal/Audio.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 306 "modules/audio/openal/Audio.cpp"
  $7 = (FUNCTION_TABLE_ii[$6 & 8191]($this)|0); //@line 306 "modules/audio/openal/Audio.cpp"
  $sd$0 = $7;
 } else {
  $sd$0 = 0;
 }
 $8 = ((($this)) + 12|0); //@line 308 "modules/audio/openal/Audio.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 308 "modules/audio/openal/Audio.cpp"
 _alcCaptureStop(($9|0)); //@line 308 "modules/audio/openal/Audio.cpp"
 $$0 = $sd$0;
 return ($$0|0); //@line 310 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio9canRecordEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 314 "modules/audio/openal/Audio.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 314 "modules/audio/openal/Audio.cpp"
 $2 = ($1|0)!=(0|0); //@line 314 "modules/audio/openal/Audio.cpp"
 return ($2|0); //@line 314 "modules/audio/openal/Audio.cpp"
}
function __ZNK4love5audio6openal5Audio16getDistanceModelEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 319 "modules/audio/openal/Audio.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 319 "modules/audio/openal/Audio.cpp"
 return ($1|0); //@line 319 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5Audio16setDistanceModelENS0_5Audio13DistanceModelE($this,$distanceModel) {
 $this = $this|0;
 $distanceModel = $distanceModel|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 324 "modules/audio/openal/Audio.cpp"
 HEAP32[$0>>2] = $distanceModel; //@line 324 "modules/audio/openal/Audio.cpp"
 switch ($distanceModel|0) {
 case 0:  {
  _alDistanceModel(0); //@line 329 "modules/audio/openal/Audio.cpp"
  break;
 }
 case 1:  {
  _alDistanceModel(53249); //@line 333 "modules/audio/openal/Audio.cpp"
  break;
 }
 case 2:  {
  _alDistanceModel(53250); //@line 337 "modules/audio/openal/Audio.cpp"
  break;
 }
 case 3:  {
  _alDistanceModel(53251); //@line 341 "modules/audio/openal/Audio.cpp"
  break;
 }
 case 4:  {
  _alDistanceModel(53252); //@line 345 "modules/audio/openal/Audio.cpp"
  break;
 }
 case 5:  {
  _alDistanceModel(53253); //@line 349 "modules/audio/openal/Audio.cpp"
  break;
 }
 case 6:  {
  _alDistanceModel(53254); //@line 353 "modules/audio/openal/Audio.cpp"
  break;
 }
 default: {
 }
 }
 return; //@line 359 "modules/audio/openal/Audio.cpp"
}
function __ZN4love5audio6openal5AudioC2Ev($this) {
 $this = $this|0;
 var $$0 = 0, $$0$i = 0, $$01 = 0, $$01$i = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $eh$lpad$body$index6Z2D = 0, $eh$lpad$body$indexZ2D = 0, $vararg_buffer = 0, $vararg_buffer1 = 0;
 var $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./common/Module.h"
 HEAP32[$this>>2] = (22380); //@line 77 "modules/audio/openal/Audio.cpp"
 $0 = ((($this)) + 8|0); //@line 71 "modules/audio/openal/Audio.cpp"
 $1 = ((($this)) + 16|0); //@line 73 "modules/audio/openal/Audio.cpp"
 $2 = ((($this)) + 20|0); //@line 74 "modules/audio/openal/Audio.cpp"
 $3 = ((($this)) + 24|0); //@line 75 "modules/audio/openal/Audio.cpp"
 $4 = ((($this)) + 28|0); //@line 76 "modules/audio/openal/Audio.cpp"
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;HEAP32[$0+12>>2]=0|0;HEAP32[$0+16>>2]=0|0; //@line 72 "modules/audio/openal/Audio.cpp"
 HEAP32[$4>>2] = 2; //@line 76 "modules/audio/openal/Audio.cpp"
 __THREW__ = 0;
 $5 = (invoke_ii(3913,(0|0))|0); //@line 79 "modules/audio/openal/Audio.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 L1: do {
  if (!($7)) {
   HEAP32[$0>>2] = $5; //@line 79 "modules/audio/openal/Audio.cpp"
   $8 = ($5|0)==(0|0); //@line 81 "modules/audio/openal/Audio.cpp"
   if ($8) {
    $9 = (___cxa_allocate_exception(16)|0); //@line 82 "modules/audio/openal/Audio.cpp"
    __THREW__ = 0;
    invoke_viii(3349,($9|0),(364912|0),($vararg_buffer|0)); //@line 82 "modules/audio/openal/Audio.cpp"
    $10 = __THREW__; __THREW__ = 0;
    $11 = $10&1;
    if (!($11)) {
     __THREW__ = 0;
     invoke_viii(3350,($9|0),(96|0),(344|0)); //@line 82 "modules/audio/openal/Audio.cpp"
     $12 = __THREW__; __THREW__ = 0;
     break;
    }
    $15 = ___cxa_find_matching_catch_2()|0;
    $16 = tempRet0;
    ___cxa_free_exception(($9|0)); //@line 82 "modules/audio/openal/Audio.cpp"
    $$1 = $15;$$12 = $16;
    __ZN4love6ModuleD2Ev($this); //@line 128 "modules/audio/openal/Audio.cpp"
    ___resumeException($$1|0);
    // unreachable;
   }
   __THREW__ = 0;
   $17 = (invoke_iii(3914,($5|0),(0|0))|0); //@line 84 "modules/audio/openal/Audio.cpp"
   $18 = __THREW__; __THREW__ = 0;
   $19 = $18&1;
   if (!($19)) {
    HEAP32[$1>>2] = $17; //@line 84 "modules/audio/openal/Audio.cpp"
    $20 = ($17|0)==(0|0); //@line 86 "modules/audio/openal/Audio.cpp"
    if ($20) {
     $21 = (___cxa_allocate_exception(16)|0); //@line 87 "modules/audio/openal/Audio.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($21|0),(364935|0),($vararg_buffer1|0)); //@line 87 "modules/audio/openal/Audio.cpp"
     $22 = __THREW__; __THREW__ = 0;
     $23 = $22&1;
     if (!($23)) {
      __THREW__ = 0;
      invoke_viii(3350,($21|0),(96|0),(344|0)); //@line 87 "modules/audio/openal/Audio.cpp"
      $24 = __THREW__; __THREW__ = 0;
      break;
     }
     $25 = ___cxa_find_matching_catch_2()|0;
     $26 = tempRet0;
     ___cxa_free_exception(($21|0)); //@line 87 "modules/audio/openal/Audio.cpp"
     $$1 = $25;$$12 = $26;
     __ZN4love6ModuleD2Ev($this); //@line 128 "modules/audio/openal/Audio.cpp"
     ___resumeException($$1|0);
     // unreachable;
    }
    __THREW__ = 0;
    $27 = (invoke_ii(3907,($17|0))|0); //@line 89 "modules/audio/openal/Audio.cpp"
    $28 = __THREW__; __THREW__ = 0;
    $29 = $28&1;
    if (!($29)) {
     $30 = ($27<<24>>24)==(0); //@line 89 "modules/audio/openal/Audio.cpp"
     if (!($30)) {
      $31 = HEAP32[$0>>2]|0; //@line 89 "modules/audio/openal/Audio.cpp"
      __THREW__ = 0;
      $32 = (invoke_ii(3915,($31|0))|0); //@line 89 "modules/audio/openal/Audio.cpp"
      $33 = __THREW__; __THREW__ = 0;
      $34 = $33&1;
      if ($34) {
       break;
      }
      $35 = ($32|0)==(0); //@line 89 "modules/audio/openal/Audio.cpp"
      if ($35) {
       __THREW__ = 0;
       $42 = (invoke_ii(3348,300)|0); //@line 115 "modules/audio/openal/Audio.cpp"
       $43 = __THREW__; __THREW__ = 0;
       $44 = $43&1;
       do {
        if ($44) {
         $51 = ___cxa_find_matching_catch_3(96|0)|0;
         $52 = tempRet0;
         $$0 = $51;$$01 = $52;
        } else {
         __THREW__ = 0;
         invoke_vi(3916,($42|0)); //@line 115 "modules/audio/openal/Audio.cpp"
         $45 = __THREW__; __THREW__ = 0;
         $46 = $45&1;
         if ($46) {
          $53 = ___cxa_find_matching_catch_3(96|0)|0;
          $54 = tempRet0;
          __ZdlPv($42); //@line 115 "modules/audio/openal/Audio.cpp"
          $$0 = $53;$$01 = $54;
          break;
         }
         HEAP32[$2>>2] = $42; //@line 115 "modules/audio/openal/Audio.cpp"
         $47 = $42; //@line 126 "modules/audio/openal/Audio.cpp"
         __THREW__ = 0;
         $48 = (invoke_ii(3348,32)|0); //@line 126 "modules/audio/openal/Audio.cpp"
         $49 = __THREW__; __THREW__ = 0;
         $50 = $49&1;
         if ($50) {
          break L1;
         }
         __THREW__ = 0;
         invoke_vi(3917,($48|0)); //@line 35 "modules/audio/openal/Audio.cpp"
         $70 = __THREW__; __THREW__ = 0;
         $71 = $70&1;
         if ($71) {
          $86 = ___cxa_find_matching_catch_2()|0;
          $87 = tempRet0;
          $eh$lpad$body$index6Z2D = $87;$eh$lpad$body$indexZ2D = $86;
         } else {
          HEAP32[$48>>2] = (22360); //@line 38 "modules/audio/openal/Audio.cpp"
          $72 = ((($48)) + 20|0); //@line 36 "modules/audio/openal/Audio.cpp"
          HEAP32[$72>>2] = $47; //@line 36 "modules/audio/openal/Audio.cpp"
          $73 = ((($48)) + 24|0); //@line 37 "modules/audio/openal/Audio.cpp"
          HEAP8[$73>>0] = 0; //@line 37 "modules/audio/openal/Audio.cpp"
          $74 = ((($48)) + 28|0); //@line 35 "modules/audio/openal/Audio.cpp"
          __THREW__ = 0;
          invoke_vi(3788,($74|0)); //@line 35 "modules/audio/openal/Audio.cpp"
          $75 = __THREW__; __THREW__ = 0;
          $76 = $75&1;
          do {
           if ($76) {
            $80 = ___cxa_find_matching_catch_2()|0;
            $81 = tempRet0;
            $$0$i = $81;$$01$i = $80;
           } else {
            $77 = ((($48)) + 8|0); //@line 39 "modules/audio/openal/Audio.cpp"
            __THREW__ = 0;
            (invoke_iii(3360,($77|0),(364902|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $78 = __THREW__; __THREW__ = 0;
            $79 = $78&1;
            if ($79) {
             $82 = ___cxa_find_matching_catch_2()|0;
             $83 = tempRet0;
             __ZN4love6thread8MutexRefD2Ev($74); //@line 40 "modules/audio/openal/Audio.cpp"
             $$0$i = $83;$$01$i = $82;
             break;
            }
            HEAP32[$3>>2] = $48; //@line 126 "modules/audio/openal/Audio.cpp"
            __THREW__ = 0;
            (invoke_ii(3787,($48|0))|0); //@line 127 "modules/audio/openal/Audio.cpp"
            $84 = __THREW__; __THREW__ = 0;
            $85 = $84&1;
            if ($85) {
             break L1;
            }
            STACKTOP = sp;return; //@line 128 "modules/audio/openal/Audio.cpp"
           }
          } while(0);
          __ZN4love6thread10ThreadableD2Ev($48); //@line 40 "modules/audio/openal/Audio.cpp"
          $eh$lpad$body$index6Z2D = $$0$i;$eh$lpad$body$indexZ2D = $$01$i;
         }
         __ZdlPv($48); //@line 126 "modules/audio/openal/Audio.cpp"
         $$1 = $eh$lpad$body$indexZ2D;$$12 = $eh$lpad$body$index6Z2D;
         __ZN4love6ModuleD2Ev($this); //@line 128 "modules/audio/openal/Audio.cpp"
         ___resumeException($$1|0);
         // unreachable;
        }
       } while(0);
       $55 = (_llvm_eh_typeid_for((96|0))|0); //@line 116 "modules/audio/openal/Audio.cpp"
       $56 = ($$01|0)==($55|0); //@line 116 "modules/audio/openal/Audio.cpp"
       if (!($56)) {
        $$1 = $$0;$$12 = $$01;
        __ZN4love6ModuleD2Ev($this); //@line 128 "modules/audio/openal/Audio.cpp"
        ___resumeException($$1|0);
        // unreachable;
       }
       (___cxa_begin_catch(($$0|0))|0); //@line 116 "modules/audio/openal/Audio.cpp"
       __THREW__ = 0;
       (invoke_ii(3907,(0|0))|0); //@line 119 "modules/audio/openal/Audio.cpp"
       $57 = __THREW__; __THREW__ = 0;
       $58 = $57&1;
       if (!($58)) {
        $59 = HEAP32[$1>>2]|0; //@line 120 "modules/audio/openal/Audio.cpp"
        __THREW__ = 0;
        invoke_vi(3908,($59|0)); //@line 120 "modules/audio/openal/Audio.cpp"
        $60 = __THREW__; __THREW__ = 0;
        $61 = $60&1;
        if (!($61)) {
         $62 = HEAP32[$0>>2]|0; //@line 122 "modules/audio/openal/Audio.cpp"
         __THREW__ = 0;
         (invoke_ii(3909,($62|0))|0); //@line 122 "modules/audio/openal/Audio.cpp"
         $63 = __THREW__; __THREW__ = 0;
         $64 = $63&1;
         if (!($64)) {
          __THREW__ = 0;
          invoke_v(3656); //@line 123 "modules/audio/openal/Audio.cpp"
          $65 = __THREW__; __THREW__ = 0;
         }
        }
       }
       $66 = ___cxa_find_matching_catch_2()|0;
       $67 = tempRet0;
       __THREW__ = 0;
       invoke_v(3378); //@line 124 "modules/audio/openal/Audio.cpp"
       $68 = __THREW__; __THREW__ = 0;
       $69 = $68&1;
       if ($69) {
        $88 = ___cxa_find_matching_catch_3(0|0)|0;
        $89 = tempRet0;
        ___clang_call_terminate($88); //@line 124 "modules/audio/openal/Audio.cpp"
        // unreachable; //@line 124 "modules/audio/openal/Audio.cpp"
       } else {
        $$1 = $66;$$12 = $67;
        __ZN4love6ModuleD2Ev($this); //@line 128 "modules/audio/openal/Audio.cpp"
        ___resumeException($$1|0);
        // unreachable;
       }
      }
     }
     $36 = (___cxa_allocate_exception(16)|0); //@line 90 "modules/audio/openal/Audio.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($36|0),(364961|0),($vararg_buffer3|0)); //@line 90 "modules/audio/openal/Audio.cpp"
     $37 = __THREW__; __THREW__ = 0;
     $38 = $37&1;
     if (!($38)) {
      __THREW__ = 0;
      invoke_viii(3350,($36|0),(96|0),(344|0)); //@line 90 "modules/audio/openal/Audio.cpp"
      $39 = __THREW__; __THREW__ = 0;
      break;
     }
     $40 = ___cxa_find_matching_catch_2()|0;
     $41 = tempRet0;
     ___cxa_free_exception(($36|0)); //@line 90 "modules/audio/openal/Audio.cpp"
     $$1 = $40;$$12 = $41;
     __ZN4love6ModuleD2Ev($this); //@line 128 "modules/audio/openal/Audio.cpp"
     ___resumeException($$1|0);
     // unreachable;
    }
   }
  }
 } while(0);
 $13 = ___cxa_find_matching_catch_2()|0;
 $14 = tempRet0;
 $$1 = $13;$$12 = $14;
 __ZN4love6ModuleD2Ev($this); //@line 128 "modules/audio/openal/Audio.cpp"
 ___resumeException($$1|0);
 // unreachable;
}
function __ZN4love5audio6openal4PoolC2Ev($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$pre$i$i = 0, $$pre1$i$i = 0, $$pre10 = 0, $$pre11 = 0, $$pre2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $8 = 0, $9 = 0, $i$08 = 0, $i1$07 = 0, $lpad$phi$index = 0, $lpad$phi$index3 = 0, $sum$i$i$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 256|0); //@line 34 "modules/audio/openal/Pool.cpp"
 $1 = ((($this)) + 260|0); //@line 32 "modules/audio/openal/Pool.cpp"
 $2 = ((($this)) + 284|0); //@line 32 "modules/audio/openal/Pool.cpp"
 $3 = ((($this)) + 288|0); //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$3>>2] = 0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $4 = ((($this)) + 292|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$4>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ((($this)) + 288|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memset(($this|0),0,284)|0; //@line 34 "modules/audio/openal/Pool.cpp"
 HEAP32[$2>>2] = $5; //@line 1130 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $6 = ((($this)) + 296|0); //@line 32 "modules/audio/openal/Pool.cpp"
 __THREW__ = 0;
 invoke_vi(3788,($6|0)); //@line 32 "modules/audio/openal/Pool.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $11 = ___cxa_find_matching_catch_2()|0;
  $12 = tempRet0;
  $$1 = $12;$$12 = $11;
  $72 = HEAP32[$5>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIPN4love5audio6openal6SourceEjEENS_19__map_value_compareIS6_S7_NS_4lessIS6_EELb1EEENS_9allocatorIS7_EEE7destroyEPNS_11__tree_nodeIS7_PvEE($2,$72); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__15queueIjNS_5dequeIjNS_9allocatorIjEEEEED2Ev($1); //@line 72 "modules/audio/openal/Pool.cpp"
  ___resumeException($$12|0);
  // unreachable;
 }
 __THREW__ = 0;
 (invoke_i(3918)|0); //@line 37 "modules/audio/openal/Pool.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 do {
  if ($10) {
   label = 6;
  } else {
   $i$08 = 0;
   while(1) {
    $19 = (($this) + ($i$08<<2)|0); //@line 42 "modules/audio/openal/Pool.cpp"
    __THREW__ = 0;
    invoke_vii(3919,1,($19|0)); //@line 42 "modules/audio/openal/Pool.cpp"
    $20 = __THREW__; __THREW__ = 0;
    $21 = $20&1;
    if ($21) {
     label = 5;
     break;
    }
    __THREW__ = 0;
    $22 = (invoke_i(3918)|0); //@line 46 "modules/audio/openal/Pool.cpp"
    $23 = __THREW__; __THREW__ = 0;
    $24 = $23&1;
    if ($24) {
     label = 5;
     break;
    }
    $25 = ($22|0)==(0); //@line 46 "modules/audio/openal/Pool.cpp"
    $26 = HEAP32[$0>>2]|0; //@line 49 "modules/audio/openal/Pool.cpp"
    if (!($25)) {
     $31 = $26;
     break;
    }
    $27 = (($26) + 1)|0; //@line 49 "modules/audio/openal/Pool.cpp"
    HEAP32[$0>>2] = $27; //@line 49 "modules/audio/openal/Pool.cpp"
    $28 = (($i$08) + 1)|0; //@line 40 "modules/audio/openal/Pool.cpp"
    $29 = ($28|0)<(64); //@line 40 "modules/audio/openal/Pool.cpp"
    if ($29) {
     $i$08 = $28;
    } else {
     $31 = $27;
     break;
    }
   }
   if ((label|0) == 5) {
    $15 = ___cxa_find_matching_catch_2()|0;
    $16 = tempRet0;
    $lpad$phi$index = $15;$lpad$phi$index3 = $16;
    label = 7;
    break;
   }
   $30 = ($31|0)<(4); //@line 52 "modules/audio/openal/Pool.cpp"
   if ($30) {
    $36 = (___cxa_allocate_exception(16)|0); //@line 53 "modules/audio/openal/Pool.cpp"
    __THREW__ = 0;
    invoke_viii(3349,($36|0),(364993|0),($vararg_buffer|0)); //@line 53 "modules/audio/openal/Pool.cpp"
    $37 = __THREW__; __THREW__ = 0;
    $38 = $37&1;
    if ($38) {
     $40 = ___cxa_find_matching_catch_2()|0;
     $41 = tempRet0;
     ___cxa_free_exception(($36|0)); //@line 53 "modules/audio/openal/Pool.cpp"
     $$0 = $41;$$01 = $40;
     break;
    } else {
     __THREW__ = 0;
     invoke_viii(3350,($36|0),(96|0),(344|0)); //@line 53 "modules/audio/openal/Pool.cpp"
     $39 = __THREW__; __THREW__ = 0;
     label = 6;
     break;
    }
   }
   $32 = ((($this)) + 268|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $33 = ((($this)) + 264|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $34 = ((($this)) + 276|0); //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $35 = ((($this)) + 280|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $$pre10 = HEAP32[$33>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $$pre11 = HEAP32[$35>>2]|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $45 = $$pre10;$51 = $$pre11;$i1$07 = 0;
   while(1) {
    $42 = (($this) + ($i1$07<<2)|0); //@line 70 "modules/audio/openal/Pool.cpp"
    $43 = HEAP32[$32>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $44 = (($43) - ($45))|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $46 = ($43|0)==($45|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $47 = $44 << 8; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $48 = (($47) + -1)|0; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $49 = $46 ? 0 : $48; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $50 = HEAP32[$34>>2]|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $sum$i$i$i = (($51) + ($50))|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $52 = ($49|0)==($sum$i$i$i|0); //@line 1769 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $53 = $45; //@line 1769 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    if ($52) {
     __THREW__ = 0;
     invoke_vi(3920,($1|0)); //@line 1770 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
     $54 = __THREW__; __THREW__ = 0;
     $55 = $54&1;
     if ($55) {
      label = 4;
      break;
     }
     $$pre$i$i = HEAP32[$35>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
     $$pre1$i$i = HEAP32[$34>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
     $$pre2$i$i = HEAP32[$33>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $56 = $$pre2$i$i; //@line 1770 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
     $58 = $$pre$i$i;$59 = $$pre1$i$i;$62 = $$pre2$i$i;$73 = $56;
    } else {
     $58 = $51;$59 = $50;$62 = $53;$73 = $45;
    }
    $57 = (($58) + ($59))|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $60 = $57 >>> 10; //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $61 = (($62) + ($60<<2)|0); //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $63 = HEAP32[$61>>2]|0; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $64 = $57 & 1023; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $65 = (($63) + ($64<<2)|0); //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $66 = HEAP32[$42>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$65>>2] = $66; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $67 = HEAP32[$35>>2]|0; //@line 1773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $68 = (($67) + 1)|0; //@line 1773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    HEAP32[$35>>2] = $68; //@line 1773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $69 = (($i1$07) + 1)|0; //@line 60 "modules/audio/openal/Pool.cpp"
    $70 = HEAP32[$0>>2]|0; //@line 60 "modules/audio/openal/Pool.cpp"
    $71 = ($69|0)<($70|0); //@line 60 "modules/audio/openal/Pool.cpp"
    if ($71) {
     $45 = $73;$51 = $68;$i1$07 = $69;
    } else {
     label = 17;
     break;
    }
   }
   if ((label|0) == 4) {
    $13 = ___cxa_find_matching_catch_2()|0;
    $14 = tempRet0;
    $lpad$phi$index = $13;$lpad$phi$index3 = $14;
    label = 7;
    break;
   }
   else if ((label|0) == 17) {
    STACKTOP = sp;return; //@line 72 "modules/audio/openal/Pool.cpp"
   }
  }
 } while(0);
 if ((label|0) == 6) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  $lpad$phi$index = $17;$lpad$phi$index3 = $18;
  label = 7;
 }
 if ((label|0) == 7) {
  $$0 = $lpad$phi$index3;$$01 = $lpad$phi$index;
 }
 __ZN4love6thread8MutexRefD2Ev($6); //@line 72 "modules/audio/openal/Pool.cpp"
 $$1 = $$0;$$12 = $$01;
 $72 = HEAP32[$5>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeIPN4love5audio6openal6SourceEjEENS_19__map_value_compareIS6_S7_NS_4lessIS6_EELb1EEENS_9allocatorIS7_EEE7destroyEPNS_11__tree_nodeIS7_PvEE($2,$72); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__15queueIjNS_5dequeIjNS_9allocatorIjEEEEED2Ev($1); //@line 72 "modules/audio/openal/Pool.cpp"
 ___resumeException($$12|0);
 // unreachable;
}
function __ZNSt3__15dequeIjNS_9allocatorIjEEE19__add_back_capacityEv($this) {
 $this = $this|0;
 var $$0161 = 0, $$0162 = 0, $$063 = 0, $$064 = 0, $$in$i = 0, $$in$i30 = 0, $$in$i4 = 0, $$lcssa = 0, $$lcssa77 = 0, $$pre = 0, $$pre$i32 = 0, $$pre$i6 = 0, $$pre4$i10 = 0, $$pre4$i37 = 0, $$pre59 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $__buf$sroa$0$0 = 0, $__buf$sroa$0$1$lcssa = 0, $__buf$sroa$0$147 = 0, $__buf$sroa$0$147$lcssa = 0, $__buf$sroa$0$2 = 0, $__buf$sroa$0$360 = 0, $__buf$sroa$13$0 = 0, $__buf$sroa$13$0$in = 0, $__buf$sroa$13$1$lcssa = 0, $__buf$sroa$13$148 = 0, $__buf$sroa$13$2 = 0, $__buf$sroa$28$1$lcssa = 0, $__buf$sroa$28$149 = 0, $__buf$sroa$28$2 = 0, $__buf$sroa$46$0 = 0, $__buf$sroa$46$1$lcssa = 0, $__buf$sroa$46$150 = 0, $__buf$sroa$46$2 = 0, $__i$051 = 0, $__t$sroa$10$0$i = 0;
 var $__t$sroa$10$0$i29 = 0, $__t$sroa$10$0$i3 = 0, $__t$sroa$10$1$i = 0, $__t$sroa$10$1$i34 = 0, $__t$sroa$10$1$i8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 4|0;
 $1 = sp;
 $2 = ((($this)) + 16|0); //@line 1374 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $3 = HEAP32[$2>>2]|0; //@line 1374 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $4 = ($3>>>0)>(1023); //@line 2390 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($4) {
  $5 = (($3) + -1024)|0; //@line 2392 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  HEAP32[$2>>2] = $5; //@line 2392 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $6 = ((($this)) + 4|0); //@line 90 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $7 = HEAP32[$6>>2]|0; //@line 90 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $8 = HEAP32[$7>>2]|0; //@line 2393 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $9 = ((($7)) + 4|0); //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  HEAP32[$6>>2] = $9; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $10 = ((($this)) + 8|0); //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $11 = HEAP32[$10>>2]|0; //@line 562 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $12 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $13 = HEAP32[$12>>2]|0; //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $14 = ($11|0)==($13|0); //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $15 = $13; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $16 = $11; //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  do {
   if ($14) {
    $17 = HEAP32[$this>>2]|0; //@line 550 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $18 = ($9>>>0)>($17>>>0); //@line 550 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $19 = $17; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($18) {
     $20 = $9; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $21 = (($20) - ($19))|0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $22 = $21 >> 2; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $23 = (($22) + 1)|0; //@line 553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $24 = (($23|0) / -2)&-1; //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $25 = (($9) + ($24<<2)|0); //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $26 = (($16) - ($20))|0; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $27 = $26 >> 2; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     _memmove(($25|0),($9|0),($26|0))|0; //@line 1893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $28 = (($25) + ($27<<2)|0); //@line 1894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     HEAP32[$10>>2] = $28; //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $29 = HEAP32[$6>>2]|0; //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $30 = (($29) + ($24<<2)|0); //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     HEAP32[$6>>2] = $30; //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $54 = $28;
     break;
    }
    $31 = (($15) - ($19))|0; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $32 = $31 >> 1; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $33 = ($32|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $34 = $33 ? 1 : $32; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $35 = $34 >>> 2; //@line 560 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $36 = $34 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $37 = (__Znwj($36)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $38 = $37; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $39 = (($37) + ($35<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $40 = $39; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $41 = (($37) + ($34<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $42 = $41; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $43 = ($9|0)==($11|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($43) {
     $52 = $17;$__t$sroa$10$1$i8 = $40;
    } else {
     $$in$i4 = $39;$45 = $9;$__t$sroa$10$0$i3 = $40;
     while(1) {
      $44 = HEAP32[$45>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$$in$i4>>2] = $44; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $46 = $__t$sroa$10$0$i3; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $47 = ((($46)) + 4|0); //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $48 = $47; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $49 = ((($45)) + 4|0); //@line 954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $50 = ($49|0)==($11|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($50) {
       $$lcssa = $48;
       break;
      } else {
       $$in$i4 = $47;$45 = $49;$__t$sroa$10$0$i3 = $48;
      }
     }
     $$pre$i6 = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $52 = $$pre$i6;$__t$sroa$10$1$i8 = $$lcssa;
    }
    HEAP32[$this>>2] = $38; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$6>>2] = $40; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$10>>2] = $__t$sroa$10$1$i8; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$12>>2] = $42; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $51 = ($52|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $53 = $__t$sroa$10$1$i8; //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($51) {
     $54 = $53;
    } else {
     __ZdlPv($52); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $$pre4$i10 = HEAP32[$10>>2]|0; //@line 569 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $54 = $$pre4$i10;
    }
   } else {
    $54 = $11;
   }
  } while(0);
  HEAP32[$54>>2] = $8; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $55 = HEAP32[$10>>2]|0; //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $56 = ((($55)) + 4|0); //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  HEAP32[$10>>2] = $56; //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  STACKTOP = sp;return; //@line 2441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 }
 $57 = ((($this)) + 8|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $58 = HEAP32[$57>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $59 = ((($this)) + 4|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $60 = HEAP32[$59>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $61 = (($58) - ($60))|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $62 = $61 >> 2; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $63 = ((($this)) + 12|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $64 = HEAP32[$63>>2]|0; //@line 86 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $65 = HEAP32[$this>>2]|0; //@line 86 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $66 = (($64) - ($65))|0; //@line 86 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $67 = $66 >> 2; //@line 86 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $68 = ($62>>>0)<($67>>>0); //@line 2398 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $69 = $58; //@line 2398 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $70 = $60; //@line 2398 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($68) {
  $71 = ($64|0)==($58|0); //@line 2402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $72 = (__Znwj(4096)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  if (!($71)) {
   HEAP32[$0>>2] = $72; //@line 2403 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   __ZNSt3__114__split_bufferIPjNS_9allocatorIS1_EEE9push_backEOS1_($this,$0); //@line 2403 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   STACKTOP = sp;return; //@line 2441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  }
  HEAP32[$1>>2] = $72; //@line 2406 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  __ZNSt3__114__split_bufferIPjNS_9allocatorIS1_EEE10push_frontEOS1_($this,$1); //@line 2406 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $73 = HEAP32[$59>>2]|0; //@line 90 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $74 = HEAP32[$73>>2]|0; //@line 2408 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $75 = ((($73)) + 4|0); //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  HEAP32[$59>>2] = $75; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $76 = HEAP32[$57>>2]|0; //@line 562 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $77 = HEAP32[$63>>2]|0; //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $78 = ($76|0)==($77|0); //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $79 = $77; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $80 = $76; //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  do {
   if ($78) {
    $81 = HEAP32[$this>>2]|0; //@line 550 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $82 = ($75>>>0)>($81>>>0); //@line 550 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $83 = $81; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($82) {
     $84 = $75; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $85 = (($84) - ($83))|0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $86 = $85 >> 2; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $87 = (($86) + 1)|0; //@line 553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $88 = (($87|0) / -2)&-1; //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $89 = (($75) + ($88<<2)|0); //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $90 = (($80) - ($84))|0; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $91 = $90 >> 2; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     _memmove(($89|0),($75|0),($90|0))|0; //@line 1893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $92 = (($89) + ($91<<2)|0); //@line 1894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     HEAP32[$57>>2] = $92; //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $93 = HEAP32[$59>>2]|0; //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $94 = (($93) + ($88<<2)|0); //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     HEAP32[$59>>2] = $94; //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $118 = $92;
     break;
    }
    $95 = (($79) - ($83))|0; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $96 = $95 >> 1; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $97 = ($96|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $98 = $97 ? 1 : $96; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $99 = $98 >>> 2; //@line 560 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $100 = $98 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $101 = (__Znwj($100)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $102 = $101; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $103 = (($101) + ($99<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $104 = $103; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $105 = (($101) + ($98<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $106 = $105; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $107 = ($75|0)==($76|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($107) {
     $116 = $81;$__t$sroa$10$1$i34 = $104;
    } else {
     $$in$i30 = $103;$109 = $75;$__t$sroa$10$0$i29 = $104;
     while(1) {
      $108 = HEAP32[$109>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$$in$i30>>2] = $108; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $110 = $__t$sroa$10$0$i29; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $111 = ((($110)) + 4|0); //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $112 = $111; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $113 = ((($109)) + 4|0); //@line 954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $114 = ($113|0)==($76|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($114) {
       $$lcssa77 = $112;
       break;
      } else {
       $$in$i30 = $111;$109 = $113;$__t$sroa$10$0$i29 = $112;
      }
     }
     $$pre$i32 = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $116 = $$pre$i32;$__t$sroa$10$1$i34 = $$lcssa77;
    }
    HEAP32[$this>>2] = $102; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$59>>2] = $104; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$57>>2] = $__t$sroa$10$1$i34; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$63>>2] = $106; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $115 = ($116|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $117 = $__t$sroa$10$1$i34; //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($115) {
     $118 = $117;
    } else {
     __ZdlPv($116); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $$pre4$i37 = HEAP32[$57>>2]|0; //@line 569 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $118 = $$pre4$i37;
    }
   } else {
    $118 = $76;
   }
  } while(0);
  HEAP32[$118>>2] = $74; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $119 = HEAP32[$57>>2]|0; //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $120 = ((($119)) + 4|0); //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  HEAP32[$57>>2] = $120; //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  STACKTOP = sp;return; //@line 2441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 }
 $121 = $66 >> 1; //@line 2417 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $122 = ($121|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $123 = $122 ? 1 : $121; //@line 2417 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $124 = $123 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $125 = (__Znwj($124)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $126 = $125; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $127 = (($125) + ($62<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $128 = (($125) + ($123<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $129 = $128; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 __THREW__ = 0;
 $130 = (invoke_ii(3348,4096)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $131 = __THREW__; __THREW__ = 0;
 $132 = $131&1;
 L37: do {
  if ($132) {
   label = 29;
  } else {
   $133 = $130; //@line 2424 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $134 = ($62|0)==($123|0); //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   do {
    if ($134) {
     $135 = ($61|0)>(0); //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     if ($135) {
      $136 = (($62) + 1)|0; //@line 584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $137 = (($136|0) / -2)&-1; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $138 = (($127) + ($137<<2)|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $163 = $69;$164 = $70;$__buf$sroa$0$0 = $126;$__buf$sroa$13$0$in = $138;$__buf$sroa$46$0 = $129;
      break;
     }
     $139 = $61 >> 1; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $140 = ($139|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $141 = $140 ? 1 : $139; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $142 = $141 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     __THREW__ = 0;
     $143 = (invoke_ii(3348,($142|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $144 = __THREW__; __THREW__ = 0;
     $145 = $144&1;
     if ($145) {
      label = 29;
      break L37;
     }
     $146 = $141 >>> 2; //@line 591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $147 = (($143) + ($146<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $148 = (($143) + ($141<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $149 = $148; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $150 = $143; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     __ZdlPv($125); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $$pre = HEAP32[$57>>2]|0; //@line 78 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $$pre59 = HEAP32[$59>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $163 = $$pre;$164 = $$pre59;$__buf$sroa$0$0 = $150;$__buf$sroa$13$0$in = $147;$__buf$sroa$46$0 = $149;
    } else {
     $163 = $69;$164 = $70;$__buf$sroa$0$0 = $126;$__buf$sroa$13$0$in = $127;$__buf$sroa$46$0 = $129;
    }
   } while(0);
   $__buf$sroa$13$0 = $__buf$sroa$13$0$in; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   HEAP32[$__buf$sroa$13$0$in>>2] = $133; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $160 = ((($__buf$sroa$13$0$in)) + 4|0); //@line 602 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $161 = $160; //@line 602 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $162 = ($163|0)==($164|0); //@line 2434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   L46: do {
    if ($162) {
     $__buf$sroa$0$1$lcssa = $__buf$sroa$0$0;$__buf$sroa$13$1$lcssa = $__buf$sroa$13$0;$__buf$sroa$28$1$lcssa = $161;$__buf$sroa$46$1$lcssa = $__buf$sroa$46$0;
    } else {
     $__buf$sroa$0$147 = $__buf$sroa$0$0;$__buf$sroa$13$148 = $__buf$sroa$13$0;$__buf$sroa$28$149 = $161;$__buf$sroa$46$150 = $__buf$sroa$46$0;$__i$051 = $163;
     L47: while(1) {
      $167 = ((($__i$051)) + -4|0); //@line 2435 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
      $168 = $__buf$sroa$13$148; //@line 497 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $169 = $__buf$sroa$0$147; //@line 484 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $170 = ($168|0)==($169|0); //@line 484 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      do {
       if ($170) {
        $171 = $__buf$sroa$28$149; //@line 498 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $172 = $__buf$sroa$46$150; //@line 486 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $173 = ($171>>>0)<($172>>>0); //@line 486 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        if ($173) {
         $174 = (($__buf$sroa$46$150) - ($__buf$sroa$28$149))|0; //@line 488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $175 = $174 >> 2; //@line 488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $176 = (($175) + 1)|0; //@line 489 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $177 = (($176|0) / 2)&-1; //@line 489 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $178 = (($171) + ($177<<2)|0); //@line 490 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $179 = (($__buf$sroa$28$149) - ($__buf$sroa$13$148))|0; //@line 1927 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $180 = $179 >> 2; //@line 1927 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $181 = (0 - ($180))|0; //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $182 = (($178) + ($181<<2)|0); //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $183 = $__buf$sroa$13$148; //@line 1929 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         _memmove(($182|0),($183|0),($179|0))|0; //@line 1929 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $184 = $182; //@line 490 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $185 = $178; //@line 491 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $212 = $182;$__buf$sroa$0$2 = $__buf$sroa$0$147;$__buf$sroa$13$2 = $184;$__buf$sroa$28$2 = $185;$__buf$sroa$46$2 = $__buf$sroa$46$150;
         break;
        }
        $186 = (($__buf$sroa$46$150) - ($__buf$sroa$0$147))|0; //@line 495 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $187 = $186 >> 1; //@line 495 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $188 = ($187|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $189 = $188 ? 1 : $187; //@line 495 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $190 = $189 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        __THREW__ = 0;
        $191 = (invoke_ii(3348,($190|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        $192 = __THREW__; __THREW__ = 0;
        $193 = $192&1;
        if ($193) {
         $__buf$sroa$0$147$lcssa = $__buf$sroa$0$147;
         break L47;
        }
        $194 = (($189) + 3)|0; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $195 = $194 >>> 2; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $196 = $191; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $197 = (($191) + ($195<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $198 = $197; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $199 = (($191) + ($189<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $200 = $199; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $201 = ($168|0)==($171|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        if ($201) {
         $__t$sroa$10$1$i = $198;
        } else {
         $$in$i = $197;$203 = $168;$__t$sroa$10$0$i = $198;
         while(1) {
          $202 = HEAP32[$203>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          HEAP32[$$in$i>>2] = $202; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $204 = $__t$sroa$10$0$i; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
          $205 = ((($204)) + 4|0); //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
          $206 = $205; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
          $207 = ((($203)) + 4|0); //@line 954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
          $208 = ($207|0)==($171|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
          if ($208) {
           $__t$sroa$10$1$i = $206;
           break;
          } else {
           $$in$i = $205;$203 = $207;$__t$sroa$10$0$i = $206;
          }
         }
        }
        $209 = ($__buf$sroa$13$148|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        if ($209) {
         $212 = $197;$__buf$sroa$0$2 = $196;$__buf$sroa$13$2 = $198;$__buf$sroa$28$2 = $__t$sroa$10$1$i;$__buf$sroa$46$2 = $200;
        } else {
         $210 = $__buf$sroa$13$148; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         __ZdlPv($210); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
         $212 = $197;$__buf$sroa$0$2 = $196;$__buf$sroa$13$2 = $198;$__buf$sroa$28$2 = $__t$sroa$10$1$i;$__buf$sroa$46$2 = $200;
        }
       } else {
        $212 = $168;$__buf$sroa$0$2 = $__buf$sroa$0$147;$__buf$sroa$13$2 = $__buf$sroa$13$148;$__buf$sroa$28$2 = $__buf$sroa$28$149;$__buf$sroa$46$2 = $__buf$sroa$46$150;
       }
      } while(0);
      $211 = ((($212)) + -4|0); //@line 505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $213 = HEAP32[$167>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$211>>2] = $213; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $214 = $__buf$sroa$13$2; //@line 506 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $215 = ((($214)) + -4|0); //@line 506 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $216 = $215; //@line 506 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $217 = HEAP32[$59>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $218 = ($167|0)==($217|0); //@line 2434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
      if ($218) {
       $__buf$sroa$0$1$lcssa = $__buf$sroa$0$2;$__buf$sroa$13$1$lcssa = $216;$__buf$sroa$28$1$lcssa = $__buf$sroa$28$2;$__buf$sroa$46$1$lcssa = $__buf$sroa$46$2;
       break L46;
      } else {
       $__buf$sroa$0$147 = $__buf$sroa$0$2;$__buf$sroa$13$148 = $216;$__buf$sroa$28$149 = $__buf$sroa$28$2;$__buf$sroa$46$150 = $__buf$sroa$46$2;$__i$051 = $167;
      }
     }
     $219 = ___cxa_find_matching_catch_2()|0;
     $220 = tempRet0;
     $221 = ($__buf$sroa$0$147$lcssa|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     if ($221) {
      $$0162 = $219;$$064 = $220;
     } else {
      $$0161 = $219;$$063 = $220;$__buf$sroa$0$360 = $__buf$sroa$0$147$lcssa;
      break L37;
     }
     ___resumeException($$0162|0);
     // unreachable;
    }
   } while(0);
   $165 = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$this>>2] = $__buf$sroa$0$1$lcssa; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$59>>2] = $__buf$sroa$13$1$lcssa; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$57>>2] = $__buf$sroa$28$1$lcssa; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$63>>2] = $__buf$sroa$46$1$lcssa; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $166 = ($165|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($166) {
    STACKTOP = sp;return; //@line 2441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   }
   __ZdlPv($165); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   STACKTOP = sp;return; //@line 2441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  }
 } while(0);
 if ((label|0) == 29) {
  $151 = ___cxa_find_matching_catch_3(0|0)|0;
  $152 = tempRet0;
  (___cxa_begin_catch(($151|0))|0); //@line 2426 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $153 = ((($127)) + -4|0); //@line 92 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $154 = HEAP32[$153>>2]|0; //@line 2429 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  __ZdlPv($154); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  __THREW__ = 0;
  invoke_v(3656); //@line 2430 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $155 = __THREW__; __THREW__ = 0;
  $156 = ___cxa_find_matching_catch_2()|0;
  $157 = tempRet0;
  __THREW__ = 0;
  invoke_v(3378); //@line 2431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $158 = __THREW__; __THREW__ = 0;
  $159 = $158&1;
  if ($159) {
   $223 = ___cxa_find_matching_catch_3(0|0)|0;
   $224 = tempRet0;
   ___clang_call_terminate($223); //@line 2431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   // unreachable; //@line 2431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  } else {
   $$0161 = $156;$$063 = $157;$__buf$sroa$0$360 = $126;
  }
 }
 $222 = $__buf$sroa$0$360; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($222); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $$0162 = $$0161;$$064 = $$063;
 ___resumeException($$0162|0);
 // unreachable;
}
function __ZNSt3__114__split_bufferIPjNS_9allocatorIS1_EEE9push_backEOS1_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__t$sroa$10$0 = 0, $__t$sroa$10$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $1 = HEAP32[$0>>2]|0; //@line 593 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $2 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $4 = ($1|0)==($3|0); //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $5 = $3; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $6 = $1; //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 do {
  if ($4) {
   $7 = ((($this)) + 4|0); //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $8 = HEAP32[$7>>2]|0; //@line 592 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $9 = HEAP32[$this>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $10 = ($8>>>0)>($9>>>0); //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $11 = $9; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($10) {
    $12 = $8; //@line 583 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $13 = (($12) - ($11))|0; //@line 583 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $14 = $13 >> 2; //@line 583 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $15 = (($14) + 1)|0; //@line 584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $16 = (($15|0) / -2)&-1; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $17 = (($8) + ($16<<2)|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $18 = (($6) - ($12))|0; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $19 = $18 >> 2; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    _memmove(($17|0),($8|0),($18|0))|0; //@line 1893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $20 = (($17) + ($19<<2)|0); //@line 1894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    HEAP32[$0>>2] = $20; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $21 = HEAP32[$7>>2]|0; //@line 586 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $22 = (($21) + ($16<<2)|0); //@line 586 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    HEAP32[$7>>2] = $22; //@line 586 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $47 = $20;
    break;
   }
   $23 = (($5) - ($11))|0; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $24 = $23 >> 1; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $25 = ($24|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $26 = $25 ? 1 : $24; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $27 = $26 >>> 2; //@line 591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $28 = $26 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $29 = (__Znwj($28)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $30 = $29; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $31 = (($29) + ($27<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $32 = $31; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $33 = (($29) + ($26<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $34 = $33; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $35 = ($8|0)==($1|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($35) {
    $44 = $9;$__t$sroa$10$1 = $32;
   } else {
    $$in = $31;$37 = $8;$__t$sroa$10$0 = $32;
    while(1) {
     $36 = HEAP32[$37>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$$in>>2] = $36; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $38 = $__t$sroa$10$0; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $39 = ((($38)) + 4|0); //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $40 = $39; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $41 = ((($37)) + 4|0); //@line 954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $42 = ($41|0)==($1|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($42) {
      $$lcssa = $40;
      break;
     } else {
      $$in = $39;$37 = $41;$__t$sroa$10$0 = $40;
     }
    }
    $$pre = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $44 = $$pre;$__t$sroa$10$1 = $$lcssa;
   }
   HEAP32[$this>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$7>>2] = $32; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$0>>2] = $__t$sroa$10$1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$2>>2] = $34; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $43 = ($44|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $45 = $__t$sroa$10$1; //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($43) {
    $47 = $45;
   } else {
    __ZdlPv($44); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $$pre4 = HEAP32[$0>>2]|0; //@line 600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $47 = $$pre4;
   }
  } else {
   $47 = $1;
  }
 } while(0);
 $46 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$47>>2] = $46; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $48 = HEAP32[$0>>2]|0; //@line 602 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $49 = ((($48)) + 4|0); //@line 602 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 HEAP32[$0>>2] = $49; //@line 602 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 return; //@line 603 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
}
function __ZNSt3__114__split_bufferIPjNS_9allocatorIS1_EEE10push_frontEOS1_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$cast = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__t$sroa$10$0 = 0, $__t$sroa$10$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 515 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $1 = HEAP32[$0>>2]|0; //@line 528 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $2 = HEAP32[$this>>2]|0; //@line 515 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $3 = ($1|0)==($2|0); //@line 515 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $4 = $2; //@line 526 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 do {
  if ($3) {
   $5 = ((($this)) + 8|0); //@line 517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $6 = HEAP32[$5>>2]|0; //@line 529 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $7 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $8 = HEAP32[$7>>2]|0; //@line 517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $9 = ($6>>>0)<($8>>>0); //@line 517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $$cast = $8; //@line 519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($9) {
    $10 = $6; //@line 519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $11 = (($$cast) - ($10))|0; //@line 519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $12 = $11 >> 2; //@line 519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $13 = (($12) + 1)|0; //@line 520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $14 = (($13|0) / 2)&-1; //@line 520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $15 = (($6) + ($14<<2)|0); //@line 521 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $16 = $1; //@line 1927 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $17 = (($10) - ($16))|0; //@line 1927 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $18 = $17 >> 2; //@line 1927 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $19 = (0 - ($18))|0; //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $20 = (($15) + ($19<<2)|0); //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    _memmove(($20|0),($1|0),($17|0))|0; //@line 1929 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    HEAP32[$0>>2] = $20; //@line 521 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $21 = HEAP32[$5>>2]|0; //@line 522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $22 = (($21) + ($14<<2)|0); //@line 522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    HEAP32[$5>>2] = $22; //@line 522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $47 = $20;
    break;
   }
   $23 = (($$cast) - ($4))|0; //@line 526 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $24 = $23 >> 1; //@line 526 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $25 = ($24|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $26 = $25 ? 1 : $24; //@line 526 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $27 = (($26) + 3)|0; //@line 527 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $28 = $27 >>> 2; //@line 527 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $29 = $26 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $30 = (__Znwj($29)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $31 = $30; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $32 = (($30) + ($28<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $33 = $32; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $34 = (($30) + ($26<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $35 = $34; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $36 = ($1|0)==($6|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($36) {
    $45 = $1;$__t$sroa$10$1 = $33;
   } else {
    $$in = $32;$38 = $1;$__t$sroa$10$0 = $33;
    while(1) {
     $37 = HEAP32[$38>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$$in>>2] = $37; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $39 = $__t$sroa$10$0; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $40 = ((($39)) + 4|0); //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $41 = $40; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $42 = ((($38)) + 4|0); //@line 954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $43 = ($42|0)==($6|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($43) {
      $$lcssa = $41;
      break;
     } else {
      $$in = $40;$38 = $42;$__t$sroa$10$0 = $41;
     }
    }
    $$pre = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $45 = $$pre;$__t$sroa$10$1 = $$lcssa;
   }
   HEAP32[$this>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$0>>2] = $33; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$5>>2] = $__t$sroa$10$1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$7>>2] = $35; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $44 = ($45|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($44) {
    $47 = $32;
   } else {
    __ZdlPv($45); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $$pre4 = HEAP32[$0>>2]|0; //@line 536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $47 = $$pre4;
   }
  } else {
   $47 = $1;
  }
 } while(0);
 $46 = ((($47)) + -4|0); //@line 536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $48 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$46>>2] = $48; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $49 = HEAP32[$0>>2]|0; //@line 538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $50 = ((($49)) + -4|0); //@line 538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 HEAP32[$0>>2] = $50; //@line 538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 return; //@line 539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
}
function __ZNSt3__16__treeINS_12__value_typeIPN4love5audio6openal6SourceEjEENS_19__map_value_compareIS6_S7_NS_4lessIS6_EELb1EEENS_9allocatorIS7_EEE7destroyEPNS_11__tree_nodeIS7_PvEE($this,$__nd) {
 $this = $this|0;
 $__nd = $__nd|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__nd|0)==(0|0); //@line 1440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($0) {
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 } else {
  $1 = HEAP32[$__nd>>2]|0; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIPN4love5audio6openal6SourceEjEENS_19__map_value_compareIS6_S7_NS_4lessIS6_EELb1EEENS_9allocatorIS7_EEE7destroyEPNS_11__tree_nodeIS7_PvEE($this,$1); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2 = ((($__nd)) + 4|0); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3 = HEAP32[$2>>2]|0; //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIPN4love5audio6openal6SourceEjEENS_19__map_value_compareIS6_S7_NS_4lessIS6_EELb1EEENS_9allocatorIS7_EEE7destroyEPNS_11__tree_nodeIS7_PvEE($this,$3); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZdlPv($__nd); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
}
function __ZNSt3__15queueIjNS_5dequeIjNS_9allocatorIjEEEEED2Ev($this) {
 $this = $this|0;
 var $$cast$i = 0, $$in$i = 0, $$lcssa$i = 0, $$pre$i = 0, $$pre2$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $__i$01$i = 0, $__i$sroa$0$0$ph$i = 0, $__i$sroa$5$0$i = 0, $__i$sroa$5$0$ph$i = 0, $__new_last2$i$i$i$i$i = 0, $phitmp$i = 0, $scevgep$i$i$i$i$i = 0, $scevgep4$i$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $1 = HEAP32[$0>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $2 = ((($this)) + 16|0); //@line 1060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $3 = HEAP32[$2>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $4 = $3 >>> 10; //@line 1060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $5 = (($1) + ($4<<2)|0); //@line 1060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $6 = ((($this)) + 8|0); //@line 85 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $7 = HEAP32[$6>>2]|0; //@line 85 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $8 = ($7|0)==($1|0); //@line 85 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $9 = $7; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $10 = $1; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($8) {
  $11 = ((($this)) + 20|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $25 = 0;$26 = $11;$57 = 0;
 } else {
  $12 = HEAP32[$5>>2]|0; //@line 1061 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $13 = $3 & 1023; //@line 1061 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $14 = (($12) + ($13<<2)|0); //@line 1061 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $phitmp$i = $14; //@line 1061 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $15 = ((($this)) + 20|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $16 = HEAP32[$15>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $17 = (($16) + ($3))|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $18 = $17 >>> 10; //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $19 = (($1) + ($18<<2)|0); //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $20 = HEAP32[$19>>2]|0; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $21 = $17 & 1023; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $22 = (($20) + ($21<<2)|0); //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $25 = $22;$26 = $15;$57 = $phitmp$i;
 }
 $__i$sroa$0$0$ph$i = $5;$__i$sroa$5$0$ph$i = $57;
 L5: while(1) {
  $__i$sroa$5$0$i = $__i$sroa$5$0$ph$i;
  while(1) {
   $23 = $__i$sroa$5$0$i; //@line 390 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $24 = ($23|0)==($25|0); //@line 394 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   if ($24) {
    break L5;
   }
   $30 = ((($23)) + 4|0); //@line 299 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $31 = $30; //@line 299 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $32 = HEAP32[$__i$sroa$0$0$ph$i>>2]|0; //@line 299 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $33 = (($31) - ($32))|0; //@line 299 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $34 = ($33|0)==(4096); //@line 299 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   if ($34) {
    break;
   } else {
    $__i$sroa$5$0$i = $31;
   }
  }
  $35 = ((($__i$sroa$0$0$ph$i)) + 4|0); //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $36 = HEAP32[$35>>2]|0; //@line 302 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $__i$sroa$0$0$ph$i = $35;$__i$sroa$5$0$ph$i = $36;
 }
 HEAP32[$26>>2] = 0; //@line 1164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $27 = (($9) - ($10))|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27 >> 2; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = ($28>>>0)>(2); //@line 1165 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($29) {
  $$in$i = $1;
  while(1) {
   $37 = HEAP32[$$in$i>>2]|0; //@line 1167 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   __ZdlPv($37); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $38 = HEAP32[$0>>2]|0; //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $39 = ((($38)) + 4|0); //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   HEAP32[$0>>2] = $39; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $40 = HEAP32[$6>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $$cast$i = $39; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $41 = (($40) - ($$cast$i))|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $42 = $41 >> 2; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $43 = ($42>>>0)>(2); //@line 1165 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   if ($43) {
    $$in$i = $39;
   } else {
    $$lcssa$i = $42;
    break;
   }
  }
 } else {
  $$lcssa$i = $28;
 }
 switch ($$lcssa$i|0) {
 case 1:  {
  HEAP32[$2>>2] = 512; //@line 1173 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  break;
 }
 case 2:  {
  HEAP32[$2>>2] = 1024; //@line 1176 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  break;
 }
 default: {
 }
 }
 $44 = HEAP32[$0>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $45 = HEAP32[$6>>2]|0; //@line 78 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $46 = ($44|0)==($45|0); //@line 1107 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if (!($46)) {
  $__i$01$i = $44;
  while(1) {
   $47 = HEAP32[$__i$01$i>>2]|0; //@line 1108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   __ZdlPv($47); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $48 = ((($__i$01$i)) + 4|0); //@line 1107 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $49 = ($48|0)==($45|0); //@line 1107 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   if ($49) {
    break;
   } else {
    $__i$01$i = $48;
   }
  }
  $$pre$i = HEAP32[$0>>2]|0; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $$pre2$i = HEAP32[$6>>2]|0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $50 = ($$pre2$i|0)==($$pre$i|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if (!($50)) {
   $__new_last2$i$i$i$i$i = $$pre$i; //@line 136 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $scevgep$i$i$i$i$i = ((($$pre2$i)) + -4|0);
   $51 = $scevgep$i$i$i$i$i;
   $52 = (($51) - ($__new_last2$i$i$i$i$i))|0;
   $53 = $52 >>> 2;
   $54 = $53 ^ -1;
   $scevgep4$i$i$i$i$i = (($$pre2$i) + ($54<<2)|0);
   HEAP32[$6>>2] = $scevgep4$i$i$i$i$i; //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  }
 }
 $55 = HEAP32[$this>>2]|0; //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $56 = ($55|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($56) {
  return; //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/queue"
 }
 __ZdlPv($55); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/queue"
}
function __ZN4love5audio6openal4PoolD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __THREW__ = 0;
 invoke_vi(3921,($this|0)); //@line 76 "modules/audio/openal/Pool.cpp"
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 if (!($1)) {
  $2 = ((($this)) + 256|0); //@line 79 "modules/audio/openal/Pool.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 79 "modules/audio/openal/Pool.cpp"
  __THREW__ = 0;
  invoke_vii(3922,($3|0),($this|0)); //@line 79 "modules/audio/openal/Pool.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if (!($5)) {
   $6 = ((($this)) + 296|0); //@line 80 "modules/audio/openal/Pool.cpp"
   __ZN4love6thread8MutexRefD2Ev($6); //@line 80 "modules/audio/openal/Pool.cpp"
   $7 = ((($this)) + 284|0); //@line 783 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $8 = ((($this)) + 288|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $9 = HEAP32[$8>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIPN4love5audio6openal6SourceEjEENS_19__map_value_compareIS6_S7_NS_4lessIS6_EELb1EEENS_9allocatorIS7_EEE7destroyEPNS_11__tree_nodeIS7_PvEE($7,$9); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $10 = ((($this)) + 260|0); //@line 80 "modules/audio/openal/Pool.cpp"
   __ZNSt3__15queueIjNS_5dequeIjNS_9allocatorIjEEEEED2Ev($10); //@line 80 "modules/audio/openal/Pool.cpp"
   return; //@line 80 "modules/audio/openal/Pool.cpp"
  }
 }
 $11 = ___cxa_find_matching_catch_3(0|0)|0;
 $12 = tempRet0;
 $13 = ((($this)) + 296|0); //@line 80 "modules/audio/openal/Pool.cpp"
 __ZN4love6thread8MutexRefD2Ev($13); //@line 80 "modules/audio/openal/Pool.cpp"
 $14 = ((($this)) + 284|0); //@line 783 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $15 = ((($this)) + 288|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $16 = HEAP32[$15>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeIPN4love5audio6openal6SourceEjEENS_19__map_value_compareIS6_S7_NS_4lessIS6_EELb1EEENS_9allocatorIS7_EEE7destroyEPNS_11__tree_nodeIS7_PvEE($14,$16); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $17 = ((($this)) + 260|0); //@line 80 "modules/audio/openal/Pool.cpp"
 __ZNSt3__15queueIjNS_5dequeIjNS_9allocatorIjEEEEED2Ev($17); //@line 80 "modules/audio/openal/Pool.cpp"
 ___clang_call_terminate($11); //@line 80 "modules/audio/openal/Pool.cpp"
 // unreachable; //@line 80 "modules/audio/openal/Pool.cpp"
}
function __ZN4love5audio6openal4Pool4stopEv($this) {
 $this = $this|0;
 var $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$01$i$i$i = 0, $$pre$i$i = 0, $$pre1$i$i = 0, $$pre2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$02 = 0, $lock = 0, $sum$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ((($this)) + 296|0); //@line 175 "modules/audio/openal/Pool.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 175 "modules/audio/openal/Pool.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$1); //@line 175 "modules/audio/openal/Pool.cpp"
 $2 = ((($this)) + 284|0); //@line 176 "modules/audio/openal/Pool.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 911 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $4 = ((($this)) + 288|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ($3|0)==($4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($5) {
  $11 = HEAP32[$4>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIPN4love5audio6openal6SourceEjEENS_19__map_value_compareIS6_S7_NS_4lessIS6_EELb1EEENS_9allocatorIS7_EEE7destroyEPNS_11__tree_nodeIS7_PvEE($2,$11); //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $12 = ((($this)) + 292|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$12>>2] = 0; //@line 1478 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2>>2] = $4; //@line 1479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4>>2] = 0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZN4love6thread4LockD2Ev($lock); //@line 185 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 185 "modules/audio/openal/Pool.cpp"
 }
 $6 = ((($this)) + 268|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $7 = ((($this)) + 264|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $8 = ((($this)) + 276|0); //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $9 = ((($this)) + 280|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $10 = ((($this)) + 260|0); //@line 286 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/queue"
 $__begin$sroa$0$02 = $3;
 while(1) {
  $15 = ((($__begin$sroa$0$02)) + 16|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $16 = HEAP32[$15>>2]|0; //@line 178 "modules/audio/openal/Pool.cpp"
  __THREW__ = 0;
  invoke_vi(3923,($16|0)); //@line 178 "modules/audio/openal/Pool.cpp"
  $17 = __THREW__; __THREW__ = 0;
  $18 = $17&1;
  if ($18) {
   label = 4;
   break;
  }
  $19 = HEAP32[$15>>2]|0; //@line 179 "modules/audio/openal/Pool.cpp"
  __THREW__ = 0;
  invoke_vi(3924,($19|0)); //@line 179 "modules/audio/openal/Pool.cpp"
  $20 = __THREW__; __THREW__ = 0;
  $21 = $20&1;
  if ($21) {
   label = 4;
   break;
  }
  $22 = HEAP32[$15>>2]|0; //@line 180 "modules/audio/openal/Pool.cpp"
  $23 = HEAP32[$22>>2]|0; //@line 180 "modules/audio/openal/Pool.cpp"
  $24 = ((($23)) + 12|0); //@line 180 "modules/audio/openal/Pool.cpp"
  $25 = HEAP32[$24>>2]|0; //@line 180 "modules/audio/openal/Pool.cpp"
  __THREW__ = 0;
  invoke_vi($25|0,($22|0)); //@line 180 "modules/audio/openal/Pool.cpp"
  $26 = __THREW__; __THREW__ = 0;
  $27 = $26&1;
  if ($27) {
   label = 4;
   break;
  }
  $28 = ((($__begin$sroa$0$02)) + 20|0); //@line 181 "modules/audio/openal/Pool.cpp"
  $29 = HEAP32[$6>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $30 = HEAP32[$7>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $31 = (($29) - ($30))|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $32 = ($29|0)==($30|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $33 = $31 << 8; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $34 = (($33) + -1)|0; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $35 = $32 ? 0 : $34; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $36 = HEAP32[$8>>2]|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $37 = HEAP32[$9>>2]|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $sum$i$i$i = (($37) + ($36))|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $38 = ($35|0)==($sum$i$i$i|0); //@line 1769 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $39 = $30; //@line 1769 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  if ($38) {
   __THREW__ = 0;
   invoke_vi(3920,($10|0)); //@line 1770 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $40 = __THREW__; __THREW__ = 0;
   $41 = $40&1;
   if ($41) {
    label = 4;
    break;
   }
   $$pre$i$i = HEAP32[$9>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $$pre1$i$i = HEAP32[$8>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $$pre2$i$i = HEAP32[$7>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $43 = $$pre$i$i;$44 = $$pre1$i$i;$47 = $$pre2$i$i;
  } else {
   $43 = $37;$44 = $36;$47 = $39;
  }
  $42 = (($43) + ($44))|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $45 = $42 >>> 10; //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $46 = (($47) + ($45<<2)|0); //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $48 = HEAP32[$46>>2]|0; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $49 = $42 & 1023; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $50 = (($48) + ($49<<2)|0); //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $51 = HEAP32[$28>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$50>>2] = $51; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $52 = HEAP32[$9>>2]|0; //@line 1773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $53 = (($52) + 1)|0; //@line 1773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  HEAP32[$9>>2] = $53; //@line 1773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $54 = ((($__begin$sroa$0$02)) + 4|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $55 = HEAP32[$54>>2]|0; //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $56 = ($55|0)==(0|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($56) {
   $$0$i$i$i = $__begin$sroa$0$02;
   while(1) {
    $59 = ((($$0$i$i$i)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $60 = HEAP32[$59>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $61 = HEAP32[$60>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $62 = ($61|0)==($$0$i$i$i|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($62) {
     $$01$i$i$i = $60;
     break;
    } else {
     $$0$i$i$i = $60;
    }
   }
  } else {
   $$0$i$i$i$i = $55;
   while(1) {
    $57 = HEAP32[$$0$i$i$i$i>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $58 = ($57|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($58) {
     $$01$i$i$i = $$0$i$i$i$i;
     break;
    } else {
     $$0$i$i$i$i = $57;
    }
   }
  }
  $63 = ($$01$i$i$i|0)==($4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($63) {
   label = 3;
   break;
  } else {
   $__begin$sroa$0$02 = $$01$i$i$i;
  }
 }
 if ((label|0) == 3) {
  $11 = HEAP32[$4>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIPN4love5audio6openal6SourceEjEENS_19__map_value_compareIS6_S7_NS_4lessIS6_EELb1EEENS_9allocatorIS7_EEE7destroyEPNS_11__tree_nodeIS7_PvEE($2,$11); //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $12 = ((($this)) + 292|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$12>>2] = 0; //@line 1478 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2>>2] = $4; //@line 1479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4>>2] = 0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZN4love6thread4LockD2Ev($lock); //@line 185 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 185 "modules/audio/openal/Pool.cpp"
 }
 else if ((label|0) == 4) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 185 "modules/audio/openal/Pool.cpp"
  ___resumeException($13|0);
  // unreachable;
 }
}
function __ZN4love5audio6openal4Pool6updateEv($this) {
 $this = $this|0;
 var $$0$i$i$i$i = 0, $$0$i$i$i$i$i = 0, $$0$i$i$i$i$i10 = 0, $$0$i$i$i$i$i10$lcssa = 0, $$0$i$i$i$i$i2 = 0, $$0$i$i$i$i12 = 0, $$0$i$i$i$i4 = 0, $$01$i$i$i$i15 = 0, $$01$i$i$i$i7 = 0, $$pre$i$i = 0, $$pre$phiZ2D = 0, $$pre1$i$i = 0, $$pre2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $i$sroa$0$0$be$in = 0, $i$sroa$0$024 = 0, $lock = 0, $sum$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ((($this)) + 296|0); //@line 104 "modules/audio/openal/Pool.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 104 "modules/audio/openal/Pool.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$1); //@line 104 "modules/audio/openal/Pool.cpp"
 $2 = ((($this)) + 284|0); //@line 106 "modules/audio/openal/Pool.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 911 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $4 = ((($this)) + 288|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ($3|0)==($4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($5) {
  __ZN4love6thread4LockD2Ev($lock); //@line 121 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 121 "modules/audio/openal/Pool.cpp"
 }
 $6 = ((($this)) + 268|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $7 = ((($this)) + 264|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $8 = ((($this)) + 276|0); //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $9 = ((($this)) + 280|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $10 = ((($this)) + 260|0); //@line 286 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/queue"
 $11 = ((($this)) + 292|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $12 = ((($this)) + 288|0); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $i$sroa$0$024 = $3;
 L4: while(1) {
  $13 = ((($i$sroa$0$024)) + 16|0); //@line 677 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $14 = HEAP32[$13>>2]|0; //@line 110 "modules/audio/openal/Pool.cpp"
  $15 = HEAP32[$14>>2]|0; //@line 110 "modules/audio/openal/Pool.cpp"
  $16 = ((($15)) + 52|0); //@line 110 "modules/audio/openal/Pool.cpp"
  $17 = HEAP32[$16>>2]|0; //@line 110 "modules/audio/openal/Pool.cpp"
  __THREW__ = 0;
  $18 = (invoke_ii($17|0,($14|0))|0); //@line 110 "modules/audio/openal/Pool.cpp"
  $19 = __THREW__; __THREW__ = 0;
  $20 = $19&1;
  if ($20) {
   label = 22;
   break;
  }
  L7: do {
   if ($18) {
    $82 = ((($i$sroa$0$024)) + 4|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $83 = HEAP32[$82>>2]|0; //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $84 = ($83|0)==(0|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if (!($84)) {
     $$0$i$i$i$i$i = $83;
     while(1) {
      $85 = HEAP32[$$0$i$i$i$i$i>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $86 = ($85|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      if ($86) {
       $i$sroa$0$0$be$in = $$0$i$i$i$i$i;
       break L7;
      } else {
       $$0$i$i$i$i$i = $85;
      }
     }
    }
    $$0$i$i$i$i = $i$sroa$0$024;
    while(1) {
     $87 = ((($$0$i$i$i$i)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $88 = HEAP32[$87>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $89 = HEAP32[$88>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $90 = ($89|0)==($$0$i$i$i$i|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     if ($90) {
      $i$sroa$0$0$be$in = $88;
      break;
     } else {
      $$0$i$i$i$i = $88;
     }
    }
   } else {
    $21 = HEAP32[$13>>2]|0; //@line 112 "modules/audio/openal/Pool.cpp"
    __THREW__ = 0;
    invoke_vi(3923,($21|0)); //@line 112 "modules/audio/openal/Pool.cpp"
    $22 = __THREW__; __THREW__ = 0;
    $23 = $22&1;
    if ($23) {
     label = 22;
     break L4;
    }
    $24 = HEAP32[$13>>2]|0; //@line 113 "modules/audio/openal/Pool.cpp"
    __THREW__ = 0;
    invoke_vi(3924,($24|0)); //@line 113 "modules/audio/openal/Pool.cpp"
    $25 = __THREW__; __THREW__ = 0;
    $26 = $25&1;
    if ($26) {
     label = 22;
     break L4;
    }
    $27 = HEAP32[$13>>2]|0; //@line 114 "modules/audio/openal/Pool.cpp"
    $28 = HEAP32[$27>>2]|0; //@line 114 "modules/audio/openal/Pool.cpp"
    $29 = ((($28)) + 12|0); //@line 114 "modules/audio/openal/Pool.cpp"
    $30 = HEAP32[$29>>2]|0; //@line 114 "modules/audio/openal/Pool.cpp"
    __THREW__ = 0;
    invoke_vi($30|0,($27|0)); //@line 114 "modules/audio/openal/Pool.cpp"
    $31 = __THREW__; __THREW__ = 0;
    $32 = $31&1;
    if ($32) {
     label = 22;
     break L4;
    }
    $33 = ((($i$sroa$0$024)) + 20|0); //@line 115 "modules/audio/openal/Pool.cpp"
    $34 = HEAP32[$6>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $35 = HEAP32[$7>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $36 = (($34) - ($35))|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $37 = ($34|0)==($35|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $38 = $36 << 8; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $39 = (($38) + -1)|0; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $40 = $37 ? 0 : $39; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $41 = HEAP32[$8>>2]|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $42 = HEAP32[$9>>2]|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $sum$i$i$i = (($42) + ($41))|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $43 = ($40|0)==($sum$i$i$i|0); //@line 1769 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $44 = $35; //@line 1769 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    if ($43) {
     __THREW__ = 0;
     invoke_vi(3920,($10|0)); //@line 1770 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
     $45 = __THREW__; __THREW__ = 0;
     $46 = $45&1;
     if ($46) {
      label = 22;
      break L4;
     }
     $$pre$i$i = HEAP32[$9>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
     $$pre1$i$i = HEAP32[$8>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
     $$pre2$i$i = HEAP32[$7>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $48 = $$pre$i$i;$49 = $$pre1$i$i;$52 = $$pre2$i$i;
    } else {
     $48 = $42;$49 = $41;$52 = $44;
    }
    $47 = (($48) + ($49))|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $50 = $47 >>> 10; //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $51 = (($52) + ($50<<2)|0); //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $53 = HEAP32[$51>>2]|0; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $54 = $47 & 1023; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $55 = (($53) + ($54<<2)|0); //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $56 = HEAP32[$33>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$55>>2] = $56; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $57 = HEAP32[$9>>2]|0; //@line 1773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $58 = (($57) + 1)|0; //@line 1773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    HEAP32[$9>>2] = $58; //@line 1773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $59 = ((($i$sroa$0$024)) + 4|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $60 = HEAP32[$59>>2]|0; //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $61 = ($60|0)==(0|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($61) {
     $$0$i$i$i$i12 = $i$sroa$0$024;
     while(1) {
      $64 = ((($$0$i$i$i$i12)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $65 = HEAP32[$64>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $66 = HEAP32[$65>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $67 = ($66|0)==($$0$i$i$i$i12|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      if ($67) {
       $$01$i$i$i$i15 = $65;$$pre$phiZ2D = $i$sroa$0$024;
       break;
      } else {
       $$0$i$i$i$i12 = $65;
      }
     }
    } else {
     $$0$i$i$i$i$i10 = $60;
     while(1) {
      $62 = HEAP32[$$0$i$i$i$i$i10>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $63 = ($62|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      if ($63) {
       $$0$i$i$i$i$i10$lcssa = $$0$i$i$i$i$i10;
       break;
      } else {
       $$0$i$i$i$i$i10 = $62;
      }
     }
     $$01$i$i$i$i15 = $$0$i$i$i$i$i10$lcssa;$$pre$phiZ2D = $i$sroa$0$024;
    }
    if ($61) {
     $$0$i$i$i$i4 = $$pre$phiZ2D;
     while(1) {
      $70 = ((($$0$i$i$i$i4)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $71 = HEAP32[$70>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $72 = HEAP32[$71>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $73 = ($72|0)==($$0$i$i$i$i4|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      if ($73) {
       $$01$i$i$i$i7 = $71;
       break;
      } else {
       $$0$i$i$i$i4 = $71;
      }
     }
    } else {
     $$0$i$i$i$i$i2 = $60;
     while(1) {
      $68 = HEAP32[$$0$i$i$i$i$i2>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $69 = ($68|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      if ($69) {
       $$01$i$i$i$i7 = $$0$i$i$i$i$i2;
       break;
      } else {
       $$0$i$i$i$i$i2 = $68;
      }
     }
    }
    $74 = HEAP32[$2>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $75 = ($74|0)==($i$sroa$0$024|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($75) {
     $76 = $$01$i$i$i$i7; //@line 656 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     HEAP32[$2>>2] = $76; //@line 1980 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    }
    $77 = HEAP32[$11>>2]|0; //@line 1981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $78 = (($77) + -1)|0; //@line 1981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP32[$11>>2] = $78; //@line 1981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $79 = HEAP32[$12>>2]|0; //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    __ZNSt3__113__tree_removeIPNS_16__tree_node_baseIPvEEEEvT_S5_($79,$$pre$phiZ2D); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    __ZdlPv($i$sroa$0$024); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $i$sroa$0$0$be$in = $$01$i$i$i$i15;
   }
  } while(0);
  $91 = ($i$sroa$0$0$be$in|0)==($4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($91) {
   label = 28;
   break;
  } else {
   $i$sroa$0$024 = $i$sroa$0$0$be$in;
  }
 }
 if ((label|0) == 22) {
  $80 = ___cxa_find_matching_catch_2()|0;
  $81 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 121 "modules/audio/openal/Pool.cpp"
  ___resumeException($80|0);
  // unreachable;
 }
 else if ((label|0) == 28) {
  __ZN4love6thread4LockD2Ev($lock); //@line 121 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 121 "modules/audio/openal/Pool.cpp"
 }
}
function __ZNK4love5audio6openal4Pool14getSourceCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 292|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 1002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 return ($1|0); //@line 125 "modules/audio/openal/Pool.cpp"
}
function __ZNK4love5audio6openal4Pool13getMaxSourcesEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 256|0); //@line 130 "modules/audio/openal/Pool.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 130 "modules/audio/openal/Pool.cpp"
 return ($1|0); //@line 130 "modules/audio/openal/Pool.cpp"
}
function __ZN4love5audio6openal4Pool4playEPNS1_6SourceERj($this,$source,$out) {
 $this = $this|0;
 $source = $source|0;
 $out = $out|0;
 var $$0$i$i$i = 0, $$0$ph$lcssa$i$i$i$i = 0, $$0$ph6$i$i$i$i = 0, $$01$ph5$i$i$i$i = 0, $$013$i$i$i$i = 0, $$013$i$i$i$i$lcssa = 0, $$lcssa = 0, $$lcssa35 = 0, $$pre$i$i$i = 0, $$sink$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $__nd$0$i$i$i = 0, $__nd$0$i$i$i$lcssa = 0, $__nd$0$i$i$i$lcssa33 = 0, $__nd$0$i$i$i$lcssa34 = 0, $__parent$i$i = 0, $lock = 0, $ok$0$off0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__parent$i$i = sp + 4|0;
 $lock = sp;
 $0 = $source;
 $1 = ((($this)) + 296|0); //@line 135 "modules/audio/openal/Pool.cpp"
 $2 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($1)|0); //@line 135 "modules/audio/openal/Pool.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$2); //@line 135 "modules/audio/openal/Pool.cpp"
 HEAP32[$out>>2] = 0; //@line 138 "modules/audio/openal/Pool.cpp"
 $3 = ((($this)) + 288|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $4 = HEAP32[$3>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $5 = ($4|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if (!($5)) {
  $$0$ph6$i$i$i$i = $3;$$01$ph5$i$i$i$i = $4;
  L2: while(1) {
   $$013$i$i$i$i = $$01$ph5$i$i$i$i;
   while(1) {
    $6 = ((($$013$i$i$i$i)) + 16|0); //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $7 = HEAP32[$6>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $8 = ($7>>>0)<($source>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if (!($8)) {
     $$013$i$i$i$i$lcssa = $$013$i$i$i$i;
     break;
    }
    $11 = ((($$013$i$i$i$i)) + 4|0); //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $12 = HEAP32[$11>>2]|0; //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $13 = ($12|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($13) {
     $$0$ph$lcssa$i$i$i$i = $$0$ph6$i$i$i$i;
     break L2;
    } else {
     $$013$i$i$i$i = $12;
    }
   }
   $9 = HEAP32[$$013$i$i$i$i$lcssa>>2]|0; //@line 2105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $10 = ($9|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($10) {
    $$0$ph$lcssa$i$i$i$i = $$013$i$i$i$i$lcssa;
    break;
   } else {
    $$0$ph6$i$i$i$i = $$013$i$i$i$i$lcssa;$$01$ph5$i$i$i$i = $9;
   }
  }
  $14 = ($$0$ph$lcssa$i$i$i$i|0)==($3|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if (!($14)) {
   $15 = ((($$0$ph$lcssa$i$i$i$i)) + 16|0); //@line 463 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $16 = HEAP32[$15>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   $17 = ($16>>>0)>($source>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   if (!($17)) {
    $18 = ((($$0$ph$lcssa$i$i$i$i)) + 20|0); //@line 289 "modules/audio/openal/Pool.cpp"
    $19 = HEAP32[$18>>2]|0; //@line 289 "modules/audio/openal/Pool.cpp"
    HEAP32[$out>>2] = $19; //@line 289 "modules/audio/openal/Pool.cpp"
    $ok$0$off0 = 1;
    __ZN4love6thread4LockD2Ev($lock); //@line 171 "modules/audio/openal/Pool.cpp"
    STACKTOP = sp;return ($ok$0$off0|0); //@line 171 "modules/audio/openal/Pool.cpp"
   }
  }
 }
 $20 = ((($this)) + 280|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $21 = HEAP32[$20>>2]|0; //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $22 = ($21|0)==(0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($22) {
  $ok$0$off0 = 0;
  __ZN4love6thread4LockD2Ev($lock); //@line 171 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return ($ok$0$off0|0); //@line 171 "modules/audio/openal/Pool.cpp"
 }
 $23 = ((($this)) + 264|0); //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $24 = HEAP32[$23>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $25 = ((($this)) + 276|0); //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $26 = HEAP32[$25>>2]|0; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $27 = $26 >>> 10; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $28 = (($24) + ($27<<2)|0); //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $29 = HEAP32[$28>>2]|0; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $30 = $26 & 1023; //@line 1734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $31 = (($29) + ($30<<2)|0); //@line 1734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $32 = HEAP32[$31>>2]|0; //@line 148 "modules/audio/openal/Pool.cpp"
 HEAP32[$out>>2] = $32; //@line 148 "modules/audio/openal/Pool.cpp"
 $33 = HEAP32[$20>>2]|0; //@line 2538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $34 = (($33) + -1)|0; //@line 2538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 HEAP32[$20>>2] = $34; //@line 2538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $35 = HEAP32[$25>>2]|0; //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $36 = (($35) + 1)|0; //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 HEAP32[$25>>2] = $36; //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $37 = ($36>>>0)>(2047); //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($37) {
  $38 = HEAP32[$24>>2]|0; //@line 2541 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  __ZdlPv($38); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $39 = HEAP32[$23>>2]|0; //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $40 = ((($39)) + 4|0); //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  HEAP32[$23>>2] = $40; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $41 = HEAP32[$25>>2]|0; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $42 = (($41) + -1024)|0; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  HEAP32[$25>>2] = $42; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 }
 $43 = ((($this)) + 284|0); //@line 154 "modules/audio/openal/Pool.cpp"
 $44 = HEAP32[$out>>2]|0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 __THREW__ = 0;
 $45 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $46 = __THREW__; __THREW__ = 0;
 $47 = $46&1;
 if ($47) {
  $88 = ___cxa_find_matching_catch_2()|0;
  $89 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 171 "modules/audio/openal/Pool.cpp"
  ___resumeException($88|0);
  // unreachable;
 }
 $48 = ((($45)) + 16|0); //@line 1731 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$48>>2] = $0; //@line 315 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $49 = ((($45)) + 20|0); //@line 316 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 HEAP32[$49>>2] = $44; //@line 316 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $50 = HEAP32[$3>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $51 = ($50|0)==(0|0); //@line 1609 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 do {
  if ($51) {
   HEAP32[$__parent$i$i>>2] = $3; //@line 1640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $64 = $3; //@line 1641 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$0$i$i$i = $3;$69 = $64;
  } else {
   $__nd$0$i$i$i = $50;
   while(1) {
    $52 = ((($__nd$0$i$i$i)) + 16|0); //@line 457 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $53 = HEAP32[$52>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $54 = ($53>>>0)>($source>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($54) {
     $55 = HEAP32[$__nd$0$i$i$i>>2]|0; //@line 1615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $56 = ($55|0)==(0|0); //@line 1615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     if ($56) {
      $$lcssa35 = $__nd$0$i$i$i;$__nd$0$i$i$i$lcssa34 = $__nd$0$i$i$i;
      label = 16;
      break;
     } else {
      $$sink$i$i$i = $55;
     }
    } else {
     $58 = ($53>>>0)<($source>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($58)) {
      $__nd$0$i$i$i$lcssa = $__nd$0$i$i$i;
      label = 21;
      break;
     }
     $59 = ((($__nd$0$i$i$i)) + 4|0); //@line 1625 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $60 = HEAP32[$59>>2]|0; //@line 1625 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $61 = ($60|0)==(0|0); //@line 1625 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     if ($61) {
      $$lcssa = $59;$__nd$0$i$i$i$lcssa33 = $__nd$0$i$i$i;
      label = 20;
      break;
     } else {
      $$sink$i$i$i = $60;
     }
    }
    $__nd$0$i$i$i = $$sink$i$i$i;
   }
   if ((label|0) == 16) {
    HEAP32[$__parent$i$i>>2] = $__nd$0$i$i$i$lcssa34; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $57 = $__nd$0$i$i$i$lcssa34; //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $$0$i$i$i = $$lcssa35;$69 = $57;
    break;
   }
   else if ((label|0) == 20) {
    HEAP32[$__parent$i$i>>2] = $__nd$0$i$i$i$lcssa33; //@line 1629 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $62 = $__nd$0$i$i$i$lcssa33; //@line 1630 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $$0$i$i$i = $$lcssa;$69 = $62;
    break;
   }
   else if ((label|0) == 21) {
    HEAP32[$__parent$i$i>>2] = $__nd$0$i$i$i$lcssa; //@line 1635 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $63 = $__nd$0$i$i$i$lcssa; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $$0$i$i$i = $__parent$i$i;$69 = $63;
    break;
   }
  }
 } while(0);
 $65 = HEAP32[$$0$i$i$i>>2]|0; //@line 1924 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $66 = ($65|0)==(0|0); //@line 1926 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($66) {
  HEAP32[$45>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $67 = ((($45)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$67>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $68 = ((($45)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$68>>2] = $69; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i$i$i>>2] = $45; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $70 = HEAP32[$43>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $71 = HEAP32[$70>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $72 = ($71|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($72) {
   $76 = $45;
  } else {
   $73 = $71; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$43>>2] = $73; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i$i = HEAP32[$$0$i$i$i>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $76 = $$pre$i$i$i;
  }
  $74 = ((($this)) + 288|0); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $75 = HEAP32[$74>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($75,$76); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $77 = ((($this)) + 292|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $78 = HEAP32[$77>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $79 = (($78) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$77>>2] = $79; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 } else {
  __ZdlPv($45); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $80 = HEAP32[$source>>2]|0; //@line 156 "modules/audio/openal/Pool.cpp"
 $81 = ((($80)) + 8|0); //@line 156 "modules/audio/openal/Pool.cpp"
 $82 = HEAP32[$81>>2]|0; //@line 156 "modules/audio/openal/Pool.cpp"
 __THREW__ = 0;
 invoke_vi($82|0,($source|0)); //@line 156 "modules/audio/openal/Pool.cpp"
 $83 = __THREW__; __THREW__ = 0;
 $84 = $83&1;
 if ($84) {
  $88 = ___cxa_find_matching_catch_2()|0;
  $89 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 171 "modules/audio/openal/Pool.cpp"
  ___resumeException($88|0);
  // unreachable;
 }
 __THREW__ = 0;
 $85 = (invoke_ii(3925,($source|0))|0); //@line 158 "modules/audio/openal/Pool.cpp"
 $86 = __THREW__; __THREW__ = 0;
 $87 = $86&1;
 if ($87) {
  $88 = ___cxa_find_matching_catch_2()|0;
  $89 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 171 "modules/audio/openal/Pool.cpp"
  ___resumeException($88|0);
  // unreachable;
 } else {
  $ok$0$off0 = $85;
  __ZN4love6thread4LockD2Ev($lock); //@line 171 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return ($ok$0$off0|0); //@line 171 "modules/audio/openal/Pool.cpp"
 }
 return (0)|0;
}
function __ZN4love5audio6openal4Pool4stopEPNS1_6SourceE($this,$source) {
 $this = $this|0;
 $source = $source|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $lock = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ((($this)) + 296|0); //@line 189 "modules/audio/openal/Pool.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 189 "modules/audio/openal/Pool.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$1); //@line 189 "modules/audio/openal/Pool.cpp"
 __THREW__ = 0;
 (invoke_iii(3926,($this|0),($source|0))|0); //@line 190 "modules/audio/openal/Pool.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 191 "modules/audio/openal/Pool.cpp"
  ___resumeException($4|0);
  // unreachable;
 } else {
  __ZN4love6thread4LockD2Ev($lock); //@line 191 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 191 "modules/audio/openal/Pool.cpp"
 }
}
function __ZN4love5audio6openal4Pool12removeSourceEPNS1_6SourceE($this,$source) {
 $this = $this|0;
 $source = $source|0;
 var $$0 = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i$i = 0, $$0$i$i$i$i$i1 = 0, $$0$i$i$i$i3 = 0, $$0$ph$lcssa$i$i$i = 0, $$0$ph6$i$i$i = 0, $$01$i$i$i$i = 0, $$01$ph5$i$i$i = 0, $$013$i$i$i = 0, $$013$i$i$i$lcssa = 0, $$pre$i$i = 0, $$pre$phiZ2D = 0, $$pre1$i$i = 0, $$pre2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $8 = 0, $9 = 0, $sum$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 284|0); //@line 296 "modules/audio/openal/Pool.cpp"
 $1 = ((($this)) + 288|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = HEAP32[$1>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $3 = ($2|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($3) {
  $$0 = 0;
  return ($$0|0); //@line 308 "modules/audio/openal/Pool.cpp"
 } else {
  $$0$ph6$i$i$i = $1;$$01$ph5$i$i$i = $2;
 }
 L3: while(1) {
  $$013$i$i$i = $$01$ph5$i$i$i;
  while(1) {
   $4 = ((($$013$i$i$i)) + 16|0); //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5 = HEAP32[$4>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   $6 = ($5>>>0)<($source>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   if (!($6)) {
    $$013$i$i$i$lcssa = $$013$i$i$i;
    break;
   }
   $9 = ((($$013$i$i$i)) + 4|0); //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $10 = HEAP32[$9>>2]|0; //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $11 = ($10|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($11) {
    $$0$ph$lcssa$i$i$i = $$0$ph6$i$i$i;
    break L3;
   } else {
    $$013$i$i$i = $10;
   }
  }
  $7 = HEAP32[$$013$i$i$i$lcssa>>2]|0; //@line 2105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $8 = ($7|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($8) {
   $$0$ph$lcssa$i$i$i = $$013$i$i$i$lcssa;
   break;
  } else {
   $$0$ph6$i$i$i = $$013$i$i$i$lcssa;$$01$ph5$i$i$i = $7;
  }
 }
 $12 = ($$0$ph$lcssa$i$i$i|0)==($1|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($12) {
  $$0 = 0;
  return ($$0|0); //@line 308 "modules/audio/openal/Pool.cpp"
 }
 $13 = ((($$0$ph$lcssa$i$i$i)) + 16|0); //@line 463 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $14 = HEAP32[$13>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
 $15 = ($14>>>0)>($source>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
 if ($15) {
  $$0 = 0;
  return ($$0|0); //@line 308 "modules/audio/openal/Pool.cpp"
 }
 __ZN4love5audio6openal6Source10stopAtomicEv($source); //@line 300 "modules/audio/openal/Pool.cpp"
 $16 = ((($$0$ph$lcssa$i$i$i)) + 20|0); //@line 301 "modules/audio/openal/Pool.cpp"
 $17 = ((($this)) + 268|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $18 = HEAP32[$17>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $19 = ((($this)) + 264|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $20 = HEAP32[$19>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $21 = (($18) - ($20))|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $22 = ($18|0)==($20|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $23 = $21 << 8; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $24 = (($23) + -1)|0; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $25 = $22 ? 0 : $24; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $26 = ((($this)) + 276|0); //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $27 = HEAP32[$26>>2]|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $28 = ((($this)) + 280|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $29 = HEAP32[$28>>2]|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $sum$i$i$i = (($29) + ($27))|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $30 = ($25|0)==($sum$i$i$i|0); //@line 1769 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $31 = $20; //@line 1769 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($30) {
  $32 = ((($this)) + 260|0); //@line 286 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/queue"
  __ZNSt3__15dequeIjNS_9allocatorIjEEE19__add_back_capacityEv($32); //@line 1770 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $$pre$i$i = HEAP32[$28>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $$pre1$i$i = HEAP32[$26>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $$pre2$i$i = HEAP32[$19>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $34 = $$pre$i$i;$35 = $$pre1$i$i;$38 = $$pre2$i$i;
 } else {
  $34 = $29;$35 = $27;$38 = $31;
 }
 $33 = (($34) + ($35))|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $36 = $33 >>> 10; //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $37 = (($38) + ($36<<2)|0); //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $39 = HEAP32[$37>>2]|0; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $40 = $33 & 1023; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $41 = (($39) + ($40<<2)|0); //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $42 = HEAP32[$16>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$41>>2] = $42; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $43 = HEAP32[$28>>2]|0; //@line 1773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $44 = (($43) + 1)|0; //@line 1773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 HEAP32[$28>>2] = $44; //@line 1773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $45 = ((($$0$ph$lcssa$i$i$i)) + 4|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $46 = HEAP32[$45>>2]|0; //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $47 = ($46|0)==(0|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($47) {
  $$0$i$i$i$i3 = $$0$ph$lcssa$i$i$i;
  while(1) {
   $50 = ((($$0$i$i$i$i3)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $51 = HEAP32[$50>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $52 = HEAP32[$51>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $53 = ($52|0)==($$0$i$i$i$i3|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($53) {
    $$pre$phiZ2D = $$0$ph$lcssa$i$i$i;
    break;
   } else {
    $$0$i$i$i$i3 = $51;
   }
  }
 } else {
  $$0$i$i$i$i$i1 = $46;
  while(1) {
   $48 = HEAP32[$$0$i$i$i$i$i1>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $49 = ($48|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($49) {
    break;
   } else {
    $$0$i$i$i$i$i1 = $48;
   }
  }
  $$pre$phiZ2D = $$0$ph$lcssa$i$i$i;
 }
 if ($47) {
  $$0$i$i$i$i = $$pre$phiZ2D;
  while(1) {
   $56 = ((($$0$i$i$i$i)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $57 = HEAP32[$56>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $58 = HEAP32[$57>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $59 = ($58|0)==($$0$i$i$i$i|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($59) {
    $$01$i$i$i$i = $57;
    break;
   } else {
    $$0$i$i$i$i = $57;
   }
  }
 } else {
  $$0$i$i$i$i$i = $46;
  while(1) {
   $54 = HEAP32[$$0$i$i$i$i$i>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $55 = ($54|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($55) {
    $$01$i$i$i$i = $$0$i$i$i$i$i;
    break;
   } else {
    $$0$i$i$i$i$i = $54;
   }
  }
 }
 $60 = HEAP32[$0>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $61 = ($60|0)==($$0$ph$lcssa$i$i$i|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($61) {
  $62 = $$01$i$i$i$i; //@line 656 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$0>>2] = $62; //@line 1980 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
 $63 = ((($this)) + 292|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $64 = HEAP32[$63>>2]|0; //@line 1981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $65 = (($64) + -1)|0; //@line 1981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$63>>2] = $65; //@line 1981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $66 = ((($this)) + 288|0); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $67 = HEAP32[$66>>2]|0; //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__113__tree_removeIPNS_16__tree_node_baseIPvEEEEvT_S5_($67,$$pre$phiZ2D); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZdlPv($$0$ph$lcssa$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $68 = HEAP32[$source>>2]|0; //@line 303 "modules/audio/openal/Pool.cpp"
 $69 = ((($68)) + 12|0); //@line 303 "modules/audio/openal/Pool.cpp"
 $70 = HEAP32[$69>>2]|0; //@line 303 "modules/audio/openal/Pool.cpp"
 FUNCTION_TABLE_vi[$70 & 8191]($source); //@line 303 "modules/audio/openal/Pool.cpp"
 $$0 = 1;
 return ($$0|0); //@line 308 "modules/audio/openal/Pool.cpp"
}
function __ZN4love5audio6openal4Pool5pauseEv($this) {
 $this = $this|0;
 var $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$01$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$02 = 0, $lock = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ((($this)) + 296|0); //@line 195 "modules/audio/openal/Pool.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 195 "modules/audio/openal/Pool.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$1); //@line 195 "modules/audio/openal/Pool.cpp"
 $2 = ((($this)) + 284|0); //@line 854 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $3 = HEAP32[$2>>2]|0; //@line 911 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $4 = ((($this)) + 288|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ($3|0)==($4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($5) {
  __ZN4love6thread4LockD2Ev($lock); //@line 198 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 198 "modules/audio/openal/Pool.cpp"
 }
 $__begin$sroa$0$02 = $3;
 while(1) {
  $8 = ((($__begin$sroa$0$02)) + 16|0); //@line 197 "modules/audio/openal/Pool.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 197 "modules/audio/openal/Pool.cpp"
  __THREW__ = 0;
  invoke_vi(3927,($9|0)); //@line 197 "modules/audio/openal/Pool.cpp"
  $10 = __THREW__; __THREW__ = 0;
  $11 = $10&1;
  if ($11) {
   label = 4;
   break;
  }
  $12 = ((($__begin$sroa$0$02)) + 4|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $13 = HEAP32[$12>>2]|0; //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $14 = ($13|0)==(0|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($14) {
   $$0$i$i$i = $__begin$sroa$0$02;
   while(1) {
    $17 = ((($$0$i$i$i)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $18 = HEAP32[$17>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $19 = HEAP32[$18>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $20 = ($19|0)==($$0$i$i$i|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($20) {
     $$01$i$i$i = $18;
     break;
    } else {
     $$0$i$i$i = $18;
    }
   }
  } else {
   $$0$i$i$i$i = $13;
   while(1) {
    $15 = HEAP32[$$0$i$i$i$i>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $16 = ($15|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($16) {
     $$01$i$i$i = $$0$i$i$i$i;
     break;
    } else {
     $$0$i$i$i$i = $15;
    }
   }
  }
  $21 = ($$01$i$i$i|0)==($4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($21) {
   label = 3;
   break;
  } else {
   $__begin$sroa$0$02 = $$01$i$i$i;
  }
 }
 if ((label|0) == 3) {
  __ZN4love6thread4LockD2Ev($lock); //@line 198 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 198 "modules/audio/openal/Pool.cpp"
 }
 else if ((label|0) == 4) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 198 "modules/audio/openal/Pool.cpp"
  ___resumeException($6|0);
  // unreachable;
 }
}
function __ZN4love5audio6openal4Pool5pauseEPNS1_6SourceE($this,$source) {
 $this = $this|0;
 $source = $source|0;
 var $$0$ph$lcssa$i$i$i$i = 0, $$0$ph6$i$i$i$i = 0, $$01$ph5$i$i$i$i = 0, $$013$i$i$i$i = 0, $$013$i$i$i$i$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lock = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ((($this)) + 296|0); //@line 202 "modules/audio/openal/Pool.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 202 "modules/audio/openal/Pool.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$1); //@line 202 "modules/audio/openal/Pool.cpp"
 $2 = ((($this)) + 288|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $4 = ($3|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($4) {
  __ZN4love6thread4LockD2Ev($lock); //@line 206 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 206 "modules/audio/openal/Pool.cpp"
 } else {
  $$0$ph6$i$i$i$i = $2;$$01$ph5$i$i$i$i = $3;
 }
 L3: while(1) {
  $$013$i$i$i$i = $$01$ph5$i$i$i$i;
  while(1) {
   $5 = ((($$013$i$i$i$i)) + 16|0); //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $6 = HEAP32[$5>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   $7 = ($6>>>0)<($source>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   if (!($7)) {
    $$013$i$i$i$i$lcssa = $$013$i$i$i$i;
    break;
   }
   $10 = ((($$013$i$i$i$i)) + 4|0); //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $11 = HEAP32[$10>>2]|0; //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $12 = ($11|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($12) {
    $$0$ph$lcssa$i$i$i$i = $$0$ph6$i$i$i$i;
    break L3;
   } else {
    $$013$i$i$i$i = $11;
   }
  }
  $8 = HEAP32[$$013$i$i$i$i$lcssa>>2]|0; //@line 2105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $9 = ($8|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($9) {
   $$0$ph$lcssa$i$i$i$i = $$013$i$i$i$i$lcssa;
   break;
  } else {
   $$0$ph6$i$i$i$i = $$013$i$i$i$i$lcssa;$$01$ph5$i$i$i$i = $8;
  }
 }
 $13 = ($$0$ph$lcssa$i$i$i$i|0)==($2|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($13) {
  __ZN4love6thread4LockD2Ev($lock); //@line 206 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 206 "modules/audio/openal/Pool.cpp"
 }
 $14 = ((($$0$ph$lcssa$i$i$i$i)) + 16|0); //@line 463 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $15 = HEAP32[$14>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
 $16 = ($15>>>0)>($source>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
 if ($16) {
  __ZN4love6thread4LockD2Ev($lock); //@line 206 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 206 "modules/audio/openal/Pool.cpp"
 }
 __THREW__ = 0;
 invoke_vi(3927,($source|0)); //@line 205 "modules/audio/openal/Pool.cpp"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if ($18) {
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 206 "modules/audio/openal/Pool.cpp"
  ___resumeException($19|0);
  // unreachable;
 } else {
  __ZN4love6thread4LockD2Ev($lock); //@line 206 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 206 "modules/audio/openal/Pool.cpp"
 }
}
function __ZN4love5audio6openal4Pool6resumeEv($this) {
 $this = $this|0;
 var $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$01$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$02 = 0, $lock = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ((($this)) + 296|0); //@line 210 "modules/audio/openal/Pool.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 210 "modules/audio/openal/Pool.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$1); //@line 210 "modules/audio/openal/Pool.cpp"
 $2 = ((($this)) + 284|0); //@line 854 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $3 = HEAP32[$2>>2]|0; //@line 911 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $4 = ((($this)) + 288|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ($3|0)==($4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($5) {
  __ZN4love6thread4LockD2Ev($lock); //@line 213 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 213 "modules/audio/openal/Pool.cpp"
 }
 $__begin$sroa$0$02 = $3;
 while(1) {
  $8 = ((($__begin$sroa$0$02)) + 16|0); //@line 212 "modules/audio/openal/Pool.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 212 "modules/audio/openal/Pool.cpp"
  __THREW__ = 0;
  invoke_vi(3928,($9|0)); //@line 212 "modules/audio/openal/Pool.cpp"
  $10 = __THREW__; __THREW__ = 0;
  $11 = $10&1;
  if ($11) {
   label = 4;
   break;
  }
  $12 = ((($__begin$sroa$0$02)) + 4|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $13 = HEAP32[$12>>2]|0; //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $14 = ($13|0)==(0|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($14) {
   $$0$i$i$i = $__begin$sroa$0$02;
   while(1) {
    $17 = ((($$0$i$i$i)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $18 = HEAP32[$17>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $19 = HEAP32[$18>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $20 = ($19|0)==($$0$i$i$i|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($20) {
     $$01$i$i$i = $18;
     break;
    } else {
     $$0$i$i$i = $18;
    }
   }
  } else {
   $$0$i$i$i$i = $13;
   while(1) {
    $15 = HEAP32[$$0$i$i$i$i>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $16 = ($15|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($16) {
     $$01$i$i$i = $$0$i$i$i$i;
     break;
    } else {
     $$0$i$i$i$i = $15;
    }
   }
  }
  $21 = ($$01$i$i$i|0)==($4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($21) {
   label = 3;
   break;
  } else {
   $__begin$sroa$0$02 = $$01$i$i$i;
  }
 }
 if ((label|0) == 3) {
  __ZN4love6thread4LockD2Ev($lock); //@line 213 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 213 "modules/audio/openal/Pool.cpp"
 }
 else if ((label|0) == 4) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 213 "modules/audio/openal/Pool.cpp"
  ___resumeException($6|0);
  // unreachable;
 }
}
function __ZN4love5audio6openal4Pool6resumeEPNS1_6SourceE($this,$source) {
 $this = $this|0;
 $source = $source|0;
 var $$0$ph$lcssa$i$i$i$i = 0, $$0$ph6$i$i$i$i = 0, $$01$ph5$i$i$i$i = 0, $$013$i$i$i$i = 0, $$013$i$i$i$i$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lock = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ((($this)) + 296|0); //@line 217 "modules/audio/openal/Pool.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 217 "modules/audio/openal/Pool.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$1); //@line 217 "modules/audio/openal/Pool.cpp"
 $2 = ((($this)) + 288|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $4 = ($3|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($4) {
  __ZN4love6thread4LockD2Ev($lock); //@line 221 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 221 "modules/audio/openal/Pool.cpp"
 } else {
  $$0$ph6$i$i$i$i = $2;$$01$ph5$i$i$i$i = $3;
 }
 L3: while(1) {
  $$013$i$i$i$i = $$01$ph5$i$i$i$i;
  while(1) {
   $5 = ((($$013$i$i$i$i)) + 16|0); //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $6 = HEAP32[$5>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   $7 = ($6>>>0)<($source>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   if (!($7)) {
    $$013$i$i$i$i$lcssa = $$013$i$i$i$i;
    break;
   }
   $10 = ((($$013$i$i$i$i)) + 4|0); //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $11 = HEAP32[$10>>2]|0; //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $12 = ($11|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($12) {
    $$0$ph$lcssa$i$i$i$i = $$0$ph6$i$i$i$i;
    break L3;
   } else {
    $$013$i$i$i$i = $11;
   }
  }
  $8 = HEAP32[$$013$i$i$i$i$lcssa>>2]|0; //@line 2105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $9 = ($8|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($9) {
   $$0$ph$lcssa$i$i$i$i = $$013$i$i$i$i$lcssa;
   break;
  } else {
   $$0$ph6$i$i$i$i = $$013$i$i$i$i$lcssa;$$01$ph5$i$i$i$i = $8;
  }
 }
 $13 = ($$0$ph$lcssa$i$i$i$i|0)==($2|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($13) {
  __ZN4love6thread4LockD2Ev($lock); //@line 221 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 221 "modules/audio/openal/Pool.cpp"
 }
 $14 = ((($$0$ph$lcssa$i$i$i$i)) + 16|0); //@line 463 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $15 = HEAP32[$14>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
 $16 = ($15>>>0)>($source>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
 if ($16) {
  __ZN4love6thread4LockD2Ev($lock); //@line 221 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 221 "modules/audio/openal/Pool.cpp"
 }
 __THREW__ = 0;
 invoke_vi(3928,($source|0)); //@line 220 "modules/audio/openal/Pool.cpp"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if ($18) {
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 221 "modules/audio/openal/Pool.cpp"
  ___resumeException($19|0);
  // unreachable;
 } else {
  __ZN4love6thread4LockD2Ev($lock); //@line 221 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 221 "modules/audio/openal/Pool.cpp"
 }
}
function __ZN4love5audio6openal4Pool6rewindEv($this) {
 $this = $this|0;
 var $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$01$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$02 = 0, $lock = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ((($this)) + 296|0); //@line 225 "modules/audio/openal/Pool.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 225 "modules/audio/openal/Pool.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$1); //@line 225 "modules/audio/openal/Pool.cpp"
 $2 = ((($this)) + 284|0); //@line 854 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $3 = HEAP32[$2>>2]|0; //@line 911 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $4 = ((($this)) + 288|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ($3|0)==($4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($5) {
  __ZN4love6thread4LockD2Ev($lock); //@line 228 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 228 "modules/audio/openal/Pool.cpp"
 }
 $__begin$sroa$0$02 = $3;
 while(1) {
  $8 = ((($__begin$sroa$0$02)) + 16|0); //@line 227 "modules/audio/openal/Pool.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 227 "modules/audio/openal/Pool.cpp"
  __THREW__ = 0;
  invoke_vi(3924,($9|0)); //@line 227 "modules/audio/openal/Pool.cpp"
  $10 = __THREW__; __THREW__ = 0;
  $11 = $10&1;
  if ($11) {
   label = 4;
   break;
  }
  $12 = ((($__begin$sroa$0$02)) + 4|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $13 = HEAP32[$12>>2]|0; //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $14 = ($13|0)==(0|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($14) {
   $$0$i$i$i = $__begin$sroa$0$02;
   while(1) {
    $17 = ((($$0$i$i$i)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $18 = HEAP32[$17>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $19 = HEAP32[$18>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $20 = ($19|0)==($$0$i$i$i|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($20) {
     $$01$i$i$i = $18;
     break;
    } else {
     $$0$i$i$i = $18;
    }
   }
  } else {
   $$0$i$i$i$i = $13;
   while(1) {
    $15 = HEAP32[$$0$i$i$i$i>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $16 = ($15|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($16) {
     $$01$i$i$i = $$0$i$i$i$i;
     break;
    } else {
     $$0$i$i$i$i = $15;
    }
   }
  }
  $21 = ($$01$i$i$i|0)==($4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($21) {
   label = 3;
   break;
  } else {
   $__begin$sroa$0$02 = $$01$i$i$i;
  }
 }
 if ((label|0) == 3) {
  __ZN4love6thread4LockD2Ev($lock); //@line 228 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 228 "modules/audio/openal/Pool.cpp"
 }
 else if ((label|0) == 4) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 228 "modules/audio/openal/Pool.cpp"
  ___resumeException($6|0);
  // unreachable;
 }
}
function __ZN4love5audio6openal4Pool10softRewindEPNS1_6SourceE($this,$source) {
 $this = $this|0;
 $source = $source|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $lock = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ((($this)) + 296|0); //@line 233 "modules/audio/openal/Pool.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 233 "modules/audio/openal/Pool.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$1); //@line 233 "modules/audio/openal/Pool.cpp"
 __THREW__ = 0;
 invoke_vi(3924,($source|0)); //@line 234 "modules/audio/openal/Pool.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 235 "modules/audio/openal/Pool.cpp"
  ___resumeException($4|0);
  // unreachable;
 } else {
  __ZN4love6thread4LockD2Ev($lock); //@line 235 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 235 "modules/audio/openal/Pool.cpp"
 }
}
function __ZN4love5audio6openal4Pool6rewindEPNS1_6SourceE($this,$source) {
 $this = $this|0;
 $source = $source|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $lock = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ((($this)) + 296|0); //@line 239 "modules/audio/openal/Pool.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 239 "modules/audio/openal/Pool.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$1); //@line 239 "modules/audio/openal/Pool.cpp"
 __THREW__ = 0;
 invoke_vi(3924,($source|0)); //@line 240 "modules/audio/openal/Pool.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 241 "modules/audio/openal/Pool.cpp"
  ___resumeException($4|0);
  // unreachable;
 } else {
  __ZN4love6thread4LockD2Ev($lock); //@line 241 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 241 "modules/audio/openal/Pool.cpp"
 }
}
function __ZN4love5audio6openal4Pool4seekEPNS1_6SourceEfPv($this,$source,$offset,$unit) {
 $this = $this|0;
 $source = $source|0;
 $offset = +$offset;
 $unit = $unit|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $lock = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ((($this)) + 296|0); //@line 256 "modules/audio/openal/Pool.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 256 "modules/audio/openal/Pool.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$1); //@line 256 "modules/audio/openal/Pool.cpp"
 $2 = HEAP32[$source>>2]|0; //@line 257 "modules/audio/openal/Pool.cpp"
 $3 = ((($2)) + 180|0); //@line 257 "modules/audio/openal/Pool.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 257 "modules/audio/openal/Pool.cpp"
 __THREW__ = 0;
 invoke_vidi($4|0,($source|0),(+$offset),($unit|0)); //@line 257 "modules/audio/openal/Pool.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 258 "modules/audio/openal/Pool.cpp"
  ___resumeException($7|0);
  // unreachable;
 } else {
  __ZN4love6thread4LockD2Ev($lock); //@line 258 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return; //@line 258 "modules/audio/openal/Pool.cpp"
 }
}
function __ZN4love5audio6openal4Pool4tellEPNS1_6SourceEPv($this,$source,$unit) {
 $this = $this|0;
 $source = $source|0;
 $unit = $unit|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lock = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ((($this)) + 296|0); //@line 262 "modules/audio/openal/Pool.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 262 "modules/audio/openal/Pool.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$1); //@line 262 "modules/audio/openal/Pool.cpp"
 $2 = HEAP32[$source>>2]|0; //@line 263 "modules/audio/openal/Pool.cpp"
 $3 = ((($2)) + 184|0); //@line 263 "modules/audio/openal/Pool.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 263 "modules/audio/openal/Pool.cpp"
 __THREW__ = 0;
 $5 = (+invoke_dii($4|0,($source|0),($unit|0))); //@line 263 "modules/audio/openal/Pool.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $8 = ___cxa_find_matching_catch_2()|0;
  $9 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 264 "modules/audio/openal/Pool.cpp"
  ___resumeException($8|0);
  // unreachable;
 } else {
  __ZN4love6thread4LockD2Ev($lock); //@line 264 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return (+$5); //@line 264 "modules/audio/openal/Pool.cpp"
 }
 return +(0.0);
}
function __ZN4love5audio6openal4Pool11getDurationEPNS1_6SourceEPv($this,$source,$unit) {
 $this = $this|0;
 $source = $source|0;
 $unit = $unit|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lock = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ((($this)) + 296|0); //@line 268 "modules/audio/openal/Pool.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 268 "modules/audio/openal/Pool.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$1); //@line 268 "modules/audio/openal/Pool.cpp"
 $2 = HEAP32[$source>>2]|0; //@line 269 "modules/audio/openal/Pool.cpp"
 $3 = ((($2)) + 188|0); //@line 269 "modules/audio/openal/Pool.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 269 "modules/audio/openal/Pool.cpp"
 __THREW__ = 0;
 $5 = (+invoke_dii($4|0,($source|0),($unit|0))); //@line 269 "modules/audio/openal/Pool.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $8 = ___cxa_find_matching_catch_2()|0;
  $9 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 270 "modules/audio/openal/Pool.cpp"
  ___resumeException($8|0);
  // unreachable;
 } else {
  __ZN4love6thread4LockD2Ev($lock); //@line 270 "modules/audio/openal/Pool.cpp"
  STACKTOP = sp;return (+$5); //@line 270 "modules/audio/openal/Pool.cpp"
 }
 return +(0.0);
}
function __ZN4love5audio6openal16StaticDataBufferD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (22532); //@line 75 "modules/audio/openal/Source.cpp"
 $0 = ((($this)) + 8|0); //@line 76 "modules/audio/openal/Source.cpp"
 __THREW__ = 0;
 invoke_vii(3929,1,($0|0)); //@line 76 "modules/audio/openal/Source.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_3(0|0)|0;
  $4 = tempRet0;
  ___clang_call_terminate($3); //@line 77 "modules/audio/openal/Source.cpp"
  // unreachable; //@line 77 "modules/audio/openal/Source.cpp"
 } else {
  return; //@line 77 "modules/audio/openal/Source.cpp"
 }
}
function __ZN4love5audio6openal16StaticDataBufferD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (22532); //@line 75 "modules/audio/openal/Source.cpp"
 $0 = ((($this)) + 8|0); //@line 76 "modules/audio/openal/Source.cpp"
 __THREW__ = 0;
 invoke_vii(3929,1,($0|0)); //@line 76 "modules/audio/openal/Source.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_3(0|0)|0;
  $4 = tempRet0;
  ___clang_call_terminate($3); //@line 77 "modules/audio/openal/Source.cpp"
  // unreachable; //@line 77 "modules/audio/openal/Source.cpp"
 } else {
  __ZdlPv($this); //@line 75 "modules/audio/openal/Source.cpp"
  return; //@line 77 "modules/audio/openal/Source.cpp"
 }
}
function __ZN4love5audio6openal6SourceD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (22556); //@line 197 "modules/audio/openal/Source.cpp"
 $0 = ((($this)) + 20|0); //@line 198 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 198 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 198 "modules/audio/openal/Source.cpp"
 if ($2) {
  label = 10;
 } else {
  $3 = ((($this)) + 12|0); //@line 199 "modules/audio/openal/Source.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 199 "modules/audio/openal/Source.cpp"
  __THREW__ = 0;
  invoke_vii(3930,($4|0),($this|0)); //@line 199 "modules/audio/openal/Source.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   label = 10;
  }
 }
 do {
  if ((label|0) == 10) {
   $29 = ((($this)) + 8|0); //@line 201 "modules/audio/openal/Source.cpp"
   $30 = HEAP32[$29>>2]|0; //@line 201 "modules/audio/openal/Source.cpp"
   $31 = ($30|0)==(1); //@line 201 "modules/audio/openal/Source.cpp"
   if ($31) {
    $32 = ((($this)) + 24|0); //@line 202 "modules/audio/openal/Source.cpp"
    __THREW__ = 0;
    invoke_vii(3929,8,($32|0)); //@line 202 "modules/audio/openal/Source.cpp"
    $33 = __THREW__; __THREW__ = 0;
    $34 = $33&1;
    if ($34) {
     break;
    }
   }
   $35 = ((($this)) + 160|0); //@line 106 "./common/Object.h"
   $36 = HEAP32[$35>>2]|0; //@line 106 "./common/Object.h"
   $37 = ($36|0)==(0|0); //@line 106 "./common/Object.h"
   if (!($37)) {
    $38 = HEAP32[$36>>2]|0; //@line 106 "./common/Object.h"
    $39 = ((($38)) + 12|0); //@line 106 "./common/Object.h"
    $40 = HEAP32[$39>>2]|0; //@line 106 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($40|0,($36|0)); //@line 106 "./common/Object.h"
    $41 = __THREW__; __THREW__ = 0;
    $42 = $41&1;
    if ($42) {
     $43 = ___cxa_find_matching_catch_3(0|0)|0;
     $44 = tempRet0;
     ___clang_call_terminate($43); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   }
   $45 = ((($this)) + 56|0); //@line 106 "./common/Object.h"
   $46 = HEAP32[$45>>2]|0; //@line 106 "./common/Object.h"
   $47 = ($46|0)==(0|0); //@line 106 "./common/Object.h"
   if ($47) {
    __ZN4love5audio6SourceD2Ev($this); //@line 203 "modules/audio/openal/Source.cpp"
    return; //@line 203 "modules/audio/openal/Source.cpp"
   }
   $48 = HEAP32[$46>>2]|0; //@line 106 "./common/Object.h"
   $49 = ((($48)) + 12|0); //@line 106 "./common/Object.h"
   $50 = HEAP32[$49>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($50|0,($46|0)); //@line 106 "./common/Object.h"
   $51 = __THREW__; __THREW__ = 0;
   $52 = $51&1;
   if ($52) {
    $53 = ___cxa_find_matching_catch_3(0|0)|0;
    $54 = tempRet0;
    ___clang_call_terminate($53); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   } else {
    __ZN4love5audio6SourceD2Ev($this); //@line 203 "modules/audio/openal/Source.cpp"
    return; //@line 203 "modules/audio/openal/Source.cpp"
   }
  }
 } while(0);
 $7 = ___cxa_find_matching_catch_3(0|0)|0;
 $8 = tempRet0;
 $9 = ((($this)) + 160|0); //@line 106 "./common/Object.h"
 $10 = HEAP32[$9>>2]|0; //@line 106 "./common/Object.h"
 $11 = ($10|0)==(0|0); //@line 106 "./common/Object.h"
 if (!($11)) {
  $12 = HEAP32[$10>>2]|0; //@line 106 "./common/Object.h"
  $13 = ((($12)) + 12|0); //@line 106 "./common/Object.h"
  $14 = HEAP32[$13>>2]|0; //@line 106 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($14|0,($10|0)); //@line 106 "./common/Object.h"
  $15 = __THREW__; __THREW__ = 0;
  $16 = $15&1;
  if ($16) {
   $17 = ___cxa_find_matching_catch_3(0|0)|0;
   $18 = tempRet0;
   ___clang_call_terminate($17); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 }
 $19 = ((($this)) + 56|0); //@line 106 "./common/Object.h"
 $20 = HEAP32[$19>>2]|0; //@line 106 "./common/Object.h"
 $21 = ($20|0)==(0|0); //@line 106 "./common/Object.h"
 if ($21) {
  __ZN4love5audio6SourceD2Ev($this); //@line 203 "modules/audio/openal/Source.cpp"
  ___clang_call_terminate($7); //@line 203 "modules/audio/openal/Source.cpp"
  // unreachable; //@line 203 "modules/audio/openal/Source.cpp"
 }
 $22 = HEAP32[$20>>2]|0; //@line 106 "./common/Object.h"
 $23 = ((($22)) + 12|0); //@line 106 "./common/Object.h"
 $24 = HEAP32[$23>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($24|0,($20|0)); //@line 106 "./common/Object.h"
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if ($26) {
  $27 = ___cxa_find_matching_catch_3(0|0)|0;
  $28 = tempRet0;
  ___clang_call_terminate($27); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  __ZN4love5audio6SourceD2Ev($this); //@line 203 "modules/audio/openal/Source.cpp"
  ___clang_call_terminate($7); //@line 203 "modules/audio/openal/Source.cpp"
  // unreachable; //@line 203 "modules/audio/openal/Source.cpp"
 }
}
function __ZN4love5audio6openal6SourceD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5audio6openal6SourceD2Ev($this); //@line 197 "modules/audio/openal/Source.cpp"
 __ZdlPv($this); //@line 197 "modules/audio/openal/Source.cpp"
 return; //@line 203 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source5cloneEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(168)|0); //@line 207 "modules/audio/openal/Source.cpp"
 __THREW__ = 0;
 invoke_vii(3931,($0|0),($this|0)); //@line 207 "modules/audio/openal/Source.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 207 "modules/audio/openal/Source.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 207 "modules/audio/openal/Source.cpp"
 }
 return (0)|0;
}
function __ZN4love5audio6openal6SourceC2ERKS2_($this,$s) {
 $this = $this|0;
 $s = $s|0;
 var $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($s)) + 8|0); //@line 156 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 156 "modules/audio/openal/Source.cpp"
 __ZN4love5audio6SourceC2ENS1_4TypeE($this,$1); //@line 156 "modules/audio/openal/Source.cpp"
 HEAP32[$this>>2] = (22556); //@line 178 "modules/audio/openal/Source.cpp"
 $2 = ((($this)) + 12|0); //@line 157 "modules/audio/openal/Source.cpp"
 $3 = ((($s)) + 12|0); //@line 157 "modules/audio/openal/Source.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 157 "modules/audio/openal/Source.cpp"
 HEAP32[$2>>2] = $4; //@line 157 "modules/audio/openal/Source.cpp"
 $5 = ((($this)) + 20|0); //@line 158 "modules/audio/openal/Source.cpp"
 HEAP8[$5>>0] = 0; //@line 158 "modules/audio/openal/Source.cpp"
 $6 = ((($this)) + 56|0); //@line 99 "./common/Object.h"
 $7 = ((($s)) + 56|0); //@line 139 "./common/Object.h"
 $8 = HEAP32[$7>>2]|0; //@line 139 "./common/Object.h"
 HEAP32[$6>>2] = $8; //@line 99 "./common/Object.h"
 $9 = ($8|0)==(0|0); //@line 101 "./common/Object.h"
 if (!($9)) {
  $10 = HEAP32[$8>>2]|0; //@line 101 "./common/Object.h"
  $11 = ((($10)) + 8|0); //@line 101 "./common/Object.h"
  $12 = HEAP32[$11>>2]|0; //@line 101 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($12|0,($8|0)); //@line 101 "./common/Object.h"
  $13 = __THREW__; __THREW__ = 0;
  $14 = $13&1;
  if ($14) {
   $89 = ___cxa_find_matching_catch_2()|0;
   $90 = tempRet0;
   $$1 = $89;$$12 = $90;
   __ZN4love5audio6SourceD2Ev($this); //@line 194 "modules/audio/openal/Source.cpp"
   ___resumeException($$1|0);
   // unreachable;
  }
 }
 $15 = ((($this)) + 60|0); //@line 160 "modules/audio/openal/Source.cpp"
 $16 = ((($s)) + 60|0); //@line 160 "modules/audio/openal/Source.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 160 "modules/audio/openal/Source.cpp"
 HEAP32[$15>>2] = $17; //@line 160 "modules/audio/openal/Source.cpp"
 $18 = ((($this)) + 64|0); //@line 161 "modules/audio/openal/Source.cpp"
 $19 = ((($s)) + 64|0); //@line 161 "modules/audio/openal/Source.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 161 "modules/audio/openal/Source.cpp"
 HEAP32[$18>>2] = $20; //@line 161 "modules/audio/openal/Source.cpp"
 $21 = ((($this)) + 104|0); //@line 162 "modules/audio/openal/Source.cpp"
 $22 = ((($s)) + 104|0); //@line 162 "modules/audio/openal/Source.cpp"
 $23 = HEAP16[$22>>1]|0; //@line 162 "modules/audio/openal/Source.cpp"
 $24 = $23&255; //@line 162 "modules/audio/openal/Source.cpp"
 HEAP8[$21>>0] = $24; //@line 162 "modules/audio/openal/Source.cpp"
 $25 = ((($this)) + 105|0); //@line 163 "modules/audio/openal/Source.cpp"
 $26 = ($23&65535) >>> 8; //@line 163 "modules/audio/openal/Source.cpp"
 $27 = $26&255; //@line 163 "modules/audio/openal/Source.cpp"
 HEAP8[$25>>0] = $27; //@line 163 "modules/audio/openal/Source.cpp"
 $28 = ((($this)) + 106|0); //@line 164 "modules/audio/openal/Source.cpp"
 HEAP8[$28>>0] = 0; //@line 164 "modules/audio/openal/Source.cpp"
 $29 = ((($this)) + 108|0); //@line 165 "modules/audio/openal/Source.cpp"
 $30 = ((($s)) + 108|0); //@line 165 "modules/audio/openal/Source.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 165 "modules/audio/openal/Source.cpp"
 HEAP32[$29>>2] = $31; //@line 165 "modules/audio/openal/Source.cpp"
 $32 = ((($this)) + 112|0); //@line 166 "modules/audio/openal/Source.cpp"
 $33 = ((($s)) + 112|0); //@line 166 "modules/audio/openal/Source.cpp"
 $34 = HEAP32[$33>>2]|0; //@line 166 "modules/audio/openal/Source.cpp"
 HEAP32[$32>>2] = $34; //@line 166 "modules/audio/openal/Source.cpp"
 $35 = ((($this)) + 116|0); //@line 167 "modules/audio/openal/Source.cpp"
 $36 = ((($s)) + 116|0); //@line 167 "modules/audio/openal/Source.cpp"
 $37 = HEAP32[$36>>2]|0; //@line 167 "modules/audio/openal/Source.cpp"
 HEAP32[$35>>2] = $37; //@line 167 "modules/audio/openal/Source.cpp"
 $38 = ((($this)) + 120|0); //@line 168 "modules/audio/openal/Source.cpp"
 $39 = ((($s)) + 120|0); //@line 168 "modules/audio/openal/Source.cpp"
 $40 = HEAP32[$39>>2]|0; //@line 168 "modules/audio/openal/Source.cpp"
 HEAP32[$38>>2] = $40; //@line 168 "modules/audio/openal/Source.cpp"
 $41 = ((($this)) + 124|0); //@line 169 "modules/audio/openal/Source.cpp"
 $42 = ((($s)) + 124|0); //@line 169 "modules/audio/openal/Source.cpp"
 $43 = HEAP32[$42>>2]|0; //@line 169 "modules/audio/openal/Source.cpp"
 HEAP32[$41>>2] = $43; //@line 169 "modules/audio/openal/Source.cpp"
 $44 = ((($this)) + 128|0); //@line 170 "modules/audio/openal/Source.cpp"
 $45 = ((($s)) + 128|0); //@line 170 "modules/audio/openal/Source.cpp"
 ;HEAP32[$44>>2]=HEAP32[$45>>2]|0;HEAP32[$44+4>>2]=HEAP32[$45+4>>2]|0;HEAP32[$44+8>>2]=HEAP32[$45+8>>2]|0; //@line 170 "modules/audio/openal/Source.cpp"
 $46 = ((($this)) + 140|0); //@line 171 "modules/audio/openal/Source.cpp"
 HEAPF32[$46>>2] = 0.0; //@line 171 "modules/audio/openal/Source.cpp"
 $47 = ((($this)) + 144|0); //@line 172 "modules/audio/openal/Source.cpp"
 HEAPF32[$47>>2] = 0.0; //@line 172 "modules/audio/openal/Source.cpp"
 $48 = ((($this)) + 148|0); //@line 173 "modules/audio/openal/Source.cpp"
 $49 = ((($s)) + 148|0); //@line 173 "modules/audio/openal/Source.cpp"
 $50 = HEAP32[$49>>2]|0; //@line 173 "modules/audio/openal/Source.cpp"
 HEAP32[$48>>2] = $50; //@line 173 "modules/audio/openal/Source.cpp"
 $51 = ((($this)) + 152|0); //@line 174 "modules/audio/openal/Source.cpp"
 $52 = ((($s)) + 152|0); //@line 174 "modules/audio/openal/Source.cpp"
 $53 = HEAP32[$52>>2]|0; //@line 174 "modules/audio/openal/Source.cpp"
 HEAP32[$51>>2] = $53; //@line 174 "modules/audio/openal/Source.cpp"
 $54 = ((($this)) + 156|0); //@line 175 "modules/audio/openal/Source.cpp"
 $55 = ((($s)) + 156|0); //@line 175 "modules/audio/openal/Source.cpp"
 $56 = HEAP32[$55>>2]|0; //@line 175 "modules/audio/openal/Source.cpp"
 HEAP32[$54>>2] = $56; //@line 175 "modules/audio/openal/Source.cpp"
 $57 = ((($this)) + 160|0); //@line 93 "./common/Object.h"
 HEAP32[$57>>2] = 0; //@line 93 "./common/Object.h"
 $58 = ((($this)) + 164|0); //@line 177 "modules/audio/openal/Source.cpp"
 HEAP32[$58>>2] = 0; //@line 177 "modules/audio/openal/Source.cpp"
 $59 = ((($this)) + 8|0); //@line 179 "modules/audio/openal/Source.cpp"
 $60 = HEAP32[$59>>2]|0; //@line 179 "modules/audio/openal/Source.cpp"
 $61 = ($60|0)==(1); //@line 179 "modules/audio/openal/Source.cpp"
 do {
  if ($61) {
   $62 = ((($s)) + 160|0); //@line 139 "./common/Object.h"
   $63 = HEAP32[$62>>2]|0; //@line 139 "./common/Object.h"
   $64 = ($63|0)==(0|0); //@line 181 "modules/audio/openal/Source.cpp"
   do {
    if ($64) {
     label = 15;
    } else {
     $65 = HEAP32[$63>>2]|0; //@line 183 "modules/audio/openal/Source.cpp"
     $66 = ((($65)) + 16|0); //@line 183 "modules/audio/openal/Source.cpp"
     $67 = HEAP32[$66>>2]|0; //@line 183 "modules/audio/openal/Source.cpp"
     __THREW__ = 0;
     $68 = (invoke_ii($67|0,($63|0))|0); //@line 183 "modules/audio/openal/Source.cpp"
     $69 = __THREW__; __THREW__ = 0;
     $70 = $69&1;
     if (!($70)) {
      $71 = ($68|0)==(0|0); //@line 132 "./common/Object.h"
      if (!($71)) {
       $72 = HEAP32[$68>>2]|0; //@line 132 "./common/Object.h"
       $73 = ((($72)) + 8|0); //@line 132 "./common/Object.h"
       $74 = HEAP32[$73>>2]|0; //@line 132 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($74|0,($68|0)); //@line 132 "./common/Object.h"
       $75 = __THREW__; __THREW__ = 0;
       $76 = $75&1;
       if ($76) {
        break;
       }
      }
      $77 = HEAP32[$57>>2]|0; //@line 133 "./common/Object.h"
      $78 = ($77|0)==(0|0); //@line 133 "./common/Object.h"
      if (!($78)) {
       $79 = HEAP32[$77>>2]|0; //@line 133 "./common/Object.h"
       $80 = ((($79)) + 12|0); //@line 133 "./common/Object.h"
       $81 = HEAP32[$80>>2]|0; //@line 133 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($81|0,($77|0)); //@line 133 "./common/Object.h"
       $82 = __THREW__; __THREW__ = 0;
       $83 = $82&1;
       if ($83) {
        break;
       }
      }
      HEAP32[$57>>2] = $68; //@line 134 "./common/Object.h"
      $84 = HEAP32[$68>>2]|0; //@line 185 "modules/audio/openal/Source.cpp"
      $85 = ((($84)) + 12|0); //@line 185 "modules/audio/openal/Source.cpp"
      $86 = HEAP32[$85>>2]|0; //@line 185 "modules/audio/openal/Source.cpp"
      __THREW__ = 0;
      invoke_vi($86|0,($68|0)); //@line 185 "modules/audio/openal/Source.cpp"
      $87 = __THREW__; __THREW__ = 0;
      $88 = $87&1;
      if (!($88)) {
       label = 15;
      }
     }
    }
   } while(0);
   if ((label|0) == 15) {
    $102 = ((($this)) + 24|0); //@line 188 "modules/audio/openal/Source.cpp"
    __THREW__ = 0;
    invoke_vii(3932,8,($102|0)); //@line 188 "modules/audio/openal/Source.cpp"
    $103 = __THREW__; __THREW__ = 0;
    $104 = $103&1;
    if (!($104)) {
     break;
    }
   }
   $91 = ___cxa_find_matching_catch_2()|0;
   $92 = tempRet0;
   $93 = HEAP32[$57>>2]|0; //@line 106 "./common/Object.h"
   $94 = ($93|0)==(0|0); //@line 106 "./common/Object.h"
   if (!($94)) {
    $95 = HEAP32[$93>>2]|0; //@line 106 "./common/Object.h"
    $96 = ((($95)) + 12|0); //@line 106 "./common/Object.h"
    $97 = HEAP32[$96>>2]|0; //@line 106 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($97|0,($93|0)); //@line 106 "./common/Object.h"
    $98 = __THREW__; __THREW__ = 0;
    $99 = $98&1;
    if ($99) {
     $100 = ___cxa_find_matching_catch_3(0|0)|0;
     $101 = tempRet0;
     ___clang_call_terminate($100); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   }
   $132 = HEAP32[$6>>2]|0; //@line 106 "./common/Object.h"
   $133 = ($132|0)==(0|0); //@line 106 "./common/Object.h"
   if ($133) {
    $$1 = $91;$$12 = $92;
    __ZN4love5audio6SourceD2Ev($this); //@line 194 "modules/audio/openal/Source.cpp"
    ___resumeException($$1|0);
    // unreachable;
   }
   $134 = HEAP32[$132>>2]|0; //@line 106 "./common/Object.h"
   $135 = ((($134)) + 12|0); //@line 106 "./common/Object.h"
   $136 = HEAP32[$135>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($136|0,($132|0)); //@line 106 "./common/Object.h"
   $137 = __THREW__; __THREW__ = 0;
   $138 = $137&1;
   if ($138) {
    $139 = ___cxa_find_matching_catch_3(0|0)|0;
    $140 = tempRet0;
    ___clang_call_terminate($139); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   } else {
    $$1 = $91;$$12 = $92;
    __ZN4love5audio6SourceD2Ev($this); //@line 194 "modules/audio/openal/Source.cpp"
    ___resumeException($$1|0);
    // unreachable;
   }
  }
 } while(0);
 $105 = ((($this)) + 68|0); //@line 191 "modules/audio/openal/Source.cpp"
 $106 = ((($s)) + 68|0); //@line 191 "modules/audio/openal/Source.cpp"
 $107 = HEAP32[$106>>2]|0; //@line 743 "modules/audio/openal/Source.cpp"
 HEAP32[$105>>2] = $107; //@line 743 "modules/audio/openal/Source.cpp"
 $108 = ((($s)) + 72|0); //@line 744 "modules/audio/openal/Source.cpp"
 $109 = HEAP32[$108>>2]|0; //@line 744 "modules/audio/openal/Source.cpp"
 $110 = ((($this)) + 72|0); //@line 744 "modules/audio/openal/Source.cpp"
 HEAP32[$110>>2] = $109; //@line 744 "modules/audio/openal/Source.cpp"
 $111 = ((($s)) + 76|0); //@line 745 "modules/audio/openal/Source.cpp"
 $112 = HEAP32[$111>>2]|0; //@line 745 "modules/audio/openal/Source.cpp"
 $113 = ((($this)) + 76|0); //@line 745 "modules/audio/openal/Source.cpp"
 HEAP32[$113>>2] = $112; //@line 745 "modules/audio/openal/Source.cpp"
 $114 = ((($this)) + 80|0); //@line 192 "modules/audio/openal/Source.cpp"
 $115 = ((($s)) + 80|0); //@line 192 "modules/audio/openal/Source.cpp"
 $116 = HEAP32[$115>>2]|0; //@line 743 "modules/audio/openal/Source.cpp"
 HEAP32[$114>>2] = $116; //@line 743 "modules/audio/openal/Source.cpp"
 $117 = ((($s)) + 84|0); //@line 744 "modules/audio/openal/Source.cpp"
 $118 = HEAP32[$117>>2]|0; //@line 744 "modules/audio/openal/Source.cpp"
 $119 = ((($this)) + 84|0); //@line 744 "modules/audio/openal/Source.cpp"
 HEAP32[$119>>2] = $118; //@line 744 "modules/audio/openal/Source.cpp"
 $120 = ((($s)) + 88|0); //@line 745 "modules/audio/openal/Source.cpp"
 $121 = HEAP32[$120>>2]|0; //@line 745 "modules/audio/openal/Source.cpp"
 $122 = ((($this)) + 88|0); //@line 745 "modules/audio/openal/Source.cpp"
 HEAP32[$122>>2] = $121; //@line 745 "modules/audio/openal/Source.cpp"
 $123 = ((($this)) + 92|0); //@line 193 "modules/audio/openal/Source.cpp"
 $124 = ((($s)) + 92|0); //@line 193 "modules/audio/openal/Source.cpp"
 $125 = HEAP32[$124>>2]|0; //@line 743 "modules/audio/openal/Source.cpp"
 HEAP32[$123>>2] = $125; //@line 743 "modules/audio/openal/Source.cpp"
 $126 = ((($s)) + 96|0); //@line 744 "modules/audio/openal/Source.cpp"
 $127 = HEAP32[$126>>2]|0; //@line 744 "modules/audio/openal/Source.cpp"
 $128 = ((($this)) + 96|0); //@line 744 "modules/audio/openal/Source.cpp"
 HEAP32[$128>>2] = $127; //@line 744 "modules/audio/openal/Source.cpp"
 $129 = ((($s)) + 100|0); //@line 745 "modules/audio/openal/Source.cpp"
 $130 = HEAP32[$129>>2]|0; //@line 745 "modules/audio/openal/Source.cpp"
 $131 = ((($this)) + 100|0); //@line 745 "modules/audio/openal/Source.cpp"
 HEAP32[$131>>2] = $130; //@line 745 "modules/audio/openal/Source.cpp"
 return; //@line 194 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source4playEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 212 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 212 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 212 "modules/audio/openal/Source.cpp"
 if (!($2)) {
  $3 = ((($this)) + 106|0); //@line 212 "modules/audio/openal/Source.cpp"
  $4 = HEAP8[$3>>0]|0; //@line 212 "modules/audio/openal/Source.cpp"
  $5 = ($4<<24>>24)==(0); //@line 212 "modules/audio/openal/Source.cpp"
  if (!($5)) {
   $6 = ((($this)) + 12|0); //@line 214 "modules/audio/openal/Source.cpp"
   $7 = HEAP32[$6>>2]|0; //@line 214 "modules/audio/openal/Source.cpp"
   __ZN4love5audio6openal4Pool6resumeEPNS1_6SourceE($7,$this); //@line 214 "modules/audio/openal/Source.cpp"
   $$0 = 1;
   return ($$0|0); //@line 220 "modules/audio/openal/Source.cpp"
  }
 }
 $8 = ((($this)) + 12|0); //@line 218 "modules/audio/openal/Source.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 218 "modules/audio/openal/Source.cpp"
 $10 = ((($this)) + 16|0); //@line 218 "modules/audio/openal/Source.cpp"
 $11 = (__ZN4love5audio6openal4Pool4playEPNS1_6SourceERj($9,$this,$10)|0); //@line 218 "modules/audio/openal/Source.cpp"
 $12 = $11&1; //@line 218 "modules/audio/openal/Source.cpp"
 HEAP8[$0>>0] = $12; //@line 218 "modules/audio/openal/Source.cpp"
 $$0 = $11;
 return ($$0|0); //@line 220 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source4stopEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 224 "modules/audio/openal/Source.cpp"
 $1 = ((($0)) + 40|0); //@line 224 "modules/audio/openal/Source.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 224 "modules/audio/openal/Source.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 224 "modules/audio/openal/Source.cpp"
 if ($3) {
  return; //@line 229 "modules/audio/openal/Source.cpp"
 }
 $4 = ((($this)) + 12|0); //@line 226 "modules/audio/openal/Source.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 226 "modules/audio/openal/Source.cpp"
 __ZN4love5audio6openal4Pool4stopEPNS1_6SourceE($5,$this); //@line 226 "modules/audio/openal/Source.cpp"
 $6 = HEAP32[$4>>2]|0; //@line 227 "modules/audio/openal/Source.cpp"
 __ZN4love5audio6openal4Pool10softRewindEPNS1_6SourceE($6,$this); //@line 227 "modules/audio/openal/Source.cpp"
 return; //@line 229 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source5pauseEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 233 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 233 "modules/audio/openal/Source.cpp"
 __ZN4love5audio6openal4Pool5pauseEPNS1_6SourceE($1,$this); //@line 233 "modules/audio/openal/Source.cpp"
 return; //@line 234 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source6resumeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 238 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 238 "modules/audio/openal/Source.cpp"
 __ZN4love5audio6openal4Pool6resumeEPNS1_6SourceE($1,$this); //@line 238 "modules/audio/openal/Source.cpp"
 return; //@line 239 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source6rewindEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 243 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 243 "modules/audio/openal/Source.cpp"
 __ZN4love5audio6openal4Pool6rewindEPNS1_6SourceE($1,$this); //@line 243 "modules/audio/openal/Source.cpp"
 return; //@line 244 "modules/audio/openal/Source.cpp"
}
function __ZNK4love5audio6openal6Source9isStoppedEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $state = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $state = sp;
 $0 = ((($this)) + 20|0); //@line 248 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 248 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 248 "modules/audio/openal/Source.cpp"
 if ($2) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 256 "modules/audio/openal/Source.cpp"
 }
 $3 = ((($this)) + 16|0); //@line 251 "modules/audio/openal/Source.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 251 "modules/audio/openal/Source.cpp"
 _alGetSourcei(($4|0),4112,($state|0)); //@line 251 "modules/audio/openal/Source.cpp"
 $5 = HEAP32[$state>>2]|0; //@line 252 "modules/audio/openal/Source.cpp"
 $6 = ($5|0)==(4116); //@line 252 "modules/audio/openal/Source.cpp"
 $$0 = $6;
 STACKTOP = sp;return ($$0|0); //@line 256 "modules/audio/openal/Source.cpp"
}
function __ZNK4love5audio6openal6Source8isPausedEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $state = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $state = sp;
 $0 = ((($this)) + 20|0); //@line 260 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 260 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 260 "modules/audio/openal/Source.cpp"
 if ($2) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 268 "modules/audio/openal/Source.cpp"
 }
 $3 = ((($this)) + 16|0); //@line 263 "modules/audio/openal/Source.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 263 "modules/audio/openal/Source.cpp"
 _alGetSourcei(($4|0),4112,($state|0)); //@line 263 "modules/audio/openal/Source.cpp"
 $5 = HEAP32[$state>>2]|0; //@line 264 "modules/audio/openal/Source.cpp"
 $6 = ($5|0)==(4115); //@line 264 "modules/audio/openal/Source.cpp"
 $$0 = $6;
 STACKTOP = sp;return ($$0|0); //@line 268 "modules/audio/openal/Source.cpp"
}
function __ZNK4love5audio6openal6Source10isFinishedEv($this) {
 $this = $this|0;
 var $$mux = 0, $$not = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $brmerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 272 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 272 "modules/audio/openal/Source.cpp"
 $2 = ($1|0)==(0); //@line 272 "modules/audio/openal/Source.cpp"
 $3 = HEAP32[$this>>2]|0; //@line 272 "modules/audio/openal/Source.cpp"
 $4 = ((($3)) + 40|0); //@line 272 "modules/audio/openal/Source.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 272 "modules/audio/openal/Source.cpp"
 $6 = (FUNCTION_TABLE_ii[$5 & 8191]($this)|0); //@line 272 "modules/audio/openal/Source.cpp"
 $$not = $6 ^ 1; //@line 272 "modules/audio/openal/Source.cpp"
 $brmerge = $2 | $$not; //@line 272 "modules/audio/openal/Source.cpp"
 $$mux = $2 & $6; //@line 272 "modules/audio/openal/Source.cpp"
 if ($brmerge) {
  $17 = $$mux;
  return ($17|0); //@line 272 "modules/audio/openal/Source.cpp"
 }
 $7 = HEAP32[$this>>2]|0; //@line 272 "modules/audio/openal/Source.cpp"
 $8 = ((($7)) + 128|0); //@line 272 "modules/audio/openal/Source.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 272 "modules/audio/openal/Source.cpp"
 $10 = (FUNCTION_TABLE_ii[$9 & 8191]($this)|0); //@line 272 "modules/audio/openal/Source.cpp"
 if ($10) {
  $17 = 0;
  return ($17|0); //@line 272 "modules/audio/openal/Source.cpp"
 }
 $11 = ((($this)) + 160|0); //@line 117 "./common/Object.h"
 $12 = HEAP32[$11>>2]|0; //@line 117 "./common/Object.h"
 $13 = HEAP32[$12>>2]|0; //@line 272 "modules/audio/openal/Source.cpp"
 $14 = ((($13)) + 44|0); //@line 272 "modules/audio/openal/Source.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 272 "modules/audio/openal/Source.cpp"
 $16 = (FUNCTION_TABLE_ii[$15 & 8191]($12)|0); //@line 272 "modules/audio/openal/Source.cpp"
 $17 = $16;
 return ($17|0); //@line 272 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source6updateEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0;
 var $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $buffer = 0, $curOffsetSamples = 0, $newOffsetSamples = 0, $processed = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $processed = sp + 12|0;
 $buffer = sp + 8|0;
 $curOffsetSamples = sp + 4|0;
 $newOffsetSamples = sp;
 $0 = ((($this)) + 20|0); //@line 277 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 277 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 277 "modules/audio/openal/Source.cpp"
 if ($2) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 326 "modules/audio/openal/Source.cpp"
 }
 $3 = ((($this)) + 8|0); //@line 280 "modules/audio/openal/Source.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 280 "modules/audio/openal/Source.cpp"
 switch ($4|0) {
 case 0:  {
  $5 = ((($this)) + 16|0); //@line 283 "modules/audio/openal/Source.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 283 "modules/audio/openal/Source.cpp"
  $7 = HEAP32[$this>>2]|0; //@line 283 "modules/audio/openal/Source.cpp"
  $8 = ((($7)) + 128|0); //@line 283 "modules/audio/openal/Source.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 283 "modules/audio/openal/Source.cpp"
  $10 = (FUNCTION_TABLE_ii[$9 & 8191]($this)|0); //@line 283 "modules/audio/openal/Source.cpp"
  $11 = $10&1; //@line 283 "modules/audio/openal/Source.cpp"
  _alSourcei(($6|0),4103,($11|0)); //@line 283 "modules/audio/openal/Source.cpp"
  $12 = HEAP32[$this>>2]|0; //@line 284 "modules/audio/openal/Source.cpp"
  $13 = ((($12)) + 40|0); //@line 284 "modules/audio/openal/Source.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 284 "modules/audio/openal/Source.cpp"
  $15 = (FUNCTION_TABLE_ii[$14 & 8191]($this)|0); //@line 284 "modules/audio/openal/Source.cpp"
  $16 = $15 ^ 1; //@line 284 "modules/audio/openal/Source.cpp"
  $$0 = $16;
  STACKTOP = sp;return ($$0|0); //@line 326 "modules/audio/openal/Source.cpp"
  break;
 }
 case 1:  {
  $17 = HEAP32[$this>>2]|0; //@line 286 "modules/audio/openal/Source.cpp"
  $18 = ((($17)) + 128|0); //@line 286 "modules/audio/openal/Source.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 286 "modules/audio/openal/Source.cpp"
  $20 = (FUNCTION_TABLE_ii[$19 & 8191]($this)|0); //@line 286 "modules/audio/openal/Source.cpp"
  if (!($20)) {
   $21 = HEAP32[$this>>2]|0; //@line 286 "modules/audio/openal/Source.cpp"
   $22 = ((($21)) + 48|0); //@line 286 "modules/audio/openal/Source.cpp"
   $23 = HEAP32[$22>>2]|0; //@line 286 "modules/audio/openal/Source.cpp"
   $24 = (FUNCTION_TABLE_ii[$23 & 8191]($this)|0); //@line 286 "modules/audio/openal/Source.cpp"
   if ($24) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0); //@line 326 "modules/audio/openal/Source.cpp"
   }
  }
  HEAP32[$processed>>2] = 0; //@line 289 "modules/audio/openal/Source.cpp"
  $25 = ((($this)) + 16|0); //@line 291 "modules/audio/openal/Source.cpp"
  $26 = HEAP32[$25>>2]|0; //@line 291 "modules/audio/openal/Source.cpp"
  _alGetSourcei(($26|0),4118,($processed|0)); //@line 291 "modules/audio/openal/Source.cpp"
  $27 = HEAP32[$processed>>2]|0; //@line 293 "modules/audio/openal/Source.cpp"
  $28 = (($27) + -1)|0; //@line 293 "modules/audio/openal/Source.cpp"
  HEAP32[$processed>>2] = $28; //@line 293 "modules/audio/openal/Source.cpp"
  $29 = ($27|0)==(0); //@line 293 "modules/audio/openal/Source.cpp"
  if (!($29)) {
   $30 = ((($this)) + 160|0); //@line 117 "./common/Object.h"
   $31 = ((($this)) + 140|0); //@line 312 "modules/audio/openal/Source.cpp"
   $32 = ((($this)) + 144|0); //@line 313 "modules/audio/openal/Source.cpp"
   while(1) {
    $33 = HEAP32[$25>>2]|0; //@line 299 "modules/audio/openal/Source.cpp"
    _alGetSourcef(($33|0),4133,($curOffsetSamples|0)); //@line 299 "modules/audio/openal/Source.cpp"
    $34 = HEAP32[$30>>2]|0; //@line 117 "./common/Object.h"
    $35 = HEAP32[$34>>2]|0; //@line 301 "modules/audio/openal/Source.cpp"
    $36 = ((($35)) + 56|0); //@line 301 "modules/audio/openal/Source.cpp"
    $37 = HEAP32[$36>>2]|0; //@line 301 "modules/audio/openal/Source.cpp"
    $38 = (FUNCTION_TABLE_ii[$37 & 8191]($34)|0); //@line 301 "modules/audio/openal/Source.cpp"
    $39 = +HEAPF32[$curOffsetSamples>>2]; //@line 302 "modules/audio/openal/Source.cpp"
    $40 = (+($38|0)); //@line 302 "modules/audio/openal/Source.cpp"
    $41 = $39 / $40; //@line 302 "modules/audio/openal/Source.cpp"
    $42 = HEAP32[$25>>2]|0; //@line 305 "modules/audio/openal/Source.cpp"
    _alSourceUnqueueBuffers(($42|0),1,($buffer|0)); //@line 305 "modules/audio/openal/Source.cpp"
    $43 = HEAP32[$25>>2]|0; //@line 309 "modules/audio/openal/Source.cpp"
    _alGetSourcef(($43|0),4133,($newOffsetSamples|0)); //@line 309 "modules/audio/openal/Source.cpp"
    $44 = +HEAPF32[$newOffsetSamples>>2]; //@line 310 "modules/audio/openal/Source.cpp"
    $45 = $44 / $40; //@line 310 "modules/audio/openal/Source.cpp"
    $46 = +HEAPF32[$curOffsetSamples>>2]; //@line 312 "modules/audio/openal/Source.cpp"
    $47 = $46 - $44; //@line 312 "modules/audio/openal/Source.cpp"
    $48 = +HEAPF32[$31>>2]; //@line 312 "modules/audio/openal/Source.cpp"
    $49 = $48 + $47; //@line 312 "modules/audio/openal/Source.cpp"
    HEAPF32[$31>>2] = $49; //@line 312 "modules/audio/openal/Source.cpp"
    $50 = $41 - $45; //@line 313 "modules/audio/openal/Source.cpp"
    $51 = +HEAPF32[$32>>2]; //@line 313 "modules/audio/openal/Source.cpp"
    $52 = $50 + $51; //@line 313 "modules/audio/openal/Source.cpp"
    HEAPF32[$32>>2] = $52; //@line 313 "modules/audio/openal/Source.cpp"
    $53 = HEAP32[$buffer>>2]|0; //@line 318 "modules/audio/openal/Source.cpp"
    $54 = HEAP32[$30>>2]|0; //@line 139 "./common/Object.h"
    $55 = (__ZN4love5audio6openal6Source12streamAtomicEjPNS_5sound7DecoderE($this,$53,$54)|0); //@line 318 "modules/audio/openal/Source.cpp"
    $56 = ($55|0)>(0); //@line 318 "modules/audio/openal/Source.cpp"
    if ($56) {
     $57 = HEAP32[$25>>2]|0; //@line 319 "modules/audio/openal/Source.cpp"
     _alSourceQueueBuffers(($57|0),1,($buffer|0)); //@line 319 "modules/audio/openal/Source.cpp"
    }
    $58 = HEAP32[$processed>>2]|0; //@line 293 "modules/audio/openal/Source.cpp"
    $59 = (($58) + -1)|0; //@line 293 "modules/audio/openal/Source.cpp"
    HEAP32[$processed>>2] = $59; //@line 293 "modules/audio/openal/Source.cpp"
    $60 = ($58|0)==(0); //@line 293 "modules/audio/openal/Source.cpp"
    if ($60) {
     break;
    }
   }
  }
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 326 "modules/audio/openal/Source.cpp"
  break;
 }
 default: {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 326 "modules/audio/openal/Source.cpp"
 }
 }
 return (0)|0;
}
function __ZN4love5audio6openal6Source12streamAtomicEjPNS_5sound7DecoderE($this,$buffer,$d) {
 $this = $this|0;
 $buffer = $buffer|0;
 $d = $d|0;
 var $$ = 0, $$0$i$ph = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $decoded$1 = 0, $or$cond$i = 0, $or$cond3$i = 0;
 var $or$cond5$i = 0, $or$cond7$i = 0, $processed = 0, $queued = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $queued = sp + 4|0;
 $processed = sp;
 $0 = HEAP32[$d>>2]|0; //@line 779 "modules/audio/openal/Source.cpp"
 $1 = ((($0)) + 20|0); //@line 779 "modules/audio/openal/Source.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 779 "modules/audio/openal/Source.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($d)|0); //@line 779 "modules/audio/openal/Source.cpp"
 $4 = ($3|0)<(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $5 = $4 ? 0 : $3; //@line 779 "modules/audio/openal/Source.cpp"
 $6 = ($5|0)>(0); //@line 782 "modules/audio/openal/Source.cpp"
 do {
  if ($6) {
   $7 = HEAP32[$d>>2]|0; //@line 784 "modules/audio/openal/Source.cpp"
   $8 = ((($7)) + 48|0); //@line 784 "modules/audio/openal/Source.cpp"
   $9 = HEAP32[$8>>2]|0; //@line 784 "modules/audio/openal/Source.cpp"
   $10 = (FUNCTION_TABLE_ii[$9 & 8191]($d)|0); //@line 784 "modules/audio/openal/Source.cpp"
   $11 = HEAP32[$d>>2]|0; //@line 784 "modules/audio/openal/Source.cpp"
   $12 = ((($11)) + 52|0); //@line 784 "modules/audio/openal/Source.cpp"
   $13 = HEAP32[$12>>2]|0; //@line 784 "modules/audio/openal/Source.cpp"
   $14 = (FUNCTION_TABLE_ii[$13 & 8191]($d)|0); //@line 784 "modules/audio/openal/Source.cpp"
   $15 = ($10|0)==(1); //@line 750 "modules/audio/openal/Source.cpp"
   $16 = ($14|0)==(8); //@line 750 "modules/audio/openal/Source.cpp"
   $or$cond$i = $15 & $16; //@line 750 "modules/audio/openal/Source.cpp"
   if ($or$cond$i) {
    $$0$i$ph = 4352;
   } else {
    $17 = ($14|0)==(16); //@line 752 "modules/audio/openal/Source.cpp"
    $or$cond3$i = $15 & $17; //@line 752 "modules/audio/openal/Source.cpp"
    if ($or$cond3$i) {
     $$0$i$ph = 4353;
    } else {
     $18 = ($10|0)==(2); //@line 754 "modules/audio/openal/Source.cpp"
     $or$cond5$i = $18 & $16; //@line 754 "modules/audio/openal/Source.cpp"
     if ($or$cond5$i) {
      $$0$i$ph = 4354;
     } else {
      $or$cond7$i = $18 & $17; //@line 756 "modules/audio/openal/Source.cpp"
      if ($or$cond7$i) {
       $$0$i$ph = 4355;
      } else {
       $decoded$1 = 0;
       break;
      }
     }
    }
   }
   $19 = HEAP32[$d>>2]|0; //@line 787 "modules/audio/openal/Source.cpp"
   $20 = ((($19)) + 28|0); //@line 787 "modules/audio/openal/Source.cpp"
   $21 = HEAP32[$20>>2]|0; //@line 787 "modules/audio/openal/Source.cpp"
   $22 = (FUNCTION_TABLE_ii[$21 & 8191]($d)|0); //@line 787 "modules/audio/openal/Source.cpp"
   $23 = HEAP32[$d>>2]|0; //@line 787 "modules/audio/openal/Source.cpp"
   $24 = ((($23)) + 56|0); //@line 787 "modules/audio/openal/Source.cpp"
   $25 = HEAP32[$24>>2]|0; //@line 787 "modules/audio/openal/Source.cpp"
   $26 = (FUNCTION_TABLE_ii[$25 & 8191]($d)|0); //@line 787 "modules/audio/openal/Source.cpp"
   _alBufferData(($buffer|0),($$0$i$ph|0),($22|0),($5|0),($26|0)); //@line 787 "modules/audio/openal/Source.cpp"
   $decoded$1 = $5;
  } else {
   $decoded$1 = $5;
  }
 } while(0);
 $27 = ((($this)) + 160|0); //@line 117 "./common/Object.h"
 $28 = HEAP32[$27>>2]|0; //@line 117 "./common/Object.h"
 $29 = HEAP32[$28>>2]|0; //@line 792 "modules/audio/openal/Source.cpp"
 $30 = ((($29)) + 44|0); //@line 792 "modules/audio/openal/Source.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 792 "modules/audio/openal/Source.cpp"
 $32 = (FUNCTION_TABLE_ii[$31 & 8191]($28)|0); //@line 792 "modules/audio/openal/Source.cpp"
 if ($32) {
  $33 = HEAP32[$this>>2]|0; //@line 792 "modules/audio/openal/Source.cpp"
  $34 = ((($33)) + 128|0); //@line 792 "modules/audio/openal/Source.cpp"
  $35 = HEAP32[$34>>2]|0; //@line 792 "modules/audio/openal/Source.cpp"
  $36 = (FUNCTION_TABLE_ii[$35 & 8191]($this)|0); //@line 792 "modules/audio/openal/Source.cpp"
  if ($36) {
   $37 = ((($this)) + 16|0); //@line 795 "modules/audio/openal/Source.cpp"
   $38 = HEAP32[$37>>2]|0; //@line 795 "modules/audio/openal/Source.cpp"
   _alGetSourcei(($38|0),4117,($queued|0)); //@line 795 "modules/audio/openal/Source.cpp"
   $39 = HEAP32[$37>>2]|0; //@line 796 "modules/audio/openal/Source.cpp"
   _alGetSourcei(($39|0),4118,($processed|0)); //@line 796 "modules/audio/openal/Source.cpp"
   $40 = HEAP32[$queued>>2]|0; //@line 797 "modules/audio/openal/Source.cpp"
   $41 = HEAP32[$processed>>2]|0; //@line 797 "modules/audio/openal/Source.cpp"
   $42 = ($40|0)>($41|0); //@line 797 "modules/audio/openal/Source.cpp"
   $$ = $42 ? $40 : 8; //@line 797 "modules/audio/openal/Source.cpp"
   $$sink = (($$) - ($41))|0; //@line 800 "modules/audio/openal/Source.cpp"
   $43 = ((($this)) + 164|0); //@line 800 "modules/audio/openal/Source.cpp"
   HEAP32[$43>>2] = $$sink; //@line 800 "modules/audio/openal/Source.cpp"
   $44 = HEAP32[$d>>2]|0; //@line 801 "modules/audio/openal/Source.cpp"
   $45 = ((($44)) + 36|0); //@line 801 "modules/audio/openal/Source.cpp"
   $46 = HEAP32[$45>>2]|0; //@line 801 "modules/audio/openal/Source.cpp"
   (FUNCTION_TABLE_ii[$46 & 8191]($d)|0); //@line 801 "modules/audio/openal/Source.cpp"
  }
 }
 $47 = ((($this)) + 164|0); //@line 804 "modules/audio/openal/Source.cpp"
 $48 = HEAP32[$47>>2]|0; //@line 804 "modules/audio/openal/Source.cpp"
 $49 = ($48|0)==(0); //@line 804 "modules/audio/openal/Source.cpp"
 if ($49) {
  STACKTOP = sp;return ($decoded$1|0); //@line 813 "modules/audio/openal/Source.cpp"
 }
 $50 = (($48) + -1)|0; //@line 806 "modules/audio/openal/Source.cpp"
 HEAP32[$47>>2] = $50; //@line 806 "modules/audio/openal/Source.cpp"
 $51 = ($50|0)==(0); //@line 806 "modules/audio/openal/Source.cpp"
 if (!($51)) {
  STACKTOP = sp;return ($decoded$1|0); //@line 813 "modules/audio/openal/Source.cpp"
 }
 $52 = ((($this)) + 140|0); //@line 808 "modules/audio/openal/Source.cpp"
 HEAPF32[$52>>2] = 0.0; //@line 808 "modules/audio/openal/Source.cpp"
 $53 = ((($this)) + 144|0); //@line 809 "modules/audio/openal/Source.cpp"
 HEAPF32[$53>>2] = 0.0; //@line 809 "modules/audio/openal/Source.cpp"
 STACKTOP = sp;return ($decoded$1|0); //@line 813 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source8setPitchEf($this,$pitch) {
 $this = $this|0;
 $pitch = +$pitch;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 330 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 330 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 330 "modules/audio/openal/Source.cpp"
 if (!($2)) {
  $3 = ((($this)) + 16|0); //@line 331 "modules/audio/openal/Source.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 331 "modules/audio/openal/Source.cpp"
  _alSourcef(($4|0),4099,(+$pitch)); //@line 331 "modules/audio/openal/Source.cpp"
 }
 $5 = ((($this)) + 60|0); //@line 333 "modules/audio/openal/Source.cpp"
 HEAPF32[$5>>2] = $pitch; //@line 333 "modules/audio/openal/Source.cpp"
 return; //@line 334 "modules/audio/openal/Source.cpp"
}
function __ZNK4love5audio6openal6Source8getPitchEv($this) {
 $this = $this|0;
 var $$0 = 0.0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $f = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f = sp;
 $0 = ((($this)) + 20|0); //@line 338 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 338 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 338 "modules/audio/openal/Source.cpp"
 if ($2) {
  $6 = ((($this)) + 60|0); //@line 346 "modules/audio/openal/Source.cpp"
  $7 = +HEAPF32[$6>>2]; //@line 346 "modules/audio/openal/Source.cpp"
  $$0 = $7;
  STACKTOP = sp;return (+$$0); //@line 347 "modules/audio/openal/Source.cpp"
 } else {
  $3 = ((($this)) + 16|0); //@line 341 "modules/audio/openal/Source.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 341 "modules/audio/openal/Source.cpp"
  _alGetSourcef(($4|0),4099,($f|0)); //@line 341 "modules/audio/openal/Source.cpp"
  $5 = +HEAPF32[$f>>2]; //@line 342 "modules/audio/openal/Source.cpp"
  $$0 = $5;
  STACKTOP = sp;return (+$$0); //@line 347 "modules/audio/openal/Source.cpp"
 }
 return +(0.0);
}
function __ZN4love5audio6openal6Source9setVolumeEf($this,$volume) {
 $this = $this|0;
 $volume = +$volume;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 351 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 351 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 351 "modules/audio/openal/Source.cpp"
 if (!($2)) {
  $3 = ((($this)) + 16|0); //@line 352 "modules/audio/openal/Source.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 352 "modules/audio/openal/Source.cpp"
  _alSourcef(($4|0),4106,(+$volume)); //@line 352 "modules/audio/openal/Source.cpp"
 }
 $5 = ((($this)) + 64|0); //@line 354 "modules/audio/openal/Source.cpp"
 HEAPF32[$5>>2] = $volume; //@line 354 "modules/audio/openal/Source.cpp"
 return; //@line 355 "modules/audio/openal/Source.cpp"
}
function __ZNK4love5audio6openal6Source9getVolumeEv($this) {
 $this = $this|0;
 var $$0 = 0.0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $f = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f = sp;
 $0 = ((($this)) + 20|0); //@line 359 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 359 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 359 "modules/audio/openal/Source.cpp"
 if ($2) {
  $6 = ((($this)) + 64|0); //@line 367 "modules/audio/openal/Source.cpp"
  $7 = +HEAPF32[$6>>2]; //@line 367 "modules/audio/openal/Source.cpp"
  $$0 = $7;
  STACKTOP = sp;return (+$$0); //@line 368 "modules/audio/openal/Source.cpp"
 } else {
  $3 = ((($this)) + 16|0); //@line 362 "modules/audio/openal/Source.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 362 "modules/audio/openal/Source.cpp"
  _alGetSourcef(($4|0),4106,($f|0)); //@line 362 "modules/audio/openal/Source.cpp"
  $5 = +HEAPF32[$f>>2]; //@line 363 "modules/audio/openal/Source.cpp"
  $$0 = $5;
  STACKTOP = sp;return (+$$0); //@line 368 "modules/audio/openal/Source.cpp"
 }
 return +(0.0);
}
function __ZN4love5audio6openal6Source4seekEfNS0_6Source4UnitE($this,$offset,$unit) {
 $this = $this|0;
 $offset = +$offset;
 $unit = $unit|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 HEAP32[$0>>2] = $unit;
 $1 = ((($this)) + 12|0); //@line 415 "modules/audio/openal/Source.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 415 "modules/audio/openal/Source.cpp"
 __ZN4love5audio6openal4Pool4seekEPNS1_6SourceEfPv($2,$this,$offset,$0); //@line 415 "modules/audio/openal/Source.cpp"
 STACKTOP = sp;return; //@line 415 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source4tellENS0_6Source4UnitE($this,$unit) {
 $this = $this|0;
 $unit = $unit|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 HEAP32[$0>>2] = $unit;
 $1 = ((($this)) + 12|0); //@line 445 "modules/audio/openal/Source.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 445 "modules/audio/openal/Source.cpp"
 $3 = (+__ZN4love5audio6openal4Pool4tellEPNS1_6SourceEPv($2,$this,$0)); //@line 445 "modules/audio/openal/Source.cpp"
 STACKTOP = sp;return (+$3); //@line 445 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source11getDurationENS0_6Source4UnitE($this,$unit) {
 $this = $this|0;
 $unit = $unit|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 HEAP32[$0>>2] = $unit;
 $1 = ((($this)) + 12|0); //@line 475 "modules/audio/openal/Source.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 475 "modules/audio/openal/Source.cpp"
 $3 = (+__ZN4love5audio6openal4Pool11getDurationEPNS1_6SourceEPv($2,$this,$0)); //@line 475 "modules/audio/openal/Source.cpp"
 STACKTOP = sp;return (+$3); //@line 475 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source11setPositionEPf($this,$v) {
 $this = $this|0;
 $v = $v|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 152|0); //@line 480 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 480 "modules/audio/openal/Source.cpp"
 $2 = ($1|0)>(1); //@line 480 "modules/audio/openal/Source.cpp"
 if ($2) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 481 "modules/audio/openal/Source.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($3|0),(365088|0),($vararg_buffer|0)); //@line 60 "modules/audio/openal/Source.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $6 = ___cxa_find_matching_catch_2()|0;
   $7 = tempRet0;
   ___cxa_free_exception(($3|0)); //@line 481 "modules/audio/openal/Source.cpp"
   ___resumeException($6|0);
   // unreachable;
  } else {
   HEAP32[$3>>2] = (22756); //@line 62 "modules/audio/openal/Source.cpp"
   ___cxa_throw(($3|0),(1488|0),(344|0)); //@line 481 "modules/audio/openal/Source.cpp"
   // unreachable; //@line 481 "modules/audio/openal/Source.cpp"
  }
 } else {
  $8 = ((($this)) + 20|0); //@line 483 "modules/audio/openal/Source.cpp"
  $9 = HEAP8[$8>>0]|0; //@line 483 "modules/audio/openal/Source.cpp"
  $10 = ($9<<24>>24)==(0); //@line 483 "modules/audio/openal/Source.cpp"
  if (!($10)) {
   $11 = ((($this)) + 16|0); //@line 484 "modules/audio/openal/Source.cpp"
   $12 = HEAP32[$11>>2]|0; //@line 484 "modules/audio/openal/Source.cpp"
   _alSourcefv(($12|0),4100,($v|0)); //@line 484 "modules/audio/openal/Source.cpp"
  }
  $13 = ((($this)) + 68|0); //@line 486 "modules/audio/openal/Source.cpp"
  $14 = HEAP32[$v>>2]|0; //@line 743 "modules/audio/openal/Source.cpp"
  HEAP32[$13>>2] = $14; //@line 743 "modules/audio/openal/Source.cpp"
  $15 = ((($v)) + 4|0); //@line 744 "modules/audio/openal/Source.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 744 "modules/audio/openal/Source.cpp"
  $17 = ((($this)) + 72|0); //@line 744 "modules/audio/openal/Source.cpp"
  HEAP32[$17>>2] = $16; //@line 744 "modules/audio/openal/Source.cpp"
  $18 = ((($v)) + 8|0); //@line 745 "modules/audio/openal/Source.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 745 "modules/audio/openal/Source.cpp"
  $20 = ((($this)) + 76|0); //@line 745 "modules/audio/openal/Source.cpp"
  HEAP32[$20>>2] = $19; //@line 745 "modules/audio/openal/Source.cpp"
  STACKTOP = sp;return; //@line 487 "modules/audio/openal/Source.cpp"
 }
}
function __ZN4love5audio6openal23SpatialSupportExceptionD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love9ExceptionD2Ev($this); //@line 55 "modules/audio/openal/Source.cpp"
 __ZdlPv($this); //@line 55 "modules/audio/openal/Source.cpp"
 return; //@line 55 "modules/audio/openal/Source.cpp"
}
function __ZNK4love5audio6openal6Source11getPositionEPf($this,$v) {
 $this = $this|0;
 $v = $v|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 152|0); //@line 491 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 491 "modules/audio/openal/Source.cpp"
 $2 = ($1|0)>(1); //@line 491 "modules/audio/openal/Source.cpp"
 if ($2) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 492 "modules/audio/openal/Source.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($3|0),(365088|0),($vararg_buffer|0)); //@line 60 "modules/audio/openal/Source.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $6 = ___cxa_find_matching_catch_2()|0;
   $7 = tempRet0;
   ___cxa_free_exception(($3|0)); //@line 492 "modules/audio/openal/Source.cpp"
   ___resumeException($6|0);
   // unreachable;
  } else {
   HEAP32[$3>>2] = (22756); //@line 62 "modules/audio/openal/Source.cpp"
   ___cxa_throw(($3|0),(1488|0),(344|0)); //@line 492 "modules/audio/openal/Source.cpp"
   // unreachable; //@line 492 "modules/audio/openal/Source.cpp"
  }
 }
 $8 = ((($this)) + 20|0); //@line 494 "modules/audio/openal/Source.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 494 "modules/audio/openal/Source.cpp"
 $10 = ($9<<24>>24)==(0); //@line 494 "modules/audio/openal/Source.cpp"
 if ($10) {
  $13 = ((($this)) + 68|0); //@line 497 "modules/audio/openal/Source.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 743 "modules/audio/openal/Source.cpp"
  HEAP32[$v>>2] = $14; //@line 743 "modules/audio/openal/Source.cpp"
  $15 = ((($this)) + 72|0); //@line 744 "modules/audio/openal/Source.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 744 "modules/audio/openal/Source.cpp"
  $17 = ((($v)) + 4|0); //@line 744 "modules/audio/openal/Source.cpp"
  HEAP32[$17>>2] = $16; //@line 744 "modules/audio/openal/Source.cpp"
  $18 = ((($this)) + 76|0); //@line 745 "modules/audio/openal/Source.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 745 "modules/audio/openal/Source.cpp"
  $20 = ((($v)) + 8|0); //@line 745 "modules/audio/openal/Source.cpp"
  HEAP32[$20>>2] = $19; //@line 745 "modules/audio/openal/Source.cpp"
  STACKTOP = sp;return; //@line 498 "modules/audio/openal/Source.cpp"
 } else {
  $11 = ((($this)) + 16|0); //@line 495 "modules/audio/openal/Source.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 495 "modules/audio/openal/Source.cpp"
  _alGetSourcefv(($12|0),4100,($v|0)); //@line 495 "modules/audio/openal/Source.cpp"
  STACKTOP = sp;return; //@line 498 "modules/audio/openal/Source.cpp"
 }
}
function __ZN4love5audio6openal6Source11setVelocityEPf($this,$v) {
 $this = $this|0;
 $v = $v|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 152|0); //@line 502 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 502 "modules/audio/openal/Source.cpp"
 $2 = ($1|0)>(1); //@line 502 "modules/audio/openal/Source.cpp"
 if ($2) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 503 "modules/audio/openal/Source.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($3|0),(365088|0),($vararg_buffer|0)); //@line 60 "modules/audio/openal/Source.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $6 = ___cxa_find_matching_catch_2()|0;
   $7 = tempRet0;
   ___cxa_free_exception(($3|0)); //@line 503 "modules/audio/openal/Source.cpp"
   ___resumeException($6|0);
   // unreachable;
  } else {
   HEAP32[$3>>2] = (22756); //@line 62 "modules/audio/openal/Source.cpp"
   ___cxa_throw(($3|0),(1488|0),(344|0)); //@line 503 "modules/audio/openal/Source.cpp"
   // unreachable; //@line 503 "modules/audio/openal/Source.cpp"
  }
 } else {
  $8 = ((($this)) + 20|0); //@line 505 "modules/audio/openal/Source.cpp"
  $9 = HEAP8[$8>>0]|0; //@line 505 "modules/audio/openal/Source.cpp"
  $10 = ($9<<24>>24)==(0); //@line 505 "modules/audio/openal/Source.cpp"
  if (!($10)) {
   $11 = ((($this)) + 16|0); //@line 506 "modules/audio/openal/Source.cpp"
   $12 = HEAP32[$11>>2]|0; //@line 506 "modules/audio/openal/Source.cpp"
   _alSourcefv(($12|0),4102,($v|0)); //@line 506 "modules/audio/openal/Source.cpp"
  }
  $13 = ((($this)) + 80|0); //@line 508 "modules/audio/openal/Source.cpp"
  $14 = HEAP32[$v>>2]|0; //@line 743 "modules/audio/openal/Source.cpp"
  HEAP32[$13>>2] = $14; //@line 743 "modules/audio/openal/Source.cpp"
  $15 = ((($v)) + 4|0); //@line 744 "modules/audio/openal/Source.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 744 "modules/audio/openal/Source.cpp"
  $17 = ((($this)) + 84|0); //@line 744 "modules/audio/openal/Source.cpp"
  HEAP32[$17>>2] = $16; //@line 744 "modules/audio/openal/Source.cpp"
  $18 = ((($v)) + 8|0); //@line 745 "modules/audio/openal/Source.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 745 "modules/audio/openal/Source.cpp"
  $20 = ((($this)) + 88|0); //@line 745 "modules/audio/openal/Source.cpp"
  HEAP32[$20>>2] = $19; //@line 745 "modules/audio/openal/Source.cpp"
  STACKTOP = sp;return; //@line 509 "modules/audio/openal/Source.cpp"
 }
}
function __ZNK4love5audio6openal6Source11getVelocityEPf($this,$v) {
 $this = $this|0;
 $v = $v|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 152|0); //@line 513 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 513 "modules/audio/openal/Source.cpp"
 $2 = ($1|0)>(1); //@line 513 "modules/audio/openal/Source.cpp"
 if ($2) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 514 "modules/audio/openal/Source.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($3|0),(365088|0),($vararg_buffer|0)); //@line 60 "modules/audio/openal/Source.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $6 = ___cxa_find_matching_catch_2()|0;
   $7 = tempRet0;
   ___cxa_free_exception(($3|0)); //@line 514 "modules/audio/openal/Source.cpp"
   ___resumeException($6|0);
   // unreachable;
  } else {
   HEAP32[$3>>2] = (22756); //@line 62 "modules/audio/openal/Source.cpp"
   ___cxa_throw(($3|0),(1488|0),(344|0)); //@line 514 "modules/audio/openal/Source.cpp"
   // unreachable; //@line 514 "modules/audio/openal/Source.cpp"
  }
 }
 $8 = ((($this)) + 20|0); //@line 516 "modules/audio/openal/Source.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 516 "modules/audio/openal/Source.cpp"
 $10 = ($9<<24>>24)==(0); //@line 516 "modules/audio/openal/Source.cpp"
 if ($10) {
  $13 = ((($this)) + 80|0); //@line 519 "modules/audio/openal/Source.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 743 "modules/audio/openal/Source.cpp"
  HEAP32[$v>>2] = $14; //@line 743 "modules/audio/openal/Source.cpp"
  $15 = ((($this)) + 84|0); //@line 744 "modules/audio/openal/Source.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 744 "modules/audio/openal/Source.cpp"
  $17 = ((($v)) + 4|0); //@line 744 "modules/audio/openal/Source.cpp"
  HEAP32[$17>>2] = $16; //@line 744 "modules/audio/openal/Source.cpp"
  $18 = ((($this)) + 88|0); //@line 745 "modules/audio/openal/Source.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 745 "modules/audio/openal/Source.cpp"
  $20 = ((($v)) + 8|0); //@line 745 "modules/audio/openal/Source.cpp"
  HEAP32[$20>>2] = $19; //@line 745 "modules/audio/openal/Source.cpp"
  STACKTOP = sp;return; //@line 520 "modules/audio/openal/Source.cpp"
 } else {
  $11 = ((($this)) + 16|0); //@line 517 "modules/audio/openal/Source.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 517 "modules/audio/openal/Source.cpp"
  _alGetSourcefv(($12|0),4102,($v|0)); //@line 517 "modules/audio/openal/Source.cpp"
  STACKTOP = sp;return; //@line 520 "modules/audio/openal/Source.cpp"
 }
}
function __ZN4love5audio6openal6Source12setDirectionEPf($this,$v) {
 $this = $this|0;
 $v = $v|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 152|0); //@line 524 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 524 "modules/audio/openal/Source.cpp"
 $2 = ($1|0)>(1); //@line 524 "modules/audio/openal/Source.cpp"
 if ($2) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 525 "modules/audio/openal/Source.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($3|0),(365088|0),($vararg_buffer|0)); //@line 60 "modules/audio/openal/Source.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $6 = ___cxa_find_matching_catch_2()|0;
   $7 = tempRet0;
   ___cxa_free_exception(($3|0)); //@line 525 "modules/audio/openal/Source.cpp"
   ___resumeException($6|0);
   // unreachable;
  } else {
   HEAP32[$3>>2] = (22756); //@line 62 "modules/audio/openal/Source.cpp"
   ___cxa_throw(($3|0),(1488|0),(344|0)); //@line 525 "modules/audio/openal/Source.cpp"
   // unreachable; //@line 525 "modules/audio/openal/Source.cpp"
  }
 }
 $8 = ((($this)) + 20|0); //@line 527 "modules/audio/openal/Source.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 527 "modules/audio/openal/Source.cpp"
 $10 = ($9<<24>>24)==(0); //@line 527 "modules/audio/openal/Source.cpp"
 if ($10) {
  $13 = ((($this)) + 92|0); //@line 530 "modules/audio/openal/Source.cpp"
  $14 = HEAP32[$v>>2]|0; //@line 743 "modules/audio/openal/Source.cpp"
  HEAP32[$13>>2] = $14; //@line 743 "modules/audio/openal/Source.cpp"
  $15 = ((($v)) + 4|0); //@line 744 "modules/audio/openal/Source.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 744 "modules/audio/openal/Source.cpp"
  $17 = ((($this)) + 96|0); //@line 744 "modules/audio/openal/Source.cpp"
  HEAP32[$17>>2] = $16; //@line 744 "modules/audio/openal/Source.cpp"
  $18 = ((($v)) + 8|0); //@line 745 "modules/audio/openal/Source.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 745 "modules/audio/openal/Source.cpp"
  $20 = ((($this)) + 100|0); //@line 745 "modules/audio/openal/Source.cpp"
  HEAP32[$20>>2] = $19; //@line 745 "modules/audio/openal/Source.cpp"
  STACKTOP = sp;return; //@line 531 "modules/audio/openal/Source.cpp"
 } else {
  $11 = ((($this)) + 16|0); //@line 528 "modules/audio/openal/Source.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 528 "modules/audio/openal/Source.cpp"
  _alSourcefv(($12|0),4101,($v|0)); //@line 528 "modules/audio/openal/Source.cpp"
  STACKTOP = sp;return; //@line 531 "modules/audio/openal/Source.cpp"
 }
}
function __ZNK4love5audio6openal6Source12getDirectionEPf($this,$v) {
 $this = $this|0;
 $v = $v|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 152|0); //@line 535 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 535 "modules/audio/openal/Source.cpp"
 $2 = ($1|0)>(1); //@line 535 "modules/audio/openal/Source.cpp"
 if ($2) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 536 "modules/audio/openal/Source.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($3|0),(365088|0),($vararg_buffer|0)); //@line 60 "modules/audio/openal/Source.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $6 = ___cxa_find_matching_catch_2()|0;
   $7 = tempRet0;
   ___cxa_free_exception(($3|0)); //@line 536 "modules/audio/openal/Source.cpp"
   ___resumeException($6|0);
   // unreachable;
  } else {
   HEAP32[$3>>2] = (22756); //@line 62 "modules/audio/openal/Source.cpp"
   ___cxa_throw(($3|0),(1488|0),(344|0)); //@line 536 "modules/audio/openal/Source.cpp"
   // unreachable; //@line 536 "modules/audio/openal/Source.cpp"
  }
 }
 $8 = ((($this)) + 20|0); //@line 538 "modules/audio/openal/Source.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 538 "modules/audio/openal/Source.cpp"
 $10 = ($9<<24>>24)==(0); //@line 538 "modules/audio/openal/Source.cpp"
 if ($10) {
  $13 = ((($this)) + 92|0); //@line 541 "modules/audio/openal/Source.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 743 "modules/audio/openal/Source.cpp"
  HEAP32[$v>>2] = $14; //@line 743 "modules/audio/openal/Source.cpp"
  $15 = ((($this)) + 96|0); //@line 744 "modules/audio/openal/Source.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 744 "modules/audio/openal/Source.cpp"
  $17 = ((($v)) + 4|0); //@line 744 "modules/audio/openal/Source.cpp"
  HEAP32[$17>>2] = $16; //@line 744 "modules/audio/openal/Source.cpp"
  $18 = ((($this)) + 100|0); //@line 745 "modules/audio/openal/Source.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 745 "modules/audio/openal/Source.cpp"
  $20 = ((($v)) + 8|0); //@line 745 "modules/audio/openal/Source.cpp"
  HEAP32[$20>>2] = $19; //@line 745 "modules/audio/openal/Source.cpp"
  STACKTOP = sp;return; //@line 542 "modules/audio/openal/Source.cpp"
 } else {
  $11 = ((($this)) + 16|0); //@line 539 "modules/audio/openal/Source.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 539 "modules/audio/openal/Source.cpp"
  _alGetSourcefv(($12|0),4101,($v|0)); //@line 539 "modules/audio/openal/Source.cpp"
  STACKTOP = sp;return; //@line 542 "modules/audio/openal/Source.cpp"
 }
}
function __ZN4love5audio6openal6Source7setConeEfff($this,$innerAngle,$outerAngle,$outerVolume) {
 $this = $this|0;
 $innerAngle = +$innerAngle;
 $outerAngle = +$outerAngle;
 $outerVolume = +$outerVolume;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0.0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0.0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 152|0); //@line 546 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 546 "modules/audio/openal/Source.cpp"
 $2 = ($1|0)>(1); //@line 546 "modules/audio/openal/Source.cpp"
 if ($2) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 547 "modules/audio/openal/Source.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($3|0),(365088|0),($vararg_buffer|0)); //@line 60 "modules/audio/openal/Source.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $6 = ___cxa_find_matching_catch_2()|0;
   $7 = tempRet0;
   ___cxa_free_exception(($3|0)); //@line 547 "modules/audio/openal/Source.cpp"
   ___resumeException($6|0);
   // unreachable;
  } else {
   HEAP32[$3>>2] = (22756); //@line 62 "modules/audio/openal/Source.cpp"
   ___cxa_throw(($3|0),(1488|0),(344|0)); //@line 547 "modules/audio/openal/Source.cpp"
   // unreachable; //@line 547 "modules/audio/openal/Source.cpp"
  }
 }
 $8 = $innerAngle * 57.295780181884766; //@line 549 "modules/audio/openal/Source.cpp"
 $9 = (~~(($8))); //@line 555 "modules/audio/openal/Source.cpp"
 $10 = ((($this)) + 128|0); //@line 549 "modules/audio/openal/Source.cpp"
 HEAP32[$10>>2] = $9; //@line 549 "modules/audio/openal/Source.cpp"
 $11 = $outerAngle * 57.295780181884766; //@line 550 "modules/audio/openal/Source.cpp"
 $12 = (~~(($11))); //@line 550 "modules/audio/openal/Source.cpp"
 $13 = ((($this)) + 132|0); //@line 550 "modules/audio/openal/Source.cpp"
 HEAP32[$13>>2] = $12; //@line 550 "modules/audio/openal/Source.cpp"
 $14 = ((($this)) + 136|0); //@line 551 "modules/audio/openal/Source.cpp"
 HEAPF32[$14>>2] = $outerVolume; //@line 551 "modules/audio/openal/Source.cpp"
 $15 = ((($this)) + 20|0); //@line 553 "modules/audio/openal/Source.cpp"
 $16 = HEAP8[$15>>0]|0; //@line 553 "modules/audio/openal/Source.cpp"
 $17 = ($16<<24>>24)==(0); //@line 553 "modules/audio/openal/Source.cpp"
 if ($17) {
  STACKTOP = sp;return; //@line 559 "modules/audio/openal/Source.cpp"
 }
 $18 = ((($this)) + 16|0); //@line 555 "modules/audio/openal/Source.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 555 "modules/audio/openal/Source.cpp"
 _alSourcei(($19|0),4097,($9|0)); //@line 555 "modules/audio/openal/Source.cpp"
 $20 = HEAP32[$18>>2]|0; //@line 556 "modules/audio/openal/Source.cpp"
 $21 = HEAP32[$13>>2]|0; //@line 556 "modules/audio/openal/Source.cpp"
 _alSourcei(($20|0),4098,($21|0)); //@line 556 "modules/audio/openal/Source.cpp"
 $22 = HEAP32[$18>>2]|0; //@line 557 "modules/audio/openal/Source.cpp"
 $23 = +HEAPF32[$14>>2]; //@line 557 "modules/audio/openal/Source.cpp"
 _alSourcef(($22|0),4130,(+$23)); //@line 557 "modules/audio/openal/Source.cpp"
 STACKTOP = sp;return; //@line 559 "modules/audio/openal/Source.cpp"
}
function __ZNK4love5audio6openal6Source7getConeERfS3_S3_($this,$innerAngle,$outerAngle,$outerVolume) {
 $this = $this|0;
 $innerAngle = $innerAngle|0;
 $outerAngle = $outerAngle|0;
 $outerVolume = $outerVolume|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 152|0); //@line 563 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 563 "modules/audio/openal/Source.cpp"
 $2 = ($1|0)>(1); //@line 563 "modules/audio/openal/Source.cpp"
 if (!($2)) {
  $8 = ((($this)) + 128|0); //@line 566 "modules/audio/openal/Source.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 566 "modules/audio/openal/Source.cpp"
  $10 = (+($9|0)); //@line 566 "modules/audio/openal/Source.cpp"
  $11 = $10 * 0.01745329238474369; //@line 566 "modules/audio/openal/Source.cpp"
  HEAPF32[$innerAngle>>2] = $11; //@line 566 "modules/audio/openal/Source.cpp"
  $12 = ((($this)) + 132|0); //@line 567 "modules/audio/openal/Source.cpp"
  $13 = HEAP32[$12>>2]|0; //@line 567 "modules/audio/openal/Source.cpp"
  $14 = (+($13|0)); //@line 567 "modules/audio/openal/Source.cpp"
  $15 = $14 * 0.01745329238474369; //@line 567 "modules/audio/openal/Source.cpp"
  HEAPF32[$outerAngle>>2] = $15; //@line 567 "modules/audio/openal/Source.cpp"
  $16 = ((($this)) + 136|0); //@line 568 "modules/audio/openal/Source.cpp"
  $17 = HEAP32[$16>>2]|0; //@line 568 "modules/audio/openal/Source.cpp"
  HEAP32[$outerVolume>>2] = $17; //@line 568 "modules/audio/openal/Source.cpp"
  STACKTOP = sp;return; //@line 569 "modules/audio/openal/Source.cpp"
 }
 $3 = (___cxa_allocate_exception(16)|0); //@line 564 "modules/audio/openal/Source.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($3|0),(365088|0),($vararg_buffer|0)); //@line 60 "modules/audio/openal/Source.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  ___cxa_free_exception(($3|0)); //@line 564 "modules/audio/openal/Source.cpp"
  ___resumeException($6|0);
  // unreachable;
 } else {
  HEAP32[$3>>2] = (22756); //@line 62 "modules/audio/openal/Source.cpp"
  ___cxa_throw(($3|0),(1488|0),(344|0)); //@line 564 "modules/audio/openal/Source.cpp"
  // unreachable; //@line 564 "modules/audio/openal/Source.cpp"
 }
}
function __ZN4love5audio6openal6Source11setRelativeEb($this,$enable) {
 $this = $this|0;
 $enable = $enable|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 152|0); //@line 573 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 573 "modules/audio/openal/Source.cpp"
 $2 = ($1|0)>(1); //@line 573 "modules/audio/openal/Source.cpp"
 if ($2) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 574 "modules/audio/openal/Source.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($3|0),(365088|0),($vararg_buffer|0)); //@line 60 "modules/audio/openal/Source.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $6 = ___cxa_find_matching_catch_2()|0;
   $7 = tempRet0;
   ___cxa_free_exception(($3|0)); //@line 574 "modules/audio/openal/Source.cpp"
   ___resumeException($6|0);
   // unreachable;
  } else {
   HEAP32[$3>>2] = (22756); //@line 62 "modules/audio/openal/Source.cpp"
   ___cxa_throw(($3|0),(1488|0),(344|0)); //@line 574 "modules/audio/openal/Source.cpp"
   // unreachable; //@line 574 "modules/audio/openal/Source.cpp"
  }
 }
 $8 = ((($this)) + 20|0); //@line 576 "modules/audio/openal/Source.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 576 "modules/audio/openal/Source.cpp"
 $10 = ($9<<24>>24)==(0); //@line 576 "modules/audio/openal/Source.cpp"
 if ($10) {
  $14 = ((($this)) + 104|0); //@line 579 "modules/audio/openal/Source.cpp"
  $15 = $enable&1; //@line 579 "modules/audio/openal/Source.cpp"
  HEAP8[$14>>0] = $15; //@line 579 "modules/audio/openal/Source.cpp"
  STACKTOP = sp;return; //@line 580 "modules/audio/openal/Source.cpp"
 }
 $11 = ((($this)) + 16|0); //@line 577 "modules/audio/openal/Source.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 577 "modules/audio/openal/Source.cpp"
 $13 = $enable&1; //@line 577 "modules/audio/openal/Source.cpp"
 _alSourcei(($12|0),514,($13|0)); //@line 577 "modules/audio/openal/Source.cpp"
 $14 = ((($this)) + 104|0); //@line 579 "modules/audio/openal/Source.cpp"
 $15 = $enable&1; //@line 579 "modules/audio/openal/Source.cpp"
 HEAP8[$14>>0] = $15; //@line 579 "modules/audio/openal/Source.cpp"
 STACKTOP = sp;return; //@line 580 "modules/audio/openal/Source.cpp"
}
function __ZNK4love5audio6openal6Source10isRelativeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 152|0); //@line 584 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 584 "modules/audio/openal/Source.cpp"
 $2 = ($1|0)>(1); //@line 584 "modules/audio/openal/Source.cpp"
 if (!($2)) {
  $8 = ((($this)) + 104|0); //@line 587 "modules/audio/openal/Source.cpp"
  $9 = HEAP8[$8>>0]|0; //@line 587 "modules/audio/openal/Source.cpp"
  $10 = ($9<<24>>24)!=(0); //@line 587 "modules/audio/openal/Source.cpp"
  STACKTOP = sp;return ($10|0); //@line 587 "modules/audio/openal/Source.cpp"
 }
 $3 = (___cxa_allocate_exception(16)|0); //@line 585 "modules/audio/openal/Source.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($3|0),(365088|0),($vararg_buffer|0)); //@line 60 "modules/audio/openal/Source.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  ___cxa_free_exception(($3|0)); //@line 585 "modules/audio/openal/Source.cpp"
  ___resumeException($6|0);
  // unreachable;
 } else {
  HEAP32[$3>>2] = (22756); //@line 62 "modules/audio/openal/Source.cpp"
  ___cxa_throw(($3|0),(1488|0),(344|0)); //@line 585 "modules/audio/openal/Source.cpp"
  // unreachable; //@line 585 "modules/audio/openal/Source.cpp"
 }
 return (0)|0;
}
function __ZN4love5audio6openal6Source10setLoopingEb($this,$enable) {
 $this = $this|0;
 $enable = $enable|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 592 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 592 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 592 "modules/audio/openal/Source.cpp"
 if (!($2)) {
  $3 = ((($this)) + 8|0); //@line 592 "modules/audio/openal/Source.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 592 "modules/audio/openal/Source.cpp"
  $5 = ($4|0)==(0); //@line 592 "modules/audio/openal/Source.cpp"
  if ($5) {
   $6 = ((($this)) + 16|0); //@line 593 "modules/audio/openal/Source.cpp"
   $7 = HEAP32[$6>>2]|0; //@line 593 "modules/audio/openal/Source.cpp"
   $8 = $enable&1; //@line 593 "modules/audio/openal/Source.cpp"
   _alSourcei(($7|0),4103,($8|0)); //@line 593 "modules/audio/openal/Source.cpp"
  }
 }
 $9 = ((($this)) + 105|0); //@line 595 "modules/audio/openal/Source.cpp"
 $10 = $enable&1; //@line 595 "modules/audio/openal/Source.cpp"
 HEAP8[$9>>0] = $10; //@line 595 "modules/audio/openal/Source.cpp"
 return; //@line 596 "modules/audio/openal/Source.cpp"
}
function __ZNK4love5audio6openal6Source9isLoopingEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 105|0); //@line 600 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 600 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 600 "modules/audio/openal/Source.cpp"
 return ($2|0); //@line 600 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source12setMinVolumeEf($this,$volume) {
 $this = $this|0;
 $volume = +$volume;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 818 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 818 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 818 "modules/audio/openal/Source.cpp"
 if (!($2)) {
  $3 = ((($this)) + 16|0); //@line 819 "modules/audio/openal/Source.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 819 "modules/audio/openal/Source.cpp"
  _alSourcef(($4|0),4109,(+$volume)); //@line 819 "modules/audio/openal/Source.cpp"
 }
 $5 = ((($this)) + 108|0); //@line 821 "modules/audio/openal/Source.cpp"
 HEAPF32[$5>>2] = $volume; //@line 821 "modules/audio/openal/Source.cpp"
 return; //@line 822 "modules/audio/openal/Source.cpp"
}
function __ZNK4love5audio6openal6Source12getMinVolumeEv($this) {
 $this = $this|0;
 var $$0 = 0.0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $f = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f = sp;
 $0 = ((($this)) + 20|0); //@line 826 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 826 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 826 "modules/audio/openal/Source.cpp"
 if ($2) {
  $6 = ((($this)) + 108|0); //@line 834 "modules/audio/openal/Source.cpp"
  $7 = +HEAPF32[$6>>2]; //@line 834 "modules/audio/openal/Source.cpp"
  $$0 = $7;
  STACKTOP = sp;return (+$$0); //@line 835 "modules/audio/openal/Source.cpp"
 } else {
  $3 = ((($this)) + 16|0); //@line 829 "modules/audio/openal/Source.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 829 "modules/audio/openal/Source.cpp"
  _alGetSourcef(($4|0),4109,($f|0)); //@line 829 "modules/audio/openal/Source.cpp"
  $5 = +HEAPF32[$f>>2]; //@line 830 "modules/audio/openal/Source.cpp"
  $$0 = $5;
  STACKTOP = sp;return (+$$0); //@line 835 "modules/audio/openal/Source.cpp"
 }
 return +(0.0);
}
function __ZN4love5audio6openal6Source12setMaxVolumeEf($this,$volume) {
 $this = $this|0;
 $volume = +$volume;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 839 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 839 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 839 "modules/audio/openal/Source.cpp"
 if (!($2)) {
  $3 = ((($this)) + 16|0); //@line 840 "modules/audio/openal/Source.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 840 "modules/audio/openal/Source.cpp"
  _alSourcef(($4|0),4110,(+$volume)); //@line 840 "modules/audio/openal/Source.cpp"
 }
 $5 = ((($this)) + 112|0); //@line 842 "modules/audio/openal/Source.cpp"
 HEAPF32[$5>>2] = $volume; //@line 842 "modules/audio/openal/Source.cpp"
 return; //@line 843 "modules/audio/openal/Source.cpp"
}
function __ZNK4love5audio6openal6Source12getMaxVolumeEv($this) {
 $this = $this|0;
 var $$0 = 0.0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $f = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f = sp;
 $0 = ((($this)) + 20|0); //@line 847 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 847 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 847 "modules/audio/openal/Source.cpp"
 if ($2) {
  $6 = ((($this)) + 112|0); //@line 855 "modules/audio/openal/Source.cpp"
  $7 = +HEAPF32[$6>>2]; //@line 855 "modules/audio/openal/Source.cpp"
  $$0 = $7;
  STACKTOP = sp;return (+$$0); //@line 856 "modules/audio/openal/Source.cpp"
 } else {
  $3 = ((($this)) + 16|0); //@line 850 "modules/audio/openal/Source.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 850 "modules/audio/openal/Source.cpp"
  _alGetSourcef(($4|0),4110,($f|0)); //@line 850 "modules/audio/openal/Source.cpp"
  $5 = +HEAPF32[$f>>2]; //@line 851 "modules/audio/openal/Source.cpp"
  $$0 = $5;
  STACKTOP = sp;return (+$$0); //@line 856 "modules/audio/openal/Source.cpp"
 }
 return +(0.0);
}
function __ZN4love5audio6openal6Source20setReferenceDistanceEf($this,$distance) {
 $this = $this|0;
 $distance = +$distance;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 152|0); //@line 860 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 860 "modules/audio/openal/Source.cpp"
 $2 = ($1|0)>(1); //@line 860 "modules/audio/openal/Source.cpp"
 if ($2) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 861 "modules/audio/openal/Source.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($3|0),(365088|0),($vararg_buffer|0)); //@line 60 "modules/audio/openal/Source.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $6 = ___cxa_find_matching_catch_2()|0;
   $7 = tempRet0;
   ___cxa_free_exception(($3|0)); //@line 861 "modules/audio/openal/Source.cpp"
   ___resumeException($6|0);
   // unreachable;
  } else {
   HEAP32[$3>>2] = (22756); //@line 62 "modules/audio/openal/Source.cpp"
   ___cxa_throw(($3|0),(1488|0),(344|0)); //@line 861 "modules/audio/openal/Source.cpp"
   // unreachable; //@line 861 "modules/audio/openal/Source.cpp"
  }
 }
 $8 = ((($this)) + 20|0); //@line 863 "modules/audio/openal/Source.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 863 "modules/audio/openal/Source.cpp"
 $10 = ($9<<24>>24)==(0); //@line 863 "modules/audio/openal/Source.cpp"
 if ($10) {
  $13 = ((($this)) + 116|0); //@line 866 "modules/audio/openal/Source.cpp"
  HEAPF32[$13>>2] = $distance; //@line 866 "modules/audio/openal/Source.cpp"
  STACKTOP = sp;return; //@line 867 "modules/audio/openal/Source.cpp"
 }
 $11 = ((($this)) + 16|0); //@line 864 "modules/audio/openal/Source.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 864 "modules/audio/openal/Source.cpp"
 _alSourcef(($12|0),4128,(+$distance)); //@line 864 "modules/audio/openal/Source.cpp"
 $13 = ((($this)) + 116|0); //@line 866 "modules/audio/openal/Source.cpp"
 HEAPF32[$13>>2] = $distance; //@line 866 "modules/audio/openal/Source.cpp"
 STACKTOP = sp;return; //@line 867 "modules/audio/openal/Source.cpp"
}
function __ZNK4love5audio6openal6Source20getReferenceDistanceEv($this) {
 $this = $this|0;
 var $$0 = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $f = sp + 4|0;
 $0 = ((($this)) + 152|0); //@line 871 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 871 "modules/audio/openal/Source.cpp"
 $2 = ($1|0)>(1); //@line 871 "modules/audio/openal/Source.cpp"
 if ($2) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 872 "modules/audio/openal/Source.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($3|0),(365088|0),($vararg_buffer|0)); //@line 60 "modules/audio/openal/Source.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $6 = ___cxa_find_matching_catch_2()|0;
   $7 = tempRet0;
   ___cxa_free_exception(($3|0)); //@line 872 "modules/audio/openal/Source.cpp"
   ___resumeException($6|0);
   // unreachable;
  } else {
   HEAP32[$3>>2] = (22756); //@line 62 "modules/audio/openal/Source.cpp"
   ___cxa_throw(($3|0),(1488|0),(344|0)); //@line 872 "modules/audio/openal/Source.cpp"
   // unreachable; //@line 872 "modules/audio/openal/Source.cpp"
  }
 }
 $8 = ((($this)) + 20|0); //@line 874 "modules/audio/openal/Source.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 874 "modules/audio/openal/Source.cpp"
 $10 = ($9<<24>>24)==(0); //@line 874 "modules/audio/openal/Source.cpp"
 if ($10) {
  $14 = ((($this)) + 116|0); //@line 882 "modules/audio/openal/Source.cpp"
  $15 = +HEAPF32[$14>>2]; //@line 882 "modules/audio/openal/Source.cpp"
  $$0 = $15;
  STACKTOP = sp;return (+$$0); //@line 883 "modules/audio/openal/Source.cpp"
 } else {
  $11 = ((($this)) + 16|0); //@line 877 "modules/audio/openal/Source.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 877 "modules/audio/openal/Source.cpp"
  _alGetSourcef(($12|0),4128,($f|0)); //@line 877 "modules/audio/openal/Source.cpp"
  $13 = +HEAPF32[$f>>2]; //@line 878 "modules/audio/openal/Source.cpp"
  $$0 = $13;
  STACKTOP = sp;return (+$$0); //@line 883 "modules/audio/openal/Source.cpp"
 }
 return +(0.0);
}
function __ZN4love5audio6openal6Source16setRolloffFactorEf($this,$factor) {
 $this = $this|0;
 $factor = +$factor;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 152|0); //@line 887 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 887 "modules/audio/openal/Source.cpp"
 $2 = ($1|0)>(1); //@line 887 "modules/audio/openal/Source.cpp"
 if ($2) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 888 "modules/audio/openal/Source.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($3|0),(365088|0),($vararg_buffer|0)); //@line 60 "modules/audio/openal/Source.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $6 = ___cxa_find_matching_catch_2()|0;
   $7 = tempRet0;
   ___cxa_free_exception(($3|0)); //@line 888 "modules/audio/openal/Source.cpp"
   ___resumeException($6|0);
   // unreachable;
  } else {
   HEAP32[$3>>2] = (22756); //@line 62 "modules/audio/openal/Source.cpp"
   ___cxa_throw(($3|0),(1488|0),(344|0)); //@line 888 "modules/audio/openal/Source.cpp"
   // unreachable; //@line 888 "modules/audio/openal/Source.cpp"
  }
 }
 $8 = ((($this)) + 20|0); //@line 890 "modules/audio/openal/Source.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 890 "modules/audio/openal/Source.cpp"
 $10 = ($9<<24>>24)==(0); //@line 890 "modules/audio/openal/Source.cpp"
 if ($10) {
  $13 = ((($this)) + 120|0); //@line 893 "modules/audio/openal/Source.cpp"
  HEAPF32[$13>>2] = $factor; //@line 893 "modules/audio/openal/Source.cpp"
  STACKTOP = sp;return; //@line 894 "modules/audio/openal/Source.cpp"
 }
 $11 = ((($this)) + 16|0); //@line 891 "modules/audio/openal/Source.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 891 "modules/audio/openal/Source.cpp"
 _alSourcef(($12|0),4129,(+$factor)); //@line 891 "modules/audio/openal/Source.cpp"
 $13 = ((($this)) + 120|0); //@line 893 "modules/audio/openal/Source.cpp"
 HEAPF32[$13>>2] = $factor; //@line 893 "modules/audio/openal/Source.cpp"
 STACKTOP = sp;return; //@line 894 "modules/audio/openal/Source.cpp"
}
function __ZNK4love5audio6openal6Source16getRolloffFactorEv($this) {
 $this = $this|0;
 var $$0 = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $f = sp + 4|0;
 $0 = ((($this)) + 152|0); //@line 898 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 898 "modules/audio/openal/Source.cpp"
 $2 = ($1|0)>(1); //@line 898 "modules/audio/openal/Source.cpp"
 if ($2) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 899 "modules/audio/openal/Source.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($3|0),(365088|0),($vararg_buffer|0)); //@line 60 "modules/audio/openal/Source.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $6 = ___cxa_find_matching_catch_2()|0;
   $7 = tempRet0;
   ___cxa_free_exception(($3|0)); //@line 899 "modules/audio/openal/Source.cpp"
   ___resumeException($6|0);
   // unreachable;
  } else {
   HEAP32[$3>>2] = (22756); //@line 62 "modules/audio/openal/Source.cpp"
   ___cxa_throw(($3|0),(1488|0),(344|0)); //@line 899 "modules/audio/openal/Source.cpp"
   // unreachable; //@line 899 "modules/audio/openal/Source.cpp"
  }
 }
 $8 = ((($this)) + 20|0); //@line 901 "modules/audio/openal/Source.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 901 "modules/audio/openal/Source.cpp"
 $10 = ($9<<24>>24)==(0); //@line 901 "modules/audio/openal/Source.cpp"
 if ($10) {
  $14 = ((($this)) + 120|0); //@line 909 "modules/audio/openal/Source.cpp"
  $15 = +HEAPF32[$14>>2]; //@line 909 "modules/audio/openal/Source.cpp"
  $$0 = $15;
  STACKTOP = sp;return (+$$0); //@line 910 "modules/audio/openal/Source.cpp"
 } else {
  $11 = ((($this)) + 16|0); //@line 904 "modules/audio/openal/Source.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 904 "modules/audio/openal/Source.cpp"
  _alGetSourcef(($12|0),4129,($f|0)); //@line 904 "modules/audio/openal/Source.cpp"
  $13 = +HEAPF32[$f>>2]; //@line 905 "modules/audio/openal/Source.cpp"
  $$0 = $13;
  STACKTOP = sp;return (+$$0); //@line 910 "modules/audio/openal/Source.cpp"
 }
 return +(0.0);
}
function __ZN4love5audio6openal6Source14setMaxDistanceEf($this,$distance) {
 $this = $this|0;
 $distance = +$distance;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 152|0); //@line 914 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 914 "modules/audio/openal/Source.cpp"
 $2 = ($1|0)>(1); //@line 914 "modules/audio/openal/Source.cpp"
 if ($2) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 915 "modules/audio/openal/Source.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($3|0),(365088|0),($vararg_buffer|0)); //@line 60 "modules/audio/openal/Source.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $6 = ___cxa_find_matching_catch_2()|0;
   $7 = tempRet0;
   ___cxa_free_exception(($3|0)); //@line 915 "modules/audio/openal/Source.cpp"
   ___resumeException($6|0);
   // unreachable;
  } else {
   HEAP32[$3>>2] = (22756); //@line 62 "modules/audio/openal/Source.cpp"
   ___cxa_throw(($3|0),(1488|0),(344|0)); //@line 915 "modules/audio/openal/Source.cpp"
   // unreachable; //@line 915 "modules/audio/openal/Source.cpp"
  }
 }
 $8 = $distance > 3.4028234663852886E+38; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $9 = $8 ? 3.4028234663852886E+38 : $distance; //@line 917 "modules/audio/openal/Source.cpp"
 $10 = ((($this)) + 20|0); //@line 919 "modules/audio/openal/Source.cpp"
 $11 = HEAP8[$10>>0]|0; //@line 919 "modules/audio/openal/Source.cpp"
 $12 = ($11<<24>>24)==(0); //@line 919 "modules/audio/openal/Source.cpp"
 if ($12) {
  $15 = ((($this)) + 124|0); //@line 922 "modules/audio/openal/Source.cpp"
  HEAPF32[$15>>2] = $9; //@line 922 "modules/audio/openal/Source.cpp"
  STACKTOP = sp;return; //@line 923 "modules/audio/openal/Source.cpp"
 }
 $13 = ((($this)) + 16|0); //@line 920 "modules/audio/openal/Source.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 920 "modules/audio/openal/Source.cpp"
 _alSourcef(($14|0),4131,(+$9)); //@line 920 "modules/audio/openal/Source.cpp"
 $15 = ((($this)) + 124|0); //@line 922 "modules/audio/openal/Source.cpp"
 HEAPF32[$15>>2] = $9; //@line 922 "modules/audio/openal/Source.cpp"
 STACKTOP = sp;return; //@line 923 "modules/audio/openal/Source.cpp"
}
function __ZNK4love5audio6openal6Source14getMaxDistanceEv($this) {
 $this = $this|0;
 var $$0 = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $f = sp + 4|0;
 $0 = ((($this)) + 152|0); //@line 927 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 927 "modules/audio/openal/Source.cpp"
 $2 = ($1|0)>(1); //@line 927 "modules/audio/openal/Source.cpp"
 if ($2) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 928 "modules/audio/openal/Source.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($3|0),(365088|0),($vararg_buffer|0)); //@line 60 "modules/audio/openal/Source.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $6 = ___cxa_find_matching_catch_2()|0;
   $7 = tempRet0;
   ___cxa_free_exception(($3|0)); //@line 928 "modules/audio/openal/Source.cpp"
   ___resumeException($6|0);
   // unreachable;
  } else {
   HEAP32[$3>>2] = (22756); //@line 62 "modules/audio/openal/Source.cpp"
   ___cxa_throw(($3|0),(1488|0),(344|0)); //@line 928 "modules/audio/openal/Source.cpp"
   // unreachable; //@line 928 "modules/audio/openal/Source.cpp"
  }
 }
 $8 = ((($this)) + 20|0); //@line 930 "modules/audio/openal/Source.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 930 "modules/audio/openal/Source.cpp"
 $10 = ($9<<24>>24)==(0); //@line 930 "modules/audio/openal/Source.cpp"
 if ($10) {
  $14 = ((($this)) + 124|0); //@line 938 "modules/audio/openal/Source.cpp"
  $15 = +HEAPF32[$14>>2]; //@line 938 "modules/audio/openal/Source.cpp"
  $$0 = $15;
  STACKTOP = sp;return (+$$0); //@line 939 "modules/audio/openal/Source.cpp"
 } else {
  $11 = ((($this)) + 16|0); //@line 933 "modules/audio/openal/Source.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 933 "modules/audio/openal/Source.cpp"
  _alGetSourcef(($12|0),4131,($f|0)); //@line 933 "modules/audio/openal/Source.cpp"
  $13 = +HEAPF32[$f>>2]; //@line 934 "modules/audio/openal/Source.cpp"
  $$0 = $13;
  STACKTOP = sp;return (+$$0); //@line 939 "modules/audio/openal/Source.cpp"
 }
 return +(0.0);
}
function __ZNK4love5audio6openal6Source11getChannelsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 152|0); //@line 943 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 943 "modules/audio/openal/Source.cpp"
 return ($1|0); //@line 943 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source10seekAtomicEfPv($this,$offset,$unit) {
 $this = $this|0;
 $offset = +$offset;
 $unit = $unit|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 372 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 372 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 372 "modules/audio/openal/Source.cpp"
 if ($2) {
  return; //@line 411 "modules/audio/openal/Source.cpp"
 }
 $3 = HEAP32[$unit>>2]|0; //@line 374 "modules/audio/openal/Source.cpp"
 $cond = ($3|0)==(1); //@line 374 "modules/audio/openal/Source.cpp"
 $4 = ((($this)) + 8|0); //@line 377 "modules/audio/openal/Source.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 377 "modules/audio/openal/Source.cpp"
 $6 = ($5|0)==(1); //@line 377 "modules/audio/openal/Source.cpp"
 do {
  if ($cond) {
   if ($6) {
    $7 = ((($this)) + 140|0); //@line 379 "modules/audio/openal/Source.cpp"
    HEAPF32[$7>>2] = $offset; //@line 379 "modules/audio/openal/Source.cpp"
    $8 = ((($this)) + 160|0); //@line 117 "./common/Object.h"
    $9 = HEAP32[$8>>2]|0; //@line 117 "./common/Object.h"
    $10 = HEAP32[$9>>2]|0; //@line 380 "modules/audio/openal/Source.cpp"
    $11 = ((($10)) + 56|0); //@line 380 "modules/audio/openal/Source.cpp"
    $12 = HEAP32[$11>>2]|0; //@line 380 "modules/audio/openal/Source.cpp"
    $13 = (FUNCTION_TABLE_ii[$12 & 8191]($9)|0); //@line 380 "modules/audio/openal/Source.cpp"
    $14 = (+($13|0)); //@line 380 "modules/audio/openal/Source.cpp"
    $15 = $offset / $14; //@line 380 "modules/audio/openal/Source.cpp"
    $16 = ((($this)) + 144|0); //@line 381 "modules/audio/openal/Source.cpp"
    HEAPF32[$16>>2] = $15; //@line 381 "modules/audio/openal/Source.cpp"
    $17 = HEAP32[$8>>2]|0; //@line 117 "./common/Object.h"
    $18 = HEAP32[$17>>2]|0; //@line 382 "modules/audio/openal/Source.cpp"
    $19 = ((($18)) + 32|0); //@line 382 "modules/audio/openal/Source.cpp"
    $20 = HEAP32[$19>>2]|0; //@line 382 "modules/audio/openal/Source.cpp"
    (FUNCTION_TABLE_iid[$20 & 4095]($17,$15)|0); //@line 382 "modules/audio/openal/Source.cpp"
    break;
   } else {
    $21 = ((($this)) + 16|0); //@line 385 "modules/audio/openal/Source.cpp"
    $22 = HEAP32[$21>>2]|0; //@line 385 "modules/audio/openal/Source.cpp"
    _alSourcef(($22|0),4133,(+$offset)); //@line 385 "modules/audio/openal/Source.cpp"
    break;
   }
  } else {
   if ($6) {
    $23 = ((($this)) + 144|0); //@line 391 "modules/audio/openal/Source.cpp"
    HEAPF32[$23>>2] = $offset; //@line 391 "modules/audio/openal/Source.cpp"
    $24 = ((($this)) + 160|0); //@line 117 "./common/Object.h"
    $25 = HEAP32[$24>>2]|0; //@line 117 "./common/Object.h"
    $26 = HEAP32[$25>>2]|0; //@line 392 "modules/audio/openal/Source.cpp"
    $27 = ((($26)) + 32|0); //@line 392 "modules/audio/openal/Source.cpp"
    $28 = HEAP32[$27>>2]|0; //@line 392 "modules/audio/openal/Source.cpp"
    (FUNCTION_TABLE_iid[$28 & 4095]($25,$offset)|0); //@line 392 "modules/audio/openal/Source.cpp"
    $29 = HEAP32[$24>>2]|0; //@line 117 "./common/Object.h"
    $30 = HEAP32[$29>>2]|0; //@line 393 "modules/audio/openal/Source.cpp"
    $31 = ((($30)) + 56|0); //@line 393 "modules/audio/openal/Source.cpp"
    $32 = HEAP32[$31>>2]|0; //@line 393 "modules/audio/openal/Source.cpp"
    $33 = (FUNCTION_TABLE_ii[$32 & 8191]($29)|0); //@line 393 "modules/audio/openal/Source.cpp"
    $34 = (+($33|0)); //@line 393 "modules/audio/openal/Source.cpp"
    $35 = $34 * $offset; //@line 393 "modules/audio/openal/Source.cpp"
    $36 = ((($this)) + 140|0); //@line 393 "modules/audio/openal/Source.cpp"
    HEAPF32[$36>>2] = $35; //@line 393 "modules/audio/openal/Source.cpp"
    break;
   } else {
    $37 = ((($this)) + 16|0); //@line 396 "modules/audio/openal/Source.cpp"
    $38 = HEAP32[$37>>2]|0; //@line 396 "modules/audio/openal/Source.cpp"
    _alSourcef(($38|0),4132,(+$offset)); //@line 396 "modules/audio/openal/Source.cpp"
    break;
   }
  }
 } while(0);
 $39 = HEAP32[$4>>2]|0; //@line 399 "modules/audio/openal/Source.cpp"
 $40 = ($39|0)==(1); //@line 399 "modules/audio/openal/Source.cpp"
 if (!($40)) {
  return; //@line 411 "modules/audio/openal/Source.cpp"
 }
 $41 = ((($this)) + 106|0); //@line 401 "modules/audio/openal/Source.cpp"
 $42 = HEAP8[$41>>0]|0; //@line 401 "modules/audio/openal/Source.cpp"
 $43 = ($42<<24>>24)==(0); //@line 401 "modules/audio/openal/Source.cpp"
 __ZN4love5audio6openal6Source10stopAtomicEv($this); //@line 405 "modules/audio/openal/Source.cpp"
 (__ZN4love5audio6openal6Source10playAtomicEv($this)|0); //@line 406 "modules/audio/openal/Source.cpp"
 if ($43) {
  return; //@line 411 "modules/audio/openal/Source.cpp"
 }
 $44 = HEAP8[$0>>0]|0; //@line 676 "modules/audio/openal/Source.cpp"
 $45 = ($44<<24>>24)==(0); //@line 676 "modules/audio/openal/Source.cpp"
 if ($45) {
  return; //@line 411 "modules/audio/openal/Source.cpp"
 }
 $46 = ((($this)) + 16|0); //@line 678 "modules/audio/openal/Source.cpp"
 $47 = HEAP32[$46>>2]|0; //@line 678 "modules/audio/openal/Source.cpp"
 _alSourcePause(($47|0)); //@line 678 "modules/audio/openal/Source.cpp"
 HEAP8[$41>>0] = 1; //@line 679 "modules/audio/openal/Source.cpp"
 return; //@line 411 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source10stopAtomicEv($this) {
 $this = $this|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $buffer = 0, $queued = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $queued = sp + 4|0;
 $buffer = sp;
 $0 = ((($this)) + 20|0); //@line 650 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 650 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 650 "modules/audio/openal/Source.cpp"
 if ($2) {
  $18 = ((($this)) + 164|0); //@line 670 "modules/audio/openal/Source.cpp"
  HEAP32[$18>>2] = 0; //@line 670 "modules/audio/openal/Source.cpp"
  HEAP8[$0>>0] = 0; //@line 671 "modules/audio/openal/Source.cpp"
  STACKTOP = sp;return; //@line 672 "modules/audio/openal/Source.cpp"
 }
 $3 = ((($this)) + 8|0); //@line 652 "modules/audio/openal/Source.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 652 "modules/audio/openal/Source.cpp"
 switch ($4|0) {
 case 0:  {
  $5 = ((($this)) + 16|0); //@line 653 "modules/audio/openal/Source.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 653 "modules/audio/openal/Source.cpp"
  _alSourceStop(($6|0)); //@line 653 "modules/audio/openal/Source.cpp"
  $$pre$phiZ2D = $5;
  break;
 }
 case 1:  {
  $7 = ((($this)) + 16|0); //@line 656 "modules/audio/openal/Source.cpp"
  $8 = HEAP32[$7>>2]|0; //@line 656 "modules/audio/openal/Source.cpp"
  _alSourceStop(($8|0)); //@line 656 "modules/audio/openal/Source.cpp"
  HEAP32[$queued>>2] = 0; //@line 657 "modules/audio/openal/Source.cpp"
  $9 = HEAP32[$7>>2]|0; //@line 658 "modules/audio/openal/Source.cpp"
  _alGetSourcei(($9|0),4117,($queued|0)); //@line 658 "modules/audio/openal/Source.cpp"
  $10 = HEAP32[$queued>>2]|0; //@line 660 "modules/audio/openal/Source.cpp"
  $11 = (($10) + -1)|0; //@line 660 "modules/audio/openal/Source.cpp"
  HEAP32[$queued>>2] = $11; //@line 660 "modules/audio/openal/Source.cpp"
  $12 = ($10|0)==(0); //@line 660 "modules/audio/openal/Source.cpp"
  if (!($12)) {
   while(1) {
    $13 = HEAP32[$7>>2]|0; //@line 663 "modules/audio/openal/Source.cpp"
    _alSourceUnqueueBuffers(($13|0),1,($buffer|0)); //@line 663 "modules/audio/openal/Source.cpp"
    $14 = HEAP32[$queued>>2]|0; //@line 660 "modules/audio/openal/Source.cpp"
    $15 = (($14) + -1)|0; //@line 660 "modules/audio/openal/Source.cpp"
    HEAP32[$queued>>2] = $15; //@line 660 "modules/audio/openal/Source.cpp"
    $16 = ($14|0)==(0); //@line 660 "modules/audio/openal/Source.cpp"
    if ($16) {
     break;
    }
   }
  }
  $$pre$phiZ2D = $7;
  break;
 }
 default: {
  $$pre = ((($this)) + 16|0); //@line 667 "modules/audio/openal/Source.cpp"
  $$pre$phiZ2D = $$pre;
 }
 }
 $17 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 667 "modules/audio/openal/Source.cpp"
 _alSourcei(($17|0),4105,0); //@line 667 "modules/audio/openal/Source.cpp"
 $18 = ((($this)) + 164|0); //@line 670 "modules/audio/openal/Source.cpp"
 HEAP32[$18>>2] = 0; //@line 670 "modules/audio/openal/Source.cpp"
 HEAP8[$0>>0] = 0; //@line 671 "modules/audio/openal/Source.cpp"
 STACKTOP = sp;return; //@line 672 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source10playAtomicEv($this) {
 $this = $this|0;
 var $$not = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $usedBuffers$02 = 0, $usedBuffers$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 605 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 605 "modules/audio/openal/Source.cpp"
 switch ($1|0) {
 case 0:  {
  $3 = ((($this)) + 16|0); //@line 607 "modules/audio/openal/Source.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 607 "modules/audio/openal/Source.cpp"
  $5 = ((($this)) + 56|0); //@line 117 "./common/Object.h"
  $6 = HEAP32[$5>>2]|0; //@line 117 "./common/Object.h"
  $7 = ((($6)) + 8|0); //@line 65 "modules/audio/openal/Source.h"
  $8 = HEAP32[$7>>2]|0; //@line 65 "modules/audio/openal/Source.h"
  _alSourcei(($4|0),4105,($8|0)); //@line 607 "modules/audio/openal/Source.cpp"
  break;
 }
 case 1:  {
  $2 = ((($this)) + 160|0); //@line 139 "./common/Object.h"
  $usedBuffers$02 = 0;
  while(1) {
   $9 = (((($this)) + 24|0) + ($usedBuffers$02<<2)|0); //@line 615 "modules/audio/openal/Source.cpp"
   $10 = HEAP32[$9>>2]|0; //@line 615 "modules/audio/openal/Source.cpp"
   $11 = HEAP32[$2>>2]|0; //@line 139 "./common/Object.h"
   $12 = (__ZN4love5audio6openal6Source12streamAtomicEjPNS_5sound7DecoderE($this,$10,$11)|0); //@line 615 "modules/audio/openal/Source.cpp"
   $13 = ($12|0)==(0); //@line 615 "modules/audio/openal/Source.cpp"
   if ($13) {
    $usedBuffers$1 = $usedBuffers$02;
    break;
   }
   $14 = (($usedBuffers$02) + 1)|0; //@line 618 "modules/audio/openal/Source.cpp"
   $15 = HEAP32[$2>>2]|0; //@line 117 "./common/Object.h"
   $16 = HEAP32[$15>>2]|0; //@line 620 "modules/audio/openal/Source.cpp"
   $17 = ((($16)) + 44|0); //@line 620 "modules/audio/openal/Source.cpp"
   $18 = HEAP32[$17>>2]|0; //@line 620 "modules/audio/openal/Source.cpp"
   $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 620 "modules/audio/openal/Source.cpp"
   $$not = $19 ^ 1; //@line 620 "modules/audio/openal/Source.cpp"
   $20 = ($14>>>0)<(8); //@line 613 "modules/audio/openal/Source.cpp"
   $or$cond = $20 & $$not; //@line 620 "modules/audio/openal/Source.cpp"
   if ($or$cond) {
    $usedBuffers$02 = $14;
   } else {
    $usedBuffers$1 = $14;
    break;
   }
  }
  $21 = ($usedBuffers$1|0)>(0); //@line 624 "modules/audio/openal/Source.cpp"
  if ($21) {
   $22 = ((($this)) + 16|0); //@line 625 "modules/audio/openal/Source.cpp"
   $23 = HEAP32[$22>>2]|0; //@line 625 "modules/audio/openal/Source.cpp"
   $24 = ((($this)) + 24|0); //@line 625 "modules/audio/openal/Source.cpp"
   _alSourceQueueBuffers(($23|0),($usedBuffers$1|0),($24|0)); //@line 625 "modules/audio/openal/Source.cpp"
  }
  break;
 }
 default: {
 }
 }
 __ZN4love5audio6openal6Source5resetEv($this); //@line 631 "modules/audio/openal/Source.cpp"
 (_alGetError()|0); //@line 634 "modules/audio/openal/Source.cpp"
 $25 = ((($this)) + 16|0); //@line 636 "modules/audio/openal/Source.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 636 "modules/audio/openal/Source.cpp"
 _alSourcePlay(($26|0)); //@line 636 "modules/audio/openal/Source.cpp"
 $27 = (_alGetError()|0); //@line 640 "modules/audio/openal/Source.cpp"
 $28 = ($27|0)==(0); //@line 640 "modules/audio/openal/Source.cpp"
 $29 = ((($this)) + 20|0); //@line 642 "modules/audio/openal/Source.cpp"
 HEAP8[$29>>0] = 1; //@line 642 "modules/audio/openal/Source.cpp"
 return ($28|0); //@line 645 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source5resetEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $23 = 0, $24 = 0.0, $25 = 0, $26 = 0;
 var $27 = 0.0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 724 "modules/audio/openal/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 724 "modules/audio/openal/Source.cpp"
 $2 = ((($this)) + 68|0); //@line 724 "modules/audio/openal/Source.cpp"
 _alSourcefv(($1|0),4100,($2|0)); //@line 724 "modules/audio/openal/Source.cpp"
 $3 = HEAP32[$0>>2]|0; //@line 725 "modules/audio/openal/Source.cpp"
 $4 = ((($this)) + 80|0); //@line 725 "modules/audio/openal/Source.cpp"
 _alSourcefv(($3|0),4102,($4|0)); //@line 725 "modules/audio/openal/Source.cpp"
 $5 = HEAP32[$0>>2]|0; //@line 726 "modules/audio/openal/Source.cpp"
 $6 = ((($this)) + 92|0); //@line 726 "modules/audio/openal/Source.cpp"
 _alSourcefv(($5|0),4101,($6|0)); //@line 726 "modules/audio/openal/Source.cpp"
 $7 = HEAP32[$0>>2]|0; //@line 727 "modules/audio/openal/Source.cpp"
 $8 = ((($this)) + 60|0); //@line 727 "modules/audio/openal/Source.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 727 "modules/audio/openal/Source.cpp"
 _alSourcef(($7|0),4099,(+$9)); //@line 727 "modules/audio/openal/Source.cpp"
 $10 = HEAP32[$0>>2]|0; //@line 728 "modules/audio/openal/Source.cpp"
 $11 = ((($this)) + 64|0); //@line 728 "modules/audio/openal/Source.cpp"
 $12 = +HEAPF32[$11>>2]; //@line 728 "modules/audio/openal/Source.cpp"
 _alSourcef(($10|0),4106,(+$12)); //@line 728 "modules/audio/openal/Source.cpp"
 $13 = HEAP32[$0>>2]|0; //@line 729 "modules/audio/openal/Source.cpp"
 $14 = ((($this)) + 108|0); //@line 729 "modules/audio/openal/Source.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 729 "modules/audio/openal/Source.cpp"
 _alSourcef(($13|0),4109,(+$15)); //@line 729 "modules/audio/openal/Source.cpp"
 $16 = HEAP32[$0>>2]|0; //@line 730 "modules/audio/openal/Source.cpp"
 $17 = ((($this)) + 112|0); //@line 730 "modules/audio/openal/Source.cpp"
 $18 = +HEAPF32[$17>>2]; //@line 730 "modules/audio/openal/Source.cpp"
 _alSourcef(($16|0),4110,(+$18)); //@line 730 "modules/audio/openal/Source.cpp"
 $19 = HEAP32[$0>>2]|0; //@line 731 "modules/audio/openal/Source.cpp"
 $20 = ((($this)) + 116|0); //@line 731 "modules/audio/openal/Source.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 731 "modules/audio/openal/Source.cpp"
 _alSourcef(($19|0),4128,(+$21)); //@line 731 "modules/audio/openal/Source.cpp"
 $22 = HEAP32[$0>>2]|0; //@line 732 "modules/audio/openal/Source.cpp"
 $23 = ((($this)) + 120|0); //@line 732 "modules/audio/openal/Source.cpp"
 $24 = +HEAPF32[$23>>2]; //@line 732 "modules/audio/openal/Source.cpp"
 _alSourcef(($22|0),4129,(+$24)); //@line 732 "modules/audio/openal/Source.cpp"
 $25 = HEAP32[$0>>2]|0; //@line 733 "modules/audio/openal/Source.cpp"
 $26 = ((($this)) + 124|0); //@line 733 "modules/audio/openal/Source.cpp"
 $27 = +HEAPF32[$26>>2]; //@line 733 "modules/audio/openal/Source.cpp"
 _alSourcef(($25|0),4131,(+$27)); //@line 733 "modules/audio/openal/Source.cpp"
 $28 = HEAP32[$0>>2]|0; //@line 734 "modules/audio/openal/Source.cpp"
 $29 = ((($this)) + 8|0); //@line 734 "modules/audio/openal/Source.cpp"
 $30 = HEAP32[$29>>2]|0; //@line 734 "modules/audio/openal/Source.cpp"
 $31 = ($30|0)==(0); //@line 734 "modules/audio/openal/Source.cpp"
 if ($31) {
  $32 = HEAP32[$this>>2]|0; //@line 734 "modules/audio/openal/Source.cpp"
  $33 = ((($32)) + 128|0); //@line 734 "modules/audio/openal/Source.cpp"
  $34 = HEAP32[$33>>2]|0; //@line 734 "modules/audio/openal/Source.cpp"
  $35 = (FUNCTION_TABLE_ii[$34 & 8191]($this)|0); //@line 734 "modules/audio/openal/Source.cpp"
  $37 = $35;
 } else {
  $37 = 0;
 }
 $36 = $37&1; //@line 734 "modules/audio/openal/Source.cpp"
 _alSourcei(($28|0),4103,($36|0)); //@line 734 "modules/audio/openal/Source.cpp"
 $38 = HEAP32[$0>>2]|0; //@line 735 "modules/audio/openal/Source.cpp"
 $39 = ((($this)) + 104|0); //@line 735 "modules/audio/openal/Source.cpp"
 $40 = HEAP8[$39>>0]|0; //@line 735 "modules/audio/openal/Source.cpp"
 $41 = $40&255; //@line 735 "modules/audio/openal/Source.cpp"
 _alSourcei(($38|0),514,($41|0)); //@line 735 "modules/audio/openal/Source.cpp"
 $42 = HEAP32[$0>>2]|0; //@line 736 "modules/audio/openal/Source.cpp"
 $43 = ((($this)) + 128|0); //@line 736 "modules/audio/openal/Source.cpp"
 $44 = HEAP32[$43>>2]|0; //@line 736 "modules/audio/openal/Source.cpp"
 _alSourcei(($42|0),4097,($44|0)); //@line 736 "modules/audio/openal/Source.cpp"
 $45 = HEAP32[$0>>2]|0; //@line 737 "modules/audio/openal/Source.cpp"
 $46 = ((($this)) + 132|0); //@line 737 "modules/audio/openal/Source.cpp"
 $47 = HEAP32[$46>>2]|0; //@line 737 "modules/audio/openal/Source.cpp"
 _alSourcei(($45|0),4098,($47|0)); //@line 737 "modules/audio/openal/Source.cpp"
 $48 = HEAP32[$0>>2]|0; //@line 738 "modules/audio/openal/Source.cpp"
 $49 = ((($this)) + 136|0); //@line 738 "modules/audio/openal/Source.cpp"
 $50 = +HEAPF32[$49>>2]; //@line 738 "modules/audio/openal/Source.cpp"
 _alSourcef(($48|0),4130,(+$50)); //@line 738 "modules/audio/openal/Source.cpp"
 return; //@line 739 "modules/audio/openal/Source.cpp"
}
function __ZNK4love5audio6openal6Source10tellAtomicEPv($this,$unit) {
 $this = $this|0;
 $unit = $unit|0;
 var $$0 = 0.0, $$pre = 0.0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0.0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cond = 0, $offset = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $offset = sp;
 $0 = ((($this)) + 20|0); //@line 420 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 420 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 420 "modules/audio/openal/Source.cpp"
 if ($2) {
  $$0 = 0.0;
  STACKTOP = sp;return (+$$0); //@line 441 "modules/audio/openal/Source.cpp"
 }
 $3 = HEAP32[$unit>>2]|0; //@line 423 "modules/audio/openal/Source.cpp"
 $cond = ($3|0)==(1); //@line 423 "modules/audio/openal/Source.cpp"
 $4 = ((($this)) + 16|0); //@line 426 "modules/audio/openal/Source.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 426 "modules/audio/openal/Source.cpp"
 _alGetSourcef(($5|0),4133,($offset|0)); //@line 426 "modules/audio/openal/Source.cpp"
 do {
  if ($cond) {
   $6 = ((($this)) + 8|0); //@line 427 "modules/audio/openal/Source.cpp"
   $7 = HEAP32[$6>>2]|0; //@line 427 "modules/audio/openal/Source.cpp"
   $8 = ($7|0)==(1); //@line 427 "modules/audio/openal/Source.cpp"
   if ($8) {
    $9 = ((($this)) + 140|0); //@line 427 "modules/audio/openal/Source.cpp"
    $10 = +HEAPF32[$9>>2]; //@line 427 "modules/audio/openal/Source.cpp"
    $11 = +HEAPF32[$offset>>2]; //@line 427 "modules/audio/openal/Source.cpp"
    $12 = $10 + $11; //@line 427 "modules/audio/openal/Source.cpp"
    HEAPF32[$offset>>2] = $12; //@line 427 "modules/audio/openal/Source.cpp"
    $24 = $12;
    break;
   } else {
    $$pre = +HEAPF32[$offset>>2]; //@line 438 "modules/audio/openal/Source.cpp"
    $24 = $$pre;
    break;
   }
  } else {
   $13 = ((($this)) + 148|0); //@line 433 "modules/audio/openal/Source.cpp"
   $14 = HEAP32[$13>>2]|0; //@line 433 "modules/audio/openal/Source.cpp"
   $15 = (+($14|0)); //@line 433 "modules/audio/openal/Source.cpp"
   $16 = +HEAPF32[$offset>>2]; //@line 433 "modules/audio/openal/Source.cpp"
   $17 = $16 / $15; //@line 433 "modules/audio/openal/Source.cpp"
   HEAPF32[$offset>>2] = $17; //@line 433 "modules/audio/openal/Source.cpp"
   $18 = ((($this)) + 8|0); //@line 434 "modules/audio/openal/Source.cpp"
   $19 = HEAP32[$18>>2]|0; //@line 434 "modules/audio/openal/Source.cpp"
   $20 = ($19|0)==(1); //@line 434 "modules/audio/openal/Source.cpp"
   if ($20) {
    $21 = ((($this)) + 144|0); //@line 434 "modules/audio/openal/Source.cpp"
    $22 = +HEAPF32[$21>>2]; //@line 434 "modules/audio/openal/Source.cpp"
    $23 = $17 + $22; //@line 434 "modules/audio/openal/Source.cpp"
    HEAPF32[$offset>>2] = $23; //@line 434 "modules/audio/openal/Source.cpp"
    $24 = $23;
   } else {
    $24 = $17;
   }
  }
 } while(0);
 $$0 = $24;
 STACKTOP = sp;return (+$$0); //@line 441 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source17getDurationAtomicEPv($this,$vunit) {
 $this = $this|0;
 $vunit = $vunit|0;
 var $$2 = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$vunit>>2]|0; //@line 450 "modules/audio/openal/Source.cpp"
 $1 = ((($this)) + 8|0); //@line 452 "modules/audio/openal/Source.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 452 "modules/audio/openal/Source.cpp"
 $3 = ($2|0)==(1); //@line 452 "modules/audio/openal/Source.cpp"
 if ($3) {
  $4 = ((($this)) + 160|0); //@line 117 "./common/Object.h"
  $5 = HEAP32[$4>>2]|0; //@line 117 "./common/Object.h"
  $6 = HEAP32[$5>>2]|0; //@line 454 "modules/audio/openal/Source.cpp"
  $7 = ((($6)) + 60|0); //@line 454 "modules/audio/openal/Source.cpp"
  $8 = HEAP32[$7>>2]|0; //@line 454 "modules/audio/openal/Source.cpp"
  $9 = (+FUNCTION_TABLE_di[$8 & 4095]($5)); //@line 454 "modules/audio/openal/Source.cpp"
  $10 = ($0|0)==(0); //@line 456 "modules/audio/openal/Source.cpp"
  if ($10) {
   $$2 = $9;
   return (+$$2); //@line 471 "modules/audio/openal/Source.cpp"
  }
  $11 = HEAP32[$4>>2]|0; //@line 117 "./common/Object.h"
  $12 = HEAP32[$11>>2]|0; //@line 459 "modules/audio/openal/Source.cpp"
  $13 = ((($12)) + 56|0); //@line 459 "modules/audio/openal/Source.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 459 "modules/audio/openal/Source.cpp"
  $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 459 "modules/audio/openal/Source.cpp"
  $16 = (+($15|0)); //@line 459 "modules/audio/openal/Source.cpp"
  $17 = $9 * $16; //@line 459 "modules/audio/openal/Source.cpp"
  $$2 = $17;
  return (+$$2); //@line 471 "modules/audio/openal/Source.cpp"
 } else {
  $18 = ((($this)) + 56|0); //@line 117 "./common/Object.h"
  $19 = HEAP32[$18>>2]|0; //@line 117 "./common/Object.h"
  $20 = ((($19)) + 12|0); //@line 70 "modules/audio/openal/Source.h"
  $21 = HEAP32[$20>>2]|0; //@line 70 "modules/audio/openal/Source.h"
  $22 = ((($this)) + 152|0); //@line 464 "modules/audio/openal/Source.cpp"
  $23 = HEAP32[$22>>2]|0; //@line 464 "modules/audio/openal/Source.cpp"
  $24 = (($21|0) / ($23|0))&-1; //@line 464 "modules/audio/openal/Source.cpp"
  $25 = ((($this)) + 156|0); //@line 464 "modules/audio/openal/Source.cpp"
  $26 = HEAP32[$25>>2]|0; //@line 464 "modules/audio/openal/Source.cpp"
  $27 = (($26|0) / 8)&-1; //@line 464 "modules/audio/openal/Source.cpp"
  $28 = (($24|0) / ($27|0))&-1; //@line 464 "modules/audio/openal/Source.cpp"
  $29 = ($0|0)==(1); //@line 466 "modules/audio/openal/Source.cpp"
  $30 = (+($28|0)); //@line 467 "modules/audio/openal/Source.cpp"
  if ($29) {
   $$2 = $30;
   return (+$$2); //@line 471 "modules/audio/openal/Source.cpp"
  }
  $31 = ((($this)) + 148|0); //@line 469 "modules/audio/openal/Source.cpp"
  $32 = HEAP32[$31>>2]|0; //@line 469 "modules/audio/openal/Source.cpp"
  $33 = (+($32|0)); //@line 469 "modules/audio/openal/Source.cpp"
  $34 = $30 / $33; //@line 469 "modules/audio/openal/Source.cpp"
  $$2 = $34;
  return (+$$2); //@line 471 "modules/audio/openal/Source.cpp"
 }
 return +(0.0);
}
function __ZN4love5audio6openal6SourceC2EPNS1_4PoolEPNS_5sound9SoundDataE($this,$pool,$soundData) {
 $this = $this|0;
 $pool = $pool|0;
 $soundData = $soundData|0;
 var $$0 = 0, $$0$i$ph = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index3Z2D = 0, $eh$lpad$body$indexZ2D = 0;
 var $or$cond$i = 0, $or$cond3$i = 0, $or$cond5$i = 0, $or$cond7$i = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 __ZN4love5audio6SourceC2ENS1_4TypeE($this,0); //@line 80 "modules/audio/openal/Source.cpp"
 HEAP32[$this>>2] = (22556); //@line 102 "modules/audio/openal/Source.cpp"
 $0 = ((($this)) + 12|0); //@line 81 "modules/audio/openal/Source.cpp"
 HEAP32[$0>>2] = $pool; //@line 81 "modules/audio/openal/Source.cpp"
 $1 = ((($this)) + 20|0); //@line 82 "modules/audio/openal/Source.cpp"
 HEAP8[$1>>0] = 0; //@line 82 "modules/audio/openal/Source.cpp"
 $2 = ((($this)) + 56|0); //@line 93 "./common/Object.h"
 HEAP32[$2>>2] = 0; //@line 93 "./common/Object.h"
 $3 = ((($this)) + 60|0); //@line 84 "modules/audio/openal/Source.cpp"
 HEAPF32[$3>>2] = 1.0; //@line 84 "modules/audio/openal/Source.cpp"
 $4 = ((($this)) + 64|0); //@line 85 "modules/audio/openal/Source.cpp"
 HEAPF32[$4>>2] = 1.0; //@line 85 "modules/audio/openal/Source.cpp"
 $5 = ((($this)) + 104|0); //@line 86 "modules/audio/openal/Source.cpp"
 HEAP8[$5>>0] = 0; //@line 86 "modules/audio/openal/Source.cpp"
 $6 = ((($this)) + 105|0); //@line 87 "modules/audio/openal/Source.cpp"
 HEAP8[$6>>0] = 0; //@line 87 "modules/audio/openal/Source.cpp"
 $7 = ((($this)) + 106|0); //@line 88 "modules/audio/openal/Source.cpp"
 HEAP8[$7>>0] = 0; //@line 88 "modules/audio/openal/Source.cpp"
 $8 = ((($this)) + 108|0); //@line 89 "modules/audio/openal/Source.cpp"
 HEAPF32[$8>>2] = 0.0; //@line 89 "modules/audio/openal/Source.cpp"
 $9 = ((($this)) + 112|0); //@line 90 "modules/audio/openal/Source.cpp"
 HEAPF32[$9>>2] = 1.0; //@line 90 "modules/audio/openal/Source.cpp"
 $10 = ((($this)) + 116|0); //@line 91 "modules/audio/openal/Source.cpp"
 HEAPF32[$10>>2] = 1.0; //@line 91 "modules/audio/openal/Source.cpp"
 $11 = ((($this)) + 120|0); //@line 92 "modules/audio/openal/Source.cpp"
 HEAPF32[$11>>2] = 1.0; //@line 92 "modules/audio/openal/Source.cpp"
 $12 = ((($this)) + 124|0); //@line 93 "modules/audio/openal/Source.cpp"
 HEAPF32[$12>>2] = 3.4028234663852886E+38; //@line 93 "modules/audio/openal/Source.cpp"
 $13 = ((($this)) + 128|0); //@line 94 "modules/audio/openal/Source.cpp"
 $14 = $13; //@line 94 "modules/audio/openal/Source.cpp"
 $15 = $14; //@line 94 "modules/audio/openal/Source.cpp"
 HEAP32[$15>>2] = 0; //@line 94 "modules/audio/openal/Source.cpp"
 $16 = (($14) + 4)|0; //@line 94 "modules/audio/openal/Source.cpp"
 $17 = $16; //@line 94 "modules/audio/openal/Source.cpp"
 HEAP32[$17>>2] = 0; //@line 94 "modules/audio/openal/Source.cpp"
 HEAP32[$13>>2] = 360; //@line 181 "modules/audio/openal/Source.h"
 $18 = ((($this)) + 132|0); //@line 182 "modules/audio/openal/Source.h"
 HEAP32[$18>>2] = 360; //@line 182 "modules/audio/openal/Source.h"
 $19 = ((($this)) + 136|0); //@line 183 "modules/audio/openal/Source.h"
 HEAPF32[$19>>2] = 0.0; //@line 183 "modules/audio/openal/Source.h"
 $20 = ((($this)) + 140|0); //@line 95 "modules/audio/openal/Source.cpp"
 HEAPF32[$20>>2] = 0.0; //@line 95 "modules/audio/openal/Source.cpp"
 $21 = ((($this)) + 144|0); //@line 96 "modules/audio/openal/Source.cpp"
 HEAPF32[$21>>2] = 0.0; //@line 96 "modules/audio/openal/Source.cpp"
 $22 = HEAP32[$soundData>>2]|0; //@line 97 "modules/audio/openal/Source.cpp"
 $23 = ((($22)) + 32|0); //@line 97 "modules/audio/openal/Source.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 97 "modules/audio/openal/Source.cpp"
 __THREW__ = 0;
 $25 = (invoke_ii($24|0,($soundData|0))|0); //@line 97 "modules/audio/openal/Source.cpp"
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  label = 14;
 } else {
  $28 = ((($this)) + 148|0); //@line 97 "modules/audio/openal/Source.cpp"
  HEAP32[$28>>2] = $25; //@line 97 "modules/audio/openal/Source.cpp"
  $29 = HEAP32[$soundData>>2]|0; //@line 98 "modules/audio/openal/Source.cpp"
  $30 = ((($29)) + 24|0); //@line 98 "modules/audio/openal/Source.cpp"
  $31 = HEAP32[$30>>2]|0; //@line 98 "modules/audio/openal/Source.cpp"
  __THREW__ = 0;
  $32 = (invoke_ii($31|0,($soundData|0))|0); //@line 98 "modules/audio/openal/Source.cpp"
  $33 = __THREW__; __THREW__ = 0;
  $34 = $33&1;
  if ($34) {
   label = 14;
  } else {
   $35 = ((($this)) + 152|0); //@line 98 "modules/audio/openal/Source.cpp"
   HEAP32[$35>>2] = $32; //@line 98 "modules/audio/openal/Source.cpp"
   $36 = HEAP32[$soundData>>2]|0; //@line 99 "modules/audio/openal/Source.cpp"
   $37 = ((($36)) + 28|0); //@line 99 "modules/audio/openal/Source.cpp"
   $38 = HEAP32[$37>>2]|0; //@line 99 "modules/audio/openal/Source.cpp"
   __THREW__ = 0;
   $39 = (invoke_ii($38|0,($soundData|0))|0); //@line 99 "modules/audio/openal/Source.cpp"
   $40 = __THREW__; __THREW__ = 0;
   $41 = $40&1;
   if ($41) {
    label = 14;
   } else {
    $42 = ((($this)) + 156|0); //@line 99 "modules/audio/openal/Source.cpp"
    HEAP32[$42>>2] = $39; //@line 99 "modules/audio/openal/Source.cpp"
    $43 = ((($this)) + 160|0); //@line 93 "./common/Object.h"
    HEAP32[$43>>2] = 0; //@line 93 "./common/Object.h"
    $44 = ((($this)) + 164|0); //@line 101 "modules/audio/openal/Source.cpp"
    HEAP32[$44>>2] = 0; //@line 101 "modules/audio/openal/Source.cpp"
    $45 = HEAP32[$soundData>>2]|0; //@line 103 "modules/audio/openal/Source.cpp"
    $46 = ((($45)) + 24|0); //@line 103 "modules/audio/openal/Source.cpp"
    $47 = HEAP32[$46>>2]|0; //@line 103 "modules/audio/openal/Source.cpp"
    __THREW__ = 0;
    $48 = (invoke_ii($47|0,($soundData|0))|0); //@line 103 "modules/audio/openal/Source.cpp"
    $49 = __THREW__; __THREW__ = 0;
    $50 = $49&1;
    do {
     if ($50) {
      label = 15;
     } else {
      $51 = HEAP32[$soundData>>2]|0; //@line 103 "modules/audio/openal/Source.cpp"
      $52 = ((($51)) + 28|0); //@line 103 "modules/audio/openal/Source.cpp"
      $53 = HEAP32[$52>>2]|0; //@line 103 "modules/audio/openal/Source.cpp"
      __THREW__ = 0;
      $54 = (invoke_ii($53|0,($soundData|0))|0); //@line 103 "modules/audio/openal/Source.cpp"
      $55 = __THREW__; __THREW__ = 0;
      $56 = $55&1;
      if ($56) {
       label = 15;
      } else {
       $57 = ($48|0)==(1); //@line 750 "modules/audio/openal/Source.cpp"
       $58 = ($54|0)==(8); //@line 750 "modules/audio/openal/Source.cpp"
       $or$cond$i = $57 & $58; //@line 750 "modules/audio/openal/Source.cpp"
       if ($or$cond$i) {
        $$0$i$ph = 4352;
       } else {
        $59 = ($54|0)==(16); //@line 752 "modules/audio/openal/Source.cpp"
        $or$cond3$i = $57 & $59; //@line 752 "modules/audio/openal/Source.cpp"
        if ($or$cond3$i) {
         $$0$i$ph = 4353;
        } else {
         $60 = ($48|0)==(2); //@line 754 "modules/audio/openal/Source.cpp"
         $or$cond5$i = $60 & $58; //@line 754 "modules/audio/openal/Source.cpp"
         if ($or$cond5$i) {
          $$0$i$ph = 4354;
         } else {
          $or$cond7$i = $60 & $59; //@line 756 "modules/audio/openal/Source.cpp"
          if ($or$cond7$i) {
           $$0$i$ph = 4355;
          } else {
           $61 = (___cxa_allocate_exception(16)|0); //@line 105 "modules/audio/openal/Source.cpp"
           $62 = HEAP32[$soundData>>2]|0; //@line 105 "modules/audio/openal/Source.cpp"
           $63 = ((($62)) + 24|0); //@line 105 "modules/audio/openal/Source.cpp"
           $64 = HEAP32[$63>>2]|0; //@line 105 "modules/audio/openal/Source.cpp"
           __THREW__ = 0;
           $65 = (invoke_ii($64|0,($soundData|0))|0); //@line 105 "modules/audio/openal/Source.cpp"
           $66 = __THREW__; __THREW__ = 0;
           $67 = $66&1;
           if (!($67)) {
            $68 = HEAP32[$soundData>>2]|0; //@line 105 "modules/audio/openal/Source.cpp"
            $69 = ((($68)) + 28|0); //@line 105 "modules/audio/openal/Source.cpp"
            $70 = HEAP32[$69>>2]|0; //@line 105 "modules/audio/openal/Source.cpp"
            __THREW__ = 0;
            $71 = (invoke_ii($70|0,($soundData|0))|0); //@line 105 "modules/audio/openal/Source.cpp"
            $72 = __THREW__; __THREW__ = 0;
            $73 = $72&1;
            if (!($73)) {
             __THREW__ = 0;
             HEAP32[$vararg_buffer>>2] = $65; //@line 49 "modules/audio/openal/Source.cpp"
             $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 49 "modules/audio/openal/Source.cpp"
             HEAP32[$vararg_ptr1>>2] = $71; //@line 49 "modules/audio/openal/Source.cpp"
             invoke_viii(3349,($61|0),(365272|0),($vararg_buffer|0)); //@line 49 "modules/audio/openal/Source.cpp"
             $74 = __THREW__; __THREW__ = 0;
             $75 = $74&1;
             if (!($75)) {
              HEAP32[$61>>2] = (22776); //@line 50 "modules/audio/openal/Source.cpp"
              __THREW__ = 0;
              invoke_viii(3350,($61|0),(1504|0),(344|0)); //@line 105 "modules/audio/openal/Source.cpp"
              $76 = __THREW__; __THREW__ = 0;
              label = 15;
              break;
             }
            }
           }
           $81 = ___cxa_find_matching_catch_2()|0;
           $82 = tempRet0;
           ___cxa_free_exception(($61|0)); //@line 105 "modules/audio/openal/Source.cpp"
           $$0 = $81;$$01 = $82;
           break;
          }
         }
        }
       }
       __THREW__ = 0;
       $83 = (invoke_ii(3348,16)|0); //@line 107 "modules/audio/openal/Source.cpp"
       $84 = __THREW__; __THREW__ = 0;
       $85 = $84&1;
       if ($85) {
        label = 15;
       } else {
        $86 = HEAP32[$soundData>>2]|0; //@line 107 "modules/audio/openal/Source.cpp"
        $87 = ((($86)) + 16|0); //@line 107 "modules/audio/openal/Source.cpp"
        $88 = HEAP32[$87>>2]|0; //@line 107 "modules/audio/openal/Source.cpp"
        __THREW__ = 0;
        $89 = (invoke_ii($88|0,($soundData|0))|0); //@line 107 "modules/audio/openal/Source.cpp"
        $90 = __THREW__; __THREW__ = 0;
        $91 = $90&1;
        if ($91) {
         label = 30;
        } else {
         $92 = HEAP32[$soundData>>2]|0; //@line 107 "modules/audio/openal/Source.cpp"
         $93 = ((($92)) + 20|0); //@line 107 "modules/audio/openal/Source.cpp"
         $94 = HEAP32[$93>>2]|0; //@line 107 "modules/audio/openal/Source.cpp"
         __THREW__ = 0;
         $95 = (invoke_ii($94|0,($soundData|0))|0); //@line 107 "modules/audio/openal/Source.cpp"
         $96 = __THREW__; __THREW__ = 0;
         $97 = $96&1;
         if ($97) {
          label = 30;
         } else {
          $98 = HEAP32[$soundData>>2]|0; //@line 107 "modules/audio/openal/Source.cpp"
          $99 = ((($98)) + 32|0); //@line 107 "modules/audio/openal/Source.cpp"
          $100 = HEAP32[$99>>2]|0; //@line 107 "modules/audio/openal/Source.cpp"
          __THREW__ = 0;
          $101 = (invoke_ii($100|0,($soundData|0))|0); //@line 107 "modules/audio/openal/Source.cpp"
          $102 = __THREW__; __THREW__ = 0;
          $103 = $102&1;
          if ($103) {
           label = 30;
          } else {
           __THREW__ = 0;
           invoke_vi(3431,($83|0)); //@line 67 "modules/audio/openal/Source.cpp"
           $104 = __THREW__; __THREW__ = 0;
           $105 = $104&1;
           if ($105) {
            label = 30;
           } else {
            HEAP32[$83>>2] = (22532); //@line 69 "modules/audio/openal/Source.cpp"
            $106 = ((($83)) + 12|0); //@line 68 "modules/audio/openal/Source.cpp"
            HEAP32[$106>>2] = $95; //@line 68 "modules/audio/openal/Source.cpp"
            $107 = ((($83)) + 8|0); //@line 70 "modules/audio/openal/Source.cpp"
            __THREW__ = 0;
            invoke_vii(3932,1,($107|0)); //@line 70 "modules/audio/openal/Source.cpp"
            $108 = __THREW__; __THREW__ = 0;
            $109 = $108&1;
            if (!($109)) {
             $110 = HEAP32[$107>>2]|0; //@line 71 "modules/audio/openal/Source.cpp"
             __THREW__ = 0;
             invoke_viiiii(3933,($110|0),($$0$i$ph|0),($89|0),($95|0),($101|0)); //@line 71 "modules/audio/openal/Source.cpp"
             $111 = __THREW__; __THREW__ = 0;
             $112 = $111&1;
             if (!($112)) {
              $115 = HEAP32[$83>>2]|0; //@line 132 "./common/Object.h"
              $116 = ((($115)) + 8|0); //@line 132 "./common/Object.h"
              $117 = HEAP32[$116>>2]|0; //@line 132 "./common/Object.h"
              __THREW__ = 0;
              invoke_vi($117|0,($83|0)); //@line 132 "./common/Object.h"
              $118 = __THREW__; __THREW__ = 0;
              $119 = $118&1;
              if ($119) {
               label = 15;
               break;
              }
              $120 = HEAP32[$2>>2]|0; //@line 133 "./common/Object.h"
              $121 = ($120|0)==(0|0); //@line 133 "./common/Object.h"
              if (!($121)) {
               $122 = HEAP32[$120>>2]|0; //@line 133 "./common/Object.h"
               $123 = ((($122)) + 12|0); //@line 133 "./common/Object.h"
               $124 = HEAP32[$123>>2]|0; //@line 133 "./common/Object.h"
               __THREW__ = 0;
               invoke_vi($124|0,($120|0)); //@line 133 "./common/Object.h"
               $125 = __THREW__; __THREW__ = 0;
               $126 = $125&1;
               if ($126) {
                label = 15;
                break;
               }
              }
              HEAP32[$2>>2] = $83; //@line 134 "./common/Object.h"
              $127 = HEAP32[$83>>2]|0; //@line 110 "modules/audio/openal/Source.cpp"
              $128 = ((($127)) + 12|0); //@line 110 "modules/audio/openal/Source.cpp"
              $129 = HEAP32[$128>>2]|0; //@line 110 "modules/audio/openal/Source.cpp"
              __THREW__ = 0;
              invoke_vi($129|0,($83|0)); //@line 110 "modules/audio/openal/Source.cpp"
              $130 = __THREW__; __THREW__ = 0;
              $131 = $130&1;
              if ($131) {
               label = 15;
               break;
              }
              $132 = ((($this)) + 68|0); //@line 114 "modules/audio/openal/Source.cpp"
              dest=$132; stop=dest+36|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 744 "modules/audio/openal/Source.cpp"
              STACKTOP = sp;return; //@line 117 "modules/audio/openal/Source.cpp"
             }
            }
            $113 = ___cxa_find_matching_catch_2()|0;
            $114 = tempRet0;
            $eh$lpad$body$index3Z2D = $114;$eh$lpad$body$indexZ2D = $113;
           }
          }
         }
        }
        if ((label|0) == 30) {
         $133 = ___cxa_find_matching_catch_2()|0;
         $134 = tempRet0;
         $eh$lpad$body$index3Z2D = $134;$eh$lpad$body$indexZ2D = $133;
        }
        __ZdlPv($83); //@line 107 "modules/audio/openal/Source.cpp"
        $$0 = $eh$lpad$body$indexZ2D;$$01 = $eh$lpad$body$index3Z2D;
       }
      }
     }
    } while(0);
    if ((label|0) == 15) {
     $79 = ___cxa_find_matching_catch_2()|0;
     $80 = tempRet0;
     $$0 = $79;$$01 = $80;
    }
    $135 = HEAP32[$43>>2]|0; //@line 106 "./common/Object.h"
    $136 = ($135|0)==(0|0); //@line 106 "./common/Object.h"
    if ($136) {
     $$1 = $$0;$$12 = $$01;
    } else {
     $137 = HEAP32[$135>>2]|0; //@line 106 "./common/Object.h"
     $138 = ((($137)) + 12|0); //@line 106 "./common/Object.h"
     $139 = HEAP32[$138>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($139|0,($135|0)); //@line 106 "./common/Object.h"
     $140 = __THREW__; __THREW__ = 0;
     $141 = $140&1;
     if ($141) {
      $142 = ___cxa_find_matching_catch_3(0|0)|0;
      $143 = tempRet0;
      ___clang_call_terminate($142); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     } else {
      $$1 = $$0;$$12 = $$01;
     }
    }
   }
  }
 }
 if ((label|0) == 14) {
  $77 = ___cxa_find_matching_catch_2()|0;
  $78 = tempRet0;
  $$1 = $77;$$12 = $78;
 }
 $144 = HEAP32[$2>>2]|0; //@line 106 "./common/Object.h"
 $145 = ($144|0)==(0|0); //@line 106 "./common/Object.h"
 if ($145) {
  __ZN4love5audio6SourceD2Ev($this); //@line 117 "modules/audio/openal/Source.cpp"
  ___resumeException($$1|0);
  // unreachable;
 }
 $146 = HEAP32[$144>>2]|0; //@line 106 "./common/Object.h"
 $147 = ((($146)) + 12|0); //@line 106 "./common/Object.h"
 $148 = HEAP32[$147>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($148|0,($144|0)); //@line 106 "./common/Object.h"
 $149 = __THREW__; __THREW__ = 0;
 $150 = $149&1;
 if ($150) {
  $151 = ___cxa_find_matching_catch_3(0|0)|0;
  $152 = tempRet0;
  ___clang_call_terminate($151); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  __ZN4love5audio6SourceD2Ev($this); //@line 117 "modules/audio/openal/Source.cpp"
  ___resumeException($$1|0);
  // unreachable;
 }
}
function __ZN4love5audio6openal22InvalidFormatExceptionD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love9ExceptionD2Ev($this); //@line 44 "modules/audio/openal/Source.cpp"
 __ZdlPv($this); //@line 44 "modules/audio/openal/Source.cpp"
 return; //@line 44 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6SourceC2EPNS1_4PoolEPNS_5sound7DecoderE($this,$pool,$decoder) {
 $this = $this|0;
 $pool = $pool|0;
 $decoder = $decoder|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, $or$cond$i = 0, $or$cond3$i = 0, $or$cond7$i5 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0;
 var dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 __ZN4love5audio6SourceC2ENS1_4TypeE($this,1); //@line 120 "modules/audio/openal/Source.cpp"
 HEAP32[$this>>2] = (22556); //@line 142 "modules/audio/openal/Source.cpp"
 $0 = ((($this)) + 12|0); //@line 121 "modules/audio/openal/Source.cpp"
 HEAP32[$0>>2] = $pool; //@line 121 "modules/audio/openal/Source.cpp"
 $1 = ((($this)) + 20|0); //@line 122 "modules/audio/openal/Source.cpp"
 HEAP8[$1>>0] = 0; //@line 122 "modules/audio/openal/Source.cpp"
 $2 = ((($this)) + 56|0); //@line 93 "./common/Object.h"
 HEAP32[$2>>2] = 0; //@line 93 "./common/Object.h"
 $3 = ((($this)) + 60|0); //@line 124 "modules/audio/openal/Source.cpp"
 HEAPF32[$3>>2] = 1.0; //@line 124 "modules/audio/openal/Source.cpp"
 $4 = ((($this)) + 64|0); //@line 125 "modules/audio/openal/Source.cpp"
 HEAPF32[$4>>2] = 1.0; //@line 125 "modules/audio/openal/Source.cpp"
 $5 = ((($this)) + 104|0); //@line 126 "modules/audio/openal/Source.cpp"
 HEAP8[$5>>0] = 0; //@line 126 "modules/audio/openal/Source.cpp"
 $6 = ((($this)) + 105|0); //@line 127 "modules/audio/openal/Source.cpp"
 HEAP8[$6>>0] = 0; //@line 127 "modules/audio/openal/Source.cpp"
 $7 = ((($this)) + 106|0); //@line 128 "modules/audio/openal/Source.cpp"
 HEAP8[$7>>0] = 0; //@line 128 "modules/audio/openal/Source.cpp"
 $8 = ((($this)) + 108|0); //@line 129 "modules/audio/openal/Source.cpp"
 HEAPF32[$8>>2] = 0.0; //@line 129 "modules/audio/openal/Source.cpp"
 $9 = ((($this)) + 112|0); //@line 130 "modules/audio/openal/Source.cpp"
 HEAPF32[$9>>2] = 1.0; //@line 130 "modules/audio/openal/Source.cpp"
 $10 = ((($this)) + 116|0); //@line 131 "modules/audio/openal/Source.cpp"
 HEAPF32[$10>>2] = 1.0; //@line 131 "modules/audio/openal/Source.cpp"
 $11 = ((($this)) + 120|0); //@line 132 "modules/audio/openal/Source.cpp"
 HEAPF32[$11>>2] = 1.0; //@line 132 "modules/audio/openal/Source.cpp"
 $12 = ((($this)) + 124|0); //@line 133 "modules/audio/openal/Source.cpp"
 HEAPF32[$12>>2] = 3.4028234663852886E+38; //@line 133 "modules/audio/openal/Source.cpp"
 $13 = ((($this)) + 128|0); //@line 134 "modules/audio/openal/Source.cpp"
 $14 = $13; //@line 134 "modules/audio/openal/Source.cpp"
 $15 = $14; //@line 134 "modules/audio/openal/Source.cpp"
 HEAP32[$15>>2] = 0; //@line 134 "modules/audio/openal/Source.cpp"
 $16 = (($14) + 4)|0; //@line 134 "modules/audio/openal/Source.cpp"
 $17 = $16; //@line 134 "modules/audio/openal/Source.cpp"
 HEAP32[$17>>2] = 0; //@line 134 "modules/audio/openal/Source.cpp"
 HEAP32[$13>>2] = 360; //@line 181 "modules/audio/openal/Source.h"
 $18 = ((($this)) + 132|0); //@line 182 "modules/audio/openal/Source.h"
 HEAP32[$18>>2] = 360; //@line 182 "modules/audio/openal/Source.h"
 $19 = ((($this)) + 136|0); //@line 183 "modules/audio/openal/Source.h"
 HEAPF32[$19>>2] = 0.0; //@line 183 "modules/audio/openal/Source.h"
 $20 = ((($this)) + 140|0); //@line 135 "modules/audio/openal/Source.cpp"
 HEAPF32[$20>>2] = 0.0; //@line 135 "modules/audio/openal/Source.cpp"
 $21 = ((($this)) + 144|0); //@line 136 "modules/audio/openal/Source.cpp"
 HEAPF32[$21>>2] = 0.0; //@line 136 "modules/audio/openal/Source.cpp"
 $22 = HEAP32[$decoder>>2]|0; //@line 137 "modules/audio/openal/Source.cpp"
 $23 = ((($22)) + 56|0); //@line 137 "modules/audio/openal/Source.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 137 "modules/audio/openal/Source.cpp"
 __THREW__ = 0;
 $25 = (invoke_ii($24|0,($decoder|0))|0); //@line 137 "modules/audio/openal/Source.cpp"
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  label = 14;
 } else {
  $28 = ((($this)) + 148|0); //@line 137 "modules/audio/openal/Source.cpp"
  HEAP32[$28>>2] = $25; //@line 137 "modules/audio/openal/Source.cpp"
  $29 = HEAP32[$decoder>>2]|0; //@line 138 "modules/audio/openal/Source.cpp"
  $30 = ((($29)) + 48|0); //@line 138 "modules/audio/openal/Source.cpp"
  $31 = HEAP32[$30>>2]|0; //@line 138 "modules/audio/openal/Source.cpp"
  __THREW__ = 0;
  $32 = (invoke_ii($31|0,($decoder|0))|0); //@line 138 "modules/audio/openal/Source.cpp"
  $33 = __THREW__; __THREW__ = 0;
  $34 = $33&1;
  if ($34) {
   label = 14;
  } else {
   $35 = ((($this)) + 152|0); //@line 138 "modules/audio/openal/Source.cpp"
   HEAP32[$35>>2] = $32; //@line 138 "modules/audio/openal/Source.cpp"
   $36 = HEAP32[$decoder>>2]|0; //@line 139 "modules/audio/openal/Source.cpp"
   $37 = ((($36)) + 52|0); //@line 139 "modules/audio/openal/Source.cpp"
   $38 = HEAP32[$37>>2]|0; //@line 139 "modules/audio/openal/Source.cpp"
   __THREW__ = 0;
   $39 = (invoke_ii($38|0,($decoder|0))|0); //@line 139 "modules/audio/openal/Source.cpp"
   $40 = __THREW__; __THREW__ = 0;
   $41 = $40&1;
   if ($41) {
    label = 14;
   } else {
    $42 = ((($this)) + 156|0); //@line 139 "modules/audio/openal/Source.cpp"
    HEAP32[$42>>2] = $39; //@line 139 "modules/audio/openal/Source.cpp"
    $43 = ((($this)) + 160|0); //@line 93 "./common/Object.h"
    HEAP32[$43>>2] = $decoder; //@line 93 "./common/Object.h"
    $44 = HEAP32[$decoder>>2]|0; //@line 95 "./common/Object.h"
    $45 = ((($44)) + 8|0); //@line 95 "./common/Object.h"
    $46 = HEAP32[$45>>2]|0; //@line 95 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($46|0,($decoder|0)); //@line 95 "./common/Object.h"
    $47 = __THREW__; __THREW__ = 0;
    $48 = $47&1;
    if ($48) {
     label = 14;
    } else {
     $49 = ((($this)) + 164|0); //@line 141 "modules/audio/openal/Source.cpp"
     HEAP32[$49>>2] = 0; //@line 141 "modules/audio/openal/Source.cpp"
     $50 = HEAP32[$decoder>>2]|0; //@line 143 "modules/audio/openal/Source.cpp"
     $51 = ((($50)) + 48|0); //@line 143 "modules/audio/openal/Source.cpp"
     $52 = HEAP32[$51>>2]|0; //@line 143 "modules/audio/openal/Source.cpp"
     __THREW__ = 0;
     $53 = (invoke_ii($52|0,($decoder|0))|0); //@line 143 "modules/audio/openal/Source.cpp"
     $54 = __THREW__; __THREW__ = 0;
     $55 = $54&1;
     do {
      if ($55) {
       label = 15;
      } else {
       $56 = HEAP32[$decoder>>2]|0; //@line 143 "modules/audio/openal/Source.cpp"
       $57 = ((($56)) + 52|0); //@line 143 "modules/audio/openal/Source.cpp"
       $58 = HEAP32[$57>>2]|0; //@line 143 "modules/audio/openal/Source.cpp"
       __THREW__ = 0;
       $59 = (invoke_ii($58|0,($decoder|0))|0); //@line 143 "modules/audio/openal/Source.cpp"
       $60 = __THREW__; __THREW__ = 0;
       $61 = $60&1;
       if ($61) {
        label = 15;
       } else {
        $62 = ($53|0)==(1); //@line 750 "modules/audio/openal/Source.cpp"
        $63 = ($59|0)==(8); //@line 750 "modules/audio/openal/Source.cpp"
        $or$cond$i = $62 & $63; //@line 750 "modules/audio/openal/Source.cpp"
        if (!($or$cond$i)) {
         $64 = ($59|0)==(16); //@line 752 "modules/audio/openal/Source.cpp"
         $or$cond3$i = $62 & $64; //@line 752 "modules/audio/openal/Source.cpp"
         if (!($or$cond3$i)) {
          $65 = ($53|0)==(2); //@line 754 "modules/audio/openal/Source.cpp"
          $or$cond7$i5 = $63 | $64; //@line 754 "modules/audio/openal/Source.cpp"
          $or$cond = $65 & $or$cond7$i5; //@line 754 "modules/audio/openal/Source.cpp"
          if (!($or$cond)) {
           $66 = (___cxa_allocate_exception(16)|0); //@line 144 "modules/audio/openal/Source.cpp"
           $67 = HEAP32[$decoder>>2]|0; //@line 144 "modules/audio/openal/Source.cpp"
           $68 = ((($67)) + 48|0); //@line 144 "modules/audio/openal/Source.cpp"
           $69 = HEAP32[$68>>2]|0; //@line 144 "modules/audio/openal/Source.cpp"
           __THREW__ = 0;
           $70 = (invoke_ii($69|0,($decoder|0))|0); //@line 144 "modules/audio/openal/Source.cpp"
           $71 = __THREW__; __THREW__ = 0;
           $72 = $71&1;
           if (!($72)) {
            $73 = HEAP32[$decoder>>2]|0; //@line 144 "modules/audio/openal/Source.cpp"
            $74 = ((($73)) + 52|0); //@line 144 "modules/audio/openal/Source.cpp"
            $75 = HEAP32[$74>>2]|0; //@line 144 "modules/audio/openal/Source.cpp"
            __THREW__ = 0;
            $76 = (invoke_ii($75|0,($decoder|0))|0); //@line 144 "modules/audio/openal/Source.cpp"
            $77 = __THREW__; __THREW__ = 0;
            $78 = $77&1;
            if (!($78)) {
             __THREW__ = 0;
             HEAP32[$vararg_buffer>>2] = $70; //@line 49 "modules/audio/openal/Source.cpp"
             $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 49 "modules/audio/openal/Source.cpp"
             HEAP32[$vararg_ptr1>>2] = $76; //@line 49 "modules/audio/openal/Source.cpp"
             invoke_viii(3349,($66|0),(365272|0),($vararg_buffer|0)); //@line 49 "modules/audio/openal/Source.cpp"
             $79 = __THREW__; __THREW__ = 0;
             $80 = $79&1;
             if (!($80)) {
              HEAP32[$66>>2] = (22776); //@line 50 "modules/audio/openal/Source.cpp"
              __THREW__ = 0;
              invoke_viii(3350,($66|0),(1504|0),(344|0)); //@line 144 "modules/audio/openal/Source.cpp"
              $81 = __THREW__; __THREW__ = 0;
              label = 15;
              break;
             }
            }
           }
           $86 = ___cxa_find_matching_catch_2()|0;
           $87 = tempRet0;
           ___cxa_free_exception(($66|0)); //@line 144 "modules/audio/openal/Source.cpp"
           $$0 = $87;$$01 = $86;
           break;
          }
         }
        }
        $88 = ((($this)) + 24|0); //@line 146 "modules/audio/openal/Source.cpp"
        __THREW__ = 0;
        invoke_vii(3932,8,($88|0)); //@line 146 "modules/audio/openal/Source.cpp"
        $89 = __THREW__; __THREW__ = 0;
        $90 = $89&1;
        if ($90) {
         label = 15;
        } else {
         $91 = ((($this)) + 68|0); //@line 150 "modules/audio/openal/Source.cpp"
         dest=$91; stop=dest+36|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 744 "modules/audio/openal/Source.cpp"
         STACKTOP = sp;return; //@line 153 "modules/audio/openal/Source.cpp"
        }
       }
      }
     } while(0);
     if ((label|0) == 15) {
      $84 = ___cxa_find_matching_catch_2()|0;
      $85 = tempRet0;
      $$0 = $85;$$01 = $84;
     }
     $92 = HEAP32[$43>>2]|0; //@line 106 "./common/Object.h"
     $93 = ($92|0)==(0|0); //@line 106 "./common/Object.h"
     if ($93) {
      $$1 = $$0;$$12 = $$01;
     } else {
      $94 = HEAP32[$92>>2]|0; //@line 106 "./common/Object.h"
      $95 = ((($94)) + 12|0); //@line 106 "./common/Object.h"
      $96 = HEAP32[$95>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($96|0,($92|0)); //@line 106 "./common/Object.h"
      $97 = __THREW__; __THREW__ = 0;
      $98 = $97&1;
      if ($98) {
       $99 = ___cxa_find_matching_catch_3(0|0)|0;
       $100 = tempRet0;
       ___clang_call_terminate($99); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      } else {
       $$1 = $$0;$$12 = $$01;
      }
     }
    }
   }
  }
 }
 if ((label|0) == 14) {
  $82 = ___cxa_find_matching_catch_2()|0;
  $83 = tempRet0;
  $$1 = $83;$$12 = $82;
 }
 $101 = HEAP32[$2>>2]|0; //@line 106 "./common/Object.h"
 $102 = ($101|0)==(0|0); //@line 106 "./common/Object.h"
 if ($102) {
  __ZN4love5audio6SourceD2Ev($this); //@line 153 "modules/audio/openal/Source.cpp"
  ___resumeException($$12|0);
  // unreachable;
 }
 $103 = HEAP32[$101>>2]|0; //@line 106 "./common/Object.h"
 $104 = ((($103)) + 12|0); //@line 106 "./common/Object.h"
 $105 = HEAP32[$104>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($105|0,($101|0)); //@line 106 "./common/Object.h"
 $106 = __THREW__; __THREW__ = 0;
 $107 = $106&1;
 if ($107) {
  $108 = ___cxa_find_matching_catch_3(0|0)|0;
  $109 = tempRet0;
  ___clang_call_terminate($108); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  __ZN4love5audio6SourceD2Ev($this); //@line 153 "modules/audio/openal/Source.cpp"
  ___resumeException($$12|0);
  // unreachable;
 }
}
function __ZN4love5audio6openal6Source11pauseAtomicEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 676 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 676 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 676 "modules/audio/openal/Source.cpp"
 if ($2) {
  return; //@line 681 "modules/audio/openal/Source.cpp"
 }
 $3 = ((($this)) + 16|0); //@line 678 "modules/audio/openal/Source.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 678 "modules/audio/openal/Source.cpp"
 _alSourcePause(($4|0)); //@line 678 "modules/audio/openal/Source.cpp"
 $5 = ((($this)) + 106|0); //@line 679 "modules/audio/openal/Source.cpp"
 HEAP8[$5>>0] = 1; //@line 679 "modules/audio/openal/Source.cpp"
 return; //@line 681 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source12resumeAtomicEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 685 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 685 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 685 "modules/audio/openal/Source.cpp"
 if ($2) {
  return; //@line 690 "modules/audio/openal/Source.cpp"
 }
 $3 = ((($this)) + 106|0); //@line 685 "modules/audio/openal/Source.cpp"
 $4 = HEAP8[$3>>0]|0; //@line 685 "modules/audio/openal/Source.cpp"
 $5 = ($4<<24>>24)==(0); //@line 685 "modules/audio/openal/Source.cpp"
 if ($5) {
  return; //@line 690 "modules/audio/openal/Source.cpp"
 }
 $6 = ((($this)) + 16|0); //@line 687 "modules/audio/openal/Source.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 687 "modules/audio/openal/Source.cpp"
 _alSourcePlay(($7|0)); //@line 687 "modules/audio/openal/Source.cpp"
 HEAP8[$3>>0] = 0; //@line 688 "modules/audio/openal/Source.cpp"
 return; //@line 690 "modules/audio/openal/Source.cpp"
}
function __ZN4love5audio6openal6Source12rewindAtomicEv($this) {
 $this = $this|0;
 var $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 694 "modules/audio/openal/Source.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 700 "modules/audio/openal/Source.cpp"
 $2 = ($1<<24>>24)==(0); //@line 694 "modules/audio/openal/Source.cpp"
 $$phi$trans$insert = ((($this)) + 8|0); //@line 714 "modules/audio/openal/Source.cpp"
 $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 714 "modules/audio/openal/Source.cpp"
 if ($2) {
  $23 = ($$pre|0)==(1); //@line 714 "modules/audio/openal/Source.cpp"
  if (!($23)) {
   return; //@line 720 "modules/audio/openal/Source.cpp"
  }
  $24 = ((($this)) + 160|0); //@line 117 "./common/Object.h"
  $25 = HEAP32[$24>>2]|0; //@line 117 "./common/Object.h"
  $26 = HEAP32[$25>>2]|0; //@line 716 "modules/audio/openal/Source.cpp"
  $27 = ((($26)) + 36|0); //@line 716 "modules/audio/openal/Source.cpp"
  $28 = HEAP32[$27>>2]|0; //@line 716 "modules/audio/openal/Source.cpp"
  (FUNCTION_TABLE_ii[$28 & 8191]($25)|0); //@line 716 "modules/audio/openal/Source.cpp"
  $29 = ((($this)) + 140|0); //@line 717 "modules/audio/openal/Source.cpp"
  HEAPF32[$29>>2] = 0.0; //@line 717 "modules/audio/openal/Source.cpp"
  $30 = ((($this)) + 144|0); //@line 718 "modules/audio/openal/Source.cpp"
  HEAPF32[$30>>2] = 0.0; //@line 718 "modules/audio/openal/Source.cpp"
  return; //@line 720 "modules/audio/openal/Source.cpp"
 }
 switch ($$pre|0) {
 case 0:  {
  $3 = ((($this)) + 16|0); //@line 696 "modules/audio/openal/Source.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 696 "modules/audio/openal/Source.cpp"
  _alSourceRewind(($4|0)); //@line 696 "modules/audio/openal/Source.cpp"
  $5 = ((($this)) + 106|0); //@line 697 "modules/audio/openal/Source.cpp"
  $6 = HEAP8[$5>>0]|0; //@line 697 "modules/audio/openal/Source.cpp"
  $7 = ($6<<24>>24)==(0); //@line 697 "modules/audio/openal/Source.cpp"
  if (!($7)) {
   return; //@line 720 "modules/audio/openal/Source.cpp"
  }
  $8 = HEAP32[$3>>2]|0; //@line 698 "modules/audio/openal/Source.cpp"
  _alSourcePlay(($8|0)); //@line 698 "modules/audio/openal/Source.cpp"
  return; //@line 720 "modules/audio/openal/Source.cpp"
  break;
 }
 case 1:  {
  $9 = ((($this)) + 106|0); //@line 702 "modules/audio/openal/Source.cpp"
  $10 = HEAP8[$9>>0]|0; //@line 702 "modules/audio/openal/Source.cpp"
  $11 = ($10<<24>>24)==(0); //@line 702 "modules/audio/openal/Source.cpp"
  $12 = ((($this)) + 160|0); //@line 117 "./common/Object.h"
  $13 = HEAP32[$12>>2]|0; //@line 117 "./common/Object.h"
  $14 = HEAP32[$13>>2]|0; //@line 703 "modules/audio/openal/Source.cpp"
  $15 = ((($14)) + 36|0); //@line 703 "modules/audio/openal/Source.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 703 "modules/audio/openal/Source.cpp"
  (FUNCTION_TABLE_ii[$16 & 8191]($13)|0); //@line 703 "modules/audio/openal/Source.cpp"
  __ZN4love5audio6openal6Source10stopAtomicEv($this); //@line 707 "modules/audio/openal/Source.cpp"
  (__ZN4love5audio6openal6Source10playAtomicEv($this)|0); //@line 708 "modules/audio/openal/Source.cpp"
  if (!($11)) {
   $17 = HEAP8[$0>>0]|0; //@line 676 "modules/audio/openal/Source.cpp"
   $18 = ($17<<24>>24)==(0); //@line 676 "modules/audio/openal/Source.cpp"
   if (!($18)) {
    $19 = ((($this)) + 16|0); //@line 678 "modules/audio/openal/Source.cpp"
    $20 = HEAP32[$19>>2]|0; //@line 678 "modules/audio/openal/Source.cpp"
    _alSourcePause(($20|0)); //@line 678 "modules/audio/openal/Source.cpp"
    HEAP8[$9>>0] = 1; //@line 679 "modules/audio/openal/Source.cpp"
   }
  }
  $21 = ((($this)) + 140|0); //@line 711 "modules/audio/openal/Source.cpp"
  HEAPF32[$21>>2] = 0.0; //@line 711 "modules/audio/openal/Source.cpp"
  $22 = ((($this)) + 144|0); //@line 712 "modules/audio/openal/Source.cpp"
  HEAPF32[$22>>2] = 0.0; //@line 712 "modules/audio/openal/Source.cpp"
  return; //@line 720 "modules/audio/openal/Source.cpp"
  break;
 }
 default: {
  return; //@line 720 "modules/audio/openal/Source.cpp"
 }
 }
}
function __GLOBAL__sub_I_Audio_cpp() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[(632400)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632412)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632424)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632436)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632448)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632460)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632472)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632484)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632496)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632508)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632520)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632532)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632544)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632556)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(632560)>>2]=0|0;HEAP32[(632560)+4>>2]=0|0;HEAP32[(632560)+8>>2]=0|0;HEAP32[(632560)+12>>2]=0|0;HEAP32[(632560)+16>>2]=0|0;HEAP32[(632560)+20>>2]=0|0;HEAP32[(632560)+24>>2]=0|0; //@line 44 "./common/StringMap.h"
 $0 = HEAP32[5697]|0; //@line 49 "./common/StringMap.h"
 $1 = HEAP32[(22792)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_5audio5Audio13DistanceModelELj7EE3addEPKcS3_(632392,$0,$1)|0); //@line 49 "./common/StringMap.h"
 $2 = HEAP32[(22796)>>2]|0; //@line 49 "./common/StringMap.h"
 $3 = HEAP32[(22800)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_5audio5Audio13DistanceModelELj7EE3addEPKcS3_(632392,$2,$3)|0); //@line 49 "./common/StringMap.h"
 $4 = HEAP32[(22804)>>2]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(22808)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_5audio5Audio13DistanceModelELj7EE3addEPKcS3_(632392,$4,$5)|0); //@line 49 "./common/StringMap.h"
 $6 = HEAP32[(22812)>>2]|0; //@line 49 "./common/StringMap.h"
 $7 = HEAP32[(22816)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_5audio5Audio13DistanceModelELj7EE3addEPKcS3_(632392,$6,$7)|0); //@line 49 "./common/StringMap.h"
 $8 = HEAP32[(22820)>>2]|0; //@line 49 "./common/StringMap.h"
 $9 = HEAP32[(22824)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_5audio5Audio13DistanceModelELj7EE3addEPKcS3_(632392,$8,$9)|0); //@line 49 "./common/StringMap.h"
 $10 = HEAP32[(22828)>>2]|0; //@line 49 "./common/StringMap.h"
 $11 = HEAP32[(22832)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_5audio5Audio13DistanceModelELj7EE3addEPKcS3_(632392,$10,$11)|0); //@line 49 "./common/StringMap.h"
 $12 = HEAP32[(22836)>>2]|0; //@line 49 "./common/StringMap.h"
 $13 = HEAP32[(22840)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_5audio5Audio13DistanceModelELj7EE3addEPKcS3_(632392,$12,$13)|0); //@line 49 "./common/StringMap.h"
 return;
}
function __ZN4love9StringMapINS_5audio5Audio13DistanceModelELj7EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $42 = (($hash$0$lcssa$i) + 6)|0; //@line 112 "./common/StringMap.h"
       $43 = (($42>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
       $44 = (((($this) + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
       $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
       $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
       if ($46) {
        $$lcssa = $43;$$lcssa10 = $44;
        label = 5;
       } else {
        $47 = (($hash$0$lcssa$i) + 7)|0; //@line 112 "./common/StringMap.h"
        $48 = (($47>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
        $49 = (((($this) + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
        $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
        $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
        if ($51) {
         $$lcssa = $48;$$lcssa10 = $49;
         label = 5;
        } else {
         $52 = (($hash$0$lcssa$i) + 8)|0; //@line 112 "./common/StringMap.h"
         $53 = (($52>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
         $54 = (((($this) + (($53*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
         $55 = HEAP8[$54>>0]|0; //@line 114 "./common/StringMap.h"
         $56 = ($55<<24>>24)==(0); //@line 114 "./common/StringMap.h"
         if ($56) {
          $$lcssa = $53;$$lcssa10 = $54;
          label = 5;
         } else {
          $57 = (($hash$0$lcssa$i) + 9)|0; //@line 112 "./common/StringMap.h"
          $58 = (($57>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
          $59 = (((($this) + (($58*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
          $60 = HEAP8[$59>>0]|0; //@line 114 "./common/StringMap.h"
          $61 = ($60<<24>>24)==(0); //@line 114 "./common/StringMap.h"
          if ($61) {
           $$lcssa = $58;$$lcssa10 = $59;
           label = 5;
          } else {
           $62 = (($hash$0$lcssa$i) + 10)|0; //@line 112 "./common/StringMap.h"
           $63 = (($62>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
           $64 = (((($this) + (($63*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
           $65 = HEAP8[$64>>0]|0; //@line 114 "./common/StringMap.h"
           $66 = ($65<<24>>24)==(0); //@line 114 "./common/StringMap.h"
           if ($66) {
            $$lcssa = $63;$$lcssa10 = $64;
            label = 5;
           } else {
            $67 = (($hash$0$lcssa$i) + 11)|0; //@line 112 "./common/StringMap.h"
            $68 = (($67>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
            $69 = (((($this) + (($68*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
            $70 = HEAP8[$69>>0]|0; //@line 114 "./common/StringMap.h"
            $71 = ($70<<24>>24)==(0); //@line 114 "./common/StringMap.h"
            if ($71) {
             $$lcssa = $68;$$lcssa10 = $69;
             label = 5;
            } else {
             $72 = (($hash$0$lcssa$i) + 12)|0; //@line 112 "./common/StringMap.h"
             $73 = (($72>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
             $74 = (((($this) + (($73*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
             $75 = HEAP8[$74>>0]|0; //@line 114 "./common/StringMap.h"
             $76 = ($75<<24>>24)==(0); //@line 114 "./common/StringMap.h"
             if ($76) {
              $$lcssa = $73;$$lcssa10 = $74;
              label = 5;
             } else {
              $77 = (($hash$0$lcssa$i) + 13)|0; //@line 112 "./common/StringMap.h"
              $78 = (($77>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
              $79 = (((($this) + (($78*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
              $80 = HEAP8[$79>>0]|0; //@line 114 "./common/StringMap.h"
              $81 = ($80<<24>>24)==(0); //@line 114 "./common/StringMap.h"
              if ($81) {
               $$lcssa = $78;$$lcssa10 = $79;
               label = 5;
              } else {
               $inserted$2$off0 = 0;
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(6); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 168|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love5audio5Audio11getConstantEPKcRNS1_13DistanceModelE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 14)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((632392 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (632392 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(14); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((632392 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 43 "modules/audio/Audio.cpp"
  return ($$$2$i|0); //@line 43 "modules/audio/Audio.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 43 "modules/audio/Audio.cpp"
  return ($$$2$i|0); //@line 43 "modules/audio/Audio.cpp"
 }
 return (0)|0;
}
function __ZN4love5audio5Audio11getConstantENS1_13DistanceModelERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(6); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (632560 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 48 "modules/audio/Audio.cpp"
}
function __ZN4love5audio6SourceD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 35 "modules/audio/Source.cpp"
}
function __ZN4love5audio6SourceD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 34 "modules/audio/Source.cpp"
 return; //@line 35 "modules/audio/Source.cpp"
}
function __ZNK4love5audio6Source7getTypeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 39 "modules/audio/Source.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 39 "modules/audio/Source.cpp"
 return ($1|0); //@line 39 "modules/audio/Source.cpp"
}
function __GLOBAL__sub_I_Source_cpp() {
 var $$02$i$i$1$i = 0, $$02$i$i$1$i8 = 0, $$02$i$i$i = 0, $$02$i$i$i1 = 0, $$lcssa$i$1$i = 0, $$lcssa$i$1$i13 = 0, $$lcssa$i$i = 0, $$lcssa$i$i6 = 0, $$lcssa10$i$1$i = 0, $$lcssa10$i$1$i12 = 0, $$lcssa10$i$i = 0, $$lcssa10$i$i5 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $hash$0$lcssa$i$i$1$i = 0, $hash$0$lcssa$i$i$1$i11 = 0, $hash$0$lcssa$i$i$i = 0, $hash$0$lcssa$i$i$i4 = 0;
 var $hash$01$i$i$1$i = 0, $hash$01$i$i$1$i9 = 0, $hash$01$i$i$i = 0, $hash$01$i$i$i2 = 0, $vararg_buffer = 0, $vararg_buffer10 = 0, $vararg_buffer2 = 0, $vararg_buffer6 = 0, $vararg_ptr1 = 0, $vararg_ptr13 = 0, $vararg_ptr5 = 0, $vararg_ptr9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer10 = sp + 24|0;
 $vararg_buffer6 = sp + 16|0;
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 HEAP8[(632596)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632608)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632620)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632632)>>0] = 0; //@line 155 "./common/StringMap.h"
 $0 = (632636); //@line 44 "./common/StringMap.h"
 $1 = $0; //@line 44 "./common/StringMap.h"
 HEAP32[$1>>2] = 0; //@line 44 "./common/StringMap.h"
 $2 = (($0) + 4)|0; //@line 44 "./common/StringMap.h"
 $3 = $2; //@line 44 "./common/StringMap.h"
 HEAP32[$3>>2] = 0; //@line 44 "./common/StringMap.h"
 $4 = HEAP32[5758]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(23036)>>2]|0; //@line 49 "./common/StringMap.h"
 $6 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
 $7 = ($6<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($7) {
  $hash$0$lcssa$i$i$i = 5381;
 } else {
  $$02$i$i$i = $4;$9 = $6;$hash$01$i$i$i = 5381;
  while(1) {
   $8 = $9 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $10 = ((($$02$i$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $11 = ($hash$01$i$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $12 = (($11) + ($8))|0; //@line 143 "./common/StringMap.h"
   $13 = HEAP8[$10>>0]|0; //@line 142 "./common/StringMap.h"
   $14 = ($13<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($14) {
    $hash$0$lcssa$i$i$i = $12;
    break;
   } else {
    $$02$i$i$i = $10;$9 = $13;$hash$01$i$i$i = $12;
   }
  }
 }
 $15 = $hash$0$lcssa$i$i$i & 3; //@line 112 "./common/StringMap.h"
 $16 = (((632588 + (($15*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $17 = HEAP8[$16>>0]|0; //@line 114 "./common/StringMap.h"
 $18 = ($17<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($18) {
  $$lcssa$i$i = $15;$$lcssa10$i$i = $16;
  label = 5;
 } else {
  $19 = (($hash$0$lcssa$i$i$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $20 = $19 & 3; //@line 112 "./common/StringMap.h"
  $21 = (((632588 + (($20*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $22 = HEAP8[$21>>0]|0; //@line 114 "./common/StringMap.h"
  $23 = ($22<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($23) {
   $$lcssa$i$i = $20;$$lcssa10$i$i = $21;
   label = 5;
  } else {
   $28 = (($hash$0$lcssa$i$i$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $29 = $28 & 3; //@line 112 "./common/StringMap.h"
   $30 = (((632588 + (($29*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $31 = HEAP8[$30>>0]|0; //@line 114 "./common/StringMap.h"
   $32 = ($31<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($32) {
    $$lcssa$i$i = $29;$$lcssa10$i$i = $30;
    label = 5;
   } else {
    $33 = (($hash$0$lcssa$i$i$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $34 = $33 & 3; //@line 112 "./common/StringMap.h"
    $35 = (((632588 + (($34*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $36 = HEAP8[$35>>0]|0; //@line 114 "./common/StringMap.h"
    $37 = ($36<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($37) {
     $$lcssa$i$i = $34;$$lcssa10$i$i = $35;
     label = 5;
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10$i$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $24 = (632588 + (($$lcssa$i$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$24>>2] = $4; //@line 118 "./common/StringMap.h"
  $25 = (((632588 + (($$lcssa$i$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$25>>2] = $5; //@line 119 "./common/StringMap.h"
 }
 $26 = ($5>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($26) {
  HEAP32[$vararg_buffer>>2] = $4; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $5; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
 } else {
  $27 = (632636 + ($5<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$27>>2] = $4; //@line 132 "./common/StringMap.h"
 }
 $38 = HEAP32[(23040)>>2]|0; //@line 49 "./common/StringMap.h"
 $39 = HEAP32[(23044)>>2]|0; //@line 49 "./common/StringMap.h"
 $40 = HEAP8[$38>>0]|0; //@line 142 "./common/StringMap.h"
 $41 = ($40<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($41) {
  $hash$0$lcssa$i$i$1$i = 5381;
 } else {
  $$02$i$i$1$i = $38;$43 = $40;$hash$01$i$i$1$i = 5381;
  while(1) {
   $42 = $43 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $44 = ((($$02$i$i$1$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $45 = ($hash$01$i$i$1$i*33)|0; //@line 143 "./common/StringMap.h"
   $46 = (($45) + ($42))|0; //@line 143 "./common/StringMap.h"
   $47 = HEAP8[$44>>0]|0; //@line 142 "./common/StringMap.h"
   $48 = ($47<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($48) {
    $hash$0$lcssa$i$i$1$i = $46;
    break;
   } else {
    $$02$i$i$1$i = $44;$43 = $47;$hash$01$i$i$1$i = $46;
   }
  }
 }
 $49 = $hash$0$lcssa$i$i$1$i & 3; //@line 112 "./common/StringMap.h"
 $50 = (((632588 + (($49*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $51 = HEAP8[$50>>0]|0; //@line 114 "./common/StringMap.h"
 $52 = ($51<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($52) {
  $$lcssa$i$1$i = $49;$$lcssa10$i$1$i = $50;
  label = 17;
 } else {
  $53 = (($hash$0$lcssa$i$i$1$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $54 = $53 & 3; //@line 112 "./common/StringMap.h"
  $55 = (((632588 + (($54*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $56 = HEAP8[$55>>0]|0; //@line 114 "./common/StringMap.h"
  $57 = ($56<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($57) {
   $$lcssa$i$1$i = $54;$$lcssa10$i$1$i = $55;
   label = 17;
  } else {
   $58 = (($hash$0$lcssa$i$i$1$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $59 = $58 & 3; //@line 112 "./common/StringMap.h"
   $60 = (((632588 + (($59*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $61 = HEAP8[$60>>0]|0; //@line 114 "./common/StringMap.h"
   $62 = ($61<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($62) {
    $$lcssa$i$1$i = $59;$$lcssa10$i$1$i = $60;
    label = 17;
   } else {
    $63 = (($hash$0$lcssa$i$i$1$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $64 = $63 & 3; //@line 112 "./common/StringMap.h"
    $65 = (((632588 + (($64*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $66 = HEAP8[$65>>0]|0; //@line 114 "./common/StringMap.h"
    $67 = ($66<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($67) {
     $$lcssa$i$1$i = $64;$$lcssa10$i$1$i = $65;
     label = 17;
    }
   }
  }
 }
 if ((label|0) == 17) {
  HEAP8[$$lcssa10$i$1$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $68 = (632588 + (($$lcssa$i$1$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$68>>2] = $38; //@line 118 "./common/StringMap.h"
  $69 = (((632588 + (($$lcssa$i$1$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$69>>2] = $39; //@line 119 "./common/StringMap.h"
 }
 $70 = ($39>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($70) {
  HEAP32[$vararg_buffer2>>2] = $38; //@line 128 "./common/StringMap.h"
  $vararg_ptr5 = ((($vararg_buffer2)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr5>>2] = $39; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer2)|0); //@line 128 "./common/StringMap.h"
 } else {
  $71 = (632636 + ($39<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$71>>2] = $38; //@line 132 "./common/StringMap.h"
 }
 HEAP8[(632652)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632664)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632676)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632688)>>0] = 0; //@line 155 "./common/StringMap.h"
 $72 = (632692); //@line 44 "./common/StringMap.h"
 $73 = $72; //@line 44 "./common/StringMap.h"
 HEAP32[$73>>2] = 0; //@line 44 "./common/StringMap.h"
 $74 = (($72) + 4)|0; //@line 44 "./common/StringMap.h"
 $75 = $74; //@line 44 "./common/StringMap.h"
 HEAP32[$75>>2] = 0; //@line 44 "./common/StringMap.h"
 $76 = HEAP32[5762]|0; //@line 49 "./common/StringMap.h"
 $77 = HEAP32[(23052)>>2]|0; //@line 49 "./common/StringMap.h"
 $78 = HEAP8[$76>>0]|0; //@line 142 "./common/StringMap.h"
 $79 = ($78<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($79) {
  $hash$0$lcssa$i$i$i4 = 5381;
 } else {
  $$02$i$i$i1 = $76;$81 = $78;$hash$01$i$i$i2 = 5381;
  while(1) {
   $80 = $81 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $82 = ((($$02$i$i$i1)) + 1|0); //@line 142 "./common/StringMap.h"
   $83 = ($hash$01$i$i$i2*33)|0; //@line 143 "./common/StringMap.h"
   $84 = (($83) + ($80))|0; //@line 143 "./common/StringMap.h"
   $85 = HEAP8[$82>>0]|0; //@line 142 "./common/StringMap.h"
   $86 = ($85<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($86) {
    $hash$0$lcssa$i$i$i4 = $84;
    break;
   } else {
    $$02$i$i$i1 = $82;$81 = $85;$hash$01$i$i$i2 = $84;
   }
  }
 }
 $87 = $hash$0$lcssa$i$i$i4 & 3; //@line 112 "./common/StringMap.h"
 $88 = (((632644 + (($87*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $89 = HEAP8[$88>>0]|0; //@line 114 "./common/StringMap.h"
 $90 = ($89<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($90) {
  $$lcssa$i$i6 = $87;$$lcssa10$i$i5 = $88;
  label = 25;
 } else {
  $91 = (($hash$0$lcssa$i$i$i4) + 1)|0; //@line 112 "./common/StringMap.h"
  $92 = $91 & 3; //@line 112 "./common/StringMap.h"
  $93 = (((632644 + (($92*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $94 = HEAP8[$93>>0]|0; //@line 114 "./common/StringMap.h"
  $95 = ($94<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($95) {
   $$lcssa$i$i6 = $92;$$lcssa10$i$i5 = $93;
   label = 25;
  } else {
   $100 = (($hash$0$lcssa$i$i$i4) + 2)|0; //@line 112 "./common/StringMap.h"
   $101 = $100 & 3; //@line 112 "./common/StringMap.h"
   $102 = (((632644 + (($101*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $103 = HEAP8[$102>>0]|0; //@line 114 "./common/StringMap.h"
   $104 = ($103<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($104) {
    $$lcssa$i$i6 = $101;$$lcssa10$i$i5 = $102;
    label = 25;
   } else {
    $105 = (($hash$0$lcssa$i$i$i4) + 3)|0; //@line 112 "./common/StringMap.h"
    $106 = $105 & 3; //@line 112 "./common/StringMap.h"
    $107 = (((632644 + (($106*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $108 = HEAP8[$107>>0]|0; //@line 114 "./common/StringMap.h"
    $109 = ($108<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($109) {
     $$lcssa$i$i6 = $106;$$lcssa10$i$i5 = $107;
     label = 25;
    }
   }
  }
 }
 if ((label|0) == 25) {
  HEAP8[$$lcssa10$i$i5>>0] = 1; //@line 117 "./common/StringMap.h"
  $96 = (632644 + (($$lcssa$i$i6*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$96>>2] = $76; //@line 118 "./common/StringMap.h"
  $97 = (((632644 + (($$lcssa$i$i6*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$97>>2] = $77; //@line 119 "./common/StringMap.h"
 }
 $98 = ($77>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($98) {
  HEAP32[$vararg_buffer6>>2] = $76; //@line 128 "./common/StringMap.h"
  $vararg_ptr9 = ((($vararg_buffer6)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr9>>2] = $77; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer6)|0); //@line 128 "./common/StringMap.h"
 } else {
  $99 = (632692 + ($77<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$99>>2] = $76; //@line 132 "./common/StringMap.h"
 }
 $110 = HEAP32[(23056)>>2]|0; //@line 49 "./common/StringMap.h"
 $111 = HEAP32[(23060)>>2]|0; //@line 49 "./common/StringMap.h"
 $112 = HEAP8[$110>>0]|0; //@line 142 "./common/StringMap.h"
 $113 = ($112<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($113) {
  $hash$0$lcssa$i$i$1$i11 = 5381;
 } else {
  $$02$i$i$1$i8 = $110;$115 = $112;$hash$01$i$i$1$i9 = 5381;
  while(1) {
   $114 = $115 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $116 = ((($$02$i$i$1$i8)) + 1|0); //@line 142 "./common/StringMap.h"
   $117 = ($hash$01$i$i$1$i9*33)|0; //@line 143 "./common/StringMap.h"
   $118 = (($117) + ($114))|0; //@line 143 "./common/StringMap.h"
   $119 = HEAP8[$116>>0]|0; //@line 142 "./common/StringMap.h"
   $120 = ($119<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($120) {
    $hash$0$lcssa$i$i$1$i11 = $118;
    break;
   } else {
    $$02$i$i$1$i8 = $116;$115 = $119;$hash$01$i$i$1$i9 = $118;
   }
  }
 }
 $121 = $hash$0$lcssa$i$i$1$i11 & 3; //@line 112 "./common/StringMap.h"
 $122 = (((632644 + (($121*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $123 = HEAP8[$122>>0]|0; //@line 114 "./common/StringMap.h"
 $124 = ($123<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($124) {
  $$lcssa$i$1$i13 = $121;$$lcssa10$i$1$i12 = $122;
  label = 37;
 } else {
  $125 = (($hash$0$lcssa$i$i$1$i11) + 1)|0; //@line 112 "./common/StringMap.h"
  $126 = $125 & 3; //@line 112 "./common/StringMap.h"
  $127 = (((632644 + (($126*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $128 = HEAP8[$127>>0]|0; //@line 114 "./common/StringMap.h"
  $129 = ($128<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($129) {
   $$lcssa$i$1$i13 = $126;$$lcssa10$i$1$i12 = $127;
   label = 37;
  } else {
   $130 = (($hash$0$lcssa$i$i$1$i11) + 2)|0; //@line 112 "./common/StringMap.h"
   $131 = $130 & 3; //@line 112 "./common/StringMap.h"
   $132 = (((632644 + (($131*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $133 = HEAP8[$132>>0]|0; //@line 114 "./common/StringMap.h"
   $134 = ($133<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($134) {
    $$lcssa$i$1$i13 = $131;$$lcssa10$i$1$i12 = $132;
    label = 37;
   } else {
    $135 = (($hash$0$lcssa$i$i$1$i11) + 3)|0; //@line 112 "./common/StringMap.h"
    $136 = $135 & 3; //@line 112 "./common/StringMap.h"
    $137 = (((632644 + (($136*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $138 = HEAP8[$137>>0]|0; //@line 114 "./common/StringMap.h"
    $139 = ($138<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($139) {
     $$lcssa$i$1$i13 = $136;$$lcssa10$i$1$i12 = $137;
     label = 37;
    }
   }
  }
 }
 if ((label|0) == 37) {
  HEAP8[$$lcssa10$i$1$i12>>0] = 1; //@line 117 "./common/StringMap.h"
  $140 = (632644 + (($$lcssa$i$1$i13*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$140>>2] = $110; //@line 118 "./common/StringMap.h"
  $141 = (((632644 + (($$lcssa$i$1$i13*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$141>>2] = $111; //@line 119 "./common/StringMap.h"
 }
 $142 = ($111>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($142) {
  HEAP32[$vararg_buffer10>>2] = $110; //@line 128 "./common/StringMap.h"
  $vararg_ptr13 = ((($vararg_buffer10)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr13>>2] = $111; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer10)|0); //@line 128 "./common/StringMap.h"
  STACKTOP = sp;return;
 } else {
  $143 = (632692 + ($111<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$143>>2] = $110; //@line 132 "./common/StringMap.h"
  STACKTOP = sp;return;
 }
}
function __ZN4love5audio6SourceC2ENS1_4TypeE($this,$type) {
 $this = $this|0;
 $type = $type|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 28 "modules/audio/Source.cpp"
 HEAP32[$this>>2] = (22852); //@line 30 "modules/audio/Source.cpp"
 $0 = ((($this)) + 8|0); //@line 29 "modules/audio/Source.cpp"
 HEAP32[$0>>2] = $type; //@line 29 "modules/audio/Source.cpp"
 return; //@line 31 "modules/audio/Source.cpp"
}
function __ZN4love5audio6Source11getConstantEPKcRNS1_4TypeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love9StringMapINS_5audio6Source4TypeELj2EE4findEPKcRS3_(632588,$in,$out)|0); //@line 44 "modules/audio/Source.cpp"
 return ($0|0); //@line 44 "modules/audio/Source.cpp"
}
function __ZN4love9StringMapINS_5audio6Source4TypeELj2EE4findEPKcRS3_($this,$key,$t) {
 $this = $this|0;
 $key = $key|0;
 $t = $t|0;
 var $$$2 = 0, $$011$i$137 = 0, $$011$i$233 = 0, $$011$i$330 = 0, $$011$i41 = 0, $$0110$i$138 = 0, $$0110$i$234 = 0, $$0110$i$331 = 0, $$0110$i42 = 0, $$02$i = 0, $$13 = 0, $$2 = 0, $$lcssa12 = 0, $$pre14 = 0, $$pre16 = 0, $$pre18 = 0, $$pre20 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $or$cond$i = 0, $or$cond$i$1 = 0, $or$cond$i$136 = 0, $or$cond$i$2 = 0, $or$cond$i$232 = 0, $or$cond$i$3 = 0, $or$cond$i$329 = 0, $or$cond$i40 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 3; //@line 72 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 74 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 74 "./common/StringMap.h"
 if ($12) {
  $$13 = 1;$$2 = 0;
  $$$2 = $$13 & $$2;
  return ($$$2|0); //@line 85 "./common/StringMap.h"
 }
 $13 = (($this) + (($9*12)|0)|0); //@line 77 "./common/StringMap.h"
 $14 = HEAP32[$13>>2]|0; //@line 77 "./common/StringMap.h"
 $15 = HEAP8[$14>>0]|0; //@line 54 "./common/StringMap.h"
 $16 = ($15<<24>>24)==(0); //@line 54 "./common/StringMap.h"
 L8: do {
  if ($16) {
   $26 = $0;
   label = 8;
  } else {
   $17 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
   $18 = ($15<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
   $or$cond$i40 = $17 & $18; //@line 54 "./common/StringMap.h"
   if ($or$cond$i40) {
    $$011$i41 = $key;$$0110$i42 = $14;
    while(1) {
     $19 = ((($$0110$i42)) + 1|0); //@line 59 "./common/StringMap.h"
     $20 = ((($$011$i41)) + 1|0); //@line 60 "./common/StringMap.h"
     $21 = HEAP8[$19>>0]|0; //@line 54 "./common/StringMap.h"
     $22 = ($21<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     $$pre14 = HEAP8[$20>>0]|0; //@line 63 "./common/StringMap.h"
     if ($22) {
      $26 = $$pre14;
      label = 8;
      break L8;
     }
     $23 = ($$pre14<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $24 = ($21<<24>>24)==($$pre14<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i = $23 & $24; //@line 54 "./common/StringMap.h"
     if ($or$cond$i) {
      $$011$i41 = $20;$$0110$i42 = $19;
     } else {
      label = 10;
      break;
     }
    }
   } else {
    label = 10;
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $25 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
  if ($25) {
   $$lcssa12 = $9;
  } else {
   label = 10;
  }
 }
 do {
  if ((label|0) == 10) {
   $29 = (($hash$0$lcssa$i) + 1)|0; //@line 72 "./common/StringMap.h"
   $30 = $29 & 3; //@line 72 "./common/StringMap.h"
   $31 = (((($this) + (($30*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $32 = HEAP8[$31>>0]|0; //@line 74 "./common/StringMap.h"
   $33 = ($32<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($33) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $34 = (($this) + (($30*12)|0)|0); //@line 77 "./common/StringMap.h"
   $35 = HEAP32[$34>>2]|0; //@line 77 "./common/StringMap.h"
   $36 = HEAP8[$35>>0]|0; //@line 54 "./common/StringMap.h"
   $37 = ($36<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L20: do {
    if ($37) {
     $47 = $0;
     label = 16;
    } else {
     $38 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $39 = ($36<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$136 = $38 & $39; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$136) {
      $$011$i$137 = $key;$$0110$i$138 = $35;
      while(1) {
       $40 = ((($$0110$i$138)) + 1|0); //@line 59 "./common/StringMap.h"
       $41 = ((($$011$i$137)) + 1|0); //@line 60 "./common/StringMap.h"
       $42 = HEAP8[$40>>0]|0; //@line 54 "./common/StringMap.h"
       $43 = ($42<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre16 = HEAP8[$41>>0]|0; //@line 63 "./common/StringMap.h"
       if ($43) {
        $47 = $$pre16;
        label = 16;
        break L20;
       }
       $44 = ($$pre16<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $45 = ($42<<24>>24)==($$pre16<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$1 = $44 & $45; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$1) {
        $$011$i$137 = $41;$$0110$i$138 = $40;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 16) {
    $46 = ($47<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($46) {
     $$lcssa12 = $30;
     break;
    }
   }
   $48 = (($hash$0$lcssa$i) + 2)|0; //@line 72 "./common/StringMap.h"
   $49 = $48 & 3; //@line 72 "./common/StringMap.h"
   $50 = (((($this) + (($49*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $51 = HEAP8[$50>>0]|0; //@line 74 "./common/StringMap.h"
   $52 = ($51<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($52) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $53 = (($this) + (($49*12)|0)|0); //@line 77 "./common/StringMap.h"
   $54 = HEAP32[$53>>2]|0; //@line 77 "./common/StringMap.h"
   $55 = HEAP8[$54>>0]|0; //@line 54 "./common/StringMap.h"
   $56 = ($55<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L31: do {
    if ($56) {
     $66 = $0;
     label = 22;
    } else {
     $57 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $58 = ($55<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$232 = $57 & $58; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$232) {
      $$011$i$233 = $key;$$0110$i$234 = $54;
      while(1) {
       $59 = ((($$0110$i$234)) + 1|0); //@line 59 "./common/StringMap.h"
       $60 = ((($$011$i$233)) + 1|0); //@line 60 "./common/StringMap.h"
       $61 = HEAP8[$59>>0]|0; //@line 54 "./common/StringMap.h"
       $62 = ($61<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre18 = HEAP8[$60>>0]|0; //@line 63 "./common/StringMap.h"
       if ($62) {
        $66 = $$pre18;
        label = 22;
        break L31;
       }
       $63 = ($$pre18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $64 = ($61<<24>>24)==($$pre18<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$2 = $63 & $64; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$2) {
        $$011$i$233 = $60;$$0110$i$234 = $59;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 22) {
    $65 = ($66<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($65) {
     $$lcssa12 = $49;
     break;
    }
   }
   $67 = (($hash$0$lcssa$i) + 3)|0; //@line 72 "./common/StringMap.h"
   $68 = $67 & 3; //@line 72 "./common/StringMap.h"
   $69 = (((($this) + (($68*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $70 = HEAP8[$69>>0]|0; //@line 74 "./common/StringMap.h"
   $71 = ($70<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($71) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $72 = (($this) + (($68*12)|0)|0); //@line 77 "./common/StringMap.h"
   $73 = HEAP32[$72>>2]|0; //@line 77 "./common/StringMap.h"
   $74 = HEAP8[$73>>0]|0; //@line 54 "./common/StringMap.h"
   $75 = ($74<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L42: do {
    if ($75) {
     $85 = $0;
    } else {
     $76 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $77 = ($74<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$329 = $76 & $77; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$329) {
      $$011$i$330 = $key;$$0110$i$331 = $73;
     } else {
      $$13 = 0;$$2 = 0;
      $$$2 = $$13 & $$2;
      return ($$$2|0); //@line 85 "./common/StringMap.h"
     }
     while(1) {
      $78 = ((($$0110$i$331)) + 1|0); //@line 59 "./common/StringMap.h"
      $79 = ((($$011$i$330)) + 1|0); //@line 60 "./common/StringMap.h"
      $80 = HEAP8[$78>>0]|0; //@line 54 "./common/StringMap.h"
      $81 = ($80<<24>>24)==(0); //@line 54 "./common/StringMap.h"
      $$pre20 = HEAP8[$79>>0]|0; //@line 63 "./common/StringMap.h"
      if ($81) {
       $85 = $$pre20;
       break L42;
      }
      $82 = ($$pre20<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
      $83 = ($80<<24>>24)==($$pre20<<24>>24); //@line 56 "./common/StringMap.h"
      $or$cond$i$3 = $82 & $83; //@line 54 "./common/StringMap.h"
      if ($or$cond$i$3) {
       $$011$i$330 = $79;$$0110$i$331 = $78;
      } else {
       $$13 = 0;$$2 = 0;
       break;
      }
     }
     $$$2 = $$13 & $$2;
     return ($$$2|0); //@line 85 "./common/StringMap.h"
    }
   } while(0);
   $84 = ($85<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($84) {
    $$lcssa12 = $68;
   } else {
    $$13 = 0;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
  }
 } while(0);
 $27 = (((($this) + (($$lcssa12*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
 $28 = HEAP32[$27>>2]|0; //@line 79 "./common/StringMap.h"
 HEAP32[$t>>2] = $28; //@line 79 "./common/StringMap.h"
 $$13 = 1;$$2 = 1;
 $$$2 = $$13 & $$2;
 return ($$$2|0); //@line 85 "./common/StringMap.h"
}
function __ZN4love5audio6Source11getConstantENS1_4TypeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(1); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (632636 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 49 "modules/audio/Source.cpp"
}
function __ZN4love5audio6Source11getConstantEPKcRNS1_4UnitE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love9StringMapINS_5audio6Source4UnitELj2EE4findEPKcRS3_(632644,$in,$out)|0); //@line 54 "modules/audio/Source.cpp"
 return ($0|0); //@line 54 "modules/audio/Source.cpp"
}
function __ZN4love9StringMapINS_5audio6Source4UnitELj2EE4findEPKcRS3_($this,$key,$t) {
 $this = $this|0;
 $key = $key|0;
 $t = $t|0;
 var $$$2 = 0, $$011$i$137 = 0, $$011$i$233 = 0, $$011$i$330 = 0, $$011$i41 = 0, $$0110$i$138 = 0, $$0110$i$234 = 0, $$0110$i$331 = 0, $$0110$i42 = 0, $$02$i = 0, $$13 = 0, $$2 = 0, $$lcssa12 = 0, $$pre14 = 0, $$pre16 = 0, $$pre18 = 0, $$pre20 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $or$cond$i = 0, $or$cond$i$1 = 0, $or$cond$i$136 = 0, $or$cond$i$2 = 0, $or$cond$i$232 = 0, $or$cond$i$3 = 0, $or$cond$i$329 = 0, $or$cond$i40 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 3; //@line 72 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 74 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 74 "./common/StringMap.h"
 if ($12) {
  $$13 = 1;$$2 = 0;
  $$$2 = $$13 & $$2;
  return ($$$2|0); //@line 85 "./common/StringMap.h"
 }
 $13 = (($this) + (($9*12)|0)|0); //@line 77 "./common/StringMap.h"
 $14 = HEAP32[$13>>2]|0; //@line 77 "./common/StringMap.h"
 $15 = HEAP8[$14>>0]|0; //@line 54 "./common/StringMap.h"
 $16 = ($15<<24>>24)==(0); //@line 54 "./common/StringMap.h"
 L8: do {
  if ($16) {
   $26 = $0;
   label = 8;
  } else {
   $17 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
   $18 = ($15<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
   $or$cond$i40 = $17 & $18; //@line 54 "./common/StringMap.h"
   if ($or$cond$i40) {
    $$011$i41 = $key;$$0110$i42 = $14;
    while(1) {
     $19 = ((($$0110$i42)) + 1|0); //@line 59 "./common/StringMap.h"
     $20 = ((($$011$i41)) + 1|0); //@line 60 "./common/StringMap.h"
     $21 = HEAP8[$19>>0]|0; //@line 54 "./common/StringMap.h"
     $22 = ($21<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     $$pre14 = HEAP8[$20>>0]|0; //@line 63 "./common/StringMap.h"
     if ($22) {
      $26 = $$pre14;
      label = 8;
      break L8;
     }
     $23 = ($$pre14<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $24 = ($21<<24>>24)==($$pre14<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i = $23 & $24; //@line 54 "./common/StringMap.h"
     if ($or$cond$i) {
      $$011$i41 = $20;$$0110$i42 = $19;
     } else {
      label = 10;
      break;
     }
    }
   } else {
    label = 10;
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $25 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
  if ($25) {
   $$lcssa12 = $9;
  } else {
   label = 10;
  }
 }
 do {
  if ((label|0) == 10) {
   $29 = (($hash$0$lcssa$i) + 1)|0; //@line 72 "./common/StringMap.h"
   $30 = $29 & 3; //@line 72 "./common/StringMap.h"
   $31 = (((($this) + (($30*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $32 = HEAP8[$31>>0]|0; //@line 74 "./common/StringMap.h"
   $33 = ($32<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($33) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $34 = (($this) + (($30*12)|0)|0); //@line 77 "./common/StringMap.h"
   $35 = HEAP32[$34>>2]|0; //@line 77 "./common/StringMap.h"
   $36 = HEAP8[$35>>0]|0; //@line 54 "./common/StringMap.h"
   $37 = ($36<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L20: do {
    if ($37) {
     $47 = $0;
     label = 16;
    } else {
     $38 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $39 = ($36<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$136 = $38 & $39; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$136) {
      $$011$i$137 = $key;$$0110$i$138 = $35;
      while(1) {
       $40 = ((($$0110$i$138)) + 1|0); //@line 59 "./common/StringMap.h"
       $41 = ((($$011$i$137)) + 1|0); //@line 60 "./common/StringMap.h"
       $42 = HEAP8[$40>>0]|0; //@line 54 "./common/StringMap.h"
       $43 = ($42<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre16 = HEAP8[$41>>0]|0; //@line 63 "./common/StringMap.h"
       if ($43) {
        $47 = $$pre16;
        label = 16;
        break L20;
       }
       $44 = ($$pre16<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $45 = ($42<<24>>24)==($$pre16<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$1 = $44 & $45; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$1) {
        $$011$i$137 = $41;$$0110$i$138 = $40;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 16) {
    $46 = ($47<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($46) {
     $$lcssa12 = $30;
     break;
    }
   }
   $48 = (($hash$0$lcssa$i) + 2)|0; //@line 72 "./common/StringMap.h"
   $49 = $48 & 3; //@line 72 "./common/StringMap.h"
   $50 = (((($this) + (($49*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $51 = HEAP8[$50>>0]|0; //@line 74 "./common/StringMap.h"
   $52 = ($51<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($52) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $53 = (($this) + (($49*12)|0)|0); //@line 77 "./common/StringMap.h"
   $54 = HEAP32[$53>>2]|0; //@line 77 "./common/StringMap.h"
   $55 = HEAP8[$54>>0]|0; //@line 54 "./common/StringMap.h"
   $56 = ($55<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L31: do {
    if ($56) {
     $66 = $0;
     label = 22;
    } else {
     $57 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $58 = ($55<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$232 = $57 & $58; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$232) {
      $$011$i$233 = $key;$$0110$i$234 = $54;
      while(1) {
       $59 = ((($$0110$i$234)) + 1|0); //@line 59 "./common/StringMap.h"
       $60 = ((($$011$i$233)) + 1|0); //@line 60 "./common/StringMap.h"
       $61 = HEAP8[$59>>0]|0; //@line 54 "./common/StringMap.h"
       $62 = ($61<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre18 = HEAP8[$60>>0]|0; //@line 63 "./common/StringMap.h"
       if ($62) {
        $66 = $$pre18;
        label = 22;
        break L31;
       }
       $63 = ($$pre18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $64 = ($61<<24>>24)==($$pre18<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$2 = $63 & $64; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$2) {
        $$011$i$233 = $60;$$0110$i$234 = $59;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 22) {
    $65 = ($66<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($65) {
     $$lcssa12 = $49;
     break;
    }
   }
   $67 = (($hash$0$lcssa$i) + 3)|0; //@line 72 "./common/StringMap.h"
   $68 = $67 & 3; //@line 72 "./common/StringMap.h"
   $69 = (((($this) + (($68*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $70 = HEAP8[$69>>0]|0; //@line 74 "./common/StringMap.h"
   $71 = ($70<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($71) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $72 = (($this) + (($68*12)|0)|0); //@line 77 "./common/StringMap.h"
   $73 = HEAP32[$72>>2]|0; //@line 77 "./common/StringMap.h"
   $74 = HEAP8[$73>>0]|0; //@line 54 "./common/StringMap.h"
   $75 = ($74<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L42: do {
    if ($75) {
     $85 = $0;
    } else {
     $76 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $77 = ($74<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$329 = $76 & $77; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$329) {
      $$011$i$330 = $key;$$0110$i$331 = $73;
     } else {
      $$13 = 0;$$2 = 0;
      $$$2 = $$13 & $$2;
      return ($$$2|0); //@line 85 "./common/StringMap.h"
     }
     while(1) {
      $78 = ((($$0110$i$331)) + 1|0); //@line 59 "./common/StringMap.h"
      $79 = ((($$011$i$330)) + 1|0); //@line 60 "./common/StringMap.h"
      $80 = HEAP8[$78>>0]|0; //@line 54 "./common/StringMap.h"
      $81 = ($80<<24>>24)==(0); //@line 54 "./common/StringMap.h"
      $$pre20 = HEAP8[$79>>0]|0; //@line 63 "./common/StringMap.h"
      if ($81) {
       $85 = $$pre20;
       break L42;
      }
      $82 = ($$pre20<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
      $83 = ($80<<24>>24)==($$pre20<<24>>24); //@line 56 "./common/StringMap.h"
      $or$cond$i$3 = $82 & $83; //@line 54 "./common/StringMap.h"
      if ($or$cond$i$3) {
       $$011$i$330 = $79;$$0110$i$331 = $78;
      } else {
       $$13 = 0;$$2 = 0;
       break;
      }
     }
     $$$2 = $$13 & $$2;
     return ($$$2|0); //@line 85 "./common/StringMap.h"
    }
   } while(0);
   $84 = ($85<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($84) {
    $$lcssa12 = $68;
   } else {
    $$13 = 0;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
  }
 } while(0);
 $27 = (((($this) + (($$lcssa12*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
 $28 = HEAP32[$27>>2]|0; //@line 79 "./common/StringMap.h"
 HEAP32[$t>>2] = $28; //@line 79 "./common/StringMap.h"
 $$13 = 1;$$2 = 1;
 $$$2 = $$13 & $$2;
 return ($$$2|0); //@line 85 "./common/StringMap.h"
}
function __ZN4love5event3sdl5EventD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (23072); //@line 107 "modules/event/sdl/Event.cpp"
 __THREW__ = 0;
 invoke_vii(3934,(3935|0),($this|0)); //@line 108 "modules/event/sdl/Event.cpp"
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 if (!($1)) {
  __THREW__ = 0;
  invoke_vi(3800,16384); //@line 109 "modules/event/sdl/Event.cpp"
  $2 = __THREW__; __THREW__ = 0;
  $3 = $2&1;
  if (!($3)) {
   __ZN4love5event5EventD2Ev($this); //@line 110 "modules/event/sdl/Event.cpp"
   return; //@line 110 "modules/event/sdl/Event.cpp"
  }
 }
 $4 = ___cxa_find_matching_catch_3(0|0)|0;
 $5 = tempRet0;
 __ZN4love5event5EventD2Ev($this); //@line 110 "modules/event/sdl/Event.cpp"
 ___clang_call_terminate($4); //@line 110 "modules/event/sdl/Event.cpp"
 // unreachable; //@line 110 "modules/event/sdl/Event.cpp"
}
function __ZN4love5event3sdlL14watchAppEventsEPvP9SDL_Event($0,$event) {
 $0 = $0|0;
 $event = $event|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $or$cond = 0, $switch = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $2 = HEAP32[$event>>2]|0; //@line 76 "modules/event/sdl/Event.cpp"
 $3 = $2 & -2; //@line 76 "modules/event/sdl/Event.cpp"
 $switch = ($3|0)==(260); //@line 76 "modules/event/sdl/Event.cpp"
 $4 = ($1|0)!=(0|0); //@line 83 "modules/event/sdl/Event.cpp"
 $or$cond = $4 & $switch; //@line 76 "modules/event/sdl/Event.cpp"
 if (!($or$cond)) {
  return 1; //@line 90 "modules/event/sdl/Event.cpp"
 }
 $5 = HEAP32[$1>>2]|0; //@line 84 "modules/event/sdl/Event.cpp"
 $6 = ((($5)) + 36|0); //@line 84 "modules/event/sdl/Event.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 84 "modules/event/sdl/Event.cpp"
 $8 = ($2|0)==(261); //@line 84 "modules/event/sdl/Event.cpp"
 FUNCTION_TABLE_vii[$7 & 8191]($1,$8); //@line 84 "modules/event/sdl/Event.cpp"
 return 1; //@line 90 "modules/event/sdl/Event.cpp"
}
function __ZN4love5event3sdl5EventD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (23072); //@line 107 "modules/event/sdl/Event.cpp"
 __THREW__ = 0;
 invoke_vii(3934,(3935|0),($this|0)); //@line 108 "modules/event/sdl/Event.cpp"
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 if (!($1)) {
  __THREW__ = 0;
  invoke_vi(3800,16384); //@line 109 "modules/event/sdl/Event.cpp"
  $2 = __THREW__; __THREW__ = 0;
  $3 = $2&1;
  if (!($3)) {
   __ZN4love5event5EventD2Ev($this); //@line 110 "modules/event/sdl/Event.cpp"
   __ZdlPv($this); //@line 107 "modules/event/sdl/Event.cpp"
   return; //@line 110 "modules/event/sdl/Event.cpp"
  }
 }
 $4 = ___cxa_find_matching_catch_3(0|0)|0;
 $5 = tempRet0;
 __ZN4love5event5EventD2Ev($this); //@line 110 "modules/event/sdl/Event.cpp"
 ___clang_call_terminate($4); //@line 110 "modules/event/sdl/Event.cpp"
 // unreachable; //@line 110 "modules/event/sdl/Event.cpp"
}
function __ZNK4love5event5Event13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1; //@line 65 "./modules/event/Event.h"
}
function __ZNK4love5event3sdl5Event7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (365502|0); //@line 95 "modules/event/sdl/Event.cpp"
}
function __ZN4love5event3sdl5Event5clearEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $e = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $e = sp;
 while(1) {
  $0 = (_SDL_PollEvent($e)|0); //@line 141 "modules/event/sdl/Event.cpp"
  $1 = ($0|0)==(0); //@line 141 "modules/event/sdl/Event.cpp"
  if ($1) {
   break;
  }
 }
 __ZN4love5event5Event5clearEv($this); //@line 146 "modules/event/sdl/Event.cpp"
 STACKTOP = sp;return; //@line 147 "modules/event/sdl/Event.cpp"
}
function __ZN4love5event3sdl5Event4pumpEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $e = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $e = sp;
 $0 = (_SDL_PollEvent($e)|0); //@line 116 "modules/event/sdl/Event.cpp"
 $1 = ($0|0)==(0); //@line 116 "modules/event/sdl/Event.cpp"
 if ($1) {
  STACKTOP = sp;return; //@line 125 "modules/event/sdl/Event.cpp"
 }
 while(1) {
  $2 = (__ZNK4love5event3sdl5Event7convertERK9SDL_Event($this,$e)|0); //@line 118 "modules/event/sdl/Event.cpp"
  $3 = ($2|0)==(0|0); //@line 119 "modules/event/sdl/Event.cpp"
  if (!($3)) {
   __ZN4love5event5Event4pushEPNS0_7MessageE($this,$2); //@line 121 "modules/event/sdl/Event.cpp"
   $6 = HEAP32[$2>>2]|0; //@line 122 "modules/event/sdl/Event.cpp"
   $7 = ((($6)) + 12|0); //@line 122 "modules/event/sdl/Event.cpp"
   $8 = HEAP32[$7>>2]|0; //@line 122 "modules/event/sdl/Event.cpp"
   FUNCTION_TABLE_vi[$8 & 8191]($2); //@line 122 "modules/event/sdl/Event.cpp"
  }
  $4 = (_SDL_PollEvent($e)|0); //@line 116 "modules/event/sdl/Event.cpp"
  $5 = ($4|0)==(0); //@line 116 "modules/event/sdl/Event.cpp"
  if ($5) {
   break;
  }
 }
 STACKTOP = sp;return; //@line 125 "modules/event/sdl/Event.cpp"
}
function __ZNK4love5event3sdl5Event7convertERK9SDL_Event($this,$e) {
 $this = $this|0;
 $e = $e|0;
 var $$0 = 0, $$0$ph$lcssa$i$i$i = 0, $$0$ph$lcssa$i$i$i64 = 0, $$0$ph6$i$i$i = 0, $$0$ph6$i$i$i59 = 0, $$01$ph5$i$i$i = 0, $$01$ph5$i$i$i60 = 0, $$013$i$i$i = 0, $$013$i$i$i$lcssa = 0, $$013$i$i$i62 = 0, $$013$i$i$i62$lcssa = 0, $$02 = 0, $$1 = 0, $$10 = 0, $$1013 = 0, $$11 = 0, $$1114 = 0, $$12 = 0, $$1215 = 0, $$13 = 0;
 var $$1316 = 0, $$14 = 0, $$2 = 0, $$25 = 0, $$3 = 0, $$36 = 0, $$4 = 0, $$47 = 0, $$5 = 0, $$58 = 0, $$6 = 0, $$69 = 0, $$7 = 0, $$710 = 0, $$8 = 0, $$811 = 0, $$9 = 0, $$912 = 0, $$pr = 0, $$pre = 0.0;
 var $$pre$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i20 = 0, $$pre$i$i$i29 = 0, $$pre$i$i$i334 = 0, $$pre$i$i$i343 = 0, $$pre$i$i$i354 = 0, $$pre$i102 = 0, $$pre$i110 = 0, $$pre$i118 = 0, $$pre$i134 = 0, $$pre$i142 = 0, $$pre$i150 = 0, $$pre$i158 = 0, $$pre$i166 = 0, $$pre$i183 = 0, $$pre$i191 = 0, $$pre$i199 = 0, $$pre$i207 = 0, $$pre$i221 = 0;
 var $$pre$i229 = 0, $$pre$i23 = 0, $$pre$i248 = 0, $$pre$i256 = 0, $$pre$i264 = 0, $$pre$i272 = 0, $$pre$i280 = 0, $$pre$i288 = 0, $$pre$i308 = 0, $$pre$i32 = 0, $$pre$i321 = 0, $$pre$i337 = 0, $$pre$i346 = 0, $$pre$i357 = 0, $$pre$i36 = 0, $$pre$i40 = 0, $$pre$i48 = 0, $$pre$i71 = 0, $$pre$i79 = 0, $$pre$i91 = 0;
 var $$pre$phi397Z2D = 0, $$pre386 = 0, $$pre387 = 0.0, $$pre388 = 0, $$pre389 = 0, $$pre391 = 0, $$pre392 = 0, $$pre393 = 0, $$pre395 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0;
 var $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0.0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0;
 var $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0;
 var $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0;
 var $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0;
 var $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0.0, $1085 = 0.0, $1086 = 0, $1087 = 0, $1088 = 0.0, $1089 = 0.0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0.0, $1093 = 0.0, $1094 = 0, $1095 = 0, $1096 = 0.0, $1097 = 0.0;
 var $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0.0, $1101 = 0.0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0.0, $1111 = 0, $1112 = 0.0, $1113 = 0.0, $1114 = 0.0;
 var $1115 = 0, $1116 = 0.0, $1117 = 0.0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0.0, $1126 = 0.0, $1127 = 0, $1128 = 0.0, $1129 = 0.0, $113 = 0, $1130 = 0.0, $1131 = 0, $1132 = 0.0;
 var $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0;
 var $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0;
 var $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0.0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0;
 var $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0;
 var $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0.0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0;
 var $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0;
 var $1241 = 0.0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0;
 var $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0, $1267 = 0, $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0.0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0;
 var $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0, $1285 = 0, $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0;
 var $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0.0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0, $1302 = 0, $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0;
 var $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0, $1320 = 0, $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0;
 var $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0, $1339 = 0, $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0;
 var $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0, $1357 = 0, $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0;
 var $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0, $1375 = 0, $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0;
 var $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0, $1393 = 0, $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0;
 var $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0, $1410 = 0, $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0;
 var $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0, $1429 = 0, $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0;
 var $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0, $1447 = 0, $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0;
 var $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0, $1465 = 0, $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0;
 var $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0, $1483 = 0, $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0;
 var $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0, $1500 = 0, $1501 = 0, $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0;
 var $1511 = 0, $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0, $1518 = 0, $1519 = 0, $152 = 0, $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0;
 var $153 = 0, $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0, $1536 = 0, $1537 = 0, $1538 = 0, $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0;
 var $1548 = 0, $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0, $1554 = 0, $1555 = 0, $1556 = 0, $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $1565 = 0;
 var $1566 = 0, $1567 = 0, $1568 = 0, $1569 = 0, $157 = 0, $1570 = 0, $1571 = 0, $1572 = 0, $1573 = 0, $1574 = 0, $1575 = 0, $1576 = 0, $1577 = 0, $1578 = 0, $1579 = 0, $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0, $1583 = 0;
 var $1584 = 0, $1585 = 0, $1586 = 0, $1587 = 0, $1588 = 0, $1589 = 0, $159 = 0, $1590 = 0, $1591 = 0, $1592 = 0, $1593 = 0, $1594 = 0, $1595 = 0, $1596 = 0, $1597 = 0, $1598 = 0, $1599 = 0, $16 = 0, $160 = 0, $1600 = 0;
 var $1601 = 0, $1602 = 0, $1603 = 0, $1604 = 0, $1605 = 0, $1606 = 0, $1607 = 0, $1608 = 0, $1609 = 0, $161 = 0, $1610 = 0, $1611 = 0, $1612 = 0, $1613 = 0, $1614 = 0, $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0, $1619 = 0;
 var $162 = 0, $1620 = 0, $1621 = 0, $1622 = 0, $1623 = 0, $1624 = 0, $1625 = 0, $1626 = 0, $1627 = 0, $1628 = 0, $1629 = 0, $163 = 0, $1630 = 0, $1631 = 0, $1632 = 0, $1633 = 0, $1634 = 0, $1635 = 0, $1636 = 0, $1637 = 0;
 var $1638 = 0, $1639 = 0, $164 = 0, $1640 = 0, $1641 = 0, $1642 = 0, $1643 = 0, $1644 = 0, $1645 = 0, $1646 = 0, $1647 = 0, $1648 = 0, $1649 = 0, $165 = 0, $1650 = 0, $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0, $1655 = 0;
 var $1656 = 0, $1657 = 0, $1658 = 0, $1659 = 0, $166 = 0, $1660 = 0, $1661 = 0, $1662 = 0, $1663 = 0, $1664 = 0, $1665 = 0, $1666 = 0, $1667 = 0, $1668 = 0, $1669 = 0, $167 = 0, $1670 = 0, $1671 = 0, $1672 = 0, $1673 = 0;
 var $1674 = 0, $1675 = 0, $1676 = 0, $1677 = 0, $1678 = 0, $1679 = 0, $168 = 0, $1680 = 0, $1681 = 0, $1682 = 0, $1683 = 0, $1684 = 0, $1685 = 0, $1686 = 0, $1687 = 0, $1688 = 0, $1689 = 0, $169 = 0, $1690 = 0, $1691 = 0;
 var $1692 = 0, $1693 = 0, $1694 = 0, $1695 = 0, $1696 = 0, $1697 = 0, $1698 = 0, $1699 = 0, $17 = 0, $170 = 0, $1700 = 0, $1701 = 0, $1702 = 0, $1703 = 0, $1704 = 0, $1705 = 0, $1706 = 0, $1707 = 0, $1708 = 0, $1709 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0;
 var $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0;
 var $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0;
 var $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0;
 var $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0;
 var $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0;
 var $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0;
 var $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0;
 var $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0;
 var $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0;
 var $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0;
 var $441 = 0, $442 = 0.0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0;
 var $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0.0, $474 = 0, $475 = 0, $476 = 0, $477 = 0;
 var $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0;
 var $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0;
 var $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0;
 var $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0.0, $546 = 0, $547 = 0, $548 = 0.0, $549 = 0;
 var $55 = 0, $550 = 0, $551 = 0.0, $552 = 0, $553 = 0, $554 = 0.0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0;
 var $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0.0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0;
 var $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0.0, $602 = 0;
 var $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0;
 var $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0.0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0;
 var $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0;
 var $658 = 0, $659 = 0.0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0;
 var $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0;
 var $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0;
 var $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0;
 var $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0;
 var $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0;
 var $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0;
 var $784 = 0, $785 = 0, $786 = 0, $787 = 0.0, $788 = 0, $789 = 0, $79 = 0, $790 = 0.0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0;
 var $801 = 0.0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0;
 var $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0.0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0;
 var $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0;
 var $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0.0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0;
 var $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0;
 var $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0;
 var $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0;
 var $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0;
 var $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0;
 var $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0.0;
 var $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $__begin$sroa$0$0378 = 0;
 var $button$0 = 0, $h$i = 0, $h$i241 = 0, $key$0 = 0, $key$1 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $msg$0 = 0, $msg$1 = 0, $proxy = 0, $scancode = 0, $touchinfo = 0, $txt = 0, $txt2 = 0, $vargs = 0, $w$i = 0, $w$i240 = 0, $x = 0, $x1 = 0, $xrel = 0;
 var $y = 0, $y2 = 0, $yrel = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 480|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $w$i240 = sp + 472|0;
 $h$i241 = sp + 468|0;
 $w$i = sp + 464|0;
 $h$i = sp + 460|0;
 $vargs = sp + 448|0;
 $scancode = sp + 444|0;
 $txt = sp + 440|0;
 $txt2 = sp + 436|0;
 $touchinfo = sp + 48|0;
 $0 = sp + 432|0;
 $1 = sp + 428|0;
 $2 = sp + 424|0;
 $3 = sp + 412|0;
 $4 = sp + 408|0;
 $5 = sp + 404|0;
 $6 = sp + 392|0;
 $7 = sp + 388|0;
 $8 = sp + 376|0;
 $9 = sp + 372|0;
 $10 = sp + 368|0;
 $11 = sp + 364|0;
 $12 = sp + 352|0;
 $x = sp + 40|0;
 $y = sp + 32|0;
 $xrel = sp + 24|0;
 $yrel = sp + 16|0;
 $13 = sp + 348|0;
 $14 = sp + 344|0;
 $15 = sp + 340|0;
 $16 = sp + 336|0;
 $17 = sp + 332|0;
 $18 = sp + 320|0;
 $x1 = sp + 8|0;
 $y2 = sp;
 $19 = sp + 316|0;
 $20 = sp + 312|0;
 $21 = sp + 308|0;
 $22 = sp + 304|0;
 $23 = sp + 292|0;
 $24 = sp + 288|0;
 $25 = sp + 284|0;
 $26 = sp + 272|0;
 $27 = sp + 260|0;
 $28 = sp + 256|0;
 $29 = sp + 252|0;
 $30 = sp + 248|0;
 $31 = sp + 244|0;
 $32 = sp + 240|0;
 $33 = sp + 236|0;
 $34 = sp + 224|0;
 $35 = sp + 212|0;
 $36 = sp + 200|0;
 $37 = sp + 196|0;
 $38 = sp + 184|0;
 $proxy = sp + 176|0;
 $39 = sp + 160|0;
 $40 = sp + 156|0;
 $41 = sp + 144|0;
 $42 = sp + 132|0;
 $43 = sp + 120|0;
 $44 = sp + 108|0;
 $45 = sp + 96|0;
 HEAP32[$vargs>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $46 = ((($vargs)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$46>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $47 = ((($vargs)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$47>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 invoke_vii(3936,($vargs|0),4); //@line 154 "modules/event/sdl/Event.cpp"
 $48 = __THREW__; __THREW__ = 0;
 $49 = $48&1;
 L1: do {
  if ($49) {
   label = 4;
  } else {
   HEAP32[$scancode>>2] = 0; //@line 159 "modules/event/sdl/Event.cpp"
   $50 = HEAP32[$e>>2]|0; //@line 174 "modules/event/sdl/Event.cpp"
   L3: do {
    switch ($50|0) {
    case 768:  {
     $55 = ((($e)) + 13|0); //@line 177 "modules/event/sdl/Event.cpp"
     $56 = HEAP8[$55>>0]|0; //@line 177 "modules/event/sdl/Event.cpp"
     $57 = ($56<<24>>24)==(0); //@line 177 "modules/event/sdl/Event.cpp"
     if (!($57)) {
      $58 = HEAP32[(625636)>>2]|0; //@line 98 "./common/Module.h"
      $59 = ($58|0)==(0|0); //@line 180 "modules/event/sdl/Event.cpp"
      if (!($59)) {
       $60 = HEAP32[$58>>2]|0; //@line 180 "modules/event/sdl/Event.cpp"
       $61 = ((($60)) + 28|0); //@line 180 "modules/event/sdl/Event.cpp"
       $62 = HEAP32[$61>>2]|0; //@line 180 "modules/event/sdl/Event.cpp"
       __THREW__ = 0;
       $63 = (invoke_ii($62|0,($58|0))|0); //@line 180 "modules/event/sdl/Event.cpp"
       $64 = __THREW__; __THREW__ = 0;
       $65 = $64&1;
       if ($65) {
        label = 4;
        break L1;
       }
       if (!($63)) {
        $msg$1 = 0;
        break L3;
       }
      }
     }
     $66 = ((($e)) + 16|0); //@line 184 "modules/event/sdl/Event.cpp"
     $67 = HEAP32[(632704)>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $68 = ($67|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     if ($68) {
      $key$0 = 0;
     } else {
      $69 = ((($66)) + 4|0); //@line 184 "modules/event/sdl/Event.cpp"
      $70 = HEAP32[$69>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
      $$0$ph6$i$i$i = (632704);$$01$ph5$i$i$i = $67;
      L12: while(1) {
       $$013$i$i$i = $$01$ph5$i$i$i;
       while(1) {
        $71 = ((($$013$i$i$i)) + 16|0); //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
        $72 = HEAP32[$71>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
        $73 = ($72|0)<($70|0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
        if (!($73)) {
         $$013$i$i$i$lcssa = $$013$i$i$i;
         break;
        }
        $76 = ((($$013$i$i$i)) + 4|0); //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        $77 = HEAP32[$76>>2]|0; //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        $78 = ($77|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        if ($78) {
         $$0$ph$lcssa$i$i$i = $$0$ph6$i$i$i;
         break L12;
        } else {
         $$013$i$i$i = $77;
        }
       }
       $74 = HEAP32[$$013$i$i$i$lcssa>>2]|0; //@line 2105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
       $75 = ($74|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
       if ($75) {
        $$0$ph$lcssa$i$i$i = $$013$i$i$i$lcssa;
        break;
       } else {
        $$0$ph6$i$i$i = $$013$i$i$i$lcssa;$$01$ph5$i$i$i = $74;
       }
      }
      $79 = ($$0$ph$lcssa$i$i$i|0)==((632704)|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      if ($79) {
       $key$0 = 0;
      } else {
       $80 = ((($$0$ph$lcssa$i$i$i)) + 16|0); //@line 463 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
       $81 = HEAP32[$80>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
       $82 = ($70|0)<($81|0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
       if ($82) {
        $key$0 = 0;
       } else {
        $83 = ((($$0$ph$lcssa$i$i$i)) + 20|0); //@line 186 "modules/event/sdl/Event.cpp"
        $84 = HEAP32[$83>>2]|0; //@line 186 "modules/event/sdl/Event.cpp"
        $key$0 = $84;
       }
      }
     }
     __THREW__ = 0;
     $85 = (invoke_iii(3937,($key$0|0),($txt|0))|0); //@line 188 "modules/event/sdl/Event.cpp"
     $86 = __THREW__; __THREW__ = 0;
     $87 = $86&1;
     if ($87) {
      label = 4;
      break L1;
     }
     if (!($85)) {
      HEAP32[$txt>>2] = 442424; //@line 189 "modules/event/sdl/Event.cpp"
     }
     $88 = HEAP32[$66>>2]|0; //@line 191 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     (invoke_iii(3938,($88|0),($scancode|0))|0); //@line 191 "modules/event/sdl/Event.cpp"
     $89 = __THREW__; __THREW__ = 0;
     $90 = $89&1;
     if ($90) {
      label = 4;
      break L1;
     }
     $91 = HEAP32[$scancode>>2]|0; //@line 192 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     $92 = (invoke_iii(3939,($91|0),($txt2|0))|0); //@line 192 "modules/event/sdl/Event.cpp"
     $93 = __THREW__; __THREW__ = 0;
     $94 = $93&1;
     if ($94) {
      label = 4;
      break L1;
     }
     if (!($92)) {
      HEAP32[$txt2>>2] = 442424; //@line 193 "modules/event/sdl/Event.cpp"
     }
     __THREW__ = 0;
     $95 = (invoke_ii(3348,32)|0); //@line 195 "modules/event/sdl/Event.cpp"
     $96 = __THREW__; __THREW__ = 0;
     $97 = $96&1;
     if ($97) {
      label = 4;
      break L1;
     }
     $98 = HEAP32[$txt>>2]|0; //@line 195 "modules/event/sdl/Event.cpp"
     $99 = (_strlen($98)|0); //@line 195 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_viii(3940,($95|0),($98|0),($99|0)); //@line 195 "modules/event/sdl/Event.cpp"
     $100 = __THREW__; __THREW__ = 0;
     $101 = $100&1;
     if ($101) {
      $193 = ___cxa_find_matching_catch_2()|0;
      $194 = tempRet0;
      __ZdlPv($95); //@line 195 "modules/event/sdl/Event.cpp"
      $$13 = $193;$$1316 = $194;
      break L1;
     }
     HEAP32[$0>>2] = $95; //@line 93 "./common/Object.h"
     $102 = HEAP32[$95>>2]|0; //@line 95 "./common/Object.h"
     $103 = ((($102)) + 8|0); //@line 95 "./common/Object.h"
     $104 = HEAP32[$103>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($104|0,($95|0)); //@line 95 "./common/Object.h"
     $105 = __THREW__; __THREW__ = 0;
     $106 = $105&1;
     if ($106) {
      label = 4;
      break L1;
     }
     $107 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $108 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $109 = HEAP32[$108>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $110 = ($107>>>0)<($109>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($110) {
      HEAP32[$107>>2] = $95; //@line 99 "./common/Object.h"
      $111 = HEAP32[$95>>2]|0; //@line 101 "./common/Object.h"
      $112 = ((($111)) + 8|0); //@line 101 "./common/Object.h"
      $113 = HEAP32[$112>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($113|0,($95|0)); //@line 101 "./common/Object.h"
      $114 = __THREW__; __THREW__ = 0;
      $115 = $114&1;
      if ($115) {
       label = 58;
      } else {
       $$pre$i36 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $116 = ((($$pre$i36)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $116; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($vargs|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $117 = __THREW__; __THREW__ = 0;
      $118 = $117&1;
      if ($118) {
       label = 58;
      }
     }
     if ((label|0) == 58) {
      $195 = ___cxa_find_matching_catch_2()|0;
      $196 = tempRet0;
      $197 = HEAP32[$95>>2]|0; //@line 106 "./common/Object.h"
      $198 = ((($197)) + 12|0); //@line 106 "./common/Object.h"
      $199 = HEAP32[$198>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($199|0,($95|0)); //@line 106 "./common/Object.h"
      $200 = __THREW__; __THREW__ = 0;
      $201 = $200&1;
      if (!($201)) {
       $$13 = $195;$$1316 = $196;
       break L1;
      }
      $202 = ___cxa_find_matching_catch_3(0|0)|0;
      $203 = tempRet0;
      ___clang_call_terminate($202); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $119 = HEAP32[$95>>2]|0; //@line 106 "./common/Object.h"
     $120 = ((($119)) + 12|0); //@line 106 "./common/Object.h"
     $121 = HEAP32[$120>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($121|0,($95|0)); //@line 106 "./common/Object.h"
     $122 = __THREW__; __THREW__ = 0;
     $123 = $122&1;
     if ($123) {
      $124 = ___cxa_find_matching_catch_3(0|0)|0;
      $125 = tempRet0;
      ___clang_call_terminate($124); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $126 = (invoke_ii(3348,32)|0); //@line 196 "modules/event/sdl/Event.cpp"
     $127 = __THREW__; __THREW__ = 0;
     $128 = $127&1;
     if ($128) {
      label = 4;
      break L1;
     }
     $129 = HEAP32[$txt2>>2]|0; //@line 196 "modules/event/sdl/Event.cpp"
     $130 = (_strlen($129)|0); //@line 196 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_viii(3940,($126|0),($129|0),($130|0)); //@line 196 "modules/event/sdl/Event.cpp"
     $131 = __THREW__; __THREW__ = 0;
     $132 = $131&1;
     if ($132) {
      $204 = ___cxa_find_matching_catch_2()|0;
      $205 = tempRet0;
      __ZdlPv($126); //@line 196 "modules/event/sdl/Event.cpp"
      $$13 = $204;$$1316 = $205;
      break L1;
     }
     HEAP32[$1>>2] = $126; //@line 93 "./common/Object.h"
     $133 = HEAP32[$126>>2]|0; //@line 95 "./common/Object.h"
     $134 = ((($133)) + 8|0); //@line 95 "./common/Object.h"
     $135 = HEAP32[$134>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($135|0,($126|0)); //@line 95 "./common/Object.h"
     $136 = __THREW__; __THREW__ = 0;
     $137 = $136&1;
     if ($137) {
      label = 4;
      break L1;
     }
     $138 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $139 = HEAP32[$108>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $140 = ($138>>>0)<($139>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     do {
      if ($140) {
       HEAP32[$138>>2] = $126; //@line 99 "./common/Object.h"
       $141 = HEAP32[$126>>2]|0; //@line 101 "./common/Object.h"
       $142 = ((($141)) + 8|0); //@line 101 "./common/Object.h"
       $143 = HEAP32[$142>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($143|0,($126|0)); //@line 101 "./common/Object.h"
       $144 = __THREW__; __THREW__ = 0;
       $145 = $144&1;
       if ($145) {
        label = 61;
        break;
       }
       $$pre$i40 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $146 = ((($$pre$i40)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $146; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      } else {
       __THREW__ = 0;
       invoke_vii(3507,($vargs|0),($1|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $147 = __THREW__; __THREW__ = 0;
       $148 = $147&1;
       if ($148) {
        label = 61;
       }
      }
     } while(0);
     if ((label|0) == 61) {
      $206 = ___cxa_find_matching_catch_2()|0;
      $207 = tempRet0;
      $208 = HEAP32[$126>>2]|0; //@line 106 "./common/Object.h"
      $209 = ((($208)) + 12|0); //@line 106 "./common/Object.h"
      $210 = HEAP32[$209>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($210|0,($126|0)); //@line 106 "./common/Object.h"
      $211 = __THREW__; __THREW__ = 0;
      $212 = $211&1;
      if (!($212)) {
       $$13 = $206;$$1316 = $207;
       break L1;
      }
      $213 = ___cxa_find_matching_catch_3(0|0)|0;
      $214 = tempRet0;
      ___clang_call_terminate($213); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $149 = HEAP32[$126>>2]|0; //@line 106 "./common/Object.h"
     $150 = ((($149)) + 12|0); //@line 106 "./common/Object.h"
     $151 = HEAP32[$150>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($151|0,($126|0)); //@line 106 "./common/Object.h"
     $152 = __THREW__; __THREW__ = 0;
     $153 = $152&1;
     if ($153) {
      $154 = ___cxa_find_matching_catch_3(0|0)|0;
      $155 = tempRet0;
      ___clang_call_terminate($154); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $156 = (invoke_ii(3348,32)|0); //@line 197 "modules/event/sdl/Event.cpp"
     $157 = __THREW__; __THREW__ = 0;
     $158 = $157&1;
     if ($158) {
      label = 4;
      break L1;
     }
     $159 = HEAP8[$55>>0]|0; //@line 197 "modules/event/sdl/Event.cpp"
     $160 = ($159<<24>>24)!=(0); //@line 197 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vii(3941,($156|0),($160|0)); //@line 197 "modules/event/sdl/Event.cpp"
     $161 = __THREW__; __THREW__ = 0;
     $162 = $161&1;
     if ($162) {
      $215 = ___cxa_find_matching_catch_2()|0;
      $216 = tempRet0;
      __ZdlPv($156); //@line 197 "modules/event/sdl/Event.cpp"
      $$13 = $215;$$1316 = $216;
      break L1;
     }
     HEAP32[$2>>2] = $156; //@line 93 "./common/Object.h"
     $163 = HEAP32[$156>>2]|0; //@line 95 "./common/Object.h"
     $164 = ((($163)) + 8|0); //@line 95 "./common/Object.h"
     $165 = HEAP32[$164>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($165|0,($156|0)); //@line 95 "./common/Object.h"
     $166 = __THREW__; __THREW__ = 0;
     $167 = $166&1;
     if ($167) {
      label = 4;
      break L1;
     }
     $168 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $169 = HEAP32[$108>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $170 = ($168>>>0)<($169>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     do {
      if ($170) {
       HEAP32[$168>>2] = $156; //@line 99 "./common/Object.h"
       $171 = HEAP32[$156>>2]|0; //@line 101 "./common/Object.h"
       $172 = ((($171)) + 8|0); //@line 101 "./common/Object.h"
       $173 = HEAP32[$172>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($173|0,($156|0)); //@line 101 "./common/Object.h"
       $174 = __THREW__; __THREW__ = 0;
       $175 = $174&1;
       if ($175) {
        label = 64;
        break;
       }
       $$pre$i48 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $176 = ((($$pre$i48)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $176; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      } else {
       __THREW__ = 0;
       invoke_vii(3507,($vargs|0),($2|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $177 = __THREW__; __THREW__ = 0;
       $178 = $177&1;
       if ($178) {
        label = 64;
       }
      }
     } while(0);
     if ((label|0) == 64) {
      $217 = ___cxa_find_matching_catch_2()|0;
      $218 = tempRet0;
      $219 = HEAP32[$156>>2]|0; //@line 106 "./common/Object.h"
      $220 = ((($219)) + 12|0); //@line 106 "./common/Object.h"
      $221 = HEAP32[$220>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($221|0,($156|0)); //@line 106 "./common/Object.h"
      $222 = __THREW__; __THREW__ = 0;
      $223 = $222&1;
      if (!($223)) {
       $$13 = $217;$$1316 = $218;
       break L1;
      }
      $224 = ___cxa_find_matching_catch_3(0|0)|0;
      $225 = tempRet0;
      ___clang_call_terminate($224); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $179 = HEAP32[$156>>2]|0; //@line 106 "./common/Object.h"
     $180 = ((($179)) + 12|0); //@line 106 "./common/Object.h"
     $181 = HEAP32[$180>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($181|0,($156|0)); //@line 106 "./common/Object.h"
     $182 = __THREW__; __THREW__ = 0;
     $183 = $182&1;
     if ($183) {
      $184 = ___cxa_find_matching_catch_3(0|0)|0;
      $185 = tempRet0;
      ___clang_call_terminate($184); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $186 = (invoke_ii(3348,32)|0); //@line 198 "modules/event/sdl/Event.cpp"
     $187 = __THREW__; __THREW__ = 0;
     $188 = $187&1;
     if ($188) {
      label = 4;
      break L1;
     }
     __THREW__ = 0;
     invoke_viii(3460,($3|0),(365517|0),10); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $189 = __THREW__; __THREW__ = 0;
     $190 = $189&1;
     do {
      if ($190) {
       $226 = ___cxa_find_matching_catch_2()|0;
       $227 = tempRet0;
       $$0 = $226;$$02 = $227;
      } else {
       __THREW__ = 0;
       invoke_viii(3508,($186|0),($3|0),($vargs|0)); //@line 198 "modules/event/sdl/Event.cpp"
       $191 = __THREW__; __THREW__ = 0;
       $192 = $191&1;
       if ($192) {
        $228 = ___cxa_find_matching_catch_2()|0;
        $229 = tempRet0;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 198 "modules/event/sdl/Event.cpp"
        $$0 = $228;$$02 = $229;
        break;
       } else {
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 198 "modules/event/sdl/Event.cpp"
        $msg$1 = $186;
        break L3;
       }
      }
     } while(0);
     __ZdlPv($186); //@line 198 "modules/event/sdl/Event.cpp"
     $$13 = $$0;$$1316 = $$02;
     break L1;
     break;
    }
    case 769:  {
     $230 = ((($e)) + 16|0); //@line 201 "modules/event/sdl/Event.cpp"
     $231 = HEAP32[(632704)>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $232 = ($231|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     if ($232) {
      $key$1 = 0;
     } else {
      $233 = ((($230)) + 4|0); //@line 201 "modules/event/sdl/Event.cpp"
      $234 = HEAP32[$233>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
      $$0$ph6$i$i$i59 = (632704);$$01$ph5$i$i$i60 = $231;
      L90: while(1) {
       $$013$i$i$i62 = $$01$ph5$i$i$i60;
       while(1) {
        $235 = ((($$013$i$i$i62)) + 16|0); //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
        $236 = HEAP32[$235>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
        $237 = ($236|0)<($234|0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
        if (!($237)) {
         $$013$i$i$i62$lcssa = $$013$i$i$i62;
         break;
        }
        $240 = ((($$013$i$i$i62)) + 4|0); //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        $241 = HEAP32[$240>>2]|0; //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        $242 = ($241|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        if ($242) {
         $$0$ph$lcssa$i$i$i64 = $$0$ph6$i$i$i59;
         break L90;
        } else {
         $$013$i$i$i62 = $241;
        }
       }
       $238 = HEAP32[$$013$i$i$i62$lcssa>>2]|0; //@line 2105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
       $239 = ($238|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
       if ($239) {
        $$0$ph$lcssa$i$i$i64 = $$013$i$i$i62$lcssa;
        break;
       } else {
        $$0$ph6$i$i$i59 = $$013$i$i$i62$lcssa;$$01$ph5$i$i$i60 = $238;
       }
      }
      $243 = ($$0$ph$lcssa$i$i$i64|0)==((632704)|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      if ($243) {
       $key$1 = 0;
      } else {
       $244 = ((($$0$ph$lcssa$i$i$i64)) + 16|0); //@line 463 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
       $245 = HEAP32[$244>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
       $246 = ($234|0)<($245|0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
       if ($246) {
        $key$1 = 0;
       } else {
        $247 = ((($$0$ph$lcssa$i$i$i64)) + 20|0); //@line 203 "modules/event/sdl/Event.cpp"
        $248 = HEAP32[$247>>2]|0; //@line 203 "modules/event/sdl/Event.cpp"
        $key$1 = $248;
       }
      }
     }
     __THREW__ = 0;
     $249 = (invoke_iii(3937,($key$1|0),($txt|0))|0); //@line 205 "modules/event/sdl/Event.cpp"
     $250 = __THREW__; __THREW__ = 0;
     $251 = $250&1;
     if ($251) {
      label = 4;
      break L1;
     }
     if (!($249)) {
      HEAP32[$txt>>2] = 442424; //@line 206 "modules/event/sdl/Event.cpp"
     }
     $252 = HEAP32[$230>>2]|0; //@line 208 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     (invoke_iii(3938,($252|0),($scancode|0))|0); //@line 208 "modules/event/sdl/Event.cpp"
     $253 = __THREW__; __THREW__ = 0;
     $254 = $253&1;
     if ($254) {
      label = 4;
      break L1;
     }
     $255 = HEAP32[$scancode>>2]|0; //@line 209 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     $256 = (invoke_iii(3939,($255|0),($txt2|0))|0); //@line 209 "modules/event/sdl/Event.cpp"
     $257 = __THREW__; __THREW__ = 0;
     $258 = $257&1;
     if ($258) {
      label = 4;
      break L1;
     }
     if (!($256)) {
      HEAP32[$txt2>>2] = 442424; //@line 210 "modules/event/sdl/Event.cpp"
     }
     __THREW__ = 0;
     $259 = (invoke_ii(3348,32)|0); //@line 212 "modules/event/sdl/Event.cpp"
     $260 = __THREW__; __THREW__ = 0;
     $261 = $260&1;
     if ($261) {
      label = 4;
      break L1;
     }
     $262 = HEAP32[$txt>>2]|0; //@line 212 "modules/event/sdl/Event.cpp"
     $263 = (_strlen($262)|0); //@line 212 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_viii(3940,($259|0),($262|0),($263|0)); //@line 212 "modules/event/sdl/Event.cpp"
     $264 = __THREW__; __THREW__ = 0;
     $265 = $264&1;
     if ($265) {
      $327 = ___cxa_find_matching_catch_2()|0;
      $328 = tempRet0;
      __ZdlPv($259); //@line 212 "modules/event/sdl/Event.cpp"
      $$13 = $327;$$1316 = $328;
      break L1;
     }
     HEAP32[$4>>2] = $259; //@line 93 "./common/Object.h"
     $266 = HEAP32[$259>>2]|0; //@line 95 "./common/Object.h"
     $267 = ((($266)) + 8|0); //@line 95 "./common/Object.h"
     $268 = HEAP32[$267>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($268|0,($259|0)); //@line 95 "./common/Object.h"
     $269 = __THREW__; __THREW__ = 0;
     $270 = $269&1;
     if ($270) {
      label = 4;
      break L1;
     }
     $271 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $272 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $273 = HEAP32[$272>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $274 = ($271>>>0)<($273>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($274) {
      HEAP32[$271>>2] = $259; //@line 99 "./common/Object.h"
      $275 = HEAP32[$259>>2]|0; //@line 101 "./common/Object.h"
      $276 = ((($275)) + 8|0); //@line 101 "./common/Object.h"
      $277 = HEAP32[$276>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($277|0,($259|0)); //@line 101 "./common/Object.h"
      $278 = __THREW__; __THREW__ = 0;
      $279 = $278&1;
      if ($279) {
       label = 108;
      } else {
       $$pre$i71 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $280 = ((($$pre$i71)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $280; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($vargs|0),($4|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $281 = __THREW__; __THREW__ = 0;
      $282 = $281&1;
      if ($282) {
       label = 108;
      }
     }
     if ((label|0) == 108) {
      $329 = ___cxa_find_matching_catch_2()|0;
      $330 = tempRet0;
      $331 = HEAP32[$259>>2]|0; //@line 106 "./common/Object.h"
      $332 = ((($331)) + 12|0); //@line 106 "./common/Object.h"
      $333 = HEAP32[$332>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($333|0,($259|0)); //@line 106 "./common/Object.h"
      $334 = __THREW__; __THREW__ = 0;
      $335 = $334&1;
      if (!($335)) {
       $$13 = $329;$$1316 = $330;
       break L1;
      }
      $336 = ___cxa_find_matching_catch_3(0|0)|0;
      $337 = tempRet0;
      ___clang_call_terminate($336); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $283 = HEAP32[$259>>2]|0; //@line 106 "./common/Object.h"
     $284 = ((($283)) + 12|0); //@line 106 "./common/Object.h"
     $285 = HEAP32[$284>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($285|0,($259|0)); //@line 106 "./common/Object.h"
     $286 = __THREW__; __THREW__ = 0;
     $287 = $286&1;
     if ($287) {
      $288 = ___cxa_find_matching_catch_3(0|0)|0;
      $289 = tempRet0;
      ___clang_call_terminate($288); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $290 = (invoke_ii(3348,32)|0); //@line 213 "modules/event/sdl/Event.cpp"
     $291 = __THREW__; __THREW__ = 0;
     $292 = $291&1;
     if ($292) {
      label = 4;
      break L1;
     }
     $293 = HEAP32[$txt2>>2]|0; //@line 213 "modules/event/sdl/Event.cpp"
     $294 = (_strlen($293)|0); //@line 213 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_viii(3940,($290|0),($293|0),($294|0)); //@line 213 "modules/event/sdl/Event.cpp"
     $295 = __THREW__; __THREW__ = 0;
     $296 = $295&1;
     if ($296) {
      $338 = ___cxa_find_matching_catch_2()|0;
      $339 = tempRet0;
      __ZdlPv($290); //@line 213 "modules/event/sdl/Event.cpp"
      $$13 = $338;$$1316 = $339;
      break L1;
     }
     HEAP32[$5>>2] = $290; //@line 93 "./common/Object.h"
     $297 = HEAP32[$290>>2]|0; //@line 95 "./common/Object.h"
     $298 = ((($297)) + 8|0); //@line 95 "./common/Object.h"
     $299 = HEAP32[$298>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($299|0,($290|0)); //@line 95 "./common/Object.h"
     $300 = __THREW__; __THREW__ = 0;
     $301 = $300&1;
     if ($301) {
      label = 4;
      break L1;
     }
     $302 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $303 = HEAP32[$272>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $304 = ($302>>>0)<($303>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($304) {
      HEAP32[$302>>2] = $290; //@line 99 "./common/Object.h"
      $305 = HEAP32[$290>>2]|0; //@line 101 "./common/Object.h"
      $306 = ((($305)) + 8|0); //@line 101 "./common/Object.h"
      $307 = HEAP32[$306>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($307|0,($290|0)); //@line 101 "./common/Object.h"
      $308 = __THREW__; __THREW__ = 0;
      $309 = $308&1;
      if ($309) {
       label = 111;
      } else {
       $$pre$i79 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $310 = ((($$pre$i79)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $310; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($vargs|0),($5|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $311 = __THREW__; __THREW__ = 0;
      $312 = $311&1;
      if ($312) {
       label = 111;
      }
     }
     if ((label|0) == 111) {
      $340 = ___cxa_find_matching_catch_2()|0;
      $341 = tempRet0;
      $342 = HEAP32[$290>>2]|0; //@line 106 "./common/Object.h"
      $343 = ((($342)) + 12|0); //@line 106 "./common/Object.h"
      $344 = HEAP32[$343>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($344|0,($290|0)); //@line 106 "./common/Object.h"
      $345 = __THREW__; __THREW__ = 0;
      $346 = $345&1;
      if (!($346)) {
       $$13 = $340;$$1316 = $341;
       break L1;
      }
      $347 = ___cxa_find_matching_catch_3(0|0)|0;
      $348 = tempRet0;
      ___clang_call_terminate($347); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $313 = HEAP32[$290>>2]|0; //@line 106 "./common/Object.h"
     $314 = ((($313)) + 12|0); //@line 106 "./common/Object.h"
     $315 = HEAP32[$314>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($315|0,($290|0)); //@line 106 "./common/Object.h"
     $316 = __THREW__; __THREW__ = 0;
     $317 = $316&1;
     if ($317) {
      $318 = ___cxa_find_matching_catch_3(0|0)|0;
      $319 = tempRet0;
      ___clang_call_terminate($318); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $320 = (invoke_ii(3348,32)|0); //@line 214 "modules/event/sdl/Event.cpp"
     $321 = __THREW__; __THREW__ = 0;
     $322 = $321&1;
     if ($322) {
      label = 4;
      break L1;
     }
     __THREW__ = 0;
     invoke_viii(3460,($6|0),(365528|0),11); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $323 = __THREW__; __THREW__ = 0;
     $324 = $323&1;
     do {
      if ($324) {
       $349 = ___cxa_find_matching_catch_2()|0;
       $350 = tempRet0;
       $$1 = $349;$$14 = $350;
      } else {
       __THREW__ = 0;
       invoke_viii(3508,($320|0),($6|0),($vargs|0)); //@line 214 "modules/event/sdl/Event.cpp"
       $325 = __THREW__; __THREW__ = 0;
       $326 = $325&1;
       if ($326) {
        $351 = ___cxa_find_matching_catch_2()|0;
        $352 = tempRet0;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 214 "modules/event/sdl/Event.cpp"
        $$1 = $351;$$14 = $352;
        break;
       } else {
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 214 "modules/event/sdl/Event.cpp"
        $msg$1 = $320;
        break L3;
       }
      }
     } while(0);
     __ZdlPv($320); //@line 214 "modules/event/sdl/Event.cpp"
     $$13 = $$1;$$1316 = $$14;
     break L1;
     break;
    }
    case 771:  {
     $353 = ((($e)) + 12|0); //@line 218 "modules/event/sdl/Event.cpp"
     HEAP32[$txt>>2] = $353; //@line 217 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     $354 = (invoke_ii(3348,32)|0); //@line 218 "modules/event/sdl/Event.cpp"
     $355 = __THREW__; __THREW__ = 0;
     $356 = $355&1;
     if ($356) {
      label = 4;
      break L1;
     }
     $357 = (_strlen($353)|0); //@line 218 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_viii(3940,($354|0),($353|0),($357|0)); //@line 218 "modules/event/sdl/Event.cpp"
     $358 = __THREW__; __THREW__ = 0;
     $359 = $358&1;
     if ($359) {
      $391 = ___cxa_find_matching_catch_2()|0;
      $392 = tempRet0;
      __ZdlPv($354); //@line 218 "modules/event/sdl/Event.cpp"
      $$13 = $391;$$1316 = $392;
      break L1;
     }
     HEAP32[$7>>2] = $354; //@line 93 "./common/Object.h"
     $360 = HEAP32[$354>>2]|0; //@line 95 "./common/Object.h"
     $361 = ((($360)) + 8|0); //@line 95 "./common/Object.h"
     $362 = HEAP32[$361>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($362|0,($354|0)); //@line 95 "./common/Object.h"
     $363 = __THREW__; __THREW__ = 0;
     $364 = $363&1;
     if ($364) {
      label = 4;
      break L1;
     }
     $365 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $366 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $367 = HEAP32[$366>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $368 = ($365>>>0)<($367>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($368) {
      HEAP32[$365>>2] = $354; //@line 99 "./common/Object.h"
      $369 = HEAP32[$354>>2]|0; //@line 101 "./common/Object.h"
      $370 = ((($369)) + 8|0); //@line 101 "./common/Object.h"
      $371 = HEAP32[$370>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($371|0,($354|0)); //@line 101 "./common/Object.h"
      $372 = __THREW__; __THREW__ = 0;
      $373 = $372&1;
      if ($373) {
       label = 130;
      } else {
       $$pre$i91 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $374 = ((($$pre$i91)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $374; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($vargs|0),($7|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $375 = __THREW__; __THREW__ = 0;
      $376 = $375&1;
      if ($376) {
       label = 130;
      }
     }
     if ((label|0) == 130) {
      $393 = ___cxa_find_matching_catch_2()|0;
      $394 = tempRet0;
      $395 = HEAP32[$354>>2]|0; //@line 106 "./common/Object.h"
      $396 = ((($395)) + 12|0); //@line 106 "./common/Object.h"
      $397 = HEAP32[$396>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($397|0,($354|0)); //@line 106 "./common/Object.h"
      $398 = __THREW__; __THREW__ = 0;
      $399 = $398&1;
      if (!($399)) {
       $$13 = $393;$$1316 = $394;
       break L1;
      }
      $400 = ___cxa_find_matching_catch_3(0|0)|0;
      $401 = tempRet0;
      ___clang_call_terminate($400); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $377 = HEAP32[$354>>2]|0; //@line 106 "./common/Object.h"
     $378 = ((($377)) + 12|0); //@line 106 "./common/Object.h"
     $379 = HEAP32[$378>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($379|0,($354|0)); //@line 106 "./common/Object.h"
     $380 = __THREW__; __THREW__ = 0;
     $381 = $380&1;
     if ($381) {
      $382 = ___cxa_find_matching_catch_3(0|0)|0;
      $383 = tempRet0;
      ___clang_call_terminate($382); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $384 = (invoke_ii(3348,32)|0); //@line 219 "modules/event/sdl/Event.cpp"
     $385 = __THREW__; __THREW__ = 0;
     $386 = $385&1;
     if ($386) {
      label = 4;
      break L1;
     }
     __THREW__ = 0;
     invoke_viii(3460,($8|0),(365540|0),9); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $387 = __THREW__; __THREW__ = 0;
     $388 = $387&1;
     do {
      if ($388) {
       $402 = ___cxa_find_matching_catch_2()|0;
       $403 = tempRet0;
       $$2 = $402;$$25 = $403;
      } else {
       __THREW__ = 0;
       invoke_viii(3508,($384|0),($8|0),($vargs|0)); //@line 219 "modules/event/sdl/Event.cpp"
       $389 = __THREW__; __THREW__ = 0;
       $390 = $389&1;
       if ($390) {
        $404 = ___cxa_find_matching_catch_2()|0;
        $405 = tempRet0;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8); //@line 219 "modules/event/sdl/Event.cpp"
        $$2 = $404;$$25 = $405;
        break;
       } else {
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8); //@line 219 "modules/event/sdl/Event.cpp"
        $msg$1 = $384;
        break L3;
       }
      }
     } while(0);
     __ZdlPv($384); //@line 219 "modules/event/sdl/Event.cpp"
     $$13 = $$2;$$1316 = $$25;
     break L1;
     break;
    }
    case 770:  {
     $406 = ((($e)) + 12|0); //@line 223 "modules/event/sdl/Event.cpp"
     HEAP32[$txt>>2] = $406; //@line 222 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     $407 = (invoke_ii(3348,32)|0); //@line 223 "modules/event/sdl/Event.cpp"
     $408 = __THREW__; __THREW__ = 0;
     $409 = $408&1;
     if ($409) {
      label = 4;
      break L1;
     }
     $410 = (_strlen($406)|0); //@line 223 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_viii(3940,($407|0),($406|0),($410|0)); //@line 223 "modules/event/sdl/Event.cpp"
     $411 = __THREW__; __THREW__ = 0;
     $412 = $411&1;
     if ($412) {
      $506 = ___cxa_find_matching_catch_2()|0;
      $507 = tempRet0;
      __ZdlPv($407); //@line 223 "modules/event/sdl/Event.cpp"
      $$13 = $506;$$1316 = $507;
      break L1;
     }
     HEAP32[$9>>2] = $407; //@line 93 "./common/Object.h"
     $413 = HEAP32[$407>>2]|0; //@line 95 "./common/Object.h"
     $414 = ((($413)) + 8|0); //@line 95 "./common/Object.h"
     $415 = HEAP32[$414>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($415|0,($407|0)); //@line 95 "./common/Object.h"
     $416 = __THREW__; __THREW__ = 0;
     $417 = $416&1;
     if ($417) {
      label = 4;
      break L1;
     }
     $418 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $419 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $420 = HEAP32[$419>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $421 = ($418>>>0)<($420>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($421) {
      HEAP32[$418>>2] = $407; //@line 99 "./common/Object.h"
      $422 = HEAP32[$407>>2]|0; //@line 101 "./common/Object.h"
      $423 = ((($422)) + 8|0); //@line 101 "./common/Object.h"
      $424 = HEAP32[$423>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($424|0,($407|0)); //@line 101 "./common/Object.h"
      $425 = __THREW__; __THREW__ = 0;
      $426 = $425&1;
      if ($426) {
       label = 167;
      } else {
       $$pre$i102 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $427 = ((($$pre$i102)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $427; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($vargs|0),($9|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $428 = __THREW__; __THREW__ = 0;
      $429 = $428&1;
      if ($429) {
       label = 167;
      }
     }
     if ((label|0) == 167) {
      $508 = ___cxa_find_matching_catch_2()|0;
      $509 = tempRet0;
      $510 = HEAP32[$407>>2]|0; //@line 106 "./common/Object.h"
      $511 = ((($510)) + 12|0); //@line 106 "./common/Object.h"
      $512 = HEAP32[$511>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($512|0,($407|0)); //@line 106 "./common/Object.h"
      $513 = __THREW__; __THREW__ = 0;
      $514 = $513&1;
      if (!($514)) {
       $$13 = $508;$$1316 = $509;
       break L1;
      }
      $515 = ___cxa_find_matching_catch_3(0|0)|0;
      $516 = tempRet0;
      ___clang_call_terminate($515); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $430 = HEAP32[$407>>2]|0; //@line 106 "./common/Object.h"
     $431 = ((($430)) + 12|0); //@line 106 "./common/Object.h"
     $432 = HEAP32[$431>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($432|0,($407|0)); //@line 106 "./common/Object.h"
     $433 = __THREW__; __THREW__ = 0;
     $434 = $433&1;
     if ($434) {
      $435 = ___cxa_find_matching_catch_3(0|0)|0;
      $436 = tempRet0;
      ___clang_call_terminate($435); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $437 = (invoke_ii(3348,32)|0); //@line 224 "modules/event/sdl/Event.cpp"
     $438 = __THREW__; __THREW__ = 0;
     $439 = $438&1;
     if ($439) {
      label = 4;
      break L1;
     }
     $440 = ((($e)) + 44|0); //@line 224 "modules/event/sdl/Event.cpp"
     $441 = HEAP32[$440>>2]|0; //@line 224 "modules/event/sdl/Event.cpp"
     $442 = (+($441|0)); //@line 224 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vid(3942,($437|0),(+$442)); //@line 224 "modules/event/sdl/Event.cpp"
     $443 = __THREW__; __THREW__ = 0;
     $444 = $443&1;
     if ($444) {
      $517 = ___cxa_find_matching_catch_2()|0;
      $518 = tempRet0;
      __ZdlPv($437); //@line 224 "modules/event/sdl/Event.cpp"
      $$13 = $517;$$1316 = $518;
      break L1;
     }
     HEAP32[$10>>2] = $437; //@line 93 "./common/Object.h"
     $445 = HEAP32[$437>>2]|0; //@line 95 "./common/Object.h"
     $446 = ((($445)) + 8|0); //@line 95 "./common/Object.h"
     $447 = HEAP32[$446>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($447|0,($437|0)); //@line 95 "./common/Object.h"
     $448 = __THREW__; __THREW__ = 0;
     $449 = $448&1;
     if ($449) {
      label = 4;
      break L1;
     }
     $450 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $451 = HEAP32[$419>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $452 = ($450>>>0)<($451>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($452) {
      HEAP32[$450>>2] = $437; //@line 99 "./common/Object.h"
      $453 = HEAP32[$437>>2]|0; //@line 101 "./common/Object.h"
      $454 = ((($453)) + 8|0); //@line 101 "./common/Object.h"
      $455 = HEAP32[$454>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($455|0,($437|0)); //@line 101 "./common/Object.h"
      $456 = __THREW__; __THREW__ = 0;
      $457 = $456&1;
      if ($457) {
       label = 170;
      } else {
       $$pre$i110 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $458 = ((($$pre$i110)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $458; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($vargs|0),($10|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $459 = __THREW__; __THREW__ = 0;
      $460 = $459&1;
      if ($460) {
       label = 170;
      }
     }
     if ((label|0) == 170) {
      $519 = ___cxa_find_matching_catch_2()|0;
      $520 = tempRet0;
      $521 = HEAP32[$437>>2]|0; //@line 106 "./common/Object.h"
      $522 = ((($521)) + 12|0); //@line 106 "./common/Object.h"
      $523 = HEAP32[$522>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($523|0,($437|0)); //@line 106 "./common/Object.h"
      $524 = __THREW__; __THREW__ = 0;
      $525 = $524&1;
      if (!($525)) {
       $$13 = $519;$$1316 = $520;
       break L1;
      }
      $526 = ___cxa_find_matching_catch_3(0|0)|0;
      $527 = tempRet0;
      ___clang_call_terminate($526); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $461 = HEAP32[$437>>2]|0; //@line 106 "./common/Object.h"
     $462 = ((($461)) + 12|0); //@line 106 "./common/Object.h"
     $463 = HEAP32[$462>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($463|0,($437|0)); //@line 106 "./common/Object.h"
     $464 = __THREW__; __THREW__ = 0;
     $465 = $464&1;
     if ($465) {
      $466 = ___cxa_find_matching_catch_3(0|0)|0;
      $467 = tempRet0;
      ___clang_call_terminate($466); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $468 = (invoke_ii(3348,32)|0); //@line 225 "modules/event/sdl/Event.cpp"
     $469 = __THREW__; __THREW__ = 0;
     $470 = $469&1;
     if ($470) {
      label = 4;
      break L1;
     }
     $471 = ((($e)) + 48|0); //@line 225 "modules/event/sdl/Event.cpp"
     $472 = HEAP32[$471>>2]|0; //@line 225 "modules/event/sdl/Event.cpp"
     $473 = (+($472|0)); //@line 225 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vid(3942,($468|0),(+$473)); //@line 225 "modules/event/sdl/Event.cpp"
     $474 = __THREW__; __THREW__ = 0;
     $475 = $474&1;
     if ($475) {
      $528 = ___cxa_find_matching_catch_2()|0;
      $529 = tempRet0;
      __ZdlPv($468); //@line 225 "modules/event/sdl/Event.cpp"
      $$13 = $528;$$1316 = $529;
      break L1;
     }
     HEAP32[$11>>2] = $468; //@line 93 "./common/Object.h"
     $476 = HEAP32[$468>>2]|0; //@line 95 "./common/Object.h"
     $477 = ((($476)) + 8|0); //@line 95 "./common/Object.h"
     $478 = HEAP32[$477>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($478|0,($468|0)); //@line 95 "./common/Object.h"
     $479 = __THREW__; __THREW__ = 0;
     $480 = $479&1;
     if ($480) {
      label = 4;
      break L1;
     }
     $481 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $482 = HEAP32[$419>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $483 = ($481>>>0)<($482>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($483) {
      HEAP32[$481>>2] = $468; //@line 99 "./common/Object.h"
      $484 = HEAP32[$468>>2]|0; //@line 101 "./common/Object.h"
      $485 = ((($484)) + 8|0); //@line 101 "./common/Object.h"
      $486 = HEAP32[$485>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($486|0,($468|0)); //@line 101 "./common/Object.h"
      $487 = __THREW__; __THREW__ = 0;
      $488 = $487&1;
      if ($488) {
       label = 173;
      } else {
       $$pre$i118 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $489 = ((($$pre$i118)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $489; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($vargs|0),($11|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $490 = __THREW__; __THREW__ = 0;
      $491 = $490&1;
      if ($491) {
       label = 173;
      }
     }
     if ((label|0) == 173) {
      $530 = ___cxa_find_matching_catch_2()|0;
      $531 = tempRet0;
      $532 = HEAP32[$468>>2]|0; //@line 106 "./common/Object.h"
      $533 = ((($532)) + 12|0); //@line 106 "./common/Object.h"
      $534 = HEAP32[$533>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($534|0,($468|0)); //@line 106 "./common/Object.h"
      $535 = __THREW__; __THREW__ = 0;
      $536 = $535&1;
      if (!($536)) {
       $$13 = $530;$$1316 = $531;
       break L1;
      }
      $537 = ___cxa_find_matching_catch_3(0|0)|0;
      $538 = tempRet0;
      ___clang_call_terminate($537); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $492 = HEAP32[$468>>2]|0; //@line 106 "./common/Object.h"
     $493 = ((($492)) + 12|0); //@line 106 "./common/Object.h"
     $494 = HEAP32[$493>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($494|0,($468|0)); //@line 106 "./common/Object.h"
     $495 = __THREW__; __THREW__ = 0;
     $496 = $495&1;
     if ($496) {
      $497 = ___cxa_find_matching_catch_3(0|0)|0;
      $498 = tempRet0;
      ___clang_call_terminate($497); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $499 = (invoke_ii(3348,32)|0); //@line 226 "modules/event/sdl/Event.cpp"
     $500 = __THREW__; __THREW__ = 0;
     $501 = $500&1;
     if ($501) {
      label = 4;
      break L1;
     }
     __THREW__ = 0;
     invoke_viii(3460,($12|0),(365550|0),10); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $502 = __THREW__; __THREW__ = 0;
     $503 = $502&1;
     do {
      if ($503) {
       $539 = ___cxa_find_matching_catch_2()|0;
       $540 = tempRet0;
       $$3 = $539;$$36 = $540;
      } else {
       __THREW__ = 0;
       invoke_viii(3508,($499|0),($12|0),($vargs|0)); //@line 226 "modules/event/sdl/Event.cpp"
       $504 = __THREW__; __THREW__ = 0;
       $505 = $504&1;
       if ($505) {
        $541 = ___cxa_find_matching_catch_2()|0;
        $542 = tempRet0;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12); //@line 226 "modules/event/sdl/Event.cpp"
        $$3 = $541;$$36 = $542;
        break;
       } else {
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12); //@line 226 "modules/event/sdl/Event.cpp"
        $msg$1 = $499;
        break L3;
       }
      }
     } while(0);
     __ZdlPv($499); //@line 226 "modules/event/sdl/Event.cpp"
     $$13 = $$3;$$1316 = $$36;
     break L1;
     break;
    }
    case 1024:  {
     $543 = ((($e)) + 20|0); //@line 230 "modules/event/sdl/Event.cpp"
     $544 = HEAP32[$543>>2]|0; //@line 230 "modules/event/sdl/Event.cpp"
     $545 = (+($544|0)); //@line 230 "modules/event/sdl/Event.cpp"
     HEAPF64[$x>>3] = $545; //@line 230 "modules/event/sdl/Event.cpp"
     $546 = ((($e)) + 24|0); //@line 231 "modules/event/sdl/Event.cpp"
     $547 = HEAP32[$546>>2]|0; //@line 231 "modules/event/sdl/Event.cpp"
     $548 = (+($547|0)); //@line 231 "modules/event/sdl/Event.cpp"
     HEAPF64[$y>>3] = $548; //@line 231 "modules/event/sdl/Event.cpp"
     $549 = ((($e)) + 28|0); //@line 232 "modules/event/sdl/Event.cpp"
     $550 = HEAP32[$549>>2]|0; //@line 232 "modules/event/sdl/Event.cpp"
     $551 = (+($550|0)); //@line 232 "modules/event/sdl/Event.cpp"
     HEAPF64[$xrel>>3] = $551; //@line 232 "modules/event/sdl/Event.cpp"
     $552 = ((($e)) + 32|0); //@line 233 "modules/event/sdl/Event.cpp"
     $553 = HEAP32[$552>>2]|0; //@line 233 "modules/event/sdl/Event.cpp"
     $554 = (+($553|0)); //@line 233 "modules/event/sdl/Event.cpp"
     HEAPF64[$yrel>>3] = $554; //@line 233 "modules/event/sdl/Event.cpp"
     $555 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
     $556 = ($555|0)==(0|0); //@line 49 "modules/event/sdl/Event.cpp"
     if (!($556)) {
      $557 = HEAP32[$555>>2]|0; //@line 50 "modules/event/sdl/Event.cpp"
      $558 = ((($557)) + 144|0); //@line 50 "modules/event/sdl/Event.cpp"
      $559 = HEAP32[$558>>2]|0; //@line 50 "modules/event/sdl/Event.cpp"
      __THREW__ = 0;
      invoke_viii($559|0,($555|0),($x|0),($y|0)); //@line 50 "modules/event/sdl/Event.cpp"
      $560 = __THREW__; __THREW__ = 0;
      $561 = $560&1;
      if ($561) {
       label = 4;
       break L1;
      }
      $$pr = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
      $562 = ($$pr|0)==(0|0); //@line 49 "modules/event/sdl/Event.cpp"
      if (!($562)) {
       $563 = HEAP32[$$pr>>2]|0; //@line 50 "modules/event/sdl/Event.cpp"
       $564 = ((($563)) + 144|0); //@line 50 "modules/event/sdl/Event.cpp"
       $565 = HEAP32[$564>>2]|0; //@line 50 "modules/event/sdl/Event.cpp"
       __THREW__ = 0;
       invoke_viii($565|0,($$pr|0),($xrel|0),($yrel|0)); //@line 50 "modules/event/sdl/Event.cpp"
       $566 = __THREW__; __THREW__ = 0;
       $567 = $566&1;
       if ($567) {
        label = 4;
        break L1;
       }
      }
     }
     __THREW__ = 0;
     $568 = (invoke_ii(3348,32)|0); //@line 236 "modules/event/sdl/Event.cpp"
     $569 = __THREW__; __THREW__ = 0;
     $570 = $569&1;
     if ($570) {
      label = 4;
      break L1;
     }
     $571 = +HEAPF64[$x>>3]; //@line 236 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vid(3942,($568|0),(+$571)); //@line 236 "modules/event/sdl/Event.cpp"
     $572 = __THREW__; __THREW__ = 0;
     $573 = $572&1;
     if ($573) {
      $723 = ___cxa_find_matching_catch_2()|0;
      $724 = tempRet0;
      __ZdlPv($568); //@line 236 "modules/event/sdl/Event.cpp"
      $$13 = $723;$$1316 = $724;
      break L1;
     }
     HEAP32[$13>>2] = $568; //@line 93 "./common/Object.h"
     $574 = HEAP32[$568>>2]|0; //@line 95 "./common/Object.h"
     $575 = ((($574)) + 8|0); //@line 95 "./common/Object.h"
     $576 = HEAP32[$575>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($576|0,($568|0)); //@line 95 "./common/Object.h"
     $577 = __THREW__; __THREW__ = 0;
     $578 = $577&1;
     if ($578) {
      label = 4;
      break L1;
     }
     $579 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $580 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $581 = HEAP32[$580>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $582 = ($579>>>0)<($581>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($582) {
      HEAP32[$579>>2] = $568; //@line 99 "./common/Object.h"
      $583 = HEAP32[$568>>2]|0; //@line 101 "./common/Object.h"
      $584 = ((($583)) + 8|0); //@line 101 "./common/Object.h"
      $585 = HEAP32[$584>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($585|0,($568|0)); //@line 101 "./common/Object.h"
      $586 = __THREW__; __THREW__ = 0;
      $587 = $586&1;
      if ($587) {
       label = 232;
      } else {
       $$pre$i134 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $588 = ((($$pre$i134)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $588; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($vargs|0),($13|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $589 = __THREW__; __THREW__ = 0;
      $590 = $589&1;
      if ($590) {
       label = 232;
      }
     }
     if ((label|0) == 232) {
      $725 = ___cxa_find_matching_catch_2()|0;
      $726 = tempRet0;
      $727 = HEAP32[$568>>2]|0; //@line 106 "./common/Object.h"
      $728 = ((($727)) + 12|0); //@line 106 "./common/Object.h"
      $729 = HEAP32[$728>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($729|0,($568|0)); //@line 106 "./common/Object.h"
      $730 = __THREW__; __THREW__ = 0;
      $731 = $730&1;
      if (!($731)) {
       $$13 = $725;$$1316 = $726;
       break L1;
      }
      $732 = ___cxa_find_matching_catch_3(0|0)|0;
      $733 = tempRet0;
      ___clang_call_terminate($732); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $591 = HEAP32[$568>>2]|0; //@line 106 "./common/Object.h"
     $592 = ((($591)) + 12|0); //@line 106 "./common/Object.h"
     $593 = HEAP32[$592>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($593|0,($568|0)); //@line 106 "./common/Object.h"
     $594 = __THREW__; __THREW__ = 0;
     $595 = $594&1;
     if ($595) {
      $596 = ___cxa_find_matching_catch_3(0|0)|0;
      $597 = tempRet0;
      ___clang_call_terminate($596); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $598 = (invoke_ii(3348,32)|0); //@line 237 "modules/event/sdl/Event.cpp"
     $599 = __THREW__; __THREW__ = 0;
     $600 = $599&1;
     if ($600) {
      label = 4;
      break L1;
     }
     $601 = +HEAPF64[$y>>3]; //@line 237 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vid(3942,($598|0),(+$601)); //@line 237 "modules/event/sdl/Event.cpp"
     $602 = __THREW__; __THREW__ = 0;
     $603 = $602&1;
     if ($603) {
      $734 = ___cxa_find_matching_catch_2()|0;
      $735 = tempRet0;
      __ZdlPv($598); //@line 237 "modules/event/sdl/Event.cpp"
      $$13 = $734;$$1316 = $735;
      break L1;
     }
     HEAP32[$14>>2] = $598; //@line 93 "./common/Object.h"
     $604 = HEAP32[$598>>2]|0; //@line 95 "./common/Object.h"
     $605 = ((($604)) + 8|0); //@line 95 "./common/Object.h"
     $606 = HEAP32[$605>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($606|0,($598|0)); //@line 95 "./common/Object.h"
     $607 = __THREW__; __THREW__ = 0;
     $608 = $607&1;
     if ($608) {
      label = 4;
      break L1;
     }
     $609 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $610 = HEAP32[$580>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $611 = ($609>>>0)<($610>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($611) {
      HEAP32[$609>>2] = $598; //@line 99 "./common/Object.h"
      $612 = HEAP32[$598>>2]|0; //@line 101 "./common/Object.h"
      $613 = ((($612)) + 8|0); //@line 101 "./common/Object.h"
      $614 = HEAP32[$613>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($614|0,($598|0)); //@line 101 "./common/Object.h"
      $615 = __THREW__; __THREW__ = 0;
      $616 = $615&1;
      if ($616) {
       label = 235;
      } else {
       $$pre$i142 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $617 = ((($$pre$i142)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $617; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($vargs|0),($14|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $618 = __THREW__; __THREW__ = 0;
      $619 = $618&1;
      if ($619) {
       label = 235;
      }
     }
     if ((label|0) == 235) {
      $736 = ___cxa_find_matching_catch_2()|0;
      $737 = tempRet0;
      $738 = HEAP32[$598>>2]|0; //@line 106 "./common/Object.h"
      $739 = ((($738)) + 12|0); //@line 106 "./common/Object.h"
      $740 = HEAP32[$739>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($740|0,($598|0)); //@line 106 "./common/Object.h"
      $741 = __THREW__; __THREW__ = 0;
      $742 = $741&1;
      if (!($742)) {
       $$13 = $736;$$1316 = $737;
       break L1;
      }
      $743 = ___cxa_find_matching_catch_3(0|0)|0;
      $744 = tempRet0;
      ___clang_call_terminate($743); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $620 = HEAP32[$598>>2]|0; //@line 106 "./common/Object.h"
     $621 = ((($620)) + 12|0); //@line 106 "./common/Object.h"
     $622 = HEAP32[$621>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($622|0,($598|0)); //@line 106 "./common/Object.h"
     $623 = __THREW__; __THREW__ = 0;
     $624 = $623&1;
     if ($624) {
      $625 = ___cxa_find_matching_catch_3(0|0)|0;
      $626 = tempRet0;
      ___clang_call_terminate($625); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $627 = (invoke_ii(3348,32)|0); //@line 238 "modules/event/sdl/Event.cpp"
     $628 = __THREW__; __THREW__ = 0;
     $629 = $628&1;
     if ($629) {
      label = 4;
      break L1;
     }
     $630 = +HEAPF64[$xrel>>3]; //@line 238 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vid(3942,($627|0),(+$630)); //@line 238 "modules/event/sdl/Event.cpp"
     $631 = __THREW__; __THREW__ = 0;
     $632 = $631&1;
     if ($632) {
      $745 = ___cxa_find_matching_catch_2()|0;
      $746 = tempRet0;
      __ZdlPv($627); //@line 238 "modules/event/sdl/Event.cpp"
      $$13 = $745;$$1316 = $746;
      break L1;
     }
     HEAP32[$15>>2] = $627; //@line 93 "./common/Object.h"
     $633 = HEAP32[$627>>2]|0; //@line 95 "./common/Object.h"
     $634 = ((($633)) + 8|0); //@line 95 "./common/Object.h"
     $635 = HEAP32[$634>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($635|0,($627|0)); //@line 95 "./common/Object.h"
     $636 = __THREW__; __THREW__ = 0;
     $637 = $636&1;
     if ($637) {
      label = 4;
      break L1;
     }
     $638 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $639 = HEAP32[$580>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $640 = ($638>>>0)<($639>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($640) {
      HEAP32[$638>>2] = $627; //@line 99 "./common/Object.h"
      $641 = HEAP32[$627>>2]|0; //@line 101 "./common/Object.h"
      $642 = ((($641)) + 8|0); //@line 101 "./common/Object.h"
      $643 = HEAP32[$642>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($643|0,($627|0)); //@line 101 "./common/Object.h"
      $644 = __THREW__; __THREW__ = 0;
      $645 = $644&1;
      if ($645) {
       label = 238;
      } else {
       $$pre$i150 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $646 = ((($$pre$i150)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $646; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($vargs|0),($15|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $647 = __THREW__; __THREW__ = 0;
      $648 = $647&1;
      if ($648) {
       label = 238;
      }
     }
     if ((label|0) == 238) {
      $747 = ___cxa_find_matching_catch_2()|0;
      $748 = tempRet0;
      $749 = HEAP32[$627>>2]|0; //@line 106 "./common/Object.h"
      $750 = ((($749)) + 12|0); //@line 106 "./common/Object.h"
      $751 = HEAP32[$750>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($751|0,($627|0)); //@line 106 "./common/Object.h"
      $752 = __THREW__; __THREW__ = 0;
      $753 = $752&1;
      if (!($753)) {
       $$13 = $747;$$1316 = $748;
       break L1;
      }
      $754 = ___cxa_find_matching_catch_3(0|0)|0;
      $755 = tempRet0;
      ___clang_call_terminate($754); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $649 = HEAP32[$627>>2]|0; //@line 106 "./common/Object.h"
     $650 = ((($649)) + 12|0); //@line 106 "./common/Object.h"
     $651 = HEAP32[$650>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($651|0,($627|0)); //@line 106 "./common/Object.h"
     $652 = __THREW__; __THREW__ = 0;
     $653 = $652&1;
     if ($653) {
      $654 = ___cxa_find_matching_catch_3(0|0)|0;
      $655 = tempRet0;
      ___clang_call_terminate($654); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $656 = (invoke_ii(3348,32)|0); //@line 239 "modules/event/sdl/Event.cpp"
     $657 = __THREW__; __THREW__ = 0;
     $658 = $657&1;
     if ($658) {
      label = 4;
      break L1;
     }
     $659 = +HEAPF64[$yrel>>3]; //@line 239 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vid(3942,($656|0),(+$659)); //@line 239 "modules/event/sdl/Event.cpp"
     $660 = __THREW__; __THREW__ = 0;
     $661 = $660&1;
     if ($661) {
      $756 = ___cxa_find_matching_catch_2()|0;
      $757 = tempRet0;
      __ZdlPv($656); //@line 239 "modules/event/sdl/Event.cpp"
      $$13 = $756;$$1316 = $757;
      break L1;
     }
     HEAP32[$16>>2] = $656; //@line 93 "./common/Object.h"
     $662 = HEAP32[$656>>2]|0; //@line 95 "./common/Object.h"
     $663 = ((($662)) + 8|0); //@line 95 "./common/Object.h"
     $664 = HEAP32[$663>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($664|0,($656|0)); //@line 95 "./common/Object.h"
     $665 = __THREW__; __THREW__ = 0;
     $666 = $665&1;
     if ($666) {
      label = 4;
      break L1;
     }
     $667 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $668 = HEAP32[$580>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $669 = ($667>>>0)<($668>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     do {
      if ($669) {
       HEAP32[$667>>2] = $656; //@line 99 "./common/Object.h"
       $670 = HEAP32[$656>>2]|0; //@line 101 "./common/Object.h"
       $671 = ((($670)) + 8|0); //@line 101 "./common/Object.h"
       $672 = HEAP32[$671>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($672|0,($656|0)); //@line 101 "./common/Object.h"
       $673 = __THREW__; __THREW__ = 0;
       $674 = $673&1;
       if ($674) {
        label = 241;
        break;
       }
       $$pre$i158 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $675 = ((($$pre$i158)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $675; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      } else {
       __THREW__ = 0;
       invoke_vii(3507,($vargs|0),($16|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $676 = __THREW__; __THREW__ = 0;
       $677 = $676&1;
       if ($677) {
        label = 241;
       }
      }
     } while(0);
     if ((label|0) == 241) {
      $758 = ___cxa_find_matching_catch_2()|0;
      $759 = tempRet0;
      $760 = HEAP32[$656>>2]|0; //@line 106 "./common/Object.h"
      $761 = ((($760)) + 12|0); //@line 106 "./common/Object.h"
      $762 = HEAP32[$761>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($762|0,($656|0)); //@line 106 "./common/Object.h"
      $763 = __THREW__; __THREW__ = 0;
      $764 = $763&1;
      if (!($764)) {
       $$13 = $758;$$1316 = $759;
       break L1;
      }
      $765 = ___cxa_find_matching_catch_3(0|0)|0;
      $766 = tempRet0;
      ___clang_call_terminate($765); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $678 = HEAP32[$656>>2]|0; //@line 106 "./common/Object.h"
     $679 = ((($678)) + 12|0); //@line 106 "./common/Object.h"
     $680 = HEAP32[$679>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($680|0,($656|0)); //@line 106 "./common/Object.h"
     $681 = __THREW__; __THREW__ = 0;
     $682 = $681&1;
     if ($682) {
      $683 = ___cxa_find_matching_catch_3(0|0)|0;
      $684 = tempRet0;
      ___clang_call_terminate($683); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $685 = (invoke_ii(3348,32)|0); //@line 240 "modules/event/sdl/Event.cpp"
     $686 = __THREW__; __THREW__ = 0;
     $687 = $686&1;
     if ($687) {
      label = 4;
      break L1;
     }
     $688 = ((($e)) + 12|0); //@line 240 "modules/event/sdl/Event.cpp"
     $689 = HEAP32[$688>>2]|0; //@line 240 "modules/event/sdl/Event.cpp"
     $690 = ($689|0)==(-1); //@line 240 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vii(3941,($685|0),($690|0)); //@line 240 "modules/event/sdl/Event.cpp"
     $691 = __THREW__; __THREW__ = 0;
     $692 = $691&1;
     if ($692) {
      $767 = ___cxa_find_matching_catch_2()|0;
      $768 = tempRet0;
      __ZdlPv($685); //@line 240 "modules/event/sdl/Event.cpp"
      $$13 = $767;$$1316 = $768;
      break L1;
     }
     HEAP32[$17>>2] = $685; //@line 93 "./common/Object.h"
     $693 = HEAP32[$685>>2]|0; //@line 95 "./common/Object.h"
     $694 = ((($693)) + 8|0); //@line 95 "./common/Object.h"
     $695 = HEAP32[$694>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($695|0,($685|0)); //@line 95 "./common/Object.h"
     $696 = __THREW__; __THREW__ = 0;
     $697 = $696&1;
     if ($697) {
      label = 4;
      break L1;
     }
     $698 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $699 = HEAP32[$580>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $700 = ($698>>>0)<($699>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     do {
      if ($700) {
       HEAP32[$698>>2] = $685; //@line 99 "./common/Object.h"
       $701 = HEAP32[$685>>2]|0; //@line 101 "./common/Object.h"
       $702 = ((($701)) + 8|0); //@line 101 "./common/Object.h"
       $703 = HEAP32[$702>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($703|0,($685|0)); //@line 101 "./common/Object.h"
       $704 = __THREW__; __THREW__ = 0;
       $705 = $704&1;
       if ($705) {
        label = 244;
        break;
       }
       $$pre$i166 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $706 = ((($$pre$i166)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $706; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      } else {
       __THREW__ = 0;
       invoke_vii(3507,($vargs|0),($17|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $707 = __THREW__; __THREW__ = 0;
       $708 = $707&1;
       if ($708) {
        label = 244;
       }
      }
     } while(0);
     if ((label|0) == 244) {
      $769 = ___cxa_find_matching_catch_2()|0;
      $770 = tempRet0;
      $771 = HEAP32[$685>>2]|0; //@line 106 "./common/Object.h"
      $772 = ((($771)) + 12|0); //@line 106 "./common/Object.h"
      $773 = HEAP32[$772>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($773|0,($685|0)); //@line 106 "./common/Object.h"
      $774 = __THREW__; __THREW__ = 0;
      $775 = $774&1;
      if (!($775)) {
       $$13 = $769;$$1316 = $770;
       break L1;
      }
      $776 = ___cxa_find_matching_catch_3(0|0)|0;
      $777 = tempRet0;
      ___clang_call_terminate($776); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $709 = HEAP32[$685>>2]|0; //@line 106 "./common/Object.h"
     $710 = ((($709)) + 12|0); //@line 106 "./common/Object.h"
     $711 = HEAP32[$710>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($711|0,($685|0)); //@line 106 "./common/Object.h"
     $712 = __THREW__; __THREW__ = 0;
     $713 = $712&1;
     if ($713) {
      $714 = ___cxa_find_matching_catch_3(0|0)|0;
      $715 = tempRet0;
      ___clang_call_terminate($714); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $716 = (invoke_ii(3348,32)|0); //@line 241 "modules/event/sdl/Event.cpp"
     $717 = __THREW__; __THREW__ = 0;
     $718 = $717&1;
     if ($718) {
      label = 4;
      break L1;
     }
     __THREW__ = 0;
     invoke_viii(3460,($18|0),(365561|0),10); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $719 = __THREW__; __THREW__ = 0;
     $720 = $719&1;
     do {
      if ($720) {
       $778 = ___cxa_find_matching_catch_2()|0;
       $779 = tempRet0;
       $$4 = $778;$$47 = $779;
      } else {
       __THREW__ = 0;
       invoke_viii(3508,($716|0),($18|0),($vargs|0)); //@line 241 "modules/event/sdl/Event.cpp"
       $721 = __THREW__; __THREW__ = 0;
       $722 = $721&1;
       if ($722) {
        $780 = ___cxa_find_matching_catch_2()|0;
        $781 = tempRet0;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($18); //@line 241 "modules/event/sdl/Event.cpp"
        $$4 = $780;$$47 = $781;
        break;
       } else {
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($18); //@line 241 "modules/event/sdl/Event.cpp"
        $msg$1 = $716;
        break L3;
       }
      }
     } while(0);
     __ZdlPv($716); //@line 241 "modules/event/sdl/Event.cpp"
     $$13 = $$4;$$1316 = $$47;
     break L1;
     break;
    }
    case 1026: case 1025:  {
     $782 = ((($e)) + 16|0); //@line 249 "modules/event/sdl/Event.cpp"
     $783 = HEAP8[$782>>0]|0; //@line 249 "modules/event/sdl/Event.cpp"
     $784 = $783&255; //@line 249 "modules/event/sdl/Event.cpp"
     switch ($784|0) {
     case 3:  {
      $button$0 = 2;
      break;
     }
     case 2:  {
      $button$0 = 3;
      break;
     }
     default: {
      $button$0 = $784;
     }
     }
     $785 = ((($e)) + 20|0); //@line 260 "modules/event/sdl/Event.cpp"
     $786 = HEAP32[$785>>2]|0; //@line 260 "modules/event/sdl/Event.cpp"
     $787 = (+($786|0)); //@line 260 "modules/event/sdl/Event.cpp"
     HEAPF64[$x1>>3] = $787; //@line 260 "modules/event/sdl/Event.cpp"
     $788 = ((($e)) + 24|0); //@line 261 "modules/event/sdl/Event.cpp"
     $789 = HEAP32[$788>>2]|0; //@line 261 "modules/event/sdl/Event.cpp"
     $790 = (+($789|0)); //@line 261 "modules/event/sdl/Event.cpp"
     HEAPF64[$y2>>3] = $790; //@line 261 "modules/event/sdl/Event.cpp"
     $791 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
     $792 = ($791|0)==(0|0); //@line 49 "modules/event/sdl/Event.cpp"
     if (!($792)) {
      $793 = HEAP32[$791>>2]|0; //@line 50 "modules/event/sdl/Event.cpp"
      $794 = ((($793)) + 144|0); //@line 50 "modules/event/sdl/Event.cpp"
      $795 = HEAP32[$794>>2]|0; //@line 50 "modules/event/sdl/Event.cpp"
      __THREW__ = 0;
      invoke_viii($795|0,($791|0),($x1|0),($y2|0)); //@line 50 "modules/event/sdl/Event.cpp"
      $796 = __THREW__; __THREW__ = 0;
      $797 = $796&1;
      if ($797) {
       label = 4;
       break L1;
      }
     }
     __THREW__ = 0;
     $798 = (invoke_ii(3348,32)|0); //@line 263 "modules/event/sdl/Event.cpp"
     $799 = __THREW__; __THREW__ = 0;
     $800 = $799&1;
     if ($800) {
      label = 4;
      break L1;
     }
     $801 = +HEAPF64[$x1>>3]; //@line 263 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vid(3942,($798|0),(+$801)); //@line 263 "modules/event/sdl/Event.cpp"
     $802 = __THREW__; __THREW__ = 0;
     $803 = $802&1;
     if ($803) {
      $928 = ___cxa_find_matching_catch_2()|0;
      $929 = tempRet0;
      __ZdlPv($798); //@line 263 "modules/event/sdl/Event.cpp"
      $$13 = $928;$$1316 = $929;
      break L1;
     }
     HEAP32[$19>>2] = $798; //@line 93 "./common/Object.h"
     $804 = HEAP32[$798>>2]|0; //@line 95 "./common/Object.h"
     $805 = ((($804)) + 8|0); //@line 95 "./common/Object.h"
     $806 = HEAP32[$805>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($806|0,($798|0)); //@line 95 "./common/Object.h"
     $807 = __THREW__; __THREW__ = 0;
     $808 = $807&1;
     if ($808) {
      label = 4;
      break L1;
     }
     $809 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $810 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $811 = HEAP32[$810>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $812 = ($809>>>0)<($811>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($812) {
      HEAP32[$809>>2] = $798; //@line 99 "./common/Object.h"
      $813 = HEAP32[$798>>2]|0; //@line 101 "./common/Object.h"
      $814 = ((($813)) + 8|0); //@line 101 "./common/Object.h"
      $815 = HEAP32[$814>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($815|0,($798|0)); //@line 101 "./common/Object.h"
      $816 = __THREW__; __THREW__ = 0;
      $817 = $816&1;
      if ($817) {
       label = 295;
      } else {
       $$pre$i183 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $818 = ((($$pre$i183)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $818; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($vargs|0),($19|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $819 = __THREW__; __THREW__ = 0;
      $820 = $819&1;
      if ($820) {
       label = 295;
      }
     }
     if ((label|0) == 295) {
      $930 = ___cxa_find_matching_catch_2()|0;
      $931 = tempRet0;
      $932 = HEAP32[$798>>2]|0; //@line 106 "./common/Object.h"
      $933 = ((($932)) + 12|0); //@line 106 "./common/Object.h"
      $934 = HEAP32[$933>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($934|0,($798|0)); //@line 106 "./common/Object.h"
      $935 = __THREW__; __THREW__ = 0;
      $936 = $935&1;
      if (!($936)) {
       $$13 = $930;$$1316 = $931;
       break L1;
      }
      $937 = ___cxa_find_matching_catch_3(0|0)|0;
      $938 = tempRet0;
      ___clang_call_terminate($937); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $821 = HEAP32[$798>>2]|0; //@line 106 "./common/Object.h"
     $822 = ((($821)) + 12|0); //@line 106 "./common/Object.h"
     $823 = HEAP32[$822>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($823|0,($798|0)); //@line 106 "./common/Object.h"
     $824 = __THREW__; __THREW__ = 0;
     $825 = $824&1;
     if ($825) {
      $826 = ___cxa_find_matching_catch_3(0|0)|0;
      $827 = tempRet0;
      ___clang_call_terminate($826); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $828 = (invoke_ii(3348,32)|0); //@line 264 "modules/event/sdl/Event.cpp"
     $829 = __THREW__; __THREW__ = 0;
     $830 = $829&1;
     if ($830) {
      label = 4;
      break L1;
     }
     $831 = +HEAPF64[$y2>>3]; //@line 264 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vid(3942,($828|0),(+$831)); //@line 264 "modules/event/sdl/Event.cpp"
     $832 = __THREW__; __THREW__ = 0;
     $833 = $832&1;
     if ($833) {
      $939 = ___cxa_find_matching_catch_2()|0;
      $940 = tempRet0;
      __ZdlPv($828); //@line 264 "modules/event/sdl/Event.cpp"
      $$13 = $939;$$1316 = $940;
      break L1;
     }
     HEAP32[$20>>2] = $828; //@line 93 "./common/Object.h"
     $834 = HEAP32[$828>>2]|0; //@line 95 "./common/Object.h"
     $835 = ((($834)) + 8|0); //@line 95 "./common/Object.h"
     $836 = HEAP32[$835>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($836|0,($828|0)); //@line 95 "./common/Object.h"
     $837 = __THREW__; __THREW__ = 0;
     $838 = $837&1;
     if ($838) {
      label = 4;
      break L1;
     }
     $839 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $840 = HEAP32[$810>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $841 = ($839>>>0)<($840>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($841) {
      HEAP32[$839>>2] = $828; //@line 99 "./common/Object.h"
      $842 = HEAP32[$828>>2]|0; //@line 101 "./common/Object.h"
      $843 = ((($842)) + 8|0); //@line 101 "./common/Object.h"
      $844 = HEAP32[$843>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($844|0,($828|0)); //@line 101 "./common/Object.h"
      $845 = __THREW__; __THREW__ = 0;
      $846 = $845&1;
      if ($846) {
       label = 298;
      } else {
       $$pre$i191 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $847 = ((($$pre$i191)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $847; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($vargs|0),($20|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $848 = __THREW__; __THREW__ = 0;
      $849 = $848&1;
      if ($849) {
       label = 298;
      }
     }
     if ((label|0) == 298) {
      $941 = ___cxa_find_matching_catch_2()|0;
      $942 = tempRet0;
      $943 = HEAP32[$828>>2]|0; //@line 106 "./common/Object.h"
      $944 = ((($943)) + 12|0); //@line 106 "./common/Object.h"
      $945 = HEAP32[$944>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($945|0,($828|0)); //@line 106 "./common/Object.h"
      $946 = __THREW__; __THREW__ = 0;
      $947 = $946&1;
      if (!($947)) {
       $$13 = $941;$$1316 = $942;
       break L1;
      }
      $948 = ___cxa_find_matching_catch_3(0|0)|0;
      $949 = tempRet0;
      ___clang_call_terminate($948); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $850 = HEAP32[$828>>2]|0; //@line 106 "./common/Object.h"
     $851 = ((($850)) + 12|0); //@line 106 "./common/Object.h"
     $852 = HEAP32[$851>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($852|0,($828|0)); //@line 106 "./common/Object.h"
     $853 = __THREW__; __THREW__ = 0;
     $854 = $853&1;
     if ($854) {
      $855 = ___cxa_find_matching_catch_3(0|0)|0;
      $856 = tempRet0;
      ___clang_call_terminate($855); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $857 = (invoke_ii(3348,32)|0); //@line 265 "modules/event/sdl/Event.cpp"
     $858 = __THREW__; __THREW__ = 0;
     $859 = $858&1;
     if ($859) {
      label = 4;
      break L1;
     }
     $860 = (+($button$0|0)); //@line 265 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vid(3942,($857|0),(+$860)); //@line 265 "modules/event/sdl/Event.cpp"
     $861 = __THREW__; __THREW__ = 0;
     $862 = $861&1;
     if ($862) {
      $950 = ___cxa_find_matching_catch_2()|0;
      $951 = tempRet0;
      __ZdlPv($857); //@line 265 "modules/event/sdl/Event.cpp"
      $$13 = $950;$$1316 = $951;
      break L1;
     }
     HEAP32[$21>>2] = $857; //@line 93 "./common/Object.h"
     $863 = HEAP32[$857>>2]|0; //@line 95 "./common/Object.h"
     $864 = ((($863)) + 8|0); //@line 95 "./common/Object.h"
     $865 = HEAP32[$864>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($865|0,($857|0)); //@line 95 "./common/Object.h"
     $866 = __THREW__; __THREW__ = 0;
     $867 = $866&1;
     if ($867) {
      label = 4;
      break L1;
     }
     $868 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $869 = HEAP32[$810>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $870 = ($868>>>0)<($869>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($870) {
      HEAP32[$868>>2] = $857; //@line 99 "./common/Object.h"
      $871 = HEAP32[$857>>2]|0; //@line 101 "./common/Object.h"
      $872 = ((($871)) + 8|0); //@line 101 "./common/Object.h"
      $873 = HEAP32[$872>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($873|0,($857|0)); //@line 101 "./common/Object.h"
      $874 = __THREW__; __THREW__ = 0;
      $875 = $874&1;
      if ($875) {
       label = 301;
      } else {
       $$pre$i199 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $876 = ((($$pre$i199)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $876; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($vargs|0),($21|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $877 = __THREW__; __THREW__ = 0;
      $878 = $877&1;
      if ($878) {
       label = 301;
      }
     }
     if ((label|0) == 301) {
      $952 = ___cxa_find_matching_catch_2()|0;
      $953 = tempRet0;
      $954 = HEAP32[$857>>2]|0; //@line 106 "./common/Object.h"
      $955 = ((($954)) + 12|0); //@line 106 "./common/Object.h"
      $956 = HEAP32[$955>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($956|0,($857|0)); //@line 106 "./common/Object.h"
      $957 = __THREW__; __THREW__ = 0;
      $958 = $957&1;
      if (!($958)) {
       $$13 = $952;$$1316 = $953;
       break L1;
      }
      $959 = ___cxa_find_matching_catch_3(0|0)|0;
      $960 = tempRet0;
      ___clang_call_terminate($959); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $879 = HEAP32[$857>>2]|0; //@line 106 "./common/Object.h"
     $880 = ((($879)) + 12|0); //@line 106 "./common/Object.h"
     $881 = HEAP32[$880>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($881|0,($857|0)); //@line 106 "./common/Object.h"
     $882 = __THREW__; __THREW__ = 0;
     $883 = $882&1;
     if ($883) {
      $884 = ___cxa_find_matching_catch_3(0|0)|0;
      $885 = tempRet0;
      ___clang_call_terminate($884); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $886 = (invoke_ii(3348,32)|0); //@line 266 "modules/event/sdl/Event.cpp"
     $887 = __THREW__; __THREW__ = 0;
     $888 = $887&1;
     if ($888) {
      label = 4;
      break L1;
     }
     $889 = ((($e)) + 12|0); //@line 266 "modules/event/sdl/Event.cpp"
     $890 = HEAP32[$889>>2]|0; //@line 266 "modules/event/sdl/Event.cpp"
     $891 = ($890|0)==(-1); //@line 266 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vii(3941,($886|0),($891|0)); //@line 266 "modules/event/sdl/Event.cpp"
     $892 = __THREW__; __THREW__ = 0;
     $893 = $892&1;
     if ($893) {
      $961 = ___cxa_find_matching_catch_2()|0;
      $962 = tempRet0;
      __ZdlPv($886); //@line 266 "modules/event/sdl/Event.cpp"
      $$13 = $961;$$1316 = $962;
      break L1;
     }
     HEAP32[$22>>2] = $886; //@line 93 "./common/Object.h"
     $894 = HEAP32[$886>>2]|0; //@line 95 "./common/Object.h"
     $895 = ((($894)) + 8|0); //@line 95 "./common/Object.h"
     $896 = HEAP32[$895>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($896|0,($886|0)); //@line 95 "./common/Object.h"
     $897 = __THREW__; __THREW__ = 0;
     $898 = $897&1;
     if ($898) {
      label = 4;
      break L1;
     }
     $899 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $900 = HEAP32[$810>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $901 = ($899>>>0)<($900>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     do {
      if ($901) {
       HEAP32[$899>>2] = $886; //@line 99 "./common/Object.h"
       $902 = HEAP32[$886>>2]|0; //@line 101 "./common/Object.h"
       $903 = ((($902)) + 8|0); //@line 101 "./common/Object.h"
       $904 = HEAP32[$903>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($904|0,($886|0)); //@line 101 "./common/Object.h"
       $905 = __THREW__; __THREW__ = 0;
       $906 = $905&1;
       if ($906) {
        label = 304;
        break;
       }
       $$pre$i207 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $907 = ((($$pre$i207)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $907; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      } else {
       __THREW__ = 0;
       invoke_vii(3507,($vargs|0),($22|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $908 = __THREW__; __THREW__ = 0;
       $909 = $908&1;
       if ($909) {
        label = 304;
       }
      }
     } while(0);
     if ((label|0) == 304) {
      $963 = ___cxa_find_matching_catch_2()|0;
      $964 = tempRet0;
      $965 = HEAP32[$886>>2]|0; //@line 106 "./common/Object.h"
      $966 = ((($965)) + 12|0); //@line 106 "./common/Object.h"
      $967 = HEAP32[$966>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($967|0,($886|0)); //@line 106 "./common/Object.h"
      $968 = __THREW__; __THREW__ = 0;
      $969 = $968&1;
      if (!($969)) {
       $$13 = $963;$$1316 = $964;
       break L1;
      }
      $970 = ___cxa_find_matching_catch_3(0|0)|0;
      $971 = tempRet0;
      ___clang_call_terminate($970); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $910 = HEAP32[$886>>2]|0; //@line 106 "./common/Object.h"
     $911 = ((($910)) + 12|0); //@line 106 "./common/Object.h"
     $912 = HEAP32[$911>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($912|0,($886|0)); //@line 106 "./common/Object.h"
     $913 = __THREW__; __THREW__ = 0;
     $914 = $913&1;
     if ($914) {
      $915 = ___cxa_find_matching_catch_3(0|0)|0;
      $916 = tempRet0;
      ___clang_call_terminate($915); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $917 = (invoke_ii(3348,32)|0); //@line 267 "modules/event/sdl/Event.cpp"
     $918 = __THREW__; __THREW__ = 0;
     $919 = $918&1;
     if ($919) {
      label = 4;
      break L1;
     }
     $920 = HEAP32[$e>>2]|0; //@line 267 "modules/event/sdl/Event.cpp"
     $921 = ($920|0)==(1025); //@line 267 "modules/event/sdl/Event.cpp"
     $922 = $921 ? 365572 : 365585; //@line 267 "modules/event/sdl/Event.cpp"
     $923 = $921 ? 12 : 13; //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     __THREW__ = 0;
     invoke_viii(3460,($23|0),($922|0),($923|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $924 = __THREW__; __THREW__ = 0;
     $925 = $924&1;
     do {
      if ($925) {
       $972 = ___cxa_find_matching_catch_2()|0;
       $973 = tempRet0;
       $$5 = $972;$$58 = $973;
      } else {
       __THREW__ = 0;
       invoke_viii(3508,($917|0),($23|0),($vargs|0)); //@line 267 "modules/event/sdl/Event.cpp"
       $926 = __THREW__; __THREW__ = 0;
       $927 = $926&1;
       if ($927) {
        $974 = ___cxa_find_matching_catch_2()|0;
        $975 = tempRet0;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($23); //@line 267 "modules/event/sdl/Event.cpp"
        $$5 = $974;$$58 = $975;
        break;
       } else {
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($23); //@line 267 "modules/event/sdl/Event.cpp"
        $msg$1 = $917;
        break L3;
       }
      }
     } while(0);
     __ZdlPv($917); //@line 267 "modules/event/sdl/Event.cpp"
     $$13 = $$5;$$1316 = $$58;
     break L1;
     break;
    }
    case 1027:  {
     __THREW__ = 0;
     $976 = (invoke_ii(3348,32)|0); //@line 273 "modules/event/sdl/Event.cpp"
     $977 = __THREW__; __THREW__ = 0;
     $978 = $977&1;
     if ($978) {
      label = 4;
      break L1;
     }
     $979 = ((($e)) + 16|0); //@line 273 "modules/event/sdl/Event.cpp"
     $980 = HEAP32[$979>>2]|0; //@line 273 "modules/event/sdl/Event.cpp"
     $981 = (+($980|0)); //@line 273 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vid(3942,($976|0),(+$981)); //@line 273 "modules/event/sdl/Event.cpp"
     $982 = __THREW__; __THREW__ = 0;
     $983 = $982&1;
     if ($983) {
      $1046 = ___cxa_find_matching_catch_2()|0;
      $1047 = tempRet0;
      __ZdlPv($976); //@line 273 "modules/event/sdl/Event.cpp"
      $$13 = $1046;$$1316 = $1047;
      break L1;
     }
     HEAP32[$24>>2] = $976; //@line 93 "./common/Object.h"
     $984 = HEAP32[$976>>2]|0; //@line 95 "./common/Object.h"
     $985 = ((($984)) + 8|0); //@line 95 "./common/Object.h"
     $986 = HEAP32[$985>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($986|0,($976|0)); //@line 95 "./common/Object.h"
     $987 = __THREW__; __THREW__ = 0;
     $988 = $987&1;
     if ($988) {
      label = 4;
      break L1;
     }
     $989 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $990 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $991 = HEAP32[$990>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $992 = ($989>>>0)<($991>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($992) {
      HEAP32[$989>>2] = $976; //@line 99 "./common/Object.h"
      $993 = HEAP32[$976>>2]|0; //@line 101 "./common/Object.h"
      $994 = ((($993)) + 8|0); //@line 101 "./common/Object.h"
      $995 = HEAP32[$994>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($995|0,($976|0)); //@line 101 "./common/Object.h"
      $996 = __THREW__; __THREW__ = 0;
      $997 = $996&1;
      if ($997) {
       label = 332;
      } else {
       $$pre$i221 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $998 = ((($$pre$i221)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $998; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($vargs|0),($24|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $999 = __THREW__; __THREW__ = 0;
      $1000 = $999&1;
      if ($1000) {
       label = 332;
      }
     }
     if ((label|0) == 332) {
      $1048 = ___cxa_find_matching_catch_2()|0;
      $1049 = tempRet0;
      $1050 = HEAP32[$976>>2]|0; //@line 106 "./common/Object.h"
      $1051 = ((($1050)) + 12|0); //@line 106 "./common/Object.h"
      $1052 = HEAP32[$1051>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($1052|0,($976|0)); //@line 106 "./common/Object.h"
      $1053 = __THREW__; __THREW__ = 0;
      $1054 = $1053&1;
      if (!($1054)) {
       $$13 = $1048;$$1316 = $1049;
       break L1;
      }
      $1055 = ___cxa_find_matching_catch_3(0|0)|0;
      $1056 = tempRet0;
      ___clang_call_terminate($1055); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $1001 = HEAP32[$976>>2]|0; //@line 106 "./common/Object.h"
     $1002 = ((($1001)) + 12|0); //@line 106 "./common/Object.h"
     $1003 = HEAP32[$1002>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($1003|0,($976|0)); //@line 106 "./common/Object.h"
     $1004 = __THREW__; __THREW__ = 0;
     $1005 = $1004&1;
     if ($1005) {
      $1006 = ___cxa_find_matching_catch_3(0|0)|0;
      $1007 = tempRet0;
      ___clang_call_terminate($1006); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $1008 = (invoke_ii(3348,32)|0); //@line 274 "modules/event/sdl/Event.cpp"
     $1009 = __THREW__; __THREW__ = 0;
     $1010 = $1009&1;
     if ($1010) {
      label = 4;
      break L1;
     }
     $1011 = ((($e)) + 20|0); //@line 274 "modules/event/sdl/Event.cpp"
     $1012 = HEAP32[$1011>>2]|0; //@line 274 "modules/event/sdl/Event.cpp"
     $1013 = (+($1012|0)); //@line 274 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vid(3942,($1008|0),(+$1013)); //@line 274 "modules/event/sdl/Event.cpp"
     $1014 = __THREW__; __THREW__ = 0;
     $1015 = $1014&1;
     if ($1015) {
      $1057 = ___cxa_find_matching_catch_2()|0;
      $1058 = tempRet0;
      __ZdlPv($1008); //@line 274 "modules/event/sdl/Event.cpp"
      $$13 = $1057;$$1316 = $1058;
      break L1;
     }
     HEAP32[$25>>2] = $1008; //@line 93 "./common/Object.h"
     $1016 = HEAP32[$1008>>2]|0; //@line 95 "./common/Object.h"
     $1017 = ((($1016)) + 8|0); //@line 95 "./common/Object.h"
     $1018 = HEAP32[$1017>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($1018|0,($1008|0)); //@line 95 "./common/Object.h"
     $1019 = __THREW__; __THREW__ = 0;
     $1020 = $1019&1;
     if ($1020) {
      label = 4;
      break L1;
     }
     $1021 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $1022 = HEAP32[$990>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $1023 = ($1021>>>0)<($1022>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($1023) {
      HEAP32[$1021>>2] = $1008; //@line 99 "./common/Object.h"
      $1024 = HEAP32[$1008>>2]|0; //@line 101 "./common/Object.h"
      $1025 = ((($1024)) + 8|0); //@line 101 "./common/Object.h"
      $1026 = HEAP32[$1025>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($1026|0,($1008|0)); //@line 101 "./common/Object.h"
      $1027 = __THREW__; __THREW__ = 0;
      $1028 = $1027&1;
      if ($1028) {
       label = 335;
      } else {
       $$pre$i229 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $1029 = ((($$pre$i229)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $1029; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($vargs|0),($25|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $1030 = __THREW__; __THREW__ = 0;
      $1031 = $1030&1;
      if ($1031) {
       label = 335;
      }
     }
     if ((label|0) == 335) {
      $1059 = ___cxa_find_matching_catch_2()|0;
      $1060 = tempRet0;
      $1061 = HEAP32[$1008>>2]|0; //@line 106 "./common/Object.h"
      $1062 = ((($1061)) + 12|0); //@line 106 "./common/Object.h"
      $1063 = HEAP32[$1062>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($1063|0,($1008|0)); //@line 106 "./common/Object.h"
      $1064 = __THREW__; __THREW__ = 0;
      $1065 = $1064&1;
      if (!($1065)) {
       $$13 = $1059;$$1316 = $1060;
       break L1;
      }
      $1066 = ___cxa_find_matching_catch_3(0|0)|0;
      $1067 = tempRet0;
      ___clang_call_terminate($1066); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $1032 = HEAP32[$1008>>2]|0; //@line 106 "./common/Object.h"
     $1033 = ((($1032)) + 12|0); //@line 106 "./common/Object.h"
     $1034 = HEAP32[$1033>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($1034|0,($1008|0)); //@line 106 "./common/Object.h"
     $1035 = __THREW__; __THREW__ = 0;
     $1036 = $1035&1;
     if ($1036) {
      $1037 = ___cxa_find_matching_catch_3(0|0)|0;
      $1038 = tempRet0;
      ___clang_call_terminate($1037); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $1039 = (invoke_ii(3348,32)|0); //@line 275 "modules/event/sdl/Event.cpp"
     $1040 = __THREW__; __THREW__ = 0;
     $1041 = $1040&1;
     if ($1041) {
      label = 4;
      break L1;
     }
     __THREW__ = 0;
     invoke_viii(3460,($26|0),(365599|0),10); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $1042 = __THREW__; __THREW__ = 0;
     $1043 = $1042&1;
     do {
      if ($1043) {
       $1068 = ___cxa_find_matching_catch_2()|0;
       $1069 = tempRet0;
       $$6 = $1068;$$69 = $1069;
      } else {
       __THREW__ = 0;
       invoke_viii(3508,($1039|0),($26|0),($vargs|0)); //@line 275 "modules/event/sdl/Event.cpp"
       $1044 = __THREW__; __THREW__ = 0;
       $1045 = $1044&1;
       if ($1045) {
        $1070 = ___cxa_find_matching_catch_2()|0;
        $1071 = tempRet0;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($26); //@line 275 "modules/event/sdl/Event.cpp"
        $$6 = $1070;$$69 = $1071;
        break;
       } else {
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($26); //@line 275 "modules/event/sdl/Event.cpp"
        $msg$1 = $1039;
        break L3;
       }
      }
     } while(0);
     __ZdlPv($1039); //@line 275 "modules/event/sdl/Event.cpp"
     $$13 = $$6;$$1316 = $$69;
     break L1;
     break;
    }
    case 1794: case 1793: case 1792:  {
     $1072 = ((($e)) + 16|0); //@line 285 "modules/event/sdl/Event.cpp"
     $1073 = $1072; //@line 285 "modules/event/sdl/Event.cpp"
     $1074 = $1073; //@line 285 "modules/event/sdl/Event.cpp"
     $1075 = HEAP32[$1074>>2]|0; //@line 285 "modules/event/sdl/Event.cpp"
     $1076 = (($1073) + 4)|0; //@line 285 "modules/event/sdl/Event.cpp"
     $1077 = $1076; //@line 285 "modules/event/sdl/Event.cpp"
     $1078 = HEAP32[$1077>>2]|0; //@line 285 "modules/event/sdl/Event.cpp"
     $1079 = $touchinfo; //@line 285 "modules/event/sdl/Event.cpp"
     $1080 = $1079; //@line 285 "modules/event/sdl/Event.cpp"
     HEAP32[$1080>>2] = $1075; //@line 285 "modules/event/sdl/Event.cpp"
     $1081 = (($1079) + 4)|0; //@line 285 "modules/event/sdl/Event.cpp"
     $1082 = $1081; //@line 285 "modules/event/sdl/Event.cpp"
     HEAP32[$1082>>2] = $1078; //@line 285 "modules/event/sdl/Event.cpp"
     $1083 = ((($e)) + 24|0); //@line 286 "modules/event/sdl/Event.cpp"
     $1084 = +HEAPF32[$1083>>2]; //@line 286 "modules/event/sdl/Event.cpp"
     $1085 = $1084; //@line 286 "modules/event/sdl/Event.cpp"
     $1086 = ((($touchinfo)) + 8|0); //@line 286 "modules/event/sdl/Event.cpp"
     HEAPF64[$1086>>3] = $1085; //@line 286 "modules/event/sdl/Event.cpp"
     $1087 = ((($e)) + 28|0); //@line 287 "modules/event/sdl/Event.cpp"
     $1088 = +HEAPF32[$1087>>2]; //@line 287 "modules/event/sdl/Event.cpp"
     $1089 = $1088; //@line 287 "modules/event/sdl/Event.cpp"
     $1090 = ((($touchinfo)) + 16|0); //@line 287 "modules/event/sdl/Event.cpp"
     HEAPF64[$1090>>3] = $1089; //@line 287 "modules/event/sdl/Event.cpp"
     $1091 = ((($e)) + 32|0); //@line 288 "modules/event/sdl/Event.cpp"
     $1092 = +HEAPF32[$1091>>2]; //@line 288 "modules/event/sdl/Event.cpp"
     $1093 = $1092; //@line 288 "modules/event/sdl/Event.cpp"
     $1094 = ((($touchinfo)) + 24|0); //@line 288 "modules/event/sdl/Event.cpp"
     HEAPF64[$1094>>3] = $1093; //@line 288 "modules/event/sdl/Event.cpp"
     $1095 = ((($e)) + 36|0); //@line 289 "modules/event/sdl/Event.cpp"
     $1096 = +HEAPF32[$1095>>2]; //@line 289 "modules/event/sdl/Event.cpp"
     $1097 = $1096; //@line 289 "modules/event/sdl/Event.cpp"
     $1098 = ((($touchinfo)) + 32|0); //@line 289 "modules/event/sdl/Event.cpp"
     HEAPF64[$1098>>3] = $1097; //@line 289 "modules/event/sdl/Event.cpp"
     $1099 = ((($e)) + 40|0); //@line 290 "modules/event/sdl/Event.cpp"
     $1100 = +HEAPF32[$1099>>2]; //@line 290 "modules/event/sdl/Event.cpp"
     $1101 = $1100; //@line 290 "modules/event/sdl/Event.cpp"
     $1102 = ((($touchinfo)) + 40|0); //@line 290 "modules/event/sdl/Event.cpp"
     HEAPF64[$1102>>3] = $1101; //@line 290 "modules/event/sdl/Event.cpp"
     $1103 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
     HEAP32[$w$i>>2] = 1; //@line 57 "modules/event/sdl/Event.cpp"
     HEAP32[$h$i>>2] = 1; //@line 57 "modules/event/sdl/Event.cpp"
     $1104 = ($1103|0)==(0|0); //@line 59 "modules/event/sdl/Event.cpp"
     if ($1104) {
      $1111 = 1;$1113 = $1085;$1115 = 1;$1117 = $1089;$1119 = 0;
     } else {
      $1105 = HEAP32[$1103>>2]|0; //@line 60 "modules/event/sdl/Event.cpp"
      $1106 = ((($1105)) + 140|0); //@line 60 "modules/event/sdl/Event.cpp"
      $1107 = HEAP32[$1106>>2]|0; //@line 60 "modules/event/sdl/Event.cpp"
      __THREW__ = 0;
      invoke_viii($1107|0,($1103|0),($w$i|0),($h$i|0)); //@line 60 "modules/event/sdl/Event.cpp"
      $1108 = __THREW__; __THREW__ = 0;
      $1109 = $1108&1;
      if ($1109) {
       label = 4;
       break L1;
      }
      $$pre = +HEAPF64[$1086>>3]; //@line 63 "modules/event/sdl/Event.cpp"
      $$pre386 = HEAP32[$w$i>>2]|0; //@line 63 "modules/event/sdl/Event.cpp"
      $$pre387 = +HEAPF64[$1090>>3]; //@line 65 "modules/event/sdl/Event.cpp"
      $$pre388 = HEAP32[$h$i>>2]|0; //@line 65 "modules/event/sdl/Event.cpp"
      $$pre389 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
      $1111 = $$pre386;$1113 = $$pre;$1115 = $$pre388;$1117 = $$pre387;$1119 = $$pre389;
     }
     $1110 = (+($1111|0)); //@line 63 "modules/event/sdl/Event.cpp"
     $1112 = $1113 * $1110; //@line 63 "modules/event/sdl/Event.cpp"
     HEAPF64[$1086>>3] = $1112; //@line 63 "modules/event/sdl/Event.cpp"
     $1114 = (+($1115|0)); //@line 65 "modules/event/sdl/Event.cpp"
     $1116 = $1117 * $1114; //@line 65 "modules/event/sdl/Event.cpp"
     HEAPF64[$1090>>3] = $1116; //@line 65 "modules/event/sdl/Event.cpp"
     HEAP32[$w$i240>>2] = 1; //@line 57 "modules/event/sdl/Event.cpp"
     HEAP32[$h$i241>>2] = 1; //@line 57 "modules/event/sdl/Event.cpp"
     $1118 = ($1119|0)==(0|0); //@line 59 "modules/event/sdl/Event.cpp"
     if ($1118) {
      $1127 = 1;$1131 = 1;
     } else {
      $1120 = HEAP32[$1119>>2]|0; //@line 60 "modules/event/sdl/Event.cpp"
      $1121 = ((($1120)) + 140|0); //@line 60 "modules/event/sdl/Event.cpp"
      $1122 = HEAP32[$1121>>2]|0; //@line 60 "modules/event/sdl/Event.cpp"
      __THREW__ = 0;
      invoke_viii($1122|0,($1119|0),($w$i240|0),($h$i241|0)); //@line 60 "modules/event/sdl/Event.cpp"
      $1123 = __THREW__; __THREW__ = 0;
      $1124 = $1123&1;
      if ($1124) {
       label = 4;
       break L1;
      }
      $$pre391 = HEAP32[$w$i240>>2]|0; //@line 63 "modules/event/sdl/Event.cpp"
      $$pre392 = HEAP32[$h$i241>>2]|0; //@line 65 "modules/event/sdl/Event.cpp"
      $1127 = $$pre391;$1131 = $$pre392;
     }
     $1125 = +HEAPF64[$1094>>3]; //@line 63 "modules/event/sdl/Event.cpp"
     $1126 = (+($1127|0)); //@line 63 "modules/event/sdl/Event.cpp"
     $1128 = $1125 * $1126; //@line 63 "modules/event/sdl/Event.cpp"
     HEAPF64[$1094>>3] = $1128; //@line 63 "modules/event/sdl/Event.cpp"
     $1129 = +HEAPF64[$1098>>3]; //@line 65 "modules/event/sdl/Event.cpp"
     $1130 = (+($1131|0)); //@line 65 "modules/event/sdl/Event.cpp"
     $1132 = $1129 * $1130; //@line 65 "modules/event/sdl/Event.cpp"
     HEAPF64[$1098>>3] = $1132; //@line 65 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_viii(3460,($27|0),(365610|0),14); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $1133 = __THREW__; __THREW__ = 0;
     $1134 = $1133&1;
     if ($1134) {
      label = 4;
      break L1;
     }
     __THREW__ = 0;
     $1135 = (invoke_ii(3943,($27|0))|0); //@line 310 "modules/event/sdl/Event.cpp"
     $1136 = __THREW__; __THREW__ = 0;
     $1137 = $1136&1;
     if ($1137) {
      $1142 = ___cxa_find_matching_catch_2()|0;
      $1143 = tempRet0;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($27); //@line 310 "modules/event/sdl/Event.cpp"
      $$13 = $1142;$$1316 = $1143;
      break L1;
     }
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($27); //@line 310 "modules/event/sdl/Event.cpp"
     $1138 = ($1135|0)==(0|0); //@line 311 "modules/event/sdl/Event.cpp"
     if (!($1138)) {
      $1139 = HEAP32[$e>>2]|0; //@line 312 "modules/event/sdl/Event.cpp"
      __THREW__ = 0;
      invoke_viii(3944,($1135|0),($1139|0),($touchinfo|0)); //@line 312 "modules/event/sdl/Event.cpp"
      $1140 = __THREW__; __THREW__ = 0;
      $1141 = $1140&1;
      if ($1141) {
       label = 4;
       break L1;
      }
     }
     __THREW__ = 0;
     $1144 = (invoke_ii(3348,32)|0); //@line 319 "modules/event/sdl/Event.cpp"
     $1145 = __THREW__; __THREW__ = 0;
     $1146 = $1145&1;
     if ($1146) {
      label = 4;
      break L1;
     }
     $1147 = $touchinfo; //@line 319 "modules/event/sdl/Event.cpp"
     $1148 = $1147; //@line 319 "modules/event/sdl/Event.cpp"
     $1149 = HEAP32[$1148>>2]|0; //@line 319 "modules/event/sdl/Event.cpp"
     $1150 = (($1147) + 4)|0; //@line 319 "modules/event/sdl/Event.cpp"
     $1151 = $1150; //@line 319 "modules/event/sdl/Event.cpp"
     $1152 = HEAP32[$1151>>2]|0; //@line 319 "modules/event/sdl/Event.cpp"
     $1153 = $1149; //@line 319 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vii(3945,($1144|0),($1153|0)); //@line 319 "modules/event/sdl/Event.cpp"
     $1154 = __THREW__; __THREW__ = 0;
     $1155 = $1154&1;
     if ($1155) {
      $1326 = ___cxa_find_matching_catch_2()|0;
      $1327 = tempRet0;
      __ZdlPv($1144); //@line 319 "modules/event/sdl/Event.cpp"
      $$13 = $1326;$$1316 = $1327;
      break L1;
     }
     HEAP32[$28>>2] = $1144; //@line 93 "./common/Object.h"
     $1156 = HEAP32[$1144>>2]|0; //@line 95 "./common/Object.h"
     $1157 = ((($1156)) + 8|0); //@line 95 "./common/Object.h"
     $1158 = HEAP32[$1157>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($1158|0,($1144|0)); //@line 95 "./common/Object.h"
     $1159 = __THREW__; __THREW__ = 0;
     $1160 = $1159&1;
     if ($1160) {
      label = 4;
      break L1;
     }
     $1161 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $1162 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $1163 = HEAP32[$1162>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $1164 = ($1161>>>0)<($1163>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($1164) {
      HEAP32[$1161>>2] = $1144; //@line 99 "./common/Object.h"
      $1165 = HEAP32[$1144>>2]|0; //@line 101 "./common/Object.h"
      $1166 = ((($1165)) + 8|0); //@line 101 "./common/Object.h"
      $1167 = HEAP32[$1166>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($1167|0,($1144|0)); //@line 101 "./common/Object.h"
      $1168 = __THREW__; __THREW__ = 0;
      $1169 = $1168&1;
      if ($1169) {
       label = 408;
      } else {
       $$pre$i248 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $1170 = ((($$pre$i248)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $1170; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($vargs|0),($28|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $1171 = __THREW__; __THREW__ = 0;
      $1172 = $1171&1;
      if ($1172) {
       label = 408;
      }
     }
     if ((label|0) == 408) {
      $1328 = ___cxa_find_matching_catch_2()|0;
      $1329 = tempRet0;
      $1330 = HEAP32[$1144>>2]|0; //@line 106 "./common/Object.h"
      $1331 = ((($1330)) + 12|0); //@line 106 "./common/Object.h"
      $1332 = HEAP32[$1331>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($1332|0,($1144|0)); //@line 106 "./common/Object.h"
      $1333 = __THREW__; __THREW__ = 0;
      $1334 = $1333&1;
      if (!($1334)) {
       $$13 = $1328;$$1316 = $1329;
       break L1;
      }
      $1335 = ___cxa_find_matching_catch_3(0|0)|0;
      $1336 = tempRet0;
      ___clang_call_terminate($1335); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $1173 = HEAP32[$1144>>2]|0; //@line 106 "./common/Object.h"
     $1174 = ((($1173)) + 12|0); //@line 106 "./common/Object.h"
     $1175 = HEAP32[$1174>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($1175|0,($1144|0)); //@line 106 "./common/Object.h"
     $1176 = __THREW__; __THREW__ = 0;
     $1177 = $1176&1;
     if ($1177) {
      $1178 = ___cxa_find_matching_catch_3(0|0)|0;
      $1179 = tempRet0;
      ___clang_call_terminate($1178); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $1180 = (invoke_ii(3348,32)|0); //@line 320 "modules/event/sdl/Event.cpp"
     $1181 = __THREW__; __THREW__ = 0;
     $1182 = $1181&1;
     if ($1182) {
      label = 4;
      break L1;
     }
     $1183 = +HEAPF64[$1086>>3]; //@line 320 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vid(3942,($1180|0),(+$1183)); //@line 320 "modules/event/sdl/Event.cpp"
     $1184 = __THREW__; __THREW__ = 0;
     $1185 = $1184&1;
     if ($1185) {
      $1337 = ___cxa_find_matching_catch_2()|0;
      $1338 = tempRet0;
      __ZdlPv($1180); //@line 320 "modules/event/sdl/Event.cpp"
      $$13 = $1337;$$1316 = $1338;
      break L1;
     }
     HEAP32[$29>>2] = $1180; //@line 93 "./common/Object.h"
     $1186 = HEAP32[$1180>>2]|0; //@line 95 "./common/Object.h"
     $1187 = ((($1186)) + 8|0); //@line 95 "./common/Object.h"
     $1188 = HEAP32[$1187>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($1188|0,($1180|0)); //@line 95 "./common/Object.h"
     $1189 = __THREW__; __THREW__ = 0;
     $1190 = $1189&1;
     if ($1190) {
      label = 4;
      break L1;
     }
     $1191 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $1192 = HEAP32[$1162>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $1193 = ($1191>>>0)<($1192>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($1193) {
      HEAP32[$1191>>2] = $1180; //@line 99 "./common/Object.h"
      $1194 = HEAP32[$1180>>2]|0; //@line 101 "./common/Object.h"
      $1195 = ((($1194)) + 8|0); //@line 101 "./common/Object.h"
      $1196 = HEAP32[$1195>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($1196|0,($1180|0)); //@line 101 "./common/Object.h"
      $1197 = __THREW__; __THREW__ = 0;
      $1198 = $1197&1;
      if ($1198) {
       label = 411;
      } else {
       $$pre$i256 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $1199 = ((($$pre$i256)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $1199; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
     } else {
      __THREW__ = 0;
      invoke_vii(3507,($vargs|0),($29|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $1200 = __THREW__; __THREW__ = 0;
      $1201 = $1200&1;
      if ($1201) {
       label = 411;
      }
     }
     if ((label|0) == 411) {
      $1339 = ___cxa_find_matching_catch_2()|0;
      $1340 = tempRet0;
      $1341 = HEAP32[$1180>>2]|0; //@line 106 "./common/Object.h"
      $1342 = ((($1341)) + 12|0); //@line 106 "./common/Object.h"
      $1343 = HEAP32[$1342>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($1343|0,($1180|0)); //@line 106 "./common/Object.h"
      $1344 = __THREW__; __THREW__ = 0;
      $1345 = $1344&1;
      if (!($1345)) {
       $$13 = $1339;$$1316 = $1340;
       break L1;
      }
      $1346 = ___cxa_find_matching_catch_3(0|0)|0;
      $1347 = tempRet0;
      ___clang_call_terminate($1346); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $1202 = HEAP32[$1180>>2]|0; //@line 106 "./common/Object.h"
     $1203 = ((($1202)) + 12|0); //@line 106 "./common/Object.h"
     $1204 = HEAP32[$1203>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($1204|0,($1180|0)); //@line 106 "./common/Object.h"
     $1205 = __THREW__; __THREW__ = 0;
     $1206 = $1205&1;
     if ($1206) {
      $1207 = ___cxa_find_matching_catch_3(0|0)|0;
      $1208 = tempRet0;
      ___clang_call_terminate($1207); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $1209 = (invoke_ii(3348,32)|0); //@line 321 "modules/event/sdl/Event.cpp"
     $1210 = __THREW__; __THREW__ = 0;
     $1211 = $1210&1;
     if ($1211) {
      label = 4;
      break L1;
     }
     $1212 = +HEAPF64[$1090>>3]; //@line 321 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vid(3942,($1209|0),(+$1212)); //@line 321 "modules/event/sdl/Event.cpp"
     $1213 = __THREW__; __THREW__ = 0;
     $1214 = $1213&1;
     if ($1214) {
      $1348 = ___cxa_find_matching_catch_2()|0;
      $1349 = tempRet0;
      __ZdlPv($1209); //@line 321 "modules/event/sdl/Event.cpp"
      $$13 = $1348;$$1316 = $1349;
      break L1;
     }
     HEAP32[$30>>2] = $1209; //@line 93 "./common/Object.h"
     $1215 = HEAP32[$1209>>2]|0; //@line 95 "./common/Object.h"
     $1216 = ((($1215)) + 8|0); //@line 95 "./common/Object.h"
     $1217 = HEAP32[$1216>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($1217|0,($1209|0)); //@line 95 "./common/Object.h"
     $1218 = __THREW__; __THREW__ = 0;
     $1219 = $1218&1;
     if ($1219) {
      label = 4;
      break L1;
     }
     $1220 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $1221 = HEAP32[$1162>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $1222 = ($1220>>>0)<($1221>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     do {
      if ($1222) {
       HEAP32[$1220>>2] = $1209; //@line 99 "./common/Object.h"
       $1223 = HEAP32[$1209>>2]|0; //@line 101 "./common/Object.h"
       $1224 = ((($1223)) + 8|0); //@line 101 "./common/Object.h"
       $1225 = HEAP32[$1224>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($1225|0,($1209|0)); //@line 101 "./common/Object.h"
       $1226 = __THREW__; __THREW__ = 0;
       $1227 = $1226&1;
       if ($1227) {
        label = 414;
        break;
       }
       $$pre$i264 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $1228 = ((($$pre$i264)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $1228; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      } else {
       __THREW__ = 0;
       invoke_vii(3507,($vargs|0),($30|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $1229 = __THREW__; __THREW__ = 0;
       $1230 = $1229&1;
       if ($1230) {
        label = 414;
       }
      }
     } while(0);
     if ((label|0) == 414) {
      $1350 = ___cxa_find_matching_catch_2()|0;
      $1351 = tempRet0;
      $1352 = HEAP32[$1209>>2]|0; //@line 106 "./common/Object.h"
      $1353 = ((($1352)) + 12|0); //@line 106 "./common/Object.h"
      $1354 = HEAP32[$1353>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($1354|0,($1209|0)); //@line 106 "./common/Object.h"
      $1355 = __THREW__; __THREW__ = 0;
      $1356 = $1355&1;
      if (!($1356)) {
       $$13 = $1350;$$1316 = $1351;
       break L1;
      }
      $1357 = ___cxa_find_matching_catch_3(0|0)|0;
      $1358 = tempRet0;
      ___clang_call_terminate($1357); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $1231 = HEAP32[$1209>>2]|0; //@line 106 "./common/Object.h"
     $1232 = ((($1231)) + 12|0); //@line 106 "./common/Object.h"
     $1233 = HEAP32[$1232>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($1233|0,($1209|0)); //@line 106 "./common/Object.h"
     $1234 = __THREW__; __THREW__ = 0;
     $1235 = $1234&1;
     if ($1235) {
      $1236 = ___cxa_find_matching_catch_3(0|0)|0;
      $1237 = tempRet0;
      ___clang_call_terminate($1236); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $1238 = (invoke_ii(3348,32)|0); //@line 322 "modules/event/sdl/Event.cpp"
     $1239 = __THREW__; __THREW__ = 0;
     $1240 = $1239&1;
     if ($1240) {
      label = 4;
      break L1;
     }
     $1241 = +HEAPF64[$1094>>3]; //@line 322 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vid(3942,($1238|0),(+$1241)); //@line 322 "modules/event/sdl/Event.cpp"
     $1242 = __THREW__; __THREW__ = 0;
     $1243 = $1242&1;
     if ($1243) {
      $1359 = ___cxa_find_matching_catch_2()|0;
      $1360 = tempRet0;
      __ZdlPv($1238); //@line 322 "modules/event/sdl/Event.cpp"
      $$13 = $1359;$$1316 = $1360;
      break L1;
     }
     HEAP32[$31>>2] = $1238; //@line 93 "./common/Object.h"
     $1244 = HEAP32[$1238>>2]|0; //@line 95 "./common/Object.h"
     $1245 = ((($1244)) + 8|0); //@line 95 "./common/Object.h"
     $1246 = HEAP32[$1245>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($1246|0,($1238|0)); //@line 95 "./common/Object.h"
     $1247 = __THREW__; __THREW__ = 0;
     $1248 = $1247&1;
     if ($1248) {
      label = 4;
      break L1;
     }
     $1249 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $1250 = HEAP32[$1162>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $1251 = ($1249>>>0)<($1250>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     do {
      if ($1251) {
       HEAP32[$1249>>2] = $1238; //@line 99 "./common/Object.h"
       $1252 = HEAP32[$1238>>2]|0; //@line 101 "./common/Object.h"
       $1253 = ((($1252)) + 8|0); //@line 101 "./common/Object.h"
       $1254 = HEAP32[$1253>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($1254|0,($1238|0)); //@line 101 "./common/Object.h"
       $1255 = __THREW__; __THREW__ = 0;
       $1256 = $1255&1;
       if ($1256) {
        label = 417;
        break;
       }
       $$pre$i272 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $1257 = ((($$pre$i272)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $1257; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      } else {
       __THREW__ = 0;
       invoke_vii(3507,($vargs|0),($31|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $1258 = __THREW__; __THREW__ = 0;
       $1259 = $1258&1;
       if ($1259) {
        label = 417;
       }
      }
     } while(0);
     if ((label|0) == 417) {
      $1361 = ___cxa_find_matching_catch_2()|0;
      $1362 = tempRet0;
      $1363 = HEAP32[$1238>>2]|0; //@line 106 "./common/Object.h"
      $1364 = ((($1363)) + 12|0); //@line 106 "./common/Object.h"
      $1365 = HEAP32[$1364>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($1365|0,($1238|0)); //@line 106 "./common/Object.h"
      $1366 = __THREW__; __THREW__ = 0;
      $1367 = $1366&1;
      if (!($1367)) {
       $$13 = $1361;$$1316 = $1362;
       break L1;
      }
      $1368 = ___cxa_find_matching_catch_3(0|0)|0;
      $1369 = tempRet0;
      ___clang_call_terminate($1368); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $1260 = HEAP32[$1238>>2]|0; //@line 106 "./common/Object.h"
     $1261 = ((($1260)) + 12|0); //@line 106 "./common/Object.h"
     $1262 = HEAP32[$1261>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($1262|0,($1238|0)); //@line 106 "./common/Object.h"
     $1263 = __THREW__; __THREW__ = 0;
     $1264 = $1263&1;
     if ($1264) {
      $1265 = ___cxa_find_matching_catch_3(0|0)|0;
      $1266 = tempRet0;
      ___clang_call_terminate($1265); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $1267 = (invoke_ii(3348,32)|0); //@line 323 "modules/event/sdl/Event.cpp"
     $1268 = __THREW__; __THREW__ = 0;
     $1269 = $1268&1;
     if ($1269) {
      label = 4;
      break L1;
     }
     $1270 = +HEAPF64[$1098>>3]; //@line 323 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vid(3942,($1267|0),(+$1270)); //@line 323 "modules/event/sdl/Event.cpp"
     $1271 = __THREW__; __THREW__ = 0;
     $1272 = $1271&1;
     if ($1272) {
      $1370 = ___cxa_find_matching_catch_2()|0;
      $1371 = tempRet0;
      __ZdlPv($1267); //@line 323 "modules/event/sdl/Event.cpp"
      $$13 = $1370;$$1316 = $1371;
      break L1;
     }
     HEAP32[$32>>2] = $1267; //@line 93 "./common/Object.h"
     $1273 = HEAP32[$1267>>2]|0; //@line 95 "./common/Object.h"
     $1274 = ((($1273)) + 8|0); //@line 95 "./common/Object.h"
     $1275 = HEAP32[$1274>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($1275|0,($1267|0)); //@line 95 "./common/Object.h"
     $1276 = __THREW__; __THREW__ = 0;
     $1277 = $1276&1;
     if ($1277) {
      label = 4;
      break L1;
     }
     $1278 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $1279 = HEAP32[$1162>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $1280 = ($1278>>>0)<($1279>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     do {
      if ($1280) {
       HEAP32[$1278>>2] = $1267; //@line 99 "./common/Object.h"
       $1281 = HEAP32[$1267>>2]|0; //@line 101 "./common/Object.h"
       $1282 = ((($1281)) + 8|0); //@line 101 "./common/Object.h"
       $1283 = HEAP32[$1282>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($1283|0,($1267|0)); //@line 101 "./common/Object.h"
       $1284 = __THREW__; __THREW__ = 0;
       $1285 = $1284&1;
       if ($1285) {
        label = 420;
        break;
       }
       $$pre$i280 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $1286 = ((($$pre$i280)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $1286; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      } else {
       __THREW__ = 0;
       invoke_vii(3507,($vargs|0),($32|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $1287 = __THREW__; __THREW__ = 0;
       $1288 = $1287&1;
       if ($1288) {
        label = 420;
       }
      }
     } while(0);
     if ((label|0) == 420) {
      $1372 = ___cxa_find_matching_catch_2()|0;
      $1373 = tempRet0;
      $1374 = HEAP32[$1267>>2]|0; //@line 106 "./common/Object.h"
      $1375 = ((($1374)) + 12|0); //@line 106 "./common/Object.h"
      $1376 = HEAP32[$1375>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($1376|0,($1267|0)); //@line 106 "./common/Object.h"
      $1377 = __THREW__; __THREW__ = 0;
      $1378 = $1377&1;
      if (!($1378)) {
       $$13 = $1372;$$1316 = $1373;
       break L1;
      }
      $1379 = ___cxa_find_matching_catch_3(0|0)|0;
      $1380 = tempRet0;
      ___clang_call_terminate($1379); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $1289 = HEAP32[$1267>>2]|0; //@line 106 "./common/Object.h"
     $1290 = ((($1289)) + 12|0); //@line 106 "./common/Object.h"
     $1291 = HEAP32[$1290>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($1291|0,($1267|0)); //@line 106 "./common/Object.h"
     $1292 = __THREW__; __THREW__ = 0;
     $1293 = $1292&1;
     if ($1293) {
      $1294 = ___cxa_find_matching_catch_3(0|0)|0;
      $1295 = tempRet0;
      ___clang_call_terminate($1294); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     __THREW__ = 0;
     $1296 = (invoke_ii(3348,32)|0); //@line 324 "modules/event/sdl/Event.cpp"
     $1297 = __THREW__; __THREW__ = 0;
     $1298 = $1297&1;
     if ($1298) {
      label = 4;
      break L1;
     }
     $1299 = +HEAPF64[$1102>>3]; //@line 324 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vid(3942,($1296|0),(+$1299)); //@line 324 "modules/event/sdl/Event.cpp"
     $1300 = __THREW__; __THREW__ = 0;
     $1301 = $1300&1;
     if ($1301) {
      $1381 = ___cxa_find_matching_catch_2()|0;
      $1382 = tempRet0;
      __ZdlPv($1296); //@line 324 "modules/event/sdl/Event.cpp"
      $$13 = $1381;$$1316 = $1382;
      break L1;
     }
     HEAP32[$33>>2] = $1296; //@line 93 "./common/Object.h"
     $1302 = HEAP32[$1296>>2]|0; //@line 95 "./common/Object.h"
     $1303 = ((($1302)) + 8|0); //@line 95 "./common/Object.h"
     $1304 = HEAP32[$1303>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($1304|0,($1296|0)); //@line 95 "./common/Object.h"
     $1305 = __THREW__; __THREW__ = 0;
     $1306 = $1305&1;
     if ($1306) {
      label = 4;
      break L1;
     }
     $1307 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $1308 = HEAP32[$1162>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $1309 = ($1307>>>0)<($1308>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     do {
      if ($1309) {
       HEAP32[$1307>>2] = $1296; //@line 99 "./common/Object.h"
       $1310 = HEAP32[$1296>>2]|0; //@line 101 "./common/Object.h"
       $1311 = ((($1310)) + 8|0); //@line 101 "./common/Object.h"
       $1312 = HEAP32[$1311>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($1312|0,($1296|0)); //@line 101 "./common/Object.h"
       $1313 = __THREW__; __THREW__ = 0;
       $1314 = $1313&1;
       if ($1314) {
        label = 423;
        break;
       }
       $$pre$i288 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $1315 = ((($$pre$i288)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$46>>2] = $1315; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      } else {
       __THREW__ = 0;
       invoke_vii(3507,($vargs|0),($33|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $1316 = __THREW__; __THREW__ = 0;
       $1317 = $1316&1;
       if ($1317) {
        label = 423;
       }
      }
     } while(0);
     if ((label|0) == 423) {
      $1383 = ___cxa_find_matching_catch_2()|0;
      $1384 = tempRet0;
      $1385 = HEAP32[$1296>>2]|0; //@line 106 "./common/Object.h"
      $1386 = ((($1385)) + 12|0); //@line 106 "./common/Object.h"
      $1387 = HEAP32[$1386>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($1387|0,($1296|0)); //@line 106 "./common/Object.h"
      $1388 = __THREW__; __THREW__ = 0;
      $1389 = $1388&1;
      if (!($1389)) {
       $$13 = $1383;$$1316 = $1384;
       break L1;
      }
      $1390 = ___cxa_find_matching_catch_3(0|0)|0;
      $1391 = tempRet0;
      ___clang_call_terminate($1390); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $1318 = HEAP32[$1296>>2]|0; //@line 106 "./common/Object.h"
     $1319 = ((($1318)) + 12|0); //@line 106 "./common/Object.h"
     $1320 = HEAP32[$1319>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($1320|0,($1296|0)); //@line 106 "./common/Object.h"
     $1321 = __THREW__; __THREW__ = 0;
     $1322 = $1321&1;
     if ($1322) {
      $1323 = ___cxa_find_matching_catch_3(0|0)|0;
      $1324 = tempRet0;
      ___clang_call_terminate($1323); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     $1325 = HEAP32[$e>>2]|0; //@line 326 "modules/event/sdl/Event.cpp"
     switch ($1325|0) {
     case 1792:  {
      HEAP32[$txt>>2] = 365625; //@line 327 "modules/event/sdl/Event.cpp"
      $1395 = 365625;
      break;
     }
     case 1793:  {
      HEAP32[$txt>>2] = 365638; //@line 329 "modules/event/sdl/Event.cpp"
      $1395 = 365638;
      break;
     }
     default: {
      HEAP32[$txt>>2] = 365652; //@line 331 "modules/event/sdl/Event.cpp"
      $1395 = 365652;
     }
     }
     __THREW__ = 0;
     $1392 = (invoke_ii(3348,32)|0); //@line 332 "modules/event/sdl/Event.cpp"
     $1393 = __THREW__; __THREW__ = 0;
     $1394 = $1393&1;
     if ($1394) {
      label = 4;
      break L1;
     }
     $1396 = (_strlen($1395)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     __THREW__ = 0;
     invoke_viii(3460,($34|0),($1395|0),($1396|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $1397 = __THREW__; __THREW__ = 0;
     $1398 = $1397&1;
     do {
      if ($1398) {
       $1401 = ___cxa_find_matching_catch_2()|0;
       $1402 = tempRet0;
       $$7 = $1401;$$710 = $1402;
      } else {
       __THREW__ = 0;
       invoke_viii(3508,($1392|0),($34|0),($vargs|0)); //@line 332 "modules/event/sdl/Event.cpp"
       $1399 = __THREW__; __THREW__ = 0;
       $1400 = $1399&1;
       if ($1400) {
        $1403 = ___cxa_find_matching_catch_2()|0;
        $1404 = tempRet0;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 332 "modules/event/sdl/Event.cpp"
        $$7 = $1403;$$710 = $1404;
        break;
       } else {
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 332 "modules/event/sdl/Event.cpp"
        $msg$1 = $1392;
        break L3;
       }
      }
     } while(0);
     __ZdlPv($1392); //@line 332 "modules/event/sdl/Event.cpp"
     $$13 = $$7;$$1316 = $$710;
     break L1;
     break;
    }
    case 1616: case 1618: case 1617: case 1542: case 1541: case 1538: case 1537: case 1536: case 1540: case 1539:  {
     __THREW__ = 0;
     $1405 = (invoke_iii(3946,0,($e|0))|0); //@line 345 "modules/event/sdl/Event.cpp"
     $1406 = __THREW__; __THREW__ = 0;
     $1407 = $1406&1;
     if ($1407) {
      label = 4;
      break L1;
     } else {
      $msg$1 = $1405;
     }
     break;
    }
    case 512:  {
     __THREW__ = 0;
     $1408 = (invoke_iii(3947,0,($e|0))|0); //@line 348 "modules/event/sdl/Event.cpp"
     $1409 = __THREW__; __THREW__ = 0;
     $1410 = $1409&1;
     if ($1410) {
      label = 4;
      break L1;
     } else {
      $msg$1 = $1408;
     }
     break;
    }
    case 4096:  {
     $1411 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
     $1412 = ($1411|0)==(0|0); //@line 352 "modules/event/sdl/Event.cpp"
     L574: do {
      if ($1412) {
       $$pre395 = ((($e)) + 8|0); //@line 372 "modules/event/sdl/Event.cpp"
       $$pre$phi397Z2D = $$pre395;$msg$0 = 0;
      } else {
       $1413 = HEAP32[$1411>>2]|0; //@line 355 "modules/event/sdl/Event.cpp"
       $1414 = ((($1413)) + 168|0); //@line 355 "modules/event/sdl/Event.cpp"
       $1415 = HEAP32[$1414>>2]|0; //@line 355 "modules/event/sdl/Event.cpp"
       $1416 = ((($e)) + 8|0); //@line 355 "modules/event/sdl/Event.cpp"
       $1417 = HEAP32[$1416>>2]|0; //@line 355 "modules/event/sdl/Event.cpp"
       $1418 = (_strlen($1417)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       __THREW__ = 0;
       invoke_viii(3460,($35|0),($1417|0),($1418|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $1419 = __THREW__; __THREW__ = 0;
       $1420 = $1419&1;
       if ($1420) {
        label = 4;
        break L1;
       }
       __THREW__ = 0;
       invoke_vii($1415|0,($1411|0),($35|0)); //@line 355 "modules/event/sdl/Event.cpp"
       $1421 = __THREW__; __THREW__ = 0;
       $1422 = $1421&1;
       if ($1422) {
        $1471 = ___cxa_find_matching_catch_2()|0;
        $1472 = tempRet0;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($35); //@line 355 "modules/event/sdl/Event.cpp"
        $$13 = $1471;$$1316 = $1472;
        break L1;
       }
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($35); //@line 355 "modules/event/sdl/Event.cpp"
       $1423 = HEAP32[$1411>>2]|0; //@line 357 "modules/event/sdl/Event.cpp"
       $1424 = ((($1423)) + 172|0); //@line 357 "modules/event/sdl/Event.cpp"
       $1425 = HEAP32[$1424>>2]|0; //@line 357 "modules/event/sdl/Event.cpp"
       $1426 = HEAP32[$1416>>2]|0; //@line 357 "modules/event/sdl/Event.cpp"
       $1427 = (_strlen($1426)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       __THREW__ = 0;
       invoke_viii(3460,($36|0),($1426|0),($1427|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $1428 = __THREW__; __THREW__ = 0;
       $1429 = $1428&1;
       if ($1429) {
        label = 4;
        break L1;
       }
       __THREW__ = 0;
       $1430 = (invoke_iii($1425|0,($1411|0),($36|0))|0); //@line 357 "modules/event/sdl/Event.cpp"
       $1431 = __THREW__; __THREW__ = 0;
       $1432 = $1431&1;
       if ($1432) {
        $1473 = ___cxa_find_matching_catch_2()|0;
        $1474 = tempRet0;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($36); //@line 357 "modules/event/sdl/Event.cpp"
        $$13 = $1473;$$1316 = $1474;
        break L1;
       }
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($36); //@line 357 "modules/event/sdl/Event.cpp"
       if ($1430) {
        __THREW__ = 0;
        $1433 = (invoke_ii(3348,32)|0); //@line 359 "modules/event/sdl/Event.cpp"
        $1434 = __THREW__; __THREW__ = 0;
        $1435 = $1434&1;
        if ($1435) {
         label = 4;
         break L1;
        }
        $1436 = HEAP32[$1416>>2]|0; //@line 359 "modules/event/sdl/Event.cpp"
        $1437 = (_strlen($1436)|0); //@line 359 "modules/event/sdl/Event.cpp"
        __THREW__ = 0;
        invoke_viii(3940,($1433|0),($1436|0),($1437|0)); //@line 359 "modules/event/sdl/Event.cpp"
        $1438 = __THREW__; __THREW__ = 0;
        $1439 = $1438&1;
        if ($1439) {
         $1475 = ___cxa_find_matching_catch_2()|0;
         $1476 = tempRet0;
         __ZdlPv($1433); //@line 359 "modules/event/sdl/Event.cpp"
         $$13 = $1475;$$1316 = $1476;
         break L1;
        }
        HEAP32[$37>>2] = $1433; //@line 93 "./common/Object.h"
        $1440 = HEAP32[$1433>>2]|0; //@line 95 "./common/Object.h"
        $1441 = ((($1440)) + 8|0); //@line 95 "./common/Object.h"
        $1442 = HEAP32[$1441>>2]|0; //@line 95 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($1442|0,($1433|0)); //@line 95 "./common/Object.h"
        $1443 = __THREW__; __THREW__ = 0;
        $1444 = $1443&1;
        if ($1444) {
         label = 4;
         break L1;
        }
        $1445 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $1446 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $1447 = HEAP32[$1446>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $1448 = ($1445>>>0)<($1447>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($1448) {
         HEAP32[$1445>>2] = $1433; //@line 99 "./common/Object.h"
         $1449 = HEAP32[$1433>>2]|0; //@line 101 "./common/Object.h"
         $1450 = ((($1449)) + 8|0); //@line 101 "./common/Object.h"
         $1451 = HEAP32[$1450>>2]|0; //@line 101 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($1451|0,($1433|0)); //@line 101 "./common/Object.h"
         $1452 = __THREW__; __THREW__ = 0;
         $1453 = $1452&1;
         if ($1453) {
          label = 459;
         } else {
          $$pre$i308 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $1454 = ((($$pre$i308)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$46>>2] = $1454; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         }
        } else {
         __THREW__ = 0;
         invoke_vii(3507,($vargs|0),($37|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $1455 = __THREW__; __THREW__ = 0;
         $1456 = $1455&1;
         if ($1456) {
          label = 459;
         }
        }
        if ((label|0) == 459) {
         $1477 = ___cxa_find_matching_catch_2()|0;
         $1478 = tempRet0;
         $1479 = HEAP32[$1433>>2]|0; //@line 106 "./common/Object.h"
         $1480 = ((($1479)) + 12|0); //@line 106 "./common/Object.h"
         $1481 = HEAP32[$1480>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($1481|0,($1433|0)); //@line 106 "./common/Object.h"
         $1482 = __THREW__; __THREW__ = 0;
         $1483 = $1482&1;
         if (!($1483)) {
          $$13 = $1477;$$1316 = $1478;
          break L1;
         }
         $1484 = ___cxa_find_matching_catch_3(0|0)|0;
         $1485 = tempRet0;
         ___clang_call_terminate($1484); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        $1457 = HEAP32[$1433>>2]|0; //@line 106 "./common/Object.h"
        $1458 = ((($1457)) + 12|0); //@line 106 "./common/Object.h"
        $1459 = HEAP32[$1458>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($1459|0,($1433|0)); //@line 106 "./common/Object.h"
        $1460 = __THREW__; __THREW__ = 0;
        $1461 = $1460&1;
        if ($1461) {
         $1462 = ___cxa_find_matching_catch_3(0|0)|0;
         $1463 = tempRet0;
         ___clang_call_terminate($1462); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        __THREW__ = 0;
        $1464 = (invoke_ii(3348,32)|0); //@line 360 "modules/event/sdl/Event.cpp"
        $1465 = __THREW__; __THREW__ = 0;
        $1466 = $1465&1;
        if ($1466) {
         label = 4;
         break L1;
        }
        __THREW__ = 0;
        invoke_viii(3460,($38|0),(365826|0),16); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $1467 = __THREW__; __THREW__ = 0;
        $1468 = $1467&1;
        do {
         if ($1468) {
          $1486 = ___cxa_find_matching_catch_2()|0;
          $1487 = tempRet0;
          $$8 = $1486;$$811 = $1487;
         } else {
          __THREW__ = 0;
          invoke_viii(3508,($1464|0),($38|0),($vargs|0)); //@line 360 "modules/event/sdl/Event.cpp"
          $1469 = __THREW__; __THREW__ = 0;
          $1470 = $1469&1;
          if ($1470) {
           $1488 = ___cxa_find_matching_catch_2()|0;
           $1489 = tempRet0;
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($38); //@line 360 "modules/event/sdl/Event.cpp"
           $$8 = $1488;$$811 = $1489;
           break;
          } else {
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($38); //@line 360 "modules/event/sdl/Event.cpp"
           $$pre$phi397Z2D = $1416;$msg$0 = $1464;
           break L574;
          }
         }
        } while(0);
        __ZdlPv($1464); //@line 360 "modules/event/sdl/Event.cpp"
        $$13 = $$8;$$1316 = $$811;
        break L1;
       }
       __THREW__ = 0;
       $1490 = (invoke_ii(3348,40)|0); //@line 365 "modules/event/sdl/Event.cpp"
       $1491 = __THREW__; __THREW__ = 0;
       $1492 = $1491&1;
       if ($1492) {
        label = 4;
        break L1;
       }
       $1493 = HEAP32[$1416>>2]|0; //@line 365 "modules/event/sdl/Event.cpp"
       $1494 = (_strlen($1493)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       __THREW__ = 0;
       invoke_viii(3460,($39|0),($1493|0),($1494|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $1495 = __THREW__; __THREW__ = 0;
       $1496 = $1495&1;
       do {
        if ($1496) {
         $1542 = ___cxa_find_matching_catch_2()|0;
         $1543 = tempRet0;
         $$9 = $1542;$$912 = $1543;
        } else {
         __THREW__ = 0;
         invoke_vii(3948,($1490|0),($39|0)); //@line 365 "modules/event/sdl/Event.cpp"
         $1497 = __THREW__; __THREW__ = 0;
         $1498 = $1497&1;
         if ($1498) {
          $1544 = ___cxa_find_matching_catch_2()|0;
          $1545 = tempRet0;
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($39); //@line 365 "modules/event/sdl/Event.cpp"
          $$9 = $1544;$$912 = $1545;
          break;
         }
         $1499 = ((($proxy)) + 4|0); //@line 365 "modules/event/sdl/Event.cpp"
         HEAP32[$1499>>2] = $1490; //@line 365 "modules/event/sdl/Event.cpp"
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($39); //@line 365 "modules/event/sdl/Event.cpp"
         HEAP32[$proxy>>2] = 6; //@line 366 "modules/event/sdl/Event.cpp"
         __THREW__ = 0;
         $1500 = (invoke_ii(3348,32)|0); //@line 367 "modules/event/sdl/Event.cpp"
         $1501 = __THREW__; __THREW__ = 0;
         $1502 = $1501&1;
         if ($1502) {
          label = 4;
          break L1;
         }
         __THREW__ = 0;
         invoke_viii(3949,($1500|0),6,($proxy|0)); //@line 367 "modules/event/sdl/Event.cpp"
         $1503 = __THREW__; __THREW__ = 0;
         $1504 = $1503&1;
         if ($1504) {
          $1546 = ___cxa_find_matching_catch_2()|0;
          $1547 = tempRet0;
          __ZdlPv($1500); //@line 367 "modules/event/sdl/Event.cpp"
          $$13 = $1546;$$1316 = $1547;
          break L1;
         }
         HEAP32[$40>>2] = $1500; //@line 93 "./common/Object.h"
         $1505 = HEAP32[$1500>>2]|0; //@line 95 "./common/Object.h"
         $1506 = ((($1505)) + 8|0); //@line 95 "./common/Object.h"
         $1507 = HEAP32[$1506>>2]|0; //@line 95 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($1507|0,($1500|0)); //@line 95 "./common/Object.h"
         $1508 = __THREW__; __THREW__ = 0;
         $1509 = $1508&1;
         if ($1509) {
          label = 4;
          break L1;
         }
         $1510 = HEAP32[$46>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $1511 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $1512 = HEAP32[$1511>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $1513 = ($1510>>>0)<($1512>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($1513) {
          HEAP32[$1510>>2] = $1500; //@line 99 "./common/Object.h"
          $1514 = HEAP32[$1500>>2]|0; //@line 101 "./common/Object.h"
          $1515 = ((($1514)) + 8|0); //@line 101 "./common/Object.h"
          $1516 = HEAP32[$1515>>2]|0; //@line 101 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($1516|0,($1500|0)); //@line 101 "./common/Object.h"
          $1517 = __THREW__; __THREW__ = 0;
          $1518 = $1517&1;
          if ($1518) {
           label = 485;
          } else {
           $$pre$i321 = HEAP32[$46>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $1519 = ((($$pre$i321)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           HEAP32[$46>>2] = $1519; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          }
         } else {
          __THREW__ = 0;
          invoke_vii(3507,($vargs|0),($40|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $1520 = __THREW__; __THREW__ = 0;
          $1521 = $1520&1;
          if ($1521) {
           label = 485;
          }
         }
         if ((label|0) == 485) {
          $1548 = ___cxa_find_matching_catch_2()|0;
          $1549 = tempRet0;
          $1550 = HEAP32[$1500>>2]|0; //@line 106 "./common/Object.h"
          $1551 = ((($1550)) + 12|0); //@line 106 "./common/Object.h"
          $1552 = HEAP32[$1551>>2]|0; //@line 106 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($1552|0,($1500|0)); //@line 106 "./common/Object.h"
          $1553 = __THREW__; __THREW__ = 0;
          $1554 = $1553&1;
          if (!($1554)) {
           $$13 = $1548;$$1316 = $1549;
           break L1;
          }
          $1555 = ___cxa_find_matching_catch_3(0|0)|0;
          $1556 = tempRet0;
          ___clang_call_terminate($1555); //@line 106 "./common/Object.h"
          // unreachable; //@line 106 "./common/Object.h"
         }
         $1522 = HEAP32[$1500>>2]|0; //@line 106 "./common/Object.h"
         $1523 = ((($1522)) + 12|0); //@line 106 "./common/Object.h"
         $1524 = HEAP32[$1523>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($1524|0,($1500|0)); //@line 106 "./common/Object.h"
         $1525 = __THREW__; __THREW__ = 0;
         $1526 = $1525&1;
         if ($1526) {
          $1527 = ___cxa_find_matching_catch_3(0|0)|0;
          $1528 = tempRet0;
          ___clang_call_terminate($1527); //@line 106 "./common/Object.h"
          // unreachable; //@line 106 "./common/Object.h"
         }
         __THREW__ = 0;
         $1529 = (invoke_ii(3348,32)|0); //@line 368 "modules/event/sdl/Event.cpp"
         $1530 = __THREW__; __THREW__ = 0;
         $1531 = $1530&1;
         if ($1531) {
          label = 4;
          break L1;
         }
         __THREW__ = 0;
         invoke_viii(3460,($41|0),(365843|0),11); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $1532 = __THREW__; __THREW__ = 0;
         $1533 = $1532&1;
         do {
          if ($1533) {
           $1557 = ___cxa_find_matching_catch_2()|0;
           $1558 = tempRet0;
           $$10 = $1557;$$1013 = $1558;
          } else {
           __THREW__ = 0;
           invoke_viii(3508,($1529|0),($41|0),($vargs|0)); //@line 368 "modules/event/sdl/Event.cpp"
           $1534 = __THREW__; __THREW__ = 0;
           $1535 = $1534&1;
           if ($1535) {
            $1559 = ___cxa_find_matching_catch_2()|0;
            $1560 = tempRet0;
            __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($41); //@line 368 "modules/event/sdl/Event.cpp"
            $$10 = $1559;$$1013 = $1560;
            break;
           }
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($41); //@line 368 "modules/event/sdl/Event.cpp"
           $1536 = HEAP32[$1499>>2]|0; //@line 369 "modules/event/sdl/Event.cpp"
           $1537 = HEAP32[$1536>>2]|0; //@line 369 "modules/event/sdl/Event.cpp"
           $1538 = ((($1537)) + 12|0); //@line 369 "modules/event/sdl/Event.cpp"
           $1539 = HEAP32[$1538>>2]|0; //@line 369 "modules/event/sdl/Event.cpp"
           __THREW__ = 0;
           invoke_vi($1539|0,($1536|0)); //@line 369 "modules/event/sdl/Event.cpp"
           $1540 = __THREW__; __THREW__ = 0;
           $1541 = $1540&1;
           if ($1541) {
            label = 4;
            break L1;
           }
           $$pre$phi397Z2D = $1416;$msg$0 = $1529;
           break L574;
          }
         } while(0);
         __ZdlPv($1529); //@line 368 "modules/event/sdl/Event.cpp"
         $$13 = $$10;$$1316 = $$1013;
         break L1;
        }
       } while(0);
       __ZdlPv($1490); //@line 365 "modules/event/sdl/Event.cpp"
       $$13 = $$9;$$1316 = $$912;
       break L1;
      }
     } while(0);
     $1561 = HEAP32[$$pre$phi397Z2D>>2]|0; //@line 372 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     invoke_vi(3775,($1561|0)); //@line 372 "modules/event/sdl/Event.cpp"
     $1562 = __THREW__; __THREW__ = 0;
     $1563 = $1562&1;
     if ($1563) {
      label = 4;
      break L1;
     } else {
      $msg$1 = $msg$0;
     }
     break;
    }
    case 257: case 256:  {
     __THREW__ = 0;
     $1564 = (invoke_ii(3348,32)|0); //@line 376 "modules/event/sdl/Event.cpp"
     $1565 = __THREW__; __THREW__ = 0;
     $1566 = $1565&1;
     if ($1566) {
      label = 4;
      break L1;
     }
     __THREW__ = 0;
     invoke_viii(3460,($42|0),(365855|0),4); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $1567 = __THREW__; __THREW__ = 0;
     $1568 = $1567&1;
     do {
      if ($1568) {
       $1591 = ___cxa_find_matching_catch_2()|0;
       $1592 = tempRet0;
       $$11 = $1591;$$1114 = $1592;
      } else {
       HEAP32[$43>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $1569 = ((($43)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$1569>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $1570 = ((($43)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$1570>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       __THREW__ = 0;
       invoke_viii(3508,($1564|0),($42|0),($43|0)); //@line 376 "modules/event/sdl/Event.cpp"
       $1571 = __THREW__; __THREW__ = 0;
       $1572 = $1571&1;
       if ($1572) {
        $1593 = ___cxa_find_matching_catch_2()|0;
        $1594 = tempRet0;
        $1595 = HEAP32[$43>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $1596 = ($1595|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($1596)) {
         $1597 = HEAP32[$1569>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $1598 = ($1597|0)==($1595|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         do {
          if ($1598) {
           $1612 = $1595;
          } else {
           $1600 = $1597;
           while(1) {
            $1599 = ((($1600)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            HEAP32[$1569>>2] = $1599; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $1601 = HEAP32[$1599>>2]|0; //@line 106 "./common/Object.h"
            $1602 = ($1601|0)==(0|0); //@line 106 "./common/Object.h"
            if ($1602) {
             $1604 = $1599;
            } else {
             $1605 = HEAP32[$1601>>2]|0; //@line 106 "./common/Object.h"
             $1606 = ((($1605)) + 12|0); //@line 106 "./common/Object.h"
             $1607 = HEAP32[$1606>>2]|0; //@line 106 "./common/Object.h"
             __THREW__ = 0;
             invoke_vi($1607|0,($1601|0)); //@line 106 "./common/Object.h"
             $1608 = __THREW__; __THREW__ = 0;
             $1609 = $1608&1;
             if ($1609) {
              label = 511;
              break;
             }
             $$pre$i$i$i343 = HEAP32[$1569>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $1604 = $$pre$i$i$i343;
            }
            $1603 = ($1604|0)==($1595|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            if ($1603) {
             label = 512;
             break;
            } else {
             $1600 = $1604;
            }
           }
           if ((label|0) == 511) {
            $1610 = ___cxa_find_matching_catch_3(0|0)|0;
            $1611 = tempRet0;
            ___clang_call_terminate($1610); //@line 106 "./common/Object.h"
            // unreachable; //@line 106 "./common/Object.h"
           }
           else if ((label|0) == 512) {
            $$pre$i346 = HEAP32[$43>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $1612 = $$pre$i346;
            break;
           }
          }
         } while(0);
         __ZdlPv($1612); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        }
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($42); //@line 376 "modules/event/sdl/Event.cpp"
        $$11 = $1593;$$1114 = $1594;
        break;
       } else {
        $1573 = HEAP32[$43>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $1574 = ($1573|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($1574)) {
         $1575 = HEAP32[$1569>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $1576 = ($1575|0)==($1573|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         do {
          if ($1576) {
           $1590 = $1573;
          } else {
           $1578 = $1575;
           while(1) {
            $1577 = ((($1578)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            HEAP32[$1569>>2] = $1577; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $1579 = HEAP32[$1577>>2]|0; //@line 106 "./common/Object.h"
            $1580 = ($1579|0)==(0|0); //@line 106 "./common/Object.h"
            if ($1580) {
             $1582 = $1577;
            } else {
             $1583 = HEAP32[$1579>>2]|0; //@line 106 "./common/Object.h"
             $1584 = ((($1583)) + 12|0); //@line 106 "./common/Object.h"
             $1585 = HEAP32[$1584>>2]|0; //@line 106 "./common/Object.h"
             __THREW__ = 0;
             invoke_vi($1585|0,($1579|0)); //@line 106 "./common/Object.h"
             $1586 = __THREW__; __THREW__ = 0;
             $1587 = $1586&1;
             if ($1587) {
              label = 500;
              break;
             }
             $$pre$i$i$i334 = HEAP32[$1569>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $1582 = $$pre$i$i$i334;
            }
            $1581 = ($1582|0)==($1573|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            if ($1581) {
             label = 501;
             break;
            } else {
             $1578 = $1582;
            }
           }
           if ((label|0) == 500) {
            $1588 = ___cxa_find_matching_catch_3(0|0)|0;
            $1589 = tempRet0;
            ___clang_call_terminate($1588); //@line 106 "./common/Object.h"
            // unreachable; //@line 106 "./common/Object.h"
           }
           else if ((label|0) == 501) {
            $$pre$i337 = HEAP32[$43>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $1590 = $$pre$i337;
            break;
           }
          }
         } while(0);
         __ZdlPv($1590); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        }
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($42); //@line 376 "modules/event/sdl/Event.cpp"
        $msg$1 = $1564;
        break L3;
       }
      }
     } while(0);
     __ZdlPv($1564); //@line 376 "modules/event/sdl/Event.cpp"
     $$13 = $$11;$$1316 = $$1114;
     break L1;
     break;
    }
    case 258:  {
     __THREW__ = 0;
     $1613 = (invoke_ii(3348,32)|0); //@line 379 "modules/event/sdl/Event.cpp"
     $1614 = __THREW__; __THREW__ = 0;
     $1615 = $1614&1;
     if ($1615) {
      label = 4;
      break L1;
     }
     __THREW__ = 0;
     invoke_viii(3460,($44|0),(365860|0),9); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $1616 = __THREW__; __THREW__ = 0;
     $1617 = $1616&1;
     do {
      if ($1617) {
       $1640 = ___cxa_find_matching_catch_2()|0;
       $1641 = tempRet0;
       $$12 = $1640;$$1215 = $1641;
      } else {
       HEAP32[$45>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $1618 = ((($45)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$1618>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $1619 = ((($45)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$1619>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       __THREW__ = 0;
       invoke_viii(3508,($1613|0),($44|0),($45|0)); //@line 379 "modules/event/sdl/Event.cpp"
       $1620 = __THREW__; __THREW__ = 0;
       $1621 = $1620&1;
       if ($1621) {
        $1642 = ___cxa_find_matching_catch_2()|0;
        $1643 = tempRet0;
        $1644 = HEAP32[$45>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $1645 = ($1644|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($1645)) {
         $1646 = HEAP32[$1618>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $1647 = ($1646|0)==($1644|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         do {
          if ($1647) {
           $1661 = $1644;
          } else {
           $1649 = $1646;
           while(1) {
            $1648 = ((($1649)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            HEAP32[$1618>>2] = $1648; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $1650 = HEAP32[$1648>>2]|0; //@line 106 "./common/Object.h"
            $1651 = ($1650|0)==(0|0); //@line 106 "./common/Object.h"
            if ($1651) {
             $1653 = $1648;
            } else {
             $1654 = HEAP32[$1650>>2]|0; //@line 106 "./common/Object.h"
             $1655 = ((($1654)) + 12|0); //@line 106 "./common/Object.h"
             $1656 = HEAP32[$1655>>2]|0; //@line 106 "./common/Object.h"
             __THREW__ = 0;
             invoke_vi($1656|0,($1650|0)); //@line 106 "./common/Object.h"
             $1657 = __THREW__; __THREW__ = 0;
             $1658 = $1657&1;
             if ($1658) {
              label = 536;
              break;
             }
             $$pre$i$i$i29 = HEAP32[$1618>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $1653 = $$pre$i$i$i29;
            }
            $1652 = ($1653|0)==($1644|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            if ($1652) {
             label = 537;
             break;
            } else {
             $1649 = $1653;
            }
           }
           if ((label|0) == 536) {
            $1659 = ___cxa_find_matching_catch_3(0|0)|0;
            $1660 = tempRet0;
            ___clang_call_terminate($1659); //@line 106 "./common/Object.h"
            // unreachable; //@line 106 "./common/Object.h"
           }
           else if ((label|0) == 537) {
            $$pre$i32 = HEAP32[$45>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $1661 = $$pre$i32;
            break;
           }
          }
         } while(0);
         __ZdlPv($1661); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        }
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($44); //@line 379 "modules/event/sdl/Event.cpp"
        $$12 = $1642;$$1215 = $1643;
        break;
       } else {
        $1622 = HEAP32[$45>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $1623 = ($1622|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($1623)) {
         $1624 = HEAP32[$1618>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $1625 = ($1624|0)==($1622|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         do {
          if ($1625) {
           $1639 = $1622;
          } else {
           $1627 = $1624;
           while(1) {
            $1626 = ((($1627)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            HEAP32[$1618>>2] = $1626; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $1628 = HEAP32[$1626>>2]|0; //@line 106 "./common/Object.h"
            $1629 = ($1628|0)==(0|0); //@line 106 "./common/Object.h"
            if ($1629) {
             $1631 = $1626;
            } else {
             $1632 = HEAP32[$1628>>2]|0; //@line 106 "./common/Object.h"
             $1633 = ((($1632)) + 12|0); //@line 106 "./common/Object.h"
             $1634 = HEAP32[$1633>>2]|0; //@line 106 "./common/Object.h"
             __THREW__ = 0;
             invoke_vi($1634|0,($1628|0)); //@line 106 "./common/Object.h"
             $1635 = __THREW__; __THREW__ = 0;
             $1636 = $1635&1;
             if ($1636) {
              label = 525;
              break;
             }
             $$pre$i$i$i354 = HEAP32[$1618>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $1631 = $$pre$i$i$i354;
            }
            $1630 = ($1631|0)==($1622|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            if ($1630) {
             label = 526;
             break;
            } else {
             $1627 = $1631;
            }
           }
           if ((label|0) == 525) {
            $1637 = ___cxa_find_matching_catch_3(0|0)|0;
            $1638 = tempRet0;
            ___clang_call_terminate($1637); //@line 106 "./common/Object.h"
            // unreachable; //@line 106 "./common/Object.h"
           }
           else if ((label|0) == 526) {
            $$pre$i357 = HEAP32[$45>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $1639 = $$pre$i357;
            break;
           }
          }
         } while(0);
         __ZdlPv($1639); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        }
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($44); //@line 379 "modules/event/sdl/Event.cpp"
        $msg$1 = $1613;
        break L3;
       }
      }
     } while(0);
     __ZdlPv($1613); //@line 379 "modules/event/sdl/Event.cpp"
     $$13 = $$12;$$1316 = $$1215;
     break L1;
     break;
    }
    default: {
     $msg$1 = 0;
    }
    }
   } while(0);
   $1662 = HEAP32[$vargs>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $1663 = HEAP32[$46>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $1664 = ($1662|0)==($1663|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   do {
    if ($1664) {
     $1666 = $1662;
    } else {
     $__begin$sroa$0$0378 = $1662;
     while(1) {
      $1683 = HEAP32[$__begin$sroa$0$0378>>2]|0; //@line 139 "./common/Object.h"
      $1684 = ($1683|0)==(0|0); //@line 388 "modules/event/sdl/Event.cpp"
      if (!($1684)) {
       $1685 = HEAP32[$1683>>2]|0; //@line 389 "modules/event/sdl/Event.cpp"
       $1686 = ((($1685)) + 12|0); //@line 389 "modules/event/sdl/Event.cpp"
       $1687 = HEAP32[$1686>>2]|0; //@line 389 "modules/event/sdl/Event.cpp"
       __THREW__ = 0;
       invoke_vi($1687|0,($1683|0)); //@line 389 "modules/event/sdl/Event.cpp"
       $1688 = __THREW__; __THREW__ = 0;
       $1689 = $1688&1;
       if ($1689) {
        label = 3;
        break;
       }
      }
      $1690 = ((($__begin$sroa$0$0378)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $1691 = ($1690|0)==($1663|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($1691) {
       label = 542;
       break;
      } else {
       $__begin$sroa$0$0378 = $1690;
      }
     }
     if ((label|0) == 3) {
      $51 = ___cxa_find_matching_catch_2()|0;
      $52 = tempRet0;
      $lpad$phi$index = $51;$lpad$phi$index2 = $52;
      label = 5;
      break L1;
     }
     else if ((label|0) == 542) {
      $$pre393 = HEAP32[$vargs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $1666 = $$pre393;
      break;
     }
    }
   } while(0);
   $1665 = ($1666|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($1665) {
    STACKTOP = sp;return ($msg$1|0); //@line 393 "modules/event/sdl/Event.cpp"
   }
   $1667 = HEAP32[$46>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $1668 = ($1667|0)==($1666|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   do {
    if ($1668) {
     $1682 = $1666;
    } else {
     $1670 = $1667;
     while(1) {
      $1669 = ((($1670)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$46>>2] = $1669; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $1671 = HEAP32[$1669>>2]|0; //@line 106 "./common/Object.h"
      $1672 = ($1671|0)==(0|0); //@line 106 "./common/Object.h"
      if ($1672) {
       $1674 = $1669;
      } else {
       $1675 = HEAP32[$1671>>2]|0; //@line 106 "./common/Object.h"
       $1676 = ((($1675)) + 12|0); //@line 106 "./common/Object.h"
       $1677 = HEAP32[$1676>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($1677|0,($1671|0)); //@line 106 "./common/Object.h"
       $1678 = __THREW__; __THREW__ = 0;
       $1679 = $1678&1;
       if ($1679) {
        label = 549;
        break;
       }
       $$pre$i$i$i20 = HEAP32[$46>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $1674 = $$pre$i$i$i20;
      }
      $1673 = ($1674|0)==($1666|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($1673) {
       label = 550;
       break;
      } else {
       $1670 = $1674;
      }
     }
     if ((label|0) == 549) {
      $1680 = ___cxa_find_matching_catch_3(0|0)|0;
      $1681 = tempRet0;
      ___clang_call_terminate($1680); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     else if ((label|0) == 550) {
      $$pre$i23 = HEAP32[$vargs>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $1682 = $$pre$i23;
      break;
     }
    }
   } while(0);
   __ZdlPv($1682); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   STACKTOP = sp;return ($msg$1|0); //@line 393 "modules/event/sdl/Event.cpp"
  }
 } while(0);
 if ((label|0) == 4) {
  $53 = ___cxa_find_matching_catch_2()|0;
  $54 = tempRet0;
  $lpad$phi$index = $53;$lpad$phi$index2 = $54;
  label = 5;
 }
 if ((label|0) == 5) {
  $$13 = $lpad$phi$index;$$1316 = $lpad$phi$index2;
 }
 $1692 = HEAP32[$vargs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1693 = ($1692|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($1693) {
  ___resumeException($$13|0);
  // unreachable;
 }
 $1694 = HEAP32[$46>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1695 = ($1694|0)==($1692|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($1695) {
   $1709 = $1692;
  } else {
   $1697 = $1694;
   while(1) {
    $1696 = ((($1697)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$46>>2] = $1696; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $1698 = HEAP32[$1696>>2]|0; //@line 106 "./common/Object.h"
    $1699 = ($1698|0)==(0|0); //@line 106 "./common/Object.h"
    if ($1699) {
     $1701 = $1696;
    } else {
     $1702 = HEAP32[$1698>>2]|0; //@line 106 "./common/Object.h"
     $1703 = ((($1702)) + 12|0); //@line 106 "./common/Object.h"
     $1704 = HEAP32[$1703>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($1704|0,($1698|0)); //@line 106 "./common/Object.h"
     $1705 = __THREW__; __THREW__ = 0;
     $1706 = $1705&1;
     if ($1706) {
      label = 562;
      break;
     }
     $$pre$i$i$i = HEAP32[$46>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $1701 = $$pre$i$i$i;
    }
    $1700 = ($1701|0)==($1692|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($1700) {
     label = 563;
     break;
    } else {
     $1697 = $1701;
    }
   }
   if ((label|0) == 562) {
    $1707 = ___cxa_find_matching_catch_3(0|0)|0;
    $1708 = tempRet0;
    ___clang_call_terminate($1707); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 563) {
    $$pre$i = HEAP32[$vargs>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $1709 = $$pre$i;
    break;
   }
  }
 } while(0);
 __ZdlPv($1709); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$13|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__16vectorIN4love9StrongRefINS1_7VariantEEENS_9allocatorIS4_EEE7reserveEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$01$i$i = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $8 = 0, $9 = 0, $__v$sroa$6$0 = 0, $__v$sroa$6$0$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($4>>>0)<($__n>>>0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = $2; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($5)) {
  return; //@line 1545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $7 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = HEAP32[$7>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($__n|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $12 = $8; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $16 = 0;
 } else {
  $13 = $__n << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $14 = (__Znwj($13)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $16 = $14;
 }
 $15 = (($16) + ($10<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $17 = $15; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $18 = (($16) + ($__n<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $19 = $18; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $20 = ($12|0)==($6|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 do {
  if ($20) {
   $35 = $17;$37 = $2;$39 = $6;
  } else {
   $$01$i$i = $12;$23 = $15;$__v$sroa$6$0 = $17;
   while(1) {
    $21 = ((($$01$i$i)) + -4|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $22 = ((($23)) + -4|0); //@line 99 "./common/Object.h"
    $24 = HEAP32[$21>>2]|0; //@line 139 "./common/Object.h"
    HEAP32[$22>>2] = $24; //@line 99 "./common/Object.h"
    $25 = ($24|0)==(0|0); //@line 101 "./common/Object.h"
    if (!($25)) {
     $26 = HEAP32[$24>>2]|0; //@line 101 "./common/Object.h"
     $27 = ((($26)) + 8|0); //@line 101 "./common/Object.h"
     $28 = HEAP32[$27>>2]|0; //@line 101 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($28|0,($24|0)); //@line 101 "./common/Object.h"
     $29 = __THREW__; __THREW__ = 0;
     $30 = $29&1;
     if ($30) {
      $__v$sroa$6$0$lcssa = $__v$sroa$6$0;
      break;
     }
    }
    $31 = $__v$sroa$6$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $32 = ((($31)) + -4|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $33 = $32; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $34 = ($21|0)==($6|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    if ($34) {
     $$lcssa = $33;
     label = 8;
     break;
    } else {
     $$01$i$i = $21;$23 = $32;$__v$sroa$6$0 = $33;
    }
   }
   if ((label|0) == 8) {
    $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $$pre = HEAP32[$7>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $35 = $$lcssa;$37 = $$pre$i;$39 = $$pre;
    break;
   }
   $54 = ___cxa_find_matching_catch_2()|0;
   $55 = tempRet0;
   $56 = $__v$sroa$6$0$lcssa; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $57 = ($15|0)==($56|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   L16: do {
    if (!($57)) {
     $59 = $15;
     while(1) {
      $58 = ((($59)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $60 = HEAP32[$58>>2]|0; //@line 106 "./common/Object.h"
      $61 = ($60|0)==(0|0); //@line 106 "./common/Object.h"
      if (!($61)) {
       $63 = HEAP32[$60>>2]|0; //@line 106 "./common/Object.h"
       $64 = ((($63)) + 12|0); //@line 106 "./common/Object.h"
       $65 = HEAP32[$64>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($65|0,($60|0)); //@line 106 "./common/Object.h"
       $66 = __THREW__; __THREW__ = 0;
       $67 = $66&1;
       if ($67) {
        break;
       }
      }
      $62 = ($58|0)==($56|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if ($62) {
       break L16;
      } else {
       $59 = $58;
      }
     }
     $68 = ___cxa_find_matching_catch_3(0|0)|0;
     $69 = tempRet0;
     ___clang_call_terminate($68); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   } while(0);
   $70 = ($16|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($70) {
    ___resumeException($54|0);
    // unreachable;
   }
   __ZdlPv($16); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($54|0);
   // unreachable;
  }
 } while(0);
 HEAP32[$this>>2] = $35; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $17; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $19; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $36 = $37; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $38 = ($39|0)==($36|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 L29: do {
  if (!($38)) {
   $41 = $39;
   while(1) {
    $40 = ((($41)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $42 = HEAP32[$40>>2]|0; //@line 106 "./common/Object.h"
    $43 = ($42|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($43)) {
     $45 = HEAP32[$42>>2]|0; //@line 106 "./common/Object.h"
     $46 = ((($45)) + 12|0); //@line 106 "./common/Object.h"
     $47 = HEAP32[$46>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($47|0,($42|0)); //@line 106 "./common/Object.h"
     $48 = __THREW__; __THREW__ = 0;
     $49 = $48&1;
     if ($49) {
      break;
     }
    }
    $44 = ($40|0)==($36|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($44) {
     break L29;
    } else {
     $41 = $40;
    }
   }
   $50 = ___cxa_find_matching_catch_3(0|0)|0;
   $51 = tempRet0;
   ___clang_call_terminate($50); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 } while(0);
 $52 = ($37|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($52) {
  return; //@line 1545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $53 = $37; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($53); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNK4love5event3sdl5Event20convertJoystickEventERK9SDL_Event($this,$e) {
 $this = $this|0;
 $e = $e|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $$1 = 0, $$13 = 0, $$2 = 0, $$24 = 0, $$3 = 0, $$35 = 0, $$4 = 0, $$46 = 0, $$5 = 0, $$57 = 0, $$6 = 0, $$68 = 0, $$7 = 0, $$79 = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i$i = 0;
 var $$pre$i$i$i154 = 0, $$pre$i10 = 0, $$pre$i103 = 0, $$pre$i111 = 0, $$pre$i119 = 0, $$pre$i132 = 0, $$pre$i14 = 0, $$pre$i143 = 0, $$pre$i157 = 0, $$pre$i25 = 0, $$pre$i33 = 0, $$pre$i41 = 0, $$pre$i54 = 0, $$pre$i62 = 0, $$pre$i70 = 0, $$pre$i83 = 0, $$pre$i91 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0;
 var $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0;
 var $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0;
 var $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0;
 var $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0.0;
 var $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0;
 var $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0.0, $219 = 0.0, $22 = 0, $220 = 0.0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0.0;
 var $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0;
 var $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0;
 var $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0;
 var $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0;
 var $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0;
 var $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0;
 var $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0.0;
 var $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0;
 var $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0;
 var $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0;
 var $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0;
 var $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0;
 var $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0;
 var $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0;
 var $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0;
 var $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0;
 var $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0;
 var $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0;
 var $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0;
 var $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0;
 var $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0;
 var $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0;
 var $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0;
 var $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0.0, $652 = 0.0, $653 = 0.0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0.0;
 var $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0;
 var $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0;
 var $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0;
 var $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0;
 var $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0;
 var $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0;
 var $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0;
 var $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0.0, $800 = 0, $801 = 0, $802 = 0;
 var $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0;
 var $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0;
 var $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0;
 var $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0;
 var $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0;
 var $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $__begin$sroa$0$0161 = 0, $hat = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $msg$0 = 0, $padaxis = 0, $padbutton = 0, $proxy = 0, $txt = 0, $vargs = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vargs = sp + 168|0;
 $proxy = sp + 160|0;
 $hat = sp + 156|0;
 $padbutton = sp + 152|0;
 $padaxis = sp + 148|0;
 $txt = sp + 144|0;
 $0 = sp + 140|0;
 $1 = sp + 136|0;
 $2 = sp + 124|0;
 $3 = sp + 120|0;
 $4 = sp + 116|0;
 $5 = sp + 112|0;
 $6 = sp + 100|0;
 $7 = sp + 96|0;
 $8 = sp + 92|0;
 $9 = sp + 88|0;
 $10 = sp + 76|0;
 $11 = sp + 72|0;
 $12 = sp + 68|0;
 $13 = sp + 56|0;
 $14 = sp + 52|0;
 $15 = sp + 48|0;
 $16 = sp + 44|0;
 $17 = sp + 32|0;
 $18 = sp + 28|0;
 $19 = sp + 16|0;
 $20 = sp + 12|0;
 $21 = sp;
 $22 = HEAP32[(625632)>>2]|0; //@line 98 "./common/Module.h"
 $23 = ($22|0)==(0|0); //@line 398 "modules/event/sdl/Event.cpp"
 if ($23) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 525 "modules/event/sdl/Event.cpp"
 }
 HEAP32[$vargs>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $24 = ((($vargs)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$24>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $25 = ((($vargs)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$25>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 invoke_vii(3936,($vargs|0),4); //@line 404 "modules/event/sdl/Event.cpp"
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 L4: do {
  if ($27) {
   label = 5;
  } else {
   $28 = HEAP32[$e>>2]|0; //@line 412 "modules/event/sdl/Event.cpp"
   L6: do {
    switch ($28|0) {
    case 1540: case 1539:  {
     HEAP32[$proxy>>2] = 24; //@line 416 "modules/event/sdl/Event.cpp"
     $33 = HEAP32[$22>>2]|0; //@line 417 "modules/event/sdl/Event.cpp"
     $34 = ((($33)) + 32|0); //@line 417 "modules/event/sdl/Event.cpp"
     $35 = HEAP32[$34>>2]|0; //@line 417 "modules/event/sdl/Event.cpp"
     $36 = ((($e)) + 8|0); //@line 417 "modules/event/sdl/Event.cpp"
     $37 = HEAP32[$36>>2]|0; //@line 417 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     $38 = (invoke_iii($35|0,($22|0),($37|0))|0); //@line 417 "modules/event/sdl/Event.cpp"
     $39 = __THREW__; __THREW__ = 0;
     $40 = $39&1;
     if ($40) {
      label = 5;
      break L4;
     }
     $41 = ((($proxy)) + 4|0); //@line 417 "modules/event/sdl/Event.cpp"
     HEAP32[$41>>2] = $38; //@line 417 "modules/event/sdl/Event.cpp"
     $42 = ($38|0)==(0|0); //@line 418 "modules/event/sdl/Event.cpp"
     if ($42) {
      $msg$0 = 0;
     } else {
      __THREW__ = 0;
      $43 = (invoke_ii(3348,32)|0); //@line 421 "modules/event/sdl/Event.cpp"
      $44 = __THREW__; __THREW__ = 0;
      $45 = $44&1;
      if ($45) {
       label = 5;
       break L4;
      }
      $46 = HEAP32[$proxy>>2]|0; //@line 421 "modules/event/sdl/Event.cpp"
      __THREW__ = 0;
      invoke_viii(3949,($43|0),($46|0),($proxy|0)); //@line 421 "modules/event/sdl/Event.cpp"
      $47 = __THREW__; __THREW__ = 0;
      $48 = $47&1;
      if ($48) {
       $117 = ___cxa_find_matching_catch_2()|0;
       $118 = tempRet0;
       __ZdlPv($43); //@line 421 "modules/event/sdl/Event.cpp"
       $$7 = $117;$$79 = $118;
       break L4;
      }
      HEAP32[$0>>2] = $43; //@line 93 "./common/Object.h"
      $49 = HEAP32[$43>>2]|0; //@line 95 "./common/Object.h"
      $50 = ((($49)) + 8|0); //@line 95 "./common/Object.h"
      $51 = HEAP32[$50>>2]|0; //@line 95 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($51|0,($43|0)); //@line 95 "./common/Object.h"
      $52 = __THREW__; __THREW__ = 0;
      $53 = $52&1;
      if ($53) {
       label = 5;
       break L4;
      }
      $54 = HEAP32[$24>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $55 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $56 = HEAP32[$55>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $57 = ($54>>>0)<($56>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($57) {
       HEAP32[$54>>2] = $43; //@line 99 "./common/Object.h"
       $58 = HEAP32[$43>>2]|0; //@line 101 "./common/Object.h"
       $59 = ((($58)) + 8|0); //@line 101 "./common/Object.h"
       $60 = HEAP32[$59>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($60|0,($43|0)); //@line 101 "./common/Object.h"
       $61 = __THREW__; __THREW__ = 0;
       $62 = $61&1;
       if ($62) {
        label = 32;
       } else {
        $$pre$i10 = HEAP32[$24>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $63 = ((($$pre$i10)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$24>>2] = $63; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
      } else {
       __THREW__ = 0;
       invoke_vii(3507,($vargs|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $64 = __THREW__; __THREW__ = 0;
       $65 = $64&1;
       if ($65) {
        label = 32;
       }
      }
      if ((label|0) == 32) {
       $119 = ___cxa_find_matching_catch_2()|0;
       $120 = tempRet0;
       $121 = HEAP32[$43>>2]|0; //@line 106 "./common/Object.h"
       $122 = ((($121)) + 12|0); //@line 106 "./common/Object.h"
       $123 = HEAP32[$122>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($123|0,($43|0)); //@line 106 "./common/Object.h"
       $124 = __THREW__; __THREW__ = 0;
       $125 = $124&1;
       if (!($125)) {
        $$7 = $119;$$79 = $120;
        break L4;
       }
       $126 = ___cxa_find_matching_catch_3(0|0)|0;
       $127 = tempRet0;
       ___clang_call_terminate($126); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      $66 = HEAP32[$43>>2]|0; //@line 106 "./common/Object.h"
      $67 = ((($66)) + 12|0); //@line 106 "./common/Object.h"
      $68 = HEAP32[$67>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($68|0,($43|0)); //@line 106 "./common/Object.h"
      $69 = __THREW__; __THREW__ = 0;
      $70 = $69&1;
      if ($70) {
       $71 = ___cxa_find_matching_catch_3(0|0)|0;
       $72 = tempRet0;
       ___clang_call_terminate($71); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      __THREW__ = 0;
      $73 = (invoke_ii(3348,32)|0); //@line 422 "modules/event/sdl/Event.cpp"
      $74 = __THREW__; __THREW__ = 0;
      $75 = $74&1;
      if ($75) {
       label = 5;
       break L4;
      }
      $76 = ((($e)) + 12|0); //@line 422 "modules/event/sdl/Event.cpp"
      $77 = HEAP8[$76>>0]|0; //@line 422 "modules/event/sdl/Event.cpp"
      $78 = $77&255; //@line 422 "modules/event/sdl/Event.cpp"
      $79 = (($78) + 1)|0; //@line 422 "modules/event/sdl/Event.cpp"
      $80 = (+($79|0)); //@line 422 "modules/event/sdl/Event.cpp"
      __THREW__ = 0;
      invoke_vid(3942,($73|0),(+$80)); //@line 422 "modules/event/sdl/Event.cpp"
      $81 = __THREW__; __THREW__ = 0;
      $82 = $81&1;
      if ($82) {
       $128 = ___cxa_find_matching_catch_2()|0;
       $129 = tempRet0;
       __ZdlPv($73); //@line 422 "modules/event/sdl/Event.cpp"
       $$7 = $128;$$79 = $129;
       break L4;
      }
      HEAP32[$1>>2] = $73; //@line 93 "./common/Object.h"
      $83 = HEAP32[$73>>2]|0; //@line 95 "./common/Object.h"
      $84 = ((($83)) + 8|0); //@line 95 "./common/Object.h"
      $85 = HEAP32[$84>>2]|0; //@line 95 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($85|0,($73|0)); //@line 95 "./common/Object.h"
      $86 = __THREW__; __THREW__ = 0;
      $87 = $86&1;
      if ($87) {
       label = 5;
       break L4;
      }
      $88 = HEAP32[$24>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $89 = HEAP32[$55>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $90 = ($88>>>0)<($89>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($90) {
       HEAP32[$88>>2] = $73; //@line 99 "./common/Object.h"
       $91 = HEAP32[$73>>2]|0; //@line 101 "./common/Object.h"
       $92 = ((($91)) + 8|0); //@line 101 "./common/Object.h"
       $93 = HEAP32[$92>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($93|0,($73|0)); //@line 101 "./common/Object.h"
       $94 = __THREW__; __THREW__ = 0;
       $95 = $94&1;
       if ($95) {
        label = 35;
       } else {
        $$pre$i14 = HEAP32[$24>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $96 = ((($$pre$i14)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$24>>2] = $96; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
      } else {
       __THREW__ = 0;
       invoke_vii(3507,($vargs|0),($1|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $97 = __THREW__; __THREW__ = 0;
       $98 = $97&1;
       if ($98) {
        label = 35;
       }
      }
      if ((label|0) == 35) {
       $130 = ___cxa_find_matching_catch_2()|0;
       $131 = tempRet0;
       $132 = HEAP32[$73>>2]|0; //@line 106 "./common/Object.h"
       $133 = ((($132)) + 12|0); //@line 106 "./common/Object.h"
       $134 = HEAP32[$133>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($134|0,($73|0)); //@line 106 "./common/Object.h"
       $135 = __THREW__; __THREW__ = 0;
       $136 = $135&1;
       if (!($136)) {
        $$7 = $130;$$79 = $131;
        break L4;
       }
       $137 = ___cxa_find_matching_catch_3(0|0)|0;
       $138 = tempRet0;
       ___clang_call_terminate($137); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      $99 = HEAP32[$73>>2]|0; //@line 106 "./common/Object.h"
      $100 = ((($99)) + 12|0); //@line 106 "./common/Object.h"
      $101 = HEAP32[$100>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($101|0,($73|0)); //@line 106 "./common/Object.h"
      $102 = __THREW__; __THREW__ = 0;
      $103 = $102&1;
      if ($103) {
       $104 = ___cxa_find_matching_catch_3(0|0)|0;
       $105 = tempRet0;
       ___clang_call_terminate($104); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      __THREW__ = 0;
      $106 = (invoke_ii(3348,32)|0); //@line 423 "modules/event/sdl/Event.cpp"
      $107 = __THREW__; __THREW__ = 0;
      $108 = $107&1;
      if ($108) {
       label = 5;
       break L4;
      }
      $109 = HEAP32[$e>>2]|0; //@line 423 "modules/event/sdl/Event.cpp"
      $110 = ($109|0)==(1539); //@line 423 "modules/event/sdl/Event.cpp"
      $111 = $110 ? 365663 : 365679; //@line 423 "modules/event/sdl/Event.cpp"
      $112 = $110 ? 15 : 16; //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __THREW__ = 0;
      invoke_viii(3460,($2|0),($111|0),($112|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $113 = __THREW__; __THREW__ = 0;
      $114 = $113&1;
      do {
       if ($114) {
        $139 = ___cxa_find_matching_catch_2()|0;
        $140 = tempRet0;
        $$01 = $139;$$02 = $140;
       } else {
        __THREW__ = 0;
        invoke_viii(3508,($106|0),($2|0),($vargs|0)); //@line 423 "modules/event/sdl/Event.cpp"
        $115 = __THREW__; __THREW__ = 0;
        $116 = $115&1;
        if ($116) {
         $141 = ___cxa_find_matching_catch_2()|0;
         $142 = tempRet0;
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 423 "modules/event/sdl/Event.cpp"
         $$01 = $141;$$02 = $142;
         break;
        } else {
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 423 "modules/event/sdl/Event.cpp"
         $msg$0 = $106;
         break L6;
        }
       }
      } while(0);
      __ZdlPv($106); //@line 423 "modules/event/sdl/Event.cpp"
      $$7 = $$01;$$79 = $$02;
      break L4;
     }
     break;
    }
    case 1536:  {
     HEAP32[$proxy>>2] = 24; //@line 429 "modules/event/sdl/Event.cpp"
     $143 = HEAP32[$22>>2]|0; //@line 430 "modules/event/sdl/Event.cpp"
     $144 = ((($143)) + 32|0); //@line 430 "modules/event/sdl/Event.cpp"
     $145 = HEAP32[$144>>2]|0; //@line 430 "modules/event/sdl/Event.cpp"
     $146 = ((($e)) + 8|0); //@line 430 "modules/event/sdl/Event.cpp"
     $147 = HEAP32[$146>>2]|0; //@line 430 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     $148 = (invoke_iii($145|0,($22|0),($147|0))|0); //@line 430 "modules/event/sdl/Event.cpp"
     $149 = __THREW__; __THREW__ = 0;
     $150 = $149&1;
     if ($150) {
      label = 5;
      break L4;
     }
     $151 = ((($proxy)) + 4|0); //@line 430 "modules/event/sdl/Event.cpp"
     HEAP32[$151>>2] = $148; //@line 430 "modules/event/sdl/Event.cpp"
     $152 = ($148|0)==(0|0); //@line 431 "modules/event/sdl/Event.cpp"
     if ($152) {
      $msg$0 = 0;
     } else {
      __THREW__ = 0;
      $153 = (invoke_ii(3348,32)|0); //@line 434 "modules/event/sdl/Event.cpp"
      $154 = __THREW__; __THREW__ = 0;
      $155 = $154&1;
      if ($155) {
       label = 5;
       break L4;
      }
      $156 = HEAP32[$proxy>>2]|0; //@line 434 "modules/event/sdl/Event.cpp"
      __THREW__ = 0;
      invoke_viii(3949,($153|0),($156|0),($proxy|0)); //@line 434 "modules/event/sdl/Event.cpp"
      $157 = __THREW__; __THREW__ = 0;
      $158 = $157&1;
      if ($158) {
       $259 = ___cxa_find_matching_catch_2()|0;
       $260 = tempRet0;
       __ZdlPv($153); //@line 434 "modules/event/sdl/Event.cpp"
       $$7 = $259;$$79 = $260;
       break L4;
      }
      HEAP32[$3>>2] = $153; //@line 93 "./common/Object.h"
      $159 = HEAP32[$153>>2]|0; //@line 95 "./common/Object.h"
      $160 = ((($159)) + 8|0); //@line 95 "./common/Object.h"
      $161 = HEAP32[$160>>2]|0; //@line 95 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($161|0,($153|0)); //@line 95 "./common/Object.h"
      $162 = __THREW__; __THREW__ = 0;
      $163 = $162&1;
      if ($163) {
       label = 5;
       break L4;
      }
      $164 = HEAP32[$24>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $165 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $166 = HEAP32[$165>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $167 = ($164>>>0)<($166>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($167) {
       HEAP32[$164>>2] = $153; //@line 99 "./common/Object.h"
       $168 = HEAP32[$153>>2]|0; //@line 101 "./common/Object.h"
       $169 = ((($168)) + 8|0); //@line 101 "./common/Object.h"
       $170 = HEAP32[$169>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($170|0,($153|0)); //@line 101 "./common/Object.h"
       $171 = __THREW__; __THREW__ = 0;
       $172 = $171&1;
       if ($172) {
        label = 75;
       } else {
        $$pre$i25 = HEAP32[$24>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $173 = ((($$pre$i25)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$24>>2] = $173; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
      } else {
       __THREW__ = 0;
       invoke_vii(3507,($vargs|0),($3|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $174 = __THREW__; __THREW__ = 0;
       $175 = $174&1;
       if ($175) {
        label = 75;
       }
      }
      if ((label|0) == 75) {
       $261 = ___cxa_find_matching_catch_2()|0;
       $262 = tempRet0;
       $263 = HEAP32[$153>>2]|0; //@line 106 "./common/Object.h"
       $264 = ((($263)) + 12|0); //@line 106 "./common/Object.h"
       $265 = HEAP32[$264>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($265|0,($153|0)); //@line 106 "./common/Object.h"
       $266 = __THREW__; __THREW__ = 0;
       $267 = $266&1;
       if (!($267)) {
        $$7 = $261;$$79 = $262;
        break L4;
       }
       $268 = ___cxa_find_matching_catch_3(0|0)|0;
       $269 = tempRet0;
       ___clang_call_terminate($268); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      $176 = HEAP32[$153>>2]|0; //@line 106 "./common/Object.h"
      $177 = ((($176)) + 12|0); //@line 106 "./common/Object.h"
      $178 = HEAP32[$177>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($178|0,($153|0)); //@line 106 "./common/Object.h"
      $179 = __THREW__; __THREW__ = 0;
      $180 = $179&1;
      if ($180) {
       $181 = ___cxa_find_matching_catch_3(0|0)|0;
       $182 = tempRet0;
       ___clang_call_terminate($181); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      __THREW__ = 0;
      $183 = (invoke_ii(3348,32)|0); //@line 435 "modules/event/sdl/Event.cpp"
      $184 = __THREW__; __THREW__ = 0;
      $185 = $184&1;
      if ($185) {
       label = 5;
       break L4;
      }
      $186 = ((($e)) + 12|0); //@line 435 "modules/event/sdl/Event.cpp"
      $187 = HEAP8[$186>>0]|0; //@line 435 "modules/event/sdl/Event.cpp"
      $188 = $187&255; //@line 435 "modules/event/sdl/Event.cpp"
      $189 = (($188) + 1)|0; //@line 435 "modules/event/sdl/Event.cpp"
      $190 = (+($189|0)); //@line 435 "modules/event/sdl/Event.cpp"
      __THREW__ = 0;
      invoke_vid(3942,($183|0),(+$190)); //@line 435 "modules/event/sdl/Event.cpp"
      $191 = __THREW__; __THREW__ = 0;
      $192 = $191&1;
      if ($192) {
       $270 = ___cxa_find_matching_catch_2()|0;
       $271 = tempRet0;
       __ZdlPv($183); //@line 435 "modules/event/sdl/Event.cpp"
       $$7 = $270;$$79 = $271;
       break L4;
      }
      HEAP32[$4>>2] = $183; //@line 93 "./common/Object.h"
      $193 = HEAP32[$183>>2]|0; //@line 95 "./common/Object.h"
      $194 = ((($193)) + 8|0); //@line 95 "./common/Object.h"
      $195 = HEAP32[$194>>2]|0; //@line 95 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($195|0,($183|0)); //@line 95 "./common/Object.h"
      $196 = __THREW__; __THREW__ = 0;
      $197 = $196&1;
      if ($197) {
       label = 5;
       break L4;
      }
      $198 = HEAP32[$24>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $199 = HEAP32[$165>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $200 = ($198>>>0)<($199>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($200) {
       HEAP32[$198>>2] = $183; //@line 99 "./common/Object.h"
       $201 = HEAP32[$183>>2]|0; //@line 101 "./common/Object.h"
       $202 = ((($201)) + 8|0); //@line 101 "./common/Object.h"
       $203 = HEAP32[$202>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($203|0,($183|0)); //@line 101 "./common/Object.h"
       $204 = __THREW__; __THREW__ = 0;
       $205 = $204&1;
       if ($205) {
        label = 78;
       } else {
        $$pre$i33 = HEAP32[$24>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $206 = ((($$pre$i33)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$24>>2] = $206; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
      } else {
       __THREW__ = 0;
       invoke_vii(3507,($vargs|0),($4|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $207 = __THREW__; __THREW__ = 0;
       $208 = $207&1;
       if ($208) {
        label = 78;
       }
      }
      if ((label|0) == 78) {
       $272 = ___cxa_find_matching_catch_2()|0;
       $273 = tempRet0;
       $274 = HEAP32[$183>>2]|0; //@line 106 "./common/Object.h"
       $275 = ((($274)) + 12|0); //@line 106 "./common/Object.h"
       $276 = HEAP32[$275>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($276|0,($183|0)); //@line 106 "./common/Object.h"
       $277 = __THREW__; __THREW__ = 0;
       $278 = $277&1;
       if (!($278)) {
        $$7 = $272;$$79 = $273;
        break L4;
       }
       $279 = ___cxa_find_matching_catch_3(0|0)|0;
       $280 = tempRet0;
       ___clang_call_terminate($279); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      $209 = HEAP32[$183>>2]|0; //@line 106 "./common/Object.h"
      $210 = ((($209)) + 12|0); //@line 106 "./common/Object.h"
      $211 = HEAP32[$210>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($211|0,($183|0)); //@line 106 "./common/Object.h"
      $212 = __THREW__; __THREW__ = 0;
      $213 = $212&1;
      if ($213) {
       $214 = ___cxa_find_matching_catch_3(0|0)|0;
       $215 = tempRet0;
       ___clang_call_terminate($214); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      $216 = ((($e)) + 16|0); //@line 436 "modules/event/sdl/Event.cpp"
      $217 = HEAP16[$216>>1]|0; //@line 436 "modules/event/sdl/Event.cpp"
      $218 = (+($217<<16>>16)); //@line 436 "modules/event/sdl/Event.cpp"
      $219 = $218 * 3.0517578125E-5; //@line 436 "modules/event/sdl/Event.cpp"
      __THREW__ = 0;
      $220 = (+invoke_dd(3950,(+$219))); //@line 436 "modules/event/sdl/Event.cpp"
      $221 = __THREW__; __THREW__ = 0;
      $222 = $221&1;
      if ($222) {
       label = 5;
       break L4;
      }
      __THREW__ = 0;
      $223 = (invoke_ii(3348,32)|0); //@line 437 "modules/event/sdl/Event.cpp"
      $224 = __THREW__; __THREW__ = 0;
      $225 = $224&1;
      if ($225) {
       label = 5;
       break L4;
      }
      $226 = $220; //@line 437 "modules/event/sdl/Event.cpp"
      __THREW__ = 0;
      invoke_vid(3942,($223|0),(+$226)); //@line 437 "modules/event/sdl/Event.cpp"
      $227 = __THREW__; __THREW__ = 0;
      $228 = $227&1;
      if ($228) {
       $281 = ___cxa_find_matching_catch_2()|0;
       $282 = tempRet0;
       __ZdlPv($223); //@line 437 "modules/event/sdl/Event.cpp"
       $$7 = $281;$$79 = $282;
       break L4;
      }
      HEAP32[$5>>2] = $223; //@line 93 "./common/Object.h"
      $229 = HEAP32[$223>>2]|0; //@line 95 "./common/Object.h"
      $230 = ((($229)) + 8|0); //@line 95 "./common/Object.h"
      $231 = HEAP32[$230>>2]|0; //@line 95 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($231|0,($223|0)); //@line 95 "./common/Object.h"
      $232 = __THREW__; __THREW__ = 0;
      $233 = $232&1;
      if ($233) {
       label = 5;
       break L4;
      }
      $234 = HEAP32[$24>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $235 = HEAP32[$165>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $236 = ($234>>>0)<($235>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      do {
       if ($236) {
        HEAP32[$234>>2] = $223; //@line 99 "./common/Object.h"
        $237 = HEAP32[$223>>2]|0; //@line 101 "./common/Object.h"
        $238 = ((($237)) + 8|0); //@line 101 "./common/Object.h"
        $239 = HEAP32[$238>>2]|0; //@line 101 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($239|0,($223|0)); //@line 101 "./common/Object.h"
        $240 = __THREW__; __THREW__ = 0;
        $241 = $240&1;
        if ($241) {
         label = 81;
         break;
        }
        $$pre$i41 = HEAP32[$24>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $242 = ((($$pre$i41)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$24>>2] = $242; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       } else {
        __THREW__ = 0;
        invoke_vii(3507,($vargs|0),($5|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $243 = __THREW__; __THREW__ = 0;
        $244 = $243&1;
        if ($244) {
         label = 81;
        }
       }
      } while(0);
      if ((label|0) == 81) {
       $283 = ___cxa_find_matching_catch_2()|0;
       $284 = tempRet0;
       $285 = HEAP32[$223>>2]|0; //@line 106 "./common/Object.h"
       $286 = ((($285)) + 12|0); //@line 106 "./common/Object.h"
       $287 = HEAP32[$286>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($287|0,($223|0)); //@line 106 "./common/Object.h"
       $288 = __THREW__; __THREW__ = 0;
       $289 = $288&1;
       if (!($289)) {
        $$7 = $283;$$79 = $284;
        break L4;
       }
       $290 = ___cxa_find_matching_catch_3(0|0)|0;
       $291 = tempRet0;
       ___clang_call_terminate($290); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      $245 = HEAP32[$223>>2]|0; //@line 106 "./common/Object.h"
      $246 = ((($245)) + 12|0); //@line 106 "./common/Object.h"
      $247 = HEAP32[$246>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($247|0,($223|0)); //@line 106 "./common/Object.h"
      $248 = __THREW__; __THREW__ = 0;
      $249 = $248&1;
      if ($249) {
       $250 = ___cxa_find_matching_catch_3(0|0)|0;
       $251 = tempRet0;
       ___clang_call_terminate($250); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      __THREW__ = 0;
      $252 = (invoke_ii(3348,32)|0); //@line 438 "modules/event/sdl/Event.cpp"
      $253 = __THREW__; __THREW__ = 0;
      $254 = $253&1;
      if ($254) {
       label = 5;
       break L4;
      }
      __THREW__ = 0;
      invoke_viii(3460,($6|0),(365696|0),12); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $255 = __THREW__; __THREW__ = 0;
      $256 = $255&1;
      do {
       if ($256) {
        $292 = ___cxa_find_matching_catch_2()|0;
        $293 = tempRet0;
        $$1 = $292;$$13 = $293;
       } else {
        __THREW__ = 0;
        invoke_viii(3508,($252|0),($6|0),($vargs|0)); //@line 438 "modules/event/sdl/Event.cpp"
        $257 = __THREW__; __THREW__ = 0;
        $258 = $257&1;
        if ($258) {
         $294 = ___cxa_find_matching_catch_2()|0;
         $295 = tempRet0;
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 438 "modules/event/sdl/Event.cpp"
         $$1 = $294;$$13 = $295;
         break;
        } else {
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 438 "modules/event/sdl/Event.cpp"
         $msg$0 = $252;
         break L6;
        }
       }
      } while(0);
      __ZdlPv($252); //@line 438 "modules/event/sdl/Event.cpp"
      $$7 = $$1;$$79 = $$13;
      break L4;
     }
     break;
    }
    case 1538:  {
     $296 = ((($e)) + 13|0); //@line 442 "modules/event/sdl/Event.cpp"
     $297 = HEAP8[$296>>0]|0; //@line 442 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     $298 = (invoke_iii(3951,($297|0),($hat|0))|0); //@line 442 "modules/event/sdl/Event.cpp"
     $299 = __THREW__; __THREW__ = 0;
     $300 = $299&1;
     if ($300) {
      label = 5;
      break L4;
     }
     if ($298) {
      $301 = HEAP32[$hat>>2]|0; //@line 442 "modules/event/sdl/Event.cpp"
      __THREW__ = 0;
      $302 = (invoke_iii(3646,($301|0),($txt|0))|0); //@line 442 "modules/event/sdl/Event.cpp"
      $303 = __THREW__; __THREW__ = 0;
      $304 = $303&1;
      if ($304) {
       label = 5;
       break L4;
      }
      if ($302) {
       HEAP32[$proxy>>2] = 24; //@line 445 "modules/event/sdl/Event.cpp"
       $305 = HEAP32[$22>>2]|0; //@line 446 "modules/event/sdl/Event.cpp"
       $306 = ((($305)) + 32|0); //@line 446 "modules/event/sdl/Event.cpp"
       $307 = HEAP32[$306>>2]|0; //@line 446 "modules/event/sdl/Event.cpp"
       $308 = ((($e)) + 8|0); //@line 446 "modules/event/sdl/Event.cpp"
       $309 = HEAP32[$308>>2]|0; //@line 446 "modules/event/sdl/Event.cpp"
       __THREW__ = 0;
       $310 = (invoke_iii($307|0,($22|0),($309|0))|0); //@line 446 "modules/event/sdl/Event.cpp"
       $311 = __THREW__; __THREW__ = 0;
       $312 = $311&1;
       if ($312) {
        label = 5;
        break L4;
       }
       $313 = ((($proxy)) + 4|0); //@line 446 "modules/event/sdl/Event.cpp"
       HEAP32[$313>>2] = $310; //@line 446 "modules/event/sdl/Event.cpp"
       $314 = ($310|0)==(0|0); //@line 447 "modules/event/sdl/Event.cpp"
       if ($314) {
        $msg$0 = 0;
       } else {
        __THREW__ = 0;
        $315 = (invoke_ii(3348,32)|0); //@line 450 "modules/event/sdl/Event.cpp"
        $316 = __THREW__; __THREW__ = 0;
        $317 = $316&1;
        if ($317) {
         label = 5;
         break L4;
        }
        $318 = HEAP32[$proxy>>2]|0; //@line 450 "modules/event/sdl/Event.cpp"
        __THREW__ = 0;
        invoke_viii(3949,($315|0),($318|0),($proxy|0)); //@line 450 "modules/event/sdl/Event.cpp"
        $319 = __THREW__; __THREW__ = 0;
        $320 = $319&1;
        if ($320) {
         $415 = ___cxa_find_matching_catch_2()|0;
         $416 = tempRet0;
         __ZdlPv($315); //@line 450 "modules/event/sdl/Event.cpp"
         $$7 = $415;$$79 = $416;
         break L4;
        }
        HEAP32[$7>>2] = $315; //@line 93 "./common/Object.h"
        $321 = HEAP32[$315>>2]|0; //@line 95 "./common/Object.h"
        $322 = ((($321)) + 8|0); //@line 95 "./common/Object.h"
        $323 = HEAP32[$322>>2]|0; //@line 95 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($323|0,($315|0)); //@line 95 "./common/Object.h"
        $324 = __THREW__; __THREW__ = 0;
        $325 = $324&1;
        if ($325) {
         label = 5;
         break L4;
        }
        $326 = HEAP32[$24>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $327 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $328 = HEAP32[$327>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $329 = ($326>>>0)<($328>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($329) {
         HEAP32[$326>>2] = $315; //@line 99 "./common/Object.h"
         $330 = HEAP32[$315>>2]|0; //@line 101 "./common/Object.h"
         $331 = ((($330)) + 8|0); //@line 101 "./common/Object.h"
         $332 = HEAP32[$331>>2]|0; //@line 101 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($332|0,($315|0)); //@line 101 "./common/Object.h"
         $333 = __THREW__; __THREW__ = 0;
         $334 = $333&1;
         if ($334) {
          label = 124;
         } else {
          $$pre$i54 = HEAP32[$24>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $335 = ((($$pre$i54)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$24>>2] = $335; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         }
        } else {
         __THREW__ = 0;
         invoke_vii(3507,($vargs|0),($7|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $336 = __THREW__; __THREW__ = 0;
         $337 = $336&1;
         if ($337) {
          label = 124;
         }
        }
        if ((label|0) == 124) {
         $417 = ___cxa_find_matching_catch_2()|0;
         $418 = tempRet0;
         $419 = HEAP32[$315>>2]|0; //@line 106 "./common/Object.h"
         $420 = ((($419)) + 12|0); //@line 106 "./common/Object.h"
         $421 = HEAP32[$420>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($421|0,($315|0)); //@line 106 "./common/Object.h"
         $422 = __THREW__; __THREW__ = 0;
         $423 = $422&1;
         if (!($423)) {
          $$7 = $417;$$79 = $418;
          break L4;
         }
         $424 = ___cxa_find_matching_catch_3(0|0)|0;
         $425 = tempRet0;
         ___clang_call_terminate($424); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        $338 = HEAP32[$315>>2]|0; //@line 106 "./common/Object.h"
        $339 = ((($338)) + 12|0); //@line 106 "./common/Object.h"
        $340 = HEAP32[$339>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($340|0,($315|0)); //@line 106 "./common/Object.h"
        $341 = __THREW__; __THREW__ = 0;
        $342 = $341&1;
        if ($342) {
         $343 = ___cxa_find_matching_catch_3(0|0)|0;
         $344 = tempRet0;
         ___clang_call_terminate($343); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        __THREW__ = 0;
        $345 = (invoke_ii(3348,32)|0); //@line 451 "modules/event/sdl/Event.cpp"
        $346 = __THREW__; __THREW__ = 0;
        $347 = $346&1;
        if ($347) {
         label = 5;
         break L4;
        }
        $348 = ((($e)) + 12|0); //@line 451 "modules/event/sdl/Event.cpp"
        $349 = HEAP8[$348>>0]|0; //@line 451 "modules/event/sdl/Event.cpp"
        $350 = $349&255; //@line 451 "modules/event/sdl/Event.cpp"
        $351 = (($350) + 1)|0; //@line 451 "modules/event/sdl/Event.cpp"
        $352 = (+($351|0)); //@line 451 "modules/event/sdl/Event.cpp"
        __THREW__ = 0;
        invoke_vid(3942,($345|0),(+$352)); //@line 451 "modules/event/sdl/Event.cpp"
        $353 = __THREW__; __THREW__ = 0;
        $354 = $353&1;
        if ($354) {
         $426 = ___cxa_find_matching_catch_2()|0;
         $427 = tempRet0;
         __ZdlPv($345); //@line 451 "modules/event/sdl/Event.cpp"
         $$7 = $426;$$79 = $427;
         break L4;
        }
        HEAP32[$8>>2] = $345; //@line 93 "./common/Object.h"
        $355 = HEAP32[$345>>2]|0; //@line 95 "./common/Object.h"
        $356 = ((($355)) + 8|0); //@line 95 "./common/Object.h"
        $357 = HEAP32[$356>>2]|0; //@line 95 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($357|0,($345|0)); //@line 95 "./common/Object.h"
        $358 = __THREW__; __THREW__ = 0;
        $359 = $358&1;
        if ($359) {
         label = 5;
         break L4;
        }
        $360 = HEAP32[$24>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $361 = HEAP32[$327>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $362 = ($360>>>0)<($361>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        do {
         if ($362) {
          HEAP32[$360>>2] = $345; //@line 99 "./common/Object.h"
          $363 = HEAP32[$345>>2]|0; //@line 101 "./common/Object.h"
          $364 = ((($363)) + 8|0); //@line 101 "./common/Object.h"
          $365 = HEAP32[$364>>2]|0; //@line 101 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($365|0,($345|0)); //@line 101 "./common/Object.h"
          $366 = __THREW__; __THREW__ = 0;
          $367 = $366&1;
          if ($367) {
           label = 127;
           break;
          }
          $$pre$i62 = HEAP32[$24>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $368 = ((($$pre$i62)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$24>>2] = $368; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         } else {
          __THREW__ = 0;
          invoke_vii(3507,($vargs|0),($8|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $369 = __THREW__; __THREW__ = 0;
          $370 = $369&1;
          if ($370) {
           label = 127;
          }
         }
        } while(0);
        if ((label|0) == 127) {
         $428 = ___cxa_find_matching_catch_2()|0;
         $429 = tempRet0;
         $430 = HEAP32[$345>>2]|0; //@line 106 "./common/Object.h"
         $431 = ((($430)) + 12|0); //@line 106 "./common/Object.h"
         $432 = HEAP32[$431>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($432|0,($345|0)); //@line 106 "./common/Object.h"
         $433 = __THREW__; __THREW__ = 0;
         $434 = $433&1;
         if (!($434)) {
          $$7 = $428;$$79 = $429;
          break L4;
         }
         $435 = ___cxa_find_matching_catch_3(0|0)|0;
         $436 = tempRet0;
         ___clang_call_terminate($435); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        $371 = HEAP32[$345>>2]|0; //@line 106 "./common/Object.h"
        $372 = ((($371)) + 12|0); //@line 106 "./common/Object.h"
        $373 = HEAP32[$372>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($373|0,($345|0)); //@line 106 "./common/Object.h"
        $374 = __THREW__; __THREW__ = 0;
        $375 = $374&1;
        if ($375) {
         $376 = ___cxa_find_matching_catch_3(0|0)|0;
         $377 = tempRet0;
         ___clang_call_terminate($376); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        __THREW__ = 0;
        $378 = (invoke_ii(3348,32)|0); //@line 452 "modules/event/sdl/Event.cpp"
        $379 = __THREW__; __THREW__ = 0;
        $380 = $379&1;
        if ($380) {
         label = 5;
         break L4;
        }
        $381 = HEAP32[$txt>>2]|0; //@line 452 "modules/event/sdl/Event.cpp"
        $382 = (_strlen($381)|0); //@line 452 "modules/event/sdl/Event.cpp"
        __THREW__ = 0;
        invoke_viii(3940,($378|0),($381|0),($382|0)); //@line 452 "modules/event/sdl/Event.cpp"
        $383 = __THREW__; __THREW__ = 0;
        $384 = $383&1;
        if ($384) {
         $437 = ___cxa_find_matching_catch_2()|0;
         $438 = tempRet0;
         __ZdlPv($378); //@line 452 "modules/event/sdl/Event.cpp"
         $$7 = $437;$$79 = $438;
         break L4;
        }
        HEAP32[$9>>2] = $378; //@line 93 "./common/Object.h"
        $385 = HEAP32[$378>>2]|0; //@line 95 "./common/Object.h"
        $386 = ((($385)) + 8|0); //@line 95 "./common/Object.h"
        $387 = HEAP32[$386>>2]|0; //@line 95 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($387|0,($378|0)); //@line 95 "./common/Object.h"
        $388 = __THREW__; __THREW__ = 0;
        $389 = $388&1;
        if ($389) {
         label = 5;
         break L4;
        }
        $390 = HEAP32[$24>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $391 = HEAP32[$327>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $392 = ($390>>>0)<($391>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        do {
         if ($392) {
          HEAP32[$390>>2] = $378; //@line 99 "./common/Object.h"
          $393 = HEAP32[$378>>2]|0; //@line 101 "./common/Object.h"
          $394 = ((($393)) + 8|0); //@line 101 "./common/Object.h"
          $395 = HEAP32[$394>>2]|0; //@line 101 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($395|0,($378|0)); //@line 101 "./common/Object.h"
          $396 = __THREW__; __THREW__ = 0;
          $397 = $396&1;
          if ($397) {
           label = 130;
           break;
          }
          $$pre$i70 = HEAP32[$24>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $398 = ((($$pre$i70)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$24>>2] = $398; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         } else {
          __THREW__ = 0;
          invoke_vii(3507,($vargs|0),($9|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $399 = __THREW__; __THREW__ = 0;
          $400 = $399&1;
          if ($400) {
           label = 130;
          }
         }
        } while(0);
        if ((label|0) == 130) {
         $439 = ___cxa_find_matching_catch_2()|0;
         $440 = tempRet0;
         $441 = HEAP32[$378>>2]|0; //@line 106 "./common/Object.h"
         $442 = ((($441)) + 12|0); //@line 106 "./common/Object.h"
         $443 = HEAP32[$442>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($443|0,($378|0)); //@line 106 "./common/Object.h"
         $444 = __THREW__; __THREW__ = 0;
         $445 = $444&1;
         if (!($445)) {
          $$7 = $439;$$79 = $440;
          break L4;
         }
         $446 = ___cxa_find_matching_catch_3(0|0)|0;
         $447 = tempRet0;
         ___clang_call_terminate($446); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        $401 = HEAP32[$378>>2]|0; //@line 106 "./common/Object.h"
        $402 = ((($401)) + 12|0); //@line 106 "./common/Object.h"
        $403 = HEAP32[$402>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($403|0,($378|0)); //@line 106 "./common/Object.h"
        $404 = __THREW__; __THREW__ = 0;
        $405 = $404&1;
        if ($405) {
         $406 = ___cxa_find_matching_catch_3(0|0)|0;
         $407 = tempRet0;
         ___clang_call_terminate($406); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        __THREW__ = 0;
        $408 = (invoke_ii(3348,32)|0); //@line 453 "modules/event/sdl/Event.cpp"
        $409 = __THREW__; __THREW__ = 0;
        $410 = $409&1;
        if ($410) {
         label = 5;
         break L4;
        }
        __THREW__ = 0;
        invoke_viii(3460,($10|0),(365709|0),11); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $411 = __THREW__; __THREW__ = 0;
        $412 = $411&1;
        do {
         if ($412) {
          $448 = ___cxa_find_matching_catch_2()|0;
          $449 = tempRet0;
          $$2 = $448;$$24 = $449;
         } else {
          __THREW__ = 0;
          invoke_viii(3508,($408|0),($10|0),($vargs|0)); //@line 453 "modules/event/sdl/Event.cpp"
          $413 = __THREW__; __THREW__ = 0;
          $414 = $413&1;
          if ($414) {
           $450 = ___cxa_find_matching_catch_2()|0;
           $451 = tempRet0;
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10); //@line 453 "modules/event/sdl/Event.cpp"
           $$2 = $450;$$24 = $451;
           break;
          } else {
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10); //@line 453 "modules/event/sdl/Event.cpp"
           $msg$0 = $408;
           break L6;
          }
         }
        } while(0);
        __ZdlPv($408); //@line 453 "modules/event/sdl/Event.cpp"
        $$7 = $$2;$$79 = $$24;
        break L4;
       }
      } else {
       $msg$0 = 0;
      }
     } else {
      $msg$0 = 0;
     }
     break;
    }
    case 1618: case 1617:  {
     $452 = ((($e)) + 12|0); //@line 457 "modules/event/sdl/Event.cpp"
     $453 = HEAP8[$452>>0]|0; //@line 457 "modules/event/sdl/Event.cpp"
     $454 = $453&255; //@line 457 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     $455 = (invoke_iii(3952,($454|0),($padbutton|0))|0); //@line 457 "modules/event/sdl/Event.cpp"
     $456 = __THREW__; __THREW__ = 0;
     $457 = $456&1;
     if ($457) {
      label = 5;
      break L4;
     }
     if ($455) {
      $458 = HEAP32[$padbutton>>2]|0; //@line 460 "modules/event/sdl/Event.cpp"
      __THREW__ = 0;
      $459 = (invoke_iii(3953,($458|0),($txt|0))|0); //@line 460 "modules/event/sdl/Event.cpp"
      $460 = __THREW__; __THREW__ = 0;
      $461 = $460&1;
      if ($461) {
       label = 5;
       break L4;
      }
      if ($459) {
       HEAP32[$proxy>>2] = 24; //@line 463 "modules/event/sdl/Event.cpp"
       $462 = HEAP32[$22>>2]|0; //@line 464 "modules/event/sdl/Event.cpp"
       $463 = ((($462)) + 32|0); //@line 464 "modules/event/sdl/Event.cpp"
       $464 = HEAP32[$463>>2]|0; //@line 464 "modules/event/sdl/Event.cpp"
       $465 = ((($e)) + 8|0); //@line 464 "modules/event/sdl/Event.cpp"
       $466 = HEAP32[$465>>2]|0; //@line 464 "modules/event/sdl/Event.cpp"
       __THREW__ = 0;
       $467 = (invoke_iii($464|0,($22|0),($466|0))|0); //@line 464 "modules/event/sdl/Event.cpp"
       $468 = __THREW__; __THREW__ = 0;
       $469 = $468&1;
       if ($469) {
        label = 5;
        break L4;
       }
       $470 = ((($proxy)) + 4|0); //@line 464 "modules/event/sdl/Event.cpp"
       HEAP32[$470>>2] = $467; //@line 464 "modules/event/sdl/Event.cpp"
       $471 = ($467|0)==(0|0); //@line 465 "modules/event/sdl/Event.cpp"
       if ($471) {
        $msg$0 = 0;
       } else {
        __THREW__ = 0;
        $472 = (invoke_ii(3348,32)|0); //@line 468 "modules/event/sdl/Event.cpp"
        $473 = __THREW__; __THREW__ = 0;
        $474 = $473&1;
        if ($474) {
         label = 5;
         break L4;
        }
        $475 = HEAP32[$proxy>>2]|0; //@line 468 "modules/event/sdl/Event.cpp"
        __THREW__ = 0;
        invoke_viii(3949,($472|0),($475|0),($proxy|0)); //@line 468 "modules/event/sdl/Event.cpp"
        $476 = __THREW__; __THREW__ = 0;
        $477 = $476&1;
        if ($477) {
         $543 = ___cxa_find_matching_catch_2()|0;
         $544 = tempRet0;
         __ZdlPv($472); //@line 468 "modules/event/sdl/Event.cpp"
         $$7 = $543;$$79 = $544;
         break L4;
        }
        HEAP32[$11>>2] = $472; //@line 93 "./common/Object.h"
        $478 = HEAP32[$472>>2]|0; //@line 95 "./common/Object.h"
        $479 = ((($478)) + 8|0); //@line 95 "./common/Object.h"
        $480 = HEAP32[$479>>2]|0; //@line 95 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($480|0,($472|0)); //@line 95 "./common/Object.h"
        $481 = __THREW__; __THREW__ = 0;
        $482 = $481&1;
        if ($482) {
         label = 5;
         break L4;
        }
        $483 = HEAP32[$24>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $484 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $485 = HEAP32[$484>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $486 = ($483>>>0)<($485>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($486) {
         HEAP32[$483>>2] = $472; //@line 99 "./common/Object.h"
         $487 = HEAP32[$472>>2]|0; //@line 101 "./common/Object.h"
         $488 = ((($487)) + 8|0); //@line 101 "./common/Object.h"
         $489 = HEAP32[$488>>2]|0; //@line 101 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($489|0,($472|0)); //@line 101 "./common/Object.h"
         $490 = __THREW__; __THREW__ = 0;
         $491 = $490&1;
         if ($491) {
          label = 164;
         } else {
          $$pre$i83 = HEAP32[$24>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $492 = ((($$pre$i83)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$24>>2] = $492; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         }
        } else {
         __THREW__ = 0;
         invoke_vii(3507,($vargs|0),($11|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $493 = __THREW__; __THREW__ = 0;
         $494 = $493&1;
         if ($494) {
          label = 164;
         }
        }
        if ((label|0) == 164) {
         $545 = ___cxa_find_matching_catch_2()|0;
         $546 = tempRet0;
         $547 = HEAP32[$472>>2]|0; //@line 106 "./common/Object.h"
         $548 = ((($547)) + 12|0); //@line 106 "./common/Object.h"
         $549 = HEAP32[$548>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($549|0,($472|0)); //@line 106 "./common/Object.h"
         $550 = __THREW__; __THREW__ = 0;
         $551 = $550&1;
         if (!($551)) {
          $$7 = $545;$$79 = $546;
          break L4;
         }
         $552 = ___cxa_find_matching_catch_3(0|0)|0;
         $553 = tempRet0;
         ___clang_call_terminate($552); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        $495 = HEAP32[$472>>2]|0; //@line 106 "./common/Object.h"
        $496 = ((($495)) + 12|0); //@line 106 "./common/Object.h"
        $497 = HEAP32[$496>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($497|0,($472|0)); //@line 106 "./common/Object.h"
        $498 = __THREW__; __THREW__ = 0;
        $499 = $498&1;
        if ($499) {
         $500 = ___cxa_find_matching_catch_3(0|0)|0;
         $501 = tempRet0;
         ___clang_call_terminate($500); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        __THREW__ = 0;
        $502 = (invoke_ii(3348,32)|0); //@line 469 "modules/event/sdl/Event.cpp"
        $503 = __THREW__; __THREW__ = 0;
        $504 = $503&1;
        if ($504) {
         label = 5;
         break L4;
        }
        $505 = HEAP32[$txt>>2]|0; //@line 469 "modules/event/sdl/Event.cpp"
        $506 = (_strlen($505)|0); //@line 469 "modules/event/sdl/Event.cpp"
        __THREW__ = 0;
        invoke_viii(3940,($502|0),($505|0),($506|0)); //@line 469 "modules/event/sdl/Event.cpp"
        $507 = __THREW__; __THREW__ = 0;
        $508 = $507&1;
        if ($508) {
         $554 = ___cxa_find_matching_catch_2()|0;
         $555 = tempRet0;
         __ZdlPv($502); //@line 469 "modules/event/sdl/Event.cpp"
         $$7 = $554;$$79 = $555;
         break L4;
        }
        HEAP32[$12>>2] = $502; //@line 93 "./common/Object.h"
        $509 = HEAP32[$502>>2]|0; //@line 95 "./common/Object.h"
        $510 = ((($509)) + 8|0); //@line 95 "./common/Object.h"
        $511 = HEAP32[$510>>2]|0; //@line 95 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($511|0,($502|0)); //@line 95 "./common/Object.h"
        $512 = __THREW__; __THREW__ = 0;
        $513 = $512&1;
        if ($513) {
         label = 5;
         break L4;
        }
        $514 = HEAP32[$24>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $515 = HEAP32[$484>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $516 = ($514>>>0)<($515>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        do {
         if ($516) {
          HEAP32[$514>>2] = $502; //@line 99 "./common/Object.h"
          $517 = HEAP32[$502>>2]|0; //@line 101 "./common/Object.h"
          $518 = ((($517)) + 8|0); //@line 101 "./common/Object.h"
          $519 = HEAP32[$518>>2]|0; //@line 101 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($519|0,($502|0)); //@line 101 "./common/Object.h"
          $520 = __THREW__; __THREW__ = 0;
          $521 = $520&1;
          if ($521) {
           label = 167;
           break;
          }
          $$pre$i91 = HEAP32[$24>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $522 = ((($$pre$i91)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$24>>2] = $522; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         } else {
          __THREW__ = 0;
          invoke_vii(3507,($vargs|0),($12|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $523 = __THREW__; __THREW__ = 0;
          $524 = $523&1;
          if ($524) {
           label = 167;
          }
         }
        } while(0);
        if ((label|0) == 167) {
         $556 = ___cxa_find_matching_catch_2()|0;
         $557 = tempRet0;
         $558 = HEAP32[$502>>2]|0; //@line 106 "./common/Object.h"
         $559 = ((($558)) + 12|0); //@line 106 "./common/Object.h"
         $560 = HEAP32[$559>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($560|0,($502|0)); //@line 106 "./common/Object.h"
         $561 = __THREW__; __THREW__ = 0;
         $562 = $561&1;
         if (!($562)) {
          $$7 = $556;$$79 = $557;
          break L4;
         }
         $563 = ___cxa_find_matching_catch_3(0|0)|0;
         $564 = tempRet0;
         ___clang_call_terminate($563); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        $525 = HEAP32[$502>>2]|0; //@line 106 "./common/Object.h"
        $526 = ((($525)) + 12|0); //@line 106 "./common/Object.h"
        $527 = HEAP32[$526>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($527|0,($502|0)); //@line 106 "./common/Object.h"
        $528 = __THREW__; __THREW__ = 0;
        $529 = $528&1;
        if ($529) {
         $530 = ___cxa_find_matching_catch_3(0|0)|0;
         $531 = tempRet0;
         ___clang_call_terminate($530); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        __THREW__ = 0;
        $532 = (invoke_ii(3348,32)|0); //@line 470 "modules/event/sdl/Event.cpp"
        $533 = __THREW__; __THREW__ = 0;
        $534 = $533&1;
        if ($534) {
         label = 5;
         break L4;
        }
        $535 = HEAP32[$e>>2]|0; //@line 470 "modules/event/sdl/Event.cpp"
        $536 = ($535|0)==(1617); //@line 470 "modules/event/sdl/Event.cpp"
        $537 = $536 ? 365721 : 365736; //@line 470 "modules/event/sdl/Event.cpp"
        $538 = $536 ? 14 : 15; //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        __THREW__ = 0;
        invoke_viii(3460,($13|0),($537|0),($538|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $539 = __THREW__; __THREW__ = 0;
        $540 = $539&1;
        do {
         if ($540) {
          $565 = ___cxa_find_matching_catch_2()|0;
          $566 = tempRet0;
          $$3 = $565;$$35 = $566;
         } else {
          __THREW__ = 0;
          invoke_viii(3508,($532|0),($13|0),($vargs|0)); //@line 470 "modules/event/sdl/Event.cpp"
          $541 = __THREW__; __THREW__ = 0;
          $542 = $541&1;
          if ($542) {
           $567 = ___cxa_find_matching_catch_2()|0;
           $568 = tempRet0;
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13); //@line 470 "modules/event/sdl/Event.cpp"
           $$3 = $567;$$35 = $568;
           break;
          } else {
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13); //@line 470 "modules/event/sdl/Event.cpp"
           $msg$0 = $532;
           break L6;
          }
         }
        } while(0);
        __ZdlPv($532); //@line 470 "modules/event/sdl/Event.cpp"
        $$7 = $$3;$$79 = $$35;
        break L4;
       }
      } else {
       $msg$0 = 0;
      }
     } else {
      $msg$0 = 0;
     }
     break;
    }
    case 1616:  {
     $569 = ((($e)) + 12|0); //@line 474 "modules/event/sdl/Event.cpp"
     $570 = HEAP8[$569>>0]|0; //@line 474 "modules/event/sdl/Event.cpp"
     $571 = $570&255; //@line 474 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     $572 = (invoke_iii(3954,($571|0),($padaxis|0))|0); //@line 474 "modules/event/sdl/Event.cpp"
     $573 = __THREW__; __THREW__ = 0;
     $574 = $573&1;
     if ($574) {
      label = 5;
      break L4;
     }
     if ($572) {
      $575 = HEAP32[$padaxis>>2]|0; //@line 476 "modules/event/sdl/Event.cpp"
      __THREW__ = 0;
      $576 = (invoke_iii(3955,($575|0),($txt|0))|0); //@line 476 "modules/event/sdl/Event.cpp"
      $577 = __THREW__; __THREW__ = 0;
      $578 = $577&1;
      if ($578) {
       label = 5;
       break L4;
      }
      if ($576) {
       HEAP32[$proxy>>2] = 24; //@line 479 "modules/event/sdl/Event.cpp"
       $579 = HEAP32[$22>>2]|0; //@line 480 "modules/event/sdl/Event.cpp"
       $580 = ((($579)) + 32|0); //@line 480 "modules/event/sdl/Event.cpp"
       $581 = HEAP32[$580>>2]|0; //@line 480 "modules/event/sdl/Event.cpp"
       $582 = ((($e)) + 8|0); //@line 480 "modules/event/sdl/Event.cpp"
       $583 = HEAP32[$582>>2]|0; //@line 480 "modules/event/sdl/Event.cpp"
       __THREW__ = 0;
       $584 = (invoke_iii($581|0,($22|0),($583|0))|0); //@line 480 "modules/event/sdl/Event.cpp"
       $585 = __THREW__; __THREW__ = 0;
       $586 = $585&1;
       if ($586) {
        label = 5;
        break L4;
       }
       $587 = ((($proxy)) + 4|0); //@line 480 "modules/event/sdl/Event.cpp"
       HEAP32[$587>>2] = $584; //@line 480 "modules/event/sdl/Event.cpp"
       $588 = ($584|0)==(0|0); //@line 481 "modules/event/sdl/Event.cpp"
       if ($588) {
        $msg$0 = 0;
       } else {
        __THREW__ = 0;
        $589 = (invoke_ii(3348,32)|0); //@line 484 "modules/event/sdl/Event.cpp"
        $590 = __THREW__; __THREW__ = 0;
        $591 = $590&1;
        if ($591) {
         label = 5;
         break L4;
        }
        $592 = HEAP32[$proxy>>2]|0; //@line 484 "modules/event/sdl/Event.cpp"
        __THREW__ = 0;
        invoke_viii(3949,($589|0),($592|0),($proxy|0)); //@line 484 "modules/event/sdl/Event.cpp"
        $593 = __THREW__; __THREW__ = 0;
        $594 = $593&1;
        if ($594) {
         $692 = ___cxa_find_matching_catch_2()|0;
         $693 = tempRet0;
         __ZdlPv($589); //@line 484 "modules/event/sdl/Event.cpp"
         $$7 = $692;$$79 = $693;
         break L4;
        }
        HEAP32[$14>>2] = $589; //@line 93 "./common/Object.h"
        $595 = HEAP32[$589>>2]|0; //@line 95 "./common/Object.h"
        $596 = ((($595)) + 8|0); //@line 95 "./common/Object.h"
        $597 = HEAP32[$596>>2]|0; //@line 95 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($597|0,($589|0)); //@line 95 "./common/Object.h"
        $598 = __THREW__; __THREW__ = 0;
        $599 = $598&1;
        if ($599) {
         label = 5;
         break L4;
        }
        $600 = HEAP32[$24>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $601 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $602 = HEAP32[$601>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $603 = ($600>>>0)<($602>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($603) {
         HEAP32[$600>>2] = $589; //@line 99 "./common/Object.h"
         $604 = HEAP32[$589>>2]|0; //@line 101 "./common/Object.h"
         $605 = ((($604)) + 8|0); //@line 101 "./common/Object.h"
         $606 = HEAP32[$605>>2]|0; //@line 101 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($606|0,($589|0)); //@line 101 "./common/Object.h"
         $607 = __THREW__; __THREW__ = 0;
         $608 = $607&1;
         if ($608) {
          label = 211;
         } else {
          $$pre$i103 = HEAP32[$24>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $609 = ((($$pre$i103)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$24>>2] = $609; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         }
        } else {
         __THREW__ = 0;
         invoke_vii(3507,($vargs|0),($14|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $610 = __THREW__; __THREW__ = 0;
         $611 = $610&1;
         if ($611) {
          label = 211;
         }
        }
        if ((label|0) == 211) {
         $694 = ___cxa_find_matching_catch_2()|0;
         $695 = tempRet0;
         $696 = HEAP32[$589>>2]|0; //@line 106 "./common/Object.h"
         $697 = ((($696)) + 12|0); //@line 106 "./common/Object.h"
         $698 = HEAP32[$697>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($698|0,($589|0)); //@line 106 "./common/Object.h"
         $699 = __THREW__; __THREW__ = 0;
         $700 = $699&1;
         if (!($700)) {
          $$7 = $694;$$79 = $695;
          break L4;
         }
         $701 = ___cxa_find_matching_catch_3(0|0)|0;
         $702 = tempRet0;
         ___clang_call_terminate($701); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        $612 = HEAP32[$589>>2]|0; //@line 106 "./common/Object.h"
        $613 = ((($612)) + 12|0); //@line 106 "./common/Object.h"
        $614 = HEAP32[$613>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($614|0,($589|0)); //@line 106 "./common/Object.h"
        $615 = __THREW__; __THREW__ = 0;
        $616 = $615&1;
        if ($616) {
         $617 = ___cxa_find_matching_catch_3(0|0)|0;
         $618 = tempRet0;
         ___clang_call_terminate($617); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        __THREW__ = 0;
        $619 = (invoke_ii(3348,32)|0); //@line 486 "modules/event/sdl/Event.cpp"
        $620 = __THREW__; __THREW__ = 0;
        $621 = $620&1;
        if ($621) {
         label = 5;
         break L4;
        }
        $622 = HEAP32[$txt>>2]|0; //@line 486 "modules/event/sdl/Event.cpp"
        $623 = (_strlen($622)|0); //@line 486 "modules/event/sdl/Event.cpp"
        __THREW__ = 0;
        invoke_viii(3940,($619|0),($622|0),($623|0)); //@line 486 "modules/event/sdl/Event.cpp"
        $624 = __THREW__; __THREW__ = 0;
        $625 = $624&1;
        if ($625) {
         $703 = ___cxa_find_matching_catch_2()|0;
         $704 = tempRet0;
         __ZdlPv($619); //@line 486 "modules/event/sdl/Event.cpp"
         $$7 = $703;$$79 = $704;
         break L4;
        }
        HEAP32[$15>>2] = $619; //@line 93 "./common/Object.h"
        $626 = HEAP32[$619>>2]|0; //@line 95 "./common/Object.h"
        $627 = ((($626)) + 8|0); //@line 95 "./common/Object.h"
        $628 = HEAP32[$627>>2]|0; //@line 95 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($628|0,($619|0)); //@line 95 "./common/Object.h"
        $629 = __THREW__; __THREW__ = 0;
        $630 = $629&1;
        if ($630) {
         label = 5;
         break L4;
        }
        $631 = HEAP32[$24>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $632 = HEAP32[$601>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $633 = ($631>>>0)<($632>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        do {
         if ($633) {
          HEAP32[$631>>2] = $619; //@line 99 "./common/Object.h"
          $634 = HEAP32[$619>>2]|0; //@line 101 "./common/Object.h"
          $635 = ((($634)) + 8|0); //@line 101 "./common/Object.h"
          $636 = HEAP32[$635>>2]|0; //@line 101 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($636|0,($619|0)); //@line 101 "./common/Object.h"
          $637 = __THREW__; __THREW__ = 0;
          $638 = $637&1;
          if ($638) {
           label = 214;
           break;
          }
          $$pre$i111 = HEAP32[$24>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $639 = ((($$pre$i111)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$24>>2] = $639; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         } else {
          __THREW__ = 0;
          invoke_vii(3507,($vargs|0),($15|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $640 = __THREW__; __THREW__ = 0;
          $641 = $640&1;
          if ($641) {
           label = 214;
          }
         }
        } while(0);
        if ((label|0) == 214) {
         $705 = ___cxa_find_matching_catch_2()|0;
         $706 = tempRet0;
         $707 = HEAP32[$619>>2]|0; //@line 106 "./common/Object.h"
         $708 = ((($707)) + 12|0); //@line 106 "./common/Object.h"
         $709 = HEAP32[$708>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($709|0,($619|0)); //@line 106 "./common/Object.h"
         $710 = __THREW__; __THREW__ = 0;
         $711 = $710&1;
         if (!($711)) {
          $$7 = $705;$$79 = $706;
          break L4;
         }
         $712 = ___cxa_find_matching_catch_3(0|0)|0;
         $713 = tempRet0;
         ___clang_call_terminate($712); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        $642 = HEAP32[$619>>2]|0; //@line 106 "./common/Object.h"
        $643 = ((($642)) + 12|0); //@line 106 "./common/Object.h"
        $644 = HEAP32[$643>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($644|0,($619|0)); //@line 106 "./common/Object.h"
        $645 = __THREW__; __THREW__ = 0;
        $646 = $645&1;
        if ($646) {
         $647 = ___cxa_find_matching_catch_3(0|0)|0;
         $648 = tempRet0;
         ___clang_call_terminate($647); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        $649 = ((($e)) + 16|0); //@line 487 "modules/event/sdl/Event.cpp"
        $650 = HEAP16[$649>>1]|0; //@line 487 "modules/event/sdl/Event.cpp"
        $651 = (+($650<<16>>16)); //@line 487 "modules/event/sdl/Event.cpp"
        $652 = $651 * 3.0517578125E-5; //@line 487 "modules/event/sdl/Event.cpp"
        __THREW__ = 0;
        $653 = (+invoke_dd(3950,(+$652))); //@line 487 "modules/event/sdl/Event.cpp"
        $654 = __THREW__; __THREW__ = 0;
        $655 = $654&1;
        if ($655) {
         label = 5;
         break L4;
        }
        __THREW__ = 0;
        $656 = (invoke_ii(3348,32)|0); //@line 488 "modules/event/sdl/Event.cpp"
        $657 = __THREW__; __THREW__ = 0;
        $658 = $657&1;
        if ($658) {
         label = 5;
         break L4;
        }
        $659 = $653; //@line 488 "modules/event/sdl/Event.cpp"
        __THREW__ = 0;
        invoke_vid(3942,($656|0),(+$659)); //@line 488 "modules/event/sdl/Event.cpp"
        $660 = __THREW__; __THREW__ = 0;
        $661 = $660&1;
        if ($661) {
         $714 = ___cxa_find_matching_catch_2()|0;
         $715 = tempRet0;
         __ZdlPv($656); //@line 488 "modules/event/sdl/Event.cpp"
         $$7 = $714;$$79 = $715;
         break L4;
        }
        HEAP32[$16>>2] = $656; //@line 93 "./common/Object.h"
        $662 = HEAP32[$656>>2]|0; //@line 95 "./common/Object.h"
        $663 = ((($662)) + 8|0); //@line 95 "./common/Object.h"
        $664 = HEAP32[$663>>2]|0; //@line 95 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($664|0,($656|0)); //@line 95 "./common/Object.h"
        $665 = __THREW__; __THREW__ = 0;
        $666 = $665&1;
        if ($666) {
         label = 5;
         break L4;
        }
        $667 = HEAP32[$24>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $668 = HEAP32[$601>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $669 = ($667>>>0)<($668>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        do {
         if ($669) {
          HEAP32[$667>>2] = $656; //@line 99 "./common/Object.h"
          $670 = HEAP32[$656>>2]|0; //@line 101 "./common/Object.h"
          $671 = ((($670)) + 8|0); //@line 101 "./common/Object.h"
          $672 = HEAP32[$671>>2]|0; //@line 101 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($672|0,($656|0)); //@line 101 "./common/Object.h"
          $673 = __THREW__; __THREW__ = 0;
          $674 = $673&1;
          if ($674) {
           label = 217;
           break;
          }
          $$pre$i119 = HEAP32[$24>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $675 = ((($$pre$i119)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$24>>2] = $675; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         } else {
          __THREW__ = 0;
          invoke_vii(3507,($vargs|0),($16|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $676 = __THREW__; __THREW__ = 0;
          $677 = $676&1;
          if ($677) {
           label = 217;
          }
         }
        } while(0);
        if ((label|0) == 217) {
         $716 = ___cxa_find_matching_catch_2()|0;
         $717 = tempRet0;
         $718 = HEAP32[$656>>2]|0; //@line 106 "./common/Object.h"
         $719 = ((($718)) + 12|0); //@line 106 "./common/Object.h"
         $720 = HEAP32[$719>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($720|0,($656|0)); //@line 106 "./common/Object.h"
         $721 = __THREW__; __THREW__ = 0;
         $722 = $721&1;
         if (!($722)) {
          $$7 = $716;$$79 = $717;
          break L4;
         }
         $723 = ___cxa_find_matching_catch_3(0|0)|0;
         $724 = tempRet0;
         ___clang_call_terminate($723); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        $678 = HEAP32[$656>>2]|0; //@line 106 "./common/Object.h"
        $679 = ((($678)) + 12|0); //@line 106 "./common/Object.h"
        $680 = HEAP32[$679>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($680|0,($656|0)); //@line 106 "./common/Object.h"
        $681 = __THREW__; __THREW__ = 0;
        $682 = $681&1;
        if ($682) {
         $683 = ___cxa_find_matching_catch_3(0|0)|0;
         $684 = tempRet0;
         ___clang_call_terminate($683); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        __THREW__ = 0;
        $685 = (invoke_ii(3348,32)|0); //@line 489 "modules/event/sdl/Event.cpp"
        $686 = __THREW__; __THREW__ = 0;
        $687 = $686&1;
        if ($687) {
         label = 5;
         break L4;
        }
        __THREW__ = 0;
        invoke_viii(3460,($17|0),(365752|0),11); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $688 = __THREW__; __THREW__ = 0;
        $689 = $688&1;
        do {
         if ($689) {
          $725 = ___cxa_find_matching_catch_2()|0;
          $726 = tempRet0;
          $$4 = $725;$$46 = $726;
         } else {
          __THREW__ = 0;
          invoke_viii(3508,($685|0),($17|0),($vargs|0)); //@line 489 "modules/event/sdl/Event.cpp"
          $690 = __THREW__; __THREW__ = 0;
          $691 = $690&1;
          if ($691) {
           $727 = ___cxa_find_matching_catch_2()|0;
           $728 = tempRet0;
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($17); //@line 489 "modules/event/sdl/Event.cpp"
           $$4 = $727;$$46 = $728;
           break;
          } else {
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($17); //@line 489 "modules/event/sdl/Event.cpp"
           $msg$0 = $685;
           break L6;
          }
         }
        } while(0);
        __ZdlPv($685); //@line 489 "modules/event/sdl/Event.cpp"
        $$7 = $$4;$$79 = $$46;
        break L4;
       }
      } else {
       $msg$0 = 0;
      }
     } else {
      $msg$0 = 0;
     }
     break;
    }
    case 1541:  {
     $729 = HEAP32[$22>>2]|0; //@line 494 "modules/event/sdl/Event.cpp"
     $730 = ((($729)) + 24|0); //@line 494 "modules/event/sdl/Event.cpp"
     $731 = HEAP32[$730>>2]|0; //@line 494 "modules/event/sdl/Event.cpp"
     $732 = ((($e)) + 8|0); //@line 494 "modules/event/sdl/Event.cpp"
     $733 = HEAP32[$732>>2]|0; //@line 494 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     $734 = (invoke_iii($731|0,($22|0),($733|0))|0); //@line 494 "modules/event/sdl/Event.cpp"
     $735 = __THREW__; __THREW__ = 0;
     $736 = $735&1;
     if ($736) {
      label = 5;
      break L4;
     }
     $737 = ((($proxy)) + 4|0); //@line 494 "modules/event/sdl/Event.cpp"
     HEAP32[$737>>2] = $734; //@line 494 "modules/event/sdl/Event.cpp"
     HEAP32[$proxy>>2] = 24; //@line 495 "modules/event/sdl/Event.cpp"
     $738 = ($734|0)==(0|0); //@line 496 "modules/event/sdl/Event.cpp"
     if ($738) {
      $msg$0 = 0;
     } else {
      __THREW__ = 0;
      $739 = (invoke_ii(3348,32)|0); //@line 498 "modules/event/sdl/Event.cpp"
      $740 = __THREW__; __THREW__ = 0;
      $741 = $740&1;
      if ($741) {
       label = 5;
       break L4;
      }
      __THREW__ = 0;
      invoke_viii(3949,($739|0),24,($proxy|0)); //@line 498 "modules/event/sdl/Event.cpp"
      $742 = __THREW__; __THREW__ = 0;
      $743 = $742&1;
      if ($743) {
       $775 = ___cxa_find_matching_catch_2()|0;
       $776 = tempRet0;
       __ZdlPv($739); //@line 498 "modules/event/sdl/Event.cpp"
       $$7 = $775;$$79 = $776;
       break L4;
      }
      HEAP32[$18>>2] = $739; //@line 93 "./common/Object.h"
      $744 = HEAP32[$739>>2]|0; //@line 95 "./common/Object.h"
      $745 = ((($744)) + 8|0); //@line 95 "./common/Object.h"
      $746 = HEAP32[$745>>2]|0; //@line 95 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($746|0,($739|0)); //@line 95 "./common/Object.h"
      $747 = __THREW__; __THREW__ = 0;
      $748 = $747&1;
      if ($748) {
       label = 5;
       break L4;
      }
      $749 = HEAP32[$24>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $750 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $751 = HEAP32[$750>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $752 = ($749>>>0)<($751>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($752) {
       HEAP32[$749>>2] = $739; //@line 99 "./common/Object.h"
       $753 = HEAP32[$739>>2]|0; //@line 101 "./common/Object.h"
       $754 = ((($753)) + 8|0); //@line 101 "./common/Object.h"
       $755 = HEAP32[$754>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($755|0,($739|0)); //@line 101 "./common/Object.h"
       $756 = __THREW__; __THREW__ = 0;
       $757 = $756&1;
       if ($757) {
        label = 238;
       } else {
        $$pre$i132 = HEAP32[$24>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $758 = ((($$pre$i132)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$24>>2] = $758; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
      } else {
       __THREW__ = 0;
       invoke_vii(3507,($vargs|0),($18|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $759 = __THREW__; __THREW__ = 0;
       $760 = $759&1;
       if ($760) {
        label = 238;
       }
      }
      if ((label|0) == 238) {
       $777 = ___cxa_find_matching_catch_2()|0;
       $778 = tempRet0;
       $779 = HEAP32[$739>>2]|0; //@line 106 "./common/Object.h"
       $780 = ((($779)) + 12|0); //@line 106 "./common/Object.h"
       $781 = HEAP32[$780>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($781|0,($739|0)); //@line 106 "./common/Object.h"
       $782 = __THREW__; __THREW__ = 0;
       $783 = $782&1;
       if (!($783)) {
        $$7 = $777;$$79 = $778;
        break L4;
       }
       $784 = ___cxa_find_matching_catch_3(0|0)|0;
       $785 = tempRet0;
       ___clang_call_terminate($784); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      $761 = HEAP32[$739>>2]|0; //@line 106 "./common/Object.h"
      $762 = ((($761)) + 12|0); //@line 106 "./common/Object.h"
      $763 = HEAP32[$762>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($763|0,($739|0)); //@line 106 "./common/Object.h"
      $764 = __THREW__; __THREW__ = 0;
      $765 = $764&1;
      if ($765) {
       $766 = ___cxa_find_matching_catch_3(0|0)|0;
       $767 = tempRet0;
       ___clang_call_terminate($766); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      __THREW__ = 0;
      $768 = (invoke_ii(3348,32)|0); //@line 499 "modules/event/sdl/Event.cpp"
      $769 = __THREW__; __THREW__ = 0;
      $770 = $769&1;
      if ($770) {
       label = 5;
       break L4;
      }
      __THREW__ = 0;
      invoke_viii(3460,($19|0),(365764|0),13); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $771 = __THREW__; __THREW__ = 0;
      $772 = $771&1;
      do {
       if ($772) {
        $786 = ___cxa_find_matching_catch_2()|0;
        $787 = tempRet0;
        $$5 = $786;$$57 = $787;
       } else {
        __THREW__ = 0;
        invoke_viii(3508,($768|0),($19|0),($vargs|0)); //@line 499 "modules/event/sdl/Event.cpp"
        $773 = __THREW__; __THREW__ = 0;
        $774 = $773&1;
        if ($774) {
         $788 = ___cxa_find_matching_catch_2()|0;
         $789 = tempRet0;
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($19); //@line 499 "modules/event/sdl/Event.cpp"
         $$5 = $788;$$57 = $789;
         break;
        } else {
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($19); //@line 499 "modules/event/sdl/Event.cpp"
         $msg$0 = $768;
         break L6;
        }
       }
      } while(0);
      __ZdlPv($768); //@line 499 "modules/event/sdl/Event.cpp"
      $$7 = $$5;$$79 = $$57;
      break L4;
     }
     break;
    }
    case 1542:  {
     $790 = HEAP32[$22>>2]|0; //@line 504 "modules/event/sdl/Event.cpp"
     $791 = ((($790)) + 32|0); //@line 504 "modules/event/sdl/Event.cpp"
     $792 = HEAP32[$791>>2]|0; //@line 504 "modules/event/sdl/Event.cpp"
     $793 = ((($e)) + 8|0); //@line 504 "modules/event/sdl/Event.cpp"
     $794 = HEAP32[$793>>2]|0; //@line 504 "modules/event/sdl/Event.cpp"
     __THREW__ = 0;
     $795 = (invoke_iii($792|0,($22|0),($794|0))|0); //@line 504 "modules/event/sdl/Event.cpp"
     $796 = __THREW__; __THREW__ = 0;
     $797 = $796&1;
     if ($797) {
      label = 5;
      break L4;
     }
     $798 = ((($proxy)) + 4|0); //@line 504 "modules/event/sdl/Event.cpp"
     HEAP32[$798>>2] = $795; //@line 504 "modules/event/sdl/Event.cpp"
     HEAP32[$proxy>>2] = 24; //@line 505 "modules/event/sdl/Event.cpp"
     $799 = ($795|0)==(0|0); //@line 506 "modules/event/sdl/Event.cpp"
     if ($799) {
      $msg$0 = 0;
     } else {
      $800 = HEAP32[$22>>2]|0; //@line 508 "modules/event/sdl/Event.cpp"
      $801 = ((($800)) + 28|0); //@line 508 "modules/event/sdl/Event.cpp"
      $802 = HEAP32[$801>>2]|0; //@line 508 "modules/event/sdl/Event.cpp"
      __THREW__ = 0;
      invoke_vii($802|0,($22|0),($795|0)); //@line 508 "modules/event/sdl/Event.cpp"
      $803 = __THREW__; __THREW__ = 0;
      $804 = $803&1;
      if ($804) {
       label = 5;
       break L4;
      }
      __THREW__ = 0;
      $805 = (invoke_ii(3348,32)|0); //@line 509 "modules/event/sdl/Event.cpp"
      $806 = __THREW__; __THREW__ = 0;
      $807 = $806&1;
      if ($807) {
       label = 5;
       break L4;
      }
      $808 = HEAP32[$proxy>>2]|0; //@line 509 "modules/event/sdl/Event.cpp"
      __THREW__ = 0;
      invoke_viii(3949,($805|0),($808|0),($proxy|0)); //@line 509 "modules/event/sdl/Event.cpp"
      $809 = __THREW__; __THREW__ = 0;
      $810 = $809&1;
      if ($810) {
       $842 = ___cxa_find_matching_catch_2()|0;
       $843 = tempRet0;
       __ZdlPv($805); //@line 509 "modules/event/sdl/Event.cpp"
       $$7 = $842;$$79 = $843;
       break L4;
      }
      HEAP32[$20>>2] = $805; //@line 93 "./common/Object.h"
      $811 = HEAP32[$805>>2]|0; //@line 95 "./common/Object.h"
      $812 = ((($811)) + 8|0); //@line 95 "./common/Object.h"
      $813 = HEAP32[$812>>2]|0; //@line 95 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($813|0,($805|0)); //@line 95 "./common/Object.h"
      $814 = __THREW__; __THREW__ = 0;
      $815 = $814&1;
      if ($815) {
       label = 5;
       break L4;
      }
      $816 = HEAP32[$24>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $817 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $818 = HEAP32[$817>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $819 = ($816>>>0)<($818>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($819) {
       HEAP32[$816>>2] = $805; //@line 99 "./common/Object.h"
       $820 = HEAP32[$805>>2]|0; //@line 101 "./common/Object.h"
       $821 = ((($820)) + 8|0); //@line 101 "./common/Object.h"
       $822 = HEAP32[$821>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($822|0,($805|0)); //@line 101 "./common/Object.h"
       $823 = __THREW__; __THREW__ = 0;
       $824 = $823&1;
       if ($824) {
        label = 260;
       } else {
        $$pre$i143 = HEAP32[$24>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $825 = ((($$pre$i143)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$24>>2] = $825; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
      } else {
       __THREW__ = 0;
       invoke_vii(3507,($vargs|0),($20|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $826 = __THREW__; __THREW__ = 0;
       $827 = $826&1;
       if ($827) {
        label = 260;
       }
      }
      if ((label|0) == 260) {
       $844 = ___cxa_find_matching_catch_2()|0;
       $845 = tempRet0;
       $846 = HEAP32[$805>>2]|0; //@line 106 "./common/Object.h"
       $847 = ((($846)) + 12|0); //@line 106 "./common/Object.h"
       $848 = HEAP32[$847>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($848|0,($805|0)); //@line 106 "./common/Object.h"
       $849 = __THREW__; __THREW__ = 0;
       $850 = $849&1;
       if (!($850)) {
        $$7 = $844;$$79 = $845;
        break L4;
       }
       $851 = ___cxa_find_matching_catch_3(0|0)|0;
       $852 = tempRet0;
       ___clang_call_terminate($851); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      $828 = HEAP32[$805>>2]|0; //@line 106 "./common/Object.h"
      $829 = ((($828)) + 12|0); //@line 106 "./common/Object.h"
      $830 = HEAP32[$829>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($830|0,($805|0)); //@line 106 "./common/Object.h"
      $831 = __THREW__; __THREW__ = 0;
      $832 = $831&1;
      if ($832) {
       $833 = ___cxa_find_matching_catch_3(0|0)|0;
       $834 = tempRet0;
       ___clang_call_terminate($833); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      __THREW__ = 0;
      $835 = (invoke_ii(3348,32)|0); //@line 510 "modules/event/sdl/Event.cpp"
      $836 = __THREW__; __THREW__ = 0;
      $837 = $836&1;
      if ($837) {
       label = 5;
       break L4;
      }
      __THREW__ = 0;
      invoke_viii(3460,($21|0),(365778|0),15); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $838 = __THREW__; __THREW__ = 0;
      $839 = $838&1;
      do {
       if ($839) {
        $853 = ___cxa_find_matching_catch_2()|0;
        $854 = tempRet0;
        $$6 = $853;$$68 = $854;
       } else {
        __THREW__ = 0;
        invoke_viii(3508,($835|0),($21|0),($vargs|0)); //@line 510 "modules/event/sdl/Event.cpp"
        $840 = __THREW__; __THREW__ = 0;
        $841 = $840&1;
        if ($841) {
         $855 = ___cxa_find_matching_catch_2()|0;
         $856 = tempRet0;
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($21); //@line 510 "modules/event/sdl/Event.cpp"
         $$6 = $855;$$68 = $856;
         break;
        } else {
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($21); //@line 510 "modules/event/sdl/Event.cpp"
         $msg$0 = $835;
         break L6;
        }
       }
      } while(0);
      __ZdlPv($835); //@line 510 "modules/event/sdl/Event.cpp"
      $$7 = $$6;$$79 = $$68;
      break L4;
     }
     break;
    }
    default: {
     $msg$0 = 0;
    }
    }
   } while(0);
   $857 = HEAP32[$vargs>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $858 = HEAP32[$24>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $859 = ($857|0)==($858|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   do {
    if ($859) {
     $861 = $857;
    } else {
     $__begin$sroa$0$0161 = $857;
     while(1) {
      $878 = HEAP32[$__begin$sroa$0$0161>>2]|0; //@line 139 "./common/Object.h"
      $879 = ($878|0)==(0|0); //@line 520 "modules/event/sdl/Event.cpp"
      if (!($879)) {
       $880 = HEAP32[$878>>2]|0; //@line 521 "modules/event/sdl/Event.cpp"
       $881 = ((($880)) + 12|0); //@line 521 "modules/event/sdl/Event.cpp"
       $882 = HEAP32[$881>>2]|0; //@line 521 "modules/event/sdl/Event.cpp"
       __THREW__ = 0;
       invoke_vi($882|0,($878|0)); //@line 521 "modules/event/sdl/Event.cpp"
       $883 = __THREW__; __THREW__ = 0;
       $884 = $883&1;
       if ($884) {
        label = 4;
        break;
       }
      }
      $885 = ((($__begin$sroa$0$0161)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $886 = ($885|0)==($858|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($886) {
       label = 266;
       break;
      } else {
       $__begin$sroa$0$0161 = $885;
      }
     }
     if ((label|0) == 4) {
      $29 = ___cxa_find_matching_catch_2()|0;
      $30 = tempRet0;
      $lpad$phi$index = $29;$lpad$phi$index2 = $30;
      label = 6;
      break L4;
     }
     else if ((label|0) == 266) {
      $$pre = HEAP32[$vargs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $861 = $$pre;
      break;
     }
    }
   } while(0);
   $860 = ($861|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($860)) {
    $862 = HEAP32[$24>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $863 = ($862|0)==($861|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    do {
     if ($863) {
      $877 = $861;
     } else {
      $865 = $862;
      while(1) {
       $864 = ((($865)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$24>>2] = $864; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $866 = HEAP32[$864>>2]|0; //@line 106 "./common/Object.h"
       $867 = ($866|0)==(0|0); //@line 106 "./common/Object.h"
       if ($867) {
        $869 = $864;
       } else {
        $870 = HEAP32[$866>>2]|0; //@line 106 "./common/Object.h"
        $871 = ((($870)) + 12|0); //@line 106 "./common/Object.h"
        $872 = HEAP32[$871>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($872|0,($866|0)); //@line 106 "./common/Object.h"
        $873 = __THREW__; __THREW__ = 0;
        $874 = $873&1;
        if ($874) {
         label = 273;
         break;
        }
        $$pre$i$i$i154 = HEAP32[$24>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $869 = $$pre$i$i$i154;
       }
       $868 = ($869|0)==($861|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($868) {
        label = 274;
        break;
       } else {
        $865 = $869;
       }
      }
      if ((label|0) == 273) {
       $875 = ___cxa_find_matching_catch_3(0|0)|0;
       $876 = tempRet0;
       ___clang_call_terminate($875); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      else if ((label|0) == 274) {
       $$pre$i157 = HEAP32[$vargs>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $877 = $$pre$i157;
       break;
      }
     }
    } while(0);
    __ZdlPv($877); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $$0 = $msg$0;
   STACKTOP = sp;return ($$0|0); //@line 525 "modules/event/sdl/Event.cpp"
  }
 } while(0);
 if ((label|0) == 5) {
  $31 = ___cxa_find_matching_catch_2()|0;
  $32 = tempRet0;
  $lpad$phi$index = $31;$lpad$phi$index2 = $32;
  label = 6;
 }
 if ((label|0) == 6) {
  $$7 = $lpad$phi$index;$$79 = $lpad$phi$index2;
 }
 $887 = HEAP32[$vargs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $888 = ($887|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($888) {
  ___resumeException($$7|0);
  // unreachable;
 }
 $889 = HEAP32[$24>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $890 = ($889|0)==($887|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($890) {
   $904 = $887;
  } else {
   $892 = $889;
   while(1) {
    $891 = ((($892)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$24>>2] = $891; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $893 = HEAP32[$891>>2]|0; //@line 106 "./common/Object.h"
    $894 = ($893|0)==(0|0); //@line 106 "./common/Object.h"
    if ($894) {
     $896 = $891;
    } else {
     $897 = HEAP32[$893>>2]|0; //@line 106 "./common/Object.h"
     $898 = ((($897)) + 12|0); //@line 106 "./common/Object.h"
     $899 = HEAP32[$898>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($899|0,($893|0)); //@line 106 "./common/Object.h"
     $900 = __THREW__; __THREW__ = 0;
     $901 = $900&1;
     if ($901) {
      label = 286;
      break;
     }
     $$pre$i$i$i = HEAP32[$24>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $896 = $$pre$i$i$i;
    }
    $895 = ($896|0)==($887|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($895) {
     label = 287;
     break;
    } else {
     $892 = $896;
    }
   }
   if ((label|0) == 286) {
    $902 = ___cxa_find_matching_catch_3(0|0)|0;
    $903 = tempRet0;
    ___clang_call_terminate($902); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 287) {
    $$pre$i = HEAP32[$vargs>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $904 = $$pre$i;
    break;
   }
  }
 } while(0);
 __ZdlPv($904); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$7|0);
 // unreachable;
 return (0)|0;
}
function __ZNK4love5event3sdl5Event18convertWindowEventERK9SDL_Event($this,$e) {
 $this = $this|0;
 $e = $e|0;
 var $$0 = 0, $$0$ph = 0, $$01 = 0, $$02 = 0, $$1 = 0, $$13 = 0, $$2 = 0, $$24 = 0, $$3 = 0, $$35 = 0, $$4 = 0, $$46 = 0, $$pr = 0, $$pre$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i73 = 0, $$pre$i13 = 0, $$pre$i24 = 0, $$pre$i35 = 0, $$pre$i43 = 0;
 var $$pre$i51 = 0, $$pre$i59 = 0, $$pre$i7 = 0, $$pre$i76 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0.0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0.0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0.0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0;
 var $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0.0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0;
 var $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0;
 var $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0;
 var $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0;
 var $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0;
 var $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0;
 var $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0;
 var $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0;
 var $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$081 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $msg$0 = 0, $px_h = 0, $px_w = 0, $vargs = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vargs = sp + 84|0;
 $0 = sp + 80|0;
 $1 = sp + 68|0;
 $2 = sp + 64|0;
 $3 = sp + 52|0;
 $4 = sp + 48|0;
 $5 = sp + 36|0;
 $px_w = sp + 32|0;
 $px_h = sp + 28|0;
 $6 = sp + 24|0;
 $7 = sp + 20|0;
 $8 = sp + 16|0;
 $9 = sp + 12|0;
 $10 = sp;
 HEAP32[$vargs>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ((($vargs)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$11>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ((($vargs)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$12>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 invoke_vii(3936,($vargs|0),4); //@line 532 "modules/event/sdl/Event.cpp"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 L1: do {
  if ($14) {
   label = 4;
  } else {
   $15 = HEAP32[$e>>2]|0; //@line 536 "modules/event/sdl/Event.cpp"
   $16 = ($15|0)==(512); //@line 536 "modules/event/sdl/Event.cpp"
   L3: do {
    if ($16) {
     $21 = ((($e)) + 12|0); //@line 539 "modules/event/sdl/Event.cpp"
     $22 = HEAP8[$21>>0]|0; //@line 543 "modules/event/sdl/Event.cpp"
     $23 = $22&255; //@line 539 "modules/event/sdl/Event.cpp"
     L5: do {
      switch ($23|0) {
      case 13: case 12:  {
       __THREW__ = 0;
       $24 = (invoke_ii(3348,32)|0); //@line 543 "modules/event/sdl/Event.cpp"
       $25 = __THREW__; __THREW__ = 0;
       $26 = $25&1;
       if ($26) {
        label = 4;
        break L1;
       }
       $27 = ($22<<24>>24)==(12); //@line 543 "modules/event/sdl/Event.cpp"
       __THREW__ = 0;
       invoke_vii(3941,($24|0),($27|0)); //@line 543 "modules/event/sdl/Event.cpp"
       $28 = __THREW__; __THREW__ = 0;
       $29 = $28&1;
       if ($29) {
        $61 = ___cxa_find_matching_catch_2()|0;
        $62 = tempRet0;
        __ZdlPv($24); //@line 543 "modules/event/sdl/Event.cpp"
        $$4 = $61;$$46 = $62;
        break L1;
       }
       HEAP32[$0>>2] = $24; //@line 93 "./common/Object.h"
       $30 = HEAP32[$24>>2]|0; //@line 95 "./common/Object.h"
       $31 = ((($30)) + 8|0); //@line 95 "./common/Object.h"
       $32 = HEAP32[$31>>2]|0; //@line 95 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($32|0,($24|0)); //@line 95 "./common/Object.h"
       $33 = __THREW__; __THREW__ = 0;
       $34 = $33&1;
       if ($34) {
        label = 4;
        break L1;
       }
       $35 = HEAP32[$11>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $36 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $37 = HEAP32[$36>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $38 = ($35>>>0)<($37>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($38) {
        HEAP32[$35>>2] = $24; //@line 99 "./common/Object.h"
        $39 = HEAP32[$24>>2]|0; //@line 101 "./common/Object.h"
        $40 = ((($39)) + 8|0); //@line 101 "./common/Object.h"
        $41 = HEAP32[$40>>2]|0; //@line 101 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($41|0,($24|0)); //@line 101 "./common/Object.h"
        $42 = __THREW__; __THREW__ = 0;
        $43 = $42&1;
        if ($43) {
         label = 21;
        } else {
         $$pre$i7 = HEAP32[$11>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $44 = ((($$pre$i7)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$11>>2] = $44; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
       } else {
        __THREW__ = 0;
        invoke_vii(3507,($vargs|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $45 = __THREW__; __THREW__ = 0;
        $46 = $45&1;
        if ($46) {
         label = 21;
        }
       }
       if ((label|0) == 21) {
        $63 = ___cxa_find_matching_catch_2()|0;
        $64 = tempRet0;
        $65 = HEAP32[$24>>2]|0; //@line 106 "./common/Object.h"
        $66 = ((($65)) + 12|0); //@line 106 "./common/Object.h"
        $67 = HEAP32[$66>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($67|0,($24|0)); //@line 106 "./common/Object.h"
        $68 = __THREW__; __THREW__ = 0;
        $69 = $68&1;
        if (!($69)) {
         $$4 = $63;$$46 = $64;
         break L1;
        }
        $70 = ___cxa_find_matching_catch_3(0|0)|0;
        $71 = tempRet0;
        ___clang_call_terminate($70); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       $47 = HEAP32[$24>>2]|0; //@line 106 "./common/Object.h"
       $48 = ((($47)) + 12|0); //@line 106 "./common/Object.h"
       $49 = HEAP32[$48>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($49|0,($24|0)); //@line 106 "./common/Object.h"
       $50 = __THREW__; __THREW__ = 0;
       $51 = $50&1;
       if ($51) {
        $52 = ___cxa_find_matching_catch_3(0|0)|0;
        $53 = tempRet0;
        ___clang_call_terminate($52); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       __THREW__ = 0;
       $54 = (invoke_ii(3348,32)|0); //@line 544 "modules/event/sdl/Event.cpp"
       $55 = __THREW__; __THREW__ = 0;
       $56 = $55&1;
       if ($56) {
        label = 4;
        break L1;
       }
       __THREW__ = 0;
       invoke_viii(3460,($1|0),(365794|0),5); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $57 = __THREW__; __THREW__ = 0;
       $58 = $57&1;
       do {
        if ($58) {
         $72 = ___cxa_find_matching_catch_2()|0;
         $73 = tempRet0;
         $$01 = $72;$$02 = $73;
        } else {
         __THREW__ = 0;
         invoke_viii(3508,($54|0),($1|0),($vargs|0)); //@line 544 "modules/event/sdl/Event.cpp"
         $59 = __THREW__; __THREW__ = 0;
         $60 = $59&1;
         if ($60) {
          $74 = ___cxa_find_matching_catch_2()|0;
          $75 = tempRet0;
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 544 "modules/event/sdl/Event.cpp"
          $$01 = $74;$$02 = $75;
          break;
         } else {
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 544 "modules/event/sdl/Event.cpp"
          $msg$0 = $54;
          break L5;
         }
        }
       } while(0);
       __ZdlPv($54); //@line 544 "modules/event/sdl/Event.cpp"
       $$4 = $$01;$$46 = $$02;
       break L1;
       break;
      }
      case 11: case 10:  {
       __THREW__ = 0;
       $76 = (invoke_ii(3348,32)|0); //@line 548 "modules/event/sdl/Event.cpp"
       $77 = __THREW__; __THREW__ = 0;
       $78 = $77&1;
       if ($78) {
        label = 4;
        break L1;
       }
       $79 = ($22<<24>>24)==(10); //@line 548 "modules/event/sdl/Event.cpp"
       __THREW__ = 0;
       invoke_vii(3941,($76|0),($79|0)); //@line 548 "modules/event/sdl/Event.cpp"
       $80 = __THREW__; __THREW__ = 0;
       $81 = $80&1;
       if ($81) {
        $113 = ___cxa_find_matching_catch_2()|0;
        $114 = tempRet0;
        __ZdlPv($76); //@line 548 "modules/event/sdl/Event.cpp"
        $$4 = $113;$$46 = $114;
        break L1;
       }
       HEAP32[$2>>2] = $76; //@line 93 "./common/Object.h"
       $82 = HEAP32[$76>>2]|0; //@line 95 "./common/Object.h"
       $83 = ((($82)) + 8|0); //@line 95 "./common/Object.h"
       $84 = HEAP32[$83>>2]|0; //@line 95 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($84|0,($76|0)); //@line 95 "./common/Object.h"
       $85 = __THREW__; __THREW__ = 0;
       $86 = $85&1;
       if ($86) {
        label = 4;
        break L1;
       }
       $87 = HEAP32[$11>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $88 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $89 = HEAP32[$88>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $90 = ($87>>>0)<($89>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($90) {
        HEAP32[$87>>2] = $76; //@line 99 "./common/Object.h"
        $91 = HEAP32[$76>>2]|0; //@line 101 "./common/Object.h"
        $92 = ((($91)) + 8|0); //@line 101 "./common/Object.h"
        $93 = HEAP32[$92>>2]|0; //@line 101 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($93|0,($76|0)); //@line 101 "./common/Object.h"
        $94 = __THREW__; __THREW__ = 0;
        $95 = $94&1;
        if ($95) {
         label = 40;
        } else {
         $$pre$i13 = HEAP32[$11>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $96 = ((($$pre$i13)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$11>>2] = $96; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
       } else {
        __THREW__ = 0;
        invoke_vii(3507,($vargs|0),($2|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $97 = __THREW__; __THREW__ = 0;
        $98 = $97&1;
        if ($98) {
         label = 40;
        }
       }
       if ((label|0) == 40) {
        $115 = ___cxa_find_matching_catch_2()|0;
        $116 = tempRet0;
        $117 = HEAP32[$76>>2]|0; //@line 106 "./common/Object.h"
        $118 = ((($117)) + 12|0); //@line 106 "./common/Object.h"
        $119 = HEAP32[$118>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($119|0,($76|0)); //@line 106 "./common/Object.h"
        $120 = __THREW__; __THREW__ = 0;
        $121 = $120&1;
        if (!($121)) {
         $$4 = $115;$$46 = $116;
         break L1;
        }
        $122 = ___cxa_find_matching_catch_3(0|0)|0;
        $123 = tempRet0;
        ___clang_call_terminate($122); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       $99 = HEAP32[$76>>2]|0; //@line 106 "./common/Object.h"
       $100 = ((($99)) + 12|0); //@line 106 "./common/Object.h"
       $101 = HEAP32[$100>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($101|0,($76|0)); //@line 106 "./common/Object.h"
       $102 = __THREW__; __THREW__ = 0;
       $103 = $102&1;
       if ($103) {
        $104 = ___cxa_find_matching_catch_3(0|0)|0;
        $105 = tempRet0;
        ___clang_call_terminate($104); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       __THREW__ = 0;
       $106 = (invoke_ii(3348,32)|0); //@line 549 "modules/event/sdl/Event.cpp"
       $107 = __THREW__; __THREW__ = 0;
       $108 = $107&1;
       if ($108) {
        label = 4;
        break L1;
       }
       __THREW__ = 0;
       invoke_viii(3460,($3|0),(365800|0),10); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $109 = __THREW__; __THREW__ = 0;
       $110 = $109&1;
       do {
        if ($110) {
         $124 = ___cxa_find_matching_catch_2()|0;
         $125 = tempRet0;
         $$1 = $124;$$13 = $125;
        } else {
         __THREW__ = 0;
         invoke_viii(3508,($106|0),($3|0),($vargs|0)); //@line 549 "modules/event/sdl/Event.cpp"
         $111 = __THREW__; __THREW__ = 0;
         $112 = $111&1;
         if ($112) {
          $126 = ___cxa_find_matching_catch_2()|0;
          $127 = tempRet0;
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 549 "modules/event/sdl/Event.cpp"
          $$1 = $126;$$13 = $127;
          break;
         } else {
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 549 "modules/event/sdl/Event.cpp"
          $msg$0 = $106;
          break L5;
         }
        }
       } while(0);
       __ZdlPv($106); //@line 549 "modules/event/sdl/Event.cpp"
       $$4 = $$1;$$46 = $$13;
       break L1;
       break;
      }
      case 2: case 1:  {
       __THREW__ = 0;
       $128 = (invoke_ii(3348,32)|0); //@line 553 "modules/event/sdl/Event.cpp"
       $129 = __THREW__; __THREW__ = 0;
       $130 = $129&1;
       if ($130) {
        label = 4;
        break L1;
       }
       $131 = ($22<<24>>24)==(1); //@line 553 "modules/event/sdl/Event.cpp"
       __THREW__ = 0;
       invoke_vii(3941,($128|0),($131|0)); //@line 553 "modules/event/sdl/Event.cpp"
       $132 = __THREW__; __THREW__ = 0;
       $133 = $132&1;
       if ($133) {
        $165 = ___cxa_find_matching_catch_2()|0;
        $166 = tempRet0;
        __ZdlPv($128); //@line 553 "modules/event/sdl/Event.cpp"
        $$4 = $165;$$46 = $166;
        break L1;
       }
       HEAP32[$4>>2] = $128; //@line 93 "./common/Object.h"
       $134 = HEAP32[$128>>2]|0; //@line 95 "./common/Object.h"
       $135 = ((($134)) + 8|0); //@line 95 "./common/Object.h"
       $136 = HEAP32[$135>>2]|0; //@line 95 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($136|0,($128|0)); //@line 95 "./common/Object.h"
       $137 = __THREW__; __THREW__ = 0;
       $138 = $137&1;
       if ($138) {
        label = 4;
        break L1;
       }
       $139 = HEAP32[$11>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $140 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $141 = HEAP32[$140>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $142 = ($139>>>0)<($141>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($142) {
        HEAP32[$139>>2] = $128; //@line 99 "./common/Object.h"
        $143 = HEAP32[$128>>2]|0; //@line 101 "./common/Object.h"
        $144 = ((($143)) + 8|0); //@line 101 "./common/Object.h"
        $145 = HEAP32[$144>>2]|0; //@line 101 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($145|0,($128|0)); //@line 101 "./common/Object.h"
        $146 = __THREW__; __THREW__ = 0;
        $147 = $146&1;
        if ($147) {
         label = 59;
        } else {
         $$pre$i24 = HEAP32[$11>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $148 = ((($$pre$i24)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$11>>2] = $148; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
       } else {
        __THREW__ = 0;
        invoke_vii(3507,($vargs|0),($4|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $149 = __THREW__; __THREW__ = 0;
        $150 = $149&1;
        if ($150) {
         label = 59;
        }
       }
       if ((label|0) == 59) {
        $167 = ___cxa_find_matching_catch_2()|0;
        $168 = tempRet0;
        $169 = HEAP32[$128>>2]|0; //@line 106 "./common/Object.h"
        $170 = ((($169)) + 12|0); //@line 106 "./common/Object.h"
        $171 = HEAP32[$170>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($171|0,($128|0)); //@line 106 "./common/Object.h"
        $172 = __THREW__; __THREW__ = 0;
        $173 = $172&1;
        if (!($173)) {
         $$4 = $167;$$46 = $168;
         break L1;
        }
        $174 = ___cxa_find_matching_catch_3(0|0)|0;
        $175 = tempRet0;
        ___clang_call_terminate($174); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       $151 = HEAP32[$128>>2]|0; //@line 106 "./common/Object.h"
       $152 = ((($151)) + 12|0); //@line 106 "./common/Object.h"
       $153 = HEAP32[$152>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($153|0,($128|0)); //@line 106 "./common/Object.h"
       $154 = __THREW__; __THREW__ = 0;
       $155 = $154&1;
       if ($155) {
        $156 = ___cxa_find_matching_catch_3(0|0)|0;
        $157 = tempRet0;
        ___clang_call_terminate($156); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       __THREW__ = 0;
       $158 = (invoke_ii(3348,32)|0); //@line 554 "modules/event/sdl/Event.cpp"
       $159 = __THREW__; __THREW__ = 0;
       $160 = $159&1;
       if ($160) {
        label = 4;
        break L1;
       }
       __THREW__ = 0;
       invoke_viii(3460,($5|0),(365811|0),7); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $161 = __THREW__; __THREW__ = 0;
       $162 = $161&1;
       do {
        if ($162) {
         $176 = ___cxa_find_matching_catch_2()|0;
         $177 = tempRet0;
         $$2 = $176;$$24 = $177;
        } else {
         __THREW__ = 0;
         invoke_viii(3508,($158|0),($5|0),($vargs|0)); //@line 554 "modules/event/sdl/Event.cpp"
         $163 = __THREW__; __THREW__ = 0;
         $164 = $163&1;
         if ($164) {
          $178 = ___cxa_find_matching_catch_2()|0;
          $179 = tempRet0;
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5); //@line 554 "modules/event/sdl/Event.cpp"
          $$2 = $178;$$24 = $179;
          break;
         } else {
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5); //@line 554 "modules/event/sdl/Event.cpp"
          $msg$0 = $158;
          break L5;
         }
        }
       } while(0);
       __ZdlPv($158); //@line 554 "modules/event/sdl/Event.cpp"
       $$4 = $$2;$$46 = $$24;
       break L1;
       break;
      }
      case 5:  {
       $180 = ((($e)) + 16|0); //@line 558 "modules/event/sdl/Event.cpp"
       $181 = HEAP32[$180>>2]|0; //@line 558 "modules/event/sdl/Event.cpp"
       HEAP32[$px_w>>2] = $181; //@line 558 "modules/event/sdl/Event.cpp"
       $182 = ((($e)) + 20|0); //@line 559 "modules/event/sdl/Event.cpp"
       $183 = HEAP32[$182>>2]|0; //@line 559 "modules/event/sdl/Event.cpp"
       HEAP32[$px_h>>2] = $183; //@line 559 "modules/event/sdl/Event.cpp"
       $184 = ((($e)) + 8|0); //@line 561 "modules/event/sdl/Event.cpp"
       $185 = HEAP32[$184>>2]|0; //@line 561 "modules/event/sdl/Event.cpp"
       __THREW__ = 0;
       $186 = (invoke_ii(3956,($185|0))|0); //@line 561 "modules/event/sdl/Event.cpp"
       $187 = __THREW__; __THREW__ = 0;
       $188 = $187&1;
       if ($188) {
        label = 4;
        break L1;
       }
       $189 = ($186|0)==(0|0); //@line 562 "modules/event/sdl/Event.cpp"
       if (!($189)) {
        __THREW__ = 0;
        invoke_viii(3957,($186|0),($px_w|0),($px_h|0)); //@line 563 "modules/event/sdl/Event.cpp"
        $190 = __THREW__; __THREW__ = 0;
        $191 = $190&1;
        if ($191) {
         label = 4;
         break L1;
        }
       }
       __THREW__ = 0;
       $192 = (invoke_ii(3348,32)|0); //@line 565 "modules/event/sdl/Event.cpp"
       $193 = __THREW__; __THREW__ = 0;
       $194 = $193&1;
       if ($194) {
        label = 4;
        break L1;
       }
       $195 = HEAP32[$px_w>>2]|0; //@line 565 "modules/event/sdl/Event.cpp"
       $196 = (+($195|0)); //@line 565 "modules/event/sdl/Event.cpp"
       __THREW__ = 0;
       invoke_vid(3942,($192|0),(+$196)); //@line 565 "modules/event/sdl/Event.cpp"
       $197 = __THREW__; __THREW__ = 0;
       $198 = $197&1;
       if ($198) {
        $320 = ___cxa_find_matching_catch_2()|0;
        $321 = tempRet0;
        __ZdlPv($192); //@line 565 "modules/event/sdl/Event.cpp"
        $$4 = $320;$$46 = $321;
        break L1;
       }
       HEAP32[$6>>2] = $192; //@line 93 "./common/Object.h"
       $199 = HEAP32[$192>>2]|0; //@line 95 "./common/Object.h"
       $200 = ((($199)) + 8|0); //@line 95 "./common/Object.h"
       $201 = HEAP32[$200>>2]|0; //@line 95 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($201|0,($192|0)); //@line 95 "./common/Object.h"
       $202 = __THREW__; __THREW__ = 0;
       $203 = $202&1;
       if ($203) {
        label = 4;
        break L1;
       }
       $204 = HEAP32[$11>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $205 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $206 = HEAP32[$205>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $207 = ($204>>>0)<($206>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($207) {
        HEAP32[$204>>2] = $192; //@line 99 "./common/Object.h"
        $208 = HEAP32[$192>>2]|0; //@line 101 "./common/Object.h"
        $209 = ((($208)) + 8|0); //@line 101 "./common/Object.h"
        $210 = HEAP32[$209>>2]|0; //@line 101 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($210|0,($192|0)); //@line 101 "./common/Object.h"
        $211 = __THREW__; __THREW__ = 0;
        $212 = $211&1;
        if ($212) {
         label = 108;
        } else {
         $$pre$i35 = HEAP32[$11>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $213 = ((($$pre$i35)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$11>>2] = $213; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
       } else {
        __THREW__ = 0;
        invoke_vii(3507,($vargs|0),($6|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $214 = __THREW__; __THREW__ = 0;
        $215 = $214&1;
        if ($215) {
         label = 108;
        }
       }
       if ((label|0) == 108) {
        $322 = ___cxa_find_matching_catch_2()|0;
        $323 = tempRet0;
        $324 = HEAP32[$192>>2]|0; //@line 106 "./common/Object.h"
        $325 = ((($324)) + 12|0); //@line 106 "./common/Object.h"
        $326 = HEAP32[$325>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($326|0,($192|0)); //@line 106 "./common/Object.h"
        $327 = __THREW__; __THREW__ = 0;
        $328 = $327&1;
        if (!($328)) {
         $$4 = $322;$$46 = $323;
         break L1;
        }
        $329 = ___cxa_find_matching_catch_3(0|0)|0;
        $330 = tempRet0;
        ___clang_call_terminate($329); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       $216 = HEAP32[$192>>2]|0; //@line 106 "./common/Object.h"
       $217 = ((($216)) + 12|0); //@line 106 "./common/Object.h"
       $218 = HEAP32[$217>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($218|0,($192|0)); //@line 106 "./common/Object.h"
       $219 = __THREW__; __THREW__ = 0;
       $220 = $219&1;
       if ($220) {
        $221 = ___cxa_find_matching_catch_3(0|0)|0;
        $222 = tempRet0;
        ___clang_call_terminate($221); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       __THREW__ = 0;
       $223 = (invoke_ii(3348,32)|0); //@line 566 "modules/event/sdl/Event.cpp"
       $224 = __THREW__; __THREW__ = 0;
       $225 = $224&1;
       if ($225) {
        label = 4;
        break L1;
       }
       $226 = HEAP32[$px_h>>2]|0; //@line 566 "modules/event/sdl/Event.cpp"
       $227 = (+($226|0)); //@line 566 "modules/event/sdl/Event.cpp"
       __THREW__ = 0;
       invoke_vid(3942,($223|0),(+$227)); //@line 566 "modules/event/sdl/Event.cpp"
       $228 = __THREW__; __THREW__ = 0;
       $229 = $228&1;
       if ($229) {
        $331 = ___cxa_find_matching_catch_2()|0;
        $332 = tempRet0;
        __ZdlPv($223); //@line 566 "modules/event/sdl/Event.cpp"
        $$4 = $331;$$46 = $332;
        break L1;
       }
       HEAP32[$7>>2] = $223; //@line 93 "./common/Object.h"
       $230 = HEAP32[$223>>2]|0; //@line 95 "./common/Object.h"
       $231 = ((($230)) + 8|0); //@line 95 "./common/Object.h"
       $232 = HEAP32[$231>>2]|0; //@line 95 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($232|0,($223|0)); //@line 95 "./common/Object.h"
       $233 = __THREW__; __THREW__ = 0;
       $234 = $233&1;
       if ($234) {
        label = 4;
        break L1;
       }
       $235 = HEAP32[$11>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $236 = HEAP32[$205>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $237 = ($235>>>0)<($236>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($237) {
        HEAP32[$235>>2] = $223; //@line 99 "./common/Object.h"
        $238 = HEAP32[$223>>2]|0; //@line 101 "./common/Object.h"
        $239 = ((($238)) + 8|0); //@line 101 "./common/Object.h"
        $240 = HEAP32[$239>>2]|0; //@line 101 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($240|0,($223|0)); //@line 101 "./common/Object.h"
        $241 = __THREW__; __THREW__ = 0;
        $242 = $241&1;
        if ($242) {
         label = 111;
        } else {
         $$pre$i43 = HEAP32[$11>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $243 = ((($$pre$i43)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$11>>2] = $243; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
       } else {
        __THREW__ = 0;
        invoke_vii(3507,($vargs|0),($7|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $244 = __THREW__; __THREW__ = 0;
        $245 = $244&1;
        if ($245) {
         label = 111;
        }
       }
       if ((label|0) == 111) {
        $333 = ___cxa_find_matching_catch_2()|0;
        $334 = tempRet0;
        $335 = HEAP32[$223>>2]|0; //@line 106 "./common/Object.h"
        $336 = ((($335)) + 12|0); //@line 106 "./common/Object.h"
        $337 = HEAP32[$336>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($337|0,($223|0)); //@line 106 "./common/Object.h"
        $338 = __THREW__; __THREW__ = 0;
        $339 = $338&1;
        if (!($339)) {
         $$4 = $333;$$46 = $334;
         break L1;
        }
        $340 = ___cxa_find_matching_catch_3(0|0)|0;
        $341 = tempRet0;
        ___clang_call_terminate($340); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       $246 = HEAP32[$223>>2]|0; //@line 106 "./common/Object.h"
       $247 = ((($246)) + 12|0); //@line 106 "./common/Object.h"
       $248 = HEAP32[$247>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($248|0,($223|0)); //@line 106 "./common/Object.h"
       $249 = __THREW__; __THREW__ = 0;
       $250 = $249&1;
       if ($250) {
        $251 = ___cxa_find_matching_catch_3(0|0)|0;
        $252 = tempRet0;
        ___clang_call_terminate($251); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       __THREW__ = 0;
       $253 = (invoke_ii(3348,32)|0); //@line 567 "modules/event/sdl/Event.cpp"
       $254 = __THREW__; __THREW__ = 0;
       $255 = $254&1;
       if ($255) {
        label = 4;
        break L1;
       }
       $256 = HEAP32[$180>>2]|0; //@line 567 "modules/event/sdl/Event.cpp"
       $257 = (+($256|0)); //@line 567 "modules/event/sdl/Event.cpp"
       __THREW__ = 0;
       invoke_vid(3942,($253|0),(+$257)); //@line 567 "modules/event/sdl/Event.cpp"
       $258 = __THREW__; __THREW__ = 0;
       $259 = $258&1;
       if ($259) {
        $342 = ___cxa_find_matching_catch_2()|0;
        $343 = tempRet0;
        __ZdlPv($253); //@line 567 "modules/event/sdl/Event.cpp"
        $$4 = $342;$$46 = $343;
        break L1;
       }
       HEAP32[$8>>2] = $253; //@line 93 "./common/Object.h"
       $260 = HEAP32[$253>>2]|0; //@line 95 "./common/Object.h"
       $261 = ((($260)) + 8|0); //@line 95 "./common/Object.h"
       $262 = HEAP32[$261>>2]|0; //@line 95 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($262|0,($253|0)); //@line 95 "./common/Object.h"
       $263 = __THREW__; __THREW__ = 0;
       $264 = $263&1;
       if ($264) {
        label = 4;
        break L1;
       }
       $265 = HEAP32[$11>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $266 = HEAP32[$205>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $267 = ($265>>>0)<($266>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       do {
        if ($267) {
         HEAP32[$265>>2] = $253; //@line 99 "./common/Object.h"
         $268 = HEAP32[$253>>2]|0; //@line 101 "./common/Object.h"
         $269 = ((($268)) + 8|0); //@line 101 "./common/Object.h"
         $270 = HEAP32[$269>>2]|0; //@line 101 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($270|0,($253|0)); //@line 101 "./common/Object.h"
         $271 = __THREW__; __THREW__ = 0;
         $272 = $271&1;
         if ($272) {
          label = 114;
          break;
         }
         $$pre$i51 = HEAP32[$11>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $273 = ((($$pre$i51)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$11>>2] = $273; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        } else {
         __THREW__ = 0;
         invoke_vii(3507,($vargs|0),($8|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $274 = __THREW__; __THREW__ = 0;
         $275 = $274&1;
         if ($275) {
          label = 114;
         }
        }
       } while(0);
       if ((label|0) == 114) {
        $344 = ___cxa_find_matching_catch_2()|0;
        $345 = tempRet0;
        $346 = HEAP32[$253>>2]|0; //@line 106 "./common/Object.h"
        $347 = ((($346)) + 12|0); //@line 106 "./common/Object.h"
        $348 = HEAP32[$347>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($348|0,($253|0)); //@line 106 "./common/Object.h"
        $349 = __THREW__; __THREW__ = 0;
        $350 = $349&1;
        if (!($350)) {
         $$4 = $344;$$46 = $345;
         break L1;
        }
        $351 = ___cxa_find_matching_catch_3(0|0)|0;
        $352 = tempRet0;
        ___clang_call_terminate($351); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       $276 = HEAP32[$253>>2]|0; //@line 106 "./common/Object.h"
       $277 = ((($276)) + 12|0); //@line 106 "./common/Object.h"
       $278 = HEAP32[$277>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($278|0,($253|0)); //@line 106 "./common/Object.h"
       $279 = __THREW__; __THREW__ = 0;
       $280 = $279&1;
       if ($280) {
        $281 = ___cxa_find_matching_catch_3(0|0)|0;
        $282 = tempRet0;
        ___clang_call_terminate($281); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       __THREW__ = 0;
       $283 = (invoke_ii(3348,32)|0); //@line 568 "modules/event/sdl/Event.cpp"
       $284 = __THREW__; __THREW__ = 0;
       $285 = $284&1;
       if ($285) {
        label = 4;
        break L1;
       }
       $286 = HEAP32[$182>>2]|0; //@line 568 "modules/event/sdl/Event.cpp"
       $287 = (+($286|0)); //@line 568 "modules/event/sdl/Event.cpp"
       __THREW__ = 0;
       invoke_vid(3942,($283|0),(+$287)); //@line 568 "modules/event/sdl/Event.cpp"
       $288 = __THREW__; __THREW__ = 0;
       $289 = $288&1;
       if ($289) {
        $353 = ___cxa_find_matching_catch_2()|0;
        $354 = tempRet0;
        __ZdlPv($283); //@line 568 "modules/event/sdl/Event.cpp"
        $$4 = $353;$$46 = $354;
        break L1;
       }
       HEAP32[$9>>2] = $283; //@line 93 "./common/Object.h"
       $290 = HEAP32[$283>>2]|0; //@line 95 "./common/Object.h"
       $291 = ((($290)) + 8|0); //@line 95 "./common/Object.h"
       $292 = HEAP32[$291>>2]|0; //@line 95 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($292|0,($283|0)); //@line 95 "./common/Object.h"
       $293 = __THREW__; __THREW__ = 0;
       $294 = $293&1;
       if ($294) {
        label = 4;
        break L1;
       }
       $295 = HEAP32[$11>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $296 = HEAP32[$205>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $297 = ($295>>>0)<($296>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       do {
        if ($297) {
         HEAP32[$295>>2] = $283; //@line 99 "./common/Object.h"
         $298 = HEAP32[$283>>2]|0; //@line 101 "./common/Object.h"
         $299 = ((($298)) + 8|0); //@line 101 "./common/Object.h"
         $300 = HEAP32[$299>>2]|0; //@line 101 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($300|0,($283|0)); //@line 101 "./common/Object.h"
         $301 = __THREW__; __THREW__ = 0;
         $302 = $301&1;
         if ($302) {
          label = 117;
          break;
         }
         $$pre$i59 = HEAP32[$11>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $303 = ((($$pre$i59)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$11>>2] = $303; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        } else {
         __THREW__ = 0;
         invoke_vii(3507,($vargs|0),($9|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $304 = __THREW__; __THREW__ = 0;
         $305 = $304&1;
         if ($305) {
          label = 117;
         }
        }
       } while(0);
       if ((label|0) == 117) {
        $355 = ___cxa_find_matching_catch_2()|0;
        $356 = tempRet0;
        $357 = HEAP32[$283>>2]|0; //@line 106 "./common/Object.h"
        $358 = ((($357)) + 12|0); //@line 106 "./common/Object.h"
        $359 = HEAP32[$358>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($359|0,($283|0)); //@line 106 "./common/Object.h"
        $360 = __THREW__; __THREW__ = 0;
        $361 = $360&1;
        if (!($361)) {
         $$4 = $355;$$46 = $356;
         break L1;
        }
        $362 = ___cxa_find_matching_catch_3(0|0)|0;
        $363 = tempRet0;
        ___clang_call_terminate($362); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       $306 = HEAP32[$283>>2]|0; //@line 106 "./common/Object.h"
       $307 = ((($306)) + 12|0); //@line 106 "./common/Object.h"
       $308 = HEAP32[$307>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($308|0,($283|0)); //@line 106 "./common/Object.h"
       $309 = __THREW__; __THREW__ = 0;
       $310 = $309&1;
       if ($310) {
        $311 = ___cxa_find_matching_catch_3(0|0)|0;
        $312 = tempRet0;
        ___clang_call_terminate($311); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       __THREW__ = 0;
       $313 = (invoke_ii(3348,32)|0); //@line 569 "modules/event/sdl/Event.cpp"
       $314 = __THREW__; __THREW__ = 0;
       $315 = $314&1;
       if ($315) {
        label = 4;
        break L1;
       }
       __THREW__ = 0;
       invoke_viii(3460,($10|0),(365819|0),6); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $316 = __THREW__; __THREW__ = 0;
       $317 = $316&1;
       do {
        if ($317) {
         $364 = ___cxa_find_matching_catch_2()|0;
         $365 = tempRet0;
         $$3 = $364;$$35 = $365;
        } else {
         __THREW__ = 0;
         invoke_viii(3508,($313|0),($10|0),($vargs|0)); //@line 569 "modules/event/sdl/Event.cpp"
         $318 = __THREW__; __THREW__ = 0;
         $319 = $318&1;
         if ($319) {
          $366 = ___cxa_find_matching_catch_2()|0;
          $367 = tempRet0;
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10); //@line 569 "modules/event/sdl/Event.cpp"
          $$3 = $366;$$35 = $367;
          break;
         } else {
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10); //@line 569 "modules/event/sdl/Event.cpp"
          $msg$0 = $313;
          break L5;
         }
        }
       } while(0);
       __ZdlPv($313); //@line 569 "modules/event/sdl/Event.cpp"
       $$4 = $$3;$$46 = $$35;
       break L1;
       break;
      }
      case 6:  {
       $368 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
       $369 = ($368|0)==(0|0); //@line 574 "modules/event/sdl/Event.cpp"
       if ($369) {
        $msg$0 = 0;
       } else {
        $370 = HEAP32[$368>>2]|0; //@line 575 "modules/event/sdl/Event.cpp"
        $371 = ((($370)) + 44|0); //@line 575 "modules/event/sdl/Event.cpp"
        $372 = HEAP32[$371>>2]|0; //@line 575 "modules/event/sdl/Event.cpp"
        $373 = ((($e)) + 16|0); //@line 575 "modules/event/sdl/Event.cpp"
        $374 = HEAP32[$373>>2]|0; //@line 575 "modules/event/sdl/Event.cpp"
        $375 = ((($e)) + 20|0); //@line 575 "modules/event/sdl/Event.cpp"
        $376 = HEAP32[$375>>2]|0; //@line 575 "modules/event/sdl/Event.cpp"
        __THREW__ = 0;
        (invoke_iiii($372|0,($368|0),($374|0),($376|0))|0); //@line 575 "modules/event/sdl/Event.cpp"
        $377 = __THREW__; __THREW__ = 0;
        $378 = $377&1;
        if ($378) {
         label = 4;
         break L1;
        } else {
         $msg$0 = 0;
        }
       }
       break;
      }
      default: {
       $msg$0 = 0;
      }
      }
     } while(0);
     $379 = HEAP32[$vargs>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $380 = HEAP32[$11>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $381 = ($379|0)==($380|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($381) {
      $$0 = $msg$0;$392 = $379;
     } else {
      $__begin$sroa$0$081 = $379;
      while(1) {
       $382 = HEAP32[$__begin$sroa$0$081>>2]|0; //@line 139 "./common/Object.h"
       $383 = ($382|0)==(0|0); //@line 597 "modules/event/sdl/Event.cpp"
       if (!($383)) {
        $384 = HEAP32[$382>>2]|0; //@line 598 "modules/event/sdl/Event.cpp"
        $385 = ((($384)) + 12|0); //@line 598 "modules/event/sdl/Event.cpp"
        $386 = HEAP32[$385>>2]|0; //@line 598 "modules/event/sdl/Event.cpp"
        __THREW__ = 0;
        invoke_vi($386|0,($382|0)); //@line 598 "modules/event/sdl/Event.cpp"
        $387 = __THREW__; __THREW__ = 0;
        $388 = $387&1;
        if ($388) {
         break;
        }
       }
       $389 = ((($__begin$sroa$0$081)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       $390 = ($389|0)==($380|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       if ($390) {
        $$0$ph = $msg$0;
        label = 128;
        break L3;
       } else {
        $__begin$sroa$0$081 = $389;
       }
      }
      $17 = ___cxa_find_matching_catch_2()|0;
      $18 = tempRet0;
      $lpad$phi$index = $17;$lpad$phi$index2 = $18;
      label = 5;
      break L1;
     }
    } else {
     $$0$ph = 0;
     label = 128;
    }
   } while(0);
   if ((label|0) == 128) {
    $$pr = HEAP32[$vargs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$0 = $$0$ph;$392 = $$pr;
   }
   $391 = ($392|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($391) {
    STACKTOP = sp;return ($$0|0); //@line 602 "modules/event/sdl/Event.cpp"
   }
   $393 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $394 = ($393|0)==($392|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   do {
    if ($394) {
     $408 = $392;
    } else {
     $396 = $393;
     while(1) {
      $395 = ((($396)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$11>>2] = $395; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $397 = HEAP32[$395>>2]|0; //@line 106 "./common/Object.h"
      $398 = ($397|0)==(0|0); //@line 106 "./common/Object.h"
      if ($398) {
       $400 = $395;
      } else {
       $401 = HEAP32[$397>>2]|0; //@line 106 "./common/Object.h"
       $402 = ((($401)) + 12|0); //@line 106 "./common/Object.h"
       $403 = HEAP32[$402>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($403|0,($397|0)); //@line 106 "./common/Object.h"
       $404 = __THREW__; __THREW__ = 0;
       $405 = $404&1;
       if ($405) {
        label = 135;
        break;
       }
       $$pre$i$i$i73 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $400 = $$pre$i$i$i73;
      }
      $399 = ($400|0)==($392|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($399) {
       label = 136;
       break;
      } else {
       $396 = $400;
      }
     }
     if ((label|0) == 135) {
      $406 = ___cxa_find_matching_catch_3(0|0)|0;
      $407 = tempRet0;
      ___clang_call_terminate($406); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     else if ((label|0) == 136) {
      $$pre$i76 = HEAP32[$vargs>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $408 = $$pre$i76;
      break;
     }
    }
   } while(0);
   __ZdlPv($408); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   STACKTOP = sp;return ($$0|0); //@line 602 "modules/event/sdl/Event.cpp"
  }
 } while(0);
 if ((label|0) == 4) {
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  $lpad$phi$index = $19;$lpad$phi$index2 = $20;
  label = 5;
 }
 if ((label|0) == 5) {
  $$4 = $lpad$phi$index;$$46 = $lpad$phi$index2;
 }
 $409 = HEAP32[$vargs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $410 = ($409|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($410) {
  ___resumeException($$4|0);
  // unreachable;
 }
 $411 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $412 = ($411|0)==($409|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($412) {
   $426 = $409;
  } else {
   $414 = $411;
   while(1) {
    $413 = ((($414)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$11>>2] = $413; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $415 = HEAP32[$413>>2]|0; //@line 106 "./common/Object.h"
    $416 = ($415|0)==(0|0); //@line 106 "./common/Object.h"
    if ($416) {
     $418 = $413;
    } else {
     $419 = HEAP32[$415>>2]|0; //@line 106 "./common/Object.h"
     $420 = ((($419)) + 12|0); //@line 106 "./common/Object.h"
     $421 = HEAP32[$420>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($421|0,($415|0)); //@line 106 "./common/Object.h"
     $422 = __THREW__; __THREW__ = 0;
     $423 = $422&1;
     if ($423) {
      label = 145;
      break;
     }
     $$pre$i$i$i = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $418 = $$pre$i$i$i;
    }
    $417 = ($418|0)==($409|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($417) {
     label = 146;
     break;
    } else {
     $414 = $418;
    }
   }
   if ((label|0) == 145) {
    $424 = ___cxa_find_matching_catch_3(0|0)|0;
    $425 = tempRet0;
    ___clang_call_terminate($424); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 146) {
    $$pre$i = HEAP32[$vargs>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $426 = $$pre$i;
    break;
   }
  }
 } while(0);
 __ZdlPv($426); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$4|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love5event3sdl5Event4waitEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $e = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $e = sp;
 $0 = (_SDL_WaitEvent($e)|0); //@line 131 "modules/event/sdl/Event.cpp"
 $1 = ($0|0)==(1); //@line 131 "modules/event/sdl/Event.cpp"
 if ($1) {
  $2 = (__ZNK4love5event3sdl5Event7convertERK9SDL_Event($this,$e)|0); //@line 134 "modules/event/sdl/Event.cpp"
  $$0 = $2;
 } else {
  $$0 = 0;
 }
 STACKTOP = sp;return ($$0|0); //@line 135 "modules/event/sdl/Event.cpp"
}
function __GLOBAL__sub_I_Event_cpp() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5event3sdl5Event12createKeyMapEv(632700); //@line 822 "modules/event/sdl/Event.cpp"
 (___cxa_atexit((3958|0),(632700|0),(___dso_handle|0))|0); //@line 822 "modules/event/sdl/Event.cpp"
 return;
}
function __ZN4love5event3sdl5Event12createKeyMapEv($agg$result) {
 $agg$result = $agg$result|0;
 var $$0$i7$i1014 = 0, $$0$i7$i1029 = 0, $$0$i7$i1044 = 0, $$0$i7$i1059 = 0, $$0$i7$i1074 = 0, $$0$i7$i1089 = 0, $$0$i7$i1104 = 0, $$0$i7$i1119 = 0, $$0$i7$i1134 = 0, $$0$i7$i114 = 0, $$0$i7$i1149 = 0, $$0$i7$i1164 = 0, $$0$i7$i1179 = 0, $$0$i7$i1194 = 0, $$0$i7$i1209 = 0, $$0$i7$i1224 = 0, $$0$i7$i1239 = 0, $$0$i7$i1254 = 0, $$0$i7$i1269 = 0, $$0$i7$i1284 = 0;
 var $$0$i7$i129 = 0, $$0$i7$i1299 = 0, $$0$i7$i1314 = 0, $$0$i7$i1329 = 0, $$0$i7$i1344 = 0, $$0$i7$i1359 = 0, $$0$i7$i1374 = 0, $$0$i7$i1389 = 0, $$0$i7$i1404 = 0, $$0$i7$i1419 = 0, $$0$i7$i1434 = 0, $$0$i7$i144 = 0, $$0$i7$i1449 = 0, $$0$i7$i1464 = 0, $$0$i7$i1479 = 0, $$0$i7$i1494 = 0, $$0$i7$i1509 = 0, $$0$i7$i1524 = 0, $$0$i7$i1539 = 0, $$0$i7$i1554 = 0;
 var $$0$i7$i1569 = 0, $$0$i7$i1584 = 0, $$0$i7$i159 = 0, $$0$i7$i1599 = 0, $$0$i7$i1614 = 0, $$0$i7$i1629 = 0, $$0$i7$i1644 = 0, $$0$i7$i1659 = 0, $$0$i7$i1674 = 0, $$0$i7$i1689 = 0, $$0$i7$i1704 = 0, $$0$i7$i1719 = 0, $$0$i7$i1734 = 0, $$0$i7$i174 = 0, $$0$i7$i1749 = 0, $$0$i7$i1764 = 0, $$0$i7$i1779 = 0, $$0$i7$i1794 = 0, $$0$i7$i1809 = 0, $$0$i7$i1824 = 0;
 var $$0$i7$i1839 = 0, $$0$i7$i1854 = 0, $$0$i7$i1869 = 0, $$0$i7$i1884 = 0, $$0$i7$i189 = 0, $$0$i7$i1899 = 0, $$0$i7$i1914 = 0, $$0$i7$i1929 = 0, $$0$i7$i1944 = 0, $$0$i7$i1959 = 0, $$0$i7$i1974 = 0, $$0$i7$i1989 = 0, $$0$i7$i2004 = 0, $$0$i7$i2019 = 0, $$0$i7$i2034 = 0, $$0$i7$i204 = 0, $$0$i7$i2049 = 0, $$0$i7$i2064 = 0, $$0$i7$i2079 = 0, $$0$i7$i2094 = 0;
 var $$0$i7$i2109 = 0, $$0$i7$i2124 = 0, $$0$i7$i2139 = 0, $$0$i7$i2154 = 0, $$0$i7$i2169 = 0, $$0$i7$i2184 = 0, $$0$i7$i219 = 0, $$0$i7$i2199 = 0, $$0$i7$i2214 = 0, $$0$i7$i2229 = 0, $$0$i7$i2244 = 0, $$0$i7$i2259 = 0, $$0$i7$i2274 = 0, $$0$i7$i2289 = 0, $$0$i7$i2304 = 0, $$0$i7$i2319 = 0, $$0$i7$i2334 = 0, $$0$i7$i234 = 0, $$0$i7$i2349 = 0, $$0$i7$i2364 = 0;
 var $$0$i7$i2379 = 0, $$0$i7$i2394 = 0, $$0$i7$i24 = 0, $$0$i7$i2409 = 0, $$0$i7$i2424 = 0, $$0$i7$i2439 = 0, $$0$i7$i2454 = 0, $$0$i7$i2469 = 0, $$0$i7$i2484 = 0, $$0$i7$i249 = 0, $$0$i7$i2499 = 0, $$0$i7$i2514 = 0, $$0$i7$i2529 = 0, $$0$i7$i2544 = 0, $$0$i7$i2559 = 0, $$0$i7$i2574 = 0, $$0$i7$i2589 = 0, $$0$i7$i2604 = 0, $$0$i7$i2619 = 0, $$0$i7$i2634 = 0;
 var $$0$i7$i264 = 0, $$0$i7$i2649 = 0, $$0$i7$i2664 = 0, $$0$i7$i2679 = 0, $$0$i7$i2694 = 0, $$0$i7$i2709 = 0, $$0$i7$i2724 = 0, $$0$i7$i2739 = 0, $$0$i7$i2754 = 0, $$0$i7$i2769 = 0, $$0$i7$i2784 = 0, $$0$i7$i279 = 0, $$0$i7$i2799 = 0, $$0$i7$i2814 = 0, $$0$i7$i2829 = 0, $$0$i7$i2844 = 0, $$0$i7$i294 = 0, $$0$i7$i309 = 0, $$0$i7$i324 = 0, $$0$i7$i339 = 0;
 var $$0$i7$i354 = 0, $$0$i7$i369 = 0, $$0$i7$i384 = 0, $$0$i7$i39 = 0, $$0$i7$i399 = 0, $$0$i7$i414 = 0, $$0$i7$i429 = 0, $$0$i7$i444 = 0, $$0$i7$i459 = 0, $$0$i7$i474 = 0, $$0$i7$i489 = 0, $$0$i7$i504 = 0, $$0$i7$i519 = 0, $$0$i7$i534 = 0, $$0$i7$i54 = 0, $$0$i7$i549 = 0, $$0$i7$i564 = 0, $$0$i7$i579 = 0, $$0$i7$i594 = 0, $$0$i7$i609 = 0;
 var $$0$i7$i624 = 0, $$0$i7$i639 = 0, $$0$i7$i654 = 0, $$0$i7$i669 = 0, $$0$i7$i684 = 0, $$0$i7$i69 = 0, $$0$i7$i699 = 0, $$0$i7$i714 = 0, $$0$i7$i729 = 0, $$0$i7$i744 = 0, $$0$i7$i759 = 0, $$0$i7$i774 = 0, $$0$i7$i789 = 0, $$0$i7$i804 = 0, $$0$i7$i819 = 0, $$0$i7$i834 = 0, $$0$i7$i84 = 0, $$0$i7$i849 = 0, $$0$i7$i864 = 0, $$0$i7$i879 = 0;
 var $$0$i7$i894 = 0, $$0$i7$i9 = 0, $$0$i7$i909 = 0, $$0$i7$i924 = 0, $$0$i7$i939 = 0, $$0$i7$i954 = 0, $$0$i7$i969 = 0, $$0$i7$i984 = 0, $$0$i7$i99 = 0, $$0$i7$i999 = 0, $$lcssa = 0, $$lcssa6844 = 0, $$lcssa6847 = 0, $$lcssa6848 = 0, $$lcssa6851 = 0, $$lcssa6852 = 0, $$lcssa6855 = 0, $$lcssa6856 = 0, $$lcssa6859 = 0, $$lcssa6860 = 0;
 var $$lcssa6863 = 0, $$lcssa6864 = 0, $$lcssa6867 = 0, $$lcssa6868 = 0, $$lcssa6871 = 0, $$lcssa6872 = 0, $$lcssa6875 = 0, $$lcssa6876 = 0, $$lcssa6879 = 0, $$lcssa6880 = 0, $$lcssa6883 = 0, $$lcssa6884 = 0, $$lcssa6887 = 0, $$lcssa6888 = 0, $$lcssa6891 = 0, $$lcssa6892 = 0, $$lcssa6895 = 0, $$lcssa6896 = 0, $$lcssa6899 = 0, $$lcssa6900 = 0;
 var $$lcssa6903 = 0, $$lcssa6904 = 0, $$lcssa6907 = 0, $$lcssa6908 = 0, $$lcssa6911 = 0, $$lcssa6912 = 0, $$lcssa6915 = 0, $$lcssa6916 = 0, $$lcssa6919 = 0, $$lcssa6920 = 0, $$lcssa6923 = 0, $$lcssa6924 = 0, $$lcssa6927 = 0, $$lcssa6928 = 0, $$lcssa6931 = 0, $$lcssa6932 = 0, $$lcssa6935 = 0, $$lcssa6936 = 0, $$lcssa6939 = 0, $$lcssa6940 = 0;
 var $$lcssa6943 = 0, $$lcssa6944 = 0, $$lcssa6947 = 0, $$lcssa6948 = 0, $$lcssa6951 = 0, $$lcssa6952 = 0, $$lcssa6955 = 0, $$lcssa6956 = 0, $$lcssa6959 = 0, $$lcssa6960 = 0, $$lcssa6963 = 0, $$lcssa6964 = 0, $$lcssa6967 = 0, $$lcssa6968 = 0, $$lcssa6971 = 0, $$lcssa6972 = 0, $$lcssa6975 = 0, $$lcssa6976 = 0, $$lcssa6979 = 0, $$lcssa6980 = 0;
 var $$lcssa6983 = 0, $$lcssa6984 = 0, $$lcssa6987 = 0, $$lcssa6988 = 0, $$lcssa6991 = 0, $$lcssa6992 = 0, $$lcssa6995 = 0, $$lcssa6996 = 0, $$lcssa6999 = 0, $$lcssa7000 = 0, $$lcssa7003 = 0, $$lcssa7004 = 0, $$lcssa7007 = 0, $$lcssa7008 = 0, $$lcssa7011 = 0, $$lcssa7012 = 0, $$lcssa7015 = 0, $$lcssa7016 = 0, $$lcssa7019 = 0, $$lcssa7020 = 0;
 var $$lcssa7023 = 0, $$lcssa7024 = 0, $$lcssa7027 = 0, $$lcssa7028 = 0, $$lcssa7031 = 0, $$lcssa7032 = 0, $$lcssa7035 = 0, $$lcssa7036 = 0, $$lcssa7039 = 0, $$lcssa7040 = 0, $$lcssa7043 = 0, $$lcssa7044 = 0, $$lcssa7047 = 0, $$lcssa7048 = 0, $$lcssa7051 = 0, $$lcssa7052 = 0, $$lcssa7055 = 0, $$lcssa7056 = 0, $$lcssa7059 = 0, $$lcssa7060 = 0;
 var $$lcssa7063 = 0, $$lcssa7064 = 0, $$lcssa7067 = 0, $$lcssa7068 = 0, $$lcssa7071 = 0, $$lcssa7072 = 0, $$lcssa7075 = 0, $$lcssa7076 = 0, $$lcssa7079 = 0, $$lcssa7080 = 0, $$lcssa7083 = 0, $$lcssa7084 = 0, $$lcssa7087 = 0, $$lcssa7088 = 0, $$lcssa7091 = 0, $$lcssa7092 = 0, $$lcssa7095 = 0, $$lcssa7096 = 0, $$lcssa7099 = 0, $$lcssa7100 = 0;
 var $$lcssa7103 = 0, $$lcssa7104 = 0, $$lcssa7107 = 0, $$lcssa7108 = 0, $$lcssa7111 = 0, $$lcssa7112 = 0, $$lcssa7115 = 0, $$lcssa7116 = 0, $$lcssa7119 = 0, $$lcssa7120 = 0, $$lcssa7123 = 0, $$lcssa7124 = 0, $$lcssa7127 = 0, $$lcssa7128 = 0, $$lcssa7131 = 0, $$lcssa7132 = 0, $$lcssa7135 = 0, $$lcssa7136 = 0, $$lcssa7139 = 0, $$lcssa7140 = 0;
 var $$lcssa7143 = 0, $$lcssa7144 = 0, $$lcssa7147 = 0, $$lcssa7148 = 0, $$lcssa7151 = 0, $$lcssa7152 = 0, $$lcssa7155 = 0, $$lcssa7156 = 0, $$lcssa7159 = 0, $$lcssa7160 = 0, $$lcssa7163 = 0, $$lcssa7164 = 0, $$lcssa7167 = 0, $$lcssa7168 = 0, $$lcssa7171 = 0, $$lcssa7172 = 0, $$lcssa7175 = 0, $$lcssa7176 = 0, $$lcssa7179 = 0, $$lcssa7180 = 0;
 var $$lcssa7183 = 0, $$lcssa7184 = 0, $$lcssa7187 = 0, $$lcssa7188 = 0, $$lcssa7191 = 0, $$lcssa7192 = 0, $$lcssa7195 = 0, $$lcssa7196 = 0, $$lcssa7199 = 0, $$lcssa7200 = 0, $$lcssa7203 = 0, $$lcssa7204 = 0, $$lcssa7207 = 0, $$lcssa7208 = 0, $$lcssa7211 = 0, $$lcssa7212 = 0, $$lcssa7215 = 0, $$lcssa7216 = 0, $$lcssa7219 = 0, $$lcssa7220 = 0;
 var $$lcssa7223 = 0, $$lcssa7224 = 0, $$lcssa7227 = 0, $$lcssa7228 = 0, $$lcssa7231 = 0, $$lcssa7232 = 0, $$lcssa7235 = 0, $$lcssa7236 = 0, $$lcssa7239 = 0, $$lcssa7240 = 0, $$lcssa7243 = 0, $$lcssa7244 = 0, $$lcssa7247 = 0, $$lcssa7248 = 0, $$lcssa7251 = 0, $$lcssa7252 = 0, $$lcssa7255 = 0, $$lcssa7256 = 0, $$lcssa7259 = 0, $$lcssa7260 = 0;
 var $$lcssa7263 = 0, $$lcssa7264 = 0, $$lcssa7267 = 0, $$lcssa7268 = 0, $$lcssa7271 = 0, $$lcssa7272 = 0, $$lcssa7275 = 0, $$lcssa7276 = 0, $$lcssa7279 = 0, $$lcssa7280 = 0, $$lcssa7283 = 0, $$lcssa7284 = 0, $$lcssa7287 = 0, $$lcssa7288 = 0, $$lcssa7291 = 0, $$lcssa7292 = 0, $$lcssa7295 = 0, $$lcssa7296 = 0, $$lcssa7299 = 0, $$lcssa7300 = 0;
 var $$lcssa7303 = 0, $$lcssa7304 = 0, $$lcssa7307 = 0, $$lcssa7308 = 0, $$lcssa7311 = 0, $$lcssa7312 = 0, $$lcssa7315 = 0, $$lcssa7316 = 0, $$lcssa7319 = 0, $$lcssa7320 = 0, $$lcssa7323 = 0, $$lcssa7324 = 0, $$lcssa7327 = 0, $$lcssa7328 = 0, $$lcssa7331 = 0, $$lcssa7332 = 0, $$lcssa7335 = 0, $$lcssa7336 = 0, $$lcssa7339 = 0, $$lcssa7340 = 0;
 var $$lcssa7343 = 0, $$lcssa7344 = 0, $$lcssa7347 = 0, $$lcssa7348 = 0, $$lcssa7351 = 0, $$lcssa7352 = 0, $$lcssa7355 = 0, $$lcssa7356 = 0, $$lcssa7359 = 0, $$lcssa7360 = 0, $$lcssa7363 = 0, $$lcssa7364 = 0, $$lcssa7367 = 0, $$lcssa7368 = 0, $$lcssa7371 = 0, $$lcssa7372 = 0, $$lcssa7375 = 0, $$lcssa7376 = 0, $$lcssa7379 = 0, $$lcssa7380 = 0;
 var $$lcssa7383 = 0, $$lcssa7384 = 0, $$lcssa7387 = 0, $$lcssa7388 = 0, $$lcssa7391 = 0, $$lcssa7392 = 0, $$lcssa7395 = 0, $$lcssa7396 = 0, $$lcssa7399 = 0, $$lcssa7400 = 0, $$lcssa7403 = 0, $$lcssa7404 = 0, $$lcssa7407 = 0, $$lcssa7408 = 0, $$lcssa7411 = 0, $$lcssa7412 = 0, $$lcssa7415 = 0, $$lcssa7416 = 0, $$lcssa7419 = 0, $$lcssa7420 = 0;
 var $$lcssa7423 = 0, $$lcssa7424 = 0, $$lcssa7427 = 0, $$lcssa7428 = 0, $$lcssa7431 = 0, $$lcssa7432 = 0, $$lcssa7435 = 0, $$lcssa7436 = 0, $$lcssa7439 = 0, $$lcssa7440 = 0, $$lcssa7443 = 0, $$lcssa7444 = 0, $$lcssa7447 = 0, $$lcssa7448 = 0, $$lcssa7451 = 0, $$lcssa7452 = 0, $$lcssa7455 = 0, $$lcssa7456 = 0, $$lcssa7459 = 0, $$lcssa7460 = 0;
 var $$lcssa7463 = 0, $$lcssa7464 = 0, $$lcssa7467 = 0, $$lcssa7468 = 0, $$lcssa7471 = 0, $$lcssa7472 = 0, $$lcssa7475 = 0, $$lcssa7476 = 0, $$lcssa7479 = 0, $$lcssa7480 = 0, $$lcssa7483 = 0, $$lcssa7484 = 0, $$lcssa7487 = 0, $$lcssa7488 = 0, $$lcssa7491 = 0, $$lcssa7492 = 0, $$lcssa7495 = 0, $$lcssa7496 = 0, $$lcssa7499 = 0, $$lcssa7500 = 0;
 var $$lcssa7503 = 0, $$lcssa7504 = 0, $$lcssa7507 = 0, $$lcssa7508 = 0, $$lcssa7511 = 0, $$lcssa7512 = 0, $$lcssa7515 = 0, $$lcssa7516 = 0, $$lcssa7519 = 0, $$lcssa7520 = 0, $$lcssa7523 = 0, $$lcssa7524 = 0, $$lcssa7527 = 0, $$lcssa7528 = 0, $$lcssa7531 = 0, $$lcssa7532 = 0, $$lcssa7535 = 0, $$lcssa7536 = 0, $$lcssa7539 = 0, $$lcssa7540 = 0;
 var $$lcssa7543 = 0, $$lcssa7544 = 0, $$lcssa7547 = 0, $$lcssa7548 = 0, $$lcssa7551 = 0, $$lcssa7552 = 0, $$lcssa7555 = 0, $$lcssa7556 = 0, $$lcssa7559 = 0, $$lcssa7560 = 0, $$lcssa7563 = 0, $$lcssa7564 = 0, $$lcssa7567 = 0, $$lcssa7568 = 0, $$lcssa7571 = 0, $$lcssa7572 = 0, $$lcssa7575 = 0, $$lcssa7576 = 0, $$lcssa7579 = 0, $$lcssa7580 = 0;
 var $$lcssa7583 = 0, $$lcssa7584 = 0, $$lcssa7587 = 0, $$lcssa7588 = 0, $$lcssa7591 = 0, $$lcssa7592 = 0, $$lcssa7595 = 0, $$lcssa7596 = 0, $$lcssa7599 = 0, $$lcssa7600 = 0, $$pre = 0, $$pre$i$i1001 = 0, $$pre$i$i101 = 0, $$pre$i$i1016 = 0, $$pre$i$i1031 = 0, $$pre$i$i1046 = 0, $$pre$i$i1061 = 0, $$pre$i$i1076 = 0, $$pre$i$i1091 = 0, $$pre$i$i11 = 0;
 var $$pre$i$i1106 = 0, $$pre$i$i1121 = 0, $$pre$i$i1136 = 0, $$pre$i$i1151 = 0, $$pre$i$i116 = 0, $$pre$i$i1166 = 0, $$pre$i$i1181 = 0, $$pre$i$i1196 = 0, $$pre$i$i1211 = 0, $$pre$i$i1226 = 0, $$pre$i$i1241 = 0, $$pre$i$i1256 = 0, $$pre$i$i1271 = 0, $$pre$i$i1286 = 0, $$pre$i$i1301 = 0, $$pre$i$i131 = 0, $$pre$i$i1316 = 0, $$pre$i$i1331 = 0, $$pre$i$i1346 = 0, $$pre$i$i1361 = 0;
 var $$pre$i$i1376 = 0, $$pre$i$i1391 = 0, $$pre$i$i1406 = 0, $$pre$i$i1421 = 0, $$pre$i$i1436 = 0, $$pre$i$i1451 = 0, $$pre$i$i146 = 0, $$pre$i$i1466 = 0, $$pre$i$i1481 = 0, $$pre$i$i1496 = 0, $$pre$i$i1511 = 0, $$pre$i$i1526 = 0, $$pre$i$i1541 = 0, $$pre$i$i1556 = 0, $$pre$i$i1571 = 0, $$pre$i$i1586 = 0, $$pre$i$i1601 = 0, $$pre$i$i161 = 0, $$pre$i$i1616 = 0, $$pre$i$i1631 = 0;
 var $$pre$i$i1646 = 0, $$pre$i$i1661 = 0, $$pre$i$i1676 = 0, $$pre$i$i1691 = 0, $$pre$i$i1706 = 0, $$pre$i$i1721 = 0, $$pre$i$i1736 = 0, $$pre$i$i1751 = 0, $$pre$i$i176 = 0, $$pre$i$i1766 = 0, $$pre$i$i1781 = 0, $$pre$i$i1796 = 0, $$pre$i$i1811 = 0, $$pre$i$i1826 = 0, $$pre$i$i1841 = 0, $$pre$i$i1856 = 0, $$pre$i$i1871 = 0, $$pre$i$i1886 = 0, $$pre$i$i1901 = 0, $$pre$i$i191 = 0;
 var $$pre$i$i1916 = 0, $$pre$i$i1931 = 0, $$pre$i$i1946 = 0, $$pre$i$i1961 = 0, $$pre$i$i1976 = 0, $$pre$i$i1991 = 0, $$pre$i$i2006 = 0, $$pre$i$i2021 = 0, $$pre$i$i2036 = 0, $$pre$i$i2051 = 0, $$pre$i$i206 = 0, $$pre$i$i2066 = 0, $$pre$i$i2081 = 0, $$pre$i$i2096 = 0, $$pre$i$i2111 = 0, $$pre$i$i2126 = 0, $$pre$i$i2141 = 0, $$pre$i$i2156 = 0, $$pre$i$i2171 = 0, $$pre$i$i2186 = 0;
 var $$pre$i$i2201 = 0, $$pre$i$i221 = 0, $$pre$i$i2216 = 0, $$pre$i$i2231 = 0, $$pre$i$i2246 = 0, $$pre$i$i2261 = 0, $$pre$i$i2276 = 0, $$pre$i$i2291 = 0, $$pre$i$i2306 = 0, $$pre$i$i2321 = 0, $$pre$i$i2336 = 0, $$pre$i$i2351 = 0, $$pre$i$i236 = 0, $$pre$i$i2366 = 0, $$pre$i$i2381 = 0, $$pre$i$i2396 = 0, $$pre$i$i2411 = 0, $$pre$i$i2426 = 0, $$pre$i$i2441 = 0, $$pre$i$i2456 = 0;
 var $$pre$i$i2471 = 0, $$pre$i$i2486 = 0, $$pre$i$i2501 = 0, $$pre$i$i251 = 0, $$pre$i$i2516 = 0, $$pre$i$i2531 = 0, $$pre$i$i2546 = 0, $$pre$i$i2561 = 0, $$pre$i$i2576 = 0, $$pre$i$i2591 = 0, $$pre$i$i26 = 0, $$pre$i$i2606 = 0, $$pre$i$i2621 = 0, $$pre$i$i2636 = 0, $$pre$i$i2651 = 0, $$pre$i$i266 = 0, $$pre$i$i2666 = 0, $$pre$i$i2681 = 0, $$pre$i$i2696 = 0, $$pre$i$i2711 = 0;
 var $$pre$i$i2726 = 0, $$pre$i$i2741 = 0, $$pre$i$i2756 = 0, $$pre$i$i2771 = 0, $$pre$i$i2786 = 0, $$pre$i$i2801 = 0, $$pre$i$i281 = 0, $$pre$i$i2816 = 0, $$pre$i$i2831 = 0, $$pre$i$i2846 = 0, $$pre$i$i296 = 0, $$pre$i$i311 = 0, $$pre$i$i326 = 0, $$pre$i$i341 = 0, $$pre$i$i356 = 0, $$pre$i$i371 = 0, $$pre$i$i386 = 0, $$pre$i$i401 = 0, $$pre$i$i41 = 0, $$pre$i$i416 = 0;
 var $$pre$i$i431 = 0, $$pre$i$i446 = 0, $$pre$i$i461 = 0, $$pre$i$i476 = 0, $$pre$i$i491 = 0, $$pre$i$i506 = 0, $$pre$i$i521 = 0, $$pre$i$i536 = 0, $$pre$i$i551 = 0, $$pre$i$i56 = 0, $$pre$i$i566 = 0, $$pre$i$i581 = 0, $$pre$i$i596 = 0, $$pre$i$i611 = 0, $$pre$i$i626 = 0, $$pre$i$i641 = 0, $$pre$i$i656 = 0, $$pre$i$i671 = 0, $$pre$i$i686 = 0, $$pre$i$i701 = 0;
 var $$pre$i$i71 = 0, $$pre$i$i716 = 0, $$pre$i$i731 = 0, $$pre$i$i746 = 0, $$pre$i$i761 = 0, $$pre$i$i776 = 0, $$pre$i$i791 = 0, $$pre$i$i806 = 0, $$pre$i$i821 = 0, $$pre$i$i836 = 0, $$pre$i$i851 = 0, $$pre$i$i86 = 0, $$pre$i$i866 = 0, $$pre$i$i881 = 0, $$pre$i$i896 = 0, $$pre$i$i911 = 0, $$pre$i$i926 = 0, $$pre$i$i941 = 0, $$pre$i$i956 = 0, $$pre$i$i971 = 0;
 var $$pre$i$i986 = 0, $$pre4377 = 0, $$pre4378 = 0, $$pre4379 = 0, $$pre4380 = 0, $$pre4381 = 0, $$pre4382 = 0, $$pre4383 = 0, $$pre4384 = 0, $$pre4385 = 0, $$pre4386 = 0, $$pre4387 = 0, $$pre4388 = 0, $$pre4389 = 0, $$pre4390 = 0, $$pre4391 = 0, $$pre4392 = 0, $$pre4393 = 0, $$pre4394 = 0, $$pre4395 = 0;
 var $$pre4396 = 0, $$pre4397 = 0, $$pre4398 = 0, $$pre4399 = 0, $$pre4400 = 0, $$pre4401 = 0, $$pre4402 = 0, $$pre4403 = 0, $$pre4404 = 0, $$pre4405 = 0, $$pre4406 = 0, $$pre4407 = 0, $$pre4408 = 0, $$pre4409 = 0, $$pre4410 = 0, $$pre4411 = 0, $$pre4412 = 0, $$pre4413 = 0, $$pre4414 = 0, $$pre4415 = 0;
 var $$pre4416 = 0, $$pre4417 = 0, $$pre4418 = 0, $$pre4419 = 0, $$pre4420 = 0, $$pre4421 = 0, $$pre4422 = 0, $$pre4423 = 0, $$pre4424 = 0, $$pre4425 = 0, $$pre4426 = 0, $$pre4427 = 0, $$pre4428 = 0, $$pre4429 = 0, $$pre4430 = 0, $$pre4431 = 0, $$pre4432 = 0, $$pre4433 = 0, $$pre4434 = 0, $$pre4435 = 0;
 var $$pre4436 = 0, $$pre4437 = 0, $$pre4438 = 0, $$pre4439 = 0, $$pre4440 = 0, $$pre4441 = 0, $$pre4442 = 0, $$pre4443 = 0, $$pre4444 = 0, $$pre4445 = 0, $$pre4446 = 0, $$pre4447 = 0, $$pre4448 = 0, $$pre4449 = 0, $$pre4450 = 0, $$pre4451 = 0, $$pre4452 = 0, $$pre4453 = 0, $$pre4454 = 0, $$pre4455 = 0;
 var $$pre4456 = 0, $$pre4457 = 0, $$pre4458 = 0, $$pre4459 = 0, $$pre4460 = 0, $$pre4461 = 0, $$pre4462 = 0, $$pre4463 = 0, $$pre4464 = 0, $$pre4465 = 0, $$pre4466 = 0, $$pre4467 = 0, $$pre4468 = 0, $$pre4469 = 0, $$pre4470 = 0, $$pre4471 = 0, $$pre4472 = 0, $$pre4473 = 0, $$pre4474 = 0, $$pre4475 = 0;
 var $$pre4476 = 0, $$pre4477 = 0, $$pre4478 = 0, $$pre4479 = 0, $$pre4480 = 0, $$pre4481 = 0, $$pre4482 = 0, $$pre4483 = 0, $$pre4484 = 0, $$pre4485 = 0, $$pre4486 = 0, $$pre4487 = 0, $$pre4488 = 0, $$pre4489 = 0, $$pre4490 = 0, $$pre4491 = 0, $$pre4492 = 0, $$pre4493 = 0, $$pre4494 = 0, $$pre4495 = 0;
 var $$pre4496 = 0, $$pre4497 = 0, $$pre4498 = 0, $$pre4499 = 0, $$pre4500 = 0, $$pre4501 = 0, $$pre4502 = 0, $$pre4503 = 0, $$pre4504 = 0, $$pre4505 = 0, $$pre4506 = 0, $$pre4507 = 0, $$pre4508 = 0, $$pre4509 = 0, $$pre4510 = 0, $$pre4511 = 0, $$pre4512 = 0, $$pre4513 = 0, $$pre4514 = 0, $$pre4515 = 0;
 var $$pre4516 = 0, $$pre4517 = 0, $$pre4518 = 0, $$pre4519 = 0, $$pre4520 = 0, $$pre4521 = 0, $$pre4522 = 0, $$pre4523 = 0, $$pre4524 = 0, $$pre4525 = 0, $$pre4526 = 0, $$pre4527 = 0, $$pre4528 = 0, $$pre4529 = 0, $$pre4530 = 0, $$pre4531 = 0, $$pre4532 = 0, $$pre4533 = 0, $$pre4534 = 0, $$pre4535 = 0;
 var $$pre4536 = 0, $$pre4537 = 0, $$pre4538 = 0, $$pre4539 = 0, $$pre4540 = 0, $$pre4541 = 0, $$pre4542 = 0, $$pre4543 = 0, $$pre4544 = 0, $$pre4545 = 0, $$pre4546 = 0, $$pre4547 = 0, $$pre4548 = 0, $$pre4549 = 0, $$pre4550 = 0, $$pre4551 = 0, $$pre4552 = 0, $$pre4553 = 0, $$pre4554 = 0, $$pre4555 = 0;
 var $$pre4556 = 0, $$pre4557 = 0, $$pre4558 = 0, $$pre4559 = 0, $$pre4560 = 0, $$pre4561 = 0, $$pre4562 = 0, $$pre4563 = 0, $$pre4564 = 0, $$sink$i$i1009 = 0, $$sink$i$i1024 = 0, $$sink$i$i1039 = 0, $$sink$i$i1054 = 0, $$sink$i$i1069 = 0, $$sink$i$i1084 = 0, $$sink$i$i109 = 0, $$sink$i$i1099 = 0, $$sink$i$i1114 = 0, $$sink$i$i1129 = 0, $$sink$i$i1144 = 0;
 var $$sink$i$i1159 = 0, $$sink$i$i1174 = 0, $$sink$i$i1189 = 0, $$sink$i$i1204 = 0, $$sink$i$i1219 = 0, $$sink$i$i1234 = 0, $$sink$i$i124 = 0, $$sink$i$i1249 = 0, $$sink$i$i1264 = 0, $$sink$i$i1279 = 0, $$sink$i$i1294 = 0, $$sink$i$i1309 = 0, $$sink$i$i1324 = 0, $$sink$i$i1339 = 0, $$sink$i$i1354 = 0, $$sink$i$i1369 = 0, $$sink$i$i1384 = 0, $$sink$i$i139 = 0, $$sink$i$i1399 = 0, $$sink$i$i1414 = 0;
 var $$sink$i$i1429 = 0, $$sink$i$i1444 = 0, $$sink$i$i1459 = 0, $$sink$i$i1474 = 0, $$sink$i$i1489 = 0, $$sink$i$i1504 = 0, $$sink$i$i1519 = 0, $$sink$i$i1534 = 0, $$sink$i$i154 = 0, $$sink$i$i1549 = 0, $$sink$i$i1564 = 0, $$sink$i$i1579 = 0, $$sink$i$i1594 = 0, $$sink$i$i1609 = 0, $$sink$i$i1624 = 0, $$sink$i$i1639 = 0, $$sink$i$i1654 = 0, $$sink$i$i1669 = 0, $$sink$i$i1684 = 0, $$sink$i$i169 = 0;
 var $$sink$i$i1699 = 0, $$sink$i$i1714 = 0, $$sink$i$i1729 = 0, $$sink$i$i1744 = 0, $$sink$i$i1759 = 0, $$sink$i$i1774 = 0, $$sink$i$i1789 = 0, $$sink$i$i1804 = 0, $$sink$i$i1819 = 0, $$sink$i$i1834 = 0, $$sink$i$i184 = 0, $$sink$i$i1849 = 0, $$sink$i$i1864 = 0, $$sink$i$i1879 = 0, $$sink$i$i1894 = 0, $$sink$i$i19 = 0, $$sink$i$i1909 = 0, $$sink$i$i1924 = 0, $$sink$i$i1939 = 0, $$sink$i$i1954 = 0;
 var $$sink$i$i1969 = 0, $$sink$i$i1984 = 0, $$sink$i$i199 = 0, $$sink$i$i1999 = 0, $$sink$i$i2014 = 0, $$sink$i$i2029 = 0, $$sink$i$i2044 = 0, $$sink$i$i2059 = 0, $$sink$i$i2074 = 0, $$sink$i$i2089 = 0, $$sink$i$i2104 = 0, $$sink$i$i2119 = 0, $$sink$i$i2134 = 0, $$sink$i$i214 = 0, $$sink$i$i2149 = 0, $$sink$i$i2164 = 0, $$sink$i$i2179 = 0, $$sink$i$i2194 = 0, $$sink$i$i2209 = 0, $$sink$i$i2224 = 0;
 var $$sink$i$i2239 = 0, $$sink$i$i2254 = 0, $$sink$i$i2269 = 0, $$sink$i$i2284 = 0, $$sink$i$i229 = 0, $$sink$i$i2299 = 0, $$sink$i$i2314 = 0, $$sink$i$i2329 = 0, $$sink$i$i2344 = 0, $$sink$i$i2359 = 0, $$sink$i$i2374 = 0, $$sink$i$i2389 = 0, $$sink$i$i2404 = 0, $$sink$i$i2419 = 0, $$sink$i$i2434 = 0, $$sink$i$i244 = 0, $$sink$i$i2449 = 0, $$sink$i$i2464 = 0, $$sink$i$i2479 = 0, $$sink$i$i2494 = 0;
 var $$sink$i$i2509 = 0, $$sink$i$i2524 = 0, $$sink$i$i2539 = 0, $$sink$i$i2554 = 0, $$sink$i$i2569 = 0, $$sink$i$i2584 = 0, $$sink$i$i259 = 0, $$sink$i$i2599 = 0, $$sink$i$i2614 = 0, $$sink$i$i2629 = 0, $$sink$i$i2644 = 0, $$sink$i$i2659 = 0, $$sink$i$i2674 = 0, $$sink$i$i2689 = 0, $$sink$i$i2704 = 0, $$sink$i$i2719 = 0, $$sink$i$i2734 = 0, $$sink$i$i274 = 0, $$sink$i$i2749 = 0, $$sink$i$i2764 = 0;
 var $$sink$i$i2779 = 0, $$sink$i$i2794 = 0, $$sink$i$i2809 = 0, $$sink$i$i2824 = 0, $$sink$i$i2839 = 0, $$sink$i$i289 = 0, $$sink$i$i304 = 0, $$sink$i$i319 = 0, $$sink$i$i334 = 0, $$sink$i$i34 = 0, $$sink$i$i349 = 0, $$sink$i$i364 = 0, $$sink$i$i379 = 0, $$sink$i$i394 = 0, $$sink$i$i4 = 0, $$sink$i$i409 = 0, $$sink$i$i424 = 0, $$sink$i$i439 = 0, $$sink$i$i454 = 0, $$sink$i$i469 = 0;
 var $$sink$i$i484 = 0, $$sink$i$i49 = 0, $$sink$i$i499 = 0, $$sink$i$i514 = 0, $$sink$i$i529 = 0, $$sink$i$i544 = 0, $$sink$i$i559 = 0, $$sink$i$i574 = 0, $$sink$i$i589 = 0, $$sink$i$i604 = 0, $$sink$i$i619 = 0, $$sink$i$i634 = 0, $$sink$i$i64 = 0, $$sink$i$i649 = 0, $$sink$i$i664 = 0, $$sink$i$i679 = 0, $$sink$i$i694 = 0, $$sink$i$i709 = 0, $$sink$i$i724 = 0, $$sink$i$i739 = 0;
 var $$sink$i$i754 = 0, $$sink$i$i769 = 0, $$sink$i$i784 = 0, $$sink$i$i79 = 0, $$sink$i$i799 = 0, $$sink$i$i814 = 0, $$sink$i$i829 = 0, $$sink$i$i844 = 0, $$sink$i$i859 = 0, $$sink$i$i874 = 0, $$sink$i$i889 = 0, $$sink$i$i904 = 0, $$sink$i$i919 = 0, $$sink$i$i934 = 0, $$sink$i$i94 = 0, $$sink$i$i949 = 0, $$sink$i$i964 = 0, $$sink$i$i979 = 0, $$sink$i$i994 = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0;
 var $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0;
 var $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0;
 var $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0;
 var $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0, $1088 = 0;
 var $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0;
 var $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0;
 var $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0;
 var $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0, $116 = 0;
 var $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0, $1178 = 0;
 var $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0, $1196 = 0;
 var $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0, $1213 = 0;
 var $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0, $1231 = 0;
 var $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0, $125 = 0;
 var $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0, $1267 = 0, $1268 = 0;
 var $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0, $1285 = 0, $1286 = 0;
 var $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0, $1302 = 0, $1303 = 0;
 var $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0, $1320 = 0, $1321 = 0;
 var $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0, $1339 = 0, $134 = 0;
 var $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0, $1357 = 0, $1358 = 0;
 var $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0, $1375 = 0, $1376 = 0;
 var $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0, $1393 = 0, $1394 = 0;
 var $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0, $1410 = 0, $1411 = 0;
 var $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0, $1429 = 0, $143 = 0;
 var $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0, $1447 = 0, $1448 = 0;
 var $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0, $1465 = 0, $1466 = 0;
 var $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0, $1483 = 0, $1484 = 0;
 var $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0, $1500 = 0, $1501 = 0;
 var $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0, $1511 = 0, $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0, $1518 = 0, $1519 = 0, $152 = 0;
 var $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0, $153 = 0, $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0, $1536 = 0, $1537 = 0, $1538 = 0;
 var $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0, $1548 = 0, $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0, $1554 = 0, $1555 = 0, $1556 = 0;
 var $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $1565 = 0, $1566 = 0, $1567 = 0, $1568 = 0, $1569 = 0, $157 = 0, $1570 = 0, $1571 = 0, $1572 = 0, $1573 = 0, $1574 = 0;
 var $1575 = 0, $1576 = 0, $1577 = 0, $1578 = 0, $1579 = 0, $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0, $1583 = 0, $1584 = 0, $1585 = 0, $1586 = 0, $1587 = 0, $1588 = 0, $1589 = 0, $159 = 0, $1590 = 0, $1591 = 0, $1592 = 0;
 var $1593 = 0, $1594 = 0, $1595 = 0, $1596 = 0, $1597 = 0, $1598 = 0, $1599 = 0, $16 = 0, $160 = 0, $1600 = 0, $1601 = 0, $1602 = 0, $1603 = 0, $1604 = 0, $1605 = 0, $1606 = 0, $1607 = 0, $1608 = 0, $1609 = 0, $161 = 0;
 var $1610 = 0, $1611 = 0, $1612 = 0, $1613 = 0, $1614 = 0, $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0, $1619 = 0, $162 = 0, $1620 = 0, $1621 = 0, $1622 = 0, $1623 = 0, $1624 = 0, $1625 = 0, $1626 = 0, $1627 = 0, $1628 = 0;
 var $1629 = 0, $163 = 0, $1630 = 0, $1631 = 0, $1632 = 0, $1633 = 0, $1634 = 0, $1635 = 0, $1636 = 0, $1637 = 0, $1638 = 0, $1639 = 0, $164 = 0, $1640 = 0, $1641 = 0, $1642 = 0, $1643 = 0, $1644 = 0, $1645 = 0, $1646 = 0;
 var $1647 = 0, $1648 = 0, $1649 = 0, $165 = 0, $1650 = 0, $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0, $1655 = 0, $1656 = 0, $1657 = 0, $1658 = 0, $1659 = 0, $166 = 0, $1660 = 0, $1661 = 0, $1662 = 0, $1663 = 0, $1664 = 0;
 var $1665 = 0, $1666 = 0, $1667 = 0, $1668 = 0, $1669 = 0, $167 = 0, $1670 = 0, $1671 = 0, $1672 = 0, $1673 = 0, $1674 = 0, $1675 = 0, $1676 = 0, $1677 = 0, $1678 = 0, $1679 = 0, $168 = 0, $1680 = 0, $1681 = 0, $1682 = 0;
 var $1683 = 0, $1684 = 0, $1685 = 0, $1686 = 0, $1687 = 0, $1688 = 0, $1689 = 0, $169 = 0, $1690 = 0, $1691 = 0, $1692 = 0, $1693 = 0, $1694 = 0, $1695 = 0, $1696 = 0, $1697 = 0, $1698 = 0, $1699 = 0, $17 = 0, $170 = 0;
 var $1700 = 0, $1701 = 0, $1702 = 0, $1703 = 0, $1704 = 0, $1705 = 0, $1706 = 0, $1707 = 0, $1708 = 0, $1709 = 0, $171 = 0, $1710 = 0, $1711 = 0, $1712 = 0, $1713 = 0, $1714 = 0, $1715 = 0, $1716 = 0, $1717 = 0, $1718 = 0;
 var $1719 = 0, $172 = 0, $1720 = 0, $1721 = 0, $1722 = 0, $1723 = 0, $1724 = 0, $1725 = 0, $1726 = 0, $1727 = 0, $1728 = 0, $1729 = 0, $173 = 0, $1730 = 0, $1731 = 0, $1732 = 0, $1733 = 0, $1734 = 0, $1735 = 0, $1736 = 0;
 var $1737 = 0, $1738 = 0, $1739 = 0, $174 = 0, $1740 = 0, $1741 = 0, $1742 = 0, $1743 = 0, $1744 = 0, $1745 = 0, $1746 = 0, $1747 = 0, $1748 = 0, $1749 = 0, $175 = 0, $1750 = 0, $1751 = 0, $1752 = 0, $1753 = 0, $1754 = 0;
 var $1755 = 0, $1756 = 0, $1757 = 0, $1758 = 0, $1759 = 0, $176 = 0, $1760 = 0, $1761 = 0, $1762 = 0, $1763 = 0, $1764 = 0, $1765 = 0, $1766 = 0, $1767 = 0, $1768 = 0, $1769 = 0, $177 = 0, $1770 = 0, $1771 = 0, $1772 = 0;
 var $1773 = 0, $1774 = 0, $1775 = 0, $1776 = 0, $1777 = 0, $1778 = 0, $1779 = 0, $178 = 0, $1780 = 0, $1781 = 0, $1782 = 0, $1783 = 0, $1784 = 0, $1785 = 0, $1786 = 0, $1787 = 0, $1788 = 0, $1789 = 0, $179 = 0, $1790 = 0;
 var $1791 = 0, $1792 = 0, $1793 = 0, $1794 = 0, $1795 = 0, $1796 = 0, $1797 = 0, $1798 = 0, $1799 = 0, $18 = 0, $180 = 0, $1800 = 0, $1801 = 0, $1802 = 0, $1803 = 0, $1804 = 0, $1805 = 0, $1806 = 0, $1807 = 0, $1808 = 0;
 var $1809 = 0, $181 = 0, $1810 = 0, $1811 = 0, $1812 = 0, $1813 = 0, $1814 = 0, $1815 = 0, $1816 = 0, $1817 = 0, $1818 = 0, $1819 = 0, $182 = 0, $1820 = 0, $1821 = 0, $1822 = 0, $1823 = 0, $1824 = 0, $1825 = 0, $1826 = 0;
 var $1827 = 0, $1828 = 0, $1829 = 0, $183 = 0, $1830 = 0, $1831 = 0, $1832 = 0, $1833 = 0, $1834 = 0, $1835 = 0, $1836 = 0, $1837 = 0, $1838 = 0, $1839 = 0, $184 = 0, $1840 = 0, $1841 = 0, $1842 = 0, $1843 = 0, $1844 = 0;
 var $1845 = 0, $1846 = 0, $1847 = 0, $1848 = 0, $1849 = 0, $185 = 0, $1850 = 0, $1851 = 0, $1852 = 0, $1853 = 0, $1854 = 0, $1855 = 0, $1856 = 0, $1857 = 0, $1858 = 0, $1859 = 0, $186 = 0, $1860 = 0, $1861 = 0, $1862 = 0;
 var $1863 = 0, $1864 = 0, $1865 = 0, $1866 = 0, $1867 = 0, $1868 = 0, $1869 = 0, $187 = 0, $1870 = 0, $1871 = 0, $1872 = 0, $1873 = 0, $1874 = 0, $1875 = 0, $1876 = 0, $1877 = 0, $1878 = 0, $1879 = 0, $188 = 0, $1880 = 0;
 var $1881 = 0, $1882 = 0, $1883 = 0, $1884 = 0, $1885 = 0, $1886 = 0, $1887 = 0, $1888 = 0, $1889 = 0, $189 = 0, $1890 = 0, $1891 = 0, $1892 = 0, $1893 = 0, $1894 = 0, $1895 = 0, $1896 = 0, $1897 = 0, $1898 = 0, $1899 = 0;
 var $19 = 0, $190 = 0, $1900 = 0, $1901 = 0, $1902 = 0, $1903 = 0, $1904 = 0, $1905 = 0, $1906 = 0, $1907 = 0, $1908 = 0, $1909 = 0, $191 = 0, $1910 = 0, $1911 = 0, $1912 = 0, $1913 = 0, $1914 = 0, $1915 = 0, $1916 = 0;
 var $1917 = 0, $1918 = 0, $1919 = 0, $192 = 0, $1920 = 0, $1921 = 0, $1922 = 0, $1923 = 0, $1924 = 0, $1925 = 0, $1926 = 0, $1927 = 0, $1928 = 0, $1929 = 0, $193 = 0, $1930 = 0, $1931 = 0, $1932 = 0, $1933 = 0, $1934 = 0;
 var $1935 = 0, $1936 = 0, $1937 = 0, $1938 = 0, $1939 = 0, $194 = 0, $1940 = 0, $1941 = 0, $1942 = 0, $1943 = 0, $1944 = 0, $1945 = 0, $1946 = 0, $1947 = 0, $1948 = 0, $1949 = 0, $195 = 0, $1950 = 0, $1951 = 0, $1952 = 0;
 var $1953 = 0, $1954 = 0, $1955 = 0, $1956 = 0, $1957 = 0, $1958 = 0, $1959 = 0, $196 = 0, $1960 = 0, $1961 = 0, $1962 = 0, $1963 = 0, $1964 = 0, $1965 = 0, $1966 = 0, $1967 = 0, $1968 = 0, $1969 = 0, $197 = 0, $1970 = 0;
 var $1971 = 0, $1972 = 0, $1973 = 0, $1974 = 0, $1975 = 0, $1976 = 0, $1977 = 0, $1978 = 0, $1979 = 0, $198 = 0, $1980 = 0, $1981 = 0, $1982 = 0, $1983 = 0, $1984 = 0, $1985 = 0, $1986 = 0, $1987 = 0, $1988 = 0, $1989 = 0;
 var $199 = 0, $1990 = 0, $1991 = 0, $1992 = 0, $1993 = 0, $1994 = 0, $1995 = 0, $1996 = 0, $1997 = 0, $1998 = 0, $1999 = 0, $2 = 0, $20 = 0, $200 = 0, $2000 = 0, $2001 = 0, $2002 = 0, $2003 = 0, $2004 = 0, $2005 = 0;
 var $2006 = 0, $2007 = 0, $2008 = 0, $2009 = 0, $201 = 0, $2010 = 0, $2011 = 0, $2012 = 0, $2013 = 0, $2014 = 0, $2015 = 0, $2016 = 0, $2017 = 0, $2018 = 0, $2019 = 0, $202 = 0, $2020 = 0, $2021 = 0, $2022 = 0, $2023 = 0;
 var $2024 = 0, $2025 = 0, $2026 = 0, $2027 = 0, $2028 = 0, $2029 = 0, $203 = 0, $2030 = 0, $2031 = 0, $2032 = 0, $2033 = 0, $2034 = 0, $2035 = 0, $2036 = 0, $2037 = 0, $2038 = 0, $2039 = 0, $204 = 0, $2040 = 0, $2041 = 0;
 var $2042 = 0, $2043 = 0, $2044 = 0, $2045 = 0, $2046 = 0, $2047 = 0, $2048 = 0, $2049 = 0, $205 = 0, $2050 = 0, $2051 = 0, $2052 = 0, $2053 = 0, $2054 = 0, $2055 = 0, $2056 = 0, $2057 = 0, $2058 = 0, $2059 = 0, $206 = 0;
 var $2060 = 0, $2061 = 0, $2062 = 0, $2063 = 0, $2064 = 0, $2065 = 0, $2066 = 0, $2067 = 0, $2068 = 0, $2069 = 0, $207 = 0, $2070 = 0, $2071 = 0, $2072 = 0, $2073 = 0, $2074 = 0, $2075 = 0, $2076 = 0, $2077 = 0, $2078 = 0;
 var $2079 = 0, $208 = 0, $2080 = 0, $2081 = 0, $2082 = 0, $2083 = 0, $2084 = 0, $2085 = 0, $2086 = 0, $2087 = 0, $2088 = 0, $2089 = 0, $209 = 0, $2090 = 0, $2091 = 0, $2092 = 0, $2093 = 0, $2094 = 0, $2095 = 0, $2096 = 0;
 var $2097 = 0, $2098 = 0, $2099 = 0, $21 = 0, $210 = 0, $2100 = 0, $2101 = 0, $2102 = 0, $2103 = 0, $2104 = 0, $2105 = 0, $2106 = 0, $2107 = 0, $2108 = 0, $2109 = 0, $211 = 0, $2110 = 0, $2111 = 0, $2112 = 0, $2113 = 0;
 var $2114 = 0, $2115 = 0, $2116 = 0, $2117 = 0, $2118 = 0, $2119 = 0, $212 = 0, $2120 = 0, $2121 = 0, $2122 = 0, $2123 = 0, $2124 = 0, $2125 = 0, $2126 = 0, $2127 = 0, $2128 = 0, $2129 = 0, $213 = 0, $2130 = 0, $2131 = 0;
 var $2132 = 0, $2133 = 0, $2134 = 0, $2135 = 0, $2136 = 0, $2137 = 0, $2138 = 0, $2139 = 0, $214 = 0, $2140 = 0, $2141 = 0, $2142 = 0, $2143 = 0, $2144 = 0, $2145 = 0, $2146 = 0, $2147 = 0, $2148 = 0, $2149 = 0, $215 = 0;
 var $2150 = 0, $2151 = 0, $2152 = 0, $2153 = 0, $2154 = 0, $2155 = 0, $2156 = 0, $2157 = 0, $2158 = 0, $2159 = 0, $216 = 0, $2160 = 0, $2161 = 0, $2162 = 0, $2163 = 0, $2164 = 0, $2165 = 0, $2166 = 0, $2167 = 0, $2168 = 0;
 var $2169 = 0, $217 = 0, $2170 = 0, $2171 = 0, $2172 = 0, $2173 = 0, $2174 = 0, $2175 = 0, $2176 = 0, $2177 = 0, $2178 = 0, $2179 = 0, $218 = 0, $2180 = 0, $2181 = 0, $2182 = 0, $2183 = 0, $2184 = 0, $2185 = 0, $2186 = 0;
 var $2187 = 0, $2188 = 0, $2189 = 0, $219 = 0, $2190 = 0, $2191 = 0, $2192 = 0, $2193 = 0, $2194 = 0, $2195 = 0, $2196 = 0, $2197 = 0, $2198 = 0, $2199 = 0, $22 = 0, $220 = 0, $2200 = 0, $2201 = 0, $2202 = 0, $2203 = 0;
 var $2204 = 0, $2205 = 0, $2206 = 0, $2207 = 0, $2208 = 0, $2209 = 0, $221 = 0, $2210 = 0, $2211 = 0, $2212 = 0, $2213 = 0, $2214 = 0, $2215 = 0, $2216 = 0, $2217 = 0, $2218 = 0, $2219 = 0, $222 = 0, $2220 = 0, $2221 = 0;
 var $2222 = 0, $2223 = 0, $2224 = 0, $2225 = 0, $2226 = 0, $2227 = 0, $2228 = 0, $2229 = 0, $223 = 0, $2230 = 0, $2231 = 0, $2232 = 0, $2233 = 0, $2234 = 0, $2235 = 0, $2236 = 0, $2237 = 0, $2238 = 0, $2239 = 0, $224 = 0;
 var $2240 = 0, $2241 = 0, $2242 = 0, $2243 = 0, $2244 = 0, $2245 = 0, $2246 = 0, $2247 = 0, $2248 = 0, $2249 = 0, $225 = 0, $2250 = 0, $2251 = 0, $2252 = 0, $2253 = 0, $2254 = 0, $2255 = 0, $2256 = 0, $2257 = 0, $2258 = 0;
 var $2259 = 0, $226 = 0, $2260 = 0, $2261 = 0, $2262 = 0, $2263 = 0, $2264 = 0, $2265 = 0, $2266 = 0, $2267 = 0, $2268 = 0, $2269 = 0, $227 = 0, $2270 = 0, $2271 = 0, $2272 = 0, $2273 = 0, $2274 = 0, $2275 = 0, $2276 = 0;
 var $2277 = 0, $2278 = 0, $2279 = 0, $228 = 0, $2280 = 0, $2281 = 0, $2282 = 0, $2283 = 0, $2284 = 0, $2285 = 0, $2286 = 0, $2287 = 0, $2288 = 0, $2289 = 0, $229 = 0, $2290 = 0, $2291 = 0, $2292 = 0, $2293 = 0, $2294 = 0;
 var $2295 = 0, $2296 = 0, $2297 = 0, $2298 = 0, $2299 = 0, $23 = 0, $230 = 0, $2300 = 0, $2301 = 0, $2302 = 0, $2303 = 0, $2304 = 0, $2305 = 0, $2306 = 0, $2307 = 0, $2308 = 0, $2309 = 0, $231 = 0, $2310 = 0, $2311 = 0;
 var $2312 = 0, $2313 = 0, $2314 = 0, $2315 = 0, $2316 = 0, $2317 = 0, $2318 = 0, $2319 = 0, $232 = 0, $2320 = 0, $2321 = 0, $2322 = 0, $2323 = 0, $2324 = 0, $2325 = 0, $2326 = 0, $2327 = 0, $2328 = 0, $2329 = 0, $233 = 0;
 var $2330 = 0, $2331 = 0, $2332 = 0, $2333 = 0, $2334 = 0, $2335 = 0, $2336 = 0, $2337 = 0, $2338 = 0, $2339 = 0, $234 = 0, $2340 = 0, $2341 = 0, $2342 = 0, $2343 = 0, $2344 = 0, $2345 = 0, $2346 = 0, $2347 = 0, $2348 = 0;
 var $2349 = 0, $235 = 0, $2350 = 0, $2351 = 0, $2352 = 0, $2353 = 0, $2354 = 0, $2355 = 0, $2356 = 0, $2357 = 0, $2358 = 0, $2359 = 0, $236 = 0, $2360 = 0, $2361 = 0, $2362 = 0, $2363 = 0, $2364 = 0, $2365 = 0, $2366 = 0;
 var $2367 = 0, $2368 = 0, $2369 = 0, $237 = 0, $2370 = 0, $2371 = 0, $2372 = 0, $2373 = 0, $2374 = 0, $2375 = 0, $2376 = 0, $2377 = 0, $2378 = 0, $2379 = 0, $238 = 0, $2380 = 0, $2381 = 0, $2382 = 0, $2383 = 0, $2384 = 0;
 var $2385 = 0, $2386 = 0, $2387 = 0, $2388 = 0, $2389 = 0, $239 = 0, $2390 = 0, $2391 = 0, $2392 = 0, $2393 = 0, $2394 = 0, $2395 = 0, $2396 = 0, $2397 = 0, $2398 = 0, $2399 = 0, $24 = 0, $240 = 0, $2400 = 0, $2401 = 0;
 var $2402 = 0, $2403 = 0, $2404 = 0, $2405 = 0, $2406 = 0, $2407 = 0, $2408 = 0, $2409 = 0, $241 = 0, $2410 = 0, $2411 = 0, $2412 = 0, $2413 = 0, $2414 = 0, $2415 = 0, $2416 = 0, $2417 = 0, $2418 = 0, $2419 = 0, $242 = 0;
 var $2420 = 0, $2421 = 0, $2422 = 0, $2423 = 0, $2424 = 0, $2425 = 0, $2426 = 0, $2427 = 0, $2428 = 0, $2429 = 0, $243 = 0, $2430 = 0, $2431 = 0, $2432 = 0, $2433 = 0, $2434 = 0, $2435 = 0, $2436 = 0, $2437 = 0, $2438 = 0;
 var $2439 = 0, $244 = 0, $2440 = 0, $2441 = 0, $2442 = 0, $2443 = 0, $2444 = 0, $2445 = 0, $2446 = 0, $2447 = 0, $2448 = 0, $2449 = 0, $245 = 0, $2450 = 0, $2451 = 0, $2452 = 0, $2453 = 0, $2454 = 0, $2455 = 0, $2456 = 0;
 var $2457 = 0, $2458 = 0, $2459 = 0, $246 = 0, $2460 = 0, $2461 = 0, $2462 = 0, $2463 = 0, $2464 = 0, $2465 = 0, $2466 = 0, $2467 = 0, $2468 = 0, $2469 = 0, $247 = 0, $2470 = 0, $2471 = 0, $2472 = 0, $2473 = 0, $2474 = 0;
 var $2475 = 0, $2476 = 0, $2477 = 0, $2478 = 0, $2479 = 0, $248 = 0, $2480 = 0, $2481 = 0, $2482 = 0, $2483 = 0, $2484 = 0, $2485 = 0, $2486 = 0, $2487 = 0, $2488 = 0, $2489 = 0, $249 = 0, $2490 = 0, $2491 = 0, $2492 = 0;
 var $2493 = 0, $2494 = 0, $2495 = 0, $2496 = 0, $2497 = 0, $2498 = 0, $2499 = 0, $25 = 0, $250 = 0, $2500 = 0, $2501 = 0, $2502 = 0, $2503 = 0, $2504 = 0, $2505 = 0, $2506 = 0, $2507 = 0, $2508 = 0, $2509 = 0, $251 = 0;
 var $2510 = 0, $2511 = 0, $2512 = 0, $2513 = 0, $2514 = 0, $2515 = 0, $2516 = 0, $2517 = 0, $2518 = 0, $2519 = 0, $252 = 0, $2520 = 0, $2521 = 0, $2522 = 0, $2523 = 0, $2524 = 0, $2525 = 0, $2526 = 0, $2527 = 0, $2528 = 0;
 var $2529 = 0, $253 = 0, $2530 = 0, $2531 = 0, $2532 = 0, $2533 = 0, $2534 = 0, $2535 = 0, $2536 = 0, $2537 = 0, $2538 = 0, $2539 = 0, $254 = 0, $2540 = 0, $2541 = 0, $2542 = 0, $2543 = 0, $2544 = 0, $2545 = 0, $2546 = 0;
 var $2547 = 0, $2548 = 0, $2549 = 0, $255 = 0, $2550 = 0, $2551 = 0, $2552 = 0, $2553 = 0, $2554 = 0, $2555 = 0, $2556 = 0, $2557 = 0, $2558 = 0, $2559 = 0, $256 = 0, $2560 = 0, $2561 = 0, $2562 = 0, $2563 = 0, $2564 = 0;
 var $2565 = 0, $2566 = 0, $2567 = 0, $2568 = 0, $2569 = 0, $257 = 0, $2570 = 0, $2571 = 0, $2572 = 0, $2573 = 0, $2574 = 0, $2575 = 0, $2576 = 0, $2577 = 0, $2578 = 0, $2579 = 0, $258 = 0, $2580 = 0, $2581 = 0, $2582 = 0;
 var $2583 = 0, $2584 = 0, $2585 = 0, $2586 = 0, $2587 = 0, $2588 = 0, $2589 = 0, $259 = 0, $2590 = 0, $2591 = 0, $2592 = 0, $2593 = 0, $2594 = 0, $2595 = 0, $2596 = 0, $2597 = 0, $2598 = 0, $2599 = 0, $26 = 0, $260 = 0;
 var $2600 = 0, $2601 = 0, $2602 = 0, $2603 = 0, $2604 = 0, $2605 = 0, $2606 = 0, $2607 = 0, $2608 = 0, $2609 = 0, $261 = 0, $2610 = 0, $2611 = 0, $2612 = 0, $2613 = 0, $2614 = 0, $2615 = 0, $2616 = 0, $2617 = 0, $2618 = 0;
 var $2619 = 0, $262 = 0, $2620 = 0, $2621 = 0, $2622 = 0, $2623 = 0, $2624 = 0, $2625 = 0, $2626 = 0, $2627 = 0, $2628 = 0, $2629 = 0, $263 = 0, $2630 = 0, $2631 = 0, $2632 = 0, $2633 = 0, $2634 = 0, $2635 = 0, $2636 = 0;
 var $2637 = 0, $2638 = 0, $2639 = 0, $264 = 0, $2640 = 0, $2641 = 0, $2642 = 0, $2643 = 0, $2644 = 0, $2645 = 0, $2646 = 0, $2647 = 0, $2648 = 0, $2649 = 0, $265 = 0, $2650 = 0, $2651 = 0, $2652 = 0, $2653 = 0, $2654 = 0;
 var $2655 = 0, $2656 = 0, $2657 = 0, $2658 = 0, $2659 = 0, $266 = 0, $2660 = 0, $2661 = 0, $2662 = 0, $2663 = 0, $2664 = 0, $2665 = 0, $2666 = 0, $2667 = 0, $2668 = 0, $2669 = 0, $267 = 0, $2670 = 0, $2671 = 0, $2672 = 0;
 var $2673 = 0, $2674 = 0, $2675 = 0, $2676 = 0, $2677 = 0, $2678 = 0, $2679 = 0, $268 = 0, $2680 = 0, $2681 = 0, $2682 = 0, $2683 = 0, $2684 = 0, $2685 = 0, $2686 = 0, $2687 = 0, $2688 = 0, $2689 = 0, $269 = 0, $2690 = 0;
 var $2691 = 0, $2692 = 0, $2693 = 0, $2694 = 0, $2695 = 0, $2696 = 0, $2697 = 0, $2698 = 0, $2699 = 0, $27 = 0, $270 = 0, $2700 = 0, $2701 = 0, $2702 = 0, $2703 = 0, $2704 = 0, $2705 = 0, $2706 = 0, $2707 = 0, $2708 = 0;
 var $2709 = 0, $271 = 0, $2710 = 0, $2711 = 0, $2712 = 0, $2713 = 0, $2714 = 0, $2715 = 0, $2716 = 0, $2717 = 0, $2718 = 0, $2719 = 0, $272 = 0, $2720 = 0, $2721 = 0, $2722 = 0, $2723 = 0, $2724 = 0, $2725 = 0, $2726 = 0;
 var $2727 = 0, $2728 = 0, $2729 = 0, $273 = 0, $2730 = 0, $2731 = 0, $2732 = 0, $2733 = 0, $2734 = 0, $2735 = 0, $2736 = 0, $2737 = 0, $2738 = 0, $2739 = 0, $274 = 0, $2740 = 0, $2741 = 0, $2742 = 0, $2743 = 0, $2744 = 0;
 var $2745 = 0, $2746 = 0, $2747 = 0, $2748 = 0, $2749 = 0, $275 = 0, $2750 = 0, $2751 = 0, $2752 = 0, $2753 = 0, $2754 = 0, $2755 = 0, $2756 = 0, $2757 = 0, $2758 = 0, $2759 = 0, $276 = 0, $2760 = 0, $2761 = 0, $2762 = 0;
 var $2763 = 0, $2764 = 0, $2765 = 0, $2766 = 0, $2767 = 0, $2768 = 0, $2769 = 0, $277 = 0, $2770 = 0, $2771 = 0, $2772 = 0, $2773 = 0, $2774 = 0, $2775 = 0, $2776 = 0, $2777 = 0, $2778 = 0, $2779 = 0, $278 = 0, $2780 = 0;
 var $2781 = 0, $2782 = 0, $2783 = 0, $2784 = 0, $2785 = 0, $2786 = 0, $2787 = 0, $2788 = 0, $2789 = 0, $279 = 0, $2790 = 0, $2791 = 0, $2792 = 0, $2793 = 0, $2794 = 0, $2795 = 0, $2796 = 0, $2797 = 0, $2798 = 0, $2799 = 0;
 var $28 = 0, $280 = 0, $2800 = 0, $2801 = 0, $2802 = 0, $2803 = 0, $2804 = 0, $2805 = 0, $2806 = 0, $2807 = 0, $2808 = 0, $2809 = 0, $281 = 0, $2810 = 0, $2811 = 0, $2812 = 0, $2813 = 0, $2814 = 0, $2815 = 0, $2816 = 0;
 var $2817 = 0, $2818 = 0, $2819 = 0, $282 = 0, $2820 = 0, $2821 = 0, $2822 = 0, $2823 = 0, $2824 = 0, $2825 = 0, $2826 = 0, $2827 = 0, $2828 = 0, $2829 = 0, $283 = 0, $2830 = 0, $2831 = 0, $2832 = 0, $2833 = 0, $2834 = 0;
 var $2835 = 0, $2836 = 0, $2837 = 0, $2838 = 0, $2839 = 0, $284 = 0, $2840 = 0, $2841 = 0, $2842 = 0, $2843 = 0, $2844 = 0, $2845 = 0, $2846 = 0, $2847 = 0, $2848 = 0, $2849 = 0, $285 = 0, $2850 = 0, $2851 = 0, $2852 = 0;
 var $2853 = 0, $2854 = 0, $2855 = 0, $2856 = 0, $2857 = 0, $2858 = 0, $2859 = 0, $286 = 0, $2860 = 0, $2861 = 0, $2862 = 0, $2863 = 0, $2864 = 0, $2865 = 0, $2866 = 0, $2867 = 0, $2868 = 0, $2869 = 0, $287 = 0, $2870 = 0;
 var $2871 = 0, $2872 = 0, $2873 = 0, $2874 = 0, $2875 = 0, $2876 = 0, $2877 = 0, $2878 = 0, $2879 = 0, $288 = 0, $2880 = 0, $2881 = 0, $2882 = 0, $2883 = 0, $2884 = 0, $2885 = 0, $2886 = 0, $2887 = 0, $2888 = 0, $2889 = 0;
 var $289 = 0, $2890 = 0, $2891 = 0, $2892 = 0, $2893 = 0, $2894 = 0, $2895 = 0, $2896 = 0, $2897 = 0, $2898 = 0, $2899 = 0, $29 = 0, $290 = 0, $2900 = 0, $2901 = 0, $2902 = 0, $2903 = 0, $2904 = 0, $2905 = 0, $2906 = 0;
 var $2907 = 0, $2908 = 0, $2909 = 0, $291 = 0, $2910 = 0, $2911 = 0, $2912 = 0, $2913 = 0, $2914 = 0, $2915 = 0, $2916 = 0, $2917 = 0, $2918 = 0, $2919 = 0, $292 = 0, $2920 = 0, $2921 = 0, $2922 = 0, $2923 = 0, $2924 = 0;
 var $2925 = 0, $2926 = 0, $2927 = 0, $2928 = 0, $2929 = 0, $293 = 0, $2930 = 0, $2931 = 0, $2932 = 0, $2933 = 0, $2934 = 0, $2935 = 0, $2936 = 0, $2937 = 0, $2938 = 0, $2939 = 0, $294 = 0, $2940 = 0, $2941 = 0, $2942 = 0;
 var $2943 = 0, $2944 = 0, $2945 = 0, $2946 = 0, $2947 = 0, $2948 = 0, $2949 = 0, $295 = 0, $2950 = 0, $2951 = 0, $2952 = 0, $2953 = 0, $2954 = 0, $2955 = 0, $2956 = 0, $2957 = 0, $2958 = 0, $2959 = 0, $296 = 0, $2960 = 0;
 var $2961 = 0, $2962 = 0, $2963 = 0, $2964 = 0, $2965 = 0, $2966 = 0, $2967 = 0, $2968 = 0, $2969 = 0, $297 = 0, $2970 = 0, $2971 = 0, $2972 = 0, $2973 = 0, $2974 = 0, $2975 = 0, $2976 = 0, $2977 = 0, $2978 = 0, $2979 = 0;
 var $298 = 0, $2980 = 0, $2981 = 0, $2982 = 0, $2983 = 0, $2984 = 0, $2985 = 0, $2986 = 0, $2987 = 0, $2988 = 0, $2989 = 0, $299 = 0, $2990 = 0, $2991 = 0, $2992 = 0, $2993 = 0, $2994 = 0, $2995 = 0, $2996 = 0, $2997 = 0;
 var $2998 = 0, $2999 = 0, $3 = 0, $30 = 0, $300 = 0, $3000 = 0, $3001 = 0, $3002 = 0, $3003 = 0, $3004 = 0, $3005 = 0, $3006 = 0, $3007 = 0, $3008 = 0, $3009 = 0, $301 = 0, $3010 = 0, $3011 = 0, $3012 = 0, $3013 = 0;
 var $3014 = 0, $3015 = 0, $3016 = 0, $3017 = 0, $3018 = 0, $3019 = 0, $302 = 0, $3020 = 0, $3021 = 0, $3022 = 0, $3023 = 0, $3024 = 0, $3025 = 0, $3026 = 0, $3027 = 0, $3028 = 0, $3029 = 0, $303 = 0, $3030 = 0, $3031 = 0;
 var $3032 = 0, $3033 = 0, $3034 = 0, $3035 = 0, $3036 = 0, $3037 = 0, $3038 = 0, $3039 = 0, $304 = 0, $3040 = 0, $3041 = 0, $3042 = 0, $3043 = 0, $3044 = 0, $3045 = 0, $3046 = 0, $3047 = 0, $3048 = 0, $3049 = 0, $305 = 0;
 var $3050 = 0, $3051 = 0, $3052 = 0, $3053 = 0, $3054 = 0, $3055 = 0, $3056 = 0, $3057 = 0, $3058 = 0, $3059 = 0, $306 = 0, $3060 = 0, $3061 = 0, $3062 = 0, $3063 = 0, $3064 = 0, $3065 = 0, $3066 = 0, $3067 = 0, $3068 = 0;
 var $3069 = 0, $307 = 0, $3070 = 0, $3071 = 0, $3072 = 0, $3073 = 0, $3074 = 0, $3075 = 0, $3076 = 0, $3077 = 0, $3078 = 0, $3079 = 0, $308 = 0, $3080 = 0, $3081 = 0, $3082 = 0, $3083 = 0, $3084 = 0, $3085 = 0, $3086 = 0;
 var $3087 = 0, $3088 = 0, $3089 = 0, $309 = 0, $3090 = 0, $3091 = 0, $3092 = 0, $3093 = 0, $3094 = 0, $3095 = 0, $3096 = 0, $3097 = 0, $3098 = 0, $3099 = 0, $31 = 0, $310 = 0, $3100 = 0, $3101 = 0, $3102 = 0, $3103 = 0;
 var $3104 = 0, $3105 = 0, $3106 = 0, $3107 = 0, $3108 = 0, $3109 = 0, $311 = 0, $3110 = 0, $3111 = 0, $3112 = 0, $3113 = 0, $3114 = 0, $3115 = 0, $3116 = 0, $3117 = 0, $3118 = 0, $3119 = 0, $312 = 0, $3120 = 0, $3121 = 0;
 var $3122 = 0, $3123 = 0, $3124 = 0, $3125 = 0, $3126 = 0, $3127 = 0, $3128 = 0, $3129 = 0, $313 = 0, $3130 = 0, $3131 = 0, $3132 = 0, $3133 = 0, $3134 = 0, $3135 = 0, $3136 = 0, $3137 = 0, $3138 = 0, $3139 = 0, $314 = 0;
 var $3140 = 0, $3141 = 0, $3142 = 0, $3143 = 0, $3144 = 0, $3145 = 0, $3146 = 0, $3147 = 0, $3148 = 0, $3149 = 0, $315 = 0, $3150 = 0, $3151 = 0, $3152 = 0, $3153 = 0, $3154 = 0, $3155 = 0, $3156 = 0, $3157 = 0, $3158 = 0;
 var $3159 = 0, $316 = 0, $3160 = 0, $3161 = 0, $3162 = 0, $3163 = 0, $3164 = 0, $3165 = 0, $3166 = 0, $3167 = 0, $3168 = 0, $3169 = 0, $317 = 0, $3170 = 0, $3171 = 0, $3172 = 0, $3173 = 0, $3174 = 0, $3175 = 0, $3176 = 0;
 var $3177 = 0, $3178 = 0, $3179 = 0, $318 = 0, $3180 = 0, $3181 = 0, $3182 = 0, $3183 = 0, $3184 = 0, $3185 = 0, $3186 = 0, $3187 = 0, $3188 = 0, $3189 = 0, $319 = 0, $3190 = 0, $3191 = 0, $3192 = 0, $3193 = 0, $3194 = 0;
 var $3195 = 0, $3196 = 0, $3197 = 0, $3198 = 0, $3199 = 0, $32 = 0, $320 = 0, $3200 = 0, $3201 = 0, $3202 = 0, $3203 = 0, $3204 = 0, $3205 = 0, $3206 = 0, $3207 = 0, $3208 = 0, $3209 = 0, $321 = 0, $3210 = 0, $3211 = 0;
 var $3212 = 0, $3213 = 0, $3214 = 0, $3215 = 0, $3216 = 0, $3217 = 0, $3218 = 0, $3219 = 0, $322 = 0, $3220 = 0, $3221 = 0, $3222 = 0, $3223 = 0, $3224 = 0, $3225 = 0, $3226 = 0, $3227 = 0, $3228 = 0, $3229 = 0, $323 = 0;
 var $3230 = 0, $3231 = 0, $3232 = 0, $3233 = 0, $3234 = 0, $3235 = 0, $3236 = 0, $3237 = 0, $3238 = 0, $3239 = 0, $324 = 0, $3240 = 0, $3241 = 0, $3242 = 0, $3243 = 0, $3244 = 0, $3245 = 0, $3246 = 0, $3247 = 0, $3248 = 0;
 var $3249 = 0, $325 = 0, $3250 = 0, $3251 = 0, $3252 = 0, $3253 = 0, $3254 = 0, $3255 = 0, $3256 = 0, $3257 = 0, $3258 = 0, $3259 = 0, $326 = 0, $3260 = 0, $3261 = 0, $3262 = 0, $3263 = 0, $3264 = 0, $3265 = 0, $3266 = 0;
 var $3267 = 0, $3268 = 0, $3269 = 0, $327 = 0, $3270 = 0, $3271 = 0, $3272 = 0, $3273 = 0, $3274 = 0, $3275 = 0, $3276 = 0, $3277 = 0, $3278 = 0, $3279 = 0, $328 = 0, $3280 = 0, $3281 = 0, $3282 = 0, $3283 = 0, $3284 = 0;
 var $3285 = 0, $3286 = 0, $3287 = 0, $3288 = 0, $3289 = 0, $329 = 0, $3290 = 0, $3291 = 0, $3292 = 0, $3293 = 0, $3294 = 0, $3295 = 0, $3296 = 0, $3297 = 0, $3298 = 0, $3299 = 0, $33 = 0, $330 = 0, $3300 = 0, $3301 = 0;
 var $3302 = 0, $3303 = 0, $3304 = 0, $3305 = 0, $3306 = 0, $3307 = 0, $3308 = 0, $3309 = 0, $331 = 0, $3310 = 0, $3311 = 0, $3312 = 0, $3313 = 0, $3314 = 0, $3315 = 0, $3316 = 0, $3317 = 0, $3318 = 0, $3319 = 0, $332 = 0;
 var $3320 = 0, $3321 = 0, $3322 = 0, $3323 = 0, $3324 = 0, $3325 = 0, $3326 = 0, $3327 = 0, $3328 = 0, $3329 = 0, $333 = 0, $3330 = 0, $3331 = 0, $3332 = 0, $3333 = 0, $3334 = 0, $3335 = 0, $3336 = 0, $3337 = 0, $3338 = 0;
 var $3339 = 0, $334 = 0, $3340 = 0, $3341 = 0, $3342 = 0, $3343 = 0, $3344 = 0, $3345 = 0, $3346 = 0, $3347 = 0, $3348 = 0, $3349 = 0, $335 = 0, $3350 = 0, $3351 = 0, $3352 = 0, $3353 = 0, $3354 = 0, $3355 = 0, $3356 = 0;
 var $3357 = 0, $3358 = 0, $3359 = 0, $336 = 0, $3360 = 0, $3361 = 0, $3362 = 0, $3363 = 0, $3364 = 0, $3365 = 0, $3366 = 0, $3367 = 0, $3368 = 0, $3369 = 0, $337 = 0, $3370 = 0, $3371 = 0, $3372 = 0, $3373 = 0, $3374 = 0;
 var $3375 = 0, $3376 = 0, $3377 = 0, $3378 = 0, $3379 = 0, $338 = 0, $3380 = 0, $3381 = 0, $3382 = 0, $3383 = 0, $3384 = 0, $3385 = 0, $3386 = 0, $3387 = 0, $3388 = 0, $3389 = 0, $339 = 0, $3390 = 0, $3391 = 0, $3392 = 0;
 var $3393 = 0, $3394 = 0, $3395 = 0, $3396 = 0, $3397 = 0, $3398 = 0, $3399 = 0, $34 = 0, $340 = 0, $3400 = 0, $3401 = 0, $3402 = 0, $3403 = 0, $3404 = 0, $3405 = 0, $3406 = 0, $3407 = 0, $3408 = 0, $3409 = 0, $341 = 0;
 var $3410 = 0, $3411 = 0, $3412 = 0, $3413 = 0, $3414 = 0, $3415 = 0, $3416 = 0, $3417 = 0, $3418 = 0, $3419 = 0, $342 = 0, $3420 = 0, $3421 = 0, $3422 = 0, $3423 = 0, $3424 = 0, $3425 = 0, $3426 = 0, $3427 = 0, $3428 = 0;
 var $3429 = 0, $343 = 0, $3430 = 0, $3431 = 0, $3432 = 0, $3433 = 0, $3434 = 0, $3435 = 0, $3436 = 0, $3437 = 0, $3438 = 0, $3439 = 0, $344 = 0, $3440 = 0, $3441 = 0, $3442 = 0, $3443 = 0, $3444 = 0, $3445 = 0, $3446 = 0;
 var $3447 = 0, $3448 = 0, $3449 = 0, $345 = 0, $3450 = 0, $3451 = 0, $3452 = 0, $3453 = 0, $3454 = 0, $3455 = 0, $3456 = 0, $3457 = 0, $3458 = 0, $3459 = 0, $346 = 0, $3460 = 0, $3461 = 0, $3462 = 0, $3463 = 0, $3464 = 0;
 var $3465 = 0, $3466 = 0, $3467 = 0, $3468 = 0, $3469 = 0, $347 = 0, $3470 = 0, $3471 = 0, $3472 = 0, $3473 = 0, $3474 = 0, $3475 = 0, $3476 = 0, $3477 = 0, $3478 = 0, $3479 = 0, $348 = 0, $3480 = 0, $3481 = 0, $3482 = 0;
 var $3483 = 0, $3484 = 0, $3485 = 0, $3486 = 0, $3487 = 0, $3488 = 0, $3489 = 0, $349 = 0, $3490 = 0, $3491 = 0, $3492 = 0, $3493 = 0, $3494 = 0, $3495 = 0, $3496 = 0, $3497 = 0, $3498 = 0, $3499 = 0, $35 = 0, $350 = 0;
 var $3500 = 0, $3501 = 0, $3502 = 0, $3503 = 0, $3504 = 0, $3505 = 0, $3506 = 0, $3507 = 0, $3508 = 0, $3509 = 0, $351 = 0, $3510 = 0, $3511 = 0, $3512 = 0, $3513 = 0, $3514 = 0, $3515 = 0, $3516 = 0, $3517 = 0, $3518 = 0;
 var $3519 = 0, $352 = 0, $3520 = 0, $3521 = 0, $3522 = 0, $3523 = 0, $3524 = 0, $3525 = 0, $3526 = 0, $3527 = 0, $3528 = 0, $3529 = 0, $353 = 0, $3530 = 0, $3531 = 0, $3532 = 0, $3533 = 0, $3534 = 0, $3535 = 0, $3536 = 0;
 var $3537 = 0, $3538 = 0, $3539 = 0, $354 = 0, $3540 = 0, $3541 = 0, $3542 = 0, $3543 = 0, $3544 = 0, $3545 = 0, $3546 = 0, $3547 = 0, $3548 = 0, $3549 = 0, $355 = 0, $3550 = 0, $3551 = 0, $3552 = 0, $3553 = 0, $3554 = 0;
 var $3555 = 0, $3556 = 0, $3557 = 0, $3558 = 0, $3559 = 0, $356 = 0, $3560 = 0, $3561 = 0, $3562 = 0, $3563 = 0, $3564 = 0, $3565 = 0, $3566 = 0, $3567 = 0, $3568 = 0, $3569 = 0, $357 = 0, $3570 = 0, $3571 = 0, $3572 = 0;
 var $3573 = 0, $3574 = 0, $3575 = 0, $3576 = 0, $3577 = 0, $3578 = 0, $3579 = 0, $358 = 0, $3580 = 0, $3581 = 0, $3582 = 0, $3583 = 0, $3584 = 0, $3585 = 0, $3586 = 0, $3587 = 0, $3588 = 0, $3589 = 0, $359 = 0, $3590 = 0;
 var $3591 = 0, $3592 = 0, $3593 = 0, $3594 = 0, $3595 = 0, $3596 = 0, $3597 = 0, $3598 = 0, $3599 = 0, $36 = 0, $360 = 0, $3600 = 0, $3601 = 0, $3602 = 0, $3603 = 0, $3604 = 0, $3605 = 0, $3606 = 0, $3607 = 0, $3608 = 0;
 var $3609 = 0, $361 = 0, $3610 = 0, $3611 = 0, $3612 = 0, $3613 = 0, $3614 = 0, $3615 = 0, $3616 = 0, $3617 = 0, $3618 = 0, $3619 = 0, $362 = 0, $3620 = 0, $3621 = 0, $3622 = 0, $3623 = 0, $3624 = 0, $3625 = 0, $3626 = 0;
 var $3627 = 0, $3628 = 0, $3629 = 0, $363 = 0, $3630 = 0, $3631 = 0, $3632 = 0, $3633 = 0, $3634 = 0, $3635 = 0, $3636 = 0, $3637 = 0, $3638 = 0, $3639 = 0, $364 = 0, $3640 = 0, $3641 = 0, $3642 = 0, $3643 = 0, $3644 = 0;
 var $3645 = 0, $3646 = 0, $3647 = 0, $3648 = 0, $3649 = 0, $365 = 0, $3650 = 0, $3651 = 0, $3652 = 0, $3653 = 0, $3654 = 0, $3655 = 0, $3656 = 0, $3657 = 0, $3658 = 0, $3659 = 0, $366 = 0, $3660 = 0, $3661 = 0, $3662 = 0;
 var $3663 = 0, $3664 = 0, $3665 = 0, $3666 = 0, $3667 = 0, $3668 = 0, $3669 = 0, $367 = 0, $3670 = 0, $3671 = 0, $3672 = 0, $3673 = 0, $3674 = 0, $3675 = 0, $3676 = 0, $3677 = 0, $3678 = 0, $3679 = 0, $368 = 0, $3680 = 0;
 var $3681 = 0, $3682 = 0, $3683 = 0, $3684 = 0, $3685 = 0, $3686 = 0, $3687 = 0, $3688 = 0, $3689 = 0, $369 = 0, $3690 = 0, $3691 = 0, $3692 = 0, $3693 = 0, $3694 = 0, $3695 = 0, $3696 = 0, $3697 = 0, $3698 = 0, $3699 = 0;
 var $37 = 0, $370 = 0, $3700 = 0, $3701 = 0, $3702 = 0, $3703 = 0, $3704 = 0, $3705 = 0, $3706 = 0, $3707 = 0, $3708 = 0, $3709 = 0, $371 = 0, $3710 = 0, $3711 = 0, $3712 = 0, $3713 = 0, $3714 = 0, $3715 = 0, $3716 = 0;
 var $3717 = 0, $3718 = 0, $3719 = 0, $372 = 0, $3720 = 0, $3721 = 0, $3722 = 0, $3723 = 0, $3724 = 0, $3725 = 0, $3726 = 0, $3727 = 0, $3728 = 0, $3729 = 0, $373 = 0, $3730 = 0, $3731 = 0, $3732 = 0, $3733 = 0, $3734 = 0;
 var $3735 = 0, $3736 = 0, $3737 = 0, $3738 = 0, $3739 = 0, $374 = 0, $3740 = 0, $3741 = 0, $3742 = 0, $3743 = 0, $3744 = 0, $3745 = 0, $3746 = 0, $3747 = 0, $3748 = 0, $3749 = 0, $375 = 0, $3750 = 0, $3751 = 0, $3752 = 0;
 var $3753 = 0, $3754 = 0, $3755 = 0, $3756 = 0, $3757 = 0, $3758 = 0, $3759 = 0, $376 = 0, $3760 = 0, $3761 = 0, $3762 = 0, $3763 = 0, $3764 = 0, $3765 = 0, $3766 = 0, $3767 = 0, $3768 = 0, $3769 = 0, $377 = 0, $3770 = 0;
 var $3771 = 0, $3772 = 0, $3773 = 0, $3774 = 0, $3775 = 0, $3776 = 0, $3777 = 0, $3778 = 0, $3779 = 0, $378 = 0, $3780 = 0, $3781 = 0, $3782 = 0, $3783 = 0, $3784 = 0, $3785 = 0, $3786 = 0, $3787 = 0, $3788 = 0, $3789 = 0;
 var $379 = 0, $3790 = 0, $3791 = 0, $3792 = 0, $3793 = 0, $3794 = 0, $3795 = 0, $3796 = 0, $3797 = 0, $3798 = 0, $3799 = 0, $38 = 0, $380 = 0, $3800 = 0, $3801 = 0, $3802 = 0, $3803 = 0, $3804 = 0, $3805 = 0, $3806 = 0;
 var $3807 = 0, $3808 = 0, $3809 = 0, $381 = 0, $3810 = 0, $3811 = 0, $3812 = 0, $3813 = 0, $3814 = 0, $3815 = 0, $3816 = 0, $3817 = 0, $3818 = 0, $3819 = 0, $382 = 0, $3820 = 0, $3821 = 0, $3822 = 0, $3823 = 0, $3824 = 0;
 var $3825 = 0, $3826 = 0, $3827 = 0, $3828 = 0, $3829 = 0, $383 = 0, $3830 = 0, $3831 = 0, $3832 = 0, $3833 = 0, $3834 = 0, $3835 = 0, $3836 = 0, $3837 = 0, $3838 = 0, $3839 = 0, $384 = 0, $3840 = 0, $3841 = 0, $3842 = 0;
 var $3843 = 0, $3844 = 0, $3845 = 0, $3846 = 0, $3847 = 0, $3848 = 0, $3849 = 0, $385 = 0, $3850 = 0, $3851 = 0, $3852 = 0, $3853 = 0, $3854 = 0, $3855 = 0, $3856 = 0, $3857 = 0, $3858 = 0, $3859 = 0, $386 = 0, $3860 = 0;
 var $3861 = 0, $3862 = 0, $3863 = 0, $3864 = 0, $3865 = 0, $3866 = 0, $3867 = 0, $3868 = 0, $3869 = 0, $387 = 0, $3870 = 0, $3871 = 0, $3872 = 0, $3873 = 0, $3874 = 0, $3875 = 0, $3876 = 0, $3877 = 0, $3878 = 0, $3879 = 0;
 var $388 = 0, $3880 = 0, $3881 = 0, $3882 = 0, $3883 = 0, $3884 = 0, $3885 = 0, $3886 = 0, $3887 = 0, $3888 = 0, $3889 = 0, $389 = 0, $3890 = 0, $3891 = 0, $3892 = 0, $3893 = 0, $3894 = 0, $3895 = 0, $3896 = 0, $3897 = 0;
 var $3898 = 0, $3899 = 0, $39 = 0, $390 = 0, $3900 = 0, $3901 = 0, $3902 = 0, $3903 = 0, $3904 = 0, $3905 = 0, $3906 = 0, $3907 = 0, $3908 = 0, $3909 = 0, $391 = 0, $3910 = 0, $3911 = 0, $3912 = 0, $3913 = 0, $3914 = 0;
 var $3915 = 0, $3916 = 0, $3917 = 0, $3918 = 0, $3919 = 0, $392 = 0, $3920 = 0, $3921 = 0, $3922 = 0, $3923 = 0, $3924 = 0, $3925 = 0, $3926 = 0, $3927 = 0, $3928 = 0, $3929 = 0, $393 = 0, $3930 = 0, $3931 = 0, $3932 = 0;
 var $3933 = 0, $3934 = 0, $3935 = 0, $3936 = 0, $3937 = 0, $3938 = 0, $3939 = 0, $394 = 0, $3940 = 0, $3941 = 0, $3942 = 0, $3943 = 0, $3944 = 0, $3945 = 0, $3946 = 0, $3947 = 0, $3948 = 0, $3949 = 0, $395 = 0, $3950 = 0;
 var $3951 = 0, $3952 = 0, $3953 = 0, $3954 = 0, $3955 = 0, $3956 = 0, $3957 = 0, $3958 = 0, $3959 = 0, $396 = 0, $3960 = 0, $3961 = 0, $3962 = 0, $3963 = 0, $3964 = 0, $3965 = 0, $3966 = 0, $3967 = 0, $3968 = 0, $3969 = 0;
 var $397 = 0, $3970 = 0, $3971 = 0, $3972 = 0, $3973 = 0, $3974 = 0, $3975 = 0, $3976 = 0, $3977 = 0, $3978 = 0, $3979 = 0, $398 = 0, $3980 = 0, $3981 = 0, $3982 = 0, $3983 = 0, $3984 = 0, $3985 = 0, $3986 = 0, $3987 = 0;
 var $3988 = 0, $3989 = 0, $399 = 0, $3990 = 0, $3991 = 0, $3992 = 0, $3993 = 0, $3994 = 0, $3995 = 0, $3996 = 0, $3997 = 0, $3998 = 0, $3999 = 0, $4 = 0, $40 = 0, $400 = 0, $4000 = 0, $4001 = 0, $4002 = 0, $4003 = 0;
 var $4004 = 0, $4005 = 0, $4006 = 0, $4007 = 0, $4008 = 0, $4009 = 0, $401 = 0, $4010 = 0, $4011 = 0, $4012 = 0, $4013 = 0, $4014 = 0, $4015 = 0, $4016 = 0, $4017 = 0, $4018 = 0, $4019 = 0, $402 = 0, $4020 = 0, $4021 = 0;
 var $4022 = 0, $4023 = 0, $4024 = 0, $4025 = 0, $4026 = 0, $4027 = 0, $4028 = 0, $4029 = 0, $403 = 0, $4030 = 0, $4031 = 0, $4032 = 0, $4033 = 0, $4034 = 0, $4035 = 0, $4036 = 0, $4037 = 0, $4038 = 0, $4039 = 0, $404 = 0;
 var $4040 = 0, $4041 = 0, $4042 = 0, $4043 = 0, $4044 = 0, $4045 = 0, $4046 = 0, $4047 = 0, $4048 = 0, $4049 = 0, $405 = 0, $4050 = 0, $4051 = 0, $4052 = 0, $4053 = 0, $4054 = 0, $4055 = 0, $4056 = 0, $4057 = 0, $4058 = 0;
 var $4059 = 0, $406 = 0, $4060 = 0, $4061 = 0, $4062 = 0, $4063 = 0, $4064 = 0, $4065 = 0, $4066 = 0, $4067 = 0, $4068 = 0, $4069 = 0, $407 = 0, $4070 = 0, $4071 = 0, $4072 = 0, $4073 = 0, $4074 = 0, $4075 = 0, $4076 = 0;
 var $4077 = 0, $4078 = 0, $4079 = 0, $408 = 0, $4080 = 0, $4081 = 0, $4082 = 0, $4083 = 0, $4084 = 0, $4085 = 0, $4086 = 0, $4087 = 0, $4088 = 0, $4089 = 0, $409 = 0, $4090 = 0, $4091 = 0, $4092 = 0, $4093 = 0, $4094 = 0;
 var $4095 = 0, $4096 = 0, $4097 = 0, $4098 = 0, $4099 = 0, $41 = 0, $410 = 0, $4100 = 0, $4101 = 0, $4102 = 0, $4103 = 0, $4104 = 0, $4105 = 0, $4106 = 0, $4107 = 0, $4108 = 0, $4109 = 0, $411 = 0, $4110 = 0, $4111 = 0;
 var $4112 = 0, $4113 = 0, $4114 = 0, $4115 = 0, $4116 = 0, $4117 = 0, $4118 = 0, $4119 = 0, $412 = 0, $4120 = 0, $4121 = 0, $4122 = 0, $4123 = 0, $4124 = 0, $4125 = 0, $4126 = 0, $4127 = 0, $4128 = 0, $4129 = 0, $413 = 0;
 var $4130 = 0, $4131 = 0, $4132 = 0, $4133 = 0, $4134 = 0, $4135 = 0, $4136 = 0, $4137 = 0, $4138 = 0, $4139 = 0, $414 = 0, $4140 = 0, $4141 = 0, $4142 = 0, $4143 = 0, $4144 = 0, $4145 = 0, $4146 = 0, $4147 = 0, $4148 = 0;
 var $4149 = 0, $415 = 0, $4150 = 0, $4151 = 0, $4152 = 0, $4153 = 0, $4154 = 0, $4155 = 0, $4156 = 0, $4157 = 0, $4158 = 0, $4159 = 0, $416 = 0, $4160 = 0, $4161 = 0, $4162 = 0, $4163 = 0, $4164 = 0, $4165 = 0, $4166 = 0;
 var $4167 = 0, $4168 = 0, $4169 = 0, $417 = 0, $4170 = 0, $4171 = 0, $4172 = 0, $4173 = 0, $4174 = 0, $4175 = 0, $4176 = 0, $4177 = 0, $4178 = 0, $4179 = 0, $418 = 0, $4180 = 0, $4181 = 0, $4182 = 0, $4183 = 0, $4184 = 0;
 var $4185 = 0, $4186 = 0, $4187 = 0, $4188 = 0, $4189 = 0, $419 = 0, $4190 = 0, $4191 = 0, $4192 = 0, $4193 = 0, $4194 = 0, $4195 = 0, $4196 = 0, $4197 = 0, $4198 = 0, $4199 = 0, $42 = 0, $420 = 0, $4200 = 0, $4201 = 0;
 var $4202 = 0, $4203 = 0, $4204 = 0, $4205 = 0, $4206 = 0, $4207 = 0, $4208 = 0, $4209 = 0, $421 = 0, $4210 = 0, $4211 = 0, $4212 = 0, $4213 = 0, $4214 = 0, $4215 = 0, $4216 = 0, $4217 = 0, $4218 = 0, $4219 = 0, $422 = 0;
 var $4220 = 0, $4221 = 0, $4222 = 0, $4223 = 0, $4224 = 0, $4225 = 0, $4226 = 0, $4227 = 0, $4228 = 0, $4229 = 0, $423 = 0, $4230 = 0, $4231 = 0, $4232 = 0, $4233 = 0, $4234 = 0, $4235 = 0, $4236 = 0, $4237 = 0, $4238 = 0;
 var $4239 = 0, $424 = 0, $4240 = 0, $4241 = 0, $4242 = 0, $4243 = 0, $4244 = 0, $4245 = 0, $4246 = 0, $4247 = 0, $4248 = 0, $4249 = 0, $425 = 0, $4250 = 0, $4251 = 0, $4252 = 0, $4253 = 0, $4254 = 0, $4255 = 0, $4256 = 0;
 var $4257 = 0, $4258 = 0, $4259 = 0, $426 = 0, $4260 = 0, $4261 = 0, $4262 = 0, $4263 = 0, $4264 = 0, $4265 = 0, $4266 = 0, $4267 = 0, $4268 = 0, $4269 = 0, $427 = 0, $4270 = 0, $4271 = 0, $4272 = 0, $4273 = 0, $4274 = 0;
 var $4275 = 0, $4276 = 0, $4277 = 0, $4278 = 0, $4279 = 0, $428 = 0, $4280 = 0, $4281 = 0, $4282 = 0, $4283 = 0, $4284 = 0, $4285 = 0, $4286 = 0, $4287 = 0, $4288 = 0, $4289 = 0, $429 = 0, $4290 = 0, $4291 = 0, $4292 = 0;
 var $4293 = 0, $4294 = 0, $4295 = 0, $4296 = 0, $4297 = 0, $4298 = 0, $4299 = 0, $43 = 0, $430 = 0, $4300 = 0, $4301 = 0, $4302 = 0, $4303 = 0, $4304 = 0, $4305 = 0, $4306 = 0, $4307 = 0, $4308 = 0, $4309 = 0, $431 = 0;
 var $4310 = 0, $4311 = 0, $4312 = 0, $4313 = 0, $4314 = 0, $4315 = 0, $4316 = 0, $4317 = 0, $4318 = 0, $4319 = 0, $432 = 0, $4320 = 0, $4321 = 0, $4322 = 0, $4323 = 0, $4324 = 0, $4325 = 0, $4326 = 0, $4327 = 0, $4328 = 0;
 var $4329 = 0, $433 = 0, $4330 = 0, $4331 = 0, $4332 = 0, $4333 = 0, $4334 = 0, $4335 = 0, $4336 = 0, $4337 = 0, $4338 = 0, $4339 = 0, $434 = 0, $4340 = 0, $4341 = 0, $4342 = 0, $4343 = 0, $4344 = 0, $4345 = 0, $4346 = 0;
 var $4347 = 0, $4348 = 0, $4349 = 0, $435 = 0, $4350 = 0, $4351 = 0, $4352 = 0, $4353 = 0, $4354 = 0, $4355 = 0, $4356 = 0, $4357 = 0, $4358 = 0, $4359 = 0, $436 = 0, $4360 = 0, $4361 = 0, $4362 = 0, $4363 = 0, $4364 = 0;
 var $4365 = 0, $4366 = 0, $4367 = 0, $4368 = 0, $4369 = 0, $437 = 0, $4370 = 0, $4371 = 0, $4372 = 0, $4373 = 0, $4374 = 0, $4375 = 0, $4376 = 0, $4377 = 0, $4378 = 0, $4379 = 0, $438 = 0, $4380 = 0, $4381 = 0, $4382 = 0;
 var $4383 = 0, $4384 = 0, $4385 = 0, $4386 = 0, $4387 = 0, $4388 = 0, $4389 = 0, $439 = 0, $4390 = 0, $4391 = 0, $4392 = 0, $4393 = 0, $4394 = 0, $4395 = 0, $4396 = 0, $4397 = 0, $4398 = 0, $4399 = 0, $44 = 0, $440 = 0;
 var $4400 = 0, $4401 = 0, $4402 = 0, $4403 = 0, $4404 = 0, $4405 = 0, $4406 = 0, $4407 = 0, $4408 = 0, $4409 = 0, $441 = 0, $4410 = 0, $4411 = 0, $4412 = 0, $4413 = 0, $4414 = 0, $4415 = 0, $4416 = 0, $4417 = 0, $4418 = 0;
 var $4419 = 0, $442 = 0, $4420 = 0, $4421 = 0, $4422 = 0, $4423 = 0, $4424 = 0, $4425 = 0, $4426 = 0, $4427 = 0, $4428 = 0, $4429 = 0, $443 = 0, $4430 = 0, $4431 = 0, $4432 = 0, $4433 = 0, $4434 = 0, $4435 = 0, $4436 = 0;
 var $4437 = 0, $4438 = 0, $4439 = 0, $444 = 0, $4440 = 0, $4441 = 0, $4442 = 0, $4443 = 0, $4444 = 0, $4445 = 0, $4446 = 0, $4447 = 0, $4448 = 0, $4449 = 0, $445 = 0, $4450 = 0, $4451 = 0, $4452 = 0, $4453 = 0, $4454 = 0;
 var $4455 = 0, $4456 = 0, $4457 = 0, $4458 = 0, $4459 = 0, $446 = 0, $4460 = 0, $4461 = 0, $4462 = 0, $4463 = 0, $4464 = 0, $4465 = 0, $4466 = 0, $4467 = 0, $4468 = 0, $4469 = 0, $447 = 0, $4470 = 0, $4471 = 0, $4472 = 0;
 var $4473 = 0, $4474 = 0, $4475 = 0, $4476 = 0, $4477 = 0, $4478 = 0, $4479 = 0, $448 = 0, $4480 = 0, $4481 = 0, $4482 = 0, $4483 = 0, $4484 = 0, $4485 = 0, $4486 = 0, $4487 = 0, $4488 = 0, $4489 = 0, $449 = 0, $4490 = 0;
 var $4491 = 0, $4492 = 0, $4493 = 0, $4494 = 0, $4495 = 0, $4496 = 0, $4497 = 0, $4498 = 0, $4499 = 0, $45 = 0, $450 = 0, $4500 = 0, $4501 = 0, $4502 = 0, $4503 = 0, $4504 = 0, $4505 = 0, $4506 = 0, $4507 = 0, $4508 = 0;
 var $4509 = 0, $451 = 0, $4510 = 0, $4511 = 0, $4512 = 0, $4513 = 0, $4514 = 0, $4515 = 0, $4516 = 0, $4517 = 0, $4518 = 0, $4519 = 0, $452 = 0, $4520 = 0, $4521 = 0, $4522 = 0, $4523 = 0, $4524 = 0, $4525 = 0, $4526 = 0;
 var $4527 = 0, $4528 = 0, $4529 = 0, $453 = 0, $4530 = 0, $4531 = 0, $4532 = 0, $4533 = 0, $4534 = 0, $4535 = 0, $4536 = 0, $4537 = 0, $4538 = 0, $4539 = 0, $454 = 0, $4540 = 0, $4541 = 0, $4542 = 0, $4543 = 0, $4544 = 0;
 var $4545 = 0, $4546 = 0, $4547 = 0, $4548 = 0, $4549 = 0, $455 = 0, $4550 = 0, $4551 = 0, $4552 = 0, $4553 = 0, $4554 = 0, $4555 = 0, $4556 = 0, $4557 = 0, $4558 = 0, $4559 = 0, $456 = 0, $4560 = 0, $4561 = 0, $4562 = 0;
 var $4563 = 0, $4564 = 0, $4565 = 0, $4566 = 0, $4567 = 0, $4568 = 0, $4569 = 0, $457 = 0, $4570 = 0, $4571 = 0, $4572 = 0, $4573 = 0, $4574 = 0, $4575 = 0, $4576 = 0, $4577 = 0, $4578 = 0, $4579 = 0, $458 = 0, $4580 = 0;
 var $4581 = 0, $4582 = 0, $4583 = 0, $4584 = 0, $4585 = 0, $4586 = 0, $4587 = 0, $4588 = 0, $4589 = 0, $459 = 0, $4590 = 0, $4591 = 0, $4592 = 0, $4593 = 0, $4594 = 0, $4595 = 0, $4596 = 0, $4597 = 0, $4598 = 0, $4599 = 0;
 var $46 = 0, $460 = 0, $4600 = 0, $4601 = 0, $4602 = 0, $4603 = 0, $4604 = 0, $4605 = 0, $4606 = 0, $4607 = 0, $4608 = 0, $4609 = 0, $461 = 0, $4610 = 0, $4611 = 0, $4612 = 0, $4613 = 0, $4614 = 0, $4615 = 0, $4616 = 0;
 var $4617 = 0, $4618 = 0, $4619 = 0, $462 = 0, $4620 = 0, $4621 = 0, $4622 = 0, $4623 = 0, $4624 = 0, $4625 = 0, $4626 = 0, $4627 = 0, $4628 = 0, $4629 = 0, $463 = 0, $4630 = 0, $4631 = 0, $4632 = 0, $4633 = 0, $4634 = 0;
 var $4635 = 0, $4636 = 0, $4637 = 0, $4638 = 0, $4639 = 0, $464 = 0, $4640 = 0, $4641 = 0, $4642 = 0, $4643 = 0, $4644 = 0, $4645 = 0, $4646 = 0, $4647 = 0, $4648 = 0, $4649 = 0, $465 = 0, $4650 = 0, $4651 = 0, $4652 = 0;
 var $4653 = 0, $4654 = 0, $4655 = 0, $4656 = 0, $4657 = 0, $4658 = 0, $4659 = 0, $466 = 0, $4660 = 0, $4661 = 0, $4662 = 0, $4663 = 0, $4664 = 0, $4665 = 0, $4666 = 0, $4667 = 0, $4668 = 0, $4669 = 0, $467 = 0, $4670 = 0;
 var $4671 = 0, $4672 = 0, $4673 = 0, $4674 = 0, $4675 = 0, $4676 = 0, $4677 = 0, $4678 = 0, $4679 = 0, $468 = 0, $4680 = 0, $4681 = 0, $4682 = 0, $4683 = 0, $4684 = 0, $4685 = 0, $4686 = 0, $4687 = 0, $4688 = 0, $4689 = 0;
 var $469 = 0, $4690 = 0, $4691 = 0, $4692 = 0, $4693 = 0, $4694 = 0, $4695 = 0, $4696 = 0, $4697 = 0, $4698 = 0, $4699 = 0, $47 = 0, $470 = 0, $4700 = 0, $4701 = 0, $4702 = 0, $4703 = 0, $4704 = 0, $4705 = 0, $4706 = 0;
 var $4707 = 0, $4708 = 0, $4709 = 0, $471 = 0, $4710 = 0, $4711 = 0, $4712 = 0, $4713 = 0, $4714 = 0, $4715 = 0, $4716 = 0, $4717 = 0, $4718 = 0, $4719 = 0, $472 = 0, $4720 = 0, $4721 = 0, $4722 = 0, $4723 = 0, $4724 = 0;
 var $4725 = 0, $4726 = 0, $4727 = 0, $4728 = 0, $4729 = 0, $473 = 0, $4730 = 0, $4731 = 0, $4732 = 0, $4733 = 0, $4734 = 0, $4735 = 0, $4736 = 0, $4737 = 0, $4738 = 0, $4739 = 0, $474 = 0, $4740 = 0, $4741 = 0, $4742 = 0;
 var $4743 = 0, $4744 = 0, $4745 = 0, $4746 = 0, $4747 = 0, $4748 = 0, $4749 = 0, $475 = 0, $4750 = 0, $4751 = 0, $4752 = 0, $4753 = 0, $4754 = 0, $4755 = 0, $4756 = 0, $4757 = 0, $4758 = 0, $4759 = 0, $476 = 0, $4760 = 0;
 var $4761 = 0, $4762 = 0, $4763 = 0, $4764 = 0, $4765 = 0, $4766 = 0, $4767 = 0, $4768 = 0, $4769 = 0, $477 = 0, $4770 = 0, $4771 = 0, $4772 = 0, $4773 = 0, $4774 = 0, $4775 = 0, $4776 = 0, $4777 = 0, $4778 = 0, $4779 = 0;
 var $478 = 0, $4780 = 0, $4781 = 0, $4782 = 0, $4783 = 0, $4784 = 0, $4785 = 0, $4786 = 0, $4787 = 0, $4788 = 0, $4789 = 0, $479 = 0, $4790 = 0, $4791 = 0, $4792 = 0, $4793 = 0, $4794 = 0, $4795 = 0, $4796 = 0, $4797 = 0;
 var $4798 = 0, $4799 = 0, $48 = 0, $480 = 0, $4800 = 0, $4801 = 0, $4802 = 0, $4803 = 0, $4804 = 0, $4805 = 0, $4806 = 0, $4807 = 0, $4808 = 0, $4809 = 0, $481 = 0, $4810 = 0, $4811 = 0, $4812 = 0, $4813 = 0, $4814 = 0;
 var $4815 = 0, $4816 = 0, $4817 = 0, $4818 = 0, $4819 = 0, $482 = 0, $4820 = 0, $4821 = 0, $4822 = 0, $4823 = 0, $4824 = 0, $4825 = 0, $4826 = 0, $4827 = 0, $4828 = 0, $4829 = 0, $483 = 0, $4830 = 0, $4831 = 0, $4832 = 0;
 var $4833 = 0, $4834 = 0, $4835 = 0, $4836 = 0, $4837 = 0, $4838 = 0, $4839 = 0, $484 = 0, $4840 = 0, $4841 = 0, $4842 = 0, $4843 = 0, $4844 = 0, $4845 = 0, $4846 = 0, $4847 = 0, $4848 = 0, $4849 = 0, $485 = 0, $4850 = 0;
 var $4851 = 0, $4852 = 0, $4853 = 0, $4854 = 0, $4855 = 0, $4856 = 0, $4857 = 0, $4858 = 0, $4859 = 0, $486 = 0, $4860 = 0, $4861 = 0, $4862 = 0, $4863 = 0, $4864 = 0, $4865 = 0, $4866 = 0, $4867 = 0, $4868 = 0, $4869 = 0;
 var $487 = 0, $4870 = 0, $4871 = 0, $4872 = 0, $4873 = 0, $4874 = 0, $4875 = 0, $4876 = 0, $4877 = 0, $4878 = 0, $4879 = 0, $488 = 0, $4880 = 0, $4881 = 0, $4882 = 0, $4883 = 0, $4884 = 0, $4885 = 0, $4886 = 0, $4887 = 0;
 var $4888 = 0, $4889 = 0, $489 = 0, $4890 = 0, $4891 = 0, $4892 = 0, $4893 = 0, $4894 = 0, $4895 = 0, $4896 = 0, $4897 = 0, $4898 = 0, $4899 = 0, $49 = 0, $490 = 0, $4900 = 0, $4901 = 0, $4902 = 0, $4903 = 0, $4904 = 0;
 var $4905 = 0, $4906 = 0, $4907 = 0, $4908 = 0, $4909 = 0, $491 = 0, $4910 = 0, $4911 = 0, $4912 = 0, $4913 = 0, $4914 = 0, $4915 = 0, $4916 = 0, $4917 = 0, $4918 = 0, $4919 = 0, $492 = 0, $4920 = 0, $4921 = 0, $4922 = 0;
 var $4923 = 0, $4924 = 0, $4925 = 0, $4926 = 0, $4927 = 0, $4928 = 0, $4929 = 0, $493 = 0, $4930 = 0, $4931 = 0, $4932 = 0, $4933 = 0, $4934 = 0, $4935 = 0, $4936 = 0, $4937 = 0, $4938 = 0, $4939 = 0, $494 = 0, $4940 = 0;
 var $4941 = 0, $4942 = 0, $4943 = 0, $4944 = 0, $4945 = 0, $4946 = 0, $4947 = 0, $4948 = 0, $4949 = 0, $495 = 0, $4950 = 0, $4951 = 0, $4952 = 0, $4953 = 0, $4954 = 0, $4955 = 0, $4956 = 0, $4957 = 0, $4958 = 0, $4959 = 0;
 var $496 = 0, $4960 = 0, $4961 = 0, $4962 = 0, $4963 = 0, $4964 = 0, $4965 = 0, $4966 = 0, $4967 = 0, $4968 = 0, $4969 = 0, $497 = 0, $4970 = 0, $4971 = 0, $4972 = 0, $4973 = 0, $4974 = 0, $4975 = 0, $4976 = 0, $4977 = 0;
 var $4978 = 0, $4979 = 0, $498 = 0, $4980 = 0, $4981 = 0, $4982 = 0, $4983 = 0, $4984 = 0, $4985 = 0, $4986 = 0, $4987 = 0, $4988 = 0, $4989 = 0, $499 = 0, $4990 = 0, $4991 = 0, $4992 = 0, $4993 = 0, $4994 = 0, $4995 = 0;
 var $4996 = 0, $4997 = 0, $4998 = 0, $4999 = 0, $5 = 0, $50 = 0, $500 = 0, $5000 = 0, $5001 = 0, $5002 = 0, $5003 = 0, $5004 = 0, $5005 = 0, $5006 = 0, $5007 = 0, $5008 = 0, $5009 = 0, $501 = 0, $5010 = 0, $5011 = 0;
 var $5012 = 0, $5013 = 0, $5014 = 0, $5015 = 0, $5016 = 0, $5017 = 0, $5018 = 0, $5019 = 0, $502 = 0, $5020 = 0, $5021 = 0, $5022 = 0, $5023 = 0, $5024 = 0, $5025 = 0, $5026 = 0, $5027 = 0, $5028 = 0, $5029 = 0, $503 = 0;
 var $5030 = 0, $5031 = 0, $5032 = 0, $5033 = 0, $5034 = 0, $5035 = 0, $5036 = 0, $5037 = 0, $5038 = 0, $5039 = 0, $504 = 0, $5040 = 0, $5041 = 0, $5042 = 0, $5043 = 0, $5044 = 0, $5045 = 0, $5046 = 0, $5047 = 0, $5048 = 0;
 var $5049 = 0, $505 = 0, $5050 = 0, $5051 = 0, $5052 = 0, $5053 = 0, $5054 = 0, $5055 = 0, $5056 = 0, $5057 = 0, $5058 = 0, $5059 = 0, $506 = 0, $5060 = 0, $5061 = 0, $5062 = 0, $5063 = 0, $5064 = 0, $5065 = 0, $5066 = 0;
 var $5067 = 0, $5068 = 0, $5069 = 0, $507 = 0, $5070 = 0, $5071 = 0, $5072 = 0, $5073 = 0, $5074 = 0, $5075 = 0, $5076 = 0, $5077 = 0, $5078 = 0, $5079 = 0, $508 = 0, $5080 = 0, $5081 = 0, $5082 = 0, $5083 = 0, $5084 = 0;
 var $5085 = 0, $5086 = 0, $5087 = 0, $5088 = 0, $5089 = 0, $509 = 0, $5090 = 0, $5091 = 0, $5092 = 0, $5093 = 0, $5094 = 0, $5095 = 0, $5096 = 0, $5097 = 0, $5098 = 0, $5099 = 0, $51 = 0, $510 = 0, $5100 = 0, $5101 = 0;
 var $5102 = 0, $5103 = 0, $5104 = 0, $5105 = 0, $5106 = 0, $5107 = 0, $5108 = 0, $5109 = 0, $511 = 0, $5110 = 0, $5111 = 0, $5112 = 0, $5113 = 0, $5114 = 0, $5115 = 0, $5116 = 0, $5117 = 0, $5118 = 0, $5119 = 0, $512 = 0;
 var $5120 = 0, $5121 = 0, $5122 = 0, $5123 = 0, $5124 = 0, $5125 = 0, $5126 = 0, $5127 = 0, $5128 = 0, $5129 = 0, $513 = 0, $5130 = 0, $5131 = 0, $5132 = 0, $5133 = 0, $5134 = 0, $5135 = 0, $5136 = 0, $5137 = 0, $5138 = 0;
 var $5139 = 0, $514 = 0, $5140 = 0, $5141 = 0, $5142 = 0, $5143 = 0, $5144 = 0, $5145 = 0, $5146 = 0, $5147 = 0, $5148 = 0, $5149 = 0, $515 = 0, $5150 = 0, $5151 = 0, $5152 = 0, $5153 = 0, $5154 = 0, $5155 = 0, $5156 = 0;
 var $5157 = 0, $5158 = 0, $5159 = 0, $516 = 0, $5160 = 0, $5161 = 0, $5162 = 0, $5163 = 0, $5164 = 0, $5165 = 0, $5166 = 0, $5167 = 0, $5168 = 0, $5169 = 0, $517 = 0, $5170 = 0, $5171 = 0, $5172 = 0, $5173 = 0, $5174 = 0;
 var $5175 = 0, $5176 = 0, $5177 = 0, $5178 = 0, $5179 = 0, $518 = 0, $5180 = 0, $5181 = 0, $5182 = 0, $5183 = 0, $5184 = 0, $5185 = 0, $5186 = 0, $5187 = 0, $5188 = 0, $5189 = 0, $519 = 0, $5190 = 0, $5191 = 0, $5192 = 0;
 var $5193 = 0, $5194 = 0, $5195 = 0, $5196 = 0, $5197 = 0, $5198 = 0, $5199 = 0, $52 = 0, $520 = 0, $5200 = 0, $5201 = 0, $5202 = 0, $5203 = 0, $5204 = 0, $5205 = 0, $5206 = 0, $5207 = 0, $5208 = 0, $5209 = 0, $521 = 0;
 var $5210 = 0, $5211 = 0, $5212 = 0, $5213 = 0, $5214 = 0, $5215 = 0, $5216 = 0, $5217 = 0, $5218 = 0, $5219 = 0, $522 = 0, $5220 = 0, $5221 = 0, $5222 = 0, $5223 = 0, $5224 = 0, $5225 = 0, $5226 = 0, $5227 = 0, $5228 = 0;
 var $5229 = 0, $523 = 0, $5230 = 0, $5231 = 0, $5232 = 0, $5233 = 0, $5234 = 0, $5235 = 0, $5236 = 0, $5237 = 0, $5238 = 0, $5239 = 0, $524 = 0, $5240 = 0, $5241 = 0, $5242 = 0, $5243 = 0, $5244 = 0, $5245 = 0, $5246 = 0;
 var $5247 = 0, $5248 = 0, $5249 = 0, $525 = 0, $5250 = 0, $5251 = 0, $5252 = 0, $5253 = 0, $5254 = 0, $5255 = 0, $5256 = 0, $5257 = 0, $5258 = 0, $5259 = 0, $526 = 0, $5260 = 0, $5261 = 0, $5262 = 0, $5263 = 0, $5264 = 0;
 var $5265 = 0, $5266 = 0, $5267 = 0, $5268 = 0, $5269 = 0, $527 = 0, $5270 = 0, $5271 = 0, $5272 = 0, $5273 = 0, $5274 = 0, $5275 = 0, $5276 = 0, $5277 = 0, $5278 = 0, $5279 = 0, $528 = 0, $5280 = 0, $5281 = 0, $5282 = 0;
 var $5283 = 0, $5284 = 0, $5285 = 0, $5286 = 0, $5287 = 0, $5288 = 0, $5289 = 0, $529 = 0, $5290 = 0, $5291 = 0, $5292 = 0, $5293 = 0, $5294 = 0, $5295 = 0, $5296 = 0, $5297 = 0, $5298 = 0, $5299 = 0, $53 = 0, $530 = 0;
 var $5300 = 0, $5301 = 0, $5302 = 0, $5303 = 0, $5304 = 0, $5305 = 0, $5306 = 0, $5307 = 0, $5308 = 0, $5309 = 0, $531 = 0, $5310 = 0, $5311 = 0, $5312 = 0, $5313 = 0, $5314 = 0, $5315 = 0, $5316 = 0, $5317 = 0, $5318 = 0;
 var $5319 = 0, $532 = 0, $5320 = 0, $5321 = 0, $5322 = 0, $5323 = 0, $5324 = 0, $5325 = 0, $5326 = 0, $5327 = 0, $5328 = 0, $5329 = 0, $533 = 0, $5330 = 0, $5331 = 0, $5332 = 0, $5333 = 0, $5334 = 0, $5335 = 0, $5336 = 0;
 var $5337 = 0, $5338 = 0, $5339 = 0, $534 = 0, $5340 = 0, $5341 = 0, $5342 = 0, $5343 = 0, $5344 = 0, $5345 = 0, $5346 = 0, $5347 = 0, $5348 = 0, $5349 = 0, $535 = 0, $5350 = 0, $5351 = 0, $5352 = 0, $5353 = 0, $5354 = 0;
 var $5355 = 0, $5356 = 0, $5357 = 0, $5358 = 0, $5359 = 0, $536 = 0, $5360 = 0, $5361 = 0, $5362 = 0, $5363 = 0, $5364 = 0, $5365 = 0, $5366 = 0, $5367 = 0, $5368 = 0, $5369 = 0, $537 = 0, $5370 = 0, $5371 = 0, $5372 = 0;
 var $5373 = 0, $5374 = 0, $5375 = 0, $5376 = 0, $5377 = 0, $5378 = 0, $5379 = 0, $538 = 0, $5380 = 0, $5381 = 0, $5382 = 0, $5383 = 0, $5384 = 0, $5385 = 0, $5386 = 0, $5387 = 0, $5388 = 0, $5389 = 0, $539 = 0, $5390 = 0;
 var $5391 = 0, $5392 = 0, $5393 = 0, $5394 = 0, $5395 = 0, $5396 = 0, $5397 = 0, $5398 = 0, $5399 = 0, $54 = 0, $540 = 0, $5400 = 0, $5401 = 0, $5402 = 0, $5403 = 0, $5404 = 0, $5405 = 0, $5406 = 0, $5407 = 0, $5408 = 0;
 var $5409 = 0, $541 = 0, $5410 = 0, $5411 = 0, $5412 = 0, $5413 = 0, $5414 = 0, $5415 = 0, $5416 = 0, $5417 = 0, $5418 = 0, $5419 = 0, $542 = 0, $5420 = 0, $5421 = 0, $5422 = 0, $5423 = 0, $5424 = 0, $5425 = 0, $5426 = 0;
 var $5427 = 0, $5428 = 0, $5429 = 0, $543 = 0, $5430 = 0, $5431 = 0, $5432 = 0, $5433 = 0, $5434 = 0, $5435 = 0, $5436 = 0, $5437 = 0, $5438 = 0, $5439 = 0, $544 = 0, $5440 = 0, $5441 = 0, $5442 = 0, $5443 = 0, $5444 = 0;
 var $5445 = 0, $5446 = 0, $5447 = 0, $5448 = 0, $5449 = 0, $545 = 0, $5450 = 0, $5451 = 0, $5452 = 0, $5453 = 0, $5454 = 0, $5455 = 0, $5456 = 0, $5457 = 0, $5458 = 0, $5459 = 0, $546 = 0, $5460 = 0, $5461 = 0, $5462 = 0;
 var $5463 = 0, $5464 = 0, $5465 = 0, $5466 = 0, $5467 = 0, $5468 = 0, $5469 = 0, $547 = 0, $5470 = 0, $5471 = 0, $5472 = 0, $5473 = 0, $5474 = 0, $5475 = 0, $5476 = 0, $5477 = 0, $5478 = 0, $5479 = 0, $548 = 0, $5480 = 0;
 var $5481 = 0, $5482 = 0, $5483 = 0, $5484 = 0, $5485 = 0, $5486 = 0, $5487 = 0, $5488 = 0, $5489 = 0, $549 = 0, $5490 = 0, $5491 = 0, $5492 = 0, $5493 = 0, $5494 = 0, $5495 = 0, $5496 = 0, $5497 = 0, $5498 = 0, $5499 = 0;
 var $55 = 0, $550 = 0, $5500 = 0, $5501 = 0, $5502 = 0, $5503 = 0, $5504 = 0, $5505 = 0, $5506 = 0, $5507 = 0, $5508 = 0, $5509 = 0, $551 = 0, $5510 = 0, $5511 = 0, $5512 = 0, $5513 = 0, $5514 = 0, $5515 = 0, $5516 = 0;
 var $5517 = 0, $5518 = 0, $5519 = 0, $552 = 0, $5520 = 0, $5521 = 0, $5522 = 0, $5523 = 0, $5524 = 0, $5525 = 0, $5526 = 0, $5527 = 0, $5528 = 0, $5529 = 0, $553 = 0, $5530 = 0, $5531 = 0, $5532 = 0, $5533 = 0, $5534 = 0;
 var $5535 = 0, $5536 = 0, $5537 = 0, $5538 = 0, $5539 = 0, $554 = 0, $5540 = 0, $5541 = 0, $5542 = 0, $5543 = 0, $5544 = 0, $5545 = 0, $5546 = 0, $5547 = 0, $5548 = 0, $5549 = 0, $555 = 0, $5550 = 0, $5551 = 0, $5552 = 0;
 var $5553 = 0, $5554 = 0, $5555 = 0, $5556 = 0, $5557 = 0, $5558 = 0, $5559 = 0, $556 = 0, $5560 = 0, $5561 = 0, $5562 = 0, $5563 = 0, $5564 = 0, $5565 = 0, $5566 = 0, $5567 = 0, $5568 = 0, $5569 = 0, $557 = 0, $5570 = 0;
 var $5571 = 0, $5572 = 0, $5573 = 0, $5574 = 0, $5575 = 0, $5576 = 0, $5577 = 0, $5578 = 0, $5579 = 0, $558 = 0, $5580 = 0, $5581 = 0, $5582 = 0, $5583 = 0, $5584 = 0, $5585 = 0, $5586 = 0, $5587 = 0, $5588 = 0, $5589 = 0;
 var $559 = 0, $5590 = 0, $5591 = 0, $5592 = 0, $5593 = 0, $5594 = 0, $5595 = 0, $5596 = 0, $5597 = 0, $5598 = 0, $5599 = 0, $56 = 0, $560 = 0, $5600 = 0, $5601 = 0, $5602 = 0, $5603 = 0, $5604 = 0, $5605 = 0, $5606 = 0;
 var $5607 = 0, $5608 = 0, $5609 = 0, $561 = 0, $5610 = 0, $5611 = 0, $5612 = 0, $5613 = 0, $5614 = 0, $5615 = 0, $5616 = 0, $5617 = 0, $5618 = 0, $5619 = 0, $562 = 0, $5620 = 0, $5621 = 0, $5622 = 0, $5623 = 0, $5624 = 0;
 var $5625 = 0, $5626 = 0, $5627 = 0, $5628 = 0, $5629 = 0, $563 = 0, $5630 = 0, $5631 = 0, $5632 = 0, $5633 = 0, $5634 = 0, $5635 = 0, $5636 = 0, $5637 = 0, $5638 = 0, $5639 = 0, $564 = 0, $5640 = 0, $5641 = 0, $5642 = 0;
 var $5643 = 0, $5644 = 0, $5645 = 0, $5646 = 0, $5647 = 0, $5648 = 0, $5649 = 0, $565 = 0, $5650 = 0, $5651 = 0, $5652 = 0, $5653 = 0, $5654 = 0, $5655 = 0, $5656 = 0, $5657 = 0, $5658 = 0, $5659 = 0, $566 = 0, $5660 = 0;
 var $5661 = 0, $5662 = 0, $5663 = 0, $5664 = 0, $5665 = 0, $5666 = 0, $5667 = 0, $5668 = 0, $5669 = 0, $567 = 0, $5670 = 0, $5671 = 0, $5672 = 0, $5673 = 0, $5674 = 0, $5675 = 0, $5676 = 0, $5677 = 0, $5678 = 0, $5679 = 0;
 var $568 = 0, $5680 = 0, $5681 = 0, $5682 = 0, $5683 = 0, $5684 = 0, $5685 = 0, $5686 = 0, $5687 = 0, $5688 = 0, $5689 = 0, $569 = 0, $5690 = 0, $5691 = 0, $5692 = 0, $5693 = 0, $5694 = 0, $5695 = 0, $5696 = 0, $5697 = 0;
 var $5698 = 0, $5699 = 0, $57 = 0, $570 = 0, $5700 = 0, $5701 = 0, $5702 = 0, $5703 = 0, $5704 = 0, $5705 = 0, $5706 = 0, $5707 = 0, $5708 = 0, $5709 = 0, $571 = 0, $5710 = 0, $5711 = 0, $5712 = 0, $5713 = 0, $5714 = 0;
 var $5715 = 0, $5716 = 0, $5717 = 0, $5718 = 0, $5719 = 0, $572 = 0, $5720 = 0, $5721 = 0, $5722 = 0, $5723 = 0, $5724 = 0, $5725 = 0, $5726 = 0, $5727 = 0, $5728 = 0, $5729 = 0, $573 = 0, $5730 = 0, $5731 = 0, $5732 = 0;
 var $5733 = 0, $5734 = 0, $5735 = 0, $5736 = 0, $5737 = 0, $5738 = 0, $5739 = 0, $574 = 0, $5740 = 0, $5741 = 0, $5742 = 0, $5743 = 0, $5744 = 0, $5745 = 0, $5746 = 0, $5747 = 0, $5748 = 0, $5749 = 0, $575 = 0, $5750 = 0;
 var $5751 = 0, $5752 = 0, $5753 = 0, $5754 = 0, $5755 = 0, $5756 = 0, $5757 = 0, $5758 = 0, $5759 = 0, $576 = 0, $5760 = 0, $5761 = 0, $5762 = 0, $5763 = 0, $5764 = 0, $5765 = 0, $5766 = 0, $5767 = 0, $5768 = 0, $5769 = 0;
 var $577 = 0, $5770 = 0, $5771 = 0, $5772 = 0, $5773 = 0, $5774 = 0, $5775 = 0, $5776 = 0, $5777 = 0, $5778 = 0, $5779 = 0, $578 = 0, $5780 = 0, $5781 = 0, $5782 = 0, $5783 = 0, $5784 = 0, $5785 = 0, $5786 = 0, $5787 = 0;
 var $5788 = 0, $5789 = 0, $579 = 0, $5790 = 0, $5791 = 0, $5792 = 0, $5793 = 0, $5794 = 0, $5795 = 0, $5796 = 0, $5797 = 0, $5798 = 0, $5799 = 0, $58 = 0, $580 = 0, $5800 = 0, $5801 = 0, $5802 = 0, $5803 = 0, $5804 = 0;
 var $5805 = 0, $5806 = 0, $5807 = 0, $5808 = 0, $5809 = 0, $581 = 0, $5810 = 0, $5811 = 0, $5812 = 0, $5813 = 0, $5814 = 0, $5815 = 0, $5816 = 0, $5817 = 0, $5818 = 0, $5819 = 0, $582 = 0, $5820 = 0, $5821 = 0, $5822 = 0;
 var $5823 = 0, $5824 = 0, $5825 = 0, $5826 = 0, $5827 = 0, $5828 = 0, $5829 = 0, $583 = 0, $5830 = 0, $5831 = 0, $5832 = 0, $5833 = 0, $5834 = 0, $5835 = 0, $5836 = 0, $5837 = 0, $5838 = 0, $5839 = 0, $584 = 0, $5840 = 0;
 var $5841 = 0, $5842 = 0, $5843 = 0, $5844 = 0, $5845 = 0, $5846 = 0, $5847 = 0, $5848 = 0, $5849 = 0, $585 = 0, $5850 = 0, $5851 = 0, $5852 = 0, $5853 = 0, $5854 = 0, $5855 = 0, $5856 = 0, $5857 = 0, $5858 = 0, $5859 = 0;
 var $586 = 0, $5860 = 0, $5861 = 0, $5862 = 0, $5863 = 0, $5864 = 0, $5865 = 0, $5866 = 0, $5867 = 0, $5868 = 0, $5869 = 0, $587 = 0, $5870 = 0, $5871 = 0, $5872 = 0, $5873 = 0, $5874 = 0, $5875 = 0, $5876 = 0, $5877 = 0;
 var $5878 = 0, $5879 = 0, $588 = 0, $5880 = 0, $5881 = 0, $5882 = 0, $5883 = 0, $5884 = 0, $5885 = 0, $5886 = 0, $5887 = 0, $5888 = 0, $5889 = 0, $589 = 0, $5890 = 0, $5891 = 0, $5892 = 0, $5893 = 0, $5894 = 0, $5895 = 0;
 var $5896 = 0, $5897 = 0, $5898 = 0, $5899 = 0, $59 = 0, $590 = 0, $5900 = 0, $5901 = 0, $5902 = 0, $5903 = 0, $5904 = 0, $5905 = 0, $5906 = 0, $5907 = 0, $5908 = 0, $5909 = 0, $591 = 0, $5910 = 0, $5911 = 0, $5912 = 0;
 var $5913 = 0, $5914 = 0, $5915 = 0, $5916 = 0, $5917 = 0, $5918 = 0, $5919 = 0, $592 = 0, $5920 = 0, $5921 = 0, $5922 = 0, $5923 = 0, $5924 = 0, $5925 = 0, $5926 = 0, $5927 = 0, $5928 = 0, $5929 = 0, $593 = 0, $5930 = 0;
 var $5931 = 0, $5932 = 0, $5933 = 0, $5934 = 0, $5935 = 0, $5936 = 0, $5937 = 0, $5938 = 0, $5939 = 0, $594 = 0, $5940 = 0, $5941 = 0, $5942 = 0, $5943 = 0, $5944 = 0, $5945 = 0, $5946 = 0, $5947 = 0, $5948 = 0, $5949 = 0;
 var $595 = 0, $5950 = 0, $5951 = 0, $5952 = 0, $5953 = 0, $5954 = 0, $5955 = 0, $5956 = 0, $5957 = 0, $5958 = 0, $5959 = 0, $596 = 0, $5960 = 0, $5961 = 0, $5962 = 0, $5963 = 0, $5964 = 0, $5965 = 0, $5966 = 0, $5967 = 0;
 var $5968 = 0, $5969 = 0, $597 = 0, $5970 = 0, $5971 = 0, $5972 = 0, $5973 = 0, $5974 = 0, $5975 = 0, $5976 = 0, $5977 = 0, $5978 = 0, $5979 = 0, $598 = 0, $5980 = 0, $5981 = 0, $5982 = 0, $5983 = 0, $5984 = 0, $5985 = 0;
 var $5986 = 0, $5987 = 0, $5988 = 0, $5989 = 0, $599 = 0, $5990 = 0, $5991 = 0, $5992 = 0, $5993 = 0, $5994 = 0, $5995 = 0, $5996 = 0, $5997 = 0, $5998 = 0, $5999 = 0, $6 = 0, $60 = 0, $600 = 0, $6000 = 0, $6001 = 0;
 var $6002 = 0, $6003 = 0, $6004 = 0, $6005 = 0, $6006 = 0, $6007 = 0, $6008 = 0, $6009 = 0, $601 = 0, $6010 = 0, $6011 = 0, $6012 = 0, $6013 = 0, $6014 = 0, $6015 = 0, $6016 = 0, $6017 = 0, $6018 = 0, $6019 = 0, $602 = 0;
 var $6020 = 0, $6021 = 0, $6022 = 0, $6023 = 0, $6024 = 0, $6025 = 0, $6026 = 0, $6027 = 0, $6028 = 0, $6029 = 0, $603 = 0, $6030 = 0, $6031 = 0, $6032 = 0, $6033 = 0, $6034 = 0, $6035 = 0, $6036 = 0, $6037 = 0, $6038 = 0;
 var $6039 = 0, $604 = 0, $6040 = 0, $6041 = 0, $6042 = 0, $6043 = 0, $6044 = 0, $6045 = 0, $6046 = 0, $6047 = 0, $6048 = 0, $6049 = 0, $605 = 0, $6050 = 0, $6051 = 0, $6052 = 0, $6053 = 0, $6054 = 0, $6055 = 0, $6056 = 0;
 var $6057 = 0, $6058 = 0, $6059 = 0, $606 = 0, $6060 = 0, $6061 = 0, $6062 = 0, $6063 = 0, $6064 = 0, $6065 = 0, $6066 = 0, $6067 = 0, $6068 = 0, $6069 = 0, $607 = 0, $6070 = 0, $6071 = 0, $6072 = 0, $6073 = 0, $6074 = 0;
 var $6075 = 0, $6076 = 0, $6077 = 0, $6078 = 0, $6079 = 0, $608 = 0, $6080 = 0, $6081 = 0, $6082 = 0, $6083 = 0, $6084 = 0, $6085 = 0, $6086 = 0, $6087 = 0, $6088 = 0, $6089 = 0, $609 = 0, $6090 = 0, $6091 = 0, $6092 = 0;
 var $6093 = 0, $6094 = 0, $6095 = 0, $6096 = 0, $6097 = 0, $6098 = 0, $6099 = 0, $61 = 0, $610 = 0, $6100 = 0, $6101 = 0, $6102 = 0, $6103 = 0, $6104 = 0, $6105 = 0, $6106 = 0, $6107 = 0, $6108 = 0, $6109 = 0, $611 = 0;
 var $6110 = 0, $6111 = 0, $6112 = 0, $6113 = 0, $6114 = 0, $6115 = 0, $6116 = 0, $6117 = 0, $6118 = 0, $6119 = 0, $612 = 0, $6120 = 0, $6121 = 0, $6122 = 0, $6123 = 0, $6124 = 0, $6125 = 0, $6126 = 0, $6127 = 0, $6128 = 0;
 var $6129 = 0, $613 = 0, $6130 = 0, $6131 = 0, $6132 = 0, $6133 = 0, $6134 = 0, $6135 = 0, $6136 = 0, $6137 = 0, $6138 = 0, $6139 = 0, $614 = 0, $6140 = 0, $6141 = 0, $6142 = 0, $6143 = 0, $6144 = 0, $6145 = 0, $6146 = 0;
 var $6147 = 0, $6148 = 0, $6149 = 0, $615 = 0, $6150 = 0, $6151 = 0, $6152 = 0, $6153 = 0, $6154 = 0, $6155 = 0, $6156 = 0, $6157 = 0, $6158 = 0, $6159 = 0, $616 = 0, $6160 = 0, $6161 = 0, $6162 = 0, $6163 = 0, $6164 = 0;
 var $6165 = 0, $6166 = 0, $6167 = 0, $6168 = 0, $6169 = 0, $617 = 0, $6170 = 0, $6171 = 0, $6172 = 0, $6173 = 0, $6174 = 0, $6175 = 0, $6176 = 0, $6177 = 0, $6178 = 0, $6179 = 0, $618 = 0, $6180 = 0, $6181 = 0, $6182 = 0;
 var $6183 = 0, $6184 = 0, $6185 = 0, $6186 = 0, $6187 = 0, $6188 = 0, $6189 = 0, $619 = 0, $6190 = 0, $6191 = 0, $6192 = 0, $6193 = 0, $6194 = 0, $6195 = 0, $6196 = 0, $6197 = 0, $6198 = 0, $6199 = 0, $62 = 0, $620 = 0;
 var $6200 = 0, $6201 = 0, $6202 = 0, $6203 = 0, $6204 = 0, $6205 = 0, $6206 = 0, $6207 = 0, $6208 = 0, $6209 = 0, $621 = 0, $6210 = 0, $6211 = 0, $6212 = 0, $6213 = 0, $6214 = 0, $6215 = 0, $6216 = 0, $6217 = 0, $6218 = 0;
 var $6219 = 0, $622 = 0, $6220 = 0, $6221 = 0, $6222 = 0, $6223 = 0, $6224 = 0, $6225 = 0, $6226 = 0, $6227 = 0, $6228 = 0, $6229 = 0, $623 = 0, $6230 = 0, $6231 = 0, $6232 = 0, $6233 = 0, $6234 = 0, $6235 = 0, $6236 = 0;
 var $6237 = 0, $6238 = 0, $6239 = 0, $624 = 0, $6240 = 0, $6241 = 0, $6242 = 0, $6243 = 0, $6244 = 0, $6245 = 0, $6246 = 0, $6247 = 0, $6248 = 0, $6249 = 0, $625 = 0, $6250 = 0, $6251 = 0, $6252 = 0, $6253 = 0, $6254 = 0;
 var $6255 = 0, $6256 = 0, $6257 = 0, $6258 = 0, $6259 = 0, $626 = 0, $6260 = 0, $6261 = 0, $6262 = 0, $6263 = 0, $6264 = 0, $6265 = 0, $6266 = 0, $6267 = 0, $6268 = 0, $6269 = 0, $627 = 0, $6270 = 0, $6271 = 0, $6272 = 0;
 var $6273 = 0, $6274 = 0, $6275 = 0, $6276 = 0, $6277 = 0, $6278 = 0, $6279 = 0, $628 = 0, $6280 = 0, $6281 = 0, $6282 = 0, $6283 = 0, $6284 = 0, $6285 = 0, $6286 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0;
 var $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0;
 var $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0;
 var $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0;
 var $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0;
 var $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0;
 var $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0;
 var $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0;
 var $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0;
 var $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0;
 var $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0;
 var $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0;
 var $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0;
 var $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0;
 var $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0;
 var $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0;
 var $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0;
 var $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0;
 var $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0;
 var $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0;
 var $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0;
 var $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $__nd$0$i$i1008 = 0, $__nd$0$i$i1008$lcssa = 0, $__nd$0$i$i1008$lcssa7329 = 0, $__nd$0$i$i1008$lcssa7330 = 0, $__nd$0$i$i1023 = 0, $__nd$0$i$i1023$lcssa = 0, $__nd$0$i$i1023$lcssa7325 = 0, $__nd$0$i$i1023$lcssa7326 = 0, $__nd$0$i$i1038 = 0, $__nd$0$i$i1038$lcssa = 0, $__nd$0$i$i1038$lcssa7321 = 0, $__nd$0$i$i1038$lcssa7322 = 0, $__nd$0$i$i1053 = 0, $__nd$0$i$i1053$lcssa = 0;
 var $__nd$0$i$i1053$lcssa7317 = 0, $__nd$0$i$i1053$lcssa7318 = 0, $__nd$0$i$i1068 = 0, $__nd$0$i$i1068$lcssa = 0, $__nd$0$i$i1068$lcssa7313 = 0, $__nd$0$i$i1068$lcssa7314 = 0, $__nd$0$i$i108 = 0, $__nd$0$i$i108$lcssa = 0, $__nd$0$i$i108$lcssa7569 = 0, $__nd$0$i$i108$lcssa7570 = 0, $__nd$0$i$i1083 = 0, $__nd$0$i$i1083$lcssa = 0, $__nd$0$i$i1083$lcssa7309 = 0, $__nd$0$i$i1083$lcssa7310 = 0, $__nd$0$i$i1098 = 0, $__nd$0$i$i1098$lcssa = 0, $__nd$0$i$i1098$lcssa7305 = 0, $__nd$0$i$i1098$lcssa7306 = 0, $__nd$0$i$i1113 = 0, $__nd$0$i$i1113$lcssa = 0;
 var $__nd$0$i$i1113$lcssa7301 = 0, $__nd$0$i$i1113$lcssa7302 = 0, $__nd$0$i$i1128 = 0, $__nd$0$i$i1128$lcssa = 0, $__nd$0$i$i1128$lcssa7297 = 0, $__nd$0$i$i1128$lcssa7298 = 0, $__nd$0$i$i1143 = 0, $__nd$0$i$i1143$lcssa = 0, $__nd$0$i$i1143$lcssa7293 = 0, $__nd$0$i$i1143$lcssa7294 = 0, $__nd$0$i$i1158 = 0, $__nd$0$i$i1158$lcssa = 0, $__nd$0$i$i1158$lcssa7289 = 0, $__nd$0$i$i1158$lcssa7290 = 0, $__nd$0$i$i1173 = 0, $__nd$0$i$i1173$lcssa = 0, $__nd$0$i$i1173$lcssa7285 = 0, $__nd$0$i$i1173$lcssa7286 = 0, $__nd$0$i$i1188 = 0, $__nd$0$i$i1188$lcssa = 0;
 var $__nd$0$i$i1188$lcssa7281 = 0, $__nd$0$i$i1188$lcssa7282 = 0, $__nd$0$i$i1203 = 0, $__nd$0$i$i1203$lcssa = 0, $__nd$0$i$i1203$lcssa7277 = 0, $__nd$0$i$i1203$lcssa7278 = 0, $__nd$0$i$i1218 = 0, $__nd$0$i$i1218$lcssa = 0, $__nd$0$i$i1218$lcssa7273 = 0, $__nd$0$i$i1218$lcssa7274 = 0, $__nd$0$i$i123 = 0, $__nd$0$i$i123$lcssa = 0, $__nd$0$i$i123$lcssa7565 = 0, $__nd$0$i$i123$lcssa7566 = 0, $__nd$0$i$i1233 = 0, $__nd$0$i$i1233$lcssa = 0, $__nd$0$i$i1233$lcssa7269 = 0, $__nd$0$i$i1233$lcssa7270 = 0, $__nd$0$i$i1248 = 0, $__nd$0$i$i1248$lcssa = 0;
 var $__nd$0$i$i1248$lcssa7265 = 0, $__nd$0$i$i1248$lcssa7266 = 0, $__nd$0$i$i1263 = 0, $__nd$0$i$i1263$lcssa = 0, $__nd$0$i$i1263$lcssa7261 = 0, $__nd$0$i$i1263$lcssa7262 = 0, $__nd$0$i$i1278 = 0, $__nd$0$i$i1278$lcssa = 0, $__nd$0$i$i1278$lcssa7257 = 0, $__nd$0$i$i1278$lcssa7258 = 0, $__nd$0$i$i1293 = 0, $__nd$0$i$i1293$lcssa = 0, $__nd$0$i$i1293$lcssa7253 = 0, $__nd$0$i$i1293$lcssa7254 = 0, $__nd$0$i$i1308 = 0, $__nd$0$i$i1308$lcssa = 0, $__nd$0$i$i1308$lcssa7249 = 0, $__nd$0$i$i1308$lcssa7250 = 0, $__nd$0$i$i1323 = 0, $__nd$0$i$i1323$lcssa = 0;
 var $__nd$0$i$i1323$lcssa7245 = 0, $__nd$0$i$i1323$lcssa7246 = 0, $__nd$0$i$i1338 = 0, $__nd$0$i$i1338$lcssa = 0, $__nd$0$i$i1338$lcssa7241 = 0, $__nd$0$i$i1338$lcssa7242 = 0, $__nd$0$i$i1353 = 0, $__nd$0$i$i1353$lcssa = 0, $__nd$0$i$i1353$lcssa7237 = 0, $__nd$0$i$i1353$lcssa7238 = 0, $__nd$0$i$i1368 = 0, $__nd$0$i$i1368$lcssa = 0, $__nd$0$i$i1368$lcssa7233 = 0, $__nd$0$i$i1368$lcssa7234 = 0, $__nd$0$i$i138 = 0, $__nd$0$i$i138$lcssa = 0, $__nd$0$i$i138$lcssa7561 = 0, $__nd$0$i$i138$lcssa7562 = 0, $__nd$0$i$i1383 = 0, $__nd$0$i$i1383$lcssa = 0;
 var $__nd$0$i$i1383$lcssa7229 = 0, $__nd$0$i$i1383$lcssa7230 = 0, $__nd$0$i$i1398 = 0, $__nd$0$i$i1398$lcssa = 0, $__nd$0$i$i1398$lcssa7225 = 0, $__nd$0$i$i1398$lcssa7226 = 0, $__nd$0$i$i1413 = 0, $__nd$0$i$i1413$lcssa = 0, $__nd$0$i$i1413$lcssa7221 = 0, $__nd$0$i$i1413$lcssa7222 = 0, $__nd$0$i$i1428 = 0, $__nd$0$i$i1428$lcssa = 0, $__nd$0$i$i1428$lcssa7217 = 0, $__nd$0$i$i1428$lcssa7218 = 0, $__nd$0$i$i1443 = 0, $__nd$0$i$i1443$lcssa = 0, $__nd$0$i$i1443$lcssa7213 = 0, $__nd$0$i$i1443$lcssa7214 = 0, $__nd$0$i$i1458 = 0, $__nd$0$i$i1458$lcssa = 0;
 var $__nd$0$i$i1458$lcssa7209 = 0, $__nd$0$i$i1458$lcssa7210 = 0, $__nd$0$i$i1473 = 0, $__nd$0$i$i1473$lcssa = 0, $__nd$0$i$i1473$lcssa7205 = 0, $__nd$0$i$i1473$lcssa7206 = 0, $__nd$0$i$i1488 = 0, $__nd$0$i$i1488$lcssa = 0, $__nd$0$i$i1488$lcssa7201 = 0, $__nd$0$i$i1488$lcssa7202 = 0, $__nd$0$i$i1503 = 0, $__nd$0$i$i1503$lcssa = 0, $__nd$0$i$i1503$lcssa7197 = 0, $__nd$0$i$i1503$lcssa7198 = 0, $__nd$0$i$i1518 = 0, $__nd$0$i$i1518$lcssa = 0, $__nd$0$i$i1518$lcssa7193 = 0, $__nd$0$i$i1518$lcssa7194 = 0, $__nd$0$i$i153 = 0, $__nd$0$i$i153$lcssa = 0;
 var $__nd$0$i$i153$lcssa7557 = 0, $__nd$0$i$i153$lcssa7558 = 0, $__nd$0$i$i1533 = 0, $__nd$0$i$i1533$lcssa = 0, $__nd$0$i$i1533$lcssa7189 = 0, $__nd$0$i$i1533$lcssa7190 = 0, $__nd$0$i$i1548 = 0, $__nd$0$i$i1548$lcssa = 0, $__nd$0$i$i1548$lcssa7185 = 0, $__nd$0$i$i1548$lcssa7186 = 0, $__nd$0$i$i1563 = 0, $__nd$0$i$i1563$lcssa = 0, $__nd$0$i$i1563$lcssa7181 = 0, $__nd$0$i$i1563$lcssa7182 = 0, $__nd$0$i$i1578 = 0, $__nd$0$i$i1578$lcssa = 0, $__nd$0$i$i1578$lcssa7177 = 0, $__nd$0$i$i1578$lcssa7178 = 0, $__nd$0$i$i1593 = 0, $__nd$0$i$i1593$lcssa = 0;
 var $__nd$0$i$i1593$lcssa7173 = 0, $__nd$0$i$i1593$lcssa7174 = 0, $__nd$0$i$i1608 = 0, $__nd$0$i$i1608$lcssa = 0, $__nd$0$i$i1608$lcssa7169 = 0, $__nd$0$i$i1608$lcssa7170 = 0, $__nd$0$i$i1623 = 0, $__nd$0$i$i1623$lcssa = 0, $__nd$0$i$i1623$lcssa7165 = 0, $__nd$0$i$i1623$lcssa7166 = 0, $__nd$0$i$i1638 = 0, $__nd$0$i$i1638$lcssa = 0, $__nd$0$i$i1638$lcssa7161 = 0, $__nd$0$i$i1638$lcssa7162 = 0, $__nd$0$i$i1653 = 0, $__nd$0$i$i1653$lcssa = 0, $__nd$0$i$i1653$lcssa7157 = 0, $__nd$0$i$i1653$lcssa7158 = 0, $__nd$0$i$i1668 = 0, $__nd$0$i$i1668$lcssa = 0;
 var $__nd$0$i$i1668$lcssa7153 = 0, $__nd$0$i$i1668$lcssa7154 = 0, $__nd$0$i$i168 = 0, $__nd$0$i$i168$lcssa = 0, $__nd$0$i$i168$lcssa7553 = 0, $__nd$0$i$i168$lcssa7554 = 0, $__nd$0$i$i1683 = 0, $__nd$0$i$i1683$lcssa = 0, $__nd$0$i$i1683$lcssa7149 = 0, $__nd$0$i$i1683$lcssa7150 = 0, $__nd$0$i$i1698 = 0, $__nd$0$i$i1698$lcssa = 0, $__nd$0$i$i1698$lcssa7145 = 0, $__nd$0$i$i1698$lcssa7146 = 0, $__nd$0$i$i1713 = 0, $__nd$0$i$i1713$lcssa = 0, $__nd$0$i$i1713$lcssa7141 = 0, $__nd$0$i$i1713$lcssa7142 = 0, $__nd$0$i$i1728 = 0, $__nd$0$i$i1728$lcssa = 0;
 var $__nd$0$i$i1728$lcssa7137 = 0, $__nd$0$i$i1728$lcssa7138 = 0, $__nd$0$i$i1743 = 0, $__nd$0$i$i1743$lcssa = 0, $__nd$0$i$i1743$lcssa7133 = 0, $__nd$0$i$i1743$lcssa7134 = 0, $__nd$0$i$i1758 = 0, $__nd$0$i$i1758$lcssa = 0, $__nd$0$i$i1758$lcssa7129 = 0, $__nd$0$i$i1758$lcssa7130 = 0, $__nd$0$i$i1773 = 0, $__nd$0$i$i1773$lcssa = 0, $__nd$0$i$i1773$lcssa7125 = 0, $__nd$0$i$i1773$lcssa7126 = 0, $__nd$0$i$i1788 = 0, $__nd$0$i$i1788$lcssa = 0, $__nd$0$i$i1788$lcssa7121 = 0, $__nd$0$i$i1788$lcssa7122 = 0, $__nd$0$i$i18 = 0, $__nd$0$i$i18$lcssa = 0;
 var $__nd$0$i$i18$lcssa7593 = 0, $__nd$0$i$i18$lcssa7594 = 0, $__nd$0$i$i1803 = 0, $__nd$0$i$i1803$lcssa = 0, $__nd$0$i$i1803$lcssa7117 = 0, $__nd$0$i$i1803$lcssa7118 = 0, $__nd$0$i$i1818 = 0, $__nd$0$i$i1818$lcssa = 0, $__nd$0$i$i1818$lcssa7113 = 0, $__nd$0$i$i1818$lcssa7114 = 0, $__nd$0$i$i183 = 0, $__nd$0$i$i183$lcssa = 0, $__nd$0$i$i183$lcssa7549 = 0, $__nd$0$i$i183$lcssa7550 = 0, $__nd$0$i$i1833 = 0, $__nd$0$i$i1833$lcssa = 0, $__nd$0$i$i1833$lcssa7109 = 0, $__nd$0$i$i1833$lcssa7110 = 0, $__nd$0$i$i1848 = 0, $__nd$0$i$i1848$lcssa = 0;
 var $__nd$0$i$i1848$lcssa7105 = 0, $__nd$0$i$i1848$lcssa7106 = 0, $__nd$0$i$i1863 = 0, $__nd$0$i$i1863$lcssa = 0, $__nd$0$i$i1863$lcssa7101 = 0, $__nd$0$i$i1863$lcssa7102 = 0, $__nd$0$i$i1878 = 0, $__nd$0$i$i1878$lcssa = 0, $__nd$0$i$i1878$lcssa7097 = 0, $__nd$0$i$i1878$lcssa7098 = 0, $__nd$0$i$i1893 = 0, $__nd$0$i$i1893$lcssa = 0, $__nd$0$i$i1893$lcssa7093 = 0, $__nd$0$i$i1893$lcssa7094 = 0, $__nd$0$i$i1908 = 0, $__nd$0$i$i1908$lcssa = 0, $__nd$0$i$i1908$lcssa7089 = 0, $__nd$0$i$i1908$lcssa7090 = 0, $__nd$0$i$i1923 = 0, $__nd$0$i$i1923$lcssa = 0;
 var $__nd$0$i$i1923$lcssa7085 = 0, $__nd$0$i$i1923$lcssa7086 = 0, $__nd$0$i$i1938 = 0, $__nd$0$i$i1938$lcssa = 0, $__nd$0$i$i1938$lcssa7081 = 0, $__nd$0$i$i1938$lcssa7082 = 0, $__nd$0$i$i1953 = 0, $__nd$0$i$i1953$lcssa = 0, $__nd$0$i$i1953$lcssa7077 = 0, $__nd$0$i$i1953$lcssa7078 = 0, $__nd$0$i$i1968 = 0, $__nd$0$i$i1968$lcssa = 0, $__nd$0$i$i1968$lcssa7073 = 0, $__nd$0$i$i1968$lcssa7074 = 0, $__nd$0$i$i198 = 0, $__nd$0$i$i198$lcssa = 0, $__nd$0$i$i198$lcssa7545 = 0, $__nd$0$i$i198$lcssa7546 = 0, $__nd$0$i$i1983 = 0, $__nd$0$i$i1983$lcssa = 0;
 var $__nd$0$i$i1983$lcssa7069 = 0, $__nd$0$i$i1983$lcssa7070 = 0, $__nd$0$i$i1998 = 0, $__nd$0$i$i1998$lcssa = 0, $__nd$0$i$i1998$lcssa7065 = 0, $__nd$0$i$i1998$lcssa7066 = 0, $__nd$0$i$i2013 = 0, $__nd$0$i$i2013$lcssa = 0, $__nd$0$i$i2013$lcssa7061 = 0, $__nd$0$i$i2013$lcssa7062 = 0, $__nd$0$i$i2028 = 0, $__nd$0$i$i2028$lcssa = 0, $__nd$0$i$i2028$lcssa7057 = 0, $__nd$0$i$i2028$lcssa7058 = 0, $__nd$0$i$i2043 = 0, $__nd$0$i$i2043$lcssa = 0, $__nd$0$i$i2043$lcssa7053 = 0, $__nd$0$i$i2043$lcssa7054 = 0, $__nd$0$i$i2058 = 0, $__nd$0$i$i2058$lcssa = 0;
 var $__nd$0$i$i2058$lcssa7049 = 0, $__nd$0$i$i2058$lcssa7050 = 0, $__nd$0$i$i2073 = 0, $__nd$0$i$i2073$lcssa = 0, $__nd$0$i$i2073$lcssa7045 = 0, $__nd$0$i$i2073$lcssa7046 = 0, $__nd$0$i$i2088 = 0, $__nd$0$i$i2088$lcssa = 0, $__nd$0$i$i2088$lcssa7041 = 0, $__nd$0$i$i2088$lcssa7042 = 0, $__nd$0$i$i2103 = 0, $__nd$0$i$i2103$lcssa = 0, $__nd$0$i$i2103$lcssa7037 = 0, $__nd$0$i$i2103$lcssa7038 = 0, $__nd$0$i$i2118 = 0, $__nd$0$i$i2118$lcssa = 0, $__nd$0$i$i2118$lcssa7033 = 0, $__nd$0$i$i2118$lcssa7034 = 0, $__nd$0$i$i213 = 0, $__nd$0$i$i213$lcssa = 0;
 var $__nd$0$i$i213$lcssa7541 = 0, $__nd$0$i$i213$lcssa7542 = 0, $__nd$0$i$i2133 = 0, $__nd$0$i$i2133$lcssa = 0, $__nd$0$i$i2133$lcssa7029 = 0, $__nd$0$i$i2133$lcssa7030 = 0, $__nd$0$i$i2148 = 0, $__nd$0$i$i2148$lcssa = 0, $__nd$0$i$i2148$lcssa7025 = 0, $__nd$0$i$i2148$lcssa7026 = 0, $__nd$0$i$i2163 = 0, $__nd$0$i$i2163$lcssa = 0, $__nd$0$i$i2163$lcssa7021 = 0, $__nd$0$i$i2163$lcssa7022 = 0, $__nd$0$i$i2178 = 0, $__nd$0$i$i2178$lcssa = 0, $__nd$0$i$i2178$lcssa7017 = 0, $__nd$0$i$i2178$lcssa7018 = 0, $__nd$0$i$i2193 = 0, $__nd$0$i$i2193$lcssa = 0;
 var $__nd$0$i$i2193$lcssa7013 = 0, $__nd$0$i$i2193$lcssa7014 = 0, $__nd$0$i$i2208 = 0, $__nd$0$i$i2208$lcssa = 0, $__nd$0$i$i2208$lcssa7009 = 0, $__nd$0$i$i2208$lcssa7010 = 0, $__nd$0$i$i2223 = 0, $__nd$0$i$i2223$lcssa = 0, $__nd$0$i$i2223$lcssa7005 = 0, $__nd$0$i$i2223$lcssa7006 = 0, $__nd$0$i$i2238 = 0, $__nd$0$i$i2238$lcssa = 0, $__nd$0$i$i2238$lcssa7001 = 0, $__nd$0$i$i2238$lcssa7002 = 0, $__nd$0$i$i2253 = 0, $__nd$0$i$i2253$lcssa = 0, $__nd$0$i$i2253$lcssa6997 = 0, $__nd$0$i$i2253$lcssa6998 = 0, $__nd$0$i$i2268 = 0, $__nd$0$i$i2268$lcssa = 0;
 var $__nd$0$i$i2268$lcssa6993 = 0, $__nd$0$i$i2268$lcssa6994 = 0, $__nd$0$i$i228 = 0, $__nd$0$i$i228$lcssa = 0, $__nd$0$i$i228$lcssa7537 = 0, $__nd$0$i$i228$lcssa7538 = 0, $__nd$0$i$i2283 = 0, $__nd$0$i$i2283$lcssa = 0, $__nd$0$i$i2283$lcssa6989 = 0, $__nd$0$i$i2283$lcssa6990 = 0, $__nd$0$i$i2298 = 0, $__nd$0$i$i2298$lcssa = 0, $__nd$0$i$i2298$lcssa6985 = 0, $__nd$0$i$i2298$lcssa6986 = 0, $__nd$0$i$i2313 = 0, $__nd$0$i$i2313$lcssa = 0, $__nd$0$i$i2313$lcssa6981 = 0, $__nd$0$i$i2313$lcssa6982 = 0, $__nd$0$i$i2328 = 0, $__nd$0$i$i2328$lcssa = 0;
 var $__nd$0$i$i2328$lcssa6977 = 0, $__nd$0$i$i2328$lcssa6978 = 0, $__nd$0$i$i2343 = 0, $__nd$0$i$i2343$lcssa = 0, $__nd$0$i$i2343$lcssa6973 = 0, $__nd$0$i$i2343$lcssa6974 = 0, $__nd$0$i$i2358 = 0, $__nd$0$i$i2358$lcssa = 0, $__nd$0$i$i2358$lcssa6969 = 0, $__nd$0$i$i2358$lcssa6970 = 0, $__nd$0$i$i2373 = 0, $__nd$0$i$i2373$lcssa = 0, $__nd$0$i$i2373$lcssa6965 = 0, $__nd$0$i$i2373$lcssa6966 = 0, $__nd$0$i$i2388 = 0, $__nd$0$i$i2388$lcssa = 0, $__nd$0$i$i2388$lcssa6961 = 0, $__nd$0$i$i2388$lcssa6962 = 0, $__nd$0$i$i2403 = 0, $__nd$0$i$i2403$lcssa = 0;
 var $__nd$0$i$i2403$lcssa6957 = 0, $__nd$0$i$i2403$lcssa6958 = 0, $__nd$0$i$i2418 = 0, $__nd$0$i$i2418$lcssa = 0, $__nd$0$i$i2418$lcssa6953 = 0, $__nd$0$i$i2418$lcssa6954 = 0, $__nd$0$i$i243 = 0, $__nd$0$i$i243$lcssa = 0, $__nd$0$i$i243$lcssa7533 = 0, $__nd$0$i$i243$lcssa7534 = 0, $__nd$0$i$i2433 = 0, $__nd$0$i$i2433$lcssa = 0, $__nd$0$i$i2433$lcssa6949 = 0, $__nd$0$i$i2433$lcssa6950 = 0, $__nd$0$i$i2448 = 0, $__nd$0$i$i2448$lcssa = 0, $__nd$0$i$i2448$lcssa6945 = 0, $__nd$0$i$i2448$lcssa6946 = 0, $__nd$0$i$i2463 = 0, $__nd$0$i$i2463$lcssa = 0;
 var $__nd$0$i$i2463$lcssa6941 = 0, $__nd$0$i$i2463$lcssa6942 = 0, $__nd$0$i$i2478 = 0, $__nd$0$i$i2478$lcssa = 0, $__nd$0$i$i2478$lcssa6937 = 0, $__nd$0$i$i2478$lcssa6938 = 0, $__nd$0$i$i2493 = 0, $__nd$0$i$i2493$lcssa = 0, $__nd$0$i$i2493$lcssa6933 = 0, $__nd$0$i$i2493$lcssa6934 = 0, $__nd$0$i$i2508 = 0, $__nd$0$i$i2508$lcssa = 0, $__nd$0$i$i2508$lcssa6929 = 0, $__nd$0$i$i2508$lcssa6930 = 0, $__nd$0$i$i2523 = 0, $__nd$0$i$i2523$lcssa = 0, $__nd$0$i$i2523$lcssa6925 = 0, $__nd$0$i$i2523$lcssa6926 = 0, $__nd$0$i$i2538 = 0, $__nd$0$i$i2538$lcssa = 0;
 var $__nd$0$i$i2538$lcssa6921 = 0, $__nd$0$i$i2538$lcssa6922 = 0, $__nd$0$i$i2553 = 0, $__nd$0$i$i2553$lcssa = 0, $__nd$0$i$i2553$lcssa6917 = 0, $__nd$0$i$i2553$lcssa6918 = 0, $__nd$0$i$i2568 = 0, $__nd$0$i$i2568$lcssa = 0, $__nd$0$i$i2568$lcssa6913 = 0, $__nd$0$i$i2568$lcssa6914 = 0, $__nd$0$i$i258 = 0, $__nd$0$i$i258$lcssa = 0, $__nd$0$i$i258$lcssa7529 = 0, $__nd$0$i$i258$lcssa7530 = 0, $__nd$0$i$i2583 = 0, $__nd$0$i$i2583$lcssa = 0, $__nd$0$i$i2583$lcssa6909 = 0, $__nd$0$i$i2583$lcssa6910 = 0, $__nd$0$i$i2598 = 0, $__nd$0$i$i2598$lcssa = 0;
 var $__nd$0$i$i2598$lcssa6905 = 0, $__nd$0$i$i2598$lcssa6906 = 0, $__nd$0$i$i2613 = 0, $__nd$0$i$i2613$lcssa = 0, $__nd$0$i$i2613$lcssa6901 = 0, $__nd$0$i$i2613$lcssa6902 = 0, $__nd$0$i$i2628 = 0, $__nd$0$i$i2628$lcssa = 0, $__nd$0$i$i2628$lcssa6897 = 0, $__nd$0$i$i2628$lcssa6898 = 0, $__nd$0$i$i2643 = 0, $__nd$0$i$i2643$lcssa = 0, $__nd$0$i$i2643$lcssa6893 = 0, $__nd$0$i$i2643$lcssa6894 = 0, $__nd$0$i$i2658 = 0, $__nd$0$i$i2658$lcssa = 0, $__nd$0$i$i2658$lcssa6889 = 0, $__nd$0$i$i2658$lcssa6890 = 0, $__nd$0$i$i2673 = 0, $__nd$0$i$i2673$lcssa = 0;
 var $__nd$0$i$i2673$lcssa6885 = 0, $__nd$0$i$i2673$lcssa6886 = 0, $__nd$0$i$i2688 = 0, $__nd$0$i$i2688$lcssa = 0, $__nd$0$i$i2688$lcssa6881 = 0, $__nd$0$i$i2688$lcssa6882 = 0, $__nd$0$i$i2703 = 0, $__nd$0$i$i2703$lcssa = 0, $__nd$0$i$i2703$lcssa6877 = 0, $__nd$0$i$i2703$lcssa6878 = 0, $__nd$0$i$i2718 = 0, $__nd$0$i$i2718$lcssa = 0, $__nd$0$i$i2718$lcssa6873 = 0, $__nd$0$i$i2718$lcssa6874 = 0, $__nd$0$i$i273 = 0, $__nd$0$i$i273$lcssa = 0, $__nd$0$i$i273$lcssa7525 = 0, $__nd$0$i$i273$lcssa7526 = 0, $__nd$0$i$i2733 = 0, $__nd$0$i$i2733$lcssa = 0;
 var $__nd$0$i$i2733$lcssa6869 = 0, $__nd$0$i$i2733$lcssa6870 = 0, $__nd$0$i$i2748 = 0, $__nd$0$i$i2748$lcssa = 0, $__nd$0$i$i2748$lcssa6865 = 0, $__nd$0$i$i2748$lcssa6866 = 0, $__nd$0$i$i2763 = 0, $__nd$0$i$i2763$lcssa = 0, $__nd$0$i$i2763$lcssa6861 = 0, $__nd$0$i$i2763$lcssa6862 = 0, $__nd$0$i$i2778 = 0, $__nd$0$i$i2778$lcssa = 0, $__nd$0$i$i2778$lcssa6857 = 0, $__nd$0$i$i2778$lcssa6858 = 0, $__nd$0$i$i2793 = 0, $__nd$0$i$i2793$lcssa = 0, $__nd$0$i$i2793$lcssa6853 = 0, $__nd$0$i$i2793$lcssa6854 = 0, $__nd$0$i$i2808 = 0, $__nd$0$i$i2808$lcssa = 0;
 var $__nd$0$i$i2808$lcssa6849 = 0, $__nd$0$i$i2808$lcssa6850 = 0, $__nd$0$i$i2823 = 0, $__nd$0$i$i2823$lcssa = 0, $__nd$0$i$i2823$lcssa6845 = 0, $__nd$0$i$i2823$lcssa6846 = 0, $__nd$0$i$i2838 = 0, $__nd$0$i$i2838$lcssa = 0, $__nd$0$i$i2838$lcssa6842 = 0, $__nd$0$i$i2838$lcssa6843 = 0, $__nd$0$i$i288 = 0, $__nd$0$i$i288$lcssa = 0, $__nd$0$i$i288$lcssa7521 = 0, $__nd$0$i$i288$lcssa7522 = 0, $__nd$0$i$i3 = 0, $__nd$0$i$i3$lcssa = 0, $__nd$0$i$i3$lcssa7597 = 0, $__nd$0$i$i3$lcssa7598 = 0, $__nd$0$i$i303 = 0, $__nd$0$i$i303$lcssa = 0;
 var $__nd$0$i$i303$lcssa7517 = 0, $__nd$0$i$i303$lcssa7518 = 0, $__nd$0$i$i318 = 0, $__nd$0$i$i318$lcssa = 0, $__nd$0$i$i318$lcssa7513 = 0, $__nd$0$i$i318$lcssa7514 = 0, $__nd$0$i$i33 = 0, $__nd$0$i$i33$lcssa = 0, $__nd$0$i$i33$lcssa7589 = 0, $__nd$0$i$i33$lcssa7590 = 0, $__nd$0$i$i333 = 0, $__nd$0$i$i333$lcssa = 0, $__nd$0$i$i333$lcssa7509 = 0, $__nd$0$i$i333$lcssa7510 = 0, $__nd$0$i$i348 = 0, $__nd$0$i$i348$lcssa = 0, $__nd$0$i$i348$lcssa7505 = 0, $__nd$0$i$i348$lcssa7506 = 0, $__nd$0$i$i363 = 0, $__nd$0$i$i363$lcssa = 0;
 var $__nd$0$i$i363$lcssa7501 = 0, $__nd$0$i$i363$lcssa7502 = 0, $__nd$0$i$i378 = 0, $__nd$0$i$i378$lcssa = 0, $__nd$0$i$i378$lcssa7497 = 0, $__nd$0$i$i378$lcssa7498 = 0, $__nd$0$i$i393 = 0, $__nd$0$i$i393$lcssa = 0, $__nd$0$i$i393$lcssa7493 = 0, $__nd$0$i$i393$lcssa7494 = 0, $__nd$0$i$i408 = 0, $__nd$0$i$i408$lcssa = 0, $__nd$0$i$i408$lcssa7489 = 0, $__nd$0$i$i408$lcssa7490 = 0, $__nd$0$i$i423 = 0, $__nd$0$i$i423$lcssa = 0, $__nd$0$i$i423$lcssa7485 = 0, $__nd$0$i$i423$lcssa7486 = 0, $__nd$0$i$i438 = 0, $__nd$0$i$i438$lcssa = 0;
 var $__nd$0$i$i438$lcssa7481 = 0, $__nd$0$i$i438$lcssa7482 = 0, $__nd$0$i$i453 = 0, $__nd$0$i$i453$lcssa = 0, $__nd$0$i$i453$lcssa7477 = 0, $__nd$0$i$i453$lcssa7478 = 0, $__nd$0$i$i468 = 0, $__nd$0$i$i468$lcssa = 0, $__nd$0$i$i468$lcssa7473 = 0, $__nd$0$i$i468$lcssa7474 = 0, $__nd$0$i$i48 = 0, $__nd$0$i$i48$lcssa = 0, $__nd$0$i$i48$lcssa7585 = 0, $__nd$0$i$i48$lcssa7586 = 0, $__nd$0$i$i483 = 0, $__nd$0$i$i483$lcssa = 0, $__nd$0$i$i483$lcssa7469 = 0, $__nd$0$i$i483$lcssa7470 = 0, $__nd$0$i$i498 = 0, $__nd$0$i$i498$lcssa = 0;
 var $__nd$0$i$i498$lcssa7465 = 0, $__nd$0$i$i498$lcssa7466 = 0, $__nd$0$i$i513 = 0, $__nd$0$i$i513$lcssa = 0, $__nd$0$i$i513$lcssa7461 = 0, $__nd$0$i$i513$lcssa7462 = 0, $__nd$0$i$i528 = 0, $__nd$0$i$i528$lcssa = 0, $__nd$0$i$i528$lcssa7457 = 0, $__nd$0$i$i528$lcssa7458 = 0, $__nd$0$i$i543 = 0, $__nd$0$i$i543$lcssa = 0, $__nd$0$i$i543$lcssa7453 = 0, $__nd$0$i$i543$lcssa7454 = 0, $__nd$0$i$i558 = 0, $__nd$0$i$i558$lcssa = 0, $__nd$0$i$i558$lcssa7449 = 0, $__nd$0$i$i558$lcssa7450 = 0, $__nd$0$i$i573 = 0, $__nd$0$i$i573$lcssa = 0;
 var $__nd$0$i$i573$lcssa7445 = 0, $__nd$0$i$i573$lcssa7446 = 0, $__nd$0$i$i588 = 0, $__nd$0$i$i588$lcssa = 0, $__nd$0$i$i588$lcssa7441 = 0, $__nd$0$i$i588$lcssa7442 = 0, $__nd$0$i$i603 = 0, $__nd$0$i$i603$lcssa = 0, $__nd$0$i$i603$lcssa7437 = 0, $__nd$0$i$i603$lcssa7438 = 0, $__nd$0$i$i618 = 0, $__nd$0$i$i618$lcssa = 0, $__nd$0$i$i618$lcssa7433 = 0, $__nd$0$i$i618$lcssa7434 = 0, $__nd$0$i$i63 = 0, $__nd$0$i$i63$lcssa = 0, $__nd$0$i$i63$lcssa7581 = 0, $__nd$0$i$i63$lcssa7582 = 0, $__nd$0$i$i633 = 0, $__nd$0$i$i633$lcssa = 0;
 var $__nd$0$i$i633$lcssa7429 = 0, $__nd$0$i$i633$lcssa7430 = 0, $__nd$0$i$i648 = 0, $__nd$0$i$i648$lcssa = 0, $__nd$0$i$i648$lcssa7425 = 0, $__nd$0$i$i648$lcssa7426 = 0, $__nd$0$i$i663 = 0, $__nd$0$i$i663$lcssa = 0, $__nd$0$i$i663$lcssa7421 = 0, $__nd$0$i$i663$lcssa7422 = 0, $__nd$0$i$i678 = 0, $__nd$0$i$i678$lcssa = 0, $__nd$0$i$i678$lcssa7417 = 0, $__nd$0$i$i678$lcssa7418 = 0, $__nd$0$i$i693 = 0, $__nd$0$i$i693$lcssa = 0, $__nd$0$i$i693$lcssa7413 = 0, $__nd$0$i$i693$lcssa7414 = 0, $__nd$0$i$i708 = 0, $__nd$0$i$i708$lcssa = 0;
 var $__nd$0$i$i708$lcssa7409 = 0, $__nd$0$i$i708$lcssa7410 = 0, $__nd$0$i$i723 = 0, $__nd$0$i$i723$lcssa = 0, $__nd$0$i$i723$lcssa7405 = 0, $__nd$0$i$i723$lcssa7406 = 0, $__nd$0$i$i738 = 0, $__nd$0$i$i738$lcssa = 0, $__nd$0$i$i738$lcssa7401 = 0, $__nd$0$i$i738$lcssa7402 = 0, $__nd$0$i$i753 = 0, $__nd$0$i$i753$lcssa = 0, $__nd$0$i$i753$lcssa7397 = 0, $__nd$0$i$i753$lcssa7398 = 0, $__nd$0$i$i768 = 0, $__nd$0$i$i768$lcssa = 0, $__nd$0$i$i768$lcssa7393 = 0, $__nd$0$i$i768$lcssa7394 = 0, $__nd$0$i$i78 = 0, $__nd$0$i$i78$lcssa = 0;
 var $__nd$0$i$i78$lcssa7577 = 0, $__nd$0$i$i78$lcssa7578 = 0, $__nd$0$i$i783 = 0, $__nd$0$i$i783$lcssa = 0, $__nd$0$i$i783$lcssa7389 = 0, $__nd$0$i$i783$lcssa7390 = 0, $__nd$0$i$i798 = 0, $__nd$0$i$i798$lcssa = 0, $__nd$0$i$i798$lcssa7385 = 0, $__nd$0$i$i798$lcssa7386 = 0, $__nd$0$i$i813 = 0, $__nd$0$i$i813$lcssa = 0, $__nd$0$i$i813$lcssa7381 = 0, $__nd$0$i$i813$lcssa7382 = 0, $__nd$0$i$i828 = 0, $__nd$0$i$i828$lcssa = 0, $__nd$0$i$i828$lcssa7377 = 0, $__nd$0$i$i828$lcssa7378 = 0, $__nd$0$i$i843 = 0, $__nd$0$i$i843$lcssa = 0;
 var $__nd$0$i$i843$lcssa7373 = 0, $__nd$0$i$i843$lcssa7374 = 0, $__nd$0$i$i858 = 0, $__nd$0$i$i858$lcssa = 0, $__nd$0$i$i858$lcssa7369 = 0, $__nd$0$i$i858$lcssa7370 = 0, $__nd$0$i$i873 = 0, $__nd$0$i$i873$lcssa = 0, $__nd$0$i$i873$lcssa7365 = 0, $__nd$0$i$i873$lcssa7366 = 0, $__nd$0$i$i888 = 0, $__nd$0$i$i888$lcssa = 0, $__nd$0$i$i888$lcssa7361 = 0, $__nd$0$i$i888$lcssa7362 = 0, $__nd$0$i$i903 = 0, $__nd$0$i$i903$lcssa = 0, $__nd$0$i$i903$lcssa7357 = 0, $__nd$0$i$i903$lcssa7358 = 0, $__nd$0$i$i918 = 0, $__nd$0$i$i918$lcssa = 0;
 var $__nd$0$i$i918$lcssa7353 = 0, $__nd$0$i$i918$lcssa7354 = 0, $__nd$0$i$i93 = 0, $__nd$0$i$i93$lcssa = 0, $__nd$0$i$i93$lcssa7573 = 0, $__nd$0$i$i93$lcssa7574 = 0, $__nd$0$i$i933 = 0, $__nd$0$i$i933$lcssa = 0, $__nd$0$i$i933$lcssa7349 = 0, $__nd$0$i$i933$lcssa7350 = 0, $__nd$0$i$i948 = 0, $__nd$0$i$i948$lcssa = 0, $__nd$0$i$i948$lcssa7345 = 0, $__nd$0$i$i948$lcssa7346 = 0, $__nd$0$i$i963 = 0, $__nd$0$i$i963$lcssa = 0, $__nd$0$i$i963$lcssa7341 = 0, $__nd$0$i$i963$lcssa7342 = 0, $__nd$0$i$i978 = 0, $__nd$0$i$i978$lcssa = 0;
 var $__nd$0$i$i978$lcssa7337 = 0, $__nd$0$i$i978$lcssa7338 = 0, $__nd$0$i$i993 = 0, $__nd$0$i$i993$lcssa = 0, $__nd$0$i$i993$lcssa7333 = 0, $__nd$0$i$i993$lcssa7334 = 0, $__parent$i1 = 0, $__parent$i1006 = 0, $__parent$i1021 = 0, $__parent$i1036 = 0, $__parent$i1051 = 0, $__parent$i106 = 0, $__parent$i1066 = 0, $__parent$i1081 = 0, $__parent$i1096 = 0, $__parent$i1111 = 0, $__parent$i1126 = 0, $__parent$i1141 = 0, $__parent$i1156 = 0, $__parent$i1171 = 0;
 var $__parent$i1186 = 0, $__parent$i1201 = 0, $__parent$i121 = 0, $__parent$i1216 = 0, $__parent$i1231 = 0, $__parent$i1246 = 0, $__parent$i1261 = 0, $__parent$i1276 = 0, $__parent$i1291 = 0, $__parent$i1306 = 0, $__parent$i1321 = 0, $__parent$i1336 = 0, $__parent$i1351 = 0, $__parent$i136 = 0, $__parent$i1366 = 0, $__parent$i1381 = 0, $__parent$i1396 = 0, $__parent$i1411 = 0, $__parent$i1426 = 0, $__parent$i1441 = 0;
 var $__parent$i1456 = 0, $__parent$i1471 = 0, $__parent$i1486 = 0, $__parent$i1501 = 0, $__parent$i151 = 0, $__parent$i1516 = 0, $__parent$i1531 = 0, $__parent$i1546 = 0, $__parent$i1561 = 0, $__parent$i1576 = 0, $__parent$i1591 = 0, $__parent$i16 = 0, $__parent$i1606 = 0, $__parent$i1621 = 0, $__parent$i1636 = 0, $__parent$i1651 = 0, $__parent$i166 = 0, $__parent$i1666 = 0, $__parent$i1681 = 0, $__parent$i1696 = 0;
 var $__parent$i1711 = 0, $__parent$i1726 = 0, $__parent$i1741 = 0, $__parent$i1756 = 0, $__parent$i1771 = 0, $__parent$i1786 = 0, $__parent$i1801 = 0, $__parent$i181 = 0, $__parent$i1816 = 0, $__parent$i1831 = 0, $__parent$i1846 = 0, $__parent$i1861 = 0, $__parent$i1876 = 0, $__parent$i1891 = 0, $__parent$i1906 = 0, $__parent$i1921 = 0, $__parent$i1936 = 0, $__parent$i1951 = 0, $__parent$i196 = 0, $__parent$i1966 = 0;
 var $__parent$i1981 = 0, $__parent$i1996 = 0, $__parent$i2011 = 0, $__parent$i2026 = 0, $__parent$i2041 = 0, $__parent$i2056 = 0, $__parent$i2071 = 0, $__parent$i2086 = 0, $__parent$i2101 = 0, $__parent$i211 = 0, $__parent$i2116 = 0, $__parent$i2131 = 0, $__parent$i2146 = 0, $__parent$i2161 = 0, $__parent$i2176 = 0, $__parent$i2191 = 0, $__parent$i2206 = 0, $__parent$i2221 = 0, $__parent$i2236 = 0, $__parent$i2251 = 0;
 var $__parent$i226 = 0, $__parent$i2266 = 0, $__parent$i2281 = 0, $__parent$i2296 = 0, $__parent$i2311 = 0, $__parent$i2326 = 0, $__parent$i2341 = 0, $__parent$i2356 = 0, $__parent$i2371 = 0, $__parent$i2386 = 0, $__parent$i2401 = 0, $__parent$i241 = 0, $__parent$i2416 = 0, $__parent$i2431 = 0, $__parent$i2446 = 0, $__parent$i2461 = 0, $__parent$i2476 = 0, $__parent$i2491 = 0, $__parent$i2506 = 0, $__parent$i2521 = 0;
 var $__parent$i2536 = 0, $__parent$i2551 = 0, $__parent$i256 = 0, $__parent$i2566 = 0, $__parent$i2581 = 0, $__parent$i2596 = 0, $__parent$i2611 = 0, $__parent$i2626 = 0, $__parent$i2641 = 0, $__parent$i2656 = 0, $__parent$i2671 = 0, $__parent$i2686 = 0, $__parent$i2701 = 0, $__parent$i271 = 0, $__parent$i2716 = 0, $__parent$i2731 = 0, $__parent$i2746 = 0, $__parent$i2761 = 0, $__parent$i2776 = 0, $__parent$i2791 = 0;
 var $__parent$i2806 = 0, $__parent$i2821 = 0, $__parent$i2836 = 0, $__parent$i286 = 0, $__parent$i301 = 0, $__parent$i31 = 0, $__parent$i316 = 0, $__parent$i331 = 0, $__parent$i346 = 0, $__parent$i361 = 0, $__parent$i376 = 0, $__parent$i391 = 0, $__parent$i406 = 0, $__parent$i421 = 0, $__parent$i436 = 0, $__parent$i451 = 0, $__parent$i46 = 0, $__parent$i466 = 0, $__parent$i481 = 0, $__parent$i496 = 0;
 var $__parent$i511 = 0, $__parent$i526 = 0, $__parent$i541 = 0, $__parent$i556 = 0, $__parent$i571 = 0, $__parent$i586 = 0, $__parent$i601 = 0, $__parent$i61 = 0, $__parent$i616 = 0, $__parent$i631 = 0, $__parent$i646 = 0, $__parent$i661 = 0, $__parent$i676 = 0, $__parent$i691 = 0, $__parent$i706 = 0, $__parent$i721 = 0, $__parent$i736 = 0, $__parent$i751 = 0, $__parent$i76 = 0, $__parent$i766 = 0;
 var $__parent$i781 = 0, $__parent$i796 = 0, $__parent$i811 = 0, $__parent$i826 = 0, $__parent$i841 = 0, $__parent$i856 = 0, $__parent$i871 = 0, $__parent$i886 = 0, $__parent$i901 = 0, $__parent$i91 = 0, $__parent$i916 = 0, $__parent$i931 = 0, $__parent$i946 = 0, $__parent$i961 = 0, $__parent$i976 = 0, $__parent$i991 = 0, $__r$0$i1003 = 0, $__r$0$i1018 = 0, $__r$0$i103 = 0, $__r$0$i1033 = 0;
 var $__r$0$i1048 = 0, $__r$0$i1063 = 0, $__r$0$i1078 = 0, $__r$0$i1093 = 0, $__r$0$i1108 = 0, $__r$0$i1123 = 0, $__r$0$i1138 = 0, $__r$0$i1153 = 0, $__r$0$i1168 = 0, $__r$0$i118 = 0, $__r$0$i1183 = 0, $__r$0$i1198 = 0, $__r$0$i1213 = 0, $__r$0$i1228 = 0, $__r$0$i1243 = 0, $__r$0$i1258 = 0, $__r$0$i1273 = 0, $__r$0$i1288 = 0, $__r$0$i13 = 0, $__r$0$i1303 = 0;
 var $__r$0$i1318 = 0, $__r$0$i133 = 0, $__r$0$i1333 = 0, $__r$0$i1348 = 0, $__r$0$i1363 = 0, $__r$0$i1378 = 0, $__r$0$i1393 = 0, $__r$0$i1408 = 0, $__r$0$i1423 = 0, $__r$0$i1438 = 0, $__r$0$i1453 = 0, $__r$0$i1468 = 0, $__r$0$i148 = 0, $__r$0$i1483 = 0, $__r$0$i1498 = 0, $__r$0$i1513 = 0, $__r$0$i1528 = 0, $__r$0$i1543 = 0, $__r$0$i1558 = 0, $__r$0$i1573 = 0;
 var $__r$0$i1588 = 0, $__r$0$i1603 = 0, $__r$0$i1618 = 0, $__r$0$i163 = 0, $__r$0$i1633 = 0, $__r$0$i1648 = 0, $__r$0$i1663 = 0, $__r$0$i1678 = 0, $__r$0$i1693 = 0, $__r$0$i1708 = 0, $__r$0$i1723 = 0, $__r$0$i1738 = 0, $__r$0$i1753 = 0, $__r$0$i1768 = 0, $__r$0$i178 = 0, $__r$0$i1783 = 0, $__r$0$i1798 = 0, $__r$0$i1813 = 0, $__r$0$i1828 = 0, $__r$0$i1843 = 0;
 var $__r$0$i1858 = 0, $__r$0$i1873 = 0, $__r$0$i1888 = 0, $__r$0$i1903 = 0, $__r$0$i1918 = 0, $__r$0$i193 = 0, $__r$0$i1933 = 0, $__r$0$i1948 = 0, $__r$0$i1963 = 0, $__r$0$i1978 = 0, $__r$0$i1993 = 0, $__r$0$i2008 = 0, $__r$0$i2023 = 0, $__r$0$i2038 = 0, $__r$0$i2053 = 0, $__r$0$i2068 = 0, $__r$0$i208 = 0, $__r$0$i2083 = 0, $__r$0$i2098 = 0, $__r$0$i2113 = 0;
 var $__r$0$i2128 = 0, $__r$0$i2143 = 0, $__r$0$i2158 = 0, $__r$0$i2173 = 0, $__r$0$i2188 = 0, $__r$0$i2203 = 0, $__r$0$i2218 = 0, $__r$0$i223 = 0, $__r$0$i2233 = 0, $__r$0$i2248 = 0, $__r$0$i2263 = 0, $__r$0$i2278 = 0, $__r$0$i2293 = 0, $__r$0$i2308 = 0, $__r$0$i2323 = 0, $__r$0$i2338 = 0, $__r$0$i2353 = 0, $__r$0$i2368 = 0, $__r$0$i238 = 0, $__r$0$i2383 = 0;
 var $__r$0$i2398 = 0, $__r$0$i2413 = 0, $__r$0$i2428 = 0, $__r$0$i2443 = 0, $__r$0$i2458 = 0, $__r$0$i2473 = 0, $__r$0$i2488 = 0, $__r$0$i2503 = 0, $__r$0$i2518 = 0, $__r$0$i253 = 0, $__r$0$i2533 = 0, $__r$0$i2548 = 0, $__r$0$i2563 = 0, $__r$0$i2578 = 0, $__r$0$i2593 = 0, $__r$0$i2608 = 0, $__r$0$i2623 = 0, $__r$0$i2638 = 0, $__r$0$i2653 = 0, $__r$0$i2668 = 0;
 var $__r$0$i268 = 0, $__r$0$i2683 = 0, $__r$0$i2698 = 0, $__r$0$i2713 = 0, $__r$0$i2728 = 0, $__r$0$i2743 = 0, $__r$0$i2758 = 0, $__r$0$i2773 = 0, $__r$0$i2788 = 0, $__r$0$i28 = 0, $__r$0$i2803 = 0, $__r$0$i2818 = 0, $__r$0$i283 = 0, $__r$0$i2833 = 0, $__r$0$i2848 = 0, $__r$0$i298 = 0, $__r$0$i313 = 0, $__r$0$i328 = 0, $__r$0$i343 = 0, $__r$0$i358 = 0;
 var $__r$0$i373 = 0, $__r$0$i388 = 0, $__r$0$i403 = 0, $__r$0$i418 = 0, $__r$0$i43 = 0, $__r$0$i433 = 0, $__r$0$i448 = 0, $__r$0$i463 = 0, $__r$0$i478 = 0, $__r$0$i493 = 0, $__r$0$i508 = 0, $__r$0$i523 = 0, $__r$0$i538 = 0, $__r$0$i553 = 0, $__r$0$i568 = 0, $__r$0$i58 = 0, $__r$0$i583 = 0, $__r$0$i598 = 0, $__r$0$i613 = 0, $__r$0$i628 = 0;
 var $__r$0$i643 = 0, $__r$0$i658 = 0, $__r$0$i673 = 0, $__r$0$i688 = 0, $__r$0$i703 = 0, $__r$0$i718 = 0, $__r$0$i73 = 0, $__r$0$i733 = 0, $__r$0$i748 = 0, $__r$0$i763 = 0, $__r$0$i778 = 0, $__r$0$i793 = 0, $__r$0$i808 = 0, $__r$0$i823 = 0, $__r$0$i838 = 0, $__r$0$i853 = 0, $__r$0$i868 = 0, $__r$0$i88 = 0, $__r$0$i883 = 0, $__r$0$i898 = 0;
 var $__r$0$i913 = 0, $__r$0$i928 = 0, $__r$0$i943 = 0, $__r$0$i958 = 0, $__r$0$i973 = 0, $__r$0$i988 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 768|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__parent$i2836 = sp + 756|0;
 $__parent$i2821 = sp + 752|0;
 $__parent$i2806 = sp + 748|0;
 $__parent$i2791 = sp + 744|0;
 $__parent$i2776 = sp + 740|0;
 $__parent$i2761 = sp + 736|0;
 $__parent$i2746 = sp + 732|0;
 $__parent$i2731 = sp + 728|0;
 $__parent$i2716 = sp + 724|0;
 $__parent$i2701 = sp + 720|0;
 $__parent$i2686 = sp + 716|0;
 $__parent$i2671 = sp + 712|0;
 $__parent$i2656 = sp + 708|0;
 $__parent$i2641 = sp + 704|0;
 $__parent$i2626 = sp + 700|0;
 $__parent$i2611 = sp + 696|0;
 $__parent$i2596 = sp + 692|0;
 $__parent$i2581 = sp + 688|0;
 $__parent$i2566 = sp + 684|0;
 $__parent$i2551 = sp + 680|0;
 $__parent$i2536 = sp + 676|0;
 $__parent$i2521 = sp + 672|0;
 $__parent$i2506 = sp + 668|0;
 $__parent$i2491 = sp + 664|0;
 $__parent$i2476 = sp + 660|0;
 $__parent$i2461 = sp + 656|0;
 $__parent$i2446 = sp + 652|0;
 $__parent$i2431 = sp + 648|0;
 $__parent$i2416 = sp + 644|0;
 $__parent$i2401 = sp + 640|0;
 $__parent$i2386 = sp + 636|0;
 $__parent$i2371 = sp + 632|0;
 $__parent$i2356 = sp + 628|0;
 $__parent$i2341 = sp + 624|0;
 $__parent$i2326 = sp + 620|0;
 $__parent$i2311 = sp + 616|0;
 $__parent$i2296 = sp + 612|0;
 $__parent$i2281 = sp + 608|0;
 $__parent$i2266 = sp + 604|0;
 $__parent$i2251 = sp + 600|0;
 $__parent$i2236 = sp + 596|0;
 $__parent$i2221 = sp + 592|0;
 $__parent$i2206 = sp + 588|0;
 $__parent$i2191 = sp + 584|0;
 $__parent$i2176 = sp + 580|0;
 $__parent$i2161 = sp + 576|0;
 $__parent$i2146 = sp + 572|0;
 $__parent$i2131 = sp + 568|0;
 $__parent$i2116 = sp + 564|0;
 $__parent$i2101 = sp + 560|0;
 $__parent$i2086 = sp + 556|0;
 $__parent$i2071 = sp + 552|0;
 $__parent$i2056 = sp + 548|0;
 $__parent$i2041 = sp + 544|0;
 $__parent$i2026 = sp + 540|0;
 $__parent$i2011 = sp + 536|0;
 $__parent$i1996 = sp + 532|0;
 $__parent$i1981 = sp + 528|0;
 $__parent$i1966 = sp + 524|0;
 $__parent$i1951 = sp + 520|0;
 $__parent$i1936 = sp + 516|0;
 $__parent$i1921 = sp + 512|0;
 $__parent$i1906 = sp + 508|0;
 $__parent$i1891 = sp + 504|0;
 $__parent$i1876 = sp + 500|0;
 $__parent$i1861 = sp + 496|0;
 $__parent$i1846 = sp + 492|0;
 $__parent$i1831 = sp + 488|0;
 $__parent$i1816 = sp + 484|0;
 $__parent$i1801 = sp + 480|0;
 $__parent$i1786 = sp + 476|0;
 $__parent$i1771 = sp + 472|0;
 $__parent$i1756 = sp + 468|0;
 $__parent$i1741 = sp + 464|0;
 $__parent$i1726 = sp + 460|0;
 $__parent$i1711 = sp + 456|0;
 $__parent$i1696 = sp + 452|0;
 $__parent$i1681 = sp + 448|0;
 $__parent$i1666 = sp + 444|0;
 $__parent$i1651 = sp + 440|0;
 $__parent$i1636 = sp + 436|0;
 $__parent$i1621 = sp + 432|0;
 $__parent$i1606 = sp + 428|0;
 $__parent$i1591 = sp + 424|0;
 $__parent$i1576 = sp + 420|0;
 $__parent$i1561 = sp + 416|0;
 $__parent$i1546 = sp + 412|0;
 $__parent$i1531 = sp + 408|0;
 $__parent$i1516 = sp + 404|0;
 $__parent$i1501 = sp + 400|0;
 $__parent$i1486 = sp + 396|0;
 $__parent$i1471 = sp + 392|0;
 $__parent$i1456 = sp + 388|0;
 $__parent$i1441 = sp + 384|0;
 $__parent$i1426 = sp + 380|0;
 $__parent$i1411 = sp + 376|0;
 $__parent$i1396 = sp + 372|0;
 $__parent$i1381 = sp + 368|0;
 $__parent$i1366 = sp + 364|0;
 $__parent$i1351 = sp + 360|0;
 $__parent$i1336 = sp + 356|0;
 $__parent$i1321 = sp + 352|0;
 $__parent$i1306 = sp + 348|0;
 $__parent$i1291 = sp + 344|0;
 $__parent$i1276 = sp + 340|0;
 $__parent$i1261 = sp + 336|0;
 $__parent$i1246 = sp + 332|0;
 $__parent$i1231 = sp + 328|0;
 $__parent$i1216 = sp + 324|0;
 $__parent$i1201 = sp + 320|0;
 $__parent$i1186 = sp + 316|0;
 $__parent$i1171 = sp + 312|0;
 $__parent$i1156 = sp + 308|0;
 $__parent$i1141 = sp + 304|0;
 $__parent$i1126 = sp + 300|0;
 $__parent$i1111 = sp + 296|0;
 $__parent$i1096 = sp + 292|0;
 $__parent$i1081 = sp + 288|0;
 $__parent$i1066 = sp + 284|0;
 $__parent$i1051 = sp + 280|0;
 $__parent$i1036 = sp + 276|0;
 $__parent$i1021 = sp + 272|0;
 $__parent$i1006 = sp + 268|0;
 $__parent$i991 = sp + 264|0;
 $__parent$i976 = sp + 260|0;
 $__parent$i961 = sp + 256|0;
 $__parent$i946 = sp + 252|0;
 $__parent$i931 = sp + 248|0;
 $__parent$i916 = sp + 244|0;
 $__parent$i901 = sp + 240|0;
 $__parent$i886 = sp + 236|0;
 $__parent$i871 = sp + 232|0;
 $__parent$i856 = sp + 228|0;
 $__parent$i841 = sp + 224|0;
 $__parent$i826 = sp + 220|0;
 $__parent$i811 = sp + 216|0;
 $__parent$i796 = sp + 212|0;
 $__parent$i781 = sp + 208|0;
 $__parent$i766 = sp + 204|0;
 $__parent$i751 = sp + 200|0;
 $__parent$i736 = sp + 196|0;
 $__parent$i721 = sp + 192|0;
 $__parent$i706 = sp + 188|0;
 $__parent$i691 = sp + 184|0;
 $__parent$i676 = sp + 180|0;
 $__parent$i661 = sp + 176|0;
 $__parent$i646 = sp + 172|0;
 $__parent$i631 = sp + 168|0;
 $__parent$i616 = sp + 164|0;
 $__parent$i601 = sp + 160|0;
 $__parent$i586 = sp + 156|0;
 $__parent$i571 = sp + 152|0;
 $__parent$i556 = sp + 148|0;
 $__parent$i541 = sp + 144|0;
 $__parent$i526 = sp + 140|0;
 $__parent$i511 = sp + 136|0;
 $__parent$i496 = sp + 132|0;
 $__parent$i481 = sp + 128|0;
 $__parent$i466 = sp + 124|0;
 $__parent$i451 = sp + 120|0;
 $__parent$i436 = sp + 116|0;
 $__parent$i421 = sp + 112|0;
 $__parent$i406 = sp + 108|0;
 $__parent$i391 = sp + 104|0;
 $__parent$i376 = sp + 100|0;
 $__parent$i361 = sp + 96|0;
 $__parent$i346 = sp + 92|0;
 $__parent$i331 = sp + 88|0;
 $__parent$i316 = sp + 84|0;
 $__parent$i301 = sp + 80|0;
 $__parent$i286 = sp + 76|0;
 $__parent$i271 = sp + 72|0;
 $__parent$i256 = sp + 68|0;
 $__parent$i241 = sp + 64|0;
 $__parent$i226 = sp + 60|0;
 $__parent$i211 = sp + 56|0;
 $__parent$i196 = sp + 52|0;
 $__parent$i181 = sp + 48|0;
 $__parent$i166 = sp + 44|0;
 $__parent$i151 = sp + 40|0;
 $__parent$i136 = sp + 36|0;
 $__parent$i121 = sp + 32|0;
 $__parent$i106 = sp + 28|0;
 $__parent$i91 = sp + 24|0;
 $__parent$i76 = sp + 20|0;
 $__parent$i61 = sp + 16|0;
 $__parent$i46 = sp + 12|0;
 $__parent$i31 = sp + 8|0;
 $__parent$i16 = sp + 4|0;
 $__parent$i1 = sp;
 $0 = ((($agg$result)) + 4|0); //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$0>>2] = 0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $1 = ((($agg$result)) + 8|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$1>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($agg$result)) + 4|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$agg$result>>2] = $2; //@line 1130 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __THREW__ = 0;
 $3 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6284 = ___cxa_find_matching_catch_2()|0;
  $6285 = tempRet0;
  $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  ___resumeException($6284|0);
  // unreachable;
 }
 $6 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $7 = ((($3)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$7>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = ((($3)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$8>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$3>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $9 = ((($3)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$9>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $10 = ((($3)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$10>>2] = $6; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$2>>2] = $3; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $11 = $3; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$agg$result>>2] = $11; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3,$3); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $12 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $13 = (($12) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$1>>2] = $13; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $$pre = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $14 = ((($3)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$14>>2] = 0; //@line 610 "modules/event/sdl/Event.cpp"
 $15 = ($$pre|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($15) {
   HEAP32[$__parent$i1>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $27 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i9 = $2;$37 = $27;
   label = 12;
  } else {
   $__nd$0$i$i3 = $$pre;
   while(1) {
    $16 = ((($__nd$0$i$i3)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $17 = HEAP32[$16>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $18 = ($17|0)>(13); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($18) {
     $19 = HEAP32[$__nd$0$i$i3>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $20 = ($19|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($20) {
      $$lcssa7600 = $__nd$0$i$i3;$__nd$0$i$i3$lcssa7598 = $__nd$0$i$i3;
      label = 5;
      break;
     } else {
      $$sink$i$i4 = $19;
     }
    } else {
     $22 = ($17|0)<(13); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($22)) {
      $__nd$0$i$i3$lcssa = $__nd$0$i$i3;
      label = 11;
      break;
     }
     $23 = ((($__nd$0$i$i3)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $24 = HEAP32[$23>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $25 = ($24|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($25) {
      $$lcssa7599 = $23;$__nd$0$i$i3$lcssa7597 = $__nd$0$i$i3;
      label = 9;
      break;
     } else {
      $$sink$i$i4 = $24;
     }
    }
    $__nd$0$i$i3 = $$sink$i$i4;
   }
   if ((label|0) == 5) {
    HEAP32[$__parent$i1>>2] = $__nd$0$i$i3$lcssa7598; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $21 = $__nd$0$i$i3$lcssa7598; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i9 = $$lcssa7600;$37 = $21;
    label = 12;
    break;
   }
   else if ((label|0) == 9) {
    HEAP32[$__parent$i1>>2] = $__nd$0$i$i3$lcssa7597; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $26 = $__nd$0$i$i3$lcssa7597; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i9 = $$lcssa7599;$37 = $26;
    label = 12;
    break;
   }
   else if ((label|0) == 11) {
    HEAP32[$__parent$i1>>2] = $__nd$0$i$i3$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $28 = $__nd$0$i$i3$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $29 = ($__nd$0$i$i3$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($29) {
     $$0$i7$i9 = $__parent$i1;$37 = $28;
     label = 12;
     break;
    } else {
     $48 = $$pre;$__r$0$i13 = $__nd$0$i$i3$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 12) {
  __THREW__ = 0;
  $30 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $31 = __THREW__; __THREW__ = 0;
  $32 = $31&1;
  if ($32) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $33 = ((($30)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$33>>2] = 13; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $34 = ((($30)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$34>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$30>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $35 = ((($30)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$35>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $36 = ((($30)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$36>>2] = $37; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i9>>2] = $30; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $38 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $39 = HEAP32[$38>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $40 = ($39|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($40) {
   $43 = $30;
  } else {
   $41 = $39; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $41; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i11 = HEAP32[$$0$i7$i9>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $43 = $$pre$i$i11;
  }
  $42 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($42,$43); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $44 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $45 = (($44) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $45; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4377 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $48 = $$pre4377;$__r$0$i13 = $30;
 }
 $46 = ((($__r$0$i13)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$46>>2] = 1; //@line 612 "modules/event/sdl/Event.cpp"
 $47 = ($48|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($47) {
   HEAP32[$__parent$i16>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $60 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i24 = $2;$70 = $60;
   label = 26;
  } else {
   $__nd$0$i$i18 = $48;
   while(1) {
    $49 = ((($__nd$0$i$i18)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $50 = HEAP32[$49>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $51 = ($50|0)>(27); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($51) {
     $52 = HEAP32[$__nd$0$i$i18>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $53 = ($52|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($53) {
      $$lcssa7596 = $__nd$0$i$i18;$__nd$0$i$i18$lcssa7594 = $__nd$0$i$i18;
      label = 19;
      break;
     } else {
      $$sink$i$i19 = $52;
     }
    } else {
     $55 = ($50|0)<(27); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($55)) {
      $__nd$0$i$i18$lcssa = $__nd$0$i$i18;
      label = 25;
      break;
     }
     $56 = ((($__nd$0$i$i18)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $57 = HEAP32[$56>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $58 = ($57|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($58) {
      $$lcssa7595 = $56;$__nd$0$i$i18$lcssa7593 = $__nd$0$i$i18;
      label = 23;
      break;
     } else {
      $$sink$i$i19 = $57;
     }
    }
    $__nd$0$i$i18 = $$sink$i$i19;
   }
   if ((label|0) == 19) {
    HEAP32[$__parent$i16>>2] = $__nd$0$i$i18$lcssa7594; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $54 = $__nd$0$i$i18$lcssa7594; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i24 = $$lcssa7596;$70 = $54;
    label = 26;
    break;
   }
   else if ((label|0) == 23) {
    HEAP32[$__parent$i16>>2] = $__nd$0$i$i18$lcssa7593; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $59 = $__nd$0$i$i18$lcssa7593; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i24 = $$lcssa7595;$70 = $59;
    label = 26;
    break;
   }
   else if ((label|0) == 25) {
    HEAP32[$__parent$i16>>2] = $__nd$0$i$i18$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $61 = $__nd$0$i$i18$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $62 = ($__nd$0$i$i18$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($62) {
     $$0$i7$i24 = $__parent$i16;$70 = $61;
     label = 26;
     break;
    } else {
     $81 = $48;$__r$0$i28 = $__nd$0$i$i18$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 26) {
  __THREW__ = 0;
  $63 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $64 = __THREW__; __THREW__ = 0;
  $65 = $64&1;
  if ($65) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $66 = ((($63)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$66>>2] = 27; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $67 = ((($63)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$67>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$63>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $68 = ((($63)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$68>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $69 = ((($63)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$69>>2] = $70; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i24>>2] = $63; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $71 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $72 = HEAP32[$71>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $73 = ($72|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($73) {
   $76 = $63;
  } else {
   $74 = $72; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $74; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i26 = HEAP32[$$0$i7$i24>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $76 = $$pre$i$i26;
  }
  $75 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($75,$76); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $77 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $78 = (($77) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $78; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4378 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $81 = $$pre4378;$__r$0$i28 = $63;
 }
 $79 = ((($__r$0$i28)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$79>>2] = 2; //@line 613 "modules/event/sdl/Event.cpp"
 $80 = ($81|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($80) {
   HEAP32[$__parent$i31>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $93 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i39 = $2;$103 = $93;
   label = 40;
  } else {
   $__nd$0$i$i33 = $81;
   while(1) {
    $82 = ((($__nd$0$i$i33)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $83 = HEAP32[$82>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $84 = ($83|0)>(8); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($84) {
     $85 = HEAP32[$__nd$0$i$i33>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $86 = ($85|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($86) {
      $$lcssa7592 = $__nd$0$i$i33;$__nd$0$i$i33$lcssa7590 = $__nd$0$i$i33;
      label = 33;
      break;
     } else {
      $$sink$i$i34 = $85;
     }
    } else {
     $88 = ($83|0)<(8); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($88)) {
      $__nd$0$i$i33$lcssa = $__nd$0$i$i33;
      label = 39;
      break;
     }
     $89 = ((($__nd$0$i$i33)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $90 = HEAP32[$89>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $91 = ($90|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($91) {
      $$lcssa7591 = $89;$__nd$0$i$i33$lcssa7589 = $__nd$0$i$i33;
      label = 37;
      break;
     } else {
      $$sink$i$i34 = $90;
     }
    }
    $__nd$0$i$i33 = $$sink$i$i34;
   }
   if ((label|0) == 33) {
    HEAP32[$__parent$i31>>2] = $__nd$0$i$i33$lcssa7590; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $87 = $__nd$0$i$i33$lcssa7590; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i39 = $$lcssa7592;$103 = $87;
    label = 40;
    break;
   }
   else if ((label|0) == 37) {
    HEAP32[$__parent$i31>>2] = $__nd$0$i$i33$lcssa7589; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $92 = $__nd$0$i$i33$lcssa7589; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i39 = $$lcssa7591;$103 = $92;
    label = 40;
    break;
   }
   else if ((label|0) == 39) {
    HEAP32[$__parent$i31>>2] = $__nd$0$i$i33$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $94 = $__nd$0$i$i33$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $95 = ($__nd$0$i$i33$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($95) {
     $$0$i7$i39 = $__parent$i31;$103 = $94;
     label = 40;
     break;
    } else {
     $114 = $81;$__r$0$i43 = $__nd$0$i$i33$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 40) {
  __THREW__ = 0;
  $96 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $97 = __THREW__; __THREW__ = 0;
  $98 = $97&1;
  if ($98) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $99 = ((($96)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$99>>2] = 8; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $100 = ((($96)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$100>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$96>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $101 = ((($96)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$101>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $102 = ((($96)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$102>>2] = $103; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i39>>2] = $96; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $104 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $105 = HEAP32[$104>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $106 = ($105|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($106) {
   $109 = $96;
  } else {
   $107 = $105; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $107; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i41 = HEAP32[$$0$i7$i39>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $109 = $$pre$i$i41;
  }
  $108 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($108,$109); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $110 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $111 = (($110) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $111; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4379 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $114 = $$pre4379;$__r$0$i43 = $96;
 }
 $112 = ((($__r$0$i43)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$112>>2] = 3; //@line 614 "modules/event/sdl/Event.cpp"
 $113 = ($114|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($113) {
   HEAP32[$__parent$i46>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $126 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i54 = $2;$136 = $126;
   label = 54;
  } else {
   $__nd$0$i$i48 = $114;
   while(1) {
    $115 = ((($__nd$0$i$i48)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $116 = HEAP32[$115>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $117 = ($116|0)>(9); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($117) {
     $118 = HEAP32[$__nd$0$i$i48>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $119 = ($118|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($119) {
      $$lcssa7588 = $__nd$0$i$i48;$__nd$0$i$i48$lcssa7586 = $__nd$0$i$i48;
      label = 47;
      break;
     } else {
      $$sink$i$i49 = $118;
     }
    } else {
     $121 = ($116|0)<(9); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($121)) {
      $__nd$0$i$i48$lcssa = $__nd$0$i$i48;
      label = 53;
      break;
     }
     $122 = ((($__nd$0$i$i48)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $123 = HEAP32[$122>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $124 = ($123|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($124) {
      $$lcssa7587 = $122;$__nd$0$i$i48$lcssa7585 = $__nd$0$i$i48;
      label = 51;
      break;
     } else {
      $$sink$i$i49 = $123;
     }
    }
    $__nd$0$i$i48 = $$sink$i$i49;
   }
   if ((label|0) == 47) {
    HEAP32[$__parent$i46>>2] = $__nd$0$i$i48$lcssa7586; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $120 = $__nd$0$i$i48$lcssa7586; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i54 = $$lcssa7588;$136 = $120;
    label = 54;
    break;
   }
   else if ((label|0) == 51) {
    HEAP32[$__parent$i46>>2] = $__nd$0$i$i48$lcssa7585; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $125 = $__nd$0$i$i48$lcssa7585; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i54 = $$lcssa7587;$136 = $125;
    label = 54;
    break;
   }
   else if ((label|0) == 53) {
    HEAP32[$__parent$i46>>2] = $__nd$0$i$i48$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $127 = $__nd$0$i$i48$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $128 = ($__nd$0$i$i48$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($128) {
     $$0$i7$i54 = $__parent$i46;$136 = $127;
     label = 54;
     break;
    } else {
     $147 = $114;$__r$0$i58 = $__nd$0$i$i48$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 54) {
  __THREW__ = 0;
  $129 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $130 = __THREW__; __THREW__ = 0;
  $131 = $130&1;
  if ($131) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $132 = ((($129)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$132>>2] = 9; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $133 = ((($129)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$133>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$129>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $134 = ((($129)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$134>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $135 = ((($129)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$135>>2] = $136; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i54>>2] = $129; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $137 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $138 = HEAP32[$137>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $139 = ($138|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($139) {
   $142 = $129;
  } else {
   $140 = $138; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $140; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i56 = HEAP32[$$0$i7$i54>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $142 = $$pre$i$i56;
  }
  $141 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($141,$142); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $143 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $144 = (($143) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $144; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4380 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $147 = $$pre4380;$__r$0$i58 = $129;
 }
 $145 = ((($__r$0$i58)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$145>>2] = 4; //@line 615 "modules/event/sdl/Event.cpp"
 $146 = ($147|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($146) {
   HEAP32[$__parent$i61>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $159 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i69 = $2;$169 = $159;
   label = 68;
  } else {
   $__nd$0$i$i63 = $147;
   while(1) {
    $148 = ((($__nd$0$i$i63)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $149 = HEAP32[$148>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $150 = ($149|0)>(32); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($150) {
     $151 = HEAP32[$__nd$0$i$i63>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $152 = ($151|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($152) {
      $$lcssa7584 = $__nd$0$i$i63;$__nd$0$i$i63$lcssa7582 = $__nd$0$i$i63;
      label = 61;
      break;
     } else {
      $$sink$i$i64 = $151;
     }
    } else {
     $154 = ($149|0)<(32); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($154)) {
      $__nd$0$i$i63$lcssa = $__nd$0$i$i63;
      label = 67;
      break;
     }
     $155 = ((($__nd$0$i$i63)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $156 = HEAP32[$155>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $157 = ($156|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($157) {
      $$lcssa7583 = $155;$__nd$0$i$i63$lcssa7581 = $__nd$0$i$i63;
      label = 65;
      break;
     } else {
      $$sink$i$i64 = $156;
     }
    }
    $__nd$0$i$i63 = $$sink$i$i64;
   }
   if ((label|0) == 61) {
    HEAP32[$__parent$i61>>2] = $__nd$0$i$i63$lcssa7582; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $153 = $__nd$0$i$i63$lcssa7582; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i69 = $$lcssa7584;$169 = $153;
    label = 68;
    break;
   }
   else if ((label|0) == 65) {
    HEAP32[$__parent$i61>>2] = $__nd$0$i$i63$lcssa7581; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $158 = $__nd$0$i$i63$lcssa7581; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i69 = $$lcssa7583;$169 = $158;
    label = 68;
    break;
   }
   else if ((label|0) == 67) {
    HEAP32[$__parent$i61>>2] = $__nd$0$i$i63$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $160 = $__nd$0$i$i63$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $161 = ($__nd$0$i$i63$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($161) {
     $$0$i7$i69 = $__parent$i61;$169 = $160;
     label = 68;
     break;
    } else {
     $180 = $147;$__r$0$i73 = $__nd$0$i$i63$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 68) {
  __THREW__ = 0;
  $162 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $163 = __THREW__; __THREW__ = 0;
  $164 = $163&1;
  if ($164) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $165 = ((($162)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$165>>2] = 32; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $166 = ((($162)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$166>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$162>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $167 = ((($162)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$167>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $168 = ((($162)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$168>>2] = $169; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i69>>2] = $162; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $170 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $171 = HEAP32[$170>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $172 = ($171|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($172) {
   $175 = $162;
  } else {
   $173 = $171; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $173; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i71 = HEAP32[$$0$i7$i69>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $175 = $$pre$i$i71;
  }
  $174 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($174,$175); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $176 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $177 = (($176) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $177; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4381 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $180 = $$pre4381;$__r$0$i73 = $162;
 }
 $178 = ((($__r$0$i73)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$178>>2] = 5; //@line 616 "modules/event/sdl/Event.cpp"
 $179 = ($180|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($179) {
   HEAP32[$__parent$i76>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $192 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i84 = $2;$202 = $192;
   label = 82;
  } else {
   $__nd$0$i$i78 = $180;
   while(1) {
    $181 = ((($__nd$0$i$i78)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $182 = HEAP32[$181>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $183 = ($182|0)>(33); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($183) {
     $184 = HEAP32[$__nd$0$i$i78>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $185 = ($184|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($185) {
      $$lcssa7580 = $__nd$0$i$i78;$__nd$0$i$i78$lcssa7578 = $__nd$0$i$i78;
      label = 75;
      break;
     } else {
      $$sink$i$i79 = $184;
     }
    } else {
     $187 = ($182|0)<(33); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($187)) {
      $__nd$0$i$i78$lcssa = $__nd$0$i$i78;
      label = 81;
      break;
     }
     $188 = ((($__nd$0$i$i78)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $189 = HEAP32[$188>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $190 = ($189|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($190) {
      $$lcssa7579 = $188;$__nd$0$i$i78$lcssa7577 = $__nd$0$i$i78;
      label = 79;
      break;
     } else {
      $$sink$i$i79 = $189;
     }
    }
    $__nd$0$i$i78 = $$sink$i$i79;
   }
   if ((label|0) == 75) {
    HEAP32[$__parent$i76>>2] = $__nd$0$i$i78$lcssa7578; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $186 = $__nd$0$i$i78$lcssa7578; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i84 = $$lcssa7580;$202 = $186;
    label = 82;
    break;
   }
   else if ((label|0) == 79) {
    HEAP32[$__parent$i76>>2] = $__nd$0$i$i78$lcssa7577; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $191 = $__nd$0$i$i78$lcssa7577; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i84 = $$lcssa7579;$202 = $191;
    label = 82;
    break;
   }
   else if ((label|0) == 81) {
    HEAP32[$__parent$i76>>2] = $__nd$0$i$i78$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $193 = $__nd$0$i$i78$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $194 = ($__nd$0$i$i78$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($194) {
     $$0$i7$i84 = $__parent$i76;$202 = $193;
     label = 82;
     break;
    } else {
     $213 = $180;$__r$0$i88 = $__nd$0$i$i78$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 82) {
  __THREW__ = 0;
  $195 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $196 = __THREW__; __THREW__ = 0;
  $197 = $196&1;
  if ($197) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $198 = ((($195)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$198>>2] = 33; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $199 = ((($195)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$199>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$195>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $200 = ((($195)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$200>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $201 = ((($195)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$201>>2] = $202; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i84>>2] = $195; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $203 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $204 = HEAP32[$203>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $205 = ($204|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($205) {
   $208 = $195;
  } else {
   $206 = $204; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $206; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i86 = HEAP32[$$0$i7$i84>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $208 = $$pre$i$i86;
  }
  $207 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($207,$208); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $209 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $210 = (($209) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $210; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4382 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $213 = $$pre4382;$__r$0$i88 = $195;
 }
 $211 = ((($__r$0$i88)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$211>>2] = 6; //@line 617 "modules/event/sdl/Event.cpp"
 $212 = ($213|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($212) {
   HEAP32[$__parent$i91>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $225 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i99 = $2;$235 = $225;
   label = 96;
  } else {
   $__nd$0$i$i93 = $213;
   while(1) {
    $214 = ((($__nd$0$i$i93)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $215 = HEAP32[$214>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $216 = ($215|0)>(34); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($216) {
     $217 = HEAP32[$__nd$0$i$i93>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $218 = ($217|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($218) {
      $$lcssa7576 = $__nd$0$i$i93;$__nd$0$i$i93$lcssa7574 = $__nd$0$i$i93;
      label = 89;
      break;
     } else {
      $$sink$i$i94 = $217;
     }
    } else {
     $220 = ($215|0)<(34); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($220)) {
      $__nd$0$i$i93$lcssa = $__nd$0$i$i93;
      label = 95;
      break;
     }
     $221 = ((($__nd$0$i$i93)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $222 = HEAP32[$221>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $223 = ($222|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($223) {
      $$lcssa7575 = $221;$__nd$0$i$i93$lcssa7573 = $__nd$0$i$i93;
      label = 93;
      break;
     } else {
      $$sink$i$i94 = $222;
     }
    }
    $__nd$0$i$i93 = $$sink$i$i94;
   }
   if ((label|0) == 89) {
    HEAP32[$__parent$i91>>2] = $__nd$0$i$i93$lcssa7574; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $219 = $__nd$0$i$i93$lcssa7574; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i99 = $$lcssa7576;$235 = $219;
    label = 96;
    break;
   }
   else if ((label|0) == 93) {
    HEAP32[$__parent$i91>>2] = $__nd$0$i$i93$lcssa7573; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $224 = $__nd$0$i$i93$lcssa7573; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i99 = $$lcssa7575;$235 = $224;
    label = 96;
    break;
   }
   else if ((label|0) == 95) {
    HEAP32[$__parent$i91>>2] = $__nd$0$i$i93$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $226 = $__nd$0$i$i93$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $227 = ($__nd$0$i$i93$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($227) {
     $$0$i7$i99 = $__parent$i91;$235 = $226;
     label = 96;
     break;
    } else {
     $246 = $213;$__r$0$i103 = $__nd$0$i$i93$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 96) {
  __THREW__ = 0;
  $228 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $229 = __THREW__; __THREW__ = 0;
  $230 = $229&1;
  if ($230) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $231 = ((($228)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$231>>2] = 34; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $232 = ((($228)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$232>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$228>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $233 = ((($228)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$233>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $234 = ((($228)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$234>>2] = $235; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i99>>2] = $228; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $236 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $237 = HEAP32[$236>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $238 = ($237|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($238) {
   $241 = $228;
  } else {
   $239 = $237; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $239; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i101 = HEAP32[$$0$i7$i99>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $241 = $$pre$i$i101;
  }
  $240 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($240,$241); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $242 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $243 = (($242) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $243; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4383 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $246 = $$pre4383;$__r$0$i103 = $228;
 }
 $244 = ((($__r$0$i103)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$244>>2] = 7; //@line 618 "modules/event/sdl/Event.cpp"
 $245 = ($246|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($245) {
   HEAP32[$__parent$i106>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $258 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i114 = $2;$268 = $258;
   label = 110;
  } else {
   $__nd$0$i$i108 = $246;
   while(1) {
    $247 = ((($__nd$0$i$i108)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $248 = HEAP32[$247>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $249 = ($248|0)>(35); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($249) {
     $250 = HEAP32[$__nd$0$i$i108>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $251 = ($250|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($251) {
      $$lcssa7572 = $__nd$0$i$i108;$__nd$0$i$i108$lcssa7570 = $__nd$0$i$i108;
      label = 103;
      break;
     } else {
      $$sink$i$i109 = $250;
     }
    } else {
     $253 = ($248|0)<(35); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($253)) {
      $__nd$0$i$i108$lcssa = $__nd$0$i$i108;
      label = 109;
      break;
     }
     $254 = ((($__nd$0$i$i108)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $255 = HEAP32[$254>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $256 = ($255|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($256) {
      $$lcssa7571 = $254;$__nd$0$i$i108$lcssa7569 = $__nd$0$i$i108;
      label = 107;
      break;
     } else {
      $$sink$i$i109 = $255;
     }
    }
    $__nd$0$i$i108 = $$sink$i$i109;
   }
   if ((label|0) == 103) {
    HEAP32[$__parent$i106>>2] = $__nd$0$i$i108$lcssa7570; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $252 = $__nd$0$i$i108$lcssa7570; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i114 = $$lcssa7572;$268 = $252;
    label = 110;
    break;
   }
   else if ((label|0) == 107) {
    HEAP32[$__parent$i106>>2] = $__nd$0$i$i108$lcssa7569; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $257 = $__nd$0$i$i108$lcssa7569; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i114 = $$lcssa7571;$268 = $257;
    label = 110;
    break;
   }
   else if ((label|0) == 109) {
    HEAP32[$__parent$i106>>2] = $__nd$0$i$i108$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $259 = $__nd$0$i$i108$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $260 = ($__nd$0$i$i108$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($260) {
     $$0$i7$i114 = $__parent$i106;$268 = $259;
     label = 110;
     break;
    } else {
     $279 = $246;$__r$0$i118 = $__nd$0$i$i108$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 110) {
  __THREW__ = 0;
  $261 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $262 = __THREW__; __THREW__ = 0;
  $263 = $262&1;
  if ($263) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $264 = ((($261)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$264>>2] = 35; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $265 = ((($261)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$265>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$261>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $266 = ((($261)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$266>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $267 = ((($261)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$267>>2] = $268; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i114>>2] = $261; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $269 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $270 = HEAP32[$269>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $271 = ($270|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($271) {
   $274 = $261;
  } else {
   $272 = $270; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $272; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i116 = HEAP32[$$0$i7$i114>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $274 = $$pre$i$i116;
  }
  $273 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($273,$274); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $275 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $276 = (($275) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $276; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4384 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $279 = $$pre4384;$__r$0$i118 = $261;
 }
 $277 = ((($__r$0$i118)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$277>>2] = 8; //@line 619 "modules/event/sdl/Event.cpp"
 $278 = ($279|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($278) {
   HEAP32[$__parent$i121>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $291 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i129 = $2;$301 = $291;
   label = 124;
  } else {
   $__nd$0$i$i123 = $279;
   while(1) {
    $280 = ((($__nd$0$i$i123)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $281 = HEAP32[$280>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $282 = ($281|0)>(37); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($282) {
     $283 = HEAP32[$__nd$0$i$i123>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $284 = ($283|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($284) {
      $$lcssa7568 = $__nd$0$i$i123;$__nd$0$i$i123$lcssa7566 = $__nd$0$i$i123;
      label = 117;
      break;
     } else {
      $$sink$i$i124 = $283;
     }
    } else {
     $286 = ($281|0)<(37); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($286)) {
      $__nd$0$i$i123$lcssa = $__nd$0$i$i123;
      label = 123;
      break;
     }
     $287 = ((($__nd$0$i$i123)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $288 = HEAP32[$287>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $289 = ($288|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($289) {
      $$lcssa7567 = $287;$__nd$0$i$i123$lcssa7565 = $__nd$0$i$i123;
      label = 121;
      break;
     } else {
      $$sink$i$i124 = $288;
     }
    }
    $__nd$0$i$i123 = $$sink$i$i124;
   }
   if ((label|0) == 117) {
    HEAP32[$__parent$i121>>2] = $__nd$0$i$i123$lcssa7566; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $285 = $__nd$0$i$i123$lcssa7566; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i129 = $$lcssa7568;$301 = $285;
    label = 124;
    break;
   }
   else if ((label|0) == 121) {
    HEAP32[$__parent$i121>>2] = $__nd$0$i$i123$lcssa7565; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $290 = $__nd$0$i$i123$lcssa7565; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i129 = $$lcssa7567;$301 = $290;
    label = 124;
    break;
   }
   else if ((label|0) == 123) {
    HEAP32[$__parent$i121>>2] = $__nd$0$i$i123$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $292 = $__nd$0$i$i123$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $293 = ($__nd$0$i$i123$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($293) {
     $$0$i7$i129 = $__parent$i121;$301 = $292;
     label = 124;
     break;
    } else {
     $312 = $279;$__r$0$i133 = $__nd$0$i$i123$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 124) {
  __THREW__ = 0;
  $294 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $295 = __THREW__; __THREW__ = 0;
  $296 = $295&1;
  if ($296) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $297 = ((($294)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$297>>2] = 37; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $298 = ((($294)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$298>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$294>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $299 = ((($294)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$299>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $300 = ((($294)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$300>>2] = $301; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i129>>2] = $294; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $302 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $303 = HEAP32[$302>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $304 = ($303|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($304) {
   $307 = $294;
  } else {
   $305 = $303; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $305; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i131 = HEAP32[$$0$i7$i129>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $307 = $$pre$i$i131;
  }
  $306 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($306,$307); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $308 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $309 = (($308) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $309; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4385 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $312 = $$pre4385;$__r$0$i133 = $294;
 }
 $310 = ((($__r$0$i133)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$310>>2] = 9; //@line 620 "modules/event/sdl/Event.cpp"
 $311 = ($312|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($311) {
   HEAP32[$__parent$i136>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $324 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i144 = $2;$334 = $324;
   label = 138;
  } else {
   $__nd$0$i$i138 = $312;
   while(1) {
    $313 = ((($__nd$0$i$i138)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $314 = HEAP32[$313>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $315 = ($314|0)>(36); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($315) {
     $316 = HEAP32[$__nd$0$i$i138>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $317 = ($316|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($317) {
      $$lcssa7564 = $__nd$0$i$i138;$__nd$0$i$i138$lcssa7562 = $__nd$0$i$i138;
      label = 131;
      break;
     } else {
      $$sink$i$i139 = $316;
     }
    } else {
     $319 = ($314|0)<(36); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($319)) {
      $__nd$0$i$i138$lcssa = $__nd$0$i$i138;
      label = 137;
      break;
     }
     $320 = ((($__nd$0$i$i138)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $321 = HEAP32[$320>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $322 = ($321|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($322) {
      $$lcssa7563 = $320;$__nd$0$i$i138$lcssa7561 = $__nd$0$i$i138;
      label = 135;
      break;
     } else {
      $$sink$i$i139 = $321;
     }
    }
    $__nd$0$i$i138 = $$sink$i$i139;
   }
   if ((label|0) == 131) {
    HEAP32[$__parent$i136>>2] = $__nd$0$i$i138$lcssa7562; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $318 = $__nd$0$i$i138$lcssa7562; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i144 = $$lcssa7564;$334 = $318;
    label = 138;
    break;
   }
   else if ((label|0) == 135) {
    HEAP32[$__parent$i136>>2] = $__nd$0$i$i138$lcssa7561; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $323 = $__nd$0$i$i138$lcssa7561; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i144 = $$lcssa7563;$334 = $323;
    label = 138;
    break;
   }
   else if ((label|0) == 137) {
    HEAP32[$__parent$i136>>2] = $__nd$0$i$i138$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $325 = $__nd$0$i$i138$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $326 = ($__nd$0$i$i138$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($326) {
     $$0$i7$i144 = $__parent$i136;$334 = $325;
     label = 138;
     break;
    } else {
     $345 = $312;$__r$0$i148 = $__nd$0$i$i138$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 138) {
  __THREW__ = 0;
  $327 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $328 = __THREW__; __THREW__ = 0;
  $329 = $328&1;
  if ($329) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $330 = ((($327)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$330>>2] = 36; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $331 = ((($327)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$331>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$327>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $332 = ((($327)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$332>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $333 = ((($327)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$333>>2] = $334; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i144>>2] = $327; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $335 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $336 = HEAP32[$335>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $337 = ($336|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($337) {
   $340 = $327;
  } else {
   $338 = $336; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $338; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i146 = HEAP32[$$0$i7$i144>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $340 = $$pre$i$i146;
  }
  $339 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($339,$340); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $341 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $342 = (($341) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $342; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4386 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $345 = $$pre4386;$__r$0$i148 = $327;
 }
 $343 = ((($__r$0$i148)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$343>>2] = 10; //@line 621 "modules/event/sdl/Event.cpp"
 $344 = ($345|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($344) {
   HEAP32[$__parent$i151>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $357 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i159 = $2;$367 = $357;
   label = 152;
  } else {
   $__nd$0$i$i153 = $345;
   while(1) {
    $346 = ((($__nd$0$i$i153)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $347 = HEAP32[$346>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $348 = ($347|0)>(38); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($348) {
     $349 = HEAP32[$__nd$0$i$i153>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $350 = ($349|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($350) {
      $$lcssa7560 = $__nd$0$i$i153;$__nd$0$i$i153$lcssa7558 = $__nd$0$i$i153;
      label = 145;
      break;
     } else {
      $$sink$i$i154 = $349;
     }
    } else {
     $352 = ($347|0)<(38); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($352)) {
      $__nd$0$i$i153$lcssa = $__nd$0$i$i153;
      label = 151;
      break;
     }
     $353 = ((($__nd$0$i$i153)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $354 = HEAP32[$353>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $355 = ($354|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($355) {
      $$lcssa7559 = $353;$__nd$0$i$i153$lcssa7557 = $__nd$0$i$i153;
      label = 149;
      break;
     } else {
      $$sink$i$i154 = $354;
     }
    }
    $__nd$0$i$i153 = $$sink$i$i154;
   }
   if ((label|0) == 145) {
    HEAP32[$__parent$i151>>2] = $__nd$0$i$i153$lcssa7558; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $351 = $__nd$0$i$i153$lcssa7558; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i159 = $$lcssa7560;$367 = $351;
    label = 152;
    break;
   }
   else if ((label|0) == 149) {
    HEAP32[$__parent$i151>>2] = $__nd$0$i$i153$lcssa7557; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $356 = $__nd$0$i$i153$lcssa7557; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i159 = $$lcssa7559;$367 = $356;
    label = 152;
    break;
   }
   else if ((label|0) == 151) {
    HEAP32[$__parent$i151>>2] = $__nd$0$i$i153$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $358 = $__nd$0$i$i153$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $359 = ($__nd$0$i$i153$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($359) {
     $$0$i7$i159 = $__parent$i151;$367 = $358;
     label = 152;
     break;
    } else {
     $378 = $345;$__r$0$i163 = $__nd$0$i$i153$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 152) {
  __THREW__ = 0;
  $360 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $361 = __THREW__; __THREW__ = 0;
  $362 = $361&1;
  if ($362) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $363 = ((($360)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$363>>2] = 38; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $364 = ((($360)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$364>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$360>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $365 = ((($360)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$365>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $366 = ((($360)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$366>>2] = $367; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i159>>2] = $360; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $368 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $369 = HEAP32[$368>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $370 = ($369|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($370) {
   $373 = $360;
  } else {
   $371 = $369; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $371; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i161 = HEAP32[$$0$i7$i159>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $373 = $$pre$i$i161;
  }
  $372 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($372,$373); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $374 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $375 = (($374) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $375; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4387 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $378 = $$pre4387;$__r$0$i163 = $360;
 }
 $376 = ((($__r$0$i163)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$376>>2] = 11; //@line 622 "modules/event/sdl/Event.cpp"
 $377 = ($378|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($377) {
   HEAP32[$__parent$i166>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $390 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i174 = $2;$400 = $390;
   label = 166;
  } else {
   $__nd$0$i$i168 = $378;
   while(1) {
    $379 = ((($__nd$0$i$i168)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $380 = HEAP32[$379>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $381 = ($380|0)>(39); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($381) {
     $382 = HEAP32[$__nd$0$i$i168>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $383 = ($382|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($383) {
      $$lcssa7556 = $__nd$0$i$i168;$__nd$0$i$i168$lcssa7554 = $__nd$0$i$i168;
      label = 159;
      break;
     } else {
      $$sink$i$i169 = $382;
     }
    } else {
     $385 = ($380|0)<(39); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($385)) {
      $__nd$0$i$i168$lcssa = $__nd$0$i$i168;
      label = 165;
      break;
     }
     $386 = ((($__nd$0$i$i168)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $387 = HEAP32[$386>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $388 = ($387|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($388) {
      $$lcssa7555 = $386;$__nd$0$i$i168$lcssa7553 = $__nd$0$i$i168;
      label = 163;
      break;
     } else {
      $$sink$i$i169 = $387;
     }
    }
    $__nd$0$i$i168 = $$sink$i$i169;
   }
   if ((label|0) == 159) {
    HEAP32[$__parent$i166>>2] = $__nd$0$i$i168$lcssa7554; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $384 = $__nd$0$i$i168$lcssa7554; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i174 = $$lcssa7556;$400 = $384;
    label = 166;
    break;
   }
   else if ((label|0) == 163) {
    HEAP32[$__parent$i166>>2] = $__nd$0$i$i168$lcssa7553; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $389 = $__nd$0$i$i168$lcssa7553; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i174 = $$lcssa7555;$400 = $389;
    label = 166;
    break;
   }
   else if ((label|0) == 165) {
    HEAP32[$__parent$i166>>2] = $__nd$0$i$i168$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $391 = $__nd$0$i$i168$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $392 = ($__nd$0$i$i168$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($392) {
     $$0$i7$i174 = $__parent$i166;$400 = $391;
     label = 166;
     break;
    } else {
     $411 = $378;$__r$0$i178 = $__nd$0$i$i168$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 166) {
  __THREW__ = 0;
  $393 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $394 = __THREW__; __THREW__ = 0;
  $395 = $394&1;
  if ($395) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $396 = ((($393)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$396>>2] = 39; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $397 = ((($393)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$397>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$393>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $398 = ((($393)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$398>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $399 = ((($393)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$399>>2] = $400; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i174>>2] = $393; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $401 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $402 = HEAP32[$401>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $403 = ($402|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($403) {
   $406 = $393;
  } else {
   $404 = $402; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $404; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i176 = HEAP32[$$0$i7$i174>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $406 = $$pre$i$i176;
  }
  $405 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($405,$406); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $407 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $408 = (($407) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $408; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4388 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $411 = $$pre4388;$__r$0$i178 = $393;
 }
 $409 = ((($__r$0$i178)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$409>>2] = 12; //@line 623 "modules/event/sdl/Event.cpp"
 $410 = ($411|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($410) {
   HEAP32[$__parent$i181>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $423 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i189 = $2;$433 = $423;
   label = 180;
  } else {
   $__nd$0$i$i183 = $411;
   while(1) {
    $412 = ((($__nd$0$i$i183)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $413 = HEAP32[$412>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $414 = ($413|0)>(40); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($414) {
     $415 = HEAP32[$__nd$0$i$i183>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $416 = ($415|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($416) {
      $$lcssa7552 = $__nd$0$i$i183;$__nd$0$i$i183$lcssa7550 = $__nd$0$i$i183;
      label = 173;
      break;
     } else {
      $$sink$i$i184 = $415;
     }
    } else {
     $418 = ($413|0)<(40); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($418)) {
      $__nd$0$i$i183$lcssa = $__nd$0$i$i183;
      label = 179;
      break;
     }
     $419 = ((($__nd$0$i$i183)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $420 = HEAP32[$419>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $421 = ($420|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($421) {
      $$lcssa7551 = $419;$__nd$0$i$i183$lcssa7549 = $__nd$0$i$i183;
      label = 177;
      break;
     } else {
      $$sink$i$i184 = $420;
     }
    }
    $__nd$0$i$i183 = $$sink$i$i184;
   }
   if ((label|0) == 173) {
    HEAP32[$__parent$i181>>2] = $__nd$0$i$i183$lcssa7550; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $417 = $__nd$0$i$i183$lcssa7550; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i189 = $$lcssa7552;$433 = $417;
    label = 180;
    break;
   }
   else if ((label|0) == 177) {
    HEAP32[$__parent$i181>>2] = $__nd$0$i$i183$lcssa7549; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $422 = $__nd$0$i$i183$lcssa7549; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i189 = $$lcssa7551;$433 = $422;
    label = 180;
    break;
   }
   else if ((label|0) == 179) {
    HEAP32[$__parent$i181>>2] = $__nd$0$i$i183$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $424 = $__nd$0$i$i183$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $425 = ($__nd$0$i$i183$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($425) {
     $$0$i7$i189 = $__parent$i181;$433 = $424;
     label = 180;
     break;
    } else {
     $444 = $411;$__r$0$i193 = $__nd$0$i$i183$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 180) {
  __THREW__ = 0;
  $426 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $427 = __THREW__; __THREW__ = 0;
  $428 = $427&1;
  if ($428) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $429 = ((($426)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$429>>2] = 40; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $430 = ((($426)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$430>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$426>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $431 = ((($426)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$431>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $432 = ((($426)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$432>>2] = $433; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i189>>2] = $426; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $434 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $435 = HEAP32[$434>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $436 = ($435|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($436) {
   $439 = $426;
  } else {
   $437 = $435; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $437; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i191 = HEAP32[$$0$i7$i189>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $439 = $$pre$i$i191;
  }
  $438 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($438,$439); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $440 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $441 = (($440) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $441; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4389 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $444 = $$pre4389;$__r$0$i193 = $426;
 }
 $442 = ((($__r$0$i193)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$442>>2] = 13; //@line 624 "modules/event/sdl/Event.cpp"
 $443 = ($444|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($443) {
   HEAP32[$__parent$i196>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $456 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i204 = $2;$466 = $456;
   label = 194;
  } else {
   $__nd$0$i$i198 = $444;
   while(1) {
    $445 = ((($__nd$0$i$i198)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $446 = HEAP32[$445>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $447 = ($446|0)>(41); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($447) {
     $448 = HEAP32[$__nd$0$i$i198>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $449 = ($448|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($449) {
      $$lcssa7548 = $__nd$0$i$i198;$__nd$0$i$i198$lcssa7546 = $__nd$0$i$i198;
      label = 187;
      break;
     } else {
      $$sink$i$i199 = $448;
     }
    } else {
     $451 = ($446|0)<(41); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($451)) {
      $__nd$0$i$i198$lcssa = $__nd$0$i$i198;
      label = 193;
      break;
     }
     $452 = ((($__nd$0$i$i198)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $453 = HEAP32[$452>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $454 = ($453|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($454) {
      $$lcssa7547 = $452;$__nd$0$i$i198$lcssa7545 = $__nd$0$i$i198;
      label = 191;
      break;
     } else {
      $$sink$i$i199 = $453;
     }
    }
    $__nd$0$i$i198 = $$sink$i$i199;
   }
   if ((label|0) == 187) {
    HEAP32[$__parent$i196>>2] = $__nd$0$i$i198$lcssa7546; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $450 = $__nd$0$i$i198$lcssa7546; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i204 = $$lcssa7548;$466 = $450;
    label = 194;
    break;
   }
   else if ((label|0) == 191) {
    HEAP32[$__parent$i196>>2] = $__nd$0$i$i198$lcssa7545; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $455 = $__nd$0$i$i198$lcssa7545; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i204 = $$lcssa7547;$466 = $455;
    label = 194;
    break;
   }
   else if ((label|0) == 193) {
    HEAP32[$__parent$i196>>2] = $__nd$0$i$i198$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $457 = $__nd$0$i$i198$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $458 = ($__nd$0$i$i198$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($458) {
     $$0$i7$i204 = $__parent$i196;$466 = $457;
     label = 194;
     break;
    } else {
     $477 = $444;$__r$0$i208 = $__nd$0$i$i198$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 194) {
  __THREW__ = 0;
  $459 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $460 = __THREW__; __THREW__ = 0;
  $461 = $460&1;
  if ($461) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $462 = ((($459)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$462>>2] = 41; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $463 = ((($459)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$463>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$459>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $464 = ((($459)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$464>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $465 = ((($459)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$465>>2] = $466; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i204>>2] = $459; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $467 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $468 = HEAP32[$467>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $469 = ($468|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($469) {
   $472 = $459;
  } else {
   $470 = $468; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $470; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i206 = HEAP32[$$0$i7$i204>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $472 = $$pre$i$i206;
  }
  $471 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($471,$472); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $473 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $474 = (($473) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $474; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4390 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $477 = $$pre4390;$__r$0$i208 = $459;
 }
 $475 = ((($__r$0$i208)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$475>>2] = 14; //@line 625 "modules/event/sdl/Event.cpp"
 $476 = ($477|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($476) {
   HEAP32[$__parent$i211>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $489 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i219 = $2;$499 = $489;
   label = 208;
  } else {
   $__nd$0$i$i213 = $477;
   while(1) {
    $478 = ((($__nd$0$i$i213)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $479 = HEAP32[$478>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $480 = ($479|0)>(42); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($480) {
     $481 = HEAP32[$__nd$0$i$i213>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $482 = ($481|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($482) {
      $$lcssa7544 = $__nd$0$i$i213;$__nd$0$i$i213$lcssa7542 = $__nd$0$i$i213;
      label = 201;
      break;
     } else {
      $$sink$i$i214 = $481;
     }
    } else {
     $484 = ($479|0)<(42); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($484)) {
      $__nd$0$i$i213$lcssa = $__nd$0$i$i213;
      label = 207;
      break;
     }
     $485 = ((($__nd$0$i$i213)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $486 = HEAP32[$485>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $487 = ($486|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($487) {
      $$lcssa7543 = $485;$__nd$0$i$i213$lcssa7541 = $__nd$0$i$i213;
      label = 205;
      break;
     } else {
      $$sink$i$i214 = $486;
     }
    }
    $__nd$0$i$i213 = $$sink$i$i214;
   }
   if ((label|0) == 201) {
    HEAP32[$__parent$i211>>2] = $__nd$0$i$i213$lcssa7542; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $483 = $__nd$0$i$i213$lcssa7542; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i219 = $$lcssa7544;$499 = $483;
    label = 208;
    break;
   }
   else if ((label|0) == 205) {
    HEAP32[$__parent$i211>>2] = $__nd$0$i$i213$lcssa7541; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $488 = $__nd$0$i$i213$lcssa7541; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i219 = $$lcssa7543;$499 = $488;
    label = 208;
    break;
   }
   else if ((label|0) == 207) {
    HEAP32[$__parent$i211>>2] = $__nd$0$i$i213$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $490 = $__nd$0$i$i213$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $491 = ($__nd$0$i$i213$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($491) {
     $$0$i7$i219 = $__parent$i211;$499 = $490;
     label = 208;
     break;
    } else {
     $510 = $477;$__r$0$i223 = $__nd$0$i$i213$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 208) {
  __THREW__ = 0;
  $492 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $493 = __THREW__; __THREW__ = 0;
  $494 = $493&1;
  if ($494) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $495 = ((($492)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$495>>2] = 42; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $496 = ((($492)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$496>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$492>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $497 = ((($492)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$497>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $498 = ((($492)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$498>>2] = $499; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i219>>2] = $492; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $500 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $501 = HEAP32[$500>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $502 = ($501|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($502) {
   $505 = $492;
  } else {
   $503 = $501; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $503; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i221 = HEAP32[$$0$i7$i219>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $505 = $$pre$i$i221;
  }
  $504 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($504,$505); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $506 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $507 = (($506) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $507; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4391 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $510 = $$pre4391;$__r$0$i223 = $492;
 }
 $508 = ((($__r$0$i223)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$508>>2] = 15; //@line 626 "modules/event/sdl/Event.cpp"
 $509 = ($510|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($509) {
   HEAP32[$__parent$i226>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $522 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i234 = $2;$532 = $522;
   label = 222;
  } else {
   $__nd$0$i$i228 = $510;
   while(1) {
    $511 = ((($__nd$0$i$i228)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $512 = HEAP32[$511>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $513 = ($512|0)>(43); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($513) {
     $514 = HEAP32[$__nd$0$i$i228>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $515 = ($514|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($515) {
      $$lcssa7540 = $__nd$0$i$i228;$__nd$0$i$i228$lcssa7538 = $__nd$0$i$i228;
      label = 215;
      break;
     } else {
      $$sink$i$i229 = $514;
     }
    } else {
     $517 = ($512|0)<(43); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($517)) {
      $__nd$0$i$i228$lcssa = $__nd$0$i$i228;
      label = 221;
      break;
     }
     $518 = ((($__nd$0$i$i228)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $519 = HEAP32[$518>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $520 = ($519|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($520) {
      $$lcssa7539 = $518;$__nd$0$i$i228$lcssa7537 = $__nd$0$i$i228;
      label = 219;
      break;
     } else {
      $$sink$i$i229 = $519;
     }
    }
    $__nd$0$i$i228 = $$sink$i$i229;
   }
   if ((label|0) == 215) {
    HEAP32[$__parent$i226>>2] = $__nd$0$i$i228$lcssa7538; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $516 = $__nd$0$i$i228$lcssa7538; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i234 = $$lcssa7540;$532 = $516;
    label = 222;
    break;
   }
   else if ((label|0) == 219) {
    HEAP32[$__parent$i226>>2] = $__nd$0$i$i228$lcssa7537; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $521 = $__nd$0$i$i228$lcssa7537; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i234 = $$lcssa7539;$532 = $521;
    label = 222;
    break;
   }
   else if ((label|0) == 221) {
    HEAP32[$__parent$i226>>2] = $__nd$0$i$i228$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $523 = $__nd$0$i$i228$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $524 = ($__nd$0$i$i228$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($524) {
     $$0$i7$i234 = $__parent$i226;$532 = $523;
     label = 222;
     break;
    } else {
     $543 = $510;$__r$0$i238 = $__nd$0$i$i228$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 222) {
  __THREW__ = 0;
  $525 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $526 = __THREW__; __THREW__ = 0;
  $527 = $526&1;
  if ($527) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $528 = ((($525)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$528>>2] = 43; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $529 = ((($525)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$529>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$525>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $530 = ((($525)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$530>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $531 = ((($525)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$531>>2] = $532; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i234>>2] = $525; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $533 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $534 = HEAP32[$533>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $535 = ($534|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($535) {
   $538 = $525;
  } else {
   $536 = $534; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $536; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i236 = HEAP32[$$0$i7$i234>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $538 = $$pre$i$i236;
  }
  $537 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($537,$538); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $539 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $540 = (($539) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $540; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4392 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $543 = $$pre4392;$__r$0$i238 = $525;
 }
 $541 = ((($__r$0$i238)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$541>>2] = 16; //@line 627 "modules/event/sdl/Event.cpp"
 $542 = ($543|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($542) {
   HEAP32[$__parent$i241>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $555 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i249 = $2;$565 = $555;
   label = 236;
  } else {
   $__nd$0$i$i243 = $543;
   while(1) {
    $544 = ((($__nd$0$i$i243)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $545 = HEAP32[$544>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $546 = ($545|0)>(44); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($546) {
     $547 = HEAP32[$__nd$0$i$i243>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $548 = ($547|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($548) {
      $$lcssa7536 = $__nd$0$i$i243;$__nd$0$i$i243$lcssa7534 = $__nd$0$i$i243;
      label = 229;
      break;
     } else {
      $$sink$i$i244 = $547;
     }
    } else {
     $550 = ($545|0)<(44); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($550)) {
      $__nd$0$i$i243$lcssa = $__nd$0$i$i243;
      label = 235;
      break;
     }
     $551 = ((($__nd$0$i$i243)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $552 = HEAP32[$551>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $553 = ($552|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($553) {
      $$lcssa7535 = $551;$__nd$0$i$i243$lcssa7533 = $__nd$0$i$i243;
      label = 233;
      break;
     } else {
      $$sink$i$i244 = $552;
     }
    }
    $__nd$0$i$i243 = $$sink$i$i244;
   }
   if ((label|0) == 229) {
    HEAP32[$__parent$i241>>2] = $__nd$0$i$i243$lcssa7534; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $549 = $__nd$0$i$i243$lcssa7534; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i249 = $$lcssa7536;$565 = $549;
    label = 236;
    break;
   }
   else if ((label|0) == 233) {
    HEAP32[$__parent$i241>>2] = $__nd$0$i$i243$lcssa7533; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $554 = $__nd$0$i$i243$lcssa7533; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i249 = $$lcssa7535;$565 = $554;
    label = 236;
    break;
   }
   else if ((label|0) == 235) {
    HEAP32[$__parent$i241>>2] = $__nd$0$i$i243$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $556 = $__nd$0$i$i243$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $557 = ($__nd$0$i$i243$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($557) {
     $$0$i7$i249 = $__parent$i241;$565 = $556;
     label = 236;
     break;
    } else {
     $576 = $543;$__r$0$i253 = $__nd$0$i$i243$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 236) {
  __THREW__ = 0;
  $558 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $559 = __THREW__; __THREW__ = 0;
  $560 = $559&1;
  if ($560) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $561 = ((($558)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$561>>2] = 44; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $562 = ((($558)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$562>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$558>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $563 = ((($558)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$563>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $564 = ((($558)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$564>>2] = $565; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i249>>2] = $558; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $566 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $567 = HEAP32[$566>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $568 = ($567|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($568) {
   $571 = $558;
  } else {
   $569 = $567; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $569; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i251 = HEAP32[$$0$i7$i249>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $571 = $$pre$i$i251;
  }
  $570 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($570,$571); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $572 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $573 = (($572) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $573; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4393 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $576 = $$pre4393;$__r$0$i253 = $558;
 }
 $574 = ((($__r$0$i253)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$574>>2] = 17; //@line 628 "modules/event/sdl/Event.cpp"
 $575 = ($576|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($575) {
   HEAP32[$__parent$i256>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $588 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i264 = $2;$598 = $588;
   label = 250;
  } else {
   $__nd$0$i$i258 = $576;
   while(1) {
    $577 = ((($__nd$0$i$i258)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $578 = HEAP32[$577>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $579 = ($578|0)>(45); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($579) {
     $580 = HEAP32[$__nd$0$i$i258>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $581 = ($580|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($581) {
      $$lcssa7532 = $__nd$0$i$i258;$__nd$0$i$i258$lcssa7530 = $__nd$0$i$i258;
      label = 243;
      break;
     } else {
      $$sink$i$i259 = $580;
     }
    } else {
     $583 = ($578|0)<(45); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($583)) {
      $__nd$0$i$i258$lcssa = $__nd$0$i$i258;
      label = 249;
      break;
     }
     $584 = ((($__nd$0$i$i258)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $585 = HEAP32[$584>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $586 = ($585|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($586) {
      $$lcssa7531 = $584;$__nd$0$i$i258$lcssa7529 = $__nd$0$i$i258;
      label = 247;
      break;
     } else {
      $$sink$i$i259 = $585;
     }
    }
    $__nd$0$i$i258 = $$sink$i$i259;
   }
   if ((label|0) == 243) {
    HEAP32[$__parent$i256>>2] = $__nd$0$i$i258$lcssa7530; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $582 = $__nd$0$i$i258$lcssa7530; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i264 = $$lcssa7532;$598 = $582;
    label = 250;
    break;
   }
   else if ((label|0) == 247) {
    HEAP32[$__parent$i256>>2] = $__nd$0$i$i258$lcssa7529; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $587 = $__nd$0$i$i258$lcssa7529; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i264 = $$lcssa7531;$598 = $587;
    label = 250;
    break;
   }
   else if ((label|0) == 249) {
    HEAP32[$__parent$i256>>2] = $__nd$0$i$i258$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $589 = $__nd$0$i$i258$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $590 = ($__nd$0$i$i258$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($590) {
     $$0$i7$i264 = $__parent$i256;$598 = $589;
     label = 250;
     break;
    } else {
     $609 = $576;$__r$0$i268 = $__nd$0$i$i258$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 250) {
  __THREW__ = 0;
  $591 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $592 = __THREW__; __THREW__ = 0;
  $593 = $592&1;
  if ($593) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $594 = ((($591)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$594>>2] = 45; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $595 = ((($591)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$595>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$591>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $596 = ((($591)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$596>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $597 = ((($591)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$597>>2] = $598; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i264>>2] = $591; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $599 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $600 = HEAP32[$599>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $601 = ($600|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($601) {
   $604 = $591;
  } else {
   $602 = $600; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $602; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i266 = HEAP32[$$0$i7$i264>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $604 = $$pre$i$i266;
  }
  $603 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($603,$604); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $605 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $606 = (($605) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $606; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4394 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $609 = $$pre4394;$__r$0$i268 = $591;
 }
 $607 = ((($__r$0$i268)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$607>>2] = 18; //@line 629 "modules/event/sdl/Event.cpp"
 $608 = ($609|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($608) {
   HEAP32[$__parent$i271>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $621 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i279 = $2;$631 = $621;
   label = 264;
  } else {
   $__nd$0$i$i273 = $609;
   while(1) {
    $610 = ((($__nd$0$i$i273)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $611 = HEAP32[$610>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $612 = ($611|0)>(46); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($612) {
     $613 = HEAP32[$__nd$0$i$i273>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $614 = ($613|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($614) {
      $$lcssa7528 = $__nd$0$i$i273;$__nd$0$i$i273$lcssa7526 = $__nd$0$i$i273;
      label = 257;
      break;
     } else {
      $$sink$i$i274 = $613;
     }
    } else {
     $616 = ($611|0)<(46); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($616)) {
      $__nd$0$i$i273$lcssa = $__nd$0$i$i273;
      label = 263;
      break;
     }
     $617 = ((($__nd$0$i$i273)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $618 = HEAP32[$617>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $619 = ($618|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($619) {
      $$lcssa7527 = $617;$__nd$0$i$i273$lcssa7525 = $__nd$0$i$i273;
      label = 261;
      break;
     } else {
      $$sink$i$i274 = $618;
     }
    }
    $__nd$0$i$i273 = $$sink$i$i274;
   }
   if ((label|0) == 257) {
    HEAP32[$__parent$i271>>2] = $__nd$0$i$i273$lcssa7526; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $615 = $__nd$0$i$i273$lcssa7526; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i279 = $$lcssa7528;$631 = $615;
    label = 264;
    break;
   }
   else if ((label|0) == 261) {
    HEAP32[$__parent$i271>>2] = $__nd$0$i$i273$lcssa7525; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $620 = $__nd$0$i$i273$lcssa7525; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i279 = $$lcssa7527;$631 = $620;
    label = 264;
    break;
   }
   else if ((label|0) == 263) {
    HEAP32[$__parent$i271>>2] = $__nd$0$i$i273$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $622 = $__nd$0$i$i273$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $623 = ($__nd$0$i$i273$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($623) {
     $$0$i7$i279 = $__parent$i271;$631 = $622;
     label = 264;
     break;
    } else {
     $642 = $609;$__r$0$i283 = $__nd$0$i$i273$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 264) {
  __THREW__ = 0;
  $624 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $625 = __THREW__; __THREW__ = 0;
  $626 = $625&1;
  if ($626) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $627 = ((($624)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$627>>2] = 46; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $628 = ((($624)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$628>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$624>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $629 = ((($624)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$629>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $630 = ((($624)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$630>>2] = $631; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i279>>2] = $624; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $632 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $633 = HEAP32[$632>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $634 = ($633|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($634) {
   $637 = $624;
  } else {
   $635 = $633; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $635; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i281 = HEAP32[$$0$i7$i279>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $637 = $$pre$i$i281;
  }
  $636 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($636,$637); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $638 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $639 = (($638) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $639; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4395 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $642 = $$pre4395;$__r$0$i283 = $624;
 }
 $640 = ((($__r$0$i283)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$640>>2] = 19; //@line 630 "modules/event/sdl/Event.cpp"
 $641 = ($642|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($641) {
   HEAP32[$__parent$i286>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $654 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i294 = $2;$664 = $654;
   label = 278;
  } else {
   $__nd$0$i$i288 = $642;
   while(1) {
    $643 = ((($__nd$0$i$i288)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $644 = HEAP32[$643>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $645 = ($644|0)>(47); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($645) {
     $646 = HEAP32[$__nd$0$i$i288>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $647 = ($646|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($647) {
      $$lcssa7524 = $__nd$0$i$i288;$__nd$0$i$i288$lcssa7522 = $__nd$0$i$i288;
      label = 271;
      break;
     } else {
      $$sink$i$i289 = $646;
     }
    } else {
     $649 = ($644|0)<(47); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($649)) {
      $__nd$0$i$i288$lcssa = $__nd$0$i$i288;
      label = 277;
      break;
     }
     $650 = ((($__nd$0$i$i288)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $651 = HEAP32[$650>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $652 = ($651|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($652) {
      $$lcssa7523 = $650;$__nd$0$i$i288$lcssa7521 = $__nd$0$i$i288;
      label = 275;
      break;
     } else {
      $$sink$i$i289 = $651;
     }
    }
    $__nd$0$i$i288 = $$sink$i$i289;
   }
   if ((label|0) == 271) {
    HEAP32[$__parent$i286>>2] = $__nd$0$i$i288$lcssa7522; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $648 = $__nd$0$i$i288$lcssa7522; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i294 = $$lcssa7524;$664 = $648;
    label = 278;
    break;
   }
   else if ((label|0) == 275) {
    HEAP32[$__parent$i286>>2] = $__nd$0$i$i288$lcssa7521; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $653 = $__nd$0$i$i288$lcssa7521; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i294 = $$lcssa7523;$664 = $653;
    label = 278;
    break;
   }
   else if ((label|0) == 277) {
    HEAP32[$__parent$i286>>2] = $__nd$0$i$i288$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $655 = $__nd$0$i$i288$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $656 = ($__nd$0$i$i288$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($656) {
     $$0$i7$i294 = $__parent$i286;$664 = $655;
     label = 278;
     break;
    } else {
     $675 = $642;$__r$0$i298 = $__nd$0$i$i288$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 278) {
  __THREW__ = 0;
  $657 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $658 = __THREW__; __THREW__ = 0;
  $659 = $658&1;
  if ($659) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $660 = ((($657)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$660>>2] = 47; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $661 = ((($657)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$661>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$657>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $662 = ((($657)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$662>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $663 = ((($657)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$663>>2] = $664; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i294>>2] = $657; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $665 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $666 = HEAP32[$665>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $667 = ($666|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($667) {
   $670 = $657;
  } else {
   $668 = $666; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $668; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i296 = HEAP32[$$0$i7$i294>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $670 = $$pre$i$i296;
  }
  $669 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($669,$670); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $671 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $672 = (($671) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $672; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4396 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $675 = $$pre4396;$__r$0$i298 = $657;
 }
 $673 = ((($__r$0$i298)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$673>>2] = 20; //@line 631 "modules/event/sdl/Event.cpp"
 $674 = ($675|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($674) {
   HEAP32[$__parent$i301>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $687 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i309 = $2;$697 = $687;
   label = 292;
  } else {
   $__nd$0$i$i303 = $675;
   while(1) {
    $676 = ((($__nd$0$i$i303)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $677 = HEAP32[$676>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $678 = ($677|0)>(48); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($678) {
     $679 = HEAP32[$__nd$0$i$i303>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $680 = ($679|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($680) {
      $$lcssa7520 = $__nd$0$i$i303;$__nd$0$i$i303$lcssa7518 = $__nd$0$i$i303;
      label = 285;
      break;
     } else {
      $$sink$i$i304 = $679;
     }
    } else {
     $682 = ($677|0)<(48); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($682)) {
      $__nd$0$i$i303$lcssa = $__nd$0$i$i303;
      label = 291;
      break;
     }
     $683 = ((($__nd$0$i$i303)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $684 = HEAP32[$683>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $685 = ($684|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($685) {
      $$lcssa7519 = $683;$__nd$0$i$i303$lcssa7517 = $__nd$0$i$i303;
      label = 289;
      break;
     } else {
      $$sink$i$i304 = $684;
     }
    }
    $__nd$0$i$i303 = $$sink$i$i304;
   }
   if ((label|0) == 285) {
    HEAP32[$__parent$i301>>2] = $__nd$0$i$i303$lcssa7518; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $681 = $__nd$0$i$i303$lcssa7518; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i309 = $$lcssa7520;$697 = $681;
    label = 292;
    break;
   }
   else if ((label|0) == 289) {
    HEAP32[$__parent$i301>>2] = $__nd$0$i$i303$lcssa7517; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $686 = $__nd$0$i$i303$lcssa7517; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i309 = $$lcssa7519;$697 = $686;
    label = 292;
    break;
   }
   else if ((label|0) == 291) {
    HEAP32[$__parent$i301>>2] = $__nd$0$i$i303$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $688 = $__nd$0$i$i303$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $689 = ($__nd$0$i$i303$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($689) {
     $$0$i7$i309 = $__parent$i301;$697 = $688;
     label = 292;
     break;
    } else {
     $708 = $675;$__r$0$i313 = $__nd$0$i$i303$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 292) {
  __THREW__ = 0;
  $690 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $691 = __THREW__; __THREW__ = 0;
  $692 = $691&1;
  if ($692) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $693 = ((($690)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$693>>2] = 48; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $694 = ((($690)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$694>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$690>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $695 = ((($690)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$695>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $696 = ((($690)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$696>>2] = $697; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i309>>2] = $690; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $698 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $699 = HEAP32[$698>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $700 = ($699|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($700) {
   $703 = $690;
  } else {
   $701 = $699; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $701; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i311 = HEAP32[$$0$i7$i309>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $703 = $$pre$i$i311;
  }
  $702 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($702,$703); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $704 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $705 = (($704) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $705; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4397 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $708 = $$pre4397;$__r$0$i313 = $690;
 }
 $706 = ((($__r$0$i313)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$706>>2] = 21; //@line 632 "modules/event/sdl/Event.cpp"
 $707 = ($708|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($707) {
   HEAP32[$__parent$i316>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $720 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i324 = $2;$730 = $720;
   label = 306;
  } else {
   $__nd$0$i$i318 = $708;
   while(1) {
    $709 = ((($__nd$0$i$i318)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $710 = HEAP32[$709>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $711 = ($710|0)>(49); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($711) {
     $712 = HEAP32[$__nd$0$i$i318>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $713 = ($712|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($713) {
      $$lcssa7516 = $__nd$0$i$i318;$__nd$0$i$i318$lcssa7514 = $__nd$0$i$i318;
      label = 299;
      break;
     } else {
      $$sink$i$i319 = $712;
     }
    } else {
     $715 = ($710|0)<(49); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($715)) {
      $__nd$0$i$i318$lcssa = $__nd$0$i$i318;
      label = 305;
      break;
     }
     $716 = ((($__nd$0$i$i318)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $717 = HEAP32[$716>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $718 = ($717|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($718) {
      $$lcssa7515 = $716;$__nd$0$i$i318$lcssa7513 = $__nd$0$i$i318;
      label = 303;
      break;
     } else {
      $$sink$i$i319 = $717;
     }
    }
    $__nd$0$i$i318 = $$sink$i$i319;
   }
   if ((label|0) == 299) {
    HEAP32[$__parent$i316>>2] = $__nd$0$i$i318$lcssa7514; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $714 = $__nd$0$i$i318$lcssa7514; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i324 = $$lcssa7516;$730 = $714;
    label = 306;
    break;
   }
   else if ((label|0) == 303) {
    HEAP32[$__parent$i316>>2] = $__nd$0$i$i318$lcssa7513; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $719 = $__nd$0$i$i318$lcssa7513; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i324 = $$lcssa7515;$730 = $719;
    label = 306;
    break;
   }
   else if ((label|0) == 305) {
    HEAP32[$__parent$i316>>2] = $__nd$0$i$i318$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $721 = $__nd$0$i$i318$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $722 = ($__nd$0$i$i318$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($722) {
     $$0$i7$i324 = $__parent$i316;$730 = $721;
     label = 306;
     break;
    } else {
     $741 = $708;$__r$0$i328 = $__nd$0$i$i318$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 306) {
  __THREW__ = 0;
  $723 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $724 = __THREW__; __THREW__ = 0;
  $725 = $724&1;
  if ($725) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $726 = ((($723)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$726>>2] = 49; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $727 = ((($723)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$727>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$723>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $728 = ((($723)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$728>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $729 = ((($723)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$729>>2] = $730; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i324>>2] = $723; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $731 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $732 = HEAP32[$731>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $733 = ($732|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($733) {
   $736 = $723;
  } else {
   $734 = $732; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $734; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i326 = HEAP32[$$0$i7$i324>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $736 = $$pre$i$i326;
  }
  $735 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($735,$736); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $737 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $738 = (($737) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $738; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4398 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $741 = $$pre4398;$__r$0$i328 = $723;
 }
 $739 = ((($__r$0$i328)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$739>>2] = 22; //@line 633 "modules/event/sdl/Event.cpp"
 $740 = ($741|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($740) {
   HEAP32[$__parent$i331>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $753 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i339 = $2;$763 = $753;
   label = 320;
  } else {
   $__nd$0$i$i333 = $741;
   while(1) {
    $742 = ((($__nd$0$i$i333)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $743 = HEAP32[$742>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $744 = ($743|0)>(50); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($744) {
     $745 = HEAP32[$__nd$0$i$i333>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $746 = ($745|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($746) {
      $$lcssa7512 = $__nd$0$i$i333;$__nd$0$i$i333$lcssa7510 = $__nd$0$i$i333;
      label = 313;
      break;
     } else {
      $$sink$i$i334 = $745;
     }
    } else {
     $748 = ($743|0)<(50); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($748)) {
      $__nd$0$i$i333$lcssa = $__nd$0$i$i333;
      label = 319;
      break;
     }
     $749 = ((($__nd$0$i$i333)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $750 = HEAP32[$749>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $751 = ($750|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($751) {
      $$lcssa7511 = $749;$__nd$0$i$i333$lcssa7509 = $__nd$0$i$i333;
      label = 317;
      break;
     } else {
      $$sink$i$i334 = $750;
     }
    }
    $__nd$0$i$i333 = $$sink$i$i334;
   }
   if ((label|0) == 313) {
    HEAP32[$__parent$i331>>2] = $__nd$0$i$i333$lcssa7510; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $747 = $__nd$0$i$i333$lcssa7510; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i339 = $$lcssa7512;$763 = $747;
    label = 320;
    break;
   }
   else if ((label|0) == 317) {
    HEAP32[$__parent$i331>>2] = $__nd$0$i$i333$lcssa7509; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $752 = $__nd$0$i$i333$lcssa7509; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i339 = $$lcssa7511;$763 = $752;
    label = 320;
    break;
   }
   else if ((label|0) == 319) {
    HEAP32[$__parent$i331>>2] = $__nd$0$i$i333$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $754 = $__nd$0$i$i333$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $755 = ($__nd$0$i$i333$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($755) {
     $$0$i7$i339 = $__parent$i331;$763 = $754;
     label = 320;
     break;
    } else {
     $774 = $741;$__r$0$i343 = $__nd$0$i$i333$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 320) {
  __THREW__ = 0;
  $756 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $757 = __THREW__; __THREW__ = 0;
  $758 = $757&1;
  if ($758) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $759 = ((($756)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$759>>2] = 50; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $760 = ((($756)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$760>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$756>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $761 = ((($756)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$761>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $762 = ((($756)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$762>>2] = $763; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i339>>2] = $756; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $764 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $765 = HEAP32[$764>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $766 = ($765|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($766) {
   $769 = $756;
  } else {
   $767 = $765; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $767; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i341 = HEAP32[$$0$i7$i339>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $769 = $$pre$i$i341;
  }
  $768 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($768,$769); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $770 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $771 = (($770) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $771; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4399 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $774 = $$pre4399;$__r$0$i343 = $756;
 }
 $772 = ((($__r$0$i343)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$772>>2] = 23; //@line 634 "modules/event/sdl/Event.cpp"
 $773 = ($774|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($773) {
   HEAP32[$__parent$i346>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $786 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i354 = $2;$796 = $786;
   label = 334;
  } else {
   $__nd$0$i$i348 = $774;
   while(1) {
    $775 = ((($__nd$0$i$i348)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $776 = HEAP32[$775>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $777 = ($776|0)>(51); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($777) {
     $778 = HEAP32[$__nd$0$i$i348>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $779 = ($778|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($779) {
      $$lcssa7508 = $__nd$0$i$i348;$__nd$0$i$i348$lcssa7506 = $__nd$0$i$i348;
      label = 327;
      break;
     } else {
      $$sink$i$i349 = $778;
     }
    } else {
     $781 = ($776|0)<(51); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($781)) {
      $__nd$0$i$i348$lcssa = $__nd$0$i$i348;
      label = 333;
      break;
     }
     $782 = ((($__nd$0$i$i348)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $783 = HEAP32[$782>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $784 = ($783|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($784) {
      $$lcssa7507 = $782;$__nd$0$i$i348$lcssa7505 = $__nd$0$i$i348;
      label = 331;
      break;
     } else {
      $$sink$i$i349 = $783;
     }
    }
    $__nd$0$i$i348 = $$sink$i$i349;
   }
   if ((label|0) == 327) {
    HEAP32[$__parent$i346>>2] = $__nd$0$i$i348$lcssa7506; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $780 = $__nd$0$i$i348$lcssa7506; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i354 = $$lcssa7508;$796 = $780;
    label = 334;
    break;
   }
   else if ((label|0) == 331) {
    HEAP32[$__parent$i346>>2] = $__nd$0$i$i348$lcssa7505; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $785 = $__nd$0$i$i348$lcssa7505; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i354 = $$lcssa7507;$796 = $785;
    label = 334;
    break;
   }
   else if ((label|0) == 333) {
    HEAP32[$__parent$i346>>2] = $__nd$0$i$i348$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $787 = $__nd$0$i$i348$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $788 = ($__nd$0$i$i348$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($788) {
     $$0$i7$i354 = $__parent$i346;$796 = $787;
     label = 334;
     break;
    } else {
     $807 = $774;$__r$0$i358 = $__nd$0$i$i348$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 334) {
  __THREW__ = 0;
  $789 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $790 = __THREW__; __THREW__ = 0;
  $791 = $790&1;
  if ($791) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $792 = ((($789)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$792>>2] = 51; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $793 = ((($789)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$793>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$789>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $794 = ((($789)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$794>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $795 = ((($789)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$795>>2] = $796; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i354>>2] = $789; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $797 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $798 = HEAP32[$797>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $799 = ($798|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($799) {
   $802 = $789;
  } else {
   $800 = $798; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $800; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i356 = HEAP32[$$0$i7$i354>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $802 = $$pre$i$i356;
  }
  $801 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($801,$802); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $803 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $804 = (($803) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $804; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4400 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $807 = $$pre4400;$__r$0$i358 = $789;
 }
 $805 = ((($__r$0$i358)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$805>>2] = 24; //@line 635 "modules/event/sdl/Event.cpp"
 $806 = ($807|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($806) {
   HEAP32[$__parent$i361>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $819 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i369 = $2;$829 = $819;
   label = 348;
  } else {
   $__nd$0$i$i363 = $807;
   while(1) {
    $808 = ((($__nd$0$i$i363)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $809 = HEAP32[$808>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $810 = ($809|0)>(52); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($810) {
     $811 = HEAP32[$__nd$0$i$i363>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $812 = ($811|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($812) {
      $$lcssa7504 = $__nd$0$i$i363;$__nd$0$i$i363$lcssa7502 = $__nd$0$i$i363;
      label = 341;
      break;
     } else {
      $$sink$i$i364 = $811;
     }
    } else {
     $814 = ($809|0)<(52); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($814)) {
      $__nd$0$i$i363$lcssa = $__nd$0$i$i363;
      label = 347;
      break;
     }
     $815 = ((($__nd$0$i$i363)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $816 = HEAP32[$815>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $817 = ($816|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($817) {
      $$lcssa7503 = $815;$__nd$0$i$i363$lcssa7501 = $__nd$0$i$i363;
      label = 345;
      break;
     } else {
      $$sink$i$i364 = $816;
     }
    }
    $__nd$0$i$i363 = $$sink$i$i364;
   }
   if ((label|0) == 341) {
    HEAP32[$__parent$i361>>2] = $__nd$0$i$i363$lcssa7502; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $813 = $__nd$0$i$i363$lcssa7502; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i369 = $$lcssa7504;$829 = $813;
    label = 348;
    break;
   }
   else if ((label|0) == 345) {
    HEAP32[$__parent$i361>>2] = $__nd$0$i$i363$lcssa7501; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $818 = $__nd$0$i$i363$lcssa7501; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i369 = $$lcssa7503;$829 = $818;
    label = 348;
    break;
   }
   else if ((label|0) == 347) {
    HEAP32[$__parent$i361>>2] = $__nd$0$i$i363$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $820 = $__nd$0$i$i363$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $821 = ($__nd$0$i$i363$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($821) {
     $$0$i7$i369 = $__parent$i361;$829 = $820;
     label = 348;
     break;
    } else {
     $840 = $807;$__r$0$i373 = $__nd$0$i$i363$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 348) {
  __THREW__ = 0;
  $822 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $823 = __THREW__; __THREW__ = 0;
  $824 = $823&1;
  if ($824) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $825 = ((($822)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$825>>2] = 52; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $826 = ((($822)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$826>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$822>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $827 = ((($822)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$827>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $828 = ((($822)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$828>>2] = $829; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i369>>2] = $822; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $830 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $831 = HEAP32[$830>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $832 = ($831|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($832) {
   $835 = $822;
  } else {
   $833 = $831; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $833; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i371 = HEAP32[$$0$i7$i369>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $835 = $$pre$i$i371;
  }
  $834 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($834,$835); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $836 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $837 = (($836) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $837; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4401 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $840 = $$pre4401;$__r$0$i373 = $822;
 }
 $838 = ((($__r$0$i373)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$838>>2] = 25; //@line 636 "modules/event/sdl/Event.cpp"
 $839 = ($840|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($839) {
   HEAP32[$__parent$i376>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $852 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i384 = $2;$862 = $852;
   label = 362;
  } else {
   $__nd$0$i$i378 = $840;
   while(1) {
    $841 = ((($__nd$0$i$i378)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $842 = HEAP32[$841>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $843 = ($842|0)>(53); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($843) {
     $844 = HEAP32[$__nd$0$i$i378>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $845 = ($844|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($845) {
      $$lcssa7500 = $__nd$0$i$i378;$__nd$0$i$i378$lcssa7498 = $__nd$0$i$i378;
      label = 355;
      break;
     } else {
      $$sink$i$i379 = $844;
     }
    } else {
     $847 = ($842|0)<(53); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($847)) {
      $__nd$0$i$i378$lcssa = $__nd$0$i$i378;
      label = 361;
      break;
     }
     $848 = ((($__nd$0$i$i378)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $849 = HEAP32[$848>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $850 = ($849|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($850) {
      $$lcssa7499 = $848;$__nd$0$i$i378$lcssa7497 = $__nd$0$i$i378;
      label = 359;
      break;
     } else {
      $$sink$i$i379 = $849;
     }
    }
    $__nd$0$i$i378 = $$sink$i$i379;
   }
   if ((label|0) == 355) {
    HEAP32[$__parent$i376>>2] = $__nd$0$i$i378$lcssa7498; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $846 = $__nd$0$i$i378$lcssa7498; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i384 = $$lcssa7500;$862 = $846;
    label = 362;
    break;
   }
   else if ((label|0) == 359) {
    HEAP32[$__parent$i376>>2] = $__nd$0$i$i378$lcssa7497; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $851 = $__nd$0$i$i378$lcssa7497; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i384 = $$lcssa7499;$862 = $851;
    label = 362;
    break;
   }
   else if ((label|0) == 361) {
    HEAP32[$__parent$i376>>2] = $__nd$0$i$i378$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $853 = $__nd$0$i$i378$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $854 = ($__nd$0$i$i378$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($854) {
     $$0$i7$i384 = $__parent$i376;$862 = $853;
     label = 362;
     break;
    } else {
     $873 = $840;$__r$0$i388 = $__nd$0$i$i378$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 362) {
  __THREW__ = 0;
  $855 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $856 = __THREW__; __THREW__ = 0;
  $857 = $856&1;
  if ($857) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $858 = ((($855)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$858>>2] = 53; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $859 = ((($855)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$859>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$855>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $860 = ((($855)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$860>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $861 = ((($855)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$861>>2] = $862; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i384>>2] = $855; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $863 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $864 = HEAP32[$863>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $865 = ($864|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($865) {
   $868 = $855;
  } else {
   $866 = $864; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $866; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i386 = HEAP32[$$0$i7$i384>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $868 = $$pre$i$i386;
  }
  $867 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($867,$868); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $869 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $870 = (($869) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $870; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4402 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $873 = $$pre4402;$__r$0$i388 = $855;
 }
 $871 = ((($__r$0$i388)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$871>>2] = 26; //@line 637 "modules/event/sdl/Event.cpp"
 $872 = ($873|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($872) {
   HEAP32[$__parent$i391>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $885 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i399 = $2;$895 = $885;
   label = 376;
  } else {
   $__nd$0$i$i393 = $873;
   while(1) {
    $874 = ((($__nd$0$i$i393)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $875 = HEAP32[$874>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $876 = ($875|0)>(54); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($876) {
     $877 = HEAP32[$__nd$0$i$i393>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $878 = ($877|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($878) {
      $$lcssa7496 = $__nd$0$i$i393;$__nd$0$i$i393$lcssa7494 = $__nd$0$i$i393;
      label = 369;
      break;
     } else {
      $$sink$i$i394 = $877;
     }
    } else {
     $880 = ($875|0)<(54); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($880)) {
      $__nd$0$i$i393$lcssa = $__nd$0$i$i393;
      label = 375;
      break;
     }
     $881 = ((($__nd$0$i$i393)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $882 = HEAP32[$881>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $883 = ($882|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($883) {
      $$lcssa7495 = $881;$__nd$0$i$i393$lcssa7493 = $__nd$0$i$i393;
      label = 373;
      break;
     } else {
      $$sink$i$i394 = $882;
     }
    }
    $__nd$0$i$i393 = $$sink$i$i394;
   }
   if ((label|0) == 369) {
    HEAP32[$__parent$i391>>2] = $__nd$0$i$i393$lcssa7494; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $879 = $__nd$0$i$i393$lcssa7494; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i399 = $$lcssa7496;$895 = $879;
    label = 376;
    break;
   }
   else if ((label|0) == 373) {
    HEAP32[$__parent$i391>>2] = $__nd$0$i$i393$lcssa7493; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $884 = $__nd$0$i$i393$lcssa7493; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i399 = $$lcssa7495;$895 = $884;
    label = 376;
    break;
   }
   else if ((label|0) == 375) {
    HEAP32[$__parent$i391>>2] = $__nd$0$i$i393$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $886 = $__nd$0$i$i393$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $887 = ($__nd$0$i$i393$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($887) {
     $$0$i7$i399 = $__parent$i391;$895 = $886;
     label = 376;
     break;
    } else {
     $906 = $873;$__r$0$i403 = $__nd$0$i$i393$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 376) {
  __THREW__ = 0;
  $888 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $889 = __THREW__; __THREW__ = 0;
  $890 = $889&1;
  if ($890) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $891 = ((($888)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$891>>2] = 54; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $892 = ((($888)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$892>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$888>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $893 = ((($888)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$893>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $894 = ((($888)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$894>>2] = $895; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i399>>2] = $888; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $896 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $897 = HEAP32[$896>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $898 = ($897|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($898) {
   $901 = $888;
  } else {
   $899 = $897; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $899; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i401 = HEAP32[$$0$i7$i399>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $901 = $$pre$i$i401;
  }
  $900 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($900,$901); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $902 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $903 = (($902) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $903; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4403 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $906 = $$pre4403;$__r$0$i403 = $888;
 }
 $904 = ((($__r$0$i403)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$904>>2] = 27; //@line 638 "modules/event/sdl/Event.cpp"
 $905 = ($906|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($905) {
   HEAP32[$__parent$i406>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $918 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i414 = $2;$928 = $918;
   label = 390;
  } else {
   $__nd$0$i$i408 = $906;
   while(1) {
    $907 = ((($__nd$0$i$i408)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $908 = HEAP32[$907>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $909 = ($908|0)>(55); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($909) {
     $910 = HEAP32[$__nd$0$i$i408>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $911 = ($910|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($911) {
      $$lcssa7492 = $__nd$0$i$i408;$__nd$0$i$i408$lcssa7490 = $__nd$0$i$i408;
      label = 383;
      break;
     } else {
      $$sink$i$i409 = $910;
     }
    } else {
     $913 = ($908|0)<(55); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($913)) {
      $__nd$0$i$i408$lcssa = $__nd$0$i$i408;
      label = 389;
      break;
     }
     $914 = ((($__nd$0$i$i408)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $915 = HEAP32[$914>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $916 = ($915|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($916) {
      $$lcssa7491 = $914;$__nd$0$i$i408$lcssa7489 = $__nd$0$i$i408;
      label = 387;
      break;
     } else {
      $$sink$i$i409 = $915;
     }
    }
    $__nd$0$i$i408 = $$sink$i$i409;
   }
   if ((label|0) == 383) {
    HEAP32[$__parent$i406>>2] = $__nd$0$i$i408$lcssa7490; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $912 = $__nd$0$i$i408$lcssa7490; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i414 = $$lcssa7492;$928 = $912;
    label = 390;
    break;
   }
   else if ((label|0) == 387) {
    HEAP32[$__parent$i406>>2] = $__nd$0$i$i408$lcssa7489; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $917 = $__nd$0$i$i408$lcssa7489; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i414 = $$lcssa7491;$928 = $917;
    label = 390;
    break;
   }
   else if ((label|0) == 389) {
    HEAP32[$__parent$i406>>2] = $__nd$0$i$i408$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $919 = $__nd$0$i$i408$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $920 = ($__nd$0$i$i408$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($920) {
     $$0$i7$i414 = $__parent$i406;$928 = $919;
     label = 390;
     break;
    } else {
     $939 = $906;$__r$0$i418 = $__nd$0$i$i408$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 390) {
  __THREW__ = 0;
  $921 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $922 = __THREW__; __THREW__ = 0;
  $923 = $922&1;
  if ($923) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $924 = ((($921)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$924>>2] = 55; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $925 = ((($921)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$925>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$921>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $926 = ((($921)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$926>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $927 = ((($921)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$927>>2] = $928; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i414>>2] = $921; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $929 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $930 = HEAP32[$929>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $931 = ($930|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($931) {
   $934 = $921;
  } else {
   $932 = $930; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $932; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i416 = HEAP32[$$0$i7$i414>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $934 = $$pre$i$i416;
  }
  $933 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($933,$934); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $935 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $936 = (($935) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $936; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4404 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $939 = $$pre4404;$__r$0$i418 = $921;
 }
 $937 = ((($__r$0$i418)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$937>>2] = 28; //@line 639 "modules/event/sdl/Event.cpp"
 $938 = ($939|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($938) {
   HEAP32[$__parent$i421>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $951 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i429 = $2;$961 = $951;
   label = 404;
  } else {
   $__nd$0$i$i423 = $939;
   while(1) {
    $940 = ((($__nd$0$i$i423)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $941 = HEAP32[$940>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $942 = ($941|0)>(56); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($942) {
     $943 = HEAP32[$__nd$0$i$i423>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $944 = ($943|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($944) {
      $$lcssa7488 = $__nd$0$i$i423;$__nd$0$i$i423$lcssa7486 = $__nd$0$i$i423;
      label = 397;
      break;
     } else {
      $$sink$i$i424 = $943;
     }
    } else {
     $946 = ($941|0)<(56); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($946)) {
      $__nd$0$i$i423$lcssa = $__nd$0$i$i423;
      label = 403;
      break;
     }
     $947 = ((($__nd$0$i$i423)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $948 = HEAP32[$947>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $949 = ($948|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($949) {
      $$lcssa7487 = $947;$__nd$0$i$i423$lcssa7485 = $__nd$0$i$i423;
      label = 401;
      break;
     } else {
      $$sink$i$i424 = $948;
     }
    }
    $__nd$0$i$i423 = $$sink$i$i424;
   }
   if ((label|0) == 397) {
    HEAP32[$__parent$i421>>2] = $__nd$0$i$i423$lcssa7486; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $945 = $__nd$0$i$i423$lcssa7486; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i429 = $$lcssa7488;$961 = $945;
    label = 404;
    break;
   }
   else if ((label|0) == 401) {
    HEAP32[$__parent$i421>>2] = $__nd$0$i$i423$lcssa7485; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $950 = $__nd$0$i$i423$lcssa7485; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i429 = $$lcssa7487;$961 = $950;
    label = 404;
    break;
   }
   else if ((label|0) == 403) {
    HEAP32[$__parent$i421>>2] = $__nd$0$i$i423$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $952 = $__nd$0$i$i423$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $953 = ($__nd$0$i$i423$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($953) {
     $$0$i7$i429 = $__parent$i421;$961 = $952;
     label = 404;
     break;
    } else {
     $972 = $939;$__r$0$i433 = $__nd$0$i$i423$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 404) {
  __THREW__ = 0;
  $954 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $955 = __THREW__; __THREW__ = 0;
  $956 = $955&1;
  if ($956) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $957 = ((($954)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$957>>2] = 56; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $958 = ((($954)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$958>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$954>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $959 = ((($954)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$959>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $960 = ((($954)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$960>>2] = $961; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i429>>2] = $954; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $962 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $963 = HEAP32[$962>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $964 = ($963|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($964) {
   $967 = $954;
  } else {
   $965 = $963; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $965; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i431 = HEAP32[$$0$i7$i429>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $967 = $$pre$i$i431;
  }
  $966 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($966,$967); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $968 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $969 = (($968) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $969; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4405 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $972 = $$pre4405;$__r$0$i433 = $954;
 }
 $970 = ((($__r$0$i433)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$970>>2] = 29; //@line 640 "modules/event/sdl/Event.cpp"
 $971 = ($972|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($971) {
   HEAP32[$__parent$i436>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $984 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i444 = $2;$994 = $984;
   label = 418;
  } else {
   $__nd$0$i$i438 = $972;
   while(1) {
    $973 = ((($__nd$0$i$i438)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $974 = HEAP32[$973>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $975 = ($974|0)>(57); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($975) {
     $976 = HEAP32[$__nd$0$i$i438>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $977 = ($976|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($977) {
      $$lcssa7484 = $__nd$0$i$i438;$__nd$0$i$i438$lcssa7482 = $__nd$0$i$i438;
      label = 411;
      break;
     } else {
      $$sink$i$i439 = $976;
     }
    } else {
     $979 = ($974|0)<(57); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($979)) {
      $__nd$0$i$i438$lcssa = $__nd$0$i$i438;
      label = 417;
      break;
     }
     $980 = ((($__nd$0$i$i438)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $981 = HEAP32[$980>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $982 = ($981|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($982) {
      $$lcssa7483 = $980;$__nd$0$i$i438$lcssa7481 = $__nd$0$i$i438;
      label = 415;
      break;
     } else {
      $$sink$i$i439 = $981;
     }
    }
    $__nd$0$i$i438 = $$sink$i$i439;
   }
   if ((label|0) == 411) {
    HEAP32[$__parent$i436>>2] = $__nd$0$i$i438$lcssa7482; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $978 = $__nd$0$i$i438$lcssa7482; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i444 = $$lcssa7484;$994 = $978;
    label = 418;
    break;
   }
   else if ((label|0) == 415) {
    HEAP32[$__parent$i436>>2] = $__nd$0$i$i438$lcssa7481; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $983 = $__nd$0$i$i438$lcssa7481; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i444 = $$lcssa7483;$994 = $983;
    label = 418;
    break;
   }
   else if ((label|0) == 417) {
    HEAP32[$__parent$i436>>2] = $__nd$0$i$i438$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $985 = $__nd$0$i$i438$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $986 = ($__nd$0$i$i438$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($986) {
     $$0$i7$i444 = $__parent$i436;$994 = $985;
     label = 418;
     break;
    } else {
     $1005 = $972;$__r$0$i448 = $__nd$0$i$i438$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 418) {
  __THREW__ = 0;
  $987 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $988 = __THREW__; __THREW__ = 0;
  $989 = $988&1;
  if ($989) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $990 = ((($987)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$990>>2] = 57; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $991 = ((($987)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$991>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$987>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $992 = ((($987)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$992>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $993 = ((($987)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$993>>2] = $994; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i444>>2] = $987; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $995 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $996 = HEAP32[$995>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $997 = ($996|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($997) {
   $1000 = $987;
  } else {
   $998 = $996; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $998; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i446 = HEAP32[$$0$i7$i444>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1000 = $$pre$i$i446;
  }
  $999 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($999,$1000); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1001 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1002 = (($1001) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1002; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4406 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1005 = $$pre4406;$__r$0$i448 = $987;
 }
 $1003 = ((($__r$0$i448)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1003>>2] = 30; //@line 641 "modules/event/sdl/Event.cpp"
 $1004 = ($1005|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1004) {
   HEAP32[$__parent$i451>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1017 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i459 = $2;$1027 = $1017;
   label = 432;
  } else {
   $__nd$0$i$i453 = $1005;
   while(1) {
    $1006 = ((($__nd$0$i$i453)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1007 = HEAP32[$1006>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1008 = ($1007|0)>(58); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1008) {
     $1009 = HEAP32[$__nd$0$i$i453>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1010 = ($1009|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1010) {
      $$lcssa7480 = $__nd$0$i$i453;$__nd$0$i$i453$lcssa7478 = $__nd$0$i$i453;
      label = 425;
      break;
     } else {
      $$sink$i$i454 = $1009;
     }
    } else {
     $1012 = ($1007|0)<(58); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1012)) {
      $__nd$0$i$i453$lcssa = $__nd$0$i$i453;
      label = 431;
      break;
     }
     $1013 = ((($__nd$0$i$i453)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1014 = HEAP32[$1013>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1015 = ($1014|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1015) {
      $$lcssa7479 = $1013;$__nd$0$i$i453$lcssa7477 = $__nd$0$i$i453;
      label = 429;
      break;
     } else {
      $$sink$i$i454 = $1014;
     }
    }
    $__nd$0$i$i453 = $$sink$i$i454;
   }
   if ((label|0) == 425) {
    HEAP32[$__parent$i451>>2] = $__nd$0$i$i453$lcssa7478; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1011 = $__nd$0$i$i453$lcssa7478; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i459 = $$lcssa7480;$1027 = $1011;
    label = 432;
    break;
   }
   else if ((label|0) == 429) {
    HEAP32[$__parent$i451>>2] = $__nd$0$i$i453$lcssa7477; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1016 = $__nd$0$i$i453$lcssa7477; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i459 = $$lcssa7479;$1027 = $1016;
    label = 432;
    break;
   }
   else if ((label|0) == 431) {
    HEAP32[$__parent$i451>>2] = $__nd$0$i$i453$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1018 = $__nd$0$i$i453$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1019 = ($__nd$0$i$i453$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1019) {
     $$0$i7$i459 = $__parent$i451;$1027 = $1018;
     label = 432;
     break;
    } else {
     $1038 = $1005;$__r$0$i463 = $__nd$0$i$i453$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 432) {
  __THREW__ = 0;
  $1020 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1021 = __THREW__; __THREW__ = 0;
  $1022 = $1021&1;
  if ($1022) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1023 = ((($1020)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1023>>2] = 58; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1024 = ((($1020)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1024>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1020>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1025 = ((($1020)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1025>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1026 = ((($1020)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1026>>2] = $1027; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i459>>2] = $1020; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1028 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1029 = HEAP32[$1028>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1030 = ($1029|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1030) {
   $1033 = $1020;
  } else {
   $1031 = $1029; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1031; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i461 = HEAP32[$$0$i7$i459>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1033 = $$pre$i$i461;
  }
  $1032 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1032,$1033); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1034 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1035 = (($1034) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1035; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4407 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1038 = $$pre4407;$__r$0$i463 = $1020;
 }
 $1036 = ((($__r$0$i463)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1036>>2] = 31; //@line 642 "modules/event/sdl/Event.cpp"
 $1037 = ($1038|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1037) {
   HEAP32[$__parent$i466>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1050 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i474 = $2;$1060 = $1050;
   label = 446;
  } else {
   $__nd$0$i$i468 = $1038;
   while(1) {
    $1039 = ((($__nd$0$i$i468)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1040 = HEAP32[$1039>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1041 = ($1040|0)>(59); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1041) {
     $1042 = HEAP32[$__nd$0$i$i468>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1043 = ($1042|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1043) {
      $$lcssa7476 = $__nd$0$i$i468;$__nd$0$i$i468$lcssa7474 = $__nd$0$i$i468;
      label = 439;
      break;
     } else {
      $$sink$i$i469 = $1042;
     }
    } else {
     $1045 = ($1040|0)<(59); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1045)) {
      $__nd$0$i$i468$lcssa = $__nd$0$i$i468;
      label = 445;
      break;
     }
     $1046 = ((($__nd$0$i$i468)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1047 = HEAP32[$1046>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1048 = ($1047|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1048) {
      $$lcssa7475 = $1046;$__nd$0$i$i468$lcssa7473 = $__nd$0$i$i468;
      label = 443;
      break;
     } else {
      $$sink$i$i469 = $1047;
     }
    }
    $__nd$0$i$i468 = $$sink$i$i469;
   }
   if ((label|0) == 439) {
    HEAP32[$__parent$i466>>2] = $__nd$0$i$i468$lcssa7474; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1044 = $__nd$0$i$i468$lcssa7474; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i474 = $$lcssa7476;$1060 = $1044;
    label = 446;
    break;
   }
   else if ((label|0) == 443) {
    HEAP32[$__parent$i466>>2] = $__nd$0$i$i468$lcssa7473; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1049 = $__nd$0$i$i468$lcssa7473; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i474 = $$lcssa7475;$1060 = $1049;
    label = 446;
    break;
   }
   else if ((label|0) == 445) {
    HEAP32[$__parent$i466>>2] = $__nd$0$i$i468$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1051 = $__nd$0$i$i468$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1052 = ($__nd$0$i$i468$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1052) {
     $$0$i7$i474 = $__parent$i466;$1060 = $1051;
     label = 446;
     break;
    } else {
     $1071 = $1038;$__r$0$i478 = $__nd$0$i$i468$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 446) {
  __THREW__ = 0;
  $1053 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1054 = __THREW__; __THREW__ = 0;
  $1055 = $1054&1;
  if ($1055) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1056 = ((($1053)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1056>>2] = 59; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1057 = ((($1053)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1057>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1053>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1058 = ((($1053)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1058>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1059 = ((($1053)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1059>>2] = $1060; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i474>>2] = $1053; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1061 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1062 = HEAP32[$1061>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1063 = ($1062|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1063) {
   $1066 = $1053;
  } else {
   $1064 = $1062; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1064; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i476 = HEAP32[$$0$i7$i474>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1066 = $$pre$i$i476;
  }
  $1065 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1065,$1066); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1067 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1068 = (($1067) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1068; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4408 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1071 = $$pre4408;$__r$0$i478 = $1053;
 }
 $1069 = ((($__r$0$i478)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1069>>2] = 32; //@line 643 "modules/event/sdl/Event.cpp"
 $1070 = ($1071|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1070) {
   HEAP32[$__parent$i481>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1083 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i489 = $2;$1093 = $1083;
   label = 460;
  } else {
   $__nd$0$i$i483 = $1071;
   while(1) {
    $1072 = ((($__nd$0$i$i483)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1073 = HEAP32[$1072>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1074 = ($1073|0)>(60); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1074) {
     $1075 = HEAP32[$__nd$0$i$i483>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1076 = ($1075|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1076) {
      $$lcssa7472 = $__nd$0$i$i483;$__nd$0$i$i483$lcssa7470 = $__nd$0$i$i483;
      label = 453;
      break;
     } else {
      $$sink$i$i484 = $1075;
     }
    } else {
     $1078 = ($1073|0)<(60); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1078)) {
      $__nd$0$i$i483$lcssa = $__nd$0$i$i483;
      label = 459;
      break;
     }
     $1079 = ((($__nd$0$i$i483)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1080 = HEAP32[$1079>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1081 = ($1080|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1081) {
      $$lcssa7471 = $1079;$__nd$0$i$i483$lcssa7469 = $__nd$0$i$i483;
      label = 457;
      break;
     } else {
      $$sink$i$i484 = $1080;
     }
    }
    $__nd$0$i$i483 = $$sink$i$i484;
   }
   if ((label|0) == 453) {
    HEAP32[$__parent$i481>>2] = $__nd$0$i$i483$lcssa7470; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1077 = $__nd$0$i$i483$lcssa7470; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i489 = $$lcssa7472;$1093 = $1077;
    label = 460;
    break;
   }
   else if ((label|0) == 457) {
    HEAP32[$__parent$i481>>2] = $__nd$0$i$i483$lcssa7469; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1082 = $__nd$0$i$i483$lcssa7469; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i489 = $$lcssa7471;$1093 = $1082;
    label = 460;
    break;
   }
   else if ((label|0) == 459) {
    HEAP32[$__parent$i481>>2] = $__nd$0$i$i483$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1084 = $__nd$0$i$i483$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1085 = ($__nd$0$i$i483$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1085) {
     $$0$i7$i489 = $__parent$i481;$1093 = $1084;
     label = 460;
     break;
    } else {
     $1104 = $1071;$__r$0$i493 = $__nd$0$i$i483$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 460) {
  __THREW__ = 0;
  $1086 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1087 = __THREW__; __THREW__ = 0;
  $1088 = $1087&1;
  if ($1088) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1089 = ((($1086)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1089>>2] = 60; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1090 = ((($1086)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1090>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1086>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1091 = ((($1086)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1091>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1092 = ((($1086)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1092>>2] = $1093; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i489>>2] = $1086; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1094 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1095 = HEAP32[$1094>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1096 = ($1095|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1096) {
   $1099 = $1086;
  } else {
   $1097 = $1095; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1097; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i491 = HEAP32[$$0$i7$i489>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1099 = $$pre$i$i491;
  }
  $1098 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1098,$1099); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1100 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1101 = (($1100) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1101; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4409 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1104 = $$pre4409;$__r$0$i493 = $1086;
 }
 $1102 = ((($__r$0$i493)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1102>>2] = 33; //@line 644 "modules/event/sdl/Event.cpp"
 $1103 = ($1104|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1103) {
   HEAP32[$__parent$i496>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1116 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i504 = $2;$1126 = $1116;
   label = 474;
  } else {
   $__nd$0$i$i498 = $1104;
   while(1) {
    $1105 = ((($__nd$0$i$i498)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1106 = HEAP32[$1105>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1107 = ($1106|0)>(61); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1107) {
     $1108 = HEAP32[$__nd$0$i$i498>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1109 = ($1108|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1109) {
      $$lcssa7468 = $__nd$0$i$i498;$__nd$0$i$i498$lcssa7466 = $__nd$0$i$i498;
      label = 467;
      break;
     } else {
      $$sink$i$i499 = $1108;
     }
    } else {
     $1111 = ($1106|0)<(61); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1111)) {
      $__nd$0$i$i498$lcssa = $__nd$0$i$i498;
      label = 473;
      break;
     }
     $1112 = ((($__nd$0$i$i498)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1113 = HEAP32[$1112>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1114 = ($1113|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1114) {
      $$lcssa7467 = $1112;$__nd$0$i$i498$lcssa7465 = $__nd$0$i$i498;
      label = 471;
      break;
     } else {
      $$sink$i$i499 = $1113;
     }
    }
    $__nd$0$i$i498 = $$sink$i$i499;
   }
   if ((label|0) == 467) {
    HEAP32[$__parent$i496>>2] = $__nd$0$i$i498$lcssa7466; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1110 = $__nd$0$i$i498$lcssa7466; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i504 = $$lcssa7468;$1126 = $1110;
    label = 474;
    break;
   }
   else if ((label|0) == 471) {
    HEAP32[$__parent$i496>>2] = $__nd$0$i$i498$lcssa7465; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1115 = $__nd$0$i$i498$lcssa7465; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i504 = $$lcssa7467;$1126 = $1115;
    label = 474;
    break;
   }
   else if ((label|0) == 473) {
    HEAP32[$__parent$i496>>2] = $__nd$0$i$i498$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1117 = $__nd$0$i$i498$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1118 = ($__nd$0$i$i498$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1118) {
     $$0$i7$i504 = $__parent$i496;$1126 = $1117;
     label = 474;
     break;
    } else {
     $1137 = $1104;$__r$0$i508 = $__nd$0$i$i498$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 474) {
  __THREW__ = 0;
  $1119 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1120 = __THREW__; __THREW__ = 0;
  $1121 = $1120&1;
  if ($1121) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1122 = ((($1119)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1122>>2] = 61; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1123 = ((($1119)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1123>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1119>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1124 = ((($1119)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1124>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1125 = ((($1119)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1125>>2] = $1126; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i504>>2] = $1119; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1127 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1128 = HEAP32[$1127>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1129 = ($1128|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1129) {
   $1132 = $1119;
  } else {
   $1130 = $1128; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1130; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i506 = HEAP32[$$0$i7$i504>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1132 = $$pre$i$i506;
  }
  $1131 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1131,$1132); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1133 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1134 = (($1133) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1134; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4410 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1137 = $$pre4410;$__r$0$i508 = $1119;
 }
 $1135 = ((($__r$0$i508)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1135>>2] = 34; //@line 645 "modules/event/sdl/Event.cpp"
 $1136 = ($1137|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1136) {
   HEAP32[$__parent$i511>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1149 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i519 = $2;$1159 = $1149;
   label = 488;
  } else {
   $__nd$0$i$i513 = $1137;
   while(1) {
    $1138 = ((($__nd$0$i$i513)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1139 = HEAP32[$1138>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1140 = ($1139|0)>(62); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1140) {
     $1141 = HEAP32[$__nd$0$i$i513>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1142 = ($1141|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1142) {
      $$lcssa7464 = $__nd$0$i$i513;$__nd$0$i$i513$lcssa7462 = $__nd$0$i$i513;
      label = 481;
      break;
     } else {
      $$sink$i$i514 = $1141;
     }
    } else {
     $1144 = ($1139|0)<(62); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1144)) {
      $__nd$0$i$i513$lcssa = $__nd$0$i$i513;
      label = 487;
      break;
     }
     $1145 = ((($__nd$0$i$i513)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1146 = HEAP32[$1145>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1147 = ($1146|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1147) {
      $$lcssa7463 = $1145;$__nd$0$i$i513$lcssa7461 = $__nd$0$i$i513;
      label = 485;
      break;
     } else {
      $$sink$i$i514 = $1146;
     }
    }
    $__nd$0$i$i513 = $$sink$i$i514;
   }
   if ((label|0) == 481) {
    HEAP32[$__parent$i511>>2] = $__nd$0$i$i513$lcssa7462; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1143 = $__nd$0$i$i513$lcssa7462; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i519 = $$lcssa7464;$1159 = $1143;
    label = 488;
    break;
   }
   else if ((label|0) == 485) {
    HEAP32[$__parent$i511>>2] = $__nd$0$i$i513$lcssa7461; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1148 = $__nd$0$i$i513$lcssa7461; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i519 = $$lcssa7463;$1159 = $1148;
    label = 488;
    break;
   }
   else if ((label|0) == 487) {
    HEAP32[$__parent$i511>>2] = $__nd$0$i$i513$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1150 = $__nd$0$i$i513$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1151 = ($__nd$0$i$i513$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1151) {
     $$0$i7$i519 = $__parent$i511;$1159 = $1150;
     label = 488;
     break;
    } else {
     $1170 = $1137;$__r$0$i523 = $__nd$0$i$i513$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 488) {
  __THREW__ = 0;
  $1152 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1153 = __THREW__; __THREW__ = 0;
  $1154 = $1153&1;
  if ($1154) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1155 = ((($1152)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1155>>2] = 62; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1156 = ((($1152)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1156>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1152>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1157 = ((($1152)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1157>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1158 = ((($1152)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1158>>2] = $1159; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i519>>2] = $1152; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1160 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1161 = HEAP32[$1160>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1162 = ($1161|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1162) {
   $1165 = $1152;
  } else {
   $1163 = $1161; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1163; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i521 = HEAP32[$$0$i7$i519>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1165 = $$pre$i$i521;
  }
  $1164 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1164,$1165); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1166 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1167 = (($1166) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1167; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4411 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1170 = $$pre4411;$__r$0$i523 = $1152;
 }
 $1168 = ((($__r$0$i523)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1168>>2] = 35; //@line 646 "modules/event/sdl/Event.cpp"
 $1169 = ($1170|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1169) {
   HEAP32[$__parent$i526>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1182 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i534 = $2;$1192 = $1182;
   label = 502;
  } else {
   $__nd$0$i$i528 = $1170;
   while(1) {
    $1171 = ((($__nd$0$i$i528)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1172 = HEAP32[$1171>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1173 = ($1172|0)>(63); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1173) {
     $1174 = HEAP32[$__nd$0$i$i528>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1175 = ($1174|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1175) {
      $$lcssa7460 = $__nd$0$i$i528;$__nd$0$i$i528$lcssa7458 = $__nd$0$i$i528;
      label = 495;
      break;
     } else {
      $$sink$i$i529 = $1174;
     }
    } else {
     $1177 = ($1172|0)<(63); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1177)) {
      $__nd$0$i$i528$lcssa = $__nd$0$i$i528;
      label = 501;
      break;
     }
     $1178 = ((($__nd$0$i$i528)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1179 = HEAP32[$1178>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1180 = ($1179|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1180) {
      $$lcssa7459 = $1178;$__nd$0$i$i528$lcssa7457 = $__nd$0$i$i528;
      label = 499;
      break;
     } else {
      $$sink$i$i529 = $1179;
     }
    }
    $__nd$0$i$i528 = $$sink$i$i529;
   }
   if ((label|0) == 495) {
    HEAP32[$__parent$i526>>2] = $__nd$0$i$i528$lcssa7458; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1176 = $__nd$0$i$i528$lcssa7458; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i534 = $$lcssa7460;$1192 = $1176;
    label = 502;
    break;
   }
   else if ((label|0) == 499) {
    HEAP32[$__parent$i526>>2] = $__nd$0$i$i528$lcssa7457; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1181 = $__nd$0$i$i528$lcssa7457; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i534 = $$lcssa7459;$1192 = $1181;
    label = 502;
    break;
   }
   else if ((label|0) == 501) {
    HEAP32[$__parent$i526>>2] = $__nd$0$i$i528$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1183 = $__nd$0$i$i528$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1184 = ($__nd$0$i$i528$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1184) {
     $$0$i7$i534 = $__parent$i526;$1192 = $1183;
     label = 502;
     break;
    } else {
     $1203 = $1170;$__r$0$i538 = $__nd$0$i$i528$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 502) {
  __THREW__ = 0;
  $1185 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1186 = __THREW__; __THREW__ = 0;
  $1187 = $1186&1;
  if ($1187) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1188 = ((($1185)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1188>>2] = 63; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1189 = ((($1185)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1189>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1185>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1190 = ((($1185)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1190>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1191 = ((($1185)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1191>>2] = $1192; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i534>>2] = $1185; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1193 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1194 = HEAP32[$1193>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1195 = ($1194|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1195) {
   $1198 = $1185;
  } else {
   $1196 = $1194; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1196; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i536 = HEAP32[$$0$i7$i534>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1198 = $$pre$i$i536;
  }
  $1197 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1197,$1198); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1199 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1200 = (($1199) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1200; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4412 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1203 = $$pre4412;$__r$0$i538 = $1185;
 }
 $1201 = ((($__r$0$i538)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1201>>2] = 36; //@line 647 "modules/event/sdl/Event.cpp"
 $1202 = ($1203|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1202) {
   HEAP32[$__parent$i541>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1215 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i549 = $2;$1225 = $1215;
   label = 516;
  } else {
   $__nd$0$i$i543 = $1203;
   while(1) {
    $1204 = ((($__nd$0$i$i543)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1205 = HEAP32[$1204>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1206 = ($1205|0)>(64); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1206) {
     $1207 = HEAP32[$__nd$0$i$i543>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1208 = ($1207|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1208) {
      $$lcssa7456 = $__nd$0$i$i543;$__nd$0$i$i543$lcssa7454 = $__nd$0$i$i543;
      label = 509;
      break;
     } else {
      $$sink$i$i544 = $1207;
     }
    } else {
     $1210 = ($1205|0)<(64); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1210)) {
      $__nd$0$i$i543$lcssa = $__nd$0$i$i543;
      label = 515;
      break;
     }
     $1211 = ((($__nd$0$i$i543)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1212 = HEAP32[$1211>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1213 = ($1212|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1213) {
      $$lcssa7455 = $1211;$__nd$0$i$i543$lcssa7453 = $__nd$0$i$i543;
      label = 513;
      break;
     } else {
      $$sink$i$i544 = $1212;
     }
    }
    $__nd$0$i$i543 = $$sink$i$i544;
   }
   if ((label|0) == 509) {
    HEAP32[$__parent$i541>>2] = $__nd$0$i$i543$lcssa7454; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1209 = $__nd$0$i$i543$lcssa7454; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i549 = $$lcssa7456;$1225 = $1209;
    label = 516;
    break;
   }
   else if ((label|0) == 513) {
    HEAP32[$__parent$i541>>2] = $__nd$0$i$i543$lcssa7453; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1214 = $__nd$0$i$i543$lcssa7453; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i549 = $$lcssa7455;$1225 = $1214;
    label = 516;
    break;
   }
   else if ((label|0) == 515) {
    HEAP32[$__parent$i541>>2] = $__nd$0$i$i543$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1216 = $__nd$0$i$i543$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1217 = ($__nd$0$i$i543$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1217) {
     $$0$i7$i549 = $__parent$i541;$1225 = $1216;
     label = 516;
     break;
    } else {
     $1236 = $1203;$__r$0$i553 = $__nd$0$i$i543$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 516) {
  __THREW__ = 0;
  $1218 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1219 = __THREW__; __THREW__ = 0;
  $1220 = $1219&1;
  if ($1220) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1221 = ((($1218)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1221>>2] = 64; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1222 = ((($1218)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1222>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1218>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1223 = ((($1218)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1223>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1224 = ((($1218)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1224>>2] = $1225; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i549>>2] = $1218; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1226 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1227 = HEAP32[$1226>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1228 = ($1227|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1228) {
   $1231 = $1218;
  } else {
   $1229 = $1227; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1229; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i551 = HEAP32[$$0$i7$i549>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1231 = $$pre$i$i551;
  }
  $1230 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1230,$1231); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1232 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1233 = (($1232) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1233; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4413 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1236 = $$pre4413;$__r$0$i553 = $1218;
 }
 $1234 = ((($__r$0$i553)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1234>>2] = 37; //@line 648 "modules/event/sdl/Event.cpp"
 $1235 = ($1236|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1235) {
   HEAP32[$__parent$i556>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1248 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i564 = $2;$1258 = $1248;
   label = 530;
  } else {
   $__nd$0$i$i558 = $1236;
   while(1) {
    $1237 = ((($__nd$0$i$i558)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1238 = HEAP32[$1237>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1239 = ($1238|0)>(91); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1239) {
     $1240 = HEAP32[$__nd$0$i$i558>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1241 = ($1240|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1241) {
      $$lcssa7452 = $__nd$0$i$i558;$__nd$0$i$i558$lcssa7450 = $__nd$0$i$i558;
      label = 523;
      break;
     } else {
      $$sink$i$i559 = $1240;
     }
    } else {
     $1243 = ($1238|0)<(91); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1243)) {
      $__nd$0$i$i558$lcssa = $__nd$0$i$i558;
      label = 529;
      break;
     }
     $1244 = ((($__nd$0$i$i558)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1245 = HEAP32[$1244>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1246 = ($1245|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1246) {
      $$lcssa7451 = $1244;$__nd$0$i$i558$lcssa7449 = $__nd$0$i$i558;
      label = 527;
      break;
     } else {
      $$sink$i$i559 = $1245;
     }
    }
    $__nd$0$i$i558 = $$sink$i$i559;
   }
   if ((label|0) == 523) {
    HEAP32[$__parent$i556>>2] = $__nd$0$i$i558$lcssa7450; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1242 = $__nd$0$i$i558$lcssa7450; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i564 = $$lcssa7452;$1258 = $1242;
    label = 530;
    break;
   }
   else if ((label|0) == 527) {
    HEAP32[$__parent$i556>>2] = $__nd$0$i$i558$lcssa7449; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1247 = $__nd$0$i$i558$lcssa7449; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i564 = $$lcssa7451;$1258 = $1247;
    label = 530;
    break;
   }
   else if ((label|0) == 529) {
    HEAP32[$__parent$i556>>2] = $__nd$0$i$i558$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1249 = $__nd$0$i$i558$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1250 = ($__nd$0$i$i558$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1250) {
     $$0$i7$i564 = $__parent$i556;$1258 = $1249;
     label = 530;
     break;
    } else {
     $1269 = $1236;$__r$0$i568 = $__nd$0$i$i558$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 530) {
  __THREW__ = 0;
  $1251 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1252 = __THREW__; __THREW__ = 0;
  $1253 = $1252&1;
  if ($1253) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1254 = ((($1251)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1254>>2] = 91; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1255 = ((($1251)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1255>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1251>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1256 = ((($1251)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1256>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1257 = ((($1251)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1257>>2] = $1258; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i564>>2] = $1251; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1259 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1260 = HEAP32[$1259>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1261 = ($1260|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1261) {
   $1264 = $1251;
  } else {
   $1262 = $1260; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1262; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i566 = HEAP32[$$0$i7$i564>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1264 = $$pre$i$i566;
  }
  $1263 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1263,$1264); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1265 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1266 = (($1265) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1266; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4414 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1269 = $$pre4414;$__r$0$i568 = $1251;
 }
 $1267 = ((($__r$0$i568)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1267>>2] = 38; //@line 650 "modules/event/sdl/Event.cpp"
 $1268 = ($1269|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1268) {
   HEAP32[$__parent$i571>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1281 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i579 = $2;$1291 = $1281;
   label = 544;
  } else {
   $__nd$0$i$i573 = $1269;
   while(1) {
    $1270 = ((($__nd$0$i$i573)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1271 = HEAP32[$1270>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1272 = ($1271|0)>(92); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1272) {
     $1273 = HEAP32[$__nd$0$i$i573>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1274 = ($1273|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1274) {
      $$lcssa7448 = $__nd$0$i$i573;$__nd$0$i$i573$lcssa7446 = $__nd$0$i$i573;
      label = 537;
      break;
     } else {
      $$sink$i$i574 = $1273;
     }
    } else {
     $1276 = ($1271|0)<(92); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1276)) {
      $__nd$0$i$i573$lcssa = $__nd$0$i$i573;
      label = 543;
      break;
     }
     $1277 = ((($__nd$0$i$i573)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1278 = HEAP32[$1277>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1279 = ($1278|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1279) {
      $$lcssa7447 = $1277;$__nd$0$i$i573$lcssa7445 = $__nd$0$i$i573;
      label = 541;
      break;
     } else {
      $$sink$i$i574 = $1278;
     }
    }
    $__nd$0$i$i573 = $$sink$i$i574;
   }
   if ((label|0) == 537) {
    HEAP32[$__parent$i571>>2] = $__nd$0$i$i573$lcssa7446; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1275 = $__nd$0$i$i573$lcssa7446; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i579 = $$lcssa7448;$1291 = $1275;
    label = 544;
    break;
   }
   else if ((label|0) == 541) {
    HEAP32[$__parent$i571>>2] = $__nd$0$i$i573$lcssa7445; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1280 = $__nd$0$i$i573$lcssa7445; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i579 = $$lcssa7447;$1291 = $1280;
    label = 544;
    break;
   }
   else if ((label|0) == 543) {
    HEAP32[$__parent$i571>>2] = $__nd$0$i$i573$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1282 = $__nd$0$i$i573$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1283 = ($__nd$0$i$i573$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1283) {
     $$0$i7$i579 = $__parent$i571;$1291 = $1282;
     label = 544;
     break;
    } else {
     $1302 = $1269;$__r$0$i583 = $__nd$0$i$i573$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 544) {
  __THREW__ = 0;
  $1284 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1285 = __THREW__; __THREW__ = 0;
  $1286 = $1285&1;
  if ($1286) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1287 = ((($1284)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1287>>2] = 92; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1288 = ((($1284)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1288>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1284>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1289 = ((($1284)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1289>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1290 = ((($1284)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1290>>2] = $1291; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i579>>2] = $1284; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1292 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1293 = HEAP32[$1292>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1294 = ($1293|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1294) {
   $1297 = $1284;
  } else {
   $1295 = $1293; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1295; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i581 = HEAP32[$$0$i7$i579>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1297 = $$pre$i$i581;
  }
  $1296 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1296,$1297); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1298 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1299 = (($1298) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1299; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4415 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1302 = $$pre4415;$__r$0$i583 = $1284;
 }
 $1300 = ((($__r$0$i583)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1300>>2] = 39; //@line 651 "modules/event/sdl/Event.cpp"
 $1301 = ($1302|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1301) {
   HEAP32[$__parent$i586>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1314 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i594 = $2;$1324 = $1314;
   label = 558;
  } else {
   $__nd$0$i$i588 = $1302;
   while(1) {
    $1303 = ((($__nd$0$i$i588)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1304 = HEAP32[$1303>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1305 = ($1304|0)>(93); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1305) {
     $1306 = HEAP32[$__nd$0$i$i588>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1307 = ($1306|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1307) {
      $$lcssa7444 = $__nd$0$i$i588;$__nd$0$i$i588$lcssa7442 = $__nd$0$i$i588;
      label = 551;
      break;
     } else {
      $$sink$i$i589 = $1306;
     }
    } else {
     $1309 = ($1304|0)<(93); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1309)) {
      $__nd$0$i$i588$lcssa = $__nd$0$i$i588;
      label = 557;
      break;
     }
     $1310 = ((($__nd$0$i$i588)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1311 = HEAP32[$1310>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1312 = ($1311|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1312) {
      $$lcssa7443 = $1310;$__nd$0$i$i588$lcssa7441 = $__nd$0$i$i588;
      label = 555;
      break;
     } else {
      $$sink$i$i589 = $1311;
     }
    }
    $__nd$0$i$i588 = $$sink$i$i589;
   }
   if ((label|0) == 551) {
    HEAP32[$__parent$i586>>2] = $__nd$0$i$i588$lcssa7442; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1308 = $__nd$0$i$i588$lcssa7442; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i594 = $$lcssa7444;$1324 = $1308;
    label = 558;
    break;
   }
   else if ((label|0) == 555) {
    HEAP32[$__parent$i586>>2] = $__nd$0$i$i588$lcssa7441; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1313 = $__nd$0$i$i588$lcssa7441; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i594 = $$lcssa7443;$1324 = $1313;
    label = 558;
    break;
   }
   else if ((label|0) == 557) {
    HEAP32[$__parent$i586>>2] = $__nd$0$i$i588$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1315 = $__nd$0$i$i588$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1316 = ($__nd$0$i$i588$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1316) {
     $$0$i7$i594 = $__parent$i586;$1324 = $1315;
     label = 558;
     break;
    } else {
     $1335 = $1302;$__r$0$i598 = $__nd$0$i$i588$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 558) {
  __THREW__ = 0;
  $1317 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1318 = __THREW__; __THREW__ = 0;
  $1319 = $1318&1;
  if ($1319) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1320 = ((($1317)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1320>>2] = 93; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1321 = ((($1317)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1321>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1317>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1322 = ((($1317)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1322>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1323 = ((($1317)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1323>>2] = $1324; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i594>>2] = $1317; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1325 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1326 = HEAP32[$1325>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1327 = ($1326|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1327) {
   $1330 = $1317;
  } else {
   $1328 = $1326; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1328; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i596 = HEAP32[$$0$i7$i594>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1330 = $$pre$i$i596;
  }
  $1329 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1329,$1330); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1331 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1332 = (($1331) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1332; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4416 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1335 = $$pre4416;$__r$0$i598 = $1317;
 }
 $1333 = ((($__r$0$i598)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1333>>2] = 40; //@line 652 "modules/event/sdl/Event.cpp"
 $1334 = ($1335|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1334) {
   HEAP32[$__parent$i601>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1347 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i609 = $2;$1357 = $1347;
   label = 572;
  } else {
   $__nd$0$i$i603 = $1335;
   while(1) {
    $1336 = ((($__nd$0$i$i603)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1337 = HEAP32[$1336>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1338 = ($1337|0)>(94); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1338) {
     $1339 = HEAP32[$__nd$0$i$i603>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1340 = ($1339|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1340) {
      $$lcssa7440 = $__nd$0$i$i603;$__nd$0$i$i603$lcssa7438 = $__nd$0$i$i603;
      label = 565;
      break;
     } else {
      $$sink$i$i604 = $1339;
     }
    } else {
     $1342 = ($1337|0)<(94); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1342)) {
      $__nd$0$i$i603$lcssa = $__nd$0$i$i603;
      label = 571;
      break;
     }
     $1343 = ((($__nd$0$i$i603)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1344 = HEAP32[$1343>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1345 = ($1344|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1345) {
      $$lcssa7439 = $1343;$__nd$0$i$i603$lcssa7437 = $__nd$0$i$i603;
      label = 569;
      break;
     } else {
      $$sink$i$i604 = $1344;
     }
    }
    $__nd$0$i$i603 = $$sink$i$i604;
   }
   if ((label|0) == 565) {
    HEAP32[$__parent$i601>>2] = $__nd$0$i$i603$lcssa7438; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1341 = $__nd$0$i$i603$lcssa7438; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i609 = $$lcssa7440;$1357 = $1341;
    label = 572;
    break;
   }
   else if ((label|0) == 569) {
    HEAP32[$__parent$i601>>2] = $__nd$0$i$i603$lcssa7437; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1346 = $__nd$0$i$i603$lcssa7437; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i609 = $$lcssa7439;$1357 = $1346;
    label = 572;
    break;
   }
   else if ((label|0) == 571) {
    HEAP32[$__parent$i601>>2] = $__nd$0$i$i603$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1348 = $__nd$0$i$i603$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1349 = ($__nd$0$i$i603$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1349) {
     $$0$i7$i609 = $__parent$i601;$1357 = $1348;
     label = 572;
     break;
    } else {
     $1368 = $1335;$__r$0$i613 = $__nd$0$i$i603$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 572) {
  __THREW__ = 0;
  $1350 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1351 = __THREW__; __THREW__ = 0;
  $1352 = $1351&1;
  if ($1352) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1353 = ((($1350)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1353>>2] = 94; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1354 = ((($1350)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1354>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1350>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1355 = ((($1350)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1355>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1356 = ((($1350)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1356>>2] = $1357; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i609>>2] = $1350; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1358 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1359 = HEAP32[$1358>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1360 = ($1359|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1360) {
   $1363 = $1350;
  } else {
   $1361 = $1359; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1361; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i611 = HEAP32[$$0$i7$i609>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1363 = $$pre$i$i611;
  }
  $1362 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1362,$1363); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1364 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1365 = (($1364) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1365; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4417 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1368 = $$pre4417;$__r$0$i613 = $1350;
 }
 $1366 = ((($__r$0$i613)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1366>>2] = 41; //@line 653 "modules/event/sdl/Event.cpp"
 $1367 = ($1368|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1367) {
   HEAP32[$__parent$i616>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1380 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i624 = $2;$1390 = $1380;
   label = 586;
  } else {
   $__nd$0$i$i618 = $1368;
   while(1) {
    $1369 = ((($__nd$0$i$i618)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1370 = HEAP32[$1369>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1371 = ($1370|0)>(95); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1371) {
     $1372 = HEAP32[$__nd$0$i$i618>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1373 = ($1372|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1373) {
      $$lcssa7436 = $__nd$0$i$i618;$__nd$0$i$i618$lcssa7434 = $__nd$0$i$i618;
      label = 579;
      break;
     } else {
      $$sink$i$i619 = $1372;
     }
    } else {
     $1375 = ($1370|0)<(95); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1375)) {
      $__nd$0$i$i618$lcssa = $__nd$0$i$i618;
      label = 585;
      break;
     }
     $1376 = ((($__nd$0$i$i618)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1377 = HEAP32[$1376>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1378 = ($1377|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1378) {
      $$lcssa7435 = $1376;$__nd$0$i$i618$lcssa7433 = $__nd$0$i$i618;
      label = 583;
      break;
     } else {
      $$sink$i$i619 = $1377;
     }
    }
    $__nd$0$i$i618 = $$sink$i$i619;
   }
   if ((label|0) == 579) {
    HEAP32[$__parent$i616>>2] = $__nd$0$i$i618$lcssa7434; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1374 = $__nd$0$i$i618$lcssa7434; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i624 = $$lcssa7436;$1390 = $1374;
    label = 586;
    break;
   }
   else if ((label|0) == 583) {
    HEAP32[$__parent$i616>>2] = $__nd$0$i$i618$lcssa7433; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1379 = $__nd$0$i$i618$lcssa7433; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i624 = $$lcssa7435;$1390 = $1379;
    label = 586;
    break;
   }
   else if ((label|0) == 585) {
    HEAP32[$__parent$i616>>2] = $__nd$0$i$i618$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1381 = $__nd$0$i$i618$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1382 = ($__nd$0$i$i618$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1382) {
     $$0$i7$i624 = $__parent$i616;$1390 = $1381;
     label = 586;
     break;
    } else {
     $1401 = $1368;$__r$0$i628 = $__nd$0$i$i618$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 586) {
  __THREW__ = 0;
  $1383 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1384 = __THREW__; __THREW__ = 0;
  $1385 = $1384&1;
  if ($1385) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1386 = ((($1383)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1386>>2] = 95; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1387 = ((($1383)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1387>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1383>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1388 = ((($1383)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1388>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1389 = ((($1383)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1389>>2] = $1390; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i624>>2] = $1383; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1391 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1392 = HEAP32[$1391>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1393 = ($1392|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1393) {
   $1396 = $1383;
  } else {
   $1394 = $1392; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1394; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i626 = HEAP32[$$0$i7$i624>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1396 = $$pre$i$i626;
  }
  $1395 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1395,$1396); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1397 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1398 = (($1397) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1398; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4418 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1401 = $$pre4418;$__r$0$i628 = $1383;
 }
 $1399 = ((($__r$0$i628)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1399>>2] = 42; //@line 654 "modules/event/sdl/Event.cpp"
 $1400 = ($1401|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1400) {
   HEAP32[$__parent$i631>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1413 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i639 = $2;$1423 = $1413;
   label = 600;
  } else {
   $__nd$0$i$i633 = $1401;
   while(1) {
    $1402 = ((($__nd$0$i$i633)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1403 = HEAP32[$1402>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1404 = ($1403|0)>(96); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1404) {
     $1405 = HEAP32[$__nd$0$i$i633>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1406 = ($1405|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1406) {
      $$lcssa7432 = $__nd$0$i$i633;$__nd$0$i$i633$lcssa7430 = $__nd$0$i$i633;
      label = 593;
      break;
     } else {
      $$sink$i$i634 = $1405;
     }
    } else {
     $1408 = ($1403|0)<(96); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1408)) {
      $__nd$0$i$i633$lcssa = $__nd$0$i$i633;
      label = 599;
      break;
     }
     $1409 = ((($__nd$0$i$i633)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1410 = HEAP32[$1409>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1411 = ($1410|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1411) {
      $$lcssa7431 = $1409;$__nd$0$i$i633$lcssa7429 = $__nd$0$i$i633;
      label = 597;
      break;
     } else {
      $$sink$i$i634 = $1410;
     }
    }
    $__nd$0$i$i633 = $$sink$i$i634;
   }
   if ((label|0) == 593) {
    HEAP32[$__parent$i631>>2] = $__nd$0$i$i633$lcssa7430; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1407 = $__nd$0$i$i633$lcssa7430; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i639 = $$lcssa7432;$1423 = $1407;
    label = 600;
    break;
   }
   else if ((label|0) == 597) {
    HEAP32[$__parent$i631>>2] = $__nd$0$i$i633$lcssa7429; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1412 = $__nd$0$i$i633$lcssa7429; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i639 = $$lcssa7431;$1423 = $1412;
    label = 600;
    break;
   }
   else if ((label|0) == 599) {
    HEAP32[$__parent$i631>>2] = $__nd$0$i$i633$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1414 = $__nd$0$i$i633$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1415 = ($__nd$0$i$i633$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1415) {
     $$0$i7$i639 = $__parent$i631;$1423 = $1414;
     label = 600;
     break;
    } else {
     $1434 = $1401;$__r$0$i643 = $__nd$0$i$i633$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 600) {
  __THREW__ = 0;
  $1416 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1417 = __THREW__; __THREW__ = 0;
  $1418 = $1417&1;
  if ($1418) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1419 = ((($1416)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1419>>2] = 96; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1420 = ((($1416)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1420>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1416>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1421 = ((($1416)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1421>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1422 = ((($1416)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1422>>2] = $1423; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i639>>2] = $1416; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1424 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1425 = HEAP32[$1424>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1426 = ($1425|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1426) {
   $1429 = $1416;
  } else {
   $1427 = $1425; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1427; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i641 = HEAP32[$$0$i7$i639>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1429 = $$pre$i$i641;
  }
  $1428 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1428,$1429); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1430 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1431 = (($1430) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1431; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4419 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1434 = $$pre4419;$__r$0$i643 = $1416;
 }
 $1432 = ((($__r$0$i643)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1432>>2] = 43; //@line 655 "modules/event/sdl/Event.cpp"
 $1433 = ($1434|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1433) {
   HEAP32[$__parent$i646>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1446 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i654 = $2;$1456 = $1446;
   label = 614;
  } else {
   $__nd$0$i$i648 = $1434;
   while(1) {
    $1435 = ((($__nd$0$i$i648)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1436 = HEAP32[$1435>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1437 = ($1436|0)>(97); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1437) {
     $1438 = HEAP32[$__nd$0$i$i648>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1439 = ($1438|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1439) {
      $$lcssa7428 = $__nd$0$i$i648;$__nd$0$i$i648$lcssa7426 = $__nd$0$i$i648;
      label = 607;
      break;
     } else {
      $$sink$i$i649 = $1438;
     }
    } else {
     $1441 = ($1436|0)<(97); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1441)) {
      $__nd$0$i$i648$lcssa = $__nd$0$i$i648;
      label = 613;
      break;
     }
     $1442 = ((($__nd$0$i$i648)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1443 = HEAP32[$1442>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1444 = ($1443|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1444) {
      $$lcssa7427 = $1442;$__nd$0$i$i648$lcssa7425 = $__nd$0$i$i648;
      label = 611;
      break;
     } else {
      $$sink$i$i649 = $1443;
     }
    }
    $__nd$0$i$i648 = $$sink$i$i649;
   }
   if ((label|0) == 607) {
    HEAP32[$__parent$i646>>2] = $__nd$0$i$i648$lcssa7426; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1440 = $__nd$0$i$i648$lcssa7426; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i654 = $$lcssa7428;$1456 = $1440;
    label = 614;
    break;
   }
   else if ((label|0) == 611) {
    HEAP32[$__parent$i646>>2] = $__nd$0$i$i648$lcssa7425; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1445 = $__nd$0$i$i648$lcssa7425; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i654 = $$lcssa7427;$1456 = $1445;
    label = 614;
    break;
   }
   else if ((label|0) == 613) {
    HEAP32[$__parent$i646>>2] = $__nd$0$i$i648$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1447 = $__nd$0$i$i648$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1448 = ($__nd$0$i$i648$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1448) {
     $$0$i7$i654 = $__parent$i646;$1456 = $1447;
     label = 614;
     break;
    } else {
     $1467 = $1434;$__r$0$i658 = $__nd$0$i$i648$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 614) {
  __THREW__ = 0;
  $1449 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1450 = __THREW__; __THREW__ = 0;
  $1451 = $1450&1;
  if ($1451) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1452 = ((($1449)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1452>>2] = 97; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1453 = ((($1449)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1453>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1449>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1454 = ((($1449)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1454>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1455 = ((($1449)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1455>>2] = $1456; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i654>>2] = $1449; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1457 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1458 = HEAP32[$1457>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1459 = ($1458|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1459) {
   $1462 = $1449;
  } else {
   $1460 = $1458; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1460; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i656 = HEAP32[$$0$i7$i654>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1462 = $$pre$i$i656;
  }
  $1461 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1461,$1462); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1463 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1464 = (($1463) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1464; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4420 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1467 = $$pre4420;$__r$0$i658 = $1449;
 }
 $1465 = ((($__r$0$i658)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1465>>2] = 44; //@line 656 "modules/event/sdl/Event.cpp"
 $1466 = ($1467|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1466) {
   HEAP32[$__parent$i661>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1479 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i669 = $2;$1489 = $1479;
   label = 628;
  } else {
   $__nd$0$i$i663 = $1467;
   while(1) {
    $1468 = ((($__nd$0$i$i663)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1469 = HEAP32[$1468>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1470 = ($1469|0)>(98); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1470) {
     $1471 = HEAP32[$__nd$0$i$i663>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1472 = ($1471|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1472) {
      $$lcssa7424 = $__nd$0$i$i663;$__nd$0$i$i663$lcssa7422 = $__nd$0$i$i663;
      label = 621;
      break;
     } else {
      $$sink$i$i664 = $1471;
     }
    } else {
     $1474 = ($1469|0)<(98); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1474)) {
      $__nd$0$i$i663$lcssa = $__nd$0$i$i663;
      label = 627;
      break;
     }
     $1475 = ((($__nd$0$i$i663)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1476 = HEAP32[$1475>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1477 = ($1476|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1477) {
      $$lcssa7423 = $1475;$__nd$0$i$i663$lcssa7421 = $__nd$0$i$i663;
      label = 625;
      break;
     } else {
      $$sink$i$i664 = $1476;
     }
    }
    $__nd$0$i$i663 = $$sink$i$i664;
   }
   if ((label|0) == 621) {
    HEAP32[$__parent$i661>>2] = $__nd$0$i$i663$lcssa7422; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1473 = $__nd$0$i$i663$lcssa7422; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i669 = $$lcssa7424;$1489 = $1473;
    label = 628;
    break;
   }
   else if ((label|0) == 625) {
    HEAP32[$__parent$i661>>2] = $__nd$0$i$i663$lcssa7421; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1478 = $__nd$0$i$i663$lcssa7421; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i669 = $$lcssa7423;$1489 = $1478;
    label = 628;
    break;
   }
   else if ((label|0) == 627) {
    HEAP32[$__parent$i661>>2] = $__nd$0$i$i663$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1480 = $__nd$0$i$i663$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1481 = ($__nd$0$i$i663$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1481) {
     $$0$i7$i669 = $__parent$i661;$1489 = $1480;
     label = 628;
     break;
    } else {
     $1500 = $1467;$__r$0$i673 = $__nd$0$i$i663$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 628) {
  __THREW__ = 0;
  $1482 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1483 = __THREW__; __THREW__ = 0;
  $1484 = $1483&1;
  if ($1484) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1485 = ((($1482)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1485>>2] = 98; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1486 = ((($1482)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1486>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1482>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1487 = ((($1482)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1487>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1488 = ((($1482)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1488>>2] = $1489; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i669>>2] = $1482; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1490 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1491 = HEAP32[$1490>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1492 = ($1491|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1492) {
   $1495 = $1482;
  } else {
   $1493 = $1491; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1493; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i671 = HEAP32[$$0$i7$i669>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1495 = $$pre$i$i671;
  }
  $1494 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1494,$1495); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1496 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1497 = (($1496) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1497; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4421 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1500 = $$pre4421;$__r$0$i673 = $1482;
 }
 $1498 = ((($__r$0$i673)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1498>>2] = 45; //@line 657 "modules/event/sdl/Event.cpp"
 $1499 = ($1500|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1499) {
   HEAP32[$__parent$i676>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1512 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i684 = $2;$1522 = $1512;
   label = 642;
  } else {
   $__nd$0$i$i678 = $1500;
   while(1) {
    $1501 = ((($__nd$0$i$i678)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1502 = HEAP32[$1501>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1503 = ($1502|0)>(99); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1503) {
     $1504 = HEAP32[$__nd$0$i$i678>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1505 = ($1504|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1505) {
      $$lcssa7420 = $__nd$0$i$i678;$__nd$0$i$i678$lcssa7418 = $__nd$0$i$i678;
      label = 635;
      break;
     } else {
      $$sink$i$i679 = $1504;
     }
    } else {
     $1507 = ($1502|0)<(99); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1507)) {
      $__nd$0$i$i678$lcssa = $__nd$0$i$i678;
      label = 641;
      break;
     }
     $1508 = ((($__nd$0$i$i678)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1509 = HEAP32[$1508>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1510 = ($1509|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1510) {
      $$lcssa7419 = $1508;$__nd$0$i$i678$lcssa7417 = $__nd$0$i$i678;
      label = 639;
      break;
     } else {
      $$sink$i$i679 = $1509;
     }
    }
    $__nd$0$i$i678 = $$sink$i$i679;
   }
   if ((label|0) == 635) {
    HEAP32[$__parent$i676>>2] = $__nd$0$i$i678$lcssa7418; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1506 = $__nd$0$i$i678$lcssa7418; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i684 = $$lcssa7420;$1522 = $1506;
    label = 642;
    break;
   }
   else if ((label|0) == 639) {
    HEAP32[$__parent$i676>>2] = $__nd$0$i$i678$lcssa7417; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1511 = $__nd$0$i$i678$lcssa7417; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i684 = $$lcssa7419;$1522 = $1511;
    label = 642;
    break;
   }
   else if ((label|0) == 641) {
    HEAP32[$__parent$i676>>2] = $__nd$0$i$i678$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1513 = $__nd$0$i$i678$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1514 = ($__nd$0$i$i678$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1514) {
     $$0$i7$i684 = $__parent$i676;$1522 = $1513;
     label = 642;
     break;
    } else {
     $1533 = $1500;$__r$0$i688 = $__nd$0$i$i678$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 642) {
  __THREW__ = 0;
  $1515 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1516 = __THREW__; __THREW__ = 0;
  $1517 = $1516&1;
  if ($1517) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1518 = ((($1515)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1518>>2] = 99; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1519 = ((($1515)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1519>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1515>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1520 = ((($1515)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1520>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1521 = ((($1515)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1521>>2] = $1522; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i684>>2] = $1515; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1523 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1524 = HEAP32[$1523>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1525 = ($1524|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1525) {
   $1528 = $1515;
  } else {
   $1526 = $1524; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1526; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i686 = HEAP32[$$0$i7$i684>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1528 = $$pre$i$i686;
  }
  $1527 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1527,$1528); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1529 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1530 = (($1529) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1530; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4422 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1533 = $$pre4422;$__r$0$i688 = $1515;
 }
 $1531 = ((($__r$0$i688)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1531>>2] = 46; //@line 658 "modules/event/sdl/Event.cpp"
 $1532 = ($1533|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1532) {
   HEAP32[$__parent$i691>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1545 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i699 = $2;$1555 = $1545;
   label = 656;
  } else {
   $__nd$0$i$i693 = $1533;
   while(1) {
    $1534 = ((($__nd$0$i$i693)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1535 = HEAP32[$1534>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1536 = ($1535|0)>(100); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1536) {
     $1537 = HEAP32[$__nd$0$i$i693>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1538 = ($1537|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1538) {
      $$lcssa7416 = $__nd$0$i$i693;$__nd$0$i$i693$lcssa7414 = $__nd$0$i$i693;
      label = 649;
      break;
     } else {
      $$sink$i$i694 = $1537;
     }
    } else {
     $1540 = ($1535|0)<(100); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1540)) {
      $__nd$0$i$i693$lcssa = $__nd$0$i$i693;
      label = 655;
      break;
     }
     $1541 = ((($__nd$0$i$i693)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1542 = HEAP32[$1541>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1543 = ($1542|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1543) {
      $$lcssa7415 = $1541;$__nd$0$i$i693$lcssa7413 = $__nd$0$i$i693;
      label = 653;
      break;
     } else {
      $$sink$i$i694 = $1542;
     }
    }
    $__nd$0$i$i693 = $$sink$i$i694;
   }
   if ((label|0) == 649) {
    HEAP32[$__parent$i691>>2] = $__nd$0$i$i693$lcssa7414; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1539 = $__nd$0$i$i693$lcssa7414; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i699 = $$lcssa7416;$1555 = $1539;
    label = 656;
    break;
   }
   else if ((label|0) == 653) {
    HEAP32[$__parent$i691>>2] = $__nd$0$i$i693$lcssa7413; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1544 = $__nd$0$i$i693$lcssa7413; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i699 = $$lcssa7415;$1555 = $1544;
    label = 656;
    break;
   }
   else if ((label|0) == 655) {
    HEAP32[$__parent$i691>>2] = $__nd$0$i$i693$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1546 = $__nd$0$i$i693$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1547 = ($__nd$0$i$i693$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1547) {
     $$0$i7$i699 = $__parent$i691;$1555 = $1546;
     label = 656;
     break;
    } else {
     $1566 = $1533;$__r$0$i703 = $__nd$0$i$i693$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 656) {
  __THREW__ = 0;
  $1548 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1549 = __THREW__; __THREW__ = 0;
  $1550 = $1549&1;
  if ($1550) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1551 = ((($1548)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1551>>2] = 100; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1552 = ((($1548)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1552>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1548>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1553 = ((($1548)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1553>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1554 = ((($1548)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1554>>2] = $1555; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i699>>2] = $1548; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1556 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1557 = HEAP32[$1556>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1558 = ($1557|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1558) {
   $1561 = $1548;
  } else {
   $1559 = $1557; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1559; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i701 = HEAP32[$$0$i7$i699>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1561 = $$pre$i$i701;
  }
  $1560 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1560,$1561); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1562 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1563 = (($1562) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1563; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4423 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1566 = $$pre4423;$__r$0$i703 = $1548;
 }
 $1564 = ((($__r$0$i703)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1564>>2] = 47; //@line 659 "modules/event/sdl/Event.cpp"
 $1565 = ($1566|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1565) {
   HEAP32[$__parent$i706>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1578 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i714 = $2;$1588 = $1578;
   label = 670;
  } else {
   $__nd$0$i$i708 = $1566;
   while(1) {
    $1567 = ((($__nd$0$i$i708)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1568 = HEAP32[$1567>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1569 = ($1568|0)>(101); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1569) {
     $1570 = HEAP32[$__nd$0$i$i708>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1571 = ($1570|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1571) {
      $$lcssa7412 = $__nd$0$i$i708;$__nd$0$i$i708$lcssa7410 = $__nd$0$i$i708;
      label = 663;
      break;
     } else {
      $$sink$i$i709 = $1570;
     }
    } else {
     $1573 = ($1568|0)<(101); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1573)) {
      $__nd$0$i$i708$lcssa = $__nd$0$i$i708;
      label = 669;
      break;
     }
     $1574 = ((($__nd$0$i$i708)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1575 = HEAP32[$1574>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1576 = ($1575|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1576) {
      $$lcssa7411 = $1574;$__nd$0$i$i708$lcssa7409 = $__nd$0$i$i708;
      label = 667;
      break;
     } else {
      $$sink$i$i709 = $1575;
     }
    }
    $__nd$0$i$i708 = $$sink$i$i709;
   }
   if ((label|0) == 663) {
    HEAP32[$__parent$i706>>2] = $__nd$0$i$i708$lcssa7410; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1572 = $__nd$0$i$i708$lcssa7410; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i714 = $$lcssa7412;$1588 = $1572;
    label = 670;
    break;
   }
   else if ((label|0) == 667) {
    HEAP32[$__parent$i706>>2] = $__nd$0$i$i708$lcssa7409; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1577 = $__nd$0$i$i708$lcssa7409; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i714 = $$lcssa7411;$1588 = $1577;
    label = 670;
    break;
   }
   else if ((label|0) == 669) {
    HEAP32[$__parent$i706>>2] = $__nd$0$i$i708$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1579 = $__nd$0$i$i708$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1580 = ($__nd$0$i$i708$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1580) {
     $$0$i7$i714 = $__parent$i706;$1588 = $1579;
     label = 670;
     break;
    } else {
     $1599 = $1566;$__r$0$i718 = $__nd$0$i$i708$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 670) {
  __THREW__ = 0;
  $1581 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1582 = __THREW__; __THREW__ = 0;
  $1583 = $1582&1;
  if ($1583) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1584 = ((($1581)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1584>>2] = 101; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1585 = ((($1581)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1585>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1581>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1586 = ((($1581)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1586>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1587 = ((($1581)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1587>>2] = $1588; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i714>>2] = $1581; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1589 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1590 = HEAP32[$1589>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1591 = ($1590|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1591) {
   $1594 = $1581;
  } else {
   $1592 = $1590; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1592; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i716 = HEAP32[$$0$i7$i714>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1594 = $$pre$i$i716;
  }
  $1593 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1593,$1594); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1595 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1596 = (($1595) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1596; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4424 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1599 = $$pre4424;$__r$0$i718 = $1581;
 }
 $1597 = ((($__r$0$i718)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1597>>2] = 48; //@line 660 "modules/event/sdl/Event.cpp"
 $1598 = ($1599|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1598) {
   HEAP32[$__parent$i721>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1611 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i729 = $2;$1621 = $1611;
   label = 684;
  } else {
   $__nd$0$i$i723 = $1599;
   while(1) {
    $1600 = ((($__nd$0$i$i723)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1601 = HEAP32[$1600>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1602 = ($1601|0)>(102); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1602) {
     $1603 = HEAP32[$__nd$0$i$i723>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1604 = ($1603|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1604) {
      $$lcssa7408 = $__nd$0$i$i723;$__nd$0$i$i723$lcssa7406 = $__nd$0$i$i723;
      label = 677;
      break;
     } else {
      $$sink$i$i724 = $1603;
     }
    } else {
     $1606 = ($1601|0)<(102); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1606)) {
      $__nd$0$i$i723$lcssa = $__nd$0$i$i723;
      label = 683;
      break;
     }
     $1607 = ((($__nd$0$i$i723)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1608 = HEAP32[$1607>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1609 = ($1608|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1609) {
      $$lcssa7407 = $1607;$__nd$0$i$i723$lcssa7405 = $__nd$0$i$i723;
      label = 681;
      break;
     } else {
      $$sink$i$i724 = $1608;
     }
    }
    $__nd$0$i$i723 = $$sink$i$i724;
   }
   if ((label|0) == 677) {
    HEAP32[$__parent$i721>>2] = $__nd$0$i$i723$lcssa7406; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1605 = $__nd$0$i$i723$lcssa7406; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i729 = $$lcssa7408;$1621 = $1605;
    label = 684;
    break;
   }
   else if ((label|0) == 681) {
    HEAP32[$__parent$i721>>2] = $__nd$0$i$i723$lcssa7405; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1610 = $__nd$0$i$i723$lcssa7405; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i729 = $$lcssa7407;$1621 = $1610;
    label = 684;
    break;
   }
   else if ((label|0) == 683) {
    HEAP32[$__parent$i721>>2] = $__nd$0$i$i723$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1612 = $__nd$0$i$i723$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1613 = ($__nd$0$i$i723$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1613) {
     $$0$i7$i729 = $__parent$i721;$1621 = $1612;
     label = 684;
     break;
    } else {
     $1632 = $1599;$__r$0$i733 = $__nd$0$i$i723$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 684) {
  __THREW__ = 0;
  $1614 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1615 = __THREW__; __THREW__ = 0;
  $1616 = $1615&1;
  if ($1616) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1617 = ((($1614)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1617>>2] = 102; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1618 = ((($1614)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1618>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1614>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1619 = ((($1614)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1619>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1620 = ((($1614)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1620>>2] = $1621; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i729>>2] = $1614; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1622 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1623 = HEAP32[$1622>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1624 = ($1623|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1624) {
   $1627 = $1614;
  } else {
   $1625 = $1623; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1625; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i731 = HEAP32[$$0$i7$i729>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1627 = $$pre$i$i731;
  }
  $1626 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1626,$1627); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1628 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1629 = (($1628) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1629; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4425 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1632 = $$pre4425;$__r$0$i733 = $1614;
 }
 $1630 = ((($__r$0$i733)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1630>>2] = 49; //@line 661 "modules/event/sdl/Event.cpp"
 $1631 = ($1632|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1631) {
   HEAP32[$__parent$i736>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1644 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i744 = $2;$1654 = $1644;
   label = 698;
  } else {
   $__nd$0$i$i738 = $1632;
   while(1) {
    $1633 = ((($__nd$0$i$i738)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1634 = HEAP32[$1633>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1635 = ($1634|0)>(103); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1635) {
     $1636 = HEAP32[$__nd$0$i$i738>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1637 = ($1636|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1637) {
      $$lcssa7404 = $__nd$0$i$i738;$__nd$0$i$i738$lcssa7402 = $__nd$0$i$i738;
      label = 691;
      break;
     } else {
      $$sink$i$i739 = $1636;
     }
    } else {
     $1639 = ($1634|0)<(103); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1639)) {
      $__nd$0$i$i738$lcssa = $__nd$0$i$i738;
      label = 697;
      break;
     }
     $1640 = ((($__nd$0$i$i738)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1641 = HEAP32[$1640>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1642 = ($1641|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1642) {
      $$lcssa7403 = $1640;$__nd$0$i$i738$lcssa7401 = $__nd$0$i$i738;
      label = 695;
      break;
     } else {
      $$sink$i$i739 = $1641;
     }
    }
    $__nd$0$i$i738 = $$sink$i$i739;
   }
   if ((label|0) == 691) {
    HEAP32[$__parent$i736>>2] = $__nd$0$i$i738$lcssa7402; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1638 = $__nd$0$i$i738$lcssa7402; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i744 = $$lcssa7404;$1654 = $1638;
    label = 698;
    break;
   }
   else if ((label|0) == 695) {
    HEAP32[$__parent$i736>>2] = $__nd$0$i$i738$lcssa7401; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1643 = $__nd$0$i$i738$lcssa7401; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i744 = $$lcssa7403;$1654 = $1643;
    label = 698;
    break;
   }
   else if ((label|0) == 697) {
    HEAP32[$__parent$i736>>2] = $__nd$0$i$i738$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1645 = $__nd$0$i$i738$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1646 = ($__nd$0$i$i738$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1646) {
     $$0$i7$i744 = $__parent$i736;$1654 = $1645;
     label = 698;
     break;
    } else {
     $1665 = $1632;$__r$0$i748 = $__nd$0$i$i738$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 698) {
  __THREW__ = 0;
  $1647 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1648 = __THREW__; __THREW__ = 0;
  $1649 = $1648&1;
  if ($1649) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1650 = ((($1647)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1650>>2] = 103; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1651 = ((($1647)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1651>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1647>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1652 = ((($1647)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1652>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1653 = ((($1647)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1653>>2] = $1654; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i744>>2] = $1647; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1655 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1656 = HEAP32[$1655>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1657 = ($1656|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1657) {
   $1660 = $1647;
  } else {
   $1658 = $1656; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1658; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i746 = HEAP32[$$0$i7$i744>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1660 = $$pre$i$i746;
  }
  $1659 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1659,$1660); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1661 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1662 = (($1661) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1662; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4426 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1665 = $$pre4426;$__r$0$i748 = $1647;
 }
 $1663 = ((($__r$0$i748)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1663>>2] = 50; //@line 662 "modules/event/sdl/Event.cpp"
 $1664 = ($1665|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1664) {
   HEAP32[$__parent$i751>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1677 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i759 = $2;$1687 = $1677;
   label = 712;
  } else {
   $__nd$0$i$i753 = $1665;
   while(1) {
    $1666 = ((($__nd$0$i$i753)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1667 = HEAP32[$1666>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1668 = ($1667|0)>(104); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1668) {
     $1669 = HEAP32[$__nd$0$i$i753>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1670 = ($1669|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1670) {
      $$lcssa7400 = $__nd$0$i$i753;$__nd$0$i$i753$lcssa7398 = $__nd$0$i$i753;
      label = 705;
      break;
     } else {
      $$sink$i$i754 = $1669;
     }
    } else {
     $1672 = ($1667|0)<(104); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1672)) {
      $__nd$0$i$i753$lcssa = $__nd$0$i$i753;
      label = 711;
      break;
     }
     $1673 = ((($__nd$0$i$i753)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1674 = HEAP32[$1673>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1675 = ($1674|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1675) {
      $$lcssa7399 = $1673;$__nd$0$i$i753$lcssa7397 = $__nd$0$i$i753;
      label = 709;
      break;
     } else {
      $$sink$i$i754 = $1674;
     }
    }
    $__nd$0$i$i753 = $$sink$i$i754;
   }
   if ((label|0) == 705) {
    HEAP32[$__parent$i751>>2] = $__nd$0$i$i753$lcssa7398; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1671 = $__nd$0$i$i753$lcssa7398; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i759 = $$lcssa7400;$1687 = $1671;
    label = 712;
    break;
   }
   else if ((label|0) == 709) {
    HEAP32[$__parent$i751>>2] = $__nd$0$i$i753$lcssa7397; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1676 = $__nd$0$i$i753$lcssa7397; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i759 = $$lcssa7399;$1687 = $1676;
    label = 712;
    break;
   }
   else if ((label|0) == 711) {
    HEAP32[$__parent$i751>>2] = $__nd$0$i$i753$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1678 = $__nd$0$i$i753$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1679 = ($__nd$0$i$i753$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1679) {
     $$0$i7$i759 = $__parent$i751;$1687 = $1678;
     label = 712;
     break;
    } else {
     $1698 = $1665;$__r$0$i763 = $__nd$0$i$i753$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 712) {
  __THREW__ = 0;
  $1680 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1681 = __THREW__; __THREW__ = 0;
  $1682 = $1681&1;
  if ($1682) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1683 = ((($1680)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1683>>2] = 104; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1684 = ((($1680)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1684>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1680>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1685 = ((($1680)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1685>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1686 = ((($1680)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1686>>2] = $1687; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i759>>2] = $1680; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1688 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1689 = HEAP32[$1688>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1690 = ($1689|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1690) {
   $1693 = $1680;
  } else {
   $1691 = $1689; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1691; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i761 = HEAP32[$$0$i7$i759>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1693 = $$pre$i$i761;
  }
  $1692 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1692,$1693); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1694 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1695 = (($1694) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1695; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4427 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1698 = $$pre4427;$__r$0$i763 = $1680;
 }
 $1696 = ((($__r$0$i763)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1696>>2] = 51; //@line 663 "modules/event/sdl/Event.cpp"
 $1697 = ($1698|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1697) {
   HEAP32[$__parent$i766>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1710 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i774 = $2;$1720 = $1710;
   label = 726;
  } else {
   $__nd$0$i$i768 = $1698;
   while(1) {
    $1699 = ((($__nd$0$i$i768)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1700 = HEAP32[$1699>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1701 = ($1700|0)>(105); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1701) {
     $1702 = HEAP32[$__nd$0$i$i768>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1703 = ($1702|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1703) {
      $$lcssa7396 = $__nd$0$i$i768;$__nd$0$i$i768$lcssa7394 = $__nd$0$i$i768;
      label = 719;
      break;
     } else {
      $$sink$i$i769 = $1702;
     }
    } else {
     $1705 = ($1700|0)<(105); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1705)) {
      $__nd$0$i$i768$lcssa = $__nd$0$i$i768;
      label = 725;
      break;
     }
     $1706 = ((($__nd$0$i$i768)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1707 = HEAP32[$1706>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1708 = ($1707|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1708) {
      $$lcssa7395 = $1706;$__nd$0$i$i768$lcssa7393 = $__nd$0$i$i768;
      label = 723;
      break;
     } else {
      $$sink$i$i769 = $1707;
     }
    }
    $__nd$0$i$i768 = $$sink$i$i769;
   }
   if ((label|0) == 719) {
    HEAP32[$__parent$i766>>2] = $__nd$0$i$i768$lcssa7394; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1704 = $__nd$0$i$i768$lcssa7394; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i774 = $$lcssa7396;$1720 = $1704;
    label = 726;
    break;
   }
   else if ((label|0) == 723) {
    HEAP32[$__parent$i766>>2] = $__nd$0$i$i768$lcssa7393; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1709 = $__nd$0$i$i768$lcssa7393; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i774 = $$lcssa7395;$1720 = $1709;
    label = 726;
    break;
   }
   else if ((label|0) == 725) {
    HEAP32[$__parent$i766>>2] = $__nd$0$i$i768$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1711 = $__nd$0$i$i768$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1712 = ($__nd$0$i$i768$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1712) {
     $$0$i7$i774 = $__parent$i766;$1720 = $1711;
     label = 726;
     break;
    } else {
     $1731 = $1698;$__r$0$i778 = $__nd$0$i$i768$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 726) {
  __THREW__ = 0;
  $1713 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1714 = __THREW__; __THREW__ = 0;
  $1715 = $1714&1;
  if ($1715) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1716 = ((($1713)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1716>>2] = 105; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1717 = ((($1713)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1717>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1713>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1718 = ((($1713)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1718>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1719 = ((($1713)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1719>>2] = $1720; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i774>>2] = $1713; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1721 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1722 = HEAP32[$1721>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1723 = ($1722|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1723) {
   $1726 = $1713;
  } else {
   $1724 = $1722; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1724; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i776 = HEAP32[$$0$i7$i774>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1726 = $$pre$i$i776;
  }
  $1725 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1725,$1726); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1727 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1728 = (($1727) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1728; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4428 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1731 = $$pre4428;$__r$0$i778 = $1713;
 }
 $1729 = ((($__r$0$i778)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1729>>2] = 52; //@line 664 "modules/event/sdl/Event.cpp"
 $1730 = ($1731|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1730) {
   HEAP32[$__parent$i781>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1743 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i789 = $2;$1753 = $1743;
   label = 740;
  } else {
   $__nd$0$i$i783 = $1731;
   while(1) {
    $1732 = ((($__nd$0$i$i783)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1733 = HEAP32[$1732>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1734 = ($1733|0)>(106); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1734) {
     $1735 = HEAP32[$__nd$0$i$i783>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1736 = ($1735|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1736) {
      $$lcssa7392 = $__nd$0$i$i783;$__nd$0$i$i783$lcssa7390 = $__nd$0$i$i783;
      label = 733;
      break;
     } else {
      $$sink$i$i784 = $1735;
     }
    } else {
     $1738 = ($1733|0)<(106); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1738)) {
      $__nd$0$i$i783$lcssa = $__nd$0$i$i783;
      label = 739;
      break;
     }
     $1739 = ((($__nd$0$i$i783)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1740 = HEAP32[$1739>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1741 = ($1740|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1741) {
      $$lcssa7391 = $1739;$__nd$0$i$i783$lcssa7389 = $__nd$0$i$i783;
      label = 737;
      break;
     } else {
      $$sink$i$i784 = $1740;
     }
    }
    $__nd$0$i$i783 = $$sink$i$i784;
   }
   if ((label|0) == 733) {
    HEAP32[$__parent$i781>>2] = $__nd$0$i$i783$lcssa7390; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1737 = $__nd$0$i$i783$lcssa7390; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i789 = $$lcssa7392;$1753 = $1737;
    label = 740;
    break;
   }
   else if ((label|0) == 737) {
    HEAP32[$__parent$i781>>2] = $__nd$0$i$i783$lcssa7389; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1742 = $__nd$0$i$i783$lcssa7389; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i789 = $$lcssa7391;$1753 = $1742;
    label = 740;
    break;
   }
   else if ((label|0) == 739) {
    HEAP32[$__parent$i781>>2] = $__nd$0$i$i783$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1744 = $__nd$0$i$i783$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1745 = ($__nd$0$i$i783$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1745) {
     $$0$i7$i789 = $__parent$i781;$1753 = $1744;
     label = 740;
     break;
    } else {
     $1764 = $1731;$__r$0$i793 = $__nd$0$i$i783$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 740) {
  __THREW__ = 0;
  $1746 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1747 = __THREW__; __THREW__ = 0;
  $1748 = $1747&1;
  if ($1748) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1749 = ((($1746)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1749>>2] = 106; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1750 = ((($1746)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1750>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1746>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1751 = ((($1746)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1751>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1752 = ((($1746)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1752>>2] = $1753; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i789>>2] = $1746; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1754 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1755 = HEAP32[$1754>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1756 = ($1755|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1756) {
   $1759 = $1746;
  } else {
   $1757 = $1755; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1757; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i791 = HEAP32[$$0$i7$i789>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1759 = $$pre$i$i791;
  }
  $1758 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1758,$1759); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1760 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1761 = (($1760) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1761; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4429 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1764 = $$pre4429;$__r$0$i793 = $1746;
 }
 $1762 = ((($__r$0$i793)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1762>>2] = 53; //@line 665 "modules/event/sdl/Event.cpp"
 $1763 = ($1764|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1763) {
   HEAP32[$__parent$i796>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1776 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i804 = $2;$1786 = $1776;
   label = 754;
  } else {
   $__nd$0$i$i798 = $1764;
   while(1) {
    $1765 = ((($__nd$0$i$i798)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1766 = HEAP32[$1765>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1767 = ($1766|0)>(107); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1767) {
     $1768 = HEAP32[$__nd$0$i$i798>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1769 = ($1768|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1769) {
      $$lcssa7388 = $__nd$0$i$i798;$__nd$0$i$i798$lcssa7386 = $__nd$0$i$i798;
      label = 747;
      break;
     } else {
      $$sink$i$i799 = $1768;
     }
    } else {
     $1771 = ($1766|0)<(107); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1771)) {
      $__nd$0$i$i798$lcssa = $__nd$0$i$i798;
      label = 753;
      break;
     }
     $1772 = ((($__nd$0$i$i798)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1773 = HEAP32[$1772>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1774 = ($1773|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1774) {
      $$lcssa7387 = $1772;$__nd$0$i$i798$lcssa7385 = $__nd$0$i$i798;
      label = 751;
      break;
     } else {
      $$sink$i$i799 = $1773;
     }
    }
    $__nd$0$i$i798 = $$sink$i$i799;
   }
   if ((label|0) == 747) {
    HEAP32[$__parent$i796>>2] = $__nd$0$i$i798$lcssa7386; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1770 = $__nd$0$i$i798$lcssa7386; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i804 = $$lcssa7388;$1786 = $1770;
    label = 754;
    break;
   }
   else if ((label|0) == 751) {
    HEAP32[$__parent$i796>>2] = $__nd$0$i$i798$lcssa7385; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1775 = $__nd$0$i$i798$lcssa7385; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i804 = $$lcssa7387;$1786 = $1775;
    label = 754;
    break;
   }
   else if ((label|0) == 753) {
    HEAP32[$__parent$i796>>2] = $__nd$0$i$i798$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1777 = $__nd$0$i$i798$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1778 = ($__nd$0$i$i798$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1778) {
     $$0$i7$i804 = $__parent$i796;$1786 = $1777;
     label = 754;
     break;
    } else {
     $1797 = $1764;$__r$0$i808 = $__nd$0$i$i798$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 754) {
  __THREW__ = 0;
  $1779 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1780 = __THREW__; __THREW__ = 0;
  $1781 = $1780&1;
  if ($1781) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1782 = ((($1779)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1782>>2] = 107; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1783 = ((($1779)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1783>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1779>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1784 = ((($1779)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1784>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1785 = ((($1779)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1785>>2] = $1786; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i804>>2] = $1779; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1787 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1788 = HEAP32[$1787>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1789 = ($1788|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1789) {
   $1792 = $1779;
  } else {
   $1790 = $1788; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1790; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i806 = HEAP32[$$0$i7$i804>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1792 = $$pre$i$i806;
  }
  $1791 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1791,$1792); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1793 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1794 = (($1793) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1794; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4430 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1797 = $$pre4430;$__r$0$i808 = $1779;
 }
 $1795 = ((($__r$0$i808)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1795>>2] = 54; //@line 666 "modules/event/sdl/Event.cpp"
 $1796 = ($1797|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1796) {
   HEAP32[$__parent$i811>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1809 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i819 = $2;$1819 = $1809;
   label = 768;
  } else {
   $__nd$0$i$i813 = $1797;
   while(1) {
    $1798 = ((($__nd$0$i$i813)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1799 = HEAP32[$1798>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1800 = ($1799|0)>(108); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1800) {
     $1801 = HEAP32[$__nd$0$i$i813>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1802 = ($1801|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1802) {
      $$lcssa7384 = $__nd$0$i$i813;$__nd$0$i$i813$lcssa7382 = $__nd$0$i$i813;
      label = 761;
      break;
     } else {
      $$sink$i$i814 = $1801;
     }
    } else {
     $1804 = ($1799|0)<(108); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1804)) {
      $__nd$0$i$i813$lcssa = $__nd$0$i$i813;
      label = 767;
      break;
     }
     $1805 = ((($__nd$0$i$i813)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1806 = HEAP32[$1805>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1807 = ($1806|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1807) {
      $$lcssa7383 = $1805;$__nd$0$i$i813$lcssa7381 = $__nd$0$i$i813;
      label = 765;
      break;
     } else {
      $$sink$i$i814 = $1806;
     }
    }
    $__nd$0$i$i813 = $$sink$i$i814;
   }
   if ((label|0) == 761) {
    HEAP32[$__parent$i811>>2] = $__nd$0$i$i813$lcssa7382; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1803 = $__nd$0$i$i813$lcssa7382; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i819 = $$lcssa7384;$1819 = $1803;
    label = 768;
    break;
   }
   else if ((label|0) == 765) {
    HEAP32[$__parent$i811>>2] = $__nd$0$i$i813$lcssa7381; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1808 = $__nd$0$i$i813$lcssa7381; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i819 = $$lcssa7383;$1819 = $1808;
    label = 768;
    break;
   }
   else if ((label|0) == 767) {
    HEAP32[$__parent$i811>>2] = $__nd$0$i$i813$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1810 = $__nd$0$i$i813$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1811 = ($__nd$0$i$i813$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1811) {
     $$0$i7$i819 = $__parent$i811;$1819 = $1810;
     label = 768;
     break;
    } else {
     $1830 = $1797;$__r$0$i823 = $__nd$0$i$i813$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 768) {
  __THREW__ = 0;
  $1812 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1813 = __THREW__; __THREW__ = 0;
  $1814 = $1813&1;
  if ($1814) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1815 = ((($1812)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1815>>2] = 108; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1816 = ((($1812)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1816>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1812>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1817 = ((($1812)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1817>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1818 = ((($1812)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1818>>2] = $1819; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i819>>2] = $1812; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1820 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1821 = HEAP32[$1820>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1822 = ($1821|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1822) {
   $1825 = $1812;
  } else {
   $1823 = $1821; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1823; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i821 = HEAP32[$$0$i7$i819>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1825 = $$pre$i$i821;
  }
  $1824 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1824,$1825); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1826 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1827 = (($1826) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1827; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4431 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1830 = $$pre4431;$__r$0$i823 = $1812;
 }
 $1828 = ((($__r$0$i823)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1828>>2] = 55; //@line 667 "modules/event/sdl/Event.cpp"
 $1829 = ($1830|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1829) {
   HEAP32[$__parent$i826>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1842 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i834 = $2;$1852 = $1842;
   label = 782;
  } else {
   $__nd$0$i$i828 = $1830;
   while(1) {
    $1831 = ((($__nd$0$i$i828)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1832 = HEAP32[$1831>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1833 = ($1832|0)>(109); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1833) {
     $1834 = HEAP32[$__nd$0$i$i828>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1835 = ($1834|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1835) {
      $$lcssa7380 = $__nd$0$i$i828;$__nd$0$i$i828$lcssa7378 = $__nd$0$i$i828;
      label = 775;
      break;
     } else {
      $$sink$i$i829 = $1834;
     }
    } else {
     $1837 = ($1832|0)<(109); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1837)) {
      $__nd$0$i$i828$lcssa = $__nd$0$i$i828;
      label = 781;
      break;
     }
     $1838 = ((($__nd$0$i$i828)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1839 = HEAP32[$1838>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1840 = ($1839|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1840) {
      $$lcssa7379 = $1838;$__nd$0$i$i828$lcssa7377 = $__nd$0$i$i828;
      label = 779;
      break;
     } else {
      $$sink$i$i829 = $1839;
     }
    }
    $__nd$0$i$i828 = $$sink$i$i829;
   }
   if ((label|0) == 775) {
    HEAP32[$__parent$i826>>2] = $__nd$0$i$i828$lcssa7378; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1836 = $__nd$0$i$i828$lcssa7378; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i834 = $$lcssa7380;$1852 = $1836;
    label = 782;
    break;
   }
   else if ((label|0) == 779) {
    HEAP32[$__parent$i826>>2] = $__nd$0$i$i828$lcssa7377; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1841 = $__nd$0$i$i828$lcssa7377; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i834 = $$lcssa7379;$1852 = $1841;
    label = 782;
    break;
   }
   else if ((label|0) == 781) {
    HEAP32[$__parent$i826>>2] = $__nd$0$i$i828$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1843 = $__nd$0$i$i828$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1844 = ($__nd$0$i$i828$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1844) {
     $$0$i7$i834 = $__parent$i826;$1852 = $1843;
     label = 782;
     break;
    } else {
     $1863 = $1830;$__r$0$i838 = $__nd$0$i$i828$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 782) {
  __THREW__ = 0;
  $1845 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1846 = __THREW__; __THREW__ = 0;
  $1847 = $1846&1;
  if ($1847) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1848 = ((($1845)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1848>>2] = 109; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1849 = ((($1845)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1849>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1845>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1850 = ((($1845)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1850>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1851 = ((($1845)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1851>>2] = $1852; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i834>>2] = $1845; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1853 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1854 = HEAP32[$1853>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1855 = ($1854|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1855) {
   $1858 = $1845;
  } else {
   $1856 = $1854; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1856; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i836 = HEAP32[$$0$i7$i834>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1858 = $$pre$i$i836;
  }
  $1857 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1857,$1858); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1859 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1860 = (($1859) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1860; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4432 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1863 = $$pre4432;$__r$0$i838 = $1845;
 }
 $1861 = ((($__r$0$i838)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1861>>2] = 56; //@line 668 "modules/event/sdl/Event.cpp"
 $1862 = ($1863|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1862) {
   HEAP32[$__parent$i841>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1875 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i849 = $2;$1885 = $1875;
   label = 796;
  } else {
   $__nd$0$i$i843 = $1863;
   while(1) {
    $1864 = ((($__nd$0$i$i843)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1865 = HEAP32[$1864>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1866 = ($1865|0)>(110); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1866) {
     $1867 = HEAP32[$__nd$0$i$i843>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1868 = ($1867|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1868) {
      $$lcssa7376 = $__nd$0$i$i843;$__nd$0$i$i843$lcssa7374 = $__nd$0$i$i843;
      label = 789;
      break;
     } else {
      $$sink$i$i844 = $1867;
     }
    } else {
     $1870 = ($1865|0)<(110); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1870)) {
      $__nd$0$i$i843$lcssa = $__nd$0$i$i843;
      label = 795;
      break;
     }
     $1871 = ((($__nd$0$i$i843)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1872 = HEAP32[$1871>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1873 = ($1872|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1873) {
      $$lcssa7375 = $1871;$__nd$0$i$i843$lcssa7373 = $__nd$0$i$i843;
      label = 793;
      break;
     } else {
      $$sink$i$i844 = $1872;
     }
    }
    $__nd$0$i$i843 = $$sink$i$i844;
   }
   if ((label|0) == 789) {
    HEAP32[$__parent$i841>>2] = $__nd$0$i$i843$lcssa7374; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1869 = $__nd$0$i$i843$lcssa7374; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i849 = $$lcssa7376;$1885 = $1869;
    label = 796;
    break;
   }
   else if ((label|0) == 793) {
    HEAP32[$__parent$i841>>2] = $__nd$0$i$i843$lcssa7373; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1874 = $__nd$0$i$i843$lcssa7373; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i849 = $$lcssa7375;$1885 = $1874;
    label = 796;
    break;
   }
   else if ((label|0) == 795) {
    HEAP32[$__parent$i841>>2] = $__nd$0$i$i843$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1876 = $__nd$0$i$i843$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1877 = ($__nd$0$i$i843$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1877) {
     $$0$i7$i849 = $__parent$i841;$1885 = $1876;
     label = 796;
     break;
    } else {
     $1896 = $1863;$__r$0$i853 = $__nd$0$i$i843$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 796) {
  __THREW__ = 0;
  $1878 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1879 = __THREW__; __THREW__ = 0;
  $1880 = $1879&1;
  if ($1880) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1881 = ((($1878)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1881>>2] = 110; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1882 = ((($1878)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1882>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1878>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1883 = ((($1878)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1883>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1884 = ((($1878)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1884>>2] = $1885; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i849>>2] = $1878; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1886 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1887 = HEAP32[$1886>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1888 = ($1887|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1888) {
   $1891 = $1878;
  } else {
   $1889 = $1887; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1889; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i851 = HEAP32[$$0$i7$i849>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1891 = $$pre$i$i851;
  }
  $1890 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1890,$1891); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1892 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1893 = (($1892) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1893; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4433 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1896 = $$pre4433;$__r$0$i853 = $1878;
 }
 $1894 = ((($__r$0$i853)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1894>>2] = 57; //@line 669 "modules/event/sdl/Event.cpp"
 $1895 = ($1896|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1895) {
   HEAP32[$__parent$i856>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1908 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i864 = $2;$1918 = $1908;
   label = 810;
  } else {
   $__nd$0$i$i858 = $1896;
   while(1) {
    $1897 = ((($__nd$0$i$i858)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1898 = HEAP32[$1897>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1899 = ($1898|0)>(111); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1899) {
     $1900 = HEAP32[$__nd$0$i$i858>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1901 = ($1900|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1901) {
      $$lcssa7372 = $__nd$0$i$i858;$__nd$0$i$i858$lcssa7370 = $__nd$0$i$i858;
      label = 803;
      break;
     } else {
      $$sink$i$i859 = $1900;
     }
    } else {
     $1903 = ($1898|0)<(111); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1903)) {
      $__nd$0$i$i858$lcssa = $__nd$0$i$i858;
      label = 809;
      break;
     }
     $1904 = ((($__nd$0$i$i858)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1905 = HEAP32[$1904>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1906 = ($1905|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1906) {
      $$lcssa7371 = $1904;$__nd$0$i$i858$lcssa7369 = $__nd$0$i$i858;
      label = 807;
      break;
     } else {
      $$sink$i$i859 = $1905;
     }
    }
    $__nd$0$i$i858 = $$sink$i$i859;
   }
   if ((label|0) == 803) {
    HEAP32[$__parent$i856>>2] = $__nd$0$i$i858$lcssa7370; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1902 = $__nd$0$i$i858$lcssa7370; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i864 = $$lcssa7372;$1918 = $1902;
    label = 810;
    break;
   }
   else if ((label|0) == 807) {
    HEAP32[$__parent$i856>>2] = $__nd$0$i$i858$lcssa7369; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1907 = $__nd$0$i$i858$lcssa7369; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i864 = $$lcssa7371;$1918 = $1907;
    label = 810;
    break;
   }
   else if ((label|0) == 809) {
    HEAP32[$__parent$i856>>2] = $__nd$0$i$i858$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1909 = $__nd$0$i$i858$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1910 = ($__nd$0$i$i858$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1910) {
     $$0$i7$i864 = $__parent$i856;$1918 = $1909;
     label = 810;
     break;
    } else {
     $1929 = $1896;$__r$0$i868 = $__nd$0$i$i858$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 810) {
  __THREW__ = 0;
  $1911 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1912 = __THREW__; __THREW__ = 0;
  $1913 = $1912&1;
  if ($1913) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1914 = ((($1911)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1914>>2] = 111; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1915 = ((($1911)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1915>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1911>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1916 = ((($1911)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1916>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1917 = ((($1911)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1917>>2] = $1918; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i864>>2] = $1911; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1919 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1920 = HEAP32[$1919>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1921 = ($1920|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1921) {
   $1924 = $1911;
  } else {
   $1922 = $1920; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1922; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i866 = HEAP32[$$0$i7$i864>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1924 = $$pre$i$i866;
  }
  $1923 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1923,$1924); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1925 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1926 = (($1925) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1926; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4434 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1929 = $$pre4434;$__r$0$i868 = $1911;
 }
 $1927 = ((($__r$0$i868)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1927>>2] = 58; //@line 670 "modules/event/sdl/Event.cpp"
 $1928 = ($1929|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1928) {
   HEAP32[$__parent$i871>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1941 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i879 = $2;$1951 = $1941;
   label = 824;
  } else {
   $__nd$0$i$i873 = $1929;
   while(1) {
    $1930 = ((($__nd$0$i$i873)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1931 = HEAP32[$1930>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1932 = ($1931|0)>(112); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1932) {
     $1933 = HEAP32[$__nd$0$i$i873>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1934 = ($1933|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1934) {
      $$lcssa7368 = $__nd$0$i$i873;$__nd$0$i$i873$lcssa7366 = $__nd$0$i$i873;
      label = 817;
      break;
     } else {
      $$sink$i$i874 = $1933;
     }
    } else {
     $1936 = ($1931|0)<(112); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1936)) {
      $__nd$0$i$i873$lcssa = $__nd$0$i$i873;
      label = 823;
      break;
     }
     $1937 = ((($__nd$0$i$i873)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1938 = HEAP32[$1937>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1939 = ($1938|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1939) {
      $$lcssa7367 = $1937;$__nd$0$i$i873$lcssa7365 = $__nd$0$i$i873;
      label = 821;
      break;
     } else {
      $$sink$i$i874 = $1938;
     }
    }
    $__nd$0$i$i873 = $$sink$i$i874;
   }
   if ((label|0) == 817) {
    HEAP32[$__parent$i871>>2] = $__nd$0$i$i873$lcssa7366; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1935 = $__nd$0$i$i873$lcssa7366; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i879 = $$lcssa7368;$1951 = $1935;
    label = 824;
    break;
   }
   else if ((label|0) == 821) {
    HEAP32[$__parent$i871>>2] = $__nd$0$i$i873$lcssa7365; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1940 = $__nd$0$i$i873$lcssa7365; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i879 = $$lcssa7367;$1951 = $1940;
    label = 824;
    break;
   }
   else if ((label|0) == 823) {
    HEAP32[$__parent$i871>>2] = $__nd$0$i$i873$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1942 = $__nd$0$i$i873$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1943 = ($__nd$0$i$i873$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1943) {
     $$0$i7$i879 = $__parent$i871;$1951 = $1942;
     label = 824;
     break;
    } else {
     $1962 = $1929;$__r$0$i883 = $__nd$0$i$i873$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 824) {
  __THREW__ = 0;
  $1944 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1945 = __THREW__; __THREW__ = 0;
  $1946 = $1945&1;
  if ($1946) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1947 = ((($1944)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1947>>2] = 112; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1948 = ((($1944)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1948>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1944>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1949 = ((($1944)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1949>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1950 = ((($1944)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1950>>2] = $1951; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i879>>2] = $1944; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1952 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1953 = HEAP32[$1952>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1954 = ($1953|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1954) {
   $1957 = $1944;
  } else {
   $1955 = $1953; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1955; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i881 = HEAP32[$$0$i7$i879>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1957 = $$pre$i$i881;
  }
  $1956 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1956,$1957); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1958 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1959 = (($1958) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1959; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4435 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1962 = $$pre4435;$__r$0$i883 = $1944;
 }
 $1960 = ((($__r$0$i883)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1960>>2] = 59; //@line 671 "modules/event/sdl/Event.cpp"
 $1961 = ($1962|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1961) {
   HEAP32[$__parent$i886>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $1974 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i894 = $2;$1984 = $1974;
   label = 838;
  } else {
   $__nd$0$i$i888 = $1962;
   while(1) {
    $1963 = ((($__nd$0$i$i888)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1964 = HEAP32[$1963>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1965 = ($1964|0)>(113); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1965) {
     $1966 = HEAP32[$__nd$0$i$i888>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1967 = ($1966|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1967) {
      $$lcssa7364 = $__nd$0$i$i888;$__nd$0$i$i888$lcssa7362 = $__nd$0$i$i888;
      label = 831;
      break;
     } else {
      $$sink$i$i889 = $1966;
     }
    } else {
     $1969 = ($1964|0)<(113); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($1969)) {
      $__nd$0$i$i888$lcssa = $__nd$0$i$i888;
      label = 837;
      break;
     }
     $1970 = ((($__nd$0$i$i888)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1971 = HEAP32[$1970>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $1972 = ($1971|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($1972) {
      $$lcssa7363 = $1970;$__nd$0$i$i888$lcssa7361 = $__nd$0$i$i888;
      label = 835;
      break;
     } else {
      $$sink$i$i889 = $1971;
     }
    }
    $__nd$0$i$i888 = $$sink$i$i889;
   }
   if ((label|0) == 831) {
    HEAP32[$__parent$i886>>2] = $__nd$0$i$i888$lcssa7362; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1968 = $__nd$0$i$i888$lcssa7362; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i894 = $$lcssa7364;$1984 = $1968;
    label = 838;
    break;
   }
   else if ((label|0) == 835) {
    HEAP32[$__parent$i886>>2] = $__nd$0$i$i888$lcssa7361; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1973 = $__nd$0$i$i888$lcssa7361; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i894 = $$lcssa7363;$1984 = $1973;
    label = 838;
    break;
   }
   else if ((label|0) == 837) {
    HEAP32[$__parent$i886>>2] = $__nd$0$i$i888$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1975 = $__nd$0$i$i888$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1976 = ($__nd$0$i$i888$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($1976) {
     $$0$i7$i894 = $__parent$i886;$1984 = $1975;
     label = 838;
     break;
    } else {
     $1995 = $1962;$__r$0$i898 = $__nd$0$i$i888$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 838) {
  __THREW__ = 0;
  $1977 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $1978 = __THREW__; __THREW__ = 0;
  $1979 = $1978&1;
  if ($1979) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $1980 = ((($1977)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1980>>2] = 113; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $1981 = ((($1977)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$1981>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$1977>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1982 = ((($1977)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1982>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1983 = ((($1977)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1983>>2] = $1984; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i894>>2] = $1977; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1985 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1986 = HEAP32[$1985>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1987 = ($1986|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($1987) {
   $1990 = $1977;
  } else {
   $1988 = $1986; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $1988; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i896 = HEAP32[$$0$i7$i894>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $1990 = $$pre$i$i896;
  }
  $1989 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($1989,$1990); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1991 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1992 = (($1991) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $1992; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4436 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $1995 = $$pre4436;$__r$0$i898 = $1977;
 }
 $1993 = ((($__r$0$i898)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$1993>>2] = 60; //@line 672 "modules/event/sdl/Event.cpp"
 $1994 = ($1995|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($1994) {
   HEAP32[$__parent$i901>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2007 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i909 = $2;$2017 = $2007;
   label = 852;
  } else {
   $__nd$0$i$i903 = $1995;
   while(1) {
    $1996 = ((($__nd$0$i$i903)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $1997 = HEAP32[$1996>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $1998 = ($1997|0)>(114); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($1998) {
     $1999 = HEAP32[$__nd$0$i$i903>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2000 = ($1999|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2000) {
      $$lcssa7360 = $__nd$0$i$i903;$__nd$0$i$i903$lcssa7358 = $__nd$0$i$i903;
      label = 845;
      break;
     } else {
      $$sink$i$i904 = $1999;
     }
    } else {
     $2002 = ($1997|0)<(114); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2002)) {
      $__nd$0$i$i903$lcssa = $__nd$0$i$i903;
      label = 851;
      break;
     }
     $2003 = ((($__nd$0$i$i903)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2004 = HEAP32[$2003>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2005 = ($2004|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2005) {
      $$lcssa7359 = $2003;$__nd$0$i$i903$lcssa7357 = $__nd$0$i$i903;
      label = 849;
      break;
     } else {
      $$sink$i$i904 = $2004;
     }
    }
    $__nd$0$i$i903 = $$sink$i$i904;
   }
   if ((label|0) == 845) {
    HEAP32[$__parent$i901>>2] = $__nd$0$i$i903$lcssa7358; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2001 = $__nd$0$i$i903$lcssa7358; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i909 = $$lcssa7360;$2017 = $2001;
    label = 852;
    break;
   }
   else if ((label|0) == 849) {
    HEAP32[$__parent$i901>>2] = $__nd$0$i$i903$lcssa7357; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2006 = $__nd$0$i$i903$lcssa7357; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i909 = $$lcssa7359;$2017 = $2006;
    label = 852;
    break;
   }
   else if ((label|0) == 851) {
    HEAP32[$__parent$i901>>2] = $__nd$0$i$i903$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2008 = $__nd$0$i$i903$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2009 = ($__nd$0$i$i903$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2009) {
     $$0$i7$i909 = $__parent$i901;$2017 = $2008;
     label = 852;
     break;
    } else {
     $2028 = $1995;$__r$0$i913 = $__nd$0$i$i903$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 852) {
  __THREW__ = 0;
  $2010 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2011 = __THREW__; __THREW__ = 0;
  $2012 = $2011&1;
  if ($2012) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2013 = ((($2010)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2013>>2] = 114; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2014 = ((($2010)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2014>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2010>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2015 = ((($2010)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2015>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2016 = ((($2010)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2016>>2] = $2017; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i909>>2] = $2010; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2018 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2019 = HEAP32[$2018>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2020 = ($2019|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2020) {
   $2023 = $2010;
  } else {
   $2021 = $2019; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2021; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i911 = HEAP32[$$0$i7$i909>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2023 = $$pre$i$i911;
  }
  $2022 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2022,$2023); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2024 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2025 = (($2024) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2025; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4437 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2028 = $$pre4437;$__r$0$i913 = $2010;
 }
 $2026 = ((($__r$0$i913)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2026>>2] = 61; //@line 673 "modules/event/sdl/Event.cpp"
 $2027 = ($2028|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2027) {
   HEAP32[$__parent$i916>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2040 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i924 = $2;$2050 = $2040;
   label = 866;
  } else {
   $__nd$0$i$i918 = $2028;
   while(1) {
    $2029 = ((($__nd$0$i$i918)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2030 = HEAP32[$2029>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2031 = ($2030|0)>(115); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2031) {
     $2032 = HEAP32[$__nd$0$i$i918>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2033 = ($2032|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2033) {
      $$lcssa7356 = $__nd$0$i$i918;$__nd$0$i$i918$lcssa7354 = $__nd$0$i$i918;
      label = 859;
      break;
     } else {
      $$sink$i$i919 = $2032;
     }
    } else {
     $2035 = ($2030|0)<(115); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2035)) {
      $__nd$0$i$i918$lcssa = $__nd$0$i$i918;
      label = 865;
      break;
     }
     $2036 = ((($__nd$0$i$i918)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2037 = HEAP32[$2036>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2038 = ($2037|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2038) {
      $$lcssa7355 = $2036;$__nd$0$i$i918$lcssa7353 = $__nd$0$i$i918;
      label = 863;
      break;
     } else {
      $$sink$i$i919 = $2037;
     }
    }
    $__nd$0$i$i918 = $$sink$i$i919;
   }
   if ((label|0) == 859) {
    HEAP32[$__parent$i916>>2] = $__nd$0$i$i918$lcssa7354; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2034 = $__nd$0$i$i918$lcssa7354; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i924 = $$lcssa7356;$2050 = $2034;
    label = 866;
    break;
   }
   else if ((label|0) == 863) {
    HEAP32[$__parent$i916>>2] = $__nd$0$i$i918$lcssa7353; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2039 = $__nd$0$i$i918$lcssa7353; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i924 = $$lcssa7355;$2050 = $2039;
    label = 866;
    break;
   }
   else if ((label|0) == 865) {
    HEAP32[$__parent$i916>>2] = $__nd$0$i$i918$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2041 = $__nd$0$i$i918$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2042 = ($__nd$0$i$i918$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2042) {
     $$0$i7$i924 = $__parent$i916;$2050 = $2041;
     label = 866;
     break;
    } else {
     $2061 = $2028;$__r$0$i928 = $__nd$0$i$i918$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 866) {
  __THREW__ = 0;
  $2043 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2044 = __THREW__; __THREW__ = 0;
  $2045 = $2044&1;
  if ($2045) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2046 = ((($2043)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2046>>2] = 115; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2047 = ((($2043)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2047>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2043>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2048 = ((($2043)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2048>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2049 = ((($2043)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2049>>2] = $2050; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i924>>2] = $2043; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2051 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2052 = HEAP32[$2051>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2053 = ($2052|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2053) {
   $2056 = $2043;
  } else {
   $2054 = $2052; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2054; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i926 = HEAP32[$$0$i7$i924>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2056 = $$pre$i$i926;
  }
  $2055 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2055,$2056); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2057 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2058 = (($2057) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2058; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4438 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2061 = $$pre4438;$__r$0$i928 = $2043;
 }
 $2059 = ((($__r$0$i928)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2059>>2] = 62; //@line 674 "modules/event/sdl/Event.cpp"
 $2060 = ($2061|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2060) {
   HEAP32[$__parent$i931>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2073 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i939 = $2;$2083 = $2073;
   label = 880;
  } else {
   $__nd$0$i$i933 = $2061;
   while(1) {
    $2062 = ((($__nd$0$i$i933)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2063 = HEAP32[$2062>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2064 = ($2063|0)>(116); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2064) {
     $2065 = HEAP32[$__nd$0$i$i933>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2066 = ($2065|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2066) {
      $$lcssa7352 = $__nd$0$i$i933;$__nd$0$i$i933$lcssa7350 = $__nd$0$i$i933;
      label = 873;
      break;
     } else {
      $$sink$i$i934 = $2065;
     }
    } else {
     $2068 = ($2063|0)<(116); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2068)) {
      $__nd$0$i$i933$lcssa = $__nd$0$i$i933;
      label = 879;
      break;
     }
     $2069 = ((($__nd$0$i$i933)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2070 = HEAP32[$2069>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2071 = ($2070|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2071) {
      $$lcssa7351 = $2069;$__nd$0$i$i933$lcssa7349 = $__nd$0$i$i933;
      label = 877;
      break;
     } else {
      $$sink$i$i934 = $2070;
     }
    }
    $__nd$0$i$i933 = $$sink$i$i934;
   }
   if ((label|0) == 873) {
    HEAP32[$__parent$i931>>2] = $__nd$0$i$i933$lcssa7350; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2067 = $__nd$0$i$i933$lcssa7350; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i939 = $$lcssa7352;$2083 = $2067;
    label = 880;
    break;
   }
   else if ((label|0) == 877) {
    HEAP32[$__parent$i931>>2] = $__nd$0$i$i933$lcssa7349; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2072 = $__nd$0$i$i933$lcssa7349; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i939 = $$lcssa7351;$2083 = $2072;
    label = 880;
    break;
   }
   else if ((label|0) == 879) {
    HEAP32[$__parent$i931>>2] = $__nd$0$i$i933$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2074 = $__nd$0$i$i933$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2075 = ($__nd$0$i$i933$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2075) {
     $$0$i7$i939 = $__parent$i931;$2083 = $2074;
     label = 880;
     break;
    } else {
     $2094 = $2061;$__r$0$i943 = $__nd$0$i$i933$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 880) {
  __THREW__ = 0;
  $2076 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2077 = __THREW__; __THREW__ = 0;
  $2078 = $2077&1;
  if ($2078) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2079 = ((($2076)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2079>>2] = 116; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2080 = ((($2076)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2080>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2076>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2081 = ((($2076)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2081>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2082 = ((($2076)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2082>>2] = $2083; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i939>>2] = $2076; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2084 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2085 = HEAP32[$2084>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2086 = ($2085|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2086) {
   $2089 = $2076;
  } else {
   $2087 = $2085; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2087; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i941 = HEAP32[$$0$i7$i939>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2089 = $$pre$i$i941;
  }
  $2088 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2088,$2089); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2090 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2091 = (($2090) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2091; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4439 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2094 = $$pre4439;$__r$0$i943 = $2076;
 }
 $2092 = ((($__r$0$i943)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2092>>2] = 63; //@line 675 "modules/event/sdl/Event.cpp"
 $2093 = ($2094|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2093) {
   HEAP32[$__parent$i946>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2106 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i954 = $2;$2116 = $2106;
   label = 894;
  } else {
   $__nd$0$i$i948 = $2094;
   while(1) {
    $2095 = ((($__nd$0$i$i948)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2096 = HEAP32[$2095>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2097 = ($2096|0)>(117); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2097) {
     $2098 = HEAP32[$__nd$0$i$i948>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2099 = ($2098|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2099) {
      $$lcssa7348 = $__nd$0$i$i948;$__nd$0$i$i948$lcssa7346 = $__nd$0$i$i948;
      label = 887;
      break;
     } else {
      $$sink$i$i949 = $2098;
     }
    } else {
     $2101 = ($2096|0)<(117); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2101)) {
      $__nd$0$i$i948$lcssa = $__nd$0$i$i948;
      label = 893;
      break;
     }
     $2102 = ((($__nd$0$i$i948)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2103 = HEAP32[$2102>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2104 = ($2103|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2104) {
      $$lcssa7347 = $2102;$__nd$0$i$i948$lcssa7345 = $__nd$0$i$i948;
      label = 891;
      break;
     } else {
      $$sink$i$i949 = $2103;
     }
    }
    $__nd$0$i$i948 = $$sink$i$i949;
   }
   if ((label|0) == 887) {
    HEAP32[$__parent$i946>>2] = $__nd$0$i$i948$lcssa7346; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2100 = $__nd$0$i$i948$lcssa7346; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i954 = $$lcssa7348;$2116 = $2100;
    label = 894;
    break;
   }
   else if ((label|0) == 891) {
    HEAP32[$__parent$i946>>2] = $__nd$0$i$i948$lcssa7345; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2105 = $__nd$0$i$i948$lcssa7345; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i954 = $$lcssa7347;$2116 = $2105;
    label = 894;
    break;
   }
   else if ((label|0) == 893) {
    HEAP32[$__parent$i946>>2] = $__nd$0$i$i948$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2107 = $__nd$0$i$i948$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2108 = ($__nd$0$i$i948$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2108) {
     $$0$i7$i954 = $__parent$i946;$2116 = $2107;
     label = 894;
     break;
    } else {
     $2127 = $2094;$__r$0$i958 = $__nd$0$i$i948$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 894) {
  __THREW__ = 0;
  $2109 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2110 = __THREW__; __THREW__ = 0;
  $2111 = $2110&1;
  if ($2111) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2112 = ((($2109)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2112>>2] = 117; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2113 = ((($2109)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2113>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2109>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2114 = ((($2109)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2114>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2115 = ((($2109)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2115>>2] = $2116; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i954>>2] = $2109; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2117 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2118 = HEAP32[$2117>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2119 = ($2118|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2119) {
   $2122 = $2109;
  } else {
   $2120 = $2118; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2120; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i956 = HEAP32[$$0$i7$i954>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2122 = $$pre$i$i956;
  }
  $2121 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2121,$2122); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2123 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2124 = (($2123) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2124; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4440 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2127 = $$pre4440;$__r$0$i958 = $2109;
 }
 $2125 = ((($__r$0$i958)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2125>>2] = 64; //@line 676 "modules/event/sdl/Event.cpp"
 $2126 = ($2127|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2126) {
   HEAP32[$__parent$i961>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2139 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i969 = $2;$2149 = $2139;
   label = 908;
  } else {
   $__nd$0$i$i963 = $2127;
   while(1) {
    $2128 = ((($__nd$0$i$i963)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2129 = HEAP32[$2128>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2130 = ($2129|0)>(118); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2130) {
     $2131 = HEAP32[$__nd$0$i$i963>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2132 = ($2131|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2132) {
      $$lcssa7344 = $__nd$0$i$i963;$__nd$0$i$i963$lcssa7342 = $__nd$0$i$i963;
      label = 901;
      break;
     } else {
      $$sink$i$i964 = $2131;
     }
    } else {
     $2134 = ($2129|0)<(118); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2134)) {
      $__nd$0$i$i963$lcssa = $__nd$0$i$i963;
      label = 907;
      break;
     }
     $2135 = ((($__nd$0$i$i963)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2136 = HEAP32[$2135>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2137 = ($2136|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2137) {
      $$lcssa7343 = $2135;$__nd$0$i$i963$lcssa7341 = $__nd$0$i$i963;
      label = 905;
      break;
     } else {
      $$sink$i$i964 = $2136;
     }
    }
    $__nd$0$i$i963 = $$sink$i$i964;
   }
   if ((label|0) == 901) {
    HEAP32[$__parent$i961>>2] = $__nd$0$i$i963$lcssa7342; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2133 = $__nd$0$i$i963$lcssa7342; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i969 = $$lcssa7344;$2149 = $2133;
    label = 908;
    break;
   }
   else if ((label|0) == 905) {
    HEAP32[$__parent$i961>>2] = $__nd$0$i$i963$lcssa7341; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2138 = $__nd$0$i$i963$lcssa7341; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i969 = $$lcssa7343;$2149 = $2138;
    label = 908;
    break;
   }
   else if ((label|0) == 907) {
    HEAP32[$__parent$i961>>2] = $__nd$0$i$i963$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2140 = $__nd$0$i$i963$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2141 = ($__nd$0$i$i963$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2141) {
     $$0$i7$i969 = $__parent$i961;$2149 = $2140;
     label = 908;
     break;
    } else {
     $2160 = $2127;$__r$0$i973 = $__nd$0$i$i963$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 908) {
  __THREW__ = 0;
  $2142 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2143 = __THREW__; __THREW__ = 0;
  $2144 = $2143&1;
  if ($2144) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2145 = ((($2142)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2145>>2] = 118; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2146 = ((($2142)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2146>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2142>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2147 = ((($2142)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2147>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2148 = ((($2142)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2148>>2] = $2149; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i969>>2] = $2142; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2150 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2151 = HEAP32[$2150>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2152 = ($2151|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2152) {
   $2155 = $2142;
  } else {
   $2153 = $2151; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2153; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i971 = HEAP32[$$0$i7$i969>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2155 = $$pre$i$i971;
  }
  $2154 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2154,$2155); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2156 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2157 = (($2156) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2157; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4441 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2160 = $$pre4441;$__r$0$i973 = $2142;
 }
 $2158 = ((($__r$0$i973)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2158>>2] = 65; //@line 677 "modules/event/sdl/Event.cpp"
 $2159 = ($2160|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2159) {
   HEAP32[$__parent$i976>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2172 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i984 = $2;$2182 = $2172;
   label = 922;
  } else {
   $__nd$0$i$i978 = $2160;
   while(1) {
    $2161 = ((($__nd$0$i$i978)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2162 = HEAP32[$2161>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2163 = ($2162|0)>(119); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2163) {
     $2164 = HEAP32[$__nd$0$i$i978>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2165 = ($2164|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2165) {
      $$lcssa7340 = $__nd$0$i$i978;$__nd$0$i$i978$lcssa7338 = $__nd$0$i$i978;
      label = 915;
      break;
     } else {
      $$sink$i$i979 = $2164;
     }
    } else {
     $2167 = ($2162|0)<(119); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2167)) {
      $__nd$0$i$i978$lcssa = $__nd$0$i$i978;
      label = 921;
      break;
     }
     $2168 = ((($__nd$0$i$i978)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2169 = HEAP32[$2168>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2170 = ($2169|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2170) {
      $$lcssa7339 = $2168;$__nd$0$i$i978$lcssa7337 = $__nd$0$i$i978;
      label = 919;
      break;
     } else {
      $$sink$i$i979 = $2169;
     }
    }
    $__nd$0$i$i978 = $$sink$i$i979;
   }
   if ((label|0) == 915) {
    HEAP32[$__parent$i976>>2] = $__nd$0$i$i978$lcssa7338; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2166 = $__nd$0$i$i978$lcssa7338; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i984 = $$lcssa7340;$2182 = $2166;
    label = 922;
    break;
   }
   else if ((label|0) == 919) {
    HEAP32[$__parent$i976>>2] = $__nd$0$i$i978$lcssa7337; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2171 = $__nd$0$i$i978$lcssa7337; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i984 = $$lcssa7339;$2182 = $2171;
    label = 922;
    break;
   }
   else if ((label|0) == 921) {
    HEAP32[$__parent$i976>>2] = $__nd$0$i$i978$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2173 = $__nd$0$i$i978$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2174 = ($__nd$0$i$i978$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2174) {
     $$0$i7$i984 = $__parent$i976;$2182 = $2173;
     label = 922;
     break;
    } else {
     $2193 = $2160;$__r$0$i988 = $__nd$0$i$i978$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 922) {
  __THREW__ = 0;
  $2175 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2176 = __THREW__; __THREW__ = 0;
  $2177 = $2176&1;
  if ($2177) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2178 = ((($2175)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2178>>2] = 119; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2179 = ((($2175)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2179>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2175>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2180 = ((($2175)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2180>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2181 = ((($2175)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2181>>2] = $2182; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i984>>2] = $2175; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2183 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2184 = HEAP32[$2183>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2185 = ($2184|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2185) {
   $2188 = $2175;
  } else {
   $2186 = $2184; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2186; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i986 = HEAP32[$$0$i7$i984>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2188 = $$pre$i$i986;
  }
  $2187 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2187,$2188); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2189 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2190 = (($2189) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2190; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4442 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2193 = $$pre4442;$__r$0$i988 = $2175;
 }
 $2191 = ((($__r$0$i988)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2191>>2] = 66; //@line 678 "modules/event/sdl/Event.cpp"
 $2192 = ($2193|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2192) {
   HEAP32[$__parent$i991>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2205 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i999 = $2;$2215 = $2205;
   label = 936;
  } else {
   $__nd$0$i$i993 = $2193;
   while(1) {
    $2194 = ((($__nd$0$i$i993)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2195 = HEAP32[$2194>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2196 = ($2195|0)>(120); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2196) {
     $2197 = HEAP32[$__nd$0$i$i993>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2198 = ($2197|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2198) {
      $$lcssa7336 = $__nd$0$i$i993;$__nd$0$i$i993$lcssa7334 = $__nd$0$i$i993;
      label = 929;
      break;
     } else {
      $$sink$i$i994 = $2197;
     }
    } else {
     $2200 = ($2195|0)<(120); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2200)) {
      $__nd$0$i$i993$lcssa = $__nd$0$i$i993;
      label = 935;
      break;
     }
     $2201 = ((($__nd$0$i$i993)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2202 = HEAP32[$2201>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2203 = ($2202|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2203) {
      $$lcssa7335 = $2201;$__nd$0$i$i993$lcssa7333 = $__nd$0$i$i993;
      label = 933;
      break;
     } else {
      $$sink$i$i994 = $2202;
     }
    }
    $__nd$0$i$i993 = $$sink$i$i994;
   }
   if ((label|0) == 929) {
    HEAP32[$__parent$i991>>2] = $__nd$0$i$i993$lcssa7334; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2199 = $__nd$0$i$i993$lcssa7334; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i999 = $$lcssa7336;$2215 = $2199;
    label = 936;
    break;
   }
   else if ((label|0) == 933) {
    HEAP32[$__parent$i991>>2] = $__nd$0$i$i993$lcssa7333; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2204 = $__nd$0$i$i993$lcssa7333; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i999 = $$lcssa7335;$2215 = $2204;
    label = 936;
    break;
   }
   else if ((label|0) == 935) {
    HEAP32[$__parent$i991>>2] = $__nd$0$i$i993$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2206 = $__nd$0$i$i993$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2207 = ($__nd$0$i$i993$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2207) {
     $$0$i7$i999 = $__parent$i991;$2215 = $2206;
     label = 936;
     break;
    } else {
     $2226 = $2193;$__r$0$i1003 = $__nd$0$i$i993$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 936) {
  __THREW__ = 0;
  $2208 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2209 = __THREW__; __THREW__ = 0;
  $2210 = $2209&1;
  if ($2210) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2211 = ((($2208)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2211>>2] = 120; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2212 = ((($2208)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2212>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2208>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2213 = ((($2208)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2213>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2214 = ((($2208)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2214>>2] = $2215; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i999>>2] = $2208; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2216 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2217 = HEAP32[$2216>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2218 = ($2217|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2218) {
   $2221 = $2208;
  } else {
   $2219 = $2217; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2219; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1001 = HEAP32[$$0$i7$i999>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2221 = $$pre$i$i1001;
  }
  $2220 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2220,$2221); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2222 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2223 = (($2222) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2223; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4443 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2226 = $$pre4443;$__r$0$i1003 = $2208;
 }
 $2224 = ((($__r$0$i1003)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2224>>2] = 67; //@line 679 "modules/event/sdl/Event.cpp"
 $2225 = ($2226|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2225) {
   HEAP32[$__parent$i1006>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2238 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1014 = $2;$2248 = $2238;
   label = 950;
  } else {
   $__nd$0$i$i1008 = $2226;
   while(1) {
    $2227 = ((($__nd$0$i$i1008)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2228 = HEAP32[$2227>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2229 = ($2228|0)>(121); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2229) {
     $2230 = HEAP32[$__nd$0$i$i1008>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2231 = ($2230|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2231) {
      $$lcssa7332 = $__nd$0$i$i1008;$__nd$0$i$i1008$lcssa7330 = $__nd$0$i$i1008;
      label = 943;
      break;
     } else {
      $$sink$i$i1009 = $2230;
     }
    } else {
     $2233 = ($2228|0)<(121); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2233)) {
      $__nd$0$i$i1008$lcssa = $__nd$0$i$i1008;
      label = 949;
      break;
     }
     $2234 = ((($__nd$0$i$i1008)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2235 = HEAP32[$2234>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2236 = ($2235|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2236) {
      $$lcssa7331 = $2234;$__nd$0$i$i1008$lcssa7329 = $__nd$0$i$i1008;
      label = 947;
      break;
     } else {
      $$sink$i$i1009 = $2235;
     }
    }
    $__nd$0$i$i1008 = $$sink$i$i1009;
   }
   if ((label|0) == 943) {
    HEAP32[$__parent$i1006>>2] = $__nd$0$i$i1008$lcssa7330; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2232 = $__nd$0$i$i1008$lcssa7330; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1014 = $$lcssa7332;$2248 = $2232;
    label = 950;
    break;
   }
   else if ((label|0) == 947) {
    HEAP32[$__parent$i1006>>2] = $__nd$0$i$i1008$lcssa7329; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2237 = $__nd$0$i$i1008$lcssa7329; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1014 = $$lcssa7331;$2248 = $2237;
    label = 950;
    break;
   }
   else if ((label|0) == 949) {
    HEAP32[$__parent$i1006>>2] = $__nd$0$i$i1008$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2239 = $__nd$0$i$i1008$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2240 = ($__nd$0$i$i1008$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2240) {
     $$0$i7$i1014 = $__parent$i1006;$2248 = $2239;
     label = 950;
     break;
    } else {
     $2259 = $2226;$__r$0$i1018 = $__nd$0$i$i1008$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 950) {
  __THREW__ = 0;
  $2241 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2242 = __THREW__; __THREW__ = 0;
  $2243 = $2242&1;
  if ($2243) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2244 = ((($2241)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2244>>2] = 121; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2245 = ((($2241)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2245>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2241>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2246 = ((($2241)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2246>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2247 = ((($2241)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2247>>2] = $2248; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1014>>2] = $2241; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2249 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2250 = HEAP32[$2249>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2251 = ($2250|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2251) {
   $2254 = $2241;
  } else {
   $2252 = $2250; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2252; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1016 = HEAP32[$$0$i7$i1014>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2254 = $$pre$i$i1016;
  }
  $2253 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2253,$2254); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2255 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2256 = (($2255) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2256; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4444 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2259 = $$pre4444;$__r$0$i1018 = $2241;
 }
 $2257 = ((($__r$0$i1018)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2257>>2] = 68; //@line 680 "modules/event/sdl/Event.cpp"
 $2258 = ($2259|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2258) {
   HEAP32[$__parent$i1021>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2271 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1029 = $2;$2281 = $2271;
   label = 964;
  } else {
   $__nd$0$i$i1023 = $2259;
   while(1) {
    $2260 = ((($__nd$0$i$i1023)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2261 = HEAP32[$2260>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2262 = ($2261|0)>(122); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2262) {
     $2263 = HEAP32[$__nd$0$i$i1023>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2264 = ($2263|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2264) {
      $$lcssa7328 = $__nd$0$i$i1023;$__nd$0$i$i1023$lcssa7326 = $__nd$0$i$i1023;
      label = 957;
      break;
     } else {
      $$sink$i$i1024 = $2263;
     }
    } else {
     $2266 = ($2261|0)<(122); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2266)) {
      $__nd$0$i$i1023$lcssa = $__nd$0$i$i1023;
      label = 963;
      break;
     }
     $2267 = ((($__nd$0$i$i1023)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2268 = HEAP32[$2267>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2269 = ($2268|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2269) {
      $$lcssa7327 = $2267;$__nd$0$i$i1023$lcssa7325 = $__nd$0$i$i1023;
      label = 961;
      break;
     } else {
      $$sink$i$i1024 = $2268;
     }
    }
    $__nd$0$i$i1023 = $$sink$i$i1024;
   }
   if ((label|0) == 957) {
    HEAP32[$__parent$i1021>>2] = $__nd$0$i$i1023$lcssa7326; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2265 = $__nd$0$i$i1023$lcssa7326; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1029 = $$lcssa7328;$2281 = $2265;
    label = 964;
    break;
   }
   else if ((label|0) == 961) {
    HEAP32[$__parent$i1021>>2] = $__nd$0$i$i1023$lcssa7325; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2270 = $__nd$0$i$i1023$lcssa7325; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1029 = $$lcssa7327;$2281 = $2270;
    label = 964;
    break;
   }
   else if ((label|0) == 963) {
    HEAP32[$__parent$i1021>>2] = $__nd$0$i$i1023$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2272 = $__nd$0$i$i1023$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2273 = ($__nd$0$i$i1023$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2273) {
     $$0$i7$i1029 = $__parent$i1021;$2281 = $2272;
     label = 964;
     break;
    } else {
     $2292 = $2259;$__r$0$i1033 = $__nd$0$i$i1023$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 964) {
  __THREW__ = 0;
  $2274 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2275 = __THREW__; __THREW__ = 0;
  $2276 = $2275&1;
  if ($2276) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2277 = ((($2274)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2277>>2] = 122; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2278 = ((($2274)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2278>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2274>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2279 = ((($2274)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2279>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2280 = ((($2274)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2280>>2] = $2281; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1029>>2] = $2274; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2282 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2283 = HEAP32[$2282>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2284 = ($2283|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2284) {
   $2287 = $2274;
  } else {
   $2285 = $2283; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2285; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1031 = HEAP32[$$0$i7$i1029>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2287 = $$pre$i$i1031;
  }
  $2286 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2286,$2287); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2288 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2289 = (($2288) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2289; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4445 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2292 = $$pre4445;$__r$0$i1033 = $2274;
 }
 $2290 = ((($__r$0$i1033)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2290>>2] = 69; //@line 681 "modules/event/sdl/Event.cpp"
 $2291 = ($2292|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2291) {
   HEAP32[$__parent$i1036>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2304 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1044 = $2;$2314 = $2304;
   label = 978;
  } else {
   $__nd$0$i$i1038 = $2292;
   while(1) {
    $2293 = ((($__nd$0$i$i1038)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2294 = HEAP32[$2293>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2295 = ($2294|0)>(1073741881); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2295) {
     $2296 = HEAP32[$__nd$0$i$i1038>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2297 = ($2296|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2297) {
      $$lcssa7324 = $__nd$0$i$i1038;$__nd$0$i$i1038$lcssa7322 = $__nd$0$i$i1038;
      label = 971;
      break;
     } else {
      $$sink$i$i1039 = $2296;
     }
    } else {
     $2299 = ($2294|0)<(1073741881); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2299)) {
      $__nd$0$i$i1038$lcssa = $__nd$0$i$i1038;
      label = 977;
      break;
     }
     $2300 = ((($__nd$0$i$i1038)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2301 = HEAP32[$2300>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2302 = ($2301|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2302) {
      $$lcssa7323 = $2300;$__nd$0$i$i1038$lcssa7321 = $__nd$0$i$i1038;
      label = 975;
      break;
     } else {
      $$sink$i$i1039 = $2301;
     }
    }
    $__nd$0$i$i1038 = $$sink$i$i1039;
   }
   if ((label|0) == 971) {
    HEAP32[$__parent$i1036>>2] = $__nd$0$i$i1038$lcssa7322; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2298 = $__nd$0$i$i1038$lcssa7322; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1044 = $$lcssa7324;$2314 = $2298;
    label = 978;
    break;
   }
   else if ((label|0) == 975) {
    HEAP32[$__parent$i1036>>2] = $__nd$0$i$i1038$lcssa7321; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2303 = $__nd$0$i$i1038$lcssa7321; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1044 = $$lcssa7323;$2314 = $2303;
    label = 978;
    break;
   }
   else if ((label|0) == 977) {
    HEAP32[$__parent$i1036>>2] = $__nd$0$i$i1038$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2305 = $__nd$0$i$i1038$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2306 = ($__nd$0$i$i1038$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2306) {
     $$0$i7$i1044 = $__parent$i1036;$2314 = $2305;
     label = 978;
     break;
    } else {
     $2325 = $2292;$__r$0$i1048 = $__nd$0$i$i1038$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 978) {
  __THREW__ = 0;
  $2307 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2308 = __THREW__; __THREW__ = 0;
  $2309 = $2308&1;
  if ($2309) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2310 = ((($2307)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2310>>2] = 1073741881; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2311 = ((($2307)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2311>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2307>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2312 = ((($2307)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2312>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2313 = ((($2307)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2313>>2] = $2314; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1044>>2] = $2307; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2315 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2316 = HEAP32[$2315>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2317 = ($2316|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2317) {
   $2320 = $2307;
  } else {
   $2318 = $2316; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2318; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1046 = HEAP32[$$0$i7$i1044>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2320 = $$pre$i$i1046;
  }
  $2319 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2319,$2320); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2321 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2322 = (($2321) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2322; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4446 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2325 = $$pre4446;$__r$0$i1048 = $2307;
 }
 $2323 = ((($__r$0$i1048)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2323>>2] = 70; //@line 683 "modules/event/sdl/Event.cpp"
 $2324 = ($2325|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2324) {
   HEAP32[$__parent$i1051>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2337 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1059 = $2;$2347 = $2337;
   label = 992;
  } else {
   $__nd$0$i$i1053 = $2325;
   while(1) {
    $2326 = ((($__nd$0$i$i1053)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2327 = HEAP32[$2326>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2328 = ($2327|0)>(1073741882); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2328) {
     $2329 = HEAP32[$__nd$0$i$i1053>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2330 = ($2329|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2330) {
      $$lcssa7320 = $__nd$0$i$i1053;$__nd$0$i$i1053$lcssa7318 = $__nd$0$i$i1053;
      label = 985;
      break;
     } else {
      $$sink$i$i1054 = $2329;
     }
    } else {
     $2332 = ($2327|0)<(1073741882); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2332)) {
      $__nd$0$i$i1053$lcssa = $__nd$0$i$i1053;
      label = 991;
      break;
     }
     $2333 = ((($__nd$0$i$i1053)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2334 = HEAP32[$2333>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2335 = ($2334|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2335) {
      $$lcssa7319 = $2333;$__nd$0$i$i1053$lcssa7317 = $__nd$0$i$i1053;
      label = 989;
      break;
     } else {
      $$sink$i$i1054 = $2334;
     }
    }
    $__nd$0$i$i1053 = $$sink$i$i1054;
   }
   if ((label|0) == 985) {
    HEAP32[$__parent$i1051>>2] = $__nd$0$i$i1053$lcssa7318; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2331 = $__nd$0$i$i1053$lcssa7318; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1059 = $$lcssa7320;$2347 = $2331;
    label = 992;
    break;
   }
   else if ((label|0) == 989) {
    HEAP32[$__parent$i1051>>2] = $__nd$0$i$i1053$lcssa7317; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2336 = $__nd$0$i$i1053$lcssa7317; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1059 = $$lcssa7319;$2347 = $2336;
    label = 992;
    break;
   }
   else if ((label|0) == 991) {
    HEAP32[$__parent$i1051>>2] = $__nd$0$i$i1053$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2338 = $__nd$0$i$i1053$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2339 = ($__nd$0$i$i1053$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2339) {
     $$0$i7$i1059 = $__parent$i1051;$2347 = $2338;
     label = 992;
     break;
    } else {
     $2358 = $2325;$__r$0$i1063 = $__nd$0$i$i1053$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 992) {
  __THREW__ = 0;
  $2340 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2341 = __THREW__; __THREW__ = 0;
  $2342 = $2341&1;
  if ($2342) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2343 = ((($2340)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2343>>2] = 1073741882; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2344 = ((($2340)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2344>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2340>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2345 = ((($2340)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2345>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2346 = ((($2340)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2346>>2] = $2347; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1059>>2] = $2340; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2348 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2349 = HEAP32[$2348>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2350 = ($2349|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2350) {
   $2353 = $2340;
  } else {
   $2351 = $2349; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2351; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1061 = HEAP32[$$0$i7$i1059>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2353 = $$pre$i$i1061;
  }
  $2352 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2352,$2353); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2354 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2355 = (($2354) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2355; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4447 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2358 = $$pre4447;$__r$0$i1063 = $2340;
 }
 $2356 = ((($__r$0$i1063)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2356>>2] = 71; //@line 685 "modules/event/sdl/Event.cpp"
 $2357 = ($2358|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2357) {
   HEAP32[$__parent$i1066>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2370 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1074 = $2;$2380 = $2370;
   label = 1006;
  } else {
   $__nd$0$i$i1068 = $2358;
   while(1) {
    $2359 = ((($__nd$0$i$i1068)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2360 = HEAP32[$2359>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2361 = ($2360|0)>(1073741883); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2361) {
     $2362 = HEAP32[$__nd$0$i$i1068>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2363 = ($2362|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2363) {
      $$lcssa7316 = $__nd$0$i$i1068;$__nd$0$i$i1068$lcssa7314 = $__nd$0$i$i1068;
      label = 999;
      break;
     } else {
      $$sink$i$i1069 = $2362;
     }
    } else {
     $2365 = ($2360|0)<(1073741883); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2365)) {
      $__nd$0$i$i1068$lcssa = $__nd$0$i$i1068;
      label = 1005;
      break;
     }
     $2366 = ((($__nd$0$i$i1068)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2367 = HEAP32[$2366>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2368 = ($2367|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2368) {
      $$lcssa7315 = $2366;$__nd$0$i$i1068$lcssa7313 = $__nd$0$i$i1068;
      label = 1003;
      break;
     } else {
      $$sink$i$i1069 = $2367;
     }
    }
    $__nd$0$i$i1068 = $$sink$i$i1069;
   }
   if ((label|0) == 999) {
    HEAP32[$__parent$i1066>>2] = $__nd$0$i$i1068$lcssa7314; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2364 = $__nd$0$i$i1068$lcssa7314; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1074 = $$lcssa7316;$2380 = $2364;
    label = 1006;
    break;
   }
   else if ((label|0) == 1003) {
    HEAP32[$__parent$i1066>>2] = $__nd$0$i$i1068$lcssa7313; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2369 = $__nd$0$i$i1068$lcssa7313; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1074 = $$lcssa7315;$2380 = $2369;
    label = 1006;
    break;
   }
   else if ((label|0) == 1005) {
    HEAP32[$__parent$i1066>>2] = $__nd$0$i$i1068$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2371 = $__nd$0$i$i1068$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2372 = ($__nd$0$i$i1068$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2372) {
     $$0$i7$i1074 = $__parent$i1066;$2380 = $2371;
     label = 1006;
     break;
    } else {
     $2391 = $2358;$__r$0$i1078 = $__nd$0$i$i1068$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1006) {
  __THREW__ = 0;
  $2373 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2374 = __THREW__; __THREW__ = 0;
  $2375 = $2374&1;
  if ($2375) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2376 = ((($2373)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2376>>2] = 1073741883; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2377 = ((($2373)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2377>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2373>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2378 = ((($2373)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2378>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2379 = ((($2373)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2379>>2] = $2380; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1074>>2] = $2373; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2381 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2382 = HEAP32[$2381>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2383 = ($2382|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2383) {
   $2386 = $2373;
  } else {
   $2384 = $2382; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2384; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1076 = HEAP32[$$0$i7$i1074>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2386 = $$pre$i$i1076;
  }
  $2385 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2385,$2386); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2387 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2388 = (($2387) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2388; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4448 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2391 = $$pre4448;$__r$0$i1078 = $2373;
 }
 $2389 = ((($__r$0$i1078)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2389>>2] = 72; //@line 686 "modules/event/sdl/Event.cpp"
 $2390 = ($2391|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2390) {
   HEAP32[$__parent$i1081>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2403 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1089 = $2;$2413 = $2403;
   label = 1020;
  } else {
   $__nd$0$i$i1083 = $2391;
   while(1) {
    $2392 = ((($__nd$0$i$i1083)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2393 = HEAP32[$2392>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2394 = ($2393|0)>(1073741884); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2394) {
     $2395 = HEAP32[$__nd$0$i$i1083>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2396 = ($2395|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2396) {
      $$lcssa7312 = $__nd$0$i$i1083;$__nd$0$i$i1083$lcssa7310 = $__nd$0$i$i1083;
      label = 1013;
      break;
     } else {
      $$sink$i$i1084 = $2395;
     }
    } else {
     $2398 = ($2393|0)<(1073741884); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2398)) {
      $__nd$0$i$i1083$lcssa = $__nd$0$i$i1083;
      label = 1019;
      break;
     }
     $2399 = ((($__nd$0$i$i1083)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2400 = HEAP32[$2399>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2401 = ($2400|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2401) {
      $$lcssa7311 = $2399;$__nd$0$i$i1083$lcssa7309 = $__nd$0$i$i1083;
      label = 1017;
      break;
     } else {
      $$sink$i$i1084 = $2400;
     }
    }
    $__nd$0$i$i1083 = $$sink$i$i1084;
   }
   if ((label|0) == 1013) {
    HEAP32[$__parent$i1081>>2] = $__nd$0$i$i1083$lcssa7310; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2397 = $__nd$0$i$i1083$lcssa7310; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1089 = $$lcssa7312;$2413 = $2397;
    label = 1020;
    break;
   }
   else if ((label|0) == 1017) {
    HEAP32[$__parent$i1081>>2] = $__nd$0$i$i1083$lcssa7309; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2402 = $__nd$0$i$i1083$lcssa7309; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1089 = $$lcssa7311;$2413 = $2402;
    label = 1020;
    break;
   }
   else if ((label|0) == 1019) {
    HEAP32[$__parent$i1081>>2] = $__nd$0$i$i1083$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2404 = $__nd$0$i$i1083$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2405 = ($__nd$0$i$i1083$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2405) {
     $$0$i7$i1089 = $__parent$i1081;$2413 = $2404;
     label = 1020;
     break;
    } else {
     $2424 = $2391;$__r$0$i1093 = $__nd$0$i$i1083$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1020) {
  __THREW__ = 0;
  $2406 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2407 = __THREW__; __THREW__ = 0;
  $2408 = $2407&1;
  if ($2408) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2409 = ((($2406)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2409>>2] = 1073741884; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2410 = ((($2406)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2410>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2406>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2411 = ((($2406)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2411>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2412 = ((($2406)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2412>>2] = $2413; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1089>>2] = $2406; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2414 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2415 = HEAP32[$2414>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2416 = ($2415|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2416) {
   $2419 = $2406;
  } else {
   $2417 = $2415; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2417; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1091 = HEAP32[$$0$i7$i1089>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2419 = $$pre$i$i1091;
  }
  $2418 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2418,$2419); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2420 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2421 = (($2420) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2421; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4449 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2424 = $$pre4449;$__r$0$i1093 = $2406;
 }
 $2422 = ((($__r$0$i1093)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2422>>2] = 73; //@line 687 "modules/event/sdl/Event.cpp"
 $2423 = ($2424|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2423) {
   HEAP32[$__parent$i1096>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2436 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1104 = $2;$2446 = $2436;
   label = 1034;
  } else {
   $__nd$0$i$i1098 = $2424;
   while(1) {
    $2425 = ((($__nd$0$i$i1098)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2426 = HEAP32[$2425>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2427 = ($2426|0)>(1073741885); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2427) {
     $2428 = HEAP32[$__nd$0$i$i1098>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2429 = ($2428|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2429) {
      $$lcssa7308 = $__nd$0$i$i1098;$__nd$0$i$i1098$lcssa7306 = $__nd$0$i$i1098;
      label = 1027;
      break;
     } else {
      $$sink$i$i1099 = $2428;
     }
    } else {
     $2431 = ($2426|0)<(1073741885); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2431)) {
      $__nd$0$i$i1098$lcssa = $__nd$0$i$i1098;
      label = 1033;
      break;
     }
     $2432 = ((($__nd$0$i$i1098)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2433 = HEAP32[$2432>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2434 = ($2433|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2434) {
      $$lcssa7307 = $2432;$__nd$0$i$i1098$lcssa7305 = $__nd$0$i$i1098;
      label = 1031;
      break;
     } else {
      $$sink$i$i1099 = $2433;
     }
    }
    $__nd$0$i$i1098 = $$sink$i$i1099;
   }
   if ((label|0) == 1027) {
    HEAP32[$__parent$i1096>>2] = $__nd$0$i$i1098$lcssa7306; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2430 = $__nd$0$i$i1098$lcssa7306; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1104 = $$lcssa7308;$2446 = $2430;
    label = 1034;
    break;
   }
   else if ((label|0) == 1031) {
    HEAP32[$__parent$i1096>>2] = $__nd$0$i$i1098$lcssa7305; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2435 = $__nd$0$i$i1098$lcssa7305; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1104 = $$lcssa7307;$2446 = $2435;
    label = 1034;
    break;
   }
   else if ((label|0) == 1033) {
    HEAP32[$__parent$i1096>>2] = $__nd$0$i$i1098$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2437 = $__nd$0$i$i1098$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2438 = ($__nd$0$i$i1098$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2438) {
     $$0$i7$i1104 = $__parent$i1096;$2446 = $2437;
     label = 1034;
     break;
    } else {
     $2457 = $2424;$__r$0$i1108 = $__nd$0$i$i1098$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1034) {
  __THREW__ = 0;
  $2439 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2440 = __THREW__; __THREW__ = 0;
  $2441 = $2440&1;
  if ($2441) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2442 = ((($2439)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2442>>2] = 1073741885; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2443 = ((($2439)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2443>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2439>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2444 = ((($2439)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2444>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2445 = ((($2439)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2445>>2] = $2446; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1104>>2] = $2439; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2447 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2448 = HEAP32[$2447>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2449 = ($2448|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2449) {
   $2452 = $2439;
  } else {
   $2450 = $2448; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2450; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1106 = HEAP32[$$0$i7$i1104>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2452 = $$pre$i$i1106;
  }
  $2451 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2451,$2452); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2453 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2454 = (($2453) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2454; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4450 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2457 = $$pre4450;$__r$0$i1108 = $2439;
 }
 $2455 = ((($__r$0$i1108)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2455>>2] = 74; //@line 688 "modules/event/sdl/Event.cpp"
 $2456 = ($2457|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2456) {
   HEAP32[$__parent$i1111>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2469 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1119 = $2;$2479 = $2469;
   label = 1048;
  } else {
   $__nd$0$i$i1113 = $2457;
   while(1) {
    $2458 = ((($__nd$0$i$i1113)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2459 = HEAP32[$2458>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2460 = ($2459|0)>(1073741886); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2460) {
     $2461 = HEAP32[$__nd$0$i$i1113>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2462 = ($2461|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2462) {
      $$lcssa7304 = $__nd$0$i$i1113;$__nd$0$i$i1113$lcssa7302 = $__nd$0$i$i1113;
      label = 1041;
      break;
     } else {
      $$sink$i$i1114 = $2461;
     }
    } else {
     $2464 = ($2459|0)<(1073741886); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2464)) {
      $__nd$0$i$i1113$lcssa = $__nd$0$i$i1113;
      label = 1047;
      break;
     }
     $2465 = ((($__nd$0$i$i1113)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2466 = HEAP32[$2465>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2467 = ($2466|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2467) {
      $$lcssa7303 = $2465;$__nd$0$i$i1113$lcssa7301 = $__nd$0$i$i1113;
      label = 1045;
      break;
     } else {
      $$sink$i$i1114 = $2466;
     }
    }
    $__nd$0$i$i1113 = $$sink$i$i1114;
   }
   if ((label|0) == 1041) {
    HEAP32[$__parent$i1111>>2] = $__nd$0$i$i1113$lcssa7302; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2463 = $__nd$0$i$i1113$lcssa7302; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1119 = $$lcssa7304;$2479 = $2463;
    label = 1048;
    break;
   }
   else if ((label|0) == 1045) {
    HEAP32[$__parent$i1111>>2] = $__nd$0$i$i1113$lcssa7301; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2468 = $__nd$0$i$i1113$lcssa7301; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1119 = $$lcssa7303;$2479 = $2468;
    label = 1048;
    break;
   }
   else if ((label|0) == 1047) {
    HEAP32[$__parent$i1111>>2] = $__nd$0$i$i1113$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2470 = $__nd$0$i$i1113$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2471 = ($__nd$0$i$i1113$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2471) {
     $$0$i7$i1119 = $__parent$i1111;$2479 = $2470;
     label = 1048;
     break;
    } else {
     $2490 = $2457;$__r$0$i1123 = $__nd$0$i$i1113$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1048) {
  __THREW__ = 0;
  $2472 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2473 = __THREW__; __THREW__ = 0;
  $2474 = $2473&1;
  if ($2474) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2475 = ((($2472)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2475>>2] = 1073741886; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2476 = ((($2472)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2476>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2472>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2477 = ((($2472)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2477>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2478 = ((($2472)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2478>>2] = $2479; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1119>>2] = $2472; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2480 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2481 = HEAP32[$2480>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2482 = ($2481|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2482) {
   $2485 = $2472;
  } else {
   $2483 = $2481; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2483; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1121 = HEAP32[$$0$i7$i1119>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2485 = $$pre$i$i1121;
  }
  $2484 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2484,$2485); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2486 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2487 = (($2486) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2487; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4451 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2490 = $$pre4451;$__r$0$i1123 = $2472;
 }
 $2488 = ((($__r$0$i1123)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2488>>2] = 75; //@line 689 "modules/event/sdl/Event.cpp"
 $2489 = ($2490|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2489) {
   HEAP32[$__parent$i1126>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2502 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1134 = $2;$2512 = $2502;
   label = 1062;
  } else {
   $__nd$0$i$i1128 = $2490;
   while(1) {
    $2491 = ((($__nd$0$i$i1128)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2492 = HEAP32[$2491>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2493 = ($2492|0)>(1073741887); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2493) {
     $2494 = HEAP32[$__nd$0$i$i1128>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2495 = ($2494|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2495) {
      $$lcssa7300 = $__nd$0$i$i1128;$__nd$0$i$i1128$lcssa7298 = $__nd$0$i$i1128;
      label = 1055;
      break;
     } else {
      $$sink$i$i1129 = $2494;
     }
    } else {
     $2497 = ($2492|0)<(1073741887); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2497)) {
      $__nd$0$i$i1128$lcssa = $__nd$0$i$i1128;
      label = 1061;
      break;
     }
     $2498 = ((($__nd$0$i$i1128)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2499 = HEAP32[$2498>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2500 = ($2499|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2500) {
      $$lcssa7299 = $2498;$__nd$0$i$i1128$lcssa7297 = $__nd$0$i$i1128;
      label = 1059;
      break;
     } else {
      $$sink$i$i1129 = $2499;
     }
    }
    $__nd$0$i$i1128 = $$sink$i$i1129;
   }
   if ((label|0) == 1055) {
    HEAP32[$__parent$i1126>>2] = $__nd$0$i$i1128$lcssa7298; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2496 = $__nd$0$i$i1128$lcssa7298; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1134 = $$lcssa7300;$2512 = $2496;
    label = 1062;
    break;
   }
   else if ((label|0) == 1059) {
    HEAP32[$__parent$i1126>>2] = $__nd$0$i$i1128$lcssa7297; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2501 = $__nd$0$i$i1128$lcssa7297; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1134 = $$lcssa7299;$2512 = $2501;
    label = 1062;
    break;
   }
   else if ((label|0) == 1061) {
    HEAP32[$__parent$i1126>>2] = $__nd$0$i$i1128$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2503 = $__nd$0$i$i1128$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2504 = ($__nd$0$i$i1128$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2504) {
     $$0$i7$i1134 = $__parent$i1126;$2512 = $2503;
     label = 1062;
     break;
    } else {
     $2523 = $2490;$__r$0$i1138 = $__nd$0$i$i1128$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1062) {
  __THREW__ = 0;
  $2505 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2506 = __THREW__; __THREW__ = 0;
  $2507 = $2506&1;
  if ($2507) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2508 = ((($2505)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2508>>2] = 1073741887; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2509 = ((($2505)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2509>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2505>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2510 = ((($2505)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2510>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2511 = ((($2505)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2511>>2] = $2512; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1134>>2] = $2505; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2513 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2514 = HEAP32[$2513>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2515 = ($2514|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2515) {
   $2518 = $2505;
  } else {
   $2516 = $2514; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2516; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1136 = HEAP32[$$0$i7$i1134>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2518 = $$pre$i$i1136;
  }
  $2517 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2517,$2518); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2519 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2520 = (($2519) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2520; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4452 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2523 = $$pre4452;$__r$0$i1138 = $2505;
 }
 $2521 = ((($__r$0$i1138)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2521>>2] = 76; //@line 690 "modules/event/sdl/Event.cpp"
 $2522 = ($2523|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2522) {
   HEAP32[$__parent$i1141>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2535 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1149 = $2;$2545 = $2535;
   label = 1076;
  } else {
   $__nd$0$i$i1143 = $2523;
   while(1) {
    $2524 = ((($__nd$0$i$i1143)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2525 = HEAP32[$2524>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2526 = ($2525|0)>(1073741888); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2526) {
     $2527 = HEAP32[$__nd$0$i$i1143>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2528 = ($2527|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2528) {
      $$lcssa7296 = $__nd$0$i$i1143;$__nd$0$i$i1143$lcssa7294 = $__nd$0$i$i1143;
      label = 1069;
      break;
     } else {
      $$sink$i$i1144 = $2527;
     }
    } else {
     $2530 = ($2525|0)<(1073741888); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2530)) {
      $__nd$0$i$i1143$lcssa = $__nd$0$i$i1143;
      label = 1075;
      break;
     }
     $2531 = ((($__nd$0$i$i1143)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2532 = HEAP32[$2531>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2533 = ($2532|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2533) {
      $$lcssa7295 = $2531;$__nd$0$i$i1143$lcssa7293 = $__nd$0$i$i1143;
      label = 1073;
      break;
     } else {
      $$sink$i$i1144 = $2532;
     }
    }
    $__nd$0$i$i1143 = $$sink$i$i1144;
   }
   if ((label|0) == 1069) {
    HEAP32[$__parent$i1141>>2] = $__nd$0$i$i1143$lcssa7294; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2529 = $__nd$0$i$i1143$lcssa7294; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1149 = $$lcssa7296;$2545 = $2529;
    label = 1076;
    break;
   }
   else if ((label|0) == 1073) {
    HEAP32[$__parent$i1141>>2] = $__nd$0$i$i1143$lcssa7293; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2534 = $__nd$0$i$i1143$lcssa7293; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1149 = $$lcssa7295;$2545 = $2534;
    label = 1076;
    break;
   }
   else if ((label|0) == 1075) {
    HEAP32[$__parent$i1141>>2] = $__nd$0$i$i1143$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2536 = $__nd$0$i$i1143$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2537 = ($__nd$0$i$i1143$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2537) {
     $$0$i7$i1149 = $__parent$i1141;$2545 = $2536;
     label = 1076;
     break;
    } else {
     $2556 = $2523;$__r$0$i1153 = $__nd$0$i$i1143$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1076) {
  __THREW__ = 0;
  $2538 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2539 = __THREW__; __THREW__ = 0;
  $2540 = $2539&1;
  if ($2540) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2541 = ((($2538)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2541>>2] = 1073741888; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2542 = ((($2538)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2542>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2538>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2543 = ((($2538)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2543>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2544 = ((($2538)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2544>>2] = $2545; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1149>>2] = $2538; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2546 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2547 = HEAP32[$2546>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2548 = ($2547|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2548) {
   $2551 = $2538;
  } else {
   $2549 = $2547; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2549; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1151 = HEAP32[$$0$i7$i1149>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2551 = $$pre$i$i1151;
  }
  $2550 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2550,$2551); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2552 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2553 = (($2552) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2553; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4453 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2556 = $$pre4453;$__r$0$i1153 = $2538;
 }
 $2554 = ((($__r$0$i1153)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2554>>2] = 77; //@line 691 "modules/event/sdl/Event.cpp"
 $2555 = ($2556|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2555) {
   HEAP32[$__parent$i1156>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2568 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1164 = $2;$2578 = $2568;
   label = 1090;
  } else {
   $__nd$0$i$i1158 = $2556;
   while(1) {
    $2557 = ((($__nd$0$i$i1158)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2558 = HEAP32[$2557>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2559 = ($2558|0)>(1073741889); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2559) {
     $2560 = HEAP32[$__nd$0$i$i1158>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2561 = ($2560|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2561) {
      $$lcssa7292 = $__nd$0$i$i1158;$__nd$0$i$i1158$lcssa7290 = $__nd$0$i$i1158;
      label = 1083;
      break;
     } else {
      $$sink$i$i1159 = $2560;
     }
    } else {
     $2563 = ($2558|0)<(1073741889); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2563)) {
      $__nd$0$i$i1158$lcssa = $__nd$0$i$i1158;
      label = 1089;
      break;
     }
     $2564 = ((($__nd$0$i$i1158)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2565 = HEAP32[$2564>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2566 = ($2565|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2566) {
      $$lcssa7291 = $2564;$__nd$0$i$i1158$lcssa7289 = $__nd$0$i$i1158;
      label = 1087;
      break;
     } else {
      $$sink$i$i1159 = $2565;
     }
    }
    $__nd$0$i$i1158 = $$sink$i$i1159;
   }
   if ((label|0) == 1083) {
    HEAP32[$__parent$i1156>>2] = $__nd$0$i$i1158$lcssa7290; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2562 = $__nd$0$i$i1158$lcssa7290; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1164 = $$lcssa7292;$2578 = $2562;
    label = 1090;
    break;
   }
   else if ((label|0) == 1087) {
    HEAP32[$__parent$i1156>>2] = $__nd$0$i$i1158$lcssa7289; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2567 = $__nd$0$i$i1158$lcssa7289; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1164 = $$lcssa7291;$2578 = $2567;
    label = 1090;
    break;
   }
   else if ((label|0) == 1089) {
    HEAP32[$__parent$i1156>>2] = $__nd$0$i$i1158$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2569 = $__nd$0$i$i1158$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2570 = ($__nd$0$i$i1158$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2570) {
     $$0$i7$i1164 = $__parent$i1156;$2578 = $2569;
     label = 1090;
     break;
    } else {
     $2589 = $2556;$__r$0$i1168 = $__nd$0$i$i1158$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1090) {
  __THREW__ = 0;
  $2571 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2572 = __THREW__; __THREW__ = 0;
  $2573 = $2572&1;
  if ($2573) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2574 = ((($2571)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2574>>2] = 1073741889; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2575 = ((($2571)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2575>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2571>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2576 = ((($2571)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2576>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2577 = ((($2571)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2577>>2] = $2578; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1164>>2] = $2571; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2579 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2580 = HEAP32[$2579>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2581 = ($2580|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2581) {
   $2584 = $2571;
  } else {
   $2582 = $2580; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2582; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1166 = HEAP32[$$0$i7$i1164>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2584 = $$pre$i$i1166;
  }
  $2583 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2583,$2584); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2585 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2586 = (($2585) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2586; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4454 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2589 = $$pre4454;$__r$0$i1168 = $2571;
 }
 $2587 = ((($__r$0$i1168)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2587>>2] = 78; //@line 692 "modules/event/sdl/Event.cpp"
 $2588 = ($2589|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2588) {
   HEAP32[$__parent$i1171>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2601 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1179 = $2;$2611 = $2601;
   label = 1104;
  } else {
   $__nd$0$i$i1173 = $2589;
   while(1) {
    $2590 = ((($__nd$0$i$i1173)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2591 = HEAP32[$2590>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2592 = ($2591|0)>(1073741890); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2592) {
     $2593 = HEAP32[$__nd$0$i$i1173>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2594 = ($2593|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2594) {
      $$lcssa7288 = $__nd$0$i$i1173;$__nd$0$i$i1173$lcssa7286 = $__nd$0$i$i1173;
      label = 1097;
      break;
     } else {
      $$sink$i$i1174 = $2593;
     }
    } else {
     $2596 = ($2591|0)<(1073741890); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2596)) {
      $__nd$0$i$i1173$lcssa = $__nd$0$i$i1173;
      label = 1103;
      break;
     }
     $2597 = ((($__nd$0$i$i1173)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2598 = HEAP32[$2597>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2599 = ($2598|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2599) {
      $$lcssa7287 = $2597;$__nd$0$i$i1173$lcssa7285 = $__nd$0$i$i1173;
      label = 1101;
      break;
     } else {
      $$sink$i$i1174 = $2598;
     }
    }
    $__nd$0$i$i1173 = $$sink$i$i1174;
   }
   if ((label|0) == 1097) {
    HEAP32[$__parent$i1171>>2] = $__nd$0$i$i1173$lcssa7286; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2595 = $__nd$0$i$i1173$lcssa7286; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1179 = $$lcssa7288;$2611 = $2595;
    label = 1104;
    break;
   }
   else if ((label|0) == 1101) {
    HEAP32[$__parent$i1171>>2] = $__nd$0$i$i1173$lcssa7285; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2600 = $__nd$0$i$i1173$lcssa7285; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1179 = $$lcssa7287;$2611 = $2600;
    label = 1104;
    break;
   }
   else if ((label|0) == 1103) {
    HEAP32[$__parent$i1171>>2] = $__nd$0$i$i1173$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2602 = $__nd$0$i$i1173$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2603 = ($__nd$0$i$i1173$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2603) {
     $$0$i7$i1179 = $__parent$i1171;$2611 = $2602;
     label = 1104;
     break;
    } else {
     $2622 = $2589;$__r$0$i1183 = $__nd$0$i$i1173$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1104) {
  __THREW__ = 0;
  $2604 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2605 = __THREW__; __THREW__ = 0;
  $2606 = $2605&1;
  if ($2606) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2607 = ((($2604)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2607>>2] = 1073741890; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2608 = ((($2604)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2608>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2604>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2609 = ((($2604)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2609>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2610 = ((($2604)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2610>>2] = $2611; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1179>>2] = $2604; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2612 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2613 = HEAP32[$2612>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2614 = ($2613|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2614) {
   $2617 = $2604;
  } else {
   $2615 = $2613; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2615; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1181 = HEAP32[$$0$i7$i1179>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2617 = $$pre$i$i1181;
  }
  $2616 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2616,$2617); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2618 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2619 = (($2618) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2619; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4455 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2622 = $$pre4455;$__r$0$i1183 = $2604;
 }
 $2620 = ((($__r$0$i1183)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2620>>2] = 79; //@line 693 "modules/event/sdl/Event.cpp"
 $2621 = ($2622|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2621) {
   HEAP32[$__parent$i1186>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2634 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1194 = $2;$2644 = $2634;
   label = 1118;
  } else {
   $__nd$0$i$i1188 = $2622;
   while(1) {
    $2623 = ((($__nd$0$i$i1188)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2624 = HEAP32[$2623>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2625 = ($2624|0)>(1073741891); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2625) {
     $2626 = HEAP32[$__nd$0$i$i1188>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2627 = ($2626|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2627) {
      $$lcssa7284 = $__nd$0$i$i1188;$__nd$0$i$i1188$lcssa7282 = $__nd$0$i$i1188;
      label = 1111;
      break;
     } else {
      $$sink$i$i1189 = $2626;
     }
    } else {
     $2629 = ($2624|0)<(1073741891); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2629)) {
      $__nd$0$i$i1188$lcssa = $__nd$0$i$i1188;
      label = 1117;
      break;
     }
     $2630 = ((($__nd$0$i$i1188)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2631 = HEAP32[$2630>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2632 = ($2631|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2632) {
      $$lcssa7283 = $2630;$__nd$0$i$i1188$lcssa7281 = $__nd$0$i$i1188;
      label = 1115;
      break;
     } else {
      $$sink$i$i1189 = $2631;
     }
    }
    $__nd$0$i$i1188 = $$sink$i$i1189;
   }
   if ((label|0) == 1111) {
    HEAP32[$__parent$i1186>>2] = $__nd$0$i$i1188$lcssa7282; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2628 = $__nd$0$i$i1188$lcssa7282; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1194 = $$lcssa7284;$2644 = $2628;
    label = 1118;
    break;
   }
   else if ((label|0) == 1115) {
    HEAP32[$__parent$i1186>>2] = $__nd$0$i$i1188$lcssa7281; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2633 = $__nd$0$i$i1188$lcssa7281; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1194 = $$lcssa7283;$2644 = $2633;
    label = 1118;
    break;
   }
   else if ((label|0) == 1117) {
    HEAP32[$__parent$i1186>>2] = $__nd$0$i$i1188$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2635 = $__nd$0$i$i1188$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2636 = ($__nd$0$i$i1188$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2636) {
     $$0$i7$i1194 = $__parent$i1186;$2644 = $2635;
     label = 1118;
     break;
    } else {
     $2655 = $2622;$__r$0$i1198 = $__nd$0$i$i1188$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1118) {
  __THREW__ = 0;
  $2637 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2638 = __THREW__; __THREW__ = 0;
  $2639 = $2638&1;
  if ($2639) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2640 = ((($2637)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2640>>2] = 1073741891; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2641 = ((($2637)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2641>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2637>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2642 = ((($2637)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2642>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2643 = ((($2637)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2643>>2] = $2644; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1194>>2] = $2637; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2645 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2646 = HEAP32[$2645>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2647 = ($2646|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2647) {
   $2650 = $2637;
  } else {
   $2648 = $2646; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2648; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1196 = HEAP32[$$0$i7$i1194>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2650 = $$pre$i$i1196;
  }
  $2649 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2649,$2650); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2651 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2652 = (($2651) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2652; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4456 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2655 = $$pre4456;$__r$0$i1198 = $2637;
 }
 $2653 = ((($__r$0$i1198)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2653>>2] = 80; //@line 694 "modules/event/sdl/Event.cpp"
 $2654 = ($2655|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2654) {
   HEAP32[$__parent$i1201>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2667 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1209 = $2;$2677 = $2667;
   label = 1132;
  } else {
   $__nd$0$i$i1203 = $2655;
   while(1) {
    $2656 = ((($__nd$0$i$i1203)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2657 = HEAP32[$2656>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2658 = ($2657|0)>(1073741892); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2658) {
     $2659 = HEAP32[$__nd$0$i$i1203>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2660 = ($2659|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2660) {
      $$lcssa7280 = $__nd$0$i$i1203;$__nd$0$i$i1203$lcssa7278 = $__nd$0$i$i1203;
      label = 1125;
      break;
     } else {
      $$sink$i$i1204 = $2659;
     }
    } else {
     $2662 = ($2657|0)<(1073741892); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2662)) {
      $__nd$0$i$i1203$lcssa = $__nd$0$i$i1203;
      label = 1131;
      break;
     }
     $2663 = ((($__nd$0$i$i1203)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2664 = HEAP32[$2663>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2665 = ($2664|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2665) {
      $$lcssa7279 = $2663;$__nd$0$i$i1203$lcssa7277 = $__nd$0$i$i1203;
      label = 1129;
      break;
     } else {
      $$sink$i$i1204 = $2664;
     }
    }
    $__nd$0$i$i1203 = $$sink$i$i1204;
   }
   if ((label|0) == 1125) {
    HEAP32[$__parent$i1201>>2] = $__nd$0$i$i1203$lcssa7278; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2661 = $__nd$0$i$i1203$lcssa7278; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1209 = $$lcssa7280;$2677 = $2661;
    label = 1132;
    break;
   }
   else if ((label|0) == 1129) {
    HEAP32[$__parent$i1201>>2] = $__nd$0$i$i1203$lcssa7277; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2666 = $__nd$0$i$i1203$lcssa7277; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1209 = $$lcssa7279;$2677 = $2666;
    label = 1132;
    break;
   }
   else if ((label|0) == 1131) {
    HEAP32[$__parent$i1201>>2] = $__nd$0$i$i1203$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2668 = $__nd$0$i$i1203$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2669 = ($__nd$0$i$i1203$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2669) {
     $$0$i7$i1209 = $__parent$i1201;$2677 = $2668;
     label = 1132;
     break;
    } else {
     $2688 = $2655;$__r$0$i1213 = $__nd$0$i$i1203$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1132) {
  __THREW__ = 0;
  $2670 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2671 = __THREW__; __THREW__ = 0;
  $2672 = $2671&1;
  if ($2672) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2673 = ((($2670)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2673>>2] = 1073741892; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2674 = ((($2670)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2674>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2670>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2675 = ((($2670)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2675>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2676 = ((($2670)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2676>>2] = $2677; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1209>>2] = $2670; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2678 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2679 = HEAP32[$2678>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2680 = ($2679|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2680) {
   $2683 = $2670;
  } else {
   $2681 = $2679; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2681; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1211 = HEAP32[$$0$i7$i1209>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2683 = $$pre$i$i1211;
  }
  $2682 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2682,$2683); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2684 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2685 = (($2684) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2685; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4457 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2688 = $$pre4457;$__r$0$i1213 = $2670;
 }
 $2686 = ((($__r$0$i1213)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2686>>2] = 81; //@line 695 "modules/event/sdl/Event.cpp"
 $2687 = ($2688|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2687) {
   HEAP32[$__parent$i1216>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2700 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1224 = $2;$2710 = $2700;
   label = 1146;
  } else {
   $__nd$0$i$i1218 = $2688;
   while(1) {
    $2689 = ((($__nd$0$i$i1218)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2690 = HEAP32[$2689>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2691 = ($2690|0)>(1073741893); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2691) {
     $2692 = HEAP32[$__nd$0$i$i1218>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2693 = ($2692|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2693) {
      $$lcssa7276 = $__nd$0$i$i1218;$__nd$0$i$i1218$lcssa7274 = $__nd$0$i$i1218;
      label = 1139;
      break;
     } else {
      $$sink$i$i1219 = $2692;
     }
    } else {
     $2695 = ($2690|0)<(1073741893); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2695)) {
      $__nd$0$i$i1218$lcssa = $__nd$0$i$i1218;
      label = 1145;
      break;
     }
     $2696 = ((($__nd$0$i$i1218)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2697 = HEAP32[$2696>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2698 = ($2697|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2698) {
      $$lcssa7275 = $2696;$__nd$0$i$i1218$lcssa7273 = $__nd$0$i$i1218;
      label = 1143;
      break;
     } else {
      $$sink$i$i1219 = $2697;
     }
    }
    $__nd$0$i$i1218 = $$sink$i$i1219;
   }
   if ((label|0) == 1139) {
    HEAP32[$__parent$i1216>>2] = $__nd$0$i$i1218$lcssa7274; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2694 = $__nd$0$i$i1218$lcssa7274; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1224 = $$lcssa7276;$2710 = $2694;
    label = 1146;
    break;
   }
   else if ((label|0) == 1143) {
    HEAP32[$__parent$i1216>>2] = $__nd$0$i$i1218$lcssa7273; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2699 = $__nd$0$i$i1218$lcssa7273; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1224 = $$lcssa7275;$2710 = $2699;
    label = 1146;
    break;
   }
   else if ((label|0) == 1145) {
    HEAP32[$__parent$i1216>>2] = $__nd$0$i$i1218$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2701 = $__nd$0$i$i1218$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2702 = ($__nd$0$i$i1218$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2702) {
     $$0$i7$i1224 = $__parent$i1216;$2710 = $2701;
     label = 1146;
     break;
    } else {
     $2721 = $2688;$__r$0$i1228 = $__nd$0$i$i1218$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1146) {
  __THREW__ = 0;
  $2703 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2704 = __THREW__; __THREW__ = 0;
  $2705 = $2704&1;
  if ($2705) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2706 = ((($2703)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2706>>2] = 1073741893; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2707 = ((($2703)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2707>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2703>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2708 = ((($2703)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2708>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2709 = ((($2703)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2709>>2] = $2710; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1224>>2] = $2703; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2711 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2712 = HEAP32[$2711>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2713 = ($2712|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2713) {
   $2716 = $2703;
  } else {
   $2714 = $2712; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2714; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1226 = HEAP32[$$0$i7$i1224>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2716 = $$pre$i$i1226;
  }
  $2715 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2715,$2716); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2717 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2718 = (($2717) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2718; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4458 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2721 = $$pre4458;$__r$0$i1228 = $2703;
 }
 $2719 = ((($__r$0$i1228)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2719>>2] = 82; //@line 696 "modules/event/sdl/Event.cpp"
 $2720 = ($2721|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2720) {
   HEAP32[$__parent$i1231>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2733 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1239 = $2;$2743 = $2733;
   label = 1160;
  } else {
   $__nd$0$i$i1233 = $2721;
   while(1) {
    $2722 = ((($__nd$0$i$i1233)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2723 = HEAP32[$2722>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2724 = ($2723|0)>(1073741894); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2724) {
     $2725 = HEAP32[$__nd$0$i$i1233>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2726 = ($2725|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2726) {
      $$lcssa7272 = $__nd$0$i$i1233;$__nd$0$i$i1233$lcssa7270 = $__nd$0$i$i1233;
      label = 1153;
      break;
     } else {
      $$sink$i$i1234 = $2725;
     }
    } else {
     $2728 = ($2723|0)<(1073741894); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2728)) {
      $__nd$0$i$i1233$lcssa = $__nd$0$i$i1233;
      label = 1159;
      break;
     }
     $2729 = ((($__nd$0$i$i1233)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2730 = HEAP32[$2729>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2731 = ($2730|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2731) {
      $$lcssa7271 = $2729;$__nd$0$i$i1233$lcssa7269 = $__nd$0$i$i1233;
      label = 1157;
      break;
     } else {
      $$sink$i$i1234 = $2730;
     }
    }
    $__nd$0$i$i1233 = $$sink$i$i1234;
   }
   if ((label|0) == 1153) {
    HEAP32[$__parent$i1231>>2] = $__nd$0$i$i1233$lcssa7270; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2727 = $__nd$0$i$i1233$lcssa7270; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1239 = $$lcssa7272;$2743 = $2727;
    label = 1160;
    break;
   }
   else if ((label|0) == 1157) {
    HEAP32[$__parent$i1231>>2] = $__nd$0$i$i1233$lcssa7269; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2732 = $__nd$0$i$i1233$lcssa7269; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1239 = $$lcssa7271;$2743 = $2732;
    label = 1160;
    break;
   }
   else if ((label|0) == 1159) {
    HEAP32[$__parent$i1231>>2] = $__nd$0$i$i1233$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2734 = $__nd$0$i$i1233$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2735 = ($__nd$0$i$i1233$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2735) {
     $$0$i7$i1239 = $__parent$i1231;$2743 = $2734;
     label = 1160;
     break;
    } else {
     $2754 = $2721;$__r$0$i1243 = $__nd$0$i$i1233$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1160) {
  __THREW__ = 0;
  $2736 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2737 = __THREW__; __THREW__ = 0;
  $2738 = $2737&1;
  if ($2738) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2739 = ((($2736)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2739>>2] = 1073741894; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2740 = ((($2736)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2740>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2736>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2741 = ((($2736)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2741>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2742 = ((($2736)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2742>>2] = $2743; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1239>>2] = $2736; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2744 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2745 = HEAP32[$2744>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2746 = ($2745|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2746) {
   $2749 = $2736;
  } else {
   $2747 = $2745; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2747; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1241 = HEAP32[$$0$i7$i1239>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2749 = $$pre$i$i1241;
  }
  $2748 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2748,$2749); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2750 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2751 = (($2750) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2751; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4459 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2754 = $$pre4459;$__r$0$i1243 = $2736;
 }
 $2752 = ((($__r$0$i1243)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2752>>2] = 83; //@line 698 "modules/event/sdl/Event.cpp"
 $2753 = ($2754|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2753) {
   HEAP32[$__parent$i1246>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2766 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1254 = $2;$2776 = $2766;
   label = 1174;
  } else {
   $__nd$0$i$i1248 = $2754;
   while(1) {
    $2755 = ((($__nd$0$i$i1248)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2756 = HEAP32[$2755>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2757 = ($2756|0)>(1073741895); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2757) {
     $2758 = HEAP32[$__nd$0$i$i1248>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2759 = ($2758|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2759) {
      $$lcssa7268 = $__nd$0$i$i1248;$__nd$0$i$i1248$lcssa7266 = $__nd$0$i$i1248;
      label = 1167;
      break;
     } else {
      $$sink$i$i1249 = $2758;
     }
    } else {
     $2761 = ($2756|0)<(1073741895); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2761)) {
      $__nd$0$i$i1248$lcssa = $__nd$0$i$i1248;
      label = 1173;
      break;
     }
     $2762 = ((($__nd$0$i$i1248)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2763 = HEAP32[$2762>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2764 = ($2763|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2764) {
      $$lcssa7267 = $2762;$__nd$0$i$i1248$lcssa7265 = $__nd$0$i$i1248;
      label = 1171;
      break;
     } else {
      $$sink$i$i1249 = $2763;
     }
    }
    $__nd$0$i$i1248 = $$sink$i$i1249;
   }
   if ((label|0) == 1167) {
    HEAP32[$__parent$i1246>>2] = $__nd$0$i$i1248$lcssa7266; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2760 = $__nd$0$i$i1248$lcssa7266; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1254 = $$lcssa7268;$2776 = $2760;
    label = 1174;
    break;
   }
   else if ((label|0) == 1171) {
    HEAP32[$__parent$i1246>>2] = $__nd$0$i$i1248$lcssa7265; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2765 = $__nd$0$i$i1248$lcssa7265; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1254 = $$lcssa7267;$2776 = $2765;
    label = 1174;
    break;
   }
   else if ((label|0) == 1173) {
    HEAP32[$__parent$i1246>>2] = $__nd$0$i$i1248$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2767 = $__nd$0$i$i1248$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2768 = ($__nd$0$i$i1248$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2768) {
     $$0$i7$i1254 = $__parent$i1246;$2776 = $2767;
     label = 1174;
     break;
    } else {
     $2787 = $2754;$__r$0$i1258 = $__nd$0$i$i1248$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1174) {
  __THREW__ = 0;
  $2769 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2770 = __THREW__; __THREW__ = 0;
  $2771 = $2770&1;
  if ($2771) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2772 = ((($2769)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2772>>2] = 1073741895; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2773 = ((($2769)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2773>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2769>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2774 = ((($2769)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2774>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2775 = ((($2769)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2775>>2] = $2776; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1254>>2] = $2769; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2777 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2778 = HEAP32[$2777>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2779 = ($2778|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2779) {
   $2782 = $2769;
  } else {
   $2780 = $2778; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2780; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1256 = HEAP32[$$0$i7$i1254>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2782 = $$pre$i$i1256;
  }
  $2781 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2781,$2782); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2783 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2784 = (($2783) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2784; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4460 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2787 = $$pre4460;$__r$0$i1258 = $2769;
 }
 $2785 = ((($__r$0$i1258)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2785>>2] = 84; //@line 699 "modules/event/sdl/Event.cpp"
 $2786 = ($2787|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2786) {
   HEAP32[$__parent$i1261>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2799 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1269 = $2;$2809 = $2799;
   label = 1188;
  } else {
   $__nd$0$i$i1263 = $2787;
   while(1) {
    $2788 = ((($__nd$0$i$i1263)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2789 = HEAP32[$2788>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2790 = ($2789|0)>(1073741896); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2790) {
     $2791 = HEAP32[$__nd$0$i$i1263>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2792 = ($2791|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2792) {
      $$lcssa7264 = $__nd$0$i$i1263;$__nd$0$i$i1263$lcssa7262 = $__nd$0$i$i1263;
      label = 1181;
      break;
     } else {
      $$sink$i$i1264 = $2791;
     }
    } else {
     $2794 = ($2789|0)<(1073741896); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2794)) {
      $__nd$0$i$i1263$lcssa = $__nd$0$i$i1263;
      label = 1187;
      break;
     }
     $2795 = ((($__nd$0$i$i1263)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2796 = HEAP32[$2795>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2797 = ($2796|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2797) {
      $$lcssa7263 = $2795;$__nd$0$i$i1263$lcssa7261 = $__nd$0$i$i1263;
      label = 1185;
      break;
     } else {
      $$sink$i$i1264 = $2796;
     }
    }
    $__nd$0$i$i1263 = $$sink$i$i1264;
   }
   if ((label|0) == 1181) {
    HEAP32[$__parent$i1261>>2] = $__nd$0$i$i1263$lcssa7262; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2793 = $__nd$0$i$i1263$lcssa7262; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1269 = $$lcssa7264;$2809 = $2793;
    label = 1188;
    break;
   }
   else if ((label|0) == 1185) {
    HEAP32[$__parent$i1261>>2] = $__nd$0$i$i1263$lcssa7261; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2798 = $__nd$0$i$i1263$lcssa7261; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1269 = $$lcssa7263;$2809 = $2798;
    label = 1188;
    break;
   }
   else if ((label|0) == 1187) {
    HEAP32[$__parent$i1261>>2] = $__nd$0$i$i1263$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2800 = $__nd$0$i$i1263$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2801 = ($__nd$0$i$i1263$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2801) {
     $$0$i7$i1269 = $__parent$i1261;$2809 = $2800;
     label = 1188;
     break;
    } else {
     $2820 = $2787;$__r$0$i1273 = $__nd$0$i$i1263$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1188) {
  __THREW__ = 0;
  $2802 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2803 = __THREW__; __THREW__ = 0;
  $2804 = $2803&1;
  if ($2804) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2805 = ((($2802)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2805>>2] = 1073741896; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2806 = ((($2802)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2806>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2802>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2807 = ((($2802)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2807>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2808 = ((($2802)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2808>>2] = $2809; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1269>>2] = $2802; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2810 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2811 = HEAP32[$2810>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2812 = ($2811|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2812) {
   $2815 = $2802;
  } else {
   $2813 = $2811; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2813; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1271 = HEAP32[$$0$i7$i1269>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2815 = $$pre$i$i1271;
  }
  $2814 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2814,$2815); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2816 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2817 = (($2816) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2817; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4461 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2820 = $$pre4461;$__r$0$i1273 = $2802;
 }
 $2818 = ((($__r$0$i1273)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2818>>2] = 85; //@line 700 "modules/event/sdl/Event.cpp"
 $2819 = ($2820|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2819) {
   HEAP32[$__parent$i1276>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2832 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1284 = $2;$2842 = $2832;
   label = 1202;
  } else {
   $__nd$0$i$i1278 = $2820;
   while(1) {
    $2821 = ((($__nd$0$i$i1278)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2822 = HEAP32[$2821>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2823 = ($2822|0)>(1073741897); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2823) {
     $2824 = HEAP32[$__nd$0$i$i1278>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2825 = ($2824|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2825) {
      $$lcssa7260 = $__nd$0$i$i1278;$__nd$0$i$i1278$lcssa7258 = $__nd$0$i$i1278;
      label = 1195;
      break;
     } else {
      $$sink$i$i1279 = $2824;
     }
    } else {
     $2827 = ($2822|0)<(1073741897); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2827)) {
      $__nd$0$i$i1278$lcssa = $__nd$0$i$i1278;
      label = 1201;
      break;
     }
     $2828 = ((($__nd$0$i$i1278)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2829 = HEAP32[$2828>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2830 = ($2829|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2830) {
      $$lcssa7259 = $2828;$__nd$0$i$i1278$lcssa7257 = $__nd$0$i$i1278;
      label = 1199;
      break;
     } else {
      $$sink$i$i1279 = $2829;
     }
    }
    $__nd$0$i$i1278 = $$sink$i$i1279;
   }
   if ((label|0) == 1195) {
    HEAP32[$__parent$i1276>>2] = $__nd$0$i$i1278$lcssa7258; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2826 = $__nd$0$i$i1278$lcssa7258; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1284 = $$lcssa7260;$2842 = $2826;
    label = 1202;
    break;
   }
   else if ((label|0) == 1199) {
    HEAP32[$__parent$i1276>>2] = $__nd$0$i$i1278$lcssa7257; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2831 = $__nd$0$i$i1278$lcssa7257; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1284 = $$lcssa7259;$2842 = $2831;
    label = 1202;
    break;
   }
   else if ((label|0) == 1201) {
    HEAP32[$__parent$i1276>>2] = $__nd$0$i$i1278$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2833 = $__nd$0$i$i1278$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2834 = ($__nd$0$i$i1278$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2834) {
     $$0$i7$i1284 = $__parent$i1276;$2842 = $2833;
     label = 1202;
     break;
    } else {
     $2853 = $2820;$__r$0$i1288 = $__nd$0$i$i1278$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1202) {
  __THREW__ = 0;
  $2835 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2836 = __THREW__; __THREW__ = 0;
  $2837 = $2836&1;
  if ($2837) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2838 = ((($2835)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2838>>2] = 1073741897; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2839 = ((($2835)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2839>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2835>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2840 = ((($2835)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2840>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2841 = ((($2835)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2841>>2] = $2842; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1284>>2] = $2835; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2843 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2844 = HEAP32[$2843>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2845 = ($2844|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2845) {
   $2848 = $2835;
  } else {
   $2846 = $2844; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2846; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1286 = HEAP32[$$0$i7$i1284>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2848 = $$pre$i$i1286;
  }
  $2847 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2847,$2848); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2849 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2850 = (($2849) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2850; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4462 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2853 = $$pre4462;$__r$0$i1288 = $2835;
 }
 $2851 = ((($__r$0$i1288)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2851>>2] = 86; //@line 701 "modules/event/sdl/Event.cpp"
 $2852 = ($2853|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2852) {
   HEAP32[$__parent$i1291>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2865 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1299 = $2;$2875 = $2865;
   label = 1216;
  } else {
   $__nd$0$i$i1293 = $2853;
   while(1) {
    $2854 = ((($__nd$0$i$i1293)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2855 = HEAP32[$2854>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2856 = ($2855|0)>(1073741898); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2856) {
     $2857 = HEAP32[$__nd$0$i$i1293>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2858 = ($2857|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2858) {
      $$lcssa7256 = $__nd$0$i$i1293;$__nd$0$i$i1293$lcssa7254 = $__nd$0$i$i1293;
      label = 1209;
      break;
     } else {
      $$sink$i$i1294 = $2857;
     }
    } else {
     $2860 = ($2855|0)<(1073741898); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2860)) {
      $__nd$0$i$i1293$lcssa = $__nd$0$i$i1293;
      label = 1215;
      break;
     }
     $2861 = ((($__nd$0$i$i1293)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2862 = HEAP32[$2861>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2863 = ($2862|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2863) {
      $$lcssa7255 = $2861;$__nd$0$i$i1293$lcssa7253 = $__nd$0$i$i1293;
      label = 1213;
      break;
     } else {
      $$sink$i$i1294 = $2862;
     }
    }
    $__nd$0$i$i1293 = $$sink$i$i1294;
   }
   if ((label|0) == 1209) {
    HEAP32[$__parent$i1291>>2] = $__nd$0$i$i1293$lcssa7254; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2859 = $__nd$0$i$i1293$lcssa7254; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1299 = $$lcssa7256;$2875 = $2859;
    label = 1216;
    break;
   }
   else if ((label|0) == 1213) {
    HEAP32[$__parent$i1291>>2] = $__nd$0$i$i1293$lcssa7253; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2864 = $__nd$0$i$i1293$lcssa7253; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1299 = $$lcssa7255;$2875 = $2864;
    label = 1216;
    break;
   }
   else if ((label|0) == 1215) {
    HEAP32[$__parent$i1291>>2] = $__nd$0$i$i1293$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2866 = $__nd$0$i$i1293$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2867 = ($__nd$0$i$i1293$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2867) {
     $$0$i7$i1299 = $__parent$i1291;$2875 = $2866;
     label = 1216;
     break;
    } else {
     $2886 = $2853;$__r$0$i1303 = $__nd$0$i$i1293$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1216) {
  __THREW__ = 0;
  $2868 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2869 = __THREW__; __THREW__ = 0;
  $2870 = $2869&1;
  if ($2870) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2871 = ((($2868)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2871>>2] = 1073741898; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2872 = ((($2868)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2872>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2868>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2873 = ((($2868)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2873>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2874 = ((($2868)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2874>>2] = $2875; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1299>>2] = $2868; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2876 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2877 = HEAP32[$2876>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2878 = ($2877|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2878) {
   $2881 = $2868;
  } else {
   $2879 = $2877; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2879; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1301 = HEAP32[$$0$i7$i1299>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2881 = $$pre$i$i1301;
  }
  $2880 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2880,$2881); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2882 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2883 = (($2882) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2883; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4463 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2886 = $$pre4463;$__r$0$i1303 = $2868;
 }
 $2884 = ((($__r$0$i1303)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2884>>2] = 87; //@line 702 "modules/event/sdl/Event.cpp"
 $2885 = ($2886|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2885) {
   HEAP32[$__parent$i1306>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2898 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1314 = $2;$2908 = $2898;
   label = 1230;
  } else {
   $__nd$0$i$i1308 = $2886;
   while(1) {
    $2887 = ((($__nd$0$i$i1308)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2888 = HEAP32[$2887>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2889 = ($2888|0)>(1073741899); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2889) {
     $2890 = HEAP32[$__nd$0$i$i1308>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2891 = ($2890|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2891) {
      $$lcssa7252 = $__nd$0$i$i1308;$__nd$0$i$i1308$lcssa7250 = $__nd$0$i$i1308;
      label = 1223;
      break;
     } else {
      $$sink$i$i1309 = $2890;
     }
    } else {
     $2893 = ($2888|0)<(1073741899); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2893)) {
      $__nd$0$i$i1308$lcssa = $__nd$0$i$i1308;
      label = 1229;
      break;
     }
     $2894 = ((($__nd$0$i$i1308)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2895 = HEAP32[$2894>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2896 = ($2895|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2896) {
      $$lcssa7251 = $2894;$__nd$0$i$i1308$lcssa7249 = $__nd$0$i$i1308;
      label = 1227;
      break;
     } else {
      $$sink$i$i1309 = $2895;
     }
    }
    $__nd$0$i$i1308 = $$sink$i$i1309;
   }
   if ((label|0) == 1223) {
    HEAP32[$__parent$i1306>>2] = $__nd$0$i$i1308$lcssa7250; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2892 = $__nd$0$i$i1308$lcssa7250; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1314 = $$lcssa7252;$2908 = $2892;
    label = 1230;
    break;
   }
   else if ((label|0) == 1227) {
    HEAP32[$__parent$i1306>>2] = $__nd$0$i$i1308$lcssa7249; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2897 = $__nd$0$i$i1308$lcssa7249; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1314 = $$lcssa7251;$2908 = $2897;
    label = 1230;
    break;
   }
   else if ((label|0) == 1229) {
    HEAP32[$__parent$i1306>>2] = $__nd$0$i$i1308$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2899 = $__nd$0$i$i1308$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2900 = ($__nd$0$i$i1308$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2900) {
     $$0$i7$i1314 = $__parent$i1306;$2908 = $2899;
     label = 1230;
     break;
    } else {
     $2919 = $2886;$__r$0$i1318 = $__nd$0$i$i1308$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1230) {
  __THREW__ = 0;
  $2901 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2902 = __THREW__; __THREW__ = 0;
  $2903 = $2902&1;
  if ($2903) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2904 = ((($2901)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2904>>2] = 1073741899; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2905 = ((($2901)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2905>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2901>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2906 = ((($2901)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2906>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2907 = ((($2901)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2907>>2] = $2908; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1314>>2] = $2901; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2909 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2910 = HEAP32[$2909>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2911 = ($2910|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2911) {
   $2914 = $2901;
  } else {
   $2912 = $2910; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2912; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1316 = HEAP32[$$0$i7$i1314>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2914 = $$pre$i$i1316;
  }
  $2913 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2913,$2914); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2915 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2916 = (($2915) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2916; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4464 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2919 = $$pre4464;$__r$0$i1318 = $2901;
 }
 $2917 = ((($__r$0$i1318)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2917>>2] = 88; //@line 703 "modules/event/sdl/Event.cpp"
 $2918 = ($2919|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2918) {
   HEAP32[$__parent$i1321>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2931 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1329 = $2;$2941 = $2931;
   label = 1244;
  } else {
   $__nd$0$i$i1323 = $2919;
   while(1) {
    $2920 = ((($__nd$0$i$i1323)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2921 = HEAP32[$2920>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2922 = ($2921|0)>(127); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2922) {
     $2923 = HEAP32[$__nd$0$i$i1323>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2924 = ($2923|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2924) {
      $$lcssa7248 = $__nd$0$i$i1323;$__nd$0$i$i1323$lcssa7246 = $__nd$0$i$i1323;
      label = 1237;
      break;
     } else {
      $$sink$i$i1324 = $2923;
     }
    } else {
     $2926 = ($2921|0)<(127); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2926)) {
      $__nd$0$i$i1323$lcssa = $__nd$0$i$i1323;
      label = 1243;
      break;
     }
     $2927 = ((($__nd$0$i$i1323)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2928 = HEAP32[$2927>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2929 = ($2928|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2929) {
      $$lcssa7247 = $2927;$__nd$0$i$i1323$lcssa7245 = $__nd$0$i$i1323;
      label = 1241;
      break;
     } else {
      $$sink$i$i1324 = $2928;
     }
    }
    $__nd$0$i$i1323 = $$sink$i$i1324;
   }
   if ((label|0) == 1237) {
    HEAP32[$__parent$i1321>>2] = $__nd$0$i$i1323$lcssa7246; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2925 = $__nd$0$i$i1323$lcssa7246; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1329 = $$lcssa7248;$2941 = $2925;
    label = 1244;
    break;
   }
   else if ((label|0) == 1241) {
    HEAP32[$__parent$i1321>>2] = $__nd$0$i$i1323$lcssa7245; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2930 = $__nd$0$i$i1323$lcssa7245; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1329 = $$lcssa7247;$2941 = $2930;
    label = 1244;
    break;
   }
   else if ((label|0) == 1243) {
    HEAP32[$__parent$i1321>>2] = $__nd$0$i$i1323$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2932 = $__nd$0$i$i1323$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2933 = ($__nd$0$i$i1323$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2933) {
     $$0$i7$i1329 = $__parent$i1321;$2941 = $2932;
     label = 1244;
     break;
    } else {
     $2952 = $2919;$__r$0$i1333 = $__nd$0$i$i1323$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1244) {
  __THREW__ = 0;
  $2934 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2935 = __THREW__; __THREW__ = 0;
  $2936 = $2935&1;
  if ($2936) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2937 = ((($2934)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2937>>2] = 127; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2938 = ((($2934)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2938>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2934>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2939 = ((($2934)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2939>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2940 = ((($2934)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2940>>2] = $2941; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1329>>2] = $2934; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2942 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2943 = HEAP32[$2942>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2944 = ($2943|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2944) {
   $2947 = $2934;
  } else {
   $2945 = $2943; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2945; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1331 = HEAP32[$$0$i7$i1329>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2947 = $$pre$i$i1331;
  }
  $2946 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2946,$2947); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2948 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2949 = (($2948) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2949; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4465 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2952 = $$pre4465;$__r$0$i1333 = $2934;
 }
 $2950 = ((($__r$0$i1333)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2950>>2] = 89; //@line 704 "modules/event/sdl/Event.cpp"
 $2951 = ($2952|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2951) {
   HEAP32[$__parent$i1336>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2964 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1344 = $2;$2974 = $2964;
   label = 1258;
  } else {
   $__nd$0$i$i1338 = $2952;
   while(1) {
    $2953 = ((($__nd$0$i$i1338)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2954 = HEAP32[$2953>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2955 = ($2954|0)>(1073741901); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2955) {
     $2956 = HEAP32[$__nd$0$i$i1338>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2957 = ($2956|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2957) {
      $$lcssa7244 = $__nd$0$i$i1338;$__nd$0$i$i1338$lcssa7242 = $__nd$0$i$i1338;
      label = 1251;
      break;
     } else {
      $$sink$i$i1339 = $2956;
     }
    } else {
     $2959 = ($2954|0)<(1073741901); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2959)) {
      $__nd$0$i$i1338$lcssa = $__nd$0$i$i1338;
      label = 1257;
      break;
     }
     $2960 = ((($__nd$0$i$i1338)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2961 = HEAP32[$2960>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2962 = ($2961|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2962) {
      $$lcssa7243 = $2960;$__nd$0$i$i1338$lcssa7241 = $__nd$0$i$i1338;
      label = 1255;
      break;
     } else {
      $$sink$i$i1339 = $2961;
     }
    }
    $__nd$0$i$i1338 = $$sink$i$i1339;
   }
   if ((label|0) == 1251) {
    HEAP32[$__parent$i1336>>2] = $__nd$0$i$i1338$lcssa7242; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2958 = $__nd$0$i$i1338$lcssa7242; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1344 = $$lcssa7244;$2974 = $2958;
    label = 1258;
    break;
   }
   else if ((label|0) == 1255) {
    HEAP32[$__parent$i1336>>2] = $__nd$0$i$i1338$lcssa7241; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2963 = $__nd$0$i$i1338$lcssa7241; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1344 = $$lcssa7243;$2974 = $2963;
    label = 1258;
    break;
   }
   else if ((label|0) == 1257) {
    HEAP32[$__parent$i1336>>2] = $__nd$0$i$i1338$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2965 = $__nd$0$i$i1338$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2966 = ($__nd$0$i$i1338$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2966) {
     $$0$i7$i1344 = $__parent$i1336;$2974 = $2965;
     label = 1258;
     break;
    } else {
     $2985 = $2952;$__r$0$i1348 = $__nd$0$i$i1338$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1258) {
  __THREW__ = 0;
  $2967 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $2968 = __THREW__; __THREW__ = 0;
  $2969 = $2968&1;
  if ($2969) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $2970 = ((($2967)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2970>>2] = 1073741901; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $2971 = ((($2967)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$2971>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$2967>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2972 = ((($2967)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2972>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2973 = ((($2967)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$2973>>2] = $2974; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1344>>2] = $2967; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2975 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2976 = HEAP32[$2975>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2977 = ($2976|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($2977) {
   $2980 = $2967;
  } else {
   $2978 = $2976; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $2978; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1346 = HEAP32[$$0$i7$i1344>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $2980 = $$pre$i$i1346;
  }
  $2979 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($2979,$2980); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2981 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2982 = (($2981) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $2982; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4466 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2985 = $$pre4466;$__r$0$i1348 = $2967;
 }
 $2983 = ((($__r$0$i1348)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$2983>>2] = 90; //@line 705 "modules/event/sdl/Event.cpp"
 $2984 = ($2985|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($2984) {
   HEAP32[$__parent$i1351>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $2997 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1359 = $2;$3007 = $2997;
   label = 1272;
  } else {
   $__nd$0$i$i1353 = $2985;
   while(1) {
    $2986 = ((($__nd$0$i$i1353)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2987 = HEAP32[$2986>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $2988 = ($2987|0)>(1073741902); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($2988) {
     $2989 = HEAP32[$__nd$0$i$i1353>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2990 = ($2989|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2990) {
      $$lcssa7240 = $__nd$0$i$i1353;$__nd$0$i$i1353$lcssa7238 = $__nd$0$i$i1353;
      label = 1265;
      break;
     } else {
      $$sink$i$i1354 = $2989;
     }
    } else {
     $2992 = ($2987|0)<(1073741902); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($2992)) {
      $__nd$0$i$i1353$lcssa = $__nd$0$i$i1353;
      label = 1271;
      break;
     }
     $2993 = ((($__nd$0$i$i1353)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2994 = HEAP32[$2993>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $2995 = ($2994|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($2995) {
      $$lcssa7239 = $2993;$__nd$0$i$i1353$lcssa7237 = $__nd$0$i$i1353;
      label = 1269;
      break;
     } else {
      $$sink$i$i1354 = $2994;
     }
    }
    $__nd$0$i$i1353 = $$sink$i$i1354;
   }
   if ((label|0) == 1265) {
    HEAP32[$__parent$i1351>>2] = $__nd$0$i$i1353$lcssa7238; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2991 = $__nd$0$i$i1353$lcssa7238; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1359 = $$lcssa7240;$3007 = $2991;
    label = 1272;
    break;
   }
   else if ((label|0) == 1269) {
    HEAP32[$__parent$i1351>>2] = $__nd$0$i$i1353$lcssa7237; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2996 = $__nd$0$i$i1353$lcssa7237; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1359 = $$lcssa7239;$3007 = $2996;
    label = 1272;
    break;
   }
   else if ((label|0) == 1271) {
    HEAP32[$__parent$i1351>>2] = $__nd$0$i$i1353$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2998 = $__nd$0$i$i1353$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $2999 = ($__nd$0$i$i1353$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($2999) {
     $$0$i7$i1359 = $__parent$i1351;$3007 = $2998;
     label = 1272;
     break;
    } else {
     $3018 = $2985;$__r$0$i1363 = $__nd$0$i$i1353$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1272) {
  __THREW__ = 0;
  $3000 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3001 = __THREW__; __THREW__ = 0;
  $3002 = $3001&1;
  if ($3002) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3003 = ((($3000)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3003>>2] = 1073741902; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3004 = ((($3000)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3004>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3000>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3005 = ((($3000)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3005>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3006 = ((($3000)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3006>>2] = $3007; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1359>>2] = $3000; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3008 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3009 = HEAP32[$3008>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3010 = ($3009|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3010) {
   $3013 = $3000;
  } else {
   $3011 = $3009; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3011; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1361 = HEAP32[$$0$i7$i1359>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3013 = $$pre$i$i1361;
  }
  $3012 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3012,$3013); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3014 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3015 = (($3014) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3015; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4467 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3018 = $$pre4467;$__r$0$i1363 = $3000;
 }
 $3016 = ((($__r$0$i1363)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3016>>2] = 91; //@line 706 "modules/event/sdl/Event.cpp"
 $3017 = ($3018|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3017) {
   HEAP32[$__parent$i1366>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3030 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1374 = $2;$3040 = $3030;
   label = 1286;
  } else {
   $__nd$0$i$i1368 = $3018;
   while(1) {
    $3019 = ((($__nd$0$i$i1368)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3020 = HEAP32[$3019>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3021 = ($3020|0)>(1073741903); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3021) {
     $3022 = HEAP32[$__nd$0$i$i1368>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3023 = ($3022|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3023) {
      $$lcssa7236 = $__nd$0$i$i1368;$__nd$0$i$i1368$lcssa7234 = $__nd$0$i$i1368;
      label = 1279;
      break;
     } else {
      $$sink$i$i1369 = $3022;
     }
    } else {
     $3025 = ($3020|0)<(1073741903); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3025)) {
      $__nd$0$i$i1368$lcssa = $__nd$0$i$i1368;
      label = 1285;
      break;
     }
     $3026 = ((($__nd$0$i$i1368)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3027 = HEAP32[$3026>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3028 = ($3027|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3028) {
      $$lcssa7235 = $3026;$__nd$0$i$i1368$lcssa7233 = $__nd$0$i$i1368;
      label = 1283;
      break;
     } else {
      $$sink$i$i1369 = $3027;
     }
    }
    $__nd$0$i$i1368 = $$sink$i$i1369;
   }
   if ((label|0) == 1279) {
    HEAP32[$__parent$i1366>>2] = $__nd$0$i$i1368$lcssa7234; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3024 = $__nd$0$i$i1368$lcssa7234; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1374 = $$lcssa7236;$3040 = $3024;
    label = 1286;
    break;
   }
   else if ((label|0) == 1283) {
    HEAP32[$__parent$i1366>>2] = $__nd$0$i$i1368$lcssa7233; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3029 = $__nd$0$i$i1368$lcssa7233; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1374 = $$lcssa7235;$3040 = $3029;
    label = 1286;
    break;
   }
   else if ((label|0) == 1285) {
    HEAP32[$__parent$i1366>>2] = $__nd$0$i$i1368$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3031 = $__nd$0$i$i1368$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3032 = ($__nd$0$i$i1368$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3032) {
     $$0$i7$i1374 = $__parent$i1366;$3040 = $3031;
     label = 1286;
     break;
    } else {
     $3051 = $3018;$__r$0$i1378 = $__nd$0$i$i1368$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1286) {
  __THREW__ = 0;
  $3033 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3034 = __THREW__; __THREW__ = 0;
  $3035 = $3034&1;
  if ($3035) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3036 = ((($3033)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3036>>2] = 1073741903; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3037 = ((($3033)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3037>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3033>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3038 = ((($3033)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3038>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3039 = ((($3033)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3039>>2] = $3040; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1374>>2] = $3033; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3041 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3042 = HEAP32[$3041>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3043 = ($3042|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3043) {
   $3046 = $3033;
  } else {
   $3044 = $3042; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3044; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1376 = HEAP32[$$0$i7$i1374>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3046 = $$pre$i$i1376;
  }
  $3045 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3045,$3046); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3047 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3048 = (($3047) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3048; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4468 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3051 = $$pre4468;$__r$0$i1378 = $3033;
 }
 $3049 = ((($__r$0$i1378)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3049>>2] = 92; //@line 707 "modules/event/sdl/Event.cpp"
 $3050 = ($3051|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3050) {
   HEAP32[$__parent$i1381>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3063 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1389 = $2;$3073 = $3063;
   label = 1300;
  } else {
   $__nd$0$i$i1383 = $3051;
   while(1) {
    $3052 = ((($__nd$0$i$i1383)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3053 = HEAP32[$3052>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3054 = ($3053|0)>(1073741904); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3054) {
     $3055 = HEAP32[$__nd$0$i$i1383>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3056 = ($3055|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3056) {
      $$lcssa7232 = $__nd$0$i$i1383;$__nd$0$i$i1383$lcssa7230 = $__nd$0$i$i1383;
      label = 1293;
      break;
     } else {
      $$sink$i$i1384 = $3055;
     }
    } else {
     $3058 = ($3053|0)<(1073741904); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3058)) {
      $__nd$0$i$i1383$lcssa = $__nd$0$i$i1383;
      label = 1299;
      break;
     }
     $3059 = ((($__nd$0$i$i1383)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3060 = HEAP32[$3059>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3061 = ($3060|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3061) {
      $$lcssa7231 = $3059;$__nd$0$i$i1383$lcssa7229 = $__nd$0$i$i1383;
      label = 1297;
      break;
     } else {
      $$sink$i$i1384 = $3060;
     }
    }
    $__nd$0$i$i1383 = $$sink$i$i1384;
   }
   if ((label|0) == 1293) {
    HEAP32[$__parent$i1381>>2] = $__nd$0$i$i1383$lcssa7230; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3057 = $__nd$0$i$i1383$lcssa7230; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1389 = $$lcssa7232;$3073 = $3057;
    label = 1300;
    break;
   }
   else if ((label|0) == 1297) {
    HEAP32[$__parent$i1381>>2] = $__nd$0$i$i1383$lcssa7229; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3062 = $__nd$0$i$i1383$lcssa7229; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1389 = $$lcssa7231;$3073 = $3062;
    label = 1300;
    break;
   }
   else if ((label|0) == 1299) {
    HEAP32[$__parent$i1381>>2] = $__nd$0$i$i1383$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3064 = $__nd$0$i$i1383$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3065 = ($__nd$0$i$i1383$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3065) {
     $$0$i7$i1389 = $__parent$i1381;$3073 = $3064;
     label = 1300;
     break;
    } else {
     $3084 = $3051;$__r$0$i1393 = $__nd$0$i$i1383$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1300) {
  __THREW__ = 0;
  $3066 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3067 = __THREW__; __THREW__ = 0;
  $3068 = $3067&1;
  if ($3068) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3069 = ((($3066)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3069>>2] = 1073741904; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3070 = ((($3066)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3070>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3066>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3071 = ((($3066)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3071>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3072 = ((($3066)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3072>>2] = $3073; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1389>>2] = $3066; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3074 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3075 = HEAP32[$3074>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3076 = ($3075|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3076) {
   $3079 = $3066;
  } else {
   $3077 = $3075; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3077; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1391 = HEAP32[$$0$i7$i1389>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3079 = $$pre$i$i1391;
  }
  $3078 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3078,$3079); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3080 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3081 = (($3080) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3081; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4469 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3084 = $$pre4469;$__r$0$i1393 = $3066;
 }
 $3082 = ((($__r$0$i1393)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3082>>2] = 93; //@line 708 "modules/event/sdl/Event.cpp"
 $3083 = ($3084|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3083) {
   HEAP32[$__parent$i1396>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3096 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1404 = $2;$3106 = $3096;
   label = 1314;
  } else {
   $__nd$0$i$i1398 = $3084;
   while(1) {
    $3085 = ((($__nd$0$i$i1398)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3086 = HEAP32[$3085>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3087 = ($3086|0)>(1073741905); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3087) {
     $3088 = HEAP32[$__nd$0$i$i1398>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3089 = ($3088|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3089) {
      $$lcssa7228 = $__nd$0$i$i1398;$__nd$0$i$i1398$lcssa7226 = $__nd$0$i$i1398;
      label = 1307;
      break;
     } else {
      $$sink$i$i1399 = $3088;
     }
    } else {
     $3091 = ($3086|0)<(1073741905); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3091)) {
      $__nd$0$i$i1398$lcssa = $__nd$0$i$i1398;
      label = 1313;
      break;
     }
     $3092 = ((($__nd$0$i$i1398)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3093 = HEAP32[$3092>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3094 = ($3093|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3094) {
      $$lcssa7227 = $3092;$__nd$0$i$i1398$lcssa7225 = $__nd$0$i$i1398;
      label = 1311;
      break;
     } else {
      $$sink$i$i1399 = $3093;
     }
    }
    $__nd$0$i$i1398 = $$sink$i$i1399;
   }
   if ((label|0) == 1307) {
    HEAP32[$__parent$i1396>>2] = $__nd$0$i$i1398$lcssa7226; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3090 = $__nd$0$i$i1398$lcssa7226; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1404 = $$lcssa7228;$3106 = $3090;
    label = 1314;
    break;
   }
   else if ((label|0) == 1311) {
    HEAP32[$__parent$i1396>>2] = $__nd$0$i$i1398$lcssa7225; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3095 = $__nd$0$i$i1398$lcssa7225; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1404 = $$lcssa7227;$3106 = $3095;
    label = 1314;
    break;
   }
   else if ((label|0) == 1313) {
    HEAP32[$__parent$i1396>>2] = $__nd$0$i$i1398$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3097 = $__nd$0$i$i1398$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3098 = ($__nd$0$i$i1398$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3098) {
     $$0$i7$i1404 = $__parent$i1396;$3106 = $3097;
     label = 1314;
     break;
    } else {
     $3117 = $3084;$__r$0$i1408 = $__nd$0$i$i1398$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1314) {
  __THREW__ = 0;
  $3099 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3100 = __THREW__; __THREW__ = 0;
  $3101 = $3100&1;
  if ($3101) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3102 = ((($3099)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3102>>2] = 1073741905; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3103 = ((($3099)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3103>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3099>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3104 = ((($3099)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3104>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3105 = ((($3099)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3105>>2] = $3106; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1404>>2] = $3099; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3107 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3108 = HEAP32[$3107>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3109 = ($3108|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3109) {
   $3112 = $3099;
  } else {
   $3110 = $3108; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3110; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1406 = HEAP32[$$0$i7$i1404>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3112 = $$pre$i$i1406;
  }
  $3111 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3111,$3112); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3113 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3114 = (($3113) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3114; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4470 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3117 = $$pre4470;$__r$0$i1408 = $3099;
 }
 $3115 = ((($__r$0$i1408)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3115>>2] = 94; //@line 709 "modules/event/sdl/Event.cpp"
 $3116 = ($3117|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3116) {
   HEAP32[$__parent$i1411>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3129 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1419 = $2;$3139 = $3129;
   label = 1328;
  } else {
   $__nd$0$i$i1413 = $3117;
   while(1) {
    $3118 = ((($__nd$0$i$i1413)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3119 = HEAP32[$3118>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3120 = ($3119|0)>(1073741906); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3120) {
     $3121 = HEAP32[$__nd$0$i$i1413>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3122 = ($3121|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3122) {
      $$lcssa7224 = $__nd$0$i$i1413;$__nd$0$i$i1413$lcssa7222 = $__nd$0$i$i1413;
      label = 1321;
      break;
     } else {
      $$sink$i$i1414 = $3121;
     }
    } else {
     $3124 = ($3119|0)<(1073741906); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3124)) {
      $__nd$0$i$i1413$lcssa = $__nd$0$i$i1413;
      label = 1327;
      break;
     }
     $3125 = ((($__nd$0$i$i1413)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3126 = HEAP32[$3125>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3127 = ($3126|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3127) {
      $$lcssa7223 = $3125;$__nd$0$i$i1413$lcssa7221 = $__nd$0$i$i1413;
      label = 1325;
      break;
     } else {
      $$sink$i$i1414 = $3126;
     }
    }
    $__nd$0$i$i1413 = $$sink$i$i1414;
   }
   if ((label|0) == 1321) {
    HEAP32[$__parent$i1411>>2] = $__nd$0$i$i1413$lcssa7222; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3123 = $__nd$0$i$i1413$lcssa7222; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1419 = $$lcssa7224;$3139 = $3123;
    label = 1328;
    break;
   }
   else if ((label|0) == 1325) {
    HEAP32[$__parent$i1411>>2] = $__nd$0$i$i1413$lcssa7221; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3128 = $__nd$0$i$i1413$lcssa7221; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1419 = $$lcssa7223;$3139 = $3128;
    label = 1328;
    break;
   }
   else if ((label|0) == 1327) {
    HEAP32[$__parent$i1411>>2] = $__nd$0$i$i1413$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3130 = $__nd$0$i$i1413$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3131 = ($__nd$0$i$i1413$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3131) {
     $$0$i7$i1419 = $__parent$i1411;$3139 = $3130;
     label = 1328;
     break;
    } else {
     $3150 = $3117;$__r$0$i1423 = $__nd$0$i$i1413$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1328) {
  __THREW__ = 0;
  $3132 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3133 = __THREW__; __THREW__ = 0;
  $3134 = $3133&1;
  if ($3134) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3135 = ((($3132)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3135>>2] = 1073741906; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3136 = ((($3132)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3136>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3132>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3137 = ((($3132)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3137>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3138 = ((($3132)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3138>>2] = $3139; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1419>>2] = $3132; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3140 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3141 = HEAP32[$3140>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3142 = ($3141|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3142) {
   $3145 = $3132;
  } else {
   $3143 = $3141; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3143; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1421 = HEAP32[$$0$i7$i1419>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3145 = $$pre$i$i1421;
  }
  $3144 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3144,$3145); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3146 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3147 = (($3146) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3147; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4471 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3150 = $$pre4471;$__r$0$i1423 = $3132;
 }
 $3148 = ((($__r$0$i1423)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3148>>2] = 95; //@line 710 "modules/event/sdl/Event.cpp"
 $3149 = ($3150|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3149) {
   HEAP32[$__parent$i1426>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3162 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1434 = $2;$3172 = $3162;
   label = 1342;
  } else {
   $__nd$0$i$i1428 = $3150;
   while(1) {
    $3151 = ((($__nd$0$i$i1428)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3152 = HEAP32[$3151>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3153 = ($3152|0)>(1073741907); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3153) {
     $3154 = HEAP32[$__nd$0$i$i1428>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3155 = ($3154|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3155) {
      $$lcssa7220 = $__nd$0$i$i1428;$__nd$0$i$i1428$lcssa7218 = $__nd$0$i$i1428;
      label = 1335;
      break;
     } else {
      $$sink$i$i1429 = $3154;
     }
    } else {
     $3157 = ($3152|0)<(1073741907); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3157)) {
      $__nd$0$i$i1428$lcssa = $__nd$0$i$i1428;
      label = 1341;
      break;
     }
     $3158 = ((($__nd$0$i$i1428)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3159 = HEAP32[$3158>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3160 = ($3159|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3160) {
      $$lcssa7219 = $3158;$__nd$0$i$i1428$lcssa7217 = $__nd$0$i$i1428;
      label = 1339;
      break;
     } else {
      $$sink$i$i1429 = $3159;
     }
    }
    $__nd$0$i$i1428 = $$sink$i$i1429;
   }
   if ((label|0) == 1335) {
    HEAP32[$__parent$i1426>>2] = $__nd$0$i$i1428$lcssa7218; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3156 = $__nd$0$i$i1428$lcssa7218; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1434 = $$lcssa7220;$3172 = $3156;
    label = 1342;
    break;
   }
   else if ((label|0) == 1339) {
    HEAP32[$__parent$i1426>>2] = $__nd$0$i$i1428$lcssa7217; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3161 = $__nd$0$i$i1428$lcssa7217; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1434 = $$lcssa7219;$3172 = $3161;
    label = 1342;
    break;
   }
   else if ((label|0) == 1341) {
    HEAP32[$__parent$i1426>>2] = $__nd$0$i$i1428$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3163 = $__nd$0$i$i1428$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3164 = ($__nd$0$i$i1428$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3164) {
     $$0$i7$i1434 = $__parent$i1426;$3172 = $3163;
     label = 1342;
     break;
    } else {
     $3183 = $3150;$__r$0$i1438 = $__nd$0$i$i1428$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1342) {
  __THREW__ = 0;
  $3165 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3166 = __THREW__; __THREW__ = 0;
  $3167 = $3166&1;
  if ($3167) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3168 = ((($3165)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3168>>2] = 1073741907; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3169 = ((($3165)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3169>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3165>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3170 = ((($3165)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3170>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3171 = ((($3165)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3171>>2] = $3172; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1434>>2] = $3165; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3173 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3174 = HEAP32[$3173>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3175 = ($3174|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3175) {
   $3178 = $3165;
  } else {
   $3176 = $3174; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3176; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1436 = HEAP32[$$0$i7$i1434>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3178 = $$pre$i$i1436;
  }
  $3177 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3177,$3178); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3179 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3180 = (($3179) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3180; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4472 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3183 = $$pre4472;$__r$0$i1438 = $3165;
 }
 $3181 = ((($__r$0$i1438)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3181>>2] = 96; //@line 712 "modules/event/sdl/Event.cpp"
 $3182 = ($3183|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3182) {
   HEAP32[$__parent$i1441>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3195 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1449 = $2;$3205 = $3195;
   label = 1356;
  } else {
   $__nd$0$i$i1443 = $3183;
   while(1) {
    $3184 = ((($__nd$0$i$i1443)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3185 = HEAP32[$3184>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3186 = ($3185|0)>(1073741908); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3186) {
     $3187 = HEAP32[$__nd$0$i$i1443>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3188 = ($3187|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3188) {
      $$lcssa7216 = $__nd$0$i$i1443;$__nd$0$i$i1443$lcssa7214 = $__nd$0$i$i1443;
      label = 1349;
      break;
     } else {
      $$sink$i$i1444 = $3187;
     }
    } else {
     $3190 = ($3185|0)<(1073741908); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3190)) {
      $__nd$0$i$i1443$lcssa = $__nd$0$i$i1443;
      label = 1355;
      break;
     }
     $3191 = ((($__nd$0$i$i1443)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3192 = HEAP32[$3191>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3193 = ($3192|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3193) {
      $$lcssa7215 = $3191;$__nd$0$i$i1443$lcssa7213 = $__nd$0$i$i1443;
      label = 1353;
      break;
     } else {
      $$sink$i$i1444 = $3192;
     }
    }
    $__nd$0$i$i1443 = $$sink$i$i1444;
   }
   if ((label|0) == 1349) {
    HEAP32[$__parent$i1441>>2] = $__nd$0$i$i1443$lcssa7214; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3189 = $__nd$0$i$i1443$lcssa7214; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1449 = $$lcssa7216;$3205 = $3189;
    label = 1356;
    break;
   }
   else if ((label|0) == 1353) {
    HEAP32[$__parent$i1441>>2] = $__nd$0$i$i1443$lcssa7213; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3194 = $__nd$0$i$i1443$lcssa7213; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1449 = $$lcssa7215;$3205 = $3194;
    label = 1356;
    break;
   }
   else if ((label|0) == 1355) {
    HEAP32[$__parent$i1441>>2] = $__nd$0$i$i1443$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3196 = $__nd$0$i$i1443$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3197 = ($__nd$0$i$i1443$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3197) {
     $$0$i7$i1449 = $__parent$i1441;$3205 = $3196;
     label = 1356;
     break;
    } else {
     $3216 = $3183;$__r$0$i1453 = $__nd$0$i$i1443$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1356) {
  __THREW__ = 0;
  $3198 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3199 = __THREW__; __THREW__ = 0;
  $3200 = $3199&1;
  if ($3200) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3201 = ((($3198)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3201>>2] = 1073741908; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3202 = ((($3198)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3202>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3198>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3203 = ((($3198)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3203>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3204 = ((($3198)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3204>>2] = $3205; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1449>>2] = $3198; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3206 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3207 = HEAP32[$3206>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3208 = ($3207|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3208) {
   $3211 = $3198;
  } else {
   $3209 = $3207; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3209; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1451 = HEAP32[$$0$i7$i1449>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3211 = $$pre$i$i1451;
  }
  $3210 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3210,$3211); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3212 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3213 = (($3212) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3213; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4473 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3216 = $$pre4473;$__r$0$i1453 = $3198;
 }
 $3214 = ((($__r$0$i1453)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3214>>2] = 97; //@line 713 "modules/event/sdl/Event.cpp"
 $3215 = ($3216|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3215) {
   HEAP32[$__parent$i1456>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3228 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1464 = $2;$3238 = $3228;
   label = 1370;
  } else {
   $__nd$0$i$i1458 = $3216;
   while(1) {
    $3217 = ((($__nd$0$i$i1458)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3218 = HEAP32[$3217>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3219 = ($3218|0)>(1073741909); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3219) {
     $3220 = HEAP32[$__nd$0$i$i1458>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3221 = ($3220|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3221) {
      $$lcssa7212 = $__nd$0$i$i1458;$__nd$0$i$i1458$lcssa7210 = $__nd$0$i$i1458;
      label = 1363;
      break;
     } else {
      $$sink$i$i1459 = $3220;
     }
    } else {
     $3223 = ($3218|0)<(1073741909); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3223)) {
      $__nd$0$i$i1458$lcssa = $__nd$0$i$i1458;
      label = 1369;
      break;
     }
     $3224 = ((($__nd$0$i$i1458)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3225 = HEAP32[$3224>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3226 = ($3225|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3226) {
      $$lcssa7211 = $3224;$__nd$0$i$i1458$lcssa7209 = $__nd$0$i$i1458;
      label = 1367;
      break;
     } else {
      $$sink$i$i1459 = $3225;
     }
    }
    $__nd$0$i$i1458 = $$sink$i$i1459;
   }
   if ((label|0) == 1363) {
    HEAP32[$__parent$i1456>>2] = $__nd$0$i$i1458$lcssa7210; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3222 = $__nd$0$i$i1458$lcssa7210; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1464 = $$lcssa7212;$3238 = $3222;
    label = 1370;
    break;
   }
   else if ((label|0) == 1367) {
    HEAP32[$__parent$i1456>>2] = $__nd$0$i$i1458$lcssa7209; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3227 = $__nd$0$i$i1458$lcssa7209; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1464 = $$lcssa7211;$3238 = $3227;
    label = 1370;
    break;
   }
   else if ((label|0) == 1369) {
    HEAP32[$__parent$i1456>>2] = $__nd$0$i$i1458$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3229 = $__nd$0$i$i1458$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3230 = ($__nd$0$i$i1458$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3230) {
     $$0$i7$i1464 = $__parent$i1456;$3238 = $3229;
     label = 1370;
     break;
    } else {
     $3249 = $3216;$__r$0$i1468 = $__nd$0$i$i1458$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1370) {
  __THREW__ = 0;
  $3231 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3232 = __THREW__; __THREW__ = 0;
  $3233 = $3232&1;
  if ($3233) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3234 = ((($3231)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3234>>2] = 1073741909; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3235 = ((($3231)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3235>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3231>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3236 = ((($3231)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3236>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3237 = ((($3231)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3237>>2] = $3238; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1464>>2] = $3231; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3239 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3240 = HEAP32[$3239>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3241 = ($3240|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3241) {
   $3244 = $3231;
  } else {
   $3242 = $3240; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3242; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1466 = HEAP32[$$0$i7$i1464>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3244 = $$pre$i$i1466;
  }
  $3243 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3243,$3244); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3245 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3246 = (($3245) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3246; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4474 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3249 = $$pre4474;$__r$0$i1468 = $3231;
 }
 $3247 = ((($__r$0$i1468)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3247>>2] = 98; //@line 714 "modules/event/sdl/Event.cpp"
 $3248 = ($3249|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3248) {
   HEAP32[$__parent$i1471>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3261 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1479 = $2;$3271 = $3261;
   label = 1384;
  } else {
   $__nd$0$i$i1473 = $3249;
   while(1) {
    $3250 = ((($__nd$0$i$i1473)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3251 = HEAP32[$3250>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3252 = ($3251|0)>(1073741910); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3252) {
     $3253 = HEAP32[$__nd$0$i$i1473>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3254 = ($3253|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3254) {
      $$lcssa7208 = $__nd$0$i$i1473;$__nd$0$i$i1473$lcssa7206 = $__nd$0$i$i1473;
      label = 1377;
      break;
     } else {
      $$sink$i$i1474 = $3253;
     }
    } else {
     $3256 = ($3251|0)<(1073741910); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3256)) {
      $__nd$0$i$i1473$lcssa = $__nd$0$i$i1473;
      label = 1383;
      break;
     }
     $3257 = ((($__nd$0$i$i1473)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3258 = HEAP32[$3257>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3259 = ($3258|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3259) {
      $$lcssa7207 = $3257;$__nd$0$i$i1473$lcssa7205 = $__nd$0$i$i1473;
      label = 1381;
      break;
     } else {
      $$sink$i$i1474 = $3258;
     }
    }
    $__nd$0$i$i1473 = $$sink$i$i1474;
   }
   if ((label|0) == 1377) {
    HEAP32[$__parent$i1471>>2] = $__nd$0$i$i1473$lcssa7206; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3255 = $__nd$0$i$i1473$lcssa7206; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1479 = $$lcssa7208;$3271 = $3255;
    label = 1384;
    break;
   }
   else if ((label|0) == 1381) {
    HEAP32[$__parent$i1471>>2] = $__nd$0$i$i1473$lcssa7205; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3260 = $__nd$0$i$i1473$lcssa7205; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1479 = $$lcssa7207;$3271 = $3260;
    label = 1384;
    break;
   }
   else if ((label|0) == 1383) {
    HEAP32[$__parent$i1471>>2] = $__nd$0$i$i1473$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3262 = $__nd$0$i$i1473$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3263 = ($__nd$0$i$i1473$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3263) {
     $$0$i7$i1479 = $__parent$i1471;$3271 = $3262;
     label = 1384;
     break;
    } else {
     $__r$0$i1483 = $__nd$0$i$i1473$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1384) {
  __THREW__ = 0;
  $3264 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3265 = __THREW__; __THREW__ = 0;
  $3266 = $3265&1;
  if ($3266) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3267 = ((($3264)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3267>>2] = 1073741910; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3268 = ((($3264)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3268>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3264>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3269 = ((($3264)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3269>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3270 = ((($3264)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3270>>2] = $3271; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1479>>2] = $3264; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3272 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3273 = HEAP32[$3272>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3274 = ($3273|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3274) {
   $3277 = $3264;
  } else {
   $3275 = $3273; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3275; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1481 = HEAP32[$$0$i7$i1479>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3277 = $$pre$i$i1481;
  }
  $3276 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3276,$3277); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3278 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3279 = (($3278) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3279; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $__r$0$i1483 = $3264;
 }
 $3280 = ((($__r$0$i1483)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3280>>2] = 99; //@line 715 "modules/event/sdl/Event.cpp"
 $3281 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $3282 = ($3281|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3282) {
   HEAP32[$__parent$i1486>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3294 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1494 = $2;$3304 = $3294;
   label = 1398;
  } else {
   $__nd$0$i$i1488 = $3281;
   while(1) {
    $3283 = ((($__nd$0$i$i1488)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3284 = HEAP32[$3283>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3285 = ($3284|0)>(1073741911); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3285) {
     $3286 = HEAP32[$__nd$0$i$i1488>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3287 = ($3286|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3287) {
      $$lcssa7204 = $__nd$0$i$i1488;$__nd$0$i$i1488$lcssa7202 = $__nd$0$i$i1488;
      label = 1391;
      break;
     } else {
      $$sink$i$i1489 = $3286;
     }
    } else {
     $3289 = ($3284|0)<(1073741911); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3289)) {
      $__nd$0$i$i1488$lcssa = $__nd$0$i$i1488;
      label = 1397;
      break;
     }
     $3290 = ((($__nd$0$i$i1488)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3291 = HEAP32[$3290>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3292 = ($3291|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3292) {
      $$lcssa7203 = $3290;$__nd$0$i$i1488$lcssa7201 = $__nd$0$i$i1488;
      label = 1395;
      break;
     } else {
      $$sink$i$i1489 = $3291;
     }
    }
    $__nd$0$i$i1488 = $$sink$i$i1489;
   }
   if ((label|0) == 1391) {
    HEAP32[$__parent$i1486>>2] = $__nd$0$i$i1488$lcssa7202; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3288 = $__nd$0$i$i1488$lcssa7202; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1494 = $$lcssa7204;$3304 = $3288;
    label = 1398;
    break;
   }
   else if ((label|0) == 1395) {
    HEAP32[$__parent$i1486>>2] = $__nd$0$i$i1488$lcssa7201; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3293 = $__nd$0$i$i1488$lcssa7201; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1494 = $$lcssa7203;$3304 = $3293;
    label = 1398;
    break;
   }
   else if ((label|0) == 1397) {
    HEAP32[$__parent$i1486>>2] = $__nd$0$i$i1488$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3295 = $__nd$0$i$i1488$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3296 = ($__nd$0$i$i1488$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3296) {
     $$0$i7$i1494 = $__parent$i1486;$3304 = $3295;
     label = 1398;
     break;
    } else {
     $3315 = $3281;$__r$0$i1498 = $__nd$0$i$i1488$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1398) {
  __THREW__ = 0;
  $3297 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3298 = __THREW__; __THREW__ = 0;
  $3299 = $3298&1;
  if ($3299) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3300 = ((($3297)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3300>>2] = 1073741911; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3301 = ((($3297)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3301>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3297>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3302 = ((($3297)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3302>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3303 = ((($3297)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3303>>2] = $3304; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1494>>2] = $3297; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3305 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3306 = HEAP32[$3305>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3307 = ($3306|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3307) {
   $3310 = $3297;
  } else {
   $3308 = $3306; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3308; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1496 = HEAP32[$$0$i7$i1494>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3310 = $$pre$i$i1496;
  }
  $3309 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3309,$3310); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3311 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3312 = (($3311) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3312; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4475 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3315 = $$pre4475;$__r$0$i1498 = $3297;
 }
 $3313 = ((($__r$0$i1498)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3313>>2] = 100; //@line 716 "modules/event/sdl/Event.cpp"
 $3314 = ($3315|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3314) {
   HEAP32[$__parent$i1501>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3327 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1509 = $2;$3337 = $3327;
   label = 1412;
  } else {
   $__nd$0$i$i1503 = $3315;
   while(1) {
    $3316 = ((($__nd$0$i$i1503)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3317 = HEAP32[$3316>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3318 = ($3317|0)>(1073741912); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3318) {
     $3319 = HEAP32[$__nd$0$i$i1503>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3320 = ($3319|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3320) {
      $$lcssa7200 = $__nd$0$i$i1503;$__nd$0$i$i1503$lcssa7198 = $__nd$0$i$i1503;
      label = 1405;
      break;
     } else {
      $$sink$i$i1504 = $3319;
     }
    } else {
     $3322 = ($3317|0)<(1073741912); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3322)) {
      $__nd$0$i$i1503$lcssa = $__nd$0$i$i1503;
      label = 1411;
      break;
     }
     $3323 = ((($__nd$0$i$i1503)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3324 = HEAP32[$3323>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3325 = ($3324|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3325) {
      $$lcssa7199 = $3323;$__nd$0$i$i1503$lcssa7197 = $__nd$0$i$i1503;
      label = 1409;
      break;
     } else {
      $$sink$i$i1504 = $3324;
     }
    }
    $__nd$0$i$i1503 = $$sink$i$i1504;
   }
   if ((label|0) == 1405) {
    HEAP32[$__parent$i1501>>2] = $__nd$0$i$i1503$lcssa7198; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3321 = $__nd$0$i$i1503$lcssa7198; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1509 = $$lcssa7200;$3337 = $3321;
    label = 1412;
    break;
   }
   else if ((label|0) == 1409) {
    HEAP32[$__parent$i1501>>2] = $__nd$0$i$i1503$lcssa7197; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3326 = $__nd$0$i$i1503$lcssa7197; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1509 = $$lcssa7199;$3337 = $3326;
    label = 1412;
    break;
   }
   else if ((label|0) == 1411) {
    HEAP32[$__parent$i1501>>2] = $__nd$0$i$i1503$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3328 = $__nd$0$i$i1503$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3329 = ($__nd$0$i$i1503$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3329) {
     $$0$i7$i1509 = $__parent$i1501;$3337 = $3328;
     label = 1412;
     break;
    } else {
     $3348 = $3315;$__r$0$i1513 = $__nd$0$i$i1503$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1412) {
  __THREW__ = 0;
  $3330 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3331 = __THREW__; __THREW__ = 0;
  $3332 = $3331&1;
  if ($3332) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3333 = ((($3330)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3333>>2] = 1073741912; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3334 = ((($3330)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3334>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3330>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3335 = ((($3330)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3335>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3336 = ((($3330)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3336>>2] = $3337; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1509>>2] = $3330; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3338 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3339 = HEAP32[$3338>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3340 = ($3339|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3340) {
   $3343 = $3330;
  } else {
   $3341 = $3339; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3341; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1511 = HEAP32[$$0$i7$i1509>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3343 = $$pre$i$i1511;
  }
  $3342 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3342,$3343); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3344 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3345 = (($3344) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3345; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4476 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3348 = $$pre4476;$__r$0$i1513 = $3330;
 }
 $3346 = ((($__r$0$i1513)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3346>>2] = 101; //@line 717 "modules/event/sdl/Event.cpp"
 $3347 = ($3348|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3347) {
   HEAP32[$__parent$i1516>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3360 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1524 = $2;$3370 = $3360;
   label = 1426;
  } else {
   $__nd$0$i$i1518 = $3348;
   while(1) {
    $3349 = ((($__nd$0$i$i1518)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3350 = HEAP32[$3349>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3351 = ($3350|0)>(1073741922); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3351) {
     $3352 = HEAP32[$__nd$0$i$i1518>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3353 = ($3352|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3353) {
      $$lcssa7196 = $__nd$0$i$i1518;$__nd$0$i$i1518$lcssa7194 = $__nd$0$i$i1518;
      label = 1419;
      break;
     } else {
      $$sink$i$i1519 = $3352;
     }
    } else {
     $3355 = ($3350|0)<(1073741922); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3355)) {
      $__nd$0$i$i1518$lcssa = $__nd$0$i$i1518;
      label = 1425;
      break;
     }
     $3356 = ((($__nd$0$i$i1518)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3357 = HEAP32[$3356>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3358 = ($3357|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3358) {
      $$lcssa7195 = $3356;$__nd$0$i$i1518$lcssa7193 = $__nd$0$i$i1518;
      label = 1423;
      break;
     } else {
      $$sink$i$i1519 = $3357;
     }
    }
    $__nd$0$i$i1518 = $$sink$i$i1519;
   }
   if ((label|0) == 1419) {
    HEAP32[$__parent$i1516>>2] = $__nd$0$i$i1518$lcssa7194; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3354 = $__nd$0$i$i1518$lcssa7194; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1524 = $$lcssa7196;$3370 = $3354;
    label = 1426;
    break;
   }
   else if ((label|0) == 1423) {
    HEAP32[$__parent$i1516>>2] = $__nd$0$i$i1518$lcssa7193; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3359 = $__nd$0$i$i1518$lcssa7193; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1524 = $$lcssa7195;$3370 = $3359;
    label = 1426;
    break;
   }
   else if ((label|0) == 1425) {
    HEAP32[$__parent$i1516>>2] = $__nd$0$i$i1518$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3361 = $__nd$0$i$i1518$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3362 = ($__nd$0$i$i1518$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3362) {
     $$0$i7$i1524 = $__parent$i1516;$3370 = $3361;
     label = 1426;
     break;
    } else {
     $3381 = $3348;$__r$0$i1528 = $__nd$0$i$i1518$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1426) {
  __THREW__ = 0;
  $3363 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3364 = __THREW__; __THREW__ = 0;
  $3365 = $3364&1;
  if ($3365) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3366 = ((($3363)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3366>>2] = 1073741922; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3367 = ((($3363)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3367>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3363>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3368 = ((($3363)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3368>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3369 = ((($3363)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3369>>2] = $3370; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1524>>2] = $3363; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3371 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3372 = HEAP32[$3371>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3373 = ($3372|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3373) {
   $3376 = $3363;
  } else {
   $3374 = $3372; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3374; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1526 = HEAP32[$$0$i7$i1524>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3376 = $$pre$i$i1526;
  }
  $3375 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3375,$3376); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3377 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3378 = (($3377) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3378; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4477 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3381 = $$pre4477;$__r$0$i1528 = $3363;
 }
 $3379 = ((($__r$0$i1528)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3379>>2] = 111; //@line 718 "modules/event/sdl/Event.cpp"
 $3380 = ($3381|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3380) {
   HEAP32[$__parent$i1531>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3393 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1539 = $2;$3403 = $3393;
   label = 1440;
  } else {
   $__nd$0$i$i1533 = $3381;
   while(1) {
    $3382 = ((($__nd$0$i$i1533)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3383 = HEAP32[$3382>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3384 = ($3383|0)>(1073741913); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3384) {
     $3385 = HEAP32[$__nd$0$i$i1533>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3386 = ($3385|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3386) {
      $$lcssa7192 = $__nd$0$i$i1533;$__nd$0$i$i1533$lcssa7190 = $__nd$0$i$i1533;
      label = 1433;
      break;
     } else {
      $$sink$i$i1534 = $3385;
     }
    } else {
     $3388 = ($3383|0)<(1073741913); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3388)) {
      $__nd$0$i$i1533$lcssa = $__nd$0$i$i1533;
      label = 1439;
      break;
     }
     $3389 = ((($__nd$0$i$i1533)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3390 = HEAP32[$3389>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3391 = ($3390|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3391) {
      $$lcssa7191 = $3389;$__nd$0$i$i1533$lcssa7189 = $__nd$0$i$i1533;
      label = 1437;
      break;
     } else {
      $$sink$i$i1534 = $3390;
     }
    }
    $__nd$0$i$i1533 = $$sink$i$i1534;
   }
   if ((label|0) == 1433) {
    HEAP32[$__parent$i1531>>2] = $__nd$0$i$i1533$lcssa7190; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3387 = $__nd$0$i$i1533$lcssa7190; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1539 = $$lcssa7192;$3403 = $3387;
    label = 1440;
    break;
   }
   else if ((label|0) == 1437) {
    HEAP32[$__parent$i1531>>2] = $__nd$0$i$i1533$lcssa7189; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3392 = $__nd$0$i$i1533$lcssa7189; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1539 = $$lcssa7191;$3403 = $3392;
    label = 1440;
    break;
   }
   else if ((label|0) == 1439) {
    HEAP32[$__parent$i1531>>2] = $__nd$0$i$i1533$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3394 = $__nd$0$i$i1533$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3395 = ($__nd$0$i$i1533$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3395) {
     $$0$i7$i1539 = $__parent$i1531;$3403 = $3394;
     label = 1440;
     break;
    } else {
     $3414 = $3381;$__r$0$i1543 = $__nd$0$i$i1533$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1440) {
  __THREW__ = 0;
  $3396 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3397 = __THREW__; __THREW__ = 0;
  $3398 = $3397&1;
  if ($3398) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3399 = ((($3396)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3399>>2] = 1073741913; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3400 = ((($3396)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3400>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3396>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3401 = ((($3396)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3401>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3402 = ((($3396)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3402>>2] = $3403; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1539>>2] = $3396; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3404 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3405 = HEAP32[$3404>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3406 = ($3405|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3406) {
   $3409 = $3396;
  } else {
   $3407 = $3405; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3407; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1541 = HEAP32[$$0$i7$i1539>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3409 = $$pre$i$i1541;
  }
  $3408 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3408,$3409); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3410 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3411 = (($3410) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3411; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4478 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3414 = $$pre4478;$__r$0$i1543 = $3396;
 }
 $3412 = ((($__r$0$i1543)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3412>>2] = 102; //@line 719 "modules/event/sdl/Event.cpp"
 $3413 = ($3414|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3413) {
   HEAP32[$__parent$i1546>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3426 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1554 = $2;$3436 = $3426;
   label = 1454;
  } else {
   $__nd$0$i$i1548 = $3414;
   while(1) {
    $3415 = ((($__nd$0$i$i1548)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3416 = HEAP32[$3415>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3417 = ($3416|0)>(1073741914); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3417) {
     $3418 = HEAP32[$__nd$0$i$i1548>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3419 = ($3418|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3419) {
      $$lcssa7188 = $__nd$0$i$i1548;$__nd$0$i$i1548$lcssa7186 = $__nd$0$i$i1548;
      label = 1447;
      break;
     } else {
      $$sink$i$i1549 = $3418;
     }
    } else {
     $3421 = ($3416|0)<(1073741914); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3421)) {
      $__nd$0$i$i1548$lcssa = $__nd$0$i$i1548;
      label = 1453;
      break;
     }
     $3422 = ((($__nd$0$i$i1548)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3423 = HEAP32[$3422>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3424 = ($3423|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3424) {
      $$lcssa7187 = $3422;$__nd$0$i$i1548$lcssa7185 = $__nd$0$i$i1548;
      label = 1451;
      break;
     } else {
      $$sink$i$i1549 = $3423;
     }
    }
    $__nd$0$i$i1548 = $$sink$i$i1549;
   }
   if ((label|0) == 1447) {
    HEAP32[$__parent$i1546>>2] = $__nd$0$i$i1548$lcssa7186; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3420 = $__nd$0$i$i1548$lcssa7186; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1554 = $$lcssa7188;$3436 = $3420;
    label = 1454;
    break;
   }
   else if ((label|0) == 1451) {
    HEAP32[$__parent$i1546>>2] = $__nd$0$i$i1548$lcssa7185; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3425 = $__nd$0$i$i1548$lcssa7185; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1554 = $$lcssa7187;$3436 = $3425;
    label = 1454;
    break;
   }
   else if ((label|0) == 1453) {
    HEAP32[$__parent$i1546>>2] = $__nd$0$i$i1548$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3427 = $__nd$0$i$i1548$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3428 = ($__nd$0$i$i1548$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3428) {
     $$0$i7$i1554 = $__parent$i1546;$3436 = $3427;
     label = 1454;
     break;
    } else {
     $3447 = $3414;$__r$0$i1558 = $__nd$0$i$i1548$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1454) {
  __THREW__ = 0;
  $3429 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3430 = __THREW__; __THREW__ = 0;
  $3431 = $3430&1;
  if ($3431) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3432 = ((($3429)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3432>>2] = 1073741914; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3433 = ((($3429)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3433>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3429>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3434 = ((($3429)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3434>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3435 = ((($3429)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3435>>2] = $3436; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1554>>2] = $3429; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3437 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3438 = HEAP32[$3437>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3439 = ($3438|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3439) {
   $3442 = $3429;
  } else {
   $3440 = $3438; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3440; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1556 = HEAP32[$$0$i7$i1554>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3442 = $$pre$i$i1556;
  }
  $3441 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3441,$3442); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3443 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3444 = (($3443) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3444; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4479 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3447 = $$pre4479;$__r$0$i1558 = $3429;
 }
 $3445 = ((($__r$0$i1558)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3445>>2] = 103; //@line 720 "modules/event/sdl/Event.cpp"
 $3446 = ($3447|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3446) {
   HEAP32[$__parent$i1561>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3459 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1569 = $2;$3469 = $3459;
   label = 1468;
  } else {
   $__nd$0$i$i1563 = $3447;
   while(1) {
    $3448 = ((($__nd$0$i$i1563)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3449 = HEAP32[$3448>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3450 = ($3449|0)>(1073741915); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3450) {
     $3451 = HEAP32[$__nd$0$i$i1563>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3452 = ($3451|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3452) {
      $$lcssa7184 = $__nd$0$i$i1563;$__nd$0$i$i1563$lcssa7182 = $__nd$0$i$i1563;
      label = 1461;
      break;
     } else {
      $$sink$i$i1564 = $3451;
     }
    } else {
     $3454 = ($3449|0)<(1073741915); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3454)) {
      $__nd$0$i$i1563$lcssa = $__nd$0$i$i1563;
      label = 1467;
      break;
     }
     $3455 = ((($__nd$0$i$i1563)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3456 = HEAP32[$3455>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3457 = ($3456|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3457) {
      $$lcssa7183 = $3455;$__nd$0$i$i1563$lcssa7181 = $__nd$0$i$i1563;
      label = 1465;
      break;
     } else {
      $$sink$i$i1564 = $3456;
     }
    }
    $__nd$0$i$i1563 = $$sink$i$i1564;
   }
   if ((label|0) == 1461) {
    HEAP32[$__parent$i1561>>2] = $__nd$0$i$i1563$lcssa7182; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3453 = $__nd$0$i$i1563$lcssa7182; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1569 = $$lcssa7184;$3469 = $3453;
    label = 1468;
    break;
   }
   else if ((label|0) == 1465) {
    HEAP32[$__parent$i1561>>2] = $__nd$0$i$i1563$lcssa7181; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3458 = $__nd$0$i$i1563$lcssa7181; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1569 = $$lcssa7183;$3469 = $3458;
    label = 1468;
    break;
   }
   else if ((label|0) == 1467) {
    HEAP32[$__parent$i1561>>2] = $__nd$0$i$i1563$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3460 = $__nd$0$i$i1563$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3461 = ($__nd$0$i$i1563$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3461) {
     $$0$i7$i1569 = $__parent$i1561;$3469 = $3460;
     label = 1468;
     break;
    } else {
     $3480 = $3447;$__r$0$i1573 = $__nd$0$i$i1563$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1468) {
  __THREW__ = 0;
  $3462 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3463 = __THREW__; __THREW__ = 0;
  $3464 = $3463&1;
  if ($3464) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3465 = ((($3462)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3465>>2] = 1073741915; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3466 = ((($3462)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3466>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3462>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3467 = ((($3462)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3467>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3468 = ((($3462)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3468>>2] = $3469; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1569>>2] = $3462; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3470 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3471 = HEAP32[$3470>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3472 = ($3471|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3472) {
   $3475 = $3462;
  } else {
   $3473 = $3471; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3473; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1571 = HEAP32[$$0$i7$i1569>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3475 = $$pre$i$i1571;
  }
  $3474 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3474,$3475); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3476 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3477 = (($3476) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3477; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4480 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3480 = $$pre4480;$__r$0$i1573 = $3462;
 }
 $3478 = ((($__r$0$i1573)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3478>>2] = 104; //@line 721 "modules/event/sdl/Event.cpp"
 $3479 = ($3480|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3479) {
   HEAP32[$__parent$i1576>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3492 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1584 = $2;$3502 = $3492;
   label = 1482;
  } else {
   $__nd$0$i$i1578 = $3480;
   while(1) {
    $3481 = ((($__nd$0$i$i1578)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3482 = HEAP32[$3481>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3483 = ($3482|0)>(1073741916); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3483) {
     $3484 = HEAP32[$__nd$0$i$i1578>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3485 = ($3484|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3485) {
      $$lcssa7180 = $__nd$0$i$i1578;$__nd$0$i$i1578$lcssa7178 = $__nd$0$i$i1578;
      label = 1475;
      break;
     } else {
      $$sink$i$i1579 = $3484;
     }
    } else {
     $3487 = ($3482|0)<(1073741916); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3487)) {
      $__nd$0$i$i1578$lcssa = $__nd$0$i$i1578;
      label = 1481;
      break;
     }
     $3488 = ((($__nd$0$i$i1578)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3489 = HEAP32[$3488>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3490 = ($3489|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3490) {
      $$lcssa7179 = $3488;$__nd$0$i$i1578$lcssa7177 = $__nd$0$i$i1578;
      label = 1479;
      break;
     } else {
      $$sink$i$i1579 = $3489;
     }
    }
    $__nd$0$i$i1578 = $$sink$i$i1579;
   }
   if ((label|0) == 1475) {
    HEAP32[$__parent$i1576>>2] = $__nd$0$i$i1578$lcssa7178; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3486 = $__nd$0$i$i1578$lcssa7178; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1584 = $$lcssa7180;$3502 = $3486;
    label = 1482;
    break;
   }
   else if ((label|0) == 1479) {
    HEAP32[$__parent$i1576>>2] = $__nd$0$i$i1578$lcssa7177; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3491 = $__nd$0$i$i1578$lcssa7177; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1584 = $$lcssa7179;$3502 = $3491;
    label = 1482;
    break;
   }
   else if ((label|0) == 1481) {
    HEAP32[$__parent$i1576>>2] = $__nd$0$i$i1578$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3493 = $__nd$0$i$i1578$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3494 = ($__nd$0$i$i1578$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3494) {
     $$0$i7$i1584 = $__parent$i1576;$3502 = $3493;
     label = 1482;
     break;
    } else {
     $3513 = $3480;$__r$0$i1588 = $__nd$0$i$i1578$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1482) {
  __THREW__ = 0;
  $3495 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3496 = __THREW__; __THREW__ = 0;
  $3497 = $3496&1;
  if ($3497) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3498 = ((($3495)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3498>>2] = 1073741916; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3499 = ((($3495)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3499>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3495>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3500 = ((($3495)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3500>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3501 = ((($3495)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3501>>2] = $3502; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1584>>2] = $3495; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3503 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3504 = HEAP32[$3503>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3505 = ($3504|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3505) {
   $3508 = $3495;
  } else {
   $3506 = $3504; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3506; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1586 = HEAP32[$$0$i7$i1584>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3508 = $$pre$i$i1586;
  }
  $3507 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3507,$3508); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3509 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3510 = (($3509) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3510; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4481 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3513 = $$pre4481;$__r$0$i1588 = $3495;
 }
 $3511 = ((($__r$0$i1588)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3511>>2] = 105; //@line 722 "modules/event/sdl/Event.cpp"
 $3512 = ($3513|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3512) {
   HEAP32[$__parent$i1591>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3525 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1599 = $2;$3535 = $3525;
   label = 1496;
  } else {
   $__nd$0$i$i1593 = $3513;
   while(1) {
    $3514 = ((($__nd$0$i$i1593)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3515 = HEAP32[$3514>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3516 = ($3515|0)>(1073741917); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3516) {
     $3517 = HEAP32[$__nd$0$i$i1593>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3518 = ($3517|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3518) {
      $$lcssa7176 = $__nd$0$i$i1593;$__nd$0$i$i1593$lcssa7174 = $__nd$0$i$i1593;
      label = 1489;
      break;
     } else {
      $$sink$i$i1594 = $3517;
     }
    } else {
     $3520 = ($3515|0)<(1073741917); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3520)) {
      $__nd$0$i$i1593$lcssa = $__nd$0$i$i1593;
      label = 1495;
      break;
     }
     $3521 = ((($__nd$0$i$i1593)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3522 = HEAP32[$3521>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3523 = ($3522|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3523) {
      $$lcssa7175 = $3521;$__nd$0$i$i1593$lcssa7173 = $__nd$0$i$i1593;
      label = 1493;
      break;
     } else {
      $$sink$i$i1594 = $3522;
     }
    }
    $__nd$0$i$i1593 = $$sink$i$i1594;
   }
   if ((label|0) == 1489) {
    HEAP32[$__parent$i1591>>2] = $__nd$0$i$i1593$lcssa7174; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3519 = $__nd$0$i$i1593$lcssa7174; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1599 = $$lcssa7176;$3535 = $3519;
    label = 1496;
    break;
   }
   else if ((label|0) == 1493) {
    HEAP32[$__parent$i1591>>2] = $__nd$0$i$i1593$lcssa7173; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3524 = $__nd$0$i$i1593$lcssa7173; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1599 = $$lcssa7175;$3535 = $3524;
    label = 1496;
    break;
   }
   else if ((label|0) == 1495) {
    HEAP32[$__parent$i1591>>2] = $__nd$0$i$i1593$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3526 = $__nd$0$i$i1593$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3527 = ($__nd$0$i$i1593$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3527) {
     $$0$i7$i1599 = $__parent$i1591;$3535 = $3526;
     label = 1496;
     break;
    } else {
     $3546 = $3513;$__r$0$i1603 = $__nd$0$i$i1593$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1496) {
  __THREW__ = 0;
  $3528 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3529 = __THREW__; __THREW__ = 0;
  $3530 = $3529&1;
  if ($3530) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3531 = ((($3528)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3531>>2] = 1073741917; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3532 = ((($3528)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3532>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3528>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3533 = ((($3528)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3533>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3534 = ((($3528)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3534>>2] = $3535; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1599>>2] = $3528; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3536 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3537 = HEAP32[$3536>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3538 = ($3537|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3538) {
   $3541 = $3528;
  } else {
   $3539 = $3537; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3539; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1601 = HEAP32[$$0$i7$i1599>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3541 = $$pre$i$i1601;
  }
  $3540 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3540,$3541); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3542 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3543 = (($3542) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3543; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4482 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3546 = $$pre4482;$__r$0$i1603 = $3528;
 }
 $3544 = ((($__r$0$i1603)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3544>>2] = 106; //@line 723 "modules/event/sdl/Event.cpp"
 $3545 = ($3546|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3545) {
   HEAP32[$__parent$i1606>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3558 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1614 = $2;$3568 = $3558;
   label = 1510;
  } else {
   $__nd$0$i$i1608 = $3546;
   while(1) {
    $3547 = ((($__nd$0$i$i1608)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3548 = HEAP32[$3547>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3549 = ($3548|0)>(1073741918); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3549) {
     $3550 = HEAP32[$__nd$0$i$i1608>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3551 = ($3550|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3551) {
      $$lcssa7172 = $__nd$0$i$i1608;$__nd$0$i$i1608$lcssa7170 = $__nd$0$i$i1608;
      label = 1503;
      break;
     } else {
      $$sink$i$i1609 = $3550;
     }
    } else {
     $3553 = ($3548|0)<(1073741918); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3553)) {
      $__nd$0$i$i1608$lcssa = $__nd$0$i$i1608;
      label = 1509;
      break;
     }
     $3554 = ((($__nd$0$i$i1608)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3555 = HEAP32[$3554>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3556 = ($3555|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3556) {
      $$lcssa7171 = $3554;$__nd$0$i$i1608$lcssa7169 = $__nd$0$i$i1608;
      label = 1507;
      break;
     } else {
      $$sink$i$i1609 = $3555;
     }
    }
    $__nd$0$i$i1608 = $$sink$i$i1609;
   }
   if ((label|0) == 1503) {
    HEAP32[$__parent$i1606>>2] = $__nd$0$i$i1608$lcssa7170; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3552 = $__nd$0$i$i1608$lcssa7170; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1614 = $$lcssa7172;$3568 = $3552;
    label = 1510;
    break;
   }
   else if ((label|0) == 1507) {
    HEAP32[$__parent$i1606>>2] = $__nd$0$i$i1608$lcssa7169; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3557 = $__nd$0$i$i1608$lcssa7169; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1614 = $$lcssa7171;$3568 = $3557;
    label = 1510;
    break;
   }
   else if ((label|0) == 1509) {
    HEAP32[$__parent$i1606>>2] = $__nd$0$i$i1608$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3559 = $__nd$0$i$i1608$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3560 = ($__nd$0$i$i1608$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3560) {
     $$0$i7$i1614 = $__parent$i1606;$3568 = $3559;
     label = 1510;
     break;
    } else {
     $3579 = $3546;$__r$0$i1618 = $__nd$0$i$i1608$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1510) {
  __THREW__ = 0;
  $3561 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3562 = __THREW__; __THREW__ = 0;
  $3563 = $3562&1;
  if ($3563) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3564 = ((($3561)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3564>>2] = 1073741918; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3565 = ((($3561)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3565>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3561>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3566 = ((($3561)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3566>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3567 = ((($3561)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3567>>2] = $3568; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1614>>2] = $3561; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3569 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3570 = HEAP32[$3569>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3571 = ($3570|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3571) {
   $3574 = $3561;
  } else {
   $3572 = $3570; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3572; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1616 = HEAP32[$$0$i7$i1614>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3574 = $$pre$i$i1616;
  }
  $3573 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3573,$3574); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3575 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3576 = (($3575) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3576; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4483 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3579 = $$pre4483;$__r$0$i1618 = $3561;
 }
 $3577 = ((($__r$0$i1618)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3577>>2] = 107; //@line 724 "modules/event/sdl/Event.cpp"
 $3578 = ($3579|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3578) {
   HEAP32[$__parent$i1621>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3591 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1629 = $2;$3601 = $3591;
   label = 1524;
  } else {
   $__nd$0$i$i1623 = $3579;
   while(1) {
    $3580 = ((($__nd$0$i$i1623)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3581 = HEAP32[$3580>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3582 = ($3581|0)>(1073741919); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3582) {
     $3583 = HEAP32[$__nd$0$i$i1623>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3584 = ($3583|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3584) {
      $$lcssa7168 = $__nd$0$i$i1623;$__nd$0$i$i1623$lcssa7166 = $__nd$0$i$i1623;
      label = 1517;
      break;
     } else {
      $$sink$i$i1624 = $3583;
     }
    } else {
     $3586 = ($3581|0)<(1073741919); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3586)) {
      $__nd$0$i$i1623$lcssa = $__nd$0$i$i1623;
      label = 1523;
      break;
     }
     $3587 = ((($__nd$0$i$i1623)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3588 = HEAP32[$3587>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3589 = ($3588|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3589) {
      $$lcssa7167 = $3587;$__nd$0$i$i1623$lcssa7165 = $__nd$0$i$i1623;
      label = 1521;
      break;
     } else {
      $$sink$i$i1624 = $3588;
     }
    }
    $__nd$0$i$i1623 = $$sink$i$i1624;
   }
   if ((label|0) == 1517) {
    HEAP32[$__parent$i1621>>2] = $__nd$0$i$i1623$lcssa7166; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3585 = $__nd$0$i$i1623$lcssa7166; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1629 = $$lcssa7168;$3601 = $3585;
    label = 1524;
    break;
   }
   else if ((label|0) == 1521) {
    HEAP32[$__parent$i1621>>2] = $__nd$0$i$i1623$lcssa7165; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3590 = $__nd$0$i$i1623$lcssa7165; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1629 = $$lcssa7167;$3601 = $3590;
    label = 1524;
    break;
   }
   else if ((label|0) == 1523) {
    HEAP32[$__parent$i1621>>2] = $__nd$0$i$i1623$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3592 = $__nd$0$i$i1623$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3593 = ($__nd$0$i$i1623$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3593) {
     $$0$i7$i1629 = $__parent$i1621;$3601 = $3592;
     label = 1524;
     break;
    } else {
     $3612 = $3579;$__r$0$i1633 = $__nd$0$i$i1623$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1524) {
  __THREW__ = 0;
  $3594 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3595 = __THREW__; __THREW__ = 0;
  $3596 = $3595&1;
  if ($3596) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3597 = ((($3594)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3597>>2] = 1073741919; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3598 = ((($3594)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3598>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3594>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3599 = ((($3594)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3599>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3600 = ((($3594)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3600>>2] = $3601; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1629>>2] = $3594; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3602 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3603 = HEAP32[$3602>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3604 = ($3603|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3604) {
   $3607 = $3594;
  } else {
   $3605 = $3603; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3605; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1631 = HEAP32[$$0$i7$i1629>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3607 = $$pre$i$i1631;
  }
  $3606 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3606,$3607); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3608 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3609 = (($3608) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3609; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4484 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3612 = $$pre4484;$__r$0$i1633 = $3594;
 }
 $3610 = ((($__r$0$i1633)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3610>>2] = 108; //@line 725 "modules/event/sdl/Event.cpp"
 $3611 = ($3612|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3611) {
   HEAP32[$__parent$i1636>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3624 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1644 = $2;$3634 = $3624;
   label = 1538;
  } else {
   $__nd$0$i$i1638 = $3612;
   while(1) {
    $3613 = ((($__nd$0$i$i1638)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3614 = HEAP32[$3613>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3615 = ($3614|0)>(1073741920); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3615) {
     $3616 = HEAP32[$__nd$0$i$i1638>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3617 = ($3616|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3617) {
      $$lcssa7164 = $__nd$0$i$i1638;$__nd$0$i$i1638$lcssa7162 = $__nd$0$i$i1638;
      label = 1531;
      break;
     } else {
      $$sink$i$i1639 = $3616;
     }
    } else {
     $3619 = ($3614|0)<(1073741920); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3619)) {
      $__nd$0$i$i1638$lcssa = $__nd$0$i$i1638;
      label = 1537;
      break;
     }
     $3620 = ((($__nd$0$i$i1638)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3621 = HEAP32[$3620>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3622 = ($3621|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3622) {
      $$lcssa7163 = $3620;$__nd$0$i$i1638$lcssa7161 = $__nd$0$i$i1638;
      label = 1535;
      break;
     } else {
      $$sink$i$i1639 = $3621;
     }
    }
    $__nd$0$i$i1638 = $$sink$i$i1639;
   }
   if ((label|0) == 1531) {
    HEAP32[$__parent$i1636>>2] = $__nd$0$i$i1638$lcssa7162; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3618 = $__nd$0$i$i1638$lcssa7162; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1644 = $$lcssa7164;$3634 = $3618;
    label = 1538;
    break;
   }
   else if ((label|0) == 1535) {
    HEAP32[$__parent$i1636>>2] = $__nd$0$i$i1638$lcssa7161; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3623 = $__nd$0$i$i1638$lcssa7161; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1644 = $$lcssa7163;$3634 = $3623;
    label = 1538;
    break;
   }
   else if ((label|0) == 1537) {
    HEAP32[$__parent$i1636>>2] = $__nd$0$i$i1638$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3625 = $__nd$0$i$i1638$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3626 = ($__nd$0$i$i1638$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3626) {
     $$0$i7$i1644 = $__parent$i1636;$3634 = $3625;
     label = 1538;
     break;
    } else {
     $3645 = $3612;$__r$0$i1648 = $__nd$0$i$i1638$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1538) {
  __THREW__ = 0;
  $3627 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3628 = __THREW__; __THREW__ = 0;
  $3629 = $3628&1;
  if ($3629) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3630 = ((($3627)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3630>>2] = 1073741920; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3631 = ((($3627)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3631>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3627>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3632 = ((($3627)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3632>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3633 = ((($3627)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3633>>2] = $3634; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1644>>2] = $3627; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3635 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3636 = HEAP32[$3635>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3637 = ($3636|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3637) {
   $3640 = $3627;
  } else {
   $3638 = $3636; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3638; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1646 = HEAP32[$$0$i7$i1644>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3640 = $$pre$i$i1646;
  }
  $3639 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3639,$3640); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3641 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3642 = (($3641) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3642; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4485 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3645 = $$pre4485;$__r$0$i1648 = $3627;
 }
 $3643 = ((($__r$0$i1648)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3643>>2] = 109; //@line 726 "modules/event/sdl/Event.cpp"
 $3644 = ($3645|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3644) {
   HEAP32[$__parent$i1651>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3657 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1659 = $2;$3667 = $3657;
   label = 1552;
  } else {
   $__nd$0$i$i1653 = $3645;
   while(1) {
    $3646 = ((($__nd$0$i$i1653)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3647 = HEAP32[$3646>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3648 = ($3647|0)>(1073741921); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3648) {
     $3649 = HEAP32[$__nd$0$i$i1653>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3650 = ($3649|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3650) {
      $$lcssa7160 = $__nd$0$i$i1653;$__nd$0$i$i1653$lcssa7158 = $__nd$0$i$i1653;
      label = 1545;
      break;
     } else {
      $$sink$i$i1654 = $3649;
     }
    } else {
     $3652 = ($3647|0)<(1073741921); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3652)) {
      $__nd$0$i$i1653$lcssa = $__nd$0$i$i1653;
      label = 1551;
      break;
     }
     $3653 = ((($__nd$0$i$i1653)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3654 = HEAP32[$3653>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3655 = ($3654|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3655) {
      $$lcssa7159 = $3653;$__nd$0$i$i1653$lcssa7157 = $__nd$0$i$i1653;
      label = 1549;
      break;
     } else {
      $$sink$i$i1654 = $3654;
     }
    }
    $__nd$0$i$i1653 = $$sink$i$i1654;
   }
   if ((label|0) == 1545) {
    HEAP32[$__parent$i1651>>2] = $__nd$0$i$i1653$lcssa7158; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3651 = $__nd$0$i$i1653$lcssa7158; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1659 = $$lcssa7160;$3667 = $3651;
    label = 1552;
    break;
   }
   else if ((label|0) == 1549) {
    HEAP32[$__parent$i1651>>2] = $__nd$0$i$i1653$lcssa7157; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3656 = $__nd$0$i$i1653$lcssa7157; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1659 = $$lcssa7159;$3667 = $3656;
    label = 1552;
    break;
   }
   else if ((label|0) == 1551) {
    HEAP32[$__parent$i1651>>2] = $__nd$0$i$i1653$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3658 = $__nd$0$i$i1653$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3659 = ($__nd$0$i$i1653$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3659) {
     $$0$i7$i1659 = $__parent$i1651;$3667 = $3658;
     label = 1552;
     break;
    } else {
     $3678 = $3645;$__r$0$i1663 = $__nd$0$i$i1653$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1552) {
  __THREW__ = 0;
  $3660 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3661 = __THREW__; __THREW__ = 0;
  $3662 = $3661&1;
  if ($3662) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3663 = ((($3660)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3663>>2] = 1073741921; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3664 = ((($3660)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3664>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3660>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3665 = ((($3660)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3665>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3666 = ((($3660)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3666>>2] = $3667; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1659>>2] = $3660; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3668 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3669 = HEAP32[$3668>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3670 = ($3669|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3670) {
   $3673 = $3660;
  } else {
   $3671 = $3669; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3671; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1661 = HEAP32[$$0$i7$i1659>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3673 = $$pre$i$i1661;
  }
  $3672 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3672,$3673); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3674 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3675 = (($3674) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3675; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4486 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3678 = $$pre4486;$__r$0$i1663 = $3660;
 }
 $3676 = ((($__r$0$i1663)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3676>>2] = 110; //@line 727 "modules/event/sdl/Event.cpp"
 $3677 = ($3678|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3677) {
   HEAP32[$__parent$i1666>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3690 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1674 = $2;$3700 = $3690;
   label = 1566;
  } else {
   $__nd$0$i$i1668 = $3678;
   while(1) {
    $3679 = ((($__nd$0$i$i1668)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3680 = HEAP32[$3679>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3681 = ($3680|0)>(1073741923); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3681) {
     $3682 = HEAP32[$__nd$0$i$i1668>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3683 = ($3682|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3683) {
      $$lcssa7156 = $__nd$0$i$i1668;$__nd$0$i$i1668$lcssa7154 = $__nd$0$i$i1668;
      label = 1559;
      break;
     } else {
      $$sink$i$i1669 = $3682;
     }
    } else {
     $3685 = ($3680|0)<(1073741923); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3685)) {
      $__nd$0$i$i1668$lcssa = $__nd$0$i$i1668;
      label = 1565;
      break;
     }
     $3686 = ((($__nd$0$i$i1668)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3687 = HEAP32[$3686>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3688 = ($3687|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3688) {
      $$lcssa7155 = $3686;$__nd$0$i$i1668$lcssa7153 = $__nd$0$i$i1668;
      label = 1563;
      break;
     } else {
      $$sink$i$i1669 = $3687;
     }
    }
    $__nd$0$i$i1668 = $$sink$i$i1669;
   }
   if ((label|0) == 1559) {
    HEAP32[$__parent$i1666>>2] = $__nd$0$i$i1668$lcssa7154; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3684 = $__nd$0$i$i1668$lcssa7154; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1674 = $$lcssa7156;$3700 = $3684;
    label = 1566;
    break;
   }
   else if ((label|0) == 1563) {
    HEAP32[$__parent$i1666>>2] = $__nd$0$i$i1668$lcssa7153; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3689 = $__nd$0$i$i1668$lcssa7153; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1674 = $$lcssa7155;$3700 = $3689;
    label = 1566;
    break;
   }
   else if ((label|0) == 1565) {
    HEAP32[$__parent$i1666>>2] = $__nd$0$i$i1668$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3691 = $__nd$0$i$i1668$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3692 = ($__nd$0$i$i1668$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3692) {
     $$0$i7$i1674 = $__parent$i1666;$3700 = $3691;
     label = 1566;
     break;
    } else {
     $3711 = $3678;$__r$0$i1678 = $__nd$0$i$i1668$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1566) {
  __THREW__ = 0;
  $3693 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3694 = __THREW__; __THREW__ = 0;
  $3695 = $3694&1;
  if ($3695) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3696 = ((($3693)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3696>>2] = 1073741923; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3697 = ((($3693)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3697>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3693>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3698 = ((($3693)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3698>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3699 = ((($3693)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3699>>2] = $3700; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1674>>2] = $3693; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3701 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3702 = HEAP32[$3701>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3703 = ($3702|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3703) {
   $3706 = $3693;
  } else {
   $3704 = $3702; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3704; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1676 = HEAP32[$$0$i7$i1674>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3706 = $$pre$i$i1676;
  }
  $3705 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3705,$3706); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3707 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3708 = (($3707) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3708; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4487 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3711 = $$pre4487;$__r$0$i1678 = $3693;
 }
 $3709 = ((($__r$0$i1678)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3709>>2] = 112; //@line 728 "modules/event/sdl/Event.cpp"
 $3710 = ($3711|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3710) {
   HEAP32[$__parent$i1681>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3723 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1689 = $2;$3733 = $3723;
   label = 1580;
  } else {
   $__nd$0$i$i1683 = $3711;
   while(1) {
    $3712 = ((($__nd$0$i$i1683)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3713 = HEAP32[$3712>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3714 = ($3713|0)>(1073741957); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3714) {
     $3715 = HEAP32[$__nd$0$i$i1683>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3716 = ($3715|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3716) {
      $$lcssa7152 = $__nd$0$i$i1683;$__nd$0$i$i1683$lcssa7150 = $__nd$0$i$i1683;
      label = 1573;
      break;
     } else {
      $$sink$i$i1684 = $3715;
     }
    } else {
     $3718 = ($3713|0)<(1073741957); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3718)) {
      $__nd$0$i$i1683$lcssa = $__nd$0$i$i1683;
      label = 1579;
      break;
     }
     $3719 = ((($__nd$0$i$i1683)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3720 = HEAP32[$3719>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3721 = ($3720|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3721) {
      $$lcssa7151 = $3719;$__nd$0$i$i1683$lcssa7149 = $__nd$0$i$i1683;
      label = 1577;
      break;
     } else {
      $$sink$i$i1684 = $3720;
     }
    }
    $__nd$0$i$i1683 = $$sink$i$i1684;
   }
   if ((label|0) == 1573) {
    HEAP32[$__parent$i1681>>2] = $__nd$0$i$i1683$lcssa7150; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3717 = $__nd$0$i$i1683$lcssa7150; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1689 = $$lcssa7152;$3733 = $3717;
    label = 1580;
    break;
   }
   else if ((label|0) == 1577) {
    HEAP32[$__parent$i1681>>2] = $__nd$0$i$i1683$lcssa7149; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3722 = $__nd$0$i$i1683$lcssa7149; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1689 = $$lcssa7151;$3733 = $3722;
    label = 1580;
    break;
   }
   else if ((label|0) == 1579) {
    HEAP32[$__parent$i1681>>2] = $__nd$0$i$i1683$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3724 = $__nd$0$i$i1683$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3725 = ($__nd$0$i$i1683$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3725) {
     $$0$i7$i1689 = $__parent$i1681;$3733 = $3724;
     label = 1580;
     break;
    } else {
     $3744 = $3711;$__r$0$i1693 = $__nd$0$i$i1683$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1580) {
  __THREW__ = 0;
  $3726 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3727 = __THREW__; __THREW__ = 0;
  $3728 = $3727&1;
  if ($3728) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3729 = ((($3726)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3729>>2] = 1073741957; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3730 = ((($3726)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3730>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3726>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3731 = ((($3726)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3731>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3732 = ((($3726)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3732>>2] = $3733; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1689>>2] = $3726; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3734 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3735 = HEAP32[$3734>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3736 = ($3735|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3736) {
   $3739 = $3726;
  } else {
   $3737 = $3735; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3737; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1691 = HEAP32[$$0$i7$i1689>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3739 = $$pre$i$i1691;
  }
  $3738 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3738,$3739); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3740 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3741 = (($3740) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3741; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4488 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3744 = $$pre4488;$__r$0$i1693 = $3726;
 }
 $3742 = ((($__r$0$i1693)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3742>>2] = 113; //@line 729 "modules/event/sdl/Event.cpp"
 $3743 = ($3744|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3743) {
   HEAP32[$__parent$i1696>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3756 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1704 = $2;$3766 = $3756;
   label = 1594;
  } else {
   $__nd$0$i$i1698 = $3744;
   while(1) {
    $3745 = ((($__nd$0$i$i1698)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3746 = HEAP32[$3745>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3747 = ($3746|0)>(1073741927); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3747) {
     $3748 = HEAP32[$__nd$0$i$i1698>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3749 = ($3748|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3749) {
      $$lcssa7148 = $__nd$0$i$i1698;$__nd$0$i$i1698$lcssa7146 = $__nd$0$i$i1698;
      label = 1587;
      break;
     } else {
      $$sink$i$i1699 = $3748;
     }
    } else {
     $3751 = ($3746|0)<(1073741927); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3751)) {
      $__nd$0$i$i1698$lcssa = $__nd$0$i$i1698;
      label = 1593;
      break;
     }
     $3752 = ((($__nd$0$i$i1698)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3753 = HEAP32[$3752>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3754 = ($3753|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3754) {
      $$lcssa7147 = $3752;$__nd$0$i$i1698$lcssa7145 = $__nd$0$i$i1698;
      label = 1591;
      break;
     } else {
      $$sink$i$i1699 = $3753;
     }
    }
    $__nd$0$i$i1698 = $$sink$i$i1699;
   }
   if ((label|0) == 1587) {
    HEAP32[$__parent$i1696>>2] = $__nd$0$i$i1698$lcssa7146; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3750 = $__nd$0$i$i1698$lcssa7146; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1704 = $$lcssa7148;$3766 = $3750;
    label = 1594;
    break;
   }
   else if ((label|0) == 1591) {
    HEAP32[$__parent$i1696>>2] = $__nd$0$i$i1698$lcssa7145; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3755 = $__nd$0$i$i1698$lcssa7145; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1704 = $$lcssa7147;$3766 = $3755;
    label = 1594;
    break;
   }
   else if ((label|0) == 1593) {
    HEAP32[$__parent$i1696>>2] = $__nd$0$i$i1698$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3757 = $__nd$0$i$i1698$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3758 = ($__nd$0$i$i1698$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3758) {
     $$0$i7$i1704 = $__parent$i1696;$3766 = $3757;
     label = 1594;
     break;
    } else {
     $3777 = $3744;$__r$0$i1708 = $__nd$0$i$i1698$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1594) {
  __THREW__ = 0;
  $3759 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3760 = __THREW__; __THREW__ = 0;
  $3761 = $3760&1;
  if ($3761) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3762 = ((($3759)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3762>>2] = 1073741927; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3763 = ((($3759)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3763>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3759>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3764 = ((($3759)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3764>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3765 = ((($3759)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3765>>2] = $3766; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1704>>2] = $3759; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3767 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3768 = HEAP32[$3767>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3769 = ($3768|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3769) {
   $3772 = $3759;
  } else {
   $3770 = $3768; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3770; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1706 = HEAP32[$$0$i7$i1704>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3772 = $$pre$i$i1706;
  }
  $3771 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3771,$3772); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3773 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3774 = (($3773) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3774; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4489 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3777 = $$pre4489;$__r$0$i1708 = $3759;
 }
 $3775 = ((($__r$0$i1708)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3775>>2] = 114; //@line 730 "modules/event/sdl/Event.cpp"
 $3776 = ($3777|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3776) {
   HEAP32[$__parent$i1711>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3789 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1719 = $2;$3799 = $3789;
   label = 1608;
  } else {
   $__nd$0$i$i1713 = $3777;
   while(1) {
    $3778 = ((($__nd$0$i$i1713)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3779 = HEAP32[$3778>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3780 = ($3779|0)>(1073741925); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3780) {
     $3781 = HEAP32[$__nd$0$i$i1713>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3782 = ($3781|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3782) {
      $$lcssa7144 = $__nd$0$i$i1713;$__nd$0$i$i1713$lcssa7142 = $__nd$0$i$i1713;
      label = 1601;
      break;
     } else {
      $$sink$i$i1714 = $3781;
     }
    } else {
     $3784 = ($3779|0)<(1073741925); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3784)) {
      $__nd$0$i$i1713$lcssa = $__nd$0$i$i1713;
      label = 1607;
      break;
     }
     $3785 = ((($__nd$0$i$i1713)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3786 = HEAP32[$3785>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3787 = ($3786|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3787) {
      $$lcssa7143 = $3785;$__nd$0$i$i1713$lcssa7141 = $__nd$0$i$i1713;
      label = 1605;
      break;
     } else {
      $$sink$i$i1714 = $3786;
     }
    }
    $__nd$0$i$i1713 = $$sink$i$i1714;
   }
   if ((label|0) == 1601) {
    HEAP32[$__parent$i1711>>2] = $__nd$0$i$i1713$lcssa7142; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3783 = $__nd$0$i$i1713$lcssa7142; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1719 = $$lcssa7144;$3799 = $3783;
    label = 1608;
    break;
   }
   else if ((label|0) == 1605) {
    HEAP32[$__parent$i1711>>2] = $__nd$0$i$i1713$lcssa7141; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3788 = $__nd$0$i$i1713$lcssa7141; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1719 = $$lcssa7143;$3799 = $3788;
    label = 1608;
    break;
   }
   else if ((label|0) == 1607) {
    HEAP32[$__parent$i1711>>2] = $__nd$0$i$i1713$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3790 = $__nd$0$i$i1713$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3791 = ($__nd$0$i$i1713$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3791) {
     $$0$i7$i1719 = $__parent$i1711;$3799 = $3790;
     label = 1608;
     break;
    } else {
     $3810 = $3777;$__r$0$i1723 = $__nd$0$i$i1713$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1608) {
  __THREW__ = 0;
  $3792 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3793 = __THREW__; __THREW__ = 0;
  $3794 = $3793&1;
  if ($3794) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3795 = ((($3792)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3795>>2] = 1073741925; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3796 = ((($3792)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3796>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3792>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3797 = ((($3792)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3797>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3798 = ((($3792)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3798>>2] = $3799; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1719>>2] = $3792; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3800 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3801 = HEAP32[$3800>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3802 = ($3801|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3802) {
   $3805 = $3792;
  } else {
   $3803 = $3801; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3803; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1721 = HEAP32[$$0$i7$i1719>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3805 = $$pre$i$i1721;
  }
  $3804 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3804,$3805); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3806 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3807 = (($3806) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3807; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4490 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3810 = $$pre4490;$__r$0$i1723 = $3792;
 }
 $3808 = ((($__r$0$i1723)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3808>>2] = 115; //@line 732 "modules/event/sdl/Event.cpp"
 $3809 = ($3810|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3809) {
   HEAP32[$__parent$i1726>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3822 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1734 = $2;$3832 = $3822;
   label = 1622;
  } else {
   $__nd$0$i$i1728 = $3810;
   while(1) {
    $3811 = ((($__nd$0$i$i1728)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3812 = HEAP32[$3811>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3813 = ($3812|0)>(1073741926); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3813) {
     $3814 = HEAP32[$__nd$0$i$i1728>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3815 = ($3814|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3815) {
      $$lcssa7140 = $__nd$0$i$i1728;$__nd$0$i$i1728$lcssa7138 = $__nd$0$i$i1728;
      label = 1615;
      break;
     } else {
      $$sink$i$i1729 = $3814;
     }
    } else {
     $3817 = ($3812|0)<(1073741926); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3817)) {
      $__nd$0$i$i1728$lcssa = $__nd$0$i$i1728;
      label = 1621;
      break;
     }
     $3818 = ((($__nd$0$i$i1728)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3819 = HEAP32[$3818>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3820 = ($3819|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3820) {
      $$lcssa7139 = $3818;$__nd$0$i$i1728$lcssa7137 = $__nd$0$i$i1728;
      label = 1619;
      break;
     } else {
      $$sink$i$i1729 = $3819;
     }
    }
    $__nd$0$i$i1728 = $$sink$i$i1729;
   }
   if ((label|0) == 1615) {
    HEAP32[$__parent$i1726>>2] = $__nd$0$i$i1728$lcssa7138; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3816 = $__nd$0$i$i1728$lcssa7138; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1734 = $$lcssa7140;$3832 = $3816;
    label = 1622;
    break;
   }
   else if ((label|0) == 1619) {
    HEAP32[$__parent$i1726>>2] = $__nd$0$i$i1728$lcssa7137; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3821 = $__nd$0$i$i1728$lcssa7137; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1734 = $$lcssa7139;$3832 = $3821;
    label = 1622;
    break;
   }
   else if ((label|0) == 1621) {
    HEAP32[$__parent$i1726>>2] = $__nd$0$i$i1728$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3823 = $__nd$0$i$i1728$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3824 = ($__nd$0$i$i1728$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3824) {
     $$0$i7$i1734 = $__parent$i1726;$3832 = $3823;
     label = 1622;
     break;
    } else {
     $3843 = $3810;$__r$0$i1738 = $__nd$0$i$i1728$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1622) {
  __THREW__ = 0;
  $3825 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3826 = __THREW__; __THREW__ = 0;
  $3827 = $3826&1;
  if ($3827) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3828 = ((($3825)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3828>>2] = 1073741926; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3829 = ((($3825)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3829>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3825>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3830 = ((($3825)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3830>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3831 = ((($3825)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3831>>2] = $3832; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1734>>2] = $3825; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3833 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3834 = HEAP32[$3833>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3835 = ($3834|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3835) {
   $3838 = $3825;
  } else {
   $3836 = $3834; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3836; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1736 = HEAP32[$$0$i7$i1734>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3838 = $$pre$i$i1736;
  }
  $3837 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3837,$3838); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3839 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3840 = (($3839) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3840; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4491 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3843 = $$pre4491;$__r$0$i1738 = $3825;
 }
 $3841 = ((($__r$0$i1738)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3841>>2] = 116; //@line 733 "modules/event/sdl/Event.cpp"
 $3842 = ($3843|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3842) {
   HEAP32[$__parent$i1741>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3855 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1749 = $2;$3865 = $3855;
   label = 1636;
  } else {
   $__nd$0$i$i1743 = $3843;
   while(1) {
    $3844 = ((($__nd$0$i$i1743)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3845 = HEAP32[$3844>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3846 = ($3845|0)>(1073741928); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3846) {
     $3847 = HEAP32[$__nd$0$i$i1743>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3848 = ($3847|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3848) {
      $$lcssa7136 = $__nd$0$i$i1743;$__nd$0$i$i1743$lcssa7134 = $__nd$0$i$i1743;
      label = 1629;
      break;
     } else {
      $$sink$i$i1744 = $3847;
     }
    } else {
     $3850 = ($3845|0)<(1073741928); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3850)) {
      $__nd$0$i$i1743$lcssa = $__nd$0$i$i1743;
      label = 1635;
      break;
     }
     $3851 = ((($__nd$0$i$i1743)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3852 = HEAP32[$3851>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3853 = ($3852|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3853) {
      $$lcssa7135 = $3851;$__nd$0$i$i1743$lcssa7133 = $__nd$0$i$i1743;
      label = 1633;
      break;
     } else {
      $$sink$i$i1744 = $3852;
     }
    }
    $__nd$0$i$i1743 = $$sink$i$i1744;
   }
   if ((label|0) == 1629) {
    HEAP32[$__parent$i1741>>2] = $__nd$0$i$i1743$lcssa7134; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3849 = $__nd$0$i$i1743$lcssa7134; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1749 = $$lcssa7136;$3865 = $3849;
    label = 1636;
    break;
   }
   else if ((label|0) == 1633) {
    HEAP32[$__parent$i1741>>2] = $__nd$0$i$i1743$lcssa7133; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3854 = $__nd$0$i$i1743$lcssa7133; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1749 = $$lcssa7135;$3865 = $3854;
    label = 1636;
    break;
   }
   else if ((label|0) == 1635) {
    HEAP32[$__parent$i1741>>2] = $__nd$0$i$i1743$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3856 = $__nd$0$i$i1743$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3857 = ($__nd$0$i$i1743$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3857) {
     $$0$i7$i1749 = $__parent$i1741;$3865 = $3856;
     label = 1636;
     break;
    } else {
     $3876 = $3843;$__r$0$i1753 = $__nd$0$i$i1743$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1636) {
  __THREW__ = 0;
  $3858 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3859 = __THREW__; __THREW__ = 0;
  $3860 = $3859&1;
  if ($3860) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3861 = ((($3858)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3861>>2] = 1073741928; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3862 = ((($3858)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3862>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3858>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3863 = ((($3858)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3863>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3864 = ((($3858)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3864>>2] = $3865; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1749>>2] = $3858; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3866 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3867 = HEAP32[$3866>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3868 = ($3867|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3868) {
   $3871 = $3858;
  } else {
   $3869 = $3867; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3869; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1751 = HEAP32[$$0$i7$i1749>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3871 = $$pre$i$i1751;
  }
  $3870 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3870,$3871); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3872 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3873 = (($3872) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3873; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4492 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3876 = $$pre4492;$__r$0$i1753 = $3858;
 }
 $3874 = ((($__r$0$i1753)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3874>>2] = 117; //@line 734 "modules/event/sdl/Event.cpp"
 $3875 = ($3876|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3875) {
   HEAP32[$__parent$i1756>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3888 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1764 = $2;$3898 = $3888;
   label = 1650;
  } else {
   $__nd$0$i$i1758 = $3876;
   while(1) {
    $3877 = ((($__nd$0$i$i1758)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3878 = HEAP32[$3877>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3879 = ($3878|0)>(1073741929); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3879) {
     $3880 = HEAP32[$__nd$0$i$i1758>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3881 = ($3880|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3881) {
      $$lcssa7132 = $__nd$0$i$i1758;$__nd$0$i$i1758$lcssa7130 = $__nd$0$i$i1758;
      label = 1643;
      break;
     } else {
      $$sink$i$i1759 = $3880;
     }
    } else {
     $3883 = ($3878|0)<(1073741929); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3883)) {
      $__nd$0$i$i1758$lcssa = $__nd$0$i$i1758;
      label = 1649;
      break;
     }
     $3884 = ((($__nd$0$i$i1758)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3885 = HEAP32[$3884>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3886 = ($3885|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3886) {
      $$lcssa7131 = $3884;$__nd$0$i$i1758$lcssa7129 = $__nd$0$i$i1758;
      label = 1647;
      break;
     } else {
      $$sink$i$i1759 = $3885;
     }
    }
    $__nd$0$i$i1758 = $$sink$i$i1759;
   }
   if ((label|0) == 1643) {
    HEAP32[$__parent$i1756>>2] = $__nd$0$i$i1758$lcssa7130; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3882 = $__nd$0$i$i1758$lcssa7130; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1764 = $$lcssa7132;$3898 = $3882;
    label = 1650;
    break;
   }
   else if ((label|0) == 1647) {
    HEAP32[$__parent$i1756>>2] = $__nd$0$i$i1758$lcssa7129; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3887 = $__nd$0$i$i1758$lcssa7129; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1764 = $$lcssa7131;$3898 = $3887;
    label = 1650;
    break;
   }
   else if ((label|0) == 1649) {
    HEAP32[$__parent$i1756>>2] = $__nd$0$i$i1758$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3889 = $__nd$0$i$i1758$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3890 = ($__nd$0$i$i1758$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3890) {
     $$0$i7$i1764 = $__parent$i1756;$3898 = $3889;
     label = 1650;
     break;
    } else {
     $3909 = $3876;$__r$0$i1768 = $__nd$0$i$i1758$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1650) {
  __THREW__ = 0;
  $3891 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3892 = __THREW__; __THREW__ = 0;
  $3893 = $3892&1;
  if ($3893) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3894 = ((($3891)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3894>>2] = 1073741929; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3895 = ((($3891)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3895>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3891>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3896 = ((($3891)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3896>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3897 = ((($3891)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3897>>2] = $3898; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1764>>2] = $3891; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3899 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3900 = HEAP32[$3899>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3901 = ($3900|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3901) {
   $3904 = $3891;
  } else {
   $3902 = $3900; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3902; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1766 = HEAP32[$$0$i7$i1764>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3904 = $$pre$i$i1766;
  }
  $3903 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3903,$3904); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3905 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3906 = (($3905) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3906; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4493 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3909 = $$pre4493;$__r$0$i1768 = $3891;
 }
 $3907 = ((($__r$0$i1768)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3907>>2] = 118; //@line 735 "modules/event/sdl/Event.cpp"
 $3908 = ($3909|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3908) {
   HEAP32[$__parent$i1771>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3921 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1779 = $2;$3931 = $3921;
   label = 1664;
  } else {
   $__nd$0$i$i1773 = $3909;
   while(1) {
    $3910 = ((($__nd$0$i$i1773)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3911 = HEAP32[$3910>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3912 = ($3911|0)>(1073741930); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3912) {
     $3913 = HEAP32[$__nd$0$i$i1773>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3914 = ($3913|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3914) {
      $$lcssa7128 = $__nd$0$i$i1773;$__nd$0$i$i1773$lcssa7126 = $__nd$0$i$i1773;
      label = 1657;
      break;
     } else {
      $$sink$i$i1774 = $3913;
     }
    } else {
     $3916 = ($3911|0)<(1073741930); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3916)) {
      $__nd$0$i$i1773$lcssa = $__nd$0$i$i1773;
      label = 1663;
      break;
     }
     $3917 = ((($__nd$0$i$i1773)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3918 = HEAP32[$3917>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3919 = ($3918|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3919) {
      $$lcssa7127 = $3917;$__nd$0$i$i1773$lcssa7125 = $__nd$0$i$i1773;
      label = 1661;
      break;
     } else {
      $$sink$i$i1774 = $3918;
     }
    }
    $__nd$0$i$i1773 = $$sink$i$i1774;
   }
   if ((label|0) == 1657) {
    HEAP32[$__parent$i1771>>2] = $__nd$0$i$i1773$lcssa7126; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3915 = $__nd$0$i$i1773$lcssa7126; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1779 = $$lcssa7128;$3931 = $3915;
    label = 1664;
    break;
   }
   else if ((label|0) == 1661) {
    HEAP32[$__parent$i1771>>2] = $__nd$0$i$i1773$lcssa7125; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3920 = $__nd$0$i$i1773$lcssa7125; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1779 = $$lcssa7127;$3931 = $3920;
    label = 1664;
    break;
   }
   else if ((label|0) == 1663) {
    HEAP32[$__parent$i1771>>2] = $__nd$0$i$i1773$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3922 = $__nd$0$i$i1773$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3923 = ($__nd$0$i$i1773$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3923) {
     $$0$i7$i1779 = $__parent$i1771;$3931 = $3922;
     label = 1664;
     break;
    } else {
     $3942 = $3909;$__r$0$i1783 = $__nd$0$i$i1773$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1664) {
  __THREW__ = 0;
  $3924 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3925 = __THREW__; __THREW__ = 0;
  $3926 = $3925&1;
  if ($3926) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3927 = ((($3924)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3927>>2] = 1073741930; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3928 = ((($3924)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3928>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3924>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3929 = ((($3924)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3929>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3930 = ((($3924)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3930>>2] = $3931; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1779>>2] = $3924; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3932 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3933 = HEAP32[$3932>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3934 = ($3933|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3934) {
   $3937 = $3924;
  } else {
   $3935 = $3933; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3935; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1781 = HEAP32[$$0$i7$i1779>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3937 = $$pre$i$i1781;
  }
  $3936 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3936,$3937); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3938 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3939 = (($3938) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3939; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4494 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3942 = $$pre4494;$__r$0$i1783 = $3924;
 }
 $3940 = ((($__r$0$i1783)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3940>>2] = 119; //@line 736 "modules/event/sdl/Event.cpp"
 $3941 = ($3942|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3941) {
   HEAP32[$__parent$i1786>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3954 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1794 = $2;$3964 = $3954;
   label = 1678;
  } else {
   $__nd$0$i$i1788 = $3942;
   while(1) {
    $3943 = ((($__nd$0$i$i1788)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3944 = HEAP32[$3943>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3945 = ($3944|0)>(1073741931); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3945) {
     $3946 = HEAP32[$__nd$0$i$i1788>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3947 = ($3946|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3947) {
      $$lcssa7124 = $__nd$0$i$i1788;$__nd$0$i$i1788$lcssa7122 = $__nd$0$i$i1788;
      label = 1671;
      break;
     } else {
      $$sink$i$i1789 = $3946;
     }
    } else {
     $3949 = ($3944|0)<(1073741931); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3949)) {
      $__nd$0$i$i1788$lcssa = $__nd$0$i$i1788;
      label = 1677;
      break;
     }
     $3950 = ((($__nd$0$i$i1788)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3951 = HEAP32[$3950>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3952 = ($3951|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3952) {
      $$lcssa7123 = $3950;$__nd$0$i$i1788$lcssa7121 = $__nd$0$i$i1788;
      label = 1675;
      break;
     } else {
      $$sink$i$i1789 = $3951;
     }
    }
    $__nd$0$i$i1788 = $$sink$i$i1789;
   }
   if ((label|0) == 1671) {
    HEAP32[$__parent$i1786>>2] = $__nd$0$i$i1788$lcssa7122; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3948 = $__nd$0$i$i1788$lcssa7122; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1794 = $$lcssa7124;$3964 = $3948;
    label = 1678;
    break;
   }
   else if ((label|0) == 1675) {
    HEAP32[$__parent$i1786>>2] = $__nd$0$i$i1788$lcssa7121; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3953 = $__nd$0$i$i1788$lcssa7121; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1794 = $$lcssa7123;$3964 = $3953;
    label = 1678;
    break;
   }
   else if ((label|0) == 1677) {
    HEAP32[$__parent$i1786>>2] = $__nd$0$i$i1788$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3955 = $__nd$0$i$i1788$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3956 = ($__nd$0$i$i1788$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3956) {
     $$0$i7$i1794 = $__parent$i1786;$3964 = $3955;
     label = 1678;
     break;
    } else {
     $3975 = $3942;$__r$0$i1798 = $__nd$0$i$i1788$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1678) {
  __THREW__ = 0;
  $3957 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3958 = __THREW__; __THREW__ = 0;
  $3959 = $3958&1;
  if ($3959) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3960 = ((($3957)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3960>>2] = 1073741931; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3961 = ((($3957)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3961>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3957>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3962 = ((($3957)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3962>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3963 = ((($3957)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3963>>2] = $3964; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1794>>2] = $3957; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3965 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3966 = HEAP32[$3965>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3967 = ($3966|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($3967) {
   $3970 = $3957;
  } else {
   $3968 = $3966; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $3968; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1796 = HEAP32[$$0$i7$i1794>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $3970 = $$pre$i$i1796;
  }
  $3969 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($3969,$3970); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3971 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3972 = (($3971) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $3972; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4495 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3975 = $$pre4495;$__r$0$i1798 = $3957;
 }
 $3973 = ((($__r$0$i1798)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3973>>2] = 120; //@line 737 "modules/event/sdl/Event.cpp"
 $3974 = ($3975|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($3974) {
   HEAP32[$__parent$i1801>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $3987 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1809 = $2;$3997 = $3987;
   label = 1692;
  } else {
   $__nd$0$i$i1803 = $3975;
   while(1) {
    $3976 = ((($__nd$0$i$i1803)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3977 = HEAP32[$3976>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $3978 = ($3977|0)>(1073741932); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($3978) {
     $3979 = HEAP32[$__nd$0$i$i1803>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3980 = ($3979|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3980) {
      $$lcssa7120 = $__nd$0$i$i1803;$__nd$0$i$i1803$lcssa7118 = $__nd$0$i$i1803;
      label = 1685;
      break;
     } else {
      $$sink$i$i1804 = $3979;
     }
    } else {
     $3982 = ($3977|0)<(1073741932); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($3982)) {
      $__nd$0$i$i1803$lcssa = $__nd$0$i$i1803;
      label = 1691;
      break;
     }
     $3983 = ((($__nd$0$i$i1803)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3984 = HEAP32[$3983>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $3985 = ($3984|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($3985) {
      $$lcssa7119 = $3983;$__nd$0$i$i1803$lcssa7117 = $__nd$0$i$i1803;
      label = 1689;
      break;
     } else {
      $$sink$i$i1804 = $3984;
     }
    }
    $__nd$0$i$i1803 = $$sink$i$i1804;
   }
   if ((label|0) == 1685) {
    HEAP32[$__parent$i1801>>2] = $__nd$0$i$i1803$lcssa7118; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3981 = $__nd$0$i$i1803$lcssa7118; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1809 = $$lcssa7120;$3997 = $3981;
    label = 1692;
    break;
   }
   else if ((label|0) == 1689) {
    HEAP32[$__parent$i1801>>2] = $__nd$0$i$i1803$lcssa7117; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3986 = $__nd$0$i$i1803$lcssa7117; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1809 = $$lcssa7119;$3997 = $3986;
    label = 1692;
    break;
   }
   else if ((label|0) == 1691) {
    HEAP32[$__parent$i1801>>2] = $__nd$0$i$i1803$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3988 = $__nd$0$i$i1803$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $3989 = ($__nd$0$i$i1803$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($3989) {
     $$0$i7$i1809 = $__parent$i1801;$3997 = $3988;
     label = 1692;
     break;
    } else {
     $4008 = $3975;$__r$0$i1813 = $__nd$0$i$i1803$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1692) {
  __THREW__ = 0;
  $3990 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $3991 = __THREW__; __THREW__ = 0;
  $3992 = $3991&1;
  if ($3992) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $3993 = ((($3990)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3993>>2] = 1073741932; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $3994 = ((($3990)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$3994>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$3990>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3995 = ((($3990)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3995>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3996 = ((($3990)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$3996>>2] = $3997; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1809>>2] = $3990; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3998 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3999 = HEAP32[$3998>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4000 = ($3999|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4000) {
   $4003 = $3990;
  } else {
   $4001 = $3999; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4001; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1811 = HEAP32[$$0$i7$i1809>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4003 = $$pre$i$i1811;
  }
  $4002 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4002,$4003); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4004 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4005 = (($4004) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4005; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4496 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4008 = $$pre4496;$__r$0$i1813 = $3990;
 }
 $4006 = ((($__r$0$i1813)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4006>>2] = 121; //@line 738 "modules/event/sdl/Event.cpp"
 $4007 = ($4008|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4007) {
   HEAP32[$__parent$i1816>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4020 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1824 = $2;$4030 = $4020;
   label = 1706;
  } else {
   $__nd$0$i$i1818 = $4008;
   while(1) {
    $4009 = ((($__nd$0$i$i1818)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4010 = HEAP32[$4009>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4011 = ($4010|0)>(1073741933); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4011) {
     $4012 = HEAP32[$__nd$0$i$i1818>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4013 = ($4012|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4013) {
      $$lcssa7116 = $__nd$0$i$i1818;$__nd$0$i$i1818$lcssa7114 = $__nd$0$i$i1818;
      label = 1699;
      break;
     } else {
      $$sink$i$i1819 = $4012;
     }
    } else {
     $4015 = ($4010|0)<(1073741933); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4015)) {
      $__nd$0$i$i1818$lcssa = $__nd$0$i$i1818;
      label = 1705;
      break;
     }
     $4016 = ((($__nd$0$i$i1818)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4017 = HEAP32[$4016>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4018 = ($4017|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4018) {
      $$lcssa7115 = $4016;$__nd$0$i$i1818$lcssa7113 = $__nd$0$i$i1818;
      label = 1703;
      break;
     } else {
      $$sink$i$i1819 = $4017;
     }
    }
    $__nd$0$i$i1818 = $$sink$i$i1819;
   }
   if ((label|0) == 1699) {
    HEAP32[$__parent$i1816>>2] = $__nd$0$i$i1818$lcssa7114; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4014 = $__nd$0$i$i1818$lcssa7114; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1824 = $$lcssa7116;$4030 = $4014;
    label = 1706;
    break;
   }
   else if ((label|0) == 1703) {
    HEAP32[$__parent$i1816>>2] = $__nd$0$i$i1818$lcssa7113; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4019 = $__nd$0$i$i1818$lcssa7113; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1824 = $$lcssa7115;$4030 = $4019;
    label = 1706;
    break;
   }
   else if ((label|0) == 1705) {
    HEAP32[$__parent$i1816>>2] = $__nd$0$i$i1818$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4021 = $__nd$0$i$i1818$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4022 = ($__nd$0$i$i1818$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4022) {
     $$0$i7$i1824 = $__parent$i1816;$4030 = $4021;
     label = 1706;
     break;
    } else {
     $4041 = $4008;$__r$0$i1828 = $__nd$0$i$i1818$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1706) {
  __THREW__ = 0;
  $4023 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4024 = __THREW__; __THREW__ = 0;
  $4025 = $4024&1;
  if ($4025) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4026 = ((($4023)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4026>>2] = 1073741933; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4027 = ((($4023)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4027>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4023>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4028 = ((($4023)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4028>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4029 = ((($4023)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4029>>2] = $4030; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1824>>2] = $4023; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4031 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4032 = HEAP32[$4031>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4033 = ($4032|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4033) {
   $4036 = $4023;
  } else {
   $4034 = $4032; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4034; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1826 = HEAP32[$$0$i7$i1824>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4036 = $$pre$i$i1826;
  }
  $4035 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4035,$4036); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4037 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4038 = (($4037) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4038; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4497 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4041 = $$pre4497;$__r$0$i1828 = $4023;
 }
 $4039 = ((($__r$0$i1828)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4039>>2] = 122; //@line 739 "modules/event/sdl/Event.cpp"
 $4040 = ($4041|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4040) {
   HEAP32[$__parent$i1831>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4053 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1839 = $2;$4063 = $4053;
   label = 1720;
  } else {
   $__nd$0$i$i1833 = $4041;
   while(1) {
    $4042 = ((($__nd$0$i$i1833)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4043 = HEAP32[$4042>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4044 = ($4043|0)>(1073741934); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4044) {
     $4045 = HEAP32[$__nd$0$i$i1833>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4046 = ($4045|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4046) {
      $$lcssa7112 = $__nd$0$i$i1833;$__nd$0$i$i1833$lcssa7110 = $__nd$0$i$i1833;
      label = 1713;
      break;
     } else {
      $$sink$i$i1834 = $4045;
     }
    } else {
     $4048 = ($4043|0)<(1073741934); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4048)) {
      $__nd$0$i$i1833$lcssa = $__nd$0$i$i1833;
      label = 1719;
      break;
     }
     $4049 = ((($__nd$0$i$i1833)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4050 = HEAP32[$4049>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4051 = ($4050|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4051) {
      $$lcssa7111 = $4049;$__nd$0$i$i1833$lcssa7109 = $__nd$0$i$i1833;
      label = 1717;
      break;
     } else {
      $$sink$i$i1834 = $4050;
     }
    }
    $__nd$0$i$i1833 = $$sink$i$i1834;
   }
   if ((label|0) == 1713) {
    HEAP32[$__parent$i1831>>2] = $__nd$0$i$i1833$lcssa7110; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4047 = $__nd$0$i$i1833$lcssa7110; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1839 = $$lcssa7112;$4063 = $4047;
    label = 1720;
    break;
   }
   else if ((label|0) == 1717) {
    HEAP32[$__parent$i1831>>2] = $__nd$0$i$i1833$lcssa7109; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4052 = $__nd$0$i$i1833$lcssa7109; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1839 = $$lcssa7111;$4063 = $4052;
    label = 1720;
    break;
   }
   else if ((label|0) == 1719) {
    HEAP32[$__parent$i1831>>2] = $__nd$0$i$i1833$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4054 = $__nd$0$i$i1833$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4055 = ($__nd$0$i$i1833$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4055) {
     $$0$i7$i1839 = $__parent$i1831;$4063 = $4054;
     label = 1720;
     break;
    } else {
     $4074 = $4041;$__r$0$i1843 = $__nd$0$i$i1833$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1720) {
  __THREW__ = 0;
  $4056 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4057 = __THREW__; __THREW__ = 0;
  $4058 = $4057&1;
  if ($4058) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4059 = ((($4056)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4059>>2] = 1073741934; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4060 = ((($4056)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4060>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4056>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4061 = ((($4056)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4061>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4062 = ((($4056)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4062>>2] = $4063; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1839>>2] = $4056; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4064 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4065 = HEAP32[$4064>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4066 = ($4065|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4066) {
   $4069 = $4056;
  } else {
   $4067 = $4065; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4067; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1841 = HEAP32[$$0$i7$i1839>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4069 = $$pre$i$i1841;
  }
  $4068 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4068,$4069); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4070 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4071 = (($4070) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4071; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4498 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4074 = $$pre4498;$__r$0$i1843 = $4056;
 }
 $4072 = ((($__r$0$i1843)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4072>>2] = 123; //@line 740 "modules/event/sdl/Event.cpp"
 $4073 = ($4074|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4073) {
   HEAP32[$__parent$i1846>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4086 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1854 = $2;$4096 = $4086;
   label = 1734;
  } else {
   $__nd$0$i$i1848 = $4074;
   while(1) {
    $4075 = ((($__nd$0$i$i1848)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4076 = HEAP32[$4075>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4077 = ($4076|0)>(1073741935); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4077) {
     $4078 = HEAP32[$__nd$0$i$i1848>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4079 = ($4078|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4079) {
      $$lcssa7108 = $__nd$0$i$i1848;$__nd$0$i$i1848$lcssa7106 = $__nd$0$i$i1848;
      label = 1727;
      break;
     } else {
      $$sink$i$i1849 = $4078;
     }
    } else {
     $4081 = ($4076|0)<(1073741935); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4081)) {
      $__nd$0$i$i1848$lcssa = $__nd$0$i$i1848;
      label = 1733;
      break;
     }
     $4082 = ((($__nd$0$i$i1848)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4083 = HEAP32[$4082>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4084 = ($4083|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4084) {
      $$lcssa7107 = $4082;$__nd$0$i$i1848$lcssa7105 = $__nd$0$i$i1848;
      label = 1731;
      break;
     } else {
      $$sink$i$i1849 = $4083;
     }
    }
    $__nd$0$i$i1848 = $$sink$i$i1849;
   }
   if ((label|0) == 1727) {
    HEAP32[$__parent$i1846>>2] = $__nd$0$i$i1848$lcssa7106; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4080 = $__nd$0$i$i1848$lcssa7106; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1854 = $$lcssa7108;$4096 = $4080;
    label = 1734;
    break;
   }
   else if ((label|0) == 1731) {
    HEAP32[$__parent$i1846>>2] = $__nd$0$i$i1848$lcssa7105; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4085 = $__nd$0$i$i1848$lcssa7105; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1854 = $$lcssa7107;$4096 = $4085;
    label = 1734;
    break;
   }
   else if ((label|0) == 1733) {
    HEAP32[$__parent$i1846>>2] = $__nd$0$i$i1848$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4087 = $__nd$0$i$i1848$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4088 = ($__nd$0$i$i1848$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4088) {
     $$0$i7$i1854 = $__parent$i1846;$4096 = $4087;
     label = 1734;
     break;
    } else {
     $4107 = $4074;$__r$0$i1858 = $__nd$0$i$i1848$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1734) {
  __THREW__ = 0;
  $4089 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4090 = __THREW__; __THREW__ = 0;
  $4091 = $4090&1;
  if ($4091) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4092 = ((($4089)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4092>>2] = 1073741935; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4093 = ((($4089)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4093>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4089>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4094 = ((($4089)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4094>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4095 = ((($4089)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4095>>2] = $4096; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1854>>2] = $4089; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4097 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4098 = HEAP32[$4097>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4099 = ($4098|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4099) {
   $4102 = $4089;
  } else {
   $4100 = $4098; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4100; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1856 = HEAP32[$$0$i7$i1854>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4102 = $$pre$i$i1856;
  }
  $4101 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4101,$4102); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4103 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4104 = (($4103) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4104; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4499 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4107 = $$pre4499;$__r$0$i1858 = $4089;
 }
 $4105 = ((($__r$0$i1858)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4105>>2] = 124; //@line 741 "modules/event/sdl/Event.cpp"
 $4106 = ($4107|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4106) {
   HEAP32[$__parent$i1861>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4119 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1869 = $2;$4129 = $4119;
   label = 1748;
  } else {
   $__nd$0$i$i1863 = $4107;
   while(1) {
    $4108 = ((($__nd$0$i$i1863)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4109 = HEAP32[$4108>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4110 = ($4109|0)>(1073741936); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4110) {
     $4111 = HEAP32[$__nd$0$i$i1863>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4112 = ($4111|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4112) {
      $$lcssa7104 = $__nd$0$i$i1863;$__nd$0$i$i1863$lcssa7102 = $__nd$0$i$i1863;
      label = 1741;
      break;
     } else {
      $$sink$i$i1864 = $4111;
     }
    } else {
     $4114 = ($4109|0)<(1073741936); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4114)) {
      $__nd$0$i$i1863$lcssa = $__nd$0$i$i1863;
      label = 1747;
      break;
     }
     $4115 = ((($__nd$0$i$i1863)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4116 = HEAP32[$4115>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4117 = ($4116|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4117) {
      $$lcssa7103 = $4115;$__nd$0$i$i1863$lcssa7101 = $__nd$0$i$i1863;
      label = 1745;
      break;
     } else {
      $$sink$i$i1864 = $4116;
     }
    }
    $__nd$0$i$i1863 = $$sink$i$i1864;
   }
   if ((label|0) == 1741) {
    HEAP32[$__parent$i1861>>2] = $__nd$0$i$i1863$lcssa7102; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4113 = $__nd$0$i$i1863$lcssa7102; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1869 = $$lcssa7104;$4129 = $4113;
    label = 1748;
    break;
   }
   else if ((label|0) == 1745) {
    HEAP32[$__parent$i1861>>2] = $__nd$0$i$i1863$lcssa7101; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4118 = $__nd$0$i$i1863$lcssa7101; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1869 = $$lcssa7103;$4129 = $4118;
    label = 1748;
    break;
   }
   else if ((label|0) == 1747) {
    HEAP32[$__parent$i1861>>2] = $__nd$0$i$i1863$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4120 = $__nd$0$i$i1863$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4121 = ($__nd$0$i$i1863$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4121) {
     $$0$i7$i1869 = $__parent$i1861;$4129 = $4120;
     label = 1748;
     break;
    } else {
     $4140 = $4107;$__r$0$i1873 = $__nd$0$i$i1863$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1748) {
  __THREW__ = 0;
  $4122 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4123 = __THREW__; __THREW__ = 0;
  $4124 = $4123&1;
  if ($4124) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4125 = ((($4122)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4125>>2] = 1073741936; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4126 = ((($4122)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4126>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4122>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4127 = ((($4122)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4127>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4128 = ((($4122)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4128>>2] = $4129; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1869>>2] = $4122; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4130 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4131 = HEAP32[$4130>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4132 = ($4131|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4132) {
   $4135 = $4122;
  } else {
   $4133 = $4131; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4133; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1871 = HEAP32[$$0$i7$i1869>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4135 = $$pre$i$i1871;
  }
  $4134 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4134,$4135); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4136 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4137 = (($4136) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4137; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4500 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4140 = $$pre4500;$__r$0$i1873 = $4122;
 }
 $4138 = ((($__r$0$i1873)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4138>>2] = 125; //@line 742 "modules/event/sdl/Event.cpp"
 $4139 = ($4140|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4139) {
   HEAP32[$__parent$i1876>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4152 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1884 = $2;$4162 = $4152;
   label = 1762;
  } else {
   $__nd$0$i$i1878 = $4140;
   while(1) {
    $4141 = ((($__nd$0$i$i1878)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4142 = HEAP32[$4141>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4143 = ($4142|0)>(1073741937); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4143) {
     $4144 = HEAP32[$__nd$0$i$i1878>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4145 = ($4144|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4145) {
      $$lcssa7100 = $__nd$0$i$i1878;$__nd$0$i$i1878$lcssa7098 = $__nd$0$i$i1878;
      label = 1755;
      break;
     } else {
      $$sink$i$i1879 = $4144;
     }
    } else {
     $4147 = ($4142|0)<(1073741937); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4147)) {
      $__nd$0$i$i1878$lcssa = $__nd$0$i$i1878;
      label = 1761;
      break;
     }
     $4148 = ((($__nd$0$i$i1878)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4149 = HEAP32[$4148>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4150 = ($4149|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4150) {
      $$lcssa7099 = $4148;$__nd$0$i$i1878$lcssa7097 = $__nd$0$i$i1878;
      label = 1759;
      break;
     } else {
      $$sink$i$i1879 = $4149;
     }
    }
    $__nd$0$i$i1878 = $$sink$i$i1879;
   }
   if ((label|0) == 1755) {
    HEAP32[$__parent$i1876>>2] = $__nd$0$i$i1878$lcssa7098; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4146 = $__nd$0$i$i1878$lcssa7098; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1884 = $$lcssa7100;$4162 = $4146;
    label = 1762;
    break;
   }
   else if ((label|0) == 1759) {
    HEAP32[$__parent$i1876>>2] = $__nd$0$i$i1878$lcssa7097; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4151 = $__nd$0$i$i1878$lcssa7097; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1884 = $$lcssa7099;$4162 = $4151;
    label = 1762;
    break;
   }
   else if ((label|0) == 1761) {
    HEAP32[$__parent$i1876>>2] = $__nd$0$i$i1878$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4153 = $__nd$0$i$i1878$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4154 = ($__nd$0$i$i1878$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4154) {
     $$0$i7$i1884 = $__parent$i1876;$4162 = $4153;
     label = 1762;
     break;
    } else {
     $4173 = $4140;$__r$0$i1888 = $__nd$0$i$i1878$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1762) {
  __THREW__ = 0;
  $4155 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4156 = __THREW__; __THREW__ = 0;
  $4157 = $4156&1;
  if ($4157) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4158 = ((($4155)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4158>>2] = 1073741937; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4159 = ((($4155)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4159>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4155>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4160 = ((($4155)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4160>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4161 = ((($4155)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4161>>2] = $4162; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1884>>2] = $4155; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4163 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4164 = HEAP32[$4163>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4165 = ($4164|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4165) {
   $4168 = $4155;
  } else {
   $4166 = $4164; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4166; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1886 = HEAP32[$$0$i7$i1884>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4168 = $$pre$i$i1886;
  }
  $4167 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4167,$4168); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4169 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4170 = (($4169) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4170; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4501 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4173 = $$pre4501;$__r$0$i1888 = $4155;
 }
 $4171 = ((($__r$0$i1888)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4171>>2] = 126; //@line 743 "modules/event/sdl/Event.cpp"
 $4172 = ($4173|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4172) {
   HEAP32[$__parent$i1891>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4185 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1899 = $2;$4195 = $4185;
   label = 1776;
  } else {
   $__nd$0$i$i1893 = $4173;
   while(1) {
    $4174 = ((($__nd$0$i$i1893)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4175 = HEAP32[$4174>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4176 = ($4175|0)>(1073741938); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4176) {
     $4177 = HEAP32[$__nd$0$i$i1893>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4178 = ($4177|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4178) {
      $$lcssa7096 = $__nd$0$i$i1893;$__nd$0$i$i1893$lcssa7094 = $__nd$0$i$i1893;
      label = 1769;
      break;
     } else {
      $$sink$i$i1894 = $4177;
     }
    } else {
     $4180 = ($4175|0)<(1073741938); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4180)) {
      $__nd$0$i$i1893$lcssa = $__nd$0$i$i1893;
      label = 1775;
      break;
     }
     $4181 = ((($__nd$0$i$i1893)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4182 = HEAP32[$4181>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4183 = ($4182|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4183) {
      $$lcssa7095 = $4181;$__nd$0$i$i1893$lcssa7093 = $__nd$0$i$i1893;
      label = 1773;
      break;
     } else {
      $$sink$i$i1894 = $4182;
     }
    }
    $__nd$0$i$i1893 = $$sink$i$i1894;
   }
   if ((label|0) == 1769) {
    HEAP32[$__parent$i1891>>2] = $__nd$0$i$i1893$lcssa7094; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4179 = $__nd$0$i$i1893$lcssa7094; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1899 = $$lcssa7096;$4195 = $4179;
    label = 1776;
    break;
   }
   else if ((label|0) == 1773) {
    HEAP32[$__parent$i1891>>2] = $__nd$0$i$i1893$lcssa7093; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4184 = $__nd$0$i$i1893$lcssa7093; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1899 = $$lcssa7095;$4195 = $4184;
    label = 1776;
    break;
   }
   else if ((label|0) == 1775) {
    HEAP32[$__parent$i1891>>2] = $__nd$0$i$i1893$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4186 = $__nd$0$i$i1893$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4187 = ($__nd$0$i$i1893$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4187) {
     $$0$i7$i1899 = $__parent$i1891;$4195 = $4186;
     label = 1776;
     break;
    } else {
     $4206 = $4173;$__r$0$i1903 = $__nd$0$i$i1893$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1776) {
  __THREW__ = 0;
  $4188 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4189 = __THREW__; __THREW__ = 0;
  $4190 = $4189&1;
  if ($4190) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4191 = ((($4188)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4191>>2] = 1073741938; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4192 = ((($4188)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4192>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4188>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4193 = ((($4188)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4193>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4194 = ((($4188)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4194>>2] = $4195; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1899>>2] = $4188; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4196 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4197 = HEAP32[$4196>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4198 = ($4197|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4198) {
   $4201 = $4188;
  } else {
   $4199 = $4197; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4199; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1901 = HEAP32[$$0$i7$i1899>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4201 = $$pre$i$i1901;
  }
  $4200 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4200,$4201); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4202 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4203 = (($4202) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4203; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4502 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4206 = $$pre4502;$__r$0$i1903 = $4188;
 }
 $4204 = ((($__r$0$i1903)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4204>>2] = 127; //@line 744 "modules/event/sdl/Event.cpp"
 $4205 = ($4206|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4205) {
   HEAP32[$__parent$i1906>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4218 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1914 = $2;$4228 = $4218;
   label = 1790;
  } else {
   $__nd$0$i$i1908 = $4206;
   while(1) {
    $4207 = ((($__nd$0$i$i1908)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4208 = HEAP32[$4207>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4209 = ($4208|0)>(1073741939); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4209) {
     $4210 = HEAP32[$__nd$0$i$i1908>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4211 = ($4210|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4211) {
      $$lcssa7092 = $__nd$0$i$i1908;$__nd$0$i$i1908$lcssa7090 = $__nd$0$i$i1908;
      label = 1783;
      break;
     } else {
      $$sink$i$i1909 = $4210;
     }
    } else {
     $4213 = ($4208|0)<(1073741939); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4213)) {
      $__nd$0$i$i1908$lcssa = $__nd$0$i$i1908;
      label = 1789;
      break;
     }
     $4214 = ((($__nd$0$i$i1908)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4215 = HEAP32[$4214>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4216 = ($4215|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4216) {
      $$lcssa7091 = $4214;$__nd$0$i$i1908$lcssa7089 = $__nd$0$i$i1908;
      label = 1787;
      break;
     } else {
      $$sink$i$i1909 = $4215;
     }
    }
    $__nd$0$i$i1908 = $$sink$i$i1909;
   }
   if ((label|0) == 1783) {
    HEAP32[$__parent$i1906>>2] = $__nd$0$i$i1908$lcssa7090; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4212 = $__nd$0$i$i1908$lcssa7090; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1914 = $$lcssa7092;$4228 = $4212;
    label = 1790;
    break;
   }
   else if ((label|0) == 1787) {
    HEAP32[$__parent$i1906>>2] = $__nd$0$i$i1908$lcssa7089; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4217 = $__nd$0$i$i1908$lcssa7089; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1914 = $$lcssa7091;$4228 = $4217;
    label = 1790;
    break;
   }
   else if ((label|0) == 1789) {
    HEAP32[$__parent$i1906>>2] = $__nd$0$i$i1908$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4219 = $__nd$0$i$i1908$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4220 = ($__nd$0$i$i1908$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4220) {
     $$0$i7$i1914 = $__parent$i1906;$4228 = $4219;
     label = 1790;
     break;
    } else {
     $4239 = $4206;$__r$0$i1918 = $__nd$0$i$i1908$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1790) {
  __THREW__ = 0;
  $4221 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4222 = __THREW__; __THREW__ = 0;
  $4223 = $4222&1;
  if ($4223) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4224 = ((($4221)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4224>>2] = 1073741939; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4225 = ((($4221)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4225>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4221>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4226 = ((($4221)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4226>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4227 = ((($4221)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4227>>2] = $4228; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1914>>2] = $4221; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4229 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4230 = HEAP32[$4229>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4231 = ($4230|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4231) {
   $4234 = $4221;
  } else {
   $4232 = $4230; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4232; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1916 = HEAP32[$$0$i7$i1914>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4234 = $$pre$i$i1916;
  }
  $4233 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4233,$4234); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4235 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4236 = (($4235) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4236; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4503 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4239 = $$pre4503;$__r$0$i1918 = $4221;
 }
 $4237 = ((($__r$0$i1918)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4237>>2] = 128; //@line 745 "modules/event/sdl/Event.cpp"
 $4238 = ($4239|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4238) {
   HEAP32[$__parent$i1921>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4251 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1929 = $2;$4261 = $4251;
   label = 1804;
  } else {
   $__nd$0$i$i1923 = $4239;
   while(1) {
    $4240 = ((($__nd$0$i$i1923)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4241 = HEAP32[$4240>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4242 = ($4241|0)>(1073741940); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4242) {
     $4243 = HEAP32[$__nd$0$i$i1923>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4244 = ($4243|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4244) {
      $$lcssa7088 = $__nd$0$i$i1923;$__nd$0$i$i1923$lcssa7086 = $__nd$0$i$i1923;
      label = 1797;
      break;
     } else {
      $$sink$i$i1924 = $4243;
     }
    } else {
     $4246 = ($4241|0)<(1073741940); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4246)) {
      $__nd$0$i$i1923$lcssa = $__nd$0$i$i1923;
      label = 1803;
      break;
     }
     $4247 = ((($__nd$0$i$i1923)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4248 = HEAP32[$4247>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4249 = ($4248|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4249) {
      $$lcssa7087 = $4247;$__nd$0$i$i1923$lcssa7085 = $__nd$0$i$i1923;
      label = 1801;
      break;
     } else {
      $$sink$i$i1924 = $4248;
     }
    }
    $__nd$0$i$i1923 = $$sink$i$i1924;
   }
   if ((label|0) == 1797) {
    HEAP32[$__parent$i1921>>2] = $__nd$0$i$i1923$lcssa7086; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4245 = $__nd$0$i$i1923$lcssa7086; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1929 = $$lcssa7088;$4261 = $4245;
    label = 1804;
    break;
   }
   else if ((label|0) == 1801) {
    HEAP32[$__parent$i1921>>2] = $__nd$0$i$i1923$lcssa7085; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4250 = $__nd$0$i$i1923$lcssa7085; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1929 = $$lcssa7087;$4261 = $4250;
    label = 1804;
    break;
   }
   else if ((label|0) == 1803) {
    HEAP32[$__parent$i1921>>2] = $__nd$0$i$i1923$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4252 = $__nd$0$i$i1923$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4253 = ($__nd$0$i$i1923$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4253) {
     $$0$i7$i1929 = $__parent$i1921;$4261 = $4252;
     label = 1804;
     break;
    } else {
     $4272 = $4239;$__r$0$i1933 = $__nd$0$i$i1923$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1804) {
  __THREW__ = 0;
  $4254 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4255 = __THREW__; __THREW__ = 0;
  $4256 = $4255&1;
  if ($4256) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4257 = ((($4254)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4257>>2] = 1073741940; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4258 = ((($4254)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4258>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4254>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4259 = ((($4254)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4259>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4260 = ((($4254)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4260>>2] = $4261; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1929>>2] = $4254; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4262 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4263 = HEAP32[$4262>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4264 = ($4263|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4264) {
   $4267 = $4254;
  } else {
   $4265 = $4263; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4265; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1931 = HEAP32[$$0$i7$i1929>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4267 = $$pre$i$i1931;
  }
  $4266 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4266,$4267); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4268 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4269 = (($4268) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4269; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4504 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4272 = $$pre4504;$__r$0$i1933 = $4254;
 }
 $4270 = ((($__r$0$i1933)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4270>>2] = 129; //@line 746 "modules/event/sdl/Event.cpp"
 $4271 = ($4272|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4271) {
   HEAP32[$__parent$i1936>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4284 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1944 = $2;$4294 = $4284;
   label = 1818;
  } else {
   $__nd$0$i$i1938 = $4272;
   while(1) {
    $4273 = ((($__nd$0$i$i1938)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4274 = HEAP32[$4273>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4275 = ($4274|0)>(1073741941); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4275) {
     $4276 = HEAP32[$__nd$0$i$i1938>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4277 = ($4276|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4277) {
      $$lcssa7084 = $__nd$0$i$i1938;$__nd$0$i$i1938$lcssa7082 = $__nd$0$i$i1938;
      label = 1811;
      break;
     } else {
      $$sink$i$i1939 = $4276;
     }
    } else {
     $4279 = ($4274|0)<(1073741941); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4279)) {
      $__nd$0$i$i1938$lcssa = $__nd$0$i$i1938;
      label = 1817;
      break;
     }
     $4280 = ((($__nd$0$i$i1938)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4281 = HEAP32[$4280>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4282 = ($4281|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4282) {
      $$lcssa7083 = $4280;$__nd$0$i$i1938$lcssa7081 = $__nd$0$i$i1938;
      label = 1815;
      break;
     } else {
      $$sink$i$i1939 = $4281;
     }
    }
    $__nd$0$i$i1938 = $$sink$i$i1939;
   }
   if ((label|0) == 1811) {
    HEAP32[$__parent$i1936>>2] = $__nd$0$i$i1938$lcssa7082; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4278 = $__nd$0$i$i1938$lcssa7082; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1944 = $$lcssa7084;$4294 = $4278;
    label = 1818;
    break;
   }
   else if ((label|0) == 1815) {
    HEAP32[$__parent$i1936>>2] = $__nd$0$i$i1938$lcssa7081; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4283 = $__nd$0$i$i1938$lcssa7081; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1944 = $$lcssa7083;$4294 = $4283;
    label = 1818;
    break;
   }
   else if ((label|0) == 1817) {
    HEAP32[$__parent$i1936>>2] = $__nd$0$i$i1938$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4285 = $__nd$0$i$i1938$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4286 = ($__nd$0$i$i1938$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4286) {
     $$0$i7$i1944 = $__parent$i1936;$4294 = $4285;
     label = 1818;
     break;
    } else {
     $4305 = $4272;$__r$0$i1948 = $__nd$0$i$i1938$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1818) {
  __THREW__ = 0;
  $4287 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4288 = __THREW__; __THREW__ = 0;
  $4289 = $4288&1;
  if ($4289) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4290 = ((($4287)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4290>>2] = 1073741941; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4291 = ((($4287)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4291>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4287>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4292 = ((($4287)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4292>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4293 = ((($4287)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4293>>2] = $4294; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1944>>2] = $4287; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4295 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4296 = HEAP32[$4295>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4297 = ($4296|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4297) {
   $4300 = $4287;
  } else {
   $4298 = $4296; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4298; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1946 = HEAP32[$$0$i7$i1944>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4300 = $$pre$i$i1946;
  }
  $4299 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4299,$4300); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4301 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4302 = (($4301) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4302; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4505 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4305 = $$pre4505;$__r$0$i1948 = $4287;
 }
 $4303 = ((($__r$0$i1948)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4303>>2] = 130; //@line 747 "modules/event/sdl/Event.cpp"
 $4304 = ($4305|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4304) {
   HEAP32[$__parent$i1951>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4317 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1959 = $2;$4327 = $4317;
   label = 1832;
  } else {
   $__nd$0$i$i1953 = $4305;
   while(1) {
    $4306 = ((($__nd$0$i$i1953)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4307 = HEAP32[$4306>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4308 = ($4307|0)>(1073741942); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4308) {
     $4309 = HEAP32[$__nd$0$i$i1953>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4310 = ($4309|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4310) {
      $$lcssa7080 = $__nd$0$i$i1953;$__nd$0$i$i1953$lcssa7078 = $__nd$0$i$i1953;
      label = 1825;
      break;
     } else {
      $$sink$i$i1954 = $4309;
     }
    } else {
     $4312 = ($4307|0)<(1073741942); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4312)) {
      $__nd$0$i$i1953$lcssa = $__nd$0$i$i1953;
      label = 1831;
      break;
     }
     $4313 = ((($__nd$0$i$i1953)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4314 = HEAP32[$4313>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4315 = ($4314|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4315) {
      $$lcssa7079 = $4313;$__nd$0$i$i1953$lcssa7077 = $__nd$0$i$i1953;
      label = 1829;
      break;
     } else {
      $$sink$i$i1954 = $4314;
     }
    }
    $__nd$0$i$i1953 = $$sink$i$i1954;
   }
   if ((label|0) == 1825) {
    HEAP32[$__parent$i1951>>2] = $__nd$0$i$i1953$lcssa7078; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4311 = $__nd$0$i$i1953$lcssa7078; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1959 = $$lcssa7080;$4327 = $4311;
    label = 1832;
    break;
   }
   else if ((label|0) == 1829) {
    HEAP32[$__parent$i1951>>2] = $__nd$0$i$i1953$lcssa7077; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4316 = $__nd$0$i$i1953$lcssa7077; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1959 = $$lcssa7079;$4327 = $4316;
    label = 1832;
    break;
   }
   else if ((label|0) == 1831) {
    HEAP32[$__parent$i1951>>2] = $__nd$0$i$i1953$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4318 = $__nd$0$i$i1953$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4319 = ($__nd$0$i$i1953$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4319) {
     $$0$i7$i1959 = $__parent$i1951;$4327 = $4318;
     label = 1832;
     break;
    } else {
     $4338 = $4305;$__r$0$i1963 = $__nd$0$i$i1953$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1832) {
  __THREW__ = 0;
  $4320 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4321 = __THREW__; __THREW__ = 0;
  $4322 = $4321&1;
  if ($4322) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4323 = ((($4320)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4323>>2] = 1073741942; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4324 = ((($4320)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4324>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4320>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4325 = ((($4320)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4325>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4326 = ((($4320)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4326>>2] = $4327; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1959>>2] = $4320; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4328 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4329 = HEAP32[$4328>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4330 = ($4329|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4330) {
   $4333 = $4320;
  } else {
   $4331 = $4329; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4331; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1961 = HEAP32[$$0$i7$i1959>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4333 = $$pre$i$i1961;
  }
  $4332 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4332,$4333); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4334 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4335 = (($4334) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4335; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4506 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4338 = $$pre4506;$__r$0$i1963 = $4320;
 }
 $4336 = ((($__r$0$i1963)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4336>>2] = 131; //@line 748 "modules/event/sdl/Event.cpp"
 $4337 = ($4338|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4337) {
   HEAP32[$__parent$i1966>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4350 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1974 = $2;$4360 = $4350;
   label = 1846;
  } else {
   $__nd$0$i$i1968 = $4338;
   while(1) {
    $4339 = ((($__nd$0$i$i1968)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4340 = HEAP32[$4339>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4341 = ($4340|0)>(1073741943); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4341) {
     $4342 = HEAP32[$__nd$0$i$i1968>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4343 = ($4342|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4343) {
      $$lcssa7076 = $__nd$0$i$i1968;$__nd$0$i$i1968$lcssa7074 = $__nd$0$i$i1968;
      label = 1839;
      break;
     } else {
      $$sink$i$i1969 = $4342;
     }
    } else {
     $4345 = ($4340|0)<(1073741943); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4345)) {
      $__nd$0$i$i1968$lcssa = $__nd$0$i$i1968;
      label = 1845;
      break;
     }
     $4346 = ((($__nd$0$i$i1968)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4347 = HEAP32[$4346>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4348 = ($4347|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4348) {
      $$lcssa7075 = $4346;$__nd$0$i$i1968$lcssa7073 = $__nd$0$i$i1968;
      label = 1843;
      break;
     } else {
      $$sink$i$i1969 = $4347;
     }
    }
    $__nd$0$i$i1968 = $$sink$i$i1969;
   }
   if ((label|0) == 1839) {
    HEAP32[$__parent$i1966>>2] = $__nd$0$i$i1968$lcssa7074; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4344 = $__nd$0$i$i1968$lcssa7074; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1974 = $$lcssa7076;$4360 = $4344;
    label = 1846;
    break;
   }
   else if ((label|0) == 1843) {
    HEAP32[$__parent$i1966>>2] = $__nd$0$i$i1968$lcssa7073; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4349 = $__nd$0$i$i1968$lcssa7073; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1974 = $$lcssa7075;$4360 = $4349;
    label = 1846;
    break;
   }
   else if ((label|0) == 1845) {
    HEAP32[$__parent$i1966>>2] = $__nd$0$i$i1968$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4351 = $__nd$0$i$i1968$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4352 = ($__nd$0$i$i1968$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4352) {
     $$0$i7$i1974 = $__parent$i1966;$4360 = $4351;
     label = 1846;
     break;
    } else {
     $4371 = $4338;$__r$0$i1978 = $__nd$0$i$i1968$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1846) {
  __THREW__ = 0;
  $4353 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4354 = __THREW__; __THREW__ = 0;
  $4355 = $4354&1;
  if ($4355) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4356 = ((($4353)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4356>>2] = 1073741943; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4357 = ((($4353)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4357>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4353>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4358 = ((($4353)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4358>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4359 = ((($4353)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4359>>2] = $4360; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1974>>2] = $4353; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4361 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4362 = HEAP32[$4361>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4363 = ($4362|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4363) {
   $4366 = $4353;
  } else {
   $4364 = $4362; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4364; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1976 = HEAP32[$$0$i7$i1974>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4366 = $$pre$i$i1976;
  }
  $4365 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4365,$4366); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4367 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4368 = (($4367) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4368; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4507 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4371 = $$pre4507;$__r$0$i1978 = $4353;
 }
 $4369 = ((($__r$0$i1978)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4369>>2] = 132; //@line 749 "modules/event/sdl/Event.cpp"
 $4370 = ($4371|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4370) {
   HEAP32[$__parent$i1981>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4383 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i1989 = $2;$4393 = $4383;
   label = 1860;
  } else {
   $__nd$0$i$i1983 = $4371;
   while(1) {
    $4372 = ((($__nd$0$i$i1983)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4373 = HEAP32[$4372>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4374 = ($4373|0)>(1073741944); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4374) {
     $4375 = HEAP32[$__nd$0$i$i1983>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4376 = ($4375|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4376) {
      $$lcssa7072 = $__nd$0$i$i1983;$__nd$0$i$i1983$lcssa7070 = $__nd$0$i$i1983;
      label = 1853;
      break;
     } else {
      $$sink$i$i1984 = $4375;
     }
    } else {
     $4378 = ($4373|0)<(1073741944); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4378)) {
      $__nd$0$i$i1983$lcssa = $__nd$0$i$i1983;
      label = 1859;
      break;
     }
     $4379 = ((($__nd$0$i$i1983)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4380 = HEAP32[$4379>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4381 = ($4380|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4381) {
      $$lcssa7071 = $4379;$__nd$0$i$i1983$lcssa7069 = $__nd$0$i$i1983;
      label = 1857;
      break;
     } else {
      $$sink$i$i1984 = $4380;
     }
    }
    $__nd$0$i$i1983 = $$sink$i$i1984;
   }
   if ((label|0) == 1853) {
    HEAP32[$__parent$i1981>>2] = $__nd$0$i$i1983$lcssa7070; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4377 = $__nd$0$i$i1983$lcssa7070; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1989 = $$lcssa7072;$4393 = $4377;
    label = 1860;
    break;
   }
   else if ((label|0) == 1857) {
    HEAP32[$__parent$i1981>>2] = $__nd$0$i$i1983$lcssa7069; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4382 = $__nd$0$i$i1983$lcssa7069; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i1989 = $$lcssa7071;$4393 = $4382;
    label = 1860;
    break;
   }
   else if ((label|0) == 1859) {
    HEAP32[$__parent$i1981>>2] = $__nd$0$i$i1983$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4384 = $__nd$0$i$i1983$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4385 = ($__nd$0$i$i1983$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4385) {
     $$0$i7$i1989 = $__parent$i1981;$4393 = $4384;
     label = 1860;
     break;
    } else {
     $4404 = $4371;$__r$0$i1993 = $__nd$0$i$i1983$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1860) {
  __THREW__ = 0;
  $4386 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4387 = __THREW__; __THREW__ = 0;
  $4388 = $4387&1;
  if ($4388) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4389 = ((($4386)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4389>>2] = 1073741944; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4390 = ((($4386)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4390>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4386>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4391 = ((($4386)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4391>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4392 = ((($4386)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4392>>2] = $4393; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i1989>>2] = $4386; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4394 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4395 = HEAP32[$4394>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4396 = ($4395|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4396) {
   $4399 = $4386;
  } else {
   $4397 = $4395; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4397; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i1991 = HEAP32[$$0$i7$i1989>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4399 = $$pre$i$i1991;
  }
  $4398 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4398,$4399); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4400 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4401 = (($4400) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4401; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4508 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4404 = $$pre4508;$__r$0$i1993 = $4386;
 }
 $4402 = ((($__r$0$i1993)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4402>>2] = 133; //@line 750 "modules/event/sdl/Event.cpp"
 $4403 = ($4404|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4403) {
   HEAP32[$__parent$i1996>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4416 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2004 = $2;$4426 = $4416;
   label = 1874;
  } else {
   $__nd$0$i$i1998 = $4404;
   while(1) {
    $4405 = ((($__nd$0$i$i1998)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4406 = HEAP32[$4405>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4407 = ($4406|0)>(1073741945); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4407) {
     $4408 = HEAP32[$__nd$0$i$i1998>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4409 = ($4408|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4409) {
      $$lcssa7068 = $__nd$0$i$i1998;$__nd$0$i$i1998$lcssa7066 = $__nd$0$i$i1998;
      label = 1867;
      break;
     } else {
      $$sink$i$i1999 = $4408;
     }
    } else {
     $4411 = ($4406|0)<(1073741945); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4411)) {
      $__nd$0$i$i1998$lcssa = $__nd$0$i$i1998;
      label = 1873;
      break;
     }
     $4412 = ((($__nd$0$i$i1998)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4413 = HEAP32[$4412>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4414 = ($4413|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4414) {
      $$lcssa7067 = $4412;$__nd$0$i$i1998$lcssa7065 = $__nd$0$i$i1998;
      label = 1871;
      break;
     } else {
      $$sink$i$i1999 = $4413;
     }
    }
    $__nd$0$i$i1998 = $$sink$i$i1999;
   }
   if ((label|0) == 1867) {
    HEAP32[$__parent$i1996>>2] = $__nd$0$i$i1998$lcssa7066; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4410 = $__nd$0$i$i1998$lcssa7066; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2004 = $$lcssa7068;$4426 = $4410;
    label = 1874;
    break;
   }
   else if ((label|0) == 1871) {
    HEAP32[$__parent$i1996>>2] = $__nd$0$i$i1998$lcssa7065; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4415 = $__nd$0$i$i1998$lcssa7065; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2004 = $$lcssa7067;$4426 = $4415;
    label = 1874;
    break;
   }
   else if ((label|0) == 1873) {
    HEAP32[$__parent$i1996>>2] = $__nd$0$i$i1998$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4417 = $__nd$0$i$i1998$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4418 = ($__nd$0$i$i1998$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4418) {
     $$0$i7$i2004 = $__parent$i1996;$4426 = $4417;
     label = 1874;
     break;
    } else {
     $4437 = $4404;$__r$0$i2008 = $__nd$0$i$i1998$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1874) {
  __THREW__ = 0;
  $4419 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4420 = __THREW__; __THREW__ = 0;
  $4421 = $4420&1;
  if ($4421) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4422 = ((($4419)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4422>>2] = 1073741945; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4423 = ((($4419)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4423>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4419>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4424 = ((($4419)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4424>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4425 = ((($4419)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4425>>2] = $4426; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2004>>2] = $4419; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4427 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4428 = HEAP32[$4427>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4429 = ($4428|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4429) {
   $4432 = $4419;
  } else {
   $4430 = $4428; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4430; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2006 = HEAP32[$$0$i7$i2004>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4432 = $$pre$i$i2006;
  }
  $4431 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4431,$4432); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4433 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4434 = (($4433) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4434; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4509 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4437 = $$pre4509;$__r$0$i2008 = $4419;
 }
 $4435 = ((($__r$0$i2008)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4435>>2] = 134; //@line 751 "modules/event/sdl/Event.cpp"
 $4436 = ($4437|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4436) {
   HEAP32[$__parent$i2011>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4449 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2019 = $2;$4459 = $4449;
   label = 1888;
  } else {
   $__nd$0$i$i2013 = $4437;
   while(1) {
    $4438 = ((($__nd$0$i$i2013)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4439 = HEAP32[$4438>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4440 = ($4439|0)>(1073741946); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4440) {
     $4441 = HEAP32[$__nd$0$i$i2013>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4442 = ($4441|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4442) {
      $$lcssa7064 = $__nd$0$i$i2013;$__nd$0$i$i2013$lcssa7062 = $__nd$0$i$i2013;
      label = 1881;
      break;
     } else {
      $$sink$i$i2014 = $4441;
     }
    } else {
     $4444 = ($4439|0)<(1073741946); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4444)) {
      $__nd$0$i$i2013$lcssa = $__nd$0$i$i2013;
      label = 1887;
      break;
     }
     $4445 = ((($__nd$0$i$i2013)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4446 = HEAP32[$4445>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4447 = ($4446|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4447) {
      $$lcssa7063 = $4445;$__nd$0$i$i2013$lcssa7061 = $__nd$0$i$i2013;
      label = 1885;
      break;
     } else {
      $$sink$i$i2014 = $4446;
     }
    }
    $__nd$0$i$i2013 = $$sink$i$i2014;
   }
   if ((label|0) == 1881) {
    HEAP32[$__parent$i2011>>2] = $__nd$0$i$i2013$lcssa7062; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4443 = $__nd$0$i$i2013$lcssa7062; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2019 = $$lcssa7064;$4459 = $4443;
    label = 1888;
    break;
   }
   else if ((label|0) == 1885) {
    HEAP32[$__parent$i2011>>2] = $__nd$0$i$i2013$lcssa7061; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4448 = $__nd$0$i$i2013$lcssa7061; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2019 = $$lcssa7063;$4459 = $4448;
    label = 1888;
    break;
   }
   else if ((label|0) == 1887) {
    HEAP32[$__parent$i2011>>2] = $__nd$0$i$i2013$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4450 = $__nd$0$i$i2013$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4451 = ($__nd$0$i$i2013$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4451) {
     $$0$i7$i2019 = $__parent$i2011;$4459 = $4450;
     label = 1888;
     break;
    } else {
     $4470 = $4437;$__r$0$i2023 = $__nd$0$i$i2013$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1888) {
  __THREW__ = 0;
  $4452 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4453 = __THREW__; __THREW__ = 0;
  $4454 = $4453&1;
  if ($4454) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4455 = ((($4452)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4455>>2] = 1073741946; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4456 = ((($4452)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4456>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4452>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4457 = ((($4452)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4457>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4458 = ((($4452)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4458>>2] = $4459; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2019>>2] = $4452; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4460 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4461 = HEAP32[$4460>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4462 = ($4461|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4462) {
   $4465 = $4452;
  } else {
   $4463 = $4461; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4463; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2021 = HEAP32[$$0$i7$i2019>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4465 = $$pre$i$i2021;
  }
  $4464 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4464,$4465); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4466 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4467 = (($4466) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4467; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4510 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4470 = $$pre4510;$__r$0$i2023 = $4452;
 }
 $4468 = ((($__r$0$i2023)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4468>>2] = 135; //@line 752 "modules/event/sdl/Event.cpp"
 $4469 = ($4470|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4469) {
   HEAP32[$__parent$i2026>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4482 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2034 = $2;$4492 = $4482;
   label = 1902;
  } else {
   $__nd$0$i$i2028 = $4470;
   while(1) {
    $4471 = ((($__nd$0$i$i2028)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4472 = HEAP32[$4471>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4473 = ($4472|0)>(1073741947); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4473) {
     $4474 = HEAP32[$__nd$0$i$i2028>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4475 = ($4474|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4475) {
      $$lcssa7060 = $__nd$0$i$i2028;$__nd$0$i$i2028$lcssa7058 = $__nd$0$i$i2028;
      label = 1895;
      break;
     } else {
      $$sink$i$i2029 = $4474;
     }
    } else {
     $4477 = ($4472|0)<(1073741947); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4477)) {
      $__nd$0$i$i2028$lcssa = $__nd$0$i$i2028;
      label = 1901;
      break;
     }
     $4478 = ((($__nd$0$i$i2028)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4479 = HEAP32[$4478>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4480 = ($4479|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4480) {
      $$lcssa7059 = $4478;$__nd$0$i$i2028$lcssa7057 = $__nd$0$i$i2028;
      label = 1899;
      break;
     } else {
      $$sink$i$i2029 = $4479;
     }
    }
    $__nd$0$i$i2028 = $$sink$i$i2029;
   }
   if ((label|0) == 1895) {
    HEAP32[$__parent$i2026>>2] = $__nd$0$i$i2028$lcssa7058; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4476 = $__nd$0$i$i2028$lcssa7058; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2034 = $$lcssa7060;$4492 = $4476;
    label = 1902;
    break;
   }
   else if ((label|0) == 1899) {
    HEAP32[$__parent$i2026>>2] = $__nd$0$i$i2028$lcssa7057; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4481 = $__nd$0$i$i2028$lcssa7057; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2034 = $$lcssa7059;$4492 = $4481;
    label = 1902;
    break;
   }
   else if ((label|0) == 1901) {
    HEAP32[$__parent$i2026>>2] = $__nd$0$i$i2028$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4483 = $__nd$0$i$i2028$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4484 = ($__nd$0$i$i2028$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4484) {
     $$0$i7$i2034 = $__parent$i2026;$4492 = $4483;
     label = 1902;
     break;
    } else {
     $4503 = $4470;$__r$0$i2038 = $__nd$0$i$i2028$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1902) {
  __THREW__ = 0;
  $4485 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4486 = __THREW__; __THREW__ = 0;
  $4487 = $4486&1;
  if ($4487) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4488 = ((($4485)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4488>>2] = 1073741947; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4489 = ((($4485)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4489>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4485>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4490 = ((($4485)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4490>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4491 = ((($4485)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4491>>2] = $4492; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2034>>2] = $4485; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4493 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4494 = HEAP32[$4493>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4495 = ($4494|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4495) {
   $4498 = $4485;
  } else {
   $4496 = $4494; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4496; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2036 = HEAP32[$$0$i7$i2034>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4498 = $$pre$i$i2036;
  }
  $4497 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4497,$4498); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4499 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4500 = (($4499) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4500; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4511 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4503 = $$pre4511;$__r$0$i2038 = $4485;
 }
 $4501 = ((($__r$0$i2038)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4501>>2] = 136; //@line 753 "modules/event/sdl/Event.cpp"
 $4502 = ($4503|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4502) {
   HEAP32[$__parent$i2041>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4515 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2049 = $2;$4525 = $4515;
   label = 1916;
  } else {
   $__nd$0$i$i2043 = $4503;
   while(1) {
    $4504 = ((($__nd$0$i$i2043)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4505 = HEAP32[$4504>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4506 = ($4505|0)>(1073741948); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4506) {
     $4507 = HEAP32[$__nd$0$i$i2043>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4508 = ($4507|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4508) {
      $$lcssa7056 = $__nd$0$i$i2043;$__nd$0$i$i2043$lcssa7054 = $__nd$0$i$i2043;
      label = 1909;
      break;
     } else {
      $$sink$i$i2044 = $4507;
     }
    } else {
     $4510 = ($4505|0)<(1073741948); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4510)) {
      $__nd$0$i$i2043$lcssa = $__nd$0$i$i2043;
      label = 1915;
      break;
     }
     $4511 = ((($__nd$0$i$i2043)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4512 = HEAP32[$4511>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4513 = ($4512|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4513) {
      $$lcssa7055 = $4511;$__nd$0$i$i2043$lcssa7053 = $__nd$0$i$i2043;
      label = 1913;
      break;
     } else {
      $$sink$i$i2044 = $4512;
     }
    }
    $__nd$0$i$i2043 = $$sink$i$i2044;
   }
   if ((label|0) == 1909) {
    HEAP32[$__parent$i2041>>2] = $__nd$0$i$i2043$lcssa7054; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4509 = $__nd$0$i$i2043$lcssa7054; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2049 = $$lcssa7056;$4525 = $4509;
    label = 1916;
    break;
   }
   else if ((label|0) == 1913) {
    HEAP32[$__parent$i2041>>2] = $__nd$0$i$i2043$lcssa7053; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4514 = $__nd$0$i$i2043$lcssa7053; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2049 = $$lcssa7055;$4525 = $4514;
    label = 1916;
    break;
   }
   else if ((label|0) == 1915) {
    HEAP32[$__parent$i2041>>2] = $__nd$0$i$i2043$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4516 = $__nd$0$i$i2043$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4517 = ($__nd$0$i$i2043$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4517) {
     $$0$i7$i2049 = $__parent$i2041;$4525 = $4516;
     label = 1916;
     break;
    } else {
     $4536 = $4503;$__r$0$i2053 = $__nd$0$i$i2043$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1916) {
  __THREW__ = 0;
  $4518 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4519 = __THREW__; __THREW__ = 0;
  $4520 = $4519&1;
  if ($4520) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4521 = ((($4518)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4521>>2] = 1073741948; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4522 = ((($4518)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4522>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4518>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4523 = ((($4518)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4523>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4524 = ((($4518)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4524>>2] = $4525; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2049>>2] = $4518; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4526 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4527 = HEAP32[$4526>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4528 = ($4527|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4528) {
   $4531 = $4518;
  } else {
   $4529 = $4527; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4529; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2051 = HEAP32[$$0$i7$i2049>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4531 = $$pre$i$i2051;
  }
  $4530 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4530,$4531); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4532 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4533 = (($4532) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4533; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4512 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4536 = $$pre4512;$__r$0$i2053 = $4518;
 }
 $4534 = ((($__r$0$i2053)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4534>>2] = 137; //@line 754 "modules/event/sdl/Event.cpp"
 $4535 = ($4536|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4535) {
   HEAP32[$__parent$i2056>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4548 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2064 = $2;$4558 = $4548;
   label = 1930;
  } else {
   $__nd$0$i$i2058 = $4536;
   while(1) {
    $4537 = ((($__nd$0$i$i2058)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4538 = HEAP32[$4537>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4539 = ($4538|0)>(1073741949); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4539) {
     $4540 = HEAP32[$__nd$0$i$i2058>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4541 = ($4540|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4541) {
      $$lcssa7052 = $__nd$0$i$i2058;$__nd$0$i$i2058$lcssa7050 = $__nd$0$i$i2058;
      label = 1923;
      break;
     } else {
      $$sink$i$i2059 = $4540;
     }
    } else {
     $4543 = ($4538|0)<(1073741949); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4543)) {
      $__nd$0$i$i2058$lcssa = $__nd$0$i$i2058;
      label = 1929;
      break;
     }
     $4544 = ((($__nd$0$i$i2058)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4545 = HEAP32[$4544>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4546 = ($4545|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4546) {
      $$lcssa7051 = $4544;$__nd$0$i$i2058$lcssa7049 = $__nd$0$i$i2058;
      label = 1927;
      break;
     } else {
      $$sink$i$i2059 = $4545;
     }
    }
    $__nd$0$i$i2058 = $$sink$i$i2059;
   }
   if ((label|0) == 1923) {
    HEAP32[$__parent$i2056>>2] = $__nd$0$i$i2058$lcssa7050; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4542 = $__nd$0$i$i2058$lcssa7050; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2064 = $$lcssa7052;$4558 = $4542;
    label = 1930;
    break;
   }
   else if ((label|0) == 1927) {
    HEAP32[$__parent$i2056>>2] = $__nd$0$i$i2058$lcssa7049; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4547 = $__nd$0$i$i2058$lcssa7049; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2064 = $$lcssa7051;$4558 = $4547;
    label = 1930;
    break;
   }
   else if ((label|0) == 1929) {
    HEAP32[$__parent$i2056>>2] = $__nd$0$i$i2058$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4549 = $__nd$0$i$i2058$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4550 = ($__nd$0$i$i2058$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4550) {
     $$0$i7$i2064 = $__parent$i2056;$4558 = $4549;
     label = 1930;
     break;
    } else {
     $4569 = $4536;$__r$0$i2068 = $__nd$0$i$i2058$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1930) {
  __THREW__ = 0;
  $4551 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4552 = __THREW__; __THREW__ = 0;
  $4553 = $4552&1;
  if ($4553) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4554 = ((($4551)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4554>>2] = 1073741949; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4555 = ((($4551)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4555>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4551>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4556 = ((($4551)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4556>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4557 = ((($4551)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4557>>2] = $4558; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2064>>2] = $4551; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4559 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4560 = HEAP32[$4559>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4561 = ($4560|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4561) {
   $4564 = $4551;
  } else {
   $4562 = $4560; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4562; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2066 = HEAP32[$$0$i7$i2064>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4564 = $$pre$i$i2066;
  }
  $4563 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4563,$4564); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4565 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4566 = (($4565) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4566; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4513 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4569 = $$pre4513;$__r$0$i2068 = $4551;
 }
 $4567 = ((($__r$0$i2068)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4567>>2] = 138; //@line 755 "modules/event/sdl/Event.cpp"
 $4568 = ($4569|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4568) {
   HEAP32[$__parent$i2071>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4581 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2079 = $2;$4591 = $4581;
   label = 1944;
  } else {
   $__nd$0$i$i2073 = $4569;
   while(1) {
    $4570 = ((($__nd$0$i$i2073)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4571 = HEAP32[$4570>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4572 = ($4571|0)>(1073741950); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4572) {
     $4573 = HEAP32[$__nd$0$i$i2073>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4574 = ($4573|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4574) {
      $$lcssa7048 = $__nd$0$i$i2073;$__nd$0$i$i2073$lcssa7046 = $__nd$0$i$i2073;
      label = 1937;
      break;
     } else {
      $$sink$i$i2074 = $4573;
     }
    } else {
     $4576 = ($4571|0)<(1073741950); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4576)) {
      $__nd$0$i$i2073$lcssa = $__nd$0$i$i2073;
      label = 1943;
      break;
     }
     $4577 = ((($__nd$0$i$i2073)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4578 = HEAP32[$4577>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4579 = ($4578|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4579) {
      $$lcssa7047 = $4577;$__nd$0$i$i2073$lcssa7045 = $__nd$0$i$i2073;
      label = 1941;
      break;
     } else {
      $$sink$i$i2074 = $4578;
     }
    }
    $__nd$0$i$i2073 = $$sink$i$i2074;
   }
   if ((label|0) == 1937) {
    HEAP32[$__parent$i2071>>2] = $__nd$0$i$i2073$lcssa7046; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4575 = $__nd$0$i$i2073$lcssa7046; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2079 = $$lcssa7048;$4591 = $4575;
    label = 1944;
    break;
   }
   else if ((label|0) == 1941) {
    HEAP32[$__parent$i2071>>2] = $__nd$0$i$i2073$lcssa7045; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4580 = $__nd$0$i$i2073$lcssa7045; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2079 = $$lcssa7047;$4591 = $4580;
    label = 1944;
    break;
   }
   else if ((label|0) == 1943) {
    HEAP32[$__parent$i2071>>2] = $__nd$0$i$i2073$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4582 = $__nd$0$i$i2073$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4583 = ($__nd$0$i$i2073$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4583) {
     $$0$i7$i2079 = $__parent$i2071;$4591 = $4582;
     label = 1944;
     break;
    } else {
     $4602 = $4569;$__r$0$i2083 = $__nd$0$i$i2073$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1944) {
  __THREW__ = 0;
  $4584 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4585 = __THREW__; __THREW__ = 0;
  $4586 = $4585&1;
  if ($4586) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4587 = ((($4584)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4587>>2] = 1073741950; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4588 = ((($4584)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4588>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4584>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4589 = ((($4584)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4589>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4590 = ((($4584)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4590>>2] = $4591; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2079>>2] = $4584; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4592 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4593 = HEAP32[$4592>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4594 = ($4593|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4594) {
   $4597 = $4584;
  } else {
   $4595 = $4593; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4595; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2081 = HEAP32[$$0$i7$i2079>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4597 = $$pre$i$i2081;
  }
  $4596 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4596,$4597); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4598 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4599 = (($4598) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4599; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4514 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4602 = $$pre4514;$__r$0$i2083 = $4584;
 }
 $4600 = ((($__r$0$i2083)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4600>>2] = 139; //@line 756 "modules/event/sdl/Event.cpp"
 $4601 = ($4602|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4601) {
   HEAP32[$__parent$i2086>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4614 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2094 = $2;$4624 = $4614;
   label = 1958;
  } else {
   $__nd$0$i$i2088 = $4602;
   while(1) {
    $4603 = ((($__nd$0$i$i2088)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4604 = HEAP32[$4603>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4605 = ($4604|0)>(1073741951); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4605) {
     $4606 = HEAP32[$__nd$0$i$i2088>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4607 = ($4606|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4607) {
      $$lcssa7044 = $__nd$0$i$i2088;$__nd$0$i$i2088$lcssa7042 = $__nd$0$i$i2088;
      label = 1951;
      break;
     } else {
      $$sink$i$i2089 = $4606;
     }
    } else {
     $4609 = ($4604|0)<(1073741951); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4609)) {
      $__nd$0$i$i2088$lcssa = $__nd$0$i$i2088;
      label = 1957;
      break;
     }
     $4610 = ((($__nd$0$i$i2088)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4611 = HEAP32[$4610>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4612 = ($4611|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4612) {
      $$lcssa7043 = $4610;$__nd$0$i$i2088$lcssa7041 = $__nd$0$i$i2088;
      label = 1955;
      break;
     } else {
      $$sink$i$i2089 = $4611;
     }
    }
    $__nd$0$i$i2088 = $$sink$i$i2089;
   }
   if ((label|0) == 1951) {
    HEAP32[$__parent$i2086>>2] = $__nd$0$i$i2088$lcssa7042; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4608 = $__nd$0$i$i2088$lcssa7042; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2094 = $$lcssa7044;$4624 = $4608;
    label = 1958;
    break;
   }
   else if ((label|0) == 1955) {
    HEAP32[$__parent$i2086>>2] = $__nd$0$i$i2088$lcssa7041; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4613 = $__nd$0$i$i2088$lcssa7041; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2094 = $$lcssa7043;$4624 = $4613;
    label = 1958;
    break;
   }
   else if ((label|0) == 1957) {
    HEAP32[$__parent$i2086>>2] = $__nd$0$i$i2088$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4615 = $__nd$0$i$i2088$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4616 = ($__nd$0$i$i2088$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4616) {
     $$0$i7$i2094 = $__parent$i2086;$4624 = $4615;
     label = 1958;
     break;
    } else {
     $4635 = $4602;$__r$0$i2098 = $__nd$0$i$i2088$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1958) {
  __THREW__ = 0;
  $4617 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4618 = __THREW__; __THREW__ = 0;
  $4619 = $4618&1;
  if ($4619) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4620 = ((($4617)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4620>>2] = 1073741951; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4621 = ((($4617)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4621>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4617>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4622 = ((($4617)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4622>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4623 = ((($4617)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4623>>2] = $4624; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2094>>2] = $4617; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4625 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4626 = HEAP32[$4625>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4627 = ($4626|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4627) {
   $4630 = $4617;
  } else {
   $4628 = $4626; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4628; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2096 = HEAP32[$$0$i7$i2094>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4630 = $$pre$i$i2096;
  }
  $4629 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4629,$4630); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4631 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4632 = (($4631) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4632; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4515 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4635 = $$pre4515;$__r$0$i2098 = $4617;
 }
 $4633 = ((($__r$0$i2098)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4633>>2] = 140; //@line 757 "modules/event/sdl/Event.cpp"
 $4634 = ($4635|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4634) {
   HEAP32[$__parent$i2101>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4647 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2109 = $2;$4657 = $4647;
   label = 1972;
  } else {
   $__nd$0$i$i2103 = $4635;
   while(1) {
    $4636 = ((($__nd$0$i$i2103)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4637 = HEAP32[$4636>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4638 = ($4637|0)>(1073741952); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4638) {
     $4639 = HEAP32[$__nd$0$i$i2103>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4640 = ($4639|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4640) {
      $$lcssa7040 = $__nd$0$i$i2103;$__nd$0$i$i2103$lcssa7038 = $__nd$0$i$i2103;
      label = 1965;
      break;
     } else {
      $$sink$i$i2104 = $4639;
     }
    } else {
     $4642 = ($4637|0)<(1073741952); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4642)) {
      $__nd$0$i$i2103$lcssa = $__nd$0$i$i2103;
      label = 1971;
      break;
     }
     $4643 = ((($__nd$0$i$i2103)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4644 = HEAP32[$4643>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4645 = ($4644|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4645) {
      $$lcssa7039 = $4643;$__nd$0$i$i2103$lcssa7037 = $__nd$0$i$i2103;
      label = 1969;
      break;
     } else {
      $$sink$i$i2104 = $4644;
     }
    }
    $__nd$0$i$i2103 = $$sink$i$i2104;
   }
   if ((label|0) == 1965) {
    HEAP32[$__parent$i2101>>2] = $__nd$0$i$i2103$lcssa7038; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4641 = $__nd$0$i$i2103$lcssa7038; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2109 = $$lcssa7040;$4657 = $4641;
    label = 1972;
    break;
   }
   else if ((label|0) == 1969) {
    HEAP32[$__parent$i2101>>2] = $__nd$0$i$i2103$lcssa7037; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4646 = $__nd$0$i$i2103$lcssa7037; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2109 = $$lcssa7039;$4657 = $4646;
    label = 1972;
    break;
   }
   else if ((label|0) == 1971) {
    HEAP32[$__parent$i2101>>2] = $__nd$0$i$i2103$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4648 = $__nd$0$i$i2103$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4649 = ($__nd$0$i$i2103$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4649) {
     $$0$i7$i2109 = $__parent$i2101;$4657 = $4648;
     label = 1972;
     break;
    } else {
     $4668 = $4635;$__r$0$i2113 = $__nd$0$i$i2103$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1972) {
  __THREW__ = 0;
  $4650 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4651 = __THREW__; __THREW__ = 0;
  $4652 = $4651&1;
  if ($4652) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4653 = ((($4650)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4653>>2] = 1073741952; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4654 = ((($4650)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4654>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4650>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4655 = ((($4650)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4655>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4656 = ((($4650)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4656>>2] = $4657; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2109>>2] = $4650; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4658 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4659 = HEAP32[$4658>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4660 = ($4659|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4660) {
   $4663 = $4650;
  } else {
   $4661 = $4659; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4661; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2111 = HEAP32[$$0$i7$i2109>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4663 = $$pre$i$i2111;
  }
  $4662 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4662,$4663); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4664 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4665 = (($4664) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4665; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4516 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4668 = $$pre4516;$__r$0$i2113 = $4650;
 }
 $4666 = ((($__r$0$i2113)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4666>>2] = 141; //@line 758 "modules/event/sdl/Event.cpp"
 $4667 = ($4668|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4667) {
   HEAP32[$__parent$i2116>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4680 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2124 = $2;$4690 = $4680;
   label = 1986;
  } else {
   $__nd$0$i$i2118 = $4668;
   while(1) {
    $4669 = ((($__nd$0$i$i2118)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4670 = HEAP32[$4669>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4671 = ($4670|0)>(1073741953); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4671) {
     $4672 = HEAP32[$__nd$0$i$i2118>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4673 = ($4672|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4673) {
      $$lcssa7036 = $__nd$0$i$i2118;$__nd$0$i$i2118$lcssa7034 = $__nd$0$i$i2118;
      label = 1979;
      break;
     } else {
      $$sink$i$i2119 = $4672;
     }
    } else {
     $4675 = ($4670|0)<(1073741953); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4675)) {
      $__nd$0$i$i2118$lcssa = $__nd$0$i$i2118;
      label = 1985;
      break;
     }
     $4676 = ((($__nd$0$i$i2118)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4677 = HEAP32[$4676>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4678 = ($4677|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4678) {
      $$lcssa7035 = $4676;$__nd$0$i$i2118$lcssa7033 = $__nd$0$i$i2118;
      label = 1983;
      break;
     } else {
      $$sink$i$i2119 = $4677;
     }
    }
    $__nd$0$i$i2118 = $$sink$i$i2119;
   }
   if ((label|0) == 1979) {
    HEAP32[$__parent$i2116>>2] = $__nd$0$i$i2118$lcssa7034; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4674 = $__nd$0$i$i2118$lcssa7034; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2124 = $$lcssa7036;$4690 = $4674;
    label = 1986;
    break;
   }
   else if ((label|0) == 1983) {
    HEAP32[$__parent$i2116>>2] = $__nd$0$i$i2118$lcssa7033; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4679 = $__nd$0$i$i2118$lcssa7033; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2124 = $$lcssa7035;$4690 = $4679;
    label = 1986;
    break;
   }
   else if ((label|0) == 1985) {
    HEAP32[$__parent$i2116>>2] = $__nd$0$i$i2118$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4681 = $__nd$0$i$i2118$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4682 = ($__nd$0$i$i2118$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4682) {
     $$0$i7$i2124 = $__parent$i2116;$4690 = $4681;
     label = 1986;
     break;
    } else {
     $4701 = $4668;$__r$0$i2128 = $__nd$0$i$i2118$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 1986) {
  __THREW__ = 0;
  $4683 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4684 = __THREW__; __THREW__ = 0;
  $4685 = $4684&1;
  if ($4685) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4686 = ((($4683)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4686>>2] = 1073741953; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4687 = ((($4683)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4687>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4683>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4688 = ((($4683)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4688>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4689 = ((($4683)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4689>>2] = $4690; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2124>>2] = $4683; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4691 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4692 = HEAP32[$4691>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4693 = ($4692|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4693) {
   $4696 = $4683;
  } else {
   $4694 = $4692; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4694; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2126 = HEAP32[$$0$i7$i2124>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4696 = $$pre$i$i2126;
  }
  $4695 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4695,$4696); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4697 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4698 = (($4697) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4698; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4517 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4701 = $$pre4517;$__r$0$i2128 = $4683;
 }
 $4699 = ((($__r$0$i2128)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4699>>2] = 142; //@line 759 "modules/event/sdl/Event.cpp"
 $4700 = ($4701|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4700) {
   HEAP32[$__parent$i2131>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4713 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2139 = $2;$4723 = $4713;
   label = 2000;
  } else {
   $__nd$0$i$i2133 = $4701;
   while(1) {
    $4702 = ((($__nd$0$i$i2133)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4703 = HEAP32[$4702>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4704 = ($4703|0)>(1073741977); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4704) {
     $4705 = HEAP32[$__nd$0$i$i2133>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4706 = ($4705|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4706) {
      $$lcssa7032 = $__nd$0$i$i2133;$__nd$0$i$i2133$lcssa7030 = $__nd$0$i$i2133;
      label = 1993;
      break;
     } else {
      $$sink$i$i2134 = $4705;
     }
    } else {
     $4708 = ($4703|0)<(1073741977); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4708)) {
      $__nd$0$i$i2133$lcssa = $__nd$0$i$i2133;
      label = 1999;
      break;
     }
     $4709 = ((($__nd$0$i$i2133)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4710 = HEAP32[$4709>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4711 = ($4710|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4711) {
      $$lcssa7031 = $4709;$__nd$0$i$i2133$lcssa7029 = $__nd$0$i$i2133;
      label = 1997;
      break;
     } else {
      $$sink$i$i2134 = $4710;
     }
    }
    $__nd$0$i$i2133 = $$sink$i$i2134;
   }
   if ((label|0) == 1993) {
    HEAP32[$__parent$i2131>>2] = $__nd$0$i$i2133$lcssa7030; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4707 = $__nd$0$i$i2133$lcssa7030; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2139 = $$lcssa7032;$4723 = $4707;
    label = 2000;
    break;
   }
   else if ((label|0) == 1997) {
    HEAP32[$__parent$i2131>>2] = $__nd$0$i$i2133$lcssa7029; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4712 = $__nd$0$i$i2133$lcssa7029; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2139 = $$lcssa7031;$4723 = $4712;
    label = 2000;
    break;
   }
   else if ((label|0) == 1999) {
    HEAP32[$__parent$i2131>>2] = $__nd$0$i$i2133$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4714 = $__nd$0$i$i2133$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4715 = ($__nd$0$i$i2133$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4715) {
     $$0$i7$i2139 = $__parent$i2131;$4723 = $4714;
     label = 2000;
     break;
    } else {
     $4734 = $4701;$__r$0$i2143 = $__nd$0$i$i2133$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2000) {
  __THREW__ = 0;
  $4716 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4717 = __THREW__; __THREW__ = 0;
  $4718 = $4717&1;
  if ($4718) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4719 = ((($4716)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4719>>2] = 1073741977; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4720 = ((($4716)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4720>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4716>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4721 = ((($4716)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4721>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4722 = ((($4716)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4722>>2] = $4723; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2139>>2] = $4716; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4724 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4725 = HEAP32[$4724>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4726 = ($4725|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4726) {
   $4729 = $4716;
  } else {
   $4727 = $4725; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4727; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2141 = HEAP32[$$0$i7$i2139>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4729 = $$pre$i$i2141;
  }
  $4728 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4728,$4729); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4730 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4731 = (($4730) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4731; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4518 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4734 = $$pre4518;$__r$0$i2143 = $4716;
 }
 $4732 = ((($__r$0$i2143)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4732>>2] = 143; //@line 761 "modules/event/sdl/Event.cpp"
 $4733 = ($4734|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4733) {
   HEAP32[$__parent$i2146>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4746 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2154 = $2;$4756 = $4746;
   label = 2014;
  } else {
   $__nd$0$i$i2148 = $4734;
   while(1) {
    $4735 = ((($__nd$0$i$i2148)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4736 = HEAP32[$4735>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4737 = ($4736|0)>(1073741978); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4737) {
     $4738 = HEAP32[$__nd$0$i$i2148>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4739 = ($4738|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4739) {
      $$lcssa7028 = $__nd$0$i$i2148;$__nd$0$i$i2148$lcssa7026 = $__nd$0$i$i2148;
      label = 2007;
      break;
     } else {
      $$sink$i$i2149 = $4738;
     }
    } else {
     $4741 = ($4736|0)<(1073741978); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4741)) {
      $__nd$0$i$i2148$lcssa = $__nd$0$i$i2148;
      label = 2013;
      break;
     }
     $4742 = ((($__nd$0$i$i2148)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4743 = HEAP32[$4742>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4744 = ($4743|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4744) {
      $$lcssa7027 = $4742;$__nd$0$i$i2148$lcssa7025 = $__nd$0$i$i2148;
      label = 2011;
      break;
     } else {
      $$sink$i$i2149 = $4743;
     }
    }
    $__nd$0$i$i2148 = $$sink$i$i2149;
   }
   if ((label|0) == 2007) {
    HEAP32[$__parent$i2146>>2] = $__nd$0$i$i2148$lcssa7026; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4740 = $__nd$0$i$i2148$lcssa7026; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2154 = $$lcssa7028;$4756 = $4740;
    label = 2014;
    break;
   }
   else if ((label|0) == 2011) {
    HEAP32[$__parent$i2146>>2] = $__nd$0$i$i2148$lcssa7025; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4745 = $__nd$0$i$i2148$lcssa7025; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2154 = $$lcssa7027;$4756 = $4745;
    label = 2014;
    break;
   }
   else if ((label|0) == 2013) {
    HEAP32[$__parent$i2146>>2] = $__nd$0$i$i2148$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4747 = $__nd$0$i$i2148$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4748 = ($__nd$0$i$i2148$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4748) {
     $$0$i7$i2154 = $__parent$i2146;$4756 = $4747;
     label = 2014;
     break;
    } else {
     $4767 = $4734;$__r$0$i2158 = $__nd$0$i$i2148$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2014) {
  __THREW__ = 0;
  $4749 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4750 = __THREW__; __THREW__ = 0;
  $4751 = $4750&1;
  if ($4751) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4752 = ((($4749)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4752>>2] = 1073741978; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4753 = ((($4749)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4753>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4749>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4754 = ((($4749)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4754>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4755 = ((($4749)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4755>>2] = $4756; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2154>>2] = $4749; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4757 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4758 = HEAP32[$4757>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4759 = ($4758|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4759) {
   $4762 = $4749;
  } else {
   $4760 = $4758; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4760; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2156 = HEAP32[$$0$i7$i2154>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4762 = $$pre$i$i2156;
  }
  $4761 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4761,$4762); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4763 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4764 = (($4763) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4764; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4519 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4767 = $$pre4519;$__r$0$i2158 = $4749;
 }
 $4765 = ((($__r$0$i2158)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4765>>2] = 144; //@line 762 "modules/event/sdl/Event.cpp"
 $4766 = ($4767|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4766) {
   HEAP32[$__parent$i2161>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4779 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2169 = $2;$4789 = $4779;
   label = 2028;
  } else {
   $__nd$0$i$i2163 = $4767;
   while(1) {
    $4768 = ((($__nd$0$i$i2163)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4769 = HEAP32[$4768>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4770 = ($4769|0)>(1073741979); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4770) {
     $4771 = HEAP32[$__nd$0$i$i2163>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4772 = ($4771|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4772) {
      $$lcssa7024 = $__nd$0$i$i2163;$__nd$0$i$i2163$lcssa7022 = $__nd$0$i$i2163;
      label = 2021;
      break;
     } else {
      $$sink$i$i2164 = $4771;
     }
    } else {
     $4774 = ($4769|0)<(1073741979); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4774)) {
      $__nd$0$i$i2163$lcssa = $__nd$0$i$i2163;
      label = 2027;
      break;
     }
     $4775 = ((($__nd$0$i$i2163)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4776 = HEAP32[$4775>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4777 = ($4776|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4777) {
      $$lcssa7023 = $4775;$__nd$0$i$i2163$lcssa7021 = $__nd$0$i$i2163;
      label = 2025;
      break;
     } else {
      $$sink$i$i2164 = $4776;
     }
    }
    $__nd$0$i$i2163 = $$sink$i$i2164;
   }
   if ((label|0) == 2021) {
    HEAP32[$__parent$i2161>>2] = $__nd$0$i$i2163$lcssa7022; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4773 = $__nd$0$i$i2163$lcssa7022; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2169 = $$lcssa7024;$4789 = $4773;
    label = 2028;
    break;
   }
   else if ((label|0) == 2025) {
    HEAP32[$__parent$i2161>>2] = $__nd$0$i$i2163$lcssa7021; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4778 = $__nd$0$i$i2163$lcssa7021; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2169 = $$lcssa7023;$4789 = $4778;
    label = 2028;
    break;
   }
   else if ((label|0) == 2027) {
    HEAP32[$__parent$i2161>>2] = $__nd$0$i$i2163$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4780 = $__nd$0$i$i2163$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4781 = ($__nd$0$i$i2163$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4781) {
     $$0$i7$i2169 = $__parent$i2161;$4789 = $4780;
     label = 2028;
     break;
    } else {
     $4800 = $4767;$__r$0$i2173 = $__nd$0$i$i2163$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2028) {
  __THREW__ = 0;
  $4782 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4783 = __THREW__; __THREW__ = 0;
  $4784 = $4783&1;
  if ($4784) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4785 = ((($4782)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4785>>2] = 1073741979; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4786 = ((($4782)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4786>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4782>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4787 = ((($4782)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4787>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4788 = ((($4782)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4788>>2] = $4789; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2169>>2] = $4782; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4790 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4791 = HEAP32[$4790>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4792 = ($4791|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4792) {
   $4795 = $4782;
  } else {
   $4793 = $4791; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4793; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2171 = HEAP32[$$0$i7$i2169>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4795 = $$pre$i$i2171;
  }
  $4794 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4794,$4795); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4796 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4797 = (($4796) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4797; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4520 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4800 = $$pre4520;$__r$0$i2173 = $4782;
 }
 $4798 = ((($__r$0$i2173)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4798>>2] = 145; //@line 763 "modules/event/sdl/Event.cpp"
 $4799 = ($4800|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4799) {
   HEAP32[$__parent$i2176>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4812 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2184 = $2;$4822 = $4812;
   label = 2042;
  } else {
   $__nd$0$i$i2178 = $4800;
   while(1) {
    $4801 = ((($__nd$0$i$i2178)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4802 = HEAP32[$4801>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4803 = ($4802|0)>(1073741980); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4803) {
     $4804 = HEAP32[$__nd$0$i$i2178>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4805 = ($4804|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4805) {
      $$lcssa7020 = $__nd$0$i$i2178;$__nd$0$i$i2178$lcssa7018 = $__nd$0$i$i2178;
      label = 2035;
      break;
     } else {
      $$sink$i$i2179 = $4804;
     }
    } else {
     $4807 = ($4802|0)<(1073741980); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4807)) {
      $__nd$0$i$i2178$lcssa = $__nd$0$i$i2178;
      label = 2041;
      break;
     }
     $4808 = ((($__nd$0$i$i2178)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4809 = HEAP32[$4808>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4810 = ($4809|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4810) {
      $$lcssa7019 = $4808;$__nd$0$i$i2178$lcssa7017 = $__nd$0$i$i2178;
      label = 2039;
      break;
     } else {
      $$sink$i$i2179 = $4809;
     }
    }
    $__nd$0$i$i2178 = $$sink$i$i2179;
   }
   if ((label|0) == 2035) {
    HEAP32[$__parent$i2176>>2] = $__nd$0$i$i2178$lcssa7018; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4806 = $__nd$0$i$i2178$lcssa7018; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2184 = $$lcssa7020;$4822 = $4806;
    label = 2042;
    break;
   }
   else if ((label|0) == 2039) {
    HEAP32[$__parent$i2176>>2] = $__nd$0$i$i2178$lcssa7017; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4811 = $__nd$0$i$i2178$lcssa7017; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2184 = $$lcssa7019;$4822 = $4811;
    label = 2042;
    break;
   }
   else if ((label|0) == 2041) {
    HEAP32[$__parent$i2176>>2] = $__nd$0$i$i2178$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4813 = $__nd$0$i$i2178$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4814 = ($__nd$0$i$i2178$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4814) {
     $$0$i7$i2184 = $__parent$i2176;$4822 = $4813;
     label = 2042;
     break;
    } else {
     $4833 = $4800;$__r$0$i2188 = $__nd$0$i$i2178$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2042) {
  __THREW__ = 0;
  $4815 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4816 = __THREW__; __THREW__ = 0;
  $4817 = $4816&1;
  if ($4817) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4818 = ((($4815)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4818>>2] = 1073741980; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4819 = ((($4815)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4819>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4815>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4820 = ((($4815)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4820>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4821 = ((($4815)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4821>>2] = $4822; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2184>>2] = $4815; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4823 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4824 = HEAP32[$4823>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4825 = ($4824|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4825) {
   $4828 = $4815;
  } else {
   $4826 = $4824; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4826; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2186 = HEAP32[$$0$i7$i2184>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4828 = $$pre$i$i2186;
  }
  $4827 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4827,$4828); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4829 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4830 = (($4829) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4830; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4521 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4833 = $$pre4521;$__r$0$i2188 = $4815;
 }
 $4831 = ((($__r$0$i2188)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4831>>2] = 146; //@line 764 "modules/event/sdl/Event.cpp"
 $4832 = ($4833|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4832) {
   HEAP32[$__parent$i2191>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4845 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2199 = $2;$4855 = $4845;
   label = 2056;
  } else {
   $__nd$0$i$i2193 = $4833;
   while(1) {
    $4834 = ((($__nd$0$i$i2193)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4835 = HEAP32[$4834>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4836 = ($4835|0)>(1073741981); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4836) {
     $4837 = HEAP32[$__nd$0$i$i2193>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4838 = ($4837|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4838) {
      $$lcssa7016 = $__nd$0$i$i2193;$__nd$0$i$i2193$lcssa7014 = $__nd$0$i$i2193;
      label = 2049;
      break;
     } else {
      $$sink$i$i2194 = $4837;
     }
    } else {
     $4840 = ($4835|0)<(1073741981); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4840)) {
      $__nd$0$i$i2193$lcssa = $__nd$0$i$i2193;
      label = 2055;
      break;
     }
     $4841 = ((($__nd$0$i$i2193)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4842 = HEAP32[$4841>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4843 = ($4842|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4843) {
      $$lcssa7015 = $4841;$__nd$0$i$i2193$lcssa7013 = $__nd$0$i$i2193;
      label = 2053;
      break;
     } else {
      $$sink$i$i2194 = $4842;
     }
    }
    $__nd$0$i$i2193 = $$sink$i$i2194;
   }
   if ((label|0) == 2049) {
    HEAP32[$__parent$i2191>>2] = $__nd$0$i$i2193$lcssa7014; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4839 = $__nd$0$i$i2193$lcssa7014; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2199 = $$lcssa7016;$4855 = $4839;
    label = 2056;
    break;
   }
   else if ((label|0) == 2053) {
    HEAP32[$__parent$i2191>>2] = $__nd$0$i$i2193$lcssa7013; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4844 = $__nd$0$i$i2193$lcssa7013; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2199 = $$lcssa7015;$4855 = $4844;
    label = 2056;
    break;
   }
   else if ((label|0) == 2055) {
    HEAP32[$__parent$i2191>>2] = $__nd$0$i$i2193$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4846 = $__nd$0$i$i2193$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4847 = ($__nd$0$i$i2193$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4847) {
     $$0$i7$i2199 = $__parent$i2191;$4855 = $4846;
     label = 2056;
     break;
    } else {
     $4866 = $4833;$__r$0$i2203 = $__nd$0$i$i2193$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2056) {
  __THREW__ = 0;
  $4848 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4849 = __THREW__; __THREW__ = 0;
  $4850 = $4849&1;
  if ($4850) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4851 = ((($4848)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4851>>2] = 1073741981; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4852 = ((($4848)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4852>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4848>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4853 = ((($4848)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4853>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4854 = ((($4848)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4854>>2] = $4855; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2199>>2] = $4848; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4856 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4857 = HEAP32[$4856>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4858 = ($4857|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4858) {
   $4861 = $4848;
  } else {
   $4859 = $4857; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4859; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2201 = HEAP32[$$0$i7$i2199>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4861 = $$pre$i$i2201;
  }
  $4860 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4860,$4861); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4862 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4863 = (($4862) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4863; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4522 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4866 = $$pre4522;$__r$0$i2203 = $4848;
 }
 $4864 = ((($__r$0$i2203)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4864>>2] = 147; //@line 765 "modules/event/sdl/Event.cpp"
 $4865 = ($4866|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4865) {
   HEAP32[$__parent$i2206>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4878 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2214 = $2;$4888 = $4878;
   label = 2070;
  } else {
   $__nd$0$i$i2208 = $4866;
   while(1) {
    $4867 = ((($__nd$0$i$i2208)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4868 = HEAP32[$4867>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4869 = ($4868|0)>(1073741982); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4869) {
     $4870 = HEAP32[$__nd$0$i$i2208>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4871 = ($4870|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4871) {
      $$lcssa7012 = $__nd$0$i$i2208;$__nd$0$i$i2208$lcssa7010 = $__nd$0$i$i2208;
      label = 2063;
      break;
     } else {
      $$sink$i$i2209 = $4870;
     }
    } else {
     $4873 = ($4868|0)<(1073741982); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4873)) {
      $__nd$0$i$i2208$lcssa = $__nd$0$i$i2208;
      label = 2069;
      break;
     }
     $4874 = ((($__nd$0$i$i2208)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4875 = HEAP32[$4874>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4876 = ($4875|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4876) {
      $$lcssa7011 = $4874;$__nd$0$i$i2208$lcssa7009 = $__nd$0$i$i2208;
      label = 2067;
      break;
     } else {
      $$sink$i$i2209 = $4875;
     }
    }
    $__nd$0$i$i2208 = $$sink$i$i2209;
   }
   if ((label|0) == 2063) {
    HEAP32[$__parent$i2206>>2] = $__nd$0$i$i2208$lcssa7010; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4872 = $__nd$0$i$i2208$lcssa7010; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2214 = $$lcssa7012;$4888 = $4872;
    label = 2070;
    break;
   }
   else if ((label|0) == 2067) {
    HEAP32[$__parent$i2206>>2] = $__nd$0$i$i2208$lcssa7009; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4877 = $__nd$0$i$i2208$lcssa7009; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2214 = $$lcssa7011;$4888 = $4877;
    label = 2070;
    break;
   }
   else if ((label|0) == 2069) {
    HEAP32[$__parent$i2206>>2] = $__nd$0$i$i2208$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4879 = $__nd$0$i$i2208$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4880 = ($__nd$0$i$i2208$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4880) {
     $$0$i7$i2214 = $__parent$i2206;$4888 = $4879;
     label = 2070;
     break;
    } else {
     $4899 = $4866;$__r$0$i2218 = $__nd$0$i$i2208$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2070) {
  __THREW__ = 0;
  $4881 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4882 = __THREW__; __THREW__ = 0;
  $4883 = $4882&1;
  if ($4883) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4884 = ((($4881)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4884>>2] = 1073741982; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4885 = ((($4881)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4885>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4881>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4886 = ((($4881)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4886>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4887 = ((($4881)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4887>>2] = $4888; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2214>>2] = $4881; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4889 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4890 = HEAP32[$4889>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4891 = ($4890|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4891) {
   $4894 = $4881;
  } else {
   $4892 = $4890; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4892; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2216 = HEAP32[$$0$i7$i2214>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4894 = $$pre$i$i2216;
  }
  $4893 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4893,$4894); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4895 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4896 = (($4895) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4896; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4523 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4899 = $$pre4523;$__r$0$i2218 = $4881;
 }
 $4897 = ((($__r$0$i2218)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4897>>2] = 148; //@line 766 "modules/event/sdl/Event.cpp"
 $4898 = ($4899|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4898) {
   HEAP32[$__parent$i2221>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4911 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2229 = $2;$4921 = $4911;
   label = 2084;
  } else {
   $__nd$0$i$i2223 = $4899;
   while(1) {
    $4900 = ((($__nd$0$i$i2223)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4901 = HEAP32[$4900>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4902 = ($4901|0)>(1073741983); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4902) {
     $4903 = HEAP32[$__nd$0$i$i2223>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4904 = ($4903|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4904) {
      $$lcssa7008 = $__nd$0$i$i2223;$__nd$0$i$i2223$lcssa7006 = $__nd$0$i$i2223;
      label = 2077;
      break;
     } else {
      $$sink$i$i2224 = $4903;
     }
    } else {
     $4906 = ($4901|0)<(1073741983); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4906)) {
      $__nd$0$i$i2223$lcssa = $__nd$0$i$i2223;
      label = 2083;
      break;
     }
     $4907 = ((($__nd$0$i$i2223)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4908 = HEAP32[$4907>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4909 = ($4908|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4909) {
      $$lcssa7007 = $4907;$__nd$0$i$i2223$lcssa7005 = $__nd$0$i$i2223;
      label = 2081;
      break;
     } else {
      $$sink$i$i2224 = $4908;
     }
    }
    $__nd$0$i$i2223 = $$sink$i$i2224;
   }
   if ((label|0) == 2077) {
    HEAP32[$__parent$i2221>>2] = $__nd$0$i$i2223$lcssa7006; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4905 = $__nd$0$i$i2223$lcssa7006; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2229 = $$lcssa7008;$4921 = $4905;
    label = 2084;
    break;
   }
   else if ((label|0) == 2081) {
    HEAP32[$__parent$i2221>>2] = $__nd$0$i$i2223$lcssa7005; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4910 = $__nd$0$i$i2223$lcssa7005; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2229 = $$lcssa7007;$4921 = $4910;
    label = 2084;
    break;
   }
   else if ((label|0) == 2083) {
    HEAP32[$__parent$i2221>>2] = $__nd$0$i$i2223$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4912 = $__nd$0$i$i2223$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4913 = ($__nd$0$i$i2223$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4913) {
     $$0$i7$i2229 = $__parent$i2221;$4921 = $4912;
     label = 2084;
     break;
    } else {
     $4932 = $4899;$__r$0$i2233 = $__nd$0$i$i2223$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2084) {
  __THREW__ = 0;
  $4914 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4915 = __THREW__; __THREW__ = 0;
  $4916 = $4915&1;
  if ($4916) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4917 = ((($4914)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4917>>2] = 1073741983; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4918 = ((($4914)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4918>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4914>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4919 = ((($4914)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4919>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4920 = ((($4914)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4920>>2] = $4921; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2229>>2] = $4914; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4922 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4923 = HEAP32[$4922>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4924 = ($4923|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4924) {
   $4927 = $4914;
  } else {
   $4925 = $4923; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4925; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2231 = HEAP32[$$0$i7$i2229>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4927 = $$pre$i$i2231;
  }
  $4926 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4926,$4927); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4928 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4929 = (($4928) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4929; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4524 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4932 = $$pre4524;$__r$0$i2233 = $4914;
 }
 $4930 = ((($__r$0$i2233)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4930>>2] = 149; //@line 767 "modules/event/sdl/Event.cpp"
 $4931 = ($4932|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4931) {
   HEAP32[$__parent$i2236>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4944 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2244 = $2;$4954 = $4944;
   label = 2098;
  } else {
   $__nd$0$i$i2238 = $4932;
   while(1) {
    $4933 = ((($__nd$0$i$i2238)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4934 = HEAP32[$4933>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4935 = ($4934|0)>(1073741984); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4935) {
     $4936 = HEAP32[$__nd$0$i$i2238>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4937 = ($4936|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4937) {
      $$lcssa7004 = $__nd$0$i$i2238;$__nd$0$i$i2238$lcssa7002 = $__nd$0$i$i2238;
      label = 2091;
      break;
     } else {
      $$sink$i$i2239 = $4936;
     }
    } else {
     $4939 = ($4934|0)<(1073741984); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4939)) {
      $__nd$0$i$i2238$lcssa = $__nd$0$i$i2238;
      label = 2097;
      break;
     }
     $4940 = ((($__nd$0$i$i2238)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4941 = HEAP32[$4940>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4942 = ($4941|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4942) {
      $$lcssa7003 = $4940;$__nd$0$i$i2238$lcssa7001 = $__nd$0$i$i2238;
      label = 2095;
      break;
     } else {
      $$sink$i$i2239 = $4941;
     }
    }
    $__nd$0$i$i2238 = $$sink$i$i2239;
   }
   if ((label|0) == 2091) {
    HEAP32[$__parent$i2236>>2] = $__nd$0$i$i2238$lcssa7002; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4938 = $__nd$0$i$i2238$lcssa7002; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2244 = $$lcssa7004;$4954 = $4938;
    label = 2098;
    break;
   }
   else if ((label|0) == 2095) {
    HEAP32[$__parent$i2236>>2] = $__nd$0$i$i2238$lcssa7001; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4943 = $__nd$0$i$i2238$lcssa7001; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2244 = $$lcssa7003;$4954 = $4943;
    label = 2098;
    break;
   }
   else if ((label|0) == 2097) {
    HEAP32[$__parent$i2236>>2] = $__nd$0$i$i2238$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4945 = $__nd$0$i$i2238$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4946 = ($__nd$0$i$i2238$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4946) {
     $$0$i7$i2244 = $__parent$i2236;$4954 = $4945;
     label = 2098;
     break;
    } else {
     $4965 = $4932;$__r$0$i2248 = $__nd$0$i$i2238$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2098) {
  __THREW__ = 0;
  $4947 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4948 = __THREW__; __THREW__ = 0;
  $4949 = $4948&1;
  if ($4949) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4950 = ((($4947)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4950>>2] = 1073741984; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4951 = ((($4947)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4951>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4947>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4952 = ((($4947)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4952>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4953 = ((($4947)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4953>>2] = $4954; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2244>>2] = $4947; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4955 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4956 = HEAP32[$4955>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4957 = ($4956|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4957) {
   $4960 = $4947;
  } else {
   $4958 = $4956; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4958; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2246 = HEAP32[$$0$i7$i2244>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4960 = $$pre$i$i2246;
  }
  $4959 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4959,$4960); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4961 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4962 = (($4961) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4962; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4525 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4965 = $$pre4525;$__r$0$i2248 = $4947;
 }
 $4963 = ((($__r$0$i2248)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4963>>2] = 150; //@line 768 "modules/event/sdl/Event.cpp"
 $4964 = ($4965|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4964) {
   HEAP32[$__parent$i2251>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $4977 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2259 = $2;$4987 = $4977;
   label = 2112;
  } else {
   $__nd$0$i$i2253 = $4965;
   while(1) {
    $4966 = ((($__nd$0$i$i2253)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4967 = HEAP32[$4966>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $4968 = ($4967|0)>(1073741985); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($4968) {
     $4969 = HEAP32[$__nd$0$i$i2253>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4970 = ($4969|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4970) {
      $$lcssa7000 = $__nd$0$i$i2253;$__nd$0$i$i2253$lcssa6998 = $__nd$0$i$i2253;
      label = 2105;
      break;
     } else {
      $$sink$i$i2254 = $4969;
     }
    } else {
     $4972 = ($4967|0)<(1073741985); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($4972)) {
      $__nd$0$i$i2253$lcssa = $__nd$0$i$i2253;
      label = 2111;
      break;
     }
     $4973 = ((($__nd$0$i$i2253)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4974 = HEAP32[$4973>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $4975 = ($4974|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($4975) {
      $$lcssa6999 = $4973;$__nd$0$i$i2253$lcssa6997 = $__nd$0$i$i2253;
      label = 2109;
      break;
     } else {
      $$sink$i$i2254 = $4974;
     }
    }
    $__nd$0$i$i2253 = $$sink$i$i2254;
   }
   if ((label|0) == 2105) {
    HEAP32[$__parent$i2251>>2] = $__nd$0$i$i2253$lcssa6998; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4971 = $__nd$0$i$i2253$lcssa6998; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2259 = $$lcssa7000;$4987 = $4971;
    label = 2112;
    break;
   }
   else if ((label|0) == 2109) {
    HEAP32[$__parent$i2251>>2] = $__nd$0$i$i2253$lcssa6997; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4976 = $__nd$0$i$i2253$lcssa6997; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2259 = $$lcssa6999;$4987 = $4976;
    label = 2112;
    break;
   }
   else if ((label|0) == 2111) {
    HEAP32[$__parent$i2251>>2] = $__nd$0$i$i2253$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4978 = $__nd$0$i$i2253$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4979 = ($__nd$0$i$i2253$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($4979) {
     $$0$i7$i2259 = $__parent$i2251;$4987 = $4978;
     label = 2112;
     break;
    } else {
     $4998 = $4965;$__r$0$i2263 = $__nd$0$i$i2253$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2112) {
  __THREW__ = 0;
  $4980 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $4981 = __THREW__; __THREW__ = 0;
  $4982 = $4981&1;
  if ($4982) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $4983 = ((($4980)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4983>>2] = 1073741985; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $4984 = ((($4980)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$4984>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$4980>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4985 = ((($4980)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4985>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4986 = ((($4980)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$4986>>2] = $4987; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2259>>2] = $4980; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4988 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4989 = HEAP32[$4988>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4990 = ($4989|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($4990) {
   $4993 = $4980;
  } else {
   $4991 = $4989; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $4991; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2261 = HEAP32[$$0$i7$i2259>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $4993 = $$pre$i$i2261;
  }
  $4992 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($4992,$4993); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4994 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4995 = (($4994) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $4995; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4526 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4998 = $$pre4526;$__r$0$i2263 = $4980;
 }
 $4996 = ((($__r$0$i2263)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$4996>>2] = 151; //@line 769 "modules/event/sdl/Event.cpp"
 $4997 = ($4998|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($4997) {
   HEAP32[$__parent$i2266>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5010 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2274 = $2;$5020 = $5010;
   label = 2126;
  } else {
   $__nd$0$i$i2268 = $4998;
   while(1) {
    $4999 = ((($__nd$0$i$i2268)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5000 = HEAP32[$4999>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5001 = ($5000|0)>(1073741986); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5001) {
     $5002 = HEAP32[$__nd$0$i$i2268>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5003 = ($5002|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5003) {
      $$lcssa6996 = $__nd$0$i$i2268;$__nd$0$i$i2268$lcssa6994 = $__nd$0$i$i2268;
      label = 2119;
      break;
     } else {
      $$sink$i$i2269 = $5002;
     }
    } else {
     $5005 = ($5000|0)<(1073741986); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5005)) {
      $__nd$0$i$i2268$lcssa = $__nd$0$i$i2268;
      label = 2125;
      break;
     }
     $5006 = ((($__nd$0$i$i2268)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5007 = HEAP32[$5006>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5008 = ($5007|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5008) {
      $$lcssa6995 = $5006;$__nd$0$i$i2268$lcssa6993 = $__nd$0$i$i2268;
      label = 2123;
      break;
     } else {
      $$sink$i$i2269 = $5007;
     }
    }
    $__nd$0$i$i2268 = $$sink$i$i2269;
   }
   if ((label|0) == 2119) {
    HEAP32[$__parent$i2266>>2] = $__nd$0$i$i2268$lcssa6994; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5004 = $__nd$0$i$i2268$lcssa6994; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2274 = $$lcssa6996;$5020 = $5004;
    label = 2126;
    break;
   }
   else if ((label|0) == 2123) {
    HEAP32[$__parent$i2266>>2] = $__nd$0$i$i2268$lcssa6993; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5009 = $__nd$0$i$i2268$lcssa6993; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2274 = $$lcssa6995;$5020 = $5009;
    label = 2126;
    break;
   }
   else if ((label|0) == 2125) {
    HEAP32[$__parent$i2266>>2] = $__nd$0$i$i2268$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5011 = $__nd$0$i$i2268$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5012 = ($__nd$0$i$i2268$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5012) {
     $$0$i7$i2274 = $__parent$i2266;$5020 = $5011;
     label = 2126;
     break;
    } else {
     $5031 = $4998;$__r$0$i2278 = $__nd$0$i$i2268$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2126) {
  __THREW__ = 0;
  $5013 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5014 = __THREW__; __THREW__ = 0;
  $5015 = $5014&1;
  if ($5015) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5016 = ((($5013)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5016>>2] = 1073741986; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5017 = ((($5013)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5017>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5013>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5018 = ((($5013)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5018>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5019 = ((($5013)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5019>>2] = $5020; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2274>>2] = $5013; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5021 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5022 = HEAP32[$5021>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5023 = ($5022|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5023) {
   $5026 = $5013;
  } else {
   $5024 = $5022; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5024; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2276 = HEAP32[$$0$i7$i2274>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5026 = $$pre$i$i2276;
  }
  $5025 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5025,$5026); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5027 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5028 = (($5027) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5028; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4527 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5031 = $$pre4527;$__r$0$i2278 = $5013;
 }
 $5029 = ((($__r$0$i2278)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5029>>2] = 152; //@line 770 "modules/event/sdl/Event.cpp"
 $5030 = ($5031|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5030) {
   HEAP32[$__parent$i2281>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5043 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2289 = $2;$5053 = $5043;
   label = 2140;
  } else {
   $__nd$0$i$i2283 = $5031;
   while(1) {
    $5032 = ((($__nd$0$i$i2283)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5033 = HEAP32[$5032>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5034 = ($5033|0)>(1073742002); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5034) {
     $5035 = HEAP32[$__nd$0$i$i2283>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5036 = ($5035|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5036) {
      $$lcssa6992 = $__nd$0$i$i2283;$__nd$0$i$i2283$lcssa6990 = $__nd$0$i$i2283;
      label = 2133;
      break;
     } else {
      $$sink$i$i2284 = $5035;
     }
    } else {
     $5038 = ($5033|0)<(1073742002); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5038)) {
      $__nd$0$i$i2283$lcssa = $__nd$0$i$i2283;
      label = 2139;
      break;
     }
     $5039 = ((($__nd$0$i$i2283)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5040 = HEAP32[$5039>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5041 = ($5040|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5041) {
      $$lcssa6991 = $5039;$__nd$0$i$i2283$lcssa6989 = $__nd$0$i$i2283;
      label = 2137;
      break;
     } else {
      $$sink$i$i2284 = $5040;
     }
    }
    $__nd$0$i$i2283 = $$sink$i$i2284;
   }
   if ((label|0) == 2133) {
    HEAP32[$__parent$i2281>>2] = $__nd$0$i$i2283$lcssa6990; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5037 = $__nd$0$i$i2283$lcssa6990; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2289 = $$lcssa6992;$5053 = $5037;
    label = 2140;
    break;
   }
   else if ((label|0) == 2137) {
    HEAP32[$__parent$i2281>>2] = $__nd$0$i$i2283$lcssa6989; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5042 = $__nd$0$i$i2283$lcssa6989; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2289 = $$lcssa6991;$5053 = $5042;
    label = 2140;
    break;
   }
   else if ((label|0) == 2139) {
    HEAP32[$__parent$i2281>>2] = $__nd$0$i$i2283$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5044 = $__nd$0$i$i2283$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5045 = ($__nd$0$i$i2283$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5045) {
     $$0$i7$i2289 = $__parent$i2281;$5053 = $5044;
     label = 2140;
     break;
    } else {
     $5064 = $5031;$__r$0$i2293 = $__nd$0$i$i2283$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2140) {
  __THREW__ = 0;
  $5046 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5047 = __THREW__; __THREW__ = 0;
  $5048 = $5047&1;
  if ($5048) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5049 = ((($5046)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5049>>2] = 1073742002; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5050 = ((($5046)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5050>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5046>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5051 = ((($5046)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5051>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5052 = ((($5046)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5052>>2] = $5053; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2289>>2] = $5046; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5054 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5055 = HEAP32[$5054>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5056 = ($5055|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5056) {
   $5059 = $5046;
  } else {
   $5057 = $5055; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5057; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2291 = HEAP32[$$0$i7$i2289>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5059 = $$pre$i$i2291;
  }
  $5058 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5058,$5059); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5060 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5061 = (($5060) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5061; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4528 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5064 = $$pre4528;$__r$0$i2293 = $5046;
 }
 $5062 = ((($__r$0$i2293)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5062>>2] = 153; //@line 772 "modules/event/sdl/Event.cpp"
 $5063 = ($5064|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5063) {
   HEAP32[$__parent$i2296>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5076 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2304 = $2;$5086 = $5076;
   label = 2154;
  } else {
   $__nd$0$i$i2298 = $5064;
   while(1) {
    $5065 = ((($__nd$0$i$i2298)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5066 = HEAP32[$5065>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5067 = ($5066|0)>(1073742003); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5067) {
     $5068 = HEAP32[$__nd$0$i$i2298>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5069 = ($5068|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5069) {
      $$lcssa6988 = $__nd$0$i$i2298;$__nd$0$i$i2298$lcssa6986 = $__nd$0$i$i2298;
      label = 2147;
      break;
     } else {
      $$sink$i$i2299 = $5068;
     }
    } else {
     $5071 = ($5066|0)<(1073742003); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5071)) {
      $__nd$0$i$i2298$lcssa = $__nd$0$i$i2298;
      label = 2153;
      break;
     }
     $5072 = ((($__nd$0$i$i2298)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5073 = HEAP32[$5072>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5074 = ($5073|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5074) {
      $$lcssa6987 = $5072;$__nd$0$i$i2298$lcssa6985 = $__nd$0$i$i2298;
      label = 2151;
      break;
     } else {
      $$sink$i$i2299 = $5073;
     }
    }
    $__nd$0$i$i2298 = $$sink$i$i2299;
   }
   if ((label|0) == 2147) {
    HEAP32[$__parent$i2296>>2] = $__nd$0$i$i2298$lcssa6986; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5070 = $__nd$0$i$i2298$lcssa6986; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2304 = $$lcssa6988;$5086 = $5070;
    label = 2154;
    break;
   }
   else if ((label|0) == 2151) {
    HEAP32[$__parent$i2296>>2] = $__nd$0$i$i2298$lcssa6985; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5075 = $__nd$0$i$i2298$lcssa6985; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2304 = $$lcssa6987;$5086 = $5075;
    label = 2154;
    break;
   }
   else if ((label|0) == 2153) {
    HEAP32[$__parent$i2296>>2] = $__nd$0$i$i2298$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5077 = $__nd$0$i$i2298$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5078 = ($__nd$0$i$i2298$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5078) {
     $$0$i7$i2304 = $__parent$i2296;$5086 = $5077;
     label = 2154;
     break;
    } else {
     $5097 = $5064;$__r$0$i2308 = $__nd$0$i$i2298$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2154) {
  __THREW__ = 0;
  $5079 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5080 = __THREW__; __THREW__ = 0;
  $5081 = $5080&1;
  if ($5081) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5082 = ((($5079)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5082>>2] = 1073742003; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5083 = ((($5079)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5083>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5079>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5084 = ((($5079)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5084>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5085 = ((($5079)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5085>>2] = $5086; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2304>>2] = $5079; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5087 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5088 = HEAP32[$5087>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5089 = ($5088|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5089) {
   $5092 = $5079;
  } else {
   $5090 = $5088; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5090; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2306 = HEAP32[$$0$i7$i2304>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5092 = $$pre$i$i2306;
  }
  $5091 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5091,$5092); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5093 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5094 = (($5093) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5094; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4529 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5097 = $$pre4529;$__r$0$i2308 = $5079;
 }
 $5095 = ((($__r$0$i2308)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5095>>2] = 154; //@line 773 "modules/event/sdl/Event.cpp"
 $5096 = ($5097|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5096) {
   HEAP32[$__parent$i2311>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5109 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2319 = $2;$5119 = $5109;
   label = 2168;
  } else {
   $__nd$0$i$i2313 = $5097;
   while(1) {
    $5098 = ((($__nd$0$i$i2313)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5099 = HEAP32[$5098>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5100 = ($5099|0)>(1073742004); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5100) {
     $5101 = HEAP32[$__nd$0$i$i2313>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5102 = ($5101|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5102) {
      $$lcssa6984 = $__nd$0$i$i2313;$__nd$0$i$i2313$lcssa6982 = $__nd$0$i$i2313;
      label = 2161;
      break;
     } else {
      $$sink$i$i2314 = $5101;
     }
    } else {
     $5104 = ($5099|0)<(1073742004); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5104)) {
      $__nd$0$i$i2313$lcssa = $__nd$0$i$i2313;
      label = 2167;
      break;
     }
     $5105 = ((($__nd$0$i$i2313)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5106 = HEAP32[$5105>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5107 = ($5106|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5107) {
      $$lcssa6983 = $5105;$__nd$0$i$i2313$lcssa6981 = $__nd$0$i$i2313;
      label = 2165;
      break;
     } else {
      $$sink$i$i2314 = $5106;
     }
    }
    $__nd$0$i$i2313 = $$sink$i$i2314;
   }
   if ((label|0) == 2161) {
    HEAP32[$__parent$i2311>>2] = $__nd$0$i$i2313$lcssa6982; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5103 = $__nd$0$i$i2313$lcssa6982; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2319 = $$lcssa6984;$5119 = $5103;
    label = 2168;
    break;
   }
   else if ((label|0) == 2165) {
    HEAP32[$__parent$i2311>>2] = $__nd$0$i$i2313$lcssa6981; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5108 = $__nd$0$i$i2313$lcssa6981; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2319 = $$lcssa6983;$5119 = $5108;
    label = 2168;
    break;
   }
   else if ((label|0) == 2167) {
    HEAP32[$__parent$i2311>>2] = $__nd$0$i$i2313$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5110 = $__nd$0$i$i2313$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5111 = ($__nd$0$i$i2313$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5111) {
     $$0$i7$i2319 = $__parent$i2311;$5119 = $5110;
     label = 2168;
     break;
    } else {
     $5130 = $5097;$__r$0$i2323 = $__nd$0$i$i2313$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2168) {
  __THREW__ = 0;
  $5112 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5113 = __THREW__; __THREW__ = 0;
  $5114 = $5113&1;
  if ($5114) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5115 = ((($5112)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5115>>2] = 1073742004; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5116 = ((($5112)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5116>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5112>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5117 = ((($5112)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5117>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5118 = ((($5112)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5118>>2] = $5119; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2319>>2] = $5112; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5120 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5121 = HEAP32[$5120>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5122 = ($5121|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5122) {
   $5125 = $5112;
  } else {
   $5123 = $5121; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5123; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2321 = HEAP32[$$0$i7$i2319>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5125 = $$pre$i$i2321;
  }
  $5124 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5124,$5125); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5126 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5127 = (($5126) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5127; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4530 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5130 = $$pre4530;$__r$0$i2323 = $5112;
 }
 $5128 = ((($__r$0$i2323)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5128>>2] = 155; //@line 774 "modules/event/sdl/Event.cpp"
 $5129 = ($5130|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5129) {
   HEAP32[$__parent$i2326>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5142 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2334 = $2;$5152 = $5142;
   label = 2182;
  } else {
   $__nd$0$i$i2328 = $5130;
   while(1) {
    $5131 = ((($__nd$0$i$i2328)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5132 = HEAP32[$5131>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5133 = ($5132|0)>(1073742005); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5133) {
     $5134 = HEAP32[$__nd$0$i$i2328>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5135 = ($5134|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5135) {
      $$lcssa6980 = $__nd$0$i$i2328;$__nd$0$i$i2328$lcssa6978 = $__nd$0$i$i2328;
      label = 2175;
      break;
     } else {
      $$sink$i$i2329 = $5134;
     }
    } else {
     $5137 = ($5132|0)<(1073742005); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5137)) {
      $__nd$0$i$i2328$lcssa = $__nd$0$i$i2328;
      label = 2181;
      break;
     }
     $5138 = ((($__nd$0$i$i2328)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5139 = HEAP32[$5138>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5140 = ($5139|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5140) {
      $$lcssa6979 = $5138;$__nd$0$i$i2328$lcssa6977 = $__nd$0$i$i2328;
      label = 2179;
      break;
     } else {
      $$sink$i$i2329 = $5139;
     }
    }
    $__nd$0$i$i2328 = $$sink$i$i2329;
   }
   if ((label|0) == 2175) {
    HEAP32[$__parent$i2326>>2] = $__nd$0$i$i2328$lcssa6978; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5136 = $__nd$0$i$i2328$lcssa6978; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2334 = $$lcssa6980;$5152 = $5136;
    label = 2182;
    break;
   }
   else if ((label|0) == 2179) {
    HEAP32[$__parent$i2326>>2] = $__nd$0$i$i2328$lcssa6977; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5141 = $__nd$0$i$i2328$lcssa6977; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2334 = $$lcssa6979;$5152 = $5141;
    label = 2182;
    break;
   }
   else if ((label|0) == 2181) {
    HEAP32[$__parent$i2326>>2] = $__nd$0$i$i2328$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5143 = $__nd$0$i$i2328$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5144 = ($__nd$0$i$i2328$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5144) {
     $$0$i7$i2334 = $__parent$i2326;$5152 = $5143;
     label = 2182;
     break;
    } else {
     $5163 = $5130;$__r$0$i2338 = $__nd$0$i$i2328$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2182) {
  __THREW__ = 0;
  $5145 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5146 = __THREW__; __THREW__ = 0;
  $5147 = $5146&1;
  if ($5147) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5148 = ((($5145)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5148>>2] = 1073742005; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5149 = ((($5145)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5149>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5145>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5150 = ((($5145)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5150>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5151 = ((($5145)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5151>>2] = $5152; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2334>>2] = $5145; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5153 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5154 = HEAP32[$5153>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5155 = ($5154|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5155) {
   $5158 = $5145;
  } else {
   $5156 = $5154; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5156; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2336 = HEAP32[$$0$i7$i2334>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5158 = $$pre$i$i2336;
  }
  $5157 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5157,$5158); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5159 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5160 = (($5159) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5160; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4531 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5163 = $$pre4531;$__r$0$i2338 = $5145;
 }
 $5161 = ((($__r$0$i2338)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5161>>2] = 156; //@line 775 "modules/event/sdl/Event.cpp"
 $5162 = ($5163|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5162) {
   HEAP32[$__parent$i2341>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5175 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2349 = $2;$5185 = $5175;
   label = 2196;
  } else {
   $__nd$0$i$i2343 = $5163;
   while(1) {
    $5164 = ((($__nd$0$i$i2343)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5165 = HEAP32[$5164>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5166 = ($5165|0)>(1073742048); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5166) {
     $5167 = HEAP32[$__nd$0$i$i2343>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5168 = ($5167|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5168) {
      $$lcssa6976 = $__nd$0$i$i2343;$__nd$0$i$i2343$lcssa6974 = $__nd$0$i$i2343;
      label = 2189;
      break;
     } else {
      $$sink$i$i2344 = $5167;
     }
    } else {
     $5170 = ($5165|0)<(1073742048); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5170)) {
      $__nd$0$i$i2343$lcssa = $__nd$0$i$i2343;
      label = 2195;
      break;
     }
     $5171 = ((($__nd$0$i$i2343)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5172 = HEAP32[$5171>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5173 = ($5172|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5173) {
      $$lcssa6975 = $5171;$__nd$0$i$i2343$lcssa6973 = $__nd$0$i$i2343;
      label = 2193;
      break;
     } else {
      $$sink$i$i2344 = $5172;
     }
    }
    $__nd$0$i$i2343 = $$sink$i$i2344;
   }
   if ((label|0) == 2189) {
    HEAP32[$__parent$i2341>>2] = $__nd$0$i$i2343$lcssa6974; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5169 = $__nd$0$i$i2343$lcssa6974; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2349 = $$lcssa6976;$5185 = $5169;
    label = 2196;
    break;
   }
   else if ((label|0) == 2193) {
    HEAP32[$__parent$i2341>>2] = $__nd$0$i$i2343$lcssa6973; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5174 = $__nd$0$i$i2343$lcssa6973; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2349 = $$lcssa6975;$5185 = $5174;
    label = 2196;
    break;
   }
   else if ((label|0) == 2195) {
    HEAP32[$__parent$i2341>>2] = $__nd$0$i$i2343$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5176 = $__nd$0$i$i2343$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5177 = ($__nd$0$i$i2343$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5177) {
     $$0$i7$i2349 = $__parent$i2341;$5185 = $5176;
     label = 2196;
     break;
    } else {
     $5196 = $5163;$__r$0$i2353 = $__nd$0$i$i2343$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2196) {
  __THREW__ = 0;
  $5178 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5179 = __THREW__; __THREW__ = 0;
  $5180 = $5179&1;
  if ($5180) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5181 = ((($5178)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5181>>2] = 1073742048; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5182 = ((($5178)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5182>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5178>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5183 = ((($5178)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5183>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5184 = ((($5178)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5184>>2] = $5185; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2349>>2] = $5178; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5186 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5187 = HEAP32[$5186>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5188 = ($5187|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5188) {
   $5191 = $5178;
  } else {
   $5189 = $5187; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5189; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2351 = HEAP32[$$0$i7$i2349>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5191 = $$pre$i$i2351;
  }
  $5190 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5190,$5191); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5192 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5193 = (($5192) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5193; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4532 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5196 = $$pre4532;$__r$0$i2353 = $5178;
 }
 $5194 = ((($__r$0$i2353)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5194>>2] = 157; //@line 777 "modules/event/sdl/Event.cpp"
 $5195 = ($5196|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5195) {
   HEAP32[$__parent$i2356>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5208 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2364 = $2;$5218 = $5208;
   label = 2210;
  } else {
   $__nd$0$i$i2358 = $5196;
   while(1) {
    $5197 = ((($__nd$0$i$i2358)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5198 = HEAP32[$5197>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5199 = ($5198|0)>(1073742049); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5199) {
     $5200 = HEAP32[$__nd$0$i$i2358>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5201 = ($5200|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5201) {
      $$lcssa6972 = $__nd$0$i$i2358;$__nd$0$i$i2358$lcssa6970 = $__nd$0$i$i2358;
      label = 2203;
      break;
     } else {
      $$sink$i$i2359 = $5200;
     }
    } else {
     $5203 = ($5198|0)<(1073742049); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5203)) {
      $__nd$0$i$i2358$lcssa = $__nd$0$i$i2358;
      label = 2209;
      break;
     }
     $5204 = ((($__nd$0$i$i2358)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5205 = HEAP32[$5204>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5206 = ($5205|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5206) {
      $$lcssa6971 = $5204;$__nd$0$i$i2358$lcssa6969 = $__nd$0$i$i2358;
      label = 2207;
      break;
     } else {
      $$sink$i$i2359 = $5205;
     }
    }
    $__nd$0$i$i2358 = $$sink$i$i2359;
   }
   if ((label|0) == 2203) {
    HEAP32[$__parent$i2356>>2] = $__nd$0$i$i2358$lcssa6970; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5202 = $__nd$0$i$i2358$lcssa6970; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2364 = $$lcssa6972;$5218 = $5202;
    label = 2210;
    break;
   }
   else if ((label|0) == 2207) {
    HEAP32[$__parent$i2356>>2] = $__nd$0$i$i2358$lcssa6969; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5207 = $__nd$0$i$i2358$lcssa6969; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2364 = $$lcssa6971;$5218 = $5207;
    label = 2210;
    break;
   }
   else if ((label|0) == 2209) {
    HEAP32[$__parent$i2356>>2] = $__nd$0$i$i2358$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5209 = $__nd$0$i$i2358$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5210 = ($__nd$0$i$i2358$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5210) {
     $$0$i7$i2364 = $__parent$i2356;$5218 = $5209;
     label = 2210;
     break;
    } else {
     $5229 = $5196;$__r$0$i2368 = $__nd$0$i$i2358$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2210) {
  __THREW__ = 0;
  $5211 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5212 = __THREW__; __THREW__ = 0;
  $5213 = $5212&1;
  if ($5213) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5214 = ((($5211)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5214>>2] = 1073742049; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5215 = ((($5211)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5215>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5211>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5216 = ((($5211)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5216>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5217 = ((($5211)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5217>>2] = $5218; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2364>>2] = $5211; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5219 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5220 = HEAP32[$5219>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5221 = ($5220|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5221) {
   $5224 = $5211;
  } else {
   $5222 = $5220; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5222; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2366 = HEAP32[$$0$i7$i2364>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5224 = $$pre$i$i2366;
  }
  $5223 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5223,$5224); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5225 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5226 = (($5225) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5226; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4533 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5229 = $$pre4533;$__r$0$i2368 = $5211;
 }
 $5227 = ((($__r$0$i2368)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5227>>2] = 158; //@line 778 "modules/event/sdl/Event.cpp"
 $5228 = ($5229|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5228) {
   HEAP32[$__parent$i2371>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5241 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2379 = $2;$5251 = $5241;
   label = 2224;
  } else {
   $__nd$0$i$i2373 = $5229;
   while(1) {
    $5230 = ((($__nd$0$i$i2373)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5231 = HEAP32[$5230>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5232 = ($5231|0)>(1073742050); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5232) {
     $5233 = HEAP32[$__nd$0$i$i2373>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5234 = ($5233|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5234) {
      $$lcssa6968 = $__nd$0$i$i2373;$__nd$0$i$i2373$lcssa6966 = $__nd$0$i$i2373;
      label = 2217;
      break;
     } else {
      $$sink$i$i2374 = $5233;
     }
    } else {
     $5236 = ($5231|0)<(1073742050); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5236)) {
      $__nd$0$i$i2373$lcssa = $__nd$0$i$i2373;
      label = 2223;
      break;
     }
     $5237 = ((($__nd$0$i$i2373)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5238 = HEAP32[$5237>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5239 = ($5238|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5239) {
      $$lcssa6967 = $5237;$__nd$0$i$i2373$lcssa6965 = $__nd$0$i$i2373;
      label = 2221;
      break;
     } else {
      $$sink$i$i2374 = $5238;
     }
    }
    $__nd$0$i$i2373 = $$sink$i$i2374;
   }
   if ((label|0) == 2217) {
    HEAP32[$__parent$i2371>>2] = $__nd$0$i$i2373$lcssa6966; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5235 = $__nd$0$i$i2373$lcssa6966; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2379 = $$lcssa6968;$5251 = $5235;
    label = 2224;
    break;
   }
   else if ((label|0) == 2221) {
    HEAP32[$__parent$i2371>>2] = $__nd$0$i$i2373$lcssa6965; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5240 = $__nd$0$i$i2373$lcssa6965; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2379 = $$lcssa6967;$5251 = $5240;
    label = 2224;
    break;
   }
   else if ((label|0) == 2223) {
    HEAP32[$__parent$i2371>>2] = $__nd$0$i$i2373$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5242 = $__nd$0$i$i2373$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5243 = ($__nd$0$i$i2373$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5243) {
     $$0$i7$i2379 = $__parent$i2371;$5251 = $5242;
     label = 2224;
     break;
    } else {
     $5262 = $5229;$__r$0$i2383 = $__nd$0$i$i2373$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2224) {
  __THREW__ = 0;
  $5244 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5245 = __THREW__; __THREW__ = 0;
  $5246 = $5245&1;
  if ($5246) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5247 = ((($5244)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5247>>2] = 1073742050; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5248 = ((($5244)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5248>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5244>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5249 = ((($5244)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5249>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5250 = ((($5244)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5250>>2] = $5251; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2379>>2] = $5244; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5252 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5253 = HEAP32[$5252>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5254 = ($5253|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5254) {
   $5257 = $5244;
  } else {
   $5255 = $5253; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2381 = HEAP32[$$0$i7$i2379>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5257 = $$pre$i$i2381;
  }
  $5256 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5256,$5257); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5258 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5259 = (($5258) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5259; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4534 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5262 = $$pre4534;$__r$0$i2383 = $5244;
 }
 $5260 = ((($__r$0$i2383)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5260>>2] = 159; //@line 779 "modules/event/sdl/Event.cpp"
 $5261 = ($5262|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5261) {
   HEAP32[$__parent$i2386>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5274 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2394 = $2;$5284 = $5274;
   label = 2238;
  } else {
   $__nd$0$i$i2388 = $5262;
   while(1) {
    $5263 = ((($__nd$0$i$i2388)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5264 = HEAP32[$5263>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5265 = ($5264|0)>(1073742051); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5265) {
     $5266 = HEAP32[$__nd$0$i$i2388>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5267 = ($5266|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5267) {
      $$lcssa6964 = $__nd$0$i$i2388;$__nd$0$i$i2388$lcssa6962 = $__nd$0$i$i2388;
      label = 2231;
      break;
     } else {
      $$sink$i$i2389 = $5266;
     }
    } else {
     $5269 = ($5264|0)<(1073742051); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5269)) {
      $__nd$0$i$i2388$lcssa = $__nd$0$i$i2388;
      label = 2237;
      break;
     }
     $5270 = ((($__nd$0$i$i2388)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5271 = HEAP32[$5270>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5272 = ($5271|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5272) {
      $$lcssa6963 = $5270;$__nd$0$i$i2388$lcssa6961 = $__nd$0$i$i2388;
      label = 2235;
      break;
     } else {
      $$sink$i$i2389 = $5271;
     }
    }
    $__nd$0$i$i2388 = $$sink$i$i2389;
   }
   if ((label|0) == 2231) {
    HEAP32[$__parent$i2386>>2] = $__nd$0$i$i2388$lcssa6962; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5268 = $__nd$0$i$i2388$lcssa6962; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2394 = $$lcssa6964;$5284 = $5268;
    label = 2238;
    break;
   }
   else if ((label|0) == 2235) {
    HEAP32[$__parent$i2386>>2] = $__nd$0$i$i2388$lcssa6961; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5273 = $__nd$0$i$i2388$lcssa6961; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2394 = $$lcssa6963;$5284 = $5273;
    label = 2238;
    break;
   }
   else if ((label|0) == 2237) {
    HEAP32[$__parent$i2386>>2] = $__nd$0$i$i2388$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5275 = $__nd$0$i$i2388$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5276 = ($__nd$0$i$i2388$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5276) {
     $$0$i7$i2394 = $__parent$i2386;$5284 = $5275;
     label = 2238;
     break;
    } else {
     $5295 = $5262;$__r$0$i2398 = $__nd$0$i$i2388$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2238) {
  __THREW__ = 0;
  $5277 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5278 = __THREW__; __THREW__ = 0;
  $5279 = $5278&1;
  if ($5279) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5280 = ((($5277)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5280>>2] = 1073742051; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5281 = ((($5277)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5281>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5277>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5282 = ((($5277)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5282>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5283 = ((($5277)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5283>>2] = $5284; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2394>>2] = $5277; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5285 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5286 = HEAP32[$5285>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5287 = ($5286|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5287) {
   $5290 = $5277;
  } else {
   $5288 = $5286; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5288; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2396 = HEAP32[$$0$i7$i2394>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5290 = $$pre$i$i2396;
  }
  $5289 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5289,$5290); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5291 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5292 = (($5291) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5292; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4535 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5295 = $$pre4535;$__r$0$i2398 = $5277;
 }
 $5293 = ((($__r$0$i2398)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5293>>2] = 160; //@line 780 "modules/event/sdl/Event.cpp"
 $5294 = ($5295|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5294) {
   HEAP32[$__parent$i2401>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5307 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2409 = $2;$5317 = $5307;
   label = 2252;
  } else {
   $__nd$0$i$i2403 = $5295;
   while(1) {
    $5296 = ((($__nd$0$i$i2403)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5297 = HEAP32[$5296>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5298 = ($5297|0)>(1073742052); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5298) {
     $5299 = HEAP32[$__nd$0$i$i2403>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5300 = ($5299|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5300) {
      $$lcssa6960 = $__nd$0$i$i2403;$__nd$0$i$i2403$lcssa6958 = $__nd$0$i$i2403;
      label = 2245;
      break;
     } else {
      $$sink$i$i2404 = $5299;
     }
    } else {
     $5302 = ($5297|0)<(1073742052); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5302)) {
      $__nd$0$i$i2403$lcssa = $__nd$0$i$i2403;
      label = 2251;
      break;
     }
     $5303 = ((($__nd$0$i$i2403)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5304 = HEAP32[$5303>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5305 = ($5304|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5305) {
      $$lcssa6959 = $5303;$__nd$0$i$i2403$lcssa6957 = $__nd$0$i$i2403;
      label = 2249;
      break;
     } else {
      $$sink$i$i2404 = $5304;
     }
    }
    $__nd$0$i$i2403 = $$sink$i$i2404;
   }
   if ((label|0) == 2245) {
    HEAP32[$__parent$i2401>>2] = $__nd$0$i$i2403$lcssa6958; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5301 = $__nd$0$i$i2403$lcssa6958; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2409 = $$lcssa6960;$5317 = $5301;
    label = 2252;
    break;
   }
   else if ((label|0) == 2249) {
    HEAP32[$__parent$i2401>>2] = $__nd$0$i$i2403$lcssa6957; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5306 = $__nd$0$i$i2403$lcssa6957; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2409 = $$lcssa6959;$5317 = $5306;
    label = 2252;
    break;
   }
   else if ((label|0) == 2251) {
    HEAP32[$__parent$i2401>>2] = $__nd$0$i$i2403$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5308 = $__nd$0$i$i2403$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5309 = ($__nd$0$i$i2403$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5309) {
     $$0$i7$i2409 = $__parent$i2401;$5317 = $5308;
     label = 2252;
     break;
    } else {
     $5328 = $5295;$__r$0$i2413 = $__nd$0$i$i2403$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2252) {
  __THREW__ = 0;
  $5310 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5311 = __THREW__; __THREW__ = 0;
  $5312 = $5311&1;
  if ($5312) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5313 = ((($5310)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5313>>2] = 1073742052; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5314 = ((($5310)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5314>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5310>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5315 = ((($5310)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5315>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5316 = ((($5310)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5316>>2] = $5317; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2409>>2] = $5310; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5318 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5319 = HEAP32[$5318>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5320 = ($5319|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5320) {
   $5323 = $5310;
  } else {
   $5321 = $5319; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5321; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2411 = HEAP32[$$0$i7$i2409>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5323 = $$pre$i$i2411;
  }
  $5322 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5322,$5323); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5324 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5325 = (($5324) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5325; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4536 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5328 = $$pre4536;$__r$0$i2413 = $5310;
 }
 $5326 = ((($__r$0$i2413)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5326>>2] = 161; //@line 781 "modules/event/sdl/Event.cpp"
 $5327 = ($5328|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5327) {
   HEAP32[$__parent$i2416>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5340 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2424 = $2;$5350 = $5340;
   label = 2266;
  } else {
   $__nd$0$i$i2418 = $5328;
   while(1) {
    $5329 = ((($__nd$0$i$i2418)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5330 = HEAP32[$5329>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5331 = ($5330|0)>(1073742053); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5331) {
     $5332 = HEAP32[$__nd$0$i$i2418>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5333 = ($5332|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5333) {
      $$lcssa6956 = $__nd$0$i$i2418;$__nd$0$i$i2418$lcssa6954 = $__nd$0$i$i2418;
      label = 2259;
      break;
     } else {
      $$sink$i$i2419 = $5332;
     }
    } else {
     $5335 = ($5330|0)<(1073742053); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5335)) {
      $__nd$0$i$i2418$lcssa = $__nd$0$i$i2418;
      label = 2265;
      break;
     }
     $5336 = ((($__nd$0$i$i2418)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5337 = HEAP32[$5336>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5338 = ($5337|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5338) {
      $$lcssa6955 = $5336;$__nd$0$i$i2418$lcssa6953 = $__nd$0$i$i2418;
      label = 2263;
      break;
     } else {
      $$sink$i$i2419 = $5337;
     }
    }
    $__nd$0$i$i2418 = $$sink$i$i2419;
   }
   if ((label|0) == 2259) {
    HEAP32[$__parent$i2416>>2] = $__nd$0$i$i2418$lcssa6954; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5334 = $__nd$0$i$i2418$lcssa6954; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2424 = $$lcssa6956;$5350 = $5334;
    label = 2266;
    break;
   }
   else if ((label|0) == 2263) {
    HEAP32[$__parent$i2416>>2] = $__nd$0$i$i2418$lcssa6953; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5339 = $__nd$0$i$i2418$lcssa6953; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2424 = $$lcssa6955;$5350 = $5339;
    label = 2266;
    break;
   }
   else if ((label|0) == 2265) {
    HEAP32[$__parent$i2416>>2] = $__nd$0$i$i2418$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5341 = $__nd$0$i$i2418$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5342 = ($__nd$0$i$i2418$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5342) {
     $$0$i7$i2424 = $__parent$i2416;$5350 = $5341;
     label = 2266;
     break;
    } else {
     $5361 = $5328;$__r$0$i2428 = $__nd$0$i$i2418$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2266) {
  __THREW__ = 0;
  $5343 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5344 = __THREW__; __THREW__ = 0;
  $5345 = $5344&1;
  if ($5345) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5346 = ((($5343)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5346>>2] = 1073742053; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5347 = ((($5343)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5347>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5343>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5348 = ((($5343)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5348>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5349 = ((($5343)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5349>>2] = $5350; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2424>>2] = $5343; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5351 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5352 = HEAP32[$5351>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5353 = ($5352|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5353) {
   $5356 = $5343;
  } else {
   $5354 = $5352; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5354; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2426 = HEAP32[$$0$i7$i2424>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5356 = $$pre$i$i2426;
  }
  $5355 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5355,$5356); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5357 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5358 = (($5357) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5358; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4537 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5361 = $$pre4537;$__r$0$i2428 = $5343;
 }
 $5359 = ((($__r$0$i2428)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5359>>2] = 162; //@line 782 "modules/event/sdl/Event.cpp"
 $5360 = ($5361|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5360) {
   HEAP32[$__parent$i2431>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5373 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2439 = $2;$5383 = $5373;
   label = 2280;
  } else {
   $__nd$0$i$i2433 = $5361;
   while(1) {
    $5362 = ((($__nd$0$i$i2433)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5363 = HEAP32[$5362>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5364 = ($5363|0)>(1073742054); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5364) {
     $5365 = HEAP32[$__nd$0$i$i2433>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5366 = ($5365|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5366) {
      $$lcssa6952 = $__nd$0$i$i2433;$__nd$0$i$i2433$lcssa6950 = $__nd$0$i$i2433;
      label = 2273;
      break;
     } else {
      $$sink$i$i2434 = $5365;
     }
    } else {
     $5368 = ($5363|0)<(1073742054); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5368)) {
      $__nd$0$i$i2433$lcssa = $__nd$0$i$i2433;
      label = 2279;
      break;
     }
     $5369 = ((($__nd$0$i$i2433)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5370 = HEAP32[$5369>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5371 = ($5370|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5371) {
      $$lcssa6951 = $5369;$__nd$0$i$i2433$lcssa6949 = $__nd$0$i$i2433;
      label = 2277;
      break;
     } else {
      $$sink$i$i2434 = $5370;
     }
    }
    $__nd$0$i$i2433 = $$sink$i$i2434;
   }
   if ((label|0) == 2273) {
    HEAP32[$__parent$i2431>>2] = $__nd$0$i$i2433$lcssa6950; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5367 = $__nd$0$i$i2433$lcssa6950; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2439 = $$lcssa6952;$5383 = $5367;
    label = 2280;
    break;
   }
   else if ((label|0) == 2277) {
    HEAP32[$__parent$i2431>>2] = $__nd$0$i$i2433$lcssa6949; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5372 = $__nd$0$i$i2433$lcssa6949; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2439 = $$lcssa6951;$5383 = $5372;
    label = 2280;
    break;
   }
   else if ((label|0) == 2279) {
    HEAP32[$__parent$i2431>>2] = $__nd$0$i$i2433$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5374 = $__nd$0$i$i2433$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5375 = ($__nd$0$i$i2433$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5375) {
     $$0$i7$i2439 = $__parent$i2431;$5383 = $5374;
     label = 2280;
     break;
    } else {
     $5394 = $5361;$__r$0$i2443 = $__nd$0$i$i2433$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2280) {
  __THREW__ = 0;
  $5376 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5377 = __THREW__; __THREW__ = 0;
  $5378 = $5377&1;
  if ($5378) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5379 = ((($5376)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5379>>2] = 1073742054; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5380 = ((($5376)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5380>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5376>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5381 = ((($5376)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5381>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5382 = ((($5376)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5382>>2] = $5383; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2439>>2] = $5376; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5384 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5385 = HEAP32[$5384>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5386 = ($5385|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5386) {
   $5389 = $5376;
  } else {
   $5387 = $5385; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5387; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2441 = HEAP32[$$0$i7$i2439>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5389 = $$pre$i$i2441;
  }
  $5388 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5388,$5389); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5390 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5391 = (($5390) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5391; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4538 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5394 = $$pre4538;$__r$0$i2443 = $5376;
 }
 $5392 = ((($__r$0$i2443)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5392>>2] = 163; //@line 783 "modules/event/sdl/Event.cpp"
 $5393 = ($5394|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5393) {
   HEAP32[$__parent$i2446>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5406 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2454 = $2;$5416 = $5406;
   label = 2294;
  } else {
   $__nd$0$i$i2448 = $5394;
   while(1) {
    $5395 = ((($__nd$0$i$i2448)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5396 = HEAP32[$5395>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5397 = ($5396|0)>(1073742055); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5397) {
     $5398 = HEAP32[$__nd$0$i$i2448>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5399 = ($5398|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5399) {
      $$lcssa6948 = $__nd$0$i$i2448;$__nd$0$i$i2448$lcssa6946 = $__nd$0$i$i2448;
      label = 2287;
      break;
     } else {
      $$sink$i$i2449 = $5398;
     }
    } else {
     $5401 = ($5396|0)<(1073742055); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5401)) {
      $__nd$0$i$i2448$lcssa = $__nd$0$i$i2448;
      label = 2293;
      break;
     }
     $5402 = ((($__nd$0$i$i2448)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5403 = HEAP32[$5402>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5404 = ($5403|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5404) {
      $$lcssa6947 = $5402;$__nd$0$i$i2448$lcssa6945 = $__nd$0$i$i2448;
      label = 2291;
      break;
     } else {
      $$sink$i$i2449 = $5403;
     }
    }
    $__nd$0$i$i2448 = $$sink$i$i2449;
   }
   if ((label|0) == 2287) {
    HEAP32[$__parent$i2446>>2] = $__nd$0$i$i2448$lcssa6946; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5400 = $__nd$0$i$i2448$lcssa6946; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2454 = $$lcssa6948;$5416 = $5400;
    label = 2294;
    break;
   }
   else if ((label|0) == 2291) {
    HEAP32[$__parent$i2446>>2] = $__nd$0$i$i2448$lcssa6945; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5405 = $__nd$0$i$i2448$lcssa6945; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2454 = $$lcssa6947;$5416 = $5405;
    label = 2294;
    break;
   }
   else if ((label|0) == 2293) {
    HEAP32[$__parent$i2446>>2] = $__nd$0$i$i2448$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5407 = $__nd$0$i$i2448$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5408 = ($__nd$0$i$i2448$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5408) {
     $$0$i7$i2454 = $__parent$i2446;$5416 = $5407;
     label = 2294;
     break;
    } else {
     $5427 = $5394;$__r$0$i2458 = $__nd$0$i$i2448$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2294) {
  __THREW__ = 0;
  $5409 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5410 = __THREW__; __THREW__ = 0;
  $5411 = $5410&1;
  if ($5411) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5412 = ((($5409)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5412>>2] = 1073742055; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5413 = ((($5409)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5413>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5409>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5414 = ((($5409)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5414>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5415 = ((($5409)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5415>>2] = $5416; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2454>>2] = $5409; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5417 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5418 = HEAP32[$5417>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5419 = ($5418|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5419) {
   $5422 = $5409;
  } else {
   $5420 = $5418; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5420; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2456 = HEAP32[$$0$i7$i2454>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5422 = $$pre$i$i2456;
  }
  $5421 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5421,$5422); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5423 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5424 = (($5423) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5424; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4539 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5427 = $$pre4539;$__r$0$i2458 = $5409;
 }
 $5425 = ((($__r$0$i2458)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5425>>2] = 164; //@line 784 "modules/event/sdl/Event.cpp"
 $5426 = ($5427|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5426) {
   HEAP32[$__parent$i2461>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5439 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2469 = $2;$5449 = $5439;
   label = 2308;
  } else {
   $__nd$0$i$i2463 = $5427;
   while(1) {
    $5428 = ((($__nd$0$i$i2463)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5429 = HEAP32[$5428>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5430 = ($5429|0)>(1073742081); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5430) {
     $5431 = HEAP32[$__nd$0$i$i2463>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5432 = ($5431|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5432) {
      $$lcssa6944 = $__nd$0$i$i2463;$__nd$0$i$i2463$lcssa6942 = $__nd$0$i$i2463;
      label = 2301;
      break;
     } else {
      $$sink$i$i2464 = $5431;
     }
    } else {
     $5434 = ($5429|0)<(1073742081); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5434)) {
      $__nd$0$i$i2463$lcssa = $__nd$0$i$i2463;
      label = 2307;
      break;
     }
     $5435 = ((($__nd$0$i$i2463)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5436 = HEAP32[$5435>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5437 = ($5436|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5437) {
      $$lcssa6943 = $5435;$__nd$0$i$i2463$lcssa6941 = $__nd$0$i$i2463;
      label = 2305;
      break;
     } else {
      $$sink$i$i2464 = $5436;
     }
    }
    $__nd$0$i$i2463 = $$sink$i$i2464;
   }
   if ((label|0) == 2301) {
    HEAP32[$__parent$i2461>>2] = $__nd$0$i$i2463$lcssa6942; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5433 = $__nd$0$i$i2463$lcssa6942; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2469 = $$lcssa6944;$5449 = $5433;
    label = 2308;
    break;
   }
   else if ((label|0) == 2305) {
    HEAP32[$__parent$i2461>>2] = $__nd$0$i$i2463$lcssa6941; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5438 = $__nd$0$i$i2463$lcssa6941; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2469 = $$lcssa6943;$5449 = $5438;
    label = 2308;
    break;
   }
   else if ((label|0) == 2307) {
    HEAP32[$__parent$i2461>>2] = $__nd$0$i$i2463$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5440 = $__nd$0$i$i2463$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5441 = ($__nd$0$i$i2463$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5441) {
     $$0$i7$i2469 = $__parent$i2461;$5449 = $5440;
     label = 2308;
     break;
    } else {
     $5460 = $5427;$__r$0$i2473 = $__nd$0$i$i2463$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2308) {
  __THREW__ = 0;
  $5442 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5443 = __THREW__; __THREW__ = 0;
  $5444 = $5443&1;
  if ($5444) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5445 = ((($5442)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5445>>2] = 1073742081; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5446 = ((($5442)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5446>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5442>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5447 = ((($5442)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5447>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5448 = ((($5442)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5448>>2] = $5449; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2469>>2] = $5442; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5450 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5451 = HEAP32[$5450>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5452 = ($5451|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5452) {
   $5455 = $5442;
  } else {
   $5453 = $5451; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5453; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2471 = HEAP32[$$0$i7$i2469>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5455 = $$pre$i$i2471;
  }
  $5454 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5454,$5455); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5456 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5457 = (($5456) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5457; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4540 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5460 = $$pre4540;$__r$0$i2473 = $5442;
 }
 $5458 = ((($__r$0$i2473)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5458>>2] = 165; //@line 786 "modules/event/sdl/Event.cpp"
 $5459 = ($5460|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5459) {
   HEAP32[$__parent$i2476>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5472 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2484 = $2;$5482 = $5472;
   label = 2322;
  } else {
   $__nd$0$i$i2478 = $5460;
   while(1) {
    $5461 = ((($__nd$0$i$i2478)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5462 = HEAP32[$5461>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5463 = ($5462|0)>(1073742082); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5463) {
     $5464 = HEAP32[$__nd$0$i$i2478>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5465 = ($5464|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5465) {
      $$lcssa6940 = $__nd$0$i$i2478;$__nd$0$i$i2478$lcssa6938 = $__nd$0$i$i2478;
      label = 2315;
      break;
     } else {
      $$sink$i$i2479 = $5464;
     }
    } else {
     $5467 = ($5462|0)<(1073742082); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5467)) {
      $__nd$0$i$i2478$lcssa = $__nd$0$i$i2478;
      label = 2321;
      break;
     }
     $5468 = ((($__nd$0$i$i2478)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5469 = HEAP32[$5468>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5470 = ($5469|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5470) {
      $$lcssa6939 = $5468;$__nd$0$i$i2478$lcssa6937 = $__nd$0$i$i2478;
      label = 2319;
      break;
     } else {
      $$sink$i$i2479 = $5469;
     }
    }
    $__nd$0$i$i2478 = $$sink$i$i2479;
   }
   if ((label|0) == 2315) {
    HEAP32[$__parent$i2476>>2] = $__nd$0$i$i2478$lcssa6938; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5466 = $__nd$0$i$i2478$lcssa6938; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2484 = $$lcssa6940;$5482 = $5466;
    label = 2322;
    break;
   }
   else if ((label|0) == 2319) {
    HEAP32[$__parent$i2476>>2] = $__nd$0$i$i2478$lcssa6937; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5471 = $__nd$0$i$i2478$lcssa6937; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2484 = $$lcssa6939;$5482 = $5471;
    label = 2322;
    break;
   }
   else if ((label|0) == 2321) {
    HEAP32[$__parent$i2476>>2] = $__nd$0$i$i2478$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5473 = $__nd$0$i$i2478$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5474 = ($__nd$0$i$i2478$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5474) {
     $$0$i7$i2484 = $__parent$i2476;$5482 = $5473;
     label = 2322;
     break;
    } else {
     $5493 = $5460;$__r$0$i2488 = $__nd$0$i$i2478$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2322) {
  __THREW__ = 0;
  $5475 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5476 = __THREW__; __THREW__ = 0;
  $5477 = $5476&1;
  if ($5477) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5478 = ((($5475)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5478>>2] = 1073742082; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5479 = ((($5475)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5479>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5475>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5480 = ((($5475)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5480>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5481 = ((($5475)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5481>>2] = $5482; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2484>>2] = $5475; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5483 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5484 = HEAP32[$5483>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5485 = ($5484|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5485) {
   $5488 = $5475;
  } else {
   $5486 = $5484; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5486; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2486 = HEAP32[$$0$i7$i2484>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5488 = $$pre$i$i2486;
  }
  $5487 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5487,$5488); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5489 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5490 = (($5489) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5490; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4541 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5493 = $$pre4541;$__r$0$i2488 = $5475;
 }
 $5491 = ((($__r$0$i2488)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5491>>2] = 166; //@line 788 "modules/event/sdl/Event.cpp"
 $5492 = ($5493|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5492) {
   HEAP32[$__parent$i2491>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5505 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2499 = $2;$5515 = $5505;
   label = 2336;
  } else {
   $__nd$0$i$i2493 = $5493;
   while(1) {
    $5494 = ((($__nd$0$i$i2493)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5495 = HEAP32[$5494>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5496 = ($5495|0)>(1073742083); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5496) {
     $5497 = HEAP32[$__nd$0$i$i2493>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5498 = ($5497|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5498) {
      $$lcssa6936 = $__nd$0$i$i2493;$__nd$0$i$i2493$lcssa6934 = $__nd$0$i$i2493;
      label = 2329;
      break;
     } else {
      $$sink$i$i2494 = $5497;
     }
    } else {
     $5500 = ($5495|0)<(1073742083); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5500)) {
      $__nd$0$i$i2493$lcssa = $__nd$0$i$i2493;
      label = 2335;
      break;
     }
     $5501 = ((($__nd$0$i$i2493)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5502 = HEAP32[$5501>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5503 = ($5502|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5503) {
      $$lcssa6935 = $5501;$__nd$0$i$i2493$lcssa6933 = $__nd$0$i$i2493;
      label = 2333;
      break;
     } else {
      $$sink$i$i2494 = $5502;
     }
    }
    $__nd$0$i$i2493 = $$sink$i$i2494;
   }
   if ((label|0) == 2329) {
    HEAP32[$__parent$i2491>>2] = $__nd$0$i$i2493$lcssa6934; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5499 = $__nd$0$i$i2493$lcssa6934; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2499 = $$lcssa6936;$5515 = $5499;
    label = 2336;
    break;
   }
   else if ((label|0) == 2333) {
    HEAP32[$__parent$i2491>>2] = $__nd$0$i$i2493$lcssa6933; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5504 = $__nd$0$i$i2493$lcssa6933; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2499 = $$lcssa6935;$5515 = $5504;
    label = 2336;
    break;
   }
   else if ((label|0) == 2335) {
    HEAP32[$__parent$i2491>>2] = $__nd$0$i$i2493$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5506 = $__nd$0$i$i2493$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5507 = ($__nd$0$i$i2493$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5507) {
     $$0$i7$i2499 = $__parent$i2491;$5515 = $5506;
     label = 2336;
     break;
    } else {
     $5526 = $5493;$__r$0$i2503 = $__nd$0$i$i2493$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2336) {
  __THREW__ = 0;
  $5508 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5509 = __THREW__; __THREW__ = 0;
  $5510 = $5509&1;
  if ($5510) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5511 = ((($5508)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5511>>2] = 1073742083; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5512 = ((($5508)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5512>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5508>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5513 = ((($5508)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5513>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5514 = ((($5508)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5514>>2] = $5515; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2499>>2] = $5508; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5516 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5517 = HEAP32[$5516>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5518 = ($5517|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5518) {
   $5521 = $5508;
  } else {
   $5519 = $5517; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5519; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2501 = HEAP32[$$0$i7$i2499>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5521 = $$pre$i$i2501;
  }
  $5520 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5520,$5521); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5522 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5523 = (($5522) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5523; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4542 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5526 = $$pre4542;$__r$0$i2503 = $5508;
 }
 $5524 = ((($__r$0$i2503)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5524>>2] = 167; //@line 789 "modules/event/sdl/Event.cpp"
 $5525 = ($5526|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5525) {
   HEAP32[$__parent$i2506>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5538 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2514 = $2;$5548 = $5538;
   label = 2350;
  } else {
   $__nd$0$i$i2508 = $5526;
   while(1) {
    $5527 = ((($__nd$0$i$i2508)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5528 = HEAP32[$5527>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5529 = ($5528|0)>(1073742084); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5529) {
     $5530 = HEAP32[$__nd$0$i$i2508>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5531 = ($5530|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5531) {
      $$lcssa6932 = $__nd$0$i$i2508;$__nd$0$i$i2508$lcssa6930 = $__nd$0$i$i2508;
      label = 2343;
      break;
     } else {
      $$sink$i$i2509 = $5530;
     }
    } else {
     $5533 = ($5528|0)<(1073742084); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5533)) {
      $__nd$0$i$i2508$lcssa = $__nd$0$i$i2508;
      label = 2349;
      break;
     }
     $5534 = ((($__nd$0$i$i2508)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5535 = HEAP32[$5534>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5536 = ($5535|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5536) {
      $$lcssa6931 = $5534;$__nd$0$i$i2508$lcssa6929 = $__nd$0$i$i2508;
      label = 2347;
      break;
     } else {
      $$sink$i$i2509 = $5535;
     }
    }
    $__nd$0$i$i2508 = $$sink$i$i2509;
   }
   if ((label|0) == 2343) {
    HEAP32[$__parent$i2506>>2] = $__nd$0$i$i2508$lcssa6930; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5532 = $__nd$0$i$i2508$lcssa6930; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2514 = $$lcssa6932;$5548 = $5532;
    label = 2350;
    break;
   }
   else if ((label|0) == 2347) {
    HEAP32[$__parent$i2506>>2] = $__nd$0$i$i2508$lcssa6929; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5537 = $__nd$0$i$i2508$lcssa6929; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2514 = $$lcssa6931;$5548 = $5537;
    label = 2350;
    break;
   }
   else if ((label|0) == 2349) {
    HEAP32[$__parent$i2506>>2] = $__nd$0$i$i2508$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5539 = $__nd$0$i$i2508$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5540 = ($__nd$0$i$i2508$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5540) {
     $$0$i7$i2514 = $__parent$i2506;$5548 = $5539;
     label = 2350;
     break;
    } else {
     $5559 = $5526;$__r$0$i2518 = $__nd$0$i$i2508$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2350) {
  __THREW__ = 0;
  $5541 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5542 = __THREW__; __THREW__ = 0;
  $5543 = $5542&1;
  if ($5543) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5544 = ((($5541)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5544>>2] = 1073742084; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5545 = ((($5541)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5545>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5541>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5546 = ((($5541)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5546>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5547 = ((($5541)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5547>>2] = $5548; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2514>>2] = $5541; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5549 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5550 = HEAP32[$5549>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5551 = ($5550|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5551) {
   $5554 = $5541;
  } else {
   $5552 = $5550; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5552; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2516 = HEAP32[$$0$i7$i2514>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5554 = $$pre$i$i2516;
  }
  $5553 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5553,$5554); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5555 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5556 = (($5555) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5556; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4543 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5559 = $$pre4543;$__r$0$i2518 = $5541;
 }
 $5557 = ((($__r$0$i2518)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5557>>2] = 168; //@line 790 "modules/event/sdl/Event.cpp"
 $5558 = ($5559|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5558) {
   HEAP32[$__parent$i2521>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5571 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2529 = $2;$5581 = $5571;
   label = 2364;
  } else {
   $__nd$0$i$i2523 = $5559;
   while(1) {
    $5560 = ((($__nd$0$i$i2523)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5561 = HEAP32[$5560>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5562 = ($5561|0)>(1073742085); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5562) {
     $5563 = HEAP32[$__nd$0$i$i2523>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5564 = ($5563|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5564) {
      $$lcssa6928 = $__nd$0$i$i2523;$__nd$0$i$i2523$lcssa6926 = $__nd$0$i$i2523;
      label = 2357;
      break;
     } else {
      $$sink$i$i2524 = $5563;
     }
    } else {
     $5566 = ($5561|0)<(1073742085); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5566)) {
      $__nd$0$i$i2523$lcssa = $__nd$0$i$i2523;
      label = 2363;
      break;
     }
     $5567 = ((($__nd$0$i$i2523)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5568 = HEAP32[$5567>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5569 = ($5568|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5569) {
      $$lcssa6927 = $5567;$__nd$0$i$i2523$lcssa6925 = $__nd$0$i$i2523;
      label = 2361;
      break;
     } else {
      $$sink$i$i2524 = $5568;
     }
    }
    $__nd$0$i$i2523 = $$sink$i$i2524;
   }
   if ((label|0) == 2357) {
    HEAP32[$__parent$i2521>>2] = $__nd$0$i$i2523$lcssa6926; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5565 = $__nd$0$i$i2523$lcssa6926; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2529 = $$lcssa6928;$5581 = $5565;
    label = 2364;
    break;
   }
   else if ((label|0) == 2361) {
    HEAP32[$__parent$i2521>>2] = $__nd$0$i$i2523$lcssa6925; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5570 = $__nd$0$i$i2523$lcssa6925; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2529 = $$lcssa6927;$5581 = $5570;
    label = 2364;
    break;
   }
   else if ((label|0) == 2363) {
    HEAP32[$__parent$i2521>>2] = $__nd$0$i$i2523$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5572 = $__nd$0$i$i2523$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5573 = ($__nd$0$i$i2523$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5573) {
     $$0$i7$i2529 = $__parent$i2521;$5581 = $5572;
     label = 2364;
     break;
    } else {
     $5592 = $5559;$__r$0$i2533 = $__nd$0$i$i2523$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2364) {
  __THREW__ = 0;
  $5574 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5575 = __THREW__; __THREW__ = 0;
  $5576 = $5575&1;
  if ($5576) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5577 = ((($5574)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5577>>2] = 1073742085; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5578 = ((($5574)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5578>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5574>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5579 = ((($5574)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5579>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5580 = ((($5574)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5580>>2] = $5581; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2529>>2] = $5574; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5582 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5583 = HEAP32[$5582>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5584 = ($5583|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5584) {
   $5587 = $5574;
  } else {
   $5585 = $5583; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5585; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2531 = HEAP32[$$0$i7$i2529>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5587 = $$pre$i$i2531;
  }
  $5586 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5586,$5587); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5588 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5589 = (($5588) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5589; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4544 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5592 = $$pre4544;$__r$0$i2533 = $5574;
 }
 $5590 = ((($__r$0$i2533)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5590>>2] = 169; //@line 791 "modules/event/sdl/Event.cpp"
 $5591 = ($5592|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5591) {
   HEAP32[$__parent$i2536>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5604 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2544 = $2;$5614 = $5604;
   label = 2378;
  } else {
   $__nd$0$i$i2538 = $5592;
   while(1) {
    $5593 = ((($__nd$0$i$i2538)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5594 = HEAP32[$5593>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5595 = ($5594|0)>(1073742086); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5595) {
     $5596 = HEAP32[$__nd$0$i$i2538>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5597 = ($5596|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5597) {
      $$lcssa6924 = $__nd$0$i$i2538;$__nd$0$i$i2538$lcssa6922 = $__nd$0$i$i2538;
      label = 2371;
      break;
     } else {
      $$sink$i$i2539 = $5596;
     }
    } else {
     $5599 = ($5594|0)<(1073742086); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5599)) {
      $__nd$0$i$i2538$lcssa = $__nd$0$i$i2538;
      label = 2377;
      break;
     }
     $5600 = ((($__nd$0$i$i2538)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5601 = HEAP32[$5600>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5602 = ($5601|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5602) {
      $$lcssa6923 = $5600;$__nd$0$i$i2538$lcssa6921 = $__nd$0$i$i2538;
      label = 2375;
      break;
     } else {
      $$sink$i$i2539 = $5601;
     }
    }
    $__nd$0$i$i2538 = $$sink$i$i2539;
   }
   if ((label|0) == 2371) {
    HEAP32[$__parent$i2536>>2] = $__nd$0$i$i2538$lcssa6922; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5598 = $__nd$0$i$i2538$lcssa6922; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2544 = $$lcssa6924;$5614 = $5598;
    label = 2378;
    break;
   }
   else if ((label|0) == 2375) {
    HEAP32[$__parent$i2536>>2] = $__nd$0$i$i2538$lcssa6921; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5603 = $__nd$0$i$i2538$lcssa6921; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2544 = $$lcssa6923;$5614 = $5603;
    label = 2378;
    break;
   }
   else if ((label|0) == 2377) {
    HEAP32[$__parent$i2536>>2] = $__nd$0$i$i2538$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5605 = $__nd$0$i$i2538$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5606 = ($__nd$0$i$i2538$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5606) {
     $$0$i7$i2544 = $__parent$i2536;$5614 = $5605;
     label = 2378;
     break;
    } else {
     $5625 = $5592;$__r$0$i2548 = $__nd$0$i$i2538$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2378) {
  __THREW__ = 0;
  $5607 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5608 = __THREW__; __THREW__ = 0;
  $5609 = $5608&1;
  if ($5609) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5610 = ((($5607)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5610>>2] = 1073742086; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5611 = ((($5607)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5611>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5607>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5612 = ((($5607)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5612>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5613 = ((($5607)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5613>>2] = $5614; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2544>>2] = $5607; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5615 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5616 = HEAP32[$5615>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5617 = ($5616|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5617) {
   $5620 = $5607;
  } else {
   $5618 = $5616; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5618; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2546 = HEAP32[$$0$i7$i2544>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5620 = $$pre$i$i2546;
  }
  $5619 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5619,$5620); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5621 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5622 = (($5621) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5622; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4545 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5625 = $$pre4545;$__r$0$i2548 = $5607;
 }
 $5623 = ((($__r$0$i2548)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5623>>2] = 170; //@line 792 "modules/event/sdl/Event.cpp"
 $5624 = ($5625|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5624) {
   HEAP32[$__parent$i2551>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5637 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2559 = $2;$5647 = $5637;
   label = 2392;
  } else {
   $__nd$0$i$i2553 = $5625;
   while(1) {
    $5626 = ((($__nd$0$i$i2553)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5627 = HEAP32[$5626>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5628 = ($5627|0)>(1073742087); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5628) {
     $5629 = HEAP32[$__nd$0$i$i2553>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5630 = ($5629|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5630) {
      $$lcssa6920 = $__nd$0$i$i2553;$__nd$0$i$i2553$lcssa6918 = $__nd$0$i$i2553;
      label = 2385;
      break;
     } else {
      $$sink$i$i2554 = $5629;
     }
    } else {
     $5632 = ($5627|0)<(1073742087); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5632)) {
      $__nd$0$i$i2553$lcssa = $__nd$0$i$i2553;
      label = 2391;
      break;
     }
     $5633 = ((($__nd$0$i$i2553)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5634 = HEAP32[$5633>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5635 = ($5634|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5635) {
      $$lcssa6919 = $5633;$__nd$0$i$i2553$lcssa6917 = $__nd$0$i$i2553;
      label = 2389;
      break;
     } else {
      $$sink$i$i2554 = $5634;
     }
    }
    $__nd$0$i$i2553 = $$sink$i$i2554;
   }
   if ((label|0) == 2385) {
    HEAP32[$__parent$i2551>>2] = $__nd$0$i$i2553$lcssa6918; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5631 = $__nd$0$i$i2553$lcssa6918; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2559 = $$lcssa6920;$5647 = $5631;
    label = 2392;
    break;
   }
   else if ((label|0) == 2389) {
    HEAP32[$__parent$i2551>>2] = $__nd$0$i$i2553$lcssa6917; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5636 = $__nd$0$i$i2553$lcssa6917; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2559 = $$lcssa6919;$5647 = $5636;
    label = 2392;
    break;
   }
   else if ((label|0) == 2391) {
    HEAP32[$__parent$i2551>>2] = $__nd$0$i$i2553$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5638 = $__nd$0$i$i2553$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5639 = ($__nd$0$i$i2553$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5639) {
     $$0$i7$i2559 = $__parent$i2551;$5647 = $5638;
     label = 2392;
     break;
    } else {
     $5658 = $5625;$__r$0$i2563 = $__nd$0$i$i2553$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2392) {
  __THREW__ = 0;
  $5640 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5641 = __THREW__; __THREW__ = 0;
  $5642 = $5641&1;
  if ($5642) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5643 = ((($5640)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5643>>2] = 1073742087; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5644 = ((($5640)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5644>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5640>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5645 = ((($5640)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5645>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5646 = ((($5640)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5646>>2] = $5647; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2559>>2] = $5640; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5648 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5649 = HEAP32[$5648>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5650 = ($5649|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5650) {
   $5653 = $5640;
  } else {
   $5651 = $5649; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5651; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2561 = HEAP32[$$0$i7$i2559>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5653 = $$pre$i$i2561;
  }
  $5652 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5652,$5653); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5654 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5655 = (($5654) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5655; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4546 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5658 = $$pre4546;$__r$0$i2563 = $5640;
 }
 $5656 = ((($__r$0$i2563)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5656>>2] = 171; //@line 793 "modules/event/sdl/Event.cpp"
 $5657 = ($5658|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5657) {
   HEAP32[$__parent$i2566>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5670 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2574 = $2;$5680 = $5670;
   label = 2406;
  } else {
   $__nd$0$i$i2568 = $5658;
   while(1) {
    $5659 = ((($__nd$0$i$i2568)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5660 = HEAP32[$5659>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5661 = ($5660|0)>(1073742088); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5661) {
     $5662 = HEAP32[$__nd$0$i$i2568>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5663 = ($5662|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5663) {
      $$lcssa6916 = $__nd$0$i$i2568;$__nd$0$i$i2568$lcssa6914 = $__nd$0$i$i2568;
      label = 2399;
      break;
     } else {
      $$sink$i$i2569 = $5662;
     }
    } else {
     $5665 = ($5660|0)<(1073742088); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5665)) {
      $__nd$0$i$i2568$lcssa = $__nd$0$i$i2568;
      label = 2405;
      break;
     }
     $5666 = ((($__nd$0$i$i2568)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5667 = HEAP32[$5666>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5668 = ($5667|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5668) {
      $$lcssa6915 = $5666;$__nd$0$i$i2568$lcssa6913 = $__nd$0$i$i2568;
      label = 2403;
      break;
     } else {
      $$sink$i$i2569 = $5667;
     }
    }
    $__nd$0$i$i2568 = $$sink$i$i2569;
   }
   if ((label|0) == 2399) {
    HEAP32[$__parent$i2566>>2] = $__nd$0$i$i2568$lcssa6914; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5664 = $__nd$0$i$i2568$lcssa6914; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2574 = $$lcssa6916;$5680 = $5664;
    label = 2406;
    break;
   }
   else if ((label|0) == 2403) {
    HEAP32[$__parent$i2566>>2] = $__nd$0$i$i2568$lcssa6913; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5669 = $__nd$0$i$i2568$lcssa6913; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2574 = $$lcssa6915;$5680 = $5669;
    label = 2406;
    break;
   }
   else if ((label|0) == 2405) {
    HEAP32[$__parent$i2566>>2] = $__nd$0$i$i2568$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5671 = $__nd$0$i$i2568$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5672 = ($__nd$0$i$i2568$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5672) {
     $$0$i7$i2574 = $__parent$i2566;$5680 = $5671;
     label = 2406;
     break;
    } else {
     $5691 = $5658;$__r$0$i2578 = $__nd$0$i$i2568$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2406) {
  __THREW__ = 0;
  $5673 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5674 = __THREW__; __THREW__ = 0;
  $5675 = $5674&1;
  if ($5675) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5676 = ((($5673)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5676>>2] = 1073742088; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5677 = ((($5673)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5677>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5673>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5678 = ((($5673)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5678>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5679 = ((($5673)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5679>>2] = $5680; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2574>>2] = $5673; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5681 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5682 = HEAP32[$5681>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5683 = ($5682|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5683) {
   $5686 = $5673;
  } else {
   $5684 = $5682; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5684; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2576 = HEAP32[$$0$i7$i2574>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5686 = $$pre$i$i2576;
  }
  $5685 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5685,$5686); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5687 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5688 = (($5687) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5688; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4547 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5691 = $$pre4547;$__r$0$i2578 = $5673;
 }
 $5689 = ((($__r$0$i2578)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5689>>2] = 172; //@line 794 "modules/event/sdl/Event.cpp"
 $5690 = ($5691|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5690) {
   HEAP32[$__parent$i2581>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5703 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2589 = $2;$5713 = $5703;
   label = 2420;
  } else {
   $__nd$0$i$i2583 = $5691;
   while(1) {
    $5692 = ((($__nd$0$i$i2583)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5693 = HEAP32[$5692>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5694 = ($5693|0)>(1073742089); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5694) {
     $5695 = HEAP32[$__nd$0$i$i2583>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5696 = ($5695|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5696) {
      $$lcssa6912 = $__nd$0$i$i2583;$__nd$0$i$i2583$lcssa6910 = $__nd$0$i$i2583;
      label = 2413;
      break;
     } else {
      $$sink$i$i2584 = $5695;
     }
    } else {
     $5698 = ($5693|0)<(1073742089); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5698)) {
      $__nd$0$i$i2583$lcssa = $__nd$0$i$i2583;
      label = 2419;
      break;
     }
     $5699 = ((($__nd$0$i$i2583)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5700 = HEAP32[$5699>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5701 = ($5700|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5701) {
      $$lcssa6911 = $5699;$__nd$0$i$i2583$lcssa6909 = $__nd$0$i$i2583;
      label = 2417;
      break;
     } else {
      $$sink$i$i2584 = $5700;
     }
    }
    $__nd$0$i$i2583 = $$sink$i$i2584;
   }
   if ((label|0) == 2413) {
    HEAP32[$__parent$i2581>>2] = $__nd$0$i$i2583$lcssa6910; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5697 = $__nd$0$i$i2583$lcssa6910; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2589 = $$lcssa6912;$5713 = $5697;
    label = 2420;
    break;
   }
   else if ((label|0) == 2417) {
    HEAP32[$__parent$i2581>>2] = $__nd$0$i$i2583$lcssa6909; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5702 = $__nd$0$i$i2583$lcssa6909; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2589 = $$lcssa6911;$5713 = $5702;
    label = 2420;
    break;
   }
   else if ((label|0) == 2419) {
    HEAP32[$__parent$i2581>>2] = $__nd$0$i$i2583$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5704 = $__nd$0$i$i2583$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5705 = ($__nd$0$i$i2583$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5705) {
     $$0$i7$i2589 = $__parent$i2581;$5713 = $5704;
     label = 2420;
     break;
    } else {
     $5724 = $5691;$__r$0$i2593 = $__nd$0$i$i2583$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2420) {
  __THREW__ = 0;
  $5706 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5707 = __THREW__; __THREW__ = 0;
  $5708 = $5707&1;
  if ($5708) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5709 = ((($5706)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5709>>2] = 1073742089; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5710 = ((($5706)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5710>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5706>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5711 = ((($5706)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5711>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5712 = ((($5706)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5712>>2] = $5713; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2589>>2] = $5706; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5714 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5715 = HEAP32[$5714>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5716 = ($5715|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5716) {
   $5719 = $5706;
  } else {
   $5717 = $5715; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5717; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2591 = HEAP32[$$0$i7$i2589>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5719 = $$pre$i$i2591;
  }
  $5718 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5718,$5719); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5720 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5721 = (($5720) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5721; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4548 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5724 = $$pre4548;$__r$0$i2593 = $5706;
 }
 $5722 = ((($__r$0$i2593)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5722>>2] = 173; //@line 795 "modules/event/sdl/Event.cpp"
 $5723 = ($5724|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5723) {
   HEAP32[$__parent$i2596>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5736 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2604 = $2;$5746 = $5736;
   label = 2434;
  } else {
   $__nd$0$i$i2598 = $5724;
   while(1) {
    $5725 = ((($__nd$0$i$i2598)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5726 = HEAP32[$5725>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5727 = ($5726|0)>(1073742090); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5727) {
     $5728 = HEAP32[$__nd$0$i$i2598>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5729 = ($5728|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5729) {
      $$lcssa6908 = $__nd$0$i$i2598;$__nd$0$i$i2598$lcssa6906 = $__nd$0$i$i2598;
      label = 2427;
      break;
     } else {
      $$sink$i$i2599 = $5728;
     }
    } else {
     $5731 = ($5726|0)<(1073742090); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5731)) {
      $__nd$0$i$i2598$lcssa = $__nd$0$i$i2598;
      label = 2433;
      break;
     }
     $5732 = ((($__nd$0$i$i2598)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5733 = HEAP32[$5732>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5734 = ($5733|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5734) {
      $$lcssa6907 = $5732;$__nd$0$i$i2598$lcssa6905 = $__nd$0$i$i2598;
      label = 2431;
      break;
     } else {
      $$sink$i$i2599 = $5733;
     }
    }
    $__nd$0$i$i2598 = $$sink$i$i2599;
   }
   if ((label|0) == 2427) {
    HEAP32[$__parent$i2596>>2] = $__nd$0$i$i2598$lcssa6906; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5730 = $__nd$0$i$i2598$lcssa6906; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2604 = $$lcssa6908;$5746 = $5730;
    label = 2434;
    break;
   }
   else if ((label|0) == 2431) {
    HEAP32[$__parent$i2596>>2] = $__nd$0$i$i2598$lcssa6905; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5735 = $__nd$0$i$i2598$lcssa6905; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2604 = $$lcssa6907;$5746 = $5735;
    label = 2434;
    break;
   }
   else if ((label|0) == 2433) {
    HEAP32[$__parent$i2596>>2] = $__nd$0$i$i2598$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5737 = $__nd$0$i$i2598$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5738 = ($__nd$0$i$i2598$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5738) {
     $$0$i7$i2604 = $__parent$i2596;$5746 = $5737;
     label = 2434;
     break;
    } else {
     $5757 = $5724;$__r$0$i2608 = $__nd$0$i$i2598$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2434) {
  __THREW__ = 0;
  $5739 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5740 = __THREW__; __THREW__ = 0;
  $5741 = $5740&1;
  if ($5741) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5742 = ((($5739)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5742>>2] = 1073742090; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5743 = ((($5739)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5743>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5739>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5744 = ((($5739)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5744>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5745 = ((($5739)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5745>>2] = $5746; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2604>>2] = $5739; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5747 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5748 = HEAP32[$5747>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5749 = ($5748|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5749) {
   $5752 = $5739;
  } else {
   $5750 = $5748; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5750; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2606 = HEAP32[$$0$i7$i2604>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5752 = $$pre$i$i2606;
  }
  $5751 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5751,$5752); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5753 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5754 = (($5753) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5754; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4549 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5757 = $$pre4549;$__r$0$i2608 = $5739;
 }
 $5755 = ((($__r$0$i2608)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5755>>2] = 174; //@line 796 "modules/event/sdl/Event.cpp"
 $5756 = ($5757|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5756) {
   HEAP32[$__parent$i2611>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5769 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2619 = $2;$5779 = $5769;
   label = 2448;
  } else {
   $__nd$0$i$i2613 = $5757;
   while(1) {
    $5758 = ((($__nd$0$i$i2613)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5759 = HEAP32[$5758>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5760 = ($5759|0)>(1073742091); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5760) {
     $5761 = HEAP32[$__nd$0$i$i2613>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5762 = ($5761|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5762) {
      $$lcssa6904 = $__nd$0$i$i2613;$__nd$0$i$i2613$lcssa6902 = $__nd$0$i$i2613;
      label = 2441;
      break;
     } else {
      $$sink$i$i2614 = $5761;
     }
    } else {
     $5764 = ($5759|0)<(1073742091); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5764)) {
      $__nd$0$i$i2613$lcssa = $__nd$0$i$i2613;
      label = 2447;
      break;
     }
     $5765 = ((($__nd$0$i$i2613)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5766 = HEAP32[$5765>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5767 = ($5766|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5767) {
      $$lcssa6903 = $5765;$__nd$0$i$i2613$lcssa6901 = $__nd$0$i$i2613;
      label = 2445;
      break;
     } else {
      $$sink$i$i2614 = $5766;
     }
    }
    $__nd$0$i$i2613 = $$sink$i$i2614;
   }
   if ((label|0) == 2441) {
    HEAP32[$__parent$i2611>>2] = $__nd$0$i$i2613$lcssa6902; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5763 = $__nd$0$i$i2613$lcssa6902; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2619 = $$lcssa6904;$5779 = $5763;
    label = 2448;
    break;
   }
   else if ((label|0) == 2445) {
    HEAP32[$__parent$i2611>>2] = $__nd$0$i$i2613$lcssa6901; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5768 = $__nd$0$i$i2613$lcssa6901; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2619 = $$lcssa6903;$5779 = $5768;
    label = 2448;
    break;
   }
   else if ((label|0) == 2447) {
    HEAP32[$__parent$i2611>>2] = $__nd$0$i$i2613$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5770 = $__nd$0$i$i2613$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5771 = ($__nd$0$i$i2613$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5771) {
     $$0$i7$i2619 = $__parent$i2611;$5779 = $5770;
     label = 2448;
     break;
    } else {
     $5790 = $5757;$__r$0$i2623 = $__nd$0$i$i2613$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2448) {
  __THREW__ = 0;
  $5772 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5773 = __THREW__; __THREW__ = 0;
  $5774 = $5773&1;
  if ($5774) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5775 = ((($5772)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5775>>2] = 1073742091; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5776 = ((($5772)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5776>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5772>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5777 = ((($5772)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5777>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5778 = ((($5772)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5778>>2] = $5779; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2619>>2] = $5772; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5780 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5781 = HEAP32[$5780>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5782 = ($5781|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5782) {
   $5785 = $5772;
  } else {
   $5783 = $5781; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5783; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2621 = HEAP32[$$0$i7$i2619>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5785 = $$pre$i$i2621;
  }
  $5784 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5784,$5785); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5786 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5787 = (($5786) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5787; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4550 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5790 = $$pre4550;$__r$0$i2623 = $5772;
 }
 $5788 = ((($__r$0$i2623)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5788>>2] = 175; //@line 797 "modules/event/sdl/Event.cpp"
 $5789 = ($5790|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5789) {
   HEAP32[$__parent$i2626>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5802 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2634 = $2;$5812 = $5802;
   label = 2462;
  } else {
   $__nd$0$i$i2628 = $5790;
   while(1) {
    $5791 = ((($__nd$0$i$i2628)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5792 = HEAP32[$5791>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5793 = ($5792|0)>(1073742092); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5793) {
     $5794 = HEAP32[$__nd$0$i$i2628>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5795 = ($5794|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5795) {
      $$lcssa6900 = $__nd$0$i$i2628;$__nd$0$i$i2628$lcssa6898 = $__nd$0$i$i2628;
      label = 2455;
      break;
     } else {
      $$sink$i$i2629 = $5794;
     }
    } else {
     $5797 = ($5792|0)<(1073742092); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5797)) {
      $__nd$0$i$i2628$lcssa = $__nd$0$i$i2628;
      label = 2461;
      break;
     }
     $5798 = ((($__nd$0$i$i2628)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5799 = HEAP32[$5798>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5800 = ($5799|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5800) {
      $$lcssa6899 = $5798;$__nd$0$i$i2628$lcssa6897 = $__nd$0$i$i2628;
      label = 2459;
      break;
     } else {
      $$sink$i$i2629 = $5799;
     }
    }
    $__nd$0$i$i2628 = $$sink$i$i2629;
   }
   if ((label|0) == 2455) {
    HEAP32[$__parent$i2626>>2] = $__nd$0$i$i2628$lcssa6898; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5796 = $__nd$0$i$i2628$lcssa6898; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2634 = $$lcssa6900;$5812 = $5796;
    label = 2462;
    break;
   }
   else if ((label|0) == 2459) {
    HEAP32[$__parent$i2626>>2] = $__nd$0$i$i2628$lcssa6897; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5801 = $__nd$0$i$i2628$lcssa6897; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2634 = $$lcssa6899;$5812 = $5801;
    label = 2462;
    break;
   }
   else if ((label|0) == 2461) {
    HEAP32[$__parent$i2626>>2] = $__nd$0$i$i2628$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5803 = $__nd$0$i$i2628$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5804 = ($__nd$0$i$i2628$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5804) {
     $$0$i7$i2634 = $__parent$i2626;$5812 = $5803;
     label = 2462;
     break;
    } else {
     $5823 = $5790;$__r$0$i2638 = $__nd$0$i$i2628$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2462) {
  __THREW__ = 0;
  $5805 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5806 = __THREW__; __THREW__ = 0;
  $5807 = $5806&1;
  if ($5807) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5808 = ((($5805)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5808>>2] = 1073742092; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5809 = ((($5805)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5809>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5805>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5810 = ((($5805)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5810>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5811 = ((($5805)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5811>>2] = $5812; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2634>>2] = $5805; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5813 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5814 = HEAP32[$5813>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5815 = ($5814|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5815) {
   $5818 = $5805;
  } else {
   $5816 = $5814; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5816; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2636 = HEAP32[$$0$i7$i2634>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5818 = $$pre$i$i2636;
  }
  $5817 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5817,$5818); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5819 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5820 = (($5819) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5820; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4551 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5823 = $$pre4551;$__r$0$i2638 = $5805;
 }
 $5821 = ((($__r$0$i2638)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5821>>2] = 176; //@line 798 "modules/event/sdl/Event.cpp"
 $5822 = ($5823|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5822) {
   HEAP32[$__parent$i2641>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5835 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2649 = $2;$5845 = $5835;
   label = 2476;
  } else {
   $__nd$0$i$i2643 = $5823;
   while(1) {
    $5824 = ((($__nd$0$i$i2643)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5825 = HEAP32[$5824>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5826 = ($5825|0)>(1073742093); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5826) {
     $5827 = HEAP32[$__nd$0$i$i2643>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5828 = ($5827|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5828) {
      $$lcssa6896 = $__nd$0$i$i2643;$__nd$0$i$i2643$lcssa6894 = $__nd$0$i$i2643;
      label = 2469;
      break;
     } else {
      $$sink$i$i2644 = $5827;
     }
    } else {
     $5830 = ($5825|0)<(1073742093); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5830)) {
      $__nd$0$i$i2643$lcssa = $__nd$0$i$i2643;
      label = 2475;
      break;
     }
     $5831 = ((($__nd$0$i$i2643)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5832 = HEAP32[$5831>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5833 = ($5832|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5833) {
      $$lcssa6895 = $5831;$__nd$0$i$i2643$lcssa6893 = $__nd$0$i$i2643;
      label = 2473;
      break;
     } else {
      $$sink$i$i2644 = $5832;
     }
    }
    $__nd$0$i$i2643 = $$sink$i$i2644;
   }
   if ((label|0) == 2469) {
    HEAP32[$__parent$i2641>>2] = $__nd$0$i$i2643$lcssa6894; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5829 = $__nd$0$i$i2643$lcssa6894; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2649 = $$lcssa6896;$5845 = $5829;
    label = 2476;
    break;
   }
   else if ((label|0) == 2473) {
    HEAP32[$__parent$i2641>>2] = $__nd$0$i$i2643$lcssa6893; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5834 = $__nd$0$i$i2643$lcssa6893; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2649 = $$lcssa6895;$5845 = $5834;
    label = 2476;
    break;
   }
   else if ((label|0) == 2475) {
    HEAP32[$__parent$i2641>>2] = $__nd$0$i$i2643$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5836 = $__nd$0$i$i2643$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5837 = ($__nd$0$i$i2643$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5837) {
     $$0$i7$i2649 = $__parent$i2641;$5845 = $5836;
     label = 2476;
     break;
    } else {
     $5856 = $5823;$__r$0$i2653 = $__nd$0$i$i2643$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2476) {
  __THREW__ = 0;
  $5838 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5839 = __THREW__; __THREW__ = 0;
  $5840 = $5839&1;
  if ($5840) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5841 = ((($5838)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5841>>2] = 1073742093; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5842 = ((($5838)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5842>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5838>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5843 = ((($5838)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5843>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5844 = ((($5838)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5844>>2] = $5845; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2649>>2] = $5838; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5846 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5847 = HEAP32[$5846>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5848 = ($5847|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5848) {
   $5851 = $5838;
  } else {
   $5849 = $5847; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5849; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2651 = HEAP32[$$0$i7$i2649>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5851 = $$pre$i$i2651;
  }
  $5850 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5850,$5851); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5852 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5853 = (($5852) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5853; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4552 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5856 = $$pre4552;$__r$0$i2653 = $5838;
 }
 $5854 = ((($__r$0$i2653)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5854>>2] = 177; //@line 799 "modules/event/sdl/Event.cpp"
 $5855 = ($5856|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5855) {
   HEAP32[$__parent$i2656>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5868 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2664 = $2;$5878 = $5868;
   label = 2490;
  } else {
   $__nd$0$i$i2658 = $5856;
   while(1) {
    $5857 = ((($__nd$0$i$i2658)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5858 = HEAP32[$5857>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5859 = ($5858|0)>(1073742094); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5859) {
     $5860 = HEAP32[$__nd$0$i$i2658>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5861 = ($5860|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5861) {
      $$lcssa6892 = $__nd$0$i$i2658;$__nd$0$i$i2658$lcssa6890 = $__nd$0$i$i2658;
      label = 2483;
      break;
     } else {
      $$sink$i$i2659 = $5860;
     }
    } else {
     $5863 = ($5858|0)<(1073742094); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5863)) {
      $__nd$0$i$i2658$lcssa = $__nd$0$i$i2658;
      label = 2489;
      break;
     }
     $5864 = ((($__nd$0$i$i2658)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5865 = HEAP32[$5864>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5866 = ($5865|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5866) {
      $$lcssa6891 = $5864;$__nd$0$i$i2658$lcssa6889 = $__nd$0$i$i2658;
      label = 2487;
      break;
     } else {
      $$sink$i$i2659 = $5865;
     }
    }
    $__nd$0$i$i2658 = $$sink$i$i2659;
   }
   if ((label|0) == 2483) {
    HEAP32[$__parent$i2656>>2] = $__nd$0$i$i2658$lcssa6890; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5862 = $__nd$0$i$i2658$lcssa6890; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2664 = $$lcssa6892;$5878 = $5862;
    label = 2490;
    break;
   }
   else if ((label|0) == 2487) {
    HEAP32[$__parent$i2656>>2] = $__nd$0$i$i2658$lcssa6889; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5867 = $__nd$0$i$i2658$lcssa6889; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2664 = $$lcssa6891;$5878 = $5867;
    label = 2490;
    break;
   }
   else if ((label|0) == 2489) {
    HEAP32[$__parent$i2656>>2] = $__nd$0$i$i2658$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5869 = $__nd$0$i$i2658$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5870 = ($__nd$0$i$i2658$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5870) {
     $$0$i7$i2664 = $__parent$i2656;$5878 = $5869;
     label = 2490;
     break;
    } else {
     $5889 = $5856;$__r$0$i2668 = $__nd$0$i$i2658$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2490) {
  __THREW__ = 0;
  $5871 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5872 = __THREW__; __THREW__ = 0;
  $5873 = $5872&1;
  if ($5873) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5874 = ((($5871)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5874>>2] = 1073742094; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5875 = ((($5871)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5875>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5871>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5876 = ((($5871)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5876>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5877 = ((($5871)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5877>>2] = $5878; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2664>>2] = $5871; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5879 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5880 = HEAP32[$5879>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5881 = ($5880|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5881) {
   $5884 = $5871;
  } else {
   $5882 = $5880; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5882; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2666 = HEAP32[$$0$i7$i2664>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5884 = $$pre$i$i2666;
  }
  $5883 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5883,$5884); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5885 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5886 = (($5885) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5886; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4553 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5889 = $$pre4553;$__r$0$i2668 = $5871;
 }
 $5887 = ((($__r$0$i2668)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5887>>2] = 178; //@line 800 "modules/event/sdl/Event.cpp"
 $5888 = ($5889|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5888) {
   HEAP32[$__parent$i2671>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5901 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2679 = $2;$5911 = $5901;
   label = 2504;
  } else {
   $__nd$0$i$i2673 = $5889;
   while(1) {
    $5890 = ((($__nd$0$i$i2673)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5891 = HEAP32[$5890>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5892 = ($5891|0)>(1073742095); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5892) {
     $5893 = HEAP32[$__nd$0$i$i2673>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5894 = ($5893|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5894) {
      $$lcssa6888 = $__nd$0$i$i2673;$__nd$0$i$i2673$lcssa6886 = $__nd$0$i$i2673;
      label = 2497;
      break;
     } else {
      $$sink$i$i2674 = $5893;
     }
    } else {
     $5896 = ($5891|0)<(1073742095); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5896)) {
      $__nd$0$i$i2673$lcssa = $__nd$0$i$i2673;
      label = 2503;
      break;
     }
     $5897 = ((($__nd$0$i$i2673)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5898 = HEAP32[$5897>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5899 = ($5898|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5899) {
      $$lcssa6887 = $5897;$__nd$0$i$i2673$lcssa6885 = $__nd$0$i$i2673;
      label = 2501;
      break;
     } else {
      $$sink$i$i2674 = $5898;
     }
    }
    $__nd$0$i$i2673 = $$sink$i$i2674;
   }
   if ((label|0) == 2497) {
    HEAP32[$__parent$i2671>>2] = $__nd$0$i$i2673$lcssa6886; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5895 = $__nd$0$i$i2673$lcssa6886; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2679 = $$lcssa6888;$5911 = $5895;
    label = 2504;
    break;
   }
   else if ((label|0) == 2501) {
    HEAP32[$__parent$i2671>>2] = $__nd$0$i$i2673$lcssa6885; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5900 = $__nd$0$i$i2673$lcssa6885; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2679 = $$lcssa6887;$5911 = $5900;
    label = 2504;
    break;
   }
   else if ((label|0) == 2503) {
    HEAP32[$__parent$i2671>>2] = $__nd$0$i$i2673$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5902 = $__nd$0$i$i2673$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5903 = ($__nd$0$i$i2673$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5903) {
     $$0$i7$i2679 = $__parent$i2671;$5911 = $5902;
     label = 2504;
     break;
    } else {
     $5922 = $5889;$__r$0$i2683 = $__nd$0$i$i2673$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2504) {
  __THREW__ = 0;
  $5904 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5905 = __THREW__; __THREW__ = 0;
  $5906 = $5905&1;
  if ($5906) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5907 = ((($5904)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5907>>2] = 1073742095; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5908 = ((($5904)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5908>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5904>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5909 = ((($5904)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5909>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5910 = ((($5904)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5910>>2] = $5911; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2679>>2] = $5904; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5912 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5913 = HEAP32[$5912>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5914 = ($5913|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5914) {
   $5917 = $5904;
  } else {
   $5915 = $5913; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5915; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2681 = HEAP32[$$0$i7$i2679>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5917 = $$pre$i$i2681;
  }
  $5916 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5916,$5917); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5918 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5919 = (($5918) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5919; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4554 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5922 = $$pre4554;$__r$0$i2683 = $5904;
 }
 $5920 = ((($__r$0$i2683)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5920>>2] = 179; //@line 801 "modules/event/sdl/Event.cpp"
 $5921 = ($5922|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5921) {
   HEAP32[$__parent$i2686>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5934 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2694 = $2;$5944 = $5934;
   label = 2518;
  } else {
   $__nd$0$i$i2688 = $5922;
   while(1) {
    $5923 = ((($__nd$0$i$i2688)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5924 = HEAP32[$5923>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5925 = ($5924|0)>(1073742096); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5925) {
     $5926 = HEAP32[$__nd$0$i$i2688>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5927 = ($5926|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5927) {
      $$lcssa6884 = $__nd$0$i$i2688;$__nd$0$i$i2688$lcssa6882 = $__nd$0$i$i2688;
      label = 2511;
      break;
     } else {
      $$sink$i$i2689 = $5926;
     }
    } else {
     $5929 = ($5924|0)<(1073742096); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5929)) {
      $__nd$0$i$i2688$lcssa = $__nd$0$i$i2688;
      label = 2517;
      break;
     }
     $5930 = ((($__nd$0$i$i2688)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5931 = HEAP32[$5930>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5932 = ($5931|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5932) {
      $$lcssa6883 = $5930;$__nd$0$i$i2688$lcssa6881 = $__nd$0$i$i2688;
      label = 2515;
      break;
     } else {
      $$sink$i$i2689 = $5931;
     }
    }
    $__nd$0$i$i2688 = $$sink$i$i2689;
   }
   if ((label|0) == 2511) {
    HEAP32[$__parent$i2686>>2] = $__nd$0$i$i2688$lcssa6882; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5928 = $__nd$0$i$i2688$lcssa6882; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2694 = $$lcssa6884;$5944 = $5928;
    label = 2518;
    break;
   }
   else if ((label|0) == 2515) {
    HEAP32[$__parent$i2686>>2] = $__nd$0$i$i2688$lcssa6881; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5933 = $__nd$0$i$i2688$lcssa6881; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2694 = $$lcssa6883;$5944 = $5933;
    label = 2518;
    break;
   }
   else if ((label|0) == 2517) {
    HEAP32[$__parent$i2686>>2] = $__nd$0$i$i2688$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5935 = $__nd$0$i$i2688$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5936 = ($__nd$0$i$i2688$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5936) {
     $$0$i7$i2694 = $__parent$i2686;$5944 = $5935;
     label = 2518;
     break;
    } else {
     $5955 = $5922;$__r$0$i2698 = $__nd$0$i$i2688$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2518) {
  __THREW__ = 0;
  $5937 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5938 = __THREW__; __THREW__ = 0;
  $5939 = $5938&1;
  if ($5939) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5940 = ((($5937)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5940>>2] = 1073742096; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5941 = ((($5937)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5941>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5937>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5942 = ((($5937)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5942>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5943 = ((($5937)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5943>>2] = $5944; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2694>>2] = $5937; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5945 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5946 = HEAP32[$5945>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5947 = ($5946|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5947) {
   $5950 = $5937;
  } else {
   $5948 = $5946; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5948; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2696 = HEAP32[$$0$i7$i2694>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5950 = $$pre$i$i2696;
  }
  $5949 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5949,$5950); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5951 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5952 = (($5951) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5952; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4555 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5955 = $$pre4555;$__r$0$i2698 = $5937;
 }
 $5953 = ((($__r$0$i2698)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5953>>2] = 180; //@line 802 "modules/event/sdl/Event.cpp"
 $5954 = ($5955|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5954) {
   HEAP32[$__parent$i2701>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $5967 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2709 = $2;$5977 = $5967;
   label = 2532;
  } else {
   $__nd$0$i$i2703 = $5955;
   while(1) {
    $5956 = ((($__nd$0$i$i2703)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5957 = HEAP32[$5956>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5958 = ($5957|0)>(1073742097); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5958) {
     $5959 = HEAP32[$__nd$0$i$i2703>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5960 = ($5959|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5960) {
      $$lcssa6880 = $__nd$0$i$i2703;$__nd$0$i$i2703$lcssa6878 = $__nd$0$i$i2703;
      label = 2525;
      break;
     } else {
      $$sink$i$i2704 = $5959;
     }
    } else {
     $5962 = ($5957|0)<(1073742097); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5962)) {
      $__nd$0$i$i2703$lcssa = $__nd$0$i$i2703;
      label = 2531;
      break;
     }
     $5963 = ((($__nd$0$i$i2703)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5964 = HEAP32[$5963>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5965 = ($5964|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5965) {
      $$lcssa6879 = $5963;$__nd$0$i$i2703$lcssa6877 = $__nd$0$i$i2703;
      label = 2529;
      break;
     } else {
      $$sink$i$i2704 = $5964;
     }
    }
    $__nd$0$i$i2703 = $$sink$i$i2704;
   }
   if ((label|0) == 2525) {
    HEAP32[$__parent$i2701>>2] = $__nd$0$i$i2703$lcssa6878; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5961 = $__nd$0$i$i2703$lcssa6878; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2709 = $$lcssa6880;$5977 = $5961;
    label = 2532;
    break;
   }
   else if ((label|0) == 2529) {
    HEAP32[$__parent$i2701>>2] = $__nd$0$i$i2703$lcssa6877; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5966 = $__nd$0$i$i2703$lcssa6877; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2709 = $$lcssa6879;$5977 = $5966;
    label = 2532;
    break;
   }
   else if ((label|0) == 2531) {
    HEAP32[$__parent$i2701>>2] = $__nd$0$i$i2703$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5968 = $__nd$0$i$i2703$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5969 = ($__nd$0$i$i2703$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($5969) {
     $$0$i7$i2709 = $__parent$i2701;$5977 = $5968;
     label = 2532;
     break;
    } else {
     $5988 = $5955;$__r$0$i2713 = $__nd$0$i$i2703$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2532) {
  __THREW__ = 0;
  $5970 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $5971 = __THREW__; __THREW__ = 0;
  $5972 = $5971&1;
  if ($5972) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $5973 = ((($5970)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5973>>2] = 1073742097; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5974 = ((($5970)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$5974>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$5970>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5975 = ((($5970)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5975>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5976 = ((($5970)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$5976>>2] = $5977; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2709>>2] = $5970; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5978 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5979 = HEAP32[$5978>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5980 = ($5979|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($5980) {
   $5983 = $5970;
  } else {
   $5981 = $5979; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $5981; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2711 = HEAP32[$$0$i7$i2709>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $5983 = $$pre$i$i2711;
  }
  $5982 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($5982,$5983); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5984 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5985 = (($5984) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $5985; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4556 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $5988 = $$pre4556;$__r$0$i2713 = $5970;
 }
 $5986 = ((($__r$0$i2713)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$5986>>2] = 181; //@line 803 "modules/event/sdl/Event.cpp"
 $5987 = ($5988|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($5987) {
   HEAP32[$__parent$i2716>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $6000 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2724 = $2;$6010 = $6000;
   label = 2546;
  } else {
   $__nd$0$i$i2718 = $5988;
   while(1) {
    $5989 = ((($__nd$0$i$i2718)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5990 = HEAP32[$5989>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5991 = ($5990|0)>(1073742098); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($5991) {
     $5992 = HEAP32[$__nd$0$i$i2718>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5993 = ($5992|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5993) {
      $$lcssa6876 = $__nd$0$i$i2718;$__nd$0$i$i2718$lcssa6874 = $__nd$0$i$i2718;
      label = 2539;
      break;
     } else {
      $$sink$i$i2719 = $5992;
     }
    } else {
     $5995 = ($5990|0)<(1073742098); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($5995)) {
      $__nd$0$i$i2718$lcssa = $__nd$0$i$i2718;
      label = 2545;
      break;
     }
     $5996 = ((($__nd$0$i$i2718)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5997 = HEAP32[$5996>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $5998 = ($5997|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($5998) {
      $$lcssa6875 = $5996;$__nd$0$i$i2718$lcssa6873 = $__nd$0$i$i2718;
      label = 2543;
      break;
     } else {
      $$sink$i$i2719 = $5997;
     }
    }
    $__nd$0$i$i2718 = $$sink$i$i2719;
   }
   if ((label|0) == 2539) {
    HEAP32[$__parent$i2716>>2] = $__nd$0$i$i2718$lcssa6874; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5994 = $__nd$0$i$i2718$lcssa6874; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2724 = $$lcssa6876;$6010 = $5994;
    label = 2546;
    break;
   }
   else if ((label|0) == 2543) {
    HEAP32[$__parent$i2716>>2] = $__nd$0$i$i2718$lcssa6873; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5999 = $__nd$0$i$i2718$lcssa6873; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2724 = $$lcssa6875;$6010 = $5999;
    label = 2546;
    break;
   }
   else if ((label|0) == 2545) {
    HEAP32[$__parent$i2716>>2] = $__nd$0$i$i2718$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6001 = $__nd$0$i$i2718$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6002 = ($__nd$0$i$i2718$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($6002) {
     $$0$i7$i2724 = $__parent$i2716;$6010 = $6001;
     label = 2546;
     break;
    } else {
     $6021 = $5988;$__r$0$i2728 = $__nd$0$i$i2718$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2546) {
  __THREW__ = 0;
  $6003 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $6004 = __THREW__; __THREW__ = 0;
  $6005 = $6004&1;
  if ($6005) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $6006 = ((($6003)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$6006>>2] = 1073742098; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $6007 = ((($6003)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$6007>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$6003>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6008 = ((($6003)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$6008>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6009 = ((($6003)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$6009>>2] = $6010; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2724>>2] = $6003; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6011 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6012 = HEAP32[$6011>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6013 = ($6012|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($6013) {
   $6016 = $6003;
  } else {
   $6014 = $6012; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $6014; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2726 = HEAP32[$$0$i7$i2724>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $6016 = $$pre$i$i2726;
  }
  $6015 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($6015,$6016); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6017 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6018 = (($6017) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $6018; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4557 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6021 = $$pre4557;$__r$0$i2728 = $6003;
 }
 $6019 = ((($__r$0$i2728)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$6019>>2] = 182; //@line 804 "modules/event/sdl/Event.cpp"
 $6020 = ($6021|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($6020) {
   HEAP32[$__parent$i2731>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $6033 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2739 = $2;$6043 = $6033;
   label = 2560;
  } else {
   $__nd$0$i$i2733 = $6021;
   while(1) {
    $6022 = ((($__nd$0$i$i2733)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6023 = HEAP32[$6022>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $6024 = ($6023|0)>(1073742099); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($6024) {
     $6025 = HEAP32[$__nd$0$i$i2733>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6026 = ($6025|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($6026) {
      $$lcssa6872 = $__nd$0$i$i2733;$__nd$0$i$i2733$lcssa6870 = $__nd$0$i$i2733;
      label = 2553;
      break;
     } else {
      $$sink$i$i2734 = $6025;
     }
    } else {
     $6028 = ($6023|0)<(1073742099); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($6028)) {
      $__nd$0$i$i2733$lcssa = $__nd$0$i$i2733;
      label = 2559;
      break;
     }
     $6029 = ((($__nd$0$i$i2733)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6030 = HEAP32[$6029>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6031 = ($6030|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($6031) {
      $$lcssa6871 = $6029;$__nd$0$i$i2733$lcssa6869 = $__nd$0$i$i2733;
      label = 2557;
      break;
     } else {
      $$sink$i$i2734 = $6030;
     }
    }
    $__nd$0$i$i2733 = $$sink$i$i2734;
   }
   if ((label|0) == 2553) {
    HEAP32[$__parent$i2731>>2] = $__nd$0$i$i2733$lcssa6870; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6027 = $__nd$0$i$i2733$lcssa6870; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2739 = $$lcssa6872;$6043 = $6027;
    label = 2560;
    break;
   }
   else if ((label|0) == 2557) {
    HEAP32[$__parent$i2731>>2] = $__nd$0$i$i2733$lcssa6869; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6032 = $__nd$0$i$i2733$lcssa6869; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2739 = $$lcssa6871;$6043 = $6032;
    label = 2560;
    break;
   }
   else if ((label|0) == 2559) {
    HEAP32[$__parent$i2731>>2] = $__nd$0$i$i2733$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6034 = $__nd$0$i$i2733$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6035 = ($__nd$0$i$i2733$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($6035) {
     $$0$i7$i2739 = $__parent$i2731;$6043 = $6034;
     label = 2560;
     break;
    } else {
     $6054 = $6021;$__r$0$i2743 = $__nd$0$i$i2733$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2560) {
  __THREW__ = 0;
  $6036 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $6037 = __THREW__; __THREW__ = 0;
  $6038 = $6037&1;
  if ($6038) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $6039 = ((($6036)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$6039>>2] = 1073742099; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $6040 = ((($6036)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$6040>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$6036>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6041 = ((($6036)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$6041>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6042 = ((($6036)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$6042>>2] = $6043; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2739>>2] = $6036; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6044 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6045 = HEAP32[$6044>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6046 = ($6045|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($6046) {
   $6049 = $6036;
  } else {
   $6047 = $6045; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $6047; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2741 = HEAP32[$$0$i7$i2739>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $6049 = $$pre$i$i2741;
  }
  $6048 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($6048,$6049); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6050 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6051 = (($6050) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $6051; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4558 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6054 = $$pre4558;$__r$0$i2743 = $6036;
 }
 $6052 = ((($__r$0$i2743)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$6052>>2] = 183; //@line 806 "modules/event/sdl/Event.cpp"
 $6053 = ($6054|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($6053) {
   HEAP32[$__parent$i2746>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $6066 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2754 = $2;$6076 = $6066;
   label = 2574;
  } else {
   $__nd$0$i$i2748 = $6054;
   while(1) {
    $6055 = ((($__nd$0$i$i2748)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6056 = HEAP32[$6055>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $6057 = ($6056|0)>(1073742100); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($6057) {
     $6058 = HEAP32[$__nd$0$i$i2748>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6059 = ($6058|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($6059) {
      $$lcssa6868 = $__nd$0$i$i2748;$__nd$0$i$i2748$lcssa6866 = $__nd$0$i$i2748;
      label = 2567;
      break;
     } else {
      $$sink$i$i2749 = $6058;
     }
    } else {
     $6061 = ($6056|0)<(1073742100); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($6061)) {
      $__nd$0$i$i2748$lcssa = $__nd$0$i$i2748;
      label = 2573;
      break;
     }
     $6062 = ((($__nd$0$i$i2748)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6063 = HEAP32[$6062>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6064 = ($6063|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($6064) {
      $$lcssa6867 = $6062;$__nd$0$i$i2748$lcssa6865 = $__nd$0$i$i2748;
      label = 2571;
      break;
     } else {
      $$sink$i$i2749 = $6063;
     }
    }
    $__nd$0$i$i2748 = $$sink$i$i2749;
   }
   if ((label|0) == 2567) {
    HEAP32[$__parent$i2746>>2] = $__nd$0$i$i2748$lcssa6866; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6060 = $__nd$0$i$i2748$lcssa6866; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2754 = $$lcssa6868;$6076 = $6060;
    label = 2574;
    break;
   }
   else if ((label|0) == 2571) {
    HEAP32[$__parent$i2746>>2] = $__nd$0$i$i2748$lcssa6865; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6065 = $__nd$0$i$i2748$lcssa6865; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2754 = $$lcssa6867;$6076 = $6065;
    label = 2574;
    break;
   }
   else if ((label|0) == 2573) {
    HEAP32[$__parent$i2746>>2] = $__nd$0$i$i2748$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6067 = $__nd$0$i$i2748$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6068 = ($__nd$0$i$i2748$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($6068) {
     $$0$i7$i2754 = $__parent$i2746;$6076 = $6067;
     label = 2574;
     break;
    } else {
     $6087 = $6054;$__r$0$i2758 = $__nd$0$i$i2748$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2574) {
  __THREW__ = 0;
  $6069 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $6070 = __THREW__; __THREW__ = 0;
  $6071 = $6070&1;
  if ($6071) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $6072 = ((($6069)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$6072>>2] = 1073742100; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $6073 = ((($6069)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$6073>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$6069>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6074 = ((($6069)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$6074>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6075 = ((($6069)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$6075>>2] = $6076; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2754>>2] = $6069; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6077 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6078 = HEAP32[$6077>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6079 = ($6078|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($6079) {
   $6082 = $6069;
  } else {
   $6080 = $6078; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $6080; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2756 = HEAP32[$$0$i7$i2754>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $6082 = $$pre$i$i2756;
  }
  $6081 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($6081,$6082); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6083 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6084 = (($6083) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $6084; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4559 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6087 = $$pre4559;$__r$0$i2758 = $6069;
 }
 $6085 = ((($__r$0$i2758)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$6085>>2] = 184; //@line 807 "modules/event/sdl/Event.cpp"
 $6086 = ($6087|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($6086) {
   HEAP32[$__parent$i2761>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $6099 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2769 = $2;$6109 = $6099;
   label = 2588;
  } else {
   $__nd$0$i$i2763 = $6087;
   while(1) {
    $6088 = ((($__nd$0$i$i2763)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6089 = HEAP32[$6088>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $6090 = ($6089|0)>(1073742101); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($6090) {
     $6091 = HEAP32[$__nd$0$i$i2763>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6092 = ($6091|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($6092) {
      $$lcssa6864 = $__nd$0$i$i2763;$__nd$0$i$i2763$lcssa6862 = $__nd$0$i$i2763;
      label = 2581;
      break;
     } else {
      $$sink$i$i2764 = $6091;
     }
    } else {
     $6094 = ($6089|0)<(1073742101); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($6094)) {
      $__nd$0$i$i2763$lcssa = $__nd$0$i$i2763;
      label = 2587;
      break;
     }
     $6095 = ((($__nd$0$i$i2763)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6096 = HEAP32[$6095>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6097 = ($6096|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($6097) {
      $$lcssa6863 = $6095;$__nd$0$i$i2763$lcssa6861 = $__nd$0$i$i2763;
      label = 2585;
      break;
     } else {
      $$sink$i$i2764 = $6096;
     }
    }
    $__nd$0$i$i2763 = $$sink$i$i2764;
   }
   if ((label|0) == 2581) {
    HEAP32[$__parent$i2761>>2] = $__nd$0$i$i2763$lcssa6862; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6093 = $__nd$0$i$i2763$lcssa6862; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2769 = $$lcssa6864;$6109 = $6093;
    label = 2588;
    break;
   }
   else if ((label|0) == 2585) {
    HEAP32[$__parent$i2761>>2] = $__nd$0$i$i2763$lcssa6861; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6098 = $__nd$0$i$i2763$lcssa6861; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2769 = $$lcssa6863;$6109 = $6098;
    label = 2588;
    break;
   }
   else if ((label|0) == 2587) {
    HEAP32[$__parent$i2761>>2] = $__nd$0$i$i2763$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6100 = $__nd$0$i$i2763$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6101 = ($__nd$0$i$i2763$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($6101) {
     $$0$i7$i2769 = $__parent$i2761;$6109 = $6100;
     label = 2588;
     break;
    } else {
     $6120 = $6087;$__r$0$i2773 = $__nd$0$i$i2763$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2588) {
  __THREW__ = 0;
  $6102 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $6103 = __THREW__; __THREW__ = 0;
  $6104 = $6103&1;
  if ($6104) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $6105 = ((($6102)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$6105>>2] = 1073742101; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $6106 = ((($6102)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$6106>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$6102>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6107 = ((($6102)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$6107>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6108 = ((($6102)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$6108>>2] = $6109; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2769>>2] = $6102; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6110 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6111 = HEAP32[$6110>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6112 = ($6111|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($6112) {
   $6115 = $6102;
  } else {
   $6113 = $6111; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $6113; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2771 = HEAP32[$$0$i7$i2769>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $6115 = $$pre$i$i2771;
  }
  $6114 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($6114,$6115); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6116 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6117 = (($6116) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $6117; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4560 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6120 = $$pre4560;$__r$0$i2773 = $6102;
 }
 $6118 = ((($__r$0$i2773)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$6118>>2] = 185; //@line 808 "modules/event/sdl/Event.cpp"
 $6119 = ($6120|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($6119) {
   HEAP32[$__parent$i2776>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $6132 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2784 = $2;$6142 = $6132;
   label = 2602;
  } else {
   $__nd$0$i$i2778 = $6120;
   while(1) {
    $6121 = ((($__nd$0$i$i2778)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6122 = HEAP32[$6121>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $6123 = ($6122|0)>(1073742102); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($6123) {
     $6124 = HEAP32[$__nd$0$i$i2778>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6125 = ($6124|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($6125) {
      $$lcssa6860 = $__nd$0$i$i2778;$__nd$0$i$i2778$lcssa6858 = $__nd$0$i$i2778;
      label = 2595;
      break;
     } else {
      $$sink$i$i2779 = $6124;
     }
    } else {
     $6127 = ($6122|0)<(1073742102); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($6127)) {
      $__nd$0$i$i2778$lcssa = $__nd$0$i$i2778;
      label = 2601;
      break;
     }
     $6128 = ((($__nd$0$i$i2778)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6129 = HEAP32[$6128>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6130 = ($6129|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($6130) {
      $$lcssa6859 = $6128;$__nd$0$i$i2778$lcssa6857 = $__nd$0$i$i2778;
      label = 2599;
      break;
     } else {
      $$sink$i$i2779 = $6129;
     }
    }
    $__nd$0$i$i2778 = $$sink$i$i2779;
   }
   if ((label|0) == 2595) {
    HEAP32[$__parent$i2776>>2] = $__nd$0$i$i2778$lcssa6858; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6126 = $__nd$0$i$i2778$lcssa6858; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2784 = $$lcssa6860;$6142 = $6126;
    label = 2602;
    break;
   }
   else if ((label|0) == 2599) {
    HEAP32[$__parent$i2776>>2] = $__nd$0$i$i2778$lcssa6857; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6131 = $__nd$0$i$i2778$lcssa6857; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2784 = $$lcssa6859;$6142 = $6131;
    label = 2602;
    break;
   }
   else if ((label|0) == 2601) {
    HEAP32[$__parent$i2776>>2] = $__nd$0$i$i2778$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6133 = $__nd$0$i$i2778$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6134 = ($__nd$0$i$i2778$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($6134) {
     $$0$i7$i2784 = $__parent$i2776;$6142 = $6133;
     label = 2602;
     break;
    } else {
     $6153 = $6120;$__r$0$i2788 = $__nd$0$i$i2778$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2602) {
  __THREW__ = 0;
  $6135 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $6136 = __THREW__; __THREW__ = 0;
  $6137 = $6136&1;
  if ($6137) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $6138 = ((($6135)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$6138>>2] = 1073742102; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $6139 = ((($6135)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$6139>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$6135>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6140 = ((($6135)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$6140>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6141 = ((($6135)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$6141>>2] = $6142; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2784>>2] = $6135; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6143 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6144 = HEAP32[$6143>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6145 = ($6144|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($6145) {
   $6148 = $6135;
  } else {
   $6146 = $6144; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $6146; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2786 = HEAP32[$$0$i7$i2784>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $6148 = $$pre$i$i2786;
  }
  $6147 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($6147,$6148); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6149 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6150 = (($6149) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $6150; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4561 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6153 = $$pre4561;$__r$0$i2788 = $6135;
 }
 $6151 = ((($__r$0$i2788)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$6151>>2] = 186; //@line 809 "modules/event/sdl/Event.cpp"
 $6152 = ($6153|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($6152) {
   HEAP32[$__parent$i2791>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $6165 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2799 = $2;$6175 = $6165;
   label = 2616;
  } else {
   $__nd$0$i$i2793 = $6153;
   while(1) {
    $6154 = ((($__nd$0$i$i2793)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6155 = HEAP32[$6154>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $6156 = ($6155|0)>(1073742103); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($6156) {
     $6157 = HEAP32[$__nd$0$i$i2793>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6158 = ($6157|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($6158) {
      $$lcssa6856 = $__nd$0$i$i2793;$__nd$0$i$i2793$lcssa6854 = $__nd$0$i$i2793;
      label = 2609;
      break;
     } else {
      $$sink$i$i2794 = $6157;
     }
    } else {
     $6160 = ($6155|0)<(1073742103); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($6160)) {
      $__nd$0$i$i2793$lcssa = $__nd$0$i$i2793;
      label = 2615;
      break;
     }
     $6161 = ((($__nd$0$i$i2793)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6162 = HEAP32[$6161>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6163 = ($6162|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($6163) {
      $$lcssa6855 = $6161;$__nd$0$i$i2793$lcssa6853 = $__nd$0$i$i2793;
      label = 2613;
      break;
     } else {
      $$sink$i$i2794 = $6162;
     }
    }
    $__nd$0$i$i2793 = $$sink$i$i2794;
   }
   if ((label|0) == 2609) {
    HEAP32[$__parent$i2791>>2] = $__nd$0$i$i2793$lcssa6854; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6159 = $__nd$0$i$i2793$lcssa6854; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2799 = $$lcssa6856;$6175 = $6159;
    label = 2616;
    break;
   }
   else if ((label|0) == 2613) {
    HEAP32[$__parent$i2791>>2] = $__nd$0$i$i2793$lcssa6853; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6164 = $__nd$0$i$i2793$lcssa6853; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2799 = $$lcssa6855;$6175 = $6164;
    label = 2616;
    break;
   }
   else if ((label|0) == 2615) {
    HEAP32[$__parent$i2791>>2] = $__nd$0$i$i2793$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6166 = $__nd$0$i$i2793$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6167 = ($__nd$0$i$i2793$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($6167) {
     $$0$i7$i2799 = $__parent$i2791;$6175 = $6166;
     label = 2616;
     break;
    } else {
     $6186 = $6153;$__r$0$i2803 = $__nd$0$i$i2793$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2616) {
  __THREW__ = 0;
  $6168 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $6169 = __THREW__; __THREW__ = 0;
  $6170 = $6169&1;
  if ($6170) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $6171 = ((($6168)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$6171>>2] = 1073742103; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $6172 = ((($6168)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$6172>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$6168>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6173 = ((($6168)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$6173>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6174 = ((($6168)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$6174>>2] = $6175; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2799>>2] = $6168; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6176 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6177 = HEAP32[$6176>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6178 = ($6177|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($6178) {
   $6181 = $6168;
  } else {
   $6179 = $6177; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $6179; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2801 = HEAP32[$$0$i7$i2799>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $6181 = $$pre$i$i2801;
  }
  $6180 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($6180,$6181); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6182 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6183 = (($6182) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $6183; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4562 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6186 = $$pre4562;$__r$0$i2803 = $6168;
 }
 $6184 = ((($__r$0$i2803)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$6184>>2] = 187; //@line 810 "modules/event/sdl/Event.cpp"
 $6185 = ($6186|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($6185) {
   HEAP32[$__parent$i2806>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $6198 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2814 = $2;$6208 = $6198;
   label = 2630;
  } else {
   $__nd$0$i$i2808 = $6186;
   while(1) {
    $6187 = ((($__nd$0$i$i2808)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6188 = HEAP32[$6187>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $6189 = ($6188|0)>(1073742104); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($6189) {
     $6190 = HEAP32[$__nd$0$i$i2808>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6191 = ($6190|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($6191) {
      $$lcssa6852 = $__nd$0$i$i2808;$__nd$0$i$i2808$lcssa6850 = $__nd$0$i$i2808;
      label = 2623;
      break;
     } else {
      $$sink$i$i2809 = $6190;
     }
    } else {
     $6193 = ($6188|0)<(1073742104); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($6193)) {
      $__nd$0$i$i2808$lcssa = $__nd$0$i$i2808;
      label = 2629;
      break;
     }
     $6194 = ((($__nd$0$i$i2808)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6195 = HEAP32[$6194>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6196 = ($6195|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($6196) {
      $$lcssa6851 = $6194;$__nd$0$i$i2808$lcssa6849 = $__nd$0$i$i2808;
      label = 2627;
      break;
     } else {
      $$sink$i$i2809 = $6195;
     }
    }
    $__nd$0$i$i2808 = $$sink$i$i2809;
   }
   if ((label|0) == 2623) {
    HEAP32[$__parent$i2806>>2] = $__nd$0$i$i2808$lcssa6850; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6192 = $__nd$0$i$i2808$lcssa6850; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2814 = $$lcssa6852;$6208 = $6192;
    label = 2630;
    break;
   }
   else if ((label|0) == 2627) {
    HEAP32[$__parent$i2806>>2] = $__nd$0$i$i2808$lcssa6849; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6197 = $__nd$0$i$i2808$lcssa6849; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2814 = $$lcssa6851;$6208 = $6197;
    label = 2630;
    break;
   }
   else if ((label|0) == 2629) {
    HEAP32[$__parent$i2806>>2] = $__nd$0$i$i2808$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6199 = $__nd$0$i$i2808$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6200 = ($__nd$0$i$i2808$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($6200) {
     $$0$i7$i2814 = $__parent$i2806;$6208 = $6199;
     label = 2630;
     break;
    } else {
     $6219 = $6186;$__r$0$i2818 = $__nd$0$i$i2808$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2630) {
  __THREW__ = 0;
  $6201 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $6202 = __THREW__; __THREW__ = 0;
  $6203 = $6202&1;
  if ($6203) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $6204 = ((($6201)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$6204>>2] = 1073742104; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $6205 = ((($6201)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$6205>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$6201>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6206 = ((($6201)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$6206>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6207 = ((($6201)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$6207>>2] = $6208; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2814>>2] = $6201; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6209 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6210 = HEAP32[$6209>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6211 = ($6210|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($6211) {
   $6214 = $6201;
  } else {
   $6212 = $6210; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $6212; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2816 = HEAP32[$$0$i7$i2814>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $6214 = $$pre$i$i2816;
  }
  $6213 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($6213,$6214); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6215 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6216 = (($6215) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $6216; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4563 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6219 = $$pre4563;$__r$0$i2818 = $6201;
 }
 $6217 = ((($__r$0$i2818)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$6217>>2] = 188; //@line 811 "modules/event/sdl/Event.cpp"
 $6218 = ($6219|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($6218) {
   HEAP32[$__parent$i2821>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $6231 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2829 = $2;$6241 = $6231;
   label = 2644;
  } else {
   $__nd$0$i$i2823 = $6219;
   while(1) {
    $6220 = ((($__nd$0$i$i2823)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6221 = HEAP32[$6220>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $6222 = ($6221|0)>(1073742105); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($6222) {
     $6223 = HEAP32[$__nd$0$i$i2823>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6224 = ($6223|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($6224) {
      $$lcssa6848 = $__nd$0$i$i2823;$__nd$0$i$i2823$lcssa6846 = $__nd$0$i$i2823;
      label = 2637;
      break;
     } else {
      $$sink$i$i2824 = $6223;
     }
    } else {
     $6226 = ($6221|0)<(1073742105); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($6226)) {
      $__nd$0$i$i2823$lcssa = $__nd$0$i$i2823;
      label = 2643;
      break;
     }
     $6227 = ((($__nd$0$i$i2823)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6228 = HEAP32[$6227>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6229 = ($6228|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($6229) {
      $$lcssa6847 = $6227;$__nd$0$i$i2823$lcssa6845 = $__nd$0$i$i2823;
      label = 2641;
      break;
     } else {
      $$sink$i$i2824 = $6228;
     }
    }
    $__nd$0$i$i2823 = $$sink$i$i2824;
   }
   if ((label|0) == 2637) {
    HEAP32[$__parent$i2821>>2] = $__nd$0$i$i2823$lcssa6846; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6225 = $__nd$0$i$i2823$lcssa6846; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2829 = $$lcssa6848;$6241 = $6225;
    label = 2644;
    break;
   }
   else if ((label|0) == 2641) {
    HEAP32[$__parent$i2821>>2] = $__nd$0$i$i2823$lcssa6845; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6230 = $__nd$0$i$i2823$lcssa6845; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2829 = $$lcssa6847;$6241 = $6230;
    label = 2644;
    break;
   }
   else if ((label|0) == 2643) {
    HEAP32[$__parent$i2821>>2] = $__nd$0$i$i2823$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6232 = $__nd$0$i$i2823$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6233 = ($__nd$0$i$i2823$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($6233) {
     $$0$i7$i2829 = $__parent$i2821;$6241 = $6232;
     label = 2644;
     break;
    } else {
     $6252 = $6219;$__r$0$i2833 = $__nd$0$i$i2823$lcssa;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 2644) {
  __THREW__ = 0;
  $6234 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $6235 = __THREW__; __THREW__ = 0;
  $6236 = $6235&1;
  if ($6236) {
   $6284 = ___cxa_find_matching_catch_2()|0;
   $6285 = tempRet0;
   $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   ___resumeException($6284|0);
   // unreachable;
  }
  $6237 = ((($6234)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$6237>>2] = 1073742105; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $6238 = ((($6234)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  HEAP32[$6238>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$6234>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6239 = ((($6234)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$6239>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6240 = ((($6234)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$6240>>2] = $6241; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$$0$i7$i2829>>2] = $6234; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6242 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6243 = HEAP32[$6242>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6244 = ($6243|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($6244) {
   $6247 = $6234;
  } else {
   $6245 = $6243; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$agg$result>>2] = $6245; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $$pre$i$i2831 = HEAP32[$$0$i7$i2829>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $6247 = $$pre$i$i2831;
  }
  $6246 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($6246,$6247); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6248 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6249 = (($6248) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$1>>2] = $6249; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre4564 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6252 = $$pre4564;$__r$0$i2833 = $6234;
 }
 $6250 = ((($__r$0$i2833)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$6250>>2] = 189; //@line 812 "modules/event/sdl/Event.cpp"
 $6251 = ($6252|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($6251) {
   HEAP32[$__parent$i2836>>2] = $2; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $6264 = $2; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i2844 = $2;$6274 = $6264;
  } else {
   $__nd$0$i$i2838 = $6252;
   while(1) {
    $6253 = ((($__nd$0$i$i2838)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6254 = HEAP32[$6253>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $6255 = ($6254|0)>(1073742106); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($6255) {
     $6256 = HEAP32[$__nd$0$i$i2838>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6257 = ($6256|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($6257) {
      $$lcssa6844 = $__nd$0$i$i2838;$__nd$0$i$i2838$lcssa6843 = $__nd$0$i$i2838;
      label = 2651;
      break;
     } else {
      $$sink$i$i2839 = $6256;
     }
    } else {
     $6259 = ($6254|0)<(1073742106); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($6259)) {
      $__nd$0$i$i2838$lcssa = $__nd$0$i$i2838;
      label = 2657;
      break;
     }
     $6260 = ((($__nd$0$i$i2838)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6261 = HEAP32[$6260>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $6262 = ($6261|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($6262) {
      $$lcssa = $6260;$__nd$0$i$i2838$lcssa6842 = $__nd$0$i$i2838;
      label = 2655;
      break;
     } else {
      $$sink$i$i2839 = $6261;
     }
    }
    $__nd$0$i$i2838 = $$sink$i$i2839;
   }
   if ((label|0) == 2651) {
    HEAP32[$__parent$i2836>>2] = $__nd$0$i$i2838$lcssa6843; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6258 = $__nd$0$i$i2838$lcssa6843; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2844 = $$lcssa6844;$6274 = $6258;
    break;
   }
   else if ((label|0) == 2655) {
    HEAP32[$__parent$i2836>>2] = $__nd$0$i$i2838$lcssa6842; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6263 = $__nd$0$i$i2838$lcssa6842; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i2844 = $$lcssa;$6274 = $6263;
    break;
   }
   else if ((label|0) == 2657) {
    HEAP32[$__parent$i2836>>2] = $__nd$0$i$i2838$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6265 = $__nd$0$i$i2838$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $6266 = ($__nd$0$i$i2838$lcssa|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($6266) {
     $$0$i7$i2844 = $__parent$i2836;$6274 = $6265;
     break;
    } else {
     $__r$0$i2848 = $__nd$0$i$i2838$lcssa;
    }
    $6283 = ((($__r$0$i2848)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    HEAP32[$6283>>2] = 190; //@line 813 "modules/event/sdl/Event.cpp"
    STACKTOP = sp;return; //@line 820 "modules/event/sdl/Event.cpp"
   }
  }
 } while(0);
 __THREW__ = 0;
 $6267 = (invoke_ii(3348,24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $6268 = __THREW__; __THREW__ = 0;
 $6269 = $6268&1;
 if ($6269) {
  $6284 = ___cxa_find_matching_catch_2()|0;
  $6285 = tempRet0;
  $6286 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($agg$result,$6286); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  ___resumeException($6284|0);
  // unreachable;
 }
 $6270 = ((($6267)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$6270>>2] = 1073742106; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6271 = ((($6267)) + 20|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$6271>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$6267>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $6272 = ((($6267)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$6272>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $6273 = ((($6267)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$6273>>2] = $6274; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$$0$i7$i2844>>2] = $6267; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $6275 = HEAP32[$agg$result>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $6276 = HEAP32[$6275>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $6277 = ($6276|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($6277) {
  $6280 = $6267;
 } else {
  $6278 = $6276; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$agg$result>>2] = $6278; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre$i$i2846 = HEAP32[$$0$i7$i2844>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $6280 = $$pre$i$i2846;
 }
 $6279 = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($6279,$6280); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $6281 = HEAP32[$1>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $6282 = (($6281) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$1>>2] = $6282; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $__r$0$i2848 = $6267;
 $6283 = ((($__r$0$i2848)) + 20|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$6283>>2] = 190; //@line 813 "modules/event/sdl/Event.cpp"
 STACKTOP = sp;return; //@line 820 "modules/event/sdl/Event.cpp"
}
function __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($this,$__nd) {
 $this = $this|0;
 $__nd = $__nd|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__nd|0)==(0|0); //@line 1440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($0) {
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 } else {
  $1 = HEAP32[$__nd>>2]|0; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($this,$1); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2 = ((($__nd)) + 4|0); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3 = HEAP32[$2>>2]|0; //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($this,$3); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZdlPv($__nd); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
}
function __ZNSt3__13mapIiN4love8keyboard8Keyboard3KeyENS_4lessIiEENS_9allocatorINS_4pairIKiS4_EEEEED2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeIiN4love8keyboard8Keyboard3KeyEEENS_19__map_value_compareIiS6_NS_4lessIiEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($this,$1); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 return; //@line 783 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
}
function __ZN4love5event3sdl5EventC2Ev($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./common/Module.h"
 HEAP32[$this>>2] = (23140); //@line 59 "./modules/event/Event.h"
 $0 = ((($this)) + 8|0); //@line 59 "./modules/event/Event.h"
 __THREW__ = 0;
 invoke_vi(3788,($0|0)); //@line 59 "./modules/event/Event.h"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZN4love6ModuleD2Ev($this); //@line 59 "./modules/event/Event.h"
  ___resumeException($3|0);
  // unreachable;
 }
 $5 = ((($this)) + 12|0); //@line 59 "./modules/event/Event.h"
 ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;HEAP32[$5+12>>2]=0|0;HEAP32[$5+16>>2]=0|0;HEAP32[$5+20>>2]=0|0; //@line 1094 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 HEAP32[$this>>2] = (23072); //@line 99 "modules/event/sdl/Event.cpp"
 __THREW__ = 0;
 $6 = (invoke_ii(3813,16384)|0); //@line 100 "modules/event/sdl/Event.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 do {
  if (!($8)) {
   $9 = ($6|0)<(0); //@line 100 "modules/event/sdl/Event.cpp"
   if (!($9)) {
    __THREW__ = 0;
    invoke_vii(3959,(3935|0),($this|0)); //@line 103 "modules/event/sdl/Event.cpp"
    $21 = __THREW__; __THREW__ = 0;
    $22 = $21&1;
    if ($22) {
     break;
    }
    STACKTOP = sp;return; //@line 104 "modules/event/sdl/Event.cpp"
   }
   $10 = (___cxa_allocate_exception(16)|0); //@line 101 "modules/event/sdl/Event.cpp"
   __THREW__ = 0;
   $11 = (invoke_i(3814)|0); //@line 101 "modules/event/sdl/Event.cpp"
   $12 = __THREW__; __THREW__ = 0;
   $13 = $12&1;
   if (!($13)) {
    __THREW__ = 0;
    HEAP32[$vararg_buffer>>2] = $11; //@line 101 "modules/event/sdl/Event.cpp"
    invoke_viii(3349,($10|0),(365870|0),($vararg_buffer|0)); //@line 101 "modules/event/sdl/Event.cpp"
    $14 = __THREW__; __THREW__ = 0;
    $15 = $14&1;
    if (!($15)) {
     __THREW__ = 0;
     invoke_viii(3350,($10|0),(96|0),(344|0)); //@line 101 "modules/event/sdl/Event.cpp"
     $16 = __THREW__; __THREW__ = 0;
     break;
    }
   }
   $19 = ___cxa_find_matching_catch_2()|0;
   $20 = tempRet0;
   ___cxa_free_exception(($10|0)); //@line 101 "modules/event/sdl/Event.cpp"
   $$0 = $20;$$01 = $19;
   __ZN4love5event5EventD2Ev($this); //@line 104 "modules/event/sdl/Event.cpp"
   ___resumeException($$01|0);
   // unreachable;
  }
 } while(0);
 $17 = ___cxa_find_matching_catch_2()|0;
 $18 = tempRet0;
 $$0 = $18;$$01 = $17;
 __ZN4love5event5EventD2Ev($this); //@line 104 "modules/event/sdl/Event.cpp"
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN4love5event7MessageD2Ev($this) {
 $this = $this|0;
 var $$pre$i = 0, $$pre$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (23116); //@line 38 "modules/event/Event.cpp"
 $0 = ((($this)) + 20|0); //@line 39 "modules/event/Event.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ($1|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($2) {
  $20 = ((($this)) + 8|0); //@line 39 "modules/event/Event.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($20); //@line 39 "modules/event/Event.cpp"
  return; //@line 39 "modules/event/Event.cpp"
 }
 $3 = ((($this)) + 24|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$3>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($4|0)==($1|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($5) {
   $19 = $1;
  } else {
   $7 = $4;
   while(1) {
    $6 = ((($7)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$3>>2] = $6; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $8 = HEAP32[$6>>2]|0; //@line 106 "./common/Object.h"
    $9 = ($8|0)==(0|0); //@line 106 "./common/Object.h"
    if ($9) {
     $11 = $6;
    } else {
     $12 = HEAP32[$8>>2]|0; //@line 106 "./common/Object.h"
     $13 = ((($12)) + 12|0); //@line 106 "./common/Object.h"
     $14 = HEAP32[$13>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($14|0,($8|0)); //@line 106 "./common/Object.h"
     $15 = __THREW__; __THREW__ = 0;
     $16 = $15&1;
     if ($16) {
      label = 7;
      break;
     }
     $$pre$i$i$i = HEAP32[$3>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $11 = $$pre$i$i$i;
    }
    $10 = ($11|0)==($1|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($10) {
     label = 8;
     break;
    } else {
     $7 = $11;
    }
   }
   if ((label|0) == 7) {
    $17 = ___cxa_find_matching_catch_3(0|0)|0;
    $18 = tempRet0;
    ___clang_call_terminate($17); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 8) {
    $$pre$i = HEAP32[$0>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $19 = $$pre$i;
    break;
   }
  }
 } while(0);
 __ZdlPv($19); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $20 = ((($this)) + 8|0); //@line 39 "modules/event/Event.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($20); //@line 39 "modules/event/Event.cpp"
 return; //@line 39 "modules/event/Event.cpp"
}
function __ZN4love5event7MessageD0Ev($this) {
 $this = $this|0;
 var $$pre$i$i = 0, $$pre$i$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (23116); //@line 38 "modules/event/Event.cpp"
 $0 = ((($this)) + 20|0); //@line 39 "modules/event/Event.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ($1|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($2) {
  $20 = ((($this)) + 8|0); //@line 39 "modules/event/Event.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($20); //@line 39 "modules/event/Event.cpp"
  __ZdlPv($this); //@line 38 "modules/event/Event.cpp"
  return; //@line 39 "modules/event/Event.cpp"
 }
 $3 = ((($this)) + 24|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$3>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($4|0)==($1|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($5) {
   $19 = $1;
  } else {
   $7 = $4;
   while(1) {
    $6 = ((($7)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$3>>2] = $6; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $8 = HEAP32[$6>>2]|0; //@line 106 "./common/Object.h"
    $9 = ($8|0)==(0|0); //@line 106 "./common/Object.h"
    if ($9) {
     $11 = $6;
    } else {
     $12 = HEAP32[$8>>2]|0; //@line 106 "./common/Object.h"
     $13 = ((($12)) + 12|0); //@line 106 "./common/Object.h"
     $14 = HEAP32[$13>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($14|0,($8|0)); //@line 106 "./common/Object.h"
     $15 = __THREW__; __THREW__ = 0;
     $16 = $15&1;
     if ($16) {
      label = 7;
      break;
     }
     $$pre$i$i$i$i = HEAP32[$3>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $11 = $$pre$i$i$i$i;
    }
    $10 = ($11|0)==($1|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($10) {
     label = 8;
     break;
    } else {
     $7 = $11;
    }
   }
   if ((label|0) == 7) {
    $17 = ___cxa_find_matching_catch_3(0|0)|0;
    $18 = tempRet0;
    ___clang_call_terminate($17); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 8) {
    $$pre$i$i = HEAP32[$0>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $19 = $$pre$i$i;
    break;
   }
  }
 } while(0);
 __ZdlPv($19); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $20 = ((($this)) + 8|0); //@line 39 "modules/event/Event.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($20); //@line 39 "modules/event/Event.cpp"
 __ZdlPv($this); //@line 38 "modules/event/Event.cpp"
 return; //@line 39 "modules/event/Event.cpp"
}
function __ZN4love5event5EventD2Ev($this) {
 $this = $this|0;
 var $$cast$i = 0, $$in$i = 0, $$lcssa$i = 0, $$pre$i$i = 0, $$pre2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__i$01$i$i = 0, $__i$sroa$0$0$ph$i = 0, $__i$sroa$5$0$i = 0, $__i$sroa$5$0$ph$i = 0, $__new_last2$i$i$i$i$i$i = 0, $phitmp$i = 0, $scevgep$i$i$i$i$i$i = 0, $scevgep4$i$i$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (23140); //@line 85 "modules/event/Event.cpp"
 $0 = ((($this)) + 16|0); //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $1 = HEAP32[$0>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $2 = ((($this)) + 28|0); //@line 1060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $3 = HEAP32[$2>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $4 = $3 >>> 10; //@line 1060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $5 = (($1) + ($4<<2)|0); //@line 1060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $6 = ((($this)) + 20|0); //@line 85 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $7 = HEAP32[$6>>2]|0; //@line 85 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $8 = ($7|0)==($1|0); //@line 85 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $9 = $7; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $10 = $1; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($8) {
  $11 = ((($this)) + 32|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $25 = 0;$26 = $11;$59 = 0;
 } else {
  $12 = HEAP32[$5>>2]|0; //@line 1061 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $13 = $3 & 1023; //@line 1061 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $14 = (($12) + ($13<<2)|0); //@line 1061 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $phitmp$i = $14; //@line 1061 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $15 = ((($this)) + 32|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $16 = HEAP32[$15>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $17 = (($16) + ($3))|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $18 = $17 >>> 10; //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $19 = (($1) + ($18<<2)|0); //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $20 = HEAP32[$19>>2]|0; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $21 = $17 & 1023; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $22 = (($20) + ($21<<2)|0); //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $25 = $22;$26 = $15;$59 = $phitmp$i;
 }
 $__i$sroa$0$0$ph$i = $5;$__i$sroa$5$0$ph$i = $59;
 L5: while(1) {
  $__i$sroa$5$0$i = $__i$sroa$5$0$ph$i;
  while(1) {
   $23 = $__i$sroa$5$0$i; //@line 390 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $24 = ($23|0)==($25|0); //@line 394 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   if ($24) {
    break L5;
   }
   $30 = ((($23)) + 4|0); //@line 299 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $31 = $30; //@line 299 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $32 = HEAP32[$__i$sroa$0$0$ph$i>>2]|0; //@line 299 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $33 = (($31) - ($32))|0; //@line 299 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $34 = ($33|0)==(4096); //@line 299 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   if ($34) {
    break;
   } else {
    $__i$sroa$5$0$i = $31;
   }
  }
  $35 = ((($__i$sroa$0$0$ph$i)) + 4|0); //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $36 = HEAP32[$35>>2]|0; //@line 302 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $__i$sroa$0$0$ph$i = $35;$__i$sroa$5$0$ph$i = $36;
 }
 HEAP32[$26>>2] = 0; //@line 1164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $27 = (($9) - ($10))|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27 >> 2; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = ($28>>>0)>(2); //@line 1165 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($29) {
  $$in$i = $1;
  while(1) {
   $37 = HEAP32[$$in$i>>2]|0; //@line 1167 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   __ZdlPv($37); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $38 = HEAP32[$0>>2]|0; //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $39 = ((($38)) + 4|0); //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   HEAP32[$0>>2] = $39; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $40 = HEAP32[$6>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $$cast$i = $39; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $41 = (($40) - ($$cast$i))|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $42 = $41 >> 2; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $43 = ($42>>>0)>(2); //@line 1165 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   if ($43) {
    $$in$i = $39;
   } else {
    $$lcssa$i = $42;
    break;
   }
  }
 } else {
  $$lcssa$i = $28;
 }
 switch ($$lcssa$i|0) {
 case 1:  {
  HEAP32[$2>>2] = 512; //@line 1173 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  break;
 }
 case 2:  {
  HEAP32[$2>>2] = 1024; //@line 1176 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  break;
 }
 default: {
 }
 }
 $44 = HEAP32[$0>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $45 = HEAP32[$6>>2]|0; //@line 78 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $46 = ($44|0)==($45|0); //@line 1107 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if (!($46)) {
  $__i$01$i$i = $44;
  while(1) {
   $47 = HEAP32[$__i$01$i$i>>2]|0; //@line 1108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   __ZdlPv($47); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $48 = ((($__i$01$i$i)) + 4|0); //@line 1107 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $49 = ($48|0)==($45|0); //@line 1107 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   if ($49) {
    break;
   } else {
    $__i$01$i$i = $48;
   }
  }
  $$pre$i$i = HEAP32[$0>>2]|0; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $$pre2$i$i = HEAP32[$6>>2]|0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $50 = ($$pre2$i$i|0)==($$pre$i$i|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if (!($50)) {
   $__new_last2$i$i$i$i$i$i = $$pre$i$i; //@line 136 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $scevgep$i$i$i$i$i$i = ((($$pre2$i$i)) + -4|0);
   $51 = $scevgep$i$i$i$i$i$i;
   $52 = (($51) - ($__new_last2$i$i$i$i$i$i))|0;
   $53 = $52 >>> 2;
   $54 = $53 ^ -1;
   $scevgep4$i$i$i$i$i$i = (($$pre2$i$i) + ($54<<2)|0);
   HEAP32[$6>>2] = $scevgep4$i$i$i$i$i$i; //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  }
 }
 $55 = ((($this)) + 12|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $56 = HEAP32[$55>>2]|0; //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $57 = ($56|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($57) {
  $58 = ((($this)) + 8|0); //@line 86 "modules/event/Event.cpp"
  __ZN4love6thread8MutexRefD2Ev($58); //@line 86 "modules/event/Event.cpp"
  __ZN4love6ModuleD2Ev($this); //@line 86 "modules/event/Event.cpp"
  return; //@line 86 "modules/event/Event.cpp"
 }
 __ZdlPv($56); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $58 = ((($this)) + 8|0); //@line 86 "modules/event/Event.cpp"
 __ZN4love6thread8MutexRefD2Ev($58); //@line 86 "modules/event/Event.cpp"
 __ZN4love6ModuleD2Ev($this); //@line 86 "modules/event/Event.cpp"
 return; //@line 86 "modules/event/Event.cpp"
}
function __ZN4love5event5EventD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5event5EventD2Ev($this); //@line 85 "modules/event/Event.cpp"
 __ZdlPv($this); //@line 85 "modules/event/Event.cpp"
 return; //@line 86 "modules/event/Event.cpp"
}
function __ZN4love5event5Event5clearEv($this) {
 $this = $this|0;
 var $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lock = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ((($this)) + 8|0); //@line 107 "modules/event/Event.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 107 "modules/event/Event.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$1); //@line 107 "modules/event/Event.cpp"
 $2 = ((($this)) + 32|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $4 = ($3|0)==(0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($4) {
  __ZN4love6thread4LockD2Ev($lock); //@line 114 "modules/event/Event.cpp"
  STACKTOP = sp;return; //@line 114 "modules/event/Event.cpp"
 }
 $5 = ((($this)) + 16|0); //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $6 = ((($this)) + 28|0); //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $$pre = HEAP32[$6>>2]|0; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $9 = $$pre;
 while(1) {
  $7 = HEAP32[$5>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $8 = $9 >>> 10; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $10 = (($7) + ($8<<2)|0); //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $11 = HEAP32[$10>>2]|0; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $12 = $9 & 1023; //@line 1734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $13 = (($11) + ($12<<2)|0); //@line 1734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $14 = HEAP32[$13>>2]|0; //@line 111 "modules/event/Event.cpp"
  $15 = HEAP32[$14>>2]|0; //@line 111 "modules/event/Event.cpp"
  $16 = ((($15)) + 12|0); //@line 111 "modules/event/Event.cpp"
  $17 = HEAP32[$16>>2]|0; //@line 111 "modules/event/Event.cpp"
  __THREW__ = 0;
  invoke_vi($17|0,($14|0)); //@line 111 "modules/event/Event.cpp"
  $18 = __THREW__; __THREW__ = 0;
  $19 = $18&1;
  if ($19) {
   label = 7;
   break;
  }
  $20 = HEAP32[$2>>2]|0; //@line 2538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $21 = (($20) + -1)|0; //@line 2538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  HEAP32[$2>>2] = $21; //@line 2538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $22 = HEAP32[$6>>2]|0; //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $23 = (($22) + 1)|0; //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  HEAP32[$6>>2] = $23; //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $24 = ($23>>>0)>(2047); //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  if ($24) {
   $27 = HEAP32[$5>>2]|0; //@line 90 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $28 = HEAP32[$27>>2]|0; //@line 2541 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   __ZdlPv($28); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $29 = HEAP32[$5>>2]|0; //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $30 = ((($29)) + 4|0); //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   HEAP32[$5>>2] = $30; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $31 = HEAP32[$6>>2]|0; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $32 = (($31) + -1024)|0; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   HEAP32[$6>>2] = $32; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $$pre1 = HEAP32[$2>>2]|0; //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $26 = $$pre1;$35 = $32;
  } else {
   $26 = $21;$35 = $23;
  }
  $25 = ($26|0)==(0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  if ($25) {
   label = 8;
   break;
  } else {
   $9 = $35;
  }
 }
 if ((label|0) == 7) {
  $33 = ___cxa_find_matching_catch_2()|0;
  $34 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 114 "modules/event/Event.cpp"
  ___resumeException($33|0);
  // unreachable;
 }
 else if ((label|0) == 8) {
  __ZN4love6thread4LockD2Ev($lock); //@line 114 "modules/event/Event.cpp"
  STACKTOP = sp;return; //@line 114 "modules/event/Event.cpp"
 }
}
function __ZN4love5event7MessageC2ERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEERKNS2_6vectorINS_9StrongRefINS_7VariantEEENS6_ISE_EEEE($this,$name,$vargs) {
 $this = $this|0;
 $name = $name|0;
 $vargs = $vargs|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 31 "modules/event/Event.cpp"
 HEAP32[$this>>2] = (23116); //@line 34 "modules/event/Event.cpp"
 $0 = ((($this)) + 8|0); //@line 32 "modules/event/Event.cpp"
 __THREW__ = 0;
 invoke_vii(3351,($0|0),($name|0)); //@line 32 "modules/event/Event.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  $$0 = $7;$$01 = $6;
  ___resumeException($$01|0);
  // unreachable;
 }
 $3 = ((($this)) + 20|0); //@line 33 "modules/event/Event.cpp"
 __THREW__ = 0;
 invoke_vii(3960,($3|0),($vargs|0)); //@line 33 "modules/event/Event.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  return; //@line 35 "modules/event/Event.cpp"
 }
 $8 = ___cxa_find_matching_catch_2()|0;
 $9 = tempRet0;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 35 "modules/event/Event.cpp"
 $$0 = $9;$$01 = $8;
 ___resumeException($$01|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love9StrongRefINS1_7VariantEEENS_9allocatorIS4_EEEC2ERKS7_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$01$i = 0, $$pre$i = 0, $$pre$i$i$i = 0, $$pre2$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $0 = ((($this)) + 4|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($this)) + 8|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($__x)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$__x>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = $5 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ($6|0)==(0); //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($7) {
  return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $8 = ($6>>>0)>(1073741823); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($8) {
  __THREW__ = 0;
  invoke_vi(3528,($this|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if ($10) {
   label = 12;
  } else {
   label = 4;
  }
 } else {
  label = 4;
 }
 do {
  if ((label|0) == 4) {
   __THREW__ = 0;
   $11 = (invoke_ii(3348,($5|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $12 = __THREW__; __THREW__ = 0;
   $13 = $12&1;
   if ($13) {
    label = 12;
   } else {
    HEAP32[$0>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$this>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $14 = (($11) + ($6<<2)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $15 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$15>>2] = $14; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $16 = HEAP32[$__x>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $17 = HEAP32[$2>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $18 = ($16|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($18) {
     return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    $$01$i = $16;$20 = $11;
    while(1) {
     $19 = HEAP32[$$01$i>>2]|0; //@line 139 "./common/Object.h"
     HEAP32[$20>>2] = $19; //@line 99 "./common/Object.h"
     $21 = ($19|0)==(0|0); //@line 101 "./common/Object.h"
     if ($21) {
      $28 = $20;
     } else {
      $22 = HEAP32[$19>>2]|0; //@line 101 "./common/Object.h"
      $23 = ((($22)) + 8|0); //@line 101 "./common/Object.h"
      $24 = HEAP32[$23>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($24|0,($19|0)); //@line 101 "./common/Object.h"
      $25 = __THREW__; __THREW__ = 0;
      $26 = $25&1;
      if ($26) {
       label = 11;
       break;
      }
      $$pre2$i = HEAP32[$0>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $28 = $$pre2$i;
     }
     $27 = ((($28)) + 4|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$0>>2] = $27; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $29 = ((($$01$i)) + 4|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $30 = ($29|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($30) {
      label = 23;
      break;
     } else {
      $$01$i = $29;$20 = $27;
     }
    }
    if ((label|0) == 11) {
     $31 = ___cxa_find_matching_catch_2()|0;
     $32 = tempRet0;
     $lpad$phi$index = $31;$lpad$phi$index2 = $32;
     break;
    }
    else if ((label|0) == 23) {
     return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
   }
  }
 } while(0);
 if ((label|0) == 12) {
  $33 = ___cxa_find_matching_catch_2()|0;
  $34 = tempRet0;
  $lpad$phi$index = $33;$lpad$phi$index2 = $34;
 }
 $35 = HEAP32[$this>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $36 = ($35|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($36) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $37 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $38 = ($37|0)==($35|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($38) {
   $52 = $35;
  } else {
   $40 = $37;
   while(1) {
    $39 = ((($40)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$0>>2] = $39; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $41 = HEAP32[$39>>2]|0; //@line 106 "./common/Object.h"
    $42 = ($41|0)==(0|0); //@line 106 "./common/Object.h"
    if ($42) {
     $44 = $39;
    } else {
     $45 = HEAP32[$41>>2]|0; //@line 106 "./common/Object.h"
     $46 = ((($45)) + 12|0); //@line 106 "./common/Object.h"
     $47 = HEAP32[$46>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($47|0,($41|0)); //@line 106 "./common/Object.h"
     $48 = __THREW__; __THREW__ = 0;
     $49 = $48&1;
     if ($49) {
      label = 19;
      break;
     }
     $$pre$i$i$i = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $44 = $$pre$i$i$i;
    }
    $43 = ($44|0)==($35|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($43) {
     label = 20;
     break;
    } else {
     $40 = $44;
    }
   }
   if ((label|0) == 19) {
    $50 = ___cxa_find_matching_catch_3(0|0)|0;
    $51 = tempRet0;
    ___clang_call_terminate($50); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 20) {
    $$pre$i = HEAP32[$this>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $52 = $$pre$i;
    break;
   }
  }
 } while(0);
 __ZdlPv($52); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZN4love5event7Message5toLuaEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $$pre = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$01 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 43 "modules/event/Event.cpp"
 __ZN4love15luax_pushstringEP9lua_StateRKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE($L,$0); //@line 43 "modules/event/Event.cpp"
 $1 = ((($this)) + 20|0); //@line 45 "modules/event/Event.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($this)) + 24|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$3>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($2|0)==($4|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($5) {
  $6 = $2; //@line 45 "modules/event/Event.cpp"
  $7 = $4; //@line 45 "modules/event/Event.cpp"
  $10 = $6;$9 = $7;
  $8 = (($9) - ($10))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $11 = $8 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $12 = (($11) + 1)|0; //@line 53 "modules/event/Event.cpp"
  return ($12|0); //@line 53 "modules/event/Event.cpp"
 } else {
  $__begin$sroa$0$01 = $2;
 }
 while(1) {
  $13 = HEAP32[$__begin$sroa$0$01>>2]|0; //@line 139 "./common/Object.h"
  $14 = ($13|0)==(0|0); //@line 47 "modules/event/Event.cpp"
  if ($14) {
   _lua_pushnil($L); //@line 50 "modules/event/Event.cpp"
  } else {
   __ZN4love7Variant5toLuaEP9lua_State($13,$L); //@line 48 "modules/event/Event.cpp"
  }
  $15 = ((($__begin$sroa$0$01)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $16 = ($15|0)==($4|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($16) {
   break;
  } else {
   $__begin$sroa$0$01 = $15;
  }
 }
 $$pre = HEAP32[$3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $$pre3 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $$pre3;$9 = $$pre;
 $8 = (($9) - ($10))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $8 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = (($11) + 1)|0; //@line 53 "modules/event/Event.cpp"
 return ($12|0); //@line 53 "modules/event/Event.cpp"
}
function __ZN4love5event7Message7fromLuaEP9lua_Statei($L,$n) {
 $L = $L|0;
 $n = $n|0;
 var $$0$i = 0, $$01 = 0, $$01$i = 0, $$02 = 0, $$1 = 0, $$lcssa79 = 0, $$lcssa80 = 0, $$lcssa87 = 0, $$pre$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i10 = 0, $$pre$i$i$i17 = 0, $$pre$i20 = 0, $$pre$i5 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index7Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i$030 = 0, $lpad$phi$index = 0;
 var $lpad$phi$index2 = 0, $name = 0, $vararg_buffer = 0, $vargs = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $name = sp + 20|0;
 $vargs = sp + 8|0;
 $0 = sp + 4|0;
 __ZN4love16luax_checkstringEP9lua_Statei($name,$L,$n); //@line 58 "modules/event/Event.cpp"
 HEAP32[$vargs>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($vargs)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($vargs)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 $3 = (invoke_ii(3385,($L|0))|0); //@line 61 "modules/event/Event.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 L1: do {
  if ($5) {
   label = 5;
  } else {
   $6 = (($3) - ($n))|0; //@line 61 "modules/event/Event.cpp"
   $7 = (($n) + 1)|0; //@line 62 "modules/event/Event.cpp"
   $8 = ($6|0)>(0); //@line 64 "modules/event/Event.cpp"
   L3: do {
    if ($8) {
     $9 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $i$030 = 0;
     while(1) {
      $14 = (($7) + ($i$030))|0; //@line 66 "modules/event/Event.cpp"
      __THREW__ = 0;
      $15 = (invoke_iii(3358,($L|0),($14|0))|0); //@line 66 "modules/event/Event.cpp"
      $16 = __THREW__; __THREW__ = 0;
      $17 = $16&1;
      if ($17) {
       label = 4;
       break;
      }
      $18 = ($15|0)<(1); //@line 66 "modules/event/Event.cpp"
      if ($18) {
       label = 35;
       break L3;
      }
      __THREW__ = 0;
      $19 = (invoke_iiii(3506,($L|0),($14|0),1)|0); //@line 106 "./common/Object.h"
      $20 = __THREW__; __THREW__ = 0;
      $21 = $20&1;
      if ($21) {
       label = 4;
       break;
      }
      HEAP32[$0>>2] = $19; //@line 93 "./common/Object.h"
      $22 = ($19|0)==(0|0); //@line 95 "./common/Object.h"
      if (!($22)) {
       $23 = HEAP32[$19>>2]|0; //@line 95 "./common/Object.h"
       $24 = ((($23)) + 8|0); //@line 95 "./common/Object.h"
       $25 = HEAP32[$24>>2]|0; //@line 95 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($25|0,($19|0)); //@line 95 "./common/Object.h"
       $26 = __THREW__; __THREW__ = 0;
       $27 = $26&1;
       if ($27) {
        label = 4;
        break;
       }
      }
      $28 = HEAP32[$1>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $29 = HEAP32[$9>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $30 = ($28>>>0)<($29>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($30) {
       HEAP32[$28>>2] = $19; //@line 99 "./common/Object.h"
       if ($22) {
        $39 = $28;
       } else {
        $31 = HEAP32[$19>>2]|0; //@line 101 "./common/Object.h"
        $32 = ((($31)) + 8|0); //@line 101 "./common/Object.h"
        $33 = HEAP32[$32>>2]|0; //@line 101 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($33|0,($19|0)); //@line 101 "./common/Object.h"
        $34 = __THREW__; __THREW__ = 0;
        $35 = $34&1;
        if ($35) {
         $$lcssa80 = $19;
         label = 15;
         break;
        }
        $$pre$i5 = HEAP32[$1>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $39 = $$pre$i5;
       }
       $38 = ((($39)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$1>>2] = $38; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      } else {
       __THREW__ = 0;
       invoke_vii(3507,($vargs|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $40 = __THREW__; __THREW__ = 0;
       $41 = $40&1;
       if ($41) {
        label = 30;
        break;
       }
      }
      if (!($22)) {
       $42 = HEAP32[$19>>2]|0; //@line 106 "./common/Object.h"
       $43 = ((($42)) + 12|0); //@line 106 "./common/Object.h"
       $44 = HEAP32[$43>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($44|0,($19|0)); //@line 106 "./common/Object.h"
       $45 = __THREW__; __THREW__ = 0;
       $46 = $45&1;
       if ($46) {
        label = 21;
        break;
       }
      }
      $49 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $50 = ((($49)) + -4|0); //@line 139 "./common/Object.h"
      $51 = HEAP32[$50>>2]|0; //@line 139 "./common/Object.h"
      $52 = ($51|0)==(0|0); //@line 71 "modules/event/Event.cpp"
      if ($52) {
       $$lcssa79 = $14;$$lcssa87 = $49;
       label = 23;
       break;
      }
      $80 = HEAP32[$51>>2]|0; //@line 78 "modules/event/Event.cpp"
      $81 = ((($80)) + 12|0); //@line 78 "modules/event/Event.cpp"
      $82 = HEAP32[$81>>2]|0; //@line 78 "modules/event/Event.cpp"
      __THREW__ = 0;
      invoke_vi($82|0,($51|0)); //@line 78 "modules/event/Event.cpp"
      $83 = __THREW__; __THREW__ = 0;
      $84 = $83&1;
      if ($84) {
       label = 4;
       break;
      }
      $85 = (($i$030) + 1)|0; //@line 64 "modules/event/Event.cpp"
      $86 = ($85|0)<($6|0); //@line 64 "modules/event/Event.cpp"
      if ($86) {
       $i$030 = $85;
      } else {
       label = 35;
       break L3;
      }
     }
     if ((label|0) == 4) {
      $10 = ___cxa_find_matching_catch_2()|0;
      $11 = tempRet0;
      $lpad$phi$index = $10;$lpad$phi$index2 = $11;
      label = 6;
      break L1;
     }
     else if ((label|0) == 15) {
      $36 = ___cxa_find_matching_catch_2()|0;
      $37 = tempRet0;
      $140 = $37;$141 = $36;$73 = $$lcssa80;
     }
     else if ((label|0) == 21) {
      $47 = ___cxa_find_matching_catch_3(0|0)|0;
      $48 = tempRet0;
      ___clang_call_terminate($47); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     else if ((label|0) == 23) {
      $53 = HEAP32[$vargs>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $54 = ($$lcssa87|0)==($53|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      L29: do {
       if (!($54)) {
        $56 = $$lcssa87;
        while(1) {
         $55 = ((($56)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$1>>2] = $55; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $57 = HEAP32[$55>>2]|0; //@line 106 "./common/Object.h"
         $58 = ($57|0)==(0|0); //@line 106 "./common/Object.h"
         if ($58) {
          $60 = $55;
         } else {
          $61 = HEAP32[$57>>2]|0; //@line 106 "./common/Object.h"
          $62 = ((($61)) + 12|0); //@line 106 "./common/Object.h"
          $63 = HEAP32[$62>>2]|0; //@line 106 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($63|0,($57|0)); //@line 106 "./common/Object.h"
          $64 = __THREW__; __THREW__ = 0;
          $65 = $64&1;
          if ($65) {
           break;
          }
          $$pre$i$i$i10 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $60 = $$pre$i$i$i10;
         }
         $59 = ($60|0)==($53|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($59) {
          break L29;
         } else {
          $56 = $60;
         }
        }
        $66 = ___cxa_find_matching_catch_3(0|0)|0;
        $67 = tempRet0;
        ___clang_call_terminate($66); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
      } while(0);
      __THREW__ = 0;
      HEAP32[$vararg_buffer>>2] = $$lcssa79; //@line 74 "modules/event/Event.cpp"
      (invoke_iiii(3389,($L|0),(365959|0),($vararg_buffer|0))|0); //@line 74 "modules/event/Event.cpp"
      $68 = __THREW__; __THREW__ = 0;
      $69 = $68&1;
      if ($69) {
       label = 5;
       break L1;
      } else {
       $$1 = 0;
       break;
      }
     }
     else if ((label|0) == 30) {
      $70 = ___cxa_find_matching_catch_2()|0;
      $71 = tempRet0;
      if ($22) {
       $$01 = $71;$$02 = $70;
       break L1;
      } else {
       $140 = $71;$141 = $70;$73 = $19;
      }
     }
     $72 = HEAP32[$73>>2]|0; //@line 106 "./common/Object.h"
     $74 = ((($72)) + 12|0); //@line 106 "./common/Object.h"
     $75 = HEAP32[$74>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($75|0,($73|0)); //@line 106 "./common/Object.h"
     $76 = __THREW__; __THREW__ = 0;
     $77 = $76&1;
     if (!($77)) {
      $$01 = $140;$$02 = $141;
      break L1;
     }
     $78 = ___cxa_find_matching_catch_3(0|0)|0;
     $79 = tempRet0;
     ___clang_call_terminate($78); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    } else {
     label = 35;
    }
   } while(0);
   do {
    if ((label|0) == 35) {
     __THREW__ = 0;
     $87 = (invoke_ii(3348,32)|0); //@line 81 "modules/event/Event.cpp"
     $88 = __THREW__; __THREW__ = 0;
     $89 = $88&1;
     if ($89) {
      label = 5;
      break L1;
     }
     __THREW__ = 0;
     invoke_vi(3431,($87|0)); //@line 31 "modules/event/Event.cpp"
     $90 = __THREW__; __THREW__ = 0;
     $91 = $90&1;
     if ($91) {
      $102 = ___cxa_find_matching_catch_2()|0;
      $103 = tempRet0;
      $eh$lpad$body$index7Z2D = $103;$eh$lpad$body$indexZ2D = $102;
     } else {
      HEAP32[$87>>2] = (23116); //@line 34 "modules/event/Event.cpp"
      $92 = ((($87)) + 8|0); //@line 32 "modules/event/Event.cpp"
      __THREW__ = 0;
      invoke_vii(3351,($92|0),($name|0)); //@line 32 "modules/event/Event.cpp"
      $93 = __THREW__; __THREW__ = 0;
      $94 = $93&1;
      if ($94) {
       $98 = ___cxa_find_matching_catch_2()|0;
       $99 = tempRet0;
       $$0$i = $99;$$01$i = $98;
      } else {
       $95 = ((($87)) + 20|0); //@line 33 "modules/event/Event.cpp"
       __THREW__ = 0;
       invoke_vii(3960,($95|0),($vargs|0)); //@line 33 "modules/event/Event.cpp"
       $96 = __THREW__; __THREW__ = 0;
       $97 = $96&1;
       if (!($97)) {
        $$1 = $87;
        break;
       }
       $100 = ___cxa_find_matching_catch_2()|0;
       $101 = tempRet0;
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($92); //@line 35 "modules/event/Event.cpp"
       $$0$i = $101;$$01$i = $100;
      }
      $eh$lpad$body$index7Z2D = $$0$i;$eh$lpad$body$indexZ2D = $$01$i;
     }
     __ZdlPv($87); //@line 81 "modules/event/Event.cpp"
     $$01 = $eh$lpad$body$index7Z2D;$$02 = $eh$lpad$body$indexZ2D;
     break L1;
    }
   } while(0);
   $104 = HEAP32[$vargs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $105 = ($104|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($105) {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($name); //@line 82 "modules/event/Event.cpp"
    STACKTOP = sp;return ($$1|0); //@line 82 "modules/event/Event.cpp"
   }
   $106 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $107 = ($106|0)==($104|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   do {
    if ($107) {
     $121 = $104;
    } else {
     $109 = $106;
     while(1) {
      $108 = ((($109)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$1>>2] = $108; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $110 = HEAP32[$108>>2]|0; //@line 106 "./common/Object.h"
      $111 = ($110|0)==(0|0); //@line 106 "./common/Object.h"
      if ($111) {
       $113 = $108;
      } else {
       $114 = HEAP32[$110>>2]|0; //@line 106 "./common/Object.h"
       $115 = ((($114)) + 12|0); //@line 106 "./common/Object.h"
       $116 = HEAP32[$115>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($116|0,($110|0)); //@line 106 "./common/Object.h"
       $117 = __THREW__; __THREW__ = 0;
       $118 = $117&1;
       if ($118) {
        label = 50;
        break;
       }
       $$pre$i$i$i17 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $113 = $$pre$i$i$i17;
      }
      $112 = ($113|0)==($104|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($112) {
       label = 51;
       break;
      } else {
       $109 = $113;
      }
     }
     if ((label|0) == 50) {
      $119 = ___cxa_find_matching_catch_3(0|0)|0;
      $120 = tempRet0;
      ___clang_call_terminate($119); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     else if ((label|0) == 51) {
      $$pre$i20 = HEAP32[$vargs>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $121 = $$pre$i20;
      break;
     }
    }
   } while(0);
   __ZdlPv($121); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($name); //@line 82 "modules/event/Event.cpp"
   STACKTOP = sp;return ($$1|0); //@line 82 "modules/event/Event.cpp"
  }
 } while(0);
 if ((label|0) == 5) {
  $12 = ___cxa_find_matching_catch_2()|0;
  $13 = tempRet0;
  $lpad$phi$index = $12;$lpad$phi$index2 = $13;
  label = 6;
 }
 if ((label|0) == 6) {
  $$01 = $lpad$phi$index2;$$02 = $lpad$phi$index;
 }
 $122 = HEAP32[$vargs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $123 = ($122|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($123) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($name); //@line 82 "modules/event/Event.cpp"
  ___resumeException($$02|0);
  // unreachable;
 }
 $124 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $125 = ($124|0)==($122|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($125) {
   $139 = $122;
  } else {
   $127 = $124;
   while(1) {
    $126 = ((($127)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$1>>2] = $126; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $128 = HEAP32[$126>>2]|0; //@line 106 "./common/Object.h"
    $129 = ($128|0)==(0|0); //@line 106 "./common/Object.h"
    if ($129) {
     $131 = $126;
    } else {
     $132 = HEAP32[$128>>2]|0; //@line 106 "./common/Object.h"
     $133 = ((($132)) + 12|0); //@line 106 "./common/Object.h"
     $134 = HEAP32[$133>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($134|0,($128|0)); //@line 106 "./common/Object.h"
     $135 = __THREW__; __THREW__ = 0;
     $136 = $135&1;
     if ($136) {
      label = 60;
      break;
     }
     $$pre$i$i$i = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $131 = $$pre$i$i$i;
    }
    $130 = ($131|0)==($122|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($130) {
     label = 61;
     break;
    } else {
     $127 = $131;
    }
   }
   if ((label|0) == 60) {
    $137 = ___cxa_find_matching_catch_3(0|0)|0;
    $138 = tempRet0;
    ___clang_call_terminate($137); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 61) {
    $$pre$i = HEAP32[$vargs>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $139 = $$pre$i;
    break;
   }
  }
 } while(0);
 __ZdlPv($139); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($name); //@line 82 "modules/event/Event.cpp"
 ___resumeException($$02|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love5event5Event4pushEPNS0_7MessageE($this,$msg) {
 $this = $this|0;
 $msg = $msg|0;
 var $$pre$i$i = 0, $$pre1$i$i = 0, $$pre2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $lock = 0, $sum$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = $msg;
 $1 = ((($this)) + 8|0); //@line 90 "modules/event/Event.cpp"
 $2 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($1)|0); //@line 90 "modules/event/Event.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$2); //@line 90 "modules/event/Event.cpp"
 $3 = HEAP32[$msg>>2]|0; //@line 91 "modules/event/Event.cpp"
 $4 = ((($3)) + 8|0); //@line 91 "modules/event/Event.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 91 "modules/event/Event.cpp"
 __THREW__ = 0;
 invoke_vi($5|0,($msg|0)); //@line 91 "modules/event/Event.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $36 = ___cxa_find_matching_catch_2()|0;
  $37 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 93 "modules/event/Event.cpp"
  ___resumeException($36|0);
  // unreachable;
 }
 $8 = ((($this)) + 20|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $9 = HEAP32[$8>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $10 = ((($this)) + 16|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $11 = HEAP32[$10>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $12 = (($9) - ($11))|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $13 = ($9|0)==($11|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $14 = $12 << 8; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $15 = (($14) + -1)|0; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $16 = $13 ? 0 : $15; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $17 = ((($this)) + 28|0); //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $18 = HEAP32[$17>>2]|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $19 = ((($this)) + 32|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $20 = HEAP32[$19>>2]|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $sum$i$i$i = (($20) + ($18))|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $21 = ($16|0)==($sum$i$i$i|0); //@line 1769 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $22 = $11; //@line 1769 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 do {
  if ($21) {
   $23 = ((($this)) + 12|0); //@line 286 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/queue"
   __THREW__ = 0;
   invoke_vi(3961,($23|0)); //@line 1770 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $24 = __THREW__; __THREW__ = 0;
   $25 = $24&1;
   if ($25) {
    $36 = ___cxa_find_matching_catch_2()|0;
    $37 = tempRet0;
    __ZN4love6thread4LockD2Ev($lock); //@line 93 "modules/event/Event.cpp"
    ___resumeException($36|0);
    // unreachable;
   } else {
    $$pre$i$i = HEAP32[$19>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $$pre1$i$i = HEAP32[$17>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $$pre2$i$i = HEAP32[$10>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $27 = $$pre$i$i;$28 = $$pre1$i$i;$31 = $$pre2$i$i;
    break;
   }
  } else {
   $27 = $20;$28 = $18;$31 = $22;
  }
 } while(0);
 $26 = (($27) + ($28))|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $29 = $26 >>> 10; //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $30 = (($31) + ($29<<2)|0); //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $32 = HEAP32[$30>>2]|0; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $33 = $26 & 1023; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $34 = (($32) + ($33<<2)|0); //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 HEAP32[$34>>2] = $0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (($27) + 1)|0; //@line 1773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 HEAP32[$19>>2] = $35; //@line 1773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 __ZN4love6thread4LockD2Ev($lock); //@line 93 "modules/event/Event.cpp"
 STACKTOP = sp;return; //@line 93 "modules/event/Event.cpp"
}
function __ZNSt3__15dequeIPN4love5event7MessageENS_9allocatorIS4_EEE19__add_back_capacityEv($this) {
 $this = $this|0;
 var $$0161 = 0, $$0162 = 0, $$063 = 0, $$064 = 0, $$in$i = 0, $$in$i30 = 0, $$in$i4 = 0, $$lcssa = 0, $$lcssa77 = 0, $$pre = 0, $$pre$i32 = 0, $$pre$i6 = 0, $$pre4$i10 = 0, $$pre4$i37 = 0, $$pre59 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $__buf$sroa$0$0 = 0, $__buf$sroa$0$1$lcssa = 0, $__buf$sroa$0$147 = 0, $__buf$sroa$0$147$lcssa = 0, $__buf$sroa$0$2 = 0, $__buf$sroa$0$360 = 0, $__buf$sroa$13$0 = 0, $__buf$sroa$13$0$in = 0, $__buf$sroa$13$1$lcssa = 0, $__buf$sroa$13$148 = 0, $__buf$sroa$13$2 = 0, $__buf$sroa$28$1$lcssa = 0, $__buf$sroa$28$149 = 0, $__buf$sroa$28$2 = 0, $__buf$sroa$46$0 = 0, $__buf$sroa$46$1$lcssa = 0, $__buf$sroa$46$150 = 0, $__buf$sroa$46$2 = 0, $__i$051 = 0, $__t$sroa$10$0$i = 0;
 var $__t$sroa$10$0$i29 = 0, $__t$sroa$10$0$i3 = 0, $__t$sroa$10$1$i = 0, $__t$sroa$10$1$i34 = 0, $__t$sroa$10$1$i8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 4|0;
 $1 = sp;
 $2 = ((($this)) + 16|0); //@line 1374 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $3 = HEAP32[$2>>2]|0; //@line 1374 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $4 = ($3>>>0)>(1023); //@line 2390 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($4) {
  $5 = (($3) + -1024)|0; //@line 2392 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  HEAP32[$2>>2] = $5; //@line 2392 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $6 = ((($this)) + 4|0); //@line 90 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $7 = HEAP32[$6>>2]|0; //@line 90 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $8 = HEAP32[$7>>2]|0; //@line 2393 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $9 = ((($7)) + 4|0); //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  HEAP32[$6>>2] = $9; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $10 = ((($this)) + 8|0); //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $11 = HEAP32[$10>>2]|0; //@line 562 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $12 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $13 = HEAP32[$12>>2]|0; //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $14 = ($11|0)==($13|0); //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $15 = $13; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $16 = $11; //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  do {
   if ($14) {
    $17 = HEAP32[$this>>2]|0; //@line 550 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $18 = ($9>>>0)>($17>>>0); //@line 550 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $19 = $17; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($18) {
     $20 = $9; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $21 = (($20) - ($19))|0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $22 = $21 >> 2; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $23 = (($22) + 1)|0; //@line 553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $24 = (($23|0) / -2)&-1; //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $25 = (($9) + ($24<<2)|0); //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $26 = (($16) - ($20))|0; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $27 = $26 >> 2; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     _memmove(($25|0),($9|0),($26|0))|0; //@line 1893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $28 = (($25) + ($27<<2)|0); //@line 1894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     HEAP32[$10>>2] = $28; //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $29 = HEAP32[$6>>2]|0; //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $30 = (($29) + ($24<<2)|0); //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     HEAP32[$6>>2] = $30; //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $54 = $28;
     break;
    }
    $31 = (($15) - ($19))|0; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $32 = $31 >> 1; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $33 = ($32|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $34 = $33 ? 1 : $32; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $35 = $34 >>> 2; //@line 560 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $36 = $34 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $37 = (__Znwj($36)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $38 = $37; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $39 = (($37) + ($35<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $40 = $39; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $41 = (($37) + ($34<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $42 = $41; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $43 = ($9|0)==($11|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($43) {
     $52 = $17;$__t$sroa$10$1$i8 = $40;
    } else {
     $$in$i4 = $39;$45 = $9;$__t$sroa$10$0$i3 = $40;
     while(1) {
      $44 = HEAP32[$45>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$$in$i4>>2] = $44; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $46 = $__t$sroa$10$0$i3; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $47 = ((($46)) + 4|0); //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $48 = $47; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $49 = ((($45)) + 4|0); //@line 954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $50 = ($49|0)==($11|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($50) {
       $$lcssa = $48;
       break;
      } else {
       $$in$i4 = $47;$45 = $49;$__t$sroa$10$0$i3 = $48;
      }
     }
     $$pre$i6 = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $52 = $$pre$i6;$__t$sroa$10$1$i8 = $$lcssa;
    }
    HEAP32[$this>>2] = $38; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$6>>2] = $40; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$10>>2] = $__t$sroa$10$1$i8; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$12>>2] = $42; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $51 = ($52|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $53 = $__t$sroa$10$1$i8; //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($51) {
     $54 = $53;
    } else {
     __ZdlPv($52); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $$pre4$i10 = HEAP32[$10>>2]|0; //@line 569 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $54 = $$pre4$i10;
    }
   } else {
    $54 = $11;
   }
  } while(0);
  HEAP32[$54>>2] = $8; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $55 = HEAP32[$10>>2]|0; //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $56 = ((($55)) + 4|0); //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  HEAP32[$10>>2] = $56; //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  STACKTOP = sp;return; //@line 2441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 }
 $57 = ((($this)) + 8|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $58 = HEAP32[$57>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $59 = ((($this)) + 4|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $60 = HEAP32[$59>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $61 = (($58) - ($60))|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $62 = $61 >> 2; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $63 = ((($this)) + 12|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $64 = HEAP32[$63>>2]|0; //@line 86 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $65 = HEAP32[$this>>2]|0; //@line 86 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $66 = (($64) - ($65))|0; //@line 86 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $67 = $66 >> 2; //@line 86 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $68 = ($62>>>0)<($67>>>0); //@line 2398 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $69 = $58; //@line 2398 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $70 = $60; //@line 2398 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($68) {
  $71 = ($64|0)==($58|0); //@line 2402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $72 = (__Znwj(4096)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  if (!($71)) {
   HEAP32[$0>>2] = $72; //@line 2403 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   __ZNSt3__114__split_bufferIPPN4love5event7MessageENS_9allocatorIS5_EEE9push_backEOS5_($this,$0); //@line 2403 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   STACKTOP = sp;return; //@line 2441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  }
  HEAP32[$1>>2] = $72; //@line 2406 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  __ZNSt3__114__split_bufferIPPN4love5event7MessageENS_9allocatorIS5_EEE10push_frontEOS5_($this,$1); //@line 2406 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $73 = HEAP32[$59>>2]|0; //@line 90 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $74 = HEAP32[$73>>2]|0; //@line 2408 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $75 = ((($73)) + 4|0); //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  HEAP32[$59>>2] = $75; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $76 = HEAP32[$57>>2]|0; //@line 562 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $77 = HEAP32[$63>>2]|0; //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $78 = ($76|0)==($77|0); //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $79 = $77; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $80 = $76; //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  do {
   if ($78) {
    $81 = HEAP32[$this>>2]|0; //@line 550 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $82 = ($75>>>0)>($81>>>0); //@line 550 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $83 = $81; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($82) {
     $84 = $75; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $85 = (($84) - ($83))|0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $86 = $85 >> 2; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $87 = (($86) + 1)|0; //@line 553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $88 = (($87|0) / -2)&-1; //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $89 = (($75) + ($88<<2)|0); //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $90 = (($80) - ($84))|0; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $91 = $90 >> 2; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     _memmove(($89|0),($75|0),($90|0))|0; //@line 1893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $92 = (($89) + ($91<<2)|0); //@line 1894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     HEAP32[$57>>2] = $92; //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $93 = HEAP32[$59>>2]|0; //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $94 = (($93) + ($88<<2)|0); //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     HEAP32[$59>>2] = $94; //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $118 = $92;
     break;
    }
    $95 = (($79) - ($83))|0; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $96 = $95 >> 1; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $97 = ($96|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $98 = $97 ? 1 : $96; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $99 = $98 >>> 2; //@line 560 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $100 = $98 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $101 = (__Znwj($100)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $102 = $101; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $103 = (($101) + ($99<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $104 = $103; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $105 = (($101) + ($98<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $106 = $105; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $107 = ($75|0)==($76|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($107) {
     $116 = $81;$__t$sroa$10$1$i34 = $104;
    } else {
     $$in$i30 = $103;$109 = $75;$__t$sroa$10$0$i29 = $104;
     while(1) {
      $108 = HEAP32[$109>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$$in$i30>>2] = $108; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $110 = $__t$sroa$10$0$i29; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $111 = ((($110)) + 4|0); //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $112 = $111; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $113 = ((($109)) + 4|0); //@line 954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $114 = ($113|0)==($76|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($114) {
       $$lcssa77 = $112;
       break;
      } else {
       $$in$i30 = $111;$109 = $113;$__t$sroa$10$0$i29 = $112;
      }
     }
     $$pre$i32 = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $116 = $$pre$i32;$__t$sroa$10$1$i34 = $$lcssa77;
    }
    HEAP32[$this>>2] = $102; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$59>>2] = $104; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$57>>2] = $__t$sroa$10$1$i34; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$63>>2] = $106; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $115 = ($116|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $117 = $__t$sroa$10$1$i34; //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($115) {
     $118 = $117;
    } else {
     __ZdlPv($116); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $$pre4$i37 = HEAP32[$57>>2]|0; //@line 569 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $118 = $$pre4$i37;
    }
   } else {
    $118 = $76;
   }
  } while(0);
  HEAP32[$118>>2] = $74; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $119 = HEAP32[$57>>2]|0; //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $120 = ((($119)) + 4|0); //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  HEAP32[$57>>2] = $120; //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  STACKTOP = sp;return; //@line 2441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 }
 $121 = $66 >> 1; //@line 2417 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $122 = ($121|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $123 = $122 ? 1 : $121; //@line 2417 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $124 = $123 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $125 = (__Znwj($124)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $126 = $125; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $127 = (($125) + ($62<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $128 = (($125) + ($123<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $129 = $128; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 __THREW__ = 0;
 $130 = (invoke_ii(3348,4096)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $131 = __THREW__; __THREW__ = 0;
 $132 = $131&1;
 L37: do {
  if ($132) {
   label = 29;
  } else {
   $133 = $130; //@line 2424 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $134 = ($62|0)==($123|0); //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   do {
    if ($134) {
     $135 = ($61|0)>(0); //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     if ($135) {
      $136 = (($62) + 1)|0; //@line 584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $137 = (($136|0) / -2)&-1; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $138 = (($127) + ($137<<2)|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $163 = $69;$164 = $70;$__buf$sroa$0$0 = $126;$__buf$sroa$13$0$in = $138;$__buf$sroa$46$0 = $129;
      break;
     }
     $139 = $61 >> 1; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $140 = ($139|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $141 = $140 ? 1 : $139; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $142 = $141 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     __THREW__ = 0;
     $143 = (invoke_ii(3348,($142|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $144 = __THREW__; __THREW__ = 0;
     $145 = $144&1;
     if ($145) {
      label = 29;
      break L37;
     }
     $146 = $141 >>> 2; //@line 591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $147 = (($143) + ($146<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $148 = (($143) + ($141<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $149 = $148; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $150 = $143; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     __ZdlPv($125); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $$pre = HEAP32[$57>>2]|0; //@line 78 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $$pre59 = HEAP32[$59>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $163 = $$pre;$164 = $$pre59;$__buf$sroa$0$0 = $150;$__buf$sroa$13$0$in = $147;$__buf$sroa$46$0 = $149;
    } else {
     $163 = $69;$164 = $70;$__buf$sroa$0$0 = $126;$__buf$sroa$13$0$in = $127;$__buf$sroa$46$0 = $129;
    }
   } while(0);
   $__buf$sroa$13$0 = $__buf$sroa$13$0$in; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   HEAP32[$__buf$sroa$13$0$in>>2] = $133; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $160 = ((($__buf$sroa$13$0$in)) + 4|0); //@line 602 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $161 = $160; //@line 602 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $162 = ($163|0)==($164|0); //@line 2434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   L46: do {
    if ($162) {
     $__buf$sroa$0$1$lcssa = $__buf$sroa$0$0;$__buf$sroa$13$1$lcssa = $__buf$sroa$13$0;$__buf$sroa$28$1$lcssa = $161;$__buf$sroa$46$1$lcssa = $__buf$sroa$46$0;
    } else {
     $__buf$sroa$0$147 = $__buf$sroa$0$0;$__buf$sroa$13$148 = $__buf$sroa$13$0;$__buf$sroa$28$149 = $161;$__buf$sroa$46$150 = $__buf$sroa$46$0;$__i$051 = $163;
     L47: while(1) {
      $167 = ((($__i$051)) + -4|0); //@line 2435 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
      $168 = $__buf$sroa$13$148; //@line 497 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $169 = $__buf$sroa$0$147; //@line 484 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $170 = ($168|0)==($169|0); //@line 484 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      do {
       if ($170) {
        $171 = $__buf$sroa$28$149; //@line 498 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $172 = $__buf$sroa$46$150; //@line 486 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $173 = ($171>>>0)<($172>>>0); //@line 486 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        if ($173) {
         $174 = (($__buf$sroa$46$150) - ($__buf$sroa$28$149))|0; //@line 488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $175 = $174 >> 2; //@line 488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $176 = (($175) + 1)|0; //@line 489 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $177 = (($176|0) / 2)&-1; //@line 489 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $178 = (($171) + ($177<<2)|0); //@line 490 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $179 = (($__buf$sroa$28$149) - ($__buf$sroa$13$148))|0; //@line 1927 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $180 = $179 >> 2; //@line 1927 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $181 = (0 - ($180))|0; //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $182 = (($178) + ($181<<2)|0); //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $183 = $__buf$sroa$13$148; //@line 1929 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         _memmove(($182|0),($183|0),($179|0))|0; //@line 1929 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $184 = $182; //@line 490 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $185 = $178; //@line 491 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $212 = $182;$__buf$sroa$0$2 = $__buf$sroa$0$147;$__buf$sroa$13$2 = $184;$__buf$sroa$28$2 = $185;$__buf$sroa$46$2 = $__buf$sroa$46$150;
         break;
        }
        $186 = (($__buf$sroa$46$150) - ($__buf$sroa$0$147))|0; //@line 495 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $187 = $186 >> 1; //@line 495 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $188 = ($187|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $189 = $188 ? 1 : $187; //@line 495 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $190 = $189 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        __THREW__ = 0;
        $191 = (invoke_ii(3348,($190|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        $192 = __THREW__; __THREW__ = 0;
        $193 = $192&1;
        if ($193) {
         $__buf$sroa$0$147$lcssa = $__buf$sroa$0$147;
         break L47;
        }
        $194 = (($189) + 3)|0; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $195 = $194 >>> 2; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $196 = $191; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $197 = (($191) + ($195<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $198 = $197; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $199 = (($191) + ($189<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $200 = $199; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $201 = ($168|0)==($171|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        if ($201) {
         $__t$sroa$10$1$i = $198;
        } else {
         $$in$i = $197;$203 = $168;$__t$sroa$10$0$i = $198;
         while(1) {
          $202 = HEAP32[$203>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          HEAP32[$$in$i>>2] = $202; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $204 = $__t$sroa$10$0$i; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
          $205 = ((($204)) + 4|0); //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
          $206 = $205; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
          $207 = ((($203)) + 4|0); //@line 954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
          $208 = ($207|0)==($171|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
          if ($208) {
           $__t$sroa$10$1$i = $206;
           break;
          } else {
           $$in$i = $205;$203 = $207;$__t$sroa$10$0$i = $206;
          }
         }
        }
        $209 = ($__buf$sroa$13$148|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        if ($209) {
         $212 = $197;$__buf$sroa$0$2 = $196;$__buf$sroa$13$2 = $198;$__buf$sroa$28$2 = $__t$sroa$10$1$i;$__buf$sroa$46$2 = $200;
        } else {
         $210 = $__buf$sroa$13$148; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         __ZdlPv($210); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
         $212 = $197;$__buf$sroa$0$2 = $196;$__buf$sroa$13$2 = $198;$__buf$sroa$28$2 = $__t$sroa$10$1$i;$__buf$sroa$46$2 = $200;
        }
       } else {
        $212 = $168;$__buf$sroa$0$2 = $__buf$sroa$0$147;$__buf$sroa$13$2 = $__buf$sroa$13$148;$__buf$sroa$28$2 = $__buf$sroa$28$149;$__buf$sroa$46$2 = $__buf$sroa$46$150;
       }
      } while(0);
      $211 = ((($212)) + -4|0); //@line 505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $213 = HEAP32[$167>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$211>>2] = $213; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $214 = $__buf$sroa$13$2; //@line 506 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $215 = ((($214)) + -4|0); //@line 506 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $216 = $215; //@line 506 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $217 = HEAP32[$59>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $218 = ($167|0)==($217|0); //@line 2434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
      if ($218) {
       $__buf$sroa$0$1$lcssa = $__buf$sroa$0$2;$__buf$sroa$13$1$lcssa = $216;$__buf$sroa$28$1$lcssa = $__buf$sroa$28$2;$__buf$sroa$46$1$lcssa = $__buf$sroa$46$2;
       break L46;
      } else {
       $__buf$sroa$0$147 = $__buf$sroa$0$2;$__buf$sroa$13$148 = $216;$__buf$sroa$28$149 = $__buf$sroa$28$2;$__buf$sroa$46$150 = $__buf$sroa$46$2;$__i$051 = $167;
      }
     }
     $219 = ___cxa_find_matching_catch_2()|0;
     $220 = tempRet0;
     $221 = ($__buf$sroa$0$147$lcssa|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     if ($221) {
      $$0162 = $219;$$064 = $220;
     } else {
      $$0161 = $219;$$063 = $220;$__buf$sroa$0$360 = $__buf$sroa$0$147$lcssa;
      break L37;
     }
     ___resumeException($$0162|0);
     // unreachable;
    }
   } while(0);
   $165 = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$this>>2] = $__buf$sroa$0$1$lcssa; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$59>>2] = $__buf$sroa$13$1$lcssa; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$57>>2] = $__buf$sroa$28$1$lcssa; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$63>>2] = $__buf$sroa$46$1$lcssa; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $166 = ($165|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($166) {
    STACKTOP = sp;return; //@line 2441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   }
   __ZdlPv($165); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   STACKTOP = sp;return; //@line 2441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  }
 } while(0);
 if ((label|0) == 29) {
  $151 = ___cxa_find_matching_catch_3(0|0)|0;
  $152 = tempRet0;
  (___cxa_begin_catch(($151|0))|0); //@line 2426 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $153 = ((($127)) + -4|0); //@line 92 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $154 = HEAP32[$153>>2]|0; //@line 2429 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  __ZdlPv($154); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  __THREW__ = 0;
  invoke_v(3656); //@line 2430 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $155 = __THREW__; __THREW__ = 0;
  $156 = ___cxa_find_matching_catch_2()|0;
  $157 = tempRet0;
  __THREW__ = 0;
  invoke_v(3378); //@line 2431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $158 = __THREW__; __THREW__ = 0;
  $159 = $158&1;
  if ($159) {
   $223 = ___cxa_find_matching_catch_3(0|0)|0;
   $224 = tempRet0;
   ___clang_call_terminate($223); //@line 2431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   // unreachable; //@line 2431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  } else {
   $$0161 = $156;$$063 = $157;$__buf$sroa$0$360 = $126;
  }
 }
 $222 = $__buf$sroa$0$360; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($222); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $$0162 = $$0161;$$064 = $$063;
 ___resumeException($$0162|0);
 // unreachable;
}
function __ZNSt3__114__split_bufferIPPN4love5event7MessageENS_9allocatorIS5_EEE9push_backEOS5_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__t$sroa$10$0 = 0, $__t$sroa$10$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $1 = HEAP32[$0>>2]|0; //@line 593 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $2 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $4 = ($1|0)==($3|0); //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $5 = $3; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $6 = $1; //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 do {
  if ($4) {
   $7 = ((($this)) + 4|0); //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $8 = HEAP32[$7>>2]|0; //@line 592 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $9 = HEAP32[$this>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $10 = ($8>>>0)>($9>>>0); //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $11 = $9; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($10) {
    $12 = $8; //@line 583 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $13 = (($12) - ($11))|0; //@line 583 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $14 = $13 >> 2; //@line 583 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $15 = (($14) + 1)|0; //@line 584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $16 = (($15|0) / -2)&-1; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $17 = (($8) + ($16<<2)|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $18 = (($6) - ($12))|0; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $19 = $18 >> 2; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    _memmove(($17|0),($8|0),($18|0))|0; //@line 1893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $20 = (($17) + ($19<<2)|0); //@line 1894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    HEAP32[$0>>2] = $20; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $21 = HEAP32[$7>>2]|0; //@line 586 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $22 = (($21) + ($16<<2)|0); //@line 586 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    HEAP32[$7>>2] = $22; //@line 586 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $47 = $20;
    break;
   }
   $23 = (($5) - ($11))|0; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $24 = $23 >> 1; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $25 = ($24|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $26 = $25 ? 1 : $24; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $27 = $26 >>> 2; //@line 591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $28 = $26 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $29 = (__Znwj($28)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $30 = $29; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $31 = (($29) + ($27<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $32 = $31; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $33 = (($29) + ($26<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $34 = $33; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $35 = ($8|0)==($1|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($35) {
    $44 = $9;$__t$sroa$10$1 = $32;
   } else {
    $$in = $31;$37 = $8;$__t$sroa$10$0 = $32;
    while(1) {
     $36 = HEAP32[$37>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$$in>>2] = $36; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $38 = $__t$sroa$10$0; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $39 = ((($38)) + 4|0); //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $40 = $39; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $41 = ((($37)) + 4|0); //@line 954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $42 = ($41|0)==($1|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($42) {
      $$lcssa = $40;
      break;
     } else {
      $$in = $39;$37 = $41;$__t$sroa$10$0 = $40;
     }
    }
    $$pre = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $44 = $$pre;$__t$sroa$10$1 = $$lcssa;
   }
   HEAP32[$this>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$7>>2] = $32; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$0>>2] = $__t$sroa$10$1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$2>>2] = $34; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $43 = ($44|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $45 = $__t$sroa$10$1; //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($43) {
    $47 = $45;
   } else {
    __ZdlPv($44); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $$pre4 = HEAP32[$0>>2]|0; //@line 600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $47 = $$pre4;
   }
  } else {
   $47 = $1;
  }
 } while(0);
 $46 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$47>>2] = $46; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $48 = HEAP32[$0>>2]|0; //@line 602 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $49 = ((($48)) + 4|0); //@line 602 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 HEAP32[$0>>2] = $49; //@line 602 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 return; //@line 603 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
}
function __ZNSt3__114__split_bufferIPPN4love5event7MessageENS_9allocatorIS5_EEE10push_frontEOS5_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$cast = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__t$sroa$10$0 = 0, $__t$sroa$10$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 515 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $1 = HEAP32[$0>>2]|0; //@line 528 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $2 = HEAP32[$this>>2]|0; //@line 515 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $3 = ($1|0)==($2|0); //@line 515 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $4 = $2; //@line 526 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 do {
  if ($3) {
   $5 = ((($this)) + 8|0); //@line 517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $6 = HEAP32[$5>>2]|0; //@line 529 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $7 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $8 = HEAP32[$7>>2]|0; //@line 517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $9 = ($6>>>0)<($8>>>0); //@line 517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $$cast = $8; //@line 519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($9) {
    $10 = $6; //@line 519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $11 = (($$cast) - ($10))|0; //@line 519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $12 = $11 >> 2; //@line 519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $13 = (($12) + 1)|0; //@line 520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $14 = (($13|0) / 2)&-1; //@line 520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $15 = (($6) + ($14<<2)|0); //@line 521 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $16 = $1; //@line 1927 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $17 = (($10) - ($16))|0; //@line 1927 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $18 = $17 >> 2; //@line 1927 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $19 = (0 - ($18))|0; //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $20 = (($15) + ($19<<2)|0); //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    _memmove(($20|0),($1|0),($17|0))|0; //@line 1929 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    HEAP32[$0>>2] = $20; //@line 521 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $21 = HEAP32[$5>>2]|0; //@line 522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $22 = (($21) + ($14<<2)|0); //@line 522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    HEAP32[$5>>2] = $22; //@line 522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $47 = $20;
    break;
   }
   $23 = (($$cast) - ($4))|0; //@line 526 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $24 = $23 >> 1; //@line 526 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $25 = ($24|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $26 = $25 ? 1 : $24; //@line 526 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $27 = (($26) + 3)|0; //@line 527 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $28 = $27 >>> 2; //@line 527 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $29 = $26 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $30 = (__Znwj($29)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $31 = $30; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $32 = (($30) + ($28<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $33 = $32; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $34 = (($30) + ($26<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $35 = $34; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $36 = ($1|0)==($6|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($36) {
    $45 = $1;$__t$sroa$10$1 = $33;
   } else {
    $$in = $32;$38 = $1;$__t$sroa$10$0 = $33;
    while(1) {
     $37 = HEAP32[$38>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$$in>>2] = $37; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $39 = $__t$sroa$10$0; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $40 = ((($39)) + 4|0); //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $41 = $40; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $42 = ((($38)) + 4|0); //@line 954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $43 = ($42|0)==($6|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($43) {
      $$lcssa = $41;
      break;
     } else {
      $$in = $40;$38 = $42;$__t$sroa$10$0 = $41;
     }
    }
    $$pre = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $45 = $$pre;$__t$sroa$10$1 = $$lcssa;
   }
   HEAP32[$this>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$0>>2] = $33; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$5>>2] = $__t$sroa$10$1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$7>>2] = $35; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $44 = ($45|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($44) {
    $47 = $32;
   } else {
    __ZdlPv($45); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $$pre4 = HEAP32[$0>>2]|0; //@line 536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $47 = $$pre4;
   }
  } else {
   $47 = $1;
  }
 } while(0);
 $46 = ((($47)) + -4|0); //@line 536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $48 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$46>>2] = $48; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $49 = HEAP32[$0>>2]|0; //@line 538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $50 = ((($49)) + -4|0); //@line 538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 HEAP32[$0>>2] = $50; //@line 538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 return; //@line 539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
}
function __ZN4love5event5Event4pollERPNS0_7MessageE($this,$msg) {
 $this = $this|0;
 $msg = $msg|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lock = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ((($this)) + 8|0); //@line 97 "modules/event/Event.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 97 "modules/event/Event.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$1); //@line 97 "modules/event/Event.cpp"
 $2 = ((($this)) + 32|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 2538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $4 = ($3|0)==(0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($4) {
  $$0 = 0;
  __ZN4love6thread4LockD2Ev($lock); //@line 103 "modules/event/Event.cpp"
  STACKTOP = sp;return ($$0|0); //@line 103 "modules/event/Event.cpp"
 }
 $5 = ((($this)) + 16|0); //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $6 = HEAP32[$5>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $7 = ((($this)) + 28|0); //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $8 = HEAP32[$7>>2]|0; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $9 = $8 >>> 10; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $10 = (($6) + ($9<<2)|0); //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $11 = HEAP32[$10>>2]|0; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $12 = $8 & 1023; //@line 1734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $13 = (($11) + ($12<<2)|0); //@line 1734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $14 = HEAP32[$13>>2]|0; //@line 100 "modules/event/Event.cpp"
 HEAP32[$msg>>2] = $14; //@line 100 "modules/event/Event.cpp"
 $15 = (($3) + -1)|0; //@line 2538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 HEAP32[$2>>2] = $15; //@line 2538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $16 = (($8) + 1)|0; //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 HEAP32[$7>>2] = $16; //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $17 = ($16>>>0)>(2047); //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if (!($17)) {
  $$0 = 1;
  __ZN4love6thread4LockD2Ev($lock); //@line 103 "modules/event/Event.cpp"
  STACKTOP = sp;return ($$0|0); //@line 103 "modules/event/Event.cpp"
 }
 $18 = HEAP32[$5>>2]|0; //@line 90 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $19 = HEAP32[$18>>2]|0; //@line 2541 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 __ZdlPv($19); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $20 = HEAP32[$5>>2]|0; //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $21 = ((($20)) + 4|0); //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 HEAP32[$5>>2] = $21; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $22 = HEAP32[$7>>2]|0; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $23 = (($22) + -1024)|0; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 HEAP32[$7>>2] = $23; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $$0 = 1;
 __ZN4love6thread4LockD2Ev($lock); //@line 103 "modules/event/Event.cpp"
 STACKTOP = sp;return ($$0|0); //@line 103 "modules/event/Event.cpp"
}
function __ZN4love10filesystem6physfs10FilesystemD2Ev($this) {
 $this = $this|0;
 var $$pre$i = 0, $$pre$i15 = 0, $$pre$i3 = 0, $$pre$i9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (23184); //@line 113 "modules/filesystem/physfs/Filesystem.cpp"
 __THREW__ = 0;
 $0 = (invoke_i(3962)|0); //@line 114 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 do {
  if (!($2)) {
   $3 = ($0|0)==(0); //@line 114 "modules/filesystem/physfs/Filesystem.cpp"
   if (!($3)) {
    __THREW__ = 0;
    (invoke_i(3963)|0); //@line 115 "modules/filesystem/physfs/Filesystem.cpp"
    $4 = __THREW__; __THREW__ = 0;
    $5 = $4&1;
    if ($5) {
     break;
    }
   }
   $36 = ((($this)) + 100|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
   $37 = HEAP32[$36>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $38 = ($37|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($38)) {
    $39 = ((($this)) + 104|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $40 = HEAP32[$39>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $41 = ($40|0)==($37|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($41) {
     $46 = $37;
    } else {
     $43 = $40;
     while(1) {
      $42 = ((($43)) + -12|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$39>>2] = $42; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($42); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $44 = HEAP32[$39>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $45 = ($44|0)==($37|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($45) {
       break;
      } else {
       $43 = $44;
      }
     }
     $$pre$i9 = HEAP32[$36>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $46 = $$pre$i9;
    }
    __ZdlPv($46); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $47 = ((($this)) + 88|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
   $48 = HEAP32[$47>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $49 = ($48|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($49) {
    $58 = ((($this)) + 72|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($58); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
    $59 = ((($this)) + 60|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($59); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
    $60 = ((($this)) + 48|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($60); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
    $61 = ((($this)) + 36|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($61); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
    $62 = ((($this)) + 24|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($62); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
    $63 = ((($this)) + 12|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($63); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
    __ZN4love10filesystem10FilesystemD2Ev($this); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
    return; //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
   }
   $50 = ((($this)) + 92|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $51 = HEAP32[$50>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $52 = ($51|0)==($48|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($52) {
    $57 = $48;
   } else {
    $54 = $51;
    while(1) {
     $53 = ((($54)) + -12|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$50>>2] = $53; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($53); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $55 = HEAP32[$50>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $56 = ($55|0)==($48|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($56) {
      break;
     } else {
      $54 = $55;
     }
    }
    $$pre$i15 = HEAP32[$47>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $57 = $$pre$i15;
   }
   __ZdlPv($57); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $58 = ((($this)) + 72|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($58); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
   $59 = ((($this)) + 60|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($59); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
   $60 = ((($this)) + 48|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($60); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
   $61 = ((($this)) + 36|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($61); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
   $62 = ((($this)) + 24|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($62); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
   $63 = ((($this)) + 12|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($63); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
   __ZN4love10filesystem10FilesystemD2Ev($this); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
   return; //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
  }
 } while(0);
 $6 = ___cxa_find_matching_catch_3(0|0)|0;
 $7 = tempRet0;
 $8 = ((($this)) + 100|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = ($9|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($10)) {
  $11 = ((($this)) + 104|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $12 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = ($12|0)==($9|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($13) {
   $18 = $9;
  } else {
   $15 = $12;
   while(1) {
    $14 = ((($15)) + -12|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$11>>2] = $14; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($14); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $16 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $17 = ($16|0)==($9|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($17) {
     break;
    } else {
     $15 = $16;
    }
   }
   $$pre$i = HEAP32[$8>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $18 = $$pre$i;
  }
  __ZdlPv($18); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $19 = ((($this)) + 88|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = ($20|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($21) {
  $30 = ((($this)) + 72|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($30); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
  $31 = ((($this)) + 60|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($31); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
  $32 = ((($this)) + 48|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($32); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
  $33 = ((($this)) + 36|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($33); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
  $34 = ((($this)) + 24|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
  $35 = ((($this)) + 12|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($35); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
  __ZN4love10filesystem10FilesystemD2Ev($this); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
  ___clang_call_terminate($6); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
  // unreachable; //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $22 = ((($this)) + 92|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $23 = HEAP32[$22>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $24 = ($23|0)==($20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($24) {
  $29 = $20;
 } else {
  $26 = $23;
  while(1) {
   $25 = ((($26)) + -12|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$22>>2] = $25; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($25); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $27 = HEAP32[$22>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $28 = ($27|0)==($20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($28) {
    break;
   } else {
    $26 = $27;
   }
  }
  $$pre$i3 = HEAP32[$19>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $29 = $$pre$i3;
 }
 __ZdlPv($29); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $30 = ((($this)) + 72|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($30); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
 $31 = ((($this)) + 60|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($31); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
 $32 = ((($this)) + 48|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($32); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
 $33 = ((($this)) + 36|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($33); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
 $34 = ((($this)) + 24|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
 $35 = ((($this)) + 12|0); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($35); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
 __ZN4love10filesystem10FilesystemD2Ev($this); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
 ___clang_call_terminate($6); //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
 // unreachable; //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZN4love10filesystem6physfs10FilesystemD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love10filesystem6physfs10FilesystemD2Ev($this); //@line 113 "modules/filesystem/physfs/Filesystem.cpp"
 __ZdlPv($this); //@line 113 "modules/filesystem/physfs/Filesystem.cpp"
 return; //@line 116 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZNK4love10filesystem10Filesystem13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 2; //@line 68 "./modules/filesystem/Filesystem.h"
}
function __ZNK4love10filesystem6physfs10Filesystem7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (366079|0); //@line 120 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZN4love10filesystem6physfs10Filesystem4initEPKc($this,$arg0) {
 $this = $this|0;
 $arg0 = $arg0|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_PHYSFS_init($arg0)|0); //@line 125 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = ($0|0)==(0); //@line 125 "modules/filesystem/physfs/Filesystem.cpp"
 if (!($1)) {
  $10 = HEAP32[$this>>2]|0; //@line 129 "modules/filesystem/physfs/Filesystem.cpp"
  $11 = ((($10)) + 156|0); //@line 129 "modules/filesystem/physfs/Filesystem.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 129 "modules/filesystem/physfs/Filesystem.cpp"
  FUNCTION_TABLE_vii[$12 & 8191]($this,1); //@line 129 "modules/filesystem/physfs/Filesystem.cpp"
  STACKTOP = sp;return; //@line 130 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $2 = (___cxa_allocate_exception(16)|0); //@line 126 "modules/filesystem/physfs/Filesystem.cpp"
 __THREW__ = 0;
 $3 = (invoke_i(3964)|0); //@line 126 "modules/filesystem/physfs/Filesystem.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $8 = ___cxa_find_matching_catch_2()|0;
  $9 = tempRet0;
  ___cxa_free_exception(($2|0)); //@line 126 "modules/filesystem/physfs/Filesystem.cpp"
  ___resumeException($8|0);
  // unreachable;
 }
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $3; //@line 126 "modules/filesystem/physfs/Filesystem.cpp"
 invoke_viii(3349,($2|0),(595867|0),($vararg_buffer|0)); //@line 126 "modules/filesystem/physfs/Filesystem.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $8 = ___cxa_find_matching_catch_2()|0;
  $9 = tempRet0;
  ___cxa_free_exception(($2|0)); //@line 126 "modules/filesystem/physfs/Filesystem.cpp"
  ___resumeException($8|0);
  // unreachable;
 } else {
  ___cxa_throw(($2|0),(96|0),(344|0)); //@line 126 "modules/filesystem/physfs/Filesystem.cpp"
  // unreachable; //@line 126 "modules/filesystem/physfs/Filesystem.cpp"
 }
}
function __ZN4love10filesystem6physfs10Filesystem8setFusedEb($this,$fused) {
 $this = $this|0;
 $fused = $fused|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 85|0); //@line 134 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 134 "modules/filesystem/physfs/Filesystem.cpp"
 $2 = ($1<<24>>24)==(0); //@line 134 "modules/filesystem/physfs/Filesystem.cpp"
 if (!($2)) {
  return; //@line 138 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $3 = ((($this)) + 84|0); //@line 136 "modules/filesystem/physfs/Filesystem.cpp"
 $4 = $fused&1; //@line 136 "modules/filesystem/physfs/Filesystem.cpp"
 HEAP8[$3>>0] = $4; //@line 136 "modules/filesystem/physfs/Filesystem.cpp"
 HEAP8[$0>>0] = 1; //@line 137 "modules/filesystem/physfs/Filesystem.cpp"
 return; //@line 138 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZNK4love10filesystem6physfs10Filesystem7isFusedEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 85|0); //@line 142 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 142 "modules/filesystem/physfs/Filesystem.cpp"
 $2 = ($1<<24>>24)==(0); //@line 142 "modules/filesystem/physfs/Filesystem.cpp"
 if ($2) {
  $$0 = 0;
  return ($$0|0); //@line 145 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $3 = ((($this)) + 84|0); //@line 144 "modules/filesystem/physfs/Filesystem.cpp"
 $4 = HEAP8[$3>>0]|0; //@line 144 "modules/filesystem/physfs/Filesystem.cpp"
 $5 = ($4<<24>>24)!=(0); //@line 144 "modules/filesystem/physfs/Filesystem.cpp"
 $$0 = $5;
 return ($$0|0); //@line 145 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZN4love10filesystem6physfs10Filesystem19setupWriteDirectoryEv($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$016$i$i$i = 0, $$016$i$i$i$lcssa = 0, $$03 = 0, $$1 = 0, $$12 = 0, $$14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0;
 var $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0;
 var $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0;
 var $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $__m1$0$i$i$i = 0, $__m2$0$i$i$i = 0, $__ps$06$i$i = 0, $__ps$06$i$i$lcssa = 0, $i$03$i$i = 0, $i$03$i$i11 = 0, $temp_createdir = 0, $temp_writedir = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $temp_writedir = sp + 72|0;
 $temp_createdir = sp + 60|0;
 $0 = sp + 48|0;
 $1 = sp + 36|0;
 $2 = sp + 24|0;
 $3 = sp + 12|0;
 $4 = sp;
 $5 = (_PHYSFS_isInit()|0); //@line 312 "modules/filesystem/physfs/Filesystem.cpp"
 $6 = ($5|0)==(0); //@line 312 "modules/filesystem/physfs/Filesystem.cpp"
 if ($6) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 364 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $7 = ((($this)) + 36|0); //@line 316 "modules/filesystem/physfs/Filesystem.cpp"
 $8 = HEAP8[$7>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $8 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = ($9<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($10) {
  $13 = $8&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $14 = $13 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $16 = $14;
 } else {
  $11 = ((($this)) + 40|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $12 = HEAP32[$11>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $16 = $12;
 }
 $15 = ($16|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($15) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 364 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $17 = ((($this)) + 60|0); //@line 316 "modules/filesystem/physfs/Filesystem.cpp"
 $18 = HEAP8[$17>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $19 = $18 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $20 = ($19<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($20) {
  $23 = $18&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $24 = $23 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $26 = $24;
 } else {
  $21 = ((($this)) + 64|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $22 = HEAP32[$21>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $26 = $22;
 }
 $25 = ($26|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($25) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 364 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $27 = ((($this)) + 48|0); //@line 316 "modules/filesystem/physfs/Filesystem.cpp"
 $28 = HEAP8[$27>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $29 = $28 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $30 = ($29<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($30) {
  $33 = $28&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $34 = $33 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $36 = $34;
 } else {
  $31 = ((($this)) + 52|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $32 = HEAP32[$31>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $36 = $32;
 }
 $35 = ($36|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($35) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 364 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $37 = ((($this)) + 64|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $38 = HEAP32[$37>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $39 = $18&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $40 = $39 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $41 = $20 ? $40 : $38; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $42 = ($41|0)==(0); //@line 64 "modules/filesystem/physfs/Filesystem.cpp"
 L25: do {
  if ($42) {
   $52 = 0;
  } else {
   $43 = ((($this)) + 68|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $44 = HEAP32[$43>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $45 = ((($17)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $46 = $20 ? $45 : $44; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $i$03$i$i = 0;
   while(1) {
    $47 = (($46) + ($i$03$i$i)|0); //@line 3262 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $48 = HEAP8[$47>>0]|0; //@line 65 "modules/filesystem/physfs/Filesystem.cpp"
    switch ($48<<24>>24) {
    case 47: case 92:  {
     $52 = $i$03$i$i;
     break L25;
     break;
    }
    default: {
    }
    }
    $49 = (($i$03$i$i) + 1)|0; //@line 64 "modules/filesystem/physfs/Filesystem.cpp"
    $50 = ($49>>>0)<($41>>>0); //@line 64 "modules/filesystem/physfs/Filesystem.cpp"
    if ($50) {
     $i$03$i$i = $49;
    } else {
     $52 = 0;
     break L25;
    }
   }
  }
 } while(0);
 $51 = (($52) + 1)|0; //@line 73 "modules/filesystem/physfs/Filesystem.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_jjRKS4_($temp_writedir,$17,0,$51,$17); //@line 3345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $53 = HEAP8[$17>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $54 = $53 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $55 = ($54<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $56 = HEAP32[$37>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $57 = $53&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $58 = $57 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $59 = $55 ? $58 : $56; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $60 = ($59|0)==(0); //@line 64 "modules/filesystem/physfs/Filesystem.cpp"
 L31: do {
  if ($60) {
   $70 = 0;
  } else {
   $61 = ((($this)) + 68|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $62 = HEAP32[$61>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $63 = ((($17)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $64 = $55 ? $63 : $62; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $i$03$i$i11 = 0;
   while(1) {
    $65 = (($64) + ($i$03$i$i11)|0); //@line 3262 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $66 = HEAP8[$65>>0]|0; //@line 65 "modules/filesystem/physfs/Filesystem.cpp"
    switch ($66<<24>>24) {
    case 47: case 92:  {
     $70 = $i$03$i$i11;
     break L31;
     break;
    }
    default: {
    }
    }
    $67 = (($i$03$i$i11) + 1)|0; //@line 64 "modules/filesystem/physfs/Filesystem.cpp"
    $68 = ($67>>>0)<($59>>>0); //@line 64 "modules/filesystem/physfs/Filesystem.cpp"
    if ($68) {
     $i$03$i$i11 = $67;
    } else {
     $70 = 0;
     break L31;
    }
   }
  }
 } while(0);
 $69 = (($70) + 1)|0; //@line 78 "modules/filesystem/physfs/Filesystem.cpp"
 __THREW__ = 0;
 invoke_viiiii(3511,($temp_createdir|0),($17|0),($69|0),-1,($17|0)); //@line 3345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $71 = __THREW__; __THREW__ = 0;
 $72 = $71&1;
 if ($72) {
  $196 = ___cxa_find_matching_catch_2()|0;
  $197 = tempRet0;
  $$12 = $197;$$14 = $196;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($temp_writedir); //@line 364 "modules/filesystem/physfs/Filesystem.cpp"
  ___resumeException($$14|0);
  // unreachable;
 }
 $73 = HEAP32[$this>>2]|0; //@line 328 "modules/filesystem/physfs/Filesystem.cpp"
 $74 = ((($73)) + 88|0); //@line 328 "modules/filesystem/physfs/Filesystem.cpp"
 $75 = HEAP32[$74>>2]|0; //@line 328 "modules/filesystem/physfs/Filesystem.cpp"
 __THREW__ = 0;
 invoke_vii($75|0,($0|0),($this|0)); //@line 328 "modules/filesystem/physfs/Filesystem.cpp"
 $76 = __THREW__; __THREW__ = 0;
 $77 = $76&1;
 L41: do {
  if ($77) {
   label = 66;
  } else {
   $78 = HEAP8[$17>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $79 = $78 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $80 = ($79<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if ($80) {
    $84 = ((($17)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $85 = $78&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $86 = $85 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $100 = $86;$103 = $84;
   } else {
    $81 = ((($this)) + 68|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $82 = HEAP32[$81>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $83 = HEAP32[$37>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $100 = $83;$103 = $82;
   }
   $87 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $88 = $87 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $89 = ($88<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $90 = ((($0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $91 = HEAP32[$90>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $92 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $93 = $89 ? $92 : $91; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $94 = ((($0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $95 = HEAP32[$94>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $96 = $87&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $97 = $96 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $98 = $89 ? $97 : $95; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $99 = ($100>>>0)<($98>>>0); //@line 962 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   L47: do {
    if ($99) {
     label = 38;
    } else {
     $101 = ($98|0)==(0); //@line 964 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($101) {
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 328 "modules/filesystem/physfs/Filesystem.cpp"
     } else {
      $102 = (($103) + ($100)|0); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $104 = (($93) + ($98)|0); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $105 = ($100|0)<($98|0); //@line 1467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if ($105) {
       label = 38;
       break;
      }
      $106 = (1 - ($98))|0; //@line 1469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $107 = (($102) + ($106)|0); //@line 1469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $108 = ($107|0)==($103|0); //@line 1475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if ($108) {
       label = 38;
       break;
      }
      $109 = HEAP8[$93>>0]|0; //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $$016$i$i$i = $103;
      L54: while(1) {
       $110 = HEAP8[$$016$i$i$i>>0]|0; //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $111 = ($110<<24>>24)==($109<<24>>24); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($111) {
        $__m1$0$i$i$i = $$016$i$i$i;$__m2$0$i$i$i = $93;
        while(1) {
         $114 = ((($__m2$0$i$i$i)) + 1|0); //@line 1517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $115 = ($114|0)==($104|0); //@line 1517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         if ($115) {
          $$016$i$i$i$lcssa = $$016$i$i$i;
          break L54;
         }
         $116 = ((($__m1$0$i$i$i)) + 1|0); //@line 1519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $117 = HEAP8[$116>>0]|0; //@line 1520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $118 = HEAP8[$114>>0]|0; //@line 1520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $119 = ($117<<24>>24)==($118<<24>>24); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         if ($119) {
          $__m1$0$i$i$i = $116;$__m2$0$i$i$i = $114;
         } else {
          break;
         }
        }
       }
       $112 = ((($$016$i$i$i)) + 1|0); //@line 1479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $113 = ($112|0)==($107|0); //@line 1475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       if ($113) {
        label = 38;
        break L47;
       } else {
        $$016$i$i$i = $112;
       }
      }
      $120 = ($$016$i$i$i$lcssa|0)==($102|0); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($120) {
       label = 38;
       break;
      }
      $121 = ($$016$i$i$i$lcssa|0)==($103|0); //@line 328 "modules/filesystem/physfs/Filesystem.cpp"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 328 "modules/filesystem/physfs/Filesystem.cpp"
      if (!($121)) {
       break;
      }
     }
     $122 = HEAP32[$this>>2]|0; //@line 330 "modules/filesystem/physfs/Filesystem.cpp"
     $123 = ((($122)) + 88|0); //@line 330 "modules/filesystem/physfs/Filesystem.cpp"
     $124 = HEAP32[$123>>2]|0; //@line 330 "modules/filesystem/physfs/Filesystem.cpp"
     __THREW__ = 0;
     invoke_vii($124|0,($1|0),($this|0)); //@line 330 "modules/filesystem/physfs/Filesystem.cpp"
     $125 = __THREW__; __THREW__ = 0;
     $126 = $125&1;
     if ($126) {
      label = 66;
      break L41;
     }
     $127 = HEAP8[$temp_writedir>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $128 = $127 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $129 = ($128<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($129) {
      $133 = ((($temp_writedir)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$133>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$temp_writedir>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     } else {
      $130 = ((($temp_writedir)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $131 = HEAP32[$130>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$131>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $132 = ((($temp_writedir)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP32[$132>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     }
     __THREW__ = 0;
     invoke_vii(3368,($temp_writedir|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $134 = __THREW__; __THREW__ = 0;
     $135 = $134&1;
     if ($135) {
      $136 = ___cxa_find_matching_catch_3(0|0)|0;
      $137 = tempRet0;
      ___clang_call_terminate($136); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     }
     ;HEAP32[$temp_writedir>>2]=HEAP32[$1>>2]|0;HEAP32[$temp_writedir+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$temp_writedir+8>>2]=HEAP32[$1+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 330 "modules/filesystem/physfs/Filesystem.cpp"
     $138 = HEAP32[$this>>2]|0; //@line 331 "modules/filesystem/physfs/Filesystem.cpp"
     $139 = ((($138)) + 88|0); //@line 331 "modules/filesystem/physfs/Filesystem.cpp"
     $140 = HEAP32[$139>>2]|0; //@line 331 "modules/filesystem/physfs/Filesystem.cpp"
     __THREW__ = 0;
     invoke_vii($140|0,($3|0),($this|0)); //@line 331 "modules/filesystem/physfs/Filesystem.cpp"
     $141 = __THREW__; __THREW__ = 0;
     $142 = $141&1;
     if ($142) {
      label = 66;
      break L41;
     }
     $143 = HEAP8[$3>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $144 = $143 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $145 = ($144<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $146 = ((($3)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $147 = HEAP32[$146>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $148 = $143&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $149 = $148 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $150 = $145 ? $149 : $147; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     __THREW__ = 0;
     invoke_viiiii(3511,($2|0),($17|0),($150|0),-1,($17|0)); //@line 3345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $151 = __THREW__; __THREW__ = 0;
     $152 = $151&1;
     if ($152) {
      $200 = ___cxa_find_matching_catch_2()|0;
      $201 = tempRet0;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 331 "modules/filesystem/physfs/Filesystem.cpp"
      $$01 = $201;$$03 = $200;
      break L41;
     }
     $153 = HEAP8[$temp_createdir>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $154 = $153 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $155 = ($154<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($155) {
      $159 = ((($temp_createdir)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$159>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$temp_createdir>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     } else {
      $156 = ((($temp_createdir)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $157 = HEAP32[$156>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$157>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $158 = ((($temp_createdir)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP32[$158>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     }
     __THREW__ = 0;
     invoke_vii(3368,($temp_createdir|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $160 = __THREW__; __THREW__ = 0;
     $161 = $160&1;
     if ($161) {
      $162 = ___cxa_find_matching_catch_3(0|0)|0;
      $163 = tempRet0;
      ___clang_call_terminate($162); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     }
     ;HEAP32[$temp_createdir>>2]=HEAP32[$2>>2]|0;HEAP32[$temp_createdir+4>>2]=HEAP32[$2+4>>2]|0;HEAP32[$temp_createdir+8>>2]=HEAP32[$2+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     ;HEAP32[$2>>2]=0|0;HEAP32[$2+4>>2]=0|0;HEAP32[$2+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 331 "modules/filesystem/physfs/Filesystem.cpp"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 331 "modules/filesystem/physfs/Filesystem.cpp"
     $164 = HEAP8[$temp_createdir>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $165 = $164 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $166 = ($165<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $167 = ((($temp_createdir)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $168 = HEAP32[$167>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $169 = ((($temp_createdir)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $170 = HEAP32[$169>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $171 = ((($temp_createdir)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $172 = $164&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $173 = $172 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $174 = $166 ? $171 : $168; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $175 = $166 ? $173 : $170; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $176 = ($175|0)==(0); //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($176) {
      break;
     }
     $177 = (($174) + ($175)|0); //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $__ps$06$i$i = $174;
     while(1) {
      $178 = HEAP8[$__ps$06$i$i>>0]|0; //@line 1080 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $179 = ($178<<24>>24)==(47); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if (!($179)) {
       $__ps$06$i$i$lcssa = $__ps$06$i$i;
       break;
      }
      $180 = ((($__ps$06$i$i)) + 1|0); //@line 1079 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $181 = ($180|0)==($177|0); //@line 1079 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($181) {
       break L47;
      } else {
       $__ps$06$i$i = $180;
      }
     }
     $182 = $__ps$06$i$i$lcssa; //@line 1081 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $183 = $174; //@line 1081 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $184 = (($182) - ($183))|0; //@line 1081 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $185 = ($184|0)==(-1); //@line 335 "modules/filesystem/physfs/Filesystem.cpp"
     if ($185) {
      break;
     }
     __THREW__ = 0;
     invoke_viiiii(3511,($4|0),($temp_createdir|0),($184|0),-1,($temp_createdir|0)); //@line 3345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $186 = __THREW__; __THREW__ = 0;
     $187 = $186&1;
     if ($187) {
      label = 66;
      break L41;
     }
     $188 = HEAP8[$temp_createdir>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $189 = $188 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $190 = ($189<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($190) {
      HEAP8[$171>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$temp_createdir>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     } else {
      $191 = HEAP32[$167>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$191>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP32[$169>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     }
     __THREW__ = 0;
     invoke_vii(3368,($temp_createdir|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $192 = __THREW__; __THREW__ = 0;
     $193 = $192&1;
     if ($193) {
      $194 = ___cxa_find_matching_catch_3(0|0)|0;
      $195 = tempRet0;
      ___clang_call_terminate($194); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     } else {
      ;HEAP32[$temp_createdir>>2]=HEAP32[$4>>2]|0;HEAP32[$temp_createdir+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$temp_createdir+8>>2]=HEAP32[$4+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 336 "modules/filesystem/physfs/Filesystem.cpp"
      break;
     }
    }
   } while(0);
   if ((label|0) == 38) {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 328 "modules/filesystem/physfs/Filesystem.cpp"
   }
   $202 = HEAP8[$temp_writedir>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $203 = $202 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $204 = ($203<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $205 = ((($temp_writedir)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $206 = HEAP32[$205>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $207 = ((($temp_writedir)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $208 = $204 ? $207 : $206; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   $209 = (invoke_ii(3965,($208|0))|0); //@line 341 "modules/filesystem/physfs/Filesystem.cpp"
   $210 = __THREW__; __THREW__ = 0;
   $211 = $210&1;
   if ($211) {
    label = 66;
   } else {
    $212 = ($209|0)==(0); //@line 341 "modules/filesystem/physfs/Filesystem.cpp"
    do {
     if ($212) {
      $$0 = 0;
     } else {
      $213 = HEAP32[$this>>2]|0; //@line 345 "modules/filesystem/physfs/Filesystem.cpp"
      $214 = ((($213)) + 124|0); //@line 345 "modules/filesystem/physfs/Filesystem.cpp"
      $215 = HEAP32[$214>>2]|0; //@line 345 "modules/filesystem/physfs/Filesystem.cpp"
      $216 = HEAP8[$temp_createdir>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $217 = $216 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $218 = ($217<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $219 = ((($temp_createdir)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $220 = HEAP32[$219>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $221 = ((($temp_createdir)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $222 = $218 ? $221 : $220; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __THREW__ = 0;
      $223 = (invoke_iii($215|0,($this|0),($222|0))|0); //@line 345 "modules/filesystem/physfs/Filesystem.cpp"
      $224 = __THREW__; __THREW__ = 0;
      $225 = $224&1;
      if ($225) {
       label = 66;
       break L41;
      }
      if (!($223)) {
       __THREW__ = 0;
       (invoke_ii(3965,(0|0))|0); //@line 348 "modules/filesystem/physfs/Filesystem.cpp"
       $226 = __THREW__; __THREW__ = 0;
       $227 = $226&1;
       if ($227) {
        label = 66;
        break L41;
       } else {
        $$0 = 0;
        break;
       }
      }
      $228 = HEAP8[$17>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $229 = $228 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $230 = ($229<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($230) {
       $233 = ((($17)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $234 = $233;
      } else {
       $231 = ((($this)) + 68|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $232 = HEAP32[$231>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $234 = $232;
      }
      __THREW__ = 0;
      $235 = (invoke_ii(3965,($234|0))|0); //@line 353 "modules/filesystem/physfs/Filesystem.cpp"
      $236 = __THREW__; __THREW__ = 0;
      $237 = $236&1;
      if ($237) {
       label = 66;
       break L41;
      }
      $238 = ($235|0)==(0); //@line 353 "modules/filesystem/physfs/Filesystem.cpp"
      if ($238) {
       $$0 = 0;
       break;
      }
      $239 = HEAP8[$17>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $240 = $239 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $241 = ($240<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($241) {
       $244 = ((($17)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $245 = $244;
      } else {
       $242 = ((($this)) + 68|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $243 = HEAP32[$242>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $245 = $243;
      }
      __THREW__ = 0;
      $246 = (invoke_iiii(3966,($245|0),(0|0),0)|0); //@line 357 "modules/filesystem/physfs/Filesystem.cpp"
      $247 = __THREW__; __THREW__ = 0;
      $248 = $247&1;
      if ($248) {
       label = 66;
       break L41;
      }
      $249 = ($246|0)==(0); //@line 357 "modules/filesystem/physfs/Filesystem.cpp"
      if (!($249)) {
       $$0 = 1;
       break;
      }
      __THREW__ = 0;
      (invoke_ii(3965,(0|0))|0); //@line 359 "modules/filesystem/physfs/Filesystem.cpp"
      $250 = __THREW__; __THREW__ = 0;
      $251 = $250&1;
      if ($251) {
       label = 66;
       break L41;
      } else {
       $$0 = 0;
      }
     }
    } while(0);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($temp_createdir); //@line 364 "modules/filesystem/physfs/Filesystem.cpp"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($temp_writedir); //@line 364 "modules/filesystem/physfs/Filesystem.cpp"
    $$1 = $$0;
    STACKTOP = sp;return ($$1|0); //@line 364 "modules/filesystem/physfs/Filesystem.cpp"
   }
  }
 } while(0);
 if ((label|0) == 66) {
  $198 = ___cxa_find_matching_catch_2()|0;
  $199 = tempRet0;
  $$01 = $199;$$03 = $198;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($temp_createdir); //@line 364 "modules/filesystem/physfs/Filesystem.cpp"
 $$12 = $$01;$$14 = $$03;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($temp_writedir); //@line 364 "modules/filesystem/physfs/Filesystem.cpp"
 ___resumeException($$14|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love10filesystem6physfs10Filesystem11setIdentityEPKcb($this,$ident,$appendToPath) {
 $this = $this|0;
 $ident = $ident|0;
 $appendToPath = $appendToPath|0;
 var $$0 = 0, $$01 = 0, $$04 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0;
 var $182 = 0, $183 = 0, $184 = 0, $185 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $old_save_path = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $old_save_path = sp + 108|0;
 $0 = sp + 96|0;
 $1 = sp + 84|0;
 $2 = sp + 72|0;
 $3 = sp + 60|0;
 $4 = sp + 48|0;
 $5 = sp + 36|0;
 $6 = sp + 24|0;
 $7 = sp + 12|0;
 $8 = sp;
 $9 = (_PHYSFS_isInit()|0); //@line 149 "modules/filesystem/physfs/Filesystem.cpp"
 $10 = ($9|0)==(0); //@line 149 "modules/filesystem/physfs/Filesystem.cpp"
 if ($10) {
  $$04 = 0;
  STACKTOP = sp;return ($$04|0); //@line 217 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $11 = ((($this)) + 60|0); //@line 152 "modules/filesystem/physfs/Filesystem.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_($old_save_path,$11); //@line 152 "modules/filesystem/physfs/Filesystem.cpp"
 $12 = ((($this)) + 36|0); //@line 155 "modules/filesystem/physfs/Filesystem.cpp"
 $13 = (_strlen($ident)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viii(3460,($0|0),($ident|0),($13|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 L4: do {
  if (!($15)) {
   $16 = HEAP8[$12>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $17 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $18 = ($17<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if ($18) {
    $22 = ((($12)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$22>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$12>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   } else {
    $19 = ((($this)) + 44|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $20 = HEAP32[$19>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$20>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $21 = ((($this)) + 40|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP32[$21>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   }
   __THREW__ = 0;
   invoke_vii(3368,($12|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $23 = __THREW__; __THREW__ = 0;
   $24 = $23&1;
   if ($24) {
    $25 = ___cxa_find_matching_catch_3(0|0)|0;
    $26 = tempRet0;
    ___clang_call_terminate($25); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   }
   ;HEAP32[$12>>2]=HEAP32[$0>>2]|0;HEAP32[$12+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$12+8>>2]=HEAP32[$0+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 155 "modules/filesystem/physfs/Filesystem.cpp"
   $27 = ((($this)) + 48|0); //@line 158 "modules/filesystem/physfs/Filesystem.cpp"
   __THREW__ = 0;
   invoke_viii(3460,($2|0),(366102|0),5); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $28 = __THREW__; __THREW__ = 0;
   $29 = $28&1;
   if (!($29)) {
    $30 = HEAP8[$12>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $31 = $30 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $32 = ($31<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $33 = ((($this)) + 44|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $34 = HEAP32[$33>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $35 = ((($12)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $36 = $32 ? $35 : $34; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $37 = ((($this)) + 40|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $38 = HEAP32[$37>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $39 = $30&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $40 = $39 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $41 = $32 ? $40 : $38; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    __THREW__ = 0;
    $42 = (invoke_iiii(3462,($2|0),($36|0),($41|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $43 = __THREW__; __THREW__ = 0;
    $44 = $43&1;
    if ($44) {
     $122 = ___cxa_find_matching_catch_2()|0;
     $123 = tempRet0;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 158 "modules/filesystem/physfs/Filesystem.cpp"
     $$2 = $123;$$23 = $122;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($old_save_path); //@line 217 "modules/filesystem/physfs/Filesystem.cpp"
     ___resumeException($$23|0);
     // unreachable;
    }
    ;HEAP32[$1>>2]=HEAP32[$42>>2]|0;HEAP32[$1+4>>2]=HEAP32[$42+4>>2]|0;HEAP32[$1+8>>2]=HEAP32[$42+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    ;HEAP32[$42>>2]=0|0;HEAP32[$42+4>>2]=0|0;HEAP32[$42+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $45 = HEAP8[$27>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $46 = $45 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $47 = ($46<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if ($47) {
     $51 = ((($27)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP8[$51>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP8[$27>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    } else {
     $48 = ((($this)) + 56|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $49 = HEAP32[$48>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP8[$49>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $50 = ((($this)) + 52|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP32[$50>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    }
    __THREW__ = 0;
    invoke_vii(3368,($27|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $52 = __THREW__; __THREW__ = 0;
    $53 = $52&1;
    if ($53) {
     $54 = ___cxa_find_matching_catch_3(0|0)|0;
     $55 = tempRet0;
     ___clang_call_terminate($54); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    }
    ;HEAP32[$27>>2]=HEAP32[$1>>2]|0;HEAP32[$27+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$27+8>>2]=HEAP32[$1+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 158 "modules/filesystem/physfs/Filesystem.cpp"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 158 "modules/filesystem/physfs/Filesystem.cpp"
    $56 = HEAP32[$this>>2]|0; //@line 161 "modules/filesystem/physfs/Filesystem.cpp"
    $57 = ((($56)) + 92|0); //@line 161 "modules/filesystem/physfs/Filesystem.cpp"
    $58 = HEAP32[$57>>2]|0; //@line 161 "modules/filesystem/physfs/Filesystem.cpp"
    __THREW__ = 0;
    invoke_vii($58|0,($4|0),($this|0)); //@line 161 "modules/filesystem/physfs/Filesystem.cpp"
    $59 = __THREW__; __THREW__ = 0;
    $60 = $59&1;
    if (!($60)) {
     __THREW__ = 0;
     invoke_viii(3460,($5|0),(523303|0),1); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $61 = __THREW__; __THREW__ = 0;
     $62 = $61&1;
     do {
      if ($62) {
       $124 = ___cxa_find_matching_catch_2()|0;
       $125 = tempRet0;
       $$0 = $125;$$01 = $124;
      } else {
       $63 = HEAP8[$5>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $64 = $63 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $65 = ($64<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $66 = ((($5)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $67 = HEAP32[$66>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $68 = ((($5)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $69 = $65 ? $68 : $67; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $70 = ((($5)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $71 = HEAP32[$70>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $72 = $63&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $73 = $72 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $74 = $65 ? $73 : $71; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       __THREW__ = 0;
       $75 = (invoke_iiii(3462,($4|0),($69|0),($74|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $76 = __THREW__; __THREW__ = 0;
       $77 = $76&1;
       if ($77) {
        $126 = ___cxa_find_matching_catch_2()|0;
        $127 = tempRet0;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5); //@line 161 "modules/filesystem/physfs/Filesystem.cpp"
        $$0 = $127;$$01 = $126;
        break;
       }
       ;HEAP32[$3>>2]=HEAP32[$75>>2]|0;HEAP32[$3+4>>2]=HEAP32[$75+4>>2]|0;HEAP32[$3+8>>2]=HEAP32[$75+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       ;HEAP32[$75>>2]=0|0;HEAP32[$75+4>>2]=0|0;HEAP32[$75+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $78 = HEAP8[$11>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $79 = $78 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $80 = ($79<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($80) {
        $84 = ((($11)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP8[$84>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP8[$11>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       } else {
        $81 = ((($this)) + 68|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $82 = HEAP32[$81>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP8[$82>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $83 = ((($this)) + 64|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP32[$83>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       }
       __THREW__ = 0;
       invoke_vii(3368,($11|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $85 = __THREW__; __THREW__ = 0;
       $86 = $85&1;
       if ($86) {
        $87 = ___cxa_find_matching_catch_3(0|0)|0;
        $88 = tempRet0;
        ___clang_call_terminate($87); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       }
       ;HEAP32[$11>>2]=HEAP32[$3>>2]|0;HEAP32[$11+4>>2]=HEAP32[$3+4>>2]|0;HEAP32[$11+8>>2]=HEAP32[$3+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       ;HEAP32[$3>>2]=0|0;HEAP32[$3+4>>2]=0|0;HEAP32[$3+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 161 "modules/filesystem/physfs/Filesystem.cpp"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5); //@line 161 "modules/filesystem/physfs/Filesystem.cpp"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 161 "modules/filesystem/physfs/Filesystem.cpp"
       $89 = ((($this)) + 84|0); //@line 162 "modules/filesystem/physfs/Filesystem.cpp"
       $90 = HEAP8[$89>>0]|0; //@line 162 "modules/filesystem/physfs/Filesystem.cpp"
       $91 = ($90<<24>>24)==(0); //@line 162 "modules/filesystem/physfs/Filesystem.cpp"
       L39: do {
        if ($91) {
         $132 = HEAP8[$27>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $133 = $132 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $134 = ($133<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $135 = ((($this)) + 56|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $136 = HEAP32[$135>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $137 = ((($27)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $138 = $134 ? $137 : $136; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $139 = ((($this)) + 52|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $140 = HEAP32[$139>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $141 = $132&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $142 = $141 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $143 = $134 ? $142 : $140; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         __THREW__ = 0;
         (invoke_iiii(3462,($11|0),($138|0),($143|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $144 = __THREW__; __THREW__ = 0;
         $145 = $144&1;
         if ($145) {
          break L4;
         }
        } else {
         __THREW__ = 0;
         invoke_viii(3460,($7|0),(1282466|0),0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $92 = __THREW__; __THREW__ = 0;
         $93 = $92&1;
         if ($93) {
          break L4;
         }
         $94 = HEAP8[$12>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $95 = $94 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $96 = ($95<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $97 = HEAP32[$33>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $98 = $96 ? $35 : $97; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $99 = HEAP32[$37>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $100 = $94&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $101 = $100 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $102 = $96 ? $101 : $99; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         __THREW__ = 0;
         $103 = (invoke_iiii(3462,($7|0),($98|0),($102|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $104 = __THREW__; __THREW__ = 0;
         $105 = $104&1;
         do {
          if ($105) {
           $128 = ___cxa_find_matching_catch_2()|0;
           $129 = tempRet0;
           $$1 = $129;$$12 = $128;
          } else {
           ;HEAP32[$6>>2]=HEAP32[$103>>2]|0;HEAP32[$6+4>>2]=HEAP32[$103+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$103+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           ;HEAP32[$103>>2]=0|0;HEAP32[$103+4>>2]=0|0;HEAP32[$103+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $106 = HEAP8[$6>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $107 = $106 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $108 = ($107<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $109 = ((($6)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $110 = HEAP32[$109>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $111 = ((($6)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $112 = $108 ? $111 : $110; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $113 = ((($6)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $114 = HEAP32[$113>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $115 = $106&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $116 = $115 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $117 = $108 ? $116 : $114; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           __THREW__ = 0;
           (invoke_iiii(3462,($11|0),($112|0),($117|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $118 = __THREW__; __THREW__ = 0;
           $119 = $118&1;
           if ($119) {
            $130 = ___cxa_find_matching_catch_2()|0;
            $131 = tempRet0;
            __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 163 "modules/filesystem/physfs/Filesystem.cpp"
            $$1 = $131;$$12 = $130;
            break;
           } else {
            __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 163 "modules/filesystem/physfs/Filesystem.cpp"
            __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7); //@line 163 "modules/filesystem/physfs/Filesystem.cpp"
            break L39;
           }
          }
         } while(0);
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7); //@line 163 "modules/filesystem/physfs/Filesystem.cpp"
         $$2 = $$1;$$23 = $$12;
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($old_save_path); //@line 217 "modules/filesystem/physfs/Filesystem.cpp"
         ___resumeException($$23|0);
         // unreachable;
        }
       } while(0);
       __THREW__ = 0;
       invoke_vii(3967,($8|0),($11|0)); //@line 167 "modules/filesystem/physfs/Filesystem.cpp"
       $146 = __THREW__; __THREW__ = 0;
       $147 = $146&1;
       if ($147) {
        break L4;
       }
       $148 = HEAP8[$11>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $149 = $148 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $150 = ($149<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($150) {
        $154 = ((($11)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP8[$154>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP8[$11>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       } else {
        $151 = ((($this)) + 68|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $152 = HEAP32[$151>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP8[$152>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $153 = ((($this)) + 64|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP32[$153>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       }
       __THREW__ = 0;
       invoke_vii(3368,($11|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $155 = __THREW__; __THREW__ = 0;
       $156 = $155&1;
       if ($156) {
        $157 = ___cxa_find_matching_catch_3(0|0)|0;
        $158 = tempRet0;
        ___clang_call_terminate($157); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       }
       ;HEAP32[$11>>2]=HEAP32[$8>>2]|0;HEAP32[$11+4>>2]=HEAP32[$8+4>>2]|0;HEAP32[$11+8>>2]=HEAP32[$8+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       ;HEAP32[$8>>2]=0|0;HEAP32[$8+4>>2]=0|0;HEAP32[$8+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8); //@line 167 "modules/filesystem/physfs/Filesystem.cpp"
       $159 = HEAP8[$old_save_path>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $160 = $159 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $161 = ($160<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $162 = ((($old_save_path)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $163 = HEAP32[$162>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $164 = $159&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $165 = $164 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $166 = $161 ? $165 : $163; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $167 = ($166|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if (!($167)) {
        $168 = ((($old_save_path)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $169 = HEAP32[$168>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $170 = ((($old_save_path)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $171 = $161 ? $170 : $169; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        __THREW__ = 0;
        (invoke_ii(3968,($171|0))|0); //@line 203 "modules/filesystem/physfs/Filesystem.cpp"
        $172 = __THREW__; __THREW__ = 0;
        $173 = $172&1;
        if ($173) {
         break L4;
        }
       }
       $174 = HEAP8[$11>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $175 = $174 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $176 = ($175<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($176) {
        $179 = ((($11)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $181 = $179;
       } else {
        $177 = ((($this)) + 68|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $178 = HEAP32[$177>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $181 = $178;
       }
       $180 = $appendToPath&1; //@line 209 "modules/filesystem/physfs/Filesystem.cpp"
       __THREW__ = 0;
       (invoke_iiii(3966,($181|0),(0|0),($180|0))|0); //@line 209 "modules/filesystem/physfs/Filesystem.cpp"
       $182 = __THREW__; __THREW__ = 0;
       $183 = $182&1;
       if ($183) {
        break L4;
       }
       __THREW__ = 0;
       (invoke_ii(3965,(0|0))|0); //@line 214 "modules/filesystem/physfs/Filesystem.cpp"
       $184 = __THREW__; __THREW__ = 0;
       $185 = $184&1;
       if ($185) {
        break L4;
       }
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($old_save_path); //@line 217 "modules/filesystem/physfs/Filesystem.cpp"
       $$04 = 1;
       STACKTOP = sp;return ($$04|0); //@line 217 "modules/filesystem/physfs/Filesystem.cpp"
      }
     } while(0);
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 161 "modules/filesystem/physfs/Filesystem.cpp"
     $$2 = $$0;$$23 = $$01;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($old_save_path); //@line 217 "modules/filesystem/physfs/Filesystem.cpp"
     ___resumeException($$23|0);
     // unreachable;
    }
   }
  }
 } while(0);
 $120 = ___cxa_find_matching_catch_2()|0;
 $121 = tempRet0;
 $$2 = $121;$$23 = $120;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($old_save_path); //@line 217 "modules/filesystem/physfs/Filesystem.cpp"
 ___resumeException($$23|0);
 // unreachable;
 return (0)|0;
}
function __ZN12_GLOBAL__N_19normalizeERKNSt3__112basic_stringIcNS0_11char_traitsIcEENS0_9allocatorIcEEEE($agg$result,$input) {
 $agg$result = $agg$result|0;
 $input = $input|0;
 var $$0$i = 0, $$01$i = 0, $$pre = 0, $$pre$phiZ2D = 0, $$pre4 = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $7 = 0, $8 = 0, $9 = 0, $brmerge$demorgan = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $i$02 = 0, $lpad$phi$index = 0, $lpad$phi$index7 = 0, $out = 0, $seenSep$0$off03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 160|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 156|0;
 $1 = sp + 144|0;
 $out = sp;
 $2 = ((($out)) + 64|0); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $3 = ((($out)) + 8|0); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$3>>2] = (9320); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $4 = ((($out)) + 12|0); //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$out>>2] = ((9524)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 HEAP32[$2>>2] = ((9544)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $5 = ((($out)) + 4|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 HEAP32[$5>>2] = 0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $6 = ((($out)) + 64|0); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 __THREW__ = 0;
 invoke_vii(3362,($6|0),($4|0)); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  $$0$i = $20;$$01$i = $19;
  __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($2); //@line 900 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  ___resumeException($$01$i|0);
  // unreachable;
 }
 $9 = ((($out)) + 136|0); //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$9>>2] = 0; //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 $10 = ((($out)) + 140|0); //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$10>>2] = -1; //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$out>>2] = (9300); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$2>>2] = (9340); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$3>>2] = (9320); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 __THREW__ = 0;
 invoke_vi(3363,($4|0)); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 do {
  if ($12) {
   $21 = ___cxa_find_matching_catch_2()|0;
   $22 = tempRet0;
   $eh$lpad$body$i$index2Z2D = $22;$eh$lpad$body$i$indexZ2D = $21;
  } else {
   HEAP32[$4>>2] = (9356); //@line 243 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $13 = ((($out)) + 44|0); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $14 = ((($out)) + 60|0); //@line 242 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   ;HEAP32[$13>>2]=0|0;HEAP32[$13+4>>2]=0|0;HEAP32[$13+8>>2]=0|0;HEAP32[$13+12>>2]=0|0; //@line 241 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   HEAP32[$14>>2] = 24; //@line 242 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   invoke_vii(3364,($4|0),($1|0)); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if ($16) {
    $17 = ___cxa_find_matching_catch_2()|0;
    $18 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13); //@line 245 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($4); //@line 245 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $eh$lpad$body$i$index2Z2D = $18;$eh$lpad$body$i$indexZ2D = $17;
    break;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $23 = HEAP8[$input>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $24 = $23 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $25 = ($24<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $26 = ((($input)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $27 = HEAP32[$26>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $28 = $23&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $29 = $28 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $30 = $25 ? $29 : $27; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $31 = ($30|0)==(0); //@line 85 "modules/filesystem/physfs/Filesystem.cpp"
   L11: do {
    if (!($31)) {
     $32 = ((($input)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $33 = ((($input)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $34 = ((($out)) + 8|0); //@line 89 "modules/filesystem/physfs/Filesystem.cpp"
     $38 = $23;$61 = $27;$i$02 = 0;$seenSep$0$off03 = 0;
     while(1) {
      $37 = $38 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $39 = ($37<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $40 = HEAP32[$32>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $41 = $39 ? $33 : $40; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $42 = (($41) + ($i$02)|0); //@line 3262 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $43 = HEAP8[$42>>0]|0; //@line 89 "modules/filesystem/physfs/Filesystem.cpp"
      $44 = ($43<<24>>24)==(47); //@line 87 "modules/filesystem/physfs/Filesystem.cpp"
      $brmerge$demorgan = $seenSep$0$off03 & $44; //@line 88 "modules/filesystem/physfs/Filesystem.cpp"
      if ($brmerge$demorgan) {
       $53 = $38;$58 = $61;
      } else {
       HEAP8[$0>>0] = $43; //@line 89 "modules/filesystem/physfs/Filesystem.cpp"
       __THREW__ = 0;
       (invoke_iiii(3366,($34|0),($0|0),1)|0); //@line 820 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
       $45 = __THREW__; __THREW__ = 0;
       $46 = $45&1;
       if ($46) {
        break;
       }
       $$pre = HEAP8[$input>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$pre4 = HEAP32[$26>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $53 = $$pre;$58 = $$pre4;
      }
      $51 = (($i$02) + 1)|0; //@line 85 "modules/filesystem/physfs/Filesystem.cpp"
      $52 = $53 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $54 = ($52<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $55 = $53&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $56 = $55 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $57 = $54 ? $56 : $58; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $59 = ($51>>>0)<($57>>>0); //@line 85 "modules/filesystem/physfs/Filesystem.cpp"
      if ($59) {
       $38 = $53;$61 = $58;$i$02 = $51;$seenSep$0$off03 = $44;
      } else {
       break L11;
      }
     }
     $47 = ___cxa_find_matching_catch_2()|0;
     $48 = tempRet0;
     $$pre$phiZ2D = $34;$lpad$phi$index = $47;$lpad$phi$index7 = $48;
     HEAP32[$out>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     HEAP32[$2>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     HEAP32[$$pre$phiZ2D>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     HEAP32[$4>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($4); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($2); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     ___resumeException($lpad$phi$index|0);
     // unreachable;
    }
   } while(0);
   __THREW__ = 0;
   invoke_vii(3367,($agg$result|0),($4|0)); //@line 964 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $35 = __THREW__; __THREW__ = 0;
   $36 = $35&1;
   if (!($36)) {
    HEAP32[$out>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    HEAP32[$2>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    $60 = ((($out)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    HEAP32[$60>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    HEAP32[$4>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($4); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($2); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    STACKTOP = sp;return; //@line 94 "modules/filesystem/physfs/Filesystem.cpp"
   }
   $49 = ___cxa_find_matching_catch_2()|0;
   $50 = tempRet0;
   $$pre5 = ((($out)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   $$pre$phiZ2D = $$pre5;$lpad$phi$index = $49;$lpad$phi$index7 = $50;
   HEAP32[$out>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$2>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$$pre$phiZ2D>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$4>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($4); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($2); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   ___resumeException($lpad$phi$index|0);
   // unreachable;
  }
 } while(0);
 $$0$i = $eh$lpad$body$i$index2Z2D;$$01$i = $eh$lpad$body$i$indexZ2D;
 __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($2); //@line 900 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 ___resumeException($$01$i|0);
 // unreachable;
}
function __ZNK4love10filesystem6physfs10Filesystem11getIdentityEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 36|0); //@line 221 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = $1 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ($2<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($3) {
  $6 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = $6;
  return ($7|0); //@line 221 "modules/filesystem/physfs/Filesystem.cpp"
 } else {
  $4 = ((($this)) + 44|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $5 = HEAP32[$4>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = $5;
  return ($7|0); //@line 221 "modules/filesystem/physfs/Filesystem.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem6physfs10Filesystem9setSourceEPKc($this,$source) {
 $this = $this|0;
 $source = $source|0;
 var $$0 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $new_search_path = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $new_search_path = sp;
 $0 = (_PHYSFS_isInit()|0); //@line 226 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = ($0|0)==(0); //@line 226 "modules/filesystem/physfs/Filesystem.cpp"
 if ($1) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 303 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $2 = ((($this)) + 72|0); //@line 230 "modules/filesystem/physfs/Filesystem.cpp"
 $3 = HEAP8[$2>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = $3 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ($4<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($5) {
  $8 = $3&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = $8 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $11 = $9;
 } else {
  $6 = ((($this)) + 76|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = HEAP32[$6>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $11 = $7;
 }
 $10 = ($11|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if (!($10)) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 303 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $12 = (_strlen($source)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($new_search_path,$source,$12); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = HEAP8[$new_search_path>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = $13 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $15 = ($14<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $16 = ((($new_search_path)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $17 = HEAP32[$16>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $18 = ((($new_search_path)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $19 = $15 ? $18 : $17; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 $20 = (invoke_iiii(3966,($19|0),(0|0),1)|0); //@line 295 "modules/filesystem/physfs/Filesystem.cpp"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $24 = ___cxa_find_matching_catch_2()|0;
  $25 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($new_search_path); //@line 303 "modules/filesystem/physfs/Filesystem.cpp"
  ___resumeException($24|0);
  // unreachable;
 }
 $23 = ($20|0)==(0); //@line 295 "modules/filesystem/physfs/Filesystem.cpp"
 if ($23) {
  $$0 = 0;
 } else {
  __THREW__ = 0;
  (invoke_iii(3578,($2|0),($new_search_path|0))|0); //@line 300 "modules/filesystem/physfs/Filesystem.cpp"
  $26 = __THREW__; __THREW__ = 0;
  $27 = $26&1;
  if ($27) {
   $24 = ___cxa_find_matching_catch_2()|0;
   $25 = tempRet0;
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($new_search_path); //@line 303 "modules/filesystem/physfs/Filesystem.cpp"
   ___resumeException($24|0);
   // unreachable;
  } else {
   $$0 = 1;
  }
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($new_search_path); //@line 303 "modules/filesystem/physfs/Filesystem.cpp"
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0); //@line 303 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZNK4love10filesystem6physfs10Filesystem9getSourceEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 72|0); //@line 307 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = $1 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ($2<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($3) {
  $6 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = $6;
  return ($7|0); //@line 307 "modules/filesystem/physfs/Filesystem.cpp"
 } else {
  $4 = ((($this)) + 80|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $5 = HEAP32[$4>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = $5;
  return ($7|0); //@line 307 "modules/filesystem/physfs/Filesystem.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem6physfs10Filesystem5mountEPKcS4_b($this,$archive,$mountpoint,$appendToPath) {
 $this = $this|0;
 $archive = $archive|0;
 $mountpoint = $mountpoint|0;
 $appendToPath = $appendToPath|0;
 var $$016$i$i$i = 0, $$016$i$i$i$lcssa = 0, $$03 = 0, $$04 = 0, $$1 = 0, $$2 = 0, $$ph = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $__m1$0$i$i$i = 0, $__m2$0$i$i$i = 0, $or$cond = 0, $or$cond12 = 0, $realPath = 0, $sourceBase = 0, $strlenfirst = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $realPath = sp + 12|0;
 $sourceBase = sp;
 $0 = (_PHYSFS_isInit()|0); //@line 368 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = ($0|0)!=(0); //@line 368 "modules/filesystem/physfs/Filesystem.cpp"
 $2 = ($archive|0)!=(0|0); //@line 368 "modules/filesystem/physfs/Filesystem.cpp"
 $or$cond = $2 & $1; //@line 368 "modules/filesystem/physfs/Filesystem.cpp"
 if (!($or$cond)) {
  $$2 = 0;
  STACKTOP = sp;return ($$2|0); //@line 410 "modules/filesystem/physfs/Filesystem.cpp"
 }
 ;HEAP32[$realPath>>2]=0|0;HEAP32[$realPath+4>>2]=0|0;HEAP32[$realPath+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = HEAP32[$this>>2]|0; //@line 372 "modules/filesystem/physfs/Filesystem.cpp"
 $4 = ((($3)) + 100|0); //@line 372 "modules/filesystem/physfs/Filesystem.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 372 "modules/filesystem/physfs/Filesystem.cpp"
 __THREW__ = 0;
 invoke_vii($5|0,($sourceBase|0),($this|0)); //@line 372 "modules/filesystem/physfs/Filesystem.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  $$03 = $22;$$04 = $21;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($realPath); //@line 410 "modules/filesystem/physfs/Filesystem.cpp"
  ___resumeException($$04|0);
  // unreachable;
 }
 $8 = ((($this)) + 100|0); //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = HEAP32[$8>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = ((($this)) + 104|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = HEAP32[$10>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($9|0)==($11|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($12) {
  label = 10;
 } else {
  $13 = $9;
  while(1) {
   $14 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7compareEPKc($13,$archive)|0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $15 = ($14|0)==(0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if ($15) {
    $$ph = $13;
    break;
   }
   $16 = ((($13)) + 12|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $17 = ($16|0)==($11|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($17) {
    $$ph = $11;
    break;
   } else {
    $13 = $16;
   }
  }
  $$pre = HEAP32[$10>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ($$ph|0)==($$pre|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($18) {
   label = 10;
  } else {
   __THREW__ = 0;
   (invoke_iii(3578,($realPath|0),($$ph|0))|0); //@line 378 "modules/filesystem/physfs/Filesystem.cpp"
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   if ($20) {
    label = 9;
   } else {
    label = 32;
   }
  }
 }
 L14: do {
  if ((label|0) == 10) {
   $25 = HEAP32[$this>>2]|0; //@line 379 "modules/filesystem/physfs/Filesystem.cpp"
   $26 = ((($25)) + 32|0); //@line 379 "modules/filesystem/physfs/Filesystem.cpp"
   $27 = HEAP32[$26>>2]|0; //@line 379 "modules/filesystem/physfs/Filesystem.cpp"
   __THREW__ = 0;
   $28 = (invoke_ii($27|0,($this|0))|0); //@line 379 "modules/filesystem/physfs/Filesystem.cpp"
   $29 = __THREW__; __THREW__ = 0;
   $30 = $29&1;
   if ($30) {
    label = 9;
   } else {
    if ($28) {
     $31 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7compareEPKc($sourceBase,$archive)|0); //@line 379 "modules/filesystem/physfs/Filesystem.cpp"
     $32 = ($31|0)==(0); //@line 379 "modules/filesystem/physfs/Filesystem.cpp"
     if ($32) {
      __THREW__ = 0;
      (invoke_iii(3578,($realPath|0),($sourceBase|0))|0); //@line 383 "modules/filesystem/physfs/Filesystem.cpp"
      $33 = __THREW__; __THREW__ = 0;
      $34 = $33&1;
      if ($34) {
       label = 9;
       break;
      } else {
       label = 32;
       break;
      }
     }
    }
    $strlenfirst = HEAP8[$archive>>0]|0; //@line 388 "modules/filesystem/physfs/Filesystem.cpp"
    $35 = ($strlenfirst<<24>>24)==(0); //@line 388 "modules/filesystem/physfs/Filesystem.cpp"
    if ($35) {
     $$1 = 0;
     label = 35;
    } else {
     $36 = (_strstr($archive,519838)|0); //@line 388 "modules/filesystem/physfs/Filesystem.cpp"
     $37 = ($36|0)==(0|0); //@line 388 "modules/filesystem/physfs/Filesystem.cpp"
     if ($37) {
      $38 = (_strcmp($archive,523303)|0); //@line 388 "modules/filesystem/physfs/Filesystem.cpp"
      $39 = ($38|0)==(0); //@line 388 "modules/filesystem/physfs/Filesystem.cpp"
      if ($39) {
       $$1 = 0;
       label = 35;
      } else {
       __THREW__ = 0;
       $40 = (invoke_ii(3969,($archive|0))|0); //@line 391 "modules/filesystem/physfs/Filesystem.cpp"
       $41 = __THREW__; __THREW__ = 0;
       $42 = $41&1;
       if ($42) {
        label = 9;
       } else {
        $43 = ($40|0)==(0|0); //@line 392 "modules/filesystem/physfs/Filesystem.cpp"
        if ($43) {
         $$1 = 0;
         label = 35;
        } else {
         __THREW__ = 0;
         (invoke_iii(3360,($realPath|0),($40|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $44 = __THREW__; __THREW__ = 0;
         $45 = $44&1;
         if ($45) {
          label = 9;
         } else {
          $46 = ((($this)) + 72|0); //@line 399 "modules/filesystem/physfs/Filesystem.cpp"
          $47 = HEAP8[$realPath>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $48 = $47 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $49 = ($48<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $50 = ((($realPath)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $51 = HEAP32[$50>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $52 = ((($realPath)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $53 = HEAP32[$52>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $54 = ((($realPath)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $55 = $47&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $56 = $55 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $57 = $49 ? $54 : $51; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $58 = $49 ? $56 : $53; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $59 = HEAP8[$46>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $60 = $59 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $61 = ($60<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $62 = ((($this)) + 80|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $63 = HEAP32[$62>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $64 = ((($46)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $65 = $61 ? $64 : $63; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $66 = ((($this)) + 76|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $67 = HEAP32[$66>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $68 = $59&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $69 = $68 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $70 = $61 ? $69 : $67; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $71 = ($58>>>0)<($70>>>0); //@line 962 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          L27: do {
           if (!($71)) {
            $72 = ($70|0)==(0); //@line 964 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            if ($72) {
             $$1 = 0;
             label = 35;
             break L14;
            }
            $73 = (($57) + ($58)|0); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $74 = (($65) + ($70)|0); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $75 = ($58|0)<($70|0); //@line 1467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
            if (!($75)) {
             $76 = (1 - ($70))|0; //@line 1469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
             $77 = (($73) + ($76)|0); //@line 1469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
             $78 = ($77|0)==($57|0); //@line 1475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
             if (!($78)) {
              $79 = HEAP8[$65>>0]|0; //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
              $$016$i$i$i = $57;
              L32: while(1) {
               $80 = HEAP8[$$016$i$i$i>>0]|0; //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
               $81 = ($80<<24>>24)==($79<<24>>24); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               if ($81) {
                $__m1$0$i$i$i = $$016$i$i$i;$__m2$0$i$i$i = $65;
                while(1) {
                 $84 = ((($__m2$0$i$i$i)) + 1|0); //@line 1517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                 $85 = ($84|0)==($74|0); //@line 1517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                 if ($85) {
                  $$016$i$i$i$lcssa = $$016$i$i$i;
                  break L32;
                 }
                 $86 = ((($__m1$0$i$i$i)) + 1|0); //@line 1519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                 $87 = HEAP8[$86>>0]|0; //@line 1520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                 $88 = HEAP8[$84>>0]|0; //@line 1520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                 $89 = ($87<<24>>24)==($88<<24>>24); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                 if ($89) {
                  $__m1$0$i$i$i = $86;$__m2$0$i$i$i = $84;
                 } else {
                  break;
                 }
                }
               }
               $82 = ((($$016$i$i$i)) + 1|0); //@line 1479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
               $83 = ($82|0)==($77|0); //@line 1475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
               if ($83) {
                break L27;
               } else {
                $$016$i$i$i = $82;
               }
              }
              $90 = ($$016$i$i$i$lcssa|0)!=($73|0); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              $91 = ($$016$i$i$i$lcssa|0)==($57|0); //@line 399 "modules/filesystem/physfs/Filesystem.cpp"
              $or$cond12 = $90 & $91;
              if ($or$cond12) {
               $$1 = 0;
               label = 35;
               break L14;
              }
             }
            }
           }
          } while(0);
          __THREW__ = 0;
          (invoke_iii(3806,($realPath|0),(523303|0))|0); //@line 1454 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $92 = __THREW__; __THREW__ = 0;
          $93 = $92&1;
          if ($93) {
           label = 9;
          } else {
           __THREW__ = 0;
           (invoke_iii(3806,($realPath|0),($archive|0))|0); //@line 1454 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $94 = __THREW__; __THREW__ = 0;
           $95 = $94&1;
           if ($95) {
            label = 9;
           } else {
            label = 32;
           }
          }
         }
        }
       }
      }
     } else {
      $$1 = 0;
      label = 35;
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 32) {
  $96 = HEAP8[$realPath>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $97 = $96 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $98 = ($97<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $99 = ((($realPath)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $100 = HEAP32[$99>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $101 = $96&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $102 = $101 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $103 = $98 ? $102 : $100; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $104 = ($103|0)==(0); //@line 406 "modules/filesystem/physfs/Filesystem.cpp"
  if ($104) {
   $$1 = 0;
   label = 35;
  } else {
   $105 = ((($realPath)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $106 = HEAP32[$105>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $107 = ((($realPath)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $108 = $98 ? $107 : $106; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $109 = $appendToPath&1; //@line 409 "modules/filesystem/physfs/Filesystem.cpp"
   __THREW__ = 0;
   $110 = (invoke_iiii(3966,($108|0),($mountpoint|0),($109|0))|0); //@line 409 "modules/filesystem/physfs/Filesystem.cpp"
   $111 = __THREW__; __THREW__ = 0;
   $112 = $111&1;
   if ($112) {
    label = 9;
   } else {
    $113 = ($110|0)!=(0); //@line 409 "modules/filesystem/physfs/Filesystem.cpp"
    $$1 = $113;
    label = 35;
   }
  }
 }
 if ((label|0) == 9) {
  $23 = ___cxa_find_matching_catch_2()|0;
  $24 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($sourceBase); //@line 410 "modules/filesystem/physfs/Filesystem.cpp"
  $$03 = $24;$$04 = $23;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($realPath); //@line 410 "modules/filesystem/physfs/Filesystem.cpp"
  ___resumeException($$04|0);
  // unreachable;
 }
 else if ((label|0) == 35) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($sourceBase); //@line 410 "modules/filesystem/physfs/Filesystem.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($realPath); //@line 410 "modules/filesystem/physfs/Filesystem.cpp"
  $$2 = $$1;
  STACKTOP = sp;return ($$2|0); //@line 410 "modules/filesystem/physfs/Filesystem.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem6physfs10Filesystem7unmountEPKc($this,$archive) {
 $this = $this|0;
 $archive = $archive|0;
 var $$02 = 0, $$04 = 0, $$2 = 0, $$3 = 0, $$ph = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $realPath = 0, $sourceBase = 0, $strlenfirst = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $realPath = sp + 12|0;
 $sourceBase = sp;
 $0 = (_PHYSFS_isInit()|0); //@line 414 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = ($0|0)!=(0); //@line 414 "modules/filesystem/physfs/Filesystem.cpp"
 $2 = ($archive|0)!=(0|0); //@line 414 "modules/filesystem/physfs/Filesystem.cpp"
 $or$cond = $2 & $1; //@line 414 "modules/filesystem/physfs/Filesystem.cpp"
 if (!($or$cond)) {
  $$3 = 0;
  STACKTOP = sp;return ($$3|0); //@line 455 "modules/filesystem/physfs/Filesystem.cpp"
 }
 ;HEAP32[$realPath>>2]=0|0;HEAP32[$realPath+4>>2]=0|0;HEAP32[$realPath+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = HEAP32[$this>>2]|0; //@line 418 "modules/filesystem/physfs/Filesystem.cpp"
 $4 = ((($3)) + 100|0); //@line 418 "modules/filesystem/physfs/Filesystem.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 418 "modules/filesystem/physfs/Filesystem.cpp"
 __THREW__ = 0;
 invoke_vii($5|0,($sourceBase|0),($this|0)); //@line 418 "modules/filesystem/physfs/Filesystem.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  $$02 = $21;$$04 = $22;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($realPath); //@line 455 "modules/filesystem/physfs/Filesystem.cpp"
  ___resumeException($$02|0);
  // unreachable;
 }
 $8 = ((($this)) + 100|0); //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = HEAP32[$8>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = ((($this)) + 104|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = HEAP32[$10>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($9|0)==($11|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($12) {
  label = 10;
 } else {
  $13 = $9;
  while(1) {
   $14 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7compareEPKc($13,$archive)|0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $15 = ($14|0)==(0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if ($15) {
    $$ph = $13;
    break;
   }
   $16 = ((($13)) + 12|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $17 = ($16|0)==($11|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($17) {
    $$ph = $11;
    break;
   } else {
    $13 = $16;
   }
  }
  $$pre = HEAP32[$10>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ($$ph|0)==($$pre|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($18) {
   label = 10;
  } else {
   __THREW__ = 0;
   (invoke_iii(3578,($realPath|0),($$ph|0))|0); //@line 424 "modules/filesystem/physfs/Filesystem.cpp"
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   if ($20) {
    label = 9;
   } else {
    label = 22;
   }
  }
 }
 do {
  if ((label|0) == 10) {
   $25 = HEAP32[$this>>2]|0; //@line 425 "modules/filesystem/physfs/Filesystem.cpp"
   $26 = ((($25)) + 32|0); //@line 425 "modules/filesystem/physfs/Filesystem.cpp"
   $27 = HEAP32[$26>>2]|0; //@line 425 "modules/filesystem/physfs/Filesystem.cpp"
   __THREW__ = 0;
   $28 = (invoke_ii($27|0,($this|0))|0); //@line 425 "modules/filesystem/physfs/Filesystem.cpp"
   $29 = __THREW__; __THREW__ = 0;
   $30 = $29&1;
   if ($30) {
    label = 9;
   } else {
    if ($28) {
     $31 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7compareEPKc($sourceBase,$archive)|0); //@line 425 "modules/filesystem/physfs/Filesystem.cpp"
     $32 = ($31|0)==(0); //@line 425 "modules/filesystem/physfs/Filesystem.cpp"
     if ($32) {
      __THREW__ = 0;
      (invoke_iii(3578,($realPath|0),($sourceBase|0))|0); //@line 429 "modules/filesystem/physfs/Filesystem.cpp"
      $33 = __THREW__; __THREW__ = 0;
      $34 = $33&1;
      if ($34) {
       label = 9;
       break;
      } else {
       label = 22;
       break;
      }
     }
    }
    $strlenfirst = HEAP8[$archive>>0]|0; //@line 434 "modules/filesystem/physfs/Filesystem.cpp"
    $35 = ($strlenfirst<<24>>24)==(0); //@line 434 "modules/filesystem/physfs/Filesystem.cpp"
    if ($35) {
     $$2 = 0;
     label = 26;
    } else {
     $36 = (_strstr($archive,519838)|0); //@line 434 "modules/filesystem/physfs/Filesystem.cpp"
     $37 = ($36|0)==(0|0); //@line 434 "modules/filesystem/physfs/Filesystem.cpp"
     if ($37) {
      $38 = (_strcmp($archive,523303)|0); //@line 434 "modules/filesystem/physfs/Filesystem.cpp"
      $39 = ($38|0)==(0); //@line 434 "modules/filesystem/physfs/Filesystem.cpp"
      if ($39) {
       $$2 = 0;
       label = 26;
      } else {
       __THREW__ = 0;
       $40 = (invoke_ii(3969,($archive|0))|0); //@line 437 "modules/filesystem/physfs/Filesystem.cpp"
       $41 = __THREW__; __THREW__ = 0;
       $42 = $41&1;
       if ($42) {
        label = 9;
       } else {
        $43 = ($40|0)==(0|0); //@line 438 "modules/filesystem/physfs/Filesystem.cpp"
        if ($43) {
         $$2 = 0;
         label = 26;
        } else {
         __THREW__ = 0;
         (invoke_iii(3360,($realPath|0),($40|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $44 = __THREW__; __THREW__ = 0;
         $45 = $44&1;
         if ($45) {
          label = 9;
         } else {
          __THREW__ = 0;
          (invoke_iii(3806,($realPath|0),(523303|0))|0); //@line 1454 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $46 = __THREW__; __THREW__ = 0;
          $47 = $46&1;
          if ($47) {
           label = 9;
          } else {
           __THREW__ = 0;
           (invoke_iii(3806,($realPath|0),($archive|0))|0); //@line 1454 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $48 = __THREW__; __THREW__ = 0;
           $49 = $48&1;
           if ($49) {
            label = 9;
           } else {
            label = 22;
           }
          }
         }
        }
       }
      }
     } else {
      $$2 = 0;
      label = 26;
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 22) {
  $50 = HEAP8[$realPath>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $51 = $50 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $52 = ($51<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $53 = ((($realPath)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $54 = HEAP32[$53>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $55 = ((($realPath)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $56 = $52 ? $55 : $54; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  __THREW__ = 0;
  $57 = (invoke_ii(3970,($56|0))|0); //@line 446 "modules/filesystem/physfs/Filesystem.cpp"
  $58 = __THREW__; __THREW__ = 0;
  $59 = $58&1;
  if ($59) {
   label = 9;
  } else {
   $60 = ($57|0)==(0|0); //@line 447 "modules/filesystem/physfs/Filesystem.cpp"
   if ($60) {
    $$2 = 0;
    label = 26;
   } else {
    $61 = HEAP8[$realPath>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $62 = $61 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $63 = ($62<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $64 = HEAP32[$53>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $65 = $63 ? $55 : $64; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    __THREW__ = 0;
    $66 = (invoke_ii(3968,($65|0))|0); //@line 453 "modules/filesystem/physfs/Filesystem.cpp"
    $67 = __THREW__; __THREW__ = 0;
    $68 = $67&1;
    if ($68) {
     label = 9;
    } else {
     $69 = ($66|0)!=(0); //@line 453 "modules/filesystem/physfs/Filesystem.cpp"
     $$2 = $69;
     label = 26;
    }
   }
  }
 }
 if ((label|0) == 9) {
  $23 = ___cxa_find_matching_catch_2()|0;
  $24 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($sourceBase); //@line 455 "modules/filesystem/physfs/Filesystem.cpp"
  $$02 = $23;$$04 = $24;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($realPath); //@line 455 "modules/filesystem/physfs/Filesystem.cpp"
  ___resumeException($$02|0);
  // unreachable;
 }
 else if ((label|0) == 26) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($sourceBase); //@line 455 "modules/filesystem/physfs/Filesystem.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($realPath); //@line 455 "modules/filesystem/physfs/Filesystem.cpp"
  $$3 = $$2;
  STACKTOP = sp;return ($$3|0); //@line 455 "modules/filesystem/physfs/Filesystem.cpp"
 }
 return (0)|0;
}
function __ZNK4love10filesystem6physfs10Filesystem7newFileEPKc($this,$filename) {
 $this = $this|0;
 $filename = $filename|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__Znwj(40)|0); //@line 459 "modules/filesystem/physfs/Filesystem.cpp"
 $2 = (_strlen($filename)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viii(3460,($0|0),($filename|0),($2|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  $$0 = $8;$$01 = $7;
  __ZdlPv($1); //@line 459 "modules/filesystem/physfs/Filesystem.cpp"
  ___resumeException($$01|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_vii(3971,($1|0),($0|0)); //@line 459 "modules/filesystem/physfs/Filesystem.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 459 "modules/filesystem/physfs/Filesystem.cpp"
  STACKTOP = sp;return ($1|0); //@line 459 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $9 = ___cxa_find_matching_catch_2()|0;
 $10 = tempRet0;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 459 "modules/filesystem/physfs/Filesystem.cpp"
 $$0 = $10;$$01 = $9;
 __ZdlPv($1); //@line 459 "modules/filesystem/physfs/Filesystem.cpp"
 ___resumeException($$01|0);
 // unreachable;
 return (0)|0;
}
function __ZNK4love10filesystem6physfs10Filesystem11newFileDataEPvjPKc($this,$data,$size,$filename) {
 $this = $this|0;
 $data = $data|0;
 $size = $size|0;
 $filename = $filename|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__Znwj(48)|0); //@line 464 "modules/filesystem/physfs/Filesystem.cpp"
 $2 = (_strlen($filename)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viii(3460,($0|0),($filename|0),($2|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $11 = ___cxa_find_matching_catch_2()|0;
  $12 = tempRet0;
  $$0 = $12;$$01 = $11;
  __ZdlPv($1); //@line 464 "modules/filesystem/physfs/Filesystem.cpp"
  ___resumeException($$01|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_viiii(3972,($1|0),($size|0),0,($0|0)); //@line 464 "modules/filesystem/physfs/Filesystem.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 464 "modules/filesystem/physfs/Filesystem.cpp"
  $7 = HEAP32[$1>>2]|0; //@line 467 "modules/filesystem/physfs/Filesystem.cpp"
  $8 = ((($7)) + 16|0); //@line 467 "modules/filesystem/physfs/Filesystem.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 467 "modules/filesystem/physfs/Filesystem.cpp"
  $10 = (FUNCTION_TABLE_ii[$9 & 8191]($1)|0); //@line 467 "modules/filesystem/physfs/Filesystem.cpp"
  _memcpy(($10|0),($data|0),($size|0))|0; //@line 467 "modules/filesystem/physfs/Filesystem.cpp"
  STACKTOP = sp;return ($1|0); //@line 469 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $13 = ___cxa_find_matching_catch_2()|0;
 $14 = tempRet0;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 464 "modules/filesystem/physfs/Filesystem.cpp"
 $$0 = $14;$$01 = $13;
 __ZdlPv($1); //@line 464 "modules/filesystem/physfs/Filesystem.cpp"
 ___resumeException($$01|0);
 // unreachable;
 return (0)|0;
}
function __ZNK4love10filesystem6physfs10Filesystem11newFileDataEPKcS4_($this,$b64,$filename) {
 $this = $this|0;
 $b64 = $b64|0;
 $filename = $filename|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $outsize = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $outsize = sp + 12|0;
 $0 = sp;
 $1 = (_strlen($b64)|0); //@line 474 "modules/filesystem/physfs/Filesystem.cpp"
 HEAP32[$outsize>>2] = 0; //@line 475 "modules/filesystem/physfs/Filesystem.cpp"
 $2 = (__ZN4love10b64_decodeEPKciRi($b64,$1,$outsize)|0); //@line 476 "modules/filesystem/physfs/Filesystem.cpp"
 $3 = (__Znwj(48)|0); //@line 477 "modules/filesystem/physfs/Filesystem.cpp"
 $4 = HEAP32[$outsize>>2]|0; //@line 477 "modules/filesystem/physfs/Filesystem.cpp"
 $5 = (_strlen($filename)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viii(3460,($0|0),($filename|0),($5|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $18 = ___cxa_find_matching_catch_2()|0;
  $19 = tempRet0;
  $$0 = $19;$$01 = $18;
  __ZdlPv($3); //@line 477 "modules/filesystem/physfs/Filesystem.cpp"
  ___resumeException($$01|0);
  // unreachable;
 }
 $8 = ($4|0)<(0); //@line 477 "modules/filesystem/physfs/Filesystem.cpp"
 $9 = $8 << 31 >> 31; //@line 477 "modules/filesystem/physfs/Filesystem.cpp"
 __THREW__ = 0;
 invoke_viiii(3972,($3|0),($4|0),($9|0),($0|0)); //@line 477 "modules/filesystem/physfs/Filesystem.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $20 = ___cxa_find_matching_catch_2()|0;
  $21 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 477 "modules/filesystem/physfs/Filesystem.cpp"
  $$0 = $21;$$01 = $20;
  __ZdlPv($3); //@line 477 "modules/filesystem/physfs/Filesystem.cpp"
  ___resumeException($$01|0);
  // unreachable;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 477 "modules/filesystem/physfs/Filesystem.cpp"
 $12 = HEAP32[$3>>2]|0; //@line 480 "modules/filesystem/physfs/Filesystem.cpp"
 $13 = ((($12)) + 16|0); //@line 480 "modules/filesystem/physfs/Filesystem.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 480 "modules/filesystem/physfs/Filesystem.cpp"
 $15 = (FUNCTION_TABLE_ii[$14 & 8191]($3)|0); //@line 480 "modules/filesystem/physfs/Filesystem.cpp"
 $16 = HEAP32[$outsize>>2]|0; //@line 480 "modules/filesystem/physfs/Filesystem.cpp"
 _memcpy(($15|0),($2|0),($16|0))|0; //@line 480 "modules/filesystem/physfs/Filesystem.cpp"
 $17 = ($2|0)==(0|0); //@line 481 "modules/filesystem/physfs/Filesystem.cpp"
 if ($17) {
  STACKTOP = sp;return ($3|0); //@line 483 "modules/filesystem/physfs/Filesystem.cpp"
 }
 __ZdaPv($2); //@line 481 "modules/filesystem/physfs/Filesystem.cpp"
 STACKTOP = sp;return ($3|0); //@line 483 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZN4love10filesystem6physfs10Filesystem19getWorkingDirectoryEv($this) {
 $this = $this|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 488 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = $1 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ($2<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($3) {
  $6 = $1&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = $6 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = $7;
 } else {
  $4 = ((($this)) + 16|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $5 = HEAP32[$4>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = $5;
 }
 $8 = ($9|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($8) {
  $10 = (__Znaj(4096)|0); //@line 497 "modules/filesystem/physfs/Filesystem.cpp"
  $11 = (_getcwd($10,4096)|0); //@line 499 "modules/filesystem/physfs/Filesystem.cpp"
  $12 = ($11|0)==(0|0); //@line 499 "modules/filesystem/physfs/Filesystem.cpp"
  if (!($12)) {
   (__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc($0,$10)|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  }
  __ZdaPv($10); //@line 502 "modules/filesystem/physfs/Filesystem.cpp"
  $$pre = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $14 = $$pre;
 } else {
  $14 = $1;
 }
 $13 = $14 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $15 = ($13<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($15) {
  $18 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $19 = $18;
  return ($19|0); //@line 506 "modules/filesystem/physfs/Filesystem.cpp"
 } else {
  $16 = ((($this)) + 20|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $17 = HEAP32[$16>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $19 = $17;
  return ($19|0); //@line 506 "modules/filesystem/physfs/Filesystem.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem6physfs10Filesystem16getUserDirectoryEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = HEAP8[624064]|0; //@line 515 "modules/filesystem/physfs/Filesystem.cpp"
 $2 = ($1<<24>>24)==(0); //@line 515 "modules/filesystem/physfs/Filesystem.cpp"
 do {
  if ($2) {
   $3 = (___cxa_guard_acquire((624064|0))|0); //@line 515 "modules/filesystem/physfs/Filesystem.cpp"
   $4 = ($3|0)==(0); //@line 515 "modules/filesystem/physfs/Filesystem.cpp"
   if (!($4)) {
    __THREW__ = 0;
    $5 = (invoke_i(3973)|0); //@line 515 "modules/filesystem/physfs/Filesystem.cpp"
    $6 = __THREW__; __THREW__ = 0;
    $7 = $6&1;
    if (!($7)) {
     $8 = (_strlen($5)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     __THREW__ = 0;
     invoke_viii(3460,($0|0),($5|0),($8|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $9 = __THREW__; __THREW__ = 0;
     $10 = $9&1;
     if (!($10)) {
      __THREW__ = 0;
      invoke_vii(3967,(632712|0),($0|0)); //@line 515 "modules/filesystem/physfs/Filesystem.cpp"
      $11 = __THREW__; __THREW__ = 0;
      $12 = $11&1;
      if (!($12)) {
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 515 "modules/filesystem/physfs/Filesystem.cpp"
       (___cxa_atexit((3974|0),(632712|0),(___dso_handle|0))|0); //@line 515 "modules/filesystem/physfs/Filesystem.cpp"
       ___cxa_guard_release((624064|0)); //@line 515 "modules/filesystem/physfs/Filesystem.cpp"
       break;
      }
      $15 = ___cxa_find_matching_catch_2()|0;
      $16 = tempRet0;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 515 "modules/filesystem/physfs/Filesystem.cpp"
      $$0 = $16;$$01 = $15;
      ___cxa_guard_abort((624064|0)); //@line 515 "modules/filesystem/physfs/Filesystem.cpp"
      ___resumeException($$01|0);
      // unreachable;
     }
    }
    $13 = ___cxa_find_matching_catch_2()|0;
    $14 = tempRet0;
    $$0 = $14;$$01 = $13;
    ___cxa_guard_abort((624064|0)); //@line 515 "modules/filesystem/physfs/Filesystem.cpp"
    ___resumeException($$01|0);
    // unreachable;
   }
  }
 } while(0);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_($agg$result,632712); //@line 518 "modules/filesystem/physfs/Filesystem.cpp"
 STACKTOP = sp;return; //@line 518 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZN4love10filesystem6physfs10Filesystem19getAppdataDirectoryEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = ((($this)) + 24|0); //@line 523 "modules/filesystem/physfs/Filesystem.cpp"
 $2 = HEAP8[$1>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = $2 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ($3<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($4) {
  $7 = $2&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $8 = $7 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $10 = $8;
 } else {
  $5 = ((($this)) + 28|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $6 = HEAP32[$5>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $10 = $6;
 }
 $9 = ($10|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if (!($9)) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_($agg$result,$1); //@line 545 "modules/filesystem/physfs/Filesystem.cpp"
  STACKTOP = sp;return; //@line 545 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $11 = HEAP32[$this>>2]|0; //@line 542 "modules/filesystem/physfs/Filesystem.cpp"
 $12 = ((($11)) + 88|0); //@line 542 "modules/filesystem/physfs/Filesystem.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 542 "modules/filesystem/physfs/Filesystem.cpp"
 FUNCTION_TABLE_vii[$13 & 8191]($0,$this); //@line 542 "modules/filesystem/physfs/Filesystem.cpp"
 $14 = HEAP8[$1>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $15 = $14 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $16 = ($15<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($16) {
  $20 = ((($1)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$20>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$1>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 } else {
  $17 = ((($this)) + 32|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $18 = HEAP32[$17>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$18>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $19 = ((($this)) + 28|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$19>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 __THREW__ = 0;
 invoke_vii(3368,($1|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $23 = ___cxa_find_matching_catch_3(0|0)|0;
  $24 = tempRet0;
  ___clang_call_terminate($23); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 ;HEAP32[$1>>2]=HEAP32[$0>>2]|0;HEAP32[$1+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$1+8>>2]=HEAP32[$0+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 542 "modules/filesystem/physfs/Filesystem.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_($agg$result,$1); //@line 545 "modules/filesystem/physfs/Filesystem.cpp"
 STACKTOP = sp;return; //@line 545 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZN4love10filesystem6physfs10Filesystem16getSaveDirectoryEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 60|0); //@line 551 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = $1 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ($2<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($3) {
  $6 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = $6;
  return ($7|0); //@line 551 "modules/filesystem/physfs/Filesystem.cpp"
 } else {
  $4 = ((($this)) + 68|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $5 = HEAP32[$4>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = $5;
  return ($7|0); //@line 551 "modules/filesystem/physfs/Filesystem.cpp"
 }
 return (0)|0;
}
function __ZNK4love10filesystem6physfs10Filesystem22getSourceBaseDirectoryEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $base_end_pos$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 72|0); //@line 556 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = $1 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ($2<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($3) {
  $6 = $1&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = $6 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = $7;
 } else {
  $4 = ((($this)) + 76|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $5 = HEAP32[$4>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = $5;
 }
 $8 = ($9|0)==(0); //@line 558 "modules/filesystem/physfs/Filesystem.cpp"
 if ($8) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($agg$result,1282466,0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  return; //@line 578 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $10 = (($9) + -2)|0; //@line 567 "modules/filesystem/physfs/Filesystem.cpp"
 $11 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE5rfindEcj($0,47,$10)|0); //@line 3544 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 switch ($11|0) {
 case -1:  {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($agg$result,1282466,0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  return; //@line 578 "modules/filesystem/physfs/Filesystem.cpp"
  break;
 }
 case 0:  {
  $base_end_pos$0 = 1;
  break;
 }
 default: {
  $base_end_pos$0 = $11;
 }
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_jjRKS4_($agg$result,$0,0,$base_end_pos$0,$0); //@line 3345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 return; //@line 578 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZNK4love10filesystem6physfs10Filesystem16getRealDirectoryEPKc($agg$result,$this,$filename) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $filename = $filename|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_PHYSFS_isInit()|0); //@line 582 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = ($0|0)==(0); //@line 582 "modules/filesystem/physfs/Filesystem.cpp"
 if ($1) {
  $2 = (___cxa_allocate_exception(16)|0); //@line 583 "modules/filesystem/physfs/Filesystem.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($2|0),(507115|0),($vararg_buffer|0)); //@line 583 "modules/filesystem/physfs/Filesystem.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if (!($4)) {
   ___cxa_throw(($2|0),(96|0),(344|0)); //@line 583 "modules/filesystem/physfs/Filesystem.cpp"
   // unreachable; //@line 583 "modules/filesystem/physfs/Filesystem.cpp"
  }
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  ___cxa_free_exception(($2|0)); //@line 583 "modules/filesystem/physfs/Filesystem.cpp"
  $$0 = $5;$$01 = $6;
  ___resumeException($$0|0);
  // unreachable;
 }
 $7 = (_PHYSFS_getRealDir($filename)|0); //@line 585 "modules/filesystem/physfs/Filesystem.cpp"
 $8 = ($7|0)==(0|0); //@line 587 "modules/filesystem/physfs/Filesystem.cpp"
 if (!($8)) {
  $14 = (_strlen($7)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($agg$result,$7,$14); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  STACKTOP = sp;return; //@line 591 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $9 = (___cxa_allocate_exception(16)|0); //@line 588 "modules/filesystem/physfs/Filesystem.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($9|0),(366108|0),($vararg_buffer1|0)); //@line 588 "modules/filesystem/physfs/Filesystem.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if (!($11)) {
  ___cxa_throw(($9|0),(96|0),(344|0)); //@line 588 "modules/filesystem/physfs/Filesystem.cpp"
  // unreachable; //@line 588 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $12 = ___cxa_find_matching_catch_2()|0;
 $13 = tempRet0;
 ___cxa_free_exception(($9|0)); //@line 588 "modules/filesystem/physfs/Filesystem.cpp"
 $$0 = $12;$$01 = $13;
 ___resumeException($$0|0);
 // unreachable;
}
function __ZNK4love10filesystem6physfs10Filesystem6existsEPKc($this,$path) {
 $this = $this|0;
 $path = $path|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_PHYSFS_isInit()|0); //@line 595 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = ($0|0)==(0); //@line 595 "modules/filesystem/physfs/Filesystem.cpp"
 if ($1) {
  $$0 = 0;
  return ($$0|0); //@line 599 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $2 = (_PHYSFS_exists($path)|0); //@line 598 "modules/filesystem/physfs/Filesystem.cpp"
 $3 = ($2|0)!=(0); //@line 598 "modules/filesystem/physfs/Filesystem.cpp"
 $$0 = $3;
 return ($$0|0); //@line 599 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZNK4love10filesystem6physfs10Filesystem11isDirectoryEPKc($this,$dir) {
 $this = $this|0;
 $dir = $dir|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_PHYSFS_isInit()|0); //@line 603 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = ($0|0)==(0); //@line 603 "modules/filesystem/physfs/Filesystem.cpp"
 if ($1) {
  $$0 = 0;
  return ($$0|0); //@line 615 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $2 = (_PHYSFS_isDirectory($dir)|0); //@line 613 "modules/filesystem/physfs/Filesystem.cpp"
 $3 = ($2|0)!=(0); //@line 613 "modules/filesystem/physfs/Filesystem.cpp"
 $$0 = $3;
 return ($$0|0); //@line 615 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZNK4love10filesystem6physfs10Filesystem6isFileEPKc($this,$file) {
 $this = $this|0;
 $file = $file|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_PHYSFS_isInit()|0); //@line 619 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = ($0|0)==(0); //@line 619 "modules/filesystem/physfs/Filesystem.cpp"
 if ($1) {
  $$0 = 0;
  return ($$0|0); //@line 623 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $2 = (_PHYSFS_exists($file)|0); //@line 622 "modules/filesystem/physfs/Filesystem.cpp"
 $3 = ($2|0)==(0); //@line 622 "modules/filesystem/physfs/Filesystem.cpp"
 if ($3) {
  $$0 = 0;
  return ($$0|0); //@line 623 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $4 = HEAP32[$this>>2]|0; //@line 622 "modules/filesystem/physfs/Filesystem.cpp"
 $5 = ((($4)) + 112|0); //@line 622 "modules/filesystem/physfs/Filesystem.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 622 "modules/filesystem/physfs/Filesystem.cpp"
 $7 = (FUNCTION_TABLE_iii[$6 & 8191]($this,$file)|0); //@line 622 "modules/filesystem/physfs/Filesystem.cpp"
 $8 = $7 ^ 1; //@line 622 "modules/filesystem/physfs/Filesystem.cpp"
 $$0 = $8;
 return ($$0|0); //@line 623 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZNK4love10filesystem6physfs10Filesystem9isSymlinkEPKc($this,$filename) {
 $this = $this|0;
 $filename = $filename|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_PHYSFS_isInit()|0); //@line 627 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = ($0|0)==(0); //@line 627 "modules/filesystem/physfs/Filesystem.cpp"
 if ($1) {
  $$0 = 0;
  return ($$0|0); //@line 639 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $2 = (_PHYSFS_isSymbolicLink($filename)|0); //@line 637 "modules/filesystem/physfs/Filesystem.cpp"
 $3 = ($2|0)!=(0); //@line 637 "modules/filesystem/physfs/Filesystem.cpp"
 $$0 = $3;
 return ($$0|0); //@line 639 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZN4love10filesystem6physfs10Filesystem15createDirectoryEPKc($this,$dir) {
 $this = $this|0;
 $dir = $dir|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_PHYSFS_isInit()|0); //@line 643 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = ($0|0)==(0); //@line 643 "modules/filesystem/physfs/Filesystem.cpp"
 if ($1) {
  $$0 = 0;
  return ($$0|0); //@line 653 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $2 = (_PHYSFS_getWriteDir()|0); //@line 646 "modules/filesystem/physfs/Filesystem.cpp"
 $3 = ($2|0)==(0|0); //@line 646 "modules/filesystem/physfs/Filesystem.cpp"
 if ($3) {
  $4 = HEAP32[$this>>2]|0; //@line 646 "modules/filesystem/physfs/Filesystem.cpp"
  $5 = ((($4)) + 36|0); //@line 646 "modules/filesystem/physfs/Filesystem.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 646 "modules/filesystem/physfs/Filesystem.cpp"
  $7 = (FUNCTION_TABLE_ii[$6 & 8191]($this)|0); //@line 646 "modules/filesystem/physfs/Filesystem.cpp"
  if (!($7)) {
   $$0 = 0;
   return ($$0|0); //@line 653 "modules/filesystem/physfs/Filesystem.cpp"
  }
 }
 $8 = (_PHYSFS_mkdir($dir)|0); //@line 649 "modules/filesystem/physfs/Filesystem.cpp"
 $not$ = ($8|0)!=(0); //@line 650 "modules/filesystem/physfs/Filesystem.cpp"
 $$0 = $not$;
 return ($$0|0); //@line 653 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZN4love10filesystem6physfs10Filesystem6removeEPKc($this,$file) {
 $this = $this|0;
 $file = $file|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_PHYSFS_isInit()|0); //@line 657 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = ($0|0)==(0); //@line 657 "modules/filesystem/physfs/Filesystem.cpp"
 if ($1) {
  $$0 = 0;
  return ($$0|0); //@line 667 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $2 = (_PHYSFS_getWriteDir()|0); //@line 660 "modules/filesystem/physfs/Filesystem.cpp"
 $3 = ($2|0)==(0|0); //@line 660 "modules/filesystem/physfs/Filesystem.cpp"
 if ($3) {
  $4 = HEAP32[$this>>2]|0; //@line 660 "modules/filesystem/physfs/Filesystem.cpp"
  $5 = ((($4)) + 36|0); //@line 660 "modules/filesystem/physfs/Filesystem.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 660 "modules/filesystem/physfs/Filesystem.cpp"
  $7 = (FUNCTION_TABLE_ii[$6 & 8191]($this)|0); //@line 660 "modules/filesystem/physfs/Filesystem.cpp"
  if (!($7)) {
   $$0 = 0;
   return ($$0|0); //@line 667 "modules/filesystem/physfs/Filesystem.cpp"
  }
 }
 $8 = (_PHYSFS_delete($file)|0); //@line 663 "modules/filesystem/physfs/Filesystem.cpp"
 $not$ = ($8|0)!=(0); //@line 664 "modules/filesystem/physfs/Filesystem.cpp"
 $$0 = $not$;
 return ($$0|0); //@line 667 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZNK4love10filesystem6physfs10Filesystem4readEPKcx($this,$filename,$0,$1) {
 $this = $this|0;
 $filename = $filename|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$01 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $file = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $file = sp;
 $2 = sp + 40|0;
 $3 = (_strlen($filename)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($2,$filename,$3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_vii(3971,($file|0),($2|0)); //@line 671 "modules/filesystem/physfs/Filesystem.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $14 = ___cxa_find_matching_catch_2()|0;
  $15 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 671 "modules/filesystem/physfs/Filesystem.cpp"
  $$0 = $15;$$01 = $14;
  ___resumeException($$01|0);
  // unreachable;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 671 "modules/filesystem/physfs/Filesystem.cpp"
 __THREW__ = 0;
 (invoke_iii(2047,($file|0),1)|0); //@line 673 "modules/filesystem/physfs/Filesystem.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  $8 = HEAP32[$file>>2]|0; //@line 676 "modules/filesystem/physfs/Filesystem.cpp"
  $9 = ((($8)) + 32|0); //@line 676 "modules/filesystem/physfs/Filesystem.cpp"
  $10 = HEAP32[$9>>2]|0; //@line 676 "modules/filesystem/physfs/Filesystem.cpp"
  __THREW__ = 0;
  $11 = (invoke_iiii($10|0,($file|0),($0|0),($1|0))|0); //@line 676 "modules/filesystem/physfs/Filesystem.cpp"
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   __ZN4love10filesystem6physfs4FileD2Ev($file); //@line 677 "modules/filesystem/physfs/Filesystem.cpp"
   STACKTOP = sp;return ($11|0); //@line 677 "modules/filesystem/physfs/Filesystem.cpp"
  }
 }
 $16 = ___cxa_find_matching_catch_2()|0;
 $17 = tempRet0;
 __ZN4love10filesystem6physfs4FileD2Ev($file); //@line 677 "modules/filesystem/physfs/Filesystem.cpp"
 $$0 = $17;$$01 = $16;
 ___resumeException($$01|0);
 // unreachable;
 return (0)|0;
}
function __ZNK4love10filesystem6physfs10Filesystem5writeEPKcPKvx($this,$filename,$data,$0,$1) {
 $this = $this|0;
 $filename = $filename|0;
 $data = $data|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $file = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp + 40|0;
 $file = sp;
 $2 = sp + 44|0;
 $3 = (_strlen($filename)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($2,$filename,$3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_vii(3971,($file|0),($2|0)); //@line 681 "modules/filesystem/physfs/Filesystem.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 681 "modules/filesystem/physfs/Filesystem.cpp"
  $$1 = $16;$$12 = $15;
  ___resumeException($$12|0);
  // unreachable;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 681 "modules/filesystem/physfs/Filesystem.cpp"
 __THREW__ = 0;
 (invoke_iii(2047,($file|0),2)|0); //@line 683 "modules/filesystem/physfs/Filesystem.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 do {
  if ($7) {
   label = 8;
  } else {
   __THREW__ = 0;
   $8 = (invoke_iiiii(2052,($file|0),($data|0),($0|0),($1|0))|0); //@line 686 "modules/filesystem/physfs/Filesystem.cpp"
   $9 = __THREW__; __THREW__ = 0;
   $10 = $9&1;
   if ($10) {
    label = 8;
   } else {
    if ($8) {
     __ZN4love10filesystem6physfs4FileD2Ev($file); //@line 688 "modules/filesystem/physfs/Filesystem.cpp"
     STACKTOP = sp;return; //@line 688 "modules/filesystem/physfs/Filesystem.cpp"
    }
    $11 = (___cxa_allocate_exception(16)|0); //@line 687 "modules/filesystem/physfs/Filesystem.cpp"
    __THREW__ = 0;
    invoke_viii(3349,($11|0),(366129|0),($vararg_buffer|0)); //@line 687 "modules/filesystem/physfs/Filesystem.cpp"
    $12 = __THREW__; __THREW__ = 0;
    $13 = $12&1;
    if ($13) {
     $19 = ___cxa_find_matching_catch_2()|0;
     $20 = tempRet0;
     ___cxa_free_exception(($11|0)); //@line 687 "modules/filesystem/physfs/Filesystem.cpp"
     $$0 = $20;$$01 = $19;
     break;
    } else {
     __THREW__ = 0;
     invoke_viii(3350,($11|0),(96|0),(344|0)); //@line 687 "modules/filesystem/physfs/Filesystem.cpp"
     $14 = __THREW__; __THREW__ = 0;
     label = 8;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  $$0 = $18;$$01 = $17;
 }
 __ZN4love10filesystem6physfs4FileD2Ev($file); //@line 688 "modules/filesystem/physfs/Filesystem.cpp"
 $$1 = $$0;$$12 = $$01;
 ___resumeException($$12|0);
 // unreachable;
}
function __ZNK4love10filesystem6physfs10Filesystem6appendEPKcPKvx($this,$filename,$data,$0,$1) {
 $this = $this|0;
 $filename = $filename|0;
 $data = $data|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $file = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp + 40|0;
 $file = sp;
 $2 = sp + 44|0;
 $3 = (_strlen($filename)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($2,$filename,$3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_vii(3971,($file|0),($2|0)); //@line 692 "modules/filesystem/physfs/Filesystem.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 692 "modules/filesystem/physfs/Filesystem.cpp"
  $$1 = $16;$$12 = $15;
  ___resumeException($$12|0);
  // unreachable;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 692 "modules/filesystem/physfs/Filesystem.cpp"
 __THREW__ = 0;
 (invoke_iii(2047,($file|0),3)|0); //@line 694 "modules/filesystem/physfs/Filesystem.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 do {
  if ($7) {
   label = 8;
  } else {
   __THREW__ = 0;
   $8 = (invoke_iiiii(2052,($file|0),($data|0),($0|0),($1|0))|0); //@line 697 "modules/filesystem/physfs/Filesystem.cpp"
   $9 = __THREW__; __THREW__ = 0;
   $10 = $9&1;
   if ($10) {
    label = 8;
   } else {
    if ($8) {
     __ZN4love10filesystem6physfs4FileD2Ev($file); //@line 699 "modules/filesystem/physfs/Filesystem.cpp"
     STACKTOP = sp;return; //@line 699 "modules/filesystem/physfs/Filesystem.cpp"
    }
    $11 = (___cxa_allocate_exception(16)|0); //@line 698 "modules/filesystem/physfs/Filesystem.cpp"
    __THREW__ = 0;
    invoke_viii(3349,($11|0),(366129|0),($vararg_buffer|0)); //@line 698 "modules/filesystem/physfs/Filesystem.cpp"
    $12 = __THREW__; __THREW__ = 0;
    $13 = $12&1;
    if ($13) {
     $19 = ___cxa_find_matching_catch_2()|0;
     $20 = tempRet0;
     ___cxa_free_exception(($11|0)); //@line 698 "modules/filesystem/physfs/Filesystem.cpp"
     $$0 = $20;$$01 = $19;
     break;
    } else {
     __THREW__ = 0;
     invoke_viii(3350,($11|0),(96|0),(344|0)); //@line 698 "modules/filesystem/physfs/Filesystem.cpp"
     $14 = __THREW__; __THREW__ = 0;
     label = 8;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  $$0 = $18;$$01 = $17;
 }
 __ZN4love10filesystem6physfs4FileD2Ev($file); //@line 699 "modules/filesystem/physfs/Filesystem.cpp"
 $$1 = $$0;$$12 = $$01;
 ___resumeException($$12|0);
 // unreachable;
}
function __ZN4love10filesystem6physfs10Filesystem17getDirectoryItemsEPKcRNSt3__16vectorINS5_12basic_stringIcNS5_11char_traitsIcEENS5_9allocatorIcEEEENSA_ISC_EEEE($this,$dir,$items) {
 $this = $this|0;
 $dir = $dir|0;
 $items = $items|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (_PHYSFS_isInit()|0); //@line 703 "modules/filesystem/physfs/Filesystem.cpp"
 $2 = ($1|0)==(0); //@line 703 "modules/filesystem/physfs/Filesystem.cpp"
 if ($2) {
  STACKTOP = sp;return; //@line 715 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $3 = (_PHYSFS_enumerateFiles($dir)|0); //@line 706 "modules/filesystem/physfs/Filesystem.cpp"
 $4 = ($3|0)==(0|0); //@line 708 "modules/filesystem/physfs/Filesystem.cpp"
 if ($4) {
  STACKTOP = sp;return; //@line 715 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $5 = HEAP32[$3>>2]|0; //@line 711 "modules/filesystem/physfs/Filesystem.cpp"
 $6 = ($5|0)==(0|0); //@line 711 "modules/filesystem/physfs/Filesystem.cpp"
 L7: do {
  if (!($6)) {
   $7 = ((($items)) + 4|0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $8 = ((($items)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $9 = $5;$i$01 = $3;
   while(1) {
    $10 = (_strlen($9)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($0,$9,$10); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $11 = HEAP32[$7>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $12 = HEAP32[$8>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $13 = ($11>>>0)<($12>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($13) {
     ;HEAP32[$11>>2]=HEAP32[$0>>2]|0;HEAP32[$11+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$11+8>>2]=HEAP32[$0+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $14 = HEAP32[$7>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $15 = ((($14)) + 12|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$7>>2] = $15; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    } else {
     __THREW__ = 0;
     invoke_vii(3472,($items|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $16 = __THREW__; __THREW__ = 0;
     $17 = $16&1;
     if ($17) {
      break;
     }
    }
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 712 "modules/filesystem/physfs/Filesystem.cpp"
    $18 = ((($i$01)) + 4|0); //@line 711 "modules/filesystem/physfs/Filesystem.cpp"
    $19 = HEAP32[$18>>2]|0; //@line 711 "modules/filesystem/physfs/Filesystem.cpp"
    $20 = ($19|0)==(0|0); //@line 711 "modules/filesystem/physfs/Filesystem.cpp"
    if ($20) {
     break L7;
    } else {
     $9 = $19;$i$01 = $18;
    }
   }
   $21 = ___cxa_find_matching_catch_2()|0;
   $22 = tempRet0;
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 712 "modules/filesystem/physfs/Filesystem.cpp"
   ___resumeException($21|0);
   // unreachable;
  }
 } while(0);
 _PHYSFS_freeList($3); //@line 714 "modules/filesystem/physfs/Filesystem.cpp"
 STACKTOP = sp;return; //@line 715 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZNK4love10filesystem6physfs10Filesystem15getLastModifiedEPKc($this,$filename) {
 $this = $this|0;
 $filename = $filename|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_PHYSFS_isInit()|0); //@line 721 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = ($0|0)==(0); //@line 721 "modules/filesystem/physfs/Filesystem.cpp"
 if ($1) {
  $12 = -1;$13 = -1;
  tempRet0 = ($12); //@line 736 "modules/filesystem/physfs/Filesystem.cpp"
  STACKTOP = sp;return ($13|0); //@line 736 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $2 = (_PHYSFS_getLastModTime($filename)|0); //@line 729 "modules/filesystem/physfs/Filesystem.cpp"
 $3 = tempRet0; //@line 729 "modules/filesystem/physfs/Filesystem.cpp"
 $4 = ($2|0)==(-1); //@line 732 "modules/filesystem/physfs/Filesystem.cpp"
 $5 = ($3|0)==(-1); //@line 732 "modules/filesystem/physfs/Filesystem.cpp"
 $6 = $4 & $5; //@line 732 "modules/filesystem/physfs/Filesystem.cpp"
 if (!($6)) {
  $12 = $3;$13 = $2;
  tempRet0 = ($12); //@line 736 "modules/filesystem/physfs/Filesystem.cpp"
  STACKTOP = sp;return ($13|0); //@line 736 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $7 = (___cxa_allocate_exception(16)|0); //@line 733 "modules/filesystem/physfs/Filesystem.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($7|0),(366156|0),($vararg_buffer|0)); //@line 733 "modules/filesystem/physfs/Filesystem.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  ___cxa_free_exception(($7|0)); //@line 733 "modules/filesystem/physfs/Filesystem.cpp"
  ___resumeException($10|0);
  // unreachable;
 } else {
  ___cxa_throw(($7|0),(96|0),(344|0)); //@line 733 "modules/filesystem/physfs/Filesystem.cpp"
  // unreachable; //@line 733 "modules/filesystem/physfs/Filesystem.cpp"
 }
 return (0)|0;
}
function __ZNK4love10filesystem6physfs10Filesystem7getSizeEPKc($this,$filename) {
 $this = $this|0;
 $filename = $filename|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $file = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $file = sp;
 $0 = sp + 40|0;
 $1 = (_strlen($filename)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($0,$filename,$1); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_vii(3971,($file|0),($0|0)); //@line 740 "modules/filesystem/physfs/Filesystem.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $8 = ___cxa_find_matching_catch_2()|0;
  $9 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 740 "modules/filesystem/physfs/Filesystem.cpp"
  $$0 = $8;$$01 = $9;
  ___resumeException($$0|0);
  // unreachable;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 740 "modules/filesystem/physfs/Filesystem.cpp"
 __THREW__ = 0;
 $4 = (invoke_ii(2050,($file|0))|0); //@line 741 "modules/filesystem/physfs/Filesystem.cpp"
 $5 = tempRet0; //@line 741 "modules/filesystem/physfs/Filesystem.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  __ZN4love10filesystem6physfs4FileD2Ev($file); //@line 743 "modules/filesystem/physfs/Filesystem.cpp"
  tempRet0 = ($5); //@line 743 "modules/filesystem/physfs/Filesystem.cpp"
  STACKTOP = sp;return ($4|0); //@line 743 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $10 = ___cxa_find_matching_catch_2()|0;
 $11 = tempRet0;
 __ZN4love10filesystem6physfs4FileD2Ev($file); //@line 743 "modules/filesystem/physfs/Filesystem.cpp"
 $$0 = $10;$$01 = $11;
 ___resumeException($$0|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love10filesystem6physfs10Filesystem18setSymlinksEnabledEb($this,$enable) {
 $this = $this|0;
 $enable = $enable|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $version = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $version = sp;
 $0 = (_PHYSFS_isInit()|0); //@line 747 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = ($0|0)==(0); //@line 747 "modules/filesystem/physfs/Filesystem.cpp"
 if ($1) {
  STACKTOP = sp;return; //@line 762 "modules/filesystem/physfs/Filesystem.cpp"
 }
 if (!($enable)) {
  ;HEAP8[$version>>0]=0|0;HEAP8[$version+1>>0]=0|0;HEAP8[$version+2>>0]=0|0; //@line 752 "modules/filesystem/physfs/Filesystem.cpp"
  _PHYSFS_getLinkedVersion($version); //@line 753 "modules/filesystem/physfs/Filesystem.cpp"
  $2 = HEAP8[$version>>0]|0; //@line 757 "modules/filesystem/physfs/Filesystem.cpp"
  $3 = ($2<<24>>24)==(2); //@line 757 "modules/filesystem/physfs/Filesystem.cpp"
  if ($3) {
   $4 = ((($version)) + 1|0); //@line 757 "modules/filesystem/physfs/Filesystem.cpp"
   $5 = HEAP8[$4>>0]|0; //@line 757 "modules/filesystem/physfs/Filesystem.cpp"
   $6 = ($5<<24>>24)==(1); //@line 757 "modules/filesystem/physfs/Filesystem.cpp"
   if ($6) {
    STACKTOP = sp;return; //@line 762 "modules/filesystem/physfs/Filesystem.cpp"
   }
  }
 }
 $7 = $enable&1; //@line 761 "modules/filesystem/physfs/Filesystem.cpp"
 _PHYSFS_permitSymbolicLinks($7); //@line 761 "modules/filesystem/physfs/Filesystem.cpp"
 STACKTOP = sp;return; //@line 762 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZNK4love10filesystem6physfs10Filesystem18areSymlinksEnabledEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_PHYSFS_isInit()|0); //@line 766 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = ($0|0)==(0); //@line 766 "modules/filesystem/physfs/Filesystem.cpp"
 if ($1) {
  $$0 = 0;
  return ($$0|0); //@line 770 "modules/filesystem/physfs/Filesystem.cpp"
 }
 $2 = (_PHYSFS_symbolicLinksPermitted()|0); //@line 769 "modules/filesystem/physfs/Filesystem.cpp"
 $3 = ($2|0)!=(0); //@line 769 "modules/filesystem/physfs/Filesystem.cpp"
 $$0 = $3;
 return ($$0|0); //@line 770 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZN4love10filesystem6physfs10Filesystem14getRequirePathEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 88|0); //@line 774 "modules/filesystem/physfs/Filesystem.cpp"
 return ($0|0); //@line 774 "modules/filesystem/physfs/Filesystem.cpp"
}
function __ZN4love10filesystem6physfs10Filesystem20allowMountingForPathERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($this,$path) {
 $this = $this|0;
 $path = $path|0;
 var $$mask$i = 0, $$pre$i = 0, $$pre8$i = 0, $$pre9$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__lhs_sz$03$i$i10 = 0, $__lp$04$i$i9 = 0, $__rp$05$i$i8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 100|0); //@line 779 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($this)) + 104|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($1|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 L1: do {
  if ($4) {
   $44 = $1;
   label = 12;
  } else {
   $5 = ((($path)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $6 = ((($path)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $7 = ((($path)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$pre$i = HEAP8[$path>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$pre8$i = HEAP32[$5>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $8 = $$pre$i & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $9 = ($8<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $10 = $$pre$i&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $11 = $10 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $12 = $9 ? $11 : $$pre8$i; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $13 = HEAP32[$6>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $14 = $9 ? $7 : $13; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $15 = ($12|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $17 = $1;
   while(1) {
    $16 = HEAP16[$17>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $18 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $19 = ($18<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $20 = ((($17)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $21 = HEAP32[$20>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $$mask$i = $16 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $22 = $$mask$i&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $23 = $22 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $24 = $19 ? $23 : $21; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $25 = ($24|0)==($12|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $26 = ($16&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $27 = $26&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    do {
     if ($25) {
      if (!($19)) {
       $37 = ((($17)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $38 = HEAP32[$37>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $39 = (_memcmp($38,$14,$12)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $40 = ($39|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($40) {
        $44 = $17;
        label = 12;
        break L1;
       } else {
        break;
       }
      }
      if ($15) {
       $44 = $17;
       label = 12;
       break L1;
      }
      $28 = HEAP8[$14>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $29 = ($27<<24>>24)==($28<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($29) {
       $30 = ((($17)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $__lhs_sz$03$i$i10 = $23;$__lp$04$i$i9 = $30;$__rp$05$i$i8 = $14;
       while(1) {
        $31 = (($__lhs_sz$03$i$i10) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $32 = ((($__lp$04$i$i9)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $33 = ($31|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if ($33) {
         $44 = $17;
         label = 12;
         break L1;
        }
        $34 = ((($__rp$05$i$i8)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $$pre9$i = HEAP8[$32>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $35 = HEAP8[$34>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $36 = ($$pre9$i<<24>>24)==($35<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if ($36) {
         $__lhs_sz$03$i$i10 = $31;$__lp$04$i$i9 = $32;$__rp$05$i$i8 = $34;
        } else {
         break;
        }
       }
      }
     }
    } while(0);
    $41 = ((($17)) + 12|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $42 = ($41|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($42) {
     break;
    } else {
     $17 = $41;
    }
   }
  }
 } while(0);
 if ((label|0) == 12) {
  $43 = ($44|0)==($3|0); //@line 1295 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if (!($43)) {
   return; //@line 781 "modules/filesystem/physfs/Filesystem.cpp"
  }
 }
 $45 = ((($this)) + 108|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $46 = HEAP32[$45>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $47 = ($3|0)==($46|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($47) {
  __ZNSt3__16vectorINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEENS4_IS6_EEE21__push_back_slow_pathIRKS6_EEvOT_($0,$path); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  return; //@line 781 "modules/filesystem/physfs/Filesystem.cpp"
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_($3,$path); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $48 = HEAP32[$2>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $49 = ((($48)) + 12|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$2>>2] = $49; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  return; //@line 781 "modules/filesystem/physfs/Filesystem.cpp"
 }
}
function __ZN4love10filesystem6physfs10FilesystemC2Ev($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $$pre$i = 0, $$pre$i7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 __ZN4love10filesystem10FilesystemC2Ev($this); //@line 105 "modules/filesystem/physfs/Filesystem.cpp"
 HEAP32[$this>>2] = (23184); //@line 108 "modules/filesystem/physfs/Filesystem.cpp"
 $1 = ((($this)) + 12|0); //@line 105 "modules/filesystem/physfs/Filesystem.cpp"
 $2 = ((($this)) + 24|0); //@line 105 "modules/filesystem/physfs/Filesystem.cpp"
 $3 = ((($this)) + 36|0); //@line 105 "modules/filesystem/physfs/Filesystem.cpp"
 $4 = ((($this)) + 48|0); //@line 105 "modules/filesystem/physfs/Filesystem.cpp"
 $5 = ((($this)) + 60|0); //@line 105 "modules/filesystem/physfs/Filesystem.cpp"
 $6 = ((($this)) + 72|0); //@line 105 "modules/filesystem/physfs/Filesystem.cpp"
 $7 = ((($this)) + 88|0); //@line 105 "modules/filesystem/physfs/Filesystem.cpp"
 $8 = ((($this)) + 92|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ((($this)) + 100|0); //@line 105 "modules/filesystem/physfs/Filesystem.cpp"
 $10 = ((($this)) + 104|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 dest=$1; stop=dest+72|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));HEAP16[$1+72>>1]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;HEAP32[$7+12>>2]=0|0;HEAP32[$7+16>>2]=0|0;HEAP32[$7+20>>2]=0|0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 __THREW__ = 0;
 invoke_viii(3460,($0|0),(366200|0),5); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 do {
  if ($12) {
   $$02 = $0;
   label = 5;
  } else {
   $13 = ((($0)) + 12|0); //@line 109 "modules/filesystem/physfs/Filesystem.cpp"
   __THREW__ = 0;
   invoke_viii(3460,($13|0),(366206|0),10); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $14 = __THREW__; __THREW__ = 0;
   $15 = $14&1;
   if ($15) {
    $$02 = $13;
    label = 5;
   } else {
    $16 = ((($0)) + 24|0); //@line 93 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/initializer_list"
    __THREW__ = 0;
    invoke_viii(3975,($7|0),($0|0),($16|0)); //@line 584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if ($18) {
     $26 = ___cxa_find_matching_catch_2()|0;
     $27 = tempRet0;
     $28 = ((($0)) + 12|0); //@line 109 "modules/filesystem/physfs/Filesystem.cpp"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($28); //@line 109 "modules/filesystem/physfs/Filesystem.cpp"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 109 "modules/filesystem/physfs/Filesystem.cpp"
     $$0 = $27;$$01 = $26;
     break;
    } else {
     $19 = ((($0)) + 12|0); //@line 109 "modules/filesystem/physfs/Filesystem.cpp"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($19); //@line 109 "modules/filesystem/physfs/Filesystem.cpp"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 109 "modules/filesystem/physfs/Filesystem.cpp"
     STACKTOP = sp;return; //@line 110 "modules/filesystem/physfs/Filesystem.cpp"
    }
   }
  }
 } while(0);
 if ((label|0) == 5) {
  $20 = ___cxa_find_matching_catch_2()|0;
  $21 = tempRet0;
  $22 = ($0|0)==($$02|0); //@line 109 "modules/filesystem/physfs/Filesystem.cpp"
  if ($22) {
   $$0 = $21;$$01 = $20;
  } else {
   $24 = $$02;
   while(1) {
    $23 = ((($24)) + -12|0); //@line 109 "modules/filesystem/physfs/Filesystem.cpp"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($23); //@line 109 "modules/filesystem/physfs/Filesystem.cpp"
    $25 = ($23|0)==($0|0); //@line 109 "modules/filesystem/physfs/Filesystem.cpp"
    if ($25) {
     $$0 = $21;$$01 = $20;
     break;
    } else {
     $24 = $23;
    }
   }
  }
 }
 $29 = HEAP32[$9>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = ($29|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($30)) {
  $31 = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $32 = ($31|0)==($29|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($32) {
   $37 = $29;
  } else {
   $34 = $31;
   while(1) {
    $33 = ((($34)) + -12|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$10>>2] = $33; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($33); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $35 = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $36 = ($35|0)==($29|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($36) {
     break;
    } else {
     $34 = $35;
    }
   }
   $$pre$i7 = HEAP32[$9>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $37 = $$pre$i7;
  }
  __ZdlPv($37); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $38 = HEAP32[$7>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $39 = ($38|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($39) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 110 "modules/filesystem/physfs/Filesystem.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5); //@line 110 "modules/filesystem/physfs/Filesystem.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 110 "modules/filesystem/physfs/Filesystem.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 110 "modules/filesystem/physfs/Filesystem.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 110 "modules/filesystem/physfs/Filesystem.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 110 "modules/filesystem/physfs/Filesystem.cpp"
  __ZN4love10filesystem10FilesystemD2Ev($this); //@line 110 "modules/filesystem/physfs/Filesystem.cpp"
  ___resumeException($$01|0);
  // unreachable;
 }
 $40 = HEAP32[$8>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $41 = ($40|0)==($38|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($41) {
  $46 = $38;
 } else {
  $43 = $40;
  while(1) {
   $42 = ((($43)) + -12|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$8>>2] = $42; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($42); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $44 = HEAP32[$8>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $45 = ($44|0)==($38|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($45) {
    break;
   } else {
    $43 = $44;
   }
  }
  $$pre$i = HEAP32[$7>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $46 = $$pre$i;
 }
 __ZdlPv($46); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 110 "modules/filesystem/physfs/Filesystem.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5); //@line 110 "modules/filesystem/physfs/Filesystem.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 110 "modules/filesystem/physfs/Filesystem.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 110 "modules/filesystem/physfs/Filesystem.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 110 "modules/filesystem/physfs/Filesystem.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 110 "modules/filesystem/physfs/Filesystem.cpp"
 __ZN4love10filesystem10FilesystemD2Ev($this); //@line 110 "modules/filesystem/physfs/Filesystem.cpp"
 ___resumeException($$01|0);
 // unreachable;
}
function __ZNSt3__16vectorINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEENS4_IS6_EEE6assignIPKS6_EENS_9enable_ifIXaasr21__is_forward_iteratorIT_EE5valuesr16is_constructibleIS6_NS_15iterator_traitsISD_E9referenceEEE5valueEvE4typeESD_SD_($this,$__first,$__last) {
 $this = $this|0;
 $__first = $__first|0;
 $__last = $__last|0;
 var $$0$i7 = 0, $$0$lcssa$i$i = 0, $$01$i = 0, $$01$i4 = 0, $$012$i$i = 0, $$03$i$i = 0, $$__last = 0, $$mux = 0, $$not = 0, $$not18 = 0, $$pre = 0, $$pre$i1 = 0, $$pre$i2 = 0, $$pre15 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0, $9 = 0, $brmerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $__first;
 $1 = $__last; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $2 = (($1) - ($0))|0; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $3 = (($2|0) / 12)&-1; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $4 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = HEAP32[$4>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = (($5) - ($6))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = (($7|0) / 12)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ($3>>>0)>($8>>>0); //@line 1398 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $6; //@line 1408 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($9)) {
  $11 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $12 = HEAP32[$11>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($6))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = ($3>>>0)>($14>>>0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = (($__first) + (($14*12)|0)|0); //@line 467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $$__last = $15 ? $16 : $__last; //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = ($$__last|0)==($__first|0); //@line 1748 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if ($17) {
   $$0$lcssa$i$i = $10;
  } else {
   $$012$i$i = $__first;$$03$i$i = $10;
   while(1) {
    (__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEaSERKS5_($$03$i$i,$$012$i$i)|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $18 = ((($$012$i$i)) + 12|0); //@line 1748 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $19 = ((($$03$i$i)) + 12|0); //@line 1748 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $20 = ($18|0)==($$__last|0); //@line 1748 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    if ($20) {
     $$0$lcssa$i$i = $19;
     break;
    } else {
     $$012$i$i = $18;$$03$i$i = $19;
    }
   }
  }
  if (!($15)) {
   $27 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $28 = ($27|0)==($$0$lcssa$i$i|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($28) {
    return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   } else {
    $30 = $27;
   }
   while(1) {
    $29 = ((($30)) + -12|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$11>>2] = $29; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($29); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $31 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $32 = ($31|0)==($$0$lcssa$i$i|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($32) {
     break;
    } else {
     $30 = $31;
    }
   }
   return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  $21 = ($$__last|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($21) {
   return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  $$pre$i2 = HEAP32[$11>>2]|0; //@line 1030 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$01$i4 = $16;$22 = $$pre$i2;
  while(1) {
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_($22,$$01$i4); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $23 = HEAP32[$11>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $24 = ((($23)) + 12|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$11>>2] = $24; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $25 = ((($$01$i4)) + 12|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $26 = ($25|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($26) {
    break;
   } else {
    $$01$i4 = $25;$22 = $24;
   }
  }
  return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $33 = ($6|0)==(0); //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($33) {
  $60 = $5;
 } else {
  $34 = $6; //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $35 = ((($this)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $36 = HEAP32[$35>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $37 = ($36|0)==($10|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($37) {
   $42 = $34;
  } else {
   $39 = $36;
   while(1) {
    $38 = ((($39)) + -12|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$35>>2] = $38; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($38); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $40 = HEAP32[$35>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $41 = ($40|0)==($10|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($41) {
     break;
    } else {
     $39 = $40;
    }
   }
   $$pre$i1 = HEAP32[$this>>2]|0; //@line 948 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $42 = $$pre$i1;
  }
  __ZdlPv($42); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  HEAP32[$4>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$35>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$this>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $60 = 0;
 }
 $43 = ($3>>>0)>(357913941); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($43) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$4>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre15 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $45 = $$pre;$46 = $$pre15;
 } else {
  $45 = $60;$46 = 0;
 }
 $44 = (($45) - ($46))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $47 = (($44|0) / 12)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $48 = ($47>>>0)<(178956970); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($48) {
  $49 = $47 << 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$not = ($49>>>0)>=($3>>>0);
  $$not18 = $43 ^ 1;
  $brmerge = $$not | $$not18;
  $$mux = $$not ? $49 : $3;
  if ($brmerge) {
   $$0$i7 = $$mux;
  } else {
   __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$0$i7 = $3;
  }
 } else {
  $$0$i7 = 357913941;
 }
 $50 = ($$0$i7*12)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $51 = (__Znwj($50)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $52 = ((($this)) + 4|0); //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$52>>2] = $51; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$this>>2] = $51; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $53 = (($51) + (($$0$i7*12)|0)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$4>>2] = $53; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $54 = ($__first|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($54) {
  return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $$01$i = $__first;$55 = $51;
 while(1) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_($55,$$01$i); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $56 = HEAP32[$52>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $57 = ((($56)) + 12|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$52>>2] = $57; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $58 = ((($$01$i)) + 12|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $59 = ($58|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($59) {
   break;
  } else {
   $$01$i = $58;$55 = $57;
  }
 }
 return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love10filesystem11DroppedFileD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (23372); //@line 50 "modules/filesystem/DroppedFile.cpp"
 $0 = ((($this)) + 24|0); //@line 51 "modules/filesystem/DroppedFile.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 51 "modules/filesystem/DroppedFile.cpp"
 $2 = ($1|0)==(0); //@line 51 "modules/filesystem/DroppedFile.cpp"
 if (!($2)) {
  $3 = ((($this)) + 20|0); //@line 91 "modules/filesystem/DroppedFile.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 91 "modules/filesystem/DroppedFile.cpp"
  $5 = ($4|0)==(0|0); //@line 91 "modules/filesystem/DroppedFile.cpp"
  if (!($5)) {
   $6 = (_fclose($4)|0); //@line 91 "modules/filesystem/DroppedFile.cpp"
   $7 = ($6|0)==(0); //@line 91 "modules/filesystem/DroppedFile.cpp"
   if ($7) {
    HEAP32[$0>>2] = 0; //@line 94 "modules/filesystem/DroppedFile.cpp"
    HEAP32[$3>>2] = 0; //@line 95 "modules/filesystem/DroppedFile.cpp"
   }
  }
 }
 $8 = ((($this)) + 8|0); //@line 53 "modules/filesystem/DroppedFile.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8); //@line 53 "modules/filesystem/DroppedFile.cpp"
 __ZN4love10filesystem4FileD2Ev($this); //@line 53 "modules/filesystem/DroppedFile.cpp"
 return; //@line 53 "modules/filesystem/DroppedFile.cpp"
}
function __ZN4love10filesystem11DroppedFileD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (23372); //@line 50 "modules/filesystem/DroppedFile.cpp"
 $0 = ((($this)) + 24|0); //@line 51 "modules/filesystem/DroppedFile.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 51 "modules/filesystem/DroppedFile.cpp"
 $2 = ($1|0)==(0); //@line 51 "modules/filesystem/DroppedFile.cpp"
 if (!($2)) {
  $3 = ((($this)) + 20|0); //@line 91 "modules/filesystem/DroppedFile.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 91 "modules/filesystem/DroppedFile.cpp"
  $5 = ($4|0)==(0|0); //@line 91 "modules/filesystem/DroppedFile.cpp"
  if (!($5)) {
   $6 = (_fclose($4)|0); //@line 91 "modules/filesystem/DroppedFile.cpp"
   $7 = ($6|0)==(0); //@line 91 "modules/filesystem/DroppedFile.cpp"
   if ($7) {
    HEAP32[$0>>2] = 0; //@line 94 "modules/filesystem/DroppedFile.cpp"
    HEAP32[$3>>2] = 0; //@line 95 "modules/filesystem/DroppedFile.cpp"
   }
  }
 }
 $8 = ((($this)) + 8|0); //@line 53 "modules/filesystem/DroppedFile.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8); //@line 53 "modules/filesystem/DroppedFile.cpp"
 __ZN4love10filesystem4FileD2Ev($this); //@line 53 "modules/filesystem/DroppedFile.cpp"
 __ZdlPv($this); //@line 50 "modules/filesystem/DroppedFile.cpp"
 return; //@line 53 "modules/filesystem/DroppedFile.cpp"
}
function __ZN4love10filesystem11DroppedFile4openENS0_4File4ModeE($this,$newmode) {
 $this = $this|0;
 $newmode = $newmode|0;
 var $$0 = 0, $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($newmode|0)==(0); //@line 57 "modules/filesystem/DroppedFile.cpp"
 if ($0) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 87 "modules/filesystem/DroppedFile.cpp"
 }
 $1 = ((($this)) + 20|0); //@line 61 "modules/filesystem/DroppedFile.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 61 "modules/filesystem/DroppedFile.cpp"
 $3 = ($2|0)==(0|0); //@line 61 "modules/filesystem/DroppedFile.cpp"
 if (!($3)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 87 "modules/filesystem/DroppedFile.cpp"
 }
 $4 = ((($this)) + 8|0); //@line 71 "modules/filesystem/DroppedFile.cpp"
 $5 = HEAP8[$4>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = $5 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ($6<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($7) {
  $10 = ((($4)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $11 = $10;
 } else {
  $8 = ((($this)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = HEAP32[$8>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $11 = $9;
 }
 switch ($newmode|0) {
 case 3:  {
  $$0$i = 366250;
  break;
 }
 case 1:  {
  $$0$i = 530591;
  break;
 }
 case 2:  {
  $$0$i = 608916;
  break;
 }
 default: {
  $$0$i = 442507;
 }
 }
 $12 = (_fopen($11,$$0$i)|0); //@line 79 "modules/filesystem/DroppedFile.cpp"
 HEAP32[$1>>2] = $12; //@line 71 "modules/filesystem/DroppedFile.cpp"
 $13 = ($newmode|0)==(1); //@line 74 "modules/filesystem/DroppedFile.cpp"
 $14 = ($12|0)==(0|0); //@line 74 "modules/filesystem/DroppedFile.cpp"
 $or$cond = $13 & $14; //@line 74 "modules/filesystem/DroppedFile.cpp"
 if (!($or$cond)) {
  $27 = ((($this)) + 24|0); //@line 77 "modules/filesystem/DroppedFile.cpp"
  HEAP32[$27>>2] = $newmode; //@line 77 "modules/filesystem/DroppedFile.cpp"
  if (!($14)) {
   $28 = HEAP32[$this>>2]|0; //@line 79 "modules/filesystem/DroppedFile.cpp"
   $29 = ((($28)) + 64|0); //@line 79 "modules/filesystem/DroppedFile.cpp"
   $30 = HEAP32[$29>>2]|0; //@line 79 "modules/filesystem/DroppedFile.cpp"
   $31 = ((($this)) + 28|0); //@line 79 "modules/filesystem/DroppedFile.cpp"
   $32 = HEAP32[$31>>2]|0; //@line 79 "modules/filesystem/DroppedFile.cpp"
   $33 = ((($this)) + 32|0); //@line 79 "modules/filesystem/DroppedFile.cpp"
   $34 = $33; //@line 79 "modules/filesystem/DroppedFile.cpp"
   $35 = $34; //@line 79 "modules/filesystem/DroppedFile.cpp"
   $36 = HEAP32[$35>>2]|0; //@line 79 "modules/filesystem/DroppedFile.cpp"
   $37 = (($34) + 4)|0; //@line 79 "modules/filesystem/DroppedFile.cpp"
   $38 = $37; //@line 79 "modules/filesystem/DroppedFile.cpp"
   $39 = HEAP32[$38>>2]|0; //@line 79 "modules/filesystem/DroppedFile.cpp"
   $40 = (FUNCTION_TABLE_iiiii[$30 & 8191]($this,$32,$36,$39)|0); //@line 79 "modules/filesystem/DroppedFile.cpp"
   if (!($40)) {
    HEAP32[$31>>2] = 0; //@line 82 "modules/filesystem/DroppedFile.cpp"
    $41 = $33; //@line 83 "modules/filesystem/DroppedFile.cpp"
    $42 = $41; //@line 83 "modules/filesystem/DroppedFile.cpp"
    HEAP32[$42>>2] = 0; //@line 83 "modules/filesystem/DroppedFile.cpp"
    $43 = (($41) + 4)|0; //@line 83 "modules/filesystem/DroppedFile.cpp"
    $44 = $43; //@line 83 "modules/filesystem/DroppedFile.cpp"
    HEAP32[$44>>2] = 0; //@line 83 "modules/filesystem/DroppedFile.cpp"
   }
  }
  $45 = HEAP32[$1>>2]|0; //@line 86 "modules/filesystem/DroppedFile.cpp"
  $46 = ($45|0)!=(0|0); //@line 86 "modules/filesystem/DroppedFile.cpp"
  $$0 = $46;
  STACKTOP = sp;return ($$0|0); //@line 87 "modules/filesystem/DroppedFile.cpp"
 }
 $15 = (___cxa_allocate_exception(16)|0); //@line 75 "modules/filesystem/DroppedFile.cpp"
 $16 = HEAP8[$4>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $17 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $18 = ($17<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($18) {
  $21 = ((($4)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $22 = $21;
 } else {
  $19 = ((($this)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $20 = HEAP32[$19>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $22 = $20;
 }
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $22; //@line 75 "modules/filesystem/DroppedFile.cpp"
 invoke_viii(3349,($15|0),(507142|0),($vararg_buffer|0)); //@line 75 "modules/filesystem/DroppedFile.cpp"
 $23 = __THREW__; __THREW__ = 0;
 $24 = $23&1;
 if ($24) {
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  ___cxa_free_exception(($15|0)); //@line 75 "modules/filesystem/DroppedFile.cpp"
  ___resumeException($25|0);
  // unreachable;
 } else {
  ___cxa_throw(($15|0),(96|0),(344|0)); //@line 75 "modules/filesystem/DroppedFile.cpp"
  // unreachable; //@line 75 "modules/filesystem/DroppedFile.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem11DroppedFile5closeEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 91 "modules/filesystem/DroppedFile.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 91 "modules/filesystem/DroppedFile.cpp"
 $2 = ($1|0)==(0|0); //@line 91 "modules/filesystem/DroppedFile.cpp"
 if ($2) {
  $$0 = 0;
 } else {
  $3 = (_fclose($1)|0); //@line 91 "modules/filesystem/DroppedFile.cpp"
  $4 = ($3|0)==(0); //@line 91 "modules/filesystem/DroppedFile.cpp"
  if ($4) {
   $5 = ((($this)) + 24|0); //@line 94 "modules/filesystem/DroppedFile.cpp"
   HEAP32[$5>>2] = 0; //@line 94 "modules/filesystem/DroppedFile.cpp"
   HEAP32[$0>>2] = 0; //@line 95 "modules/filesystem/DroppedFile.cpp"
   $$0 = 1;
  } else {
   $$0 = 0;
  }
 }
 return ($$0|0); //@line 98 "modules/filesystem/DroppedFile.cpp"
}
function __ZNK4love10filesystem11DroppedFile6isOpenEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 102 "modules/filesystem/DroppedFile.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 102 "modules/filesystem/DroppedFile.cpp"
 $2 = ($1|0)==(0); //@line 102 "modules/filesystem/DroppedFile.cpp"
 if ($2) {
  $6 = 0;
  return ($6|0); //@line 102 "modules/filesystem/DroppedFile.cpp"
 }
 $3 = ((($this)) + 20|0); //@line 102 "modules/filesystem/DroppedFile.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 102 "modules/filesystem/DroppedFile.cpp"
 $5 = ($4|0)!=(0|0); //@line 102 "modules/filesystem/DroppedFile.cpp"
 $6 = $5;
 return ($6|0); //@line 102 "modules/filesystem/DroppedFile.cpp"
}
function __ZN4love10filesystem11DroppedFile7getSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buf = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $buf = sp;
 $0 = ((($this)) + 8|0); //@line 122 "modules/filesystem/DroppedFile.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = $1 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ($2<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($3) {
  $6 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = $6;
 } else {
  $4 = ((($this)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $5 = HEAP32[$4>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = $5;
 }
 $8 = (_stat($7,$buf)|0); //@line 122 "modules/filesystem/DroppedFile.cpp"
 $9 = ($8|0)==(0); //@line 122 "modules/filesystem/DroppedFile.cpp"
 $10 = ((($buf)) + 36|0); //@line 125 "modules/filesystem/DroppedFile.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 125 "modules/filesystem/DroppedFile.cpp"
 $12 = ($11|0)<(0); //@line 125 "modules/filesystem/DroppedFile.cpp"
 $13 = $12 << 31 >> 31; //@line 125 "modules/filesystem/DroppedFile.cpp"
 $14 = $9 ? $11 : -1; //@line 122 "modules/filesystem/DroppedFile.cpp"
 $15 = $9 ? $13 : -1; //@line 122 "modules/filesystem/DroppedFile.cpp"
 tempRet0 = ($15); //@line 128 "modules/filesystem/DroppedFile.cpp"
 STACKTOP = sp;return ($14|0); //@line 128 "modules/filesystem/DroppedFile.cpp"
}
function __ZN4love10filesystem11DroppedFile4readEPvx($this,$dst,$0,$1) {
 $this = $this|0;
 $dst = $dst|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$01 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $2 = ((($this)) + 20|0); //@line 132 "modules/filesystem/DroppedFile.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 132 "modules/filesystem/DroppedFile.cpp"
 $4 = ($3|0)==(0|0); //@line 132 "modules/filesystem/DroppedFile.cpp"
 if (!($4)) {
  $5 = ((($this)) + 24|0); //@line 132 "modules/filesystem/DroppedFile.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 132 "modules/filesystem/DroppedFile.cpp"
  $7 = ($6|0)==(1); //@line 132 "modules/filesystem/DroppedFile.cpp"
  if ($7) {
   $13 = ($1|0)<(0); //@line 135 "modules/filesystem/DroppedFile.cpp"
   if (!($13)) {
    $19 = (_fread($dst,1,$0,$3)|0); //@line 138 "modules/filesystem/DroppedFile.cpp"
    tempRet0 = (0); //@line 140 "modules/filesystem/DroppedFile.cpp"
    STACKTOP = sp;return ($19|0); //@line 140 "modules/filesystem/DroppedFile.cpp"
   }
   $14 = (___cxa_allocate_exception(16)|0); //@line 136 "modules/filesystem/DroppedFile.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($14|0),(507273|0),($vararg_buffer1|0)); //@line 136 "modules/filesystem/DroppedFile.cpp"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if (!($16)) {
    ___cxa_throw(($14|0),(96|0),(344|0)); //@line 136 "modules/filesystem/DroppedFile.cpp"
    // unreachable; //@line 136 "modules/filesystem/DroppedFile.cpp"
   }
   $17 = ___cxa_find_matching_catch_2()|0;
   $18 = tempRet0;
   ___cxa_free_exception(($14|0)); //@line 136 "modules/filesystem/DroppedFile.cpp"
   $$0 = $17;$$01 = $18;
   ___resumeException($$0|0);
   // unreachable;
  }
 }
 $8 = (___cxa_allocate_exception(16)|0); //@line 133 "modules/filesystem/DroppedFile.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($8|0),(507241|0),($vararg_buffer|0)); //@line 133 "modules/filesystem/DroppedFile.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if (!($10)) {
  ___cxa_throw(($8|0),(96|0),(344|0)); //@line 133 "modules/filesystem/DroppedFile.cpp"
  // unreachable; //@line 133 "modules/filesystem/DroppedFile.cpp"
 }
 $11 = ___cxa_find_matching_catch_2()|0;
 $12 = tempRet0;
 ___cxa_free_exception(($8|0)); //@line 133 "modules/filesystem/DroppedFile.cpp"
 $$0 = $11;$$01 = $12;
 ___resumeException($$0|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love10filesystem11DroppedFile5writeEPKvx($this,$data,$0,$1) {
 $this = $this|0;
 $data = $data|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$01 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $switch = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $2 = ((($this)) + 20|0); //@line 145 "modules/filesystem/DroppedFile.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 145 "modules/filesystem/DroppedFile.cpp"
 $4 = ($3|0)==(0|0); //@line 145 "modules/filesystem/DroppedFile.cpp"
 if (!($4)) {
  $5 = ((($this)) + 24|0); //@line 145 "modules/filesystem/DroppedFile.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 145 "modules/filesystem/DroppedFile.cpp"
  $7 = $6 & -2; //@line 145 "modules/filesystem/DroppedFile.cpp"
  $switch = ($7|0)==(2); //@line 145 "modules/filesystem/DroppedFile.cpp"
  if ($switch) {
   $13 = ($1|0)<(0); //@line 148 "modules/filesystem/DroppedFile.cpp"
   if (!($13)) {
    $19 = (_fwrite($data,1,$0,$3)|0); //@line 151 "modules/filesystem/DroppedFile.cpp"
    $20 = ($19|0)==($0|0); //@line 153 "modules/filesystem/DroppedFile.cpp"
    $21 = (0)==($1|0); //@line 153 "modules/filesystem/DroppedFile.cpp"
    $22 = $20 & $21; //@line 153 "modules/filesystem/DroppedFile.cpp"
    STACKTOP = sp;return ($22|0); //@line 153 "modules/filesystem/DroppedFile.cpp"
   }
   $14 = (___cxa_allocate_exception(16)|0); //@line 149 "modules/filesystem/DroppedFile.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($14|0),(507324|0),($vararg_buffer1|0)); //@line 149 "modules/filesystem/DroppedFile.cpp"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if (!($16)) {
    ___cxa_throw(($14|0),(96|0),(344|0)); //@line 149 "modules/filesystem/DroppedFile.cpp"
    // unreachable; //@line 149 "modules/filesystem/DroppedFile.cpp"
   }
   $17 = ___cxa_find_matching_catch_2()|0;
   $18 = tempRet0;
   ___cxa_free_exception(($14|0)); //@line 149 "modules/filesystem/DroppedFile.cpp"
   $$0 = $17;$$01 = $18;
   ___resumeException($$0|0);
   // unreachable;
  }
 }
 $8 = (___cxa_allocate_exception(16)|0); //@line 146 "modules/filesystem/DroppedFile.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($8|0),(507292|0),($vararg_buffer|0)); //@line 146 "modules/filesystem/DroppedFile.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if (!($10)) {
  ___cxa_throw(($8|0),(96|0),(344|0)); //@line 146 "modules/filesystem/DroppedFile.cpp"
  // unreachable; //@line 146 "modules/filesystem/DroppedFile.cpp"
 }
 $11 = ___cxa_find_matching_catch_2()|0;
 $12 = tempRet0;
 ___cxa_free_exception(($8|0)); //@line 146 "modules/filesystem/DroppedFile.cpp"
 $$0 = $11;$$01 = $12;
 ___resumeException($$0|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love10filesystem11DroppedFile5flushEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $switch = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 20|0); //@line 158 "modules/filesystem/DroppedFile.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 158 "modules/filesystem/DroppedFile.cpp"
 $2 = ($1|0)==(0|0); //@line 158 "modules/filesystem/DroppedFile.cpp"
 if (!($2)) {
  $3 = ((($this)) + 24|0); //@line 158 "modules/filesystem/DroppedFile.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 158 "modules/filesystem/DroppedFile.cpp"
  $5 = $4 & -2; //@line 158 "modules/filesystem/DroppedFile.cpp"
  $switch = ($5|0)==(2); //@line 158 "modules/filesystem/DroppedFile.cpp"
  if ($switch) {
   $11 = (_fflush($1)|0); //@line 161 "modules/filesystem/DroppedFile.cpp"
   $12 = ($11|0)==(0); //@line 161 "modules/filesystem/DroppedFile.cpp"
   STACKTOP = sp;return ($12|0); //@line 161 "modules/filesystem/DroppedFile.cpp"
  }
 }
 $6 = (___cxa_allocate_exception(16)|0); //@line 159 "modules/filesystem/DroppedFile.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($6|0),(507292|0),($vararg_buffer|0)); //@line 159 "modules/filesystem/DroppedFile.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  ___cxa_free_exception(($6|0)); //@line 159 "modules/filesystem/DroppedFile.cpp"
  ___resumeException($9|0);
  // unreachable;
 } else {
  ___cxa_throw(($6|0),(96|0),(344|0)); //@line 159 "modules/filesystem/DroppedFile.cpp"
  // unreachable; //@line 159 "modules/filesystem/DroppedFile.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem11DroppedFile5isEOFEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 166 "modules/filesystem/DroppedFile.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 166 "modules/filesystem/DroppedFile.cpp"
 $2 = ($1|0)==(0|0); //@line 166 "modules/filesystem/DroppedFile.cpp"
 if ($2) {
  $5 = 1;
  return ($5|0); //@line 166 "modules/filesystem/DroppedFile.cpp"
 }
 $3 = (_feof($1)|0); //@line 166 "modules/filesystem/DroppedFile.cpp"
 $4 = ($3|0)!=(0); //@line 166 "modules/filesystem/DroppedFile.cpp"
 $5 = $4;
 return ($5|0); //@line 166 "modules/filesystem/DroppedFile.cpp"
}
function __ZN4love10filesystem11DroppedFile4tellEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 171 "modules/filesystem/DroppedFile.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 171 "modules/filesystem/DroppedFile.cpp"
 $2 = ($1|0)==(0|0); //@line 171 "modules/filesystem/DroppedFile.cpp"
 if ($2) {
  $6 = -1;$7 = -1;
 } else {
  $3 = (_ftell($1)|0); //@line 174 "modules/filesystem/DroppedFile.cpp"
  $4 = ($3|0)<(0); //@line 174 "modules/filesystem/DroppedFile.cpp"
  $5 = $4 << 31 >> 31; //@line 174 "modules/filesystem/DroppedFile.cpp"
  $6 = $5;$7 = $3;
 }
 tempRet0 = ($6); //@line 175 "modules/filesystem/DroppedFile.cpp"
 return ($7|0); //@line 175 "modules/filesystem/DroppedFile.cpp"
}
function __ZN4love10filesystem11DroppedFile4seekEy($this,$0,$1) {
 $this = $this|0;
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($this)) + 20|0); //@line 179 "modules/filesystem/DroppedFile.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 179 "modules/filesystem/DroppedFile.cpp"
 $4 = ($3|0)==(0|0); //@line 179 "modules/filesystem/DroppedFile.cpp"
 if ($4) {
  $7 = 0;
  return ($7|0); //@line 179 "modules/filesystem/DroppedFile.cpp"
 }
 $5 = (_fseek($3,$0,0)|0); //@line 179 "modules/filesystem/DroppedFile.cpp"
 $6 = ($5|0)==(0); //@line 179 "modules/filesystem/DroppedFile.cpp"
 $7 = $6;
 return ($7|0); //@line 179 "modules/filesystem/DroppedFile.cpp"
}
function __ZN4love10filesystem11DroppedFile9setBufferENS0_4File10BufferModeEx($this,$bufmode,$0,$1) {
 $this = $this|0;
 $bufmode = $bufmode|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $switch$select = 0, $switch$select3 = 0, $switch$selectcmp = 0, $switch$selectcmp2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($1|0)<(0); //@line 184 "modules/filesystem/DroppedFile.cpp"
 if ($2) {
  $$1 = 0;
  return ($$1|0); //@line 221 "modules/filesystem/DroppedFile.cpp"
 }
 $3 = ($bufmode|0)==(0); //@line 187 "modules/filesystem/DroppedFile.cpp"
 $4 = $3 ? 0 : $0; //@line 187 "modules/filesystem/DroppedFile.cpp"
 $5 = $3 ? 0 : $1; //@line 187 "modules/filesystem/DroppedFile.cpp"
 $6 = HEAP32[$this>>2]|0; //@line 192 "modules/filesystem/DroppedFile.cpp"
 $7 = ((($6)) + 24|0); //@line 192 "modules/filesystem/DroppedFile.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 192 "modules/filesystem/DroppedFile.cpp"
 $9 = (FUNCTION_TABLE_ii[$8 & 8191]($this)|0); //@line 192 "modules/filesystem/DroppedFile.cpp"
 if (!($9)) {
  $10 = ((($this)) + 28|0); //@line 194 "modules/filesystem/DroppedFile.cpp"
  HEAP32[$10>>2] = $bufmode; //@line 194 "modules/filesystem/DroppedFile.cpp"
  $11 = ((($this)) + 32|0); //@line 195 "modules/filesystem/DroppedFile.cpp"
  $12 = $11; //@line 195 "modules/filesystem/DroppedFile.cpp"
  $13 = $12; //@line 195 "modules/filesystem/DroppedFile.cpp"
  HEAP32[$13>>2] = $4; //@line 195 "modules/filesystem/DroppedFile.cpp"
  $14 = (($12) + 4)|0; //@line 195 "modules/filesystem/DroppedFile.cpp"
  $15 = $14; //@line 195 "modules/filesystem/DroppedFile.cpp"
  HEAP32[$15>>2] = $5; //@line 195 "modules/filesystem/DroppedFile.cpp"
  $$1 = 1;
  return ($$1|0); //@line 221 "modules/filesystem/DroppedFile.cpp"
 }
 $switch$selectcmp = ($bufmode|0)==(1); //@line 200 "modules/filesystem/DroppedFile.cpp"
 $switch$select = $switch$selectcmp ? 1 : 2; //@line 200 "modules/filesystem/DroppedFile.cpp"
 $switch$selectcmp2 = ($bufmode|0)==(2); //@line 200 "modules/filesystem/DroppedFile.cpp"
 $switch$select3 = $switch$selectcmp2 ? 0 : $switch$select; //@line 200 "modules/filesystem/DroppedFile.cpp"
 $16 = ((($this)) + 20|0); //@line 214 "modules/filesystem/DroppedFile.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 214 "modules/filesystem/DroppedFile.cpp"
 $18 = (_setvbuf($17,0,$switch$select3,$4)|0); //@line 214 "modules/filesystem/DroppedFile.cpp"
 $19 = ($18|0)==(0); //@line 214 "modules/filesystem/DroppedFile.cpp"
 if (!($19)) {
  $$1 = 0;
  return ($$1|0); //@line 221 "modules/filesystem/DroppedFile.cpp"
 }
 $20 = ((($this)) + 28|0); //@line 217 "modules/filesystem/DroppedFile.cpp"
 HEAP32[$20>>2] = $bufmode; //@line 217 "modules/filesystem/DroppedFile.cpp"
 $21 = ((($this)) + 32|0); //@line 218 "modules/filesystem/DroppedFile.cpp"
 $22 = $21; //@line 218 "modules/filesystem/DroppedFile.cpp"
 $23 = $22; //@line 218 "modules/filesystem/DroppedFile.cpp"
 HEAP32[$23>>2] = $4; //@line 218 "modules/filesystem/DroppedFile.cpp"
 $24 = (($22) + 4)|0; //@line 218 "modules/filesystem/DroppedFile.cpp"
 $25 = $24; //@line 218 "modules/filesystem/DroppedFile.cpp"
 HEAP32[$25>>2] = $5; //@line 218 "modules/filesystem/DroppedFile.cpp"
 $$1 = 1;
 return ($$1|0); //@line 221 "modules/filesystem/DroppedFile.cpp"
}
function __ZNK4love10filesystem11DroppedFile9getBufferERx($this,$size) {
 $this = $this|0;
 $size = $size|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 225 "modules/filesystem/DroppedFile.cpp"
 $1 = $0; //@line 225 "modules/filesystem/DroppedFile.cpp"
 $2 = $1; //@line 225 "modules/filesystem/DroppedFile.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 225 "modules/filesystem/DroppedFile.cpp"
 $4 = (($1) + 4)|0; //@line 225 "modules/filesystem/DroppedFile.cpp"
 $5 = $4; //@line 225 "modules/filesystem/DroppedFile.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 225 "modules/filesystem/DroppedFile.cpp"
 $7 = $size; //@line 225 "modules/filesystem/DroppedFile.cpp"
 $8 = $7; //@line 225 "modules/filesystem/DroppedFile.cpp"
 HEAP32[$8>>2] = $3; //@line 225 "modules/filesystem/DroppedFile.cpp"
 $9 = (($7) + 4)|0; //@line 225 "modules/filesystem/DroppedFile.cpp"
 $10 = $9; //@line 225 "modules/filesystem/DroppedFile.cpp"
 HEAP32[$10>>2] = $6; //@line 225 "modules/filesystem/DroppedFile.cpp"
 $11 = ((($this)) + 28|0); //@line 226 "modules/filesystem/DroppedFile.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 226 "modules/filesystem/DroppedFile.cpp"
 return ($12|0); //@line 226 "modules/filesystem/DroppedFile.cpp"
}
function __ZNK4love10filesystem11DroppedFile7getModeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 236 "modules/filesystem/DroppedFile.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 236 "modules/filesystem/DroppedFile.cpp"
 return ($1|0); //@line 236 "modules/filesystem/DroppedFile.cpp"
}
function __ZNK4love10filesystem11DroppedFile11getFilenameEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 231 "modules/filesystem/DroppedFile.cpp"
 return ($0|0); //@line 231 "modules/filesystem/DroppedFile.cpp"
}
function __ZN4love10filesystem11DroppedFileC2ERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE($this,$filename) {
 $this = $this|0;
 $filename = $filename|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 43 "modules/filesystem/File.h"
 HEAP32[$this>>2] = (23372); //@line 46 "modules/filesystem/DroppedFile.cpp"
 $0 = ((($this)) + 8|0); //@line 41 "modules/filesystem/DroppedFile.cpp"
 __THREW__ = 0;
 invoke_vii(3351,($0|0),($filename|0)); //@line 41 "modules/filesystem/DroppedFile.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  __ZN4love10filesystem4FileD2Ev($this); //@line 47 "modules/filesystem/DroppedFile.cpp"
  ___resumeException($4|0);
  // unreachable;
 } else {
  $3 = ((($this)) + 20|0); //@line 42 "modules/filesystem/DroppedFile.cpp"
  ;HEAP32[$3>>2]=0|0;HEAP32[$3+4>>2]=0|0;HEAP32[$3+8>>2]=0|0;HEAP32[$3+12>>2]=0|0;HEAP32[$3+16>>2]=0|0; //@line 43 "modules/filesystem/DroppedFile.cpp"
  return; //@line 47 "modules/filesystem/DroppedFile.cpp"
 }
}
function __ZN4love10filesystem4FileD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 30 "modules/filesystem/File.cpp"
}
function __ZN4love10filesystem4File4readEx($this,$0,$1) {
 $this = $this|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$01 = 0, $$02 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fileData$0 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $2 = HEAP32[$this>>2]|0; //@line 34 "modules/filesystem/File.cpp"
 $3 = ((($2)) + 24|0); //@line 34 "modules/filesystem/File.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 34 "modules/filesystem/File.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($this)|0); //@line 34 "modules/filesystem/File.cpp"
 if (!($5)) {
  $6 = HEAP32[$this>>2]|0; //@line 36 "modules/filesystem/File.cpp"
  $7 = ((($6)) + 16|0); //@line 36 "modules/filesystem/File.cpp"
  $8 = HEAP32[$7>>2]|0; //@line 36 "modules/filesystem/File.cpp"
  $9 = (FUNCTION_TABLE_iii[$8 & 8191]($this,1)|0); //@line 36 "modules/filesystem/File.cpp"
  if (!($9)) {
   $10 = (___cxa_allocate_exception(16)|0); //@line 37 "modules/filesystem/File.cpp"
   $11 = HEAP32[$this>>2]|0; //@line 37 "modules/filesystem/File.cpp"
   $12 = ((($11)) + 76|0); //@line 37 "modules/filesystem/File.cpp"
   $13 = HEAP32[$12>>2]|0; //@line 37 "modules/filesystem/File.cpp"
   __THREW__ = 0;
   $14 = (invoke_ii($13|0,($this|0))|0); //@line 37 "modules/filesystem/File.cpp"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if (!($16)) {
    $17 = HEAP8[$14>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $18 = $17 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $19 = ($18<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $20 = ((($14)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $21 = HEAP32[$20>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $22 = ((($14)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $23 = $19 ? $22 : $21; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    __THREW__ = 0;
    HEAP32[$vararg_buffer>>2] = $23; //@line 37 "modules/filesystem/File.cpp"
    invoke_viii(3349,($10|0),(366278|0),($vararg_buffer|0)); //@line 37 "modules/filesystem/File.cpp"
    $24 = __THREW__; __THREW__ = 0;
    $25 = $24&1;
    if (!($25)) {
     ___cxa_throw(($10|0),(96|0),(344|0)); //@line 37 "modules/filesystem/File.cpp"
     // unreachable; //@line 37 "modules/filesystem/File.cpp"
    }
   }
   $26 = ___cxa_find_matching_catch_2()|0;
   $27 = tempRet0;
   ___cxa_free_exception(($10|0)); //@line 37 "modules/filesystem/File.cpp"
   $$01 = $26;$$02 = $27;
   ___resumeException($$01|0);
   // unreachable;
  }
 }
 $28 = HEAP32[$this>>2]|0; //@line 39 "modules/filesystem/File.cpp"
 $29 = ((($28)) + 28|0); //@line 39 "modules/filesystem/File.cpp"
 $30 = HEAP32[$29>>2]|0; //@line 39 "modules/filesystem/File.cpp"
 $31 = (FUNCTION_TABLE_ii[$30 & 8191]($this)|0); //@line 39 "modules/filesystem/File.cpp"
 $32 = tempRet0; //@line 39 "modules/filesystem/File.cpp"
 $33 = HEAP32[$this>>2]|0; //@line 40 "modules/filesystem/File.cpp"
 $34 = ((($33)) + 56|0); //@line 40 "modules/filesystem/File.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 40 "modules/filesystem/File.cpp"
 $36 = (FUNCTION_TABLE_ii[$35 & 8191]($this)|0); //@line 40 "modules/filesystem/File.cpp"
 $37 = tempRet0; //@line 40 "modules/filesystem/File.cpp"
 $38 = ($0|0)==(-1); //@line 41 "modules/filesystem/File.cpp"
 $39 = ($1|0)==(-1); //@line 41 "modules/filesystem/File.cpp"
 $40 = $38 & $39; //@line 41 "modules/filesystem/File.cpp"
 $41 = $40 ? $31 : $0; //@line 41 "modules/filesystem/File.cpp"
 $42 = $40 ? $32 : $1; //@line 41 "modules/filesystem/File.cpp"
 $43 = ($42|0)<(0); //@line 43 "modules/filesystem/File.cpp"
 if ($43) {
  $44 = (___cxa_allocate_exception(16)|0); //@line 44 "modules/filesystem/File.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($44|0),(507273|0),($vararg_buffer1|0)); //@line 44 "modules/filesystem/File.cpp"
  $45 = __THREW__; __THREW__ = 0;
  $46 = $45&1;
  if (!($46)) {
   ___cxa_throw(($44|0),(96|0),(344|0)); //@line 44 "modules/filesystem/File.cpp"
   // unreachable; //@line 44 "modules/filesystem/File.cpp"
  }
  $47 = ___cxa_find_matching_catch_2()|0;
  $48 = tempRet0;
  ___cxa_free_exception(($44|0)); //@line 44 "modules/filesystem/File.cpp"
  $$01 = $47;$$02 = $48;
  ___resumeException($$01|0);
  // unreachable;
 }
 $49 = ($37|0)<(0); //@line 47 "modules/filesystem/File.cpp"
 $50 = ($37|0)>($32|0); //@line 49 "modules/filesystem/File.cpp"
 $51 = ($36>>>0)>($31>>>0); //@line 49 "modules/filesystem/File.cpp"
 $52 = ($37|0)==($32|0); //@line 49 "modules/filesystem/File.cpp"
 $53 = $52 & $51; //@line 49 "modules/filesystem/File.cpp"
 $54 = $50 | $53; //@line 49 "modules/filesystem/File.cpp"
 $55 = $54 ? $31 : $36; //@line 49 "modules/filesystem/File.cpp"
 $56 = $54 ? $32 : $37; //@line 49 "modules/filesystem/File.cpp"
 $57 = $49 ? 0 : $55; //@line 47 "modules/filesystem/File.cpp"
 $58 = $49 ? 0 : $56; //@line 47 "modules/filesystem/File.cpp"
 $59 = (_i64Add(($57|0),($58|0),($41|0),($42|0))|0); //@line 52 "modules/filesystem/File.cpp"
 $60 = tempRet0; //@line 52 "modules/filesystem/File.cpp"
 $61 = ($60|0)>($32|0); //@line 52 "modules/filesystem/File.cpp"
 $62 = ($59>>>0)>($31>>>0); //@line 52 "modules/filesystem/File.cpp"
 $63 = ($60|0)==($32|0); //@line 52 "modules/filesystem/File.cpp"
 $64 = $63 & $62; //@line 52 "modules/filesystem/File.cpp"
 $65 = $61 | $64; //@line 52 "modules/filesystem/File.cpp"
 $66 = (_i64Subtract(($31|0),($32|0),($57|0),($58|0))|0); //@line 53 "modules/filesystem/File.cpp"
 $67 = tempRet0; //@line 53 "modules/filesystem/File.cpp"
 $68 = $65 ? $66 : $41; //@line 52 "modules/filesystem/File.cpp"
 $69 = $65 ? $67 : $42; //@line 52 "modules/filesystem/File.cpp"
 $70 = (__Znwj(48)|0); //@line 55 "modules/filesystem/File.cpp"
 $71 = HEAP32[$this>>2]|0; //@line 55 "modules/filesystem/File.cpp"
 $72 = ((($71)) + 76|0); //@line 55 "modules/filesystem/File.cpp"
 $73 = HEAP32[$72>>2]|0; //@line 55 "modules/filesystem/File.cpp"
 __THREW__ = 0;
 $74 = (invoke_ii($73|0,($this|0))|0); //@line 55 "modules/filesystem/File.cpp"
 $75 = __THREW__; __THREW__ = 0;
 $76 = $75&1;
 if (!($76)) {
  __THREW__ = 0;
  invoke_viiii(3972,($70|0),($68|0),($69|0),($74|0)); //@line 55 "modules/filesystem/File.cpp"
  $77 = __THREW__; __THREW__ = 0;
  $78 = $77&1;
  if (!($78)) {
   $79 = HEAP32[$this>>2]|0; //@line 56 "modules/filesystem/File.cpp"
   $80 = ((($79)) + 36|0); //@line 56 "modules/filesystem/File.cpp"
   $81 = HEAP32[$80>>2]|0; //@line 56 "modules/filesystem/File.cpp"
   $82 = HEAP32[$70>>2]|0; //@line 56 "modules/filesystem/File.cpp"
   $83 = ((($82)) + 16|0); //@line 56 "modules/filesystem/File.cpp"
   $84 = HEAP32[$83>>2]|0; //@line 56 "modules/filesystem/File.cpp"
   $85 = (FUNCTION_TABLE_ii[$84 & 8191]($70)|0); //@line 56 "modules/filesystem/File.cpp"
   $86 = (FUNCTION_TABLE_iiiii[$81 & 8191]($this,$85,$68,$69)|0); //@line 56 "modules/filesystem/File.cpp"
   $87 = tempRet0; //@line 56 "modules/filesystem/File.cpp"
   $88 = ($87|0)<(0); //@line 58 "modules/filesystem/File.cpp"
   if (!($88)) {
    $89 = ($86|0)!=(0); //@line 58 "modules/filesystem/File.cpp"
    $90 = ($87|0)!=(0); //@line 58 "modules/filesystem/File.cpp"
    $91 = $89 | $90; //@line 58 "modules/filesystem/File.cpp"
    $92 = ($86|0)==($68|0); //@line 58 "modules/filesystem/File.cpp"
    $93 = ($87|0)==($69|0); //@line 58 "modules/filesystem/File.cpp"
    $94 = $92 & $93; //@line 58 "modules/filesystem/File.cpp"
    $or$cond = $91 | $94; //@line 58 "modules/filesystem/File.cpp"
    if ($or$cond) {
     $105 = ($87|0)<($69|0); //@line 64 "modules/filesystem/File.cpp"
     $106 = ($86>>>0)<($68>>>0); //@line 64 "modules/filesystem/File.cpp"
     $107 = ($87|0)==($69|0); //@line 64 "modules/filesystem/File.cpp"
     $108 = $107 & $106; //@line 64 "modules/filesystem/File.cpp"
     $109 = $105 | $108; //@line 64 "modules/filesystem/File.cpp"
     do {
      if ($109) {
       $110 = (__Znwj(48)|0); //@line 66 "modules/filesystem/File.cpp"
       $111 = HEAP32[$this>>2]|0; //@line 66 "modules/filesystem/File.cpp"
       $112 = ((($111)) + 76|0); //@line 66 "modules/filesystem/File.cpp"
       $113 = HEAP32[$112>>2]|0; //@line 66 "modules/filesystem/File.cpp"
       __THREW__ = 0;
       $114 = (invoke_ii($113|0,($this|0))|0); //@line 66 "modules/filesystem/File.cpp"
       $115 = __THREW__; __THREW__ = 0;
       $116 = $115&1;
       if (!($116)) {
        __THREW__ = 0;
        invoke_viiii(3972,($110|0),($86|0),($87|0),($114|0)); //@line 66 "modules/filesystem/File.cpp"
        $117 = __THREW__; __THREW__ = 0;
        $118 = $117&1;
        if (!($118)) {
         $119 = HEAP32[$110>>2]|0; //@line 67 "modules/filesystem/File.cpp"
         $120 = ((($119)) + 16|0); //@line 67 "modules/filesystem/File.cpp"
         $121 = HEAP32[$120>>2]|0; //@line 67 "modules/filesystem/File.cpp"
         $122 = (FUNCTION_TABLE_ii[$121 & 8191]($110)|0); //@line 67 "modules/filesystem/File.cpp"
         $123 = HEAP32[$70>>2]|0; //@line 67 "modules/filesystem/File.cpp"
         $124 = ((($123)) + 16|0); //@line 67 "modules/filesystem/File.cpp"
         $125 = HEAP32[$124>>2]|0; //@line 67 "modules/filesystem/File.cpp"
         $126 = (FUNCTION_TABLE_ii[$125 & 8191]($70)|0); //@line 67 "modules/filesystem/File.cpp"
         _memcpy(($122|0),($126|0),($86|0))|0; //@line 67 "modules/filesystem/File.cpp"
         $127 = HEAP32[$70>>2]|0; //@line 68 "modules/filesystem/File.cpp"
         $128 = ((($127)) + 12|0); //@line 68 "modules/filesystem/File.cpp"
         $129 = HEAP32[$128>>2]|0; //@line 68 "modules/filesystem/File.cpp"
         FUNCTION_TABLE_vi[$129 & 8191]($70); //@line 68 "modules/filesystem/File.cpp"
         $fileData$0 = $110;
         break;
        }
       }
       $130 = ___cxa_find_matching_catch_2()|0;
       $131 = tempRet0;
       __ZdlPv($110); //@line 66 "modules/filesystem/File.cpp"
       $$01 = $130;$$02 = $131;
       ___resumeException($$01|0);
       // unreachable;
      } else {
       $fileData$0 = $70;
      }
     } while(0);
     if ($5) {
      STACKTOP = sp;return ($fileData$0|0); //@line 75 "modules/filesystem/File.cpp"
     }
     $132 = HEAP32[$this>>2]|0; //@line 73 "modules/filesystem/File.cpp"
     $133 = ((($132)) + 20|0); //@line 73 "modules/filesystem/File.cpp"
     $134 = HEAP32[$133>>2]|0; //@line 73 "modules/filesystem/File.cpp"
     (FUNCTION_TABLE_ii[$134 & 8191]($this)|0); //@line 73 "modules/filesystem/File.cpp"
     STACKTOP = sp;return ($fileData$0|0); //@line 75 "modules/filesystem/File.cpp"
    }
   }
   $95 = HEAP32[$70>>2]|0; //@line 60 "modules/filesystem/File.cpp"
   $96 = ((($95)) + 4|0); //@line 60 "modules/filesystem/File.cpp"
   $97 = HEAP32[$96>>2]|0; //@line 60 "modules/filesystem/File.cpp"
   FUNCTION_TABLE_vi[$97 & 8191]($70); //@line 60 "modules/filesystem/File.cpp"
   $98 = (___cxa_allocate_exception(16)|0); //@line 61 "modules/filesystem/File.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($98|0),(366568|0),($vararg_buffer3|0)); //@line 61 "modules/filesystem/File.cpp"
   $99 = __THREW__; __THREW__ = 0;
   $100 = $99&1;
   if (!($100)) {
    ___cxa_throw(($98|0),(96|0),(344|0)); //@line 61 "modules/filesystem/File.cpp"
    // unreachable; //@line 61 "modules/filesystem/File.cpp"
   }
   $103 = ___cxa_find_matching_catch_2()|0;
   $104 = tempRet0;
   ___cxa_free_exception(($98|0)); //@line 61 "modules/filesystem/File.cpp"
   $$01 = $103;$$02 = $104;
   ___resumeException($$01|0);
   // unreachable;
  }
 }
 $101 = ___cxa_find_matching_catch_2()|0;
 $102 = tempRet0;
 __ZdlPv($70); //@line 55 "modules/filesystem/File.cpp"
 $$01 = $101;$$02 = $102;
 ___resumeException($$01|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love10filesystem4File5writeEPKNS_4DataEx($this,$data,$0,$1) {
 $this = $this|0;
 $data = $data|0;
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[$this>>2]|0; //@line 80 "modules/filesystem/File.cpp"
 $3 = ((($2)) + 40|0); //@line 80 "modules/filesystem/File.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 80 "modules/filesystem/File.cpp"
 $5 = HEAP32[$data>>2]|0; //@line 80 "modules/filesystem/File.cpp"
 $6 = ((($5)) + 16|0); //@line 80 "modules/filesystem/File.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 80 "modules/filesystem/File.cpp"
 $8 = (FUNCTION_TABLE_ii[$7 & 8191]($data)|0); //@line 80 "modules/filesystem/File.cpp"
 $9 = ($0|0)==(-1); //@line 80 "modules/filesystem/File.cpp"
 $10 = ($1|0)==(-1); //@line 80 "modules/filesystem/File.cpp"
 $11 = $9 & $10; //@line 80 "modules/filesystem/File.cpp"
 if (!($11)) {
  $16 = $0;$17 = $1;
  $18 = (FUNCTION_TABLE_iiiii[$4 & 8191]($this,$8,$16,$17)|0); //@line 80 "modules/filesystem/File.cpp"
  return ($18|0); //@line 80 "modules/filesystem/File.cpp"
 }
 $12 = HEAP32[$data>>2]|0; //@line 80 "modules/filesystem/File.cpp"
 $13 = ((($12)) + 20|0); //@line 80 "modules/filesystem/File.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 80 "modules/filesystem/File.cpp"
 $15 = (FUNCTION_TABLE_ii[$14 & 8191]($data)|0); //@line 80 "modules/filesystem/File.cpp"
 $16 = $15;$17 = 0;
 $18 = (FUNCTION_TABLE_iiiii[$4 & 8191]($this,$8,$16,$17)|0); //@line 80 "modules/filesystem/File.cpp"
 return ($18|0); //@line 80 "modules/filesystem/File.cpp"
}
function __ZNK4love10filesystem4File12getExtensionEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 85 "modules/filesystem/File.cpp"
 $1 = ((($0)) + 76|0); //@line 85 "modules/filesystem/File.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 85 "modules/filesystem/File.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 85 "modules/filesystem/File.cpp"
 $4 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE5rfindEcj($3,46,-1)|0); //@line 86 "modules/filesystem/File.cpp"
 $5 = ($4|0)==(-1); //@line 88 "modules/filesystem/File.cpp"
 if ($5) {
  ;HEAP32[$agg$result>>2]=0|0;HEAP32[$agg$result+4>>2]=0|0;HEAP32[$agg$result+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  return; //@line 92 "modules/filesystem/File.cpp"
 } else {
  $6 = (($4) + 1)|0; //@line 89 "modules/filesystem/File.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_jjRKS4_($agg$result,$3,$6,-1,$3); //@line 3345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  return; //@line 92 "modules/filesystem/File.cpp"
 }
}
function __GLOBAL__sub_I_File_cpp() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[(632732)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632744)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632756)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632768)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632780)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632792)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632804)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632816)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(632820)>>2]=0|0;HEAP32[(632820)+4>>2]=0|0;HEAP32[(632820)+8>>2]=0|0;HEAP32[(632820)+12>>2]=0|0; //@line 44 "./common/StringMap.h"
 $0 = HEAP32[5864]|0; //@line 49 "./common/StringMap.h"
 $1 = HEAP32[(23460)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_10filesystem4File4ModeELj4EE3addEPKcS3_(632724,$0,$1)|0); //@line 49 "./common/StringMap.h"
 $2 = HEAP32[(23464)>>2]|0; //@line 49 "./common/StringMap.h"
 $3 = HEAP32[(23468)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_10filesystem4File4ModeELj4EE3addEPKcS3_(632724,$2,$3)|0); //@line 49 "./common/StringMap.h"
 $4 = HEAP32[(23472)>>2]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(23476)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_10filesystem4File4ModeELj4EE3addEPKcS3_(632724,$4,$5)|0); //@line 49 "./common/StringMap.h"
 $6 = HEAP32[(23480)>>2]|0; //@line 49 "./common/StringMap.h"
 $7 = HEAP32[(23484)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_10filesystem4File4ModeELj4EE3addEPKcS3_(632724,$6,$7)|0); //@line 49 "./common/StringMap.h"
 HEAP8[(632844)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632856)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632868)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632880)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632892)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632904)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(632908)>>2]=0|0;HEAP32[(632908)+4>>2]=0|0;HEAP32[(632908)+8>>2]=0|0; //@line 44 "./common/StringMap.h"
 $8 = HEAP32[5872]|0; //@line 49 "./common/StringMap.h"
 $9 = HEAP32[(23492)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_10filesystem4File10BufferModeELj3EE3addEPKcS3_(632836,$8,$9)|0); //@line 49 "./common/StringMap.h"
 $10 = HEAP32[(23496)>>2]|0; //@line 49 "./common/StringMap.h"
 $11 = HEAP32[(23500)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_10filesystem4File10BufferModeELj3EE3addEPKcS3_(632836,$10,$11)|0); //@line 49 "./common/StringMap.h"
 $12 = HEAP32[(23504)>>2]|0; //@line 49 "./common/StringMap.h"
 $13 = HEAP32[(23508)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_10filesystem4File10BufferModeELj3EE3addEPKcS3_(632836,$12,$13)|0); //@line 49 "./common/StringMap.h"
 return;
}
function __ZN4love9StringMapINS_10filesystem4File4ModeELj4EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0;
 var $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 7; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = $13 & 7; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = $22 & 7; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = $27 & 7; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = $32 & 7; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = $37 & 7; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $42 = (($hash$0$lcssa$i) + 6)|0; //@line 112 "./common/StringMap.h"
       $43 = $42 & 7; //@line 112 "./common/StringMap.h"
       $44 = (((($this) + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
       $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
       $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
       if ($46) {
        $$lcssa = $43;$$lcssa10 = $44;
        label = 5;
       } else {
        $47 = (($hash$0$lcssa$i) + 7)|0; //@line 112 "./common/StringMap.h"
        $48 = $47 & 7; //@line 112 "./common/StringMap.h"
        $49 = (((($this) + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
        $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
        $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
        if ($51) {
         $$lcssa = $48;$$lcssa10 = $49;
         label = 5;
        } else {
         $inserted$2$off0 = 0;
        }
       }
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(3); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 96|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love9StringMapINS_10filesystem4File10BufferModeELj3EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $inserted$2$off0 = 0;
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(2); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 72|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love10filesystem4File11getConstantEPKcRNS1_4ModeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = $9 & 7; //@line 72 "./common/StringMap.h"
  $11 = (((632724 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (632724 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(8); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((632724 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 96 "modules/filesystem/File.cpp"
  return ($$$2$i|0); //@line 96 "modules/filesystem/File.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 96 "modules/filesystem/File.cpp"
  return ($$$2$i|0); //@line 96 "modules/filesystem/File.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem4File11getConstantENS1_4ModeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(3); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (632820 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 101 "modules/filesystem/File.cpp"
}
function __ZN4love10filesystem4File11getConstantEPKcRNS1_10BufferModeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 6)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((632836 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (632836 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(6); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((632836 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 106 "modules/filesystem/File.cpp"
  return ($$$2$i|0); //@line 106 "modules/filesystem/File.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 106 "modules/filesystem/File.cpp"
  return ($$$2$i|0); //@line 106 "modules/filesystem/File.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem4File11getConstantENS1_10BufferModeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(2); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (632908 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 111 "modules/filesystem/File.cpp"
}
function __ZN4love10filesystem10FilesystemD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ModuleD2Ev($this); //@line 51 "modules/filesystem/Filesystem.cpp"
 return; //@line 51 "modules/filesystem/Filesystem.cpp"
}
function __ZN4love10filesystem10FilesystemD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ModuleD2Ev($this); //@line 51 "modules/filesystem/Filesystem.cpp"
 __ZdlPv($this); //@line 50 "modules/filesystem/Filesystem.cpp"
 return; //@line 51 "modules/filesystem/Filesystem.cpp"
}
function __ZN4love10filesystem10Filesystem22setAndroidSaveExternalEb($this,$useExternal) {
 $this = $this|0;
 $useExternal = $useExternal|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 55 "modules/filesystem/Filesystem.cpp"
 $1 = $useExternal&1; //@line 55 "modules/filesystem/Filesystem.cpp"
 HEAP8[$0>>0] = $1; //@line 55 "modules/filesystem/Filesystem.cpp"
 return; //@line 56 "modules/filesystem/Filesystem.cpp"
}
function __ZNK4love10filesystem10Filesystem21isAndroidSaveExternalEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 60 "modules/filesystem/Filesystem.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 60 "modules/filesystem/Filesystem.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 60 "modules/filesystem/Filesystem.cpp"
 return ($2|0); //@line 60 "modules/filesystem/Filesystem.cpp"
}
function __ZNK4love10filesystem10Filesystem15isRealDirectoryERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE($this,$path) {
 $this = $this|0;
 $path = $path|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buf = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $buf = sp;
 $0 = HEAP8[$path>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $1 = $0 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = ($1<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ((($path)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = HEAP32[$3>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($path)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = $2 ? $5 : $4; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = (_stat($6,$buf)|0); //@line 77 "modules/filesystem/Filesystem.cpp"
 $8 = ($7|0)==(0); //@line 77 "modules/filesystem/Filesystem.cpp"
 if (!($8)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 82 "modules/filesystem/Filesystem.cpp"
 }
 $9 = ((($buf)) + 12|0); //@line 80 "modules/filesystem/Filesystem.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 80 "modules/filesystem/Filesystem.cpp"
 $11 = $10 & 61440; //@line 80 "modules/filesystem/Filesystem.cpp"
 $12 = ($11|0)==(16384); //@line 80 "modules/filesystem/Filesystem.cpp"
 $$0 = $12;
 STACKTOP = sp;return ($$0|0); //@line 82 "modules/filesystem/Filesystem.cpp"
}
function __ZNK4love10filesystem10Filesystem17getExecutablePathEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 2048|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $buffer = sp;
 _memset(($buffer|0),0,2048)|0; //@line 101 "modules/filesystem/Filesystem.cpp"
 $0 = (_readlink(517112,$buffer,2048)|0); //@line 103 "modules/filesystem/Filesystem.cpp"
 $1 = ($0|0)<(1); //@line 104 "modules/filesystem/Filesystem.cpp"
 if ($1) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($agg$result,1282466,0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  STACKTOP = sp;return; //@line 111 "modules/filesystem/Filesystem.cpp"
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($agg$result,$buffer,$0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  STACKTOP = sp;return; //@line 111 "modules/filesystem/Filesystem.cpp"
 }
}
function __ZN4love10filesystem10FilesystemC2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./common/Module.h"
 HEAP32[$this>>2] = (23520); //@line 46 "modules/filesystem/Filesystem.cpp"
 return; //@line 47 "modules/filesystem/Filesystem.cpp"
}
function _luaopen_droppedfile($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 34 "modules/filesystem/wrap_DroppedFile.cpp"
 HEAP32[$vararg_buffer>>2] = 23700; //@line 36 "modules/filesystem/wrap_DroppedFile.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 36 "modules/filesystem/wrap_DroppedFile.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 36 "modules/filesystem/wrap_DroppedFile.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,6,366334,$vararg_buffer)|0); //@line 36 "modules/filesystem/wrap_DroppedFile.cpp"
 STACKTOP = sp;return ($0|0); //@line 36 "modules/filesystem/wrap_DroppedFile.cpp"
}
function __ZN4love10filesystem14w_File_getSizeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0; //@line 49 "modules/filesystem/wrap_File.cpp"
 $vararg_buffer1 = sp + 8|0; //@line 49 "modules/filesystem/wrap_File.cpp"
 $vararg_buffer = sp; //@line 49 "modules/filesystem/wrap_File.cpp"
 $0 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,1,5)|0); //@line 46 "modules/filesystem/wrap_File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 56 "modules/filesystem/wrap_File.cpp"
 $2 = ((($1)) + 28|0); //@line 56 "modules/filesystem/wrap_File.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 56 "modules/filesystem/wrap_File.cpp"
 __THREW__ = 0;
 $4 = (invoke_ii($3|0,($0|0))|0); //@line 56 "modules/filesystem/wrap_File.cpp"
 $5 = tempRet0; //@line 56 "modules/filesystem/wrap_File.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $11 = ___cxa_find_matching_catch_3(96|0)|0;
  $12 = tempRet0;
  $13 = (_llvm_eh_typeid_for((96|0))|0); //@line 57 "modules/filesystem/wrap_File.cpp"
  $14 = ($12|0)==($13|0); //@line 57 "modules/filesystem/wrap_File.cpp"
  if (!($14)) {
   $$01 = $12;$$02 = $11;
   ___resumeException($$02|0);
   // unreachable;
  }
  $15 = (___cxa_begin_catch(($11|0))|0); //@line 57 "modules/filesystem/wrap_File.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 60 "modules/filesystem/wrap_File.cpp"
  $17 = ((($16)) + 8|0); //@line 60 "modules/filesystem/wrap_File.cpp"
  $18 = HEAP32[$17>>2]|0; //@line 60 "modules/filesystem/wrap_File.cpp"
  $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 60 "modules/filesystem/wrap_File.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer>>2] = $19; //@line 60 "modules/filesystem/wrap_File.cpp"
  (invoke_iiii(3512,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 60 "modules/filesystem/wrap_File.cpp"
  $20 = __THREW__; __THREW__ = 0;
  $21 = $20&1;
  if (!($21)) {
   ___cxa_end_catch(); //@line 61 "modules/filesystem/wrap_File.cpp"
   $$0 = 2;
   STACKTOP = sp;return ($$0|0); //@line 71 "modules/filesystem/wrap_File.cpp"
  }
  $22 = ___cxa_find_matching_catch_2()|0;
  $23 = tempRet0;
  __THREW__ = 0;
  invoke_v(3378); //@line 61 "modules/filesystem/wrap_File.cpp"
  $24 = __THREW__; __THREW__ = 0;
  $25 = $24&1;
  if ($25) {
   $32 = ___cxa_find_matching_catch_3(0|0)|0;
   $33 = tempRet0;
   ___clang_call_terminate($32); //@line 61 "modules/filesystem/wrap_File.cpp"
   // unreachable; //@line 61 "modules/filesystem/wrap_File.cpp"
  } else {
   $$01 = $23;$$02 = $22;
   ___resumeException($$02|0);
   // unreachable;
  }
 } else {
  $8 = ($4|0)==(-1); //@line 64 "modules/filesystem/wrap_File.cpp"
  $9 = ($5|0)==(-1); //@line 64 "modules/filesystem/wrap_File.cpp"
  $10 = $8 & $9; //@line 64 "modules/filesystem/wrap_File.cpp"
  if ($10) {
   (__ZN4love10filesystem12luax_ioErrorEP9lua_StatePKcz($L,366354,$vararg_buffer1)|0); //@line 65 "modules/filesystem/wrap_File.cpp"
   $$0 = 2;
   STACKTOP = sp;return ($$0|0); //@line 71 "modules/filesystem/wrap_File.cpp"
  }
  $26 = ($5|0)>(2097151); //@line 66 "modules/filesystem/wrap_File.cpp"
  $27 = ($4>>>0)>(4294967295); //@line 66 "modules/filesystem/wrap_File.cpp"
  $28 = ($5|0)==(2097151); //@line 66 "modules/filesystem/wrap_File.cpp"
  $29 = $28 & $27; //@line 66 "modules/filesystem/wrap_File.cpp"
  $30 = $26 | $29; //@line 66 "modules/filesystem/wrap_File.cpp"
  if ($30) {
   (__ZN4love10filesystem12luax_ioErrorEP9lua_StatePKcz($L,366385,$vararg_buffer3)|0); //@line 67 "modules/filesystem/wrap_File.cpp"
   $$0 = 2;
   STACKTOP = sp;return ($$0|0); //@line 71 "modules/filesystem/wrap_File.cpp"
  } else {
   $31 = (+($4>>>0)) + (4294967296.0*(+($5|0))); //@line 69 "modules/filesystem/wrap_File.cpp"
   _lua_pushnumber($L,$31); //@line 69 "modules/filesystem/wrap_File.cpp"
   $$0 = 1;
   STACKTOP = sp;return ($$0|0); //@line 71 "modules/filesystem/wrap_File.cpp"
  }
 }
 return (0)|0;
}
function __ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love10filesystem12luax_ioErrorEP9lua_StatePKcz($L,$fmt,$varargs) {
 $L = $L|0;
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $args = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $args = sp;
 HEAP32[$args>>2] = $varargs; //@line 35 "modules/filesystem/wrap_File.cpp"
 _lua_pushnil($L); //@line 37 "modules/filesystem/wrap_File.cpp"
 (_lua_pushvfstring($L,$fmt,$args)|0); //@line 38 "modules/filesystem/wrap_File.cpp"
 STACKTOP = sp;return 2; //@line 41 "modules/filesystem/wrap_File.cpp"
}
function __ZN4love10filesystem11w_File_openEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $mode = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $mode = sp + 12|0;
 $0 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,1,5)|0); //@line 46 "modules/filesystem/wrap_File.cpp"
 $1 = (_luaL_checklstring($L,2,0)|0); //@line 76 "modules/filesystem/wrap_File.cpp"
 $2 = (__ZN4love10filesystem4File11getConstantEPKcRNS1_4ModeE($1,$mode)|0); //@line 79 "modules/filesystem/wrap_File.cpp"
 if (!($2)) {
  HEAP32[$vararg_buffer>>2] = $1; //@line 80 "modules/filesystem/wrap_File.cpp"
  $3 = (_luaL_error($L,366404,$vararg_buffer)|0); //@line 80 "modules/filesystem/wrap_File.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 92 "modules/filesystem/wrap_File.cpp"
 }
 $4 = HEAP32[$0>>2]|0; //@line 84 "modules/filesystem/wrap_File.cpp"
 $5 = ((($4)) + 16|0); //@line 84 "modules/filesystem/wrap_File.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 84 "modules/filesystem/wrap_File.cpp"
 $7 = HEAP32[$mode>>2]|0; //@line 84 "modules/filesystem/wrap_File.cpp"
 __THREW__ = 0;
 $8 = (invoke_iii($6|0,($0|0),($7|0))|0); //@line 84 "modules/filesystem/wrap_File.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if (!($10)) {
  __THREW__ = 0;
  invoke_vii(3397,($L|0),($8|0)); //@line 84 "modules/filesystem/wrap_File.cpp"
  $11 = __THREW__; __THREW__ = 0;
  $12 = $11&1;
  if (!($12)) {
   $$0 = 1;
   STACKTOP = sp;return ($$0|0); //@line 92 "modules/filesystem/wrap_File.cpp"
  }
 }
 $13 = ___cxa_find_matching_catch_3(96|0)|0;
 $14 = tempRet0;
 $15 = (_llvm_eh_typeid_for((96|0))|0); //@line 85 "modules/filesystem/wrap_File.cpp"
 $16 = ($14|0)==($15|0); //@line 85 "modules/filesystem/wrap_File.cpp"
 if (!($16)) {
  $$01 = $14;$$02 = $13;
  ___resumeException($$02|0);
  // unreachable;
 }
 $17 = (___cxa_begin_catch(($13|0))|0); //@line 85 "modules/filesystem/wrap_File.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 88 "modules/filesystem/wrap_File.cpp"
 $19 = ((($18)) + 8|0); //@line 88 "modules/filesystem/wrap_File.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 88 "modules/filesystem/wrap_File.cpp"
 $21 = (FUNCTION_TABLE_ii[$20 & 8191]($17)|0); //@line 88 "modules/filesystem/wrap_File.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer1>>2] = $21; //@line 88 "modules/filesystem/wrap_File.cpp"
 (invoke_iiii(3512,($L|0),(595867|0),($vararg_buffer1|0))|0); //@line 88 "modules/filesystem/wrap_File.cpp"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if (!($23)) {
  ___cxa_end_catch(); //@line 89 "modules/filesystem/wrap_File.cpp"
  $$0 = 2;
  STACKTOP = sp;return ($$0|0); //@line 92 "modules/filesystem/wrap_File.cpp"
 }
 $24 = ___cxa_find_matching_catch_2()|0;
 $25 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 89 "modules/filesystem/wrap_File.cpp"
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  $28 = ___cxa_find_matching_catch_3(0|0)|0;
  $29 = tempRet0;
  ___clang_call_terminate($28); //@line 89 "modules/filesystem/wrap_File.cpp"
  // unreachable; //@line 89 "modules/filesystem/wrap_File.cpp"
 } else {
  $$01 = $25;$$02 = $24;
  ___resumeException($$02|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love10filesystem12w_File_closeEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,1,5)|0); //@line 46 "modules/filesystem/wrap_File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 97 "modules/filesystem/wrap_File.cpp"
 $2 = ((($1)) + 20|0); //@line 97 "modules/filesystem/wrap_File.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 97 "modules/filesystem/wrap_File.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 97 "modules/filesystem/wrap_File.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 97 "modules/filesystem/wrap_File.cpp"
 return 1; //@line 98 "modules/filesystem/wrap_File.cpp"
}
function __ZN4love10filesystem13w_File_isOpenEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,1,5)|0); //@line 46 "modules/filesystem/wrap_File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 104 "modules/filesystem/wrap_File.cpp"
 $2 = ((($1)) + 24|0); //@line 104 "modules/filesystem/wrap_File.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 104 "modules/filesystem/wrap_File.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 104 "modules/filesystem/wrap_File.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 104 "modules/filesystem/wrap_File.cpp"
 return 1; //@line 105 "modules/filesystem/wrap_File.cpp"
}
function __ZN4love10filesystem11w_File_readEP9lua_State($L) {
 $L = $L|0;
 var $$01 = 0, $$02 = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0.0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 108 "modules/filesystem/wrap_File.cpp"
 $0 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,1,5)|0); //@line 46 "modules/filesystem/wrap_File.cpp"
 $1 = (+_luaL_optnumber($L,2,-1.0)); //@line 113 "modules/filesystem/wrap_File.cpp"
 $2 = (~~$1)>>>0; //@line 113 "modules/filesystem/wrap_File.cpp"
 $3 = +Math_abs($1) >= 1.0 ? $1 > 0.0 ? (~~+Math_min(+Math_floor($1 / 4294967296.0), 4294967295.0)) >>> 0 : ~~+Math_ceil(($1 - +(~~$1 >>> 0)) / 4294967296.0) >>> 0 : 0; //@line 113 "modules/filesystem/wrap_File.cpp"
 $4 = HEAP32[$0>>2]|0; //@line 117 "modules/filesystem/wrap_File.cpp"
 $5 = ((($4)) + 32|0); //@line 117 "modules/filesystem/wrap_File.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 117 "modules/filesystem/wrap_File.cpp"
 __THREW__ = 0;
 $7 = (invoke_iiii($6|0,($0|0),($2|0),($3|0))|0); //@line 117 "modules/filesystem/wrap_File.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if (!($9)) {
  $10 = HEAP32[$7>>2]|0; //@line 124 "modules/filesystem/wrap_File.cpp"
  $11 = ((($10)) + 16|0); //@line 124 "modules/filesystem/wrap_File.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 124 "modules/filesystem/wrap_File.cpp"
  $13 = (FUNCTION_TABLE_ii[$12 & 8191]($7)|0); //@line 124 "modules/filesystem/wrap_File.cpp"
  $14 = HEAP32[$7>>2]|0; //@line 124 "modules/filesystem/wrap_File.cpp"
  $15 = ((($14)) + 20|0); //@line 124 "modules/filesystem/wrap_File.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 124 "modules/filesystem/wrap_File.cpp"
  $17 = (FUNCTION_TABLE_ii[$16 & 8191]($7)|0); //@line 124 "modules/filesystem/wrap_File.cpp"
  _lua_pushlstring($L,$13,$17); //@line 124 "modules/filesystem/wrap_File.cpp"
  $18 = HEAP32[$7>>2]|0; //@line 125 "modules/filesystem/wrap_File.cpp"
  $19 = ((($18)) + 20|0); //@line 125 "modules/filesystem/wrap_File.cpp"
  $20 = HEAP32[$19>>2]|0; //@line 125 "modules/filesystem/wrap_File.cpp"
  $21 = (FUNCTION_TABLE_ii[$20 & 8191]($7)|0); //@line 125 "modules/filesystem/wrap_File.cpp"
  $22 = (+($21>>>0)); //@line 125 "modules/filesystem/wrap_File.cpp"
  _lua_pushnumber($L,$22); //@line 125 "modules/filesystem/wrap_File.cpp"
  $23 = HEAP32[$7>>2]|0; //@line 126 "modules/filesystem/wrap_File.cpp"
  $24 = ((($23)) + 12|0); //@line 126 "modules/filesystem/wrap_File.cpp"
  $25 = HEAP32[$24>>2]|0; //@line 126 "modules/filesystem/wrap_File.cpp"
  FUNCTION_TABLE_vi[$25 & 8191]($7); //@line 126 "modules/filesystem/wrap_File.cpp"
  STACKTOP = sp;return 2; //@line 128 "modules/filesystem/wrap_File.cpp"
 }
 $26 = ___cxa_find_matching_catch_3(96|0)|0;
 $27 = tempRet0;
 $28 = (_llvm_eh_typeid_for((96|0))|0); //@line 118 "modules/filesystem/wrap_File.cpp"
 $29 = ($27|0)==($28|0); //@line 118 "modules/filesystem/wrap_File.cpp"
 if (!($29)) {
  $$01 = $27;$$02 = $26;
  ___resumeException($$02|0);
  // unreachable;
 }
 $30 = (___cxa_begin_catch(($26|0))|0); //@line 118 "modules/filesystem/wrap_File.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 121 "modules/filesystem/wrap_File.cpp"
 $32 = ((($31)) + 8|0); //@line 121 "modules/filesystem/wrap_File.cpp"
 $33 = HEAP32[$32>>2]|0; //@line 121 "modules/filesystem/wrap_File.cpp"
 $34 = (FUNCTION_TABLE_ii[$33 & 8191]($30)|0); //@line 121 "modules/filesystem/wrap_File.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $34; //@line 121 "modules/filesystem/wrap_File.cpp"
 (invoke_iiii(3512,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 121 "modules/filesystem/wrap_File.cpp"
 $35 = __THREW__; __THREW__ = 0;
 $36 = $35&1;
 if (!($36)) {
  ___cxa_end_catch(); //@line 122 "modules/filesystem/wrap_File.cpp"
  STACKTOP = sp;return 2; //@line 128 "modules/filesystem/wrap_File.cpp"
 }
 $37 = ___cxa_find_matching_catch_2()|0;
 $38 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 122 "modules/filesystem/wrap_File.cpp"
 $39 = __THREW__; __THREW__ = 0;
 $40 = $39&1;
 if ($40) {
  $41 = ___cxa_find_matching_catch_3(0|0)|0;
  $42 = tempRet0;
  ___clang_call_terminate($41); //@line 122 "modules/filesystem/wrap_File.cpp"
  // unreachable; //@line 122 "modules/filesystem/wrap_File.cpp"
 } else {
  $$01 = $38;$$02 = $37;
  ___resumeException($$02|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love10filesystem12w_File_writeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $8 = 0, $9 = 0;
 var $datasize = 0, $result$0$in = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $datasize = sp + 12|0;
 $0 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,1,5)|0); //@line 46 "modules/filesystem/wrap_File.cpp"
 $1 = (_lua_isstring($L,2)|0); //@line 135 "modules/filesystem/wrap_File.cpp"
 $2 = ($1|0)==(0); //@line 135 "modules/filesystem/wrap_File.cpp"
 L1: do {
  if ($2) {
   $35 = (__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE($L,2,2)|0); //@line 152 "modules/filesystem/wrap_File.cpp"
   if (!($35)) {
    $73 = (_luaL_argerror($L,2,366440)|0); //@line 166 "modules/filesystem/wrap_File.cpp"
    $$0 = $73;
    STACKTOP = sp;return ($$0|0); //@line 171 "modules/filesystem/wrap_File.cpp"
   }
   __THREW__ = 0;
   $36 = (invoke_iii(3499,($L|0),2)|0); //@line 516 "./common/runtime.h"
   $37 = __THREW__; __THREW__ = 0;
   $38 = $37&1;
   if (!($38)) {
    $39 = ((($36)) + 4|0); //@line 516 "./common/runtime.h"
    $40 = HEAP32[$39>>2]|0; //@line 516 "./common/runtime.h"
    $41 = HEAP32[$0>>2]|0; //@line 157 "modules/filesystem/wrap_File.cpp"
    $42 = ((($41)) + 44|0); //@line 157 "modules/filesystem/wrap_File.cpp"
    $43 = HEAP32[$42>>2]|0; //@line 157 "modules/filesystem/wrap_File.cpp"
    $44 = HEAP32[$40>>2]|0; //@line 157 "modules/filesystem/wrap_File.cpp"
    $45 = ((($44)) + 20|0); //@line 157 "modules/filesystem/wrap_File.cpp"
    $46 = HEAP32[$45>>2]|0; //@line 157 "modules/filesystem/wrap_File.cpp"
    __THREW__ = 0;
    $47 = (invoke_ii($46|0,($40|0))|0); //@line 157 "modules/filesystem/wrap_File.cpp"
    $48 = __THREW__; __THREW__ = 0;
    $49 = $48&1;
    if (!($49)) {
     __THREW__ = 0;
     $50 = (invoke_iiii(3976,($L|0),3,($47|0))|0); //@line 157 "modules/filesystem/wrap_File.cpp"
     $51 = __THREW__; __THREW__ = 0;
     $52 = $51&1;
     if (!($52)) {
      $53 = ($50|0)<(0); //@line 157 "modules/filesystem/wrap_File.cpp"
      $54 = $53 << 31 >> 31; //@line 157 "modules/filesystem/wrap_File.cpp"
      __THREW__ = 0;
      $55 = (invoke_iiiii($43|0,($0|0),($40|0),($50|0),($54|0))|0); //@line 157 "modules/filesystem/wrap_File.cpp"
      $56 = __THREW__; __THREW__ = 0;
      $57 = $56&1;
      if (!($57)) {
       $result$0$in = $55;
       break;
      }
     }
    }
   }
   $58 = ___cxa_find_matching_catch_3(96|0)|0;
   $59 = tempRet0;
   $60 = (_llvm_eh_typeid_for((96|0))|0); //@line 158 "modules/filesystem/wrap_File.cpp"
   $61 = ($59|0)==($60|0); //@line 158 "modules/filesystem/wrap_File.cpp"
   if (!($61)) {
    $$01 = $58;$$02 = $59;
    ___resumeException($$01|0);
    // unreachable;
   }
   $62 = (___cxa_begin_catch(($58|0))|0); //@line 158 "modules/filesystem/wrap_File.cpp"
   $63 = HEAP32[$62>>2]|0; //@line 161 "modules/filesystem/wrap_File.cpp"
   $64 = ((($63)) + 8|0); //@line 161 "modules/filesystem/wrap_File.cpp"
   $65 = HEAP32[$64>>2]|0; //@line 161 "modules/filesystem/wrap_File.cpp"
   $66 = (FUNCTION_TABLE_ii[$65 & 8191]($62)|0); //@line 161 "modules/filesystem/wrap_File.cpp"
   __THREW__ = 0;
   HEAP32[$vararg_buffer1>>2] = $66; //@line 161 "modules/filesystem/wrap_File.cpp"
   (invoke_iiii(3512,($L|0),(595867|0),($vararg_buffer1|0))|0); //@line 161 "modules/filesystem/wrap_File.cpp"
   $67 = __THREW__; __THREW__ = 0;
   $68 = $67&1;
   if (!($68)) {
    ___cxa_end_catch(); //@line 162 "modules/filesystem/wrap_File.cpp"
    $$0 = 2;
    STACKTOP = sp;return ($$0|0); //@line 171 "modules/filesystem/wrap_File.cpp"
   }
   $69 = ___cxa_find_matching_catch_2()|0;
   $70 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 162 "modules/filesystem/wrap_File.cpp"
   $71 = __THREW__; __THREW__ = 0;
   $72 = $71&1;
   if ($72) {
    $74 = ___cxa_find_matching_catch_3(0|0)|0;
    $75 = tempRet0;
    ___clang_call_terminate($74); //@line 150 "modules/filesystem/wrap_File.cpp"
    // unreachable; //@line 150 "modules/filesystem/wrap_File.cpp"
   } else {
    $$01 = $69;$$02 = $70;
    ___resumeException($$01|0);
    // unreachable;
   }
  } else {
   HEAP32[$datasize>>2] = 0; //@line 139 "modules/filesystem/wrap_File.cpp"
   __THREW__ = 0;
   $3 = (invoke_iiii(3392,($L|0),2,($datasize|0))|0); //@line 140 "modules/filesystem/wrap_File.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   do {
    if (!($5)) {
     __THREW__ = 0;
     $6 = (invoke_iii(3358,($L|0),3)|0); //@line 142 "modules/filesystem/wrap_File.cpp"
     $7 = __THREW__; __THREW__ = 0;
     $8 = $7&1;
     if (!($8)) {
      $9 = ($6|0)<(1); //@line 142 "modules/filesystem/wrap_File.cpp"
      if ($9) {
       $$pre = HEAP32[$datasize>>2]|0; //@line 145 "modules/filesystem/wrap_File.cpp"
       $27 = $$pre;
      } else {
       __THREW__ = 0;
       $10 = (invoke_iii(3596,($L|0),3)|0); //@line 143 "modules/filesystem/wrap_File.cpp"
       $11 = __THREW__; __THREW__ = 0;
       $12 = $11&1;
       if ($12) {
        break;
       }
       HEAP32[$datasize>>2] = $10; //@line 143 "modules/filesystem/wrap_File.cpp"
       $27 = $10;
      }
      $24 = HEAP32[$0>>2]|0; //@line 145 "modules/filesystem/wrap_File.cpp"
      $25 = ((($24)) + 40|0); //@line 145 "modules/filesystem/wrap_File.cpp"
      $26 = HEAP32[$25>>2]|0; //@line 145 "modules/filesystem/wrap_File.cpp"
      __THREW__ = 0;
      $28 = (invoke_iiiii($26|0,($0|0),($3|0),($27|0),0)|0); //@line 145 "modules/filesystem/wrap_File.cpp"
      $29 = __THREW__; __THREW__ = 0;
      $30 = $29&1;
      if (!($30)) {
       $result$0$in = $28;
       break L1;
      }
     }
    }
   } while(0);
   $13 = ___cxa_find_matching_catch_3(96|0)|0;
   $14 = tempRet0;
   $15 = (_llvm_eh_typeid_for((96|0))|0); //@line 146 "modules/filesystem/wrap_File.cpp"
   $16 = ($14|0)==($15|0); //@line 146 "modules/filesystem/wrap_File.cpp"
   if (!($16)) {
    $$01 = $13;$$02 = $14;
    ___resumeException($$01|0);
    // unreachable;
   }
   $17 = (___cxa_begin_catch(($13|0))|0); //@line 146 "modules/filesystem/wrap_File.cpp"
   $18 = HEAP32[$17>>2]|0; //@line 149 "modules/filesystem/wrap_File.cpp"
   $19 = ((($18)) + 8|0); //@line 149 "modules/filesystem/wrap_File.cpp"
   $20 = HEAP32[$19>>2]|0; //@line 149 "modules/filesystem/wrap_File.cpp"
   $21 = (FUNCTION_TABLE_ii[$20 & 8191]($17)|0); //@line 149 "modules/filesystem/wrap_File.cpp"
   __THREW__ = 0;
   HEAP32[$vararg_buffer>>2] = $21; //@line 149 "modules/filesystem/wrap_File.cpp"
   (invoke_iiii(3512,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 149 "modules/filesystem/wrap_File.cpp"
   $22 = __THREW__; __THREW__ = 0;
   $23 = $22&1;
   if (!($23)) {
    ___cxa_end_catch(); //@line 150 "modules/filesystem/wrap_File.cpp"
    $$0 = 2;
    STACKTOP = sp;return ($$0|0); //@line 171 "modules/filesystem/wrap_File.cpp"
   }
   $31 = ___cxa_find_matching_catch_2()|0;
   $32 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 150 "modules/filesystem/wrap_File.cpp"
   $33 = __THREW__; __THREW__ = 0;
   $34 = $33&1;
   if ($34) {
    $74 = ___cxa_find_matching_catch_3(0|0)|0;
    $75 = tempRet0;
    ___clang_call_terminate($74); //@line 150 "modules/filesystem/wrap_File.cpp"
    // unreachable; //@line 150 "modules/filesystem/wrap_File.cpp"
   } else {
    $$01 = $31;$$02 = $32;
    ___resumeException($$01|0);
    // unreachable;
   }
  }
 } while(0);
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$result$0$in); //@line 169 "modules/filesystem/wrap_File.cpp"
 $$0 = 1;
 STACKTOP = sp;return ($$0|0); //@line 171 "modules/filesystem/wrap_File.cpp"
}
function __ZN4love10filesystem12w_File_flushEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 173 "modules/filesystem/wrap_File.cpp"
 $0 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,1,5)|0); //@line 46 "modules/filesystem/wrap_File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 179 "modules/filesystem/wrap_File.cpp"
 $2 = ((($1)) + 48|0); //@line 179 "modules/filesystem/wrap_File.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 179 "modules/filesystem/wrap_File.cpp"
 __THREW__ = 0;
 $4 = (invoke_ii($3|0,($0|0))|0); //@line 179 "modules/filesystem/wrap_File.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 185 "modules/filesystem/wrap_File.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 187 "modules/filesystem/wrap_File.cpp"
 }
 $7 = ___cxa_find_matching_catch_3(96|0)|0;
 $8 = tempRet0;
 $9 = (_llvm_eh_typeid_for((96|0))|0); //@line 180 "modules/filesystem/wrap_File.cpp"
 $10 = ($8|0)==($9|0); //@line 180 "modules/filesystem/wrap_File.cpp"
 if (!($10)) {
  $$01 = $8;$$02 = $7;
  ___resumeException($$02|0);
  // unreachable;
 }
 $11 = (___cxa_begin_catch(($7|0))|0); //@line 180 "modules/filesystem/wrap_File.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 183 "modules/filesystem/wrap_File.cpp"
 $13 = ((($12)) + 8|0); //@line 183 "modules/filesystem/wrap_File.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 183 "modules/filesystem/wrap_File.cpp"
 $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0); //@line 183 "modules/filesystem/wrap_File.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $15; //@line 183 "modules/filesystem/wrap_File.cpp"
 (invoke_iiii(3512,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 183 "modules/filesystem/wrap_File.cpp"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if (!($17)) {
  ___cxa_end_catch(); //@line 184 "modules/filesystem/wrap_File.cpp"
  $$0 = 2;
  STACKTOP = sp;return ($$0|0); //@line 187 "modules/filesystem/wrap_File.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 184 "modules/filesystem/wrap_File.cpp"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 184 "modules/filesystem/wrap_File.cpp"
  // unreachable; //@line 184 "modules/filesystem/wrap_File.cpp"
 } else {
  $$01 = $19;$$02 = $18;
  ___resumeException($$02|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love10filesystem12w_File_isEOFEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,1,5)|0); //@line 46 "modules/filesystem/wrap_File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 192 "modules/filesystem/wrap_File.cpp"
 $2 = ((($1)) + 52|0); //@line 192 "modules/filesystem/wrap_File.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 192 "modules/filesystem/wrap_File.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 192 "modules/filesystem/wrap_File.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 192 "modules/filesystem/wrap_File.cpp"
 return 1; //@line 193 "modules/filesystem/wrap_File.cpp"
}
function __ZN4love10filesystem11w_File_tellEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 196 "modules/filesystem/wrap_File.cpp"
 $vararg_buffer = sp; //@line 196 "modules/filesystem/wrap_File.cpp"
 $0 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,1,5)|0); //@line 46 "modules/filesystem/wrap_File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 199 "modules/filesystem/wrap_File.cpp"
 $2 = ((($1)) + 56|0); //@line 199 "modules/filesystem/wrap_File.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 199 "modules/filesystem/wrap_File.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 199 "modules/filesystem/wrap_File.cpp"
 $5 = tempRet0; //@line 199 "modules/filesystem/wrap_File.cpp"
 $6 = ($4|0)==(-1); //@line 201 "modules/filesystem/wrap_File.cpp"
 $7 = ($5|0)==(-1); //@line 201 "modules/filesystem/wrap_File.cpp"
 $8 = $6 & $7; //@line 201 "modules/filesystem/wrap_File.cpp"
 if ($8) {
  (__ZN4love10filesystem12luax_ioErrorEP9lua_StatePKcz($L,366475,$vararg_buffer)|0); //@line 202 "modules/filesystem/wrap_File.cpp"
  $$0 = 2;
  STACKTOP = sp;return ($$0|0); //@line 208 "modules/filesystem/wrap_File.cpp"
 }
 $9 = ($5|0)>(2097151); //@line 203 "modules/filesystem/wrap_File.cpp"
 $10 = ($4>>>0)>(4294967295); //@line 203 "modules/filesystem/wrap_File.cpp"
 $11 = ($5|0)==(2097151); //@line 203 "modules/filesystem/wrap_File.cpp"
 $12 = $11 & $10; //@line 203 "modules/filesystem/wrap_File.cpp"
 $13 = $9 | $12; //@line 203 "modules/filesystem/wrap_File.cpp"
 if ($13) {
  (__ZN4love10filesystem12luax_ioErrorEP9lua_StatePKcz($L,366493,$vararg_buffer1)|0); //@line 204 "modules/filesystem/wrap_File.cpp"
  $$0 = 2;
  STACKTOP = sp;return ($$0|0); //@line 208 "modules/filesystem/wrap_File.cpp"
 } else {
  $14 = (+($4>>>0)) + (4294967296.0*(+($5|0))); //@line 206 "modules/filesystem/wrap_File.cpp"
  _lua_pushnumber($L,$14); //@line 206 "modules/filesystem/wrap_File.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 208 "modules/filesystem/wrap_File.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem11w_File_seekEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,1,5)|0); //@line 46 "modules/filesystem/wrap_File.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 213 "modules/filesystem/wrap_File.cpp"
 $2 = $1 < 0.0; //@line 217 "modules/filesystem/wrap_File.cpp"
 $3 = $1 >= 9007199254740992.0; //@line 217 "modules/filesystem/wrap_File.cpp"
 $or$cond = $2 | $3; //@line 217 "modules/filesystem/wrap_File.cpp"
 if ($or$cond) {
  __ZN4love16luax_pushbooleanEP9lua_Stateb($L,0); //@line 218 "modules/filesystem/wrap_File.cpp"
  return 1; //@line 221 "modules/filesystem/wrap_File.cpp"
 } else {
  $4 = HEAP32[$0>>2]|0; //@line 220 "modules/filesystem/wrap_File.cpp"
  $5 = ((($4)) + 60|0); //@line 220 "modules/filesystem/wrap_File.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 220 "modules/filesystem/wrap_File.cpp"
  $7 = (~~$1)>>>0; //@line 220 "modules/filesystem/wrap_File.cpp"
  $8 = +Math_abs($1) >= 1.0 ? $1 > 0.0 ? (~~+Math_min(+Math_floor($1 / 4294967296.0), 4294967295.0)) >>> 0 : ~~+Math_ceil(($1 - +(~~$1 >>> 0)) / 4294967296.0) >>> 0 : 0; //@line 220 "modules/filesystem/wrap_File.cpp"
  $9 = (FUNCTION_TABLE_iiii[$6 & 8191]($0,$7,$8)|0); //@line 220 "modules/filesystem/wrap_File.cpp"
  __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$9); //@line 220 "modules/filesystem/wrap_File.cpp"
  return 1; //@line 221 "modules/filesystem/wrap_File.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem12w_File_linesEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 326 "modules/filesystem/wrap_File.cpp"
 $vararg_buffer = sp; //@line 326 "modules/filesystem/wrap_File.cpp"
 $0 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,1,5)|0); //@line 46 "modules/filesystem/wrap_File.cpp"
 _lua_pushnumber($L,0.0); //@line 330 "modules/filesystem/wrap_File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 331 "modules/filesystem/wrap_File.cpp"
 $2 = ((($1)) + 72|0); //@line 331 "modules/filesystem/wrap_File.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 331 "modules/filesystem/wrap_File.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 331 "modules/filesystem/wrap_File.cpp"
 $5 = ($4|0)!=(0); //@line 331 "modules/filesystem/wrap_File.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$5); //@line 331 "modules/filesystem/wrap_File.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 333 "modules/filesystem/wrap_File.cpp"
 $7 = ((($6)) + 72|0); //@line 333 "modules/filesystem/wrap_File.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 333 "modules/filesystem/wrap_File.cpp"
 $9 = (FUNCTION_TABLE_ii[$8 & 8191]($0)|0); //@line 333 "modules/filesystem/wrap_File.cpp"
 $10 = ($9|0)==(1); //@line 333 "modules/filesystem/wrap_File.cpp"
 L1: do {
  if (!($10)) {
   $11 = HEAP32[$0>>2]|0; //@line 335 "modules/filesystem/wrap_File.cpp"
   $12 = ((($11)) + 72|0); //@line 335 "modules/filesystem/wrap_File.cpp"
   $13 = HEAP32[$12>>2]|0; //@line 335 "modules/filesystem/wrap_File.cpp"
   $14 = (FUNCTION_TABLE_ii[$13 & 8191]($0)|0); //@line 335 "modules/filesystem/wrap_File.cpp"
   $15 = ($14|0)==(0); //@line 335 "modules/filesystem/wrap_File.cpp"
   if (!($15)) {
    $16 = HEAP32[$0>>2]|0; //@line 336 "modules/filesystem/wrap_File.cpp"
    $17 = ((($16)) + 20|0); //@line 336 "modules/filesystem/wrap_File.cpp"
    $18 = HEAP32[$17>>2]|0; //@line 336 "modules/filesystem/wrap_File.cpp"
    (FUNCTION_TABLE_ii[$18 & 8191]($0)|0); //@line 336 "modules/filesystem/wrap_File.cpp"
   }
   $19 = HEAP32[$0>>2]|0; //@line 339 "modules/filesystem/wrap_File.cpp"
   $20 = ((($19)) + 16|0); //@line 339 "modules/filesystem/wrap_File.cpp"
   $21 = HEAP32[$20>>2]|0; //@line 339 "modules/filesystem/wrap_File.cpp"
   __THREW__ = 0;
   $22 = (invoke_iii($21|0,($0|0),1)|0); //@line 339 "modules/filesystem/wrap_File.cpp"
   $23 = __THREW__; __THREW__ = 0;
   $24 = $23&1;
   do {
    if ($24) {
     $25 = ___cxa_find_matching_catch_3(8992|0)|0;
     $26 = tempRet0;
     $27 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
     $28 = ($26|0)==($27|0); //@line 535 "./common/runtime.h"
     if (!($28)) {
      $$01$i = $26;$$02$i = $25;
      ___resumeException($$02$i|0);
      // unreachable;
     }
     $29 = (___cxa_begin_catch(($25|0))|0); //@line 535 "./common/runtime.h"
     $30 = HEAP32[$29>>2]|0; //@line 539 "./common/runtime.h"
     $31 = ((($30)) + 8|0); //@line 539 "./common/runtime.h"
     $32 = HEAP32[$31>>2]|0; //@line 539 "./common/runtime.h"
     $33 = (FUNCTION_TABLE_ii[$32 & 8191]($29)|0); //@line 539 "./common/runtime.h"
     __THREW__ = 0;
     invoke_vii(3380,($L|0),($33|0)); //@line 539 "./common/runtime.h"
     $34 = __THREW__; __THREW__ = 0;
     $35 = $34&1;
     if (!($35)) {
      ___cxa_end_catch(); //@line 540 "./common/runtime.h"
      $36 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
      HEAP32[$vararg_buffer>>2] = $36; //@line 543 "./common/runtime.h"
      (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
      break;
     }
     $37 = ___cxa_find_matching_catch_2()|0;
     $38 = tempRet0;
     __THREW__ = 0;
     invoke_v(3378); //@line 540 "./common/runtime.h"
     $39 = __THREW__; __THREW__ = 0;
     $40 = $39&1;
     if ($40) {
      $41 = ___cxa_find_matching_catch_3(0|0)|0;
      $42 = tempRet0;
      ___clang_call_terminate($41); //@line 540 "./common/runtime.h"
      // unreachable; //@line 540 "./common/runtime.h"
     } else {
      $$01$i = $38;$$02$i = $37;
      ___resumeException($$02$i|0);
      // unreachable;
     }
    } else {
     if ($22) {
      break L1;
     }
    }
   } while(0);
   $43 = (_luaL_error($L,366514,$vararg_buffer1)|0); //@line 342 "modules/filesystem/wrap_File.cpp"
   $$1 = $43;
   STACKTOP = sp;return ($$1|0); //@line 347 "modules/filesystem/wrap_File.cpp"
  }
 } while(0);
 _lua_pushcclosure($L,3513,3); //@line 345 "modules/filesystem/wrap_File.cpp"
 $$1 = 1;
 STACKTOP = sp;return ($$1|0); //@line 347 "modules/filesystem/wrap_File.cpp"
}
function __ZN4love10filesystem14w_File_lines_iEP9lua_State($L) {
 $L = $L|0;
 var $$4 = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0.0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $13 = 0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $buf = 0, $i$020 = 0, $i$020$lcssa = 0, $linesize$016 = 0, $linesize$026$lcssa = 0, $linesize$02636 = 0, $linesize$02636$lcssa43 = 0, $linesize$1 = 0, $linesize$129 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0;
 var $vararg_buffer5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1056|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $buf = sp + 32|0;
 $0 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,-10003,5)|0); //@line 231 "modules/filesystem/wrap_File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 234 "modules/filesystem/wrap_File.cpp"
 $2 = ((($1)) + 72|0); //@line 234 "modules/filesystem/wrap_File.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 234 "modules/filesystem/wrap_File.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 234 "modules/filesystem/wrap_File.cpp"
 $5 = ($4|0)==(1); //@line 234 "modules/filesystem/wrap_File.cpp"
 if (!($5)) {
  $6 = (_luaL_error($L,366535,$vararg_buffer)|0); //@line 235 "modules/filesystem/wrap_File.cpp"
  $$4 = $6;
  STACKTOP = sp;return ($$4|0); //@line 324 "modules/filesystem/wrap_File.cpp"
 }
 $7 = HEAP32[$0>>2]|0; //@line 237 "modules/filesystem/wrap_File.cpp"
 $8 = ((($7)) + 56|0); //@line 237 "modules/filesystem/wrap_File.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 237 "modules/filesystem/wrap_File.cpp"
 $10 = (FUNCTION_TABLE_ii[$9 & 8191]($0)|0); //@line 237 "modules/filesystem/wrap_File.cpp"
 $11 = tempRet0; //@line 237 "modules/filesystem/wrap_File.cpp"
 $12 = (_lua_type($L,-10004)|0); //@line 240 "modules/filesystem/wrap_File.cpp"
 $13 = ($12|0)<(1); //@line 240 "modules/filesystem/wrap_File.cpp"
 if ($13) {
  $57 = -1;$63 = $10;$64 = $11;$99 = -1;
 } else {
  $14 = (+_lua_tonumber($L,-10004)); //@line 244 "modules/filesystem/wrap_File.cpp"
  $15 = (~~$14)>>>0; //@line 244 "modules/filesystem/wrap_File.cpp"
  $16 = +Math_abs($14) >= 1.0 ? $14 > 0.0 ? (~~+Math_min(+Math_floor($14 / 4294967296.0), 4294967295.0)) >>> 0 : ~~+Math_ceil(($14 - +(~~$14 >>> 0)) / 4294967296.0) >>> 0 : 0; //@line 244 "modules/filesystem/wrap_File.cpp"
  $17 = ($10|0)==($15|0); //@line 245 "modules/filesystem/wrap_File.cpp"
  $18 = ($11|0)==($16|0); //@line 245 "modules/filesystem/wrap_File.cpp"
  $19 = $17 & $18; //@line 245 "modules/filesystem/wrap_File.cpp"
  if ($19) {
   $57 = $11;$63 = $10;$64 = $11;$99 = $10;
  } else {
   $20 = HEAP32[$0>>2]|0; //@line 246 "modules/filesystem/wrap_File.cpp"
   $21 = ((($20)) + 60|0); //@line 246 "modules/filesystem/wrap_File.cpp"
   $22 = HEAP32[$21>>2]|0; //@line 246 "modules/filesystem/wrap_File.cpp"
   (FUNCTION_TABLE_iiii[$22 & 8191]($0,$15,$16)|0); //@line 246 "modules/filesystem/wrap_File.cpp"
   $57 = $11;$63 = $15;$64 = $16;$99 = $10;
  }
 }
 $23 = HEAP32[$0>>2]|0; //@line 249 "modules/filesystem/wrap_File.cpp"
 $24 = ((($23)) + 52|0); //@line 249 "modules/filesystem/wrap_File.cpp"
 $25 = HEAP32[$24>>2]|0; //@line 249 "modules/filesystem/wrap_File.cpp"
 $26 = (FUNCTION_TABLE_ii[$25 & 8191]($0)|0); //@line 249 "modules/filesystem/wrap_File.cpp"
 L9: do {
  if ($26) {
   $linesize$026$lcssa = 0;
   label = 13;
  } else {
   $linesize$02636 = 0;
   L10: while(1) {
    $27 = HEAP32[$0>>2]|0; //@line 252 "modules/filesystem/wrap_File.cpp"
    $28 = ((($27)) + 36|0); //@line 252 "modules/filesystem/wrap_File.cpp"
    $29 = HEAP32[$28>>2]|0; //@line 252 "modules/filesystem/wrap_File.cpp"
    $30 = (FUNCTION_TABLE_iiiii[$29 & 8191]($0,$buf,1024,0)|0); //@line 252 "modules/filesystem/wrap_File.cpp"
    $31 = tempRet0; //@line 252 "modules/filesystem/wrap_File.cpp"
    $32 = ($30|0)<(0); //@line 253 "modules/filesystem/wrap_File.cpp"
    if ($32) {
     break;
    }
    $33 = ($30|0)>(0); //@line 258 "modules/filesystem/wrap_File.cpp"
    if ($33) {
     $i$020 = 0;
     while(1) {
      $34 = (($buf) + ($i$020)|0); //@line 260 "modules/filesystem/wrap_File.cpp"
      $35 = HEAP8[$34>>0]|0; //@line 260 "modules/filesystem/wrap_File.cpp"
      $36 = ($35<<24>>24)==(10); //@line 260 "modules/filesystem/wrap_File.cpp"
      if ($36) {
       $i$020$lcssa = $i$020;$linesize$02636$lcssa43 = $linesize$02636;
       label = 14;
       break L10;
      }
      $37 = (($i$020) + 1)|0; //@line 258 "modules/filesystem/wrap_File.cpp"
      $38 = ($30|0)>($37|0); //@line 258 "modules/filesystem/wrap_File.cpp"
      if ($38) {
       $i$020 = $37;
      } else {
       break;
      }
     }
    }
    $linesize$129 = (($30) + ($linesize$02636))|0; //@line 262 "modules/filesystem/wrap_File.cpp"
    $39 = HEAP32[$0>>2]|0; //@line 249 "modules/filesystem/wrap_File.cpp"
    $40 = ((($39)) + 52|0); //@line 249 "modules/filesystem/wrap_File.cpp"
    $41 = HEAP32[$40>>2]|0; //@line 249 "modules/filesystem/wrap_File.cpp"
    $42 = (FUNCTION_TABLE_ii[$41 & 8191]($0)|0); //@line 249 "modules/filesystem/wrap_File.cpp"
    if ($42) {
     $linesize$026$lcssa = $linesize$129;
     label = 13;
     break L9;
    } else {
     $linesize$02636 = $linesize$129;
    }
   }
   if ((label|0) == 14) {
    $linesize$1 = (($i$020$lcssa) + ($linesize$02636$lcssa43))|0; //@line 262 "modules/filesystem/wrap_File.cpp"
    $linesize$016 = $linesize$1;
    break;
   }
   $43 = (_luaL_error($L,366568,$vararg_buffer1)|0); //@line 254 "modules/filesystem/wrap_File.cpp"
   $$4 = $43;
   STACKTOP = sp;return ($$4|0); //@line 324 "modules/filesystem/wrap_File.cpp"
  }
 } while(0);
 if ((label|0) == 13) {
  $44 = HEAP32[$0>>2]|0; //@line 269 "modules/filesystem/wrap_File.cpp"
  $45 = ((($44)) + 52|0); //@line 269 "modules/filesystem/wrap_File.cpp"
  $46 = HEAP32[$45>>2]|0; //@line 269 "modules/filesystem/wrap_File.cpp"
  $47 = (FUNCTION_TABLE_ii[$46 & 8191]($0)|0); //@line 269 "modules/filesystem/wrap_File.cpp"
  $48 = ($linesize$026$lcssa|0)>(0); //@line 269 "modules/filesystem/wrap_File.cpp"
  $or$cond = $48 & $47; //@line 269 "modules/filesystem/wrap_File.cpp"
  if ($or$cond) {
   $linesize$016 = $linesize$026$lcssa;
  } else {
   $113 = ($57|0)>(-1); //@line 318 "modules/filesystem/wrap_File.cpp"
   $114 = ($99>>>0)>(4294967295); //@line 318 "modules/filesystem/wrap_File.cpp"
   $115 = ($57|0)==(-1); //@line 318 "modules/filesystem/wrap_File.cpp"
   $116 = $115 & $114; //@line 318 "modules/filesystem/wrap_File.cpp"
   $117 = $113 | $116; //@line 318 "modules/filesystem/wrap_File.cpp"
   if ($117) {
    $118 = (__ZN4love14luax_tobooleanEP9lua_Statei($L,-10005)|0); //@line 318 "modules/filesystem/wrap_File.cpp"
    if ($118) {
     $119 = HEAP32[$0>>2]|0; //@line 319 "modules/filesystem/wrap_File.cpp"
     $120 = ((($119)) + 60|0); //@line 319 "modules/filesystem/wrap_File.cpp"
     $121 = HEAP32[$120>>2]|0; //@line 319 "modules/filesystem/wrap_File.cpp"
     (FUNCTION_TABLE_iiii[$121 & 8191]($0,$99,$57)|0); //@line 319 "modules/filesystem/wrap_File.cpp"
     $$4 = 0;
     STACKTOP = sp;return ($$4|0); //@line 324 "modules/filesystem/wrap_File.cpp"
    }
   }
   $122 = HEAP32[$0>>2]|0; //@line 321 "modules/filesystem/wrap_File.cpp"
   $123 = ((($122)) + 20|0); //@line 321 "modules/filesystem/wrap_File.cpp"
   $124 = HEAP32[$123>>2]|0; //@line 321 "modules/filesystem/wrap_File.cpp"
   (FUNCTION_TABLE_ii[$124 & 8191]($0)|0); //@line 321 "modules/filesystem/wrap_File.cpp"
   $$4 = 0;
   STACKTOP = sp;return ($$4|0); //@line 324 "modules/filesystem/wrap_File.cpp"
  }
 }
 $49 = ($linesize$016|0)<(1024); //@line 271 "modules/filesystem/wrap_File.cpp"
 do {
  if ($49) {
   $50 = ($linesize$016|0)>(0); //@line 274 "modules/filesystem/wrap_File.cpp"
   if ($50) {
    $51 = (($linesize$016) + -1)|0; //@line 274 "modules/filesystem/wrap_File.cpp"
    $52 = (($buf) + ($51)|0); //@line 274 "modules/filesystem/wrap_File.cpp"
    $53 = HEAP8[$52>>0]|0; //@line 274 "modules/filesystem/wrap_File.cpp"
    $54 = ($53<<24>>24)==(13); //@line 274 "modules/filesystem/wrap_File.cpp"
    if ($54) {
     $55 = $51;
    } else {
     label = 18;
    }
   } else {
    label = 18;
   }
   if ((label|0) == 18) {
    $55 = $linesize$016;
   }
   _lua_pushlstring($L,$buf,$55); //@line 274 "modules/filesystem/wrap_File.cpp"
   $56 = ($57|0)<(0); //@line 275 "modules/filesystem/wrap_File.cpp"
   if ($56) {
    $58 = HEAP32[$0>>2]|0; //@line 276 "modules/filesystem/wrap_File.cpp"
    $59 = ((($58)) + 60|0); //@line 276 "modules/filesystem/wrap_File.cpp"
    $60 = HEAP32[$59>>2]|0; //@line 276 "modules/filesystem/wrap_File.cpp"
    $61 = ($linesize$016|0)<(0); //@line 276 "modules/filesystem/wrap_File.cpp"
    $62 = $61 << 31 >> 31; //@line 276 "modules/filesystem/wrap_File.cpp"
    $65 = (_i64Add(($63|0),($64|0),1,0)|0); //@line 276 "modules/filesystem/wrap_File.cpp"
    $66 = tempRet0; //@line 276 "modules/filesystem/wrap_File.cpp"
    $67 = (_i64Add(($65|0),($66|0),($linesize$016|0),($62|0))|0); //@line 276 "modules/filesystem/wrap_File.cpp"
    $68 = tempRet0; //@line 276 "modules/filesystem/wrap_File.cpp"
    (FUNCTION_TABLE_iiii[$60 & 8191]($0,$67,$68)|0); //@line 276 "modules/filesystem/wrap_File.cpp"
    $$4 = 1;
    STACKTOP = sp;return ($$4|0); //@line 324 "modules/filesystem/wrap_File.cpp"
   } else {
    $$pre$phiZ2D = $0;
   }
  } else {
   $69 = (($linesize$016) + 1)|0; //@line 283 "modules/filesystem/wrap_File.cpp"
   __THREW__ = 0;
   $70 = (invoke_ii(3492,($69|0))|0); //@line 283 "modules/filesystem/wrap_File.cpp"
   $71 = __THREW__; __THREW__ = 0;
   $72 = $71&1;
   if ($72) {
    $73 = ___cxa_find_matching_catch_3(9056|0)|0;
    $74 = tempRet0;
    $75 = (_llvm_eh_typeid_for((9056|0))|0); //@line 284 "modules/filesystem/wrap_File.cpp"
    $76 = ($74|0)==($75|0); //@line 284 "modules/filesystem/wrap_File.cpp"
    if (!($76)) {
     ___resumeException($73|0);
     // unreachable;
    }
    (___cxa_begin_catch(($73|0))|0); //@line 284 "modules/filesystem/wrap_File.cpp"
    ___cxa_end_catch(); //@line 288 "modules/filesystem/wrap_File.cpp"
    $77 = (_luaL_error($L,507610,$vararg_buffer3)|0); //@line 291 "modules/filesystem/wrap_File.cpp"
    $$4 = $77;
    STACKTOP = sp;return ($$4|0); //@line 324 "modules/filesystem/wrap_File.cpp"
   }
   $78 = HEAP32[$0>>2]|0; //@line 293 "modules/filesystem/wrap_File.cpp"
   $79 = ((($78)) + 60|0); //@line 293 "modules/filesystem/wrap_File.cpp"
   $80 = HEAP32[$79>>2]|0; //@line 293 "modules/filesystem/wrap_File.cpp"
   (FUNCTION_TABLE_iiii[$80 & 8191]($0,$63,$64)|0); //@line 293 "modules/filesystem/wrap_File.cpp"
   $81 = HEAP32[$0>>2]|0; //@line 296 "modules/filesystem/wrap_File.cpp"
   $82 = ((($81)) + 36|0); //@line 296 "modules/filesystem/wrap_File.cpp"
   $83 = HEAP32[$82>>2]|0; //@line 296 "modules/filesystem/wrap_File.cpp"
   $84 = ($69|0)<(0); //@line 296 "modules/filesystem/wrap_File.cpp"
   $85 = $84 << 31 >> 31; //@line 296 "modules/filesystem/wrap_File.cpp"
   $86 = (FUNCTION_TABLE_iiiii[$83 & 8191]($0,$70,$69,$85)|0); //@line 296 "modules/filesystem/wrap_File.cpp"
   $87 = tempRet0; //@line 296 "modules/filesystem/wrap_File.cpp"
   $88 = ($86|0)==(-1); //@line 296 "modules/filesystem/wrap_File.cpp"
   $89 = ($87|0)==(-1); //@line 296 "modules/filesystem/wrap_File.cpp"
   $90 = $88 & $89; //@line 296 "modules/filesystem/wrap_File.cpp"
   if ($90) {
    __ZdaPv($70); //@line 298 "modules/filesystem/wrap_File.cpp"
    $91 = (_luaL_error($L,366568,$vararg_buffer5)|0); //@line 299 "modules/filesystem/wrap_File.cpp"
    $$4 = $91;
    STACKTOP = sp;return ($$4|0); //@line 324 "modules/filesystem/wrap_File.cpp"
   } else {
    $92 = (($linesize$016) + -1)|0; //@line 302 "modules/filesystem/wrap_File.cpp"
    $93 = (($70) + ($92)|0); //@line 302 "modules/filesystem/wrap_File.cpp"
    $94 = HEAP8[$93>>0]|0; //@line 302 "modules/filesystem/wrap_File.cpp"
    $95 = ($94<<24>>24)==(13); //@line 302 "modules/filesystem/wrap_File.cpp"
    $96 = $95 ? $92 : $linesize$016; //@line 302 "modules/filesystem/wrap_File.cpp"
    _lua_pushlstring($L,$70,$96); //@line 302 "modules/filesystem/wrap_File.cpp"
    __ZdaPv($70); //@line 303 "modules/filesystem/wrap_File.cpp"
    $97 = ($57|0)>(-1); //@line 306 "modules/filesystem/wrap_File.cpp"
    $98 = ($99>>>0)>(4294967295); //@line 306 "modules/filesystem/wrap_File.cpp"
    $100 = ($57|0)==(-1); //@line 306 "modules/filesystem/wrap_File.cpp"
    $101 = $100 & $98; //@line 306 "modules/filesystem/wrap_File.cpp"
    $102 = $97 | $101; //@line 306 "modules/filesystem/wrap_File.cpp"
    if ($102) {
     $$pre$phiZ2D = $0;
     break;
    } else {
     $$4 = 1;
    }
    STACKTOP = sp;return ($$4|0); //@line 324 "modules/filesystem/wrap_File.cpp"
   }
  }
 } while(0);
 $103 = ($linesize$016|0)<(0); //@line 309 "modules/filesystem/wrap_File.cpp"
 $104 = $103 << 31 >> 31; //@line 309 "modules/filesystem/wrap_File.cpp"
 $105 = (_i64Add(($63|0),($64|0),1,0)|0); //@line 309 "modules/filesystem/wrap_File.cpp"
 $106 = tempRet0; //@line 309 "modules/filesystem/wrap_File.cpp"
 $107 = (_i64Add(($105|0),($106|0),($linesize$016|0),($104|0))|0); //@line 309 "modules/filesystem/wrap_File.cpp"
 $108 = tempRet0; //@line 309 "modules/filesystem/wrap_File.cpp"
 $109 = (+($107>>>0)) + (4294967296.0*(+($108|0))); //@line 309 "modules/filesystem/wrap_File.cpp"
 _lua_pushnumber($L,$109); //@line 309 "modules/filesystem/wrap_File.cpp"
 _lua_replace($L,-10004); //@line 310 "modules/filesystem/wrap_File.cpp"
 $110 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 311 "modules/filesystem/wrap_File.cpp"
 $111 = ((($110)) + 60|0); //@line 311 "modules/filesystem/wrap_File.cpp"
 $112 = HEAP32[$111>>2]|0; //@line 311 "modules/filesystem/wrap_File.cpp"
 (FUNCTION_TABLE_iiii[$112 & 8191]($0,$99,$57)|0); //@line 311 "modules/filesystem/wrap_File.cpp"
 $$4 = 1;
 STACKTOP = sp;return ($$4|0); //@line 324 "modules/filesystem/wrap_File.cpp"
}
function __ZN4love10filesystem16w_File_setBufferEP9lua_State($L) {
 $L = $L|0;
 var $$01 = 0, $$02 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $bufmode = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $bufmode = sp + 12|0;
 $0 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,1,5)|0); //@line 46 "modules/filesystem/wrap_File.cpp"
 $1 = (_luaL_checklstring($L,2,0)|0); //@line 352 "modules/filesystem/wrap_File.cpp"
 $2 = (+_luaL_optnumber($L,3,0.0)); //@line 353 "modules/filesystem/wrap_File.cpp"
 $3 = (__ZN4love10filesystem4File11getConstantEPKcRNS1_10BufferModeE($1,$bufmode)|0); //@line 356 "modules/filesystem/wrap_File.cpp"
 if (!($3)) {
  HEAP32[$vararg_buffer>>2] = $1; //@line 357 "modules/filesystem/wrap_File.cpp"
  $4 = (_luaL_error($L,366604,$vararg_buffer)|0); //@line 357 "modules/filesystem/wrap_File.cpp"
  $$1 = $4;
  STACKTOP = sp;return ($$1|0); //@line 371 "modules/filesystem/wrap_File.cpp"
 }
 $5 = (~~$2)>>>0; //@line 353 "modules/filesystem/wrap_File.cpp"
 $6 = +Math_abs($2) >= 1.0 ? $2 > 0.0 ? (~~+Math_min(+Math_floor($2 / 4294967296.0), 4294967295.0)) >>> 0 : ~~+Math_ceil(($2 - +(~~$2 >>> 0)) / 4294967296.0) >>> 0 : 0; //@line 353 "modules/filesystem/wrap_File.cpp"
 $7 = HEAP32[$0>>2]|0; //@line 362 "modules/filesystem/wrap_File.cpp"
 $8 = ((($7)) + 64|0); //@line 362 "modules/filesystem/wrap_File.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 362 "modules/filesystem/wrap_File.cpp"
 $10 = HEAP32[$bufmode>>2]|0; //@line 362 "modules/filesystem/wrap_File.cpp"
 __THREW__ = 0;
 $11 = (invoke_iiiii($9|0,($0|0),($10|0),($5|0),($6|0))|0); //@line 362 "modules/filesystem/wrap_File.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if (!($13)) {
  __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$11); //@line 369 "modules/filesystem/wrap_File.cpp"
  $$1 = 1;
  STACKTOP = sp;return ($$1|0); //@line 371 "modules/filesystem/wrap_File.cpp"
 }
 $14 = ___cxa_find_matching_catch_3(96|0)|0;
 $15 = tempRet0;
 $16 = (_llvm_eh_typeid_for((96|0))|0); //@line 363 "modules/filesystem/wrap_File.cpp"
 $17 = ($15|0)==($16|0); //@line 363 "modules/filesystem/wrap_File.cpp"
 if (!($17)) {
  $$01 = $15;$$02 = $14;
  ___resumeException($$02|0);
  // unreachable;
 }
 $18 = (___cxa_begin_catch(($14|0))|0); //@line 363 "modules/filesystem/wrap_File.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 366 "modules/filesystem/wrap_File.cpp"
 $20 = ((($19)) + 8|0); //@line 366 "modules/filesystem/wrap_File.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 366 "modules/filesystem/wrap_File.cpp"
 $22 = (FUNCTION_TABLE_ii[$21 & 8191]($18)|0); //@line 366 "modules/filesystem/wrap_File.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer1>>2] = $22; //@line 366 "modules/filesystem/wrap_File.cpp"
 (invoke_iiii(3512,($L|0),(595867|0),($vararg_buffer1|0))|0); //@line 366 "modules/filesystem/wrap_File.cpp"
 $23 = __THREW__; __THREW__ = 0;
 $24 = $23&1;
 if (!($24)) {
  ___cxa_end_catch(); //@line 367 "modules/filesystem/wrap_File.cpp"
  $$1 = 2;
  STACKTOP = sp;return ($$1|0); //@line 371 "modules/filesystem/wrap_File.cpp"
 }
 $25 = ___cxa_find_matching_catch_2()|0;
 $26 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 367 "modules/filesystem/wrap_File.cpp"
 $27 = __THREW__; __THREW__ = 0;
 $28 = $27&1;
 if ($28) {
  $29 = ___cxa_find_matching_catch_3(0|0)|0;
  $30 = tempRet0;
  ___clang_call_terminate($29); //@line 367 "modules/filesystem/wrap_File.cpp"
  // unreachable; //@line 367 "modules/filesystem/wrap_File.cpp"
 } else {
  $$01 = $26;$$02 = $25;
  ___resumeException($$02|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love10filesystem16w_File_getBufferEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $size = 0, $str = 0;
 var $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp + 8|0;
 $size = sp;
 $str = sp + 12|0;
 $0 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,1,5)|0); //@line 46 "modules/filesystem/wrap_File.cpp"
 $1 = $size; //@line 376 "modules/filesystem/wrap_File.cpp"
 $2 = $1; //@line 376 "modules/filesystem/wrap_File.cpp"
 HEAP32[$2>>2] = 0; //@line 376 "modules/filesystem/wrap_File.cpp"
 $3 = (($1) + 4)|0; //@line 376 "modules/filesystem/wrap_File.cpp"
 $4 = $3; //@line 376 "modules/filesystem/wrap_File.cpp"
 HEAP32[$4>>2] = 0; //@line 376 "modules/filesystem/wrap_File.cpp"
 $5 = HEAP32[$0>>2]|0; //@line 377 "modules/filesystem/wrap_File.cpp"
 $6 = ((($5)) + 68|0); //@line 377 "modules/filesystem/wrap_File.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 377 "modules/filesystem/wrap_File.cpp"
 $8 = (FUNCTION_TABLE_iii[$7 & 8191]($0,$size)|0); //@line 377 "modules/filesystem/wrap_File.cpp"
 HEAP32[$str>>2] = 0; //@line 378 "modules/filesystem/wrap_File.cpp"
 $9 = (__ZN4love10filesystem4File11getConstantENS1_10BufferModeERPKc($8,$str)|0); //@line 380 "modules/filesystem/wrap_File.cpp"
 if ($9) {
  $10 = HEAP32[$str>>2]|0; //@line 383 "modules/filesystem/wrap_File.cpp"
  _lua_pushstring($L,$10); //@line 383 "modules/filesystem/wrap_File.cpp"
  $11 = $size; //@line 384 "modules/filesystem/wrap_File.cpp"
  $12 = $11; //@line 384 "modules/filesystem/wrap_File.cpp"
  $13 = HEAP32[$12>>2]|0; //@line 384 "modules/filesystem/wrap_File.cpp"
  $14 = (($11) + 4)|0; //@line 384 "modules/filesystem/wrap_File.cpp"
  $15 = $14; //@line 384 "modules/filesystem/wrap_File.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 384 "modules/filesystem/wrap_File.cpp"
  $17 = (+($13>>>0)) + (4294967296.0*(+($16|0))); //@line 384 "modules/filesystem/wrap_File.cpp"
  _lua_pushnumber($L,$17); //@line 384 "modules/filesystem/wrap_File.cpp"
  STACKTOP = sp;return 2; //@line 386 "modules/filesystem/wrap_File.cpp"
 } else {
  (__ZN4love10filesystem12luax_ioErrorEP9lua_StatePKcz($L,366645,$vararg_buffer)|0); //@line 381 "modules/filesystem/wrap_File.cpp"
  STACKTOP = sp;return 2; //@line 386 "modules/filesystem/wrap_File.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem14w_File_getModeEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $str = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $str = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,1,5)|0); //@line 46 "modules/filesystem/wrap_File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 392 "modules/filesystem/wrap_File.cpp"
 $2 = ((($1)) + 72|0); //@line 392 "modules/filesystem/wrap_File.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 392 "modules/filesystem/wrap_File.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 392 "modules/filesystem/wrap_File.cpp"
 HEAP32[$str>>2] = 0; //@line 393 "modules/filesystem/wrap_File.cpp"
 $5 = (__ZN4love10filesystem4File11getConstantENS1_4ModeERPKc($4,$str)|0); //@line 395 "modules/filesystem/wrap_File.cpp"
 if ($5) {
  $6 = HEAP32[$str>>2]|0; //@line 398 "modules/filesystem/wrap_File.cpp"
  _lua_pushstring($L,$6); //@line 398 "modules/filesystem/wrap_File.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 400 "modules/filesystem/wrap_File.cpp"
 } else {
  (__ZN4love10filesystem12luax_ioErrorEP9lua_StatePKcz($L,366679,$vararg_buffer)|0); //@line 396 "modules/filesystem/wrap_File.cpp"
  $$0 = 2;
  STACKTOP = sp;return ($$0|0); //@line 400 "modules/filesystem/wrap_File.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem18w_File_getFilenameEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,1,5)|0); //@line 46 "modules/filesystem/wrap_File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 405 "modules/filesystem/wrap_File.cpp"
 $2 = ((($1)) + 76|0); //@line 405 "modules/filesystem/wrap_File.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 405 "modules/filesystem/wrap_File.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 405 "modules/filesystem/wrap_File.cpp"
 __ZN4love15luax_pushstringEP9lua_StateRKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE($L,$4); //@line 405 "modules/filesystem/wrap_File.cpp"
 return 1; //@line 406 "modules/filesystem/wrap_File.cpp"
}
function __ZN4love10filesystem19w_File_getExtensionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,1,5)|0); //@line 46 "modules/filesystem/wrap_File.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 412 "modules/filesystem/wrap_File.cpp"
 $3 = ((($2)) + 80|0); //@line 412 "modules/filesystem/wrap_File.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 412 "modules/filesystem/wrap_File.cpp"
 FUNCTION_TABLE_vii[$4 & 8191]($0,$1); //@line 412 "modules/filesystem/wrap_File.cpp"
 __THREW__ = 0;
 invoke_vii(3405,($L|0),($0|0)); //@line 412 "modules/filesystem/wrap_File.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 412 "modules/filesystem/wrap_File.cpp"
  ___resumeException($7|0);
  // unreachable;
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 412 "modules/filesystem/wrap_File.cpp"
  STACKTOP = sp;return 1; //@line 413 "modules/filesystem/wrap_File.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem14luax_checkfileEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_10filesystem4FileEEEPT_P9lua_StateiNS_4TypeE($L,$idx,5)|0); //@line 46 "modules/filesystem/wrap_File.cpp"
 return ($0|0); //@line 46 "modules/filesystem/wrap_File.cpp"
}
function _luaopen_file($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 437 "modules/filesystem/wrap_File.cpp"
 HEAP32[$vararg_buffer>>2] = 23700; //@line 439 "modules/filesystem/wrap_File.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 439 "modules/filesystem/wrap_File.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 439 "modules/filesystem/wrap_File.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,5,366698,$vararg_buffer)|0); //@line 439 "modules/filesystem/wrap_File.cpp"
 STACKTOP = sp;return ($0|0); //@line 439 "modules/filesystem/wrap_File.cpp"
}
function __ZN4love10filesystem18luax_checkfiledataEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_10filesystem8FileDataEEEPT_P9lua_StateiNS_4TypeE($L,$idx,7)|0); //@line 32 "modules/filesystem/wrap_FileData.cpp"
 return ($0|0); //@line 32 "modules/filesystem/wrap_FileData.cpp"
}
function __ZN4love10filesystem22w_FileData_getFilenameEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_10filesystem8FileDataEEEPT_P9lua_StateiNS_4TypeE($L,1,7)|0); //@line 32 "modules/filesystem/wrap_FileData.cpp"
 $1 = (__ZNK4love10filesystem8FileData11getFilenameEv($0)|0); //@line 38 "modules/filesystem/wrap_FileData.cpp"
 $2 = HEAP8[$1>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = $2 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ($3<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($1)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = HEAP32[$5>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($1)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $4 ? $7 : $6; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 _lua_pushstring($L,$8); //@line 38 "modules/filesystem/wrap_FileData.cpp"
 return 1; //@line 39 "modules/filesystem/wrap_FileData.cpp"
}
function __ZN4love10filesystem23w_FileData_getExtensionEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_10filesystem8FileDataEEEPT_P9lua_StateiNS_4TypeE($L,1,7)|0); //@line 32 "modules/filesystem/wrap_FileData.cpp"
 $1 = (__ZNK4love10filesystem8FileData12getExtensionEv($0)|0); //@line 45 "modules/filesystem/wrap_FileData.cpp"
 $2 = HEAP8[$1>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = $2 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ($3<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($1)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = HEAP32[$5>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($1)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $4 ? $7 : $6; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 _lua_pushstring($L,$8); //@line 45 "modules/filesystem/wrap_FileData.cpp"
 return 1; //@line 46 "modules/filesystem/wrap_FileData.cpp"
}
function _luaopen_filedata($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 57 "modules/filesystem/wrap_FileData.cpp"
 HEAP32[$vararg_buffer>>2] = 12272; //@line 59 "modules/filesystem/wrap_FileData.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 59 "modules/filesystem/wrap_FileData.cpp"
 HEAP32[$vararg_ptr1>>2] = 23836; //@line 59 "modules/filesystem/wrap_FileData.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 59 "modules/filesystem/wrap_FileData.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 59 "modules/filesystem/wrap_FileData.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,7,366703,$vararg_buffer)|0); //@line 59 "modules/filesystem/wrap_FileData.cpp"
 STACKTOP = sp;return ($0|0); //@line 59 "modules/filesystem/wrap_FileData.cpp"
}
function __ZN4love4font8freetype4FontD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (23868); //@line 42 "modules/font/freetype/Font.cpp"
 $0 = ((($this)) + 8|0); //@line 43 "modules/font/freetype/Font.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 43 "modules/font/freetype/Font.cpp"
 __THREW__ = 0;
 (invoke_ii(3977,($1|0))|0); //@line 43 "modules/font/freetype/Font.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  __ZN4love6ModuleD2Ev($this); //@line 44 "modules/font/freetype/Font.cpp"
  ___clang_call_terminate($4); //@line 44 "modules/font/freetype/Font.cpp"
  // unreachable; //@line 44 "modules/font/freetype/Font.cpp"
 } else {
  __ZN4love6ModuleD2Ev($this); //@line 44 "modules/font/freetype/Font.cpp"
  return; //@line 44 "modules/font/freetype/Font.cpp"
 }
}
function __ZN4love4font8freetype4FontD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (23868); //@line 42 "modules/font/freetype/Font.cpp"
 $0 = ((($this)) + 8|0); //@line 43 "modules/font/freetype/Font.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 43 "modules/font/freetype/Font.cpp"
 __THREW__ = 0;
 (invoke_ii(3977,($1|0))|0); //@line 43 "modules/font/freetype/Font.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  __ZN4love6ModuleD2Ev($this); //@line 44 "modules/font/freetype/Font.cpp"
  ___clang_call_terminate($4); //@line 44 "modules/font/freetype/Font.cpp"
  // unreachable; //@line 44 "modules/font/freetype/Font.cpp"
 } else {
  __ZN4love6ModuleD2Ev($this); //@line 44 "modules/font/freetype/Font.cpp"
  __ZdlPv($this); //@line 42 "modules/font/freetype/Font.cpp"
  return; //@line 44 "modules/font/freetype/Font.cpp"
 }
}
function __ZNK4love4font4Font13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 3; //@line 62 "./modules/font/Font.h"
}
function __ZNK4love4font8freetype4Font7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (366764|0); //@line 63 "modules/font/freetype/Font.cpp"
}
function __ZN4love4font8freetype4Font13newRasterizerEPNS_10filesystem8FileDataE($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i3 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i4 = 0, $vararg_buffer = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 $1 = ((($this)) + 8|0); //@line 48 "modules/font/freetype/Font.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 48 "modules/font/freetype/Font.cpp"
 $3 = (__ZN4love4font8freetype18TrueTypeRasterizer7acceptsEP14FT_LibraryRec_PNS_4DataE($2,$data)|0); //@line 48 "modules/font/freetype/Font.cpp"
 if ($3) {
  $4 = HEAP32[$this>>2]|0; //@line 49 "modules/font/freetype/Font.cpp"
  $5 = ((($4)) + 32|0); //@line 49 "modules/font/freetype/Font.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 49 "modules/font/freetype/Font.cpp"
  $7 = (FUNCTION_TABLE_iiiii[$6 & 8191]($this,$data,12,0)|0); //@line 49 "modules/font/freetype/Font.cpp"
  $$02 = $7;
  STACKTOP = sp;return ($$02|0); //@line 54 "modules/font/freetype/Font.cpp"
 }
 $8 = (__ZN4love4font16BMFontRasterizer7acceptsEPNS_10filesystem8FileDataE($data)|0); //@line 50 "modules/font/freetype/Font.cpp"
 if (!($8)) {
  $37 = (___cxa_allocate_exception(16)|0); //@line 53 "modules/font/freetype/Font.cpp"
  __THREW__ = 0;
  $38 = (invoke_ii(3461,($data|0))|0); //@line 53 "modules/font/freetype/Font.cpp"
  $39 = __THREW__; __THREW__ = 0;
  $40 = $39&1;
  if (!($40)) {
   $41 = HEAP8[$38>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $42 = $41 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $43 = ($42<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $44 = ((($38)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $45 = HEAP32[$44>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $46 = ((($38)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $47 = $43 ? $46 : $45; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   HEAP32[$vararg_buffer>>2] = $47; //@line 53 "modules/font/freetype/Font.cpp"
   invoke_viii(3349,($37|0),(366783|0),($vararg_buffer|0)); //@line 53 "modules/font/freetype/Font.cpp"
   $48 = __THREW__; __THREW__ = 0;
   $49 = $48&1;
   if (!($49)) {
    ___cxa_throw(($37|0),(96|0),(344|0)); //@line 53 "modules/font/freetype/Font.cpp"
    // unreachable; //@line 53 "modules/font/freetype/Font.cpp"
   }
  }
  $50 = ___cxa_find_matching_catch_2()|0;
  $51 = tempRet0;
  ___cxa_free_exception(($37|0)); //@line 53 "modules/font/freetype/Font.cpp"
  $$0 = $51;$$01 = $50;
  ___resumeException($$01|0);
  // unreachable;
 }
 $9 = HEAP32[$this>>2]|0; //@line 51 "modules/font/freetype/Font.cpp"
 $10 = ((($9)) + 36|0); //@line 51 "modules/font/freetype/Font.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 51 "modules/font/freetype/Font.cpp"
 HEAP32[$0>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ((($0)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$12>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ((($0)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$13>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 $14 = (invoke_iiii($11|0,($this|0),($data|0),($0|0))|0); //@line 51 "modules/font/freetype/Font.cpp"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if ($16) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  $28 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $29 = ($28|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $30 = $28; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($29) {
   $$0 = $27;$$01 = $26;
   ___resumeException($$01|0);
   // unreachable;
  }
  $31 = HEAP32[$12>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $32 = ($31|0)==($28|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($32)) {
   $scevgep$i$i$i = ((($31)) + -4|0);
   $33 = $scevgep$i$i$i;
   $34 = (($33) - ($30))|0;
   $35 = $34 >>> 2;
   $36 = $35 ^ -1;
   $scevgep4$i$i$i = (($31) + ($36<<2)|0);
   HEAP32[$12>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($28); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0 = $27;$$01 = $26;
  ___resumeException($$01|0);
  // unreachable;
 } else {
  $17 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ($17|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = $17; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($18) {
   $$02 = $14;
   STACKTOP = sp;return ($$02|0); //@line 54 "modules/font/freetype/Font.cpp"
  }
  $20 = HEAP32[$12>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($20|0)==($17|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($21)) {
   $scevgep$i$i$i3 = ((($20)) + -4|0);
   $22 = $scevgep$i$i$i3;
   $23 = (($22) - ($19))|0;
   $24 = $23 >>> 2;
   $25 = $24 ^ -1;
   $scevgep4$i$i$i4 = (($20) + ($25<<2)|0);
   HEAP32[$12>>2] = $scevgep4$i$i$i4; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($17); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$02 = $14;
  STACKTOP = sp;return ($$02|0); //@line 54 "modules/font/freetype/Font.cpp"
 }
 return (0)|0;
}
function __ZN4love4font8freetype4Font21newTrueTypeRasterizerEPNS_4DataEiNS0_18TrueTypeRasterizer7HintingE($this,$data,$size,$hinting) {
 $this = $this|0;
 $data = $data|0;
 $size = $size|0;
 $hinting = $hinting|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(36)|0); //@line 58 "modules/font/freetype/Font.cpp"
 $1 = ((($this)) + 8|0); //@line 58 "modules/font/freetype/Font.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 58 "modules/font/freetype/Font.cpp"
 __THREW__ = 0;
 invoke_viiiii(3978,($0|0),($2|0),($data|0),($size|0),($hinting|0)); //@line 58 "modules/font/freetype/Font.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  __ZdlPv($0); //@line 58 "modules/font/freetype/Font.cpp"
  ___resumeException($5|0);
  // unreachable;
 } else {
  return ($0|0); //@line 58 "modules/font/freetype/Font.cpp"
 }
 return (0)|0;
}
function __ZN4love4font8freetype4FontC2Ev($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./common/Module.h"
 HEAP32[$this>>2] = (23868); //@line 36 "modules/font/freetype/Font.cpp"
 $0 = ((($this)) + 8|0); //@line 37 "modules/font/freetype/Font.cpp"
 __THREW__ = 0;
 $1 = (invoke_ii(3979,($0|0))|0); //@line 37 "modules/font/freetype/Font.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 do {
  if (!($3)) {
   $4 = ($1|0)==(0); //@line 37 "modules/font/freetype/Font.cpp"
   if ($4) {
    STACKTOP = sp;return; //@line 39 "modules/font/freetype/Font.cpp"
   }
   $5 = (___cxa_allocate_exception(16)|0); //@line 38 "modules/font/freetype/Font.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($5|0),(366805|0),($vararg_buffer|0)); //@line 38 "modules/font/freetype/Font.cpp"
   $6 = __THREW__; __THREW__ = 0;
   $7 = $6&1;
   if (!($7)) {
    __THREW__ = 0;
    invoke_viii(3350,($5|0),(96|0),(344|0)); //@line 38 "modules/font/freetype/Font.cpp"
    $8 = __THREW__; __THREW__ = 0;
    break;
   }
   $11 = ___cxa_find_matching_catch_2()|0;
   $12 = tempRet0;
   ___cxa_free_exception(($5|0)); //@line 38 "modules/font/freetype/Font.cpp"
   $$0 = $12;$$01 = $11;
   __ZN4love6ModuleD2Ev($this); //@line 39 "modules/font/freetype/Font.cpp"
   ___resumeException($$01|0);
   // unreachable;
  }
 } while(0);
 $9 = ___cxa_find_matching_catch_2()|0;
 $10 = tempRet0;
 $$0 = $10;$$01 = $9;
 __ZN4love6ModuleD2Ev($this); //@line 39 "modules/font/freetype/Font.cpp"
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN4love4font8freetype18TrueTypeRasterizerD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (23932); //@line 64 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $0 = ((($this)) + 24|0); //@line 65 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 65 "modules/font/freetype/TrueTypeRasterizer.cpp"
 __THREW__ = 0;
 (invoke_ii(3980,($1|0))|0); //@line 65 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $14 = ___cxa_find_matching_catch_3(0|0)|0;
  $15 = tempRet0;
  $16 = ((($this)) + 28|0); //@line 106 "./common/Object.h"
  $17 = HEAP32[$16>>2]|0; //@line 106 "./common/Object.h"
  $18 = ($17|0)==(0|0); //@line 106 "./common/Object.h"
  if ($18) {
   __ZN4love4font10RasterizerD2Ev($this); //@line 66 "modules/font/freetype/TrueTypeRasterizer.cpp"
   ___clang_call_terminate($14); //@line 66 "modules/font/freetype/TrueTypeRasterizer.cpp"
   // unreachable; //@line 66 "modules/font/freetype/TrueTypeRasterizer.cpp"
  }
  $19 = HEAP32[$17>>2]|0; //@line 106 "./common/Object.h"
  $20 = ((($19)) + 12|0); //@line 106 "./common/Object.h"
  $21 = HEAP32[$20>>2]|0; //@line 106 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($21|0,($17|0)); //@line 106 "./common/Object.h"
  $22 = __THREW__; __THREW__ = 0;
  $23 = $22&1;
  if ($23) {
   $24 = ___cxa_find_matching_catch_3(0|0)|0;
   $25 = tempRet0;
   ___clang_call_terminate($24); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  } else {
   __ZN4love4font10RasterizerD2Ev($this); //@line 66 "modules/font/freetype/TrueTypeRasterizer.cpp"
   ___clang_call_terminate($14); //@line 66 "modules/font/freetype/TrueTypeRasterizer.cpp"
   // unreachable; //@line 66 "modules/font/freetype/TrueTypeRasterizer.cpp"
  }
 } else {
  $4 = ((($this)) + 28|0); //@line 106 "./common/Object.h"
  $5 = HEAP32[$4>>2]|0; //@line 106 "./common/Object.h"
  $6 = ($5|0)==(0|0); //@line 106 "./common/Object.h"
  if ($6) {
   __ZN4love4font10RasterizerD2Ev($this); //@line 66 "modules/font/freetype/TrueTypeRasterizer.cpp"
   return; //@line 66 "modules/font/freetype/TrueTypeRasterizer.cpp"
  }
  $7 = HEAP32[$5>>2]|0; //@line 106 "./common/Object.h"
  $8 = ((($7)) + 12|0); //@line 106 "./common/Object.h"
  $9 = HEAP32[$8>>2]|0; //@line 106 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($9|0,($5|0)); //@line 106 "./common/Object.h"
  $10 = __THREW__; __THREW__ = 0;
  $11 = $10&1;
  if ($11) {
   $12 = ___cxa_find_matching_catch_3(0|0)|0;
   $13 = tempRet0;
   ___clang_call_terminate($12); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  } else {
   __ZN4love4font10RasterizerD2Ev($this); //@line 66 "modules/font/freetype/TrueTypeRasterizer.cpp"
   return; //@line 66 "modules/font/freetype/TrueTypeRasterizer.cpp"
  }
 }
}
function __ZN4love4font8freetype18TrueTypeRasterizerD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love4font8freetype18TrueTypeRasterizerD2Ev($this); //@line 64 "modules/font/freetype/TrueTypeRasterizer.cpp"
 __ZdlPv($this); //@line 64 "modules/font/freetype/TrueTypeRasterizer.cpp"
 return; //@line 66 "modules/font/freetype/TrueTypeRasterizer.cpp"
}
function __ZNK4love4font8freetype18TrueTypeRasterizer13getLineHeightEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 70 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $1 = ((($0)) + 16|0); //@line 70 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 70 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 70 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $4 = (+($3|0)); //@line 70 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $5 = $4 * 1.25; //@line 70 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $6 = (~~(($5))); //@line 70 "modules/font/freetype/TrueTypeRasterizer.cpp"
 return ($6|0); //@line 70 "modules/font/freetype/TrueTypeRasterizer.cpp"
}
function __ZNK4love4font8freetype18TrueTypeRasterizer12getGlyphDataEj($this,$glyph) {
 $this = $this|0;
 $glyph = $glyph|0;
 var $$ = 0, $$0 = 0, $$0$i = 0, $$01 = 0, $$byval_copy = 0, $$lcssa = 0, $$lcssa26 = 0, $$pre = 0, $$pre17 = 0, $$pre18 = 0, $$pre19 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $ftglyph = 0, $pixels$07 = 0, $pixels$113 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, $vararg_buffer7 = 0, $x$06 = 0, $x2$010 = 0, $y$08 = 0, $y1$014 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 52|0;
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $ftglyph = sp + 48|0;
 $0 = sp + 28|0;
 $1 = ((($this)) + 32|0); //@line 79 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 79 "modules/font/freetype/TrueTypeRasterizer.cpp"
 switch ($2|0) {
 case 3:  {
  $$0$i = 2;
  break;
 }
 case 1:  {
  $$0$i = 65536;
  break;
 }
 case 2:  {
  $$0$i = 131072;
  break;
 }
 default: {
  $$0$i = 0;
 }
 }
 $3 = ((($this)) + 24|0); //@line 82 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 82 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $5 = (_FT_Get_Char_Index($4,$glyph)|0); //@line 82 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $6 = (_FT_Load_Glyph($4,$5,$$0$i)|0); //@line 82 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $7 = ($6|0)==(0); //@line 84 "modules/font/freetype/TrueTypeRasterizer.cpp"
 if (!($7)) {
  $8 = (___cxa_allocate_exception(16)|0); //@line 85 "modules/font/freetype/TrueTypeRasterizer.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer>>2] = $6; //@line 85 "modules/font/freetype/TrueTypeRasterizer.cpp"
  invoke_viii(3349,($8|0),(366932|0),($vararg_buffer|0)); //@line 85 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if (!($10)) {
   ___cxa_throw(($8|0),(96|0),(344|0)); //@line 85 "modules/font/freetype/TrueTypeRasterizer.cpp"
   // unreachable; //@line 85 "modules/font/freetype/TrueTypeRasterizer.cpp"
  }
  $11 = ___cxa_find_matching_catch_2()|0;
  $12 = tempRet0;
  ___cxa_free_exception(($8|0)); //@line 85 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $$0 = $12;$$01 = $11;
  ___resumeException($$01|0);
  // unreachable;
 }
 $13 = HEAP32[$3>>2]|0; //@line 87 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $14 = ((($13)) + 84|0); //@line 87 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 87 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $16 = (_FT_Get_Glyph($15,$ftglyph)|0); //@line 87 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $17 = ($16|0)==(0); //@line 89 "modules/font/freetype/TrueTypeRasterizer.cpp"
 if (!($17)) {
  $18 = (___cxa_allocate_exception(16)|0); //@line 90 "modules/font/freetype/TrueTypeRasterizer.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer1>>2] = $16; //@line 90 "modules/font/freetype/TrueTypeRasterizer.cpp"
  invoke_viii(3349,($18|0),(366987|0),($vararg_buffer1|0)); //@line 90 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $19 = __THREW__; __THREW__ = 0;
  $20 = $19&1;
  if (!($20)) {
   ___cxa_throw(($18|0),(96|0),(344|0)); //@line 90 "modules/font/freetype/TrueTypeRasterizer.cpp"
   // unreachable; //@line 90 "modules/font/freetype/TrueTypeRasterizer.cpp"
  }
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  ___cxa_free_exception(($18|0)); //@line 90 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $$0 = $22;$$01 = $21;
  ___resumeException($$01|0);
  // unreachable;
 }
 $23 = HEAP32[$1>>2]|0; //@line 93 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $24 = ($23|0)==(2); //@line 93 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $$ = $24 ? 2 : 0; //@line 93 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $25 = (_FT_Glyph_To_Bitmap($ftglyph,$$,0,1)|0); //@line 96 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $26 = ($25|0)==(0); //@line 98 "modules/font/freetype/TrueTypeRasterizer.cpp"
 if (!($26)) {
  $27 = (___cxa_allocate_exception(16)|0); //@line 99 "modules/font/freetype/TrueTypeRasterizer.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer4>>2] = $25; //@line 99 "modules/font/freetype/TrueTypeRasterizer.cpp"
  invoke_viii(3349,($27|0),(367041|0),($vararg_buffer4|0)); //@line 99 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $28 = __THREW__; __THREW__ = 0;
  $29 = $28&1;
  if (!($29)) {
   ___cxa_throw(($27|0),(96|0),(344|0)); //@line 99 "modules/font/freetype/TrueTypeRasterizer.cpp"
   // unreachable; //@line 99 "modules/font/freetype/TrueTypeRasterizer.cpp"
  }
  $30 = ___cxa_find_matching_catch_2()|0;
  $31 = tempRet0;
  ___cxa_free_exception(($27|0)); //@line 99 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $$0 = $31;$$01 = $30;
  ___resumeException($$01|0);
  // unreachable;
 }
 $32 = HEAP32[$ftglyph>>2]|0; //@line 101 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $33 = ((($32)) + 28|0); //@line 102 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $34 = ((($32)) + 20|0); //@line 105 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 105 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $36 = ((($32)) + 24|0); //@line 106 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $37 = HEAP32[$36>>2]|0; //@line 106 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $38 = HEAP32[$33>>2]|0; //@line 107 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $39 = ((($33)) + 4|0); //@line 108 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $40 = HEAP32[$39>>2]|0; //@line 108 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $41 = ((($32)) + 12|0); //@line 109 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $42 = HEAP32[$41>>2]|0; //@line 109 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $43 = $42 >> 16; //@line 109 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $44 = (__Znwj(40)|0); //@line 111 "modules/font/freetype/TrueTypeRasterizer.cpp"
 HEAP32[$0>>2] = $38; //@line 111 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $45 = ((($0)) + 4|0); //@line 111 "modules/font/freetype/TrueTypeRasterizer.cpp"
 HEAP32[$45>>2] = $40; //@line 111 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $46 = ((($0)) + 8|0); //@line 111 "modules/font/freetype/TrueTypeRasterizer.cpp"
 HEAP32[$46>>2] = $43; //@line 111 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $47 = ((($0)) + 12|0); //@line 111 "modules/font/freetype/TrueTypeRasterizer.cpp"
 HEAP32[$47>>2] = $35; //@line 111 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $48 = ((($0)) + 16|0); //@line 111 "modules/font/freetype/TrueTypeRasterizer.cpp"
 HEAP32[$48>>2] = $37; //@line 111 "modules/font/freetype/TrueTypeRasterizer.cpp"
 __THREW__ = 0;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;HEAP32[$$byval_copy+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$$byval_copy+8>>2]=HEAP32[$0+8>>2]|0;HEAP32[$$byval_copy+12>>2]=HEAP32[$0+12>>2]|0;HEAP32[$$byval_copy+16>>2]=HEAP32[$0+16>>2]|0; //@line 111 "modules/font/freetype/TrueTypeRasterizer.cpp"
 invoke_viiii(3981,($44|0),($glyph|0),($$byval_copy|0),0); //@line 111 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $49 = __THREW__; __THREW__ = 0;
 $50 = $49&1;
 if ($50) {
  $67 = ___cxa_find_matching_catch_2()|0;
  $68 = tempRet0;
  __ZdlPv($44); //@line 111 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $$0 = $68;$$01 = $67;
  ___resumeException($$01|0);
  // unreachable;
 }
 $51 = ((($33)) + 12|0); //@line 113 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $52 = HEAP32[$51>>2]|0; //@line 113 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $53 = HEAP32[$44>>2]|0; //@line 114 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $54 = ((($53)) + 16|0); //@line 114 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $55 = HEAP32[$54>>2]|0; //@line 114 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $56 = (FUNCTION_TABLE_ii[$55 & 8191]($44)|0); //@line 114 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $57 = ((($33)) + 18|0); //@line 117 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $58 = HEAP8[$57>>0]|0; //@line 117 "modules/font/freetype/TrueTypeRasterizer.cpp"
 switch ($58<<24>>24) {
 case 1:  {
  $62 = HEAP32[$33>>2]|0; //@line 119 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $63 = ($62|0)>(0); //@line 119 "modules/font/freetype/TrueTypeRasterizer.cpp"
  if (!($63)) {
   $130 = HEAP32[$ftglyph>>2]|0; //@line 153 "modules/font/freetype/TrueTypeRasterizer.cpp"
   _FT_Done_Glyph($130); //@line 153 "modules/font/freetype/TrueTypeRasterizer.cpp"
   STACKTOP = sp;return ($44|0); //@line 155 "modules/font/freetype/TrueTypeRasterizer.cpp"
  }
  $64 = ((($33)) + 8|0); //@line 129 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $$pre18 = HEAP32[$39>>2]|0; //@line 121 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $132 = $62;$66 = $$pre18;$pixels$07 = $52;$y$08 = 0;
  while(1) {
   $65 = ($66|0)>(0); //@line 121 "modules/font/freetype/TrueTypeRasterizer.cpp"
   if ($65) {
    $85 = $66;$x$06 = 0;
    while(1) {
     $74 = (($x$06|0) / 8)&-1; //@line 124 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $75 = (($pixels$07) + ($74)|0); //@line 124 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $76 = HEAP8[$75>>0]|0; //@line 124 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $77 = $76&255; //@line 124 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $78 = (($x$06|0) % 8)&-1; //@line 124 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $79 = (7 - ($78))|0; //@line 124 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $80 = 1 << $79; //@line 124 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $81 = $77 & $80; //@line 124 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $82 = ($81|0)!=(0); //@line 124 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $83 = $82 << 31 >> 31; //@line 124 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $84 = Math_imul($85, $y$08)|0; //@line 125 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $86 = (($84) + ($x$06))|0; //@line 125 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $87 = $86 << 1; //@line 125 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $88 = (($56) + ($87)|0); //@line 125 "modules/font/freetype/TrueTypeRasterizer.cpp"
     HEAP8[$88>>0] = -1; //@line 125 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $89 = HEAP32[$39>>2]|0; //@line 126 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $90 = Math_imul($89, $y$08)|0; //@line 126 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $91 = (($90) + ($x$06))|0; //@line 126 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $92 = $91 << 1; //@line 126 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $93 = $92 | 1; //@line 126 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $94 = (($56) + ($93)|0); //@line 126 "modules/font/freetype/TrueTypeRasterizer.cpp"
     HEAP8[$94>>0] = $83; //@line 126 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $95 = (($x$06) + 1)|0; //@line 121 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $96 = HEAP32[$39>>2]|0; //@line 121 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $97 = ($95|0)<($96|0); //@line 121 "modules/font/freetype/TrueTypeRasterizer.cpp"
     if ($97) {
      $85 = $96;$x$06 = $95;
     } else {
      $$lcssa = $96;
      break;
     }
    }
    $$pre19 = HEAP32[$33>>2]|0; //@line 119 "modules/font/freetype/TrueTypeRasterizer.cpp"
    $133 = $$lcssa;$73 = $$pre19;
   } else {
    $133 = $66;$73 = $132;
   }
   $69 = HEAP32[$64>>2]|0; //@line 129 "modules/font/freetype/TrueTypeRasterizer.cpp"
   $70 = (($pixels$07) + ($69)|0); //@line 129 "modules/font/freetype/TrueTypeRasterizer.cpp"
   $71 = (($y$08) + 1)|0; //@line 119 "modules/font/freetype/TrueTypeRasterizer.cpp"
   $72 = ($71|0)<($73|0); //@line 119 "modules/font/freetype/TrueTypeRasterizer.cpp"
   if ($72) {
    $132 = $73;$66 = $133;$pixels$07 = $70;$y$08 = $71;
   } else {
    break;
   }
  }
  $130 = HEAP32[$ftglyph>>2]|0; //@line 153 "modules/font/freetype/TrueTypeRasterizer.cpp"
  _FT_Done_Glyph($130); //@line 153 "modules/font/freetype/TrueTypeRasterizer.cpp"
  STACKTOP = sp;return ($44|0); //@line 155 "modules/font/freetype/TrueTypeRasterizer.cpp"
  break;
 }
 case 2:  {
  $59 = HEAP32[$33>>2]|0; //@line 134 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $60 = ($59|0)>(0); //@line 134 "modules/font/freetype/TrueTypeRasterizer.cpp"
  if (!($60)) {
   $130 = HEAP32[$ftglyph>>2]|0; //@line 153 "modules/font/freetype/TrueTypeRasterizer.cpp"
   _FT_Done_Glyph($130); //@line 153 "modules/font/freetype/TrueTypeRasterizer.cpp"
   STACKTOP = sp;return ($44|0); //@line 155 "modules/font/freetype/TrueTypeRasterizer.cpp"
  }
  $61 = ((($33)) + 8|0); //@line 142 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $$pre = HEAP32[$39>>2]|0; //@line 136 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $131 = $59;$99 = $$pre;$pixels$113 = $52;$y1$014 = 0;
  while(1) {
   $98 = ($99|0)>(0); //@line 136 "modules/font/freetype/TrueTypeRasterizer.cpp"
   if ($98) {
    $106 = $99;$x2$010 = 0;
    while(1) {
     $105 = Math_imul($106, $y1$014)|0; //@line 138 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $107 = (($105) + ($x2$010))|0; //@line 138 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $108 = $107 << 1; //@line 138 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $109 = (($56) + ($108)|0); //@line 138 "modules/font/freetype/TrueTypeRasterizer.cpp"
     HEAP8[$109>>0] = -1; //@line 138 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $110 = (($pixels$113) + ($x2$010)|0); //@line 139 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $111 = HEAP8[$110>>0]|0; //@line 139 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $112 = HEAP32[$39>>2]|0; //@line 139 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $113 = Math_imul($112, $y1$014)|0; //@line 139 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $114 = (($113) + ($x2$010))|0; //@line 139 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $115 = $114 << 1; //@line 139 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $116 = $115 | 1; //@line 139 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $117 = (($56) + ($116)|0); //@line 139 "modules/font/freetype/TrueTypeRasterizer.cpp"
     HEAP8[$117>>0] = $111; //@line 139 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $118 = (($x2$010) + 1)|0; //@line 136 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $119 = HEAP32[$39>>2]|0; //@line 136 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $120 = ($118|0)<($119|0); //@line 136 "modules/font/freetype/TrueTypeRasterizer.cpp"
     if ($120) {
      $106 = $119;$x2$010 = $118;
     } else {
      $$lcssa26 = $119;
      break;
     }
    }
    $$pre17 = HEAP32[$33>>2]|0; //@line 134 "modules/font/freetype/TrueTypeRasterizer.cpp"
    $104 = $$pre17;$134 = $$lcssa26;
   } else {
    $104 = $131;$134 = $99;
   }
   $100 = HEAP32[$61>>2]|0; //@line 142 "modules/font/freetype/TrueTypeRasterizer.cpp"
   $101 = (($pixels$113) + ($100)|0); //@line 142 "modules/font/freetype/TrueTypeRasterizer.cpp"
   $102 = (($y1$014) + 1)|0; //@line 134 "modules/font/freetype/TrueTypeRasterizer.cpp"
   $103 = ($102|0)<($104|0); //@line 134 "modules/font/freetype/TrueTypeRasterizer.cpp"
   if ($103) {
    $131 = $104;$99 = $134;$pixels$113 = $101;$y1$014 = $102;
   } else {
    break;
   }
  }
  $130 = HEAP32[$ftglyph>>2]|0; //@line 153 "modules/font/freetype/TrueTypeRasterizer.cpp"
  _FT_Done_Glyph($130); //@line 153 "modules/font/freetype/TrueTypeRasterizer.cpp"
  STACKTOP = sp;return ($44|0); //@line 155 "modules/font/freetype/TrueTypeRasterizer.cpp"
  break;
 }
 default: {
  $121 = HEAP32[$44>>2]|0; //@line 147 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $122 = ((($121)) + 4|0); //@line 147 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $123 = HEAP32[$122>>2]|0; //@line 147 "modules/font/freetype/TrueTypeRasterizer.cpp"
  FUNCTION_TABLE_vi[$123 & 8191]($44); //@line 147 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $124 = HEAP32[$ftglyph>>2]|0; //@line 148 "modules/font/freetype/TrueTypeRasterizer.cpp"
  _FT_Done_Glyph($124); //@line 148 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $125 = (___cxa_allocate_exception(16)|0); //@line 149 "modules/font/freetype/TrueTypeRasterizer.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($125|0),(367101|0),($vararg_buffer7|0)); //@line 149 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $126 = __THREW__; __THREW__ = 0;
  $127 = $126&1;
  if (!($127)) {
   ___cxa_throw(($125|0),(96|0),(344|0)); //@line 149 "modules/font/freetype/TrueTypeRasterizer.cpp"
   // unreachable; //@line 149 "modules/font/freetype/TrueTypeRasterizer.cpp"
  }
  $128 = ___cxa_find_matching_catch_2()|0;
  $129 = tempRet0;
  ___cxa_free_exception(($125|0)); //@line 149 "modules/font/freetype/TrueTypeRasterizer.cpp"
  $$0 = $129;$$01 = $128;
  ___resumeException($$01|0);
  // unreachable;
 }
 }
 return (0)|0;
}
function __ZNK4love4font8freetype18TrueTypeRasterizer13getGlyphCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 160 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 160 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $2 = ((($1)) + 16|0); //@line 160 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 160 "modules/font/freetype/TrueTypeRasterizer.cpp"
 return ($3|0); //@line 160 "modules/font/freetype/TrueTypeRasterizer.cpp"
}
function __ZNK4love4font8freetype18TrueTypeRasterizer8hasGlyphEj($this,$glyph) {
 $this = $this|0;
 $glyph = $glyph|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 165 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 165 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $2 = (_FT_Get_Char_Index($1,$glyph)|0); //@line 165 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $3 = ($2|0)!=(0); //@line 165 "modules/font/freetype/TrueTypeRasterizer.cpp"
 return ($3|0); //@line 165 "modules/font/freetype/TrueTypeRasterizer.cpp"
}
function __ZNK4love4font8freetype18TrueTypeRasterizer10getKerningEjj($this,$leftglyph,$rightglyph) {
 $this = $this|0;
 $leftglyph = $leftglyph|0;
 $rightglyph = $rightglyph|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $kerning = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $kerning = sp;
 $0 = $kerning; //@line 170 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $1 = $0; //@line 170 "modules/font/freetype/TrueTypeRasterizer.cpp"
 HEAP32[$1>>2] = 0; //@line 170 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $2 = (($0) + 4)|0; //@line 170 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $3 = $2; //@line 170 "modules/font/freetype/TrueTypeRasterizer.cpp"
 HEAP32[$3>>2] = 0; //@line 170 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $4 = ((($this)) + 24|0); //@line 171 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 171 "modules/font/freetype/TrueTypeRasterizer.cpp"
 (_FT_Get_Kerning($5,$leftglyph,$rightglyph,0,$kerning)|0); //@line 171 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $6 = HEAP32[$kerning>>2]|0; //@line 172 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $7 = $6 >> 6; //@line 172 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $8 = (+($7|0)); //@line 172 "modules/font/freetype/TrueTypeRasterizer.cpp"
 STACKTOP = sp;return (+$8); //@line 172 "modules/font/freetype/TrueTypeRasterizer.cpp"
}
function __ZN4love4font8freetype18TrueTypeRasterizerC2EP14FT_LibraryRec_PNS_4DataEiNS0_18TrueTypeRasterizer7HintingE($this,$library,$data,$size,$hinting) {
 $this = $this|0;
 $library = $library|0;
 $data = $data|0;
 $size = $size|0;
 $hinting = $hinting|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 __ZN4love6ObjectC2Ev($this); //@line 48 "./modules/font/Rasterizer.h"
 HEAP32[$this>>2] = (23932); //@line 36 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $0 = ((($this)) + 28|0); //@line 93 "./common/Object.h"
 HEAP32[$0>>2] = $data; //@line 93 "./common/Object.h"
 $1 = ($data|0)==(0|0); //@line 95 "./common/Object.h"
 if (!($1)) {
  $2 = HEAP32[$data>>2]|0; //@line 95 "./common/Object.h"
  $3 = ((($2)) + 8|0); //@line 95 "./common/Object.h"
  $4 = HEAP32[$3>>2]|0; //@line 95 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($4|0,($data|0)); //@line 95 "./common/Object.h"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if ($6) {
   $13 = ___cxa_find_matching_catch_2()|0;
   $14 = tempRet0;
   $$1 = $13;$$12 = $14;
   __ZN4love4font10RasterizerD2Ev($this); //@line 61 "modules/font/freetype/TrueTypeRasterizer.cpp"
   ___resumeException($$1|0);
   // unreachable;
  }
 }
 $7 = ((($this)) + 32|0); //@line 35 "modules/font/freetype/TrueTypeRasterizer.cpp"
 HEAP32[$7>>2] = $hinting; //@line 35 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $8 = ($size|0)<(1); //@line 37 "modules/font/freetype/TrueTypeRasterizer.cpp"
 do {
  if ($8) {
   $9 = (___cxa_allocate_exception(16)|0); //@line 38 "modules/font/freetype/TrueTypeRasterizer.cpp"
   __THREW__ = 0;
   HEAP32[$vararg_buffer>>2] = $size; //@line 38 "modules/font/freetype/TrueTypeRasterizer.cpp"
   invoke_viii(3349,($9|0),(367136|0),($vararg_buffer|0)); //@line 38 "modules/font/freetype/TrueTypeRasterizer.cpp"
   $10 = __THREW__; __THREW__ = 0;
   $11 = $10&1;
   if ($11) {
    $15 = ___cxa_find_matching_catch_2()|0;
    $16 = tempRet0;
    ___cxa_free_exception(($9|0)); //@line 38 "modules/font/freetype/TrueTypeRasterizer.cpp"
    $$0 = $15;$$01 = $16;
    break;
   } else {
    __THREW__ = 0;
    invoke_viii(3350,($9|0),(96|0),(344|0)); //@line 38 "modules/font/freetype/TrueTypeRasterizer.cpp"
    $12 = __THREW__; __THREW__ = 0;
    label = 8;
    break;
   }
  } else {
   $19 = HEAP32[$data>>2]|0; //@line 42 "modules/font/freetype/TrueTypeRasterizer.cpp"
   $20 = ((($19)) + 16|0); //@line 42 "modules/font/freetype/TrueTypeRasterizer.cpp"
   $21 = HEAP32[$20>>2]|0; //@line 42 "modules/font/freetype/TrueTypeRasterizer.cpp"
   __THREW__ = 0;
   $22 = (invoke_ii($21|0,($data|0))|0); //@line 42 "modules/font/freetype/TrueTypeRasterizer.cpp"
   $23 = __THREW__; __THREW__ = 0;
   $24 = $23&1;
   if ($24) {
    label = 8;
   } else {
    $25 = HEAP32[$data>>2]|0; //@line 43 "modules/font/freetype/TrueTypeRasterizer.cpp"
    $26 = ((($25)) + 20|0); //@line 43 "modules/font/freetype/TrueTypeRasterizer.cpp"
    $27 = HEAP32[$26>>2]|0; //@line 43 "modules/font/freetype/TrueTypeRasterizer.cpp"
    __THREW__ = 0;
    $28 = (invoke_ii($27|0,($data|0))|0); //@line 43 "modules/font/freetype/TrueTypeRasterizer.cpp"
    $29 = __THREW__; __THREW__ = 0;
    $30 = $29&1;
    if ($30) {
     label = 8;
    } else {
     $31 = ((($this)) + 24|0); //@line 45 "modules/font/freetype/TrueTypeRasterizer.cpp"
     __THREW__ = 0;
     $32 = (invoke_iiiiii(3982,($library|0),($22|0),($28|0),0,($31|0))|0); //@line 41 "modules/font/freetype/TrueTypeRasterizer.cpp"
     $33 = __THREW__; __THREW__ = 0;
     $34 = $33&1;
     if ($34) {
      label = 8;
     } else {
      $35 = ($32|0)==(0); //@line 47 "modules/font/freetype/TrueTypeRasterizer.cpp"
      if (!($35)) {
       $36 = (___cxa_allocate_exception(16)|0); //@line 48 "modules/font/freetype/TrueTypeRasterizer.cpp"
       __THREW__ = 0;
       HEAP32[$vararg_buffer1>>2] = $32; //@line 48 "modules/font/freetype/TrueTypeRasterizer.cpp"
       invoke_viii(3349,($36|0),(367167|0),($vararg_buffer1|0)); //@line 48 "modules/font/freetype/TrueTypeRasterizer.cpp"
       $37 = __THREW__; __THREW__ = 0;
       $38 = $37&1;
       if ($38) {
        $40 = ___cxa_find_matching_catch_2()|0;
        $41 = tempRet0;
        ___cxa_free_exception(($36|0)); //@line 48 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $$0 = $40;$$01 = $41;
        break;
       } else {
        __THREW__ = 0;
        invoke_viii(3350,($36|0),(96|0),(344|0)); //@line 48 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $39 = __THREW__; __THREW__ = 0;
        label = 8;
        break;
       }
      }
      $42 = HEAP32[$31>>2]|0; //@line 50 "modules/font/freetype/TrueTypeRasterizer.cpp"
      __THREW__ = 0;
      $43 = (invoke_iiii(3983,($42|0),($size|0),($size|0))|0); //@line 50 "modules/font/freetype/TrueTypeRasterizer.cpp"
      $44 = __THREW__; __THREW__ = 0;
      $45 = $44&1;
      if ($45) {
       label = 8;
      } else {
       $46 = ($43|0)==(0); //@line 52 "modules/font/freetype/TrueTypeRasterizer.cpp"
       if ($46) {
        $53 = HEAP32[$31>>2]|0; //@line 56 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $54 = ((($53)) + 88|0); //@line 56 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $55 = HEAP32[$54>>2]|0; //@line 56 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $56 = ((($55)) + 24|0); //@line 56 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $57 = HEAP32[$56>>2]|0; //@line 56 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $58 = ((($55)) + 28|0); //@line 56 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $59 = HEAP32[$58>>2]|0; //@line 56 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $60 = ((($55)) + 32|0); //@line 56 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $61 = HEAP32[$60>>2]|0; //@line 56 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $62 = ((($55)) + 36|0); //@line 56 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $63 = HEAP32[$62>>2]|0; //@line 56 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $64 = $63 >> 6; //@line 57 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $65 = ((($this)) + 8|0); //@line 57 "modules/font/freetype/TrueTypeRasterizer.cpp"
        HEAP32[$65>>2] = $64; //@line 57 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $66 = $57 >> 6; //@line 58 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $67 = ((($this)) + 12|0); //@line 58 "modules/font/freetype/TrueTypeRasterizer.cpp"
        HEAP32[$67>>2] = $66; //@line 58 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $68 = $59 >> 6; //@line 59 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $69 = ((($this)) + 16|0); //@line 59 "modules/font/freetype/TrueTypeRasterizer.cpp"
        HEAP32[$69>>2] = $68; //@line 59 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $70 = $61 >> 6; //@line 60 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $71 = ((($this)) + 20|0); //@line 60 "modules/font/freetype/TrueTypeRasterizer.cpp"
        HEAP32[$71>>2] = $70; //@line 60 "modules/font/freetype/TrueTypeRasterizer.cpp"
        STACKTOP = sp;return; //@line 61 "modules/font/freetype/TrueTypeRasterizer.cpp"
       }
       $47 = (___cxa_allocate_exception(16)|0); //@line 53 "modules/font/freetype/TrueTypeRasterizer.cpp"
       __THREW__ = 0;
       HEAP32[$vararg_buffer4>>2] = $43; //@line 53 "modules/font/freetype/TrueTypeRasterizer.cpp"
       invoke_viii(3349,($47|0),(367247|0),($vararg_buffer4|0)); //@line 53 "modules/font/freetype/TrueTypeRasterizer.cpp"
       $48 = __THREW__; __THREW__ = 0;
       $49 = $48&1;
       if ($49) {
        $51 = ___cxa_find_matching_catch_2()|0;
        $52 = tempRet0;
        ___cxa_free_exception(($47|0)); //@line 53 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $$0 = $51;$$01 = $52;
        break;
       } else {
        __THREW__ = 0;
        invoke_viii(3350,($47|0),(96|0),(344|0)); //@line 53 "modules/font/freetype/TrueTypeRasterizer.cpp"
        $50 = __THREW__; __THREW__ = 0;
        label = 8;
        break;
       }
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  $$0 = $17;$$01 = $18;
 }
 $72 = HEAP32[$0>>2]|0; //@line 106 "./common/Object.h"
 $73 = ($72|0)==(0|0); //@line 106 "./common/Object.h"
 if ($73) {
  $$1 = $$0;$$12 = $$01;
  __ZN4love4font10RasterizerD2Ev($this); //@line 61 "modules/font/freetype/TrueTypeRasterizer.cpp"
  ___resumeException($$1|0);
  // unreachable;
 }
 $74 = HEAP32[$72>>2]|0; //@line 106 "./common/Object.h"
 $75 = ((($74)) + 12|0); //@line 106 "./common/Object.h"
 $76 = HEAP32[$75>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($76|0,($72|0)); //@line 106 "./common/Object.h"
 $77 = __THREW__; __THREW__ = 0;
 $78 = $77&1;
 if ($78) {
  $79 = ___cxa_find_matching_catch_3(0|0)|0;
  $80 = tempRet0;
  ___clang_call_terminate($79); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  $$1 = $$0;$$12 = $$01;
  __ZN4love4font10RasterizerD2Ev($this); //@line 61 "modules/font/freetype/TrueTypeRasterizer.cpp"
  ___resumeException($$1|0);
  // unreachable;
 }
}
function __ZN4love4font8freetype18TrueTypeRasterizer7acceptsEP14FT_LibraryRec_PNS_4DataE($library,$data) {
 $library = $library|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$data>>2]|0; //@line 177 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $1 = ((($0)) + 16|0); //@line 177 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 177 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($data)|0); //@line 177 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $4 = HEAP32[$data>>2]|0; //@line 178 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $5 = ((($4)) + 20|0); //@line 178 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 178 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $7 = (FUNCTION_TABLE_ii[$6 & 8191]($data)|0); //@line 178 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $8 = (_FT_New_Memory_Face($library,$3,$7,-1,0)|0); //@line 181 "modules/font/freetype/TrueTypeRasterizer.cpp"
 $9 = ($8|0)==(0); //@line 181 "modules/font/freetype/TrueTypeRasterizer.cpp"
 return ($9|0); //@line 181 "modules/font/freetype/TrueTypeRasterizer.cpp"
}
function __ZN4love4font16BMFontRasterizerD2Ev($this) {
 $this = $this|0;
 var $$01$i$i$i = 0, $$01$i$i$i1 = 0, $$01$i$i$i3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (24000); //@line 155 "modules/font/BMFontRasterizer.cpp"
 $0 = ((($this)) + 84|0); //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $1 = HEAP32[$0>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $2 = ($1|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($2)) {
  $$01$i$i$i = $1;
  while(1) {
   $3 = HEAP32[$$01$i$i$i>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   __ZdlPv($$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $4 = ($3|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($4) {
    break;
   } else {
    $$01$i$i$i = $3;
   }
  }
 }
 $5 = ((($this)) + 76|0); //@line 2014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = HEAP32[$5>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$5>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $7 = ($6|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($7)) {
  __ZdlPv($6); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $8 = ((($this)) + 64|0); //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $9 = HEAP32[$8>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $10 = ($9|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($10)) {
  $$01$i$i$i1 = $9;
  while(1) {
   $11 = HEAP32[$$01$i$i$i1>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   __ZdlPv($$01$i$i$i1); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $12 = ($11|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($12) {
    break;
   } else {
    $$01$i$i$i1 = $11;
   }
  }
 }
 $13 = ((($this)) + 56|0); //@line 2014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = HEAP32[$13>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$13>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $15 = ($14|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($15)) {
  __ZdlPv($14); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $16 = ((($this)) + 44|0); //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $17 = HEAP32[$16>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $18 = ($17|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L15: do {
  if (!($18)) {
   $$01$i$i$i3 = $17;
   while(1) {
    $19 = HEAP32[$$01$i$i$i3>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $20 = ((($$01$i$i$i3)) + 12|0); //@line 106 "./common/Object.h"
    $21 = HEAP32[$20>>2]|0; //@line 106 "./common/Object.h"
    $22 = ($21|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($22)) {
     $23 = HEAP32[$21>>2]|0; //@line 106 "./common/Object.h"
     $24 = ((($23)) + 12|0); //@line 106 "./common/Object.h"
     $25 = HEAP32[$24>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($25|0,($21|0)); //@line 106 "./common/Object.h"
     $26 = __THREW__; __THREW__ = 0;
     $27 = $26&1;
     if ($27) {
      break;
     }
    }
    __ZdlPv($$01$i$i$i3); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $30 = ($19|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($30) {
     break L15;
    } else {
     $$01$i$i$i3 = $19;
    }
   }
   $28 = ___cxa_find_matching_catch_3(0|0)|0;
   $29 = tempRet0;
   ___clang_call_terminate($28); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 } while(0);
 $31 = ((($this)) + 36|0); //@line 2014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $32 = HEAP32[$31>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$31>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = ($32|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($33) {
  $34 = ((($this)) + 24|0); //@line 156 "modules/font/BMFontRasterizer.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 156 "modules/font/BMFontRasterizer.cpp"
  __ZN4love4font10RasterizerD2Ev($this); //@line 156 "modules/font/BMFontRasterizer.cpp"
  return; //@line 156 "modules/font/BMFontRasterizer.cpp"
 }
 __ZdlPv($32); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $34 = ((($this)) + 24|0); //@line 156 "modules/font/BMFontRasterizer.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 156 "modules/font/BMFontRasterizer.cpp"
 __ZN4love4font10RasterizerD2Ev($this); //@line 156 "modules/font/BMFontRasterizer.cpp"
 return; //@line 156 "modules/font/BMFontRasterizer.cpp"
}
function __ZN4love4font16BMFontRasterizerD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love4font16BMFontRasterizerD2Ev($this); //@line 155 "modules/font/BMFontRasterizer.cpp"
 __ZdlPv($this); //@line 155 "modules/font/BMFontRasterizer.cpp"
 return; //@line 156 "modules/font/BMFontRasterizer.cpp"
}
function __ZNK4love4font16BMFontRasterizer13getLineHeightEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 104|0); //@line 277 "modules/font/BMFontRasterizer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 277 "modules/font/BMFontRasterizer.cpp"
 return ($1|0); //@line 277 "modules/font/BMFontRasterizer.cpp"
}
function __ZNK4love4font16BMFontRasterizer12getGlyphDataEj($this,$glyph) {
 $this = $this|0;
 $glyph = $glyph|0;
 var $$01 = 0, $$02 = 0, $$1 = 0, $$byval_copy = 0, $$byval_copy1 = 0, $$byval_copy2 = 0, $$pn$i$i = 0, $$pn$i$i10 = 0, $$pn$us$i$i = 0, $$pn$us$i$i5 = 0, $$sroa$0$0$i$i = 0, $$sroa$0$0$i$i15 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__nd$0$i$i = 0, $__nd$0$i$i12 = 0, $__nd$0$us$i$i = 0, $__nd$0$us$i$i7 = 0, $lock = 0, $y$027 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy2 = sp + 84|0;
 $$byval_copy1 = sp + 64|0;
 $$byval_copy = sp + 44|0;
 $0 = sp + 24|0;
 $1 = sp + 4|0;
 $lock = sp;
 $2 = ((($this)) + 60|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $4 = ($3|0)==(0); //@line 2047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L1: do {
  if (!($4)) {
   $5 = (($3) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $6 = $5 & $3; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $7 = ($6|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($7) {
    $8 = $5 & $glyph; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $13 = $8;
   } else {
    $9 = (($glyph>>>0) % ($3>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $13 = $9;
   }
   $10 = ((($this)) + 56|0); //@line 2015 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $11 = HEAP32[$10>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $12 = (($11) + ($13<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $14 = HEAP32[$12>>2]|0; //@line 2050 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $15 = ($14|0)==(0|0); //@line 2051 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($15)) {
    if ($7) {
     $$pn$us$i$i = $14;
     while(1) {
      $__nd$0$us$i$i = HEAP32[$$pn$us$i$i>>2]|0; //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $16 = ($__nd$0$us$i$i|0)==(0|0); //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($16) {
       break L1;
      }
      $17 = ((($__nd$0$us$i$i)) + 4|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $18 = HEAP32[$17>>2]|0; //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $19 = $18 & $5; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $20 = ($19|0)==($13|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($20)) {
       break L1;
      }
      $21 = ((($__nd$0$us$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $22 = HEAP32[$21>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $23 = ($22|0)==($glyph|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($23) {
       $$sroa$0$0$i$i = $__nd$0$us$i$i;
       break;
      } else {
       $$pn$us$i$i = $__nd$0$us$i$i;
      }
     }
    } else {
     $$pn$i$i = $14;
     while(1) {
      $__nd$0$i$i = HEAP32[$$pn$i$i>>2]|0; //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $24 = ($__nd$0$i$i|0)==(0|0); //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($24) {
       break L1;
      }
      $25 = ((($__nd$0$i$i)) + 4|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $26 = HEAP32[$25>>2]|0; //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $27 = (($26>>>0) % ($3>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $28 = ($27|0)==($13|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($28)) {
       break L1;
      }
      $29 = ((($__nd$0$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $30 = HEAP32[$29>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $31 = ($30|0)==($glyph|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($31) {
       $$sroa$0$0$i$i = $__nd$0$i$i;
       break;
      } else {
       $$pn$i$i = $__nd$0$i$i;
      }
     }
    }
    $37 = (__Znwj(40)|0); //@line 289 "modules/font/BMFontRasterizer.cpp"
    $38 = ((($$sroa$0$0$i$i)) + 24|0); //@line 289 "modules/font/BMFontRasterizer.cpp"
    __THREW__ = 0;
    ;HEAP32[$$byval_copy1>>2]=HEAP32[$38>>2]|0;HEAP32[$$byval_copy1+4>>2]=HEAP32[$38+4>>2]|0;HEAP32[$$byval_copy1+8>>2]=HEAP32[$38+8>>2]|0;HEAP32[$$byval_copy1+12>>2]=HEAP32[$38+12>>2]|0;HEAP32[$$byval_copy1+16>>2]=HEAP32[$38+16>>2]|0; //@line 289 "modules/font/BMFontRasterizer.cpp"
    invoke_viiii(3981,($37|0),($glyph|0),($$byval_copy1|0),1); //@line 289 "modules/font/BMFontRasterizer.cpp"
    $39 = __THREW__; __THREW__ = 0;
    $40 = $39&1;
    if ($40) {
     $79 = ___cxa_find_matching_catch_2()|0;
     $80 = tempRet0;
     __ZdlPv($37); //@line 289 "modules/font/BMFontRasterizer.cpp"
     $$01 = $79;$$02 = $80;
     ___resumeException($$01|0);
     // unreachable;
    }
    $41 = ((($$sroa$0$0$i$i)) + 20|0); //@line 291 "modules/font/BMFontRasterizer.cpp"
    $42 = HEAP32[$41>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    $43 = ((($this)) + 40|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $44 = HEAP32[$43>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $45 = ($44|0)==(0); //@line 2047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    L22: do {
     if (!($45)) {
      $46 = (($44) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $47 = $46 & $44; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $48 = ($47|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($48) {
       $49 = $46 & $42; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $54 = $49;
      } else {
       $50 = (($42>>>0) % ($44>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $54 = $50;
      }
      $51 = ((($this)) + 36|0); //@line 2015 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $52 = HEAP32[$51>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $53 = (($52) + ($54<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $55 = HEAP32[$53>>2]|0; //@line 2050 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $56 = ($55|0)==(0|0); //@line 2051 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($56)) {
       if ($48) {
        $$pn$us$i$i5 = $55;
        while(1) {
         $__nd$0$us$i$i7 = HEAP32[$$pn$us$i$i5>>2]|0; //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $57 = ($__nd$0$us$i$i7|0)==(0|0); //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         if ($57) {
          break L22;
         }
         $58 = ((($__nd$0$us$i$i7)) + 4|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $59 = HEAP32[$58>>2]|0; //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $60 = $59 & $46; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $61 = ($60|0)==($54|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         if (!($61)) {
          break L22;
         }
         $62 = ((($__nd$0$us$i$i7)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
         $63 = HEAP32[$62>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
         $64 = ($63|0)==($42|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
         if ($64) {
          $$sroa$0$0$i$i15 = $__nd$0$us$i$i7;
          break;
         } else {
          $$pn$us$i$i5 = $__nd$0$us$i$i7;
         }
        }
       } else {
        $$pn$i$i10 = $55;
        while(1) {
         $__nd$0$i$i12 = HEAP32[$$pn$i$i10>>2]|0; //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $65 = ($__nd$0$i$i12|0)==(0|0); //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         if ($65) {
          break L22;
         }
         $66 = ((($__nd$0$i$i12)) + 4|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $67 = HEAP32[$66>>2]|0; //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $68 = (($67>>>0) % ($44>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $69 = ($68|0)==($54|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         if (!($69)) {
          break L22;
         }
         $70 = ((($__nd$0$i$i12)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
         $71 = HEAP32[$70>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
         $72 = ($71|0)==($42|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
         if ($72) {
          $$sroa$0$0$i$i15 = $__nd$0$i$i12;
          break;
         } else {
          $$pn$i$i10 = $__nd$0$i$i12;
         }
        }
       }
       $83 = ((($$sroa$0$0$i$i15)) + 12|0); //@line 139 "./common/Object.h"
       $84 = HEAP32[$83>>2]|0; //@line 139 "./common/Object.h"
       $85 = HEAP32[$37>>2]|0; //@line 300 "modules/font/BMFontRasterizer.cpp"
       $86 = ((($85)) + 16|0); //@line 300 "modules/font/BMFontRasterizer.cpp"
       $87 = HEAP32[$86>>2]|0; //@line 300 "modules/font/BMFontRasterizer.cpp"
       $88 = (FUNCTION_TABLE_ii[$87 & 8191]($37)|0); //@line 300 "modules/font/BMFontRasterizer.cpp"
       $89 = HEAP32[$84>>2]|0; //@line 301 "modules/font/BMFontRasterizer.cpp"
       $90 = ((($89)) + 16|0); //@line 301 "modules/font/BMFontRasterizer.cpp"
       $91 = HEAP32[$90>>2]|0; //@line 301 "modules/font/BMFontRasterizer.cpp"
       $92 = (FUNCTION_TABLE_ii[$91 & 8191]($84)|0); //@line 301 "modules/font/BMFontRasterizer.cpp"
       $93 = (__ZNK4love5image9ImageData8getMutexEv($84)|0); //@line 303 "modules/font/BMFontRasterizer.cpp"
       __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$93); //@line 303 "modules/font/BMFontRasterizer.cpp"
       $94 = HEAP32[$38>>2]|0; //@line 306 "modules/font/BMFontRasterizer.cpp"
       $95 = ($94|0)>(0); //@line 306 "modules/font/BMFontRasterizer.cpp"
       L39: do {
        if ($95) {
         $96 = ((($$sroa$0$0$i$i)) + 16|0); //@line 308 "modules/font/BMFontRasterizer.cpp"
         $97 = ((($$sroa$0$0$i$i)) + 12|0); //@line 308 "modules/font/BMFontRasterizer.cpp"
         $98 = ((($$sroa$0$0$i$i)) + 28|0); //@line 309 "modules/font/BMFontRasterizer.cpp"
         $y$027 = 0;
         while(1) {
          $99 = HEAP32[$96>>2]|0; //@line 308 "modules/font/BMFontRasterizer.cpp"
          __THREW__ = 0;
          $100 = (invoke_ii(3984,($84|0))|0); //@line 308 "modules/font/BMFontRasterizer.cpp"
          $101 = __THREW__; __THREW__ = 0;
          $102 = $101&1;
          if ($102) {
           break;
          }
          $103 = (($99) + ($y$027))|0; //@line 308 "modules/font/BMFontRasterizer.cpp"
          $104 = Math_imul($103, $100)|0; //@line 308 "modules/font/BMFontRasterizer.cpp"
          $105 = HEAP32[$97>>2]|0; //@line 308 "modules/font/BMFontRasterizer.cpp"
          $106 = (($105) + ($104))|0; //@line 308 "modules/font/BMFontRasterizer.cpp"
          $107 = HEAP32[$98>>2]|0; //@line 309 "modules/font/BMFontRasterizer.cpp"
          $108 = Math_imul($107, $y$027)|0; //@line 309 "modules/font/BMFontRasterizer.cpp"
          $109 = (($88) + ($108<<2)|0); //@line 309 "modules/font/BMFontRasterizer.cpp"
          $110 = (($92) + ($106<<2)|0); //@line 309 "modules/font/BMFontRasterizer.cpp"
          $111 = $107 << 2; //@line 309 "modules/font/BMFontRasterizer.cpp"
          _memcpy(($109|0),($110|0),($111|0))|0; //@line 309 "modules/font/BMFontRasterizer.cpp"
          $112 = (($y$027) + 1)|0; //@line 306 "modules/font/BMFontRasterizer.cpp"
          $113 = HEAP32[$38>>2]|0; //@line 306 "modules/font/BMFontRasterizer.cpp"
          $114 = ($112|0)<($113|0); //@line 306 "modules/font/BMFontRasterizer.cpp"
          if ($114) {
           $y$027 = $112;
          } else {
           break L39;
          }
         }
         $115 = ___cxa_find_matching_catch_2()|0;
         $116 = tempRet0;
         __ZN4love6thread4LockD2Ev($lock); //@line 313 "modules/font/BMFontRasterizer.cpp"
         $$01 = $115;$$02 = $116;
         ___resumeException($$01|0);
         // unreachable;
        }
       } while(0);
       __ZN4love6thread4LockD2Ev($lock); //@line 313 "modules/font/BMFontRasterizer.cpp"
       $$1 = $37;
       STACKTOP = sp;return ($$1|0); //@line 313 "modules/font/BMFontRasterizer.cpp"
      }
     }
    } while(0);
    $73 = HEAP32[$37>>2]|0; //@line 295 "modules/font/BMFontRasterizer.cpp"
    $74 = ((($73)) + 12|0); //@line 295 "modules/font/BMFontRasterizer.cpp"
    $75 = HEAP32[$74>>2]|0; //@line 295 "modules/font/BMFontRasterizer.cpp"
    FUNCTION_TABLE_vi[$75 & 8191]($37); //@line 295 "modules/font/BMFontRasterizer.cpp"
    $76 = (__Znwj(40)|0); //@line 296 "modules/font/BMFontRasterizer.cpp"
    ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0;HEAP32[$1+12>>2]=0|0;HEAP32[$1+16>>2]=0|0; //@line 296 "modules/font/BMFontRasterizer.cpp"
    __THREW__ = 0;
    ;HEAP32[$$byval_copy2>>2]=HEAP32[$1>>2]|0;HEAP32[$$byval_copy2+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$$byval_copy2+8>>2]=HEAP32[$1+8>>2]|0;HEAP32[$$byval_copy2+12>>2]=HEAP32[$1+12>>2]|0;HEAP32[$$byval_copy2+16>>2]=HEAP32[$1+16>>2]|0; //@line 296 "modules/font/BMFontRasterizer.cpp"
    invoke_viiii(3981,($76|0),($glyph|0),($$byval_copy2|0),1); //@line 296 "modules/font/BMFontRasterizer.cpp"
    $77 = __THREW__; __THREW__ = 0;
    $78 = $77&1;
    if (!($78)) {
     $$1 = $76;
     STACKTOP = sp;return ($$1|0); //@line 313 "modules/font/BMFontRasterizer.cpp"
    }
    $81 = ___cxa_find_matching_catch_2()|0;
    $82 = tempRet0;
    __ZdlPv($76); //@line 296 "modules/font/BMFontRasterizer.cpp"
    $$01 = $81;$$02 = $82;
    ___resumeException($$01|0);
    // unreachable;
   }
  }
 } while(0);
 $32 = (__Znwj(40)|0); //@line 286 "modules/font/BMFontRasterizer.cpp"
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;HEAP32[$0+12>>2]=0|0;HEAP32[$0+16>>2]=0|0; //@line 286 "modules/font/BMFontRasterizer.cpp"
 __THREW__ = 0;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;HEAP32[$$byval_copy+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$$byval_copy+8>>2]=HEAP32[$0+8>>2]|0;HEAP32[$$byval_copy+12>>2]=HEAP32[$0+12>>2]|0;HEAP32[$$byval_copy+16>>2]=HEAP32[$0+16>>2]|0; //@line 286 "modules/font/BMFontRasterizer.cpp"
 invoke_viiii(3981,($32|0),($glyph|0),($$byval_copy|0),1); //@line 286 "modules/font/BMFontRasterizer.cpp"
 $33 = __THREW__; __THREW__ = 0;
 $34 = $33&1;
 if (!($34)) {
  $$1 = $32;
  STACKTOP = sp;return ($$1|0); //@line 313 "modules/font/BMFontRasterizer.cpp"
 }
 $35 = ___cxa_find_matching_catch_2()|0;
 $36 = tempRet0;
 __ZdlPv($32); //@line 286 "modules/font/BMFontRasterizer.cpp"
 $$01 = $35;$$02 = $36;
 ___resumeException($$01|0);
 // unreachable;
 return (0)|0;
}
function __ZNK4love4font16BMFontRasterizer13getGlyphCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 68|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 return ($1|0); //@line 317 "modules/font/BMFontRasterizer.cpp"
}
function __ZNK4love4font16BMFontRasterizer8hasGlyphEj($this,$glyph) {
 $this = $this|0;
 $glyph = $glyph|0;
 var $$pn$i$i = 0, $$pn$us$i$i = 0, $$sroa$0$0$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__nd$0$i$i = 0, $__nd$0$us$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 60|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $2 = ($1|0)==(0); //@line 2047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L1: do {
  if ($2) {
   $$sroa$0$0$i$i = 0;
  } else {
   $3 = (($1) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $4 = $3 & $1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $5 = ($4|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($5) {
    $6 = $3 & $glyph; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $11 = $6;
   } else {
    $7 = (($glyph>>>0) % ($1>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $11 = $7;
   }
   $8 = ((($this)) + 56|0); //@line 2015 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $9 = HEAP32[$8>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $10 = (($9) + ($11<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $12 = HEAP32[$10>>2]|0; //@line 2050 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $13 = ($12|0)==(0|0); //@line 2051 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($13) {
    $$sroa$0$0$i$i = 0;
   } else {
    if ($5) {
     $$pn$us$i$i = $12;
     while(1) {
      $__nd$0$us$i$i = HEAP32[$$pn$us$i$i>>2]|0; //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $14 = ($__nd$0$us$i$i|0)==(0|0); //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($14) {
       $$sroa$0$0$i$i = 0;
       break L1;
      }
      $15 = ((($__nd$0$us$i$i)) + 4|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $16 = HEAP32[$15>>2]|0; //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $17 = $16 & $3; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $18 = ($17|0)==($11|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($18)) {
       $$sroa$0$0$i$i = 0;
       break L1;
      }
      $19 = ((($__nd$0$us$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $20 = HEAP32[$19>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $21 = ($20|0)==($glyph|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($21) {
       $$sroa$0$0$i$i = $__nd$0$us$i$i;
       break;
      } else {
       $$pn$us$i$i = $__nd$0$us$i$i;
      }
     }
    } else {
     $$pn$i$i = $12;
     while(1) {
      $__nd$0$i$i = HEAP32[$$pn$i$i>>2]|0; //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $22 = ($__nd$0$i$i|0)==(0|0); //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($22) {
       $$sroa$0$0$i$i = 0;
       break L1;
      }
      $23 = ((($__nd$0$i$i)) + 4|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $24 = HEAP32[$23>>2]|0; //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $25 = (($24>>>0) % ($1>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $26 = ($25|0)==($11|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($26)) {
       $$sroa$0$0$i$i = 0;
       break L1;
      }
      $27 = ((($__nd$0$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $28 = HEAP32[$27>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $29 = ($28|0)==($glyph|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($29) {
       $$sroa$0$0$i$i = $__nd$0$i$i;
       break;
      } else {
       $$pn$i$i = $__nd$0$i$i;
      }
     }
    }
   }
  }
 } while(0);
 $30 = ($$sroa$0$0$i$i|0)!=(0|0); //@line 340 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 return ($30|0); //@line 322 "modules/font/BMFontRasterizer.cpp"
}
function __ZNK4love4font16BMFontRasterizer10getKerningEjj($this,$leftglyph,$rightglyph) {
 $this = $this|0;
 $leftglyph = $leftglyph|0;
 $rightglyph = $rightglyph|0;
 var $$0 = 0.0, $$pn$i$i = 0, $$pn$us$i$i = 0, $$sroa$0$0$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0.0, $7 = 0, $8 = 0, $9 = 0, $__nd$0$i$i = 0, $__nd$0$us$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = Math_imul($rightglyph, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = $0 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = $1 ^ $0; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = Math_imul($2, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $4 = $3 ^ -561034072; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = Math_imul($leftglyph, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = $5 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $7 = $6 ^ $5; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = Math_imul($7, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = Math_imul($4, 1540483477)|0; //@line 3104 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $10 = $9 ^ $8; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $11 = $10 >>> 13; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $12 = $11 ^ $10; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $13 = Math_imul($12, 1540483477)|0; //@line 3118 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = $13 >>> 15; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $15 = $14 ^ $13; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $16 = ((($this)) + 80|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $17 = HEAP32[$16>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $18 = ($17|0)==(0); //@line 2047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($18) {
  $$0 = 0.0;
  return (+$$0); //@line 334 "modules/font/BMFontRasterizer.cpp"
 }
 $19 = (($17) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $20 = $19 & $17; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $21 = ($20|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($21) {
  $22 = $19 & $15; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $22;
 } else {
  $23 = (($15>>>0) % ($17>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $23;
 }
 $24 = ((($this)) + 76|0); //@line 2015 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $25 = HEAP32[$24>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $26 = (($25) + ($27<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $28 = HEAP32[$26>>2]|0; //@line 2050 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $29 = ($28|0)==(0|0); //@line 2051 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($29) {
  $$0 = 0.0;
  return (+$$0); //@line 334 "modules/font/BMFontRasterizer.cpp"
 }
 L11: do {
  if ($21) {
   $$pn$us$i$i = $28;
   while(1) {
    $__nd$0$us$i$i = HEAP32[$$pn$us$i$i>>2]|0; //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $30 = ($__nd$0$us$i$i|0)==(0|0); //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($30) {
     $$0 = 0.0;
     label = 14;
     break;
    }
    $31 = ((($__nd$0$us$i$i)) + 4|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $32 = HEAP32[$31>>2]|0; //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $33 = $32 & $19; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $34 = ($33|0)==($27|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if (!($34)) {
     $$0 = 0.0;
     label = 14;
     break;
    }
    $35 = ((($__nd$0$us$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
    $36 = $35; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    $37 = $36; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    $38 = HEAP32[$37>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    $39 = (($36) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    $40 = $39; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    $41 = HEAP32[$40>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    $42 = ($38|0)==($rightglyph|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    $43 = ($41|0)==($leftglyph|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    $44 = $42 & $43; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    if ($44) {
     $$sroa$0$0$i$i = $__nd$0$us$i$i;
     break L11;
    } else {
     $$pn$us$i$i = $__nd$0$us$i$i;
    }
   }
   if ((label|0) == 14) {
    return (+$$0); //@line 334 "modules/font/BMFontRasterizer.cpp"
   }
  } else {
   $$pn$i$i = $28;
   while(1) {
    $__nd$0$i$i = HEAP32[$$pn$i$i>>2]|0; //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $45 = ($__nd$0$i$i|0)==(0|0); //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($45) {
     $$0 = 0.0;
     label = 14;
     break;
    }
    $46 = ((($__nd$0$i$i)) + 4|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $47 = HEAP32[$46>>2]|0; //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $48 = (($47>>>0) % ($17>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $49 = ($48|0)==($27|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if (!($49)) {
     $$0 = 0.0;
     label = 14;
     break;
    }
    $50 = ((($__nd$0$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
    $51 = $50; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    $52 = $51; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    $53 = HEAP32[$52>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    $54 = (($51) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    $55 = $54; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    $56 = HEAP32[$55>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    $57 = ($53|0)==($rightglyph|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    $58 = ($56|0)==($leftglyph|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    $59 = $57 & $58; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
    if ($59) {
     $$sroa$0$0$i$i = $__nd$0$i$i;
     break L11;
    } else {
     $$pn$i$i = $__nd$0$i$i;
    }
   }
   if ((label|0) == 14) {
    return (+$$0); //@line 334 "modules/font/BMFontRasterizer.cpp"
   }
  }
 } while(0);
 $60 = ((($$sroa$0$0$i$i)) + 16|0); //@line 331 "modules/font/BMFontRasterizer.cpp"
 $61 = HEAP32[$60>>2]|0; //@line 331 "modules/font/BMFontRasterizer.cpp"
 $62 = (+($61|0)); //@line 331 "modules/font/BMFontRasterizer.cpp"
 $$0 = $62;
 return (+$$0); //@line 334 "modules/font/BMFontRasterizer.cpp"
}
function __ZN4love4font16BMFontRasterizerC2EPNS_10filesystem8FileDataERKNSt3__16vectorIPNS_5image9ImageDataENS5_9allocatorIS9_EEEE($this,$fontdef,$imagelist) {
 $this = $this|0;
 $fontdef = $fontdef|0;
 $imagelist = $imagelist|0;
 var $$0 = 0, $$01 = 0, $$01$i$i$i = 0, $$01$i$i$i2 = 0, $$01$i$i$i8 = 0, $$lcssa = 0, $$lcssa52 = 0, $$pre$phi24Z2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $configtext = 0, $i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 16|0;
 $i = sp + 12|0;
 $configtext = sp;
 __ZN4love6ObjectC2Ev($this); //@line 48 "modules/font/Rasterizer.h"
 HEAP32[$this>>2] = (24000); //@line 138 "modules/font/BMFontRasterizer.cpp"
 $1 = ((($this)) + 24|0); //@line 134 "modules/font/BMFontRasterizer.cpp"
 $2 = ((($this)) + 36|0); //@line 134 "modules/font/BMFontRasterizer.cpp"
 $3 = ((($this)) + 52|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0;HEAP32[$1+12>>2]=0|0;HEAP32[$1+16>>2]=0|0;HEAP32[$1+20>>2]=0|0;HEAP32[$1+24>>2]=0|0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAPF32[$3>>2] = 1.0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $4 = ((($this)) + 56|0); //@line 134 "modules/font/BMFontRasterizer.cpp"
 $5 = ((($this)) + 72|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;HEAP32[$4+12>>2]=0|0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAPF32[$5>>2] = 1.0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = ((($this)) + 76|0); //@line 134 "modules/font/BMFontRasterizer.cpp"
 $7 = ((($this)) + 92|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;HEAP32[$6+12>>2]=0|0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAPF32[$7>>2] = 1.0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = ((($this)) + 96|0); //@line 135 "modules/font/BMFontRasterizer.cpp"
 HEAP32[$8>>2] = 0; //@line 135 "modules/font/BMFontRasterizer.cpp"
 $9 = ((($this)) + 100|0); //@line 136 "modules/font/BMFontRasterizer.cpp"
 HEAP8[$9>>0] = 0; //@line 136 "modules/font/BMFontRasterizer.cpp"
 $10 = ((($this)) + 104|0); //@line 137 "modules/font/BMFontRasterizer.cpp"
 HEAP32[$10>>2] = 0; //@line 137 "modules/font/BMFontRasterizer.cpp"
 __THREW__ = 0;
 $11 = (invoke_ii(3461,($fontdef|0))|0); //@line 139 "modules/font/BMFontRasterizer.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 L1: do {
  if ($13) {
   label = 11;
  } else {
   $14 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE5rfindEcj($11,47,-1)|0); //@line 141 "modules/font/BMFontRasterizer.cpp"
   $15 = ($14|0)==(-1); //@line 142 "modules/font/BMFontRasterizer.cpp"
   do {
    if (!($15)) {
     __THREW__ = 0;
     invoke_viiiii(3511,($0|0),($11|0),0,($14|0),($11|0)); //@line 3345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $16 = __THREW__; __THREW__ = 0;
     $17 = $16&1;
     if ($17) {
      label = 11;
      break L1;
     }
     $18 = HEAP8[$1>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $19 = $18 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $20 = ($19<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($20) {
      $24 = ((($1)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$24>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$1>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     } else {
      $21 = ((($this)) + 32|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $22 = HEAP32[$21>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP8[$22>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $23 = ((($this)) + 28|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      HEAP32[$23>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     }
     __THREW__ = 0;
     invoke_vii(3368,($1|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $25 = __THREW__; __THREW__ = 0;
     $26 = $25&1;
     if ($26) {
      $27 = ___cxa_find_matching_catch_3(0|0)|0;
      $28 = tempRet0;
      ___clang_call_terminate($27); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     } else {
      ;HEAP32[$1>>2]=HEAP32[$0>>2]|0;HEAP32[$1+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$1+8>>2]=HEAP32[$0+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 143 "modules/font/BMFontRasterizer.cpp"
      break;
     }
    }
   } while(0);
   HEAP32[$i>>2] = 0; //@line 146 "modules/font/BMFontRasterizer.cpp"
   $33 = ((($imagelist)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $34 = HEAP32[$33>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $35 = HEAP32[$imagelist>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $36 = (($34) - ($35))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $37 = ($36|0)>(0); //@line 146 "modules/font/BMFontRasterizer.cpp"
   L14: do {
    if ($37) {
     $49 = 0;
     while(1) {
      __THREW__ = 0;
      $44 = (invoke_iii(3985,($2|0),($i|0))|0); //@line 147 "modules/font/BMFontRasterizer.cpp"
      $45 = __THREW__; __THREW__ = 0;
      $46 = $45&1;
      if ($46) {
       label = 10;
       break;
      }
      $47 = HEAP32[$imagelist>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $48 = (($47) + ($49<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $50 = HEAP32[$48>>2]|0; //@line 147 "modules/font/BMFontRasterizer.cpp"
      $51 = ($50|0)==(0|0); //@line 95 "./common/Object.h"
      if (!($51)) {
       $52 = HEAP32[$50>>2]|0; //@line 95 "./common/Object.h"
       $53 = ((($52)) + 8|0); //@line 95 "./common/Object.h"
       $54 = HEAP32[$53>>2]|0; //@line 95 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($54|0,($50|0)); //@line 95 "./common/Object.h"
       $55 = __THREW__; __THREW__ = 0;
       $56 = $55&1;
       if ($56) {
        label = 10;
        break;
       }
       $57 = HEAP32[$50>>2]|0; //@line 132 "./common/Object.h"
       $58 = ((($57)) + 8|0); //@line 132 "./common/Object.h"
       $59 = HEAP32[$58>>2]|0; //@line 132 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($59|0,($50|0)); //@line 132 "./common/Object.h"
       $60 = __THREW__; __THREW__ = 0;
       $61 = $60&1;
       if ($61) {
        $$lcssa = $50;$$lcssa52 = $50;
        label = 20;
        break;
       }
      }
      $64 = HEAP32[$44>>2]|0; //@line 133 "./common/Object.h"
      $65 = ($64|0)==(0|0); //@line 133 "./common/Object.h"
      if (!($65)) {
       $66 = HEAP32[$64>>2]|0; //@line 133 "./common/Object.h"
       $67 = ((($66)) + 12|0); //@line 133 "./common/Object.h"
       $68 = HEAP32[$67>>2]|0; //@line 133 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($68|0,($64|0)); //@line 133 "./common/Object.h"
       $69 = __THREW__; __THREW__ = 0;
       $70 = $69&1;
       if ($70) {
        label = 27;
        break;
       }
      }
      HEAP32[$44>>2] = $50; //@line 134 "./common/Object.h"
      if (!($51)) {
       $71 = HEAP32[$50>>2]|0; //@line 106 "./common/Object.h"
       $72 = ((($71)) + 12|0); //@line 106 "./common/Object.h"
       $73 = HEAP32[$72>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($73|0,($50|0)); //@line 106 "./common/Object.h"
       $74 = __THREW__; __THREW__ = 0;
       $75 = $74&1;
       if ($75) {
        label = 25;
        break;
       }
      }
      $78 = (($49) + 1)|0; //@line 146 "modules/font/BMFontRasterizer.cpp"
      HEAP32[$i>>2] = $78; //@line 146 "modules/font/BMFontRasterizer.cpp"
      $79 = HEAP32[$33>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $80 = HEAP32[$imagelist>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $81 = (($79) - ($80))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $82 = $81 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $83 = ($78|0)<($82|0); //@line 146 "modules/font/BMFontRasterizer.cpp"
      if ($83) {
       $49 = $78;
      } else {
       break L14;
      }
     }
     if ((label|0) == 10) {
      $29 = ___cxa_find_matching_catch_2()|0;
      $30 = tempRet0;
      $lpad$phi$index = $29;$lpad$phi$index2 = $30;
      label = 12;
      break L1;
     }
     else if ((label|0) == 20) {
      $62 = ___cxa_find_matching_catch_2()|0;
      $63 = tempRet0;
      $$pre$phi24Z2D = $$lcssa52;$$pre$phiZ2D = $$lcssa;$136 = $63;$137 = $62;
     }
     else if ((label|0) == 25) {
      $76 = ___cxa_find_matching_catch_3(0|0)|0;
      $77 = tempRet0;
      ___clang_call_terminate($76); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     else if ((label|0) == 27) {
      $84 = ___cxa_find_matching_catch_2()|0;
      $85 = tempRet0;
      if ($51) {
       $$0 = $84;$$01 = $85;
       break L1;
      }
      $$pre$phi24Z2D = $50;$$pre$phiZ2D = $50;$136 = $85;$137 = $84;
     }
     $86 = HEAP32[$$pre$phi24Z2D>>2]|0; //@line 106 "./common/Object.h"
     $87 = ((($86)) + 12|0); //@line 106 "./common/Object.h"
     $88 = HEAP32[$87>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($88|0,($$pre$phiZ2D|0)); //@line 106 "./common/Object.h"
     $89 = __THREW__; __THREW__ = 0;
     $90 = $89&1;
     if (!($90)) {
      $$0 = $137;$$01 = $136;
      break L1;
     }
     $91 = ___cxa_find_matching_catch_3(0|0)|0;
     $92 = tempRet0;
     ___clang_call_terminate($91); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   } while(0);
   $38 = HEAP32[$fontdef>>2]|0; //@line 149 "modules/font/BMFontRasterizer.cpp"
   $39 = ((($38)) + 16|0); //@line 149 "modules/font/BMFontRasterizer.cpp"
   $40 = HEAP32[$39>>2]|0; //@line 149 "modules/font/BMFontRasterizer.cpp"
   __THREW__ = 0;
   $41 = (invoke_ii($40|0,($fontdef|0))|0); //@line 149 "modules/font/BMFontRasterizer.cpp"
   $42 = __THREW__; __THREW__ = 0;
   $43 = $42&1;
   if ($43) {
    label = 11;
   } else {
    $93 = HEAP32[$fontdef>>2]|0; //@line 149 "modules/font/BMFontRasterizer.cpp"
    $94 = ((($93)) + 20|0); //@line 149 "modules/font/BMFontRasterizer.cpp"
    $95 = HEAP32[$94>>2]|0; //@line 149 "modules/font/BMFontRasterizer.cpp"
    __THREW__ = 0;
    $96 = (invoke_ii($95|0,($fontdef|0))|0); //@line 149 "modules/font/BMFontRasterizer.cpp"
    $97 = __THREW__; __THREW__ = 0;
    $98 = $97&1;
    if ($98) {
     label = 11;
    } else {
     __THREW__ = 0;
     invoke_viii(3460,($configtext|0),($41|0),($96|0)); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $99 = __THREW__; __THREW__ = 0;
     $100 = $99&1;
     if ($100) {
      label = 11;
     } else {
      __THREW__ = 0;
      invoke_vii(3986,($this|0),($configtext|0)); //@line 151 "modules/font/BMFontRasterizer.cpp"
      $101 = __THREW__; __THREW__ = 0;
      $102 = $101&1;
      if ($102) {
       $103 = ___cxa_find_matching_catch_2()|0;
       $104 = tempRet0;
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($configtext); //@line 152 "modules/font/BMFontRasterizer.cpp"
       $$0 = $103;$$01 = $104;
       break;
      } else {
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($configtext); //@line 152 "modules/font/BMFontRasterizer.cpp"
       STACKTOP = sp;return; //@line 152 "modules/font/BMFontRasterizer.cpp"
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 11) {
  $31 = ___cxa_find_matching_catch_2()|0;
  $32 = tempRet0;
  $lpad$phi$index = $31;$lpad$phi$index2 = $32;
  label = 12;
 }
 if ((label|0) == 12) {
  $$0 = $lpad$phi$index;$$01 = $lpad$phi$index2;
 }
 $105 = ((($this)) + 84|0); //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $106 = HEAP32[$105>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $107 = ($106|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($107)) {
  $$01$i$i$i8 = $106;
  while(1) {
   $108 = HEAP32[$$01$i$i$i8>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   __ZdlPv($$01$i$i$i8); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $109 = ($108|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($109) {
    break;
   } else {
    $$01$i$i$i8 = $108;
   }
  }
 }
 $110 = HEAP32[$6>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$6>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $111 = ($110|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($111)) {
  __ZdlPv($110); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $112 = ((($this)) + 64|0); //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $113 = HEAP32[$112>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $114 = ($113|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($114)) {
  $$01$i$i$i2 = $113;
  while(1) {
   $115 = HEAP32[$$01$i$i$i2>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   __ZdlPv($$01$i$i$i2); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $116 = ($115|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($116) {
    break;
   } else {
    $$01$i$i$i2 = $115;
   }
  }
 }
 $117 = HEAP32[$4>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$4>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $118 = ($117|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($118)) {
  __ZdlPv($117); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $119 = ((($this)) + 44|0); //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $120 = HEAP32[$119>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $121 = ($120|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L63: do {
  if (!($121)) {
   $$01$i$i$i = $120;
   while(1) {
    $122 = HEAP32[$$01$i$i$i>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $123 = ((($$01$i$i$i)) + 12|0); //@line 106 "./common/Object.h"
    $124 = HEAP32[$123>>2]|0; //@line 106 "./common/Object.h"
    $125 = ($124|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($125)) {
     $126 = HEAP32[$124>>2]|0; //@line 106 "./common/Object.h"
     $127 = ((($126)) + 12|0); //@line 106 "./common/Object.h"
     $128 = HEAP32[$127>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($128|0,($124|0)); //@line 106 "./common/Object.h"
     $129 = __THREW__; __THREW__ = 0;
     $130 = $129&1;
     if ($130) {
      break;
     }
    }
    __ZdlPv($$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $133 = ($122|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($133) {
     break L63;
    } else {
     $$01$i$i$i = $122;
    }
   }
   $131 = ___cxa_find_matching_catch_3(0|0)|0;
   $132 = tempRet0;
   ___clang_call_terminate($131); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 } while(0);
 $134 = HEAP32[$2>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$2>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $135 = ($134|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($135) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 152 "modules/font/BMFontRasterizer.cpp"
  __ZN4love4font10RasterizerD2Ev($this); //@line 152 "modules/font/BMFontRasterizer.cpp"
  ___resumeException($$0|0);
  // unreachable;
 }
 __ZdlPv($134); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 152 "modules/font/BMFontRasterizer.cpp"
 __ZN4love4font10RasterizerD2Ev($this); //@line 152 "modules/font/BMFontRasterizer.cpp"
 ___resumeException($$0|0);
 // unreachable;
}
function __ZNSt3__113unordered_mapIiN4love9StrongRefINS1_5image9ImageDataEEENS_4hashIiEENS_8equal_toIiEENS_9allocatorINS_4pairIKiS5_EEEEEixERSC_($this,$__k) {
 $this = $this|0;
 $__k = $__k|0;
 var $$0 = 0, $$pn$i$i = 0, $$pn$us$i$i = 0, $$sroa$0$0$i$i$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__nd$0$i$i = 0, $__nd$0$us$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = HEAP32[$__k>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $4 = ($3|0)==(0); //@line 2018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L1: do {
  if (!($4)) {
   $5 = (($3) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $6 = $5 & $3; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $7 = ($6|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($7) {
    $8 = $5 & $1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $12 = $8;
   } else {
    $9 = (($1>>>0) % ($3>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $12 = $9;
   }
   $10 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $11 = (($10) + ($12<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $13 = HEAP32[$11>>2]|0; //@line 2021 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $14 = ($13|0)==(0|0); //@line 2022 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($14)) {
    if ($7) {
     $$pn$us$i$i = $13;
     while(1) {
      $__nd$0$us$i$i = HEAP32[$$pn$us$i$i>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $15 = ($__nd$0$us$i$i|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($15) {
       break L1;
      }
      $16 = ((($__nd$0$us$i$i)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $17 = HEAP32[$16>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $18 = $17 & $5; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $19 = ($18|0)==($12|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($19)) {
       break L1;
      }
      $20 = ((($__nd$0$us$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $21 = HEAP32[$20>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $22 = ($21|0)==($1|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($22) {
       $$sroa$0$0$i$i$sink = $__nd$0$us$i$i;
       break;
      } else {
       $$pn$us$i$i = $__nd$0$us$i$i;
      }
     }
     $$0 = ((($$sroa$0$0$i$i$sink)) + 12|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
     STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
    } else {
     $$pn$i$i = $13;
     while(1) {
      $__nd$0$i$i = HEAP32[$$pn$i$i>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $23 = ($__nd$0$i$i|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($23) {
       break L1;
      }
      $24 = ((($__nd$0$i$i)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $25 = HEAP32[$24>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $26 = (($25>>>0) % ($3>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $27 = ($26|0)==($12|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($27)) {
       break L1;
      }
      $28 = ((($__nd$0$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $29 = HEAP32[$28>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $30 = ($29|0)==($1|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($30) {
       $$sroa$0$0$i$i$sink = $__nd$0$i$i;
       break;
      } else {
       $$pn$i$i = $__nd$0$i$i;
      }
     }
     $$0 = ((($$sroa$0$0$i$i$sink)) + 12|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
     STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
    }
   }
  }
 } while(0);
 $31 = (__Znwj(16)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $32 = ((($31)) + 8|0); //@line 1340 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 HEAP32[$32>>2] = $1; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = ((($31)) + 12|0); //@line 1342 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 HEAP32[$33>>2] = 0; //@line 88 "./common/Object.h"
 __THREW__ = 0;
 invoke_viii(3987,($0|0),($this|0),($31|0)); //@line 1366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $34 = __THREW__; __THREW__ = 0;
 $35 = $34&1;
 if (!($35)) {
  $36 = HEAP32[$0>>2]|0; //@line 315 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
  $$sroa$0$0$i$i$sink = $36;
  $$0 = ((($$sroa$0$0$i$i$sink)) + 12|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
  STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 }
 $37 = ___cxa_find_matching_catch_2()|0;
 $38 = tempRet0;
 $39 = HEAP32[$33>>2]|0; //@line 106 "./common/Object.h"
 $40 = ($39|0)==(0|0); //@line 106 "./common/Object.h"
 if ($40) {
  __ZdlPv($31); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($37|0);
  // unreachable;
 }
 $41 = HEAP32[$39>>2]|0; //@line 106 "./common/Object.h"
 $42 = ((($41)) + 12|0); //@line 106 "./common/Object.h"
 $43 = HEAP32[$42>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($43|0,($39|0)); //@line 106 "./common/Object.h"
 $44 = __THREW__; __THREW__ = 0;
 $45 = $44&1;
 if ($45) {
  $46 = ___cxa_find_matching_catch_3(0|0)|0;
  $47 = tempRet0;
  ___clang_call_terminate($46); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  __ZdlPv($31); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($37|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeIiN4love9StrongRefINS2_5image9ImageDataEEEEENS_22__unordered_map_hasherIiS7_NS_4hashIiEELb1EEENS_21__unordered_map_equalIiS7_NS_8equal_toIiEELb1EEENS_9allocatorIS7_EEE20__node_insert_uniqueEPNS_11__hash_nodeIS7_PvEE($agg$result,$this,$__nd) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $__nd = $__nd|0;
 var $$pn = 0, $$pn$us = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $__bc$0 = 0, $__chash$0 = 0, $__chash$1 = 0, $__inserted$0 = 0, $__ndptr$0 = 0, $__ndptr$0$us = 0, $__ndptr$1 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($__nd)) + 8|0); //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $1 = HEAP32[$0>>2]|0; //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $2 = ((($__nd)) + 4|0); //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$2>>2] = $1; //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $3 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $4 = HEAP32[$3>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $5 = ($4|0)==(0); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L1: do {
  if ($5) {
   $__chash$0 = 0;
  } else {
   $6 = (($4) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $7 = $6 & $4; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $8 = ($7|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($8) {
    $9 = $6 & $1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $13 = $9;
   } else {
    $10 = (($1>>>0) % ($4>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $13 = $10;
   }
   $11 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $12 = (($11) + ($13<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $14 = HEAP32[$12>>2]|0; //@line 1603 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $15 = ($14|0)==(0|0); //@line 1604 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($15) {
    $__chash$0 = $13;
   } else {
    if ($8) {
     $$pn$us = $14;
     while(1) {
      $__ndptr$0$us = HEAP32[$$pn$us>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $16 = ($__ndptr$0$us|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($16) {
       $__chash$0 = $13;
       break L1;
      }
      $17 = ((($__ndptr$0$us)) + 4|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $18 = HEAP32[$17>>2]|0; //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $19 = $18 & $6; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $20 = ($19|0)==($13|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($20)) {
       $__chash$0 = $13;
       break L1;
      }
      $21 = ((($__ndptr$0$us)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $22 = HEAP32[$21>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $23 = ($22|0)==($1|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($23) {
       $__inserted$0 = 0;$__ndptr$1 = $__ndptr$0$us;
       break;
      } else {
       $$pn$us = $__ndptr$0$us;
      }
     }
     $85 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     HEAP32[$agg$result>>2] = $85; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     $86 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     HEAP8[$86>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    } else {
     $$pn = $14;
     while(1) {
      $__ndptr$0 = HEAP32[$$pn>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $24 = ($__ndptr$0|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($24) {
       $__chash$0 = $13;
       break L1;
      }
      $25 = ((($__ndptr$0)) + 4|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $26 = HEAP32[$25>>2]|0; //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $27 = (($26>>>0) % ($4>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $28 = ($27|0)==($13|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($28)) {
       $__chash$0 = $13;
       break L1;
      }
      $29 = ((($__ndptr$0)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $30 = HEAP32[$29>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $31 = ($30|0)==($1|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($31) {
       $__inserted$0 = 0;$__ndptr$1 = $__ndptr$0;
       break;
      } else {
       $$pn = $__ndptr$0;
      }
     }
     $85 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     HEAP32[$agg$result>>2] = $85; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     $86 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     HEAP8[$86>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    }
   }
  }
 } while(0);
 $32 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = HEAP32[$32>>2]|0; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $34 = (($33) + 1)|0; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $35 = (+($34>>>0)); //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $36 = (+($4>>>0)); //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $37 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = +HEAPF32[$37>>2]; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $39 = $36 * $38; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $40 = $35 > $39; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $or$cond = $5 | $40; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 do {
  if ($or$cond) {
   $41 = $4 << 1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $42 = ($4>>>0)>(2); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($42) {
    $43 = (($4) + -1)|0; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $44 = $43 & $4; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $45 = ($44|0)==(0); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $47 = $45;
   } else {
    $47 = 0;
   }
   $46 = $47&1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $48 = $46 | $41; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $49 = $48 ^ 1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $50 = $35 / $38; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $51 = (+Math_ceil((+$50))); //@line 750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/cmath"
   $52 = (~~(($51))>>>0); //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $53 = ($49>>>0)<($52>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $54 = $53 ? $52 : $49; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   __ZNSt3__112__hash_tableINS_17__hash_value_typeIiN4love9StrongRefINS2_5image9ImageDataEEEEENS_22__unordered_map_hasherIiS7_NS_4hashIiEELb1EEENS_21__unordered_map_equalIiS7_NS_8equal_toIiEELb1EEENS_9allocatorIS7_EEE6rehashEj($this,$54); //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $55 = HEAP32[$3>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $56 = HEAP32[$2>>2]|0; //@line 1621 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $57 = (($55) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $58 = $57 & $55; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $59 = ($58|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($59) {
    $60 = $57 & $56; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__bc$0 = $55;$__chash$1 = $60;
    break;
   } else {
    $61 = (($56>>>0) % ($55>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__bc$0 = $55;$__chash$1 = $61;
    break;
   }
  } else {
   $__bc$0 = $4;$__chash$1 = $__chash$0;
  }
 } while(0);
 $62 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $63 = (($62) + ($__chash$1<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $64 = HEAP32[$63>>2]|0; //@line 1624 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $65 = ($64|0)==(0|0); //@line 1625 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($65) {
  $66 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $67 = HEAP32[$66>>2]|0; //@line 1628 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$__nd>>2] = $67; //@line 1628 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$66>>2] = $__nd; //@line 1629 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $68 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $69 = (($68) + ($__chash$1<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$69>>2] = $66; //@line 1631 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $70 = HEAP32[$__nd>>2]|0; //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $71 = ($70|0)==(0|0); //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if (!($71)) {
   $72 = ((($70)) + 4|0); //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $73 = HEAP32[$72>>2]|0; //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $74 = (($__bc$0) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $75 = $74 & $__bc$0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $76 = ($75|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($76) {
    $77 = $73 & $74; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $81 = $77;
   } else {
    $78 = (($73>>>0) % ($__bc$0>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $81 = $78;
   }
   $79 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $80 = (($79) + ($81<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$80>>2] = $__nd; //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 } else {
  $82 = HEAP32[$64>>2]|0; //@line 1637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$__nd>>2] = $82; //@line 1637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$64>>2] = $__nd; //@line 1638 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $83 = HEAP32[$32>>2]|0; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $84 = (($83) + 1)|0; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$32>>2] = $84; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__inserted$0 = 1;$__ndptr$1 = $__nd;
 $85 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$agg$result>>2] = $85; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $86 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 HEAP8[$86>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeIiN4love9StrongRefINS2_5image9ImageDataEEEEENS_22__unordered_map_hasherIiS7_NS_4hashIiEELb1EEENS_21__unordered_map_equalIiS7_NS_8equal_toIiEELb1EEENS_9allocatorIS7_EEE6rehashEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__n|0)==(1); //@line 1937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($0) {
  $$0 = 2;
 } else {
  $1 = (($__n) + -1)|0; //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $2 = $1 & $__n; //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $3 = ($2|0)==(0); //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($3) {
   $$0 = $__n;
  } else {
   $4 = (__ZNSt3__112__next_primeEj($__n)|0); //@line 1940 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $$0 = $4;
  }
 }
 $5 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = HEAP32[$5>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $7 = ($$0>>>0)>($6>>>0); //@line 1942 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($7) {
  __ZNSt3__112__hash_tableINS_17__hash_value_typeIiN4love9StrongRefINS2_5image9ImageDataEEEEENS_22__unordered_map_hasherIiS7_NS_4hashIiEELb1EEENS_21__unordered_map_equalIiS7_NS_8equal_toIiEELb1EEENS_9allocatorIS7_EEE8__rehashEj($this,$$0); //@line 1943 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $8 = ($$0>>>0)<($6>>>0); //@line 1944 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($8)) {
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $9 = ($6>>>0)>(2); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($9) {
  $10 = (($6) + -1)|0; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $11 = $10 & $6; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $12 = ($11|0)==(0); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $30 = $12;
 } else {
  $30 = 0;
 }
 $13 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = HEAP32[$13>>2]|0; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $15 = (+($14>>>0)); //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $16 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $17 = +HEAPF32[$16>>2]; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $18 = $15 / $17; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $19 = (+Math_ceil((+$18))); //@line 750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/cmath"
 $20 = (~~(($19))>>>0); //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($30) {
  $21 = (($20) + -1)|0; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $22 = (Math_clz32(($21|0))|0); //@line 785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $23 = (32 - ($22))|0; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $24 = 1 << $23; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $24;
 } else {
  $25 = (__ZNSt3__112__next_primeEj($20)|0); //@line 1950 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $25;
 }
 $26 = ($$0>>>0)<($27>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $28 = $26 ? $27 : $$0; //@line 1946 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $29 = ($28>>>0)<($6>>>0); //@line 1952 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($29)) {
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 __ZNSt3__112__hash_tableINS_17__hash_value_typeIiN4love9StrongRefINS2_5image9ImageDataEEEEENS_22__unordered_map_hasherIiS7_NS_4hashIiEELb1EEENS_21__unordered_map_equalIiS7_NS_8equal_toIiEELb1EEENS_9allocatorIS7_EEE8__rehashEj($this,$28); //@line 1953 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeIiN4love9StrongRefINS2_5image9ImageDataEEEEENS_22__unordered_map_hasherIiS7_NS_4hashIiEELb1EEENS_21__unordered_map_equalIiS7_NS_8equal_toIiEELb1EEENS_9allocatorIS7_EEE8__rehashEj($this,$__nbc) {
 $this = $this|0;
 $__nbc = $__nbc|0;
 var $$lcssa = 0, $$lcssa133 = 0, $$lcssa143 = 0, $$lcssa6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $8 = 0, $9 = 0;
 var $__cp$0 = 0, $__cp$0$in1019 = 0, $__cp$0$in13 = 0, $__cp$0$in13$us = 0, $__cp$0$lcssa5 = 0, $__cp$0$us$us = 0, $__cp$08 = 0, $__cp$08$us = 0, $__cp$0811 = 0, $__cp$081116 = 0, $__cp$081120 = 0, $__cp$0814 = 0, $__cp$0814$us = 0, $__cp$09 = 0, $__cp$09$lcssa = 0, $__cp$09$us$us = 0, $__cp$09$us$us$lcssa = 0, $__i$021 = 0, $__np$0 = 0, $__np$0$lcssa = 0;
 var $__np$0$us$us = 0, $__np$0$us$us$lcssa = 0, $__phash$0$ph$ph18 = 0, $__pp$0$ph$lcssa = 0, $__pp$0$ph$ph17 = 0, $__pp$0$ph12 = 0, $__pp$0$ph12$us = 0, $exitcond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = ($__nbc|0)==(0); //@line 1965 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($1) {
  $74 = HEAP32[$this>>2]|0; //@line 2822 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$this>>2] = 0; //@line 2823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $75 = ($74|0)==(0|0); //@line 2824 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  if (!($75)) {
   __ZdlPv($74); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  HEAP32[$0>>2] = 0; //@line 1967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $2 = $__nbc << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = (__Znwj($2)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $4 = HEAP32[$this>>2]|0; //@line 2822 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $3; //@line 2823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ($4|0)==(0|0); //@line 2824 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($5)) {
  __ZdlPv($4); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 HEAP32[$0>>2] = $__nbc; //@line 1967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__i$021 = 0;
 while(1) {
  $9 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $10 = (($9) + ($__i$021<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$10>>2] = 0; //@line 1971 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $11 = (($__i$021) + 1)|0; //@line 1970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $exitcond = ($11|0)==($__nbc|0); //@line 1970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($exitcond) {
   break;
  } else {
   $__i$021 = $11;
  }
 }
 $6 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $7 = HEAP32[$6>>2]|0; //@line 1973 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $8 = ($7|0)==(0|0); //@line 1974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($8) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $12 = ((($7)) + 4|0); //@line 1976 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $13 = HEAP32[$12>>2]|0; //@line 1976 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $14 = (($__nbc) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $15 = $14 & $__nbc; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $16 = ($15|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($16) {
  $17 = $13 & $14; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $21 = $17;
 } else {
  $18 = (($13>>>0) % ($__nbc>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $21 = $18;
 }
 $19 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $20 = (($19) + ($21<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$20>>2] = $6; //@line 1977 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__cp$081116 = HEAP32[$7>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $22 = ($__cp$081116|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($22) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 } else {
  $__cp$0$in1019 = $7;$__cp$081120 = $__cp$081116;$__phash$0$ph$ph18 = $21;$__pp$0$ph$ph17 = $7;
 }
 L23: while(1) {
  L25: do {
   if ($16) {
    $__cp$0$in13$us = $__cp$0$in1019;$__cp$0814$us = $__cp$081120;$__pp$0$ph12$us = $__pp$0$ph$ph17;
    while(1) {
     $__cp$09$us$us = $__cp$0814$us;
     while(1) {
      $34 = ((($__cp$09$us$us)) + 4|0); //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $35 = HEAP32[$34>>2]|0; //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $36 = $35 & $14; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $37 = ($36|0)==($__phash$0$ph$ph18|0); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($37) {
       $__cp$09$us$us$lcssa = $__cp$09$us$us;
       break;
      }
      $38 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $39 = (($38) + ($36<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $40 = HEAP32[$39>>2]|0; //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $41 = ($40|0)==(0|0); //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($41) {
       $$lcssa = $39;$$lcssa6 = $36;$__cp$0$lcssa5 = $__cp$09$us$us;$__pp$0$ph$lcssa = $__pp$0$ph12$us;
       break L25;
      }
      $51 = ((($__cp$09$us$us)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $__np$0$us$us = $__cp$09$us$us;
      while(1) {
       $52 = HEAP32[$__np$0$us$us>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $53 = ($52|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($53) {
        $$lcssa143 = $52;$__np$0$us$us$lcssa = $__np$0$us$us;
        break;
       }
       $54 = ((($52)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
       $55 = HEAP32[$51>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $56 = HEAP32[$54>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $57 = ($55|0)==($56|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       if ($57) {
        $__np$0$us$us = $52;
       } else {
        $$lcssa143 = $52;$__np$0$us$us$lcssa = $__np$0$us$us;
        break;
       }
      }
      $42 = $$lcssa143; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__pp$0$ph12$us>>2] = $42; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $43 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $44 = (($43) + ($36<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $45 = HEAP32[$44>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $46 = HEAP32[$45>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__np$0$us$us$lcssa>>2] = $46; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $47 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $48 = (($47) + ($36<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $49 = HEAP32[$48>>2]|0; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$49>>2] = $__cp$09$us$us; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $__cp$0$us$us = HEAP32[$__cp$0$in13$us>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $50 = ($__cp$0$us$us|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($50) {
       label = 32;
       break L23;
      } else {
       $__cp$09$us$us = $__cp$0$us$us;
      }
     }
     $__cp$08$us = HEAP32[$__cp$09$us$us$lcssa>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $58 = ($__cp$08$us|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     if ($58) {
      label = 32;
      break L23;
     } else {
      $__cp$0$in13$us = $__cp$09$us$us$lcssa;$__cp$0814$us = $__cp$08$us;$__pp$0$ph12$us = $__cp$09$us$us$lcssa;
     }
    }
   } else {
    $__cp$0$in13 = $__cp$0$in1019;$__cp$0814 = $__cp$081120;$__pp$0$ph12 = $__pp$0$ph$ph17;
    while(1) {
     $__cp$09 = $__cp$0814;
     while(1) {
      $24 = ((($__cp$09)) + 4|0); //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $25 = HEAP32[$24>>2]|0; //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $26 = (($25>>>0) % ($__nbc>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $27 = ($26|0)==($__phash$0$ph$ph18|0); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($27) {
       $__cp$09$lcssa = $__cp$09;
       break;
      }
      $28 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $29 = (($28) + ($26<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $30 = HEAP32[$29>>2]|0; //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $31 = ($30|0)==(0|0); //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($31) {
       $$lcssa = $29;$$lcssa6 = $26;$__cp$0$lcssa5 = $__cp$09;$__pp$0$ph$lcssa = $__pp$0$ph12;
       break L25;
      }
      $32 = ((($__cp$09)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $__np$0 = $__cp$09;
      while(1) {
       $59 = HEAP32[$__np$0>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $60 = ($59|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($60) {
        $$lcssa133 = $59;$__np$0$lcssa = $__np$0;
        break;
       }
       $61 = ((($59)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
       $62 = HEAP32[$32>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $63 = HEAP32[$61>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $64 = ($62|0)==($63|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       if ($64) {
        $__np$0 = $59;
       } else {
        $$lcssa133 = $59;$__np$0$lcssa = $__np$0;
        break;
       }
      }
      $65 = $$lcssa133; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__pp$0$ph12>>2] = $65; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $66 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $67 = (($66) + ($26<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $68 = HEAP32[$67>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $69 = HEAP32[$68>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__np$0$lcssa>>2] = $69; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $70 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $71 = (($70) + ($26<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $72 = HEAP32[$71>>2]|0; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$72>>2] = $__cp$09; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $__cp$0 = HEAP32[$__cp$0$in13>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $73 = ($__cp$0|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($73) {
       label = 32;
       break L23;
      } else {
       $__cp$09 = $__cp$0;
      }
     }
     $__cp$08 = HEAP32[$__cp$09$lcssa>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $23 = ($__cp$08|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     if ($23) {
      label = 32;
      break L23;
     } else {
      $__cp$0$in13 = $__cp$09$lcssa;$__cp$0814 = $__cp$08;$__pp$0$ph12 = $__cp$09$lcssa;
     }
    }
   }
  } while(0);
  HEAP32[$$lcssa>>2] = $__pp$0$ph$lcssa; //@line 1989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $__cp$0811 = HEAP32[$__cp$0$lcssa5>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $33 = ($__cp$0811|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($33) {
   label = 32;
   break;
  } else {
   $__cp$0$in1019 = $__cp$0$lcssa5;$__cp$081120 = $__cp$0811;$__phash$0$ph$ph18 = $$lcssa6;$__pp$0$ph$ph17 = $__cp$0$lcssa5;
  }
 }
 if ((label|0) == 32) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
}
function __ZN4love4font16BMFontRasterizer11parseConfigERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE($this,$configtext) {
 $this = $this|0;
 $configtext = $configtext|0;
 var $$0 = 0, $$0$i = 0, $$0$i11 = 0, $$0$i15 = 0, $$0$i19 = 0, $$0$i23 = 0, $$0$i38 = 0, $$0$i42 = 0, $$0$i46 = 0, $$0$i50 = 0, $$0$i54 = 0, $$0$i58 = 0, $$0$i62 = 0, $$0$i66 = 0, $$0$i70 = 0, $$0$i74 = 0, $$0$i78 = 0, $$0$i8 = 0, $$0$i82 = 0, $$01 = 0;
 var $$01$i = 0, $$01$i$i$i$i = 0, $$01$i$i$i$i85 = 0, $$01$i$i$i$i89 = 0, $$01$i5 = 0, $$02$i = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$lcssa = 0, $$lcssa111 = 0, $$lcssa117 = 0, $$lcssa476 = 0, $$pre$lcssa = 0, $$pre228$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0;
 var $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0;
 var $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0;
 var $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0;
 var $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0;
 var $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0;
 var $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0;
 var $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0;
 var $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0;
 var $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0;
 var $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0;
 var $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0;
 var $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0;
 var $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0;
 var $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0;
 var $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0;
 var $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0;
 var $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0;
 var $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0;
 var $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0;
 var $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0;
 var $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0;
 var $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0;
 var $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0;
 var $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0;
 var $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0;
 var $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0;
 var $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0;
 var $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0;
 var $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0;
 var $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$0$us = 0, $__begin$sroa$0$0156 = 0, $__begin$sroa$0$0157 = 0, $__begin$sroa$0$0157$us = 0, $cline = 0, $eh$lpad$body$i$index21Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $eh$lpad$body$index35Z2D = 0, $eh$lpad$body$indexZ2D = 0;
 var $eh$lpad$body31$index45Z2D = 0, $eh$lpad$body31$indexZ2D = 0, $filename = 0, $id = 0, $key$i = 0, $line = 0, $lpad$phi$i$index = 0, $lpad$phi$i$index26 = 0, $packedids = 0, $pageindex = 0, $ss = 0, $startpos$010$i = 0, $storemerge$i = 0, $storemerge$in$i = 0, $valstart$0$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer12 = 0, $vararg_buffer16 = 0, $vararg_buffer3 = 0;
 var $vararg_buffer5 = 0, $vararg_buffer7 = 0, $vararg_buffer9 = 0, $vararg_ptr15 = 0, $vararg_ptr19 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 576|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer16 = sp + 64|0;
 $vararg_buffer12 = sp + 56|0;
 $vararg_buffer9 = sp + 48|0;
 $vararg_buffer7 = sp + 40|0;
 $vararg_buffer5 = sp + 32|0;
 $vararg_buffer3 = sp + 24|0;
 $vararg_buffer1 = sp + 16|0;
 $vararg_buffer = sp + 8|0;
 $0 = sp + 556|0;
 $1 = sp + 544|0;
 $2 = sp + 532|0;
 $3 = sp + 520|0;
 $4 = sp + 508|0;
 $5 = sp + 496|0;
 $6 = sp + 484|0;
 $7 = sp + 472|0;
 $8 = sp + 460|0;
 $9 = sp + 448|0;
 $10 = sp + 436|0;
 $11 = sp + 424|0;
 $12 = sp + 412|0;
 $13 = sp + 400|0;
 $14 = sp + 388|0;
 $15 = sp + 376|0;
 $16 = sp + 364|0;
 $17 = sp + 352|0;
 $18 = sp + 340|0;
 $key$i = sp + 328|0;
 $19 = sp + 316|0;
 $20 = sp + 312|0;
 $ss = sp + 168|0;
 $line = sp + 152|0;
 $cline = sp + 120|0;
 $pageindex = sp + 112|0;
 $filename = sp + 100|0;
 $21 = sp + 88|0;
 $22 = sp + 76|0;
 $id = sp + 72|0;
 $packedids = sp;
 $23 = ((($ss)) + 64|0); //@line 908 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $24 = ((($ss)) + 8|0); //@line 908 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$24>>2] = (9320); //@line 908 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $25 = ((($ss)) + 12|0); //@line 906 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$ss>>2] = ((9524)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 HEAP32[$23>>2] = ((9544)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $26 = ((($ss)) + 4|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 HEAP32[$26>>2] = 0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $27 = ((($ss)) + 64|0); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 __THREW__ = 0;
 invoke_vii(3362,($27|0),($25|0)); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if ($29) {
  $40 = ___cxa_find_matching_catch_2()|0;
  $41 = tempRet0;
  $$0$i = $41;$$01$i = $40;
  __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($23); //@line 909 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  ___resumeException($$01$i|0);
  // unreachable;
 }
 $30 = ((($ss)) + 136|0); //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$30>>2] = 0; //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 $31 = ((($ss)) + 140|0); //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$31>>2] = -1; //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$ss>>2] = (9300); //@line 908 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$23>>2] = (9340); //@line 908 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$24>>2] = (9320); //@line 908 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 __THREW__ = 0;
 invoke_vi(3363,($25|0)); //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $32 = __THREW__; __THREW__ = 0;
 $33 = $32&1;
 do {
  if ($33) {
   $42 = ___cxa_find_matching_catch_2()|0;
   $43 = tempRet0;
   $eh$lpad$body$i$index21Z2D = $43;$eh$lpad$body$i$indexZ2D = $42;
  } else {
   HEAP32[$25>>2] = (9356); //@line 253 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $34 = ((($ss)) + 44|0); //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $35 = ((($ss)) + 60|0); //@line 252 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   ;HEAP32[$34>>2]=0|0;HEAP32[$34+4>>2]=0|0;HEAP32[$34+8>>2]=0|0;HEAP32[$34+12>>2]=0|0; //@line 251 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   HEAP32[$35>>2] = 24; //@line 252 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   __THREW__ = 0;
   invoke_vii(3364,($25|0),($configtext|0)); //@line 254 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $36 = __THREW__; __THREW__ = 0;
   $37 = $36&1;
   if ($37) {
    $38 = ___cxa_find_matching_catch_2()|0;
    $39 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 255 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($25); //@line 255 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $eh$lpad$body$i$index21Z2D = $39;$eh$lpad$body$i$indexZ2D = $38;
    break;
   }
   ;HEAP32[$line>>2]=0|0;HEAP32[$line+4>>2]=0|0;HEAP32[$line+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $44 = ((($cline)) + 12|0); //@line 67 "modules/font/BMFontRasterizer.cpp"
   $45 = ((($cline)) + 28|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $46 = ((($cline)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $47 = ((($line)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $48 = ((($cline)) + 12|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
   $49 = ((($this)) + 96|0); //@line 171 "modules/font/BMFontRasterizer.cpp"
   $50 = ((($this)) + 100|0); //@line 172 "modules/font/BMFontRasterizer.cpp"
   $51 = ((($cline)) + 20|0); //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $52 = ((($cline)) + 12|0); //@line 2014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $53 = ((($this)) + 104|0); //@line 176 "modules/font/BMFontRasterizer.cpp"
   $54 = ((($this)) + 12|0); //@line 177 "modules/font/BMFontRasterizer.cpp"
   $55 = ((($this)) + 24|0); //@line 185 "modules/font/BMFontRasterizer.cpp"
   $56 = ((($this)) + 36|0); //@line 189 "modules/font/BMFontRasterizer.cpp"
   $57 = ((($filename)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $58 = ((($filename)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $59 = ((($this)) + 28|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $60 = ((($this)) + 32|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $61 = ((($55)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $62 = ((($filename)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $63 = ((($this)) + 56|0); //@line 225 "modules/font/BMFontRasterizer.cpp"
   $64 = ((($this)) + 76|0); //@line 234 "modules/font/BMFontRasterizer.cpp"
   $65 = ((($line)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $66 = ((($line)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $67 = ((($cline)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $68 = ((($cline)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   L11: while(1) {
    $69 = HEAP32[$ss>>2]|0; //@line 1652 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
    $70 = ((($69)) + -12|0); //@line 1652 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
    $71 = HEAP32[$70>>2]|0; //@line 1652 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
    $72 = (($ss) + ($71)|0); //@line 1652 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
    __THREW__ = 0;
    $73 = (invoke_ii(3373,($72|0))|0); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    $74 = __THREW__; __THREW__ = 0;
    $75 = $74&1;
    if ($75) {
     label = 62;
     break;
    }
    HEAP32[$20>>2] = $73; //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    __THREW__ = 0;
    $76 = (invoke_iii(3374,($20|0),(1207576|0))|0); //@line 181 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
    $77 = __THREW__; __THREW__ = 0;
    $78 = $77&1;
    if ($78) {
     label = 13;
     break;
    }
    $79 = HEAP32[$76>>2]|0; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
    $80 = ((($79)) + 28|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
    $81 = HEAP32[$80>>2]|0; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
    __THREW__ = 0;
    $82 = (invoke_iii($81|0,($76|0),10)|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
    $83 = __THREW__; __THREW__ = 0;
    $84 = $83&1;
    if ($84) {
     label = 13;
     break;
    }
    __ZNSt3__16localeD2Ev($20); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    __THREW__ = 0;
    $87 = (invoke_iiii(3516,($ss|0),($line|0),($82|0))|0); //@line 1652 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
    $88 = __THREW__; __THREW__ = 0;
    $89 = $88&1;
    if ($89) {
     label = 62;
     break;
    }
    $90 = HEAP32[$87>>2]|0; //@line 163 "modules/font/BMFontRasterizer.cpp"
    $91 = ((($90)) + -12|0); //@line 163 "modules/font/BMFontRasterizer.cpp"
    $92 = HEAP32[$91>>2]|0; //@line 163 "modules/font/BMFontRasterizer.cpp"
    $93 = (($87) + ($92)|0); //@line 163 "modules/font/BMFontRasterizer.cpp"
    $94 = ((($93)) + 16|0); //@line 545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    $95 = HEAP32[$94>>2]|0; //@line 545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    $96 = $95 & 5; //@line 545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    $97 = ($96|0)==(0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    if (!($97)) {
     label = 227;
     break;
    }
    ;HEAP32[$cline>>2]=0|0;HEAP32[$cline+4>>2]=0|0;HEAP32[$cline+8>>2]=0|0;HEAP32[$cline+12>>2]=0|0;HEAP32[$cline+16>>2]=0|0;HEAP32[$cline+20>>2]=0|0;HEAP32[$cline+24>>2]=0|0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAPF32[$45>>2] = 1.0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $98 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE4findEcj($line,32,0)|0); //@line 70 "modules/font/BMFontRasterizer.cpp"
    __THREW__ = 0;
    invoke_viiiii(3511,($18|0),($line|0),0,($98|0),($line|0)); //@line 3345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $99 = __THREW__; __THREW__ = 0;
    $100 = $99&1;
    if ($100) {
     label = 26;
     break;
    }
    $101 = HEAP8[$cline>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $102 = $101 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $103 = ($102<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if ($103) {
     HEAP8[$46>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP8[$cline>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    } else {
     $104 = HEAP32[$67>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP8[$104>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP32[$68>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    }
    __THREW__ = 0;
    invoke_vii(3368,($cline|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $105 = __THREW__; __THREW__ = 0;
    $106 = $105&1;
    if ($106) {
     label = 21;
     break;
    }
    ;HEAP32[$cline>>2]=HEAP32[$18>>2]|0;HEAP32[$cline+4>>2]=HEAP32[$18+4>>2]|0;HEAP32[$cline+8>>2]=HEAP32[$18+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    ;HEAP32[$18>>2]=0|0;HEAP32[$18+4>>2]=0|0;HEAP32[$18+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($18); //@line 70 "modules/font/BMFontRasterizer.cpp"
    $109 = HEAP8[$line>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $110 = $109 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $111 = ($110<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $112 = HEAP32[$47>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $113 = $109&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $114 = $113 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $115 = $111 ? $114 : $112; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $116 = ($115|0)==(0); //@line 74 "modules/font/BMFontRasterizer.cpp"
    L24: do {
     if (!($116)) {
      $startpos$010$i = 0;
      while(1) {
       $117 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE4findEcj($line,61,$startpos$010$i)|0); //@line 77 "modules/font/BMFontRasterizer.cpp"
       $118 = ($117|0)==(-1); //@line 78 "modules/font/BMFontRasterizer.cpp"
       if ($118) {
        break L24;
       }
       $119 = (($117) + 1)|0; //@line 78 "modules/font/BMFontRasterizer.cpp"
       $120 = HEAP8[$line>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $121 = $120 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $122 = ($121<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $123 = HEAP32[$47>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $124 = $120&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $125 = $124 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $126 = $122 ? $125 : $123; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $127 = ($119>>>0)<($126>>>0); //@line 78 "modules/font/BMFontRasterizer.cpp"
       if (!($127)) {
        break L24;
       }
       $132 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE5rfindEcj($line,32,$117)|0); //@line 82 "modules/font/BMFontRasterizer.cpp"
       $133 = ($132|0)==(-1); //@line 83 "modules/font/BMFontRasterizer.cpp"
       if ($133) {
        break L24;
       }
       $134 = (($132) + 1)|0; //@line 86 "modules/font/BMFontRasterizer.cpp"
       $135 = (($117) - ($134))|0; //@line 88 "modules/font/BMFontRasterizer.cpp"
       __THREW__ = 0;
       invoke_viiiii(3511,($key$i|0),($line|0),($134|0),($135|0),($line|0)); //@line 3345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $136 = __THREW__; __THREW__ = 0;
       $137 = $136&1;
       if ($137) {
        label = 25;
        break L11;
       }
       $138 = (($117) + 2)|0; //@line 91 "modules/font/BMFontRasterizer.cpp"
       $139 = HEAP8[$line>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $140 = $139 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $141 = ($140<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $142 = HEAP32[$65>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $143 = $141 ? $66 : $142; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $144 = (($143) + ($119)|0); //@line 3262 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $145 = HEAP8[$144>>0]|0; //@line 93 "modules/font/BMFontRasterizer.cpp"
       $146 = ($145<<24>>24)==(34); //@line 93 "modules/font/BMFontRasterizer.cpp"
       if ($146) {
        $147 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE4findEcj($line,34,$138)|0); //@line 97 "modules/font/BMFontRasterizer.cpp"
        $storemerge$in$i = $147;$valstart$0$i = $138;
       } else {
        $150 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE4findEcj($line,32,$138)|0); //@line 102 "modules/font/BMFontRasterizer.cpp"
        $storemerge$in$i = $150;$valstart$0$i = $119;
       }
       $storemerge$i = (($storemerge$in$i) + -1)|0; //@line 102 "modules/font/BMFontRasterizer.cpp"
       $151 = HEAP8[$line>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $152 = $151 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $153 = ($152<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $154 = HEAP32[$47>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $155 = $151&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $156 = $155 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $157 = $153 ? $156 : $154; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $158 = (($157) + -1)|0; //@line 105 "modules/font/BMFontRasterizer.cpp"
       $159 = ($158>>>0)<($storemerge$i>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $160 = $159 ? $158 : $storemerge$i; //@line 105 "modules/font/BMFontRasterizer.cpp"
       __THREW__ = 0;
       $161 = (invoke_iii(3988,($44|0),($key$i|0))|0); //@line 107 "modules/font/BMFontRasterizer.cpp"
       $162 = __THREW__; __THREW__ = 0;
       $163 = $162&1;
       if ($163) {
        label = 32;
        break L11;
       }
       $164 = (1 - ($valstart$0$i))|0; //@line 107 "modules/font/BMFontRasterizer.cpp"
       $165 = (($164) + ($160))|0; //@line 107 "modules/font/BMFontRasterizer.cpp"
       __THREW__ = 0;
       invoke_viiiii(3511,($19|0),($line|0),($valstart$0$i|0),($165|0),($line|0)); //@line 3345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $166 = __THREW__; __THREW__ = 0;
       $167 = $166&1;
       if ($167) {
        label = 32;
        break L11;
       }
       $168 = HEAP8[$161>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $169 = $168 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $170 = ($169<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($170) {
        $174 = ((($161)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP8[$174>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP8[$161>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       } else {
        $171 = ((($161)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $172 = HEAP32[$171>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP8[$172>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $173 = ((($161)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP32[$173>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       }
       __THREW__ = 0;
       invoke_vii(3368,($161|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $175 = __THREW__; __THREW__ = 0;
       $176 = $175&1;
       if ($176) {
        label = 40;
        break L11;
       }
       ;HEAP32[$161>>2]=HEAP32[$19>>2]|0;HEAP32[$161+4>>2]=HEAP32[$19+4>>2]|0;HEAP32[$161+8>>2]=HEAP32[$19+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       ;HEAP32[$19>>2]=0|0;HEAP32[$19+4>>2]=0|0;HEAP32[$19+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($19); //@line 107 "modules/font/BMFontRasterizer.cpp"
       $179 = (($160) + 1)|0; //@line 109 "modules/font/BMFontRasterizer.cpp"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($key$i); //@line 110 "modules/font/BMFontRasterizer.cpp"
       $180 = HEAP8[$line>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $181 = $180 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $182 = ($181<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $183 = HEAP32[$47>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $184 = $180&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $185 = $184 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $186 = $182 ? $185 : $183; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $187 = ($179>>>0)<($186>>>0); //@line 74 "modules/font/BMFontRasterizer.cpp"
       if ($187) {
        $startpos$010$i = $179;
       } else {
        break;
       }
      }
     }
    } while(0);
    $196 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7compareEPKc($cline,367355)|0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $197 = ($196|0)==(0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    do {
     if ($197) {
      __THREW__ = 0;
      invoke_viii(3460,($17|0),(367360|0),4); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $198 = __THREW__; __THREW__ = 0;
      $199 = $198&1;
      if ($199) {
       label = 65;
       break L11;
      }
      $200 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$17)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($17); //@line 115 "modules/font/BMFontRasterizer.cpp"
      $201 = ($200|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($201) {
       $$0$i8 = 0;
      } else {
       $202 = ((($200)) + 20|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $203 = HEAP8[$202>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $204 = $203 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $205 = ($204<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($205) {
        $208 = ((($202)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $209 = $208;
       } else {
        $206 = ((($200)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $207 = HEAP32[$206>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $209 = $207;
       }
       $210 = (_strtol($209,0,10)|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $$0$i8 = $210;
      }
      HEAP32[$49>>2] = $$0$i8; //@line 171 "modules/font/BMFontRasterizer.cpp"
      __THREW__ = 0;
      invoke_viii(3460,($16|0),(367365|0),7); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $211 = __THREW__; __THREW__ = 0;
      $212 = $211&1;
      if ($212) {
       label = 65;
       break L11;
      }
      $213 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$16)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($16); //@line 115 "modules/font/BMFontRasterizer.cpp"
      $214 = ($213|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($214) {
       $$0$i11 = 0;
      } else {
       $215 = ((($213)) + 20|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $216 = HEAP8[$215>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $217 = $216 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $218 = ($217<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($218) {
        $221 = ((($215)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $222 = $221;
       } else {
        $219 = ((($213)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $220 = HEAP32[$219>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $222 = $220;
       }
       $223 = (_strtol($222,0,10)|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $$0$i11 = $223;
      }
      $224 = ($$0$i11|0)>(0); //@line 172 "modules/font/BMFontRasterizer.cpp"
      $225 = $224&1; //@line 172 "modules/font/BMFontRasterizer.cpp"
      HEAP8[$50>>0] = $225; //@line 172 "modules/font/BMFontRasterizer.cpp"
     } else {
      $234 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7compareEPKc($cline,367373)|0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $235 = ($234|0)==(0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($235) {
       __THREW__ = 0;
       invoke_viii(3460,($15|0),(367380|0),10); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $236 = __THREW__; __THREW__ = 0;
       $237 = $236&1;
       if ($237) {
        label = 65;
        break L11;
       }
       $238 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$15)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($15); //@line 115 "modules/font/BMFontRasterizer.cpp"
       $239 = ($238|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($239) {
        $$0$i15 = 0;
       } else {
        $240 = ((($238)) + 20|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
        $241 = HEAP8[$240>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $242 = $241 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $243 = ($242<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if ($243) {
         $246 = ((($240)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $247 = $246;
        } else {
         $244 = ((($238)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $245 = HEAP32[$244>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $247 = $245;
        }
        $248 = (_strtol($247,0,10)|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
        $$0$i15 = $248;
       }
       HEAP32[$53>>2] = $$0$i15; //@line 176 "modules/font/BMFontRasterizer.cpp"
       __THREW__ = 0;
       invoke_viii(3460,($14|0),(367391|0),4); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $249 = __THREW__; __THREW__ = 0;
       $250 = $249&1;
       if ($250) {
        label = 65;
        break L11;
       }
       $251 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$14)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($14); //@line 115 "modules/font/BMFontRasterizer.cpp"
       $252 = ($251|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($252) {
        $$0$i19 = 0;
       } else {
        $253 = ((($251)) + 20|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
        $254 = HEAP8[$253>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $255 = $254 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $256 = ($255<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if ($256) {
         $259 = ((($253)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $260 = $259;
        } else {
         $257 = ((($251)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $258 = HEAP32[$257>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $260 = $258;
        }
        $261 = (_strtol($260,0,10)|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
        $$0$i19 = $261;
       }
       HEAP32[$54>>2] = $$0$i19; //@line 177 "modules/font/BMFontRasterizer.cpp"
       break;
      }
      $262 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7compareEPKc($cline,367396)|0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $263 = ($262|0)==(0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($263) {
       __THREW__ = 0;
       invoke_viii(3460,($13|0),(597634|0),2); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $264 = __THREW__; __THREW__ = 0;
       $265 = $264&1;
       if ($265) {
        label = 65;
        break L11;
       }
       $266 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$13)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13); //@line 115 "modules/font/BMFontRasterizer.cpp"
       $267 = ($266|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($267) {
        $$0$i23 = 0;
       } else {
        $268 = ((($266)) + 20|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
        $269 = HEAP8[$268>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $270 = $269 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $271 = ($270<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if ($271) {
         $274 = ((($268)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $275 = $274;
        } else {
         $272 = ((($266)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $273 = HEAP32[$272>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $275 = $273;
        }
        $276 = (_strtol($275,0,10)|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
        $$0$i23 = $276;
       }
       HEAP32[$pageindex>>2] = $$0$i23; //@line 181 "modules/font/BMFontRasterizer.cpp"
       __THREW__ = 0;
       invoke_viii(3460,($12|0),(521802|0),4); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $277 = __THREW__; __THREW__ = 0;
       $278 = $277&1;
       if ($278) {
        label = 65;
        break L11;
       }
       $279 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$12)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12); //@line 124 "modules/font/BMFontRasterizer.cpp"
       $280 = ($279|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($280) {
        __THREW__ = 0;
        invoke_viii(3460,($filename|0),(1282466|0),0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $281 = __THREW__; __THREW__ = 0;
        $282 = $281&1;
        if ($282) {
         label = 65;
         break L11;
        }
       } else {
        $283 = ((($279)) + 20|0); //@line 128 "modules/font/BMFontRasterizer.cpp"
        __THREW__ = 0;
        invoke_vii(3351,($filename|0),($283|0)); //@line 128 "modules/font/BMFontRasterizer.cpp"
        $284 = __THREW__; __THREW__ = 0;
        $285 = $284&1;
        if ($285) {
         label = 65;
         break L11;
        }
       }
       $286 = HEAP8[$55>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $287 = $286 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $288 = ($287<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($288) {
        $290 = $286&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $291 = $290 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $293 = $291;
       } else {
        $289 = HEAP32[$59>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $293 = $289;
       }
       $292 = ($293|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if (!($292)) {
        ;HEAP32[$22>>2]=0|0;HEAP32[$22+4>>2]=0|0;HEAP32[$22+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $294 = HEAP32[$59>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $295 = $286&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $296 = $295 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $297 = $288 ? $296 : $294; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $298 = HEAP32[$60>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $299 = $288 ? $61 : $298; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $300 = (($297) + 1)|0; //@line 3980 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        __THREW__ = 0;
        invoke_viiii(3514,($22|0),($299|0),($297|0),($300|0)); //@line 3980 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $301 = __THREW__; __THREW__ = 0;
        $302 = $301&1;
        if ($302) {
         label = 97;
         break L11;
        }
        __THREW__ = 0;
        (invoke_iiii(3462,($22|0),(523303|0),1)|0); //@line 3981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $303 = __THREW__; __THREW__ = 0;
        $304 = $303&1;
        if ($304) {
         label = 97;
         break L11;
        }
        $307 = HEAP8[$filename>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $308 = $307 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $309 = ($308<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $310 = HEAP32[$57>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $311 = $309 ? $58 : $310; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $312 = HEAP32[$62>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $313 = $307&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $314 = $313 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $315 = $309 ? $314 : $312; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        __THREW__ = 0;
        $316 = (invoke_iiii(3462,($22|0),($311|0),($315|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $317 = __THREW__; __THREW__ = 0;
        $318 = $317&1;
        if ($318) {
         label = 108;
         break L11;
        }
        ;HEAP32[$21>>2]=HEAP32[$316>>2]|0;HEAP32[$21+4>>2]=HEAP32[$316+4>>2]|0;HEAP32[$21+8>>2]=HEAP32[$316+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        ;HEAP32[$316>>2]=0|0;HEAP32[$316+4>>2]=0|0;HEAP32[$316+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $319 = HEAP8[$filename>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $320 = $319 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $321 = ($320<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if ($321) {
         HEAP8[$58>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         HEAP8[$filename>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        } else {
         $322 = HEAP32[$57>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         HEAP8[$322>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         HEAP32[$62>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        }
        __THREW__ = 0;
        invoke_vii(3368,($filename|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $323 = __THREW__; __THREW__ = 0;
        $324 = $323&1;
        if ($324) {
         label = 103;
         break L11;
        }
        ;HEAP32[$filename>>2]=HEAP32[$21>>2]|0;HEAP32[$filename+4>>2]=HEAP32[$21+4>>2]|0;HEAP32[$filename+8>>2]=HEAP32[$21+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        ;HEAP32[$21>>2]=0|0;HEAP32[$21+4>>2]=0|0;HEAP32[$21+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($21); //@line 186 "modules/font/BMFontRasterizer.cpp"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($22); //@line 186 "modules/font/BMFontRasterizer.cpp"
       }
       __THREW__ = 0;
       $333 = (invoke_iii(3985,($56|0),($pageindex|0))|0); //@line 189 "modules/font/BMFontRasterizer.cpp"
       $334 = __THREW__; __THREW__ = 0;
       $335 = $334&1;
       if ($335) {
        label = 105;
        break L11;
       }
       $336 = HEAP32[$333>>2]|0; //@line 139 "./common/Object.h"
       $337 = ($336|0)==(0|0); //@line 189 "modules/font/BMFontRasterizer.cpp"
       do {
        if ($337) {
         $338 = HEAP32[(625616)>>2]|0; //@line 98 "./common/Module.h"
         $339 = HEAP32[(625628)>>2]|0; //@line 98 "./common/Module.h"
         $340 = ($338|0)==(0|0); //@line 196 "modules/font/BMFontRasterizer.cpp"
         if ($340) {
          label = 112;
          break L11;
         }
         $347 = ($339|0)==(0|0); //@line 198 "modules/font/BMFontRasterizer.cpp"
         if ($347) {
          label = 116;
          break L11;
         }
         $354 = HEAP32[$338>>2]|0; //@line 202 "modules/font/BMFontRasterizer.cpp"
         $355 = ((($354)) + 132|0); //@line 202 "modules/font/BMFontRasterizer.cpp"
         $356 = HEAP32[$355>>2]|0; //@line 202 "modules/font/BMFontRasterizer.cpp"
         $357 = HEAP8[$filename>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $358 = $357 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $359 = ($358<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $360 = HEAP32[$57>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $361 = $359 ? $58 : $360; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         __THREW__ = 0;
         $362 = (invoke_iiiii($356|0,($338|0),($361|0),-1,-1)|0); //@line 202 "modules/font/BMFontRasterizer.cpp"
         $363 = __THREW__; __THREW__ = 0;
         $364 = $363&1;
         if ($364) {
          label = 105;
          break L11;
         }
         $365 = ($362|0)==(0|0); //@line 95 "./common/Object.h"
         if (!($365)) {
          $366 = HEAP32[$362>>2]|0; //@line 95 "./common/Object.h"
          $367 = ((($366)) + 8|0); //@line 95 "./common/Object.h"
          $368 = HEAP32[$367>>2]|0; //@line 95 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($368|0,($362|0)); //@line 95 "./common/Object.h"
          $369 = __THREW__; __THREW__ = 0;
          $370 = $369&1;
          if ($370) {
           label = 105;
           break L11;
          }
         }
         $371 = HEAP32[$362>>2]|0; //@line 203 "modules/font/BMFontRasterizer.cpp"
         $372 = ((($371)) + 12|0); //@line 203 "modules/font/BMFontRasterizer.cpp"
         $373 = HEAP32[$372>>2]|0; //@line 203 "modules/font/BMFontRasterizer.cpp"
         __THREW__ = 0;
         invoke_vi($373|0,($362|0)); //@line 203 "modules/font/BMFontRasterizer.cpp"
         $374 = __THREW__; __THREW__ = 0;
         $375 = $374&1;
         if ($375) {
          $$lcssa476 = $365;$$pre$lcssa = $362;$$pre228$lcssa = $362;
          label = 134;
          break L11;
         }
         __THREW__ = 0;
         $376 = (invoke_iii(3985,($56|0),($pageindex|0))|0); //@line 205 "modules/font/BMFontRasterizer.cpp"
         $377 = __THREW__; __THREW__ = 0;
         $378 = $377&1;
         if ($378) {
          $$lcssa476 = $365;$$pre$lcssa = $362;$$pre228$lcssa = $362;
          label = 134;
          break L11;
         }
         $379 = HEAP32[$339>>2]|0; //@line 205 "modules/font/BMFontRasterizer.cpp"
         $380 = ((($379)) + 24|0); //@line 205 "modules/font/BMFontRasterizer.cpp"
         $381 = HEAP32[$380>>2]|0; //@line 205 "modules/font/BMFontRasterizer.cpp"
         __THREW__ = 0;
         $382 = (invoke_iii($381|0,($339|0),($362|0))|0); //@line 205 "modules/font/BMFontRasterizer.cpp"
         $383 = __THREW__; __THREW__ = 0;
         $384 = $383&1;
         if ($384) {
          $$lcssa476 = $365;$$pre$lcssa = $362;$$pre228$lcssa = $362;
          label = 134;
          break L11;
         }
         $385 = ($382|0)==(0|0); //@line 132 "./common/Object.h"
         if (!($385)) {
          $386 = HEAP32[$382>>2]|0; //@line 132 "./common/Object.h"
          $387 = ((($386)) + 8|0); //@line 132 "./common/Object.h"
          $388 = HEAP32[$387>>2]|0; //@line 132 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($388|0,($382|0)); //@line 132 "./common/Object.h"
          $389 = __THREW__; __THREW__ = 0;
          $390 = $389&1;
          if ($390) {
           $$lcssa476 = $365;$$pre$lcssa = $362;$$pre228$lcssa = $362;
           label = 134;
           break L11;
          }
         }
         $391 = HEAP32[$376>>2]|0; //@line 133 "./common/Object.h"
         $392 = ($391|0)==(0|0); //@line 133 "./common/Object.h"
         if (!($392)) {
          $393 = HEAP32[$391>>2]|0; //@line 133 "./common/Object.h"
          $394 = ((($393)) + 12|0); //@line 133 "./common/Object.h"
          $395 = HEAP32[$394>>2]|0; //@line 133 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($395|0,($391|0)); //@line 133 "./common/Object.h"
          $396 = __THREW__; __THREW__ = 0;
          $397 = $396&1;
          if ($397) {
           $$lcssa476 = $365;$$pre$lcssa = $362;$$pre228$lcssa = $362;
           label = 134;
           break L11;
          }
         }
         HEAP32[$376>>2] = $382; //@line 134 "./common/Object.h"
         __THREW__ = 0;
         $398 = (invoke_iii(3985,($56|0),($pageindex|0))|0); //@line 206 "modules/font/BMFontRasterizer.cpp"
         $399 = __THREW__; __THREW__ = 0;
         $400 = $399&1;
         if ($400) {
          $$lcssa476 = $365;$$pre$lcssa = $362;$$pre228$lcssa = $362;
          label = 134;
          break L11;
         }
         $401 = HEAP32[$398>>2]|0; //@line 117 "./common/Object.h"
         $402 = HEAP32[$401>>2]|0; //@line 206 "modules/font/BMFontRasterizer.cpp"
         $403 = ((($402)) + 12|0); //@line 206 "modules/font/BMFontRasterizer.cpp"
         $404 = HEAP32[$403>>2]|0; //@line 206 "modules/font/BMFontRasterizer.cpp"
         __THREW__ = 0;
         invoke_vi($404|0,($401|0)); //@line 206 "modules/font/BMFontRasterizer.cpp"
         $405 = __THREW__; __THREW__ = 0;
         $406 = $405&1;
         if ($406) {
          $$lcssa476 = $365;$$pre$lcssa = $362;$$pre228$lcssa = $362;
          label = 134;
          break L11;
         }
         if ($365) {
          break;
         }
         $407 = HEAP32[$362>>2]|0; //@line 106 "./common/Object.h"
         $408 = ((($407)) + 12|0); //@line 106 "./common/Object.h"
         $409 = HEAP32[$408>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($409|0,($362|0)); //@line 106 "./common/Object.h"
         $410 = __THREW__; __THREW__ = 0;
         $411 = $410&1;
         if ($411) {
          label = 133;
          break L11;
         }
        }
       } while(0);
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($filename); //@line 208 "modules/font/BMFontRasterizer.cpp"
       break;
      }
      $423 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7compareEPKc($cline,522523)|0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $424 = ($423|0)==(0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if (!($424)) {
       $553 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7compareEPKc($cline,593878)|0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $554 = ($553|0)==(0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if (!($554)) {
        break;
       }
       __THREW__ = 0;
       invoke_viii(3460,($2|0),(367481|0),5); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $555 = __THREW__; __THREW__ = 0;
       $556 = $555&1;
       if ($556) {
        label = 65;
        break L11;
       }
       $557 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$2)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 115 "modules/font/BMFontRasterizer.cpp"
       $558 = ($557|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($558) {
        $$0$i74 = 0;
       } else {
        $559 = ((($557)) + 20|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
        $560 = HEAP8[$559>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $561 = $560 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $562 = ($561<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if ($562) {
         $565 = ((($559)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $566 = $565;
        } else {
         $563 = ((($557)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $564 = HEAP32[$563>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $566 = $564;
        }
        $567 = (_strtol($566,0,10)|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
        $$0$i74 = $567;
       }
       __THREW__ = 0;
       invoke_viii(3460,($1|0),(367487|0),6); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $568 = __THREW__; __THREW__ = 0;
       $569 = $568&1;
       if ($569) {
        label = 65;
        break L11;
       }
       $570 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$1)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 115 "modules/font/BMFontRasterizer.cpp"
       $571 = ($570|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($571) {
        $$0$i78 = 0;
       } else {
        $572 = ((($570)) + 20|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
        $573 = HEAP8[$572>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $574 = $573 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $575 = ($574<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if ($575) {
         $578 = ((($572)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $579 = $578;
        } else {
         $576 = ((($570)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $577 = HEAP32[$576>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $579 = $577;
        }
        $580 = (_strtol($579,0,10)|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
        $$0$i78 = $580;
       }
       $581 = $packedids; //@line 232 "modules/font/BMFontRasterizer.cpp"
       $582 = $581; //@line 232 "modules/font/BMFontRasterizer.cpp"
       HEAP32[$582>>2] = $$0$i78; //@line 232 "modules/font/BMFontRasterizer.cpp"
       $583 = (($581) + 4)|0; //@line 232 "modules/font/BMFontRasterizer.cpp"
       $584 = $583; //@line 232 "modules/font/BMFontRasterizer.cpp"
       HEAP32[$584>>2] = $$0$i74; //@line 232 "modules/font/BMFontRasterizer.cpp"
       __THREW__ = 0;
       invoke_viii(3460,($0|0),(367494|0),6); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $585 = __THREW__; __THREW__ = 0;
       $586 = $585&1;
       if ($586) {
        label = 65;
        break L11;
       }
       $587 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$0)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 115 "modules/font/BMFontRasterizer.cpp"
       $588 = ($587|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($588) {
        $$0$i82 = 0;
       } else {
        $589 = ((($587)) + 20|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
        $590 = HEAP8[$589>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $591 = $590 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $592 = ($591<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if ($592) {
         $595 = ((($589)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $596 = $595;
        } else {
         $593 = ((($587)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $594 = HEAP32[$593>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $596 = $594;
        }
        $597 = (_strtol($596,0,10)|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
        $$0$i82 = $597;
       }
       __THREW__ = 0;
       $598 = (invoke_iii(3990,($64|0),($packedids|0))|0); //@line 234 "modules/font/BMFontRasterizer.cpp"
       $599 = __THREW__; __THREW__ = 0;
       $600 = $599&1;
       if ($600) {
        label = 65;
        break L11;
       }
       HEAP32[$598>>2] = $$0$i82; //@line 234 "modules/font/BMFontRasterizer.cpp"
       break;
      }
      __THREW__ = 0;
      invoke_viii(3460,($11|0),(597634|0),2); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $425 = __THREW__; __THREW__ = 0;
      $426 = $425&1;
      if ($426) {
       label = 65;
       break L11;
      }
      $427 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$11)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($11); //@line 115 "modules/font/BMFontRasterizer.cpp"
      $428 = ($427|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($428) {
       $$0$i38 = 0;
      } else {
       $429 = ((($427)) + 20|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $430 = HEAP8[$429>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $431 = $430 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $432 = ($431<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($432) {
        $435 = ((($429)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $436 = $435;
       } else {
        $433 = ((($427)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $434 = HEAP32[$433>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $436 = $434;
       }
       $437 = (_strtol($436,0,10)|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $$0$i38 = $437;
      }
      HEAP32[$id>>2] = $$0$i38; //@line 213 "modules/font/BMFontRasterizer.cpp"
      __THREW__ = 0;
      invoke_viii(3460,($10|0),(610853|0),1); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $438 = __THREW__; __THREW__ = 0;
      $439 = $438&1;
      if ($439) {
       label = 65;
       break L11;
      }
      $440 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$10)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10); //@line 115 "modules/font/BMFontRasterizer.cpp"
      $441 = ($440|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($441) {
       $$0$i42 = 0;
      } else {
       $442 = ((($440)) + 20|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $443 = HEAP8[$442>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $444 = $443 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $445 = ($444<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($445) {
        $448 = ((($442)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $449 = $448;
       } else {
        $446 = ((($440)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $447 = HEAP32[$446>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $449 = $447;
       }
       $450 = (_strtol($449,0,10)|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $$0$i42 = $450;
      }
      __THREW__ = 0;
      invoke_viii(3460,($9|0),(610855|0),1); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $451 = __THREW__; __THREW__ = 0;
      $452 = $451&1;
      if ($452) {
       label = 65;
       break L11;
      }
      $453 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$9)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9); //@line 115 "modules/font/BMFontRasterizer.cpp"
      $454 = ($453|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($454) {
       $$0$i46 = 0;
      } else {
       $455 = ((($453)) + 20|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $456 = HEAP8[$455>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $457 = $456 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $458 = ($457<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($458) {
        $461 = ((($455)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $462 = $461;
       } else {
        $459 = ((($453)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $460 = HEAP32[$459>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $462 = $460;
       }
       $463 = (_strtol($462,0,10)|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $$0$i46 = $463;
      }
      __THREW__ = 0;
      invoke_viii(3460,($8|0),(367396|0),4); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $464 = __THREW__; __THREW__ = 0;
      $465 = $464&1;
      if ($465) {
       label = 65;
       break L11;
      }
      $466 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$8)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8); //@line 115 "modules/font/BMFontRasterizer.cpp"
      $467 = ($466|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($467) {
       $$0$i50 = 0;
      } else {
       $468 = ((($466)) + 20|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $469 = HEAP8[$468>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $470 = $469 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $471 = ($470<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($471) {
        $474 = ((($468)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $475 = $474;
       } else {
        $472 = ((($466)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $473 = HEAP32[$472>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $475 = $473;
       }
       $476 = (_strtol($475,0,10)|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $$0$i50 = $476;
      }
      __THREW__ = 0;
      invoke_viii(3460,($7|0),(609941|0),5); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $477 = __THREW__; __THREW__ = 0;
      $478 = $477&1;
      if ($478) {
       label = 65;
       break L11;
      }
      $479 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$7)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7); //@line 115 "modules/font/BMFontRasterizer.cpp"
      $480 = ($479|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($480) {
       $$0$i54 = 0;
      } else {
       $481 = ((($479)) + 20|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $482 = HEAP8[$481>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $483 = $482 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $484 = ($483<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($484) {
        $487 = ((($481)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $488 = $487;
       } else {
        $485 = ((($479)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $486 = HEAP32[$485>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $488 = $486;
       }
       $489 = (_strtol($488,0,10)|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $$0$i54 = $489;
      }
      __THREW__ = 0;
      invoke_viii(3460,($6|0),(609947|0),6); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $490 = __THREW__; __THREW__ = 0;
      $491 = $490&1;
      if ($491) {
       label = 65;
       break L11;
      }
      $492 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$6)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 115 "modules/font/BMFontRasterizer.cpp"
      $493 = ($492|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($493) {
       $$0$i58 = 0;
      } else {
       $494 = ((($492)) + 20|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $495 = HEAP8[$494>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $496 = $495 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $497 = ($496<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($497) {
        $500 = ((($494)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $501 = $500;
       } else {
        $498 = ((($492)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $499 = HEAP32[$498>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $501 = $499;
       }
       $502 = (_strtol($501,0,10)|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $$0$i58 = $502;
      }
      __THREW__ = 0;
      invoke_viii(3460,($5|0),(367456|0),7); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $503 = __THREW__; __THREW__ = 0;
      $504 = $503&1;
      if ($504) {
       label = 65;
       break L11;
      }
      $505 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$5)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5); //@line 115 "modules/font/BMFontRasterizer.cpp"
      $506 = ($505|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($506) {
       $$0$i62 = 0;
      } else {
       $507 = ((($505)) + 20|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $508 = HEAP8[$507>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $509 = $508 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $510 = ($509<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($510) {
        $513 = ((($507)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $514 = $513;
       } else {
        $511 = ((($505)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $512 = HEAP32[$511>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $514 = $512;
       }
       $515 = (_strtol($514,0,10)|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $$0$i62 = $515;
      }
      __THREW__ = 0;
      invoke_viii(3460,($4|0),(367464|0),7); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $516 = __THREW__; __THREW__ = 0;
      $517 = $516&1;
      if ($517) {
       label = 65;
       break L11;
      }
      $518 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$4)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 115 "modules/font/BMFontRasterizer.cpp"
      $519 = ($518|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($519) {
       $$0$i66 = 0;
      } else {
       $520 = ((($518)) + 20|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $521 = HEAP8[$520>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $522 = $521 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $523 = ($522<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($523) {
        $526 = ((($520)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $527 = $526;
       } else {
        $524 = ((($518)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $525 = HEAP32[$524>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $527 = $525;
       }
       $528 = (_strtol($527,0,10)|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $$0$i66 = $528;
      }
      $529 = (0 - ($$0$i66))|0; //@line 222 "modules/font/BMFontRasterizer.cpp"
      __THREW__ = 0;
      invoke_viii(3460,($3|0),(367472|0),8); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $530 = __THREW__; __THREW__ = 0;
      $531 = $530&1;
      if ($531) {
       label = 65;
       break L11;
      }
      $532 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($48,$3)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 115 "modules/font/BMFontRasterizer.cpp"
      $533 = ($532|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($533) {
       $$0$i70 = 0;
      } else {
       $534 = ((($532)) + 20|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $535 = HEAP8[$534>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $536 = $535 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $537 = ($536<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($537) {
        $540 = ((($534)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $541 = $540;
       } else {
        $538 = ((($532)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $539 = HEAP32[$538>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $541 = $539;
       }
       $542 = (_strtol($541,0,10)|0); //@line 119 "modules/font/BMFontRasterizer.cpp"
       $$0$i70 = $542;
      }
      __THREW__ = 0;
      $543 = (invoke_iii(3989,($63|0),($id|0))|0); //@line 225 "modules/font/BMFontRasterizer.cpp"
      $544 = __THREW__; __THREW__ = 0;
      $545 = $544&1;
      if ($545) {
       label = 65;
       break L11;
      }
      HEAP32[$543>>2] = $$0$i42; //@line 225 "modules/font/BMFontRasterizer.cpp"
      $546 = ((($543)) + 4|0); //@line 225 "modules/font/BMFontRasterizer.cpp"
      HEAP32[$546>>2] = $$0$i46; //@line 225 "modules/font/BMFontRasterizer.cpp"
      $547 = ((($543)) + 8|0); //@line 225 "modules/font/BMFontRasterizer.cpp"
      HEAP32[$547>>2] = $$0$i50; //@line 225 "modules/font/BMFontRasterizer.cpp"
      $548 = ((($543)) + 12|0); //@line 225 "modules/font/BMFontRasterizer.cpp"
      HEAP32[$548>>2] = $$0$i58; //@line 225 "modules/font/BMFontRasterizer.cpp"
      $549 = ((($543)) + 16|0); //@line 225 "modules/font/BMFontRasterizer.cpp"
      HEAP32[$549>>2] = $$0$i54; //@line 225 "modules/font/BMFontRasterizer.cpp"
      $550 = ((($543)) + 20|0); //@line 225 "modules/font/BMFontRasterizer.cpp"
      HEAP32[$550>>2] = $$0$i70; //@line 225 "modules/font/BMFontRasterizer.cpp"
      $551 = ((($543)) + 24|0); //@line 225 "modules/font/BMFontRasterizer.cpp"
      HEAP32[$551>>2] = $$0$i62; //@line 225 "modules/font/BMFontRasterizer.cpp"
      $552 = ((($543)) + 28|0); //@line 225 "modules/font/BMFontRasterizer.cpp"
      HEAP32[$552>>2] = $529; //@line 225 "modules/font/BMFontRasterizer.cpp"
     }
    } while(0);
    $601 = HEAP32[$51>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $602 = ($601|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if (!($602)) {
     $$01$i$i$i$i85 = $601;
     while(1) {
      $603 = HEAP32[$$01$i$i$i$i85>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $604 = ((($$01$i$i$i$i85)) + 20|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($604); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
      $605 = ((($$01$i$i$i$i85)) + 8|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($605); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
      __ZdlPv($$01$i$i$i$i85); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $606 = ($603|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($606) {
       break;
      } else {
       $$01$i$i$i$i85 = $603;
      }
     }
    }
    $607 = HEAP32[$52>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$52>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $608 = ($607|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    if (!($608)) {
     __ZdlPv($607); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    }
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($cline); //@line 48 "modules/font/BMFontRasterizer.cpp"
   }
   L244:    switch (label|0) {
    case 13: {
     $85 = ___cxa_find_matching_catch_2()|0;
     $86 = tempRet0;
     __ZNSt3__16localeD2Ev($20); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     $eh$lpad$body$index35Z2D = $86;$eh$lpad$body$indexZ2D = $85;
     label = 64;
     break;
    }
    case 21: {
     $107 = ___cxa_find_matching_catch_3(0|0)|0;
     $108 = tempRet0;
     ___clang_call_terminate($107); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     break;
    }
    case 25: {
     $128 = ___cxa_find_matching_catch_2()|0;
     $129 = tempRet0;
     $lpad$phi$i$index = $128;$lpad$phi$i$index26 = $129;
     label = 27;
     break;
    }
    case 26: {
     $130 = ___cxa_find_matching_catch_2()|0;
     $131 = tempRet0;
     $lpad$phi$i$index = $130;$lpad$phi$i$index26 = $131;
     label = 27;
     break;
    }
    case 32: {
     $148 = ___cxa_find_matching_catch_2()|0;
     $149 = tempRet0;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($key$i); //@line 110 "modules/font/BMFontRasterizer.cpp"
     $$01$i5 = $148;$$02$i = $149;
     label = 42;
     break;
    }
    case 40: {
     $177 = ___cxa_find_matching_catch_3(0|0)|0;
     $178 = tempRet0;
     ___clang_call_terminate($177); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     break;
    }
    case 62: {
     $228 = ___cxa_find_matching_catch_2()|0;
     $229 = tempRet0;
     $eh$lpad$body$index35Z2D = $229;$eh$lpad$body$indexZ2D = $228;
     label = 64;
     break;
    }
    case 65: {
     $232 = ___cxa_find_matching_catch_2()|0;
     $233 = tempRet0;
     $$1 = $232;$$12 = $233;
     break;
    }
    case 97: {
     $305 = ___cxa_find_matching_catch_2()|0;
     $306 = tempRet0;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($22); //@line 3983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $eh$lpad$body31$index45Z2D = $306;$eh$lpad$body31$indexZ2D = $305;
     label = 107;
     break;
    }
    case 103: {
     $325 = ___cxa_find_matching_catch_3(0|0)|0;
     $326 = tempRet0;
     ___clang_call_terminate($325); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     break;
    }
    case 105: {
     $327 = ___cxa_find_matching_catch_2()|0;
     $328 = tempRet0;
     $eh$lpad$body31$index45Z2D = $328;$eh$lpad$body31$indexZ2D = $327;
     label = 107;
     break;
    }
    case 108: {
     $331 = ___cxa_find_matching_catch_2()|0;
     $332 = tempRet0;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($22); //@line 186 "modules/font/BMFontRasterizer.cpp"
     $$0 = $331;$$01 = $332;
     label = 138;
     break;
    }
    case 112: {
     $341 = (___cxa_allocate_exception(16)|0); //@line 197 "modules/font/BMFontRasterizer.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($341|0),(367401|0),($vararg_buffer|0)); //@line 197 "modules/font/BMFontRasterizer.cpp"
     $342 = __THREW__; __THREW__ = 0;
     $343 = $342&1;
     if ($343) {
      $345 = ___cxa_find_matching_catch_2()|0;
      $346 = tempRet0;
      ___cxa_free_exception(($341|0)); //@line 197 "modules/font/BMFontRasterizer.cpp"
      $$0 = $345;$$01 = $346;
      label = 138;
      break;
     } else {
      __THREW__ = 0;
      invoke_viii(3350,($341|0),(96|0),(344|0)); //@line 197 "modules/font/BMFontRasterizer.cpp"
      $344 = __THREW__; __THREW__ = 0;
      label = 106;
      break;
     }
     break;
    }
    case 116: {
     $348 = (___cxa_allocate_exception(16)|0); //@line 199 "modules/font/BMFontRasterizer.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($348|0),(367431|0),($vararg_buffer1|0)); //@line 199 "modules/font/BMFontRasterizer.cpp"
     $349 = __THREW__; __THREW__ = 0;
     $350 = $349&1;
     if ($350) {
      $352 = ___cxa_find_matching_catch_2()|0;
      $353 = tempRet0;
      ___cxa_free_exception(($348|0)); //@line 199 "modules/font/BMFontRasterizer.cpp"
      $$0 = $352;$$01 = $353;
      label = 138;
      break;
     } else {
      __THREW__ = 0;
      invoke_viii(3350,($348|0),(96|0),(344|0)); //@line 199 "modules/font/BMFontRasterizer.cpp"
      $351 = __THREW__; __THREW__ = 0;
      label = 106;
      break;
     }
     break;
    }
    case 133: {
     $412 = ___cxa_find_matching_catch_3(0|0)|0;
     $413 = tempRet0;
     ___clang_call_terminate($412); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
     break;
    }
    case 134: {
     $414 = ___cxa_find_matching_catch_2()|0;
     $415 = tempRet0;
     if ($$lcssa476) {
      $$0 = $414;$$01 = $415;
      label = 138;
     } else {
      $416 = HEAP32[$$pre228$lcssa>>2]|0; //@line 106 "./common/Object.h"
      $417 = ((($416)) + 12|0); //@line 106 "./common/Object.h"
      $418 = HEAP32[$417>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($418|0,($$pre$lcssa|0)); //@line 106 "./common/Object.h"
      $419 = __THREW__; __THREW__ = 0;
      $420 = $419&1;
      if ($420) {
       $421 = ___cxa_find_matching_catch_3(0|0)|0;
       $422 = tempRet0;
       ___clang_call_terminate($421); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      } else {
       $$0 = $414;$$01 = $415;
       label = 138;
      }
     }
     break;
    }
    case 227: {
     $617 = ((($this)) + 68|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $618 = HEAP32[$617>>2]|0; //@line 823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $619 = ($618|0)==(0); //@line 238 "modules/font/BMFontRasterizer.cpp"
     do {
      if ($619) {
       $620 = (___cxa_allocate_exception(16)|0); //@line 239 "modules/font/BMFontRasterizer.cpp"
       __THREW__ = 0;
       invoke_viii(3349,($620|0),(367501|0),($vararg_buffer3|0)); //@line 239 "modules/font/BMFontRasterizer.cpp"
       $621 = __THREW__; __THREW__ = 0;
       $622 = $621&1;
       if (!($622)) {
        __THREW__ = 0;
        invoke_viii(3350,($620|0),(96|0),(344|0)); //@line 239 "modules/font/BMFontRasterizer.cpp"
        $623 = __THREW__; __THREW__ = 0;
        break;
       }
       $624 = ___cxa_find_matching_catch_2()|0;
       $625 = tempRet0;
       ___cxa_free_exception(($620|0)); //@line 239 "modules/font/BMFontRasterizer.cpp"
       $$2 = $624;$$23 = $625;
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($line); //@line 273 "modules/font/BMFontRasterizer.cpp"
       HEAP32[$ss>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
       HEAP32[$23>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
       $773 = ((($ss)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
       HEAP32[$773>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
       HEAP32[$25>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
       __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($25); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
       __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($23); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
       ___resumeException($$2|0);
       // unreachable;
      } else {
       $626 = ((($this)) + 64|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $__begin$sroa$0$0156 = HEAP32[$626>>2]|0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $627 = ($__begin$sroa$0$0156|0)==(0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($627) {
        $683 = HEAP32[$53>>2]|0; //@line 272 "modules/font/BMFontRasterizer.cpp"
        $684 = ((($this)) + 20|0); //@line 272 "modules/font/BMFontRasterizer.cpp"
        HEAP32[$684>>2] = $683; //@line 272 "modules/font/BMFontRasterizer.cpp"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($line); //@line 273 "modules/font/BMFontRasterizer.cpp"
        HEAP32[$ss>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$23>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        $685 = ((($ss)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$685>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$25>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($25); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($23); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        STACKTOP = sp;return; //@line 273 "modules/font/BMFontRasterizer.cpp"
       }
       $628 = HEAP32[$53>>2]|0; //@line 242 "modules/font/BMFontRasterizer.cpp"
       $629 = ($628|0)==(0); //@line 242 "modules/font/BMFontRasterizer.cpp"
       L270: do {
        if ($629) {
         $__begin$sroa$0$0157$us = $__begin$sroa$0$0156;
         while(1) {
          $630 = $__begin$sroa$0$0157$us; //@line 189 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          $631 = ((($630)) + 28|0); //@line 248 "modules/font/BMFontRasterizer.cpp"
          $632 = HEAP32[$631>>2]|0; //@line 248 "modules/font/BMFontRasterizer.cpp"
          $633 = ((($630)) + 24|0); //@line 249 "modules/font/BMFontRasterizer.cpp"
          $634 = HEAP32[$633>>2]|0; //@line 249 "modules/font/BMFontRasterizer.cpp"
          $635 = HEAP8[$50>>0]|0; //@line 251 "modules/font/BMFontRasterizer.cpp"
          $636 = ($635<<24>>24)==(0); //@line 251 "modules/font/BMFontRasterizer.cpp"
          if ($636) {
           $637 = ((($630)) + 8|0); //@line 251 "modules/font/BMFontRasterizer.cpp"
           $638 = HEAP32[$637>>2]|0; //@line 251 "modules/font/BMFontRasterizer.cpp"
           $639 = ($638>>>0)>(127); //@line 251 "modules/font/BMFontRasterizer.cpp"
           if ($639) {
            label = 252;
            break L270;
           }
          }
          $640 = ((($630)) + 20|0); //@line 254 "modules/font/BMFontRasterizer.cpp"
          $641 = HEAP32[$640>>2]|0; //@line 254 "modules/font/BMFontRasterizer.cpp"
          $642 = ($641|0)<(0); //@line 254 "modules/font/BMFontRasterizer.cpp"
          if ($642) {
           $$lcssa = $640;
           label = 258;
           break L270;
          }
          __THREW__ = 0;
          $643 = (invoke_iii(3985,($56|0),($640|0))|0); //@line 254 "modules/font/BMFontRasterizer.cpp"
          $644 = __THREW__; __THREW__ = 0;
          $645 = $644&1;
          if ($645) {
           label = 248;
           break;
          }
          $646 = HEAP32[$643>>2]|0; //@line 139 "./common/Object.h"
          $647 = ($646|0)==(0|0); //@line 254 "modules/font/BMFontRasterizer.cpp"
          if ($647) {
           $$lcssa = $640;
           label = 258;
           break L270;
          }
          __THREW__ = 0;
          $648 = (invoke_iii(3985,($56|0),($640|0))|0); //@line 257 "modules/font/BMFontRasterizer.cpp"
          $649 = __THREW__; __THREW__ = 0;
          $650 = $649&1;
          if ($650) {
           label = 248;
           break;
          }
          $651 = HEAP32[$648>>2]|0; //@line 139 "./common/Object.h"
          $652 = ((($630)) + 12|0); //@line 259 "modules/font/BMFontRasterizer.cpp"
          $653 = HEAP32[$652>>2]|0; //@line 259 "modules/font/BMFontRasterizer.cpp"
          $654 = ((($630)) + 16|0); //@line 259 "modules/font/BMFontRasterizer.cpp"
          $655 = HEAP32[$654>>2]|0; //@line 259 "modules/font/BMFontRasterizer.cpp"
          __THREW__ = 0;
          $656 = (invoke_iiii(3991,($651|0),($653|0),($655|0))|0); //@line 259 "modules/font/BMFontRasterizer.cpp"
          $657 = __THREW__; __THREW__ = 0;
          $658 = $657&1;
          if ($658) {
           label = 248;
           break;
          }
          if (!($656)) {
           $730 = $630;
           label = 264;
           break L270;
          }
          $659 = ($632|0)>(0); //@line 262 "modules/font/BMFontRasterizer.cpp"
          if ($659) {
           $660 = HEAP32[$652>>2]|0; //@line 262 "modules/font/BMFontRasterizer.cpp"
           $661 = (($632) + -1)|0; //@line 262 "modules/font/BMFontRasterizer.cpp"
           $662 = (($661) + ($660))|0; //@line 262 "modules/font/BMFontRasterizer.cpp"
           $663 = HEAP32[$654>>2]|0; //@line 262 "modules/font/BMFontRasterizer.cpp"
           __THREW__ = 0;
           $664 = (invoke_iiii(3991,($651|0),($662|0),($663|0))|0); //@line 262 "modules/font/BMFontRasterizer.cpp"
           $665 = __THREW__; __THREW__ = 0;
           $666 = $665&1;
           if ($666) {
            label = 248;
            break;
           }
           if (!($664)) {
            $$lcssa111 = $632;$747 = $630;
            label = 270;
            break L270;
           }
          }
          $667 = ($634|0)>(0); //@line 265 "modules/font/BMFontRasterizer.cpp"
          if ($667) {
           $668 = HEAP32[$652>>2]|0; //@line 265 "modules/font/BMFontRasterizer.cpp"
           $669 = HEAP32[$654>>2]|0; //@line 265 "modules/font/BMFontRasterizer.cpp"
           $670 = (($634) + -1)|0; //@line 265 "modules/font/BMFontRasterizer.cpp"
           $671 = (($670) + ($669))|0; //@line 265 "modules/font/BMFontRasterizer.cpp"
           __THREW__ = 0;
           $672 = (invoke_iiii(3991,($651|0),($668|0),($671|0))|0); //@line 265 "modules/font/BMFontRasterizer.cpp"
           $673 = __THREW__; __THREW__ = 0;
           $674 = $673&1;
           if ($674) {
            label = 248;
            break;
           }
           if (!($672)) {
            $$lcssa117 = $634;$764 = $630;
            label = 276;
            break L270;
           }
          }
          $675 = HEAP32[$53>>2]|0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
          $676 = HEAP32[$633>>2]|0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
          $677 = ($675|0)<($676|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
          $678 = $677 ? $676 : $675; //@line 269 "modules/font/BMFontRasterizer.cpp"
          HEAP32[$53>>2] = $678; //@line 269 "modules/font/BMFontRasterizer.cpp"
          $679 = $__begin$sroa$0$0157$us; //@line 174 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          $__begin$sroa$0$0$us = HEAP32[$679>>2]|0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          $680 = ($__begin$sroa$0$0$us|0)==(0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          if ($680) {
           label = 249;
           break;
          } else {
           $__begin$sroa$0$0157$us = $__begin$sroa$0$0$us;
          }
         }
         if ((label|0) == 248) {
          $681 = ___cxa_find_matching_catch_2()|0;
          $682 = tempRet0;
          $eh$lpad$body$index35Z2D = $682;$eh$lpad$body$indexZ2D = $681;
          label = 64;
          break L244;
         }
         else if ((label|0) == 249) {
          $683 = HEAP32[$53>>2]|0; //@line 272 "modules/font/BMFontRasterizer.cpp"
          $684 = ((($this)) + 20|0); //@line 272 "modules/font/BMFontRasterizer.cpp"
          HEAP32[$684>>2] = $683; //@line 272 "modules/font/BMFontRasterizer.cpp"
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($line); //@line 273 "modules/font/BMFontRasterizer.cpp"
          HEAP32[$ss>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
          HEAP32[$23>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
          $685 = ((($ss)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
          HEAP32[$685>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
          HEAP32[$25>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
          __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($25); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
          __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($23); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
          STACKTOP = sp;return; //@line 273 "modules/font/BMFontRasterizer.cpp"
         }
        } else {
         $__begin$sroa$0$0157 = $__begin$sroa$0$0156;
         while(1) {
          $686 = $__begin$sroa$0$0157; //@line 189 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          $687 = ((($686)) + 28|0); //@line 248 "modules/font/BMFontRasterizer.cpp"
          $688 = HEAP32[$687>>2]|0; //@line 248 "modules/font/BMFontRasterizer.cpp"
          $689 = ((($686)) + 24|0); //@line 249 "modules/font/BMFontRasterizer.cpp"
          $690 = HEAP32[$689>>2]|0; //@line 249 "modules/font/BMFontRasterizer.cpp"
          $691 = HEAP8[$50>>0]|0; //@line 251 "modules/font/BMFontRasterizer.cpp"
          $692 = ($691<<24>>24)==(0); //@line 251 "modules/font/BMFontRasterizer.cpp"
          if ($692) {
           $693 = ((($686)) + 8|0); //@line 251 "modules/font/BMFontRasterizer.cpp"
           $694 = HEAP32[$693>>2]|0; //@line 251 "modules/font/BMFontRasterizer.cpp"
           $695 = ($694>>>0)>(127); //@line 251 "modules/font/BMFontRasterizer.cpp"
           if ($695) {
            label = 252;
            break L270;
           }
          }
          $702 = ((($686)) + 20|0); //@line 254 "modules/font/BMFontRasterizer.cpp"
          $703 = HEAP32[$702>>2]|0; //@line 254 "modules/font/BMFontRasterizer.cpp"
          $704 = ($703|0)<(0); //@line 254 "modules/font/BMFontRasterizer.cpp"
          if ($704) {
           $$lcssa = $702;
           label = 258;
           break L270;
          }
          __THREW__ = 0;
          $705 = (invoke_iii(3985,($56|0),($702|0))|0); //@line 254 "modules/font/BMFontRasterizer.cpp"
          $706 = __THREW__; __THREW__ = 0;
          $707 = $706&1;
          if ($707) {
           label = 61;
           break;
          }
          $708 = HEAP32[$705>>2]|0; //@line 139 "./common/Object.h"
          $709 = ($708|0)==(0|0); //@line 254 "modules/font/BMFontRasterizer.cpp"
          if ($709) {
           $$lcssa = $702;
           label = 258;
           break L270;
          }
          __THREW__ = 0;
          $717 = (invoke_iii(3985,($56|0),($702|0))|0); //@line 257 "modules/font/BMFontRasterizer.cpp"
          $718 = __THREW__; __THREW__ = 0;
          $719 = $718&1;
          if ($719) {
           label = 61;
           break;
          }
          $720 = HEAP32[$717>>2]|0; //@line 139 "./common/Object.h"
          $721 = ((($686)) + 12|0); //@line 259 "modules/font/BMFontRasterizer.cpp"
          $722 = HEAP32[$721>>2]|0; //@line 259 "modules/font/BMFontRasterizer.cpp"
          $723 = ((($686)) + 16|0); //@line 259 "modules/font/BMFontRasterizer.cpp"
          $724 = HEAP32[$723>>2]|0; //@line 259 "modules/font/BMFontRasterizer.cpp"
          __THREW__ = 0;
          $725 = (invoke_iiii(3991,($720|0),($722|0),($724|0))|0); //@line 259 "modules/font/BMFontRasterizer.cpp"
          $726 = __THREW__; __THREW__ = 0;
          $727 = $726&1;
          if ($727) {
           label = 61;
           break;
          }
          if (!($725)) {
           $730 = $686;
           label = 264;
           break L270;
          }
          $737 = ($688|0)>(0); //@line 262 "modules/font/BMFontRasterizer.cpp"
          if ($737) {
           $738 = HEAP32[$721>>2]|0; //@line 262 "modules/font/BMFontRasterizer.cpp"
           $739 = (($688) + -1)|0; //@line 262 "modules/font/BMFontRasterizer.cpp"
           $740 = (($739) + ($738))|0; //@line 262 "modules/font/BMFontRasterizer.cpp"
           $741 = HEAP32[$723>>2]|0; //@line 262 "modules/font/BMFontRasterizer.cpp"
           __THREW__ = 0;
           $742 = (invoke_iiii(3991,($720|0),($740|0),($741|0))|0); //@line 262 "modules/font/BMFontRasterizer.cpp"
           $743 = __THREW__; __THREW__ = 0;
           $744 = $743&1;
           if ($744) {
            label = 61;
            break;
           }
           if (!($742)) {
            $$lcssa111 = $688;$747 = $686;
            label = 270;
            break L270;
           }
          }
          $754 = ($690|0)>(0); //@line 265 "modules/font/BMFontRasterizer.cpp"
          if ($754) {
           $755 = HEAP32[$721>>2]|0; //@line 265 "modules/font/BMFontRasterizer.cpp"
           $756 = HEAP32[$723>>2]|0; //@line 265 "modules/font/BMFontRasterizer.cpp"
           $757 = (($690) + -1)|0; //@line 265 "modules/font/BMFontRasterizer.cpp"
           $758 = (($757) + ($756))|0; //@line 265 "modules/font/BMFontRasterizer.cpp"
           __THREW__ = 0;
           $759 = (invoke_iiii(3991,($720|0),($755|0),($758|0))|0); //@line 265 "modules/font/BMFontRasterizer.cpp"
           $760 = __THREW__; __THREW__ = 0;
           $761 = $760&1;
           if ($761) {
            label = 61;
            break;
           }
           if (!($759)) {
            $$lcssa117 = $690;$764 = $686;
            label = 276;
            break L270;
           }
          }
          $771 = $__begin$sroa$0$0157; //@line 174 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          $__begin$sroa$0$0 = HEAP32[$771>>2]|0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          $772 = ($__begin$sroa$0$0|0)==(0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          if ($772) {
           label = 249;
           break;
          } else {
           $__begin$sroa$0$0157 = $__begin$sroa$0$0;
          }
         }
         if ((label|0) == 61) {
          $226 = ___cxa_find_matching_catch_2()|0;
          $227 = tempRet0;
          $eh$lpad$body$index35Z2D = $227;$eh$lpad$body$indexZ2D = $226;
          label = 64;
          break L244;
         }
         else if ((label|0) == 249) {
          $683 = HEAP32[$53>>2]|0; //@line 272 "modules/font/BMFontRasterizer.cpp"
          $684 = ((($this)) + 20|0); //@line 272 "modules/font/BMFontRasterizer.cpp"
          HEAP32[$684>>2] = $683; //@line 272 "modules/font/BMFontRasterizer.cpp"
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($line); //@line 273 "modules/font/BMFontRasterizer.cpp"
          HEAP32[$ss>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
          HEAP32[$23>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
          $685 = ((($ss)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
          HEAP32[$685>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
          HEAP32[$25>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
          __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($25); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
          __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($23); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
          STACKTOP = sp;return; //@line 273 "modules/font/BMFontRasterizer.cpp"
         }
        }
       } while(0);
       if ((label|0) == 252) {
        $696 = (___cxa_allocate_exception(16)|0); //@line 252 "modules/font/BMFontRasterizer.cpp"
        __THREW__ = 0;
        invoke_viii(3349,($696|0),(367549|0),($vararg_buffer5|0)); //@line 252 "modules/font/BMFontRasterizer.cpp"
        $697 = __THREW__; __THREW__ = 0;
        $698 = $697&1;
        if (!($698)) {
         __THREW__ = 0;
         invoke_viii(3350,($696|0),(96|0),(344|0)); //@line 252 "modules/font/BMFontRasterizer.cpp"
         $699 = __THREW__; __THREW__ = 0;
         break;
        }
        $700 = ___cxa_find_matching_catch_2()|0;
        $701 = tempRet0;
        ___cxa_free_exception(($696|0)); //@line 252 "modules/font/BMFontRasterizer.cpp"
        $$2 = $700;$$23 = $701;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($line); //@line 273 "modules/font/BMFontRasterizer.cpp"
        HEAP32[$ss>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$23>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        $773 = ((($ss)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$773>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$25>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($25); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($23); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        ___resumeException($$2|0);
        // unreachable;
       }
       else if ((label|0) == 258) {
        $710 = (___cxa_allocate_exception(16)|0); //@line 255 "modules/font/BMFontRasterizer.cpp"
        $711 = HEAP32[$$lcssa>>2]|0; //@line 255 "modules/font/BMFontRasterizer.cpp"
        __THREW__ = 0;
        HEAP32[$vararg_buffer7>>2] = $711; //@line 255 "modules/font/BMFontRasterizer.cpp"
        invoke_viii(3349,($710|0),(367616|0),($vararg_buffer7|0)); //@line 255 "modules/font/BMFontRasterizer.cpp"
        $712 = __THREW__; __THREW__ = 0;
        $713 = $712&1;
        if (!($713)) {
         __THREW__ = 0;
         invoke_viii(3350,($710|0),(96|0),(344|0)); //@line 255 "modules/font/BMFontRasterizer.cpp"
         $714 = __THREW__; __THREW__ = 0;
         break;
        }
        $715 = ___cxa_find_matching_catch_2()|0;
        $716 = tempRet0;
        ___cxa_free_exception(($710|0)); //@line 255 "modules/font/BMFontRasterizer.cpp"
        $$2 = $715;$$23 = $716;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($line); //@line 273 "modules/font/BMFontRasterizer.cpp"
        HEAP32[$ss>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$23>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        $773 = ((($ss)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$773>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$25>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($25); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($23); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        ___resumeException($$2|0);
        // unreachable;
       }
       else if ((label|0) == 264) {
        $728 = (___cxa_allocate_exception(16)|0); //@line 260 "modules/font/BMFontRasterizer.cpp"
        $729 = ((($730)) + 8|0); //@line 260 "modules/font/BMFontRasterizer.cpp"
        $731 = HEAP32[$729>>2]|0; //@line 260 "modules/font/BMFontRasterizer.cpp"
        __THREW__ = 0;
        HEAP32[$vararg_buffer9>>2] = $731; //@line 260 "modules/font/BMFontRasterizer.cpp"
        invoke_viii(3349,($728|0),(367653|0),($vararg_buffer9|0)); //@line 260 "modules/font/BMFontRasterizer.cpp"
        $732 = __THREW__; __THREW__ = 0;
        $733 = $732&1;
        if (!($733)) {
         __THREW__ = 0;
         invoke_viii(3350,($728|0),(96|0),(344|0)); //@line 260 "modules/font/BMFontRasterizer.cpp"
         $734 = __THREW__; __THREW__ = 0;
         break;
        }
        $735 = ___cxa_find_matching_catch_2()|0;
        $736 = tempRet0;
        ___cxa_free_exception(($728|0)); //@line 260 "modules/font/BMFontRasterizer.cpp"
        $$2 = $735;$$23 = $736;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($line); //@line 273 "modules/font/BMFontRasterizer.cpp"
        HEAP32[$ss>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$23>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        $773 = ((($ss)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$773>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$25>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($25); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($23); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        ___resumeException($$2|0);
        // unreachable;
       }
       else if ((label|0) == 270) {
        $745 = (___cxa_allocate_exception(16)|0); //@line 263 "modules/font/BMFontRasterizer.cpp"
        $746 = ((($747)) + 8|0); //@line 263 "modules/font/BMFontRasterizer.cpp"
        $748 = HEAP32[$746>>2]|0; //@line 263 "modules/font/BMFontRasterizer.cpp"
        __THREW__ = 0;
        HEAP32[$vararg_buffer12>>2] = $$lcssa111; //@line 263 "modules/font/BMFontRasterizer.cpp"
        $vararg_ptr15 = ((($vararg_buffer12)) + 4|0); //@line 263 "modules/font/BMFontRasterizer.cpp"
        HEAP32[$vararg_ptr15>>2] = $748; //@line 263 "modules/font/BMFontRasterizer.cpp"
        invoke_viii(3349,($745|0),(367698|0),($vararg_buffer12|0)); //@line 263 "modules/font/BMFontRasterizer.cpp"
        $749 = __THREW__; __THREW__ = 0;
        $750 = $749&1;
        if (!($750)) {
         __THREW__ = 0;
         invoke_viii(3350,($745|0),(96|0),(344|0)); //@line 263 "modules/font/BMFontRasterizer.cpp"
         $751 = __THREW__; __THREW__ = 0;
         break;
        }
        $752 = ___cxa_find_matching_catch_2()|0;
        $753 = tempRet0;
        ___cxa_free_exception(($745|0)); //@line 263 "modules/font/BMFontRasterizer.cpp"
        $$2 = $752;$$23 = $753;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($line); //@line 273 "modules/font/BMFontRasterizer.cpp"
        HEAP32[$ss>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$23>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        $773 = ((($ss)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$773>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$25>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($25); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($23); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        ___resumeException($$2|0);
        // unreachable;
       }
       else if ((label|0) == 276) {
        $762 = (___cxa_allocate_exception(16)|0); //@line 266 "modules/font/BMFontRasterizer.cpp"
        $763 = ((($764)) + 8|0); //@line 266 "modules/font/BMFontRasterizer.cpp"
        $765 = HEAP32[$763>>2]|0; //@line 266 "modules/font/BMFontRasterizer.cpp"
        __THREW__ = 0;
        HEAP32[$vararg_buffer16>>2] = $$lcssa117; //@line 266 "modules/font/BMFontRasterizer.cpp"
        $vararg_ptr19 = ((($vararg_buffer16)) + 4|0); //@line 266 "modules/font/BMFontRasterizer.cpp"
        HEAP32[$vararg_ptr19>>2] = $765; //@line 266 "modules/font/BMFontRasterizer.cpp"
        invoke_viii(3349,($762|0),(367740|0),($vararg_buffer16|0)); //@line 266 "modules/font/BMFontRasterizer.cpp"
        $766 = __THREW__; __THREW__ = 0;
        $767 = $766&1;
        if (!($767)) {
         __THREW__ = 0;
         invoke_viii(3350,($762|0),(96|0),(344|0)); //@line 266 "modules/font/BMFontRasterizer.cpp"
         $768 = __THREW__; __THREW__ = 0;
         break;
        }
        $769 = ___cxa_find_matching_catch_2()|0;
        $770 = tempRet0;
        ___cxa_free_exception(($762|0)); //@line 266 "modules/font/BMFontRasterizer.cpp"
        $$2 = $769;$$23 = $770;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($line); //@line 273 "modules/font/BMFontRasterizer.cpp"
        HEAP32[$ss>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$23>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        $773 = ((($ss)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$773>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        HEAP32[$25>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($25); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($23); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
        ___resumeException($$2|0);
        // unreachable;
       }
      }
     } while(0);
     $230 = ___cxa_find_matching_catch_2()|0;
     $231 = tempRet0;
     $eh$lpad$body$index35Z2D = $231;$eh$lpad$body$indexZ2D = $230;
     label = 64;
     break;
    }
   }
   if ((label|0) == 27) {
    $$01$i5 = $lpad$phi$i$index;$$02$i = $lpad$phi$i$index26;
    label = 42;
   }
   else if ((label|0) == 106) {
    $329 = ___cxa_find_matching_catch_2()|0;
    $330 = tempRet0;
    $eh$lpad$body31$index45Z2D = $330;$eh$lpad$body31$indexZ2D = $329;
    label = 107;
   }
   if ((label|0) == 42) {
    $188 = HEAP32[$51>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $189 = ($188|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if (!($189)) {
     $$01$i$i$i$i = $188;
     while(1) {
      $190 = HEAP32[$$01$i$i$i$i>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $191 = ((($$01$i$i$i$i)) + 20|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($191); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
      $192 = ((($$01$i$i$i$i)) + 8|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($192); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
      __ZdlPv($$01$i$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $193 = ($190|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($193) {
       break;
      } else {
       $$01$i$i$i$i = $190;
      }
     }
    }
    $194 = HEAP32[$44>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$44>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $195 = ($194|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    if (!($195)) {
     __ZdlPv($194); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    }
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($cline); //@line 111 "modules/font/BMFontRasterizer.cpp"
    $eh$lpad$body$index35Z2D = $$02$i;$eh$lpad$body$indexZ2D = $$01$i5;
    label = 64;
   }
   else if ((label|0) == 107) {
    $$0 = $eh$lpad$body31$indexZ2D;$$01 = $eh$lpad$body31$index45Z2D;
    label = 138;
   }
   if ((label|0) == 64) {
    $$2 = $eh$lpad$body$indexZ2D;$$23 = $eh$lpad$body$index35Z2D;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($line); //@line 273 "modules/font/BMFontRasterizer.cpp"
    HEAP32[$ss>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    HEAP32[$23>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    $773 = ((($ss)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    HEAP32[$773>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    HEAP32[$25>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($25); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($23); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    ___resumeException($$2|0);
    // unreachable;
   }
   else if ((label|0) == 138) {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($filename); //@line 208 "modules/font/BMFontRasterizer.cpp"
    $$1 = $$0;$$12 = $$01;
   }
   $609 = HEAP32[$51>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $610 = ($609|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($610)) {
    $$01$i$i$i$i89 = $609;
    while(1) {
     $611 = HEAP32[$$01$i$i$i$i89>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $612 = ((($$01$i$i$i$i89)) + 20|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($612); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     $613 = ((($$01$i$i$i$i89)) + 8|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($613); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     __ZdlPv($$01$i$i$i$i89); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $614 = ($611|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     if ($614) {
      break;
     } else {
      $$01$i$i$i$i89 = $611;
     }
    }
   }
   $615 = HEAP32[$52>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$52>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $616 = ($615|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if (!($616)) {
    __ZdlPv($615); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($cline); //@line 48 "modules/font/BMFontRasterizer.cpp"
   $$2 = $$1;$$23 = $$12;
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($line); //@line 273 "modules/font/BMFontRasterizer.cpp"
   HEAP32[$ss>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$23>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   $773 = ((($ss)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$773>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$25>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($25); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($23); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   ___resumeException($$2|0);
   // unreachable;
  }
 } while(0);
 $$0$i = $eh$lpad$body$i$index21Z2D;$$01$i = $eh$lpad$body$i$indexZ2D;
 __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($23); //@line 909 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 ___resumeException($$01$i|0);
 // unreachable;
}
function __ZNSt3__113unordered_mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES6_NS_4hashIS6_EENS_8equal_toIS6_EENS4_INS_4pairIKS6_S6_EEEEEixERSC_($this,$__k) {
 $this = $this|0;
 $__k = $__k|0;
 var $$0 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_15__hash_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($this,$__k)|0); //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $2 = ($1|0)==(0|0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($2)) {
  $$sink = $1;
  $$0 = ((($$sink)) + 20|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
  STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 }
 $3 = (__Znwj(32)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $4 = ((($3)) + 8|0); //@line 1340 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 __THREW__ = 0;
 invoke_vii(3351,($4|0),($__k|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZdlPv($3); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($7|0);
  // unreachable;
 }
 $9 = ((($3)) + 20|0); //@line 1342 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viii(3992,($0|0),($this|0),($3|0)); //@line 1366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9); //@line 1589 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 1589 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  __ZdlPv($3); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($13|0);
  // unreachable;
 }
 $12 = HEAP32[$0>>2]|0; //@line 315 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $$sink = $12;
 $$0 = ((($$sink)) + 20|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_15__hash_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($this,$__k) {
 $this = $this|0;
 $__k = $__k|0;
 var $$0$lcssa$i$i$i$i = 0, $$03$i$i$i$i = 0, $$lcssa = 0, $$mask$i$i = 0, $$mask$i$i$us = 0, $$pre$i$i = 0, $$pre$i$i$us = 0, $$sroa$0$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__data$0$lcssa$i$i$i$i = 0, $__data$02$i$i$i$i = 0, $__h$0$lcssa$i$i$i$i = 0, $__h$01$i$i$i$i = 0, $__h$1$i$i$i$i = 0, $__h$2$i$i$i$i = 0, $__h$3$i$i$i$i = 0, $__lhs_sz$03$i$i3$i = 0, $__lhs_sz$03$i$i3$i$us = 0, $__lp$04$i$i2$i = 0, $__lp$04$i$i2$i$us = 0;
 var $__nd$0 = 0, $__nd$0$us = 0, $__nd$08 = 0, $__nd$09 = 0, $__nd$09$us = 0, $__rp$05$i$i1$i = 0, $__rp$05$i$i1$i$us = 0, $scevgep$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$__k>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $1 = $0 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = ($1<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ((($__k)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = HEAP32[$3>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($__k)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = $2 ? $5 : $4; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($__k)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = HEAP32[$7>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $0&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $9 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $2 ? $10 : $8; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = ($11>>>0)>(3); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($12) {
  $$03$i$i$i$i = $11;$__data$02$i$i$i$i = $6;$__h$01$i$i$i$i = $11;
  while(1) {
   $13 = HEAPU8[$__data$02$i$i$i$i>>0]|(HEAPU8[$__data$02$i$i$i$i+1>>0]<<8)|(HEAPU8[$__data$02$i$i$i$i+2>>0]<<16)|(HEAPU8[$__data$02$i$i$i$i+3>>0]<<24); //@line 3073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $14 = Math_imul($13, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $15 = $14 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $16 = $15 ^ $14; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $17 = Math_imul($16, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $18 = Math_imul($__h$01$i$i$i$i, 1540483477)|0; //@line 3104 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $19 = $17 ^ $18; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $20 = ((($__data$02$i$i$i$i)) + 4|0); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $21 = (($$03$i$i$i$i) + -4)|0; //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $22 = ($21>>>0)>(3); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if ($22) {
    $$03$i$i$i$i = $21;$__data$02$i$i$i$i = $20;$__h$01$i$i$i$i = $19;
   } else {
    $$lcssa = $19;
    break;
   }
  }
  $23 = (($11) + -4)|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = $23 & -4; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $25 = (($24) + 4)|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (($23) - ($24))|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $scevgep$i$i$i$i = (($6) + ($25)|0); //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $$0$lcssa$i$i$i$i = $26;$__data$0$lcssa$i$i$i$i = $scevgep$i$i$i$i;$__h$0$lcssa$i$i$i$i = $$lcssa;
 } else {
  $$0$lcssa$i$i$i$i = $11;$__data$0$lcssa$i$i$i$i = $6;$__h$0$lcssa$i$i$i$i = $11;
 }
 switch ($$0$lcssa$i$i$i$i|0) {
 case 3:  {
  $27 = ((($__data$0$lcssa$i$i$i$i)) + 2|0); //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $28 = HEAP8[$27>>0]|0; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $29 = $28&255; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $30 = $29 << 16; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $31 = $30 ^ $__h$0$lcssa$i$i$i$i; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$1$i$i$i$i = $31;
  label = 6;
  break;
 }
 case 2:  {
  $__h$1$i$i$i$i = $__h$0$lcssa$i$i$i$i;
  label = 6;
  break;
 }
 case 1:  {
  $__h$2$i$i$i$i = $__h$0$lcssa$i$i$i$i;
  label = 7;
  break;
 }
 default: {
  $__h$3$i$i$i$i = $__h$0$lcssa$i$i$i$i;
 }
 }
 if ((label|0) == 6) {
  $32 = ((($__data$0$lcssa$i$i$i$i)) + 1|0); //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $33 = HEAP8[$32>>0]|0; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $34 = $33&255; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $35 = $34 << 8; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $36 = $35 ^ $__h$1$i$i$i$i; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$2$i$i$i$i = $36;
  label = 7;
 }
 if ((label|0) == 7) {
  $37 = HEAP8[$__data$0$lcssa$i$i$i$i>>0]|0; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $38 = $37&255; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $39 = $38 ^ $__h$2$i$i$i$i; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $40 = Math_imul($39, 1540483477)|0; //@line 3115 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$3$i$i$i$i = $40;
 }
 $41 = $__h$3$i$i$i$i >>> 13; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $42 = $41 ^ $__h$3$i$i$i$i; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $43 = Math_imul($42, 1540483477)|0; //@line 3118 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $44 = $43 >>> 15; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $45 = $44 ^ $43; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $46 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = HEAP32[$46>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $48 = ($47|0)==(0); //@line 2018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($48) {
  $$sroa$0$0 = 0;
  return ($$sroa$0$0|0); //@line 2038 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $49 = (($47) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $50 = $49 & $47; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $51 = ($50|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($51) {
  $52 = $45 & $49; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $56 = $52;
 } else {
  $53 = (($45>>>0) % ($47>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $56 = $53;
 }
 $54 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $55 = (($54) + ($56<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $57 = HEAP32[$55>>2]|0; //@line 2021 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $58 = ($57|0)==(0|0); //@line 2022 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($58) {
  $$sroa$0$0 = 0;
  return ($$sroa$0$0|0); //@line 2038 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $__nd$08 = HEAP32[$57>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $59 = ($__nd$08|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($59) {
  $$sroa$0$0 = 0;
  return ($$sroa$0$0|0); //@line 2038 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $60 = ($11|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($51) {
  $__nd$09$us = $__nd$08;
  L27: while(1) {
   $61 = ((($__nd$09$us)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $62 = HEAP32[$61>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $63 = $62 & $49; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $64 = ($63|0)==($56|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($64)) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   }
   $65 = ((($__nd$09$us)) + 8|0); //@line 2028 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $66 = HEAP16[$65>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $67 = $66 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $68 = ($67<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $69 = ((($__nd$09$us)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $70 = HEAP32[$69>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$mask$i$i$us = $66 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $71 = $$mask$i$i$us&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $72 = $71 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $73 = $68 ? $72 : $70; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $74 = ($73|0)==($11|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $75 = ($66&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $76 = $75&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   do {
    if ($74) {
     if (!($68)) {
      $77 = ((($__nd$09$us)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $78 = HEAP32[$77>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $79 = (_memcmp($78,$6,$11)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $80 = ($79|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($80) {
       $$sroa$0$0 = $__nd$09$us;
       label = 35;
       break L27;
      } else {
       break;
      }
     }
     if ($60) {
      $$sroa$0$0 = $__nd$09$us;
      label = 35;
      break L27;
     }
     $81 = HEAP8[$6>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $82 = ($76<<24>>24)==($81<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($82) {
      $83 = ((($65)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $__lhs_sz$03$i$i3$i$us = $72;$__lp$04$i$i2$i$us = $83;$__rp$05$i$i1$i$us = $6;
      while(1) {
       $84 = (($__lhs_sz$03$i$i3$i$us) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $85 = ((($__lp$04$i$i2$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $86 = ($84|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($86) {
        $$sroa$0$0 = $__nd$09$us;
        label = 35;
        break L27;
       }
       $87 = ((($__rp$05$i$i1$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$pre$i$i$us = HEAP8[$85>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $88 = HEAP8[$87>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $89 = ($$pre$i$i$us<<24>>24)==($88<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($89) {
        $__lhs_sz$03$i$i3$i$us = $84;$__lp$04$i$i2$i$us = $85;$__rp$05$i$i1$i$us = $87;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   $__nd$0$us = HEAP32[$__nd$09$us>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $90 = ($__nd$0$us|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($90) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   } else {
    $__nd$09$us = $__nd$0$us;
   }
  }
  if ((label|0) == 35) {
   return ($$sroa$0$0|0); //@line 2038 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 } else {
  $__nd$09 = $__nd$08;
  L47: while(1) {
   $91 = ((($__nd$09)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $92 = HEAP32[$91>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $93 = (($92>>>0) % ($47>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $94 = ($93|0)==($56|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($94)) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   }
   $95 = ((($__nd$09)) + 8|0); //@line 2028 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $96 = HEAP16[$95>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $97 = $96 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $98 = ($97<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $99 = ((($__nd$09)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $100 = HEAP32[$99>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$mask$i$i = $96 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $101 = $$mask$i$i&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $102 = $101 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $103 = $98 ? $102 : $100; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $104 = ($103|0)==($11|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $105 = ($96&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $106 = $105&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   do {
    if ($104) {
     if (!($98)) {
      $116 = ((($__nd$09)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $117 = HEAP32[$116>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $118 = (_memcmp($117,$6,$11)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $119 = ($118|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($119) {
       $$sroa$0$0 = $__nd$09;
       label = 35;
       break L47;
      } else {
       break;
      }
     }
     if ($60) {
      $$sroa$0$0 = $__nd$09;
      label = 35;
      break L47;
     }
     $107 = HEAP8[$6>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $108 = ($106<<24>>24)==($107<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($108) {
      $109 = ((($95)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $__lhs_sz$03$i$i3$i = $102;$__lp$04$i$i2$i = $109;$__rp$05$i$i1$i = $6;
      while(1) {
       $110 = (($__lhs_sz$03$i$i3$i) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $111 = ((($__lp$04$i$i2$i)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $112 = ($110|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($112) {
        $$sroa$0$0 = $__nd$09;
        label = 35;
        break L47;
       }
       $113 = ((($__rp$05$i$i1$i)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$pre$i$i = HEAP8[$111>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $114 = HEAP8[$113>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $115 = ($$pre$i$i<<24>>24)==($114<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($115) {
        $__lhs_sz$03$i$i3$i = $110;$__lp$04$i$i2$i = $111;$__rp$05$i$i1$i = $113;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   $__nd$0 = HEAP32[$__nd$09>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $120 = ($__nd$0|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($120) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   } else {
    $__nd$09 = $__nd$0;
   }
  }
  if ((label|0) == 35) {
   return ($$sroa$0$0|0); //@line 2038 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 }
 return (0)|0;
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE20__node_insert_uniqueEPNS_11__hash_nodeIS8_PvEE($agg$result,$this,$__nd) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $__nd = $__nd|0;
 var $$0$lcssa$i$i$i$i = 0, $$03$i$i$i$i = 0, $$lcssa = 0, $$mask$i$i = 0, $$mask$i$i$us = 0, $$pre$i$i = 0, $$pre$i$i$us = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0.0, $127 = 0.0;
 var $128 = 0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0.0, $142 = 0.0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__bc$0 = 0, $__chash$0 = 0, $__chash$1 = 0, $__data$0$lcssa$i$i$i$i = 0, $__data$02$i$i$i$i = 0, $__h$0$lcssa$i$i$i$i = 0, $__h$01$i$i$i$i = 0, $__h$1$i$i$i$i = 0, $__h$2$i$i$i$i = 0, $__h$3$i$i$i$i = 0, $__inserted$0 = 0, $__lhs_sz$03$i$i3$i = 0, $__lhs_sz$03$i$i3$i$us = 0, $__lp$04$i$i2$i = 0, $__lp$04$i$i2$i$us = 0;
 var $__ndptr$0 = 0, $__ndptr$0$us = 0, $__ndptr$011 = 0, $__ndptr$012 = 0, $__ndptr$012$us = 0, $__ndptr$1 = 0, $__rp$05$i$i1$i = 0, $__rp$05$i$i1$i$us = 0, $or$cond = 0, $scevgep$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($__nd)) + 8|0); //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $1 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = $1 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ($2<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ((($__nd)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = HEAP32[$4>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = $3 ? $6 : $5; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = ((($__nd)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = HEAP32[$8>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $1&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $10 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = $3 ? $11 : $9; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ($12>>>0)>(3); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($13) {
  $$03$i$i$i$i = $12;$__data$02$i$i$i$i = $7;$__h$01$i$i$i$i = $12;
  while(1) {
   $14 = HEAPU8[$__data$02$i$i$i$i>>0]|(HEAPU8[$__data$02$i$i$i$i+1>>0]<<8)|(HEAPU8[$__data$02$i$i$i$i+2>>0]<<16)|(HEAPU8[$__data$02$i$i$i$i+3>>0]<<24); //@line 3073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $15 = Math_imul($14, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $16 = $15 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $17 = $16 ^ $15; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $18 = Math_imul($17, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $19 = Math_imul($__h$01$i$i$i$i, 1540483477)|0; //@line 3104 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $20 = $18 ^ $19; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $21 = ((($__data$02$i$i$i$i)) + 4|0); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $22 = (($$03$i$i$i$i) + -4)|0; //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $23 = ($22>>>0)>(3); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if ($23) {
    $$03$i$i$i$i = $22;$__data$02$i$i$i$i = $21;$__h$01$i$i$i$i = $20;
   } else {
    $$lcssa = $20;
    break;
   }
  }
  $24 = (($12) + -4)|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $25 = $24 & -4; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (($25) + 4)|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $27 = (($24) - ($25))|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $scevgep$i$i$i$i = (($7) + ($26)|0); //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $$0$lcssa$i$i$i$i = $27;$__data$0$lcssa$i$i$i$i = $scevgep$i$i$i$i;$__h$0$lcssa$i$i$i$i = $$lcssa;
 } else {
  $$0$lcssa$i$i$i$i = $12;$__data$0$lcssa$i$i$i$i = $7;$__h$0$lcssa$i$i$i$i = $12;
 }
 switch ($$0$lcssa$i$i$i$i|0) {
 case 3:  {
  $28 = ((($__data$0$lcssa$i$i$i$i)) + 2|0); //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $29 = HEAP8[$28>>0]|0; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $30 = $29&255; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $31 = $30 << 16; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $32 = $31 ^ $__h$0$lcssa$i$i$i$i; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$1$i$i$i$i = $32;
  label = 6;
  break;
 }
 case 2:  {
  $__h$1$i$i$i$i = $__h$0$lcssa$i$i$i$i;
  label = 6;
  break;
 }
 case 1:  {
  $__h$2$i$i$i$i = $__h$0$lcssa$i$i$i$i;
  label = 7;
  break;
 }
 default: {
  $__h$3$i$i$i$i = $__h$0$lcssa$i$i$i$i;
 }
 }
 if ((label|0) == 6) {
  $33 = ((($__data$0$lcssa$i$i$i$i)) + 1|0); //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $34 = HEAP8[$33>>0]|0; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $35 = $34&255; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $36 = $35 << 8; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $37 = $36 ^ $__h$1$i$i$i$i; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$2$i$i$i$i = $37;
  label = 7;
 }
 if ((label|0) == 7) {
  $38 = HEAP8[$__data$0$lcssa$i$i$i$i>>0]|0; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $39 = $38&255; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $40 = $39 ^ $__h$2$i$i$i$i; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $41 = Math_imul($40, 1540483477)|0; //@line 3115 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$3$i$i$i$i = $41;
 }
 $42 = $__h$3$i$i$i$i >>> 13; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $43 = $42 ^ $__h$3$i$i$i$i; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $44 = Math_imul($43, 1540483477)|0; //@line 3118 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $45 = $44 >>> 15; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $46 = $45 ^ $44; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = ((($__nd)) + 4|0); //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$47>>2] = $46; //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $48 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $49 = HEAP32[$48>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $50 = ($49|0)==(0); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L13: do {
  if ($50) {
   $__chash$0 = 0;
  } else {
   $51 = (($49) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $52 = $51 & $49; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $53 = ($52|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($53) {
    $54 = $46 & $51; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $58 = $54;
   } else {
    $55 = (($46>>>0) % ($49>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $58 = $55;
   }
   $56 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $57 = (($56) + ($58<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $59 = HEAP32[$57>>2]|0; //@line 1603 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $60 = ($59|0)==(0|0); //@line 1604 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($60) {
    $__chash$0 = $58;
   } else {
    $__ndptr$011 = HEAP32[$59>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $61 = ($__ndptr$011|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($61) {
     $__chash$0 = $58;
    } else {
     $62 = ($12|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($53) {
      $__ndptr$012$us = $__ndptr$011;
      L22: while(1) {
       $63 = ((($__ndptr$012$us)) + 4|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $64 = HEAP32[$63>>2]|0; //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $65 = $64 & $51; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $66 = ($65|0)==($58|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if (!($66)) {
        $__chash$0 = $58;
        break L13;
       }
       $67 = ((($__ndptr$012$us)) + 8|0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $68 = HEAP16[$67>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $69 = $68 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $70 = ($69<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $71 = ((($__ndptr$012$us)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $72 = HEAP32[$71>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$mask$i$i$us = $68 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $73 = $$mask$i$i$us&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $74 = $73 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $75 = $70 ? $74 : $72; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $76 = ($75|0)==($12|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $77 = ($68&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $78 = $77&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       do {
        if ($76) {
         if (!($70)) {
          $79 = ((($__ndptr$012$us)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $80 = HEAP32[$79>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $81 = (_memcmp($80,$7,$12)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $82 = ($81|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          if ($82) {
           $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012$us;
           label = 49;
           break L22;
          } else {
           break;
          }
         }
         if ($62) {
          $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012$us;
          label = 49;
          break L22;
         }
         $83 = HEAP8[$7>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $84 = ($78<<24>>24)==($83<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         if ($84) {
          $85 = ((($67)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $__lhs_sz$03$i$i3$i$us = $74;$__lp$04$i$i2$i$us = $85;$__rp$05$i$i1$i$us = $7;
          while(1) {
           $86 = (($__lhs_sz$03$i$i3$i$us) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $87 = ((($__lp$04$i$i2$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $88 = ($86|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($88) {
            $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012$us;
            label = 49;
            break L22;
           }
           $89 = ((($__rp$05$i$i1$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $$pre$i$i$us = HEAP8[$87>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $90 = HEAP8[$89>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $91 = ($$pre$i$i$us<<24>>24)==($90<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($91) {
            $__lhs_sz$03$i$i3$i$us = $86;$__lp$04$i$i2$i$us = $87;$__rp$05$i$i1$i$us = $89;
           } else {
            break;
           }
          }
         }
        }
       } while(0);
       $__ndptr$0$us = HEAP32[$__ndptr$012$us>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $92 = ($__ndptr$0$us|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($92) {
        $__chash$0 = $58;
        break L13;
       } else {
        $__ndptr$012$us = $__ndptr$0$us;
       }
      }
      if ((label|0) == 49) {
       $176 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       HEAP32[$agg$result>>2] = $176; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       $177 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       HEAP8[$177>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      }
     } else {
      $__ndptr$012 = $__ndptr$011;
      L40: while(1) {
       $93 = ((($__ndptr$012)) + 4|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $94 = HEAP32[$93>>2]|0; //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $95 = (($94>>>0) % ($49>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $96 = ($95|0)==($58|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if (!($96)) {
        $__chash$0 = $58;
        break L13;
       }
       $97 = ((($__ndptr$012)) + 8|0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $98 = HEAP16[$97>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $99 = $98 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $100 = ($99<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $101 = ((($__ndptr$012)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $102 = HEAP32[$101>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$mask$i$i = $98 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $103 = $$mask$i$i&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $104 = $103 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $105 = $100 ? $104 : $102; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $106 = ($105|0)==($12|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $107 = ($98&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $108 = $107&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       do {
        if ($106) {
         if (!($100)) {
          $118 = ((($__ndptr$012)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $119 = HEAP32[$118>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $120 = (_memcmp($119,$7,$12)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $121 = ($120|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          if ($121) {
           $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012;
           label = 49;
           break L40;
          } else {
           break;
          }
         }
         if ($62) {
          $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012;
          label = 49;
          break L40;
         }
         $109 = HEAP8[$7>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $110 = ($108<<24>>24)==($109<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         if ($110) {
          $111 = ((($97)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $__lhs_sz$03$i$i3$i = $104;$__lp$04$i$i2$i = $111;$__rp$05$i$i1$i = $7;
          while(1) {
           $112 = (($__lhs_sz$03$i$i3$i) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $113 = ((($__lp$04$i$i2$i)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $114 = ($112|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($114) {
            $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012;
            label = 49;
            break L40;
           }
           $115 = ((($__rp$05$i$i1$i)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $$pre$i$i = HEAP8[$113>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $116 = HEAP8[$115>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $117 = ($$pre$i$i<<24>>24)==($116<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($117) {
            $__lhs_sz$03$i$i3$i = $112;$__lp$04$i$i2$i = $113;$__rp$05$i$i1$i = $115;
           } else {
            break;
           }
          }
         }
        }
       } while(0);
       $__ndptr$0 = HEAP32[$__ndptr$012>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $122 = ($__ndptr$0|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($122) {
        $__chash$0 = $58;
        break L13;
       } else {
        $__ndptr$012 = $__ndptr$0;
       }
      }
      if ((label|0) == 49) {
       $176 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       HEAP32[$agg$result>>2] = $176; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       $177 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       HEAP8[$177>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      }
     }
    }
   }
  }
 } while(0);
 $123 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $124 = HEAP32[$123>>2]|0; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $125 = (($124) + 1)|0; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $126 = (+($125>>>0)); //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $127 = (+($49>>>0)); //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $128 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $129 = +HEAPF32[$128>>2]; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $130 = $127 * $129; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $131 = $126 > $130; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $or$cond = $50 | $131; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 do {
  if ($or$cond) {
   $132 = $49 << 1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $133 = ($49>>>0)>(2); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($133) {
    $134 = (($49) + -1)|0; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $135 = $134 & $49; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $136 = ($135|0)==(0); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $138 = $136;
   } else {
    $138 = 0;
   }
   $137 = $138&1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $139 = $137 | $132; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $140 = $139 ^ 1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $141 = $126 / $129; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $142 = (+Math_ceil((+$141))); //@line 750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/cmath"
   $143 = (~~(($142))>>>0); //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $144 = ($140>>>0)<($143>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $145 = $144 ? $143 : $140; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE6rehashEj($this,$145); //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $146 = HEAP32[$48>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $147 = HEAP32[$47>>2]|0; //@line 1621 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $148 = (($146) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $149 = $148 & $146; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $150 = ($149|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($150) {
    $151 = $148 & $147; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__bc$0 = $146;$__chash$1 = $151;
    break;
   } else {
    $152 = (($147>>>0) % ($146>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__bc$0 = $146;$__chash$1 = $152;
    break;
   }
  } else {
   $__bc$0 = $49;$__chash$1 = $__chash$0;
  }
 } while(0);
 $153 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $154 = (($153) + ($__chash$1<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $155 = HEAP32[$154>>2]|0; //@line 1624 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $156 = ($155|0)==(0|0); //@line 1625 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($156) {
  $157 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $158 = HEAP32[$157>>2]|0; //@line 1628 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$__nd>>2] = $158; //@line 1628 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$157>>2] = $__nd; //@line 1629 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $159 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $160 = (($159) + ($__chash$1<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$160>>2] = $157; //@line 1631 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $161 = HEAP32[$__nd>>2]|0; //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $162 = ($161|0)==(0|0); //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if (!($162)) {
   $163 = ((($161)) + 4|0); //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $164 = HEAP32[$163>>2]|0; //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $165 = (($__bc$0) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $166 = $165 & $__bc$0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $167 = ($166|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($167) {
    $168 = $164 & $165; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $172 = $168;
   } else {
    $169 = (($164>>>0) % ($__bc$0>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $172 = $169;
   }
   $170 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $171 = (($170) + ($172<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$171>>2] = $__nd; //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 } else {
  $173 = HEAP32[$155>>2]|0; //@line 1637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$__nd>>2] = $173; //@line 1637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$155>>2] = $__nd; //@line 1638 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $174 = HEAP32[$123>>2]|0; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $175 = (($174) + 1)|0; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$123>>2] = $175; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__inserted$0 = 1;$__ndptr$1 = $__nd;
 $176 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$agg$result>>2] = $176; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $177 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 HEAP8[$177>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE6rehashEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__n|0)==(1); //@line 1937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($0) {
  $$0 = 2;
 } else {
  $1 = (($__n) + -1)|0; //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $2 = $1 & $__n; //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $3 = ($2|0)==(0); //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($3) {
   $$0 = $__n;
  } else {
   $4 = (__ZNSt3__112__next_primeEj($__n)|0); //@line 1940 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $$0 = $4;
  }
 }
 $5 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = HEAP32[$5>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $7 = ($$0>>>0)>($6>>>0); //@line 1942 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($7) {
  __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE8__rehashEj($this,$$0); //@line 1943 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $8 = ($$0>>>0)<($6>>>0); //@line 1944 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($8)) {
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $9 = ($6>>>0)>(2); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($9) {
  $10 = (($6) + -1)|0; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $11 = $10 & $6; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $12 = ($11|0)==(0); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $30 = $12;
 } else {
  $30 = 0;
 }
 $13 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = HEAP32[$13>>2]|0; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $15 = (+($14>>>0)); //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $16 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $17 = +HEAPF32[$16>>2]; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $18 = $15 / $17; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $19 = (+Math_ceil((+$18))); //@line 750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/cmath"
 $20 = (~~(($19))>>>0); //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($30) {
  $21 = (($20) + -1)|0; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $22 = (Math_clz32(($21|0))|0); //@line 785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $23 = (32 - ($22))|0; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $24 = 1 << $23; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $24;
 } else {
  $25 = (__ZNSt3__112__next_primeEj($20)|0); //@line 1950 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $25;
 }
 $26 = ($$0>>>0)<($27>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $28 = $26 ? $27 : $$0; //@line 1946 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $29 = ($28>>>0)<($6>>>0); //@line 1952 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($29)) {
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE8__rehashEj($this,$28); //@line 1953 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE8__rehashEj($this,$__nbc) {
 $this = $this|0;
 $__nbc = $__nbc|0;
 var $$lcssa = 0, $$lcssa113 = 0, $$mask$i$i = 0, $$pre$i$i$us = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__cp$0 = 0, $__cp$0$in2332 = 0, $__cp$0$in26 = 0, $__cp$018 = 0, $__cp$01824 = 0, $__cp$0182429 = 0, $__cp$0182433 = 0, $__cp$01827 = 0, $__cp$019 = 0, $__cp$019$lcssa = 0, $__cp$019$lcssa111 = 0, $__i$034 = 0, $__lhs_sz$03$i$i3$i$us = 0, $__lp$04$i$i2$i$us = 0, $__np$013 = 0, $__np$013$phi = 0;
 var $__np$013$us = 0, $__np$013$us$phi = 0, $__np$013$us36 = 0, $__np$013$us36$phi = 0, $__np$08 = 0, $__phash$0$ph$ph31 = 0, $__pp$0$ph$ph30 = 0, $__pp$0$ph25 = 0, $__pp$0$ph25$lcssa = 0, $__rp$05$i$i1$i$us = 0, $exitcond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = ($__nbc|0)==(0); //@line 1965 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($1) {
  $118 = HEAP32[$this>>2]|0; //@line 2822 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$this>>2] = 0; //@line 2823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $119 = ($118|0)==(0|0); //@line 2824 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  if (!($119)) {
   __ZdlPv($118); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  HEAP32[$0>>2] = 0; //@line 1967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $2 = $__nbc << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = (__Znwj($2)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $4 = HEAP32[$this>>2]|0; //@line 2822 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $3; //@line 2823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ($4|0)==(0|0); //@line 2824 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($5)) {
  __ZdlPv($4); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 HEAP32[$0>>2] = $__nbc; //@line 1967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__i$034 = 0;
 while(1) {
  $9 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $10 = (($9) + ($__i$034<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$10>>2] = 0; //@line 1971 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $11 = (($__i$034) + 1)|0; //@line 1970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $exitcond = ($11|0)==($__nbc|0); //@line 1970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($exitcond) {
   break;
  } else {
   $__i$034 = $11;
  }
 }
 $6 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $7 = HEAP32[$6>>2]|0; //@line 1973 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $8 = ($7|0)==(0|0); //@line 1974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($8) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $12 = ((($7)) + 4|0); //@line 1976 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $13 = HEAP32[$12>>2]|0; //@line 1976 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $14 = (($__nbc) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $15 = $14 & $__nbc; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $16 = ($15|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($16) {
  $17 = $13 & $14; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $21 = $17;
 } else {
  $18 = (($13>>>0) % ($__nbc>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $21 = $18;
 }
 $19 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $20 = (($19) + ($21<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$20>>2] = $6; //@line 1977 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__cp$0182429 = HEAP32[$7>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $22 = ($__cp$0182429|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($22) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 } else {
  $__cp$0$in2332 = $7;$__cp$0182433 = $__cp$0182429;$__phash$0$ph$ph31 = $21;$__pp$0$ph$ph30 = $7;
 }
 L23: while(1) {
  $__cp$0$in26 = $__cp$0$in2332;$__cp$01827 = $__cp$0182433;$__pp$0$ph25 = $__pp$0$ph$ph30;
  L25: while(1) {
   $__cp$019 = $__cp$01827;
   while(1) {
    $24 = ((($__cp$019)) + 4|0); //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $25 = HEAP32[$24>>2]|0; //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($16) {
     $26 = $25 & $14; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $29 = $26;
    } else {
     $27 = (($25>>>0) % ($__nbc>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $29 = $27;
    }
    $28 = ($29|0)==($__phash$0$ph$ph31|0); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($28) {
     $__cp$019$lcssa = $__cp$019;
     break;
    }
    $30 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $31 = (($30) + ($29<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $32 = HEAP32[$31>>2]|0; //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $33 = ($32|0)==(0|0); //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($33) {
     $$lcssa = $31;$$lcssa113 = $29;$__cp$019$lcssa111 = $__cp$019;$__pp$0$ph25$lcssa = $__pp$0$ph25;
     break L25;
    }
    $34 = HEAP32[$__cp$019>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $35 = ($34|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    L35: do {
     if ($35) {
      $__np$08 = $__cp$019;
     } else {
      $36 = ((($__cp$019)) + 8|0); //@line 1997 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $37 = HEAP16[$36>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $38 = $37 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $39 = ($38<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $40 = ((($__cp$019)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $41 = HEAP32[$40>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $$mask$i$i = $37 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $42 = $$mask$i$i&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $43 = $42 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $44 = $39 ? $43 : $41; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $45 = ($37&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $46 = $45&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $47 = ($44|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $48 = ((($36)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $49 = ((($__cp$019)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($39) {
       $51 = $34;$__np$013$us = $__cp$019;
       while(1) {
        $50 = ((($51)) + 8|0); //@line 1997 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $52 = HEAP8[$50>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $53 = $52 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $54 = ($53<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $55 = ((($51)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $56 = HEAP32[$55>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $57 = $52&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $58 = $57 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $59 = $54 ? $58 : $56; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $60 = ($44|0)==($59|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if (!($60)) {
         $__np$08 = $__np$013$us;
         break L35;
        }
        $61 = ((($51)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $62 = HEAP32[$61>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $63 = ((($50)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $64 = $54 ? $63 : $62; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        L41: do {
         if (!($47)) {
          $65 = HEAP8[$64>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $66 = ($46<<24>>24)==($65<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          if ($66) {
           $__lhs_sz$03$i$i3$i$us = $43;$__lp$04$i$i2$i$us = $48;$__rp$05$i$i1$i$us = $64;
          } else {
           $__np$08 = $__np$013$us;
           break L35;
          }
          while(1) {
           $67 = (($__lhs_sz$03$i$i3$i$us) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $68 = ((($__lp$04$i$i2$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $69 = ($67|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($69) {
            break L41;
           }
           $70 = ((($__rp$05$i$i1$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $$pre$i$i$us = HEAP8[$68>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $71 = HEAP8[$70>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $72 = ($$pre$i$i$us<<24>>24)==($71<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($72) {
            $__lhs_sz$03$i$i3$i$us = $67;$__lp$04$i$i2$i$us = $68;$__rp$05$i$i1$i$us = $70;
           } else {
            $__np$08 = $__np$013$us;
            break L35;
           }
          }
         }
        } while(0);
        $73 = HEAP32[$51>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $74 = ($73|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        if ($74) {
         $__np$08 = $51;
         break;
        } else {
         $__np$013$us$phi = $51;$51 = $73;$__np$013$us = $__np$013$us$phi;
        }
       }
      } else {
       if ($47) {
        $76 = $34;$__np$013$us36 = $__cp$019;
        while(1) {
         $75 = ((($76)) + 8|0); //@line 1997 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $77 = HEAP8[$75>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $78 = $77 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $79 = ($78<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $80 = ((($76)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $81 = HEAP32[$80>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $82 = $77&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $83 = $82 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $84 = $79 ? $83 : $81; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $85 = ($84|0)==(0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         if (!($85)) {
          $__np$08 = $__np$013$us36;
          break L35;
         }
         $86 = HEAP32[$76>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $87 = ($86|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         if ($87) {
          $__np$08 = $76;
          break L35;
         } else {
          $__np$013$us36$phi = $76;$76 = $86;$__np$013$us36 = $__np$013$us36$phi;
         }
        }
       } else {
        $90 = $34;$__np$013 = $__cp$019;
       }
       while(1) {
        $89 = ((($90)) + 8|0); //@line 1997 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $91 = HEAP8[$89>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $92 = $91 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $93 = ($92<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $94 = ((($90)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $95 = HEAP32[$94>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $96 = $91&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $97 = $96 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $98 = $93 ? $97 : $95; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $99 = ($44|0)==($98|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if (!($99)) {
         $__np$08 = $__np$013;
         break L35;
        }
        $100 = ((($90)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $101 = HEAP32[$100>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $102 = ((($89)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $103 = $93 ? $102 : $101; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $104 = HEAP32[$49>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $105 = (_memcmp($104,$103,$44)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $106 = ($105|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if (!($106)) {
         $__np$08 = $__np$013;
         break L35;
        }
        $107 = HEAP32[$90>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $108 = ($107|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        if ($108) {
         $__np$08 = $90;
         break;
        } else {
         $__np$013$phi = $90;$90 = $107;$__np$013 = $__np$013$phi;
        }
       }
      }
     }
    } while(0);
    $109 = HEAP32[$__np$08>>2]|0; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    HEAP32[$__pp$0$ph25>>2] = $109; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $110 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $111 = (($110) + ($29<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $112 = HEAP32[$111>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $113 = HEAP32[$112>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    HEAP32[$__np$08>>2] = $113; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $114 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $115 = (($114) + ($29<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $116 = HEAP32[$115>>2]|0; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    HEAP32[$116>>2] = $__cp$019; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__cp$0 = HEAP32[$__cp$0$in26>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $117 = ($__cp$0|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($117) {
     label = 38;
     break L23;
    } else {
     $__cp$019 = $__cp$0;
    }
   }
   $__cp$018 = HEAP32[$__cp$019$lcssa>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $23 = ($__cp$018|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($23) {
    label = 38;
    break L23;
   } else {
    $__cp$0$in26 = $__cp$019$lcssa;$__cp$01827 = $__cp$018;$__pp$0$ph25 = $__cp$019$lcssa;
   }
  }
  HEAP32[$$lcssa>>2] = $__pp$0$ph25$lcssa; //@line 1989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $__cp$01824 = HEAP32[$__cp$019$lcssa111>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $88 = ($__cp$01824|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($88) {
   label = 38;
   break;
  } else {
   $__cp$0$in2332 = $__cp$019$lcssa111;$__cp$0182433 = $__cp$01824;$__phash$0$ph$ph31 = $$lcssa113;$__pp$0$ph$ph30 = $__cp$019$lcssa111;
  }
 }
 if ((label|0) == 38) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
}
function __ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeIS8_PvEEEERKT_($this,$__k) {
 $this = $this|0;
 $__k = $__k|0;
 var $$0$lcssa$i$i$i$i = 0, $$03$i$i$i$i = 0, $$lcssa = 0, $$mask$i$i = 0, $$mask$i$i$us = 0, $$pre$i$i = 0, $$pre$i$i$us = 0, $$sroa$0$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__data$0$lcssa$i$i$i$i = 0, $__data$02$i$i$i$i = 0, $__h$0$lcssa$i$i$i$i = 0, $__h$01$i$i$i$i = 0, $__h$1$i$i$i$i = 0, $__h$2$i$i$i$i = 0, $__h$3$i$i$i$i = 0, $__lhs_sz$03$i$i3$i = 0, $__lhs_sz$03$i$i3$i$us = 0, $__lp$04$i$i2$i = 0, $__lp$04$i$i2$i$us = 0;
 var $__nd$0 = 0, $__nd$0$us = 0, $__nd$08 = 0, $__nd$09 = 0, $__nd$09$us = 0, $__rp$05$i$i1$i = 0, $__rp$05$i$i1$i$us = 0, $scevgep$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$__k>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $1 = $0 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = ($1<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ((($__k)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = HEAP32[$3>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($__k)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = $2 ? $5 : $4; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($__k)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = HEAP32[$7>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $0&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $9 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $2 ? $10 : $8; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = ($11>>>0)>(3); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($12) {
  $$03$i$i$i$i = $11;$__data$02$i$i$i$i = $6;$__h$01$i$i$i$i = $11;
  while(1) {
   $13 = HEAPU8[$__data$02$i$i$i$i>>0]|(HEAPU8[$__data$02$i$i$i$i+1>>0]<<8)|(HEAPU8[$__data$02$i$i$i$i+2>>0]<<16)|(HEAPU8[$__data$02$i$i$i$i+3>>0]<<24); //@line 3073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $14 = Math_imul($13, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $15 = $14 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $16 = $15 ^ $14; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $17 = Math_imul($16, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $18 = Math_imul($__h$01$i$i$i$i, 1540483477)|0; //@line 3104 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $19 = $17 ^ $18; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $20 = ((($__data$02$i$i$i$i)) + 4|0); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $21 = (($$03$i$i$i$i) + -4)|0; //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $22 = ($21>>>0)>(3); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if ($22) {
    $$03$i$i$i$i = $21;$__data$02$i$i$i$i = $20;$__h$01$i$i$i$i = $19;
   } else {
    $$lcssa = $19;
    break;
   }
  }
  $23 = (($11) + -4)|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = $23 & -4; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $25 = (($24) + 4)|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (($23) - ($24))|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $scevgep$i$i$i$i = (($6) + ($25)|0); //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $$0$lcssa$i$i$i$i = $26;$__data$0$lcssa$i$i$i$i = $scevgep$i$i$i$i;$__h$0$lcssa$i$i$i$i = $$lcssa;
 } else {
  $$0$lcssa$i$i$i$i = $11;$__data$0$lcssa$i$i$i$i = $6;$__h$0$lcssa$i$i$i$i = $11;
 }
 switch ($$0$lcssa$i$i$i$i|0) {
 case 3:  {
  $27 = ((($__data$0$lcssa$i$i$i$i)) + 2|0); //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $28 = HEAP8[$27>>0]|0; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $29 = $28&255; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $30 = $29 << 16; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $31 = $30 ^ $__h$0$lcssa$i$i$i$i; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$1$i$i$i$i = $31;
  label = 6;
  break;
 }
 case 2:  {
  $__h$1$i$i$i$i = $__h$0$lcssa$i$i$i$i;
  label = 6;
  break;
 }
 case 1:  {
  $__h$2$i$i$i$i = $__h$0$lcssa$i$i$i$i;
  label = 7;
  break;
 }
 default: {
  $__h$3$i$i$i$i = $__h$0$lcssa$i$i$i$i;
 }
 }
 if ((label|0) == 6) {
  $32 = ((($__data$0$lcssa$i$i$i$i)) + 1|0); //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $33 = HEAP8[$32>>0]|0; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $34 = $33&255; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $35 = $34 << 8; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $36 = $35 ^ $__h$1$i$i$i$i; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$2$i$i$i$i = $36;
  label = 7;
 }
 if ((label|0) == 7) {
  $37 = HEAP8[$__data$0$lcssa$i$i$i$i>>0]|0; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $38 = $37&255; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $39 = $38 ^ $__h$2$i$i$i$i; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $40 = Math_imul($39, 1540483477)|0; //@line 3115 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$3$i$i$i$i = $40;
 }
 $41 = $__h$3$i$i$i$i >>> 13; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $42 = $41 ^ $__h$3$i$i$i$i; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $43 = Math_imul($42, 1540483477)|0; //@line 3118 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $44 = $43 >>> 15; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $45 = $44 ^ $43; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $46 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = HEAP32[$46>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $48 = ($47|0)==(0); //@line 2047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($48) {
  $$sroa$0$0 = 0;
  return ($$sroa$0$0|0); //@line 2068 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $49 = (($47) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $50 = $49 & $47; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $51 = ($50|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($51) {
  $52 = $45 & $49; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $56 = $52;
 } else {
  $53 = (($45>>>0) % ($47>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $56 = $53;
 }
 $54 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $55 = (($54) + ($56<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $57 = HEAP32[$55>>2]|0; //@line 2050 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $58 = ($57|0)==(0|0); //@line 2051 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($58) {
  $$sroa$0$0 = 0;
  return ($$sroa$0$0|0); //@line 2068 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $__nd$08 = HEAP32[$57>>2]|0; //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $59 = ($__nd$08|0)==(0|0); //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($59) {
  $$sroa$0$0 = 0;
  return ($$sroa$0$0|0); //@line 2068 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $60 = ($11|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($51) {
  $__nd$09$us = $__nd$08;
  L27: while(1) {
   $61 = ((($__nd$09$us)) + 4|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $62 = HEAP32[$61>>2]|0; //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $63 = $62 & $49; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $64 = ($63|0)==($56|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($64)) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   }
   $65 = ((($__nd$09$us)) + 8|0); //@line 2057 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $66 = HEAP16[$65>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $67 = $66 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $68 = ($67<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $69 = ((($__nd$09$us)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $70 = HEAP32[$69>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$mask$i$i$us = $66 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $71 = $$mask$i$i$us&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $72 = $71 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $73 = $68 ? $72 : $70; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $74 = ($73|0)==($11|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $75 = ($66&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $76 = $75&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   do {
    if ($74) {
     if (!($68)) {
      $77 = ((($__nd$09$us)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $78 = HEAP32[$77>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $79 = (_memcmp($78,$6,$11)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $80 = ($79|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($80) {
       $$sroa$0$0 = $__nd$09$us;
       label = 35;
       break L27;
      } else {
       break;
      }
     }
     if ($60) {
      $$sroa$0$0 = $__nd$09$us;
      label = 35;
      break L27;
     }
     $81 = HEAP8[$6>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $82 = ($76<<24>>24)==($81<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($82) {
      $83 = ((($65)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $__lhs_sz$03$i$i3$i$us = $72;$__lp$04$i$i2$i$us = $83;$__rp$05$i$i1$i$us = $6;
      while(1) {
       $84 = (($__lhs_sz$03$i$i3$i$us) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $85 = ((($__lp$04$i$i2$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $86 = ($84|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($86) {
        $$sroa$0$0 = $__nd$09$us;
        label = 35;
        break L27;
       }
       $87 = ((($__rp$05$i$i1$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$pre$i$i$us = HEAP8[$85>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $88 = HEAP8[$87>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $89 = ($$pre$i$i$us<<24>>24)==($88<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($89) {
        $__lhs_sz$03$i$i3$i$us = $84;$__lp$04$i$i2$i$us = $85;$__rp$05$i$i1$i$us = $87;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   $__nd$0$us = HEAP32[$__nd$09$us>>2]|0; //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $90 = ($__nd$0$us|0)==(0|0); //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($90) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   } else {
    $__nd$09$us = $__nd$0$us;
   }
  }
  if ((label|0) == 35) {
   return ($$sroa$0$0|0); //@line 2068 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 } else {
  $__nd$09 = $__nd$08;
  L47: while(1) {
   $91 = ((($__nd$09)) + 4|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $92 = HEAP32[$91>>2]|0; //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $93 = (($92>>>0) % ($47>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $94 = ($93|0)==($56|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($94)) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   }
   $95 = ((($__nd$09)) + 8|0); //@line 2057 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $96 = HEAP16[$95>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $97 = $96 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $98 = ($97<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $99 = ((($__nd$09)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $100 = HEAP32[$99>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$mask$i$i = $96 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $101 = $$mask$i$i&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $102 = $101 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $103 = $98 ? $102 : $100; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $104 = ($103|0)==($11|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $105 = ($96&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $106 = $105&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   do {
    if ($104) {
     if (!($98)) {
      $116 = ((($__nd$09)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $117 = HEAP32[$116>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $118 = (_memcmp($117,$6,$11)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $119 = ($118|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($119) {
       $$sroa$0$0 = $__nd$09;
       label = 35;
       break L47;
      } else {
       break;
      }
     }
     if ($60) {
      $$sroa$0$0 = $__nd$09;
      label = 35;
      break L47;
     }
     $107 = HEAP8[$6>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $108 = ($106<<24>>24)==($107<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($108) {
      $109 = ((($95)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $__lhs_sz$03$i$i3$i = $102;$__lp$04$i$i2$i = $109;$__rp$05$i$i1$i = $6;
      while(1) {
       $110 = (($__lhs_sz$03$i$i3$i) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $111 = ((($__lp$04$i$i2$i)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $112 = ($110|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($112) {
        $$sroa$0$0 = $__nd$09;
        label = 35;
        break L47;
       }
       $113 = ((($__rp$05$i$i1$i)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$pre$i$i = HEAP8[$111>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $114 = HEAP8[$113>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $115 = ($$pre$i$i<<24>>24)==($114<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($115) {
        $__lhs_sz$03$i$i3$i = $110;$__lp$04$i$i2$i = $111;$__rp$05$i$i1$i = $113;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   $__nd$0 = HEAP32[$__nd$09>>2]|0; //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $120 = ($__nd$0|0)==(0|0); //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($120) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   } else {
    $__nd$09 = $__nd$0;
   }
  }
  if ((label|0) == 35) {
   return ($$sroa$0$0|0); //@line 2068 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 }
 return (0)|0;
}
function __ZNSt3__113unordered_mapIjN4love4font16BMFontRasterizer15BMFontCharacterENS_4hashIjEENS_8equal_toIjEENS_9allocatorINS_4pairIKjS4_EEEEEixERSB_($this,$__k) {
 $this = $this|0;
 $__k = $__k|0;
 var $$0 = 0, $$pn$i$i = 0, $$pn$us$i$i = 0, $$sroa$0$0$i$i$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $__nd$0$i$i = 0, $__nd$0$us$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = HEAP32[$__k>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $4 = ($3|0)==(0); //@line 2018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L1: do {
  if (!($4)) {
   $5 = (($3) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $6 = $5 & $3; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $7 = ($6|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($7) {
    $8 = $5 & $1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $12 = $8;
   } else {
    $9 = (($1>>>0) % ($3>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $12 = $9;
   }
   $10 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $11 = (($10) + ($12<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $13 = HEAP32[$11>>2]|0; //@line 2021 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $14 = ($13|0)==(0|0); //@line 2022 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($14)) {
    if ($7) {
     $$pn$us$i$i = $13;
     while(1) {
      $__nd$0$us$i$i = HEAP32[$$pn$us$i$i>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $15 = ($__nd$0$us$i$i|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($15) {
       break L1;
      }
      $16 = ((($__nd$0$us$i$i)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $17 = HEAP32[$16>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $18 = $17 & $5; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $19 = ($18|0)==($12|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($19)) {
       break L1;
      }
      $20 = ((($__nd$0$us$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $21 = HEAP32[$20>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $22 = ($21|0)==($1|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($22) {
       $$sroa$0$0$i$i$sink = $__nd$0$us$i$i;
       break;
      } else {
       $$pn$us$i$i = $__nd$0$us$i$i;
      }
     }
     $$0 = ((($$sroa$0$0$i$i$sink)) + 12|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
     STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
    } else {
     $$pn$i$i = $13;
     while(1) {
      $__nd$0$i$i = HEAP32[$$pn$i$i>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $23 = ($__nd$0$i$i|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($23) {
       break L1;
      }
      $24 = ((($__nd$0$i$i)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $25 = HEAP32[$24>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $26 = (($25>>>0) % ($3>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $27 = ($26|0)==($12|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($27)) {
       break L1;
      }
      $28 = ((($__nd$0$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $29 = HEAP32[$28>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $30 = ($29|0)==($1|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($30) {
       $$sroa$0$0$i$i$sink = $__nd$0$i$i;
       break;
      } else {
       $$pn$i$i = $__nd$0$i$i;
      }
     }
     $$0 = ((($$sroa$0$0$i$i$sink)) + 12|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
     STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
    }
   }
  }
 } while(0);
 $31 = (__Znwj(44)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $32 = ((($31)) + 8|0); //@line 1340 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 HEAP32[$32>>2] = $1; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = ((($31)) + 12|0); //@line 1342 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 ;HEAP32[$33>>2]=0|0;HEAP32[$33+4>>2]=0|0;HEAP32[$33+8>>2]=0|0;HEAP32[$33+12>>2]=0|0;HEAP32[$33+16>>2]=0|0;HEAP32[$33+20>>2]=0|0;HEAP32[$33+24>>2]=0|0;HEAP32[$33+28>>2]=0|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 invoke_viii(3993,($0|0),($this|0),($31|0)); //@line 1366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $34 = __THREW__; __THREW__ = 0;
 $35 = $34&1;
 if ($35) {
  $37 = ___cxa_find_matching_catch_2()|0;
  $38 = tempRet0;
  __ZdlPv($31); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($37|0);
  // unreachable;
 }
 $36 = HEAP32[$0>>2]|0; //@line 315 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $$sroa$0$0$i$i$sink = $36;
 $$0 = ((($$sroa$0$0$i$i$sink)) + 12|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeIjN4love4font16BMFontRasterizer15BMFontCharacterEEENS_22__unordered_map_hasherIjS6_NS_4hashIjEELb1EEENS_21__unordered_map_equalIjS6_NS_8equal_toIjEELb1EEENS_9allocatorIS6_EEE20__node_insert_uniqueEPNS_11__hash_nodeIS6_PvEE($agg$result,$this,$__nd) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $__nd = $__nd|0;
 var $$pn = 0, $$pn$us = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $__bc$0 = 0, $__chash$0 = 0, $__chash$1 = 0, $__inserted$0 = 0, $__ndptr$0 = 0, $__ndptr$0$us = 0, $__ndptr$1 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($__nd)) + 8|0); //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $1 = HEAP32[$0>>2]|0; //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $2 = ((($__nd)) + 4|0); //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$2>>2] = $1; //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $3 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $4 = HEAP32[$3>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $5 = ($4|0)==(0); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L1: do {
  if ($5) {
   $__chash$0 = 0;
  } else {
   $6 = (($4) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $7 = $6 & $4; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $8 = ($7|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($8) {
    $9 = $6 & $1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $13 = $9;
   } else {
    $10 = (($1>>>0) % ($4>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $13 = $10;
   }
   $11 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $12 = (($11) + ($13<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $14 = HEAP32[$12>>2]|0; //@line 1603 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $15 = ($14|0)==(0|0); //@line 1604 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($15) {
    $__chash$0 = $13;
   } else {
    if ($8) {
     $$pn$us = $14;
     while(1) {
      $__ndptr$0$us = HEAP32[$$pn$us>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $16 = ($__ndptr$0$us|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($16) {
       $__chash$0 = $13;
       break L1;
      }
      $17 = ((($__ndptr$0$us)) + 4|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $18 = HEAP32[$17>>2]|0; //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $19 = $18 & $6; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $20 = ($19|0)==($13|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($20)) {
       $__chash$0 = $13;
       break L1;
      }
      $21 = ((($__ndptr$0$us)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $22 = HEAP32[$21>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $23 = ($22|0)==($1|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($23) {
       $__inserted$0 = 0;$__ndptr$1 = $__ndptr$0$us;
       break;
      } else {
       $$pn$us = $__ndptr$0$us;
      }
     }
     $85 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     HEAP32[$agg$result>>2] = $85; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     $86 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     HEAP8[$86>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    } else {
     $$pn = $14;
     while(1) {
      $__ndptr$0 = HEAP32[$$pn>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $24 = ($__ndptr$0|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($24) {
       $__chash$0 = $13;
       break L1;
      }
      $25 = ((($__ndptr$0)) + 4|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $26 = HEAP32[$25>>2]|0; //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $27 = (($26>>>0) % ($4>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $28 = ($27|0)==($13|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($28)) {
       $__chash$0 = $13;
       break L1;
      }
      $29 = ((($__ndptr$0)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $30 = HEAP32[$29>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $31 = ($30|0)==($1|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($31) {
       $__inserted$0 = 0;$__ndptr$1 = $__ndptr$0;
       break;
      } else {
       $$pn = $__ndptr$0;
      }
     }
     $85 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     HEAP32[$agg$result>>2] = $85; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     $86 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     HEAP8[$86>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    }
   }
  }
 } while(0);
 $32 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = HEAP32[$32>>2]|0; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $34 = (($33) + 1)|0; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $35 = (+($34>>>0)); //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $36 = (+($4>>>0)); //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $37 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = +HEAPF32[$37>>2]; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $39 = $36 * $38; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $40 = $35 > $39; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $or$cond = $5 | $40; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 do {
  if ($or$cond) {
   $41 = $4 << 1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $42 = ($4>>>0)>(2); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($42) {
    $43 = (($4) + -1)|0; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $44 = $43 & $4; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $45 = ($44|0)==(0); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $47 = $45;
   } else {
    $47 = 0;
   }
   $46 = $47&1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $48 = $46 | $41; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $49 = $48 ^ 1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $50 = $35 / $38; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $51 = (+Math_ceil((+$50))); //@line 750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/cmath"
   $52 = (~~(($51))>>>0); //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $53 = ($49>>>0)<($52>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $54 = $53 ? $52 : $49; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   __ZNSt3__112__hash_tableINS_17__hash_value_typeIjN4love4font16BMFontRasterizer15BMFontCharacterEEENS_22__unordered_map_hasherIjS6_NS_4hashIjEELb1EEENS_21__unordered_map_equalIjS6_NS_8equal_toIjEELb1EEENS_9allocatorIS6_EEE6rehashEj($this,$54); //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $55 = HEAP32[$3>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $56 = HEAP32[$2>>2]|0; //@line 1621 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $57 = (($55) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $58 = $57 & $55; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $59 = ($58|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($59) {
    $60 = $57 & $56; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__bc$0 = $55;$__chash$1 = $60;
    break;
   } else {
    $61 = (($56>>>0) % ($55>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__bc$0 = $55;$__chash$1 = $61;
    break;
   }
  } else {
   $__bc$0 = $4;$__chash$1 = $__chash$0;
  }
 } while(0);
 $62 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $63 = (($62) + ($__chash$1<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $64 = HEAP32[$63>>2]|0; //@line 1624 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $65 = ($64|0)==(0|0); //@line 1625 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($65) {
  $66 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $67 = HEAP32[$66>>2]|0; //@line 1628 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$__nd>>2] = $67; //@line 1628 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$66>>2] = $__nd; //@line 1629 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $68 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $69 = (($68) + ($__chash$1<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$69>>2] = $66; //@line 1631 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $70 = HEAP32[$__nd>>2]|0; //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $71 = ($70|0)==(0|0); //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if (!($71)) {
   $72 = ((($70)) + 4|0); //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $73 = HEAP32[$72>>2]|0; //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $74 = (($__bc$0) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $75 = $74 & $__bc$0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $76 = ($75|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($76) {
    $77 = $73 & $74; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $81 = $77;
   } else {
    $78 = (($73>>>0) % ($__bc$0>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $81 = $78;
   }
   $79 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $80 = (($79) + ($81<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$80>>2] = $__nd; //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 } else {
  $82 = HEAP32[$64>>2]|0; //@line 1637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$__nd>>2] = $82; //@line 1637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$64>>2] = $__nd; //@line 1638 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $83 = HEAP32[$32>>2]|0; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $84 = (($83) + 1)|0; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$32>>2] = $84; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__inserted$0 = 1;$__ndptr$1 = $__nd;
 $85 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$agg$result>>2] = $85; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $86 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 HEAP8[$86>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeIjN4love4font16BMFontRasterizer15BMFontCharacterEEENS_22__unordered_map_hasherIjS6_NS_4hashIjEELb1EEENS_21__unordered_map_equalIjS6_NS_8equal_toIjEELb1EEENS_9allocatorIS6_EEE6rehashEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__n|0)==(1); //@line 1937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($0) {
  $$0 = 2;
 } else {
  $1 = (($__n) + -1)|0; //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $2 = $1 & $__n; //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $3 = ($2|0)==(0); //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($3) {
   $$0 = $__n;
  } else {
   $4 = (__ZNSt3__112__next_primeEj($__n)|0); //@line 1940 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $$0 = $4;
  }
 }
 $5 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = HEAP32[$5>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $7 = ($$0>>>0)>($6>>>0); //@line 1942 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($7) {
  __ZNSt3__112__hash_tableINS_17__hash_value_typeIjN4love4font16BMFontRasterizer15BMFontCharacterEEENS_22__unordered_map_hasherIjS6_NS_4hashIjEELb1EEENS_21__unordered_map_equalIjS6_NS_8equal_toIjEELb1EEENS_9allocatorIS6_EEE8__rehashEj($this,$$0); //@line 1943 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $8 = ($$0>>>0)<($6>>>0); //@line 1944 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($8)) {
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $9 = ($6>>>0)>(2); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($9) {
  $10 = (($6) + -1)|0; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $11 = $10 & $6; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $12 = ($11|0)==(0); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $30 = $12;
 } else {
  $30 = 0;
 }
 $13 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = HEAP32[$13>>2]|0; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $15 = (+($14>>>0)); //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $16 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $17 = +HEAPF32[$16>>2]; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $18 = $15 / $17; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $19 = (+Math_ceil((+$18))); //@line 750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/cmath"
 $20 = (~~(($19))>>>0); //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($30) {
  $21 = (($20) + -1)|0; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $22 = (Math_clz32(($21|0))|0); //@line 785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $23 = (32 - ($22))|0; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $24 = 1 << $23; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $24;
 } else {
  $25 = (__ZNSt3__112__next_primeEj($20)|0); //@line 1950 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $25;
 }
 $26 = ($$0>>>0)<($27>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $28 = $26 ? $27 : $$0; //@line 1946 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $29 = ($28>>>0)<($6>>>0); //@line 1952 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($29)) {
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 __ZNSt3__112__hash_tableINS_17__hash_value_typeIjN4love4font16BMFontRasterizer15BMFontCharacterEEENS_22__unordered_map_hasherIjS6_NS_4hashIjEELb1EEENS_21__unordered_map_equalIjS6_NS_8equal_toIjEELb1EEENS_9allocatorIS6_EEE8__rehashEj($this,$28); //@line 1953 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeIjN4love4font16BMFontRasterizer15BMFontCharacterEEENS_22__unordered_map_hasherIjS6_NS_4hashIjEELb1EEENS_21__unordered_map_equalIjS6_NS_8equal_toIjEELb1EEENS_9allocatorIS6_EEE8__rehashEj($this,$__nbc) {
 $this = $this|0;
 $__nbc = $__nbc|0;
 var $$lcssa = 0, $$lcssa133 = 0, $$lcssa143 = 0, $$lcssa6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $8 = 0, $9 = 0;
 var $__cp$0 = 0, $__cp$0$in1019 = 0, $__cp$0$in13 = 0, $__cp$0$in13$us = 0, $__cp$0$lcssa5 = 0, $__cp$0$us$us = 0, $__cp$08 = 0, $__cp$08$us = 0, $__cp$0811 = 0, $__cp$081116 = 0, $__cp$081120 = 0, $__cp$0814 = 0, $__cp$0814$us = 0, $__cp$09 = 0, $__cp$09$lcssa = 0, $__cp$09$us$us = 0, $__cp$09$us$us$lcssa = 0, $__i$021 = 0, $__np$0 = 0, $__np$0$lcssa = 0;
 var $__np$0$us$us = 0, $__np$0$us$us$lcssa = 0, $__phash$0$ph$ph18 = 0, $__pp$0$ph$lcssa = 0, $__pp$0$ph$ph17 = 0, $__pp$0$ph12 = 0, $__pp$0$ph12$us = 0, $exitcond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = ($__nbc|0)==(0); //@line 1965 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($1) {
  $74 = HEAP32[$this>>2]|0; //@line 2822 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$this>>2] = 0; //@line 2823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $75 = ($74|0)==(0|0); //@line 2824 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  if (!($75)) {
   __ZdlPv($74); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  HEAP32[$0>>2] = 0; //@line 1967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $2 = $__nbc << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = (__Znwj($2)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $4 = HEAP32[$this>>2]|0; //@line 2822 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $3; //@line 2823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ($4|0)==(0|0); //@line 2824 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($5)) {
  __ZdlPv($4); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 HEAP32[$0>>2] = $__nbc; //@line 1967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__i$021 = 0;
 while(1) {
  $9 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $10 = (($9) + ($__i$021<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$10>>2] = 0; //@line 1971 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $11 = (($__i$021) + 1)|0; //@line 1970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $exitcond = ($11|0)==($__nbc|0); //@line 1970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($exitcond) {
   break;
  } else {
   $__i$021 = $11;
  }
 }
 $6 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $7 = HEAP32[$6>>2]|0; //@line 1973 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $8 = ($7|0)==(0|0); //@line 1974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($8) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $12 = ((($7)) + 4|0); //@line 1976 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $13 = HEAP32[$12>>2]|0; //@line 1976 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $14 = (($__nbc) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $15 = $14 & $__nbc; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $16 = ($15|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($16) {
  $17 = $13 & $14; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $21 = $17;
 } else {
  $18 = (($13>>>0) % ($__nbc>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $21 = $18;
 }
 $19 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $20 = (($19) + ($21<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$20>>2] = $6; //@line 1977 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__cp$081116 = HEAP32[$7>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $22 = ($__cp$081116|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($22) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 } else {
  $__cp$0$in1019 = $7;$__cp$081120 = $__cp$081116;$__phash$0$ph$ph18 = $21;$__pp$0$ph$ph17 = $7;
 }
 L23: while(1) {
  L25: do {
   if ($16) {
    $__cp$0$in13$us = $__cp$0$in1019;$__cp$0814$us = $__cp$081120;$__pp$0$ph12$us = $__pp$0$ph$ph17;
    while(1) {
     $__cp$09$us$us = $__cp$0814$us;
     while(1) {
      $34 = ((($__cp$09$us$us)) + 4|0); //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $35 = HEAP32[$34>>2]|0; //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $36 = $35 & $14; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $37 = ($36|0)==($__phash$0$ph$ph18|0); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($37) {
       $__cp$09$us$us$lcssa = $__cp$09$us$us;
       break;
      }
      $38 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $39 = (($38) + ($36<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $40 = HEAP32[$39>>2]|0; //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $41 = ($40|0)==(0|0); //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($41) {
       $$lcssa = $39;$$lcssa6 = $36;$__cp$0$lcssa5 = $__cp$09$us$us;$__pp$0$ph$lcssa = $__pp$0$ph12$us;
       break L25;
      }
      $51 = ((($__cp$09$us$us)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $__np$0$us$us = $__cp$09$us$us;
      while(1) {
       $52 = HEAP32[$__np$0$us$us>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $53 = ($52|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($53) {
        $$lcssa143 = $52;$__np$0$us$us$lcssa = $__np$0$us$us;
        break;
       }
       $54 = ((($52)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
       $55 = HEAP32[$51>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $56 = HEAP32[$54>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $57 = ($55|0)==($56|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       if ($57) {
        $__np$0$us$us = $52;
       } else {
        $$lcssa143 = $52;$__np$0$us$us$lcssa = $__np$0$us$us;
        break;
       }
      }
      $42 = $$lcssa143; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__pp$0$ph12$us>>2] = $42; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $43 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $44 = (($43) + ($36<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $45 = HEAP32[$44>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $46 = HEAP32[$45>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__np$0$us$us$lcssa>>2] = $46; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $47 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $48 = (($47) + ($36<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $49 = HEAP32[$48>>2]|0; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$49>>2] = $__cp$09$us$us; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $__cp$0$us$us = HEAP32[$__cp$0$in13$us>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $50 = ($__cp$0$us$us|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($50) {
       label = 32;
       break L23;
      } else {
       $__cp$09$us$us = $__cp$0$us$us;
      }
     }
     $__cp$08$us = HEAP32[$__cp$09$us$us$lcssa>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $58 = ($__cp$08$us|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     if ($58) {
      label = 32;
      break L23;
     } else {
      $__cp$0$in13$us = $__cp$09$us$us$lcssa;$__cp$0814$us = $__cp$08$us;$__pp$0$ph12$us = $__cp$09$us$us$lcssa;
     }
    }
   } else {
    $__cp$0$in13 = $__cp$0$in1019;$__cp$0814 = $__cp$081120;$__pp$0$ph12 = $__pp$0$ph$ph17;
    while(1) {
     $__cp$09 = $__cp$0814;
     while(1) {
      $24 = ((($__cp$09)) + 4|0); //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $25 = HEAP32[$24>>2]|0; //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $26 = (($25>>>0) % ($__nbc>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $27 = ($26|0)==($__phash$0$ph$ph18|0); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($27) {
       $__cp$09$lcssa = $__cp$09;
       break;
      }
      $28 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $29 = (($28) + ($26<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $30 = HEAP32[$29>>2]|0; //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $31 = ($30|0)==(0|0); //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($31) {
       $$lcssa = $29;$$lcssa6 = $26;$__cp$0$lcssa5 = $__cp$09;$__pp$0$ph$lcssa = $__pp$0$ph12;
       break L25;
      }
      $32 = ((($__cp$09)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $__np$0 = $__cp$09;
      while(1) {
       $59 = HEAP32[$__np$0>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $60 = ($59|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($60) {
        $$lcssa133 = $59;$__np$0$lcssa = $__np$0;
        break;
       }
       $61 = ((($59)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
       $62 = HEAP32[$32>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $63 = HEAP32[$61>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $64 = ($62|0)==($63|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       if ($64) {
        $__np$0 = $59;
       } else {
        $$lcssa133 = $59;$__np$0$lcssa = $__np$0;
        break;
       }
      }
      $65 = $$lcssa133; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__pp$0$ph12>>2] = $65; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $66 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $67 = (($66) + ($26<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $68 = HEAP32[$67>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $69 = HEAP32[$68>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__np$0$lcssa>>2] = $69; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $70 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $71 = (($70) + ($26<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $72 = HEAP32[$71>>2]|0; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$72>>2] = $__cp$09; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $__cp$0 = HEAP32[$__cp$0$in13>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $73 = ($__cp$0|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($73) {
       label = 32;
       break L23;
      } else {
       $__cp$09 = $__cp$0;
      }
     }
     $__cp$08 = HEAP32[$__cp$09$lcssa>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $23 = ($__cp$08|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     if ($23) {
      label = 32;
      break L23;
     } else {
      $__cp$0$in13 = $__cp$09$lcssa;$__cp$0814 = $__cp$08;$__pp$0$ph12 = $__cp$09$lcssa;
     }
    }
   }
  } while(0);
  HEAP32[$$lcssa>>2] = $__pp$0$ph$lcssa; //@line 1989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $__cp$0811 = HEAP32[$__cp$0$lcssa5>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $33 = ($__cp$0811|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($33) {
   label = 32;
   break;
  } else {
   $__cp$0$in1019 = $__cp$0$lcssa5;$__cp$081120 = $__cp$0811;$__phash$0$ph$ph18 = $$lcssa6;$__pp$0$ph$ph17 = $__cp$0$lcssa5;
  }
 }
 if ((label|0) == 32) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
}
function __ZNSt3__113unordered_mapIyiNS_4hashIyEENS_8equal_toIyEENS_9allocatorINS_4pairIKyiEEEEEixERS7_($this,$__k) {
 $this = $this|0;
 $__k = $__k|0;
 var $$0 = 0, $$pn$i$i = 0, $$pn$us$i$i = 0, $$sroa$0$0$i$i$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $8 = 0, $9 = 0, $__nd$0$i$i = 0, $__nd$0$us$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = $__k; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = $1; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $4 = (($1) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = $4; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = HEAP32[$5>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $7 = Math_imul($3, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = $7 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = $8 ^ $7; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $10 = Math_imul($9, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $11 = $10 ^ -561034072; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $12 = Math_imul($6, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $13 = $12 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = $13 ^ $12; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $15 = Math_imul($14, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $16 = Math_imul($11, 1540483477)|0; //@line 3104 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $17 = $16 ^ $15; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $18 = $17 >>> 13; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $19 = $18 ^ $17; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $20 = Math_imul($19, 1540483477)|0; //@line 3118 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $21 = $20 >>> 15; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $22 = $21 ^ $20; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $23 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $24 = HEAP32[$23>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $25 = ($24|0)==(0); //@line 2018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L1: do {
  if (!($25)) {
   $26 = (($24) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $27 = $26 & $24; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $28 = ($27|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($28) {
    $29 = $22 & $26; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $33 = $29;
   } else {
    $30 = (($22>>>0) % ($24>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $33 = $30;
   }
   $31 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $32 = (($31) + ($33<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $34 = HEAP32[$32>>2]|0; //@line 2021 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $35 = ($34|0)==(0|0); //@line 2022 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($35)) {
    if ($28) {
     $$pn$us$i$i = $34;
     while(1) {
      $__nd$0$us$i$i = HEAP32[$$pn$us$i$i>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $36 = ($__nd$0$us$i$i|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($36) {
       break L1;
      }
      $37 = ((($__nd$0$us$i$i)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $38 = HEAP32[$37>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $39 = $38 & $26; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $40 = ($39|0)==($33|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($40)) {
       break L1;
      }
      $41 = ((($__nd$0$us$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $42 = $41; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $43 = $42; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $44 = HEAP32[$43>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $45 = (($42) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $46 = $45; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $47 = HEAP32[$46>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $48 = ($44|0)==($3|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $49 = ($47|0)==($6|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $50 = $48 & $49; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($50) {
       $$sroa$0$0$i$i$sink = $__nd$0$us$i$i;
       break;
      } else {
       $$pn$us$i$i = $__nd$0$us$i$i;
      }
     }
     $$0 = ((($$sroa$0$0$i$i$sink)) + 16|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
     STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
    } else {
     $$pn$i$i = $34;
     while(1) {
      $__nd$0$i$i = HEAP32[$$pn$i$i>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $51 = ($__nd$0$i$i|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($51) {
       break L1;
      }
      $52 = ((($__nd$0$i$i)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $53 = HEAP32[$52>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $54 = (($53>>>0) % ($24>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $55 = ($54|0)==($33|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($55)) {
       break L1;
      }
      $56 = ((($__nd$0$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $57 = $56; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $58 = $57; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $59 = HEAP32[$58>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $60 = (($57) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $61 = $60; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $62 = HEAP32[$61>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $63 = ($59|0)==($3|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $64 = ($62|0)==($6|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $65 = $63 & $64; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($65) {
       $$sroa$0$0$i$i$sink = $__nd$0$i$i;
       break;
      } else {
       $$pn$i$i = $__nd$0$i$i;
      }
     }
     $$0 = ((($$sroa$0$0$i$i$sink)) + 16|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
     STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
    }
   }
  }
 } while(0);
 $66 = (__Znwj(24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $67 = ((($66)) + 8|0); //@line 1340 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $68 = $67; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $69 = $68; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$69>>2] = $3; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $70 = (($68) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $71 = $70; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$71>>2] = $6; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $72 = ((($66)) + 16|0); //@line 1342 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 HEAP32[$72>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 invoke_viii(3994,($0|0),($this|0),($66|0)); //@line 1366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $73 = __THREW__; __THREW__ = 0;
 $74 = $73&1;
 if ($74) {
  $76 = ___cxa_find_matching_catch_2()|0;
  $77 = tempRet0;
  __ZdlPv($66); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($76|0);
  // unreachable;
 }
 $75 = HEAP32[$0>>2]|0; //@line 315 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $$sroa$0$0$i$i$sink = $75;
 $$0 = ((($$sroa$0$0$i$i$sink)) + 16|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeIyiEENS_22__unordered_map_hasherIyS2_NS_4hashIyEELb1EEENS_21__unordered_map_equalIyS2_NS_8equal_toIyEELb1EEENS_9allocatorIS2_EEE20__node_insert_uniqueEPNS_11__hash_nodeIS2_PvEE($agg$result,$this,$__nd) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $__nd = $__nd|0;
 var $$pn = 0, $$pn$us = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0, $73 = 0.0, $74 = 0.0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0.0, $86 = 0.0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__bc$0 = 0, $__chash$0 = 0, $__chash$1 = 0, $__inserted$0 = 0, $__ndptr$0 = 0, $__ndptr$0$us = 0, $__ndptr$1 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($__nd)) + 8|0); //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $1 = $0; //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $2 = $1; //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $3 = HEAP32[$2>>2]|0; //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $4 = (($1) + 4)|0; //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $5 = $4; //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $6 = HEAP32[$5>>2]|0; //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $7 = Math_imul($3, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = $7 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = $8 ^ $7; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $10 = Math_imul($9, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $11 = $10 ^ -561034072; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $12 = Math_imul($6, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $13 = $12 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = $13 ^ $12; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $15 = Math_imul($14, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $16 = Math_imul($11, 1540483477)|0; //@line 3104 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $17 = $16 ^ $15; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $18 = $17 >>> 13; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $19 = $18 ^ $17; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $20 = Math_imul($19, 1540483477)|0; //@line 3118 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $21 = $20 >>> 15; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $22 = $21 ^ $20; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $23 = ((($__nd)) + 4|0); //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$23>>2] = $22; //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $24 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $25 = HEAP32[$24>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $26 = ($25|0)==(0); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L1: do {
  if ($26) {
   $__chash$0 = 0;
  } else {
   $27 = (($25) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $28 = $27 & $25; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $29 = ($28|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($29) {
    $30 = $22 & $27; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $34 = $30;
   } else {
    $31 = (($22>>>0) % ($25>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $34 = $31;
   }
   $32 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $33 = (($32) + ($34<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $35 = HEAP32[$33>>2]|0; //@line 1603 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $36 = ($35|0)==(0|0); //@line 1604 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($36) {
    $__chash$0 = $34;
   } else {
    if ($29) {
     $$pn$us = $35;
     while(1) {
      $__ndptr$0$us = HEAP32[$$pn$us>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $37 = ($__ndptr$0$us|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($37) {
       $__chash$0 = $34;
       break L1;
      }
      $38 = ((($__ndptr$0$us)) + 4|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $39 = HEAP32[$38>>2]|0; //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $40 = $39 & $27; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $41 = ($40|0)==($34|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($41)) {
       $__chash$0 = $34;
       break L1;
      }
      $42 = ((($__ndptr$0$us)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $43 = $42; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $44 = $43; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $45 = HEAP32[$44>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $46 = (($43) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $47 = $46; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $48 = HEAP32[$47>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $49 = ($45|0)==($3|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $50 = ($48|0)==($6|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $51 = $49 & $50; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($51) {
       $__inserted$0 = 0;$__ndptr$1 = $__ndptr$0$us;
       break;
      } else {
       $$pn$us = $__ndptr$0$us;
      }
     }
     $120 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     HEAP32[$agg$result>>2] = $120; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     $121 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     HEAP8[$121>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    } else {
     $$pn = $35;
     while(1) {
      $__ndptr$0 = HEAP32[$$pn>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $52 = ($__ndptr$0|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($52) {
       $__chash$0 = $34;
       break L1;
      }
      $53 = ((($__ndptr$0)) + 4|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $54 = HEAP32[$53>>2]|0; //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $55 = (($54>>>0) % ($25>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $56 = ($55|0)==($34|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($56)) {
       $__chash$0 = $34;
       break L1;
      }
      $57 = ((($__ndptr$0)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $58 = $57; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $59 = $58; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $60 = HEAP32[$59>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $61 = (($58) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $62 = $61; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $63 = HEAP32[$62>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $64 = ($60|0)==($3|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $65 = ($63|0)==($6|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $66 = $64 & $65; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($66) {
       $__inserted$0 = 0;$__ndptr$1 = $__ndptr$0;
       break;
      } else {
       $$pn = $__ndptr$0;
      }
     }
     $120 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     HEAP32[$agg$result>>2] = $120; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     $121 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     HEAP8[$121>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    }
   }
  }
 } while(0);
 $67 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $68 = HEAP32[$67>>2]|0; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $69 = (($68) + 1)|0; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $70 = (+($69>>>0)); //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $71 = (+($25>>>0)); //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $72 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $73 = +HEAPF32[$72>>2]; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $74 = $71 * $73; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $75 = $70 > $74; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $or$cond = $26 | $75; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 do {
  if ($or$cond) {
   $76 = $25 << 1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $77 = ($25>>>0)>(2); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($77) {
    $78 = (($25) + -1)|0; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $79 = $78 & $25; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $80 = ($79|0)==(0); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $82 = $80;
   } else {
    $82 = 0;
   }
   $81 = $82&1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $83 = $81 | $76; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $84 = $83 ^ 1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $85 = $70 / $73; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $86 = (+Math_ceil((+$85))); //@line 750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/cmath"
   $87 = (~~(($86))>>>0); //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $88 = ($84>>>0)<($87>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $89 = $88 ? $87 : $84; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   __ZNSt3__112__hash_tableINS_17__hash_value_typeIyiEENS_22__unordered_map_hasherIyS2_NS_4hashIyEELb1EEENS_21__unordered_map_equalIyS2_NS_8equal_toIyEELb1EEENS_9allocatorIS2_EEE6rehashEj($this,$89); //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $90 = HEAP32[$24>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $91 = HEAP32[$23>>2]|0; //@line 1621 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $92 = (($90) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $93 = $92 & $90; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $94 = ($93|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($94) {
    $95 = $92 & $91; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__bc$0 = $90;$__chash$1 = $95;
    break;
   } else {
    $96 = (($91>>>0) % ($90>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__bc$0 = $90;$__chash$1 = $96;
    break;
   }
  } else {
   $__bc$0 = $25;$__chash$1 = $__chash$0;
  }
 } while(0);
 $97 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $98 = (($97) + ($__chash$1<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $99 = HEAP32[$98>>2]|0; //@line 1624 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $100 = ($99|0)==(0|0); //@line 1625 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($100) {
  $101 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $102 = HEAP32[$101>>2]|0; //@line 1628 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$__nd>>2] = $102; //@line 1628 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$101>>2] = $__nd; //@line 1629 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $103 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $104 = (($103) + ($__chash$1<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$104>>2] = $101; //@line 1631 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $105 = HEAP32[$__nd>>2]|0; //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $106 = ($105|0)==(0|0); //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if (!($106)) {
   $107 = ((($105)) + 4|0); //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $108 = HEAP32[$107>>2]|0; //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $109 = (($__bc$0) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $110 = $109 & $__bc$0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $111 = ($110|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($111) {
    $112 = $108 & $109; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $116 = $112;
   } else {
    $113 = (($108>>>0) % ($__bc$0>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $116 = $113;
   }
   $114 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $115 = (($114) + ($116<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$115>>2] = $__nd; //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 } else {
  $117 = HEAP32[$99>>2]|0; //@line 1637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$__nd>>2] = $117; //@line 1637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$99>>2] = $__nd; //@line 1638 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $118 = HEAP32[$67>>2]|0; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $119 = (($118) + 1)|0; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$67>>2] = $119; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__inserted$0 = 1;$__ndptr$1 = $__nd;
 $120 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$agg$result>>2] = $120; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $121 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 HEAP8[$121>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeIyiEENS_22__unordered_map_hasherIyS2_NS_4hashIyEELb1EEENS_21__unordered_map_equalIyS2_NS_8equal_toIyEELb1EEENS_9allocatorIS2_EEE6rehashEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__n|0)==(1); //@line 1937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($0) {
  $$0 = 2;
 } else {
  $1 = (($__n) + -1)|0; //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $2 = $1 & $__n; //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $3 = ($2|0)==(0); //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($3) {
   $$0 = $__n;
  } else {
   $4 = (__ZNSt3__112__next_primeEj($__n)|0); //@line 1940 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $$0 = $4;
  }
 }
 $5 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = HEAP32[$5>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $7 = ($$0>>>0)>($6>>>0); //@line 1942 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($7) {
  __ZNSt3__112__hash_tableINS_17__hash_value_typeIyiEENS_22__unordered_map_hasherIyS2_NS_4hashIyEELb1EEENS_21__unordered_map_equalIyS2_NS_8equal_toIyEELb1EEENS_9allocatorIS2_EEE8__rehashEj($this,$$0); //@line 1943 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $8 = ($$0>>>0)<($6>>>0); //@line 1944 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($8)) {
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $9 = ($6>>>0)>(2); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($9) {
  $10 = (($6) + -1)|0; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $11 = $10 & $6; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $12 = ($11|0)==(0); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $30 = $12;
 } else {
  $30 = 0;
 }
 $13 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = HEAP32[$13>>2]|0; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $15 = (+($14>>>0)); //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $16 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $17 = +HEAPF32[$16>>2]; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $18 = $15 / $17; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $19 = (+Math_ceil((+$18))); //@line 750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/cmath"
 $20 = (~~(($19))>>>0); //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($30) {
  $21 = (($20) + -1)|0; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $22 = (Math_clz32(($21|0))|0); //@line 785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $23 = (32 - ($22))|0; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $24 = 1 << $23; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $24;
 } else {
  $25 = (__ZNSt3__112__next_primeEj($20)|0); //@line 1950 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $25;
 }
 $26 = ($$0>>>0)<($27>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $28 = $26 ? $27 : $$0; //@line 1946 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $29 = ($28>>>0)<($6>>>0); //@line 1952 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($29)) {
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 __ZNSt3__112__hash_tableINS_17__hash_value_typeIyiEENS_22__unordered_map_hasherIyS2_NS_4hashIyEELb1EEENS_21__unordered_map_equalIyS2_NS_8equal_toIyEELb1EEENS_9allocatorIS2_EEE8__rehashEj($this,$28); //@line 1953 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeIyiEENS_22__unordered_map_hasherIyS2_NS_4hashIyEELb1EEENS_21__unordered_map_equalIyS2_NS_8equal_toIyEELb1EEENS_9allocatorIS2_EEE8__rehashEj($this,$__nbc) {
 $this = $this|0;
 $__nbc = $__nbc|0;
 var $$lcssa = 0, $$lcssa133 = 0, $$lcssa143 = 0, $$lcssa6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__cp$0 = 0, $__cp$0$in1019 = 0, $__cp$0$in13 = 0, $__cp$0$in13$us = 0, $__cp$0$lcssa5 = 0, $__cp$0$us$us = 0, $__cp$08 = 0, $__cp$08$us = 0, $__cp$0811 = 0, $__cp$081116 = 0, $__cp$081120 = 0, $__cp$0814 = 0, $__cp$0814$us = 0, $__cp$09 = 0, $__cp$09$lcssa = 0, $__cp$09$us$us = 0;
 var $__cp$09$us$us$lcssa = 0, $__i$021 = 0, $__np$0 = 0, $__np$0$lcssa = 0, $__np$0$us$us = 0, $__np$0$us$us$lcssa = 0, $__phash$0$ph$ph18 = 0, $__pp$0$ph$lcssa = 0, $__pp$0$ph$ph17 = 0, $__pp$0$ph12 = 0, $__pp$0$ph12$us = 0, $exitcond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = ($__nbc|0)==(0); //@line 1965 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($1) {
  $98 = HEAP32[$this>>2]|0; //@line 2822 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$this>>2] = 0; //@line 2823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $99 = ($98|0)==(0|0); //@line 2824 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  if (!($99)) {
   __ZdlPv($98); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  HEAP32[$0>>2] = 0; //@line 1967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $2 = $__nbc << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = (__Znwj($2)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $4 = HEAP32[$this>>2]|0; //@line 2822 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $3; //@line 2823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ($4|0)==(0|0); //@line 2824 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($5)) {
  __ZdlPv($4); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 HEAP32[$0>>2] = $__nbc; //@line 1967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__i$021 = 0;
 while(1) {
  $9 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $10 = (($9) + ($__i$021<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$10>>2] = 0; //@line 1971 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $11 = (($__i$021) + 1)|0; //@line 1970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $exitcond = ($11|0)==($__nbc|0); //@line 1970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($exitcond) {
   break;
  } else {
   $__i$021 = $11;
  }
 }
 $6 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $7 = HEAP32[$6>>2]|0; //@line 1973 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $8 = ($7|0)==(0|0); //@line 1974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($8) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $12 = ((($7)) + 4|0); //@line 1976 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $13 = HEAP32[$12>>2]|0; //@line 1976 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $14 = (($__nbc) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $15 = $14 & $__nbc; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $16 = ($15|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($16) {
  $17 = $13 & $14; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $21 = $17;
 } else {
  $18 = (($13>>>0) % ($__nbc>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $21 = $18;
 }
 $19 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $20 = (($19) + ($21<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$20>>2] = $6; //@line 1977 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__cp$081116 = HEAP32[$7>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $22 = ($__cp$081116|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($22) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 } else {
  $__cp$0$in1019 = $7;$__cp$081120 = $__cp$081116;$__phash$0$ph$ph18 = $21;$__pp$0$ph$ph17 = $7;
 }
 L23: while(1) {
  L25: do {
   if ($16) {
    $__cp$0$in13$us = $__cp$0$in1019;$__cp$0814$us = $__cp$081120;$__pp$0$ph12$us = $__pp$0$ph$ph17;
    while(1) {
     $__cp$09$us$us = $__cp$0814$us;
     while(1) {
      $34 = ((($__cp$09$us$us)) + 4|0); //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $35 = HEAP32[$34>>2]|0; //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $36 = $35 & $14; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $37 = ($36|0)==($__phash$0$ph$ph18|0); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($37) {
       $__cp$09$us$us$lcssa = $__cp$09$us$us;
       break;
      }
      $38 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $39 = (($38) + ($36<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $40 = HEAP32[$39>>2]|0; //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $41 = ($40|0)==(0|0); //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($41) {
       $$lcssa = $39;$$lcssa6 = $36;$__cp$0$lcssa5 = $__cp$09$us$us;$__pp$0$ph$lcssa = $__pp$0$ph12$us;
       break L25;
      }
      $51 = ((($__cp$09$us$us)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $__np$0$us$us = $__cp$09$us$us;
      while(1) {
       $52 = HEAP32[$__np$0$us$us>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $53 = ($52|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($53) {
        $$lcssa143 = $52;$__np$0$us$us$lcssa = $__np$0$us$us;
        break;
       }
       $54 = ((($52)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
       $55 = $51; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $56 = $55; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $57 = HEAP32[$56>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $58 = (($55) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $59 = $58; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $60 = HEAP32[$59>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $61 = $54; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $62 = $61; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $63 = HEAP32[$62>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $64 = (($61) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $65 = $64; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $66 = HEAP32[$65>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $67 = ($57|0)==($63|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $68 = ($60|0)==($66|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $69 = $67 & $68; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       if ($69) {
        $__np$0$us$us = $52;
       } else {
        $$lcssa143 = $52;$__np$0$us$us$lcssa = $__np$0$us$us;
        break;
       }
      }
      $42 = $$lcssa143; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__pp$0$ph12$us>>2] = $42; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $43 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $44 = (($43) + ($36<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $45 = HEAP32[$44>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $46 = HEAP32[$45>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__np$0$us$us$lcssa>>2] = $46; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $47 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $48 = (($47) + ($36<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $49 = HEAP32[$48>>2]|0; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$49>>2] = $__cp$09$us$us; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $__cp$0$us$us = HEAP32[$__cp$0$in13$us>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $50 = ($__cp$0$us$us|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($50) {
       label = 32;
       break L23;
      } else {
       $__cp$09$us$us = $__cp$0$us$us;
      }
     }
     $__cp$08$us = HEAP32[$__cp$09$us$us$lcssa>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $70 = ($__cp$08$us|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     if ($70) {
      label = 32;
      break L23;
     } else {
      $__cp$0$in13$us = $__cp$09$us$us$lcssa;$__cp$0814$us = $__cp$08$us;$__pp$0$ph12$us = $__cp$09$us$us$lcssa;
     }
    }
   } else {
    $__cp$0$in13 = $__cp$0$in1019;$__cp$0814 = $__cp$081120;$__pp$0$ph12 = $__pp$0$ph$ph17;
    while(1) {
     $__cp$09 = $__cp$0814;
     while(1) {
      $24 = ((($__cp$09)) + 4|0); //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $25 = HEAP32[$24>>2]|0; //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $26 = (($25>>>0) % ($__nbc>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $27 = ($26|0)==($__phash$0$ph$ph18|0); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($27) {
       $__cp$09$lcssa = $__cp$09;
       break;
      }
      $28 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $29 = (($28) + ($26<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $30 = HEAP32[$29>>2]|0; //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $31 = ($30|0)==(0|0); //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($31) {
       $$lcssa = $29;$$lcssa6 = $26;$__cp$0$lcssa5 = $__cp$09;$__pp$0$ph$lcssa = $__pp$0$ph12;
       break L25;
      }
      $32 = ((($__cp$09)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $__np$0 = $__cp$09;
      while(1) {
       $71 = HEAP32[$__np$0>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $72 = ($71|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($72) {
        $$lcssa133 = $71;$__np$0$lcssa = $__np$0;
        break;
       }
       $73 = ((($71)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
       $74 = $32; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $75 = $74; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $76 = HEAP32[$75>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $77 = (($74) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $78 = $77; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $79 = HEAP32[$78>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $80 = $73; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $81 = $80; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $82 = HEAP32[$81>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $83 = (($80) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $84 = $83; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $85 = HEAP32[$84>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $86 = ($76|0)==($82|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $87 = ($79|0)==($85|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $88 = $86 & $87; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       if ($88) {
        $__np$0 = $71;
       } else {
        $$lcssa133 = $71;$__np$0$lcssa = $__np$0;
        break;
       }
      }
      $89 = $$lcssa133; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__pp$0$ph12>>2] = $89; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $90 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $91 = (($90) + ($26<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $92 = HEAP32[$91>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $93 = HEAP32[$92>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__np$0$lcssa>>2] = $93; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $94 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $95 = (($94) + ($26<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $96 = HEAP32[$95>>2]|0; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$96>>2] = $__cp$09; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $__cp$0 = HEAP32[$__cp$0$in13>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $97 = ($__cp$0|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($97) {
       label = 32;
       break L23;
      } else {
       $__cp$09 = $__cp$0;
      }
     }
     $__cp$08 = HEAP32[$__cp$09$lcssa>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $23 = ($__cp$08|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     if ($23) {
      label = 32;
      break L23;
     } else {
      $__cp$0$in13 = $__cp$09$lcssa;$__cp$0814 = $__cp$08;$__pp$0$ph12 = $__cp$09$lcssa;
     }
    }
   }
  } while(0);
  HEAP32[$$lcssa>>2] = $__pp$0$ph$lcssa; //@line 1989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $__cp$0811 = HEAP32[$__cp$0$lcssa5>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $33 = ($__cp$0811|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($33) {
   label = 32;
   break;
  } else {
   $__cp$0$in1019 = $__cp$0$lcssa5;$__cp$081120 = $__cp$0811;$__phash$0$ph$ph18 = $$lcssa6;$__pp$0$ph$ph17 = $__cp$0$lcssa5;
  }
 }
 if ((label|0) == 32) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
}
function __ZN4love4font16BMFontRasterizer7acceptsEPNS_10filesystem8FileDataE($fontdef) {
 $fontdef = $fontdef|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$fontdef>>2]|0; //@line 338 "modules/font/BMFontRasterizer.cpp"
 $1 = ((($0)) + 16|0); //@line 338 "modules/font/BMFontRasterizer.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 338 "modules/font/BMFontRasterizer.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($fontdef)|0); //@line 338 "modules/font/BMFontRasterizer.cpp"
 $4 = HEAP32[$fontdef>>2]|0; //@line 342 "modules/font/BMFontRasterizer.cpp"
 $5 = ((($4)) + 20|0); //@line 342 "modules/font/BMFontRasterizer.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 342 "modules/font/BMFontRasterizer.cpp"
 $7 = (FUNCTION_TABLE_ii[$6 & 8191]($fontdef)|0); //@line 342 "modules/font/BMFontRasterizer.cpp"
 $8 = ($7>>>0)>(4); //@line 342 "modules/font/BMFontRasterizer.cpp"
 if (!($8)) {
  $11 = 0;
  return ($11|0); //@line 342 "modules/font/BMFontRasterizer.cpp"
 }
 $9 = (_memcmp($3,367355,4)|0); //@line 342 "modules/font/BMFontRasterizer.cpp"
 $10 = ($9|0)==(0); //@line 342 "modules/font/BMFontRasterizer.cpp"
 $11 = $10;
 return ($11|0); //@line 342 "modules/font/BMFontRasterizer.cpp"
}
function __ZN4love4font4Font21newTrueTypeRasterizerEiNS0_18TrueTypeRasterizer7HintingE($this,$size,$hinting) {
 $this = $this|0;
 $size = $size|0;
 $hinting = $hinting|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(8)|0); //@line 46 "modules/font/Font.cpp"
 __THREW__ = 0;
 invoke_vi(3431,($0|0)); //@line 37 "./common/Data.h"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  __ZdlPv($0); //@line 46 "modules/font/Font.cpp"
  $$0 = $22;$$01 = $21;
  ___resumeException($$01|0);
  // unreachable;
 }
 HEAP32[$0>>2] = (24068); //@line 36 "modules/font/Font.cpp"
 __ZN4love6Object6retainEv($0); //@line 95 "./common/Object.h"
 $3 = HEAP32[$0>>2]|0; //@line 47 "modules/font/Font.cpp"
 $4 = ((($3)) + 12|0); //@line 47 "modules/font/Font.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 47 "modules/font/Font.cpp"
 __THREW__ = 0;
 invoke_vi($5|0,($0|0)); //@line 47 "modules/font/Font.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  $8 = HEAP32[$this>>2]|0; //@line 49 "modules/font/Font.cpp"
  $9 = ((($8)) + 32|0); //@line 49 "modules/font/Font.cpp"
  $10 = HEAP32[$9>>2]|0; //@line 49 "modules/font/Font.cpp"
  __THREW__ = 0;
  $11 = (invoke_iiiii($10|0,($this|0),($0|0),($size|0),($hinting|0))|0); //@line 49 "modules/font/Font.cpp"
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   $14 = HEAP32[$0>>2]|0; //@line 106 "./common/Object.h"
   $15 = ((($14)) + 12|0); //@line 106 "./common/Object.h"
   $16 = HEAP32[$15>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($16|0,($0|0)); //@line 106 "./common/Object.h"
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if ($18) {
    $19 = ___cxa_find_matching_catch_3(0|0)|0;
    $20 = tempRet0;
    ___clang_call_terminate($19); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   } else {
    return ($11|0); //@line 50 "modules/font/Font.cpp"
   }
  }
 }
 $23 = ___cxa_find_matching_catch_2()|0;
 $24 = tempRet0;
 $25 = HEAP32[$0>>2]|0; //@line 106 "./common/Object.h"
 $26 = ((($25)) + 12|0); //@line 106 "./common/Object.h"
 $27 = HEAP32[$26>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($27|0,($0|0)); //@line 106 "./common/Object.h"
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if ($29) {
  $30 = ___cxa_find_matching_catch_3(0|0)|0;
  $31 = tempRet0;
  ___clang_call_terminate($30); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  $$0 = $24;$$01 = $23;
  ___resumeException($$01|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love4font15DefaultFontDataD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 36 "modules/font/Font.cpp"
 return; //@line 36 "modules/font/Font.cpp"
}
function __ZNK4love4font15DefaultFontData7getDataEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (367783|0); //@line 40 "modules/font/Font.cpp"
}
function __ZNK4love4font15DefaultFontData7getSizeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 65932; //@line 41 "modules/font/Font.cpp"
}
function __ZN4love4font4Font19newBMFontRasterizerEPNS_10filesystem8FileDataERKNSt3__16vectorIPNS_5image9ImageDataENS5_9allocatorIS9_EEEE($this,$fontdef,$images) {
 $this = $this|0;
 $fontdef = $fontdef|0;
 $images = $images|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(108)|0); //@line 54 "modules/font/Font.cpp"
 __THREW__ = 0;
 invoke_viii(3995,($0|0),($fontdef|0),($images|0)); //@line 54 "modules/font/Font.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 54 "modules/font/Font.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 54 "modules/font/Font.cpp"
 }
 return (0)|0;
}
function __ZN4love4font4Font18newImageRasterizerEPNS_5image9ImageDataERKNSt3__112basic_stringIcNS5_11char_traitsIcEENS5_9allocatorIcEEEEi($this,$data,$text,$extraspacing) {
 $this = $this|0;
 $data = $data|0;
 $text = $text|0;
 $extraspacing = $extraspacing|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$byval_copy = 0, $$byval_copy10 = 0, $$cast = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index4Z2D = 0, $eh$lpad$body$indexZ2D = 0, $glyphs = 0, $i = 0, $or$cond32 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i4 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i5 = 0, $temp$i = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy10 = sp + 48|0;
 $$byval_copy = sp + 44|0;
 $vararg_buffer = sp;
 $temp$i = sp + 40|0;
 $0 = sp + 36|0;
 $1 = sp + 32|0;
 $glyphs = sp + 20|0;
 $i = sp + 8|0;
 $2 = sp + 4|0;
 HEAP32[$glyphs>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($glyphs)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$3>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ((($glyphs)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$4>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = HEAP8[$text>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = $5 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ($6<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = ((($text)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = HEAP32[$8>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $5&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $10 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = $7 ? $11 : $9; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ((($glyphs)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = ($12|0)==(0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($14) {
  label = 4;
 } else {
  $15 = $12 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  __THREW__ = 0;
  $16 = (invoke_ii(3348,($15|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $17 = __THREW__; __THREW__ = 0;
  $18 = $17&1;
  if ($18) {
   label = 19;
  } else {
   $19 = $16; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $20 = (($16) + ($12<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $21 = $20; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   HEAP32[$glyphs>>2] = $19; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$3>>2] = $19; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$13>>2] = $21; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   label = 4;
  }
 }
 L4: do {
  if ((label|0) == 4) {
   $22 = ((($text)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $23 = HEAP32[$22>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $24 = ((($text)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $25 = $7 ? $24 : $23; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $26 = $25; //@line 64 "modules/font/Font.cpp"
   $27 = (($25) + ($12)|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $28 = $27; //@line 64 "modules/font/Font.cpp"
   HEAP32[$i>>2] = $26; //@line 277 "./libraries/utf8/utf8/checked.h"
   $29 = ((($i)) + 4|0); //@line 277 "./libraries/utf8/utf8/checked.h"
   HEAP32[$29>>2] = $26; //@line 277 "./libraries/utf8/utf8/checked.h"
   $30 = ((($i)) + 8|0); //@line 277 "./libraries/utf8/utf8/checked.h"
   HEAP32[$30>>2] = $28; //@line 277 "./libraries/utf8/utf8/checked.h"
   $31 = ($12|0)<(0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   L6: do {
    if ($31) {
     $32 = (___cxa_allocate_exception(8)|0); //@line 280 "./libraries/utf8/utf8/checked.h"
     __THREW__ = 0;
     invoke_vii(3996,($32|0),(434517|0)); //@line 133 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/stdexcept"
     $33 = __THREW__; __THREW__ = 0;
     $34 = $33&1;
     if ($34) {
      $36 = ___cxa_find_matching_catch_3(1744|0)|0;
      $37 = tempRet0;
      ___cxa_free_exception(($32|0)); //@line 280 "./libraries/utf8/utf8/checked.h"
      $eh$lpad$body$index4Z2D = $37;$eh$lpad$body$indexZ2D = $36;
      break;
     } else {
      HEAP32[$32>>2] = (157912); //@line 133 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/stdexcept"
      __THREW__ = 0;
      invoke_viii(3350,($32|0),(9120|0),(3328|0)); //@line 280 "./libraries/utf8/utf8/checked.h"
      $35 = __THREW__; __THREW__ = 0;
      label = 21;
      break;
     }
    } else {
     $38 = ((($i)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $39 = ((($i)) + 8|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $40 = ($12|0)==(0); //@line 297 "./libraries/utf8/utf8/checked.h"
     L12: do {
      if (!($40)) {
       $47 = $28;$50 = $26;
       while(1) {
        HEAP32[$1>>2] = $47; //@line 307 "./libraries/utf8/utf8/checked.h"
        __THREW__ = 0;
        ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0; //@line 307 "./libraries/utf8/utf8/checked.h"
        (invoke_iii(3997,($i|0),($$byval_copy|0))|0); //@line 307 "./libraries/utf8/utf8/checked.h"
        $48 = __THREW__; __THREW__ = 0;
        $49 = $48&1;
        if ($49) {
         label = 20;
         break;
        }
        HEAP32[$temp$i>>2] = $50; //@line 286 "./libraries/utf8/utf8/checked.h"
        HEAP32[$0>>2] = $47; //@line 287 "./libraries/utf8/utf8/checked.h"
        __THREW__ = 0;
        ;HEAP32[$$byval_copy10>>2]=HEAP32[$0>>2]|0; //@line 287 "./libraries/utf8/utf8/checked.h"
        $51 = (invoke_iii(3997,($temp$i|0),($$byval_copy10|0))|0); //@line 287 "./libraries/utf8/utf8/checked.h"
        $52 = __THREW__; __THREW__ = 0;
        $53 = $52&1;
        if ($53) {
         label = 20;
         break;
        }
        HEAP32[$2>>2] = $51; //@line 68 "modules/font/Font.cpp"
        $54 = HEAP32[$3>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $55 = HEAP32[$13>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $56 = ($54>>>0)<($55>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($56) {
         HEAP32[$54>>2] = $51; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $57 = ((($54)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$3>>2] = $57; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        } else {
         __THREW__ = 0;
         invoke_vii(3597,($glyphs|0),($2|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $58 = __THREW__; __THREW__ = 0;
         $59 = $58&1;
         if ($59) {
          label = 20;
          break;
         }
        }
        $60 = HEAP32[$38>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        $61 = ($60|0)==($25|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        $62 = HEAP32[$39>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        $63 = ($62|0)==($27|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        $or$cond32 = $61 & $63; //@line 291 "./libraries/utf8/utf8/checked.h"
        if (!($or$cond32)) {
         break;
        }
        $64 = $62; //@line 291 "./libraries/utf8/utf8/checked.h"
        $$pre = HEAP32[$i>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        $65 = $$pre; //@line 291 "./libraries/utf8/utf8/checked.h"
        $66 = ($$pre|0)==($27|0); //@line 297 "./libraries/utf8/utf8/checked.h"
        if ($66) {
         break L12;
        } else {
         $47 = $64;$50 = $65;
        }
       }
       if ((label|0) == 20) {
        $69 = ___cxa_find_matching_catch_3(1744|0)|0;
        $70 = tempRet0;
        $eh$lpad$body$index4Z2D = $70;$eh$lpad$body$indexZ2D = $69;
        break L6;
       }
       $41 = (___cxa_allocate_exception(8)|0); //@line 292 "./libraries/utf8/utf8/checked.h"
       __THREW__ = 0;
       invoke_vii(3996,($41|0),(434549|0)); //@line 292 "./libraries/utf8/utf8/checked.h"
       $42 = __THREW__; __THREW__ = 0;
       $43 = $42&1;
       if ($43) {
        $45 = ___cxa_find_matching_catch_3(1744|0)|0;
        $46 = tempRet0;
        ___cxa_free_exception(($41|0)); //@line 292 "./libraries/utf8/utf8/checked.h"
        $eh$lpad$body$index4Z2D = $46;$eh$lpad$body$indexZ2D = $45;
        break L6;
       } else {
        __THREW__ = 0;
        invoke_viii(3350,($41|0),(9072|0),(3328|0)); //@line 292 "./libraries/utf8/utf8/checked.h"
        $44 = __THREW__; __THREW__ = 0;
        label = 21;
        break L6;
       }
      }
     } while(0);
     $86 = HEAP32[$this>>2]|0; //@line 75 "modules/font/Font.cpp"
     $87 = ((($86)) + 44|0); //@line 75 "modules/font/Font.cpp"
     $88 = HEAP32[$87>>2]|0; //@line 75 "modules/font/Font.cpp"
     $89 = HEAP32[$glyphs>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $90 = HEAP32[$3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $$cast = $89; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $91 = (($90) - ($$cast))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $92 = $91 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     __THREW__ = 0;
     $93 = (invoke_iiiiii($88|0,($this|0),($data|0),($89|0),($92|0),($extraspacing|0))|0); //@line 75 "modules/font/Font.cpp"
     $94 = __THREW__; __THREW__ = 0;
     $95 = $94&1;
     if ($95) {
      label = 19;
      break L4;
     }
     $100 = HEAP32[$glyphs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $101 = ($100|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $102 = $100; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($101) {
      STACKTOP = sp;return ($93|0); //@line 76 "modules/font/Font.cpp"
     }
     $103 = HEAP32[$3>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $104 = ($103|0)==($100|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($104)) {
      $scevgep$i$i$i4 = ((($103)) + -4|0);
      $105 = $scevgep$i$i$i4;
      $106 = (($105) - ($102))|0;
      $107 = $106 >>> 2;
      $108 = $107 ^ -1;
      $scevgep4$i$i$i5 = (($103) + ($108<<2)|0);
      HEAP32[$3>>2] = $scevgep4$i$i$i5; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __ZdlPv($100); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     STACKTOP = sp;return ($93|0); //@line 76 "modules/font/Font.cpp"
    }
   } while(0);
   if ((label|0) == 21) {
    $71 = ___cxa_find_matching_catch_3(1744|0)|0;
    $72 = tempRet0;
    $eh$lpad$body$index4Z2D = $72;$eh$lpad$body$indexZ2D = $71;
   }
   $73 = (_llvm_eh_typeid_for((1744|0))|0); //@line 69 "modules/font/Font.cpp"
   $74 = ($eh$lpad$body$index4Z2D|0)==($73|0); //@line 69 "modules/font/Font.cpp"
   if ($74) {
    $75 = (___cxa_begin_catch(($eh$lpad$body$indexZ2D|0))|0); //@line 69 "modules/font/Font.cpp"
    $76 = (___cxa_allocate_exception(16)|0); //@line 72 "modules/font/Font.cpp"
    $77 = HEAP32[$75>>2]|0; //@line 72 "modules/font/Font.cpp"
    $78 = ((($77)) + 8|0); //@line 72 "modules/font/Font.cpp"
    $79 = HEAP32[$78>>2]|0; //@line 72 "modules/font/Font.cpp"
    $80 = (FUNCTION_TABLE_ii[$79 & 8191]($75)|0); //@line 72 "modules/font/Font.cpp"
    __THREW__ = 0;
    HEAP32[$vararg_buffer>>2] = $80; //@line 72 "modules/font/Font.cpp"
    invoke_viii(3349,($76|0),(434605|0),($vararg_buffer|0)); //@line 72 "modules/font/Font.cpp"
    $81 = __THREW__; __THREW__ = 0;
    $82 = $81&1;
    if ($82) {
     $96 = ___cxa_find_matching_catch_2()|0;
     $97 = tempRet0;
     ___cxa_free_exception(($76|0)); //@line 72 "modules/font/Font.cpp"
     $$0 = $96;$$01 = $97;
    } else {
     __THREW__ = 0;
     invoke_viii(3350,($76|0),(96|0),(344|0)); //@line 72 "modules/font/Font.cpp"
     $83 = __THREW__; __THREW__ = 0;
     $84 = ___cxa_find_matching_catch_2()|0;
     $85 = tempRet0;
     $$0 = $84;$$01 = $85;
    }
    __THREW__ = 0;
    invoke_v(3378); //@line 73 "modules/font/Font.cpp"
    $98 = __THREW__; __THREW__ = 0;
    $99 = $98&1;
    if ($99) {
     $118 = ___cxa_find_matching_catch_3(0|0)|0;
     $119 = tempRet0;
     ___clang_call_terminate($118); //@line 73 "modules/font/Font.cpp"
     // unreachable; //@line 73 "modules/font/Font.cpp"
    } else {
     $$1 = $$0;$$12 = $$01;
    }
   } else {
    $$1 = $eh$lpad$body$indexZ2D;$$12 = $eh$lpad$body$index4Z2D;
   }
  }
 } while(0);
 if ((label|0) == 19) {
  $67 = ___cxa_find_matching_catch_2()|0;
  $68 = tempRet0;
  $$1 = $67;$$12 = $68;
 }
 $109 = HEAP32[$glyphs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $110 = ($109|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $111 = $109; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($110) {
  ___resumeException($$1|0);
  // unreachable;
 }
 $112 = HEAP32[$3>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $113 = ($112|0)==($109|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($113)) {
  $scevgep$i$i$i = ((($112)) + -4|0);
  $114 = $scevgep$i$i$i;
  $115 = (($114) - ($111))|0;
  $116 = $115 >>> 2;
  $117 = $116 ^ -1;
  $scevgep4$i$i$i = (($112) + ($117<<2)|0);
  HEAP32[$3>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($109); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZN4utf84nextINSt3__111__wrap_iterIPKcEEEEjRT_S6_($it,$end) {
 $it = $it|0;
 $end = $end|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 8|0;
 $cp = sp + 4|0;
 $0 = sp;
 HEAP32[$cp>>2] = 0; //@line 139 "./libraries/utf8/utf8/checked.h"
 $1 = HEAP32[$end>>2]|0; //@line 140 "./libraries/utf8/utf8/checked.h"
 HEAP32[$0>>2] = $1; //@line 140 "./libraries/utf8/utf8/checked.h"
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0; //@line 140 "./libraries/utf8/utf8/checked.h"
 $2 = (__ZN4utf88internal13validate_nextINSt3__111__wrap_iterIPKcEEEENS0_9utf_errorERT_S8_Rj($it,$$byval_copy,$cp)|0); //@line 140 "./libraries/utf8/utf8/checked.h"
 switch ($2|0) {
 case 5:  {
  $8 = (___cxa_allocate_exception(8)|0); //@line 151 "./libraries/utf8/utf8/checked.h"
  $9 = HEAP32[$cp>>2]|0; //@line 151 "./libraries/utf8/utf8/checked.h"
  HEAP32[$8>>2] = (24140); //@line 44 "./libraries/utf8/utf8/checked.h"
  $10 = ((($8)) + 4|0); //@line 44 "./libraries/utf8/utf8/checked.h"
  HEAP32[$10>>2] = $9; //@line 44 "./libraries/utf8/utf8/checked.h"
  ___cxa_throw(($8|0),(1792|0),(1696|0)); //@line 151 "./libraries/utf8/utf8/checked.h"
  // unreachable; //@line 151 "./libraries/utf8/utf8/checked.h"
  break;
 }
 case 1:  {
  $3 = (___cxa_allocate_exception(4)|0); //@line 145 "./libraries/utf8/utf8/checked.h"
  HEAP32[$3>>2] = (24100); //@line 65 "./libraries/utf8/utf8/checked.h"
  ___cxa_throw(($3|0),(1760|0),(1696|0)); //@line 145 "./libraries/utf8/utf8/checked.h"
  // unreachable; //@line 145 "./libraries/utf8/utf8/checked.h"
  break;
 }
 case 4: case 3: case 2:  {
  $4 = (___cxa_allocate_exception(8)|0); //@line 149 "./libraries/utf8/utf8/checked.h"
  $5 = HEAP32[$it>>2]|0; //@line 1163 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $6 = HEAP8[$5>>0]|0; //@line 149 "./libraries/utf8/utf8/checked.h"
  HEAP32[$4>>2] = (24120); //@line 52 "./libraries/utf8/utf8/checked.h"
  $7 = ((($4)) + 4|0); //@line 52 "./libraries/utf8/utf8/checked.h"
  HEAP8[$7>>0] = $6; //@line 52 "./libraries/utf8/utf8/checked.h"
  ___cxa_throw(($4|0),(1776|0),(1696|0)); //@line 149 "./libraries/utf8/utf8/checked.h"
  // unreachable; //@line 149 "./libraries/utf8/utf8/checked.h"
  break;
 }
 default: {
  $11 = HEAP32[$cp>>2]|0; //@line 153 "./libraries/utf8/utf8/checked.h"
  STACKTOP = sp;return ($11|0); //@line 153 "./libraries/utf8/utf8/checked.h"
 }
 }
 return (0)|0;
}
function __ZN4utf88internal13validate_nextINSt3__111__wrap_iterIPKcEEEENS0_9utf_errorERT_S8_Rj($it,$end,$code_point) {
 $it = $it|0;
 $end = $end|0;
 $code_point = $code_point|0;
 var $$0 = 0, $$0$i16 = 0, $$mask$i = 0, $$mask$i$i$i = 0, $$mask$i$i$i2 = 0, $$mask$i$i$i4 = 0, $$mask$i$i13$i = 0, $$mask$i$i6$i = 0, $$mask$i$i9$i = 0, $$mask1$i = 0, $$mask2$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $cp$4 = 0, $err$1 = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$it>>2]|0; //@line 227 "./libraries/utf8/utf8/core.h"
 $1 = $0; //@line 1163 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $2 = HEAP8[$1>>0]|0; //@line 158 "./libraries/utf8/utf8/core.h"
 $3 = $2&255; //@line 103 "./libraries/utf8/utf8/core.h"
 $4 = ($2<<24>>24)>(-1); //@line 103 "./libraries/utf8/utf8/core.h"
 do {
  if ($4) {
   $8 = HEAP32[$end>>2]|0; //@line 240 "./libraries/utf8/utf8/core.h"
   $9 = ($1|0)==($8|0); //@line 1295 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($9) {
    $err$1 = 1;
   } else {
    $$0$i16 = 1;$78 = $1;$cp$4 = $3;
    label = 24;
   }
  } else {
   $$mask$i = $3 & 224; //@line 105 "./libraries/utf8/utf8/core.h"
   $5 = ($$mask$i|0)==(192); //@line 105 "./libraries/utf8/utf8/core.h"
   if ($5) {
    $10 = HEAP32[$end>>2]|0; //@line 243 "./libraries/utf8/utf8/core.h"
    $11 = ($1|0)==($10|0); //@line 1295 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($11) {
     $err$1 = 1;
     break;
    }
    $12 = ((($1)) + 1|0); //@line 1163 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    HEAP32[$it>>2] = $12; //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $13 = ($12|0)==($10|0); //@line 1295 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($13) {
     $err$1 = 1;
     break;
    }
    $14 = HEAP8[$12>>0]|0; //@line 173 "./libraries/utf8/utf8/core.h"
    $$mask$i$i$i4 = $14 & -64; //@line 71 "./libraries/utf8/utf8/core.h"
    $15 = ($$mask$i$i$i4<<24>>24)==(-128); //@line 71 "./libraries/utf8/utf8/core.h"
    if (!($15)) {
     $err$1 = 3;
     break;
    }
    $16 = $3 << 6; //@line 173 "./libraries/utf8/utf8/core.h"
    $17 = $16 & 1984; //@line 173 "./libraries/utf8/utf8/core.h"
    $18 = $14&255; //@line 173 "./libraries/utf8/utf8/core.h"
    $19 = $18 & 63; //@line 173 "./libraries/utf8/utf8/core.h"
    $20 = $19 | $17; //@line 173 "./libraries/utf8/utf8/core.h"
    $$0$i16 = 2;$78 = $12;$cp$4 = $20;
    label = 24;
    break;
   }
   $$mask1$i = $3 & 240; //@line 107 "./libraries/utf8/utf8/core.h"
   $6 = ($$mask1$i|0)==(224); //@line 107 "./libraries/utf8/utf8/core.h"
   if ($6) {
    $21 = HEAP32[$end>>2]|0; //@line 246 "./libraries/utf8/utf8/core.h"
    $22 = ($1|0)==($21|0); //@line 1295 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($22) {
     $err$1 = 1;
     break;
    }
    $23 = ((($1)) + 1|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    HEAP32[$it>>2] = $23; //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $24 = ($23|0)==($21|0); //@line 1295 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($24) {
     $err$1 = 1;
     break;
    }
    $25 = HEAP8[$23>>0]|0; //@line 188 "./libraries/utf8/utf8/core.h"
    $$mask$i$i6$i = $25 & -64; //@line 71 "./libraries/utf8/utf8/core.h"
    $26 = ($$mask$i$i6$i<<24>>24)==(-128); //@line 71 "./libraries/utf8/utf8/core.h"
    if (!($26)) {
     $err$1 = 3;
     break;
    }
    $27 = $3 << 12; //@line 188 "./libraries/utf8/utf8/core.h"
    $28 = $27 & 61440; //@line 188 "./libraries/utf8/utf8/core.h"
    $29 = $25&255; //@line 188 "./libraries/utf8/utf8/core.h"
    $30 = $29 << 6; //@line 188 "./libraries/utf8/utf8/core.h"
    $31 = $30 & 4032; //@line 188 "./libraries/utf8/utf8/core.h"
    $32 = $31 | $28; //@line 192 "./libraries/utf8/utf8/core.h"
    $33 = ((($1)) + 2|0); //@line 1163 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    HEAP32[$it>>2] = $33; //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $34 = ($33|0)==($21|0); //@line 1295 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($34) {
     $err$1 = 1;
     break;
    }
    $35 = HEAP8[$33>>0]|0; //@line 192 "./libraries/utf8/utf8/core.h"
    $$mask$i$i$i2 = $35 & -64; //@line 71 "./libraries/utf8/utf8/core.h"
    $36 = ($$mask$i$i$i2<<24>>24)==(-128); //@line 71 "./libraries/utf8/utf8/core.h"
    if (!($36)) {
     $err$1 = 3;
     break;
    }
    $37 = $35&255; //@line 192 "./libraries/utf8/utf8/core.h"
    $38 = $37 & 63; //@line 192 "./libraries/utf8/utf8/core.h"
    $39 = $32 | $38; //@line 192 "./libraries/utf8/utf8/core.h"
    $$0$i16 = 3;$78 = $33;$cp$4 = $39;
    label = 24;
    break;
   }
   $$mask2$i = $3 & 248; //@line 109 "./libraries/utf8/utf8/core.h"
   $7 = ($$mask2$i|0)==(240); //@line 109 "./libraries/utf8/utf8/core.h"
   if (!($7)) {
    $$0 = 2;
    return ($$0|0); //@line 272 "./libraries/utf8/utf8/core.h"
   }
   $40 = HEAP32[$end>>2]|0; //@line 249 "./libraries/utf8/utf8/core.h"
   $41 = ($1|0)==($40|0); //@line 1295 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($41) {
    $err$1 = 1;
   } else {
    $42 = ((($1)) + 1|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    HEAP32[$it>>2] = $42; //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $43 = ($42|0)==($40|0); //@line 1295 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($43) {
     $err$1 = 1;
    } else {
     $44 = HEAP8[$42>>0]|0; //@line 207 "./libraries/utf8/utf8/core.h"
     $$mask$i$i13$i = $44 & -64; //@line 71 "./libraries/utf8/utf8/core.h"
     $45 = ($$mask$i$i13$i<<24>>24)==(-128); //@line 71 "./libraries/utf8/utf8/core.h"
     if ($45) {
      $46 = $3 << 18; //@line 207 "./libraries/utf8/utf8/core.h"
      $47 = $46 & 1835008; //@line 207 "./libraries/utf8/utf8/core.h"
      $48 = $44&255; //@line 207 "./libraries/utf8/utf8/core.h"
      $49 = $48 << 12; //@line 207 "./libraries/utf8/utf8/core.h"
      $50 = $49 & 258048; //@line 207 "./libraries/utf8/utf8/core.h"
      $51 = $50 | $47; //@line 211 "./libraries/utf8/utf8/core.h"
      $52 = ((($1)) + 2|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      HEAP32[$it>>2] = $52; //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $53 = ($52|0)==($40|0); //@line 1295 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($53) {
       $err$1 = 1;
      } else {
       $54 = HEAP8[$52>>0]|0; //@line 211 "./libraries/utf8/utf8/core.h"
       $$mask$i$i9$i = $54 & -64; //@line 71 "./libraries/utf8/utf8/core.h"
       $55 = ($$mask$i$i9$i<<24>>24)==(-128); //@line 71 "./libraries/utf8/utf8/core.h"
       if ($55) {
        $56 = $54&255; //@line 211 "./libraries/utf8/utf8/core.h"
        $57 = $56 << 6; //@line 211 "./libraries/utf8/utf8/core.h"
        $58 = $57 & 4032; //@line 211 "./libraries/utf8/utf8/core.h"
        $59 = $51 | $58; //@line 215 "./libraries/utf8/utf8/core.h"
        $60 = ((($1)) + 3|0); //@line 1163 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        HEAP32[$it>>2] = $60; //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        $61 = ($60|0)==($40|0); //@line 1295 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        if ($61) {
         $err$1 = 1;
        } else {
         $62 = HEAP8[$60>>0]|0; //@line 215 "./libraries/utf8/utf8/core.h"
         $$mask$i$i$i = $62 & -64; //@line 71 "./libraries/utf8/utf8/core.h"
         $63 = ($$mask$i$i$i<<24>>24)==(-128); //@line 71 "./libraries/utf8/utf8/core.h"
         if ($63) {
          $64 = $62&255; //@line 215 "./libraries/utf8/utf8/core.h"
          $65 = $64 & 63; //@line 215 "./libraries/utf8/utf8/core.h"
          $66 = $59 | $65; //@line 215 "./libraries/utf8/utf8/core.h"
          $$0$i16 = 4;$78 = $60;$cp$4 = $66;
          label = 24;
         } else {
          $err$1 = 3;
         }
        }
       } else {
        $err$1 = 3;
       }
      }
     } else {
      $err$1 = 3;
     }
    }
   }
  }
 } while(0);
 L28: do {
  if ((label|0) == 24) {
   $67 = ($cp$4>>>0)<(1114112); //@line 95 "./libraries/utf8/utf8/core.h"
   $68 = $cp$4 & -2048; //@line 89 "./libraries/utf8/utf8/core.h"
   $69 = ($68|0)!=(55296); //@line 95 "./libraries/utf8/utf8/core.h"
   $70 = $67 & $69; //@line 95 "./libraries/utf8/utf8/core.h"
   if ($70) {
    $71 = ($cp$4>>>0)<(128); //@line 118 "./libraries/utf8/utf8/core.h"
    do {
     if ($71) {
      $72 = ($$0$i16|0)==(1); //@line 119 "./libraries/utf8/utf8/core.h"
      if (!($72)) {
       $err$1 = 4;
       break L28;
      }
     } else {
      $73 = ($cp$4>>>0)<(2048); //@line 122 "./libraries/utf8/utf8/core.h"
      if ($73) {
       $74 = ($$0$i16|0)==(2); //@line 123 "./libraries/utf8/utf8/core.h"
       if ($74) {
        break;
       } else {
        $err$1 = 4;
        break L28;
       }
      } else {
       $75 = ($cp$4>>>0)<(65536); //@line 126 "./libraries/utf8/utf8/core.h"
       $76 = ($$0$i16|0)!=(3); //@line 127 "./libraries/utf8/utf8/core.h"
       $or$cond$i = $76 & $75; //@line 126 "./libraries/utf8/utf8/core.h"
       if ($or$cond$i) {
        $err$1 = 4;
        break L28;
       } else {
        break;
       }
      }
     }
    } while(0);
    HEAP32[$code_point>>2] = $cp$4; //@line 258 "./libraries/utf8/utf8/core.h"
    $77 = ((($78)) + 1|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    HEAP32[$it>>2] = $77; //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $$0 = 0;
    return ($$0|0); //@line 272 "./libraries/utf8/utf8/core.h"
   } else {
    $err$1 = 5;
   }
  }
 } while(0);
 HEAP32[$it>>2] = $0; //@line 270 "./libraries/utf8/utf8/core.h"
 $$0 = $err$1;
 return ($$0|0); //@line 272 "./libraries/utf8/utf8/core.h"
}
function __ZN4utf815not_enough_roomD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 65 "./libraries/utf8/utf8/checked.h"
 return; //@line 65 "./libraries/utf8/utf8/checked.h"
}
function __ZNK4utf815not_enough_room4whatEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (433806|0); //@line 67 "./libraries/utf8/utf8/checked.h"
}
function __ZN4utf812invalid_utf8D0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 49 "./libraries/utf8/utf8/checked.h"
 return; //@line 49 "./libraries/utf8/utf8/checked.h"
}
function __ZNK4utf812invalid_utf84whatEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (433845|0); //@line 53 "./libraries/utf8/utf8/checked.h"
}
function __ZN4utf818invalid_code_pointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 41 "./libraries/utf8/utf8/checked.h"
 return; //@line 41 "./libraries/utf8/utf8/checked.h"
}
function __ZNK4utf818invalid_code_point4whatEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (433887|0); //@line 45 "./libraries/utf8/utf8/checked.h"
}
function __ZN4love4font4Font18newImageRasterizerEPNS_5image9ImageDataEPjii($this,$data,$glyphs,$numglyphs,$extraspacing) {
 $this = $this|0;
 $data = $data|0;
 $glyphs = $glyphs|0;
 $numglyphs = $numglyphs|0;
 $extraspacing = $extraspacing|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(56)|0); //@line 80 "modules/font/Font.cpp"
 __THREW__ = 0;
 invoke_viiiii(3998,($0|0),($data|0),($glyphs|0),($numglyphs|0),($extraspacing|0)); //@line 80 "modules/font/Font.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 80 "modules/font/Font.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 80 "modules/font/Font.cpp"
 }
 return (0)|0;
}
function __ZN4love4font4Font12newGlyphDataEPNS0_10RasterizerERKNSt3__112basic_stringIcNS4_11char_traitsIcEENS4_9allocatorIcEEEE($this,$r,$text) {
 $this = $this|0;
 $r = $r|0;
 $text = $text|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 16|0;
 $vararg_buffer = sp + 8|0;
 $0 = sp + 12|0;
 $1 = sp;
 $2 = HEAP8[$text>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = $2 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ($3<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($text)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = HEAP32[$5>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($text)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $4 ? $7 : $6; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = ((($text)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = HEAP32[$9>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $2&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = $11 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = $4 ? $12 : $10; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = (($8) + ($13)|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $15 = $14; //@line 89 "modules/font/Font.cpp"
 HEAP32[$1>>2] = $8; //@line 89 "modules/font/Font.cpp"
 HEAP32[$0>>2] = $15; //@line 159 "./libraries/utf8/utf8/checked.h"
 __THREW__ = 0;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0; //@line 159 "./libraries/utf8/utf8/checked.h"
 $16 = (invoke_iii(3997,($1|0),($$byval_copy|0))|0); //@line 159 "./libraries/utf8/utf8/checked.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if (!($18)) {
  $19 = HEAP32[$r>>2]|0; //@line 96 "modules/font/Font.cpp"
  $20 = ((($19)) + 36|0); //@line 96 "modules/font/Font.cpp"
  $21 = HEAP32[$20>>2]|0; //@line 96 "modules/font/Font.cpp"
  $22 = (FUNCTION_TABLE_iii[$21 & 8191]($r,$16)|0); //@line 96 "modules/font/Font.cpp"
  STACKTOP = sp;return ($22|0); //@line 96 "modules/font/Font.cpp"
 }
 $23 = ___cxa_find_matching_catch_3(1744|0)|0;
 $24 = tempRet0;
 $25 = (_llvm_eh_typeid_for((1744|0))|0); //@line 90 "modules/font/Font.cpp"
 $26 = ($24|0)==($25|0); //@line 90 "modules/font/Font.cpp"
 if (!($26)) {
  $$1 = $24;$$12 = $23;
  ___resumeException($$12|0);
  // unreachable;
 }
 $27 = (___cxa_begin_catch(($23|0))|0); //@line 90 "modules/font/Font.cpp"
 $28 = (___cxa_allocate_exception(16)|0); //@line 93 "modules/font/Font.cpp"
 $29 = HEAP32[$27>>2]|0; //@line 93 "modules/font/Font.cpp"
 $30 = ((($29)) + 8|0); //@line 93 "modules/font/Font.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 93 "modules/font/Font.cpp"
 $32 = (FUNCTION_TABLE_ii[$31 & 8191]($27)|0); //@line 93 "modules/font/Font.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $32; //@line 93 "modules/font/Font.cpp"
 invoke_viii(3349,($28|0),(434605|0),($vararg_buffer|0)); //@line 93 "modules/font/Font.cpp"
 $33 = __THREW__; __THREW__ = 0;
 $34 = $33&1;
 if ($34) {
  $38 = ___cxa_find_matching_catch_2()|0;
  $39 = tempRet0;
  ___cxa_free_exception(($28|0)); //@line 93 "modules/font/Font.cpp"
  $$0 = $39;$$01 = $38;
 } else {
  __THREW__ = 0;
  invoke_viii(3350,($28|0),(96|0),(344|0)); //@line 93 "modules/font/Font.cpp"
  $35 = __THREW__; __THREW__ = 0;
  $36 = ___cxa_find_matching_catch_2()|0;
  $37 = tempRet0;
  $$0 = $37;$$01 = $36;
 }
 __THREW__ = 0;
 invoke_v(3378); //@line 94 "modules/font/Font.cpp"
 $40 = __THREW__; __THREW__ = 0;
 $41 = $40&1;
 if ($41) {
  $42 = ___cxa_find_matching_catch_3(0|0)|0;
  $43 = tempRet0;
  ___clang_call_terminate($42); //@line 94 "modules/font/Font.cpp"
  // unreachable; //@line 94 "modules/font/Font.cpp"
 } else {
  $$1 = $$0;$$12 = $$01;
  ___resumeException($$12|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love4font4Font12newGlyphDataEPNS0_10RasterizerEj($this,$r,$glyph) {
 $this = $this|0;
 $r = $r|0;
 $glyph = $glyph|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$r>>2]|0; //@line 101 "modules/font/Font.cpp"
 $1 = ((($0)) + 36|0); //@line 101 "modules/font/Font.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 101 "modules/font/Font.cpp"
 $3 = (FUNCTION_TABLE_iii[$2 & 8191]($r,$glyph)|0); //@line 101 "modules/font/Font.cpp"
 return ($3|0); //@line 101 "modules/font/Font.cpp"
}
function __ZN4love4font9GlyphDataD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (24160); //@line 58 "modules/font/GlyphData.cpp"
 $0 = ((($this)) + 32|0); //@line 59 "modules/font/GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 59 "modules/font/GlyphData.cpp"
 $2 = ($1|0)==(0|0); //@line 59 "modules/font/GlyphData.cpp"
 if ($2) {
  return; //@line 60 "modules/font/GlyphData.cpp"
 }
 __ZdaPv($1); //@line 59 "modules/font/GlyphData.cpp"
 return; //@line 60 "modules/font/GlyphData.cpp"
}
function __ZN4love4font9GlyphDataD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (24160); //@line 58 "modules/font/GlyphData.cpp"
 $0 = ((($this)) + 32|0); //@line 59 "modules/font/GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 59 "modules/font/GlyphData.cpp"
 $2 = ($1|0)==(0|0); //@line 59 "modules/font/GlyphData.cpp"
 if (!($2)) {
  __ZdaPv($1); //@line 59 "modules/font/GlyphData.cpp"
 }
 __ZdlPv($this); //@line 58 "modules/font/GlyphData.cpp"
 return; //@line 60 "modules/font/GlyphData.cpp"
}
function __ZNK4love4font9GlyphData7getDataEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 64 "modules/font/GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 64 "modules/font/GlyphData.cpp"
 return ($1|0); //@line 64 "modules/font/GlyphData.cpp"
}
function __ZNK4love4font9GlyphData7getSizeEv($this) {
 $this = $this|0;
 var $$0 = 0, $$0$v = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 36|0); //@line 69 "modules/font/GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 69 "modules/font/GlyphData.cpp"
 $cond = ($1|0)==(0); //@line 69 "modules/font/GlyphData.cpp"
 $2 = HEAP32[$this>>2]|0; //@line 72 "modules/font/GlyphData.cpp"
 $3 = ((($2)) + 28|0); //@line 72 "modules/font/GlyphData.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 72 "modules/font/GlyphData.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($this)|0); //@line 72 "modules/font/GlyphData.cpp"
 $6 = HEAP32[$this>>2]|0; //@line 72 "modules/font/GlyphData.cpp"
 $7 = ((($6)) + 24|0); //@line 72 "modules/font/GlyphData.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 72 "modules/font/GlyphData.cpp"
 $9 = (FUNCTION_TABLE_ii[$8 & 8191]($this)|0); //@line 72 "modules/font/GlyphData.cpp"
 $10 = Math_imul($9, $5)|0; //@line 72 "modules/font/GlyphData.cpp"
 $$0$v = $cond ? 1 : 2; //@line 69 "modules/font/GlyphData.cpp"
 $$0 = $10 << $$0$v; //@line 69 "modules/font/GlyphData.cpp"
 return ($$0|0); //@line 80 "modules/font/GlyphData.cpp"
}
function __ZNK4love4font9GlyphData9getHeightEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 84 "modules/font/GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 84 "modules/font/GlyphData.cpp"
 return ($1|0); //@line 84 "modules/font/GlyphData.cpp"
}
function __ZNK4love4font9GlyphData8getWidthEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 89 "modules/font/GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 89 "modules/font/GlyphData.cpp"
 return ($1|0); //@line 89 "modules/font/GlyphData.cpp"
}
function __GLOBAL__sub_I_GlyphData_cpp() {
 var $$02$i$i$1$i = 0, $$02$i$i$i = 0, $$lcssa$i$1$i = 0, $$lcssa$i$i = 0, $$lcssa10$i$1$i = 0, $$lcssa10$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i$i$1$i = 0, $hash$0$lcssa$i$i$i = 0;
 var $hash$01$i$i$1$i = 0, $hash$01$i$i$i = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 HEAP8[(632928)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632940)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632952)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632964)>>0] = 0; //@line 155 "./common/StringMap.h"
 $0 = (632968); //@line 44 "./common/StringMap.h"
 $1 = $0; //@line 44 "./common/StringMap.h"
 HEAP32[$1>>2] = 0; //@line 44 "./common/StringMap.h"
 $2 = (($0) + 4)|0; //@line 44 "./common/StringMap.h"
 $3 = $2; //@line 44 "./common/StringMap.h"
 HEAP32[$3>>2] = 0; //@line 44 "./common/StringMap.h"
 $4 = HEAP32[6048]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(24196)>>2]|0; //@line 49 "./common/StringMap.h"
 $6 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
 $7 = ($6<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($7) {
  $hash$0$lcssa$i$i$i = 5381;
 } else {
  $$02$i$i$i = $4;$9 = $6;$hash$01$i$i$i = 5381;
  while(1) {
   $8 = $9 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $10 = ((($$02$i$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $11 = ($hash$01$i$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $12 = (($11) + ($8))|0; //@line 143 "./common/StringMap.h"
   $13 = HEAP8[$10>>0]|0; //@line 142 "./common/StringMap.h"
   $14 = ($13<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($14) {
    $hash$0$lcssa$i$i$i = $12;
    break;
   } else {
    $$02$i$i$i = $10;$9 = $13;$hash$01$i$i$i = $12;
   }
  }
 }
 $15 = $hash$0$lcssa$i$i$i & 3; //@line 112 "./common/StringMap.h"
 $16 = (((632920 + (($15*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $17 = HEAP8[$16>>0]|0; //@line 114 "./common/StringMap.h"
 $18 = ($17<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($18) {
  $$lcssa$i$i = $15;$$lcssa10$i$i = $16;
  label = 5;
 } else {
  $19 = (($hash$0$lcssa$i$i$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $20 = $19 & 3; //@line 112 "./common/StringMap.h"
  $21 = (((632920 + (($20*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $22 = HEAP8[$21>>0]|0; //@line 114 "./common/StringMap.h"
  $23 = ($22<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($23) {
   $$lcssa$i$i = $20;$$lcssa10$i$i = $21;
   label = 5;
  } else {
   $28 = (($hash$0$lcssa$i$i$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $29 = $28 & 3; //@line 112 "./common/StringMap.h"
   $30 = (((632920 + (($29*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $31 = HEAP8[$30>>0]|0; //@line 114 "./common/StringMap.h"
   $32 = ($31<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($32) {
    $$lcssa$i$i = $29;$$lcssa10$i$i = $30;
    label = 5;
   } else {
    $33 = (($hash$0$lcssa$i$i$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $34 = $33 & 3; //@line 112 "./common/StringMap.h"
    $35 = (((632920 + (($34*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $36 = HEAP8[$35>>0]|0; //@line 114 "./common/StringMap.h"
    $37 = ($36<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($37) {
     $$lcssa$i$i = $34;$$lcssa10$i$i = $35;
     label = 5;
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10$i$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $24 = (632920 + (($$lcssa$i$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$24>>2] = $4; //@line 118 "./common/StringMap.h"
  $25 = (((632920 + (($$lcssa$i$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$25>>2] = $5; //@line 119 "./common/StringMap.h"
 }
 $26 = ($5>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($26) {
  HEAP32[$vararg_buffer>>2] = $4; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $5; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
 } else {
  $27 = (632968 + ($5<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$27>>2] = $4; //@line 132 "./common/StringMap.h"
 }
 $38 = HEAP32[(24200)>>2]|0; //@line 49 "./common/StringMap.h"
 $39 = HEAP32[(24204)>>2]|0; //@line 49 "./common/StringMap.h"
 $40 = HEAP8[$38>>0]|0; //@line 142 "./common/StringMap.h"
 $41 = ($40<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($41) {
  $hash$0$lcssa$i$i$1$i = 5381;
 } else {
  $$02$i$i$1$i = $38;$43 = $40;$hash$01$i$i$1$i = 5381;
  while(1) {
   $42 = $43 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $44 = ((($$02$i$i$1$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $45 = ($hash$01$i$i$1$i*33)|0; //@line 143 "./common/StringMap.h"
   $46 = (($45) + ($42))|0; //@line 143 "./common/StringMap.h"
   $47 = HEAP8[$44>>0]|0; //@line 142 "./common/StringMap.h"
   $48 = ($47<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($48) {
    $hash$0$lcssa$i$i$1$i = $46;
    break;
   } else {
    $$02$i$i$1$i = $44;$43 = $47;$hash$01$i$i$1$i = $46;
   }
  }
 }
 $49 = $hash$0$lcssa$i$i$1$i & 3; //@line 112 "./common/StringMap.h"
 $50 = (((632920 + (($49*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $51 = HEAP8[$50>>0]|0; //@line 114 "./common/StringMap.h"
 $52 = ($51<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($52) {
  $$lcssa$i$1$i = $49;$$lcssa10$i$1$i = $50;
  label = 17;
 } else {
  $53 = (($hash$0$lcssa$i$i$1$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $54 = $53 & 3; //@line 112 "./common/StringMap.h"
  $55 = (((632920 + (($54*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $56 = HEAP8[$55>>0]|0; //@line 114 "./common/StringMap.h"
  $57 = ($56<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($57) {
   $$lcssa$i$1$i = $54;$$lcssa10$i$1$i = $55;
   label = 17;
  } else {
   $58 = (($hash$0$lcssa$i$i$1$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $59 = $58 & 3; //@line 112 "./common/StringMap.h"
   $60 = (((632920 + (($59*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $61 = HEAP8[$60>>0]|0; //@line 114 "./common/StringMap.h"
   $62 = ($61<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($62) {
    $$lcssa$i$1$i = $59;$$lcssa10$i$1$i = $60;
    label = 17;
   } else {
    $63 = (($hash$0$lcssa$i$i$1$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $64 = $63 & 3; //@line 112 "./common/StringMap.h"
    $65 = (((632920 + (($64*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $66 = HEAP8[$65>>0]|0; //@line 114 "./common/StringMap.h"
    $67 = ($66<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($67) {
     $$lcssa$i$1$i = $64;$$lcssa10$i$1$i = $65;
     label = 17;
    }
   }
  }
 }
 if ((label|0) == 17) {
  HEAP8[$$lcssa10$i$1$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $68 = (632920 + (($$lcssa$i$1$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$68>>2] = $38; //@line 118 "./common/StringMap.h"
  $69 = (((632920 + (($$lcssa$i$1$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$69>>2] = $39; //@line 119 "./common/StringMap.h"
 }
 $70 = ($39>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($70) {
  HEAP32[$vararg_buffer2>>2] = $38; //@line 128 "./common/StringMap.h"
  $vararg_ptr5 = ((($vararg_buffer2)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr5>>2] = $39; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer2)|0); //@line 128 "./common/StringMap.h"
  STACKTOP = sp;return;
 } else {
  $71 = (632968 + ($39<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$71>>2] = $38; //@line 132 "./common/StringMap.h"
  STACKTOP = sp;return;
 }
}
function __ZN4love4font9GlyphDataC2EjNS0_12GlyphMetricsENS1_6FormatE($this,$glyph,$glyphMetrics,$f) {
 $this = $this|0;
 $glyph = $glyph|0;
 $glyphMetrics = $glyphMetrics|0;
 $f = $f|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 37 "./common/Data.h"
 HEAP32[$this>>2] = (24160); //@line 41 "modules/font/GlyphData.cpp"
 $0 = ((($this)) + 8|0); //@line 37 "modules/font/GlyphData.cpp"
 HEAP32[$0>>2] = $glyph; //@line 37 "modules/font/GlyphData.cpp"
 $1 = ((($this)) + 12|0); //@line 38 "modules/font/GlyphData.cpp"
 ;HEAP32[$1>>2]=HEAP32[$glyphMetrics>>2]|0;HEAP32[$1+4>>2]=HEAP32[$glyphMetrics+4>>2]|0;HEAP32[$1+8>>2]=HEAP32[$glyphMetrics+8>>2]|0;HEAP32[$1+12>>2]=HEAP32[$glyphMetrics+12>>2]|0;HEAP32[$1+16>>2]=HEAP32[$glyphMetrics+16>>2]|0; //@line 38 "modules/font/GlyphData.cpp"
 $2 = ((($this)) + 32|0); //@line 39 "modules/font/GlyphData.cpp"
 HEAP32[$2>>2] = 0; //@line 39 "modules/font/GlyphData.cpp"
 $3 = ((($this)) + 36|0); //@line 40 "modules/font/GlyphData.cpp"
 HEAP32[$3>>2] = $f; //@line 40 "modules/font/GlyphData.cpp"
 $4 = ((($this)) + 16|0); //@line 42 "modules/font/GlyphData.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 42 "modules/font/GlyphData.cpp"
 $6 = ($5|0)>(0); //@line 42 "modules/font/GlyphData.cpp"
 if (!($6)) {
  return; //@line 55 "modules/font/GlyphData.cpp"
 }
 $7 = HEAP32[$1>>2]|0; //@line 42 "modules/font/GlyphData.cpp"
 $8 = ($7|0)>(0); //@line 42 "modules/font/GlyphData.cpp"
 if (!($8)) {
  return; //@line 55 "modules/font/GlyphData.cpp"
 }
 $cond = ($f|0)==(0); //@line 44 "modules/font/GlyphData.cpp"
 $9 = Math_imul($7, $5)|0; //@line 47 "modules/font/GlyphData.cpp"
 if ($cond) {
  $10 = $9 << 1; //@line 47 "modules/font/GlyphData.cpp"
  $11 = ($9|0)<(0); //@line 47 "modules/font/GlyphData.cpp"
  $12 = $11 ? -1 : $10; //@line 47 "modules/font/GlyphData.cpp"
  __THREW__ = 0;
  $13 = (invoke_ii(3492,($12|0))|0); //@line 47 "modules/font/GlyphData.cpp"
  $14 = __THREW__; __THREW__ = 0;
  $15 = $14&1;
  if ($15) {
   $16 = ___cxa_find_matching_catch_2()|0;
   $17 = tempRet0;
   ___resumeException($16|0);
   // unreachable;
  }
  HEAP32[$2>>2] = $13; //@line 47 "modules/font/GlyphData.cpp"
  return; //@line 55 "modules/font/GlyphData.cpp"
 } else {
  $18 = $9 << 2; //@line 51 "modules/font/GlyphData.cpp"
  $19 = ($9|0)<(0); //@line 51 "modules/font/GlyphData.cpp"
  $20 = $19 ? -1 : $18; //@line 51 "modules/font/GlyphData.cpp"
  __THREW__ = 0;
  $21 = (invoke_ii(3492,($20|0))|0); //@line 51 "modules/font/GlyphData.cpp"
  $22 = __THREW__; __THREW__ = 0;
  $23 = $22&1;
  if ($23) {
   $16 = ___cxa_find_matching_catch_2()|0;
   $17 = tempRet0;
   ___resumeException($16|0);
   // unreachable;
  }
  HEAP32[$2>>2] = $21; //@line 51 "modules/font/GlyphData.cpp"
  return; //@line 55 "modules/font/GlyphData.cpp"
 }
}
function __ZNK4love4font9GlyphData8getGlyphEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 94 "modules/font/GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 94 "modules/font/GlyphData.cpp"
 return ($1|0); //@line 94 "modules/font/GlyphData.cpp"
}
function __ZNK4love4font9GlyphData14getGlyphStringEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $u = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $u = sp + 4|0;
 ;HEAP8[$u>>0]=0|0;HEAP8[$u+1>>0]=0|0;HEAP8[$u+2>>0]=0|0;HEAP8[$u+3>>0]=0|0;HEAP8[$u+4>>0]=0|0; //@line 99 "modules/font/GlyphData.cpp"
 $0 = ((($this)) + 8|0); //@line 104 "modules/font/GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 104 "modules/font/GlyphData.cpp"
 __THREW__ = 0;
 $2 = (invoke_iii(3999,($1|0),($u|0))|0); //@line 104 "modules/font/GlyphData.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  $5 = $2; //@line 105 "modules/font/GlyphData.cpp"
  $6 = $u; //@line 105 "modules/font/GlyphData.cpp"
  $7 = (($5) - ($6))|0; //@line 105 "modules/font/GlyphData.cpp"
  $8 = ($7|0)<(0); //@line 113 "modules/font/GlyphData.cpp"
  if ($8) {
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($agg$result,1282466,0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   STACKTOP = sp;return; //@line 117 "modules/font/GlyphData.cpp"
  } else {
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($agg$result,$u,$7); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   STACKTOP = sp;return; //@line 117 "modules/font/GlyphData.cpp"
  }
 }
 $9 = ___cxa_find_matching_catch_3(1744|0)|0;
 $10 = tempRet0;
 $11 = (_llvm_eh_typeid_for((1744|0))|0); //@line 106 "modules/font/GlyphData.cpp"
 $12 = ($10|0)==($11|0); //@line 106 "modules/font/GlyphData.cpp"
 if (!($12)) {
  $$1 = $10;$$12 = $9;
  ___resumeException($$12|0);
  // unreachable;
 }
 $13 = (___cxa_begin_catch(($9|0))|0); //@line 106 "modules/font/GlyphData.cpp"
 $14 = (___cxa_allocate_exception(16)|0); //@line 109 "modules/font/GlyphData.cpp"
 $15 = HEAP32[$13>>2]|0; //@line 109 "modules/font/GlyphData.cpp"
 $16 = ((($15)) + 8|0); //@line 109 "modules/font/GlyphData.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 109 "modules/font/GlyphData.cpp"
 $18 = (FUNCTION_TABLE_ii[$17 & 8191]($13)|0); //@line 109 "modules/font/GlyphData.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $18; //@line 109 "modules/font/GlyphData.cpp"
 invoke_viii(3349,($14|0),(434605|0),($vararg_buffer|0)); //@line 109 "modules/font/GlyphData.cpp"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $24 = ___cxa_find_matching_catch_2()|0;
  $25 = tempRet0;
  ___cxa_free_exception(($14|0)); //@line 109 "modules/font/GlyphData.cpp"
  $$0 = $25;$$01 = $24;
 } else {
  __THREW__ = 0;
  invoke_viii(3350,($14|0),(96|0),(344|0)); //@line 109 "modules/font/GlyphData.cpp"
  $21 = __THREW__; __THREW__ = 0;
  $22 = ___cxa_find_matching_catch_2()|0;
  $23 = tempRet0;
  $$0 = $23;$$01 = $22;
 }
 __THREW__ = 0;
 invoke_v(3378); //@line 110 "modules/font/GlyphData.cpp"
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  $28 = ___cxa_find_matching_catch_3(0|0)|0;
  $29 = tempRet0;
  ___clang_call_terminate($28); //@line 110 "modules/font/GlyphData.cpp"
  // unreachable; //@line 110 "modules/font/GlyphData.cpp"
 } else {
  $$1 = $$0;$$12 = $$01;
  ___resumeException($$12|0);
  // unreachable;
 }
}
function __ZN4utf86appendIPcEET_jS2_($cp,$result) {
 $cp = $cp|0;
 $result = $result|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($cp>>>0)<(1114112); //@line 95 "./libraries/utf8/utf8/core.h"
 $1 = $cp & -2048; //@line 89 "./libraries/utf8/utf8/core.h"
 $2 = ($1|0)!=(55296); //@line 95 "./libraries/utf8/utf8/core.h"
 $3 = $0 & $2; //@line 95 "./libraries/utf8/utf8/core.h"
 if (!($3)) {
  $4 = (___cxa_allocate_exception(8)|0); //@line 76 "./libraries/utf8/utf8/checked.h"
  HEAP32[$4>>2] = (24140); //@line 44 "./libraries/utf8/utf8/checked.h"
  $5 = ((($4)) + 4|0); //@line 44 "./libraries/utf8/utf8/checked.h"
  HEAP32[$5>>2] = $cp; //@line 44 "./libraries/utf8/utf8/checked.h"
  ___cxa_throw(($4|0),(1792|0),(1696|0)); //@line 76 "./libraries/utf8/utf8/checked.h"
  // unreachable; //@line 76 "./libraries/utf8/utf8/checked.h"
 }
 $6 = ($cp>>>0)<(128); //@line 78 "./libraries/utf8/utf8/checked.h"
 if ($6) {
  $7 = $cp&255; //@line 79 "./libraries/utf8/utf8/checked.h"
  $8 = ((($result)) + 1|0); //@line 79 "./libraries/utf8/utf8/checked.h"
  HEAP8[$result>>0] = $7; //@line 79 "./libraries/utf8/utf8/checked.h"
  $$0 = $8;
  return ($$0|0); //@line 95 "./libraries/utf8/utf8/checked.h"
 }
 $9 = ($cp>>>0)<(2048); //@line 80 "./libraries/utf8/utf8/checked.h"
 if ($9) {
  $10 = $cp >>> 6; //@line 81 "./libraries/utf8/utf8/checked.h"
  $11 = $10 | 192; //@line 81 "./libraries/utf8/utf8/checked.h"
  $12 = $11&255; //@line 81 "./libraries/utf8/utf8/checked.h"
  $13 = ((($result)) + 1|0); //@line 81 "./libraries/utf8/utf8/checked.h"
  HEAP8[$result>>0] = $12; //@line 81 "./libraries/utf8/utf8/checked.h"
  $14 = $cp & 63; //@line 82 "./libraries/utf8/utf8/checked.h"
  $15 = $14 | 128; //@line 82 "./libraries/utf8/utf8/checked.h"
  $16 = $15&255; //@line 82 "./libraries/utf8/utf8/checked.h"
  $17 = ((($result)) + 2|0); //@line 82 "./libraries/utf8/utf8/checked.h"
  HEAP8[$13>>0] = $16; //@line 82 "./libraries/utf8/utf8/checked.h"
  $$0 = $17;
  return ($$0|0); //@line 95 "./libraries/utf8/utf8/checked.h"
 }
 $18 = ($cp>>>0)<(65536); //@line 84 "./libraries/utf8/utf8/checked.h"
 if ($18) {
  $19 = $cp >>> 12; //@line 85 "./libraries/utf8/utf8/checked.h"
  $20 = $19 | 224; //@line 85 "./libraries/utf8/utf8/checked.h"
  $21 = $20&255; //@line 85 "./libraries/utf8/utf8/checked.h"
  $22 = ((($result)) + 1|0); //@line 85 "./libraries/utf8/utf8/checked.h"
  HEAP8[$result>>0] = $21; //@line 85 "./libraries/utf8/utf8/checked.h"
  $23 = $cp >>> 6; //@line 86 "./libraries/utf8/utf8/checked.h"
  $24 = $23 & 63; //@line 86 "./libraries/utf8/utf8/checked.h"
  $25 = $24 | 128; //@line 86 "./libraries/utf8/utf8/checked.h"
  $26 = $25&255; //@line 86 "./libraries/utf8/utf8/checked.h"
  $27 = ((($result)) + 2|0); //@line 86 "./libraries/utf8/utf8/checked.h"
  HEAP8[$22>>0] = $26; //@line 86 "./libraries/utf8/utf8/checked.h"
  $28 = $cp & 63; //@line 87 "./libraries/utf8/utf8/checked.h"
  $29 = $28 | 128; //@line 87 "./libraries/utf8/utf8/checked.h"
  $30 = $29&255; //@line 87 "./libraries/utf8/utf8/checked.h"
  $31 = ((($result)) + 3|0); //@line 87 "./libraries/utf8/utf8/checked.h"
  HEAP8[$27>>0] = $30; //@line 87 "./libraries/utf8/utf8/checked.h"
  $$0 = $31;
  return ($$0|0); //@line 95 "./libraries/utf8/utf8/checked.h"
 } else {
  $32 = $cp >>> 18; //@line 90 "./libraries/utf8/utf8/checked.h"
  $33 = $32 | 240; //@line 90 "./libraries/utf8/utf8/checked.h"
  $34 = $33&255; //@line 90 "./libraries/utf8/utf8/checked.h"
  $35 = ((($result)) + 1|0); //@line 90 "./libraries/utf8/utf8/checked.h"
  HEAP8[$result>>0] = $34; //@line 90 "./libraries/utf8/utf8/checked.h"
  $36 = $cp >>> 12; //@line 91 "./libraries/utf8/utf8/checked.h"
  $37 = $36 & 63; //@line 91 "./libraries/utf8/utf8/checked.h"
  $38 = $37 | 128; //@line 91 "./libraries/utf8/utf8/checked.h"
  $39 = $38&255; //@line 91 "./libraries/utf8/utf8/checked.h"
  $40 = ((($result)) + 2|0); //@line 91 "./libraries/utf8/utf8/checked.h"
  HEAP8[$35>>0] = $39; //@line 91 "./libraries/utf8/utf8/checked.h"
  $41 = $cp >>> 6; //@line 92 "./libraries/utf8/utf8/checked.h"
  $42 = $41 & 63; //@line 92 "./libraries/utf8/utf8/checked.h"
  $43 = $42 | 128; //@line 92 "./libraries/utf8/utf8/checked.h"
  $44 = $43&255; //@line 92 "./libraries/utf8/utf8/checked.h"
  $45 = ((($result)) + 3|0); //@line 92 "./libraries/utf8/utf8/checked.h"
  HEAP8[$40>>0] = $44; //@line 92 "./libraries/utf8/utf8/checked.h"
  $46 = $cp & 63; //@line 93 "./libraries/utf8/utf8/checked.h"
  $47 = $46 | 128; //@line 93 "./libraries/utf8/utf8/checked.h"
  $48 = $47&255; //@line 93 "./libraries/utf8/utf8/checked.h"
  $49 = ((($result)) + 4|0); //@line 93 "./libraries/utf8/utf8/checked.h"
  HEAP8[$45>>0] = $48; //@line 93 "./libraries/utf8/utf8/checked.h"
  $$0 = $49;
  return ($$0|0); //@line 95 "./libraries/utf8/utf8/checked.h"
 }
 return (0)|0;
}
function __ZNK4love4font9GlyphData10getAdvanceEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 121 "modules/font/GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 121 "modules/font/GlyphData.cpp"
 return ($1|0); //@line 121 "modules/font/GlyphData.cpp"
}
function __ZNK4love4font9GlyphData11getBearingXEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 126 "modules/font/GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 126 "modules/font/GlyphData.cpp"
 return ($1|0); //@line 126 "modules/font/GlyphData.cpp"
}
function __ZNK4love4font9GlyphData11getBearingYEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 131 "modules/font/GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 131 "modules/font/GlyphData.cpp"
 return ($1|0); //@line 131 "modules/font/GlyphData.cpp"
}
function __ZNK4love4font9GlyphData7getMinXEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 126 "modules/font/GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 126 "modules/font/GlyphData.cpp"
 return ($1|0); //@line 136 "modules/font/GlyphData.cpp"
}
function __ZNK4love4font9GlyphData7getMinYEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 141 "modules/font/GlyphData.cpp"
 $1 = ((($0)) + 24|0); //@line 141 "modules/font/GlyphData.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 141 "modules/font/GlyphData.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 141 "modules/font/GlyphData.cpp"
 $4 = ((($this)) + 28|0); //@line 131 "modules/font/GlyphData.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 131 "modules/font/GlyphData.cpp"
 $6 = (($3) - ($5))|0; //@line 141 "modules/font/GlyphData.cpp"
 return ($6|0); //@line 141 "modules/font/GlyphData.cpp"
}
function __ZNK4love4font9GlyphData7getMaxXEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 126 "modules/font/GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 126 "modules/font/GlyphData.cpp"
 $2 = HEAP32[$this>>2]|0; //@line 146 "modules/font/GlyphData.cpp"
 $3 = ((($2)) + 28|0); //@line 146 "modules/font/GlyphData.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 146 "modules/font/GlyphData.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($this)|0); //@line 146 "modules/font/GlyphData.cpp"
 $6 = (($5) + ($1))|0; //@line 146 "modules/font/GlyphData.cpp"
 return ($6|0); //@line 146 "modules/font/GlyphData.cpp"
}
function __ZNK4love4font9GlyphData7getMaxYEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 131 "modules/font/GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 131 "modules/font/GlyphData.cpp"
 return ($1|0); //@line 151 "modules/font/GlyphData.cpp"
}
function __ZNK4love4font9GlyphData9getFormatEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 36|0); //@line 156 "modules/font/GlyphData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 156 "modules/font/GlyphData.cpp"
 return ($1|0); //@line 156 "modules/font/GlyphData.cpp"
}
function __ZN4love4font9GlyphData11getConstantENS1_6FormatERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(1); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (632968 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 166 "modules/font/GlyphData.cpp"
}
function __ZN4love4font15ImageRasterizerD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (24216); //@line 47 "modules/font/ImageRasterizer.cpp"
 $0 = ((($this)) + 40|0); //@line 783 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $1 = ((($this)) + 44|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = HEAP32[$1>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeIjN4love4font15ImageRasterizer14ImageGlyphDataEEENS_19__map_value_compareIjS6_NS_4lessIjEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($0,$2); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $3 = ((($this)) + 24|0); //@line 106 "./common/Object.h"
 $4 = HEAP32[$3>>2]|0; //@line 106 "./common/Object.h"
 $5 = ($4|0)==(0|0); //@line 106 "./common/Object.h"
 if ($5) {
  __ZN4love4font10RasterizerD2Ev($this); //@line 48 "modules/font/ImageRasterizer.cpp"
  return; //@line 48 "modules/font/ImageRasterizer.cpp"
 }
 $6 = HEAP32[$4>>2]|0; //@line 106 "./common/Object.h"
 $7 = ((($6)) + 12|0); //@line 106 "./common/Object.h"
 $8 = HEAP32[$7>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($8|0,($4|0)); //@line 106 "./common/Object.h"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $11 = ___cxa_find_matching_catch_3(0|0)|0;
  $12 = tempRet0;
  ___clang_call_terminate($11); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  __ZN4love4font10RasterizerD2Ev($this); //@line 48 "modules/font/ImageRasterizer.cpp"
  return; //@line 48 "modules/font/ImageRasterizer.cpp"
 }
}
function __ZNSt3__16__treeINS_12__value_typeIjN4love4font15ImageRasterizer14ImageGlyphDataEEENS_19__map_value_compareIjS6_NS_4lessIjEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($this,$__nd) {
 $this = $this|0;
 $__nd = $__nd|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__nd|0)==(0|0); //@line 1440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($0) {
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 } else {
  $1 = HEAP32[$__nd>>2]|0; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIjN4love4font15ImageRasterizer14ImageGlyphDataEEENS_19__map_value_compareIjS6_NS_4lessIjEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($this,$1); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2 = ((($__nd)) + 4|0); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3 = HEAP32[$2>>2]|0; //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIjN4love4font15ImageRasterizer14ImageGlyphDataEEENS_19__map_value_compareIjS6_NS_4lessIjEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($this,$3); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZdlPv($__nd); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
}
function __ZN4love4font15ImageRasterizerD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (24216); //@line 47 "modules/font/ImageRasterizer.cpp"
 $0 = ((($this)) + 40|0); //@line 783 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $1 = ((($this)) + 44|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = HEAP32[$1>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeIjN4love4font15ImageRasterizer14ImageGlyphDataEEENS_19__map_value_compareIjS6_NS_4lessIjEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($0,$2); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $3 = ((($this)) + 24|0); //@line 106 "./common/Object.h"
 $4 = HEAP32[$3>>2]|0; //@line 106 "./common/Object.h"
 $5 = ($4|0)==(0|0); //@line 106 "./common/Object.h"
 if ($5) {
  __ZN4love4font10RasterizerD2Ev($this); //@line 48 "modules/font/ImageRasterizer.cpp"
  __ZdlPv($this); //@line 47 "modules/font/ImageRasterizer.cpp"
  return; //@line 48 "modules/font/ImageRasterizer.cpp"
 }
 $6 = HEAP32[$4>>2]|0; //@line 106 "./common/Object.h"
 $7 = ((($6)) + 12|0); //@line 106 "./common/Object.h"
 $8 = HEAP32[$7>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($8|0,($4|0)); //@line 106 "./common/Object.h"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $11 = ___cxa_find_matching_catch_3(0|0)|0;
  $12 = tempRet0;
  ___clang_call_terminate($11); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  __ZN4love4font10RasterizerD2Ev($this); //@line 48 "modules/font/ImageRasterizer.cpp"
  __ZdlPv($this); //@line 47 "modules/font/ImageRasterizer.cpp"
  return; //@line 48 "modules/font/ImageRasterizer.cpp"
 }
}
function __ZNK4love4font15ImageRasterizer13getLineHeightEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 52 "modules/font/ImageRasterizer.cpp"
 $1 = ((($0)) + 16|0); //@line 52 "modules/font/ImageRasterizer.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 52 "modules/font/ImageRasterizer.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 52 "modules/font/ImageRasterizer.cpp"
 return ($3|0); //@line 52 "modules/font/ImageRasterizer.cpp"
}
function __ZNK4love4font15ImageRasterizer12getGlyphDataEj($this,$glyph) {
 $this = $this|0;
 $glyph = $glyph|0;
 var $$0$ph$lcssa$i$i$i = 0, $$0$ph6$i$i$i = 0, $$01 = 0, $$01$ph5$i$i$i = 0, $$013$i$i$i = 0, $$013$i$i$i$lcssa = 0, $$02 = 0, $$byval_copy = 0, $$sroa$0$0$i$i3 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $gm$sroa$5$0 = 0, $gm$sroa$9$0 = 0;
 var $i$0 = 0, $lock = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 24|0;
 $0 = sp + 4|0;
 $lock = sp;
 $1 = ((($this)) + 44|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = HEAP32[$1>>2]|0; //@line 878 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $3 = ($2|0)==(0|0); //@line 2120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($3) {
  $$sroa$0$0$i$i3 = $1;$gm$sroa$5$0 = 0;$gm$sroa$9$0 = 0;
 } else {
  $$0$ph6$i$i$i = $1;$$01$ph5$i$i$i = $2;
  L2: while(1) {
   $$013$i$i$i = $$01$ph5$i$i$i;
   while(1) {
    $4 = ((($$013$i$i$i)) + 16|0); //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $5 = HEAP32[$4>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $6 = ($5>>>0)<($glyph>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if (!($6)) {
     $$013$i$i$i$lcssa = $$013$i$i$i;
     break;
    }
    $9 = ((($$013$i$i$i)) + 4|0); //@line 2128 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $10 = HEAP32[$9>>2]|0; //@line 2128 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $11 = ($10|0)==(0|0); //@line 2120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($11) {
     $$0$ph$lcssa$i$i$i = $$0$ph6$i$i$i;
     break L2;
    } else {
     $$013$i$i$i = $10;
    }
   }
   $7 = HEAP32[$$013$i$i$i$lcssa>>2]|0; //@line 2125 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $8 = ($7|0)==(0|0); //@line 2120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($8) {
    $$0$ph$lcssa$i$i$i = $$013$i$i$i$lcssa;
    break;
   } else {
    $$0$ph6$i$i$i = $$013$i$i$i$lcssa;$$01$ph5$i$i$i = $7;
   }
  }
  $12 = ($$0$ph$lcssa$i$i$i|0)==($1|0); //@line 766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($12) {
   $$sroa$0$0$i$i3 = $1;$gm$sroa$5$0 = 0;$gm$sroa$9$0 = 0;
  } else {
   $13 = ((($$0$ph$lcssa$i$i$i)) + 16|0); //@line 463 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $14 = HEAP32[$13>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   $15 = ($14>>>0)>($glyph>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   if ($15) {
    $$sroa$0$0$i$i3 = $1;$gm$sroa$5$0 = 0;$gm$sroa$9$0 = 0;
   } else {
    $16 = ((($$0$ph$lcssa$i$i$i)) + 24|0); //@line 63 "modules/font/ImageRasterizer.cpp"
    $17 = HEAP32[$16>>2]|0; //@line 63 "modules/font/ImageRasterizer.cpp"
    $18 = ((($this)) + 36|0); //@line 64 "modules/font/ImageRasterizer.cpp"
    $19 = HEAP32[$18>>2]|0; //@line 64 "modules/font/ImageRasterizer.cpp"
    $20 = (($19) + ($17))|0; //@line 64 "modules/font/ImageRasterizer.cpp"
    $$sroa$0$0$i$i3 = $$0$ph$lcssa$i$i$i;$gm$sroa$5$0 = $17;$gm$sroa$9$0 = $20;
   }
  }
 }
 $21 = ((($this)) + 20|0); //@line 67 "modules/font/ImageRasterizer.cpp"
 $22 = HEAP32[$21>>2]|0; //@line 67 "modules/font/ImageRasterizer.cpp"
 $23 = (__Znwj(40)|0); //@line 69 "modules/font/ImageRasterizer.cpp"
 HEAP32[$0>>2] = $22; //@line 69 "modules/font/ImageRasterizer.cpp"
 $24 = ((($0)) + 4|0); //@line 69 "modules/font/ImageRasterizer.cpp"
 HEAP32[$24>>2] = $gm$sroa$5$0; //@line 69 "modules/font/ImageRasterizer.cpp"
 $25 = ((($0)) + 8|0); //@line 69 "modules/font/ImageRasterizer.cpp"
 HEAP32[$25>>2] = $gm$sroa$9$0; //@line 69 "modules/font/ImageRasterizer.cpp"
 $26 = ((($0)) + 12|0); //@line 69 "modules/font/ImageRasterizer.cpp"
 $27 = $26; //@line 69 "modules/font/ImageRasterizer.cpp"
 $28 = $27; //@line 69 "modules/font/ImageRasterizer.cpp"
 HEAP32[$28>>2] = 0; //@line 69 "modules/font/ImageRasterizer.cpp"
 $29 = (($27) + 4)|0; //@line 69 "modules/font/ImageRasterizer.cpp"
 $30 = $29; //@line 69 "modules/font/ImageRasterizer.cpp"
 HEAP32[$30>>2] = 0; //@line 69 "modules/font/ImageRasterizer.cpp"
 __THREW__ = 0;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;HEAP32[$$byval_copy+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$$byval_copy+8>>2]=HEAP32[$0+8>>2]|0;HEAP32[$$byval_copy+12>>2]=HEAP32[$0+12>>2]|0;HEAP32[$$byval_copy+16>>2]=HEAP32[$0+16>>2]|0; //@line 69 "modules/font/ImageRasterizer.cpp"
 invoke_viiii(3981,($23|0),($glyph|0),($$byval_copy|0),1); //@line 69 "modules/font/ImageRasterizer.cpp"
 $31 = __THREW__; __THREW__ = 0;
 $32 = $31&1;
 if ($32) {
  $34 = ___cxa_find_matching_catch_2()|0;
  $35 = tempRet0;
  __ZdlPv($23); //@line 69 "modules/font/ImageRasterizer.cpp"
  $$01 = $35;$$02 = $34;
  ___resumeException($$02|0);
  // unreachable;
 }
 $33 = ($gm$sroa$5$0|0)==(0); //@line 71 "modules/font/ImageRasterizer.cpp"
 if ($33) {
  STACKTOP = sp;return ($23|0); //@line 93 "modules/font/ImageRasterizer.cpp"
 }
 $36 = ((($this)) + 24|0); //@line 117 "./common/Object.h"
 $37 = HEAP32[$36>>2]|0; //@line 117 "./common/Object.h"
 $38 = (__ZNK4love5image9ImageData8getMutexEv($37)|0); //@line 75 "modules/font/ImageRasterizer.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$38); //@line 75 "modules/font/ImageRasterizer.cpp"
 $39 = HEAP32[$23>>2]|0; //@line 77 "modules/font/ImageRasterizer.cpp"
 $40 = ((($39)) + 16|0); //@line 77 "modules/font/ImageRasterizer.cpp"
 $41 = HEAP32[$40>>2]|0; //@line 77 "modules/font/ImageRasterizer.cpp"
 __THREW__ = 0;
 $42 = (invoke_ii($41|0,($23|0))|0); //@line 77 "modules/font/ImageRasterizer.cpp"
 $43 = __THREW__; __THREW__ = 0;
 $44 = $43&1;
 do {
  if ($44) {
   label = 20;
  } else {
   $45 = HEAP32[$36>>2]|0; //@line 117 "./common/Object.h"
   $46 = HEAP32[$45>>2]|0; //@line 78 "modules/font/ImageRasterizer.cpp"
   $47 = ((($46)) + 16|0); //@line 78 "modules/font/ImageRasterizer.cpp"
   $48 = HEAP32[$47>>2]|0; //@line 78 "modules/font/ImageRasterizer.cpp"
   __THREW__ = 0;
   $49 = (invoke_ii($48|0,($45|0))|0); //@line 78 "modules/font/ImageRasterizer.cpp"
   $50 = __THREW__; __THREW__ = 0;
   $51 = $50&1;
   if ($51) {
    label = 20;
   } else {
    $52 = ((($$sroa$0$0$i$i3)) + 20|0); //@line 83 "modules/font/ImageRasterizer.cpp"
    $53 = ((($this)) + 52|0); //@line 34 "modules/font/ImageRasterizer.cpp"
    $54 = ((($this)) + 53|0); //@line 34 "modules/font/ImageRasterizer.cpp"
    $55 = ((($this)) + 54|0); //@line 34 "modules/font/ImageRasterizer.cpp"
    $56 = ((($this)) + 55|0); //@line 34 "modules/font/ImageRasterizer.cpp"
    $i$0 = 0;
    while(1) {
     $57 = HEAP32[$23>>2]|0; //@line 81 "modules/font/ImageRasterizer.cpp"
     $58 = ((($57)) + 28|0); //@line 81 "modules/font/ImageRasterizer.cpp"
     $59 = HEAP32[$58>>2]|0; //@line 81 "modules/font/ImageRasterizer.cpp"
     __THREW__ = 0;
     $60 = (invoke_ii($59|0,($23|0))|0); //@line 81 "modules/font/ImageRasterizer.cpp"
     $61 = __THREW__; __THREW__ = 0;
     $62 = $61&1;
     if ($62) {
      label = 19;
      break;
     }
     $63 = HEAP32[$23>>2]|0; //@line 81 "modules/font/ImageRasterizer.cpp"
     $64 = ((($63)) + 24|0); //@line 81 "modules/font/ImageRasterizer.cpp"
     $65 = HEAP32[$64>>2]|0; //@line 81 "modules/font/ImageRasterizer.cpp"
     __THREW__ = 0;
     $66 = (invoke_ii($65|0,($23|0))|0); //@line 81 "modules/font/ImageRasterizer.cpp"
     $67 = __THREW__; __THREW__ = 0;
     $68 = $67&1;
     if ($68) {
      label = 19;
      break;
     }
     $69 = Math_imul($66, $60)|0; //@line 81 "modules/font/ImageRasterizer.cpp"
     $70 = ($i$0|0)<($69|0); //@line 81 "modules/font/ImageRasterizer.cpp"
     if (!($70)) {
      break;
     }
     $75 = HEAP32[$52>>2]|0; //@line 83 "modules/font/ImageRasterizer.cpp"
     $76 = HEAP32[$36>>2]|0; //@line 117 "./common/Object.h"
     __THREW__ = 0;
     $77 = (invoke_ii(3984,($76|0))|0); //@line 83 "modules/font/ImageRasterizer.cpp"
     $78 = __THREW__; __THREW__ = 0;
     $79 = $78&1;
     if ($79) {
      label = 19;
      break;
     }
     $80 = (($i$0|0) % ($gm$sroa$5$0|0))&-1; //@line 83 "modules/font/ImageRasterizer.cpp"
     $81 = (($80) + ($75))|0; //@line 83 "modules/font/ImageRasterizer.cpp"
     $82 = (($i$0|0) / ($gm$sroa$5$0|0))&-1; //@line 83 "modules/font/ImageRasterizer.cpp"
     $83 = Math_imul($82, $77)|0; //@line 83 "modules/font/ImageRasterizer.cpp"
     $84 = (($81) + ($83))|0; //@line 83 "modules/font/ImageRasterizer.cpp"
     $85 = (($49) + ($84<<2)|0); //@line 83 "modules/font/ImageRasterizer.cpp"
     $86 = HEAPU8[$85>>0]|(HEAPU8[$85+1>>0]<<8)|(HEAPU8[$85+2>>0]<<16)|(HEAPU8[$85+3>>0]<<24); //@line 83 "modules/font/ImageRasterizer.cpp"
     $87 = $86&255; //@line 83 "modules/font/ImageRasterizer.cpp"
     $88 = $86 >>> 16; //@line 83 "modules/font/ImageRasterizer.cpp"
     $89 = $88&255; //@line 83 "modules/font/ImageRasterizer.cpp"
     $90 = $86 >>> 24; //@line 83 "modules/font/ImageRasterizer.cpp"
     $91 = $90&255; //@line 83 "modules/font/ImageRasterizer.cpp"
     $92 = HEAP8[$53>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
     $93 = ($87<<24>>24)==($92<<24>>24); //@line 34 "modules/font/ImageRasterizer.cpp"
     if ($93) {
      $94 = $86 >>> 8; //@line 83 "modules/font/ImageRasterizer.cpp"
      $95 = $94&255; //@line 83 "modules/font/ImageRasterizer.cpp"
      $96 = HEAP8[$54>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
      $97 = ($95<<24>>24)==($96<<24>>24); //@line 34 "modules/font/ImageRasterizer.cpp"
      if ($97) {
       $98 = HEAP8[$55>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
       $99 = ($89<<24>>24)==($98<<24>>24); //@line 34 "modules/font/ImageRasterizer.cpp"
       if ($99) {
        $101 = HEAP8[$56>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
        $102 = ($91<<24>>24)==($101<<24>>24); //@line 34 "modules/font/ImageRasterizer.cpp"
        $103 = (($42) + ($i$0<<2)|0); //@line 87 "modules/font/ImageRasterizer.cpp"
        if ($102) {
         HEAP8[$103>>0]=0&255;HEAP8[$103+1>>0]=(0>>8)&255;HEAP8[$103+2>>0]=(0>>16)&255;HEAP8[$103+3>>0]=0>>24; //@line 87 "modules/font/ImageRasterizer.cpp"
        } else {
         $107 = $103;
         label = 29;
        }
       } else {
        label = 26;
       }
      } else {
       label = 26;
      }
     } else {
      label = 26;
     }
     if ((label|0) == 26) {
      label = 0;
      $100 = (($42) + ($i$0<<2)|0); //@line 87 "modules/font/ImageRasterizer.cpp"
      $107 = $100;
      label = 29;
     }
     if ((label|0) == 29) {
      label = 0;
      $104 = $90 << 24; //@line 89 "modules/font/ImageRasterizer.cpp"
      $105 = $86 & 16777215; //@line 89 "modules/font/ImageRasterizer.cpp"
      $106 = $105 | $104; //@line 89 "modules/font/ImageRasterizer.cpp"
      HEAP8[$107>>0]=$106&255;HEAP8[$107+1>>0]=($106>>8)&255;HEAP8[$107+2>>0]=($106>>16)&255;HEAP8[$107+3>>0]=$106>>24; //@line 89 "modules/font/ImageRasterizer.cpp"
     }
     $108 = (($i$0) + 1)|0; //@line 81 "modules/font/ImageRasterizer.cpp"
     $i$0 = $108;
    }
    if ((label|0) == 19) {
     $71 = ___cxa_find_matching_catch_2()|0;
     $72 = tempRet0;
     $lpad$phi$index = $71;$lpad$phi$index2 = $72;
     break;
    }
    __ZN4love6thread4LockD2Ev($lock); //@line 93 "modules/font/ImageRasterizer.cpp"
    STACKTOP = sp;return ($23|0); //@line 93 "modules/font/ImageRasterizer.cpp"
   }
  }
 } while(0);
 if ((label|0) == 20) {
  $73 = ___cxa_find_matching_catch_2()|0;
  $74 = tempRet0;
  $lpad$phi$index = $73;$lpad$phi$index2 = $74;
 }
 __ZN4love6thread4LockD2Ev($lock); //@line 93 "modules/font/ImageRasterizer.cpp"
 $$01 = $lpad$phi$index2;$$02 = $lpad$phi$index;
 ___resumeException($$02|0);
 // unreachable;
 return (0)|0;
}
function __ZNK4love4font15ImageRasterizer13getGlyphCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 141 "modules/font/ImageRasterizer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 141 "modules/font/ImageRasterizer.cpp"
 return ($1|0); //@line 141 "modules/font/ImageRasterizer.cpp"
}
function __ZNK4love4font15ImageRasterizer8hasGlyphEj($this,$glyph) {
 $this = $this|0;
 $glyph = $glyph|0;
 var $$0$ph$lcssa$i$i$i = 0, $$0$ph6$i$i$i = 0, $$01$ph5$i$i$i = 0, $$013$i$i$i = 0, $$013$i$i$i$lcssa = 0, $$sroa$0$0$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 44|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 878 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $2 = ($1|0)==(0|0); //@line 2120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if (!($2)) {
  $$0$ph6$i$i$i = $0;$$01$ph5$i$i$i = $1;
  L2: while(1) {
   $$013$i$i$i = $$01$ph5$i$i$i;
   while(1) {
    $3 = ((($$013$i$i$i)) + 16|0); //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $4 = HEAP32[$3>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $5 = ($4>>>0)<($glyph>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if (!($5)) {
     $$013$i$i$i$lcssa = $$013$i$i$i;
     break;
    }
    $8 = ((($$013$i$i$i)) + 4|0); //@line 2128 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $9 = HEAP32[$8>>2]|0; //@line 2128 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $10 = ($9|0)==(0|0); //@line 2120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($10) {
     $$0$ph$lcssa$i$i$i = $$0$ph6$i$i$i;
     break L2;
    } else {
     $$013$i$i$i = $9;
    }
   }
   $6 = HEAP32[$$013$i$i$i$lcssa>>2]|0; //@line 2125 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $7 = ($6|0)==(0|0); //@line 2120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($7) {
    $$0$ph$lcssa$i$i$i = $$013$i$i$i$lcssa;
    break;
   } else {
    $$0$ph6$i$i$i = $$013$i$i$i$lcssa;$$01$ph5$i$i$i = $6;
   }
  }
  $11 = ($$0$ph$lcssa$i$i$i|0)==($0|0); //@line 766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if (!($11)) {
   $12 = ((($$0$ph$lcssa$i$i$i)) + 16|0); //@line 463 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $13 = HEAP32[$12>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   $14 = ($13>>>0)>($glyph>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
   if (!($14)) {
    $$sroa$0$0$i$i = $$0$ph$lcssa$i$i$i;
    $15 = ($$sroa$0$0$i$i|0)!=($0|0); //@line 766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    return ($15|0); //@line 146 "modules/font/ImageRasterizer.cpp"
   }
  }
 }
 $$sroa$0$0$i$i = $0;
 $15 = ($$sroa$0$0$i$i|0)!=($0|0); //@line 766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 return ($15|0); //@line 146 "modules/font/ImageRasterizer.cpp"
}
function __ZN4love4font15ImageRasterizerC2EPNS_5image9ImageDataEPjii($this,$data,$glyphs,$numglyphs,$extraspacing) {
 $this = $this|0;
 $data = $data|0;
 $glyphs = $glyphs|0;
 $numglyphs = $numglyphs|0;
 $extraspacing = $extraspacing|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 48 "./modules/font/Rasterizer.h"
 HEAP32[$this>>2] = (24216); //@line 42 "modules/font/ImageRasterizer.cpp"
 $0 = ((($this)) + 24|0); //@line 93 "./common/Object.h"
 HEAP32[$0>>2] = $data; //@line 93 "./common/Object.h"
 $1 = ($data|0)==(0|0); //@line 95 "./common/Object.h"
 if (!($1)) {
  $2 = HEAP32[$data>>2]|0; //@line 95 "./common/Object.h"
  $3 = ((($2)) + 8|0); //@line 95 "./common/Object.h"
  $4 = HEAP32[$3>>2]|0; //@line 95 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($4|0,($data|0)); //@line 95 "./common/Object.h"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if ($6) {
   $16 = ___cxa_find_matching_catch_2()|0;
   $17 = tempRet0;
   $$0 = $17;$$01 = $16;
   __ZN4love4font10RasterizerD2Ev($this); //@line 44 "modules/font/ImageRasterizer.cpp"
   ___resumeException($$01|0);
   // unreachable;
  }
 }
 $7 = ((($this)) + 28|0); //@line 39 "modules/font/ImageRasterizer.cpp"
 HEAP32[$7>>2] = $glyphs; //@line 39 "modules/font/ImageRasterizer.cpp"
 $8 = ((($this)) + 32|0); //@line 40 "modules/font/ImageRasterizer.cpp"
 HEAP32[$8>>2] = $numglyphs; //@line 40 "modules/font/ImageRasterizer.cpp"
 $9 = ((($this)) + 36|0); //@line 41 "modules/font/ImageRasterizer.cpp"
 HEAP32[$9>>2] = $extraspacing; //@line 41 "modules/font/ImageRasterizer.cpp"
 $10 = ((($this)) + 40|0); //@line 37 "modules/font/ImageRasterizer.cpp"
 $11 = ((($this)) + 44|0); //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$11>>2] = 0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $12 = ((($this)) + 48|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$12>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $13 = ((($this)) + 44|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$10>>2] = $13; //@line 1130 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __THREW__ = 0;
 invoke_vi(4000,($this|0)); //@line 43 "modules/font/ImageRasterizer.cpp"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if (!($15)) {
  return; //@line 44 "modules/font/ImageRasterizer.cpp"
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 $20 = HEAP32[$13>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeIjN4love4font15ImageRasterizer14ImageGlyphDataEEENS_19__map_value_compareIjS6_NS_4lessIjEELb1EEENS_9allocatorIS6_EEE7destroyEPNS_11__tree_nodeIS6_PvEE($10,$20); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $21 = HEAP32[$0>>2]|0; //@line 106 "./common/Object.h"
 $22 = ($21|0)==(0|0); //@line 106 "./common/Object.h"
 if ($22) {
  $$0 = $19;$$01 = $18;
  __ZN4love4font10RasterizerD2Ev($this); //@line 44 "modules/font/ImageRasterizer.cpp"
  ___resumeException($$01|0);
  // unreachable;
 }
 $23 = HEAP32[$21>>2]|0; //@line 106 "./common/Object.h"
 $24 = ((($23)) + 12|0); //@line 106 "./common/Object.h"
 $25 = HEAP32[$24>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($25|0,($21|0)); //@line 106 "./common/Object.h"
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  $28 = ___cxa_find_matching_catch_3(0|0)|0;
  $29 = tempRet0;
  ___clang_call_terminate($28); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  $$0 = $19;$$01 = $18;
  __ZN4love4font10RasterizerD2Ev($this); //@line 44 "modules/font/ImageRasterizer.cpp"
  ___resumeException($$01|0);
  // unreachable;
 }
}
function __ZN4love4font15ImageRasterizer4loadEv($this) {
 $this = $this|0;
 var $$0$i7$i = 0, $$lcssa = 0, $$lcssa34 = 0, $$pre$i$i = 0, $$sink$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__nd$0$i$i = 0, $__nd$0$i$i$lcssa = 0;
 var $__nd$0$i$i$lcssa32 = 0, $__nd$0$i$i$lcssa33 = 0, $__parent$i = 0, $__r$0$i = 0, $end$019 = 0, $end$1$lcssa = 0, $end$114 = 0, $i$018 = 0, $lock = 0, $start$0$lcssa = 0, $start$06 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__parent$i = sp + 4|0;
 $lock = sp;
 $0 = ((($this)) + 24|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 97 "modules/font/ImageRasterizer.cpp"
 $3 = ((($2)) + 16|0); //@line 97 "modules/font/ImageRasterizer.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 97 "modules/font/ImageRasterizer.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($1)|0); //@line 97 "modules/font/ImageRasterizer.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $7 = (__ZNK4love5image9ImageData8getWidthEv($6)|0); //@line 99 "modules/font/ImageRasterizer.cpp"
 $8 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $9 = (__ZNK4love5image9ImageData9getHeightEv($8)|0); //@line 100 "modules/font/ImageRasterizer.cpp"
 $10 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $11 = (__ZNK4love5image9ImageData8getMutexEv($10)|0); //@line 103 "modules/font/ImageRasterizer.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$11); //@line 103 "modules/font/ImageRasterizer.cpp"
 $12 = ((($this)) + 20|0); //@line 106 "modules/font/ImageRasterizer.cpp"
 HEAP32[$12>>2] = $9; //@line 106 "modules/font/ImageRasterizer.cpp"
 $13 = ((($this)) + 52|0); //@line 109 "modules/font/ImageRasterizer.cpp"
 $14 = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8)|(HEAPU8[$5+2>>0]<<16)|(HEAPU8[$5+3>>0]<<24); //@line 109 "modules/font/ImageRasterizer.cpp"
 HEAP8[$13>>0]=$14&255;HEAP8[$13+1>>0]=($14>>8)&255;HEAP8[$13+2>>0]=($14>>16)&255;HEAP8[$13+3>>0]=$14>>24; //@line 109 "modules/font/ImageRasterizer.cpp"
 $15 = ((($this)) + 32|0); //@line 114 "modules/font/ImageRasterizer.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 114 "modules/font/ImageRasterizer.cpp"
 $17 = ($16|0)>(0); //@line 114 "modules/font/ImageRasterizer.cpp"
 if (!($17)) {
  __ZN4love6thread4LockD2Ev($lock); //@line 137 "modules/font/ImageRasterizer.cpp"
  STACKTOP = sp;return; //@line 137 "modules/font/ImageRasterizer.cpp"
 }
 $18 = ((($this)) + 40|0); //@line 135 "modules/font/ImageRasterizer.cpp"
 $19 = ((($this)) + 28|0); //@line 135 "modules/font/ImageRasterizer.cpp"
 $20 = ((($this)) + 44|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $21 = $20; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $22 = ((($this)) + 44|0); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $23 = ((($this)) + 48|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $24 = ((($this)) + 53|0); //@line 34 "modules/font/ImageRasterizer.cpp"
 $25 = ((($this)) + 54|0); //@line 34 "modules/font/ImageRasterizer.cpp"
 $26 = ((($this)) + 55|0); //@line 34 "modules/font/ImageRasterizer.cpp"
 $end$019 = 0;$i$018 = 0;
 while(1) {
  $27 = ($end$019|0)<($7|0); //@line 119 "modules/font/ImageRasterizer.cpp"
  L6: do {
   if ($27) {
    $28 = HEAP8[$13>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
    $start$06 = $end$019;
    while(1) {
     $29 = (($5) + ($start$06<<2)|0); //@line 34 "modules/font/ImageRasterizer.cpp"
     $30 = HEAP8[$29>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
     $31 = ($30<<24>>24)==($28<<24>>24); //@line 34 "modules/font/ImageRasterizer.cpp"
     if (!($31)) {
      $start$0$lcssa = $start$06;
      break L6;
     }
     $34 = (((($5) + ($start$06<<2)|0)) + 1|0); //@line 34 "modules/font/ImageRasterizer.cpp"
     $35 = HEAP8[$34>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
     $36 = HEAP8[$24>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
     $37 = ($35<<24>>24)==($36<<24>>24); //@line 34 "modules/font/ImageRasterizer.cpp"
     if (!($37)) {
      $start$0$lcssa = $start$06;
      break L6;
     }
     $38 = (((($5) + ($start$06<<2)|0)) + 2|0); //@line 34 "modules/font/ImageRasterizer.cpp"
     $39 = HEAP8[$38>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
     $40 = HEAP8[$25>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
     $41 = ($39<<24>>24)==($40<<24>>24); //@line 34 "modules/font/ImageRasterizer.cpp"
     if (!($41)) {
      $start$0$lcssa = $start$06;
      break L6;
     }
     $42 = (((($5) + ($start$06<<2)|0)) + 3|0); //@line 34 "modules/font/ImageRasterizer.cpp"
     $43 = HEAP8[$42>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
     $44 = HEAP8[$26>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
     $45 = ($43<<24>>24)==($44<<24>>24); //@line 34 "modules/font/ImageRasterizer.cpp"
     if (!($45)) {
      $start$0$lcssa = $start$06;
      break L6;
     }
     $46 = (($start$06) + 1)|0; //@line 120 "modules/font/ImageRasterizer.cpp"
     $47 = ($46|0)<($7|0); //@line 119 "modules/font/ImageRasterizer.cpp"
     if ($47) {
      $start$06 = $46;
     } else {
      $start$0$lcssa = $46;
      break;
     }
    }
   } else {
    $start$0$lcssa = $end$019;
   }
  } while(0);
  $32 = ($start$0$lcssa|0)<($7|0); //@line 125 "modules/font/ImageRasterizer.cpp"
  if (!($32)) {
   label = 35;
   break;
  }
  $33 = HEAP8[$13>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
  $end$114 = $start$0$lcssa;
  while(1) {
   $50 = (($5) + ($end$114<<2)|0); //@line 34 "modules/font/ImageRasterizer.cpp"
   $51 = HEAP8[$50>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
   $52 = ($51<<24>>24)==($33<<24>>24); //@line 34 "modules/font/ImageRasterizer.cpp"
   if ($52) {
    $53 = (((($5) + ($end$114<<2)|0)) + 1|0); //@line 34 "modules/font/ImageRasterizer.cpp"
    $54 = HEAP8[$53>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
    $55 = HEAP8[$24>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
    $56 = ($54<<24>>24)==($55<<24>>24); //@line 34 "modules/font/ImageRasterizer.cpp"
    if ($56) {
     $57 = (((($5) + ($end$114<<2)|0)) + 2|0); //@line 34 "modules/font/ImageRasterizer.cpp"
     $58 = HEAP8[$57>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
     $59 = HEAP8[$25>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
     $60 = ($58<<24>>24)==($59<<24>>24); //@line 34 "modules/font/ImageRasterizer.cpp"
     if ($60) {
      $61 = (((($5) + ($end$114<<2)|0)) + 3|0); //@line 34 "modules/font/ImageRasterizer.cpp"
      $62 = HEAP8[$61>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
      $63 = HEAP8[$26>>0]|0; //@line 34 "modules/font/ImageRasterizer.cpp"
      $64 = ($62<<24>>24)==($63<<24>>24); //@line 34 "modules/font/ImageRasterizer.cpp"
      if ($64) {
       $end$1$lcssa = $end$114;
       break;
      }
     }
    }
   }
   $65 = (($end$114) + 1)|0; //@line 126 "modules/font/ImageRasterizer.cpp"
   $66 = ($65|0)<($7|0); //@line 125 "modules/font/ImageRasterizer.cpp"
   if ($66) {
    $end$114 = $65;
   } else {
    $end$1$lcssa = $65;
    break;
   }
  }
  $67 = ($end$1$lcssa|0)>($start$0$lcssa|0); //@line 128 "modules/font/ImageRasterizer.cpp"
  if (!($67)) {
   label = 35;
   break;
  }
  $68 = HEAP32[$19>>2]|0; //@line 135 "modules/font/ImageRasterizer.cpp"
  $69 = (($68) + ($i$018<<2)|0); //@line 135 "modules/font/ImageRasterizer.cpp"
  $70 = HEAP32[$20>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $71 = ($70|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  do {
   if ($71) {
    HEAP32[$__parent$i>>2] = $20; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i = $20;$98 = $21;
    label = 30;
   } else {
    $72 = HEAP32[$69>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $__nd$0$i$i = $70;
    while(1) {
     $73 = ((($__nd$0$i$i)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $74 = HEAP32[$73>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     $75 = ($72>>>0)<($74>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if ($75) {
      $76 = HEAP32[$__nd$0$i$i>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
      $77 = ($76|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
      if ($77) {
       $$lcssa34 = $__nd$0$i$i;$__nd$0$i$i$lcssa33 = $__nd$0$i$i;
       label = 23;
       break;
      } else {
       $$sink$i$i = $76;
      }
     } else {
      $79 = ($74>>>0)<($72>>>0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
      if (!($79)) {
       $__nd$0$i$i$lcssa = $__nd$0$i$i;
       label = 29;
       break;
      }
      $80 = ((($__nd$0$i$i)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
      $81 = HEAP32[$80>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
      $82 = ($81|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
      if ($82) {
       $$lcssa = $80;$__nd$0$i$i$lcssa32 = $__nd$0$i$i;
       label = 27;
       break;
      } else {
       $$sink$i$i = $81;
      }
     }
     $__nd$0$i$i = $$sink$i$i;
    }
    if ((label|0) == 23) {
     label = 0;
     HEAP32[$__parent$i>>2] = $__nd$0$i$i$lcssa33; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $78 = $__nd$0$i$i$lcssa33; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $$0$i7$i = $$lcssa34;$98 = $78;
     label = 30;
     break;
    }
    else if ((label|0) == 27) {
     label = 0;
     HEAP32[$__parent$i>>2] = $__nd$0$i$i$lcssa32; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $83 = $__nd$0$i$i$lcssa32; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $$0$i7$i = $$lcssa;$98 = $83;
     label = 30;
     break;
    }
    else if ((label|0) == 29) {
     label = 0;
     HEAP32[$__parent$i>>2] = $__nd$0$i$i$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $84 = $__nd$0$i$i$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $85 = ($__nd$0$i$i$lcssa|0)==(0|0); //@line 1378 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($85) {
      $$0$i7$i = $__parent$i;$98 = $84;
      label = 30;
      break;
     } else {
      $__r$0$i = $__nd$0$i$i$lcssa;
      break;
     }
    }
   }
  } while(0);
  if ((label|0) == 30) {
   label = 0;
   __THREW__ = 0;
   $86 = (invoke_ii(3348,28)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $87 = __THREW__; __THREW__ = 0;
   $88 = $87&1;
   if ($88) {
    label = 12;
    break;
   }
   $89 = ((($86)) + 16|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $90 = HEAP32[$69>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$89>>2] = $90; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $91 = ((($86)) + 20|0); //@line 1366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $92 = $91; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $93 = $92; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$93>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $94 = (($92) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $95 = $94; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$95>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$86>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $96 = ((($86)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$96>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $97 = ((($86)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$97>>2] = $98; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$$0$i7$i>>2] = $86; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $99 = HEAP32[$18>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $100 = HEAP32[$99>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $101 = ($100|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($101) {
    $104 = $86;
   } else {
    $102 = $100; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP32[$18>>2] = $102; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $$pre$i$i = HEAP32[$$0$i7$i>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $104 = $$pre$i$i;
   }
   $103 = HEAP32[$22>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($103,$104); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $105 = HEAP32[$23>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $106 = (($105) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$23>>2] = $106; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $__r$0$i = $86;
  }
  $107 = (($end$1$lcssa) - ($start$0$lcssa))|0; //@line 133 "modules/font/ImageRasterizer.cpp"
  $108 = ((($__r$0$i)) + 20|0); //@line 135 "modules/font/ImageRasterizer.cpp"
  HEAP32[$108>>2] = $start$0$lcssa; //@line 135 "modules/font/ImageRasterizer.cpp"
  $109 = ((($__r$0$i)) + 24|0); //@line 135 "modules/font/ImageRasterizer.cpp"
  HEAP32[$109>>2] = $107; //@line 135 "modules/font/ImageRasterizer.cpp"
  $110 = (($i$018) + 1)|0; //@line 114 "modules/font/ImageRasterizer.cpp"
  $111 = HEAP32[$15>>2]|0; //@line 114 "modules/font/ImageRasterizer.cpp"
  $112 = ($110|0)<($111|0); //@line 114 "modules/font/ImageRasterizer.cpp"
  if ($112) {
   $end$019 = $end$1$lcssa;$i$018 = $110;
  } else {
   label = 35;
   break;
  }
 }
 if ((label|0) == 12) {
  $48 = ___cxa_find_matching_catch_2()|0;
  $49 = tempRet0;
  __ZN4love6thread4LockD2Ev($lock); //@line 137 "modules/font/ImageRasterizer.cpp"
  ___resumeException($48|0);
  // unreachable;
 }
 else if ((label|0) == 35) {
  __ZN4love6thread4LockD2Ev($lock); //@line 137 "modules/font/ImageRasterizer.cpp"
  STACKTOP = sp;return; //@line 137 "modules/font/ImageRasterizer.cpp"
 }
}
function __ZN4love4font10RasterizerD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 34 "modules/font/Rasterizer.cpp"
}
function __ZNK4love4font10Rasterizer9getHeightEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 38 "modules/font/Rasterizer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 38 "modules/font/Rasterizer.cpp"
 return ($1|0); //@line 38 "modules/font/Rasterizer.cpp"
}
function __ZNK4love4font10Rasterizer10getAdvanceEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 43 "modules/font/Rasterizer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 43 "modules/font/Rasterizer.cpp"
 return ($1|0); //@line 43 "modules/font/Rasterizer.cpp"
}
function __ZNK4love4font10Rasterizer9getAscentEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 48 "modules/font/Rasterizer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 48 "modules/font/Rasterizer.cpp"
 return ($1|0); //@line 48 "modules/font/Rasterizer.cpp"
}
function __ZNK4love4font10Rasterizer10getDescentEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 53 "modules/font/Rasterizer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 53 "modules/font/Rasterizer.cpp"
 return ($1|0); //@line 53 "modules/font/Rasterizer.cpp"
}
function __ZNK4love4font10Rasterizer12getGlyphDataERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE($this,$text) {
 $this = $this|0;
 $text = $text|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 16|0;
 $vararg_buffer = sp + 8|0;
 $0 = sp + 12|0;
 $1 = sp;
 $2 = HEAP8[$text>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = $2 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ($3<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($text)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = HEAP32[$5>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($text)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $4 ? $7 : $6; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = ((($text)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = HEAP32[$9>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $2&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = $11 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = $4 ? $12 : $10; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = (($8) + ($13)|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $15 = $14; //@line 62 "modules/font/Rasterizer.cpp"
 HEAP32[$1>>2] = $8; //@line 62 "modules/font/Rasterizer.cpp"
 HEAP32[$0>>2] = $15; //@line 159 "./libraries/utf8/utf8/checked.h"
 __THREW__ = 0;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0; //@line 159 "./libraries/utf8/utf8/checked.h"
 $16 = (invoke_iii(3997,($1|0),($$byval_copy|0))|0); //@line 159 "./libraries/utf8/utf8/checked.h"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if (!($18)) {
  $19 = HEAP32[$this>>2]|0; //@line 69 "modules/font/Rasterizer.cpp"
  $20 = ((($19)) + 36|0); //@line 69 "modules/font/Rasterizer.cpp"
  $21 = HEAP32[$20>>2]|0; //@line 69 "modules/font/Rasterizer.cpp"
  $22 = (FUNCTION_TABLE_iii[$21 & 8191]($this,$16)|0); //@line 69 "modules/font/Rasterizer.cpp"
  STACKTOP = sp;return ($22|0); //@line 69 "modules/font/Rasterizer.cpp"
 }
 $23 = ___cxa_find_matching_catch_3(1744|0)|0;
 $24 = tempRet0;
 $25 = (_llvm_eh_typeid_for((1744|0))|0); //@line 63 "modules/font/Rasterizer.cpp"
 $26 = ($24|0)==($25|0); //@line 63 "modules/font/Rasterizer.cpp"
 if (!($26)) {
  $$1 = $24;$$12 = $23;
  ___resumeException($$12|0);
  // unreachable;
 }
 $27 = (___cxa_begin_catch(($23|0))|0); //@line 63 "modules/font/Rasterizer.cpp"
 $28 = (___cxa_allocate_exception(16)|0); //@line 66 "modules/font/Rasterizer.cpp"
 $29 = HEAP32[$27>>2]|0; //@line 66 "modules/font/Rasterizer.cpp"
 $30 = ((($29)) + 8|0); //@line 66 "modules/font/Rasterizer.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 66 "modules/font/Rasterizer.cpp"
 $32 = (FUNCTION_TABLE_ii[$31 & 8191]($27)|0); //@line 66 "modules/font/Rasterizer.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $32; //@line 66 "modules/font/Rasterizer.cpp"
 invoke_viii(3349,($28|0),(434605|0),($vararg_buffer|0)); //@line 66 "modules/font/Rasterizer.cpp"
 $33 = __THREW__; __THREW__ = 0;
 $34 = $33&1;
 if ($34) {
  $38 = ___cxa_find_matching_catch_2()|0;
  $39 = tempRet0;
  ___cxa_free_exception(($28|0)); //@line 66 "modules/font/Rasterizer.cpp"
  $$0 = $39;$$01 = $38;
 } else {
  __THREW__ = 0;
  invoke_viii(3350,($28|0),(96|0),(344|0)); //@line 66 "modules/font/Rasterizer.cpp"
  $35 = __THREW__; __THREW__ = 0;
  $36 = ___cxa_find_matching_catch_2()|0;
  $37 = tempRet0;
  $$0 = $37;$$01 = $36;
 }
 __THREW__ = 0;
 invoke_v(3378); //@line 67 "modules/font/Rasterizer.cpp"
 $40 = __THREW__; __THREW__ = 0;
 $41 = $40&1;
 if ($41) {
  $42 = ___cxa_find_matching_catch_3(0|0)|0;
  $43 = tempRet0;
  ___clang_call_terminate($42); //@line 67 "modules/font/Rasterizer.cpp"
  // unreachable; //@line 67 "modules/font/Rasterizer.cpp"
 } else {
  $$1 = $$0;$$12 = $$01;
  ___resumeException($$12|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZNK4love4font10Rasterizer9hasGlyphsERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE($this,$text) {
 $this = $this|0;
 $text = $text|0;
 var $$03 = 0, $$05 = 0, $$12 = 0, $$14 = 0, $$16 = 0, $$byval_copy = 0, $$byval_copy10 = 0, $$pre = 0, $$pre30 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index4Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i = 0, $or$cond29 = 0, $temp$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy10 = sp + 32|0;
 $$byval_copy = sp + 28|0;
 $vararg_buffer = sp;
 $0 = sp + 24|0;
 $temp$i = sp + 20|0;
 $1 = sp + 16|0;
 $i = sp + 4|0;
 $2 = HEAP8[$text>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = $2 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ($3<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($text)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = HEAP32[$5>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = $2&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $7 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $4 ? $8 : $6; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = ($9|0)==(0); //@line 74 "modules/font/Rasterizer.cpp"
 if ($10) {
  STACKTOP = sp;return 0; //@line 96 "modules/font/Rasterizer.cpp"
 }
 $11 = ((($text)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = HEAP32[$11>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ((($text)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = $4 ? $13 : $12; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $15 = $14; //@line 79 "modules/font/Rasterizer.cpp"
 $16 = (($14) + ($9)|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $17 = $16; //@line 79 "modules/font/Rasterizer.cpp"
 HEAP32[$i>>2] = $15; //@line 277 "./libraries/utf8/utf8/checked.h"
 $18 = ((($i)) + 4|0); //@line 277 "./libraries/utf8/utf8/checked.h"
 HEAP32[$18>>2] = $15; //@line 277 "./libraries/utf8/utf8/checked.h"
 $19 = ((($i)) + 8|0); //@line 277 "./libraries/utf8/utf8/checked.h"
 HEAP32[$19>>2] = $17; //@line 277 "./libraries/utf8/utf8/checked.h"
 $20 = ($9|0)<(0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 do {
  if ($20) {
   $21 = (___cxa_allocate_exception(8)|0); //@line 280 "./libraries/utf8/utf8/checked.h"
   __THREW__ = 0;
   invoke_vii(3996,($21|0),(434517|0)); //@line 133 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/stdexcept"
   $22 = __THREW__; __THREW__ = 0;
   $23 = $22&1;
   if ($23) {
    $25 = ___cxa_find_matching_catch_3(1744|0)|0;
    $26 = tempRet0;
    ___cxa_free_exception(($21|0)); //@line 280 "./libraries/utf8/utf8/checked.h"
    $eh$lpad$body$index4Z2D = $26;$eh$lpad$body$indexZ2D = $25;
    break;
   } else {
    HEAP32[$21>>2] = (157912); //@line 133 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/stdexcept"
    __THREW__ = 0;
    invoke_viii(3350,($21|0),(9120|0),(3328|0)); //@line 280 "./libraries/utf8/utf8/checked.h"
    $24 = __THREW__; __THREW__ = 0;
    label = 17;
    break;
   }
  } else {
   $27 = ((($i)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $28 = ((($i)) + 8|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $38 = $17;
   while(1) {
    $35 = HEAP32[$i>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $36 = ($35|0)==($16|0); //@line 297 "./libraries/utf8/utf8/checked.h"
    $37 = $35; //@line 306 "./libraries/utf8/utf8/checked.h"
    if ($36) {
     $$12 = 1;
     label = 21;
     break;
    }
    HEAP32[$0>>2] = $38; //@line 307 "./libraries/utf8/utf8/checked.h"
    __THREW__ = 0;
    ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0; //@line 307 "./libraries/utf8/utf8/checked.h"
    (invoke_iii(3997,($i|0),($$byval_copy|0))|0); //@line 307 "./libraries/utf8/utf8/checked.h"
    $39 = __THREW__; __THREW__ = 0;
    $40 = $39&1;
    if ($40) {
     label = 16;
     break;
    }
    HEAP32[$temp$i>>2] = $37; //@line 286 "./libraries/utf8/utf8/checked.h"
    HEAP32[$1>>2] = $38; //@line 287 "./libraries/utf8/utf8/checked.h"
    __THREW__ = 0;
    ;HEAP32[$$byval_copy10>>2]=HEAP32[$1>>2]|0; //@line 287 "./libraries/utf8/utf8/checked.h"
    $41 = (invoke_iii(3997,($temp$i|0),($$byval_copy10|0))|0); //@line 287 "./libraries/utf8/utf8/checked.h"
    $42 = __THREW__; __THREW__ = 0;
    $43 = $42&1;
    if ($43) {
     label = 16;
     break;
    }
    $44 = HEAP32[$this>>2]|0; //@line 86 "modules/font/Rasterizer.cpp"
    $45 = ((($44)) + 48|0); //@line 86 "modules/font/Rasterizer.cpp"
    $46 = HEAP32[$45>>2]|0; //@line 86 "modules/font/Rasterizer.cpp"
    __THREW__ = 0;
    $47 = (invoke_iii($46|0,($this|0),($41|0))|0); //@line 86 "modules/font/Rasterizer.cpp"
    $48 = __THREW__; __THREW__ = 0;
    $49 = $48&1;
    if ($49) {
     label = 16;
     break;
    }
    if (!($47)) {
     $$12 = 0;
     label = 21;
     break;
    }
    $$pre = HEAP32[$27>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $$pre30 = HEAP32[$28>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $50 = $$pre30;
    $51 = ($$pre|0)==($14|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $52 = ($$pre30|0)==($16|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $or$cond29 = $51 & $52; //@line 291 "./libraries/utf8/utf8/checked.h"
    if ($or$cond29) {
     $38 = $50;
    } else {
     label = 7;
     break;
    }
   }
   if ((label|0) == 7) {
    $29 = (___cxa_allocate_exception(8)|0); //@line 292 "./libraries/utf8/utf8/checked.h"
    __THREW__ = 0;
    invoke_vii(3996,($29|0),(434549|0)); //@line 292 "./libraries/utf8/utf8/checked.h"
    $30 = __THREW__; __THREW__ = 0;
    $31 = $30&1;
    if ($31) {
     $33 = ___cxa_find_matching_catch_3(1744|0)|0;
     $34 = tempRet0;
     ___cxa_free_exception(($29|0)); //@line 292 "./libraries/utf8/utf8/checked.h"
     $eh$lpad$body$index4Z2D = $34;$eh$lpad$body$indexZ2D = $33;
     break;
    } else {
     __THREW__ = 0;
     invoke_viii(3350,($29|0),(9072|0),(3328|0)); //@line 292 "./libraries/utf8/utf8/checked.h"
     $32 = __THREW__; __THREW__ = 0;
     label = 17;
     break;
    }
   }
   else if ((label|0) == 16) {
    $53 = ___cxa_find_matching_catch_3(1744|0)|0;
    $54 = tempRet0;
    $eh$lpad$body$index4Z2D = $54;$eh$lpad$body$indexZ2D = $53;
    break;
   }
   else if ((label|0) == 21) {
    STACKTOP = sp;return ($$12|0);
   }
  }
 } while(0);
 if ((label|0) == 17) {
  $55 = ___cxa_find_matching_catch_3(1744|0)|0;
  $56 = tempRet0;
  $eh$lpad$body$index4Z2D = $56;$eh$lpad$body$indexZ2D = $55;
 }
 $57 = (_llvm_eh_typeid_for((1744|0))|0); //@line 89 "modules/font/Rasterizer.cpp"
 $58 = ($eh$lpad$body$index4Z2D|0)==($57|0); //@line 89 "modules/font/Rasterizer.cpp"
 if (!($58)) {
  $$14 = $eh$lpad$body$indexZ2D;$$16 = $eh$lpad$body$index4Z2D;
  ___resumeException($$14|0);
  // unreachable;
 }
 $59 = (___cxa_begin_catch(($eh$lpad$body$indexZ2D|0))|0); //@line 89 "modules/font/Rasterizer.cpp"
 $60 = (___cxa_allocate_exception(16)|0); //@line 92 "modules/font/Rasterizer.cpp"
 $61 = HEAP32[$59>>2]|0; //@line 92 "modules/font/Rasterizer.cpp"
 $62 = ((($61)) + 8|0); //@line 92 "modules/font/Rasterizer.cpp"
 $63 = HEAP32[$62>>2]|0; //@line 92 "modules/font/Rasterizer.cpp"
 $64 = (FUNCTION_TABLE_ii[$63 & 8191]($59)|0); //@line 92 "modules/font/Rasterizer.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $64; //@line 92 "modules/font/Rasterizer.cpp"
 invoke_viii(3349,($60|0),(434605|0),($vararg_buffer|0)); //@line 92 "modules/font/Rasterizer.cpp"
 $65 = __THREW__; __THREW__ = 0;
 $66 = $65&1;
 if ($66) {
  $70 = ___cxa_find_matching_catch_2()|0;
  $71 = tempRet0;
  ___cxa_free_exception(($60|0)); //@line 92 "modules/font/Rasterizer.cpp"
  $$03 = $70;$$05 = $71;
 } else {
  __THREW__ = 0;
  invoke_viii(3350,($60|0),(96|0),(344|0)); //@line 92 "modules/font/Rasterizer.cpp"
  $67 = __THREW__; __THREW__ = 0;
  $68 = ___cxa_find_matching_catch_2()|0;
  $69 = tempRet0;
  $$03 = $68;$$05 = $69;
 }
 __THREW__ = 0;
 invoke_v(3378); //@line 93 "modules/font/Rasterizer.cpp"
 $72 = __THREW__; __THREW__ = 0;
 $73 = $72&1;
 if ($73) {
  $74 = ___cxa_find_matching_catch_3(0|0)|0;
  $75 = tempRet0;
  ___clang_call_terminate($74); //@line 93 "modules/font/Rasterizer.cpp"
  // unreachable; //@line 93 "modules/font/Rasterizer.cpp"
 } else {
  $$14 = $$03;$$16 = $$05;
  ___resumeException($$14|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZNK4love4font10Rasterizer10getKerningEjj($this,$0,$1) {
 $this = $this|0;
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return +0; //@line 100 "modules/font/Rasterizer.cpp"
}
function __GLOBAL__sub_I_TrueTypeRasterizer_cpp() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[(632984)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(632996)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633008)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633020)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633032)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633044)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633056)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633068)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(633072)>>2]=0|0;HEAP32[(633072)+4>>2]=0|0;HEAP32[(633072)+8>>2]=0|0;HEAP32[(633072)+12>>2]=0|0; //@line 44 "./common/StringMap.h"
 $0 = HEAP32[6069]|0; //@line 49 "./common/StringMap.h"
 $1 = HEAP32[(24280)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_4font18TrueTypeRasterizer7HintingELj4EE3addEPKcS3_(632976,$0,$1)|0); //@line 49 "./common/StringMap.h"
 $2 = HEAP32[(24284)>>2]|0; //@line 49 "./common/StringMap.h"
 $3 = HEAP32[(24288)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_4font18TrueTypeRasterizer7HintingELj4EE3addEPKcS3_(632976,$2,$3)|0); //@line 49 "./common/StringMap.h"
 $4 = HEAP32[(24292)>>2]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(24296)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_4font18TrueTypeRasterizer7HintingELj4EE3addEPKcS3_(632976,$4,$5)|0); //@line 49 "./common/StringMap.h"
 $6 = HEAP32[(24300)>>2]|0; //@line 49 "./common/StringMap.h"
 $7 = HEAP32[(24304)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_4font18TrueTypeRasterizer7HintingELj4EE3addEPKcS3_(632976,$6,$7)|0); //@line 49 "./common/StringMap.h"
 return;
}
function __ZN4love9StringMapINS_4font18TrueTypeRasterizer7HintingELj4EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0;
 var $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 7; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = $13 & 7; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = $22 & 7; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = $27 & 7; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = $32 & 7; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = $37 & 7; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $42 = (($hash$0$lcssa$i) + 6)|0; //@line 112 "./common/StringMap.h"
       $43 = $42 & 7; //@line 112 "./common/StringMap.h"
       $44 = (((($this) + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
       $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
       $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
       if ($46) {
        $$lcssa = $43;$$lcssa10 = $44;
        label = 5;
       } else {
        $47 = (($hash$0$lcssa$i) + 7)|0; //@line 112 "./common/StringMap.h"
        $48 = $47 & 7; //@line 112 "./common/StringMap.h"
        $49 = (((($this) + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
        $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
        $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
        if ($51) {
         $$lcssa = $48;$$lcssa10 = $49;
         label = 5;
        } else {
         $inserted$2$off0 = 0;
        }
       }
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(3); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 96|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love4font18TrueTypeRasterizer11getConstantEPKcRNS1_7HintingE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = $9 & 7; //@line 72 "./common/StringMap.h"
  $11 = (((632976 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (632976 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(8); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((632976 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 30 "modules/font/TrueTypeRasterizer.cpp"
  return ($$$2$i|0); //@line 30 "modules/font/TrueTypeRasterizer.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 30 "modules/font/TrueTypeRasterizer.cpp"
  return ($$$2$i|0); //@line 30 "modules/font/TrueTypeRasterizer.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl6CanvasD2Ev($this) {
 $this = $this|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i1 = 0, $scevgep4$i$i$i = 0;
 var $scevgep4$i$i$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (24316); //@line 144 "modules/graphics/opengl/Canvas.cpp"
 $0 = ((($this)) + 120|0); //@line 144 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$0>>2] = (24388); //@line 144 "modules/graphics/opengl/Canvas.cpp"
 $1 = HEAP32[158277]|0; //@line 145 "modules/graphics/opengl/Canvas.cpp"
 $2 = (($1) + -1)|0; //@line 145 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[158277] = $2; //@line 145 "modules/graphics/opengl/Canvas.cpp"
 $3 = HEAP32[158272]|0; //@line 148 "modules/graphics/opengl/Canvas.cpp"
 $4 = ($3|0)==($this|0); //@line 148 "modules/graphics/opengl/Canvas.cpp"
 if ($4) {
  __THREW__ = 0;
  invoke_vii(4001,($this|0),0); //@line 149 "modules/graphics/opengl/Canvas.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   $$pre = HEAP32[$this>>2]|0; //@line 151 "modules/graphics/opengl/Canvas.cpp"
   $21 = $$pre;
   label = 9;
  }
 } else {
  $21 = (24316);
  label = 9;
 }
 if ((label|0) == 9) {
  $20 = ((($21)) + 60|0); //@line 151 "modules/graphics/opengl/Canvas.cpp"
  $22 = HEAP32[$20>>2]|0; //@line 151 "modules/graphics/opengl/Canvas.cpp"
  __THREW__ = 0;
  invoke_vi($22|0,($this|0)); //@line 151 "modules/graphics/opengl/Canvas.cpp"
  $23 = __THREW__; __THREW__ = 0;
  $24 = $23&1;
  if (!($24)) {
   $25 = ((($this)) + 152|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $26 = HEAP32[$25>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $27 = ($26|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $28 = $26; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($27) {
    __ZN4love8graphics8VolatileD2Ev($0); //@line 152 "modules/graphics/opengl/Canvas.cpp"
    __ZN4love8graphics7TextureD2Ev($this); //@line 152 "modules/graphics/opengl/Canvas.cpp"
    return; //@line 152 "modules/graphics/opengl/Canvas.cpp"
   }
   $29 = ((($this)) + 156|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $30 = HEAP32[$29>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $31 = ($30|0)==($26|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($31)) {
    $scevgep$i$i$i1 = ((($30)) + -4|0);
    $32 = $scevgep$i$i$i1;
    $33 = (($32) - ($28))|0;
    $34 = $33 >>> 2;
    $35 = $34 ^ -1;
    $scevgep4$i$i$i2 = (($30) + ($35<<2)|0);
    HEAP32[$29>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($26); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   __ZN4love8graphics8VolatileD2Ev($0); //@line 152 "modules/graphics/opengl/Canvas.cpp"
   __ZN4love8graphics7TextureD2Ev($this); //@line 152 "modules/graphics/opengl/Canvas.cpp"
   return; //@line 152 "modules/graphics/opengl/Canvas.cpp"
  }
 }
 $7 = ___cxa_find_matching_catch_3(0|0)|0;
 $8 = tempRet0;
 $9 = ((($this)) + 152|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = HEAP32[$9>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = $10; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  __ZN4love8graphics8VolatileD2Ev($0); //@line 152 "modules/graphics/opengl/Canvas.cpp"
  __ZN4love8graphics7TextureD2Ev($this); //@line 152 "modules/graphics/opengl/Canvas.cpp"
  ___clang_call_terminate($7); //@line 152 "modules/graphics/opengl/Canvas.cpp"
  // unreachable; //@line 152 "modules/graphics/opengl/Canvas.cpp"
 }
 $13 = ((($this)) + 156|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = HEAP32[$13>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = ($14|0)==($10|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($15)) {
  $scevgep$i$i$i = ((($14)) + -4|0);
  $16 = $scevgep$i$i$i;
  $17 = (($16) - ($12))|0;
  $18 = $17 >>> 2;
  $19 = $18 ^ -1;
  $scevgep4$i$i$i = (($14) + ($19<<2)|0);
  HEAP32[$13>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($10); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 __ZN4love8graphics8VolatileD2Ev($0); //@line 152 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics7TextureD2Ev($this); //@line 152 "modules/graphics/opengl/Canvas.cpp"
 ___clang_call_terminate($7); //@line 152 "modules/graphics/opengl/Canvas.cpp"
 // unreachable; //@line 152 "modules/graphics/opengl/Canvas.cpp"
}
function __ZN4love8graphics6opengl6Canvas8stopGrabEb($this,$switchingToOtherCanvas) {
 $this = $this|0;
 $switchingToOtherCanvas = $switchingToOtherCanvas|0;
 var $$pr = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[158272]|0; //@line 516 "modules/graphics/opengl/Canvas.cpp"
 $1 = ($0|0)==($this|0); //@line 516 "modules/graphics/opengl/Canvas.cpp"
 if (!($1)) {
  return; //@line 541 "modules/graphics/opengl/Canvas.cpp"
 }
 $2 = ((($this)) + 128|0); //@line 635 "modules/graphics/opengl/Canvas.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 635 "modules/graphics/opengl/Canvas.cpp"
 $4 = ($3|0)==(0); //@line 635 "modules/graphics/opengl/Canvas.cpp"
 do {
  if (!($4)) {
   $5 = ((($this)) + 136|0); //@line 635 "modules/graphics/opengl/Canvas.cpp"
   $6 = HEAP32[$5>>2]|0; //@line 635 "modules/graphics/opengl/Canvas.cpp"
   $7 = ($6|0)==(0); //@line 635 "modules/graphics/opengl/Canvas.cpp"
   if (!($7)) {
    $8 = ((($this)) + 8|0); //@line 640 "modules/graphics/opengl/Canvas.cpp"
    $9 = HEAP32[$8>>2]|0; //@line 640 "modules/graphics/opengl/Canvas.cpp"
    $10 = ((($this)) + 12|0); //@line 641 "modules/graphics/opengl/Canvas.cpp"
    $11 = HEAP32[$10>>2]|0; //@line 641 "modules/graphics/opengl/Canvas.cpp"
    $12 = ((($this)) + 124|0); //@line 647 "modules/graphics/opengl/Canvas.cpp"
    $13 = HEAP32[$12>>2]|0; //@line 647 "modules/graphics/opengl/Canvas.cpp"
    __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36008,$13); //@line 647 "modules/graphics/opengl/Canvas.cpp"
    $14 = HEAP32[$2>>2]|0; //@line 648 "modules/graphics/opengl/Canvas.cpp"
    __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36009,$14); //@line 648 "modules/graphics/opengl/Canvas.cpp"
    $15 = HEAP8[1252362]|0; //@line 650 "modules/graphics/opengl/Canvas.cpp"
    $16 = ($15<<24>>24)==(0); //@line 650 "modules/graphics/opengl/Canvas.cpp"
    if ($16) {
     $18 = HEAP32[164725]|0; //@line 677 "./libraries/glad/gladfuncs.hpp"
     FUNCTION_TABLE_viiiiiiiiii[$18 & 8191](0,0,$9,$11,0,0,$9,$11,16384,9728); //@line 677 "./libraries/glad/gladfuncs.hpp"
     break;
    } else {
     $17 = HEAP32[166329]|0; //@line 1996 "./libraries/glad/gladfuncs.hpp"
     FUNCTION_TABLE_v[$17 & 8191](); //@line 1996 "./libraries/glad/gladfuncs.hpp"
     break;
    }
   }
  }
 } while(0);
 $19 = HEAP32[(634060)>>2]|0; //@line 1666 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = ((($19)) + -64|0); //@line 1666 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $22 = $19;
 while(1) {
  $21 = ((($22)) + -64|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[(634060)>>2] = $21; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  __ZN4love7Matrix4D2Ev($21); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = HEAP32[(634060)>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $24 = ($23|0)==($20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($24) {
   break;
  } else {
   $22 = $23;
  }
 }
 if ($switchingToOtherCanvas) {
  return; //@line 541 "modules/graphics/opengl/Canvas.cpp"
 }
 $25 = (__ZNK4love8graphics6opengl6OpenGL13getDefaultFBOEv(634044)|0); //@line 529 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$25); //@line 529 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[158272] = 0; //@line 530 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL11setViewportERKNS2_8ViewportE(634044,633092); //@line 531 "modules/graphics/opengl/Canvas.cpp"
 $26 = HEAP8[1251994]|0; //@line 533 "modules/graphics/opengl/Canvas.cpp"
 $27 = HEAP8[1252225]|0; //@line 533 "modules/graphics/opengl/Canvas.cpp"
 $28 = $27 | $26; //@line 533 "modules/graphics/opengl/Canvas.cpp"
 $29 = ($28<<24>>24)==(0); //@line 533 "modules/graphics/opengl/Canvas.cpp"
 if ($29) {
  return; //@line 541 "modules/graphics/opengl/Canvas.cpp"
 }
 $30 = HEAP8[1251895]|0; //@line 535 "modules/graphics/opengl/Canvas.cpp"
 $31 = ($30<<24>>24)==(0); //@line 535 "modules/graphics/opengl/Canvas.cpp"
 do {
  if (!($31)) {
   $32 = (__ZNK4love8graphics6opengl6OpenGL18hasFramebufferSRGBEv(634044)|0); //@line 535 "modules/graphics/opengl/Canvas.cpp"
   if ($32) {
    $$pr = HEAP8[1251895]|0; //@line 537 "modules/graphics/opengl/Canvas.cpp"
    $33 = ($$pr<<24>>24)==(0); //@line 537 "modules/graphics/opengl/Canvas.cpp"
    if ($33) {
     break;
    }
    return; //@line 541 "modules/graphics/opengl/Canvas.cpp"
   } else {
    __ZN4love8graphics6opengl6OpenGL18setFramebufferSRGBEb(634044,1); //@line 536 "modules/graphics/opengl/Canvas.cpp"
    return; //@line 541 "modules/graphics/opengl/Canvas.cpp"
   }
  }
 } while(0);
 $34 = (__ZNK4love8graphics6opengl6OpenGL18hasFramebufferSRGBEv(634044)|0); //@line 537 "modules/graphics/opengl/Canvas.cpp"
 if (!($34)) {
  return; //@line 541 "modules/graphics/opengl/Canvas.cpp"
 }
 __ZN4love8graphics6opengl6OpenGL18setFramebufferSRGBEb(634044,0); //@line 538 "modules/graphics/opengl/Canvas.cpp"
 return; //@line 541 "modules/graphics/opengl/Canvas.cpp"
}
function __ZN4love8graphics6opengl6CanvasD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl6CanvasD2Ev($this); //@line 144 "modules/graphics/opengl/Canvas.cpp"
 __ZdlPv($this); //@line 144 "modules/graphics/opengl/Canvas.cpp"
 return; //@line 152 "modules/graphics/opengl/Canvas.cpp"
}
function __ZN4love8graphics6opengl6Canvas4drawEfffffffff($this,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $t = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $t = sp;
 __ZN4love7Matrix4C2Efffffffff($t,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky); //@line 327 "modules/graphics/opengl/Canvas.cpp"
 $0 = ((($this)) + 40|0); //@line 329 "modules/graphics/opengl/Canvas.cpp"
 __THREW__ = 0;
 invoke_viii(4002,($this|0),($t|0),($0|0)); //@line 329 "modules/graphics/opengl/Canvas.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZN4love7Matrix4D2Ev($t); //@line 330 "modules/graphics/opengl/Canvas.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  __ZN4love7Matrix4D2Ev($t); //@line 330 "modules/graphics/opengl/Canvas.cpp"
  STACKTOP = sp;return; //@line 330 "modules/graphics/opengl/Canvas.cpp"
 }
}
function __ZN4love8graphics6opengl6Canvas5drawvERKNS_7Matrix4EPKNS_6VertexE($this,$t,$v) {
 $this = $this|0;
 $t = $t|0;
 $v = $v|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP32[158272]|0; //@line 306 "modules/graphics/opengl/Canvas.cpp"
 $1 = ($0|0)==($this|0); //@line 306 "modules/graphics/opengl/Canvas.cpp"
 if ($1) {
  $2 = (___cxa_allocate_exception(16)|0); //@line 307 "modules/graphics/opengl/Canvas.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($2|0),(434046|0),($vararg_buffer|0)); //@line 307 "modules/graphics/opengl/Canvas.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if (!($4)) {
   ___cxa_throw(($2|0),(96|0),(344|0)); //@line 307 "modules/graphics/opengl/Canvas.cpp"
   // unreachable; //@line 307 "modules/graphics/opengl/Canvas.cpp"
  }
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  ___cxa_free_exception(($2|0)); //@line 307 "modules/graphics/opengl/Canvas.cpp"
  $$0 = $6;$$01 = $5;
  ___resumeException($$01|0);
  // unreachable;
 }
 __ZN4love8graphics6opengl6OpenGL13pushTransformEv(634044); //@line 125 "modules/graphics/opengl/OpenGL.h"
 __THREW__ = 0;
 $7 = (invoke_ii(4003,(634044|0))|0); //@line 135 "modules/graphics/opengl/OpenGL.h"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if (!($9)) {
  __THREW__ = 0;
  invoke_vii(4004,($7|0),($t|0)); //@line 312 "modules/graphics/opengl/Canvas.cpp"
  $10 = __THREW__; __THREW__ = 0;
  $11 = $10&1;
  if (!($11)) {
   $12 = ((($this)) + 132|0); //@line 314 "modules/graphics/opengl/Canvas.cpp"
   $13 = HEAP32[$12>>2]|0; //@line 314 "modules/graphics/opengl/Canvas.cpp"
   __THREW__ = 0;
   invoke_vii(4005,(634044|0),($13|0)); //@line 314 "modules/graphics/opengl/Canvas.cpp"
   $14 = __THREW__; __THREW__ = 0;
   $15 = $14&1;
   if (!($15)) {
    __THREW__ = 0;
    invoke_vii(4006,(634044|0),3); //@line 316 "modules/graphics/opengl/Canvas.cpp"
    $16 = __THREW__; __THREW__ = 0;
    $17 = $16&1;
    if (!($17)) {
     $18 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_viiiiii($18|0,0,2,5126,0,20,($v|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
     $19 = __THREW__; __THREW__ = 0;
     $20 = $19&1;
     if (!($20)) {
      $21 = ((($v)) + 8|0); //@line 319 "modules/graphics/opengl/Canvas.cpp"
      $22 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_viiiiii($22|0,1,2,5126,0,20,($21|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
      $23 = __THREW__; __THREW__ = 0;
      $24 = $23&1;
      if (!($24)) {
       __THREW__ = 0;
       invoke_vi(4007,(634044|0)); //@line 321 "modules/graphics/opengl/Canvas.cpp"
       $25 = __THREW__; __THREW__ = 0;
       $26 = $25&1;
       if (!($26)) {
        __THREW__ = 0;
        invoke_viiii(4008,(634044|0),5,0,4); //@line 322 "modules/graphics/opengl/Canvas.cpp"
        $27 = __THREW__; __THREW__ = 0;
        $28 = $27&1;
        if (!($28)) {
         __THREW__ = 0;
         invoke_vi(4009,(634044|0)); //@line 130 "modules/graphics/opengl/OpenGL.h"
         $29 = __THREW__; __THREW__ = 0;
         $30 = $29&1;
         if ($30) {
          $31 = ___cxa_find_matching_catch_3(0|0)|0;
          $32 = tempRet0;
          ___clang_call_terminate($31); //@line 130 "modules/graphics/opengl/OpenGL.h"
          // unreachable; //@line 130 "modules/graphics/opengl/OpenGL.h"
         } else {
          STACKTOP = sp;return; //@line 323 "modules/graphics/opengl/Canvas.cpp"
         }
        }
       }
      }
     }
    }
   }
  }
 }
 $33 = ___cxa_find_matching_catch_2()|0;
 $34 = tempRet0;
 __THREW__ = 0;
 invoke_vi(4009,(634044|0)); //@line 130 "modules/graphics/opengl/OpenGL.h"
 $35 = __THREW__; __THREW__ = 0;
 $36 = $35&1;
 if ($36) {
  $37 = ___cxa_find_matching_catch_3(0|0)|0;
  $38 = tempRet0;
  ___clang_call_terminate($37); //@line 130 "modules/graphics/opengl/OpenGL.h"
  // unreachable; //@line 130 "modules/graphics/opengl/OpenGL.h"
 } else {
  $$0 = $34;$$01 = $33;
  ___resumeException($$01|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl6Canvas5drawqEPNS0_4QuadEfffffffff($this,$quad,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $quad = $quad|0;
 $x = +$x;
 $y = +$y;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $t = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $t = sp;
 __ZN4love7Matrix4C2Efffffffff($t,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky); //@line 334 "modules/graphics/opengl/Canvas.cpp"
 __THREW__ = 0;
 $0 = (invoke_ii(4010,($quad|0))|0); //@line 336 "modules/graphics/opengl/Canvas.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if (!($2)) {
  __THREW__ = 0;
  invoke_viii(4002,($this|0),($t|0),($0|0)); //@line 336 "modules/graphics/opengl/Canvas.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if (!($4)) {
   __ZN4love7Matrix4D2Ev($t); //@line 337 "modules/graphics/opengl/Canvas.cpp"
   STACKTOP = sp;return; //@line 337 "modules/graphics/opengl/Canvas.cpp"
  }
 }
 $5 = ___cxa_find_matching_catch_2()|0;
 $6 = tempRet0;
 __ZN4love7Matrix4D2Ev($t); //@line 337 "modules/graphics/opengl/Canvas.cpp"
 ___resumeException($5|0);
 // unreachable;
}
function __ZN4love8graphics6opengl6Canvas9setFilterERKNS0_7Texture6FilterE($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (__ZN4love8graphics7Texture14validateFilterERKNS1_6FilterEb($f,0)|0); //@line 341 "modules/graphics/opengl/Canvas.cpp"
 if ($0) {
  $6 = ((($this)) + 16|0); //@line 344 "modules/graphics/opengl/Canvas.cpp"
  ;HEAP32[$6>>2]=HEAP32[$f>>2]|0;HEAP32[$6+4>>2]=HEAP32[$f+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$f+8>>2]|0;HEAP32[$6+12>>2]=HEAP32[$f+12>>2]|0; //@line 344 "modules/graphics/opengl/Canvas.cpp"
  $7 = ((($this)) + 132|0); //@line 345 "modules/graphics/opengl/Canvas.cpp"
  $8 = HEAP32[$7>>2]|0; //@line 345 "modules/graphics/opengl/Canvas.cpp"
  __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$8); //@line 345 "modules/graphics/opengl/Canvas.cpp"
  __ZN4love8graphics6opengl6OpenGL16setTextureFilterERNS0_7Texture6FilterE(634044,$6); //@line 346 "modules/graphics/opengl/Canvas.cpp"
  STACKTOP = sp;return; //@line 347 "modules/graphics/opengl/Canvas.cpp"
 }
 $1 = (___cxa_allocate_exception(16)|0); //@line 342 "modules/graphics/opengl/Canvas.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($1|0),(439276|0),($vararg_buffer|0)); //@line 342 "modules/graphics/opengl/Canvas.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  ___cxa_free_exception(($1|0)); //@line 342 "modules/graphics/opengl/Canvas.cpp"
  ___resumeException($4|0);
  // unreachable;
 } else {
  ___cxa_throw(($1|0),(96|0),(344|0)); //@line 342 "modules/graphics/opengl/Canvas.cpp"
  // unreachable; //@line 342 "modules/graphics/opengl/Canvas.cpp"
 }
}
function __ZN4love8graphics6opengl6Canvas7setWrapERKNS0_7Texture4WrapE($this,$w) {
 $this = $this|0;
 $w = $w|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $success$1$off0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 352 "modules/graphics/opengl/Canvas.cpp"
 $1 = $w; //@line 352 "modules/graphics/opengl/Canvas.cpp"
 $2 = $1; //@line 352 "modules/graphics/opengl/Canvas.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 352 "modules/graphics/opengl/Canvas.cpp"
 $4 = (($1) + 4)|0; //@line 352 "modules/graphics/opengl/Canvas.cpp"
 $5 = $4; //@line 352 "modules/graphics/opengl/Canvas.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 352 "modules/graphics/opengl/Canvas.cpp"
 $7 = $0; //@line 352 "modules/graphics/opengl/Canvas.cpp"
 $8 = $7; //@line 352 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$8>>2] = $3; //@line 352 "modules/graphics/opengl/Canvas.cpp"
 $9 = (($7) + 4)|0; //@line 352 "modules/graphics/opengl/Canvas.cpp"
 $10 = $9; //@line 352 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$10>>2] = $6; //@line 352 "modules/graphics/opengl/Canvas.cpp"
 $11 = HEAP8[1252012]|0; //@line 354 "modules/graphics/opengl/Canvas.cpp"
 $12 = ($11<<24>>24)==(0); //@line 354 "modules/graphics/opengl/Canvas.cpp"
 $13 = HEAP8[1252013]|0; //@line 354 "modules/graphics/opengl/Canvas.cpp"
 $14 = HEAP8[1252305]|0; //@line 354 "modules/graphics/opengl/Canvas.cpp"
 $15 = $14 | $13; //@line 354 "modules/graphics/opengl/Canvas.cpp"
 $16 = ($15<<24>>24)!=(0); //@line 354 "modules/graphics/opengl/Canvas.cpp"
 $17 = $12 | $16; //@line 354 "modules/graphics/opengl/Canvas.cpp"
 do {
  if ($17) {
   $success$1$off0 = 1;
  } else {
   $18 = ((($this)) + 8|0); //@line 355 "modules/graphics/opengl/Canvas.cpp"
   $19 = HEAP32[$18>>2]|0; //@line 355 "modules/graphics/opengl/Canvas.cpp"
   $20 = ($19|0)==(0); //@line 81 "./common/math.h"
   $21 = $20&1; //@line 81 "./common/math.h"
   $22 = (($19) + -1)|0; //@line 81 "./common/math.h"
   $23 = (($22) + ($21))|0; //@line 82 "./common/math.h"
   $24 = $23 >> 1; //@line 83 "./common/math.h"
   $25 = $24 | $23; //@line 83 "./common/math.h"
   $26 = $25 >> 2; //@line 83 "./common/math.h"
   $27 = $26 | $25; //@line 83 "./common/math.h"
   $28 = $27 >> 4; //@line 83 "./common/math.h"
   $29 = $28 | $27; //@line 83 "./common/math.h"
   $30 = $29 >> 8; //@line 83 "./common/math.h"
   $31 = $30 | $29; //@line 83 "./common/math.h"
   $32 = $31 >> 16; //@line 83 "./common/math.h"
   $33 = $32 | $31; //@line 83 "./common/math.h"
   $34 = (($33) + 1)|0; //@line 84 "./common/math.h"
   $35 = ($19|0)==($34|0); //@line 355 "modules/graphics/opengl/Canvas.cpp"
   if ($35) {
    $36 = ((($this)) + 12|0); //@line 355 "modules/graphics/opengl/Canvas.cpp"
    $37 = HEAP32[$36>>2]|0; //@line 355 "modules/graphics/opengl/Canvas.cpp"
    $38 = ($37|0)==(0); //@line 81 "./common/math.h"
    $39 = $38&1; //@line 81 "./common/math.h"
    $40 = (($37) + -1)|0; //@line 81 "./common/math.h"
    $41 = (($40) + ($39))|0; //@line 82 "./common/math.h"
    $42 = $41 >> 1; //@line 83 "./common/math.h"
    $43 = $42 | $41; //@line 83 "./common/math.h"
    $44 = $43 >> 2; //@line 83 "./common/math.h"
    $45 = $44 | $43; //@line 83 "./common/math.h"
    $46 = $45 >> 4; //@line 83 "./common/math.h"
    $47 = $46 | $45; //@line 83 "./common/math.h"
    $48 = $47 >> 8; //@line 83 "./common/math.h"
    $49 = $48 | $47; //@line 83 "./common/math.h"
    $50 = $49 >> 16; //@line 83 "./common/math.h"
    $51 = $50 | $49; //@line 83 "./common/math.h"
    $52 = (($51) + 1)|0; //@line 84 "./common/math.h"
    $53 = ($37|0)==($52|0); //@line 355 "modules/graphics/opengl/Canvas.cpp"
    if ($53) {
     $success$1$off0 = 1;
     break;
    }
   }
   $54 = ($3|0)==(0); //@line 357 "modules/graphics/opengl/Canvas.cpp"
   $55 = ((($this)) + 36|0); //@line 357 "modules/graphics/opengl/Canvas.cpp"
   $56 = ($6>>>0)<(1); //@line 357 "modules/graphics/opengl/Canvas.cpp"
   $57 = ($3>>>0)<(0); //@line 357 "modules/graphics/opengl/Canvas.cpp"
   $58 = ($6|0)==(1); //@line 357 "modules/graphics/opengl/Canvas.cpp"
   $59 = $58 & $57; //@line 357 "modules/graphics/opengl/Canvas.cpp"
   $60 = $56 | $59; //@line 357 "modules/graphics/opengl/Canvas.cpp"
   $or$cond = $54 & $60; //@line 357 "modules/graphics/opengl/Canvas.cpp"
   HEAP32[$55>>2] = 0; //@line 361 "modules/graphics/opengl/Canvas.cpp"
   HEAP32[$0>>2] = 0; //@line 361 "modules/graphics/opengl/Canvas.cpp"
   $success$1$off0 = $or$cond;
  }
 } while(0);
 $61 = (__ZNK4love8graphics6opengl6OpenGL31isClampZeroTextureWrapSupportedEv(634044)|0); //@line 364 "modules/graphics/opengl/Canvas.cpp"
 if ($61) {
  $67 = ((($this)) + 132|0); //@line 372 "modules/graphics/opengl/Canvas.cpp"
  $68 = HEAP32[$67>>2]|0; //@line 372 "modules/graphics/opengl/Canvas.cpp"
  __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$68); //@line 372 "modules/graphics/opengl/Canvas.cpp"
  __ZN4love8graphics6opengl6OpenGL14setTextureWrapERKNS0_7Texture4WrapE(634044,$0); //@line 373 "modules/graphics/opengl/Canvas.cpp"
  return ($success$1$off0|0); //@line 375 "modules/graphics/opengl/Canvas.cpp"
 }
 $62 = HEAP32[$0>>2]|0; //@line 366 "modules/graphics/opengl/Canvas.cpp"
 $63 = ($62|0)==(1); //@line 366 "modules/graphics/opengl/Canvas.cpp"
 if ($63) {
  HEAP32[$0>>2] = 0; //@line 367 "modules/graphics/opengl/Canvas.cpp"
 }
 $64 = ((($this)) + 36|0); //@line 368 "modules/graphics/opengl/Canvas.cpp"
 $65 = HEAP32[$64>>2]|0; //@line 368 "modules/graphics/opengl/Canvas.cpp"
 $66 = ($65|0)==(1); //@line 368 "modules/graphics/opengl/Canvas.cpp"
 if (!($66)) {
  $67 = ((($this)) + 132|0); //@line 372 "modules/graphics/opengl/Canvas.cpp"
  $68 = HEAP32[$67>>2]|0; //@line 372 "modules/graphics/opengl/Canvas.cpp"
  __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$68); //@line 372 "modules/graphics/opengl/Canvas.cpp"
  __ZN4love8graphics6opengl6OpenGL14setTextureWrapERKNS0_7Texture4WrapE(634044,$0); //@line 373 "modules/graphics/opengl/Canvas.cpp"
  return ($success$1$off0|0); //@line 375 "modules/graphics/opengl/Canvas.cpp"
 }
 HEAP32[$64>>2] = 0; //@line 369 "modules/graphics/opengl/Canvas.cpp"
 $67 = ((($this)) + 132|0); //@line 372 "modules/graphics/opengl/Canvas.cpp"
 $68 = HEAP32[$67>>2]|0; //@line 372 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$68); //@line 372 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL14setTextureWrapERKNS0_7Texture4WrapE(634044,$0); //@line 373 "modules/graphics/opengl/Canvas.cpp"
 return ($success$1$off0|0); //@line 375 "modules/graphics/opengl/Canvas.cpp"
}
function __ZNK4love8graphics6opengl6Canvas9getHandleEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 132|0); //@line 380 "modules/graphics/opengl/Canvas.cpp"
 return ($0|0); //@line 380 "modules/graphics/opengl/Canvas.cpp"
}
function __ZN4love8graphics6opengl6Canvas12loadVolatileEv($this) {
 $this = $this|0;
 var $$ = 0, $$$i$i = 0, $$$i$i3 = 0, $$0$i = 0, $$0$i$i = 0, $$0$i$i4 = 0, $$1 = 0, $$6 = 0, $$7 = 0, $$8 = 0, $$9 = 0, $$pr = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $current_fbo$i = 0, $externalformat$2 = 0, $internalformat$0 = 0, $or$cond = 0, $or$cond$i = 0, $textype$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $current_fbo$i = sp;
 $0 = ((($this)) + 132|0); //@line 203 "modules/graphics/opengl/Canvas.cpp"
 $1 = ((($this)) + 124|0); //@line 203 "modules/graphics/opengl/Canvas.cpp"
 $2 = ((($this)) + 136|0); //@line 204 "modules/graphics/opengl/Canvas.cpp"
 $3 = ((($this)) + 148|0); //@line 205 "modules/graphics/opengl/Canvas.cpp"
 ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0;HEAP32[$1+12>>2]=0|0;HEAP32[$1+16>>2]=0|0; //@line 203 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$3>>2] = 36053; //@line 205 "modules/graphics/opengl/Canvas.cpp"
 $4 = ((($this)) + 8|0); //@line 208 "modules/graphics/opengl/Canvas.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 208 "modules/graphics/opengl/Canvas.cpp"
 $6 = (__ZNK4love8graphics6opengl6OpenGL17getMaxTextureSizeEv(634044)|0); //@line 208 "modules/graphics/opengl/Canvas.cpp"
 $7 = ($5|0)>($6|0); //@line 208 "modules/graphics/opengl/Canvas.cpp"
 if (!($7)) {
  $8 = ((($this)) + 12|0); //@line 208 "modules/graphics/opengl/Canvas.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 208 "modules/graphics/opengl/Canvas.cpp"
  $10 = (__ZNK4love8graphics6opengl6OpenGL17getMaxTextureSizeEv(634044)|0); //@line 208 "modules/graphics/opengl/Canvas.cpp"
  $11 = ($9|0)>($10|0); //@line 208 "modules/graphics/opengl/Canvas.cpp"
  if (!($11)) {
   $12 = ((($this)) + 164|0); //@line 216 "modules/graphics/opengl/Canvas.cpp"
   $13 = (__ZNK4love8graphics6opengl6OpenGL25getMaxRenderbufferSamplesEv(634044)|0); //@line 216 "modules/graphics/opengl/Canvas.cpp"
   $14 = HEAP32[$12>>2]|0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $15 = ($13|0)<($14|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $16 = $15 ? $13 : $14; //@line 216 "modules/graphics/opengl/Canvas.cpp"
   $17 = ($16|0)<(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $18 = $17 ? 0 : $16; //@line 217 "modules/graphics/opengl/Canvas.cpp"
   HEAP32[$12>>2] = $18; //@line 217 "modules/graphics/opengl/Canvas.cpp"
   $19 = HEAP32[164922]|0; //@line 355 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_vii[$19 & 8191](1,$0); //@line 355 "./libraries/glad/gladfuncs.hpp"
   $20 = HEAP32[$0>>2]|0; //@line 220 "modules/graphics/opengl/Canvas.cpp"
   __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$20); //@line 220 "modules/graphics/opengl/Canvas.cpp"
   $21 = HEAP8[1252351]|0; //@line 222 "modules/graphics/opengl/Canvas.cpp"
   $22 = ($21<<24>>24)==(0); //@line 222 "modules/graphics/opengl/Canvas.cpp"
   if (!($22)) {
    $23 = HEAP32[165108]|0; //@line 44 "./libraries/glad/gladfuncs.hpp"
    FUNCTION_TABLE_viii[$23 & 8191](3553,37794,37795); //@line 44 "./libraries/glad/gladfuncs.hpp"
   }
   $24 = HEAP32[$this>>2]|0; //@line 225 "modules/graphics/opengl/Canvas.cpp"
   $25 = ((($24)) + 32|0); //@line 225 "modules/graphics/opengl/Canvas.cpp"
   $26 = HEAP32[$25>>2]|0; //@line 225 "modules/graphics/opengl/Canvas.cpp"
   $27 = ((($this)) + 16|0); //@line 225 "modules/graphics/opengl/Canvas.cpp"
   FUNCTION_TABLE_vii[$26 & 8191]($this,$27); //@line 225 "modules/graphics/opengl/Canvas.cpp"
   $28 = HEAP32[$this>>2]|0; //@line 226 "modules/graphics/opengl/Canvas.cpp"
   $29 = ((($28)) + 40|0); //@line 226 "modules/graphics/opengl/Canvas.cpp"
   $30 = HEAP32[$29>>2]|0; //@line 226 "modules/graphics/opengl/Canvas.cpp"
   $31 = ((($this)) + 32|0); //@line 226 "modules/graphics/opengl/Canvas.cpp"
   (FUNCTION_TABLE_iii[$30 & 8191]($this,$31)|0); //@line 226 "modules/graphics/opengl/Canvas.cpp"
   $32 = ((($this)) + 144|0); //@line 232 "modules/graphics/opengl/Canvas.cpp"
   $33 = HEAP32[$32>>2]|0; //@line 232 "modules/graphics/opengl/Canvas.cpp"
   switch ($33|0) {
   case 0:  {
    $34 = (__ZN4love8graphics14isGammaCorrectEv()|0); //@line 669 "modules/graphics/opengl/Canvas.cpp"
    if ($34) {
     label = 26;
    } else {
     $35 = HEAP8[1252012]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $36 = ($35<<24>>24)==(0); //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $37 = HEAP8[1252013]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $38 = HEAP8[1252284]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $39 = $38 | $37; //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $40 = HEAP8[1252377]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $41 = $39 | $40; //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $42 = ($41<<24>>24)!=(0); //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $43 = $36 | $42; //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $$$i$i3 = $43 ? 7 : 2; //@line 675 "modules/graphics/opengl/Canvas.cpp"
     $$0$i$i4 = $$$i$i3;
     label = 9;
    }
    break;
   }
   case 1:  {
    label = 21;
    break;
   }
   default: {
    $$0$i$i4 = $33;
    label = 9;
   }
   }
   L10: do {
    if ((label|0) == 9) {
     do {
      switch ($$0$i$i4|0) {
      case 2:  {
       $externalformat$2 = 6408;$internalformat$0 = 32854;$textype$0 = 32819;
       break L10;
       break;
      }
      case 16:  {
       label = 26;
       break L10;
       break;
      }
      case 12:  {
       label = 21;
       break L10;
       break;
      }
      case 3:  {
       $externalformat$2 = 6408;$internalformat$0 = 32855;$textype$0 = 32820;
       break L10;
       break;
      }
      case 4:  {
       $externalformat$2 = 6407;$internalformat$0 = 36194;$textype$0 = 33635;
       break L10;
       break;
      }
      case 5:  {
       $externalformat$2 = 6403;$internalformat$0 = 33321;$textype$0 = 5121;
       break L10;
       break;
      }
      case 6:  {
       $externalformat$2 = 33319;$internalformat$0 = 33323;$textype$0 = 5121;
       break L10;
       break;
      }
      case 8:  {
       $externalformat$2 = 6408;$internalformat$0 = 32857;$textype$0 = 33640;
       break L10;
       break;
      }
      case 9:  {
       $externalformat$2 = 6407;$internalformat$0 = 35898;$textype$0 = 35899;
       break L10;
       break;
      }
      case 10:  {
       $44 = HEAP8[1252303]|0; //@line 730 "modules/graphics/opengl/Canvas.cpp"
       $45 = ($44<<24>>24)==(0); //@line 730 "modules/graphics/opengl/Canvas.cpp"
       if (!($45)) {
        $externalformat$2 = 6403;$internalformat$0 = 33325;$textype$0 = 36193;
        break L10;
       }
       $46 = HEAP8[1251994]|0; //@line 732 "modules/graphics/opengl/Canvas.cpp"
       $47 = ($46<<24>>24)==(0); //@line 732 "modules/graphics/opengl/Canvas.cpp"
       $$6 = $47 ? 5131 : 5126; //@line 732 "modules/graphics/opengl/Canvas.cpp"
       $externalformat$2 = 6403;$internalformat$0 = 33325;$textype$0 = $$6;
       break L10;
       break;
      }
      case 11:  {
       $48 = HEAP8[1252303]|0; //@line 740 "modules/graphics/opengl/Canvas.cpp"
       $49 = ($48<<24>>24)==(0); //@line 740 "modules/graphics/opengl/Canvas.cpp"
       if (!($49)) {
        $externalformat$2 = 33319;$internalformat$0 = 33327;$textype$0 = 36193;
        break L10;
       }
       $50 = HEAP8[1251994]|0; //@line 742 "modules/graphics/opengl/Canvas.cpp"
       $51 = ($50<<24>>24)==(0); //@line 742 "modules/graphics/opengl/Canvas.cpp"
       $$7 = $51 ? 5131 : 5126; //@line 742 "modules/graphics/opengl/Canvas.cpp"
       $externalformat$2 = 33319;$internalformat$0 = 33327;$textype$0 = $$7;
       break L10;
       break;
      }
      case 13:  {
       $externalformat$2 = 6403;$internalformat$0 = 33326;$textype$0 = 5126;
       break L10;
       break;
      }
      case 14:  {
       $externalformat$2 = 33319;$internalformat$0 = 33328;$textype$0 = 5126;
       break L10;
       break;
      }
      case 15:  {
       $externalformat$2 = 6408;$internalformat$0 = 34836;$textype$0 = 5126;
       break L10;
       break;
      }
      default: {
       $externalformat$2 = 6408;$internalformat$0 = 32856;$textype$0 = 5121;
       break L10;
      }
      }
     } while(0);
    }
   } while(0);
   if ((label|0) == 21) {
    $52 = HEAP8[1252303]|0; //@line 749 "modules/graphics/opengl/Canvas.cpp"
    $53 = ($52<<24>>24)==(0); //@line 749 "modules/graphics/opengl/Canvas.cpp"
    if ($53) {
     $54 = HEAP8[1251994]|0; //@line 751 "modules/graphics/opengl/Canvas.cpp"
     $55 = ($54<<24>>24)==(0); //@line 751 "modules/graphics/opengl/Canvas.cpp"
     $$8 = $55 ? 5131 : 5126; //@line 751 "modules/graphics/opengl/Canvas.cpp"
     $externalformat$2 = 6408;$internalformat$0 = 34842;$textype$0 = $$8;
    } else {
     $externalformat$2 = 6408;$internalformat$0 = 34842;$textype$0 = 36193;
    }
   }
   else if ((label|0) == 26) {
    $56 = HEAP8[1252012]|0; //@line 773 "modules/graphics/opengl/Canvas.cpp"
    $57 = ($56<<24>>24)==(0); //@line 773 "modules/graphics/opengl/Canvas.cpp"
    $58 = HEAP8[1252013]|0; //@line 773 "modules/graphics/opengl/Canvas.cpp"
    $59 = ($58<<24>>24)!=(0); //@line 773 "modules/graphics/opengl/Canvas.cpp"
    $or$cond$i = $57 | $59; //@line 773 "modules/graphics/opengl/Canvas.cpp"
    $$9 = $or$cond$i ? 6408 : 35906; //@line 773 "modules/graphics/opengl/Canvas.cpp"
    $externalformat$2 = $$9;$internalformat$0 = 35907;$textype$0 = 5121;
   }
   $60 = HEAP8[1252012]|0; //@line 236 "modules/graphics/opengl/Canvas.cpp"
   $61 = ($60<<24>>24)==(0); //@line 236 "modules/graphics/opengl/Canvas.cpp"
   $62 = HEAP8[1252013]|0; //@line 236 "modules/graphics/opengl/Canvas.cpp"
   $63 = ($62<<24>>24)!=(0); //@line 236 "modules/graphics/opengl/Canvas.cpp"
   $or$cond = $61 | $63; //@line 236 "modules/graphics/opengl/Canvas.cpp"
   while(1) {
    $64 = HEAP32[165571]|0; //@line 71 "./libraries/glad/gladfuncs.hpp"
    $65 = (FUNCTION_TABLE_i[$64 & 8191]()|0); //@line 71 "./libraries/glad/gladfuncs.hpp"
    $66 = ($65|0)==(0); //@line 239 "modules/graphics/opengl/Canvas.cpp"
    if ($66) {
     break;
    }
   }
   $$ = $or$cond ? $internalformat$0 : $externalformat$2; //@line 236 "modules/graphics/opengl/Canvas.cpp"
   $67 = HEAP32[$4>>2]|0; //@line 242 "modules/graphics/opengl/Canvas.cpp"
   $68 = HEAP32[$8>>2]|0; //@line 242 "modules/graphics/opengl/Canvas.cpp"
   $69 = HEAP32[165602]|0; //@line 47 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_viiiiiiiii[$69 & 8191](3553,0,$$,$67,$68,0,$externalformat$2,$textype$0,0); //@line 47 "./libraries/glad/gladfuncs.hpp"
   $70 = HEAP32[165571]|0; //@line 71 "./libraries/glad/gladfuncs.hpp"
   $71 = (FUNCTION_TABLE_i[$70 & 8191]()|0); //@line 71 "./libraries/glad/gladfuncs.hpp"
   $72 = ($71|0)==(0); //@line 245 "modules/graphics/opengl/Canvas.cpp"
   if (!($72)) {
    $73 = HEAP32[$0>>2]|0; //@line 247 "modules/graphics/opengl/Canvas.cpp"
    __ZN4love8graphics6opengl6OpenGL13deleteTextureEj(634044,$73); //@line 247 "modules/graphics/opengl/Canvas.cpp"
    HEAP32[$0>>2] = 0; //@line 248 "modules/graphics/opengl/Canvas.cpp"
    HEAP32[$3>>2] = 36054; //@line 249 "modules/graphics/opengl/Canvas.cpp"
    $$1 = 0;
    STACKTOP = sp;return ($$1|0); //@line 277 "modules/graphics/opengl/Canvas.cpp"
   }
   $74 = (__ZN4love8graphics6opengl6Canvas13createMSAAFBOEj($this,$internalformat$0)|0); //@line 255 "modules/graphics/opengl/Canvas.cpp"
   if ($74) {
    $$pr = HEAP32[$3>>2]|0; //@line 258 "modules/graphics/opengl/Canvas.cpp"
    $87 = $$pr;
   } else {
    $75 = HEAP32[$0>>2]|0; //@line 256 "modules/graphics/opengl/Canvas.cpp"
    $76 = HEAP32[165587]|0; //@line 73 "./libraries/glad/gladfuncs.hpp"
    FUNCTION_TABLE_vii[$76 & 8191](36006,$current_fbo$i); //@line 73 "./libraries/glad/gladfuncs.hpp"
    $77 = HEAP32[164797]|0; //@line 669 "./libraries/glad/gladfuncs.hpp"
    FUNCTION_TABLE_vii[$77 & 8191](1,$1); //@line 669 "./libraries/glad/gladfuncs.hpp"
    $78 = HEAP32[$1>>2]|0; //@line 43 "modules/graphics/opengl/Canvas.cpp"
    __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$78); //@line 43 "modules/graphics/opengl/Canvas.cpp"
    $79 = ($75|0)==(0); //@line 45 "modules/graphics/opengl/Canvas.cpp"
    if (!($79)) {
     $80 = HEAP32[164777]|0; //@line 672 "./libraries/glad/gladfuncs.hpp"
     FUNCTION_TABLE_viiiii[$80 & 8191](36160,36064,3553,$75,0); //@line 672 "./libraries/glad/gladfuncs.hpp"
     $81 = HEAP32[165395]|0; //@line 50 "./libraries/glad/gladfuncs.hpp"
     FUNCTION_TABLE_vdddd[$81 & 8191](0.0,0.0,0.0,0.0); //@line 50 "./libraries/glad/gladfuncs.hpp"
     $82 = HEAP32[164768]|0; //@line 49 "./libraries/glad/gladfuncs.hpp"
     FUNCTION_TABLE_vi[$82 & 8191](16384); //@line 49 "./libraries/glad/gladfuncs.hpp"
    }
    $83 = HEAP32[165498]|0; //@line 670 "./libraries/glad/gladfuncs.hpp"
    $84 = (FUNCTION_TABLE_ii[$83 & 8191](36160)|0); //@line 670 "./libraries/glad/gladfuncs.hpp"
    $85 = HEAP32[$current_fbo$i>>2]|0; //@line 57 "modules/graphics/opengl/Canvas.cpp"
    __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$85); //@line 57 "modules/graphics/opengl/Canvas.cpp"
    HEAP32[$3>>2] = $84; //@line 256 "modules/graphics/opengl/Canvas.cpp"
    $87 = $84;
   }
   $86 = ($87|0)==(36053); //@line 258 "modules/graphics/opengl/Canvas.cpp"
   if (!($86)) {
    $88 = HEAP32[$1>>2]|0; //@line 260 "modules/graphics/opengl/Canvas.cpp"
    $89 = ($88|0)==(0); //@line 260 "modules/graphics/opengl/Canvas.cpp"
    if ($89) {
     $$1 = 0;
     STACKTOP = sp;return ($$1|0); //@line 277 "modules/graphics/opengl/Canvas.cpp"
    }
    $90 = HEAP32[164758]|0; //@line 668 "./libraries/glad/gladfuncs.hpp"
    FUNCTION_TABLE_vii[$90 & 8191](1,$1); //@line 668 "./libraries/glad/gladfuncs.hpp"
    HEAP32[$1>>2] = 0; //@line 263 "modules/graphics/opengl/Canvas.cpp"
    $$1 = 0;
    STACKTOP = sp;return ($$1|0); //@line 277 "modules/graphics/opengl/Canvas.cpp"
   }
   $91 = ((($this)) + 172|0); //@line 268 "modules/graphics/opengl/Canvas.cpp"
   $92 = HEAP32[$91>>2]|0; //@line 268 "modules/graphics/opengl/Canvas.cpp"
   $93 = HEAP32[$32>>2]|0; //@line 270 "modules/graphics/opengl/Canvas.cpp"
   switch ($93|0) {
   case 0:  {
    $94 = (__ZN4love8graphics14isGammaCorrectEv()|0); //@line 669 "modules/graphics/opengl/Canvas.cpp"
    if ($94) {
     label = 44;
    } else {
     $95 = HEAP8[1252012]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $96 = ($95<<24>>24)==(0); //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $97 = HEAP8[1252013]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $98 = HEAP8[1252284]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $99 = $98 | $97; //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $100 = HEAP8[1252377]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $101 = $99 | $100; //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $102 = ($101<<24>>24)!=(0); //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $103 = $96 | $102; //@line 671 "modules/graphics/opengl/Canvas.cpp"
     $$$i$i = $103 ? 7 : 2; //@line 675 "modules/graphics/opengl/Canvas.cpp"
     $$0$i$i = $$$i$i;
     label = 42;
    }
    break;
   }
   case 1:  {
    label = 45;
    break;
   }
   default: {
    $$0$i$i = $93;
    label = 42;
   }
   }
   L56: do {
    if ((label|0) == 42) {
     switch ($$0$i$i|0) {
     case 5:  {
      $$0$i = 8;
      break L56;
      break;
     }
     case 12: case 14:  {
      label = 45;
      break L56;
      break;
     }
     case 10: case 6: case 4: case 3: case 2:  {
      $$0$i = 16;
      break L56;
      break;
     }
     case 15:  {
      $$0$i = 128;
      break L56;
      break;
     }
     default: {
      label = 44;
      break L56;
     }
     }
    }
   } while(0);
   if ((label|0) == 44) {
    $$0$i = 32;
   }
   else if ((label|0) == 45) {
    $$0$i = 64;
   }
   $104 = HEAP32[$4>>2]|0; //@line 270 "modules/graphics/opengl/Canvas.cpp"
   $105 = Math_imul($104, $$0$i)|0; //@line 270 "modules/graphics/opengl/Canvas.cpp"
   $106 = HEAP32[$8>>2]|0; //@line 270 "modules/graphics/opengl/Canvas.cpp"
   $107 = Math_imul($105, $106)|0; //@line 270 "modules/graphics/opengl/Canvas.cpp"
   $108 = $107 >>> 3; //@line 270 "modules/graphics/opengl/Canvas.cpp"
   HEAP32[$91>>2] = $108; //@line 270 "modules/graphics/opengl/Canvas.cpp"
   $109 = HEAP32[$2>>2]|0; //@line 271 "modules/graphics/opengl/Canvas.cpp"
   $110 = ($109|0)==(0); //@line 271 "modules/graphics/opengl/Canvas.cpp"
   if ($110) {
    $115 = $108;
   } else {
    $111 = ((($this)) + 168|0); //@line 272 "modules/graphics/opengl/Canvas.cpp"
    $112 = HEAP32[$111>>2]|0; //@line 272 "modules/graphics/opengl/Canvas.cpp"
    $113 = Math_imul($112, $108)|0; //@line 272 "modules/graphics/opengl/Canvas.cpp"
    $114 = (($113) + ($108))|0; //@line 272 "modules/graphics/opengl/Canvas.cpp"
    HEAP32[$91>>2] = $114; //@line 272 "modules/graphics/opengl/Canvas.cpp"
    $115 = $114;
   }
   __ZN4love8graphics6opengl6OpenGL23updateTextureMemorySizeEjj(634044,$92,$115); //@line 274 "modules/graphics/opengl/Canvas.cpp"
   $$1 = 1;
   STACKTOP = sp;return ($$1|0); //@line 277 "modules/graphics/opengl/Canvas.cpp"
  }
 }
 HEAP32[$3>>2] = 36054; //@line 210 "modules/graphics/opengl/Canvas.cpp"
 $$1 = 0;
 STACKTOP = sp;return ($$1|0); //@line 277 "modules/graphics/opengl/Canvas.cpp"
}
function __ZN4love8graphics6opengl6Canvas13createMSAAFBOEj($this,$internalformat) {
 $this = $this|0;
 $internalformat = $internalformat|0;
 var $$0 = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $current_fbo$i = 0, $current_fbo$i1 = 0, $previous$0 = 0, $previous$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $current_fbo$i1 = sp + 4|0;
 $current_fbo$i = sp;
 $0 = ((($this)) + 164|0); //@line 156 "modules/graphics/opengl/Canvas.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 156 "modules/graphics/opengl/Canvas.cpp"
 $2 = ((($this)) + 168|0); //@line 156 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$2>>2] = $1; //@line 156 "modules/graphics/opengl/Canvas.cpp"
 $3 = ($1|0)<(2); //@line 158 "modules/graphics/opengl/Canvas.cpp"
 if ($3) {
  HEAP32[$2>>2] = 0; //@line 160 "modules/graphics/opengl/Canvas.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 197 "modules/graphics/opengl/Canvas.cpp"
 }
 $4 = ((($this)) + 124|0); //@line 165 "modules/graphics/opengl/Canvas.cpp"
 $5 = HEAP32[165587]|0; //@line 73 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$5 & 8191](36006,$current_fbo$i); //@line 73 "./libraries/glad/gladfuncs.hpp"
 $6 = HEAP32[164797]|0; //@line 669 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$6 & 8191](1,$4); //@line 669 "./libraries/glad/gladfuncs.hpp"
 $7 = HEAP32[$4>>2]|0; //@line 43 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$7); //@line 43 "modules/graphics/opengl/Canvas.cpp"
 $8 = HEAP32[165498]|0; //@line 670 "./libraries/glad/gladfuncs.hpp"
 $9 = (FUNCTION_TABLE_ii[$8 & 8191](36160)|0); //@line 670 "./libraries/glad/gladfuncs.hpp"
 $10 = HEAP32[$current_fbo$i>>2]|0; //@line 57 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$10); //@line 57 "modules/graphics/opengl/Canvas.cpp"
 $11 = ((($this)) + 148|0); //@line 165 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$11>>2] = $9; //@line 165 "modules/graphics/opengl/Canvas.cpp"
 $12 = (__ZNK4love8graphics6opengl6OpenGL13getDefaultFBOEv(634044)|0); //@line 167 "modules/graphics/opengl/Canvas.cpp"
 $13 = HEAP32[158272]|0; //@line 168 "modules/graphics/opengl/Canvas.cpp"
 $14 = ($13|0)==($this|0); //@line 168 "modules/graphics/opengl/Canvas.cpp"
 if ($14) {
  $previous$1 = $12;
 } else {
  $15 = ($13|0)==(0|0); //@line 170 "modules/graphics/opengl/Canvas.cpp"
  if ($15) {
   $previous$0 = $12;
  } else {
   $16 = ((($13)) + 124|0); //@line 171 "modules/graphics/opengl/Canvas.cpp"
   $17 = HEAP32[$16>>2]|0; //@line 171 "modules/graphics/opengl/Canvas.cpp"
   $previous$0 = $17;
  }
  $18 = HEAP32[$4>>2]|0; //@line 173 "modules/graphics/opengl/Canvas.cpp"
  __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$18); //@line 173 "modules/graphics/opengl/Canvas.cpp"
  $previous$1 = $previous$0;
 }
 $19 = ((($this)) + 8|0); //@line 177 "modules/graphics/opengl/Canvas.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 177 "modules/graphics/opengl/Canvas.cpp"
 $21 = ((($this)) + 12|0); //@line 177 "modules/graphics/opengl/Canvas.cpp"
 $22 = HEAP32[$21>>2]|0; //@line 177 "modules/graphics/opengl/Canvas.cpp"
 $23 = ((($this)) + 136|0); //@line 177 "modules/graphics/opengl/Canvas.cpp"
 $24 = HEAP32[164676]|0; //@line 663 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$24 & 8191](1,$23); //@line 663 "./libraries/glad/gladfuncs.hpp"
 $25 = HEAP32[$23>>2]|0; //@line 65 "modules/graphics/opengl/Canvas.cpp"
 $26 = HEAP32[165274]|0; //@line 661 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$26 & 8191](36161,$25); //@line 661 "./libraries/glad/gladfuncs.hpp"
 $27 = HEAP32[$2>>2]|0; //@line 67 "modules/graphics/opengl/Canvas.cpp"
 $28 = HEAP32[165298]|0; //@line 678 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiiii[$28 & 8191](36161,$27,$internalformat,$20,$22); //@line 678 "./libraries/glad/gladfuncs.hpp"
 $29 = HEAP32[$23>>2]|0; //@line 68 "modules/graphics/opengl/Canvas.cpp"
 $30 = HEAP32[164630]|0; //@line 674 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiii[$30 & 8191](36160,36064,36161,$29); //@line 674 "./libraries/glad/gladfuncs.hpp"
 $31 = HEAP32[164884]|0; //@line 665 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$31 & 8191](36161,36011,$2); //@line 665 "./libraries/glad/gladfuncs.hpp"
 $32 = HEAP32[165274]|0; //@line 661 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$32 & 8191](36161,0); //@line 661 "./libraries/glad/gladfuncs.hpp"
 $33 = HEAP32[165498]|0; //@line 670 "./libraries/glad/gladfuncs.hpp"
 $34 = (FUNCTION_TABLE_ii[$33 & 8191](36160)|0); //@line 670 "./libraries/glad/gladfuncs.hpp"
 $35 = ($34|0)==(36053); //@line 76 "modules/graphics/opengl/Canvas.cpp"
 if ($35) {
  $37 = HEAP32[165395]|0; //@line 50 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vdddd[$37 & 8191](0.0,0.0,0.0,0.0); //@line 50 "./libraries/glad/gladfuncs.hpp"
  $38 = HEAP32[164768]|0; //@line 49 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$38 & 8191](16384); //@line 49 "./libraries/glad/gladfuncs.hpp"
  HEAP32[$11>>2] = 36053; //@line 177 "modules/graphics/opengl/Canvas.cpp"
  $39 = ((($this)) + 128|0); //@line 181 "modules/graphics/opengl/Canvas.cpp"
  $40 = ((($this)) + 132|0); //@line 181 "modules/graphics/opengl/Canvas.cpp"
  $41 = HEAP32[$40>>2]|0; //@line 181 "modules/graphics/opengl/Canvas.cpp"
  $42 = HEAP32[165587]|0; //@line 73 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$42 & 8191](36006,$current_fbo$i1); //@line 73 "./libraries/glad/gladfuncs.hpp"
  $43 = HEAP32[164797]|0; //@line 669 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$43 & 8191](1,$39); //@line 669 "./libraries/glad/gladfuncs.hpp"
  $44 = HEAP32[$39>>2]|0; //@line 43 "modules/graphics/opengl/Canvas.cpp"
  __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$44); //@line 43 "modules/graphics/opengl/Canvas.cpp"
  $45 = ($41|0)==(0); //@line 45 "modules/graphics/opengl/Canvas.cpp"
  if (!($45)) {
   $46 = HEAP32[164777]|0; //@line 672 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_viiiii[$46 & 8191](36160,36064,3553,$41,0); //@line 672 "./libraries/glad/gladfuncs.hpp"
   $47 = HEAP32[165395]|0; //@line 50 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_vdddd[$47 & 8191](0.0,0.0,0.0,0.0); //@line 50 "./libraries/glad/gladfuncs.hpp"
   $48 = HEAP32[164768]|0; //@line 49 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_vi[$48 & 8191](16384); //@line 49 "./libraries/glad/gladfuncs.hpp"
  }
  $49 = HEAP32[165498]|0; //@line 670 "./libraries/glad/gladfuncs.hpp"
  $50 = (FUNCTION_TABLE_ii[$49 & 8191](36160)|0); //@line 670 "./libraries/glad/gladfuncs.hpp"
  $51 = HEAP32[$current_fbo$i1>>2]|0; //@line 57 "modules/graphics/opengl/Canvas.cpp"
  __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$51); //@line 57 "modules/graphics/opengl/Canvas.cpp"
  HEAP32[$11>>2] = $50; //@line 181 "modules/graphics/opengl/Canvas.cpp"
  $52 = ($50|0)==(36053); //@line 183 "modules/graphics/opengl/Canvas.cpp"
  if (!($52)) {
   $$pre$phiZ2D = $39;
   label = 12;
  }
 } else {
  $36 = HEAP32[164936]|0; //@line 662 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$36 & 8191](1,$23); //@line 662 "./libraries/glad/gladfuncs.hpp"
  HEAP32[$23>>2] = 0; //@line 85 "modules/graphics/opengl/Canvas.cpp"
  HEAP32[$11>>2] = $34; //@line 177 "modules/graphics/opengl/Canvas.cpp"
  $$pre = ((($this)) + 128|0); //@line 187 "modules/graphics/opengl/Canvas.cpp"
  $$pre$phiZ2D = $$pre;
  label = 12;
 }
 if ((label|0) == 12) {
  $53 = HEAP32[164758]|0; //@line 668 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$53 & 8191](1,$4); //@line 668 "./libraries/glad/gladfuncs.hpp"
  $54 = HEAP32[164758]|0; //@line 668 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$54 & 8191](1,$$pre$phiZ2D); //@line 668 "./libraries/glad/gladfuncs.hpp"
  $55 = HEAP32[164936]|0; //@line 662 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$55 & 8191](1,$23); //@line 662 "./libraries/glad/gladfuncs.hpp"
  HEAP32[$$pre$phiZ2D>>2] = 0; //@line 189 "modules/graphics/opengl/Canvas.cpp"
  HEAP32[$23>>2] = 0; //@line 189 "modules/graphics/opengl/Canvas.cpp"
  HEAP32[$4>>2] = 0; //@line 189 "modules/graphics/opengl/Canvas.cpp"
  HEAP32[$2>>2] = 0; //@line 190 "modules/graphics/opengl/Canvas.cpp"
 }
 $56 = HEAP32[158272]|0; //@line 193 "modules/graphics/opengl/Canvas.cpp"
 $57 = ($56|0)==($this|0); //@line 193 "modules/graphics/opengl/Canvas.cpp"
 if (!($57)) {
  __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$previous$1); //@line 194 "modules/graphics/opengl/Canvas.cpp"
 }
 $58 = HEAP32[$11>>2]|0; //@line 196 "modules/graphics/opengl/Canvas.cpp"
 $59 = ($58|0)==(36053); //@line 196 "modules/graphics/opengl/Canvas.cpp"
 $$0 = $59;
 STACKTOP = sp;return ($$0|0); //@line 197 "modules/graphics/opengl/Canvas.cpp"
}
function __ZN4love8graphics6opengl6Canvas14unloadVolatileEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $__new_last2$i$i$i = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 124|0); //@line 281 "modules/graphics/opengl/Canvas.cpp"
 $1 = HEAP32[164758]|0; //@line 668 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$1 & 8191](1,$0); //@line 668 "./libraries/glad/gladfuncs.hpp"
 $2 = ((($this)) + 128|0); //@line 282 "modules/graphics/opengl/Canvas.cpp"
 $3 = HEAP32[164758]|0; //@line 668 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$3 & 8191](1,$2); //@line 668 "./libraries/glad/gladfuncs.hpp"
 $4 = ((($this)) + 140|0); //@line 284 "modules/graphics/opengl/Canvas.cpp"
 $5 = HEAP32[164936]|0; //@line 662 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$5 & 8191](1,$4); //@line 662 "./libraries/glad/gladfuncs.hpp"
 $6 = ((($this)) + 136|0); //@line 285 "modules/graphics/opengl/Canvas.cpp"
 $7 = HEAP32[164936]|0; //@line 662 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$7 & 8191](1,$6); //@line 662 "./libraries/glad/gladfuncs.hpp"
 $8 = ((($this)) + 132|0); //@line 287 "modules/graphics/opengl/Canvas.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 287 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL13deleteTextureEj(634044,$9); //@line 287 "modules/graphics/opengl/Canvas.cpp"
 $10 = ((($this)) + 152|0); //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;HEAP32[$0+12>>2]=0|0;HEAP32[$0+16>>2]=0|0; //@line 290 "modules/graphics/opengl/Canvas.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ((($this)) + 156|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = HEAP32[$12>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = ($13|0)==($11|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($14) {
  $19 = ((($this)) + 172|0); //@line 297 "modules/graphics/opengl/Canvas.cpp"
  $20 = HEAP32[$19>>2]|0; //@line 297 "modules/graphics/opengl/Canvas.cpp"
  __ZN4love8graphics6opengl6OpenGL23updateTextureMemorySizeEjj(634044,$20,0); //@line 297 "modules/graphics/opengl/Canvas.cpp"
  HEAP32[$19>>2] = 0; //@line 298 "modules/graphics/opengl/Canvas.cpp"
  return; //@line 299 "modules/graphics/opengl/Canvas.cpp"
 }
 $__new_last2$i$i$i = $11; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $scevgep$i$i$i = ((($13)) + -4|0);
 $15 = $scevgep$i$i$i;
 $16 = (($15) - ($__new_last2$i$i$i))|0;
 $17 = $16 >>> 2;
 $18 = $17 ^ -1;
 $scevgep4$i$i$i = (($13) + ($18<<2)|0);
 HEAP32[$12>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $19 = ((($this)) + 172|0); //@line 297 "modules/graphics/opengl/Canvas.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 297 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL23updateTextureMemorySizeEjj(634044,$20,0); //@line 297 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$19>>2] = 0; //@line 298 "modules/graphics/opengl/Canvas.cpp"
 return; //@line 299 "modules/graphics/opengl/Canvas.cpp"
}
function __ZThn120_N4love8graphics6opengl6CanvasD1Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -120|0); //@line 143 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6CanvasD2Ev($0); //@line 143 "modules/graphics/opengl/Canvas.cpp"
 return; //@line 143 "modules/graphics/opengl/Canvas.cpp"
}
function __ZThn120_N4love8graphics6opengl6CanvasD0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -120|0); //@line 143 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6CanvasD2Ev($0); //@line 144 "modules/graphics/opengl/Canvas.cpp"
 __ZdlPv($0); //@line 144 "modules/graphics/opengl/Canvas.cpp"
 return; //@line 143 "modules/graphics/opengl/Canvas.cpp"
}
function __ZThn120_N4love8graphics6opengl6Canvas12loadVolatileEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -120|0); //@line 199 "modules/graphics/opengl/Canvas.cpp"
 $1 = (__ZN4love8graphics6opengl6Canvas12loadVolatileEv($0)|0); //@line 199 "modules/graphics/opengl/Canvas.cpp"
 return ($1|0); //@line 199 "modules/graphics/opengl/Canvas.cpp"
}
function __ZThn120_N4love8graphics6opengl6Canvas14unloadVolatileEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $__new_last2$i$i$i$i = 0, $scevgep$i$i$i$i = 0, $scevgep4$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -120|0); //@line 279 "modules/graphics/opengl/Canvas.cpp"
 $1 = ((($0)) + 124|0); //@line 281 "modules/graphics/opengl/Canvas.cpp"
 $2 = HEAP32[164758]|0; //@line 668 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$2 & 8191](1,$1); //@line 668 "./libraries/glad/gladfuncs.hpp"
 $3 = ((($0)) + 128|0); //@line 282 "modules/graphics/opengl/Canvas.cpp"
 $4 = HEAP32[164758]|0; //@line 668 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$4 & 8191](1,$3); //@line 668 "./libraries/glad/gladfuncs.hpp"
 $5 = ((($0)) + 140|0); //@line 284 "modules/graphics/opengl/Canvas.cpp"
 $6 = HEAP32[164936]|0; //@line 662 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$6 & 8191](1,$5); //@line 662 "./libraries/glad/gladfuncs.hpp"
 $7 = ((($0)) + 136|0); //@line 285 "modules/graphics/opengl/Canvas.cpp"
 $8 = HEAP32[164936]|0; //@line 662 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$8 & 8191](1,$7); //@line 662 "./libraries/glad/gladfuncs.hpp"
 $9 = ((($0)) + 132|0); //@line 287 "modules/graphics/opengl/Canvas.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 287 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL13deleteTextureEj(634044,$10); //@line 287 "modules/graphics/opengl/Canvas.cpp"
 $11 = ((($0)) + 152|0); //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0;HEAP32[$1+12>>2]=0|0;HEAP32[$1+16>>2]=0|0; //@line 290 "modules/graphics/opengl/Canvas.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ((($0)) + 156|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = HEAP32[$13>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = ($14|0)==($12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($15) {
  $20 = ((($0)) + 172|0); //@line 297 "modules/graphics/opengl/Canvas.cpp"
  $21 = HEAP32[$20>>2]|0; //@line 297 "modules/graphics/opengl/Canvas.cpp"
  __ZN4love8graphics6opengl6OpenGL23updateTextureMemorySizeEjj(634044,$21,0); //@line 297 "modules/graphics/opengl/Canvas.cpp"
  HEAP32[$20>>2] = 0; //@line 298 "modules/graphics/opengl/Canvas.cpp"
  return; //@line 279 "modules/graphics/opengl/Canvas.cpp"
 }
 $__new_last2$i$i$i$i = $12; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $scevgep$i$i$i$i = ((($14)) + -4|0);
 $16 = $scevgep$i$i$i$i;
 $17 = (($16) - ($__new_last2$i$i$i$i))|0;
 $18 = $17 >>> 2;
 $19 = $18 ^ -1;
 $scevgep4$i$i$i$i = (($14) + ($19<<2)|0);
 HEAP32[$13>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = ((($0)) + 172|0); //@line 297 "modules/graphics/opengl/Canvas.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 297 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL23updateTextureMemorySizeEjj(634044,$21,0); //@line 297 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$20>>2] = 0; //@line 298 "modules/graphics/opengl/Canvas.cpp"
 return; //@line 279 "modules/graphics/opengl/Canvas.cpp"
}
function __GLOBAL__sub_I_Canvas_cpp() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love9StringMapINS_8graphics6opengl6Canvas6FormatELj17EEC2EPKNS5_5EntryEj(633112,24404,136); //@line 964 "modules/graphics/opengl/Canvas.cpp"
 return;
}
function __ZN4love9StringMapINS_8graphics6opengl6Canvas6FormatELj17EEC2EPKNS5_5EntryEj($this,$entries,$num) {
 $this = $this|0;
 $entries = $entries|0;
 $num = $num|0;
 var $$02$i$i = 0, $$lcssa26 = 0, $$lcssa28 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$08$i = 0, $i1$05 = 0, $scevgep = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 8|0); //@line 155 "./common/StringMap.h"
 HEAP8[$0>>0] = 0; //@line 155 "./common/StringMap.h"
 $1 = ((($this)) + 20|0); //@line 155 "./common/StringMap.h"
 HEAP8[$1>>0] = 0; //@line 155 "./common/StringMap.h"
 $2 = ((($this)) + 32|0); //@line 155 "./common/StringMap.h"
 HEAP8[$2>>0] = 0; //@line 155 "./common/StringMap.h"
 $3 = ((($this)) + 44|0); //@line 155 "./common/StringMap.h"
 HEAP8[$3>>0] = 0; //@line 155 "./common/StringMap.h"
 $4 = ((($this)) + 56|0); //@line 155 "./common/StringMap.h"
 HEAP8[$4>>0] = 0; //@line 155 "./common/StringMap.h"
 $5 = ((($this)) + 68|0); //@line 155 "./common/StringMap.h"
 HEAP8[$5>>0] = 0; //@line 155 "./common/StringMap.h"
 $6 = ((($this)) + 80|0); //@line 155 "./common/StringMap.h"
 HEAP8[$6>>0] = 0; //@line 155 "./common/StringMap.h"
 $7 = ((($this)) + 92|0); //@line 155 "./common/StringMap.h"
 HEAP8[$7>>0] = 0; //@line 155 "./common/StringMap.h"
 $8 = ((($this)) + 104|0); //@line 155 "./common/StringMap.h"
 HEAP8[$8>>0] = 0; //@line 155 "./common/StringMap.h"
 $9 = ((($this)) + 116|0); //@line 155 "./common/StringMap.h"
 HEAP8[$9>>0] = 0; //@line 155 "./common/StringMap.h"
 $10 = ((($this)) + 128|0); //@line 155 "./common/StringMap.h"
 HEAP8[$10>>0] = 0; //@line 155 "./common/StringMap.h"
 $11 = ((($this)) + 140|0); //@line 155 "./common/StringMap.h"
 HEAP8[$11>>0] = 0; //@line 155 "./common/StringMap.h"
 $12 = ((($this)) + 152|0); //@line 155 "./common/StringMap.h"
 HEAP8[$12>>0] = 0; //@line 155 "./common/StringMap.h"
 $13 = ((($this)) + 164|0); //@line 155 "./common/StringMap.h"
 HEAP8[$13>>0] = 0; //@line 155 "./common/StringMap.h"
 $14 = ((($this)) + 176|0); //@line 155 "./common/StringMap.h"
 HEAP8[$14>>0] = 0; //@line 155 "./common/StringMap.h"
 $15 = ((($this)) + 188|0); //@line 155 "./common/StringMap.h"
 HEAP8[$15>>0] = 0; //@line 155 "./common/StringMap.h"
 $16 = ((($this)) + 200|0); //@line 155 "./common/StringMap.h"
 HEAP8[$16>>0] = 0; //@line 155 "./common/StringMap.h"
 $17 = ((($this)) + 212|0); //@line 155 "./common/StringMap.h"
 HEAP8[$17>>0] = 0; //@line 155 "./common/StringMap.h"
 $18 = ((($this)) + 224|0); //@line 155 "./common/StringMap.h"
 HEAP8[$18>>0] = 0; //@line 155 "./common/StringMap.h"
 $19 = ((($this)) + 236|0); //@line 155 "./common/StringMap.h"
 HEAP8[$19>>0] = 0; //@line 155 "./common/StringMap.h"
 $20 = ((($this)) + 248|0); //@line 155 "./common/StringMap.h"
 HEAP8[$20>>0] = 0; //@line 155 "./common/StringMap.h"
 $21 = ((($this)) + 260|0); //@line 155 "./common/StringMap.h"
 HEAP8[$21>>0] = 0; //@line 155 "./common/StringMap.h"
 $22 = ((($this)) + 272|0); //@line 155 "./common/StringMap.h"
 HEAP8[$22>>0] = 0; //@line 155 "./common/StringMap.h"
 $23 = ((($this)) + 284|0); //@line 155 "./common/StringMap.h"
 HEAP8[$23>>0] = 0; //@line 155 "./common/StringMap.h"
 $24 = ((($this)) + 296|0); //@line 155 "./common/StringMap.h"
 HEAP8[$24>>0] = 0; //@line 155 "./common/StringMap.h"
 $25 = ((($this)) + 308|0); //@line 155 "./common/StringMap.h"
 HEAP8[$25>>0] = 0; //@line 155 "./common/StringMap.h"
 $26 = ((($this)) + 320|0); //@line 155 "./common/StringMap.h"
 HEAP8[$26>>0] = 0; //@line 155 "./common/StringMap.h"
 $27 = ((($this)) + 332|0); //@line 155 "./common/StringMap.h"
 HEAP8[$27>>0] = 0; //@line 155 "./common/StringMap.h"
 $28 = ((($this)) + 344|0); //@line 155 "./common/StringMap.h"
 HEAP8[$28>>0] = 0; //@line 155 "./common/StringMap.h"
 $29 = ((($this)) + 356|0); //@line 155 "./common/StringMap.h"
 HEAP8[$29>>0] = 0; //@line 155 "./common/StringMap.h"
 $30 = ((($this)) + 368|0); //@line 155 "./common/StringMap.h"
 HEAP8[$30>>0] = 0; //@line 155 "./common/StringMap.h"
 $31 = ((($this)) + 380|0); //@line 155 "./common/StringMap.h"
 HEAP8[$31>>0] = 0; //@line 155 "./common/StringMap.h"
 $32 = ((($this)) + 392|0); //@line 155 "./common/StringMap.h"
 HEAP8[$32>>0] = 0; //@line 155 "./common/StringMap.h"
 $33 = ((($this)) + 404|0); //@line 155 "./common/StringMap.h"
 HEAP8[$33>>0] = 0; //@line 155 "./common/StringMap.h"
 $scevgep = ((($this)) + 408|0); //@line 44 "./common/StringMap.h"
 dest=$scevgep; stop=dest+68|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 44 "./common/StringMap.h"
 $34 = $num >>> 3; //@line 46 "./common/StringMap.h"
 $35 = ($34|0)==(0); //@line 48 "./common/StringMap.h"
 if ($35) {
  STACKTOP = sp;return; //@line 50 "./common/StringMap.h"
 } else {
  $i1$05 = 0;
 }
 while(1) {
  $36 = (($entries) + ($i1$05<<3)|0); //@line 49 "./common/StringMap.h"
  $37 = HEAP32[$36>>2]|0; //@line 49 "./common/StringMap.h"
  $38 = (((($entries) + ($i1$05<<3)|0)) + 4|0); //@line 49 "./common/StringMap.h"
  $39 = HEAP32[$38>>2]|0; //@line 49 "./common/StringMap.h"
  $40 = HEAP8[$37>>0]|0; //@line 142 "./common/StringMap.h"
  $41 = ($40<<24>>24)==(0); //@line 142 "./common/StringMap.h"
  if ($41) {
   $hash$0$lcssa$i$i = 5381;
  } else {
   $$02$i$i = $37;$43 = $40;$hash$01$i$i = 5381;
   while(1) {
    $42 = $43 << 24 >> 24; //@line 142 "./common/StringMap.h"
    $44 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
    $45 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
    $46 = (($45) + ($42))|0; //@line 143 "./common/StringMap.h"
    $47 = HEAP8[$44>>0]|0; //@line 142 "./common/StringMap.h"
    $48 = ($47<<24>>24)==(0); //@line 142 "./common/StringMap.h"
    if ($48) {
     $hash$0$lcssa$i$i = $46;
     break;
    } else {
     $$02$i$i = $44;$43 = $47;$hash$01$i$i = $46;
    }
   }
  }
  $i$08$i = 0;
  while(1) {
   $51 = (($i$08$i) + ($hash$0$lcssa$i$i))|0; //@line 112 "./common/StringMap.h"
   $52 = (($51>>>0) % 34)&-1; //@line 112 "./common/StringMap.h"
   $53 = (((($this) + (($52*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $54 = HEAP8[$53>>0]|0; //@line 114 "./common/StringMap.h"
   $55 = ($54<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   $50 = (($i$08$i) + 1)|0; //@line 110 "./common/StringMap.h"
   if ($55) {
    $$lcssa26 = $52;$$lcssa28 = $53;
    label = 8;
    break;
   }
   $49 = ($50>>>0)<(34); //@line 110 "./common/StringMap.h"
   if ($49) {
    $i$08$i = $50;
   } else {
    break;
   }
  }
  if ((label|0) == 8) {
   label = 0;
   HEAP8[$$lcssa28>>0] = 1; //@line 117 "./common/StringMap.h"
   $56 = (($this) + (($$lcssa26*12)|0)|0); //@line 118 "./common/StringMap.h"
   HEAP32[$56>>2] = $37; //@line 118 "./common/StringMap.h"
   $57 = (((($this) + (($$lcssa26*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
   HEAP32[$57>>2] = $39; //@line 119 "./common/StringMap.h"
  }
  $58 = ($39>>>0)>(16); //@line 126 "./common/StringMap.h"
  if ($58) {
   HEAP32[$vararg_buffer>>2] = $37; //@line 128 "./common/StringMap.h"
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
   HEAP32[$vararg_ptr1>>2] = $39; //@line 128 "./common/StringMap.h"
   (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  } else {
   $59 = (((($this)) + 408|0) + ($39<<2)|0); //@line 132 "./common/StringMap.h"
   HEAP32[$59>>2] = $37; //@line 132 "./common/StringMap.h"
  }
  $60 = (($i1$05) + 1)|0; //@line 48 "./common/StringMap.h"
  $exitcond = ($60|0)==($34|0); //@line 48 "./common/StringMap.h"
  if ($exitcond) {
   break;
  } else {
   $i1$05 = $60;
  }
 }
 STACKTOP = sp;return; //@line 50 "./common/StringMap.h"
}
function __ZN4love8graphics6opengl6CanvasC2EiiNS2_6FormatEi($this,$width,$height,$format,$msaa) {
 $this = $this|0;
 $width = $width|0;
 $height = $height|0;
 $format = $format|0;
 $msaa = $msaa|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics7TextureC2Ev($this); //@line 96 "modules/graphics/opengl/Canvas.cpp"
 $0 = ((($this)) + 120|0); //@line 106 "modules/graphics/opengl/Canvas.cpp"
 __THREW__ = 0;
 invoke_vi(4011,($0|0)); //@line 96 "modules/graphics/opengl/Canvas.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $35 = ___cxa_find_matching_catch_2()|0;
  $36 = tempRet0;
  $$0 = $35;$$01 = $36;
  __ZN4love8graphics7TextureD2Ev($this); //@line 141 "modules/graphics/opengl/Canvas.cpp"
  ___resumeException($$0|0);
  // unreachable;
 }
 HEAP32[$this>>2] = (24316); //@line 106 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$0>>2] = (24388); //@line 106 "modules/graphics/opengl/Canvas.cpp"
 $3 = ((($this)) + 124|0); //@line 97 "modules/graphics/opengl/Canvas.cpp"
 $4 = ((($this)) + 144|0); //@line 102 "modules/graphics/opengl/Canvas.cpp"
 ;HEAP32[$3>>2]=0|0;HEAP32[$3+4>>2]=0|0;HEAP32[$3+8>>2]=0|0;HEAP32[$3+12>>2]=0|0;HEAP32[$3+16>>2]=0|0; //@line 98 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$4>>2] = $format; //@line 102 "modules/graphics/opengl/Canvas.cpp"
 $5 = ((($this)) + 152|0); //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$5>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ((($this)) + 156|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$6>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ((($this)) + 160|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$7>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = ((($this)) + 164|0); //@line 103 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$8>>2] = $msaa; //@line 103 "modules/graphics/opengl/Canvas.cpp"
 $9 = ((($this)) + 168|0); //@line 104 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$9>>2] = 0; //@line 104 "modules/graphics/opengl/Canvas.cpp"
 $10 = ((($this)) + 172|0); //@line 105 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$10>>2] = 0; //@line 105 "modules/graphics/opengl/Canvas.cpp"
 $11 = ((($this)) + 8|0); //@line 107 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$11>>2] = $width; //@line 107 "modules/graphics/opengl/Canvas.cpp"
 $12 = ((($this)) + 12|0); //@line 108 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$12>>2] = $height; //@line 108 "modules/graphics/opengl/Canvas.cpp"
 $13 = (+($width|0)); //@line 110 "modules/graphics/opengl/Canvas.cpp"
 $14 = (+($height|0)); //@line 111 "modules/graphics/opengl/Canvas.cpp"
 $15 = ((($this)) + 40|0); //@line 119 "modules/graphics/opengl/Canvas.cpp"
 HEAPF32[$15>>2] = 0.0; //@line 119 "modules/graphics/opengl/Canvas.cpp"
 $16 = ((($this)) + 44|0); //@line 120 "modules/graphics/opengl/Canvas.cpp"
 HEAPF32[$16>>2] = 0.0; //@line 120 "modules/graphics/opengl/Canvas.cpp"
 $17 = ((($this)) + 60|0); //@line 121 "modules/graphics/opengl/Canvas.cpp"
 HEAPF32[$17>>2] = 0.0; //@line 121 "modules/graphics/opengl/Canvas.cpp"
 $18 = ((($this)) + 64|0); //@line 122 "modules/graphics/opengl/Canvas.cpp"
 HEAPF32[$18>>2] = $14; //@line 122 "modules/graphics/opengl/Canvas.cpp"
 $19 = ((($this)) + 80|0); //@line 123 "modules/graphics/opengl/Canvas.cpp"
 HEAPF32[$19>>2] = $13; //@line 123 "modules/graphics/opengl/Canvas.cpp"
 $20 = ((($this)) + 84|0); //@line 124 "modules/graphics/opengl/Canvas.cpp"
 HEAPF32[$20>>2] = 0.0; //@line 124 "modules/graphics/opengl/Canvas.cpp"
 $21 = ((($this)) + 100|0); //@line 125 "modules/graphics/opengl/Canvas.cpp"
 HEAPF32[$21>>2] = $13; //@line 125 "modules/graphics/opengl/Canvas.cpp"
 $22 = ((($this)) + 104|0); //@line 126 "modules/graphics/opengl/Canvas.cpp"
 HEAPF32[$22>>2] = $14; //@line 126 "modules/graphics/opengl/Canvas.cpp"
 $23 = ((($this)) + 48|0); //@line 129 "modules/graphics/opengl/Canvas.cpp"
 HEAPF32[$23>>2] = 0.0; //@line 129 "modules/graphics/opengl/Canvas.cpp"
 $24 = ((($this)) + 52|0); //@line 130 "modules/graphics/opengl/Canvas.cpp"
 HEAPF32[$24>>2] = 0.0; //@line 130 "modules/graphics/opengl/Canvas.cpp"
 $25 = ((($this)) + 68|0); //@line 131 "modules/graphics/opengl/Canvas.cpp"
 HEAPF32[$25>>2] = 0.0; //@line 131 "modules/graphics/opengl/Canvas.cpp"
 $26 = ((($this)) + 72|0); //@line 132 "modules/graphics/opengl/Canvas.cpp"
 HEAPF32[$26>>2] = 1.0; //@line 132 "modules/graphics/opengl/Canvas.cpp"
 $27 = ((($this)) + 88|0); //@line 133 "modules/graphics/opengl/Canvas.cpp"
 HEAPF32[$27>>2] = 1.0; //@line 133 "modules/graphics/opengl/Canvas.cpp"
 $28 = ((($this)) + 92|0); //@line 134 "modules/graphics/opengl/Canvas.cpp"
 HEAPF32[$28>>2] = 0.0; //@line 134 "modules/graphics/opengl/Canvas.cpp"
 $29 = ((($this)) + 108|0); //@line 135 "modules/graphics/opengl/Canvas.cpp"
 HEAPF32[$29>>2] = 1.0; //@line 135 "modules/graphics/opengl/Canvas.cpp"
 $30 = ((($this)) + 112|0); //@line 136 "modules/graphics/opengl/Canvas.cpp"
 HEAPF32[$30>>2] = 1.0; //@line 136 "modules/graphics/opengl/Canvas.cpp"
 __THREW__ = 0;
 (invoke_ii(1723,($this|0))|0); //@line 138 "modules/graphics/opengl/Canvas.cpp"
 $31 = __THREW__; __THREW__ = 0;
 $32 = $31&1;
 if (!($32)) {
  $33 = HEAP32[158277]|0; //@line 140 "modules/graphics/opengl/Canvas.cpp"
  $34 = (($33) + 1)|0; //@line 140 "modules/graphics/opengl/Canvas.cpp"
  HEAP32[158277] = $34; //@line 140 "modules/graphics/opengl/Canvas.cpp"
  return; //@line 141 "modules/graphics/opengl/Canvas.cpp"
 }
 $37 = ___cxa_find_matching_catch_2()|0;
 $38 = tempRet0;
 $39 = HEAP32[$5>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $40 = ($39|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $41 = $39; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($40)) {
  $42 = HEAP32[$6>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $43 = ($42|0)==($39|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($43)) {
   $scevgep$i$i$i = ((($42)) + -4|0);
   $44 = $scevgep$i$i$i;
   $45 = (($44) - ($41))|0;
   $46 = $45 >>> 2;
   $47 = $46 ^ -1;
   $scevgep4$i$i$i = (($42) + ($47<<2)|0);
   HEAP32[$6>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($39); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 __ZN4love8graphics8VolatileD2Ev($0); //@line 141 "modules/graphics/opengl/Canvas.cpp"
 $$0 = $37;$$01 = $38;
 __ZN4love8graphics7TextureD2Ev($this); //@line 141 "modules/graphics/opengl/Canvas.cpp"
 ___resumeException($$0|0);
 // unreachable;
}
function __ZN4love8graphics6opengl6Canvas9setupGrabEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 80|0;
 $1 = sp + 64|0;
 $2 = sp;
 $3 = HEAP32[158272]|0; //@line 386 "modules/graphics/opengl/Canvas.cpp"
 $4 = ($3|0)==($this|0); //@line 386 "modules/graphics/opengl/Canvas.cpp"
 if ($4) {
  STACKTOP = sp;return; //@line 407 "modules/graphics/opengl/Canvas.cpp"
 }
 $5 = ($3|0)==(0|0); //@line 390 "modules/graphics/opengl/Canvas.cpp"
 if ($5) {
  __ZNK4love8graphics6opengl6OpenGL11getViewportEv($0,634044); //@line 396 "modules/graphics/opengl/Canvas.cpp"
  ;HEAP32[633092>>2]=HEAP32[$0>>2]|0;HEAP32[633092+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[633092+8>>2]=HEAP32[$0+8>>2]|0;HEAP32[633092+12>>2]=HEAP32[$0+12>>2]|0; //@line 396 "modules/graphics/opengl/Canvas.cpp"
 } else {
  __ZN4love8graphics6opengl6Canvas8stopGrabEb($3,1); //@line 393 "modules/graphics/opengl/Canvas.cpp"
 }
 HEAP32[158272] = $this; //@line 399 "modules/graphics/opengl/Canvas.cpp"
 $6 = ((($this)) + 124|0); //@line 402 "modules/graphics/opengl/Canvas.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 402 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$7); //@line 402 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$1>>2] = 0; //@line 403 "modules/graphics/opengl/Canvas.cpp"
 $8 = ((($1)) + 4|0); //@line 403 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$8>>2] = 0; //@line 403 "modules/graphics/opengl/Canvas.cpp"
 $9 = ((($1)) + 8|0); //@line 403 "modules/graphics/opengl/Canvas.cpp"
 $10 = ((($this)) + 8|0); //@line 403 "modules/graphics/opengl/Canvas.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 403 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$9>>2] = $11; //@line 403 "modules/graphics/opengl/Canvas.cpp"
 $12 = ((($1)) + 12|0); //@line 403 "modules/graphics/opengl/Canvas.cpp"
 $13 = ((($this)) + 12|0); //@line 403 "modules/graphics/opengl/Canvas.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 403 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$12>>2] = $14; //@line 403 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL11setViewportERKNS2_8ViewportE(634044,$1); //@line 403 "modules/graphics/opengl/Canvas.cpp"
 $15 = HEAP32[$10>>2]|0; //@line 406 "modules/graphics/opengl/Canvas.cpp"
 $16 = (+($15|0)); //@line 406 "modules/graphics/opengl/Canvas.cpp"
 $17 = HEAP32[$13>>2]|0; //@line 406 "modules/graphics/opengl/Canvas.cpp"
 $18 = (+($17|0)); //@line 406 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love7Matrix45orthoEffff($2,0.0,$16,0.0,$18); //@line 406 "modules/graphics/opengl/Canvas.cpp"
 $19 = HEAP32[(634060)>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = HEAP32[(634064)>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = ($19>>>0)<($20>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($21) {
  dest=$19; src=$2; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $22 = HEAP32[(634060)>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = ((($22)) + 64|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[(634060)>>2] = $23; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 } else {
  __THREW__ = 0;
  invoke_vii(4012,((634056)|0),($2|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $24 = __THREW__; __THREW__ = 0;
  $25 = $24&1;
  if ($25) {
   $26 = ___cxa_find_matching_catch_2()|0;
   $27 = tempRet0;
   __ZN4love7Matrix4D2Ev($2); //@line 406 "modules/graphics/opengl/Canvas.cpp"
   ___resumeException($26|0);
   // unreachable;
  }
 }
 __ZN4love7Matrix4D2Ev($2); //@line 406 "modules/graphics/opengl/Canvas.cpp"
 STACKTOP = sp;return; //@line 407 "modules/graphics/opengl/Canvas.cpp"
}
function __ZNSt3__16vectorIN4love7Matrix4ENS_9allocatorIS2_EEE21__push_back_slow_pathIS2_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$01$i$i = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__v$sroa$5$0 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 6; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(67108863); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = $$in; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = HEAP32[$8>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $10 >> 6; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($11>>>0)<(33554431); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $17 = $10 >> 5; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ($17>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $19 = $18 ? $5 : $17; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = (($20) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = $21 >> 6; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = ($19|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $24 = $20; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($23) {
   $$0$i2 = 0;$28 = 0;$29 = $22;$36 = $24;
  } else {
   $$0$i1 = $19;$54 = $24;$55 = $22;
   label = 6;
  }
 } else {
  $13 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = $14 >> 6; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = $13; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $$0$i1 = 67108863;$54 = $16;$55 = $15;
  label = 6;
 }
 if ((label|0) == 6) {
  $25 = $$0$i1 << 6; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (__Znwj($25)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$28 = $26;$29 = $55;$36 = $54;
 }
 $27 = (($28) + ($29<<6)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = $27; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $31 = (($28) + ($$0$i2<<6)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $32 = $31; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 dest=$27; src=$__x; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = ((($27)) + 64|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $34 = $33; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $35 = ($36|0)==($7|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($35) {
  $44 = $30;$46 = $$in;$48 = $36;
 } else {
  $$01$i$i = $36;$38 = $27;$__v$sroa$5$0 = $30;
  while(1) {
   $37 = ((($38)) + -64|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $39 = ((($$01$i$i)) + -64|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   dest=$37; src=$39; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $40 = $__v$sroa$5$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $41 = ((($40)) + -64|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $42 = $41; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $43 = ($39|0)==($7|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if ($43) {
    $$lcssa = $42;
    break;
   } else {
    $$01$i$i = $39;$38 = $41;$__v$sroa$5$0 = $42;
   }
  }
  $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $$pre3 = HEAP32[$0>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $44 = $$lcssa;$46 = $$pre$i;$48 = $$pre3;
 }
 HEAP32[$this>>2] = $44; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $34; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$8>>2] = $32; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $45 = $46; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $47 = ($48|0)==($45|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if (!($47)) {
  $50 = $48;
  while(1) {
   $49 = ((($50)) + -64|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   __ZN4love7Matrix4D2Ev($49); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $51 = ($49|0)==($45|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($51) {
    break;
   } else {
    $50 = $49;
   }
  }
 }
 $52 = ($46|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($52) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $53 = $46; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($53); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl6Canvas9startGrabERKNSt3__16vectorIPS2_NS3_9allocatorIS5_EEEE($this,$canvases) {
 $this = $this|0;
 $canvases = $canvases|0;
 var $$$i = 0, $$$i19 = 0, $$$i19$us = 0, $$0 = 0, $$0$i = 0, $$0$i20 = 0, $$0$i20$us = 0, $$02 = 0, $$cast = 0, $$hasSRGBcanvas$0$off0 = 0, $$hasSRGBcanvas$0$off0$us = 0, $$in = 0, $$in40 = 0, $$pre = 0, $$pre39 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0;
 var $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0;
 var $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $canvaseschanged$0$off0$lcssa = 0;
 var $canvaseschanged$0$off031 = 0, $canvaseschanged$0$off031$us = 0, $canvaseschanged$1$off0 = 0, $canvaseschanged$1$off0$us = 0, $drawbuffers = 0, $hasSRGBcanvas$0$off0$lcssa = 0, $hasSRGBcanvas$0$off029 = 0, $hasSRGBcanvas$0$off029$us = 0, $i$030 = 0, $i$030$us = 0, $i1$028 = 0, $lpad$phi$index = 0, $lpad$phi$index10 = 0, $not$ = 0, $not$$us = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i10 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i11 = 0, $vararg_buffer = 0;
 var $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $drawbuffers = sp + 44|0;
 $0 = sp + 40|0;
 $1 = sp + 36|0;
 $2 = ((($canvases)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$canvases>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ((($this)) + 152|0); //@line 413 "modules/graphics/opengl/Canvas.cpp"
 $7 = ((($this)) + 156|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = HEAP32[$7>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = HEAP32[$6>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($8) - ($9))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($5|0)!=($10|0); //@line 413 "modules/graphics/opengl/Canvas.cpp"
 $12 = ((($this)) + 144|0); //@line 414 "modules/graphics/opengl/Canvas.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 414 "modules/graphics/opengl/Canvas.cpp"
 switch ($13|0) {
 case 0:  {
  $14 = (__ZN4love8graphics14isGammaCorrectEv()|0); //@line 669 "modules/graphics/opengl/Canvas.cpp"
  if ($14) {
   $$0$i = 16;
  } else {
   $15 = HEAP8[1252012]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $16 = ($15<<24>>24)==(0); //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $17 = HEAP8[1252013]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $18 = HEAP8[1252284]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $19 = $18 | $17; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $20 = HEAP8[1252377]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $21 = $19 | $20; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $22 = ($21<<24>>24)!=(0); //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $23 = $16 | $22; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $$$i = $23 ? 7 : 2; //@line 675 "modules/graphics/opengl/Canvas.cpp"
   $$0$i = $$$i;
  }
  break;
 }
 case 1:  {
  $$0$i = 12;
  break;
 }
 default: {
  $$0$i = $13;
 }
 }
 $24 = ($$0$i|0)==(16); //@line 414 "modules/graphics/opengl/Canvas.cpp"
 $25 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $26 = HEAP32[$canvases>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = (($25) - ($26))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $28 = $27 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $29 = ($28|0)==(0); //@line 416 "modules/graphics/opengl/Canvas.cpp"
 if (!($29)) {
  $30 = (__ZNK4love8graphics6opengl6OpenGL19getMaxRenderTargetsEv(634044)|0); //@line 418 "modules/graphics/opengl/Canvas.cpp"
  $31 = ($28|0)<($30|0); //@line 418 "modules/graphics/opengl/Canvas.cpp"
  if (!($31)) {
   $32 = (___cxa_allocate_exception(16)|0); //@line 419 "modules/graphics/opengl/Canvas.cpp"
   $33 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $34 = HEAP32[$canvases>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $35 = (($33) - ($34))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $36 = $35 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $37 = (($36) + 1)|0; //@line 419 "modules/graphics/opengl/Canvas.cpp"
   __THREW__ = 0;
   HEAP32[$vararg_buffer>>2] = $37; //@line 419 "modules/graphics/opengl/Canvas.cpp"
   invoke_viii(3349,($32|0),(434169|0),($vararg_buffer|0)); //@line 419 "modules/graphics/opengl/Canvas.cpp"
   $38 = __THREW__; __THREW__ = 0;
   $39 = $38&1;
   if (!($39)) {
    ___cxa_throw(($32|0),(96|0),(344|0)); //@line 419 "modules/graphics/opengl/Canvas.cpp"
    // unreachable; //@line 419 "modules/graphics/opengl/Canvas.cpp"
   }
   $40 = ___cxa_find_matching_catch_2()|0;
   $41 = tempRet0;
   ___cxa_free_exception(($32|0)); //@line 419 "modules/graphics/opengl/Canvas.cpp"
   $$0 = $41;$$02 = $40;
   ___resumeException($$02|0);
   // unreachable;
  }
  $42 = ((($this)) + 168|0); //@line 421 "modules/graphics/opengl/Canvas.cpp"
  $43 = HEAP32[$42>>2]|0; //@line 421 "modules/graphics/opengl/Canvas.cpp"
  $44 = ($43|0)==(0); //@line 421 "modules/graphics/opengl/Canvas.cpp"
  if (!($44)) {
   $45 = (___cxa_allocate_exception(16)|0); //@line 422 "modules/graphics/opengl/Canvas.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($45|0),(434225|0),($vararg_buffer1|0)); //@line 422 "modules/graphics/opengl/Canvas.cpp"
   $46 = __THREW__; __THREW__ = 0;
   $47 = $46&1;
   if (!($47)) {
    ___cxa_throw(($45|0),(96|0),(344|0)); //@line 422 "modules/graphics/opengl/Canvas.cpp"
    // unreachable; //@line 422 "modules/graphics/opengl/Canvas.cpp"
   }
   $48 = ___cxa_find_matching_catch_2()|0;
   $49 = tempRet0;
   ___cxa_free_exception(($45|0)); //@line 422 "modules/graphics/opengl/Canvas.cpp"
   $$0 = $49;$$02 = $48;
   ___resumeException($$02|0);
   // unreachable;
  }
 }
 $50 = (__ZNK4love8graphics6opengl6OpenGL19getMaxRenderTargetsEv(634044)|0); //@line 814 "modules/graphics/opengl/Canvas.cpp"
 $51 = ($50|0)>(1); //@line 814 "modules/graphics/opengl/Canvas.cpp"
 if ($51) {
  $52 = HEAP8[1252013]|0; //@line 814 "modules/graphics/opengl/Canvas.cpp"
  $53 = HEAP8[1252002]|0; //@line 814 "modules/graphics/opengl/Canvas.cpp"
  $54 = $53 | $52; //@line 814 "modules/graphics/opengl/Canvas.cpp"
  $55 = ($54<<24>>24)==(0); //@line 814 "modules/graphics/opengl/Canvas.cpp"
  if ($55) {
   $56 = HEAP8[1252065]|0; //@line 814 "modules/graphics/opengl/Canvas.cpp"
   $57 = ($56<<24>>24)!=(0); //@line 814 "modules/graphics/opengl/Canvas.cpp"
   $263 = $57;
  } else {
   $263 = 1;
  }
 } else {
  $263 = 0;
 }
 $58 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $59 = HEAP32[$canvases>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $60 = ($58|0)==($59|0); //@line 427 "modules/graphics/opengl/Canvas.cpp"
 $61 = $59; //@line 427 "modules/graphics/opengl/Canvas.cpp"
 L25: do {
  if ($60) {
   $canvaseschanged$0$off0$lcssa = $11;$hasSRGBcanvas$0$off0$lcssa = $24;
  } else {
   $62 = ((($this)) + 8|0); //@line 429 "modules/graphics/opengl/Canvas.cpp"
   $63 = ((($this)) + 12|0); //@line 429 "modules/graphics/opengl/Canvas.cpp"
   L27: do {
    if ($263) {
     $65 = $61;$canvaseschanged$0$off031$us = $11;$hasSRGBcanvas$0$off029$us = $24;$i$030$us = 0;
     while(1) {
      $64 = (($65) + ($i$030$us<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $66 = HEAP32[$64>>2]|0; //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $67 = HEAP32[$66>>2]|0; //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $68 = ((($67)) + 24|0); //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $69 = HEAP32[$68>>2]|0; //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $70 = (FUNCTION_TABLE_ii[$69 & 8191]($66)|0); //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $71 = HEAP32[$62>>2]|0; //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $72 = ($70|0)==($71|0); //@line 429 "modules/graphics/opengl/Canvas.cpp"
      if (!($72)) {
       label = 32;
       break L27;
      }
      $73 = HEAP32[$canvases>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $74 = (($73) + ($i$030$us<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $75 = HEAP32[$74>>2]|0; //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $76 = HEAP32[$75>>2]|0; //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $77 = ((($76)) + 28|0); //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $78 = HEAP32[$77>>2]|0; //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $79 = (FUNCTION_TABLE_ii[$78 & 8191]($75)|0); //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $80 = HEAP32[$63>>2]|0; //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $81 = ($79|0)==($80|0); //@line 429 "modules/graphics/opengl/Canvas.cpp"
      if (!($81)) {
       label = 32;
       break L27;
      }
      $82 = HEAP32[$canvases>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $83 = (($82) + ($i$030$us<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $84 = HEAP32[$83>>2]|0; //@line 432 "modules/graphics/opengl/Canvas.cpp"
      $85 = ((($84)) + 144|0); //@line 112 "modules/graphics/opengl/Canvas.h"
      $86 = HEAP32[$85>>2]|0; //@line 112 "modules/graphics/opengl/Canvas.h"
      $87 = ((($84)) + 168|0); //@line 117 "modules/graphics/opengl/Canvas.h"
      $88 = HEAP32[$87>>2]|0; //@line 117 "modules/graphics/opengl/Canvas.h"
      $89 = ($88|0)==(0); //@line 437 "modules/graphics/opengl/Canvas.cpp"
      if (!($89)) {
       label = 40;
       break L27;
      }
      if ($canvaseschanged$0$off031$us) {
       $canvaseschanged$1$off0$us = 1;
      } else {
       $90 = HEAP32[$6>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $91 = (($90) + ($i$030$us<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $92 = HEAP32[$91>>2]|0; //@line 440 "modules/graphics/opengl/Canvas.cpp"
       $not$$us = ($84|0)!=($92|0); //@line 440 "modules/graphics/opengl/Canvas.cpp"
       $canvaseschanged$1$off0$us = $not$$us;
      }
      switch ($86|0) {
      case 0:  {
       $93 = (__ZN4love8graphics14isGammaCorrectEv()|0); //@line 669 "modules/graphics/opengl/Canvas.cpp"
       if ($93) {
        $$0$i20$us = 16;
       } else {
        $94 = HEAP8[1252012]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $95 = ($94<<24>>24)==(0); //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $96 = HEAP8[1252013]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $97 = HEAP8[1252284]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $98 = $97 | $96; //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $99 = HEAP8[1252377]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $100 = $98 | $99; //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $101 = ($100<<24>>24)!=(0); //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $102 = $95 | $101; //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $$$i19$us = $102 ? 7 : 2; //@line 675 "modules/graphics/opengl/Canvas.cpp"
        $$0$i20$us = $$$i19$us;
       }
       break;
      }
      case 1:  {
       $$0$i20$us = 12;
       break;
      }
      default: {
       $$0$i20$us = $86;
      }
      }
      $103 = ($$0$i20$us|0)==(16); //@line 443 "modules/graphics/opengl/Canvas.cpp"
      $$hasSRGBcanvas$0$off0$us = $hasSRGBcanvas$0$off029$us | $103; //@line 443 "modules/graphics/opengl/Canvas.cpp"
      $104 = (($i$030$us) + 1)|0; //@line 427 "modules/graphics/opengl/Canvas.cpp"
      $105 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $106 = HEAP32[$canvases>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $107 = (($105) - ($106))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $108 = $107 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $109 = ($104>>>0)<($108>>>0); //@line 427 "modules/graphics/opengl/Canvas.cpp"
      $110 = $106; //@line 427 "modules/graphics/opengl/Canvas.cpp"
      if ($109) {
       $65 = $110;$canvaseschanged$0$off031$us = $canvaseschanged$1$off0$us;$hasSRGBcanvas$0$off029$us = $$hasSRGBcanvas$0$off0$us;$i$030$us = $104;
      } else {
       $canvaseschanged$0$off0$lcssa = $canvaseschanged$1$off0$us;$hasSRGBcanvas$0$off0$lcssa = $$hasSRGBcanvas$0$off0$us;
       break L25;
      }
     }
    } else {
     $116 = $61;$canvaseschanged$0$off031 = $11;$hasSRGBcanvas$0$off029 = $24;$i$030 = 0;
     while(1) {
      $115 = (($116) + ($i$030<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $117 = HEAP32[$115>>2]|0; //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $118 = HEAP32[$117>>2]|0; //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $119 = ((($118)) + 24|0); //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $120 = HEAP32[$119>>2]|0; //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $121 = (FUNCTION_TABLE_ii[$120 & 8191]($117)|0); //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $122 = HEAP32[$62>>2]|0; //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $123 = ($121|0)==($122|0); //@line 429 "modules/graphics/opengl/Canvas.cpp"
      if (!($123)) {
       label = 32;
       break L27;
      }
      $124 = HEAP32[$canvases>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $125 = (($124) + ($i$030<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $126 = HEAP32[$125>>2]|0; //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $127 = HEAP32[$126>>2]|0; //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $128 = ((($127)) + 28|0); //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $129 = HEAP32[$128>>2]|0; //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $130 = (FUNCTION_TABLE_ii[$129 & 8191]($126)|0); //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $131 = HEAP32[$63>>2]|0; //@line 429 "modules/graphics/opengl/Canvas.cpp"
      $132 = ($130|0)==($131|0); //@line 429 "modules/graphics/opengl/Canvas.cpp"
      if (!($132)) {
       label = 32;
       break L27;
      }
      $138 = HEAP32[$canvases>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $139 = (($138) + ($i$030<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $140 = HEAP32[$139>>2]|0; //@line 432 "modules/graphics/opengl/Canvas.cpp"
      $141 = ((($140)) + 144|0); //@line 112 "modules/graphics/opengl/Canvas.h"
      $142 = HEAP32[$141>>2]|0; //@line 112 "modules/graphics/opengl/Canvas.h"
      $143 = HEAP32[$12>>2]|0; //@line 434 "modules/graphics/opengl/Canvas.cpp"
      $144 = ($142|0)==($143|0); //@line 434 "modules/graphics/opengl/Canvas.cpp"
      if (!($144)) {
       break;
      }
      $150 = ((($140)) + 168|0); //@line 117 "modules/graphics/opengl/Canvas.h"
      $151 = HEAP32[$150>>2]|0; //@line 117 "modules/graphics/opengl/Canvas.h"
      $152 = ($151|0)==(0); //@line 437 "modules/graphics/opengl/Canvas.cpp"
      if (!($152)) {
       label = 40;
       break L27;
      }
      if ($canvaseschanged$0$off031) {
       $canvaseschanged$1$off0 = 1;
      } else {
       $158 = HEAP32[$6>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $159 = (($158) + ($i$030<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $160 = HEAP32[$159>>2]|0; //@line 440 "modules/graphics/opengl/Canvas.cpp"
       $not$ = ($140|0)!=($160|0); //@line 440 "modules/graphics/opengl/Canvas.cpp"
       $canvaseschanged$1$off0 = $not$;
      }
      switch ($142|0) {
      case 0:  {
       $161 = (__ZN4love8graphics14isGammaCorrectEv()|0); //@line 669 "modules/graphics/opengl/Canvas.cpp"
       if ($161) {
        $$0$i20 = 16;
       } else {
        $162 = HEAP8[1252012]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $163 = ($162<<24>>24)==(0); //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $164 = HEAP8[1252013]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $165 = HEAP8[1252284]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $166 = $165 | $164; //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $167 = HEAP8[1252377]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $168 = $166 | $167; //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $169 = ($168<<24>>24)!=(0); //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $170 = $163 | $169; //@line 671 "modules/graphics/opengl/Canvas.cpp"
        $$$i19 = $170 ? 7 : 2; //@line 675 "modules/graphics/opengl/Canvas.cpp"
        $$0$i20 = $$$i19;
       }
       break;
      }
      case 1:  {
       $$0$i20 = 12;
       break;
      }
      default: {
       $$0$i20 = $142;
      }
      }
      $171 = ($$0$i20|0)==(16); //@line 443 "modules/graphics/opengl/Canvas.cpp"
      $$hasSRGBcanvas$0$off0 = $hasSRGBcanvas$0$off029 | $171; //@line 443 "modules/graphics/opengl/Canvas.cpp"
      $172 = (($i$030) + 1)|0; //@line 427 "modules/graphics/opengl/Canvas.cpp"
      $173 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $174 = HEAP32[$canvases>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $175 = (($173) - ($174))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $176 = $175 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $177 = ($172>>>0)<($176>>>0); //@line 427 "modules/graphics/opengl/Canvas.cpp"
      $178 = $174; //@line 427 "modules/graphics/opengl/Canvas.cpp"
      if ($177) {
       $116 = $178;$canvaseschanged$0$off031 = $canvaseschanged$1$off0;$hasSRGBcanvas$0$off029 = $$hasSRGBcanvas$0$off0;$i$030 = $172;
      } else {
       $canvaseschanged$0$off0$lcssa = $canvaseschanged$1$off0;$hasSRGBcanvas$0$off0$lcssa = $$hasSRGBcanvas$0$off0;
       break L25;
      }
     }
     $145 = (___cxa_allocate_exception(16)|0); //@line 435 "modules/graphics/opengl/Canvas.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($145|0),(434320|0),($vararg_buffer5|0)); //@line 435 "modules/graphics/opengl/Canvas.cpp"
     $146 = __THREW__; __THREW__ = 0;
     $147 = $146&1;
     if (!($147)) {
      ___cxa_throw(($145|0),(96|0),(344|0)); //@line 435 "modules/graphics/opengl/Canvas.cpp"
      // unreachable; //@line 435 "modules/graphics/opengl/Canvas.cpp"
     }
     $148 = ___cxa_find_matching_catch_2()|0;
     $149 = tempRet0;
     ___cxa_free_exception(($145|0)); //@line 435 "modules/graphics/opengl/Canvas.cpp"
     $$0 = $149;$$02 = $148;
     ___resumeException($$02|0);
     // unreachable;
    }
   } while(0);
   if ((label|0) == 32) {
    $133 = (___cxa_allocate_exception(16)|0); //@line 430 "modules/graphics/opengl/Canvas.cpp"
    __THREW__ = 0;
    invoke_viii(3349,($133|0),(434276|0),($vararg_buffer3|0)); //@line 430 "modules/graphics/opengl/Canvas.cpp"
    $134 = __THREW__; __THREW__ = 0;
    $135 = $134&1;
    if (!($135)) {
     ___cxa_throw(($133|0),(96|0),(344|0)); //@line 430 "modules/graphics/opengl/Canvas.cpp"
     // unreachable; //@line 430 "modules/graphics/opengl/Canvas.cpp"
    }
    $136 = ___cxa_find_matching_catch_2()|0;
    $137 = tempRet0;
    ___cxa_free_exception(($133|0)); //@line 430 "modules/graphics/opengl/Canvas.cpp"
    $$0 = $137;$$02 = $136;
    ___resumeException($$02|0);
    // unreachable;
   }
   else if ((label|0) == 40) {
    $153 = (___cxa_allocate_exception(16)|0); //@line 438 "modules/graphics/opengl/Canvas.cpp"
    __THREW__ = 0;
    invoke_viii(3349,($153|0),(434225|0),($vararg_buffer7|0)); //@line 438 "modules/graphics/opengl/Canvas.cpp"
    $154 = __THREW__; __THREW__ = 0;
    $155 = $154&1;
    if (!($155)) {
     ___cxa_throw(($153|0),(96|0),(344|0)); //@line 438 "modules/graphics/opengl/Canvas.cpp"
     // unreachable; //@line 438 "modules/graphics/opengl/Canvas.cpp"
    }
    $156 = ___cxa_find_matching_catch_2()|0;
    $157 = tempRet0;
    ___cxa_free_exception(($153|0)); //@line 438 "modules/graphics/opengl/Canvas.cpp"
    $$0 = $157;$$02 = $156;
    ___resumeException($$02|0);
    // unreachable;
   }
  }
 } while(0);
 __ZN4love8graphics6opengl6Canvas9setupGrabEv($this); //@line 449 "modules/graphics/opengl/Canvas.cpp"
 $111 = HEAP8[1251994]|0; //@line 452 "modules/graphics/opengl/Canvas.cpp"
 $112 = HEAP8[1252225]|0; //@line 452 "modules/graphics/opengl/Canvas.cpp"
 $113 = $112 | $111; //@line 452 "modules/graphics/opengl/Canvas.cpp"
 $114 = ($113<<24>>24)==(0); //@line 452 "modules/graphics/opengl/Canvas.cpp"
 do {
  if (!($114)) {
   $179 = (__ZNK4love8graphics6opengl6OpenGL18hasFramebufferSRGBEv(634044)|0); //@line 454 "modules/graphics/opengl/Canvas.cpp"
   if ($hasSRGBcanvas$0$off0$lcssa) {
    if ($179) {
     break;
    }
    __ZN4love8graphics6opengl6OpenGL18setFramebufferSRGBEb(634044,1); //@line 455 "modules/graphics/opengl/Canvas.cpp"
    break;
   } else {
    if (!($179)) {
     break;
    }
    __ZN4love8graphics6opengl6OpenGL18setFramebufferSRGBEb(634044,0); //@line 457 "modules/graphics/opengl/Canvas.cpp"
    break;
   }
  }
 } while(0);
 if (!($canvaseschanged$0$off0$lcssa)) {
  STACKTOP = sp;return; //@line 486 "modules/graphics/opengl/Canvas.cpp"
 }
 HEAP32[$drawbuffers>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $180 = ((($drawbuffers)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$180>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $181 = ((($drawbuffers)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$181>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $182 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $183 = HEAP32[$canvases>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $184 = (($182) - ($183))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $185 = $184 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $186 = (($185) + 1)|0; //@line 466 "modules/graphics/opengl/Canvas.cpp"
 $187 = ((($drawbuffers)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $188 = ($186|0)==(0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($188) {
  HEAP32[$0>>2] = 36064; //@line 468 "modules/graphics/opengl/Canvas.cpp"
  label = 61;
 } else {
  $189 = $186 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  __THREW__ = 0;
  $190 = (invoke_ii(3348,($189|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $191 = __THREW__; __THREW__ = 0;
  $192 = $191&1;
  if ($192) {
   label = 66;
  } else {
   $193 = $190; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $194 = (($190) + ($186<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $195 = $194; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   HEAP32[$drawbuffers>>2] = $193; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$180>>2] = $193; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$187>>2] = $195; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$0>>2] = 36064; //@line 468 "modules/graphics/opengl/Canvas.cpp"
   $196 = ($190>>>0)<($194>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($196) {
    HEAP32[$190>>2] = 36064; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $197 = ((($190)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$180>>2] = $197; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$in = $183;$201 = $182;
    label = 63;
   } else {
    label = 61;
   }
  }
 }
 if ((label|0) == 61) {
  __THREW__ = 0;
  invoke_vii(3597,($drawbuffers|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $198 = __THREW__; __THREW__ = 0;
  $199 = $198&1;
  if ($199) {
   label = 66;
  } else {
   $$pre = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$pre39 = HEAP32[$canvases>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$in = $$pre39;$201 = $$pre;
   label = 63;
  }
 }
 L91: do {
  if ((label|0) == 63) {
   $200 = (($201) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $202 = ($200|0)>(0); //@line 471 "modules/graphics/opengl/Canvas.cpp"
   L93: do {
    if ($202) {
     $$in40 = $$in;$i1$028 = 0;
     while(1) {
      $223 = $$in40; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $224 = (($i1$028) + 36065)|0; //@line 473 "modules/graphics/opengl/Canvas.cpp"
      $225 = (($223) + ($i1$028<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $226 = HEAP32[$225>>2]|0; //@line 474 "modules/graphics/opengl/Canvas.cpp"
      $227 = HEAP32[$226>>2]|0; //@line 474 "modules/graphics/opengl/Canvas.cpp"
      $228 = ((($227)) + 52|0); //@line 474 "modules/graphics/opengl/Canvas.cpp"
      $229 = HEAP32[$228>>2]|0; //@line 474 "modules/graphics/opengl/Canvas.cpp"
      __THREW__ = 0;
      $230 = (invoke_ii($229|0,($226|0))|0); //@line 474 "modules/graphics/opengl/Canvas.cpp"
      $231 = __THREW__; __THREW__ = 0;
      $232 = $231&1;
      if ($232) {
       break;
      }
      $233 = HEAP32[$230>>2]|0; //@line 474 "modules/graphics/opengl/Canvas.cpp"
      $234 = HEAP32[164777]|0; //@line 672 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_viiiii($234|0,36160,($224|0),3553,($233|0),0); //@line 672 "./libraries/glad/gladfuncs.hpp"
      $235 = __THREW__; __THREW__ = 0;
      $236 = $235&1;
      if ($236) {
       break;
      }
      HEAP32[$1>>2] = $224; //@line 476 "modules/graphics/opengl/Canvas.cpp"
      $237 = HEAP32[$180>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $238 = HEAP32[$187>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $239 = ($237>>>0)<($238>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($239) {
       HEAP32[$237>>2] = $224; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $240 = ((($237)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$180>>2] = $240; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      } else {
       __THREW__ = 0;
       invoke_vii(3597,($drawbuffers|0),($1|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $241 = __THREW__; __THREW__ = 0;
       $242 = $241&1;
       if ($242) {
        break;
       }
      }
      $243 = (($i1$028) + 1)|0; //@line 471 "modules/graphics/opengl/Canvas.cpp"
      $244 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $245 = HEAP32[$canvases>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $246 = (($244) - ($245))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $247 = $246 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $248 = ($243|0)<($247|0); //@line 471 "modules/graphics/opengl/Canvas.cpp"
      if ($248) {
       $$in40 = $245;$i1$028 = $243;
      } else {
       break L93;
      }
     }
     $210 = ___cxa_find_matching_catch_2()|0;
     $211 = tempRet0;
     $lpad$phi$index = $210;$lpad$phi$index10 = $211;
     break L91;
    }
   } while(0);
   $203 = HEAP32[$180>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $204 = HEAP32[$drawbuffers>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $205 = (($203) - ($204))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $206 = $205 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$cast = $204; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $207 = HEAP32[165046]|0; //@line 500 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   invoke_vii($207|0,($206|0),($$cast|0)); //@line 500 "./libraries/glad/gladfuncs.hpp"
   $208 = __THREW__; __THREW__ = 0;
   $209 = $208&1;
   if ($209) {
    label = 66;
   } else {
    $249 = ($6|0)==($canvases|0); //@line 1359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($249)) {
     $250 = HEAP32[$canvases>>2]|0; //@line 1362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $251 = HEAP32[$2>>2]|0; //@line 1362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     __THREW__ = 0;
     invoke_viii(4013,($6|0),($250|0),($251|0)); //@line 1362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $252 = __THREW__; __THREW__ = 0;
     $253 = $252&1;
     if ($253) {
      label = 66;
      break;
     }
    }
    $254 = HEAP32[$drawbuffers>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $255 = ($254|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $256 = $254; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($255)) {
     $257 = HEAP32[$180>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $258 = ($257|0)==($254|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($258)) {
      $scevgep$i$i$i = ((($257)) + -4|0);
      $259 = $scevgep$i$i$i;
      $260 = (($259) - ($256))|0;
      $261 = $260 >>> 2;
      $262 = $261 ^ -1;
      $scevgep4$i$i$i = (($257) + ($262<<2)|0);
      HEAP32[$180>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __ZdlPv($254); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    }
    STACKTOP = sp;return; //@line 486 "modules/graphics/opengl/Canvas.cpp"
   }
  }
 } while(0);
 if ((label|0) == 66) {
  $212 = ___cxa_find_matching_catch_2()|0;
  $213 = tempRet0;
  $lpad$phi$index = $212;$lpad$phi$index10 = $213;
 }
 $214 = HEAP32[$drawbuffers>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $215 = ($214|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $216 = $214; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($215) {
  $$0 = $lpad$phi$index10;$$02 = $lpad$phi$index;
  ___resumeException($$02|0);
  // unreachable;
 }
 $217 = HEAP32[$180>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $218 = ($217|0)==($214|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($218)) {
  $scevgep$i$i$i10 = ((($217)) + -4|0);
  $219 = $scevgep$i$i$i10;
  $220 = (($219) - ($216))|0;
  $221 = $220 >>> 2;
  $222 = $221 ^ -1;
  $scevgep4$i$i$i11 = (($217) + ($222<<2)|0);
  HEAP32[$180>>2] = $scevgep4$i$i$i11; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($214); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $$0 = $lpad$phi$index10;$$02 = $lpad$phi$index;
 ___resumeException($$02|0);
 // unreachable;
}
function __ZNSt3__16vectorIPN4love8graphics6opengl6CanvasENS_9allocatorIS5_EEE6assignIPS5_EENS_9enable_ifIXaasr21__is_forward_iteratorIT_EE5valuesr16is_constructibleIS5_NS_15iterator_traitsISC_E9referenceEEE5valueEvE4typeESC_SC_($this,$__first,$__last) {
 $this = $this|0;
 $__first = $__first|0;
 $__last = $__last|0;
 var $$0$i5 = 0, $$01$i = 0, $$01$i3 = 0, $$__last = 0, $$pre = 0, $$pre$i1 = 0, $$pre8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $7 = 0, $8 = 0, $9 = 0, $__new_last2$i$i = 0, $scevgep$i$i = 0, $scevgep$i$i$i$i = 0, $scevgep4$i$i = 0, $scevgep4$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $__last; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $1 = $__first; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $2 = (($0) - ($1))|0; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $3 = $2 >> 2; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $4 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = HEAP32[$4>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = (($5) - ($6))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = $7 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ($3>>>0)>($8>>>0); //@line 1398 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $6; //@line 1408 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($9)) {
  $11 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $12 = HEAP32[$11>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($6))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = ($3>>>0)>($14>>>0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = (($__first) + ($14<<2)|0); //@line 467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $$__last = $15 ? $16 : $__last; //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = $$__last; //@line 1763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $18 = (($17) - ($1))|0; //@line 1763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $19 = $18 >> 2; //@line 1763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $20 = $6; //@line 1764 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  _memmove(($20|0),($__first|0),($18|0))|0; //@line 1764 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $21 = (($10) + ($19<<2)|0); //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($15)) {
   $29 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $30 = ($29|0)==($21|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($30) {
    return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $__new_last2$i$i = $21; //@line 818 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $scevgep$i$i = ((($29)) + -4|0);
   $31 = $scevgep$i$i;
   $32 = (($31) - ($__new_last2$i$i))|0;
   $33 = $32 >>> 2;
   $34 = $33 ^ -1;
   $scevgep4$i$i = (($29) + ($34<<2)|0);
   HEAP32[$11>>2] = $scevgep4$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  $22 = ($$__last|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($22) {
   return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  $$pre$i1 = HEAP32[$11>>2]|0; //@line 1030 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$01$i3 = $16;$24 = $$pre$i1;
  while(1) {
   $23 = HEAP32[$$01$i3>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$24>>2] = $23; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $25 = HEAP32[$11>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $26 = ((($25)) + 4|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$11>>2] = $26; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $27 = ((($$01$i3)) + 4|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $28 = ($27|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($28) {
    break;
   } else {
    $$01$i3 = $27;$24 = $26;
   }
  }
  return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $35 = ($6|0)==(0); //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $36 = $6; //@line 948 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($35) {
  $65 = $5;
 } else {
  $37 = ((($this)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $38 = HEAP32[$37>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $39 = ($38|0)==($10|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($39)) {
   $scevgep$i$i$i$i = ((($38)) + -4|0);
   $40 = $scevgep$i$i$i$i;
   $41 = (($40) - ($6))|0;
   $42 = $41 >>> 2;
   $43 = $42 ^ -1;
   $scevgep4$i$i$i$i = (($38) + ($43<<2)|0);
   HEAP32[$37>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($36); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  HEAP32[$4>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$37>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$this>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $65 = 0;
 }
 $44 = ($3>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($44) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$4>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre8 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $46 = $$pre;$47 = $$pre8;
 } else {
  $46 = $65;$47 = 0;
 }
 $45 = (($46) - ($47))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $48 = $45 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $49 = ($48>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($49) {
  $50 = $45 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $51 = ($50>>>0)<($3>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $52 = $51 ? $3 : $50; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $53 = ($52>>>0)>(1073741823); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($53) {
   __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$0$i5 = $52;
  } else {
   $$0$i5 = $52;
  }
 } else {
  $$0$i5 = 1073741823;
 }
 $54 = $$0$i5 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $55 = (__Znwj($54)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $56 = ((($this)) + 4|0); //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$56>>2] = $55; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$this>>2] = $55; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $57 = (($55) + ($$0$i5<<2)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$4>>2] = $57; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $58 = ($__first|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($58) {
  return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $$01$i = $__first;$60 = $55;
 while(1) {
  $59 = HEAP32[$$01$i>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$60>>2] = $59; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $61 = HEAP32[$56>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $62 = ((($61)) + 4|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$56>>2] = $62; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $63 = ((($$01$i)) + 4|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $64 = ($63|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($64) {
   break;
  } else {
   $$01$i = $63;$60 = $62;
  }
 }
 return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl6Canvas14getSizedFormatENS2_6FormatE($format) {
 $format = $format|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 switch ($format|0) {
 case 0:  {
  $0 = (__ZN4love8graphics14isGammaCorrectEv()|0); //@line 669 "modules/graphics/opengl/Canvas.cpp"
  if ($0) {
   $$0 = 16;
  } else {
   $1 = HEAP8[1252012]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $2 = ($1<<24>>24)==(0); //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $3 = HEAP8[1252013]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $4 = HEAP8[1252284]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $5 = $4 | $3; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $6 = HEAP8[1252377]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $7 = $5 | $6; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $8 = ($7<<24>>24)!=(0); //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $9 = $2 | $8; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $$ = $9 ? 7 : 2; //@line 675 "modules/graphics/opengl/Canvas.cpp"
   $$0 = $$;
  }
  break;
 }
 case 1:  {
  $$0 = 12;
  break;
 }
 default: {
  $$0 = $format;
 }
 }
 return ($$0|0); //@line 681 "modules/graphics/opengl/Canvas.cpp"
}
function __ZN4love8graphics6opengl6Canvas33isMultiFormatMultiCanvasSupportedEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNK4love8graphics6opengl6OpenGL19getMaxRenderTargetsEv(634044)|0); //@line 814 "modules/graphics/opengl/Canvas.cpp"
 $1 = ($0|0)>(1); //@line 814 "modules/graphics/opengl/Canvas.cpp"
 if (!($1)) {
  $8 = 0;
  return ($8|0); //@line 814 "modules/graphics/opengl/Canvas.cpp"
 }
 $2 = HEAP8[1252013]|0; //@line 814 "modules/graphics/opengl/Canvas.cpp"
 $3 = HEAP8[1252002]|0; //@line 814 "modules/graphics/opengl/Canvas.cpp"
 $4 = $3 | $2; //@line 814 "modules/graphics/opengl/Canvas.cpp"
 $5 = ($4<<24>>24)==(0); //@line 814 "modules/graphics/opengl/Canvas.cpp"
 if (!($5)) {
  $8 = 1;
  return ($8|0); //@line 814 "modules/graphics/opengl/Canvas.cpp"
 }
 $6 = HEAP8[1252065]|0; //@line 814 "modules/graphics/opengl/Canvas.cpp"
 $7 = ($6<<24>>24)!=(0); //@line 814 "modules/graphics/opengl/Canvas.cpp"
 $8 = $7;
 return ($8|0); //@line 814 "modules/graphics/opengl/Canvas.cpp"
}
function __ZN4love8graphics6opengl6Canvas9startGrabEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $__new_last2$i$i$i = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl6Canvas9setupGrabEv($this); //@line 492 "modules/graphics/opengl/Canvas.cpp"
 $0 = HEAP8[1251994]|0; //@line 495 "modules/graphics/opengl/Canvas.cpp"
 $1 = HEAP8[1252225]|0; //@line 495 "modules/graphics/opengl/Canvas.cpp"
 $2 = $1 | $0; //@line 495 "modules/graphics/opengl/Canvas.cpp"
 $3 = ($2<<24>>24)==(0); //@line 495 "modules/graphics/opengl/Canvas.cpp"
 do {
  if (!($3)) {
   $4 = ((($this)) + 144|0); //@line 497 "modules/graphics/opengl/Canvas.cpp"
   $5 = HEAP32[$4>>2]|0; //@line 497 "modules/graphics/opengl/Canvas.cpp"
   switch ($5|0) {
   case 0:  {
    $6 = (__ZN4love8graphics14isGammaCorrectEv()|0); //@line 669 "modules/graphics/opengl/Canvas.cpp"
    if ($6) {
     label = 4;
    } else {
     label = 6;
    }
    break;
   }
   case 16:  {
    label = 4;
    break;
   }
   default: {
    label = 6;
   }
   }
   if ((label|0) == 4) {
    $7 = (__ZNK4love8graphics6opengl6OpenGL18hasFramebufferSRGBEv(634044)|0); //@line 498 "modules/graphics/opengl/Canvas.cpp"
    if ($7) {
     break;
    }
    __ZN4love8graphics6opengl6OpenGL18setFramebufferSRGBEb(634044,1); //@line 499 "modules/graphics/opengl/Canvas.cpp"
    break;
   }
   else if ((label|0) == 6) {
    $8 = (__ZNK4love8graphics6opengl6OpenGL18hasFramebufferSRGBEv(634044)|0); //@line 500 "modules/graphics/opengl/Canvas.cpp"
    if (!($8)) {
     break;
    }
    __ZN4love8graphics6opengl6OpenGL18setFramebufferSRGBEb(634044,0); //@line 501 "modules/graphics/opengl/Canvas.cpp"
    break;
   }
  }
 } while(0);
 $9 = ((($this)) + 152|0); //@line 504 "modules/graphics/opengl/Canvas.cpp"
 $10 = ((($this)) + 156|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = HEAP32[$10>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = HEAP32[$9>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ($11|0)==($12|0); //@line 504 "modules/graphics/opengl/Canvas.cpp"
 if ($13) {
  return; //@line 511 "modules/graphics/opengl/Canvas.cpp"
 }
 $14 = HEAP32[164868]|0; //@line 48 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$14 & 8191](36064); //@line 48 "./libraries/glad/gladfuncs.hpp"
 $15 = HEAP32[$9>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ($16|0)==($15|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($17) {
  return; //@line 511 "modules/graphics/opengl/Canvas.cpp"
 }
 $__new_last2$i$i$i = $15; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $scevgep$i$i$i = ((($16)) + -4|0);
 $18 = $scevgep$i$i$i;
 $19 = (($18) - ($__new_last2$i$i$i))|0;
 $20 = $19 >>> 2;
 $21 = $20 ^ -1;
 $scevgep4$i$i$i = (($16) + ($21<<2)|0);
 HEAP32[$10>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 return; //@line 511 "modules/graphics/opengl/Canvas.cpp"
}
function __ZN4love8graphics6opengl6Canvas18checkCreateStencilEv($this) {
 $this = $this|0;
 var $$ = 0, $$0 = 0, $$10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 140|0); //@line 546 "modules/graphics/opengl/Canvas.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 546 "modules/graphics/opengl/Canvas.cpp"
 $2 = ($1|0)==(0); //@line 546 "modules/graphics/opengl/Canvas.cpp"
 if (!($2)) {
  $$0 = 1;
  return ($$0|0); //@line 595 "modules/graphics/opengl/Canvas.cpp"
 }
 $3 = HEAP32[158272]|0; //@line 551 "modules/graphics/opengl/Canvas.cpp"
 $4 = ($3|0)==($this|0); //@line 551 "modules/graphics/opengl/Canvas.cpp"
 if (!($4)) {
  $5 = ((($this)) + 124|0); //@line 552 "modules/graphics/opengl/Canvas.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 552 "modules/graphics/opengl/Canvas.cpp"
  __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$6); //@line 552 "modules/graphics/opengl/Canvas.cpp"
 }
 $7 = HEAP8[1252013]|0; //@line 558 "modules/graphics/opengl/Canvas.cpp"
 $8 = HEAP8[1252002]|0; //@line 558 "modules/graphics/opengl/Canvas.cpp"
 $9 = $8 | $7; //@line 558 "modules/graphics/opengl/Canvas.cpp"
 $10 = HEAP8[1252065]|0; //@line 558 "modules/graphics/opengl/Canvas.cpp"
 $11 = $9 | $10; //@line 558 "modules/graphics/opengl/Canvas.cpp"
 $12 = HEAP8[1252200]|0; //@line 559 "modules/graphics/opengl/Canvas.cpp"
 $13 = $11 | $12; //@line 558 "modules/graphics/opengl/Canvas.cpp"
 $14 = HEAP8[1252281]|0; //@line 559 "modules/graphics/opengl/Canvas.cpp"
 $15 = $13 | $14; //@line 558 "modules/graphics/opengl/Canvas.cpp"
 $16 = ($15<<24>>24)==(0); //@line 558 "modules/graphics/opengl/Canvas.cpp"
 $$ = $16 ? 36168 : 35056; //@line 558 "modules/graphics/opengl/Canvas.cpp"
 $$10 = $16 ? 36128 : 33306; //@line 558 "modules/graphics/opengl/Canvas.cpp"
 $17 = HEAP32[164676]|0; //@line 663 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$17 & 8191](1,$0); //@line 663 "./libraries/glad/gladfuncs.hpp"
 $18 = HEAP32[$0>>2]|0; //@line 566 "modules/graphics/opengl/Canvas.cpp"
 $19 = HEAP32[165274]|0; //@line 661 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$19 & 8191](36161,$18); //@line 661 "./libraries/glad/gladfuncs.hpp"
 $20 = ((($this)) + 164|0); //@line 568 "modules/graphics/opengl/Canvas.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 568 "modules/graphics/opengl/Canvas.cpp"
 $22 = ($21|0)>(1); //@line 568 "modules/graphics/opengl/Canvas.cpp"
 $23 = ((($this)) + 8|0); //@line 569 "modules/graphics/opengl/Canvas.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 569 "modules/graphics/opengl/Canvas.cpp"
 $25 = ((($this)) + 12|0); //@line 569 "modules/graphics/opengl/Canvas.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 569 "modules/graphics/opengl/Canvas.cpp"
 if ($22) {
  $27 = HEAP32[165298]|0; //@line 678 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiiii[$27 & 8191](36161,$21,$$,$24,$26); //@line 678 "./libraries/glad/gladfuncs.hpp"
 } else {
  $28 = HEAP32[164704]|0; //@line 664 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiii[$28 & 8191](36161,$$,$24,$26); //@line 664 "./libraries/glad/gladfuncs.hpp"
 }
 $29 = HEAP32[$0>>2]|0; //@line 574 "modules/graphics/opengl/Canvas.cpp"
 $30 = HEAP32[164630]|0; //@line 674 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiii[$30 & 8191](36160,$$10,36161,$29); //@line 674 "./libraries/glad/gladfuncs.hpp"
 $31 = HEAP32[165274]|0; //@line 661 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$31 & 8191](36161,0); //@line 661 "./libraries/glad/gladfuncs.hpp"
 $32 = HEAP32[165498]|0; //@line 670 "./libraries/glad/gladfuncs.hpp"
 $33 = (FUNCTION_TABLE_ii[$32 & 8191](36160)|0); //@line 670 "./libraries/glad/gladfuncs.hpp"
 $34 = ($33|0)==(36053); //@line 578 "modules/graphics/opengl/Canvas.cpp"
 if ($34) {
  $35 = HEAP32[164768]|0; //@line 49 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$35 & 8191](1280); //@line 49 "./libraries/glad/gladfuncs.hpp"
 } else {
  $36 = HEAP32[164936]|0; //@line 662 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$36 & 8191](1,$0); //@line 662 "./libraries/glad/gladfuncs.hpp"
  HEAP32[$0>>2] = 0; //@line 586 "modules/graphics/opengl/Canvas.cpp"
 }
 $37 = HEAP32[158272]|0; //@line 589 "modules/graphics/opengl/Canvas.cpp"
 $38 = ($37|0)!=(0|0); //@line 589 "modules/graphics/opengl/Canvas.cpp"
 $39 = ($37|0)!=($this|0); //@line 589 "modules/graphics/opengl/Canvas.cpp"
 $or$cond9 = $38 & $39; //@line 589 "modules/graphics/opengl/Canvas.cpp"
 if ($or$cond9) {
  $40 = ((($37)) + 124|0); //@line 590 "modules/graphics/opengl/Canvas.cpp"
  $41 = HEAP32[$40>>2]|0; //@line 590 "modules/graphics/opengl/Canvas.cpp"
  __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$41); //@line 590 "modules/graphics/opengl/Canvas.cpp"
  $$0 = $34;
  return ($$0|0); //@line 595 "modules/graphics/opengl/Canvas.cpp"
 }
 if ($38) {
  $$0 = $34;
  return ($$0|0); //@line 595 "modules/graphics/opengl/Canvas.cpp"
 }
 $42 = (__ZNK4love8graphics6opengl6OpenGL13getDefaultFBOEv(634044)|0); //@line 592 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$42); //@line 592 "modules/graphics/opengl/Canvas.cpp"
 $$0 = $34;
 return ($$0|0); //@line 595 "modules/graphics/opengl/Canvas.cpp"
}
function __ZN4love8graphics6opengl6Canvas12newImageDataEPNS_5image5ImageEiiii($this,$image,$x,$y,$w,$h) {
 $this = $this|0;
 $image = $image|0;
 $x = $x|0;
 $y = $y|0;
 $w = $w|0;
 $h = $h|0;
 var $$0 = 0, $$06 = 0, $$1 = 0, $$17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond3 = 0, $or$cond5 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = $y | $x; //@line 599 "modules/graphics/opengl/Canvas.cpp"
 $1 = ($0|0)<(0); //@line 599 "modules/graphics/opengl/Canvas.cpp"
 $2 = ($w|0)<(1); //@line 599 "modules/graphics/opengl/Canvas.cpp"
 $or$cond3 = $1 | $2; //@line 599 "modules/graphics/opengl/Canvas.cpp"
 $3 = ($h|0)<(1); //@line 599 "modules/graphics/opengl/Canvas.cpp"
 $or$cond5 = $or$cond3 | $3; //@line 599 "modules/graphics/opengl/Canvas.cpp"
 if (!($or$cond5)) {
  $4 = (($w) + ($x))|0; //@line 599 "modules/graphics/opengl/Canvas.cpp"
  $5 = ((($this)) + 8|0); //@line 599 "modules/graphics/opengl/Canvas.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 640 "modules/graphics/opengl/Canvas.cpp"
  $7 = ($4|0)>($6|0); //@line 599 "modules/graphics/opengl/Canvas.cpp"
  if (!($7)) {
   $8 = (($h) + ($y))|0; //@line 599 "modules/graphics/opengl/Canvas.cpp"
   $9 = ((($this)) + 12|0); //@line 599 "modules/graphics/opengl/Canvas.cpp"
   $10 = HEAP32[$9>>2]|0; //@line 641 "modules/graphics/opengl/Canvas.cpp"
   $11 = ($8|0)>($10|0); //@line 599 "modules/graphics/opengl/Canvas.cpp"
   if (!($11)) {
    $17 = $w << 2; //@line 602 "modules/graphics/opengl/Canvas.cpp"
    $18 = Math_imul($17, $h)|0; //@line 603 "modules/graphics/opengl/Canvas.cpp"
    $19 = ($18|0)>(-1); //@line 607 "modules/graphics/opengl/Canvas.cpp"
    $20 = $19 ? $18 : -1; //@line 607 "modules/graphics/opengl/Canvas.cpp"
    __THREW__ = 0;
    $21 = (invoke_ii(3492,($20|0))|0); //@line 607 "modules/graphics/opengl/Canvas.cpp"
    $22 = __THREW__; __THREW__ = 0;
    $23 = $22&1;
    if ($23) {
     $27 = ___cxa_find_matching_catch_3(9056|0)|0;
     $28 = tempRet0;
     $29 = (_llvm_eh_typeid_for((9056|0))|0); //@line 608 "modules/graphics/opengl/Canvas.cpp"
     $30 = ($28|0)==($29|0); //@line 608 "modules/graphics/opengl/Canvas.cpp"
     if (!($30)) {
      $$1 = $27;$$17 = $28;
      ___resumeException($$1|0);
      // unreachable;
     }
     (___cxa_begin_catch(($27|0))|0); //@line 608 "modules/graphics/opengl/Canvas.cpp"
     $31 = (___cxa_allocate_exception(16)|0); //@line 611 "modules/graphics/opengl/Canvas.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($31|0),(507610|0),($vararg_buffer1|0)); //@line 611 "modules/graphics/opengl/Canvas.cpp"
     $32 = __THREW__; __THREW__ = 0;
     $33 = $32&1;
     if ($33) {
      $37 = ___cxa_find_matching_catch_2()|0;
      $38 = tempRet0;
      ___cxa_free_exception(($31|0)); //@line 611 "modules/graphics/opengl/Canvas.cpp"
      $$0 = $37;$$06 = $38;
     } else {
      __THREW__ = 0;
      invoke_viii(3350,($31|0),(96|0),(344|0)); //@line 611 "modules/graphics/opengl/Canvas.cpp"
      $34 = __THREW__; __THREW__ = 0;
      $35 = ___cxa_find_matching_catch_2()|0;
      $36 = tempRet0;
      $$0 = $35;$$06 = $36;
     }
     __THREW__ = 0;
     invoke_v(3378); //@line 612 "modules/graphics/opengl/Canvas.cpp"
     $39 = __THREW__; __THREW__ = 0;
     $40 = $39&1;
     if ($40) {
      $68 = ___cxa_find_matching_catch_3(0|0)|0;
      $69 = tempRet0;
      ___clang_call_terminate($68); //@line 612 "modules/graphics/opengl/Canvas.cpp"
      // unreachable; //@line 612 "modules/graphics/opengl/Canvas.cpp"
     } else {
      $$1 = $$0;$$17 = $$06;
      ___resumeException($$1|0);
      // unreachable;
     }
    }
    $24 = HEAP32[158272]|0; //@line 615 "modules/graphics/opengl/Canvas.cpp"
    $25 = ($24|0)==($this|0); //@line 615 "modules/graphics/opengl/Canvas.cpp"
    $26 = ((($this)) + 128|0); //@line 635 "modules/graphics/opengl/Canvas.cpp"
    do {
     if ($25) {
      $41 = HEAP32[$26>>2]|0; //@line 635 "modules/graphics/opengl/Canvas.cpp"
      $42 = ($41|0)==(0); //@line 635 "modules/graphics/opengl/Canvas.cpp"
      if (!($42)) {
       $43 = ((($this)) + 136|0); //@line 635 "modules/graphics/opengl/Canvas.cpp"
       $44 = HEAP32[$43>>2]|0; //@line 635 "modules/graphics/opengl/Canvas.cpp"
       $45 = ($44|0)==(0); //@line 635 "modules/graphics/opengl/Canvas.cpp"
       if (!($45)) {
        $46 = ((($this)) + 124|0); //@line 647 "modules/graphics/opengl/Canvas.cpp"
        $47 = HEAP32[$46>>2]|0; //@line 647 "modules/graphics/opengl/Canvas.cpp"
        __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36008,$47); //@line 647 "modules/graphics/opengl/Canvas.cpp"
        $48 = HEAP32[$26>>2]|0; //@line 648 "modules/graphics/opengl/Canvas.cpp"
        __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36009,$48); //@line 648 "modules/graphics/opengl/Canvas.cpp"
        $49 = HEAP8[1252362]|0; //@line 650 "modules/graphics/opengl/Canvas.cpp"
        $50 = ($49<<24>>24)==(0); //@line 650 "modules/graphics/opengl/Canvas.cpp"
        if ($50) {
         $52 = HEAP32[164725]|0; //@line 677 "./libraries/glad/gladfuncs.hpp"
         FUNCTION_TABLE_viiiiiiiiii[$52 & 8191](0,0,$6,$10,0,0,$6,$10,16384,9728); //@line 677 "./libraries/glad/gladfuncs.hpp"
         break;
        } else {
         $51 = HEAP32[166329]|0; //@line 1996 "./libraries/glad/gladfuncs.hpp"
         FUNCTION_TABLE_v[$51 & 8191](); //@line 1996 "./libraries/glad/gladfuncs.hpp"
         break;
        }
       }
      }
     }
    } while(0);
    $53 = HEAP32[$26>>2]|0; //@line 619 "modules/graphics/opengl/Canvas.cpp"
    $54 = ($53|0)==(0); //@line 619 "modules/graphics/opengl/Canvas.cpp"
    if ($54) {
     $55 = ((($this)) + 124|0); //@line 622 "modules/graphics/opengl/Canvas.cpp"
     $56 = HEAP32[$55>>2]|0; //@line 622 "modules/graphics/opengl/Canvas.cpp"
     __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$56); //@line 622 "modules/graphics/opengl/Canvas.cpp"
    } else {
     __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36008,$53); //@line 620 "modules/graphics/opengl/Canvas.cpp"
    }
    $57 = HEAP32[165294]|0; //@line 68 "./libraries/glad/gladfuncs.hpp"
    FUNCTION_TABLE_viiiiiii[$57 & 8191]($x,$y,$w,$h,6408,5121,$21); //@line 68 "./libraries/glad/gladfuncs.hpp"
    $58 = HEAP32[158272]|0; //@line 626 "modules/graphics/opengl/Canvas.cpp"
    $59 = ($58|0)==(0|0); //@line 626 "modules/graphics/opengl/Canvas.cpp"
    if ($59) {
     $62 = (__ZNK4love8graphics6opengl6OpenGL13getDefaultFBOEv(634044)|0); //@line 626 "modules/graphics/opengl/Canvas.cpp"
     $63 = $62;
     __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$63); //@line 627 "modules/graphics/opengl/Canvas.cpp"
     $64 = HEAP32[$image>>2]|0; //@line 630 "modules/graphics/opengl/Canvas.cpp"
     $65 = ((($64)) + 32|0); //@line 630 "modules/graphics/opengl/Canvas.cpp"
     $66 = HEAP32[$65>>2]|0; //@line 630 "modules/graphics/opengl/Canvas.cpp"
     $67 = (FUNCTION_TABLE_iiiiii[$66 & 8191]($image,$w,$h,$21,1)|0); //@line 630 "modules/graphics/opengl/Canvas.cpp"
     STACKTOP = sp;return ($67|0); //@line 630 "modules/graphics/opengl/Canvas.cpp"
    } else {
     $60 = ((($58)) + 124|0); //@line 626 "modules/graphics/opengl/Canvas.cpp"
     $61 = HEAP32[$60>>2]|0; //@line 626 "modules/graphics/opengl/Canvas.cpp"
     $63 = $61;
     __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$63); //@line 627 "modules/graphics/opengl/Canvas.cpp"
     $64 = HEAP32[$image>>2]|0; //@line 630 "modules/graphics/opengl/Canvas.cpp"
     $65 = ((($64)) + 32|0); //@line 630 "modules/graphics/opengl/Canvas.cpp"
     $66 = HEAP32[$65>>2]|0; //@line 630 "modules/graphics/opengl/Canvas.cpp"
     $67 = (FUNCTION_TABLE_iiiiii[$66 & 8191]($image,$w,$h,$21,1)|0); //@line 630 "modules/graphics/opengl/Canvas.cpp"
     STACKTOP = sp;return ($67|0); //@line 630 "modules/graphics/opengl/Canvas.cpp"
    }
   }
  }
 }
 $12 = (___cxa_allocate_exception(16)|0); //@line 600 "modules/graphics/opengl/Canvas.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($12|0),(434402|0),($vararg_buffer|0)); //@line 600 "modules/graphics/opengl/Canvas.cpp"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if (!($14)) {
  ___cxa_throw(($12|0),(96|0),(344|0)); //@line 600 "modules/graphics/opengl/Canvas.cpp"
  // unreachable; //@line 600 "modules/graphics/opengl/Canvas.cpp"
 }
 $15 = ___cxa_find_matching_catch_2()|0;
 $16 = tempRet0;
 ___cxa_free_exception(($12|0)); //@line 600 "modules/graphics/opengl/Canvas.cpp"
 $$1 = $15;$$17 = $16;
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8graphics6opengl6Canvas11isSupportedEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[1252012]|0; //@line 809 "modules/graphics/opengl/Canvas.cpp"
 $1 = HEAP8[1252002]|0; //@line 809 "modules/graphics/opengl/Canvas.cpp"
 $2 = $1 | $0; //@line 809 "modules/graphics/opengl/Canvas.cpp"
 $3 = HEAP8[1252065]|0; //@line 809 "modules/graphics/opengl/Canvas.cpp"
 $4 = $2 | $3; //@line 809 "modules/graphics/opengl/Canvas.cpp"
 $5 = HEAP8[1252185]|0; //@line 809 "modules/graphics/opengl/Canvas.cpp"
 $6 = $4 | $5; //@line 809 "modules/graphics/opengl/Canvas.cpp"
 $7 = ($6<<24>>24)!=(0); //@line 809 "modules/graphics/opengl/Canvas.cpp"
 return ($7|0); //@line 809 "modules/graphics/opengl/Canvas.cpp"
}
function __ZN4love8graphics6opengl6Canvas17isFormatSupportedENS2_6FormatE($format) {
 $format = $format|0;
 var $$ = 0, $$$i = 0, $$$i$i = 0, $$0$i = 0, $$0$i$i = 0, $$0$i25 = 0, $$1 = 0, $$28 = 0, $$29 = 0, $$30 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $current_fbo$i = 0, $externalformat$2 = 0, $f = 0, $fbo = 0, $internalformat$0 = 0, $internalformat$0$externalformat$2 = 0, $or$cond$i = 0, $or$cond11 = 0, $texture = 0, $textype$0 = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $current_fbo$i = sp + 36|0;
 $texture = sp + 32|0;
 $f = sp + 16|0;
 $w = sp + 8|0;
 $fbo = sp;
 $0 = HEAP8[1252012]|0; //@line 809 "modules/graphics/opengl/Canvas.cpp"
 $1 = HEAP8[1252002]|0; //@line 809 "modules/graphics/opengl/Canvas.cpp"
 $2 = $1 | $0; //@line 809 "modules/graphics/opengl/Canvas.cpp"
 $3 = HEAP8[1252065]|0; //@line 809 "modules/graphics/opengl/Canvas.cpp"
 $4 = $2 | $3; //@line 809 "modules/graphics/opengl/Canvas.cpp"
 $5 = HEAP8[1252185]|0; //@line 809 "modules/graphics/opengl/Canvas.cpp"
 $6 = $4 | $5; //@line 809 "modules/graphics/opengl/Canvas.cpp"
 $7 = ($6<<24>>24)==(0); //@line 809 "modules/graphics/opengl/Canvas.cpp"
 if ($7) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
 }
 switch ($format|0) {
 case 0:  {
  $8 = (__ZN4love8graphics14isGammaCorrectEv()|0); //@line 669 "modules/graphics/opengl/Canvas.cpp"
  if ($8) {
   label = 28;
  } else {
   $9 = HEAP8[1252012]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $10 = ($9<<24>>24)==(0); //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $11 = HEAP8[1252013]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $12 = HEAP8[1252284]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $13 = $12 | $11; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $14 = HEAP8[1252377]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $15 = $13 | $14; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $16 = ($15<<24>>24)!=(0); //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $17 = $10 | $16; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $$$i = $17 ? 7 : 2; //@line 675 "modules/graphics/opengl/Canvas.cpp"
   $$0$i = $$$i;$20 = $9;
   label = 5;
  }
  break;
 }
 case 1:  {
  $78 = $0;
  label = 20;
  break;
 }
 default: {
  $$0$i = $format;$20 = $0;
  label = 5;
 }
 }
 L7: do {
  if ((label|0) == 5) {
   switch ($$0$i|0) {
   case 3: case 2:  {
    $$0$i25 = $$0$i;
    break L7;
    break;
   }
   case 12:  {
    $78 = $20;
    label = 20;
    break L7;
    break;
   }
   case 16:  {
    label = 28;
    break L7;
    break;
   }
   case 4:  {
    $18 = HEAP8[1252008]|0; //@line 835 "modules/graphics/opengl/Canvas.cpp"
    $19 = $18 | $20; //@line 835 "modules/graphics/opengl/Canvas.cpp"
    $21 = HEAP8[1252055]|0; //@line 835 "modules/graphics/opengl/Canvas.cpp"
    $22 = $19 | $21; //@line 835 "modules/graphics/opengl/Canvas.cpp"
    $23 = ($22<<24>>24)==(0); //@line 835 "modules/graphics/opengl/Canvas.cpp"
    if ($23) {
     $$1 = 0;
    } else {
     $$0$i25 = 4;
     break L7;
    }
    STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
    break;
   }
   case 6: case 5:  {
    $24 = HEAP8[1251994]|0; //@line 839 "modules/graphics/opengl/Canvas.cpp"
    $25 = ($24<<24>>24)==(0); //@line 839 "modules/graphics/opengl/Canvas.cpp"
    if (!($25)) {
     $26 = HEAP8[1252002]|0; //@line 840 "modules/graphics/opengl/Canvas.cpp"
     $27 = HEAP8[1252142]|0; //@line 840 "modules/graphics/opengl/Canvas.cpp"
     $28 = $27 | $26; //@line 840 "modules/graphics/opengl/Canvas.cpp"
     $29 = ($28<<24>>24)==(0); //@line 840 "modules/graphics/opengl/Canvas.cpp"
     if ($29) {
      $$1 = 0;
     } else {
      $$0$i25 = $$0$i;
      break L7;
     }
     STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
    }
    $30 = ($20<<24>>24)==(0); //@line 841 "modules/graphics/opengl/Canvas.cpp"
    if ($30) {
     $$0$i25 = $$0$i;
     break L7;
    }
    $31 = HEAP8[1252013]|0; //@line 842 "modules/graphics/opengl/Canvas.cpp"
    $32 = HEAP8[1252244]|0; //@line 842 "modules/graphics/opengl/Canvas.cpp"
    $33 = $32 | $31; //@line 842 "modules/graphics/opengl/Canvas.cpp"
    $34 = ($33<<24>>24)==(0); //@line 842 "modules/graphics/opengl/Canvas.cpp"
    if ($34) {
     $$1 = 0;
    } else {
     $$0$i25 = $$0$i;
     break L7;
    }
    STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
    break;
   }
   case 7:  {
    $35 = HEAP8[1251994]|0; //@line 845 "modules/graphics/opengl/Canvas.cpp"
    $36 = HEAP8[1252013]|0; //@line 845 "modules/graphics/opengl/Canvas.cpp"
    $37 = $36 | $35; //@line 845 "modules/graphics/opengl/Canvas.cpp"
    $38 = HEAP8[1252284]|0; //@line 845 "modules/graphics/opengl/Canvas.cpp"
    $39 = $37 | $38; //@line 845 "modules/graphics/opengl/Canvas.cpp"
    $40 = HEAP8[1252377]|0; //@line 845 "modules/graphics/opengl/Canvas.cpp"
    $41 = $39 | $40; //@line 845 "modules/graphics/opengl/Canvas.cpp"
    $42 = ($41<<24>>24)==(0); //@line 845 "modules/graphics/opengl/Canvas.cpp"
    if ($42) {
     $$1 = 0;
    } else {
     $$0$i25 = 7;
     break L7;
    }
    STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
    break;
   }
   case 8:  {
    $43 = HEAP8[1252013]|0; //@line 848 "modules/graphics/opengl/Canvas.cpp"
    $44 = HEAP8[1251994]|0; //@line 848 "modules/graphics/opengl/Canvas.cpp"
    $45 = $44 | $43; //@line 848 "modules/graphics/opengl/Canvas.cpp"
    $46 = ($45<<24>>24)==(0); //@line 848 "modules/graphics/opengl/Canvas.cpp"
    if ($46) {
     $$1 = 0;
    } else {
     $$0$i25 = 8;
     break L7;
    }
    STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
    break;
   }
   case 9:  {
    $47 = HEAP8[1252002]|0; //@line 851 "modules/graphics/opengl/Canvas.cpp"
    $48 = HEAP8[1252201]|0; //@line 851 "modules/graphics/opengl/Canvas.cpp"
    $49 = $48 | $47; //@line 851 "modules/graphics/opengl/Canvas.cpp"
    $50 = HEAP8[1252356]|0; //@line 851 "modules/graphics/opengl/Canvas.cpp"
    $51 = $49 | $50; //@line 851 "modules/graphics/opengl/Canvas.cpp"
    $52 = ($51<<24>>24)==(0); //@line 851 "modules/graphics/opengl/Canvas.cpp"
    if ($52) {
     $$1 = 0;
    } else {
     $$0$i25 = 9;
     break L7;
    }
    STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
    break;
   }
   case 11: case 10:  {
    $53 = HEAP8[1251994]|0; //@line 855 "modules/graphics/opengl/Canvas.cpp"
    $54 = ($53<<24>>24)==(0); //@line 855 "modules/graphics/opengl/Canvas.cpp"
    if (!($54)) {
     $55 = HEAP8[1252002]|0; //@line 856 "modules/graphics/opengl/Canvas.cpp"
     $56 = ($55<<24>>24)==(0); //@line 856 "modules/graphics/opengl/Canvas.cpp"
     if (!($56)) {
      $$0$i25 = $$0$i;
      break L7;
     }
     $57 = HEAP8[1252135]|0; //@line 856 "modules/graphics/opengl/Canvas.cpp"
     $58 = ($57<<24>>24)!=(0); //@line 856 "modules/graphics/opengl/Canvas.cpp"
     $59 = HEAP8[1252142]|0; //@line 856 "modules/graphics/opengl/Canvas.cpp"
     $60 = ($59<<24>>24)!=(0); //@line 856 "modules/graphics/opengl/Canvas.cpp"
     $61 = $58 & $60; //@line 856 "modules/graphics/opengl/Canvas.cpp"
     if ($61) {
      $$0$i25 = $$0$i;
      break L7;
     } else {
      $$1 = 0;
     }
     STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
    }
    $62 = HEAP8[1252169]|0; //@line 858 "modules/graphics/opengl/Canvas.cpp"
    $63 = ($62<<24>>24)==(0); //@line 858 "modules/graphics/opengl/Canvas.cpp"
    if ($63) {
     $$1 = 0;
     STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
    }
    $64 = HEAP8[1252013]|0; //@line 858 "modules/graphics/opengl/Canvas.cpp"
    $65 = ($64<<24>>24)==(0); //@line 858 "modules/graphics/opengl/Canvas.cpp"
    if (!($65)) {
     $$0$i25 = $$0$i;
     break L7;
    }
    $66 = HEAP8[1252303]|0; //@line 858 "modules/graphics/opengl/Canvas.cpp"
    $67 = ($66<<24>>24)!=(0); //@line 858 "modules/graphics/opengl/Canvas.cpp"
    $68 = HEAP8[1252244]|0; //@line 858 "modules/graphics/opengl/Canvas.cpp"
    $69 = ($68<<24>>24)!=(0); //@line 858 "modules/graphics/opengl/Canvas.cpp"
    $70 = $67 & $69; //@line 858 "modules/graphics/opengl/Canvas.cpp"
    if ($70) {
     $$0$i25 = $$0$i;
     break L7;
    } else {
     $$1 = 0;
    }
    STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
    break;
   }
   case 14: case 13:  {
    $85 = HEAP8[1252002]|0; //@line 868 "modules/graphics/opengl/Canvas.cpp"
    $86 = ($85<<24>>24)==(0); //@line 868 "modules/graphics/opengl/Canvas.cpp"
    if (!($86)) {
     $$0$i25 = $$0$i;
     break L7;
    }
    $87 = HEAP8[1252135]|0; //@line 868 "modules/graphics/opengl/Canvas.cpp"
    $88 = ($87<<24>>24)!=(0); //@line 868 "modules/graphics/opengl/Canvas.cpp"
    $89 = HEAP8[1252142]|0; //@line 868 "modules/graphics/opengl/Canvas.cpp"
    $90 = ($89<<24>>24)!=(0); //@line 868 "modules/graphics/opengl/Canvas.cpp"
    $91 = $88 & $90; //@line 868 "modules/graphics/opengl/Canvas.cpp"
    if ($91) {
     $$0$i25 = $$0$i;
     break L7;
    } else {
     $$1 = 0;
    }
    STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
    break;
   }
   case 15:  {
    $92 = HEAP8[1252002]|0; //@line 871 "modules/graphics/opengl/Canvas.cpp"
    $93 = HEAP8[1252135]|0; //@line 871 "modules/graphics/opengl/Canvas.cpp"
    $94 = $93 | $92; //@line 871 "modules/graphics/opengl/Canvas.cpp"
    $95 = ($94<<24>>24)==(0); //@line 871 "modules/graphics/opengl/Canvas.cpp"
    if ($95) {
     $$1 = 0;
    } else {
     $$0$i25 = 15;
     break L7;
    }
    STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
    break;
   }
   default: {
    $$1 = 0;
    STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
   }
   }
  }
 } while(0);
 do {
  if ((label|0) == 20) {
   $71 = HEAP8[1251994]|0; //@line 861 "modules/graphics/opengl/Canvas.cpp"
   $72 = ($71<<24>>24)==(0); //@line 861 "modules/graphics/opengl/Canvas.cpp"
   if (!($72)) {
    $73 = HEAP8[1252002]|0; //@line 862 "modules/graphics/opengl/Canvas.cpp"
    $74 = HEAP8[1252135]|0; //@line 862 "modules/graphics/opengl/Canvas.cpp"
    $75 = $74 | $73; //@line 862 "modules/graphics/opengl/Canvas.cpp"
    $76 = ($75<<24>>24)==(0); //@line 862 "modules/graphics/opengl/Canvas.cpp"
    if ($76) {
     $$1 = 0;
    } else {
     $$0$i25 = 12;
     break;
    }
    STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
   }
   $77 = ($78<<24>>24)==(0); //@line 863 "modules/graphics/opengl/Canvas.cpp"
   if ($77) {
    $$0$i25 = 12;
   } else {
    $79 = HEAP8[1252169]|0; //@line 864 "modules/graphics/opengl/Canvas.cpp"
    $80 = ($79<<24>>24)==(0); //@line 864 "modules/graphics/opengl/Canvas.cpp"
    if ($80) {
     $$1 = 0;
     STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
    }
    $81 = HEAP8[1252013]|0; //@line 864 "modules/graphics/opengl/Canvas.cpp"
    $82 = HEAP8[1252303]|0; //@line 864 "modules/graphics/opengl/Canvas.cpp"
    $83 = $82 | $81; //@line 864 "modules/graphics/opengl/Canvas.cpp"
    $84 = ($83<<24>>24)==(0); //@line 864 "modules/graphics/opengl/Canvas.cpp"
    if ($84) {
     $$1 = 0;
     STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
    } else {
     $$0$i25 = 12;
    }
   }
  }
  else if ((label|0) == 28) {
   $96 = HEAP8[1251994]|0; //@line 874 "modules/graphics/opengl/Canvas.cpp"
   $97 = ($96<<24>>24)==(0); //@line 874 "modules/graphics/opengl/Canvas.cpp"
   if ($97) {
    $104 = HEAP8[1252013]|0; //@line 880 "modules/graphics/opengl/Canvas.cpp"
    $105 = HEAP8[1252224]|0; //@line 880 "modules/graphics/opengl/Canvas.cpp"
    $106 = $105 | $104; //@line 880 "modules/graphics/opengl/Canvas.cpp"
    $107 = ($106<<24>>24)==(0); //@line 880 "modules/graphics/opengl/Canvas.cpp"
    if ($107) {
     $$1 = 0;
    } else {
     $$0$i25 = 16;
     break;
    }
    STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
   }
   $98 = HEAP8[1252002]|0; //@line 876 "modules/graphics/opengl/Canvas.cpp"
   $99 = ($98<<24>>24)==(0); //@line 876 "modules/graphics/opengl/Canvas.cpp"
   if ($99) {
    $100 = HEAP8[1252066]|0; //@line 876 "modules/graphics/opengl/Canvas.cpp"
    $101 = HEAP8[1252186]|0; //@line 876 "modules/graphics/opengl/Canvas.cpp"
    $102 = $101 | $100; //@line 876 "modules/graphics/opengl/Canvas.cpp"
    $103 = ($102<<24>>24)==(0); //@line 876 "modules/graphics/opengl/Canvas.cpp"
    if ($103) {
     $$1 = 0;
     STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
    }
    $108 = HEAP8[1252001]|0; //@line 877 "modules/graphics/opengl/Canvas.cpp"
    $109 = HEAP8[1252247]|0; //@line 877 "modules/graphics/opengl/Canvas.cpp"
    $110 = $109 | $108; //@line 877 "modules/graphics/opengl/Canvas.cpp"
    $111 = ($110<<24>>24)==(0); //@line 877 "modules/graphics/opengl/Canvas.cpp"
    if ($111) {
     $$1 = 0;
     STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
    } else {
     $$0$i25 = 16;
    }
   } else {
    $$0$i25 = 16;
   }
  }
 } while(0);
 $112 = (1251913 + ($$0$i25)|0); //@line 890 "modules/graphics/opengl/Canvas.cpp"
 $113 = HEAP8[$112>>0]|0; //@line 890 "modules/graphics/opengl/Canvas.cpp"
 $114 = ($113<<24>>24)==(0); //@line 890 "modules/graphics/opengl/Canvas.cpp"
 if (!($114)) {
  $115 = (1251896 + ($$0$i25)|0); //@line 891 "modules/graphics/opengl/Canvas.cpp"
  $116 = HEAP8[$115>>0]|0; //@line 891 "modules/graphics/opengl/Canvas.cpp"
  $117 = ($116<<24>>24)!=(0); //@line 891 "modules/graphics/opengl/Canvas.cpp"
  $$1 = $117;
  STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
 }
 switch ($$0$i25|0) {
 case 0:  {
  $118 = (__ZN4love8graphics14isGammaCorrectEv()|0); //@line 669 "modules/graphics/opengl/Canvas.cpp"
  if ($118) {
   label = 55;
  } else {
   $119 = HEAP8[1252012]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $120 = ($119<<24>>24)==(0); //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $121 = HEAP8[1252013]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $122 = HEAP8[1252284]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $123 = $122 | $121; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $124 = HEAP8[1252377]|0; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $125 = $123 | $124; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $126 = ($125<<24>>24)!=(0); //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $127 = $120 | $126; //@line 671 "modules/graphics/opengl/Canvas.cpp"
   $$$i$i = $127 ? 7 : 2; //@line 675 "modules/graphics/opengl/Canvas.cpp"
   $$0$i$i = $$$i$i;
   label = 38;
  }
  break;
 }
 case 1:  {
  label = 50;
  break;
 }
 default: {
  $$0$i$i = $$0$i25;
  label = 38;
 }
 }
 L71: do {
  if ((label|0) == 38) {
   do {
    switch ($$0$i$i|0) {
    case 2:  {
     $externalformat$2 = 6408;$internalformat$0 = 32854;$textype$0 = 32819;
     break L71;
     break;
    }
    case 16:  {
     label = 55;
     break L71;
     break;
    }
    case 12:  {
     label = 50;
     break L71;
     break;
    }
    case 3:  {
     $externalformat$2 = 6408;$internalformat$0 = 32855;$textype$0 = 32820;
     break L71;
     break;
    }
    case 4:  {
     $externalformat$2 = 6407;$internalformat$0 = 36194;$textype$0 = 33635;
     break L71;
     break;
    }
    case 5:  {
     $externalformat$2 = 6403;$internalformat$0 = 33321;$textype$0 = 5121;
     break L71;
     break;
    }
    case 6:  {
     $externalformat$2 = 33319;$internalformat$0 = 33323;$textype$0 = 5121;
     break L71;
     break;
    }
    case 8:  {
     $externalformat$2 = 6408;$internalformat$0 = 32857;$textype$0 = 33640;
     break L71;
     break;
    }
    case 9:  {
     $externalformat$2 = 6407;$internalformat$0 = 35898;$textype$0 = 35899;
     break L71;
     break;
    }
    case 10:  {
     $128 = HEAP8[1252303]|0; //@line 730 "modules/graphics/opengl/Canvas.cpp"
     $129 = ($128<<24>>24)==(0); //@line 730 "modules/graphics/opengl/Canvas.cpp"
     if (!($129)) {
      $externalformat$2 = 6403;$internalformat$0 = 33325;$textype$0 = 36193;
      break L71;
     }
     $130 = HEAP8[1251994]|0; //@line 732 "modules/graphics/opengl/Canvas.cpp"
     $131 = ($130<<24>>24)==(0); //@line 732 "modules/graphics/opengl/Canvas.cpp"
     $$ = $131 ? 5131 : 5126; //@line 732 "modules/graphics/opengl/Canvas.cpp"
     $externalformat$2 = 6403;$internalformat$0 = 33325;$textype$0 = $$;
     break L71;
     break;
    }
    case 11:  {
     $132 = HEAP8[1252303]|0; //@line 740 "modules/graphics/opengl/Canvas.cpp"
     $133 = ($132<<24>>24)==(0); //@line 740 "modules/graphics/opengl/Canvas.cpp"
     if (!($133)) {
      $externalformat$2 = 33319;$internalformat$0 = 33327;$textype$0 = 36193;
      break L71;
     }
     $134 = HEAP8[1251994]|0; //@line 742 "modules/graphics/opengl/Canvas.cpp"
     $135 = ($134<<24>>24)==(0); //@line 742 "modules/graphics/opengl/Canvas.cpp"
     $$28 = $135 ? 5131 : 5126; //@line 742 "modules/graphics/opengl/Canvas.cpp"
     $externalformat$2 = 33319;$internalformat$0 = 33327;$textype$0 = $$28;
     break L71;
     break;
    }
    case 13:  {
     $externalformat$2 = 6403;$internalformat$0 = 33326;$textype$0 = 5126;
     break L71;
     break;
    }
    case 14:  {
     $externalformat$2 = 33319;$internalformat$0 = 33328;$textype$0 = 5126;
     break L71;
     break;
    }
    case 15:  {
     $externalformat$2 = 6408;$internalformat$0 = 34836;$textype$0 = 5126;
     break L71;
     break;
    }
    default: {
     $externalformat$2 = 6408;$internalformat$0 = 32856;$textype$0 = 5121;
     break L71;
    }
    }
   } while(0);
  }
 } while(0);
 if ((label|0) == 50) {
  $136 = HEAP8[1252303]|0; //@line 749 "modules/graphics/opengl/Canvas.cpp"
  $137 = ($136<<24>>24)==(0); //@line 749 "modules/graphics/opengl/Canvas.cpp"
  if ($137) {
   $138 = HEAP8[1251994]|0; //@line 751 "modules/graphics/opengl/Canvas.cpp"
   $139 = ($138<<24>>24)==(0); //@line 751 "modules/graphics/opengl/Canvas.cpp"
   $$29 = $139 ? 5131 : 5126; //@line 751 "modules/graphics/opengl/Canvas.cpp"
   $externalformat$2 = 6408;$internalformat$0 = 34842;$textype$0 = $$29;
  } else {
   $externalformat$2 = 6408;$internalformat$0 = 34842;$textype$0 = 36193;
  }
 }
 else if ((label|0) == 55) {
  $140 = HEAP8[1252012]|0; //@line 773 "modules/graphics/opengl/Canvas.cpp"
  $141 = ($140<<24>>24)==(0); //@line 773 "modules/graphics/opengl/Canvas.cpp"
  $142 = HEAP8[1252013]|0; //@line 773 "modules/graphics/opengl/Canvas.cpp"
  $143 = ($142<<24>>24)!=(0); //@line 773 "modules/graphics/opengl/Canvas.cpp"
  $or$cond$i = $141 | $143; //@line 773 "modules/graphics/opengl/Canvas.cpp"
  $$30 = $or$cond$i ? 6408 : 35906; //@line 773 "modules/graphics/opengl/Canvas.cpp"
  $externalformat$2 = $$30;$internalformat$0 = 35907;$textype$0 = 5121;
 }
 $144 = HEAP8[1252012]|0; //@line 904 "modules/graphics/opengl/Canvas.cpp"
 $145 = ($144<<24>>24)==(0); //@line 904 "modules/graphics/opengl/Canvas.cpp"
 $146 = HEAP8[1252013]|0; //@line 904 "modules/graphics/opengl/Canvas.cpp"
 $147 = ($146<<24>>24)!=(0); //@line 904 "modules/graphics/opengl/Canvas.cpp"
 $or$cond11 = $145 | $147; //@line 904 "modules/graphics/opengl/Canvas.cpp"
 $internalformat$0$externalformat$2 = $or$cond11 ? $internalformat$0 : $externalformat$2; //@line 904 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$texture>>2] = 0; //@line 907 "modules/graphics/opengl/Canvas.cpp"
 $148 = HEAP32[164922]|0; //@line 355 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$148 & 8191](1,$texture); //@line 355 "./libraries/glad/gladfuncs.hpp"
 $149 = HEAP32[$texture>>2]|0; //@line 909 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$149); //@line 909 "modules/graphics/opengl/Canvas.cpp"
 $150 = ((($f)) + 4|0); //@line 63 "./modules/graphics/Texture.h"
 $151 = ((($f)) + 8|0); //@line 64 "./modules/graphics/Texture.h"
 HEAP32[$151>>2] = 0; //@line 64 "./modules/graphics/Texture.h"
 $152 = ((($f)) + 12|0); //@line 65 "./modules/graphics/Texture.h"
 HEAPF32[$152>>2] = 1.0; //@line 65 "./modules/graphics/Texture.h"
 HEAP32[$150>>2] = 2; //@line 912 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$f>>2] = 2; //@line 912 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL16setTextureFilterERNS0_7Texture6FilterE(634044,$f); //@line 913 "modules/graphics/opengl/Canvas.cpp"
 HEAP32[$w>>2] = 0; //@line 70 "./modules/graphics/Texture.h"
 $153 = ((($w)) + 4|0); //@line 71 "./modules/graphics/Texture.h"
 HEAP32[$153>>2] = 0; //@line 71 "./modules/graphics/Texture.h"
 __ZN4love8graphics6opengl6OpenGL14setTextureWrapERKNS0_7Texture4WrapE(634044,$w); //@line 916 "modules/graphics/opengl/Canvas.cpp"
 $154 = HEAP32[165602]|0; //@line 47 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiiiiiiii[$154 & 8191](3553,0,$internalformat$0$externalformat$2,2,2,0,$externalformat$2,$textype$0,0); //@line 47 "./libraries/glad/gladfuncs.hpp"
 HEAP32[$fbo>>2] = 0; //@line 920 "modules/graphics/opengl/Canvas.cpp"
 $155 = HEAP32[$texture>>2]|0; //@line 921 "modules/graphics/opengl/Canvas.cpp"
 $156 = HEAP32[165587]|0; //@line 73 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$156 & 8191](36006,$current_fbo$i); //@line 73 "./libraries/glad/gladfuncs.hpp"
 $157 = HEAP32[164797]|0; //@line 669 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$157 & 8191](1,$fbo); //@line 669 "./libraries/glad/gladfuncs.hpp"
 $158 = HEAP32[$fbo>>2]|0; //@line 43 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$158); //@line 43 "modules/graphics/opengl/Canvas.cpp"
 $159 = ($155|0)==(0); //@line 45 "modules/graphics/opengl/Canvas.cpp"
 if (!($159)) {
  $160 = HEAP32[164777]|0; //@line 672 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiiii[$160 & 8191](36160,36064,3553,$155,0); //@line 672 "./libraries/glad/gladfuncs.hpp"
  $161 = HEAP32[165395]|0; //@line 50 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vdddd[$161 & 8191](0.0,0.0,0.0,0.0); //@line 50 "./libraries/glad/gladfuncs.hpp"
  $162 = HEAP32[164768]|0; //@line 49 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$162 & 8191](16384); //@line 49 "./libraries/glad/gladfuncs.hpp"
 }
 $163 = HEAP32[165498]|0; //@line 670 "./libraries/glad/gladfuncs.hpp"
 $164 = (FUNCTION_TABLE_ii[$163 & 8191](36160)|0); //@line 670 "./libraries/glad/gladfuncs.hpp"
 $165 = HEAP32[$current_fbo$i>>2]|0; //@line 57 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj(634044,36160,$165); //@line 57 "modules/graphics/opengl/Canvas.cpp"
 $166 = ($164|0)==(36053); //@line 921 "modules/graphics/opengl/Canvas.cpp"
 $167 = HEAP32[164758]|0; //@line 668 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$167 & 8191](1,$fbo); //@line 668 "./libraries/glad/gladfuncs.hpp"
 $168 = HEAP32[$texture>>2]|0; //@line 924 "modules/graphics/opengl/Canvas.cpp"
 __ZN4love8graphics6opengl6OpenGL13deleteTextureEj(634044,$168); //@line 924 "modules/graphics/opengl/Canvas.cpp"
 HEAP8[$112>>0] = 1; //@line 927 "modules/graphics/opengl/Canvas.cpp"
 $169 = (1251896 + ($$0$i25)|0); //@line 928 "modules/graphics/opengl/Canvas.cpp"
 $170 = $166&1; //@line 928 "modules/graphics/opengl/Canvas.cpp"
 HEAP8[$169>>0] = $170; //@line 928 "modules/graphics/opengl/Canvas.cpp"
 $$1 = $166;
 STACKTOP = sp;return ($$1|0); //@line 931 "modules/graphics/opengl/Canvas.cpp"
}
function __ZN4love8graphics6opengl6Canvas11getConstantEPKcRNS2_6FormatE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 34)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((633112 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (633112 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(34); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((633112 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 935 "modules/graphics/opengl/Canvas.cpp"
  return ($$$2$i|0); //@line 935 "modules/graphics/opengl/Canvas.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 935 "modules/graphics/opengl/Canvas.cpp"
  return ($$$2$i|0); //@line 935 "modules/graphics/opengl/Canvas.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl6Canvas11getConstantENS2_6FormatERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(16); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (633520 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 940 "modules/graphics/opengl/Canvas.cpp"
}
function __ZN4love8graphics6opengl4FontD2Ev($this) {
 $this = $this|0;
 var $$01$i$i$i = 0, $$01$i$i$i1 = 0, $$01$i$i$i4 = 0, $$01$i$i$i8 = 0, $$pre$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i19 = 0, $$pre$i22 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i12 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i13 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (24548); //@line 91 "modules/graphics/opengl/Font.cpp"
 $0 = ((($this)) + 8|0); //@line 91 "modules/graphics/opengl/Font.cpp"
 HEAP32[$0>>2] = (24580); //@line 91 "modules/graphics/opengl/Font.cpp"
 __THREW__ = 0;
 invoke_vi(1732,($this|0)); //@line 92 "modules/graphics/opengl/Font.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $53 = ___cxa_find_matching_catch_3(0|0)|0;
  $54 = tempRet0;
  $55 = ((($this)) + 128|0); //@line 95 "modules/graphics/opengl/Font.cpp"
  __ZN4love8graphics6opengl11QuadIndicesD2Ev($55); //@line 95 "modules/graphics/opengl/Font.cpp"
  $56 = ((($this)) + 80|0); //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $57 = HEAP32[$56>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $58 = ($57|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if (!($58)) {
   $$01$i$i$i4 = $57;
   while(1) {
    $59 = HEAP32[$$01$i$i$i4>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    __ZdlPv($$01$i$i$i4); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $60 = ($59|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($60) {
     break;
    } else {
     $$01$i$i$i4 = $59;
    }
   }
  }
  $61 = ((($this)) + 72|0); //@line 2014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $62 = HEAP32[$61>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$61>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $63 = ($62|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  if (!($63)) {
   __ZdlPv($62); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  $64 = ((($this)) + 60|0); //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $65 = HEAP32[$64>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $66 = ($65|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if (!($66)) {
   $$01$i$i$i8 = $65;
   while(1) {
    $67 = HEAP32[$$01$i$i$i8>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    __ZdlPv($$01$i$i$i8); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $68 = ($67|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($68) {
     break;
    } else {
     $$01$i$i$i8 = $67;
    }
   }
  }
  $69 = ((($this)) + 52|0); //@line 2014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $70 = HEAP32[$69>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$69>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $71 = ($70|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  if (!($71)) {
   __ZdlPv($70); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  $72 = ((($this)) + 40|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $73 = HEAP32[$72>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $74 = ($73|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $75 = $73; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($74)) {
   $76 = ((($this)) + 44|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $77 = HEAP32[$76>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $78 = ($77|0)==($73|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($78)) {
    $scevgep$i$i$i12 = ((($77)) + -4|0);
    $79 = $scevgep$i$i$i12;
    $80 = (($79) - ($75))|0;
    $81 = $80 >>> 2;
    $82 = $81 ^ -1;
    $scevgep4$i$i$i13 = (($77) + ($82<<2)|0);
    HEAP32[$76>>2] = $scevgep4$i$i$i13; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($73); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  $83 = ((($this)) + 12|0); //@line 95 "modules/graphics/opengl/Font.cpp"
  $84 = HEAP32[$83>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $85 = ($84|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($85) {
   __ZN4love8graphics8VolatileD2Ev($0); //@line 95 "modules/graphics/opengl/Font.cpp"
   ___clang_call_terminate($53); //@line 95 "modules/graphics/opengl/Font.cpp"
   // unreachable; //@line 95 "modules/graphics/opengl/Font.cpp"
  }
  $86 = ((($this)) + 16|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $87 = HEAP32[$86>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $88 = ($87|0)==($84|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  do {
   if ($88) {
    $102 = $84;
   } else {
    $90 = $87;
    while(1) {
     $89 = ((($90)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$86>>2] = $89; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $91 = HEAP32[$89>>2]|0; //@line 106 "./common/Object.h"
     $92 = ($91|0)==(0|0); //@line 106 "./common/Object.h"
     if ($92) {
      $94 = $89;
     } else {
      $95 = HEAP32[$91>>2]|0; //@line 106 "./common/Object.h"
      $96 = ((($95)) + 12|0); //@line 106 "./common/Object.h"
      $97 = HEAP32[$96>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($97|0,($91|0)); //@line 106 "./common/Object.h"
      $98 = __THREW__; __THREW__ = 0;
      $99 = $98&1;
      if ($99) {
       label = 42;
       break;
      }
      $$pre$i$i$i19 = HEAP32[$86>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $94 = $$pre$i$i$i19;
     }
     $93 = ($94|0)==($84|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($93) {
      label = 43;
      break;
     } else {
      $90 = $94;
     }
    }
    if ((label|0) == 42) {
     $100 = ___cxa_find_matching_catch_3(0|0)|0;
     $101 = tempRet0;
     ___clang_call_terminate($100); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
    else if ((label|0) == 43) {
     $$pre$i22 = HEAP32[$83>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $102 = $$pre$i22;
     break;
    }
   }
  } while(0);
  __ZdlPv($102); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  __ZN4love8graphics8VolatileD2Ev($0); //@line 95 "modules/graphics/opengl/Font.cpp"
  ___clang_call_terminate($53); //@line 95 "modules/graphics/opengl/Font.cpp"
  // unreachable; //@line 95 "modules/graphics/opengl/Font.cpp"
 } else {
  $3 = HEAP32[158397]|0; //@line 94 "modules/graphics/opengl/Font.cpp"
  $4 = (($3) + -1)|0; //@line 94 "modules/graphics/opengl/Font.cpp"
  HEAP32[158397] = $4; //@line 94 "modules/graphics/opengl/Font.cpp"
  $5 = ((($this)) + 128|0); //@line 95 "modules/graphics/opengl/Font.cpp"
  __ZN4love8graphics6opengl11QuadIndicesD2Ev($5); //@line 95 "modules/graphics/opengl/Font.cpp"
  $6 = ((($this)) + 80|0); //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $7 = HEAP32[$6>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $8 = ($7|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if (!($8)) {
   $$01$i$i$i = $7;
   while(1) {
    $9 = HEAP32[$$01$i$i$i>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    __ZdlPv($$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $10 = ($9|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($10) {
     break;
    } else {
     $$01$i$i$i = $9;
    }
   }
  }
  $11 = ((($this)) + 72|0); //@line 2014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $12 = HEAP32[$11>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$11>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $13 = ($12|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  if (!($13)) {
   __ZdlPv($12); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  $14 = ((($this)) + 60|0); //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $15 = HEAP32[$14>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $16 = ($15|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if (!($16)) {
   $$01$i$i$i1 = $15;
   while(1) {
    $17 = HEAP32[$$01$i$i$i1>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    __ZdlPv($$01$i$i$i1); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $18 = ($17|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($18) {
     break;
    } else {
     $$01$i$i$i1 = $17;
    }
   }
  }
  $19 = ((($this)) + 52|0); //@line 2014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $20 = HEAP32[$19>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$19>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $21 = ($20|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  if (!($21)) {
   __ZdlPv($20); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  $22 = ((($this)) + 40|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = HEAP32[$22>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $24 = ($23|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $25 = $23; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($24)) {
   $26 = ((($this)) + 44|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $27 = HEAP32[$26>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $28 = ($27|0)==($23|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($28)) {
    $scevgep$i$i$i = ((($27)) + -4|0);
    $29 = $scevgep$i$i$i;
    $30 = (($29) - ($25))|0;
    $31 = $30 >>> 2;
    $32 = $31 ^ -1;
    $scevgep4$i$i$i = (($27) + ($32<<2)|0);
    HEAP32[$26>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($23); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  $33 = ((($this)) + 12|0); //@line 95 "modules/graphics/opengl/Font.cpp"
  $34 = HEAP32[$33>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $35 = ($34|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($35) {
   __ZN4love8graphics8VolatileD2Ev($0); //@line 95 "modules/graphics/opengl/Font.cpp"
   return; //@line 95 "modules/graphics/opengl/Font.cpp"
  }
  $36 = ((($this)) + 16|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $37 = HEAP32[$36>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $38 = ($37|0)==($34|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  do {
   if ($38) {
    $52 = $34;
   } else {
    $40 = $37;
    while(1) {
     $39 = ((($40)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$36>>2] = $39; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $41 = HEAP32[$39>>2]|0; //@line 106 "./common/Object.h"
     $42 = ($41|0)==(0|0); //@line 106 "./common/Object.h"
     if ($42) {
      $44 = $39;
     } else {
      $45 = HEAP32[$41>>2]|0; //@line 106 "./common/Object.h"
      $46 = ((($45)) + 12|0); //@line 106 "./common/Object.h"
      $47 = HEAP32[$46>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($47|0,($41|0)); //@line 106 "./common/Object.h"
      $48 = __THREW__; __THREW__ = 0;
      $49 = $48&1;
      if ($49) {
       label = 20;
       break;
      }
      $$pre$i$i$i = HEAP32[$36>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $44 = $$pre$i$i$i;
     }
     $43 = ($44|0)==($34|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($43) {
      label = 21;
      break;
     } else {
      $40 = $44;
     }
    }
    if ((label|0) == 20) {
     $50 = ___cxa_find_matching_catch_3(0|0)|0;
     $51 = tempRet0;
     ___clang_call_terminate($50); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
    else if ((label|0) == 21) {
     $$pre$i = HEAP32[$33>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $52 = $$pre$i;
     break;
    }
   }
  } while(0);
  __ZdlPv($52); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  __ZN4love8graphics8VolatileD2Ev($0); //@line 95 "modules/graphics/opengl/Font.cpp"
  return; //@line 95 "modules/graphics/opengl/Font.cpp"
 }
}
function __ZN4love8graphics6opengl4Font14unloadVolatileEv($this) {
 $this = $this|0;
 var $$01$i$i$i = 0, $$pre = 0, $$pre2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$01 = 0, $__i$01$i$i = 0, $__new_last2$i$i$i = 0, $exitcond$i$i = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 64|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 1580 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $2 = ($1|0)==(0); //@line 1580 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($2)) {
  $3 = ((($this)) + 60|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $4 = HEAP32[$3>>2]|0; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $5 = ($4|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if (!($5)) {
   $$01$i$i$i = $4;
   while(1) {
    $6 = HEAP32[$$01$i$i$i>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    __ZdlPv($$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $7 = ($6|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($7) {
     break;
    } else {
     $$01$i$i$i = $6;
    }
   }
  }
  HEAP32[$3>>2] = 0; //@line 1583 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $8 = ((($this)) + 56|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $9 = HEAP32[$8>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $10 = ($9|0)==(0); //@line 1585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if (!($10)) {
   $11 = ((($this)) + 52|0); //@line 2015 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $__i$01$i$i = 0;
   while(1) {
    $12 = HEAP32[$11>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $13 = (($12) + ($__i$01$i$i<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$13>>2] = 0; //@line 1586 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $14 = (($__i$01$i$i) + 1)|0; //@line 1585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $exitcond$i$i = ($14|0)==($9|0); //@line 1585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($exitcond$i$i) {
     break;
    } else {
     $__i$01$i$i = $14;
    }
   }
  }
  HEAP32[$0>>2] = 0; //@line 1587 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $15 = ((($this)) + 40|0); //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = HEAP32[$15>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ((($this)) + 44|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $18 = HEAP32[$17>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $19 = ($16|0)==($18|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($19) {
  $25 = ((($this)) + 136|0); //@line 977 "modules/graphics/opengl/Font.cpp"
  $26 = HEAP32[$25>>2]|0; //@line 977 "modules/graphics/opengl/Font.cpp"
  __ZN4love8graphics6opengl6OpenGL23updateTextureMemorySizeEjj(634044,$26,0); //@line 977 "modules/graphics/opengl/Font.cpp"
  HEAP32[$25>>2] = 0; //@line 978 "modules/graphics/opengl/Font.cpp"
  return; //@line 979 "modules/graphics/opengl/Font.cpp"
 } else {
  $__begin$sroa$0$01 = $16;
 }
 while(1) {
  $27 = HEAP32[$__begin$sroa$0$01>>2]|0; //@line 972 "modules/graphics/opengl/Font.cpp"
  __ZN4love8graphics6opengl6OpenGL13deleteTextureEj(634044,$27); //@line 973 "modules/graphics/opengl/Font.cpp"
  $28 = ((($__begin$sroa$0$01)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $29 = ($28|0)==($18|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($29) {
   break;
  } else {
   $__begin$sroa$0$01 = $28;
  }
 }
 $$pre = HEAP32[$15>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $$pre2 = HEAP32[$17>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = ($$pre2|0)==($$pre|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($20) {
  $25 = ((($this)) + 136|0); //@line 977 "modules/graphics/opengl/Font.cpp"
  $26 = HEAP32[$25>>2]|0; //@line 977 "modules/graphics/opengl/Font.cpp"
  __ZN4love8graphics6opengl6OpenGL23updateTextureMemorySizeEjj(634044,$26,0); //@line 977 "modules/graphics/opengl/Font.cpp"
  HEAP32[$25>>2] = 0; //@line 978 "modules/graphics/opengl/Font.cpp"
  return; //@line 979 "modules/graphics/opengl/Font.cpp"
 }
 $__new_last2$i$i$i = $$pre; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $scevgep$i$i$i = ((($$pre2)) + -4|0);
 $21 = $scevgep$i$i$i;
 $22 = (($21) - ($__new_last2$i$i$i))|0;
 $23 = $22 >>> 2;
 $24 = $23 ^ -1;
 $scevgep4$i$i$i = (($$pre2) + ($24<<2)|0);
 HEAP32[$17>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $25 = ((($this)) + 136|0); //@line 977 "modules/graphics/opengl/Font.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 977 "modules/graphics/opengl/Font.cpp"
 __ZN4love8graphics6opengl6OpenGL23updateTextureMemorySizeEjj(634044,$26,0); //@line 977 "modules/graphics/opengl/Font.cpp"
 HEAP32[$25>>2] = 0; //@line 978 "modules/graphics/opengl/Font.cpp"
 return; //@line 979 "modules/graphics/opengl/Font.cpp"
}
function __ZN4love8graphics6opengl4FontD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl4FontD2Ev($this); //@line 91 "modules/graphics/opengl/Font.cpp"
 __ZdlPv($this); //@line 91 "modules/graphics/opengl/Font.cpp"
 return; //@line 95 "modules/graphics/opengl/Font.cpp"
}
function __ZN4love8graphics6opengl4Font12loadVolatileEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl4Font13createTextureEv($this); //@line 961 "modules/graphics/opengl/Font.cpp"
 $0 = ((($this)) + 132|0); //@line 962 "modules/graphics/opengl/Font.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 962 "modules/graphics/opengl/Font.cpp"
 $2 = (($1) + 1)|0; //@line 962 "modules/graphics/opengl/Font.cpp"
 HEAP32[$0>>2] = $2; //@line 962 "modules/graphics/opengl/Font.cpp"
 return 1; //@line 963 "modules/graphics/opengl/Font.cpp"
}
function __ZN4love8graphics6opengl4Font13createTextureEv($this) {
 $this = $this|0;
 var $$ = 0, $$0 = 0, $$0$i$i = 0, $$01 = 0, $$01$i$i$i = 0, $$pre = 0, $$pre$phi34Z2D = 0, $$pre28 = 0, $$pre29 = 0, $$pre30 = 0, $$pre32 = 0, $$pre33 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$023 = 0, $__begin$sroa$0$024 = 0;
 var $__begin2$sroa$0$022 = 0, $__i$01$i$i = 0, $emptydata = 0, $exitcond$i$i = 0, $glyphstoadd = 0, $lpad$phi$index = 0, $lpad$phi$index7 = 0, $lpad$phi21$index = 0, $lpad$phi21$index2 = 0, $nextsize$sroa$0$013 = 0, $nextsize$sroa$0$014 = 0, $nextsize$sroa$6$015 = 0, $nextsize$sroa$6$016 = 0, $or$cond$i = 0, $recreatetexture$0$off0 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i3 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i4 = 0, $size$sroa$0$0 = 0;
 var $size$sroa$8$0 = 0, $t = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $t = sp + 28|0;
 $emptydata = sp + 16|0;
 $glyphstoadd = sp + 4|0;
 $0 = ((($this)) + 92|0); //@line 119 "modules/graphics/opengl/Font.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 119 "modules/graphics/opengl/Font.cpp"
 $2 = ($1|0)==(0); //@line 119 "modules/graphics/opengl/Font.cpp"
 $3 = $2 ? 6410 : 6408; //@line 119 "modules/graphics/opengl/Font.cpp"
 $4 = $2 ? 2 : 4; //@line 120 "modules/graphics/opengl/Font.cpp"
 $5 = ((($this)) + 136|0); //@line 122 "modules/graphics/opengl/Font.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 122 "modules/graphics/opengl/Font.cpp"
 $7 = ($6|0)==(0); //@line 123 "modules/graphics/opengl/Font.cpp"
 $$pre32 = ((($this)) + 32|0); //@line 130 "modules/graphics/opengl/Font.cpp"
 if ($7) {
  $$pre33 = ((($this)) + 36|0); //@line 130 "modules/graphics/opengl/Font.cpp"
  $$pre$phi34Z2D = $$pre33;
 } else {
  $8 = HEAP32[$$pre32>>2]|0; //@line 125 "modules/graphics/opengl/Font.cpp"
  $9 = ((($this)) + 36|0); //@line 125 "modules/graphics/opengl/Font.cpp"
  $10 = HEAP32[$9>>2]|0; //@line 125 "modules/graphics/opengl/Font.cpp"
  $11 = Math_imul($8, $4)|0; //@line 125 "modules/graphics/opengl/Font.cpp"
  $12 = Math_imul($11, $10)|0; //@line 125 "modules/graphics/opengl/Font.cpp"
  $13 = (($6) - ($12))|0; //@line 125 "modules/graphics/opengl/Font.cpp"
  HEAP32[$5>>2] = $13; //@line 125 "modules/graphics/opengl/Font.cpp"
  __ZN4love8graphics6opengl6OpenGL23updateTextureMemorySizeEjj(634044,$6,$13); //@line 126 "modules/graphics/opengl/Font.cpp"
  $$pre$phi34Z2D = $9;
 }
 HEAP32[$t>>2] = 0; //@line 129 "modules/graphics/opengl/Font.cpp"
 $14 = HEAP32[$$pre32>>2]|0; //@line 130 "modules/graphics/opengl/Font.cpp"
 $15 = HEAP32[$$pre$phi34Z2D>>2]|0; //@line 130 "modules/graphics/opengl/Font.cpp"
 $16 = (__ZNK4love8graphics6opengl6OpenGL17getMaxTextureSizeEv(634044)|0); //@line 101 "modules/graphics/opengl/Font.cpp"
 $17 = ($16|0)<(4096); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $18 = $17 ? $16 : 4096; //@line 101 "modules/graphics/opengl/Font.cpp"
 $19 = $14 << 1; //@line 103 "modules/graphics/opengl/Font.cpp"
 $20 = ($19|0)>($18|0); //@line 103 "modules/graphics/opengl/Font.cpp"
 $21 = $15 << 1; //@line 103 "modules/graphics/opengl/Font.cpp"
 $22 = ($21|0)>($18|0); //@line 103 "modules/graphics/opengl/Font.cpp"
 $or$cond$i = $20 & $22; //@line 103 "modules/graphics/opengl/Font.cpp"
 do {
  if ($or$cond$i) {
   label = 10;
  } else {
   $23 = ($14|0)==($15|0); //@line 106 "modules/graphics/opengl/Font.cpp"
   if ($23) {
    $24 = ($19|0)>($14|0); //@line 137 "modules/graphics/opengl/Font.cpp"
    if ($24) {
     $nextsize$sroa$0$014 = $19;$nextsize$sroa$6$016 = $14;
    } else {
     $nextsize$sroa$0$013 = $19;$nextsize$sroa$6$015 = $14;
     label = 7;
    }
   } else {
    $nextsize$sroa$0$013 = $14;$nextsize$sroa$6$015 = $21;
    label = 7;
   }
   if ((label|0) == 7) {
    $25 = ($nextsize$sroa$6$015|0)>($15|0); //@line 137 "modules/graphics/opengl/Font.cpp"
    if ($25) {
     $nextsize$sroa$0$014 = $nextsize$sroa$0$013;$nextsize$sroa$6$016 = $nextsize$sroa$6$015;
    } else {
     label = 10;
     break;
    }
   }
   $26 = ((($this)) + 40|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $27 = HEAP32[$26>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $28 = ((($this)) + 44|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $29 = HEAP32[$28>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $30 = ($27|0)==($29|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($30) {
    label = 10;
   } else {
    $31 = ((($29)) + -4|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $32 = HEAP32[$31>>2]|0; //@line 142 "modules/graphics/opengl/Font.cpp"
    HEAP32[$t>>2] = $32; //@line 142 "modules/graphics/opengl/Font.cpp"
    $34 = $32;$recreatetexture$0$off0 = 1;$size$sroa$0$0 = $nextsize$sroa$0$014;$size$sroa$8$0 = $nextsize$sroa$6$016;
   }
  }
 } while(0);
 if ((label|0) == 10) {
  $33 = HEAP32[164922]|0; //@line 355 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$33 & 8191](1,$t); //@line 355 "./libraries/glad/gladfuncs.hpp"
  $$pre = HEAP32[$t>>2]|0; //@line 147 "modules/graphics/opengl/Font.cpp"
  $34 = $$pre;$recreatetexture$0$off0 = 0;$size$sroa$0$0 = $14;$size$sroa$8$0 = $15;
 }
 __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$34); //@line 147 "modules/graphics/opengl/Font.cpp"
 $35 = ((($this)) + 96|0); //@line 149 "modules/graphics/opengl/Font.cpp"
 __ZN4love8graphics6opengl6OpenGL16setTextureFilterERNS0_7Texture6FilterE(634044,$35); //@line 149 "modules/graphics/opengl/Font.cpp"
 $36 = HEAP32[165108]|0; //@line 44 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$36 & 8191](3553,10242,33071); //@line 44 "./libraries/glad/gladfuncs.hpp"
 $37 = HEAP32[165108]|0; //@line 44 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$37 & 8191](3553,10243,33071); //@line 44 "./libraries/glad/gladfuncs.hpp"
 $38 = HEAP32[$0>>2]|0; //@line 154 "modules/graphics/opengl/Font.cpp"
 $39 = ($38|0)==(0); //@line 154 "modules/graphics/opengl/Font.cpp"
 $40 = $39 ? 32837 : 32856; //@line 154 "modules/graphics/opengl/Font.cpp"
 $41 = HEAP8[1252012]|0; //@line 159 "modules/graphics/opengl/Font.cpp"
 $42 = ($41<<24>>24)==(0); //@line 159 "modules/graphics/opengl/Font.cpp"
 $$ = $42 ? $40 : $3; //@line 159 "modules/graphics/opengl/Font.cpp"
 $43 = Math_imul($size$sroa$0$0, $4)|0; //@line 163 "modules/graphics/opengl/Font.cpp"
 $44 = Math_imul($43, $size$sroa$8$0)|0; //@line 163 "modules/graphics/opengl/Font.cpp"
 HEAP32[$emptydata>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $45 = ((($emptydata)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$45>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $46 = ((($emptydata)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$46>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = ($44|0)==(0); //@line 1109 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 L16: do {
  if (!($47)) {
   $48 = ($44|0)<(0); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($48) {
    __THREW__ = 0;
    invoke_vi(3528,($emptydata|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $49 = __THREW__; __THREW__ = 0;
    $50 = $49&1;
    if (!($50)) {
     label = 14;
    }
   } else {
    label = 14;
   }
   if ((label|0) == 14) {
    __THREW__ = 0;
    $51 = (invoke_ii(3348,($44|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $52 = __THREW__; __THREW__ = 0;
    $53 = $52&1;
    if (!($53)) {
     HEAP32[$45>>2] = $51; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$emptydata>>2] = $51; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $54 = (($51) + ($44)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $55 = ((($emptydata)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$55>>2] = $54; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $$0$i$i = $44;$56 = $51;
     while(1) {
      HEAP8[$56>>0] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $57 = HEAP32[$45>>2]|0; //@line 1011 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $58 = ((($57)) + 1|0); //@line 1011 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$45>>2] = $58; //@line 1011 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $59 = (($$0$i$i) + -1)|0; //@line 1012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $60 = ($59|0)==(0); //@line 1014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($60) {
       break L16;
      } else {
       $$0$i$i = $59;$56 = $58;
      }
     }
    }
   }
   $61 = ___cxa_find_matching_catch_2()|0;
   $62 = tempRet0;
   $63 = HEAP32[$emptydata>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $64 = ($63|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($64) {
    ___resumeException($61|0);
    // unreachable;
   }
   $65 = HEAP32[$45>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $66 = ($65|0)==($63|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($66)) {
    HEAP32[$45>>2] = $63; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($63); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($61|0);
   // unreachable;
  }
 } while(0);
 while(1) {
  $67 = HEAP32[165571]|0; //@line 71 "./libraries/glad/gladfuncs.hpp"
  __THREW__ = 0;
  $68 = (invoke_i($67|0)|0); //@line 71 "./libraries/glad/gladfuncs.hpp"
  $69 = __THREW__; __THREW__ = 0;
  $70 = $69&1;
  if ($70) {
   label = 24;
   break;
  }
  $71 = ($68|0)==(0); //@line 166 "modules/graphics/opengl/Font.cpp"
  if ($71) {
   label = 27;
   break;
  }
 }
 L36: do {
  if ((label|0) == 24) {
   $72 = ___cxa_find_matching_catch_2()|0;
   $73 = tempRet0;
   $lpad$phi21$index = $72;$lpad$phi21$index2 = $73;
   label = 26;
  }
  else if ((label|0) == 27) {
   $76 = HEAP32[$emptydata>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $77 = HEAP32[165602]|0; //@line 47 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   invoke_viiiiiiiii($77|0,3553,0,($$|0),($size$sroa$0$0|0),($size$sroa$8$0|0),0,($3|0),5121,($76|0)); //@line 47 "./libraries/glad/gladfuncs.hpp"
   $78 = __THREW__; __THREW__ = 0;
   $79 = $78&1;
   L39: do {
    if (!($79)) {
     $80 = HEAP32[165571]|0; //@line 71 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     $81 = (invoke_i($80|0)|0); //@line 71 "./libraries/glad/gladfuncs.hpp"
     $82 = __THREW__; __THREW__ = 0;
     $83 = $82&1;
     if (!($83)) {
      $84 = ($81|0)==(0); //@line 171 "modules/graphics/opengl/Font.cpp"
      if (!($84)) {
       if (!($recreatetexture$0$off0)) {
        $85 = HEAP32[$t>>2]|0; //@line 174 "modules/graphics/opengl/Font.cpp"
        __THREW__ = 0;
        invoke_vii(4014,(634044|0),($85|0)); //@line 174 "modules/graphics/opengl/Font.cpp"
        $86 = __THREW__; __THREW__ = 0;
        $87 = $86&1;
        if ($87) {
         break;
        }
       }
       $88 = (___cxa_allocate_exception(16)|0); //@line 175 "modules/graphics/opengl/Font.cpp"
       __THREW__ = 0;
       invoke_viii(3349,($88|0),(434471|0),($vararg_buffer|0)); //@line 175 "modules/graphics/opengl/Font.cpp"
       $89 = __THREW__; __THREW__ = 0;
       $90 = $89&1;
       if ($90) {
        $92 = ___cxa_find_matching_catch_2()|0;
        $93 = tempRet0;
        ___cxa_free_exception(($88|0)); //@line 175 "modules/graphics/opengl/Font.cpp"
        $$0 = $93;$$01 = $92;
        break L36;
       } else {
        __THREW__ = 0;
        invoke_viii(3350,($88|0),(96|0),(344|0)); //@line 175 "modules/graphics/opengl/Font.cpp"
        $91 = __THREW__; __THREW__ = 0;
        break;
       }
      }
      HEAP32[$$pre32>>2] = $size$sroa$0$0; //@line 178 "modules/graphics/opengl/Font.cpp"
      HEAP32[$$pre$phi34Z2D>>2] = $size$sroa$8$0; //@line 179 "modules/graphics/opengl/Font.cpp"
      $94 = ((($this)) + 116|0); //@line 181 "modules/graphics/opengl/Font.cpp"
      HEAP32[$94>>2] = 1; //@line 181 "modules/graphics/opengl/Font.cpp"
      $95 = ((($this)) + 112|0); //@line 181 "modules/graphics/opengl/Font.cpp"
      HEAP32[$95>>2] = 1; //@line 181 "modules/graphics/opengl/Font.cpp"
      $96 = ((($this)) + 120|0); //@line 181 "modules/graphics/opengl/Font.cpp"
      HEAP32[$96>>2] = 1; //@line 181 "modules/graphics/opengl/Font.cpp"
      $97 = HEAP32[$5>>2]|0; //@line 183 "modules/graphics/opengl/Font.cpp"
      $98 = HEAP32[$45>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $99 = HEAP32[$emptydata>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $100 = (($98) - ($99))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $101 = (($100) + ($97))|0; //@line 184 "modules/graphics/opengl/Font.cpp"
      HEAP32[$5>>2] = $101; //@line 184 "modules/graphics/opengl/Font.cpp"
      __THREW__ = 0;
      invoke_viii(4015,(634044|0),($97|0),($101|0)); //@line 185 "modules/graphics/opengl/Font.cpp"
      $102 = __THREW__; __THREW__ = 0;
      $103 = $102&1;
      if (!($103)) {
       L52: do {
        if ($recreatetexture$0$off0) {
         $104 = ((($this)) + 132|0); //@line 190 "modules/graphics/opengl/Font.cpp"
         $105 = HEAP32[$104>>2]|0; //@line 190 "modules/graphics/opengl/Font.cpp"
         $106 = (($105) + 1)|0; //@line 190 "modules/graphics/opengl/Font.cpp"
         HEAP32[$104>>2] = $106; //@line 190 "modules/graphics/opengl/Font.cpp"
         HEAP32[$glyphstoadd>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $107 = ((($glyphstoadd)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$107>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $108 = ((($glyphstoadd)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$108>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $109 = ((($this)) + 60|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $__begin$sroa$0$023 = HEAP32[$109>>2]|0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $110 = ($__begin$sroa$0$023|0)==(0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         L54: do {
          if ($110) {
           label = 39;
          } else {
           $111 = ((($glyphstoadd)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $145 = 0;$146 = 0;$__begin$sroa$0$024 = $__begin$sroa$0$023;
           while(1) {
            $142 = $__begin$sroa$0$024; //@line 189 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
            $143 = ((($142)) + 8|0); //@line 195 "modules/graphics/opengl/Font.cpp"
            $144 = ($145|0)==($146|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            if ($144) {
             __THREW__ = 0;
             invoke_vii(4016,($glyphstoadd|0),($143|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $149 = __THREW__; __THREW__ = 0;
             $150 = $149&1;
             if ($150) {
              break;
             }
            } else {
             $147 = HEAP32[$143>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
             HEAP32[$145>>2] = $147; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
             $148 = ((($145)) + 4|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             HEAP32[$107>>2] = $148; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            }
            $151 = $__begin$sroa$0$024; //@line 174 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
            $__begin$sroa$0$0 = HEAP32[$151>>2]|0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
            $152 = ($__begin$sroa$0$0|0)==(0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
            if ($152) {
             label = 39;
             break L54;
            }
            $$pre28 = HEAP32[$107>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $$pre29 = HEAP32[$111>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $145 = $$pre28;$146 = $$pre29;$__begin$sroa$0$024 = $__begin$sroa$0$0;
           }
           $131 = ___cxa_find_matching_catch_2()|0;
           $132 = tempRet0;
           $lpad$phi$index = $131;$lpad$phi$index7 = $132;
          }
         } while(0);
         L64: do {
          if ((label|0) == 39) {
           $112 = ((($this)) + 64|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $113 = HEAP32[$112>>2]|0; //@line 1580 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
           $114 = ($113|0)==(0); //@line 1580 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
           if (!($114)) {
            $115 = HEAP32[$109>>2]|0; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
            $116 = ($115|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
            if (!($116)) {
             $$01$i$i$i = $115;
             while(1) {
              $117 = HEAP32[$$01$i$i$i>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
              __ZdlPv($$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
              $118 = ($117|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
              if ($118) {
               break;
              } else {
               $$01$i$i$i = $117;
              }
             }
            }
            HEAP32[$109>>2] = 0; //@line 1583 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
            $119 = ((($this)) + 56|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
            $120 = HEAP32[$119>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
            $121 = ($120|0)==(0); //@line 1585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
            if (!($121)) {
             $122 = ((($this)) + 52|0); //@line 2015 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
             $__i$01$i$i = 0;
             while(1) {
              $123 = HEAP32[$122>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
              $124 = (($123) + ($__i$01$i$i<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
              HEAP32[$124>>2] = 0; //@line 1586 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
              $125 = (($__i$01$i$i) + 1)|0; //@line 1585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
              $exitcond$i$i = ($125|0)==($120|0); //@line 1585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
              if ($exitcond$i$i) {
               break;
              } else {
               $__i$01$i$i = $125;
              }
             }
            }
            HEAP32[$112>>2] = 0; //@line 1587 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
           }
           $126 = HEAP32[$glyphstoadd>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $127 = HEAP32[$107>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $128 = ($126|0)==($127|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
           do {
            if ($128) {
             $154 = $126;
            } else {
             $__begin2$sroa$0$022 = $126;
             while(1) {
              $162 = HEAP32[$__begin2$sroa$0$022>>2]|0; //@line 199 "modules/graphics/opengl/Font.cpp"
              __THREW__ = 0;
              (invoke_iii(4017,($this|0),($162|0))|0); //@line 200 "modules/graphics/opengl/Font.cpp"
              $163 = __THREW__; __THREW__ = 0;
              $164 = $163&1;
              if ($164) {
               label = 47;
               break;
              }
              $165 = ((($__begin2$sroa$0$022)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
              $166 = ($165|0)==($127|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
              if ($166) {
               label = 58;
               break;
              } else {
               $__begin2$sroa$0$022 = $165;
              }
             }
             if ((label|0) == 47) {
              $129 = ___cxa_find_matching_catch_2()|0;
              $130 = tempRet0;
              $lpad$phi$index = $129;$lpad$phi$index7 = $130;
              break L64;
             }
             else if ((label|0) == 58) {
              $$pre30 = HEAP32[$glyphstoadd>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $154 = $$pre30;
              break;
             }
            }
           } while(0);
           $153 = ($154|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $155 = $154; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           if (!($153)) {
            $156 = HEAP32[$107>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $157 = ($156|0)==($154|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            if (!($157)) {
             $scevgep$i$i$i3 = ((($156)) + -4|0);
             $158 = $scevgep$i$i$i3;
             $159 = (($158) - ($155))|0;
             $160 = $159 >>> 2;
             $161 = $160 ^ -1;
             $scevgep4$i$i$i4 = (($156) + ($161<<2)|0);
             HEAP32[$107>>2] = $scevgep4$i$i$i4; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            }
            __ZdlPv($154); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
           }
           break L52;
          }
         } while(0);
         $133 = HEAP32[$glyphstoadd>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $134 = ($133|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $135 = $133; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($134) {
          $$0 = $lpad$phi$index7;$$01 = $lpad$phi$index;
          break L36;
         }
         $136 = HEAP32[$107>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $137 = ($136|0)==($133|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if (!($137)) {
          $scevgep$i$i$i = ((($136)) + -4|0);
          $138 = $scevgep$i$i$i;
          $139 = (($138) - ($135))|0;
          $140 = $139 >>> 2;
          $141 = $140 ^ -1;
          $scevgep4$i$i$i = (($136) + ($141<<2)|0);
          HEAP32[$107>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         }
         __ZdlPv($133); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
         $$0 = $lpad$phi$index7;$$01 = $lpad$phi$index;
         break L36;
        } else {
         $167 = ((($this)) + 44|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $168 = HEAP32[$167>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $169 = ((($this)) + 48|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $170 = HEAP32[$169>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $171 = ($168|0)==($170|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($171) {
          $174 = ((($this)) + 40|0); //@line 203 "modules/graphics/opengl/Font.cpp"
          __THREW__ = 0;
          invoke_vii(4016,($174|0),($t|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $175 = __THREW__; __THREW__ = 0;
          $176 = $175&1;
          if ($176) {
           break L39;
          } else {
           break;
          }
         } else {
          $172 = HEAP32[$t>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          HEAP32[$168>>2] = $172; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $173 = ((($168)) + 4|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$167>>2] = $173; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          break;
         }
        }
       } while(0);
       $177 = HEAP32[$emptydata>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $178 = ($177|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($178) {
        STACKTOP = sp;return; //@line 204 "modules/graphics/opengl/Font.cpp"
       }
       $179 = HEAP32[$45>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $180 = ($179|0)==($177|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($180)) {
        HEAP32[$45>>2] = $177; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       __ZdlPv($177); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       STACKTOP = sp;return; //@line 204 "modules/graphics/opengl/Font.cpp"
      }
     }
    }
   } while(0);
   $74 = ___cxa_find_matching_catch_2()|0;
   $75 = tempRet0;
   $lpad$phi21$index = $74;$lpad$phi21$index2 = $75;
   label = 26;
  }
 } while(0);
 if ((label|0) == 26) {
  $$0 = $lpad$phi21$index2;$$01 = $lpad$phi21$index;
 }
 $181 = HEAP32[$emptydata>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $182 = ($181|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($182) {
  ___resumeException($$01|0);
  // unreachable;
 }
 $183 = HEAP32[$45>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $184 = ($183|0)==($181|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($184)) {
  HEAP32[$45>>2] = $181; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($181); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$01|0);
 // unreachable;
}
function __ZNSt3__16vectorIjNS_9allocatorIjEEE21__push_back_slow_pathIRKjEEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl4Font8addGlyphEj($this,$glyph) {
 $this = $this|0;
 $glyph = $glyph|0;
 var $$0 = 0, $$02 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0.0, $11 = 0, $110 = 0.0;
 var $111 = 0.0, $112 = 0, $113 = 0.0, $114 = 0, $115 = 0.0, $116 = 0.0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0, $124 = 0.0, $125 = 0, $126 = 0.0, $127 = 0.0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0.0, $132 = 0.0, $133 = 0.0, $134 = 0, $135 = 0.0, $136 = 0, $137 = 0.0, $138 = 0.0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0, $146 = 0.0, $147 = 0;
 var $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0, $72 = 0.0, $73 = 0.0, $74 = 0, $75 = 0, $76 = 0.0, $77 = 0.0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0.0, $83 = 0, $84 = 0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0.0, $93 = 0, $94 = 0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0;
 var $99 = 0, $g = 0, $not$$i$i = 0, $or$cond = 0, $verts = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 136|0;
 $g = sp + 64|0;
 $verts = sp;
 $1 = (__ZN4love8graphics6opengl4Font22getRasterizerGlyphDataEj($this,$glyph)|0); //@line 235 "modules/graphics/opengl/Font.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 237 "modules/graphics/opengl/Font.cpp"
 $3 = ((($2)) + 28|0); //@line 237 "modules/graphics/opengl/Font.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 237 "modules/graphics/opengl/Font.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($1)|0); //@line 237 "modules/graphics/opengl/Font.cpp"
 $6 = HEAP32[$1>>2]|0; //@line 238 "modules/graphics/opengl/Font.cpp"
 $7 = ((($6)) + 24|0); //@line 238 "modules/graphics/opengl/Font.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 238 "modules/graphics/opengl/Font.cpp"
 $9 = (FUNCTION_TABLE_ii[$8 & 8191]($1)|0); //@line 238 "modules/graphics/opengl/Font.cpp"
 $10 = ((($this)) + 112|0); //@line 240 "modules/graphics/opengl/Font.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 240 "modules/graphics/opengl/Font.cpp"
 $12 = (($11) + ($5))|0; //@line 240 "modules/graphics/opengl/Font.cpp"
 $13 = ((($this)) + 32|0); //@line 240 "modules/graphics/opengl/Font.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 240 "modules/graphics/opengl/Font.cpp"
 $15 = ($12|0)<($14|0); //@line 240 "modules/graphics/opengl/Font.cpp"
 if ($15) {
  $$phi$trans$insert = ((($this)) + 116|0); //@line 247 "modules/graphics/opengl/Font.cpp"
  $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 247 "modules/graphics/opengl/Font.cpp"
  $$pre$phiZ2D = $$phi$trans$insert;$22 = $$pre;
 } else {
  HEAP32[$10>>2] = 1; //@line 243 "modules/graphics/opengl/Font.cpp"
  $16 = ((($this)) + 120|0); //@line 244 "modules/graphics/opengl/Font.cpp"
  $17 = HEAP32[$16>>2]|0; //@line 244 "modules/graphics/opengl/Font.cpp"
  $18 = ((($this)) + 116|0); //@line 244 "modules/graphics/opengl/Font.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 244 "modules/graphics/opengl/Font.cpp"
  $20 = (($19) + ($17))|0; //@line 244 "modules/graphics/opengl/Font.cpp"
  HEAP32[$18>>2] = $20; //@line 244 "modules/graphics/opengl/Font.cpp"
  HEAP32[$16>>2] = 1; //@line 245 "modules/graphics/opengl/Font.cpp"
  $$pre$phiZ2D = $18;$22 = $20;
 }
 $21 = (($22) + ($9))|0; //@line 247 "modules/graphics/opengl/Font.cpp"
 $23 = ((($this)) + 36|0); //@line 247 "modules/graphics/opengl/Font.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 247 "modules/graphics/opengl/Font.cpp"
 $25 = ($21|0)<($24|0); //@line 247 "modules/graphics/opengl/Font.cpp"
 if (!($25)) {
  __THREW__ = 0;
  invoke_vi(4018,($this|0)); //@line 252 "modules/graphics/opengl/Font.cpp"
  $26 = __THREW__; __THREW__ = 0;
  $27 = $26&1;
  if ($27) {
   $28 = ___cxa_find_matching_catch_3(96|0)|0;
   $29 = tempRet0;
   $30 = (_llvm_eh_typeid_for((96|0))|0); //@line 253 "modules/graphics/opengl/Font.cpp"
   $31 = ($29|0)==($30|0); //@line 253 "modules/graphics/opengl/Font.cpp"
   if (!($31)) {
    $$0 = $28;$$02 = $29;
    ___resumeException($$0|0);
    // unreachable;
   }
   (___cxa_begin_catch(($28|0))|0); //@line 253 "modules/graphics/opengl/Font.cpp"
   $32 = HEAP32[$1>>2]|0; //@line 256 "modules/graphics/opengl/Font.cpp"
   $33 = ((($32)) + 12|0); //@line 256 "modules/graphics/opengl/Font.cpp"
   $34 = HEAP32[$33>>2]|0; //@line 256 "modules/graphics/opengl/Font.cpp"
   __THREW__ = 0;
   invoke_vi($34|0,($1|0)); //@line 256 "modules/graphics/opengl/Font.cpp"
   $35 = __THREW__; __THREW__ = 0;
   $36 = $35&1;
   if (!($36)) {
    __THREW__ = 0;
    invoke_v(3656); //@line 257 "modules/graphics/opengl/Font.cpp"
    $37 = __THREW__; __THREW__ = 0;
   }
   $38 = ___cxa_find_matching_catch_2()|0;
   $39 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 258 "modules/graphics/opengl/Font.cpp"
   $40 = __THREW__; __THREW__ = 0;
   $41 = $40&1;
   if ($41) {
    $174 = ___cxa_find_matching_catch_3(0|0)|0;
    $175 = tempRet0;
    ___clang_call_terminate($174); //@line 258 "modules/graphics/opengl/Font.cpp"
    // unreachable; //@line 258 "modules/graphics/opengl/Font.cpp"
   } else {
    $$0 = $38;$$02 = $39;
    ___resumeException($$0|0);
    // unreachable;
   }
  }
 }
 HEAP32[$g>>2] = 0; //@line 263 "modules/graphics/opengl/Font.cpp"
 $42 = (__ZNK4love4font9GlyphData10getAdvanceEv($1)|0); //@line 264 "modules/graphics/opengl/Font.cpp"
 $43 = ((($g)) + 4|0); //@line 264 "modules/graphics/opengl/Font.cpp"
 HEAP32[$43>>2] = $42; //@line 264 "modules/graphics/opengl/Font.cpp"
 $44 = ((($g)) + 8|0); //@line 266 "modules/graphics/opengl/Font.cpp"
 dest=$44; stop=dest+64|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 266 "modules/graphics/opengl/Font.cpp"
 $45 = ($5|0)>(0); //@line 269 "modules/graphics/opengl/Font.cpp"
 $46 = ($9|0)>(0); //@line 269 "modules/graphics/opengl/Font.cpp"
 $or$cond = $45 & $46; //@line 269 "modules/graphics/opengl/Font.cpp"
 if ($or$cond) {
  $47 = ((($this)) + 44|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $48 = HEAP32[$47>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $49 = ((($48)) + -4|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $50 = HEAP32[$49>>2]|0; //@line 271 "modules/graphics/opengl/Font.cpp"
  __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$50); //@line 273 "modules/graphics/opengl/Font.cpp"
  $51 = HEAP32[$10>>2]|0; //@line 274 "modules/graphics/opengl/Font.cpp"
  $52 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 274 "modules/graphics/opengl/Font.cpp"
  $53 = ((($this)) + 92|0); //@line 275 "modules/graphics/opengl/Font.cpp"
  $54 = HEAP32[$53>>2]|0; //@line 275 "modules/graphics/opengl/Font.cpp"
  $55 = ($54|0)==(0); //@line 275 "modules/graphics/opengl/Font.cpp"
  $56 = $55 ? 6410 : 6408; //@line 275 "modules/graphics/opengl/Font.cpp"
  $57 = HEAP32[$1>>2]|0; //@line 276 "modules/graphics/opengl/Font.cpp"
  $58 = ((($57)) + 16|0); //@line 276 "modules/graphics/opengl/Font.cpp"
  $59 = HEAP32[$58>>2]|0; //@line 276 "modules/graphics/opengl/Font.cpp"
  $60 = (FUNCTION_TABLE_ii[$59 & 8191]($1)|0); //@line 276 "modules/graphics/opengl/Font.cpp"
  $61 = HEAP32[165192]|0; //@line 352 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiiiiiiii[$61 & 8191](3553,0,$51,$52,$5,$9,$56,5121,$60); //@line 352 "./libraries/glad/gladfuncs.hpp"
  HEAP32[$g>>2] = $50; //@line 278 "modules/graphics/opengl/Font.cpp"
  $62 = HEAP32[$10>>2]|0; //@line 280 "modules/graphics/opengl/Font.cpp"
  $63 = (+($62|0)); //@line 280 "modules/graphics/opengl/Font.cpp"
  $64 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 280 "modules/graphics/opengl/Font.cpp"
  $65 = (+($64|0)); //@line 280 "modules/graphics/opengl/Font.cpp"
  $66 = HEAP32[$13>>2]|0; //@line 281 "modules/graphics/opengl/Font.cpp"
  $67 = (+($66|0)); //@line 281 "modules/graphics/opengl/Font.cpp"
  $68 = HEAP32[$23>>2]|0; //@line 281 "modules/graphics/opengl/Font.cpp"
  $69 = (+($68|0)); //@line 281 "modules/graphics/opengl/Font.cpp"
  HEAPF32[$verts>>2] = 0.0; //@line 290 "modules/graphics/opengl/Font.cpp"
  $70 = ((($verts)) + 4|0); //@line 290 "modules/graphics/opengl/Font.cpp"
  HEAPF32[$70>>2] = 0.0; //@line 290 "modules/graphics/opengl/Font.cpp"
  $71 = ((($verts)) + 8|0); //@line 290 "modules/graphics/opengl/Font.cpp"
  $72 = $63 / $67; //@line 290 "modules/graphics/opengl/Font.cpp"
  $73 = $72 * 65535.0; //@line 43 "modules/graphics/opengl/Font.cpp"
  $74 = (~~(($73))&65535); //@line 43 "modules/graphics/opengl/Font.cpp"
  HEAP16[$71>>1] = $74; //@line 290 "modules/graphics/opengl/Font.cpp"
  $75 = ((($verts)) + 10|0); //@line 290 "modules/graphics/opengl/Font.cpp"
  $76 = $65 / $69; //@line 290 "modules/graphics/opengl/Font.cpp"
  $77 = $76 * 65535.0; //@line 43 "modules/graphics/opengl/Font.cpp"
  $78 = (~~(($77))&65535); //@line 43 "modules/graphics/opengl/Font.cpp"
  HEAP16[$75>>1] = $78; //@line 290 "modules/graphics/opengl/Font.cpp"
  $79 = ((($verts)) + 12|0); //@line 290 "modules/graphics/opengl/Font.cpp"
  HEAP32[$79>>2] = -1; //@line 290 "modules/graphics/opengl/Font.cpp"
  $80 = ((($verts)) + 16|0); //@line 291 "modules/graphics/opengl/Font.cpp"
  HEAPF32[$80>>2] = 0.0; //@line 291 "modules/graphics/opengl/Font.cpp"
  $81 = ((($verts)) + 20|0); //@line 291 "modules/graphics/opengl/Font.cpp"
  $82 = (+($9|0)); //@line 291 "modules/graphics/opengl/Font.cpp"
  HEAPF32[$81>>2] = $82; //@line 291 "modules/graphics/opengl/Font.cpp"
  $83 = ((($verts)) + 24|0); //@line 291 "modules/graphics/opengl/Font.cpp"
  HEAP16[$83>>1] = $74; //@line 291 "modules/graphics/opengl/Font.cpp"
  $84 = ((($verts)) + 26|0); //@line 291 "modules/graphics/opengl/Font.cpp"
  $85 = (+($9|0)); //@line 291 "modules/graphics/opengl/Font.cpp"
  $86 = $85 + $65; //@line 291 "modules/graphics/opengl/Font.cpp"
  $87 = $86 / $69; //@line 291 "modules/graphics/opengl/Font.cpp"
  $88 = $87 * 65535.0; //@line 43 "modules/graphics/opengl/Font.cpp"
  $89 = (~~(($88))&65535); //@line 43 "modules/graphics/opengl/Font.cpp"
  HEAP16[$84>>1] = $89; //@line 291 "modules/graphics/opengl/Font.cpp"
  $90 = ((($verts)) + 28|0); //@line 291 "modules/graphics/opengl/Font.cpp"
  HEAP32[$90>>2] = -1; //@line 291 "modules/graphics/opengl/Font.cpp"
  $91 = ((($verts)) + 32|0); //@line 292 "modules/graphics/opengl/Font.cpp"
  $92 = (+($5|0)); //@line 292 "modules/graphics/opengl/Font.cpp"
  HEAPF32[$91>>2] = $92; //@line 292 "modules/graphics/opengl/Font.cpp"
  $93 = ((($verts)) + 36|0); //@line 292 "modules/graphics/opengl/Font.cpp"
  HEAPF32[$93>>2] = 0.0; //@line 292 "modules/graphics/opengl/Font.cpp"
  $94 = ((($verts)) + 40|0); //@line 292 "modules/graphics/opengl/Font.cpp"
  $95 = (+($5|0)); //@line 292 "modules/graphics/opengl/Font.cpp"
  $96 = $95 + $63; //@line 292 "modules/graphics/opengl/Font.cpp"
  $97 = $96 / $67; //@line 292 "modules/graphics/opengl/Font.cpp"
  $98 = $97 * 65535.0; //@line 43 "modules/graphics/opengl/Font.cpp"
  $99 = (~~(($98))&65535); //@line 43 "modules/graphics/opengl/Font.cpp"
  HEAP16[$94>>1] = $99; //@line 292 "modules/graphics/opengl/Font.cpp"
  $100 = ((($verts)) + 42|0); //@line 292 "modules/graphics/opengl/Font.cpp"
  HEAP16[$100>>1] = $78; //@line 292 "modules/graphics/opengl/Font.cpp"
  $101 = ((($verts)) + 44|0); //@line 292 "modules/graphics/opengl/Font.cpp"
  HEAP32[$101>>2] = -1; //@line 292 "modules/graphics/opengl/Font.cpp"
  $102 = ((($verts)) + 48|0); //@line 293 "modules/graphics/opengl/Font.cpp"
  HEAPF32[$102>>2] = $92; //@line 293 "modules/graphics/opengl/Font.cpp"
  $103 = ((($verts)) + 52|0); //@line 293 "modules/graphics/opengl/Font.cpp"
  HEAPF32[$103>>2] = $82; //@line 293 "modules/graphics/opengl/Font.cpp"
  $104 = ((($verts)) + 56|0); //@line 293 "modules/graphics/opengl/Font.cpp"
  HEAP16[$104>>1] = $99; //@line 293 "modules/graphics/opengl/Font.cpp"
  $105 = ((($verts)) + 58|0); //@line 293 "modules/graphics/opengl/Font.cpp"
  HEAP16[$105>>1] = $89; //@line 293 "modules/graphics/opengl/Font.cpp"
  $106 = ((($verts)) + 60|0); //@line 293 "modules/graphics/opengl/Font.cpp"
  HEAP32[$106>>2] = -1; //@line 293 "modules/graphics/opengl/Font.cpp"
  $107 = ((($g)) + 8|0); //@line 299 "modules/graphics/opengl/Font.cpp"
  ;HEAP32[$107>>2]=HEAP32[$verts>>2]|0;HEAP32[$107+4>>2]=HEAP32[$verts+4>>2]|0;HEAP32[$107+8>>2]=HEAP32[$verts+8>>2]|0;HEAP32[$107+12>>2]=HEAP32[$verts+12>>2]|0; //@line 299 "modules/graphics/opengl/Font.cpp"
  $108 = (__ZNK4love4font9GlyphData11getBearingXEv($1)|0); //@line 300 "modules/graphics/opengl/Font.cpp"
  $109 = (+($108|0)); //@line 300 "modules/graphics/opengl/Font.cpp"
  $110 = +HEAPF32[$107>>2]; //@line 300 "modules/graphics/opengl/Font.cpp"
  $111 = $110 + $109; //@line 300 "modules/graphics/opengl/Font.cpp"
  HEAPF32[$107>>2] = $111; //@line 300 "modules/graphics/opengl/Font.cpp"
  $112 = (__ZNK4love4font9GlyphData11getBearingYEv($1)|0); //@line 301 "modules/graphics/opengl/Font.cpp"
  $113 = (+($112|0)); //@line 301 "modules/graphics/opengl/Font.cpp"
  $114 = ((($g)) + 12|0); //@line 301 "modules/graphics/opengl/Font.cpp"
  $115 = +HEAPF32[$114>>2]; //@line 301 "modules/graphics/opengl/Font.cpp"
  $116 = $115 - $113; //@line 301 "modules/graphics/opengl/Font.cpp"
  HEAPF32[$114>>2] = $116; //@line 301 "modules/graphics/opengl/Font.cpp"
  $117 = ((($g)) + 24|0); //@line 299 "modules/graphics/opengl/Font.cpp"
  $118 = ((($verts)) + 16|0); //@line 299 "modules/graphics/opengl/Font.cpp"
  ;HEAP32[$117>>2]=HEAP32[$118>>2]|0;HEAP32[$117+4>>2]=HEAP32[$118+4>>2]|0;HEAP32[$117+8>>2]=HEAP32[$118+8>>2]|0;HEAP32[$117+12>>2]=HEAP32[$118+12>>2]|0; //@line 299 "modules/graphics/opengl/Font.cpp"
  $119 = (__ZNK4love4font9GlyphData11getBearingXEv($1)|0); //@line 300 "modules/graphics/opengl/Font.cpp"
  $120 = (+($119|0)); //@line 300 "modules/graphics/opengl/Font.cpp"
  $121 = +HEAPF32[$117>>2]; //@line 300 "modules/graphics/opengl/Font.cpp"
  $122 = $121 + $120; //@line 300 "modules/graphics/opengl/Font.cpp"
  HEAPF32[$117>>2] = $122; //@line 300 "modules/graphics/opengl/Font.cpp"
  $123 = (__ZNK4love4font9GlyphData11getBearingYEv($1)|0); //@line 301 "modules/graphics/opengl/Font.cpp"
  $124 = (+($123|0)); //@line 301 "modules/graphics/opengl/Font.cpp"
  $125 = ((($g)) + 28|0); //@line 301 "modules/graphics/opengl/Font.cpp"
  $126 = +HEAPF32[$125>>2]; //@line 301 "modules/graphics/opengl/Font.cpp"
  $127 = $126 - $124; //@line 301 "modules/graphics/opengl/Font.cpp"
  HEAPF32[$125>>2] = $127; //@line 301 "modules/graphics/opengl/Font.cpp"
  $128 = ((($g)) + 40|0); //@line 299 "modules/graphics/opengl/Font.cpp"
  $129 = ((($verts)) + 32|0); //@line 299 "modules/graphics/opengl/Font.cpp"
  ;HEAP32[$128>>2]=HEAP32[$129>>2]|0;HEAP32[$128+4>>2]=HEAP32[$129+4>>2]|0;HEAP32[$128+8>>2]=HEAP32[$129+8>>2]|0;HEAP32[$128+12>>2]=HEAP32[$129+12>>2]|0; //@line 299 "modules/graphics/opengl/Font.cpp"
  $130 = (__ZNK4love4font9GlyphData11getBearingXEv($1)|0); //@line 300 "modules/graphics/opengl/Font.cpp"
  $131 = (+($130|0)); //@line 300 "modules/graphics/opengl/Font.cpp"
  $132 = +HEAPF32[$128>>2]; //@line 300 "modules/graphics/opengl/Font.cpp"
  $133 = $132 + $131; //@line 300 "modules/graphics/opengl/Font.cpp"
  HEAPF32[$128>>2] = $133; //@line 300 "modules/graphics/opengl/Font.cpp"
  $134 = (__ZNK4love4font9GlyphData11getBearingYEv($1)|0); //@line 301 "modules/graphics/opengl/Font.cpp"
  $135 = (+($134|0)); //@line 301 "modules/graphics/opengl/Font.cpp"
  $136 = ((($g)) + 44|0); //@line 301 "modules/graphics/opengl/Font.cpp"
  $137 = +HEAPF32[$136>>2]; //@line 301 "modules/graphics/opengl/Font.cpp"
  $138 = $137 - $135; //@line 301 "modules/graphics/opengl/Font.cpp"
  HEAPF32[$136>>2] = $138; //@line 301 "modules/graphics/opengl/Font.cpp"
  $139 = ((($g)) + 56|0); //@line 299 "modules/graphics/opengl/Font.cpp"
  $140 = ((($verts)) + 48|0); //@line 299 "modules/graphics/opengl/Font.cpp"
  ;HEAP32[$139>>2]=HEAP32[$140>>2]|0;HEAP32[$139+4>>2]=HEAP32[$140+4>>2]|0;HEAP32[$139+8>>2]=HEAP32[$140+8>>2]|0;HEAP32[$139+12>>2]=HEAP32[$140+12>>2]|0; //@line 299 "modules/graphics/opengl/Font.cpp"
  $141 = (__ZNK4love4font9GlyphData11getBearingXEv($1)|0); //@line 300 "modules/graphics/opengl/Font.cpp"
  $142 = (+($141|0)); //@line 300 "modules/graphics/opengl/Font.cpp"
  $143 = +HEAPF32[$139>>2]; //@line 300 "modules/graphics/opengl/Font.cpp"
  $144 = $143 + $142; //@line 300 "modules/graphics/opengl/Font.cpp"
  HEAPF32[$139>>2] = $144; //@line 300 "modules/graphics/opengl/Font.cpp"
  $145 = (__ZNK4love4font9GlyphData11getBearingYEv($1)|0); //@line 301 "modules/graphics/opengl/Font.cpp"
  $146 = (+($145|0)); //@line 301 "modules/graphics/opengl/Font.cpp"
  $147 = ((($g)) + 60|0); //@line 301 "modules/graphics/opengl/Font.cpp"
  $148 = +HEAPF32[$147>>2]; //@line 301 "modules/graphics/opengl/Font.cpp"
  $149 = $148 - $146; //@line 301 "modules/graphics/opengl/Font.cpp"
  HEAPF32[$147>>2] = $149; //@line 301 "modules/graphics/opengl/Font.cpp"
 }
 if ($45) {
  $150 = (($5) + 1)|0; //@line 306 "modules/graphics/opengl/Font.cpp"
  $151 = HEAP32[$10>>2]|0; //@line 306 "modules/graphics/opengl/Font.cpp"
  $152 = (($150) + ($151))|0; //@line 306 "modules/graphics/opengl/Font.cpp"
  HEAP32[$10>>2] = $152; //@line 306 "modules/graphics/opengl/Font.cpp"
 }
 if ($46) {
  $153 = ((($this)) + 120|0); //@line 309 "modules/graphics/opengl/Font.cpp"
  $154 = (($9) + 1)|0; //@line 309 "modules/graphics/opengl/Font.cpp"
  $155 = HEAP32[$153>>2]|0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $156 = ($155|0)<($154|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $157 = $156 ? $154 : $155; //@line 309 "modules/graphics/opengl/Font.cpp"
  HEAP32[$153>>2] = $157; //@line 309 "modules/graphics/opengl/Font.cpp"
 }
 $158 = HEAP32[$1>>2]|0; //@line 311 "modules/graphics/opengl/Font.cpp"
 $159 = ((($158)) + 12|0); //@line 311 "modules/graphics/opengl/Font.cpp"
 $160 = HEAP32[$159>>2]|0; //@line 311 "modules/graphics/opengl/Font.cpp"
 FUNCTION_TABLE_vi[$160 & 8191]($1); //@line 311 "modules/graphics/opengl/Font.cpp"
 $161 = ((($this)) + 52|0); //@line 909 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $162 = (__Znwj(84)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $163 = ((($162)) + 8|0); //@line 315 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 HEAP32[$163>>2] = $glyph; //@line 315 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $164 = ((($162)) + 12|0); //@line 316 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 dest=$164; src=$g; stop=dest+72|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 316 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $165 = ((($162)) + 4|0); //@line 2082 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$165>>2] = $glyph; //@line 2082 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$162>>2] = 0; //@line 2083 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 __THREW__ = 0;
 invoke_viii(4019,($0|0),($161|0),($162|0)); //@line 1865 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $166 = __THREW__; __THREW__ = 0;
 $167 = $166&1;
 if ($167) {
  $170 = ___cxa_find_matching_catch_2()|0;
  $171 = tempRet0;
  __ZdlPv($162); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($170|0);
  // unreachable;
 }
 $168 = ((($0)) + 4|0); //@line 1866 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $169 = HEAP8[$168>>0]|0; //@line 316 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $not$$i$i = ($169<<24>>24)==(0); //@line 2629 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($not$$i$i)) {
  $172 = HEAP32[$0>>2]|0; //@line 315 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
  $173 = ((($172)) + 12|0); //@line 315 "modules/graphics/opengl/Font.cpp"
  STACKTOP = sp;return ($173|0); //@line 315 "modules/graphics/opengl/Font.cpp"
 }
 __ZdlPv($162); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $172 = HEAP32[$0>>2]|0; //@line 315 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $173 = ((($172)) + 12|0); //@line 315 "modules/graphics/opengl/Font.cpp"
 STACKTOP = sp;return ($173|0); //@line 315 "modules/graphics/opengl/Font.cpp"
}
function __ZN4love8graphics6opengl4Font22getRasterizerGlyphDataEj($this,$glyph) {
 $this = $this|0;
 $glyph = $glyph|0;
 var $$3 = 0, $$byval_copy = 0, $$lcssa = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$06 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 20|0;
 $0 = sp;
 $1 = ($glyph|0)==(9); //@line 209 "modules/graphics/opengl/Font.cpp"
 if ($1) {
  $2 = ((($this)) + 124|0); //@line 209 "modules/graphics/opengl/Font.cpp"
  $3 = HEAP8[$2>>0]|0; //@line 209 "modules/graphics/opengl/Font.cpp"
  $4 = ($3<<24>>24)==(0); //@line 209 "modules/graphics/opengl/Font.cpp"
  if (!($4)) {
   $5 = ((($this)) + 12|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $6 = HEAP32[$5>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $7 = HEAP32[$6>>2]|0; //@line 117 "./common/Object.h"
   $8 = HEAP32[$7>>2]|0; //@line 211 "modules/graphics/opengl/Font.cpp"
   $9 = ((($8)) + 36|0); //@line 211 "modules/graphics/opengl/Font.cpp"
   $10 = HEAP32[$9>>2]|0; //@line 211 "modules/graphics/opengl/Font.cpp"
   $11 = (FUNCTION_TABLE_iii[$10 & 8191]($7,32)|0); //@line 211 "modules/graphics/opengl/Font.cpp"
   $12 = (__ZNK4love4font9GlyphData9getFormatEv($11)|0); //@line 212 "modules/graphics/opengl/Font.cpp"
   $13 = (__ZNK4love4font9GlyphData10getAdvanceEv($11)|0); //@line 215 "modules/graphics/opengl/Font.cpp"
   $14 = $13 << 2; //@line 215 "modules/graphics/opengl/Font.cpp"
   $15 = (__ZNK4love4font9GlyphData11getBearingXEv($11)|0); //@line 216 "modules/graphics/opengl/Font.cpp"
   $16 = (__ZNK4love4font9GlyphData11getBearingYEv($11)|0); //@line 217 "modules/graphics/opengl/Font.cpp"
   $17 = HEAP32[$11>>2]|0; //@line 219 "modules/graphics/opengl/Font.cpp"
   $18 = ((($17)) + 12|0); //@line 219 "modules/graphics/opengl/Font.cpp"
   $19 = HEAP32[$18>>2]|0; //@line 219 "modules/graphics/opengl/Font.cpp"
   FUNCTION_TABLE_vi[$19 & 8191]($11); //@line 219 "modules/graphics/opengl/Font.cpp"
   $20 = (__Znwj(40)|0); //@line 221 "modules/graphics/opengl/Font.cpp"
   $21 = $0; //@line 221 "modules/graphics/opengl/Font.cpp"
   $22 = $21; //@line 221 "modules/graphics/opengl/Font.cpp"
   HEAP32[$22>>2] = 0; //@line 221 "modules/graphics/opengl/Font.cpp"
   $23 = (($21) + 4)|0; //@line 221 "modules/graphics/opengl/Font.cpp"
   $24 = $23; //@line 221 "modules/graphics/opengl/Font.cpp"
   HEAP32[$24>>2] = 0; //@line 221 "modules/graphics/opengl/Font.cpp"
   $25 = ((($0)) + 8|0); //@line 221 "modules/graphics/opengl/Font.cpp"
   HEAP32[$25>>2] = $14; //@line 221 "modules/graphics/opengl/Font.cpp"
   $26 = ((($0)) + 12|0); //@line 221 "modules/graphics/opengl/Font.cpp"
   HEAP32[$26>>2] = $15; //@line 221 "modules/graphics/opengl/Font.cpp"
   $27 = ((($0)) + 16|0); //@line 221 "modules/graphics/opengl/Font.cpp"
   HEAP32[$27>>2] = $16; //@line 221 "modules/graphics/opengl/Font.cpp"
   __THREW__ = 0;
   ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;HEAP32[$$byval_copy+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$$byval_copy+8>>2]=HEAP32[$0+8>>2]|0;HEAP32[$$byval_copy+12>>2]=HEAP32[$0+12>>2]|0;HEAP32[$$byval_copy+16>>2]=HEAP32[$0+16>>2]|0; //@line 221 "modules/graphics/opengl/Font.cpp"
   invoke_viiii(3981,($20|0),9,($$byval_copy|0),($12|0)); //@line 221 "modules/graphics/opengl/Font.cpp"
   $28 = __THREW__; __THREW__ = 0;
   $29 = $28&1;
   if ($29) {
    $30 = ___cxa_find_matching_catch_2()|0;
    $31 = tempRet0;
    __ZdlPv($20); //@line 221 "modules/graphics/opengl/Font.cpp"
    ___resumeException($30|0);
    // unreachable;
   } else {
    $$3 = $20;
    STACKTOP = sp;return ($$3|0); //@line 231 "modules/graphics/opengl/Font.cpp"
   }
  }
 }
 $32 = ((($this)) + 12|0); //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $33 = HEAP32[$32>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $34 = ((($this)) + 16|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $35 = HEAP32[$34>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $36 = ($33|0)==($35|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 do {
  if ($36) {
   $50 = $33;
  } else {
   $__begin$sroa$0$06 = $33;
   while(1) {
    $39 = HEAP32[$__begin$sroa$0$06>>2]|0; //@line 117 "./common/Object.h"
    $40 = HEAP32[$39>>2]|0; //@line 226 "modules/graphics/opengl/Font.cpp"
    $41 = ((($40)) + 48|0); //@line 226 "modules/graphics/opengl/Font.cpp"
    $42 = HEAP32[$41>>2]|0; //@line 226 "modules/graphics/opengl/Font.cpp"
    $43 = (FUNCTION_TABLE_iii[$42 & 8191]($39,$glyph)|0); //@line 226 "modules/graphics/opengl/Font.cpp"
    $38 = ((($__begin$sroa$0$06)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($43) {
     $$lcssa = $__begin$sroa$0$06;
     break;
    }
    $37 = ($38|0)==($35|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($37) {
     label = 9;
     break;
    } else {
     $__begin$sroa$0$06 = $38;
    }
   }
   if ((label|0) == 9) {
    $$pre = HEAP32[$32>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $50 = $$pre;
    break;
   }
   $44 = HEAP32[$$lcssa>>2]|0; //@line 117 "./common/Object.h"
   $45 = HEAP32[$44>>2]|0; //@line 227 "modules/graphics/opengl/Font.cpp"
   $46 = ((($45)) + 36|0); //@line 227 "modules/graphics/opengl/Font.cpp"
   $47 = HEAP32[$46>>2]|0; //@line 227 "modules/graphics/opengl/Font.cpp"
   $48 = (FUNCTION_TABLE_iii[$47 & 8191]($44,$glyph)|0); //@line 227 "modules/graphics/opengl/Font.cpp"
   $$3 = $48;
   STACKTOP = sp;return ($$3|0); //@line 231 "modules/graphics/opengl/Font.cpp"
  }
 } while(0);
 $49 = HEAP32[$50>>2]|0; //@line 117 "./common/Object.h"
 $51 = HEAP32[$49>>2]|0; //@line 230 "modules/graphics/opengl/Font.cpp"
 $52 = ((($51)) + 36|0); //@line 230 "modules/graphics/opengl/Font.cpp"
 $53 = HEAP32[$52>>2]|0; //@line 230 "modules/graphics/opengl/Font.cpp"
 $54 = (FUNCTION_TABLE_iii[$53 & 8191]($49,$glyph)|0); //@line 230 "modules/graphics/opengl/Font.cpp"
 $$3 = $54;
 STACKTOP = sp;return ($$3|0); //@line 231 "modules/graphics/opengl/Font.cpp"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeIjN4love8graphics6opengl4Font5GlyphEEENS_22__unordered_map_hasherIjS7_NS_4hashIjEELb1EEENS_21__unordered_map_equalIjS7_NS_8equal_toIjEELb1EEENS_9allocatorIS7_EEE20__node_insert_uniqueEPNS_11__hash_nodeIS7_PvEE($agg$result,$this,$__nd) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $__nd = $__nd|0;
 var $$pn = 0, $$pn$us = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $__bc$0 = 0, $__chash$0 = 0, $__chash$1 = 0, $__inserted$0 = 0, $__ndptr$0 = 0, $__ndptr$0$us = 0, $__ndptr$1 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($__nd)) + 8|0); //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $1 = HEAP32[$0>>2]|0; //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $2 = ((($__nd)) + 4|0); //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$2>>2] = $1; //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $3 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $4 = HEAP32[$3>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $5 = ($4|0)==(0); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L1: do {
  if ($5) {
   $__chash$0 = 0;
  } else {
   $6 = (($4) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $7 = $6 & $4; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $8 = ($7|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($8) {
    $9 = $6 & $1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $13 = $9;
   } else {
    $10 = (($1>>>0) % ($4>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $13 = $10;
   }
   $11 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $12 = (($11) + ($13<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $14 = HEAP32[$12>>2]|0; //@line 1603 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $15 = ($14|0)==(0|0); //@line 1604 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($15) {
    $__chash$0 = $13;
   } else {
    if ($8) {
     $$pn$us = $14;
     while(1) {
      $__ndptr$0$us = HEAP32[$$pn$us>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $16 = ($__ndptr$0$us|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($16) {
       $__chash$0 = $13;
       break L1;
      }
      $17 = ((($__ndptr$0$us)) + 4|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $18 = HEAP32[$17>>2]|0; //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $19 = $18 & $6; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $20 = ($19|0)==($13|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($20)) {
       $__chash$0 = $13;
       break L1;
      }
      $21 = ((($__ndptr$0$us)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $22 = HEAP32[$21>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $23 = ($22|0)==($1|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($23) {
       $__inserted$0 = 0;$__ndptr$1 = $__ndptr$0$us;
       break;
      } else {
       $$pn$us = $__ndptr$0$us;
      }
     }
     $85 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     HEAP32[$agg$result>>2] = $85; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     $86 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     HEAP8[$86>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    } else {
     $$pn = $14;
     while(1) {
      $__ndptr$0 = HEAP32[$$pn>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $24 = ($__ndptr$0|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($24) {
       $__chash$0 = $13;
       break L1;
      }
      $25 = ((($__ndptr$0)) + 4|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $26 = HEAP32[$25>>2]|0; //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $27 = (($26>>>0) % ($4>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $28 = ($27|0)==($13|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($28)) {
       $__chash$0 = $13;
       break L1;
      }
      $29 = ((($__ndptr$0)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $30 = HEAP32[$29>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $31 = ($30|0)==($1|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($31) {
       $__inserted$0 = 0;$__ndptr$1 = $__ndptr$0;
       break;
      } else {
       $$pn = $__ndptr$0;
      }
     }
     $85 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     HEAP32[$agg$result>>2] = $85; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     $86 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     HEAP8[$86>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    }
   }
  }
 } while(0);
 $32 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = HEAP32[$32>>2]|0; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $34 = (($33) + 1)|0; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $35 = (+($34>>>0)); //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $36 = (+($4>>>0)); //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $37 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = +HEAPF32[$37>>2]; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $39 = $36 * $38; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $40 = $35 > $39; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $or$cond = $5 | $40; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 do {
  if ($or$cond) {
   $41 = $4 << 1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $42 = ($4>>>0)>(2); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($42) {
    $43 = (($4) + -1)|0; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $44 = $43 & $4; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $45 = ($44|0)==(0); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $47 = $45;
   } else {
    $47 = 0;
   }
   $46 = $47&1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $48 = $46 | $41; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $49 = $48 ^ 1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $50 = $35 / $38; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $51 = (+Math_ceil((+$50))); //@line 750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/cmath"
   $52 = (~~(($51))>>>0); //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $53 = ($49>>>0)<($52>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $54 = $53 ? $52 : $49; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   __ZNSt3__112__hash_tableINS_17__hash_value_typeIjN4love8graphics6opengl4Font5GlyphEEENS_22__unordered_map_hasherIjS7_NS_4hashIjEELb1EEENS_21__unordered_map_equalIjS7_NS_8equal_toIjEELb1EEENS_9allocatorIS7_EEE6rehashEj($this,$54); //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $55 = HEAP32[$3>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $56 = HEAP32[$2>>2]|0; //@line 1621 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $57 = (($55) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $58 = $57 & $55; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $59 = ($58|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($59) {
    $60 = $57 & $56; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__bc$0 = $55;$__chash$1 = $60;
    break;
   } else {
    $61 = (($56>>>0) % ($55>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__bc$0 = $55;$__chash$1 = $61;
    break;
   }
  } else {
   $__bc$0 = $4;$__chash$1 = $__chash$0;
  }
 } while(0);
 $62 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $63 = (($62) + ($__chash$1<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $64 = HEAP32[$63>>2]|0; //@line 1624 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $65 = ($64|0)==(0|0); //@line 1625 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($65) {
  $66 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $67 = HEAP32[$66>>2]|0; //@line 1628 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$__nd>>2] = $67; //@line 1628 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$66>>2] = $__nd; //@line 1629 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $68 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $69 = (($68) + ($__chash$1<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$69>>2] = $66; //@line 1631 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $70 = HEAP32[$__nd>>2]|0; //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $71 = ($70|0)==(0|0); //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if (!($71)) {
   $72 = ((($70)) + 4|0); //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $73 = HEAP32[$72>>2]|0; //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $74 = (($__bc$0) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $75 = $74 & $__bc$0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $76 = ($75|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($76) {
    $77 = $73 & $74; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $81 = $77;
   } else {
    $78 = (($73>>>0) % ($__bc$0>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $81 = $78;
   }
   $79 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $80 = (($79) + ($81<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$80>>2] = $__nd; //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 } else {
  $82 = HEAP32[$64>>2]|0; //@line 1637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$__nd>>2] = $82; //@line 1637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$64>>2] = $__nd; //@line 1638 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $83 = HEAP32[$32>>2]|0; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $84 = (($83) + 1)|0; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$32>>2] = $84; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__inserted$0 = 1;$__ndptr$1 = $__nd;
 $85 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$agg$result>>2] = $85; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $86 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 HEAP8[$86>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeIjN4love8graphics6opengl4Font5GlyphEEENS_22__unordered_map_hasherIjS7_NS_4hashIjEELb1EEENS_21__unordered_map_equalIjS7_NS_8equal_toIjEELb1EEENS_9allocatorIS7_EEE6rehashEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__n|0)==(1); //@line 1937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($0) {
  $$0 = 2;
 } else {
  $1 = (($__n) + -1)|0; //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $2 = $1 & $__n; //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $3 = ($2|0)==(0); //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($3) {
   $$0 = $__n;
  } else {
   $4 = (__ZNSt3__112__next_primeEj($__n)|0); //@line 1940 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $$0 = $4;
  }
 }
 $5 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = HEAP32[$5>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $7 = ($$0>>>0)>($6>>>0); //@line 1942 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($7) {
  __ZNSt3__112__hash_tableINS_17__hash_value_typeIjN4love8graphics6opengl4Font5GlyphEEENS_22__unordered_map_hasherIjS7_NS_4hashIjEELb1EEENS_21__unordered_map_equalIjS7_NS_8equal_toIjEELb1EEENS_9allocatorIS7_EEE8__rehashEj($this,$$0); //@line 1943 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $8 = ($$0>>>0)<($6>>>0); //@line 1944 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($8)) {
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $9 = ($6>>>0)>(2); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($9) {
  $10 = (($6) + -1)|0; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $11 = $10 & $6; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $12 = ($11|0)==(0); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $30 = $12;
 } else {
  $30 = 0;
 }
 $13 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = HEAP32[$13>>2]|0; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $15 = (+($14>>>0)); //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $16 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $17 = +HEAPF32[$16>>2]; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $18 = $15 / $17; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $19 = (+Math_ceil((+$18))); //@line 750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/cmath"
 $20 = (~~(($19))>>>0); //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($30) {
  $21 = (($20) + -1)|0; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $22 = (Math_clz32(($21|0))|0); //@line 785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $23 = (32 - ($22))|0; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $24 = 1 << $23; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $24;
 } else {
  $25 = (__ZNSt3__112__next_primeEj($20)|0); //@line 1950 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $25;
 }
 $26 = ($$0>>>0)<($27>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $28 = $26 ? $27 : $$0; //@line 1946 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $29 = ($28>>>0)<($6>>>0); //@line 1952 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($29)) {
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 __ZNSt3__112__hash_tableINS_17__hash_value_typeIjN4love8graphics6opengl4Font5GlyphEEENS_22__unordered_map_hasherIjS7_NS_4hashIjEELb1EEENS_21__unordered_map_equalIjS7_NS_8equal_toIjEELb1EEENS_9allocatorIS7_EEE8__rehashEj($this,$28); //@line 1953 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeIjN4love8graphics6opengl4Font5GlyphEEENS_22__unordered_map_hasherIjS7_NS_4hashIjEELb1EEENS_21__unordered_map_equalIjS7_NS_8equal_toIjEELb1EEENS_9allocatorIS7_EEE8__rehashEj($this,$__nbc) {
 $this = $this|0;
 $__nbc = $__nbc|0;
 var $$lcssa = 0, $$lcssa133 = 0, $$lcssa143 = 0, $$lcssa6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $8 = 0, $9 = 0;
 var $__cp$0 = 0, $__cp$0$in1019 = 0, $__cp$0$in13 = 0, $__cp$0$in13$us = 0, $__cp$0$lcssa5 = 0, $__cp$0$us$us = 0, $__cp$08 = 0, $__cp$08$us = 0, $__cp$0811 = 0, $__cp$081116 = 0, $__cp$081120 = 0, $__cp$0814 = 0, $__cp$0814$us = 0, $__cp$09 = 0, $__cp$09$lcssa = 0, $__cp$09$us$us = 0, $__cp$09$us$us$lcssa = 0, $__i$021 = 0, $__np$0 = 0, $__np$0$lcssa = 0;
 var $__np$0$us$us = 0, $__np$0$us$us$lcssa = 0, $__phash$0$ph$ph18 = 0, $__pp$0$ph$lcssa = 0, $__pp$0$ph$ph17 = 0, $__pp$0$ph12 = 0, $__pp$0$ph12$us = 0, $exitcond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = ($__nbc|0)==(0); //@line 1965 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($1) {
  $74 = HEAP32[$this>>2]|0; //@line 2822 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$this>>2] = 0; //@line 2823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $75 = ($74|0)==(0|0); //@line 2824 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  if (!($75)) {
   __ZdlPv($74); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  HEAP32[$0>>2] = 0; //@line 1967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $2 = $__nbc << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = (__Znwj($2)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $4 = HEAP32[$this>>2]|0; //@line 2822 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $3; //@line 2823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ($4|0)==(0|0); //@line 2824 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($5)) {
  __ZdlPv($4); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 HEAP32[$0>>2] = $__nbc; //@line 1967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__i$021 = 0;
 while(1) {
  $9 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $10 = (($9) + ($__i$021<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$10>>2] = 0; //@line 1971 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $11 = (($__i$021) + 1)|0; //@line 1970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $exitcond = ($11|0)==($__nbc|0); //@line 1970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($exitcond) {
   break;
  } else {
   $__i$021 = $11;
  }
 }
 $6 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $7 = HEAP32[$6>>2]|0; //@line 1973 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $8 = ($7|0)==(0|0); //@line 1974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($8) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $12 = ((($7)) + 4|0); //@line 1976 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $13 = HEAP32[$12>>2]|0; //@line 1976 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $14 = (($__nbc) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $15 = $14 & $__nbc; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $16 = ($15|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($16) {
  $17 = $13 & $14; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $21 = $17;
 } else {
  $18 = (($13>>>0) % ($__nbc>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $21 = $18;
 }
 $19 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $20 = (($19) + ($21<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$20>>2] = $6; //@line 1977 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__cp$081116 = HEAP32[$7>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $22 = ($__cp$081116|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($22) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 } else {
  $__cp$0$in1019 = $7;$__cp$081120 = $__cp$081116;$__phash$0$ph$ph18 = $21;$__pp$0$ph$ph17 = $7;
 }
 L23: while(1) {
  L25: do {
   if ($16) {
    $__cp$0$in13$us = $__cp$0$in1019;$__cp$0814$us = $__cp$081120;$__pp$0$ph12$us = $__pp$0$ph$ph17;
    while(1) {
     $__cp$09$us$us = $__cp$0814$us;
     while(1) {
      $34 = ((($__cp$09$us$us)) + 4|0); //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $35 = HEAP32[$34>>2]|0; //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $36 = $35 & $14; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $37 = ($36|0)==($__phash$0$ph$ph18|0); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($37) {
       $__cp$09$us$us$lcssa = $__cp$09$us$us;
       break;
      }
      $38 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $39 = (($38) + ($36<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $40 = HEAP32[$39>>2]|0; //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $41 = ($40|0)==(0|0); //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($41) {
       $$lcssa = $39;$$lcssa6 = $36;$__cp$0$lcssa5 = $__cp$09$us$us;$__pp$0$ph$lcssa = $__pp$0$ph12$us;
       break L25;
      }
      $51 = ((($__cp$09$us$us)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $__np$0$us$us = $__cp$09$us$us;
      while(1) {
       $52 = HEAP32[$__np$0$us$us>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $53 = ($52|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($53) {
        $$lcssa143 = $52;$__np$0$us$us$lcssa = $__np$0$us$us;
        break;
       }
       $54 = ((($52)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
       $55 = HEAP32[$51>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $56 = HEAP32[$54>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $57 = ($55|0)==($56|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       if ($57) {
        $__np$0$us$us = $52;
       } else {
        $$lcssa143 = $52;$__np$0$us$us$lcssa = $__np$0$us$us;
        break;
       }
      }
      $42 = $$lcssa143; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__pp$0$ph12$us>>2] = $42; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $43 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $44 = (($43) + ($36<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $45 = HEAP32[$44>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $46 = HEAP32[$45>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__np$0$us$us$lcssa>>2] = $46; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $47 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $48 = (($47) + ($36<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $49 = HEAP32[$48>>2]|0; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$49>>2] = $__cp$09$us$us; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $__cp$0$us$us = HEAP32[$__cp$0$in13$us>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $50 = ($__cp$0$us$us|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($50) {
       label = 32;
       break L23;
      } else {
       $__cp$09$us$us = $__cp$0$us$us;
      }
     }
     $__cp$08$us = HEAP32[$__cp$09$us$us$lcssa>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $58 = ($__cp$08$us|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     if ($58) {
      label = 32;
      break L23;
     } else {
      $__cp$0$in13$us = $__cp$09$us$us$lcssa;$__cp$0814$us = $__cp$08$us;$__pp$0$ph12$us = $__cp$09$us$us$lcssa;
     }
    }
   } else {
    $__cp$0$in13 = $__cp$0$in1019;$__cp$0814 = $__cp$081120;$__pp$0$ph12 = $__pp$0$ph$ph17;
    while(1) {
     $__cp$09 = $__cp$0814;
     while(1) {
      $24 = ((($__cp$09)) + 4|0); //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $25 = HEAP32[$24>>2]|0; //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $26 = (($25>>>0) % ($__nbc>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $27 = ($26|0)==($__phash$0$ph$ph18|0); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($27) {
       $__cp$09$lcssa = $__cp$09;
       break;
      }
      $28 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $29 = (($28) + ($26<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $30 = HEAP32[$29>>2]|0; //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $31 = ($30|0)==(0|0); //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($31) {
       $$lcssa = $29;$$lcssa6 = $26;$__cp$0$lcssa5 = $__cp$09;$__pp$0$ph$lcssa = $__pp$0$ph12;
       break L25;
      }
      $32 = ((($__cp$09)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $__np$0 = $__cp$09;
      while(1) {
       $59 = HEAP32[$__np$0>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $60 = ($59|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($60) {
        $$lcssa133 = $59;$__np$0$lcssa = $__np$0;
        break;
       }
       $61 = ((($59)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
       $62 = HEAP32[$32>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $63 = HEAP32[$61>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $64 = ($62|0)==($63|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       if ($64) {
        $__np$0 = $59;
       } else {
        $$lcssa133 = $59;$__np$0$lcssa = $__np$0;
        break;
       }
      }
      $65 = $$lcssa133; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__pp$0$ph12>>2] = $65; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $66 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $67 = (($66) + ($26<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $68 = HEAP32[$67>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $69 = HEAP32[$68>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__np$0$lcssa>>2] = $69; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $70 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $71 = (($70) + ($26<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $72 = HEAP32[$71>>2]|0; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$72>>2] = $__cp$09; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $__cp$0 = HEAP32[$__cp$0$in13>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $73 = ($__cp$0|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($73) {
       label = 32;
       break L23;
      } else {
       $__cp$09 = $__cp$0;
      }
     }
     $__cp$08 = HEAP32[$__cp$09$lcssa>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $23 = ($__cp$08|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     if ($23) {
      label = 32;
      break L23;
     } else {
      $__cp$0$in13 = $__cp$09$lcssa;$__cp$0814 = $__cp$08;$__pp$0$ph12 = $__cp$09$lcssa;
     }
    }
   }
  } while(0);
  HEAP32[$$lcssa>>2] = $__pp$0$ph$lcssa; //@line 1989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $__cp$0811 = HEAP32[$__cp$0$lcssa5>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $33 = ($__cp$0811|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($33) {
   label = 32;
   break;
  } else {
   $__cp$0$in1019 = $__cp$0$lcssa5;$__cp$081120 = $__cp$0811;$__phash$0$ph$ph18 = $$lcssa6;$__pp$0$ph$ph17 = $__cp$0$lcssa5;
  }
 }
 if ((label|0) == 32) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
}
function __ZThn8_N4love8graphics6opengl4FontD1Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 90 "modules/graphics/opengl/Font.cpp"
 __ZN4love8graphics6opengl4FontD2Ev($0); //@line 90 "modules/graphics/opengl/Font.cpp"
 return; //@line 90 "modules/graphics/opengl/Font.cpp"
}
function __ZThn8_N4love8graphics6opengl4FontD0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 90 "modules/graphics/opengl/Font.cpp"
 __ZN4love8graphics6opengl4FontD2Ev($0); //@line 91 "modules/graphics/opengl/Font.cpp"
 __ZdlPv($0); //@line 91 "modules/graphics/opengl/Font.cpp"
 return; //@line 90 "modules/graphics/opengl/Font.cpp"
}
function __ZThn8_N4love8graphics6opengl4Font12loadVolatileEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 959 "modules/graphics/opengl/Font.cpp"
 __ZN4love8graphics6opengl4Font13createTextureEv($0); //@line 961 "modules/graphics/opengl/Font.cpp"
 $1 = ((($0)) + 132|0); //@line 962 "modules/graphics/opengl/Font.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 962 "modules/graphics/opengl/Font.cpp"
 $3 = (($2) + 1)|0; //@line 962 "modules/graphics/opengl/Font.cpp"
 HEAP32[$1>>2] = $3; //@line 962 "modules/graphics/opengl/Font.cpp"
 return 1; //@line 959 "modules/graphics/opengl/Font.cpp"
}
function __ZThn8_N4love8graphics6opengl4Font14unloadVolatileEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 966 "modules/graphics/opengl/Font.cpp"
 __ZN4love8graphics6opengl4Font14unloadVolatileEv($0); //@line 966 "modules/graphics/opengl/Font.cpp"
 return; //@line 966 "modules/graphics/opengl/Font.cpp"
}
function __GLOBAL__sub_I_Font_cpp() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[(633600)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633612)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633624)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633636)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633648)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633660)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633672)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633684)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(633688)>>2]=0|0;HEAP32[(633688)+4>>2]=0|0;HEAP32[(633688)+8>>2]=0|0;HEAP32[(633688)+12>>2]=0|0; //@line 44 "./common/StringMap.h"
 $0 = HEAP32[6149]|0; //@line 49 "./common/StringMap.h"
 $1 = HEAP32[(24600)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl4Font9AlignModeELj4EE3addEPKcS4_(633592,$0,$1)|0); //@line 49 "./common/StringMap.h"
 $2 = HEAP32[(24604)>>2]|0; //@line 49 "./common/StringMap.h"
 $3 = HEAP32[(24608)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl4Font9AlignModeELj4EE3addEPKcS4_(633592,$2,$3)|0); //@line 49 "./common/StringMap.h"
 $4 = HEAP32[(24612)>>2]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(24616)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl4Font9AlignModeELj4EE3addEPKcS4_(633592,$4,$5)|0); //@line 49 "./common/StringMap.h"
 $6 = HEAP32[(24620)>>2]|0; //@line 49 "./common/StringMap.h"
 $7 = HEAP32[(24624)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl4Font9AlignModeELj4EE3addEPKcS4_(633592,$6,$7)|0); //@line 49 "./common/StringMap.h"
 return;
}
function __ZN4love9StringMapINS_8graphics6opengl4Font9AlignModeELj4EE3addEPKcS4_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0;
 var $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 7; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = $13 & 7; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = $22 & 7; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = $27 & 7; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = $32 & 7; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = $37 & 7; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $42 = (($hash$0$lcssa$i) + 6)|0; //@line 112 "./common/StringMap.h"
       $43 = $42 & 7; //@line 112 "./common/StringMap.h"
       $44 = (((($this) + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
       $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
       $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
       if ($46) {
        $$lcssa = $43;$$lcssa10 = $44;
        label = 5;
       } else {
        $47 = (($hash$0$lcssa$i) + 7)|0; //@line 112 "./common/StringMap.h"
        $48 = $47 & 7; //@line 112 "./common/StringMap.h"
        $49 = (((($this) + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
        $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
        $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
        if ($51) {
         $$lcssa = $48;$$lcssa10 = $49;
         label = 5;
        } else {
         $inserted$2$off0 = 0;
        }
       }
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(3); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 96|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl4FontC2EPNS_4font10RasterizerERKNS0_7Texture6FilterE($this,$r,$filter) {
 $this = $this|0;
 $r = $r|0;
 $filter = $filter|0;
 var $$ = 0, $$0 = 0, $$01$i$i$i = 0, $$01$i$i$i8 = 0, $$02 = 0, $$1 = 0, $$12 = 0, $$13 = 0, $$2 = 0, $$24 = 0, $$3 = 0, $$35 = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i$i = 0, $$pre2$i$i = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0, $126 = 0.0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0.0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $lpad$phi$i14$index = 0;
 var $lpad$phi$i14$index2 = 0, $lpad$phi$index = 0, $lpad$phi$index7 = 0, $nextsize$sroa$0$0 = 0, $nextsize$sroa$7$0 = 0, $not$ = 0, $or$cond$i = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 48 "modules/graphics/opengl/Font.cpp"
 $0 = ((($this)) + 8|0); //@line 59 "modules/graphics/opengl/Font.cpp"
 __THREW__ = 0;
 invoke_vi(4011,($0|0)); //@line 48 "modules/graphics/opengl/Font.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $85 = ___cxa_find_matching_catch_2()|0;
  $86 = tempRet0;
  $$3 = $85;$$35 = $86;
  ___resumeException($$3|0);
  // unreachable;
 }
 HEAP32[$this>>2] = (24548); //@line 59 "modules/graphics/opengl/Font.cpp"
 HEAP32[$0>>2] = (24580); //@line 59 "modules/graphics/opengl/Font.cpp"
 $3 = ((($this)) + 12|0); //@line 49 "modules/graphics/opengl/Font.cpp"
 $4 = ($r|0)==(0|0); //@line 95 "./common/Object.h"
 if ($4) {
  label = 4;
 } else {
  $5 = HEAP32[$r>>2]|0; //@line 95 "./common/Object.h"
  $6 = ((($5)) + 8|0); //@line 95 "./common/Object.h"
  $7 = HEAP32[$6>>2]|0; //@line 95 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($7|0,($r|0)); //@line 95 "./common/Object.h"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if ($9) {
   $87 = ___cxa_find_matching_catch_2()|0;
   $88 = tempRet0;
   $$2 = $87;$$24 = $88;
  } else {
   label = 4;
  }
 }
 L8: do {
  if ((label|0) == 4) {
   HEAP32[$3>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $10 = ((($this)) + 16|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$10>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $11 = ((($this)) + 20|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$11>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __THREW__ = 0;
   $12 = (invoke_ii(3348,4)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $13 = __THREW__; __THREW__ = 0;
   $14 = $13&1;
   L10: do {
    if ($14) {
     $23 = ___cxa_find_matching_catch_2()|0;
     $24 = tempRet0;
     $lpad$phi$i14$index = $23;$lpad$phi$i14$index2 = $24;
     label = 26;
    } else {
     HEAP32[$10>>2] = $12; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$3>>2] = $12; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $15 = ((($12)) + 4|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $16 = ((($this)) + 20|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$16>>2] = $15; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$12>>2] = $r; //@line 99 "./common/Object.h"
     do {
      if ($4) {
       $17 = ((($12)) + 4|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$10>>2] = $17; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      } else {
       $18 = HEAP32[$r>>2]|0; //@line 101 "./common/Object.h"
       $19 = ((($18)) + 8|0); //@line 101 "./common/Object.h"
       $20 = HEAP32[$19>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($20|0,($r|0)); //@line 101 "./common/Object.h"
       $21 = __THREW__; __THREW__ = 0;
       $22 = $21&1;
       if (!($22)) {
        $$pre2$i$i = HEAP32[$10>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $44 = ((($$pre2$i$i)) + 4|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$10>>2] = $44; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $45 = HEAP32[$r>>2]|0; //@line 106 "./common/Object.h"
        $46 = ((($45)) + 12|0); //@line 106 "./common/Object.h"
        $47 = HEAP32[$46>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($47|0,($r|0)); //@line 106 "./common/Object.h"
        $48 = __THREW__; __THREW__ = 0;
        $49 = $48&1;
        if (!($49)) {
         break;
        }
        $50 = ___cxa_find_matching_catch_3(0|0)|0;
        $51 = tempRet0;
        ___clang_call_terminate($50); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       $25 = ___cxa_find_matching_catch_2()|0;
       $26 = tempRet0;
       $$pre = HEAP32[$3>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $27 = ($$pre|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($27) {
        $199 = $26;$200 = $25;
        break L10;
       }
       $28 = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $29 = ($28|0)==($$pre|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       do {
        if ($29) {
         $43 = $$pre;
        } else {
         $31 = $28;
         while(1) {
          $30 = ((($31)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$10>>2] = $30; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $32 = HEAP32[$30>>2]|0; //@line 106 "./common/Object.h"
          $33 = ($32|0)==(0|0); //@line 106 "./common/Object.h"
          if ($33) {
           $35 = $30;
          } else {
           $36 = HEAP32[$32>>2]|0; //@line 106 "./common/Object.h"
           $37 = ((($36)) + 12|0); //@line 106 "./common/Object.h"
           $38 = HEAP32[$37>>2]|0; //@line 106 "./common/Object.h"
           __THREW__ = 0;
           invoke_vi($38|0,($32|0)); //@line 106 "./common/Object.h"
           $39 = __THREW__; __THREW__ = 0;
           $40 = $39&1;
           if ($40) {
            label = 16;
            break;
           }
           $$pre$i$i$i$i = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $35 = $$pre$i$i$i$i;
          }
          $34 = ($35|0)==($$pre|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          if ($34) {
           label = 17;
           break;
          } else {
           $31 = $35;
          }
         }
         if ((label|0) == 16) {
          $41 = ___cxa_find_matching_catch_3(0|0)|0;
          $42 = tempRet0;
          ___clang_call_terminate($41); //@line 106 "./common/Object.h"
          // unreachable; //@line 106 "./common/Object.h"
         }
         else if ((label|0) == 17) {
          $$pre$i$i = HEAP32[$3>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $43 = $$pre$i$i;
          break;
         }
        }
       } while(0);
       __ZdlPv($43); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       $lpad$phi$i14$index = $25;$lpad$phi$i14$index2 = $26;
       label = 26;
       break L10;
      }
     } while(0);
     $52 = ((($this)) + 24|0); //@line 50 "modules/graphics/opengl/Font.cpp"
     $53 = HEAP32[$r>>2]|0; //@line 50 "modules/graphics/opengl/Font.cpp"
     $54 = ((($53)) + 16|0); //@line 50 "modules/graphics/opengl/Font.cpp"
     $55 = HEAP32[$54>>2]|0; //@line 50 "modules/graphics/opengl/Font.cpp"
     __THREW__ = 0;
     $56 = (invoke_ii($55|0,($r|0))|0); //@line 50 "modules/graphics/opengl/Font.cpp"
     $57 = __THREW__; __THREW__ = 0;
     $58 = $57&1;
     if ($58) {
      $96 = ___cxa_find_matching_catch_2()|0;
      $97 = tempRet0;
      $$1 = $96;$$13 = $97;
     } else {
      HEAP32[$52>>2] = $56; //@line 50 "modules/graphics/opengl/Font.cpp"
      $59 = ((($this)) + 28|0); //@line 51 "modules/graphics/opengl/Font.cpp"
      HEAPF32[$59>>2] = 1.0; //@line 51 "modules/graphics/opengl/Font.cpp"
      $60 = ((($this)) + 32|0); //@line 52 "modules/graphics/opengl/Font.cpp"
      HEAP32[$60>>2] = 128; //@line 52 "modules/graphics/opengl/Font.cpp"
      $61 = ((($this)) + 36|0); //@line 53 "modules/graphics/opengl/Font.cpp"
      HEAP32[$61>>2] = 128; //@line 53 "modules/graphics/opengl/Font.cpp"
      $62 = ((($this)) + 40|0); //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $63 = ((($this)) + 44|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $64 = ((($this)) + 68|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      ;HEAP32[$62>>2]=0|0;HEAP32[$62+4>>2]=0|0;HEAP32[$62+8>>2]=0|0;HEAP32[$62+12>>2]=0|0;HEAP32[$62+16>>2]=0|0;HEAP32[$62+20>>2]=0|0;HEAP32[$62+24>>2]=0|0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAPF32[$64>>2] = 1.0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $65 = ((($this)) + 72|0); //@line 48 "modules/graphics/opengl/Font.cpp"
      $66 = ((($this)) + 88|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      ;HEAP32[$65>>2]=0|0;HEAP32[$65+4>>2]=0|0;HEAP32[$65+8>>2]=0|0;HEAP32[$65+12>>2]=0|0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAPF32[$66>>2] = 1.0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $67 = ((($this)) + 96|0); //@line 54 "modules/graphics/opengl/Font.cpp"
      ;HEAP32[$67>>2]=HEAP32[$filter>>2]|0;HEAP32[$67+4>>2]=HEAP32[$filter+4>>2]|0;HEAP32[$67+8>>2]=HEAP32[$filter+8>>2]|0;HEAP32[$67+12>>2]=HEAP32[$filter+12>>2]|0; //@line 54 "modules/graphics/opengl/Font.cpp"
      $68 = ((($this)) + 124|0); //@line 55 "modules/graphics/opengl/Font.cpp"
      HEAP8[$68>>0] = 0; //@line 55 "modules/graphics/opengl/Font.cpp"
      $69 = ((($this)) + 128|0); //@line 56 "modules/graphics/opengl/Font.cpp"
      __THREW__ = 0;
      invoke_vii(4020,($69|0),20); //@line 56 "modules/graphics/opengl/Font.cpp"
      $70 = __THREW__; __THREW__ = 0;
      $71 = $70&1;
      if ($71) {
       $98 = ___cxa_find_matching_catch_2()|0;
       $99 = tempRet0;
       $$0 = $98;$$02 = $99;
      } else {
       $72 = ((($this)) + 132|0); //@line 57 "modules/graphics/opengl/Font.cpp"
       HEAP32[$72>>2] = 0; //@line 57 "modules/graphics/opengl/Font.cpp"
       $73 = ((($this)) + 136|0); //@line 58 "modules/graphics/opengl/Font.cpp"
       HEAP32[$73>>2] = 0; //@line 58 "modules/graphics/opengl/Font.cpp"
       $74 = ((($this)) + 104|0); //@line 60 "modules/graphics/opengl/Font.cpp"
       HEAP32[$74>>2] = 0; //@line 60 "modules/graphics/opengl/Font.cpp"
       $75 = HEAP32[$52>>2]|0; //@line 66 "modules/graphics/opengl/Font.cpp"
       $76 = (+($75|0)); //@line 66 "modules/graphics/opengl/Font.cpp"
       $77 = $76 * 0.80000000000000004; //@line 66 "modules/graphics/opengl/Font.cpp"
       $78 = $76 * $77; //@line 66 "modules/graphics/opengl/Font.cpp"
       $79 = $78 * 30.0; //@line 66 "modules/graphics/opengl/Font.cpp"
       $80 = HEAP32[$60>>2]|0; //@line 66 "modules/graphics/opengl/Font.cpp"
       $81 = HEAP32[$61>>2]|0; //@line 66 "modules/graphics/opengl/Font.cpp"
       $82 = Math_imul($81, $80)|0; //@line 66 "modules/graphics/opengl/Font.cpp"
       $83 = (+($82|0)); //@line 66 "modules/graphics/opengl/Font.cpp"
       $84 = !($79 <= $83); //@line 66 "modules/graphics/opengl/Font.cpp"
       L41: do {
        if ($84) {
         $106 = $80;$109 = $81;
         while(1) {
          __THREW__ = 0;
          $100 = (invoke_ii(4021,(634044|0))|0); //@line 101 "modules/graphics/opengl/Font.cpp"
          $101 = __THREW__; __THREW__ = 0;
          $102 = $101&1;
          if ($102) {
           break;
          }
          $103 = ($100|0)<(4096); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
          $104 = $103 ? $100 : 4096; //@line 101 "modules/graphics/opengl/Font.cpp"
          $105 = $106 << 1; //@line 103 "modules/graphics/opengl/Font.cpp"
          $107 = ($105|0)>($104|0); //@line 103 "modules/graphics/opengl/Font.cpp"
          $108 = $109 << 1; //@line 103 "modules/graphics/opengl/Font.cpp"
          $110 = ($108|0)>($104|0); //@line 103 "modules/graphics/opengl/Font.cpp"
          $or$cond$i = $107 & $110; //@line 103 "modules/graphics/opengl/Font.cpp"
          if ($or$cond$i) {
           $nextsize$sroa$0$0 = $106;$nextsize$sroa$7$0 = $109;
          } else {
           $111 = ($106|0)==($109|0); //@line 106 "modules/graphics/opengl/Font.cpp"
           $$ = $111 ? $109 : $108; //@line 106 "modules/graphics/opengl/Font.cpp"
           $$12 = $111 ? $105 : $106; //@line 106 "modules/graphics/opengl/Font.cpp"
           $nextsize$sroa$0$0 = $$12;$nextsize$sroa$7$0 = $$;
          }
          $112 = HEAP32[$60>>2]|0; //@line 71 "modules/graphics/opengl/Font.cpp"
          $113 = ($nextsize$sroa$0$0|0)>($112|0); //@line 71 "modules/graphics/opengl/Font.cpp"
          if (!($113)) {
           $114 = HEAP32[$61>>2]|0; //@line 71 "modules/graphics/opengl/Font.cpp"
           $115 = ($nextsize$sroa$7$0|0)>($114|0); //@line 71 "modules/graphics/opengl/Font.cpp"
           if (!($115)) {
            label = 40;
            break L41;
           }
          }
          HEAP32[$60>>2] = $nextsize$sroa$0$0; //@line 74 "modules/graphics/opengl/Font.cpp"
          HEAP32[$61>>2] = $nextsize$sroa$7$0; //@line 75 "modules/graphics/opengl/Font.cpp"
          $120 = HEAP32[$52>>2]|0; //@line 66 "modules/graphics/opengl/Font.cpp"
          $121 = (+($120|0)); //@line 66 "modules/graphics/opengl/Font.cpp"
          $122 = $121 * 0.80000000000000004; //@line 66 "modules/graphics/opengl/Font.cpp"
          $123 = $121 * $122; //@line 66 "modules/graphics/opengl/Font.cpp"
          $124 = $123 * 30.0; //@line 66 "modules/graphics/opengl/Font.cpp"
          $125 = Math_imul($nextsize$sroa$7$0, $nextsize$sroa$0$0)|0; //@line 66 "modules/graphics/opengl/Font.cpp"
          $126 = (+($125|0)); //@line 66 "modules/graphics/opengl/Font.cpp"
          $127 = !($124 <= $126); //@line 66 "modules/graphics/opengl/Font.cpp"
          if ($127) {
           $106 = $nextsize$sroa$0$0;$109 = $nextsize$sroa$7$0;
          } else {
           label = 40;
           break L41;
          }
         }
         $116 = ___cxa_find_matching_catch_2()|0;
         $117 = tempRet0;
         $lpad$phi$index = $116;$lpad$phi$index7 = $117;
        } else {
         label = 40;
        }
       } while(0);
       if ((label|0) == 40) {
        $128 = HEAP32[$r>>2]|0; //@line 78 "modules/graphics/opengl/Font.cpp"
        $129 = ((($128)) + 36|0); //@line 78 "modules/graphics/opengl/Font.cpp"
        $130 = HEAP32[$129>>2]|0; //@line 78 "modules/graphics/opengl/Font.cpp"
        __THREW__ = 0;
        $131 = (invoke_iii($130|0,($r|0),32)|0); //@line 78 "modules/graphics/opengl/Font.cpp"
        $132 = __THREW__; __THREW__ = 0;
        $133 = $132&1;
        if (!($133)) {
         __THREW__ = 0;
         $134 = (invoke_ii(4022,($131|0))|0); //@line 79 "modules/graphics/opengl/Font.cpp"
         $135 = __THREW__; __THREW__ = 0;
         $136 = $135&1;
         if (!($136)) {
          $not$ = ($134|0)!=(0); //@line 79 "modules/graphics/opengl/Font.cpp"
          $137 = $not$&1; //@line 79 "modules/graphics/opengl/Font.cpp"
          $138 = ((($this)) + 92|0); //@line 79 "modules/graphics/opengl/Font.cpp"
          HEAP32[$138>>2] = $137; //@line 79 "modules/graphics/opengl/Font.cpp"
          $139 = HEAP32[$131>>2]|0; //@line 80 "modules/graphics/opengl/Font.cpp"
          $140 = ((($139)) + 12|0); //@line 80 "modules/graphics/opengl/Font.cpp"
          $141 = HEAP32[$140>>2]|0; //@line 80 "modules/graphics/opengl/Font.cpp"
          __THREW__ = 0;
          invoke_vi($141|0,($131|0)); //@line 80 "modules/graphics/opengl/Font.cpp"
          $142 = __THREW__; __THREW__ = 0;
          $143 = $142&1;
          if (!($143)) {
           $144 = HEAP32[$r>>2]|0; //@line 82 "modules/graphics/opengl/Font.cpp"
           $145 = ((($144)) + 48|0); //@line 82 "modules/graphics/opengl/Font.cpp"
           $146 = HEAP32[$145>>2]|0; //@line 82 "modules/graphics/opengl/Font.cpp"
           __THREW__ = 0;
           $147 = (invoke_iii($146|0,($r|0),9)|0); //@line 82 "modules/graphics/opengl/Font.cpp"
           $148 = __THREW__; __THREW__ = 0;
           $149 = $148&1;
           if (!($149)) {
            if (!($147)) {
             HEAP8[$68>>0] = 1; //@line 83 "modules/graphics/opengl/Font.cpp"
            }
            $150 = HEAP32[$this>>2]|0; //@line 85 "modules/graphics/opengl/Font.cpp"
            $151 = ((($150)) + 16|0); //@line 85 "modules/graphics/opengl/Font.cpp"
            $152 = HEAP32[$151>>2]|0; //@line 85 "modules/graphics/opengl/Font.cpp"
            __THREW__ = 0;
            (invoke_ii($152|0,($this|0))|0); //@line 85 "modules/graphics/opengl/Font.cpp"
            $153 = __THREW__; __THREW__ = 0;
            $154 = $153&1;
            if (!($154)) {
             $155 = HEAP32[158397]|0; //@line 87 "modules/graphics/opengl/Font.cpp"
             $156 = (($155) + 1)|0; //@line 87 "modules/graphics/opengl/Font.cpp"
             HEAP32[158397] = $156; //@line 87 "modules/graphics/opengl/Font.cpp"
             return; //@line 88 "modules/graphics/opengl/Font.cpp"
            }
           }
          }
         }
        }
        $118 = ___cxa_find_matching_catch_2()|0;
        $119 = tempRet0;
        $lpad$phi$index = $118;$lpad$phi$index7 = $119;
       }
       __ZN4love8graphics6opengl11QuadIndicesD2Ev($69); //@line 88 "modules/graphics/opengl/Font.cpp"
       $$0 = $lpad$phi$index;$$02 = $lpad$phi$index7;
      }
      $157 = ((($this)) + 80|0); //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $158 = HEAP32[$157>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $159 = ($158|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($159)) {
       $$01$i$i$i = $158;
       while(1) {
        $160 = HEAP32[$$01$i$i$i>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        __ZdlPv($$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        $161 = ($160|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        if ($161) {
         break;
        } else {
         $$01$i$i$i = $160;
        }
       }
      }
      $162 = HEAP32[$65>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$65>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $163 = ($162|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      if (!($163)) {
       __ZdlPv($162); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      }
      $164 = ((($this)) + 60|0); //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $165 = HEAP32[$164>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $166 = ($165|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($166)) {
       $$01$i$i$i8 = $165;
       while(1) {
        $167 = HEAP32[$$01$i$i$i8>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        __ZdlPv($$01$i$i$i8); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        $168 = ($167|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        if ($168) {
         break;
        } else {
         $$01$i$i$i8 = $167;
        }
       }
      }
      $169 = ((($this)) + 52|0); //@line 2014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $170 = HEAP32[$169>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$169>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $171 = ($170|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      if (!($171)) {
       __ZdlPv($170); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      }
      $172 = HEAP32[$62>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $173 = ($172|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $174 = $172; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($173) {
       $$1 = $$0;$$13 = $$02;
      } else {
       $175 = HEAP32[$63>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $176 = ($175|0)==($172|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($176)) {
        $scevgep$i$i$i = ((($175)) + -4|0);
        $177 = $scevgep$i$i$i;
        $178 = (($177) - ($174))|0;
        $179 = $178 >>> 2;
        $180 = $179 ^ -1;
        $scevgep4$i$i$i = (($175) + ($180<<2)|0);
        HEAP32[$63>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       __ZdlPv($172); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       $$1 = $$0;$$13 = $$02;
      }
     }
     $181 = HEAP32[$3>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $182 = ($181|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($182) {
      $$2 = $$1;$$24 = $$13;
      break L8;
     }
     $183 = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $184 = ($183|0)==($181|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     do {
      if ($184) {
       $198 = $181;
      } else {
       $186 = $183;
       while(1) {
        $185 = ((($186)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$10>>2] = $185; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $187 = HEAP32[$185>>2]|0; //@line 106 "./common/Object.h"
        $188 = ($187|0)==(0|0); //@line 106 "./common/Object.h"
        if ($188) {
         $190 = $185;
        } else {
         $191 = HEAP32[$187>>2]|0; //@line 106 "./common/Object.h"
         $192 = ((($191)) + 12|0); //@line 106 "./common/Object.h"
         $193 = HEAP32[$192>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($193|0,($187|0)); //@line 106 "./common/Object.h"
         $194 = __THREW__; __THREW__ = 0;
         $195 = $194&1;
         if ($195) {
          label = 66;
          break;
         }
         $$pre$i$i$i = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $190 = $$pre$i$i$i;
        }
        $189 = ($190|0)==($181|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($189) {
         label = 67;
         break;
        } else {
         $186 = $190;
        }
       }
       if ((label|0) == 66) {
        $196 = ___cxa_find_matching_catch_3(0|0)|0;
        $197 = tempRet0;
        ___clang_call_terminate($196); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       else if ((label|0) == 67) {
        $$pre$i = HEAP32[$3>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $198 = $$pre$i;
        break;
       }
      }
     } while(0);
     __ZdlPv($198); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $$2 = $$1;$$24 = $$13;
     break L8;
    }
   } while(0);
   if ((label|0) == 26) {
    if ($4) {
     $$2 = $lpad$phi$i14$index;$$24 = $lpad$phi$i14$index2;
     break;
    } else {
     $199 = $lpad$phi$i14$index2;$200 = $lpad$phi$i14$index;
    }
   }
   $89 = HEAP32[$r>>2]|0; //@line 106 "./common/Object.h"
   $90 = ((($89)) + 12|0); //@line 106 "./common/Object.h"
   $91 = HEAP32[$90>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($91|0,($r|0)); //@line 106 "./common/Object.h"
   $92 = __THREW__; __THREW__ = 0;
   $93 = $92&1;
   if ($93) {
    $94 = ___cxa_find_matching_catch_3(0|0)|0;
    $95 = tempRet0;
    ___clang_call_terminate($94); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   } else {
    $$2 = $200;$$24 = $199;
   }
  }
 } while(0);
 __ZN4love8graphics8VolatileD2Ev($0); //@line 88 "modules/graphics/opengl/Font.cpp"
 $$3 = $$2;$$35 = $$24;
 ___resumeException($$3|0);
 // unreachable;
}
function __ZN4love8graphics6opengl4Font10getKerningEjj($this,$leftglyph,$rightglyph) {
 $this = $this|0;
 $leftglyph = $leftglyph|0;
 $rightglyph = $rightglyph|0;
 var $$0 = 0.0, $$lcssa = 0, $$pn$i$i = 0, $$pn$us$i$i = 0, $$sroa$0$0$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0.0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0.0, $92 = 0, $93 = 0, $94 = 0;
 var $__begin$sroa$0$08 = 0, $__nd$0$i$i = 0, $__nd$0$us$i$i = 0, $k$2 = 0.0, $packedglyphs = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $packedglyphs = sp;
 $0 = $packedglyphs; //@line 330 "modules/graphics/opengl/Font.cpp"
 $1 = $0; //@line 330 "modules/graphics/opengl/Font.cpp"
 HEAP32[$1>>2] = $rightglyph; //@line 330 "modules/graphics/opengl/Font.cpp"
 $2 = (($0) + 4)|0; //@line 330 "modules/graphics/opengl/Font.cpp"
 $3 = $2; //@line 330 "modules/graphics/opengl/Font.cpp"
 HEAP32[$3>>2] = $leftglyph; //@line 330 "modules/graphics/opengl/Font.cpp"
 $4 = ((($this)) + 72|0); //@line 332 "modules/graphics/opengl/Font.cpp"
 $5 = Math_imul($rightglyph, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = $5 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $7 = $6 ^ $5; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = Math_imul($7, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = $8 ^ -561034072; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $10 = Math_imul($leftglyph, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $11 = $10 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $12 = $11 ^ $10; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $13 = Math_imul($12, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = Math_imul($9, 1540483477)|0; //@line 3104 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $15 = $14 ^ $13; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $16 = $15 >>> 13; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $17 = $16 ^ $15; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $18 = Math_imul($17, 1540483477)|0; //@line 3118 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $19 = $18 >>> 15; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $20 = $19 ^ $18; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $21 = ((($this)) + 76|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $22 = HEAP32[$21>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $23 = ($22|0)==(0); //@line 2018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L1: do {
  if (!($23)) {
   $24 = (($22) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $25 = $24 & $22; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $26 = ($25|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($26) {
    $27 = $24 & $20; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $31 = $27;
   } else {
    $28 = (($20>>>0) % ($22>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $31 = $28;
   }
   $29 = HEAP32[$4>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $30 = (($29) + ($31<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $32 = HEAP32[$30>>2]|0; //@line 2021 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $33 = ($32|0)==(0|0); //@line 2022 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($33)) {
    if ($26) {
     $$pn$us$i$i = $32;
     while(1) {
      $__nd$0$us$i$i = HEAP32[$$pn$us$i$i>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $34 = ($__nd$0$us$i$i|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($34) {
       break L1;
      }
      $35 = ((($__nd$0$us$i$i)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $36 = HEAP32[$35>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $37 = $36 & $24; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $38 = ($37|0)==($31|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($38)) {
       break L1;
      }
      $39 = ((($__nd$0$us$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $40 = $39; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $41 = $40; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $42 = HEAP32[$41>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $43 = (($40) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $44 = $43; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $45 = HEAP32[$44>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $46 = ($42|0)==($rightglyph|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $47 = ($45|0)==($leftglyph|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $48 = $46 & $47; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($48) {
       $$sroa$0$0$i$i = $__nd$0$us$i$i;
       break;
      } else {
       $$pn$us$i$i = $__nd$0$us$i$i;
      }
     }
    } else {
     $$pn$i$i = $32;
     while(1) {
      $__nd$0$i$i = HEAP32[$$pn$i$i>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $49 = ($__nd$0$i$i|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($49) {
       break L1;
      }
      $50 = ((($__nd$0$i$i)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $51 = HEAP32[$50>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $52 = (($51>>>0) % ($22>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $53 = ($52|0)==($31|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($53)) {
       break L1;
      }
      $54 = ((($__nd$0$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $55 = $54; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $56 = $55; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $57 = HEAP32[$56>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $58 = (($55) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $59 = $58; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $60 = HEAP32[$59>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $61 = ($57|0)==($rightglyph|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $62 = ($60|0)==($leftglyph|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $63 = $61 & $62; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($63) {
       $$sroa$0$0$i$i = $__nd$0$i$i;
       break;
      } else {
       $$pn$i$i = $__nd$0$i$i;
      }
     }
    }
    $64 = ((($$sroa$0$0$i$i)) + 16|0); //@line 334 "modules/graphics/opengl/Font.cpp"
    $65 = +HEAPF32[$64>>2]; //@line 334 "modules/graphics/opengl/Font.cpp"
    $$0 = $65;
    STACKTOP = sp;return (+$$0); //@line 349 "modules/graphics/opengl/Font.cpp"
   }
  }
 } while(0);
 $66 = ((($this)) + 12|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $67 = HEAP32[$66>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $68 = HEAP32[$67>>2]|0; //@line 117 "./common/Object.h"
 $69 = HEAP32[$68>>2]|0; //@line 336 "modules/graphics/opengl/Font.cpp"
 $70 = ((($69)) + 56|0); //@line 336 "modules/graphics/opengl/Font.cpp"
 $71 = HEAP32[$70>>2]|0; //@line 336 "modules/graphics/opengl/Font.cpp"
 $72 = (+FUNCTION_TABLE_diii[$71 & 8191]($68,$leftglyph,$rightglyph)); //@line 336 "modules/graphics/opengl/Font.cpp"
 $73 = HEAP32[$66>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $74 = ((($this)) + 16|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $75 = HEAP32[$74>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $76 = ($73|0)==($75|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 L20: do {
  if ($76) {
   $k$2 = $72;
  } else {
   $__begin$sroa$0$08 = $73;
   while(1) {
    $77 = HEAP32[$__begin$sroa$0$08>>2]|0; //@line 117 "./common/Object.h"
    $78 = HEAP32[$77>>2]|0; //@line 340 "modules/graphics/opengl/Font.cpp"
    $79 = ((($78)) + 48|0); //@line 340 "modules/graphics/opengl/Font.cpp"
    $80 = HEAP32[$79>>2]|0; //@line 340 "modules/graphics/opengl/Font.cpp"
    $81 = (FUNCTION_TABLE_iii[$80 & 8191]($77,$leftglyph)|0); //@line 340 "modules/graphics/opengl/Font.cpp"
    if ($81) {
     $82 = HEAP32[$__begin$sroa$0$08>>2]|0; //@line 117 "./common/Object.h"
     $83 = HEAP32[$82>>2]|0; //@line 340 "modules/graphics/opengl/Font.cpp"
     $84 = ((($83)) + 48|0); //@line 340 "modules/graphics/opengl/Font.cpp"
     $85 = HEAP32[$84>>2]|0; //@line 340 "modules/graphics/opengl/Font.cpp"
     $86 = (FUNCTION_TABLE_iii[$85 & 8191]($82,$rightglyph)|0); //@line 340 "modules/graphics/opengl/Font.cpp"
     if ($86) {
      $$lcssa = $__begin$sroa$0$08;
      break;
     }
    }
    $92 = ((($__begin$sroa$0$08)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $93 = ($92|0)==($75|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($93) {
     $k$2 = $72;
     break L20;
    } else {
     $__begin$sroa$0$08 = $92;
    }
   }
   $87 = HEAP32[$$lcssa>>2]|0; //@line 117 "./common/Object.h"
   $88 = HEAP32[$87>>2]|0; //@line 342 "modules/graphics/opengl/Font.cpp"
   $89 = ((($88)) + 56|0); //@line 342 "modules/graphics/opengl/Font.cpp"
   $90 = HEAP32[$89>>2]|0; //@line 342 "modules/graphics/opengl/Font.cpp"
   $91 = (+FUNCTION_TABLE_diii[$90 & 8191]($87,$leftglyph,$rightglyph)); //@line 342 "modules/graphics/opengl/Font.cpp"
   $k$2 = $91;
  }
 } while(0);
 $94 = (__ZNSt3__113unordered_mapIyfNS_4hashIyEENS_8equal_toIyEENS_9allocatorINS_4pairIKyfEEEEEixERS7_($4,$packedglyphs)|0); //@line 347 "modules/graphics/opengl/Font.cpp"
 HEAPF32[$94>>2] = $k$2; //@line 347 "modules/graphics/opengl/Font.cpp"
 $$0 = $k$2;
 STACKTOP = sp;return (+$$0); //@line 349 "modules/graphics/opengl/Font.cpp"
}
function __ZNSt3__113unordered_mapIyfNS_4hashIyEENS_8equal_toIyEENS_9allocatorINS_4pairIKyfEEEEEixERS7_($this,$__k) {
 $this = $this|0;
 $__k = $__k|0;
 var $$0 = 0, $$pn$i$i = 0, $$pn$us$i$i = 0, $$sroa$0$0$i$i$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $8 = 0, $9 = 0, $__nd$0$i$i = 0, $__nd$0$us$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = $__k; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = $1; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $4 = (($1) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = $4; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = HEAP32[$5>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $7 = Math_imul($3, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = $7 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = $8 ^ $7; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $10 = Math_imul($9, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $11 = $10 ^ -561034072; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $12 = Math_imul($6, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $13 = $12 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = $13 ^ $12; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $15 = Math_imul($14, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $16 = Math_imul($11, 1540483477)|0; //@line 3104 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $17 = $16 ^ $15; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $18 = $17 >>> 13; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $19 = $18 ^ $17; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $20 = Math_imul($19, 1540483477)|0; //@line 3118 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $21 = $20 >>> 15; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $22 = $21 ^ $20; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $23 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $24 = HEAP32[$23>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $25 = ($24|0)==(0); //@line 2018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L1: do {
  if (!($25)) {
   $26 = (($24) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $27 = $26 & $24; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $28 = ($27|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($28) {
    $29 = $22 & $26; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $33 = $29;
   } else {
    $30 = (($22>>>0) % ($24>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $33 = $30;
   }
   $31 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $32 = (($31) + ($33<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $34 = HEAP32[$32>>2]|0; //@line 2021 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $35 = ($34|0)==(0|0); //@line 2022 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($35)) {
    if ($28) {
     $$pn$us$i$i = $34;
     while(1) {
      $__nd$0$us$i$i = HEAP32[$$pn$us$i$i>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $36 = ($__nd$0$us$i$i|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($36) {
       break L1;
      }
      $37 = ((($__nd$0$us$i$i)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $38 = HEAP32[$37>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $39 = $38 & $26; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $40 = ($39|0)==($33|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($40)) {
       break L1;
      }
      $41 = ((($__nd$0$us$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $42 = $41; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $43 = $42; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $44 = HEAP32[$43>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $45 = (($42) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $46 = $45; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $47 = HEAP32[$46>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $48 = ($44|0)==($3|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $49 = ($47|0)==($6|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $50 = $48 & $49; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($50) {
       $$sroa$0$0$i$i$sink = $__nd$0$us$i$i;
       break;
      } else {
       $$pn$us$i$i = $__nd$0$us$i$i;
      }
     }
     $$0 = ((($$sroa$0$0$i$i$sink)) + 16|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
     STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
    } else {
     $$pn$i$i = $34;
     while(1) {
      $__nd$0$i$i = HEAP32[$$pn$i$i>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $51 = ($__nd$0$i$i|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($51) {
       break L1;
      }
      $52 = ((($__nd$0$i$i)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $53 = HEAP32[$52>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $54 = (($53>>>0) % ($24>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $55 = ($54|0)==($33|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($55)) {
       break L1;
      }
      $56 = ((($__nd$0$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $57 = $56; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $58 = $57; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $59 = HEAP32[$58>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $60 = (($57) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $61 = $60; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $62 = HEAP32[$61>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $63 = ($59|0)==($3|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $64 = ($62|0)==($6|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $65 = $63 & $64; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($65) {
       $$sroa$0$0$i$i$sink = $__nd$0$i$i;
       break;
      } else {
       $$pn$i$i = $__nd$0$i$i;
      }
     }
     $$0 = ((($$sroa$0$0$i$i$sink)) + 16|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
     STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
    }
   }
  }
 } while(0);
 $66 = (__Znwj(24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $67 = ((($66)) + 8|0); //@line 1340 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $68 = $67; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $69 = $68; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$69>>2] = $3; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $70 = (($68) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $71 = $70; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$71>>2] = $6; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $72 = ((($66)) + 16|0); //@line 1342 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 HEAPF32[$72>>2] = 0.0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 invoke_viii(4023,($0|0),($this|0),($66|0)); //@line 1366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $73 = __THREW__; __THREW__ = 0;
 $74 = $73&1;
 if ($74) {
  $76 = ___cxa_find_matching_catch_2()|0;
  $77 = tempRet0;
  __ZdlPv($66); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($76|0);
  // unreachable;
 }
 $75 = HEAP32[$0>>2]|0; //@line 315 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $$sroa$0$0$i$i$sink = $75;
 $$0 = ((($$sroa$0$0$i$i$sink)) + 16|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeIyfEENS_22__unordered_map_hasherIyS2_NS_4hashIyEELb1EEENS_21__unordered_map_equalIyS2_NS_8equal_toIyEELb1EEENS_9allocatorIS2_EEE20__node_insert_uniqueEPNS_11__hash_nodeIS2_PvEE($agg$result,$this,$__nd) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $__nd = $__nd|0;
 var $$pn = 0, $$pn$us = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0, $73 = 0.0, $74 = 0.0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0.0, $86 = 0.0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__bc$0 = 0, $__chash$0 = 0, $__chash$1 = 0, $__inserted$0 = 0, $__ndptr$0 = 0, $__ndptr$0$us = 0, $__ndptr$1 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($__nd)) + 8|0); //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $1 = $0; //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $2 = $1; //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $3 = HEAP32[$2>>2]|0; //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $4 = (($1) + 4)|0; //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $5 = $4; //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $6 = HEAP32[$5>>2]|0; //@line 383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $7 = Math_imul($3, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = $7 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = $8 ^ $7; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $10 = Math_imul($9, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $11 = $10 ^ -561034072; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $12 = Math_imul($6, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $13 = $12 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = $13 ^ $12; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $15 = Math_imul($14, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $16 = Math_imul($11, 1540483477)|0; //@line 3104 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $17 = $16 ^ $15; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $18 = $17 >>> 13; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $19 = $18 ^ $17; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $20 = Math_imul($19, 1540483477)|0; //@line 3118 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $21 = $20 >>> 15; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $22 = $21 ^ $20; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $23 = ((($__nd)) + 4|0); //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$23>>2] = $22; //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $24 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $25 = HEAP32[$24>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $26 = ($25|0)==(0); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L1: do {
  if ($26) {
   $__chash$0 = 0;
  } else {
   $27 = (($25) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $28 = $27 & $25; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $29 = ($28|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($29) {
    $30 = $22 & $27; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $34 = $30;
   } else {
    $31 = (($22>>>0) % ($25>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $34 = $31;
   }
   $32 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $33 = (($32) + ($34<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $35 = HEAP32[$33>>2]|0; //@line 1603 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $36 = ($35|0)==(0|0); //@line 1604 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($36) {
    $__chash$0 = $34;
   } else {
    if ($29) {
     $$pn$us = $35;
     while(1) {
      $__ndptr$0$us = HEAP32[$$pn$us>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $37 = ($__ndptr$0$us|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($37) {
       $__chash$0 = $34;
       break L1;
      }
      $38 = ((($__ndptr$0$us)) + 4|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $39 = HEAP32[$38>>2]|0; //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $40 = $39 & $27; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $41 = ($40|0)==($34|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($41)) {
       $__chash$0 = $34;
       break L1;
      }
      $42 = ((($__ndptr$0$us)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $43 = $42; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $44 = $43; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $45 = HEAP32[$44>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $46 = (($43) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $47 = $46; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $48 = HEAP32[$47>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $49 = ($45|0)==($3|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $50 = ($48|0)==($6|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $51 = $49 & $50; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($51) {
       $__inserted$0 = 0;$__ndptr$1 = $__ndptr$0$us;
       break;
      } else {
       $$pn$us = $__ndptr$0$us;
      }
     }
     $120 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     HEAP32[$agg$result>>2] = $120; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     $121 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     HEAP8[$121>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    } else {
     $$pn = $35;
     while(1) {
      $__ndptr$0 = HEAP32[$$pn>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $52 = ($__ndptr$0|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($52) {
       $__chash$0 = $34;
       break L1;
      }
      $53 = ((($__ndptr$0)) + 4|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $54 = HEAP32[$53>>2]|0; //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $55 = (($54>>>0) % ($25>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $56 = ($55|0)==($34|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if (!($56)) {
       $__chash$0 = $34;
       break L1;
      }
      $57 = ((($__ndptr$0)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $58 = $57; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $59 = $58; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $60 = HEAP32[$59>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $61 = (($58) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $62 = $61; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $63 = HEAP32[$62>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $64 = ($60|0)==($3|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $65 = ($63|0)==($6|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      $66 = $64 & $65; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
      if ($66) {
       $__inserted$0 = 0;$__ndptr$1 = $__ndptr$0;
       break;
      } else {
       $$pn = $__ndptr$0;
      }
     }
     $120 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     HEAP32[$agg$result>>2] = $120; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     $121 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     HEAP8[$121>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    }
   }
  }
 } while(0);
 $67 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $68 = HEAP32[$67>>2]|0; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $69 = (($68) + 1)|0; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $70 = (+($69>>>0)); //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $71 = (+($25>>>0)); //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $72 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $73 = +HEAPF32[$72>>2]; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $74 = $71 * $73; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $75 = $70 > $74; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $or$cond = $26 | $75; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 do {
  if ($or$cond) {
   $76 = $25 << 1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $77 = ($25>>>0)>(2); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($77) {
    $78 = (($25) + -1)|0; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $79 = $78 & $25; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $80 = ($79|0)==(0); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $82 = $80;
   } else {
    $82 = 0;
   }
   $81 = $82&1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $83 = $81 | $76; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $84 = $83 ^ 1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $85 = $70 / $73; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $86 = (+Math_ceil((+$85))); //@line 750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/cmath"
   $87 = (~~(($86))>>>0); //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $88 = ($84>>>0)<($87>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $89 = $88 ? $87 : $84; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   __ZNSt3__112__hash_tableINS_17__hash_value_typeIyfEENS_22__unordered_map_hasherIyS2_NS_4hashIyEELb1EEENS_21__unordered_map_equalIyS2_NS_8equal_toIyEELb1EEENS_9allocatorIS2_EEE6rehashEj($this,$89); //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $90 = HEAP32[$24>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $91 = HEAP32[$23>>2]|0; //@line 1621 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $92 = (($90) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $93 = $92 & $90; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $94 = ($93|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($94) {
    $95 = $92 & $91; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__bc$0 = $90;$__chash$1 = $95;
    break;
   } else {
    $96 = (($91>>>0) % ($90>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__bc$0 = $90;$__chash$1 = $96;
    break;
   }
  } else {
   $__bc$0 = $25;$__chash$1 = $__chash$0;
  }
 } while(0);
 $97 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $98 = (($97) + ($__chash$1<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $99 = HEAP32[$98>>2]|0; //@line 1624 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $100 = ($99|0)==(0|0); //@line 1625 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($100) {
  $101 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $102 = HEAP32[$101>>2]|0; //@line 1628 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$__nd>>2] = $102; //@line 1628 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$101>>2] = $__nd; //@line 1629 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $103 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $104 = (($103) + ($__chash$1<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$104>>2] = $101; //@line 1631 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $105 = HEAP32[$__nd>>2]|0; //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $106 = ($105|0)==(0|0); //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if (!($106)) {
   $107 = ((($105)) + 4|0); //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $108 = HEAP32[$107>>2]|0; //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $109 = (($__bc$0) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $110 = $109 & $__bc$0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $111 = ($110|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($111) {
    $112 = $108 & $109; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $116 = $112;
   } else {
    $113 = (($108>>>0) % ($__bc$0>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $116 = $113;
   }
   $114 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $115 = (($114) + ($116<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$115>>2] = $__nd; //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 } else {
  $117 = HEAP32[$99>>2]|0; //@line 1637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$__nd>>2] = $117; //@line 1637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$99>>2] = $__nd; //@line 1638 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $118 = HEAP32[$67>>2]|0; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $119 = (($118) + 1)|0; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$67>>2] = $119; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__inserted$0 = 1;$__ndptr$1 = $__nd;
 $120 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$agg$result>>2] = $120; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $121 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 HEAP8[$121>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeIyfEENS_22__unordered_map_hasherIyS2_NS_4hashIyEELb1EEENS_21__unordered_map_equalIyS2_NS_8equal_toIyEELb1EEENS_9allocatorIS2_EEE6rehashEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__n|0)==(1); //@line 1937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($0) {
  $$0 = 2;
 } else {
  $1 = (($__n) + -1)|0; //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $2 = $1 & $__n; //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $3 = ($2|0)==(0); //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($3) {
   $$0 = $__n;
  } else {
   $4 = (__ZNSt3__112__next_primeEj($__n)|0); //@line 1940 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $$0 = $4;
  }
 }
 $5 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = HEAP32[$5>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $7 = ($$0>>>0)>($6>>>0); //@line 1942 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($7) {
  __ZNSt3__112__hash_tableINS_17__hash_value_typeIyfEENS_22__unordered_map_hasherIyS2_NS_4hashIyEELb1EEENS_21__unordered_map_equalIyS2_NS_8equal_toIyEELb1EEENS_9allocatorIS2_EEE8__rehashEj($this,$$0); //@line 1943 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $8 = ($$0>>>0)<($6>>>0); //@line 1944 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($8)) {
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $9 = ($6>>>0)>(2); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($9) {
  $10 = (($6) + -1)|0; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $11 = $10 & $6; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $12 = ($11|0)==(0); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $30 = $12;
 } else {
  $30 = 0;
 }
 $13 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = HEAP32[$13>>2]|0; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $15 = (+($14>>>0)); //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $16 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $17 = +HEAPF32[$16>>2]; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $18 = $15 / $17; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $19 = (+Math_ceil((+$18))); //@line 750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/cmath"
 $20 = (~~(($19))>>>0); //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($30) {
  $21 = (($20) + -1)|0; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $22 = (Math_clz32(($21|0))|0); //@line 785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $23 = (32 - ($22))|0; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $24 = 1 << $23; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $24;
 } else {
  $25 = (__ZNSt3__112__next_primeEj($20)|0); //@line 1950 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $25;
 }
 $26 = ($$0>>>0)<($27>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $28 = $26 ? $27 : $$0; //@line 1946 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $29 = ($28>>>0)<($6>>>0); //@line 1952 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($29)) {
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 __ZNSt3__112__hash_tableINS_17__hash_value_typeIyfEENS_22__unordered_map_hasherIyS2_NS_4hashIyEELb1EEENS_21__unordered_map_equalIyS2_NS_8equal_toIyEELb1EEENS_9allocatorIS2_EEE8__rehashEj($this,$28); //@line 1953 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeIyfEENS_22__unordered_map_hasherIyS2_NS_4hashIyEELb1EEENS_21__unordered_map_equalIyS2_NS_8equal_toIyEELb1EEENS_9allocatorIS2_EEE8__rehashEj($this,$__nbc) {
 $this = $this|0;
 $__nbc = $__nbc|0;
 var $$lcssa = 0, $$lcssa133 = 0, $$lcssa143 = 0, $$lcssa6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__cp$0 = 0, $__cp$0$in1019 = 0, $__cp$0$in13 = 0, $__cp$0$in13$us = 0, $__cp$0$lcssa5 = 0, $__cp$0$us$us = 0, $__cp$08 = 0, $__cp$08$us = 0, $__cp$0811 = 0, $__cp$081116 = 0, $__cp$081120 = 0, $__cp$0814 = 0, $__cp$0814$us = 0, $__cp$09 = 0, $__cp$09$lcssa = 0, $__cp$09$us$us = 0;
 var $__cp$09$us$us$lcssa = 0, $__i$021 = 0, $__np$0 = 0, $__np$0$lcssa = 0, $__np$0$us$us = 0, $__np$0$us$us$lcssa = 0, $__phash$0$ph$ph18 = 0, $__pp$0$ph$lcssa = 0, $__pp$0$ph$ph17 = 0, $__pp$0$ph12 = 0, $__pp$0$ph12$us = 0, $exitcond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = ($__nbc|0)==(0); //@line 1965 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($1) {
  $98 = HEAP32[$this>>2]|0; //@line 2822 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$this>>2] = 0; //@line 2823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $99 = ($98|0)==(0|0); //@line 2824 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  if (!($99)) {
   __ZdlPv($98); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  HEAP32[$0>>2] = 0; //@line 1967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $2 = $__nbc << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = (__Znwj($2)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $4 = HEAP32[$this>>2]|0; //@line 2822 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $3; //@line 2823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ($4|0)==(0|0); //@line 2824 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($5)) {
  __ZdlPv($4); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 HEAP32[$0>>2] = $__nbc; //@line 1967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__i$021 = 0;
 while(1) {
  $9 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $10 = (($9) + ($__i$021<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$10>>2] = 0; //@line 1971 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $11 = (($__i$021) + 1)|0; //@line 1970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $exitcond = ($11|0)==($__nbc|0); //@line 1970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($exitcond) {
   break;
  } else {
   $__i$021 = $11;
  }
 }
 $6 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $7 = HEAP32[$6>>2]|0; //@line 1973 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $8 = ($7|0)==(0|0); //@line 1974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($8) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $12 = ((($7)) + 4|0); //@line 1976 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $13 = HEAP32[$12>>2]|0; //@line 1976 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $14 = (($__nbc) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $15 = $14 & $__nbc; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $16 = ($15|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($16) {
  $17 = $13 & $14; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $21 = $17;
 } else {
  $18 = (($13>>>0) % ($__nbc>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $21 = $18;
 }
 $19 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $20 = (($19) + ($21<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$20>>2] = $6; //@line 1977 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__cp$081116 = HEAP32[$7>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $22 = ($__cp$081116|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($22) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 } else {
  $__cp$0$in1019 = $7;$__cp$081120 = $__cp$081116;$__phash$0$ph$ph18 = $21;$__pp$0$ph$ph17 = $7;
 }
 L23: while(1) {
  L25: do {
   if ($16) {
    $__cp$0$in13$us = $__cp$0$in1019;$__cp$0814$us = $__cp$081120;$__pp$0$ph12$us = $__pp$0$ph$ph17;
    while(1) {
     $__cp$09$us$us = $__cp$0814$us;
     while(1) {
      $34 = ((($__cp$09$us$us)) + 4|0); //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $35 = HEAP32[$34>>2]|0; //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $36 = $35 & $14; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $37 = ($36|0)==($__phash$0$ph$ph18|0); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($37) {
       $__cp$09$us$us$lcssa = $__cp$09$us$us;
       break;
      }
      $38 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $39 = (($38) + ($36<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $40 = HEAP32[$39>>2]|0; //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $41 = ($40|0)==(0|0); //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($41) {
       $$lcssa = $39;$$lcssa6 = $36;$__cp$0$lcssa5 = $__cp$09$us$us;$__pp$0$ph$lcssa = $__pp$0$ph12$us;
       break L25;
      }
      $51 = ((($__cp$09$us$us)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $__np$0$us$us = $__cp$09$us$us;
      while(1) {
       $52 = HEAP32[$__np$0$us$us>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $53 = ($52|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($53) {
        $$lcssa143 = $52;$__np$0$us$us$lcssa = $__np$0$us$us;
        break;
       }
       $54 = ((($52)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
       $55 = $51; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $56 = $55; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $57 = HEAP32[$56>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $58 = (($55) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $59 = $58; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $60 = HEAP32[$59>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $61 = $54; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $62 = $61; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $63 = HEAP32[$62>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $64 = (($61) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $65 = $64; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $66 = HEAP32[$65>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $67 = ($57|0)==($63|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $68 = ($60|0)==($66|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $69 = $67 & $68; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       if ($69) {
        $__np$0$us$us = $52;
       } else {
        $$lcssa143 = $52;$__np$0$us$us$lcssa = $__np$0$us$us;
        break;
       }
      }
      $42 = $$lcssa143; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__pp$0$ph12$us>>2] = $42; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $43 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $44 = (($43) + ($36<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $45 = HEAP32[$44>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $46 = HEAP32[$45>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__np$0$us$us$lcssa>>2] = $46; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $47 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $48 = (($47) + ($36<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $49 = HEAP32[$48>>2]|0; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$49>>2] = $__cp$09$us$us; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $__cp$0$us$us = HEAP32[$__cp$0$in13$us>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $50 = ($__cp$0$us$us|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($50) {
       label = 32;
       break L23;
      } else {
       $__cp$09$us$us = $__cp$0$us$us;
      }
     }
     $__cp$08$us = HEAP32[$__cp$09$us$us$lcssa>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $70 = ($__cp$08$us|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     if ($70) {
      label = 32;
      break L23;
     } else {
      $__cp$0$in13$us = $__cp$09$us$us$lcssa;$__cp$0814$us = $__cp$08$us;$__pp$0$ph12$us = $__cp$09$us$us$lcssa;
     }
    }
   } else {
    $__cp$0$in13 = $__cp$0$in1019;$__cp$0814 = $__cp$081120;$__pp$0$ph12 = $__pp$0$ph$ph17;
    while(1) {
     $__cp$09 = $__cp$0814;
     while(1) {
      $24 = ((($__cp$09)) + 4|0); //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $25 = HEAP32[$24>>2]|0; //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $26 = (($25>>>0) % ($__nbc>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $27 = ($26|0)==($__phash$0$ph$ph18|0); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($27) {
       $__cp$09$lcssa = $__cp$09;
       break;
      }
      $28 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $29 = (($28) + ($26<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $30 = HEAP32[$29>>2]|0; //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $31 = ($30|0)==(0|0); //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($31) {
       $$lcssa = $29;$$lcssa6 = $26;$__cp$0$lcssa5 = $__cp$09;$__pp$0$ph$lcssa = $__pp$0$ph12;
       break L25;
      }
      $32 = ((($__cp$09)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
      $__np$0 = $__cp$09;
      while(1) {
       $71 = HEAP32[$__np$0>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $72 = ($71|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($72) {
        $$lcssa133 = $71;$__np$0$lcssa = $__np$0;
        break;
       }
       $73 = ((($71)) + 8|0); //@line 434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
       $74 = $32; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $75 = $74; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $76 = HEAP32[$75>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $77 = (($74) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $78 = $77; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $79 = HEAP32[$78>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $80 = $73; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $81 = $80; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $82 = HEAP32[$81>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $83 = (($80) + 4)|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $84 = $83; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $85 = HEAP32[$84>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $86 = ($76|0)==($82|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $87 = ($79|0)==($85|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       $88 = $86 & $87; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
       if ($88) {
        $__np$0 = $71;
       } else {
        $$lcssa133 = $71;$__np$0$lcssa = $__np$0;
        break;
       }
      }
      $89 = $$lcssa133; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__pp$0$ph12>>2] = $89; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $90 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $91 = (($90) + ($26<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $92 = HEAP32[$91>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $93 = HEAP32[$92>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$__np$0$lcssa>>2] = $93; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $94 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $95 = (($94) + ($26<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $96 = HEAP32[$95>>2]|0; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      HEAP32[$96>>2] = $__cp$09; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $__cp$0 = HEAP32[$__cp$0$in13>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $97 = ($__cp$0|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      if ($97) {
       label = 32;
       break L23;
      } else {
       $__cp$09 = $__cp$0;
      }
     }
     $__cp$08 = HEAP32[$__cp$09$lcssa>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $23 = ($__cp$08|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     if ($23) {
      label = 32;
      break L23;
     } else {
      $__cp$0$in13 = $__cp$09$lcssa;$__cp$0814 = $__cp$08;$__pp$0$ph12 = $__cp$09$lcssa;
     }
    }
   }
  } while(0);
  HEAP32[$$lcssa>>2] = $__pp$0$ph$lcssa; //@line 1989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $__cp$0811 = HEAP32[$__cp$0$lcssa5>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $33 = ($__cp$0811|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($33) {
   label = 32;
   break;
  } else {
   $__cp$0$in1019 = $__cp$0$lcssa5;$__cp$081120 = $__cp$0811;$__phash$0$ph$ph18 = $$lcssa6;$__pp$0$ph$ph17 = $__cp$0$lcssa5;
  }
 }
 if ((label|0) == 32) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
}
function __ZN4love8graphics6opengl4Font23getCodepointsFromStringERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEERNS3_6vectorIjNS7_IjEEEE($text,$codepoints) {
 $text = $text|0;
 $codepoints = $codepoints|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$byval_copy = 0, $$byval_copy10 = 0, $$pre = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index4Z2D = 0, $eh$lpad$body$indexZ2D = 0, $g = 0;
 var $i = 0, $or$cond26 = 0, $temp$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy10 = sp + 36|0;
 $$byval_copy = sp + 32|0;
 $vararg_buffer = sp;
 $temp$i = sp + 28|0;
 $0 = sp + 24|0;
 $1 = sp + 20|0;
 $i = sp + 8|0;
 $g = sp + 4|0;
 $2 = HEAP8[$text>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = $2 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ($3<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($text)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = HEAP32[$5>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = $2&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $7 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $4 ? $8 : $6; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = ((($codepoints)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $11 = HEAP32[$10>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = HEAP32[$codepoints>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = (($11) - ($12))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = $13 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = ($14>>>0)<($9>>>0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($15) {
  $16 = ((($codepoints)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = HEAP32[$16>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = (($17) - ($12))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = $18 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = ($9|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($20) {
   $24 = 0;
  } else {
   $21 = $9 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $22 = (__Znwj($21)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $24 = $22;
  }
  $23 = (($24) + ($19<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $25 = $23; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $26 = (($24) + ($9<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $27 = $26; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $28 = (0 - ($19))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $29 = (($23) + ($28<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $30 = $29; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $31 = $12; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  _memcpy(($29|0),($31|0),($18|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$codepoints>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$16>>2] = $25; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$10>>2] = $27; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $32 = ($12|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($32) {
   $44 = $6;
  } else {
   __ZdlPv($31); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $$pre = HEAP32[$5>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $44 = $$pre;
  }
 } else {
  $44 = $6;
 }
 $33 = HEAP8[$text>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $34 = $33 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $35 = ($34<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $36 = ((($text)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $37 = HEAP32[$36>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $38 = ((($text)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $39 = $35 ? $38 : $37; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $40 = $39; //@line 357 "modules/graphics/opengl/Font.cpp"
 $41 = $33&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $42 = $41 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $43 = $35 ? $42 : $44; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $45 = (($39) + ($43)|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $46 = $45; //@line 357 "modules/graphics/opengl/Font.cpp"
 HEAP32[$i>>2] = $40; //@line 277 "./libraries/utf8/utf8/checked.h"
 $47 = ((($i)) + 4|0); //@line 277 "./libraries/utf8/utf8/checked.h"
 HEAP32[$47>>2] = $40; //@line 277 "./libraries/utf8/utf8/checked.h"
 $48 = ((($i)) + 8|0); //@line 277 "./libraries/utf8/utf8/checked.h"
 HEAP32[$48>>2] = $46; //@line 277 "./libraries/utf8/utf8/checked.h"
 $49 = ($43|0)<(0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 do {
  if ($49) {
   $50 = (___cxa_allocate_exception(8)|0); //@line 280 "./libraries/utf8/utf8/checked.h"
   __THREW__ = 0;
   invoke_vii(3996,($50|0),(434517|0)); //@line 133 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/stdexcept"
   $51 = __THREW__; __THREW__ = 0;
   $52 = $51&1;
   if ($52) {
    $54 = ___cxa_find_matching_catch_3(1744|0)|0;
    $55 = tempRet0;
    ___cxa_free_exception(($50|0)); //@line 280 "./libraries/utf8/utf8/checked.h"
    $eh$lpad$body$index4Z2D = $55;$eh$lpad$body$indexZ2D = $54;
    break;
   } else {
    HEAP32[$50>>2] = (157912); //@line 133 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/stdexcept"
    __THREW__ = 0;
    invoke_viii(3350,($50|0),(9120|0),(3328|0)); //@line 280 "./libraries/utf8/utf8/checked.h"
    $53 = __THREW__; __THREW__ = 0;
    label = 22;
    break;
   }
  } else {
   $56 = ((($i)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $57 = ((($i)) + 8|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $58 = ((($codepoints)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $59 = ($43|0)==(0); //@line 297 "./libraries/utf8/utf8/checked.h"
   if ($59) {
    STACKTOP = sp;return; //@line 370 "modules/graphics/opengl/Font.cpp"
   } else {
    $66 = $46;$69 = $40;
   }
   while(1) {
    HEAP32[$1>>2] = $66; //@line 307 "./libraries/utf8/utf8/checked.h"
    __THREW__ = 0;
    ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0; //@line 307 "./libraries/utf8/utf8/checked.h"
    (invoke_iii(3997,($i|0),($$byval_copy|0))|0); //@line 307 "./libraries/utf8/utf8/checked.h"
    $67 = __THREW__; __THREW__ = 0;
    $68 = $67&1;
    if ($68) {
     label = 21;
     break;
    }
    HEAP32[$temp$i>>2] = $69; //@line 286 "./libraries/utf8/utf8/checked.h"
    HEAP32[$0>>2] = $66; //@line 287 "./libraries/utf8/utf8/checked.h"
    __THREW__ = 0;
    ;HEAP32[$$byval_copy10>>2]=HEAP32[$0>>2]|0; //@line 287 "./libraries/utf8/utf8/checked.h"
    $70 = (invoke_iii(3997,($temp$i|0),($$byval_copy10|0))|0); //@line 287 "./libraries/utf8/utf8/checked.h"
    $71 = __THREW__; __THREW__ = 0;
    $72 = $71&1;
    if ($72) {
     label = 21;
     break;
    }
    HEAP32[$g>>2] = $70; //@line 362 "modules/graphics/opengl/Font.cpp"
    $73 = HEAP32[$58>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $74 = HEAP32[$10>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $75 = ($73|0)==($74|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($75) {
     __THREW__ = 0;
     invoke_vii(4016,($codepoints|0),($g|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $77 = __THREW__; __THREW__ = 0;
     $78 = $77&1;
     if ($78) {
      label = 21;
      break;
     }
    } else {
     HEAP32[$73>>2] = $70; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $76 = ((($73)) + 4|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$58>>2] = $76; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    $79 = HEAP32[$56>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $80 = ($79|0)==($39|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $81 = HEAP32[$57>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $82 = ($81|0)==($45|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $or$cond26 = $80 & $82; //@line 291 "./libraries/utf8/utf8/checked.h"
    if (!($or$cond26)) {
     label = 11;
     break;
    }
    $83 = $81; //@line 291 "./libraries/utf8/utf8/checked.h"
    $$pre28 = HEAP32[$i>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $84 = $$pre28; //@line 291 "./libraries/utf8/utf8/checked.h"
    $85 = ($$pre28|0)==($45|0); //@line 297 "./libraries/utf8/utf8/checked.h"
    if ($85) {
     label = 26;
     break;
    } else {
     $66 = $83;$69 = $84;
    }
   }
   if ((label|0) == 11) {
    $60 = (___cxa_allocate_exception(8)|0); //@line 292 "./libraries/utf8/utf8/checked.h"
    __THREW__ = 0;
    invoke_vii(3996,($60|0),(434549|0)); //@line 292 "./libraries/utf8/utf8/checked.h"
    $61 = __THREW__; __THREW__ = 0;
    $62 = $61&1;
    if ($62) {
     $64 = ___cxa_find_matching_catch_3(1744|0)|0;
     $65 = tempRet0;
     ___cxa_free_exception(($60|0)); //@line 292 "./libraries/utf8/utf8/checked.h"
     $eh$lpad$body$index4Z2D = $65;$eh$lpad$body$indexZ2D = $64;
     break;
    } else {
     __THREW__ = 0;
     invoke_viii(3350,($60|0),(9072|0),(3328|0)); //@line 292 "./libraries/utf8/utf8/checked.h"
     $63 = __THREW__; __THREW__ = 0;
     label = 22;
     break;
    }
   }
   else if ((label|0) == 21) {
    $86 = ___cxa_find_matching_catch_3(1744|0)|0;
    $87 = tempRet0;
    $eh$lpad$body$index4Z2D = $87;$eh$lpad$body$indexZ2D = $86;
    break;
   }
   else if ((label|0) == 26) {
    STACKTOP = sp;return; //@line 370 "modules/graphics/opengl/Font.cpp"
   }
  }
 } while(0);
 if ((label|0) == 22) {
  $88 = ___cxa_find_matching_catch_3(1744|0)|0;
  $89 = tempRet0;
  $eh$lpad$body$index4Z2D = $89;$eh$lpad$body$indexZ2D = $88;
 }
 $90 = (_llvm_eh_typeid_for((1744|0))|0); //@line 365 "modules/graphics/opengl/Font.cpp"
 $91 = ($eh$lpad$body$index4Z2D|0)==($90|0); //@line 365 "modules/graphics/opengl/Font.cpp"
 if (!($91)) {
  $$1 = $eh$lpad$body$indexZ2D;$$12 = $eh$lpad$body$index4Z2D;
  ___resumeException($$1|0);
  // unreachable;
 }
 $92 = (___cxa_begin_catch(($eh$lpad$body$indexZ2D|0))|0); //@line 365 "modules/graphics/opengl/Font.cpp"
 $93 = (___cxa_allocate_exception(16)|0); //@line 368 "modules/graphics/opengl/Font.cpp"
 $94 = HEAP32[$92>>2]|0; //@line 368 "modules/graphics/opengl/Font.cpp"
 $95 = ((($94)) + 8|0); //@line 368 "modules/graphics/opengl/Font.cpp"
 $96 = HEAP32[$95>>2]|0; //@line 368 "modules/graphics/opengl/Font.cpp"
 $97 = (FUNCTION_TABLE_ii[$96 & 8191]($92)|0); //@line 368 "modules/graphics/opengl/Font.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $97; //@line 368 "modules/graphics/opengl/Font.cpp"
 invoke_viii(3349,($93|0),(434605|0),($vararg_buffer|0)); //@line 368 "modules/graphics/opengl/Font.cpp"
 $98 = __THREW__; __THREW__ = 0;
 $99 = $98&1;
 if ($99) {
  $103 = ___cxa_find_matching_catch_2()|0;
  $104 = tempRet0;
  ___cxa_free_exception(($93|0)); //@line 368 "modules/graphics/opengl/Font.cpp"
  $$0 = $103;$$01 = $104;
 } else {
  __THREW__ = 0;
  invoke_viii(3350,($93|0),(96|0),(344|0)); //@line 368 "modules/graphics/opengl/Font.cpp"
  $100 = __THREW__; __THREW__ = 0;
  $101 = ___cxa_find_matching_catch_2()|0;
  $102 = tempRet0;
  $$0 = $101;$$01 = $102;
 }
 __THREW__ = 0;
 invoke_v(3378); //@line 369 "modules/graphics/opengl/Font.cpp"
 $105 = __THREW__; __THREW__ = 0;
 $106 = $105&1;
 if ($106) {
  $107 = ___cxa_find_matching_catch_3(0|0)|0;
  $108 = tempRet0;
  ___clang_call_terminate($107); //@line 369 "modules/graphics/opengl/Font.cpp"
  // unreachable; //@line 369 "modules/graphics/opengl/Font.cpp"
 } else {
  $$1 = $$0;$$12 = $$01;
  ___resumeException($$1|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl4Font23getCodepointsFromStringERKNSt3__16vectorINS2_13ColoredStringENS3_9allocatorIS5_EEEERNS2_17ColoredCodepointsE($strs,$codepoints) {
 $strs = $strs|0;
 $codepoints = $codepoints|0;
 var $$pre = 0, $$pre$phi6Z2D = 0, $$pre$phiZ2D = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $9 = 0, $__begin$sroa$0$04 = 0, $c = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $c = sp;
 $0 = HEAP32[$strs>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($strs)) + 4|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ($0|0)==($2|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($3) {
  STACKTOP = sp;return; //@line 394 "modules/graphics/opengl/Font.cpp"
 }
 $4 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = $4 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = ($5<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($6) {
  $9 = $4&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $10 = $9 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $17 = $10;
 } else {
  $7 = ((($0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $8 = HEAP32[$7>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $17 = $8;
 }
 $11 = ((($codepoints)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $12 = HEAP32[$11>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = HEAP32[$codepoints>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = (($12) - ($13))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = $14 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = ($15>>>0)<($17>>>0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($16) {
  $18 = ((($codepoints)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = HEAP32[$18>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = (($19) - ($13))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = $20 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($22) {
   $26 = 0;
  } else {
   $23 = $17 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $24 = (__Znwj($23)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $26 = $24;
  }
  $25 = (($26) + ($21<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $27 = $25; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $28 = (($26) + ($17<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $29 = $28; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $30 = (0 - ($21))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $31 = (($25) + ($30<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $32 = $31; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $33 = $13; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  _memcpy(($31|0),($33|0),($20|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$codepoints>>2] = $32; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$18>>2] = $27; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$11>>2] = $29; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $34 = ($13|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if (!($34)) {
   __ZdlPv($33); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
 }
 $35 = HEAP32[$strs>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $36 = HEAP32[$1>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $37 = ($35|0)==($36|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($37) {
  $$pre = ((($codepoints)) + 12|0); //@line 387 "modules/graphics/opengl/Font.cpp"
  $$pre5 = ((($codepoints)) + 16|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre$phi6Z2D = $$pre5;$$pre$phiZ2D = $$pre;
 } else {
  $38 = ((($c)) + 4|0); //@line 381 "modules/graphics/opengl/Font.cpp"
  $39 = ((($codepoints)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $40 = ((($codepoints)) + 16|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $41 = ((($codepoints)) + 20|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $42 = ((($codepoints)) + 12|0); //@line 382 "modules/graphics/opengl/Font.cpp"
  $__begin$sroa$0$04 = $35;
  while(1) {
   $48 = ((($__begin$sroa$0$04)) + 12|0); //@line 381 "modules/graphics/opengl/Font.cpp"
   $49 = HEAP32[$48>>2]|0; //@line 381 "modules/graphics/opengl/Font.cpp"
   HEAP32[$c>>2] = $49; //@line 381 "modules/graphics/opengl/Font.cpp"
   $50 = HEAP32[$39>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $51 = HEAP32[$codepoints>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $52 = (($50) - ($51))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $53 = $52 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$38>>2] = $53; //@line 381 "modules/graphics/opengl/Font.cpp"
   $54 = HEAP32[$40>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $55 = HEAP32[$41>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $56 = ($54|0)==($55|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($56) {
    __ZNSt3__16vectorIN4love8graphics6opengl4Font12IndexedColorENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_($42,$c); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   } else {
    $57 = $c; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $58 = $57; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $59 = HEAP32[$58>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $60 = (($57) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $61 = $60; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $62 = HEAP32[$61>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $63 = $54; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $64 = $63; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$64>>2] = $59; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $65 = (($63) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $66 = $65; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$66>>2] = $62; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $67 = HEAP32[$40>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $68 = ((($67)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$40>>2] = $68; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZN4love8graphics6opengl4Font23getCodepointsFromStringERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEERNS3_6vectorIjNS7_IjEEEE($__begin$sroa$0$04,$codepoints); //@line 384 "modules/graphics/opengl/Font.cpp"
   $69 = ((($__begin$sroa$0$04)) + 16|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $70 = ($69|0)==($36|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($70) {
    $$pre$phi6Z2D = $40;$$pre$phiZ2D = $42;
    break;
   } else {
    $__begin$sroa$0$04 = $69;
   }
  }
 }
 $43 = HEAP32[$$pre$phi6Z2D>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $44 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $45 = (($43) - ($44))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $46 = ($45|0)==(8); //@line 387 "modules/graphics/opengl/Font.cpp"
 $47 = $43; //@line 1666 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($46)) {
  STACKTOP = sp;return; //@line 394 "modules/graphics/opengl/Font.cpp"
 }
 $71 = $44; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $72 = $71; //@line 389 "modules/graphics/opengl/Font.cpp"
 $73 = $72; //@line 389 "modules/graphics/opengl/Font.cpp"
 $74 = HEAP32[$73>>2]|0; //@line 389 "modules/graphics/opengl/Font.cpp"
 $75 = (($72) + 4)|0; //@line 389 "modules/graphics/opengl/Font.cpp"
 $76 = $75; //@line 389 "modules/graphics/opengl/Font.cpp"
 $77 = HEAP32[$76>>2]|0; //@line 389 "modules/graphics/opengl/Font.cpp"
 $78 = ($74|0)==(-1); //@line 391 "modules/graphics/opengl/Font.cpp"
 $79 = ($77|0)==(0); //@line 391 "modules/graphics/opengl/Font.cpp"
 $80 = $78 & $79; //@line 391 "modules/graphics/opengl/Font.cpp"
 if (!($80)) {
  STACKTOP = sp;return; //@line 394 "modules/graphics/opengl/Font.cpp"
 }
 $scevgep4$i$i$i = ((($47)) + -8|0);
 HEAP32[$$pre$phi6Z2D>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 STACKTOP = sp;return; //@line 394 "modules/graphics/opengl/Font.cpp"
}
function __ZNSt3__16vectorIN4love8graphics6opengl4Font12IndexedColorENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(536870911); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = $$in; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $8 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = HEAP32[$8>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $10 >> 3; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($11>>>0)<(268435455); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $16 = $10 >> 2; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = ($16>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $18 = $17 ? $5 : $16; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = (($19) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = $20 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = ($18|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($22) {
   $$0$i2 = 0;$26 = 0;$27 = $21;$43 = $19;
  } else {
   $$0$i1 = $18;$49 = $19;$50 = $21;
   label = 6;
  }
 } else {
  $13 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = $14 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 536870911;$49 = $13;$50 = $15;
  label = 6;
 }
 if ((label|0) == 6) {
  $23 = $$0$i1 << 3; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = (__Znwj($23)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$26 = $24;$27 = $50;$43 = $49;
 }
 $25 = (($26) + ($27<<3)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = (($26) + ($$0$i2<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = $28; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = $__x; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $31 = $30; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $32 = HEAP32[$31>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = (($30) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $33; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = HEAP32[$34>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = $25; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$37>>2] = $32; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = (($36) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $39 = $38; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$39>>2] = $35; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $40 = ((($25)) + 8|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $41 = $40; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $42 = (($43) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $44 = $42 >> 3; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $45 = (0 - ($44))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $46 = (($25) + ($45<<3)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = $46; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($46|0),($7|0),($42|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $47; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $41; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$8>>2] = $29; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $48 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($48) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($7); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNK4love8graphics6opengl4Font9getHeightEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 398 "modules/graphics/opengl/Font.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 398 "modules/graphics/opengl/Font.cpp"
 $2 = (+($1|0)); //@line 398 "modules/graphics/opengl/Font.cpp"
 return (+$2); //@line 398 "modules/graphics/opengl/Font.cpp"
}
function __ZN4love8graphics6opengl4Font16generateVerticesERKNS2_17ColoredCodepointsERNSt3__16vectorINS2_11GlyphVertexENS6_9allocatorIS8_EEEEfNS_6VectorEPNS2_8TextInfoE($agg$result,$this,$codepoints,$vertices,$extra_spacing,$offset,$info) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $codepoints = $codepoints|0;
 $vertices = $vertices|0;
 $extra_spacing = +$extra_spacing;
 $offset = $offset|0;
 $info = $info|0;
 var $$0$i = 0, $$in = 0, $$pn$i$i$i = 0, $$pn$us$i$i$i = 0, $$pre = 0, $$pre48 = 0, $$pre49 = 0, $$pre50 = 0, $$pre51 = 0, $$pre52 = 0, $$pre53 = 0, $$pre53$pre = 0, $$sroa$0$0$i$i$i = 0, $0 = 0.0, $1 = 0, $10 = 0.0, $100 = 0, $101 = 0.0, $102 = 0.0, $103 = 0.0;
 var $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0, $108 = 0, $109 = 0, $11 = 0.0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0;
 var $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0.0, $167 = 0, $168 = 0, $169 = 0.0, $17 = 0, $170 = 0, $171 = 0, $172 = 0.0, $173 = 0.0, $174 = 0, $175 = 0, $176 = 0;
 var $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0.0;
 var $195 = 0.0, $196 = 0, $197 = 0.0, $198 = 0.0, $199 = 0, $2 = 0.0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0;
 var $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0.0, $228 = 0.0, $229 = 0, $23 = 0;
 var $230 = 0.0, $231 = 0.0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0.0, $242 = 0, $243 = 0, $244 = 0.0, $245 = 0.0, $246 = 0, $247 = 0, $248 = 0.0;
 var $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0.0, $253 = 0, $254 = 0.0, $255 = 0.0, $256 = 0.0, $257 = 0.0, $258 = 0.0, $259 = 0.0, $26 = 0, $260 = 0.0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0;
 var $267 = 0, $268 = 0, $269 = 0.0, $27 = 0, $270 = 0.0, $271 = 0, $272 = 0.0, $273 = 0.0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0;
 var $285 = 0.0, $286 = 0.0, $287 = 0, $288 = 0.0, $289 = 0.0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0.0;
 var $302 = 0.0, $303 = 0, $304 = 0.0, $305 = 0.0, $306 = 0, $307 = 0, $308 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0.0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0.0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0.0;
 var $98 = 0, $99 = 0, $__nd$0$i$i$i = 0, $__nd$0$us$i$i$i = 0, $__new_last2$i$i$i = 0, $__new_last2$i$i$i7 = 0, $cmd = 0, $curcolor$sroa$0$031 = 0, $curcolor$sroa$0$1 = 0, $curcolor$sroa$0$2 = 0, $curcolor$sroa$6$032 = 0, $curcolor$sroa$6$1 = 0, $curcolor$sroa$6$2 = 0, $curcolor$sroa$7$033 = 0, $curcolor$sroa$7$1 = 0, $curcolor$sroa$7$2 = 0, $curcolor$sroa$8$034 = 0, $curcolor$sroa$8$1 = 0, $curcolor$sroa$8$2 = 0, $curcolori$035 = 0;
 var $curcolori$1 = 0, $curcolori$3 = 0, $dx$0$lcssa = 0.0, $dx$041 = 0.0, $dx$3 = 0.0, $dy$0$lcssa = 0.0, $dy$039 = 0.0, $dy$2 = 0.0, $i$037 = 0, $i$2 = 0, $lpad$phi$index = 0, $lpad$phi$index3 = 0, $maxwidth$0$lcssa = 0, $maxwidth$038 = 0, $maxwidth$1 = 0, $maxwidth$3 = 0, $maxwidth$4 = 0, $or$cond = 0, $prevglyph$036 = 0, $prevglyph$2 = 0;
 var $scevgep$i$i$i = 0, $scevgep$i$i$i12 = 0, $scevgep$i$i$i8 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i13 = 0, $scevgep4$i$i$i9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $cmd = sp;
 $0 = +HEAPF32[$offset>>2]; //@line 515 "modules/graphics/opengl/Font.cpp"
 $1 = ((($offset)) + 4|0); //@line 405 "modules/graphics/opengl/Font.cpp"
 $2 = +HEAPF32[$1>>2]; //@line 516 "modules/graphics/opengl/Font.cpp"
 $3 = ((($this)) + 92|0); //@line 994 "modules/graphics/opengl/Font.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 994 "modules/graphics/opengl/Font.cpp"
 $5 = ($4|0)==(0); //@line 994 "modules/graphics/opengl/Font.cpp"
 if ($5) {
  $6 = ((($this)) + 24|0); //@line 398 "modules/graphics/opengl/Font.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 398 "modules/graphics/opengl/Font.cpp"
  $8 = (+($7|0)); //@line 398 "modules/graphics/opengl/Font.cpp"
  $9 = $8 / 1.25; //@line 994 "modules/graphics/opengl/Font.cpp"
  $10 = $9 + 0.5; //@line 994 "modules/graphics/opengl/Font.cpp"
  $11 = (+Math_floor((+$10))); //@line 994 "modules/graphics/opengl/Font.cpp"
  $173 = $11;
 } else {
  $173 = 0.0;
 }
 HEAP32[$agg$result>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ((($agg$result)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$12>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ((($agg$result)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$13>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = ((($vertices)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = HEAP32[$14>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = HEAP32[$vertices>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = (($15) - ($16))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $18 = $17 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $19 = ((($codepoints)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = HEAP32[$19>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = HEAP32[$codepoints>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $22 = (($20) - ($21))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $23 = (($22) + ($18))|0; //@line 415 "modules/graphics/opengl/Font.cpp"
 $24 = ((($vertices)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $25 = HEAP32[$24>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $26 = (($25) - ($16))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = $26 >> 4; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $28 = ($27>>>0)<($23>>>0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 L4: do {
  if ($28) {
   $29 = ($23|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   do {
    if ($29) {
     $35 = 0;
    } else {
     $30 = $23 << 4; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     __THREW__ = 0;
     $31 = (invoke_ii(3348,($30|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $32 = __THREW__; __THREW__ = 0;
     $33 = $32&1;
     if ($33) {
      $67 = ___cxa_find_matching_catch_2()|0;
      $68 = tempRet0;
      $lpad$phi$index = $67;$lpad$phi$index3 = $68;
      break L4;
     } else {
      $35 = $31;
      break;
     }
    }
   } while(0);
   $34 = (($35) + ($18<<4)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $36 = $34; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $37 = (($35) + ($23<<4)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $38 = $37; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $39 = (0 - ($18))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $40 = (($34) + ($39<<4)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $41 = $40; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $42 = $16; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   _memcpy(($40|0),($42|0),($17|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$vertices>>2] = $41; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$14>>2] = $36; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$24>>2] = $38; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $43 = ($16|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($43) {
    label = 9;
   } else {
    __ZdlPv($42); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    label = 9;
   }
  } else {
   label = 9;
  }
 } while(0);
 L13: do {
  if ((label|0) == 9) {
   $44 = ((($codepoints)) + 12|0); //@line 421 "modules/graphics/opengl/Font.cpp"
   $45 = ((($codepoints)) + 16|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $46 = HEAP32[$45>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $47 = HEAP32[$44>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $48 = (($46) - ($47))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $49 = $48 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $50 = HEAP32[$19>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $51 = HEAP32[$codepoints>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $52 = (($50) - ($51))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $53 = ($52|0)>(0); //@line 423 "modules/graphics/opengl/Font.cpp"
   L15: do {
    if ($53) {
     $54 = ((($this)) + 24|0); //@line 398 "modules/graphics/opengl/Font.cpp"
     $55 = ((($this)) + 28|0); //@line 937 "modules/graphics/opengl/Font.cpp"
     $56 = ((($this)) + 132|0); //@line 441 "modules/graphics/opengl/Font.cpp"
     $57 = ((($this)) + 56|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $58 = $extra_spacing != 0.0; //@line 492 "modules/graphics/opengl/Font.cpp"
     $59 = ((($cmd)) + 4|0); //@line 479 "modules/graphics/opengl/Font.cpp"
     $60 = ((($cmd)) + 8|0); //@line 480 "modules/graphics/opengl/Font.cpp"
     $61 = ((($agg$result)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $62 = ((($this)) + 52|0); //@line 2015 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $$in = $51;$curcolor$sroa$0$031 = -1;$curcolor$sroa$6$032 = -1;$curcolor$sroa$7$033 = -1;$curcolor$sroa$8$034 = -1;$curcolori$035 = -1;$dx$041 = $0;$dy$039 = $2;$i$037 = 0;$maxwidth$038 = 0;$prevglyph$036 = 0;
     L17: while(1) {
      $78 = $$in; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $79 = (($78) + ($i$037<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $80 = HEAP32[$79>>2]|0; //@line 425 "modules/graphics/opengl/Font.cpp"
      $81 = (($curcolori$035) + 1)|0; //@line 427 "modules/graphics/opengl/Font.cpp"
      $82 = ($81|0)<($49|0); //@line 427 "modules/graphics/opengl/Font.cpp"
      if ($82) {
       $83 = HEAP32[$44>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $84 = (((($83) + ($81<<3)|0)) + 4|0); //@line 427 "modules/graphics/opengl/Font.cpp"
       $85 = HEAP32[$84>>2]|0; //@line 427 "modules/graphics/opengl/Font.cpp"
       $86 = ($85|0)==($i$037|0); //@line 427 "modules/graphics/opengl/Font.cpp"
       if ($86) {
        $87 = (($83) + ($81<<3)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $88 = HEAPU8[$87>>0]|(HEAPU8[$87+1>>0]<<8)|(HEAPU8[$87+2>>0]<<16)|(HEAPU8[$87+3>>0]<<24); //@line 428 "modules/graphics/opengl/Font.cpp"
        $89 = $88&255; //@line 428 "modules/graphics/opengl/Font.cpp"
        $90 = $88 >>> 8; //@line 428 "modules/graphics/opengl/Font.cpp"
        $91 = $90&255; //@line 428 "modules/graphics/opengl/Font.cpp"
        $92 = $88 >>> 16; //@line 428 "modules/graphics/opengl/Font.cpp"
        $93 = $92&255; //@line 428 "modules/graphics/opengl/Font.cpp"
        $94 = $88 >>> 24; //@line 428 "modules/graphics/opengl/Font.cpp"
        $95 = $94&255; //@line 428 "modules/graphics/opengl/Font.cpp"
        $curcolor$sroa$0$1 = $89;$curcolor$sroa$6$1 = $91;$curcolor$sroa$7$1 = $93;$curcolor$sroa$8$1 = $95;$curcolori$1 = $81;
       } else {
        $curcolor$sroa$0$1 = $curcolor$sroa$0$031;$curcolor$sroa$6$1 = $curcolor$sroa$6$032;$curcolor$sroa$7$1 = $curcolor$sroa$7$033;$curcolor$sroa$8$1 = $curcolor$sroa$8$034;$curcolori$1 = $curcolori$035;
       }
      } else {
       $curcolor$sroa$0$1 = $curcolor$sroa$0$031;$curcolor$sroa$6$1 = $curcolor$sroa$6$032;$curcolor$sroa$7$1 = $curcolor$sroa$7$033;$curcolor$sroa$8$1 = $curcolor$sroa$8$034;$curcolori$1 = $curcolori$035;
      }
      $96 = ($80|0)==(10); //@line 430 "modules/graphics/opengl/Font.cpp"
      do {
       if ($96) {
        $97 = (+($maxwidth$038|0)); //@line 432 "modules/graphics/opengl/Font.cpp"
        $98 = $dx$041 > $97; //@line 432 "modules/graphics/opengl/Font.cpp"
        $99 = (~~(($dx$041))); //@line 433 "modules/graphics/opengl/Font.cpp"
        $maxwidth$1 = $98 ? $99 : $maxwidth$038; //@line 432 "modules/graphics/opengl/Font.cpp"
        $100 = HEAP32[$54>>2]|0; //@line 398 "modules/graphics/opengl/Font.cpp"
        $101 = (+($100|0)); //@line 398 "modules/graphics/opengl/Font.cpp"
        $102 = +HEAPF32[$55>>2]; //@line 937 "modules/graphics/opengl/Font.cpp"
        $103 = $101 * $102; //@line 436 "modules/graphics/opengl/Font.cpp"
        $104 = $103 + 0.5; //@line 436 "modules/graphics/opengl/Font.cpp"
        $105 = (+Math_floor((+$104))); //@line 436 "modules/graphics/opengl/Font.cpp"
        $106 = $dy$039 + $105; //@line 436 "modules/graphics/opengl/Font.cpp"
        $curcolor$sroa$0$2 = $curcolor$sroa$0$1;$curcolor$sroa$6$2 = $curcolor$sroa$6$1;$curcolor$sroa$7$2 = $curcolor$sroa$7$1;$curcolor$sroa$8$2 = $curcolor$sroa$8$1;$curcolori$3 = $curcolori$1;$dx$3 = $0;$dy$2 = $106;$i$2 = $i$037;$maxwidth$3 = $maxwidth$1;$prevglyph$2 = $prevglyph$036;
       } else {
        $107 = HEAP32[$56>>2]|0; //@line 441 "modules/graphics/opengl/Font.cpp"
        $108 = HEAP32[$57>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $109 = ($108|0)==(0); //@line 2018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        L26: do {
         if ($109) {
          label = 37;
         } else {
          $110 = (($108) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          $111 = $110 & $108; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          $112 = ($111|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          if ($112) {
           $113 = $110 & $80; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
           $117 = $113;
          } else {
           $114 = (($80>>>0) % ($108>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
           $117 = $114;
          }
          $115 = HEAP32[$62>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $116 = (($115) + ($117<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $118 = HEAP32[$116>>2]|0; //@line 2021 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          $119 = ($118|0)==(0|0); //@line 2022 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          if ($119) {
           label = 37;
          } else {
           if ($112) {
            $$pn$us$i$i$i = $118;
            while(1) {
             $__nd$0$us$i$i$i = HEAP32[$$pn$us$i$i$i>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
             $120 = ($__nd$0$us$i$i$i|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
             if ($120) {
              label = 37;
              break L26;
             }
             $121 = ((($__nd$0$us$i$i$i)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
             $122 = HEAP32[$121>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
             $123 = $122 & $110; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
             $124 = ($123|0)==($117|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
             if (!($124)) {
              label = 37;
              break L26;
             }
             $125 = ((($__nd$0$us$i$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
             $126 = HEAP32[$125>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
             $127 = ($126|0)==($80|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
             if ($127) {
              $$sroa$0$0$i$i$i = $__nd$0$us$i$i$i;
              break;
             } else {
              $$pn$us$i$i$i = $__nd$0$us$i$i$i;
             }
            }
           } else {
            $$pn$i$i$i = $118;
            while(1) {
             $__nd$0$i$i$i = HEAP32[$$pn$i$i$i>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
             $128 = ($__nd$0$i$i$i|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
             if ($128) {
              label = 37;
              break L26;
             }
             $129 = ((($__nd$0$i$i$i)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
             $130 = HEAP32[$129>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
             $131 = (($130>>>0) % ($108>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
             $132 = ($131|0)==($117|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
             if (!($132)) {
              label = 37;
              break L26;
             }
             $133 = ((($__nd$0$i$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
             $134 = HEAP32[$133>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
             $135 = ($134|0)==($80|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
             if ($135) {
              $$sroa$0$0$i$i$i = $__nd$0$i$i$i;
              break;
             } else {
              $$pn$i$i$i = $__nd$0$i$i$i;
             }
            }
           }
           $136 = ((($$sroa$0$0$i$i$i)) + 12|0); //@line 323 "modules/graphics/opengl/Font.cpp"
           $$0$i = $136;$141 = $107;
          }
         }
        } while(0);
        if ((label|0) == 37) {
         label = 0;
         __THREW__ = 0;
         $137 = (invoke_iii(4017,($this|0),($80|0))|0); //@line 325 "modules/graphics/opengl/Font.cpp"
         $138 = __THREW__; __THREW__ = 0;
         $139 = $138&1;
         if ($139) {
          label = 12;
          break L17;
         }
         $$pre = HEAP32[$56>>2]|0; //@line 446 "modules/graphics/opengl/Font.cpp"
         $$0$i = $137;$141 = $$pre;
        }
        $140 = ($107|0)==($141|0); //@line 446 "modules/graphics/opengl/Font.cpp"
        if (!($140)) {
         $142 = HEAP32[$agg$result>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $143 = HEAP32[$12>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $144 = ($143|0)==($142|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if (!($144)) {
          $__new_last2$i$i$i7 = $142; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $scevgep$i$i$i8 = ((($143)) + -12|0);
          $145 = $scevgep$i$i$i8;
          $146 = (($145) - ($__new_last2$i$i$i7))|0;
          $147 = (($146>>>0) / 12)&-1;
          $148 = $147 ^ -1;
          $scevgep4$i$i$i9 = (($143) + (($148*12)|0)|0);
          HEAP32[$12>>2] = $scevgep4$i$i$i9; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         }
         $149 = HEAP32[$14>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $150 = HEAP32[$vertices>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $151 = (($149) - ($150))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $152 = $151 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $153 = ($18>>>0)>($152>>>0); //@line 1998 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $154 = $150; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $155 = $149; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($153) {
          $156 = (($18) - ($152))|0; //@line 1999 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          __THREW__ = 0;
          invoke_vii(4024,($vertices|0),($156|0)); //@line 1999 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $157 = __THREW__; __THREW__ = 0;
          $158 = $157&1;
          if ($158) {
           label = 12;
           break L17;
          } else {
           $curcolor$sroa$0$2 = -1;$curcolor$sroa$6$2 = -1;$curcolor$sroa$7$2 = -1;$curcolor$sroa$8$2 = -1;$curcolori$3 = -1;$dx$3 = $0;$dy$2 = $2;$i$2 = 0;$maxwidth$3 = 0;$prevglyph$2 = 0;
           break;
          }
         }
         $159 = ($18>>>0)<($152>>>0); //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if (!($159)) {
          $curcolor$sroa$0$2 = -1;$curcolor$sroa$6$2 = -1;$curcolor$sroa$7$2 = -1;$curcolor$sroa$8$2 = -1;$curcolori$3 = -1;$dx$3 = $0;$dy$2 = $2;$i$2 = 0;$maxwidth$3 = 0;$prevglyph$2 = 0;
          break;
         }
         $160 = (($154) + ($18<<4)|0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $161 = ($155|0)==($160|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($161) {
          $curcolor$sroa$0$2 = -1;$curcolor$sroa$6$2 = -1;$curcolor$sroa$7$2 = -1;$curcolor$sroa$8$2 = -1;$curcolori$3 = -1;$dx$3 = $0;$dy$2 = $2;$i$2 = 0;$maxwidth$3 = 0;$prevglyph$2 = 0;
          break;
         }
         $__new_last2$i$i$i = $160; //@line 818 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $scevgep$i$i$i = ((($155)) + -16|0);
         $162 = $scevgep$i$i$i;
         $163 = (($162) - ($__new_last2$i$i$i))|0;
         $164 = $163 >>> 4;
         $165 = $164 ^ -1;
         $scevgep4$i$i$i = (($155) + ($165<<4)|0);
         HEAP32[$14>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $curcolor$sroa$0$2 = -1;$curcolor$sroa$6$2 = -1;$curcolor$sroa$7$2 = -1;$curcolor$sroa$8$2 = -1;$curcolori$3 = -1;$dx$3 = $0;$dy$2 = $2;$i$2 = 0;$maxwidth$3 = 0;$prevglyph$2 = 0;
         break;
        }
        __THREW__ = 0;
        $166 = (+invoke_diii(4025,($this|0),($prevglyph$036|0),($80|0))); //@line 461 "modules/graphics/opengl/Font.cpp"
        $167 = __THREW__; __THREW__ = 0;
        $168 = $167&1;
        if ($168) {
         label = 12;
         break L17;
        }
        $169 = $dx$041 + $166; //@line 461 "modules/graphics/opengl/Font.cpp"
        $170 = HEAP32[$$0$i>>2]|0; //@line 463 "modules/graphics/opengl/Font.cpp"
        $171 = ($170|0)==(0); //@line 463 "modules/graphics/opengl/Font.cpp"
        if (!($171)) {
         $172 = $173 + $dy$039; //@line 470 "modules/graphics/opengl/Font.cpp"
         $174 = $curcolor$sroa$8$1&255; //@line 471 "modules/graphics/opengl/Font.cpp"
         $175 = $174 << 24; //@line 471 "modules/graphics/opengl/Font.cpp"
         $176 = $curcolor$sroa$7$1&255; //@line 471 "modules/graphics/opengl/Font.cpp"
         $177 = $176 << 16; //@line 471 "modules/graphics/opengl/Font.cpp"
         $178 = $curcolor$sroa$6$1&255; //@line 471 "modules/graphics/opengl/Font.cpp"
         $179 = $178 << 8; //@line 471 "modules/graphics/opengl/Font.cpp"
         $180 = $curcolor$sroa$0$1&255; //@line 471 "modules/graphics/opengl/Font.cpp"
         $181 = $179 | $180; //@line 471 "modules/graphics/opengl/Font.cpp"
         $182 = $181 | $177; //@line 471 "modules/graphics/opengl/Font.cpp"
         $183 = $182 | $175; //@line 471 "modules/graphics/opengl/Font.cpp"
         $184 = ((($$0$i)) + 8|0); //@line 468 "modules/graphics/opengl/Font.cpp"
         $185 = HEAP32[$14>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $186 = HEAP32[$24>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $187 = ($185|0)==($186|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($187) {
          __THREW__ = 0;
          invoke_vii(4026,($vertices|0),($184|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $190 = __THREW__; __THREW__ = 0;
          $191 = $190&1;
          if ($191) {
           label = 11;
           break L17;
          }
          $$pre48 = HEAP32[$14>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $193 = $$pre48;
         } else {
          ;HEAP32[$185>>2]=HEAP32[$184>>2]|0;HEAP32[$185+4>>2]=HEAP32[$184+4>>2]|0;HEAP32[$185+8>>2]=HEAP32[$184+8>>2]|0;HEAP32[$185+12>>2]=HEAP32[$184+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $188 = HEAP32[$14>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $189 = ((($188)) + 16|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$14>>2] = $189; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $193 = $189;
         }
         $192 = ((($193)) + -16|0); //@line 469 "modules/graphics/opengl/Font.cpp"
         $194 = +HEAPF32[$192>>2]; //@line 469 "modules/graphics/opengl/Font.cpp"
         $195 = $169 + $194; //@line 469 "modules/graphics/opengl/Font.cpp"
         HEAPF32[$192>>2] = $195; //@line 469 "modules/graphics/opengl/Font.cpp"
         $196 = ((($193)) + -12|0); //@line 470 "modules/graphics/opengl/Font.cpp"
         $197 = +HEAPF32[$196>>2]; //@line 470 "modules/graphics/opengl/Font.cpp"
         $198 = $172 + $197; //@line 470 "modules/graphics/opengl/Font.cpp"
         HEAPF32[$196>>2] = $198; //@line 470 "modules/graphics/opengl/Font.cpp"
         $199 = ((($193)) + -4|0); //@line 471 "modules/graphics/opengl/Font.cpp"
         HEAP8[$199>>0]=$183&255;HEAP8[$199+1>>0]=($183>>8)&255;HEAP8[$199+2>>0]=($183>>16)&255;HEAP8[$199+3>>0]=$183>>24; //@line 471 "modules/graphics/opengl/Font.cpp"
         $200 = ((($$0$i)) + 24|0); //@line 468 "modules/graphics/opengl/Font.cpp"
         $201 = HEAP32[$14>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $202 = HEAP32[$24>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $203 = ($201|0)==($202|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($203) {
          __THREW__ = 0;
          invoke_vii(4026,($vertices|0),($200|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $265 = __THREW__; __THREW__ = 0;
          $266 = $265&1;
          if ($266) {
           label = 11;
           break L17;
          }
          $$pre49 = HEAP32[$14>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $268 = $$pre49;
         } else {
          ;HEAP32[$201>>2]=HEAP32[$200>>2]|0;HEAP32[$201+4>>2]=HEAP32[$200+4>>2]|0;HEAP32[$201+8>>2]=HEAP32[$200+8>>2]|0;HEAP32[$201+12>>2]=HEAP32[$200+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $263 = HEAP32[$14>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $264 = ((($263)) + 16|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$14>>2] = $264; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $268 = $264;
         }
         $267 = ((($268)) + -16|0); //@line 469 "modules/graphics/opengl/Font.cpp"
         $269 = +HEAPF32[$267>>2]; //@line 469 "modules/graphics/opengl/Font.cpp"
         $270 = $169 + $269; //@line 469 "modules/graphics/opengl/Font.cpp"
         HEAPF32[$267>>2] = $270; //@line 469 "modules/graphics/opengl/Font.cpp"
         $271 = ((($268)) + -12|0); //@line 470 "modules/graphics/opengl/Font.cpp"
         $272 = +HEAPF32[$271>>2]; //@line 470 "modules/graphics/opengl/Font.cpp"
         $273 = $172 + $272; //@line 470 "modules/graphics/opengl/Font.cpp"
         HEAPF32[$271>>2] = $273; //@line 470 "modules/graphics/opengl/Font.cpp"
         $274 = ((($268)) + -4|0); //@line 471 "modules/graphics/opengl/Font.cpp"
         HEAP8[$274>>0]=$183&255;HEAP8[$274+1>>0]=($183>>8)&255;HEAP8[$274+2>>0]=($183>>16)&255;HEAP8[$274+3>>0]=$183>>24; //@line 471 "modules/graphics/opengl/Font.cpp"
         $275 = ((($$0$i)) + 40|0); //@line 468 "modules/graphics/opengl/Font.cpp"
         $276 = HEAP32[$14>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $277 = HEAP32[$24>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $278 = ($276|0)==($277|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($278) {
          __THREW__ = 0;
          invoke_vii(4026,($vertices|0),($275|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $281 = __THREW__; __THREW__ = 0;
          $282 = $281&1;
          if ($282) {
           label = 11;
           break L17;
          }
          $$pre50 = HEAP32[$14>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $284 = $$pre50;
         } else {
          ;HEAP32[$276>>2]=HEAP32[$275>>2]|0;HEAP32[$276+4>>2]=HEAP32[$275+4>>2]|0;HEAP32[$276+8>>2]=HEAP32[$275+8>>2]|0;HEAP32[$276+12>>2]=HEAP32[$275+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $279 = HEAP32[$14>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $280 = ((($279)) + 16|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$14>>2] = $280; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $284 = $280;
         }
         $283 = ((($284)) + -16|0); //@line 469 "modules/graphics/opengl/Font.cpp"
         $285 = +HEAPF32[$283>>2]; //@line 469 "modules/graphics/opengl/Font.cpp"
         $286 = $169 + $285; //@line 469 "modules/graphics/opengl/Font.cpp"
         HEAPF32[$283>>2] = $286; //@line 469 "modules/graphics/opengl/Font.cpp"
         $287 = ((($284)) + -12|0); //@line 470 "modules/graphics/opengl/Font.cpp"
         $288 = +HEAPF32[$287>>2]; //@line 470 "modules/graphics/opengl/Font.cpp"
         $289 = $172 + $288; //@line 470 "modules/graphics/opengl/Font.cpp"
         HEAPF32[$287>>2] = $289; //@line 470 "modules/graphics/opengl/Font.cpp"
         $290 = ((($284)) + -4|0); //@line 471 "modules/graphics/opengl/Font.cpp"
         HEAP8[$290>>0]=$183&255;HEAP8[$290+1>>0]=($183>>8)&255;HEAP8[$290+2>>0]=($183>>16)&255;HEAP8[$290+3>>0]=$183>>24; //@line 471 "modules/graphics/opengl/Font.cpp"
         $291 = ((($$0$i)) + 56|0); //@line 468 "modules/graphics/opengl/Font.cpp"
         $292 = HEAP32[$14>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $293 = HEAP32[$24>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $294 = ($292|0)==($293|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($294) {
          __THREW__ = 0;
          invoke_vii(4026,($vertices|0),($291|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $297 = __THREW__; __THREW__ = 0;
          $298 = $297&1;
          if ($298) {
           label = 11;
           break L17;
          }
          $$pre51 = HEAP32[$14>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $300 = $$pre51;
         } else {
          ;HEAP32[$292>>2]=HEAP32[$291>>2]|0;HEAP32[$292+4>>2]=HEAP32[$291+4>>2]|0;HEAP32[$292+8>>2]=HEAP32[$291+8>>2]|0;HEAP32[$292+12>>2]=HEAP32[$291+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $295 = HEAP32[$14>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $296 = ((($295)) + 16|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$14>>2] = $296; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $300 = $296;
         }
         $299 = ((($300)) + -16|0); //@line 469 "modules/graphics/opengl/Font.cpp"
         $301 = +HEAPF32[$299>>2]; //@line 469 "modules/graphics/opengl/Font.cpp"
         $302 = $169 + $301; //@line 469 "modules/graphics/opengl/Font.cpp"
         HEAPF32[$299>>2] = $302; //@line 469 "modules/graphics/opengl/Font.cpp"
         $303 = ((($300)) + -12|0); //@line 470 "modules/graphics/opengl/Font.cpp"
         $304 = +HEAPF32[$303>>2]; //@line 470 "modules/graphics/opengl/Font.cpp"
         $305 = $172 + $304; //@line 470 "modules/graphics/opengl/Font.cpp"
         HEAPF32[$303>>2] = $305; //@line 470 "modules/graphics/opengl/Font.cpp"
         $306 = ((($300)) + -4|0); //@line 471 "modules/graphics/opengl/Font.cpp"
         HEAP8[$306>>0]=$183&255;HEAP8[$306+1>>0]=($183>>8)&255;HEAP8[$306+2>>0]=($183>>16)&255;HEAP8[$306+3>>0]=$183>>24; //@line 471 "modules/graphics/opengl/Font.cpp"
         $307 = HEAP32[$agg$result>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $205 = HEAP32[$12>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $308 = ($307|0)==($205|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($308) {
          $$pre52 = HEAP32[$$0$i>>2]|0; //@line 481 "modules/graphics/opengl/Font.cpp"
          $214 = $$pre52;
          label = 55;
         } else {
          $204 = ((($205)) + -12|0); //@line 475 "modules/graphics/opengl/Font.cpp"
          $206 = HEAP32[$204>>2]|0; //@line 475 "modules/graphics/opengl/Font.cpp"
          $207 = HEAP32[$$0$i>>2]|0; //@line 475 "modules/graphics/opengl/Font.cpp"
          $208 = ($206|0)==($207|0); //@line 475 "modules/graphics/opengl/Font.cpp"
          if ($208) {
           $222 = $205;
          } else {
           $214 = $207;
           label = 55;
          }
         }
         if ((label|0) == 55) {
          label = 0;
          $209 = HEAP32[$14>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $210 = HEAP32[$vertices>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $211 = (($209) - ($210))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $212 = $211 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $213 = (($212) + -4)|0; //@line 479 "modules/graphics/opengl/Font.cpp"
          HEAP32[$59>>2] = $213; //@line 479 "modules/graphics/opengl/Font.cpp"
          HEAP32[$60>>2] = 0; //@line 480 "modules/graphics/opengl/Font.cpp"
          HEAP32[$cmd>>2] = $214; //@line 481 "modules/graphics/opengl/Font.cpp"
          $215 = HEAP32[$61>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $216 = ($205|0)==($215|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          if ($216) {
           __THREW__ = 0;
           invoke_vii(4027,($agg$result|0),($cmd|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $219 = __THREW__; __THREW__ = 0;
           $220 = $219&1;
           if ($220) {
            label = 12;
            break L17;
           }
           $$pre53$pre = HEAP32[$12>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $$pre53 = $$pre53$pre;
          } else {
           ;HEAP32[$205>>2]=HEAP32[$cmd>>2]|0;HEAP32[$205+4>>2]=HEAP32[$cmd+4>>2]|0;HEAP32[$205+8>>2]=HEAP32[$cmd+8>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $217 = HEAP32[$12>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $218 = ((($217)) + 12|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           HEAP32[$12>>2] = $218; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $$pre53 = $218;
          }
          $222 = $$pre53;
         }
         $221 = ((($222)) + -4|0); //@line 485 "modules/graphics/opengl/Font.cpp"
         $223 = HEAP32[$221>>2]|0; //@line 485 "modules/graphics/opengl/Font.cpp"
         $224 = (($223) + 4)|0; //@line 485 "modules/graphics/opengl/Font.cpp"
         HEAP32[$221>>2] = $224; //@line 485 "modules/graphics/opengl/Font.cpp"
        }
        $225 = ((($$0$i)) + 4|0); //@line 489 "modules/graphics/opengl/Font.cpp"
        $226 = HEAP32[$225>>2]|0; //@line 489 "modules/graphics/opengl/Font.cpp"
        $227 = (+($226|0)); //@line 489 "modules/graphics/opengl/Font.cpp"
        $228 = $169 + $227; //@line 489 "modules/graphics/opengl/Font.cpp"
        $229 = ($80|0)==(32); //@line 492 "modules/graphics/opengl/Font.cpp"
        $or$cond = $58 & $229; //@line 492 "modules/graphics/opengl/Font.cpp"
        if ($or$cond) {
         $230 = $228 + $extra_spacing; //@line 493 "modules/graphics/opengl/Font.cpp"
         $231 = (+Math_floor((+$230))); //@line 493 "modules/graphics/opengl/Font.cpp"
         $curcolor$sroa$0$2 = $curcolor$sroa$0$1;$curcolor$sroa$6$2 = $curcolor$sroa$6$1;$curcolor$sroa$7$2 = $curcolor$sroa$7$1;$curcolor$sroa$8$2 = $curcolor$sroa$8$1;$curcolori$3 = $curcolori$1;$dx$3 = $231;$dy$2 = $dy$039;$i$2 = $i$037;$maxwidth$3 = $maxwidth$038;$prevglyph$2 = 32;
        } else {
         $curcolor$sroa$0$2 = $curcolor$sroa$0$1;$curcolor$sroa$6$2 = $curcolor$sroa$6$1;$curcolor$sroa$7$2 = $curcolor$sroa$7$1;$curcolor$sroa$8$2 = $curcolor$sroa$8$1;$curcolori$3 = $curcolori$1;$dx$3 = $228;$dy$2 = $dy$039;$i$2 = $i$037;$maxwidth$3 = $maxwidth$038;$prevglyph$2 = $80;
        }
       }
      } while(0);
      $232 = (($i$2) + 1)|0; //@line 423 "modules/graphics/opengl/Font.cpp"
      $233 = HEAP32[$19>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $234 = HEAP32[$codepoints>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $235 = (($233) - ($234))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $236 = $235 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $237 = ($232|0)<($236|0); //@line 423 "modules/graphics/opengl/Font.cpp"
      if ($237) {
       $$in = $234;$curcolor$sroa$0$031 = $curcolor$sroa$0$2;$curcolor$sroa$6$032 = $curcolor$sroa$6$2;$curcolor$sroa$7$033 = $curcolor$sroa$7$2;$curcolor$sroa$8$034 = $curcolor$sroa$8$2;$curcolori$035 = $curcolori$3;$dx$041 = $dx$3;$dy$039 = $dy$2;$i$037 = $232;$maxwidth$038 = $maxwidth$3;$prevglyph$036 = $prevglyph$2;
      } else {
       $dx$0$lcssa = $dx$3;$dy$0$lcssa = $dy$2;$maxwidth$0$lcssa = $maxwidth$3;
       break L15;
      }
     }
     if ((label|0) == 11) {
      $63 = ___cxa_find_matching_catch_2()|0;
      $64 = tempRet0;
      $lpad$phi$index = $63;$lpad$phi$index3 = $64;
      break L13;
     }
     else if ((label|0) == 12) {
      $65 = ___cxa_find_matching_catch_2()|0;
      $66 = tempRet0;
      $lpad$phi$index = $65;$lpad$phi$index3 = $66;
      break L13;
     }
    } else {
     $dx$0$lcssa = $0;$dy$0$lcssa = $2;$maxwidth$0$lcssa = 0;
    }
   } while(0);
   $238 = HEAP32[$agg$result>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $239 = HEAP32[$12>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZNSt3__16__sortIRZN4love8graphics6opengl4Font16generateVerticesERKNS4_17ColoredCodepointsERNS_6vectorINS4_11GlyphVertexENS_9allocatorIS9_EEEEfNS1_6VectorEPNS4_8TextInfoEE3__0PNS4_11DrawCommandEEEvT0_SL_T_($238,$239); //@line 4037 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $240 = ($info|0)==(0|0); //@line 513 "modules/graphics/opengl/Font.cpp"
   if ($240) {
    STACKTOP = sp;return; //@line 520 "modules/graphics/opengl/Font.cpp"
   }
   $241 = (+($maxwidth$0$lcssa|0)); //@line 510 "modules/graphics/opengl/Font.cpp"
   $242 = $dx$0$lcssa > $241; //@line 510 "modules/graphics/opengl/Font.cpp"
   $243 = (~~(($dx$0$lcssa))); //@line 511 "modules/graphics/opengl/Font.cpp"
   $maxwidth$4 = $242 ? $243 : $maxwidth$0$lcssa; //@line 510 "modules/graphics/opengl/Font.cpp"
   $244 = (+($maxwidth$4|0)); //@line 515 "modules/graphics/opengl/Font.cpp"
   $245 = $244 - $0; //@line 515 "modules/graphics/opengl/Font.cpp"
   $246 = (~~(($245))); //@line 515 "modules/graphics/opengl/Font.cpp"
   HEAP32[$info>>2] = $246; //@line 515 "modules/graphics/opengl/Font.cpp"
   $247 = (~~(($dy$0$lcssa))); //@line 516 "modules/graphics/opengl/Font.cpp"
   $248 = (+($247|0)); //@line 516 "modules/graphics/opengl/Font.cpp"
   $249 = $dx$0$lcssa > 0.0; //@line 516 "modules/graphics/opengl/Font.cpp"
   if ($249) {
    $250 = ((($this)) + 24|0); //@line 398 "modules/graphics/opengl/Font.cpp"
    $251 = HEAP32[$250>>2]|0; //@line 398 "modules/graphics/opengl/Font.cpp"
    $252 = (+($251|0)); //@line 398 "modules/graphics/opengl/Font.cpp"
    $253 = ((($this)) + 28|0); //@line 937 "modules/graphics/opengl/Font.cpp"
    $254 = +HEAPF32[$253>>2]; //@line 937 "modules/graphics/opengl/Font.cpp"
    $255 = $252 * $254; //@line 516 "modules/graphics/opengl/Font.cpp"
    $256 = $255 + 0.5; //@line 516 "modules/graphics/opengl/Font.cpp"
    $257 = (+Math_floor((+$256))); //@line 516 "modules/graphics/opengl/Font.cpp"
    $259 = $257;
   } else {
    $259 = 0.0;
   }
   $258 = $248 + $259; //@line 516 "modules/graphics/opengl/Font.cpp"
   $260 = $258 - $2; //@line 516 "modules/graphics/opengl/Font.cpp"
   $261 = (~~(($260))); //@line 516 "modules/graphics/opengl/Font.cpp"
   $262 = ((($info)) + 4|0); //@line 516 "modules/graphics/opengl/Font.cpp"
   HEAP32[$262>>2] = $261; //@line 516 "modules/graphics/opengl/Font.cpp"
   STACKTOP = sp;return; //@line 520 "modules/graphics/opengl/Font.cpp"
  }
 } while(0);
 $69 = HEAP32[$agg$result>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $70 = ($69|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $71 = $69; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($70) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $72 = HEAP32[$12>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $73 = ($72|0)==($69|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($73)) {
  $scevgep$i$i$i12 = ((($72)) + -12|0);
  $74 = $scevgep$i$i$i12;
  $75 = (($74) - ($71))|0;
  $76 = (($75>>>0) / 12)&-1;
  $77 = $76 ^ -1;
  $scevgep4$i$i$i13 = (($72) + (($77*12)|0)|0);
  HEAP32[$12>>2] = $scevgep4$i$i$i13; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($69); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love8graphics6opengl4Font11GlyphVertexENS_9allocatorIS5_EEE8__appendEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$0$i = 0, $$0$i2 = 0, $$0$i34 = 0, $$0$i35 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__v$sroa$12$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($this)) + 4|0); //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($1) - ($3))|0; //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = $4 >> 4; //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)<($__n>>>0); //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($6)) {
  $7 = $3; //@line 988 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i = $__n;$8 = $7;
  while(1) {
   ;HEAP32[$8>>2]=0|0;HEAP32[$8+4>>2]=0|0;HEAP32[$8+8>>2]=0|0;HEAP32[$8+12>>2]=0|0; //@line 37 "./modules/graphics/Color.h"
   $9 = HEAP32[$2>>2]|0; //@line 989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $10 = ((($9)) + 16|0); //@line 989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$2>>2] = $10; //@line 989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $11 = (($$0$i) + -1)|0; //@line 990 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $12 = ($11|0)==(0); //@line 992 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($12) {
    break;
   } else {
    $$0$i = $11;$8 = $10;
   }
  }
  return; //@line 1053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $13 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = (($3) - ($13))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = $14 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = (($15) + ($__n))|0; //@line 1049 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ($16>>>0)>(268435455); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($17) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$0>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre6 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = $$pre;$20 = $$pre6;
 } else {
  $19 = $1;$20 = $13;
 }
 $18 = (($19) - ($20))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = $18 >> 4; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $22 = ($21>>>0)<(134217727); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($22) {
  $26 = $18 >> 3; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $27 = ($26>>>0)<($16>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $28 = $27 ? $16 : $26; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $29 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $30 = (($29) - ($20))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $31 = $30 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $32 = ($28|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($32) {
   $$0$i35 = 0;$36 = 0;$37 = $31;
  } else {
   $$0$i34 = $28;$55 = $31;
   label = 9;
  }
 } else {
  $23 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $24 = (($23) - ($20))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $25 = $24 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i34 = 268435455;$55 = $25;
  label = 9;
 }
 if ((label|0) == 9) {
  $33 = $$0$i34 << 4; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $34 = (__Znwj($33)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i35 = $$0$i34;$36 = $34;$37 = $55;
 }
 $35 = (($36) + ($37<<4)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $38 = $35; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $39 = (($36) + ($$0$i35<<4)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $$0$i2 = $__n;$$in = $35;$__v$sroa$12$0 = $38;
 while(1) {
  ;HEAP32[$$in>>2]=0|0;HEAP32[$$in+4>>2]=0|0;HEAP32[$$in+8>>2]=0|0;HEAP32[$$in+12>>2]=0|0; //@line 37 "./modules/graphics/Color.h"
  $40 = $__v$sroa$12$0; //@line 218 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $41 = ((($40)) + 16|0); //@line 218 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $42 = $41; //@line 218 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $43 = (($$0$i2) + -1)|0; //@line 219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $44 = ($43|0)==(0); //@line 220 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($44) {
   $$lcssa = $42;
   break;
  } else {
   $$0$i2 = $43;$$in = $41;$__v$sroa$12$0 = $42;
  }
 }
 $45 = $39; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $46 = HEAP32[$this>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $47 = HEAP32[$2>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $48 = $46; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $49 = (($47) - ($48))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $50 = $49 >> 4; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $51 = (0 - ($50))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $52 = (($35) + ($51<<4)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $53 = $52; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($52|0),($46|0),($49|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $53; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$2>>2] = $$lcssa; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $45; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $54 = ($46|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($54) {
  return; //@line 1053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($46); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNSt3__16vectorIN4love8graphics6opengl4Font11GlyphVertexENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(268435455); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 4; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(134217727); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 3; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$32 = $18;
  } else {
   $$0$i1 = $17;$40 = $18;$41 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 268435455;$40 = $12;$41 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 4; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $41;$32 = $40;
 }
 $24 = (($25) + ($26<<4)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<4)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 ;HEAP32[$24>>2]=HEAP32[$__x>>2]|0;HEAP32[$24+4>>2]=HEAP32[$__x+4>>2]|0;HEAP32[$24+8>>2]=HEAP32[$__x+8>>2]|0;HEAP32[$24+12>>2]=HEAP32[$__x+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $29 = ((($24)) + 16|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = $29; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = (($32) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = $31 >> 4; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = (0 - ($33))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (($24) + ($34<<4)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = $35; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($35|0),($37|0),($31|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $36; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $38 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($38) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $39 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($39); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNSt3__16vectorIN4love8graphics6opengl4Font11DrawCommandENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($3|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(357913941); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9|0) / 12)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(178956970); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $10 << 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = (($19|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$32 = $18;
  } else {
   $$0$i1 = $17;$39 = $18;$40 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 357913941;$39 = $12;$40 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = ($$0$i1*12)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $40;$32 = $39;
 }
 $24 = (($25) + (($26*12)|0)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + (($$0$i2*12)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 ;HEAP32[$24>>2]=HEAP32[$__x>>2]|0;HEAP32[$24+4>>2]=HEAP32[$__x+4>>2]|0;HEAP32[$24+8>>2]=HEAP32[$__x+8>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $29 = ((($24)) + 12|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = $29; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = (($32) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = (($31|0) / -12)&-1; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = (($24) + (($33*12)|0)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = $34; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($34|0),($36|0),($31|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $35; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $37 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($37) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $38 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($38); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNSt3__16__sortIRZN4love8graphics6opengl4Font16generateVerticesERKNS4_17ColoredCodepointsERNS_6vectorINS4_11GlyphVertexENS_9allocatorIS9_EEEEfNS1_6VectorEPNS4_8TextInfoEE3__0PNS4_11DrawCommandEEEvT0_SL_T_($__first,$__last) {
 $__first = $__first|0;
 $__last = $__last|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i$i = 0, $$0$i$i14 = 0, $$0$i9$i = 0, $$0$i9$i$i = 0, $$0$i9$i16 = 0, $$0$lcssa291 = 0, $$0$lcssa292 = 0, $$0$lcssa293 = 0, $$0$lcssa294 = 0, $$0$lcssa295 = 0, $$0$lcssa298 = 0, $$0$ph = 0, $$04$ph = 0, $$04$ph$lcssa307 = 0, $$__m$0 = 0, $$lcssa = 0, $$lcssa289 = 0, $$lcssa290 = 0;
 var $$lcssa313 = 0, $$lcssa314 = 0, $$lcssa319 = 0, $$lcssa320 = 0, $$lcssa327 = 0, $$lcssa328 = 0, $$lcssa335 = 0, $$lcssa336 = 0, $$phi$trans$insert$i = 0, $$pre$i = 0, $$pre$phi112Z2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0;
 var $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0;
 var $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0;
 var $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0;
 var $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__i$084 = 0, $__i$084$lcssa = 0, $__i$09$i = 0;
 var $__i$1 = 0, $__i$2 = 0, $__i$2$lcssa = 0, $__i$2$ph = 0, $__i$3 = 0, $__i$3$lcssa = 0, $__i$3$ph = 0, $__i$4 = 0, $__i$4$lcssa300 = 0, $__j$08$i = 0, $__j$08$i$phi = 0, $__j$083 = 0, $__j$083$phi = 0, $__j$1$ph = 0, $__j$2 = 0, $__j$3 = 0, $__j$4$ph = 0, $__j$5 = 0, $__k$0$lcssa$i = 0, $__k$04$i = 0;
 var $__m$0$ph = 0, $__m$2 = 0, $__n_swaps$0 = 0, $__n_swaps$3 = 0, $__n_swaps$4$ph = 0, $__n_swaps$5 = 0, $__n_swaps$6 = 0, $__t$i = 0, $__t$i$i = 0, $__t$i$i$i = 0, $__t$i$i13 = 0, $__t$i22 = 0, $__t$i25 = 0, $__t$i3$i = 0, $__t$i3$i$i = 0, $__t$i3$i12 = 0, $__t$i30 = 0, $__t$i33 = 0, $__t$i38 = 0, $__t$i4$i = 0;
 var $__t$i4$i$i = 0, $__t$i4$i11 = 0, $__t$i5$i = 0, $__t$i5$i$i = 0, $__t$i5$i10 = 0, $__t$i8$i = 0, $__t$i8$i$i = 0, $__t$i8$i9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__t$i38 = sp + 240|0;
 $__t$i33 = sp + 228|0;
 $__t$i30 = sp + 216|0;
 $__t$i25 = sp + 204|0;
 $__t$i22 = sp + 192|0;
 $__t$i8$i9 = sp + 180|0;
 $__t$i5$i10 = sp + 168|0;
 $__t$i4$i11 = sp + 156|0;
 $__t$i3$i12 = sp + 144|0;
 $__t$i$i13 = sp + 132|0;
 $__t$i8$i$i = sp + 120|0;
 $__t$i5$i$i = sp + 108|0;
 $__t$i4$i$i = sp + 96|0;
 $__t$i3$i$i = sp + 84|0;
 $__t$i$i$i = sp + 72|0;
 $__t$i8$i = sp + 60|0;
 $__t$i5$i = sp + 48|0;
 $__t$i4$i = sp + 36|0;
 $__t$i3$i = sp + 24|0;
 $__t$i$i = sp + 12|0;
 $__t$i = sp;
 $$0$ph = $__first;$$04$ph = $__last;
 L1: while(1) {
  $0 = $$04$ph; //@line 3844 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $1 = ((($$04$ph)) + -12|0); //@line 3872 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $2 = ((($$04$ph)) + -24|0); //@line 3904 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $3 = ((($$04$ph)) + -8|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $$0 = $$0$ph;
  L3: while(1) {
   $4 = $$0; //@line 3844 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $5 = (($0) - ($4))|0; //@line 3844 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $6 = (($5|0) / 12)&-1; //@line 3844 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   switch ($6|0) {
   case 2:  {
    $$0$lcssa291 = $$0;$$lcssa319 = $1;$$lcssa327 = $3;$$lcssa335 = $1;
    label = 4;
    break L1;
    break;
   }
   case 3:  {
    $$0$lcssa292 = $$0;$$lcssa320 = $1;$$lcssa328 = $3;$$lcssa336 = $1;
    label = 8;
    break L1;
    break;
   }
   case 4:  {
    $$0$lcssa293 = $$0;$$lcssa313 = $1;
    label = 26;
    break L1;
    break;
   }
   case 5:  {
    $$0$lcssa294 = $$0;$$lcssa314 = $1;
    label = 27;
    break L1;
    break;
   }
   case 1: case 0:  {
    label = 136;
    break L1;
    break;
   }
   default: {
   }
   }
   $54 = ($5|0)<(372); //@line 3864 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   if ($54) {
    $$0$lcssa295 = $$0;$$04$ph$lcssa307 = $$04$ph;
    label = 29;
    break L1;
   }
   $117 = ($5|0)>(11988); //@line 3876 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $118 = (($5|0) / 24)&-1; //@line 3878 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $119 = (($$0) + (($118*12)|0)|0); //@line 3879 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   do {
    if ($117) {
     $120 = (($5|0) / 48)&-1; //@line 3880 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $121 = (($$0) + (($120*12)|0)|0); //@line 3881 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $122 = (($119) + (($120*12)|0)|0); //@line 3881 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $123 = (__ZNSt3__17__sort5IRZN4love8graphics6opengl4Font16generateVerticesERKNS4_17ColoredCodepointsERNS_6vectorINS4_11GlyphVertexENS_9allocatorIS9_EEEEfNS1_6VectorEPNS4_8TextInfoEE3__0PNS4_11DrawCommandEEEjT0_SL_SL_SL_SL_T_($$0,$121,$119,$122,$1)|0); //@line 3881 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $$pre$phi112Z2D = $119;$$pre$phiZ2D = $$0;$__n_swaps$0 = $123;
    } else {
     $124 = HEAP32[$119>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
     $125 = HEAP32[$$0>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
     $126 = ($124|0)==($125|0); //@line 502 "modules/graphics/opengl/Font.cpp"
     if ($126) {
      $128 = (((($$0) + (($118*12)|0)|0)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      $129 = HEAP32[$128>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
      $130 = ((($$0)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      $131 = HEAP32[$130>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
      $132 = ($129|0)<($131|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      $$0$i$i14 = $132;
     } else {
      $127 = ($124>>>0)<($125>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
      $$0$i$i14 = $127;
     }
     $133 = HEAP32[$1>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
     $134 = ($133|0)==($124|0); //@line 502 "modules/graphics/opengl/Font.cpp"
     if ($134) {
      $136 = HEAP32[$3>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
      $137 = (((($$0) + (($118*12)|0)|0)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      $138 = HEAP32[$137>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
      $139 = ($136|0)<($138|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      $$0$i9$i16 = $139;
     } else {
      $135 = ($133>>>0)<($124>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
      $$0$i9$i16 = $135;
     }
     if (!($$0$i$i14)) {
      if (!($$0$i9$i16)) {
       $$pre$phi112Z2D = $119;$$pre$phiZ2D = $$0;$__n_swaps$0 = 0;
       break;
      }
      ;HEAP32[$__t$i8$i9>>2]=HEAP32[$119>>2]|0;HEAP32[$__t$i8$i9+4>>2]=HEAP32[$119+4>>2]|0;HEAP32[$__t$i8$i9+8>>2]=HEAP32[$119+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      ;HEAP32[$119>>2]=HEAP32[$1>>2]|0;HEAP32[$119+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$119+8>>2]=HEAP32[$1+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      ;HEAP32[$1>>2]=HEAP32[$__t$i8$i9>>2]|0;HEAP32[$1+4>>2]=HEAP32[$__t$i8$i9+4>>2]|0;HEAP32[$1+8>>2]=HEAP32[$__t$i8$i9+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $140 = HEAP32[$119>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
      $141 = HEAP32[$$0>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
      $142 = ($140|0)==($141|0); //@line 502 "modules/graphics/opengl/Font.cpp"
      if ($142) {
       $144 = (((($$0) + (($118*12)|0)|0)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
       $145 = HEAP32[$144>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
       $146 = ((($$0)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
       $147 = HEAP32[$146>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
       $148 = ($145|0)<($147|0); //@line 505 "modules/graphics/opengl/Font.cpp"
       if (!($148)) {
        $$pre$phi112Z2D = $119;$$pre$phiZ2D = $$0;$__n_swaps$0 = 1;
        break;
       }
      } else {
       $143 = ($140>>>0)<($141>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
       if (!($143)) {
        $$pre$phi112Z2D = $119;$$pre$phiZ2D = $$0;$__n_swaps$0 = 1;
        break;
       }
      }
      ;HEAP32[$__t$i5$i10>>2]=HEAP32[$$0>>2]|0;HEAP32[$__t$i5$i10+4>>2]=HEAP32[$$0+4>>2]|0;HEAP32[$__t$i5$i10+8>>2]=HEAP32[$$0+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      ;HEAP32[$$0>>2]=HEAP32[$119>>2]|0;HEAP32[$$0+4>>2]=HEAP32[$119+4>>2]|0;HEAP32[$$0+8>>2]=HEAP32[$119+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      ;HEAP32[$119>>2]=HEAP32[$__t$i5$i10>>2]|0;HEAP32[$119+4>>2]=HEAP32[$__t$i5$i10+4>>2]|0;HEAP32[$119+8>>2]=HEAP32[$__t$i5$i10+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $$pre$phi112Z2D = $119;$$pre$phiZ2D = $$0;$__n_swaps$0 = 2;
      break;
     }
     if ($$0$i9$i16) {
      ;HEAP32[$__t$i4$i11>>2]=HEAP32[$$0>>2]|0;HEAP32[$__t$i4$i11+4>>2]=HEAP32[$$0+4>>2]|0;HEAP32[$__t$i4$i11+8>>2]=HEAP32[$$0+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      ;HEAP32[$$0>>2]=HEAP32[$1>>2]|0;HEAP32[$$0+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$$0+8>>2]=HEAP32[$1+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      ;HEAP32[$1>>2]=HEAP32[$__t$i4$i11>>2]|0;HEAP32[$1+4>>2]=HEAP32[$__t$i4$i11+4>>2]|0;HEAP32[$1+8>>2]=HEAP32[$__t$i4$i11+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $$pre$phi112Z2D = $119;$$pre$phiZ2D = $$0;$__n_swaps$0 = 1;
      break;
     }
     ;HEAP32[$__t$i3$i12>>2]=HEAP32[$$0>>2]|0;HEAP32[$__t$i3$i12+4>>2]=HEAP32[$$0+4>>2]|0;HEAP32[$__t$i3$i12+8>>2]=HEAP32[$$0+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$$0>>2]=HEAP32[$119>>2]|0;HEAP32[$$0+4>>2]=HEAP32[$119+4>>2]|0;HEAP32[$$0+8>>2]=HEAP32[$119+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$119>>2]=HEAP32[$__t$i3$i12>>2]|0;HEAP32[$119+4>>2]=HEAP32[$__t$i3$i12+4>>2]|0;HEAP32[$119+8>>2]=HEAP32[$__t$i3$i12+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $149 = HEAP32[$1>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
     $150 = HEAP32[$119>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
     $151 = ($149|0)==($150|0); //@line 502 "modules/graphics/opengl/Font.cpp"
     if ($151) {
      $153 = HEAP32[$3>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
      $154 = (((($$0) + (($118*12)|0)|0)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      $155 = HEAP32[$154>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
      $156 = ($153|0)<($155|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      if (!($156)) {
       $$pre$phi112Z2D = $119;$$pre$phiZ2D = $$0;$__n_swaps$0 = 1;
       break;
      }
     } else {
      $152 = ($149>>>0)<($150>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
      if (!($152)) {
       $$pre$phi112Z2D = $119;$$pre$phiZ2D = $$0;$__n_swaps$0 = 1;
       break;
      }
     }
     ;HEAP32[$__t$i$i13>>2]=HEAP32[$119>>2]|0;HEAP32[$__t$i$i13+4>>2]=HEAP32[$119+4>>2]|0;HEAP32[$__t$i$i13+8>>2]=HEAP32[$119+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$119>>2]=HEAP32[$1>>2]|0;HEAP32[$119+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$119+8>>2]=HEAP32[$1+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$1>>2]=HEAP32[$__t$i$i13>>2]|0;HEAP32[$1+4>>2]=HEAP32[$__t$i$i13+4>>2]|0;HEAP32[$1+8>>2]=HEAP32[$__t$i$i13+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $$pre$phi112Z2D = $119;$$pre$phiZ2D = $$0;$__n_swaps$0 = 2;
    }
   } while(0);
   $157 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
   $158 = HEAP32[$$pre$phi112Z2D>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
   $159 = ($157|0)==($158|0); //@line 502 "modules/graphics/opengl/Font.cpp"
   if ($159) {
    $161 = ((($$0)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
    $162 = HEAP32[$161>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
    $163 = (((($$0) + (($118*12)|0)|0)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
    $164 = HEAP32[$163>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
    $165 = ($162|0)<($164|0); //@line 505 "modules/graphics/opengl/Font.cpp"
    if ($165) {
     $__j$3 = $1;$__n_swaps$3 = $__n_swaps$0;
    } else {
     label = 82;
    }
   } else {
    $160 = ($157>>>0)<($158>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
    if ($160) {
     $__j$3 = $1;$__n_swaps$3 = $__n_swaps$0;
    } else {
     label = 82;
    }
   }
   L37: do {
    if ((label|0) == 82) {
     label = 0;
     $166 = ($$0|0)==($2|0); //@line 3904 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     L39: do {
      if (!($166)) {
       $167 = (((($$0) + (($118*12)|0)|0)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
       $210 = $2;$__j$083 = $1;
       while(1) {
        $209 = HEAP32[$210>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
        $211 = ($209|0)==($158|0); //@line 502 "modules/graphics/opengl/Font.cpp"
        if ($211) {
         $213 = ((($__j$083)) + -8|0); //@line 505 "modules/graphics/opengl/Font.cpp"
         $214 = HEAP32[$213>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
         $215 = HEAP32[$167>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
         $216 = ($214|0)<($215|0); //@line 505 "modules/graphics/opengl/Font.cpp"
         if ($216) {
          $$lcssa = $210;
          break;
         }
        } else {
         $212 = ($209>>>0)<($158>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
         if ($212) {
          $$lcssa = $210;
          break;
         }
        }
        $217 = ((($210)) + -12|0); //@line 3904 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $218 = ($$0|0)==($217|0); //@line 3904 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        if ($218) {
         break L39;
        } else {
         $__j$083$phi = $210;$210 = $217;$__j$083 = $__j$083$phi;
        }
       }
       ;HEAP32[$__t$i30>>2]=HEAP32[$$0>>2]|0;HEAP32[$__t$i30+4>>2]=HEAP32[$$0+4>>2]|0;HEAP32[$__t$i30+8>>2]=HEAP32[$$0+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       ;HEAP32[$$0>>2]=HEAP32[$$lcssa>>2]|0;HEAP32[$$0+4>>2]=HEAP32[$$lcssa+4>>2]|0;HEAP32[$$0+8>>2]=HEAP32[$$lcssa+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       ;HEAP32[$$lcssa>>2]=HEAP32[$__t$i30>>2]|0;HEAP32[$$lcssa+4>>2]=HEAP32[$__t$i30+4>>2]|0;HEAP32[$$lcssa+8>>2]=HEAP32[$__t$i30+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $219 = (($__n_swaps$0) + 1)|0; //@line 3950 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $__j$3 = $$lcssa;$__n_swaps$3 = $219;
       break L37;
      }
     } while(0);
     $168 = ((($$0)) + 12|0); //@line 3908 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $169 = HEAP32[$1>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
     $170 = ($157|0)==($169|0); //@line 502 "modules/graphics/opengl/Font.cpp"
     if ($170) {
      $172 = ((($$0)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      $173 = HEAP32[$172>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
      $174 = HEAP32[$3>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
      $175 = ($173|0)<($174|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      if ($175) {
       $__i$1 = $168;
      } else {
       label = 87;
      }
     } else {
      $171 = ($157>>>0)<($169>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
      if ($171) {
       $__i$1 = $168;
      } else {
       label = 87;
      }
     }
     if ((label|0) == 87) {
      label = 0;
      $176 = ($168|0)==($1|0); //@line 3914 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if ($176) {
       label = 136;
       break L1;
      }
      $177 = ((($$0)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      $__i$084 = $168;
      while(1) {
       $178 = HEAP32[$__i$084>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
       $179 = ($157|0)==($178|0); //@line 502 "modules/graphics/opengl/Font.cpp"
       if ($179) {
        $181 = HEAP32[$177>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
        $182 = ((($__i$084)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
        $183 = HEAP32[$182>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
        $184 = ($181|0)<($183|0); //@line 505 "modules/graphics/opengl/Font.cpp"
        if ($184) {
         $__i$084$lcssa = $__i$084;
         break;
        }
       } else {
        $180 = ($157>>>0)<($178>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
        if ($180) {
         $__i$084$lcssa = $__i$084;
         break;
        }
       }
       $186 = ((($__i$084)) + 12|0); //@line 3923 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $187 = ($186|0)==($1|0); //@line 3914 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       if ($187) {
        label = 136;
        break L1;
       } else {
        $__i$084 = $186;
       }
      }
      ;HEAP32[$__t$i38>>2]=HEAP32[$__i$084$lcssa>>2]|0;HEAP32[$__t$i38+4>>2]=HEAP32[$__i$084$lcssa+4>>2]|0;HEAP32[$__t$i38+8>>2]=HEAP32[$__i$084$lcssa+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      ;HEAP32[$__i$084$lcssa>>2]=HEAP32[$1>>2]|0;HEAP32[$__i$084$lcssa+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$__i$084$lcssa+8>>2]=HEAP32[$1+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      ;HEAP32[$1>>2]=HEAP32[$__t$i38>>2]|0;HEAP32[$1+4>>2]=HEAP32[$__t$i38+4>>2]|0;HEAP32[$1+8>>2]=HEAP32[$__t$i38+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $185 = ((($__i$084$lcssa)) + 12|0); //@line 3920 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $__i$1 = $185;
     }
     $188 = ($__i$1|0)==($1|0); //@line 3927 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     if ($188) {
      label = 136;
      break L1;
     }
     $189 = ((($$0)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
     $__i$2$ph = $__i$1;$__j$1$ph = $1;
     while(1) {
      $190 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
      $__i$2 = $__i$2$ph;
      while(1) {
       $191 = HEAP32[$__i$2>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
       $192 = ($190|0)==($191|0); //@line 502 "modules/graphics/opengl/Font.cpp"
       if ($192) {
        $194 = HEAP32[$189>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
        $195 = ((($__i$2)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
        $196 = HEAP32[$195>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
        $197 = ($194|0)<($196|0); //@line 505 "modules/graphics/opengl/Font.cpp"
        if ($197) {
         $__i$2$lcssa = $__i$2;
         break;
        }
       } else {
        $193 = ($190>>>0)<($191>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
        if ($193) {
         $__i$2$lcssa = $__i$2;
         break;
        }
       }
       $198 = ((($__i$2)) + 12|0); //@line 3932 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $__i$2 = $198;
      }
      $__j$2 = $__j$1$ph;
      while(1) {
       $199 = ((($__j$2)) + -12|0); //@line 3933 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $200 = HEAP32[$199>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
       $201 = ($190|0)==($200|0); //@line 502 "modules/graphics/opengl/Font.cpp"
       if ($201) {
        $203 = HEAP32[$189>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
        $204 = ((($__j$2)) + -8|0); //@line 505 "modules/graphics/opengl/Font.cpp"
        $205 = HEAP32[$204>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
        $206 = ($203|0)<($205|0); //@line 505 "modules/graphics/opengl/Font.cpp"
        if ($206) {
         $__j$2 = $199;
         continue;
        } else {
         $$lcssa289 = $199;
         break;
        }
       } else {
        $202 = ($190>>>0)<($200>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
        if ($202) {
         $__j$2 = $199;
         continue;
        } else {
         $$lcssa289 = $199;
         break;
        }
       }
      }
      $207 = ($__i$2$lcssa>>>0)<($$lcssa289>>>0); //@line 3935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if (!($207)) {
       $$0 = $__i$2$lcssa;
       continue L3;
      }
      ;HEAP32[$__t$i33>>2]=HEAP32[$__i$2$lcssa>>2]|0;HEAP32[$__t$i33+4>>2]=HEAP32[$__i$2$lcssa+4>>2]|0;HEAP32[$__t$i33+8>>2]=HEAP32[$__i$2$lcssa+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      ;HEAP32[$__i$2$lcssa>>2]=HEAP32[$$lcssa289>>2]|0;HEAP32[$__i$2$lcssa+4>>2]=HEAP32[$$lcssa289+4>>2]|0;HEAP32[$__i$2$lcssa+8>>2]=HEAP32[$$lcssa289+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      ;HEAP32[$$lcssa289>>2]=HEAP32[$__t$i33>>2]|0;HEAP32[$$lcssa289+4>>2]=HEAP32[$__t$i33+4>>2]|0;HEAP32[$$lcssa289+8>>2]=HEAP32[$__t$i33+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $208 = ((($__i$2$lcssa)) + 12|0); //@line 3939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $__i$2$ph = $208;$__j$1$ph = $$lcssa289;
     }
    }
   } while(0);
   $220 = ((($$0)) + 12|0); //@line 3956 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $221 = ($220>>>0)<($__j$3>>>0); //@line 3959 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   L81: do {
    if ($221) {
     $__i$3$ph = $220;$__j$4$ph = $__j$3;$__m$0$ph = $119;$__n_swaps$4$ph = $__n_swaps$3;
     while(1) {
      $222 = HEAP32[$__m$0$ph>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
      $223 = ((($__m$0$ph)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      $__i$3 = $__i$3$ph;
      while(1) {
       $224 = HEAP32[$__i$3>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
       $225 = ($224|0)==($222|0); //@line 502 "modules/graphics/opengl/Font.cpp"
       if ($225) {
        $227 = ((($__i$3)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
        $228 = HEAP32[$227>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
        $229 = HEAP32[$223>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
        $230 = ($228|0)<($229|0); //@line 505 "modules/graphics/opengl/Font.cpp"
        if (!($230)) {
         $__i$3$lcssa = $__i$3;
         break;
        }
       } else {
        $226 = ($224>>>0)<($222>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
        if (!($226)) {
         $__i$3$lcssa = $__i$3;
         break;
        }
       }
       $231 = ((($__i$3)) + 12|0); //@line 3967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $__i$3 = $231;
      }
      $__j$5 = $__j$4$ph;
      while(1) {
       $232 = ((($__j$5)) + -12|0); //@line 3969 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $233 = HEAP32[$232>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
       $234 = ($233|0)==($222|0); //@line 502 "modules/graphics/opengl/Font.cpp"
       if ($234) {
        $236 = ((($__j$5)) + -8|0); //@line 505 "modules/graphics/opengl/Font.cpp"
        $237 = HEAP32[$236>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
        $238 = HEAP32[$223>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
        $239 = ($237|0)<($238|0); //@line 505 "modules/graphics/opengl/Font.cpp"
        if ($239) {
         $$lcssa290 = $232;
         break;
        } else {
         $__j$5 = $232;
         continue;
        }
       } else {
        $235 = ($233>>>0)<($222>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
        if ($235) {
         $$lcssa290 = $232;
         break;
        } else {
         $__j$5 = $232;
         continue;
        }
       }
      }
      $240 = ($__i$3$lcssa>>>0)>($$lcssa290>>>0); //@line 3971 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if ($240) {
       $__i$4 = $__i$3$lcssa;$__m$2 = $__m$0$ph;$__n_swaps$5 = $__n_swaps$4$ph;
       break L81;
      }
      ;HEAP32[$__t$i25>>2]=HEAP32[$__i$3$lcssa>>2]|0;HEAP32[$__t$i25+4>>2]=HEAP32[$__i$3$lcssa+4>>2]|0;HEAP32[$__t$i25+8>>2]=HEAP32[$__i$3$lcssa+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      ;HEAP32[$__i$3$lcssa>>2]=HEAP32[$$lcssa290>>2]|0;HEAP32[$__i$3$lcssa+4>>2]=HEAP32[$$lcssa290+4>>2]|0;HEAP32[$__i$3$lcssa+8>>2]=HEAP32[$$lcssa290+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      ;HEAP32[$$lcssa290>>2]=HEAP32[$__t$i25>>2]|0;HEAP32[$$lcssa290+4>>2]=HEAP32[$__t$i25+4>>2]|0;HEAP32[$$lcssa290+8>>2]=HEAP32[$__t$i25+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $241 = (($__n_swaps$4$ph) + 1)|0; //@line 3974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $242 = ($__m$0$ph|0)==($__i$3$lcssa|0); //@line 3977 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $$__m$0 = $242 ? $$lcssa290 : $__m$0$ph; //@line 3977 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $243 = ((($__i$3$lcssa)) + 12|0); //@line 3979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $__i$3$ph = $243;$__j$4$ph = $$lcssa290;$__m$0$ph = $$__m$0;$__n_swaps$4$ph = $241;
     }
    } else {
     $__i$4 = $220;$__m$2 = $119;$__n_swaps$5 = $__n_swaps$3;
    }
   } while(0);
   $244 = ($__i$4|0)==($__m$2|0); //@line 3983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   do {
    if ($244) {
     $__n_swaps$6 = $__n_swaps$5;
    } else {
     $245 = HEAP32[$__m$2>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
     $246 = HEAP32[$__i$4>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
     $247 = ($245|0)==($246|0); //@line 502 "modules/graphics/opengl/Font.cpp"
     if ($247) {
      $249 = ((($__m$2)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      $250 = HEAP32[$249>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
      $251 = ((($__i$4)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      $252 = HEAP32[$251>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
      $253 = ($250|0)<($252|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      if (!($253)) {
       $__n_swaps$6 = $__n_swaps$5;
       break;
      }
     } else {
      $248 = ($245>>>0)<($246>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
      if (!($248)) {
       $__n_swaps$6 = $__n_swaps$5;
       break;
      }
     }
     ;HEAP32[$__t$i22>>2]=HEAP32[$__i$4>>2]|0;HEAP32[$__t$i22+4>>2]=HEAP32[$__i$4+4>>2]|0;HEAP32[$__t$i22+8>>2]=HEAP32[$__i$4+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$__i$4>>2]=HEAP32[$__m$2>>2]|0;HEAP32[$__i$4+4>>2]=HEAP32[$__m$2+4>>2]|0;HEAP32[$__i$4+8>>2]=HEAP32[$__m$2+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$__m$2>>2]=HEAP32[$__t$i22>>2]|0;HEAP32[$__m$2+4>>2]=HEAP32[$__t$i22+4>>2]|0;HEAP32[$__m$2+8>>2]=HEAP32[$__t$i22+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $254 = (($__n_swaps$5) + 1)|0; //@line 3986 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $__n_swaps$6 = $254;
    }
   } while(0);
   $255 = ($__n_swaps$6|0)==(0); //@line 3990 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   if ($255) {
    $256 = (__ZNSt3__127__insertion_sort_incompleteIRZN4love8graphics6opengl4Font16generateVerticesERKNS4_17ColoredCodepointsERNS_6vectorINS4_11GlyphVertexENS_9allocatorIS9_EEEEfNS1_6VectorEPNS4_8TextInfoEE3__0PNS4_11DrawCommandEEEbT0_SL_T_($$0,$__i$4)|0); //@line 3992 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $257 = ((($__i$4)) + 12|0); //@line 3993 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $258 = (__ZNSt3__127__insertion_sort_incompleteIRZN4love8graphics6opengl4Font16generateVerticesERKNS4_17ColoredCodepointsERNS_6vectorINS4_11GlyphVertexENS_9allocatorIS9_EEEEfNS1_6VectorEPNS4_8TextInfoEE3__0PNS4_11DrawCommandEEEbT0_SL_T_($257,$$04$ph)|0); //@line 3993 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    if ($258) {
     label = 135;
     break;
    }
    if ($256) {
     $$0 = $257;
     continue;
    }
   }
   $259 = $__i$4; //@line 4010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $260 = (($259) - ($4))|0; //@line 4010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $261 = (($0) - ($259))|0; //@line 4010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $262 = ($260|0)<($261|0); //@line 4010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   if (!($262)) {
    $$0$lcssa298 = $$0;$__i$4$lcssa300 = $__i$4;
    label = 134;
    break;
   }
   __ZNSt3__16__sortIRZN4love8graphics6opengl4Font16generateVerticesERKNS4_17ColoredCodepointsERNS_6vectorINS4_11GlyphVertexENS_9allocatorIS9_EEEEfNS1_6VectorEPNS4_8TextInfoEE3__0PNS4_11DrawCommandEEEvT0_SL_T_($$0,$__i$4); //@line 4012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $263 = ((($__i$4)) + 12|0); //@line 4014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $$0 = $263;
  }
  if ((label|0) == 134) {
   label = 0;
   $264 = ((($__i$4$lcssa300)) + 12|0); //@line 4018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   __ZNSt3__16__sortIRZN4love8graphics6opengl4Font16generateVerticesERKNS4_17ColoredCodepointsERNS_6vectorINS4_11GlyphVertexENS_9allocatorIS9_EEEEfNS1_6VectorEPNS4_8TextInfoEE3__0PNS4_11DrawCommandEEEvT0_SL_T_($264,$$04$ph); //@line 4018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $$0$ph = $$0$lcssa298;$$04$ph = $__i$4$lcssa300;
   continue;
  }
  else if ((label|0) == 135) {
   label = 0;
   if ($256) {
    label = 136;
    break;
   } else {
    $$0$ph = $$0;$$04$ph = $__i$4;
    continue;
   }
  }
 }
 if ((label|0) == 4) {
  $7 = HEAP32[$$lcssa319>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $8 = HEAP32[$$0$lcssa291>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $9 = ($7|0)==($8|0); //@line 502 "modules/graphics/opengl/Font.cpp"
  if ($9) {
   $11 = HEAP32[$$lcssa327>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $12 = ((($$0$lcssa291)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $13 = HEAP32[$12>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $14 = ($11|0)<($13|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   if (!($14)) {
    STACKTOP = sp;return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   }
  } else {
   $10 = ($7>>>0)<($8>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
   if (!($10)) {
    STACKTOP = sp;return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   }
  }
  ;HEAP32[$__t$i>>2]=HEAP32[$$0$lcssa291>>2]|0;HEAP32[$__t$i+4>>2]=HEAP32[$$0$lcssa291+4>>2]|0;HEAP32[$__t$i+8>>2]=HEAP32[$$0$lcssa291+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  ;HEAP32[$$0$lcssa291>>2]=HEAP32[$$lcssa335>>2]|0;HEAP32[$$0$lcssa291+4>>2]=HEAP32[$$lcssa335+4>>2]|0;HEAP32[$$0$lcssa291+8>>2]=HEAP32[$$lcssa335+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  ;HEAP32[$$lcssa335>>2]=HEAP32[$__t$i>>2]|0;HEAP32[$$lcssa335+4>>2]=HEAP32[$__t$i+4>>2]|0;HEAP32[$$lcssa335+8>>2]=HEAP32[$__t$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  STACKTOP = sp;return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
 else if ((label|0) == 8) {
  $15 = ((($$0$lcssa292)) + 12|0); //@line 3855 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $16 = HEAP32[$15>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $17 = HEAP32[$$0$lcssa292>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $18 = ($16|0)==($17|0); //@line 502 "modules/graphics/opengl/Font.cpp"
  if ($18) {
   $20 = ((($$0$lcssa292)) + 16|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $21 = HEAP32[$20>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $22 = ((($$0$lcssa292)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $23 = HEAP32[$22>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $24 = ($21|0)<($23|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $$0$i$i = $24;
  } else {
   $19 = ($16>>>0)<($17>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
   $$0$i$i = $19;
  }
  $25 = HEAP32[$$lcssa320>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $26 = ($25|0)==($16|0); //@line 502 "modules/graphics/opengl/Font.cpp"
  if ($26) {
   $28 = HEAP32[$$lcssa328>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $29 = ((($$0$lcssa292)) + 16|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $30 = HEAP32[$29>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $31 = ($28|0)<($30|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $$0$i9$i = $31;
  } else {
   $27 = ($25>>>0)<($16>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
   $$0$i9$i = $27;
  }
  if (!($$0$i$i)) {
   if (!($$0$i9$i)) {
    STACKTOP = sp;return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   }
   ;HEAP32[$__t$i8$i>>2]=HEAP32[$15>>2]|0;HEAP32[$__t$i8$i+4>>2]=HEAP32[$15+4>>2]|0;HEAP32[$__t$i8$i+8>>2]=HEAP32[$15+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   ;HEAP32[$15>>2]=HEAP32[$$lcssa336>>2]|0;HEAP32[$15+4>>2]=HEAP32[$$lcssa336+4>>2]|0;HEAP32[$15+8>>2]=HEAP32[$$lcssa336+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   ;HEAP32[$$lcssa336>>2]=HEAP32[$__t$i8$i>>2]|0;HEAP32[$$lcssa336+4>>2]=HEAP32[$__t$i8$i+4>>2]|0;HEAP32[$$lcssa336+8>>2]=HEAP32[$__t$i8$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $32 = HEAP32[$15>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
   $33 = HEAP32[$$0$lcssa292>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
   $34 = ($32|0)==($33|0); //@line 502 "modules/graphics/opengl/Font.cpp"
   if ($34) {
    $36 = ((($$0$lcssa292)) + 16|0); //@line 505 "modules/graphics/opengl/Font.cpp"
    $37 = HEAP32[$36>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
    $38 = ((($$0$lcssa292)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
    $39 = HEAP32[$38>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
    $40 = ($37|0)<($39|0); //@line 505 "modules/graphics/opengl/Font.cpp"
    if (!($40)) {
     STACKTOP = sp;return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    }
   } else {
    $35 = ($32>>>0)<($33>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
    if (!($35)) {
     STACKTOP = sp;return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    }
   }
   ;HEAP32[$__t$i5$i>>2]=HEAP32[$$0$lcssa292>>2]|0;HEAP32[$__t$i5$i+4>>2]=HEAP32[$$0$lcssa292+4>>2]|0;HEAP32[$__t$i5$i+8>>2]=HEAP32[$$0$lcssa292+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   ;HEAP32[$$0$lcssa292>>2]=HEAP32[$15>>2]|0;HEAP32[$$0$lcssa292+4>>2]=HEAP32[$15+4>>2]|0;HEAP32[$$0$lcssa292+8>>2]=HEAP32[$15+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   ;HEAP32[$15>>2]=HEAP32[$__t$i5$i>>2]|0;HEAP32[$15+4>>2]=HEAP32[$__t$i5$i+4>>2]|0;HEAP32[$15+8>>2]=HEAP32[$__t$i5$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   STACKTOP = sp;return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  if ($$0$i9$i) {
   ;HEAP32[$__t$i4$i>>2]=HEAP32[$$0$lcssa292>>2]|0;HEAP32[$__t$i4$i+4>>2]=HEAP32[$$0$lcssa292+4>>2]|0;HEAP32[$__t$i4$i+8>>2]=HEAP32[$$0$lcssa292+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   ;HEAP32[$$0$lcssa292>>2]=HEAP32[$$lcssa336>>2]|0;HEAP32[$$0$lcssa292+4>>2]=HEAP32[$$lcssa336+4>>2]|0;HEAP32[$$0$lcssa292+8>>2]=HEAP32[$$lcssa336+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   ;HEAP32[$$lcssa336>>2]=HEAP32[$__t$i4$i>>2]|0;HEAP32[$$lcssa336+4>>2]=HEAP32[$__t$i4$i+4>>2]|0;HEAP32[$$lcssa336+8>>2]=HEAP32[$__t$i4$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   STACKTOP = sp;return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  ;HEAP32[$__t$i3$i>>2]=HEAP32[$$0$lcssa292>>2]|0;HEAP32[$__t$i3$i+4>>2]=HEAP32[$$0$lcssa292+4>>2]|0;HEAP32[$__t$i3$i+8>>2]=HEAP32[$$0$lcssa292+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  ;HEAP32[$$0$lcssa292>>2]=HEAP32[$15>>2]|0;HEAP32[$$0$lcssa292+4>>2]=HEAP32[$15+4>>2]|0;HEAP32[$$0$lcssa292+8>>2]=HEAP32[$15+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  ;HEAP32[$15>>2]=HEAP32[$__t$i3$i>>2]|0;HEAP32[$15+4>>2]=HEAP32[$__t$i3$i+4>>2]|0;HEAP32[$15+8>>2]=HEAP32[$__t$i3$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $41 = HEAP32[$$lcssa320>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $42 = HEAP32[$15>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $43 = ($41|0)==($42|0); //@line 502 "modules/graphics/opengl/Font.cpp"
  if ($43) {
   $45 = HEAP32[$$lcssa328>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $46 = ((($$0$lcssa292)) + 16|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $47 = HEAP32[$46>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $48 = ($45|0)<($47|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   if (!($48)) {
    STACKTOP = sp;return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   }
  } else {
   $44 = ($41>>>0)<($42>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
   if (!($44)) {
    STACKTOP = sp;return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   }
  }
  ;HEAP32[$__t$i$i>>2]=HEAP32[$15>>2]|0;HEAP32[$__t$i$i+4>>2]=HEAP32[$15+4>>2]|0;HEAP32[$__t$i$i+8>>2]=HEAP32[$15+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  ;HEAP32[$15>>2]=HEAP32[$$lcssa336>>2]|0;HEAP32[$15+4>>2]=HEAP32[$$lcssa336+4>>2]|0;HEAP32[$15+8>>2]=HEAP32[$$lcssa336+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  ;HEAP32[$$lcssa336>>2]=HEAP32[$__t$i$i>>2]|0;HEAP32[$$lcssa336+4>>2]=HEAP32[$__t$i$i+4>>2]|0;HEAP32[$$lcssa336+8>>2]=HEAP32[$__t$i$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  STACKTOP = sp;return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
 else if ((label|0) == 26) {
  $49 = ((($$0$lcssa293)) + 12|0); //@line 3858 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $50 = ((($$0$lcssa293)) + 24|0); //@line 3858 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  (__ZNSt3__17__sort4IRZN4love8graphics6opengl4Font16generateVerticesERKNS4_17ColoredCodepointsERNS_6vectorINS4_11GlyphVertexENS_9allocatorIS9_EEEEfNS1_6VectorEPNS4_8TextInfoEE3__0PNS4_11DrawCommandEEEjT0_SL_SL_SL_T_($$0$lcssa293,$49,$50,$$lcssa313)|0); //@line 3858 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  STACKTOP = sp;return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
 else if ((label|0) == 27) {
  $51 = ((($$0$lcssa294)) + 12|0); //@line 3861 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $52 = ((($$0$lcssa294)) + 24|0); //@line 3861 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $53 = ((($$0$lcssa294)) + 36|0); //@line 3861 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  (__ZNSt3__17__sort5IRZN4love8graphics6opengl4Font16generateVerticesERKNS4_17ColoredCodepointsERNS_6vectorINS4_11GlyphVertexENS_9allocatorIS9_EEEEfNS1_6VectorEPNS4_8TextInfoEE3__0PNS4_11DrawCommandEEEjT0_SL_SL_SL_SL_T_($$0$lcssa294,$51,$52,$53,$$lcssa314)|0); //@line 3861 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  STACKTOP = sp;return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
 else if ((label|0) == 29) {
  $55 = ((($$0$lcssa295)) + 24|0); //@line 3728 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $56 = ((($$0$lcssa295)) + 12|0); //@line 3729 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $57 = HEAP32[$56>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $58 = HEAP32[$$0$lcssa295>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $59 = ($57|0)==($58|0); //@line 502 "modules/graphics/opengl/Font.cpp"
  if ($59) {
   $61 = ((($$0$lcssa295)) + 16|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $62 = HEAP32[$61>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $63 = ((($$0$lcssa295)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $64 = HEAP32[$63>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $65 = ($62|0)<($64|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $$0$i$i$i = $65;
  } else {
   $60 = ($57>>>0)<($58>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
   $$0$i$i$i = $60;
  }
  $66 = HEAP32[$55>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $67 = ($66|0)==($57|0); //@line 502 "modules/graphics/opengl/Font.cpp"
  if ($67) {
   $69 = ((($$0$lcssa295)) + 28|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $70 = HEAP32[$69>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $71 = ((($$0$lcssa295)) + 16|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $72 = HEAP32[$71>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $73 = ($70|0)<($72|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $$0$i9$i$i = $73;
  } else {
   $68 = ($66>>>0)<($57>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
   $$0$i9$i$i = $68;
  }
  do {
   if ($$0$i$i$i) {
    if ($$0$i9$i$i) {
     ;HEAP32[$__t$i4$i$i>>2]=HEAP32[$$0$lcssa295>>2]|0;HEAP32[$__t$i4$i$i+4>>2]=HEAP32[$$0$lcssa295+4>>2]|0;HEAP32[$__t$i4$i$i+8>>2]=HEAP32[$$0$lcssa295+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$$0$lcssa295>>2]=HEAP32[$55>>2]|0;HEAP32[$$0$lcssa295+4>>2]=HEAP32[$55+4>>2]|0;HEAP32[$$0$lcssa295+8>>2]=HEAP32[$55+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$55>>2]=HEAP32[$__t$i4$i$i>>2]|0;HEAP32[$55+4>>2]=HEAP32[$__t$i4$i$i+4>>2]|0;HEAP32[$55+8>>2]=HEAP32[$__t$i4$i$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     break;
    }
    ;HEAP32[$__t$i3$i$i>>2]=HEAP32[$$0$lcssa295>>2]|0;HEAP32[$__t$i3$i$i+4>>2]=HEAP32[$$0$lcssa295+4>>2]|0;HEAP32[$__t$i3$i$i+8>>2]=HEAP32[$$0$lcssa295+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    ;HEAP32[$$0$lcssa295>>2]=HEAP32[$56>>2]|0;HEAP32[$$0$lcssa295+4>>2]=HEAP32[$56+4>>2]|0;HEAP32[$$0$lcssa295+8>>2]=HEAP32[$56+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    ;HEAP32[$56>>2]=HEAP32[$__t$i3$i$i>>2]|0;HEAP32[$56+4>>2]=HEAP32[$__t$i3$i$i+4>>2]|0;HEAP32[$56+8>>2]=HEAP32[$__t$i3$i$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $82 = HEAP32[$56>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
    $83 = ($66|0)==($82|0); //@line 502 "modules/graphics/opengl/Font.cpp"
    if ($83) {
     $85 = ((($$0$lcssa295)) + 28|0); //@line 505 "modules/graphics/opengl/Font.cpp"
     $86 = HEAP32[$85>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
     $87 = ((($$0$lcssa295)) + 16|0); //@line 505 "modules/graphics/opengl/Font.cpp"
     $88 = HEAP32[$87>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
     $89 = ($86|0)<($88|0); //@line 505 "modules/graphics/opengl/Font.cpp"
     if (!($89)) {
      break;
     }
    } else {
     $84 = ($66>>>0)<($82>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
     if (!($84)) {
      break;
     }
    }
    ;HEAP32[$__t$i$i$i>>2]=HEAP32[$56>>2]|0;HEAP32[$__t$i$i$i+4>>2]=HEAP32[$56+4>>2]|0;HEAP32[$__t$i$i$i+8>>2]=HEAP32[$56+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    ;HEAP32[$56>>2]=HEAP32[$55>>2]|0;HEAP32[$56+4>>2]=HEAP32[$55+4>>2]|0;HEAP32[$56+8>>2]=HEAP32[$55+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    ;HEAP32[$55>>2]=HEAP32[$__t$i$i$i>>2]|0;HEAP32[$55+4>>2]=HEAP32[$__t$i$i$i+4>>2]|0;HEAP32[$55+8>>2]=HEAP32[$__t$i$i$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   } else {
    if ($$0$i9$i$i) {
     ;HEAP32[$__t$i8$i$i>>2]=HEAP32[$56>>2]|0;HEAP32[$__t$i8$i$i+4>>2]=HEAP32[$56+4>>2]|0;HEAP32[$__t$i8$i$i+8>>2]=HEAP32[$56+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$56>>2]=HEAP32[$55>>2]|0;HEAP32[$56+4>>2]=HEAP32[$55+4>>2]|0;HEAP32[$56+8>>2]=HEAP32[$55+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$55>>2]=HEAP32[$__t$i8$i$i>>2]|0;HEAP32[$55+4>>2]=HEAP32[$__t$i8$i$i+4>>2]|0;HEAP32[$55+8>>2]=HEAP32[$__t$i8$i$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $74 = HEAP32[$56>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
     $75 = ($74|0)==($58|0); //@line 502 "modules/graphics/opengl/Font.cpp"
     if ($75) {
      $77 = ((($$0$lcssa295)) + 16|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      $78 = HEAP32[$77>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
      $79 = ((($$0$lcssa295)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      $80 = HEAP32[$79>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
      $81 = ($78|0)<($80|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      if (!($81)) {
       break;
      }
     } else {
      $76 = ($74>>>0)<($58>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
      if (!($76)) {
       break;
      }
     }
     ;HEAP32[$__t$i5$i$i>>2]=HEAP32[$$0$lcssa295>>2]|0;HEAP32[$__t$i5$i$i+4>>2]=HEAP32[$$0$lcssa295+4>>2]|0;HEAP32[$__t$i5$i$i+8>>2]=HEAP32[$$0$lcssa295+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$$0$lcssa295>>2]=HEAP32[$56>>2]|0;HEAP32[$$0$lcssa295+4>>2]=HEAP32[$56+4>>2]|0;HEAP32[$$0$lcssa295+8>>2]=HEAP32[$56+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$56>>2]=HEAP32[$__t$i5$i$i>>2]|0;HEAP32[$56+4>>2]=HEAP32[$__t$i5$i$i+4>>2]|0;HEAP32[$56+8>>2]=HEAP32[$__t$i5$i$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    }
   }
  } while(0);
  $90 = ((($$0$lcssa295)) + 36|0); //@line 3730 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $91 = ($90|0)==($$04$ph$lcssa307|0); //@line 3730 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if ($91) {
   STACKTOP = sp;return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  } else {
   $__i$09$i = $90;$__j$08$i = $55;
  }
  while(1) {
   $92 = HEAP32[$__i$09$i>>2]|0; //@line 3734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $93 = HEAP32[$__j$08$i>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
   $94 = ($92|0)==($93|0); //@line 502 "modules/graphics/opengl/Font.cpp"
   if ($94) {
    $96 = ((($__i$09$i)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
    $97 = HEAP32[$96>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
    $98 = ((($__j$08$i)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
    $99 = HEAP32[$98>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
    $100 = ($97|0)<($99|0); //@line 505 "modules/graphics/opengl/Font.cpp"
    if ($100) {
     $111 = $97;
     label = 52;
    }
   } else {
    $95 = ($92>>>0)<($93>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
    if ($95) {
     $$phi$trans$insert$i = ((($__i$09$i)) + 4|0); //@line 3734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $$pre$i = HEAP32[$$phi$trans$insert$i>>2]|0; //@line 3734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $111 = $$pre$i;
     label = 52;
    }
   }
   if ((label|0) == 52) {
    label = 0;
    $101 = ((($__i$09$i)) + 8|0); //@line 3734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $102 = HEAP32[$101>>2]|0; //@line 3734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    ;HEAP32[$__i$09$i>>2]=HEAP32[$__j$08$i>>2]|0;HEAP32[$__i$09$i+4>>2]=HEAP32[$__j$08$i+4>>2]|0;HEAP32[$__i$09$i+8>>2]=HEAP32[$__j$08$i+8>>2]|0; //@line 3739 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $103 = ($__j$08$i|0)==($$0$lcssa295|0); //@line 3741 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    L151: do {
     if ($103) {
      $__k$0$lcssa$i = $$0$lcssa295;
     } else {
      $__k$04$i = $__j$08$i;
      while(1) {
       $104 = ((($__k$04$i)) + -12|0); //@line 3741 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $105 = HEAP32[$104>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
       $106 = ($92|0)==($105|0); //@line 502 "modules/graphics/opengl/Font.cpp"
       if ($106) {
        $108 = ((($__k$04$i)) + -8|0); //@line 505 "modules/graphics/opengl/Font.cpp"
        $109 = HEAP32[$108>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
        $110 = ($111|0)<($109|0); //@line 505 "modules/graphics/opengl/Font.cpp"
        if (!($110)) {
         $__k$0$lcssa$i = $__k$04$i;
         break L151;
        }
       } else {
        $107 = ($92>>>0)<($105>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
        if (!($107)) {
         $__k$0$lcssa$i = $__k$04$i;
         break L151;
        }
       }
       ;HEAP32[$__k$04$i>>2]=HEAP32[$104>>2]|0;HEAP32[$__k$04$i+4>>2]=HEAP32[$104+4>>2]|0;HEAP32[$__k$04$i+8>>2]=HEAP32[$104+8>>2]|0; //@line 3739 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $112 = ($104|0)==($$0$lcssa295|0); //@line 3741 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       if ($112) {
        $__k$0$lcssa$i = $$0$lcssa295;
        break;
       } else {
        $__k$04$i = $104;
       }
      }
     }
    } while(0);
    HEAP32[$__k$0$lcssa$i>>2] = $92; //@line 3742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $113 = ((($__k$0$lcssa$i)) + 4|0); //@line 3742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    HEAP32[$113>>2] = $111; //@line 3742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $114 = ((($__k$0$lcssa$i)) + 8|0); //@line 3742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    HEAP32[$114>>2] = $102; //@line 3742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   }
   $115 = ((($__i$09$i)) + 12|0); //@line 3730 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $116 = ($115|0)==($$04$ph$lcssa307|0); //@line 3730 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   if ($116) {
    break;
   } else {
    $__j$08$i$phi = $__i$09$i;$__i$09$i = $115;$__j$08$i = $__j$08$i$phi;
   }
  }
  STACKTOP = sp;return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
 else if ((label|0) == 136) {
  STACKTOP = sp;return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
}
function __ZNSt3__17__sort4IRZN4love8graphics6opengl4Font16generateVerticesERKNS4_17ColoredCodepointsERNS_6vectorINS4_11GlyphVertexENS_9allocatorIS9_EEEEfNS1_6VectorEPNS4_8TextInfoEE3__0PNS4_11DrawCommandEEEjT0_SL_SL_SL_T_($__x1,$__x2,$__x3,$__x4) {
 $__x1 = $__x1|0;
 $__x2 = $__x2|0;
 $__x3 = $__x3|0;
 $__x4 = $__x4|0;
 var $$0$i = 0, $$0$i$i = 0, $$0$i9$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $__r$0 = 0, $__t$i = 0, $__t$i$i = 0, $__t$i2 = 0, $__t$i3$i = 0, $__t$i4$i = 0, $__t$i5 = 0, $__t$i5$i = 0, $__t$i8$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__t$i5 = sp + 84|0;
 $__t$i2 = sp + 72|0;
 $__t$i = sp + 60|0;
 $__t$i8$i = sp + 48|0;
 $__t$i5$i = sp + 36|0;
 $__t$i4$i = sp + 24|0;
 $__t$i3$i = sp + 12|0;
 $__t$i$i = sp;
 $0 = HEAP32[$__x2>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
 $1 = HEAP32[$__x1>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
 $2 = ($0|0)==($1|0); //@line 502 "modules/graphics/opengl/Font.cpp"
 if ($2) {
  $4 = ((($__x2)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $6 = ((($__x1)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $8 = ($5|0)<($7|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $$0$i$i = $8;
 } else {
  $3 = ($0>>>0)<($1>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
  $$0$i$i = $3;
 }
 $9 = HEAP32[$__x3>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
 $10 = ($9|0)==($0|0); //@line 502 "modules/graphics/opengl/Font.cpp"
 if ($10) {
  $12 = ((($__x3)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $13 = HEAP32[$12>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $14 = ((($__x2)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $15 = HEAP32[$14>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $16 = ($13|0)<($15|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $$0$i9$i = $16;
 } else {
  $11 = ($9>>>0)<($0>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
  $$0$i9$i = $11;
 }
 do {
  if ($$0$i$i) {
   if ($$0$i9$i) {
    ;HEAP32[$__t$i4$i>>2]=HEAP32[$__x1>>2]|0;HEAP32[$__t$i4$i+4>>2]=HEAP32[$__x1+4>>2]|0;HEAP32[$__t$i4$i+8>>2]=HEAP32[$__x1+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    ;HEAP32[$__x1>>2]=HEAP32[$__x3>>2]|0;HEAP32[$__x1+4>>2]=HEAP32[$__x3+4>>2]|0;HEAP32[$__x1+8>>2]=HEAP32[$__x3+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    ;HEAP32[$__x3>>2]=HEAP32[$__t$i4$i>>2]|0;HEAP32[$__x3+4>>2]=HEAP32[$__t$i4$i+4>>2]|0;HEAP32[$__x3+8>>2]=HEAP32[$__t$i4$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $$0$i = 1;
    break;
   }
   ;HEAP32[$__t$i3$i>>2]=HEAP32[$__x1>>2]|0;HEAP32[$__t$i3$i+4>>2]=HEAP32[$__x1+4>>2]|0;HEAP32[$__t$i3$i+8>>2]=HEAP32[$__x1+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   ;HEAP32[$__x1>>2]=HEAP32[$__x2>>2]|0;HEAP32[$__x1+4>>2]=HEAP32[$__x2+4>>2]|0;HEAP32[$__x1+8>>2]=HEAP32[$__x2+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   ;HEAP32[$__x2>>2]=HEAP32[$__t$i3$i>>2]|0;HEAP32[$__x2+4>>2]=HEAP32[$__t$i3$i+4>>2]|0;HEAP32[$__x2+8>>2]=HEAP32[$__t$i3$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $26 = HEAP32[$__x3>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
   $27 = HEAP32[$__x2>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
   $28 = ($26|0)==($27|0); //@line 502 "modules/graphics/opengl/Font.cpp"
   if ($28) {
    $30 = ((($__x3)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
    $31 = HEAP32[$30>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
    $32 = ((($__x2)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
    $33 = HEAP32[$32>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
    $34 = ($31|0)<($33|0); //@line 505 "modules/graphics/opengl/Font.cpp"
    if (!($34)) {
     $$0$i = 1;
     break;
    }
   } else {
    $29 = ($26>>>0)<($27>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
    if (!($29)) {
     $$0$i = 1;
     break;
    }
   }
   ;HEAP32[$__t$i$i>>2]=HEAP32[$__x2>>2]|0;HEAP32[$__t$i$i+4>>2]=HEAP32[$__x2+4>>2]|0;HEAP32[$__t$i$i+8>>2]=HEAP32[$__x2+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   ;HEAP32[$__x2>>2]=HEAP32[$__x3>>2]|0;HEAP32[$__x2+4>>2]=HEAP32[$__x3+4>>2]|0;HEAP32[$__x2+8>>2]=HEAP32[$__x3+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   ;HEAP32[$__x3>>2]=HEAP32[$__t$i$i>>2]|0;HEAP32[$__x3+4>>2]=HEAP32[$__t$i$i+4>>2]|0;HEAP32[$__x3+8>>2]=HEAP32[$__t$i$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $$0$i = 2;
  } else {
   if ($$0$i9$i) {
    ;HEAP32[$__t$i8$i>>2]=HEAP32[$__x2>>2]|0;HEAP32[$__t$i8$i+4>>2]=HEAP32[$__x2+4>>2]|0;HEAP32[$__t$i8$i+8>>2]=HEAP32[$__x2+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    ;HEAP32[$__x2>>2]=HEAP32[$__x3>>2]|0;HEAP32[$__x2+4>>2]=HEAP32[$__x3+4>>2]|0;HEAP32[$__x2+8>>2]=HEAP32[$__x3+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    ;HEAP32[$__x3>>2]=HEAP32[$__t$i8$i>>2]|0;HEAP32[$__x3+4>>2]=HEAP32[$__t$i8$i+4>>2]|0;HEAP32[$__x3+8>>2]=HEAP32[$__t$i8$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $17 = HEAP32[$__x2>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
    $18 = HEAP32[$__x1>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
    $19 = ($17|0)==($18|0); //@line 502 "modules/graphics/opengl/Font.cpp"
    if ($19) {
     $21 = ((($__x2)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
     $22 = HEAP32[$21>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
     $23 = ((($__x1)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
     $24 = HEAP32[$23>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
     $25 = ($22|0)<($24|0); //@line 505 "modules/graphics/opengl/Font.cpp"
     if (!($25)) {
      $$0$i = 1;
      break;
     }
    } else {
     $20 = ($17>>>0)<($18>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
     if (!($20)) {
      $$0$i = 1;
      break;
     }
    }
    ;HEAP32[$__t$i5$i>>2]=HEAP32[$__x1>>2]|0;HEAP32[$__t$i5$i+4>>2]=HEAP32[$__x1+4>>2]|0;HEAP32[$__t$i5$i+8>>2]=HEAP32[$__x1+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    ;HEAP32[$__x1>>2]=HEAP32[$__x2>>2]|0;HEAP32[$__x1+4>>2]=HEAP32[$__x2+4>>2]|0;HEAP32[$__x1+8>>2]=HEAP32[$__x2+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    ;HEAP32[$__x2>>2]=HEAP32[$__t$i5$i>>2]|0;HEAP32[$__x2+4>>2]=HEAP32[$__t$i5$i+4>>2]|0;HEAP32[$__x2+8>>2]=HEAP32[$__t$i5$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $$0$i = 2;
   } else {
    $$0$i = 0;
   }
  }
 } while(0);
 $35 = HEAP32[$__x4>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
 $36 = HEAP32[$__x3>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
 $37 = ($35|0)==($36|0); //@line 502 "modules/graphics/opengl/Font.cpp"
 if ($37) {
  $39 = ((($__x4)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $40 = HEAP32[$39>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $41 = ((($__x3)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $42 = HEAP32[$41>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $43 = ($40|0)<($42|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  if (!($43)) {
   $__r$0 = $$0$i;
   STACKTOP = sp;return ($__r$0|0); //@line 3654 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
 } else {
  $38 = ($35>>>0)<($36>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
  if (!($38)) {
   $__r$0 = $$0$i;
   STACKTOP = sp;return ($__r$0|0); //@line 3654 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
 }
 ;HEAP32[$__t$i5>>2]=HEAP32[$__x3>>2]|0;HEAP32[$__t$i5+4>>2]=HEAP32[$__x3+4>>2]|0;HEAP32[$__t$i5+8>>2]=HEAP32[$__x3+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 ;HEAP32[$__x3>>2]=HEAP32[$__x4>>2]|0;HEAP32[$__x3+4>>2]=HEAP32[$__x4+4>>2]|0;HEAP32[$__x3+8>>2]=HEAP32[$__x4+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 ;HEAP32[$__x4>>2]=HEAP32[$__t$i5>>2]|0;HEAP32[$__x4+4>>2]=HEAP32[$__t$i5+4>>2]|0;HEAP32[$__x4+8>>2]=HEAP32[$__t$i5+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $44 = (($$0$i) + 1)|0; //@line 3642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $45 = HEAP32[$__x3>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
 $46 = HEAP32[$__x2>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
 $47 = ($45|0)==($46|0); //@line 502 "modules/graphics/opengl/Font.cpp"
 if ($47) {
  $49 = ((($__x3)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $50 = HEAP32[$49>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $51 = ((($__x2)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $52 = HEAP32[$51>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $53 = ($50|0)<($52|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  if (!($53)) {
   $__r$0 = $44;
   STACKTOP = sp;return ($__r$0|0); //@line 3654 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
 } else {
  $48 = ($45>>>0)<($46>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
  if (!($48)) {
   $__r$0 = $44;
   STACKTOP = sp;return ($__r$0|0); //@line 3654 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
 }
 ;HEAP32[$__t$i2>>2]=HEAP32[$__x2>>2]|0;HEAP32[$__t$i2+4>>2]=HEAP32[$__x2+4>>2]|0;HEAP32[$__t$i2+8>>2]=HEAP32[$__x2+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 ;HEAP32[$__x2>>2]=HEAP32[$__x3>>2]|0;HEAP32[$__x2+4>>2]=HEAP32[$__x3+4>>2]|0;HEAP32[$__x2+8>>2]=HEAP32[$__x3+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 ;HEAP32[$__x3>>2]=HEAP32[$__t$i2>>2]|0;HEAP32[$__x3+4>>2]=HEAP32[$__t$i2+4>>2]|0;HEAP32[$__x3+8>>2]=HEAP32[$__t$i2+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $54 = (($$0$i) + 2)|0; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $55 = HEAP32[$__x2>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
 $56 = HEAP32[$__x1>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
 $57 = ($55|0)==($56|0); //@line 502 "modules/graphics/opengl/Font.cpp"
 if ($57) {
  $59 = ((($__x2)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $60 = HEAP32[$59>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $61 = ((($__x1)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $62 = HEAP32[$61>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $63 = ($60|0)<($62|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  if (!($63)) {
   $__r$0 = $54;
   STACKTOP = sp;return ($__r$0|0); //@line 3654 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
 } else {
  $58 = ($55>>>0)<($56>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
  if (!($58)) {
   $__r$0 = $54;
   STACKTOP = sp;return ($__r$0|0); //@line 3654 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
 }
 ;HEAP32[$__t$i>>2]=HEAP32[$__x1>>2]|0;HEAP32[$__t$i+4>>2]=HEAP32[$__x1+4>>2]|0;HEAP32[$__t$i+8>>2]=HEAP32[$__x1+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 ;HEAP32[$__x1>>2]=HEAP32[$__x2>>2]|0;HEAP32[$__x1+4>>2]=HEAP32[$__x2+4>>2]|0;HEAP32[$__x1+8>>2]=HEAP32[$__x2+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 ;HEAP32[$__x2>>2]=HEAP32[$__t$i>>2]|0;HEAP32[$__x2+4>>2]=HEAP32[$__t$i+4>>2]|0;HEAP32[$__x2+8>>2]=HEAP32[$__t$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $64 = (($$0$i) + 3)|0; //@line 3650 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $__r$0 = $64;
 STACKTOP = sp;return ($__r$0|0); //@line 3654 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
}
function __ZNSt3__17__sort5IRZN4love8graphics6opengl4Font16generateVerticesERKNS4_17ColoredCodepointsERNS_6vectorINS4_11GlyphVertexENS_9allocatorIS9_EEEEfNS1_6VectorEPNS4_8TextInfoEE3__0PNS4_11DrawCommandEEEjT0_SL_SL_SL_SL_T_($__x1,$__x2,$__x3,$__x4,$__x5) {
 $__x1 = $__x1|0;
 $__x2 = $__x2|0;
 $__x3 = $__x3|0;
 $__x4 = $__x4|0;
 $__x5 = $__x5|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $__r$0 = 0, $__t$i = 0, $__t$i3 = 0, $__t$i6 = 0, $__t$i9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__t$i9 = sp + 36|0;
 $__t$i6 = sp + 24|0;
 $__t$i3 = sp + 12|0;
 $__t$i = sp;
 $0 = (__ZNSt3__17__sort4IRZN4love8graphics6opengl4Font16generateVerticesERKNS4_17ColoredCodepointsERNS_6vectorINS4_11GlyphVertexENS_9allocatorIS9_EEEEfNS1_6VectorEPNS4_8TextInfoEE3__0PNS4_11DrawCommandEEEjT0_SL_SL_SL_T_($__x1,$__x2,$__x3,$__x4)|0); //@line 3664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $1 = HEAP32[$__x5>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
 $2 = HEAP32[$__x4>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
 $3 = ($1|0)==($2|0); //@line 502 "modules/graphics/opengl/Font.cpp"
 if ($3) {
  $5 = ((($__x5)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $7 = ((($__x4)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $8 = HEAP32[$7>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $9 = ($6|0)<($8|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  if (!($9)) {
   $__r$0 = $0;
   STACKTOP = sp;return ($__r$0|0); //@line 3685 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
 } else {
  $4 = ($1>>>0)<($2>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
  if (!($4)) {
   $__r$0 = $0;
   STACKTOP = sp;return ($__r$0|0); //@line 3685 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
 }
 ;HEAP32[$__t$i9>>2]=HEAP32[$__x4>>2]|0;HEAP32[$__t$i9+4>>2]=HEAP32[$__x4+4>>2]|0;HEAP32[$__t$i9+8>>2]=HEAP32[$__x4+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 ;HEAP32[$__x4>>2]=HEAP32[$__x5>>2]|0;HEAP32[$__x4+4>>2]=HEAP32[$__x5+4>>2]|0;HEAP32[$__x4+8>>2]=HEAP32[$__x5+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 ;HEAP32[$__x5>>2]=HEAP32[$__t$i9>>2]|0;HEAP32[$__x5+4>>2]=HEAP32[$__t$i9+4>>2]|0;HEAP32[$__x5+8>>2]=HEAP32[$__t$i9+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $10 = (($0) + 1)|0; //@line 3668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $11 = HEAP32[$__x4>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
 $12 = HEAP32[$__x3>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
 $13 = ($11|0)==($12|0); //@line 502 "modules/graphics/opengl/Font.cpp"
 if ($13) {
  $15 = ((($__x4)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $17 = ((($__x3)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $18 = HEAP32[$17>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $19 = ($16|0)<($18|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  if (!($19)) {
   $__r$0 = $10;
   STACKTOP = sp;return ($__r$0|0); //@line 3685 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
 } else {
  $14 = ($11>>>0)<($12>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
  if (!($14)) {
   $__r$0 = $10;
   STACKTOP = sp;return ($__r$0|0); //@line 3685 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
 }
 ;HEAP32[$__t$i6>>2]=HEAP32[$__x3>>2]|0;HEAP32[$__t$i6+4>>2]=HEAP32[$__x3+4>>2]|0;HEAP32[$__t$i6+8>>2]=HEAP32[$__x3+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 ;HEAP32[$__x3>>2]=HEAP32[$__x4>>2]|0;HEAP32[$__x3+4>>2]=HEAP32[$__x4+4>>2]|0;HEAP32[$__x3+8>>2]=HEAP32[$__x4+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 ;HEAP32[$__x4>>2]=HEAP32[$__t$i6>>2]|0;HEAP32[$__x4+4>>2]=HEAP32[$__t$i6+4>>2]|0;HEAP32[$__x4+8>>2]=HEAP32[$__t$i6+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $20 = (($0) + 2)|0; //@line 3672 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $21 = HEAP32[$__x3>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
 $22 = HEAP32[$__x2>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
 $23 = ($21|0)==($22|0); //@line 502 "modules/graphics/opengl/Font.cpp"
 if ($23) {
  $25 = ((($__x3)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $26 = HEAP32[$25>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $27 = ((($__x2)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $28 = HEAP32[$27>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $29 = ($26|0)<($28|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  if (!($29)) {
   $__r$0 = $20;
   STACKTOP = sp;return ($__r$0|0); //@line 3685 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
 } else {
  $24 = ($21>>>0)<($22>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
  if (!($24)) {
   $__r$0 = $20;
   STACKTOP = sp;return ($__r$0|0); //@line 3685 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
 }
 ;HEAP32[$__t$i3>>2]=HEAP32[$__x2>>2]|0;HEAP32[$__t$i3+4>>2]=HEAP32[$__x2+4>>2]|0;HEAP32[$__t$i3+8>>2]=HEAP32[$__x2+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 ;HEAP32[$__x2>>2]=HEAP32[$__x3>>2]|0;HEAP32[$__x2+4>>2]=HEAP32[$__x3+4>>2]|0;HEAP32[$__x2+8>>2]=HEAP32[$__x3+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 ;HEAP32[$__x3>>2]=HEAP32[$__t$i3>>2]|0;HEAP32[$__x3+4>>2]=HEAP32[$__t$i3+4>>2]|0;HEAP32[$__x3+8>>2]=HEAP32[$__t$i3+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $30 = (($0) + 3)|0; //@line 3676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $31 = HEAP32[$__x2>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
 $32 = HEAP32[$__x1>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
 $33 = ($31|0)==($32|0); //@line 502 "modules/graphics/opengl/Font.cpp"
 if ($33) {
  $35 = ((($__x2)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $36 = HEAP32[$35>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $37 = ((($__x1)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  $38 = HEAP32[$37>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
  $39 = ($36|0)<($38|0); //@line 505 "modules/graphics/opengl/Font.cpp"
  if (!($39)) {
   $__r$0 = $30;
   STACKTOP = sp;return ($__r$0|0); //@line 3685 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
 } else {
  $34 = ($31>>>0)<($32>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
  if (!($34)) {
   $__r$0 = $30;
   STACKTOP = sp;return ($__r$0|0); //@line 3685 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
 }
 ;HEAP32[$__t$i>>2]=HEAP32[$__x1>>2]|0;HEAP32[$__t$i+4>>2]=HEAP32[$__x1+4>>2]|0;HEAP32[$__t$i+8>>2]=HEAP32[$__x1+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 ;HEAP32[$__x1>>2]=HEAP32[$__x2>>2]|0;HEAP32[$__x1+4>>2]=HEAP32[$__x2+4>>2]|0;HEAP32[$__x1+8>>2]=HEAP32[$__x2+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 ;HEAP32[$__x2>>2]=HEAP32[$__t$i>>2]|0;HEAP32[$__x2+4>>2]=HEAP32[$__t$i+4>>2]|0;HEAP32[$__x2+8>>2]=HEAP32[$__t$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $40 = (($0) + 4)|0; //@line 3680 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $__r$0 = $40;
 STACKTOP = sp;return ($__r$0|0); //@line 3685 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
}
function __ZNSt3__127__insertion_sort_incompleteIRZN4love8graphics6opengl4Font16generateVerticesERKNS4_17ColoredCodepointsERNS_6vectorINS4_11GlyphVertexENS_9allocatorIS9_EEEEfNS1_6VectorEPNS4_8TextInfoEE3__0PNS4_11DrawCommandEEEbT0_SL_T_($__first,$__last) {
 $__first = $__first|0;
 $__last = $__last|0;
 var $$$3 = 0, $$0$i$i = 0, $$0$i$i11 = 0, $$0$i9$i = 0, $$0$i9$i13 = 0, $$1 = 0, $$3 = 0, $$5 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__count$025 = 0, $__count$1 = 0, $__i$026 = 0, $__j$023 = 0, $__j$023$phi = 0, $__k$0$lcssa = 0;
 var $__k$019 = 0, $__t$i = 0, $__t$i$i = 0, $__t$i$i10 = 0, $__t$i3$i = 0, $__t$i3$i9 = 0, $__t$i4$i = 0, $__t$i4$i8 = 0, $__t$i5$i = 0, $__t$i5$i7 = 0, $__t$i8$i = 0, $__t$i8$i6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__t$i8$i6 = sp + 120|0;
 $__t$i5$i7 = sp + 108|0;
 $__t$i4$i8 = sp + 96|0;
 $__t$i3$i9 = sp + 84|0;
 $__t$i$i10 = sp + 72|0;
 $__t$i = sp + 60|0;
 $__t$i8$i = sp + 48|0;
 $__t$i5$i = sp + 36|0;
 $__t$i4$i = sp + 24|0;
 $__t$i3$i = sp + 12|0;
 $__t$i$i = sp;
 $0 = $__last; //@line 3752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $1 = $__first; //@line 3752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $2 = (($0) - ($1))|0; //@line 3752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $3 = (($2|0) / 12)&-1; //@line 3752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 switch ($3|0) {
 case 2:  {
  $4 = ((($__last)) + -12|0); //@line 3758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $5 = HEAP32[$4>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $6 = HEAP32[$__first>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $7 = ($5|0)==($6|0); //@line 502 "modules/graphics/opengl/Font.cpp"
  if ($7) {
   $9 = ((($__last)) + -8|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $10 = HEAP32[$9>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $11 = ((($__first)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $12 = HEAP32[$11>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $13 = ($10|0)<($12|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   if (!($13)) {
    $$5 = 1;
    STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   }
  } else {
   $8 = ($5>>>0)<($6>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
   if (!($8)) {
    $$5 = 1;
    STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   }
  }
  ;HEAP32[$__t$i>>2]=HEAP32[$__first>>2]|0;HEAP32[$__t$i+4>>2]=HEAP32[$__first+4>>2]|0;HEAP32[$__t$i+8>>2]=HEAP32[$__first+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  ;HEAP32[$__first>>2]=HEAP32[$4>>2]|0;HEAP32[$__first+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$__first+8>>2]=HEAP32[$4+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  ;HEAP32[$4>>2]=HEAP32[$__t$i>>2]|0;HEAP32[$4+4>>2]=HEAP32[$__t$i+4>>2]|0;HEAP32[$4+8>>2]=HEAP32[$__t$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $$5 = 1;
  STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  break;
 }
 case 3:  {
  $14 = ((($__first)) + 12|0); //@line 3762 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $15 = ((($__last)) + -12|0); //@line 3762 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $16 = HEAP32[$14>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $17 = HEAP32[$__first>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $18 = ($16|0)==($17|0); //@line 502 "modules/graphics/opengl/Font.cpp"
  if ($18) {
   $20 = ((($__first)) + 16|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $21 = HEAP32[$20>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $22 = ((($__first)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $23 = HEAP32[$22>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $24 = ($21|0)<($23|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $$0$i$i11 = $24;
  } else {
   $19 = ($16>>>0)<($17>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
   $$0$i$i11 = $19;
  }
  $25 = HEAP32[$15>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $26 = ($25|0)==($16|0); //@line 502 "modules/graphics/opengl/Font.cpp"
  if ($26) {
   $28 = ((($__last)) + -8|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $29 = HEAP32[$28>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $30 = ((($__first)) + 16|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $31 = HEAP32[$30>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $32 = ($29|0)<($31|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $$0$i9$i13 = $32;
  } else {
   $27 = ($25>>>0)<($16>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
   $$0$i9$i13 = $27;
  }
  if (!($$0$i$i11)) {
   if (!($$0$i9$i13)) {
    $$5 = 1;
    STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   }
   ;HEAP32[$__t$i8$i6>>2]=HEAP32[$14>>2]|0;HEAP32[$__t$i8$i6+4>>2]=HEAP32[$14+4>>2]|0;HEAP32[$__t$i8$i6+8>>2]=HEAP32[$14+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   ;HEAP32[$14>>2]=HEAP32[$15>>2]|0;HEAP32[$14+4>>2]=HEAP32[$15+4>>2]|0;HEAP32[$14+8>>2]=HEAP32[$15+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   ;HEAP32[$15>>2]=HEAP32[$__t$i8$i6>>2]|0;HEAP32[$15+4>>2]=HEAP32[$__t$i8$i6+4>>2]|0;HEAP32[$15+8>>2]=HEAP32[$__t$i8$i6+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $33 = HEAP32[$14>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
   $34 = HEAP32[$__first>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
   $35 = ($33|0)==($34|0); //@line 502 "modules/graphics/opengl/Font.cpp"
   if ($35) {
    $37 = ((($__first)) + 16|0); //@line 505 "modules/graphics/opengl/Font.cpp"
    $38 = HEAP32[$37>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
    $39 = ((($__first)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
    $40 = HEAP32[$39>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
    $41 = ($38|0)<($40|0); //@line 505 "modules/graphics/opengl/Font.cpp"
    if (!($41)) {
     $$5 = 1;
     STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    }
   } else {
    $36 = ($33>>>0)<($34>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
    if (!($36)) {
     $$5 = 1;
     STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    }
   }
   ;HEAP32[$__t$i5$i7>>2]=HEAP32[$__first>>2]|0;HEAP32[$__t$i5$i7+4>>2]=HEAP32[$__first+4>>2]|0;HEAP32[$__t$i5$i7+8>>2]=HEAP32[$__first+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   ;HEAP32[$__first>>2]=HEAP32[$14>>2]|0;HEAP32[$__first+4>>2]=HEAP32[$14+4>>2]|0;HEAP32[$__first+8>>2]=HEAP32[$14+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   ;HEAP32[$14>>2]=HEAP32[$__t$i5$i7>>2]|0;HEAP32[$14+4>>2]=HEAP32[$__t$i5$i7+4>>2]|0;HEAP32[$14+8>>2]=HEAP32[$__t$i5$i7+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $$5 = 1;
   STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  if ($$0$i9$i13) {
   ;HEAP32[$__t$i4$i8>>2]=HEAP32[$__first>>2]|0;HEAP32[$__t$i4$i8+4>>2]=HEAP32[$__first+4>>2]|0;HEAP32[$__t$i4$i8+8>>2]=HEAP32[$__first+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   ;HEAP32[$__first>>2]=HEAP32[$15>>2]|0;HEAP32[$__first+4>>2]=HEAP32[$15+4>>2]|0;HEAP32[$__first+8>>2]=HEAP32[$15+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   ;HEAP32[$15>>2]=HEAP32[$__t$i4$i8>>2]|0;HEAP32[$15+4>>2]=HEAP32[$__t$i4$i8+4>>2]|0;HEAP32[$15+8>>2]=HEAP32[$__t$i4$i8+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $$5 = 1;
   STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  ;HEAP32[$__t$i3$i9>>2]=HEAP32[$__first>>2]|0;HEAP32[$__t$i3$i9+4>>2]=HEAP32[$__first+4>>2]|0;HEAP32[$__t$i3$i9+8>>2]=HEAP32[$__first+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  ;HEAP32[$__first>>2]=HEAP32[$14>>2]|0;HEAP32[$__first+4>>2]=HEAP32[$14+4>>2]|0;HEAP32[$__first+8>>2]=HEAP32[$14+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  ;HEAP32[$14>>2]=HEAP32[$__t$i3$i9>>2]|0;HEAP32[$14+4>>2]=HEAP32[$__t$i3$i9+4>>2]|0;HEAP32[$14+8>>2]=HEAP32[$__t$i3$i9+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $42 = HEAP32[$15>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $43 = HEAP32[$14>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $44 = ($42|0)==($43|0); //@line 502 "modules/graphics/opengl/Font.cpp"
  if ($44) {
   $46 = ((($__last)) + -8|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $47 = HEAP32[$46>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $48 = ((($__first)) + 16|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $49 = HEAP32[$48>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $50 = ($47|0)<($49|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   if (!($50)) {
    $$5 = 1;
    STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   }
  } else {
   $45 = ($42>>>0)<($43>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
   if (!($45)) {
    $$5 = 1;
    STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   }
  }
  ;HEAP32[$__t$i$i10>>2]=HEAP32[$14>>2]|0;HEAP32[$__t$i$i10+4>>2]=HEAP32[$14+4>>2]|0;HEAP32[$__t$i$i10+8>>2]=HEAP32[$14+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  ;HEAP32[$14>>2]=HEAP32[$15>>2]|0;HEAP32[$14+4>>2]=HEAP32[$15+4>>2]|0;HEAP32[$14+8>>2]=HEAP32[$15+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  ;HEAP32[$15>>2]=HEAP32[$__t$i$i10>>2]|0;HEAP32[$15+4>>2]=HEAP32[$__t$i$i10+4>>2]|0;HEAP32[$15+8>>2]=HEAP32[$__t$i$i10+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $$5 = 1;
  STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  break;
 }
 case 4:  {
  $51 = ((($__first)) + 12|0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $52 = ((($__first)) + 24|0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $53 = ((($__last)) + -12|0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  (__ZNSt3__17__sort4IRZN4love8graphics6opengl4Font16generateVerticesERKNS4_17ColoredCodepointsERNS_6vectorINS4_11GlyphVertexENS_9allocatorIS9_EEEEfNS1_6VectorEPNS4_8TextInfoEE3__0PNS4_11DrawCommandEEEjT0_SL_SL_SL_T_($__first,$51,$52,$53)|0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $$5 = 1;
  STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  break;
 }
 case 5:  {
  $54 = ((($__first)) + 12|0); //@line 3768 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $55 = ((($__first)) + 24|0); //@line 3768 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $56 = ((($__first)) + 36|0); //@line 3768 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $57 = ((($__last)) + -12|0); //@line 3768 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  (__ZNSt3__17__sort5IRZN4love8graphics6opengl4Font16generateVerticesERKNS4_17ColoredCodepointsERNS_6vectorINS4_11GlyphVertexENS_9allocatorIS9_EEEEfNS1_6VectorEPNS4_8TextInfoEE3__0PNS4_11DrawCommandEEEjT0_SL_SL_SL_SL_T_($__first,$54,$55,$56,$57)|0); //@line 3768 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $$5 = 1;
  STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  break;
 }
 case 1: case 0:  {
  $$5 = 1;
  STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  break;
 }
 default: {
  $58 = ((($__first)) + 24|0); //@line 3772 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $59 = ((($__first)) + 12|0); //@line 3773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $60 = HEAP32[$59>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $61 = HEAP32[$__first>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $62 = ($60|0)==($61|0); //@line 502 "modules/graphics/opengl/Font.cpp"
  if ($62) {
   $64 = ((($__first)) + 16|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $65 = HEAP32[$64>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $66 = ((($__first)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $67 = HEAP32[$66>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $68 = ($65|0)<($67|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $$0$i$i = $68;
  } else {
   $63 = ($60>>>0)<($61>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
   $$0$i$i = $63;
  }
  $69 = HEAP32[$58>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
  $70 = ($69|0)==($60|0); //@line 502 "modules/graphics/opengl/Font.cpp"
  if ($70) {
   $72 = ((($__first)) + 28|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $73 = HEAP32[$72>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $74 = ((($__first)) + 16|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $75 = HEAP32[$74>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
   $76 = ($73|0)<($75|0); //@line 505 "modules/graphics/opengl/Font.cpp"
   $$0$i9$i = $76;
  } else {
   $71 = ($69>>>0)<($60>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
   $$0$i9$i = $71;
  }
  do {
   if ($$0$i$i) {
    if ($$0$i9$i) {
     ;HEAP32[$__t$i4$i>>2]=HEAP32[$__first>>2]|0;HEAP32[$__t$i4$i+4>>2]=HEAP32[$__first+4>>2]|0;HEAP32[$__t$i4$i+8>>2]=HEAP32[$__first+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$__first>>2]=HEAP32[$58>>2]|0;HEAP32[$__first+4>>2]=HEAP32[$58+4>>2]|0;HEAP32[$__first+8>>2]=HEAP32[$58+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$58>>2]=HEAP32[$__t$i4$i>>2]|0;HEAP32[$58+4>>2]=HEAP32[$__t$i4$i+4>>2]|0;HEAP32[$58+8>>2]=HEAP32[$__t$i4$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     break;
    }
    ;HEAP32[$__t$i3$i>>2]=HEAP32[$__first>>2]|0;HEAP32[$__t$i3$i+4>>2]=HEAP32[$__first+4>>2]|0;HEAP32[$__t$i3$i+8>>2]=HEAP32[$__first+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    ;HEAP32[$__first>>2]=HEAP32[$59>>2]|0;HEAP32[$__first+4>>2]=HEAP32[$59+4>>2]|0;HEAP32[$__first+8>>2]=HEAP32[$59+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    ;HEAP32[$59>>2]=HEAP32[$__t$i3$i>>2]|0;HEAP32[$59+4>>2]=HEAP32[$__t$i3$i+4>>2]|0;HEAP32[$59+8>>2]=HEAP32[$__t$i3$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $85 = HEAP32[$59>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
    $86 = ($69|0)==($85|0); //@line 502 "modules/graphics/opengl/Font.cpp"
    if ($86) {
     $88 = ((($__first)) + 28|0); //@line 505 "modules/graphics/opengl/Font.cpp"
     $89 = HEAP32[$88>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
     $90 = ((($__first)) + 16|0); //@line 505 "modules/graphics/opengl/Font.cpp"
     $91 = HEAP32[$90>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
     $92 = ($89|0)<($91|0); //@line 505 "modules/graphics/opengl/Font.cpp"
     if (!($92)) {
      break;
     }
    } else {
     $87 = ($69>>>0)<($85>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
     if (!($87)) {
      break;
     }
    }
    ;HEAP32[$__t$i$i>>2]=HEAP32[$59>>2]|0;HEAP32[$__t$i$i+4>>2]=HEAP32[$59+4>>2]|0;HEAP32[$__t$i$i+8>>2]=HEAP32[$59+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    ;HEAP32[$59>>2]=HEAP32[$58>>2]|0;HEAP32[$59+4>>2]=HEAP32[$58+4>>2]|0;HEAP32[$59+8>>2]=HEAP32[$58+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    ;HEAP32[$58>>2]=HEAP32[$__t$i$i>>2]|0;HEAP32[$58+4>>2]=HEAP32[$__t$i$i+4>>2]|0;HEAP32[$58+8>>2]=HEAP32[$__t$i$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   } else {
    if ($$0$i9$i) {
     ;HEAP32[$__t$i8$i>>2]=HEAP32[$59>>2]|0;HEAP32[$__t$i8$i+4>>2]=HEAP32[$59+4>>2]|0;HEAP32[$__t$i8$i+8>>2]=HEAP32[$59+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$59>>2]=HEAP32[$58>>2]|0;HEAP32[$59+4>>2]=HEAP32[$58+4>>2]|0;HEAP32[$59+8>>2]=HEAP32[$58+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$58>>2]=HEAP32[$__t$i8$i>>2]|0;HEAP32[$58+4>>2]=HEAP32[$__t$i8$i+4>>2]|0;HEAP32[$58+8>>2]=HEAP32[$__t$i8$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $77 = HEAP32[$59>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
     $78 = ($77|0)==($61|0); //@line 502 "modules/graphics/opengl/Font.cpp"
     if ($78) {
      $80 = ((($__first)) + 16|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      $81 = HEAP32[$80>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
      $82 = ((($__first)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      $83 = HEAP32[$82>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
      $84 = ($81|0)<($83|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      if (!($84)) {
       break;
      }
     } else {
      $79 = ($77>>>0)<($61>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
      if (!($79)) {
       break;
      }
     }
     ;HEAP32[$__t$i5$i>>2]=HEAP32[$__first>>2]|0;HEAP32[$__t$i5$i+4>>2]=HEAP32[$__first+4>>2]|0;HEAP32[$__t$i5$i+8>>2]=HEAP32[$__first+8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$__first>>2]=HEAP32[$59>>2]|0;HEAP32[$__first+4>>2]=HEAP32[$59+4>>2]|0;HEAP32[$__first+8>>2]=HEAP32[$59+8>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     ;HEAP32[$59>>2]=HEAP32[$__t$i5$i>>2]|0;HEAP32[$59+4>>2]=HEAP32[$__t$i5$i+4>>2]|0;HEAP32[$59+8>>2]=HEAP32[$__t$i5$i+8>>2]|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    }
   }
  } while(0);
  $93 = ((($__first)) + 36|0); //@line 3776 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $94 = ($93|0)==($__last|0); //@line 3776 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  L27: do {
   if ($94) {
    $$1 = 1;$$3 = 0;
   } else {
    $__count$025 = 0;$__i$026 = $93;$__j$023 = $58;
    while(1) {
     $95 = HEAP32[$__i$026>>2]|0; //@line 3780 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $96 = HEAP32[$__j$023>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
     $97 = ($95|0)==($96|0); //@line 502 "modules/graphics/opengl/Font.cpp"
     if ($97) {
      $99 = ((($__i$026)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      $100 = HEAP32[$99>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
      $101 = ((($__j$023)) + 4|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      $102 = HEAP32[$101>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
      $103 = ($100|0)<($102|0); //@line 505 "modules/graphics/opengl/Font.cpp"
      if ($103) {
       $114 = $100;
       label = 49;
      } else {
       $__count$1 = $__count$025;
      }
     } else {
      $98 = ($95>>>0)<($96>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
      if ($98) {
       $$phi$trans$insert = ((($__i$026)) + 4|0); //@line 3780 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 3780 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $114 = $$pre;
       label = 49;
      } else {
       $__count$1 = $__count$025;
      }
     }
     if ((label|0) == 49) {
      label = 0;
      $104 = ((($__i$026)) + 8|0); //@line 3780 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $105 = HEAP32[$104>>2]|0; //@line 3780 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      ;HEAP32[$__i$026>>2]=HEAP32[$__j$023>>2]|0;HEAP32[$__i$026+4>>2]=HEAP32[$__j$023+4>>2]|0;HEAP32[$__i$026+8>>2]=HEAP32[$__j$023+8>>2]|0; //@line 3785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $106 = ($__j$023|0)==($__first|0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      L36: do {
       if ($106) {
        $__k$0$lcssa = $__first;
       } else {
        $__k$019 = $__j$023;
        while(1) {
         $107 = ((($__k$019)) + -12|0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $108 = HEAP32[$107>>2]|0; //@line 502 "modules/graphics/opengl/Font.cpp"
         $109 = ($95|0)==($108|0); //@line 502 "modules/graphics/opengl/Font.cpp"
         if ($109) {
          $111 = ((($__k$019)) + -8|0); //@line 505 "modules/graphics/opengl/Font.cpp"
          $112 = HEAP32[$111>>2]|0; //@line 505 "modules/graphics/opengl/Font.cpp"
          $113 = ($114|0)<($112|0); //@line 505 "modules/graphics/opengl/Font.cpp"
          if (!($113)) {
           $__k$0$lcssa = $__k$019;
           break L36;
          }
         } else {
          $110 = ($95>>>0)<($108>>>0); //@line 503 "modules/graphics/opengl/Font.cpp"
          if (!($110)) {
           $__k$0$lcssa = $__k$019;
           break L36;
          }
         }
         ;HEAP32[$__k$019>>2]=HEAP32[$107>>2]|0;HEAP32[$__k$019+4>>2]=HEAP32[$107+4>>2]|0;HEAP32[$__k$019+8>>2]=HEAP32[$107+8>>2]|0; //@line 3785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $115 = ($107|0)==($__first|0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         if ($115) {
          $__k$0$lcssa = $__first;
          break;
         } else {
          $__k$019 = $107;
         }
        }
       }
      } while(0);
      HEAP32[$__k$0$lcssa>>2] = $95; //@line 3788 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $116 = ((($__k$0$lcssa)) + 4|0); //@line 3788 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      HEAP32[$116>>2] = $114; //@line 3788 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $117 = ((($__k$0$lcssa)) + 8|0); //@line 3788 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      HEAP32[$117>>2] = $105; //@line 3788 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $118 = (($__count$025) + 1)|0; //@line 3789 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $119 = ($118|0)==(8); //@line 3789 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $120 = ((($__i$026)) + 12|0); //@line 3790 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $121 = ($120|0)==($__last|0); //@line 3790 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if ($119) {
       $$1 = 0;$$3 = $121;
       break L27;
      } else {
       $__count$1 = $118;
      }
     }
     $122 = ((($__i$026)) + 12|0); //@line 3776 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $123 = ($122|0)==($__last|0); //@line 3776 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     if ($123) {
      $$1 = 1;$$3 = 0;
      break;
     } else {
      $__j$023$phi = $__i$026;$__count$025 = $__count$1;$__i$026 = $122;$__j$023 = $__j$023$phi;
     }
    }
   }
  } while(0);
  $$$3 = $$3 | $$1;
  $$5 = $$$3;
  STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
 }
 return (0)|0;
}
function __ZNK4love8graphics6opengl4Font11getBaselineEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 92|0); //@line 994 "modules/graphics/opengl/Font.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 994 "modules/graphics/opengl/Font.cpp"
 $2 = ($1|0)==(0); //@line 994 "modules/graphics/opengl/Font.cpp"
 if (!($2)) {
  $9 = 0.0;
  return (+$9); //@line 994 "modules/graphics/opengl/Font.cpp"
 }
 $3 = ((($this)) + 24|0); //@line 398 "modules/graphics/opengl/Font.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 398 "modules/graphics/opengl/Font.cpp"
 $5 = (+($4|0)); //@line 398 "modules/graphics/opengl/Font.cpp"
 $6 = $5 / 1.25; //@line 994 "modules/graphics/opengl/Font.cpp"
 $7 = $6 + 0.5; //@line 994 "modules/graphics/opengl/Font.cpp"
 $8 = (+Math_floor((+$7))); //@line 994 "modules/graphics/opengl/Font.cpp"
 $9 = $8;
 return (+$9); //@line 994 "modules/graphics/opengl/Font.cpp"
}
function __ZNK4love8graphics6opengl4Font13getLineHeightEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 937 "modules/graphics/opengl/Font.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 937 "modules/graphics/opengl/Font.cpp"
 return (+$1); //@line 937 "modules/graphics/opengl/Font.cpp"
}
function __ZN4love8graphics6opengl4Font25generateVerticesFormattedERKNS2_17ColoredCodepointsEfNS2_9AlignModeERNSt3__16vectorINS2_11GlyphVertexENS7_9allocatorIS9_EEEEPNS2_8TextInfoE($agg$result,$this,$text,$wrap,$align,$vertices,$info) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $text = $text|0;
 $wrap = +$wrap;
 $align = $align|0;
 $vertices = $vertices|0;
 $info = $info|0;
 var $$0 = 0, $$02 = 0, $$1 = 0, $$13 = 0, $$__r$0$i = 0, $$__r$0$i$lcssa = 0, $$byval_copy = 0, $$byval_copy6 = 0, $$byval_copy7 = 0, $$pre = 0, $$pre$i = 0, $$pre$i23 = 0, $$pre$phi2$i$iZ2D = 0, $$pre1$i$i = 0, $$pre59 = 0, $$sroa$050 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0.0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0.0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0;
 var $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0;
 var $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0, $73 = 0.0, $74 = 0, $75 = 0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0, $__new_last1$i$i$i$i$i$i$i$i$i$i = 0, $__new_last1$i$i$i$i$i$i$i$i$i$i11 = 0, $__new_last2$i$i$i = 0, $__r$01$i = 0, $extraspacing$1 = 0.0, $firstcmd$sroa$0$0 = 0;
 var $i$056 = 0, $lines = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $maxwidth$sroa$0 = 0, $newcommands = 0, $offset$sroa$0$0 = 0, $or$cond = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i$i$i$i = 0, $scevgep$i$i$i$i$i$i$i$i$i$i = 0, $scevgep$i$i$i$i$i$i$i$i$i$i12 = 0, $scevgep$i$i$i1$i$i$i$i$i$i$i = 0, $scevgep$i$i$i1$i$i$i$i$i$i$i18 = 0, $scevgep$i$i$i30 = 0, $scevgep$i$i$i35 = 0, $scevgep$i$i$i38 = 0, $scevgep$i$i$i43 = 0, $scevgep$i$i$i5 = 0, $scevgep4$i$i$i = 0;
 var $scevgep4$i$i$i$i$i$i = 0, $scevgep4$i$i$i$i$i$i$i$i$i$i = 0, $scevgep4$i$i$i$i$i$i$i$i$i$i13 = 0, $scevgep4$i$i$i2$i$i$i$i$i$i$i = 0, $scevgep4$i$i$i2$i$i$i$i$i$i$i19 = 0, $scevgep4$i$i$i31 = 0, $scevgep4$i$i$i36 = 0, $scevgep4$i$i$i39 = 0, $scevgep4$i$i$i44 = 0, $scevgep4$i$i$i6 = 0, $tmpcast$byval_copy = 0, $width = 0, $widths = 0, $y$0$lcssa = 0.0, $y$057 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy7 = sp + 104|0;
 $$byval_copy6 = sp + 100|0;
 $$byval_copy = sp + 96|0;
 $tmpcast$byval_copy = sp + 88|0;
 $$sroa$050 = sp + 80|0;
 $0 = sp + 76|0;
 $widths = sp + 64|0;
 $lines = sp + 52|0;
 $maxwidth$sroa$0 = sp + 48|0;
 $width = sp + 44|0;
 $newcommands = sp + 32|0;
 $1 = sp;
 $2 = sp + 28|0;
 $3 = sp + 24|0;
 $4 = sp + 20|0;
 $5 = sp + 8|0;
 HEAPF32[$$sroa$050>>2] = $wrap;
 HEAPF32[$0>>2] = 0.0; //@line 532 "modules/graphics/opengl/Font.cpp"
 $6 = $wrap < 0.0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $7 = $6 ? $0 : $$sroa$050; //@line 2648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $8 = HEAP32[$7>>2]|0; //@line 532 "modules/graphics/opengl/Font.cpp"
 HEAP32[$$sroa$050>>2] = $8; //@line 532 "modules/graphics/opengl/Font.cpp"
 $9 = ((($this)) + 132|0); //@line 534 "modules/graphics/opengl/Font.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 534 "modules/graphics/opengl/Font.cpp"
 HEAP32[$agg$result>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ((($agg$result)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$11>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ((($agg$result)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$12>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $13 = ((($text)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = HEAP32[$13>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = HEAP32[$text>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = (($14) - ($15))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ((($vertices)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $18 = HEAP32[$17>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $19 = HEAP32[$vertices>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = (($18) - ($19))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = $20 >> 4; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $22 = ($21>>>0)<($16>>>0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $23 = (HEAP32[tempDoublePtr>>2]=$8,+HEAPF32[tempDoublePtr>>2]); //@line 542 "modules/graphics/opengl/Font.cpp"
 L1: do {
  if ($22) {
   $24 = ((($vertices)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $25 = HEAP32[$24>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $26 = (($25) - ($19))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $27 = $26 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $28 = ($14|0)==($15|0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   do {
    if ($28) {
     $34 = 0;
    } else {
     $29 = $16 << 4; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     __THREW__ = 0;
     $30 = (invoke_ii(3348,($29|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $31 = __THREW__; __THREW__ = 0;
     $32 = $31&1;
     if ($32) {
      $59 = ___cxa_find_matching_catch_2()|0;
      $60 = tempRet0;
      $$1 = $59;$$13 = $60;
      break L1;
     } else {
      $34 = $30;
      break;
     }
    }
   } while(0);
   $33 = (($34) + ($27<<4)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $35 = $33; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $36 = (($34) + ($16<<4)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $37 = $36; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $38 = (0 - ($27))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $39 = (($33) + ($38<<4)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $40 = $39; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $41 = $19; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   _memcpy(($39|0),($41|0),($26|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$vertices>>2] = $40; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$24>>2] = $35; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$17>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $42 = ($19|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($42) {
    label = 7;
   } else {
    __ZdlPv($41); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    label = 7;
   }
  } else {
   label = 7;
  }
 } while(0);
 if ((label|0) == 7) {
  HEAP32[$widths>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $43 = ((($widths)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$43>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $44 = ((($widths)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$44>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$lines>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $45 = ((($lines)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$45>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $46 = ((($lines)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$46>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  __THREW__ = 0;
  invoke_viidii(4028,($this|0),($text|0),(+$23),($lines|0),($widths|0)); //@line 542 "modules/graphics/opengl/Font.cpp"
  $47 = __THREW__; __THREW__ = 0;
  $48 = $47&1;
  L12: do {
   if ($48) {
    label = 13;
   } else {
    HEAP32[$maxwidth$sroa$0>>2] = 0; //@line 545 "modules/graphics/opengl/Font.cpp"
    $49 = HEAP32[$45>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $50 = HEAP32[$lines>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $51 = (($49) - ($50))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $52 = ($51|0)>(0); //@line 547 "modules/graphics/opengl/Font.cpp"
    $53 = $50; //@line 547 "modules/graphics/opengl/Font.cpp"
    L14: do {
     if ($52) {
      $54 = ((($1)) + 4|0); //@line 579 "modules/graphics/opengl/Font.cpp"
      $55 = ((($newcommands)) + 4|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $56 = ((($this)) + 24|0); //@line 398 "modules/graphics/opengl/Font.cpp"
      $57 = ((($this)) + 28|0); //@line 937 "modules/graphics/opengl/Font.cpp"
      $66 = $53;$73 = 0.0;$i$056 = 0;$y$057 = 0.0;
      while(1) {
       $65 = (($66) + (($i$056*24)|0)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $67 = HEAP32[$widths>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $68 = (($67) + ($i$056<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $69 = HEAP32[$68>>2]|0; //@line 551 "modules/graphics/opengl/Font.cpp"
       $70 = (+($69|0)); //@line 560 "modules/graphics/opengl/Font.cpp"
       HEAPF32[$width>>2] = $70; //@line 551 "modules/graphics/opengl/Font.cpp"
       $71 = (+Math_floor((+$y$057))); //@line 552 "modules/graphics/opengl/Font.cpp"
       $72 = $70 < $73; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $74 = $72 ? $maxwidth$sroa$0 : $width; //@line 2648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $75 = HEAP32[$74>>2]|0; //@line 555 "modules/graphics/opengl/Font.cpp"
       HEAP32[$maxwidth$sroa$0>>2] = $75; //@line 555 "modules/graphics/opengl/Font.cpp"
       $76 = (HEAP32[tempDoublePtr>>2]=$75,+HEAPF32[tempDoublePtr>>2]); //@line 557 "modules/graphics/opengl/Font.cpp"
       switch ($align|0) {
       case 2:  {
        $77 = $23 - $70; //@line 560 "modules/graphics/opengl/Font.cpp"
        $78 = (+Math_floor((+$77))); //@line 560 "modules/graphics/opengl/Font.cpp"
        $79 = (HEAPF32[tempDoublePtr>>2]=$78,HEAP32[tempDoublePtr>>2]|0); //@line 560 "modules/graphics/opengl/Font.cpp"
        $extraspacing$1 = 0.0;$offset$sroa$0$0 = $79;
        break;
       }
       case 1:  {
        $80 = $23 - $70; //@line 563 "modules/graphics/opengl/Font.cpp"
        $81 = $80 * 0.5; //@line 563 "modules/graphics/opengl/Font.cpp"
        $82 = (+Math_floor((+$81))); //@line 563 "modules/graphics/opengl/Font.cpp"
        $83 = (HEAPF32[tempDoublePtr>>2]=$82,HEAP32[tempDoublePtr>>2]|0); //@line 563 "modules/graphics/opengl/Font.cpp"
        $extraspacing$1 = 0.0;$offset$sroa$0$0 = $83;
        break;
       }
       case 3:  {
        $84 = HEAP32[$65>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $85 = (((($66) + (($i$056*24)|0)|0)) + 4|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $86 = HEAP32[$85>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $87 = ($84|0)==($86|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        if ($87) {
         $extraspacing$1 = 0.0;$offset$sroa$0$0 = 0;
        } else {
         $89 = $84;$__r$01$i = 0;
         while(1) {
          $88 = HEAP32[$89>>2]|0; //@line 1116 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
          $90 = ($88|0)==(32); //@line 1116 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
          $91 = $90&1; //@line 1116 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
          $$__r$0$i = (($91) + ($__r$01$i))|0; //@line 1116 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
          $92 = ((($89)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
          $93 = ($92|0)==($86|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
          if ($93) {
           $$__r$0$i$lcssa = $$__r$0$i;
           break;
          } else {
           $89 = $92;$__r$01$i = $$__r$0$i;
          }
         }
         $94 = $70 < $23; //@line 568 "modules/graphics/opengl/Font.cpp"
         $95 = ($$__r$0$i$lcssa|0)>(0); //@line 568 "modules/graphics/opengl/Font.cpp"
         $or$cond = $95 & $94; //@line 568 "modules/graphics/opengl/Font.cpp"
         if ($or$cond) {
          $96 = (+($$__r$0$i$lcssa|0)); //@line 567 "modules/graphics/opengl/Font.cpp"
          $97 = $23 - $70; //@line 569 "modules/graphics/opengl/Font.cpp"
          $98 = $97 / $96; //@line 569 "modules/graphics/opengl/Font.cpp"
          $extraspacing$1 = $98;$offset$sroa$0$0 = 0;
         } else {
          $extraspacing$1 = 0.0;$offset$sroa$0$0 = 0;
         }
        }
        break;
       }
       default: {
        $extraspacing$1 = 0.0;$offset$sroa$0$0 = 0;
       }
       }
       HEAP32[$1>>2] = $offset$sroa$0$0; //@line 579 "modules/graphics/opengl/Font.cpp"
       HEAPF32[$54>>2] = $71; //@line 579 "modules/graphics/opengl/Font.cpp"
       __THREW__ = 0;
       ;HEAP32[$tmpcast$byval_copy>>2]=HEAP32[$1>>2]|0;HEAP32[$tmpcast$byval_copy+4>>2]=HEAP32[$1+4>>2]|0; //@line 579 "modules/graphics/opengl/Font.cpp"
       invoke_viiiidii(4029,($newcommands|0),($this|0),($65|0),($vertices|0),(+$extraspacing$1),($tmpcast$byval_copy|0),(0|0)); //@line 579 "modules/graphics/opengl/Font.cpp"
       $99 = __THREW__; __THREW__ = 0;
       $100 = $99&1;
       if ($100) {
        label = 12;
        break;
       }
       $101 = HEAP32[$newcommands>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $102 = HEAP32[$55>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $103 = ($101|0)==($102|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $104 = $102; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($103) {
        $146 = $101;
       } else {
        $105 = $101; //@line 583 "modules/graphics/opengl/Font.cpp"
        $106 = HEAP32[$agg$result>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $107 = HEAP32[$11>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $108 = ($106|0)==($107|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $109 = $107; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($108) {
         $firstcmd$sroa$0$0 = $105;
        } else {
         $110 = ((($107)) + -12|0); //@line 590 "modules/graphics/opengl/Font.cpp"
         $111 = HEAP32[$110>>2]|0; //@line 590 "modules/graphics/opengl/Font.cpp"
         $112 = ((($107)) + -4|0); //@line 590 "modules/graphics/opengl/Font.cpp"
         $113 = HEAP32[$112>>2]|0; //@line 590 "modules/graphics/opengl/Font.cpp"
         $114 = HEAP32[$101>>2]|0; //@line 591 "modules/graphics/opengl/Font.cpp"
         $115 = ($111|0)==($114|0); //@line 591 "modules/graphics/opengl/Font.cpp"
         if ($115) {
          $116 = ((($107)) + -8|0); //@line 590 "modules/graphics/opengl/Font.cpp"
          $117 = HEAP32[$116>>2]|0; //@line 590 "modules/graphics/opengl/Font.cpp"
          $118 = (($117) + ($113))|0; //@line 591 "modules/graphics/opengl/Font.cpp"
          $119 = ((($101)) + 4|0); //@line 591 "modules/graphics/opengl/Font.cpp"
          $120 = HEAP32[$119>>2]|0; //@line 591 "modules/graphics/opengl/Font.cpp"
          $121 = ($118|0)==($120|0); //@line 591 "modules/graphics/opengl/Font.cpp"
          if ($121) {
           $122 = ((($101)) + 8|0); //@line 593 "modules/graphics/opengl/Font.cpp"
           $123 = HEAP32[$122>>2]|0; //@line 593 "modules/graphics/opengl/Font.cpp"
           $124 = (($123) + ($113))|0; //@line 593 "modules/graphics/opengl/Font.cpp"
           HEAP32[$112>>2] = $124; //@line 593 "modules/graphics/opengl/Font.cpp"
           $125 = ((($101)) + 12|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
           $126 = $125; //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
           $firstcmd$sroa$0$0 = $126;
          } else {
           $firstcmd$sroa$0$0 = $105;
          }
         } else {
          $firstcmd$sroa$0$0 = $105;
         }
        }
        HEAP32[$2>>2] = $109; //@line 1128 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        HEAP32[$3>>2] = $firstcmd$sroa$0$0; //@line 599 "modules/graphics/opengl/Font.cpp"
        HEAP32[$4>>2] = $104; //@line 599 "modules/graphics/opengl/Font.cpp"
        __THREW__ = 0;
        ;HEAP32[$$byval_copy>>2]=HEAP32[$2>>2]|0; //@line 599 "modules/graphics/opengl/Font.cpp"
        ;HEAP32[$$byval_copy6>>2]=HEAP32[$3>>2]|0; //@line 599 "modules/graphics/opengl/Font.cpp"
        ;HEAP32[$$byval_copy7>>2]=HEAP32[$4>>2]|0; //@line 599 "modules/graphics/opengl/Font.cpp"
        (invoke_iiiii(4030,($agg$result|0),($$byval_copy|0),($$byval_copy6|0),($$byval_copy7|0))|0); //@line 599 "modules/graphics/opengl/Font.cpp"
        $138 = __THREW__; __THREW__ = 0;
        $139 = $138&1;
        if ($139) {
         break;
        }
        $$pre = HEAP32[$newcommands>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $146 = $$pre;
       }
       $140 = HEAP32[$56>>2]|0; //@line 398 "modules/graphics/opengl/Font.cpp"
       $141 = (+($140|0)); //@line 398 "modules/graphics/opengl/Font.cpp"
       $142 = +HEAPF32[$57>>2]; //@line 937 "modules/graphics/opengl/Font.cpp"
       $143 = $141 * $142; //@line 602 "modules/graphics/opengl/Font.cpp"
       $144 = $y$057 + $143; //@line 602 "modules/graphics/opengl/Font.cpp"
       $145 = ($146|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $147 = $146; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($145)) {
        $148 = HEAP32[$55>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $149 = ($148|0)==($146|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($149)) {
         $scevgep$i$i$i38 = ((($148)) + -12|0);
         $150 = $scevgep$i$i$i38;
         $151 = (($150) - ($147))|0;
         $152 = (($151>>>0) / 12)&-1;
         $153 = $152 ^ -1;
         $scevgep4$i$i$i39 = (($148) + (($153*12)|0)|0);
         HEAP32[$55>>2] = $scevgep4$i$i$i39; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
        __ZdlPv($146); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       }
       $154 = (($i$056) + 1)|0; //@line 547 "modules/graphics/opengl/Font.cpp"
       $155 = HEAP32[$45>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $156 = HEAP32[$lines>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $157 = (($155) - ($156))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $158 = (($157|0) / 24)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $159 = ($154|0)<($158|0); //@line 547 "modules/graphics/opengl/Font.cpp"
       $160 = $156; //@line 547 "modules/graphics/opengl/Font.cpp"
       if ($159) {
        $66 = $160;$73 = $76;$i$056 = $154;$y$057 = $144;
       } else {
        $162 = $76;$277 = $160;$y$0$lcssa = $144;
        break L14;
       }
      }
      if ((label|0) == 12) {
       $61 = ___cxa_find_matching_catch_2()|0;
       $62 = tempRet0;
       $lpad$phi$index = $61;$lpad$phi$index2 = $62;
       label = 14;
       break L12;
      }
      $127 = ___cxa_find_matching_catch_2()|0;
      $128 = tempRet0;
      $129 = HEAP32[$newcommands>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $130 = ($129|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $131 = $129; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($130) {
       $$0 = $127;$$02 = $128;
       break L12;
      }
      $132 = HEAP32[$55>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $133 = ($132|0)==($129|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($133)) {
       $scevgep$i$i$i43 = ((($132)) + -12|0);
       $134 = $scevgep$i$i$i43;
       $135 = (($134) - ($131))|0;
       $136 = (($135>>>0) / 12)&-1;
       $137 = $136 ^ -1;
       $scevgep4$i$i$i44 = (($132) + (($137*12)|0)|0);
       HEAP32[$55>>2] = $scevgep4$i$i$i44; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($129); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $$0 = $127;$$02 = $128;
      break L12;
     } else {
      $162 = 0.0;$277 = $53;$y$0$lcssa = 0.0;
     }
    } while(0);
    $58 = ($info|0)==(0|0); //@line 605 "modules/graphics/opengl/Font.cpp"
    if (!($58)) {
     $161 = (~~(($162))); //@line 607 "modules/graphics/opengl/Font.cpp"
     HEAP32[$info>>2] = $161; //@line 607 "modules/graphics/opengl/Font.cpp"
     $163 = (~~(($y$0$lcssa))); //@line 608 "modules/graphics/opengl/Font.cpp"
     $164 = ((($info)) + 4|0); //@line 608 "modules/graphics/opengl/Font.cpp"
     HEAP32[$164>>2] = $163; //@line 608 "modules/graphics/opengl/Font.cpp"
    }
    $165 = HEAP32[$9>>2]|0; //@line 611 "modules/graphics/opengl/Font.cpp"
    $166 = ($10|0)==($165|0); //@line 611 "modules/graphics/opengl/Font.cpp"
    if ($166) {
     $193 = $277;
    } else {
     $167 = HEAP32[$vertices>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $168 = ((($vertices)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $169 = HEAP32[$168>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $170 = ($169|0)==($167|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($170)) {
      $__new_last2$i$i$i = $167; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $scevgep$i$i$i35 = ((($169)) + -16|0);
      $171 = $scevgep$i$i$i35;
      $172 = (($171) - ($__new_last2$i$i$i))|0;
      $173 = $172 >>> 4;
      $174 = $173 ^ -1;
      $scevgep4$i$i$i36 = (($169) + ($174<<4)|0);
      HEAP32[$168>>2] = $scevgep4$i$i$i36; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __THREW__ = 0;
     invoke_viiidiii(4031,($5|0),($this|0),($text|0),(+$23),($align|0),($vertices|0),(0|0)); //@line 614 "modules/graphics/opengl/Font.cpp"
     $175 = __THREW__; __THREW__ = 0;
     $176 = $175&1;
     if ($176) {
      label = 13;
      break;
     }
     $177 = HEAP32[$agg$result>>2]|0; //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $178 = ($177|0)==(0|0); //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $179 = $177; //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($178) {
      $$pre1$i$i = ((($agg$result)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $$pre$phi2$i$iZ2D = $$pre1$i$i;
     } else {
      $180 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $181 = ($180|0)==($177|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($181)) {
       $scevgep$i$i$i$i$i$i = ((($180)) + -12|0);
       $182 = $scevgep$i$i$i$i$i$i;
       $183 = (($182) - ($179))|0;
       $184 = (($183>>>0) / 12)&-1;
       $185 = $184 ^ -1;
       $scevgep4$i$i$i$i$i$i = (($180) + (($185*12)|0)|0);
       HEAP32[$11>>2] = $scevgep4$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($177); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $186 = ((($agg$result)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$186>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$11>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$agg$result>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$pre$phi2$i$iZ2D = $186;
     }
     $187 = HEAP32[$5>>2]|0; //@line 1343 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$agg$result>>2] = $187; //@line 1343 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $188 = ((($5)) + 4|0); //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $189 = HEAP32[$188>>2]|0; //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$11>>2] = $189; //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $190 = ((($5)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $191 = HEAP32[$190>>2]|0; //@line 1345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$$pre$phi2$i$iZ2D>>2] = $191; //@line 1345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$190>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$188>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$5>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $$pre59 = HEAP32[$lines>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $193 = $$pre59;
    }
    $192 = ($193|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($192)) {
     $194 = HEAP32[$45>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $195 = ($194|0)==($193|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($195) {
      $220 = $193;
     } else {
      $197 = $194;
      while(1) {
       $196 = ((($197)) + -24|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$45>>2] = $196; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $198 = ((($197)) + -12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $199 = HEAP32[$198>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $200 = ($199|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($200)) {
        $201 = ((($197)) + -8|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $202 = HEAP32[$201>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $203 = ($202|0)==($199|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($203)) {
         $__new_last1$i$i$i$i$i$i$i$i$i$i11 = $199; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $scevgep$i$i$i$i$i$i$i$i$i$i12 = ((($202)) + -8|0);
         $204 = $scevgep$i$i$i$i$i$i$i$i$i$i12;
         $205 = (($204) - ($__new_last1$i$i$i$i$i$i$i$i$i$i11))|0;
         $206 = $205 >>> 3;
         $207 = $206 ^ -1;
         $scevgep4$i$i$i$i$i$i$i$i$i$i13 = (($202) + ($207<<3)|0);
         HEAP32[$201>>2] = $scevgep4$i$i$i$i$i$i$i$i$i$i13; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
        __ZdlPv($199); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       }
       $208 = HEAP32[$196>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $209 = ($208|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $210 = $208; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($209)) {
        $213 = ((($197)) + -20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $214 = HEAP32[$213>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $215 = ($214|0)==($208|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($215)) {
         $scevgep$i$i$i1$i$i$i$i$i$i$i18 = ((($214)) + -4|0);
         $216 = $scevgep$i$i$i1$i$i$i$i$i$i$i18;
         $217 = (($216) - ($210))|0;
         $218 = $217 >>> 2;
         $219 = $218 ^ -1;
         $scevgep4$i$i$i2$i$i$i$i$i$i$i19 = (($214) + ($219<<2)|0);
         HEAP32[$213>>2] = $scevgep4$i$i$i2$i$i$i$i$i$i$i19; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
        __ZdlPv($208); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       }
       $211 = HEAP32[$45>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $212 = ($211|0)==($193|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($212) {
        break;
       } else {
        $197 = $211;
       }
      }
      $$pre$i23 = HEAP32[$lines>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $220 = $$pre$i23;
     }
     __ZdlPv($220); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    }
    $221 = HEAP32[$widths>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $222 = ($221|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $223 = $221; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($222) {
     STACKTOP = sp;return; //@line 618 "modules/graphics/opengl/Font.cpp"
    }
    $224 = HEAP32[$43>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $225 = ($224|0)==($221|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($225)) {
     $scevgep$i$i$i5 = ((($224)) + -4|0);
     $226 = $scevgep$i$i$i5;
     $227 = (($226) - ($223))|0;
     $228 = $227 >>> 2;
     $229 = $228 ^ -1;
     $scevgep4$i$i$i6 = (($224) + ($229<<2)|0);
     HEAP32[$43>>2] = $scevgep4$i$i$i6; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($221); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    STACKTOP = sp;return; //@line 618 "modules/graphics/opengl/Font.cpp"
   }
  } while(0);
  if ((label|0) == 13) {
   $63 = ___cxa_find_matching_catch_2()|0;
   $64 = tempRet0;
   $lpad$phi$index = $63;$lpad$phi$index2 = $64;
   label = 14;
  }
  if ((label|0) == 14) {
   $$0 = $lpad$phi$index;$$02 = $lpad$phi$index2;
  }
  $230 = HEAP32[$lines>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $231 = ($230|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($231)) {
   $232 = HEAP32[$45>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $233 = ($232|0)==($230|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($233) {
    $258 = $230;
   } else {
    $235 = $232;
    while(1) {
     $234 = ((($235)) + -24|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$45>>2] = $234; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $236 = ((($235)) + -12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $237 = HEAP32[$236>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $238 = ($237|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($238)) {
      $239 = ((($235)) + -8|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $240 = HEAP32[$239>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $241 = ($240|0)==($237|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($241)) {
       $__new_last1$i$i$i$i$i$i$i$i$i$i = $237; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $scevgep$i$i$i$i$i$i$i$i$i$i = ((($240)) + -8|0);
       $242 = $scevgep$i$i$i$i$i$i$i$i$i$i;
       $243 = (($242) - ($__new_last1$i$i$i$i$i$i$i$i$i$i))|0;
       $244 = $243 >>> 3;
       $245 = $244 ^ -1;
       $scevgep4$i$i$i$i$i$i$i$i$i$i = (($240) + ($245<<3)|0);
       HEAP32[$239>>2] = $scevgep4$i$i$i$i$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($237); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $246 = HEAP32[$234>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $247 = ($246|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $248 = $246; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($247)) {
      $251 = ((($235)) + -20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $252 = HEAP32[$251>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $253 = ($252|0)==($246|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($253)) {
       $scevgep$i$i$i1$i$i$i$i$i$i$i = ((($252)) + -4|0);
       $254 = $scevgep$i$i$i1$i$i$i$i$i$i$i;
       $255 = (($254) - ($248))|0;
       $256 = $255 >>> 2;
       $257 = $256 ^ -1;
       $scevgep4$i$i$i2$i$i$i$i$i$i$i = (($252) + ($257<<2)|0);
       HEAP32[$251>>2] = $scevgep4$i$i$i2$i$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($246); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $249 = HEAP32[$45>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $250 = ($249|0)==($230|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($250) {
      break;
     } else {
      $235 = $249;
     }
    }
    $$pre$i = HEAP32[$lines>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $258 = $$pre$i;
   }
   __ZdlPv($258); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  $259 = HEAP32[$widths>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $260 = ($259|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $261 = $259; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($260) {
   $$1 = $$0;$$13 = $$02;
  } else {
   $262 = HEAP32[$43>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $263 = ($262|0)==($259|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($263)) {
    $scevgep$i$i$i = ((($262)) + -4|0);
    $264 = $scevgep$i$i$i;
    $265 = (($264) - ($261))|0;
    $266 = $265 >>> 2;
    $267 = $266 ^ -1;
    $scevgep4$i$i$i = (($262) + ($267<<2)|0);
    HEAP32[$43>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($259); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $$1 = $$0;$$13 = $$02;
  }
 }
 $268 = HEAP32[$agg$result>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $269 = ($268|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $270 = $268; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($269) {
  ___resumeException($$1|0);
  // unreachable;
 }
 $271 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $272 = ($271|0)==($268|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($272)) {
  $scevgep$i$i$i30 = ((($271)) + -12|0);
  $273 = $scevgep$i$i$i30;
  $274 = (($273) - ($270))|0;
  $275 = (($274>>>0) / 12)&-1;
  $276 = $275 ^ -1;
  $scevgep4$i$i$i31 = (($271) + (($276*12)|0)|0);
  HEAP32[$11>>2] = $scevgep4$i$i$i31; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($268); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$1|0);
 // unreachable;
}
function __ZN4love8graphics6opengl4Font7getWrapERKNS2_17ColoredCodepointsEfRNSt3__16vectorIS3_NS6_9allocatorIS3_EEEEPNS7_IiNS8_IiEEEE($this,$codepoints,$wraplimit,$lines,$linewidths) {
 $this = $this|0;
 $codepoints = $codepoints|0;
 $wraplimit = +$wraplimit;
 $lines = $lines|0;
 $linewidths = $linewidths|0;
 var $$0$i = 0, $$01$i$i = 0, $$01$i$i45 = 0, $$01$i$i88 = 0, $$in = 0, $$lcssa = 0, $$lcssa186 = 0, $$lcssa187 = 0, $$lcssa188 = 0, $$lcssa189 = 0, $$lcssa190 = 0, $$pn$i$i$i = 0, $$pn$us$i$i$i = 0, $$pre = 0, $$pre$phi148152Z2D = 0, $$pre$phi150151Z2D = 0, $$pre145 = 0, $$pre149 = 0, $$sroa$0$0$i$i$i = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0.0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0.0, $194 = 0, $195 = 0, $196 = 0.0, $197 = 0.0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0;
 var $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0;
 var $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0;
 var $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0;
 var $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0;
 var $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0;
 var $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0.0, $366 = 0, $367 = 0, $368 = 0, $369 = 0;
 var $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0;
 var $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0;
 var $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0;
 var $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0;
 var $441 = 0.0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0;
 var $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__nd$0$i$i$i = 0, $__nd$0$us$i$i$i = 0, $__new_last1$i$i$i = 0, $__new_last1$i$i$i$i = 0, $__new_last1$i$i$i$i11 = 0, $__new_last1$i$i$i$i19 = 0, $__new_last1$i$i$i$i46 = 0, $__new_last1$i$i$i$i89 = 0, $__new_last1$i$i$i63 = 0, $__new_last2$i$i$i = 0;
 var $__new_last2$i$i$i68 = 0, $addcurcolor$0$off0125 = 0, $addcurcolor$1$off0 = 0, $addcurcolor$4$off0 = 0, $c = 0, $colori$0118 = 0, $colori$0118$lcssa = 0, $curcolor$sroa$0$0123 = 0, $curcolor$sroa$0$1 = 0, $curcolor$sroa$0$2 = 0, $curcolor$sroa$0$3 = 0, $curcolor$sroa$0$4 = 0, $curcolor$sroa$6$0122 = 0, $curcolor$sroa$6$1 = 0, $curcolor$sroa$6$2 = 0, $curcolor$sroa$6$3 = 0, $curcolor$sroa$6$4 = 0, $curcolor$sroa$7$0121 = 0, $curcolor$sroa$7$1 = 0, $curcolor$sroa$7$2 = 0;
 var $curcolor$sroa$7$3 = 0, $curcolor$sroa$7$4 = 0, $curcolor$sroa$8$0120 = 0, $curcolor$sroa$8$1 = 0, $curcolor$sroa$8$2 = 0, $curcolor$sroa$8$3 = 0, $curcolor$sroa$8$4 = 0, $curcolori$0124 = 0, $curcolori$1 = 0, $curcolori$2 = 0, $curcolori$3 = 0, $curcolori$5 = 0, $eh$lpad$body$index8Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i$0126 = 0, $i$1 = 0, $i$3 = 0, $lastspaceindex$0127 = 0, $lastspaceindex$1 = 0, $lastspaceindex$3 = 0;
 var $or$cond = 0, $or$cond8 = 0, $prevglyph$0128 = 0, $prevglyph$2 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i$i = 0, $scevgep$i$i$i$i$i$i$i$i = 0, $scevgep$i$i$i$i$i$i$i$i30 = 0, $scevgep$i$i$i$i$i$i$i$i75 = 0, $scevgep$i$i$i$i12 = 0, $scevgep$i$i$i$i20 = 0, $scevgep$i$i$i$i47 = 0, $scevgep$i$i$i$i90 = 0, $scevgep$i$i$i1$i = 0, $scevgep$i$i$i1$i25 = 0, $scevgep$i$i$i55 = 0, $scevgep$i$i$i64 = 0, $scevgep$i$i$i69 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i$i = 0;
 var $scevgep4$i$i$i$i$i$i$i$i = 0, $scevgep4$i$i$i$i$i$i$i$i31 = 0, $scevgep4$i$i$i$i$i$i$i$i76 = 0, $scevgep4$i$i$i$i13 = 0, $scevgep4$i$i$i$i21 = 0, $scevgep4$i$i$i$i48 = 0, $scevgep4$i$i$i$i91 = 0, $scevgep4$i$i$i2$i = 0, $scevgep4$i$i$i2$i26 = 0, $scevgep4$i$i$i56 = 0, $scevgep4$i$i$i61 = 0, $scevgep4$i$i$i65 = 0, $scevgep4$i$i$i70 = 0, $width$0$widthbeforelastspace$0 = 0.0, $width$0131 = 0.0, $width$1 = 0.0, $width$3 = 0.0, $width$3$lcssa = 0.0, $widthbeforelastspace$0130 = 0.0, $widthbeforelastspace$3 = 0.0;
 var $widthoftrailingspace$0129 = 0.0, $widthoftrailingspace$1 = 0.0, $widthoftrailingspace$3 = 0.0, $widthoftrailingspace$3$lcssa = 0.0, $wline = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $wline = sp + 24|0;
 $c = sp + 20|0;
 $0 = sp + 16|0;
 $1 = sp + 12|0;
 $2 = sp;
 $3 = sp + 8|0;
 $4 = ((($codepoints)) + 12|0); //@line 769 "modules/graphics/opengl/Font.cpp"
 $5 = ((($codepoints)) + 16|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$5>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = HEAP32[$4>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = (($6) - ($7))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = $8 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9) + -1)|0; //@line 769 "modules/graphics/opengl/Font.cpp"
 ;HEAP32[$wline>>2]=0|0;HEAP32[$wline+4>>2]=0|0;HEAP32[$wline+8>>2]=0|0;HEAP32[$wline+12>>2]=0|0;HEAP32[$wline+16>>2]=0|0;HEAP32[$wline+20>>2]=0|0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ((($codepoints)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = HEAP32[$11>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = HEAP32[$codepoints>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = (($12) - ($13))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = ($14|0)>(0); //@line 775 "modules/graphics/opengl/Font.cpp"
 L1: do {
  if ($15) {
   $16 = ((($lines)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $17 = ((($lines)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $18 = ($linewidths|0)==(0|0); //@line 794 "modules/graphics/opengl/Font.cpp"
   $19 = ((($wline)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $20 = ((($wline)) + 12|0); //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $21 = ((($wline)) + 16|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $22 = ((($linewidths)) + 4|0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $23 = ((($linewidths)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $24 = ((($wline)) + 12|0); //@line 76 "modules/graphics/opengl/Font.h"
   $25 = ((($this)) + 56|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $26 = ((($2)) + 4|0); //@line 874 "modules/graphics/opengl/Font.cpp"
   $27 = ((($wline)) + 20|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $28 = ((($wline)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $29 = ((($this)) + 52|0); //@line 2015 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $$in = $13;$addcurcolor$0$off0125 = 0;$curcolor$sroa$0$0123 = -1;$curcolor$sroa$6$0122 = -1;$curcolor$sroa$7$0121 = -1;$curcolor$sroa$8$0120 = -1;$curcolori$0124 = -1;$i$0126 = 0;$lastspaceindex$0127 = -1;$prevglyph$0128 = 0;$width$0131 = 0.0;$widthbeforelastspace$0130 = 0.0;$widthoftrailingspace$0129 = 0.0;
   L3: while(1) {
    $30 = $$in; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $31 = (($30) + ($i$0126<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $32 = HEAP32[$31>>2]|0; //@line 882 "modules/graphics/opengl/Font.cpp"
    HEAP32[$c>>2] = $32; //@line 777 "modules/graphics/opengl/Font.cpp"
    $33 = ($curcolori$0124|0)<($10|0); //@line 781 "modules/graphics/opengl/Font.cpp"
    if ($33) {
     $34 = (($curcolori$0124) + 1)|0; //@line 781 "modules/graphics/opengl/Font.cpp"
     $35 = HEAP32[$4>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $36 = (((($35) + ($34<<3)|0)) + 4|0); //@line 781 "modules/graphics/opengl/Font.cpp"
     $37 = HEAP32[$36>>2]|0; //@line 781 "modules/graphics/opengl/Font.cpp"
     $38 = ($37|0)==($i$0126|0); //@line 781 "modules/graphics/opengl/Font.cpp"
     if ($38) {
      $39 = (($35) + ($34<<3)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $40 = HEAPU8[$39>>0]|(HEAPU8[$39+1>>0]<<8)|(HEAPU8[$39+2>>0]<<16)|(HEAPU8[$39+3>>0]<<24); //@line 783 "modules/graphics/opengl/Font.cpp"
      $41 = $40&255; //@line 783 "modules/graphics/opengl/Font.cpp"
      $42 = $40 >>> 8; //@line 783 "modules/graphics/opengl/Font.cpp"
      $43 = $42&255; //@line 783 "modules/graphics/opengl/Font.cpp"
      $44 = $40 >>> 16; //@line 783 "modules/graphics/opengl/Font.cpp"
      $45 = $44&255; //@line 783 "modules/graphics/opengl/Font.cpp"
      $46 = $40 >>> 24; //@line 783 "modules/graphics/opengl/Font.cpp"
      $47 = $46&255; //@line 783 "modules/graphics/opengl/Font.cpp"
      $addcurcolor$1$off0 = 1;$curcolor$sroa$0$1 = $41;$curcolor$sroa$6$1 = $43;$curcolor$sroa$7$1 = $45;$curcolor$sroa$8$1 = $47;$curcolori$1 = $34;
     } else {
      $addcurcolor$1$off0 = $addcurcolor$0$off0125;$curcolor$sroa$0$1 = $curcolor$sroa$0$0123;$curcolor$sroa$6$1 = $curcolor$sroa$6$0122;$curcolor$sroa$7$1 = $curcolor$sroa$7$0121;$curcolor$sroa$8$1 = $curcolor$sroa$8$0120;$curcolori$1 = $curcolori$0124;
     }
    } else {
     $addcurcolor$1$off0 = $addcurcolor$0$off0125;$curcolor$sroa$0$1 = $curcolor$sroa$0$0123;$curcolor$sroa$6$1 = $curcolor$sroa$6$0122;$curcolor$sroa$7$1 = $curcolor$sroa$7$0121;$curcolor$sroa$8$1 = $curcolor$sroa$8$0120;$curcolori$1 = $curcolori$0124;
    }
    $72 = ($32|0)==(10); //@line 789 "modules/graphics/opengl/Font.cpp"
    do {
     if ($72) {
      $73 = HEAP32[$16>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $74 = HEAP32[$17>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $75 = ($73|0)==($74|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($75) {
       __THREW__ = 0;
       invoke_vii(4033,($lines|0),($wline|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $134 = __THREW__; __THREW__ = 0;
       $135 = $134&1;
       if ($135) {
        label = 7;
        break L3;
       }
      } else {
       __THREW__ = 0;
       invoke_vii(4032,($73|0),($wline|0)); //@line 76 "modules/graphics/opengl/Font.h"
       $76 = __THREW__; __THREW__ = 0;
       $77 = $76&1;
       if ($77) {
        label = 7;
        break L3;
       }
       $78 = ((($73)) + 12|0); //@line 76 "modules/graphics/opengl/Font.h"
       HEAP32[$78>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $79 = ((($73)) + 16|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$79>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $80 = ((($73)) + 20|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$80>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $81 = HEAP32[$21>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $82 = HEAP32[$24>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $83 = (($81) - ($82))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $84 = $83 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $85 = ($84|0)==(0); //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($85)) {
        $86 = ($84>>>0)>(536870911); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($86) {
         __THREW__ = 0;
         invoke_vi(3528,($78|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $87 = __THREW__; __THREW__ = 0;
         $88 = $87&1;
         if ($88) {
          $$lcssa188 = $73;$$lcssa189 = $78;$$lcssa190 = $79;
          label = 28;
          break L3;
         }
        }
        __THREW__ = 0;
        $89 = (invoke_ii(3348,($83|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        $90 = __THREW__; __THREW__ = 0;
        $91 = $90&1;
        if ($91) {
         $$lcssa188 = $73;$$lcssa189 = $78;$$lcssa190 = $79;
         label = 28;
         break L3;
        }
        HEAP32[$79>>2] = $89; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$78>>2] = $89; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $92 = (($89) + ($84<<3)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $93 = ((($73)) + 20|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        HEAP32[$93>>2] = $92; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $94 = HEAP32[$24>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $95 = HEAP32[$21>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $96 = ($94|0)==($95|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($96)) {
         $$01$i$i45 = $94;$104 = $89;
         while(1) {
          $97 = $$01$i$i45; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $98 = $97; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $99 = HEAP32[$98>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $100 = (($97) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $101 = $100; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $102 = HEAP32[$101>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $103 = $104; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $105 = $103; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          HEAP32[$105>>2] = $99; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $106 = (($103) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $107 = $106; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          HEAP32[$107>>2] = $102; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $108 = HEAP32[$79>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $109 = ((($108)) + 8|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$79>>2] = $109; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $110 = ((($$01$i$i45)) + 8|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $111 = ($110|0)==($95|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          if ($111) {
           break;
          } else {
           $$01$i$i45 = $110;$104 = $109;
          }
         }
        }
       }
       $132 = HEAP32[$16>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $133 = ((($132)) + 24|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$16>>2] = $133; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      do {
       if (!($18)) {
        $136 = $width$0131 - $widthoftrailingspace$0129; //@line 795 "modules/graphics/opengl/Font.cpp"
        $137 = (~~(($136))); //@line 795 "modules/graphics/opengl/Font.cpp"
        HEAP32[$0>>2] = $137; //@line 795 "modules/graphics/opengl/Font.cpp"
        $138 = HEAP32[$22>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $139 = HEAP32[$23>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $140 = ($138>>>0)<($139>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($140) {
         HEAP32[$138>>2] = $137; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $141 = ((($138)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$22>>2] = $141; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         break;
        } else {
         __THREW__ = 0;
         invoke_vii(3406,($linewidths|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $142 = __THREW__; __THREW__ = 0;
         $143 = $142&1;
         if ($143) {
          label = 7;
          break L3;
         } else {
          break;
         }
        }
       }
      } while(0);
      $144 = HEAP32[$wline>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $145 = HEAP32[$19>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $146 = ($145|0)==($144|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($146)) {
       $__new_last2$i$i$i = $144; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $scevgep$i$i$i = ((($145)) + -4|0);
       $147 = $scevgep$i$i$i;
       $148 = (($147) - ($__new_last2$i$i$i))|0;
       $149 = $148 >>> 2;
       $150 = $149 ^ -1;
       $scevgep4$i$i$i = (($145) + ($150<<2)|0);
       HEAP32[$19>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      $151 = HEAP32[$20>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $152 = HEAP32[$21>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $153 = ($152|0)==($151|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($153)) {
       $__new_last1$i$i$i = $151; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $scevgep$i$i$i55 = ((($152)) + -8|0);
       $154 = $scevgep$i$i$i55;
       $155 = (($154) - ($__new_last1$i$i$i))|0;
       $156 = $155 >>> 3;
       $157 = $156 ^ -1;
       $scevgep4$i$i$i56 = (($152) + ($157<<3)|0);
       HEAP32[$21>>2] = $scevgep4$i$i$i56; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      $158 = (($i$0126) + 1)|0; //@line 805 "modules/graphics/opengl/Font.cpp"
      $addcurcolor$4$off0 = 1;$curcolor$sroa$0$4 = $curcolor$sroa$0$1;$curcolor$sroa$6$4 = $curcolor$sroa$6$1;$curcolor$sroa$7$4 = $curcolor$sroa$7$1;$curcolor$sroa$8$4 = $curcolor$sroa$8$1;$curcolori$5 = $curcolori$1;$i$3 = $158;$lastspaceindex$3 = -1;$prevglyph$2 = 0;$width$3 = 0.0;$widthbeforelastspace$3 = 0.0;$widthoftrailingspace$3 = 0.0;
     } else {
      $159 = HEAP32[$25>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $160 = ($159|0)==(0); //@line 2018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      L39: do {
       if ($160) {
        label = 60;
       } else {
        $161 = (($159) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $162 = $161 & $159; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $163 = ($162|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        if ($163) {
         $164 = $161 & $32; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $168 = $164;
        } else {
         $165 = (($32>>>0) % ($159>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $168 = $165;
        }
        $166 = HEAP32[$29>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $167 = (($166) + ($168<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $169 = HEAP32[$167>>2]|0; //@line 2021 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $170 = ($169|0)==(0|0); //@line 2022 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        if ($170) {
         label = 60;
        } else {
         if ($163) {
          $$pn$us$i$i$i = $169;
          while(1) {
           $__nd$0$us$i$i$i = HEAP32[$$pn$us$i$i$i>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
           $171 = ($__nd$0$us$i$i$i|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
           if ($171) {
            label = 60;
            break L39;
           }
           $172 = ((($__nd$0$us$i$i$i)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
           $173 = HEAP32[$172>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
           $174 = $173 & $161; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
           $175 = ($174|0)==($168|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
           if (!($175)) {
            label = 60;
            break L39;
           }
           $176 = ((($__nd$0$us$i$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
           $177 = HEAP32[$176>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
           $178 = ($177|0)==($32|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
           if ($178) {
            $$sroa$0$0$i$i$i = $__nd$0$us$i$i$i;
            break;
           } else {
            $$pn$us$i$i$i = $__nd$0$us$i$i$i;
           }
          }
         } else {
          $$pn$i$i$i = $169;
          while(1) {
           $__nd$0$i$i$i = HEAP32[$$pn$i$i$i>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
           $179 = ($__nd$0$i$i$i|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
           if ($179) {
            label = 60;
            break L39;
           }
           $180 = ((($__nd$0$i$i$i)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
           $181 = HEAP32[$180>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
           $182 = (($181>>>0) % ($159>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
           $183 = ($182|0)==($168|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
           if (!($183)) {
            label = 60;
            break L39;
           }
           $184 = ((($__nd$0$i$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
           $185 = HEAP32[$184>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
           $186 = ($185|0)==($32|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
           if ($186) {
            $$sroa$0$0$i$i$i = $__nd$0$i$i$i;
            break;
           } else {
            $$pn$i$i$i = $__nd$0$i$i$i;
           }
          }
         }
         $187 = ((($$sroa$0$0$i$i$i)) + 12|0); //@line 323 "modules/graphics/opengl/Font.cpp"
         $$0$i = $187;
        }
       }
      } while(0);
      if ((label|0) == 60) {
       label = 0;
       __THREW__ = 0;
       $188 = (invoke_iii(4017,($this|0),($32|0))|0); //@line 325 "modules/graphics/opengl/Font.cpp"
       $189 = __THREW__; __THREW__ = 0;
       $190 = $189&1;
       if ($190) {
        label = 7;
        break L3;
       } else {
        $$0$i = $188;
       }
      }
      $191 = ((($$0$i)) + 4|0); //@line 811 "modules/graphics/opengl/Font.cpp"
      $192 = HEAP32[$191>>2]|0; //@line 811 "modules/graphics/opengl/Font.cpp"
      __THREW__ = 0;
      $193 = (+invoke_diii(4025,($this|0),($prevglyph$0128|0),($32|0))); //@line 811 "modules/graphics/opengl/Font.cpp"
      $194 = __THREW__; __THREW__ = 0;
      $195 = $194&1;
      if ($195) {
       label = 7;
       break L3;
      }
      $196 = (+($192|0)); //@line 811 "modules/graphics/opengl/Font.cpp"
      $197 = $196 + $193; //@line 811 "modules/graphics/opengl/Font.cpp"
      $198 = $width$0131 + $197; //@line 812 "modules/graphics/opengl/Font.cpp"
      $199 = ($32|0)!=(32); //@line 816 "modules/graphics/opengl/Font.cpp"
      $200 = $198 > $wraplimit; //@line 816 "modules/graphics/opengl/Font.cpp"
      $or$cond8 = $200 & $199; //@line 816 "modules/graphics/opengl/Font.cpp"
      if (!($or$cond8)) {
       $326 = ($prevglyph$0128|0)!=(32); //@line 866 "modules/graphics/opengl/Font.cpp"
       $327 = ($32|0)==(32); //@line 866 "modules/graphics/opengl/Font.cpp"
       $or$cond = $326 & $327; //@line 866 "modules/graphics/opengl/Font.cpp"
       $width$0$widthbeforelastspace$0 = $or$cond ? $width$0131 : $widthbeforelastspace$0130; //@line 866 "modules/graphics/opengl/Font.cpp"
       do {
        if ($addcurcolor$1$off0) {
         $328 = $curcolor$sroa$8$1&255; //@line 874 "modules/graphics/opengl/Font.cpp"
         $329 = $328 << 24; //@line 874 "modules/graphics/opengl/Font.cpp"
         $330 = $curcolor$sroa$7$1&255; //@line 874 "modules/graphics/opengl/Font.cpp"
         $331 = $330 << 16; //@line 874 "modules/graphics/opengl/Font.cpp"
         $332 = $331 | $329; //@line 874 "modules/graphics/opengl/Font.cpp"
         $333 = $curcolor$sroa$6$1&255; //@line 874 "modules/graphics/opengl/Font.cpp"
         $334 = $333 << 8; //@line 874 "modules/graphics/opengl/Font.cpp"
         $335 = $332 | $334; //@line 874 "modules/graphics/opengl/Font.cpp"
         $336 = $curcolor$sroa$0$1&255; //@line 874 "modules/graphics/opengl/Font.cpp"
         $337 = $335 | $336; //@line 874 "modules/graphics/opengl/Font.cpp"
         HEAP32[$2>>2] = $337; //@line 874 "modules/graphics/opengl/Font.cpp"
         $338 = HEAP32[$19>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $339 = HEAP32[$wline>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $340 = (($338) - ($339))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $341 = $340 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$26>>2] = $341; //@line 874 "modules/graphics/opengl/Font.cpp"
         $342 = HEAP32[$21>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $343 = HEAP32[$27>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $344 = ($342>>>0)<($343>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($344) {
          $345 = $2; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $346 = $345; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $347 = HEAP32[$346>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $348 = (($345) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $349 = $348; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $350 = HEAP32[$349>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $351 = $342; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $352 = $351; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          HEAP32[$352>>2] = $347; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $353 = (($351) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $354 = $353; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          HEAP32[$354>>2] = $350; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $355 = HEAP32[$21>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $356 = ((($355)) + 8|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$21>>2] = $356; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          break;
         } else {
          __THREW__ = 0;
          invoke_vii(4034,($24|0),($2|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $357 = __THREW__; __THREW__ = 0;
          $358 = $357&1;
          if ($358) {
           label = 7;
           break L3;
          } else {
           break;
          }
         }
        }
       } while(0);
       $359 = HEAP32[$19>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $360 = HEAP32[$28>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $361 = ($359|0)==($360|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($361) {
        __THREW__ = 0;
        invoke_vii(4016,($wline|0),($c|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $363 = __THREW__; __THREW__ = 0;
        $364 = $363&1;
        if ($364) {
         label = 7;
         break L3;
        }
       } else {
        HEAP32[$359>>2] = $32; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $362 = ((($359)) + 4|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$19>>2] = $362; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       switch ($32|0) {
       case 32:  {
        $365 = $widthoftrailingspace$0129 + $197; //@line 885 "modules/graphics/opengl/Font.cpp"
        $lastspaceindex$1 = $i$0126;$widthoftrailingspace$1 = $365;
        break;
       }
       case 10:  {
        $lastspaceindex$1 = $lastspaceindex$0127;$widthoftrailingspace$1 = $widthoftrailingspace$0129;
        break;
       }
       default: {
        $lastspaceindex$1 = $lastspaceindex$0127;$widthoftrailingspace$1 = 0.0;
       }
       }
       $366 = (($i$0126) + 1)|0; //@line 890 "modules/graphics/opengl/Font.cpp"
       $addcurcolor$4$off0 = 0;$curcolor$sroa$0$4 = $curcolor$sroa$0$1;$curcolor$sroa$6$4 = $curcolor$sroa$6$1;$curcolor$sroa$7$4 = $curcolor$sroa$7$1;$curcolor$sroa$8$4 = $curcolor$sroa$8$1;$curcolori$5 = $curcolori$1;$i$3 = $366;$lastspaceindex$3 = $lastspaceindex$1;$prevglyph$2 = $32;$width$3 = $198;$widthbeforelastspace$3 = $width$0$widthbeforelastspace$0;$widthoftrailingspace$3 = $widthoftrailingspace$1;
       break;
      }
      $201 = HEAP32[$wline>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $202 = HEAP32[$19>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $203 = ($201|0)==($202|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $204 = $201; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($203) {
       $205 = (($i$0126) + 1)|0; //@line 821 "modules/graphics/opengl/Font.cpp"
       $curcolor$sroa$0$3 = $curcolor$sroa$0$1;$curcolor$sroa$6$3 = $curcolor$sroa$6$1;$curcolor$sroa$7$3 = $curcolor$sroa$7$1;$curcolor$sroa$8$3 = $curcolor$sroa$8$1;$curcolori$3 = $curcolori$1;$i$1 = $205;$width$1 = $width$0131;
      } else {
       $206 = ($lastspaceindex$0127|0)==(-1); //@line 822 "modules/graphics/opengl/Font.cpp"
       if ($206) {
        $curcolor$sroa$0$3 = $curcolor$sroa$0$1;$curcolor$sroa$6$3 = $curcolor$sroa$6$1;$curcolor$sroa$7$3 = $curcolor$sroa$7$1;$curcolor$sroa$8$3 = $curcolor$sroa$8$1;$curcolori$3 = $curcolori$1;$i$1 = $i$0126;$width$1 = $width$0131;
       } else {
        $207 = $202; //@line 820 "modules/graphics/opengl/Font.cpp"
        $209 = $202;$470 = $207;
        while(1) {
         $208 = ((($209)) + -4|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $210 = HEAP32[$208>>2]|0; //@line 826 "modules/graphics/opengl/Font.cpp"
         $211 = ($210|0)==(32); //@line 826 "modules/graphics/opengl/Font.cpp"
         if ($211) {
          $216 = $470;
          break;
         }
         HEAP32[$19>>2] = $208; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $218 = ($201|0)==($208|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $219 = $208; //@line 826 "modules/graphics/opengl/Font.cpp"
         if ($218) {
          $216 = $219;
          break;
         } else {
          $209 = $208;$470 = $219;
         }
        }
        $212 = HEAP32[$20>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $213 = HEAP32[$21>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $214 = ($212|0)==($213|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        L85: do {
         if (!($214)) {
          $215 = (($216) - ($204))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $217 = $215 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $223 = $213;
          while(1) {
           $222 = ((($223)) + -4|0); //@line 829 "modules/graphics/opengl/Font.cpp"
           $224 = HEAP32[$222>>2]|0; //@line 829 "modules/graphics/opengl/Font.cpp"
           $225 = ($224|0)<($217|0); //@line 829 "modules/graphics/opengl/Font.cpp"
           if ($225) {
            break L85;
           }
           $scevgep4$i$i$i61 = ((($223)) + -8|0);
           HEAP32[$21>>2] = $scevgep4$i$i$i61; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $226 = ($212|0)==($scevgep4$i$i$i61|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           if ($226) {
            break;
           } else {
            $223 = $scevgep4$i$i$i61;
           }
          }
         }
        } while(0);
        $220 = ($curcolori$1|0)>(-1); //@line 833 "modules/graphics/opengl/Font.cpp"
        L91: do {
         if ($220) {
          $221 = HEAP32[$4>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $colori$0118 = $curcolori$1;
          while(1) {
           $227 = (((($221) + ($colori$0118<<3)|0)) + 4|0); //@line 835 "modules/graphics/opengl/Font.cpp"
           $228 = HEAP32[$227>>2]|0; //@line 835 "modules/graphics/opengl/Font.cpp"
           $229 = ($228|0)>($lastspaceindex$0127|0); //@line 835 "modules/graphics/opengl/Font.cpp"
           if (!($229)) {
            $colori$0118$lcssa = $colori$0118;
            break;
           }
           $239 = (($colori$0118) + -1)|0; //@line 833 "modules/graphics/opengl/Font.cpp"
           $240 = ($colori$0118|0)>(0); //@line 833 "modules/graphics/opengl/Font.cpp"
           if ($240) {
            $colori$0118 = $239;
           } else {
            $curcolor$sroa$0$2 = $curcolor$sroa$0$1;$curcolor$sroa$6$2 = $curcolor$sroa$6$1;$curcolor$sroa$7$2 = $curcolor$sroa$7$1;$curcolor$sroa$8$2 = $curcolor$sroa$8$1;$curcolori$2 = $curcolori$1;
            break L91;
           }
          }
          $230 = (($221) + ($colori$0118$lcssa<<3)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $231 = HEAPU8[$230>>0]|(HEAPU8[$230+1>>0]<<8)|(HEAPU8[$230+2>>0]<<16)|(HEAPU8[$230+3>>0]<<24); //@line 837 "modules/graphics/opengl/Font.cpp"
          $232 = $231&255; //@line 837 "modules/graphics/opengl/Font.cpp"
          $233 = $231 >>> 8; //@line 837 "modules/graphics/opengl/Font.cpp"
          $234 = $233&255; //@line 837 "modules/graphics/opengl/Font.cpp"
          $235 = $231 >>> 16; //@line 837 "modules/graphics/opengl/Font.cpp"
          $236 = $235&255; //@line 837 "modules/graphics/opengl/Font.cpp"
          $237 = $231 >>> 24; //@line 837 "modules/graphics/opengl/Font.cpp"
          $238 = $237&255; //@line 837 "modules/graphics/opengl/Font.cpp"
          $curcolor$sroa$0$2 = $232;$curcolor$sroa$6$2 = $234;$curcolor$sroa$7$2 = $236;$curcolor$sroa$8$2 = $238;$curcolori$2 = $colori$0118$lcssa;
         } else {
          $curcolor$sroa$0$2 = $curcolor$sroa$0$1;$curcolor$sroa$6$2 = $curcolor$sroa$6$1;$curcolor$sroa$7$2 = $curcolor$sroa$7$1;$curcolor$sroa$8$2 = $curcolor$sroa$8$1;$curcolori$2 = $curcolori$1;
         }
        } while(0);
        $241 = (($lastspaceindex$0127) + 1)|0; //@line 847 "modules/graphics/opengl/Font.cpp"
        $curcolor$sroa$0$3 = $curcolor$sroa$0$2;$curcolor$sroa$6$3 = $curcolor$sroa$6$2;$curcolor$sroa$7$3 = $curcolor$sroa$7$2;$curcolor$sroa$8$3 = $curcolor$sroa$8$2;$curcolori$3 = $curcolori$2;$i$1 = $241;$width$1 = $widthbeforelastspace$0130;
       }
      }
      $242 = HEAP32[$16>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $243 = HEAP32[$17>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $244 = ($242|0)==($243|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($244) {
       __THREW__ = 0;
       invoke_vii(4033,($lines|0),($wline|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $303 = __THREW__; __THREW__ = 0;
       $304 = $303&1;
       if ($304) {
        label = 7;
        break L3;
       }
      } else {
       __THREW__ = 0;
       invoke_vii(4032,($242|0),($wline|0)); //@line 76 "modules/graphics/opengl/Font.h"
       $245 = __THREW__; __THREW__ = 0;
       $246 = $245&1;
       if ($246) {
        label = 7;
        break L3;
       }
       $247 = ((($242)) + 12|0); //@line 76 "modules/graphics/opengl/Font.h"
       HEAP32[$247>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $248 = ((($242)) + 16|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$248>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $249 = ((($242)) + 20|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$249>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $250 = HEAP32[$21>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $251 = HEAP32[$24>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $252 = (($250) - ($251))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $253 = $252 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $254 = ($253|0)==(0); //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($254)) {
        $255 = ($253>>>0)>(536870911); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($255) {
         __THREW__ = 0;
         invoke_vi(3528,($247|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $256 = __THREW__; __THREW__ = 0;
         $257 = $256&1;
         if ($257) {
          $$lcssa = $242;$$lcssa186 = $247;$$lcssa187 = $248;
          label = 88;
          break L3;
         }
        }
        __THREW__ = 0;
        $258 = (invoke_ii(3348,($252|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        $259 = __THREW__; __THREW__ = 0;
        $260 = $259&1;
        if ($260) {
         $$lcssa = $242;$$lcssa186 = $247;$$lcssa187 = $248;
         label = 88;
         break L3;
        }
        HEAP32[$248>>2] = $258; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$247>>2] = $258; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $261 = (($258) + ($253<<3)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $262 = ((($242)) + 20|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        HEAP32[$262>>2] = $261; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $263 = HEAP32[$24>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $264 = HEAP32[$21>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $265 = ($263|0)==($264|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($265)) {
         $$01$i$i88 = $263;$273 = $258;
         while(1) {
          $266 = $$01$i$i88; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $267 = $266; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $268 = HEAP32[$267>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $269 = (($266) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $270 = $269; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $271 = HEAP32[$270>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $272 = $273; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $274 = $272; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          HEAP32[$274>>2] = $268; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $275 = (($272) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $276 = $275; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          HEAP32[$276>>2] = $271; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $277 = HEAP32[$248>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $278 = ((($277)) + 8|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$248>>2] = $278; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $279 = ((($$01$i$i88)) + 8|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $280 = ($279|0)==($264|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          if ($280) {
           break;
          } else {
           $$01$i$i88 = $279;$273 = $278;
          }
         }
        }
       }
       $301 = HEAP32[$16>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $302 = ((($301)) + 24|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$16>>2] = $302; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      do {
       if (!($18)) {
        $305 = (~~(($width$1))); //@line 853 "modules/graphics/opengl/Font.cpp"
        HEAP32[$1>>2] = $305; //@line 853 "modules/graphics/opengl/Font.cpp"
        $306 = HEAP32[$22>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $307 = HEAP32[$23>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $308 = ($306>>>0)<($307>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($308) {
         HEAP32[$306>>2] = $305; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $309 = ((($306)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$22>>2] = $309; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         break;
        } else {
         __THREW__ = 0;
         invoke_vii(3406,($linewidths|0),($1|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $310 = __THREW__; __THREW__ = 0;
         $311 = $310&1;
         if ($311) {
          label = 7;
          break L3;
         } else {
          break;
         }
        }
       }
      } while(0);
      $312 = HEAP32[$wline>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $313 = HEAP32[$19>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $314 = ($313|0)==($312|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($314)) {
       $__new_last2$i$i$i68 = $312; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $scevgep$i$i$i69 = ((($313)) + -4|0);
       $315 = $scevgep$i$i$i69;
       $316 = (($315) - ($__new_last2$i$i$i68))|0;
       $317 = $316 >>> 2;
       $318 = $317 ^ -1;
       $scevgep4$i$i$i70 = (($313) + ($318<<2)|0);
       HEAP32[$19>>2] = $scevgep4$i$i$i70; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      $319 = HEAP32[$20>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $320 = HEAP32[$21>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $321 = ($320|0)==($319|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($321) {
       $addcurcolor$4$off0 = 1;$curcolor$sroa$0$4 = $curcolor$sroa$0$3;$curcolor$sroa$6$4 = $curcolor$sroa$6$3;$curcolor$sroa$7$4 = $curcolor$sroa$7$3;$curcolor$sroa$8$4 = $curcolor$sroa$8$3;$curcolori$5 = $curcolori$3;$i$3 = $i$1;$lastspaceindex$3 = -1;$prevglyph$2 = 0;$width$3 = 0.0;$widthbeforelastspace$3 = 0.0;$widthoftrailingspace$3 = 0.0;
      } else {
       $__new_last1$i$i$i63 = $319; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $scevgep$i$i$i64 = ((($320)) + -8|0);
       $322 = $scevgep$i$i$i64;
       $323 = (($322) - ($__new_last1$i$i$i63))|0;
       $324 = $323 >>> 3;
       $325 = $324 ^ -1;
       $scevgep4$i$i$i65 = (($320) + ($325<<3)|0);
       HEAP32[$21>>2] = $scevgep4$i$i$i65; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $addcurcolor$4$off0 = 1;$curcolor$sroa$0$4 = $curcolor$sroa$0$3;$curcolor$sroa$6$4 = $curcolor$sroa$6$3;$curcolor$sroa$7$4 = $curcolor$sroa$7$3;$curcolor$sroa$8$4 = $curcolor$sroa$8$3;$curcolori$5 = $curcolori$3;$i$3 = $i$1;$lastspaceindex$3 = -1;$prevglyph$2 = 0;$width$3 = 0.0;$widthbeforelastspace$3 = 0.0;$widthoftrailingspace$3 = 0.0;
      }
     }
    } while(0);
    $367 = HEAP32[$11>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $368 = HEAP32[$codepoints>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $369 = (($367) - ($368))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $370 = $369 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $371 = ($i$3|0)<($370|0); //@line 775 "modules/graphics/opengl/Font.cpp"
    if ($371) {
     $$in = $368;$addcurcolor$0$off0125 = $addcurcolor$4$off0;$curcolor$sroa$0$0123 = $curcolor$sroa$0$4;$curcolor$sroa$6$0122 = $curcolor$sroa$6$4;$curcolor$sroa$7$0121 = $curcolor$sroa$7$4;$curcolor$sroa$8$0120 = $curcolor$sroa$8$4;$curcolori$0124 = $curcolori$5;$i$0126 = $i$3;$lastspaceindex$0127 = $lastspaceindex$3;$prevglyph$0128 = $prevglyph$2;$width$0131 = $width$3;$widthbeforelastspace$0130 = $widthbeforelastspace$3;$widthoftrailingspace$0129 = $widthoftrailingspace$3;
    } else {
     $width$3$lcssa = $width$3;$widthoftrailingspace$3$lcssa = $widthoftrailingspace$3;
     label = 118;
     break;
    }
   }
   L125: do {
    if ((label|0) == 7) {
     $48 = ___cxa_find_matching_catch_2()|0;
     $49 = tempRet0;
     $eh$lpad$body$index8Z2D = $49;$eh$lpad$body$indexZ2D = $48;
    }
    else if ((label|0) == 28) {
     $112 = ___cxa_find_matching_catch_2()|0;
     $113 = tempRet0;
     $114 = HEAP32[$$lcssa189>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $115 = ($114|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($115)) {
      $116 = HEAP32[$$lcssa190>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $117 = ($116|0)==($114|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($117)) {
       $__new_last1$i$i$i$i46 = $114; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $scevgep$i$i$i$i47 = ((($116)) + -8|0);
       $118 = $scevgep$i$i$i$i47;
       $119 = (($118) - ($__new_last1$i$i$i$i46))|0;
       $120 = $119 >>> 3;
       $121 = $120 ^ -1;
       $scevgep4$i$i$i$i48 = (($116) + ($121<<3)|0);
       HEAP32[$$lcssa190>>2] = $scevgep4$i$i$i$i48; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($114); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $122 = HEAP32[$$lcssa188>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $123 = ($122|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $124 = $122; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($123) {
      $eh$lpad$body$index8Z2D = $113;$eh$lpad$body$indexZ2D = $112;
     } else {
      $125 = ((($$lcssa188)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $126 = HEAP32[$125>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $127 = ($126|0)==($122|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($127)) {
       $scevgep$i$i$i$i$i$i$i$i30 = ((($126)) + -4|0);
       $128 = $scevgep$i$i$i$i$i$i$i$i30;
       $129 = (($128) - ($124))|0;
       $130 = $129 >>> 2;
       $131 = $130 ^ -1;
       $scevgep4$i$i$i$i$i$i$i$i31 = (($126) + ($131<<2)|0);
       HEAP32[$125>>2] = $scevgep4$i$i$i$i$i$i$i$i31; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($122); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $eh$lpad$body$index8Z2D = $113;$eh$lpad$body$indexZ2D = $112;
     }
    }
    else if ((label|0) == 88) {
     $281 = ___cxa_find_matching_catch_2()|0;
     $282 = tempRet0;
     $283 = HEAP32[$$lcssa186>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $284 = ($283|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($284)) {
      $285 = HEAP32[$$lcssa187>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $286 = ($285|0)==($283|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($286)) {
       $__new_last1$i$i$i$i89 = $283; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $scevgep$i$i$i$i90 = ((($285)) + -8|0);
       $287 = $scevgep$i$i$i$i90;
       $288 = (($287) - ($__new_last1$i$i$i$i89))|0;
       $289 = $288 >>> 3;
       $290 = $289 ^ -1;
       $scevgep4$i$i$i$i91 = (($285) + ($290<<3)|0);
       HEAP32[$$lcssa187>>2] = $scevgep4$i$i$i$i91; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($283); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $291 = HEAP32[$$lcssa>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $292 = ($291|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $293 = $291; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($292) {
      $eh$lpad$body$index8Z2D = $282;$eh$lpad$body$indexZ2D = $281;
     } else {
      $294 = ((($$lcssa)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $295 = HEAP32[$294>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $296 = ($295|0)==($291|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($296)) {
       $scevgep$i$i$i$i$i$i$i$i75 = ((($295)) + -4|0);
       $297 = $scevgep$i$i$i$i$i$i$i$i75;
       $298 = (($297) - ($293))|0;
       $299 = $298 >>> 2;
       $300 = $299 ^ -1;
       $scevgep4$i$i$i$i$i$i$i$i76 = (($295) + ($300<<2)|0);
       HEAP32[$294>>2] = $scevgep4$i$i$i$i$i$i$i$i76; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($291); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $eh$lpad$body$index8Z2D = $282;$eh$lpad$body$indexZ2D = $281;
     }
    }
    else if ((label|0) == 118) {
     $$pre = HEAP32[$wline>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $$pre145 = HEAP32[$19>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $372 = ($$pre|0)==($$pre145|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($372) {
      $$pre$phi148152Z2D = $wline;$$pre$phi150151Z2D = $19;
      break L1;
     }
     $373 = ((($lines)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $374 = HEAP32[$373>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $375 = ((($lines)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $376 = HEAP32[$375>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $377 = ($374|0)==($376|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($377) {
      __THREW__ = 0;
      invoke_vii(4033,($lines|0),($wline|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $438 = __THREW__; __THREW__ = 0;
      $439 = $438&1;
      if (!($439)) {
       label = 138;
      }
     } else {
      __THREW__ = 0;
      invoke_vii(4032,($374|0),($wline|0)); //@line 76 "modules/graphics/opengl/Font.h"
      $378 = __THREW__; __THREW__ = 0;
      $379 = $378&1;
      if (!($379)) {
       $380 = ((($374)) + 12|0); //@line 76 "modules/graphics/opengl/Font.h"
       $381 = ((($wline)) + 12|0); //@line 76 "modules/graphics/opengl/Font.h"
       HEAP32[$380>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $382 = ((($374)) + 16|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$382>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $383 = ((($374)) + 20|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$383>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $384 = ((($wline)) + 16|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $385 = HEAP32[$384>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $386 = HEAP32[$381>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $387 = (($385) - ($386))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $388 = $387 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $389 = ($388|0)==(0); //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       L132: do {
        if (!($389)) {
         $390 = ($388>>>0)>(536870911); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($390) {
          __THREW__ = 0;
          invoke_vi(3528,($380|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $391 = __THREW__; __THREW__ = 0;
          $392 = $391&1;
          if (!($392)) {
           label = 124;
          }
         } else {
          label = 124;
         }
         if ((label|0) == 124) {
          __THREW__ = 0;
          $393 = (invoke_ii(3348,($387|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
          $394 = __THREW__; __THREW__ = 0;
          $395 = $394&1;
          if (!($395)) {
           HEAP32[$382>>2] = $393; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           HEAP32[$380>>2] = $393; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $396 = (($393) + ($388<<3)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $397 = ((($374)) + 20|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           HEAP32[$397>>2] = $396; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $398 = HEAP32[$381>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $399 = HEAP32[$384>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $400 = ($398|0)==($399|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           if ($400) {
            break;
           }
           $$01$i$i = $398;$408 = $393;
           while(1) {
            $401 = $$01$i$i; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
            $402 = $401; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
            $403 = HEAP32[$402>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
            $404 = (($401) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
            $405 = $404; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
            $406 = HEAP32[$405>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
            $407 = $408; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
            $409 = $407; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
            HEAP32[$409>>2] = $403; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
            $410 = (($407) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
            $411 = $410; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
            HEAP32[$411>>2] = $406; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
            $412 = HEAP32[$382>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $413 = ((($412)) + 8|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            HEAP32[$382>>2] = $413; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $414 = ((($$01$i$i)) + 8|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $415 = ($414|0)==($399|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            if ($415) {
             break L132;
            } else {
             $$01$i$i = $414;$408 = $413;
            }
           }
          }
         }
         $416 = ___cxa_find_matching_catch_2()|0;
         $417 = tempRet0;
         $418 = HEAP32[$380>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $419 = ($418|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if (!($419)) {
          $420 = HEAP32[$382>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $421 = ($420|0)==($418|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          if (!($421)) {
           $__new_last1$i$i$i$i11 = $418; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $scevgep$i$i$i$i12 = ((($420)) + -8|0);
           $422 = $scevgep$i$i$i$i12;
           $423 = (($422) - ($__new_last1$i$i$i$i11))|0;
           $424 = $423 >>> 3;
           $425 = $424 ^ -1;
           $scevgep4$i$i$i$i13 = (($420) + ($425<<3)|0);
           HEAP32[$382>>2] = $scevgep4$i$i$i$i13; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          }
          __ZdlPv($418); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
         }
         $426 = HEAP32[$374>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $427 = ($426|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $428 = $426; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($427) {
          $eh$lpad$body$index8Z2D = $417;$eh$lpad$body$indexZ2D = $416;
          break L125;
         }
         $429 = ((($374)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $430 = HEAP32[$429>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $431 = ($430|0)==($426|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if (!($431)) {
          $scevgep$i$i$i$i$i$i$i$i = ((($430)) + -4|0);
          $432 = $scevgep$i$i$i$i$i$i$i$i;
          $433 = (($432) - ($428))|0;
          $434 = $433 >>> 2;
          $435 = $434 ^ -1;
          $scevgep4$i$i$i$i$i$i$i$i = (($430) + ($435<<2)|0);
          HEAP32[$429>>2] = $scevgep4$i$i$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         }
         __ZdlPv($426); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
         $eh$lpad$body$index8Z2D = $417;$eh$lpad$body$indexZ2D = $416;
         break L125;
        }
       } while(0);
       $436 = HEAP32[$373>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $437 = ((($436)) + 24|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$373>>2] = $437; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       label = 138;
      }
     }
     do {
      if ((label|0) == 138) {
       $440 = ($linewidths|0)==(0|0); //@line 899 "modules/graphics/opengl/Font.cpp"
       if ($440) {
        $$pre$phi148152Z2D = $wline;$$pre$phi150151Z2D = $19;
        break L1;
       }
       $441 = $width$3$lcssa - $widthoftrailingspace$3$lcssa; //@line 900 "modules/graphics/opengl/Font.cpp"
       $442 = (~~(($441))); //@line 900 "modules/graphics/opengl/Font.cpp"
       HEAP32[$3>>2] = $442; //@line 900 "modules/graphics/opengl/Font.cpp"
       $443 = ((($linewidths)) + 4|0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $444 = HEAP32[$443>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $445 = ((($linewidths)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $446 = HEAP32[$445>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $447 = ($444>>>0)<($446>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($447) {
        HEAP32[$444>>2] = $442; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $448 = ((($444)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$443>>2] = $448; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $$pre$phi148152Z2D = $wline;$$pre$phi150151Z2D = $19;
        break L1;
       } else {
        __THREW__ = 0;
        invoke_vii(3406,($linewidths|0),($3|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $449 = __THREW__; __THREW__ = 0;
        $450 = $449&1;
        if ($450) {
         break;
        } else {
         $$pre$phi148152Z2D = $wline;$$pre$phi150151Z2D = $19;
         break L1;
        }
       }
      }
     } while(0);
     $50 = ___cxa_find_matching_catch_2()|0;
     $51 = tempRet0;
     $eh$lpad$body$index8Z2D = $51;$eh$lpad$body$indexZ2D = $50;
    }
   } while(0);
   $52 = ((($wline)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $53 = HEAP32[$52>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $54 = ($53|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($54)) {
    $55 = ((($wline)) + 16|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $56 = HEAP32[$55>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $57 = ($56|0)==($53|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($57)) {
     $__new_last1$i$i$i$i19 = $53; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $scevgep$i$i$i$i20 = ((($56)) + -8|0);
     $58 = $scevgep$i$i$i$i20;
     $59 = (($58) - ($__new_last1$i$i$i$i19))|0;
     $60 = $59 >>> 3;
     $61 = $60 ^ -1;
     $scevgep4$i$i$i$i21 = (($56) + ($61<<3)|0);
     HEAP32[$55>>2] = $scevgep4$i$i$i$i21; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($53); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $62 = HEAP32[$wline>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $63 = ($62|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $64 = $62; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($63) {
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
   $65 = ((($wline)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $66 = HEAP32[$65>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $67 = ($66|0)==($62|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($67)) {
    $scevgep$i$i$i1$i25 = ((($66)) + -4|0);
    $68 = $scevgep$i$i$i1$i25;
    $69 = (($68) - ($64))|0;
    $70 = $69 >>> 2;
    $71 = $70 ^ -1;
    $scevgep4$i$i$i2$i26 = (($66) + ($71<<2)|0);
    HEAP32[$65>>2] = $scevgep4$i$i$i2$i26; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($62); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($eh$lpad$body$indexZ2D|0);
   // unreachable;
  } else {
   $$pre149 = ((($wline)) + 4|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$pre$phi148152Z2D = $wline;$$pre$phi150151Z2D = $$pre149;
  }
 } while(0);
 $451 = ((($wline)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $452 = HEAP32[$451>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $453 = ($452|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($453)) {
  $454 = ((($wline)) + 16|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $455 = HEAP32[$454>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $456 = ($455|0)==($452|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($456)) {
   $__new_last1$i$i$i$i = $452; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $scevgep$i$i$i$i = ((($455)) + -8|0);
   $457 = $scevgep$i$i$i$i;
   $458 = (($457) - ($__new_last1$i$i$i$i))|0;
   $459 = $458 >>> 3;
   $460 = $459 ^ -1;
   $scevgep4$i$i$i$i = (($455) + ($460<<3)|0);
   HEAP32[$454>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($452); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $461 = HEAP32[$$pre$phi148152Z2D>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $462 = ($461|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $463 = $461; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($462) {
  STACKTOP = sp;return; //@line 902 "modules/graphics/opengl/Font.cpp"
 }
 $464 = HEAP32[$$pre$phi150151Z2D>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $465 = ($464|0)==($461|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($465)) {
  $scevgep$i$i$i1$i = ((($464)) + -4|0);
  $466 = $scevgep$i$i$i1$i;
  $467 = (($466) - ($463))|0;
  $468 = $467 >>> 2;
  $469 = $468 ^ -1;
  $scevgep4$i$i$i2$i = (($464) + ($469<<2)|0);
  HEAP32[$$pre$phi150151Z2D>>2] = $scevgep4$i$i$i2$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($461); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 STACKTOP = sp;return; //@line 902 "modules/graphics/opengl/Font.cpp"
}
function __ZNSt3__16vectorIjNS_9allocatorIjEEEC2ERKS3_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$01$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $__first3$i = 0, $scevgep$i = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, $scevgep5$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $0 = ((($this)) + 4|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($this)) + 8|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($__x)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$__x>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = $5 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ($6|0)==(0); //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($7) {
  return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $8 = ($6>>>0)>(1073741823); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($8) {
  __THREW__ = 0;
  invoke_vi(3528,($this|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if (!($10)) {
   label = 4;
  }
 } else {
  label = 4;
 }
 if ((label|0) == 4) {
  __THREW__ = 0;
  $11 = (invoke_ii(3348,($5|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   HEAP32[$0>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$this>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $14 = (($11) + ($6<<2)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $15 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$15>>2] = $14; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $16 = HEAP32[$__x>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $17 = HEAP32[$2>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $18 = ($16|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($18) {
    return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $__first3$i = $16; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $scevgep$i = ((($17)) + -4|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $19 = $scevgep$i; //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $20 = (($19) - ($__first3$i))|0; //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $21 = $20 >>> 2; //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$01$i = $16;$23 = $11;
   while(1) {
    $22 = HEAP32[$$01$i>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$23>>2] = $22; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $24 = ((($23)) + 4|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $25 = ((($$01$i)) + 4|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $26 = ($25|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($26) {
     break;
    } else {
     $$01$i = $25;$23 = $24;
    }
   }
   $27 = (($21) + 1)|0; //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $scevgep5$i = (($11) + ($27<<2)|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$0>>2] = $scevgep5$i; //@line 1030 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
 }
 $28 = ___cxa_find_matching_catch_2()|0;
 $29 = tempRet0;
 $30 = HEAP32[$this>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = ($30|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = $30; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($31) {
  ___resumeException($28|0);
  // unreachable;
 }
 $33 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $34 = ($33|0)==($30|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($34)) {
  $scevgep$i$i$i = ((($33)) + -4|0);
  $35 = $scevgep$i$i$i;
  $36 = (($35) - ($32))|0;
  $37 = $36 >>> 2;
  $38 = $37 ^ -1;
  $scevgep4$i$i$i = (($33) + ($38<<2)|0);
  HEAP32[$0>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($30); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($28|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love8graphics6opengl4Font17ColoredCodepointsENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i15 = 0, $$0$i16 = 0, $$01$i$i = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre17 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $__new_last1$i$i$i$i$i$i$i$i$i$i$i2 = 0, $__v$sroa$6$0 = 0, $eh$lpad$body18$index3Z2D = 0, $eh$lpad$body18$indexZ2D = 0, $eh$lpad$body19$index9Z2D = 0, $eh$lpad$body19$indexZ2D = 0, $scevgep$i$i$i$i$i$i$i = 0, $scevgep$i$i$i$i$i$i$i$i$i$i$i3 = 0, $scevgep$i$i$i1$i$i$i$i$i$i$i$i9 = 0, $scevgep4$i$i$i$i$i$i$i = 0, $scevgep4$i$i$i$i$i$i$i$i$i$i$i4 = 0, $scevgep4$i$i$i2$i$i$i$i$i$i$i$i10 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($3|0) / 24)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(178956970); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $10 = $$pre;
 } else {
  $10 = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($10))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = (($9|0) / 24)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($11>>>0)<(89478485); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $16 = $11 << 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = ($16>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $18 = $17 ? $5 : $16; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = (($19) - ($10))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = (($20|0) / 24)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = ($18|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($22) {
   $$0$i16 = 0;$26 = 0;$27 = $21;
  } else {
   $$0$i15 = $18;$113 = $21;
   label = 6;
  }
 } else {
  $13 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) - ($10))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = (($14|0) / 24)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i15 = 178956970;$113 = $15;
  label = 6;
 }
 if ((label|0) == 6) {
  $23 = ($$0$i15*24)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = (__Znwj($23)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i16 = $$0$i15;$26 = $24;$27 = $113;
 }
 $25 = (($26) + (($27*24)|0)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $25; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = (($26) + (($$0$i16*24)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = $29; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 __THREW__ = 0;
 invoke_vii(4032,($25|0),($__x|0)); //@line 76 "modules/graphics/opengl/Font.h"
 $31 = __THREW__; __THREW__ = 0;
 $32 = $31&1;
 do {
  if ($32) {
   $110 = ___cxa_find_matching_catch_2()|0;
   $111 = tempRet0;
   $112 = ($26|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($112) {
    $eh$lpad$body19$index9Z2D = $111;$eh$lpad$body19$indexZ2D = $110;
    ___resumeException($eh$lpad$body19$indexZ2D|0);
    // unreachable;
   } else {
    $eh$lpad$body18$index3Z2D = $111;$eh$lpad$body18$indexZ2D = $110;
   }
  } else {
   $33 = (((($26) + (($27*24)|0)|0)) + 12|0); //@line 76 "modules/graphics/opengl/Font.h"
   $34 = ((($__x)) + 12|0); //@line 76 "modules/graphics/opengl/Font.h"
   __THREW__ = 0;
   invoke_vii(4035,($33|0),($34|0)); //@line 76 "modules/graphics/opengl/Font.h"
   $35 = __THREW__; __THREW__ = 0;
   $36 = $35&1;
   if ($36) {
    $37 = ___cxa_find_matching_catch_2()|0;
    $38 = tempRet0;
    $39 = HEAP32[$25>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $40 = ($39|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $41 = $39; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($40) {
     $eh$lpad$body18$index3Z2D = $38;$eh$lpad$body18$indexZ2D = $37;
     break;
    }
    $42 = (((($26) + (($27*24)|0)|0)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $43 = HEAP32[$42>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $44 = ($43|0)==($39|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($44)) {
     $scevgep$i$i$i$i$i$i$i = ((($43)) + -4|0);
     $45 = $scevgep$i$i$i$i$i$i$i;
     $46 = (($45) - ($41))|0;
     $47 = $46 >>> 2;
     $48 = $47 ^ -1;
     $scevgep4$i$i$i$i$i$i$i = (($43) + ($48<<2)|0);
     HEAP32[$42>>2] = $scevgep4$i$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($39); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $eh$lpad$body18$index3Z2D = $38;$eh$lpad$body18$indexZ2D = $37;
    break;
   }
   $49 = ((($25)) + 24|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $50 = $49; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $51 = HEAP32[$this>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $52 = HEAP32[$0>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $53 = ($52|0)==($51|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if ($53) {
    $54 = $51; //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $80 = $28;$82 = $54;$84 = $51;
   } else {
    $$01$i$i = $52;$56 = $25;$__v$sroa$6$0 = $28;
    while(1) {
     $55 = ((($56)) + -24|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $57 = ((($$01$i$i)) + -24|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$55>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $58 = ((($56)) + -20|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$58>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $59 = ((($56)) + -16|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$59>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $60 = HEAP32[$57>>2]|0; //@line 1246 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$55>>2] = $60; //@line 1246 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $61 = ((($$01$i$i)) + -20|0); //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $62 = HEAP32[$61>>2]|0; //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$58>>2] = $62; //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $63 = ((($$01$i$i)) + -16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $64 = HEAP32[$63>>2]|0; //@line 1248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $65 = ((($56)) + -16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$65>>2] = $64; //@line 1248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$63>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$61>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$57>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $66 = ((($56)) + -12|0); //@line 76 "modules/graphics/opengl/Font.h"
     $67 = ((($$01$i$i)) + -12|0); //@line 76 "modules/graphics/opengl/Font.h"
     HEAP32[$66>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $68 = ((($56)) + -8|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$68>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $69 = ((($56)) + -4|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$69>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $70 = HEAP32[$67>>2]|0; //@line 1246 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$66>>2] = $70; //@line 1246 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $71 = ((($$01$i$i)) + -8|0); //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $72 = HEAP32[$71>>2]|0; //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$68>>2] = $72; //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $73 = ((($$01$i$i)) + -4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $74 = HEAP32[$73>>2]|0; //@line 1248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $75 = ((($56)) + -4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$75>>2] = $74; //@line 1248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$73>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$71>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$67>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $76 = $__v$sroa$6$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $77 = ((($76)) + -24|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $78 = $77; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $79 = ($57|0)==($51|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     if ($79) {
      $$lcssa = $78;
      break;
     } else {
      $$01$i$i = $57;$56 = $77;$__v$sroa$6$0 = $78;
     }
    }
    $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $$pre17 = HEAP32[$0>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $80 = $$lcssa;$82 = $$pre$i;$84 = $$pre17;
   }
   HEAP32[$this>>2] = $80; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$0>>2] = $50; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$7>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $81 = $82; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $83 = ($84|0)==($81|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if (!($83)) {
    $86 = $84;
    while(1) {
     $85 = ((($86)) + -24|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $87 = ((($86)) + -12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $88 = HEAP32[$87>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $89 = ($88|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($89)) {
      $90 = ((($86)) + -8|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $91 = HEAP32[$90>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $92 = ($91|0)==($88|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($92)) {
       $__new_last1$i$i$i$i$i$i$i$i$i$i$i2 = $88; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $scevgep$i$i$i$i$i$i$i$i$i$i$i3 = ((($91)) + -8|0);
       $93 = $scevgep$i$i$i$i$i$i$i$i$i$i$i3;
       $94 = (($93) - ($__new_last1$i$i$i$i$i$i$i$i$i$i$i2))|0;
       $95 = $94 >>> 3;
       $96 = $95 ^ -1;
       $scevgep4$i$i$i$i$i$i$i$i$i$i$i4 = (($91) + ($96<<3)|0);
       HEAP32[$90>>2] = $scevgep4$i$i$i$i$i$i$i$i$i$i$i4; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($88); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $97 = HEAP32[$85>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $98 = ($97|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $99 = $97; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($98)) {
      $101 = ((($86)) + -20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $102 = HEAP32[$101>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $103 = ($102|0)==($97|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($103)) {
       $scevgep$i$i$i1$i$i$i$i$i$i$i$i9 = ((($102)) + -4|0);
       $104 = $scevgep$i$i$i1$i$i$i$i$i$i$i$i9;
       $105 = (($104) - ($99))|0;
       $106 = $105 >>> 2;
       $107 = $106 ^ -1;
       $scevgep4$i$i$i2$i$i$i$i$i$i$i$i10 = (($102) + ($107<<2)|0);
       HEAP32[$101>>2] = $scevgep4$i$i$i2$i$i$i$i$i$i$i$i10; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($97); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $100 = ($85|0)==($81|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     if ($100) {
      break;
     } else {
      $86 = $85;
     }
    }
   }
   $108 = ($82|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($108) {
    return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $109 = $82; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __ZdlPv($109); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
 } while(0);
 __ZdlPv($26); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $eh$lpad$body19$index9Z2D = $eh$lpad$body18$index3Z2D;$eh$lpad$body19$indexZ2D = $eh$lpad$body18$indexZ2D;
 ___resumeException($eh$lpad$body19$indexZ2D|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love8graphics6opengl4Font12IndexedColorENS_9allocatorIS5_EEEC2ERKS8_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$01$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__new_last1$i$i$i = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $0 = ((($this)) + 4|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($this)) + 8|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($__x)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$__x>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = $5 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ($6|0)==(0); //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($7) {
  return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $8 = ($6>>>0)>(536870911); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($8) {
  __THREW__ = 0;
  invoke_vi(3528,($this|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if (!($10)) {
   label = 4;
  }
 } else {
  label = 4;
 }
 if ((label|0) == 4) {
  __THREW__ = 0;
  $11 = (invoke_ii(3348,($5|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   HEAP32[$0>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$this>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $14 = (($11) + ($6<<3)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $15 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$15>>2] = $14; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $16 = HEAP32[$__x>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $17 = HEAP32[$2>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $18 = ($16|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($18) {
    return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $$01$i = $16;$26 = $11;
   while(1) {
    $19 = $$01$i; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $20 = $19; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $21 = HEAP32[$20>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $22 = (($19) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $23 = $22; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $24 = HEAP32[$23>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $25 = $26; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $27 = $25; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$27>>2] = $21; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $28 = (($25) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $29 = $28; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$29>>2] = $24; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $30 = HEAP32[$0>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $31 = ((($30)) + 8|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$0>>2] = $31; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $32 = ((($$01$i)) + 8|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $33 = ($32|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($33) {
     break;
    } else {
     $$01$i = $32;$26 = $31;
    }
   }
   return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
 }
 $34 = ___cxa_find_matching_catch_2()|0;
 $35 = tempRet0;
 $36 = HEAP32[$this>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $37 = ($36|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($37) {
  ___resumeException($34|0);
  // unreachable;
 }
 $38 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $39 = ($38|0)==($36|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($39)) {
  $__new_last1$i$i$i = $36; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $scevgep$i$i$i = ((($38)) + -8|0);
  $40 = $scevgep$i$i$i;
  $41 = (($40) - ($__new_last1$i$i$i))|0;
  $42 = $41 >>> 3;
  $43 = $42 ^ -1;
  $scevgep4$i$i$i = (($38) + ($43<<3)|0);
  HEAP32[$0>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($36); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($34|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love8graphics6opengl4Font12IndexedColorENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(536870911); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = $$in; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $8 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = HEAP32[$8>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $10 >> 3; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($11>>>0)<(268435455); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $16 = $10 >> 2; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = ($16>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $18 = $17 ? $5 : $16; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = (($19) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = $20 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = ($18|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($22) {
   $$0$i2 = 0;$26 = 0;$27 = $21;$43 = $19;
  } else {
   $$0$i1 = $18;$49 = $19;$50 = $21;
   label = 6;
  }
 } else {
  $13 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = $14 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 536870911;$49 = $13;$50 = $15;
  label = 6;
 }
 if ((label|0) == 6) {
  $23 = $$0$i1 << 3; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = (__Znwj($23)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$26 = $24;$27 = $50;$43 = $49;
 }
 $25 = (($26) + ($27<<3)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = (($26) + ($$0$i2<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = $28; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = $__x; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $31 = $30; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $32 = HEAP32[$31>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = (($30) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $33; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = HEAP32[$34>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = $25; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$37>>2] = $32; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = (($36) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $39 = $38; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$39>>2] = $35; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $40 = ((($25)) + 8|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $41 = $40; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $42 = (($43) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $44 = $42 >> 3; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $45 = (0 - ($44))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $46 = (($25) + ($45<<3)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = $46; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($46|0),($7|0),($42|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $47; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $41; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$8>>2] = $29; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $48 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($48) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($7); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNSt3__16vectorIN4love8graphics6opengl4Font11DrawCommandENS_9allocatorIS5_EEE6insertINS_11__wrap_iterIPS5_EEEENS_9enable_ifIXaasr21__is_forward_iteratorIT_EE5valuesr16is_constructibleIS5_NS_15iterator_traitsISE_E9referenceEEE5valueESC_E4typeENSA_IPKS5_EESE_SE_($this,$__position,$__first,$__last) {
 $this = $this|0;
 $__position = $__position|0;
 $__first = $__first|0;
 $__last = $__last|0;
 var $$0$i21 = 0, $$0$i22 = 0, $$in = 0, $$in$i = 0, $$in25 = 0, $$lcssa = 0, $$pre = 0, $$pre23 = 0, $$pre24 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $11 = 0, $12 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__i$01$i = 0, $__m$sroa$0$020 = 0, $__p$0 = 0, $__v$sroa$12$0 = 0, $__v$sroa$12$1 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 1957 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = $0; //@line 1957 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$__position>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $3 = (($2) - ($1))|0; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $4 = (($3|0) / 12)&-1; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $5 = (($0) + (($4*12)|0)|0); //@line 1957 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$__first>>2]|0; //@line 1958 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = HEAP32[$__last>>2]|0; //@line 1958 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = (($7) - ($6))|0; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $9 = (($8|0) / 12)&-1; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $10 = ($8|0)>(0); //@line 1959 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $6; //@line 1986 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = $7; //@line 1986 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($10)) {
  $__p$0 = $5;
  return ($__p$0|0); //@line 1991 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $13 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = HEAP32[$13>>2]|0; //@line 1961 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = ((($this)) + 4|0); //@line 1961 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = HEAP32[$15>>2]|0; //@line 1961 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = $16; //@line 1961 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $18 = (($14) - ($17))|0; //@line 1961 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $19 = (($18|0) / 12)&-1; //@line 1961 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = ($9|0)>($19|0); //@line 1961 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($20)) {
  $21 = $5; //@line 1966 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = (($17) - ($21))|0; //@line 1966 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = (($22|0) / 12)&-1; //@line 1966 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $24 = ($9|0)>($23|0); //@line 1967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($24) {
   $25 = (($11) + (($23*12)|0)|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $26 = $25; //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $27 = ($25|0)==($12|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($27) {
    $102 = $16;
   } else {
    $$in25 = $16;$28 = $25;
    while(1) {
     ;HEAP32[$$in25>>2]=HEAP32[$28>>2]|0;HEAP32[$$in25+4>>2]=HEAP32[$28+4>>2]|0;HEAP32[$$in25+8>>2]=HEAP32[$28+8>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $29 = HEAP32[$15>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $30 = ((($29)) + 12|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$15>>2] = $30; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $31 = ((($28)) + 12|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $32 = ($31|0)==($12|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($32) {
      $102 = $30;
      break;
     } else {
      $$in25 = $30;$28 = $31;
     }
    }
   }
   $33 = ($22|0)>(0); //@line 1974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($33) {
    $36 = $102;$__m$sroa$0$020 = $26;
   } else {
    $__p$0 = $5;
    return ($__p$0|0); //@line 1991 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
  } else {
   $36 = $16;$__m$sroa$0$020 = $7;
  }
  $34 = (($5) + (($9*12)|0)|0); //@line 1977 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $35 = $36; //@line 1710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $37 = $34; //@line 1710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $38 = (($35) - ($37))|0; //@line 1710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $39 = (($38|0) / 12)&-1; //@line 1710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $40 = (($5) + (($39*12)|0)|0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $41 = ($40>>>0)<($16>>>0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($41) {
   $$in$i = $36;$__i$01$i = $40;
   while(1) {
    ;HEAP32[$$in$i>>2]=HEAP32[$__i$01$i>>2]|0;HEAP32[$$in$i+4>>2]=HEAP32[$__i$01$i+4>>2]|0;HEAP32[$$in$i+8>>2]=HEAP32[$__i$01$i+8>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $42 = ((($__i$01$i)) + 12|0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $43 = HEAP32[$15>>2]|0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $44 = ((($43)) + 12|0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$15>>2] = $44; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $45 = ($42>>>0)<($16>>>0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($45) {
     $$in$i = $44;$__i$01$i = $42;
    } else {
     break;
    }
   }
  }
  $46 = (($38|0) / -12)&-1; //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $47 = (($36) + (($46*12)|0)|0); //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  _memmove(($47|0),($5|0),($38|0))|0; //@line 1929 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $48 = HEAP32[$__first>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $49 = $48; //@line 1763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $50 = (($__m$sroa$0$020) - ($49))|0; //@line 1763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  _memmove(($5|0),($48|0),($50|0))|0; //@line 1764 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $__p$0 = $5;
  return ($__p$0|0); //@line 1991 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $51 = (($17) - ($1))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $52 = (($51|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $53 = (($52) + ($9))|0; //@line 1985 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $54 = ($53>>>0)>(357913941); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($54) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$13>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre23 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $55 = $$pre23; //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $103 = $55;$57 = $$pre;$58 = $$pre23;
 } else {
  $103 = $0;$57 = $14;$58 = $1;
 }
 $56 = (($57) - ($58))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $59 = (($56|0) / 12)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $60 = ($59>>>0)<(178956970); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($60) {
  $64 = $59 << 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $65 = ($64>>>0)<($53>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $66 = $65 ? $53 : $64; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $67 = $5; //@line 1985 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $68 = (($67) - ($58))|0; //@line 1985 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $69 = (($68|0) / 12)&-1; //@line 1985 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $70 = ($66|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($70) {
   $$0$i22 = 0;$74 = 0;$75 = $69;$89 = $67;
  } else {
   $$0$i21 = $66;$104 = $69;$105 = $67;
   label = 15;
  }
 } else {
  $61 = $5; //@line 1985 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $62 = (($61) - ($58))|0; //@line 1985 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $63 = (($62|0) / 12)&-1; //@line 1985 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i21 = 357913941;$104 = $63;$105 = $61;
  label = 15;
 }
 if ((label|0) == 15) {
  $71 = ($$0$i21*12)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $72 = (__Znwj($71)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i22 = $$0$i21;$74 = $72;$75 = $104;$89 = $105;
 }
 $73 = (($74) + (($75*12)|0)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $76 = $73; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $77 = (($74) + (($$0$i22*12)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $78 = $77; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $79 = ($11|0)==($12|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($79) {
  $87 = $103;$__v$sroa$12$1 = $76;
 } else {
  $$in = $73;$80 = $11;$__v$sroa$12$0 = $76;
  while(1) {
   ;HEAP32[$$in>>2]=HEAP32[$80>>2]|0;HEAP32[$$in+4>>2]=HEAP32[$80+4>>2]|0;HEAP32[$$in+8>>2]=HEAP32[$80+8>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $81 = $__v$sroa$12$0; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $82 = ((($81)) + 12|0); //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $83 = $82; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $84 = ((($80)) + 12|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $85 = ($84|0)==($12|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($85) {
    $$lcssa = $83;
    break;
   } else {
    $$in = $82;$80 = $84;$__v$sroa$12$0 = $83;
   }
  }
  $$pre24 = HEAP32[$this>>2]|0; //@line 912 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $87 = $$pre24;$__v$sroa$12$1 = $$lcssa;
 }
 $86 = $87; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $88 = (($89) - ($86))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $90 = (($88|0) / -12)&-1; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $91 = (($73) + (($90*12)|0)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $92 = $91; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($91|0),($87|0),($88|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $93 = HEAP32[$15>>2]|0; //@line 913 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $94 = (($93) - ($89))|0; //@line 1522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $95 = (($94|0) / 12)&-1; //@line 1522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $96 = $__v$sroa$12$1; //@line 1523 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($96|0),($5|0),($94|0))|0; //@line 1523 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $97 = $__v$sroa$12$1; //@line 1524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $98 = (($97) + (($95*12)|0)|0); //@line 1524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $99 = $98; //@line 1524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $100 = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$this>>2] = $92; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$15>>2] = $99; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$13>>2] = $78; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $101 = ($100|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($101) {
  $__p$0 = $73;
  return ($__p$0|0); //@line 1991 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($100); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $__p$0 = $73;
 return ($__p$0|0); //@line 1991 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl4Font12drawVerticesERKNSt3__16vectorINS2_11DrawCommandENS3_9allocatorIS5_EEEEb($this,$drawcommands,$bufferedvertices) {
 $this = $this|0;
 $drawcommands = $drawcommands|0;
 $bufferedvertices = $bufferedvertices|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$02 = 0, $__begin2$sroa$0$01 = 0, $__begin2$sroa$0$01$us = 0;
 var $phitmp = 0, $totalverts$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = HEAP32[$drawcommands>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($drawcommands)) + 4|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($1|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 do {
  if ($4) {
   $5 = ((($this)) + 128|0); //@line 629 "modules/graphics/opengl/Font.cpp"
   $23 = $5;
  } else {
   $__begin$sroa$0$02 = $1;$totalverts$03 = 0;
   while(1) {
    $9 = ((($__begin$sroa$0$02)) + 4|0); //@line 627 "modules/graphics/opengl/Font.cpp"
    $10 = HEAP32[$9>>2]|0; //@line 627 "modules/graphics/opengl/Font.cpp"
    $11 = ((($__begin$sroa$0$02)) + 8|0); //@line 627 "modules/graphics/opengl/Font.cpp"
    $12 = HEAP32[$11>>2]|0; //@line 627 "modules/graphics/opengl/Font.cpp"
    $13 = (($12) + ($10))|0; //@line 627 "modules/graphics/opengl/Font.cpp"
    $14 = ($13|0)<($totalverts$03|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $15 = $14 ? $totalverts$03 : $13; //@line 627 "modules/graphics/opengl/Font.cpp"
    $16 = ((($__begin$sroa$0$02)) + 12|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $17 = ($16|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($17) {
     $$lcssa = $15;
     break;
    } else {
     $__begin$sroa$0$02 = $16;$totalverts$03 = $15;
    }
   }
   $phitmp = $$lcssa >>> 2; //@line 625 "modules/graphics/opengl/Font.cpp"
   $6 = ((($this)) + 128|0); //@line 629 "modules/graphics/opengl/Font.cpp"
   $7 = (__ZNK4love8graphics6opengl11QuadIndices7getSizeEv($6)|0); //@line 629 "modules/graphics/opengl/Font.cpp"
   $8 = ($phitmp>>>0)>($7>>>0); //@line 629 "modules/graphics/opengl/Font.cpp"
   if ($8) {
    __ZN4love8graphics6opengl11QuadIndicesC2Ej($0,$phitmp); //@line 630 "modules/graphics/opengl/Font.cpp"
    __THREW__ = 0;
    (invoke_iii(4036,($6|0),($0|0))|0); //@line 630 "modules/graphics/opengl/Font.cpp"
    $18 = __THREW__; __THREW__ = 0;
    $19 = $18&1;
    if ($19) {
     $20 = ___cxa_find_matching_catch_2()|0;
     $21 = tempRet0;
     __ZN4love8graphics6opengl11QuadIndicesD2Ev($0); //@line 630 "modules/graphics/opengl/Font.cpp"
     ___resumeException($20|0);
     // unreachable;
    } else {
     __ZN4love8graphics6opengl11QuadIndicesD2Ev($0); //@line 630 "modules/graphics/opengl/Font.cpp"
     $23 = $6;
     break;
    }
   } else {
    $23 = $6;
   }
  }
 } while(0);
 __ZN4love8graphics6opengl6OpenGL11prepareDrawEv(634044); //@line 632 "modules/graphics/opengl/Font.cpp"
 $22 = HEAP32[158426]|0; //@line 357 "modules/graphics/opengl/GLBuffer.h"
 $24 = (__ZNK4love8graphics6opengl11QuadIndices7getTypeEj($23,$22)|0); //@line 357 "modules/graphics/opengl/GLBuffer.h"
 $25 = (__ZN4love8graphics6opengl11QuadIndices14getElementSizeEv($23)|0); //@line 635 "modules/graphics/opengl/Font.cpp"
 if ($bufferedvertices) {
  $26 = (__ZNK4love8graphics6opengl11QuadIndices9getBufferEv($23)|0); //@line 642 "modules/graphics/opengl/Font.cpp"
  __ZN4love8graphics6opengl8GLBuffer4bindEv($26); //@line 642 "modules/graphics/opengl/Font.cpp"
 }
 $27 = HEAP32[$drawcommands>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $28 = HEAP32[$2>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $29 = ($27|0)==($28|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if (!($29)) {
  $30 = ($25*6)|0; //@line 649 "modules/graphics/opengl/Font.cpp"
  if ($bufferedvertices) {
   $__begin2$sroa$0$01$us = $27;
   while(1) {
    $31 = ((($__begin2$sroa$0$01$us)) + 8|0); //@line 648 "modules/graphics/opengl/Font.cpp"
    $32 = HEAP32[$31>>2]|0; //@line 648 "modules/graphics/opengl/Font.cpp"
    $33 = (($32|0) / 4)&-1; //@line 648 "modules/graphics/opengl/Font.cpp"
    $34 = ($33*6)|0; //@line 648 "modules/graphics/opengl/Font.cpp"
    $35 = ((($__begin2$sroa$0$01$us)) + 4|0); //@line 649 "modules/graphics/opengl/Font.cpp"
    $36 = HEAP32[$35>>2]|0; //@line 649 "modules/graphics/opengl/Font.cpp"
    $37 = (($36|0) / 4)&-1; //@line 649 "modules/graphics/opengl/Font.cpp"
    $38 = Math_imul($30, $37)|0; //@line 649 "modules/graphics/opengl/Font.cpp"
    $39 = HEAP32[$__begin2$sroa$0$01$us>>2]|0; //@line 652 "modules/graphics/opengl/Font.cpp"
    __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$39); //@line 652 "modules/graphics/opengl/Font.cpp"
    $40 = (__ZNK4love8graphics6opengl11QuadIndices10getPointerEj($23,$38)|0); //@line 655 "modules/graphics/opengl/Font.cpp"
    __ZN4love8graphics6opengl6OpenGL12drawElementsEjijPKv(634044,4,$34,$24,$40); //@line 655 "modules/graphics/opengl/Font.cpp"
    $41 = ((($__begin2$sroa$0$01$us)) + 12|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $42 = ($41|0)==($28|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($42) {
     break;
    } else {
     $__begin2$sroa$0$01$us = $41;
    }
   }
  } else {
   $__begin2$sroa$0$01 = $27;
   while(1) {
    $43 = ((($__begin2$sroa$0$01)) + 8|0); //@line 648 "modules/graphics/opengl/Font.cpp"
    $44 = HEAP32[$43>>2]|0; //@line 648 "modules/graphics/opengl/Font.cpp"
    $45 = (($44|0) / 4)&-1; //@line 648 "modules/graphics/opengl/Font.cpp"
    $46 = ($45*6)|0; //@line 648 "modules/graphics/opengl/Font.cpp"
    $47 = ((($__begin2$sroa$0$01)) + 4|0); //@line 649 "modules/graphics/opengl/Font.cpp"
    $48 = HEAP32[$47>>2]|0; //@line 649 "modules/graphics/opengl/Font.cpp"
    $49 = (($48|0) / 4)&-1; //@line 649 "modules/graphics/opengl/Font.cpp"
    $50 = Math_imul($30, $49)|0; //@line 649 "modules/graphics/opengl/Font.cpp"
    $51 = HEAP32[$__begin2$sroa$0$01>>2]|0; //@line 652 "modules/graphics/opengl/Font.cpp"
    __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$51); //@line 652 "modules/graphics/opengl/Font.cpp"
    $52 = (__ZNK4love8graphics6opengl11QuadIndices10getIndicesEj($23,$50)|0); //@line 657 "modules/graphics/opengl/Font.cpp"
    __ZN4love8graphics6opengl6OpenGL12drawElementsEjijPKv(634044,4,$46,$24,$52); //@line 657 "modules/graphics/opengl/Font.cpp"
    $53 = ((($__begin2$sroa$0$01)) + 12|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $54 = ($53|0)==($28|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($54) {
     break;
    } else {
     $__begin2$sroa$0$01 = $53;
    }
   }
  }
 }
 if (!($bufferedvertices)) {
  STACKTOP = sp;return; //@line 662 "modules/graphics/opengl/Font.cpp"
 }
 $55 = (__ZNK4love8graphics6opengl11QuadIndices9getBufferEv($23)|0); //@line 661 "modules/graphics/opengl/Font.cpp"
 __ZN4love8graphics6opengl8GLBuffer6unbindEv($55); //@line 661 "modules/graphics/opengl/Font.cpp"
 STACKTOP = sp;return; //@line 662 "modules/graphics/opengl/Font.cpp"
}
function __ZN4love8graphics6opengl4Font6printvERKNS_7Matrix4ERKNSt3__16vectorINS2_11DrawCommandENS6_9allocatorIS8_EEEERKNS7_INS2_11GlyphVertexENS9_ISE_EEEE($this,$t,$drawcommands,$vertices) {
 $this = $this|0;
 $t = $t|0;
 $drawcommands = $drawcommands|0;
 $vertices = $vertices|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$vertices>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($vertices)) + 4|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ($0|0)==($2|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($3) {
  return; //@line 681 "modules/graphics/opengl/Font.cpp"
 }
 $4 = HEAP32[$drawcommands>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ((($drawcommands)) + 4|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$5>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ($4|0)==($6|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($7) {
  return; //@line 681 "modules/graphics/opengl/Font.cpp"
 }
 __ZN4love8graphics6opengl6OpenGL13pushTransformEv(634044); //@line 125 "modules/graphics/opengl/OpenGL.h"
 __THREW__ = 0;
 $8 = (invoke_ii(4003,(634044|0))|0); //@line 135 "modules/graphics/opengl/OpenGL.h"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if (!($10)) {
  __THREW__ = 0;
  invoke_vii(4004,($8|0),($t|0)); //@line 672 "modules/graphics/opengl/Font.cpp"
  $11 = __THREW__; __THREW__ = 0;
  $12 = $11&1;
  if (!($12)) {
   $13 = HEAP32[$vertices>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $14 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   invoke_viiiiii($14|0,0,2,5126,0,16,($13|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if (!($16)) {
    $17 = HEAP32[$vertices>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $18 = ((($17)) + 8|0); //@line 675 "modules/graphics/opengl/Font.cpp"
    $19 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
    __THREW__ = 0;
    invoke_viiiiii($19|0,1,2,5123,1,16,($18|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
    $20 = __THREW__; __THREW__ = 0;
    $21 = $20&1;
    if (!($21)) {
     $22 = HEAP32[$vertices>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $23 = ((($22)) + 12|0); //@line 676 "modules/graphics/opengl/Font.cpp"
     $24 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_viiiiii($24|0,2,4,5121,1,16,($23|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
     $25 = __THREW__; __THREW__ = 0;
     $26 = $25&1;
     if (!($26)) {
      __THREW__ = 0;
      invoke_vii(4006,(634044|0),7); //@line 678 "modules/graphics/opengl/Font.cpp"
      $27 = __THREW__; __THREW__ = 0;
      $28 = $27&1;
      if (!($28)) {
       __THREW__ = 0;
       invoke_viii(4037,($this|0),($drawcommands|0),0); //@line 680 "modules/graphics/opengl/Font.cpp"
       $29 = __THREW__; __THREW__ = 0;
       $30 = $29&1;
       if (!($30)) {
        __THREW__ = 0;
        invoke_vi(4009,(634044|0)); //@line 130 "modules/graphics/opengl/OpenGL.h"
        $31 = __THREW__; __THREW__ = 0;
        $32 = $31&1;
        if ($32) {
         $33 = ___cxa_find_matching_catch_3(0|0)|0;
         $34 = tempRet0;
         ___clang_call_terminate($33); //@line 130 "modules/graphics/opengl/OpenGL.h"
         // unreachable; //@line 130 "modules/graphics/opengl/OpenGL.h"
        } else {
         return; //@line 681 "modules/graphics/opengl/Font.cpp"
        }
       }
      }
     }
    }
   }
  }
 }
 $35 = ___cxa_find_matching_catch_2()|0;
 $36 = tempRet0;
 __THREW__ = 0;
 invoke_vi(4009,(634044|0)); //@line 130 "modules/graphics/opengl/OpenGL.h"
 $37 = __THREW__; __THREW__ = 0;
 $38 = $37&1;
 if ($38) {
  $39 = ___cxa_find_matching_catch_3(0|0)|0;
  $40 = tempRet0;
  ___clang_call_terminate($39); //@line 130 "modules/graphics/opengl/OpenGL.h"
  // unreachable; //@line 130 "modules/graphics/opengl/OpenGL.h"
 } else {
  ___resumeException($35|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl4Font5printERKNSt3__16vectorINS2_13ColoredStringENS3_9allocatorIS5_EEEEfffffffff($this,$text,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $text = $text|0;
 $x = +$x;
 $y = +$y;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $__new_last1$i$i$i$i = 0, $__new_last1$i$i$i$i12 = 0, $codepoints = 0, $drawcommands = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i$i = 0, $scevgep$i$i$i$i13 = 0, $scevgep$i$i$i1$i = 0, $scevgep$i$i$i1$i18 = 0, $scevgep$i$i$i23 = 0, $scevgep$i$i$i4 = 0, $scevgep$i$i$i7 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i$i = 0, $scevgep4$i$i$i$i14 = 0;
 var $scevgep4$i$i$i2$i = 0, $scevgep4$i$i$i2$i19 = 0, $scevgep4$i$i$i24 = 0, $scevgep4$i$i$i5 = 0, $scevgep4$i$i$i8 = 0, $t = 0, $vertices = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 120|0;
 $codepoints = sp + 96|0;
 $vertices = sp + 84|0;
 $drawcommands = sp + 72|0;
 $0 = sp + 64|0;
 $t = sp;
 ;HEAP32[$codepoints>>2]=0|0;HEAP32[$codepoints+4>>2]=0|0;HEAP32[$codepoints+8>>2]=0|0;HEAP32[$codepoints+12>>2]=0|0;HEAP32[$codepoints+16>>2]=0|0;HEAP32[$codepoints+20>>2]=0|0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 __THREW__ = 0;
 invoke_vii(4038,($text|0),($codepoints|0)); //@line 686 "modules/graphics/opengl/Font.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $51 = ___cxa_find_matching_catch_2()|0;
  $52 = tempRet0;
  $$2 = $52;$$23 = $51;
 } else {
  HEAP32[$vertices>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $3 = ((($vertices)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$3>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = ((($vertices)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$4>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAPF32[$0>>2] = 0.0; //@line 219 "./common/Vector.h"
  $5 = ((($0)) + 4|0); //@line 220 "./common/Vector.h"
  HEAPF32[$5>>2] = 0.0; //@line 220 "./common/Vector.h"
  __THREW__ = 0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;HEAP32[$$byval_copy+4>>2]=HEAP32[$0+4>>2]|0; //@line 689 "modules/graphics/opengl/Font.cpp"
  invoke_viiiidii(4029,($drawcommands|0),($this|0),($codepoints|0),($vertices|0),0.0,($$byval_copy|0),(0|0)); //@line 689 "modules/graphics/opengl/Font.cpp"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if ($7) {
   $53 = ___cxa_find_matching_catch_2()|0;
   $54 = tempRet0;
   $$1 = $54;$$12 = $53;
  } else {
   __THREW__ = 0;
   invoke_viddddddddd(4039,($t|0),(+$x),(+$y),(+$angle),(+$sx),(+$sy),(+$ox),(+$oy),(+$kx),(+$ky)); //@line 691 "modules/graphics/opengl/Font.cpp"
   $8 = __THREW__; __THREW__ = 0;
   $9 = $8&1;
   do {
    if ($9) {
     $55 = ___cxa_find_matching_catch_2()|0;
     $56 = tempRet0;
     $$0 = $56;$$01 = $55;
    } else {
     __THREW__ = 0;
     invoke_viiii(4040,($this|0),($t|0),($drawcommands|0),($vertices|0)); //@line 693 "modules/graphics/opengl/Font.cpp"
     $10 = __THREW__; __THREW__ = 0;
     $11 = $10&1;
     if ($11) {
      $57 = ___cxa_find_matching_catch_2()|0;
      $58 = tempRet0;
      __ZN4love7Matrix4D2Ev($t); //@line 694 "modules/graphics/opengl/Font.cpp"
      $$0 = $58;$$01 = $57;
      break;
     }
     __ZN4love7Matrix4D2Ev($t); //@line 694 "modules/graphics/opengl/Font.cpp"
     $12 = HEAP32[$drawcommands>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $13 = ($12|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $14 = $12; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($13)) {
      $15 = ((($drawcommands)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $16 = HEAP32[$15>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $17 = ($16|0)==($12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($17)) {
       $scevgep$i$i$i4 = ((($16)) + -12|0);
       $18 = $scevgep$i$i$i4;
       $19 = (($18) - ($14))|0;
       $20 = (($19>>>0) / 12)&-1;
       $21 = $20 ^ -1;
       $scevgep4$i$i$i5 = (($16) + (($21*12)|0)|0);
       HEAP32[$15>>2] = $scevgep4$i$i$i5; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($12); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $22 = HEAP32[$vertices>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $23 = ($22|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $24 = $22; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($23)) {
      $25 = HEAP32[$3>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $26 = ($25|0)==($22|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($26)) {
       $scevgep$i$i$i7 = ((($25)) + -16|0);
       $27 = $scevgep$i$i$i7;
       $28 = (($27) - ($24))|0;
       $29 = $28 >>> 4;
       $30 = $29 ^ -1;
       $scevgep4$i$i$i8 = (($25) + ($30<<4)|0);
       HEAP32[$3>>2] = $scevgep4$i$i$i8; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($22); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $31 = ((($codepoints)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $32 = HEAP32[$31>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $33 = ($32|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($33)) {
      $34 = ((($codepoints)) + 16|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $35 = HEAP32[$34>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $36 = ($35|0)==($32|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($36)) {
       $__new_last1$i$i$i$i12 = $32; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $scevgep$i$i$i$i13 = ((($35)) + -8|0);
       $37 = $scevgep$i$i$i$i13;
       $38 = (($37) - ($__new_last1$i$i$i$i12))|0;
       $39 = $38 >>> 3;
       $40 = $39 ^ -1;
       $scevgep4$i$i$i$i14 = (($35) + ($40<<3)|0);
       HEAP32[$34>>2] = $scevgep4$i$i$i$i14; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($32); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $41 = HEAP32[$codepoints>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $42 = ($41|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $43 = $41; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($42) {
      STACKTOP = sp;return; //@line 694 "modules/graphics/opengl/Font.cpp"
     }
     $44 = ((($codepoints)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $45 = HEAP32[$44>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $46 = ($45|0)==($41|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($46)) {
      $scevgep$i$i$i1$i18 = ((($45)) + -4|0);
      $47 = $scevgep$i$i$i1$i18;
      $48 = (($47) - ($43))|0;
      $49 = $48 >>> 2;
      $50 = $49 ^ -1;
      $scevgep4$i$i$i2$i19 = (($45) + ($50<<2)|0);
      HEAP32[$44>>2] = $scevgep4$i$i$i2$i19; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __ZdlPv($41); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     STACKTOP = sp;return; //@line 694 "modules/graphics/opengl/Font.cpp"
    }
   } while(0);
   $59 = HEAP32[$drawcommands>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $60 = ($59|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $61 = $59; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($60) {
    $$1 = $$0;$$12 = $$01;
   } else {
    $62 = ((($drawcommands)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $63 = HEAP32[$62>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $64 = ($63|0)==($59|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($64)) {
     $scevgep$i$i$i23 = ((($63)) + -12|0);
     $65 = $scevgep$i$i$i23;
     $66 = (($65) - ($61))|0;
     $67 = (($66>>>0) / 12)&-1;
     $68 = $67 ^ -1;
     $scevgep4$i$i$i24 = (($63) + (($68*12)|0)|0);
     HEAP32[$62>>2] = $scevgep4$i$i$i24; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($59); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $$1 = $$0;$$12 = $$01;
   }
  }
  $69 = HEAP32[$vertices>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $70 = ($69|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $71 = $69; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($70) {
   $$2 = $$1;$$23 = $$12;
  } else {
   $72 = HEAP32[$3>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $73 = ($72|0)==($69|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($73)) {
    $scevgep$i$i$i = ((($72)) + -16|0);
    $74 = $scevgep$i$i$i;
    $75 = (($74) - ($71))|0;
    $76 = $75 >>> 4;
    $77 = $76 ^ -1;
    $scevgep4$i$i$i = (($72) + ($77<<4)|0);
    HEAP32[$3>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($69); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $$2 = $$1;$$23 = $$12;
  }
 }
 $78 = ((($codepoints)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $79 = HEAP32[$78>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $80 = ($79|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($80)) {
  $81 = ((($codepoints)) + 16|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $82 = HEAP32[$81>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $83 = ($82|0)==($79|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($83)) {
   $__new_last1$i$i$i$i = $79; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $scevgep$i$i$i$i = ((($82)) + -8|0);
   $84 = $scevgep$i$i$i$i;
   $85 = (($84) - ($__new_last1$i$i$i$i))|0;
   $86 = $85 >>> 3;
   $87 = $86 ^ -1;
   $scevgep4$i$i$i$i = (($82) + ($87<<3)|0);
   HEAP32[$81>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($79); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $88 = HEAP32[$codepoints>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $89 = ($88|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $90 = $88; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($89) {
  ___resumeException($$23|0);
  // unreachable;
 }
 $91 = ((($codepoints)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $92 = HEAP32[$91>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $93 = ($92|0)==($88|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($93)) {
  $scevgep$i$i$i1$i = ((($92)) + -4|0);
  $94 = $scevgep$i$i$i1$i;
  $95 = (($94) - ($90))|0;
  $96 = $95 >>> 2;
  $97 = $96 ^ -1;
  $scevgep4$i$i$i2$i = (($92) + ($97<<2)|0);
  HEAP32[$91>>2] = $scevgep4$i$i$i2$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($88); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$23|0);
 // unreachable;
}
function __ZN4love8graphics6opengl4Font6printfERKNSt3__16vectorINS2_13ColoredStringENS3_9allocatorIS5_EEEEfffNS2_9AlignModeEfffffff($this,$text,$x,$y,$wrap,$align,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $text = $text|0;
 $x = +$x;
 $y = +$y;
 $wrap = +$wrap;
 $align = $align|0;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $__new_last1$i$i$i$i = 0, $__new_last1$i$i$i$i7 = 0, $codepoints = 0, $drawcommands = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i$i = 0, $scevgep$i$i$i$i8 = 0, $scevgep$i$i$i1$i = 0, $scevgep$i$i$i1$i13 = 0, $scevgep$i$i$i18 = 0, $scevgep$i$i$i23 = 0, $scevgep$i$i$i4 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i$i = 0, $scevgep4$i$i$i$i9 = 0, $scevgep4$i$i$i19 = 0, $scevgep4$i$i$i2$i = 0, $scevgep4$i$i$i2$i14 = 0;
 var $scevgep4$i$i$i24 = 0, $scevgep4$i$i$i5 = 0, $t = 0, $vertices = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $codepoints = sp + 88|0;
 $vertices = sp + 76|0;
 $drawcommands = sp + 64|0;
 $t = sp;
 ;HEAP32[$codepoints>>2]=0|0;HEAP32[$codepoints+4>>2]=0|0;HEAP32[$codepoints+8>>2]=0|0;HEAP32[$codepoints+12>>2]=0|0;HEAP32[$codepoints+16>>2]=0|0;HEAP32[$codepoints+20>>2]=0|0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 __THREW__ = 0;
 invoke_vii(4038,($text|0),($codepoints|0)); //@line 699 "modules/graphics/opengl/Font.cpp"
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 if ($1) {
  $49 = ___cxa_find_matching_catch_2()|0;
  $50 = tempRet0;
  $$2 = $50;$$23 = $49;
 } else {
  HEAP32[$vertices>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $2 = ((($vertices)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$2>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $3 = ((($vertices)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$3>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  __THREW__ = 0;
  invoke_viiidiii(4031,($drawcommands|0),($this|0),($codepoints|0),(+$wrap),($align|0),($vertices|0),(0|0)); //@line 702 "modules/graphics/opengl/Font.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $51 = ___cxa_find_matching_catch_2()|0;
   $52 = tempRet0;
   $$1 = $52;$$12 = $51;
  } else {
   __THREW__ = 0;
   invoke_viddddddddd(4039,($t|0),(+$x),(+$y),(+$angle),(+$sx),(+$sy),(+$ox),(+$oy),(+$kx),(+$ky)); //@line 704 "modules/graphics/opengl/Font.cpp"
   $6 = __THREW__; __THREW__ = 0;
   $7 = $6&1;
   do {
    if ($7) {
     $53 = ___cxa_find_matching_catch_2()|0;
     $54 = tempRet0;
     $$0 = $54;$$01 = $53;
    } else {
     __THREW__ = 0;
     invoke_viiii(4040,($this|0),($t|0),($drawcommands|0),($vertices|0)); //@line 706 "modules/graphics/opengl/Font.cpp"
     $8 = __THREW__; __THREW__ = 0;
     $9 = $8&1;
     if ($9) {
      $55 = ___cxa_find_matching_catch_2()|0;
      $56 = tempRet0;
      __ZN4love7Matrix4D2Ev($t); //@line 707 "modules/graphics/opengl/Font.cpp"
      $$0 = $56;$$01 = $55;
      break;
     }
     __ZN4love7Matrix4D2Ev($t); //@line 707 "modules/graphics/opengl/Font.cpp"
     $10 = HEAP32[$drawcommands>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $11 = ($10|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $12 = $10; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($11)) {
      $13 = ((($drawcommands)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $14 = HEAP32[$13>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $15 = ($14|0)==($10|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($15)) {
       $scevgep$i$i$i = ((($14)) + -12|0);
       $16 = $scevgep$i$i$i;
       $17 = (($16) - ($12))|0;
       $18 = (($17>>>0) / 12)&-1;
       $19 = $18 ^ -1;
       $scevgep4$i$i$i = (($14) + (($19*12)|0)|0);
       HEAP32[$13>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($10); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $20 = HEAP32[$vertices>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $21 = ($20|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $22 = $20; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($21)) {
      $23 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $24 = ($23|0)==($20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($24)) {
       $scevgep$i$i$i4 = ((($23)) + -16|0);
       $25 = $scevgep$i$i$i4;
       $26 = (($25) - ($22))|0;
       $27 = $26 >>> 4;
       $28 = $27 ^ -1;
       $scevgep4$i$i$i5 = (($23) + ($28<<4)|0);
       HEAP32[$2>>2] = $scevgep4$i$i$i5; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($20); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $29 = ((($codepoints)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $30 = HEAP32[$29>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $31 = ($30|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($31)) {
      $32 = ((($codepoints)) + 16|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $33 = HEAP32[$32>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $34 = ($33|0)==($30|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($34)) {
       $__new_last1$i$i$i$i7 = $30; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $scevgep$i$i$i$i8 = ((($33)) + -8|0);
       $35 = $scevgep$i$i$i$i8;
       $36 = (($35) - ($__new_last1$i$i$i$i7))|0;
       $37 = $36 >>> 3;
       $38 = $37 ^ -1;
       $scevgep4$i$i$i$i9 = (($33) + ($38<<3)|0);
       HEAP32[$32>>2] = $scevgep4$i$i$i$i9; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($30); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $39 = HEAP32[$codepoints>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $40 = ($39|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $41 = $39; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($40) {
      STACKTOP = sp;return; //@line 707 "modules/graphics/opengl/Font.cpp"
     }
     $42 = ((($codepoints)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $43 = HEAP32[$42>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $44 = ($43|0)==($39|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($44)) {
      $scevgep$i$i$i1$i13 = ((($43)) + -4|0);
      $45 = $scevgep$i$i$i1$i13;
      $46 = (($45) - ($41))|0;
      $47 = $46 >>> 2;
      $48 = $47 ^ -1;
      $scevgep4$i$i$i2$i14 = (($43) + ($48<<2)|0);
      HEAP32[$42>>2] = $scevgep4$i$i$i2$i14; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __ZdlPv($39); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     STACKTOP = sp;return; //@line 707 "modules/graphics/opengl/Font.cpp"
    }
   } while(0);
   $57 = HEAP32[$drawcommands>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $58 = ($57|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $59 = $57; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($58) {
    $$1 = $$0;$$12 = $$01;
   } else {
    $60 = ((($drawcommands)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $61 = HEAP32[$60>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $62 = ($61|0)==($57|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($62)) {
     $scevgep$i$i$i18 = ((($61)) + -12|0);
     $63 = $scevgep$i$i$i18;
     $64 = (($63) - ($59))|0;
     $65 = (($64>>>0) / 12)&-1;
     $66 = $65 ^ -1;
     $scevgep4$i$i$i19 = (($61) + (($66*12)|0)|0);
     HEAP32[$60>>2] = $scevgep4$i$i$i19; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($57); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $$1 = $$0;$$12 = $$01;
   }
  }
  $67 = HEAP32[$vertices>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $68 = ($67|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $69 = $67; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($68) {
   $$2 = $$1;$$23 = $$12;
  } else {
   $70 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $71 = ($70|0)==($67|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($71)) {
    $scevgep$i$i$i23 = ((($70)) + -16|0);
    $72 = $scevgep$i$i$i23;
    $73 = (($72) - ($69))|0;
    $74 = $73 >>> 4;
    $75 = $74 ^ -1;
    $scevgep4$i$i$i24 = (($70) + ($75<<4)|0);
    HEAP32[$2>>2] = $scevgep4$i$i$i24; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($67); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $$2 = $$1;$$23 = $$12;
  }
 }
 $76 = ((($codepoints)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $77 = HEAP32[$76>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $78 = ($77|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($78)) {
  $79 = ((($codepoints)) + 16|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $80 = HEAP32[$79>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $81 = ($80|0)==($77|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($81)) {
   $__new_last1$i$i$i$i = $77; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $scevgep$i$i$i$i = ((($80)) + -8|0);
   $82 = $scevgep$i$i$i$i;
   $83 = (($82) - ($__new_last1$i$i$i$i))|0;
   $84 = $83 >>> 3;
   $85 = $84 ^ -1;
   $scevgep4$i$i$i$i = (($80) + ($85<<3)|0);
   HEAP32[$79>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($77); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $86 = HEAP32[$codepoints>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $87 = ($86|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $88 = $86; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($87) {
  ___resumeException($$23|0);
  // unreachable;
 }
 $89 = ((($codepoints)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $90 = HEAP32[$89>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $91 = ($90|0)==($86|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($91)) {
  $scevgep$i$i$i1$i = ((($90)) + -4|0);
  $92 = $scevgep$i$i$i1$i;
  $93 = (($92) - ($88))|0;
  $94 = $93 >>> 2;
  $95 = $94 ^ -1;
  $scevgep4$i$i$i2$i = (($90) + ($95<<2)|0);
  HEAP32[$89>>2] = $scevgep4$i$i$i2$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($86); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$23|0);
 // unreachable;
}
function __ZN4love8graphics6opengl4Font8getWidthERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($this,$str) {
 $this = $this|0;
 $str = $str|0;
 var $$0 = 0, $$0$i = 0, $$0$i17 = 0, $$01 = 0, $$01$i = 0, $$02 = 0, $$1 = 0, $$13 = 0, $$byval_copy = 0, $$byval_copy15 = 0, $$pn$i$i$i = 0, $$pn$us$i$i$i = 0, $$pre = 0, $$sroa$0$0$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0.0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__nd$0$i$i$i = 0, $__nd$0$us$i$i$i = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $eh$lpad$body$index9Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i = 0, $iss = 0;
 var $line = 0, $max_width$0 = 0, $max_width$0$lcssa95 = 0, $prevglyph$04579 = 0, $temp$i = 0, $vararg_buffer = 0, $width$04480 = 0, $width$04480$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy15 = sp + 184|0;
 $$byval_copy = sp + 180|0;
 $vararg_buffer = sp;
 $temp$i = sp + 176|0;
 $0 = sp + 172|0;
 $1 = sp + 168|0;
 $iss = sp + 28|0;
 $line = sp + 16|0;
 $i = sp + 4|0;
 $2 = HEAP8[$str>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = $2 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ($3<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($str)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = HEAP32[$5>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = $2&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $7 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $4 ? $8 : $6; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = ($9|0)==(0); //@line 711 "modules/graphics/opengl/Font.cpp"
 if ($10) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 745 "modules/graphics/opengl/Font.cpp"
 }
 $11 = ((($iss)) + 60|0); //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $12 = ((($iss)) + 8|0); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$iss>>2] = ((24696)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 HEAP32[$11>>2] = ((24716)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $13 = ((($iss)) + 4|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 HEAP32[$13>>2] = 0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $14 = ((($iss)) + 60|0); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 __THREW__ = 0;
 invoke_vii(3362,($14|0),($12|0)); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if ($16) {
  $27 = ___cxa_find_matching_catch_2()|0;
  $28 = tempRet0;
  $$0$i = $28;$$01$i = $27;
  __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($11); //@line 671 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  ___resumeException($$01$i|0);
  // unreachable;
 }
 $17 = ((($iss)) + 132|0); //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$17>>2] = 0; //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 $18 = ((($iss)) + 136|0); //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$18>>2] = -1; //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$iss>>2] = (24656); //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$11>>2] = (24676); //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 __THREW__ = 0;
 invoke_vi(3363,($12|0)); //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 do {
  if ($20) {
   $29 = ___cxa_find_matching_catch_2()|0;
   $30 = tempRet0;
   $eh$lpad$body$i$index2Z2D = $30;$eh$lpad$body$i$indexZ2D = $29;
  } else {
   HEAP32[$12>>2] = (9356); //@line 253 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $21 = ((($iss)) + 40|0); //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $22 = ((($iss)) + 56|0); //@line 252 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   ;HEAP32[$21>>2]=0|0;HEAP32[$21+4>>2]=0|0;HEAP32[$21+8>>2]=0|0;HEAP32[$21+12>>2]=0|0; //@line 251 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   HEAP32[$22>>2] = 8; //@line 252 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   __THREW__ = 0;
   invoke_vii(3364,($12|0),($str|0)); //@line 254 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $23 = __THREW__; __THREW__ = 0;
   $24 = $23&1;
   if ($24) {
    $25 = ___cxa_find_matching_catch_2()|0;
    $26 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($21); //@line 255 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($12); //@line 255 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $eh$lpad$body$i$index2Z2D = $26;$eh$lpad$body$i$indexZ2D = $25;
    break;
   }
   ;HEAP32[$line>>2]=0|0;HEAP32[$line+4>>2]=0|0;HEAP32[$line+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $31 = ((($line)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $32 = ((($line)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $33 = ((($line)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $34 = ((($i)) + 4|0); //@line 277 "./libraries/utf8/utf8/checked.h"
   $35 = ((($i)) + 8|0); //@line 277 "./libraries/utf8/utf8/checked.h"
   $36 = ((($i)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $37 = ((($i)) + 8|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $38 = ((($this)) + 56|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $39 = ((($this)) + 52|0); //@line 2015 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $max_width$0 = 0;
   L14: while(1) {
    __THREW__ = 0;
    $40 = (invoke_iiii(3516,($iss|0),($line|0),10)|0); //@line 717 "modules/graphics/opengl/Font.cpp"
    $41 = __THREW__; __THREW__ = 0;
    $42 = $41&1;
    if ($42) {
     label = 40;
     break;
    }
    $43 = HEAP32[$40>>2]|0; //@line 717 "modules/graphics/opengl/Font.cpp"
    $44 = ((($43)) + -12|0); //@line 717 "modules/graphics/opengl/Font.cpp"
    $45 = HEAP32[$44>>2]|0; //@line 717 "modules/graphics/opengl/Font.cpp"
    $46 = (($40) + ($45)|0); //@line 717 "modules/graphics/opengl/Font.cpp"
    $47 = ((($46)) + 16|0); //@line 545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    $48 = HEAP32[$47>>2]|0; //@line 545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    $49 = $48 & 5; //@line 545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    $50 = ($49|0)==(0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    if (!($50)) {
     $max_width$0$lcssa95 = $max_width$0;
     label = 49;
     break;
    }
    $51 = HEAP8[$line>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $52 = $51 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $53 = ($52<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $54 = HEAP32[$31>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $55 = $53 ? $32 : $54; //@line 1755 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $56 = $55; //@line 723 "modules/graphics/opengl/Font.cpp"
    $57 = HEAP32[$33>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $58 = $51&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $59 = $58 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $60 = $53 ? $59 : $57; //@line 1755 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $61 = (($55) + ($60)|0); //@line 1399 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $62 = $61; //@line 723 "modules/graphics/opengl/Font.cpp"
    HEAP32[$i>>2] = $56; //@line 277 "./libraries/utf8/utf8/checked.h"
    HEAP32[$34>>2] = $56; //@line 277 "./libraries/utf8/utf8/checked.h"
    HEAP32[$35>>2] = $62; //@line 277 "./libraries/utf8/utf8/checked.h"
    $63 = ($60|0)<(0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($63) {
     label = 14;
     break;
    } else {
     $79 = $62;$prevglyph$04579 = 0;$width$04480 = 0;
    }
    while(1) {
     $76 = HEAP32[$i>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $77 = ($76|0)==($61|0); //@line 297 "./libraries/utf8/utf8/checked.h"
     $78 = $76; //@line 306 "./libraries/utf8/utf8/checked.h"
     if ($77) {
      $width$04480$lcssa = $width$04480;
      break;
     }
     HEAP32[$1>>2] = $79; //@line 307 "./libraries/utf8/utf8/checked.h"
     __THREW__ = 0;
     ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0; //@line 307 "./libraries/utf8/utf8/checked.h"
     (invoke_iii(3997,($i|0),($$byval_copy|0))|0); //@line 307 "./libraries/utf8/utf8/checked.h"
     $80 = __THREW__; __THREW__ = 0;
     $81 = $80&1;
     if ($81) {
      label = 41;
      break L14;
     }
     HEAP32[$temp$i>>2] = $78; //@line 286 "./libraries/utf8/utf8/checked.h"
     HEAP32[$0>>2] = $79; //@line 287 "./libraries/utf8/utf8/checked.h"
     __THREW__ = 0;
     ;HEAP32[$$byval_copy15>>2]=HEAP32[$0>>2]|0; //@line 287 "./libraries/utf8/utf8/checked.h"
     $82 = (invoke_iii(3997,($temp$i|0),($$byval_copy15|0))|0); //@line 287 "./libraries/utf8/utf8/checked.h"
     $83 = __THREW__; __THREW__ = 0;
     $84 = $83&1;
     if ($84) {
      label = 41;
      break L14;
     }
     $85 = HEAP32[$38>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $86 = ($85|0)==(0); //@line 2018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     L23: do {
      if ($86) {
       label = 36;
      } else {
       $87 = (($85) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $88 = $87 & $85; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $89 = ($88|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($89) {
        $90 = $87 & $82; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $94 = $90;
       } else {
        $91 = (($82>>>0) % ($85>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $94 = $91;
       }
       $92 = HEAP32[$39>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $93 = (($92) + ($94<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $95 = HEAP32[$93>>2]|0; //@line 2021 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $96 = ($95|0)==(0|0); //@line 2022 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($96) {
        label = 36;
       } else {
        if ($89) {
         $$pn$us$i$i$i = $95;
         while(1) {
          $__nd$0$us$i$i$i = HEAP32[$$pn$us$i$i$i>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          $97 = ($__nd$0$us$i$i$i|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          if ($97) {
           label = 36;
           break L23;
          }
          $98 = ((($__nd$0$us$i$i$i)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          $99 = HEAP32[$98>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          $100 = $99 & $87; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          $101 = ($100|0)==($94|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          if (!($101)) {
           label = 36;
           break L23;
          }
          $102 = ((($__nd$0$us$i$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
          $103 = HEAP32[$102>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
          $104 = ($103|0)==($82|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
          if ($104) {
           $$sroa$0$0$i$i$i = $__nd$0$us$i$i$i;
           break;
          } else {
           $$pn$us$i$i$i = $__nd$0$us$i$i$i;
          }
         }
        } else {
         $$pn$i$i$i = $95;
         while(1) {
          $__nd$0$i$i$i = HEAP32[$$pn$i$i$i>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          $105 = ($__nd$0$i$i$i|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          if ($105) {
           label = 36;
           break L23;
          }
          $106 = ((($__nd$0$i$i$i)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          $107 = HEAP32[$106>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          $108 = (($107>>>0) % ($85>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          $109 = ($108|0)==($94|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
          if (!($109)) {
           label = 36;
           break L23;
          }
          $110 = ((($__nd$0$i$i$i)) + 8|0); //@line 437 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
          $111 = HEAP32[$110>>2]|0; //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
          $112 = ($111|0)==($82|0); //@line 647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/functional"
          if ($112) {
           $$sroa$0$0$i$i$i = $__nd$0$i$i$i;
           break;
          } else {
           $$pn$i$i$i = $__nd$0$i$i$i;
          }
         }
        }
        $113 = ((($$sroa$0$0$i$i$i)) + 12|0); //@line 323 "modules/graphics/opengl/Font.cpp"
        $$0$i17 = $113;
       }
      }
     } while(0);
     if ((label|0) == 36) {
      label = 0;
      __THREW__ = 0;
      $114 = (invoke_iii(4017,($this|0),($82|0))|0); //@line 325 "modules/graphics/opengl/Font.cpp"
      $115 = __THREW__; __THREW__ = 0;
      $116 = $115&1;
      if ($116) {
       label = 41;
       break L14;
      } else {
       $$0$i17 = $114;
      }
     }
     $117 = ((($$0$i17)) + 4|0); //@line 731 "modules/graphics/opengl/Font.cpp"
     $118 = HEAP32[$117>>2]|0; //@line 731 "modules/graphics/opengl/Font.cpp"
     __THREW__ = 0;
     $119 = (+invoke_diii(4025,($this|0),($prevglyph$04579|0),($82|0))); //@line 731 "modules/graphics/opengl/Font.cpp"
     $120 = __THREW__; __THREW__ = 0;
     $121 = $120&1;
     if ($121) {
      label = 41;
      break L14;
     }
     $122 = HEAP32[$36>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $123 = ($122|0)==($55|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if (!($123)) {
      label = 17;
      break L14;
     }
     $124 = (+($width$04480|0)); //@line 731 "modules/graphics/opengl/Font.cpp"
     $125 = (+($118|0)); //@line 731 "modules/graphics/opengl/Font.cpp"
     $126 = $125 + $119; //@line 731 "modules/graphics/opengl/Font.cpp"
     $127 = $124 + $126; //@line 731 "modules/graphics/opengl/Font.cpp"
     $128 = (~~(($127))); //@line 731 "modules/graphics/opengl/Font.cpp"
     $$pre = HEAP32[$37>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $129 = $$pre; //@line 291 "./libraries/utf8/utf8/checked.h"
     $130 = ($$pre|0)==($61|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($130) {
      $79 = $129;$prevglyph$04579 = $82;$width$04480 = $128;
     } else {
      label = 17;
      break L14;
     }
    }
    $154 = ($max_width$0|0)<($width$04480$lcssa|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $155 = $154 ? $width$04480$lcssa : $max_width$0; //@line 741 "modules/graphics/opengl/Font.cpp"
    $max_width$0 = $155;
   }
   do {
    if ((label|0) == 14) {
     $64 = (___cxa_allocate_exception(8)|0); //@line 280 "./libraries/utf8/utf8/checked.h"
     __THREW__ = 0;
     invoke_vii(3996,($64|0),(434517|0)); //@line 133 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/stdexcept"
     $65 = __THREW__; __THREW__ = 0;
     $66 = $65&1;
     if ($66) {
      $68 = ___cxa_find_matching_catch_3(1744|0)|0;
      $69 = tempRet0;
      ___cxa_free_exception(($64|0)); //@line 280 "./libraries/utf8/utf8/checked.h"
      $eh$lpad$body$index9Z2D = $69;$eh$lpad$body$indexZ2D = $68;
      break;
     } else {
      HEAP32[$64>>2] = (157912); //@line 133 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/stdexcept"
      __THREW__ = 0;
      invoke_viii(3350,($64|0),(9120|0),(3328|0)); //@line 280 "./libraries/utf8/utf8/checked.h"
      $67 = __THREW__; __THREW__ = 0;
      label = 42;
      break;
     }
    }
    else if ((label|0) == 17) {
     $70 = (___cxa_allocate_exception(8)|0); //@line 292 "./libraries/utf8/utf8/checked.h"
     __THREW__ = 0;
     invoke_vii(3996,($70|0),(434549|0)); //@line 292 "./libraries/utf8/utf8/checked.h"
     $71 = __THREW__; __THREW__ = 0;
     $72 = $71&1;
     if ($72) {
      $74 = ___cxa_find_matching_catch_3(1744|0)|0;
      $75 = tempRet0;
      ___cxa_free_exception(($70|0)); //@line 292 "./libraries/utf8/utf8/checked.h"
      $eh$lpad$body$index9Z2D = $75;$eh$lpad$body$indexZ2D = $74;
      break;
     } else {
      __THREW__ = 0;
      invoke_viii(3350,($70|0),(9072|0),(3328|0)); //@line 292 "./libraries/utf8/utf8/checked.h"
      $73 = __THREW__; __THREW__ = 0;
      label = 42;
      break;
     }
    }
    else if ((label|0) == 40) {
     $131 = ___cxa_find_matching_catch_2()|0;
     $132 = tempRet0;
     $$1 = $131;$$13 = $132;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($line); //@line 745 "modules/graphics/opengl/Font.cpp"
     HEAP32[$iss>>2] = ((24656)); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     HEAP32[$11>>2] = ((24676)); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     HEAP32[$12>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($21); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($12); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($11); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     ___resumeException($$1|0);
     // unreachable;
    }
    else if ((label|0) == 41) {
     $133 = ___cxa_find_matching_catch_3(1744|0)|0;
     $134 = tempRet0;
     $eh$lpad$body$index9Z2D = $134;$eh$lpad$body$indexZ2D = $133;
    }
    else if ((label|0) == 49) {
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($line); //@line 745 "modules/graphics/opengl/Font.cpp"
     HEAP32[$iss>>2] = ((24656)); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     HEAP32[$11>>2] = ((24676)); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     HEAP32[$12>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($21); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($12); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($11); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
     $$0 = $max_width$0$lcssa95;
     STACKTOP = sp;return ($$0|0); //@line 745 "modules/graphics/opengl/Font.cpp"
    }
   } while(0);
   if ((label|0) == 42) {
    $135 = ___cxa_find_matching_catch_3(1744|0)|0;
    $136 = tempRet0;
    $eh$lpad$body$index9Z2D = $136;$eh$lpad$body$indexZ2D = $135;
   }
   $137 = (_llvm_eh_typeid_for((1744|0))|0); //@line 735 "modules/graphics/opengl/Font.cpp"
   $138 = ($eh$lpad$body$index9Z2D|0)==($137|0); //@line 735 "modules/graphics/opengl/Font.cpp"
   if (!($138)) {
    $$1 = $eh$lpad$body$indexZ2D;$$13 = $eh$lpad$body$index9Z2D;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($line); //@line 745 "modules/graphics/opengl/Font.cpp"
    HEAP32[$iss>>2] = ((24656)); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    HEAP32[$11>>2] = ((24676)); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    HEAP32[$12>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($21); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($12); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($11); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    ___resumeException($$1|0);
    // unreachable;
   }
   $139 = (___cxa_begin_catch(($eh$lpad$body$indexZ2D|0))|0); //@line 735 "modules/graphics/opengl/Font.cpp"
   $140 = (___cxa_allocate_exception(16)|0); //@line 738 "modules/graphics/opengl/Font.cpp"
   $141 = HEAP32[$139>>2]|0; //@line 738 "modules/graphics/opengl/Font.cpp"
   $142 = ((($141)) + 8|0); //@line 738 "modules/graphics/opengl/Font.cpp"
   $143 = HEAP32[$142>>2]|0; //@line 738 "modules/graphics/opengl/Font.cpp"
   $144 = (FUNCTION_TABLE_ii[$143 & 8191]($139)|0); //@line 738 "modules/graphics/opengl/Font.cpp"
   __THREW__ = 0;
   HEAP32[$vararg_buffer>>2] = $144; //@line 738 "modules/graphics/opengl/Font.cpp"
   invoke_viii(3349,($140|0),(434605|0),($vararg_buffer|0)); //@line 738 "modules/graphics/opengl/Font.cpp"
   $145 = __THREW__; __THREW__ = 0;
   $146 = $145&1;
   if ($146) {
    $150 = ___cxa_find_matching_catch_2()|0;
    $151 = tempRet0;
    ___cxa_free_exception(($140|0)); //@line 738 "modules/graphics/opengl/Font.cpp"
    $$01 = $150;$$02 = $151;
   } else {
    __THREW__ = 0;
    invoke_viii(3350,($140|0),(96|0),(344|0)); //@line 738 "modules/graphics/opengl/Font.cpp"
    $147 = __THREW__; __THREW__ = 0;
    $148 = ___cxa_find_matching_catch_2()|0;
    $149 = tempRet0;
    $$01 = $148;$$02 = $149;
   }
   __THREW__ = 0;
   invoke_v(3378); //@line 739 "modules/graphics/opengl/Font.cpp"
   $152 = __THREW__; __THREW__ = 0;
   $153 = $152&1;
   if ($153) {
    $156 = ___cxa_find_matching_catch_3(0|0)|0;
    $157 = tempRet0;
    ___clang_call_terminate($156); //@line 739 "modules/graphics/opengl/Font.cpp"
    // unreachable; //@line 739 "modules/graphics/opengl/Font.cpp"
   } else {
    $$1 = $$01;$$13 = $$02;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($line); //@line 745 "modules/graphics/opengl/Font.cpp"
    HEAP32[$iss>>2] = ((24656)); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    HEAP32[$11>>2] = ((24676)); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    HEAP32[$12>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($21); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($12); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($11); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
    ___resumeException($$1|0);
    // unreachable;
   }
  }
 } while(0);
 $$0$i = $eh$lpad$body$i$index2Z2D;$$01$i = $eh$lpad$body$i$indexZ2D;
 __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($11); //@line 671 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 ___resumeException($$01$i|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__119basic_istringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED1Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = ((24656)); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $0 = ((($this)) + 60|0); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$0>>2] = ((24676)); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $1 = ((($this)) + 8|0); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$1>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $2 = ((($this)) + 40|0); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($1); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($0); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 return; //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
}
function __ZNSt3__119basic_istringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = ((24656)); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $0 = ((($this)) + 60|0); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$0>>2] = ((24676)); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $1 = ((($this)) + 8|0); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$1>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $2 = ((($this)) + 40|0); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($1); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($0); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZdlPv($this); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 return; //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
}
function __ZTv0_n12_NSt3__119basic_istringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED1Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $1 = ((($0)) + -12|0); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $2 = HEAP32[$1>>2]|0; //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $3 = (($this) + ($2)|0); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$3>>2] = ((24656)); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $4 = ((($3)) + 60|0); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$4>>2] = ((24676)); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $5 = ((($3)) + 8|0); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$5>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $6 = ((($3)) + 40|0); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($5); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($4); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 return; //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
}
function __ZTv0_n12_NSt3__119basic_istringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $1 = ((($0)) + -12|0); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $2 = HEAP32[$1>>2]|0; //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $3 = (($this) + ($2)|0); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$3>>2] = ((24656)); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $4 = ((($3)) + 60|0); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$4>>2] = ((24676)); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $5 = ((($3)) + 8|0); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 HEAP32[$5>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 $6 = ((($3)) + 40|0); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($5); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($4); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 __ZdlPv($3); //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
 return; //@line 120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
}
function __ZN4love8graphics6opengl4Font7getWrapERKNSt3__16vectorINS2_13ColoredStringENS3_9allocatorIS5_EEEEfRNS4_INS3_12basic_stringIcNS3_11char_traitsIcEENS6_IcEEEENS6_ISF_EEEEPNS4_IiNS6_IiEEEE($this,$text,$wraplimit,$lines,$linewidths) {
 $this = $this|0;
 $text = $text|0;
 $wraplimit = +$wraplimit;
 $lines = $lines|0;
 $linewidths = $linewidths|0;
 var $$0 = 0, $$0$i = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$pre = 0, $$pre$i = 0, $$pre$i17 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$033 = 0;
 var $__begin2$sroa$0$032 = 0, $__new_last1$i$i$i$i = 0, $__new_last1$i$i$i$i$i$i$i$i$i$i = 0, $__new_last1$i$i$i$i$i$i$i$i$i$i5 = 0, $__new_last1$i$i$i$i21 = 0, $character = 0, $codepointlines = 0, $cps = 0, $line = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $scevgep$i$i$i$i = 0, $scevgep$i$i$i$i$i$i$i$i$i$i = 0, $scevgep$i$i$i$i$i$i$i$i$i$i6 = 0, $scevgep$i$i$i$i22 = 0, $scevgep$i$i$i1$i = 0, $scevgep$i$i$i1$i$i$i$i$i$i$i = 0, $scevgep$i$i$i1$i$i$i$i$i$i$i12 = 0, $scevgep$i$i$i1$i27 = 0, $scevgep4$i$i$i$i = 0;
 var $scevgep4$i$i$i$i$i$i$i$i$i$i = 0, $scevgep4$i$i$i$i$i$i$i$i$i$i7 = 0, $scevgep4$i$i$i$i23 = 0, $scevgep4$i$i$i2$i = 0, $scevgep4$i$i$i2$i$i$i$i$i$i$i = 0, $scevgep4$i$i$i2$i$i$i$i$i$i$i13 = 0, $scevgep4$i$i$i2$i28 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $cps = sp + 24|0;
 $codepointlines = sp + 12|0;
 $line = sp;
 $character = sp + 48|0;
 ;HEAP32[$cps>>2]=0|0;HEAP32[$cps+4>>2]=0|0;HEAP32[$cps+8>>2]=0|0;HEAP32[$cps+12>>2]=0|0;HEAP32[$cps+16>>2]=0|0;HEAP32[$cps+20>>2]=0|0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 __THREW__ = 0;
 invoke_vii(4038,($text|0),($cps|0)); //@line 907 "modules/graphics/opengl/Font.cpp"
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 if ($1) {
  $68 = ___cxa_find_matching_catch_2()|0;
  $69 = tempRet0;
  $$1 = $68;$$12 = $69;
 } else {
  HEAP32[$codepointlines>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $2 = ((($codepointlines)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$2>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $3 = ((($codepointlines)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$3>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  __THREW__ = 0;
  invoke_viidii(4028,($this|0),($cps|0),(+$wraplimit),($codepointlines|0),($linewidths|0)); //@line 910 "modules/graphics/opengl/Font.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  L4: do {
   if ($5) {
    $70 = ___cxa_find_matching_catch_2()|0;
    $71 = tempRet0;
    $$0 = $70;$$01 = $71;
   } else {
    ;HEAP32[$line>>2]=0|0;HEAP32[$line+4>>2]=0|0;HEAP32[$line+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $6 = HEAP32[$codepointlines>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $7 = HEAP32[$2>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $8 = ($6|0)==($7|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    L7: do {
     if (!($8)) {
      $9 = ((($line)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $10 = ((($lines)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $11 = ((($lines)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $12 = ((($character)) + 1|0); //@line 41 "./libraries/utf8/utf8/unchecked.h"
      $13 = $character; //@line 923 "modules/graphics/opengl/Font.cpp"
      $14 = ((($character)) + 2|0); //@line 44 "./libraries/utf8/utf8/unchecked.h"
      $15 = ((($character)) + 3|0); //@line 49 "./libraries/utf8/utf8/unchecked.h"
      $16 = ((($character)) + 4|0); //@line 55 "./libraries/utf8/utf8/unchecked.h"
      $17 = ((($line)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $18 = ((($line)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $73 = 0;$__begin$sroa$0$033 = $6;
      L9: while(1) {
       $72 = $73 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $74 = ($72<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($74) {
        HEAP8[$9>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP8[$line>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       } else {
        $75 = HEAP32[$17>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP8[$75>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP32[$18>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       }
       $76 = ((($__begin$sroa$0$033)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $77 = HEAP32[$76>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $78 = HEAP32[$__begin$sroa$0$033>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $79 = (($77) - ($78))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $80 = $79 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       __THREW__ = 0;
       invoke_vii(3368,($line|0),($80|0)); //@line 917 "modules/graphics/opengl/Font.cpp"
       $81 = __THREW__; __THREW__ = 0;
       $82 = $81&1;
       if ($82) {
        label = 39;
        break;
       }
       $83 = HEAP32[$__begin$sroa$0$033>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $84 = HEAP32[$76>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $85 = ($83|0)==($84|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       if (!($85)) {
        $__begin2$sroa$0$032 = $83;
        while(1) {
         $99 = HEAP32[$__begin2$sroa$0$032>>2]|0; //@line 919 "modules/graphics/opengl/Font.cpp"
         ;HEAP8[$character>>0]=0|0;HEAP8[$character+1>>0]=0|0;HEAP8[$character+2>>0]=0|0;HEAP8[$character+3>>0]=0|0;HEAP8[$character+4>>0]=0|0; //@line 921 "modules/graphics/opengl/Font.cpp"
         $100 = ($99>>>0)<(128); //@line 40 "./libraries/utf8/utf8/unchecked.h"
         do {
          if ($100) {
           $101 = $99&255; //@line 41 "./libraries/utf8/utf8/unchecked.h"
           HEAP8[$character>>0] = $101; //@line 41 "./libraries/utf8/utf8/unchecked.h"
           $$0$i = $12;
          } else {
           $102 = ($99>>>0)<(2048); //@line 42 "./libraries/utf8/utf8/unchecked.h"
           if ($102) {
            $103 = $99 >>> 6; //@line 43 "./libraries/utf8/utf8/unchecked.h"
            $104 = $103 | 192; //@line 43 "./libraries/utf8/utf8/unchecked.h"
            $105 = $104&255; //@line 43 "./libraries/utf8/utf8/unchecked.h"
            HEAP8[$character>>0] = $105; //@line 43 "./libraries/utf8/utf8/unchecked.h"
            $106 = $99 & 63; //@line 44 "./libraries/utf8/utf8/unchecked.h"
            $107 = $106 | 128; //@line 44 "./libraries/utf8/utf8/unchecked.h"
            $108 = $107&255; //@line 44 "./libraries/utf8/utf8/unchecked.h"
            HEAP8[$12>>0] = $108; //@line 44 "./libraries/utf8/utf8/unchecked.h"
            $$0$i = $14;
            break;
           }
           $109 = ($99>>>0)<(65536); //@line 46 "./libraries/utf8/utf8/unchecked.h"
           if ($109) {
            $110 = $99 >>> 12; //@line 47 "./libraries/utf8/utf8/unchecked.h"
            $111 = $110 | 224; //@line 47 "./libraries/utf8/utf8/unchecked.h"
            $112 = $111&255; //@line 47 "./libraries/utf8/utf8/unchecked.h"
            HEAP8[$character>>0] = $112; //@line 47 "./libraries/utf8/utf8/unchecked.h"
            $113 = $99 >>> 6; //@line 48 "./libraries/utf8/utf8/unchecked.h"
            $114 = $113 & 63; //@line 48 "./libraries/utf8/utf8/unchecked.h"
            $115 = $114 | 128; //@line 48 "./libraries/utf8/utf8/unchecked.h"
            $116 = $115&255; //@line 48 "./libraries/utf8/utf8/unchecked.h"
            HEAP8[$12>>0] = $116; //@line 48 "./libraries/utf8/utf8/unchecked.h"
            $117 = $99 & 63; //@line 49 "./libraries/utf8/utf8/unchecked.h"
            $118 = $117 | 128; //@line 49 "./libraries/utf8/utf8/unchecked.h"
            $119 = $118&255; //@line 49 "./libraries/utf8/utf8/unchecked.h"
            HEAP8[$14>>0] = $119; //@line 49 "./libraries/utf8/utf8/unchecked.h"
            $$0$i = $15;
            break;
           } else {
            $120 = $99 >>> 18; //@line 52 "./libraries/utf8/utf8/unchecked.h"
            $121 = $120 | 240; //@line 52 "./libraries/utf8/utf8/unchecked.h"
            $122 = $121&255; //@line 52 "./libraries/utf8/utf8/unchecked.h"
            HEAP8[$character>>0] = $122; //@line 52 "./libraries/utf8/utf8/unchecked.h"
            $123 = $99 >>> 12; //@line 53 "./libraries/utf8/utf8/unchecked.h"
            $124 = $123 & 63; //@line 53 "./libraries/utf8/utf8/unchecked.h"
            $125 = $124 | 128; //@line 53 "./libraries/utf8/utf8/unchecked.h"
            $126 = $125&255; //@line 53 "./libraries/utf8/utf8/unchecked.h"
            HEAP8[$12>>0] = $126; //@line 53 "./libraries/utf8/utf8/unchecked.h"
            $127 = $99 >>> 6; //@line 54 "./libraries/utf8/utf8/unchecked.h"
            $128 = $127 & 63; //@line 54 "./libraries/utf8/utf8/unchecked.h"
            $129 = $128 | 128; //@line 54 "./libraries/utf8/utf8/unchecked.h"
            $130 = $129&255; //@line 54 "./libraries/utf8/utf8/unchecked.h"
            HEAP8[$14>>0] = $130; //@line 54 "./libraries/utf8/utf8/unchecked.h"
            $131 = $99 & 63; //@line 55 "./libraries/utf8/utf8/unchecked.h"
            $132 = $131 | 128; //@line 55 "./libraries/utf8/utf8/unchecked.h"
            $133 = $132&255; //@line 55 "./libraries/utf8/utf8/unchecked.h"
            HEAP8[$15>>0] = $133; //@line 55 "./libraries/utf8/utf8/unchecked.h"
            $$0$i = $16;
            break;
           }
          }
         } while(0);
         $134 = $$0$i; //@line 923 "modules/graphics/opengl/Font.cpp"
         $135 = (($134) - ($13))|0; //@line 923 "modules/graphics/opengl/Font.cpp"
         __THREW__ = 0;
         (invoke_iiii(3462,($line|0),($character|0),($135|0))|0); //@line 923 "modules/graphics/opengl/Font.cpp"
         $136 = __THREW__; __THREW__ = 0;
         $137 = $136&1;
         if ($137) {
          label = 38;
          break L9;
         }
         $138 = ((($__begin2$sroa$0$032)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
         $139 = ($138|0)==($84|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
         if ($139) {
          break;
         } else {
          $__begin2$sroa$0$032 = $138;
         }
        }
       }
       $86 = HEAP32[$10>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $87 = HEAP32[$11>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $88 = ($86|0)==($87|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($88) {
        __THREW__ = 0;
        invoke_vii(3517,($lines|0),($line|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $93 = __THREW__; __THREW__ = 0;
        $94 = $93&1;
        if ($94) {
         label = 39;
         break;
        }
       } else {
        __THREW__ = 0;
        invoke_vii(3351,($86|0),($line|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $89 = __THREW__; __THREW__ = 0;
        $90 = $89&1;
        if ($90) {
         label = 39;
         break;
        }
        $91 = HEAP32[$10>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $92 = ((($91)) + 12|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$10>>2] = $92; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       $140 = ((($__begin$sroa$0$033)) + 24|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       $141 = ($140|0)==($7|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       if ($141) {
        break L7;
       }
       $$pre = HEAP8[$line>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $73 = $$pre;$__begin$sroa$0$033 = $140;
      }
      if ((label|0) == 38) {
       $95 = ___cxa_find_matching_catch_2()|0;
       $96 = tempRet0;
       $lpad$phi$index = $95;$lpad$phi$index2 = $96;
      }
      else if ((label|0) == 39) {
       $97 = ___cxa_find_matching_catch_2()|0;
       $98 = tempRet0;
       $lpad$phi$index = $97;$lpad$phi$index2 = $98;
      }
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($line); //@line 928 "modules/graphics/opengl/Font.cpp"
      $$0 = $lpad$phi$index;$$01 = $lpad$phi$index2;
      break L4;
     }
    } while(0);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($line); //@line 928 "modules/graphics/opengl/Font.cpp"
    $19 = HEAP32[$codepointlines>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $20 = ($19|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($20)) {
     $21 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $22 = ($21|0)==($19|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($22) {
      $47 = $19;
     } else {
      $24 = $21;
      while(1) {
       $23 = ((($24)) + -24|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$2>>2] = $23; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $25 = ((($24)) + -12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $26 = HEAP32[$25>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $27 = ($26|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($27)) {
        $28 = ((($24)) + -8|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $29 = HEAP32[$28>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $30 = ($29|0)==($26|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($30)) {
         $__new_last1$i$i$i$i$i$i$i$i$i$i5 = $26; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $scevgep$i$i$i$i$i$i$i$i$i$i6 = ((($29)) + -8|0);
         $31 = $scevgep$i$i$i$i$i$i$i$i$i$i6;
         $32 = (($31) - ($__new_last1$i$i$i$i$i$i$i$i$i$i5))|0;
         $33 = $32 >>> 3;
         $34 = $33 ^ -1;
         $scevgep4$i$i$i$i$i$i$i$i$i$i7 = (($29) + ($34<<3)|0);
         HEAP32[$28>>2] = $scevgep4$i$i$i$i$i$i$i$i$i$i7; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
        __ZdlPv($26); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       }
       $35 = HEAP32[$23>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $36 = ($35|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $37 = $35; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($36)) {
        $40 = ((($24)) + -20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $41 = HEAP32[$40>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $42 = ($41|0)==($35|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($42)) {
         $scevgep$i$i$i1$i$i$i$i$i$i$i12 = ((($41)) + -4|0);
         $43 = $scevgep$i$i$i1$i$i$i$i$i$i$i12;
         $44 = (($43) - ($37))|0;
         $45 = $44 >>> 2;
         $46 = $45 ^ -1;
         $scevgep4$i$i$i2$i$i$i$i$i$i$i13 = (($41) + ($46<<2)|0);
         HEAP32[$40>>2] = $scevgep4$i$i$i2$i$i$i$i$i$i$i13; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
        __ZdlPv($35); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       }
       $38 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $39 = ($38|0)==($19|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($39) {
        break;
       } else {
        $24 = $38;
       }
      }
      $$pre$i17 = HEAP32[$codepointlines>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $47 = $$pre$i17;
     }
     __ZdlPv($47); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    }
    $48 = ((($cps)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $49 = HEAP32[$48>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $50 = ($49|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($50)) {
     $51 = ((($cps)) + 16|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $52 = HEAP32[$51>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $53 = ($52|0)==($49|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($53)) {
      $__new_last1$i$i$i$i21 = $49; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $scevgep$i$i$i$i22 = ((($52)) + -8|0);
      $54 = $scevgep$i$i$i$i22;
      $55 = (($54) - ($__new_last1$i$i$i$i21))|0;
      $56 = $55 >>> 3;
      $57 = $56 ^ -1;
      $scevgep4$i$i$i$i23 = (($52) + ($57<<3)|0);
      HEAP32[$51>>2] = $scevgep4$i$i$i$i23; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __ZdlPv($49); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    }
    $58 = HEAP32[$cps>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $59 = ($58|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $60 = $58; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($59) {
     STACKTOP = sp;return; //@line 928 "modules/graphics/opengl/Font.cpp"
    }
    $61 = ((($cps)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $62 = HEAP32[$61>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $63 = ($62|0)==($58|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($63)) {
     $scevgep$i$i$i1$i27 = ((($62)) + -4|0);
     $64 = $scevgep$i$i$i1$i27;
     $65 = (($64) - ($60))|0;
     $66 = $65 >>> 2;
     $67 = $66 ^ -1;
     $scevgep4$i$i$i2$i28 = (($62) + ($67<<2)|0);
     HEAP32[$61>>2] = $scevgep4$i$i$i2$i28; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($58); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    STACKTOP = sp;return; //@line 928 "modules/graphics/opengl/Font.cpp"
   }
  } while(0);
  $142 = HEAP32[$codepointlines>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $143 = ($142|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($143) {
   $$1 = $$0;$$12 = $$01;
  } else {
   $144 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $145 = ($144|0)==($142|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($145) {
    $170 = $142;
   } else {
    $147 = $144;
    while(1) {
     $146 = ((($147)) + -24|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$2>>2] = $146; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $148 = ((($147)) + -12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $149 = HEAP32[$148>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $150 = ($149|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($150)) {
      $151 = ((($147)) + -8|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $152 = HEAP32[$151>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $153 = ($152|0)==($149|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($153)) {
       $__new_last1$i$i$i$i$i$i$i$i$i$i = $149; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $scevgep$i$i$i$i$i$i$i$i$i$i = ((($152)) + -8|0);
       $154 = $scevgep$i$i$i$i$i$i$i$i$i$i;
       $155 = (($154) - ($__new_last1$i$i$i$i$i$i$i$i$i$i))|0;
       $156 = $155 >>> 3;
       $157 = $156 ^ -1;
       $scevgep4$i$i$i$i$i$i$i$i$i$i = (($152) + ($157<<3)|0);
       HEAP32[$151>>2] = $scevgep4$i$i$i$i$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($149); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $158 = HEAP32[$146>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $159 = ($158|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $160 = $158; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($159)) {
      $163 = ((($147)) + -20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $164 = HEAP32[$163>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $165 = ($164|0)==($158|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($165)) {
       $scevgep$i$i$i1$i$i$i$i$i$i$i = ((($164)) + -4|0);
       $166 = $scevgep$i$i$i1$i$i$i$i$i$i$i;
       $167 = (($166) - ($160))|0;
       $168 = $167 >>> 2;
       $169 = $168 ^ -1;
       $scevgep4$i$i$i2$i$i$i$i$i$i$i = (($164) + ($169<<2)|0);
       HEAP32[$163>>2] = $scevgep4$i$i$i2$i$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($158); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $161 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $162 = ($161|0)==($142|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($162) {
      break;
     } else {
      $147 = $161;
     }
    }
    $$pre$i = HEAP32[$codepointlines>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $170 = $$pre$i;
   }
   __ZdlPv($170); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $$1 = $$0;$$12 = $$01;
  }
 }
 $171 = ((($cps)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $172 = HEAP32[$171>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $173 = ($172|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($173)) {
  $174 = ((($cps)) + 16|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $175 = HEAP32[$174>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $176 = ($175|0)==($172|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($176)) {
   $__new_last1$i$i$i$i = $172; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $scevgep$i$i$i$i = ((($175)) + -8|0);
   $177 = $scevgep$i$i$i$i;
   $178 = (($177) - ($__new_last1$i$i$i$i))|0;
   $179 = $178 >>> 3;
   $180 = $179 ^ -1;
   $scevgep4$i$i$i$i = (($175) + ($180<<3)|0);
   HEAP32[$174>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($172); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $181 = HEAP32[$cps>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $182 = ($181|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $183 = $181; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($182) {
  ___resumeException($$1|0);
  // unreachable;
 }
 $184 = ((($cps)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $185 = HEAP32[$184>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $186 = ($185|0)==($181|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($186)) {
  $scevgep$i$i$i1$i = ((($185)) + -4|0);
  $187 = $scevgep$i$i$i1$i;
  $188 = (($187) - ($183))|0;
  $189 = $188 >>> 2;
  $190 = $189 ^ -1;
  $scevgep4$i$i$i2$i = (($185) + ($190<<2)|0);
  HEAP32[$184>>2] = $scevgep4$i$i$i2$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($181); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$1|0);
 // unreachable;
}
function __ZN4love8graphics6opengl4Font13setLineHeightEf($this,$height) {
 $this = $this|0;
 $height = +$height;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 932 "modules/graphics/opengl/Font.cpp"
 HEAPF32[$0>>2] = $height; //@line 932 "modules/graphics/opengl/Font.cpp"
 return; //@line 933 "modules/graphics/opengl/Font.cpp"
}
function __ZN4love8graphics6opengl4Font9setFilterERKNS0_7Texture6FilterE($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$01 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (__ZN4love8graphics7Texture14validateFilterERKNS1_6FilterEb($f,0)|0); //@line 942 "modules/graphics/opengl/Font.cpp"
 if (!($0)) {
  $1 = (___cxa_allocate_exception(16)|0); //@line 943 "modules/graphics/opengl/Font.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($1|0),(439276|0),($vararg_buffer|0)); //@line 943 "modules/graphics/opengl/Font.cpp"
  $2 = __THREW__; __THREW__ = 0;
  $3 = $2&1;
  if ($3) {
   $4 = ___cxa_find_matching_catch_2()|0;
   $5 = tempRet0;
   ___cxa_free_exception(($1|0)); //@line 943 "modules/graphics/opengl/Font.cpp"
   ___resumeException($4|0);
   // unreachable;
  } else {
   ___cxa_throw(($1|0),(96|0),(344|0)); //@line 943 "modules/graphics/opengl/Font.cpp"
   // unreachable; //@line 943 "modules/graphics/opengl/Font.cpp"
  }
 }
 $6 = ((($this)) + 96|0); //@line 945 "modules/graphics/opengl/Font.cpp"
 ;HEAP32[$6>>2]=HEAP32[$f>>2]|0;HEAP32[$6+4>>2]=HEAP32[$f+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$f+8>>2]|0;HEAP32[$6+12>>2]=HEAP32[$f+12>>2]|0; //@line 945 "modules/graphics/opengl/Font.cpp"
 $7 = ((($this)) + 40|0); //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = HEAP32[$7>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ((($this)) + 44|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = HEAP32[$9>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($8|0)==($10|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($11) {
  STACKTOP = sp;return; //@line 952 "modules/graphics/opengl/Font.cpp"
 } else {
  $__begin$sroa$0$01 = $8;
 }
 while(1) {
  $12 = HEAP32[$__begin$sroa$0$01>>2]|0; //@line 947 "modules/graphics/opengl/Font.cpp"
  __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$12); //@line 949 "modules/graphics/opengl/Font.cpp"
  __ZN4love8graphics6opengl6OpenGL16setTextureFilterERNS0_7Texture6FilterE(634044,$6); //@line 950 "modules/graphics/opengl/Font.cpp"
  $13 = ((($__begin$sroa$0$01)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $14 = ($13|0)==($10|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($14) {
   break;
  } else {
   $__begin$sroa$0$01 = $13;
  }
 }
 STACKTOP = sp;return; //@line 952 "modules/graphics/opengl/Font.cpp"
}
function __ZN4love8graphics6opengl4Font9getFilterEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 96|0); //@line 956 "modules/graphics/opengl/Font.cpp"
 return ($0|0); //@line 956 "modules/graphics/opengl/Font.cpp"
}
function __ZNK4love8graphics6opengl4Font9getAscentEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 117 "./common/Object.h"
 $3 = HEAP32[$2>>2]|0; //@line 983 "modules/graphics/opengl/Font.cpp"
 $4 = ((($3)) + 24|0); //@line 983 "modules/graphics/opengl/Font.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 983 "modules/graphics/opengl/Font.cpp"
 $6 = (FUNCTION_TABLE_ii[$5 & 8191]($2)|0); //@line 983 "modules/graphics/opengl/Font.cpp"
 return ($6|0); //@line 983 "modules/graphics/opengl/Font.cpp"
}
function __ZNK4love8graphics6opengl4Font10getDescentEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 117 "./common/Object.h"
 $3 = HEAP32[$2>>2]|0; //@line 988 "modules/graphics/opengl/Font.cpp"
 $4 = ((($3)) + 28|0); //@line 988 "modules/graphics/opengl/Font.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 988 "modules/graphics/opengl/Font.cpp"
 $6 = (FUNCTION_TABLE_ii[$5 & 8191]($2)|0); //@line 988 "modules/graphics/opengl/Font.cpp"
 return ($6|0); //@line 988 "modules/graphics/opengl/Font.cpp"
}
function __ZNK4love8graphics6opengl4Font8hasGlyphEj($this,$glyph) {
 $this = $this|0;
 $glyph = $glyph|0;
 var $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$05 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($this)) + 16|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($1|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($4) {
  $$12 = 0;
  return ($$12|0); //@line 1006 "modules/graphics/opengl/Font.cpp"
 } else {
  $__begin$sroa$0$05 = $1;
 }
 while(1) {
  $7 = HEAP32[$__begin$sroa$0$05>>2]|0; //@line 117 "./common/Object.h"
  $8 = HEAP32[$7>>2]|0; //@line 1001 "modules/graphics/opengl/Font.cpp"
  $9 = ((($8)) + 48|0); //@line 1001 "modules/graphics/opengl/Font.cpp"
  $10 = HEAP32[$9>>2]|0; //@line 1001 "modules/graphics/opengl/Font.cpp"
  $11 = (FUNCTION_TABLE_iii[$10 & 8191]($7,$glyph)|0); //@line 1001 "modules/graphics/opengl/Font.cpp"
  $6 = ((($__begin$sroa$0$05)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($11) {
   $$12 = 1;
   label = 4;
   break;
  }
  $5 = ($6|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($5) {
   $$12 = 0;
   label = 4;
   break;
  } else {
   $__begin$sroa$0$05 = $6;
  }
 }
 if ((label|0) == 4) {
  return ($$12|0); //@line 1006 "modules/graphics/opengl/Font.cpp"
 }
 return (0)|0;
}
function __ZNK4love8graphics6opengl4Font9hasGlyphsERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($this,$text) {
 $this = $this|0;
 $text = $text|0;
 var $$03 = 0, $$05 = 0, $$12 = 0, $$14 = 0, $$16 = 0, $$byval_copy = 0, $$byval_copy12 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $9 = 0, $__begin$sroa$0$05$i = 0, $eh$lpad$body$index5Z2D = 0, $eh$lpad$body$indexZ2D = 0;
 var $i = 0, $or$cond36 = 0, $temp$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy12 = sp + 32|0;
 $$byval_copy = sp + 28|0;
 $vararg_buffer = sp;
 $0 = sp + 24|0;
 $temp$i = sp + 20|0;
 $1 = sp + 16|0;
 $i = sp + 4|0;
 $2 = HEAP8[$text>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = $2 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ($3<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($text)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = HEAP32[$5>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = $2&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $7 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $4 ? $8 : $6; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = ($9|0)==(0); //@line 1010 "modules/graphics/opengl/Font.cpp"
 if ($10) {
  STACKTOP = sp;return 0; //@line 1032 "modules/graphics/opengl/Font.cpp"
 }
 $11 = ((($text)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = HEAP32[$11>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ((($text)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = $4 ? $13 : $12; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $15 = $14; //@line 1015 "modules/graphics/opengl/Font.cpp"
 $16 = (($14) + ($9)|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $17 = $16; //@line 1015 "modules/graphics/opengl/Font.cpp"
 HEAP32[$i>>2] = $15; //@line 277 "./libraries/utf8/utf8/checked.h"
 $18 = ((($i)) + 4|0); //@line 277 "./libraries/utf8/utf8/checked.h"
 HEAP32[$18>>2] = $15; //@line 277 "./libraries/utf8/utf8/checked.h"
 $19 = ((($i)) + 8|0); //@line 277 "./libraries/utf8/utf8/checked.h"
 HEAP32[$19>>2] = $17; //@line 277 "./libraries/utf8/utf8/checked.h"
 $20 = ($9|0)<(0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 do {
  if ($20) {
   $21 = (___cxa_allocate_exception(8)|0); //@line 280 "./libraries/utf8/utf8/checked.h"
   __THREW__ = 0;
   invoke_vii(3996,($21|0),(434517|0)); //@line 133 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/stdexcept"
   $22 = __THREW__; __THREW__ = 0;
   $23 = $22&1;
   if ($23) {
    $25 = ___cxa_find_matching_catch_3(1744|0)|0;
    $26 = tempRet0;
    ___cxa_free_exception(($21|0)); //@line 280 "./libraries/utf8/utf8/checked.h"
    $eh$lpad$body$index5Z2D = $26;$eh$lpad$body$indexZ2D = $25;
    break;
   } else {
    HEAP32[$21>>2] = (157912); //@line 133 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/stdexcept"
    __THREW__ = 0;
    invoke_viii(3350,($21|0),(9120|0),(3328|0)); //@line 280 "./libraries/utf8/utf8/checked.h"
    $24 = __THREW__; __THREW__ = 0;
    label = 20;
    break;
   }
  } else {
   $27 = ((($i)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $28 = ((($i)) + 8|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $29 = ((($this)) + 12|0); //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $30 = ((($this)) + 16|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $31 = ($9|0)==(0); //@line 297 "./libraries/utf8/utf8/checked.h"
   if ($31) {
    $$12 = 1;
    STACKTOP = sp;return ($$12|0);
   } else {
    $45 = $17;$48 = $15;
   }
   L12: while(1) {
    HEAP32[$0>>2] = $45; //@line 307 "./libraries/utf8/utf8/checked.h"
    __THREW__ = 0;
    ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0; //@line 307 "./libraries/utf8/utf8/checked.h"
    (invoke_iii(3997,($i|0),($$byval_copy|0))|0); //@line 307 "./libraries/utf8/utf8/checked.h"
    $46 = __THREW__; __THREW__ = 0;
    $47 = $46&1;
    if ($47) {
     label = 19;
     break;
    }
    HEAP32[$temp$i>>2] = $48; //@line 286 "./libraries/utf8/utf8/checked.h"
    HEAP32[$1>>2] = $45; //@line 287 "./libraries/utf8/utf8/checked.h"
    __THREW__ = 0;
    ;HEAP32[$$byval_copy12>>2]=HEAP32[$1>>2]|0; //@line 287 "./libraries/utf8/utf8/checked.h"
    $49 = (invoke_iii(3997,($temp$i|0),($$byval_copy12|0))|0); //@line 287 "./libraries/utf8/utf8/checked.h"
    $50 = __THREW__; __THREW__ = 0;
    $51 = $50&1;
    if ($51) {
     label = 19;
     break;
    }
    $52 = HEAP32[$29>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $53 = HEAP32[$30>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $54 = ($52|0)==($53|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($54) {
     $$12 = 0;
     label = 24;
     break;
    } else {
     $__begin$sroa$0$05$i = $52;
    }
    while(1) {
     $57 = HEAP32[$__begin$sroa$0$05$i>>2]|0; //@line 117 "./common/Object.h"
     $58 = HEAP32[$57>>2]|0; //@line 1001 "modules/graphics/opengl/Font.cpp"
     $59 = ((($58)) + 48|0); //@line 1001 "modules/graphics/opengl/Font.cpp"
     $60 = HEAP32[$59>>2]|0; //@line 1001 "modules/graphics/opengl/Font.cpp"
     __THREW__ = 0;
     $61 = (invoke_iii($60|0,($57|0),($49|0))|0); //@line 1001 "modules/graphics/opengl/Font.cpp"
     $62 = __THREW__; __THREW__ = 0;
     $63 = $62&1;
     if ($63) {
      label = 18;
      break L12;
     }
     $56 = ((($__begin$sroa$0$05$i)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($61) {
      break;
     }
     $55 = ($56|0)==($53|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($55) {
      $$12 = 0;
      label = 24;
      break L12;
     } else {
      $__begin$sroa$0$05$i = $56;
     }
    }
    $32 = HEAP32[$27>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $33 = ($32|0)==($14|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $34 = HEAP32[$28>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $35 = ($34|0)==($16|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $or$cond36 = $33 & $35; //@line 291 "./libraries/utf8/utf8/checked.h"
    if (!($or$cond36)) {
     label = 9;
     break;
    }
    $36 = $34; //@line 291 "./libraries/utf8/utf8/checked.h"
    $$pre = HEAP32[$i>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $37 = $$pre; //@line 291 "./libraries/utf8/utf8/checked.h"
    $38 = ($$pre|0)==($16|0); //@line 297 "./libraries/utf8/utf8/checked.h"
    if ($38) {
     $$12 = 1;
     label = 24;
     break;
    } else {
     $45 = $36;$48 = $37;
    }
   }
   if ((label|0) == 9) {
    $39 = (___cxa_allocate_exception(8)|0); //@line 292 "./libraries/utf8/utf8/checked.h"
    __THREW__ = 0;
    invoke_vii(3996,($39|0),(434549|0)); //@line 292 "./libraries/utf8/utf8/checked.h"
    $40 = __THREW__; __THREW__ = 0;
    $41 = $40&1;
    if ($41) {
     $43 = ___cxa_find_matching_catch_3(1744|0)|0;
     $44 = tempRet0;
     ___cxa_free_exception(($39|0)); //@line 292 "./libraries/utf8/utf8/checked.h"
     $eh$lpad$body$index5Z2D = $44;$eh$lpad$body$indexZ2D = $43;
     break;
    } else {
     __THREW__ = 0;
     invoke_viii(3350,($39|0),(9072|0),(3328|0)); //@line 292 "./libraries/utf8/utf8/checked.h"
     $42 = __THREW__; __THREW__ = 0;
     label = 20;
     break;
    }
   }
   else if ((label|0) == 18) {
    $64 = ___cxa_find_matching_catch_3(1744|0)|0;
    $65 = tempRet0;
    $eh$lpad$body$index5Z2D = $65;$eh$lpad$body$indexZ2D = $64;
    break;
   }
   else if ((label|0) == 19) {
    $66 = ___cxa_find_matching_catch_3(1744|0)|0;
    $67 = tempRet0;
    $eh$lpad$body$index5Z2D = $67;$eh$lpad$body$indexZ2D = $66;
    break;
   }
   else if ((label|0) == 24) {
    STACKTOP = sp;return ($$12|0);
   }
  }
 } while(0);
 if ((label|0) == 20) {
  $68 = ___cxa_find_matching_catch_3(1744|0)|0;
  $69 = tempRet0;
  $eh$lpad$body$index5Z2D = $69;$eh$lpad$body$indexZ2D = $68;
 }
 $70 = (_llvm_eh_typeid_for((1744|0))|0); //@line 1025 "modules/graphics/opengl/Font.cpp"
 $71 = ($eh$lpad$body$index5Z2D|0)==($70|0); //@line 1025 "modules/graphics/opengl/Font.cpp"
 if (!($71)) {
  $$14 = $eh$lpad$body$indexZ2D;$$16 = $eh$lpad$body$index5Z2D;
  ___resumeException($$14|0);
  // unreachable;
 }
 $72 = (___cxa_begin_catch(($eh$lpad$body$indexZ2D|0))|0); //@line 1025 "modules/graphics/opengl/Font.cpp"
 $73 = (___cxa_allocate_exception(16)|0); //@line 1028 "modules/graphics/opengl/Font.cpp"
 $74 = HEAP32[$72>>2]|0; //@line 1028 "modules/graphics/opengl/Font.cpp"
 $75 = ((($74)) + 8|0); //@line 1028 "modules/graphics/opengl/Font.cpp"
 $76 = HEAP32[$75>>2]|0; //@line 1028 "modules/graphics/opengl/Font.cpp"
 $77 = (FUNCTION_TABLE_ii[$76 & 8191]($72)|0); //@line 1028 "modules/graphics/opengl/Font.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $77; //@line 1028 "modules/graphics/opengl/Font.cpp"
 invoke_viii(3349,($73|0),(434605|0),($vararg_buffer|0)); //@line 1028 "modules/graphics/opengl/Font.cpp"
 $78 = __THREW__; __THREW__ = 0;
 $79 = $78&1;
 if ($79) {
  $83 = ___cxa_find_matching_catch_2()|0;
  $84 = tempRet0;
  ___cxa_free_exception(($73|0)); //@line 1028 "modules/graphics/opengl/Font.cpp"
  $$03 = $83;$$05 = $84;
 } else {
  __THREW__ = 0;
  invoke_viii(3350,($73|0),(96|0),(344|0)); //@line 1028 "modules/graphics/opengl/Font.cpp"
  $80 = __THREW__; __THREW__ = 0;
  $81 = ___cxa_find_matching_catch_2()|0;
  $82 = tempRet0;
  $$03 = $81;$$05 = $82;
 }
 __THREW__ = 0;
 invoke_v(3378); //@line 1029 "modules/graphics/opengl/Font.cpp"
 $85 = __THREW__; __THREW__ = 0;
 $86 = $85&1;
 if ($86) {
  $87 = ___cxa_find_matching_catch_3(0|0)|0;
  $88 = tempRet0;
  ___clang_call_terminate($87); //@line 1029 "modules/graphics/opengl/Font.cpp"
  // unreachable; //@line 1029 "modules/graphics/opengl/Font.cpp"
 } else {
  $$14 = $$03;$$16 = $$05;
  ___resumeException($$14|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl4Font12setFallbacksERKNSt3__16vectorIPS2_NS3_9allocatorIS5_EEEE($this,$fallbacks) {
 $this = $this|0;
 $fallbacks = $fallbacks|0;
 var $$pre$i = 0, $$pre$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$02 = 0, $__begin2$sroa$0$01 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP32[$fallbacks>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($fallbacks)) + 4|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ($0|0)==($2|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 L1: do {
  if (!($3)) {
   $4 = ((($this)) + 92|0); //@line 1038 "modules/graphics/opengl/Font.cpp"
   $5 = HEAP32[$4>>2]|0; //@line 1038 "modules/graphics/opengl/Font.cpp"
   $__begin$sroa$0$02 = $0;
   while(1) {
    $37 = HEAP32[$__begin$sroa$0$02>>2]|0; //@line 1036 "modules/graphics/opengl/Font.cpp"
    $38 = ((($37)) + 92|0); //@line 1038 "modules/graphics/opengl/Font.cpp"
    $39 = HEAP32[$38>>2]|0; //@line 1038 "modules/graphics/opengl/Font.cpp"
    $40 = ($39|0)==($5|0); //@line 1038 "modules/graphics/opengl/Font.cpp"
    $7 = ((($__begin$sroa$0$02)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if (!($40)) {
     break;
    }
    $6 = ($7|0)==($2|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($6) {
     break L1;
    } else {
     $__begin$sroa$0$02 = $7;
    }
   }
   $41 = (___cxa_allocate_exception(16)|0); //@line 1039 "modules/graphics/opengl/Font.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($41|0),(434700|0),($vararg_buffer|0)); //@line 1039 "modules/graphics/opengl/Font.cpp"
   $42 = __THREW__; __THREW__ = 0;
   $43 = $42&1;
   if ($43) {
    $44 = ___cxa_find_matching_catch_2()|0;
    $45 = tempRet0;
    ___cxa_free_exception(($41|0)); //@line 1039 "modules/graphics/opengl/Font.cpp"
    ___resumeException($44|0);
    // unreachable;
   } else {
    ___cxa_throw(($41|0),(96|0),(344|0)); //@line 1039 "modules/graphics/opengl/Font.cpp"
    // unreachable; //@line 1039 "modules/graphics/opengl/Font.cpp"
   }
  }
 } while(0);
 $8 = ((($this)) + 12|0); //@line 1042 "modules/graphics/opengl/Font.cpp"
 $9 = ((($this)) + 16|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = HEAP32[$9>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = HEAP32[$8>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = (($10) - ($11))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = $12 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = ($13|0)==(0); //@line 1998 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = $11; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = $10; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 L11: do {
  if ($14) {
   __ZNSt3__16vectorIN4love9StrongRefINS1_4font10RasterizerEEENS_9allocatorIS5_EEE8__appendEj($8,1); //@line 1999 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  } else {
   $17 = ($13>>>0)>(1); //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($17) {
    $18 = ((($15)) + 4|0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $19 = ($16|0)==($18|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($19)) {
     $21 = $16;
     while(1) {
      $20 = ((($21)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$9>>2] = $20; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $22 = HEAP32[$20>>2]|0; //@line 106 "./common/Object.h"
      $23 = ($22|0)==(0|0); //@line 106 "./common/Object.h"
      if ($23) {
       $25 = $20;
      } else {
       $26 = HEAP32[$22>>2]|0; //@line 106 "./common/Object.h"
       $27 = ((($26)) + 12|0); //@line 106 "./common/Object.h"
       $28 = HEAP32[$27>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($28|0,($22|0)); //@line 106 "./common/Object.h"
       $29 = __THREW__; __THREW__ = 0;
       $30 = $29&1;
       if ($30) {
        break;
       }
       $$pre$i$i$i = HEAP32[$9>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $25 = $$pre$i$i$i;
      }
      $24 = ($25|0)==($18|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($24) {
       break L11;
      } else {
       $21 = $25;
      }
     }
     $31 = ___cxa_find_matching_catch_3(0|0)|0;
     $32 = tempRet0;
     ___clang_call_terminate($31); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   }
  }
 } while(0);
 $33 = HEAP32[$fallbacks>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $34 = HEAP32[$1>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $35 = ($33|0)==($34|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($35) {
  STACKTOP = sp;return; //@line 1047 "modules/graphics/opengl/Font.cpp"
 }
 $36 = ((($this)) + 20|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $__begin2$sroa$0$01 = $33;
 while(1) {
  $46 = HEAP32[$__begin2$sroa$0$01>>2]|0; //@line 1045 "modules/graphics/opengl/Font.cpp"
  $47 = ((($46)) + 12|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $48 = HEAP32[$47>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $49 = HEAP32[$9>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $50 = HEAP32[$36>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $51 = ($49|0)==($50|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($51) {
   __ZNSt3__16vectorIN4love9StrongRefINS1_4font10RasterizerEEENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_($8,$48); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  } else {
   $52 = HEAP32[$48>>2]|0; //@line 139 "./common/Object.h"
   HEAP32[$49>>2] = $52; //@line 99 "./common/Object.h"
   $53 = ($52|0)==(0|0); //@line 101 "./common/Object.h"
   if ($53) {
    $58 = $49;
   } else {
    $54 = HEAP32[$52>>2]|0; //@line 101 "./common/Object.h"
    $55 = ((($54)) + 8|0); //@line 101 "./common/Object.h"
    $56 = HEAP32[$55>>2]|0; //@line 101 "./common/Object.h"
    FUNCTION_TABLE_vi[$56 & 8191]($52); //@line 101 "./common/Object.h"
    $$pre$i = HEAP32[$9>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $58 = $$pre$i;
   }
   $57 = ((($58)) + 4|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$9>>2] = $57; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  $59 = ((($__begin2$sroa$0$01)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $60 = ($59|0)==($34|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($60) {
   break;
  } else {
   $__begin2$sroa$0$01 = $59;
  }
 }
 STACKTOP = sp;return; //@line 1047 "modules/graphics/opengl/Font.cpp"
}
function __ZNSt3__16vectorIN4love9StrongRefINS1_4font10RasterizerEEENS_9allocatorIS5_EEE8__appendEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$0$i10 = 0, $$0$i11 = 0, $$01$i$i = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre14 = 0, $$pre15 = 0, $$promoted1$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $__v$sroa$6$0 = 0, $__v$sroa$6$0$lcssa = 0, $scevgep$i = 0, $scevgep$i9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($this)) + 4|0); //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($1) - ($3))|0; //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = $4 >> 2; //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)<($__n>>>0); //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($6)) {
  $7 = $3; //@line 988 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$promoted1$i = $3; //@line 1045 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $8 = $__n << 2; //@line 985 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  _memset(($$promoted1$i|0),0,($8|0))|0; //@line 88 "./common/Object.h"
  $scevgep$i = (($7) + ($__n<<2)|0); //@line 985 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$2>>2] = $scevgep$i; //@line 988 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  return; //@line 1053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $9 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($3) - ($9))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $10 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = (($11) + ($__n))|0; //@line 1049 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ($12>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($13) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$0>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre14 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre14;$16 = $$pre;
 } else {
  $$in = $9;$16 = $1;
 }
 $14 = $$in; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = (($16) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = $15 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $18 = ($17>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($18) {
  $23 = $15 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $24 = ($23>>>0)<($12>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $25 = $24 ? $12 : $23; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $26 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $27 = (($26) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $28 = $27 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $29 = ($25|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $30 = $26; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($29) {
   $$0$i11 = 0;$34 = 0;$35 = $28;$42 = $30;
  } else {
   $$0$i10 = $25;$93 = $30;$94 = $28;
   label = 8;
  }
 } else {
  $19 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = (($19) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = $20 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = $19; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $$0$i10 = 1073741823;$93 = $22;$94 = $21;
  label = 8;
 }
 if ((label|0) == 8) {
  $31 = $$0$i10 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $32 = (__Znwj($31)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i11 = $$0$i10;$34 = $32;$35 = $94;$42 = $93;
 }
 $33 = (($34) + ($35<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $36 = $33; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $37 = (($34) + ($$0$i11<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $38 = $37; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $39 = $__n << 2; //@line 215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 _memset(($33|0),0,($39|0))|0; //@line 88 "./common/Object.h"
 $scevgep$i9 = (($33) + ($__n<<2)|0); //@line 215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $40 = $scevgep$i9; //@line 218 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $41 = ($42|0)==($14|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 do {
  if ($41) {
   $57 = $36;$59 = $$in;$61 = $42;
  } else {
   $$01$i$i = $42;$45 = $33;$__v$sroa$6$0 = $36;
   while(1) {
    $43 = ((($$01$i$i)) + -4|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $44 = ((($45)) + -4|0); //@line 99 "./common/Object.h"
    $46 = HEAP32[$43>>2]|0; //@line 139 "./common/Object.h"
    HEAP32[$44>>2] = $46; //@line 99 "./common/Object.h"
    $47 = ($46|0)==(0|0); //@line 101 "./common/Object.h"
    if (!($47)) {
     $48 = HEAP32[$46>>2]|0; //@line 101 "./common/Object.h"
     $49 = ((($48)) + 8|0); //@line 101 "./common/Object.h"
     $50 = HEAP32[$49>>2]|0; //@line 101 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($50|0,($46|0)); //@line 101 "./common/Object.h"
     $51 = __THREW__; __THREW__ = 0;
     $52 = $51&1;
     if ($52) {
      $__v$sroa$6$0$lcssa = $__v$sroa$6$0;
      break;
     }
    }
    $53 = $__v$sroa$6$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $54 = ((($53)) + -4|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $55 = $54; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $56 = ($43|0)==($14|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    if ($56) {
     $$lcssa = $55;
     label = 13;
     break;
    } else {
     $$01$i$i = $43;$45 = $54;$__v$sroa$6$0 = $55;
    }
   }
   if ((label|0) == 13) {
    $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $$pre15 = HEAP32[$2>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $57 = $$lcssa;$59 = $$pre$i;$61 = $$pre15;
    break;
   }
   $76 = ___cxa_find_matching_catch_2()|0;
   $77 = tempRet0;
   $78 = $__v$sroa$6$0$lcssa; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $79 = ($scevgep$i9|0)==($78|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   L23: do {
    if (!($79)) {
     $81 = $scevgep$i9;
     while(1) {
      $80 = ((($81)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $82 = HEAP32[$80>>2]|0; //@line 106 "./common/Object.h"
      $83 = ($82|0)==(0|0); //@line 106 "./common/Object.h"
      if (!($83)) {
       $85 = HEAP32[$82>>2]|0; //@line 106 "./common/Object.h"
       $86 = ((($85)) + 12|0); //@line 106 "./common/Object.h"
       $87 = HEAP32[$86>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($87|0,($82|0)); //@line 106 "./common/Object.h"
       $88 = __THREW__; __THREW__ = 0;
       $89 = $88&1;
       if ($89) {
        break;
       }
      }
      $84 = ($80|0)==($78|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if ($84) {
       break L23;
      } else {
       $81 = $80;
      }
     }
     $90 = ___cxa_find_matching_catch_3(0|0)|0;
     $91 = tempRet0;
     ___clang_call_terminate($90); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   } while(0);
   $92 = ($34|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($92) {
    ___resumeException($76|0);
    // unreachable;
   }
   __ZdlPv($34); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($76|0);
   // unreachable;
  }
 } while(0);
 HEAP32[$this>>2] = $57; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$2>>2] = $40; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $38; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $58 = $59; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $60 = ($61|0)==($58|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 L36: do {
  if (!($60)) {
   $63 = $61;
   while(1) {
    $62 = ((($63)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $64 = HEAP32[$62>>2]|0; //@line 106 "./common/Object.h"
    $65 = ($64|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($65)) {
     $67 = HEAP32[$64>>2]|0; //@line 106 "./common/Object.h"
     $68 = ((($67)) + 12|0); //@line 106 "./common/Object.h"
     $69 = HEAP32[$68>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($69|0,($64|0)); //@line 106 "./common/Object.h"
     $70 = __THREW__; __THREW__ = 0;
     $71 = $70&1;
     if ($71) {
      break;
     }
    }
    $66 = ($62|0)==($58|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($66) {
     break L36;
    } else {
     $63 = $62;
    }
   }
   $72 = ___cxa_find_matching_catch_3(0|0)|0;
   $73 = tempRet0;
   ___clang_call_terminate($72); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 } while(0);
 $74 = ($59|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($74) {
  return; //@line 1053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $75 = $59; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($75); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNSt3__16vectorIN4love9StrongRefINS1_4font10RasterizerEEENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i8 = 0, $$0$i9 = 0, $$01$i$i = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre12 = 0, $$pre13 = 0, $$pre14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__v$sroa$15$0 = 0, $__v$sroa$6$0 = 0, $__v$sroa$6$0$lcssa = 0, $__v$sroa$6$1 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = $$in; //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = HEAP32[$8>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $10 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($11>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $17 = $10 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ($17>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $19 = $18 ? $5 : $17; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = (($20) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = $21 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = ($19|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $24 = $20; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($23) {
   $$0$i9 = 0;$101 = $24;$28 = 0;$29 = $22;
  } else {
   $$0$i8 = $19;$100 = $22;$99 = $24;
   label = 6;
  }
 } else {
  $13 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = $14 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = $13; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $$0$i8 = 1073741823;$100 = $15;$99 = $16;
  label = 6;
 }
 if ((label|0) == 6) {
  $25 = $$0$i8 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (__Znwj($25)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i9 = $$0$i8;$101 = $99;$28 = $26;$29 = $100;
 }
 $27 = (($28) + ($29<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = $27; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $31 = (($28) + ($$0$i9<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $32 = $31; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $33 = HEAP32[$__x>>2]|0; //@line 139 "./common/Object.h"
 HEAP32[$27>>2] = $33; //@line 99 "./common/Object.h"
 $34 = ($33|0)==(0|0); //@line 101 "./common/Object.h"
 do {
  if ($34) {
   $43 = $101;$44 = $7;
   label = 10;
  } else {
   $35 = HEAP32[$33>>2]|0; //@line 101 "./common/Object.h"
   $36 = ((($35)) + 8|0); //@line 101 "./common/Object.h"
   $37 = HEAP32[$36>>2]|0; //@line 101 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($37|0,($33|0)); //@line 101 "./common/Object.h"
   $38 = __THREW__; __THREW__ = 0;
   $39 = $38&1;
   if ($39) {
    $81 = ___cxa_find_matching_catch_2()|0;
    $82 = tempRet0;
    $__v$sroa$15$0 = $30;$__v$sroa$6$1 = $30;$lpad$phi$index = $81;$lpad$phi$index2 = $82;
    break;
   } else {
    $$pre12 = HEAP32[$this>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre13 = HEAP32[$0>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $43 = $$pre13;$44 = $$pre12;
    label = 10;
    break;
   }
  }
 } while(0);
 L15: do {
  if ((label|0) == 10) {
   $40 = ((($27)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $41 = $40; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $42 = ($43|0)==($44|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   do {
    if ($42) {
     $45 = $43; //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $60 = $30;$62 = $45;$64 = $43;
    } else {
     $$01$i$i = $43;$48 = $27;$__v$sroa$6$0 = $30;
     while(1) {
      $46 = ((($$01$i$i)) + -4|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $47 = ((($48)) + -4|0); //@line 99 "./common/Object.h"
      $49 = HEAP32[$46>>2]|0; //@line 139 "./common/Object.h"
      HEAP32[$47>>2] = $49; //@line 99 "./common/Object.h"
      $50 = ($49|0)==(0|0); //@line 101 "./common/Object.h"
      if (!($50)) {
       $51 = HEAP32[$49>>2]|0; //@line 101 "./common/Object.h"
       $52 = ((($51)) + 8|0); //@line 101 "./common/Object.h"
       $53 = HEAP32[$52>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($53|0,($49|0)); //@line 101 "./common/Object.h"
       $54 = __THREW__; __THREW__ = 0;
       $55 = $54&1;
       if ($55) {
        $__v$sroa$6$0$lcssa = $__v$sroa$6$0;
        label = 24;
        break;
       }
      }
      $56 = $__v$sroa$6$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $57 = ((($56)) + -4|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $58 = $57; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $59 = ($46|0)==($44|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      if ($59) {
       $$lcssa = $58;
       label = 15;
       break;
      } else {
       $$01$i$i = $46;$48 = $57;$__v$sroa$6$0 = $58;
      }
     }
     if ((label|0) == 15) {
      $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $$pre14 = HEAP32[$0>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $60 = $$lcssa;$62 = $$pre$i;$64 = $$pre14;
      break;
     }
     else if ((label|0) == 24) {
      $79 = ___cxa_find_matching_catch_2()|0;
      $80 = tempRet0;
      $__v$sroa$15$0 = $41;$__v$sroa$6$1 = $__v$sroa$6$0$lcssa;$lpad$phi$index = $79;$lpad$phi$index2 = $80;
      break L15;
     }
    }
   } while(0);
   HEAP32[$this>>2] = $60; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$0>>2] = $41; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$8>>2] = $32; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $61 = $62; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $63 = ($64|0)==($61|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   L28: do {
    if (!($63)) {
     $66 = $64;
     while(1) {
      $65 = ((($66)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $67 = HEAP32[$65>>2]|0; //@line 106 "./common/Object.h"
      $68 = ($67|0)==(0|0); //@line 106 "./common/Object.h"
      if (!($68)) {
       $70 = HEAP32[$67>>2]|0; //@line 106 "./common/Object.h"
       $71 = ((($70)) + 12|0); //@line 106 "./common/Object.h"
       $72 = HEAP32[$71>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($72|0,($67|0)); //@line 106 "./common/Object.h"
       $73 = __THREW__; __THREW__ = 0;
       $74 = $73&1;
       if ($74) {
        break;
       }
      }
      $69 = ($65|0)==($61|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if ($69) {
       break L28;
      } else {
       $66 = $65;
      }
     }
     $75 = ___cxa_find_matching_catch_3(0|0)|0;
     $76 = tempRet0;
     ___clang_call_terminate($75); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   } while(0);
   $77 = ($62|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($77) {
    return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $78 = $62; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __ZdlPv($78); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
 } while(0);
 $83 = $__v$sroa$6$1; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $84 = $__v$sroa$15$0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $85 = ($84|0)==($83|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 L41: do {
  if (!($85)) {
   $87 = $84;
   while(1) {
    $86 = ((($87)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $88 = HEAP32[$86>>2]|0; //@line 106 "./common/Object.h"
    $89 = ($88|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($89)) {
     $91 = HEAP32[$88>>2]|0; //@line 106 "./common/Object.h"
     $92 = ((($91)) + 12|0); //@line 106 "./common/Object.h"
     $93 = HEAP32[$92>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($93|0,($88|0)); //@line 106 "./common/Object.h"
     $94 = __THREW__; __THREW__ = 0;
     $95 = $94&1;
     if ($95) {
      break;
     }
    }
    $90 = ($86|0)==($83|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($90) {
     break L41;
    } else {
     $87 = $86;
    }
   }
   $96 = ___cxa_find_matching_catch_3(0|0)|0;
   $97 = tempRet0;
   ___clang_call_terminate($96); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 } while(0);
 $98 = ($28|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($98) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 __ZdlPv($28); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZNK4love8graphics6opengl4Font17getTextureCacheIDEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 132|0); //@line 1051 "modules/graphics/opengl/Font.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1051 "modules/graphics/opengl/Font.cpp"
 return ($1|0); //@line 1051 "modules/graphics/opengl/Font.cpp"
}
function __ZN4love8graphics6opengl4Font11getConstantEPKcRNS2_9AlignModeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = $9 & 7; //@line 72 "./common/StringMap.h"
  $11 = (((633592 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (633592 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(8); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((633592 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 1056 "modules/graphics/opengl/Font.cpp"
  return ($$$2$i|0); //@line 1056 "modules/graphics/opengl/Font.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 1056 "modules/graphics/opengl/Font.cpp"
  return ($$$2$i|0); //@line 1056 "modules/graphics/opengl/Font.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl8GLBufferD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (24732); //@line 71 "modules/graphics/opengl/GLBuffer.cpp"
 $0 = ((($this)) + 20|0); //@line 72 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 72 "modules/graphics/opengl/GLBuffer.cpp"
 $2 = ($1|0)==(0); //@line 72 "modules/graphics/opengl/GLBuffer.cpp"
 do {
  if (!($2)) {
   $3 = ((($this)) + 5|0); //@line 235 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP8[$3>>0] = 0; //@line 235 "modules/graphics/opengl/GLBuffer.cpp"
   $4 = HEAP32[165381]|0; //@line 487 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   invoke_vii($4|0,1,($0|0)); //@line 487 "./libraries/glad/gladfuncs.hpp"
   $5 = __THREW__; __THREW__ = 0;
   $6 = $5&1;
   if ($6) {
    $7 = ___cxa_find_matching_catch_3(0|0)|0;
    $8 = tempRet0;
    __ZN4love8graphics8VolatileD2Ev($this); //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
    ___clang_call_terminate($7); //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
    // unreachable; //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
   } else {
    HEAP32[$0>>2] = 0; //@line 238 "modules/graphics/opengl/GLBuffer.cpp"
    break;
   }
  }
 } while(0);
 $9 = ((($this)) + 24|0); //@line 75 "modules/graphics/opengl/GLBuffer.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 75 "modules/graphics/opengl/GLBuffer.cpp"
 $11 = ($10|0)==(0|0); //@line 75 "modules/graphics/opengl/GLBuffer.cpp"
 if ($11) {
  __ZN4love8graphics8VolatileD2Ev($this); //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
  return; //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
 }
 __ZdaPv($10); //@line 75 "modules/graphics/opengl/GLBuffer.cpp"
 __ZN4love8graphics8VolatileD2Ev($this); //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
 return; //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZN4love8graphics6opengl8GLBufferD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (24732); //@line 71 "modules/graphics/opengl/GLBuffer.cpp"
 $0 = ((($this)) + 20|0); //@line 72 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 72 "modules/graphics/opengl/GLBuffer.cpp"
 $2 = ($1|0)==(0); //@line 72 "modules/graphics/opengl/GLBuffer.cpp"
 do {
  if (!($2)) {
   $3 = ((($this)) + 5|0); //@line 235 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP8[$3>>0] = 0; //@line 235 "modules/graphics/opengl/GLBuffer.cpp"
   $4 = HEAP32[165381]|0; //@line 487 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   invoke_vii($4|0,1,($0|0)); //@line 487 "./libraries/glad/gladfuncs.hpp"
   $5 = __THREW__; __THREW__ = 0;
   $6 = $5&1;
   if ($6) {
    $7 = ___cxa_find_matching_catch_3(0|0)|0;
    $8 = tempRet0;
    __ZN4love8graphics8VolatileD2Ev($this); //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
    ___clang_call_terminate($7); //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
    // unreachable; //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
   } else {
    HEAP32[$0>>2] = 0; //@line 238 "modules/graphics/opengl/GLBuffer.cpp"
    break;
   }
  }
 } while(0);
 $9 = ((($this)) + 24|0); //@line 75 "modules/graphics/opengl/GLBuffer.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 75 "modules/graphics/opengl/GLBuffer.cpp"
 $11 = ($10|0)==(0|0); //@line 75 "modules/graphics/opengl/GLBuffer.cpp"
 if ($11) {
  __ZN4love8graphics8VolatileD2Ev($this); //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
  __ZdlPv($this); //@line 71 "modules/graphics/opengl/GLBuffer.cpp"
  return; //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
 }
 __ZdaPv($10); //@line 75 "modules/graphics/opengl/GLBuffer.cpp"
 __ZN4love8graphics8VolatileD2Ev($this); //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
 __ZdlPv($this); //@line 71 "modules/graphics/opengl/GLBuffer.cpp"
 return; //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZN4love8graphics6opengl8GLBuffer12loadVolatileEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 (__ZN4love8graphics6opengl8GLBuffer4loadEb($this,1)|0); //@line 210 "modules/graphics/opengl/GLBuffer.cpp"
 return 1; //@line 210 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZN4love8graphics6opengl8GLBuffer4loadEb($this,$restore) {
 $this = $this|0;
 $restore = $restore|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 220 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = HEAP32[165357]|0; //@line 488 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$1 & 8191](1,$0); //@line 488 "./libraries/glad/gladfuncs.hpp"
 $2 = ((($this)) + 5|0); //@line 178 "modules/graphics/opengl/GLBuffer.cpp"
 $3 = HEAP8[$2>>0]|0; //@line 178 "modules/graphics/opengl/GLBuffer.cpp"
 $4 = ($3<<24>>24)==(0); //@line 178 "modules/graphics/opengl/GLBuffer.cpp"
 if ($4) {
  $5 = ((($this)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
  $6 = HEAP32[$5>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
  $7 = HEAP32[$0>>2]|0; //@line 180 "modules/graphics/opengl/GLBuffer.cpp"
  $8 = HEAP32[165557]|0; //@line 486 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$8 & 8191]($6,$7); //@line 486 "./libraries/glad/gladfuncs.hpp"
  $9 = ((($this)) + 4|0); //@line 181 "modules/graphics/opengl/GLBuffer.cpp"
  HEAP8[$9>>0] = 1; //@line 181 "modules/graphics/opengl/GLBuffer.cpp"
 }
 if ($restore) {
  $10 = ((($this)) + 24|0); //@line 225 "modules/graphics/opengl/GLBuffer.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 225 "modules/graphics/opengl/GLBuffer.cpp"
  $19 = $11;
 } else {
  $19 = 0;
 }
 $12 = ((($this)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
 $13 = HEAP32[$12>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
 $14 = ((($this)) + 8|0); //@line 76 "modules/graphics/opengl/GLBuffer.h"
 $15 = HEAP32[$14>>2]|0; //@line 76 "modules/graphics/opengl/GLBuffer.h"
 $16 = ((($this)) + 16|0); //@line 96 "modules/graphics/opengl/GLBuffer.h"
 $17 = HEAP32[$16>>2]|0; //@line 96 "modules/graphics/opengl/GLBuffer.h"
 $18 = HEAP32[165565]|0; //@line 490 "./libraries/glad/gladfuncs.hpp"
 __THREW__ = 0;
 invoke_viiii($18|0,($13|0),($15|0),($19|0),($17|0)); //@line 490 "./libraries/glad/gladfuncs.hpp"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $31 = ___cxa_find_matching_catch_2()|0;
  $32 = tempRet0;
  $33 = ((($this)) + 4|0); //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
  $34 = HEAP8[$33>>0]|0; //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
  $35 = ($34<<24>>24)==(0); //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
  if ($35) {
   HEAP8[$33>>0] = 0; //@line 190 "modules/graphics/opengl/GLBuffer.cpp"
   ___resumeException($31|0);
   // unreachable;
  }
  $36 = HEAP32[$12>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
  $37 = HEAP32[165557]|0; //@line 486 "./libraries/glad/gladfuncs.hpp"
  __THREW__ = 0;
  invoke_vii($37|0,($36|0),0); //@line 486 "./libraries/glad/gladfuncs.hpp"
  $38 = __THREW__; __THREW__ = 0;
  $39 = $38&1;
  if ($39) {
   $40 = ___cxa_find_matching_catch_3(0|0)|0;
   $41 = tempRet0;
   ___clang_call_terminate($40); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
  } else {
   HEAP8[$33>>0] = 0; //@line 190 "modules/graphics/opengl/GLBuffer.cpp"
   ___resumeException($31|0);
   // unreachable;
  }
 } else {
  $22 = ((($this)) + 4|0); //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
  $23 = HEAP8[$22>>0]|0; //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
  $24 = ($23<<24>>24)==(0); //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
  if ($24) {
   HEAP8[$22>>0] = 0; //@line 190 "modules/graphics/opengl/GLBuffer.cpp"
   return 1; //@line 231 "modules/graphics/opengl/GLBuffer.cpp"
  }
  $25 = HEAP32[$12>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
  $26 = HEAP32[165557]|0; //@line 486 "./libraries/glad/gladfuncs.hpp"
  __THREW__ = 0;
  invoke_vii($26|0,($25|0),0); //@line 486 "./libraries/glad/gladfuncs.hpp"
  $27 = __THREW__; __THREW__ = 0;
  $28 = $27&1;
  if ($28) {
   $29 = ___cxa_find_matching_catch_3(0|0)|0;
   $30 = tempRet0;
   ___clang_call_terminate($29); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
  } else {
   HEAP8[$22>>0] = 0; //@line 190 "modules/graphics/opengl/GLBuffer.cpp"
   return 1; //@line 231 "modules/graphics/opengl/GLBuffer.cpp"
  }
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl8GLBuffer14unloadVolatileEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 5|0); //@line 235 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP8[$0>>0] = 0; //@line 235 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = ((($this)) + 20|0); //@line 237 "modules/graphics/opengl/GLBuffer.cpp"
 $2 = HEAP32[165381]|0; //@line 487 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$2 & 8191](1,$1); //@line 487 "./libraries/glad/gladfuncs.hpp"
 HEAP32[$1>>2] = 0; //@line 238 "modules/graphics/opengl/GLBuffer.cpp"
 return; //@line 216 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZN4love8graphics6opengl8GLBufferC2EjPKvjjj($this,$size,$data,$target,$usage,$mapflags) {
 $this = $this|0;
 $size = $size|0;
 $data = $data|0;
 $target = $target|0;
 $usage = $usage|0;
 $mapflags = $mapflags|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 __ZN4love8graphics8VolatileC2Ev($this); //@line 37 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP32[$this>>2] = (24732); //@line 48 "modules/graphics/opengl/GLBuffer.cpp"
 $0 = ((($this)) + 4|0); //@line 38 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP8[$0>>0] = 0; //@line 38 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = ((($this)) + 5|0); //@line 39 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP8[$1>>0] = 0; //@line 39 "modules/graphics/opengl/GLBuffer.cpp"
 $2 = ((($this)) + 8|0); //@line 40 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP32[$2>>2] = $size; //@line 40 "modules/graphics/opengl/GLBuffer.cpp"
 $3 = ((($this)) + 12|0); //@line 41 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP32[$3>>2] = $target; //@line 41 "modules/graphics/opengl/GLBuffer.cpp"
 $4 = ((($this)) + 16|0); //@line 42 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP32[$4>>2] = $usage; //@line 42 "modules/graphics/opengl/GLBuffer.cpp"
 $5 = ((($this)) + 20|0); //@line 43 "modules/graphics/opengl/GLBuffer.cpp"
 $6 = ((($this)) + 36|0); //@line 47 "modules/graphics/opengl/GLBuffer.cpp"
 ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;HEAP32[$5+12>>2]=0|0; //@line 44 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP32[$6>>2] = $mapflags; //@line 47 "modules/graphics/opengl/GLBuffer.cpp"
 __THREW__ = 0;
 $7 = (invoke_ii(3492,($size|0))|0); //@line 51 "modules/graphics/opengl/GLBuffer.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if (!($9)) {
  $10 = ((($this)) + 24|0); //@line 44 "modules/graphics/opengl/GLBuffer.cpp"
  HEAP32[$10>>2] = $7; //@line 51 "modules/graphics/opengl/GLBuffer.cpp"
  $11 = ($data|0)!=(0|0); //@line 58 "modules/graphics/opengl/GLBuffer.cpp"
  if ($11) {
   _memcpy(($7|0),($data|0),($size|0))|0; //@line 59 "modules/graphics/opengl/GLBuffer.cpp"
  }
  __THREW__ = 0;
  (invoke_iii(4041,($this|0),($11|0))|0); //@line 61 "modules/graphics/opengl/GLBuffer.cpp"
  $26 = __THREW__; __THREW__ = 0;
  $27 = $26&1;
  if (!($27)) {
   STACKTOP = sp;return; //@line 68 "modules/graphics/opengl/GLBuffer.cpp"
  }
  $28 = ___cxa_find_matching_catch_2()|0;
  $29 = tempRet0;
  $$1 = $28;$$12 = $29;
  __ZN4love8graphics8VolatileD2Ev($this); //@line 68 "modules/graphics/opengl/GLBuffer.cpp"
  ___resumeException($$1|0);
  // unreachable;
 }
 $12 = ___cxa_find_matching_catch_3(9056|0)|0;
 $13 = tempRet0;
 $14 = (_llvm_eh_typeid_for((9056|0))|0); //@line 52 "modules/graphics/opengl/GLBuffer.cpp"
 $15 = ($13|0)==($14|0); //@line 52 "modules/graphics/opengl/GLBuffer.cpp"
 if (!($15)) {
  $$1 = $12;$$12 = $13;
  __ZN4love8graphics8VolatileD2Ev($this); //@line 68 "modules/graphics/opengl/GLBuffer.cpp"
  ___resumeException($$1|0);
  // unreachable;
 }
 (___cxa_begin_catch(($12|0))|0); //@line 52 "modules/graphics/opengl/GLBuffer.cpp"
 $16 = (___cxa_allocate_exception(16)|0); //@line 55 "modules/graphics/opengl/GLBuffer.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($16|0),(507610|0),($vararg_buffer|0)); //@line 55 "modules/graphics/opengl/GLBuffer.cpp"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if ($18) {
  $22 = ___cxa_find_matching_catch_2()|0;
  $23 = tempRet0;
  ___cxa_free_exception(($16|0)); //@line 55 "modules/graphics/opengl/GLBuffer.cpp"
  $$0 = $22;$$01 = $23;
 } else {
  __THREW__ = 0;
  invoke_viii(3350,($16|0),(96|0),(344|0)); //@line 55 "modules/graphics/opengl/GLBuffer.cpp"
  $19 = __THREW__; __THREW__ = 0;
  $20 = ___cxa_find_matching_catch_2()|0;
  $21 = tempRet0;
  $$0 = $20;$$01 = $21;
 }
 __THREW__ = 0;
 invoke_v(3378); //@line 56 "modules/graphics/opengl/GLBuffer.cpp"
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $30 = ___cxa_find_matching_catch_3(0|0)|0;
  $31 = tempRet0;
  ___clang_call_terminate($30); //@line 56 "modules/graphics/opengl/GLBuffer.cpp"
  // unreachable; //@line 56 "modules/graphics/opengl/GLBuffer.cpp"
 } else {
  $$1 = $$0;$$12 = $$01;
  __ZN4love8graphics8VolatileD2Ev($this); //@line 68 "modules/graphics/opengl/GLBuffer.cpp"
  ___resumeException($$1|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl8GLBuffer3mapEv($this) {
 $this = $this|0;
 var $$0 = 0, $$0$in = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 5|0); //@line 80 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 80 "modules/graphics/opengl/GLBuffer.cpp"
 $2 = ($1<<24>>24)==(0); //@line 80 "modules/graphics/opengl/GLBuffer.cpp"
 if ($2) {
  HEAP8[$0>>0] = 1; //@line 83 "modules/graphics/opengl/GLBuffer.cpp"
  $3 = ((($this)) + 28|0); //@line 85 "modules/graphics/opengl/GLBuffer.cpp"
  HEAP32[$3>>2] = 0; //@line 85 "modules/graphics/opengl/GLBuffer.cpp"
  $4 = ((($this)) + 32|0); //@line 86 "modules/graphics/opengl/GLBuffer.cpp"
  HEAP32[$4>>2] = 0; //@line 86 "modules/graphics/opengl/GLBuffer.cpp"
 }
 $$0$in = ((($this)) + 24|0); //@line 81 "modules/graphics/opengl/GLBuffer.cpp"
 $$0 = HEAP32[$$0$in>>2]|0; //@line 81 "modules/graphics/opengl/GLBuffer.cpp"
 return ($$0|0); //@line 89 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZN4love8graphics6opengl8GLBuffer5unmapEv($this) {
 $this = $this|0;
 var $$phi$trans$insert = 0, $$pre = 0, $$pre$phiZ2D = 0, $$pre2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 5|0); //@line 110 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 110 "modules/graphics/opengl/GLBuffer.cpp"
 $2 = ($1<<24>>24)==(0); //@line 110 "modules/graphics/opengl/GLBuffer.cpp"
 if ($2) {
  return; //@line 158 "modules/graphics/opengl/GLBuffer.cpp"
 }
 $3 = ((($this)) + 36|0); //@line 113 "modules/graphics/opengl/GLBuffer.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 113 "modules/graphics/opengl/GLBuffer.cpp"
 $5 = $4 & 1; //@line 113 "modules/graphics/opengl/GLBuffer.cpp"
 $6 = ($5|0)==(0); //@line 113 "modules/graphics/opengl/GLBuffer.cpp"
 $7 = ((($this)) + 28|0); //@line 115 "modules/graphics/opengl/GLBuffer.cpp"
 if ($6) {
  HEAP32[$7>>2] = 0; //@line 120 "modules/graphics/opengl/GLBuffer.cpp"
  $19 = ((($this)) + 8|0); //@line 76 "modules/graphics/opengl/GLBuffer.h"
  $20 = HEAP32[$19>>2]|0; //@line 76 "modules/graphics/opengl/GLBuffer.h"
  $21 = ((($this)) + 32|0); //@line 121 "modules/graphics/opengl/GLBuffer.cpp"
  HEAP32[$21>>2] = $20; //@line 121 "modules/graphics/opengl/GLBuffer.cpp"
  $72 = $20;
 } else {
  $8 = ((($this)) + 8|0); //@line 76 "modules/graphics/opengl/GLBuffer.h"
  $9 = HEAP32[$8>>2]|0; //@line 76 "modules/graphics/opengl/GLBuffer.h"
  $10 = (($9) + -1)|0; //@line 115 "modules/graphics/opengl/GLBuffer.cpp"
  $11 = HEAP32[$7>>2]|0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $12 = ($10>>>0)<($11>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $13 = $12 ? $10 : $11; //@line 115 "modules/graphics/opengl/GLBuffer.cpp"
  HEAP32[$7>>2] = $13; //@line 115 "modules/graphics/opengl/GLBuffer.cpp"
  $14 = ((($this)) + 32|0); //@line 116 "modules/graphics/opengl/GLBuffer.cpp"
  $15 = (($9) - ($13))|0; //@line 116 "modules/graphics/opengl/GLBuffer.cpp"
  $16 = HEAP32[$14>>2]|0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = ($15>>>0)<($16>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $18 = $17 ? $15 : $16; //@line 116 "modules/graphics/opengl/GLBuffer.cpp"
  HEAP32[$14>>2] = $18; //@line 116 "modules/graphics/opengl/GLBuffer.cpp"
  $72 = $18;
 }
 $22 = ((($this)) + 4|0); //@line 126 "modules/graphics/opengl/GLBuffer.cpp"
 $23 = HEAP8[$22>>0]|0; //@line 126 "modules/graphics/opengl/GLBuffer.cpp"
 $24 = ($23<<24>>24)==(0); //@line 126 "modules/graphics/opengl/GLBuffer.cpp"
 if ($24) {
  $25 = ((($this)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
  $26 = HEAP32[$25>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
  $27 = ((($this)) + 20|0); //@line 128 "modules/graphics/opengl/GLBuffer.cpp"
  $28 = HEAP32[$27>>2]|0; //@line 128 "modules/graphics/opengl/GLBuffer.cpp"
  $29 = HEAP32[165557]|0; //@line 486 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$29 & 8191]($26,$28); //@line 486 "./libraries/glad/gladfuncs.hpp"
  HEAP8[$22>>0] = 1; //@line 129 "modules/graphics/opengl/GLBuffer.cpp"
  $$phi$trans$insert = ((($this)) + 32|0); //@line 132 "modules/graphics/opengl/GLBuffer.cpp"
  $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 132 "modules/graphics/opengl/GLBuffer.cpp"
  $$pre$phiZ2D = $$phi$trans$insert;$31 = $$pre;
 } else {
  $$pre2 = ((($this)) + 32|0); //@line 132 "modules/graphics/opengl/GLBuffer.cpp"
  $$pre$phiZ2D = $$pre2;$31 = $72;
 }
 $30 = ($31|0)==(0); //@line 132 "modules/graphics/opengl/GLBuffer.cpp"
 L12: do {
  if (!($30)) {
   $32 = ((($this)) + 16|0); //@line 96 "modules/graphics/opengl/GLBuffer.h"
   $33 = HEAP32[$32>>2]|0; //@line 96 "modules/graphics/opengl/GLBuffer.h"
   switch ($33|0) {
   case 35044:  {
    $34 = HEAP32[$7>>2]|0; //@line 137 "modules/graphics/opengl/GLBuffer.cpp"
    $35 = ((($this)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
    $36 = HEAP32[$35>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
    $37 = ((($this)) + 24|0); //@line 97 "modules/graphics/opengl/GLBuffer.cpp"
    $38 = HEAP32[$37>>2]|0; //@line 97 "modules/graphics/opengl/GLBuffer.cpp"
    $39 = (($38) + ($34)|0); //@line 97 "modules/graphics/opengl/GLBuffer.cpp"
    $40 = HEAP32[165513]|0; //@line 491 "./libraries/glad/gladfuncs.hpp"
    FUNCTION_TABLE_viiii[$40 & 8191]($36,$34,$31,$39); //@line 491 "./libraries/glad/gladfuncs.hpp"
    break L12;
    break;
   }
   case 35040:  {
    $41 = ((($this)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
    $42 = HEAP32[$41>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
    $43 = ((($this)) + 8|0); //@line 76 "modules/graphics/opengl/GLBuffer.h"
    $44 = HEAP32[$43>>2]|0; //@line 76 "modules/graphics/opengl/GLBuffer.h"
    $45 = HEAP32[165565]|0; //@line 490 "./libraries/glad/gladfuncs.hpp"
    FUNCTION_TABLE_viiii[$45 & 8191]($42,$44,0,35040); //@line 490 "./libraries/glad/gladfuncs.hpp"
    $46 = HEAP32[$41>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
    $47 = HEAP32[$43>>2]|0; //@line 76 "modules/graphics/opengl/GLBuffer.h"
    $48 = ((($this)) + 24|0); //@line 105 "modules/graphics/opengl/GLBuffer.cpp"
    $49 = HEAP32[$48>>2]|0; //@line 105 "modules/graphics/opengl/GLBuffer.cpp"
    $50 = HEAP32[$32>>2]|0; //@line 96 "modules/graphics/opengl/GLBuffer.h"
    $51 = HEAP32[165565]|0; //@line 490 "./libraries/glad/gladfuncs.hpp"
    FUNCTION_TABLE_viiii[$51 & 8191]($46,$47,$49,$50); //@line 490 "./libraries/glad/gladfuncs.hpp"
    break L12;
    break;
   }
   default: {
    $52 = ((($this)) + 8|0); //@line 76 "modules/graphics/opengl/GLBuffer.h"
    $53 = HEAP32[$52>>2]|0; //@line 76 "modules/graphics/opengl/GLBuffer.h"
    $54 = (($53>>>0) / 3)&-1; //@line 146 "modules/graphics/opengl/GLBuffer.cpp"
    $55 = ($31>>>0)<($54>>>0); //@line 146 "modules/graphics/opengl/GLBuffer.cpp"
    if ($55) {
     $65 = HEAP32[$7>>2]|0; //@line 149 "modules/graphics/opengl/GLBuffer.cpp"
     $66 = ((($this)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
     $67 = HEAP32[$66>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
     $68 = ((($this)) + 24|0); //@line 97 "modules/graphics/opengl/GLBuffer.cpp"
     $69 = HEAP32[$68>>2]|0; //@line 97 "modules/graphics/opengl/GLBuffer.cpp"
     $70 = (($69) + ($65)|0); //@line 97 "modules/graphics/opengl/GLBuffer.cpp"
     $71 = HEAP32[165513]|0; //@line 491 "./libraries/glad/gladfuncs.hpp"
     FUNCTION_TABLE_viiii[$71 & 8191]($67,$65,$31,$70); //@line 491 "./libraries/glad/gladfuncs.hpp"
     break L12;
    } else {
     $56 = ((($this)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
     $57 = HEAP32[$56>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
     $58 = HEAP32[165565]|0; //@line 490 "./libraries/glad/gladfuncs.hpp"
     FUNCTION_TABLE_viiii[$58 & 8191]($57,$53,0,$33); //@line 490 "./libraries/glad/gladfuncs.hpp"
     $59 = HEAP32[$56>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
     $60 = HEAP32[$52>>2]|0; //@line 76 "modules/graphics/opengl/GLBuffer.h"
     $61 = ((($this)) + 24|0); //@line 105 "modules/graphics/opengl/GLBuffer.cpp"
     $62 = HEAP32[$61>>2]|0; //@line 105 "modules/graphics/opengl/GLBuffer.cpp"
     $63 = HEAP32[$32>>2]|0; //@line 96 "modules/graphics/opengl/GLBuffer.h"
     $64 = HEAP32[165565]|0; //@line 490 "./libraries/glad/gladfuncs.hpp"
     FUNCTION_TABLE_viiii[$64 & 8191]($59,$60,$62,$63); //@line 490 "./libraries/glad/gladfuncs.hpp"
     break L12;
    }
   }
   }
  }
 } while(0);
 HEAP32[$7>>2] = 0; //@line 154 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP32[$$pre$phiZ2D>>2] = 0; //@line 155 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP8[$0>>0] = 0; //@line 157 "modules/graphics/opengl/GLBuffer.cpp"
 return; //@line 158 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZN4love8graphics6opengl8GLBuffer22setMappedRangeModifiedEjj($this,$offset,$modifiedsize) {
 $this = $this|0;
 $offset = $offset|0;
 $modifiedsize = $modifiedsize|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 5|0); //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
 $2 = ($1<<24>>24)==(0); //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
 if ($2) {
  return; //@line 174 "modules/graphics/opengl/GLBuffer.cpp"
 }
 $3 = ((($this)) + 36|0); //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
 $5 = $4 & 1; //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
 $6 = ($5|0)==(0); //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
 if ($6) {
  return; //@line 174 "modules/graphics/opengl/GLBuffer.cpp"
 }
 $7 = ((($this)) + 28|0); //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
 $9 = ((($this)) + 32|0); //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
 $11 = (($10) + ($8))|0; //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
 $12 = ($8>>>0)>($offset>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $13 = $12 ? $offset : $8; //@line 170 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP32[$7>>2] = $13; //@line 170 "modules/graphics/opengl/GLBuffer.cpp"
 $14 = (($modifiedsize) + ($offset))|0; //@line 172 "modules/graphics/opengl/GLBuffer.cpp"
 $15 = ($14>>>0)<($11>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $16 = $15 ? $11 : $14; //@line 172 "modules/graphics/opengl/GLBuffer.cpp"
 $17 = (($16) - ($13))|0; //@line 173 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP32[$9>>2] = $17; //@line 173 "modules/graphics/opengl/GLBuffer.cpp"
 return; //@line 174 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZN4love8graphics6opengl8GLBuffer4bindEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 5|0); //@line 178 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 178 "modules/graphics/opengl/GLBuffer.cpp"
 $2 = ($1<<24>>24)==(0); //@line 178 "modules/graphics/opengl/GLBuffer.cpp"
 if (!($2)) {
  return; //@line 183 "modules/graphics/opengl/GLBuffer.cpp"
 }
 $3 = ((($this)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
 $4 = HEAP32[$3>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
 $5 = ((($this)) + 20|0); //@line 180 "modules/graphics/opengl/GLBuffer.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 180 "modules/graphics/opengl/GLBuffer.cpp"
 $7 = HEAP32[165557]|0; //@line 486 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$7 & 8191]($4,$6); //@line 486 "./libraries/glad/gladfuncs.hpp"
 $8 = ((($this)) + 4|0); //@line 181 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP8[$8>>0] = 1; //@line 181 "modules/graphics/opengl/GLBuffer.cpp"
 return; //@line 183 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZN4love8graphics6opengl8GLBuffer6unbindEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
 $2 = ($1<<24>>24)==(0); //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
 if ($2) {
  HEAP8[$0>>0] = 0; //@line 190 "modules/graphics/opengl/GLBuffer.cpp"
  return; //@line 191 "modules/graphics/opengl/GLBuffer.cpp"
 }
 $3 = ((($this)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
 $4 = HEAP32[$3>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
 $5 = HEAP32[165557]|0; //@line 486 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$5 & 8191]($4,0); //@line 486 "./libraries/glad/gladfuncs.hpp"
 HEAP8[$0>>0] = 0; //@line 190 "modules/graphics/opengl/GLBuffer.cpp"
 return; //@line 191 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZN4love8graphics6opengl8GLBuffer4fillEjjPKv($this,$offset,$size,$data) {
 $this = $this|0;
 $offset = $offset|0;
 $size = $size|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 195 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 195 "modules/graphics/opengl/GLBuffer.cpp"
 $2 = (($1) + ($offset)|0); //@line 195 "modules/graphics/opengl/GLBuffer.cpp"
 _memcpy(($2|0),($data|0),($size|0))|0; //@line 195 "modules/graphics/opengl/GLBuffer.cpp"
 $3 = ((($this)) + 5|0); //@line 197 "modules/graphics/opengl/GLBuffer.cpp"
 $4 = HEAP8[$3>>0]|0; //@line 197 "modules/graphics/opengl/GLBuffer.cpp"
 $5 = ($4<<24>>24)==(0); //@line 197 "modules/graphics/opengl/GLBuffer.cpp"
 if ($5) {
  $21 = ((($this)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
  $22 = HEAP32[$21>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
  $23 = HEAP32[165513]|0; //@line 491 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiii[$23 & 8191]($22,$offset,$size,$data); //@line 491 "./libraries/glad/gladfuncs.hpp"
  return; //@line 201 "modules/graphics/opengl/GLBuffer.cpp"
 }
 $6 = ((($this)) + 36|0); //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
 $8 = $7 & 1; //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
 $9 = ($8|0)==(0); //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
 if ($9) {
  return; //@line 201 "modules/graphics/opengl/GLBuffer.cpp"
 }
 $10 = ((($this)) + 28|0); //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
 $12 = ((($this)) + 32|0); //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
 $14 = (($13) + ($11))|0; //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
 $15 = ($11>>>0)>($offset>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $16 = $15 ? $offset : $11; //@line 170 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP32[$10>>2] = $16; //@line 170 "modules/graphics/opengl/GLBuffer.cpp"
 $17 = (($size) + ($offset))|0; //@line 172 "modules/graphics/opengl/GLBuffer.cpp"
 $18 = ($17>>>0)<($14>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $19 = $18 ? $14 : $17; //@line 172 "modules/graphics/opengl/GLBuffer.cpp"
 $20 = (($19) - ($16))|0; //@line 173 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP32[$12>>2] = $20; //@line 173 "modules/graphics/opengl/GLBuffer.cpp"
 return; //@line 201 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZNK4love8graphics6opengl8GLBuffer10getPointerEj($this,$offset) {
 $this = $this|0;
 $offset = $offset|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (0 + ($offset)|0); //@line 205 "modules/graphics/opengl/GLBuffer.cpp"
 return ($0|0); //@line 205 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZN4love8graphics6opengl11QuadIndicesC2Ej($this,$size) {
 $this = $this|0;
 $size = $size|0;
 var $$0 = 0, $$02 = 0, $$1 = 0, $$13 = 0, $$2 = 0, $$24 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $8 = 0;
 var $9 = 0, $newbuffer$0 = 0, $newbuffer$1 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 HEAP32[$this>>2] = $size; //@line 252 "modules/graphics/opengl/GLBuffer.cpp"
 $0 = (($size) + -1)|0; //@line 257 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = ($0>>>0)>(178956969); //@line 257 "modules/graphics/opengl/GLBuffer.cpp"
 if ($1) {
  $2 = (___cxa_allocate_exception(16)|0); //@line 258 "modules/graphics/opengl/GLBuffer.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($2|0),(434779|0),($vararg_buffer|0)); //@line 258 "modules/graphics/opengl/GLBuffer.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if (!($4)) {
   ___cxa_throw(($2|0),(96|0),(344|0)); //@line 258 "modules/graphics/opengl/GLBuffer.cpp"
   // unreachable; //@line 258 "modules/graphics/opengl/GLBuffer.cpp"
  }
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  ___cxa_free_exception(($2|0)); //@line 258 "modules/graphics/opengl/GLBuffer.cpp"
  $$2 = $5;$$24 = $6;
  ___resumeException($$2|0);
  // unreachable;
 }
 $7 = HEAP32[158429]|0; //@line 261 "modules/graphics/opengl/GLBuffer.cpp"
 $8 = ($7|0)==(0|0); //@line 261 "modules/graphics/opengl/GLBuffer.cpp"
 $9 = HEAP32[158426]|0; //@line 261 "modules/graphics/opengl/GLBuffer.cpp"
 $10 = ($9>>>0)<($size>>>0); //@line 261 "modules/graphics/opengl/GLBuffer.cpp"
 $or$cond = $8 | $10; //@line 261 "modules/graphics/opengl/GLBuffer.cpp"
 if (!($or$cond)) {
  $70 = HEAP32[158428]|0; //@line 308 "modules/graphics/opengl/GLBuffer.cpp"
  $71 = (($70) + 1)|0; //@line 308 "modules/graphics/opengl/GLBuffer.cpp"
  HEAP32[158428] = $71; //@line 308 "modules/graphics/opengl/GLBuffer.cpp"
  STACKTOP = sp;return; //@line 309 "modules/graphics/opengl/GLBuffer.cpp"
 }
 $11 = $size << 2; //@line 352 "modules/graphics/opengl/GLBuffer.cpp"
 $12 = ($11>>>0)>(65535); //@line 352 "modules/graphics/opengl/GLBuffer.cpp"
 $13 = $12&1; //@line 352 "modules/graphics/opengl/GLBuffer.cpp"
 $14 = (24748 + ($13<<2)|0); //@line 352 "modules/graphics/opengl/GLBuffer.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 352 "modules/graphics/opengl/GLBuffer.cpp"
 $16 = $12 ? 4 : 2; //@line 268 "modules/graphics/opengl/GLBuffer.cpp"
 $17 = ($size*6)|0; //@line 270 "modules/graphics/opengl/GLBuffer.cpp"
 $18 = Math_imul($17, $16)|0; //@line 270 "modules/graphics/opengl/GLBuffer.cpp"
 __THREW__ = 0;
 $19 = (invoke_ii(3348,40)|0); //@line 276 "modules/graphics/opengl/GLBuffer.cpp"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 do {
  if ($21) {
   $newbuffer$0 = 0;
   label = 10;
  } else {
   __THREW__ = 0;
   invoke_viiiiii(4042,($19|0),($18|0),(0|0),34963,35044,0); //@line 276 "modules/graphics/opengl/GLBuffer.cpp"
   $22 = __THREW__; __THREW__ = 0;
   $23 = $22&1;
   if ($23) {
    $31 = ___cxa_find_matching_catch_3(9056|0)|0;
    $32 = tempRet0;
    __ZdlPv($19); //@line 276 "modules/graphics/opengl/GLBuffer.cpp"
    $$0 = $31;$$02 = $32;$newbuffer$1 = 0;
    break;
   }
   __THREW__ = 0;
   $24 = (invoke_ii(3492,($18|0))|0); //@line 277 "modules/graphics/opengl/GLBuffer.cpp"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    $newbuffer$0 = $19;
    label = 10;
   } else {
    $27 = HEAP32[158429]|0; //@line 288 "modules/graphics/opengl/GLBuffer.cpp"
    $28 = ($27|0)==(0|0); //@line 288 "modules/graphics/opengl/GLBuffer.cpp"
    if (!($28)) {
     $65 = HEAP32[$27>>2]|0; //@line 288 "modules/graphics/opengl/GLBuffer.cpp"
     $66 = ((($65)) + 4|0); //@line 288 "modules/graphics/opengl/GLBuffer.cpp"
     $67 = HEAP32[$66>>2]|0; //@line 288 "modules/graphics/opengl/GLBuffer.cpp"
     FUNCTION_TABLE_vi[$67 & 8191]($27); //@line 288 "modules/graphics/opengl/GLBuffer.cpp"
    }
    HEAP32[158429] = $19; //@line 289 "modules/graphics/opengl/GLBuffer.cpp"
    $68 = HEAP32[158430]|0; //@line 291 "modules/graphics/opengl/GLBuffer.cpp"
    $69 = ($68|0)==(0|0); //@line 291 "modules/graphics/opengl/GLBuffer.cpp"
    if (!($69)) {
     __ZdaPv($68); //@line 291 "modules/graphics/opengl/GLBuffer.cpp"
    }
    HEAP32[158430] = $24; //@line 292 "modules/graphics/opengl/GLBuffer.cpp"
    HEAP32[158426] = $size; //@line 294 "modules/graphics/opengl/GLBuffer.cpp"
    HEAP32[158427] = $16; //@line 295 "modules/graphics/opengl/GLBuffer.cpp"
    switch ($15|0) {
    case 5123:  {
     __ZN4love8graphics6opengl11QuadIndices4fillItEEvv($this); //@line 300 "modules/graphics/opengl/GLBuffer.cpp"
     $70 = HEAP32[158428]|0; //@line 308 "modules/graphics/opengl/GLBuffer.cpp"
     $71 = (($70) + 1)|0; //@line 308 "modules/graphics/opengl/GLBuffer.cpp"
     HEAP32[158428] = $71; //@line 308 "modules/graphics/opengl/GLBuffer.cpp"
     STACKTOP = sp;return; //@line 309 "modules/graphics/opengl/GLBuffer.cpp"
     break;
    }
    case 5125:  {
     __ZN4love8graphics6opengl11QuadIndices4fillIjEEvv($this); //@line 303 "modules/graphics/opengl/GLBuffer.cpp"
     $70 = HEAP32[158428]|0; //@line 308 "modules/graphics/opengl/GLBuffer.cpp"
     $71 = (($70) + 1)|0; //@line 308 "modules/graphics/opengl/GLBuffer.cpp"
     HEAP32[158428] = $71; //@line 308 "modules/graphics/opengl/GLBuffer.cpp"
     STACKTOP = sp;return; //@line 309 "modules/graphics/opengl/GLBuffer.cpp"
     break;
    }
    default: {
     $70 = HEAP32[158428]|0; //@line 308 "modules/graphics/opengl/GLBuffer.cpp"
     $71 = (($70) + 1)|0; //@line 308 "modules/graphics/opengl/GLBuffer.cpp"
     HEAP32[158428] = $71; //@line 308 "modules/graphics/opengl/GLBuffer.cpp"
     STACKTOP = sp;return; //@line 309 "modules/graphics/opengl/GLBuffer.cpp"
    }
    }
   }
  }
 } while(0);
 if ((label|0) == 10) {
  $29 = ___cxa_find_matching_catch_3(9056|0)|0;
  $30 = tempRet0;
  $$0 = $29;$$02 = $30;$newbuffer$1 = $newbuffer$0;
 }
 $33 = (_llvm_eh_typeid_for((9056|0))|0); //@line 278 "modules/graphics/opengl/GLBuffer.cpp"
 $34 = ($$02|0)==($33|0); //@line 278 "modules/graphics/opengl/GLBuffer.cpp"
 if (!($34)) {
  $$2 = $$0;$$24 = $$02;
  ___resumeException($$2|0);
  // unreachable;
 }
 (___cxa_begin_catch(($$0|0))|0); //@line 278 "modules/graphics/opengl/GLBuffer.cpp"
 $35 = ($newbuffer$1|0)==(0|0); //@line 281 "modules/graphics/opengl/GLBuffer.cpp"
 if (!($35)) {
  $36 = ((($newbuffer$1)) + -4|0); //@line 281 "modules/graphics/opengl/GLBuffer.cpp"
  $37 = HEAP32[$36>>2]|0; //@line 281 "modules/graphics/opengl/GLBuffer.cpp"
  $38 = ($37|0)==(0); //@line 281 "modules/graphics/opengl/GLBuffer.cpp"
  L37: do {
   if (!($38)) {
    $39 = (($newbuffer$1) + (($37*40)|0)|0); //@line 281 "modules/graphics/opengl/GLBuffer.cpp"
    $41 = $39;
    while(1) {
     $40 = ((($41)) + -40|0); //@line 281 "modules/graphics/opengl/GLBuffer.cpp"
     HEAP32[$40>>2] = (24732); //@line 71 "modules/graphics/opengl/GLBuffer.cpp"
     $42 = ((($41)) + -20|0); //@line 72 "modules/graphics/opengl/GLBuffer.cpp"
     $43 = HEAP32[$42>>2]|0; //@line 72 "modules/graphics/opengl/GLBuffer.cpp"
     $44 = ($43|0)==(0); //@line 72 "modules/graphics/opengl/GLBuffer.cpp"
     if (!($44)) {
      $45 = ((($41)) + -35|0); //@line 235 "modules/graphics/opengl/GLBuffer.cpp"
      HEAP8[$45>>0] = 0; //@line 235 "modules/graphics/opengl/GLBuffer.cpp"
      $46 = HEAP32[165381]|0; //@line 487 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_vii($46|0,1,($42|0)); //@line 487 "./libraries/glad/gladfuncs.hpp"
      $47 = __THREW__; __THREW__ = 0;
      $48 = $47&1;
      if ($48) {
       $$lcssa = $40;
       break;
      }
      HEAP32[$42>>2] = 0; //@line 238 "modules/graphics/opengl/GLBuffer.cpp"
     }
     $51 = ((($41)) + -16|0); //@line 75 "modules/graphics/opengl/GLBuffer.cpp"
     $52 = HEAP32[$51>>2]|0; //@line 75 "modules/graphics/opengl/GLBuffer.cpp"
     $53 = ($52|0)==(0|0); //@line 75 "modules/graphics/opengl/GLBuffer.cpp"
     if (!($53)) {
      __ZdaPv($52); //@line 75 "modules/graphics/opengl/GLBuffer.cpp"
     }
     __ZN4love8graphics8VolatileD2Ev($40); //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
     $54 = ($40|0)==($newbuffer$1|0); //@line 281 "modules/graphics/opengl/GLBuffer.cpp"
     if ($54) {
      break L37;
     } else {
      $41 = $40;
     }
    }
    $49 = ___cxa_find_matching_catch_3(0|0)|0;
    $50 = tempRet0;
    __ZN4love8graphics8VolatileD2Ev($$lcssa); //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
    ___clang_call_terminate($49); //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
    // unreachable; //@line 76 "modules/graphics/opengl/GLBuffer.cpp"
   }
  } while(0);
  __ZdaPv($36); //@line 281 "modules/graphics/opengl/GLBuffer.cpp"
 }
 $55 = (___cxa_allocate_exception(16)|0); //@line 283 "modules/graphics/opengl/GLBuffer.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($55|0),(507610|0),($vararg_buffer1|0)); //@line 283 "modules/graphics/opengl/GLBuffer.cpp"
 $56 = __THREW__; __THREW__ = 0;
 $57 = $56&1;
 if ($57) {
  $61 = ___cxa_find_matching_catch_2()|0;
  $62 = tempRet0;
  ___cxa_free_exception(($55|0)); //@line 283 "modules/graphics/opengl/GLBuffer.cpp"
  $$1 = $61;$$13 = $62;
 } else {
  __THREW__ = 0;
  invoke_viii(3350,($55|0),(96|0),(344|0)); //@line 283 "modules/graphics/opengl/GLBuffer.cpp"
  $58 = __THREW__; __THREW__ = 0;
  $59 = ___cxa_find_matching_catch_2()|0;
  $60 = tempRet0;
  $$1 = $59;$$13 = $60;
 }
 __THREW__ = 0;
 invoke_v(3378); //@line 284 "modules/graphics/opengl/GLBuffer.cpp"
 $63 = __THREW__; __THREW__ = 0;
 $64 = $63&1;
 if ($64) {
  $72 = ___cxa_find_matching_catch_3(0|0)|0;
  $73 = tempRet0;
  ___clang_call_terminate($72); //@line 284 "modules/graphics/opengl/GLBuffer.cpp"
  // unreachable; //@line 284 "modules/graphics/opengl/GLBuffer.cpp"
 } else {
  $$2 = $$1;$$24 = $$13;
  ___resumeException($$2|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl11QuadIndices4fillItEEvv($this) {
 $this = $this|0;
 var $$pre = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $9 = 0, $i$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[158430]|0; //@line 379 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = HEAP32[158426]|0; //@line 385 "modules/graphics/opengl/GLBuffer.cpp"
 $2 = ($1|0)==(0); //@line 385 "modules/graphics/opengl/GLBuffer.cpp"
 if (!($2)) {
  $i$02 = 0;
  while(1) {
   $38 = $i$02 << 2; //@line 387 "modules/graphics/opengl/GLBuffer.cpp"
   $39 = $38&65535; //@line 387 "modules/graphics/opengl/GLBuffer.cpp"
   $40 = ($i$02*6)|0; //@line 387 "modules/graphics/opengl/GLBuffer.cpp"
   $41 = (($0) + ($40<<1)|0); //@line 387 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP16[$41>>1] = $39; //@line 387 "modules/graphics/opengl/GLBuffer.cpp"
   $42 = $38 | 1; //@line 388 "modules/graphics/opengl/GLBuffer.cpp"
   $43 = $42&65535; //@line 388 "modules/graphics/opengl/GLBuffer.cpp"
   $44 = $40 | 1; //@line 388 "modules/graphics/opengl/GLBuffer.cpp"
   $45 = (($0) + ($44<<1)|0); //@line 388 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP16[$45>>1] = $43; //@line 388 "modules/graphics/opengl/GLBuffer.cpp"
   $46 = $38 | 2; //@line 389 "modules/graphics/opengl/GLBuffer.cpp"
   $47 = $46&65535; //@line 389 "modules/graphics/opengl/GLBuffer.cpp"
   $48 = (($40) + 2)|0; //@line 389 "modules/graphics/opengl/GLBuffer.cpp"
   $49 = (($0) + ($48<<1)|0); //@line 389 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP16[$49>>1] = $47; //@line 389 "modules/graphics/opengl/GLBuffer.cpp"
   $50 = (($40) + 3)|0; //@line 391 "modules/graphics/opengl/GLBuffer.cpp"
   $51 = (($0) + ($50<<1)|0); //@line 391 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP16[$51>>1] = $47; //@line 391 "modules/graphics/opengl/GLBuffer.cpp"
   $52 = (($40) + 4)|0; //@line 392 "modules/graphics/opengl/GLBuffer.cpp"
   $53 = (($0) + ($52<<1)|0); //@line 392 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP16[$53>>1] = $43; //@line 392 "modules/graphics/opengl/GLBuffer.cpp"
   $54 = $38 | 3; //@line 393 "modules/graphics/opengl/GLBuffer.cpp"
   $55 = $54&65535; //@line 393 "modules/graphics/opengl/GLBuffer.cpp"
   $56 = (($40) + 5)|0; //@line 393 "modules/graphics/opengl/GLBuffer.cpp"
   $57 = (($0) + ($56<<1)|0); //@line 393 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP16[$57>>1] = $55; //@line 393 "modules/graphics/opengl/GLBuffer.cpp"
   $58 = (($i$02) + 1)|0; //@line 385 "modules/graphics/opengl/GLBuffer.cpp"
   $59 = ($58>>>0)<($1>>>0); //@line 385 "modules/graphics/opengl/GLBuffer.cpp"
   if ($59) {
    $i$02 = $58;
   } else {
    break;
   }
  }
 }
 $3 = HEAP32[158429]|0; //@line 396 "modules/graphics/opengl/GLBuffer.cpp"
 $4 = ((($3)) + 5|0); //@line 178 "modules/graphics/opengl/GLBuffer.cpp"
 $5 = HEAP8[$4>>0]|0; //@line 178 "modules/graphics/opengl/GLBuffer.cpp"
 $6 = ($5<<24>>24)==(0); //@line 178 "modules/graphics/opengl/GLBuffer.cpp"
 if ($6) {
  $7 = ((($3)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
  $8 = HEAP32[$7>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
  $9 = ((($3)) + 20|0); //@line 180 "modules/graphics/opengl/GLBuffer.cpp"
  $10 = HEAP32[$9>>2]|0; //@line 180 "modules/graphics/opengl/GLBuffer.cpp"
  $11 = HEAP32[165557]|0; //@line 486 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$11 & 8191]($8,$10); //@line 486 "./libraries/glad/gladfuncs.hpp"
  $12 = ((($3)) + 4|0); //@line 181 "modules/graphics/opengl/GLBuffer.cpp"
  HEAP8[$12>>0] = 1; //@line 181 "modules/graphics/opengl/GLBuffer.cpp"
  $$pre = HEAP32[158429]|0; //@line 397 "modules/graphics/opengl/GLBuffer.cpp"
  $$pre3 = HEAP32[158430]|0; //@line 397 "modules/graphics/opengl/GLBuffer.cpp"
  $14 = $$pre;$18 = $$pre3;
 } else {
  $14 = $3;$18 = $0;
 }
 $13 = ((($14)) + 8|0); //@line 76 "modules/graphics/opengl/GLBuffer.h"
 $15 = HEAP32[$13>>2]|0; //@line 76 "modules/graphics/opengl/GLBuffer.h"
 $16 = ((($14)) + 24|0); //@line 195 "modules/graphics/opengl/GLBuffer.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 195 "modules/graphics/opengl/GLBuffer.cpp"
 _memcpy(($17|0),($18|0),($15|0))|0; //@line 195 "modules/graphics/opengl/GLBuffer.cpp"
 $19 = ((($14)) + 5|0); //@line 197 "modules/graphics/opengl/GLBuffer.cpp"
 $20 = HEAP8[$19>>0]|0; //@line 197 "modules/graphics/opengl/GLBuffer.cpp"
 $21 = ($20<<24>>24)==(0); //@line 197 "modules/graphics/opengl/GLBuffer.cpp"
 if ($21) {
  $33 = ((($14)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
  $34 = HEAP32[$33>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
  $35 = HEAP32[165513]|0; //@line 491 "./libraries/glad/gladfuncs.hpp"
  __THREW__ = 0;
  invoke_viiii($35|0,($34|0),0,($15|0),($18|0)); //@line 491 "./libraries/glad/gladfuncs.hpp"
  $36 = __THREW__; __THREW__ = 0;
  $37 = $36&1;
  if ($37) {
   $70 = ___cxa_find_matching_catch_2()|0;
   $71 = tempRet0;
   $72 = ((($3)) + 4|0); //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
   $73 = HEAP8[$72>>0]|0; //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
   $74 = ($73<<24>>24)==(0); //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
   if ($74) {
    HEAP8[$72>>0] = 0; //@line 190 "modules/graphics/opengl/GLBuffer.cpp"
    ___resumeException($70|0);
    // unreachable;
   }
   $75 = ((($3)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
   $76 = HEAP32[$75>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
   $77 = HEAP32[165557]|0; //@line 486 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   invoke_vii($77|0,($76|0),0); //@line 486 "./libraries/glad/gladfuncs.hpp"
   $78 = __THREW__; __THREW__ = 0;
   $79 = $78&1;
   if ($79) {
    $80 = ___cxa_find_matching_catch_3(0|0)|0;
    $81 = tempRet0;
    ___clang_call_terminate($80); //@line 196 "modules/graphics/opengl/GLBuffer.h"
    // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
   } else {
    HEAP8[$72>>0] = 0; //@line 190 "modules/graphics/opengl/GLBuffer.cpp"
    ___resumeException($70|0);
    // unreachable;
   }
  }
 } else {
  $22 = ((($14)) + 36|0); //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
  $23 = HEAP32[$22>>2]|0; //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
  $24 = $23 & 1; //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
  $25 = ($24|0)==(0); //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
  if (!($25)) {
   $26 = ((($14)) + 28|0); //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
   $27 = HEAP32[$26>>2]|0; //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
   $28 = ((($14)) + 32|0); //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
   $29 = HEAP32[$28>>2]|0; //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
   $30 = (($29) + ($27))|0; //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP32[$26>>2] = 0; //@line 170 "modules/graphics/opengl/GLBuffer.cpp"
   $31 = ($15>>>0)<($30>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $32 = $31 ? $30 : $15; //@line 172 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP32[$28>>2] = $32; //@line 173 "modules/graphics/opengl/GLBuffer.cpp"
  }
 }
 $60 = ((($3)) + 4|0); //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
 $61 = HEAP8[$60>>0]|0; //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
 $62 = ($61<<24>>24)==(0); //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
 if ($62) {
  HEAP8[$60>>0] = 0; //@line 190 "modules/graphics/opengl/GLBuffer.cpp"
  return; //@line 398 "modules/graphics/opengl/GLBuffer.cpp"
 }
 $63 = ((($3)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
 $64 = HEAP32[$63>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
 $65 = HEAP32[165557]|0; //@line 486 "./libraries/glad/gladfuncs.hpp"
 __THREW__ = 0;
 invoke_vii($65|0,($64|0),0); //@line 486 "./libraries/glad/gladfuncs.hpp"
 $66 = __THREW__; __THREW__ = 0;
 $67 = $66&1;
 if ($67) {
  $68 = ___cxa_find_matching_catch_3(0|0)|0;
  $69 = tempRet0;
  ___clang_call_terminate($68); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
 } else {
  HEAP8[$60>>0] = 0; //@line 190 "modules/graphics/opengl/GLBuffer.cpp"
  return; //@line 398 "modules/graphics/opengl/GLBuffer.cpp"
 }
}
function __ZN4love8graphics6opengl11QuadIndices4fillIjEEvv($this) {
 $this = $this|0;
 var $$pre = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0;
 var $9 = 0, $i$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[158430]|0; //@line 379 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = HEAP32[158426]|0; //@line 385 "modules/graphics/opengl/GLBuffer.cpp"
 $2 = ($1|0)==(0); //@line 385 "modules/graphics/opengl/GLBuffer.cpp"
 if (!($2)) {
  $i$02 = 0;
  while(1) {
   $38 = $i$02 << 2; //@line 387 "modules/graphics/opengl/GLBuffer.cpp"
   $39 = ($i$02*6)|0; //@line 387 "modules/graphics/opengl/GLBuffer.cpp"
   $40 = (($0) + ($39<<2)|0); //@line 387 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP32[$40>>2] = $38; //@line 387 "modules/graphics/opengl/GLBuffer.cpp"
   $41 = $38 | 1; //@line 388 "modules/graphics/opengl/GLBuffer.cpp"
   $42 = $39 | 1; //@line 388 "modules/graphics/opengl/GLBuffer.cpp"
   $43 = (($0) + ($42<<2)|0); //@line 388 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP32[$43>>2] = $41; //@line 388 "modules/graphics/opengl/GLBuffer.cpp"
   $44 = $38 | 2; //@line 389 "modules/graphics/opengl/GLBuffer.cpp"
   $45 = (($39) + 2)|0; //@line 389 "modules/graphics/opengl/GLBuffer.cpp"
   $46 = (($0) + ($45<<2)|0); //@line 389 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP32[$46>>2] = $44; //@line 389 "modules/graphics/opengl/GLBuffer.cpp"
   $47 = (($39) + 3)|0; //@line 391 "modules/graphics/opengl/GLBuffer.cpp"
   $48 = (($0) + ($47<<2)|0); //@line 391 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP32[$48>>2] = $44; //@line 391 "modules/graphics/opengl/GLBuffer.cpp"
   $49 = (($39) + 4)|0; //@line 392 "modules/graphics/opengl/GLBuffer.cpp"
   $50 = (($0) + ($49<<2)|0); //@line 392 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP32[$50>>2] = $41; //@line 392 "modules/graphics/opengl/GLBuffer.cpp"
   $51 = $38 | 3; //@line 393 "modules/graphics/opengl/GLBuffer.cpp"
   $52 = (($39) + 5)|0; //@line 393 "modules/graphics/opengl/GLBuffer.cpp"
   $53 = (($0) + ($52<<2)|0); //@line 393 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP32[$53>>2] = $51; //@line 393 "modules/graphics/opengl/GLBuffer.cpp"
   $54 = (($i$02) + 1)|0; //@line 385 "modules/graphics/opengl/GLBuffer.cpp"
   $55 = HEAP32[158426]|0; //@line 385 "modules/graphics/opengl/GLBuffer.cpp"
   $56 = ($54>>>0)<($55>>>0); //@line 385 "modules/graphics/opengl/GLBuffer.cpp"
   if ($56) {
    $i$02 = $54;
   } else {
    break;
   }
  }
 }
 $3 = HEAP32[158429]|0; //@line 396 "modules/graphics/opengl/GLBuffer.cpp"
 $4 = ((($3)) + 5|0); //@line 178 "modules/graphics/opengl/GLBuffer.cpp"
 $5 = HEAP8[$4>>0]|0; //@line 178 "modules/graphics/opengl/GLBuffer.cpp"
 $6 = ($5<<24>>24)==(0); //@line 178 "modules/graphics/opengl/GLBuffer.cpp"
 if ($6) {
  $7 = ((($3)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
  $8 = HEAP32[$7>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
  $9 = ((($3)) + 20|0); //@line 180 "modules/graphics/opengl/GLBuffer.cpp"
  $10 = HEAP32[$9>>2]|0; //@line 180 "modules/graphics/opengl/GLBuffer.cpp"
  $11 = HEAP32[165557]|0; //@line 486 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$11 & 8191]($8,$10); //@line 486 "./libraries/glad/gladfuncs.hpp"
  $12 = ((($3)) + 4|0); //@line 181 "modules/graphics/opengl/GLBuffer.cpp"
  HEAP8[$12>>0] = 1; //@line 181 "modules/graphics/opengl/GLBuffer.cpp"
  $$pre = HEAP32[158429]|0; //@line 397 "modules/graphics/opengl/GLBuffer.cpp"
  $$pre3 = HEAP32[158430]|0; //@line 397 "modules/graphics/opengl/GLBuffer.cpp"
  $14 = $$pre;$18 = $$pre3;
 } else {
  $14 = $3;$18 = $0;
 }
 $13 = ((($14)) + 8|0); //@line 76 "modules/graphics/opengl/GLBuffer.h"
 $15 = HEAP32[$13>>2]|0; //@line 76 "modules/graphics/opengl/GLBuffer.h"
 $16 = ((($14)) + 24|0); //@line 195 "modules/graphics/opengl/GLBuffer.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 195 "modules/graphics/opengl/GLBuffer.cpp"
 _memcpy(($17|0),($18|0),($15|0))|0; //@line 195 "modules/graphics/opengl/GLBuffer.cpp"
 $19 = ((($14)) + 5|0); //@line 197 "modules/graphics/opengl/GLBuffer.cpp"
 $20 = HEAP8[$19>>0]|0; //@line 197 "modules/graphics/opengl/GLBuffer.cpp"
 $21 = ($20<<24>>24)==(0); //@line 197 "modules/graphics/opengl/GLBuffer.cpp"
 if ($21) {
  $33 = ((($14)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
  $34 = HEAP32[$33>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
  $35 = HEAP32[165513]|0; //@line 491 "./libraries/glad/gladfuncs.hpp"
  __THREW__ = 0;
  invoke_viiii($35|0,($34|0),0,($15|0),($18|0)); //@line 491 "./libraries/glad/gladfuncs.hpp"
  $36 = __THREW__; __THREW__ = 0;
  $37 = $36&1;
  if ($37) {
   $67 = ___cxa_find_matching_catch_2()|0;
   $68 = tempRet0;
   $69 = ((($3)) + 4|0); //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
   $70 = HEAP8[$69>>0]|0; //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
   $71 = ($70<<24>>24)==(0); //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
   if ($71) {
    HEAP8[$69>>0] = 0; //@line 190 "modules/graphics/opengl/GLBuffer.cpp"
    ___resumeException($67|0);
    // unreachable;
   }
   $72 = ((($3)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
   $73 = HEAP32[$72>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
   $74 = HEAP32[165557]|0; //@line 486 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   invoke_vii($74|0,($73|0),0); //@line 486 "./libraries/glad/gladfuncs.hpp"
   $75 = __THREW__; __THREW__ = 0;
   $76 = $75&1;
   if ($76) {
    $77 = ___cxa_find_matching_catch_3(0|0)|0;
    $78 = tempRet0;
    ___clang_call_terminate($77); //@line 196 "modules/graphics/opengl/GLBuffer.h"
    // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
   } else {
    HEAP8[$69>>0] = 0; //@line 190 "modules/graphics/opengl/GLBuffer.cpp"
    ___resumeException($67|0);
    // unreachable;
   }
  }
 } else {
  $22 = ((($14)) + 36|0); //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
  $23 = HEAP32[$22>>2]|0; //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
  $24 = $23 & 1; //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
  $25 = ($24|0)==(0); //@line 162 "modules/graphics/opengl/GLBuffer.cpp"
  if (!($25)) {
   $26 = ((($14)) + 28|0); //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
   $27 = HEAP32[$26>>2]|0; //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
   $28 = ((($14)) + 32|0); //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
   $29 = HEAP32[$28>>2]|0; //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
   $30 = (($29) + ($27))|0; //@line 169 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP32[$26>>2] = 0; //@line 170 "modules/graphics/opengl/GLBuffer.cpp"
   $31 = ($15>>>0)<($30>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $32 = $31 ? $30 : $15; //@line 172 "modules/graphics/opengl/GLBuffer.cpp"
   HEAP32[$28>>2] = $32; //@line 173 "modules/graphics/opengl/GLBuffer.cpp"
  }
 }
 $57 = ((($3)) + 4|0); //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
 $58 = HEAP8[$57>>0]|0; //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
 $59 = ($58<<24>>24)==(0); //@line 187 "modules/graphics/opengl/GLBuffer.cpp"
 if ($59) {
  HEAP8[$57>>0] = 0; //@line 190 "modules/graphics/opengl/GLBuffer.cpp"
  return; //@line 398 "modules/graphics/opengl/GLBuffer.cpp"
 }
 $60 = ((($3)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
 $61 = HEAP32[$60>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
 $62 = HEAP32[165557]|0; //@line 486 "./libraries/glad/gladfuncs.hpp"
 __THREW__ = 0;
 invoke_vii($62|0,($61|0),0); //@line 486 "./libraries/glad/gladfuncs.hpp"
 $63 = __THREW__; __THREW__ = 0;
 $64 = $63&1;
 if ($64) {
  $65 = ___cxa_find_matching_catch_3(0|0)|0;
  $66 = tempRet0;
  ___clang_call_terminate($65); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
 } else {
  HEAP8[$57>>0] = 0; //@line 190 "modules/graphics/opengl/GLBuffer.cpp"
  return; //@line 398 "modules/graphics/opengl/GLBuffer.cpp"
 }
}
function __ZNK4love8graphics6opengl11QuadIndices7getTypeEj($this,$s) {
 $this = $this|0;
 $s = $s|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $s << 2; //@line 352 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = ($0>>>0)>(65535); //@line 352 "modules/graphics/opengl/GLBuffer.cpp"
 $2 = $1&1; //@line 352 "modules/graphics/opengl/GLBuffer.cpp"
 $3 = (24748 + ($2<<2)|0); //@line 352 "modules/graphics/opengl/GLBuffer.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 352 "modules/graphics/opengl/GLBuffer.cpp"
 return ($4|0); //@line 352 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZN4love8graphics6opengl11QuadIndicesC2ERKS2_($this,$other) {
 $this = $this|0;
 $other = $other|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$other>>2]|0; //@line 312 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP32[$this>>2] = $0; //@line 312 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = HEAP32[158428]|0; //@line 314 "modules/graphics/opengl/GLBuffer.cpp"
 $2 = (($1) + 1)|0; //@line 314 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP32[158428] = $2; //@line 314 "modules/graphics/opengl/GLBuffer.cpp"
 return; //@line 315 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZN4love8graphics6opengl11QuadIndicesaSERKS2_($this,$other) {
 $this = $this|0;
 $other = $other|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$other>>2]|0; //@line 319 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP32[$this>>2] = $0; //@line 319 "modules/graphics/opengl/GLBuffer.cpp"
 return ($this|0); //@line 320 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZN4love8graphics6opengl11QuadIndicesD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[158428]|0; //@line 325 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = (($0) + -1)|0; //@line 325 "modules/graphics/opengl/GLBuffer.cpp"
 HEAP32[158428] = $1; //@line 325 "modules/graphics/opengl/GLBuffer.cpp"
 $2 = ($1|0)==(0); //@line 328 "modules/graphics/opengl/GLBuffer.cpp"
 if (!($2)) {
  return; //@line 336 "modules/graphics/opengl/GLBuffer.cpp"
 }
 $3 = HEAP32[158429]|0; //@line 330 "modules/graphics/opengl/GLBuffer.cpp"
 $4 = ($3|0)==(0|0); //@line 330 "modules/graphics/opengl/GLBuffer.cpp"
 if (!($4)) {
  $5 = HEAP32[$3>>2]|0; //@line 330 "modules/graphics/opengl/GLBuffer.cpp"
  $6 = ((($5)) + 4|0); //@line 330 "modules/graphics/opengl/GLBuffer.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 330 "modules/graphics/opengl/GLBuffer.cpp"
  FUNCTION_TABLE_vi[$7 & 8191]($3); //@line 330 "modules/graphics/opengl/GLBuffer.cpp"
 }
 HEAP32[158429] = 0; //@line 331 "modules/graphics/opengl/GLBuffer.cpp"
 $8 = HEAP32[158430]|0; //@line 333 "modules/graphics/opengl/GLBuffer.cpp"
 $9 = ($8|0)==(0|0); //@line 333 "modules/graphics/opengl/GLBuffer.cpp"
 if (!($9)) {
  __ZdaPv($8); //@line 333 "modules/graphics/opengl/GLBuffer.cpp"
 }
 HEAP32[158430] = 0; //@line 334 "modules/graphics/opengl/GLBuffer.cpp"
 return; //@line 336 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZNK4love8graphics6opengl11QuadIndices7getSizeEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 340 "modules/graphics/opengl/GLBuffer.cpp"
 return ($0|0); //@line 340 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZNK4love8graphics6opengl11QuadIndices13getIndexCountEj($this,$elements) {
 $this = $this|0;
 $elements = $elements|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($elements*6)|0; //@line 345 "modules/graphics/opengl/GLBuffer.cpp"
 return ($0|0); //@line 345 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZN4love8graphics6opengl11QuadIndices14getElementSizeEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[158427]|0; //@line 358 "modules/graphics/opengl/GLBuffer.cpp"
 return ($0|0); //@line 358 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZNK4love8graphics6opengl11QuadIndices9getBufferEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[158429]|0; //@line 363 "modules/graphics/opengl/GLBuffer.cpp"
 return ($0|0); //@line 363 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZNK4love8graphics6opengl11QuadIndices10getPointerEj($this,$offset) {
 $this = $this|0;
 $offset = $offset|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (0 + ($offset)|0); //@line 205 "modules/graphics/opengl/GLBuffer.cpp"
 return ($0|0); //@line 368 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZNK4love8graphics6opengl11QuadIndices10getIndicesEj($this,$offset) {
 $this = $this|0;
 $offset = $offset|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[158430]|0; //@line 373 "modules/graphics/opengl/GLBuffer.cpp"
 $1 = (($0) + ($offset)|0); //@line 373 "modules/graphics/opengl/GLBuffer.cpp"
 return ($1|0); //@line 373 "modules/graphics/opengl/GLBuffer.cpp"
}
function __ZN4love8graphics6opengl8GraphicsD2Ev($this) {
 $this = $this|0;
 var $$pre$i$i$i$i$i$i$i$i = 0, $$pre$i$i$i$i$i$i$i$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $scevgep$i$i$i = 0, $scevgep$i$i$i1 = 0, $scevgep$i$i$i4 = 0, $scevgep$i$i$i9 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i10 = 0, $scevgep4$i$i$i2 = 0, $scevgep4$i$i$i5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (24764); //@line 79 "modules/graphics/opengl/Graphics.cpp"
 $0 = ((($this)) + 44|0); //@line 81 "modules/graphics/opengl/Graphics.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($this)) + 48|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($3|0)==($1|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 L1: do {
  if (!($4)) {
   $6 = $3;
   L2: while(1) {
    $5 = ((($6)) + -136|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$2>>2] = $5; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $7 = ((($6)) + -44|0); //@line 482 "modules/graphics/opengl/Graphics.h"
    $8 = HEAP32[$7>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $9 = ($8|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($9)) {
     $10 = ((($6)) + -40|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $11 = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $12 = ($11|0)==($8|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($12) {
      $26 = $8;
     } else {
      $14 = $11;
      while(1) {
       $13 = ((($14)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$10>>2] = $13; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $15 = HEAP32[$13>>2]|0; //@line 106 "./common/Object.h"
       $16 = ($15|0)==(0|0); //@line 106 "./common/Object.h"
       if ($16) {
        $18 = $13;
       } else {
        $19 = HEAP32[$15>>2]|0; //@line 106 "./common/Object.h"
        $20 = ((($19)) + 12|0); //@line 106 "./common/Object.h"
        $21 = HEAP32[$20>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($21|0,($15|0)); //@line 106 "./common/Object.h"
        $22 = __THREW__; __THREW__ = 0;
        $23 = $22&1;
        if ($23) {
         label = 8;
         break L2;
        }
        $$pre$i$i$i$i$i$i$i$i$i$i = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $18 = $$pre$i$i$i$i$i$i$i$i$i$i;
       }
       $17 = ($18|0)==($8|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($17) {
        break;
       } else {
        $14 = $18;
       }
      }
      $$pre$i$i$i$i$i$i$i$i = HEAP32[$7>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $26 = $$pre$i$i$i$i$i$i$i$i;
     }
     __ZdlPv($26); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    }
    $27 = ((($6)) + -48|0); //@line 106 "./common/Object.h"
    $28 = HEAP32[$27>>2]|0; //@line 106 "./common/Object.h"
    $29 = ($28|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($29)) {
     $30 = HEAP32[$28>>2]|0; //@line 106 "./common/Object.h"
     $31 = ((($30)) + 12|0); //@line 106 "./common/Object.h"
     $32 = HEAP32[$31>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($32|0,($28|0)); //@line 106 "./common/Object.h"
     $33 = __THREW__; __THREW__ = 0;
     $34 = $33&1;
     if ($34) {
      label = 13;
      break;
     }
    }
    $37 = ((($6)) + -52|0); //@line 106 "./common/Object.h"
    $38 = HEAP32[$37>>2]|0; //@line 106 "./common/Object.h"
    $39 = ($38|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($39)) {
     $42 = HEAP32[$38>>2]|0; //@line 106 "./common/Object.h"
     $43 = ((($42)) + 12|0); //@line 106 "./common/Object.h"
     $44 = HEAP32[$43>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($44|0,($38|0)); //@line 106 "./common/Object.h"
     $45 = __THREW__; __THREW__ = 0;
     $46 = $45&1;
     if ($46) {
      label = 17;
      break;
     }
    }
    $40 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $41 = ($40|0)==($1|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($41) {
     break L1;
    } else {
     $6 = $40;
    }
   }
   if ((label|0) == 8) {
    $24 = ___cxa_find_matching_catch_3(0|0)|0;
    $25 = tempRet0;
    ___clang_call_terminate($24); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 13) {
    $35 = ___cxa_find_matching_catch_3(0|0)|0;
    $36 = tempRet0;
    ___clang_call_terminate($35); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 17) {
    $47 = ___cxa_find_matching_catch_3(0|0)|0;
    $48 = tempRet0;
    ___clang_call_terminate($47); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
  }
 } while(0);
 $49 = ((($this)) + 12|0); //@line 133 "./common/Object.h"
 $50 = HEAP32[$49>>2]|0; //@line 133 "./common/Object.h"
 $51 = ($50|0)==(0|0); //@line 133 "./common/Object.h"
 if ($51) {
  label = 20;
 } else {
  $52 = HEAP32[$50>>2]|0; //@line 133 "./common/Object.h"
  $53 = ((($52)) + 12|0); //@line 133 "./common/Object.h"
  $54 = HEAP32[$53>>2]|0; //@line 133 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($54|0,($50|0)); //@line 133 "./common/Object.h"
  $55 = __THREW__; __THREW__ = 0;
  $56 = $55&1;
  if (!($56)) {
   label = 20;
  }
 }
 do {
  if ((label|0) == 20) {
   HEAP32[$49>>2] = 0; //@line 134 "./common/Object.h"
   $57 = HEAP32[158582]|0; //@line 84 "modules/graphics/opengl/Graphics.cpp"
   $58 = ($57|0)==(0|0); //@line 84 "modules/graphics/opengl/Graphics.cpp"
   if (!($58)) {
    $59 = HEAP32[$57>>2]|0; //@line 86 "modules/graphics/opengl/Graphics.cpp"
    $60 = ((($59)) + 12|0); //@line 86 "modules/graphics/opengl/Graphics.cpp"
    $61 = HEAP32[$60>>2]|0; //@line 86 "modules/graphics/opengl/Graphics.cpp"
    __THREW__ = 0;
    invoke_vi($61|0,($57|0)); //@line 86 "modules/graphics/opengl/Graphics.cpp"
    $62 = __THREW__; __THREW__ = 0;
    $63 = $62&1;
    if ($63) {
     break;
    }
    HEAP32[158582] = 0; //@line 87 "modules/graphics/opengl/Graphics.cpp"
   }
   $107 = HEAP32[158583]|0; //@line 89 "modules/graphics/opengl/Graphics.cpp"
   $108 = ($107|0)==(0|0); //@line 89 "modules/graphics/opengl/Graphics.cpp"
   if (!($108)) {
    $109 = HEAP32[$107>>2]|0; //@line 91 "modules/graphics/opengl/Graphics.cpp"
    $110 = ((($109)) + 12|0); //@line 91 "modules/graphics/opengl/Graphics.cpp"
    $111 = HEAP32[$110>>2]|0; //@line 91 "modules/graphics/opengl/Graphics.cpp"
    __THREW__ = 0;
    invoke_vi($111|0,($107|0)); //@line 91 "modules/graphics/opengl/Graphics.cpp"
    $112 = __THREW__; __THREW__ = 0;
    $113 = $112&1;
    if ($113) {
     break;
    }
    HEAP32[158583] = 0; //@line 92 "modules/graphics/opengl/Graphics.cpp"
   }
   $114 = ((($this)) + 28|0); //@line 95 "modules/graphics/opengl/Graphics.cpp"
   $115 = HEAP32[$114>>2]|0; //@line 95 "modules/graphics/opengl/Graphics.cpp"
   $116 = ($115|0)==(0|0); //@line 95 "modules/graphics/opengl/Graphics.cpp"
   if (!($116)) {
    __ZN4love8graphics6opengl11QuadIndicesD2Ev($115); //@line 96 "modules/graphics/opengl/Graphics.cpp"
    __ZdlPv($115); //@line 96 "modules/graphics/opengl/Graphics.cpp"
   }
   $117 = ((($this)) + 56|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $118 = HEAP32[$117>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $119 = ($118|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $120 = $118; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($119)) {
    $121 = ((($this)) + 60|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $122 = HEAP32[$121>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $123 = ($122|0)==($118|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($123)) {
     $scevgep$i$i$i4 = ((($122)) + -4|0);
     $124 = $scevgep$i$i$i4;
     $125 = (($124) - ($120))|0;
     $126 = $125 >>> 2;
     $127 = $126 ^ -1;
     $scevgep4$i$i$i5 = (($122) + ($127<<2)|0);
     HEAP32[$121>>2] = $scevgep4$i$i$i5; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($118); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   __ZNSt3__113__vector_baseIN4love8graphics6opengl8Graphics12DisplayStateENS_9allocatorIS5_EEED2Ev($0); //@line 97 "modules/graphics/opengl/Graphics.cpp"
   $128 = ((($this)) + 16|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $129 = HEAP32[$128>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $130 = ($129|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $131 = $129; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($130)) {
    $132 = ((($this)) + 20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $133 = HEAP32[$132>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $134 = ($133|0)==($129|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($134)) {
     $scevgep$i$i$i9 = ((($133)) + -8|0);
     $135 = $scevgep$i$i$i9;
     $136 = (($135) - ($131))|0;
     $137 = $136 >>> 3;
     $138 = $137 ^ -1;
     $scevgep4$i$i$i10 = (($133) + ($138<<3)|0);
     HEAP32[$132>>2] = $scevgep4$i$i$i10; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($129); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $139 = HEAP32[$49>>2]|0; //@line 106 "./common/Object.h"
   $140 = ($139|0)==(0|0); //@line 106 "./common/Object.h"
   if (!($140)) {
    $141 = HEAP32[$139>>2]|0; //@line 106 "./common/Object.h"
    $142 = ((($141)) + 12|0); //@line 106 "./common/Object.h"
    $143 = HEAP32[$142>>2]|0; //@line 106 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($143|0,($139|0)); //@line 106 "./common/Object.h"
    $144 = __THREW__; __THREW__ = 0;
    $145 = $144&1;
    if ($145) {
     $146 = ___cxa_find_matching_catch_3(0|0)|0;
     $147 = tempRet0;
     ___clang_call_terminate($146); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   }
   $148 = ((($this)) + 8|0); //@line 106 "./common/Object.h"
   $149 = HEAP32[$148>>2]|0; //@line 106 "./common/Object.h"
   $150 = ($149|0)==(0|0); //@line 106 "./common/Object.h"
   if ($150) {
    __ZN4love8graphics8GraphicsD2Ev($this); //@line 97 "modules/graphics/opengl/Graphics.cpp"
    return; //@line 97 "modules/graphics/opengl/Graphics.cpp"
   }
   $151 = HEAP32[$149>>2]|0; //@line 106 "./common/Object.h"
   $152 = ((($151)) + 12|0); //@line 106 "./common/Object.h"
   $153 = HEAP32[$152>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($153|0,($149|0)); //@line 106 "./common/Object.h"
   $154 = __THREW__; __THREW__ = 0;
   $155 = $154&1;
   if ($155) {
    $156 = ___cxa_find_matching_catch_3(0|0)|0;
    $157 = tempRet0;
    ___clang_call_terminate($156); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   } else {
    __ZN4love8graphics8GraphicsD2Ev($this); //@line 97 "modules/graphics/opengl/Graphics.cpp"
    return; //@line 97 "modules/graphics/opengl/Graphics.cpp"
   }
  }
 } while(0);
 $64 = ___cxa_find_matching_catch_3(0|0)|0;
 $65 = tempRet0;
 $66 = ((($this)) + 56|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $67 = HEAP32[$66>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $68 = ($67|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $69 = $67; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($68)) {
  $70 = ((($this)) + 60|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $71 = HEAP32[$70>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $72 = ($71|0)==($67|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($72)) {
   $scevgep$i$i$i = ((($71)) + -4|0);
   $73 = $scevgep$i$i$i;
   $74 = (($73) - ($69))|0;
   $75 = $74 >>> 2;
   $76 = $75 ^ -1;
   $scevgep4$i$i$i = (($71) + ($76<<2)|0);
   HEAP32[$70>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($67); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 __ZNSt3__113__vector_baseIN4love8graphics6opengl8Graphics12DisplayStateENS_9allocatorIS5_EEED2Ev($0); //@line 97 "modules/graphics/opengl/Graphics.cpp"
 $77 = ((($this)) + 16|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $78 = HEAP32[$77>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $79 = ($78|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $80 = $78; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($79)) {
  $81 = ((($this)) + 20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $82 = HEAP32[$81>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $83 = ($82|0)==($78|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($83)) {
   $scevgep$i$i$i1 = ((($82)) + -8|0);
   $84 = $scevgep$i$i$i1;
   $85 = (($84) - ($80))|0;
   $86 = $85 >>> 3;
   $87 = $86 ^ -1;
   $scevgep4$i$i$i2 = (($82) + ($87<<3)|0);
   HEAP32[$81>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($78); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $88 = HEAP32[$49>>2]|0; //@line 106 "./common/Object.h"
 $89 = ($88|0)==(0|0); //@line 106 "./common/Object.h"
 if (!($89)) {
  $90 = HEAP32[$88>>2]|0; //@line 106 "./common/Object.h"
  $91 = ((($90)) + 12|0); //@line 106 "./common/Object.h"
  $92 = HEAP32[$91>>2]|0; //@line 106 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($92|0,($88|0)); //@line 106 "./common/Object.h"
  $93 = __THREW__; __THREW__ = 0;
  $94 = $93&1;
  if ($94) {
   $95 = ___cxa_find_matching_catch_3(0|0)|0;
   $96 = tempRet0;
   ___clang_call_terminate($95); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 }
 $97 = ((($this)) + 8|0); //@line 106 "./common/Object.h"
 $98 = HEAP32[$97>>2]|0; //@line 106 "./common/Object.h"
 $99 = ($98|0)==(0|0); //@line 106 "./common/Object.h"
 if ($99) {
  __ZN4love8graphics8GraphicsD2Ev($this); //@line 97 "modules/graphics/opengl/Graphics.cpp"
  ___clang_call_terminate($64); //@line 97 "modules/graphics/opengl/Graphics.cpp"
  // unreachable; //@line 97 "modules/graphics/opengl/Graphics.cpp"
 }
 $100 = HEAP32[$98>>2]|0; //@line 106 "./common/Object.h"
 $101 = ((($100)) + 12|0); //@line 106 "./common/Object.h"
 $102 = HEAP32[$101>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($102|0,($98|0)); //@line 106 "./common/Object.h"
 $103 = __THREW__; __THREW__ = 0;
 $104 = $103&1;
 if ($104) {
  $105 = ___cxa_find_matching_catch_3(0|0)|0;
  $106 = tempRet0;
  ___clang_call_terminate($105); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  __ZN4love8graphics8GraphicsD2Ev($this); //@line 97 "modules/graphics/opengl/Graphics.cpp"
  ___clang_call_terminate($64); //@line 97 "modules/graphics/opengl/Graphics.cpp"
  // unreachable; //@line 97 "modules/graphics/opengl/Graphics.cpp"
 }
}
function __ZNSt3__113__vector_baseIN4love8graphics6opengl8Graphics12DisplayStateENS_9allocatorIS5_EEED2Ev($this) {
 $this = $this|0;
 var $$pre = 0, $$pre$i$i$i$i$i$i$i = 0, $$pre$i$i$i$i$i$i$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ($0|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($1) {
  return; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $2 = ((($this)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($3|0)==($0|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($4) {
   $49 = $0;
  } else {
   $6 = $3;
   L5: while(1) {
    $5 = ((($6)) + -136|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$2>>2] = $5; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $7 = ((($6)) + -44|0); //@line 482 "modules/graphics/opengl/Graphics.h"
    $8 = HEAP32[$7>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $9 = ($8|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($9)) {
     $10 = ((($6)) + -40|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $11 = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $12 = ($11|0)==($8|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($12) {
      $26 = $8;
     } else {
      $14 = $11;
      while(1) {
       $13 = ((($14)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$10>>2] = $13; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $15 = HEAP32[$13>>2]|0; //@line 106 "./common/Object.h"
       $16 = ($15|0)==(0|0); //@line 106 "./common/Object.h"
       if ($16) {
        $18 = $13;
       } else {
        $19 = HEAP32[$15>>2]|0; //@line 106 "./common/Object.h"
        $20 = ((($19)) + 12|0); //@line 106 "./common/Object.h"
        $21 = HEAP32[$20>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($21|0,($15|0)); //@line 106 "./common/Object.h"
        $22 = __THREW__; __THREW__ = 0;
        $23 = $22&1;
        if ($23) {
         label = 9;
         break L5;
        }
        $$pre$i$i$i$i$i$i$i$i$i = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $18 = $$pre$i$i$i$i$i$i$i$i$i;
       }
       $17 = ($18|0)==($8|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($17) {
        break;
       } else {
        $14 = $18;
       }
      }
      $$pre$i$i$i$i$i$i$i = HEAP32[$7>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $26 = $$pre$i$i$i$i$i$i$i;
     }
     __ZdlPv($26); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    }
    $27 = ((($6)) + -48|0); //@line 106 "./common/Object.h"
    $28 = HEAP32[$27>>2]|0; //@line 106 "./common/Object.h"
    $29 = ($28|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($29)) {
     $30 = HEAP32[$28>>2]|0; //@line 106 "./common/Object.h"
     $31 = ((($30)) + 12|0); //@line 106 "./common/Object.h"
     $32 = HEAP32[$31>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($32|0,($28|0)); //@line 106 "./common/Object.h"
     $33 = __THREW__; __THREW__ = 0;
     $34 = $33&1;
     if ($34) {
      label = 14;
      break;
     }
    }
    $37 = ((($6)) + -52|0); //@line 106 "./common/Object.h"
    $38 = HEAP32[$37>>2]|0; //@line 106 "./common/Object.h"
    $39 = ($38|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($39)) {
     $42 = HEAP32[$38>>2]|0; //@line 106 "./common/Object.h"
     $43 = ((($42)) + 12|0); //@line 106 "./common/Object.h"
     $44 = HEAP32[$43>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($44|0,($38|0)); //@line 106 "./common/Object.h"
     $45 = __THREW__; __THREW__ = 0;
     $46 = $45&1;
     if ($46) {
      label = 18;
      break;
     }
    }
    $40 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $41 = ($40|0)==($0|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($41) {
     label = 19;
     break;
    } else {
     $6 = $40;
    }
   }
   if ((label|0) == 9) {
    $24 = ___cxa_find_matching_catch_3(0|0)|0;
    $25 = tempRet0;
    ___clang_call_terminate($24); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 14) {
    $35 = ___cxa_find_matching_catch_3(0|0)|0;
    $36 = tempRet0;
    ___clang_call_terminate($35); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 18) {
    $47 = ___cxa_find_matching_catch_3(0|0)|0;
    $48 = tempRet0;
    ___clang_call_terminate($47); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 19) {
    $$pre = HEAP32[$this>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $49 = $$pre;
    break;
   }
  }
 } while(0);
 __ZdlPv($49); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl8GraphicsD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl8GraphicsD2Ev($this); //@line 79 "modules/graphics/opengl/Graphics.cpp"
 __ZdlPv($this); //@line 79 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 97 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNK4love8graphics6opengl8Graphics7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (434837|0); //@line 101 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics15setViewportSizeEii($this,$width,$height) {
 $this = $this|0;
 $width = $width|0;
 $height = $height|0;
 var $$pre$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i$i = 0, $$pre$i$i$i3 = 0, $$pre$i6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $9 = 0, $canvases = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $canvases = sp + 96|0;
 $0 = sp + 80|0;
 $1 = sp + 64|0;
 $2 = sp;
 $3 = ((($this)) + 32|0); //@line 218 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$3>>2] = $width; //@line 218 "modules/graphics/opengl/Graphics.cpp"
 $4 = ((($this)) + 36|0); //@line 219 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$4>>2] = $height; //@line 219 "modules/graphics/opengl/Graphics.cpp"
 $5 = ((($this)) + 40|0); //@line 620 "modules/graphics/opengl/Graphics.cpp"
 $6 = HEAP8[$5>>0]|0; //@line 620 "modules/graphics/opengl/Graphics.cpp"
 $7 = ($6<<24>>24)==(0); //@line 620 "modules/graphics/opengl/Graphics.cpp"
 if ($7) {
  STACKTOP = sp;return; //@line 241 "modules/graphics/opengl/Graphics.cpp"
 }
 $8 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = HEAP32[$8>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = ((($9)) + -44|0); //@line 226 "modules/graphics/opengl/Graphics.cpp"
 __ZNSt3__16vectorIN4love9StrongRefINS1_8graphics6opengl6CanvasEEENS_9allocatorIS6_EEEC2ERKS9_($canvases,$10); //@line 226 "modules/graphics/opengl/Graphics.cpp"
 $11 = HEAP32[$8>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = HEAP32[158272]|0; //@line 1065 "modules/graphics/opengl/Graphics.cpp"
 $13 = ($12|0)==(0|0); //@line 1065 "modules/graphics/opengl/Graphics.cpp"
 if ($13) {
  label = 4;
 } else {
  __THREW__ = 0;
  invoke_vii(4001,($12|0),0); //@line 1066 "modules/graphics/opengl/Graphics.cpp"
  $14 = __THREW__; __THREW__ = 0;
  $15 = $14&1;
  if (!($15)) {
   label = 4;
  }
 }
 if ((label|0) == 4) {
  $16 = ((($11)) + -44|0); //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = HEAP32[$16>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ((($11)) + -40|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = HEAP32[$18>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = ($19|0)==($17|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  L8: do {
   if (!($20)) {
    $22 = $19;
    while(1) {
     $21 = ((($22)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$18>>2] = $21; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $23 = HEAP32[$21>>2]|0; //@line 106 "./common/Object.h"
     $24 = ($23|0)==(0|0); //@line 106 "./common/Object.h"
     if ($24) {
      $26 = $21;
     } else {
      $27 = HEAP32[$23>>2]|0; //@line 106 "./common/Object.h"
      $28 = ((($27)) + 12|0); //@line 106 "./common/Object.h"
      $29 = HEAP32[$28>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($29|0,($23|0)); //@line 106 "./common/Object.h"
      $30 = __THREW__; __THREW__ = 0;
      $31 = $30&1;
      if ($31) {
       break;
      }
      $$pre$i$i$i$i = HEAP32[$18>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $26 = $$pre$i$i$i$i;
     }
     $25 = ($26|0)==($17|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($25) {
      break L8;
     } else {
      $22 = $26;
     }
    }
    $32 = ___cxa_find_matching_catch_3(0|0)|0;
    $33 = tempRet0;
    ___clang_call_terminate($32); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
  } while(0);
  HEAP32[$0>>2] = 0; //@line 230 "modules/graphics/opengl/Graphics.cpp"
  $34 = ((($0)) + 4|0); //@line 230 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$34>>2] = 0; //@line 230 "modules/graphics/opengl/Graphics.cpp"
  $35 = ((($0)) + 8|0); //@line 230 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$35>>2] = $width; //@line 230 "modules/graphics/opengl/Graphics.cpp"
  $36 = ((($0)) + 12|0); //@line 230 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$36>>2] = $height; //@line 230 "modules/graphics/opengl/Graphics.cpp"
  __THREW__ = 0;
  invoke_vii(4043,(634044|0),($0|0)); //@line 230 "modules/graphics/opengl/Graphics.cpp"
  $37 = __THREW__; __THREW__ = 0;
  $38 = $37&1;
  if (!($38)) {
   __THREW__ = 0;
   invoke_vii(4044,($1|0),(634044|0)); //@line 234 "modules/graphics/opengl/Graphics.cpp"
   $39 = __THREW__; __THREW__ = 0;
   $40 = $39&1;
   if (!($40)) {
    ;HEAP32[633092>>2]=HEAP32[$1>>2]|0;HEAP32[633092+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[633092+8>>2]=HEAP32[$1+8>>2]|0;HEAP32[633092+12>>2]=HEAP32[$1+12>>2]|0; //@line 234 "modules/graphics/opengl/Graphics.cpp"
    $41 = HEAP32[(634060)>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $42 = (+($width|0)); //@line 237 "modules/graphics/opengl/Graphics.cpp"
    $43 = (+($height|0)); //@line 237 "modules/graphics/opengl/Graphics.cpp"
    __THREW__ = 0;
    invoke_vidddd(4045,($2|0),0.0,(+$42),(+$43),0.0); //@line 237 "modules/graphics/opengl/Graphics.cpp"
    $44 = __THREW__; __THREW__ = 0;
    $45 = $44&1;
    if (!($45)) {
     $46 = ((($41)) + -64|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     dest=$46; src=$2; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 237 "modules/graphics/opengl/Graphics.cpp"
     __ZN4love7Matrix4D2Ev($2); //@line 237 "modules/graphics/opengl/Graphics.cpp"
     __THREW__ = 0;
     invoke_vii(4046,($this|0),($canvases|0)); //@line 240 "modules/graphics/opengl/Graphics.cpp"
     $47 = __THREW__; __THREW__ = 0;
     $48 = $47&1;
     if (!($48)) {
      $49 = HEAP32[$canvases>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $50 = ($49|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($50)) {
       $51 = ((($canvases)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $52 = HEAP32[$51>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $53 = ($52|0)==($49|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       do {
        if ($53) {
         $67 = $49;
        } else {
         $55 = $52;
         while(1) {
          $54 = ((($55)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$51>>2] = $54; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $56 = HEAP32[$54>>2]|0; //@line 106 "./common/Object.h"
          $57 = ($56|0)==(0|0); //@line 106 "./common/Object.h"
          if ($57) {
           $59 = $54;
          } else {
           $60 = HEAP32[$56>>2]|0; //@line 106 "./common/Object.h"
           $61 = ((($60)) + 12|0); //@line 106 "./common/Object.h"
           $62 = HEAP32[$61>>2]|0; //@line 106 "./common/Object.h"
           __THREW__ = 0;
           invoke_vi($62|0,($56|0)); //@line 106 "./common/Object.h"
           $63 = __THREW__; __THREW__ = 0;
           $64 = $63&1;
           if ($64) {
            label = 20;
            break;
           }
           $$pre$i$i$i3 = HEAP32[$51>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $59 = $$pre$i$i$i3;
          }
          $58 = ($59|0)==($49|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          if ($58) {
           label = 21;
           break;
          } else {
           $55 = $59;
          }
         }
         if ((label|0) == 20) {
          $65 = ___cxa_find_matching_catch_3(0|0)|0;
          $66 = tempRet0;
          ___clang_call_terminate($65); //@line 106 "./common/Object.h"
          // unreachable; //@line 106 "./common/Object.h"
         }
         else if ((label|0) == 21) {
          $$pre$i6 = HEAP32[$canvases>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $67 = $$pre$i6;
          break;
         }
        }
       } while(0);
       __ZdlPv($67); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      }
      STACKTOP = sp;return; //@line 241 "modules/graphics/opengl/Graphics.cpp"
     }
    }
   }
  }
 }
 $68 = ___cxa_find_matching_catch_2()|0;
 $69 = tempRet0;
 $70 = HEAP32[$canvases>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $71 = ($70|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($71) {
  ___resumeException($68|0);
  // unreachable;
 }
 $72 = ((($canvases)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $73 = HEAP32[$72>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $74 = ($73|0)==($70|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($74) {
   $88 = $70;
  } else {
   $76 = $73;
   while(1) {
    $75 = ((($76)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$72>>2] = $75; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $77 = HEAP32[$75>>2]|0; //@line 106 "./common/Object.h"
    $78 = ($77|0)==(0|0); //@line 106 "./common/Object.h"
    if ($78) {
     $80 = $75;
    } else {
     $81 = HEAP32[$77>>2]|0; //@line 106 "./common/Object.h"
     $82 = ((($81)) + 12|0); //@line 106 "./common/Object.h"
     $83 = HEAP32[$82>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($83|0,($77|0)); //@line 106 "./common/Object.h"
     $84 = __THREW__; __THREW__ = 0;
     $85 = $84&1;
     if ($85) {
      label = 31;
      break;
     }
     $$pre$i$i$i = HEAP32[$72>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $80 = $$pre$i$i$i;
    }
    $79 = ($80|0)==($70|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($79) {
     label = 32;
     break;
    } else {
     $76 = $80;
    }
   }
   if ((label|0) == 31) {
    $86 = ___cxa_find_matching_catch_3(0|0)|0;
    $87 = tempRet0;
    ___clang_call_terminate($86); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 32) {
    $$pre$i = HEAP32[$canvases>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $88 = $$pre$i;
    break;
   }
  }
 } while(0);
 __ZdlPv($88); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($68|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love9StrongRefINS1_8graphics6opengl6CanvasEEENS_9allocatorIS6_EEEC2ERKS9_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$01$i = 0, $$pre$i = 0, $$pre$i$i$i = 0, $$pre2$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $0 = ((($this)) + 4|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($this)) + 8|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($__x)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$__x>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = $5 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ($6|0)==(0); //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($7) {
  return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $8 = ($6>>>0)>(1073741823); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($8) {
  __THREW__ = 0;
  invoke_vi(3528,($this|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if ($10) {
   label = 12;
  } else {
   label = 4;
  }
 } else {
  label = 4;
 }
 do {
  if ((label|0) == 4) {
   __THREW__ = 0;
   $11 = (invoke_ii(3348,($5|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $12 = __THREW__; __THREW__ = 0;
   $13 = $12&1;
   if ($13) {
    label = 12;
   } else {
    HEAP32[$0>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$this>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $14 = (($11) + ($6<<2)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $15 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$15>>2] = $14; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $16 = HEAP32[$__x>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $17 = HEAP32[$2>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $18 = ($16|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($18) {
     return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    $$01$i = $16;$20 = $11;
    while(1) {
     $19 = HEAP32[$$01$i>>2]|0; //@line 139 "./common/Object.h"
     HEAP32[$20>>2] = $19; //@line 99 "./common/Object.h"
     $21 = ($19|0)==(0|0); //@line 101 "./common/Object.h"
     if ($21) {
      $28 = $20;
     } else {
      $22 = HEAP32[$19>>2]|0; //@line 101 "./common/Object.h"
      $23 = ((($22)) + 8|0); //@line 101 "./common/Object.h"
      $24 = HEAP32[$23>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($24|0,($19|0)); //@line 101 "./common/Object.h"
      $25 = __THREW__; __THREW__ = 0;
      $26 = $25&1;
      if ($26) {
       label = 11;
       break;
      }
      $$pre2$i = HEAP32[$0>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $28 = $$pre2$i;
     }
     $27 = ((($28)) + 4|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$0>>2] = $27; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $29 = ((($$01$i)) + 4|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $30 = ($29|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($30) {
      label = 23;
      break;
     } else {
      $$01$i = $29;$20 = $27;
     }
    }
    if ((label|0) == 11) {
     $31 = ___cxa_find_matching_catch_2()|0;
     $32 = tempRet0;
     $lpad$phi$index = $31;$lpad$phi$index2 = $32;
     break;
    }
    else if ((label|0) == 23) {
     return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
   }
  }
 } while(0);
 if ((label|0) == 12) {
  $33 = ___cxa_find_matching_catch_2()|0;
  $34 = tempRet0;
  $lpad$phi$index = $33;$lpad$phi$index2 = $34;
 }
 $35 = HEAP32[$this>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $36 = ($35|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($36) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $37 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $38 = ($37|0)==($35|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($38) {
   $52 = $35;
  } else {
   $40 = $37;
   while(1) {
    $39 = ((($40)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$0>>2] = $39; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $41 = HEAP32[$39>>2]|0; //@line 106 "./common/Object.h"
    $42 = ($41|0)==(0|0); //@line 106 "./common/Object.h"
    if ($42) {
     $44 = $39;
    } else {
     $45 = HEAP32[$41>>2]|0; //@line 106 "./common/Object.h"
     $46 = ((($45)) + 12|0); //@line 106 "./common/Object.h"
     $47 = HEAP32[$46>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($47|0,($41|0)); //@line 106 "./common/Object.h"
     $48 = __THREW__; __THREW__ = 0;
     $49 = $48&1;
     if ($49) {
      label = 19;
      break;
     }
     $$pre$i$i$i = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $44 = $$pre$i$i$i;
    }
    $43 = ($44|0)==($35|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($43) {
     label = 20;
     break;
    } else {
     $40 = $44;
    }
   }
   if ((label|0) == 19) {
    $50 = ___cxa_find_matching_catch_3(0|0)|0;
    $51 = tempRet0;
    ___clang_call_terminate($50); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 20) {
    $$pre$i = HEAP32[$this>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $52 = $$pre$i;
    break;
   }
  }
 } while(0);
 __ZdlPv($52); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZN4love8graphics6opengl8Graphics9setCanvasERKNSt3__16vectorINS_9StrongRefINS1_6CanvasEEENS3_9allocatorIS7_EEEE($this,$canvases) {
 $this = $this|0;
 $canvases = $canvases|0;
 var $$pre = 0, $$pre8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$07 = 0, $canvaslist = 0, $lpad$phi$index = 0;
 var $lpad$phi$index2 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i2 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $canvaslist = sp + 4|0;
 $0 = sp;
 HEAP32[$canvaslist>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($canvaslist)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($canvaslist)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = ((($canvases)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$canvases>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = (($4) - ($5))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = $6 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($canvaslist)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = ($7|0)==(0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $5; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $4; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($9) {
  $53 = 0;$54 = 0;
  label = 4;
 } else {
  __THREW__ = 0;
  $12 = (invoke_ii(3348,($6|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $13 = __THREW__; __THREW__ = 0;
  $14 = $13&1;
  if ($14) {
   label = 8;
  } else {
   $15 = $12; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $16 = (($12) + ($7<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $17 = $16; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   HEAP32[$canvaslist>>2] = $15; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$1>>2] = $15; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$8>>2] = $17; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $53 = $16;$54 = $12;
   label = 4;
  }
 }
 L4: do {
  if ((label|0) == 4) {
   $18 = ($10|0)==($11|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   L6: do {
    if (!($18)) {
     $36 = $54;$37 = $53;$__begin$sroa$0$07 = $10;
     while(1) {
      $34 = HEAP32[$__begin$sroa$0$07>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$0>>2] = $34; //@line 1056 "modules/graphics/opengl/Graphics.cpp"
      $35 = ($36>>>0)<($37>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($35) {
       HEAP32[$36>>2] = $34; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $38 = HEAP32[$1>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $39 = ((($38)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$1>>2] = $39; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      } else {
       __THREW__ = 0;
       invoke_vii(3574,($canvaslist|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $40 = __THREW__; __THREW__ = 0;
       $41 = $40&1;
       if ($41) {
        break;
       }
      }
      $42 = ((($__begin$sroa$0$07)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $43 = ($42|0)==($11|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($43) {
       break L6;
      }
      $$pre = HEAP32[$1>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$pre8 = HEAP32[$8>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $36 = $$pre;$37 = $$pre8;$__begin$sroa$0$07 = $42;
     }
     $21 = ___cxa_find_matching_catch_2()|0;
     $22 = tempRet0;
     $lpad$phi$index = $21;$lpad$phi$index2 = $22;
     break L4;
    }
   } while(0);
   __THREW__ = 0;
   invoke_vii(3575,($this|0),($canvaslist|0)); //@line 1058 "modules/graphics/opengl/Graphics.cpp"
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   if ($20) {
    label = 8;
   } else {
    $44 = HEAP32[$canvaslist>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $45 = ($44|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $46 = $44; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($45) {
     STACKTOP = sp;return; //@line 1059 "modules/graphics/opengl/Graphics.cpp"
    }
    $47 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $48 = ($47|0)==($44|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($48)) {
     $scevgep$i$i$i = ((($47)) + -4|0);
     $49 = $scevgep$i$i$i;
     $50 = (($49) - ($46))|0;
     $51 = $50 >>> 2;
     $52 = $51 ^ -1;
     $scevgep4$i$i$i = (($47) + ($52<<2)|0);
     HEAP32[$1>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($44); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    STACKTOP = sp;return; //@line 1059 "modules/graphics/opengl/Graphics.cpp"
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $23 = ___cxa_find_matching_catch_2()|0;
  $24 = tempRet0;
  $lpad$phi$index = $23;$lpad$phi$index2 = $24;
 }
 $25 = HEAP32[$canvaslist>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $26 = ($25|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = $25; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($26) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $28 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $29 = ($28|0)==($25|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($29)) {
  $scevgep$i$i$i2 = ((($28)) + -4|0);
  $30 = $scevgep$i$i$i2;
  $31 = (($30) - ($27))|0;
  $32 = $31 >>> 2;
  $33 = $32 ^ -1;
  $scevgep4$i$i$i3 = (($28) + ($33<<2)|0);
  HEAP32[$1>>2] = $scevgep4$i$i$i3; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($25); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZN4love8graphics6opengl8Graphics9setCanvasERKNSt3__16vectorIPNS1_6CanvasENS3_9allocatorIS6_EEEE($this,$canvases) {
 $this = $this|0;
 $canvases = $canvases|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$lcssa53 = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i$i = 0, $$pre$i$i$i8 = 0, $$pre$i11 = 0, $$pre$i4 = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0;
 var $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0;
 var $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$027 = 0, $attachments = 0, $canvasrefs = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i$i = 0, $scevgep$i$i$i15 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i$i = 0, $scevgep4$i$i$i16 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $attachments = sp + 16|0;
 $canvasrefs = sp + 4|0;
 $0 = sp;
 $1 = ((($canvases)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$canvases>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($2) - ($3))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = $4 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = $3; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 switch ($5|0) {
 case 0:  {
  $7 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $8 = HEAP32[$7>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $9 = HEAP32[158272]|0; //@line 1065 "modules/graphics/opengl/Graphics.cpp"
  $10 = ($9|0)==(0|0); //@line 1065 "modules/graphics/opengl/Graphics.cpp"
  if (!($10)) {
   __ZN4love8graphics6opengl6Canvas8stopGrabEb($9,0); //@line 1066 "modules/graphics/opengl/Graphics.cpp"
  }
  $11 = ((($8)) + -44|0); //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $12 = HEAP32[$11>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = ((($8)) + -40|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = HEAP32[$13>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = ($14|0)==($12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($15) {
   STACKTOP = sp;return; //@line 1048 "modules/graphics/opengl/Graphics.cpp"
  } else {
   $17 = $14;
  }
  while(1) {
   $16 = ((($17)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$13>>2] = $16; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $18 = HEAP32[$16>>2]|0; //@line 106 "./common/Object.h"
   $19 = ($18|0)==(0|0); //@line 106 "./common/Object.h"
   if ($19) {
    $21 = $16;
   } else {
    $22 = HEAP32[$18>>2]|0; //@line 106 "./common/Object.h"
    $23 = ((($22)) + 12|0); //@line 106 "./common/Object.h"
    $24 = HEAP32[$23>>2]|0; //@line 106 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($24|0,($18|0)); //@line 106 "./common/Object.h"
    $25 = __THREW__; __THREW__ = 0;
    $26 = $25&1;
    if ($26) {
     label = 9;
     break;
    }
    $$pre$i$i$i$i = HEAP32[$13>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $21 = $$pre$i$i$i$i;
   }
   $20 = ($21|0)==($12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($20) {
    label = 62;
    break;
   } else {
    $17 = $21;
   }
  }
  if ((label|0) == 9) {
   $27 = ___cxa_find_matching_catch_3(0|0)|0;
   $28 = tempRet0;
   ___clang_call_terminate($27); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
  else if ((label|0) == 62) {
   STACKTOP = sp;return; //@line 1048 "modules/graphics/opengl/Graphics.cpp"
  }
  break;
 }
 case 1:  {
  $29 = HEAP32[$6>>2]|0; //@line 1034 "modules/graphics/opengl/Graphics.cpp"
  __ZN4love8graphics6opengl8Graphics9setCanvasEPNS1_6CanvasE($this,$29); //@line 1034 "modules/graphics/opengl/Graphics.cpp"
  STACKTOP = sp;return; //@line 1048 "modules/graphics/opengl/Graphics.cpp"
  break;
 }
 default: {
  $30 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $31 = HEAP32[$30>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $32 = ((($6)) + 4|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $33 = $32; //@line 786 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$attachments>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $34 = ((($attachments)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$34>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $35 = ((($attachments)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$35>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $36 = (($2) - ($33))|0; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $37 = $36 >> 2; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $38 = ($37|0)==(0); //@line 1177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $39 = $2; //@line 1180 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  L3: do {
   if (!($38)) {
    $40 = ($37>>>0)>(1073741823); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($40) {
     __THREW__ = 0;
     invoke_vi(3528,($attachments|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $41 = __THREW__; __THREW__ = 0;
     $42 = $41&1;
     if (!($42)) {
      label = 14;
     }
    } else {
     label = 14;
    }
    if ((label|0) == 14) {
     __THREW__ = 0;
     $43 = (invoke_ii(3348,($36|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $44 = __THREW__; __THREW__ = 0;
     $45 = $44&1;
     if (!($45)) {
      HEAP32[$34>>2] = $43; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$attachments>>2] = $43; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $46 = (($43) + ($37<<2)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $47 = ((($attachments)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$47>>2] = $46; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $48 = ($32|0)==($39|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($48) {
       break;
      }
      $50 = $32;$51 = $43;
      while(1) {
       $49 = HEAP32[$50>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       HEAP32[$51>>2] = $49; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $52 = HEAP32[$34>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $53 = ((($52)) + 4|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$34>>2] = $53; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $54 = ((($50)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       $55 = ($54|0)==($39|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       if ($55) {
        break L3;
       } else {
        $50 = $54;$51 = $53;
       }
      }
     }
    }
    $56 = ___cxa_find_matching_catch_2()|0;
    $57 = tempRet0;
    $58 = HEAP32[$attachments>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $59 = ($58|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $60 = $58; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($59) {
     ___resumeException($56|0);
     // unreachable;
    }
    $61 = HEAP32[$34>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $62 = ($61|0)==($58|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($62)) {
     $scevgep$i$i$i$i = ((($61)) + -4|0);
     $63 = $scevgep$i$i$i$i;
     $64 = (($63) - ($60))|0;
     $65 = $64 >>> 2;
     $66 = $65 ^ -1;
     $scevgep4$i$i$i$i = (($61) + ($66<<2)|0);
     HEAP32[$34>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($58); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    ___resumeException($56|0);
    // unreachable;
   }
  } while(0);
  $67 = HEAP32[$canvases>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $68 = HEAP32[$67>>2]|0; //@line 1039 "modules/graphics/opengl/Graphics.cpp"
  __THREW__ = 0;
  invoke_vii(4047,($68|0),($attachments|0)); //@line 1039 "modules/graphics/opengl/Graphics.cpp"
  $69 = __THREW__; __THREW__ = 0;
  $70 = $69&1;
  if ($70) {
   $119 = ___cxa_find_matching_catch_2()|0;
   $120 = tempRet0;
   $$1 = $119;$$12 = $120;
  } else {
   HEAP32[$canvasrefs>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $71 = ((($canvasrefs)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$71>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $72 = ((($canvasrefs)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$72>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $73 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $74 = HEAP32[$canvases>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $75 = (($73) - ($74))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $76 = $75 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __THREW__ = 0;
   invoke_vii(4048,($canvasrefs|0),($76|0)); //@line 1042 "modules/graphics/opengl/Graphics.cpp"
   $77 = __THREW__; __THREW__ = 0;
   $78 = $77&1;
   L25: do {
    if ($78) {
     $123 = ___cxa_find_matching_catch_2()|0;
     $124 = tempRet0;
     $lpad$phi$index = $123;$lpad$phi$index2 = $124;
     label = 45;
    } else {
     $79 = HEAP32[$canvases>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $80 = HEAP32[$1>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $81 = ($79|0)==($80|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     L28: do {
      if ($81) {
       $$pre = ((($canvasrefs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $$pre$phiZ2D = $$pre;
      } else {
       $82 = ((($canvasrefs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $__begin$sroa$0$027 = $79;
       while(1) {
        $125 = HEAP32[$__begin$sroa$0$027>>2]|0; //@line 106 "./common/Object.h"
        HEAP32[$0>>2] = $125; //@line 93 "./common/Object.h"
        $126 = ($125|0)==(0|0); //@line 95 "./common/Object.h"
        if (!($126)) {
         $127 = HEAP32[$125>>2]|0; //@line 95 "./common/Object.h"
         $128 = ((($127)) + 8|0); //@line 95 "./common/Object.h"
         $129 = HEAP32[$128>>2]|0; //@line 95 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($129|0,($125|0)); //@line 95 "./common/Object.h"
         $130 = __THREW__; __THREW__ = 0;
         $131 = $130&1;
         if ($131) {
          label = 43;
          break;
         }
        }
        $132 = HEAP32[$71>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $133 = HEAP32[$82>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $134 = ($132>>>0)<($133>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($134) {
         HEAP32[$132>>2] = $125; //@line 99 "./common/Object.h"
         if ($126) {
          $143 = $132;
         } else {
          $135 = HEAP32[$125>>2]|0; //@line 101 "./common/Object.h"
          $136 = ((($135)) + 8|0); //@line 101 "./common/Object.h"
          $137 = HEAP32[$136>>2]|0; //@line 101 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($137|0,($125|0)); //@line 101 "./common/Object.h"
          $138 = __THREW__; __THREW__ = 0;
          $139 = $138&1;
          if ($139) {
           $$lcssa53 = $125;
           label = 51;
           break;
          }
          $$pre$i4 = HEAP32[$71>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $143 = $$pre$i4;
         }
         $142 = ((($143)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$71>>2] = $142; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        } else {
         __THREW__ = 0;
         invoke_vii(4049,($canvasrefs|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $144 = __THREW__; __THREW__ = 0;
         $145 = $144&1;
         if ($145) {
          label = 59;
          break;
         }
        }
        if (!($126)) {
         $146 = HEAP32[$125>>2]|0; //@line 106 "./common/Object.h"
         $147 = ((($146)) + 12|0); //@line 106 "./common/Object.h"
         $148 = HEAP32[$147>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($148|0,($125|0)); //@line 106 "./common/Object.h"
         $149 = __THREW__; __THREW__ = 0;
         $150 = $149&1;
         if ($150) {
          label = 57;
          break;
         }
        }
        $153 = ((($__begin$sroa$0$027)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        $154 = ($153|0)==($80|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        if ($154) {
         $$pre$phiZ2D = $82;
         break L28;
        } else {
         $__begin$sroa$0$027 = $153;
        }
       }
       if ((label|0) == 43) {
        $121 = ___cxa_find_matching_catch_2()|0;
        $122 = tempRet0;
        $lpad$phi$index = $121;$lpad$phi$index2 = $122;
        label = 45;
        break L25;
       }
       else if ((label|0) == 51) {
        $140 = ___cxa_find_matching_catch_2()|0;
        $141 = tempRet0;
        $158 = $$lcssa53;$192 = $141;$193 = $140;
       }
       else if ((label|0) == 57) {
        $151 = ___cxa_find_matching_catch_3(0|0)|0;
        $152 = tempRet0;
        ___clang_call_terminate($151); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       else if ((label|0) == 59) {
        $155 = ___cxa_find_matching_catch_2()|0;
        $156 = tempRet0;
        if ($126) {
         $$0 = $155;$$01 = $156;
         break L25;
        } else {
         $158 = $125;$192 = $156;$193 = $155;
        }
       }
       $157 = HEAP32[$158>>2]|0; //@line 106 "./common/Object.h"
       $159 = ((($157)) + 12|0); //@line 106 "./common/Object.h"
       $160 = HEAP32[$159>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($160|0,($158|0)); //@line 106 "./common/Object.h"
       $161 = __THREW__; __THREW__ = 0;
       $162 = $161&1;
       if (!($162)) {
        $$0 = $193;$$01 = $192;
        break L25;
       }
       $163 = ___cxa_find_matching_catch_3(0|0)|0;
       $164 = tempRet0;
       ___clang_call_terminate($163); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
     } while(0);
     $83 = ((($31)) + -44|0); //@line 1047 "modules/graphics/opengl/Graphics.cpp"
     $84 = HEAP32[$83>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $85 = HEAP32[$canvasrefs>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$83>>2] = $85; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$canvasrefs>>2] = $84; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $86 = ((($31)) + -40|0); //@line 2026 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $87 = HEAP32[$86>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $88 = HEAP32[$71>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$86>>2] = $88; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$71>>2] = $87; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $89 = ((($31)) + -36|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $90 = HEAP32[$89>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $91 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$89>>2] = $91; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$$pre$phiZ2D>>2] = $90; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $92 = HEAP32[$canvasrefs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $93 = ($92|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($93)) {
      $94 = HEAP32[$71>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $95 = ($94|0)==($92|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      do {
       if ($95) {
        $109 = $92;
       } else {
        $97 = $94;
        while(1) {
         $96 = ((($97)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$71>>2] = $96; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $98 = HEAP32[$96>>2]|0; //@line 106 "./common/Object.h"
         $99 = ($98|0)==(0|0); //@line 106 "./common/Object.h"
         if ($99) {
          $101 = $96;
         } else {
          $102 = HEAP32[$98>>2]|0; //@line 106 "./common/Object.h"
          $103 = ((($102)) + 12|0); //@line 106 "./common/Object.h"
          $104 = HEAP32[$103>>2]|0; //@line 106 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($104|0,($98|0)); //@line 106 "./common/Object.h"
          $105 = __THREW__; __THREW__ = 0;
          $106 = $105&1;
          if ($106) {
           label = 34;
           break;
          }
          $$pre$i$i$i8 = HEAP32[$71>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $101 = $$pre$i$i$i8;
         }
         $100 = ($101|0)==($92|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($100) {
          label = 35;
          break;
         } else {
          $97 = $101;
         }
        }
        if ((label|0) == 34) {
         $107 = ___cxa_find_matching_catch_3(0|0)|0;
         $108 = tempRet0;
         ___clang_call_terminate($107); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        else if ((label|0) == 35) {
         $$pre$i11 = HEAP32[$canvasrefs>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $109 = $$pre$i11;
         break;
        }
       }
      } while(0);
      __ZdlPv($109); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $110 = HEAP32[$attachments>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $111 = ($110|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $112 = $110; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($111)) {
      $113 = HEAP32[$34>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $114 = ($113|0)==($110|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($114)) {
       $scevgep$i$i$i15 = ((($113)) + -4|0);
       $115 = $scevgep$i$i$i15;
       $116 = (($115) - ($112))|0;
       $117 = $116 >>> 2;
       $118 = $117 ^ -1;
       $scevgep4$i$i$i16 = (($113) + ($118<<2)|0);
       HEAP32[$34>>2] = $scevgep4$i$i$i16; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($110); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     STACKTOP = sp;return; //@line 1048 "modules/graphics/opengl/Graphics.cpp"
    }
   } while(0);
   if ((label|0) == 45) {
    $$0 = $lpad$phi$index;$$01 = $lpad$phi$index2;
   }
   $165 = HEAP32[$canvasrefs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $166 = ($165|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($166) {
    $$1 = $$0;$$12 = $$01;
   } else {
    $167 = HEAP32[$71>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $168 = ($167|0)==($165|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    do {
     if ($168) {
      $182 = $165;
     } else {
      $170 = $167;
      while(1) {
       $169 = ((($170)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$71>>2] = $169; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $171 = HEAP32[$169>>2]|0; //@line 106 "./common/Object.h"
       $172 = ($171|0)==(0|0); //@line 106 "./common/Object.h"
       if ($172) {
        $174 = $169;
       } else {
        $175 = HEAP32[$171>>2]|0; //@line 106 "./common/Object.h"
        $176 = ((($175)) + 12|0); //@line 106 "./common/Object.h"
        $177 = HEAP32[$176>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($177|0,($171|0)); //@line 106 "./common/Object.h"
        $178 = __THREW__; __THREW__ = 0;
        $179 = $178&1;
        if ($179) {
         label = 69;
         break;
        }
        $$pre$i$i$i = HEAP32[$71>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $174 = $$pre$i$i$i;
       }
       $173 = ($174|0)==($165|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($173) {
        label = 70;
        break;
       } else {
        $170 = $174;
       }
      }
      if ((label|0) == 69) {
       $180 = ___cxa_find_matching_catch_3(0|0)|0;
       $181 = tempRet0;
       ___clang_call_terminate($180); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      else if ((label|0) == 70) {
       $$pre$i = HEAP32[$canvasrefs>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $182 = $$pre$i;
       break;
      }
     }
    } while(0);
    __ZdlPv($182); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $$1 = $$0;$$12 = $$01;
   }
  }
  $183 = HEAP32[$attachments>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $184 = ($183|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $185 = $183; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($184) {
   ___resumeException($$1|0);
   // unreachable;
  }
  $186 = HEAP32[$34>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $187 = ($186|0)==($183|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($187)) {
   $scevgep$i$i$i = ((($186)) + -4|0);
   $188 = $scevgep$i$i$i;
   $189 = (($188) - ($185))|0;
   $190 = $189 >>> 2;
   $191 = $190 ^ -1;
   $scevgep4$i$i$i = (($186) + ($191<<2)|0);
   HEAP32[$34>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($183); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($$1|0);
  // unreachable;
 }
 }
}
function __ZN4love8graphics6opengl8Graphics9setCanvasEPNS1_6CanvasE($this,$canvas) {
 $this = $this|0;
 $canvas = $canvas|0;
 var $$0 = 0, $$01 = 0, $$pre$i$i$i = 0, $$pre$i$i$i$i = 0, $$pre$i$i$i8 = 0, $$pre$i11 = 0, $$pre$i4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $canvasref = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $canvasref = sp + 4|0;
 $0 = sp;
 $1 = ($canvas|0)==(0|0); //@line 1016 "modules/graphics/opengl/Graphics.cpp"
 $2 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($1) {
  $4 = HEAP32[158272]|0; //@line 1065 "modules/graphics/opengl/Graphics.cpp"
  $5 = ($4|0)==(0|0); //@line 1065 "modules/graphics/opengl/Graphics.cpp"
  if (!($5)) {
   __ZN4love8graphics6opengl6Canvas8stopGrabEb($4,0); //@line 1066 "modules/graphics/opengl/Graphics.cpp"
  }
  $6 = ((($3)) + -44|0); //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $7 = HEAP32[$6>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $8 = ((($3)) + -40|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $9 = HEAP32[$8>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $10 = ($9|0)==($7|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($10) {
   STACKTOP = sp;return; //@line 1027 "modules/graphics/opengl/Graphics.cpp"
  } else {
   $12 = $9;
  }
  while(1) {
   $11 = ((($12)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$8>>2] = $11; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $13 = HEAP32[$11>>2]|0; //@line 106 "./common/Object.h"
   $14 = ($13|0)==(0|0); //@line 106 "./common/Object.h"
   if ($14) {
    $16 = $11;
   } else {
    $17 = HEAP32[$13>>2]|0; //@line 106 "./common/Object.h"
    $18 = ((($17)) + 12|0); //@line 106 "./common/Object.h"
    $19 = HEAP32[$18>>2]|0; //@line 106 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($19|0,($13|0)); //@line 106 "./common/Object.h"
    $20 = __THREW__; __THREW__ = 0;
    $21 = $20&1;
    if ($21) {
     label = 9;
     break;
    }
    $$pre$i$i$i$i = HEAP32[$8>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $16 = $$pre$i$i$i$i;
   }
   $15 = ($16|0)==($7|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($15) {
    label = 24;
    break;
   } else {
    $12 = $16;
   }
  }
  if ((label|0) == 9) {
   $22 = ___cxa_find_matching_catch_3(0|0)|0;
   $23 = tempRet0;
   ___clang_call_terminate($22); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
  else if ((label|0) == 24) {
   STACKTOP = sp;return; //@line 1027 "modules/graphics/opengl/Graphics.cpp"
  }
 }
 __ZN4love8graphics6opengl6Canvas9startGrabEv($canvas); //@line 1021 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$canvasref>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $24 = ((($canvasref)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$24>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $25 = ((($canvasref)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$25>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$0>>2] = $canvas; //@line 93 "./common/Object.h"
 $26 = HEAP32[$canvas>>2]|0; //@line 95 "./common/Object.h"
 $27 = ((($26)) + 8|0); //@line 95 "./common/Object.h"
 $28 = HEAP32[$27>>2]|0; //@line 95 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($28|0,($canvas|0)); //@line 95 "./common/Object.h"
 $29 = __THREW__; __THREW__ = 0;
 $30 = $29&1;
 do {
  if ($30) {
   $68 = ___cxa_find_matching_catch_2()|0;
   $69 = tempRet0;
   $$0 = $69;$$01 = $68;
  } else {
   $31 = ((($canvasref)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __THREW__ = 0;
   invoke_vii(4049,($canvasref|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $32 = __THREW__; __THREW__ = 0;
   $33 = $32&1;
   if ($33) {
    $70 = ___cxa_find_matching_catch_2()|0;
    $71 = tempRet0;
    $72 = HEAP32[$canvas>>2]|0; //@line 106 "./common/Object.h"
    $73 = ((($72)) + 12|0); //@line 106 "./common/Object.h"
    $74 = HEAP32[$73>>2]|0; //@line 106 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($74|0,($canvas|0)); //@line 106 "./common/Object.h"
    $75 = __THREW__; __THREW__ = 0;
    $76 = $75&1;
    if (!($76)) {
     $$0 = $71;$$01 = $70;
     break;
    }
    $77 = ___cxa_find_matching_catch_3(0|0)|0;
    $78 = tempRet0;
    ___clang_call_terminate($77); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   $34 = HEAP32[$canvas>>2]|0; //@line 106 "./common/Object.h"
   $35 = ((($34)) + 12|0); //@line 106 "./common/Object.h"
   $36 = HEAP32[$35>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($36|0,($canvas|0)); //@line 106 "./common/Object.h"
   $37 = __THREW__; __THREW__ = 0;
   $38 = $37&1;
   if ($38) {
    $39 = ___cxa_find_matching_catch_3(0|0)|0;
    $40 = tempRet0;
    ___clang_call_terminate($39); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   $41 = ((($3)) + -44|0); //@line 1026 "modules/graphics/opengl/Graphics.cpp"
   $42 = HEAP32[$41>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $43 = HEAP32[$canvasref>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$41>>2] = $43; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$canvasref>>2] = $42; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $44 = ((($3)) + -40|0); //@line 2026 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $45 = HEAP32[$44>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $46 = HEAP32[$24>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$44>>2] = $46; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$24>>2] = $45; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $47 = ((($3)) + -36|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $48 = HEAP32[$47>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $49 = HEAP32[$31>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$47>>2] = $49; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$31>>2] = $48; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $50 = HEAP32[$canvasref>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $51 = ($50|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($51)) {
    $52 = HEAP32[$24>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $53 = ($52|0)==($50|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    do {
     if ($53) {
      $67 = $50;
     } else {
      $55 = $52;
      while(1) {
       $54 = ((($55)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$24>>2] = $54; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $56 = HEAP32[$54>>2]|0; //@line 106 "./common/Object.h"
       $57 = ($56|0)==(0|0); //@line 106 "./common/Object.h"
       if ($57) {
        $59 = $54;
       } else {
        $60 = HEAP32[$56>>2]|0; //@line 106 "./common/Object.h"
        $61 = ((($60)) + 12|0); //@line 106 "./common/Object.h"
        $62 = HEAP32[$61>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($62|0,($56|0)); //@line 106 "./common/Object.h"
        $63 = __THREW__; __THREW__ = 0;
        $64 = $63&1;
        if ($64) {
         label = 20;
         break;
        }
        $$pre$i$i$i = HEAP32[$24>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $59 = $$pre$i$i$i;
       }
       $58 = ($59|0)==($50|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($58) {
        label = 21;
        break;
       } else {
        $55 = $59;
       }
      }
      if ((label|0) == 20) {
       $65 = ___cxa_find_matching_catch_3(0|0)|0;
       $66 = tempRet0;
       ___clang_call_terminate($65); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      else if ((label|0) == 21) {
       $$pre$i4 = HEAP32[$canvasref>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $67 = $$pre$i4;
       break;
      }
     }
    } while(0);
    __ZdlPv($67); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   STACKTOP = sp;return; //@line 1027 "modules/graphics/opengl/Graphics.cpp"
  }
 } while(0);
 $79 = HEAP32[$canvasref>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $80 = ($79|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($80) {
  ___resumeException($$01|0);
  // unreachable;
 }
 $81 = HEAP32[$24>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $82 = ($81|0)==($79|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($82) {
   $96 = $79;
  } else {
   $84 = $81;
   while(1) {
    $83 = ((($84)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$24>>2] = $83; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $85 = HEAP32[$83>>2]|0; //@line 106 "./common/Object.h"
    $86 = ($85|0)==(0|0); //@line 106 "./common/Object.h"
    if ($86) {
     $88 = $83;
    } else {
     $89 = HEAP32[$85>>2]|0; //@line 106 "./common/Object.h"
     $90 = ((($89)) + 12|0); //@line 106 "./common/Object.h"
     $91 = HEAP32[$90>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($91|0,($85|0)); //@line 106 "./common/Object.h"
     $92 = __THREW__; __THREW__ = 0;
     $93 = $92&1;
     if ($93) {
      label = 34;
      break;
     }
     $$pre$i$i$i8 = HEAP32[$24>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $88 = $$pre$i$i$i8;
    }
    $87 = ($88|0)==($79|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($87) {
     label = 35;
     break;
    } else {
     $84 = $88;
    }
   }
   if ((label|0) == 34) {
    $94 = ___cxa_find_matching_catch_3(0|0)|0;
    $95 = tempRet0;
    ___clang_call_terminate($94); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 35) {
    $$pre$i11 = HEAP32[$canvasref>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $96 = $$pre$i11;
    break;
   }
  }
 } while(0);
 __ZdlPv($96); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$01|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love9StrongRefINS1_8graphics6opengl6CanvasEEENS_9allocatorIS6_EEE21__push_back_slow_pathIS6_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i8 = 0, $$0$i9 = 0, $$01$i$i = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre12 = 0, $$pre13 = 0, $$pre14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__v$sroa$15$0 = 0, $__v$sroa$6$0 = 0, $__v$sroa$6$0$lcssa = 0, $__v$sroa$6$1 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = $$in; //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = HEAP32[$8>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $10 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($11>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $17 = $10 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ($17>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $19 = $18 ? $5 : $17; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = (($20) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = $21 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = ($19|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $24 = $20; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($23) {
   $$0$i9 = 0;$101 = $24;$28 = 0;$29 = $22;
  } else {
   $$0$i8 = $19;$100 = $22;$99 = $24;
   label = 6;
  }
 } else {
  $13 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = $14 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = $13; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $$0$i8 = 1073741823;$100 = $15;$99 = $16;
  label = 6;
 }
 if ((label|0) == 6) {
  $25 = $$0$i8 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (__Znwj($25)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i9 = $$0$i8;$101 = $99;$28 = $26;$29 = $100;
 }
 $27 = (($28) + ($29<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = $27; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $31 = (($28) + ($$0$i9<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $32 = $31; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $33 = HEAP32[$__x>>2]|0; //@line 139 "./common/Object.h"
 HEAP32[$27>>2] = $33; //@line 99 "./common/Object.h"
 $34 = ($33|0)==(0|0); //@line 101 "./common/Object.h"
 do {
  if ($34) {
   $43 = $101;$44 = $7;
   label = 10;
  } else {
   $35 = HEAP32[$33>>2]|0; //@line 101 "./common/Object.h"
   $36 = ((($35)) + 8|0); //@line 101 "./common/Object.h"
   $37 = HEAP32[$36>>2]|0; //@line 101 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($37|0,($33|0)); //@line 101 "./common/Object.h"
   $38 = __THREW__; __THREW__ = 0;
   $39 = $38&1;
   if ($39) {
    $81 = ___cxa_find_matching_catch_2()|0;
    $82 = tempRet0;
    $__v$sroa$15$0 = $30;$__v$sroa$6$1 = $30;$lpad$phi$index = $81;$lpad$phi$index2 = $82;
    break;
   } else {
    $$pre12 = HEAP32[$this>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre13 = HEAP32[$0>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $43 = $$pre13;$44 = $$pre12;
    label = 10;
    break;
   }
  }
 } while(0);
 L15: do {
  if ((label|0) == 10) {
   $40 = ((($27)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $41 = $40; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $42 = ($43|0)==($44|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   do {
    if ($42) {
     $45 = $43; //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $60 = $30;$62 = $45;$64 = $43;
    } else {
     $$01$i$i = $43;$48 = $27;$__v$sroa$6$0 = $30;
     while(1) {
      $46 = ((($$01$i$i)) + -4|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $47 = ((($48)) + -4|0); //@line 99 "./common/Object.h"
      $49 = HEAP32[$46>>2]|0; //@line 139 "./common/Object.h"
      HEAP32[$47>>2] = $49; //@line 99 "./common/Object.h"
      $50 = ($49|0)==(0|0); //@line 101 "./common/Object.h"
      if (!($50)) {
       $51 = HEAP32[$49>>2]|0; //@line 101 "./common/Object.h"
       $52 = ((($51)) + 8|0); //@line 101 "./common/Object.h"
       $53 = HEAP32[$52>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($53|0,($49|0)); //@line 101 "./common/Object.h"
       $54 = __THREW__; __THREW__ = 0;
       $55 = $54&1;
       if ($55) {
        $__v$sroa$6$0$lcssa = $__v$sroa$6$0;
        label = 24;
        break;
       }
      }
      $56 = $__v$sroa$6$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $57 = ((($56)) + -4|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $58 = $57; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $59 = ($46|0)==($44|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      if ($59) {
       $$lcssa = $58;
       label = 15;
       break;
      } else {
       $$01$i$i = $46;$48 = $57;$__v$sroa$6$0 = $58;
      }
     }
     if ((label|0) == 15) {
      $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $$pre14 = HEAP32[$0>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $60 = $$lcssa;$62 = $$pre$i;$64 = $$pre14;
      break;
     }
     else if ((label|0) == 24) {
      $79 = ___cxa_find_matching_catch_2()|0;
      $80 = tempRet0;
      $__v$sroa$15$0 = $41;$__v$sroa$6$1 = $__v$sroa$6$0$lcssa;$lpad$phi$index = $79;$lpad$phi$index2 = $80;
      break L15;
     }
    }
   } while(0);
   HEAP32[$this>>2] = $60; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$0>>2] = $41; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$8>>2] = $32; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $61 = $62; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $63 = ($64|0)==($61|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   L28: do {
    if (!($63)) {
     $66 = $64;
     while(1) {
      $65 = ((($66)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $67 = HEAP32[$65>>2]|0; //@line 106 "./common/Object.h"
      $68 = ($67|0)==(0|0); //@line 106 "./common/Object.h"
      if (!($68)) {
       $70 = HEAP32[$67>>2]|0; //@line 106 "./common/Object.h"
       $71 = ((($70)) + 12|0); //@line 106 "./common/Object.h"
       $72 = HEAP32[$71>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($72|0,($67|0)); //@line 106 "./common/Object.h"
       $73 = __THREW__; __THREW__ = 0;
       $74 = $73&1;
       if ($74) {
        break;
       }
      }
      $69 = ($65|0)==($61|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if ($69) {
       break L28;
      } else {
       $66 = $65;
      }
     }
     $75 = ___cxa_find_matching_catch_3(0|0)|0;
     $76 = tempRet0;
     ___clang_call_terminate($75); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   } while(0);
   $77 = ($62|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($77) {
    return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $78 = $62; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __ZdlPv($78); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
 } while(0);
 $83 = $__v$sroa$6$1; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $84 = $__v$sroa$15$0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $85 = ($84|0)==($83|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 L41: do {
  if (!($85)) {
   $87 = $84;
   while(1) {
    $86 = ((($87)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $88 = HEAP32[$86>>2]|0; //@line 106 "./common/Object.h"
    $89 = ($88|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($89)) {
     $91 = HEAP32[$88>>2]|0; //@line 106 "./common/Object.h"
     $92 = ((($91)) + 12|0); //@line 106 "./common/Object.h"
     $93 = HEAP32[$92>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($93|0,($88|0)); //@line 106 "./common/Object.h"
     $94 = __THREW__; __THREW__ = 0;
     $95 = $94&1;
     if ($95) {
      break;
     }
    }
    $90 = ($86|0)==($83|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($90) {
     break L41;
    } else {
     $87 = $86;
    }
   }
   $96 = ___cxa_find_matching_catch_3(0|0)|0;
   $97 = tempRet0;
   ___clang_call_terminate($96); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 } while(0);
 $98 = ($28|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($98) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 __ZdlPv($28); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love9StrongRefINS1_8graphics6opengl6CanvasEEENS_9allocatorIS6_EEE7reserveEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$01$i$i = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $8 = 0, $9 = 0, $__v$sroa$6$0 = 0, $__v$sroa$6$0$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($4>>>0)<($__n>>>0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = $2; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($5)) {
  return; //@line 1545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $7 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = HEAP32[$7>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($__n|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $12 = $8; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $16 = 0;
 } else {
  $13 = $__n << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $14 = (__Znwj($13)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $16 = $14;
 }
 $15 = (($16) + ($10<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $17 = $15; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $18 = (($16) + ($__n<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $19 = $18; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $20 = ($12|0)==($6|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 do {
  if ($20) {
   $35 = $17;$37 = $2;$39 = $6;
  } else {
   $$01$i$i = $12;$23 = $15;$__v$sroa$6$0 = $17;
   while(1) {
    $21 = ((($$01$i$i)) + -4|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $22 = ((($23)) + -4|0); //@line 99 "./common/Object.h"
    $24 = HEAP32[$21>>2]|0; //@line 139 "./common/Object.h"
    HEAP32[$22>>2] = $24; //@line 99 "./common/Object.h"
    $25 = ($24|0)==(0|0); //@line 101 "./common/Object.h"
    if (!($25)) {
     $26 = HEAP32[$24>>2]|0; //@line 101 "./common/Object.h"
     $27 = ((($26)) + 8|0); //@line 101 "./common/Object.h"
     $28 = HEAP32[$27>>2]|0; //@line 101 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($28|0,($24|0)); //@line 101 "./common/Object.h"
     $29 = __THREW__; __THREW__ = 0;
     $30 = $29&1;
     if ($30) {
      $__v$sroa$6$0$lcssa = $__v$sroa$6$0;
      break;
     }
    }
    $31 = $__v$sroa$6$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $32 = ((($31)) + -4|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $33 = $32; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $34 = ($21|0)==($6|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    if ($34) {
     $$lcssa = $33;
     label = 8;
     break;
    } else {
     $$01$i$i = $21;$23 = $32;$__v$sroa$6$0 = $33;
    }
   }
   if ((label|0) == 8) {
    $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $$pre = HEAP32[$7>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $35 = $$lcssa;$37 = $$pre$i;$39 = $$pre;
    break;
   }
   $54 = ___cxa_find_matching_catch_2()|0;
   $55 = tempRet0;
   $56 = $__v$sroa$6$0$lcssa; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $57 = ($15|0)==($56|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   L16: do {
    if (!($57)) {
     $59 = $15;
     while(1) {
      $58 = ((($59)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $60 = HEAP32[$58>>2]|0; //@line 106 "./common/Object.h"
      $61 = ($60|0)==(0|0); //@line 106 "./common/Object.h"
      if (!($61)) {
       $63 = HEAP32[$60>>2]|0; //@line 106 "./common/Object.h"
       $64 = ((($63)) + 12|0); //@line 106 "./common/Object.h"
       $65 = HEAP32[$64>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($65|0,($60|0)); //@line 106 "./common/Object.h"
       $66 = __THREW__; __THREW__ = 0;
       $67 = $66&1;
       if ($67) {
        break;
       }
      }
      $62 = ($58|0)==($56|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if ($62) {
       break L16;
      } else {
       $59 = $58;
      }
     }
     $68 = ___cxa_find_matching_catch_3(0|0)|0;
     $69 = tempRet0;
     ___clang_call_terminate($68); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   } while(0);
   $70 = ($16|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($70) {
    ___resumeException($54|0);
    // unreachable;
   }
   __ZdlPv($16); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($54|0);
   // unreachable;
  }
 } while(0);
 HEAP32[$this>>2] = $35; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $17; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $19; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $36 = $37; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $38 = ($39|0)==($36|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 L29: do {
  if (!($38)) {
   $41 = $39;
   while(1) {
    $40 = ((($41)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $42 = HEAP32[$40>>2]|0; //@line 106 "./common/Object.h"
    $43 = ($42|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($43)) {
     $45 = HEAP32[$42>>2]|0; //@line 106 "./common/Object.h"
     $46 = ((($45)) + 12|0); //@line 106 "./common/Object.h"
     $47 = HEAP32[$46>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($47|0,($42|0)); //@line 106 "./common/Object.h"
     $48 = __THREW__; __THREW__ = 0;
     $49 = $48&1;
     if ($49) {
      break;
     }
    }
    $44 = ($40|0)==($36|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($44) {
     break L29;
    } else {
     $41 = $40;
    }
   }
   $50 = ___cxa_find_matching_catch_3(0|0)|0;
   $51 = tempRet0;
   ___clang_call_terminate($50); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 } while(0);
 $52 = ($37|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($52) {
  return; //@line 1545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $53 = $37; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($53); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl8Graphics7setModeEii($this,$width,$height) {
 $this = $this|0;
 $width = $width|0;
 $height = $height|0;
 var $$in = 0, $$pre = 0, $$pre11 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $enabledebug$0$off0 = 0, $flags = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $flags = sp + 8|0;
 $0 = sp;
 $1 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $2 = ($1|0)==(0|0); //@line 132 "./common/Object.h"
 if (!($2)) {
  $3 = HEAP32[$1>>2]|0; //@line 132 "./common/Object.h"
  $4 = ((($3)) + 8|0); //@line 132 "./common/Object.h"
  $5 = HEAP32[$4>>2]|0; //@line 132 "./common/Object.h"
  FUNCTION_TABLE_vi[$5 & 8191]($1); //@line 132 "./common/Object.h"
 }
 $6 = ((($this)) + 8|0); //@line 133 "./common/Object.h"
 $7 = HEAP32[$6>>2]|0; //@line 133 "./common/Object.h"
 $8 = ($7|0)==(0|0); //@line 133 "./common/Object.h"
 if (!($8)) {
  $9 = HEAP32[$7>>2]|0; //@line 133 "./common/Object.h"
  $10 = ((($9)) + 12|0); //@line 133 "./common/Object.h"
  $11 = HEAP32[$10>>2]|0; //@line 133 "./common/Object.h"
  FUNCTION_TABLE_vi[$11 & 8191]($7); //@line 133 "./common/Object.h"
 }
 HEAP32[$6>>2] = $1; //@line 134 "./common/Object.h"
 $12 = ((($this)) + 32|0); //@line 247 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$12>>2] = $width; //@line 247 "modules/graphics/opengl/Graphics.cpp"
 $13 = ((($this)) + 36|0); //@line 248 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$13>>2] = $height; //@line 248 "modules/graphics/opengl/Graphics.cpp"
 (__ZN4love8graphics6opengl6OpenGL11initContextEv(634044)|0); //@line 251 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl6OpenGL12setupContextEv(634044); //@line 252 "modules/graphics/opengl/Graphics.cpp"
 $14 = ((($this)) + 40|0); //@line 254 "modules/graphics/opengl/Graphics.cpp"
 HEAP8[$14>>0] = 1; //@line 254 "modules/graphics/opengl/Graphics.cpp"
 $15 = HEAP32[$this>>2]|0; //@line 256 "modules/graphics/opengl/Graphics.cpp"
 $16 = ((($15)) + 24|0); //@line 256 "modules/graphics/opengl/Graphics.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 256 "modules/graphics/opengl/Graphics.cpp"
 FUNCTION_TABLE_viii[$17 & 8191]($this,$width,$height); //@line 256 "modules/graphics/opengl/Graphics.cpp"
 $18 = HEAP32[164832]|0; //@line 57 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$18 & 8191](3042); //@line 57 "./libraries/glad/gladfuncs.hpp"
 $19 = HEAP32[165134]|0; //@line 37 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$19 & 8191](33170,4354); //@line 37 "./libraries/glad/gladfuncs.hpp"
 $20 = HEAP8[1252012]|0; //@line 264 "modules/graphics/opengl/Graphics.cpp"
 $21 = ($20<<24>>24)==(0); //@line 264 "modules/graphics/opengl/Graphics.cpp"
 if ($21) {
  $22 = HEAP32[164832]|0; //@line 57 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$22 & 8191](32925); //@line 57 "./libraries/glad/gladfuncs.hpp"
  $23 = HEAP32[164832]|0; //@line 57 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$23 & 8191](3553); //@line 57 "./libraries/glad/gladfuncs.hpp"
 }
 __ZN4love8graphics6opengl6OpenGL14setTextureUnitEi(634044,0); //@line 273 "modules/graphics/opengl/Graphics.cpp"
 $24 = HEAP32[165018]|0; //@line 66 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$24 & 8191](3317,1); //@line 66 "./libraries/glad/gladfuncs.hpp"
 $25 = HEAP8[1252002]|0; //@line 279 "modules/graphics/opengl/Graphics.cpp"
 $26 = HEAP8[1252066]|0; //@line 279 "modules/graphics/opengl/Graphics.cpp"
 $27 = $26 | $25; //@line 279 "modules/graphics/opengl/Graphics.cpp"
 $28 = HEAP8[1252186]|0; //@line 279 "modules/graphics/opengl/Graphics.cpp"
 $29 = $27 | $28; //@line 279 "modules/graphics/opengl/Graphics.cpp"
 $30 = HEAP8[1252013]|0; //@line 280 "modules/graphics/opengl/Graphics.cpp"
 $31 = $29 | $30; //@line 279 "modules/graphics/opengl/Graphics.cpp"
 $32 = HEAP8[1252224]|0; //@line 280 "modules/graphics/opengl/Graphics.cpp"
 $33 = $31 | $32; //@line 279 "modules/graphics/opengl/Graphics.cpp"
 $34 = ($33<<24>>24)==(0); //@line 279 "modules/graphics/opengl/Graphics.cpp"
 if ($34) {
  __ZN4love8graphics15setGammaCorrectEb(0); //@line 286 "modules/graphics/opengl/Graphics.cpp"
 } else {
  $35 = HEAP8[1251994]|0; //@line 282 "modules/graphics/opengl/Graphics.cpp"
  $36 = HEAP8[1252225]|0; //@line 282 "modules/graphics/opengl/Graphics.cpp"
  $37 = $36 | $35; //@line 282 "modules/graphics/opengl/Graphics.cpp"
  $38 = ($37<<24>>24)==(0); //@line 282 "modules/graphics/opengl/Graphics.cpp"
  if (!($38)) {
   $39 = (__ZN4love8graphics14isGammaCorrectEv()|0); //@line 944 "modules/graphics/opengl/Graphics.cpp"
   __ZN4love8graphics6opengl6OpenGL18setFramebufferSRGBEb(634044,$39); //@line 283 "modules/graphics/opengl/Graphics.cpp"
  }
 }
 $40 = (__ZN4love8graphics14isGammaCorrectEv()|0); //@line 944 "modules/graphics/opengl/Graphics.cpp"
 $41 = $40&1; //@line 288 "modules/graphics/opengl/Graphics.cpp"
 HEAP8[1251895] = $41; //@line 288 "modules/graphics/opengl/Graphics.cpp"
 $42 = HEAP8[1252002]|0; //@line 292 "modules/graphics/opengl/Graphics.cpp"
 $43 = ($42<<24>>24)==(0); //@line 292 "modules/graphics/opengl/Graphics.cpp"
 if ($43) {
  $enabledebug$0$off0 = 0;
 } else {
  HEAP32[$flags>>2] = 0; //@line 295 "modules/graphics/opengl/Graphics.cpp"
  $44 = HEAP32[165587]|0; //@line 73 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$44 & 8191](33310,$flags); //@line 73 "./libraries/glad/gladfuncs.hpp"
  $45 = HEAP32[$flags>>2]|0; //@line 297 "modules/graphics/opengl/Graphics.cpp"
  $46 = $45 & 2; //@line 297 "modules/graphics/opengl/Graphics.cpp"
  $47 = ($46|0)!=(0); //@line 297 "modules/graphics/opengl/Graphics.cpp"
  $enabledebug$0$off0 = $47;
 }
 __ZN4love8graphics6opengl8Graphics8setDebugEb(0,$enabledebug$0$off0); //@line 300 "modules/graphics/opengl/Graphics.cpp"
 $48 = (__ZN4love8graphics8Volatile7loadAllEv()|0); //@line 303 "modules/graphics/opengl/Graphics.cpp"
 if (!($48)) {
  (_puts(434963)|0); //@line 304 "modules/graphics/opengl/Graphics.cpp"
 }
 $49 = ((($this)) + 28|0); //@line 311 "modules/graphics/opengl/Graphics.cpp"
 $50 = HEAP32[$49>>2]|0; //@line 311 "modules/graphics/opengl/Graphics.cpp"
 $51 = ($50|0)==(0|0); //@line 311 "modules/graphics/opengl/Graphics.cpp"
 do {
  if ($51) {
   $52 = (__Znwj(4)|0); //@line 312 "modules/graphics/opengl/Graphics.cpp"
   __THREW__ = 0;
   invoke_vii(4020,($52|0),20); //@line 312 "modules/graphics/opengl/Graphics.cpp"
   $53 = __THREW__; __THREW__ = 0;
   $54 = $53&1;
   if ($54) {
    $55 = ___cxa_find_matching_catch_2()|0;
    $56 = tempRet0;
    __ZdlPv($52); //@line 312 "modules/graphics/opengl/Graphics.cpp"
    ___resumeException($55|0);
    // unreachable;
   } else {
    HEAP32[$49>>2] = $52; //@line 312 "modules/graphics/opengl/Graphics.cpp"
    break;
   }
  }
 } while(0);
 $57 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $58 = HEAP32[$57>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $59 = ((($58)) + -136|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 __ZN4love8graphics6opengl8Graphics12restoreStateERKNS2_12DisplayStateE($this,$59); //@line 315 "modules/graphics/opengl/Graphics.cpp"
 $60 = ((($this)) + 16|0); //@line 317 "modules/graphics/opengl/Graphics.cpp"
 $61 = HEAP32[$60>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $62 = ((($this)) + 20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $63 = HEAP32[$62>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $64 = ($63|0)==($61|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $65 = $61; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($64) {
  $$in = $63;$117 = $61;
 } else {
  $scevgep$i$i$i = ((($63)) + -8|0);
  $66 = $scevgep$i$i$i;
  $67 = (($66) - ($65))|0;
  $68 = $67 >>> 3;
  $69 = $68 ^ -1;
  $scevgep4$i$i$i = (($63) + ($69<<3)|0);
  HEAP32[$62>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $scevgep4$i$i$i;$117 = $scevgep4$i$i$i;
 }
 $70 = ((($this)) + 24|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $71 = HEAP32[$70>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $72 = (($71) - ($65))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $73 = $72 >> 3; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $74 = ($73>>>0)<(5); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $75 = $71; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($74) {
  $76 = $$in; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $77 = (($76) - ($65))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $78 = $77 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $79 = (__Znwj(40)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $80 = (($79) + ($78<<3)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $81 = $80; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $82 = ((($79)) + 40|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $83 = $82; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $84 = (0 - ($78))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $85 = (($80) + ($84<<3)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $86 = $85; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  _memcpy(($85|0),($61|0),($77|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$60>>2] = $86; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$62>>2] = $81; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$70>>2] = $83; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $87 = ($61|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($87) {
   $89 = $80;$90 = $82;
  } else {
   __ZdlPv($61); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $$pre = HEAP32[$62>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$pre11 = HEAP32[$70>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $89 = $$pre;$90 = $$pre11;
  }
 } else {
  $89 = $117;$90 = $75;
 }
 HEAPF64[$0>>3] = 1.0; //@line 319 "modules/graphics/opengl/Graphics.cpp"
 $88 = ($89>>>0)<($90>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($88) {
  HEAPF64[$89>>3] = 1.0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $91 = ((($89)) + 8|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$62>>2] = $91; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 } else {
  __ZNSt3__16vectorIdNS_9allocatorIdEEE21__push_back_slow_pathIdEEvOT_($60,$0); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $92 = HEAP32[158582]|0; //@line 322 "modules/graphics/opengl/Graphics.cpp"
 $93 = ($92|0)==(0|0); //@line 322 "modules/graphics/opengl/Graphics.cpp"
 do {
  if ($93) {
   $94 = HEAP8[1252012]|0; //@line 324 "modules/graphics/opengl/Graphics.cpp"
   $95 = ($94<<24>>24)!=(0); //@line 324 "modules/graphics/opengl/Graphics.cpp"
   $96 = $95&1; //@line 324 "modules/graphics/opengl/Graphics.cpp"
   $97 = (634336 + (($96*24)|0)|0); //@line 325 "modules/graphics/opengl/Graphics.cpp"
   $98 = (__Znwj(328)|0); //@line 909 "modules/graphics/opengl/Graphics.cpp"
   __THREW__ = 0;
   invoke_vii(4050,($98|0),($97|0)); //@line 909 "modules/graphics/opengl/Graphics.cpp"
   $99 = __THREW__; __THREW__ = 0;
   $100 = $99&1;
   if ($100) {
    $101 = ___cxa_find_matching_catch_2()|0;
    $102 = tempRet0;
    __ZdlPv($98); //@line 909 "modules/graphics/opengl/Graphics.cpp"
    ___resumeException($101|0);
    // unreachable;
   } else {
    HEAP32[158582] = $98; //@line 325 "modules/graphics/opengl/Graphics.cpp"
    break;
   }
  }
 } while(0);
 $103 = HEAP32[158583]|0; //@line 329 "modules/graphics/opengl/Graphics.cpp"
 $104 = ($103|0)==(0|0); //@line 329 "modules/graphics/opengl/Graphics.cpp"
 do {
  if ($104) {
   $105 = HEAP8[1252012]|0; //@line 331 "modules/graphics/opengl/Graphics.cpp"
   $106 = ($105<<24>>24)!=(0); //@line 331 "modules/graphics/opengl/Graphics.cpp"
   $107 = $106&1; //@line 331 "modules/graphics/opengl/Graphics.cpp"
   $108 = (634384 + (($107*24)|0)|0); //@line 332 "modules/graphics/opengl/Graphics.cpp"
   $109 = (__Znwj(328)|0); //@line 909 "modules/graphics/opengl/Graphics.cpp"
   __THREW__ = 0;
   invoke_vii(4050,($109|0),($108|0)); //@line 909 "modules/graphics/opengl/Graphics.cpp"
   $110 = __THREW__; __THREW__ = 0;
   $111 = $110&1;
   if ($111) {
    $112 = ___cxa_find_matching_catch_2()|0;
    $113 = tempRet0;
    __ZdlPv($109); //@line 909 "modules/graphics/opengl/Graphics.cpp"
    ___resumeException($112|0);
    // unreachable;
   } else {
    HEAP32[158583] = $109; //@line 332 "modules/graphics/opengl/Graphics.cpp"
    break;
   }
  }
 } while(0);
 $114 = HEAP32[158581]|0; //@line 337 "modules/graphics/opengl/Graphics.cpp"
 $115 = ($114|0)==(0|0); //@line 337 "modules/graphics/opengl/Graphics.cpp"
 if (!($115)) {
  STACKTOP = sp;return 1; //@line 340 "modules/graphics/opengl/Graphics.cpp"
 }
 $116 = HEAP32[158582]|0; //@line 338 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl6Shader6attachEb($116,0); //@line 338 "modules/graphics/opengl/Graphics.cpp"
 STACKTOP = sp;return 1; //@line 340 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics8setDebugEb($this,$enable) {
 $this = $this|0;
 $enable = $enable|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond13 = 0, $or$cond3$not = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[1252009]|0; //@line 389 "modules/graphics/opengl/Graphics.cpp"
 $1 = HEAP8[1252019]|0; //@line 389 "modules/graphics/opengl/Graphics.cpp"
 $2 = $1 | $0; //@line 389 "modules/graphics/opengl/Graphics.cpp"
 $3 = HEAP8[1252045]|0; //@line 389 "modules/graphics/opengl/Graphics.cpp"
 $4 = $2 | $3; //@line 389 "modules/graphics/opengl/Graphics.cpp"
 $or$cond3$not = ($4<<24>>24)==(0); //@line 389 "modules/graphics/opengl/Graphics.cpp"
 $5 = HEAP8[1252012]|0; //@line 393 "modules/graphics/opengl/Graphics.cpp"
 $6 = ($5<<24>>24)!=(0); //@line 393 "modules/graphics/opengl/Graphics.cpp"
 $or$cond13 = $or$cond3$not | $6; //@line 389 "modules/graphics/opengl/Graphics.cpp"
 if ($or$cond13) {
  return; //@line 431 "modules/graphics/opengl/Graphics.cpp"
 }
 $7 = ($3<<24>>24)==(0); //@line 397 "modules/graphics/opengl/Graphics.cpp"
 $8 = ($2<<24>>24)!=(0); //@line 397 "modules/graphics/opengl/Graphics.cpp"
 $9 = $8 | $7; //@line 397 "modules/graphics/opengl/Graphics.cpp"
 if (!($9)) {
  $10 = HEAP32[165702]|0; //@line 399 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[165100] = $10; //@line 399 "modules/graphics/opengl/Graphics.cpp"
  $11 = HEAP32[165700]|0; //@line 400 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[164693] = $11; //@line 400 "modules/graphics/opengl/Graphics.cpp"
 }
 if ($enable) {
  $18 = HEAP32[164832]|0; //@line 57 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$18 & 8191](33346); //@line 57 "./libraries/glad/gladfuncs.hpp"
  $19 = HEAP32[165100]|0; //@line 970 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$19 & 8191](4051,0); //@line 970 "./libraries/glad/gladfuncs.hpp"
  $20 = HEAP32[164693]|0; //@line 968 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiiiii[$20 & 8191](4352,4352,4352,0,0,1); //@line 968 "./libraries/glad/gladfuncs.hpp"
  $21 = HEAP32[164693]|0; //@line 968 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiiiii[$21 & 8191](33350,33357,4352,0,0,0); //@line 968 "./libraries/glad/gladfuncs.hpp"
  $22 = HEAP32[164693]|0; //@line 968 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiiiii[$22 & 8191](33352,33357,4352,0,0,0); //@line 968 "./libraries/glad/gladfuncs.hpp"
  $23 = HEAP8[1252009]|0; //@line 427 "modules/graphics/opengl/Graphics.cpp"
  $24 = HEAP8[1252019]|0; //@line 427 "modules/graphics/opengl/Graphics.cpp"
  $25 = $24 | $23; //@line 427 "modules/graphics/opengl/Graphics.cpp"
  $26 = ($25<<24>>24)==(0); //@line 427 "modules/graphics/opengl/Graphics.cpp"
  if (!($26)) {
   $27 = HEAP32[164832]|0; //@line 57 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_vi[$27 & 8191](37600); //@line 57 "./libraries/glad/gladfuncs.hpp"
  }
  (_puts(434911)|0); //@line 430 "modules/graphics/opengl/Graphics.cpp"
  return; //@line 431 "modules/graphics/opengl/Graphics.cpp"
 } else {
  $12 = HEAP32[165100]|0; //@line 970 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$12 & 8191](0,0); //@line 970 "./libraries/glad/gladfuncs.hpp"
  $13 = HEAP8[1252009]|0; //@line 409 "modules/graphics/opengl/Graphics.cpp"
  $14 = HEAP8[1252019]|0; //@line 409 "modules/graphics/opengl/Graphics.cpp"
  $15 = $14 | $13; //@line 409 "modules/graphics/opengl/Graphics.cpp"
  $16 = ($15<<24>>24)==(0); //@line 409 "modules/graphics/opengl/Graphics.cpp"
  if ($16) {
   return; //@line 431 "modules/graphics/opengl/Graphics.cpp"
  }
  $17 = HEAP32[165193]|0; //@line 56 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$17 & 8191](37600); //@line 56 "./libraries/glad/gladfuncs.hpp"
  return; //@line 431 "modules/graphics/opengl/Graphics.cpp"
 }
}
function __ZN4love8graphics6openglL7debugCBEjjjjiPKcPKv($source,$type,$id,$severity,$0,$msg,$1) {
 $source = $source|0;
 $type = $type|0;
 $id = $id|0;
 $severity = $severity|0;
 $0 = $0|0;
 $msg = $msg|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0, $vararg_ptr4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 374 "modules/graphics/opengl/Graphics.cpp"
 $2 = (__ZN4love8graphics6opengl6OpenGL17debugSourceStringEj($source)|0); //@line 377 "modules/graphics/opengl/Graphics.cpp"
 $3 = (__ZN4love8graphics6opengl6OpenGL15debugTypeStringEj($type)|0); //@line 378 "modules/graphics/opengl/Graphics.cpp"
 $4 = (__ZN4love8graphics6opengl6OpenGL19debugSeverityStringEj($severity)|0); //@line 379 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$vararg_buffer>>2] = $msg; //@line 382 "modules/graphics/opengl/Graphics.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 382 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$vararg_ptr1>>2] = $2; //@line 382 "modules/graphics/opengl/Graphics.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 382 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$vararg_ptr2>>2] = $3; //@line 382 "modules/graphics/opengl/Graphics.cpp"
 $vararg_ptr3 = ((($vararg_buffer)) + 12|0); //@line 382 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$vararg_ptr3>>2] = $4; //@line 382 "modules/graphics/opengl/Graphics.cpp"
 $vararg_ptr4 = ((($vararg_buffer)) + 16|0); //@line 382 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$vararg_ptr4>>2] = $id; //@line 382 "modules/graphics/opengl/Graphics.cpp"
 (_printf(434858,$vararg_buffer)|0); //@line 382 "modules/graphics/opengl/Graphics.cpp"
 STACKTOP = sp;return; //@line 383 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics12restoreStateERKNS2_12DisplayStateE($this,$s) {
 $this = $this|0;
 $s = $s|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0.0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0.0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0.0;
 var $152 = 0, $153 = 0, $154 = 0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0.0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0.0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0.0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0.0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $glcompare$0$i = 0, $nc$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 32|0;
 $1 = sp;
 $nc$i = sp + 16|0;
 $2 = +HEAPF32[$s>>2]; //@line 106 "modules/graphics/opengl/Graphics.cpp"
 $3 = ((($s)) + 4|0); //@line 106 "modules/graphics/opengl/Graphics.cpp"
 $4 = +HEAPF32[$3>>2]; //@line 106 "modules/graphics/opengl/Graphics.cpp"
 $5 = ((($s)) + 8|0); //@line 106 "modules/graphics/opengl/Graphics.cpp"
 $6 = +HEAPF32[$5>>2]; //@line 106 "modules/graphics/opengl/Graphics.cpp"
 $7 = ((($s)) + 12|0); //@line 106 "modules/graphics/opengl/Graphics.cpp"
 $8 = +HEAPF32[$7>>2]; //@line 106 "modules/graphics/opengl/Graphics.cpp"
 $9 = $2 / 255.0; //@line 949 "modules/graphics/opengl/Graphics.cpp"
 $10 = $4 / 255.0; //@line 949 "modules/graphics/opengl/Graphics.cpp"
 $11 = $6 / 255.0; //@line 949 "modules/graphics/opengl/Graphics.cpp"
 $12 = $8 / 255.0; //@line 949 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$nc$i>>2] = $9; //@line 38 "./modules/graphics/Color.h"
 $13 = ((($nc$i)) + 4|0); //@line 38 "./modules/graphics/Color.h"
 HEAPF32[$13>>2] = $10; //@line 38 "./modules/graphics/Color.h"
 $14 = ((($nc$i)) + 8|0); //@line 38 "./modules/graphics/Color.h"
 HEAPF32[$14>>2] = $11; //@line 38 "./modules/graphics/Color.h"
 $15 = ((($nc$i)) + 12|0); //@line 38 "./modules/graphics/Color.h"
 HEAPF32[$15>>2] = $12; //@line 38 "./modules/graphics/Color.h"
 __ZN4love8graphics17gammaCorrectColorERNS0_6ColorTIfEE($nc$i); //@line 951 "modules/graphics/opengl/Graphics.cpp"
 $16 = +HEAPF32[$nc$i>>2]; //@line 953 "modules/graphics/opengl/Graphics.cpp"
 $17 = +HEAPF32[$13>>2]; //@line 953 "modules/graphics/opengl/Graphics.cpp"
 $18 = +HEAPF32[$14>>2]; //@line 953 "modules/graphics/opengl/Graphics.cpp"
 $19 = +HEAPF32[$15>>2]; //@line 953 "modules/graphics/opengl/Graphics.cpp"
 $20 = HEAP32[165163]|0; //@line 583 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vidddd[$20 & 8191](3,$16,$17,$18,$19); //@line 583 "./libraries/glad/gladfuncs.hpp"
 $21 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $22 = HEAP32[$21>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $23 = ((($22)) + -136|0); //@line 954 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$23>>2] = $2; //@line 954 "modules/graphics/opengl/Graphics.cpp"
 $24 = ((($22)) + -132|0); //@line 954 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$24>>2] = $4; //@line 954 "modules/graphics/opengl/Graphics.cpp"
 $25 = ((($22)) + -128|0); //@line 954 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$25>>2] = $6; //@line 954 "modules/graphics/opengl/Graphics.cpp"
 $26 = ((($22)) + -124|0); //@line 954 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$26>>2] = $8; //@line 954 "modules/graphics/opengl/Graphics.cpp"
 $27 = ((($s)) + 16|0); //@line 107 "modules/graphics/opengl/Graphics.cpp"
 ;HEAP32[$1>>2]=HEAP32[$27>>2]|0;HEAP32[$1+4>>2]=HEAP32[$27+4>>2]|0;HEAP32[$1+8>>2]=HEAP32[$27+8>>2]|0;HEAP32[$1+12>>2]=HEAP32[$27+12>>2]|0;
 $28 = HEAP32[$21>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $29 = ((($28)) + -120|0); //@line 964 "modules/graphics/opengl/Graphics.cpp"
 ;HEAP32[$29>>2]=HEAP32[$1>>2]|0;HEAP32[$29+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$29+8>>2]=HEAP32[$1+8>>2]|0;HEAP32[$29+12>>2]=HEAP32[$1+12>>2]|0; //@line 964 "modules/graphics/opengl/Graphics.cpp"
 $30 = ((($s)) + 32|0); //@line 109 "modules/graphics/opengl/Graphics.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 109 "modules/graphics/opengl/Graphics.cpp"
 $32 = ((($s)) + 36|0); //@line 109 "modules/graphics/opengl/Graphics.cpp"
 $33 = HEAP32[$32>>2]|0; //@line 109 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl8Graphics12setBlendModeENS0_8Graphics9BlendModeENS3_10BlendAlphaE($this,$31,$33); //@line 109 "modules/graphics/opengl/Graphics.cpp"
 $34 = ((($s)) + 40|0); //@line 111 "modules/graphics/opengl/Graphics.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 111 "modules/graphics/opengl/Graphics.cpp"
 $36 = HEAP32[$21>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $37 = ((($36)) + -96|0); //@line 1202 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$37>>2] = $35; //@line 1202 "modules/graphics/opengl/Graphics.cpp"
 $38 = ((($s)) + 44|0); //@line 112 "modules/graphics/opengl/Graphics.cpp"
 $39 = HEAP32[$38>>2]|0; //@line 112 "modules/graphics/opengl/Graphics.cpp"
 $40 = ((($36)) + -92|0); //@line 1207 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$40>>2] = $39; //@line 1207 "modules/graphics/opengl/Graphics.cpp"
 $41 = ((($s)) + 48|0); //@line 113 "modules/graphics/opengl/Graphics.cpp"
 $42 = HEAP32[$41>>2]|0; //@line 113 "modules/graphics/opengl/Graphics.cpp"
 $43 = ((($36)) + -88|0); //@line 1212 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$43>>2] = $42; //@line 1212 "modules/graphics/opengl/Graphics.cpp"
 $44 = ((($s)) + 52|0); //@line 115 "modules/graphics/opengl/Graphics.cpp"
 $45 = +HEAPF32[$44>>2]; //@line 115 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl6OpenGL12setPointSizeEf(634044,$45); //@line 1232 "modules/graphics/opengl/Graphics.cpp"
 $46 = HEAP32[$21>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $47 = ((($46)) + -84|0); //@line 1233 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$47>>2] = $45; //@line 1233 "modules/graphics/opengl/Graphics.cpp"
 $48 = ((($s)) + 56|0); //@line 117 "modules/graphics/opengl/Graphics.cpp"
 $49 = HEAP8[$48>>0]|0; //@line 117 "modules/graphics/opengl/Graphics.cpp"
 $50 = ($49<<24>>24)==(0); //@line 117 "modules/graphics/opengl/Graphics.cpp"
 if ($50) {
  $69 = ((($46)) + -80|0); //@line 658 "modules/graphics/opengl/Graphics.cpp"
  HEAP8[$69>>0] = 0; //@line 658 "modules/graphics/opengl/Graphics.cpp"
  $70 = HEAP32[165193]|0; //@line 56 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$70 & 8191](3089); //@line 56 "./libraries/glad/gladfuncs.hpp"
 } else {
  $51 = ((($s)) + 60|0); //@line 118 "modules/graphics/opengl/Graphics.cpp"
  $52 = HEAP32[$51>>2]|0; //@line 118 "modules/graphics/opengl/Graphics.cpp"
  $53 = ((($s)) + 64|0); //@line 118 "modules/graphics/opengl/Graphics.cpp"
  $54 = HEAP32[$53>>2]|0; //@line 118 "modules/graphics/opengl/Graphics.cpp"
  $55 = ((($s)) + 68|0); //@line 118 "modules/graphics/opengl/Graphics.cpp"
  $56 = HEAP32[$55>>2]|0; //@line 118 "modules/graphics/opengl/Graphics.cpp"
  $57 = ((($s)) + 72|0); //@line 118 "modules/graphics/opengl/Graphics.cpp"
  $58 = HEAP32[$57>>2]|0; //@line 118 "modules/graphics/opengl/Graphics.cpp"
  $59 = HEAP32[164832]|0; //@line 57 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$59 & 8191](3089); //@line 57 "./libraries/glad/gladfuncs.hpp"
  HEAP32[$0>>2] = $52; //@line 629 "modules/graphics/opengl/Graphics.cpp"
  $60 = ((($0)) + 4|0); //@line 629 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$60>>2] = $54; //@line 629 "modules/graphics/opengl/Graphics.cpp"
  $61 = ((($0)) + 8|0); //@line 629 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$61>>2] = $56; //@line 629 "modules/graphics/opengl/Graphics.cpp"
  $62 = ((($0)) + 12|0); //@line 629 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$62>>2] = $58; //@line 629 "modules/graphics/opengl/Graphics.cpp"
  __ZN4love8graphics6opengl6OpenGL10setScissorERKNS2_8ViewportE(634044,$0); //@line 629 "modules/graphics/opengl/Graphics.cpp"
  $63 = HEAP32[$21>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $64 = ((($63)) + -80|0); //@line 631 "modules/graphics/opengl/Graphics.cpp"
  HEAP8[$64>>0] = 1; //@line 631 "modules/graphics/opengl/Graphics.cpp"
  $65 = ((($63)) + -76|0); //@line 632 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$65>>2] = $52; //@line 632 "modules/graphics/opengl/Graphics.cpp"
  $66 = ((($63)) + -72|0); //@line 632 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$66>>2] = $54; //@line 632 "modules/graphics/opengl/Graphics.cpp"
  $67 = ((($63)) + -68|0); //@line 632 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$67>>2] = $56; //@line 632 "modules/graphics/opengl/Graphics.cpp"
  $68 = ((($63)) + -64|0); //@line 632 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$68>>2] = $58; //@line 632 "modules/graphics/opengl/Graphics.cpp"
 }
 $71 = ((($s)) + 76|0); //@line 122 "modules/graphics/opengl/Graphics.cpp"
 $72 = HEAP32[$71>>2]|0; //@line 122 "modules/graphics/opengl/Graphics.cpp"
 $73 = ((($s)) + 80|0); //@line 122 "modules/graphics/opengl/Graphics.cpp"
 $74 = HEAP32[$73>>2]|0; //@line 122 "modules/graphics/opengl/Graphics.cpp"
 $75 = HEAP32[$21>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $76 = ((($75)) + -60|0); //@line 735 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$76>>2] = $72; //@line 735 "modules/graphics/opengl/Graphics.cpp"
 $77 = ((($75)) + -56|0); //@line 736 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$77>>2] = $74; //@line 736 "modules/graphics/opengl/Graphics.cpp"
 $78 = ((($this)) + 42|0); //@line 738 "modules/graphics/opengl/Graphics.cpp"
 $79 = HEAP8[$78>>0]|0; //@line 738 "modules/graphics/opengl/Graphics.cpp"
 $80 = ($79<<24>>24)==(0); //@line 738 "modules/graphics/opengl/Graphics.cpp"
 do {
  if ($80) {
   $81 = ($72|0)==(6); //@line 741 "modules/graphics/opengl/Graphics.cpp"
   if ($81) {
    $82 = HEAP32[165193]|0; //@line 56 "./libraries/glad/gladfuncs.hpp"
    FUNCTION_TABLE_vi[$82 & 8191](2960); //@line 56 "./libraries/glad/gladfuncs.hpp"
    break;
   }
   $83 = HEAP32[158272]|0; //@line 748 "modules/graphics/opengl/Graphics.cpp"
   $84 = ($83|0)==(0|0); //@line 748 "modules/graphics/opengl/Graphics.cpp"
   if (!($84)) {
    (__ZN4love8graphics6opengl6Canvas18checkCreateStencilEv($83)|0); //@line 749 "modules/graphics/opengl/Graphics.cpp"
   }
   switch ($72|0) {
   case 0:  {
    $glcompare$0$i = 516;
    break;
   }
   case 1:  {
    $glcompare$0$i = 518;
    break;
   }
   case 5:  {
    $glcompare$0$i = 517;
    break;
   }
   case 3:  {
    $glcompare$0$i = 515;
    break;
   }
   case 4:  {
    $glcompare$0$i = 513;
    break;
   }
   default: {
    $glcompare$0$i = 514;
   }
   }
   $85 = HEAP32[164832]|0; //@line 57 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_vi[$85 & 8191](2960); //@line 57 "./libraries/glad/gladfuncs.hpp"
   $86 = HEAP32[165322]|0; //@line 62 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_viii[$86 & 8191]($glcompare$0$i,$74,-1); //@line 62 "./libraries/glad/gladfuncs.hpp"
   $87 = HEAP32[164775]|0; //@line 63 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_viii[$87 & 8191](7680,7680,7680); //@line 63 "./libraries/glad/gladfuncs.hpp"
  }
 } while(0);
 $88 = ((($s)) + 84|0); //@line 139 "./common/Object.h"
 $89 = HEAP32[$88>>2]|0; //@line 139 "./common/Object.h"
 $90 = HEAP32[$21>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $91 = ($89|0)==(0|0); //@line 132 "./common/Object.h"
 if (!($91)) {
  $92 = HEAP32[$89>>2]|0; //@line 132 "./common/Object.h"
  $93 = ((($92)) + 8|0); //@line 132 "./common/Object.h"
  $94 = HEAP32[$93>>2]|0; //@line 132 "./common/Object.h"
  FUNCTION_TABLE_vi[$94 & 8191]($89); //@line 132 "./common/Object.h"
 }
 $95 = ((($90)) + -52|0); //@line 133 "./common/Object.h"
 $96 = HEAP32[$95>>2]|0; //@line 133 "./common/Object.h"
 $97 = ($96|0)==(0|0); //@line 133 "./common/Object.h"
 if (!($97)) {
  $98 = HEAP32[$96>>2]|0; //@line 133 "./common/Object.h"
  $99 = ((($98)) + 12|0); //@line 133 "./common/Object.h"
  $100 = HEAP32[$99>>2]|0; //@line 133 "./common/Object.h"
  FUNCTION_TABLE_vi[$100 & 8191]($96); //@line 133 "./common/Object.h"
 }
 HEAP32[$95>>2] = $89; //@line 134 "./common/Object.h"
 $101 = ((($s)) + 88|0); //@line 139 "./common/Object.h"
 $102 = HEAP32[$101>>2]|0; //@line 139 "./common/Object.h"
 $103 = ($102|0)==(0|0); //@line 989 "modules/graphics/opengl/Graphics.cpp"
 $104 = HEAP32[$21>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($103) {
  __ZN4love8graphics6opengl6Shader6detachEv(); //@line 1004 "modules/graphics/opengl/Graphics.cpp"
  $105 = ((($104)) + -48|0); //@line 133 "./common/Object.h"
  $106 = HEAP32[$105>>2]|0; //@line 133 "./common/Object.h"
  $107 = ($106|0)==(0|0); //@line 133 "./common/Object.h"
  if (!($107)) {
   $108 = HEAP32[$106>>2]|0; //@line 133 "./common/Object.h"
   $109 = ((($108)) + 12|0); //@line 133 "./common/Object.h"
   $110 = HEAP32[$109>>2]|0; //@line 133 "./common/Object.h"
   FUNCTION_TABLE_vi[$110 & 8191]($106); //@line 133 "./common/Object.h"
  }
  HEAP32[$105>>2] = 0; //@line 134 "./common/Object.h"
 } else {
  __ZN4love8graphics6opengl6Shader6attachEb($102,0); //@line 994 "modules/graphics/opengl/Graphics.cpp"
  $111 = HEAP32[$102>>2]|0; //@line 132 "./common/Object.h"
  $112 = ((($111)) + 8|0); //@line 132 "./common/Object.h"
  $113 = HEAP32[$112>>2]|0; //@line 132 "./common/Object.h"
  FUNCTION_TABLE_vi[$113 & 8191]($102); //@line 132 "./common/Object.h"
  $114 = ((($104)) + -48|0); //@line 133 "./common/Object.h"
  $115 = HEAP32[$114>>2]|0; //@line 133 "./common/Object.h"
  $116 = ($115|0)==(0|0); //@line 133 "./common/Object.h"
  if (!($116)) {
   $117 = HEAP32[$115>>2]|0; //@line 133 "./common/Object.h"
   $118 = ((($117)) + 12|0); //@line 133 "./common/Object.h"
   $119 = HEAP32[$118>>2]|0; //@line 133 "./common/Object.h"
   FUNCTION_TABLE_vi[$119 & 8191]($115); //@line 133 "./common/Object.h"
  }
  HEAP32[$114>>2] = $102; //@line 134 "./common/Object.h"
 }
 $120 = ((($s)) + 92|0); //@line 126 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl8Graphics9setCanvasERKNSt3__16vectorINS_9StrongRefINS1_6CanvasEEENS3_9allocatorIS7_EEEE($this,$120); //@line 126 "modules/graphics/opengl/Graphics.cpp"
 $121 = ((($s)) + 104|0); //@line 128 "modules/graphics/opengl/Graphics.cpp"
 $122 = HEAPU8[$121>>0]|(HEAPU8[$121+1>>0]<<8)|(HEAPU8[$121+2>>0]<<16)|(HEAPU8[$121+3>>0]<<24); //@line 128 "modules/graphics/opengl/Graphics.cpp"
 $123 = $122&255;
 $124 = $122 >>> 8;
 $125 = $124&255;
 $126 = $122 >>> 16;
 $127 = $126&255;
 $128 = $122 >>> 24;
 $129 = $128&255;
 $130 = HEAP32[165466]|0; //@line 54 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiii[$130 & 8191]($123,$125,$127,$129); //@line 54 "./libraries/glad/gladfuncs.hpp"
 $131 = HEAP32[$21>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $132 = ((($131)) + -32|0); //@line 1085 "modules/graphics/opengl/Graphics.cpp"
 $133 = $128 << 24; //@line 1085 "modules/graphics/opengl/Graphics.cpp"
 $134 = $122 & 16777215; //@line 1085 "modules/graphics/opengl/Graphics.cpp"
 $135 = $134 | $133; //@line 1085 "modules/graphics/opengl/Graphics.cpp"
 HEAP8[$132>>0]=$135&255;HEAP8[$132+1>>0]=($135>>8)&255;HEAP8[$132+2>>0]=($135>>16)&255;HEAP8[$132+3>>0]=$135>>24; //@line 1085 "modules/graphics/opengl/Graphics.cpp"
 $136 = ((($s)) + 108|0); //@line 129 "modules/graphics/opengl/Graphics.cpp"
 $137 = HEAP8[$136>>0]|0; //@line 129 "modules/graphics/opengl/Graphics.cpp"
 $138 = HEAP8[1252012]|0; //@line 1244 "modules/graphics/opengl/Graphics.cpp"
 $139 = ($138<<24>>24)==(0); //@line 1244 "modules/graphics/opengl/Graphics.cpp"
 if (!($139)) {
  $145 = ((($s)) + 112|0); //@line 131 "modules/graphics/opengl/Graphics.cpp"
  __ZN4love8graphics7Texture16setDefaultFilterERKNS1_6FilterE($145); //@line 1176 "modules/graphics/opengl/Graphics.cpp"
  $146 = HEAP32[$21>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $147 = ((($146)) + -24|0); //@line 1177 "modules/graphics/opengl/Graphics.cpp"
  ;HEAP32[$147>>2]=HEAP32[$145>>2]|0;HEAP32[$147+4>>2]=HEAP32[$145+4>>2]|0;HEAP32[$147+8>>2]=HEAP32[$145+8>>2]|0;HEAP32[$147+12>>2]=HEAP32[$145+12>>2]|0; //@line 1177 "modules/graphics/opengl/Graphics.cpp"
  $148 = ((($s)) + 128|0); //@line 132 "modules/graphics/opengl/Graphics.cpp"
  $149 = HEAP32[$148>>2]|0; //@line 132 "modules/graphics/opengl/Graphics.cpp"
  $150 = ((($s)) + 132|0); //@line 132 "modules/graphics/opengl/Graphics.cpp"
  $151 = +HEAPF32[$150>>2]; //@line 132 "modules/graphics/opengl/Graphics.cpp"
  __ZN4love8graphics6opengl5Image22setDefaultMipmapFilterENS0_7Texture10FilterModeE($149); //@line 1187 "modules/graphics/opengl/Graphics.cpp"
  __ZN4love8graphics6opengl5Image25setDefaultMipmapSharpnessEf($151); //@line 1188 "modules/graphics/opengl/Graphics.cpp"
  $152 = HEAP32[$21>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $153 = ((($152)) + -8|0); //@line 1190 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$153>>2] = $149; //@line 1190 "modules/graphics/opengl/Graphics.cpp"
  $154 = ((($152)) + -4|0); //@line 1191 "modules/graphics/opengl/Graphics.cpp"
  HEAPF32[$154>>2] = $151; //@line 1191 "modules/graphics/opengl/Graphics.cpp"
  STACKTOP = sp;return; //@line 133 "modules/graphics/opengl/Graphics.cpp"
 }
 $140 = ($137<<24>>24)!=(0); //@line 129 "modules/graphics/opengl/Graphics.cpp"
 $141 = $140 ? 6913 : 6914; //@line 1247 "modules/graphics/opengl/Graphics.cpp"
 $142 = HEAP32[165045]|0; //@line 40 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$142 & 8191](1032,$141); //@line 40 "./libraries/glad/gladfuncs.hpp"
 $143 = HEAP32[$21>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $144 = ((($143)) + -28|0); //@line 1248 "modules/graphics/opengl/Graphics.cpp"
 HEAP8[$144>>0] = $137; //@line 1248 "modules/graphics/opengl/Graphics.cpp"
 $145 = ((($s)) + 112|0); //@line 131 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics7Texture16setDefaultFilterERKNS1_6FilterE($145); //@line 1176 "modules/graphics/opengl/Graphics.cpp"
 $146 = HEAP32[$21>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $147 = ((($146)) + -24|0); //@line 1177 "modules/graphics/opengl/Graphics.cpp"
 ;HEAP32[$147>>2]=HEAP32[$145>>2]|0;HEAP32[$147+4>>2]=HEAP32[$145+4>>2]|0;HEAP32[$147+8>>2]=HEAP32[$145+8>>2]|0;HEAP32[$147+12>>2]=HEAP32[$145+12>>2]|0; //@line 1177 "modules/graphics/opengl/Graphics.cpp"
 $148 = ((($s)) + 128|0); //@line 132 "modules/graphics/opengl/Graphics.cpp"
 $149 = HEAP32[$148>>2]|0; //@line 132 "modules/graphics/opengl/Graphics.cpp"
 $150 = ((($s)) + 132|0); //@line 132 "modules/graphics/opengl/Graphics.cpp"
 $151 = +HEAPF32[$150>>2]; //@line 132 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl5Image22setDefaultMipmapFilterENS0_7Texture10FilterModeE($149); //@line 1187 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl5Image25setDefaultMipmapSharpnessEf($151); //@line 1188 "modules/graphics/opengl/Graphics.cpp"
 $152 = HEAP32[$21>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $153 = ((($152)) + -8|0); //@line 1190 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$153>>2] = $149; //@line 1190 "modules/graphics/opengl/Graphics.cpp"
 $154 = ((($152)) + -4|0); //@line 1191 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$154>>2] = $151; //@line 1191 "modules/graphics/opengl/Graphics.cpp"
 STACKTOP = sp;return; //@line 133 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics12setBlendModeENS0_8Graphics9BlendModeENS3_10BlendAlphaE($this,$mode,$alphamode) {
 $this = $this|0;
 $mode = $mode|0;
 $alphamode = $alphamode|0;
 var $$0 = 0, $$04 = 0, $$srcRGB$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dstRGB$0 = 0, $func$0 = 0, $func$1 = 0, $modestr = 0, $or$cond3 = 0, $srcA$0 = 0, $srcRGB$0 = 0, $switch = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $modestr = sp + 12|0;
 $0 = $mode | 1; //@line 1101 "modules/graphics/opengl/Graphics.cpp"
 $1 = ($0|0)==(5); //@line 1101 "modules/graphics/opengl/Graphics.cpp"
 if ($1) {
  $2 = HEAP8[1251998]|0; //@line 1686 "modules/graphics/opengl/Graphics.cpp"
  $3 = HEAP8[1252013]|0; //@line 1686 "modules/graphics/opengl/Graphics.cpp"
  $4 = $3 | $2; //@line 1686 "modules/graphics/opengl/Graphics.cpp"
  $5 = HEAP8[1252166]|0; //@line 1686 "modules/graphics/opengl/Graphics.cpp"
  $6 = $4 | $5; //@line 1686 "modules/graphics/opengl/Graphics.cpp"
  $7 = ($6<<24>>24)==(0); //@line 1686 "modules/graphics/opengl/Graphics.cpp"
  if ($7) {
   $8 = (___cxa_allocate_exception(16)|0); //@line 1104 "modules/graphics/opengl/Graphics.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($8|0),(435002|0),($vararg_buffer|0)); //@line 1104 "modules/graphics/opengl/Graphics.cpp"
   $9 = __THREW__; __THREW__ = 0;
   $10 = $9&1;
   if (!($10)) {
    ___cxa_throw(($8|0),(96|0),(344|0)); //@line 1104 "modules/graphics/opengl/Graphics.cpp"
    // unreachable; //@line 1104 "modules/graphics/opengl/Graphics.cpp"
   }
   $11 = ___cxa_find_matching_catch_2()|0;
   $12 = tempRet0;
   ___cxa_free_exception(($8|0)); //@line 1104 "modules/graphics/opengl/Graphics.cpp"
   $$0 = $12;$$04 = $11;
   ___resumeException($$04|0);
   // unreachable;
  }
 }
 $13 = ($alphamode|0)==(1); //@line 1107 "modules/graphics/opengl/Graphics.cpp"
 do {
  if (!($13)) {
   HEAP32[$modestr>>2] = 442424; //@line 1109 "modules/graphics/opengl/Graphics.cpp"
   $14 = $mode & -2; //@line 1110 "modules/graphics/opengl/Graphics.cpp"
   $switch = ($14|0)==(4); //@line 1110 "modules/graphics/opengl/Graphics.cpp"
   if (!($switch)) {
    break;
   }
   (__ZN4love8graphics8Graphics11getConstantENS1_9BlendModeERPKc($mode,$modestr)|0); //@line 1115 "modules/graphics/opengl/Graphics.cpp"
   $15 = (___cxa_allocate_exception(16)|0); //@line 1116 "modules/graphics/opengl/Graphics.cpp"
   $16 = HEAP32[$modestr>>2]|0; //@line 1116 "modules/graphics/opengl/Graphics.cpp"
   __THREW__ = 0;
   HEAP32[$vararg_buffer1>>2] = $16; //@line 1116 "modules/graphics/opengl/Graphics.cpp"
   invoke_viii(3349,($15|0),(435075|0),($vararg_buffer1|0)); //@line 1116 "modules/graphics/opengl/Graphics.cpp"
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if (!($18)) {
    ___cxa_throw(($15|0),(96|0),(344|0)); //@line 1116 "modules/graphics/opengl/Graphics.cpp"
    // unreachable; //@line 1116 "modules/graphics/opengl/Graphics.cpp"
   }
   $19 = ___cxa_find_matching_catch_2()|0;
   $20 = tempRet0;
   ___cxa_free_exception(($15|0)); //@line 1116 "modules/graphics/opengl/Graphics.cpp"
   $$0 = $20;$$04 = $19;
   ___resumeException($$04|0);
   // unreachable;
  }
 } while(0);
 switch ($mode|0) {
 case 0:  {
  $dstRGB$0 = 771;$func$1 = 32774;$srcA$0 = 1;$srcRGB$0 = 1;
  break;
 }
 case 3:  {
  $dstRGB$0 = 0;$func$1 = 32774;$srcA$0 = 774;$srcRGB$0 = 774;
  break;
 }
 case 2:  {
  $func$0 = 32779;
  label = 15;
  break;
 }
 case 1:  {
  $func$0 = 32774;
  label = 15;
  break;
 }
 case 4:  {
  $dstRGB$0 = 0;$func$1 = 32776;$srcA$0 = 1;$srcRGB$0 = 1;
  break;
 }
 case 5:  {
  $dstRGB$0 = 0;$func$1 = 32775;$srcA$0 = 1;$srcRGB$0 = 1;
  break;
 }
 case 6:  {
  $dstRGB$0 = 769;$func$1 = 32774;$srcA$0 = 1;$srcRGB$0 = 1;
  break;
 }
 default: {
  $dstRGB$0 = 0;$func$1 = 32774;$srcA$0 = 1;$srcRGB$0 = 1;
 }
 }
 if ((label|0) == 15) {
  $dstRGB$0 = 1;$func$1 = $func$0;$srcA$0 = 0;$srcRGB$0 = 1;
 }
 $21 = ($srcRGB$0|0)==(1); //@line 1158 "modules/graphics/opengl/Graphics.cpp"
 $22 = ($alphamode|0)==(0); //@line 1158 "modules/graphics/opengl/Graphics.cpp"
 $or$cond3 = $22 & $21; //@line 1158 "modules/graphics/opengl/Graphics.cpp"
 $$srcRGB$0 = $or$cond3 ? 770 : $srcRGB$0; //@line 1158 "modules/graphics/opengl/Graphics.cpp"
 $23 = HEAP32[165469]|0; //@line 475 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$23 & 8191]($func$1); //@line 475 "./libraries/glad/gladfuncs.hpp"
 $24 = HEAP32[164680]|0; //@line 429 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiii[$24 & 8191]($$srcRGB$0,$dstRGB$0,$srcA$0,$dstRGB$0); //@line 429 "./libraries/glad/gladfuncs.hpp"
 $25 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $26 = HEAP32[$25>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = ((($26)) + -104|0); //@line 1164 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$27>>2] = $mode; //@line 1164 "modules/graphics/opengl/Graphics.cpp"
 $28 = ((($26)) + -100|0); //@line 1165 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$28>>2] = $alphamode; //@line 1165 "modules/graphics/opengl/Graphics.cpp"
 STACKTOP = sp;return; //@line 1166 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNSt3__16vectorIdNS_9allocatorIdEEE21__push_back_slow_pathIdEEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(536870911); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 3; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(268435455); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 2; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 536870911;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 3; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<3)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = +HEAPF64[$__x>>3]; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAPF64[$24>>3] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 8|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 3; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<3)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl8Graphics9unSetModeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0); //@line 620 "modules/graphics/opengl/Graphics.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 620 "modules/graphics/opengl/Graphics.cpp"
 $2 = ($1<<24>>24)==(0); //@line 620 "modules/graphics/opengl/Graphics.cpp"
 if ($2) {
  return; //@line 355 "modules/graphics/opengl/Graphics.cpp"
 }
 __ZN4love8graphics8Volatile9unloadAllEv(); //@line 350 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl6OpenGL13deInitContextEv(634044); //@line 352 "modules/graphics/opengl/Graphics.cpp"
 HEAP8[$0>>0] = 0; //@line 354 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 355 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics9setActiveEb($this,$enable) {
 $this = $this|0;
 $enable = $enable|0;
 var $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $brmerge = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0); //@line 620 "modules/graphics/opengl/Graphics.cpp"
 $1 = HEAP16[$0>>1]|0; //@line 620 "modules/graphics/opengl/Graphics.cpp"
 $2 = $1&255; //@line 620 "modules/graphics/opengl/Graphics.cpp"
 $3 = ($2<<24>>24)==(0); //@line 620 "modules/graphics/opengl/Graphics.cpp"
 $$pre = ((($this)) + 41|0); //@line 364 "modules/graphics/opengl/Graphics.cpp"
 $4 = ($1&65535)<(256); //@line 361 "modules/graphics/opengl/Graphics.cpp"
 $brmerge = $4 | $enable; //@line 361 "modules/graphics/opengl/Graphics.cpp"
 $or$cond = $3 | $brmerge; //@line 361 "modules/graphics/opengl/Graphics.cpp"
 if (!($or$cond)) {
  $5 = HEAP32[165082]|0; //@line 58 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_v[$5 & 8191](); //@line 58 "./libraries/glad/gladfuncs.hpp"
 }
 $6 = $enable&1; //@line 364 "modules/graphics/opengl/Graphics.cpp"
 HEAP8[$$pre>>0] = $6; //@line 364 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 365 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNK4love8graphics6opengl8Graphics8isActiveEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 41|0); //@line 371 "modules/graphics/opengl/Graphics.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 371 "modules/graphics/opengl/Graphics.cpp"
 $2 = ($1<<24>>24)==(0); //@line 371 "modules/graphics/opengl/Graphics.cpp"
 if ($2) {
  $13 = 0;
  return ($13|0); //@line 371 "modules/graphics/opengl/Graphics.cpp"
 }
 $3 = ((($this)) + 40|0); //@line 620 "modules/graphics/opengl/Graphics.cpp"
 $4 = HEAP8[$3>>0]|0; //@line 620 "modules/graphics/opengl/Graphics.cpp"
 $5 = ($4<<24>>24)==(0); //@line 620 "modules/graphics/opengl/Graphics.cpp"
 if ($5) {
  $13 = 0;
  return ($13|0); //@line 371 "modules/graphics/opengl/Graphics.cpp"
 }
 $6 = ((($this)) + 8|0); //@line 139 "./common/Object.h"
 $7 = HEAP32[$6>>2]|0; //@line 139 "./common/Object.h"
 $8 = ($7|0)==(0|0); //@line 371 "modules/graphics/opengl/Graphics.cpp"
 if ($8) {
  $13 = 0;
  return ($13|0); //@line 371 "modules/graphics/opengl/Graphics.cpp"
 }
 $9 = HEAP32[$7>>2]|0; //@line 371 "modules/graphics/opengl/Graphics.cpp"
 $10 = ((($9)) + 72|0); //@line 371 "modules/graphics/opengl/Graphics.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 371 "modules/graphics/opengl/Graphics.cpp"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($7)|0); //@line 371 "modules/graphics/opengl/Graphics.cpp"
 $13 = $12;
 return ($13|0); //@line 371 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8GraphicsC2Ev($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$pre = 0, $$pre$i$i = 0, $$pre$i$i$i$i = 0, $$pre$i$i$i$i10 = 0, $$pre$i$i$i$i22 = 0, $$pre$i$i$i6$i = 0, $$pre$i$i13 = 0, $$pre$i$i25 = 0, $$pre$i3$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0;
 var $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0;
 var $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0;
 var $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0;
 var $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0;
 var $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0;
 var $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0;
 var $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0;
 var $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0;
 var $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0;
 var $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0;
 var $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__v$i = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body6$index8Z2D = 0, $eh$lpad$body6$indexZ2D = 0, $h = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i$i$i = 0, $scevgep$i$i$i$i$i$i$i$i = 0, $scevgep$i$i$i19 = 0, $scevgep4$i$i$i = 0;
 var $scevgep4$i$i$i$i$i = 0, $scevgep4$i$i$i$i$i$i$i$i = 0, $scevgep4$i$i$i20 = 0, $w = 0, $wsettings = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 496|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__v$i = sp + 464|0;
 $0 = sp + 208|0;
 $1 = sp + 72|0;
 $w = sp + 68|0;
 $h = sp + 64|0;
 $wsettings = sp;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./common/Module.h"
 HEAP32[$this>>2] = (24764); //@line 60 "modules/graphics/opengl/Graphics.cpp"
 $2 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $3 = ((($this)) + 8|0); //@line 93 "./common/Object.h"
 HEAP32[$3>>2] = $2; //@line 93 "./common/Object.h"
 $4 = ($2|0)==(0|0); //@line 95 "./common/Object.h"
 if (!($4)) {
  $5 = HEAP32[$2>>2]|0; //@line 95 "./common/Object.h"
  $6 = ((($5)) + 8|0); //@line 95 "./common/Object.h"
  $7 = HEAP32[$6>>2]|0; //@line 95 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($7|0,($2|0)); //@line 95 "./common/Object.h"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if ($9) {
   $274 = ___cxa_find_matching_catch_2()|0;
   $275 = tempRet0;
   $$1 = $275;$$12 = $274;
   __ZN4love8graphics8GraphicsD2Ev($this); //@line 76 "modules/graphics/opengl/Graphics.cpp"
   ___resumeException($$12|0);
   // unreachable;
  }
 }
 $10 = ((($this)) + 12|0); //@line 88 "./common/Object.h"
 $11 = ((($this)) + 16|0); //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ((($this)) + 20|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ((($this)) + 41|0); //@line 58 "modules/graphics/opengl/Graphics.cpp"
 ;HEAP32[$10>>2]=0|0;HEAP32[$10+4>>2]=0|0;HEAP32[$10+8>>2]=0|0;HEAP32[$10+12>>2]=0|0;HEAP32[$10+16>>2]=0|0;HEAP32[$10+20>>2]=0|0;HEAP32[$10+24>>2]=0|0;HEAP8[$10+28>>0]=0|0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP8[$13>>0] = 1; //@line 58 "modules/graphics/opengl/Graphics.cpp"
 $14 = ((($this)) + 42|0); //@line 59 "modules/graphics/opengl/Graphics.cpp"
 HEAP8[$14>>0] = 0; //@line 59 "modules/graphics/opengl/Graphics.cpp"
 $15 = ((($this)) + 44|0); //@line 52 "modules/graphics/opengl/Graphics.cpp"
 $16 = ((($this)) + 48|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ((($this)) + 52|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $18 = ((($this)) + 56|0); //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $19 = ((($this)) + 60|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 ;HEAP32[$15>>2]=0|0;HEAP32[$15+4>>2]=0|0;HEAP32[$15+8>>2]=0|0;HEAP32[$15+12>>2]=0|0;HEAP32[$15+16>>2]=0|0;HEAP32[$15+20>>2]=0|0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 __THREW__ = 0;
 invoke_vi(4052,($0|0)); //@line 61 "modules/graphics/opengl/Graphics.cpp"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 L6: do {
  if ($21) {
   label = 68;
  } else {
   $22 = HEAP32[158511]|0; //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $23 = ($22|0)==(0|0); //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($23)) {
    $24 = HEAP32[(634048)>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $25 = ($24|0)==($22|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($25) {
     $30 = $22;
    } else {
     $27 = $24;
     while(1) {
      $26 = ((($27)) + -64|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[(634048)>>2] = $26; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      __ZN4love7Matrix4D2Ev($26); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $28 = HEAP32[(634048)>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $29 = ($28|0)==($22|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($29) {
       break;
      } else {
       $27 = $28;
      }
     }
     $$pre$i$i$i$i22 = HEAP32[158511]|0; //@line 948 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $30 = $$pre$i$i$i$i22;
    }
    __ZdlPv($30); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    HEAP32[(634052)>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[(634048)>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[158511] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $31 = HEAP32[$0>>2]|0; //@line 1343 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[158511] = $31; //@line 1343 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $32 = ((($0)) + 4|0); //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $33 = HEAP32[$32>>2]|0; //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[(634048)>>2] = $33; //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $34 = ((($0)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $35 = HEAP32[$34>>2]|0; //@line 1345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[(634052)>>2] = $35; //@line 1345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$34>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$32>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$0>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $36 = ((($0)) + 12|0); //@line 112 "modules/graphics/opengl/OpenGL.h"
   $37 = HEAP32[(634056)>>2]|0; //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $38 = ($37|0)==(0|0); //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($38)) {
    $39 = HEAP32[(634060)>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $40 = ($39|0)==($37|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($40) {
     $45 = $37;
    } else {
     $42 = $39;
     while(1) {
      $41 = ((($42)) + -64|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[(634060)>>2] = $41; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      __ZN4love7Matrix4D2Ev($41); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $43 = HEAP32[(634060)>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $44 = ($43|0)==($37|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($44) {
       break;
      } else {
       $42 = $43;
      }
     }
     $$pre$i$i$i6$i = HEAP32[(634056)>>2]|0; //@line 948 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $45 = $$pre$i$i$i6$i;
    }
    __ZdlPv($45); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    HEAP32[(634064)>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[(634060)>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[(634056)>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $46 = HEAP32[$36>>2]|0; //@line 1343 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[(634056)>>2] = $46; //@line 1343 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $47 = ((($0)) + 16|0); //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $48 = HEAP32[$47>>2]|0; //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[(634060)>>2] = $48; //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $49 = ((($0)) + 20|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $50 = HEAP32[$49>>2]|0; //@line 1345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[(634064)>>2] = $50; //@line 1345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$49>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$47>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$36>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $51 = ((($0)) + 24|0); //@line 79 "modules/graphics/opengl/OpenGL.h"
   dest=(634068); src=$51; stop=dest+40|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 79 "modules/graphics/opengl/OpenGL.h"
   $52 = ((($0)) + 64|0); //@line 79 "modules/graphics/opengl/OpenGL.h"
   $53 = HEAP32[(634108)>>2]|0; //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $54 = ($53|0)==(0|0); //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $55 = $53; //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($54)) {
    $56 = HEAP32[(634112)>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $57 = ($56|0)==($53|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($57)) {
     $scevgep$i$i$i$i$i$i$i$i = ((($56)) + -4|0);
     $58 = $scevgep$i$i$i$i$i$i$i$i;
     $59 = (($58) - ($55))|0;
     $60 = $59 >>> 2;
     $61 = $60 ^ -1;
     $scevgep4$i$i$i$i$i$i$i$i = (($56) + ($61<<2)|0);
     HEAP32[(634112)>>2] = $scevgep4$i$i$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($53); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    HEAP32[(634116)>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[(634112)>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[(634108)>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $62 = HEAP32[$52>>2]|0; //@line 1343 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[(634108)>>2] = $62; //@line 1343 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $63 = ((($0)) + 68|0); //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $64 = HEAP32[$63>>2]|0; //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[(634112)>>2] = $64; //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $65 = ((($0)) + 72|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $66 = HEAP32[$65>>2]|0; //@line 1345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[(634116)>>2] = $66; //@line 1345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$65>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$63>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$52>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $67 = ((($0)) + 76|0); //@line 415 "modules/graphics/opengl/OpenGL.h"
   dest=(634120); src=$67; stop=dest+52|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 415 "modules/graphics/opengl/OpenGL.h"
   $68 = ((($0)) + 128|0); //@line 415 "modules/graphics/opengl/OpenGL.h"
   dest=(634172); src=$68; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 415 "modules/graphics/opengl/OpenGL.h"
   $69 = ((($0)) + 192|0); //@line 415 "modules/graphics/opengl/OpenGL.h"
   dest=(634236); src=$69; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 415 "modules/graphics/opengl/OpenGL.h"
   __ZN4love7Matrix4D2Ev($69); //@line 415 "modules/graphics/opengl/OpenGL.h"
   __ZN4love7Matrix4D2Ev($68); //@line 415 "modules/graphics/opengl/OpenGL.h"
   $70 = ((($0)) + 64|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $71 = HEAP32[$70>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $72 = ($71|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $73 = $71; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($72)) {
    $74 = HEAP32[$63>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $75 = ($74|0)==($71|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($75)) {
     $scevgep$i$i$i$i$i = ((($74)) + -4|0);
     $76 = $scevgep$i$i$i$i$i;
     $77 = (($76) - ($73))|0;
     $78 = $77 >>> 2;
     $79 = $78 ^ -1;
     $scevgep4$i$i$i$i$i = (($74) + ($79<<2)|0);
     HEAP32[$63>>2] = $scevgep4$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($71); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $80 = ((($0)) + 12|0); //@line 112 "modules/graphics/opengl/OpenGL.h"
   $81 = HEAP32[$80>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $82 = ($81|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($82)) {
    $83 = HEAP32[$47>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $84 = ($83|0)==($81|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($84) {
     $89 = $81;
    } else {
     $86 = $83;
     while(1) {
      $85 = ((($86)) + -64|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$47>>2] = $85; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      __ZN4love7Matrix4D2Ev($85); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $87 = HEAP32[$47>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $88 = ($87|0)==($81|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($88) {
       break;
      } else {
       $86 = $87;
      }
     }
     $$pre$i$i25 = HEAP32[$80>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $89 = $$pre$i$i25;
    }
    __ZdlPv($89); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $90 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $91 = ($90|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($91)) {
    $92 = HEAP32[$32>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $93 = ($92|0)==($90|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($93) {
     $98 = $90;
    } else {
     $95 = $92;
     while(1) {
      $94 = ((($95)) + -64|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$32>>2] = $94; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      __ZN4love7Matrix4D2Ev($94); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $96 = HEAP32[$32>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $97 = ($96|0)==($90|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($97) {
       break;
      } else {
       $95 = $96;
      }
     }
     $$pre$i3$i = HEAP32[$0>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $98 = $$pre$i3$i;
    }
    __ZdlPv($98); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $99 = ((($this)) + 52|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $100 = HEAP32[$99>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $101 = HEAP32[$15>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $102 = (($100) - ($101))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $103 = (($102|0) / 136)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $104 = ($103>>>0)<(10); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $105 = $100; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   do {
    if ($104) {
     $106 = HEAP32[$16>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $107 = ((($__v$i)) + 12|0); //@line 323 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     HEAP32[$107>>2] = 0; //@line 1959 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $108 = ((($__v$i)) + 16|0); //@line 1959 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$108>>2] = $17; //@line 1959 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     __THREW__ = 0;
     $109 = (invoke_ii(3348,1360)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $110 = __THREW__; __THREW__ = 0;
     $111 = $110&1;
     if ($111) {
      label = 68;
      break L6;
     }
     $112 = (($106) - ($101))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $113 = (($112|0) / 136)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$__v$i>>2] = $109; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $114 = (($109) + (($113*136)|0)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $115 = ((($__v$i)) + 8|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     HEAP32[$115>>2] = $114; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $116 = ((($__v$i)) + 4|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     HEAP32[$116>>2] = $114; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $117 = ((($109)) + 1360|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $118 = ((($__v$i)) + 12|0); //@line 2014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$118>>2] = $117; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     __THREW__ = 0;
     invoke_vii(4053,($15|0),($__v$i|0)); //@line 1543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $119 = __THREW__; __THREW__ = 0;
     $120 = $119&1;
     if ($120) {
      $121 = ___cxa_find_matching_catch_2()|0;
      $122 = tempRet0;
      __ZNSt3__114__split_bufferIN4love8graphics6opengl8Graphics12DisplayStateERNS_9allocatorIS5_EEED2Ev($__v$i); //@line 1544 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $eh$lpad$body$index2Z2D = $122;$eh$lpad$body$indexZ2D = $121;
      label = 69;
      break L6;
     } else {
      __ZNSt3__114__split_bufferIN4love8graphics6opengl8Graphics12DisplayStateERNS_9allocatorIS5_EEED2Ev($__v$i); //@line 1544 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$pre = HEAP32[$99>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $154 = $$pre;
      break;
     }
    } else {
     $154 = $105;
    }
   } while(0);
   _memset(($1|0),0,132)|0; //@line 64 "modules/graphics/opengl/Graphics.cpp"
   HEAPF32[$1>>2] = 255.0; //@line 38 "./modules/graphics/Color.h"
   $123 = ((($1)) + 4|0); //@line 38 "./modules/graphics/Color.h"
   HEAPF32[$123>>2] = 255.0; //@line 38 "./modules/graphics/Color.h"
   $124 = ((($1)) + 8|0); //@line 38 "./modules/graphics/Color.h"
   HEAPF32[$124>>2] = 255.0; //@line 38 "./modules/graphics/Color.h"
   $125 = ((($1)) + 12|0); //@line 38 "./modules/graphics/Color.h"
   HEAPF32[$125>>2] = 255.0; //@line 38 "./modules/graphics/Color.h"
   $126 = ((($1)) + 16|0); //@line 38 "./modules/graphics/Color.h"
   HEAPF32[$126>>2] = 0.0; //@line 38 "./modules/graphics/Color.h"
   $127 = ((($1)) + 20|0); //@line 38 "./modules/graphics/Color.h"
   HEAPF32[$127>>2] = 0.0; //@line 38 "./modules/graphics/Color.h"
   $128 = ((($1)) + 24|0); //@line 38 "./modules/graphics/Color.h"
   HEAPF32[$128>>2] = 0.0; //@line 38 "./modules/graphics/Color.h"
   $129 = ((($1)) + 28|0); //@line 38 "./modules/graphics/Color.h"
   HEAPF32[$129>>2] = 255.0; //@line 38 "./modules/graphics/Color.h"
   $130 = ((($1)) + 32|0); //@line 487 "modules/graphics/opengl/Graphics.h"
   HEAP32[$130>>2] = 0; //@line 487 "modules/graphics/opengl/Graphics.h"
   $131 = ((($1)) + 36|0); //@line 488 "modules/graphics/opengl/Graphics.h"
   HEAP32[$131>>2] = 0; //@line 488 "modules/graphics/opengl/Graphics.h"
   $132 = ((($1)) + 40|0); //@line 490 "modules/graphics/opengl/Graphics.h"
   HEAPF32[$132>>2] = 1.0; //@line 490 "modules/graphics/opengl/Graphics.h"
   $133 = ((($1)) + 44|0); //@line 491 "modules/graphics/opengl/Graphics.h"
   HEAP32[$133>>2] = 1; //@line 491 "modules/graphics/opengl/Graphics.h"
   $134 = ((($1)) + 48|0); //@line 492 "modules/graphics/opengl/Graphics.h"
   HEAP32[$134>>2] = 1; //@line 492 "modules/graphics/opengl/Graphics.h"
   $135 = ((($1)) + 52|0); //@line 494 "modules/graphics/opengl/Graphics.h"
   HEAPF32[$135>>2] = 1.0; //@line 494 "modules/graphics/opengl/Graphics.h"
   $136 = ((($1)) + 56|0); //@line 496 "modules/graphics/opengl/Graphics.h"
   HEAP8[$136>>0] = 0; //@line 496 "modules/graphics/opengl/Graphics.h"
   $137 = ((($1)) + 60|0); //@line 497 "modules/graphics/opengl/Graphics.h"
   ;HEAP32[$137>>2]=0|0;HEAP32[$137+4>>2]=0|0;HEAP32[$137+8>>2]=0|0;HEAP32[$137+12>>2]=0|0; //@line 497 "modules/graphics/opengl/Graphics.h"
   $138 = ((($1)) + 76|0); //@line 500 "modules/graphics/opengl/Graphics.h"
   HEAP32[$138>>2] = 6; //@line 500 "modules/graphics/opengl/Graphics.h"
   $139 = ((($1)) + 80|0); //@line 501 "modules/graphics/opengl/Graphics.h"
   $140 = ((($1)) + 104|0); //@line 208 "./modules/graphics/Graphics.h"
   ;HEAP32[$139>>2]=0|0;HEAP32[$139+4>>2]=0|0;HEAP32[$139+8>>2]=0|0;HEAP32[$139+12>>2]=0|0;HEAP32[$139+16>>2]=0|0;HEAP32[$139+20>>2]=0|0; //@line 88 "./common/Object.h"
   $141 = ((($1)) + 108|0); //@line 510 "modules/graphics/opengl/Graphics.h"
   HEAP32[$140>>2] = 16843009; //@line 208 "./modules/graphics/Graphics.h"
   HEAP8[$141>>0] = 0; //@line 510 "modules/graphics/opengl/Graphics.h"
   $142 = ((($1)) + 112|0); //@line 512 "modules/graphics/opengl/Graphics.h"
   $143 = $142; //@line 512 "modules/graphics/opengl/Graphics.h"
   $144 = $143; //@line 512 "modules/graphics/opengl/Graphics.h"
   HEAP32[$144>>2] = 0; //@line 512 "modules/graphics/opengl/Graphics.h"
   $145 = (($143) + 4)|0; //@line 512 "modules/graphics/opengl/Graphics.h"
   $146 = $145; //@line 512 "modules/graphics/opengl/Graphics.h"
   HEAP32[$146>>2] = 0; //@line 512 "modules/graphics/opengl/Graphics.h"
   HEAP32[$142>>2] = 1; //@line 62 "./modules/graphics/Texture.h"
   $147 = ((($1)) + 116|0); //@line 63 "./modules/graphics/Texture.h"
   HEAP32[$147>>2] = 1; //@line 63 "./modules/graphics/Texture.h"
   $148 = ((($1)) + 120|0); //@line 64 "./modules/graphics/Texture.h"
   HEAP32[$148>>2] = 0; //@line 64 "./modules/graphics/Texture.h"
   $149 = ((($1)) + 124|0); //@line 65 "./modules/graphics/Texture.h"
   HEAPF32[$149>>2] = 1.0; //@line 65 "./modules/graphics/Texture.h"
   $150 = ((($1)) + 128|0); //@line 514 "modules/graphics/opengl/Graphics.h"
   HEAP32[$150>>2] = 2; //@line 514 "modules/graphics/opengl/Graphics.h"
   $151 = ((($1)) + 132|0); //@line 515 "modules/graphics/opengl/Graphics.h"
   HEAPF32[$151>>2] = 0.0; //@line 515 "modules/graphics/opengl/Graphics.h"
   $152 = HEAP32[$16>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $153 = ($152>>>0)<($154>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   do {
    if ($153) {
     dest=$152; src=$1; stop=dest+84|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 482 "modules/graphics/opengl/Graphics.h"
     $155 = ((($152)) + 84|0); //@line 99 "./common/Object.h"
     $156 = ((($1)) + 84|0); //@line 139 "./common/Object.h"
     $157 = HEAP32[$156>>2]|0; //@line 139 "./common/Object.h"
     HEAP32[$155>>2] = $157; //@line 99 "./common/Object.h"
     $158 = ($157|0)==(0|0); //@line 101 "./common/Object.h"
     if (!($158)) {
      $159 = HEAP32[$157>>2]|0; //@line 101 "./common/Object.h"
      $160 = ((($159)) + 8|0); //@line 101 "./common/Object.h"
      $161 = HEAP32[$160>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($161|0,($157|0)); //@line 101 "./common/Object.h"
      $162 = __THREW__; __THREW__ = 0;
      $163 = $162&1;
      if ($163) {
       label = 70;
       break;
      }
     }
     $164 = ((($152)) + 88|0); //@line 99 "./common/Object.h"
     $165 = ((($1)) + 88|0); //@line 139 "./common/Object.h"
     $166 = HEAP32[$165>>2]|0; //@line 139 "./common/Object.h"
     HEAP32[$164>>2] = $166; //@line 99 "./common/Object.h"
     $167 = ($166|0)==(0|0); //@line 101 "./common/Object.h"
     if (!($167)) {
      $168 = HEAP32[$166>>2]|0; //@line 101 "./common/Object.h"
      $169 = ((($168)) + 8|0); //@line 101 "./common/Object.h"
      $170 = HEAP32[$169>>2]|0; //@line 101 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($170|0,($166|0)); //@line 101 "./common/Object.h"
      $171 = __THREW__; __THREW__ = 0;
      $172 = $171&1;
      if ($172) {
       $173 = ___cxa_find_matching_catch_2()|0;
       $174 = tempRet0;
       $175 = HEAP32[$155>>2]|0; //@line 106 "./common/Object.h"
       $176 = ($175|0)==(0|0); //@line 106 "./common/Object.h"
       if ($176) {
        $eh$lpad$body6$index8Z2D = $174;$eh$lpad$body6$indexZ2D = $173;
        break;
       }
       $177 = HEAP32[$175>>2]|0; //@line 106 "./common/Object.h"
       $178 = ((($177)) + 12|0); //@line 106 "./common/Object.h"
       $179 = HEAP32[$178>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($179|0,($175|0)); //@line 106 "./common/Object.h"
       $180 = __THREW__; __THREW__ = 0;
       $181 = $180&1;
       if (!($181)) {
        $eh$lpad$body6$index8Z2D = $174;$eh$lpad$body6$indexZ2D = $173;
        break;
       }
       $182 = ___cxa_find_matching_catch_3(0|0)|0;
       $183 = tempRet0;
       ___clang_call_terminate($182); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
     }
     $184 = ((($152)) + 92|0); //@line 482 "modules/graphics/opengl/Graphics.h"
     $185 = ((($1)) + 92|0); //@line 482 "modules/graphics/opengl/Graphics.h"
     HEAP32[$184>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $186 = ((($152)) + 96|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$186>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $187 = ((($152)) + 100|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$187>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $188 = HEAP32[$185>>2]|0; //@line 1246 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$184>>2] = $188; //@line 1246 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $189 = ((($1)) + 96|0); //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $190 = HEAP32[$189>>2]|0; //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$186>>2] = $190; //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $191 = ((($1)) + 100|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $192 = HEAP32[$191>>2]|0; //@line 1248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $193 = ((($152)) + 100|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$193>>2] = $192; //@line 1248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$191>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$189>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$185>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $194 = ((($152)) + 104|0); //@line 482 "modules/graphics/opengl/Graphics.h"
     ;HEAP32[$194>>2]=HEAP32[$140>>2]|0;HEAP32[$194+4>>2]=HEAP32[$140+4>>2]|0;HEAP32[$194+8>>2]=HEAP32[$140+8>>2]|0;HEAP32[$194+12>>2]=HEAP32[$140+12>>2]|0;HEAP32[$194+16>>2]=HEAP32[$140+16>>2]|0;HEAP32[$194+20>>2]=HEAP32[$140+20>>2]|0;HEAP32[$194+24>>2]=HEAP32[$140+24>>2]|0;HEAP32[$194+28>>2]=HEAP32[$140+28>>2]|0; //@line 482 "modules/graphics/opengl/Graphics.h"
     $195 = HEAP32[$16>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $196 = ((($195)) + 136|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$16>>2] = $196; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     label = 47;
    } else {
     __THREW__ = 0;
     invoke_vii(4054,($15|0),($1|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $197 = __THREW__; __THREW__ = 0;
     $198 = $197&1;
     if ($198) {
      label = 70;
     } else {
      label = 47;
     }
    }
   } while(0);
   if ((label|0) == 47) {
    $199 = ((($1)) + 92|0); //@line 482 "modules/graphics/opengl/Graphics.h"
    $200 = HEAP32[$199>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $201 = ($200|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($201)) {
     $202 = ((($1)) + 96|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $203 = HEAP32[$202>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $204 = ($203|0)==($200|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     do {
      if ($204) {
       $218 = $200;
      } else {
       $206 = $203;
       while(1) {
        $205 = ((($206)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$202>>2] = $205; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $207 = HEAP32[$205>>2]|0; //@line 106 "./common/Object.h"
        $208 = ($207|0)==(0|0); //@line 106 "./common/Object.h"
        if ($208) {
         $210 = $205;
        } else {
         $211 = HEAP32[$207>>2]|0; //@line 106 "./common/Object.h"
         $212 = ((($211)) + 12|0); //@line 106 "./common/Object.h"
         $213 = HEAP32[$212>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($213|0,($207|0)); //@line 106 "./common/Object.h"
         $214 = __THREW__; __THREW__ = 0;
         $215 = $214&1;
         if ($215) {
          label = 53;
          break;
         }
         $$pre$i$i$i$i = HEAP32[$202>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $210 = $$pre$i$i$i$i;
        }
        $209 = ($210|0)==($200|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($209) {
         label = 54;
         break;
        } else {
         $206 = $210;
        }
       }
       if ((label|0) == 53) {
        $216 = ___cxa_find_matching_catch_3(0|0)|0;
        $217 = tempRet0;
        ___clang_call_terminate($216); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       else if ((label|0) == 54) {
        $$pre$i$i = HEAP32[$199>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $218 = $$pre$i$i;
        break;
       }
      }
     } while(0);
     __ZdlPv($218); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    }
    $219 = ((($1)) + 88|0); //@line 106 "./common/Object.h"
    $220 = HEAP32[$219>>2]|0; //@line 106 "./common/Object.h"
    $221 = ($220|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($221)) {
     $222 = HEAP32[$220>>2]|0; //@line 106 "./common/Object.h"
     $223 = ((($222)) + 12|0); //@line 106 "./common/Object.h"
     $224 = HEAP32[$223>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($224|0,($220|0)); //@line 106 "./common/Object.h"
     $225 = __THREW__; __THREW__ = 0;
     $226 = $225&1;
     if ($226) {
      $227 = ___cxa_find_matching_catch_3(0|0)|0;
      $228 = tempRet0;
      ___clang_call_terminate($227); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
    }
    $229 = ((($1)) + 84|0); //@line 106 "./common/Object.h"
    $230 = HEAP32[$229>>2]|0; //@line 106 "./common/Object.h"
    $231 = ($230|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($231)) {
     $232 = HEAP32[$230>>2]|0; //@line 106 "./common/Object.h"
     $233 = ((($232)) + 12|0); //@line 106 "./common/Object.h"
     $234 = HEAP32[$233>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($234|0,($230|0)); //@line 106 "./common/Object.h"
     $235 = __THREW__; __THREW__ = 0;
     $236 = $235&1;
     if ($236) {
      $237 = ___cxa_find_matching_catch_3(0|0)|0;
      $238 = tempRet0;
      ___clang_call_terminate($237); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
    }
    $239 = HEAP32[$3>>2]|0; //@line 117 "./common/Object.h"
    $240 = ($239|0)==(0|0); //@line 66 "modules/graphics/opengl/Graphics.cpp"
    if ($240) {
     STACKTOP = sp;return; //@line 76 "modules/graphics/opengl/Graphics.cpp"
    }
    HEAP8[$wsettings>>0] = 0; //@line 204 "./modules/window/Window.h"
    $241 = ((($wsettings)) + 4|0); //@line 205 "./modules/window/Window.h"
    HEAP32[$241>>2] = 1; //@line 205 "./modules/window/Window.h"
    $242 = ((($wsettings)) + 8|0); //@line 206 "./modules/window/Window.h"
    HEAP8[$242>>0] = 1; //@line 206 "./modules/window/Window.h"
    $243 = ((($wsettings)) + 12|0); //@line 207 "./modules/window/Window.h"
    HEAP32[$243>>2] = 0; //@line 207 "./modules/window/Window.h"
    $244 = ((($wsettings)) + 16|0); //@line 208 "./modules/window/Window.h"
    HEAP8[$244>>0] = 0; //@line 208 "./modules/window/Window.h"
    $245 = ((($wsettings)) + 20|0); //@line 209 "./modules/window/Window.h"
    HEAP32[$245>>2] = 1; //@line 209 "./modules/window/Window.h"
    $246 = ((($wsettings)) + 24|0); //@line 210 "./modules/window/Window.h"
    HEAP32[$246>>2] = 1; //@line 210 "./modules/window/Window.h"
    $247 = ((($wsettings)) + 28|0); //@line 211 "./modules/window/Window.h"
    HEAP8[$247>>0] = 0; //@line 211 "./modules/window/Window.h"
    $248 = ((($wsettings)) + 29|0); //@line 212 "./modules/window/Window.h"
    HEAP8[$248>>0] = 1; //@line 212 "./modules/window/Window.h"
    $249 = ((($wsettings)) + 32|0); //@line 213 "./modules/window/Window.h"
    HEAP32[$249>>2] = 0; //@line 213 "./modules/window/Window.h"
    $250 = ((($wsettings)) + 36|0); //@line 214 "./modules/window/Window.h"
    HEAP8[$250>>0] = 0; //@line 214 "./modules/window/Window.h"
    $251 = ((($wsettings)) + 40|0); //@line 215 "./modules/window/Window.h"
    HEAPF64[$251>>3] = 0.0; //@line 215 "./modules/window/Window.h"
    $252 = ((($wsettings)) + 48|0); //@line 216 "./modules/window/Window.h"
    HEAP8[$252>>0] = 0; //@line 216 "./modules/window/Window.h"
    $253 = ((($wsettings)) + 52|0); //@line 217 "./modules/window/Window.h"
    HEAP32[$253>>2] = 0; //@line 217 "./modules/window/Window.h"
    $254 = ((($wsettings)) + 56|0); //@line 218 "./modules/window/Window.h"
    HEAP32[$254>>2] = 0; //@line 218 "./modules/window/Window.h"
    $255 = HEAP32[$239>>2]|0; //@line 71 "modules/graphics/opengl/Graphics.cpp"
    $256 = ((($255)) + 28|0); //@line 71 "modules/graphics/opengl/Graphics.cpp"
    $257 = HEAP32[$256>>2]|0; //@line 71 "modules/graphics/opengl/Graphics.cpp"
    __THREW__ = 0;
    invoke_viiii($257|0,($239|0),($w|0),($h|0),($wsettings|0)); //@line 71 "modules/graphics/opengl/Graphics.cpp"
    $258 = __THREW__; __THREW__ = 0;
    $259 = $258&1;
    if ($259) {
     label = 68;
     break;
    }
    $260 = HEAP32[$3>>2]|0; //@line 117 "./common/Object.h"
    $261 = HEAP32[$260>>2]|0; //@line 73 "modules/graphics/opengl/Graphics.cpp"
    $262 = ((($261)) + 72|0); //@line 73 "modules/graphics/opengl/Graphics.cpp"
    $263 = HEAP32[$262>>2]|0; //@line 73 "modules/graphics/opengl/Graphics.cpp"
    __THREW__ = 0;
    $264 = (invoke_ii($263|0,($260|0))|0); //@line 73 "modules/graphics/opengl/Graphics.cpp"
    $265 = __THREW__; __THREW__ = 0;
    $266 = $265&1;
    if ($266) {
     label = 68;
     break;
    }
    if ($264) {
     $267 = HEAP32[$this>>2]|0; //@line 74 "modules/graphics/opengl/Graphics.cpp"
     $268 = ((($267)) + 28|0); //@line 74 "modules/graphics/opengl/Graphics.cpp"
     $269 = HEAP32[$268>>2]|0; //@line 74 "modules/graphics/opengl/Graphics.cpp"
     $270 = HEAP32[$w>>2]|0; //@line 74 "modules/graphics/opengl/Graphics.cpp"
     $271 = HEAP32[$h>>2]|0; //@line 74 "modules/graphics/opengl/Graphics.cpp"
     __THREW__ = 0;
     (invoke_iiii($269|0,($this|0),($270|0),($271|0))|0); //@line 74 "modules/graphics/opengl/Graphics.cpp"
     $272 = __THREW__; __THREW__ = 0;
     $273 = $272&1;
     if ($273) {
      label = 68;
      break;
     }
    }
    STACKTOP = sp;return; //@line 76 "modules/graphics/opengl/Graphics.cpp"
   }
   else if ((label|0) == 70) {
    $278 = ___cxa_find_matching_catch_2()|0;
    $279 = tempRet0;
    $eh$lpad$body6$index8Z2D = $279;$eh$lpad$body6$indexZ2D = $278;
   }
   $280 = ((($1)) + 92|0); //@line 482 "modules/graphics/opengl/Graphics.h"
   $281 = HEAP32[$280>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $282 = ($281|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($282)) {
    $283 = ((($1)) + 96|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $284 = HEAP32[$283>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $285 = ($284|0)==($281|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    do {
     if ($285) {
      $299 = $281;
     } else {
      $287 = $284;
      while(1) {
       $286 = ((($287)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$283>>2] = $286; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $288 = HEAP32[$286>>2]|0; //@line 106 "./common/Object.h"
       $289 = ($288|0)==(0|0); //@line 106 "./common/Object.h"
       if ($289) {
        $291 = $286;
       } else {
        $292 = HEAP32[$288>>2]|0; //@line 106 "./common/Object.h"
        $293 = ((($292)) + 12|0); //@line 106 "./common/Object.h"
        $294 = HEAP32[$293>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($294|0,($288|0)); //@line 106 "./common/Object.h"
        $295 = __THREW__; __THREW__ = 0;
        $296 = $295&1;
        if ($296) {
         label = 77;
         break;
        }
        $$pre$i$i$i$i10 = HEAP32[$283>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $291 = $$pre$i$i$i$i10;
       }
       $290 = ($291|0)==($281|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($290) {
        label = 78;
        break;
       } else {
        $287 = $291;
       }
      }
      if ((label|0) == 77) {
       $297 = ___cxa_find_matching_catch_3(0|0)|0;
       $298 = tempRet0;
       ___clang_call_terminate($297); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      else if ((label|0) == 78) {
       $$pre$i$i13 = HEAP32[$280>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $299 = $$pre$i$i13;
       break;
      }
     }
    } while(0);
    __ZdlPv($299); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $300 = ((($1)) + 88|0); //@line 106 "./common/Object.h"
   $301 = HEAP32[$300>>2]|0; //@line 106 "./common/Object.h"
   $302 = ($301|0)==(0|0); //@line 106 "./common/Object.h"
   if (!($302)) {
    $303 = HEAP32[$301>>2]|0; //@line 106 "./common/Object.h"
    $304 = ((($303)) + 12|0); //@line 106 "./common/Object.h"
    $305 = HEAP32[$304>>2]|0; //@line 106 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($305|0,($301|0)); //@line 106 "./common/Object.h"
    $306 = __THREW__; __THREW__ = 0;
    $307 = $306&1;
    if ($307) {
     $308 = ___cxa_find_matching_catch_3(0|0)|0;
     $309 = tempRet0;
     ___clang_call_terminate($308); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   }
   $310 = ((($1)) + 84|0); //@line 106 "./common/Object.h"
   $311 = HEAP32[$310>>2]|0; //@line 106 "./common/Object.h"
   $312 = ($311|0)==(0|0); //@line 106 "./common/Object.h"
   if ($312) {
    $$0 = $eh$lpad$body6$index8Z2D;$$01 = $eh$lpad$body6$indexZ2D;
   } else {
    $313 = HEAP32[$311>>2]|0; //@line 106 "./common/Object.h"
    $314 = ((($313)) + 12|0); //@line 106 "./common/Object.h"
    $315 = HEAP32[$314>>2]|0; //@line 106 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($315|0,($311|0)); //@line 106 "./common/Object.h"
    $316 = __THREW__; __THREW__ = 0;
    $317 = $316&1;
    if ($317) {
     $318 = ___cxa_find_matching_catch_3(0|0)|0;
     $319 = tempRet0;
     ___clang_call_terminate($318); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    } else {
     $$0 = $eh$lpad$body6$index8Z2D;$$01 = $eh$lpad$body6$indexZ2D;
    }
   }
  }
 } while(0);
 if ((label|0) == 68) {
  $276 = ___cxa_find_matching_catch_2()|0;
  $277 = tempRet0;
  $eh$lpad$body$index2Z2D = $277;$eh$lpad$body$indexZ2D = $276;
  label = 69;
 }
 if ((label|0) == 69) {
  $$0 = $eh$lpad$body$index2Z2D;$$01 = $eh$lpad$body$indexZ2D;
 }
 $320 = HEAP32[$18>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $321 = ($320|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $322 = $320; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($321)) {
  $323 = HEAP32[$19>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $324 = ($323|0)==($320|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($324)) {
   $scevgep$i$i$i19 = ((($323)) + -4|0);
   $325 = $scevgep$i$i$i19;
   $326 = (($325) - ($322))|0;
   $327 = $326 >>> 2;
   $328 = $327 ^ -1;
   $scevgep4$i$i$i20 = (($323) + ($328<<2)|0);
   HEAP32[$19>>2] = $scevgep4$i$i$i20; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($320); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 __ZNSt3__113__vector_baseIN4love8graphics6opengl8Graphics12DisplayStateENS_9allocatorIS5_EEED2Ev($15); //@line 76 "modules/graphics/opengl/Graphics.cpp"
 $329 = HEAP32[$11>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $330 = ($329|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $331 = $329; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($330)) {
  $332 = HEAP32[$12>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $333 = ($332|0)==($329|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($333)) {
   $scevgep$i$i$i = ((($332)) + -8|0);
   $334 = $scevgep$i$i$i;
   $335 = (($334) - ($331))|0;
   $336 = $335 >>> 3;
   $337 = $336 ^ -1;
   $scevgep4$i$i$i = (($332) + ($337<<3)|0);
   HEAP32[$12>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($329); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $338 = HEAP32[$10>>2]|0; //@line 106 "./common/Object.h"
 $339 = ($338|0)==(0|0); //@line 106 "./common/Object.h"
 if (!($339)) {
  $340 = HEAP32[$338>>2]|0; //@line 106 "./common/Object.h"
  $341 = ((($340)) + 12|0); //@line 106 "./common/Object.h"
  $342 = HEAP32[$341>>2]|0; //@line 106 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($342|0,($338|0)); //@line 106 "./common/Object.h"
  $343 = __THREW__; __THREW__ = 0;
  $344 = $343&1;
  if ($344) {
   $345 = ___cxa_find_matching_catch_3(0|0)|0;
   $346 = tempRet0;
   ___clang_call_terminate($345); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 }
 $347 = HEAP32[$3>>2]|0; //@line 106 "./common/Object.h"
 $348 = ($347|0)==(0|0); //@line 106 "./common/Object.h"
 if ($348) {
  $$1 = $$0;$$12 = $$01;
  __ZN4love8graphics8GraphicsD2Ev($this); //@line 76 "modules/graphics/opengl/Graphics.cpp"
  ___resumeException($$12|0);
  // unreachable;
 }
 $349 = HEAP32[$347>>2]|0; //@line 106 "./common/Object.h"
 $350 = ((($349)) + 12|0); //@line 106 "./common/Object.h"
 $351 = HEAP32[$350>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($351|0,($347|0)); //@line 106 "./common/Object.h"
 $352 = __THREW__; __THREW__ = 0;
 $353 = $352&1;
 if ($353) {
  $354 = ___cxa_find_matching_catch_3(0|0)|0;
  $355 = tempRet0;
  ___clang_call_terminate($354); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  $$1 = $$0;$$12 = $$01;
  __ZN4love8graphics8GraphicsD2Ev($this); //@line 76 "modules/graphics/opengl/Graphics.cpp"
  ___resumeException($$12|0);
  // unreachable;
 }
}
function __ZNSt3__16vectorIN4love8graphics6opengl8Graphics12DisplayStateENS_9allocatorIS5_EEE26__swap_out_circular_bufferERNS_14__split_bufferIS5_RS7_EE($this,$__v) {
 $this = $this|0;
 $__v = $__v|0;
 var $$0$i$i$i$i$i = 0, $$01$i$i$i$i$i = 0, $$04$i = 0, $$lcssa = 0, $$lcssa34 = 0, $$lcssa37 = 0, $$lcssa39 = 0, $$pre = 0, $$pre$i = 0, $$pre$phi15Z2D = 0, $$pre$phiZ2D = 0, $$pre12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $7 = 0, $8 = 0, $9 = 0;
 var dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($this)) + 4|0); //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($__v)) + 4|0); //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($2|0)==($0|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 do {
  if ($4) {
   $5 = $0; //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $$pre12 = HEAP32[$3>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $$pre$phi15Z2D = $3;$$pre$phiZ2D = $this;$58 = $$pre12;$59 = $5;
  } else {
   $$pre$i = HEAP32[$3>>2]|0; //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $$04$i = $2;$7 = $$pre$i;
   while(1) {
    $6 = ((($7)) + -136|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $8 = ((($$04$i)) + -136|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    dest=$6; src=$8; stop=dest+84|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 482 "modules/graphics/opengl/Graphics.h"
    $9 = ((($7)) + -52|0); //@line 99 "./common/Object.h"
    $10 = ((($$04$i)) + -52|0); //@line 139 "./common/Object.h"
    $11 = HEAP32[$10>>2]|0; //@line 139 "./common/Object.h"
    HEAP32[$9>>2] = $11; //@line 99 "./common/Object.h"
    $12 = ($11|0)==(0|0); //@line 101 "./common/Object.h"
    if (!($12)) {
     $13 = HEAP32[$11>>2]|0; //@line 101 "./common/Object.h"
     $14 = ((($13)) + 8|0); //@line 101 "./common/Object.h"
     $15 = HEAP32[$14>>2]|0; //@line 101 "./common/Object.h"
     FUNCTION_TABLE_vi[$15 & 8191]($11); //@line 101 "./common/Object.h"
    }
    $16 = ((($7)) + -48|0); //@line 99 "./common/Object.h"
    $17 = ((($$04$i)) + -48|0); //@line 139 "./common/Object.h"
    $18 = HEAP32[$17>>2]|0; //@line 139 "./common/Object.h"
    HEAP32[$16>>2] = $18; //@line 99 "./common/Object.h"
    $19 = ($18|0)==(0|0); //@line 101 "./common/Object.h"
    if (!($19)) {
     $20 = HEAP32[$18>>2]|0; //@line 101 "./common/Object.h"
     $21 = ((($20)) + 8|0); //@line 101 "./common/Object.h"
     $22 = HEAP32[$21>>2]|0; //@line 101 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($22|0,($18|0)); //@line 101 "./common/Object.h"
     $23 = __THREW__; __THREW__ = 0;
     $24 = $23&1;
     if ($24) {
      $$lcssa = $9;
      label = 9;
      break;
     }
    }
    $25 = ((($7)) + -44|0); //@line 482 "modules/graphics/opengl/Graphics.h"
    $26 = ((($$04$i)) + -44|0); //@line 482 "modules/graphics/opengl/Graphics.h"
    __THREW__ = 0;
    invoke_vii(4055,($25|0),($26|0)); //@line 482 "modules/graphics/opengl/Graphics.h"
    $27 = __THREW__; __THREW__ = 0;
    $28 = $27&1;
    if ($28) {
     $$lcssa34 = $9;$$lcssa37 = $16;
     label = 10;
     break;
    }
    $52 = ((($7)) + -32|0); //@line 482 "modules/graphics/opengl/Graphics.h"
    $53 = ((($$04$i)) + -32|0); //@line 482 "modules/graphics/opengl/Graphics.h"
    ;HEAP32[$52>>2]=HEAP32[$53>>2]|0;HEAP32[$52+4>>2]=HEAP32[$53+4>>2]|0;HEAP32[$52+8>>2]=HEAP32[$53+8>>2]|0;HEAP32[$52+12>>2]=HEAP32[$53+12>>2]|0;HEAP32[$52+16>>2]=HEAP32[$53+16>>2]|0;HEAP32[$52+20>>2]=HEAP32[$53+20>>2]|0;HEAP32[$52+24>>2]=HEAP32[$53+24>>2]|0;HEAP32[$52+28>>2]=HEAP32[$53+28>>2]|0; //@line 482 "modules/graphics/opengl/Graphics.h"
    $54 = HEAP32[$3>>2]|0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $55 = ((($54)) + -136|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$3>>2] = $55; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $56 = ($8|0)==($0|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    if ($56) {
     $$lcssa39 = $55;
     label = 18;
     break;
    } else {
     $$04$i = $8;$7 = $55;
    }
   }
   if ((label|0) == 9) {
    $29 = ___cxa_find_matching_catch_2()|0;
    $30 = tempRet0;
    $$0$i$i$i$i$i = $30;$$01$i$i$i$i$i = $29;$43 = $$lcssa;
   }
   else if ((label|0) == 10) {
    $31 = ___cxa_find_matching_catch_2()|0;
    $32 = tempRet0;
    $33 = HEAP32[$$lcssa37>>2]|0; //@line 106 "./common/Object.h"
    $34 = ($33|0)==(0|0); //@line 106 "./common/Object.h"
    if ($34) {
     $$0$i$i$i$i$i = $32;$$01$i$i$i$i$i = $31;$43 = $$lcssa34;
    } else {
     $35 = HEAP32[$33>>2]|0; //@line 106 "./common/Object.h"
     $36 = ((($35)) + 12|0); //@line 106 "./common/Object.h"
     $37 = HEAP32[$36>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($37|0,($33|0)); //@line 106 "./common/Object.h"
     $38 = __THREW__; __THREW__ = 0;
     $39 = $38&1;
     if ($39) {
      $40 = ___cxa_find_matching_catch_3(0|0)|0;
      $41 = tempRet0;
      ___clang_call_terminate($40); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     } else {
      $$0$i$i$i$i$i = $32;$$01$i$i$i$i$i = $31;$43 = $$lcssa34;
     }
    }
   }
   else if ((label|0) == 18) {
    $57 = $$lcssa39; //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $$pre = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $$pre$phi15Z2D = $3;$$pre$phiZ2D = $this;$58 = $57;$59 = $$pre;
    break;
   }
   $42 = HEAP32[$43>>2]|0; //@line 106 "./common/Object.h"
   $44 = ($42|0)==(0|0); //@line 106 "./common/Object.h"
   if ($44) {
    ___resumeException($$01$i$i$i$i$i|0);
    // unreachable;
   }
   $45 = HEAP32[$42>>2]|0; //@line 106 "./common/Object.h"
   $46 = ((($45)) + 12|0); //@line 106 "./common/Object.h"
   $47 = HEAP32[$46>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($47|0,($42|0)); //@line 106 "./common/Object.h"
   $48 = __THREW__; __THREW__ = 0;
   $49 = $48&1;
   if ($49) {
    $50 = ___cxa_find_matching_catch_3(0|0)|0;
    $51 = tempRet0;
    ___clang_call_terminate($50); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   } else {
    ___resumeException($$01$i$i$i$i$i|0);
    // unreachable;
   }
  }
 } while(0);
 HEAP32[$$pre$phiZ2D>>2] = $58; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$$pre$phi15Z2D>>2] = $59; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $60 = ((($__v)) + 8|0); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $61 = HEAP32[$1>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $62 = HEAP32[$60>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$1>>2] = $62; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$60>>2] = $61; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $63 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $64 = ((($__v)) + 12|0); //@line 2014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $65 = HEAP32[$63>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $66 = HEAP32[$64>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$63>>2] = $66; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$64>>2] = $65; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $67 = HEAP32[$$pre$phi15Z2D>>2]|0; //@line 901 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$__v>>2] = $67; //@line 901 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 return; //@line 904 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNSt3__114__split_bufferIN4love8graphics6opengl8Graphics12DisplayStateERNS_9allocatorIS5_EEED2Ev($this) {
 $this = $this|0;
 var $$pre$i$i$i$i$i$i$i$i = 0, $$pre$i$i$i$i$i$i$i$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $1 = HEAP32[$0>>2]|0; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $2 = ((($this)) + 8|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $3 = HEAP32[$2>>2]|0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $4 = ($3|0)==($1|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 L1: do {
  if (!($4)) {
   $6 = $3;
   L2: while(1) {
    $5 = ((($6)) + -136|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    HEAP32[$2>>2] = $5; //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $7 = ((($6)) + -44|0); //@line 482 "modules/graphics/opengl/Graphics.h"
    $8 = HEAP32[$7>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $9 = ($8|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($9)) {
     $10 = ((($6)) + -40|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $11 = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $12 = ($11|0)==($8|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($12) {
      $26 = $8;
     } else {
      $14 = $11;
      while(1) {
       $13 = ((($14)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$10>>2] = $13; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $15 = HEAP32[$13>>2]|0; //@line 106 "./common/Object.h"
       $16 = ($15|0)==(0|0); //@line 106 "./common/Object.h"
       if ($16) {
        $18 = $13;
       } else {
        $19 = HEAP32[$15>>2]|0; //@line 106 "./common/Object.h"
        $20 = ((($19)) + 12|0); //@line 106 "./common/Object.h"
        $21 = HEAP32[$20>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($21|0,($15|0)); //@line 106 "./common/Object.h"
        $22 = __THREW__; __THREW__ = 0;
        $23 = $22&1;
        if ($23) {
         label = 8;
         break L2;
        }
        $$pre$i$i$i$i$i$i$i$i$i$i = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $18 = $$pre$i$i$i$i$i$i$i$i$i$i;
       }
       $17 = ($18|0)==($8|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($17) {
        break;
       } else {
        $14 = $18;
       }
      }
      $$pre$i$i$i$i$i$i$i$i = HEAP32[$7>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $26 = $$pre$i$i$i$i$i$i$i$i;
     }
     __ZdlPv($26); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    }
    $27 = ((($6)) + -48|0); //@line 106 "./common/Object.h"
    $28 = HEAP32[$27>>2]|0; //@line 106 "./common/Object.h"
    $29 = ($28|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($29)) {
     $30 = HEAP32[$28>>2]|0; //@line 106 "./common/Object.h"
     $31 = ((($30)) + 12|0); //@line 106 "./common/Object.h"
     $32 = HEAP32[$31>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($32|0,($28|0)); //@line 106 "./common/Object.h"
     $33 = __THREW__; __THREW__ = 0;
     $34 = $33&1;
     if ($34) {
      label = 13;
      break;
     }
    }
    $37 = ((($6)) + -52|0); //@line 106 "./common/Object.h"
    $38 = HEAP32[$37>>2]|0; //@line 106 "./common/Object.h"
    $39 = ($38|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($39)) {
     $42 = HEAP32[$38>>2]|0; //@line 106 "./common/Object.h"
     $43 = ((($42)) + 12|0); //@line 106 "./common/Object.h"
     $44 = HEAP32[$43>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($44|0,($38|0)); //@line 106 "./common/Object.h"
     $45 = __THREW__; __THREW__ = 0;
     $46 = $45&1;
     if ($46) {
      label = 17;
      break;
     }
    }
    $40 = HEAP32[$2>>2]|0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $41 = ($40|0)==($1|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($41) {
     break L1;
    } else {
     $6 = $40;
    }
   }
   if ((label|0) == 8) {
    $24 = ___cxa_find_matching_catch_3(0|0)|0;
    $25 = tempRet0;
    ___clang_call_terminate($24); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 13) {
    $35 = ___cxa_find_matching_catch_3(0|0)|0;
    $36 = tempRet0;
    ___clang_call_terminate($35); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 17) {
    $47 = ___cxa_find_matching_catch_3(0|0)|0;
    $48 = tempRet0;
    ___clang_call_terminate($47); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
  }
 } while(0);
 $49 = HEAP32[$this>>2]|0; //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $50 = ($49|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($50) {
  return; //@line 358 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 }
 __ZdlPv($49); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 358 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
}
function __ZNSt3__16vectorIN4love8graphics6opengl8Graphics12DisplayStateENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $8 = 0, $9 = 0, $__v = 0;
 var $eh$lpad$body$index3Z2D = 0, $eh$lpad$body$indexZ2D = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__v = sp;
 $0 = ((($this)) + 8|0); //@line 350 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($2) - ($3))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4|0) / 136)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = (($5) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ($6>>>0)>(31580641); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($7) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $11 = $$pre;
 } else {
  $11 = $3;
 }
 $8 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = HEAP32[$8>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9) - ($11))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = (($10|0) / 136)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ($12>>>0)<(15790320); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($13) {
  $14 = $12 << 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = ($14>>>0)<($6>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $16 = $15 ? $6 : $14; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i = $16;
 } else {
  $$0$i = 31580641;
 }
 $17 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $18 = (($17) - ($11))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $19 = (($18|0) / 136)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = ((($__v)) + 12|0); //@line 323 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 HEAP32[$20>>2] = 0; //@line 1959 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $21 = ((($__v)) + 16|0); //@line 1959 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$21>>2] = $0; //@line 1959 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $22 = ($$0$i|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($22) {
  $25 = 0;
 } else {
  $23 = ($$0$i*136)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = (__Znwj($23)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $25 = $24;
 }
 HEAP32[$__v>>2] = $25; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $26 = (($25) + (($19*136)|0)|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = ((($__v)) + 8|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 HEAP32[$27>>2] = $26; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = ((($__v)) + 4|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 HEAP32[$28>>2] = $26; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = (($25) + (($$0$i*136)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = ((($__v)) + 12|0); //@line 2014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$30>>2] = $29; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 dest=$26; src=$__x; stop=dest+84|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 482 "modules/graphics/opengl/Graphics.h"
 $31 = (((($25) + (($19*136)|0)|0)) + 84|0); //@line 99 "./common/Object.h"
 $32 = ((($__x)) + 84|0); //@line 139 "./common/Object.h"
 $33 = HEAP32[$32>>2]|0; //@line 139 "./common/Object.h"
 HEAP32[$31>>2] = $33; //@line 99 "./common/Object.h"
 $34 = ($33|0)==(0|0); //@line 101 "./common/Object.h"
 if ($34) {
  label = 9;
 } else {
  $35 = HEAP32[$33>>2]|0; //@line 101 "./common/Object.h"
  $36 = ((($35)) + 8|0); //@line 101 "./common/Object.h"
  $37 = HEAP32[$36>>2]|0; //@line 101 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($37|0,($33|0)); //@line 101 "./common/Object.h"
  $38 = __THREW__; __THREW__ = 0;
  $39 = $38&1;
  if (!($39)) {
   label = 9;
  }
 }
 if ((label|0) == 9) {
  $40 = (((($25) + (($19*136)|0)|0)) + 88|0); //@line 99 "./common/Object.h"
  $41 = ((($__x)) + 88|0); //@line 139 "./common/Object.h"
  $42 = HEAP32[$41>>2]|0; //@line 139 "./common/Object.h"
  HEAP32[$40>>2] = $42; //@line 99 "./common/Object.h"
  $43 = ($42|0)==(0|0); //@line 101 "./common/Object.h"
  if (!($43)) {
   $44 = HEAP32[$42>>2]|0; //@line 101 "./common/Object.h"
   $45 = ((($44)) + 8|0); //@line 101 "./common/Object.h"
   $46 = HEAP32[$45>>2]|0; //@line 101 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($46|0,($42|0)); //@line 101 "./common/Object.h"
   $47 = __THREW__; __THREW__ = 0;
   $48 = $47&1;
   if ($48) {
    $49 = ___cxa_find_matching_catch_2()|0;
    $50 = tempRet0;
    $51 = HEAP32[$31>>2]|0; //@line 106 "./common/Object.h"
    $52 = ($51|0)==(0|0); //@line 106 "./common/Object.h"
    if ($52) {
     $eh$lpad$body$index3Z2D = $50;$eh$lpad$body$indexZ2D = $49;
     __ZNSt3__114__split_bufferIN4love8graphics6opengl8Graphics12DisplayStateERNS_9allocatorIS5_EEED2Ev($__v); //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     ___resumeException($eh$lpad$body$indexZ2D|0);
     // unreachable;
    }
    $53 = HEAP32[$51>>2]|0; //@line 106 "./common/Object.h"
    $54 = ((($53)) + 12|0); //@line 106 "./common/Object.h"
    $55 = HEAP32[$54>>2]|0; //@line 106 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($55|0,($51|0)); //@line 106 "./common/Object.h"
    $56 = __THREW__; __THREW__ = 0;
    $57 = $56&1;
    if ($57) {
     $58 = ___cxa_find_matching_catch_3(0|0)|0;
     $59 = tempRet0;
     ___clang_call_terminate($58); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    } else {
     $eh$lpad$body$index3Z2D = $50;$eh$lpad$body$indexZ2D = $49;
     __ZNSt3__114__split_bufferIN4love8graphics6opengl8Graphics12DisplayStateERNS_9allocatorIS5_EEED2Ev($__v); //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     ___resumeException($eh$lpad$body$indexZ2D|0);
     // unreachable;
    }
   }
  }
  $60 = (((($25) + (($19*136)|0)|0)) + 92|0); //@line 482 "modules/graphics/opengl/Graphics.h"
  $61 = ((($__x)) + 92|0); //@line 482 "modules/graphics/opengl/Graphics.h"
  $62 = (((($25) + (($19*136)|0)|0)) + 96|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $63 = (((($25) + (($19*136)|0)|0)) + 100|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$63>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $64 = HEAP32[$61>>2]|0; //@line 1246 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$60>>2] = $64; //@line 1246 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $65 = ((($__x)) + 96|0); //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $66 = HEAP32[$65>>2]|0; //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$62>>2] = $66; //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $67 = ((($__x)) + 100|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $68 = HEAP32[$67>>2]|0; //@line 1248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $69 = (((($25) + (($19*136)|0)|0)) + 100|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$69>>2] = $68; //@line 1248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$67>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$65>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$61>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $70 = (((($25) + (($19*136)|0)|0)) + 104|0); //@line 482 "modules/graphics/opengl/Graphics.h"
  $71 = ((($__x)) + 104|0); //@line 482 "modules/graphics/opengl/Graphics.h"
  ;HEAP32[$70>>2]=HEAP32[$71>>2]|0;HEAP32[$70+4>>2]=HEAP32[$71+4>>2]|0;HEAP32[$70+8>>2]=HEAP32[$71+8>>2]|0;HEAP32[$70+12>>2]=HEAP32[$71+12>>2]|0;HEAP32[$70+16>>2]=HEAP32[$71+16>>2]|0;HEAP32[$70+20>>2]=HEAP32[$71+20>>2]|0;HEAP32[$70+24>>2]=HEAP32[$71+24>>2]|0;HEAP32[$70+28>>2]=HEAP32[$71+28>>2]|0; //@line 482 "modules/graphics/opengl/Graphics.h"
  $72 = ((($26)) + 136|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$27>>2] = $72; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  __THREW__ = 0;
  invoke_vii(4053,($this|0),($__v|0)); //@line 1583 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $73 = __THREW__; __THREW__ = 0;
  $74 = $73&1;
  if (!($74)) {
   __ZNSt3__114__split_bufferIN4love8graphics6opengl8Graphics12DisplayStateERNS_9allocatorIS5_EEED2Ev($__v); //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   STACKTOP = sp;return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
 }
 $75 = ___cxa_find_matching_catch_2()|0;
 $76 = tempRet0;
 $eh$lpad$body$index3Z2D = $76;$eh$lpad$body$indexZ2D = $75;
 __ZNSt3__114__split_bufferIN4love8graphics6opengl8Graphics12DisplayStateERNS_9allocatorIS5_EEED2Ev($__v); //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
}
function __ZN4love8graphics6opengl8Graphics8setColorENS0_6ColorTIfEE($this,$c) {
 $this = $this|0;
 $c = $c|0;
 var $0 = 0.0, $1 = 0.0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0.0;
 var $8 = 0, $9 = 0.0, $nc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $nc = sp;
 $0 = +HEAPF32[$c>>2]; //@line 949 "modules/graphics/opengl/Graphics.cpp"
 $1 = $0 / 255.0; //@line 949 "modules/graphics/opengl/Graphics.cpp"
 $2 = ((($c)) + 4|0); //@line 949 "modules/graphics/opengl/Graphics.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 949 "modules/graphics/opengl/Graphics.cpp"
 $4 = $3 / 255.0; //@line 949 "modules/graphics/opengl/Graphics.cpp"
 $5 = ((($c)) + 8|0); //@line 949 "modules/graphics/opengl/Graphics.cpp"
 $6 = +HEAPF32[$5>>2]; //@line 949 "modules/graphics/opengl/Graphics.cpp"
 $7 = $6 / 255.0; //@line 949 "modules/graphics/opengl/Graphics.cpp"
 $8 = ((($c)) + 12|0); //@line 949 "modules/graphics/opengl/Graphics.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 949 "modules/graphics/opengl/Graphics.cpp"
 $10 = $9 / 255.0; //@line 949 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$nc>>2] = $1; //@line 38 "./modules/graphics/Color.h"
 $11 = ((($nc)) + 4|0); //@line 38 "./modules/graphics/Color.h"
 HEAPF32[$11>>2] = $4; //@line 38 "./modules/graphics/Color.h"
 $12 = ((($nc)) + 8|0); //@line 38 "./modules/graphics/Color.h"
 HEAPF32[$12>>2] = $7; //@line 38 "./modules/graphics/Color.h"
 $13 = ((($nc)) + 12|0); //@line 38 "./modules/graphics/Color.h"
 HEAPF32[$13>>2] = $10; //@line 38 "./modules/graphics/Color.h"
 __ZN4love8graphics17gammaCorrectColorERNS0_6ColorTIfEE($nc); //@line 951 "modules/graphics/opengl/Graphics.cpp"
 $14 = +HEAPF32[$nc>>2]; //@line 953 "modules/graphics/opengl/Graphics.cpp"
 $15 = +HEAPF32[$11>>2]; //@line 953 "modules/graphics/opengl/Graphics.cpp"
 $16 = +HEAPF32[$12>>2]; //@line 953 "modules/graphics/opengl/Graphics.cpp"
 $17 = +HEAPF32[$13>>2]; //@line 953 "modules/graphics/opengl/Graphics.cpp"
 $18 = HEAP32[165163]|0; //@line 583 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vidddd[$18 & 8191](3,$14,$15,$16,$17); //@line 583 "./libraries/glad/gladfuncs.hpp"
 $19 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = HEAP32[$19>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = ((($20)) + -136|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 ;HEAP32[$21>>2]=HEAP32[$c>>2]|0;HEAP32[$21+4>>2]=HEAP32[$c+4>>2]|0;HEAP32[$21+8>>2]=HEAP32[$c+8>>2]|0;HEAP32[$21+12>>2]=HEAP32[$c+12>>2]|0; //@line 954 "modules/graphics/opengl/Graphics.cpp"
 STACKTOP = sp;return; //@line 955 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics18setBackgroundColorENS0_6ColorTIfEE($this,$c) {
 $this = $this|0;
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -120|0); //@line 964 "modules/graphics/opengl/Graphics.cpp"
 ;HEAP32[$2>>2]=HEAP32[$c>>2]|0;HEAP32[$2+4>>2]=HEAP32[$c+4>>2]|0;HEAP32[$2+8>>2]=HEAP32[$c+8>>2]|0;HEAP32[$2+12>>2]=HEAP32[$c+12>>2]|0; //@line 964 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 965 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics12setLineWidthEf($this,$width) {
 $this = $this|0;
 $width = +$width;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -96|0); //@line 1202 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$2>>2] = $width; //@line 1202 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1203 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics12setLineStyleENS0_8Graphics9LineStyleE($this,$style) {
 $this = $this|0;
 $style = $style|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -92|0); //@line 1207 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$2>>2] = $style; //@line 1207 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1208 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics11setLineJoinENS0_8Graphics8LineJoinE($this,$join) {
 $this = $this|0;
 $join = $join|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -88|0); //@line 1212 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$2>>2] = $join; //@line 1212 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1213 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics12setPointSizeEf($this,$size) {
 $this = $this|0;
 $size = +$size;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl6OpenGL12setPointSizeEf(634044,$size); //@line 1232 "modules/graphics/opengl/Graphics.cpp"
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -84|0); //@line 1233 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$2>>2] = $size; //@line 1233 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1234 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics10setScissorEiiii($this,$x,$y,$width,$height) {
 $this = $this|0;
 $x = $x|0;
 $y = $y|0;
 $width = $width|0;
 $height = $height|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = HEAP32[164832]|0; //@line 57 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$1 & 8191](3089); //@line 57 "./libraries/glad/gladfuncs.hpp"
 HEAP32[$0>>2] = $x; //@line 629 "modules/graphics/opengl/Graphics.cpp"
 $2 = ((($0)) + 4|0); //@line 629 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$2>>2] = $y; //@line 629 "modules/graphics/opengl/Graphics.cpp"
 $3 = ((($0)) + 8|0); //@line 629 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$3>>2] = $width; //@line 629 "modules/graphics/opengl/Graphics.cpp"
 $4 = ((($0)) + 12|0); //@line 629 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$4>>2] = $height; //@line 629 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl6OpenGL10setScissorERKNS2_8ViewportE(634044,$0); //@line 629 "modules/graphics/opengl/Graphics.cpp"
 $5 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$5>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ((($6)) + -80|0); //@line 631 "modules/graphics/opengl/Graphics.cpp"
 HEAP8[$7>>0] = 1; //@line 631 "modules/graphics/opengl/Graphics.cpp"
 $8 = ((($6)) + -76|0); //@line 632 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$8>>2] = $x; //@line 632 "modules/graphics/opengl/Graphics.cpp"
 $9 = ((($6)) + -72|0); //@line 632 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$9>>2] = $y; //@line 632 "modules/graphics/opengl/Graphics.cpp"
 $10 = ((($6)) + -68|0); //@line 632 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$10>>2] = $width; //@line 632 "modules/graphics/opengl/Graphics.cpp"
 $11 = ((($6)) + -64|0); //@line 632 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$11>>2] = $height; //@line 632 "modules/graphics/opengl/Graphics.cpp"
 STACKTOP = sp;return; //@line 633 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics10setScissorEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -80|0); //@line 658 "modules/graphics/opengl/Graphics.cpp"
 HEAP8[$2>>0] = 0; //@line 658 "modules/graphics/opengl/Graphics.cpp"
 $3 = HEAP32[165193]|0; //@line 56 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$3 & 8191](3089); //@line 56 "./libraries/glad/gladfuncs.hpp"
 return; //@line 660 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics14setStencilTestENS0_8Graphics11CompareModeEi($this,$compare,$value) {
 $this = $this|0;
 $compare = $compare|0;
 $value = $value|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $glcompare$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -60|0); //@line 735 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$2>>2] = $compare; //@line 735 "modules/graphics/opengl/Graphics.cpp"
 $3 = ((($1)) + -56|0); //@line 736 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$3>>2] = $value; //@line 736 "modules/graphics/opengl/Graphics.cpp"
 $4 = ((($this)) + 42|0); //@line 738 "modules/graphics/opengl/Graphics.cpp"
 $5 = HEAP8[$4>>0]|0; //@line 738 "modules/graphics/opengl/Graphics.cpp"
 $6 = ($5<<24>>24)==(0); //@line 738 "modules/graphics/opengl/Graphics.cpp"
 if (!($6)) {
  return; //@line 794 "modules/graphics/opengl/Graphics.cpp"
 }
 $7 = ($compare|0)==(6); //@line 741 "modules/graphics/opengl/Graphics.cpp"
 if ($7) {
  $8 = HEAP32[165193]|0; //@line 56 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$8 & 8191](2960); //@line 56 "./libraries/glad/gladfuncs.hpp"
  return; //@line 794 "modules/graphics/opengl/Graphics.cpp"
 }
 $9 = HEAP32[158272]|0; //@line 748 "modules/graphics/opengl/Graphics.cpp"
 $10 = ($9|0)==(0|0); //@line 748 "modules/graphics/opengl/Graphics.cpp"
 if (!($10)) {
  (__ZN4love8graphics6opengl6Canvas18checkCreateStencilEv($9)|0); //@line 749 "modules/graphics/opengl/Graphics.cpp"
 }
 switch ($compare|0) {
 case 0:  {
  $glcompare$0 = 516;
  break;
 }
 case 1:  {
  $glcompare$0 = 518;
  break;
 }
 case 5:  {
  $glcompare$0 = 517;
  break;
 }
 case 3:  {
  $glcompare$0 = 515;
  break;
 }
 case 4:  {
  $glcompare$0 = 513;
  break;
 }
 default: {
  $glcompare$0 = 514;
 }
 }
 $11 = HEAP32[164832]|0; //@line 57 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$11 & 8191](2960); //@line 57 "./libraries/glad/gladfuncs.hpp"
 $12 = HEAP32[165322]|0; //@line 62 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$12 & 8191]($glcompare$0,$value,-1); //@line 62 "./libraries/glad/gladfuncs.hpp"
 $13 = HEAP32[164775]|0; //@line 63 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$13 & 8191](7680,7680,7680); //@line 63 "./libraries/glad/gladfuncs.hpp"
 return; //@line 794 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics7setFontEPNS1_4FontE($this,$font) {
 $this = $this|0;
 $font = $font|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ($font|0)==(0|0); //@line 132 "./common/Object.h"
 if (!($2)) {
  $3 = HEAP32[$font>>2]|0; //@line 132 "./common/Object.h"
  $4 = ((($3)) + 8|0); //@line 132 "./common/Object.h"
  $5 = HEAP32[$4>>2]|0; //@line 132 "./common/Object.h"
  FUNCTION_TABLE_vi[$5 & 8191]($font); //@line 132 "./common/Object.h"
 }
 $6 = ((($1)) + -52|0); //@line 133 "./common/Object.h"
 $7 = HEAP32[$6>>2]|0; //@line 133 "./common/Object.h"
 $8 = ($7|0)==(0|0); //@line 133 "./common/Object.h"
 if ($8) {
  HEAP32[$6>>2] = $font; //@line 134 "./common/Object.h"
  return; //@line 979 "modules/graphics/opengl/Graphics.cpp"
 }
 $9 = HEAP32[$7>>2]|0; //@line 133 "./common/Object.h"
 $10 = ((($9)) + 12|0); //@line 133 "./common/Object.h"
 $11 = HEAP32[$10>>2]|0; //@line 133 "./common/Object.h"
 FUNCTION_TABLE_vi[$11 & 8191]($7); //@line 133 "./common/Object.h"
 HEAP32[$6>>2] = $font; //@line 134 "./common/Object.h"
 return; //@line 979 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics9setShaderEPNS1_6ShaderE($this,$shader) {
 $this = $this|0;
 $shader = $shader|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($shader|0)==(0|0); //@line 989 "modules/graphics/opengl/Graphics.cpp"
 $1 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($0) {
  __ZN4love8graphics6opengl6Shader6detachEv(); //@line 1004 "modules/graphics/opengl/Graphics.cpp"
  $3 = ((($2)) + -48|0); //@line 133 "./common/Object.h"
  $4 = HEAP32[$3>>2]|0; //@line 133 "./common/Object.h"
  $5 = ($4|0)==(0|0); //@line 133 "./common/Object.h"
  if (!($5)) {
   $6 = HEAP32[$4>>2]|0; //@line 133 "./common/Object.h"
   $7 = ((($6)) + 12|0); //@line 133 "./common/Object.h"
   $8 = HEAP32[$7>>2]|0; //@line 133 "./common/Object.h"
   FUNCTION_TABLE_vi[$8 & 8191]($4); //@line 133 "./common/Object.h"
  }
  HEAP32[$3>>2] = 0; //@line 134 "./common/Object.h"
  return; //@line 997 "modules/graphics/opengl/Graphics.cpp"
 } else {
  __ZN4love8graphics6opengl6Shader6attachEb($shader,0); //@line 994 "modules/graphics/opengl/Graphics.cpp"
  $9 = HEAP32[$shader>>2]|0; //@line 132 "./common/Object.h"
  $10 = ((($9)) + 8|0); //@line 132 "./common/Object.h"
  $11 = HEAP32[$10>>2]|0; //@line 132 "./common/Object.h"
  FUNCTION_TABLE_vi[$11 & 8191]($shader); //@line 132 "./common/Object.h"
  $12 = ((($2)) + -48|0); //@line 133 "./common/Object.h"
  $13 = HEAP32[$12>>2]|0; //@line 133 "./common/Object.h"
  $14 = ($13|0)==(0|0); //@line 133 "./common/Object.h"
  if (!($14)) {
   $15 = HEAP32[$13>>2]|0; //@line 133 "./common/Object.h"
   $16 = ((($15)) + 12|0); //@line 133 "./common/Object.h"
   $17 = HEAP32[$16>>2]|0; //@line 133 "./common/Object.h"
   FUNCTION_TABLE_vi[$17 & 8191]($13); //@line 133 "./common/Object.h"
  }
  HEAP32[$12>>2] = $shader; //@line 134 "./common/Object.h"
  return; //@line 997 "modules/graphics/opengl/Graphics.cpp"
 }
}
function __ZN4love8graphics6opengl8Graphics12setColorMaskENS0_8Graphics9ColorMaskE($this,$mask) {
 $this = $this|0;
 $mask = $mask|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$mask>>0]|0; //@line 1084 "modules/graphics/opengl/Graphics.cpp"
 $1 = ((($mask)) + 1|0); //@line 1084 "modules/graphics/opengl/Graphics.cpp"
 $2 = HEAP8[$1>>0]|0; //@line 1084 "modules/graphics/opengl/Graphics.cpp"
 $3 = ((($mask)) + 2|0); //@line 1084 "modules/graphics/opengl/Graphics.cpp"
 $4 = HEAP8[$3>>0]|0; //@line 1084 "modules/graphics/opengl/Graphics.cpp"
 $5 = ((($mask)) + 3|0); //@line 1084 "modules/graphics/opengl/Graphics.cpp"
 $6 = HEAP8[$5>>0]|0; //@line 1084 "modules/graphics/opengl/Graphics.cpp"
 $7 = HEAP32[165466]|0; //@line 54 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiii[$7 & 8191]($0,$2,$4,$6); //@line 54 "./libraries/glad/gladfuncs.hpp"
 $8 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = HEAP32[$8>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = ((($9)) + -32|0); //@line 1085 "modules/graphics/opengl/Graphics.cpp"
 $11 = HEAPU8[$mask>>0]|(HEAPU8[$mask+1>>0]<<8)|(HEAPU8[$mask+2>>0]<<16)|(HEAPU8[$mask+3>>0]<<24); //@line 1085 "modules/graphics/opengl/Graphics.cpp"
 HEAP8[$10>>0]=$11&255;HEAP8[$10+1>>0]=($11>>8)&255;HEAP8[$10+2>>0]=($11>>16)&255;HEAP8[$10+3>>0]=$11>>24; //@line 1085 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1086 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics12setWireframeEb($this,$enable) {
 $this = $this|0;
 $enable = $enable|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[1252012]|0; //@line 1244 "modules/graphics/opengl/Graphics.cpp"
 $1 = ($0<<24>>24)==(0); //@line 1244 "modules/graphics/opengl/Graphics.cpp"
 if (!($1)) {
  return; //@line 1249 "modules/graphics/opengl/Graphics.cpp"
 }
 $2 = $enable ? 6913 : 6914; //@line 1247 "modules/graphics/opengl/Graphics.cpp"
 $3 = HEAP32[165045]|0; //@line 40 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$3 & 8191](1032,$2); //@line 40 "./libraries/glad/gladfuncs.hpp"
 $4 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$4>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ((($5)) + -28|0); //@line 1248 "modules/graphics/opengl/Graphics.cpp"
 $7 = $enable&1; //@line 1248 "modules/graphics/opengl/Graphics.cpp"
 HEAP8[$6>>0] = $7; //@line 1248 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1249 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics16setDefaultFilterERKNS0_7Texture6FilterE($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics7Texture16setDefaultFilterERKNS1_6FilterE($f); //@line 1176 "modules/graphics/opengl/Graphics.cpp"
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -24|0); //@line 1177 "modules/graphics/opengl/Graphics.cpp"
 ;HEAP32[$2>>2]=HEAP32[$f>>2]|0;HEAP32[$2+4>>2]=HEAP32[$f+4>>2]|0;HEAP32[$2+8>>2]=HEAP32[$f+8>>2]|0;HEAP32[$2+12>>2]=HEAP32[$f+12>>2]|0; //@line 1177 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1178 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics22setDefaultMipmapFilterENS0_7Texture10FilterModeEf($this,$filter,$sharpness) {
 $this = $this|0;
 $filter = $filter|0;
 $sharpness = +$sharpness;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl5Image22setDefaultMipmapFilterENS0_7Texture10FilterModeE($filter); //@line 1187 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl5Image25setDefaultMipmapSharpnessEf($sharpness); //@line 1188 "modules/graphics/opengl/Graphics.cpp"
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -8|0); //@line 1190 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$2>>2] = $filter; //@line 1190 "modules/graphics/opengl/Graphics.cpp"
 $3 = ((($1)) + -4|0); //@line 1191 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$3>>2] = $sharpness; //@line 1191 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1192 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics19restoreStateCheckedERKNS2_12DisplayStateE($this,$s) {
 $this = $this|0;
 $s = $s|0;
 var $$pre11 = 0, $0 = 0, $1 = 0, $10 = 0.0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0.0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0.0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0.0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0.0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0.0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0.0, $240 = 0;
 var $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0.0, $250 = 0.0, $251 = 0, $252 = 0, $253 = 0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0.0, $30 = 0;
 var $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0.0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0.0, $67 = 0;
 var $68 = 0.0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0.0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $glcompare$0$i = 0, $i$09 = 0, $nc$i = 0, $or$cond = 0, $phitmp$i = 0;
 var $phitmp$i3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 16|0;
 $nc$i = sp;
 $1 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = +HEAPF32[$s>>2]; //@line 140 "modules/graphics/opengl/Graphics.cpp"
 $4 = ((($2)) + -136|0); //@line 58 "./modules/graphics/Color.h"
 $5 = +HEAPF32[$4>>2]; //@line 58 "./modules/graphics/Color.h"
 $6 = $3 == $5; //@line 58 "./modules/graphics/Color.h"
 $7 = ((($s)) + 4|0); //@line 58 "./modules/graphics/Color.h"
 $8 = +HEAPF32[$7>>2]; //@line 58 "./modules/graphics/Color.h"
 if ($6) {
  $9 = ((($2)) + -132|0); //@line 58 "./modules/graphics/Color.h"
  $10 = +HEAPF32[$9>>2]; //@line 58 "./modules/graphics/Color.h"
  $11 = $8 == $10; //@line 58 "./modules/graphics/Color.h"
  if ($11) {
   $12 = ((($s)) + 8|0); //@line 58 "./modules/graphics/Color.h"
   $13 = +HEAPF32[$12>>2]; //@line 58 "./modules/graphics/Color.h"
   $14 = ((($2)) + -128|0); //@line 58 "./modules/graphics/Color.h"
   $15 = +HEAPF32[$14>>2]; //@line 58 "./modules/graphics/Color.h"
   $16 = $13 == $15; //@line 58 "./modules/graphics/Color.h"
   if ($16) {
    $17 = ((($s)) + 12|0); //@line 58 "./modules/graphics/Color.h"
    $18 = +HEAPF32[$17>>2]; //@line 58 "./modules/graphics/Color.h"
    $19 = ((($2)) + -124|0); //@line 58 "./modules/graphics/Color.h"
    $20 = +HEAPF32[$19>>2]; //@line 58 "./modules/graphics/Color.h"
    $phitmp$i = $18 != $20; //@line 64 "./modules/graphics/Color.h"
    if ($phitmp$i) {
     label = 5;
    } else {
     $44 = $2;
    }
   } else {
    label = 5;
   }
  } else {
   label = 5;
  }
 } else {
  label = 5;
 }
 if ((label|0) == 5) {
  $21 = ((($s)) + 8|0); //@line 140 "modules/graphics/opengl/Graphics.cpp"
  $22 = +HEAPF32[$21>>2]; //@line 140 "modules/graphics/opengl/Graphics.cpp"
  $23 = ((($s)) + 12|0); //@line 140 "modules/graphics/opengl/Graphics.cpp"
  $24 = +HEAPF32[$23>>2]; //@line 140 "modules/graphics/opengl/Graphics.cpp"
  $25 = $3 / 255.0; //@line 949 "modules/graphics/opengl/Graphics.cpp"
  $26 = $8 / 255.0; //@line 949 "modules/graphics/opengl/Graphics.cpp"
  $27 = $22 / 255.0; //@line 949 "modules/graphics/opengl/Graphics.cpp"
  $28 = $24 / 255.0; //@line 949 "modules/graphics/opengl/Graphics.cpp"
  HEAPF32[$nc$i>>2] = $25; //@line 38 "./modules/graphics/Color.h"
  $29 = ((($nc$i)) + 4|0); //@line 38 "./modules/graphics/Color.h"
  HEAPF32[$29>>2] = $26; //@line 38 "./modules/graphics/Color.h"
  $30 = ((($nc$i)) + 8|0); //@line 38 "./modules/graphics/Color.h"
  HEAPF32[$30>>2] = $27; //@line 38 "./modules/graphics/Color.h"
  $31 = ((($nc$i)) + 12|0); //@line 38 "./modules/graphics/Color.h"
  HEAPF32[$31>>2] = $28; //@line 38 "./modules/graphics/Color.h"
  __ZN4love8graphics17gammaCorrectColorERNS0_6ColorTIfEE($nc$i); //@line 951 "modules/graphics/opengl/Graphics.cpp"
  $32 = +HEAPF32[$nc$i>>2]; //@line 953 "modules/graphics/opengl/Graphics.cpp"
  $33 = +HEAPF32[$29>>2]; //@line 953 "modules/graphics/opengl/Graphics.cpp"
  $34 = +HEAPF32[$30>>2]; //@line 953 "modules/graphics/opengl/Graphics.cpp"
  $35 = +HEAPF32[$31>>2]; //@line 953 "modules/graphics/opengl/Graphics.cpp"
  $36 = HEAP32[165163]|0; //@line 583 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vidddd[$36 & 8191](3,$32,$33,$34,$35); //@line 583 "./libraries/glad/gladfuncs.hpp"
  $37 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $38 = ((($37)) + -136|0); //@line 954 "modules/graphics/opengl/Graphics.cpp"
  HEAPF32[$38>>2] = $3; //@line 954 "modules/graphics/opengl/Graphics.cpp"
  $39 = ((($37)) + -132|0); //@line 954 "modules/graphics/opengl/Graphics.cpp"
  HEAPF32[$39>>2] = $8; //@line 954 "modules/graphics/opengl/Graphics.cpp"
  $40 = ((($37)) + -128|0); //@line 954 "modules/graphics/opengl/Graphics.cpp"
  HEAPF32[$40>>2] = $22; //@line 954 "modules/graphics/opengl/Graphics.cpp"
  $41 = ((($37)) + -124|0); //@line 954 "modules/graphics/opengl/Graphics.cpp"
  HEAPF32[$41>>2] = $24; //@line 954 "modules/graphics/opengl/Graphics.cpp"
  $$pre11 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $44 = $$pre11;
 }
 $42 = ((($s)) + 16|0); //@line 142 "modules/graphics/opengl/Graphics.cpp"
 $43 = ((($44)) + -120|0); //@line 964 "modules/graphics/opengl/Graphics.cpp"
 _memmove(($43|0),($42|0),16)|0; //@line 964 "modules/graphics/opengl/Graphics.cpp"
 $45 = ((($s)) + 32|0); //@line 144 "modules/graphics/opengl/Graphics.cpp"
 $46 = HEAP32[$45>>2]|0; //@line 145 "modules/graphics/opengl/Graphics.cpp"
 $47 = ((($2)) + -104|0); //@line 144 "modules/graphics/opengl/Graphics.cpp"
 $48 = HEAP32[$47>>2]|0; //@line 144 "modules/graphics/opengl/Graphics.cpp"
 $49 = ($46|0)==($48|0); //@line 144 "modules/graphics/opengl/Graphics.cpp"
 $50 = ((($s)) + 36|0); //@line 144 "modules/graphics/opengl/Graphics.cpp"
 $51 = HEAP32[$50>>2]|0; //@line 144 "modules/graphics/opengl/Graphics.cpp"
 if ($49) {
  $52 = ((($2)) + -100|0); //@line 144 "modules/graphics/opengl/Graphics.cpp"
  $53 = HEAP32[$52>>2]|0; //@line 144 "modules/graphics/opengl/Graphics.cpp"
  $54 = ($51|0)==($53|0); //@line 144 "modules/graphics/opengl/Graphics.cpp"
  if (!($54)) {
   label = 8;
  }
 } else {
  label = 8;
 }
 if ((label|0) == 8) {
  __ZN4love8graphics6opengl8Graphics12setBlendModeENS0_8Graphics9BlendModeENS3_10BlendAlphaE($this,$46,$51); //@line 145 "modules/graphics/opengl/Graphics.cpp"
 }
 $55 = ((($s)) + 40|0); //@line 148 "modules/graphics/opengl/Graphics.cpp"
 $56 = HEAP32[$55>>2]|0; //@line 148 "modules/graphics/opengl/Graphics.cpp"
 $57 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $58 = ((($57)) + -96|0); //@line 1202 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$58>>2] = $56; //@line 1202 "modules/graphics/opengl/Graphics.cpp"
 $59 = ((($s)) + 44|0); //@line 149 "modules/graphics/opengl/Graphics.cpp"
 $60 = HEAP32[$59>>2]|0; //@line 149 "modules/graphics/opengl/Graphics.cpp"
 $61 = ((($57)) + -92|0); //@line 1207 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$61>>2] = $60; //@line 1207 "modules/graphics/opengl/Graphics.cpp"
 $62 = ((($s)) + 48|0); //@line 150 "modules/graphics/opengl/Graphics.cpp"
 $63 = HEAP32[$62>>2]|0; //@line 150 "modules/graphics/opengl/Graphics.cpp"
 $64 = ((($57)) + -88|0); //@line 1212 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$64>>2] = $63; //@line 1212 "modules/graphics/opengl/Graphics.cpp"
 $65 = ((($s)) + 52|0); //@line 152 "modules/graphics/opengl/Graphics.cpp"
 $66 = +HEAPF32[$65>>2]; //@line 152 "modules/graphics/opengl/Graphics.cpp"
 $67 = ((($2)) + -84|0); //@line 152 "modules/graphics/opengl/Graphics.cpp"
 $68 = +HEAPF32[$67>>2]; //@line 152 "modules/graphics/opengl/Graphics.cpp"
 $69 = $66 != $68; //@line 152 "modules/graphics/opengl/Graphics.cpp"
 if ($69) {
  __ZN4love8graphics6opengl6OpenGL12setPointSizeEf(634044,$66); //@line 1232 "modules/graphics/opengl/Graphics.cpp"
  $70 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $71 = ((($70)) + -84|0); //@line 1233 "modules/graphics/opengl/Graphics.cpp"
  HEAPF32[$71>>2] = $66; //@line 1233 "modules/graphics/opengl/Graphics.cpp"
  $117 = $70;
 } else {
  $117 = $57;
 }
 $72 = ((($s)) + 56|0); //@line 155 "modules/graphics/opengl/Graphics.cpp"
 $73 = HEAP8[$72>>0]|0; //@line 157 "modules/graphics/opengl/Graphics.cpp"
 $74 = ((($2)) + -80|0); //@line 155 "modules/graphics/opengl/Graphics.cpp"
 $75 = HEAP8[$74>>0]|0; //@line 155 "modules/graphics/opengl/Graphics.cpp"
 $76 = ($73<<24>>24)==($75<<24>>24); //@line 155 "modules/graphics/opengl/Graphics.cpp"
 $77 = ($73<<24>>24)==(0); //@line 155 "modules/graphics/opengl/Graphics.cpp"
 if ($76) {
  if (!($77)) {
   $78 = ((($s)) + 60|0); //@line 229 "./modules/graphics/Graphics.h"
   $79 = HEAP32[$78>>2]|0; //@line 229 "./modules/graphics/Graphics.h"
   $80 = ((($2)) + -76|0); //@line 229 "./modules/graphics/Graphics.h"
   $81 = HEAP32[$80>>2]|0; //@line 229 "./modules/graphics/Graphics.h"
   $82 = ($79|0)==($81|0); //@line 229 "./modules/graphics/Graphics.h"
   if ($82) {
    $83 = ((($s)) + 64|0); //@line 229 "./modules/graphics/Graphics.h"
    $84 = HEAP32[$83>>2]|0; //@line 229 "./modules/graphics/Graphics.h"
    $85 = ((($2)) + -72|0); //@line 229 "./modules/graphics/Graphics.h"
    $86 = HEAP32[$85>>2]|0; //@line 229 "./modules/graphics/Graphics.h"
    $87 = ($84|0)==($86|0); //@line 229 "./modules/graphics/Graphics.h"
    if ($87) {
     $88 = ((($s)) + 68|0); //@line 229 "./modules/graphics/Graphics.h"
     $89 = HEAP32[$88>>2]|0; //@line 229 "./modules/graphics/Graphics.h"
     $90 = ((($2)) + -68|0); //@line 229 "./modules/graphics/Graphics.h"
     $91 = HEAP32[$90>>2]|0; //@line 229 "./modules/graphics/Graphics.h"
     $92 = ($89|0)==($91|0); //@line 229 "./modules/graphics/Graphics.h"
     if ($92) {
      $93 = ((($s)) + 72|0); //@line 229 "./modules/graphics/Graphics.h"
      $94 = HEAP32[$93>>2]|0; //@line 229 "./modules/graphics/Graphics.h"
      $95 = ((($2)) + -64|0); //@line 229 "./modules/graphics/Graphics.h"
      $96 = HEAP32[$95>>2]|0; //@line 229 "./modules/graphics/Graphics.h"
      $97 = ($94|0)==($96|0); //@line 229 "./modules/graphics/Graphics.h"
      if (!($97)) {
       label = 18;
      }
     } else {
      label = 18;
     }
    } else {
     label = 18;
    }
   } else {
    label = 18;
   }
  }
 } else {
  if ($77) {
   $116 = ((($117)) + -80|0); //@line 658 "modules/graphics/opengl/Graphics.cpp"
   HEAP8[$116>>0] = 0; //@line 658 "modules/graphics/opengl/Graphics.cpp"
   $118 = HEAP32[165193]|0; //@line 56 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_vi[$118 & 8191](3089); //@line 56 "./libraries/glad/gladfuncs.hpp"
  } else {
   label = 18;
  }
 }
 if ((label|0) == 18) {
  $98 = ((($s)) + 60|0); //@line 158 "modules/graphics/opengl/Graphics.cpp"
  $99 = HEAP32[$98>>2]|0; //@line 158 "modules/graphics/opengl/Graphics.cpp"
  $100 = ((($s)) + 64|0); //@line 158 "modules/graphics/opengl/Graphics.cpp"
  $101 = HEAP32[$100>>2]|0; //@line 158 "modules/graphics/opengl/Graphics.cpp"
  $102 = ((($s)) + 68|0); //@line 158 "modules/graphics/opengl/Graphics.cpp"
  $103 = HEAP32[$102>>2]|0; //@line 158 "modules/graphics/opengl/Graphics.cpp"
  $104 = ((($s)) + 72|0); //@line 158 "modules/graphics/opengl/Graphics.cpp"
  $105 = HEAP32[$104>>2]|0; //@line 158 "modules/graphics/opengl/Graphics.cpp"
  $106 = HEAP32[164832]|0; //@line 57 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$106 & 8191](3089); //@line 57 "./libraries/glad/gladfuncs.hpp"
  HEAP32[$0>>2] = $99; //@line 629 "modules/graphics/opengl/Graphics.cpp"
  $107 = ((($0)) + 4|0); //@line 629 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$107>>2] = $101; //@line 629 "modules/graphics/opengl/Graphics.cpp"
  $108 = ((($0)) + 8|0); //@line 629 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$108>>2] = $103; //@line 629 "modules/graphics/opengl/Graphics.cpp"
  $109 = ((($0)) + 12|0); //@line 629 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$109>>2] = $105; //@line 629 "modules/graphics/opengl/Graphics.cpp"
  __ZN4love8graphics6opengl6OpenGL10setScissorERKNS2_8ViewportE(634044,$0); //@line 629 "modules/graphics/opengl/Graphics.cpp"
  $110 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $111 = ((($110)) + -80|0); //@line 631 "modules/graphics/opengl/Graphics.cpp"
  HEAP8[$111>>0] = 1; //@line 631 "modules/graphics/opengl/Graphics.cpp"
  $112 = ((($110)) + -76|0); //@line 632 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$112>>2] = $99; //@line 632 "modules/graphics/opengl/Graphics.cpp"
  $113 = ((($110)) + -72|0); //@line 632 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$113>>2] = $101; //@line 632 "modules/graphics/opengl/Graphics.cpp"
  $114 = ((($110)) + -68|0); //@line 632 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$114>>2] = $103; //@line 632 "modules/graphics/opengl/Graphics.cpp"
  $115 = ((($110)) + -64|0); //@line 632 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$115>>2] = $105; //@line 632 "modules/graphics/opengl/Graphics.cpp"
 }
 $119 = ((($s)) + 76|0); //@line 163 "modules/graphics/opengl/Graphics.cpp"
 $120 = HEAP32[$119>>2]|0; //@line 164 "modules/graphics/opengl/Graphics.cpp"
 $121 = ((($2)) + -60|0); //@line 163 "modules/graphics/opengl/Graphics.cpp"
 $122 = HEAP32[$121>>2]|0; //@line 163 "modules/graphics/opengl/Graphics.cpp"
 $123 = ($120|0)==($122|0); //@line 163 "modules/graphics/opengl/Graphics.cpp"
 $124 = ((($s)) + 80|0); //@line 163 "modules/graphics/opengl/Graphics.cpp"
 $125 = HEAP32[$124>>2]|0; //@line 163 "modules/graphics/opengl/Graphics.cpp"
 if ($123) {
  $126 = ((($2)) + -56|0); //@line 163 "modules/graphics/opengl/Graphics.cpp"
  $127 = HEAP32[$126>>2]|0; //@line 163 "modules/graphics/opengl/Graphics.cpp"
  $128 = ($125|0)==($127|0); //@line 163 "modules/graphics/opengl/Graphics.cpp"
  if (!($128)) {
   label = 22;
  }
 } else {
  label = 22;
 }
 do {
  if ((label|0) == 22) {
   $129 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $130 = ((($129)) + -60|0); //@line 735 "modules/graphics/opengl/Graphics.cpp"
   HEAP32[$130>>2] = $120; //@line 735 "modules/graphics/opengl/Graphics.cpp"
   $131 = ((($129)) + -56|0); //@line 736 "modules/graphics/opengl/Graphics.cpp"
   HEAP32[$131>>2] = $125; //@line 736 "modules/graphics/opengl/Graphics.cpp"
   $132 = ((($this)) + 42|0); //@line 738 "modules/graphics/opengl/Graphics.cpp"
   $133 = HEAP8[$132>>0]|0; //@line 738 "modules/graphics/opengl/Graphics.cpp"
   $134 = ($133<<24>>24)==(0); //@line 738 "modules/graphics/opengl/Graphics.cpp"
   if ($134) {
    $135 = ($120|0)==(6); //@line 741 "modules/graphics/opengl/Graphics.cpp"
    if ($135) {
     $136 = HEAP32[165193]|0; //@line 56 "./libraries/glad/gladfuncs.hpp"
     FUNCTION_TABLE_vi[$136 & 8191](2960); //@line 56 "./libraries/glad/gladfuncs.hpp"
     break;
    }
    $137 = HEAP32[158272]|0; //@line 748 "modules/graphics/opengl/Graphics.cpp"
    $138 = ($137|0)==(0|0); //@line 748 "modules/graphics/opengl/Graphics.cpp"
    if (!($138)) {
     (__ZN4love8graphics6opengl6Canvas18checkCreateStencilEv($137)|0); //@line 749 "modules/graphics/opengl/Graphics.cpp"
    }
    switch ($120|0) {
    case 0:  {
     $glcompare$0$i = 516;
     break;
    }
    case 1:  {
     $glcompare$0$i = 518;
     break;
    }
    case 5:  {
     $glcompare$0$i = 517;
     break;
    }
    case 3:  {
     $glcompare$0$i = 515;
     break;
    }
    case 4:  {
     $glcompare$0$i = 513;
     break;
    }
    default: {
     $glcompare$0$i = 514;
    }
    }
    $139 = HEAP32[164832]|0; //@line 57 "./libraries/glad/gladfuncs.hpp"
    FUNCTION_TABLE_vi[$139 & 8191](2960); //@line 57 "./libraries/glad/gladfuncs.hpp"
    $140 = HEAP32[165322]|0; //@line 62 "./libraries/glad/gladfuncs.hpp"
    FUNCTION_TABLE_viii[$140 & 8191]($glcompare$0$i,$125,-1); //@line 62 "./libraries/glad/gladfuncs.hpp"
    $141 = HEAP32[164775]|0; //@line 63 "./libraries/glad/gladfuncs.hpp"
    FUNCTION_TABLE_viii[$141 & 8191](7680,7680,7680); //@line 63 "./libraries/glad/gladfuncs.hpp"
   }
  }
 } while(0);
 $142 = ((($s)) + 84|0); //@line 139 "./common/Object.h"
 $143 = HEAP32[$142>>2]|0; //@line 139 "./common/Object.h"
 $144 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $145 = ($143|0)==(0|0); //@line 132 "./common/Object.h"
 if (!($145)) {
  $146 = HEAP32[$143>>2]|0; //@line 132 "./common/Object.h"
  $147 = ((($146)) + 8|0); //@line 132 "./common/Object.h"
  $148 = HEAP32[$147>>2]|0; //@line 132 "./common/Object.h"
  FUNCTION_TABLE_vi[$148 & 8191]($143); //@line 132 "./common/Object.h"
 }
 $149 = ((($144)) + -52|0); //@line 133 "./common/Object.h"
 $150 = HEAP32[$149>>2]|0; //@line 133 "./common/Object.h"
 $151 = ($150|0)==(0|0); //@line 133 "./common/Object.h"
 if (!($151)) {
  $152 = HEAP32[$150>>2]|0; //@line 133 "./common/Object.h"
  $153 = ((($152)) + 12|0); //@line 133 "./common/Object.h"
  $154 = HEAP32[$153>>2]|0; //@line 133 "./common/Object.h"
  FUNCTION_TABLE_vi[$154 & 8191]($150); //@line 133 "./common/Object.h"
 }
 HEAP32[$149>>2] = $143; //@line 134 "./common/Object.h"
 $155 = ((($s)) + 88|0); //@line 139 "./common/Object.h"
 $156 = HEAP32[$155>>2]|0; //@line 139 "./common/Object.h"
 $157 = ($156|0)==(0|0); //@line 989 "modules/graphics/opengl/Graphics.cpp"
 $158 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($157) {
  __ZN4love8graphics6opengl6Shader6detachEv(); //@line 1004 "modules/graphics/opengl/Graphics.cpp"
  $159 = ((($158)) + -48|0); //@line 133 "./common/Object.h"
  $160 = HEAP32[$159>>2]|0; //@line 133 "./common/Object.h"
  $161 = ($160|0)==(0|0); //@line 133 "./common/Object.h"
  if (!($161)) {
   $162 = HEAP32[$160>>2]|0; //@line 133 "./common/Object.h"
   $163 = ((($162)) + 12|0); //@line 133 "./common/Object.h"
   $164 = HEAP32[$163>>2]|0; //@line 133 "./common/Object.h"
   FUNCTION_TABLE_vi[$164 & 8191]($160); //@line 133 "./common/Object.h"
  }
  HEAP32[$159>>2] = 0; //@line 134 "./common/Object.h"
 } else {
  __ZN4love8graphics6opengl6Shader6attachEb($156,0); //@line 994 "modules/graphics/opengl/Graphics.cpp"
  $165 = HEAP32[$156>>2]|0; //@line 132 "./common/Object.h"
  $166 = ((($165)) + 8|0); //@line 132 "./common/Object.h"
  $167 = HEAP32[$166>>2]|0; //@line 132 "./common/Object.h"
  FUNCTION_TABLE_vi[$167 & 8191]($156); //@line 132 "./common/Object.h"
  $168 = ((($158)) + -48|0); //@line 133 "./common/Object.h"
  $169 = HEAP32[$168>>2]|0; //@line 133 "./common/Object.h"
  $170 = ($169|0)==(0|0); //@line 133 "./common/Object.h"
  if (!($170)) {
   $171 = HEAP32[$169>>2]|0; //@line 133 "./common/Object.h"
   $172 = ((($171)) + 12|0); //@line 133 "./common/Object.h"
   $173 = HEAP32[$172>>2]|0; //@line 133 "./common/Object.h"
   FUNCTION_TABLE_vi[$173 & 8191]($169); //@line 133 "./common/Object.h"
  }
  HEAP32[$168>>2] = $156; //@line 134 "./common/Object.h"
 }
 $174 = ((($s)) + 92|0); //@line 169 "modules/graphics/opengl/Graphics.cpp"
 $175 = ((($s)) + 96|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $176 = HEAP32[$175>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $177 = HEAP32[$174>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $178 = (($176) - ($177))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $179 = ((($2)) + -44|0); //@line 169 "modules/graphics/opengl/Graphics.cpp"
 $180 = ((($2)) + -40|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $181 = HEAP32[$180>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $182 = HEAP32[$179>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $183 = (($181) - ($182))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $184 = ($178|0)==($183|0); //@line 169 "modules/graphics/opengl/Graphics.cpp"
 $185 = ($176|0)==($177|0); //@line 171 "modules/graphics/opengl/Graphics.cpp"
 $186 = $177; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $187 = $182; //@line 171 "modules/graphics/opengl/Graphics.cpp"
 L62: do {
  if ($185) {
   label = 50;
  } else {
   $188 = $183 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $189 = (($176) - ($177))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $190 = $189 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $i$09 = 0;
   while(1) {
    $193 = ($i$09>>>0)<($188>>>0); //@line 171 "modules/graphics/opengl/Graphics.cpp"
    if (!($193)) {
     label = 50;
     break L62;
    }
    $194 = (($186) + ($i$09<<2)|0); //@line 139 "./common/Object.h"
    $195 = HEAP32[$194>>2]|0; //@line 139 "./common/Object.h"
    $196 = (($187) + ($i$09<<2)|0); //@line 139 "./common/Object.h"
    $197 = HEAP32[$196>>2]|0; //@line 139 "./common/Object.h"
    $198 = ($195|0)==($197|0); //@line 173 "modules/graphics/opengl/Graphics.cpp"
    $192 = (($i$09) + 1)|0; //@line 171 "modules/graphics/opengl/Graphics.cpp"
    if (!($198)) {
     label = 51;
     break L62;
    }
    $191 = ($192>>>0)<($190>>>0); //@line 171 "modules/graphics/opengl/Graphics.cpp"
    if ($191) {
     $i$09 = $192;
    } else {
     label = 50;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 50) {
  if (!($184)) {
   label = 51;
  }
 }
 if ((label|0) == 51) {
  __ZN4love8graphics6opengl8Graphics9setCanvasERKNSt3__16vectorINS_9StrongRefINS1_6CanvasEEENS3_9allocatorIS7_EEEE($this,$174); //@line 181 "modules/graphics/opengl/Graphics.cpp"
 }
 $199 = ((($s)) + 104|0); //@line 183 "modules/graphics/opengl/Graphics.cpp"
 $200 = HEAP8[$199>>0]|0; //@line 213 "./modules/graphics/Graphics.h"
 $201 = ((($2)) + -32|0); //@line 213 "./modules/graphics/Graphics.h"
 $202 = HEAP8[$201>>0]|0; //@line 213 "./modules/graphics/Graphics.h"
 $203 = ($200<<24>>24)==($202<<24>>24); //@line 213 "./modules/graphics/Graphics.h"
 if ($203) {
  $204 = ((($s)) + 105|0); //@line 213 "./modules/graphics/Graphics.h"
  $205 = HEAP8[$204>>0]|0; //@line 213 "./modules/graphics/Graphics.h"
  $206 = ((($2)) + -31|0); //@line 213 "./modules/graphics/Graphics.h"
  $207 = HEAP8[$206>>0]|0; //@line 213 "./modules/graphics/Graphics.h"
  $208 = ($205<<24>>24)==($207<<24>>24); //@line 213 "./modules/graphics/Graphics.h"
  if ($208) {
   $209 = ((($s)) + 106|0); //@line 213 "./modules/graphics/Graphics.h"
   $210 = HEAP8[$209>>0]|0; //@line 213 "./modules/graphics/Graphics.h"
   $211 = ((($2)) + -30|0); //@line 213 "./modules/graphics/Graphics.h"
   $212 = HEAP8[$211>>0]|0; //@line 213 "./modules/graphics/Graphics.h"
   $213 = ($210<<24>>24)==($212<<24>>24); //@line 213 "./modules/graphics/Graphics.h"
   if ($213) {
    $214 = ((($s)) + 107|0); //@line 213 "./modules/graphics/Graphics.h"
    $215 = HEAP8[$214>>0]|0; //@line 213 "./modules/graphics/Graphics.h"
    $216 = ((($2)) + -29|0); //@line 213 "./modules/graphics/Graphics.h"
    $217 = HEAP8[$216>>0]|0; //@line 213 "./modules/graphics/Graphics.h"
    $phitmp$i3 = ($215<<24>>24)==($217<<24>>24); //@line 218 "./modules/graphics/Graphics.h"
    if (!($phitmp$i3)) {
     label = 56;
    }
   } else {
    label = 56;
   }
  } else {
   label = 56;
  }
 } else {
  label = 56;
 }
 if ((label|0) == 56) {
  $218 = HEAPU8[$199>>0]|(HEAPU8[$199+1>>0]<<8)|(HEAPU8[$199+2>>0]<<16)|(HEAPU8[$199+3>>0]<<24); //@line 184 "modules/graphics/opengl/Graphics.cpp"
  $219 = $218&255;
  $220 = $218 >>> 8;
  $221 = $220&255;
  $222 = $218 >>> 16;
  $223 = $222&255;
  $224 = $218 >>> 24;
  $225 = $224&255;
  $226 = HEAP32[165466]|0; //@line 54 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiii[$226 & 8191]($219,$221,$223,$225); //@line 54 "./libraries/glad/gladfuncs.hpp"
  $227 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $228 = ((($227)) + -32|0); //@line 1085 "modules/graphics/opengl/Graphics.cpp"
  $229 = $224 << 24; //@line 1085 "modules/graphics/opengl/Graphics.cpp"
  $230 = $218 & 16777215; //@line 1085 "modules/graphics/opengl/Graphics.cpp"
  $231 = $230 | $229; //@line 1085 "modules/graphics/opengl/Graphics.cpp"
  HEAP8[$228>>0]=$231&255;HEAP8[$228+1>>0]=($231>>8)&255;HEAP8[$228+2>>0]=($231>>16)&255;HEAP8[$228+3>>0]=$231>>24; //@line 1085 "modules/graphics/opengl/Graphics.cpp"
 }
 $232 = ((($s)) + 108|0); //@line 186 "modules/graphics/opengl/Graphics.cpp"
 $233 = HEAP8[$232>>0]|0; //@line 186 "modules/graphics/opengl/Graphics.cpp"
 $234 = ((($2)) + -28|0); //@line 186 "modules/graphics/opengl/Graphics.cpp"
 $235 = HEAP8[$234>>0]|0; //@line 186 "modules/graphics/opengl/Graphics.cpp"
 $236 = ($233<<24>>24)!=($235<<24>>24); //@line 186 "modules/graphics/opengl/Graphics.cpp"
 $237 = HEAP8[1252012]|0; //@line 1244 "modules/graphics/opengl/Graphics.cpp"
 $238 = ($237<<24>>24)==(0); //@line 1244 "modules/graphics/opengl/Graphics.cpp"
 $or$cond = $236 & $238; //@line 186 "modules/graphics/opengl/Graphics.cpp"
 if (!($or$cond)) {
  $244 = ((($s)) + 112|0); //@line 189 "modules/graphics/opengl/Graphics.cpp"
  __ZN4love8graphics7Texture16setDefaultFilterERKNS1_6FilterE($244); //@line 1176 "modules/graphics/opengl/Graphics.cpp"
  $245 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $246 = ((($245)) + -24|0); //@line 1177 "modules/graphics/opengl/Graphics.cpp"
  ;HEAP32[$246>>2]=HEAP32[$244>>2]|0;HEAP32[$246+4>>2]=HEAP32[$244+4>>2]|0;HEAP32[$246+8>>2]=HEAP32[$244+8>>2]|0;HEAP32[$246+12>>2]=HEAP32[$244+12>>2]|0; //@line 1177 "modules/graphics/opengl/Graphics.cpp"
  $247 = ((($s)) + 128|0); //@line 190 "modules/graphics/opengl/Graphics.cpp"
  $248 = HEAP32[$247>>2]|0; //@line 190 "modules/graphics/opengl/Graphics.cpp"
  $249 = ((($s)) + 132|0); //@line 190 "modules/graphics/opengl/Graphics.cpp"
  $250 = +HEAPF32[$249>>2]; //@line 190 "modules/graphics/opengl/Graphics.cpp"
  __ZN4love8graphics6opengl5Image22setDefaultMipmapFilterENS0_7Texture10FilterModeE($248); //@line 1187 "modules/graphics/opengl/Graphics.cpp"
  __ZN4love8graphics6opengl5Image25setDefaultMipmapSharpnessEf($250); //@line 1188 "modules/graphics/opengl/Graphics.cpp"
  $251 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $252 = ((($251)) + -8|0); //@line 1190 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$252>>2] = $248; //@line 1190 "modules/graphics/opengl/Graphics.cpp"
  $253 = ((($251)) + -4|0); //@line 1191 "modules/graphics/opengl/Graphics.cpp"
  HEAPF32[$253>>2] = $250; //@line 1191 "modules/graphics/opengl/Graphics.cpp"
  STACKTOP = sp;return; //@line 191 "modules/graphics/opengl/Graphics.cpp"
 }
 $239 = ($233<<24>>24)!=(0); //@line 186 "modules/graphics/opengl/Graphics.cpp"
 $240 = $239 ? 6913 : 6914; //@line 1247 "modules/graphics/opengl/Graphics.cpp"
 $241 = HEAP32[165045]|0; //@line 40 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$241 & 8191](1032,$240); //@line 40 "./libraries/glad/gladfuncs.hpp"
 $242 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $243 = ((($242)) + -28|0); //@line 1248 "modules/graphics/opengl/Graphics.cpp"
 HEAP8[$243>>0] = $233; //@line 1248 "modules/graphics/opengl/Graphics.cpp"
 $244 = ((($s)) + 112|0); //@line 189 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics7Texture16setDefaultFilterERKNS1_6FilterE($244); //@line 1176 "modules/graphics/opengl/Graphics.cpp"
 $245 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $246 = ((($245)) + -24|0); //@line 1177 "modules/graphics/opengl/Graphics.cpp"
 ;HEAP32[$246>>2]=HEAP32[$244>>2]|0;HEAP32[$246+4>>2]=HEAP32[$244+4>>2]|0;HEAP32[$246+8>>2]=HEAP32[$244+8>>2]|0;HEAP32[$246+12>>2]=HEAP32[$244+12>>2]|0; //@line 1177 "modules/graphics/opengl/Graphics.cpp"
 $247 = ((($s)) + 128|0); //@line 190 "modules/graphics/opengl/Graphics.cpp"
 $248 = HEAP32[$247>>2]|0; //@line 190 "modules/graphics/opengl/Graphics.cpp"
 $249 = ((($s)) + 132|0); //@line 190 "modules/graphics/opengl/Graphics.cpp"
 $250 = +HEAPF32[$249>>2]; //@line 190 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl5Image22setDefaultMipmapFilterENS0_7Texture10FilterModeE($248); //@line 1187 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl5Image25setDefaultMipmapSharpnessEf($250); //@line 1188 "modules/graphics/opengl/Graphics.cpp"
 $251 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $252 = ((($251)) + -8|0); //@line 1190 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$252>>2] = $248; //@line 1190 "modules/graphics/opengl/Graphics.cpp"
 $253 = ((($251)) + -4|0); //@line 1191 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$253>>2] = $250; //@line 1191 "modules/graphics/opengl/Graphics.cpp"
 STACKTOP = sp;return; //@line 191 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics19checkSetDefaultFontEv($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$pr = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -52|0); //@line 139 "./common/Object.h"
 $3 = HEAP32[$2>>2]|0; //@line 139 "./common/Object.h"
 $4 = ($3|0)==(0|0); //@line 196 "modules/graphics/opengl/Graphics.cpp"
 if (!($4)) {
  STACKTOP = sp;return; //@line 214 "modules/graphics/opengl/Graphics.cpp"
 }
 $5 = ((($this)) + 12|0); //@line 139 "./common/Object.h"
 $6 = HEAP32[$5>>2]|0; //@line 139 "./common/Object.h"
 $7 = ($6|0)==(0|0); //@line 200 "modules/graphics/opengl/Graphics.cpp"
 L4: do {
  if ($7) {
   $9 = HEAP32[(625620)>>2]|0; //@line 98 "./common/Module.h"
   $10 = ($9|0)==(0|0); //@line 203 "modules/graphics/opengl/Graphics.cpp"
   if ($10) {
    $11 = (___cxa_allocate_exception(16)|0); //@line 204 "modules/graphics/opengl/Graphics.cpp"
    __THREW__ = 0;
    invoke_viii(3349,($11|0),(435134|0),($vararg_buffer|0)); //@line 204 "modules/graphics/opengl/Graphics.cpp"
    $12 = __THREW__; __THREW__ = 0;
    $13 = $12&1;
    if (!($13)) {
     ___cxa_throw(($11|0),(96|0),(344|0)); //@line 204 "modules/graphics/opengl/Graphics.cpp"
     // unreachable; //@line 204 "modules/graphics/opengl/Graphics.cpp"
    }
    $14 = ___cxa_find_matching_catch_2()|0;
    $15 = tempRet0;
    ___cxa_free_exception(($11|0)); //@line 204 "modules/graphics/opengl/Graphics.cpp"
    $$0 = $15;$$01 = $14;
    ___resumeException($$01|0);
    // unreachable;
   }
   $16 = HEAP32[$9>>2]|0; //@line 206 "modules/graphics/opengl/Graphics.cpp"
   $17 = ((($16)) + 28|0); //@line 206 "modules/graphics/opengl/Graphics.cpp"
   $18 = HEAP32[$17>>2]|0; //@line 206 "modules/graphics/opengl/Graphics.cpp"
   $19 = (FUNCTION_TABLE_iiii[$18 & 8191]($9,12,0)|0); //@line 206 "modules/graphics/opengl/Graphics.cpp"
   $20 = ($19|0)==(0|0); //@line 95 "./common/Object.h"
   if (!($20)) {
    $21 = HEAP32[$19>>2]|0; //@line 95 "./common/Object.h"
    $22 = ((($21)) + 8|0); //@line 95 "./common/Object.h"
    $23 = HEAP32[$22>>2]|0; //@line 95 "./common/Object.h"
    FUNCTION_TABLE_vi[$23 & 8191]($19); //@line 95 "./common/Object.h"
   }
   $24 = HEAP32[$19>>2]|0; //@line 207 "modules/graphics/opengl/Graphics.cpp"
   $25 = ((($24)) + 12|0); //@line 207 "modules/graphics/opengl/Graphics.cpp"
   $26 = HEAP32[$25>>2]|0; //@line 207 "modules/graphics/opengl/Graphics.cpp"
   __THREW__ = 0;
   invoke_vi($26|0,($19|0)); //@line 207 "modules/graphics/opengl/Graphics.cpp"
   $27 = __THREW__; __THREW__ = 0;
   $28 = $27&1;
   do {
    if ($28) {
     label = 22;
    } else {
     __THREW__ = 0;
     $29 = (invoke_i(4056)|0); //@line 209 "modules/graphics/opengl/Graphics.cpp"
     $30 = __THREW__; __THREW__ = 0;
     $31 = $30&1;
     if ($31) {
      label = 22;
     } else {
      __THREW__ = 0;
      $32 = (invoke_ii(3348,140)|0); //@line 830 "modules/graphics/opengl/Graphics.cpp"
      $33 = __THREW__; __THREW__ = 0;
      $34 = $33&1;
      if ($34) {
       label = 22;
      } else {
       __THREW__ = 0;
       invoke_viii(4057,($32|0),($19|0),($29|0)); //@line 830 "modules/graphics/opengl/Graphics.cpp"
       $35 = __THREW__; __THREW__ = 0;
       $36 = $35&1;
       if ($36) {
        $37 = ___cxa_find_matching_catch_2()|0;
        $38 = tempRet0;
        __ZdlPv($32); //@line 830 "modules/graphics/opengl/Graphics.cpp"
        $eh$lpad$body$index2Z2D = $38;$eh$lpad$body$indexZ2D = $37;
        break;
       }
       $39 = HEAP32[$32>>2]|0; //@line 132 "./common/Object.h"
       $40 = ((($39)) + 8|0); //@line 132 "./common/Object.h"
       $41 = HEAP32[$40>>2]|0; //@line 132 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($41|0,($32|0)); //@line 132 "./common/Object.h"
       $42 = __THREW__; __THREW__ = 0;
       $43 = $42&1;
       if ($43) {
        label = 22;
       } else {
        $44 = HEAP32[$5>>2]|0; //@line 133 "./common/Object.h"
        $45 = ($44|0)==(0|0); //@line 133 "./common/Object.h"
        if (!($45)) {
         $46 = HEAP32[$44>>2]|0; //@line 133 "./common/Object.h"
         $47 = ((($46)) + 12|0); //@line 133 "./common/Object.h"
         $48 = HEAP32[$47>>2]|0; //@line 133 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($48|0,($44|0)); //@line 133 "./common/Object.h"
         $49 = __THREW__; __THREW__ = 0;
         $50 = $49&1;
         if ($50) {
          label = 22;
          break;
         }
        }
        HEAP32[$5>>2] = $32; //@line 134 "./common/Object.h"
        $51 = HEAP32[$32>>2]|0; //@line 210 "modules/graphics/opengl/Graphics.cpp"
        $52 = ((($51)) + 12|0); //@line 210 "modules/graphics/opengl/Graphics.cpp"
        $53 = HEAP32[$52>>2]|0; //@line 210 "modules/graphics/opengl/Graphics.cpp"
        __THREW__ = 0;
        invoke_vi($53|0,($32|0)); //@line 210 "modules/graphics/opengl/Graphics.cpp"
        $54 = __THREW__; __THREW__ = 0;
        $55 = $54&1;
        if ($55) {
         label = 22;
        } else {
         if (!($20)) {
          $56 = HEAP32[$19>>2]|0; //@line 106 "./common/Object.h"
          $57 = ((($56)) + 12|0); //@line 106 "./common/Object.h"
          $58 = HEAP32[$57>>2]|0; //@line 106 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($58|0,($19|0)); //@line 106 "./common/Object.h"
          $59 = __THREW__; __THREW__ = 0;
          $60 = $59&1;
          if ($60) {
           $61 = ___cxa_find_matching_catch_3(0|0)|0;
           $62 = tempRet0;
           ___clang_call_terminate($61); //@line 106 "./common/Object.h"
           // unreachable; //@line 106 "./common/Object.h"
          }
         }
         $$pr = HEAP32[$5>>2]|0; //@line 139 "./common/Object.h"
         $72 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $73 = ((($72)) + -52|0); //@line 213 "modules/graphics/opengl/Graphics.cpp"
         $74 = ($$pr|0)==(0|0); //@line 132 "./common/Object.h"
         if ($74) {
          $80 = $73;$85 = 0;
          break L4;
         } else {
          $76 = $$pr;$86 = $73;
          label = 27;
          break L4;
         }
        }
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 22) {
    $63 = ___cxa_find_matching_catch_2()|0;
    $64 = tempRet0;
    $eh$lpad$body$index2Z2D = $64;$eh$lpad$body$indexZ2D = $63;
   }
   if ($20) {
    $$0 = $eh$lpad$body$index2Z2D;$$01 = $eh$lpad$body$indexZ2D;
    ___resumeException($$01|0);
    // unreachable;
   }
   $65 = HEAP32[$19>>2]|0; //@line 106 "./common/Object.h"
   $66 = ((($65)) + 12|0); //@line 106 "./common/Object.h"
   $67 = HEAP32[$66>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($67|0,($19|0)); //@line 106 "./common/Object.h"
   $68 = __THREW__; __THREW__ = 0;
   $69 = $68&1;
   if ($69) {
    $70 = ___cxa_find_matching_catch_3(0|0)|0;
    $71 = tempRet0;
    ___clang_call_terminate($70); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   } else {
    $$0 = $eh$lpad$body$index2Z2D;$$01 = $eh$lpad$body$indexZ2D;
    ___resumeException($$01|0);
    // unreachable;
   }
  } else {
   $8 = ((($1)) + -52|0); //@line 213 "modules/graphics/opengl/Graphics.cpp"
   $76 = $6;$86 = $8;
   label = 27;
  }
 } while(0);
 if ((label|0) == 27) {
  $75 = HEAP32[$76>>2]|0; //@line 132 "./common/Object.h"
  $77 = ((($75)) + 8|0); //@line 132 "./common/Object.h"
  $78 = HEAP32[$77>>2]|0; //@line 132 "./common/Object.h"
  FUNCTION_TABLE_vi[$78 & 8191]($76); //@line 132 "./common/Object.h"
  $80 = $86;$85 = $76;
 }
 $79 = HEAP32[$80>>2]|0; //@line 133 "./common/Object.h"
 $81 = ($79|0)==(0|0); //@line 133 "./common/Object.h"
 if (!($81)) {
  $82 = HEAP32[$79>>2]|0; //@line 133 "./common/Object.h"
  $83 = ((($82)) + 12|0); //@line 133 "./common/Object.h"
  $84 = HEAP32[$83>>2]|0; //@line 133 "./common/Object.h"
  FUNCTION_TABLE_vi[$84 & 8191]($79); //@line 133 "./common/Object.h"
 }
 HEAP32[$80>>2] = $85; //@line 134 "./common/Object.h"
 STACKTOP = sp;return; //@line 214 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics7newFontEPNS_4font10RasterizerERKNS0_7Texture6FilterE($this,$r,$filter) {
 $this = $this|0;
 $r = $r|0;
 $filter = $filter|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(140)|0); //@line 830 "modules/graphics/opengl/Graphics.cpp"
 __THREW__ = 0;
 invoke_viii(4057,($0|0),($r|0),($filter|0)); //@line 830 "modules/graphics/opengl/Graphics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 830 "modules/graphics/opengl/Graphics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 830 "modules/graphics/opengl/Graphics.cpp"
 }
 return (0)|0;
}
function __ZNK4love8graphics6opengl8Graphics9isCreatedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0); //@line 620 "modules/graphics/opengl/Graphics.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 620 "modules/graphics/opengl/Graphics.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 620 "modules/graphics/opengl/Graphics.cpp"
 return ($2|0); //@line 620 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics9setCanvasEv($this) {
 $this = $this|0;
 var $$pre$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[158272]|0; //@line 1065 "modules/graphics/opengl/Graphics.cpp"
 $3 = ($2|0)==(0|0); //@line 1065 "modules/graphics/opengl/Graphics.cpp"
 if (!($3)) {
  __ZN4love8graphics6opengl6Canvas8stopGrabEb($2,0); //@line 1066 "modules/graphics/opengl/Graphics.cpp"
 }
 $4 = ((($1)) + -44|0); //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$4>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ((($1)) + -40|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = HEAP32[$6>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ($7|0)==($5|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($8) {
  return; //@line 1069 "modules/graphics/opengl/Graphics.cpp"
 } else {
  $10 = $7;
 }
 while(1) {
  $9 = ((($10)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$6>>2] = $9; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $11 = HEAP32[$9>>2]|0; //@line 106 "./common/Object.h"
  $12 = ($11|0)==(0|0); //@line 106 "./common/Object.h"
  if ($12) {
   $14 = $9;
  } else {
   $15 = HEAP32[$11>>2]|0; //@line 106 "./common/Object.h"
   $16 = ((($15)) + 12|0); //@line 106 "./common/Object.h"
   $17 = HEAP32[$16>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($17|0,($11|0)); //@line 106 "./common/Object.h"
   $18 = __THREW__; __THREW__ = 0;
   $19 = $18&1;
   if ($19) {
    label = 8;
    break;
   }
   $$pre$i$i$i = HEAP32[$6>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $14 = $$pre$i$i$i;
  }
  $13 = ($14|0)==($5|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($13) {
   label = 9;
   break;
  } else {
   $10 = $14;
  }
 }
 if ((label|0) == 8) {
  $20 = ___cxa_find_matching_catch_3(0|0)|0;
  $21 = tempRet0;
  ___clang_call_terminate($20); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 }
 else if ((label|0) == 9) {
  return; //@line 1069 "modules/graphics/opengl/Graphics.cpp"
 }
}
function __ZNK4love8graphics6opengl8Graphics14isGammaCorrectEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love8graphics14isGammaCorrectEv()|0); //@line 944 "modules/graphics/opengl/Graphics.cpp"
 return ($0|0); //@line 944 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics9newShaderERKNS1_6Shader12ShaderSourceE($this,$source) {
 $this = $this|0;
 $source = $source|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(328)|0); //@line 909 "modules/graphics/opengl/Graphics.cpp"
 __THREW__ = 0;
 invoke_vii(4050,($0|0),($source|0)); //@line 909 "modules/graphics/opengl/Graphics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 909 "modules/graphics/opengl/Graphics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 909 "modules/graphics/opengl/Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl8Graphics5resetEv($this) {
 $this = $this|0;
 var $$pre$i$i = 0, $$pre$i$i$i$i = 0, $$pre$i$i$i$i4 = 0, $$pre$i$i7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $s = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $s = sp;
 HEAPF32[$s>>2] = 255.0; //@line 38 "./modules/graphics/Color.h"
 $0 = ((($s)) + 4|0); //@line 38 "./modules/graphics/Color.h"
 HEAPF32[$0>>2] = 255.0; //@line 38 "./modules/graphics/Color.h"
 $1 = ((($s)) + 8|0); //@line 38 "./modules/graphics/Color.h"
 HEAPF32[$1>>2] = 255.0; //@line 38 "./modules/graphics/Color.h"
 $2 = ((($s)) + 12|0); //@line 38 "./modules/graphics/Color.h"
 HEAPF32[$2>>2] = 255.0; //@line 38 "./modules/graphics/Color.h"
 $3 = ((($s)) + 16|0); //@line 38 "./modules/graphics/Color.h"
 HEAPF32[$3>>2] = 0.0; //@line 38 "./modules/graphics/Color.h"
 $4 = ((($s)) + 20|0); //@line 38 "./modules/graphics/Color.h"
 HEAPF32[$4>>2] = 0.0; //@line 38 "./modules/graphics/Color.h"
 $5 = ((($s)) + 24|0); //@line 38 "./modules/graphics/Color.h"
 HEAPF32[$5>>2] = 0.0; //@line 38 "./modules/graphics/Color.h"
 $6 = ((($s)) + 28|0); //@line 38 "./modules/graphics/Color.h"
 HEAPF32[$6>>2] = 255.0; //@line 38 "./modules/graphics/Color.h"
 $7 = ((($s)) + 32|0); //@line 487 "modules/graphics/opengl/Graphics.h"
 HEAP32[$7>>2] = 0; //@line 487 "modules/graphics/opengl/Graphics.h"
 $8 = ((($s)) + 36|0); //@line 488 "modules/graphics/opengl/Graphics.h"
 HEAP32[$8>>2] = 0; //@line 488 "modules/graphics/opengl/Graphics.h"
 $9 = ((($s)) + 40|0); //@line 490 "modules/graphics/opengl/Graphics.h"
 HEAPF32[$9>>2] = 1.0; //@line 490 "modules/graphics/opengl/Graphics.h"
 $10 = ((($s)) + 44|0); //@line 491 "modules/graphics/opengl/Graphics.h"
 HEAP32[$10>>2] = 1; //@line 491 "modules/graphics/opengl/Graphics.h"
 $11 = ((($s)) + 48|0); //@line 492 "modules/graphics/opengl/Graphics.h"
 HEAP32[$11>>2] = 1; //@line 492 "modules/graphics/opengl/Graphics.h"
 $12 = ((($s)) + 52|0); //@line 494 "modules/graphics/opengl/Graphics.h"
 HEAPF32[$12>>2] = 1.0; //@line 494 "modules/graphics/opengl/Graphics.h"
 $13 = ((($s)) + 56|0); //@line 496 "modules/graphics/opengl/Graphics.h"
 HEAP8[$13>>0] = 0; //@line 496 "modules/graphics/opengl/Graphics.h"
 $14 = ((($s)) + 60|0); //@line 497 "modules/graphics/opengl/Graphics.h"
 ;HEAP32[$14>>2]=0|0;HEAP32[$14+4>>2]=0|0;HEAP32[$14+8>>2]=0|0;HEAP32[$14+12>>2]=0|0; //@line 497 "modules/graphics/opengl/Graphics.h"
 $15 = ((($s)) + 76|0); //@line 500 "modules/graphics/opengl/Graphics.h"
 HEAP32[$15>>2] = 6; //@line 500 "modules/graphics/opengl/Graphics.h"
 $16 = ((($s)) + 80|0); //@line 501 "modules/graphics/opengl/Graphics.h"
 $17 = ((($s)) + 104|0); //@line 208 "./modules/graphics/Graphics.h"
 ;HEAP32[$16>>2]=0|0;HEAP32[$16+4>>2]=0|0;HEAP32[$16+8>>2]=0|0;HEAP32[$16+12>>2]=0|0;HEAP32[$16+16>>2]=0|0;HEAP32[$16+20>>2]=0|0; //@line 88 "./common/Object.h"
 $18 = ((($s)) + 108|0); //@line 510 "modules/graphics/opengl/Graphics.h"
 HEAP32[$17>>2] = 16843009; //@line 208 "./modules/graphics/Graphics.h"
 HEAP8[$18>>0] = 0; //@line 510 "modules/graphics/opengl/Graphics.h"
 $19 = ((($s)) + 112|0); //@line 512 "modules/graphics/opengl/Graphics.h"
 $20 = $19; //@line 512 "modules/graphics/opengl/Graphics.h"
 $21 = $20; //@line 512 "modules/graphics/opengl/Graphics.h"
 HEAP32[$21>>2] = 0; //@line 512 "modules/graphics/opengl/Graphics.h"
 $22 = (($20) + 4)|0; //@line 512 "modules/graphics/opengl/Graphics.h"
 $23 = $22; //@line 512 "modules/graphics/opengl/Graphics.h"
 HEAP32[$23>>2] = 0; //@line 512 "modules/graphics/opengl/Graphics.h"
 HEAP32[$19>>2] = 1; //@line 62 "./modules/graphics/Texture.h"
 $24 = ((($s)) + 116|0); //@line 63 "./modules/graphics/Texture.h"
 HEAP32[$24>>2] = 1; //@line 63 "./modules/graphics/Texture.h"
 $25 = ((($s)) + 120|0); //@line 64 "./modules/graphics/Texture.h"
 HEAP32[$25>>2] = 0; //@line 64 "./modules/graphics/Texture.h"
 $26 = ((($s)) + 124|0); //@line 65 "./modules/graphics/Texture.h"
 HEAPF32[$26>>2] = 1.0; //@line 65 "./modules/graphics/Texture.h"
 $27 = ((($s)) + 128|0); //@line 514 "modules/graphics/opengl/Graphics.h"
 HEAP32[$27>>2] = 2; //@line 514 "modules/graphics/opengl/Graphics.h"
 $28 = ((($s)) + 132|0); //@line 515 "modules/graphics/opengl/Graphics.h"
 HEAPF32[$28>>2] = 0.0; //@line 515 "modules/graphics/opengl/Graphics.h"
 __THREW__ = 0;
 invoke_vi(4058,($this|0)); //@line 436 "modules/graphics/opengl/Graphics.cpp"
 $29 = __THREW__; __THREW__ = 0;
 $30 = $29&1;
 if (!($30)) {
  __THREW__ = 0;
  invoke_vii(4059,($this|0),($s|0)); //@line 437 "modules/graphics/opengl/Graphics.cpp"
  $31 = __THREW__; __THREW__ = 0;
  $32 = $31&1;
  if (!($32)) {
   __THREW__ = 0;
   $33 = (invoke_ii(4003,(634044|0))|0); //@line 1751 "modules/graphics/opengl/Graphics.cpp"
   $34 = __THREW__; __THREW__ = 0;
   $35 = $34&1;
   if (!($35)) {
    __THREW__ = 0;
    invoke_vi(4060,($33|0)); //@line 1751 "modules/graphics/opengl/Graphics.cpp"
    $36 = __THREW__; __THREW__ = 0;
    $37 = $36&1;
    if (!($37)) {
     $38 = ((($this)) + 20|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $39 = HEAP32[$38>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $40 = ((($39)) + -8|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAPF64[$40>>3] = 1.0; //@line 1752 "modules/graphics/opengl/Graphics.cpp"
     $41 = ((($s)) + 92|0); //@line 482 "modules/graphics/opengl/Graphics.h"
     $42 = HEAP32[$41>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $43 = ($42|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($43)) {
      $44 = ((($s)) + 96|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $45 = HEAP32[$44>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $46 = ($45|0)==($42|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      do {
       if ($46) {
        $60 = $42;
       } else {
        $48 = $45;
        while(1) {
         $47 = ((($48)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$44>>2] = $47; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $49 = HEAP32[$47>>2]|0; //@line 106 "./common/Object.h"
         $50 = ($49|0)==(0|0); //@line 106 "./common/Object.h"
         if ($50) {
          $52 = $47;
         } else {
          $53 = HEAP32[$49>>2]|0; //@line 106 "./common/Object.h"
          $54 = ((($53)) + 12|0); //@line 106 "./common/Object.h"
          $55 = HEAP32[$54>>2]|0; //@line 106 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($55|0,($49|0)); //@line 106 "./common/Object.h"
          $56 = __THREW__; __THREW__ = 0;
          $57 = $56&1;
          if ($57) {
           label = 11;
           break;
          }
          $$pre$i$i$i$i4 = HEAP32[$44>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $52 = $$pre$i$i$i$i4;
         }
         $51 = ($52|0)==($42|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($51) {
          label = 12;
          break;
         } else {
          $48 = $52;
         }
        }
        if ((label|0) == 11) {
         $58 = ___cxa_find_matching_catch_3(0|0)|0;
         $59 = tempRet0;
         ___clang_call_terminate($58); //@line 106 "./common/Object.h"
         // unreachable; //@line 106 "./common/Object.h"
        }
        else if ((label|0) == 12) {
         $$pre$i$i7 = HEAP32[$41>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $60 = $$pre$i$i7;
         break;
        }
       }
      } while(0);
      __ZdlPv($60); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $61 = ((($s)) + 88|0); //@line 106 "./common/Object.h"
     $62 = HEAP32[$61>>2]|0; //@line 106 "./common/Object.h"
     $63 = ($62|0)==(0|0); //@line 106 "./common/Object.h"
     if (!($63)) {
      $64 = HEAP32[$62>>2]|0; //@line 106 "./common/Object.h"
      $65 = ((($64)) + 12|0); //@line 106 "./common/Object.h"
      $66 = HEAP32[$65>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($66|0,($62|0)); //@line 106 "./common/Object.h"
      $67 = __THREW__; __THREW__ = 0;
      $68 = $67&1;
      if ($68) {
       $69 = ___cxa_find_matching_catch_3(0|0)|0;
       $70 = tempRet0;
       ___clang_call_terminate($69); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
     }
     $71 = ((($s)) + 84|0); //@line 106 "./common/Object.h"
     $72 = HEAP32[$71>>2]|0; //@line 106 "./common/Object.h"
     $73 = ($72|0)==(0|0); //@line 106 "./common/Object.h"
     if ($73) {
      STACKTOP = sp;return; //@line 439 "modules/graphics/opengl/Graphics.cpp"
     }
     $74 = HEAP32[$72>>2]|0; //@line 106 "./common/Object.h"
     $75 = ((($74)) + 12|0); //@line 106 "./common/Object.h"
     $76 = HEAP32[$75>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($76|0,($72|0)); //@line 106 "./common/Object.h"
     $77 = __THREW__; __THREW__ = 0;
     $78 = $77&1;
     if ($78) {
      $79 = ___cxa_find_matching_catch_3(0|0)|0;
      $80 = tempRet0;
      ___clang_call_terminate($79); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     } else {
      STACKTOP = sp;return; //@line 439 "modules/graphics/opengl/Graphics.cpp"
     }
    }
   }
  }
 }
 $81 = ___cxa_find_matching_catch_2()|0;
 $82 = tempRet0;
 $83 = ((($s)) + 92|0); //@line 482 "modules/graphics/opengl/Graphics.h"
 $84 = HEAP32[$83>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $85 = ($84|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($85)) {
  $86 = ((($s)) + 96|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $87 = HEAP32[$86>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $88 = ($87|0)==($84|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  do {
   if ($88) {
    $102 = $84;
   } else {
    $90 = $87;
    while(1) {
     $89 = ((($90)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$86>>2] = $89; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $91 = HEAP32[$89>>2]|0; //@line 106 "./common/Object.h"
     $92 = ($91|0)==(0|0); //@line 106 "./common/Object.h"
     if ($92) {
      $94 = $89;
     } else {
      $95 = HEAP32[$91>>2]|0; //@line 106 "./common/Object.h"
      $96 = ((($95)) + 12|0); //@line 106 "./common/Object.h"
      $97 = HEAP32[$96>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($97|0,($91|0)); //@line 106 "./common/Object.h"
      $98 = __THREW__; __THREW__ = 0;
      $99 = $98&1;
      if ($99) {
       label = 27;
       break;
      }
      $$pre$i$i$i$i = HEAP32[$86>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $94 = $$pre$i$i$i$i;
     }
     $93 = ($94|0)==($84|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($93) {
      label = 28;
      break;
     } else {
      $90 = $94;
     }
    }
    if ((label|0) == 27) {
     $100 = ___cxa_find_matching_catch_3(0|0)|0;
     $101 = tempRet0;
     ___clang_call_terminate($100); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
    else if ((label|0) == 28) {
     $$pre$i$i = HEAP32[$83>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $102 = $$pre$i$i;
     break;
    }
   }
  } while(0);
  __ZdlPv($102); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $103 = ((($s)) + 88|0); //@line 106 "./common/Object.h"
 $104 = HEAP32[$103>>2]|0; //@line 106 "./common/Object.h"
 $105 = ($104|0)==(0|0); //@line 106 "./common/Object.h"
 if (!($105)) {
  $106 = HEAP32[$104>>2]|0; //@line 106 "./common/Object.h"
  $107 = ((($106)) + 12|0); //@line 106 "./common/Object.h"
  $108 = HEAP32[$107>>2]|0; //@line 106 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($108|0,($104|0)); //@line 106 "./common/Object.h"
  $109 = __THREW__; __THREW__ = 0;
  $110 = $109&1;
  if ($110) {
   $111 = ___cxa_find_matching_catch_3(0|0)|0;
   $112 = tempRet0;
   ___clang_call_terminate($111); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 }
 $113 = ((($s)) + 84|0); //@line 106 "./common/Object.h"
 $114 = HEAP32[$113>>2]|0; //@line 106 "./common/Object.h"
 $115 = ($114|0)==(0|0); //@line 106 "./common/Object.h"
 if ($115) {
  ___resumeException($81|0);
  // unreachable;
 }
 $116 = HEAP32[$114>>2]|0; //@line 106 "./common/Object.h"
 $117 = ((($116)) + 12|0); //@line 106 "./common/Object.h"
 $118 = HEAP32[$117>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($118|0,($114|0)); //@line 106 "./common/Object.h"
 $119 = __THREW__; __THREW__ = 0;
 $120 = $119&1;
 if ($120) {
  $121 = ___cxa_find_matching_catch_3(0|0)|0;
  $122 = tempRet0;
  ___clang_call_terminate($121); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  ___resumeException($81|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl8Graphics23stopDrawToStencilBufferEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $glcompare$0$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 42|0); //@line 718 "modules/graphics/opengl/Graphics.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 718 "modules/graphics/opengl/Graphics.cpp"
 $2 = ($1<<24>>24)==(0); //@line 718 "modules/graphics/opengl/Graphics.cpp"
 if ($2) {
  return; //@line 730 "modules/graphics/opengl/Graphics.cpp"
 }
 HEAP8[$0>>0] = 0; //@line 721 "modules/graphics/opengl/Graphics.cpp"
 $3 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$3>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ((($4)) + -32|0); //@line 726 "modules/graphics/opengl/Graphics.cpp"
 $6 = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8)|(HEAPU8[$5+2>>0]<<16)|(HEAPU8[$5+3>>0]<<24); //@line 726 "modules/graphics/opengl/Graphics.cpp"
 $7 = $6&255;
 $8 = $6 >>> 8;
 $9 = $8&255;
 $10 = $6 >>> 16;
 $11 = $10&255;
 $12 = $6 >>> 24;
 $13 = $12&255;
 $14 = HEAP32[165466]|0; //@line 54 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiii[$14 & 8191]($7,$9,$11,$13); //@line 54 "./libraries/glad/gladfuncs.hpp"
 $15 = HEAP32[$3>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = ((($15)) + -32|0); //@line 1085 "modules/graphics/opengl/Graphics.cpp"
 $17 = $12 << 24; //@line 1085 "modules/graphics/opengl/Graphics.cpp"
 $18 = $6 & 16777215; //@line 1085 "modules/graphics/opengl/Graphics.cpp"
 $19 = $18 | $17; //@line 1085 "modules/graphics/opengl/Graphics.cpp"
 HEAP8[$16>>0]=$19&255;HEAP8[$16+1>>0]=($19>>8)&255;HEAP8[$16+2>>0]=($19>>16)&255;HEAP8[$16+3>>0]=$19>>24; //@line 1085 "modules/graphics/opengl/Graphics.cpp"
 $20 = ((($4)) + -60|0); //@line 729 "modules/graphics/opengl/Graphics.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 729 "modules/graphics/opengl/Graphics.cpp"
 $22 = ((($4)) + -56|0); //@line 729 "modules/graphics/opengl/Graphics.cpp"
 $23 = HEAP32[$22>>2]|0; //@line 729 "modules/graphics/opengl/Graphics.cpp"
 $24 = HEAP32[$3>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $25 = ((($24)) + -60|0); //@line 735 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$25>>2] = $21; //@line 735 "modules/graphics/opengl/Graphics.cpp"
 $26 = ((($24)) + -56|0); //@line 736 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$26>>2] = $23; //@line 736 "modules/graphics/opengl/Graphics.cpp"
 $27 = HEAP8[$0>>0]|0; //@line 738 "modules/graphics/opengl/Graphics.cpp"
 $28 = ($27<<24>>24)==(0); //@line 738 "modules/graphics/opengl/Graphics.cpp"
 if (!($28)) {
  return; //@line 730 "modules/graphics/opengl/Graphics.cpp"
 }
 $29 = ($21|0)==(6); //@line 741 "modules/graphics/opengl/Graphics.cpp"
 if ($29) {
  $30 = HEAP32[165193]|0; //@line 56 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$30 & 8191](2960); //@line 56 "./libraries/glad/gladfuncs.hpp"
  return; //@line 730 "modules/graphics/opengl/Graphics.cpp"
 }
 $31 = HEAP32[158272]|0; //@line 748 "modules/graphics/opengl/Graphics.cpp"
 $32 = ($31|0)==(0|0); //@line 748 "modules/graphics/opengl/Graphics.cpp"
 if (!($32)) {
  (__ZN4love8graphics6opengl6Canvas18checkCreateStencilEv($31)|0); //@line 749 "modules/graphics/opengl/Graphics.cpp"
 }
 switch ($21|0) {
 case 0:  {
  $glcompare$0$i = 516;
  break;
 }
 case 1:  {
  $glcompare$0$i = 518;
  break;
 }
 case 5:  {
  $glcompare$0$i = 517;
  break;
 }
 case 3:  {
  $glcompare$0$i = 515;
  break;
 }
 case 4:  {
  $glcompare$0$i = 513;
  break;
 }
 default: {
  $glcompare$0$i = 514;
 }
 }
 $33 = HEAP32[164832]|0; //@line 57 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$33 & 8191](2960); //@line 57 "./libraries/glad/gladfuncs.hpp"
 $34 = HEAP32[165322]|0; //@line 62 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$34 & 8191]($glcompare$0$i,$23,-1); //@line 62 "./libraries/glad/gladfuncs.hpp"
 $35 = HEAP32[164775]|0; //@line 63 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$35 & 8191](7680,7680,7680); //@line 63 "./libraries/glad/gladfuncs.hpp"
 return; //@line 730 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics6originEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love8graphics6opengl6OpenGL12getTransformEv(634044)|0); //@line 1751 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love7Matrix411setIdentityEv($0); //@line 1751 "modules/graphics/opengl/Graphics.cpp"
 $1 = ((($this)) + 20|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($2)) + -8|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAPF64[$3>>3] = 1.0; //@line 1752 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1753 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics5clearENS0_6ColorTIfEE($this,$c) {
 $this = $this|0;
 $c = $c|0;
 var $0 = 0.0, $1 = 0.0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0, $9 = 0.0, $nc = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $nc = sp;
 $0 = +HEAPF32[$c>>2]; //@line 443 "modules/graphics/opengl/Graphics.cpp"
 $1 = $0 / 255.0; //@line 443 "modules/graphics/opengl/Graphics.cpp"
 $2 = ((($c)) + 4|0); //@line 443 "modules/graphics/opengl/Graphics.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 443 "modules/graphics/opengl/Graphics.cpp"
 $4 = $3 / 255.0; //@line 443 "modules/graphics/opengl/Graphics.cpp"
 $5 = ((($c)) + 8|0); //@line 443 "modules/graphics/opengl/Graphics.cpp"
 $6 = +HEAPF32[$5>>2]; //@line 443 "modules/graphics/opengl/Graphics.cpp"
 $7 = $6 / 255.0; //@line 443 "modules/graphics/opengl/Graphics.cpp"
 $8 = ((($c)) + 12|0); //@line 443 "modules/graphics/opengl/Graphics.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 443 "modules/graphics/opengl/Graphics.cpp"
 $10 = $9 / 255.0; //@line 443 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$nc>>2] = $1; //@line 38 "./modules/graphics/Color.h"
 $11 = ((($nc)) + 4|0); //@line 38 "./modules/graphics/Color.h"
 HEAPF32[$11>>2] = $4; //@line 38 "./modules/graphics/Color.h"
 $12 = ((($nc)) + 8|0); //@line 38 "./modules/graphics/Color.h"
 HEAPF32[$12>>2] = $7; //@line 38 "./modules/graphics/Color.h"
 $13 = ((($nc)) + 12|0); //@line 38 "./modules/graphics/Color.h"
 HEAPF32[$13>>2] = $10; //@line 38 "./modules/graphics/Color.h"
 __ZN4love8graphics17gammaCorrectColorERNS0_6ColorTIfEE($nc); //@line 445 "modules/graphics/opengl/Graphics.cpp"
 $14 = +HEAPF32[$nc>>2]; //@line 447 "modules/graphics/opengl/Graphics.cpp"
 $15 = +HEAPF32[$11>>2]; //@line 447 "modules/graphics/opengl/Graphics.cpp"
 $16 = +HEAPF32[$12>>2]; //@line 447 "modules/graphics/opengl/Graphics.cpp"
 $17 = +HEAPF32[$13>>2]; //@line 447 "modules/graphics/opengl/Graphics.cpp"
 $18 = HEAP32[165395]|0; //@line 50 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vdddd[$18 & 8191]($14,$15,$16,$17); //@line 50 "./libraries/glad/gladfuncs.hpp"
 $19 = HEAP32[164768]|0; //@line 49 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$19 & 8191](17664); //@line 49 "./libraries/glad/gladfuncs.hpp"
 $20 = HEAP8[(634080)>>0]|0; //@line 450 "modules/graphics/opengl/Graphics.cpp"
 $21 = ($20<<24>>24)!=(0); //@line 450 "modules/graphics/opengl/Graphics.cpp"
 $22 = HEAP32[158581]|0; //@line 450 "modules/graphics/opengl/Graphics.cpp"
 $23 = ($22|0)!=(0|0); //@line 450 "modules/graphics/opengl/Graphics.cpp"
 $or$cond = $21 & $23; //@line 450 "modules/graphics/opengl/Graphics.cpp"
 if (!($or$cond)) {
  STACKTOP = sp;return; //@line 457 "modules/graphics/opengl/Graphics.cpp"
 }
 $24 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$24 & 8191](0); //@line 534 "./libraries/glad/gladfuncs.hpp"
 $25 = HEAP32[158581]|0; //@line 455 "modules/graphics/opengl/Graphics.cpp"
 $26 = ((($25)) + 48|0); //@line 199 "modules/graphics/opengl/Shader.h"
 $27 = HEAP32[$26>>2]|0; //@line 199 "modules/graphics/opengl/Shader.h"
 $28 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$28 & 8191]($27); //@line 534 "./libraries/glad/gladfuncs.hpp"
 STACKTOP = sp;return; //@line 457 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics5clearERKNSt3__16vectorINS2_14OptionalColorfENS3_9allocatorIS5_EEEE($this,$colors) {
 $this = $this|0;
 $colors = $colors|0;
 var $$0 = 0, $$04 = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $$pre26 = 0, $$pre27 = 0, $$pre28 = 0, $$pre30 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0, $181 = 0, $182 = 0, $183 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0.0, $73 = 0.0, $74 = 0, $75 = 0.0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0;
 var $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $bufs = 0, $c = 0, $drawbuffermodified$0$off019 = 0, $drawbuffermodified$1$off0 = 0, $drawbuffermodified$2$off0 = 0, $i$020 = 0, $i2$017 = 0;
 var $lpad$phi$index = 0, $lpad$phi$index2 = 0, $nc$i = 0, $or$cond$i = 0, $or$cond3 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i8 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $nc$i = sp + 40|0;
 $c = sp + 24|0;
 $bufs = sp + 8|0;
 $0 = sp + 4|0;
 $1 = ((($colors)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$colors>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($2) - ($3))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($2|0)==($3|0); //@line 461 "modules/graphics/opengl/Graphics.cpp"
 $7 = $3; //@line 461 "modules/graphics/opengl/Graphics.cpp"
 if ($6) {
  STACKTOP = sp;return; //@line 529 "modules/graphics/opengl/Graphics.cpp"
 }
 $8 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = HEAP32[$8>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = ((($9)) + -44|0); //@line 464 "modules/graphics/opengl/Graphics.cpp"
 $11 = ((($9)) + -40|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = HEAP32[$11>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = HEAP32[$10>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = (($12) - ($13))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = $14 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = ($15|0)==(0); //@line 464 "modules/graphics/opengl/Graphics.cpp"
 if ($16) {
  $17 = ((($7)) + 16|0); //@line 466 "modules/graphics/opengl/Graphics.cpp"
  $18 = HEAP8[$17>>0]|0; //@line 466 "modules/graphics/opengl/Graphics.cpp"
  $19 = ($18<<24>>24)==(0); //@line 466 "modules/graphics/opengl/Graphics.cpp"
  if ($19) {
   STACKTOP = sp;return; //@line 529 "modules/graphics/opengl/Graphics.cpp"
  }
  $20 = +HEAPF32[$7>>2]; //@line 467 "modules/graphics/opengl/Graphics.cpp"
  $21 = ((($7)) + 4|0); //@line 467 "modules/graphics/opengl/Graphics.cpp"
  $22 = +HEAPF32[$21>>2]; //@line 467 "modules/graphics/opengl/Graphics.cpp"
  $23 = ((($7)) + 8|0); //@line 467 "modules/graphics/opengl/Graphics.cpp"
  $24 = +HEAPF32[$23>>2]; //@line 467 "modules/graphics/opengl/Graphics.cpp"
  $25 = ((($7)) + 12|0); //@line 467 "modules/graphics/opengl/Graphics.cpp"
  $26 = +HEAPF32[$25>>2]; //@line 467 "modules/graphics/opengl/Graphics.cpp"
  $27 = $20 / 255.0; //@line 443 "modules/graphics/opengl/Graphics.cpp"
  $28 = $22 / 255.0; //@line 443 "modules/graphics/opengl/Graphics.cpp"
  $29 = $24 / 255.0; //@line 443 "modules/graphics/opengl/Graphics.cpp"
  $30 = $26 / 255.0; //@line 443 "modules/graphics/opengl/Graphics.cpp"
  HEAPF32[$nc$i>>2] = $27; //@line 38 "./modules/graphics/Color.h"
  $31 = ((($nc$i)) + 4|0); //@line 38 "./modules/graphics/Color.h"
  HEAPF32[$31>>2] = $28; //@line 38 "./modules/graphics/Color.h"
  $32 = ((($nc$i)) + 8|0); //@line 38 "./modules/graphics/Color.h"
  HEAPF32[$32>>2] = $29; //@line 38 "./modules/graphics/Color.h"
  $33 = ((($nc$i)) + 12|0); //@line 38 "./modules/graphics/Color.h"
  HEAPF32[$33>>2] = $30; //@line 38 "./modules/graphics/Color.h"
  __ZN4love8graphics17gammaCorrectColorERNS0_6ColorTIfEE($nc$i); //@line 445 "modules/graphics/opengl/Graphics.cpp"
  $34 = +HEAPF32[$nc$i>>2]; //@line 447 "modules/graphics/opengl/Graphics.cpp"
  $35 = +HEAPF32[$31>>2]; //@line 447 "modules/graphics/opengl/Graphics.cpp"
  $36 = +HEAPF32[$32>>2]; //@line 447 "modules/graphics/opengl/Graphics.cpp"
  $37 = +HEAPF32[$33>>2]; //@line 447 "modules/graphics/opengl/Graphics.cpp"
  $38 = HEAP32[165395]|0; //@line 50 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vdddd[$38 & 8191]($34,$35,$36,$37); //@line 50 "./libraries/glad/gladfuncs.hpp"
  $39 = HEAP32[164768]|0; //@line 49 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$39 & 8191](17664); //@line 49 "./libraries/glad/gladfuncs.hpp"
  $40 = HEAP8[(634080)>>0]|0; //@line 450 "modules/graphics/opengl/Graphics.cpp"
  $41 = ($40<<24>>24)!=(0); //@line 450 "modules/graphics/opengl/Graphics.cpp"
  $42 = HEAP32[158581]|0; //@line 450 "modules/graphics/opengl/Graphics.cpp"
  $43 = ($42|0)!=(0|0); //@line 450 "modules/graphics/opengl/Graphics.cpp"
  $or$cond$i = $41 & $43; //@line 450 "modules/graphics/opengl/Graphics.cpp"
  if ($or$cond$i) {
   $44 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_vi[$44 & 8191](0); //@line 534 "./libraries/glad/gladfuncs.hpp"
   $45 = HEAP32[158581]|0; //@line 455 "modules/graphics/opengl/Graphics.cpp"
   $46 = ((($45)) + 48|0); //@line 199 "modules/graphics/opengl/Shader.h"
   $47 = HEAP32[$46>>2]|0; //@line 199 "modules/graphics/opengl/Shader.h"
   $48 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_vi[$48 & 8191]($47); //@line 534 "./libraries/glad/gladfuncs.hpp"
  }
  STACKTOP = sp;return; //@line 529 "modules/graphics/opengl/Graphics.cpp"
 }
 $49 = ($5|0)==($15|0); //@line 472 "modules/graphics/opengl/Graphics.cpp"
 if (!($49)) {
  $54 = (___cxa_allocate_exception(16)|0); //@line 473 "modules/graphics/opengl/Graphics.cpp"
  $55 = HEAP32[$8>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $56 = ((($55)) + -44|0); //@line 473 "modules/graphics/opengl/Graphics.cpp"
  $57 = ((($55)) + -40|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $58 = HEAP32[$57>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $59 = HEAP32[$56>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $60 = (($58) - ($59))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $61 = $60 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  __THREW__ = 0;
  HEAP32[$vararg_buffer>>2] = $61; //@line 473 "modules/graphics/opengl/Graphics.cpp"
  invoke_viii(3349,($54|0),(435167|0),($vararg_buffer|0)); //@line 473 "modules/graphics/opengl/Graphics.cpp"
  $62 = __THREW__; __THREW__ = 0;
  $63 = $62&1;
  if (!($63)) {
   ___cxa_throw(($54|0),(96|0),(344|0)); //@line 473 "modules/graphics/opengl/Graphics.cpp"
   // unreachable; //@line 473 "modules/graphics/opengl/Graphics.cpp"
  }
  $64 = ___cxa_find_matching_catch_2()|0;
  $65 = tempRet0;
  ___cxa_free_exception(($54|0)); //@line 473 "modules/graphics/opengl/Graphics.cpp"
  $$0 = $65;$$04 = $64;
  ___resumeException($$04|0);
  // unreachable;
 }
 $50 = ($4|0)>(0); //@line 477 "modules/graphics/opengl/Graphics.cpp"
 L21: do {
  if ($50) {
   $51 = ((($c)) + 4|0); //@line 482 "modules/graphics/opengl/Graphics.cpp"
   $52 = ((($c)) + 8|0); //@line 482 "modules/graphics/opengl/Graphics.cpp"
   $53 = ((($c)) + 12|0); //@line 482 "modules/graphics/opengl/Graphics.cpp"
   $181 = $3;$182 = $2;$68 = $7;$drawbuffermodified$0$off019 = 0;$i$020 = 0;
   while(1) {
    $67 = (((($68) + (($i$020*20)|0)|0)) + 16|0); //@line 479 "modules/graphics/opengl/Graphics.cpp"
    $69 = HEAP8[$67>>0]|0; //@line 479 "modules/graphics/opengl/Graphics.cpp"
    $70 = ($69<<24>>24)==(0); //@line 479 "modules/graphics/opengl/Graphics.cpp"
    if ($70) {
     $105 = $182;$106 = $181;$drawbuffermodified$2$off0 = $drawbuffermodified$0$off019;
    } else {
     $71 = (($68) + (($i$020*20)|0)|0); //@line 482 "modules/graphics/opengl/Graphics.cpp"
     $72 = +HEAPF32[$71>>2]; //@line 482 "modules/graphics/opengl/Graphics.cpp"
     $73 = $72 / 255.0; //@line 482 "modules/graphics/opengl/Graphics.cpp"
     HEAPF32[$c>>2] = $73; //@line 482 "modules/graphics/opengl/Graphics.cpp"
     $74 = (((($68) + (($i$020*20)|0)|0)) + 4|0); //@line 482 "modules/graphics/opengl/Graphics.cpp"
     $75 = +HEAPF32[$74>>2]; //@line 482 "modules/graphics/opengl/Graphics.cpp"
     $76 = $75 / 255.0; //@line 482 "modules/graphics/opengl/Graphics.cpp"
     HEAPF32[$51>>2] = $76; //@line 482 "modules/graphics/opengl/Graphics.cpp"
     $77 = (((($68) + (($i$020*20)|0)|0)) + 8|0); //@line 482 "modules/graphics/opengl/Graphics.cpp"
     $78 = +HEAPF32[$77>>2]; //@line 482 "modules/graphics/opengl/Graphics.cpp"
     $79 = $78 / 255.0; //@line 482 "modules/graphics/opengl/Graphics.cpp"
     HEAPF32[$52>>2] = $79; //@line 482 "modules/graphics/opengl/Graphics.cpp"
     $80 = (((($68) + (($i$020*20)|0)|0)) + 12|0); //@line 482 "modules/graphics/opengl/Graphics.cpp"
     $81 = +HEAPF32[$80>>2]; //@line 482 "modules/graphics/opengl/Graphics.cpp"
     $82 = $81 / 255.0; //@line 482 "modules/graphics/opengl/Graphics.cpp"
     HEAPF32[$53>>2] = $82; //@line 482 "modules/graphics/opengl/Graphics.cpp"
     $83 = (__ZN4love8graphics14isGammaCorrectEv()|0); //@line 944 "modules/graphics/opengl/Graphics.cpp"
     if ($83) {
      $84 = +HEAPF32[$c>>2]; //@line 490 "modules/graphics/opengl/Graphics.cpp"
      $85 = (+__ZNK4love4math4Math13gammaToLinearEf(623984,$84)); //@line 490 "modules/graphics/opengl/Graphics.cpp"
      HEAPF32[$c>>2] = $85; //@line 490 "modules/graphics/opengl/Graphics.cpp"
      $86 = +HEAPF32[$51>>2]; //@line 490 "modules/graphics/opengl/Graphics.cpp"
      $87 = (+__ZNK4love4math4Math13gammaToLinearEf(623984,$86)); //@line 490 "modules/graphics/opengl/Graphics.cpp"
      HEAPF32[$51>>2] = $87; //@line 490 "modules/graphics/opengl/Graphics.cpp"
      $88 = +HEAPF32[$52>>2]; //@line 490 "modules/graphics/opengl/Graphics.cpp"
      $89 = (+__ZNK4love4math4Math13gammaToLinearEf(623984,$88)); //@line 490 "modules/graphics/opengl/Graphics.cpp"
      HEAPF32[$52>>2] = $89; //@line 490 "modules/graphics/opengl/Graphics.cpp"
     }
     $90 = HEAP8[1252013]|0; //@line 493 "modules/graphics/opengl/Graphics.cpp"
     $91 = HEAP8[1252002]|0; //@line 493 "modules/graphics/opengl/Graphics.cpp"
     $92 = $91 | $90; //@line 493 "modules/graphics/opengl/Graphics.cpp"
     $93 = ($92<<24>>24)==(0); //@line 493 "modules/graphics/opengl/Graphics.cpp"
     if ($93) {
      $95 = (($i$020) + 36064)|0; //@line 497 "modules/graphics/opengl/Graphics.cpp"
      $96 = HEAP32[164868]|0; //@line 48 "./libraries/glad/gladfuncs.hpp"
      FUNCTION_TABLE_vi[$96 & 8191]($95); //@line 48 "./libraries/glad/gladfuncs.hpp"
      $97 = +HEAPF32[$c>>2]; //@line 498 "modules/graphics/opengl/Graphics.cpp"
      $98 = +HEAPF32[$51>>2]; //@line 498 "modules/graphics/opengl/Graphics.cpp"
      $99 = +HEAPF32[$52>>2]; //@line 498 "modules/graphics/opengl/Graphics.cpp"
      $100 = +HEAPF32[$53>>2]; //@line 498 "modules/graphics/opengl/Graphics.cpp"
      $101 = HEAP32[165395]|0; //@line 50 "./libraries/glad/gladfuncs.hpp"
      FUNCTION_TABLE_vdddd[$101 & 8191]($97,$98,$99,$100); //@line 50 "./libraries/glad/gladfuncs.hpp"
      $102 = HEAP32[164768]|0; //@line 49 "./libraries/glad/gladfuncs.hpp"
      FUNCTION_TABLE_vi[$102 & 8191](16384); //@line 49 "./libraries/glad/gladfuncs.hpp"
      $drawbuffermodified$1$off0 = 1;
     } else {
      $94 = HEAP32[164707]|0; //@line 657 "./libraries/glad/gladfuncs.hpp"
      FUNCTION_TABLE_viii[$94 & 8191](6144,$i$020,$c); //@line 657 "./libraries/glad/gladfuncs.hpp"
      $drawbuffermodified$1$off0 = $drawbuffermodified$0$off019;
     }
     $$pre = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $$pre24 = HEAP32[$colors>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $105 = $$pre;$106 = $$pre24;$drawbuffermodified$2$off0 = $drawbuffermodified$1$off0;
    }
    $103 = (($i$020) + 1)|0; //@line 477 "modules/graphics/opengl/Graphics.cpp"
    $104 = (($105) - ($106))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $107 = (($104|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $108 = ($103|0)<($107|0); //@line 477 "modules/graphics/opengl/Graphics.cpp"
    $109 = $106; //@line 477 "modules/graphics/opengl/Graphics.cpp"
    if ($108) {
     $181 = $106;$182 = $105;$68 = $109;$drawbuffermodified$0$off019 = $drawbuffermodified$2$off0;$i$020 = $103;
    } else {
     break;
    }
   }
   $66 = HEAP32[164768]|0; //@line 49 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_vi[$66 & 8191](1280); //@line 49 "./libraries/glad/gladfuncs.hpp"
   if ($drawbuffermodified$2$off0) {
    HEAP32[$bufs>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $110 = ((($bufs)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$110>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $111 = ((($bufs)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$111>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $112 = HEAP32[$8>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $113 = ((($112)) + -44|0); //@line 513 "modules/graphics/opengl/Graphics.cpp"
    $114 = ((($112)) + -40|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $115 = HEAP32[$114>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $116 = HEAP32[$113>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $117 = (($115) - ($116))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $118 = ($117|0)>(0); //@line 513 "modules/graphics/opengl/Graphics.cpp"
    do {
     if ($118) {
      $119 = ((($bufs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $141 = 0;$142 = 0;$183 = $112;$i2$017 = 0;
      while(1) {
       $139 = (($i2$017) + 36064)|0; //@line 514 "modules/graphics/opengl/Graphics.cpp"
       HEAP32[$0>>2] = $139; //@line 514 "modules/graphics/opengl/Graphics.cpp"
       $140 = ($141>>>0)<($142>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($140) {
        HEAP32[$141>>2] = $139; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $143 = ((($141)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$110>>2] = $143; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $148 = $183;
       } else {
        __THREW__ = 0;
        invoke_vii(3597,($bufs|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $144 = __THREW__; __THREW__ = 0;
        $145 = $144&1;
        if ($145) {
         label = 26;
         break;
        }
        $$pre27 = HEAP32[$8>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $148 = $$pre27;
       }
       $146 = (($i2$017) + 1)|0; //@line 513 "modules/graphics/opengl/Graphics.cpp"
       $147 = ((($148)) + -44|0); //@line 513 "modules/graphics/opengl/Graphics.cpp"
       $149 = ((($148)) + -40|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $150 = HEAP32[$149>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $151 = HEAP32[$147>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $152 = (($150) - ($151))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $153 = $152 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $154 = ($146|0)<($153|0); //@line 513 "modules/graphics/opengl/Graphics.cpp"
       if (!($154)) {
        label = 24;
        break;
       }
       $$pre25 = HEAP32[$110>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $$pre26 = HEAP32[$119>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $141 = $$pre25;$142 = $$pre26;$183 = $148;$i2$017 = $146;
      }
      if ((label|0) == 24) {
       $$pre28 = HEAP32[$110>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $$pre30 = HEAP32[$bufs>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $120 = $$pre30; //@line 511 "modules/graphics/opengl/Graphics.cpp"
       $122 = $$pre28;$123 = $$pre30;$156 = $120;
       label = 25;
       break;
      }
      else if ((label|0) == 26) {
       $126 = ___cxa_find_matching_catch_2()|0;
       $127 = tempRet0;
       $lpad$phi$index = $126;$lpad$phi$index2 = $127;
       break;
      }
     } else {
      $122 = 0;$123 = 0;$156 = 0;
      label = 25;
     }
    } while(0);
    do {
     if ((label|0) == 25) {
      $121 = (($122) - ($123))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $124 = $121 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $125 = ($124>>>0)>(1); //@line 516 "modules/graphics/opengl/Graphics.cpp"
      if ($125) {
       $155 = HEAP32[165046]|0; //@line 500 "./libraries/glad/gladfuncs.hpp"
       __THREW__ = 0;
       invoke_vii($155|0,($124|0),($156|0)); //@line 500 "./libraries/glad/gladfuncs.hpp"
       $157 = __THREW__; __THREW__ = 0;
       $158 = $157&1;
       if ($158) {
        label = 27;
       }
      } else {
       $159 = HEAP32[164868]|0; //@line 48 "./libraries/glad/gladfuncs.hpp"
       __THREW__ = 0;
       invoke_vi($159|0,36064); //@line 48 "./libraries/glad/gladfuncs.hpp"
       $160 = __THREW__; __THREW__ = 0;
       $161 = $160&1;
       if ($161) {
        label = 27;
       }
      }
      if ((label|0) == 27) {
       $128 = ___cxa_find_matching_catch_2()|0;
       $129 = tempRet0;
       $lpad$phi$index = $128;$lpad$phi$index2 = $129;
       break;
      }
      $162 = HEAP32[$bufs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $163 = ($162|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $164 = $162; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($163)) {
       $165 = HEAP32[$110>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $166 = ($165|0)==($162|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($166)) {
        $scevgep$i$i$i = ((($165)) + -4|0);
        $167 = $scevgep$i$i$i;
        $168 = (($167) - ($164))|0;
        $169 = $168 >>> 2;
        $170 = $169 ^ -1;
        $scevgep4$i$i$i = (($165) + ($170<<2)|0);
        HEAP32[$110>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       __ZdlPv($162); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      }
      break L21;
     }
    } while(0);
    $130 = HEAP32[$bufs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $131 = ($130|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $132 = $130; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($131) {
     $$0 = $lpad$phi$index2;$$04 = $lpad$phi$index;
     ___resumeException($$04|0);
     // unreachable;
    }
    $133 = HEAP32[$110>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $134 = ($133|0)==($130|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($134)) {
     $scevgep$i$i$i8 = ((($133)) + -4|0);
     $135 = $scevgep$i$i$i8;
     $136 = (($135) - ($132))|0;
     $137 = $136 >>> 2;
     $138 = $137 ^ -1;
     $scevgep4$i$i$i9 = (($133) + ($138<<2)|0);
     HEAP32[$110>>2] = $scevgep4$i$i$i9; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($130); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $$0 = $lpad$phi$index2;$$04 = $lpad$phi$index;
    ___resumeException($$04|0);
    // unreachable;
   }
  } else {
   $171 = HEAP32[164768]|0; //@line 49 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_vi[$171 & 8191](1280); //@line 49 "./libraries/glad/gladfuncs.hpp"
  }
 } while(0);
 $172 = HEAP8[(634080)>>0]|0; //@line 522 "modules/graphics/opengl/Graphics.cpp"
 $173 = ($172<<24>>24)!=(0); //@line 522 "modules/graphics/opengl/Graphics.cpp"
 $174 = HEAP32[158581]|0; //@line 522 "modules/graphics/opengl/Graphics.cpp"
 $175 = ($174|0)!=(0|0); //@line 522 "modules/graphics/opengl/Graphics.cpp"
 $or$cond3 = $173 & $175; //@line 522 "modules/graphics/opengl/Graphics.cpp"
 if (!($or$cond3)) {
  STACKTOP = sp;return; //@line 529 "modules/graphics/opengl/Graphics.cpp"
 }
 $176 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$176 & 8191](0); //@line 534 "./libraries/glad/gladfuncs.hpp"
 $177 = HEAP32[158581]|0; //@line 527 "modules/graphics/opengl/Graphics.cpp"
 $178 = ((($177)) + 48|0); //@line 199 "modules/graphics/opengl/Shader.h"
 $179 = HEAP32[$178>>2]|0; //@line 199 "modules/graphics/opengl/Shader.h"
 $180 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$180 & 8191]($179); //@line 534 "./libraries/glad/gladfuncs.hpp"
 STACKTOP = sp;return; //@line 529 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics7discardERKNSt3__16vectorIbNS3_9allocatorIbEEEEb($this,$colorbuffers,$stencil) {
 $this = $this|0;
 $colorbuffers = $colorbuffers|0;
 $stencil = $stencil|0;
 var $$cast = 0, $$cast29 = 0, $$pr = 0, $$pre = 0, $$pre31 = 0, $$pre32 = 0, $$pre33 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $attachments = 0, $i$030 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $or$cond = 0, $rendertargetcount$0 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i14 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i15 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $attachments = sp + 24|0;
 $0 = sp + 20|0;
 $1 = sp + 16|0;
 $2 = sp + 12|0;
 $3 = sp + 8|0;
 $4 = sp + 4|0;
 $5 = sp;
 $6 = HEAP8[1252009]|0; //@line 533 "modules/graphics/opengl/Graphics.cpp"
 $7 = HEAP8[1252079]|0; //@line 533 "modules/graphics/opengl/Graphics.cpp"
 $8 = $7 | $6; //@line 533 "modules/graphics/opengl/Graphics.cpp"
 $9 = HEAP8[1252013]|0; //@line 533 "modules/graphics/opengl/Graphics.cpp"
 $10 = $8 | $9; //@line 533 "modules/graphics/opengl/Graphics.cpp"
 $11 = HEAP8[1252174]|0; //@line 533 "modules/graphics/opengl/Graphics.cpp"
 $12 = $10 | $11; //@line 533 "modules/graphics/opengl/Graphics.cpp"
 $13 = ($12<<24>>24)==(0); //@line 533 "modules/graphics/opengl/Graphics.cpp"
 if ($13) {
  STACKTOP = sp;return; //@line 575 "modules/graphics/opengl/Graphics.cpp"
 }
 HEAP32[$attachments>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = ((($attachments)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$14>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = ((($attachments)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$15>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $16 = ((($colorbuffers)) + 4|0); //@line 2255 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = HEAP32[$16>>2]|0; //@line 2255 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $18 = ((($attachments)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $19 = ($17|0)==(0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($19) {
  label = 5;
 } else {
  $20 = $17 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  __THREW__ = 0;
  $21 = (invoke_ii(3348,($20|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $22 = __THREW__; __THREW__ = 0;
  $23 = $22&1;
  if ($23) {
   label = 14;
  } else {
   $24 = $21; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $25 = (($21) + ($17<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $26 = $25; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   HEAP32[$attachments>>2] = $24; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$14>>2] = $24; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$18>>2] = $26; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   label = 5;
  }
 }
 L7: do {
  if ((label|0) == 5) {
   $27 = HEAP32[158272]|0; //@line 540 "modules/graphics/opengl/Graphics.cpp"
   $28 = ($27|0)==(0|0); //@line 540 "modules/graphics/opengl/Graphics.cpp"
   do {
    if ($28) {
     __THREW__ = 0;
     $29 = (invoke_ii(4061,(634044|0))|0); //@line 540 "modules/graphics/opengl/Graphics.cpp"
     $30 = __THREW__; __THREW__ = 0;
     $31 = $30&1;
     if ($31) {
      label = 14;
      break L7;
     }
     $32 = ($29|0)==(0); //@line 540 "modules/graphics/opengl/Graphics.cpp"
     if (!($32)) {
      $$pr = HEAP32[158272]|0; //@line 554 "modules/graphics/opengl/Graphics.cpp"
      $70 = ($$pr|0)==(0|0); //@line 554 "modules/graphics/opengl/Graphics.cpp"
      if ($70) {
       $rendertargetcount$0 = 1;
       label = 30;
       break;
      } else {
       label = 29;
       break;
      }
     }
     $33 = HEAP32[$16>>2]|0; //@line 2255 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $34 = ($33|0)==(0); //@line 542 "modules/graphics/opengl/Graphics.cpp"
     do {
      if (!($34)) {
       $35 = HEAP32[$colorbuffers>>2]|0; //@line 2387 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $36 = HEAP32[$35>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
       $37 = $36 & 1; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
       $38 = ($37|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
       if (!($38)) {
        HEAP32[$0>>2] = 6144; //@line 543 "modules/graphics/opengl/Graphics.cpp"
        $39 = HEAP32[$14>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $40 = HEAP32[$18>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $41 = ($39>>>0)<($40>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($41) {
         HEAP32[$39>>2] = 6144; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $42 = ((($39)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$14>>2] = $42; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         break;
        } else {
         __THREW__ = 0;
         invoke_vii(3597,($attachments|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $43 = __THREW__; __THREW__ = 0;
         $44 = $43&1;
         if ($44) {
          label = 14;
          break L7;
         } else {
          break;
         }
        }
       }
      }
     } while(0);
     if ($stencil) {
      HEAP32[$1>>2] = 6146; //@line 547 "modules/graphics/opengl/Graphics.cpp"
      $58 = HEAP32[$14>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $59 = HEAP32[$18>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $60 = ($58>>>0)<($59>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($60) {
       HEAP32[$58>>2] = 6146; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $61 = ((($58)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$14>>2] = $61; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $65 = $61;$66 = $59;
      } else {
       __THREW__ = 0;
       invoke_vii(3597,($attachments|0),($1|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $62 = __THREW__; __THREW__ = 0;
       $63 = $62&1;
       if ($63) {
        label = 14;
        break L7;
       }
       $$pre32 = HEAP32[$14>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $$pre33 = HEAP32[$18>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $65 = $$pre32;$66 = $$pre33;
      }
      HEAP32[$2>>2] = 6145; //@line 548 "modules/graphics/opengl/Graphics.cpp"
      $64 = ($65>>>0)<($66>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($64) {
       HEAP32[$65>>2] = 6145; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $67 = ((($65)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$14>>2] = $67; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       break;
      } else {
       __THREW__ = 0;
       invoke_vii(3597,($attachments|0),($2|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $68 = __THREW__; __THREW__ = 0;
       $69 = $68&1;
       if ($69) {
        label = 14;
        break L7;
       } else {
        break;
       }
      }
     }
    } else {
     label = 29;
    }
   } while(0);
   if ((label|0) == 29) {
    $71 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $72 = HEAP32[$71>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $73 = ((($72)) + -44|0); //@line 555 "modules/graphics/opengl/Graphics.cpp"
    $74 = ((($72)) + -40|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $75 = HEAP32[$74>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $76 = HEAP32[$73>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $77 = (($75) - ($76))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $78 = $77 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $rendertargetcount$0 = $78;
    label = 30;
   }
   do {
    if ((label|0) == 30) {
     $79 = HEAP32[$16>>2]|0; //@line 2255 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $80 = ($79|0)>(0); //@line 557 "modules/graphics/opengl/Graphics.cpp"
     L35: do {
      if ($80) {
       $i$030 = 0;
       L37: while(1) {
        $81 = HEAP32[$colorbuffers>>2]|0; //@line 2387 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $82 = $i$030 >>> 5; //@line 2387 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $83 = (($81) + ($82<<2)|0); //@line 2387 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $84 = $i$030 & 31; //@line 2387 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $85 = 1 << $84; //@line 2387 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $86 = HEAP32[$83>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $87 = $86 & $85; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $88 = ($87|0)!=(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
        $89 = ($i$030|0)<($rendertargetcount$0|0); //@line 559 "modules/graphics/opengl/Graphics.cpp"
        $or$cond = $89 & $88; //@line 559 "modules/graphics/opengl/Graphics.cpp"
        do {
         if ($or$cond) {
          $90 = (($i$030) + 36064)|0; //@line 560 "modules/graphics/opengl/Graphics.cpp"
          HEAP32[$3>>2] = $90; //@line 560 "modules/graphics/opengl/Graphics.cpp"
          $91 = HEAP32[$14>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $92 = HEAP32[$18>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $93 = ($91>>>0)<($92>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          if ($93) {
           HEAP32[$91>>2] = $90; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $94 = ((($91)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           HEAP32[$14>>2] = $94; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           break;
          } else {
           __THREW__ = 0;
           invoke_vii(3597,($attachments|0),($3|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $95 = __THREW__; __THREW__ = 0;
           $96 = $95&1;
           if ($96) {
            break L37;
           } else {
            break;
           }
          }
         }
        } while(0);
        $97 = (($i$030) + 1)|0; //@line 557 "modules/graphics/opengl/Graphics.cpp"
        $98 = HEAP32[$16>>2]|0; //@line 2255 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $99 = ($97|0)<($98|0); //@line 557 "modules/graphics/opengl/Graphics.cpp"
        if ($99) {
         $i$030 = $97;
        } else {
         break L35;
        }
       }
       $45 = ___cxa_find_matching_catch_2()|0;
       $46 = tempRet0;
       $lpad$phi$index = $45;$lpad$phi$index2 = $46;
       break L7;
      }
     } while(0);
     if ($stencil) {
      HEAP32[$4>>2] = 36128; //@line 565 "modules/graphics/opengl/Graphics.cpp"
      $100 = HEAP32[$14>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $101 = HEAP32[$18>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $102 = ($100>>>0)<($101>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($102) {
       HEAP32[$100>>2] = 36128; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $103 = ((($100)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$14>>2] = $103; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $107 = $103;$108 = $101;
      } else {
       __THREW__ = 0;
       invoke_vii(3597,($attachments|0),($4|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $104 = __THREW__; __THREW__ = 0;
       $105 = $104&1;
       if ($105) {
        label = 14;
        break L7;
       }
       $$pre = HEAP32[$14>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $$pre31 = HEAP32[$18>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $107 = $$pre;$108 = $$pre31;
      }
      HEAP32[$5>>2] = 36096; //@line 566 "modules/graphics/opengl/Graphics.cpp"
      $106 = ($107>>>0)<($108>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($106) {
       HEAP32[$107>>2] = 36096; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $109 = ((($107)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$14>>2] = $109; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       break;
      } else {
       __THREW__ = 0;
       invoke_vii(3597,($attachments|0),($5|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $110 = __THREW__; __THREW__ = 0;
       $111 = $110&1;
       if ($111) {
        label = 14;
        break L7;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   $112 = HEAP8[1252009]|0; //@line 571 "modules/graphics/opengl/Graphics.cpp"
   $113 = HEAP8[1252079]|0; //@line 571 "modules/graphics/opengl/Graphics.cpp"
   $114 = $113 | $112; //@line 571 "modules/graphics/opengl/Graphics.cpp"
   $115 = HEAP8[1252013]|0; //@line 571 "modules/graphics/opengl/Graphics.cpp"
   $116 = $114 | $115; //@line 571 "modules/graphics/opengl/Graphics.cpp"
   $117 = ($116<<24>>24)==(0); //@line 571 "modules/graphics/opengl/Graphics.cpp"
   if ($117) {
    $125 = HEAP8[1252174]|0; //@line 573 "modules/graphics/opengl/Graphics.cpp"
    $126 = ($125<<24>>24)==(0); //@line 573 "modules/graphics/opengl/Graphics.cpp"
    if (!($126)) {
     $127 = HEAP32[$14>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $128 = HEAP32[$attachments>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $129 = (($127) - ($128))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $130 = $129 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $$cast29 = $128; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $131 = HEAP32[166063]|0; //@line 1582 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_viii($131|0,36160,($130|0),($$cast29|0)); //@line 1582 "./libraries/glad/gladfuncs.hpp"
     $132 = __THREW__; __THREW__ = 0;
     $133 = $132&1;
     if ($133) {
      label = 14;
      break;
     }
    }
   } else {
    $118 = HEAP32[$14>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $119 = HEAP32[$attachments>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $120 = (($118) - ($119))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $121 = $120 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$cast = $119; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $122 = HEAP32[164995]|0; //@line 947 "./libraries/glad/gladfuncs.hpp"
    __THREW__ = 0;
    invoke_viii($122|0,36160,($121|0),($$cast|0)); //@line 947 "./libraries/glad/gladfuncs.hpp"
    $123 = __THREW__; __THREW__ = 0;
    $124 = $123&1;
    if ($124) {
     label = 14;
     break;
    }
   }
   $134 = HEAP32[$attachments>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $135 = ($134|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $136 = $134; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($135)) {
    $137 = HEAP32[$14>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $138 = ($137|0)==($134|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($138)) {
     $scevgep$i$i$i = ((($137)) + -4|0);
     $139 = $scevgep$i$i$i;
     $140 = (($139) - ($136))|0;
     $141 = $140 >>> 2;
     $142 = $141 ^ -1;
     $scevgep4$i$i$i = (($137) + ($142<<2)|0);
     HEAP32[$14>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($134); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   STACKTOP = sp;return; //@line 575 "modules/graphics/opengl/Graphics.cpp"
  }
 } while(0);
 if ((label|0) == 14) {
  $47 = ___cxa_find_matching_catch_2()|0;
  $48 = tempRet0;
  $lpad$phi$index = $47;$lpad$phi$index2 = $48;
 }
 $49 = HEAP32[$attachments>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $50 = ($49|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $51 = $49; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($50) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $52 = HEAP32[$14>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $53 = ($52|0)==($49|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($53)) {
  $scevgep$i$i$i14 = ((($52)) + -4|0);
  $54 = $scevgep$i$i$i14;
  $55 = (($54) - ($51))|0;
  $56 = $55 >>> 2;
  $57 = $56 ^ -1;
  $scevgep4$i$i$i15 = (($52) + ($57<<2)|0);
  HEAP32[$14>>2] = $scevgep4$i$i$i15; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($49); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZN4love8graphics6opengl8Graphics7presentEv($this) {
 $this = $this|0;
 var $$phi$trans$insert = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i$i = 0, $$pre$i$i$i5 = 0, $$pre$i8 = 0, $$sink$index = 0, $$sink$index2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $9 = 0, $canvases = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $canvases = sp + 12|0;
 $0 = sp;
 $1 = HEAP32[$this>>2]|0; //@line 579 "modules/graphics/opengl/Graphics.cpp"
 $2 = ((($1)) + 40|0); //@line 579 "modules/graphics/opengl/Graphics.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 579 "modules/graphics/opengl/Graphics.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($this)|0); //@line 579 "modules/graphics/opengl/Graphics.cpp"
 if (!($4)) {
  STACKTOP = sp;return; //@line 606 "modules/graphics/opengl/Graphics.cpp"
 }
 $5 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$5>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ((($6)) + -44|0); //@line 583 "modules/graphics/opengl/Graphics.cpp"
 __ZNSt3__16vectorIN4love9StrongRefINS1_8graphics6opengl6CanvasEEENS_9allocatorIS6_EEEC2ERKS9_($canvases,$7); //@line 583 "modules/graphics/opengl/Graphics.cpp"
 $8 = HEAP32[$5>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = HEAP32[158272]|0; //@line 1065 "modules/graphics/opengl/Graphics.cpp"
 $10 = ($9|0)==(0|0); //@line 1065 "modules/graphics/opengl/Graphics.cpp"
 if ($10) {
  label = 4;
 } else {
  __THREW__ = 0;
  invoke_vii(4001,($9|0),0); //@line 1066 "modules/graphics/opengl/Graphics.cpp"
  $11 = __THREW__; __THREW__ = 0;
  $12 = $11&1;
  if ($12) {
   label = 13;
  } else {
   label = 4;
  }
 }
 do {
  if ((label|0) == 4) {
   $13 = ((($8)) + -44|0); //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $14 = HEAP32[$13>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $15 = ((($8)) + -40|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $16 = HEAP32[$15>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $17 = ($16|0)==($14|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   L8: do {
    if (!($17)) {
     $19 = $16;
     while(1) {
      $18 = ((($19)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$15>>2] = $18; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $20 = HEAP32[$18>>2]|0; //@line 106 "./common/Object.h"
      $21 = ($20|0)==(0|0); //@line 106 "./common/Object.h"
      if ($21) {
       $23 = $18;
      } else {
       $24 = HEAP32[$20>>2]|0; //@line 106 "./common/Object.h"
       $25 = ((($24)) + 12|0); //@line 106 "./common/Object.h"
       $26 = HEAP32[$25>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($26|0,($20|0)); //@line 106 "./common/Object.h"
       $27 = __THREW__; __THREW__ = 0;
       $28 = $27&1;
       if ($28) {
        break;
       }
       $$pre$i$i$i$i = HEAP32[$15>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $23 = $$pre$i$i$i$i;
      }
      $22 = ($23|0)==($14|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($22) {
       break L8;
      } else {
       $19 = $23;
      }
     }
     $29 = ___cxa_find_matching_catch_3(0|0)|0;
     $30 = tempRet0;
     ___clang_call_terminate($29); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   } while(0);
   HEAP32[$0>>2] = 0; //@line 2563 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $31 = ((($0)) + 4|0); //@line 2564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$31>>2] = 0; //@line 2564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $32 = ((($0)) + 8|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$32>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __THREW__ = 0;
   invoke_viii(3533,($this|0),($0|0),1); //@line 587 "modules/graphics/opengl/Graphics.cpp"
   $33 = __THREW__; __THREW__ = 0;
   $34 = $33&1;
   if ($34) {
    $43 = ___cxa_find_matching_catch_2()|0;
    $44 = tempRet0;
    $$sink$index = $43;$$sink$index2 = $44;
    break;
   }
   $$phi$trans$insert = ((($this)) + 8|0); //@line 139 "./common/Object.h"
   $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 139 "./common/Object.h"
   $35 = ($$pre|0)==(0|0); //@line 597 "modules/graphics/opengl/Graphics.cpp"
   if (!($35)) {
    $36 = HEAP32[$$pre>>2]|0; //@line 598 "modules/graphics/opengl/Graphics.cpp"
    $37 = ((($36)) + 108|0); //@line 598 "modules/graphics/opengl/Graphics.cpp"
    $38 = HEAP32[$37>>2]|0; //@line 598 "modules/graphics/opengl/Graphics.cpp"
    __THREW__ = 0;
    invoke_vi($38|0,($$pre|0)); //@line 598 "modules/graphics/opengl/Graphics.cpp"
    $39 = __THREW__; __THREW__ = 0;
    $40 = $39&1;
    if ($40) {
     label = 13;
     break;
    }
   }
   __THREW__ = 0;
   invoke_vii(4046,($this|0),($canvases|0)); //@line 601 "modules/graphics/opengl/Graphics.cpp"
   $45 = __THREW__; __THREW__ = 0;
   $46 = $45&1;
   if ($46) {
    label = 13;
   } else {
    HEAP32[(634072)>>2] = 0; //@line 604 "modules/graphics/opengl/Graphics.cpp"
    HEAP32[(634076)>>2] = 0; //@line 605 "modules/graphics/opengl/Graphics.cpp"
    $47 = HEAP32[$canvases>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $48 = ($47|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($48)) {
     $49 = ((($canvases)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $50 = HEAP32[$49>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $51 = ($50|0)==($47|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     do {
      if ($51) {
       $65 = $47;
      } else {
       $53 = $50;
       while(1) {
        $52 = ((($53)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$49>>2] = $52; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $54 = HEAP32[$52>>2]|0; //@line 106 "./common/Object.h"
        $55 = ($54|0)==(0|0); //@line 106 "./common/Object.h"
        if ($55) {
         $57 = $52;
        } else {
         $58 = HEAP32[$54>>2]|0; //@line 106 "./common/Object.h"
         $59 = ((($58)) + 12|0); //@line 106 "./common/Object.h"
         $60 = HEAP32[$59>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($60|0,($54|0)); //@line 106 "./common/Object.h"
         $61 = __THREW__; __THREW__ = 0;
         $62 = $61&1;
         if ($62) {
          label = 22;
          break;
         }
         $$pre$i$i$i5 = HEAP32[$49>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $57 = $$pre$i$i$i5;
        }
        $56 = ($57|0)==($47|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($56) {
         label = 23;
         break;
        } else {
         $53 = $57;
        }
       }
       if ((label|0) == 22) {
        $63 = ___cxa_find_matching_catch_3(0|0)|0;
        $64 = tempRet0;
        ___clang_call_terminate($63); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       else if ((label|0) == 23) {
        $$pre$i8 = HEAP32[$canvases>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $65 = $$pre$i8;
        break;
       }
      }
     } while(0);
     __ZdlPv($65); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    }
    STACKTOP = sp;return; //@line 606 "modules/graphics/opengl/Graphics.cpp"
   }
  }
 } while(0);
 if ((label|0) == 13) {
  $41 = ___cxa_find_matching_catch_2()|0;
  $42 = tempRet0;
  $$sink$index = $41;$$sink$index2 = $42;
 }
 $66 = HEAP32[$canvases>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $67 = ($66|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($67) {
  ___resumeException($$sink$index|0);
  // unreachable;
 }
 $68 = ((($canvases)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $69 = HEAP32[$68>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $70 = ($69|0)==($66|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($70) {
   $84 = $66;
  } else {
   $72 = $69;
   while(1) {
    $71 = ((($72)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$68>>2] = $71; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $73 = HEAP32[$71>>2]|0; //@line 106 "./common/Object.h"
    $74 = ($73|0)==(0|0); //@line 106 "./common/Object.h"
    if ($74) {
     $76 = $71;
    } else {
     $77 = HEAP32[$73>>2]|0; //@line 106 "./common/Object.h"
     $78 = ((($77)) + 12|0); //@line 106 "./common/Object.h"
     $79 = HEAP32[$78>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($79|0,($73|0)); //@line 106 "./common/Object.h"
     $80 = __THREW__; __THREW__ = 0;
     $81 = $80&1;
     if ($81) {
      label = 33;
      break;
     }
     $$pre$i$i$i = HEAP32[$68>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $76 = $$pre$i$i$i;
    }
    $75 = ($76|0)==($66|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($75) {
     label = 34;
     break;
    } else {
     $72 = $76;
    }
   }
   if ((label|0) == 33) {
    $82 = ___cxa_find_matching_catch_3(0|0)|0;
    $83 = tempRet0;
    ___clang_call_terminate($82); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 34) {
    $$pre$i = HEAP32[$canvases>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $84 = $$pre$i;
    break;
   }
  }
 } while(0);
 __ZdlPv($84); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$sink$index|0);
 // unreachable;
}
function __ZNK4love8graphics6opengl8Graphics8getWidthEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 610 "modules/graphics/opengl/Graphics.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 610 "modules/graphics/opengl/Graphics.cpp"
 return ($1|0); //@line 610 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNK4love8graphics6opengl8Graphics9getHeightEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 36|0); //@line 615 "modules/graphics/opengl/Graphics.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 615 "modules/graphics/opengl/Graphics.cpp"
 return ($1|0); //@line 615 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics16intersectScissorEiiii($this,$x,$y,$width,$height) {
 $this = $this|0;
 $x = $x|0;
 $y = $y|0;
 $width = $width|0;
 $height = $height|0;
 var $$2 = 0, $$3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($2)) + -76|0); //@line 637 "modules/graphics/opengl/Graphics.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 637 "modules/graphics/opengl/Graphics.cpp"
 $5 = ((($2)) + -72|0); //@line 637 "modules/graphics/opengl/Graphics.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 637 "modules/graphics/opengl/Graphics.cpp"
 $7 = ((($2)) + -68|0); //@line 637 "modules/graphics/opengl/Graphics.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 637 "modules/graphics/opengl/Graphics.cpp"
 $9 = ((($2)) + -64|0); //@line 637 "modules/graphics/opengl/Graphics.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 637 "modules/graphics/opengl/Graphics.cpp"
 $11 = ((($2)) + -80|0); //@line 639 "modules/graphics/opengl/Graphics.cpp"
 $12 = HEAP8[$11>>0]|0; //@line 639 "modules/graphics/opengl/Graphics.cpp"
 $13 = ($12<<24>>24)==(0); //@line 639 "modules/graphics/opengl/Graphics.cpp"
 $$2 = $13 ? 0 : $6; //@line 639 "modules/graphics/opengl/Graphics.cpp"
 $$3 = $13 ? 0 : $4; //@line 639 "modules/graphics/opengl/Graphics.cpp"
 $14 = ($$3|0)<($x|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $15 = $14 ? $x : $$3; //@line 647 "modules/graphics/opengl/Graphics.cpp"
 $16 = ($$2|0)<($y|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $17 = $16 ? $y : $$2; //@line 648 "modules/graphics/opengl/Graphics.cpp"
 $18 = (($8) + ($4))|0; //@line 650 "modules/graphics/opengl/Graphics.cpp"
 $19 = $13 ? 2147483647 : $18; //@line 650 "modules/graphics/opengl/Graphics.cpp"
 $20 = (($width) + ($x))|0; //@line 650 "modules/graphics/opengl/Graphics.cpp"
 $21 = ($20|0)<($19|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $22 = $21 ? $20 : $19; //@line 650 "modules/graphics/opengl/Graphics.cpp"
 $23 = (($10) + ($6))|0; //@line 651 "modules/graphics/opengl/Graphics.cpp"
 $24 = $13 ? 2147483647 : $23; //@line 651 "modules/graphics/opengl/Graphics.cpp"
 $25 = (($height) + ($y))|0; //@line 651 "modules/graphics/opengl/Graphics.cpp"
 $26 = ($25|0)<($24|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $27 = $26 ? $25 : $24; //@line 651 "modules/graphics/opengl/Graphics.cpp"
 $28 = (($22) - ($15))|0; //@line 653 "modules/graphics/opengl/Graphics.cpp"
 $29 = ($28|0)>(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $30 = $29 ? $28 : 0; //@line 653 "modules/graphics/opengl/Graphics.cpp"
 $31 = (($27) - ($17))|0; //@line 653 "modules/graphics/opengl/Graphics.cpp"
 $32 = ($31|0)>(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $33 = $32 ? $31 : 0; //@line 653 "modules/graphics/opengl/Graphics.cpp"
 $34 = HEAP32[164832]|0; //@line 57 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$34 & 8191](3089); //@line 57 "./libraries/glad/gladfuncs.hpp"
 HEAP32[$0>>2] = $15; //@line 629 "modules/graphics/opengl/Graphics.cpp"
 $35 = ((($0)) + 4|0); //@line 629 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$35>>2] = $17; //@line 629 "modules/graphics/opengl/Graphics.cpp"
 $36 = ((($0)) + 8|0); //@line 629 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$36>>2] = $30; //@line 629 "modules/graphics/opengl/Graphics.cpp"
 $37 = ((($0)) + 12|0); //@line 629 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$37>>2] = $33; //@line 629 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl6OpenGL10setScissorERKNS2_8ViewportE(634044,$0); //@line 629 "modules/graphics/opengl/Graphics.cpp"
 $38 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $39 = ((($38)) + -80|0); //@line 631 "modules/graphics/opengl/Graphics.cpp"
 HEAP8[$39>>0] = 1; //@line 631 "modules/graphics/opengl/Graphics.cpp"
 $40 = ((($38)) + -76|0); //@line 632 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$40>>2] = $15; //@line 632 "modules/graphics/opengl/Graphics.cpp"
 $41 = ((($38)) + -72|0); //@line 632 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$41>>2] = $17; //@line 632 "modules/graphics/opengl/Graphics.cpp"
 $42 = ((($38)) + -68|0); //@line 632 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$42>>2] = $30; //@line 632 "modules/graphics/opengl/Graphics.cpp"
 $43 = ((($38)) + -64|0); //@line 632 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$43>>2] = $33; //@line 632 "modules/graphics/opengl/Graphics.cpp"
 STACKTOP = sp;return; //@line 654 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNK4love8graphics6opengl8Graphics10getScissorERiS3_S3_S3_($this,$x,$y,$width,$height) {
 $this = $this|0;
 $x = $x|0;
 $y = $y|0;
 $width = $width|0;
 $height = $height|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -76|0); //@line 666 "modules/graphics/opengl/Graphics.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 666 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$x>>2] = $3; //@line 666 "modules/graphics/opengl/Graphics.cpp"
 $4 = ((($1)) + -72|0); //@line 667 "modules/graphics/opengl/Graphics.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 667 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$y>>2] = $5; //@line 667 "modules/graphics/opengl/Graphics.cpp"
 $6 = ((($1)) + -68|0); //@line 668 "modules/graphics/opengl/Graphics.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 668 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$width>>2] = $7; //@line 668 "modules/graphics/opengl/Graphics.cpp"
 $8 = ((($1)) + -64|0); //@line 669 "modules/graphics/opengl/Graphics.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 669 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$height>>2] = $9; //@line 669 "modules/graphics/opengl/Graphics.cpp"
 $10 = ((($1)) + -80|0); //@line 671 "modules/graphics/opengl/Graphics.cpp"
 $11 = HEAP8[$10>>0]|0; //@line 671 "modules/graphics/opengl/Graphics.cpp"
 $12 = ($11<<24>>24)!=(0); //@line 671 "modules/graphics/opengl/Graphics.cpp"
 return ($12|0); //@line 671 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics19drawToStencilBufferENS0_8Graphics13StencilActionEi($this,$action,$value) {
 $this = $this|0;
 $action = $action|0;
 $value = $value|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $glaction$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 42|0); //@line 676 "modules/graphics/opengl/Graphics.cpp"
 HEAP8[$0>>0] = 1; //@line 676 "modules/graphics/opengl/Graphics.cpp"
 $1 = HEAP32[158272]|0; //@line 679 "modules/graphics/opengl/Graphics.cpp"
 $2 = ($1|0)==(0|0); //@line 679 "modules/graphics/opengl/Graphics.cpp"
 if (!($2)) {
  (__ZN4love8graphics6opengl6Canvas18checkCreateStencilEv($1)|0); //@line 680 "modules/graphics/opengl/Graphics.cpp"
 }
 $3 = HEAP32[165466]|0; //@line 54 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiii[$3 & 8191](0,0,0,0); //@line 54 "./libraries/glad/gladfuncs.hpp"
 switch ($action|0) {
 case 5:  {
  $glaction$0 = 5386;
  break;
 }
 case 1:  {
  $glaction$0 = 7682;
  break;
 }
 case 2:  {
  $glaction$0 = 7683;
  break;
 }
 case 3:  {
  $glaction$0 = 34055;
  break;
 }
 case 4:  {
  $glaction$0 = 34056;
  break;
 }
 default: {
  $glaction$0 = 7681;
 }
 }
 $4 = HEAP32[164832]|0; //@line 57 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$4 & 8191](2960); //@line 57 "./libraries/glad/gladfuncs.hpp"
 $5 = HEAP32[165322]|0; //@line 62 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$5 & 8191](519,$value,-1); //@line 62 "./libraries/glad/gladfuncs.hpp"
 $6 = HEAP32[164775]|0; //@line 63 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$6 & 8191](7680,7680,$glaction$0); //@line 63 "./libraries/glad/gladfuncs.hpp"
 return; //@line 714 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics14getStencilTestERNS0_8Graphics11CompareModeERi($this,$compare,$value) {
 $this = $this|0;
 $compare = $compare|0;
 $value = $value|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -60|0); //@line 804 "modules/graphics/opengl/Graphics.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 804 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$compare>>2] = $3; //@line 804 "modules/graphics/opengl/Graphics.cpp"
 $4 = ((($1)) + -56|0); //@line 805 "modules/graphics/opengl/Graphics.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 805 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$value>>2] = $5; //@line 805 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 806 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics12clearStencilEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[164768]|0; //@line 49 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$0 & 8191](1280); //@line 49 "./libraries/glad/gladfuncs.hpp"
 return; //@line 811 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics8newImageERKNSt3__16vectorIPNS_5image9ImageDataENS3_9allocatorIS7_EEEERKNS1_5Image5FlagsE($this,$data,$flags) {
 $this = $this|0;
 $data = $data|0;
 $flags = $flags|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(168)|0); //@line 815 "modules/graphics/opengl/Graphics.cpp"
 __THREW__ = 0;
 invoke_viii(4062,($0|0),($data|0),($flags|0)); //@line 815 "modules/graphics/opengl/Graphics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 815 "modules/graphics/opengl/Graphics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 815 "modules/graphics/opengl/Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl8Graphics8newImageERKNSt3__16vectorIPNS_5image19CompressedImageDataENS3_9allocatorIS7_EEEERKNS1_5Image5FlagsE($this,$cdata,$flags) {
 $this = $this|0;
 $cdata = $cdata|0;
 $flags = $flags|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(168)|0); //@line 820 "modules/graphics/opengl/Graphics.cpp"
 __THREW__ = 0;
 invoke_viii(4063,($0|0),($cdata|0),($flags|0)); //@line 820 "modules/graphics/opengl/Graphics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 820 "modules/graphics/opengl/Graphics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 820 "modules/graphics/opengl/Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl8Graphics7newQuadENS0_4Quad8ViewportEdd($this,$v,$sw,$sh) {
 $this = $this|0;
 $v = $v|0;
 $sw = +$sw;
 $sh = +$sh;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(136)|0); //@line 825 "modules/graphics/opengl/Graphics.cpp"
 __THREW__ = 0;
 invoke_viidd(4064,($0|0),($v|0),(+$sw),(+$sh)); //@line 825 "modules/graphics/opengl/Graphics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 825 "modules/graphics/opengl/Graphics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 825 "modules/graphics/opengl/Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl8Graphics14newSpriteBatchEPNS0_7TextureEiNS1_4Mesh5UsageE($this,$texture,$size,$usage) {
 $this = $this|0;
 $texture = $texture|0;
 $size = $size|0;
 $usage = $usage|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(52)|0); //@line 835 "modules/graphics/opengl/Graphics.cpp"
 __THREW__ = 0;
 invoke_viiii(4065,($0|0),($texture|0),($size|0),($usage|0)); //@line 835 "modules/graphics/opengl/Graphics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 835 "modules/graphics/opengl/Graphics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 835 "modules/graphics/opengl/Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl8Graphics17newParticleSystemEPNS0_7TextureEi($this,$texture,$size) {
 $this = $this|0;
 $texture = $texture|0;
 $size = $size|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(236)|0); //@line 840 "modules/graphics/opengl/Graphics.cpp"
 __THREW__ = 0;
 invoke_viii(4066,($0|0),($texture|0),($size|0)); //@line 840 "modules/graphics/opengl/Graphics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 840 "modules/graphics/opengl/Graphics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 840 "modules/graphics/opengl/Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl8Graphics9newCanvasEiiNS1_6Canvas6FormatEi($this,$width,$height,$format,$msaa) {
 $this = $this|0;
 $width = $width|0;
 $height = $height|0;
 $format = $format|0;
 $msaa = $msaa|0;
 var $$0 = 0, $$0$i = 0, $$01$i = 0, $$0114 = 0, $$0313 = 0, $$12 = 0, $$14 = 0, $$2 = 0, $$25 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$i$index13Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $error_string = 0, $fstr = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer6 = 0, $vararg_buffer9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 208|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer9 = sp + 32|0;
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 196|0;
 $fstr = sp + 192|0;
 $error_string = sp + 48|0;
 $1 = sp + 36|0;
 $2 = (__ZN4love8graphics6opengl6Canvas11isSupportedEv()|0); //@line 845 "modules/graphics/opengl/Graphics.cpp"
 if (!($2)) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 846 "modules/graphics/opengl/Graphics.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($3|0),(435237|0),($vararg_buffer|0)); //@line 846 "modules/graphics/opengl/Graphics.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if (!($5)) {
   ___cxa_throw(($3|0),(96|0),(344|0)); //@line 846 "modules/graphics/opengl/Graphics.cpp"
   // unreachable; //@line 846 "modules/graphics/opengl/Graphics.cpp"
  }
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  ___cxa_free_exception(($3|0)); //@line 846 "modules/graphics/opengl/Graphics.cpp"
  $$2 = $6;$$25 = $7;
  ___resumeException($$2|0);
  // unreachable;
 }
 $8 = (__ZN4love8graphics6opengl6Canvas17isFormatSupportedENS2_6FormatE($format)|0); //@line 848 "modules/graphics/opengl/Graphics.cpp"
 if (!($8)) {
  HEAP32[$fstr>>2] = 435288; //@line 850 "modules/graphics/opengl/Graphics.cpp"
  $9 = (__ZN4love8graphics6opengl6Canvas14getSizedFormatENS2_6FormatE($format)|0); //@line 851 "modules/graphics/opengl/Graphics.cpp"
  (__ZN4love8graphics6opengl6Canvas11getConstantENS2_6FormatERPKc($9,$fstr)|0); //@line 851 "modules/graphics/opengl/Graphics.cpp"
  $10 = (___cxa_allocate_exception(16)|0); //@line 852 "modules/graphics/opengl/Graphics.cpp"
  $11 = HEAP32[$fstr>>2]|0; //@line 852 "modules/graphics/opengl/Graphics.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer1>>2] = $11; //@line 852 "modules/graphics/opengl/Graphics.cpp"
  invoke_viii(3349,($10|0),(435294|0),($vararg_buffer1|0)); //@line 852 "modules/graphics/opengl/Graphics.cpp"
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   ___cxa_throw(($10|0),(96|0),(344|0)); //@line 852 "modules/graphics/opengl/Graphics.cpp"
   // unreachable; //@line 852 "modules/graphics/opengl/Graphics.cpp"
  }
  $14 = ___cxa_find_matching_catch_2()|0;
  $15 = tempRet0;
  ___cxa_free_exception(($10|0)); //@line 852 "modules/graphics/opengl/Graphics.cpp"
  $$2 = $14;$$25 = $15;
  ___resumeException($$2|0);
  // unreachable;
 }
 $16 = (__ZNK4love8graphics6opengl6OpenGL17getMaxTextureSizeEv(634044)|0); //@line 855 "modules/graphics/opengl/Graphics.cpp"
 $17 = ($16|0)<($width|0); //@line 855 "modules/graphics/opengl/Graphics.cpp"
 if ($17) {
  $18 = (___cxa_allocate_exception(16)|0); //@line 856 "modules/graphics/opengl/Graphics.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer3>>2] = $width; //@line 856 "modules/graphics/opengl/Graphics.cpp"
  invoke_viii(3349,($18|0),(435356|0),($vararg_buffer3|0)); //@line 856 "modules/graphics/opengl/Graphics.cpp"
  $19 = __THREW__; __THREW__ = 0;
  $20 = $19&1;
  if (!($20)) {
   ___cxa_throw(($18|0),(96|0),(344|0)); //@line 856 "modules/graphics/opengl/Graphics.cpp"
   // unreachable; //@line 856 "modules/graphics/opengl/Graphics.cpp"
  }
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  ___cxa_free_exception(($18|0)); //@line 856 "modules/graphics/opengl/Graphics.cpp"
  $$2 = $21;$$25 = $22;
  ___resumeException($$2|0);
  // unreachable;
 }
 $23 = (__ZNK4love8graphics6opengl6OpenGL17getMaxTextureSizeEv(634044)|0); //@line 857 "modules/graphics/opengl/Graphics.cpp"
 $24 = ($23|0)<($height|0); //@line 857 "modules/graphics/opengl/Graphics.cpp"
 if ($24) {
  $25 = (___cxa_allocate_exception(16)|0); //@line 858 "modules/graphics/opengl/Graphics.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer6>>2] = $height; //@line 858 "modules/graphics/opengl/Graphics.cpp"
  invoke_viii(3349,($25|0),(435427|0),($vararg_buffer6|0)); //@line 858 "modules/graphics/opengl/Graphics.cpp"
  $26 = __THREW__; __THREW__ = 0;
  $27 = $26&1;
  if (!($27)) {
   ___cxa_throw(($25|0),(96|0),(344|0)); //@line 858 "modules/graphics/opengl/Graphics.cpp"
   // unreachable; //@line 858 "modules/graphics/opengl/Graphics.cpp"
  }
  $28 = ___cxa_find_matching_catch_2()|0;
  $29 = tempRet0;
  ___cxa_free_exception(($25|0)); //@line 858 "modules/graphics/opengl/Graphics.cpp"
  $$2 = $28;$$25 = $29;
  ___resumeException($$2|0);
  // unreachable;
 }
 while(1) {
  $30 = HEAP32[165571]|0; //@line 71 "./libraries/glad/gladfuncs.hpp"
  $31 = (FUNCTION_TABLE_i[$30 & 8191]()|0); //@line 71 "./libraries/glad/gladfuncs.hpp"
  $32 = ($31|0)==(0); //@line 860 "modules/graphics/opengl/Graphics.cpp"
  if ($32) {
   break;
  }
 }
 $33 = (__Znwj(176)|0); //@line 863 "modules/graphics/opengl/Graphics.cpp"
 __THREW__ = 0;
 invoke_viiiii(4067,($33|0),($width|0),($height|0),($format|0),($msaa|0)); //@line 863 "modules/graphics/opengl/Graphics.cpp"
 $34 = __THREW__; __THREW__ = 0;
 $35 = $34&1;
 if ($35) {
  $39 = ___cxa_find_matching_catch_2()|0;
  $40 = tempRet0;
  __ZdlPv($33); //@line 863 "modules/graphics/opengl/Graphics.cpp"
  $$2 = $39;$$25 = $40;
  ___resumeException($$2|0);
  // unreachable;
 }
 $36 = ((($33)) + 148|0); //@line 107 "modules/graphics/opengl/Canvas.h"
 $37 = HEAP32[$36>>2]|0; //@line 107 "modules/graphics/opengl/Canvas.h"
 $38 = ($37|0)==(36053); //@line 867 "modules/graphics/opengl/Graphics.cpp"
 if ($38) {
  STACKTOP = sp;return ($33|0); //@line 905 "modules/graphics/opengl/Graphics.cpp"
 }
 $41 = ((($error_string)) + 64|0); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $42 = ((($error_string)) + 8|0); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$42>>2] = (9320); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $43 = ((($error_string)) + 12|0); //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$error_string>>2] = ((9524)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 HEAP32[$41>>2] = ((9544)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $44 = ((($error_string)) + 4|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 HEAP32[$44>>2] = 0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $45 = ((($error_string)) + 64|0); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 __THREW__ = 0;
 invoke_vii(3362,($45|0),($43|0)); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 $46 = __THREW__; __THREW__ = 0;
 $47 = $46&1;
 if ($47) {
  $58 = ___cxa_find_matching_catch_2()|0;
  $59 = tempRet0;
  $$0$i = $59;$$01$i = $58;
  __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($41); //@line 900 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  ___resumeException($$01$i|0);
  // unreachable;
 }
 $48 = ((($error_string)) + 136|0); //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$48>>2] = 0; //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 $49 = ((($error_string)) + 140|0); //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$49>>2] = -1; //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$error_string>>2] = (9300); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$41>>2] = (9340); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$42>>2] = (9320); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 __THREW__ = 0;
 invoke_vi(3363,($43|0)); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $50 = __THREW__; __THREW__ = 0;
 $51 = $50&1;
 do {
  if ($51) {
   $60 = ___cxa_find_matching_catch_2()|0;
   $61 = tempRet0;
   $eh$lpad$body$i$index13Z2D = $61;$eh$lpad$body$i$indexZ2D = $60;
  } else {
   HEAP32[$43>>2] = (9356); //@line 243 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $52 = ((($error_string)) + 44|0); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $53 = ((($error_string)) + 60|0); //@line 242 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   ;HEAP32[$52>>2]=0|0;HEAP32[$52+4>>2]=0|0;HEAP32[$52+8>>2]=0|0;HEAP32[$52+12>>2]=0|0; //@line 241 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   HEAP32[$53>>2] = 24; //@line 242 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   invoke_vii(3364,($43|0),($0|0)); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $54 = __THREW__; __THREW__ = 0;
   $55 = $54&1;
   if ($55) {
    $56 = ___cxa_find_matching_catch_2()|0;
    $57 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($52); //@line 245 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($43); //@line 245 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $eh$lpad$body$i$index13Z2D = $57;$eh$lpad$body$i$indexZ2D = $56;
    break;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $62 = ((($error_string)) + 8|0); //@line 872 "modules/graphics/opengl/Graphics.cpp"
   __THREW__ = 0;
   (invoke_iiii(3366,($62|0),(435499|0),22)|0); //@line 894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
   $63 = __THREW__; __THREW__ = 0;
   $64 = $63&1;
   L48: do {
    if ($64) {
     label = 32;
    } else {
     L50: do {
      switch ($37|0) {
      case 36061:  {
       __THREW__ = 0;
       (invoke_iiii(3366,($62|0),(435522|0),37)|0); //@line 894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
       $67 = __THREW__; __THREW__ = 0;
       $68 = $67&1;
       if ($68) {
        label = 32;
        break L48;
       }
       break;
      }
      case 36054:  {
       __THREW__ = 0;
       (invoke_iiii(3366,($62|0),(435560|0),52)|0); //@line 894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
       $69 = __THREW__; __THREW__ = 0;
       $70 = $69&1;
       if ($70) {
        label = 32;
        break L48;
       }
       break;
      }
      case 36182: case 36060: case 36059: case 36055: case 33305:  {
       __THREW__ = 0;
       (invoke_iiii(3366,($62|0),(435613|0),25)|0); //@line 894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
       $71 = __THREW__; __THREW__ = 0;
       $72 = $71&1;
       if ($72) {
        label = 32;
        break L48;
       }
       break;
      }
      default: {
       $73 = HEAP32[165571]|0; //@line 71 "./libraries/glad/gladfuncs.hpp"
       __THREW__ = 0;
       $74 = (invoke_i($73|0)|0); //@line 71 "./libraries/glad/gladfuncs.hpp"
       $75 = __THREW__; __THREW__ = 0;
       $76 = $75&1;
       if ($76) {
        label = 32;
        break L48;
       }
       $77 = ($74|0)==(0); //@line 893 "modules/graphics/opengl/Graphics.cpp"
       if ($77) {
        __THREW__ = 0;
        (invoke_iiii(3366,($62|0),(435639|0),44)|0); //@line 894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
        $78 = __THREW__; __THREW__ = 0;
        $79 = $78&1;
        if ($79) {
         label = 32;
         break L48;
        } else {
         break L50;
        }
       }
       __THREW__ = 0;
       $80 = (invoke_iiii(3366,($62|0),(435684|0),56)|0); //@line 894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
       $81 = __THREW__; __THREW__ = 0;
       $82 = $81&1;
       if ($82) {
        label = 32;
        break L48;
       }
       $83 = HEAP32[165571]|0; //@line 71 "./libraries/glad/gladfuncs.hpp"
       __THREW__ = 0;
       $84 = (invoke_i($83|0)|0); //@line 71 "./libraries/glad/gladfuncs.hpp"
       $85 = __THREW__; __THREW__ = 0;
       $86 = $85&1;
       if ($86) {
        label = 32;
        break L48;
       }
       __THREW__ = 0;
       $87 = (invoke_iii(4068,($80|0),($84|0))|0); //@line 899 "modules/graphics/opengl/Graphics.cpp"
       $88 = __THREW__; __THREW__ = 0;
       $89 = $88&1;
       if ($89) {
        label = 32;
        break L48;
       }
       __THREW__ = 0;
       (invoke_iiii(3366,($87|0),(442473|0),1)|0); //@line 894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
       $90 = __THREW__; __THREW__ = 0;
       $91 = $90&1;
       if ($91) {
        label = 32;
        break L48;
       }
      }
      }
     } while(0);
     $92 = HEAP32[$33>>2]|0; //@line 902 "modules/graphics/opengl/Graphics.cpp"
     $93 = ((($92)) + 12|0); //@line 902 "modules/graphics/opengl/Graphics.cpp"
     $94 = HEAP32[$93>>2]|0; //@line 902 "modules/graphics/opengl/Graphics.cpp"
     __THREW__ = 0;
     invoke_vi($94|0,($33|0)); //@line 902 "modules/graphics/opengl/Graphics.cpp"
     $95 = __THREW__; __THREW__ = 0;
     $96 = $95&1;
     if ($96) {
      label = 32;
     } else {
      $97 = (___cxa_allocate_exception(16)|0); //@line 903 "modules/graphics/opengl/Graphics.cpp"
      __THREW__ = 0;
      invoke_vii(3367,($1|0),($43|0)); //@line 964 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
      $98 = __THREW__; __THREW__ = 0;
      $99 = $98&1;
      if ($99) {
       $110 = ___cxa_find_matching_catch_2()|0;
       $111 = tempRet0;
       $$0114 = $110;$$0313 = $111;
      } else {
       $100 = HEAP8[$1>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $101 = $100 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $102 = ($101<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $103 = ((($1)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $104 = HEAP32[$103>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $105 = ((($1)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $106 = $102 ? $105 : $104; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       __THREW__ = 0;
       HEAP32[$vararg_buffer9>>2] = $106; //@line 903 "modules/graphics/opengl/Graphics.cpp"
       invoke_viii(3349,($97|0),(595867|0),($vararg_buffer9|0)); //@line 903 "modules/graphics/opengl/Graphics.cpp"
       $107 = __THREW__; __THREW__ = 0;
       $108 = $107&1;
       if ($108) {
        $$0 = 1;
       } else {
        __THREW__ = 0;
        invoke_viii(3350,($97|0),(96|0),(344|0)); //@line 903 "modules/graphics/opengl/Graphics.cpp"
        $109 = __THREW__; __THREW__ = 0;
        $$0 = 0;
       }
       $112 = ___cxa_find_matching_catch_2()|0;
       $113 = tempRet0;
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 903 "modules/graphics/opengl/Graphics.cpp"
       if ($$0) {
        $$0114 = $112;$$0313 = $113;
       } else {
        $$12 = $112;$$14 = $113;
        break;
       }
      }
      ___cxa_free_exception(($97|0)); //@line 903 "modules/graphics/opengl/Graphics.cpp"
      $$12 = $$0114;$$14 = $$0313;
     }
    }
   } while(0);
   if ((label|0) == 32) {
    $65 = ___cxa_find_matching_catch_2()|0;
    $66 = tempRet0;
    $$12 = $65;$$14 = $66;
   }
   HEAP32[$error_string>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$41>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$62>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$43>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($52); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($43); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($41); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   $$2 = $$12;$$25 = $$14;
   ___resumeException($$2|0);
   // unreachable;
  }
 } while(0);
 $$0$i = $eh$lpad$body$i$index13Z2D;$$01$i = $eh$lpad$body$i$indexZ2D;
 __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($41); //@line 900 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 ___resumeException($$01$i|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8graphics6opengl8Graphics7newMeshERKNSt3__16vectorINS_6VertexENS3_9allocatorIS5_EEEENS1_4Mesh8DrawModeENSB_5UsageE($this,$vertices,$drawmode,$usage) {
 $this = $this|0;
 $vertices = $vertices|0;
 $drawmode = $drawmode|0;
 $usage = $usage|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(100)|0); //@line 914 "modules/graphics/opengl/Graphics.cpp"
 __THREW__ = 0;
 invoke_viiii(4069,($0|0),($vertices|0),($drawmode|0),($usage|0)); //@line 914 "modules/graphics/opengl/Graphics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 914 "modules/graphics/opengl/Graphics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 914 "modules/graphics/opengl/Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl8Graphics7newMeshEiNS1_4Mesh8DrawModeENS3_5UsageE($this,$vertexcount,$drawmode,$usage) {
 $this = $this|0;
 $vertexcount = $vertexcount|0;
 $drawmode = $drawmode|0;
 $usage = $usage|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(100)|0); //@line 919 "modules/graphics/opengl/Graphics.cpp"
 __THREW__ = 0;
 invoke_viiii(4070,($0|0),($vertexcount|0),($drawmode|0),($usage|0)); //@line 919 "modules/graphics/opengl/Graphics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 919 "modules/graphics/opengl/Graphics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 919 "modules/graphics/opengl/Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl8Graphics7newMeshERKNSt3__16vectorINS1_4Mesh12AttribFormatENS3_9allocatorIS6_EEEEiNS5_8DrawModeENS5_5UsageE($this,$vertexformat,$vertexcount,$drawmode,$usage) {
 $this = $this|0;
 $vertexformat = $vertexformat|0;
 $vertexcount = $vertexcount|0;
 $drawmode = $drawmode|0;
 $usage = $usage|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(100)|0); //@line 924 "modules/graphics/opengl/Graphics.cpp"
 __THREW__ = 0;
 invoke_viiiii(4071,($0|0),($vertexformat|0),($vertexcount|0),($drawmode|0),($usage|0)); //@line 924 "modules/graphics/opengl/Graphics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 924 "modules/graphics/opengl/Graphics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 924 "modules/graphics/opengl/Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl8Graphics7newMeshERKNSt3__16vectorINS1_4Mesh12AttribFormatENS3_9allocatorIS6_EEEEPKvjNS5_8DrawModeENS5_5UsageE($this,$vertexformat,$data,$datasize,$drawmode,$usage) {
 $this = $this|0;
 $vertexformat = $vertexformat|0;
 $data = $data|0;
 $datasize = $datasize|0;
 $drawmode = $drawmode|0;
 $usage = $usage|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(100)|0); //@line 929 "modules/graphics/opengl/Graphics.cpp"
 __THREW__ = 0;
 invoke_viiiiii(4072,($0|0),($vertexformat|0),($data|0),($datasize|0),($drawmode|0),($usage|0)); //@line 929 "modules/graphics/opengl/Graphics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 929 "modules/graphics/opengl/Graphics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 929 "modules/graphics/opengl/Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl8Graphics7newTextEPNS1_4FontERKNSt3__16vectorINS3_13ColoredStringENS5_9allocatorIS7_EEEE($this,$font,$text) {
 $this = $this|0;
 $font = $font|0;
 $text = $text|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(48)|0); //@line 934 "modules/graphics/opengl/Graphics.cpp"
 __THREW__ = 0;
 invoke_viii(4073,($0|0),($font|0),($text|0)); //@line 934 "modules/graphics/opengl/Graphics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 934 "modules/graphics/opengl/Graphics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 934 "modules/graphics/opengl/Graphics.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl8Graphics8newVideoEPNS_5video11VideoStreamE($this,$stream) {
 $this = $this|0;
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(128)|0); //@line 939 "modules/graphics/opengl/Graphics.cpp"
 __THREW__ = 0;
 invoke_vii(4074,($0|0),($stream|0)); //@line 939 "modules/graphics/opengl/Graphics.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 939 "modules/graphics/opengl/Graphics.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 939 "modules/graphics/opengl/Graphics.cpp"
 }
 return (0)|0;
}
function __ZNK4love8graphics6opengl8Graphics8getColorEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -136|0); //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 ;HEAP32[$agg$result>>2]=HEAP32[$2>>2]|0;HEAP32[$agg$result+4>>2]=HEAP32[$2+4>>2]|0;HEAP32[$agg$result+8>>2]=HEAP32[$2+8>>2]|0;HEAP32[$agg$result+12>>2]=HEAP32[$2+12>>2]|0; //@line 959 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 959 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNK4love8graphics6opengl8Graphics18getBackgroundColorEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -120|0); //@line 969 "modules/graphics/opengl/Graphics.cpp"
 ;HEAP32[$agg$result>>2]=HEAP32[$2>>2]|0;HEAP32[$agg$result+4>>2]=HEAP32[$2+4>>2]|0;HEAP32[$agg$result+8>>2]=HEAP32[$2+8>>2]|0;HEAP32[$agg$result+12>>2]=HEAP32[$2+12>>2]|0; //@line 969 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 969 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics7getFontEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl8Graphics19checkSetDefaultFontEv($this); //@line 983 "modules/graphics/opengl/Graphics.cpp"
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -52|0); //@line 139 "./common/Object.h"
 $3 = HEAP32[$2>>2]|0; //@line 139 "./common/Object.h"
 return ($3|0); //@line 984 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics9setShaderEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 __ZN4love8graphics6opengl6Shader6detachEv(); //@line 1004 "modules/graphics/opengl/Graphics.cpp"
 $2 = ((($1)) + -48|0); //@line 133 "./common/Object.h"
 $3 = HEAP32[$2>>2]|0; //@line 133 "./common/Object.h"
 $4 = ($3|0)==(0|0); //@line 133 "./common/Object.h"
 if ($4) {
  HEAP32[$2>>2] = 0; //@line 134 "./common/Object.h"
  return; //@line 1007 "modules/graphics/opengl/Graphics.cpp"
 }
 $5 = HEAP32[$3>>2]|0; //@line 133 "./common/Object.h"
 $6 = ((($5)) + 12|0); //@line 133 "./common/Object.h"
 $7 = HEAP32[$6>>2]|0; //@line 133 "./common/Object.h"
 FUNCTION_TABLE_vi[$7 & 8191]($3); //@line 133 "./common/Object.h"
 HEAP32[$2>>2] = 0; //@line 134 "./common/Object.h"
 return; //@line 1007 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNK4love8graphics6opengl8Graphics9getShaderEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -48|0); //@line 139 "./common/Object.h"
 $3 = HEAP32[$2>>2]|0; //@line 139 "./common/Object.h"
 return ($3|0); //@line 1011 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNK4love8graphics6opengl8Graphics9getCanvasEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$phi$trans$insert = 0, $$pre = 0, $$pre3 = 0, $$pre4 = 0, $$pre5 = 0, $$pre6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$02 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 HEAP32[$agg$result>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($agg$result)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($agg$result)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = ((($this)) + 48|0); //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$3>>2]|0; //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ((($4)) + -44|0); //@line 1074 "modules/graphics/opengl/Graphics.cpp"
 $6 = ((($4)) + -40|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = HEAP32[$6>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = HEAP32[$5>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($7) - ($8))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ((($agg$result)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $12 = ($10|0)==(0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = $7; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = $8; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $20 = $14;$21 = $13;$42 = 0;$43 = 0;
 } else {
  $15 = (__Znwj($9)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $16 = $15; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $17 = (($15) + ($10<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $18 = $17; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  HEAP32[$agg$result>>2] = $16; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$1>>2] = $16; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$11>>2] = $18; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $$pre = HEAP32[$6>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$phi$trans$insert = ((($4)) + -44|0); //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre6 = HEAP32[$$phi$trans$insert>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $$pre6;$21 = $$pre;$42 = $17;$43 = $15;
 }
 $19 = ($20|0)==($21|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($19) {
  STACKTOP = sp;return; //@line 1080 "modules/graphics/opengl/Graphics.cpp"
 }
 $34 = $43;$35 = $42;$__begin$sroa$0$02 = $20;
 while(1) {
  $32 = HEAP32[$__begin$sroa$0$02>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$0>>2] = $32; //@line 1077 "modules/graphics/opengl/Graphics.cpp"
  $33 = ($34>>>0)<($35>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($33) {
   HEAP32[$34>>2] = $32; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $36 = HEAP32[$1>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $37 = ((($36)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$1>>2] = $37; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  } else {
   __THREW__ = 0;
   invoke_vii(3574,($agg$result|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $38 = __THREW__; __THREW__ = 0;
   $39 = $38&1;
   if ($39) {
    break;
   }
  }
  $40 = ((($__begin$sroa$0$02)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $41 = ($40|0)==($21|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($41) {
   label = 5;
   break;
  }
  $$pre3 = HEAP32[$1>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre4 = HEAP32[$11>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $34 = $$pre3;$35 = $$pre4;$__begin$sroa$0$02 = $40;
 }
 if ((label|0) == 5) {
  STACKTOP = sp;return; //@line 1080 "modules/graphics/opengl/Graphics.cpp"
 }
 $22 = ___cxa_find_matching_catch_2()|0;
 $23 = tempRet0;
 $$pre5 = HEAP32[$agg$result>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $24 = ($$pre5|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $25 = $$pre5; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($24) {
  ___resumeException($22|0);
  // unreachable;
 }
 $26 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = ($26|0)==($$pre5|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($27)) {
  $scevgep$i$i$i = ((($26)) + -4|0);
  $28 = $scevgep$i$i$i;
  $29 = (($28) - ($25))|0;
  $30 = $29 >>> 2;
  $31 = $30 ^ -1;
  $scevgep4$i$i$i = (($26) + ($31<<2)|0);
  HEAP32[$1>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($$pre5); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($22|0);
 // unreachable;
}
function __ZNK4love8graphics6opengl8Graphics12getColorMaskEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -32|0); //@line 1090 "modules/graphics/opengl/Graphics.cpp"
 $3 = HEAPU8[$2>>0]|(HEAPU8[$2+1>>0]<<8)|(HEAPU8[$2+2>>0]<<16)|(HEAPU8[$2+3>>0]<<24); //@line 1090 "modules/graphics/opengl/Graphics.cpp"
 HEAP8[$agg$result>>0]=$3&255;HEAP8[$agg$result+1>>0]=($3>>8)&255;HEAP8[$agg$result+2>>0]=($3>>16)&255;HEAP8[$agg$result+3>>0]=$3>>24; //@line 1090 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1090 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNK4love8graphics6opengl8Graphics11isSupportedENS0_8Graphics7SupportE($this,$feature) {
 $this = $this|0;
 $feature = $feature|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 switch ($feature|0) {
 case 0:  {
  $0 = (__ZN4love8graphics6opengl6Canvas33isMultiFormatMultiCanvasSupportedEv()|0); //@line 1682 "modules/graphics/opengl/Graphics.cpp"
  $$0 = $0;
  break;
 }
 case 1:  {
  $1 = (__ZNK4love8graphics6opengl6OpenGL31isClampZeroTextureWrapSupportedEv(634044)|0); //@line 1684 "modules/graphics/opengl/Graphics.cpp"
  $$0 = $1;
  break;
 }
 case 2:  {
  $2 = HEAP8[1251998]|0; //@line 1686 "modules/graphics/opengl/Graphics.cpp"
  $3 = HEAP8[1252013]|0; //@line 1686 "modules/graphics/opengl/Graphics.cpp"
  $4 = $3 | $2; //@line 1686 "modules/graphics/opengl/Graphics.cpp"
  $5 = HEAP8[1252166]|0; //@line 1686 "modules/graphics/opengl/Graphics.cpp"
  $6 = $4 | $5; //@line 1686 "modules/graphics/opengl/Graphics.cpp"
  $7 = ($6<<24>>24)!=(0); //@line 1686 "modules/graphics/opengl/Graphics.cpp"
  $$0 = $7;
  break;
 }
 default: {
  $$0 = 0;
 }
 }
 return ($$0|0); //@line 1690 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNK4love8graphics6opengl8Graphics12getBlendModeERNS0_8Graphics10BlendAlphaE($this,$alphamode) {
 $this = $this|0;
 $alphamode = $alphamode|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -100|0); //@line 1170 "modules/graphics/opengl/Graphics.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1170 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$alphamode>>2] = $3; //@line 1170 "modules/graphics/opengl/Graphics.cpp"
 $4 = ((($1)) + -104|0); //@line 1171 "modules/graphics/opengl/Graphics.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1171 "modules/graphics/opengl/Graphics.cpp"
 return ($5|0); //@line 1171 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNK4love8graphics6opengl8Graphics16getDefaultFilterEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love8graphics7Texture16getDefaultFilterEv()|0); //@line 1182 "modules/graphics/opengl/Graphics.cpp"
 return ($0|0); //@line 1182 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNK4love8graphics6opengl8Graphics22getDefaultMipmapFilterEPNS0_7Texture10FilterModeEPf($this,$filter,$sharpness) {
 $this = $this|0;
 $filter = $filter|0;
 $sharpness = $sharpness|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love8graphics6opengl5Image22getDefaultMipmapFilterEv()|0); //@line 1196 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$filter>>2] = $0; //@line 1196 "modules/graphics/opengl/Graphics.cpp"
 $1 = (+__ZN4love8graphics6opengl5Image25getDefaultMipmapSharpnessEv()); //@line 1197 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$sharpness>>2] = $1; //@line 1197 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1198 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNK4love8graphics6opengl8Graphics12getLineWidthEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -96|0); //@line 1217 "modules/graphics/opengl/Graphics.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 1217 "modules/graphics/opengl/Graphics.cpp"
 return (+$3); //@line 1217 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNK4love8graphics6opengl8Graphics12getLineStyleEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -92|0); //@line 1222 "modules/graphics/opengl/Graphics.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1222 "modules/graphics/opengl/Graphics.cpp"
 return ($3|0); //@line 1222 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNK4love8graphics6opengl8Graphics11getLineJoinEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -88|0); //@line 1227 "modules/graphics/opengl/Graphics.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1227 "modules/graphics/opengl/Graphics.cpp"
 return ($3|0); //@line 1227 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNK4love8graphics6opengl8Graphics12getPointSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -84|0); //@line 1238 "modules/graphics/opengl/Graphics.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 1238 "modules/graphics/opengl/Graphics.cpp"
 return (+$3); //@line 1238 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNK4love8graphics6opengl8Graphics11isWireframeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 687 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -28|0); //@line 1253 "modules/graphics/opengl/Graphics.cpp"
 $3 = HEAP8[$2>>0]|0; //@line 1253 "modules/graphics/opengl/Graphics.cpp"
 $4 = ($3<<24>>24)!=(0); //@line 1253 "modules/graphics/opengl/Graphics.cpp"
 return ($4|0); //@line 1253 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics5printERKNSt3__16vectorINS1_4Font13ColoredStringENS3_9allocatorIS6_EEEEfffffffff($this,$str,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $str = $str|0;
 $x = +$x;
 $y = +$y;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl8Graphics19checkSetDefaultFontEv($this); //@line 1258 "modules/graphics/opengl/Graphics.cpp"
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -52|0); //@line 139 "./common/Object.h"
 $3 = HEAP32[$2>>2]|0; //@line 139 "./common/Object.h"
 $4 = ($3|0)==(0|0); //@line 1262 "modules/graphics/opengl/Graphics.cpp"
 if ($4) {
  return; //@line 1264 "modules/graphics/opengl/Graphics.cpp"
 }
 __ZN4love8graphics6opengl4Font5printERKNSt3__16vectorINS2_13ColoredStringENS3_9allocatorIS5_EEEEfffffffff($3,$str,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky); //@line 1263 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1264 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics6printfERKNSt3__16vectorINS1_4Font13ColoredStringENS3_9allocatorIS6_EEEEfffNS5_9AlignModeEfffffff($this,$str,$x,$y,$wrap,$align,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $str = $str|0;
 $x = +$x;
 $y = +$y;
 $wrap = +$wrap;
 $align = $align|0;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl8Graphics19checkSetDefaultFontEv($this); //@line 1268 "modules/graphics/opengl/Graphics.cpp"
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -52|0); //@line 139 "./common/Object.h"
 $3 = HEAP32[$2>>2]|0; //@line 139 "./common/Object.h"
 $4 = ($3|0)==(0|0); //@line 1272 "modules/graphics/opengl/Graphics.cpp"
 if ($4) {
  return; //@line 1274 "modules/graphics/opengl/Graphics.cpp"
 }
 __ZN4love8graphics6opengl4Font6printfERKNSt3__16vectorINS2_13ColoredStringENS3_9allocatorIS5_EEEEfffNS2_9AlignModeEfffffff($3,$str,$x,$y,$wrap,$align,$angle,$sx,$sy,$ox,$oy,$kx,$ky); //@line 1273 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1274 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics6pointsEPKfPKhj($this,$coords,$colors,$numpoints) {
 $this = $this|0;
 $coords = $coords|0;
 $colors = $colors|0;
 $numpoints = $numpoints|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $attribflags$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl6OpenGL11prepareDrawEv(634044); //@line 1284 "modules/graphics/opengl/Graphics.cpp"
 $0 = (__ZNK4love8graphics6opengl6OpenGL17getDefaultTextureEv(634044)|0); //@line 1285 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$0); //@line 1285 "modules/graphics/opengl/Graphics.cpp"
 $1 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiiiii[$1 & 8191](0,2,5126,0,0,$coords); //@line 591 "./libraries/glad/gladfuncs.hpp"
 $2 = ($colors|0)==(0|0); //@line 1290 "modules/graphics/opengl/Graphics.cpp"
 if ($2) {
  $attribflags$0 = 1;
 } else {
  $3 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiiiii[$3 & 8191](2,4,5121,1,0,$colors); //@line 591 "./libraries/glad/gladfuncs.hpp"
  $attribflags$0 = 5;
 }
 __ZN4love8graphics6opengl6OpenGL21useVertexAttribArraysEj(634044,$attribflags$0); //@line 1296 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl6OpenGL10drawArraysEjii(634044,0,0,$numpoints); //@line 1297 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1298 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics8polylineEPKfj($this,$coords,$count) {
 $this = $this|0;
 $coords = $coords|0;
 $count = $count|0;
 var $$0 = 0, $$01 = 0, $$lcssa1$i = 0, $$lcssa2$i = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0.0, $59 = 0.0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0.0, $64 = 0.0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0, $85 = 0, $86 = 0.0, $87 = 0.0, $88 = 0, $89 = 0, $9 = 0.0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $i$03$i = 0, $line = 0, $line1 = 0, $line2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $line = sp + 64|0;
 $line1 = sp + 32|0;
 $line2 = sp;
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -88|0); //@line 1304 "modules/graphics/opengl/Graphics.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1304 "modules/graphics/opengl/Graphics.cpp"
 switch ($3|0) {
 case 0:  {
  $4 = ((($line)) + 4|0); //@line 50 "modules/graphics/opengl/Polyline.h"
  $5 = ((($line)) + 20|0); //@line 54 "modules/graphics/opengl/Polyline.h"
  ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;HEAP32[$4+12>>2]=0|0; //@line 51 "modules/graphics/opengl/Polyline.h"
  HEAP32[$5>>2] = 4; //@line 54 "modules/graphics/opengl/Polyline.h"
  $6 = ((($line)) + 24|0); //@line 55 "modules/graphics/opengl/Polyline.h"
  HEAP8[$6>>0] = 1; //@line 55 "modules/graphics/opengl/Polyline.h"
  $7 = ((($line)) + 28|0); //@line 56 "modules/graphics/opengl/Polyline.h"
  HEAP32[$7>>2] = 0; //@line 56 "modules/graphics/opengl/Polyline.h"
  HEAP32[$line>>2] = (25116); //@line 114 "modules/graphics/opengl/Polyline.h"
  $8 = ((($1)) + -96|0); //@line 1307 "modules/graphics/opengl/Graphics.cpp"
  $9 = +HEAPF32[$8>>2]; //@line 1307 "modules/graphics/opengl/Graphics.cpp"
  $10 = $9 * 0.5; //@line 1307 "modules/graphics/opengl/Graphics.cpp"
  $11 = ((($this)) + 20|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $12 = HEAP32[$11>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = ((($12)) + -8|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = +HEAPF64[$13>>3]; //@line 1307 "modules/graphics/opengl/Graphics.cpp"
  $15 = $14; //@line 1307 "modules/graphics/opengl/Graphics.cpp"
  $16 = ((($1)) + -92|0); //@line 1307 "modules/graphics/opengl/Graphics.cpp"
  $17 = HEAP32[$16>>2]|0; //@line 1307 "modules/graphics/opengl/Graphics.cpp"
  $18 = ($17|0)==(1); //@line 1307 "modules/graphics/opengl/Graphics.cpp"
  $19 = $count << 1; //@line 118 "modules/graphics/opengl/Polyline.h"
  $20 = (($19) + -4)|0; //@line 118 "modules/graphics/opengl/Polyline.h"
  __THREW__ = 0;
  invoke_viiiiddi(4075,($line|0),($coords|0),($count|0),($20|0),(+$10),(+$15),($18|0)); //@line 118 "modules/graphics/opengl/Polyline.h"
  $21 = __THREW__; __THREW__ = 0;
  $22 = $21&1;
  if (!($22)) {
   $23 = ((($line)) + 12|0); //@line 121 "modules/graphics/opengl/Polyline.h"
   $24 = HEAP32[$23>>2]|0; //@line 121 "modules/graphics/opengl/Polyline.h"
   $25 = ($24|0)==(4); //@line 121 "modules/graphics/opengl/Polyline.h"
   $26 = HEAP32[$4>>2]|0; //@line 122 "modules/graphics/opengl/Polyline.h"
   if ($25) {
    $$lcssa1$i = 0;$$lcssa2$i = $26;
   } else {
    $28 = $26;$i$03$i = 0;
    while(1) {
     $27 = (($28) + ($i$03$i<<3)|0); //@line 122 "modules/graphics/opengl/Polyline.h"
     $29 = (($i$03$i) + 2)|0; //@line 122 "modules/graphics/opengl/Polyline.h"
     $30 = (($28) + ($29<<3)|0); //@line 122 "modules/graphics/opengl/Polyline.h"
     $31 = $30; //@line 122 "modules/graphics/opengl/Polyline.h"
     $32 = $31; //@line 122 "modules/graphics/opengl/Polyline.h"
     $33 = HEAP32[$32>>2]|0; //@line 122 "modules/graphics/opengl/Polyline.h"
     $34 = (($31) + 4)|0; //@line 122 "modules/graphics/opengl/Polyline.h"
     $35 = $34; //@line 122 "modules/graphics/opengl/Polyline.h"
     $36 = HEAP32[$35>>2]|0; //@line 122 "modules/graphics/opengl/Polyline.h"
     $37 = $27; //@line 122 "modules/graphics/opengl/Polyline.h"
     $38 = $37; //@line 122 "modules/graphics/opengl/Polyline.h"
     HEAP32[$38>>2] = $33; //@line 122 "modules/graphics/opengl/Polyline.h"
     $39 = (($37) + 4)|0; //@line 122 "modules/graphics/opengl/Polyline.h"
     $40 = $39; //@line 122 "modules/graphics/opengl/Polyline.h"
     HEAP32[$40>>2] = $36; //@line 122 "modules/graphics/opengl/Polyline.h"
     $41 = (($i$03$i) + 1)|0; //@line 121 "modules/graphics/opengl/Polyline.h"
     $42 = HEAP32[$23>>2]|0; //@line 121 "modules/graphics/opengl/Polyline.h"
     $43 = (($42) + -4)|0; //@line 121 "modules/graphics/opengl/Polyline.h"
     $44 = ($41>>>0)<($43>>>0); //@line 121 "modules/graphics/opengl/Polyline.h"
     $45 = HEAP32[$4>>2]|0; //@line 122 "modules/graphics/opengl/Polyline.h"
     if ($44) {
      $28 = $45;$i$03$i = $41;
     } else {
      $$lcssa1$i = $43;$$lcssa2$i = $45;
      break;
     }
    }
   }
   $46 = (($$lcssa2$i) + ($$lcssa1$i<<3)|0); //@line 128 "modules/graphics/opengl/Polyline.h"
   ;HEAP32[$46>>2]=0|0;HEAP32[$46+4>>2]=0|0;HEAP32[$46+8>>2]=0|0;HEAP32[$46+12>>2]=0|0;HEAP32[$46+16>>2]=0|0;HEAP32[$46+20>>2]=0|0;HEAP32[$46+24>>2]=0|0;HEAP32[$46+28>>2]=0|0; //@line 128 "modules/graphics/opengl/Polyline.h"
   $47 = HEAP32[$23>>2]|0; //@line 130 "modules/graphics/opengl/Polyline.h"
   $48 = (($47) + -4)|0; //@line 130 "modules/graphics/opengl/Polyline.h"
   HEAP32[$23>>2] = $48; //@line 130 "modules/graphics/opengl/Polyline.h"
   __THREW__ = 0;
   invoke_vi(4076,($line|0)); //@line 1308 "modules/graphics/opengl/Graphics.cpp"
   $49 = __THREW__; __THREW__ = 0;
   $50 = $49&1;
   if (!($50)) {
    __ZN4love8graphics6opengl8PolylineD2Ev($line); //@line 1309 "modules/graphics/opengl/Graphics.cpp"
    STACKTOP = sp;return; //@line 1322 "modules/graphics/opengl/Graphics.cpp"
   }
  }
  $51 = ___cxa_find_matching_catch_2()|0;
  $52 = tempRet0;
  __ZN4love8graphics6opengl8PolylineD2Ev($line); //@line 1309 "modules/graphics/opengl/Graphics.cpp"
  $$0 = $52;$$01 = $51;
  ___resumeException($$01|0);
  // unreachable;
  break;
 }
 case 2:  {
  $53 = ((($line1)) + 4|0); //@line 50 "modules/graphics/opengl/Polyline.h"
  $54 = ((($line1)) + 20|0); //@line 54 "modules/graphics/opengl/Polyline.h"
  ;HEAP32[$53>>2]=0|0;HEAP32[$53+4>>2]=0|0;HEAP32[$53+8>>2]=0|0;HEAP32[$53+12>>2]=0|0; //@line 51 "modules/graphics/opengl/Polyline.h"
  HEAP32[$54>>2] = 5; //@line 54 "modules/graphics/opengl/Polyline.h"
  $55 = ((($line1)) + 24|0); //@line 55 "modules/graphics/opengl/Polyline.h"
  HEAP8[$55>>0] = 0; //@line 55 "modules/graphics/opengl/Polyline.h"
  $56 = ((($line1)) + 28|0); //@line 56 "modules/graphics/opengl/Polyline.h"
  HEAP32[$56>>2] = 0; //@line 56 "modules/graphics/opengl/Polyline.h"
  HEAP32[$line1>>2] = (25180); //@line 166 "modules/graphics/opengl/Polyline.h"
  $57 = ((($1)) + -96|0); //@line 1313 "modules/graphics/opengl/Graphics.cpp"
  $58 = +HEAPF32[$57>>2]; //@line 1313 "modules/graphics/opengl/Graphics.cpp"
  $59 = $58 * 0.5; //@line 1313 "modules/graphics/opengl/Graphics.cpp"
  $60 = ((($this)) + 20|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $61 = HEAP32[$60>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $62 = ((($61)) + -8|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $63 = +HEAPF64[$62>>3]; //@line 1313 "modules/graphics/opengl/Graphics.cpp"
  $64 = $63; //@line 1313 "modules/graphics/opengl/Graphics.cpp"
  $65 = ((($1)) + -92|0); //@line 1313 "modules/graphics/opengl/Graphics.cpp"
  $66 = HEAP32[$65>>2]|0; //@line 1313 "modules/graphics/opengl/Graphics.cpp"
  $67 = ($66|0)==(1); //@line 1313 "modules/graphics/opengl/Graphics.cpp"
  $68 = $count << 1; //@line 171 "modules/graphics/opengl/Polyline.h"
  $69 = (($68) + -4)|0; //@line 171 "modules/graphics/opengl/Polyline.h"
  __THREW__ = 0;
  invoke_viiiiddi(4075,($line1|0),($coords|0),($count|0),($69|0),(+$59),(+$64),($67|0)); //@line 171 "modules/graphics/opengl/Polyline.h"
  $70 = __THREW__; __THREW__ = 0;
  $71 = $70&1;
  if (!($71)) {
   __THREW__ = 0;
   invoke_vi(4076,($line1|0)); //@line 1314 "modules/graphics/opengl/Graphics.cpp"
   $72 = __THREW__; __THREW__ = 0;
   $73 = $72&1;
   if (!($73)) {
    __ZN4love8graphics6opengl8PolylineD2Ev($line1); //@line 1315 "modules/graphics/opengl/Graphics.cpp"
    STACKTOP = sp;return; //@line 1322 "modules/graphics/opengl/Graphics.cpp"
   }
  }
  $74 = ___cxa_find_matching_catch_2()|0;
  $75 = tempRet0;
  __ZN4love8graphics6opengl8PolylineD2Ev($line1); //@line 1315 "modules/graphics/opengl/Graphics.cpp"
  $$0 = $75;$$01 = $74;
  ___resumeException($$01|0);
  // unreachable;
  break;
 }
 default: {
  $76 = ((($line2)) + 4|0); //@line 50 "modules/graphics/opengl/Polyline.h"
  $77 = ((($line2)) + 20|0); //@line 54 "modules/graphics/opengl/Polyline.h"
  ;HEAP32[$76>>2]=0|0;HEAP32[$76+4>>2]=0|0;HEAP32[$76+8>>2]=0|0;HEAP32[$76+12>>2]=0|0; //@line 51 "modules/graphics/opengl/Polyline.h"
  HEAP32[$77>>2] = 5; //@line 54 "modules/graphics/opengl/Polyline.h"
  $78 = ((($line2)) + 24|0); //@line 55 "modules/graphics/opengl/Polyline.h"
  HEAP8[$78>>0] = 0; //@line 55 "modules/graphics/opengl/Polyline.h"
  $79 = ((($line2)) + 28|0); //@line 56 "modules/graphics/opengl/Polyline.h"
  HEAP32[$79>>2] = 0; //@line 56 "modules/graphics/opengl/Polyline.h"
  HEAP32[$line2>>2] = (25148); //@line 147 "modules/graphics/opengl/Polyline.h"
  $80 = ((($1)) + -96|0); //@line 1319 "modules/graphics/opengl/Graphics.cpp"
  $81 = +HEAPF32[$80>>2]; //@line 1319 "modules/graphics/opengl/Graphics.cpp"
  $82 = $81 * 0.5; //@line 1319 "modules/graphics/opengl/Graphics.cpp"
  $83 = ((($this)) + 20|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $84 = HEAP32[$83>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $85 = ((($84)) + -8|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $86 = +HEAPF64[$85>>3]; //@line 1319 "modules/graphics/opengl/Graphics.cpp"
  $87 = $86; //@line 1319 "modules/graphics/opengl/Graphics.cpp"
  $88 = ((($1)) + -92|0); //@line 1319 "modules/graphics/opengl/Graphics.cpp"
  $89 = HEAP32[$88>>2]|0; //@line 1319 "modules/graphics/opengl/Graphics.cpp"
  $90 = ($89|0)==(1); //@line 1319 "modules/graphics/opengl/Graphics.cpp"
  __THREW__ = 0;
  invoke_viiiiddi(4075,($line2|0),($coords|0),($count|0),($count|0),(+$82),(+$87),($90|0)); //@line 152 "modules/graphics/opengl/Polyline.h"
  $91 = __THREW__; __THREW__ = 0;
  $92 = $91&1;
  if (!($92)) {
   __THREW__ = 0;
   invoke_vi(4076,($line2|0)); //@line 1320 "modules/graphics/opengl/Graphics.cpp"
   $93 = __THREW__; __THREW__ = 0;
   $94 = $93&1;
   if (!($94)) {
    __ZN4love8graphics6opengl8PolylineD2Ev($line2); //@line 1321 "modules/graphics/opengl/Graphics.cpp"
    STACKTOP = sp;return; //@line 1322 "modules/graphics/opengl/Graphics.cpp"
   }
  }
  $95 = ___cxa_find_matching_catch_2()|0;
  $96 = tempRet0;
  __ZN4love8graphics6opengl8PolylineD2Ev($line2); //@line 1321 "modules/graphics/opengl/Graphics.cpp"
  $$0 = $96;$$01 = $95;
  ___resumeException($$01|0);
  // unreachable;
 }
 }
}
function __ZN4love8graphics6opengl8Graphics9rectangleENS0_8Graphics8DrawModeEffff($this,$mode,$x,$y,$w,$h) {
 $this = $this|0;
 $mode = $mode|0;
 $x = +$x;
 $y = +$y;
 $w = +$w;
 $h = +$h;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $coords = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $coords = sp;
 HEAPF32[$coords>>2] = $x; //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 $0 = ((($coords)) + 4|0); //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$0>>2] = $y; //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 $1 = ((($coords)) + 8|0); //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$1>>2] = $x; //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 $2 = ((($coords)) + 12|0); //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 $3 = $y + $h; //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$2>>2] = $3; //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 $4 = ((($coords)) + 16|0); //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 $5 = $x + $w; //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$4>>2] = $5; //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 $6 = ((($coords)) + 20|0); //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$6>>2] = $3; //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 $7 = ((($coords)) + 24|0); //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$7>>2] = $5; //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 $8 = ((($coords)) + 28|0); //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$8>>2] = $y; //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 $9 = ((($coords)) + 32|0); //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$9>>2] = $x; //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 $10 = ((($coords)) + 36|0); //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 HEAPF32[$10>>2] = $y; //@line 1326 "modules/graphics/opengl/Graphics.cpp"
 $11 = ($mode|0)==(0); //@line 1504 "modules/graphics/opengl/Graphics.cpp"
 if ($11) {
  __ZN4love8graphics6opengl8Graphics8polylineEPKfj($this,$coords,10); //@line 1506 "modules/graphics/opengl/Graphics.cpp"
  STACKTOP = sp;return; //@line 1328 "modules/graphics/opengl/Graphics.cpp"
 } else {
  __ZN4love8graphics6opengl6OpenGL11prepareDrawEv(634044); //@line 1512 "modules/graphics/opengl/Graphics.cpp"
  $12 = (__ZNK4love8graphics6opengl6OpenGL17getDefaultTextureEv(634044)|0); //@line 1513 "modules/graphics/opengl/Graphics.cpp"
  __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$12); //@line 1513 "modules/graphics/opengl/Graphics.cpp"
  __ZN4love8graphics6opengl6OpenGL21useVertexAttribArraysEj(634044,1); //@line 1514 "modules/graphics/opengl/Graphics.cpp"
  $13 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiiiii[$13 & 8191](0,2,5126,0,0,$coords); //@line 591 "./libraries/glad/gladfuncs.hpp"
  __ZN4love8graphics6opengl6OpenGL10drawArraysEjii(634044,6,0,4); //@line 1516 "modules/graphics/opengl/Graphics.cpp"
  STACKTOP = sp;return; //@line 1328 "modules/graphics/opengl/Graphics.cpp"
 }
}
function __ZN4love8graphics6opengl8Graphics7polygonENS0_8Graphics8DrawModeEPKfj($this,$mode,$coords,$count) {
 $this = $this|0;
 $mode = $mode|0;
 $coords = $coords|0;
 $count = $count|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mode|0)==(0); //@line 1504 "modules/graphics/opengl/Graphics.cpp"
 if ($0) {
  __ZN4love8graphics6opengl8Graphics8polylineEPKfj($this,$coords,$count); //@line 1506 "modules/graphics/opengl/Graphics.cpp"
  return; //@line 1518 "modules/graphics/opengl/Graphics.cpp"
 } else {
  __ZN4love8graphics6opengl6OpenGL11prepareDrawEv(634044); //@line 1512 "modules/graphics/opengl/Graphics.cpp"
  $1 = (__ZNK4love8graphics6opengl6OpenGL17getDefaultTextureEv(634044)|0); //@line 1513 "modules/graphics/opengl/Graphics.cpp"
  __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$1); //@line 1513 "modules/graphics/opengl/Graphics.cpp"
  __ZN4love8graphics6opengl6OpenGL21useVertexAttribArraysEj(634044,1); //@line 1514 "modules/graphics/opengl/Graphics.cpp"
  $2 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiiiii[$2 & 8191](0,2,5126,0,0,$coords); //@line 591 "./libraries/glad/gladfuncs.hpp"
  $3 = (($count|0) / 2)&-1; //@line 1516 "modules/graphics/opengl/Graphics.cpp"
  $4 = (($3) + -1)|0; //@line 1516 "modules/graphics/opengl/Graphics.cpp"
  __ZN4love8graphics6opengl6OpenGL10drawArraysEjii(634044,6,0,$4); //@line 1516 "modules/graphics/opengl/Graphics.cpp"
  return; //@line 1518 "modules/graphics/opengl/Graphics.cpp"
 }
}
function __ZN4love8graphics6opengl8Graphics9rectangleENS0_8Graphics8DrawModeEffffffi($this,$mode,$x,$y,$w,$h,$rx,$ry,$points) {
 $this = $this|0;
 $mode = $mode|0;
 $x = +$x;
 $y = +$y;
 $w = +$w;
 $h = +$h;
 $rx = +$rx;
 $ry = +$ry;
 $points = $points|0;
 var $$arith = 0, $$overflow = 0, $$sroa$0 = 0, $$sroa$02 = 0, $0 = 0, $1 = 0, $10 = 0.0, $100 = 0, $101 = 0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0, $107 = 0, $108 = 0, $109 = 0.0, $11 = 0, $110 = 0, $12 = 0.0;
 var $13 = 0.0, $14 = 0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0;
 var $5 = 0.0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0.0, $60 = 0, $61 = 0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0.0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $i$017 = 0, $i1$013 = 0, $i2$010 = 0, $i3$08 = 0, $or$cond = 0;
 var $phi$016 = 0.0, $phi$112 = 0.0, $phi$29 = 0.0, $phi$37 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$sroa$02 = sp + 12|0;
 $$sroa$0 = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 HEAPF32[$$sroa$02>>2] = $rx;
 HEAPF32[$$sroa$0>>2] = $ry;
 $2 = $rx == 0.0; //@line 1332 "modules/graphics/opengl/Graphics.cpp"
 $3 = $ry == 0.0; //@line 1332 "modules/graphics/opengl/Graphics.cpp"
 $or$cond = $2 | $3; //@line 1332 "modules/graphics/opengl/Graphics.cpp"
 if ($or$cond) {
  __ZN4love8graphics6opengl8Graphics9rectangleENS0_8Graphics8DrawModeEffff($this,$mode,$x,$y,$w,$h); //@line 1334 "modules/graphics/opengl/Graphics.cpp"
  STACKTOP = sp;return; //@line 1390 "modules/graphics/opengl/Graphics.cpp"
 }
 $4 = !($w >= 0.019999999552965164); //@line 1340 "modules/graphics/opengl/Graphics.cpp"
 if ($4) {
  $37 = $rx;
 } else {
  $5 = $w * 0.5; //@line 1341 "modules/graphics/opengl/Graphics.cpp"
  $6 = $5 + -0.0099999997764825821; //@line 1341 "modules/graphics/opengl/Graphics.cpp"
  HEAPF32[$0>>2] = $6; //@line 1341 "modules/graphics/opengl/Graphics.cpp"
  $7 = $6 < $rx; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $8 = $7 ? $0 : $$sroa$02; //@line 2575 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $9 = HEAP32[$8>>2]|0; //@line 1341 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$$sroa$02>>2] = $9; //@line 1341 "modules/graphics/opengl/Graphics.cpp"
  $10 = (HEAP32[tempDoublePtr>>2]=$9,+HEAPF32[tempDoublePtr>>2]); //@line 1341 "modules/graphics/opengl/Graphics.cpp"
  $37 = $10;
 }
 $11 = !($h >= 0.019999999552965164); //@line 1342 "modules/graphics/opengl/Graphics.cpp"
 if ($11) {
  $44 = $ry;
 } else {
  $12 = $h * 0.5; //@line 1343 "modules/graphics/opengl/Graphics.cpp"
  $13 = $12 + -0.0099999997764825821; //@line 1343 "modules/graphics/opengl/Graphics.cpp"
  HEAPF32[$1>>2] = $13; //@line 1343 "modules/graphics/opengl/Graphics.cpp"
  $14 = $13 < $ry; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $15 = $14 ? $1 : $$sroa$0; //@line 2575 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $16 = HEAP32[$15>>2]|0; //@line 1343 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$$sroa$0>>2] = $16; //@line 1343 "modules/graphics/opengl/Graphics.cpp"
  $17 = (HEAP32[tempDoublePtr>>2]=$16,+HEAPF32[tempDoublePtr>>2]); //@line 1343 "modules/graphics/opengl/Graphics.cpp"
  $44 = $17;
 }
 $18 = ($points|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $19 = $18 ? 1 : $points; //@line 1345 "modules/graphics/opengl/Graphics.cpp"
 $20 = (+($19|0)); //@line 1348 "modules/graphics/opengl/Graphics.cpp"
 $21 = $20 + 1.0; //@line 1348 "modules/graphics/opengl/Graphics.cpp"
 $22 = 1.5707963705062866 / $21; //@line 1348 "modules/graphics/opengl/Graphics.cpp"
 $23 = $19 << 3; //@line 1350 "modules/graphics/opengl/Graphics.cpp"
 $24 = (($23) + 16)|0; //@line 1350 "modules/graphics/opengl/Graphics.cpp"
 $25 = (($23) + 18)|0; //@line 1351 "modules/graphics/opengl/Graphics.cpp"
 $$arith = $25<<2; //@line 1351 "modules/graphics/opengl/Graphics.cpp"
 $$overflow = ($25>>>0)>(1073741823); //@line 1351 "modules/graphics/opengl/Graphics.cpp"
 $26 = $$overflow ? -1 : $$arith; //@line 1351 "modules/graphics/opengl/Graphics.cpp"
 $27 = (__Znaj($26)|0); //@line 1351 "modules/graphics/opengl/Graphics.cpp"
 $28 = (($19) + 2)|0; //@line 1354 "modules/graphics/opengl/Graphics.cpp"
 $29 = ($19|0)<(-2); //@line 1354 "modules/graphics/opengl/Graphics.cpp"
 if (!($29)) {
  $i$017 = 0;$phi$016 = 0.0;
  while(1) {
   $34 = (+Math_cos((+$phi$016))); //@line 1356 "modules/graphics/opengl/Graphics.cpp"
   $35 = 1.0 - $34; //@line 1356 "modules/graphics/opengl/Graphics.cpp"
   $36 = $37 * $35; //@line 1356 "modules/graphics/opengl/Graphics.cpp"
   $38 = $36 + $x; //@line 1356 "modules/graphics/opengl/Graphics.cpp"
   $39 = $i$017 << 1; //@line 1356 "modules/graphics/opengl/Graphics.cpp"
   $40 = (($27) + ($39<<2)|0); //@line 1356 "modules/graphics/opengl/Graphics.cpp"
   HEAPF32[$40>>2] = $38; //@line 1356 "modules/graphics/opengl/Graphics.cpp"
   $41 = (+Math_sin((+$phi$016))); //@line 1357 "modules/graphics/opengl/Graphics.cpp"
   $42 = 1.0 - $41; //@line 1357 "modules/graphics/opengl/Graphics.cpp"
   $43 = $44 * $42; //@line 1357 "modules/graphics/opengl/Graphics.cpp"
   $45 = $43 + $y; //@line 1357 "modules/graphics/opengl/Graphics.cpp"
   $46 = $39 | 1; //@line 1357 "modules/graphics/opengl/Graphics.cpp"
   $47 = (($27) + ($46<<2)|0); //@line 1357 "modules/graphics/opengl/Graphics.cpp"
   HEAPF32[$47>>2] = $45; //@line 1357 "modules/graphics/opengl/Graphics.cpp"
   $48 = (($i$017) + 1)|0; //@line 1354 "modules/graphics/opengl/Graphics.cpp"
   $49 = $22 + $phi$016; //@line 1354 "modules/graphics/opengl/Graphics.cpp"
   $50 = ($i$017|0)<($28|0); //@line 1354 "modules/graphics/opengl/Graphics.cpp"
   if ($50) {
    $i$017 = $48;$phi$016 = $49;
   } else {
    break;
   }
  }
 }
 $30 = $19 << 1; //@line 1362 "modules/graphics/opengl/Graphics.cpp"
 $31 = (($30) + 4)|0; //@line 1362 "modules/graphics/opengl/Graphics.cpp"
 $32 = ($28|0)>($31|0); //@line 1362 "modules/graphics/opengl/Graphics.cpp"
 if (!($32)) {
  $33 = $x + $w; //@line 1364 "modules/graphics/opengl/Graphics.cpp"
  $i1$013 = $28;$phi$112 = 1.5707963705062866;
  while(1) {
   $56 = (+Math_cos((+$phi$112))); //@line 1364 "modules/graphics/opengl/Graphics.cpp"
   $57 = $56 + 1.0; //@line 1364 "modules/graphics/opengl/Graphics.cpp"
   $58 = $37 * $57; //@line 1364 "modules/graphics/opengl/Graphics.cpp"
   $59 = $33 - $58; //@line 1364 "modules/graphics/opengl/Graphics.cpp"
   $60 = $i1$013 << 1; //@line 1364 "modules/graphics/opengl/Graphics.cpp"
   $61 = (($27) + ($60<<2)|0); //@line 1364 "modules/graphics/opengl/Graphics.cpp"
   HEAPF32[$61>>2] = $59; //@line 1364 "modules/graphics/opengl/Graphics.cpp"
   $62 = (+Math_sin((+$phi$112))); //@line 1365 "modules/graphics/opengl/Graphics.cpp"
   $63 = 1.0 - $62; //@line 1365 "modules/graphics/opengl/Graphics.cpp"
   $64 = $44 * $63; //@line 1365 "modules/graphics/opengl/Graphics.cpp"
   $65 = $64 + $y; //@line 1365 "modules/graphics/opengl/Graphics.cpp"
   $66 = $60 | 1; //@line 1365 "modules/graphics/opengl/Graphics.cpp"
   $67 = (($27) + ($66<<2)|0); //@line 1365 "modules/graphics/opengl/Graphics.cpp"
   HEAPF32[$67>>2] = $65; //@line 1365 "modules/graphics/opengl/Graphics.cpp"
   $68 = (($i1$013) + 1)|0; //@line 1362 "modules/graphics/opengl/Graphics.cpp"
   $69 = $22 + $phi$112; //@line 1362 "modules/graphics/opengl/Graphics.cpp"
   $70 = ($i1$013|0)<($31|0); //@line 1362 "modules/graphics/opengl/Graphics.cpp"
   if ($70) {
    $i1$013 = $68;$phi$112 = $69;
   } else {
    break;
   }
  }
 }
 $51 = ($19*3)|0; //@line 1370 "modules/graphics/opengl/Graphics.cpp"
 $52 = (($51) + 6)|0; //@line 1370 "modules/graphics/opengl/Graphics.cpp"
 $53 = ($31|0)>($52|0); //@line 1370 "modules/graphics/opengl/Graphics.cpp"
 if (!($53)) {
  $54 = $x + $w; //@line 1372 "modules/graphics/opengl/Graphics.cpp"
  $55 = $y + $h; //@line 1373 "modules/graphics/opengl/Graphics.cpp"
  $i2$010 = $31;$phi$29 = 3.1415927410125732;
  while(1) {
   $75 = (+Math_cos((+$phi$29))); //@line 1372 "modules/graphics/opengl/Graphics.cpp"
   $76 = $75 + 1.0; //@line 1372 "modules/graphics/opengl/Graphics.cpp"
   $77 = $37 * $76; //@line 1372 "modules/graphics/opengl/Graphics.cpp"
   $78 = $54 - $77; //@line 1372 "modules/graphics/opengl/Graphics.cpp"
   $79 = $i2$010 << 1; //@line 1372 "modules/graphics/opengl/Graphics.cpp"
   $80 = (($27) + ($79<<2)|0); //@line 1372 "modules/graphics/opengl/Graphics.cpp"
   HEAPF32[$80>>2] = $78; //@line 1372 "modules/graphics/opengl/Graphics.cpp"
   $81 = (+Math_sin((+$phi$29))); //@line 1373 "modules/graphics/opengl/Graphics.cpp"
   $82 = $81 + 1.0; //@line 1373 "modules/graphics/opengl/Graphics.cpp"
   $83 = $44 * $82; //@line 1373 "modules/graphics/opengl/Graphics.cpp"
   $84 = $55 - $83; //@line 1373 "modules/graphics/opengl/Graphics.cpp"
   $85 = $79 | 1; //@line 1373 "modules/graphics/opengl/Graphics.cpp"
   $86 = (($27) + ($85<<2)|0); //@line 1373 "modules/graphics/opengl/Graphics.cpp"
   HEAPF32[$86>>2] = $84; //@line 1373 "modules/graphics/opengl/Graphics.cpp"
   $87 = (($i2$010) + 1)|0; //@line 1370 "modules/graphics/opengl/Graphics.cpp"
   $88 = $22 + $phi$29; //@line 1370 "modules/graphics/opengl/Graphics.cpp"
   $89 = ($i2$010|0)<($52|0); //@line 1370 "modules/graphics/opengl/Graphics.cpp"
   if ($89) {
    $i2$010 = $87;$phi$29 = $88;
   } else {
    break;
   }
  }
 }
 $71 = $19 << 2; //@line 1378 "modules/graphics/opengl/Graphics.cpp"
 $72 = (($71) + 8)|0; //@line 1378 "modules/graphics/opengl/Graphics.cpp"
 $73 = ($52|0)>($72|0); //@line 1378 "modules/graphics/opengl/Graphics.cpp"
 if (!($73)) {
  $74 = $y + $h; //@line 1381 "modules/graphics/opengl/Graphics.cpp"
  $i3$08 = $52;$phi$37 = 4.7123889923095703;
  while(1) {
   $96 = (+Math_cos((+$phi$37))); //@line 1380 "modules/graphics/opengl/Graphics.cpp"
   $97 = 1.0 - $96; //@line 1380 "modules/graphics/opengl/Graphics.cpp"
   $98 = $37 * $97; //@line 1380 "modules/graphics/opengl/Graphics.cpp"
   $99 = $98 + $x; //@line 1380 "modules/graphics/opengl/Graphics.cpp"
   $100 = $i3$08 << 1; //@line 1380 "modules/graphics/opengl/Graphics.cpp"
   $101 = (($27) + ($100<<2)|0); //@line 1380 "modules/graphics/opengl/Graphics.cpp"
   HEAPF32[$101>>2] = $99; //@line 1380 "modules/graphics/opengl/Graphics.cpp"
   $102 = (+Math_sin((+$phi$37))); //@line 1381 "modules/graphics/opengl/Graphics.cpp"
   $103 = $102 + 1.0; //@line 1381 "modules/graphics/opengl/Graphics.cpp"
   $104 = $44 * $103; //@line 1381 "modules/graphics/opengl/Graphics.cpp"
   $105 = $74 - $104; //@line 1381 "modules/graphics/opengl/Graphics.cpp"
   $106 = $100 | 1; //@line 1381 "modules/graphics/opengl/Graphics.cpp"
   $107 = (($27) + ($106<<2)|0); //@line 1381 "modules/graphics/opengl/Graphics.cpp"
   HEAPF32[$107>>2] = $105; //@line 1381 "modules/graphics/opengl/Graphics.cpp"
   $108 = (($i3$08) + 1)|0; //@line 1378 "modules/graphics/opengl/Graphics.cpp"
   $109 = $22 + $phi$37; //@line 1378 "modules/graphics/opengl/Graphics.cpp"
   $110 = ($i3$08|0)<($72|0); //@line 1378 "modules/graphics/opengl/Graphics.cpp"
   if ($110) {
    $i3$08 = $108;$phi$37 = $109;
   } else {
    break;
   }
  }
 }
 $90 = HEAP32[$27>>2]|0; //@line 1384 "modules/graphics/opengl/Graphics.cpp"
 $91 = (($27) + ($24<<2)|0); //@line 1384 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$91>>2] = $90; //@line 1384 "modules/graphics/opengl/Graphics.cpp"
 $92 = ((($27)) + 4|0); //@line 1385 "modules/graphics/opengl/Graphics.cpp"
 $93 = HEAP32[$92>>2]|0; //@line 1385 "modules/graphics/opengl/Graphics.cpp"
 $94 = (($23) + 17)|0; //@line 1385 "modules/graphics/opengl/Graphics.cpp"
 $95 = (($27) + ($94<<2)|0); //@line 1385 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$95>>2] = $93; //@line 1385 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl8Graphics7polygonENS0_8Graphics8DrawModeEPKfj($this,$mode,$27,$25); //@line 1387 "modules/graphics/opengl/Graphics.cpp"
 __ZdaPv($27); //@line 1389 "modules/graphics/opengl/Graphics.cpp"
 STACKTOP = sp;return; //@line 1390 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics6circleENS0_8Graphics8DrawModeEfffi($this,$mode,$x,$y,$radius,$points) {
 $this = $this|0;
 $mode = $mode|0;
 $x = +$x;
 $y = +$y;
 $radius = +$radius;
 $points = $points|0;
 var $$arith = 0, $$overflow = 0, $$phi$trans$insert3$i = 0, $$points$i = 0, $$pre$i = 0, $$pre5$i = 0, $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0.0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0.0, $i$02$i = 0, $phi$01$i = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($points|0)<(1); //@line 1400 "modules/graphics/opengl/Graphics.cpp"
 $$points$i = $0 ? 1 : $points; //@line 1400 "modules/graphics/opengl/Graphics.cpp"
 $1 = (+($$points$i|0)); //@line 1401 "modules/graphics/opengl/Graphics.cpp"
 $2 = 6.2831854820251465 / $1; //@line 1401 "modules/graphics/opengl/Graphics.cpp"
 $3 = $$points$i << 1; //@line 1404 "modules/graphics/opengl/Graphics.cpp"
 $4 = (($3) + 2)|0; //@line 1404 "modules/graphics/opengl/Graphics.cpp"
 $$arith = $4<<2; //@line 1404 "modules/graphics/opengl/Graphics.cpp"
 $$overflow = ($4>>>0)>(1073741823); //@line 1404 "modules/graphics/opengl/Graphics.cpp"
 $5 = $$overflow ? -1 : $$arith; //@line 1404 "modules/graphics/opengl/Graphics.cpp"
 $6 = (__Znaj($5)|0); //@line 1404 "modules/graphics/opengl/Graphics.cpp"
 $7 = ($$points$i|0)>(0); //@line 1405 "modules/graphics/opengl/Graphics.cpp"
 if ($7) {
  $i$02$i = 0;$phi$01$i = 0.0;
  while(1) {
   $8 = (+Math_cos((+$phi$01$i))); //@line 1407 "modules/graphics/opengl/Graphics.cpp"
   $9 = $8 * $radius; //@line 1407 "modules/graphics/opengl/Graphics.cpp"
   $10 = $9 + $x; //@line 1407 "modules/graphics/opengl/Graphics.cpp"
   $11 = $i$02$i << 1; //@line 1407 "modules/graphics/opengl/Graphics.cpp"
   $12 = (($6) + ($11<<2)|0); //@line 1407 "modules/graphics/opengl/Graphics.cpp"
   HEAPF32[$12>>2] = $10; //@line 1407 "modules/graphics/opengl/Graphics.cpp"
   $13 = (+Math_sin((+$phi$01$i))); //@line 1408 "modules/graphics/opengl/Graphics.cpp"
   $14 = $13 * $radius; //@line 1408 "modules/graphics/opengl/Graphics.cpp"
   $15 = $14 + $y; //@line 1408 "modules/graphics/opengl/Graphics.cpp"
   $16 = $11 | 1; //@line 1408 "modules/graphics/opengl/Graphics.cpp"
   $17 = (($6) + ($16<<2)|0); //@line 1408 "modules/graphics/opengl/Graphics.cpp"
   HEAPF32[$17>>2] = $15; //@line 1408 "modules/graphics/opengl/Graphics.cpp"
   $18 = (($i$02$i) + 1)|0; //@line 1405 "modules/graphics/opengl/Graphics.cpp"
   $19 = $2 + $phi$01$i; //@line 1405 "modules/graphics/opengl/Graphics.cpp"
   $20 = ($18|0)<($$points$i|0); //@line 1405 "modules/graphics/opengl/Graphics.cpp"
   if ($20) {
    $i$02$i = $18;$phi$01$i = $19;
   } else {
    break;
   }
  }
  $$pre$i = HEAP32[$6>>2]|0; //@line 1411 "modules/graphics/opengl/Graphics.cpp"
  $$phi$trans$insert3$i = ((($6)) + 4|0); //@line 1412 "modules/graphics/opengl/Graphics.cpp"
  $$pre5$i = HEAP32[$$phi$trans$insert3$i>>2]|0; //@line 1412 "modules/graphics/opengl/Graphics.cpp"
  $22 = $$pre$i;$25 = $$pre5$i;
 } else {
  $22 = 0;$25 = 0;
 }
 $21 = (($6) + ($3<<2)|0); //@line 1411 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$21>>2] = $22; //@line 1411 "modules/graphics/opengl/Graphics.cpp"
 $23 = $3 | 1; //@line 1412 "modules/graphics/opengl/Graphics.cpp"
 $24 = (($6) + ($23<<2)|0); //@line 1412 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$24>>2] = $25; //@line 1412 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl8Graphics7polygonENS0_8Graphics8DrawModeEPKfj($this,$mode,$6,$4); //@line 1414 "modules/graphics/opengl/Graphics.cpp"
 __ZdaPv($6); //@line 1416 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1395 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics7ellipseENS0_8Graphics8DrawModeEffffi($this,$mode,$x,$y,$a,$b,$points) {
 $this = $this|0;
 $mode = $mode|0;
 $x = +$x;
 $y = +$y;
 $a = +$a;
 $b = +$b;
 $points = $points|0;
 var $$arith = 0, $$overflow = 0, $$phi$trans$insert3 = 0, $$points = 0, $$pre = 0, $$pre5 = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0.0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0.0, $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$02 = 0, $phi$01 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($points|0)<(1); //@line 1400 "modules/graphics/opengl/Graphics.cpp"
 $$points = $0 ? 1 : $points; //@line 1400 "modules/graphics/opengl/Graphics.cpp"
 $1 = (+($$points|0)); //@line 1401 "modules/graphics/opengl/Graphics.cpp"
 $2 = 6.2831854820251465 / $1; //@line 1401 "modules/graphics/opengl/Graphics.cpp"
 $3 = $$points << 1; //@line 1404 "modules/graphics/opengl/Graphics.cpp"
 $4 = (($3) + 2)|0; //@line 1404 "modules/graphics/opengl/Graphics.cpp"
 $$arith = $4<<2; //@line 1404 "modules/graphics/opengl/Graphics.cpp"
 $$overflow = ($4>>>0)>(1073741823); //@line 1404 "modules/graphics/opengl/Graphics.cpp"
 $5 = $$overflow ? -1 : $$arith; //@line 1404 "modules/graphics/opengl/Graphics.cpp"
 $6 = (__Znaj($5)|0); //@line 1404 "modules/graphics/opengl/Graphics.cpp"
 $7 = ($$points|0)>(0); //@line 1405 "modules/graphics/opengl/Graphics.cpp"
 if ($7) {
  $i$02 = 0;$phi$01 = 0.0;
  while(1) {
   $13 = (+Math_cos((+$phi$01))); //@line 1407 "modules/graphics/opengl/Graphics.cpp"
   $14 = $13 * $a; //@line 1407 "modules/graphics/opengl/Graphics.cpp"
   $15 = $14 + $x; //@line 1407 "modules/graphics/opengl/Graphics.cpp"
   $16 = $i$02 << 1; //@line 1407 "modules/graphics/opengl/Graphics.cpp"
   $17 = (($6) + ($16<<2)|0); //@line 1407 "modules/graphics/opengl/Graphics.cpp"
   HEAPF32[$17>>2] = $15; //@line 1407 "modules/graphics/opengl/Graphics.cpp"
   $18 = (+Math_sin((+$phi$01))); //@line 1408 "modules/graphics/opengl/Graphics.cpp"
   $19 = $18 * $b; //@line 1408 "modules/graphics/opengl/Graphics.cpp"
   $20 = $19 + $y; //@line 1408 "modules/graphics/opengl/Graphics.cpp"
   $21 = $16 | 1; //@line 1408 "modules/graphics/opengl/Graphics.cpp"
   $22 = (($6) + ($21<<2)|0); //@line 1408 "modules/graphics/opengl/Graphics.cpp"
   HEAPF32[$22>>2] = $20; //@line 1408 "modules/graphics/opengl/Graphics.cpp"
   $23 = (($i$02) + 1)|0; //@line 1405 "modules/graphics/opengl/Graphics.cpp"
   $24 = $2 + $phi$01; //@line 1405 "modules/graphics/opengl/Graphics.cpp"
   $25 = ($23|0)<($$points|0); //@line 1405 "modules/graphics/opengl/Graphics.cpp"
   if ($25) {
    $i$02 = $23;$phi$01 = $24;
   } else {
    break;
   }
  }
  $$pre = HEAP32[$6>>2]|0; //@line 1411 "modules/graphics/opengl/Graphics.cpp"
  $$phi$trans$insert3 = ((($6)) + 4|0); //@line 1412 "modules/graphics/opengl/Graphics.cpp"
  $$pre5 = HEAP32[$$phi$trans$insert3>>2]|0; //@line 1412 "modules/graphics/opengl/Graphics.cpp"
  $12 = $$pre5;$9 = $$pre;
 } else {
  $12 = 0;$9 = 0;
 }
 $8 = (($6) + ($3<<2)|0); //@line 1411 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$8>>2] = $9; //@line 1411 "modules/graphics/opengl/Graphics.cpp"
 $10 = $3 | 1; //@line 1412 "modules/graphics/opengl/Graphics.cpp"
 $11 = (($6) + ($10<<2)|0); //@line 1412 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$11>>2] = $12; //@line 1412 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love8graphics6opengl8Graphics7polygonENS0_8Graphics8DrawModeEPKfj($this,$mode,$6,$4); //@line 1414 "modules/graphics/opengl/Graphics.cpp"
 __ZdaPv($6); //@line 1416 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1417 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics3arcENS0_8Graphics8DrawModeENS3_7ArcModeEfffffi($this,$drawmode,$arcmode,$x,$y,$radius,$angle1,$angle2,$points) {
 $this = $this|0;
 $drawmode = $drawmode|0;
 $arcmode = $arcmode|0;
 $x = +$x;
 $y = +$y;
 $radius = +$radius;
 $angle1 = +$angle1;
 $angle2 = +$angle2;
 $points = $points|0;
 var $$$0 = 0, $$0 = 0, $$arith = 0, $$arith10 = 0, $$arith2 = 0, $$arith6 = 0, $$overflow = 0, $$overflow11 = 0, $$overflow3 = 0, $$overflow7 = 0, $$phi$trans$insert16 = 0, $$phi$trans$insert3$i$i = 0, $$pre = 0, $$pre$i$i = 0, $$pre18 = 0, $$pre5$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0.0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0, $78 = 0, $79 = 0.0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0, $88 = 0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0, $93 = 0, $94 = 0, $95 = 0.0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $coords$0 = 0, $exitcond = 0;
 var $exitcond14 = 0, $exitcond15 = 0, $fabsf = 0.0, $i$01$i = 0, $i$01$i6 = 0, $i$01$i9 = 0, $i$02$i$i = 0, $num_coords$0 = 0, $or$cond = 0, $or$cond3 = 0, $or$cond4 = 0, $phi$0 = 0.0, $phi$01$i$i = 0.0, $phi$1 = 0.0, $phi$2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($points|0)<(1); //@line 1422 "modules/graphics/opengl/Graphics.cpp"
 $1 = $angle1 == $angle2; //@line 1422 "modules/graphics/opengl/Graphics.cpp"
 $or$cond4 = $1 | $0; //@line 1422 "modules/graphics/opengl/Graphics.cpp"
 if ($or$cond4) {
  return; //@line 1495 "modules/graphics/opengl/Graphics.cpp"
 }
 $2 = $angle1 - $angle2; //@line 1426 "modules/graphics/opengl/Graphics.cpp"
 $fabsf = (+Math_abs((+$2))); //@line 1426 "modules/graphics/opengl/Graphics.cpp"
 $3 = !($fabsf >= 6.2831854820251465); //@line 1426 "modules/graphics/opengl/Graphics.cpp"
 if (!($3)) {
  $4 = (+($points|0)); //@line 1401 "modules/graphics/opengl/Graphics.cpp"
  $5 = 6.2831854820251465 / $4; //@line 1401 "modules/graphics/opengl/Graphics.cpp"
  $6 = $points << 1; //@line 1404 "modules/graphics/opengl/Graphics.cpp"
  $7 = (($6) + 2)|0; //@line 1404 "modules/graphics/opengl/Graphics.cpp"
  $$arith10 = $7<<2; //@line 1404 "modules/graphics/opengl/Graphics.cpp"
  $$overflow11 = ($7>>>0)>(1073741823); //@line 1404 "modules/graphics/opengl/Graphics.cpp"
  $8 = $$overflow11 ? -1 : $$arith10; //@line 1404 "modules/graphics/opengl/Graphics.cpp"
  $9 = (__Znaj($8)|0); //@line 1404 "modules/graphics/opengl/Graphics.cpp"
  $10 = ($points|0)>(0); //@line 1405 "modules/graphics/opengl/Graphics.cpp"
  if ($10) {
   $i$02$i$i = 0;$phi$01$i$i = 0.0;
   while(1) {
    $11 = (+Math_cos((+$phi$01$i$i))); //@line 1407 "modules/graphics/opengl/Graphics.cpp"
    $12 = $11 * $radius; //@line 1407 "modules/graphics/opengl/Graphics.cpp"
    $13 = $12 + $x; //@line 1407 "modules/graphics/opengl/Graphics.cpp"
    $14 = $i$02$i$i << 1; //@line 1407 "modules/graphics/opengl/Graphics.cpp"
    $15 = (($9) + ($14<<2)|0); //@line 1407 "modules/graphics/opengl/Graphics.cpp"
    HEAPF32[$15>>2] = $13; //@line 1407 "modules/graphics/opengl/Graphics.cpp"
    $16 = (+Math_sin((+$phi$01$i$i))); //@line 1408 "modules/graphics/opengl/Graphics.cpp"
    $17 = $16 * $radius; //@line 1408 "modules/graphics/opengl/Graphics.cpp"
    $18 = $17 + $y; //@line 1408 "modules/graphics/opengl/Graphics.cpp"
    $19 = $14 | 1; //@line 1408 "modules/graphics/opengl/Graphics.cpp"
    $20 = (($9) + ($19<<2)|0); //@line 1408 "modules/graphics/opengl/Graphics.cpp"
    HEAPF32[$20>>2] = $18; //@line 1408 "modules/graphics/opengl/Graphics.cpp"
    $21 = (($i$02$i$i) + 1)|0; //@line 1405 "modules/graphics/opengl/Graphics.cpp"
    $22 = $5 + $phi$01$i$i; //@line 1405 "modules/graphics/opengl/Graphics.cpp"
    $23 = ($21|0)<($points|0); //@line 1405 "modules/graphics/opengl/Graphics.cpp"
    if ($23) {
     $i$02$i$i = $21;$phi$01$i$i = $22;
    } else {
     break;
    }
   }
   $$pre$i$i = HEAP32[$9>>2]|0; //@line 1411 "modules/graphics/opengl/Graphics.cpp"
   $$phi$trans$insert3$i$i = ((($9)) + 4|0); //@line 1412 "modules/graphics/opengl/Graphics.cpp"
   $$pre5$i$i = HEAP32[$$phi$trans$insert3$i$i>>2]|0; //@line 1412 "modules/graphics/opengl/Graphics.cpp"
   $25 = $$pre$i$i;$28 = $$pre5$i$i;
  } else {
   $25 = 0;$28 = 0;
  }
  $24 = (($9) + ($6<<2)|0); //@line 1411 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$24>>2] = $25; //@line 1411 "modules/graphics/opengl/Graphics.cpp"
  $26 = $6 | 1; //@line 1412 "modules/graphics/opengl/Graphics.cpp"
  $27 = (($9) + ($26<<2)|0); //@line 1412 "modules/graphics/opengl/Graphics.cpp"
  HEAP32[$27>>2] = $28; //@line 1412 "modules/graphics/opengl/Graphics.cpp"
  __ZN4love8graphics6opengl8Graphics7polygonENS0_8Graphics8DrawModeEPKfj($this,$drawmode,$9,$7); //@line 1414 "modules/graphics/opengl/Graphics.cpp"
  __ZdaPv($9); //@line 1416 "modules/graphics/opengl/Graphics.cpp"
  return; //@line 1495 "modules/graphics/opengl/Graphics.cpp"
 }
 $29 = $angle2 - $angle1; //@line 1432 "modules/graphics/opengl/Graphics.cpp"
 $30 = (+($points|0)); //@line 1432 "modules/graphics/opengl/Graphics.cpp"
 $31 = $29 / $30; //@line 1432 "modules/graphics/opengl/Graphics.cpp"
 $32 = $31 == 0.0; //@line 1434 "modules/graphics/opengl/Graphics.cpp"
 if ($32) {
  return; //@line 1495 "modules/graphics/opengl/Graphics.cpp"
 }
 $33 = ($drawmode|0)==(0); //@line 1440 "modules/graphics/opengl/Graphics.cpp"
 $34 = ($arcmode|0)==(1); //@line 1440 "modules/graphics/opengl/Graphics.cpp"
 $or$cond = $33 & $34; //@line 1440 "modules/graphics/opengl/Graphics.cpp"
 if ($or$cond) {
  $35 = $fabsf < 0.069813169538974762; //@line 1440 "modules/graphics/opengl/Graphics.cpp"
  if ($35) {
   $$0 = 0;
  } else {
   $$0 = 1;
  }
 } else {
  $$0 = $arcmode;
 }
 $36 = ($drawmode|0)==(1); //@line 1445 "modules/graphics/opengl/Graphics.cpp"
 $37 = ($$0|0)==(0); //@line 1445 "modules/graphics/opengl/Graphics.cpp"
 $or$cond3 = $36 & $37; //@line 1445 "modules/graphics/opengl/Graphics.cpp"
 $$$0 = $or$cond3 ? 1 : $$0; //@line 1445 "modules/graphics/opengl/Graphics.cpp"
 $38 = ($$$0|0)==(2); //@line 1462 "modules/graphics/opengl/Graphics.cpp"
 L20: do {
  if ($38) {
   $39 = $points << 1; //@line 1464 "modules/graphics/opengl/Graphics.cpp"
   $40 = (($39) + 6)|0; //@line 1464 "modules/graphics/opengl/Graphics.cpp"
   $$arith6 = $40<<2; //@line 1465 "modules/graphics/opengl/Graphics.cpp"
   $$overflow7 = ($40>>>0)>(1073741823); //@line 1465 "modules/graphics/opengl/Graphics.cpp"
   $41 = $$overflow7 ? -1 : $$arith6; //@line 1465 "modules/graphics/opengl/Graphics.cpp"
   $42 = (__Znaj($41)|0); //@line 1465 "modules/graphics/opengl/Graphics.cpp"
   $43 = (($39) + 4)|0; //@line 1467 "modules/graphics/opengl/Graphics.cpp"
   $44 = (($42) + ($43<<2)|0); //@line 1467 "modules/graphics/opengl/Graphics.cpp"
   HEAPF32[$44>>2] = $x; //@line 1467 "modules/graphics/opengl/Graphics.cpp"
   HEAPF32[$42>>2] = $x; //@line 1467 "modules/graphics/opengl/Graphics.cpp"
   $45 = (($39) + 5)|0; //@line 1468 "modules/graphics/opengl/Graphics.cpp"
   $46 = (($42) + ($45<<2)|0); //@line 1468 "modules/graphics/opengl/Graphics.cpp"
   HEAPF32[$46>>2] = $y; //@line 1468 "modules/graphics/opengl/Graphics.cpp"
   $47 = ((($42)) + 4|0); //@line 1468 "modules/graphics/opengl/Graphics.cpp"
   HEAPF32[$47>>2] = $y; //@line 1468 "modules/graphics/opengl/Graphics.cpp"
   $48 = ((($42)) + 8|0); //@line 1470 "modules/graphics/opengl/Graphics.cpp"
   $49 = ($points|0)<(0); //@line 1455 "modules/graphics/opengl/Graphics.cpp"
   if ($49) {
    $coords$0 = $42;$num_coords$0 = $40;
   } else {
    $i$01$i = 0;$phi$0 = $angle1;
    while(1) {
     $50 = (+Math_cos((+$phi$0))); //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     $51 = $50 * $radius; //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     $52 = $51 + $x; //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     $53 = $i$01$i << 1; //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     $54 = (($48) + ($53<<2)|0); //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     HEAPF32[$54>>2] = $52; //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     $55 = (+Math_sin((+$phi$0))); //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     $56 = $55 * $radius; //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     $57 = $56 + $y; //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     $58 = $53 | 1; //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     $59 = (($48) + ($58<<2)|0); //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     HEAPF32[$59>>2] = $57; //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     $60 = (($i$01$i) + 1)|0; //@line 1455 "modules/graphics/opengl/Graphics.cpp"
     $61 = $31 + $phi$0; //@line 1455 "modules/graphics/opengl/Graphics.cpp"
     $exitcond = ($i$01$i|0)==($points|0); //@line 1455 "modules/graphics/opengl/Graphics.cpp"
     if ($exitcond) {
      $coords$0 = $42;$num_coords$0 = $40;
      break;
     } else {
      $i$01$i = $60;$phi$0 = $61;
     }
    }
   }
  } else {
   $62 = ($$$0|0)==(0); //@line 1472 "modules/graphics/opengl/Graphics.cpp"
   $63 = $points << 1; //@line 1474 "modules/graphics/opengl/Graphics.cpp"
   if ($62) {
    $64 = (($63) + 2)|0; //@line 1474 "modules/graphics/opengl/Graphics.cpp"
    $$arith2 = $64<<2; //@line 1475 "modules/graphics/opengl/Graphics.cpp"
    $$overflow3 = ($64>>>0)>(1073741823); //@line 1475 "modules/graphics/opengl/Graphics.cpp"
    $65 = $$overflow3 ? -1 : $$arith2; //@line 1475 "modules/graphics/opengl/Graphics.cpp"
    $66 = (__Znaj($65)|0); //@line 1475 "modules/graphics/opengl/Graphics.cpp"
    $67 = ($points|0)<(0); //@line 1455 "modules/graphics/opengl/Graphics.cpp"
    if ($67) {
     $coords$0 = $66;$num_coords$0 = $64;
     break;
    } else {
     $i$01$i6 = 0;$phi$1 = $angle1;
    }
    while(1) {
     $68 = (+Math_cos((+$phi$1))); //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     $69 = $68 * $radius; //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     $70 = $69 + $x; //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     $71 = $i$01$i6 << 1; //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     $72 = (($66) + ($71<<2)|0); //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     HEAPF32[$72>>2] = $70; //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     $73 = (+Math_sin((+$phi$1))); //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     $74 = $73 * $radius; //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     $75 = $74 + $y; //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     $76 = $71 | 1; //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     $77 = (($66) + ($76<<2)|0); //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     HEAPF32[$77>>2] = $75; //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     $78 = (($i$01$i6) + 1)|0; //@line 1455 "modules/graphics/opengl/Graphics.cpp"
     $79 = $31 + $phi$1; //@line 1455 "modules/graphics/opengl/Graphics.cpp"
     $exitcond14 = ($i$01$i6|0)==($points|0); //@line 1455 "modules/graphics/opengl/Graphics.cpp"
     if ($exitcond14) {
      $coords$0 = $66;$num_coords$0 = $64;
      break L20;
     } else {
      $i$01$i6 = $78;$phi$1 = $79;
     }
    }
   }
   $80 = (($63) + 4)|0; //@line 1481 "modules/graphics/opengl/Graphics.cpp"
   $$arith = $80<<2; //@line 1482 "modules/graphics/opengl/Graphics.cpp"
   $$overflow = ($80>>>0)>(1073741823); //@line 1482 "modules/graphics/opengl/Graphics.cpp"
   $81 = $$overflow ? -1 : $$arith; //@line 1482 "modules/graphics/opengl/Graphics.cpp"
   $82 = (__Znaj($81)|0); //@line 1482 "modules/graphics/opengl/Graphics.cpp"
   $83 = ($points|0)<(0); //@line 1455 "modules/graphics/opengl/Graphics.cpp"
   if ($83) {
    $101 = 0;$98 = 0;
   } else {
    $i$01$i9 = 0;$phi$2 = $angle1;
    while(1) {
     $84 = (+Math_cos((+$phi$2))); //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     $85 = $84 * $radius; //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     $86 = $85 + $x; //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     $87 = $i$01$i9 << 1; //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     $88 = (($82) + ($87<<2)|0); //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     HEAPF32[$88>>2] = $86; //@line 1457 "modules/graphics/opengl/Graphics.cpp"
     $89 = (+Math_sin((+$phi$2))); //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     $90 = $89 * $radius; //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     $91 = $90 + $y; //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     $92 = $87 | 1; //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     $93 = (($82) + ($92<<2)|0); //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     HEAPF32[$93>>2] = $91; //@line 1458 "modules/graphics/opengl/Graphics.cpp"
     $94 = (($i$01$i9) + 1)|0; //@line 1455 "modules/graphics/opengl/Graphics.cpp"
     $95 = $31 + $phi$2; //@line 1455 "modules/graphics/opengl/Graphics.cpp"
     $exitcond15 = ($i$01$i9|0)==($points|0); //@line 1455 "modules/graphics/opengl/Graphics.cpp"
     if ($exitcond15) {
      break;
     } else {
      $i$01$i9 = $94;$phi$2 = $95;
     }
    }
    $$pre = HEAP32[$82>>2]|0; //@line 1487 "modules/graphics/opengl/Graphics.cpp"
    $$phi$trans$insert16 = ((($82)) + 4|0); //@line 1488 "modules/graphics/opengl/Graphics.cpp"
    $$pre18 = HEAP32[$$phi$trans$insert16>>2]|0; //@line 1488 "modules/graphics/opengl/Graphics.cpp"
    $101 = $$pre18;$98 = $$pre;
   }
   $96 = (($63) + 2)|0; //@line 1487 "modules/graphics/opengl/Graphics.cpp"
   $97 = (($82) + ($96<<2)|0); //@line 1487 "modules/graphics/opengl/Graphics.cpp"
   HEAP32[$97>>2] = $98; //@line 1487 "modules/graphics/opengl/Graphics.cpp"
   $99 = (($63) + 3)|0; //@line 1488 "modules/graphics/opengl/Graphics.cpp"
   $100 = (($82) + ($99<<2)|0); //@line 1488 "modules/graphics/opengl/Graphics.cpp"
   HEAP32[$100>>2] = $101; //@line 1488 "modules/graphics/opengl/Graphics.cpp"
   $coords$0 = $82;$num_coords$0 = $80;
  }
 } while(0);
 __ZN4love8graphics6opengl8Graphics7polygonENS0_8Graphics8DrawModeEPKfj($this,$drawmode,$coords$0,$num_coords$0); //@line 1492 "modules/graphics/opengl/Graphics.cpp"
 __ZdaPv($coords$0); //@line 1494 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1495 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics13newScreenshotEPNS_5image5ImageEb($this,$image,$copyAlpha) {
 $this = $this|0;
 $image = $image|0;
 $copyAlpha = $copyAlpha|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$pre$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i$i = 0, $$pre$i$i$i6 = 0, $$pre$i9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $canvases = 0, $copyAlpha$not = 0, $dst$014 = 0, $exitcond = 0, $i$016 = 0, $i1$015 = 0, $or$cond = 0, $pixels$0 = 0, $src$013 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $canvases = sp + 4|0;
 $0 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -44|0); //@line 1524 "modules/graphics/opengl/Graphics.cpp"
 __ZNSt3__16vectorIN4love9StrongRefINS1_8graphics6opengl6CanvasEEENS_9allocatorIS6_EEEC2ERKS9_($canvases,$2); //@line 1524 "modules/graphics/opengl/Graphics.cpp"
 $3 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[158272]|0; //@line 1065 "modules/graphics/opengl/Graphics.cpp"
 $5 = ($4|0)==(0|0); //@line 1065 "modules/graphics/opengl/Graphics.cpp"
 if ($5) {
  label = 3;
 } else {
  __THREW__ = 0;
  invoke_vii(4001,($4|0),0); //@line 1066 "modules/graphics/opengl/Graphics.cpp"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if ($7) {
   label = 12;
  } else {
   label = 3;
  }
 }
 do {
  if ((label|0) == 3) {
   $8 = ((($3)) + -44|0); //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $9 = HEAP32[$8>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $10 = ((($3)) + -40|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $11 = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $12 = ($11|0)==($9|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   L5: do {
    if (!($12)) {
     $14 = $11;
     while(1) {
      $13 = ((($14)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$10>>2] = $13; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $15 = HEAP32[$13>>2]|0; //@line 106 "./common/Object.h"
      $16 = ($15|0)==(0|0); //@line 106 "./common/Object.h"
      if ($16) {
       $18 = $13;
      } else {
       $19 = HEAP32[$15>>2]|0; //@line 106 "./common/Object.h"
       $20 = ((($19)) + 12|0); //@line 106 "./common/Object.h"
       $21 = HEAP32[$20>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($21|0,($15|0)); //@line 106 "./common/Object.h"
       $22 = __THREW__; __THREW__ = 0;
       $23 = $22&1;
       if ($23) {
        break;
       }
       $$pre$i$i$i$i = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $18 = $$pre$i$i$i$i;
      }
      $17 = ($18|0)==($9|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($17) {
       break L5;
      } else {
       $14 = $18;
      }
     }
     $24 = ___cxa_find_matching_catch_3(0|0)|0;
     $25 = tempRet0;
     ___clang_call_terminate($24); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   } while(0);
   $26 = ((($this)) + 32|0); //@line 610 "modules/graphics/opengl/Graphics.cpp"
   $27 = HEAP32[$26>>2]|0; //@line 610 "modules/graphics/opengl/Graphics.cpp"
   $28 = ((($this)) + 36|0); //@line 615 "modules/graphics/opengl/Graphics.cpp"
   $29 = HEAP32[$28>>2]|0; //@line 615 "modules/graphics/opengl/Graphics.cpp"
   $30 = $27 << 2; //@line 1530 "modules/graphics/opengl/Graphics.cpp"
   $31 = Math_imul($30, $29)|0; //@line 1531 "modules/graphics/opengl/Graphics.cpp"
   $32 = ($31|0)>(-1); //@line 1538 "modules/graphics/opengl/Graphics.cpp"
   $33 = $32 ? $31 : -1; //@line 1538 "modules/graphics/opengl/Graphics.cpp"
   __THREW__ = 0;
   $34 = (invoke_ii(3492,($33|0))|0); //@line 1538 "modules/graphics/opengl/Graphics.cpp"
   $35 = __THREW__; __THREW__ = 0;
   $36 = $35&1;
   if ($36) {
    $pixels$0 = 0;
   } else {
    __THREW__ = 0;
    $37 = (invoke_ii(3492,($33|0))|0); //@line 1539 "modules/graphics/opengl/Graphics.cpp"
    $38 = __THREW__; __THREW__ = 0;
    $39 = $38&1;
    if ($39) {
     $pixels$0 = $34;
    } else {
     $40 = HEAP32[165294]|0; //@line 68 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_viiiiiii($40|0,0,0,($27|0),($29|0),6408,5121,($34|0)); //@line 68 "./libraries/glad/gladfuncs.hpp"
     $41 = __THREW__; __THREW__ = 0;
     $42 = $41&1;
     if ($42) {
      label = 12;
      break;
     }
     $copyAlpha$not = $copyAlpha ^ 1; //@line 1577 "modules/graphics/opengl/Graphics.cpp"
     $62 = ($31|0)>(3); //@line 1580 "modules/graphics/opengl/Graphics.cpp"
     $or$cond = $62 & $copyAlpha$not; //@line 1577 "modules/graphics/opengl/Graphics.cpp"
     if ($or$cond) {
      $i$016 = 3;
      while(1) {
       $63 = (($34) + ($i$016)|0); //@line 1581 "modules/graphics/opengl/Graphics.cpp"
       HEAP8[$63>>0] = -1; //@line 1581 "modules/graphics/opengl/Graphics.cpp"
       $64 = (($i$016) + 4)|0; //@line 1580 "modules/graphics/opengl/Graphics.cpp"
       $65 = ($64|0)<($31|0); //@line 1580 "modules/graphics/opengl/Graphics.cpp"
       if ($65) {
        $i$016 = $64;
       } else {
        break;
       }
      }
     }
     $66 = (0 - ($30))|0; //@line 1585 "modules/graphics/opengl/Graphics.cpp"
     $67 = ($29|0)>(0); //@line 1588 "modules/graphics/opengl/Graphics.cpp"
     if ($67) {
      $68 = (($37) + ($31)|0); //@line 1586 "modules/graphics/opengl/Graphics.cpp"
      $69 = (($34) + ($66)|0); //@line 1585 "modules/graphics/opengl/Graphics.cpp"
      $dst$014 = $68;$i1$015 = 0;$src$013 = $69;
      while(1) {
       $70 = (($dst$014) + ($66)|0); //@line 1589 "modules/graphics/opengl/Graphics.cpp"
       $71 = (($src$013) + ($30)|0); //@line 1589 "modules/graphics/opengl/Graphics.cpp"
       _memcpy(($70|0),($71|0),($30|0))|0; //@line 1589 "modules/graphics/opengl/Graphics.cpp"
       $72 = (($i1$015) + 1)|0; //@line 1588 "modules/graphics/opengl/Graphics.cpp"
       $exitcond = ($72|0)==($29|0); //@line 1588 "modules/graphics/opengl/Graphics.cpp"
       if ($exitcond) {
        break;
       } else {
        $dst$014 = $70;$i1$015 = $72;$src$013 = $71;
       }
      }
     }
     __ZdaPv($34); //@line 1591 "modules/graphics/opengl/Graphics.cpp"
     $73 = HEAP32[$image>>2]|0; //@line 1598 "modules/graphics/opengl/Graphics.cpp"
     $74 = ((($73)) + 32|0); //@line 1598 "modules/graphics/opengl/Graphics.cpp"
     $75 = HEAP32[$74>>2]|0; //@line 1598 "modules/graphics/opengl/Graphics.cpp"
     __THREW__ = 0;
     $76 = (invoke_iiiiii($75|0,($image|0),($27|0),($29|0),($37|0),1)|0); //@line 1598 "modules/graphics/opengl/Graphics.cpp"
     $77 = __THREW__; __THREW__ = 0;
     $78 = $77&1;
     if ($78) {
      $81 = ___cxa_find_matching_catch_3(96|0)|0;
      $82 = tempRet0;
      $83 = (_llvm_eh_typeid_for((96|0))|0); //@line 1599 "modules/graphics/opengl/Graphics.cpp"
      $84 = ($82|0)==($83|0); //@line 1599 "modules/graphics/opengl/Graphics.cpp"
      if (!($84)) {
       $$1 = $81;$$12 = $82;
       break;
      }
      (___cxa_begin_catch(($81|0))|0); //@line 1599 "modules/graphics/opengl/Graphics.cpp"
      __ZdaPv($37); //@line 1602 "modules/graphics/opengl/Graphics.cpp"
      __THREW__ = 0;
      invoke_vii(4046,($this|0),($canvases|0)); //@line 1603 "modules/graphics/opengl/Graphics.cpp"
      $85 = __THREW__; __THREW__ = 0;
      $86 = $85&1;
      if (!($86)) {
       __THREW__ = 0;
       invoke_v(3656); //@line 1604 "modules/graphics/opengl/Graphics.cpp"
       $87 = __THREW__; __THREW__ = 0;
      }
      $88 = ___cxa_find_matching_catch_2()|0;
      $89 = tempRet0;
      __THREW__ = 0;
      invoke_v(3378); //@line 1605 "modules/graphics/opengl/Graphics.cpp"
      $90 = __THREW__; __THREW__ = 0;
      $91 = $90&1;
      if (!($91)) {
       $$1 = $88;$$12 = $89;
       break;
      }
      $130 = ___cxa_find_matching_catch_3(0|0)|0;
      $131 = tempRet0;
      ___clang_call_terminate($130); //@line 1547 "modules/graphics/opengl/Graphics.cpp"
      // unreachable; //@line 1547 "modules/graphics/opengl/Graphics.cpp"
     }
     __THREW__ = 0;
     invoke_vii(4046,($this|0),($canvases|0)); //@line 1608 "modules/graphics/opengl/Graphics.cpp"
     $79 = __THREW__; __THREW__ = 0;
     $80 = $79&1;
     if ($80) {
      label = 12;
      break;
     }
     $92 = HEAP32[$canvases>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $93 = ($92|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($93) {
      STACKTOP = sp;return ($76|0); //@line 1611 "modules/graphics/opengl/Graphics.cpp"
     }
     $94 = ((($canvases)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $95 = HEAP32[$94>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $96 = ($95|0)==($92|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     do {
      if ($96) {
       $110 = $92;
      } else {
       $98 = $95;
       while(1) {
        $97 = ((($98)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$94>>2] = $97; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $99 = HEAP32[$97>>2]|0; //@line 106 "./common/Object.h"
        $100 = ($99|0)==(0|0); //@line 106 "./common/Object.h"
        if ($100) {
         $102 = $97;
        } else {
         $103 = HEAP32[$99>>2]|0; //@line 106 "./common/Object.h"
         $104 = ((($103)) + 12|0); //@line 106 "./common/Object.h"
         $105 = HEAP32[$104>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($105|0,($99|0)); //@line 106 "./common/Object.h"
         $106 = __THREW__; __THREW__ = 0;
         $107 = $106&1;
         if ($107) {
          label = 39;
          break;
         }
         $$pre$i$i$i = HEAP32[$94>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $102 = $$pre$i$i$i;
        }
        $101 = ($102|0)==($92|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($101) {
         label = 40;
         break;
        } else {
         $98 = $102;
        }
       }
       if ((label|0) == 39) {
        $108 = ___cxa_find_matching_catch_3(0|0)|0;
        $109 = tempRet0;
        ___clang_call_terminate($108); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       else if ((label|0) == 40) {
        $$pre$i = HEAP32[$canvases>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $110 = $$pre$i;
        break;
       }
      }
     } while(0);
     __ZdlPv($110); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     STACKTOP = sp;return ($76|0); //@line 1611 "modules/graphics/opengl/Graphics.cpp"
    }
   }
   $45 = ___cxa_find_matching_catch_3(8992|0)|0;
   $46 = tempRet0;
   $47 = (_llvm_eh_typeid_for((8992|0))|0); //@line 1540 "modules/graphics/opengl/Graphics.cpp"
   $48 = ($46|0)==($47|0); //@line 1540 "modules/graphics/opengl/Graphics.cpp"
   if ($48) {
    (___cxa_begin_catch(($45|0))|0); //@line 1540 "modules/graphics/opengl/Graphics.cpp"
    $49 = ($pixels$0|0)==(0|0); //@line 1543 "modules/graphics/opengl/Graphics.cpp"
    if (!($49)) {
     __ZdaPv($pixels$0); //@line 1543 "modules/graphics/opengl/Graphics.cpp"
    }
    __THREW__ = 0;
    invoke_vii(4046,($this|0),($canvases|0)); //@line 1545 "modules/graphics/opengl/Graphics.cpp"
    $50 = __THREW__; __THREW__ = 0;
    $51 = $50&1;
    do {
     if ($51) {
      label = 19;
     } else {
      $52 = (___cxa_allocate_exception(16)|0); //@line 1546 "modules/graphics/opengl/Graphics.cpp"
      __THREW__ = 0;
      invoke_viii(3349,($52|0),(507610|0),($vararg_buffer|0)); //@line 1546 "modules/graphics/opengl/Graphics.cpp"
      $53 = __THREW__; __THREW__ = 0;
      $54 = $53&1;
      if ($54) {
       $58 = ___cxa_find_matching_catch_2()|0;
       $59 = tempRet0;
       ___cxa_free_exception(($52|0)); //@line 1546 "modules/graphics/opengl/Graphics.cpp"
       $$0 = $58;$$01 = $59;
       break;
      } else {
       __THREW__ = 0;
       invoke_viii(3350,($52|0),(96|0),(344|0)); //@line 1546 "modules/graphics/opengl/Graphics.cpp"
       $55 = __THREW__; __THREW__ = 0;
       label = 19;
       break;
      }
     }
    } while(0);
    if ((label|0) == 19) {
     $56 = ___cxa_find_matching_catch_2()|0;
     $57 = tempRet0;
     $$0 = $56;$$01 = $57;
    }
    __THREW__ = 0;
    invoke_v(3378); //@line 1547 "modules/graphics/opengl/Graphics.cpp"
    $60 = __THREW__; __THREW__ = 0;
    $61 = $60&1;
    if ($61) {
     $130 = ___cxa_find_matching_catch_3(0|0)|0;
     $131 = tempRet0;
     ___clang_call_terminate($130); //@line 1547 "modules/graphics/opengl/Graphics.cpp"
     // unreachable; //@line 1547 "modules/graphics/opengl/Graphics.cpp"
    } else {
     $$1 = $$0;$$12 = $$01;
    }
   } else {
    $$1 = $45;$$12 = $46;
   }
  }
 } while(0);
 if ((label|0) == 12) {
  $43 = ___cxa_find_matching_catch_2()|0;
  $44 = tempRet0;
  $$1 = $43;$$12 = $44;
 }
 $111 = HEAP32[$canvases>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $112 = ($111|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($112) {
  ___resumeException($$1|0);
  // unreachable;
 }
 $113 = ((($canvases)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $114 = HEAP32[$113>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $115 = ($114|0)==($111|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($115) {
   $129 = $111;
  } else {
   $117 = $114;
   while(1) {
    $116 = ((($117)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$113>>2] = $116; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $118 = HEAP32[$116>>2]|0; //@line 106 "./common/Object.h"
    $119 = ($118|0)==(0|0); //@line 106 "./common/Object.h"
    if ($119) {
     $121 = $116;
    } else {
     $122 = HEAP32[$118>>2]|0; //@line 106 "./common/Object.h"
     $123 = ((($122)) + 12|0); //@line 106 "./common/Object.h"
     $124 = HEAP32[$123>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($124|0,($118|0)); //@line 106 "./common/Object.h"
     $125 = __THREW__; __THREW__ = 0;
     $126 = $125&1;
     if ($126) {
      label = 49;
      break;
     }
     $$pre$i$i$i6 = HEAP32[$113>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $121 = $$pre$i$i$i6;
    }
    $120 = ($121|0)==($111|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($120) {
     label = 50;
     break;
    } else {
     $117 = $121;
    }
   }
   if ((label|0) == 49) {
    $127 = ___cxa_find_matching_catch_3(0|0)|0;
    $128 = tempRet0;
    ___clang_call_terminate($127); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 50) {
    $$pre$i9 = HEAP32[$canvases>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $129 = $$pre$i9;
    break;
   }
  }
 } while(0);
 __ZdlPv($129); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNK4love8graphics6opengl8Graphics15getRendererInfoEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, dest = 0, label = 0;
 var sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 dest=$agg$result; stop=dest+48|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $0 = HEAP8[1252012]|0; //@line 1617 "modules/graphics/opengl/Graphics.cpp"
 $1 = ($0<<24>>24)==(0); //@line 1617 "modules/graphics/opengl/Graphics.cpp"
 if ($1) {
  __THREW__ = 0;
  (invoke_iii(3360,($agg$result|0),(435751|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if ($7) {
   label = 3;
  } else {
   label = 5;
  }
 } else {
  __THREW__ = 0;
  (invoke_iii(3360,($agg$result|0),(435741|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $2 = __THREW__; __THREW__ = 0;
  $3 = $2&1;
  if ($3) {
   label = 3;
  } else {
   label = 5;
  }
 }
 do {
  if ((label|0) == 5) {
   $8 = HEAP32[165140]|0; //@line 74 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   $9 = (invoke_ii($8|0,7938)|0); //@line 74 "./libraries/glad/gladfuncs.hpp"
   $10 = __THREW__; __THREW__ = 0;
   $11 = $10&1;
   if ($11) {
    label = 3;
   } else {
    $12 = ($9|0)==(0|0); //@line 1623 "modules/graphics/opengl/Graphics.cpp"
    if ($12) {
     $16 = (___cxa_allocate_exception(16)|0); //@line 1626 "modules/graphics/opengl/Graphics.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($16|0),(435758|0),($vararg_buffer|0)); //@line 1626 "modules/graphics/opengl/Graphics.cpp"
     $17 = __THREW__; __THREW__ = 0;
     $18 = $17&1;
     if ($18) {
      $20 = ___cxa_find_matching_catch_2()|0;
      $21 = tempRet0;
      ___cxa_free_exception(($16|0)); //@line 1626 "modules/graphics/opengl/Graphics.cpp"
      $$0 = $21;$$01 = $20;
      break;
     } else {
      __THREW__ = 0;
      invoke_viii(3350,($16|0),(96|0),(344|0)); //@line 1626 "modules/graphics/opengl/Graphics.cpp"
      $19 = __THREW__; __THREW__ = 0;
      label = 3;
      break;
     }
    }
    $13 = ((($agg$result)) + 12|0); //@line 1624 "modules/graphics/opengl/Graphics.cpp"
    __THREW__ = 0;
    (invoke_iii(3360,($13|0),($9|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $14 = __THREW__; __THREW__ = 0;
    $15 = $14&1;
    if ($15) {
     label = 3;
    } else {
     $22 = HEAP32[165140]|0; //@line 74 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     $23 = (invoke_ii($22|0,7936)|0); //@line 74 "./libraries/glad/gladfuncs.hpp"
     $24 = __THREW__; __THREW__ = 0;
     $25 = $24&1;
     if ($25) {
      label = 3;
     } else {
      $26 = ($23|0)==(0|0); //@line 1629 "modules/graphics/opengl/Graphics.cpp"
      if ($26) {
       $30 = (___cxa_allocate_exception(16)|0); //@line 1632 "modules/graphics/opengl/Graphics.cpp"
       __THREW__ = 0;
       invoke_viii(3349,($30|0),(435804|0),($vararg_buffer1|0)); //@line 1632 "modules/graphics/opengl/Graphics.cpp"
       $31 = __THREW__; __THREW__ = 0;
       $32 = $31&1;
       if ($32) {
        $34 = ___cxa_find_matching_catch_2()|0;
        $35 = tempRet0;
        ___cxa_free_exception(($30|0)); //@line 1632 "modules/graphics/opengl/Graphics.cpp"
        $$0 = $35;$$01 = $34;
        break;
       } else {
        __THREW__ = 0;
        invoke_viii(3350,($30|0),(96|0),(344|0)); //@line 1632 "modules/graphics/opengl/Graphics.cpp"
        $33 = __THREW__; __THREW__ = 0;
        label = 3;
        break;
       }
      }
      $27 = ((($agg$result)) + 24|0); //@line 1630 "modules/graphics/opengl/Graphics.cpp"
      __THREW__ = 0;
      (invoke_iii(3360,($27|0),($23|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $28 = __THREW__; __THREW__ = 0;
      $29 = $28&1;
      if ($29) {
       label = 3;
      } else {
       $36 = HEAP32[165140]|0; //@line 74 "./libraries/glad/gladfuncs.hpp"
       __THREW__ = 0;
       $37 = (invoke_ii($36|0,7937)|0); //@line 74 "./libraries/glad/gladfuncs.hpp"
       $38 = __THREW__; __THREW__ = 0;
       $39 = $38&1;
       if ($39) {
        label = 3;
       } else {
        $40 = ($37|0)==(0|0); //@line 1635 "modules/graphics/opengl/Graphics.cpp"
        if (!($40)) {
         $41 = ((($agg$result)) + 36|0); //@line 1636 "modules/graphics/opengl/Graphics.cpp"
         __THREW__ = 0;
         (invoke_iii(3360,($41|0),($37|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $42 = __THREW__; __THREW__ = 0;
         $43 = $42&1;
         if ($43) {
          label = 3;
          break;
         }
         STACKTOP = sp;return; //@line 1641 "modules/graphics/opengl/Graphics.cpp"
        }
        $44 = (___cxa_allocate_exception(16)|0); //@line 1638 "modules/graphics/opengl/Graphics.cpp"
        __THREW__ = 0;
        invoke_viii(3349,($44|0),(435849|0),($vararg_buffer3|0)); //@line 1638 "modules/graphics/opengl/Graphics.cpp"
        $45 = __THREW__; __THREW__ = 0;
        $46 = $45&1;
        if ($46) {
         $48 = ___cxa_find_matching_catch_2()|0;
         $49 = tempRet0;
         ___cxa_free_exception(($44|0)); //@line 1638 "modules/graphics/opengl/Graphics.cpp"
         $$0 = $49;$$01 = $48;
         break;
        } else {
         __THREW__ = 0;
         invoke_viii(3350,($44|0),(96|0),(344|0)); //@line 1638 "modules/graphics/opengl/Graphics.cpp"
         $47 = __THREW__; __THREW__ = 0;
         label = 3;
         break;
        }
       }
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  $$0 = $5;$$01 = $4;
 }
 $50 = ((($agg$result)) + 36|0); //@line 181 "./modules/graphics/Graphics.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($50); //@line 181 "./modules/graphics/Graphics.h"
 $51 = ((($agg$result)) + 24|0); //@line 181 "./modules/graphics/Graphics.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($51); //@line 181 "./modules/graphics/Graphics.h"
 $52 = ((($agg$result)) + 12|0); //@line 181 "./modules/graphics/Graphics.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($52); //@line 181 "./modules/graphics/Graphics.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($agg$result); //@line 181 "./modules/graphics/Graphics.h"
 ___resumeException($$01|0);
 // unreachable;
}
function __ZNK4love8graphics6opengl8Graphics8getStatsEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(634072)>>2]|0; //@line 1647 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$agg$result>>2] = $0; //@line 1647 "modules/graphics/opengl/Graphics.cpp"
 $1 = HEAP32[(634076)>>2]|0; //@line 1648 "modules/graphics/opengl/Graphics.cpp"
 $2 = ((($agg$result)) + 4|0); //@line 1648 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$2>>2] = $1; //@line 1648 "modules/graphics/opengl/Graphics.cpp"
 $3 = HEAP32[158277]|0; //@line 1649 "modules/graphics/opengl/Graphics.cpp"
 $4 = ((($agg$result)) + 8|0); //@line 1649 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$4>>2] = $3; //@line 1649 "modules/graphics/opengl/Graphics.cpp"
 $5 = HEAP32[158431]|0; //@line 1650 "modules/graphics/opengl/Graphics.cpp"
 $6 = ((($agg$result)) + 12|0); //@line 1650 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$6>>2] = $5; //@line 1650 "modules/graphics/opengl/Graphics.cpp"
 $7 = HEAP32[158397]|0; //@line 1651 "modules/graphics/opengl/Graphics.cpp"
 $8 = ((($agg$result)) + 16|0); //@line 1651 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$8>>2] = $7; //@line 1651 "modules/graphics/opengl/Graphics.cpp"
 $9 = HEAP32[(634068)>>2]|0; //@line 1652 "modules/graphics/opengl/Graphics.cpp"
 $10 = ((($agg$result)) + 20|0); //@line 1652 "modules/graphics/opengl/Graphics.cpp"
 HEAP32[$10>>2] = $9; //@line 1652 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1654 "modules/graphics/opengl/Graphics.cpp"
}
function __ZNK4love8graphics6opengl8Graphics14getSystemLimitENS0_8Graphics11SystemLimitE($this,$limittype) {
 $this = $this|0;
 $limittype = $limittype|0;
 var $$0 = 0.0, $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0.0, $limits = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $limits = sp;
 switch ($limittype|0) {
 case 0:  {
  $0 = HEAP32[165575]|0; //@line 72 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$0 & 8191](33901,$limits); //@line 72 "./libraries/glad/gladfuncs.hpp"
  $1 = ((($limits)) + 4|0); //@line 1665 "modules/graphics/opengl/Graphics.cpp"
  $2 = +HEAPF32[$1>>2]; //@line 1665 "modules/graphics/opengl/Graphics.cpp"
  $3 = $2; //@line 1665 "modules/graphics/opengl/Graphics.cpp"
  $$0 = $3;
  break;
 }
 case 1:  {
  $4 = (__ZNK4love8graphics6opengl6OpenGL17getMaxTextureSizeEv(634044)|0); //@line 1667 "modules/graphics/opengl/Graphics.cpp"
  $5 = (+($4|0)); //@line 1667 "modules/graphics/opengl/Graphics.cpp"
  $$0 = $5;
  break;
 }
 case 2:  {
  $6 = (__ZNK4love8graphics6opengl6OpenGL19getMaxRenderTargetsEv(634044)|0); //@line 1669 "modules/graphics/opengl/Graphics.cpp"
  $7 = (+($6|0)); //@line 1669 "modules/graphics/opengl/Graphics.cpp"
  $$0 = $7;
  break;
 }
 case 3:  {
  $8 = (__ZNK4love8graphics6opengl6OpenGL25getMaxRenderbufferSamplesEv(634044)|0); //@line 1671 "modules/graphics/opengl/Graphics.cpp"
  $9 = (+($8|0)); //@line 1671 "modules/graphics/opengl/Graphics.cpp"
  $$0 = $9;
  break;
 }
 default: {
  $$0 = 0.0;
 }
 }
 STACKTOP = sp;return (+$$0); //@line 1675 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics4pushENS0_8Graphics9StackTypeE($this,$type) {
 $this = $this|0;
 $type = $type|0;
 var $$0$i$i$i$i$i = 0, $$01$i$i$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $9 = 0, $vararg_buffer = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 HEAP32[$0>>2] = $type;
 $1 = ((($this)) + 56|0); //@line 1694 "modules/graphics/opengl/Graphics.cpp"
 $2 = ((($this)) + 60|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5|0)==(256); //@line 1694 "modules/graphics/opengl/Graphics.cpp"
 if ($6) {
  $7 = (___cxa_allocate_exception(16)|0); //@line 1695 "modules/graphics/opengl/Graphics.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($7|0),(435894|0),($vararg_buffer|0)); //@line 1695 "modules/graphics/opengl/Graphics.cpp"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if ($9) {
   $10 = ___cxa_find_matching_catch_2()|0;
   $11 = tempRet0;
   ___cxa_free_exception(($7|0)); //@line 1695 "modules/graphics/opengl/Graphics.cpp"
   ___resumeException($10|0);
   // unreachable;
  } else {
   ___cxa_throw(($7|0),(96|0),(344|0)); //@line 1695 "modules/graphics/opengl/Graphics.cpp"
   // unreachable; //@line 1695 "modules/graphics/opengl/Graphics.cpp"
  }
 }
 __ZN4love8graphics6opengl6OpenGL13pushTransformEv(634044); //@line 1697 "modules/graphics/opengl/Graphics.cpp"
 $12 = ((($this)) + 20|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = HEAP32[$12>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = ((($13)) + -8|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = ((($this)) + 24|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $16 = HEAP32[$15>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ($13|0)==($16|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($17) {
  $20 = ((($this)) + 16|0); //@line 1699 "modules/graphics/opengl/Graphics.cpp"
  __ZNSt3__16vectorIdNS_9allocatorIdEEE21__push_back_slow_pathIRKdEEvOT_($20,$14); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 } else {
  $18 = +HEAPF64[$14>>3]; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAPF64[$13>>3] = $18; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $19 = ((($13)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$12>>2] = $19; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $21 = ($type|0)==(0); //@line 1701 "modules/graphics/opengl/Graphics.cpp"
 do {
  if ($21) {
   $22 = ((($this)) + 48|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $23 = HEAP32[$22>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $24 = ((($23)) + -136|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $25 = ((($this)) + 52|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $26 = HEAP32[$25>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $27 = ($23|0)==($26|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($27) {
    $74 = ((($this)) + 44|0); //@line 1702 "modules/graphics/opengl/Graphics.cpp"
    __ZNSt3__16vectorIN4love8graphics6opengl8Graphics12DisplayStateENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_($74,$24); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    break;
   }
   dest=$23; src=$24; stop=dest+84|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 482 "modules/graphics/opengl/Graphics.h"
   $28 = ((($23)) + 84|0); //@line 99 "./common/Object.h"
   $29 = ((($23)) + -52|0); //@line 139 "./common/Object.h"
   $30 = HEAP32[$29>>2]|0; //@line 139 "./common/Object.h"
   HEAP32[$28>>2] = $30; //@line 99 "./common/Object.h"
   $31 = ($30|0)==(0|0); //@line 101 "./common/Object.h"
   if (!($31)) {
    $32 = HEAP32[$30>>2]|0; //@line 101 "./common/Object.h"
    $33 = ((($32)) + 8|0); //@line 101 "./common/Object.h"
    $34 = HEAP32[$33>>2]|0; //@line 101 "./common/Object.h"
    FUNCTION_TABLE_vi[$34 & 8191]($30); //@line 101 "./common/Object.h"
   }
   $35 = ((($23)) + 88|0); //@line 99 "./common/Object.h"
   $36 = ((($23)) + -48|0); //@line 139 "./common/Object.h"
   $37 = HEAP32[$36>>2]|0; //@line 139 "./common/Object.h"
   HEAP32[$35>>2] = $37; //@line 99 "./common/Object.h"
   $38 = ($37|0)==(0|0); //@line 101 "./common/Object.h"
   if ($38) {
    label = 14;
   } else {
    $39 = HEAP32[$37>>2]|0; //@line 101 "./common/Object.h"
    $40 = ((($39)) + 8|0); //@line 101 "./common/Object.h"
    $41 = HEAP32[$40>>2]|0; //@line 101 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($41|0,($37|0)); //@line 101 "./common/Object.h"
    $42 = __THREW__; __THREW__ = 0;
    $43 = $42&1;
    if ($43) {
     $48 = ___cxa_find_matching_catch_2()|0;
     $49 = tempRet0;
     $$0$i$i$i$i$i = $49;$$01$i$i$i$i$i = $48;
    } else {
     label = 14;
    }
   }
   if ((label|0) == 14) {
    $44 = ((($23)) + 92|0); //@line 482 "modules/graphics/opengl/Graphics.h"
    $45 = ((($23)) + -44|0); //@line 482 "modules/graphics/opengl/Graphics.h"
    __THREW__ = 0;
    invoke_vii(4055,($44|0),($45|0)); //@line 482 "modules/graphics/opengl/Graphics.h"
    $46 = __THREW__; __THREW__ = 0;
    $47 = $46&1;
    if (!($47)) {
     $70 = ((($23)) + 104|0); //@line 482 "modules/graphics/opengl/Graphics.h"
     $71 = ((($23)) + -32|0); //@line 482 "modules/graphics/opengl/Graphics.h"
     ;HEAP32[$70>>2]=HEAP32[$71>>2]|0;HEAP32[$70+4>>2]=HEAP32[$71+4>>2]|0;HEAP32[$70+8>>2]=HEAP32[$71+8>>2]|0;HEAP32[$70+12>>2]=HEAP32[$71+12>>2]|0;HEAP32[$70+16>>2]=HEAP32[$71+16>>2]|0;HEAP32[$70+20>>2]=HEAP32[$71+20>>2]|0;HEAP32[$70+24>>2]=HEAP32[$71+24>>2]|0;HEAP32[$70+28>>2]=HEAP32[$71+28>>2]|0; //@line 482 "modules/graphics/opengl/Graphics.h"
     $72 = HEAP32[$22>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $73 = ((($72)) + 136|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$22>>2] = $73; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     break;
    }
    $50 = ___cxa_find_matching_catch_2()|0;
    $51 = tempRet0;
    $52 = HEAP32[$35>>2]|0; //@line 106 "./common/Object.h"
    $53 = ($52|0)==(0|0); //@line 106 "./common/Object.h"
    if ($53) {
     $$0$i$i$i$i$i = $51;$$01$i$i$i$i$i = $50;
    } else {
     $54 = HEAP32[$52>>2]|0; //@line 106 "./common/Object.h"
     $55 = ((($54)) + 12|0); //@line 106 "./common/Object.h"
     $56 = HEAP32[$55>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($56|0,($52|0)); //@line 106 "./common/Object.h"
     $57 = __THREW__; __THREW__ = 0;
     $58 = $57&1;
     if ($58) {
      $59 = ___cxa_find_matching_catch_3(0|0)|0;
      $60 = tempRet0;
      ___clang_call_terminate($59); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     } else {
      $$0$i$i$i$i$i = $51;$$01$i$i$i$i$i = $50;
     }
    }
   }
   $61 = HEAP32[$28>>2]|0; //@line 106 "./common/Object.h"
   $62 = ($61|0)==(0|0); //@line 106 "./common/Object.h"
   if ($62) {
    ___resumeException($$01$i$i$i$i$i|0);
    // unreachable;
   }
   $63 = HEAP32[$61>>2]|0; //@line 106 "./common/Object.h"
   $64 = ((($63)) + 12|0); //@line 106 "./common/Object.h"
   $65 = HEAP32[$64>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($65|0,($61|0)); //@line 106 "./common/Object.h"
   $66 = __THREW__; __THREW__ = 0;
   $67 = $66&1;
   if ($67) {
    $68 = ___cxa_find_matching_catch_3(0|0)|0;
    $69 = tempRet0;
    ___clang_call_terminate($68); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   } else {
    ___resumeException($$01$i$i$i$i$i|0);
    // unreachable;
   }
  }
 } while(0);
 $75 = HEAP32[$2>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $76 = ((($this)) + 64|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $77 = HEAP32[$76>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $78 = ($75|0)==($77|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($78) {
  __ZNSt3__16vectorIN4love8graphics8Graphics9StackTypeENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_($1,$0); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  STACKTOP = sp;return; //@line 1705 "modules/graphics/opengl/Graphics.cpp"
 } else {
  HEAP32[$75>>2] = $type; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $79 = ((($75)) + 4|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$2>>2] = $79; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  STACKTOP = sp;return; //@line 1705 "modules/graphics/opengl/Graphics.cpp"
 }
}
function __ZNSt3__16vectorIdNS_9allocatorIdEEE21__push_back_slow_pathIRKdEEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(536870911); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 3; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(268435455); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 2; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 536870911;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 3; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<3)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = +HEAPF64[$__x>>3]; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAPF64[$24>>3] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 8|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 3; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<3)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNSt3__16vectorIN4love8graphics6opengl8Graphics12DisplayStateENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i = 0, $$0$i$i$i$i = 0, $$01$i$i$i$i = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $9 = 0, $__v = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__v = sp;
 $0 = ((($this)) + 8|0); //@line 350 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($2) - ($3))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4|0) / 136)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = (($5) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ($6>>>0)>(31580641); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($7) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $11 = $$pre;
 } else {
  $11 = $3;
 }
 $8 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = HEAP32[$8>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9) - ($11))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = (($10|0) / 136)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ($12>>>0)<(15790320); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($13) {
  $14 = $12 << 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = ($14>>>0)<($6>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $16 = $15 ? $6 : $14; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i = $16;
 } else {
  $$0$i = 31580641;
 }
 $17 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $18 = (($17) - ($11))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $19 = (($18|0) / 136)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = ((($__v)) + 12|0); //@line 323 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 HEAP32[$20>>2] = 0; //@line 1959 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $21 = ((($__v)) + 16|0); //@line 1959 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$21>>2] = $0; //@line 1959 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $22 = ($$0$i|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($22) {
  $25 = 0;
 } else {
  $23 = ($$0$i*136)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = (__Znwj($23)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $25 = $24;
 }
 HEAP32[$__v>>2] = $25; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $26 = (($25) + (($19*136)|0)|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = ((($__v)) + 8|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 HEAP32[$27>>2] = $26; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = ((($__v)) + 4|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 HEAP32[$28>>2] = $26; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = (($25) + (($$0$i*136)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = ((($__v)) + 12|0); //@line 2014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$30>>2] = $29; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 dest=$26; src=$__x; stop=dest+84|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 482 "modules/graphics/opengl/Graphics.h"
 $31 = (((($25) + (($19*136)|0)|0)) + 84|0); //@line 99 "./common/Object.h"
 $32 = ((($__x)) + 84|0); //@line 139 "./common/Object.h"
 $33 = HEAP32[$32>>2]|0; //@line 139 "./common/Object.h"
 HEAP32[$31>>2] = $33; //@line 99 "./common/Object.h"
 $34 = ($33|0)==(0|0); //@line 101 "./common/Object.h"
 if ($34) {
  label = 9;
 } else {
  $35 = HEAP32[$33>>2]|0; //@line 101 "./common/Object.h"
  $36 = ((($35)) + 8|0); //@line 101 "./common/Object.h"
  $37 = HEAP32[$36>>2]|0; //@line 101 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($37|0,($33|0)); //@line 101 "./common/Object.h"
  $38 = __THREW__; __THREW__ = 0;
  $39 = $38&1;
  if (!($39)) {
   label = 9;
  }
 }
 do {
  if ((label|0) == 9) {
   $40 = (((($25) + (($19*136)|0)|0)) + 88|0); //@line 99 "./common/Object.h"
   $41 = ((($__x)) + 88|0); //@line 139 "./common/Object.h"
   $42 = HEAP32[$41>>2]|0; //@line 139 "./common/Object.h"
   HEAP32[$40>>2] = $42; //@line 99 "./common/Object.h"
   $43 = ($42|0)==(0|0); //@line 101 "./common/Object.h"
   if ($43) {
    label = 11;
   } else {
    $44 = HEAP32[$42>>2]|0; //@line 101 "./common/Object.h"
    $45 = ((($44)) + 8|0); //@line 101 "./common/Object.h"
    $46 = HEAP32[$45>>2]|0; //@line 101 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($46|0,($42|0)); //@line 101 "./common/Object.h"
    $47 = __THREW__; __THREW__ = 0;
    $48 = $47&1;
    if ($48) {
     $53 = ___cxa_find_matching_catch_2()|0;
     $54 = tempRet0;
     $$0$i$i$i$i = $54;$$01$i$i$i$i = $53;
    } else {
     label = 11;
    }
   }
   if ((label|0) == 11) {
    $49 = (((($25) + (($19*136)|0)|0)) + 92|0); //@line 482 "modules/graphics/opengl/Graphics.h"
    $50 = ((($__x)) + 92|0); //@line 482 "modules/graphics/opengl/Graphics.h"
    __THREW__ = 0;
    invoke_vii(4055,($49|0),($50|0)); //@line 482 "modules/graphics/opengl/Graphics.h"
    $51 = __THREW__; __THREW__ = 0;
    $52 = $51&1;
    if (!($52)) {
     $75 = (((($25) + (($19*136)|0)|0)) + 104|0); //@line 482 "modules/graphics/opengl/Graphics.h"
     $76 = ((($__x)) + 104|0); //@line 482 "modules/graphics/opengl/Graphics.h"
     ;HEAP32[$75>>2]=HEAP32[$76>>2]|0;HEAP32[$75+4>>2]=HEAP32[$76+4>>2]|0;HEAP32[$75+8>>2]=HEAP32[$76+8>>2]|0;HEAP32[$75+12>>2]=HEAP32[$76+12>>2]|0;HEAP32[$75+16>>2]=HEAP32[$76+16>>2]|0;HEAP32[$75+20>>2]=HEAP32[$76+20>>2]|0;HEAP32[$75+24>>2]=HEAP32[$76+24>>2]|0;HEAP32[$75+28>>2]=HEAP32[$76+28>>2]|0; //@line 482 "modules/graphics/opengl/Graphics.h"
     $77 = ((($26)) + 136|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$27>>2] = $77; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     __THREW__ = 0;
     invoke_vii(4053,($this|0),($__v|0)); //@line 1583 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $78 = __THREW__; __THREW__ = 0;
     $79 = $78&1;
     if ($79) {
      break;
     }
     __ZNSt3__114__split_bufferIN4love8graphics6opengl8Graphics12DisplayStateERNS_9allocatorIS5_EEED2Ev($__v); //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     STACKTOP = sp;return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    $55 = ___cxa_find_matching_catch_2()|0;
    $56 = tempRet0;
    $57 = HEAP32[$40>>2]|0; //@line 106 "./common/Object.h"
    $58 = ($57|0)==(0|0); //@line 106 "./common/Object.h"
    if ($58) {
     $$0$i$i$i$i = $56;$$01$i$i$i$i = $55;
    } else {
     $59 = HEAP32[$57>>2]|0; //@line 106 "./common/Object.h"
     $60 = ((($59)) + 12|0); //@line 106 "./common/Object.h"
     $61 = HEAP32[$60>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($61|0,($57|0)); //@line 106 "./common/Object.h"
     $62 = __THREW__; __THREW__ = 0;
     $63 = $62&1;
     if ($63) {
      $64 = ___cxa_find_matching_catch_3(0|0)|0;
      $65 = tempRet0;
      ___clang_call_terminate($64); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     } else {
      $$0$i$i$i$i = $56;$$01$i$i$i$i = $55;
     }
    }
   }
   $66 = HEAP32[$31>>2]|0; //@line 106 "./common/Object.h"
   $67 = ($66|0)==(0|0); //@line 106 "./common/Object.h"
   if (!($67)) {
    $68 = HEAP32[$66>>2]|0; //@line 106 "./common/Object.h"
    $69 = ((($68)) + 12|0); //@line 106 "./common/Object.h"
    $70 = HEAP32[$69>>2]|0; //@line 106 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($70|0,($66|0)); //@line 106 "./common/Object.h"
    $71 = __THREW__; __THREW__ = 0;
    $72 = $71&1;
    if ($72) {
     $73 = ___cxa_find_matching_catch_3(0|0)|0;
     $74 = tempRet0;
     ___clang_call_terminate($73); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   }
   $eh$lpad$body$index2Z2D = $$0$i$i$i$i;$eh$lpad$body$indexZ2D = $$01$i$i$i$i;
   __ZNSt3__114__split_bufferIN4love8graphics6opengl8Graphics12DisplayStateERNS_9allocatorIS5_EEED2Ev($__v); //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   ___resumeException($eh$lpad$body$indexZ2D|0);
   // unreachable;
  }
 } while(0);
 $80 = ___cxa_find_matching_catch_2()|0;
 $81 = tempRet0;
 $eh$lpad$body$index2Z2D = $81;$eh$lpad$body$indexZ2D = $80;
 __ZNSt3__114__split_bufferIN4love8graphics6opengl8Graphics12DisplayStateERNS_9allocatorIS5_EEED2Ev($__v); //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love8graphics8Graphics9StackTypeENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl8Graphics3popEv($this) {
 $this = $this|0;
 var $$cast = 0, $$pre = 0, $$pre$i$i$i$i$i$i$i$i = 0, $$pre$i$i$i$i$i$i$i$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $8 = 0, $9 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 56|0); //@line 1709 "modules/graphics/opengl/Graphics.cpp"
 $1 = ((($this)) + 60|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($2|0)==($3|0); //@line 1709 "modules/graphics/opengl/Graphics.cpp"
 if ($4) {
  $5 = (___cxa_allocate_exception(16)|0); //@line 1710 "modules/graphics/opengl/Graphics.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($5|0),(435947|0),($vararg_buffer|0)); //@line 1710 "modules/graphics/opengl/Graphics.cpp"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if ($7) {
   $8 = ___cxa_find_matching_catch_2()|0;
   $9 = tempRet0;
   ___cxa_free_exception(($5|0)); //@line 1710 "modules/graphics/opengl/Graphics.cpp"
   ___resumeException($8|0);
   // unreachable;
  } else {
   ___cxa_throw(($5|0),(96|0),(344|0)); //@line 1710 "modules/graphics/opengl/Graphics.cpp"
   // unreachable; //@line 1710 "modules/graphics/opengl/Graphics.cpp"
  }
 }
 __ZN4love8graphics6opengl6OpenGL12popTransformEv(634044); //@line 1712 "modules/graphics/opengl/Graphics.cpp"
 $10 = ((($this)) + 20|0); //@line 1666 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = HEAP32[$10>>2]|0; //@line 1666 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $scevgep4$i$i$i1 = ((($11)) + -8|0);
 HEAP32[$10>>2] = $scevgep4$i$i$i1; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = HEAP32[$1>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ((($12)) + -4|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = HEAP32[$13>>2]|0; //@line 1715 "modules/graphics/opengl/Graphics.cpp"
 $15 = ($14|0)==(0); //@line 1715 "modules/graphics/opengl/Graphics.cpp"
 if (!($15)) {
  $70 = $12;
  $scevgep4$i$i$i = ((($70)) + -4|0);
  HEAP32[$1>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  STACKTOP = sp;return; //@line 1726 "modules/graphics/opengl/Graphics.cpp"
 }
 $16 = ((($this)) + 44|0); //@line 1717 "modules/graphics/opengl/Graphics.cpp"
 $17 = ((($this)) + 48|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $18 = HEAP32[$17>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $19 = HEAP32[$16>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = (($18) - ($19))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = (($20|0) / 136)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $22 = (($21) + -2)|0; //@line 1717 "modules/graphics/opengl/Graphics.cpp"
 $$cast = $19; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $23 = (($$cast) + (($22*136)|0)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 __ZN4love8graphics6opengl8Graphics19restoreStateCheckedERKNS2_12DisplayStateE($this,$23); //@line 1719 "modules/graphics/opengl/Graphics.cpp"
 $24 = HEAP32[$17>>2]|0; //@line 1666 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $25 = ((($24)) + -136|0); //@line 1666 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = $24;
 L10: while(1) {
  $26 = ((($27)) + -136|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$17>>2] = $26; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $28 = ((($27)) + -44|0); //@line 482 "modules/graphics/opengl/Graphics.h"
  $29 = HEAP32[$28>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $30 = ($29|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($30)) {
   $31 = ((($27)) + -40|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $32 = HEAP32[$31>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $33 = ($32|0)==($29|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($33) {
    $47 = $29;
   } else {
    $35 = $32;
    while(1) {
     $34 = ((($35)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$31>>2] = $34; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $36 = HEAP32[$34>>2]|0; //@line 106 "./common/Object.h"
     $37 = ($36|0)==(0|0); //@line 106 "./common/Object.h"
     if ($37) {
      $39 = $34;
     } else {
      $40 = HEAP32[$36>>2]|0; //@line 106 "./common/Object.h"
      $41 = ((($40)) + 12|0); //@line 106 "./common/Object.h"
      $42 = HEAP32[$41>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($42|0,($36|0)); //@line 106 "./common/Object.h"
      $43 = __THREW__; __THREW__ = 0;
      $44 = $43&1;
      if ($44) {
       label = 13;
       break L10;
      }
      $$pre$i$i$i$i$i$i$i$i$i$i = HEAP32[$31>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $39 = $$pre$i$i$i$i$i$i$i$i$i$i;
     }
     $38 = ($39|0)==($29|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($38) {
      break;
     } else {
      $35 = $39;
     }
    }
    $$pre$i$i$i$i$i$i$i$i = HEAP32[$28>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $47 = $$pre$i$i$i$i$i$i$i$i;
   }
   __ZdlPv($47); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  $48 = ((($27)) + -48|0); //@line 106 "./common/Object.h"
  $49 = HEAP32[$48>>2]|0; //@line 106 "./common/Object.h"
  $50 = ($49|0)==(0|0); //@line 106 "./common/Object.h"
  if (!($50)) {
   $51 = HEAP32[$49>>2]|0; //@line 106 "./common/Object.h"
   $52 = ((($51)) + 12|0); //@line 106 "./common/Object.h"
   $53 = HEAP32[$52>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($53|0,($49|0)); //@line 106 "./common/Object.h"
   $54 = __THREW__; __THREW__ = 0;
   $55 = $54&1;
   if ($55) {
    label = 18;
    break;
   }
  }
  $58 = ((($27)) + -52|0); //@line 106 "./common/Object.h"
  $59 = HEAP32[$58>>2]|0; //@line 106 "./common/Object.h"
  $60 = ($59|0)==(0|0); //@line 106 "./common/Object.h"
  if (!($60)) {
   $63 = HEAP32[$59>>2]|0; //@line 106 "./common/Object.h"
   $64 = ((($63)) + 12|0); //@line 106 "./common/Object.h"
   $65 = HEAP32[$64>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($65|0,($59|0)); //@line 106 "./common/Object.h"
   $66 = __THREW__; __THREW__ = 0;
   $67 = $66&1;
   if ($67) {
    label = 22;
    break;
   }
  }
  $61 = HEAP32[$17>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $62 = ($61|0)==($25|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($62) {
   label = 23;
   break;
  } else {
   $27 = $61;
  }
 }
 if ((label|0) == 13) {
  $45 = ___cxa_find_matching_catch_3(0|0)|0;
  $46 = tempRet0;
  ___clang_call_terminate($45); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 }
 else if ((label|0) == 18) {
  $56 = ___cxa_find_matching_catch_3(0|0)|0;
  $57 = tempRet0;
  ___clang_call_terminate($56); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 }
 else if ((label|0) == 22) {
  $68 = ___cxa_find_matching_catch_3(0|0)|0;
  $69 = tempRet0;
  ___clang_call_terminate($68); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 }
 else if ((label|0) == 23) {
  $$pre = HEAP32[$1>>2]|0; //@line 1666 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $70 = $$pre;
  $scevgep4$i$i$i = ((($70)) + -4|0);
  HEAP32[$1>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  STACKTOP = sp;return; //@line 1726 "modules/graphics/opengl/Graphics.cpp"
 }
}
function __ZN4love8graphics6opengl8Graphics6rotateEf($this,$r) {
 $this = $this|0;
 $r = +$r;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love8graphics6opengl6OpenGL12getTransformEv(634044)|0); //@line 1730 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love7Matrix46rotateEf($0,$r); //@line 1730 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1731 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics5scaleEff($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0.0, $fabsf = 0.0, $fabsf1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love8graphics6opengl6OpenGL12getTransformEv(634044)|0); //@line 1735 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love7Matrix45scaleEff($0,$x,$y); //@line 1735 "modules/graphics/opengl/Graphics.cpp"
 $fabsf = (+Math_abs((+$x))); //@line 1736 "modules/graphics/opengl/Graphics.cpp"
 $1 = $fabsf; //@line 1736 "modules/graphics/opengl/Graphics.cpp"
 $fabsf1 = (+Math_abs((+$y))); //@line 1736 "modules/graphics/opengl/Graphics.cpp"
 $2 = $fabsf1; //@line 1736 "modules/graphics/opengl/Graphics.cpp"
 $3 = $1 + $2; //@line 1736 "modules/graphics/opengl/Graphics.cpp"
 $4 = 2.0 / $3; //@line 1736 "modules/graphics/opengl/Graphics.cpp"
 $5 = ((($this)) + 20|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$5>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ((($6)) + -8|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = +HEAPF64[$7>>3]; //@line 1736 "modules/graphics/opengl/Graphics.cpp"
 $9 = $8 * $4; //@line 1736 "modules/graphics/opengl/Graphics.cpp"
 HEAPF64[$7>>3] = $9; //@line 1736 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1737 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics9translateEff($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love8graphics6opengl6OpenGL12getTransformEv(634044)|0); //@line 1741 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love7Matrix49translateEff($0,$x,$y); //@line 1741 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1742 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl8Graphics5shearEff($this,$kx,$ky) {
 $this = $this|0;
 $kx = +$kx;
 $ky = +$ky;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love8graphics6opengl6OpenGL12getTransformEv(634044)|0); //@line 1746 "modules/graphics/opengl/Graphics.cpp"
 __ZN4love7Matrix45shearEff($0,$kx,$ky); //@line 1746 "modules/graphics/opengl/Graphics.cpp"
 return; //@line 1747 "modules/graphics/opengl/Graphics.cpp"
}
function __ZN4love8graphics6opengl5ImageD2Ev($this) {
 $this = $this|0;
 var $$pre$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i17 = 0, $$pre$i$i$i3 = 0, $$pre$i$i$i8 = 0, $$pre$i11 = 0, $$pre$i20 = 0, $$pre$i5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (24820); //@line 153 "modules/graphics/opengl/Image.cpp"
 $0 = ((($this)) + 120|0); //@line 153 "modules/graphics/opengl/Image.cpp"
 HEAP32[$0>>2] = (24892); //@line 153 "modules/graphics/opengl/Image.cpp"
 $1 = ((($this)) + 148|0); //@line 383 "modules/graphics/opengl/Image.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 383 "modules/graphics/opengl/Image.cpp"
 $3 = ($2|0)==(0); //@line 383 "modules/graphics/opengl/Image.cpp"
 do {
  if (!($3)) {
   __THREW__ = 0;
   invoke_vii(4014,(634044|0),($2|0)); //@line 386 "modules/graphics/opengl/Image.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   if (!($5)) {
    HEAP32[$1>>2] = 0; //@line 387 "modules/graphics/opengl/Image.cpp"
    $6 = ((($this)) + 164|0); //@line 389 "modules/graphics/opengl/Image.cpp"
    $7 = HEAP32[$6>>2]|0; //@line 389 "modules/graphics/opengl/Image.cpp"
    __THREW__ = 0;
    invoke_viii(4015,(634044|0),($7|0),0); //@line 389 "modules/graphics/opengl/Image.cpp"
    $8 = __THREW__; __THREW__ = 0;
    $9 = $8&1;
    if (!($9)) {
     HEAP32[$6>>2] = 0; //@line 390 "modules/graphics/opengl/Image.cpp"
     break;
    }
   }
   $52 = ___cxa_find_matching_catch_3(0|0)|0;
   $53 = tempRet0;
   $54 = ((($this)) + 136|0); //@line 156 "modules/graphics/opengl/Image.cpp"
   $55 = HEAP32[$54>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $56 = ($55|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($56)) {
    $57 = ((($this)) + 140|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $58 = HEAP32[$57>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $59 = ($58|0)==($55|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    do {
     if ($59) {
      $73 = $55;
     } else {
      $61 = $58;
      while(1) {
       $60 = ((($61)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$57>>2] = $60; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $62 = HEAP32[$60>>2]|0; //@line 106 "./common/Object.h"
       $63 = ($62|0)==(0|0); //@line 106 "./common/Object.h"
       if ($63) {
        $65 = $60;
       } else {
        $66 = HEAP32[$62>>2]|0; //@line 106 "./common/Object.h"
        $67 = ((($66)) + 12|0); //@line 106 "./common/Object.h"
        $68 = HEAP32[$67>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($68|0,($62|0)); //@line 106 "./common/Object.h"
        $69 = __THREW__; __THREW__ = 0;
        $70 = $69&1;
        if ($70) {
         label = 30;
         break;
        }
        $$pre$i$i$i8 = HEAP32[$57>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $65 = $$pre$i$i$i8;
       }
       $64 = ($65|0)==($55|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($64) {
        label = 31;
        break;
       } else {
        $61 = $65;
       }
      }
      if ((label|0) == 30) {
       $71 = ___cxa_find_matching_catch_3(0|0)|0;
       $72 = tempRet0;
       ___clang_call_terminate($71); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      else if ((label|0) == 31) {
       $$pre$i11 = HEAP32[$54>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $73 = $$pre$i11;
       break;
      }
     }
    } while(0);
    __ZdlPv($73); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $74 = ((($this)) + 124|0); //@line 156 "modules/graphics/opengl/Image.cpp"
   $75 = HEAP32[$74>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $76 = ($75|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($76) {
    __ZN4love8graphics8VolatileD2Ev($0); //@line 156 "modules/graphics/opengl/Image.cpp"
    __ZN4love8graphics7TextureD2Ev($this); //@line 156 "modules/graphics/opengl/Image.cpp"
    ___clang_call_terminate($52); //@line 156 "modules/graphics/opengl/Image.cpp"
    // unreachable; //@line 156 "modules/graphics/opengl/Image.cpp"
   }
   $77 = ((($this)) + 128|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $78 = HEAP32[$77>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $79 = ($78|0)==($75|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   do {
    if ($79) {
     $93 = $75;
    } else {
     $81 = $78;
     while(1) {
      $80 = ((($81)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$77>>2] = $80; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $82 = HEAP32[$80>>2]|0; //@line 106 "./common/Object.h"
      $83 = ($82|0)==(0|0); //@line 106 "./common/Object.h"
      if ($83) {
       $85 = $80;
      } else {
       $86 = HEAP32[$82>>2]|0; //@line 106 "./common/Object.h"
       $87 = ((($86)) + 12|0); //@line 106 "./common/Object.h"
       $88 = HEAP32[$87>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($88|0,($82|0)); //@line 106 "./common/Object.h"
       $89 = __THREW__; __THREW__ = 0;
       $90 = $89&1;
       if ($90) {
        label = 39;
        break;
       }
       $$pre$i$i$i17 = HEAP32[$77>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $85 = $$pre$i$i$i17;
      }
      $84 = ($85|0)==($75|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($84) {
       label = 40;
       break;
      } else {
       $81 = $85;
      }
     }
     if ((label|0) == 39) {
      $91 = ___cxa_find_matching_catch_3(0|0)|0;
      $92 = tempRet0;
      ___clang_call_terminate($91); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
     else if ((label|0) == 40) {
      $$pre$i20 = HEAP32[$74>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $93 = $$pre$i20;
      break;
     }
    }
   } while(0);
   __ZdlPv($93); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   __ZN4love8graphics8VolatileD2Ev($0); //@line 156 "modules/graphics/opengl/Image.cpp"
   __ZN4love8graphics7TextureD2Ev($this); //@line 156 "modules/graphics/opengl/Image.cpp"
   ___clang_call_terminate($52); //@line 156 "modules/graphics/opengl/Image.cpp"
   // unreachable; //@line 156 "modules/graphics/opengl/Image.cpp"
  }
 } while(0);
 $10 = HEAP32[158431]|0; //@line 155 "modules/graphics/opengl/Image.cpp"
 $11 = (($10) + -1)|0; //@line 155 "modules/graphics/opengl/Image.cpp"
 HEAP32[158431] = $11; //@line 155 "modules/graphics/opengl/Image.cpp"
 $12 = ((($this)) + 136|0); //@line 156 "modules/graphics/opengl/Image.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = ($13|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($14)) {
  $15 = ((($this)) + 140|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = HEAP32[$15>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = ($16|0)==($13|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  do {
   if ($17) {
    $31 = $13;
   } else {
    $19 = $16;
    while(1) {
     $18 = ((($19)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$15>>2] = $18; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $20 = HEAP32[$18>>2]|0; //@line 106 "./common/Object.h"
     $21 = ($20|0)==(0|0); //@line 106 "./common/Object.h"
     if ($21) {
      $23 = $18;
     } else {
      $24 = HEAP32[$20>>2]|0; //@line 106 "./common/Object.h"
      $25 = ((($24)) + 12|0); //@line 106 "./common/Object.h"
      $26 = HEAP32[$25>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($26|0,($20|0)); //@line 106 "./common/Object.h"
      $27 = __THREW__; __THREW__ = 0;
      $28 = $27&1;
      if ($28) {
       label = 11;
       break;
      }
      $$pre$i$i$i = HEAP32[$15>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $23 = $$pre$i$i$i;
     }
     $22 = ($23|0)==($13|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($22) {
      label = 12;
      break;
     } else {
      $19 = $23;
     }
    }
    if ((label|0) == 11) {
     $29 = ___cxa_find_matching_catch_3(0|0)|0;
     $30 = tempRet0;
     ___clang_call_terminate($29); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
    else if ((label|0) == 12) {
     $$pre$i = HEAP32[$12>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $31 = $$pre$i;
     break;
    }
   }
  } while(0);
  __ZdlPv($31); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $32 = ((($this)) + 124|0); //@line 156 "modules/graphics/opengl/Image.cpp"
 $33 = HEAP32[$32>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $34 = ($33|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($34) {
  __ZN4love8graphics8VolatileD2Ev($0); //@line 156 "modules/graphics/opengl/Image.cpp"
  __ZN4love8graphics7TextureD2Ev($this); //@line 156 "modules/graphics/opengl/Image.cpp"
  return; //@line 156 "modules/graphics/opengl/Image.cpp"
 }
 $35 = ((($this)) + 128|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $36 = HEAP32[$35>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $37 = ($36|0)==($33|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($37) {
   $51 = $33;
  } else {
   $39 = $36;
   while(1) {
    $38 = ((($39)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$35>>2] = $38; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $40 = HEAP32[$38>>2]|0; //@line 106 "./common/Object.h"
    $41 = ($40|0)==(0|0); //@line 106 "./common/Object.h"
    if ($41) {
     $43 = $38;
    } else {
     $44 = HEAP32[$40>>2]|0; //@line 106 "./common/Object.h"
     $45 = ((($44)) + 12|0); //@line 106 "./common/Object.h"
     $46 = HEAP32[$45>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($46|0,($40|0)); //@line 106 "./common/Object.h"
     $47 = __THREW__; __THREW__ = 0;
     $48 = $47&1;
     if ($48) {
      label = 20;
      break;
     }
     $$pre$i$i$i3 = HEAP32[$35>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $43 = $$pre$i$i$i3;
    }
    $42 = ($43|0)==($33|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($42) {
     label = 21;
     break;
    } else {
     $39 = $43;
    }
   }
   if ((label|0) == 20) {
    $49 = ___cxa_find_matching_catch_3(0|0)|0;
    $50 = tempRet0;
    ___clang_call_terminate($49); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
   else if ((label|0) == 21) {
    $$pre$i5 = HEAP32[$32>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $51 = $$pre$i5;
    break;
   }
  }
 } while(0);
 __ZdlPv($51); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 __ZN4love8graphics8VolatileD2Ev($0); //@line 156 "modules/graphics/opengl/Image.cpp"
 __ZN4love8graphics7TextureD2Ev($this); //@line 156 "modules/graphics/opengl/Image.cpp"
 return; //@line 156 "modules/graphics/opengl/Image.cpp"
}
function __ZN4love8graphics6opengl5ImageD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl5ImageD2Ev($this); //@line 153 "modules/graphics/opengl/Image.cpp"
 __ZdlPv($this); //@line 153 "modules/graphics/opengl/Image.cpp"
 return; //@line 156 "modules/graphics/opengl/Image.cpp"
}
function __ZN4love8graphics6opengl5Image4drawEfffffffff($this,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $t = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $t = sp;
 __ZN4love7Matrix4C2Efffffffff($t,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky); //@line 466 "modules/graphics/opengl/Image.cpp"
 $0 = ((($this)) + 40|0); //@line 468 "modules/graphics/opengl/Image.cpp"
 __THREW__ = 0;
 invoke_viii(4077,($this|0),($t|0),($0|0)); //@line 468 "modules/graphics/opengl/Image.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZN4love7Matrix4D2Ev($t); //@line 469 "modules/graphics/opengl/Image.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  __ZN4love7Matrix4D2Ev($t); //@line 469 "modules/graphics/opengl/Image.cpp"
  STACKTOP = sp;return; //@line 469 "modules/graphics/opengl/Image.cpp"
 }
}
function __ZN4love8graphics6opengl5Image5drawvERKNS_7Matrix4EPKNS_6VertexE($this,$t,$v) {
 $this = $this|0;
 $t = $t|0;
 $v = $v|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl6OpenGL13pushTransformEv(634044); //@line 125 "modules/graphics/opengl/OpenGL.h"
 __THREW__ = 0;
 $0 = (invoke_ii(4003,(634044|0))|0); //@line 135 "modules/graphics/opengl/OpenGL.h"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if (!($2)) {
  __THREW__ = 0;
  invoke_vii(4004,($0|0),($t|0)); //@line 451 "modules/graphics/opengl/Image.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if (!($4)) {
   $5 = ((($this)) + 148|0); //@line 453 "modules/graphics/opengl/Image.cpp"
   $6 = HEAP32[$5>>2]|0; //@line 453 "modules/graphics/opengl/Image.cpp"
   __THREW__ = 0;
   invoke_vii(4005,(634044|0),($6|0)); //@line 453 "modules/graphics/opengl/Image.cpp"
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if (!($8)) {
    __THREW__ = 0;
    invoke_vii(4006,(634044|0),3); //@line 455 "modules/graphics/opengl/Image.cpp"
    $9 = __THREW__; __THREW__ = 0;
    $10 = $9&1;
    if (!($10)) {
     $11 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_viiiiii($11|0,0,2,5126,0,20,($v|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
     $12 = __THREW__; __THREW__ = 0;
     $13 = $12&1;
     if (!($13)) {
      $14 = ((($v)) + 8|0); //@line 458 "modules/graphics/opengl/Image.cpp"
      $15 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_viiiiii($15|0,1,2,5126,0,20,($14|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
      $16 = __THREW__; __THREW__ = 0;
      $17 = $16&1;
      if (!($17)) {
       __THREW__ = 0;
       invoke_vi(4007,(634044|0)); //@line 460 "modules/graphics/opengl/Image.cpp"
       $18 = __THREW__; __THREW__ = 0;
       $19 = $18&1;
       if (!($19)) {
        __THREW__ = 0;
        invoke_viiii(4008,(634044|0),5,0,4); //@line 461 "modules/graphics/opengl/Image.cpp"
        $20 = __THREW__; __THREW__ = 0;
        $21 = $20&1;
        if (!($21)) {
         __THREW__ = 0;
         invoke_vi(4009,(634044|0)); //@line 130 "modules/graphics/opengl/OpenGL.h"
         $22 = __THREW__; __THREW__ = 0;
         $23 = $22&1;
         if ($23) {
          $24 = ___cxa_find_matching_catch_3(0|0)|0;
          $25 = tempRet0;
          ___clang_call_terminate($24); //@line 130 "modules/graphics/opengl/OpenGL.h"
          // unreachable; //@line 130 "modules/graphics/opengl/OpenGL.h"
         } else {
          return; //@line 462 "modules/graphics/opengl/Image.cpp"
         }
        }
       }
      }
     }
    }
   }
  }
 }
 $26 = ___cxa_find_matching_catch_2()|0;
 $27 = tempRet0;
 __THREW__ = 0;
 invoke_vi(4009,(634044|0)); //@line 130 "modules/graphics/opengl/OpenGL.h"
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if ($29) {
  $30 = ___cxa_find_matching_catch_3(0|0)|0;
  $31 = tempRet0;
  ___clang_call_terminate($30); //@line 130 "modules/graphics/opengl/OpenGL.h"
  // unreachable; //@line 130 "modules/graphics/opengl/OpenGL.h"
 } else {
  ___resumeException($26|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl5Image5drawqEPNS0_4QuadEfffffffff($this,$quad,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $quad = $quad|0;
 $x = +$x;
 $y = +$y;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $t = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $t = sp;
 __ZN4love7Matrix4C2Efffffffff($t,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky); //@line 473 "modules/graphics/opengl/Image.cpp"
 __THREW__ = 0;
 $0 = (invoke_ii(4010,($quad|0))|0); //@line 475 "modules/graphics/opengl/Image.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if (!($2)) {
  __THREW__ = 0;
  invoke_viii(4077,($this|0),($t|0),($0|0)); //@line 475 "modules/graphics/opengl/Image.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if (!($4)) {
   __ZN4love7Matrix4D2Ev($t); //@line 476 "modules/graphics/opengl/Image.cpp"
   STACKTOP = sp;return; //@line 476 "modules/graphics/opengl/Image.cpp"
  }
 }
 $5 = ___cxa_find_matching_catch_2()|0;
 $6 = tempRet0;
 __ZN4love7Matrix4D2Ev($t); //@line 476 "modules/graphics/opengl/Image.cpp"
 ___resumeException($5|0);
 // unreachable;
}
function __ZN4love8graphics6opengl5Image9setFilterERKNS0_7Texture6FilterE($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($this)) + 157|0); //@line 495 "modules/graphics/opengl/Image.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 495 "modules/graphics/opengl/Image.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 495 "modules/graphics/opengl/Image.cpp"
 $3 = (__ZN4love8graphics7Texture14validateFilterERKNS1_6FilterEb($f,$2)|0); //@line 495 "modules/graphics/opengl/Image.cpp"
 if ($3) {
  $19 = ((($this)) + 16|0); //@line 503 "modules/graphics/opengl/Image.cpp"
  ;HEAP32[$19>>2]=HEAP32[$f>>2]|0;HEAP32[$19+4>>2]=HEAP32[$f+4>>2]|0;HEAP32[$19+8>>2]=HEAP32[$f+8>>2]|0;HEAP32[$19+12>>2]=HEAP32[$f+12>>2]|0; //@line 503 "modules/graphics/opengl/Image.cpp"
  $20 = ((($this)) + 160|0); //@line 506 "modules/graphics/opengl/Image.cpp"
  $21 = HEAP8[$20>>0]|0; //@line 506 "modules/graphics/opengl/Image.cpp"
  $22 = ($21<<24>>24)==(0); //@line 506 "modules/graphics/opengl/Image.cpp"
  if ($22) {
   $25 = ((($this)) + 148|0); //@line 512 "modules/graphics/opengl/Image.cpp"
   $26 = HEAP32[$25>>2]|0; //@line 512 "modules/graphics/opengl/Image.cpp"
   __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$26); //@line 512 "modules/graphics/opengl/Image.cpp"
   __ZN4love8graphics6opengl6OpenGL16setTextureFilterERNS0_7Texture6FilterE(634044,$19); //@line 513 "modules/graphics/opengl/Image.cpp"
   STACKTOP = sp;return; //@line 514 "modules/graphics/opengl/Image.cpp"
  }
  $23 = ((($this)) + 24|0); //@line 508 "modules/graphics/opengl/Image.cpp"
  HEAP32[$23>>2] = 0; //@line 508 "modules/graphics/opengl/Image.cpp"
  $24 = ((($this)) + 20|0); //@line 509 "modules/graphics/opengl/Image.cpp"
  HEAP32[$24>>2] = 2; //@line 509 "modules/graphics/opengl/Image.cpp"
  HEAP32[$19>>2] = 2; //@line 509 "modules/graphics/opengl/Image.cpp"
  $25 = ((($this)) + 148|0); //@line 512 "modules/graphics/opengl/Image.cpp"
  $26 = HEAP32[$25>>2]|0; //@line 512 "modules/graphics/opengl/Image.cpp"
  __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$26); //@line 512 "modules/graphics/opengl/Image.cpp"
  __ZN4love8graphics6opengl6OpenGL16setTextureFilterERNS0_7Texture6FilterE(634044,$19); //@line 513 "modules/graphics/opengl/Image.cpp"
  STACKTOP = sp;return; //@line 514 "modules/graphics/opengl/Image.cpp"
 }
 $4 = ((($f)) + 8|0); //@line 497 "modules/graphics/opengl/Image.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 497 "modules/graphics/opengl/Image.cpp"
 $6 = ($5|0)==(0); //@line 497 "modules/graphics/opengl/Image.cpp"
 if (!($6)) {
  $7 = HEAP8[$0>>0]|0; //@line 497 "modules/graphics/opengl/Image.cpp"
  $8 = ($7<<24>>24)==(0); //@line 497 "modules/graphics/opengl/Image.cpp"
  if ($8) {
   $9 = (___cxa_allocate_exception(16)|0); //@line 498 "modules/graphics/opengl/Image.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($9|0),(436030|0),($vararg_buffer|0)); //@line 498 "modules/graphics/opengl/Image.cpp"
   $10 = __THREW__; __THREW__ = 0;
   $11 = $10&1;
   if (!($11)) {
    ___cxa_throw(($9|0),(96|0),(344|0)); //@line 498 "modules/graphics/opengl/Image.cpp"
    // unreachable; //@line 498 "modules/graphics/opengl/Image.cpp"
   }
   $12 = ___cxa_find_matching_catch_2()|0;
   $13 = tempRet0;
   ___cxa_free_exception(($9|0)); //@line 498 "modules/graphics/opengl/Image.cpp"
   $$0 = $13;$$01 = $12;
   ___resumeException($$01|0);
   // unreachable;
  }
 }
 $14 = (___cxa_allocate_exception(16)|0); //@line 500 "modules/graphics/opengl/Image.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($14|0),(439276|0),($vararg_buffer1|0)); //@line 500 "modules/graphics/opengl/Image.cpp"
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if (!($16)) {
  ___cxa_throw(($14|0),(96|0),(344|0)); //@line 500 "modules/graphics/opengl/Image.cpp"
  // unreachable; //@line 500 "modules/graphics/opengl/Image.cpp"
 }
 $17 = ___cxa_find_matching_catch_2()|0;
 $18 = tempRet0;
 ___cxa_free_exception(($14|0)); //@line 500 "modules/graphics/opengl/Image.cpp"
 $$0 = $18;$$01 = $17;
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN4love8graphics6opengl5Image7setWrapERKNS0_7Texture4WrapE($this,$w) {
 $this = $this|0;
 $w = $w|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $success$1$off0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 519 "modules/graphics/opengl/Image.cpp"
 $1 = $w; //@line 519 "modules/graphics/opengl/Image.cpp"
 $2 = $1; //@line 519 "modules/graphics/opengl/Image.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 519 "modules/graphics/opengl/Image.cpp"
 $4 = (($1) + 4)|0; //@line 519 "modules/graphics/opengl/Image.cpp"
 $5 = $4; //@line 519 "modules/graphics/opengl/Image.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 519 "modules/graphics/opengl/Image.cpp"
 $7 = $0; //@line 519 "modules/graphics/opengl/Image.cpp"
 $8 = $7; //@line 519 "modules/graphics/opengl/Image.cpp"
 HEAP32[$8>>2] = $3; //@line 519 "modules/graphics/opengl/Image.cpp"
 $9 = (($7) + 4)|0; //@line 519 "modules/graphics/opengl/Image.cpp"
 $10 = $9; //@line 519 "modules/graphics/opengl/Image.cpp"
 HEAP32[$10>>2] = $6; //@line 519 "modules/graphics/opengl/Image.cpp"
 $11 = HEAP8[1252012]|0; //@line 521 "modules/graphics/opengl/Image.cpp"
 $12 = ($11<<24>>24)==(0); //@line 521 "modules/graphics/opengl/Image.cpp"
 $13 = HEAP8[1252013]|0; //@line 521 "modules/graphics/opengl/Image.cpp"
 $14 = HEAP8[1252305]|0; //@line 521 "modules/graphics/opengl/Image.cpp"
 $15 = $14 | $13; //@line 521 "modules/graphics/opengl/Image.cpp"
 $16 = ($15<<24>>24)!=(0); //@line 521 "modules/graphics/opengl/Image.cpp"
 $17 = $12 | $16; //@line 521 "modules/graphics/opengl/Image.cpp"
 do {
  if ($17) {
   $success$1$off0 = 1;
  } else {
   $18 = ((($this)) + 8|0); //@line 522 "modules/graphics/opengl/Image.cpp"
   $19 = HEAP32[$18>>2]|0; //@line 522 "modules/graphics/opengl/Image.cpp"
   $20 = ($19|0)==(0); //@line 81 "./common/math.h"
   $21 = $20&1; //@line 81 "./common/math.h"
   $22 = (($19) + -1)|0; //@line 81 "./common/math.h"
   $23 = (($22) + ($21))|0; //@line 82 "./common/math.h"
   $24 = $23 >> 1; //@line 83 "./common/math.h"
   $25 = $24 | $23; //@line 83 "./common/math.h"
   $26 = $25 >> 2; //@line 83 "./common/math.h"
   $27 = $26 | $25; //@line 83 "./common/math.h"
   $28 = $27 >> 4; //@line 83 "./common/math.h"
   $29 = $28 | $27; //@line 83 "./common/math.h"
   $30 = $29 >> 8; //@line 83 "./common/math.h"
   $31 = $30 | $29; //@line 83 "./common/math.h"
   $32 = $31 >> 16; //@line 83 "./common/math.h"
   $33 = $32 | $31; //@line 83 "./common/math.h"
   $34 = (($33) + 1)|0; //@line 84 "./common/math.h"
   $35 = ($19|0)==($34|0); //@line 522 "modules/graphics/opengl/Image.cpp"
   if ($35) {
    $36 = ((($this)) + 12|0); //@line 522 "modules/graphics/opengl/Image.cpp"
    $37 = HEAP32[$36>>2]|0; //@line 522 "modules/graphics/opengl/Image.cpp"
    $38 = ($37|0)==(0); //@line 81 "./common/math.h"
    $39 = $38&1; //@line 81 "./common/math.h"
    $40 = (($37) + -1)|0; //@line 81 "./common/math.h"
    $41 = (($40) + ($39))|0; //@line 82 "./common/math.h"
    $42 = $41 >> 1; //@line 83 "./common/math.h"
    $43 = $42 | $41; //@line 83 "./common/math.h"
    $44 = $43 >> 2; //@line 83 "./common/math.h"
    $45 = $44 | $43; //@line 83 "./common/math.h"
    $46 = $45 >> 4; //@line 83 "./common/math.h"
    $47 = $46 | $45; //@line 83 "./common/math.h"
    $48 = $47 >> 8; //@line 83 "./common/math.h"
    $49 = $48 | $47; //@line 83 "./common/math.h"
    $50 = $49 >> 16; //@line 83 "./common/math.h"
    $51 = $50 | $49; //@line 83 "./common/math.h"
    $52 = (($51) + 1)|0; //@line 84 "./common/math.h"
    $53 = ($37|0)==($52|0); //@line 522 "modules/graphics/opengl/Image.cpp"
    if ($53) {
     $success$1$off0 = 1;
     break;
    }
   }
   $54 = ($3|0)==(0); //@line 524 "modules/graphics/opengl/Image.cpp"
   $55 = ((($this)) + 36|0); //@line 524 "modules/graphics/opengl/Image.cpp"
   $56 = ($6>>>0)<(1); //@line 524 "modules/graphics/opengl/Image.cpp"
   $57 = ($3>>>0)<(0); //@line 524 "modules/graphics/opengl/Image.cpp"
   $58 = ($6|0)==(1); //@line 524 "modules/graphics/opengl/Image.cpp"
   $59 = $58 & $57; //@line 524 "modules/graphics/opengl/Image.cpp"
   $60 = $56 | $59; //@line 524 "modules/graphics/opengl/Image.cpp"
   $or$cond = $54 & $60; //@line 524 "modules/graphics/opengl/Image.cpp"
   HEAP32[$55>>2] = 0; //@line 528 "modules/graphics/opengl/Image.cpp"
   HEAP32[$0>>2] = 0; //@line 528 "modules/graphics/opengl/Image.cpp"
   $success$1$off0 = $or$cond;
  }
 } while(0);
 $61 = (__ZNK4love8graphics6opengl6OpenGL31isClampZeroTextureWrapSupportedEv(634044)|0); //@line 531 "modules/graphics/opengl/Image.cpp"
 if ($61) {
  $67 = ((($this)) + 148|0); //@line 539 "modules/graphics/opengl/Image.cpp"
  $68 = HEAP32[$67>>2]|0; //@line 539 "modules/graphics/opengl/Image.cpp"
  __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$68); //@line 539 "modules/graphics/opengl/Image.cpp"
  __ZN4love8graphics6opengl6OpenGL14setTextureWrapERKNS0_7Texture4WrapE(634044,$0); //@line 540 "modules/graphics/opengl/Image.cpp"
  return ($success$1$off0|0); //@line 542 "modules/graphics/opengl/Image.cpp"
 }
 $62 = HEAP32[$0>>2]|0; //@line 533 "modules/graphics/opengl/Image.cpp"
 $63 = ($62|0)==(1); //@line 533 "modules/graphics/opengl/Image.cpp"
 if ($63) {
  HEAP32[$0>>2] = 0; //@line 534 "modules/graphics/opengl/Image.cpp"
 }
 $64 = ((($this)) + 36|0); //@line 535 "modules/graphics/opengl/Image.cpp"
 $65 = HEAP32[$64>>2]|0; //@line 535 "modules/graphics/opengl/Image.cpp"
 $66 = ($65|0)==(1); //@line 535 "modules/graphics/opengl/Image.cpp"
 if (!($66)) {
  $67 = ((($this)) + 148|0); //@line 539 "modules/graphics/opengl/Image.cpp"
  $68 = HEAP32[$67>>2]|0; //@line 539 "modules/graphics/opengl/Image.cpp"
  __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$68); //@line 539 "modules/graphics/opengl/Image.cpp"
  __ZN4love8graphics6opengl6OpenGL14setTextureWrapERKNS0_7Texture4WrapE(634044,$0); //@line 540 "modules/graphics/opengl/Image.cpp"
  return ($success$1$off0|0); //@line 542 "modules/graphics/opengl/Image.cpp"
 }
 HEAP32[$64>>2] = 0; //@line 536 "modules/graphics/opengl/Image.cpp"
 $67 = ((($this)) + 148|0); //@line 539 "modules/graphics/opengl/Image.cpp"
 $68 = HEAP32[$67>>2]|0; //@line 539 "modules/graphics/opengl/Image.cpp"
 __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$68); //@line 539 "modules/graphics/opengl/Image.cpp"
 __ZN4love8graphics6opengl6OpenGL14setTextureWrapERKNS0_7Texture4WrapE(634044,$0); //@line 540 "modules/graphics/opengl/Image.cpp"
 return ($success$1$off0|0); //@line 542 "modules/graphics/opengl/Image.cpp"
}
function __ZNK4love8graphics6opengl5Image9getHandleEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 148|0); //@line 480 "modules/graphics/opengl/Image.cpp"
 return ($0|0); //@line 480 "modules/graphics/opengl/Image.cpp"
}
function __ZN4love8graphics6opengl5Image12loadVolatileEv($this) {
 $this = $this|0;
 var $$020 = 0, $$021 = 0, $$1 = 0, $$122 = 0, $$in = 0, $$in$in = 0, $$pr23$pre = 0, $$pre = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0.0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0.0, $172 = 0, $173 = 0, $174 = 0.0, $175 = 0.0, $176 = 0, $177 = 0, $178 = 0.0, $179 = 0.0, $18 = 0;
 var $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0.0, $185 = 0.0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0;
 var $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0;
 var $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0;
 var $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0;
 var $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0;
 var $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0.0, $282 = 0.0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, $or$cond11 = 0, $or$cond3 = 0;
 var $or$cond3$not = 0, $or$cond5 = 0, $phitmp = 0, $px$i = 0, $storemerge = 0, $str = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_buffer4 = 0, $vararg_buffer6 = 0, $vararg_ptr1 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 40|0;
 $1 = sp + 36|0;
 $2 = sp + 32|0;
 $px$i = sp + 48|0;
 $str = sp + 28|0;
 $3 = ((($this)) + 156|0); //@line 592 "modules/graphics/opengl/Image.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 592 "modules/graphics/opengl/Image.cpp"
 $5 = $4&255; //@line 592 "modules/graphics/opengl/Image.cpp"
 $6 = ($5<<24>>24)==(0); //@line 592 "modules/graphics/opengl/Image.cpp"
 do {
  if ($6) {
   $$in$in = $4;
   label = 20;
  } else {
   $7 = ((($this)) + 136|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $8 = HEAP32[$7>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $9 = HEAP32[$8>>2]|0; //@line 117 "./common/Object.h"
   $10 = (__ZNK4love5image19CompressedImageData9getFormatEv($9)|0); //@line 284 "modules/graphics/opengl/Image.cpp"
   $11 = ((($this)) + 159|0); //@line 284 "modules/graphics/opengl/Image.cpp"
   $12 = HEAP8[$11>>0]|0; //@line 284 "modules/graphics/opengl/Image.cpp"
   $13 = ($12<<24>>24)!=(0); //@line 284 "modules/graphics/opengl/Image.cpp"
   switch ($10|0) {
   case 1:  {
    $14 = HEAP8[1252236]|0; //@line 707 "modules/graphics/opengl/Image.cpp"
    $15 = HEAP8[1252233]|0; //@line 707 "modules/graphics/opengl/Image.cpp"
    $16 = $15 | $14; //@line 707 "modules/graphics/opengl/Image.cpp"
    $17 = ($16<<24>>24)==(0); //@line 707 "modules/graphics/opengl/Image.cpp"
    if (!($17)) {
     label = 19;
    }
    break;
   }
   case 2:  {
    $18 = HEAP8[1252236]|0; //@line 709 "modules/graphics/opengl/Image.cpp"
    $19 = HEAP8[1252349]|0; //@line 709 "modules/graphics/opengl/Image.cpp"
    $20 = $19 | $18; //@line 709 "modules/graphics/opengl/Image.cpp"
    $21 = ($20<<24>>24)==(0); //@line 709 "modules/graphics/opengl/Image.cpp"
    if (!($21)) {
     label = 19;
    }
    break;
   }
   case 3:  {
    $22 = HEAP8[1252236]|0; //@line 711 "modules/graphics/opengl/Image.cpp"
    $23 = HEAP8[1252350]|0; //@line 711 "modules/graphics/opengl/Image.cpp"
    $24 = $23 | $22; //@line 711 "modules/graphics/opengl/Image.cpp"
    $25 = ($24<<24>>24)==(0); //@line 711 "modules/graphics/opengl/Image.cpp"
    if (!($25)) {
     label = 19;
    }
    break;
   }
   case 7: case 6: case 5: case 4:  {
    $26 = HEAP8[1252002]|0; //@line 716 "modules/graphics/opengl/Image.cpp"
    $27 = HEAP8[1252132]|0; //@line 716 "modules/graphics/opengl/Image.cpp"
    $28 = $27 | $26; //@line 716 "modules/graphics/opengl/Image.cpp"
    $29 = HEAP8[1252235]|0; //@line 716 "modules/graphics/opengl/Image.cpp"
    $30 = $28 | $29; //@line 716 "modules/graphics/opengl/Image.cpp"
    $31 = ($30<<24>>24)==(0); //@line 716 "modules/graphics/opengl/Image.cpp"
    if (!($31)) {
     label = 19;
    }
    break;
   }
   case 10: case 9: case 8:  {
    $32 = HEAP8[1252008]|0; //@line 720 "modules/graphics/opengl/Image.cpp"
    $33 = HEAP8[1252131]|0; //@line 720 "modules/graphics/opengl/Image.cpp"
    $34 = $33 | $32; //@line 720 "modules/graphics/opengl/Image.cpp"
    $35 = ($34<<24>>24)==(0); //@line 720 "modules/graphics/opengl/Image.cpp"
    if (!($35)) {
     label = 19;
    }
    break;
   }
   case 14: case 13: case 12: case 11:  {
    $36 = HEAP8[1252207]|0; //@line 725 "modules/graphics/opengl/Image.cpp"
    $37 = HEAP8[1252391]|0; //@line 725 "modules/graphics/opengl/Image.cpp"
    $38 = $13 ? $36 : $37; //@line 725 "modules/graphics/opengl/Image.cpp"
    $39 = ($38<<24>>24)==(0); //@line 725 "modules/graphics/opengl/Image.cpp"
    if (!($39)) {
     label = 19;
    }
    break;
   }
   case 15:  {
    $40 = HEAP8[1252013]|0; //@line 728 "modules/graphics/opengl/Image.cpp"
    $41 = HEAP8[1252009]|0; //@line 728 "modules/graphics/opengl/Image.cpp"
    $42 = $41 | $40; //@line 728 "modules/graphics/opengl/Image.cpp"
    $43 = HEAP8[1252058]|0; //@line 728 "modules/graphics/opengl/Image.cpp"
    $44 = $42 | $43; //@line 728 "modules/graphics/opengl/Image.cpp"
    $45 = HEAP8[1252263]|0; //@line 728 "modules/graphics/opengl/Image.cpp"
    $46 = $44 | $45; //@line 728 "modules/graphics/opengl/Image.cpp"
    $47 = ($46<<24>>24)==(0); //@line 728 "modules/graphics/opengl/Image.cpp"
    if (!($47)) {
     label = 19;
    }
    break;
   }
   case 22: case 21: case 20: case 19: case 18: case 17: case 16:  {
    $52 = HEAP8[1252013]|0; //@line 736 "modules/graphics/opengl/Image.cpp"
    $53 = HEAP8[1252009]|0; //@line 736 "modules/graphics/opengl/Image.cpp"
    $54 = $53 | $52; //@line 736 "modules/graphics/opengl/Image.cpp"
    $55 = HEAP8[1252058]|0; //@line 736 "modules/graphics/opengl/Image.cpp"
    $56 = $54 | $55; //@line 736 "modules/graphics/opengl/Image.cpp"
    $57 = ($56<<24>>24)==(0); //@line 736 "modules/graphics/opengl/Image.cpp"
    if (!($57)) {
     label = 19;
    }
    break;
   }
   case 36: case 35: case 34: case 33: case 32: case 31: case 30: case 29: case 28: case 27: case 26: case 25: case 24: case 23:  {
    $48 = HEAP8[1252015]|0; //@line 751 "modules/graphics/opengl/Image.cpp"
    $49 = HEAP8[1252024]|0; //@line 751 "modules/graphics/opengl/Image.cpp"
    $50 = $49 | $48; //@line 751 "modules/graphics/opengl/Image.cpp"
    $51 = ($50<<24>>24)==(0); //@line 751 "modules/graphics/opengl/Image.cpp"
    if (!($51)) {
     label = 19;
    }
    break;
   }
   default: {
   }
   }
   if ((label|0) == 19) {
    $75 = HEAP32[$3>>2]|0; //@line 592 "modules/graphics/opengl/Image.cpp"
    $76 = $75&255; //@line 592 "modules/graphics/opengl/Image.cpp"
    $77 = ($76<<24>>24)==(0); //@line 592 "modules/graphics/opengl/Image.cpp"
    if ($77) {
     $$in$in = $75;
     label = 20;
     break;
    } else {
     break;
    }
   }
   $58 = HEAP32[$7>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $59 = HEAP32[$58>>2]|0; //@line 117 "./common/Object.h"
   $60 = (__ZNK4love5image19CompressedImageData9getFormatEv($59)|0); //@line 287 "modules/graphics/opengl/Image.cpp"
   $61 = (__ZN4love5image19CompressedImageData11getConstantENS1_6FormatERPKc($60,$str)|0); //@line 287 "modules/graphics/opengl/Image.cpp"
   $62 = (___cxa_allocate_exception(16)|0); //@line 289 "modules/graphics/opengl/Image.cpp"
   if ($61) {
    $63 = HEAP8[$11>>0]|0; //@line 290 "modules/graphics/opengl/Image.cpp"
    $64 = ($63<<24>>24)!=(0); //@line 290 "modules/graphics/opengl/Image.cpp"
    $65 = $64 ? 436080 : 1282466; //@line 290 "modules/graphics/opengl/Image.cpp"
    $66 = HEAP32[$str>>2]|0; //@line 290 "modules/graphics/opengl/Image.cpp"
    __THREW__ = 0;
    HEAP32[$vararg_buffer>>2] = $65; //@line 289 "modules/graphics/opengl/Image.cpp"
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 289 "modules/graphics/opengl/Image.cpp"
    HEAP32[$vararg_ptr1>>2] = $66; //@line 289 "modules/graphics/opengl/Image.cpp"
    invoke_viii(3349,($62|0),(436086|0),($vararg_buffer|0)); //@line 289 "modules/graphics/opengl/Image.cpp"
    $67 = __THREW__; __THREW__ = 0;
    $68 = $67&1;
    if (!($68)) {
     ___cxa_throw(($62|0),(96|0),(344|0)); //@line 289 "modules/graphics/opengl/Image.cpp"
     // unreachable; //@line 289 "modules/graphics/opengl/Image.cpp"
    }
    $69 = ___cxa_find_matching_catch_2()|0;
    $70 = tempRet0;
    ___cxa_free_exception(($62|0)); //@line 289 "modules/graphics/opengl/Image.cpp"
    $$1 = $69;$$122 = $70;
    ___resumeException($$1|0);
    // unreachable;
   } else {
    __THREW__ = 0;
    invoke_viii(3349,($62|0),(436164|0),($vararg_buffer2|0)); //@line 293 "modules/graphics/opengl/Image.cpp"
    $71 = __THREW__; __THREW__ = 0;
    $72 = $71&1;
    if (!($72)) {
     ___cxa_throw(($62|0),(96|0),(344|0)); //@line 293 "modules/graphics/opengl/Image.cpp"
     // unreachable; //@line 293 "modules/graphics/opengl/Image.cpp"
    }
    $73 = ___cxa_find_matching_catch_2()|0;
    $74 = tempRet0;
    ___cxa_free_exception(($62|0)); //@line 293 "modules/graphics/opengl/Image.cpp"
    $$1 = $73;$$122 = $74;
    ___resumeException($$1|0);
    // unreachable;
   }
  }
 } while(0);
 if ((label|0) == 20) {
  $$in = $$in$in >>> 24; //@line 295 "modules/graphics/opengl/Image.cpp"
  $78 = $$in&255; //@line 301 "modules/graphics/opengl/Image.cpp"
  $79 = ($78<<24>>24)==(0); //@line 297 "modules/graphics/opengl/Image.cpp"
  if ($79) {
   $$pre = HEAP8[1252224]|0; //@line 301 "modules/graphics/opengl/Image.cpp"
   $$pre25 = HEAP8[1252013]|0; //@line 301 "modules/graphics/opengl/Image.cpp"
   $97 = $$pre;$99 = $$pre25;
  } else {
   $80 = HEAP8[1252013]|0; //@line 759 "modules/graphics/opengl/Image.cpp"
   $81 = HEAP8[1252224]|0; //@line 759 "modules/graphics/opengl/Image.cpp"
   $82 = $81 | $80; //@line 759 "modules/graphics/opengl/Image.cpp"
   $83 = HEAP8[1252001]|0; //@line 759 "modules/graphics/opengl/Image.cpp"
   $84 = $82 | $83; //@line 759 "modules/graphics/opengl/Image.cpp"
   $85 = HEAP8[1252247]|0; //@line 759 "modules/graphics/opengl/Image.cpp"
   $86 = $84 | $85; //@line 759 "modules/graphics/opengl/Image.cpp"
   $87 = ($86<<24>>24)==(0); //@line 759 "modules/graphics/opengl/Image.cpp"
   if ($87) {
    $88 = (___cxa_allocate_exception(16)|0); //@line 298 "modules/graphics/opengl/Image.cpp"
    __THREW__ = 0;
    invoke_viii(3349,($88|0),(436225|0),($vararg_buffer4|0)); //@line 298 "modules/graphics/opengl/Image.cpp"
    $89 = __THREW__; __THREW__ = 0;
    $90 = $89&1;
    if (!($90)) {
     ___cxa_throw(($88|0),(96|0),(344|0)); //@line 298 "modules/graphics/opengl/Image.cpp"
     // unreachable; //@line 298 "modules/graphics/opengl/Image.cpp"
    }
    $91 = ___cxa_find_matching_catch_2()|0;
    $92 = tempRet0;
    ___cxa_free_exception(($88|0)); //@line 298 "modules/graphics/opengl/Image.cpp"
    $$1 = $91;$$122 = $92;
    ___resumeException($$1|0);
    // unreachable;
   } else {
    $97 = $81;$99 = $80;
   }
  }
  $93 = ($78<<24>>24)!=(0); //@line 301 "modules/graphics/opengl/Image.cpp"
  $94 = HEAP8[1252012]|0; //@line 301 "modules/graphics/opengl/Image.cpp"
  $95 = ($94<<24>>24)!=(0); //@line 301 "modules/graphics/opengl/Image.cpp"
  $or$cond = $93 & $95; //@line 301 "modules/graphics/opengl/Image.cpp"
  $96 = ($97<<24>>24)!=(0); //@line 301 "modules/graphics/opengl/Image.cpp"
  $or$cond3 = $or$cond & $96; //@line 301 "modules/graphics/opengl/Image.cpp"
  $or$cond3$not = $or$cond3 ^ 1; //@line 301 "modules/graphics/opengl/Image.cpp"
  $98 = ($99<<24>>24)!=(0); //@line 301 "modules/graphics/opengl/Image.cpp"
  $or$cond5 = $98 | $or$cond3$not; //@line 301 "modules/graphics/opengl/Image.cpp"
  if (!($or$cond5)) {
   $100 = ((($this)) + 124|0); //@line 302 "modules/graphics/opengl/Image.cpp"
   $101 = ((($this)) + 128|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $102 = HEAP32[$101>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $103 = HEAP32[$100>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $104 = (($102) - ($103))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $105 = $104 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $106 = ($105>>>0)<(2); //@line 302 "modules/graphics/opengl/Image.cpp"
   if ($106) {
    $107 = ((($this)) + 157|0); //@line 304 "modules/graphics/opengl/Image.cpp"
    HEAP8[$107>>0] = 0; //@line 304 "modules/graphics/opengl/Image.cpp"
    $108 = ((($this)) + 24|0); //@line 305 "modules/graphics/opengl/Image.cpp"
    HEAP32[$108>>2] = 0; //@line 305 "modules/graphics/opengl/Image.cpp"
   }
  }
 }
 $109 = HEAP8[1252012]|0; //@line 310 "modules/graphics/opengl/Image.cpp"
 $110 = ($109<<24>>24)==(0); //@line 310 "modules/graphics/opengl/Image.cpp"
 $111 = HEAP8[1252013]|0; //@line 310 "modules/graphics/opengl/Image.cpp"
 $112 = HEAP8[1252305]|0; //@line 310 "modules/graphics/opengl/Image.cpp"
 $113 = $112 | $111; //@line 310 "modules/graphics/opengl/Image.cpp"
 $114 = ($113<<24>>24)!=(0); //@line 310 "modules/graphics/opengl/Image.cpp"
 $115 = $110 | $114; //@line 310 "modules/graphics/opengl/Image.cpp"
 do {
  if (!($115)) {
   $116 = ((($this)) + 8|0); //@line 311 "modules/graphics/opengl/Image.cpp"
   $117 = HEAP32[$116>>2]|0; //@line 311 "modules/graphics/opengl/Image.cpp"
   $118 = ($117|0)==(0); //@line 81 "./common/math.h"
   $119 = $118&1; //@line 81 "./common/math.h"
   $120 = (($117) + -1)|0; //@line 81 "./common/math.h"
   $121 = (($120) + ($119))|0; //@line 82 "./common/math.h"
   $122 = $121 >> 1; //@line 83 "./common/math.h"
   $123 = $122 | $121; //@line 83 "./common/math.h"
   $124 = $123 >> 2; //@line 83 "./common/math.h"
   $125 = $124 | $123; //@line 83 "./common/math.h"
   $126 = $125 >> 4; //@line 83 "./common/math.h"
   $127 = $126 | $125; //@line 83 "./common/math.h"
   $128 = $127 >> 8; //@line 83 "./common/math.h"
   $129 = $128 | $127; //@line 83 "./common/math.h"
   $130 = $129 >> 16; //@line 83 "./common/math.h"
   $131 = $130 | $129; //@line 83 "./common/math.h"
   $132 = (($131) + 1)|0; //@line 84 "./common/math.h"
   $133 = ($117|0)==($132|0); //@line 311 "modules/graphics/opengl/Image.cpp"
   if ($133) {
    $134 = ((($this)) + 12|0); //@line 311 "modules/graphics/opengl/Image.cpp"
    $135 = HEAP32[$134>>2]|0; //@line 311 "modules/graphics/opengl/Image.cpp"
    $136 = ($135|0)==(0); //@line 81 "./common/math.h"
    $137 = $136&1; //@line 81 "./common/math.h"
    $138 = (($135) + -1)|0; //@line 81 "./common/math.h"
    $139 = (($138) + ($137))|0; //@line 82 "./common/math.h"
    $140 = $139 >> 1; //@line 83 "./common/math.h"
    $141 = $140 | $139; //@line 83 "./common/math.h"
    $142 = $141 >> 2; //@line 83 "./common/math.h"
    $143 = $142 | $141; //@line 83 "./common/math.h"
    $144 = $143 >> 4; //@line 83 "./common/math.h"
    $145 = $144 | $143; //@line 83 "./common/math.h"
    $146 = $145 >> 8; //@line 83 "./common/math.h"
    $147 = $146 | $145; //@line 83 "./common/math.h"
    $148 = $147 >> 16; //@line 83 "./common/math.h"
    $149 = $148 | $147; //@line 83 "./common/math.h"
    $150 = (($149) + 1)|0; //@line 84 "./common/math.h"
    $151 = ($135|0)==($150|0); //@line 311 "modules/graphics/opengl/Image.cpp"
    if ($151) {
     break;
    }
   }
   $152 = ((($this)) + 157|0); //@line 313 "modules/graphics/opengl/Image.cpp"
   HEAP8[$152>>0] = 0; //@line 313 "modules/graphics/opengl/Image.cpp"
   $153 = ((($this)) + 24|0); //@line 314 "modules/graphics/opengl/Image.cpp"
   HEAP32[$153>>2] = 0; //@line 314 "modules/graphics/opengl/Image.cpp"
  }
 } while(0);
 $154 = +HEAPF32[158432]; //@line 317 "modules/graphics/opengl/Image.cpp"
 $155 = $154 == 0.0; //@line 317 "modules/graphics/opengl/Image.cpp"
 $156 = HEAP8[1251998]|0; //@line 317 "modules/graphics/opengl/Image.cpp"
 $157 = ($156<<24>>24)!=(0); //@line 317 "modules/graphics/opengl/Image.cpp"
 $or$cond11 = $155 & $157; //@line 317 "modules/graphics/opengl/Image.cpp"
 if ($or$cond11) {
  $158 = HEAP32[165575]|0; //@line 72 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$158 & 8191](34045,633728); //@line 72 "./libraries/glad/gladfuncs.hpp"
 }
 $159 = ((($this)) + 148|0); //@line 320 "modules/graphics/opengl/Image.cpp"
 $160 = HEAP32[164922]|0; //@line 355 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$160 & 8191](1,$159); //@line 355 "./libraries/glad/gladfuncs.hpp"
 $161 = HEAP32[$159>>2]|0; //@line 321 "modules/graphics/opengl/Image.cpp"
 __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$161); //@line 321 "modules/graphics/opengl/Image.cpp"
 $162 = HEAP32[$this>>2]|0; //@line 323 "modules/graphics/opengl/Image.cpp"
 $163 = ((($162)) + 32|0); //@line 323 "modules/graphics/opengl/Image.cpp"
 $164 = HEAP32[$163>>2]|0; //@line 323 "modules/graphics/opengl/Image.cpp"
 $165 = ((($this)) + 16|0); //@line 323 "modules/graphics/opengl/Image.cpp"
 FUNCTION_TABLE_vii[$164 & 8191]($this,$165); //@line 323 "modules/graphics/opengl/Image.cpp"
 $166 = HEAP32[$this>>2]|0; //@line 324 "modules/graphics/opengl/Image.cpp"
 $167 = ((($166)) + 40|0); //@line 324 "modules/graphics/opengl/Image.cpp"
 $168 = HEAP32[$167>>2]|0; //@line 324 "modules/graphics/opengl/Image.cpp"
 $169 = ((($this)) + 32|0); //@line 324 "modules/graphics/opengl/Image.cpp"
 (FUNCTION_TABLE_iii[$168 & 8191]($this,$169)|0); //@line 324 "modules/graphics/opengl/Image.cpp"
 $170 = ((($this)) + 152|0); //@line 325 "modules/graphics/opengl/Image.cpp"
 $171 = +HEAPF32[$170>>2]; //@line 325 "modules/graphics/opengl/Image.cpp"
 HEAPF32[$0>>2] = $171; //@line 325 "modules/graphics/opengl/Image.cpp"
 $172 = HEAP8[1251998]|0; //@line 548 "modules/graphics/opengl/Image.cpp"
 $173 = ($172<<24>>24)==(0); //@line 548 "modules/graphics/opengl/Image.cpp"
 if (!($173)) {
  $174 = +HEAPF32[158432]; //@line 552 "modules/graphics/opengl/Image.cpp"
  $175 = 0.0099999997764825821 - $174; //@line 552 "modules/graphics/opengl/Image.cpp"
  HEAPF32[$1>>2] = $175; //@line 552 "modules/graphics/opengl/Image.cpp"
  $176 = $175 > $171; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $177 = $176 ? $1 : $0; //@line 2648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $178 = $174 + -0.0099999997764825821; //@line 552 "modules/graphics/opengl/Image.cpp"
  HEAPF32[$2>>2] = $178; //@line 552 "modules/graphics/opengl/Image.cpp"
  $179 = +HEAPF32[$177>>2]; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $180 = $178 < $179; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $181 = $180 ? $2 : $177; //@line 2575 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $182 = HEAP32[$181>>2]|0; //@line 552 "modules/graphics/opengl/Image.cpp"
  HEAP32[$170>>2] = $182; //@line 552 "modules/graphics/opengl/Image.cpp"
  $183 = HEAP32[$159>>2]|0; //@line 554 "modules/graphics/opengl/Image.cpp"
  __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$183); //@line 554 "modules/graphics/opengl/Image.cpp"
  $184 = +HEAPF32[$170>>2]; //@line 557 "modules/graphics/opengl/Image.cpp"
  $185 = -$184; //@line 557 "modules/graphics/opengl/Image.cpp"
  $186 = HEAP32[165106]|0; //@line 42 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viid[$186 & 8191](3553,34049,$185); //@line 42 "./libraries/glad/gladfuncs.hpp"
 }
 $187 = ((($this)) + 8|0); //@line 328 "modules/graphics/opengl/Image.cpp"
 $188 = HEAP32[$187>>2]|0; //@line 328 "modules/graphics/opengl/Image.cpp"
 $189 = (__ZNK4love8graphics6opengl6OpenGL17getMaxTextureSizeEv(634044)|0); //@line 328 "modules/graphics/opengl/Image.cpp"
 $190 = ($188|0)>($189|0); //@line 328 "modules/graphics/opengl/Image.cpp"
 if (!($190)) {
  $191 = ((($this)) + 12|0); //@line 328 "modules/graphics/opengl/Image.cpp"
  $192 = HEAP32[$191>>2]|0; //@line 328 "modules/graphics/opengl/Image.cpp"
  $193 = (__ZNK4love8graphics6opengl6OpenGL17getMaxTextureSizeEv(634044)|0); //@line 328 "modules/graphics/opengl/Image.cpp"
  $194 = ($192|0)>($193|0); //@line 328 "modules/graphics/opengl/Image.cpp"
  if (!($194)) {
   $201 = ((($this)) + 157|0); //@line 334 "modules/graphics/opengl/Image.cpp"
   $202 = HEAP8[$201>>0]|0; //@line 334 "modules/graphics/opengl/Image.cpp"
   $203 = ($202<<24>>24)==(0); //@line 334 "modules/graphics/opengl/Image.cpp"
   if ($203) {
    $204 = HEAP8[1252013]|0; //@line 334 "modules/graphics/opengl/Image.cpp"
    $205 = HEAP8[1251994]|0; //@line 334 "modules/graphics/opengl/Image.cpp"
    $206 = $205 | $204; //@line 334 "modules/graphics/opengl/Image.cpp"
    $207 = ($206<<24>>24)==(0); //@line 334 "modules/graphics/opengl/Image.cpp"
    if (!($207)) {
     $208 = HEAP32[165108]|0; //@line 44 "./libraries/glad/gladfuncs.hpp"
     FUNCTION_TABLE_viii[$208 & 8191](3553,33085,0); //@line 44 "./libraries/glad/gladfuncs.hpp"
     $$pr23$pre = HEAP8[$201>>0]|0; //@line 337 "modules/graphics/opengl/Image.cpp"
     $phitmp = ($$pr23$pre<<24>>24)==(0); //@line 335 "modules/graphics/opengl/Image.cpp"
     if (!($phitmp)) {
      label = 43;
     }
    }
   } else {
    label = 43;
   }
   if ((label|0) == 43) {
    $209 = HEAP8[$3>>0]|0; //@line 592 "modules/graphics/opengl/Image.cpp"
    $210 = ($209<<24>>24)==(0); //@line 592 "modules/graphics/opengl/Image.cpp"
    if ($210) {
     $211 = ((($this)) + 124|0); //@line 337 "modules/graphics/opengl/Image.cpp"
     $212 = ((($this)) + 128|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $213 = HEAP32[$212>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $214 = HEAP32[$211>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $215 = (($213) - ($214))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $216 = $215 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $217 = ($216>>>0)>(1); //@line 337 "modules/graphics/opengl/Image.cpp"
     $218 = HEAP8[1252012]|0; //@line 338 "modules/graphics/opengl/Image.cpp"
     $219 = HEAP8[1252002]|0; //@line 338 "modules/graphics/opengl/Image.cpp"
     $220 = $219 | $218; //@line 337 "modules/graphics/opengl/Image.cpp"
     $221 = HEAP8[1252065]|0; //@line 338 "modules/graphics/opengl/Image.cpp"
     $222 = $220 | $221; //@line 337 "modules/graphics/opengl/Image.cpp"
     $223 = ($222<<24>>24)!=(0); //@line 337 "modules/graphics/opengl/Image.cpp"
     $224 = $217 | $223; //@line 337 "modules/graphics/opengl/Image.cpp"
     if (!($224)) {
      $225 = HEAP32[165108]|0; //@line 44 "./libraries/glad/gladfuncs.hpp"
      FUNCTION_TABLE_viii[$225 & 8191](3553,33169,1); //@line 44 "./libraries/glad/gladfuncs.hpp"
     }
    }
   }
   while(1) {
    $226 = HEAP32[165571]|0; //@line 71 "./libraries/glad/gladfuncs.hpp"
    $227 = (FUNCTION_TABLE_i[$226 & 8191]()|0); //@line 71 "./libraries/glad/gladfuncs.hpp"
    $228 = ($227|0)==(0); //@line 345 "modules/graphics/opengl/Image.cpp"
    if ($228) {
     break;
    }
   }
   $229 = HEAP8[$3>>0]|0; //@line 592 "modules/graphics/opengl/Image.cpp"
   $230 = ($229<<24>>24)==(0); //@line 592 "modules/graphics/opengl/Image.cpp"
   if ($230) {
    __THREW__ = 0;
    invoke_vi(4079,($this|0)); //@line 352 "modules/graphics/opengl/Image.cpp"
    $235 = __THREW__; __THREW__ = 0;
    $236 = $235&1;
    if ($236) {
     label = 49;
    } else {
     label = 51;
    }
   } else {
    __THREW__ = 0;
    invoke_vi(4078,($this|0)); //@line 350 "modules/graphics/opengl/Image.cpp"
    $231 = __THREW__; __THREW__ = 0;
    $232 = $231&1;
    if ($232) {
     label = 49;
    } else {
     label = 51;
    }
   }
   do {
    if ((label|0) == 51) {
     $237 = HEAP32[165571]|0; //@line 71 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     $238 = (invoke_i($237|0)|0); //@line 71 "./libraries/glad/gladfuncs.hpp"
     $239 = __THREW__; __THREW__ = 0;
     $240 = $239&1;
     if ($240) {
      label = 49;
     } else {
      $241 = ($238|0)==(0); //@line 355 "modules/graphics/opengl/Image.cpp"
      if (!($241)) {
       $242 = (___cxa_allocate_exception(16)|0); //@line 356 "modules/graphics/opengl/Image.cpp"
       __THREW__ = 0;
       $243 = (invoke_ii(4080,($238|0))|0); //@line 356 "modules/graphics/opengl/Image.cpp"
       $244 = __THREW__; __THREW__ = 0;
       $245 = $244&1;
       if (!($245)) {
        __THREW__ = 0;
        HEAP32[$vararg_buffer6>>2] = $243; //@line 356 "modules/graphics/opengl/Image.cpp"
        invoke_viii(3349,($242|0),(436287|0),($vararg_buffer6|0)); //@line 356 "modules/graphics/opengl/Image.cpp"
        $246 = __THREW__; __THREW__ = 0;
        $247 = $246&1;
        if (!($247)) {
         __THREW__ = 0;
         invoke_viii(3350,($242|0),(96|0),(344|0)); //@line 356 "modules/graphics/opengl/Image.cpp"
         $248 = __THREW__; __THREW__ = 0;
         label = 49;
         break;
        }
       }
       $249 = ___cxa_find_matching_catch_3(96|0)|0;
       $250 = tempRet0;
       ___cxa_free_exception(($242|0)); //@line 356 "modules/graphics/opengl/Image.cpp"
       $$020 = $249;$$021 = $250;
       break;
      }
      $257 = ((($this)) + 164|0); //@line 365 "modules/graphics/opengl/Image.cpp"
      $258 = HEAP32[$257>>2]|0; //@line 365 "modules/graphics/opengl/Image.cpp"
      $259 = HEAP8[$3>>0]|0; //@line 592 "modules/graphics/opengl/Image.cpp"
      $260 = ($259<<24>>24)==(0); //@line 592 "modules/graphics/opengl/Image.cpp"
      if ($260) {
       $272 = ((($this)) + 124|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $273 = HEAP32[$272>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $274 = HEAP32[$273>>2]|0; //@line 117 "./common/Object.h"
       $275 = HEAP32[$274>>2]|0; //@line 370 "modules/graphics/opengl/Image.cpp"
       $276 = ((($275)) + 20|0); //@line 370 "modules/graphics/opengl/Image.cpp"
       $277 = HEAP32[$276>>2]|0; //@line 370 "modules/graphics/opengl/Image.cpp"
       $278 = (FUNCTION_TABLE_ii[$277 & 8191]($274)|0); //@line 370 "modules/graphics/opengl/Image.cpp"
       $storemerge = $278;
      } else {
       $265 = ((($this)) + 136|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $266 = HEAP32[$265>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $267 = HEAP32[$266>>2]|0; //@line 117 "./common/Object.h"
       $268 = HEAP32[$267>>2]|0; //@line 368 "modules/graphics/opengl/Image.cpp"
       $269 = ((($268)) + 20|0); //@line 368 "modules/graphics/opengl/Image.cpp"
       $270 = HEAP32[$269>>2]|0; //@line 368 "modules/graphics/opengl/Image.cpp"
       $271 = (FUNCTION_TABLE_ii[$270 & 8191]($267)|0); //@line 368 "modules/graphics/opengl/Image.cpp"
       $storemerge = $271;
      }
      HEAP32[$257>>2] = $storemerge; //@line 368 "modules/graphics/opengl/Image.cpp"
      $279 = HEAP8[$201>>0]|0; //@line 372 "modules/graphics/opengl/Image.cpp"
      $280 = ($279<<24>>24)==(0); //@line 372 "modules/graphics/opengl/Image.cpp"
      if ($280) {
       $284 = $storemerge;
      } else {
       $281 = (+($storemerge>>>0)); //@line 373 "modules/graphics/opengl/Image.cpp"
       $282 = $281 * 1.33334; //@line 373 "modules/graphics/opengl/Image.cpp"
       $283 = (~~(($282))>>>0); //@line 373 "modules/graphics/opengl/Image.cpp"
       HEAP32[$257>>2] = $283; //@line 373 "modules/graphics/opengl/Image.cpp"
       $284 = $283;
      }
      __ZN4love8graphics6opengl6OpenGL23updateTextureMemorySizeEjj(634044,$258,$284); //@line 375 "modules/graphics/opengl/Image.cpp"
      $285 = ((($this)) + 160|0); //@line 377 "modules/graphics/opengl/Image.cpp"
      HEAP8[$285>>0] = 0; //@line 377 "modules/graphics/opengl/Image.cpp"
      STACKTOP = sp;return 1; //@line 379 "modules/graphics/opengl/Image.cpp"
     }
    }
   } while(0);
   if ((label|0) == 49) {
    $233 = ___cxa_find_matching_catch_3(96|0)|0;
    $234 = tempRet0;
    $$020 = $233;$$021 = $234;
   }
   $251 = (_llvm_eh_typeid_for((96|0))|0); //@line 357 "modules/graphics/opengl/Image.cpp"
   $252 = ($$021|0)==($251|0); //@line 357 "modules/graphics/opengl/Image.cpp"
   if (!($252)) {
    $$1 = $$020;$$122 = $$021;
    ___resumeException($$1|0);
    // unreachable;
   }
   (___cxa_begin_catch(($$020|0))|0); //@line 357 "modules/graphics/opengl/Image.cpp"
   $253 = HEAP32[$159>>2]|0; //@line 360 "modules/graphics/opengl/Image.cpp"
   __THREW__ = 0;
   invoke_vii(4014,(634044|0),($253|0)); //@line 360 "modules/graphics/opengl/Image.cpp"
   $254 = __THREW__; __THREW__ = 0;
   $255 = $254&1;
   if (!($255)) {
    HEAP32[$159>>2] = 0; //@line 361 "modules/graphics/opengl/Image.cpp"
    __THREW__ = 0;
    invoke_v(3656); //@line 362 "modules/graphics/opengl/Image.cpp"
    $256 = __THREW__; __THREW__ = 0;
   }
   $261 = ___cxa_find_matching_catch_2()|0;
   $262 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 363 "modules/graphics/opengl/Image.cpp"
   $263 = __THREW__; __THREW__ = 0;
   $264 = $263&1;
   if ($264) {
    $286 = ___cxa_find_matching_catch_3(0|0)|0;
    $287 = tempRet0;
    ___clang_call_terminate($286); //@line 363 "modules/graphics/opengl/Image.cpp"
    // unreachable; //@line 363 "modules/graphics/opengl/Image.cpp"
   } else {
    $$1 = $261;$$122 = $262;
    ___resumeException($$1|0);
    // unreachable;
   }
  }
 }
 $195 = ((($this)) + 160|0); //@line 214 "modules/graphics/opengl/Image.cpp"
 HEAP8[$195>>0] = 1; //@line 214 "modules/graphics/opengl/Image.cpp"
 $196 = HEAP32[$159>>2]|0; //@line 216 "modules/graphics/opengl/Image.cpp"
 __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$196); //@line 216 "modules/graphics/opengl/Image.cpp"
 $197 = HEAP32[$this>>2]|0; //@line 217 "modules/graphics/opengl/Image.cpp"
 $198 = ((($197)) + 32|0); //@line 217 "modules/graphics/opengl/Image.cpp"
 $199 = HEAP32[$198>>2]|0; //@line 217 "modules/graphics/opengl/Image.cpp"
 FUNCTION_TABLE_vii[$199 & 8191]($this,$165); //@line 217 "modules/graphics/opengl/Image.cpp"
 dest=$px$i; src=436271; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 220 "modules/graphics/opengl/Image.cpp"
 $200 = HEAP32[165602]|0; //@line 47 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiiiiiiii[$200 & 8191](3553,0,6408,2,2,0,6408,5121,$px$i); //@line 47 "./libraries/glad/gladfuncs.hpp"
 STACKTOP = sp;return 1; //@line 379 "modules/graphics/opengl/Image.cpp"
}
function __ZN4love8graphics6opengl5Image22loadFromCompressedDataEv($this) {
 $this = $this|0;
 var $$cast = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $count$0 = 0, $count$02 = 0, $exitcond = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 136|0); //@line 228 "modules/graphics/opengl/Image.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 117 "./common/Object.h"
 $3 = (__ZNK4love5image19CompressedImageData9getFormatEv($2)|0); //@line 228 "modules/graphics/opengl/Image.cpp"
 $4 = ((($this)) + 159|0); //@line 228 "modules/graphics/opengl/Image.cpp"
 $5 = (__ZNK4love8graphics6opengl5Image19getCompressedFormatENS_5image19CompressedImageData6FormatERb(0,$3,$4)|0); //@line 228 "modules/graphics/opengl/Image.cpp"
 $6 = (__ZN4love8graphics14isGammaCorrectEv()|0); //@line 230 "modules/graphics/opengl/Image.cpp"
 if ($6) {
  $7 = HEAP8[$4>>0]|0; //@line 230 "modules/graphics/opengl/Image.cpp"
  $8 = ($7<<24>>24)==(0); //@line 230 "modules/graphics/opengl/Image.cpp"
  if ($8) {
   $9 = ((($this)) + 158|0); //@line 231 "modules/graphics/opengl/Image.cpp"
   HEAP8[$9>>0] = 1; //@line 231 "modules/graphics/opengl/Image.cpp"
  }
 }
 $10 = ((($this)) + 157|0); //@line 235 "modules/graphics/opengl/Image.cpp"
 $11 = HEAP8[$10>>0]|0; //@line 237 "modules/graphics/opengl/Image.cpp"
 $12 = ($11<<24>>24)==(0); //@line 235 "modules/graphics/opengl/Image.cpp"
 if ($12) {
  $count$02 = 1;
 } else {
  $13 = ((($this)) + 140|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = HEAP32[$13>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = (($14) - ($15))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = $16 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ($17>>>0)>(1); //@line 235 "modules/graphics/opengl/Image.cpp"
  if ($18) {
   $count$0 = $17;
  } else {
   $19 = $15; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $20 = HEAP32[$19>>2]|0; //@line 117 "./common/Object.h"
   $21 = (__ZNK4love5image19CompressedImageData14getMipmapCountEv($20)|0); //@line 238 "modules/graphics/opengl/Image.cpp"
   $count$0 = $21;
  }
  $22 = ($count$0|0)>(0); //@line 240 "modules/graphics/opengl/Image.cpp"
  if ($22) {
   $count$02 = $count$0;
  } else {
   return; //@line 251 "modules/graphics/opengl/Image.cpp"
  }
 }
 $23 = ((($this)) + 140|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $i$01 = 0;
 while(1) {
  $24 = HEAP32[$23>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $25 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $26 = (($24) - ($25))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $27 = $26 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $28 = ($27>>>0)>(1); //@line 244 "modules/graphics/opengl/Image.cpp"
  $$cast = $25; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $29 = (($$cast) + ($i$01<<2)|0); //@line 139 "./common/Object.h"
  $$in = $28 ? $29 : $$cast; //@line 244 "modules/graphics/opengl/Image.cpp"
  $30 = HEAP32[$$in>>2]|0; //@line 244 "modules/graphics/opengl/Image.cpp"
  $31 = $28 ? 0 : $i$01; //@line 245 "modules/graphics/opengl/Image.cpp"
  $32 = (__ZNK4love5image19CompressedImageData8getWidthEi($30,$31)|0); //@line 247 "modules/graphics/opengl/Image.cpp"
  $33 = (__ZNK4love5image19CompressedImageData9getHeightEi($30,$31)|0); //@line 248 "modules/graphics/opengl/Image.cpp"
  $34 = (__ZNK4love5image19CompressedImageData7getSizeEi($30,$31)|0); //@line 249 "modules/graphics/opengl/Image.cpp"
  $35 = (__ZNK4love5image19CompressedImageData7getDataEi($30,$31)|0); //@line 249 "modules/graphics/opengl/Image.cpp"
  $36 = HEAP32[164678]|0; //@line 384 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiiiiiii[$36 & 8191](3553,$i$01,$5,$32,$33,0,$34,$35); //@line 384 "./libraries/glad/gladfuncs.hpp"
  $37 = (($i$01) + 1)|0; //@line 240 "modules/graphics/opengl/Image.cpp"
  $exitcond = ($37|0)==($count$02|0); //@line 240 "modules/graphics/opengl/Image.cpp"
  if ($exitcond) {
   break;
  } else {
   $i$01 = $37;
  }
 }
 return; //@line 251 "modules/graphics/opengl/Image.cpp"
}
function __ZNK4love8graphics6opengl5Image19getCompressedFormatENS_5image19CompressedImageData6FormatERb($this,$cformat,$isSRGB) {
 $this = $this|0;
 $cformat = $cformat|0;
 $isSRGB = $isSRGB|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 L1: do {
  switch ($cformat|0) {
  case 1:  {
   $0 = HEAP8[$isSRGB>>0]|0; //@line 602 "modules/graphics/opengl/Image.cpp"
   $1 = ($0<<24>>24)!=(0); //@line 602 "modules/graphics/opengl/Image.cpp"
   $2 = $1 ? 35916 : 33776; //@line 602 "modules/graphics/opengl/Image.cpp"
   $$0 = $2;
   break;
  }
  case 2:  {
   $3 = HEAP8[$isSRGB>>0]|0; //@line 604 "modules/graphics/opengl/Image.cpp"
   $4 = ($3<<24>>24)!=(0); //@line 604 "modules/graphics/opengl/Image.cpp"
   $5 = $4 ? 35918 : 33778; //@line 604 "modules/graphics/opengl/Image.cpp"
   $$0 = $5;
   break;
  }
  case 3:  {
   $6 = HEAP8[$isSRGB>>0]|0; //@line 606 "modules/graphics/opengl/Image.cpp"
   $7 = ($6<<24>>24)!=(0); //@line 606 "modules/graphics/opengl/Image.cpp"
   $8 = $7 ? 35919 : 33779; //@line 606 "modules/graphics/opengl/Image.cpp"
   $$0 = $8;
   break;
  }
  case 4:  {
   HEAP8[$isSRGB>>0] = 0; //@line 608 "modules/graphics/opengl/Image.cpp"
   $$0 = 36283;
   break;
  }
  case 5:  {
   HEAP8[$isSRGB>>0] = 0; //@line 611 "modules/graphics/opengl/Image.cpp"
   $$0 = 36284;
   break;
  }
  case 6:  {
   HEAP8[$isSRGB>>0] = 0; //@line 614 "modules/graphics/opengl/Image.cpp"
   $$0 = 36285;
   break;
  }
  case 7:  {
   HEAP8[$isSRGB>>0] = 0; //@line 617 "modules/graphics/opengl/Image.cpp"
   $$0 = 36286;
   break;
  }
  case 8:  {
   HEAP8[$isSRGB>>0] = 0; //@line 620 "modules/graphics/opengl/Image.cpp"
   $$0 = 36495;
   break;
  }
  case 9:  {
   HEAP8[$isSRGB>>0] = 0; //@line 623 "modules/graphics/opengl/Image.cpp"
   $$0 = 36494;
   break;
  }
  case 10:  {
   $9 = HEAP8[$isSRGB>>0]|0; //@line 626 "modules/graphics/opengl/Image.cpp"
   $10 = ($9<<24>>24)!=(0); //@line 626 "modules/graphics/opengl/Image.cpp"
   $11 = $10 ? 36493 : 36492; //@line 626 "modules/graphics/opengl/Image.cpp"
   $$0 = $11;
   break;
  }
  case 11:  {
   $12 = HEAP8[$isSRGB>>0]|0; //@line 628 "modules/graphics/opengl/Image.cpp"
   $13 = ($12<<24>>24)!=(0); //@line 628 "modules/graphics/opengl/Image.cpp"
   $14 = $13 ? 35412 : 35841; //@line 628 "modules/graphics/opengl/Image.cpp"
   $$0 = $14;
   break;
  }
  case 12:  {
   $15 = HEAP8[$isSRGB>>0]|0; //@line 630 "modules/graphics/opengl/Image.cpp"
   $16 = ($15<<24>>24)!=(0); //@line 630 "modules/graphics/opengl/Image.cpp"
   $17 = $16 ? 35413 : 35840; //@line 630 "modules/graphics/opengl/Image.cpp"
   $$0 = $17;
   break;
  }
  case 13:  {
   $18 = HEAP8[$isSRGB>>0]|0; //@line 632 "modules/graphics/opengl/Image.cpp"
   $19 = ($18<<24>>24)!=(0); //@line 632 "modules/graphics/opengl/Image.cpp"
   $20 = $19 ? 35414 : 35843; //@line 632 "modules/graphics/opengl/Image.cpp"
   $$0 = $20;
   break;
  }
  case 14:  {
   $21 = HEAP8[$isSRGB>>0]|0; //@line 634 "modules/graphics/opengl/Image.cpp"
   $22 = ($21<<24>>24)!=(0); //@line 634 "modules/graphics/opengl/Image.cpp"
   $23 = $22 ? 35415 : 35842; //@line 634 "modules/graphics/opengl/Image.cpp"
   $$0 = $23;
   break;
  }
  case 15:  {
   $24 = HEAP8[1252013]|0; //@line 637 "modules/graphics/opengl/Image.cpp"
   $25 = HEAP8[1252009]|0; //@line 637 "modules/graphics/opengl/Image.cpp"
   $26 = $25 | $24; //@line 637 "modules/graphics/opengl/Image.cpp"
   $27 = HEAP8[1252058]|0; //@line 637 "modules/graphics/opengl/Image.cpp"
   $28 = $26 | $27; //@line 637 "modules/graphics/opengl/Image.cpp"
   $29 = ($28<<24>>24)==(0); //@line 637 "modules/graphics/opengl/Image.cpp"
   if ($29) {
    HEAP8[$isSRGB>>0] = 0; //@line 641 "modules/graphics/opengl/Image.cpp"
    $$0 = 36196;
    break L1;
   } else {
    $30 = HEAP8[$isSRGB>>0]|0; //@line 638 "modules/graphics/opengl/Image.cpp"
    $31 = ($30<<24>>24)!=(0); //@line 638 "modules/graphics/opengl/Image.cpp"
    $32 = $31 ? 37493 : 37492; //@line 638 "modules/graphics/opengl/Image.cpp"
    $$0 = $32;
    break L1;
   }
   break;
  }
  case 16:  {
   $33 = HEAP8[$isSRGB>>0]|0; //@line 645 "modules/graphics/opengl/Image.cpp"
   $34 = ($33<<24>>24)!=(0); //@line 645 "modules/graphics/opengl/Image.cpp"
   $35 = $34 ? 37493 : 37492; //@line 645 "modules/graphics/opengl/Image.cpp"
   $$0 = $35;
   break;
  }
  case 17:  {
   $36 = HEAP8[$isSRGB>>0]|0; //@line 647 "modules/graphics/opengl/Image.cpp"
   $37 = ($36<<24>>24)!=(0); //@line 647 "modules/graphics/opengl/Image.cpp"
   $38 = $37 ? 37497 : 37496; //@line 647 "modules/graphics/opengl/Image.cpp"
   $$0 = $38;
   break;
  }
  case 18:  {
   $39 = HEAP8[$isSRGB>>0]|0; //@line 649 "modules/graphics/opengl/Image.cpp"
   $40 = ($39<<24>>24)!=(0); //@line 649 "modules/graphics/opengl/Image.cpp"
   $41 = $40 ? 37495 : 37494; //@line 649 "modules/graphics/opengl/Image.cpp"
   $$0 = $41;
   break;
  }
  case 19:  {
   HEAP8[$isSRGB>>0] = 0; //@line 651 "modules/graphics/opengl/Image.cpp"
   $$0 = 37488;
   break;
  }
  case 20:  {
   HEAP8[$isSRGB>>0] = 0; //@line 654 "modules/graphics/opengl/Image.cpp"
   $$0 = 37489;
   break;
  }
  case 21:  {
   HEAP8[$isSRGB>>0] = 0; //@line 657 "modules/graphics/opengl/Image.cpp"
   $$0 = 37490;
   break;
  }
  case 22:  {
   HEAP8[$isSRGB>>0] = 0; //@line 660 "modules/graphics/opengl/Image.cpp"
   $$0 = 37491;
   break;
  }
  case 23:  {
   $42 = HEAP8[$isSRGB>>0]|0; //@line 663 "modules/graphics/opengl/Image.cpp"
   $43 = ($42<<24>>24)!=(0); //@line 663 "modules/graphics/opengl/Image.cpp"
   $44 = $43 ? 37840 : 37808; //@line 663 "modules/graphics/opengl/Image.cpp"
   $$0 = $44;
   break;
  }
  case 24:  {
   $45 = HEAP8[$isSRGB>>0]|0; //@line 665 "modules/graphics/opengl/Image.cpp"
   $46 = ($45<<24>>24)!=(0); //@line 665 "modules/graphics/opengl/Image.cpp"
   $47 = $46 ? 37841 : 37809; //@line 665 "modules/graphics/opengl/Image.cpp"
   $$0 = $47;
   break;
  }
  case 25:  {
   $48 = HEAP8[$isSRGB>>0]|0; //@line 667 "modules/graphics/opengl/Image.cpp"
   $49 = ($48<<24>>24)!=(0); //@line 667 "modules/graphics/opengl/Image.cpp"
   $50 = $49 ? 37842 : 37810; //@line 667 "modules/graphics/opengl/Image.cpp"
   $$0 = $50;
   break;
  }
  case 26:  {
   $51 = HEAP8[$isSRGB>>0]|0; //@line 669 "modules/graphics/opengl/Image.cpp"
   $52 = ($51<<24>>24)!=(0); //@line 669 "modules/graphics/opengl/Image.cpp"
   $53 = $52 ? 37843 : 37811; //@line 669 "modules/graphics/opengl/Image.cpp"
   $$0 = $53;
   break;
  }
  case 27:  {
   $54 = HEAP8[$isSRGB>>0]|0; //@line 671 "modules/graphics/opengl/Image.cpp"
   $55 = ($54<<24>>24)!=(0); //@line 671 "modules/graphics/opengl/Image.cpp"
   $56 = $55 ? 37844 : 37812; //@line 671 "modules/graphics/opengl/Image.cpp"
   $$0 = $56;
   break;
  }
  case 28:  {
   $57 = HEAP8[$isSRGB>>0]|0; //@line 673 "modules/graphics/opengl/Image.cpp"
   $58 = ($57<<24>>24)!=(0); //@line 673 "modules/graphics/opengl/Image.cpp"
   $59 = $58 ? 37845 : 37813; //@line 673 "modules/graphics/opengl/Image.cpp"
   $$0 = $59;
   break;
  }
  case 29:  {
   $60 = HEAP8[$isSRGB>>0]|0; //@line 675 "modules/graphics/opengl/Image.cpp"
   $61 = ($60<<24>>24)!=(0); //@line 675 "modules/graphics/opengl/Image.cpp"
   $62 = $61 ? 37846 : 37814; //@line 675 "modules/graphics/opengl/Image.cpp"
   $$0 = $62;
   break;
  }
  case 30:  {
   $63 = HEAP8[$isSRGB>>0]|0; //@line 677 "modules/graphics/opengl/Image.cpp"
   $64 = ($63<<24>>24)!=(0); //@line 677 "modules/graphics/opengl/Image.cpp"
   $65 = $64 ? 37847 : 37815; //@line 677 "modules/graphics/opengl/Image.cpp"
   $$0 = $65;
   break;
  }
  case 31:  {
   $66 = HEAP8[$isSRGB>>0]|0; //@line 679 "modules/graphics/opengl/Image.cpp"
   $67 = ($66<<24>>24)!=(0); //@line 679 "modules/graphics/opengl/Image.cpp"
   $68 = $67 ? 37848 : 37816; //@line 679 "modules/graphics/opengl/Image.cpp"
   $$0 = $68;
   break;
  }
  case 32:  {
   $69 = HEAP8[$isSRGB>>0]|0; //@line 681 "modules/graphics/opengl/Image.cpp"
   $70 = ($69<<24>>24)!=(0); //@line 681 "modules/graphics/opengl/Image.cpp"
   $71 = $70 ? 37849 : 37817; //@line 681 "modules/graphics/opengl/Image.cpp"
   $$0 = $71;
   break;
  }
  case 33:  {
   $72 = HEAP8[$isSRGB>>0]|0; //@line 683 "modules/graphics/opengl/Image.cpp"
   $73 = ($72<<24>>24)!=(0); //@line 683 "modules/graphics/opengl/Image.cpp"
   $74 = $73 ? 37850 : 37818; //@line 683 "modules/graphics/opengl/Image.cpp"
   $$0 = $74;
   break;
  }
  case 34:  {
   $75 = HEAP8[$isSRGB>>0]|0; //@line 685 "modules/graphics/opengl/Image.cpp"
   $76 = ($75<<24>>24)!=(0); //@line 685 "modules/graphics/opengl/Image.cpp"
   $77 = $76 ? 37851 : 37819; //@line 685 "modules/graphics/opengl/Image.cpp"
   $$0 = $77;
   break;
  }
  case 35:  {
   $78 = HEAP8[$isSRGB>>0]|0; //@line 687 "modules/graphics/opengl/Image.cpp"
   $79 = ($78<<24>>24)!=(0); //@line 687 "modules/graphics/opengl/Image.cpp"
   $80 = $79 ? 37852 : 37820; //@line 687 "modules/graphics/opengl/Image.cpp"
   $$0 = $80;
   break;
  }
  case 36:  {
   $81 = HEAP8[$isSRGB>>0]|0; //@line 689 "modules/graphics/opengl/Image.cpp"
   $82 = ($81<<24>>24)!=(0); //@line 689 "modules/graphics/opengl/Image.cpp"
   $83 = $82 ? 37853 : 37821; //@line 689 "modules/graphics/opengl/Image.cpp"
   $$0 = $83;
   break;
  }
  default: {
   $84 = HEAP8[$isSRGB>>0]|0; //@line 691 "modules/graphics/opengl/Image.cpp"
   $85 = ($84<<24>>24)!=(0); //@line 691 "modules/graphics/opengl/Image.cpp"
   $86 = $85 ? 35907 : 32856; //@line 691 "modules/graphics/opengl/Image.cpp"
   $$0 = $86;
  }
  }
 } while(0);
 return ($$0|0); //@line 693 "modules/graphics/opengl/Image.cpp"
}
function __ZN4love8graphics6opengl5Image17loadFromImageDataEv($this) {
 $this = $this|0;
 var $$ = 0, $$2 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, $i$04 = 0, $lock = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock = sp;
 $0 = ((($this)) + 159|0); //@line 255 "modules/graphics/opengl/Image.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 255 "modules/graphics/opengl/Image.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 255 "modules/graphics/opengl/Image.cpp"
 $3 = $2 ? 35907 : 32856; //@line 255 "modules/graphics/opengl/Image.cpp"
 $4 = HEAP8[1252012]|0; //@line 259 "modules/graphics/opengl/Image.cpp"
 $5 = ($4<<24>>24)==(0); //@line 259 "modules/graphics/opengl/Image.cpp"
 $6 = HEAP8[1252013]|0; //@line 259 "modules/graphics/opengl/Image.cpp"
 $7 = ($6<<24>>24)!=(0); //@line 259 "modules/graphics/opengl/Image.cpp"
 $or$cond = $5 | $7; //@line 259 "modules/graphics/opengl/Image.cpp"
 $8 = $2 ? 35906 : 6408; //@line 261 "modules/graphics/opengl/Image.cpp"
 $$ = $or$cond ? 6408 : $8; //@line 259 "modules/graphics/opengl/Image.cpp"
 $$2 = $or$cond ? $3 : $8; //@line 259 "modules/graphics/opengl/Image.cpp"
 $9 = ((($this)) + 157|0); //@line 265 "modules/graphics/opengl/Image.cpp"
 $10 = HEAP8[$9>>0]|0; //@line 265 "modules/graphics/opengl/Image.cpp"
 $11 = ($10<<24>>24)==(0); //@line 265 "modules/graphics/opengl/Image.cpp"
 $$pre = ((($this)) + 124|0); //@line 269 "modules/graphics/opengl/Image.cpp"
 if ($11) {
  $46 = 1;
  label = 3;
 } else {
  $12 = ((($this)) + 128|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = HEAP32[$12>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = HEAP32[$$pre>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = (($13) - ($14))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = $15 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = ($15|0)>(0); //@line 267 "modules/graphics/opengl/Image.cpp"
  if ($17) {
   $46 = $16;
   label = 3;
  }
 }
 L3: do {
  if ((label|0) == 3) {
   $18 = ((($this)) + 124|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $i$04 = 0;
   while(1) {
    $25 = HEAP32[$18>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $26 = (($25) + ($i$04<<2)|0); //@line 139 "./common/Object.h"
    $27 = HEAP32[$26>>2]|0; //@line 139 "./common/Object.h"
    $28 = (__ZNK4love5image9ImageData8getMutexEv($27)|0); //@line 270 "modules/graphics/opengl/Image.cpp"
    __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$28); //@line 270 "modules/graphics/opengl/Image.cpp"
    __THREW__ = 0;
    $29 = (invoke_ii(3984,($27|0))|0); //@line 272 "modules/graphics/opengl/Image.cpp"
    $30 = __THREW__; __THREW__ = 0;
    $31 = $30&1;
    if ($31) {
     label = 10;
     break;
    }
    __THREW__ = 0;
    $32 = (invoke_ii(4081,($27|0))|0); //@line 272 "modules/graphics/opengl/Image.cpp"
    $33 = __THREW__; __THREW__ = 0;
    $34 = $33&1;
    if ($34) {
     label = 10;
     break;
    }
    $35 = HEAP32[$27>>2]|0; //@line 273 "modules/graphics/opengl/Image.cpp"
    $36 = ((($35)) + 16|0); //@line 273 "modules/graphics/opengl/Image.cpp"
    $37 = HEAP32[$36>>2]|0; //@line 273 "modules/graphics/opengl/Image.cpp"
    __THREW__ = 0;
    $38 = (invoke_ii($37|0,($27|0))|0); //@line 273 "modules/graphics/opengl/Image.cpp"
    $39 = __THREW__; __THREW__ = 0;
    $40 = $39&1;
    if ($40) {
     label = 10;
     break;
    }
    $41 = HEAP32[165602]|0; //@line 47 "./libraries/glad/gladfuncs.hpp"
    __THREW__ = 0;
    invoke_viiiiiiiii($41|0,3553,($i$04|0),($$2|0),($29|0),($32|0),0,($$|0),5121,($38|0)); //@line 47 "./libraries/glad/gladfuncs.hpp"
    $42 = __THREW__; __THREW__ = 0;
    $43 = $42&1;
    if ($43) {
     label = 10;
     break;
    }
    __ZN4love6thread4LockD2Ev($lock); //@line 274 "modules/graphics/opengl/Image.cpp"
    $44 = (($i$04) + 1)|0; //@line 267 "modules/graphics/opengl/Image.cpp"
    $45 = ($44|0)<($46|0); //@line 267 "modules/graphics/opengl/Image.cpp"
    if ($45) {
     $i$04 = $44;
    } else {
     break L3;
    }
   }
   if ((label|0) == 10) {
    $47 = ___cxa_find_matching_catch_2()|0;
    $48 = tempRet0;
    __ZN4love6thread4LockD2Ev($lock); //@line 274 "modules/graphics/opengl/Image.cpp"
    ___resumeException($47|0);
    // unreachable;
   }
  }
 } while(0);
 $19 = ((($this)) + 128|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = HEAP32[$19>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = HEAP32[$$pre>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $22 = (($20) - ($21))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $23 = $22 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $24 = ($23>>>0)<(2); //@line 276 "modules/graphics/opengl/Image.cpp"
 if (!($24)) {
  STACKTOP = sp;return; //@line 278 "modules/graphics/opengl/Image.cpp"
 }
 $49 = HEAP8[$9>>0]|0; //@line 202 "modules/graphics/opengl/Image.cpp"
 $50 = ($49<<24>>24)==(0); //@line 202 "modules/graphics/opengl/Image.cpp"
 if ($50) {
  STACKTOP = sp;return; //@line 278 "modules/graphics/opengl/Image.cpp"
 }
 $51 = ((($this)) + 156|0); //@line 592 "modules/graphics/opengl/Image.cpp"
 $52 = HEAP8[$51>>0]|0; //@line 592 "modules/graphics/opengl/Image.cpp"
 $53 = ($52<<24>>24)==(0); //@line 592 "modules/graphics/opengl/Image.cpp"
 if (!($53)) {
  STACKTOP = sp;return; //@line 278 "modules/graphics/opengl/Image.cpp"
 }
 $54 = HEAP8[1252012]|0; //@line 203 "modules/graphics/opengl/Image.cpp"
 $55 = HEAP8[1252002]|0; //@line 203 "modules/graphics/opengl/Image.cpp"
 $56 = $55 | $54; //@line 203 "modules/graphics/opengl/Image.cpp"
 $57 = HEAP8[1252065]|0; //@line 203 "modules/graphics/opengl/Image.cpp"
 $58 = $56 | $57; //@line 203 "modules/graphics/opengl/Image.cpp"
 $59 = ($58<<24>>24)==(0); //@line 203 "modules/graphics/opengl/Image.cpp"
 if ($59) {
  STACKTOP = sp;return; //@line 278 "modules/graphics/opengl/Image.cpp"
 }
 $60 = HEAP8[(634081)>>0]|0; //@line 205 "modules/graphics/opengl/Image.cpp"
 $61 = ($60<<24>>24)==(0); //@line 205 "modules/graphics/opengl/Image.cpp"
 if (!($61)) {
  $62 = HEAP32[164832]|0; //@line 57 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$62 & 8191](3553); //@line 57 "./libraries/glad/gladfuncs.hpp"
 }
 $63 = HEAP32[165249]|0; //@line 676 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$63 & 8191](3553); //@line 676 "./libraries/glad/gladfuncs.hpp"
 STACKTOP = sp;return; //@line 278 "modules/graphics/opengl/Image.cpp"
}
function __ZN4love8graphics6opengl5Image14unloadVolatileEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 148|0); //@line 383 "modules/graphics/opengl/Image.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 383 "modules/graphics/opengl/Image.cpp"
 $2 = ($1|0)==(0); //@line 383 "modules/graphics/opengl/Image.cpp"
 if ($2) {
  return; //@line 391 "modules/graphics/opengl/Image.cpp"
 }
 __ZN4love8graphics6opengl6OpenGL13deleteTextureEj(634044,$1); //@line 386 "modules/graphics/opengl/Image.cpp"
 HEAP32[$0>>2] = 0; //@line 387 "modules/graphics/opengl/Image.cpp"
 $3 = ((($this)) + 164|0); //@line 389 "modules/graphics/opengl/Image.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 389 "modules/graphics/opengl/Image.cpp"
 __ZN4love8graphics6opengl6OpenGL23updateTextureMemorySizeEjj(634044,$4,0); //@line 389 "modules/graphics/opengl/Image.cpp"
 HEAP32[$3>>2] = 0; //@line 390 "modules/graphics/opengl/Image.cpp"
 return; //@line 391 "modules/graphics/opengl/Image.cpp"
}
function __ZThn120_N4love8graphics6opengl5ImageD1Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -120|0); //@line 152 "modules/graphics/opengl/Image.cpp"
 __ZN4love8graphics6opengl5ImageD2Ev($0); //@line 152 "modules/graphics/opengl/Image.cpp"
 return; //@line 152 "modules/graphics/opengl/Image.cpp"
}
function __ZThn120_N4love8graphics6opengl5ImageD0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -120|0); //@line 152 "modules/graphics/opengl/Image.cpp"
 __ZN4love8graphics6opengl5ImageD2Ev($0); //@line 153 "modules/graphics/opengl/Image.cpp"
 __ZdlPv($0); //@line 153 "modules/graphics/opengl/Image.cpp"
 return; //@line 152 "modules/graphics/opengl/Image.cpp"
}
function __ZThn120_N4love8graphics6opengl5Image12loadVolatileEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -120|0); //@line 280 "modules/graphics/opengl/Image.cpp"
 (__ZN4love8graphics6opengl5Image12loadVolatileEv($0)|0); //@line 280 "modules/graphics/opengl/Image.cpp"
 return 1; //@line 280 "modules/graphics/opengl/Image.cpp"
}
function __ZThn120_N4love8graphics6opengl5Image14unloadVolatileEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -120|0); //@line 381 "modules/graphics/opengl/Image.cpp"
 $1 = ((($0)) + 148|0); //@line 383 "modules/graphics/opengl/Image.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 383 "modules/graphics/opengl/Image.cpp"
 $3 = ($2|0)==(0); //@line 383 "modules/graphics/opengl/Image.cpp"
 if ($3) {
  return; //@line 381 "modules/graphics/opengl/Image.cpp"
 }
 __ZN4love8graphics6opengl6OpenGL13deleteTextureEj(634044,$2); //@line 386 "modules/graphics/opengl/Image.cpp"
 HEAP32[$1>>2] = 0; //@line 387 "modules/graphics/opengl/Image.cpp"
 $4 = ((($0)) + 164|0); //@line 389 "modules/graphics/opengl/Image.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 389 "modules/graphics/opengl/Image.cpp"
 __ZN4love8graphics6opengl6OpenGL23updateTextureMemorySizeEjj(634044,$5,0); //@line 389 "modules/graphics/opengl/Image.cpp"
 HEAP32[$4>>2] = 0; //@line 390 "modules/graphics/opengl/Image.cpp"
 return; //@line 381 "modules/graphics/opengl/Image.cpp"
}
function __GLOBAL__sub_I_Image_cpp() {
 var $$02$i$i$1$i = 0, $$02$i$i$i = 0, $$lcssa$i$1$i = 0, $$lcssa$i$i = 0, $$lcssa10$i$1$i = 0, $$lcssa10$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i$i$1$i = 0, $hash$0$lcssa$i$i$i = 0;
 var $hash$01$i$i$1$i = 0, $hash$01$i$i$i = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 HEAP8[(633744)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633756)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633768)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633780)>>0] = 0; //@line 155 "./common/StringMap.h"
 $0 = (633784); //@line 44 "./common/StringMap.h"
 $1 = $0; //@line 44 "./common/StringMap.h"
 HEAP32[$1>>2] = 0; //@line 44 "./common/StringMap.h"
 $2 = (($0) + 4)|0; //@line 44 "./common/StringMap.h"
 $3 = $2; //@line 44 "./common/StringMap.h"
 HEAP32[$3>>2] = 0; //@line 44 "./common/StringMap.h"
 $4 = HEAP32[6227]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(24912)>>2]|0; //@line 49 "./common/StringMap.h"
 $6 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
 $7 = ($6<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($7) {
  $hash$0$lcssa$i$i$i = 5381;
 } else {
  $$02$i$i$i = $4;$9 = $6;$hash$01$i$i$i = 5381;
  while(1) {
   $8 = $9 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $10 = ((($$02$i$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $11 = ($hash$01$i$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $12 = (($11) + ($8))|0; //@line 143 "./common/StringMap.h"
   $13 = HEAP8[$10>>0]|0; //@line 142 "./common/StringMap.h"
   $14 = ($13<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($14) {
    $hash$0$lcssa$i$i$i = $12;
    break;
   } else {
    $$02$i$i$i = $10;$9 = $13;$hash$01$i$i$i = $12;
   }
  }
 }
 $15 = $hash$0$lcssa$i$i$i & 3; //@line 112 "./common/StringMap.h"
 $16 = (((633736 + (($15*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $17 = HEAP8[$16>>0]|0; //@line 114 "./common/StringMap.h"
 $18 = ($17<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($18) {
  $$lcssa$i$i = $15;$$lcssa10$i$i = $16;
  label = 5;
 } else {
  $19 = (($hash$0$lcssa$i$i$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $20 = $19 & 3; //@line 112 "./common/StringMap.h"
  $21 = (((633736 + (($20*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $22 = HEAP8[$21>>0]|0; //@line 114 "./common/StringMap.h"
  $23 = ($22<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($23) {
   $$lcssa$i$i = $20;$$lcssa10$i$i = $21;
   label = 5;
  } else {
   $28 = (($hash$0$lcssa$i$i$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $29 = $28 & 3; //@line 112 "./common/StringMap.h"
   $30 = (((633736 + (($29*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $31 = HEAP8[$30>>0]|0; //@line 114 "./common/StringMap.h"
   $32 = ($31<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($32) {
    $$lcssa$i$i = $29;$$lcssa10$i$i = $30;
    label = 5;
   } else {
    $33 = (($hash$0$lcssa$i$i$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $34 = $33 & 3; //@line 112 "./common/StringMap.h"
    $35 = (((633736 + (($34*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $36 = HEAP8[$35>>0]|0; //@line 114 "./common/StringMap.h"
    $37 = ($36<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($37) {
     $$lcssa$i$i = $34;$$lcssa10$i$i = $35;
     label = 5;
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10$i$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $24 = (633736 + (($$lcssa$i$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$24>>2] = $4; //@line 118 "./common/StringMap.h"
  $25 = (((633736 + (($$lcssa$i$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$25>>2] = $5; //@line 119 "./common/StringMap.h"
 }
 $26 = ($5>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($26) {
  HEAP32[$vararg_buffer>>2] = $4; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $5; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
 } else {
  $27 = (633784 + ($5<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$27>>2] = $4; //@line 132 "./common/StringMap.h"
 }
 $38 = HEAP32[(24916)>>2]|0; //@line 49 "./common/StringMap.h"
 $39 = HEAP32[(24920)>>2]|0; //@line 49 "./common/StringMap.h"
 $40 = HEAP8[$38>>0]|0; //@line 142 "./common/StringMap.h"
 $41 = ($40<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($41) {
  $hash$0$lcssa$i$i$1$i = 5381;
 } else {
  $$02$i$i$1$i = $38;$43 = $40;$hash$01$i$i$1$i = 5381;
  while(1) {
   $42 = $43 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $44 = ((($$02$i$i$1$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $45 = ($hash$01$i$i$1$i*33)|0; //@line 143 "./common/StringMap.h"
   $46 = (($45) + ($42))|0; //@line 143 "./common/StringMap.h"
   $47 = HEAP8[$44>>0]|0; //@line 142 "./common/StringMap.h"
   $48 = ($47<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($48) {
    $hash$0$lcssa$i$i$1$i = $46;
    break;
   } else {
    $$02$i$i$1$i = $44;$43 = $47;$hash$01$i$i$1$i = $46;
   }
  }
 }
 $49 = $hash$0$lcssa$i$i$1$i & 3; //@line 112 "./common/StringMap.h"
 $50 = (((633736 + (($49*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $51 = HEAP8[$50>>0]|0; //@line 114 "./common/StringMap.h"
 $52 = ($51<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($52) {
  $$lcssa$i$1$i = $49;$$lcssa10$i$1$i = $50;
  label = 17;
 } else {
  $53 = (($hash$0$lcssa$i$i$1$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $54 = $53 & 3; //@line 112 "./common/StringMap.h"
  $55 = (((633736 + (($54*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $56 = HEAP8[$55>>0]|0; //@line 114 "./common/StringMap.h"
  $57 = ($56<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($57) {
   $$lcssa$i$1$i = $54;$$lcssa10$i$1$i = $55;
   label = 17;
  } else {
   $58 = (($hash$0$lcssa$i$i$1$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $59 = $58 & 3; //@line 112 "./common/StringMap.h"
   $60 = (((633736 + (($59*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $61 = HEAP8[$60>>0]|0; //@line 114 "./common/StringMap.h"
   $62 = ($61<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($62) {
    $$lcssa$i$1$i = $59;$$lcssa10$i$1$i = $60;
    label = 17;
   } else {
    $63 = (($hash$0$lcssa$i$i$1$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $64 = $63 & 3; //@line 112 "./common/StringMap.h"
    $65 = (((633736 + (($64*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $66 = HEAP8[$65>>0]|0; //@line 114 "./common/StringMap.h"
    $67 = ($66<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($67) {
     $$lcssa$i$1$i = $64;$$lcssa10$i$1$i = $65;
     label = 17;
    }
   }
  }
 }
 if ((label|0) == 17) {
  HEAP8[$$lcssa10$i$1$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $68 = (633736 + (($$lcssa$i$1$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$68>>2] = $38; //@line 118 "./common/StringMap.h"
  $69 = (((633736 + (($$lcssa$i$1$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$69>>2] = $39; //@line 119 "./common/StringMap.h"
 }
 $70 = ($39>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($70) {
  HEAP32[$vararg_buffer2>>2] = $38; //@line 128 "./common/StringMap.h"
  $vararg_ptr5 = ((($vararg_buffer2)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr5>>2] = $39; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer2)|0); //@line 128 "./common/StringMap.h"
  STACKTOP = sp;return;
 } else {
  $71 = (633784 + ($39<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$71>>2] = $38; //@line 132 "./common/StringMap.h"
  STACKTOP = sp;return;
 }
}
function __ZN4love8graphics6opengl5ImageC2ERKNSt3__16vectorIPNS_5image9ImageDataENS3_9allocatorIS7_EEEERKNS2_5FlagsE($this,$imagedata,$flags) {
 $this = $this|0;
 $imagedata = $imagedata|0;
 $flags = $flags|0;
 var $$0 = 0, $$0$i = 0, $$01 = 0, $$01$i = 0, $$1 = 0, $$12 = 0, $$lcssa108 = 0, $$lcssa112 = 0, $$lcssa98 = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i4 = 0, $$pre$i16 = 0, $$pre$i6 = 0, $$pre45 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0.0, $136 = 0, $137 = 0;
 var $138 = 0.0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0;
 var $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0.0, $64 = 0.0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$035 = 0, $eh$lpad$body$index17Z2D = 0, $eh$lpad$body$indexZ2D = 0, $height$09$i = 0, $i$010$i = 0, $i$010$i$lcssa105 = 0, $i$010$i$lcssa106 = 0, $storemerge = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, $vararg_buffer9 = 0, $vararg_ptr12 = 0;
 var $vararg_ptr13 = 0, $vararg_ptr3 = 0, $vararg_ptr7 = 0, $vararg_ptr8 = 0, $width$08$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer9 = sp + 32|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 44|0;
 __ZN4love8graphics7TextureC2Ev($this); //@line 90 "modules/graphics/opengl/Image.cpp"
 $1 = ((($this)) + 120|0); //@line 98 "modules/graphics/opengl/Image.cpp"
 __THREW__ = 0;
 invoke_vi(4011,($1|0)); //@line 90 "modules/graphics/opengl/Image.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $24 = ___cxa_find_matching_catch_2()|0;
  $25 = tempRet0;
  $$1 = $24;$$12 = $25;
  __ZN4love8graphics7TextureD2Ev($this); //@line 115 "modules/graphics/opengl/Image.cpp"
  ___resumeException($$1|0);
  // unreachable;
 }
 HEAP32[$this>>2] = (24820); //@line 98 "modules/graphics/opengl/Image.cpp"
 HEAP32[$1>>2] = (24892); //@line 98 "modules/graphics/opengl/Image.cpp"
 $4 = ((($this)) + 124|0); //@line 90 "modules/graphics/opengl/Image.cpp"
 $5 = ((($this)) + 128|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ((($this)) + 136|0); //@line 90 "modules/graphics/opengl/Image.cpp"
 $7 = ((($this)) + 140|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($this)) + 152|0); //@line 92 "modules/graphics/opengl/Image.cpp"
 ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;HEAP32[$4+12>>2]=0|0;HEAP32[$4+16>>2]=0|0;HEAP32[$4+20>>2]=0|0;HEAP32[$4+24>>2]=0|0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = HEAP32[158433]|0; //@line 92 "modules/graphics/opengl/Image.cpp"
 HEAP32[$8>>2] = $9; //@line 92 "modules/graphics/opengl/Image.cpp"
 $10 = ((($this)) + 156|0); //@line 93 "modules/graphics/opengl/Image.cpp"
 HEAP8[$10>>0] = 0; //@line 93 "modules/graphics/opengl/Image.cpp"
 $11 = ((($this)) + 157|0); //@line 94 "modules/graphics/opengl/Image.cpp"
 $12 = HEAPU8[$flags>>0]|(HEAPU8[$flags+1>>0]<<8); //@line 94 "modules/graphics/opengl/Image.cpp"
 HEAP8[$11>>0]=$12&255;HEAP8[$11+1>>0]=$12>>8; //@line 94 "modules/graphics/opengl/Image.cpp"
 $13 = ((($this)) + 159|0); //@line 95 "modules/graphics/opengl/Image.cpp"
 HEAP8[$13>>0] = 0; //@line 95 "modules/graphics/opengl/Image.cpp"
 $14 = ((($this)) + 160|0); //@line 96 "modules/graphics/opengl/Image.cpp"
 HEAP8[$14>>0] = 0; //@line 96 "modules/graphics/opengl/Image.cpp"
 $15 = ((($this)) + 164|0); //@line 97 "modules/graphics/opengl/Image.cpp"
 HEAP32[$15>>2] = 0; //@line 97 "modules/graphics/opengl/Image.cpp"
 $16 = HEAP32[$imagedata>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ((($imagedata)) + 4|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $18 = HEAP32[$17>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $19 = ($16|0)==($18|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 L5: do {
  if ($19) {
   $20 = (___cxa_allocate_exception(16)|0); //@line 100 "modules/graphics/opengl/Image.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($20|0),(1282466|0),($vararg_buffer|0)); //@line 100 "modules/graphics/opengl/Image.cpp"
   $21 = __THREW__; __THREW__ = 0;
   $22 = $21&1;
   if ($22) {
    $26 = ___cxa_find_matching_catch_2()|0;
    $27 = tempRet0;
    ___cxa_free_exception(($20|0)); //@line 100 "modules/graphics/opengl/Image.cpp"
    $$0 = $26;$$01 = $27;
    break;
   } else {
    __THREW__ = 0;
    invoke_viii(3350,($20|0),(96|0),(344|0)); //@line 100 "modules/graphics/opengl/Image.cpp"
    $23 = __THREW__; __THREW__ = 0;
    label = 9;
    break;
   }
  } else {
   $34 = HEAP32[$16>>2]|0; //@line 102 "modules/graphics/opengl/Image.cpp"
   __THREW__ = 0;
   $35 = (invoke_ii(3984,($34|0))|0); //@line 102 "modules/graphics/opengl/Image.cpp"
   $36 = __THREW__; __THREW__ = 0;
   $37 = $36&1;
   if ($37) {
    label = 9;
   } else {
    $38 = ((($this)) + 8|0); //@line 102 "modules/graphics/opengl/Image.cpp"
    HEAP32[$38>>2] = $35; //@line 102 "modules/graphics/opengl/Image.cpp"
    $39 = HEAP32[$imagedata>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $40 = HEAP32[$39>>2]|0; //@line 103 "modules/graphics/opengl/Image.cpp"
    __THREW__ = 0;
    $41 = (invoke_ii(4081,($40|0))|0); //@line 103 "modules/graphics/opengl/Image.cpp"
    $42 = __THREW__; __THREW__ = 0;
    $43 = $42&1;
    if ($43) {
     label = 9;
    } else {
     $44 = ((($this)) + 12|0); //@line 103 "modules/graphics/opengl/Image.cpp"
     HEAP32[$44>>2] = $41; //@line 103 "modules/graphics/opengl/Image.cpp"
     $45 = HEAP32[$17>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $46 = HEAP32[$imagedata>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $47 = (($45) - ($46))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $48 = $47 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $49 = ($48|0)==(1); //@line 63 "modules/graphics/opengl/Image.cpp"
     $50 = $46; //@line 63 "modules/graphics/opengl/Image.cpp"
     $51 = $45; //@line 63 "modules/graphics/opengl/Image.cpp"
     L13: do {
      if ($49) {
       $124 = $50;$125 = $51;
      } else {
       $52 = HEAP32[$50>>2]|0; //@line 66 "modules/graphics/opengl/Image.cpp"
       __THREW__ = 0;
       $53 = (invoke_ii(3984,($52|0))|0); //@line 66 "modules/graphics/opengl/Image.cpp"
       $54 = __THREW__; __THREW__ = 0;
       $55 = $54&1;
       if ($55) {
        label = 9;
        break L5;
       }
       $56 = HEAP32[$imagedata>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $57 = HEAP32[$56>>2]|0; //@line 67 "modules/graphics/opengl/Image.cpp"
       __THREW__ = 0;
       $58 = (invoke_ii(4081,($57|0))|0); //@line 67 "modules/graphics/opengl/Image.cpp"
       $59 = __THREW__; __THREW__ = 0;
       $60 = $59&1;
       if ($60) {
        label = 9;
        break L5;
       }
       $61 = ($53|0)<($58|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $62 = $61 ? $58 : $53; //@line 55 "modules/graphics/opengl/Image.cpp"
       $63 = (+($62|0)); //@line 55 "modules/graphics/opengl/Image.cpp"
       $64 = (+_log2($63)); //@line 55 "modules/graphics/opengl/Image.cpp"
       $65 = (~~(($64))); //@line 55 "modules/graphics/opengl/Image.cpp"
       $66 = (($65) + 1)|0; //@line 55 "modules/graphics/opengl/Image.cpp"
       $67 = ($48|0)==($66|0); //@line 72 "modules/graphics/opengl/Image.cpp"
       L17: do {
        if ($67) {
         $68 = ($47|0)>(4); //@line 76 "modules/graphics/opengl/Image.cpp"
         L19: do {
          if ($68) {
           $height$09$i = $58;$i$010$i = 1;$width$08$i = $53;
           while(1) {
            $75 = (($width$08$i|0) / 2)&-1; //@line 78 "modules/graphics/opengl/Image.cpp"
            $76 = ($75|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
            $77 = $76 ? 1 : $75; //@line 78 "modules/graphics/opengl/Image.cpp"
            $78 = (($height$09$i|0) / 2)&-1; //@line 79 "modules/graphics/opengl/Image.cpp"
            $79 = ($78|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
            $80 = $79 ? 1 : $78; //@line 79 "modules/graphics/opengl/Image.cpp"
            $81 = HEAP32[$imagedata>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $82 = (($81) + ($i$010$i<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $83 = HEAP32[$82>>2]|0; //@line 81 "modules/graphics/opengl/Image.cpp"
            __THREW__ = 0;
            $84 = (invoke_ii(3984,($83|0))|0); //@line 81 "modules/graphics/opengl/Image.cpp"
            $85 = __THREW__; __THREW__ = 0;
            $86 = $85&1;
            if ($86) {
             label = 8;
             break;
            }
            $87 = ($84|0)==($77|0); //@line 81 "modules/graphics/opengl/Image.cpp"
            if (!($87)) {
             $$lcssa108 = $77;$i$010$i$lcssa105 = $i$010$i;
             label = 23;
             break;
            }
            $101 = HEAP32[$imagedata>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $102 = (($101) + ($i$010$i<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $103 = HEAP32[$102>>2]|0; //@line 83 "modules/graphics/opengl/Image.cpp"
            __THREW__ = 0;
            $104 = (invoke_ii(4081,($103|0))|0); //@line 83 "modules/graphics/opengl/Image.cpp"
            $105 = __THREW__; __THREW__ = 0;
            $106 = $105&1;
            if ($106) {
             label = 8;
             break;
            }
            $107 = ($104|0)==($80|0); //@line 83 "modules/graphics/opengl/Image.cpp"
            if (!($107)) {
             $$lcssa112 = $80;$i$010$i$lcssa106 = $i$010$i;
             label = 29;
             break;
            }
            $121 = (($i$010$i) + 1)|0; //@line 76 "modules/graphics/opengl/Image.cpp"
            $122 = ($121|0)<($48|0); //@line 76 "modules/graphics/opengl/Image.cpp"
            if ($122) {
             $height$09$i = $80;$i$010$i = $121;$width$08$i = $77;
            } else {
             break L19;
            }
           }
           if ((label|0) == 8) {
            $30 = ___cxa_find_matching_catch_2()|0;
            $31 = tempRet0;
            $eh$lpad$body$index17Z2D = $31;$eh$lpad$body$indexZ2D = $30;
            label = 10;
            break L5;
           }
           else if ((label|0) == 23) {
            $88 = (___cxa_allocate_exception(16)|0); //@line 82 "modules/graphics/opengl/Image.cpp"
            $89 = HEAP32[$imagedata>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $90 = (($89) + ($i$010$i$lcssa105<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $91 = HEAP32[$90>>2]|0; //@line 82 "modules/graphics/opengl/Image.cpp"
            __THREW__ = 0;
            $92 = (invoke_ii(3984,($91|0))|0); //@line 82 "modules/graphics/opengl/Image.cpp"
            $93 = __THREW__; __THREW__ = 0;
            $94 = $93&1;
            if (!($94)) {
             $95 = (($i$010$i$lcssa105) + 1)|0; //@line 82 "modules/graphics/opengl/Image.cpp"
             __THREW__ = 0;
             HEAP32[$vararg_buffer4>>2] = $95; //@line 82 "modules/graphics/opengl/Image.cpp"
             $vararg_ptr7 = ((($vararg_buffer4)) + 4|0); //@line 82 "modules/graphics/opengl/Image.cpp"
             HEAP32[$vararg_ptr7>>2] = $$lcssa108; //@line 82 "modules/graphics/opengl/Image.cpp"
             $vararg_ptr8 = ((($vararg_buffer4)) + 8|0); //@line 82 "modules/graphics/opengl/Image.cpp"
             HEAP32[$vararg_ptr8>>2] = $92; //@line 82 "modules/graphics/opengl/Image.cpp"
             invoke_viii(3349,($88|0),(436410|0),($vararg_buffer4|0)); //@line 82 "modules/graphics/opengl/Image.cpp"
             $96 = __THREW__; __THREW__ = 0;
             $97 = $96&1;
             if (!($97)) {
              __THREW__ = 0;
              invoke_viii(3350,($88|0),(96|0),(344|0)); //@line 82 "modules/graphics/opengl/Image.cpp"
              $98 = __THREW__; __THREW__ = 0;
              label = 9;
              break L5;
             }
            }
            $99 = ___cxa_find_matching_catch_2()|0;
            $100 = tempRet0;
            ___cxa_free_exception(($88|0)); //@line 82 "modules/graphics/opengl/Image.cpp"
            $$0$i = $100;$$01$i = $99;
            break L17;
           }
           else if ((label|0) == 29) {
            $108 = (___cxa_allocate_exception(16)|0); //@line 84 "modules/graphics/opengl/Image.cpp"
            $109 = HEAP32[$imagedata>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $110 = (($109) + ($i$010$i$lcssa106<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $111 = HEAP32[$110>>2]|0; //@line 84 "modules/graphics/opengl/Image.cpp"
            __THREW__ = 0;
            $112 = (invoke_ii(4081,($111|0))|0); //@line 84 "modules/graphics/opengl/Image.cpp"
            $113 = __THREW__; __THREW__ = 0;
            $114 = $113&1;
            if (!($114)) {
             $115 = (($i$010$i$lcssa106) + 1)|0; //@line 84 "modules/graphics/opengl/Image.cpp"
             __THREW__ = 0;
             HEAP32[$vararg_buffer9>>2] = $115; //@line 84 "modules/graphics/opengl/Image.cpp"
             $vararg_ptr12 = ((($vararg_buffer9)) + 4|0); //@line 84 "modules/graphics/opengl/Image.cpp"
             HEAP32[$vararg_ptr12>>2] = $$lcssa112; //@line 84 "modules/graphics/opengl/Image.cpp"
             $vararg_ptr13 = ((($vararg_buffer9)) + 8|0); //@line 84 "modules/graphics/opengl/Image.cpp"
             HEAP32[$vararg_ptr13>>2] = $112; //@line 84 "modules/graphics/opengl/Image.cpp"
             invoke_viii(3349,($108|0),(436476|0),($vararg_buffer9|0)); //@line 84 "modules/graphics/opengl/Image.cpp"
             $116 = __THREW__; __THREW__ = 0;
             $117 = $116&1;
             if (!($117)) {
              __THREW__ = 0;
              invoke_viii(3350,($108|0),(96|0),(344|0)); //@line 84 "modules/graphics/opengl/Image.cpp"
              $118 = __THREW__; __THREW__ = 0;
              label = 9;
              break L5;
             }
            }
            $119 = ___cxa_find_matching_catch_2()|0;
            $120 = tempRet0;
            ___cxa_free_exception(($108|0)); //@line 84 "modules/graphics/opengl/Image.cpp"
            $$0$i = $120;$$01$i = $119;
            break L17;
           }
          }
         } while(0);
         HEAP8[$11>>0] = 1; //@line 106 "modules/graphics/opengl/Image.cpp"
         $$pre = HEAP32[$imagedata>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $$pre45 = HEAP32[$17>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $124 = $$pre;$125 = $$pre45;
         break L13;
        } else {
         $69 = (___cxa_allocate_exception(16)|0); //@line 73 "modules/graphics/opengl/Image.cpp"
         __THREW__ = 0;
         HEAP32[$vararg_buffer1>>2] = $66; //@line 73 "modules/graphics/opengl/Image.cpp"
         $vararg_ptr3 = ((($vararg_buffer1)) + 4|0); //@line 73 "modules/graphics/opengl/Image.cpp"
         HEAP32[$vararg_ptr3>>2] = $48; //@line 73 "modules/graphics/opengl/Image.cpp"
         invoke_viii(3349,($69|0),(436341|0),($vararg_buffer1|0)); //@line 73 "modules/graphics/opengl/Image.cpp"
         $70 = __THREW__; __THREW__ = 0;
         $71 = $70&1;
         if ($71) {
          $73 = ___cxa_find_matching_catch_2()|0;
          $74 = tempRet0;
          ___cxa_free_exception(($69|0)); //@line 73 "modules/graphics/opengl/Image.cpp"
          $$0$i = $74;$$01$i = $73;
          break;
         } else {
          __THREW__ = 0;
          invoke_viii(3350,($69|0),(96|0),(344|0)); //@line 73 "modules/graphics/opengl/Image.cpp"
          $72 = __THREW__; __THREW__ = 0;
          label = 9;
          break L5;
         }
        }
       } while(0);
       $eh$lpad$body$index17Z2D = $$0$i;$eh$lpad$body$indexZ2D = $$01$i;
       label = 10;
       break L5;
      }
     } while(0);
     $123 = ($124|0)==($125|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     L45: do {
      if (!($123)) {
       $126 = ((($this)) + 132|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $__begin$sroa$0$035 = $124;
       while(1) {
        $166 = HEAP32[$__begin$sroa$0$035>>2]|0; //@line 106 "./common/Object.h"
        HEAP32[$0>>2] = $166; //@line 93 "./common/Object.h"
        $167 = ($166|0)==(0|0); //@line 95 "./common/Object.h"
        if (!($167)) {
         $168 = HEAP32[$166>>2]|0; //@line 95 "./common/Object.h"
         $169 = ((($168)) + 8|0); //@line 95 "./common/Object.h"
         $170 = HEAP32[$169>>2]|0; //@line 95 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($170|0,($166|0)); //@line 95 "./common/Object.h"
         $171 = __THREW__; __THREW__ = 0;
         $172 = $171&1;
         if ($172) {
          label = 7;
          break;
         }
        }
        $173 = HEAP32[$5>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $174 = HEAP32[$126>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $175 = ($173>>>0)<($174>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($175) {
         HEAP32[$173>>2] = $166; //@line 99 "./common/Object.h"
         if ($167) {
          $184 = $173;
         } else {
          $176 = HEAP32[$166>>2]|0; //@line 101 "./common/Object.h"
          $177 = ((($176)) + 8|0); //@line 101 "./common/Object.h"
          $178 = HEAP32[$177>>2]|0; //@line 101 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($178|0,($166|0)); //@line 101 "./common/Object.h"
          $179 = __THREW__; __THREW__ = 0;
          $180 = $179&1;
          if ($180) {
           $$lcssa98 = $166;
           label = 52;
           break;
          }
          $$pre$i16 = HEAP32[$5>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $184 = $$pre$i16;
         }
         $183 = ((($184)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$5>>2] = $183; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        } else {
         __THREW__ = 0;
         invoke_vii(4082,($4|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $185 = __THREW__; __THREW__ = 0;
         $186 = $185&1;
         if ($186) {
          label = 60;
          break;
         }
        }
        if (!($167)) {
         $187 = HEAP32[$166>>2]|0; //@line 106 "./common/Object.h"
         $188 = ((($187)) + 12|0); //@line 106 "./common/Object.h"
         $189 = HEAP32[$188>>2]|0; //@line 106 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($189|0,($166|0)); //@line 106 "./common/Object.h"
         $190 = __THREW__; __THREW__ = 0;
         $191 = $190&1;
         if ($191) {
          label = 58;
          break;
         }
        }
        $194 = ((($__begin$sroa$0$035)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        $195 = ($194|0)==($125|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        if ($195) {
         break L45;
        } else {
         $__begin$sroa$0$035 = $194;
        }
       }
       if ((label|0) == 7) {
        $28 = ___cxa_find_matching_catch_2()|0;
        $29 = tempRet0;
        $eh$lpad$body$index17Z2D = $29;$eh$lpad$body$indexZ2D = $28;
        label = 10;
        break L5;
       }
       else if ((label|0) == 52) {
        $181 = ___cxa_find_matching_catch_2()|0;
        $182 = tempRet0;
        $199 = $$lcssa98;$249 = $182;$250 = $181;
       }
       else if ((label|0) == 58) {
        $192 = ___cxa_find_matching_catch_3(0|0)|0;
        $193 = tempRet0;
        ___clang_call_terminate($192); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
       else if ((label|0) == 60) {
        $196 = ___cxa_find_matching_catch_2()|0;
        $197 = tempRet0;
        if ($167) {
         $$0 = $196;$$01 = $197;
         break L5;
        } else {
         $199 = $166;$249 = $197;$250 = $196;
        }
       }
       $198 = HEAP32[$199>>2]|0; //@line 106 "./common/Object.h"
       $200 = ((($198)) + 12|0); //@line 106 "./common/Object.h"
       $201 = HEAP32[$200>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($201|0,($199|0)); //@line 106 "./common/Object.h"
       $202 = __THREW__; __THREW__ = 0;
       $203 = $202&1;
       if (!($203)) {
        $$0 = $250;$$01 = $249;
        break L5;
       }
       $204 = ___cxa_find_matching_catch_3(0|0)|0;
       $205 = tempRet0;
       ___clang_call_terminate($204); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
     } while(0);
     $127 = ((($this)) + 56|0); //@line 161 "modules/graphics/opengl/Image.cpp"
     $128 = ((($this)) + 76|0); //@line 161 "modules/graphics/opengl/Image.cpp"
     $129 = ((($this)) + 96|0); //@line 161 "modules/graphics/opengl/Image.cpp"
     $130 = ((($this)) + 116|0); //@line 161 "modules/graphics/opengl/Image.cpp"
     $131 = ((($this)) + 40|0); //@line 168 "modules/graphics/opengl/Image.cpp"
     HEAP32[$127>>2] = -1; //@line 161 "modules/graphics/opengl/Image.cpp"
     HEAP32[$128>>2] = -1; //@line 161 "modules/graphics/opengl/Image.cpp"
     HEAP32[$129>>2] = -1; //@line 161 "modules/graphics/opengl/Image.cpp"
     HEAP32[$130>>2] = -1; //@line 161 "modules/graphics/opengl/Image.cpp"
     HEAPF32[$131>>2] = 0.0; //@line 168 "modules/graphics/opengl/Image.cpp"
     $132 = ((($this)) + 44|0); //@line 169 "modules/graphics/opengl/Image.cpp"
     HEAPF32[$132>>2] = 0.0; //@line 169 "modules/graphics/opengl/Image.cpp"
     $133 = ((($this)) + 60|0); //@line 170 "modules/graphics/opengl/Image.cpp"
     HEAPF32[$133>>2] = 0.0; //@line 170 "modules/graphics/opengl/Image.cpp"
     $134 = HEAP32[$44>>2]|0; //@line 171 "modules/graphics/opengl/Image.cpp"
     $135 = (+($134|0)); //@line 171 "modules/graphics/opengl/Image.cpp"
     $136 = ((($this)) + 64|0); //@line 171 "modules/graphics/opengl/Image.cpp"
     HEAPF32[$136>>2] = $135; //@line 171 "modules/graphics/opengl/Image.cpp"
     $137 = HEAP32[$38>>2]|0; //@line 172 "modules/graphics/opengl/Image.cpp"
     $138 = (+($137|0)); //@line 172 "modules/graphics/opengl/Image.cpp"
     $139 = ((($this)) + 80|0); //@line 172 "modules/graphics/opengl/Image.cpp"
     HEAPF32[$139>>2] = $138; //@line 172 "modules/graphics/opengl/Image.cpp"
     $140 = ((($this)) + 84|0); //@line 173 "modules/graphics/opengl/Image.cpp"
     HEAPF32[$140>>2] = 0.0; //@line 173 "modules/graphics/opengl/Image.cpp"
     $141 = ((($this)) + 100|0); //@line 174 "modules/graphics/opengl/Image.cpp"
     HEAPF32[$141>>2] = $138; //@line 174 "modules/graphics/opengl/Image.cpp"
     $142 = ((($this)) + 104|0); //@line 175 "modules/graphics/opengl/Image.cpp"
     HEAPF32[$142>>2] = $135; //@line 175 "modules/graphics/opengl/Image.cpp"
     $143 = ((($this)) + 48|0); //@line 177 "modules/graphics/opengl/Image.cpp"
     HEAPF32[$143>>2] = 0.0; //@line 177 "modules/graphics/opengl/Image.cpp"
     $144 = ((($this)) + 52|0); //@line 178 "modules/graphics/opengl/Image.cpp"
     HEAPF32[$144>>2] = 0.0; //@line 178 "modules/graphics/opengl/Image.cpp"
     $145 = ((($this)) + 68|0); //@line 179 "modules/graphics/opengl/Image.cpp"
     HEAPF32[$145>>2] = 0.0; //@line 179 "modules/graphics/opengl/Image.cpp"
     $146 = ((($this)) + 72|0); //@line 180 "modules/graphics/opengl/Image.cpp"
     HEAPF32[$146>>2] = 1.0; //@line 180 "modules/graphics/opengl/Image.cpp"
     $147 = ((($this)) + 88|0); //@line 181 "modules/graphics/opengl/Image.cpp"
     HEAPF32[$147>>2] = 1.0; //@line 181 "modules/graphics/opengl/Image.cpp"
     $148 = ((($this)) + 92|0); //@line 182 "modules/graphics/opengl/Image.cpp"
     HEAPF32[$148>>2] = 0.0; //@line 182 "modules/graphics/opengl/Image.cpp"
     $149 = ((($this)) + 108|0); //@line 183 "modules/graphics/opengl/Image.cpp"
     HEAPF32[$149>>2] = 1.0; //@line 183 "modules/graphics/opengl/Image.cpp"
     $150 = ((($this)) + 112|0); //@line 184 "modules/graphics/opengl/Image.cpp"
     HEAPF32[$150>>2] = 1.0; //@line 184 "modules/graphics/opengl/Image.cpp"
     $151 = ((($this)) + 157|0); //@line 186 "modules/graphics/opengl/Image.cpp"
     $152 = HEAP8[$151>>0]|0; //@line 186 "modules/graphics/opengl/Image.cpp"
     $153 = ($152<<24>>24)==(0); //@line 186 "modules/graphics/opengl/Image.cpp"
     if (!($153)) {
      $154 = HEAP32[6202]|0; //@line 187 "modules/graphics/opengl/Image.cpp"
      $155 = ((($this)) + 24|0); //@line 187 "modules/graphics/opengl/Image.cpp"
      HEAP32[$155>>2] = $154; //@line 187 "modules/graphics/opengl/Image.cpp"
     }
     __THREW__ = 0;
     $156 = (invoke_i(3801)|0); //@line 189 "modules/graphics/opengl/Image.cpp"
     $157 = __THREW__; __THREW__ = 0;
     $158 = $157&1;
     if ($158) {
      label = 9;
     } else {
      if (!($156)) {
       $159 = ((($this)) + 158|0); //@line 190 "modules/graphics/opengl/Image.cpp"
       HEAP8[$159>>0] = 0; //@line 190 "modules/graphics/opengl/Image.cpp"
      }
      __THREW__ = 0;
      $160 = (invoke_i(3801)|0); //@line 192 "modules/graphics/opengl/Image.cpp"
      $161 = __THREW__; __THREW__ = 0;
      $162 = $161&1;
      if ($162) {
       label = 9;
      } else {
       if ($160) {
        $163 = ((($this)) + 158|0); //@line 192 "modules/graphics/opengl/Image.cpp"
        $164 = HEAP8[$163>>0]|0; //@line 192 "modules/graphics/opengl/Image.cpp"
        $165 = ($164<<24>>24)==(0); //@line 192 "modules/graphics/opengl/Image.cpp"
        if ($165) {
         $storemerge = 1;
        } else {
         label = 46;
        }
       } else {
        label = 46;
       }
       if ((label|0) == 46) {
        $storemerge = 0;
       }
       HEAP8[$13>>0] = $storemerge; //@line 193 "modules/graphics/opengl/Image.cpp"
       $206 = HEAP32[$this>>2]|0; //@line 112 "modules/graphics/opengl/Image.cpp"
       $207 = ((($206)) + 56|0); //@line 112 "modules/graphics/opengl/Image.cpp"
       $208 = HEAP32[$207>>2]|0; //@line 112 "modules/graphics/opengl/Image.cpp"
       __THREW__ = 0;
       (invoke_ii($208|0,($this|0))|0); //@line 112 "modules/graphics/opengl/Image.cpp"
       $209 = __THREW__; __THREW__ = 0;
       $210 = $209&1;
       if ($210) {
        label = 9;
       } else {
        $211 = HEAP32[158431]|0; //@line 114 "modules/graphics/opengl/Image.cpp"
        $212 = (($211) + 1)|0; //@line 114 "modules/graphics/opengl/Image.cpp"
        HEAP32[158431] = $212; //@line 114 "modules/graphics/opengl/Image.cpp"
        STACKTOP = sp;return; //@line 115 "modules/graphics/opengl/Image.cpp"
       }
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 9) {
  $32 = ___cxa_find_matching_catch_2()|0;
  $33 = tempRet0;
  $eh$lpad$body$index17Z2D = $33;$eh$lpad$body$indexZ2D = $32;
  label = 10;
 }
 if ((label|0) == 10) {
  $$0 = $eh$lpad$body$indexZ2D;$$01 = $eh$lpad$body$index17Z2D;
 }
 $213 = HEAP32[$6>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $214 = ($213|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($214)) {
  $215 = HEAP32[$7>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $216 = ($215|0)==($213|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  do {
   if ($216) {
    $230 = $213;
   } else {
    $218 = $215;
    while(1) {
     $217 = ((($218)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$7>>2] = $217; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $219 = HEAP32[$217>>2]|0; //@line 106 "./common/Object.h"
     $220 = ($219|0)==(0|0); //@line 106 "./common/Object.h"
     if ($220) {
      $222 = $217;
     } else {
      $223 = HEAP32[$219>>2]|0; //@line 106 "./common/Object.h"
      $224 = ((($223)) + 12|0); //@line 106 "./common/Object.h"
      $225 = HEAP32[$224>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($225|0,($219|0)); //@line 106 "./common/Object.h"
      $226 = __THREW__; __THREW__ = 0;
      $227 = $226&1;
      if ($227) {
       label = 71;
       break;
      }
      $$pre$i$i$i4 = HEAP32[$7>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $222 = $$pre$i$i$i4;
     }
     $221 = ($222|0)==($213|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($221) {
      label = 72;
      break;
     } else {
      $218 = $222;
     }
    }
    if ((label|0) == 71) {
     $228 = ___cxa_find_matching_catch_3(0|0)|0;
     $229 = tempRet0;
     ___clang_call_terminate($228); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
    else if ((label|0) == 72) {
     $$pre$i6 = HEAP32[$6>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $230 = $$pre$i6;
     break;
    }
   }
  } while(0);
  __ZdlPv($230); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $231 = HEAP32[$4>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $232 = ($231|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($232)) {
  $233 = HEAP32[$5>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $234 = ($233|0)==($231|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  do {
   if ($234) {
    $248 = $231;
   } else {
    $236 = $233;
    while(1) {
     $235 = ((($236)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$5>>2] = $235; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $237 = HEAP32[$235>>2]|0; //@line 106 "./common/Object.h"
     $238 = ($237|0)==(0|0); //@line 106 "./common/Object.h"
     if ($238) {
      $240 = $235;
     } else {
      $241 = HEAP32[$237>>2]|0; //@line 106 "./common/Object.h"
      $242 = ((($241)) + 12|0); //@line 106 "./common/Object.h"
      $243 = HEAP32[$242>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($243|0,($237|0)); //@line 106 "./common/Object.h"
      $244 = __THREW__; __THREW__ = 0;
      $245 = $244&1;
      if ($245) {
       label = 80;
       break;
      }
      $$pre$i$i$i = HEAP32[$5>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $240 = $$pre$i$i$i;
     }
     $239 = ($240|0)==($231|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($239) {
      label = 81;
      break;
     } else {
      $236 = $240;
     }
    }
    if ((label|0) == 80) {
     $246 = ___cxa_find_matching_catch_3(0|0)|0;
     $247 = tempRet0;
     ___clang_call_terminate($246); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
    else if ((label|0) == 81) {
     $$pre$i = HEAP32[$4>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $248 = $$pre$i;
     break;
    }
   }
  } while(0);
  __ZdlPv($248); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 __ZN4love8graphics8VolatileD2Ev($1); //@line 115 "modules/graphics/opengl/Image.cpp"
 $$1 = $$0;$$12 = $$01;
 __ZN4love8graphics7TextureD2Ev($this); //@line 115 "modules/graphics/opengl/Image.cpp"
 ___resumeException($$1|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love9StrongRefINS1_5image9ImageDataEEENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i8 = 0, $$0$i9 = 0, $$01$i$i = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre12 = 0, $$pre13 = 0, $$pre14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__v$sroa$15$0 = 0, $__v$sroa$6$0 = 0, $__v$sroa$6$0$lcssa = 0, $__v$sroa$6$1 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = $$in; //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = HEAP32[$8>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $10 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($11>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $17 = $10 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ($17>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $19 = $18 ? $5 : $17; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = (($20) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = $21 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = ($19|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $24 = $20; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($23) {
   $$0$i9 = 0;$101 = $24;$28 = 0;$29 = $22;
  } else {
   $$0$i8 = $19;$100 = $22;$99 = $24;
   label = 6;
  }
 } else {
  $13 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = $14 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = $13; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $$0$i8 = 1073741823;$100 = $15;$99 = $16;
  label = 6;
 }
 if ((label|0) == 6) {
  $25 = $$0$i8 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (__Znwj($25)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i9 = $$0$i8;$101 = $99;$28 = $26;$29 = $100;
 }
 $27 = (($28) + ($29<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = $27; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $31 = (($28) + ($$0$i9<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $32 = $31; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $33 = HEAP32[$__x>>2]|0; //@line 139 "./common/Object.h"
 HEAP32[$27>>2] = $33; //@line 99 "./common/Object.h"
 $34 = ($33|0)==(0|0); //@line 101 "./common/Object.h"
 do {
  if ($34) {
   $43 = $101;$44 = $7;
   label = 10;
  } else {
   $35 = HEAP32[$33>>2]|0; //@line 101 "./common/Object.h"
   $36 = ((($35)) + 8|0); //@line 101 "./common/Object.h"
   $37 = HEAP32[$36>>2]|0; //@line 101 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($37|0,($33|0)); //@line 101 "./common/Object.h"
   $38 = __THREW__; __THREW__ = 0;
   $39 = $38&1;
   if ($39) {
    $81 = ___cxa_find_matching_catch_2()|0;
    $82 = tempRet0;
    $__v$sroa$15$0 = $30;$__v$sroa$6$1 = $30;$lpad$phi$index = $81;$lpad$phi$index2 = $82;
    break;
   } else {
    $$pre12 = HEAP32[$this>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre13 = HEAP32[$0>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $43 = $$pre13;$44 = $$pre12;
    label = 10;
    break;
   }
  }
 } while(0);
 L15: do {
  if ((label|0) == 10) {
   $40 = ((($27)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $41 = $40; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $42 = ($43|0)==($44|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   do {
    if ($42) {
     $45 = $43; //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $60 = $30;$62 = $45;$64 = $43;
    } else {
     $$01$i$i = $43;$48 = $27;$__v$sroa$6$0 = $30;
     while(1) {
      $46 = ((($$01$i$i)) + -4|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $47 = ((($48)) + -4|0); //@line 99 "./common/Object.h"
      $49 = HEAP32[$46>>2]|0; //@line 139 "./common/Object.h"
      HEAP32[$47>>2] = $49; //@line 99 "./common/Object.h"
      $50 = ($49|0)==(0|0); //@line 101 "./common/Object.h"
      if (!($50)) {
       $51 = HEAP32[$49>>2]|0; //@line 101 "./common/Object.h"
       $52 = ((($51)) + 8|0); //@line 101 "./common/Object.h"
       $53 = HEAP32[$52>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($53|0,($49|0)); //@line 101 "./common/Object.h"
       $54 = __THREW__; __THREW__ = 0;
       $55 = $54&1;
       if ($55) {
        $__v$sroa$6$0$lcssa = $__v$sroa$6$0;
        label = 24;
        break;
       }
      }
      $56 = $__v$sroa$6$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $57 = ((($56)) + -4|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $58 = $57; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $59 = ($46|0)==($44|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      if ($59) {
       $$lcssa = $58;
       label = 15;
       break;
      } else {
       $$01$i$i = $46;$48 = $57;$__v$sroa$6$0 = $58;
      }
     }
     if ((label|0) == 15) {
      $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $$pre14 = HEAP32[$0>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $60 = $$lcssa;$62 = $$pre$i;$64 = $$pre14;
      break;
     }
     else if ((label|0) == 24) {
      $79 = ___cxa_find_matching_catch_2()|0;
      $80 = tempRet0;
      $__v$sroa$15$0 = $41;$__v$sroa$6$1 = $__v$sroa$6$0$lcssa;$lpad$phi$index = $79;$lpad$phi$index2 = $80;
      break L15;
     }
    }
   } while(0);
   HEAP32[$this>>2] = $60; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$0>>2] = $41; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$8>>2] = $32; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $61 = $62; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $63 = ($64|0)==($61|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   L28: do {
    if (!($63)) {
     $66 = $64;
     while(1) {
      $65 = ((($66)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $67 = HEAP32[$65>>2]|0; //@line 106 "./common/Object.h"
      $68 = ($67|0)==(0|0); //@line 106 "./common/Object.h"
      if (!($68)) {
       $70 = HEAP32[$67>>2]|0; //@line 106 "./common/Object.h"
       $71 = ((($70)) + 12|0); //@line 106 "./common/Object.h"
       $72 = HEAP32[$71>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($72|0,($67|0)); //@line 106 "./common/Object.h"
       $73 = __THREW__; __THREW__ = 0;
       $74 = $73&1;
       if ($74) {
        break;
       }
      }
      $69 = ($65|0)==($61|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if ($69) {
       break L28;
      } else {
       $66 = $65;
      }
     }
     $75 = ___cxa_find_matching_catch_3(0|0)|0;
     $76 = tempRet0;
     ___clang_call_terminate($75); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   } while(0);
   $77 = ($62|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($77) {
    return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $78 = $62; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __ZdlPv($78); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
 } while(0);
 $83 = $__v$sroa$6$1; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $84 = $__v$sroa$15$0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $85 = ($84|0)==($83|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 L41: do {
  if (!($85)) {
   $87 = $84;
   while(1) {
    $86 = ((($87)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $88 = HEAP32[$86>>2]|0; //@line 106 "./common/Object.h"
    $89 = ($88|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($89)) {
     $91 = HEAP32[$88>>2]|0; //@line 106 "./common/Object.h"
     $92 = ((($91)) + 12|0); //@line 106 "./common/Object.h"
     $93 = HEAP32[$92>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($93|0,($88|0)); //@line 106 "./common/Object.h"
     $94 = __THREW__; __THREW__ = 0;
     $95 = $94&1;
     if ($95) {
      break;
     }
    }
    $90 = ($86|0)==($83|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($90) {
     break L41;
    } else {
     $87 = $86;
    }
   }
   $96 = ___cxa_find_matching_catch_3(0|0)|0;
   $97 = tempRet0;
   ___clang_call_terminate($96); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 } while(0);
 $98 = ($28|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($98) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 __ZdlPv($28); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZN4love8graphics6opengl5ImageC2ERKNSt3__16vectorIPNS_5image19CompressedImageDataENS3_9allocatorIS7_EEEERKNS2_5FlagsE($this,$compresseddata,$flags) {
 $this = $this|0;
 $compresseddata = $compresseddata|0;
 $flags = $flags|0;
 var $$0 = 0, $$0$i = 0, $$01 = 0, $$01$i = 0, $$1 = 0, $$12 = 0, $$lcssa104 = 0, $$lcssa116 = 0, $$lcssa120 = 0, $$pre$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i4 = 0, $$pre$i16 = 0, $$pre$i6 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0.0, $124 = 0.0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0.0, $155 = 0, $156 = 0, $157 = 0.0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0;
 var $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0;
 var $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0;
 var $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0;
 var $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0;
 var $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0.0;
 var $47 = 0.0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$036 = 0, $eh$lpad$body$index19Z2D = 0;
 var $eh$lpad$body$indexZ2D = 0, $height$09$i = 0, $i$010$i = 0, $i$010$i$lcssa113 = 0, $i$010$i$lcssa114 = 0, $storemerge = 0, $vararg_buffer = 0, $vararg_buffer12 = 0, $vararg_buffer14 = 0, $vararg_buffer2 = 0, $vararg_buffer7 = 0, $vararg_ptr1 = 0, $vararg_ptr10 = 0, $vararg_ptr11 = 0, $vararg_ptr5 = 0, $vararg_ptr6 = 0, $width$08$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer14 = sp + 48|0;
 $vararg_buffer12 = sp + 40|0;
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 52|0;
 __ZN4love8graphics7TextureC2Ev($this); //@line 117 "modules/graphics/opengl/Image.cpp"
 $1 = ((($this)) + 120|0); //@line 125 "modules/graphics/opengl/Image.cpp"
 __THREW__ = 0;
 invoke_vi(4011,($1|0)); //@line 117 "modules/graphics/opengl/Image.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $106 = ___cxa_find_matching_catch_2()|0;
  $107 = tempRet0;
  $$1 = $106;$$12 = $107;
  __ZN4love8graphics7TextureD2Ev($this); //@line 150 "modules/graphics/opengl/Image.cpp"
  ___resumeException($$1|0);
  // unreachable;
 }
 HEAP32[$this>>2] = (24820); //@line 125 "modules/graphics/opengl/Image.cpp"
 HEAP32[$1>>2] = (24892); //@line 125 "modules/graphics/opengl/Image.cpp"
 $4 = ((($this)) + 124|0); //@line 117 "modules/graphics/opengl/Image.cpp"
 $5 = ((($this)) + 128|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ((($this)) + 136|0); //@line 117 "modules/graphics/opengl/Image.cpp"
 $7 = ((($this)) + 140|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($this)) + 152|0); //@line 119 "modules/graphics/opengl/Image.cpp"
 ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;HEAP32[$4+12>>2]=0|0;HEAP32[$4+16>>2]=0|0;HEAP32[$4+20>>2]=0|0;HEAP32[$4+24>>2]=0|0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = HEAP32[158433]|0; //@line 119 "modules/graphics/opengl/Image.cpp"
 HEAP32[$8>>2] = $9; //@line 119 "modules/graphics/opengl/Image.cpp"
 $10 = ((($this)) + 156|0); //@line 120 "modules/graphics/opengl/Image.cpp"
 HEAP8[$10>>0] = 1; //@line 120 "modules/graphics/opengl/Image.cpp"
 $11 = ((($this)) + 157|0); //@line 121 "modules/graphics/opengl/Image.cpp"
 $12 = HEAPU8[$flags>>0]|(HEAPU8[$flags+1>>0]<<8); //@line 121 "modules/graphics/opengl/Image.cpp"
 HEAP8[$11>>0]=$12&255;HEAP8[$11+1>>0]=$12>>8; //@line 121 "modules/graphics/opengl/Image.cpp"
 $13 = ((($this)) + 159|0); //@line 122 "modules/graphics/opengl/Image.cpp"
 HEAP8[$13>>0] = 0; //@line 122 "modules/graphics/opengl/Image.cpp"
 $14 = ((($this)) + 160|0); //@line 123 "modules/graphics/opengl/Image.cpp"
 HEAP8[$14>>0] = 0; //@line 123 "modules/graphics/opengl/Image.cpp"
 $15 = ((($this)) + 164|0); //@line 124 "modules/graphics/opengl/Image.cpp"
 HEAP32[$15>>2] = 0; //@line 124 "modules/graphics/opengl/Image.cpp"
 $16 = HEAP32[$compresseddata>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = HEAP32[$16>>2]|0; //@line 126 "modules/graphics/opengl/Image.cpp"
 __THREW__ = 0;
 $18 = (invoke_iii(3635,($17|0),0)|0); //@line 126 "modules/graphics/opengl/Image.cpp"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 L5: do {
  if ($20) {
   label = 30;
  } else {
   $21 = ((($this)) + 8|0); //@line 126 "modules/graphics/opengl/Image.cpp"
   HEAP32[$21>>2] = $18; //@line 126 "modules/graphics/opengl/Image.cpp"
   $22 = HEAP32[$compresseddata>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $23 = HEAP32[$22>>2]|0; //@line 127 "modules/graphics/opengl/Image.cpp"
   __THREW__ = 0;
   $24 = (invoke_iii(3636,($23|0),0)|0); //@line 131 "modules/graphics/opengl/Image.cpp"
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    label = 30;
   } else {
    $27 = ((($this)) + 12|0); //@line 127 "modules/graphics/opengl/Image.cpp"
    HEAP32[$27>>2] = $24; //@line 127 "modules/graphics/opengl/Image.cpp"
    $28 = ((($compresseddata)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $29 = HEAP32[$28>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $30 = HEAP32[$compresseddata>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $31 = (($29) - ($30))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $32 = $31 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $33 = ($32|0)==(1); //@line 63 "modules/graphics/opengl/Image.cpp"
    $34 = $30; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    L8: do {
     if ($33) {
      $114 = HEAP8[$flags>>0]|0; //@line 131 "modules/graphics/opengl/Image.cpp"
      $115 = ($114<<24>>24)==(0); //@line 131 "modules/graphics/opengl/Image.cpp"
      if (!($115)) {
       $116 = HEAP32[$21>>2]|0; //@line 131 "modules/graphics/opengl/Image.cpp"
       $117 = HEAP32[$34>>2]|0; //@line 131 "modules/graphics/opengl/Image.cpp"
       __THREW__ = 0;
       $118 = (invoke_ii(4083,($117|0))|0); //@line 131 "modules/graphics/opengl/Image.cpp"
       $119 = __THREW__; __THREW__ = 0;
       $120 = $119&1;
       if ($120) {
        label = 30;
        break L5;
       }
       $121 = ($116|0)<($24|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $122 = $121 ? $24 : $116; //@line 55 "modules/graphics/opengl/Image.cpp"
       $123 = (+($122|0)); //@line 55 "modules/graphics/opengl/Image.cpp"
       $124 = (+_log2($123)); //@line 55 "modules/graphics/opengl/Image.cpp"
       $125 = (~~(($124))); //@line 55 "modules/graphics/opengl/Image.cpp"
       $126 = (($125) + 1)|0; //@line 55 "modules/graphics/opengl/Image.cpp"
       $127 = ($126|0)==($118|0); //@line 131 "modules/graphics/opengl/Image.cpp"
       if (!($127)) {
        $128 = HEAP32[$compresseddata>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $129 = HEAP32[$128>>2]|0; //@line 133 "modules/graphics/opengl/Image.cpp"
        __THREW__ = 0;
        $130 = (invoke_ii(4083,($129|0))|0); //@line 133 "modules/graphics/opengl/Image.cpp"
        $131 = __THREW__; __THREW__ = 0;
        $132 = $131&1;
        if ($132) {
         label = 30;
         break L5;
        }
        $133 = ($130|0)==(1); //@line 133 "modules/graphics/opengl/Image.cpp"
        if ($133) {
         HEAP8[$11>>0] = 0; //@line 134 "modules/graphics/opengl/Image.cpp"
         break;
        }
        $134 = (___cxa_allocate_exception(16)|0); //@line 136 "modules/graphics/opengl/Image.cpp"
        __THREW__ = 0;
        invoke_viii(3349,($134|0),(436543|0),($vararg_buffer12|0)); //@line 136 "modules/graphics/opengl/Image.cpp"
        $135 = __THREW__; __THREW__ = 0;
        $136 = $135&1;
        if ($136) {
         $138 = ___cxa_find_matching_catch_2()|0;
         $139 = tempRet0;
         ___cxa_free_exception(($134|0)); //@line 136 "modules/graphics/opengl/Image.cpp"
         $$0 = $138;$$01 = $139;
         break L5;
        } else {
         __THREW__ = 0;
         invoke_viii(3350,($134|0),(96|0),(344|0)); //@line 136 "modules/graphics/opengl/Image.cpp"
         $137 = __THREW__; __THREW__ = 0;
         label = 30;
         break L5;
        }
       }
      }
     } else {
      $35 = HEAP32[$34>>2]|0; //@line 66 "modules/graphics/opengl/Image.cpp"
      __THREW__ = 0;
      $36 = (invoke_iii(3635,($35|0),0)|0); //@line 66 "modules/graphics/opengl/Image.cpp"
      $37 = __THREW__; __THREW__ = 0;
      $38 = $37&1;
      if ($38) {
       label = 30;
       break L5;
      }
      $39 = HEAP32[$compresseddata>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $40 = HEAP32[$39>>2]|0; //@line 67 "modules/graphics/opengl/Image.cpp"
      __THREW__ = 0;
      $41 = (invoke_iii(3636,($40|0),0)|0); //@line 67 "modules/graphics/opengl/Image.cpp"
      $42 = __THREW__; __THREW__ = 0;
      $43 = $42&1;
      if ($43) {
       label = 30;
       break L5;
      }
      $44 = ($36|0)<($41|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $45 = $44 ? $41 : $36; //@line 55 "modules/graphics/opengl/Image.cpp"
      $46 = (+($45|0)); //@line 55 "modules/graphics/opengl/Image.cpp"
      $47 = (+_log2($46)); //@line 55 "modules/graphics/opengl/Image.cpp"
      $48 = (~~(($47))); //@line 55 "modules/graphics/opengl/Image.cpp"
      $49 = (($48) + 1)|0; //@line 55 "modules/graphics/opengl/Image.cpp"
      $50 = ($32|0)==($49|0); //@line 72 "modules/graphics/opengl/Image.cpp"
      L23: do {
       if ($50) {
        $51 = ($31|0)>(4); //@line 76 "modules/graphics/opengl/Image.cpp"
        L25: do {
         if ($51) {
          $height$09$i = $41;$i$010$i = 1;$width$08$i = $36;
          while(1) {
           $58 = (($width$08$i|0) / 2)&-1; //@line 78 "modules/graphics/opengl/Image.cpp"
           $59 = ($58|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
           $60 = $59 ? 1 : $58; //@line 78 "modules/graphics/opengl/Image.cpp"
           $61 = (($height$09$i|0) / 2)&-1; //@line 79 "modules/graphics/opengl/Image.cpp"
           $62 = ($61|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
           $63 = $62 ? 1 : $61; //@line 79 "modules/graphics/opengl/Image.cpp"
           $64 = HEAP32[$compresseddata>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $65 = (($64) + ($i$010$i<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $66 = HEAP32[$65>>2]|0; //@line 81 "modules/graphics/opengl/Image.cpp"
           __THREW__ = 0;
           $67 = (invoke_iii(3635,($66|0),0)|0); //@line 81 "modules/graphics/opengl/Image.cpp"
           $68 = __THREW__; __THREW__ = 0;
           $69 = $68&1;
           if ($69) {
            label = 29;
            break;
           }
           $70 = ($67|0)==($60|0); //@line 81 "modules/graphics/opengl/Image.cpp"
           if (!($70)) {
            $$lcssa116 = $60;$i$010$i$lcssa113 = $i$010$i;
            label = 14;
            break;
           }
           $84 = HEAP32[$compresseddata>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $85 = (($84) + ($i$010$i<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $86 = HEAP32[$85>>2]|0; //@line 83 "modules/graphics/opengl/Image.cpp"
           __THREW__ = 0;
           $87 = (invoke_iii(3636,($86|0),0)|0); //@line 83 "modules/graphics/opengl/Image.cpp"
           $88 = __THREW__; __THREW__ = 0;
           $89 = $88&1;
           if ($89) {
            label = 29;
            break;
           }
           $90 = ($87|0)==($63|0); //@line 83 "modules/graphics/opengl/Image.cpp"
           if (!($90)) {
            $$lcssa120 = $63;$i$010$i$lcssa114 = $i$010$i;
            label = 20;
            break;
           }
           $104 = (($i$010$i) + 1)|0; //@line 76 "modules/graphics/opengl/Image.cpp"
           $105 = ($104|0)<($32|0); //@line 76 "modules/graphics/opengl/Image.cpp"
           if ($105) {
            $height$09$i = $63;$i$010$i = $104;$width$08$i = $60;
           } else {
            break L25;
           }
          }
          if ((label|0) == 14) {
           $71 = (___cxa_allocate_exception(16)|0); //@line 82 "modules/graphics/opengl/Image.cpp"
           $72 = HEAP32[$compresseddata>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $73 = (($72) + ($i$010$i$lcssa113<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $74 = HEAP32[$73>>2]|0; //@line 82 "modules/graphics/opengl/Image.cpp"
           __THREW__ = 0;
           $75 = (invoke_iii(3635,($74|0),0)|0); //@line 82 "modules/graphics/opengl/Image.cpp"
           $76 = __THREW__; __THREW__ = 0;
           $77 = $76&1;
           if (!($77)) {
            $78 = (($i$010$i$lcssa113) + 1)|0; //@line 82 "modules/graphics/opengl/Image.cpp"
            __THREW__ = 0;
            HEAP32[$vararg_buffer2>>2] = $78; //@line 82 "modules/graphics/opengl/Image.cpp"
            $vararg_ptr5 = ((($vararg_buffer2)) + 4|0); //@line 82 "modules/graphics/opengl/Image.cpp"
            HEAP32[$vararg_ptr5>>2] = $$lcssa116; //@line 82 "modules/graphics/opengl/Image.cpp"
            $vararg_ptr6 = ((($vararg_buffer2)) + 8|0); //@line 82 "modules/graphics/opengl/Image.cpp"
            HEAP32[$vararg_ptr6>>2] = $75; //@line 82 "modules/graphics/opengl/Image.cpp"
            invoke_viii(3349,($71|0),(436410|0),($vararg_buffer2|0)); //@line 82 "modules/graphics/opengl/Image.cpp"
            $79 = __THREW__; __THREW__ = 0;
            $80 = $79&1;
            if (!($80)) {
             __THREW__ = 0;
             invoke_viii(3350,($71|0),(96|0),(344|0)); //@line 82 "modules/graphics/opengl/Image.cpp"
             $81 = __THREW__; __THREW__ = 0;
             label = 30;
             break L5;
            }
           }
           $82 = ___cxa_find_matching_catch_2()|0;
           $83 = tempRet0;
           ___cxa_free_exception(($71|0)); //@line 82 "modules/graphics/opengl/Image.cpp"
           $$0$i = $83;$$01$i = $82;
           break L23;
          }
          else if ((label|0) == 20) {
           $91 = (___cxa_allocate_exception(16)|0); //@line 84 "modules/graphics/opengl/Image.cpp"
           $92 = HEAP32[$compresseddata>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $93 = (($92) + ($i$010$i$lcssa114<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $94 = HEAP32[$93>>2]|0; //@line 84 "modules/graphics/opengl/Image.cpp"
           __THREW__ = 0;
           $95 = (invoke_iii(3636,($94|0),0)|0); //@line 84 "modules/graphics/opengl/Image.cpp"
           $96 = __THREW__; __THREW__ = 0;
           $97 = $96&1;
           if (!($97)) {
            $98 = (($i$010$i$lcssa114) + 1)|0; //@line 84 "modules/graphics/opengl/Image.cpp"
            __THREW__ = 0;
            HEAP32[$vararg_buffer7>>2] = $98; //@line 84 "modules/graphics/opengl/Image.cpp"
            $vararg_ptr10 = ((($vararg_buffer7)) + 4|0); //@line 84 "modules/graphics/opengl/Image.cpp"
            HEAP32[$vararg_ptr10>>2] = $$lcssa120; //@line 84 "modules/graphics/opengl/Image.cpp"
            $vararg_ptr11 = ((($vararg_buffer7)) + 8|0); //@line 84 "modules/graphics/opengl/Image.cpp"
            HEAP32[$vararg_ptr11>>2] = $95; //@line 84 "modules/graphics/opengl/Image.cpp"
            invoke_viii(3349,($91|0),(436476|0),($vararg_buffer7|0)); //@line 84 "modules/graphics/opengl/Image.cpp"
            $99 = __THREW__; __THREW__ = 0;
            $100 = $99&1;
            if (!($100)) {
             __THREW__ = 0;
             invoke_viii(3350,($91|0),(96|0),(344|0)); //@line 84 "modules/graphics/opengl/Image.cpp"
             $101 = __THREW__; __THREW__ = 0;
             label = 30;
             break L5;
            }
           }
           $102 = ___cxa_find_matching_catch_2()|0;
           $103 = tempRet0;
           ___cxa_free_exception(($91|0)); //@line 84 "modules/graphics/opengl/Image.cpp"
           $$0$i = $103;$$01$i = $102;
           break L23;
          }
          else if ((label|0) == 29) {
           $110 = ___cxa_find_matching_catch_2()|0;
           $111 = tempRet0;
           $eh$lpad$body$index19Z2D = $111;$eh$lpad$body$indexZ2D = $110;
           label = 31;
           break L5;
          }
         }
        } while(0);
        HEAP8[$11>>0] = 1; //@line 130 "modules/graphics/opengl/Image.cpp"
        break L8;
       } else {
        $52 = (___cxa_allocate_exception(16)|0); //@line 73 "modules/graphics/opengl/Image.cpp"
        __THREW__ = 0;
        HEAP32[$vararg_buffer>>2] = $49; //@line 73 "modules/graphics/opengl/Image.cpp"
        $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 73 "modules/graphics/opengl/Image.cpp"
        HEAP32[$vararg_ptr1>>2] = $32; //@line 73 "modules/graphics/opengl/Image.cpp"
        invoke_viii(3349,($52|0),(436341|0),($vararg_buffer|0)); //@line 73 "modules/graphics/opengl/Image.cpp"
        $53 = __THREW__; __THREW__ = 0;
        $54 = $53&1;
        if ($54) {
         $56 = ___cxa_find_matching_catch_2()|0;
         $57 = tempRet0;
         ___cxa_free_exception(($52|0)); //@line 73 "modules/graphics/opengl/Image.cpp"
         $$0$i = $57;$$01$i = $56;
         break;
        } else {
         __THREW__ = 0;
         invoke_viii(3350,($52|0),(96|0),(344|0)); //@line 73 "modules/graphics/opengl/Image.cpp"
         $55 = __THREW__; __THREW__ = 0;
         label = 30;
         break L5;
        }
       }
      } while(0);
      $eh$lpad$body$index19Z2D = $$0$i;$eh$lpad$body$indexZ2D = $$01$i;
      label = 31;
      break L5;
     }
    } while(0);
    $140 = HEAP32[$compresseddata>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $141 = HEAP32[$28>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $142 = ($140|0)==($141|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    L51: do {
     if (!($142)) {
      $143 = ((($this)) + 144|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $__begin$sroa$0$036 = $140;
      while(1) {
       $185 = HEAP32[$__begin$sroa$0$036>>2]|0; //@line 106 "./common/Object.h"
       HEAP32[$0>>2] = $185; //@line 93 "./common/Object.h"
       $186 = ($185|0)==(0|0); //@line 95 "./common/Object.h"
       if (!($186)) {
        $187 = HEAP32[$185>>2]|0; //@line 95 "./common/Object.h"
        $188 = ((($187)) + 8|0); //@line 95 "./common/Object.h"
        $189 = HEAP32[$188>>2]|0; //@line 95 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($189|0,($185|0)); //@line 95 "./common/Object.h"
        $190 = __THREW__; __THREW__ = 0;
        $191 = $190&1;
        if ($191) {
         label = 28;
         break;
        }
       }
       $192 = HEAP32[$7>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $193 = HEAP32[$143>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $194 = ($192>>>0)<($193>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($194) {
        HEAP32[$192>>2] = $185; //@line 99 "./common/Object.h"
        if ($186) {
         $203 = $192;
        } else {
         $195 = HEAP32[$185>>2]|0; //@line 101 "./common/Object.h"
         $196 = ((($195)) + 8|0); //@line 101 "./common/Object.h"
         $197 = HEAP32[$196>>2]|0; //@line 101 "./common/Object.h"
         __THREW__ = 0;
         invoke_vi($197|0,($185|0)); //@line 101 "./common/Object.h"
         $198 = __THREW__; __THREW__ = 0;
         $199 = $198&1;
         if ($199) {
          $$lcssa104 = $185;
          label = 58;
          break;
         }
         $$pre$i16 = HEAP32[$7>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $203 = $$pre$i16;
        }
        $202 = ((($203)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$7>>2] = $202; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       } else {
        __THREW__ = 0;
        invoke_vii(4084,($6|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $204 = __THREW__; __THREW__ = 0;
        $205 = $204&1;
        if ($205) {
         label = 70;
         break;
        }
       }
       if (!($186)) {
        $206 = HEAP32[$185>>2]|0; //@line 106 "./common/Object.h"
        $207 = ((($206)) + 12|0); //@line 106 "./common/Object.h"
        $208 = HEAP32[$207>>2]|0; //@line 106 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($208|0,($185|0)); //@line 106 "./common/Object.h"
        $209 = __THREW__; __THREW__ = 0;
        $210 = $209&1;
        if ($210) {
         label = 64;
         break;
        }
       }
       $213 = HEAP32[$__begin$sroa$0$036>>2]|0; //@line 142 "modules/graphics/opengl/Image.cpp"
       __THREW__ = 0;
       $214 = (invoke_ii(4085,($213|0))|0); //@line 142 "modules/graphics/opengl/Image.cpp"
       $215 = __THREW__; __THREW__ = 0;
       $216 = $215&1;
       if ($216) {
        label = 28;
        break;
       }
       $217 = HEAP32[$6>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $218 = HEAP32[$217>>2]|0; //@line 117 "./common/Object.h"
       __THREW__ = 0;
       $219 = (invoke_ii(4085,($218|0))|0); //@line 142 "modules/graphics/opengl/Image.cpp"
       $220 = __THREW__; __THREW__ = 0;
       $221 = $220&1;
       if ($221) {
        label = 28;
        break;
       }
       $222 = ($214|0)==($219|0); //@line 142 "modules/graphics/opengl/Image.cpp"
       $145 = ((($__begin$sroa$0$036)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       if (!($222)) {
        label = 68;
        break;
       }
       $144 = ($145|0)==($141|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       if ($144) {
        break L51;
       } else {
        $__begin$sroa$0$036 = $145;
       }
      }
      if ((label|0) == 28) {
       $108 = ___cxa_find_matching_catch_2()|0;
       $109 = tempRet0;
       $eh$lpad$body$index19Z2D = $109;$eh$lpad$body$indexZ2D = $108;
       label = 31;
       break L5;
      }
      else if ((label|0) == 58) {
       $200 = ___cxa_find_matching_catch_2()|0;
       $201 = tempRet0;
       $230 = $$lcssa104;$282 = $201;$283 = $200;
      }
      else if ((label|0) == 64) {
       $211 = ___cxa_find_matching_catch_3(0|0)|0;
       $212 = tempRet0;
       ___clang_call_terminate($211); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
      else if ((label|0) == 68) {
       $223 = (___cxa_allocate_exception(16)|0); //@line 143 "modules/graphics/opengl/Image.cpp"
       __THREW__ = 0;
       invoke_viii(3349,($223|0),(436634|0),($vararg_buffer14|0)); //@line 143 "modules/graphics/opengl/Image.cpp"
       $224 = __THREW__; __THREW__ = 0;
       $225 = $224&1;
       if ($225) {
        $237 = ___cxa_find_matching_catch_2()|0;
        $238 = tempRet0;
        ___cxa_free_exception(($223|0)); //@line 143 "modules/graphics/opengl/Image.cpp"
        $$0 = $237;$$01 = $238;
        break L5;
       } else {
        __THREW__ = 0;
        invoke_viii(3350,($223|0),(96|0),(344|0)); //@line 143 "modules/graphics/opengl/Image.cpp"
        $226 = __THREW__; __THREW__ = 0;
        label = 30;
        break L5;
       }
      }
      else if ((label|0) == 70) {
       $227 = ___cxa_find_matching_catch_2()|0;
       $228 = tempRet0;
       if ($186) {
        $$0 = $227;$$01 = $228;
        break L5;
       } else {
        $230 = $185;$282 = $228;$283 = $227;
       }
      }
      $229 = HEAP32[$230>>2]|0; //@line 106 "./common/Object.h"
      $231 = ((($229)) + 12|0); //@line 106 "./common/Object.h"
      $232 = HEAP32[$231>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($232|0,($230|0)); //@line 106 "./common/Object.h"
      $233 = __THREW__; __THREW__ = 0;
      $234 = $233&1;
      if (!($234)) {
       $$0 = $283;$$01 = $282;
       break L5;
      }
      $235 = ___cxa_find_matching_catch_3(0|0)|0;
      $236 = tempRet0;
      ___clang_call_terminate($235); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
    } while(0);
    $146 = ((($this)) + 56|0); //@line 161 "modules/graphics/opengl/Image.cpp"
    $147 = ((($this)) + 76|0); //@line 161 "modules/graphics/opengl/Image.cpp"
    $148 = ((($this)) + 96|0); //@line 161 "modules/graphics/opengl/Image.cpp"
    $149 = ((($this)) + 116|0); //@line 161 "modules/graphics/opengl/Image.cpp"
    $150 = ((($this)) + 40|0); //@line 168 "modules/graphics/opengl/Image.cpp"
    HEAP32[$146>>2] = -1; //@line 161 "modules/graphics/opengl/Image.cpp"
    HEAP32[$147>>2] = -1; //@line 161 "modules/graphics/opengl/Image.cpp"
    HEAP32[$148>>2] = -1; //@line 161 "modules/graphics/opengl/Image.cpp"
    HEAP32[$149>>2] = -1; //@line 161 "modules/graphics/opengl/Image.cpp"
    HEAPF32[$150>>2] = 0.0; //@line 168 "modules/graphics/opengl/Image.cpp"
    $151 = ((($this)) + 44|0); //@line 169 "modules/graphics/opengl/Image.cpp"
    HEAPF32[$151>>2] = 0.0; //@line 169 "modules/graphics/opengl/Image.cpp"
    $152 = ((($this)) + 60|0); //@line 170 "modules/graphics/opengl/Image.cpp"
    HEAPF32[$152>>2] = 0.0; //@line 170 "modules/graphics/opengl/Image.cpp"
    $153 = HEAP32[$27>>2]|0; //@line 171 "modules/graphics/opengl/Image.cpp"
    $154 = (+($153|0)); //@line 171 "modules/graphics/opengl/Image.cpp"
    $155 = ((($this)) + 64|0); //@line 171 "modules/graphics/opengl/Image.cpp"
    HEAPF32[$155>>2] = $154; //@line 171 "modules/graphics/opengl/Image.cpp"
    $156 = HEAP32[$21>>2]|0; //@line 172 "modules/graphics/opengl/Image.cpp"
    $157 = (+($156|0)); //@line 172 "modules/graphics/opengl/Image.cpp"
    $158 = ((($this)) + 80|0); //@line 172 "modules/graphics/opengl/Image.cpp"
    HEAPF32[$158>>2] = $157; //@line 172 "modules/graphics/opengl/Image.cpp"
    $159 = ((($this)) + 84|0); //@line 173 "modules/graphics/opengl/Image.cpp"
    HEAPF32[$159>>2] = 0.0; //@line 173 "modules/graphics/opengl/Image.cpp"
    $160 = ((($this)) + 100|0); //@line 174 "modules/graphics/opengl/Image.cpp"
    HEAPF32[$160>>2] = $157; //@line 174 "modules/graphics/opengl/Image.cpp"
    $161 = ((($this)) + 104|0); //@line 175 "modules/graphics/opengl/Image.cpp"
    HEAPF32[$161>>2] = $154; //@line 175 "modules/graphics/opengl/Image.cpp"
    $162 = ((($this)) + 48|0); //@line 177 "modules/graphics/opengl/Image.cpp"
    HEAPF32[$162>>2] = 0.0; //@line 177 "modules/graphics/opengl/Image.cpp"
    $163 = ((($this)) + 52|0); //@line 178 "modules/graphics/opengl/Image.cpp"
    HEAPF32[$163>>2] = 0.0; //@line 178 "modules/graphics/opengl/Image.cpp"
    $164 = ((($this)) + 68|0); //@line 179 "modules/graphics/opengl/Image.cpp"
    HEAPF32[$164>>2] = 0.0; //@line 179 "modules/graphics/opengl/Image.cpp"
    $165 = ((($this)) + 72|0); //@line 180 "modules/graphics/opengl/Image.cpp"
    HEAPF32[$165>>2] = 1.0; //@line 180 "modules/graphics/opengl/Image.cpp"
    $166 = ((($this)) + 88|0); //@line 181 "modules/graphics/opengl/Image.cpp"
    HEAPF32[$166>>2] = 1.0; //@line 181 "modules/graphics/opengl/Image.cpp"
    $167 = ((($this)) + 92|0); //@line 182 "modules/graphics/opengl/Image.cpp"
    HEAPF32[$167>>2] = 0.0; //@line 182 "modules/graphics/opengl/Image.cpp"
    $168 = ((($this)) + 108|0); //@line 183 "modules/graphics/opengl/Image.cpp"
    HEAPF32[$168>>2] = 1.0; //@line 183 "modules/graphics/opengl/Image.cpp"
    $169 = ((($this)) + 112|0); //@line 184 "modules/graphics/opengl/Image.cpp"
    HEAPF32[$169>>2] = 1.0; //@line 184 "modules/graphics/opengl/Image.cpp"
    $170 = ((($this)) + 157|0); //@line 186 "modules/graphics/opengl/Image.cpp"
    $171 = HEAP8[$170>>0]|0; //@line 186 "modules/graphics/opengl/Image.cpp"
    $172 = ($171<<24>>24)==(0); //@line 186 "modules/graphics/opengl/Image.cpp"
    if (!($172)) {
     $173 = HEAP32[6202]|0; //@line 187 "modules/graphics/opengl/Image.cpp"
     $174 = ((($this)) + 24|0); //@line 187 "modules/graphics/opengl/Image.cpp"
     HEAP32[$174>>2] = $173; //@line 187 "modules/graphics/opengl/Image.cpp"
    }
    __THREW__ = 0;
    $175 = (invoke_i(3801)|0); //@line 189 "modules/graphics/opengl/Image.cpp"
    $176 = __THREW__; __THREW__ = 0;
    $177 = $176&1;
    if ($177) {
     label = 30;
    } else {
     if (!($175)) {
      $178 = ((($this)) + 158|0); //@line 190 "modules/graphics/opengl/Image.cpp"
      HEAP8[$178>>0] = 0; //@line 190 "modules/graphics/opengl/Image.cpp"
     }
     __THREW__ = 0;
     $179 = (invoke_i(3801)|0); //@line 192 "modules/graphics/opengl/Image.cpp"
     $180 = __THREW__; __THREW__ = 0;
     $181 = $180&1;
     if ($181) {
      label = 30;
     } else {
      if ($179) {
       $182 = ((($this)) + 158|0); //@line 192 "modules/graphics/opengl/Image.cpp"
       $183 = HEAP8[$182>>0]|0; //@line 192 "modules/graphics/opengl/Image.cpp"
       $184 = ($183<<24>>24)==(0); //@line 192 "modules/graphics/opengl/Image.cpp"
       if ($184) {
        $storemerge = 1;
       } else {
        label = 52;
       }
      } else {
       label = 52;
      }
      if ((label|0) == 52) {
       $storemerge = 0;
      }
      HEAP8[$13>>0] = $storemerge; //@line 193 "modules/graphics/opengl/Image.cpp"
      $239 = HEAP32[$this>>2]|0; //@line 147 "modules/graphics/opengl/Image.cpp"
      $240 = ((($239)) + 56|0); //@line 147 "modules/graphics/opengl/Image.cpp"
      $241 = HEAP32[$240>>2]|0; //@line 147 "modules/graphics/opengl/Image.cpp"
      __THREW__ = 0;
      (invoke_ii($241|0,($this|0))|0); //@line 147 "modules/graphics/opengl/Image.cpp"
      $242 = __THREW__; __THREW__ = 0;
      $243 = $242&1;
      if ($243) {
       label = 30;
      } else {
       $244 = HEAP32[158431]|0; //@line 149 "modules/graphics/opengl/Image.cpp"
       $245 = (($244) + 1)|0; //@line 149 "modules/graphics/opengl/Image.cpp"
       HEAP32[158431] = $245; //@line 149 "modules/graphics/opengl/Image.cpp"
       STACKTOP = sp;return; //@line 150 "modules/graphics/opengl/Image.cpp"
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 30) {
  $112 = ___cxa_find_matching_catch_2()|0;
  $113 = tempRet0;
  $eh$lpad$body$index19Z2D = $113;$eh$lpad$body$indexZ2D = $112;
  label = 31;
 }
 if ((label|0) == 31) {
  $$0 = $eh$lpad$body$indexZ2D;$$01 = $eh$lpad$body$index19Z2D;
 }
 $246 = HEAP32[$6>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $247 = ($246|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($247)) {
  $248 = HEAP32[$7>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $249 = ($248|0)==($246|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  do {
   if ($249) {
    $263 = $246;
   } else {
    $251 = $248;
    while(1) {
     $250 = ((($251)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$7>>2] = $250; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $252 = HEAP32[$250>>2]|0; //@line 106 "./common/Object.h"
     $253 = ($252|0)==(0|0); //@line 106 "./common/Object.h"
     if ($253) {
      $255 = $250;
     } else {
      $256 = HEAP32[$252>>2]|0; //@line 106 "./common/Object.h"
      $257 = ((($256)) + 12|0); //@line 106 "./common/Object.h"
      $258 = HEAP32[$257>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($258|0,($252|0)); //@line 106 "./common/Object.h"
      $259 = __THREW__; __THREW__ = 0;
      $260 = $259&1;
      if ($260) {
       label = 82;
       break;
      }
      $$pre$i$i$i4 = HEAP32[$7>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $255 = $$pre$i$i$i4;
     }
     $254 = ($255|0)==($246|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($254) {
      label = 83;
      break;
     } else {
      $251 = $255;
     }
    }
    if ((label|0) == 82) {
     $261 = ___cxa_find_matching_catch_3(0|0)|0;
     $262 = tempRet0;
     ___clang_call_terminate($261); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
    else if ((label|0) == 83) {
     $$pre$i6 = HEAP32[$6>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $263 = $$pre$i6;
     break;
    }
   }
  } while(0);
  __ZdlPv($263); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $264 = HEAP32[$4>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $265 = ($264|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($265)) {
  $266 = HEAP32[$5>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $267 = ($266|0)==($264|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  do {
   if ($267) {
    $281 = $264;
   } else {
    $269 = $266;
    while(1) {
     $268 = ((($269)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$5>>2] = $268; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $270 = HEAP32[$268>>2]|0; //@line 106 "./common/Object.h"
     $271 = ($270|0)==(0|0); //@line 106 "./common/Object.h"
     if ($271) {
      $273 = $268;
     } else {
      $274 = HEAP32[$270>>2]|0; //@line 106 "./common/Object.h"
      $275 = ((($274)) + 12|0); //@line 106 "./common/Object.h"
      $276 = HEAP32[$275>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($276|0,($270|0)); //@line 106 "./common/Object.h"
      $277 = __THREW__; __THREW__ = 0;
      $278 = $277&1;
      if ($278) {
       label = 91;
       break;
      }
      $$pre$i$i$i = HEAP32[$5>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $273 = $$pre$i$i$i;
     }
     $272 = ($273|0)==($264|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($272) {
      label = 92;
      break;
     } else {
      $269 = $273;
     }
    }
    if ((label|0) == 91) {
     $279 = ___cxa_find_matching_catch_3(0|0)|0;
     $280 = tempRet0;
     ___clang_call_terminate($279); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
    else if ((label|0) == 92) {
     $$pre$i = HEAP32[$4>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $281 = $$pre$i;
     break;
    }
   }
  } while(0);
  __ZdlPv($281); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 __ZN4love8graphics8VolatileD2Ev($1); //@line 150 "modules/graphics/opengl/Image.cpp"
 $$1 = $$0;$$12 = $$01;
 __ZN4love8graphics7TextureD2Ev($this); //@line 150 "modules/graphics/opengl/Image.cpp"
 ___resumeException($$1|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love9StrongRefINS1_5image19CompressedImageDataEEENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i8 = 0, $$0$i9 = 0, $$01$i$i = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre12 = 0, $$pre13 = 0, $$pre14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__v$sroa$15$0 = 0, $__v$sroa$6$0 = 0, $__v$sroa$6$0$lcssa = 0, $__v$sroa$6$1 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = $$in; //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = HEAP32[$8>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $10 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($11>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $17 = $10 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ($17>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $19 = $18 ? $5 : $17; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = (($20) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = $21 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = ($19|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $24 = $20; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($23) {
   $$0$i9 = 0;$101 = $24;$28 = 0;$29 = $22;
  } else {
   $$0$i8 = $19;$100 = $22;$99 = $24;
   label = 6;
  }
 } else {
  $13 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = $14 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = $13; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $$0$i8 = 1073741823;$100 = $15;$99 = $16;
  label = 6;
 }
 if ((label|0) == 6) {
  $25 = $$0$i8 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (__Znwj($25)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i9 = $$0$i8;$101 = $99;$28 = $26;$29 = $100;
 }
 $27 = (($28) + ($29<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = $27; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $31 = (($28) + ($$0$i9<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $32 = $31; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $33 = HEAP32[$__x>>2]|0; //@line 139 "./common/Object.h"
 HEAP32[$27>>2] = $33; //@line 99 "./common/Object.h"
 $34 = ($33|0)==(0|0); //@line 101 "./common/Object.h"
 do {
  if ($34) {
   $43 = $101;$44 = $7;
   label = 10;
  } else {
   $35 = HEAP32[$33>>2]|0; //@line 101 "./common/Object.h"
   $36 = ((($35)) + 8|0); //@line 101 "./common/Object.h"
   $37 = HEAP32[$36>>2]|0; //@line 101 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($37|0,($33|0)); //@line 101 "./common/Object.h"
   $38 = __THREW__; __THREW__ = 0;
   $39 = $38&1;
   if ($39) {
    $81 = ___cxa_find_matching_catch_2()|0;
    $82 = tempRet0;
    $__v$sroa$15$0 = $30;$__v$sroa$6$1 = $30;$lpad$phi$index = $81;$lpad$phi$index2 = $82;
    break;
   } else {
    $$pre12 = HEAP32[$this>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre13 = HEAP32[$0>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $43 = $$pre13;$44 = $$pre12;
    label = 10;
    break;
   }
  }
 } while(0);
 L15: do {
  if ((label|0) == 10) {
   $40 = ((($27)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $41 = $40; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $42 = ($43|0)==($44|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   do {
    if ($42) {
     $45 = $43; //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $60 = $30;$62 = $45;$64 = $43;
    } else {
     $$01$i$i = $43;$48 = $27;$__v$sroa$6$0 = $30;
     while(1) {
      $46 = ((($$01$i$i)) + -4|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $47 = ((($48)) + -4|0); //@line 99 "./common/Object.h"
      $49 = HEAP32[$46>>2]|0; //@line 139 "./common/Object.h"
      HEAP32[$47>>2] = $49; //@line 99 "./common/Object.h"
      $50 = ($49|0)==(0|0); //@line 101 "./common/Object.h"
      if (!($50)) {
       $51 = HEAP32[$49>>2]|0; //@line 101 "./common/Object.h"
       $52 = ((($51)) + 8|0); //@line 101 "./common/Object.h"
       $53 = HEAP32[$52>>2]|0; //@line 101 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($53|0,($49|0)); //@line 101 "./common/Object.h"
       $54 = __THREW__; __THREW__ = 0;
       $55 = $54&1;
       if ($55) {
        $__v$sroa$6$0$lcssa = $__v$sroa$6$0;
        label = 24;
        break;
       }
      }
      $56 = $__v$sroa$6$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $57 = ((($56)) + -4|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $58 = $57; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $59 = ($46|0)==($44|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      if ($59) {
       $$lcssa = $58;
       label = 15;
       break;
      } else {
       $$01$i$i = $46;$48 = $57;$__v$sroa$6$0 = $58;
      }
     }
     if ((label|0) == 15) {
      $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $$pre14 = HEAP32[$0>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $60 = $$lcssa;$62 = $$pre$i;$64 = $$pre14;
      break;
     }
     else if ((label|0) == 24) {
      $79 = ___cxa_find_matching_catch_2()|0;
      $80 = tempRet0;
      $__v$sroa$15$0 = $41;$__v$sroa$6$1 = $__v$sroa$6$0$lcssa;$lpad$phi$index = $79;$lpad$phi$index2 = $80;
      break L15;
     }
    }
   } while(0);
   HEAP32[$this>>2] = $60; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$0>>2] = $41; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$8>>2] = $32; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $61 = $62; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $63 = ($64|0)==($61|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   L28: do {
    if (!($63)) {
     $66 = $64;
     while(1) {
      $65 = ((($66)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $67 = HEAP32[$65>>2]|0; //@line 106 "./common/Object.h"
      $68 = ($67|0)==(0|0); //@line 106 "./common/Object.h"
      if (!($68)) {
       $70 = HEAP32[$67>>2]|0; //@line 106 "./common/Object.h"
       $71 = ((($70)) + 12|0); //@line 106 "./common/Object.h"
       $72 = HEAP32[$71>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($72|0,($67|0)); //@line 106 "./common/Object.h"
       $73 = __THREW__; __THREW__ = 0;
       $74 = $73&1;
       if ($74) {
        break;
       }
      }
      $69 = ($65|0)==($61|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      if ($69) {
       break L28;
      } else {
       $66 = $65;
      }
     }
     $75 = ___cxa_find_matching_catch_3(0|0)|0;
     $76 = tempRet0;
     ___clang_call_terminate($75); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   } while(0);
   $77 = ($62|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($77) {
    return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $78 = $62; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __ZdlPv($78); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
 } while(0);
 $83 = $__v$sroa$6$1; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $84 = $__v$sroa$15$0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $85 = ($84|0)==($83|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 L41: do {
  if (!($85)) {
   $87 = $84;
   while(1) {
    $86 = ((($87)) + -4|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $88 = HEAP32[$86>>2]|0; //@line 106 "./common/Object.h"
    $89 = ($88|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($89)) {
     $91 = HEAP32[$88>>2]|0; //@line 106 "./common/Object.h"
     $92 = ((($91)) + 12|0); //@line 106 "./common/Object.h"
     $93 = HEAP32[$92>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($93|0,($88|0)); //@line 106 "./common/Object.h"
     $94 = __THREW__; __THREW__ = 0;
     $95 = $94&1;
     if ($95) {
      break;
     }
    }
    $90 = ($86|0)==($83|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($90) {
     break L41;
    } else {
     $87 = $86;
    }
   }
   $96 = ___cxa_find_matching_catch_3(0|0)|0;
   $97 = tempRet0;
   ___clang_call_terminate($96); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 } while(0);
 $98 = ($28|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($98) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 __ZdlPv($28); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZNK4love8graphics6opengl5Image12isCompressedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 156|0); //@line 592 "modules/graphics/opengl/Image.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 592 "modules/graphics/opengl/Image.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 592 "modules/graphics/opengl/Image.cpp"
 return ($2|0); //@line 592 "modules/graphics/opengl/Image.cpp"
}
function __ZN4love8graphics6opengl5Image27hasCompressedTextureSupportENS_5image19CompressedImageData6FormatEb($format,$sRGB) {
 $format = $format|0;
 $sRGB = $sRGB|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 do {
  switch ($format|0) {
  case 1:  {
   $0 = HEAP8[1252236]|0; //@line 707 "modules/graphics/opengl/Image.cpp"
   $1 = HEAP8[1252233]|0; //@line 707 "modules/graphics/opengl/Image.cpp"
   $2 = $1 | $0; //@line 707 "modules/graphics/opengl/Image.cpp"
   $3 = ($2<<24>>24)!=(0); //@line 707 "modules/graphics/opengl/Image.cpp"
   $$0 = $3;
   return ($$0|0); //@line 755 "modules/graphics/opengl/Image.cpp"
   break;
  }
  case 2:  {
   $4 = HEAP8[1252236]|0; //@line 709 "modules/graphics/opengl/Image.cpp"
   $5 = HEAP8[1252349]|0; //@line 709 "modules/graphics/opengl/Image.cpp"
   $6 = $5 | $4; //@line 709 "modules/graphics/opengl/Image.cpp"
   $7 = ($6<<24>>24)!=(0); //@line 709 "modules/graphics/opengl/Image.cpp"
   $$0 = $7;
   return ($$0|0); //@line 755 "modules/graphics/opengl/Image.cpp"
   break;
  }
  case 3:  {
   $8 = HEAP8[1252236]|0; //@line 711 "modules/graphics/opengl/Image.cpp"
   $9 = HEAP8[1252350]|0; //@line 711 "modules/graphics/opengl/Image.cpp"
   $10 = $9 | $8; //@line 711 "modules/graphics/opengl/Image.cpp"
   $11 = ($10<<24>>24)!=(0); //@line 711 "modules/graphics/opengl/Image.cpp"
   $$0 = $11;
   return ($$0|0); //@line 755 "modules/graphics/opengl/Image.cpp"
   break;
  }
  case 7: case 6: case 5: case 4:  {
   $12 = HEAP8[1252002]|0; //@line 716 "modules/graphics/opengl/Image.cpp"
   $13 = HEAP8[1252132]|0; //@line 716 "modules/graphics/opengl/Image.cpp"
   $14 = $13 | $12; //@line 716 "modules/graphics/opengl/Image.cpp"
   $15 = HEAP8[1252235]|0; //@line 716 "modules/graphics/opengl/Image.cpp"
   $16 = $14 | $15; //@line 716 "modules/graphics/opengl/Image.cpp"
   $17 = ($16<<24>>24)!=(0); //@line 716 "modules/graphics/opengl/Image.cpp"
   $$0 = $17;
   return ($$0|0); //@line 755 "modules/graphics/opengl/Image.cpp"
   break;
  }
  case 10: case 9: case 8:  {
   $18 = HEAP8[1252008]|0; //@line 720 "modules/graphics/opengl/Image.cpp"
   $19 = HEAP8[1252131]|0; //@line 720 "modules/graphics/opengl/Image.cpp"
   $20 = $19 | $18; //@line 720 "modules/graphics/opengl/Image.cpp"
   $21 = ($20<<24>>24)!=(0); //@line 720 "modules/graphics/opengl/Image.cpp"
   $$0 = $21;
   return ($$0|0); //@line 755 "modules/graphics/opengl/Image.cpp"
   break;
  }
  case 14: case 13: case 12: case 11:  {
   $22 = HEAP8[1252207]|0; //@line 725 "modules/graphics/opengl/Image.cpp"
   $23 = HEAP8[1252391]|0; //@line 725 "modules/graphics/opengl/Image.cpp"
   $24 = $sRGB ? $22 : $23; //@line 725 "modules/graphics/opengl/Image.cpp"
   $25 = ($24<<24>>24)!=(0); //@line 725 "modules/graphics/opengl/Image.cpp"
   $$0 = $25;
   return ($$0|0); //@line 755 "modules/graphics/opengl/Image.cpp"
   break;
  }
  case 15:  {
   $26 = HEAP8[1252013]|0; //@line 728 "modules/graphics/opengl/Image.cpp"
   $27 = HEAP8[1252009]|0; //@line 728 "modules/graphics/opengl/Image.cpp"
   $28 = $27 | $26; //@line 728 "modules/graphics/opengl/Image.cpp"
   $29 = HEAP8[1252058]|0; //@line 728 "modules/graphics/opengl/Image.cpp"
   $30 = $28 | $29; //@line 728 "modules/graphics/opengl/Image.cpp"
   $31 = HEAP8[1252263]|0; //@line 728 "modules/graphics/opengl/Image.cpp"
   $32 = $30 | $31; //@line 728 "modules/graphics/opengl/Image.cpp"
   $33 = ($32<<24>>24)!=(0); //@line 728 "modules/graphics/opengl/Image.cpp"
   $$0 = $33;
   return ($$0|0); //@line 755 "modules/graphics/opengl/Image.cpp"
   break;
  }
  case 22: case 21: case 20: case 19: case 18: case 17: case 16:  {
   $34 = HEAP8[1252013]|0; //@line 736 "modules/graphics/opengl/Image.cpp"
   $35 = HEAP8[1252009]|0; //@line 736 "modules/graphics/opengl/Image.cpp"
   $36 = $35 | $34; //@line 736 "modules/graphics/opengl/Image.cpp"
   $37 = HEAP8[1252058]|0; //@line 736 "modules/graphics/opengl/Image.cpp"
   $38 = $36 | $37; //@line 736 "modules/graphics/opengl/Image.cpp"
   $39 = ($38<<24>>24)!=(0); //@line 736 "modules/graphics/opengl/Image.cpp"
   $$0 = $39;
   return ($$0|0); //@line 755 "modules/graphics/opengl/Image.cpp"
   break;
  }
  case 36: case 35: case 34: case 33: case 32: case 31: case 30: case 29: case 28: case 27: case 26: case 25: case 24: case 23:  {
   $40 = HEAP8[1252015]|0; //@line 751 "modules/graphics/opengl/Image.cpp"
   $41 = HEAP8[1252024]|0; //@line 751 "modules/graphics/opengl/Image.cpp"
   $42 = $41 | $40; //@line 751 "modules/graphics/opengl/Image.cpp"
   $43 = ($42<<24>>24)!=(0); //@line 751 "modules/graphics/opengl/Image.cpp"
   $$0 = $43;
   return ($$0|0); //@line 755 "modules/graphics/opengl/Image.cpp"
   break;
  }
  default: {
   $$0 = 0;
   return ($$0|0); //@line 755 "modules/graphics/opengl/Image.cpp"
  }
  }
 } while(0);
 return (0)|0;
}
function __ZN4love8graphics6opengl5Image18setMipmapSharpnessEf($this,$sharpness) {
 $this = $this|0;
 $sharpness = +$sharpness;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0.0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 8|0;
 $1 = sp + 4|0;
 $2 = sp;
 HEAPF32[$0>>2] = $sharpness;
 $3 = HEAP8[1251998]|0; //@line 548 "modules/graphics/opengl/Image.cpp"
 $4 = ($3<<24>>24)==(0); //@line 548 "modules/graphics/opengl/Image.cpp"
 if ($4) {
  STACKTOP = sp;return; //@line 558 "modules/graphics/opengl/Image.cpp"
 }
 $5 = +HEAPF32[158432]; //@line 552 "modules/graphics/opengl/Image.cpp"
 $6 = 0.0099999997764825821 - $5; //@line 552 "modules/graphics/opengl/Image.cpp"
 HEAPF32[$1>>2] = $6; //@line 552 "modules/graphics/opengl/Image.cpp"
 $7 = $6 > $sharpness; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $8 = $7 ? $1 : $0; //@line 2648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $9 = $5 + -0.0099999997764825821; //@line 552 "modules/graphics/opengl/Image.cpp"
 HEAPF32[$2>>2] = $9; //@line 552 "modules/graphics/opengl/Image.cpp"
 $10 = +HEAPF32[$8>>2]; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $11 = $9 < $10; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $12 = $11 ? $2 : $8; //@line 2575 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $13 = HEAP32[$12>>2]|0; //@line 552 "modules/graphics/opengl/Image.cpp"
 $14 = ((($this)) + 152|0); //@line 552 "modules/graphics/opengl/Image.cpp"
 HEAP32[$14>>2] = $13; //@line 552 "modules/graphics/opengl/Image.cpp"
 $15 = ((($this)) + 148|0); //@line 554 "modules/graphics/opengl/Image.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 554 "modules/graphics/opengl/Image.cpp"
 __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$16); //@line 554 "modules/graphics/opengl/Image.cpp"
 $17 = +HEAPF32[$14>>2]; //@line 557 "modules/graphics/opengl/Image.cpp"
 $18 = -$17; //@line 557 "modules/graphics/opengl/Image.cpp"
 $19 = HEAP32[165106]|0; //@line 42 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viid[$19 & 8191](3553,34049,$18); //@line 42 "./libraries/glad/gladfuncs.hpp"
 STACKTOP = sp;return; //@line 558 "modules/graphics/opengl/Image.cpp"
}
function __ZN4love8graphics6opengl5Image7refreshEiiii($this,$xoffset,$yoffset,$w,$h) {
 $this = $this|0;
 $xoffset = $xoffset|0;
 $yoffset = $yoffset|0;
 $w = $w|0;
 $h = $h|0;
 var $$ = 0, $$01020 = 0, $$011 = 0, $$012 = 0, $$01319 = 0, $$01418 = 0, $$01517 = 0, $$1 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $i$021 = 0, $lock = 0, $notlhs = 0, $notrhs = 0, $or$cond3 = 0, $or$cond5 = 0, $or$cond7$not = 0, $or$cond9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $lock = sp + 4|0;
 $0 = ((($this)) + 148|0); //@line 396 "modules/graphics/opengl/Image.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 396 "modules/graphics/opengl/Image.cpp"
 $2 = ($1|0)==(0); //@line 396 "modules/graphics/opengl/Image.cpp"
 if ($2) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 444 "modules/graphics/opengl/Image.cpp"
 }
 $3 = ((($this)) + 160|0); //@line 396 "modules/graphics/opengl/Image.cpp"
 $4 = HEAP8[$3>>0]|0; //@line 396 "modules/graphics/opengl/Image.cpp"
 $5 = ($4<<24>>24)==(0); //@line 396 "modules/graphics/opengl/Image.cpp"
 if (!($5)) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 444 "modules/graphics/opengl/Image.cpp"
 }
 $6 = $yoffset | $xoffset; //@line 399 "modules/graphics/opengl/Image.cpp"
 $7 = ($6|0)<(0); //@line 399 "modules/graphics/opengl/Image.cpp"
 $8 = ($w|0)<(1); //@line 399 "modules/graphics/opengl/Image.cpp"
 $or$cond3 = $7 | $8; //@line 399 "modules/graphics/opengl/Image.cpp"
 $9 = ($h|0)<(1); //@line 399 "modules/graphics/opengl/Image.cpp"
 $or$cond5 = $or$cond3 | $9; //@line 399 "modules/graphics/opengl/Image.cpp"
 if (!($or$cond5)) {
  $10 = (($w) + ($xoffset))|0; //@line 400 "modules/graphics/opengl/Image.cpp"
  $11 = ((($this)) + 8|0); //@line 400 "modules/graphics/opengl/Image.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 400 "modules/graphics/opengl/Image.cpp"
  $13 = ($10|0)>($12|0); //@line 400 "modules/graphics/opengl/Image.cpp"
  if (!($13)) {
   $14 = (($h) + ($yoffset))|0; //@line 400 "modules/graphics/opengl/Image.cpp"
   $15 = ((($this)) + 12|0); //@line 400 "modules/graphics/opengl/Image.cpp"
   $16 = HEAP32[$15>>2]|0; //@line 400 "modules/graphics/opengl/Image.cpp"
   $17 = ($14|0)>($16|0); //@line 400 "modules/graphics/opengl/Image.cpp"
   if (!($17)) {
    __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$1); //@line 407 "modules/graphics/opengl/Image.cpp"
    $23 = ((($this)) + 156|0); //@line 592 "modules/graphics/opengl/Image.cpp"
    $24 = HEAP32[$23>>2]|0; //@line 592 "modules/graphics/opengl/Image.cpp"
    $25 = $24&255; //@line 592 "modules/graphics/opengl/Image.cpp"
    $26 = ($25<<24>>24)==(0); //@line 592 "modules/graphics/opengl/Image.cpp"
    if (!($26)) {
     __ZN4love8graphics6opengl5Image22loadFromCompressedDataEv($this); //@line 411 "modules/graphics/opengl/Image.cpp"
     $$1 = 1;
     STACKTOP = sp;return ($$1|0); //@line 444 "modules/graphics/opengl/Image.cpp"
    }
    $27 = HEAP16[$23>>1]|0; //@line 592 "modules/graphics/opengl/Image.cpp"
    $28 = HEAP8[1252012]|0; //@line 419 "modules/graphics/opengl/Image.cpp"
    $notlhs = ($24>>>0)<(16777216); //@line 419 "modules/graphics/opengl/Image.cpp"
    $notrhs = ($28<<24>>24)==(0); //@line 419 "modules/graphics/opengl/Image.cpp"
    $or$cond7$not = $notlhs | $notrhs; //@line 419 "modules/graphics/opengl/Image.cpp"
    $29 = HEAP8[1252013]|0; //@line 419 "modules/graphics/opengl/Image.cpp"
    $30 = ($29<<24>>24)!=(0); //@line 419 "modules/graphics/opengl/Image.cpp"
    $or$cond9 = $or$cond7$not | $30; //@line 419 "modules/graphics/opengl/Image.cpp"
    $$ = $or$cond9 ? 6408 : 35906; //@line 419 "modules/graphics/opengl/Image.cpp"
    $31 = ((($this)) + 157|0); //@line 422 "modules/graphics/opengl/Image.cpp"
    $32 = ($27&65535)<(256); //@line 422 "modules/graphics/opengl/Image.cpp"
    $$pre = ((($this)) + 124|0); //@line 427 "modules/graphics/opengl/Image.cpp"
    if ($32) {
     $77 = 1;
     label = 13;
    } else {
     $33 = ((($this)) + 128|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $34 = HEAP32[$33>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $35 = HEAP32[$$pre>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $36 = (($34) - ($35))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $37 = $36 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $38 = ($36|0)>(0); //@line 425 "modules/graphics/opengl/Image.cpp"
     if ($38) {
      $77 = $37;
      label = 13;
     }
    }
    L17: do {
     if ((label|0) == 13) {
      $39 = ((($this)) + 124|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$01020 = $xoffset;$$01319 = $h;$$01418 = $w;$$01517 = $yoffset;$i$021 = 0;
      while(1) {
       $46 = HEAP32[$39>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $47 = (($46) + ($i$021<<2)|0); //@line 117 "./common/Object.h"
       $48 = HEAP32[$47>>2]|0; //@line 117 "./common/Object.h"
       $49 = HEAP32[$48>>2]|0; //@line 427 "modules/graphics/opengl/Image.cpp"
       $50 = ((($49)) + 16|0); //@line 427 "modules/graphics/opengl/Image.cpp"
       $51 = HEAP32[$50>>2]|0; //@line 427 "modules/graphics/opengl/Image.cpp"
       $52 = (FUNCTION_TABLE_ii[$51 & 8191]($48)|0); //@line 427 "modules/graphics/opengl/Image.cpp"
       $53 = HEAP32[$39>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $54 = (($53) + ($i$021<<2)|0); //@line 117 "./common/Object.h"
       $55 = HEAP32[$54>>2]|0; //@line 117 "./common/Object.h"
       $56 = (__ZNK4love5image9ImageData8getWidthEv($55)|0); //@line 428 "modules/graphics/opengl/Image.cpp"
       $57 = Math_imul($56, $$01517)|0; //@line 428 "modules/graphics/opengl/Image.cpp"
       $58 = (($57) + ($$01020))|0; //@line 428 "modules/graphics/opengl/Image.cpp"
       $59 = HEAP32[$39>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $60 = (($59) + ($i$021<<2)|0); //@line 117 "./common/Object.h"
       $61 = HEAP32[$60>>2]|0; //@line 117 "./common/Object.h"
       $62 = (__ZNK4love5image9ImageData8getMutexEv($61)|0); //@line 430 "modules/graphics/opengl/Image.cpp"
       __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$62); //@line 430 "modules/graphics/opengl/Image.cpp"
       $63 = (($52) + ($58<<2)|0); //@line 432 "modules/graphics/opengl/Image.cpp"
       $64 = HEAP32[165192]|0; //@line 352 "./libraries/glad/gladfuncs.hpp"
       __THREW__ = 0;
       invoke_viiiiiiiii($64|0,3553,($i$021|0),($$01020|0),($$01517|0),($$01418|0),($$01319|0),($$|0),5121,($63|0)); //@line 352 "./libraries/glad/gladfuncs.hpp"
       $65 = __THREW__; __THREW__ = 0;
       $66 = $65&1;
       if ($66) {
        break;
       }
       $67 = (($$01020|0) / 2)&-1; //@line 434 "modules/graphics/opengl/Image.cpp"
       $68 = (($$01517|0) / 2)&-1; //@line 435 "modules/graphics/opengl/Image.cpp"
       $69 = (($$01418|0) / 2)&-1; //@line 436 "modules/graphics/opengl/Image.cpp"
       $70 = ($69|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $71 = $70 ? 1 : $69; //@line 436 "modules/graphics/opengl/Image.cpp"
       $72 = (($$01319|0) / 2)&-1; //@line 437 "modules/graphics/opengl/Image.cpp"
       $73 = ($72|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $74 = $73 ? 1 : $72; //@line 437 "modules/graphics/opengl/Image.cpp"
       __ZN4love6thread4LockD2Ev($lock); //@line 438 "modules/graphics/opengl/Image.cpp"
       $75 = (($i$021) + 1)|0; //@line 425 "modules/graphics/opengl/Image.cpp"
       $76 = ($75|0)<($77|0); //@line 425 "modules/graphics/opengl/Image.cpp"
       if ($76) {
        $$01020 = $67;$$01319 = $74;$$01418 = $71;$$01517 = $68;$i$021 = $75;
       } else {
        break L17;
       }
      }
      $78 = ___cxa_find_matching_catch_2()|0;
      $79 = tempRet0;
      __ZN4love6thread4LockD2Ev($lock); //@line 438 "modules/graphics/opengl/Image.cpp"
      $$011 = $79;$$012 = $78;
      ___resumeException($$012|0);
      // unreachable;
     }
    } while(0);
    $40 = ((($this)) + 128|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $41 = HEAP32[$40>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $42 = HEAP32[$$pre>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $43 = (($41) - ($42))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $44 = $43 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $45 = ($44>>>0)<(2); //@line 440 "modules/graphics/opengl/Image.cpp"
    if (!($45)) {
     $$1 = 1;
     STACKTOP = sp;return ($$1|0); //@line 444 "modules/graphics/opengl/Image.cpp"
    }
    $80 = HEAP8[$31>>0]|0; //@line 202 "modules/graphics/opengl/Image.cpp"
    $81 = ($80<<24>>24)==(0); //@line 202 "modules/graphics/opengl/Image.cpp"
    if ($81) {
     $$1 = 1;
     STACKTOP = sp;return ($$1|0); //@line 444 "modules/graphics/opengl/Image.cpp"
    }
    $82 = HEAP8[$23>>0]|0; //@line 592 "modules/graphics/opengl/Image.cpp"
    $83 = ($82<<24>>24)==(0); //@line 592 "modules/graphics/opengl/Image.cpp"
    if (!($83)) {
     $$1 = 1;
     STACKTOP = sp;return ($$1|0); //@line 444 "modules/graphics/opengl/Image.cpp"
    }
    $84 = HEAP8[1252012]|0; //@line 203 "modules/graphics/opengl/Image.cpp"
    $85 = HEAP8[1252002]|0; //@line 203 "modules/graphics/opengl/Image.cpp"
    $86 = $85 | $84; //@line 203 "modules/graphics/opengl/Image.cpp"
    $87 = HEAP8[1252065]|0; //@line 203 "modules/graphics/opengl/Image.cpp"
    $88 = $86 | $87; //@line 203 "modules/graphics/opengl/Image.cpp"
    $89 = ($88<<24>>24)==(0); //@line 203 "modules/graphics/opengl/Image.cpp"
    if ($89) {
     $$1 = 1;
     STACKTOP = sp;return ($$1|0); //@line 444 "modules/graphics/opengl/Image.cpp"
    }
    $90 = HEAP8[(634081)>>0]|0; //@line 205 "modules/graphics/opengl/Image.cpp"
    $91 = ($90<<24>>24)==(0); //@line 205 "modules/graphics/opengl/Image.cpp"
    if (!($91)) {
     $92 = HEAP32[164832]|0; //@line 57 "./libraries/glad/gladfuncs.hpp"
     FUNCTION_TABLE_vi[$92 & 8191](3553); //@line 57 "./libraries/glad/gladfuncs.hpp"
    }
    $93 = HEAP32[165249]|0; //@line 676 "./libraries/glad/gladfuncs.hpp"
    FUNCTION_TABLE_vi[$93 & 8191](3553); //@line 676 "./libraries/glad/gladfuncs.hpp"
    $$1 = 1;
    STACKTOP = sp;return ($$1|0); //@line 444 "modules/graphics/opengl/Image.cpp"
   }
  }
 }
 $18 = (___cxa_allocate_exception(16)|0); //@line 402 "modules/graphics/opengl/Image.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($18|0),(436685|0),($vararg_buffer|0)); //@line 402 "modules/graphics/opengl/Image.cpp"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if (!($20)) {
  ___cxa_throw(($18|0),(96|0),(344|0)); //@line 402 "modules/graphics/opengl/Image.cpp"
  // unreachable; //@line 402 "modules/graphics/opengl/Image.cpp"
 }
 $21 = ___cxa_find_matching_catch_2()|0;
 $22 = tempRet0;
 ___cxa_free_exception(($18|0)); //@line 402 "modules/graphics/opengl/Image.cpp"
 $$011 = $22;$$012 = $21;
 ___resumeException($$012|0);
 // unreachable;
 return (0)|0;
}
function __ZNK4love8graphics6opengl5Image12getImageDataEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 124|0); //@line 485 "modules/graphics/opengl/Image.cpp"
 return ($0|0); //@line 485 "modules/graphics/opengl/Image.cpp"
}
function __ZNK4love8graphics6opengl5Image17getCompressedDataEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 136|0); //@line 490 "modules/graphics/opengl/Image.cpp"
 return ($0|0); //@line 490 "modules/graphics/opengl/Image.cpp"
}
function __ZNK4love8graphics6opengl5Image18getMipmapSharpnessEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 152|0); //@line 562 "modules/graphics/opengl/Image.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 562 "modules/graphics/opengl/Image.cpp"
 return (+$1); //@line 562 "modules/graphics/opengl/Image.cpp"
}
function __ZNK4love8graphics6opengl5Image8getFlagsEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 157|0); //@line 567 "modules/graphics/opengl/Image.cpp"
 return ($0|0); //@line 567 "modules/graphics/opengl/Image.cpp"
}
function __ZN4love8graphics6opengl5Image25setDefaultMipmapSharpnessEf($sharpness) {
 $sharpness = +$sharpness;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAPF32[158433] = $sharpness; //@line 572 "modules/graphics/opengl/Image.cpp"
 return; //@line 573 "modules/graphics/opengl/Image.cpp"
}
function __ZN4love8graphics6opengl5Image25getDefaultMipmapSharpnessEv() {
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = +HEAPF32[158433]; //@line 577 "modules/graphics/opengl/Image.cpp"
 return (+$0); //@line 577 "modules/graphics/opengl/Image.cpp"
}
function __ZN4love8graphics6opengl5Image22setDefaultMipmapFilterENS0_7Texture10FilterModeE($f) {
 $f = $f|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[6202] = $f; //@line 582 "modules/graphics/opengl/Image.cpp"
 return; //@line 583 "modules/graphics/opengl/Image.cpp"
}
function __ZN4love8graphics6opengl5Image22getDefaultMipmapFilterEv() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[6202]|0; //@line 587 "modules/graphics/opengl/Image.cpp"
 return ($0|0); //@line 587 "modules/graphics/opengl/Image.cpp"
}
function __ZN4love8graphics6opengl5Image11getConstantENS2_8FlagTypeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(1); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (633784 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 769 "modules/graphics/opengl/Image.cpp"
}
function __ZN4love8graphics6opengl4MeshD2Ev($this) {
 $this = $this|0;
 var $$01$i$i$i = 0, $$01$i$i$i9 = 0, $$pre$i = 0, $$pre$i4 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$018 = 0, $__begin$sroa$0$019 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i13 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i14 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (24932); //@line 131 "modules/graphics/opengl/Mesh.cpp"
 $0 = ((($this)) + 52|0); //@line 132 "modules/graphics/opengl/Mesh.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 132 "modules/graphics/opengl/Mesh.cpp"
 $2 = ($1|0)==(0|0); //@line 132 "modules/graphics/opengl/Mesh.cpp"
 if (!($2)) {
  $3 = HEAP32[$1>>2]|0; //@line 132 "modules/graphics/opengl/Mesh.cpp"
  $4 = ((($3)) + 4|0); //@line 132 "modules/graphics/opengl/Mesh.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 132 "modules/graphics/opengl/Mesh.cpp"
  FUNCTION_TABLE_vi[$5 & 8191]($1); //@line 132 "modules/graphics/opengl/Mesh.cpp"
 }
 $6 = ((($this)) + 68|0); //@line 133 "modules/graphics/opengl/Mesh.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 133 "modules/graphics/opengl/Mesh.cpp"
 $8 = ($7|0)==(0|0); //@line 133 "modules/graphics/opengl/Mesh.cpp"
 if (!($8)) {
  $9 = HEAP32[$7>>2]|0; //@line 133 "modules/graphics/opengl/Mesh.cpp"
  $10 = ((($9)) + 4|0); //@line 133 "modules/graphics/opengl/Mesh.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 133 "modules/graphics/opengl/Mesh.cpp"
  FUNCTION_TABLE_vi[$11 & 8191]($7); //@line 133 "modules/graphics/opengl/Mesh.cpp"
 }
 $12 = ((($this)) + 64|0); //@line 134 "modules/graphics/opengl/Mesh.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 134 "modules/graphics/opengl/Mesh.cpp"
 $14 = ($13|0)==(0|0); //@line 134 "modules/graphics/opengl/Mesh.cpp"
 if (!($14)) {
  __ZdlPv($13); //@line 134 "modules/graphics/opengl/Mesh.cpp"
 }
 $15 = ((($this)) + 32|0); //@line 136 "modules/graphics/opengl/Mesh.cpp"
 $16 = ((($this)) + 40|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__begin$sroa$0$018 = HEAP32[$16>>2]|0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $17 = ($__begin$sroa$0$018|0)==(0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L10: do {
  if (!($17)) {
   $__begin$sroa$0$019 = $__begin$sroa$0$018;
   while(1) {
    $98 = $__begin$sroa$0$019; //@line 189 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $99 = ((($98)) + 20|0); //@line 138 "modules/graphics/opengl/Mesh.cpp"
    $100 = HEAP32[$99>>2]|0; //@line 138 "modules/graphics/opengl/Mesh.cpp"
    $101 = ($100|0)==($this|0); //@line 138 "modules/graphics/opengl/Mesh.cpp"
    if (!($101)) {
     $102 = HEAP32[$100>>2]|0; //@line 139 "modules/graphics/opengl/Mesh.cpp"
     $103 = ((($102)) + 12|0); //@line 139 "modules/graphics/opengl/Mesh.cpp"
     $104 = HEAP32[$103>>2]|0; //@line 139 "modules/graphics/opengl/Mesh.cpp"
     __THREW__ = 0;
     invoke_vi($104|0,($100|0)); //@line 139 "modules/graphics/opengl/Mesh.cpp"
     $105 = __THREW__; __THREW__ = 0;
     $106 = $105&1;
     if ($106) {
      break;
     }
    }
    $107 = $__begin$sroa$0$019; //@line 174 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__begin$sroa$0$0 = HEAP32[$107>>2]|0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $108 = ($__begin$sroa$0$0|0)==(0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($108) {
     break L10;
    } else {
     $__begin$sroa$0$019 = $__begin$sroa$0$0;
    }
   }
   $57 = ___cxa_find_matching_catch_3(0|0)|0;
   $58 = tempRet0;
   $59 = ((($this)) + 96|0); //@line 106 "./common/Object.h"
   $60 = HEAP32[$59>>2]|0; //@line 106 "./common/Object.h"
   $61 = ($60|0)==(0|0); //@line 106 "./common/Object.h"
   if (!($61)) {
    $62 = HEAP32[$60>>2]|0; //@line 106 "./common/Object.h"
    $63 = ((($62)) + 12|0); //@line 106 "./common/Object.h"
    $64 = HEAP32[$63>>2]|0; //@line 106 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($64|0,($60|0)); //@line 106 "./common/Object.h"
    $65 = __THREW__; __THREW__ = 0;
    $66 = $65&1;
    if ($66) {
     $67 = ___cxa_find_matching_catch_3(0|0)|0;
     $68 = tempRet0;
     ___clang_call_terminate($67); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    }
   }
   $69 = HEAP32[$16>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $70 = ($69|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($70)) {
    $$01$i$i$i9 = $69;
    while(1) {
     $71 = HEAP32[$$01$i$i$i9>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $72 = ((($$01$i$i$i9)) + 8|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($72); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     __ZdlPv($$01$i$i$i9); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $73 = ($71|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     if ($73) {
      break;
     } else {
      $$01$i$i$i9 = $71;
     }
    }
   }
   $74 = HEAP32[$15>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$15>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $75 = ($74|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if (!($75)) {
    __ZdlPv($74); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $76 = ((($this)) + 20|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $77 = HEAP32[$76>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $78 = ($77|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $79 = $77; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($78)) {
    $80 = ((($this)) + 24|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $81 = HEAP32[$80>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $82 = ($81|0)==($77|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($82)) {
     $scevgep$i$i$i13 = ((($81)) + -4|0);
     $83 = $scevgep$i$i$i13;
     $84 = (($83) - ($79))|0;
     $85 = $84 >>> 2;
     $86 = $85 ^ -1;
     $scevgep4$i$i$i14 = (($81) + ($86<<2)|0);
     HEAP32[$80>>2] = $scevgep4$i$i$i14; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($77); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $87 = ((($this)) + 8|0); //@line 141 "modules/graphics/opengl/Mesh.cpp"
   $88 = HEAP32[$87>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $89 = ($88|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($89) {
    ___clang_call_terminate($57); //@line 141 "modules/graphics/opengl/Mesh.cpp"
    // unreachable; //@line 141 "modules/graphics/opengl/Mesh.cpp"
   }
   $90 = ((($this)) + 12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $91 = HEAP32[$90>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $92 = ($91|0)==($88|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($92) {
    $97 = $88;
   } else {
    $94 = $91;
    while(1) {
     $93 = ((($94)) + -20|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$90>>2] = $93; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($93); //@line 81 "modules/graphics/opengl/Mesh.h"
     $95 = HEAP32[$90>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $96 = ($95|0)==($88|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($96) {
      break;
     } else {
      $94 = $95;
     }
    }
    $$pre$i = HEAP32[$87>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $97 = $$pre$i;
   }
   __ZdlPv($97); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___clang_call_terminate($57); //@line 141 "modules/graphics/opengl/Mesh.cpp"
   // unreachable; //@line 141 "modules/graphics/opengl/Mesh.cpp"
  }
 } while(0);
 $18 = ((($this)) + 96|0); //@line 106 "./common/Object.h"
 $19 = HEAP32[$18>>2]|0; //@line 106 "./common/Object.h"
 $20 = ($19|0)==(0|0); //@line 106 "./common/Object.h"
 if (!($20)) {
  $21 = HEAP32[$19>>2]|0; //@line 106 "./common/Object.h"
  $22 = ((($21)) + 12|0); //@line 106 "./common/Object.h"
  $23 = HEAP32[$22>>2]|0; //@line 106 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($23|0,($19|0)); //@line 106 "./common/Object.h"
  $24 = __THREW__; __THREW__ = 0;
  $25 = $24&1;
  if ($25) {
   $26 = ___cxa_find_matching_catch_3(0|0)|0;
   $27 = tempRet0;
   ___clang_call_terminate($26); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 }
 $28 = HEAP32[$16>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $29 = ($28|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($29)) {
  $$01$i$i$i = $28;
  while(1) {
   $30 = HEAP32[$$01$i$i$i>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $31 = ((($$01$i$i$i)) + 8|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($31); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
   __ZdlPv($$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $32 = ($30|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($32) {
    break;
   } else {
    $$01$i$i$i = $30;
   }
  }
 }
 $33 = HEAP32[$15>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$15>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = ($33|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($34)) {
  __ZdlPv($33); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $35 = ((($this)) + 20|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $36 = HEAP32[$35>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $37 = ($36|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $38 = $36; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($37)) {
  $39 = ((($this)) + 24|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $40 = HEAP32[$39>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $41 = ($40|0)==($36|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($41)) {
   $scevgep$i$i$i = ((($40)) + -4|0);
   $42 = $scevgep$i$i$i;
   $43 = (($42) - ($38))|0;
   $44 = $43 >>> 2;
   $45 = $44 ^ -1;
   $scevgep4$i$i$i = (($40) + ($45<<2)|0);
   HEAP32[$39>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($36); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $46 = ((($this)) + 8|0); //@line 141 "modules/graphics/opengl/Mesh.cpp"
 $47 = HEAP32[$46>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $48 = ($47|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($48) {
  return; //@line 141 "modules/graphics/opengl/Mesh.cpp"
 }
 $49 = ((($this)) + 12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $50 = HEAP32[$49>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $51 = ($50|0)==($47|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($51) {
  $56 = $47;
 } else {
  $53 = $50;
  while(1) {
   $52 = ((($53)) + -20|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$49>>2] = $52; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($52); //@line 81 "modules/graphics/opengl/Mesh.h"
   $54 = HEAP32[$49>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $55 = ($54|0)==($47|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($55) {
    break;
   } else {
    $53 = $54;
   }
  }
  $$pre$i4 = HEAP32[$46>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $56 = $$pre$i4;
 }
 __ZdlPv($56); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 141 "modules/graphics/opengl/Mesh.cpp"
}
function __ZN4love8graphics6opengl4MeshD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl4MeshD2Ev($this); //@line 131 "modules/graphics/opengl/Mesh.cpp"
 __ZdlPv($this); //@line 131 "modules/graphics/opengl/Mesh.cpp"
 return; //@line 141 "modules/graphics/opengl/Mesh.cpp"
}
function __ZN4love8graphics6opengl4Mesh4drawEfffffffff($this,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $$0$i = 0, $$0$i6 = 0, $$0$i8 = 0, $$01 = 0, $$02 = 0, $$1 = 0, $$13 = 0, $$2 = 0, $$24 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$010 = 0, $__begin$sroa$0$012 = 0;
 var $enabledattribs$011 = 0, $enabledattribs$1 = 0, $enabledattribs$2 = 0, $enabledattribs$2$lcssa = 0, $m = 0, $max = 0, $max1 = 0, $min$0 = 0, $min2$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $m = sp + 16|0;
 $max = sp + 8|0;
 $max1 = sp + 4|0;
 $0 = ((($this)) + 40|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__begin$sroa$0$010 = HEAP32[$0>>2]|0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $1 = ($__begin$sroa$0$010|0)==(0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($1)) {
  $__begin$sroa$0$012 = $__begin$sroa$0$010;$enabledattribs$011 = 0;
  while(1) {
   $4 = $__begin$sroa$0$012; //@line 189 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $5 = ((($4)) + 28|0); //@line 581 "modules/graphics/opengl/Mesh.cpp"
   $6 = HEAP8[$5>>0]|0; //@line 581 "modules/graphics/opengl/Mesh.cpp"
   $7 = ($6<<24>>24)==(0); //@line 581 "modules/graphics/opengl/Mesh.cpp"
   if ($7) {
    $enabledattribs$2 = $enabledattribs$011;
   } else {
    $8 = ((($4)) + 20|0); //@line 584 "modules/graphics/opengl/Mesh.cpp"
    $9 = HEAP32[$8>>2]|0; //@line 584 "modules/graphics/opengl/Mesh.cpp"
    $10 = ((($4)) + 24|0); //@line 585 "modules/graphics/opengl/Mesh.cpp"
    $11 = HEAP32[$10>>2]|0; //@line 585 "modules/graphics/opengl/Mesh.cpp"
    $12 = ((($4)) + 8|0); //@line 585 "modules/graphics/opengl/Mesh.cpp"
    $13 = (__ZN4love8graphics6opengl4Mesh26bindAttributeToShaderInputEiRKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($9,$11,$12)|0); //@line 585 "modules/graphics/opengl/Mesh.cpp"
    $14 = ($13|0)>(-1); //@line 587 "modules/graphics/opengl/Mesh.cpp"
    $15 = 1 << $13; //@line 588 "modules/graphics/opengl/Mesh.cpp"
    $16 = $14 ? $15 : 0; //@line 587 "modules/graphics/opengl/Mesh.cpp"
    $enabledattribs$1 = $16 | $enabledattribs$011; //@line 587 "modules/graphics/opengl/Mesh.cpp"
    $enabledattribs$2 = $enabledattribs$1;
   }
   $17 = $__begin$sroa$0$012; //@line 174 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $__begin$sroa$0$0 = HEAP32[$17>>2]|0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $18 = ($__begin$sroa$0$0|0)==(0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($18) {
    $enabledattribs$2$lcssa = $enabledattribs$2;
    break;
   } else {
    $__begin$sroa$0$012 = $__begin$sroa$0$0;$enabledattribs$011 = $enabledattribs$2;
   }
  }
  $2 = $enabledattribs$2$lcssa & 1; //@line 592 "modules/graphics/opengl/Mesh.cpp"
  $3 = ($2|0)==(0); //@line 592 "modules/graphics/opengl/Mesh.cpp"
  if (!($3)) {
   __ZN4love8graphics6opengl6OpenGL21useVertexAttribArraysEj(634044,$enabledattribs$2$lcssa); //@line 595 "modules/graphics/opengl/Mesh.cpp"
   $24 = ((($this)) + 96|0); //@line 139 "./common/Object.h"
   $25 = HEAP32[$24>>2]|0; //@line 139 "./common/Object.h"
   $26 = ($25|0)==(0|0); //@line 597 "modules/graphics/opengl/Mesh.cpp"
   if ($26) {
    $32 = (__ZNK4love8graphics6opengl6OpenGL17getDefaultTextureEv(634044)|0); //@line 600 "modules/graphics/opengl/Mesh.cpp"
    __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$32); //@line 600 "modules/graphics/opengl/Mesh.cpp"
   } else {
    $27 = HEAP32[$25>>2]|0; //@line 598 "modules/graphics/opengl/Mesh.cpp"
    $28 = ((($27)) + 52|0); //@line 598 "modules/graphics/opengl/Mesh.cpp"
    $29 = HEAP32[$28>>2]|0; //@line 598 "modules/graphics/opengl/Mesh.cpp"
    $30 = (FUNCTION_TABLE_ii[$29 & 8191]($25)|0); //@line 598 "modules/graphics/opengl/Mesh.cpp"
    $31 = HEAP32[$30>>2]|0; //@line 598 "modules/graphics/opengl/Mesh.cpp"
    __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$31); //@line 598 "modules/graphics/opengl/Mesh.cpp"
   }
   __ZN4love7Matrix4C2Efffffffff($m,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky); //@line 602 "modules/graphics/opengl/Mesh.cpp"
   __THREW__ = 0;
   invoke_vi(4086,(634044|0)); //@line 125 "modules/graphics/opengl/OpenGL.h"
   $33 = __THREW__; __THREW__ = 0;
   $34 = $33&1;
   if ($34) {
    $64 = ___cxa_find_matching_catch_2()|0;
    $65 = tempRet0;
    $$1 = $65;$$13 = $64;
   } else {
    __THREW__ = 0;
    $35 = (invoke_ii(4003,(634044|0))|0); //@line 135 "modules/graphics/opengl/OpenGL.h"
    $36 = __THREW__; __THREW__ = 0;
    $37 = $36&1;
    L16: do {
     if ($37) {
      label = 24;
     } else {
      __THREW__ = 0;
      invoke_vii(4004,($35|0),($m|0)); //@line 605 "modules/graphics/opengl/Mesh.cpp"
      $38 = __THREW__; __THREW__ = 0;
      $39 = $38&1;
      if ($39) {
       label = 24;
      } else {
       __THREW__ = 0;
       invoke_vi(4007,(634044|0)); //@line 607 "modules/graphics/opengl/Mesh.cpp"
       $40 = __THREW__; __THREW__ = 0;
       $41 = $40&1;
       if ($41) {
        label = 24;
       } else {
        $42 = ((($this)) + 72|0); //@line 609 "modules/graphics/opengl/Mesh.cpp"
        $43 = HEAP8[$42>>0]|0; //@line 609 "modules/graphics/opengl/Mesh.cpp"
        $44 = ($43<<24>>24)==(0); //@line 609 "modules/graphics/opengl/Mesh.cpp"
        L20: do {
         if ($44) {
          label = 41;
         } else {
          $45 = ((($this)) + 68|0); //@line 609 "modules/graphics/opengl/Mesh.cpp"
          $46 = HEAP32[$45>>2]|0; //@line 609 "modules/graphics/opengl/Mesh.cpp"
          $47 = ($46|0)==(0|0); //@line 609 "modules/graphics/opengl/Mesh.cpp"
          if ($47) {
           label = 41;
          } else {
           $48 = ((($this)) + 76|0); //@line 609 "modules/graphics/opengl/Mesh.cpp"
           $49 = HEAP32[$48>>2]|0; //@line 609 "modules/graphics/opengl/Mesh.cpp"
           $50 = ($49|0)==(0); //@line 609 "modules/graphics/opengl/Mesh.cpp"
           if ($50) {
            label = 41;
           } else {
            __THREW__ = 0;
            invoke_vi(4087,($46|0)); //@line 188 "modules/graphics/opengl/GLBuffer.h"
            $51 = __THREW__; __THREW__ = 0;
            $52 = $51&1;
            if ($52) {
             label = 24;
             break L16;
            }
            $53 = HEAP32[$45>>2]|0; //@line 615 "modules/graphics/opengl/Mesh.cpp"
            __THREW__ = 0;
            invoke_vi(4088,($53|0)); //@line 615 "modules/graphics/opengl/Mesh.cpp"
            $54 = __THREW__; __THREW__ = 0;
            $55 = $54&1;
            do {
             if (!($55)) {
              $56 = HEAP32[$48>>2]|0; //@line 617 "modules/graphics/opengl/Mesh.cpp"
              $57 = (($56) + -1)|0; //@line 617 "modules/graphics/opengl/Mesh.cpp"
              HEAP32[$max>>2] = $57; //@line 617 "modules/graphics/opengl/Mesh.cpp"
              $58 = ((($this)) + 92|0); //@line 618 "modules/graphics/opengl/Mesh.cpp"
              $59 = HEAP32[$58>>2]|0; //@line 618 "modules/graphics/opengl/Mesh.cpp"
              $60 = ($59|0)>(-1); //@line 618 "modules/graphics/opengl/Mesh.cpp"
              if ($60) {
               $61 = ($56|0)<=($59|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
               $62 = $61 ? $max : $58; //@line 2575 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
               $63 = HEAP32[$62>>2]|0; //@line 619 "modules/graphics/opengl/Mesh.cpp"
               HEAP32[$max>>2] = $63; //@line 619 "modules/graphics/opengl/Mesh.cpp"
               $78 = $63;
              } else {
               $78 = $57;
              }
              $74 = ((($this)) + 88|0); //@line 622 "modules/graphics/opengl/Mesh.cpp"
              $75 = HEAP32[$74>>2]|0; //@line 622 "modules/graphics/opengl/Mesh.cpp"
              $76 = ($75|0)>(-1); //@line 622 "modules/graphics/opengl/Mesh.cpp"
              if ($76) {
               $77 = ($78|0)<($75|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
               $79 = $77 ? $max : $74; //@line 2575 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
               $80 = HEAP32[$79>>2]|0; //@line 623 "modules/graphics/opengl/Mesh.cpp"
               $min$0 = $80;
              } else {
               $min$0 = 0;
              }
              $81 = ((($this)) + 80|0); //@line 625 "modules/graphics/opengl/Mesh.cpp"
              $82 = HEAP32[$81>>2]|0; //@line 625 "modules/graphics/opengl/Mesh.cpp"
              $83 = HEAP32[$45>>2]|0; //@line 626 "modules/graphics/opengl/Mesh.cpp"
              switch ($82|0) {
              case 5121:  {
               $$0$i8 = 1;
               break;
              }
              case 5123:  {
               $$0$i8 = 2;
               break;
              }
              case 5125:  {
               $$0$i8 = 4;
               break;
              }
              default: {
               $$0$i8 = 0;
              }
              }
              $84 = Math_imul($$0$i8, $min$0)|0; //@line 626 "modules/graphics/opengl/Mesh.cpp"
              __THREW__ = 0;
              $85 = (invoke_iii(4090,($83|0),($84|0))|0); //@line 626 "modules/graphics/opengl/Mesh.cpp"
              $86 = __THREW__; __THREW__ = 0;
              $87 = $86&1;
              if (!($87)) {
               $88 = ((($this)) + 84|0); //@line 628 "modules/graphics/opengl/Mesh.cpp"
               $89 = HEAP32[$88>>2]|0; //@line 628 "modules/graphics/opengl/Mesh.cpp"
               switch ($89|0) {
               case 0:  {
                $$0$i6 = 6;
                break;
               }
               case 1:  {
                $$0$i6 = 5;
                break;
               }
               case 3:  {
                $$0$i6 = 0;
                break;
               }
               default: {
                $$0$i6 = 4;
               }
               }
               $90 = (1 - ($min$0))|0; //@line 628 "modules/graphics/opengl/Mesh.cpp"
               $91 = (($90) + ($78))|0; //@line 628 "modules/graphics/opengl/Mesh.cpp"
               __THREW__ = 0;
               invoke_viiiii(4091,(634044|0),($$0$i6|0),($91|0),($82|0),($85|0)); //@line 628 "modules/graphics/opengl/Mesh.cpp"
               $92 = __THREW__; __THREW__ = 0;
               $93 = $92&1;
               if ($93) {
                break;
               }
               __THREW__ = 0;
               invoke_vi(4089,($46|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
               $94 = __THREW__; __THREW__ = 0;
               $95 = $94&1;
               if (!($95)) {
                break L20;
               }
               $96 = ___cxa_find_matching_catch_3(0|0)|0;
               $97 = tempRet0;
               ___clang_call_terminate($96); //@line 196 "modules/graphics/opengl/GLBuffer.h"
               // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
              }
             }
            } while(0);
            $68 = ___cxa_find_matching_catch_2()|0;
            $69 = tempRet0;
            __THREW__ = 0;
            invoke_vi(4089,($46|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
            $70 = __THREW__; __THREW__ = 0;
            $71 = $70&1;
            if (!($71)) {
             $$01 = $69;$$02 = $68;
             break L16;
            }
            $72 = ___cxa_find_matching_catch_3(0|0)|0;
            $73 = tempRet0;
            ___clang_call_terminate($72); //@line 196 "modules/graphics/opengl/GLBuffer.h"
            // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
           }
          }
         }
        } while(0);
        if ((label|0) == 41) {
         $98 = ((($this)) + 56|0); //@line 632 "modules/graphics/opengl/Mesh.cpp"
         $99 = HEAP32[$98>>2]|0; //@line 632 "modules/graphics/opengl/Mesh.cpp"
         $100 = (($99) + -1)|0; //@line 632 "modules/graphics/opengl/Mesh.cpp"
         HEAP32[$max1>>2] = $100; //@line 632 "modules/graphics/opengl/Mesh.cpp"
         $101 = ((($this)) + 92|0); //@line 633 "modules/graphics/opengl/Mesh.cpp"
         $102 = HEAP32[$101>>2]|0; //@line 633 "modules/graphics/opengl/Mesh.cpp"
         $103 = ($102|0)>(-1); //@line 633 "modules/graphics/opengl/Mesh.cpp"
         if ($103) {
          $104 = ($99|0)<=($102|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
          $105 = $104 ? $max1 : $101; //@line 2575 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
          $106 = HEAP32[$105>>2]|0; //@line 634 "modules/graphics/opengl/Mesh.cpp"
          HEAP32[$max1>>2] = $106; //@line 634 "modules/graphics/opengl/Mesh.cpp"
          $111 = $106;
         } else {
          $111 = $100;
         }
         $107 = ((($this)) + 88|0); //@line 637 "modules/graphics/opengl/Mesh.cpp"
         $108 = HEAP32[$107>>2]|0; //@line 637 "modules/graphics/opengl/Mesh.cpp"
         $109 = ($108|0)>(-1); //@line 637 "modules/graphics/opengl/Mesh.cpp"
         if ($109) {
          $110 = ($111|0)<($108|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
          $112 = $110 ? $max1 : $107; //@line 2575 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
          $113 = HEAP32[$112>>2]|0; //@line 638 "modules/graphics/opengl/Mesh.cpp"
          $min2$0 = $113;
         } else {
          $min2$0 = 0;
         }
         $114 = ((($this)) + 84|0); //@line 641 "modules/graphics/opengl/Mesh.cpp"
         $115 = HEAP32[$114>>2]|0; //@line 641 "modules/graphics/opengl/Mesh.cpp"
         switch ($115|0) {
         case 0:  {
          $$0$i = 6;
          break;
         }
         case 1:  {
          $$0$i = 5;
          break;
         }
         case 3:  {
          $$0$i = 0;
          break;
         }
         default: {
          $$0$i = 4;
         }
         }
         $116 = (1 - ($min2$0))|0; //@line 641 "modules/graphics/opengl/Mesh.cpp"
         $117 = (($116) + ($111))|0; //@line 641 "modules/graphics/opengl/Mesh.cpp"
         __THREW__ = 0;
         invoke_viiii(4008,(634044|0),($$0$i|0),($min2$0|0),($117|0)); //@line 641 "modules/graphics/opengl/Mesh.cpp"
         $118 = __THREW__; __THREW__ = 0;
         $119 = $118&1;
         if ($119) {
          label = 24;
          break;
         }
        }
        __THREW__ = 0;
        invoke_vi(4009,(634044|0)); //@line 130 "modules/graphics/opengl/OpenGL.h"
        $120 = __THREW__; __THREW__ = 0;
        $121 = $120&1;
        if ($121) {
         $122 = ___cxa_find_matching_catch_3(0|0)|0;
         $123 = tempRet0;
         ___clang_call_terminate($122); //@line 130 "modules/graphics/opengl/OpenGL.h"
         // unreachable; //@line 130 "modules/graphics/opengl/OpenGL.h"
        } else {
         __ZN4love7Matrix4D2Ev($m); //@line 643 "modules/graphics/opengl/Mesh.cpp"
         STACKTOP = sp;return; //@line 643 "modules/graphics/opengl/Mesh.cpp"
        }
       }
      }
     }
    } while(0);
    if ((label|0) == 24) {
     $66 = ___cxa_find_matching_catch_2()|0;
     $67 = tempRet0;
     $$01 = $67;$$02 = $66;
    }
    __THREW__ = 0;
    invoke_vi(4009,(634044|0)); //@line 130 "modules/graphics/opengl/OpenGL.h"
    $124 = __THREW__; __THREW__ = 0;
    $125 = $124&1;
    if ($125) {
     $126 = ___cxa_find_matching_catch_3(0|0)|0;
     $127 = tempRet0;
     ___clang_call_terminate($126); //@line 130 "modules/graphics/opengl/OpenGL.h"
     // unreachable; //@line 130 "modules/graphics/opengl/OpenGL.h"
    } else {
     $$1 = $$01;$$13 = $$02;
    }
   }
   __ZN4love7Matrix4D2Ev($m); //@line 643 "modules/graphics/opengl/Mesh.cpp"
   $$2 = $$1;$$24 = $$13;
   ___resumeException($$24|0);
   // unreachable;
  }
 }
 $19 = (___cxa_allocate_exception(16)|0); //@line 593 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($19|0),(436744|0),($vararg_buffer|0)); //@line 593 "modules/graphics/opengl/Mesh.cpp"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if (!($21)) {
  ___cxa_throw(($19|0),(96|0),(344|0)); //@line 593 "modules/graphics/opengl/Mesh.cpp"
  // unreachable; //@line 593 "modules/graphics/opengl/Mesh.cpp"
 }
 $22 = ___cxa_find_matching_catch_2()|0;
 $23 = tempRet0;
 ___cxa_free_exception(($19|0)); //@line 593 "modules/graphics/opengl/Mesh.cpp"
 $$2 = $23;$$24 = $22;
 ___resumeException($$24|0);
 // unreachable;
}
function __ZN4love8graphics6opengl4Mesh26bindAttributeToShaderInputEiRKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($this,$attributeindex,$inputname) {
 $this = $this|0;
 $attributeindex = $attributeindex|0;
 $inputname = $inputname|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $attriblocation$0 = 0, $attriblocation$02 = 0, $builtinattrib = 0, $exitcond$i = 0, $i$02$i = 0, $offset$0$lcssa$i = 0, $offset$01$i = 0, $switch$select$i = 0, $switch$select2$i = 0;
 var $switch$selectcmp$i = 0, $switch$selectcmp1$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $builtinattrib = sp;
 $0 = ((($this)) + 8|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP8[$inputname>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = $2 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ($3<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($inputname)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = HEAP32[$5>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($inputname)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $4 ? $7 : $6; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = (__ZN4love8graphics6opengl6Shader11getConstantEPKcRNS1_14VertexAttribIDE($8,$builtinattrib)|0); //@line 549 "modules/graphics/opengl/Mesh.cpp"
 do {
  if ($9) {
   $10 = HEAP32[$builtinattrib>>2]|0; //@line 550 "modules/graphics/opengl/Mesh.cpp"
   $attriblocation$0 = $10;
  } else {
   $11 = HEAP32[158581]|0; //@line 551 "modules/graphics/opengl/Mesh.cpp"
   $12 = ($11|0)==(0|0); //@line 551 "modules/graphics/opengl/Mesh.cpp"
   if ($12) {
    $attriblocation$02 = -1;
    STACKTOP = sp;return ($attriblocation$02|0); //@line 571 "modules/graphics/opengl/Mesh.cpp"
   } else {
    $13 = (__ZN4love8graphics6opengl6Shader17getAttribLocationERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($11,$inputname)|0); //@line 552 "modules/graphics/opengl/Mesh.cpp"
    $attriblocation$0 = $13;
    break;
   }
  }
 } while(0);
 $14 = ($attriblocation$0|0)<(0); //@line 555 "modules/graphics/opengl/Mesh.cpp"
 if ($14) {
  $attriblocation$02 = $attriblocation$0;
  STACKTOP = sp;return ($attriblocation$02|0); //@line 571 "modules/graphics/opengl/Mesh.cpp"
 }
 $15 = ((($this)) + 52|0); //@line 559 "modules/graphics/opengl/Mesh.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 559 "modules/graphics/opengl/Mesh.cpp"
 __ZN4love8graphics6opengl8GLBuffer4bindEv($16); //@line 188 "modules/graphics/opengl/GLBuffer.h"
 $17 = HEAP32[$15>>2]|0; //@line 562 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 invoke_vi(4088,($17|0)); //@line 562 "modules/graphics/opengl/Mesh.cpp"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if (!($19)) {
  $20 = HEAP32[$15>>2]|0; //@line 564 "modules/graphics/opengl/Mesh.cpp"
  $21 = ($attributeindex|0)==(0); //@line 181 "modules/graphics/opengl/Mesh.cpp"
  if ($21) {
   $offset$0$lcssa$i = 0;
  } else {
   $22 = ((($this)) + 20|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $23 = HEAP32[$22>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $i$02$i = 0;$offset$01$i = 0;
   while(1) {
    $24 = (($23) + ($i$02$i<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $25 = HEAP32[$24>>2]|0; //@line 182 "modules/graphics/opengl/Mesh.cpp"
    $26 = (($25) + ($offset$01$i))|0; //@line 182 "modules/graphics/opengl/Mesh.cpp"
    $27 = (($i$02$i) + 1)|0; //@line 181 "modules/graphics/opengl/Mesh.cpp"
    $exitcond$i = ($27|0)==($attributeindex|0); //@line 181 "modules/graphics/opengl/Mesh.cpp"
    if ($exitcond$i) {
     $offset$0$lcssa$i = $26;
     break;
    } else {
     $i$02$i = $27;$offset$01$i = $26;
    }
   }
  }
  __THREW__ = 0;
  $28 = (invoke_iii(4090,($20|0),($offset$0$lcssa$i|0))|0); //@line 564 "modules/graphics/opengl/Mesh.cpp"
  $29 = __THREW__; __THREW__ = 0;
  $30 = $29&1;
  if (!($30)) {
   $31 = (((($1) + (($attributeindex*20)|0)|0)) + 12|0); //@line 565 "modules/graphics/opengl/Mesh.cpp"
   $32 = HEAP32[$31>>2]|0; //@line 565 "modules/graphics/opengl/Mesh.cpp"
   $switch$selectcmp$i = ($32|0)==(1); //@line 676 "modules/graphics/opengl/Mesh.cpp"
   $switch$select$i = $switch$selectcmp$i ? 5126 : 0; //@line 676 "modules/graphics/opengl/Mesh.cpp"
   $switch$selectcmp1$i = ($32|0)==(0); //@line 676 "modules/graphics/opengl/Mesh.cpp"
   $switch$select2$i = $switch$selectcmp1$i ? 5121 : $switch$select$i; //@line 676 "modules/graphics/opengl/Mesh.cpp"
   $33 = $switch$selectcmp1$i&1; //@line 566 "modules/graphics/opengl/Mesh.cpp"
   $34 = (((($1) + (($attributeindex*20)|0)|0)) + 16|0); //@line 568 "modules/graphics/opengl/Mesh.cpp"
   $35 = HEAP32[$34>>2]|0; //@line 568 "modules/graphics/opengl/Mesh.cpp"
   $36 = ((($this)) + 60|0); //@line 568 "modules/graphics/opengl/Mesh.cpp"
   $37 = HEAP32[$36>>2]|0; //@line 568 "modules/graphics/opengl/Mesh.cpp"
   $38 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   invoke_viiiiii($38|0,($attriblocation$0|0),($35|0),($switch$select2$i|0),($33|0),($37|0),($28|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
   $39 = __THREW__; __THREW__ = 0;
   $40 = $39&1;
   if (!($40)) {
    __THREW__ = 0;
    invoke_vi(4089,($16|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
    $41 = __THREW__; __THREW__ = 0;
    $42 = $41&1;
    if ($42) {
     $43 = ___cxa_find_matching_catch_3(0|0)|0;
     $44 = tempRet0;
     ___clang_call_terminate($43); //@line 196 "modules/graphics/opengl/GLBuffer.h"
     // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
    } else {
     $attriblocation$02 = $attriblocation$0;
     STACKTOP = sp;return ($attriblocation$02|0); //@line 571 "modules/graphics/opengl/Mesh.cpp"
    }
   }
  }
 }
 $45 = ___cxa_find_matching_catch_2()|0;
 $46 = tempRet0;
 __THREW__ = 0;
 invoke_vi(4089,($16|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
 $47 = __THREW__; __THREW__ = 0;
 $48 = $47&1;
 if ($48) {
  $49 = ___cxa_find_matching_catch_3(0|0)|0;
  $50 = tempRet0;
  ___clang_call_terminate($49); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
 } else {
  ___resumeException($45|0);
  // unreachable;
 }
 return (0)|0;
}
function __GLOBAL__sub_I_Mesh_cpp() {
 var $$02$i$i$1$i = 0, $$02$i$i$i = 0, $$lcssa$i$1$i = 0, $$lcssa$i$i = 0, $$lcssa10$i$1$i = 0, $$lcssa10$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $9 = 0, $hash$0$lcssa$i$i$1$i = 0, $hash$0$lcssa$i$i$i = 0, $hash$01$i$i$1$i = 0, $hash$01$i$i$i = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 HEAP8[(633800)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633812)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633824)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633836)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633848)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633860)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(633864)>>2]=0|0;HEAP32[(633864)+4>>2]=0|0;HEAP32[(633864)+8>>2]=0|0; //@line 44 "./common/StringMap.h"
 $0 = HEAP32[6238]|0; //@line 49 "./common/StringMap.h"
 $1 = HEAP32[(24956)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl4Mesh5UsageELj3EE3addEPKcS4_(633792,$0,$1)|0); //@line 49 "./common/StringMap.h"
 $2 = HEAP32[(24960)>>2]|0; //@line 49 "./common/StringMap.h"
 $3 = HEAP32[(24964)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl4Mesh5UsageELj3EE3addEPKcS4_(633792,$2,$3)|0); //@line 49 "./common/StringMap.h"
 $4 = HEAP32[(24968)>>2]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(24972)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl4Mesh5UsageELj3EE3addEPKcS4_(633792,$4,$5)|0); //@line 49 "./common/StringMap.h"
 HEAP8[(633884)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633896)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633908)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633920)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633932)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633944)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633956)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(633968)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(633972)>>2]=0|0;HEAP32[(633972)+4>>2]=0|0;HEAP32[(633972)+8>>2]=0|0;HEAP32[(633972)+12>>2]=0|0; //@line 44 "./common/StringMap.h"
 $6 = HEAP32[6244]|0; //@line 49 "./common/StringMap.h"
 $7 = HEAP32[(24980)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl4Mesh8DrawModeELj4EE3addEPKcS4_(633876,$6,$7)|0); //@line 49 "./common/StringMap.h"
 $8 = HEAP32[(24984)>>2]|0; //@line 49 "./common/StringMap.h"
 $9 = HEAP32[(24988)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl4Mesh8DrawModeELj4EE3addEPKcS4_(633876,$8,$9)|0); //@line 49 "./common/StringMap.h"
 $10 = HEAP32[(24992)>>2]|0; //@line 49 "./common/StringMap.h"
 $11 = HEAP32[(24996)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl4Mesh8DrawModeELj4EE3addEPKcS4_(633876,$10,$11)|0); //@line 49 "./common/StringMap.h"
 $12 = HEAP32[(25000)>>2]|0; //@line 49 "./common/StringMap.h"
 $13 = HEAP32[(25004)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl4Mesh8DrawModeELj4EE3addEPKcS4_(633876,$12,$13)|0); //@line 49 "./common/StringMap.h"
 HEAP8[(633996)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634008)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634020)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634032)>>0] = 0; //@line 155 "./common/StringMap.h"
 $14 = (634036); //@line 44 "./common/StringMap.h"
 $15 = $14; //@line 44 "./common/StringMap.h"
 HEAP32[$15>>2] = 0; //@line 44 "./common/StringMap.h"
 $16 = (($14) + 4)|0; //@line 44 "./common/StringMap.h"
 $17 = $16; //@line 44 "./common/StringMap.h"
 HEAP32[$17>>2] = 0; //@line 44 "./common/StringMap.h"
 $18 = HEAP32[6252]|0; //@line 49 "./common/StringMap.h"
 $19 = HEAP32[(25012)>>2]|0; //@line 49 "./common/StringMap.h"
 $20 = HEAP8[$18>>0]|0; //@line 142 "./common/StringMap.h"
 $21 = ($20<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($21) {
  $hash$0$lcssa$i$i$i = 5381;
 } else {
  $$02$i$i$i = $18;$23 = $20;$hash$01$i$i$i = 5381;
  while(1) {
   $22 = $23 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $24 = ((($$02$i$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $25 = ($hash$01$i$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $26 = (($25) + ($22))|0; //@line 143 "./common/StringMap.h"
   $27 = HEAP8[$24>>0]|0; //@line 142 "./common/StringMap.h"
   $28 = ($27<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($28) {
    $hash$0$lcssa$i$i$i = $26;
    break;
   } else {
    $$02$i$i$i = $24;$23 = $27;$hash$01$i$i$i = $26;
   }
  }
 }
 $29 = $hash$0$lcssa$i$i$i & 3; //@line 112 "./common/StringMap.h"
 $30 = (((633988 + (($29*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $31 = HEAP8[$30>>0]|0; //@line 114 "./common/StringMap.h"
 $32 = ($31<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($32) {
  $$lcssa$i$i = $29;$$lcssa10$i$i = $30;
  label = 5;
 } else {
  $33 = (($hash$0$lcssa$i$i$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $34 = $33 & 3; //@line 112 "./common/StringMap.h"
  $35 = (((633988 + (($34*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $36 = HEAP8[$35>>0]|0; //@line 114 "./common/StringMap.h"
  $37 = ($36<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($37) {
   $$lcssa$i$i = $34;$$lcssa10$i$i = $35;
   label = 5;
  } else {
   $42 = (($hash$0$lcssa$i$i$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $43 = $42 & 3; //@line 112 "./common/StringMap.h"
   $44 = (((633988 + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
   $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($46) {
    $$lcssa$i$i = $43;$$lcssa10$i$i = $44;
    label = 5;
   } else {
    $47 = (($hash$0$lcssa$i$i$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $48 = $47 & 3; //@line 112 "./common/StringMap.h"
    $49 = (((633988 + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
    $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($51) {
     $$lcssa$i$i = $48;$$lcssa10$i$i = $49;
     label = 5;
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10$i$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $38 = (633988 + (($$lcssa$i$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$38>>2] = $18; //@line 118 "./common/StringMap.h"
  $39 = (((633988 + (($$lcssa$i$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$39>>2] = $19; //@line 119 "./common/StringMap.h"
 }
 $40 = ($19>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($40) {
  HEAP32[$vararg_buffer>>2] = $18; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $19; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
 } else {
  $41 = (634036 + ($19<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$41>>2] = $18; //@line 132 "./common/StringMap.h"
 }
 $52 = HEAP32[(25016)>>2]|0; //@line 49 "./common/StringMap.h"
 $53 = HEAP32[(25020)>>2]|0; //@line 49 "./common/StringMap.h"
 $54 = HEAP8[$52>>0]|0; //@line 142 "./common/StringMap.h"
 $55 = ($54<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($55) {
  $hash$0$lcssa$i$i$1$i = 5381;
 } else {
  $$02$i$i$1$i = $52;$57 = $54;$hash$01$i$i$1$i = 5381;
  while(1) {
   $56 = $57 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $58 = ((($$02$i$i$1$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $59 = ($hash$01$i$i$1$i*33)|0; //@line 143 "./common/StringMap.h"
   $60 = (($59) + ($56))|0; //@line 143 "./common/StringMap.h"
   $61 = HEAP8[$58>>0]|0; //@line 142 "./common/StringMap.h"
   $62 = ($61<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($62) {
    $hash$0$lcssa$i$i$1$i = $60;
    break;
   } else {
    $$02$i$i$1$i = $58;$57 = $61;$hash$01$i$i$1$i = $60;
   }
  }
 }
 $63 = $hash$0$lcssa$i$i$1$i & 3; //@line 112 "./common/StringMap.h"
 $64 = (((633988 + (($63*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $65 = HEAP8[$64>>0]|0; //@line 114 "./common/StringMap.h"
 $66 = ($65<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($66) {
  $$lcssa$i$1$i = $63;$$lcssa10$i$1$i = $64;
  label = 17;
 } else {
  $67 = (($hash$0$lcssa$i$i$1$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $68 = $67 & 3; //@line 112 "./common/StringMap.h"
  $69 = (((633988 + (($68*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $70 = HEAP8[$69>>0]|0; //@line 114 "./common/StringMap.h"
  $71 = ($70<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($71) {
   $$lcssa$i$1$i = $68;$$lcssa10$i$1$i = $69;
   label = 17;
  } else {
   $72 = (($hash$0$lcssa$i$i$1$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $73 = $72 & 3; //@line 112 "./common/StringMap.h"
   $74 = (((633988 + (($73*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $75 = HEAP8[$74>>0]|0; //@line 114 "./common/StringMap.h"
   $76 = ($75<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($76) {
    $$lcssa$i$1$i = $73;$$lcssa10$i$1$i = $74;
    label = 17;
   } else {
    $77 = (($hash$0$lcssa$i$i$1$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $78 = $77 & 3; //@line 112 "./common/StringMap.h"
    $79 = (((633988 + (($78*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $80 = HEAP8[$79>>0]|0; //@line 114 "./common/StringMap.h"
    $81 = ($80<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($81) {
     $$lcssa$i$1$i = $78;$$lcssa10$i$1$i = $79;
     label = 17;
    }
   }
  }
 }
 if ((label|0) == 17) {
  HEAP8[$$lcssa10$i$1$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $82 = (633988 + (($$lcssa$i$1$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$82>>2] = $52; //@line 118 "./common/StringMap.h"
  $83 = (((633988 + (($$lcssa$i$1$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$83>>2] = $53; //@line 119 "./common/StringMap.h"
 }
 $84 = ($53>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($84) {
  HEAP32[$vararg_buffer2>>2] = $52; //@line 128 "./common/StringMap.h"
  $vararg_ptr5 = ((($vararg_buffer2)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr5>>2] = $53; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer2)|0); //@line 128 "./common/StringMap.h"
  STACKTOP = sp;return;
 } else {
  $85 = (634036 + ($53<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$85>>2] = $52; //@line 132 "./common/StringMap.h"
  STACKTOP = sp;return;
 }
}
function __ZN4love9StringMapINS_8graphics6opengl4Mesh5UsageELj3EE3addEPKcS4_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $inserted$2$off0 = 0;
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(2); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 72|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love9StringMapINS_8graphics6opengl4Mesh8DrawModeELj4EE3addEPKcS4_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0;
 var $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 7; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = $13 & 7; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = $22 & 7; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = $27 & 7; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = $32 & 7; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = $37 & 7; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $42 = (($hash$0$lcssa$i) + 6)|0; //@line 112 "./common/StringMap.h"
       $43 = $42 & 7; //@line 112 "./common/StringMap.h"
       $44 = (((($this) + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
       $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
       $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
       if ($46) {
        $$lcssa = $43;$$lcssa10 = $44;
        label = 5;
       } else {
        $47 = (($hash$0$lcssa$i) + 7)|0; //@line 112 "./common/StringMap.h"
        $48 = $47 & 7; //@line 112 "./common/StringMap.h"
        $49 = (((($this) + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
        $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
        $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
        if ($51) {
         $$lcssa = $48;$$lcssa10 = $49;
         label = 5;
        } else {
         $inserted$2$off0 = 0;
        }
       }
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(3); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 96|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl4MeshC2ERKNSt3__16vectorINS2_12AttribFormatENS3_9allocatorIS5_EEEEPKvjNS2_8DrawModeENS2_5UsageE($this,$vertexformat,$data,$datasize,$drawmode,$usage) {
 $this = $this|0;
 $vertexformat = $vertexformat|0;
 $data = $data|0;
 $datasize = $datasize|0;
 $drawmode = $drawmode|0;
 $usage = $usage|0;
 var $$$i = 0, $$0 = 0, $$0$i = 0, $$01 = 0, $$01$i$i$i = 0, $$1 = 0, $$12 = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $9 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 __ZN4love6ObjectC2Ev($this); //@line 36 "./modules/graphics/Drawable.h"
 HEAP32[$this>>2] = (24932); //@line 73 "modules/graphics/opengl/Mesh.cpp"
 $0 = ((($this)) + 8|0); //@line 62 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 invoke_vii(4092,($0|0),($vertexformat|0)); //@line 62 "modules/graphics/opengl/Mesh.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $29 = ___cxa_find_matching_catch_2()|0;
  $30 = tempRet0;
  $$1 = $30;$$12 = $29;
  ___resumeException($$12|0);
  // unreachable;
 }
 $3 = ((($this)) + 20|0); //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ((($this)) + 24|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ((($this)) + 48|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 ;HEAP32[$3>>2]=0|0;HEAP32[$3+4>>2]=0|0;HEAP32[$3+8>>2]=0|0;HEAP32[$3+12>>2]=0|0;HEAP32[$3+16>>2]=0|0;HEAP32[$3+20>>2]=0|0;HEAP32[$3+24>>2]=0|0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAPF32[$5>>2] = 1.0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = ((($this)) + 52|0); //@line 63 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$6>>2] = 0; //@line 63 "modules/graphics/opengl/Mesh.cpp"
 $7 = ((($this)) + 56|0); //@line 64 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$7>>2] = 0; //@line 64 "modules/graphics/opengl/Mesh.cpp"
 $8 = ((($this)) + 60|0); //@line 65 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$8>>2] = 0; //@line 65 "modules/graphics/opengl/Mesh.cpp"
 $9 = ((($this)) + 68|0); //@line 66 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$9>>2] = 0; //@line 66 "modules/graphics/opengl/Mesh.cpp"
 $10 = ((($this)) + 72|0); //@line 67 "modules/graphics/opengl/Mesh.cpp"
 HEAP8[$10>>0] = 0; //@line 67 "modules/graphics/opengl/Mesh.cpp"
 $11 = ((($this)) + 76|0); //@line 68 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$11>>2] = 0; //@line 68 "modules/graphics/opengl/Mesh.cpp"
 $12 = ((($this)) + 80|0); //@line 69 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$12>>2] = 0; //@line 69 "modules/graphics/opengl/Mesh.cpp"
 $13 = ((($this)) + 84|0); //@line 70 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$13>>2] = $drawmode; //@line 70 "modules/graphics/opengl/Mesh.cpp"
 $14 = ((($this)) + 88|0); //@line 71 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$14>>2] = -1; //@line 71 "modules/graphics/opengl/Mesh.cpp"
 $15 = ((($this)) + 92|0); //@line 72 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$15>>2] = -1; //@line 72 "modules/graphics/opengl/Mesh.cpp"
 $16 = ((($this)) + 96|0); //@line 88 "./common/Object.h"
 HEAP32[$16>>2] = 0; //@line 88 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi(4093,($this|0)); //@line 74 "modules/graphics/opengl/Mesh.cpp"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 do {
  if ($18) {
   label = 8;
  } else {
   __THREW__ = 0;
   invoke_vi(4094,($this|0)); //@line 75 "modules/graphics/opengl/Mesh.cpp"
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   if ($20) {
    label = 8;
   } else {
    $21 = HEAP32[$8>>2]|0; //@line 77 "modules/graphics/opengl/Mesh.cpp"
    $22 = (($datasize>>>0) / ($21>>>0))&-1; //@line 77 "modules/graphics/opengl/Mesh.cpp"
    HEAP32[$7>>2] = $22; //@line 77 "modules/graphics/opengl/Mesh.cpp"
    $23 = ($22>>>0)>(65535); //@line 689 "modules/graphics/opengl/Mesh.cpp"
    $$$i = $23 ? 5125 : 5123; //@line 690 "modules/graphics/opengl/Mesh.cpp"
    HEAP32[$12>>2] = $$$i; //@line 78 "modules/graphics/opengl/Mesh.cpp"
    $24 = ($22|0)==(0); //@line 80 "modules/graphics/opengl/Mesh.cpp"
    if ($24) {
     $25 = (___cxa_allocate_exception(16)|0); //@line 81 "modules/graphics/opengl/Mesh.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($25|0),(436989|0),($vararg_buffer|0)); //@line 81 "modules/graphics/opengl/Mesh.cpp"
     $26 = __THREW__; __THREW__ = 0;
     $27 = $26&1;
     if ($27) {
      $33 = ___cxa_find_matching_catch_2()|0;
      $34 = tempRet0;
      ___cxa_free_exception(($25|0)); //@line 81 "modules/graphics/opengl/Mesh.cpp"
      $$0 = $34;$$01 = $33;
      break;
     } else {
      __THREW__ = 0;
      invoke_viii(3350,($25|0),(96|0),(344|0)); //@line 81 "modules/graphics/opengl/Mesh.cpp"
      $28 = __THREW__; __THREW__ = 0;
      label = 8;
      break;
     }
    }
    __THREW__ = 0;
    $35 = (invoke_ii(3348,40)|0); //@line 83 "modules/graphics/opengl/Mesh.cpp"
    $36 = __THREW__; __THREW__ = 0;
    $37 = $36&1;
    if ($37) {
     label = 8;
    } else {
     switch ($usage|0) {
     case 0:  {
      $$0$i = 35040;
      break;
     }
     case 1:  {
      $$0$i = 35048;
      break;
     }
     case 2:  {
      $$0$i = 35044;
      break;
     }
     default: {
      $$0$i = 0;
     }
     }
     __THREW__ = 0;
     invoke_viiiiii(4042,($35|0),($datasize|0),($data|0),34962,($$0$i|0),1); //@line 83 "modules/graphics/opengl/Mesh.cpp"
     $38 = __THREW__; __THREW__ = 0;
     $39 = $38&1;
     if ($39) {
      $45 = ___cxa_find_matching_catch_2()|0;
      $46 = tempRet0;
      __ZdlPv($35); //@line 83 "modules/graphics/opengl/Mesh.cpp"
      $$0 = $46;$$01 = $45;
      break;
     }
     HEAP32[$6>>2] = $35; //@line 83 "modules/graphics/opengl/Mesh.cpp"
     $40 = HEAP32[$8>>2]|0; //@line 85 "modules/graphics/opengl/Mesh.cpp"
     __THREW__ = 0;
     $41 = (invoke_ii(3492,($40|0))|0); //@line 85 "modules/graphics/opengl/Mesh.cpp"
     $42 = __THREW__; __THREW__ = 0;
     $43 = $42&1;
     if ($43) {
      label = 8;
     } else {
      $44 = ((($this)) + 64|0); //@line 85 "modules/graphics/opengl/Mesh.cpp"
      HEAP32[$44>>2] = $41; //@line 85 "modules/graphics/opengl/Mesh.cpp"
      STACKTOP = sp;return; //@line 86 "modules/graphics/opengl/Mesh.cpp"
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $31 = ___cxa_find_matching_catch_2()|0;
  $32 = tempRet0;
  $$0 = $32;$$01 = $31;
 }
 $47 = HEAP32[$16>>2]|0; //@line 106 "./common/Object.h"
 $48 = ($47|0)==(0|0); //@line 106 "./common/Object.h"
 if (!($48)) {
  $49 = HEAP32[$47>>2]|0; //@line 106 "./common/Object.h"
  $50 = ((($49)) + 12|0); //@line 106 "./common/Object.h"
  $51 = HEAP32[$50>>2]|0; //@line 106 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($51|0,($47|0)); //@line 106 "./common/Object.h"
  $52 = __THREW__; __THREW__ = 0;
  $53 = $52&1;
  if ($53) {
   $54 = ___cxa_find_matching_catch_3(0|0)|0;
   $55 = tempRet0;
   ___clang_call_terminate($54); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 }
 $56 = ((($this)) + 40|0); //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $57 = HEAP32[$56>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $58 = ($57|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($58)) {
  $$01$i$i$i = $57;
  while(1) {
   $59 = HEAP32[$$01$i$i$i>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $60 = ((($$01$i$i$i)) + 8|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($60); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
   __ZdlPv($$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $61 = ($59|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($61) {
    break;
   } else {
    $$01$i$i$i = $59;
   }
  }
 }
 $62 = ((($this)) + 32|0); //@line 2014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $63 = HEAP32[$62>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$62>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $64 = ($63|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($64)) {
  __ZdlPv($63); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $65 = HEAP32[$3>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $66 = ($65|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $67 = $65; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($66)) {
  $68 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $69 = ($68|0)==($65|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($69)) {
   $scevgep$i$i$i = ((($68)) + -4|0);
   $70 = $scevgep$i$i$i;
   $71 = (($70) - ($67))|0;
   $72 = $71 >>> 2;
   $73 = $72 ^ -1;
   $scevgep4$i$i$i = (($68) + ($73<<2)|0);
   HEAP32[$4>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($65); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $74 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $75 = ($74|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($75) {
  $$1 = $$0;$$12 = $$01;
  ___resumeException($$12|0);
  // unreachable;
 }
 $76 = ((($this)) + 12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $77 = HEAP32[$76>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $78 = ($77|0)==($74|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($78) {
  $83 = $74;
 } else {
  $80 = $77;
  while(1) {
   $79 = ((($80)) + -20|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$76>>2] = $79; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($79); //@line 81 "modules/graphics/opengl/Mesh.h"
   $81 = HEAP32[$76>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $82 = ($81|0)==($74|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($82) {
    break;
   } else {
    $80 = $81;
   }
  }
  $$pre$i = HEAP32[$0>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $83 = $$pre$i;
 }
 __ZdlPv($83); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $$1 = $$0;$$12 = $$01;
 ___resumeException($$12|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love8graphics6opengl4Mesh12AttribFormatENS_9allocatorIS5_EEEC2ERKS8_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$01$i = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $0 = ((($this)) + 4|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($this)) + 8|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($__x)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$__x>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = (($5|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ($3|0)==($4|0); //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($7) {
  return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $8 = ($6>>>0)>(214748364); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($8) {
  __THREW__ = 0;
  invoke_vi(3528,($this|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if ($10) {
   label = 10;
  } else {
   label = 4;
  }
 } else {
  label = 4;
 }
 do {
  if ((label|0) == 4) {
   __THREW__ = 0;
   $11 = (invoke_ii(3348,($5|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $12 = __THREW__; __THREW__ = 0;
   $13 = $12&1;
   if ($13) {
    label = 10;
   } else {
    HEAP32[$0>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$this>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $14 = (($11) + (($6*20)|0)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $15 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$15>>2] = $14; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $16 = HEAP32[$__x>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $17 = HEAP32[$2>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $18 = ($16|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($18) {
     return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    $$01$i = $16;$19 = $11;
    while(1) {
     __THREW__ = 0;
     invoke_vii(3351,($19|0),($$01$i|0)); //@line 81 "modules/graphics/opengl/Mesh.h"
     $20 = __THREW__; __THREW__ = 0;
     $21 = $20&1;
     if ($21) {
      label = 9;
      break;
     }
     $22 = ((($19)) + 12|0); //@line 81 "modules/graphics/opengl/Mesh.h"
     $23 = ((($$01$i)) + 12|0); //@line 81 "modules/graphics/opengl/Mesh.h"
     $24 = $23; //@line 81 "modules/graphics/opengl/Mesh.h"
     $25 = $24; //@line 81 "modules/graphics/opengl/Mesh.h"
     $26 = HEAP32[$25>>2]|0; //@line 81 "modules/graphics/opengl/Mesh.h"
     $27 = (($24) + 4)|0; //@line 81 "modules/graphics/opengl/Mesh.h"
     $28 = $27; //@line 81 "modules/graphics/opengl/Mesh.h"
     $29 = HEAP32[$28>>2]|0; //@line 81 "modules/graphics/opengl/Mesh.h"
     $30 = $22; //@line 81 "modules/graphics/opengl/Mesh.h"
     $31 = $30; //@line 81 "modules/graphics/opengl/Mesh.h"
     HEAP32[$31>>2] = $26; //@line 81 "modules/graphics/opengl/Mesh.h"
     $32 = (($30) + 4)|0; //@line 81 "modules/graphics/opengl/Mesh.h"
     $33 = $32; //@line 81 "modules/graphics/opengl/Mesh.h"
     HEAP32[$33>>2] = $29; //@line 81 "modules/graphics/opengl/Mesh.h"
     $34 = HEAP32[$0>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $35 = ((($34)) + 20|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$0>>2] = $35; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $36 = ((($$01$i)) + 20|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $37 = ($36|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($37) {
      label = 17;
      break;
     } else {
      $$01$i = $36;$19 = $35;
     }
    }
    if ((label|0) == 9) {
     $38 = ___cxa_find_matching_catch_2()|0;
     $39 = tempRet0;
     $lpad$phi$index = $38;$lpad$phi$index2 = $39;
     break;
    }
    else if ((label|0) == 17) {
     return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
   }
  }
 } while(0);
 if ((label|0) == 10) {
  $40 = ___cxa_find_matching_catch_2()|0;
  $41 = tempRet0;
  $lpad$phi$index = $40;$lpad$phi$index2 = $41;
 }
 $42 = HEAP32[$this>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $43 = ($42|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($43) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $44 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $45 = ($44|0)==($42|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($45) {
  $50 = $42;
 } else {
  $47 = $44;
  while(1) {
   $46 = ((($47)) + -20|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$0>>2] = $46; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($46); //@line 81 "modules/graphics/opengl/Mesh.h"
   $48 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $49 = ($48|0)==($42|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($49) {
    break;
   } else {
    $47 = $48;
   }
  }
  $$pre$i = HEAP32[$this>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $50 = $$pre$i;
 }
 __ZdlPv($50); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZN4love8graphics6opengl4Mesh23setupAttachedAttributesEv($this) {
 $this = $this|0;
 var $$in = 0, $$lcssa = 0, $$lcssa23 = 0, $$sroa$4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$04 = 0, $i$04$lcssa = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $$sroa$4 = sp + 4|0;
 $0 = ((($this)) + 8|0); //@line 145 "modules/graphics/opengl/Mesh.cpp"
 $1 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($2|0)==($3|0); //@line 145 "modules/graphics/opengl/Mesh.cpp"
 if ($4) {
  STACKTOP = sp;return; //@line 154 "modules/graphics/opengl/Mesh.cpp"
 }
 $5 = ((($this)) + 32|0); //@line 149 "modules/graphics/opengl/Mesh.cpp"
 $$in = $3;$i$04 = 0;
 while(1) {
  $6 = $$in; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $7 = (($6) + (($i$04*20)|0)|0); //@line 147 "modules/graphics/opengl/Mesh.cpp"
  $8 = (__ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl4Mesh17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE4findIS7_EENS_15__hash_iteratorIPNS_11__hash_nodeISD_PvEEEERKT_($5,$7)|0); //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
  $9 = ($8|0)==(0|0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if (!($9)) {
   $$lcssa = $6;$$lcssa23 = $7;$i$04$lcssa = $i$04;
   break;
  }
  $22 = (__ZNSt3__113unordered_mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl4Mesh17AttachedAttributeENS_4hashIS6_EENS_8equal_toIS6_EENS4_INS_4pairIKS6_SB_EEEEEixERSH_($5,$7)|0); //@line 152 "modules/graphics/opengl/Mesh.cpp"
  HEAP32[$22>>2] = $this; //@line 152 "modules/graphics/opengl/Mesh.cpp"
  $23 = ((($22)) + 4|0); //@line 152 "modules/graphics/opengl/Mesh.cpp"
  HEAP32[$23>>2] = $i$04; //@line 152 "modules/graphics/opengl/Mesh.cpp"
  $24 = ((($22)) + 8|0); //@line 152 "modules/graphics/opengl/Mesh.cpp"
  HEAP8[$24>>0] = 1; //@line 152 "modules/graphics/opengl/Mesh.cpp"
  $25 = ((($22)) + 9|0); //@line 152 "modules/graphics/opengl/Mesh.cpp"
  ;HEAP8[$25>>0]=HEAP8[$$sroa$4>>0]|0;HEAP8[$25+1>>0]=HEAP8[$$sroa$4+1>>0]|0;HEAP8[$25+2>>0]=HEAP8[$$sroa$4+2>>0]|0; //@line 152 "modules/graphics/opengl/Mesh.cpp"
  $26 = (($i$04) + 1)|0; //@line 145 "modules/graphics/opengl/Mesh.cpp"
  $27 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $28 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $29 = (($27) - ($28))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $30 = (($29|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $31 = ($26>>>0)<($30>>>0); //@line 145 "modules/graphics/opengl/Mesh.cpp"
  if ($31) {
   $$in = $28;$i$04 = $26;
  } else {
   label = 3;
   break;
  }
 }
 if ((label|0) == 3) {
  STACKTOP = sp;return; //@line 154 "modules/graphics/opengl/Mesh.cpp"
 }
 $10 = (___cxa_allocate_exception(16)|0); //@line 150 "modules/graphics/opengl/Mesh.cpp"
 $11 = HEAP8[$$lcssa23>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = $11 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ($12<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($13) {
  $16 = ((($$lcssa23)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $17 = $16;
 } else {
  $14 = (((($$lcssa) + (($i$04$lcssa*20)|0)|0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $15 = HEAP32[$14>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $17 = $15;
 }
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $17; //@line 150 "modules/graphics/opengl/Mesh.cpp"
 invoke_viii(3349,($10|0),(436850|0),($vararg_buffer|0)); //@line 150 "modules/graphics/opengl/Mesh.cpp"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $20 = ___cxa_find_matching_catch_2()|0;
  $21 = tempRet0;
  ___cxa_free_exception(($10|0)); //@line 150 "modules/graphics/opengl/Mesh.cpp"
  ___resumeException($20|0);
  // unreachable;
 } else {
  ___cxa_throw(($10|0),(96|0),(344|0)); //@line 150 "modules/graphics/opengl/Mesh.cpp"
  // unreachable; //@line 150 "modules/graphics/opengl/Mesh.cpp"
 }
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl4Mesh17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE4findIS7_EENS_15__hash_iteratorIPNS_11__hash_nodeISD_PvEEEERKT_($this,$__k) {
 $this = $this|0;
 $__k = $__k|0;
 var $$0$lcssa$i$i$i$i = 0, $$03$i$i$i$i = 0, $$lcssa = 0, $$mask$i$i = 0, $$mask$i$i$us = 0, $$pre$i$i = 0, $$pre$i$i$us = 0, $$sroa$0$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__data$0$lcssa$i$i$i$i = 0, $__data$02$i$i$i$i = 0, $__h$0$lcssa$i$i$i$i = 0, $__h$01$i$i$i$i = 0, $__h$1$i$i$i$i = 0, $__h$2$i$i$i$i = 0, $__h$3$i$i$i$i = 0, $__lhs_sz$03$i$i3$i = 0, $__lhs_sz$03$i$i3$i$us = 0, $__lp$04$i$i2$i = 0, $__lp$04$i$i2$i$us = 0;
 var $__nd$0 = 0, $__nd$0$us = 0, $__nd$08 = 0, $__nd$09 = 0, $__nd$09$us = 0, $__rp$05$i$i1$i = 0, $__rp$05$i$i1$i$us = 0, $scevgep$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$__k>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $1 = $0 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = ($1<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ((($__k)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = HEAP32[$3>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($__k)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = $2 ? $5 : $4; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($__k)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = HEAP32[$7>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $0&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $9 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $2 ? $10 : $8; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = ($11>>>0)>(3); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($12) {
  $$03$i$i$i$i = $11;$__data$02$i$i$i$i = $6;$__h$01$i$i$i$i = $11;
  while(1) {
   $13 = HEAPU8[$__data$02$i$i$i$i>>0]|(HEAPU8[$__data$02$i$i$i$i+1>>0]<<8)|(HEAPU8[$__data$02$i$i$i$i+2>>0]<<16)|(HEAPU8[$__data$02$i$i$i$i+3>>0]<<24); //@line 3073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $14 = Math_imul($13, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $15 = $14 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $16 = $15 ^ $14; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $17 = Math_imul($16, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $18 = Math_imul($__h$01$i$i$i$i, 1540483477)|0; //@line 3104 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $19 = $17 ^ $18; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $20 = ((($__data$02$i$i$i$i)) + 4|0); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $21 = (($$03$i$i$i$i) + -4)|0; //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $22 = ($21>>>0)>(3); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if ($22) {
    $$03$i$i$i$i = $21;$__data$02$i$i$i$i = $20;$__h$01$i$i$i$i = $19;
   } else {
    $$lcssa = $19;
    break;
   }
  }
  $23 = (($11) + -4)|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = $23 & -4; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $25 = (($24) + 4)|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (($23) - ($24))|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $scevgep$i$i$i$i = (($6) + ($25)|0); //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $$0$lcssa$i$i$i$i = $26;$__data$0$lcssa$i$i$i$i = $scevgep$i$i$i$i;$__h$0$lcssa$i$i$i$i = $$lcssa;
 } else {
  $$0$lcssa$i$i$i$i = $11;$__data$0$lcssa$i$i$i$i = $6;$__h$0$lcssa$i$i$i$i = $11;
 }
 switch ($$0$lcssa$i$i$i$i|0) {
 case 3:  {
  $27 = ((($__data$0$lcssa$i$i$i$i)) + 2|0); //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $28 = HEAP8[$27>>0]|0; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $29 = $28&255; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $30 = $29 << 16; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $31 = $30 ^ $__h$0$lcssa$i$i$i$i; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$1$i$i$i$i = $31;
  label = 6;
  break;
 }
 case 2:  {
  $__h$1$i$i$i$i = $__h$0$lcssa$i$i$i$i;
  label = 6;
  break;
 }
 case 1:  {
  $__h$2$i$i$i$i = $__h$0$lcssa$i$i$i$i;
  label = 7;
  break;
 }
 default: {
  $__h$3$i$i$i$i = $__h$0$lcssa$i$i$i$i;
 }
 }
 if ((label|0) == 6) {
  $32 = ((($__data$0$lcssa$i$i$i$i)) + 1|0); //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $33 = HEAP8[$32>>0]|0; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $34 = $33&255; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $35 = $34 << 8; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $36 = $35 ^ $__h$1$i$i$i$i; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$2$i$i$i$i = $36;
  label = 7;
 }
 if ((label|0) == 7) {
  $37 = HEAP8[$__data$0$lcssa$i$i$i$i>>0]|0; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $38 = $37&255; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $39 = $38 ^ $__h$2$i$i$i$i; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $40 = Math_imul($39, 1540483477)|0; //@line 3115 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$3$i$i$i$i = $40;
 }
 $41 = $__h$3$i$i$i$i >>> 13; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $42 = $41 ^ $__h$3$i$i$i$i; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $43 = Math_imul($42, 1540483477)|0; //@line 3118 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $44 = $43 >>> 15; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $45 = $44 ^ $43; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $46 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = HEAP32[$46>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $48 = ($47|0)==(0); //@line 2018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($48) {
  $$sroa$0$0 = 0;
  return ($$sroa$0$0|0); //@line 2038 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $49 = (($47) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $50 = $49 & $47; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $51 = ($50|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($51) {
  $52 = $45 & $49; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $56 = $52;
 } else {
  $53 = (($45>>>0) % ($47>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $56 = $53;
 }
 $54 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $55 = (($54) + ($56<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $57 = HEAP32[$55>>2]|0; //@line 2021 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $58 = ($57|0)==(0|0); //@line 2022 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($58) {
  $$sroa$0$0 = 0;
  return ($$sroa$0$0|0); //@line 2038 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $__nd$08 = HEAP32[$57>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $59 = ($__nd$08|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($59) {
  $$sroa$0$0 = 0;
  return ($$sroa$0$0|0); //@line 2038 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $60 = ($11|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($51) {
  $__nd$09$us = $__nd$08;
  L27: while(1) {
   $61 = ((($__nd$09$us)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $62 = HEAP32[$61>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $63 = $62 & $49; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $64 = ($63|0)==($56|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($64)) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   }
   $65 = ((($__nd$09$us)) + 8|0); //@line 2028 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $66 = HEAP16[$65>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $67 = $66 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $68 = ($67<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $69 = ((($__nd$09$us)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $70 = HEAP32[$69>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$mask$i$i$us = $66 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $71 = $$mask$i$i$us&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $72 = $71 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $73 = $68 ? $72 : $70; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $74 = ($73|0)==($11|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $75 = ($66&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $76 = $75&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   do {
    if ($74) {
     if (!($68)) {
      $77 = ((($__nd$09$us)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $78 = HEAP32[$77>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $79 = (_memcmp($78,$6,$11)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $80 = ($79|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($80) {
       $$sroa$0$0 = $__nd$09$us;
       label = 35;
       break L27;
      } else {
       break;
      }
     }
     if ($60) {
      $$sroa$0$0 = $__nd$09$us;
      label = 35;
      break L27;
     }
     $81 = HEAP8[$6>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $82 = ($76<<24>>24)==($81<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($82) {
      $83 = ((($65)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $__lhs_sz$03$i$i3$i$us = $72;$__lp$04$i$i2$i$us = $83;$__rp$05$i$i1$i$us = $6;
      while(1) {
       $84 = (($__lhs_sz$03$i$i3$i$us) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $85 = ((($__lp$04$i$i2$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $86 = ($84|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($86) {
        $$sroa$0$0 = $__nd$09$us;
        label = 35;
        break L27;
       }
       $87 = ((($__rp$05$i$i1$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$pre$i$i$us = HEAP8[$85>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $88 = HEAP8[$87>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $89 = ($$pre$i$i$us<<24>>24)==($88<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($89) {
        $__lhs_sz$03$i$i3$i$us = $84;$__lp$04$i$i2$i$us = $85;$__rp$05$i$i1$i$us = $87;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   $__nd$0$us = HEAP32[$__nd$09$us>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $90 = ($__nd$0$us|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($90) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   } else {
    $__nd$09$us = $__nd$0$us;
   }
  }
  if ((label|0) == 35) {
   return ($$sroa$0$0|0); //@line 2038 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 } else {
  $__nd$09 = $__nd$08;
  L47: while(1) {
   $91 = ((($__nd$09)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $92 = HEAP32[$91>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $93 = (($92>>>0) % ($47>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $94 = ($93|0)==($56|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($94)) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   }
   $95 = ((($__nd$09)) + 8|0); //@line 2028 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $96 = HEAP16[$95>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $97 = $96 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $98 = ($97<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $99 = ((($__nd$09)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $100 = HEAP32[$99>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$mask$i$i = $96 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $101 = $$mask$i$i&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $102 = $101 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $103 = $98 ? $102 : $100; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $104 = ($103|0)==($11|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $105 = ($96&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $106 = $105&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   do {
    if ($104) {
     if (!($98)) {
      $116 = ((($__nd$09)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $117 = HEAP32[$116>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $118 = (_memcmp($117,$6,$11)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $119 = ($118|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($119) {
       $$sroa$0$0 = $__nd$09;
       label = 35;
       break L47;
      } else {
       break;
      }
     }
     if ($60) {
      $$sroa$0$0 = $__nd$09;
      label = 35;
      break L47;
     }
     $107 = HEAP8[$6>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $108 = ($106<<24>>24)==($107<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($108) {
      $109 = ((($95)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $__lhs_sz$03$i$i3$i = $102;$__lp$04$i$i2$i = $109;$__rp$05$i$i1$i = $6;
      while(1) {
       $110 = (($__lhs_sz$03$i$i3$i) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $111 = ((($__lp$04$i$i2$i)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $112 = ($110|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($112) {
        $$sroa$0$0 = $__nd$09;
        label = 35;
        break L47;
       }
       $113 = ((($__rp$05$i$i1$i)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$pre$i$i = HEAP8[$111>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $114 = HEAP8[$113>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $115 = ($$pre$i$i<<24>>24)==($114<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($115) {
        $__lhs_sz$03$i$i3$i = $110;$__lp$04$i$i2$i = $111;$__rp$05$i$i1$i = $113;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   $__nd$0 = HEAP32[$__nd$09>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $120 = ($__nd$0|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($120) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   } else {
    $__nd$09 = $__nd$0;
   }
  }
  if ((label|0) == 35) {
   return ($$sroa$0$0|0); //@line 2038 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 }
 return (0)|0;
}
function __ZNSt3__113unordered_mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl4Mesh17AttachedAttributeENS_4hashIS6_EENS_8equal_toIS6_EENS4_INS_4pairIKS6_SB_EEEEEixERSH_($this,$__k) {
 $this = $this|0;
 $__k = $__k|0;
 var $$0 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl4Mesh17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE4findIS7_EENS_15__hash_iteratorIPNS_11__hash_nodeISD_PvEEEERKT_($this,$__k)|0); //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $2 = ($1|0)==(0|0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($2)) {
  $$sink = $1;
  $$0 = ((($$sink)) + 20|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
  STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 }
 $3 = (__Znwj(32)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $4 = ((($3)) + 8|0); //@line 1340 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 __THREW__ = 0;
 invoke_vii(3351,($4|0),($__k|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZdlPv($3); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($7|0);
  // unreachable;
 }
 $9 = ((($3)) + 20|0); //@line 1342 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 invoke_viii(4095,($0|0),($this|0),($3|0)); //@line 1366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 1589 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  __ZdlPv($3); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($13|0);
  // unreachable;
 }
 $12 = HEAP32[$0>>2]|0; //@line 315 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $$sink = $12;
 $$0 = ((($$sink)) + 20|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl4Mesh17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE20__node_insert_uniqueEPNS_11__hash_nodeISD_PvEE($agg$result,$this,$__nd) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $__nd = $__nd|0;
 var $$0$lcssa$i$i$i$i = 0, $$03$i$i$i$i = 0, $$lcssa = 0, $$mask$i$i = 0, $$mask$i$i$us = 0, $$pre$i$i = 0, $$pre$i$i$us = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0.0, $127 = 0.0;
 var $128 = 0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0.0, $142 = 0.0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__bc$0 = 0, $__chash$0 = 0, $__chash$1 = 0, $__data$0$lcssa$i$i$i$i = 0, $__data$02$i$i$i$i = 0, $__h$0$lcssa$i$i$i$i = 0, $__h$01$i$i$i$i = 0, $__h$1$i$i$i$i = 0, $__h$2$i$i$i$i = 0, $__h$3$i$i$i$i = 0, $__inserted$0 = 0, $__lhs_sz$03$i$i3$i = 0, $__lhs_sz$03$i$i3$i$us = 0, $__lp$04$i$i2$i = 0, $__lp$04$i$i2$i$us = 0;
 var $__ndptr$0 = 0, $__ndptr$0$us = 0, $__ndptr$011 = 0, $__ndptr$012 = 0, $__ndptr$012$us = 0, $__ndptr$1 = 0, $__rp$05$i$i1$i = 0, $__rp$05$i$i1$i$us = 0, $or$cond = 0, $scevgep$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($__nd)) + 8|0); //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $1 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = $1 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ($2<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ((($__nd)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = HEAP32[$4>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = $3 ? $6 : $5; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = ((($__nd)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = HEAP32[$8>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $1&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $10 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = $3 ? $11 : $9; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ($12>>>0)>(3); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($13) {
  $$03$i$i$i$i = $12;$__data$02$i$i$i$i = $7;$__h$01$i$i$i$i = $12;
  while(1) {
   $14 = HEAPU8[$__data$02$i$i$i$i>>0]|(HEAPU8[$__data$02$i$i$i$i+1>>0]<<8)|(HEAPU8[$__data$02$i$i$i$i+2>>0]<<16)|(HEAPU8[$__data$02$i$i$i$i+3>>0]<<24); //@line 3073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $15 = Math_imul($14, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $16 = $15 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $17 = $16 ^ $15; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $18 = Math_imul($17, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $19 = Math_imul($__h$01$i$i$i$i, 1540483477)|0; //@line 3104 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $20 = $18 ^ $19; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $21 = ((($__data$02$i$i$i$i)) + 4|0); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $22 = (($$03$i$i$i$i) + -4)|0; //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $23 = ($22>>>0)>(3); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if ($23) {
    $$03$i$i$i$i = $22;$__data$02$i$i$i$i = $21;$__h$01$i$i$i$i = $20;
   } else {
    $$lcssa = $20;
    break;
   }
  }
  $24 = (($12) + -4)|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $25 = $24 & -4; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (($25) + 4)|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $27 = (($24) - ($25))|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $scevgep$i$i$i$i = (($7) + ($26)|0); //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $$0$lcssa$i$i$i$i = $27;$__data$0$lcssa$i$i$i$i = $scevgep$i$i$i$i;$__h$0$lcssa$i$i$i$i = $$lcssa;
 } else {
  $$0$lcssa$i$i$i$i = $12;$__data$0$lcssa$i$i$i$i = $7;$__h$0$lcssa$i$i$i$i = $12;
 }
 switch ($$0$lcssa$i$i$i$i|0) {
 case 3:  {
  $28 = ((($__data$0$lcssa$i$i$i$i)) + 2|0); //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $29 = HEAP8[$28>>0]|0; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $30 = $29&255; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $31 = $30 << 16; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $32 = $31 ^ $__h$0$lcssa$i$i$i$i; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$1$i$i$i$i = $32;
  label = 6;
  break;
 }
 case 2:  {
  $__h$1$i$i$i$i = $__h$0$lcssa$i$i$i$i;
  label = 6;
  break;
 }
 case 1:  {
  $__h$2$i$i$i$i = $__h$0$lcssa$i$i$i$i;
  label = 7;
  break;
 }
 default: {
  $__h$3$i$i$i$i = $__h$0$lcssa$i$i$i$i;
 }
 }
 if ((label|0) == 6) {
  $33 = ((($__data$0$lcssa$i$i$i$i)) + 1|0); //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $34 = HEAP8[$33>>0]|0; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $35 = $34&255; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $36 = $35 << 8; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $37 = $36 ^ $__h$1$i$i$i$i; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$2$i$i$i$i = $37;
  label = 7;
 }
 if ((label|0) == 7) {
  $38 = HEAP8[$__data$0$lcssa$i$i$i$i>>0]|0; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $39 = $38&255; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $40 = $39 ^ $__h$2$i$i$i$i; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $41 = Math_imul($40, 1540483477)|0; //@line 3115 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$3$i$i$i$i = $41;
 }
 $42 = $__h$3$i$i$i$i >>> 13; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $43 = $42 ^ $__h$3$i$i$i$i; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $44 = Math_imul($43, 1540483477)|0; //@line 3118 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $45 = $44 >>> 15; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $46 = $45 ^ $44; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = ((($__nd)) + 4|0); //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$47>>2] = $46; //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $48 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $49 = HEAP32[$48>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $50 = ($49|0)==(0); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L13: do {
  if ($50) {
   $__chash$0 = 0;
  } else {
   $51 = (($49) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $52 = $51 & $49; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $53 = ($52|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($53) {
    $54 = $46 & $51; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $58 = $54;
   } else {
    $55 = (($46>>>0) % ($49>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $58 = $55;
   }
   $56 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $57 = (($56) + ($58<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $59 = HEAP32[$57>>2]|0; //@line 1603 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $60 = ($59|0)==(0|0); //@line 1604 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($60) {
    $__chash$0 = $58;
   } else {
    $__ndptr$011 = HEAP32[$59>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $61 = ($__ndptr$011|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($61) {
     $__chash$0 = $58;
    } else {
     $62 = ($12|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($53) {
      $__ndptr$012$us = $__ndptr$011;
      L22: while(1) {
       $63 = ((($__ndptr$012$us)) + 4|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $64 = HEAP32[$63>>2]|0; //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $65 = $64 & $51; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $66 = ($65|0)==($58|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if (!($66)) {
        $__chash$0 = $58;
        break L13;
       }
       $67 = ((($__ndptr$012$us)) + 8|0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $68 = HEAP16[$67>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $69 = $68 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $70 = ($69<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $71 = ((($__ndptr$012$us)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $72 = HEAP32[$71>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$mask$i$i$us = $68 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $73 = $$mask$i$i$us&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $74 = $73 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $75 = $70 ? $74 : $72; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $76 = ($75|0)==($12|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $77 = ($68&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $78 = $77&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       do {
        if ($76) {
         if (!($70)) {
          $79 = ((($__ndptr$012$us)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $80 = HEAP32[$79>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $81 = (_memcmp($80,$7,$12)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $82 = ($81|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          if ($82) {
           $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012$us;
           label = 49;
           break L22;
          } else {
           break;
          }
         }
         if ($62) {
          $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012$us;
          label = 49;
          break L22;
         }
         $83 = HEAP8[$7>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $84 = ($78<<24>>24)==($83<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         if ($84) {
          $85 = ((($67)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $__lhs_sz$03$i$i3$i$us = $74;$__lp$04$i$i2$i$us = $85;$__rp$05$i$i1$i$us = $7;
          while(1) {
           $86 = (($__lhs_sz$03$i$i3$i$us) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $87 = ((($__lp$04$i$i2$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $88 = ($86|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($88) {
            $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012$us;
            label = 49;
            break L22;
           }
           $89 = ((($__rp$05$i$i1$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $$pre$i$i$us = HEAP8[$87>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $90 = HEAP8[$89>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $91 = ($$pre$i$i$us<<24>>24)==($90<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($91) {
            $__lhs_sz$03$i$i3$i$us = $86;$__lp$04$i$i2$i$us = $87;$__rp$05$i$i1$i$us = $89;
           } else {
            break;
           }
          }
         }
        }
       } while(0);
       $__ndptr$0$us = HEAP32[$__ndptr$012$us>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $92 = ($__ndptr$0$us|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($92) {
        $__chash$0 = $58;
        break L13;
       } else {
        $__ndptr$012$us = $__ndptr$0$us;
       }
      }
      if ((label|0) == 49) {
       $176 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       HEAP32[$agg$result>>2] = $176; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       $177 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       HEAP8[$177>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      }
     } else {
      $__ndptr$012 = $__ndptr$011;
      L40: while(1) {
       $93 = ((($__ndptr$012)) + 4|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $94 = HEAP32[$93>>2]|0; //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $95 = (($94>>>0) % ($49>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $96 = ($95|0)==($58|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if (!($96)) {
        $__chash$0 = $58;
        break L13;
       }
       $97 = ((($__ndptr$012)) + 8|0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $98 = HEAP16[$97>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $99 = $98 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $100 = ($99<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $101 = ((($__ndptr$012)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $102 = HEAP32[$101>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$mask$i$i = $98 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $103 = $$mask$i$i&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $104 = $103 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $105 = $100 ? $104 : $102; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $106 = ($105|0)==($12|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $107 = ($98&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $108 = $107&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       do {
        if ($106) {
         if (!($100)) {
          $118 = ((($__ndptr$012)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $119 = HEAP32[$118>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $120 = (_memcmp($119,$7,$12)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $121 = ($120|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          if ($121) {
           $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012;
           label = 49;
           break L40;
          } else {
           break;
          }
         }
         if ($62) {
          $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012;
          label = 49;
          break L40;
         }
         $109 = HEAP8[$7>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $110 = ($108<<24>>24)==($109<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         if ($110) {
          $111 = ((($97)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $__lhs_sz$03$i$i3$i = $104;$__lp$04$i$i2$i = $111;$__rp$05$i$i1$i = $7;
          while(1) {
           $112 = (($__lhs_sz$03$i$i3$i) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $113 = ((($__lp$04$i$i2$i)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $114 = ($112|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($114) {
            $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012;
            label = 49;
            break L40;
           }
           $115 = ((($__rp$05$i$i1$i)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $$pre$i$i = HEAP8[$113>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $116 = HEAP8[$115>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $117 = ($$pre$i$i<<24>>24)==($116<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($117) {
            $__lhs_sz$03$i$i3$i = $112;$__lp$04$i$i2$i = $113;$__rp$05$i$i1$i = $115;
           } else {
            break;
           }
          }
         }
        }
       } while(0);
       $__ndptr$0 = HEAP32[$__ndptr$012>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $122 = ($__ndptr$0|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($122) {
        $__chash$0 = $58;
        break L13;
       } else {
        $__ndptr$012 = $__ndptr$0;
       }
      }
      if ((label|0) == 49) {
       $176 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       HEAP32[$agg$result>>2] = $176; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       $177 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       HEAP8[$177>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      }
     }
    }
   }
  }
 } while(0);
 $123 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $124 = HEAP32[$123>>2]|0; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $125 = (($124) + 1)|0; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $126 = (+($125>>>0)); //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $127 = (+($49>>>0)); //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $128 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $129 = +HEAPF32[$128>>2]; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $130 = $127 * $129; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $131 = $126 > $130; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $or$cond = $50 | $131; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 do {
  if ($or$cond) {
   $132 = $49 << 1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $133 = ($49>>>0)>(2); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($133) {
    $134 = (($49) + -1)|0; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $135 = $134 & $49; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $136 = ($135|0)==(0); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $138 = $136;
   } else {
    $138 = 0;
   }
   $137 = $138&1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $139 = $137 | $132; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $140 = $139 ^ 1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $141 = $126 / $129; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $142 = (+Math_ceil((+$141))); //@line 750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/cmath"
   $143 = (~~(($142))>>>0); //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $144 = ($140>>>0)<($143>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $145 = $144 ? $143 : $140; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl4Mesh17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE6rehashEj($this,$145); //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $146 = HEAP32[$48>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $147 = HEAP32[$47>>2]|0; //@line 1621 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $148 = (($146) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $149 = $148 & $146; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $150 = ($149|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($150) {
    $151 = $148 & $147; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__bc$0 = $146;$__chash$1 = $151;
    break;
   } else {
    $152 = (($147>>>0) % ($146>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__bc$0 = $146;$__chash$1 = $152;
    break;
   }
  } else {
   $__bc$0 = $49;$__chash$1 = $__chash$0;
  }
 } while(0);
 $153 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $154 = (($153) + ($__chash$1<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $155 = HEAP32[$154>>2]|0; //@line 1624 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $156 = ($155|0)==(0|0); //@line 1625 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($156) {
  $157 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $158 = HEAP32[$157>>2]|0; //@line 1628 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$__nd>>2] = $158; //@line 1628 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$157>>2] = $__nd; //@line 1629 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $159 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $160 = (($159) + ($__chash$1<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$160>>2] = $157; //@line 1631 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $161 = HEAP32[$__nd>>2]|0; //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $162 = ($161|0)==(0|0); //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if (!($162)) {
   $163 = ((($161)) + 4|0); //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $164 = HEAP32[$163>>2]|0; //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $165 = (($__bc$0) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $166 = $165 & $__bc$0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $167 = ($166|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($167) {
    $168 = $164 & $165; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $172 = $168;
   } else {
    $169 = (($164>>>0) % ($__bc$0>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $172 = $169;
   }
   $170 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $171 = (($170) + ($172<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$171>>2] = $__nd; //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 } else {
  $173 = HEAP32[$155>>2]|0; //@line 1637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$__nd>>2] = $173; //@line 1637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$155>>2] = $__nd; //@line 1638 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $174 = HEAP32[$123>>2]|0; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $175 = (($174) + 1)|0; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$123>>2] = $175; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__inserted$0 = 1;$__ndptr$1 = $__nd;
 $176 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$agg$result>>2] = $176; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $177 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 HEAP8[$177>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl4Mesh17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE6rehashEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__n|0)==(1); //@line 1937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($0) {
  $$0 = 2;
 } else {
  $1 = (($__n) + -1)|0; //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $2 = $1 & $__n; //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $3 = ($2|0)==(0); //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($3) {
   $$0 = $__n;
  } else {
   $4 = (__ZNSt3__112__next_primeEj($__n)|0); //@line 1940 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $$0 = $4;
  }
 }
 $5 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = HEAP32[$5>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $7 = ($$0>>>0)>($6>>>0); //@line 1942 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($7) {
  __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl4Mesh17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE8__rehashEj($this,$$0); //@line 1943 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $8 = ($$0>>>0)<($6>>>0); //@line 1944 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($8)) {
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $9 = ($6>>>0)>(2); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($9) {
  $10 = (($6) + -1)|0; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $11 = $10 & $6; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $12 = ($11|0)==(0); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $30 = $12;
 } else {
  $30 = 0;
 }
 $13 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = HEAP32[$13>>2]|0; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $15 = (+($14>>>0)); //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $16 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $17 = +HEAPF32[$16>>2]; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $18 = $15 / $17; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $19 = (+Math_ceil((+$18))); //@line 750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/cmath"
 $20 = (~~(($19))>>>0); //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($30) {
  $21 = (($20) + -1)|0; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $22 = (Math_clz32(($21|0))|0); //@line 785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $23 = (32 - ($22))|0; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $24 = 1 << $23; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $24;
 } else {
  $25 = (__ZNSt3__112__next_primeEj($20)|0); //@line 1950 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $25;
 }
 $26 = ($$0>>>0)<($27>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $28 = $26 ? $27 : $$0; //@line 1946 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $29 = ($28>>>0)<($6>>>0); //@line 1952 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($29)) {
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl4Mesh17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE8__rehashEj($this,$28); //@line 1953 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl4Mesh17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE8__rehashEj($this,$__nbc) {
 $this = $this|0;
 $__nbc = $__nbc|0;
 var $$lcssa = 0, $$lcssa113 = 0, $$mask$i$i = 0, $$pre$i$i$us = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__cp$0 = 0, $__cp$0$in2332 = 0, $__cp$0$in26 = 0, $__cp$018 = 0, $__cp$01824 = 0, $__cp$0182429 = 0, $__cp$0182433 = 0, $__cp$01827 = 0, $__cp$019 = 0, $__cp$019$lcssa = 0, $__cp$019$lcssa111 = 0, $__i$034 = 0, $__lhs_sz$03$i$i3$i$us = 0, $__lp$04$i$i2$i$us = 0, $__np$013 = 0, $__np$013$phi = 0;
 var $__np$013$us = 0, $__np$013$us$phi = 0, $__np$013$us36 = 0, $__np$013$us36$phi = 0, $__np$08 = 0, $__phash$0$ph$ph31 = 0, $__pp$0$ph$ph30 = 0, $__pp$0$ph25 = 0, $__pp$0$ph25$lcssa = 0, $__rp$05$i$i1$i$us = 0, $exitcond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = ($__nbc|0)==(0); //@line 1965 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($1) {
  $118 = HEAP32[$this>>2]|0; //@line 2822 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$this>>2] = 0; //@line 2823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $119 = ($118|0)==(0|0); //@line 2824 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  if (!($119)) {
   __ZdlPv($118); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  HEAP32[$0>>2] = 0; //@line 1967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $2 = $__nbc << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = (__Znwj($2)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $4 = HEAP32[$this>>2]|0; //@line 2822 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $3; //@line 2823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ($4|0)==(0|0); //@line 2824 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($5)) {
  __ZdlPv($4); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 HEAP32[$0>>2] = $__nbc; //@line 1967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__i$034 = 0;
 while(1) {
  $9 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $10 = (($9) + ($__i$034<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$10>>2] = 0; //@line 1971 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $11 = (($__i$034) + 1)|0; //@line 1970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $exitcond = ($11|0)==($__nbc|0); //@line 1970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($exitcond) {
   break;
  } else {
   $__i$034 = $11;
  }
 }
 $6 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $7 = HEAP32[$6>>2]|0; //@line 1973 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $8 = ($7|0)==(0|0); //@line 1974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($8) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $12 = ((($7)) + 4|0); //@line 1976 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $13 = HEAP32[$12>>2]|0; //@line 1976 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $14 = (($__nbc) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $15 = $14 & $__nbc; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $16 = ($15|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($16) {
  $17 = $13 & $14; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $21 = $17;
 } else {
  $18 = (($13>>>0) % ($__nbc>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $21 = $18;
 }
 $19 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $20 = (($19) + ($21<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$20>>2] = $6; //@line 1977 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__cp$0182429 = HEAP32[$7>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $22 = ($__cp$0182429|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($22) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 } else {
  $__cp$0$in2332 = $7;$__cp$0182433 = $__cp$0182429;$__phash$0$ph$ph31 = $21;$__pp$0$ph$ph30 = $7;
 }
 L23: while(1) {
  $__cp$0$in26 = $__cp$0$in2332;$__cp$01827 = $__cp$0182433;$__pp$0$ph25 = $__pp$0$ph$ph30;
  L25: while(1) {
   $__cp$019 = $__cp$01827;
   while(1) {
    $24 = ((($__cp$019)) + 4|0); //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $25 = HEAP32[$24>>2]|0; //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($16) {
     $26 = $25 & $14; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $29 = $26;
    } else {
     $27 = (($25>>>0) % ($__nbc>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $29 = $27;
    }
    $28 = ($29|0)==($__phash$0$ph$ph31|0); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($28) {
     $__cp$019$lcssa = $__cp$019;
     break;
    }
    $30 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $31 = (($30) + ($29<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $32 = HEAP32[$31>>2]|0; //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $33 = ($32|0)==(0|0); //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($33) {
     $$lcssa = $31;$$lcssa113 = $29;$__cp$019$lcssa111 = $__cp$019;$__pp$0$ph25$lcssa = $__pp$0$ph25;
     break L25;
    }
    $34 = HEAP32[$__cp$019>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $35 = ($34|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    L35: do {
     if ($35) {
      $__np$08 = $__cp$019;
     } else {
      $36 = ((($__cp$019)) + 8|0); //@line 1997 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $37 = HEAP16[$36>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $38 = $37 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $39 = ($38<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $40 = ((($__cp$019)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $41 = HEAP32[$40>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $$mask$i$i = $37 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $42 = $$mask$i$i&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $43 = $42 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $44 = $39 ? $43 : $41; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $45 = ($37&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $46 = $45&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $47 = ($44|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $48 = ((($36)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $49 = ((($__cp$019)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($39) {
       $51 = $34;$__np$013$us = $__cp$019;
       while(1) {
        $50 = ((($51)) + 8|0); //@line 1997 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $52 = HEAP8[$50>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $53 = $52 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $54 = ($53<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $55 = ((($51)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $56 = HEAP32[$55>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $57 = $52&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $58 = $57 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $59 = $54 ? $58 : $56; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $60 = ($44|0)==($59|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if (!($60)) {
         $__np$08 = $__np$013$us;
         break L35;
        }
        $61 = ((($51)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $62 = HEAP32[$61>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $63 = ((($50)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $64 = $54 ? $63 : $62; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        L41: do {
         if (!($47)) {
          $65 = HEAP8[$64>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $66 = ($46<<24>>24)==($65<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          if ($66) {
           $__lhs_sz$03$i$i3$i$us = $43;$__lp$04$i$i2$i$us = $48;$__rp$05$i$i1$i$us = $64;
          } else {
           $__np$08 = $__np$013$us;
           break L35;
          }
          while(1) {
           $67 = (($__lhs_sz$03$i$i3$i$us) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $68 = ((($__lp$04$i$i2$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $69 = ($67|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($69) {
            break L41;
           }
           $70 = ((($__rp$05$i$i1$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $$pre$i$i$us = HEAP8[$68>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $71 = HEAP8[$70>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $72 = ($$pre$i$i$us<<24>>24)==($71<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($72) {
            $__lhs_sz$03$i$i3$i$us = $67;$__lp$04$i$i2$i$us = $68;$__rp$05$i$i1$i$us = $70;
           } else {
            $__np$08 = $__np$013$us;
            break L35;
           }
          }
         }
        } while(0);
        $73 = HEAP32[$51>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $74 = ($73|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        if ($74) {
         $__np$08 = $51;
         break;
        } else {
         $__np$013$us$phi = $51;$51 = $73;$__np$013$us = $__np$013$us$phi;
        }
       }
      } else {
       if ($47) {
        $76 = $34;$__np$013$us36 = $__cp$019;
        while(1) {
         $75 = ((($76)) + 8|0); //@line 1997 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $77 = HEAP8[$75>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $78 = $77 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $79 = ($78<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $80 = ((($76)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $81 = HEAP32[$80>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $82 = $77&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $83 = $82 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $84 = $79 ? $83 : $81; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $85 = ($84|0)==(0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         if (!($85)) {
          $__np$08 = $__np$013$us36;
          break L35;
         }
         $86 = HEAP32[$76>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $87 = ($86|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         if ($87) {
          $__np$08 = $76;
          break L35;
         } else {
          $__np$013$us36$phi = $76;$76 = $86;$__np$013$us36 = $__np$013$us36$phi;
         }
        }
       } else {
        $90 = $34;$__np$013 = $__cp$019;
       }
       while(1) {
        $89 = ((($90)) + 8|0); //@line 1997 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $91 = HEAP8[$89>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $92 = $91 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $93 = ($92<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $94 = ((($90)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $95 = HEAP32[$94>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $96 = $91&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $97 = $96 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $98 = $93 ? $97 : $95; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $99 = ($44|0)==($98|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if (!($99)) {
         $__np$08 = $__np$013;
         break L35;
        }
        $100 = ((($90)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $101 = HEAP32[$100>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $102 = ((($89)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $103 = $93 ? $102 : $101; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $104 = HEAP32[$49>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $105 = (_memcmp($104,$103,$44)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $106 = ($105|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if (!($106)) {
         $__np$08 = $__np$013;
         break L35;
        }
        $107 = HEAP32[$90>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $108 = ($107|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        if ($108) {
         $__np$08 = $90;
         break;
        } else {
         $__np$013$phi = $90;$90 = $107;$__np$013 = $__np$013$phi;
        }
       }
      }
     }
    } while(0);
    $109 = HEAP32[$__np$08>>2]|0; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    HEAP32[$__pp$0$ph25>>2] = $109; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $110 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $111 = (($110) + ($29<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $112 = HEAP32[$111>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $113 = HEAP32[$112>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    HEAP32[$__np$08>>2] = $113; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $114 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $115 = (($114) + ($29<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $116 = HEAP32[$115>>2]|0; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    HEAP32[$116>>2] = $__cp$019; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__cp$0 = HEAP32[$__cp$0$in26>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $117 = ($__cp$0|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($117) {
     label = 38;
     break L23;
    } else {
     $__cp$019 = $__cp$0;
    }
   }
   $__cp$018 = HEAP32[$__cp$019$lcssa>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $23 = ($__cp$018|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($23) {
    label = 38;
    break L23;
   } else {
    $__cp$0$in26 = $__cp$019$lcssa;$__cp$01827 = $__cp$018;$__pp$0$ph25 = $__cp$019$lcssa;
   }
  }
  HEAP32[$$lcssa>>2] = $__pp$0$ph25$lcssa; //@line 1989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $__cp$01824 = HEAP32[$__cp$019$lcssa111>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $88 = ($__cp$01824|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($88) {
   label = 38;
   break;
  } else {
   $__cp$0$in2332 = $__cp$019$lcssa111;$__cp$0182433 = $__cp$01824;$__phash$0$ph$ph31 = $$lcssa113;$__pp$0$ph$ph30 = $__cp$019$lcssa111;
  }
 }
 if ((label|0) == 38) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
}
function __ZN4love8graphics6opengl4Mesh23calculateAttributeSizesEv($this) {
 $this = $this|0;
 var $$ = 0, $$0 = 0, $$0$i = 0, $$01 = 0, $$off = 0, $$pre7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$04 = 0, $cond = 0, $stride$0$lcssa = 0, $stride$05 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 12|0;
 $1 = ((($this)) + 8|0); //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($this)) + 12|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$3>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($2|0)==($4|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($5) {
  $stride$0$lcssa = 0;
  $9 = ((($this)) + 60|0); //@line 174 "modules/graphics/opengl/Mesh.cpp"
  HEAP32[$9>>2] = $stride$0$lcssa; //@line 174 "modules/graphics/opengl/Mesh.cpp"
  STACKTOP = sp;return; //@line 175 "modules/graphics/opengl/Mesh.cpp"
 }
 $6 = ((($this)) + 20|0); //@line 170 "modules/graphics/opengl/Mesh.cpp"
 $7 = ((($this)) + 24|0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($this)) + 28|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $__begin$sroa$0$04 = $2;$stride$05 = 0;
 while(1) {
  $10 = ((($__begin$sroa$0$04)) + 12|0); //@line 163 "modules/graphics/opengl/Mesh.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 647 "modules/graphics/opengl/Mesh.cpp"
  $12 = ($11|0)==(0); //@line 163 "modules/graphics/opengl/Mesh.cpp"
  $13 = ((($__begin$sroa$0$04)) + 16|0); //@line 163 "modules/graphics/opengl/Mesh.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 163 "modules/graphics/opengl/Mesh.cpp"
  if ($12) {
   $15 = ($14|0)==(4); //@line 163 "modules/graphics/opengl/Mesh.cpp"
   if ($15) {
    $$0$i = 4;
   } else {
    label = 6;
    break;
   }
  } else {
   $$off = (($14) + -1)|0; //@line 166 "modules/graphics/opengl/Mesh.cpp"
   $21 = ($$off>>>0)>(3); //@line 166 "modules/graphics/opengl/Mesh.cpp"
   if ($21) {
    label = 10;
    break;
   }
   $cond = ($11|0)==(1); //@line 647 "modules/graphics/opengl/Mesh.cpp"
   $27 = $14 << 2; //@line 652 "modules/graphics/opengl/Mesh.cpp"
   $$ = $cond ? $27 : 0; //@line 647 "modules/graphics/opengl/Mesh.cpp"
   $$0$i = $$;
  }
  HEAP32[$0>>2] = $$0$i; //@line 170 "modules/graphics/opengl/Mesh.cpp"
  $28 = HEAP32[$7>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $29 = HEAP32[$8>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $30 = ($28>>>0)<($29>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($30) {
   HEAP32[$28>>2] = $$0$i; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $31 = ((($28)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$7>>2] = $31; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $33 = $31;
  } else {
   __ZNSt3__16vectorIjNS_9allocatorIjEEE21__push_back_slow_pathIjEEvOT_($6,$0); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$pre7 = HEAP32[$7>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $33 = $$pre7;
  }
  $32 = ((($33)) + -4|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $34 = HEAP32[$32>>2]|0; //@line 171 "modules/graphics/opengl/Mesh.cpp"
  $35 = (($34) + ($stride$05))|0; //@line 171 "modules/graphics/opengl/Mesh.cpp"
  $36 = ((($__begin$sroa$0$04)) + 20|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $37 = ($36|0)==($4|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($37) {
   $stride$0$lcssa = $35;
   label = 3;
   break;
  } else {
   $__begin$sroa$0$04 = $36;$stride$05 = $35;
  }
 }
 if ((label|0) == 3) {
  $9 = ((($this)) + 60|0); //@line 174 "modules/graphics/opengl/Mesh.cpp"
  HEAP32[$9>>2] = $stride$0$lcssa; //@line 174 "modules/graphics/opengl/Mesh.cpp"
  STACKTOP = sp;return; //@line 175 "modules/graphics/opengl/Mesh.cpp"
 }
 else if ((label|0) == 6) {
  $16 = (___cxa_allocate_exception(16)|0); //@line 164 "modules/graphics/opengl/Mesh.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($16|0),(436886|0),($vararg_buffer|0)); //@line 164 "modules/graphics/opengl/Mesh.cpp"
  $17 = __THREW__; __THREW__ = 0;
  $18 = $17&1;
  if (!($18)) {
   ___cxa_throw(($16|0),(96|0),(344|0)); //@line 164 "modules/graphics/opengl/Mesh.cpp"
   // unreachable; //@line 164 "modules/graphics/opengl/Mesh.cpp"
  }
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  ___cxa_free_exception(($16|0)); //@line 164 "modules/graphics/opengl/Mesh.cpp"
  $$0 = $20;$$01 = $19;
  ___resumeException($$01|0);
  // unreachable;
 }
 else if ((label|0) == 10) {
  $22 = (___cxa_allocate_exception(16)|0); //@line 167 "modules/graphics/opengl/Mesh.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($22|0),(436933|0),($vararg_buffer1|0)); //@line 167 "modules/graphics/opengl/Mesh.cpp"
  $23 = __THREW__; __THREW__ = 0;
  $24 = $23&1;
  if (!($24)) {
   ___cxa_throw(($22|0),(96|0),(344|0)); //@line 167 "modules/graphics/opengl/Mesh.cpp"
   // unreachable; //@line 167 "modules/graphics/opengl/Mesh.cpp"
  }
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  ___cxa_free_exception(($22|0)); //@line 167 "modules/graphics/opengl/Mesh.cpp"
  $$0 = $26;$$01 = $25;
  ___resumeException($$01|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl4Mesh16getGLBufferUsageENS2_5UsageE($usage) {
 $usage = $usage|0;
 var $$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 switch ($usage|0) {
 case 0:  {
  $$0 = 35040;
  break;
 }
 case 1:  {
  $$0 = 35048;
  break;
 }
 case 2:  {
  $$0 = 35044;
  break;
 }
 default: {
  $$0 = 0;
 }
 }
 return ($$0|0); //@line 723 "modules/graphics/opengl/Mesh.cpp"
}
function __ZN4love8graphics6opengl4MeshC2ERKNSt3__16vectorINS2_12AttribFormatENS3_9allocatorIS5_EEEEiNS2_8DrawModeENS2_5UsageE($this,$vertexformat,$vertexcount,$drawmode,$usage) {
 $this = $this|0;
 $vertexformat = $vertexformat|0;
 $vertexcount = $vertexcount|0;
 $drawmode = $drawmode|0;
 $usage = $usage|0;
 var $$$i = 0, $$0 = 0, $$0$i = 0, $$01 = 0, $$01$i$i$i = 0, $$1 = 0, $$12 = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, $vararg_buffer = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 __ZN4love6ObjectC2Ev($this); //@line 36 "./modules/graphics/Drawable.h"
 HEAP32[$this>>2] = (24932); //@line 100 "modules/graphics/opengl/Mesh.cpp"
 $0 = ((($this)) + 8|0); //@line 89 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 invoke_vii(4092,($0|0),($vertexformat|0)); //@line 89 "modules/graphics/opengl/Mesh.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $23 = ___cxa_find_matching_catch_2()|0;
  $24 = tempRet0;
  $$1 = $23;$$12 = $24;
  ___resumeException($$1|0);
  // unreachable;
 }
 $3 = ((($this)) + 20|0); //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ((($this)) + 24|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ((($this)) + 48|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 ;HEAP32[$3>>2]=0|0;HEAP32[$3+4>>2]=0|0;HEAP32[$3+8>>2]=0|0;HEAP32[$3+12>>2]=0|0;HEAP32[$3+16>>2]=0|0;HEAP32[$3+20>>2]=0|0;HEAP32[$3+24>>2]=0|0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAPF32[$5>>2] = 1.0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = ((($this)) + 52|0); //@line 90 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$6>>2] = 0; //@line 90 "modules/graphics/opengl/Mesh.cpp"
 $7 = ((($this)) + 56|0); //@line 91 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$7>>2] = $vertexcount; //@line 91 "modules/graphics/opengl/Mesh.cpp"
 $8 = ((($this)) + 60|0); //@line 92 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$8>>2] = 0; //@line 92 "modules/graphics/opengl/Mesh.cpp"
 $9 = ((($this)) + 68|0); //@line 93 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$9>>2] = 0; //@line 93 "modules/graphics/opengl/Mesh.cpp"
 $10 = ((($this)) + 72|0); //@line 94 "modules/graphics/opengl/Mesh.cpp"
 HEAP8[$10>>0] = 0; //@line 94 "modules/graphics/opengl/Mesh.cpp"
 $11 = ((($this)) + 76|0); //@line 95 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$11>>2] = 0; //@line 95 "modules/graphics/opengl/Mesh.cpp"
 $12 = ((($this)) + 80|0); //@line 96 "modules/graphics/opengl/Mesh.cpp"
 $13 = ($vertexcount>>>0)>(65535); //@line 689 "modules/graphics/opengl/Mesh.cpp"
 $$$i = $13 ? 5125 : 5123; //@line 690 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$12>>2] = $$$i; //@line 96 "modules/graphics/opengl/Mesh.cpp"
 $14 = ((($this)) + 84|0); //@line 97 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$14>>2] = $drawmode; //@line 97 "modules/graphics/opengl/Mesh.cpp"
 $15 = ((($this)) + 88|0); //@line 98 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$15>>2] = -1; //@line 98 "modules/graphics/opengl/Mesh.cpp"
 $16 = ((($this)) + 92|0); //@line 99 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$16>>2] = -1; //@line 99 "modules/graphics/opengl/Mesh.cpp"
 $17 = ((($this)) + 96|0); //@line 88 "./common/Object.h"
 HEAP32[$17>>2] = 0; //@line 88 "./common/Object.h"
 $18 = ($vertexcount|0)<(1); //@line 101 "modules/graphics/opengl/Mesh.cpp"
 do {
  if ($18) {
   $19 = (___cxa_allocate_exception(16)|0); //@line 102 "modules/graphics/opengl/Mesh.cpp"
   __THREW__ = 0;
   HEAP32[$vararg_buffer>>2] = $vertexcount; //@line 102 "modules/graphics/opengl/Mesh.cpp"
   invoke_viii(3349,($19|0),(437052|0),($vararg_buffer|0)); //@line 102 "modules/graphics/opengl/Mesh.cpp"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if ($21) {
    $25 = ___cxa_find_matching_catch_2()|0;
    $26 = tempRet0;
    ___cxa_free_exception(($19|0)); //@line 102 "modules/graphics/opengl/Mesh.cpp"
    $$0 = $25;$$01 = $26;
    break;
   } else {
    __THREW__ = 0;
    invoke_viii(3350,($19|0),(96|0),(344|0)); //@line 102 "modules/graphics/opengl/Mesh.cpp"
    $22 = __THREW__; __THREW__ = 0;
    label = 7;
    break;
   }
  } else {
   __THREW__ = 0;
   invoke_vi(4093,($this|0)); //@line 104 "modules/graphics/opengl/Mesh.cpp"
   $29 = __THREW__; __THREW__ = 0;
   $30 = $29&1;
   if ($30) {
    label = 7;
   } else {
    __THREW__ = 0;
    invoke_vi(4094,($this|0)); //@line 105 "modules/graphics/opengl/Mesh.cpp"
    $31 = __THREW__; __THREW__ = 0;
    $32 = $31&1;
    if ($32) {
     label = 7;
    } else {
     $33 = HEAP32[$7>>2]|0; //@line 107 "modules/graphics/opengl/Mesh.cpp"
     $34 = HEAP32[$8>>2]|0; //@line 107 "modules/graphics/opengl/Mesh.cpp"
     $35 = Math_imul($34, $33)|0; //@line 107 "modules/graphics/opengl/Mesh.cpp"
     __THREW__ = 0;
     $36 = (invoke_ii(3348,40)|0); //@line 109 "modules/graphics/opengl/Mesh.cpp"
     $37 = __THREW__; __THREW__ = 0;
     $38 = $37&1;
     if ($38) {
      label = 7;
     } else {
      switch ($usage|0) {
      case 0:  {
       $$0$i = 35040;
       break;
      }
      case 1:  {
       $$0$i = 35048;
       break;
      }
      case 2:  {
       $$0$i = 35044;
       break;
      }
      default: {
       $$0$i = 0;
      }
      }
      __THREW__ = 0;
      invoke_viiiiii(4042,($36|0),($35|0),(0|0),34962,($$0$i|0),1); //@line 109 "modules/graphics/opengl/Mesh.cpp"
      $39 = __THREW__; __THREW__ = 0;
      $40 = $39&1;
      if ($40) {
       $64 = ___cxa_find_matching_catch_2()|0;
       $65 = tempRet0;
       __ZdlPv($36); //@line 109 "modules/graphics/opengl/Mesh.cpp"
       $$0 = $64;$$01 = $65;
       break;
      }
      HEAP32[$6>>2] = $36; //@line 109 "modules/graphics/opengl/Mesh.cpp"
      __THREW__ = 0;
      invoke_vi(4087,($36|0)); //@line 188 "modules/graphics/opengl/GLBuffer.h"
      $41 = __THREW__; __THREW__ = 0;
      $42 = $41&1;
      if ($42) {
       label = 7;
      } else {
       $43 = HEAP32[$6>>2]|0; //@line 113 "modules/graphics/opengl/Mesh.cpp"
       __THREW__ = 0;
       $44 = (invoke_ii(4096,($43|0))|0); //@line 113 "modules/graphics/opengl/Mesh.cpp"
       $45 = __THREW__; __THREW__ = 0;
       $46 = $45&1;
       if (!($46)) {
        _memset(($44|0),0,($35|0))|0; //@line 113 "modules/graphics/opengl/Mesh.cpp"
        $47 = HEAP32[$6>>2]|0; //@line 114 "modules/graphics/opengl/Mesh.cpp"
        $48 = ((($47)) + 8|0); //@line 76 "modules/graphics/opengl/GLBuffer.h"
        $49 = HEAP32[$48>>2]|0; //@line 76 "modules/graphics/opengl/GLBuffer.h"
        __THREW__ = 0;
        invoke_viii(4097,($47|0),0,($49|0)); //@line 114 "modules/graphics/opengl/Mesh.cpp"
        $50 = __THREW__; __THREW__ = 0;
        $51 = $50&1;
        if (!($51)) {
         $52 = HEAP32[$6>>2]|0; //@line 115 "modules/graphics/opengl/Mesh.cpp"
         __THREW__ = 0;
         invoke_vi(4088,($52|0)); //@line 115 "modules/graphics/opengl/Mesh.cpp"
         $53 = __THREW__; __THREW__ = 0;
         $54 = $53&1;
         if (!($54)) {
          $55 = HEAP32[$8>>2]|0; //@line 117 "modules/graphics/opengl/Mesh.cpp"
          __THREW__ = 0;
          $56 = (invoke_ii(3492,($55|0))|0); //@line 117 "modules/graphics/opengl/Mesh.cpp"
          $57 = __THREW__; __THREW__ = 0;
          $58 = $57&1;
          if (!($58)) {
           $59 = ((($this)) + 64|0); //@line 117 "modules/graphics/opengl/Mesh.cpp"
           HEAP32[$59>>2] = $56; //@line 117 "modules/graphics/opengl/Mesh.cpp"
           __THREW__ = 0;
           invoke_vi(4089,($36|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
           $60 = __THREW__; __THREW__ = 0;
           $61 = $60&1;
           if ($61) {
            $62 = ___cxa_find_matching_catch_3(0|0)|0;
            $63 = tempRet0;
            ___clang_call_terminate($62); //@line 196 "modules/graphics/opengl/GLBuffer.h"
            // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
           } else {
            STACKTOP = sp;return; //@line 118 "modules/graphics/opengl/Mesh.cpp"
           }
          }
         }
        }
       }
       $66 = ___cxa_find_matching_catch_2()|0;
       $67 = tempRet0;
       __THREW__ = 0;
       invoke_vi(4089,($36|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
       $68 = __THREW__; __THREW__ = 0;
       $69 = $68&1;
       if ($69) {
        $70 = ___cxa_find_matching_catch_3(0|0)|0;
        $71 = tempRet0;
        ___clang_call_terminate($70); //@line 196 "modules/graphics/opengl/GLBuffer.h"
        // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
       } else {
        $$0 = $66;$$01 = $67;
       }
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 7) {
  $27 = ___cxa_find_matching_catch_2()|0;
  $28 = tempRet0;
  $$0 = $27;$$01 = $28;
 }
 $72 = HEAP32[$17>>2]|0; //@line 106 "./common/Object.h"
 $73 = ($72|0)==(0|0); //@line 106 "./common/Object.h"
 if (!($73)) {
  $74 = HEAP32[$72>>2]|0; //@line 106 "./common/Object.h"
  $75 = ((($74)) + 12|0); //@line 106 "./common/Object.h"
  $76 = HEAP32[$75>>2]|0; //@line 106 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($76|0,($72|0)); //@line 106 "./common/Object.h"
  $77 = __THREW__; __THREW__ = 0;
  $78 = $77&1;
  if ($78) {
   $79 = ___cxa_find_matching_catch_3(0|0)|0;
   $80 = tempRet0;
   ___clang_call_terminate($79); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 }
 $81 = ((($this)) + 40|0); //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $82 = HEAP32[$81>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $83 = ($82|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($83)) {
  $$01$i$i$i = $82;
  while(1) {
   $84 = HEAP32[$$01$i$i$i>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $85 = ((($$01$i$i$i)) + 8|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($85); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
   __ZdlPv($$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $86 = ($84|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($86) {
    break;
   } else {
    $$01$i$i$i = $84;
   }
  }
 }
 $87 = ((($this)) + 32|0); //@line 2014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $88 = HEAP32[$87>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$87>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $89 = ($88|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($89)) {
  __ZdlPv($88); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $90 = HEAP32[$3>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $91 = ($90|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $92 = $90; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($91)) {
  $93 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $94 = ($93|0)==($90|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($94)) {
   $scevgep$i$i$i = ((($93)) + -4|0);
   $95 = $scevgep$i$i$i;
   $96 = (($95) - ($92))|0;
   $97 = $96 >>> 2;
   $98 = $97 ^ -1;
   $scevgep4$i$i$i = (($93) + ($98<<2)|0);
   HEAP32[$4>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($90); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $99 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $100 = ($99|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($100) {
  $$1 = $$0;$$12 = $$01;
  ___resumeException($$1|0);
  // unreachable;
 }
 $101 = ((($this)) + 12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $102 = HEAP32[$101>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $103 = ($102|0)==($99|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($103) {
  $108 = $99;
 } else {
  $105 = $102;
  while(1) {
   $104 = ((($105)) + -20|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$101>>2] = $104; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($104); //@line 81 "modules/graphics/opengl/Mesh.h"
   $106 = HEAP32[$101>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $107 = ($106|0)==($99|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($107) {
    break;
   } else {
    $105 = $106;
   }
  }
  $$pre$i = HEAP32[$0>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $108 = $$pre$i;
 }
 __ZdlPv($108); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $$1 = $$0;$$12 = $$01;
 ___resumeException($$1|0);
 // unreachable;
}
function __ZN4love8graphics6opengl4MeshC2ERKNSt3__16vectorINS_6VertexENS3_9allocatorIS5_EEEENS2_8DrawModeENS2_5UsageE($this,$vertices,$drawmode,$usage) {
 $this = $this|0;
 $vertices = $vertices|0;
 $drawmode = $drawmode|0;
 $usage = $usage|0;
 var $$cast = 0, $$pre$i = 0, $$pre$i3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 __ZN4love8graphics6openglL22getDefaultVertexFormatEv($0); //@line 121 "modules/graphics/opengl/Mesh.cpp"
 $1 = HEAP32[$vertices>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($vertices)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $$cast = $1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($3) - ($$cast))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 __THREW__ = 0;
 invoke_viiiiii(4072,($this|0),($0|0),($1|0),($4|0),($drawmode|0),($usage|0)); //@line 121 "modules/graphics/opengl/Mesh.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  $19 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = ($19|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($20) {
   ___resumeException($17|0);
   // unreachable;
  }
  $21 = ((($0)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = HEAP32[$21>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = ($22|0)==($19|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($23) {
   $28 = $19;
  } else {
   $25 = $22;
   while(1) {
    $24 = ((($25)) + -20|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$21>>2] = $24; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($24); //@line 81 "modules/graphics/opengl/Mesh.h"
    $26 = HEAP32[$21>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $27 = ($26|0)==($19|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($27) {
     break;
    } else {
     $25 = $26;
    }
   }
   $$pre$i = HEAP32[$0>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $28 = $$pre$i;
  }
  __ZdlPv($28); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($17|0);
  // unreachable;
 } else {
  $7 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $8 = ($7|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($8) {
   STACKTOP = sp;return; //@line 123 "modules/graphics/opengl/Mesh.cpp"
  }
  $9 = ((($0)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $10 = HEAP32[$9>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $11 = ($10|0)==($7|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($11) {
   $16 = $7;
  } else {
   $13 = $10;
   while(1) {
    $12 = ((($13)) + -20|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$9>>2] = $12; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12); //@line 81 "modules/graphics/opengl/Mesh.h"
    $14 = HEAP32[$9>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $15 = ($14|0)==($7|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($15) {
     break;
    } else {
     $13 = $14;
    }
   }
   $$pre$i3 = HEAP32[$0>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $16 = $$pre$i3;
  }
  __ZdlPv($16); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  STACKTOP = sp;return; //@line 123 "modules/graphics/opengl/Mesh.cpp"
 }
}
function __ZN4love8graphics6openglL22getDefaultVertexFormatEv($agg$result) {
 $agg$result = $agg$result|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $lpad$phi$i17$index = 0, $lpad$phi$i17$index3 = 0, $name$i = 0, $name$i3 = 0, $name$i6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name$i6 = sp + 68|0;
 $name$i3 = sp + 64|0;
 $name$i = sp + 60|0;
 $0 = sp;
 HEAP32[$name$i>>2] = 1282466; //@line 40 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 (invoke_iii(4098,0,($name$i|0))|0); //@line 41 "modules/graphics/opengl/Mesh.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $$02 = $0;
 } else {
  $3 = HEAP32[$name$i>>2]|0; //@line 42 "modules/graphics/opengl/Mesh.cpp"
  $4 = (_strlen($3)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  __THREW__ = 0;
  invoke_viii(3460,($0|0),($3|0),($4|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if ($6) {
   $$02 = $0;
  } else {
   $7 = ((($0)) + 12|0); //@line 53 "modules/graphics/opengl/Mesh.cpp"
   HEAP32[$7>>2] = 1; //@line 53 "modules/graphics/opengl/Mesh.cpp"
   $8 = ((($0)) + 16|0); //@line 53 "modules/graphics/opengl/Mesh.cpp"
   HEAP32[$8>>2] = 2; //@line 53 "modules/graphics/opengl/Mesh.cpp"
   $9 = ((($0)) + 20|0); //@line 52 "modules/graphics/opengl/Mesh.cpp"
   HEAP32[$name$i3>>2] = 1282466; //@line 40 "modules/graphics/opengl/Mesh.cpp"
   __THREW__ = 0;
   (invoke_iii(4098,1,($name$i3|0))|0); //@line 41 "modules/graphics/opengl/Mesh.cpp"
   $10 = __THREW__; __THREW__ = 0;
   $11 = $10&1;
   if ($11) {
    $$02 = $9;
   } else {
    $12 = HEAP32[$name$i3>>2]|0; //@line 42 "modules/graphics/opengl/Mesh.cpp"
    $13 = (_strlen($12)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    __THREW__ = 0;
    invoke_viii(3460,($9|0),($12|0),($13|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $14 = __THREW__; __THREW__ = 0;
    $15 = $14&1;
    if ($15) {
     $$02 = $9;
    } else {
     $16 = ((($0)) + 32|0); //@line 54 "modules/graphics/opengl/Mesh.cpp"
     HEAP32[$16>>2] = 1; //@line 54 "modules/graphics/opengl/Mesh.cpp"
     $17 = ((($0)) + 36|0); //@line 54 "modules/graphics/opengl/Mesh.cpp"
     HEAP32[$17>>2] = 2; //@line 54 "modules/graphics/opengl/Mesh.cpp"
     $18 = ((($0)) + 40|0); //@line 52 "modules/graphics/opengl/Mesh.cpp"
     HEAP32[$name$i6>>2] = 1282466; //@line 40 "modules/graphics/opengl/Mesh.cpp"
     __THREW__ = 0;
     (invoke_iii(4098,2,($name$i6|0))|0); //@line 41 "modules/graphics/opengl/Mesh.cpp"
     $19 = __THREW__; __THREW__ = 0;
     $20 = $19&1;
     if ($20) {
      $$02 = $18;
     } else {
      $21 = HEAP32[$name$i6>>2]|0; //@line 42 "modules/graphics/opengl/Mesh.cpp"
      $22 = (_strlen($21)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __THREW__ = 0;
      invoke_viii(3460,($18|0),($21|0),($22|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $23 = __THREW__; __THREW__ = 0;
      $24 = $23&1;
      if ($24) {
       $$02 = $18;
      } else {
       $25 = ((($0)) + 52|0); //@line 55 "modules/graphics/opengl/Mesh.cpp"
       HEAP32[$25>>2] = 0; //@line 55 "modules/graphics/opengl/Mesh.cpp"
       $26 = ((($0)) + 56|0); //@line 55 "modules/graphics/opengl/Mesh.cpp"
       HEAP32[$26>>2] = 4; //@line 55 "modules/graphics/opengl/Mesh.cpp"
       HEAP32[$agg$result>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $27 = ((($agg$result)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$27>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $28 = ((($agg$result)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$28>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       __THREW__ = 0;
       $29 = (invoke_ii(3348,60)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       $30 = __THREW__; __THREW__ = 0;
       $31 = $30&1;
       if ($31) {
        $52 = ___cxa_find_matching_catch_2()|0;
        $53 = tempRet0;
        $lpad$phi$i17$index = $52;$lpad$phi$i17$index3 = $53;
       } else {
        HEAP32[$27>>2] = $29; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$agg$result>>2] = $29; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $32 = ((($29)) + 60|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $33 = ((($agg$result)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        HEAP32[$33>>2] = $32; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        __THREW__ = 0;
        invoke_vii(3351,($29|0),($0|0)); //@line 81 "modules/graphics/opengl/Mesh.h"
        $34 = __THREW__; __THREW__ = 0;
        $35 = $34&1;
        if ($35) {
         $58 = $29;
        } else {
         $36 = ((($29)) + 12|0); //@line 81 "modules/graphics/opengl/Mesh.h"
         $37 = ((($0)) + 12|0); //@line 81 "modules/graphics/opengl/Mesh.h"
         $38 = $37; //@line 81 "modules/graphics/opengl/Mesh.h"
         $39 = $38; //@line 81 "modules/graphics/opengl/Mesh.h"
         $40 = HEAP32[$39>>2]|0; //@line 81 "modules/graphics/opengl/Mesh.h"
         $41 = (($38) + 4)|0; //@line 81 "modules/graphics/opengl/Mesh.h"
         $42 = $41; //@line 81 "modules/graphics/opengl/Mesh.h"
         $43 = HEAP32[$42>>2]|0; //@line 81 "modules/graphics/opengl/Mesh.h"
         $44 = $36; //@line 81 "modules/graphics/opengl/Mesh.h"
         $45 = $44; //@line 81 "modules/graphics/opengl/Mesh.h"
         HEAP32[$45>>2] = $40; //@line 81 "modules/graphics/opengl/Mesh.h"
         $46 = (($44) + 4)|0; //@line 81 "modules/graphics/opengl/Mesh.h"
         $47 = $46; //@line 81 "modules/graphics/opengl/Mesh.h"
         HEAP32[$47>>2] = $43; //@line 81 "modules/graphics/opengl/Mesh.h"
         $48 = ((($29)) + 20|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$27>>2] = $48; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $49 = ((($0)) + 20|0); //@line 81 "modules/graphics/opengl/Mesh.h"
         __THREW__ = 0;
         invoke_vii(3351,($48|0),($49|0)); //@line 81 "modules/graphics/opengl/Mesh.h"
         $50 = __THREW__; __THREW__ = 0;
         $51 = $50&1;
         if ($51) {
          $58 = $48;
         } else {
          $70 = ((($29)) + 32|0); //@line 81 "modules/graphics/opengl/Mesh.h"
          $71 = ((($0)) + 32|0); //@line 81 "modules/graphics/opengl/Mesh.h"
          $72 = $71; //@line 81 "modules/graphics/opengl/Mesh.h"
          $73 = $72; //@line 81 "modules/graphics/opengl/Mesh.h"
          $74 = HEAP32[$73>>2]|0; //@line 81 "modules/graphics/opengl/Mesh.h"
          $75 = (($72) + 4)|0; //@line 81 "modules/graphics/opengl/Mesh.h"
          $76 = $75; //@line 81 "modules/graphics/opengl/Mesh.h"
          $77 = HEAP32[$76>>2]|0; //@line 81 "modules/graphics/opengl/Mesh.h"
          $78 = $70; //@line 81 "modules/graphics/opengl/Mesh.h"
          $79 = $78; //@line 81 "modules/graphics/opengl/Mesh.h"
          HEAP32[$79>>2] = $74; //@line 81 "modules/graphics/opengl/Mesh.h"
          $80 = (($78) + 4)|0; //@line 81 "modules/graphics/opengl/Mesh.h"
          $81 = $80; //@line 81 "modules/graphics/opengl/Mesh.h"
          HEAP32[$81>>2] = $77; //@line 81 "modules/graphics/opengl/Mesh.h"
          $82 = HEAP32[$27>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $83 = ((($82)) + 20|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$27>>2] = $83; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $84 = ((($0)) + 40|0); //@line 81 "modules/graphics/opengl/Mesh.h"
          __THREW__ = 0;
          invoke_vii(3351,($83|0),($84|0)); //@line 81 "modules/graphics/opengl/Mesh.h"
          $85 = __THREW__; __THREW__ = 0;
          $86 = $85&1;
          if ($86) {
           $58 = $83;
          } else {
           $87 = ((($82)) + 32|0); //@line 81 "modules/graphics/opengl/Mesh.h"
           $88 = ((($0)) + 52|0); //@line 81 "modules/graphics/opengl/Mesh.h"
           $89 = $88; //@line 81 "modules/graphics/opengl/Mesh.h"
           $90 = $89; //@line 81 "modules/graphics/opengl/Mesh.h"
           $91 = HEAP32[$90>>2]|0; //@line 81 "modules/graphics/opengl/Mesh.h"
           $92 = (($89) + 4)|0; //@line 81 "modules/graphics/opengl/Mesh.h"
           $93 = $92; //@line 81 "modules/graphics/opengl/Mesh.h"
           $94 = HEAP32[$93>>2]|0; //@line 81 "modules/graphics/opengl/Mesh.h"
           $95 = $87; //@line 81 "modules/graphics/opengl/Mesh.h"
           $96 = $95; //@line 81 "modules/graphics/opengl/Mesh.h"
           HEAP32[$96>>2] = $91; //@line 81 "modules/graphics/opengl/Mesh.h"
           $97 = (($95) + 4)|0; //@line 81 "modules/graphics/opengl/Mesh.h"
           $98 = $97; //@line 81 "modules/graphics/opengl/Mesh.h"
           HEAP32[$98>>2] = $94; //@line 81 "modules/graphics/opengl/Mesh.h"
           $99 = ((($82)) + 40|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           HEAP32[$27>>2] = $99; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $100 = ((($0)) + 40|0); //@line 81 "modules/graphics/opengl/Mesh.h"
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($100); //@line 81 "modules/graphics/opengl/Mesh.h"
           $101 = ((($0)) + 20|0); //@line 81 "modules/graphics/opengl/Mesh.h"
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($101); //@line 81 "modules/graphics/opengl/Mesh.h"
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 81 "modules/graphics/opengl/Mesh.h"
           STACKTOP = sp;return; //@line 59 "modules/graphics/opengl/Mesh.cpp"
          }
         }
        }
        $54 = ___cxa_find_matching_catch_2()|0;
        $55 = tempRet0;
        $$pre = HEAP32[$agg$result>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $56 = ($$pre|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($56) {
         $lpad$phi$i17$index = $54;$lpad$phi$i17$index3 = $55;
        } else {
         $57 = ($58|0)==($$pre|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if (!($57)) {
          $60 = $58;
          while(1) {
           $59 = ((($60)) + -20|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($59); //@line 81 "modules/graphics/opengl/Mesh.h"
           $61 = ($59|0)==($$pre|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           if ($61) {
            break;
           } else {
            $60 = $59;
           }
          }
          HEAP32[$27>>2] = $$pre; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         }
         __ZdlPv($$pre); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
         $lpad$phi$i17$index = $54;$lpad$phi$i17$index3 = $55;
        }
       }
       $68 = ((($0)) + 40|0); //@line 81 "modules/graphics/opengl/Mesh.h"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($68); //@line 81 "modules/graphics/opengl/Mesh.h"
       $69 = ((($0)) + 20|0); //@line 81 "modules/graphics/opengl/Mesh.h"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($69); //@line 81 "modules/graphics/opengl/Mesh.h"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 81 "modules/graphics/opengl/Mesh.h"
       $$0 = $lpad$phi$i17$index3;$$01 = $lpad$phi$i17$index;
       ___resumeException($$01|0);
       // unreachable;
      }
     }
    }
   }
  }
 }
 $62 = ___cxa_find_matching_catch_2()|0;
 $63 = tempRet0;
 $64 = ($0|0)==($$02|0); //@line 52 "modules/graphics/opengl/Mesh.cpp"
 if ($64) {
  $$0 = $63;$$01 = $62;
  ___resumeException($$01|0);
  // unreachable;
 } else {
  $66 = $$02;
 }
 while(1) {
  $65 = ((($66)) + -20|0); //@line 52 "modules/graphics/opengl/Mesh.cpp"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($65); //@line 81 "modules/graphics/opengl/Mesh.h"
  $67 = ($65|0)==($0|0); //@line 52 "modules/graphics/opengl/Mesh.cpp"
  if ($67) {
   $$0 = $63;$$01 = $62;
   break;
  } else {
   $66 = $65;
  }
 }
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN4love8graphics6opengl4MeshC2EiNS2_8DrawModeENS2_5UsageE($this,$vertexcount,$drawmode,$usage) {
 $this = $this|0;
 $vertexcount = $vertexcount|0;
 $drawmode = $drawmode|0;
 $usage = $usage|0;
 var $$pre$i = 0, $$pre$i3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 __ZN4love8graphics6openglL22getDefaultVertexFormatEv($0); //@line 126 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 invoke_viiiii(4071,($this|0),($0|0),($vertexcount|0),($drawmode|0),($usage|0)); //@line 126 "modules/graphics/opengl/Mesh.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  $15 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($16) {
   ___resumeException($13|0);
   // unreachable;
  }
  $17 = ((($0)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$17>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = ($18|0)==($15|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($19) {
   $24 = $15;
  } else {
   $21 = $18;
   while(1) {
    $20 = ((($21)) + -20|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$17>>2] = $20; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($20); //@line 81 "modules/graphics/opengl/Mesh.h"
    $22 = HEAP32[$17>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $23 = ($22|0)==($15|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($23) {
     break;
    } else {
     $21 = $22;
    }
   }
   $$pre$i3 = HEAP32[$0>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $24 = $$pre$i3;
  }
  __ZdlPv($24); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($13|0);
  // unreachable;
 } else {
  $3 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = ($3|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($4) {
   STACKTOP = sp;return; //@line 128 "modules/graphics/opengl/Mesh.cpp"
  }
  $5 = ((($0)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $6 = HEAP32[$5>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $7 = ($6|0)==($3|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($7) {
   $12 = $3;
  } else {
   $9 = $6;
   while(1) {
    $8 = ((($9)) + -20|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$5>>2] = $8; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8); //@line 81 "modules/graphics/opengl/Mesh.h"
    $10 = HEAP32[$5>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $11 = ($10|0)==($3|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($11) {
     break;
    } else {
     $9 = $10;
    }
   }
   $$pre$i = HEAP32[$0>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $12 = $$pre$i;
  }
  __ZdlPv($12); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  STACKTOP = sp;return; //@line 128 "modules/graphics/opengl/Mesh.cpp"
 }
}
function __ZN4love8graphics6opengl4Mesh9setVertexEjPKvj($this,$vertindex,$data,$datasize) {
 $this = $this|0;
 $vertindex = $vertindex|0;
 $data = $data|0;
 $datasize = $datasize|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 56|0); //@line 189 "modules/graphics/opengl/Mesh.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 189 "modules/graphics/opengl/Mesh.cpp"
 $2 = ($1>>>0)>($vertindex>>>0); //@line 189 "modules/graphics/opengl/Mesh.cpp"
 if (!($2)) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 190 "modules/graphics/opengl/Mesh.cpp"
  $4 = (($vertindex) + 1)|0; //@line 190 "modules/graphics/opengl/Mesh.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer>>2] = $4; //@line 190 "modules/graphics/opengl/Mesh.cpp"
  invoke_viii(3349,($3|0),(437085|0),($vararg_buffer|0)); //@line 190 "modules/graphics/opengl/Mesh.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   ___cxa_throw(($3|0),(96|0),(344|0)); //@line 190 "modules/graphics/opengl/Mesh.cpp"
   // unreachable; //@line 190 "modules/graphics/opengl/Mesh.cpp"
  }
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  ___cxa_free_exception(($3|0)); //@line 190 "modules/graphics/opengl/Mesh.cpp"
  $$0 = $7;$$01 = $8;
  ___resumeException($$0|0);
  // unreachable;
 }
 $9 = ((($this)) + 60|0); //@line 192 "modules/graphics/opengl/Mesh.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 192 "modules/graphics/opengl/Mesh.cpp"
 $11 = Math_imul($10, $vertindex)|0; //@line 192 "modules/graphics/opengl/Mesh.cpp"
 $12 = ($10>>>0)<($datasize>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $13 = $12 ? $10 : $datasize; //@line 193 "modules/graphics/opengl/Mesh.cpp"
 $14 = ((($this)) + 52|0); //@line 195 "modules/graphics/opengl/Mesh.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 195 "modules/graphics/opengl/Mesh.cpp"
 __ZN4love8graphics6opengl8GLBuffer4bindEv($15); //@line 188 "modules/graphics/opengl/GLBuffer.h"
 $16 = HEAP32[$14>>2]|0; //@line 196 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 $17 = (invoke_ii(4096,($16|0))|0); //@line 196 "modules/graphics/opengl/Mesh.cpp"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if (!($19)) {
  $20 = (($17) + ($11)|0); //@line 198 "modules/graphics/opengl/Mesh.cpp"
  _memcpy(($20|0),($data|0),($13|0))|0; //@line 198 "modules/graphics/opengl/Mesh.cpp"
  $21 = HEAP32[$14>>2]|0; //@line 200 "modules/graphics/opengl/Mesh.cpp"
  __THREW__ = 0;
  invoke_viii(4097,($21|0),($11|0),($13|0)); //@line 200 "modules/graphics/opengl/Mesh.cpp"
  $22 = __THREW__; __THREW__ = 0;
  $23 = $22&1;
  if (!($23)) {
   __THREW__ = 0;
   invoke_vi(4089,($15|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   $24 = __THREW__; __THREW__ = 0;
   $25 = $24&1;
   if ($25) {
    $26 = ___cxa_find_matching_catch_3(0|0)|0;
    $27 = tempRet0;
    ___clang_call_terminate($26); //@line 196 "modules/graphics/opengl/GLBuffer.h"
    // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
   } else {
    STACKTOP = sp;return; //@line 201 "modules/graphics/opengl/Mesh.cpp"
   }
  }
 }
 $28 = ___cxa_find_matching_catch_2()|0;
 $29 = tempRet0;
 __THREW__ = 0;
 invoke_vi(4089,($15|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
 $30 = __THREW__; __THREW__ = 0;
 $31 = $30&1;
 if ($31) {
  $32 = ___cxa_find_matching_catch_3(0|0)|0;
  $33 = tempRet0;
  ___clang_call_terminate($32); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
 } else {
  $$0 = $28;$$01 = $29;
  ___resumeException($$0|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl4Mesh9getVertexEjPvj($this,$vertindex,$data,$datasize) {
 $this = $this|0;
 $vertindex = $vertindex|0;
 $data = $data|0;
 $datasize = $datasize|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 56|0); //@line 205 "modules/graphics/opengl/Mesh.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 205 "modules/graphics/opengl/Mesh.cpp"
 $2 = ($1>>>0)>($vertindex>>>0); //@line 205 "modules/graphics/opengl/Mesh.cpp"
 if (!($2)) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 206 "modules/graphics/opengl/Mesh.cpp"
  $4 = (($vertindex) + 1)|0; //@line 206 "modules/graphics/opengl/Mesh.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer>>2] = $4; //@line 206 "modules/graphics/opengl/Mesh.cpp"
  invoke_viii(3349,($3|0),(437085|0),($vararg_buffer|0)); //@line 206 "modules/graphics/opengl/Mesh.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   ___cxa_throw(($3|0),(96|0),(344|0)); //@line 206 "modules/graphics/opengl/Mesh.cpp"
   // unreachable; //@line 206 "modules/graphics/opengl/Mesh.cpp"
  }
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  ___cxa_free_exception(($3|0)); //@line 206 "modules/graphics/opengl/Mesh.cpp"
  $$0 = $7;$$01 = $8;
  ___resumeException($$0|0);
  // unreachable;
 }
 $9 = ((($this)) + 60|0); //@line 208 "modules/graphics/opengl/Mesh.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 208 "modules/graphics/opengl/Mesh.cpp"
 $11 = ($10>>>0)<($datasize>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $12 = $11 ? $10 : $datasize; //@line 209 "modules/graphics/opengl/Mesh.cpp"
 $13 = ((($this)) + 52|0); //@line 212 "modules/graphics/opengl/Mesh.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 212 "modules/graphics/opengl/Mesh.cpp"
 __ZN4love8graphics6opengl8GLBuffer4bindEv($14); //@line 188 "modules/graphics/opengl/GLBuffer.h"
 $15 = HEAP32[$13>>2]|0; //@line 213 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 $16 = (invoke_ii(4096,($15|0))|0); //@line 213 "modules/graphics/opengl/Mesh.cpp"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if ($18) {
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  __THREW__ = 0;
  invoke_vi(4089,($14|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  $27 = __THREW__; __THREW__ = 0;
  $28 = $27&1;
  if ($28) {
   $29 = ___cxa_find_matching_catch_3(0|0)|0;
   $30 = tempRet0;
   ___clang_call_terminate($29); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
  } else {
   $$0 = $25;$$01 = $26;
   ___resumeException($$0|0);
   // unreachable;
  }
 } else {
  $19 = Math_imul($10, $vertindex)|0; //@line 208 "modules/graphics/opengl/Mesh.cpp"
  $20 = (($16) + ($19)|0); //@line 215 "modules/graphics/opengl/Mesh.cpp"
  _memcpy(($data|0),($20|0),($12|0))|0; //@line 215 "modules/graphics/opengl/Mesh.cpp"
  __THREW__ = 0;
  invoke_vi(4089,($14|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  $21 = __THREW__; __THREW__ = 0;
  $22 = $21&1;
  if ($22) {
   $23 = ___cxa_find_matching_catch_3(0|0)|0;
   $24 = tempRet0;
   ___clang_call_terminate($23); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
  } else {
   STACKTOP = sp;return ($12|0); //@line 218 "modules/graphics/opengl/Mesh.cpp"
  }
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl4Mesh22getVertexScratchBufferEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 64|0); //@line 222 "modules/graphics/opengl/Mesh.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 222 "modules/graphics/opengl/Mesh.cpp"
 return ($1|0); //@line 222 "modules/graphics/opengl/Mesh.cpp"
}
function __ZN4love8graphics6opengl4Mesh18setVertexAttributeEjiPKvj($this,$vertindex,$attribindex,$data,$datasize) {
 $this = $this|0;
 $vertindex = $vertindex|0;
 $attribindex = $attribindex|0;
 $data = $data|0;
 $datasize = $datasize|0;
 var $$0 = 0, $$01 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i = 0;
 var $i$02$i = 0, $offset$0$lcssa$i = 0, $offset$01$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($this)) + 56|0); //@line 227 "modules/graphics/opengl/Mesh.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 227 "modules/graphics/opengl/Mesh.cpp"
 $2 = ($1>>>0)>($vertindex>>>0); //@line 227 "modules/graphics/opengl/Mesh.cpp"
 if (!($2)) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 228 "modules/graphics/opengl/Mesh.cpp"
  $4 = (($vertindex) + 1)|0; //@line 228 "modules/graphics/opengl/Mesh.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer>>2] = $4; //@line 228 "modules/graphics/opengl/Mesh.cpp"
  invoke_viii(3349,($3|0),(437085|0),($vararg_buffer|0)); //@line 228 "modules/graphics/opengl/Mesh.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   ___cxa_throw(($3|0),(96|0),(344|0)); //@line 228 "modules/graphics/opengl/Mesh.cpp"
   // unreachable; //@line 228 "modules/graphics/opengl/Mesh.cpp"
  }
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  ___cxa_free_exception(($3|0)); //@line 228 "modules/graphics/opengl/Mesh.cpp"
  $$0 = $7;$$01 = $8;
  ___resumeException($$0|0);
  // unreachable;
 }
 $9 = ((($this)) + 8|0); //@line 230 "modules/graphics/opengl/Mesh.cpp"
 $10 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = HEAP32[$10>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = HEAP32[$9>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = (($11) - ($12))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = (($13|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = ($14|0)>($attribindex|0); //@line 230 "modules/graphics/opengl/Mesh.cpp"
 if (!($15)) {
  $16 = (___cxa_allocate_exception(16)|0); //@line 231 "modules/graphics/opengl/Mesh.cpp"
  $17 = (($attribindex) + 1)|0; //@line 231 "modules/graphics/opengl/Mesh.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer1>>2] = $17; //@line 231 "modules/graphics/opengl/Mesh.cpp"
  invoke_viii(3349,($16|0),(437111|0),($vararg_buffer1|0)); //@line 231 "modules/graphics/opengl/Mesh.cpp"
  $18 = __THREW__; __THREW__ = 0;
  $19 = $18&1;
  if (!($19)) {
   ___cxa_throw(($16|0),(96|0),(344|0)); //@line 231 "modules/graphics/opengl/Mesh.cpp"
   // unreachable; //@line 231 "modules/graphics/opengl/Mesh.cpp"
  }
  $20 = ___cxa_find_matching_catch_2()|0;
  $21 = tempRet0;
  ___cxa_free_exception(($16|0)); //@line 231 "modules/graphics/opengl/Mesh.cpp"
  $$0 = $20;$$01 = $21;
  ___resumeException($$0|0);
  // unreachable;
 }
 $22 = ((($this)) + 60|0); //@line 233 "modules/graphics/opengl/Mesh.cpp"
 $23 = HEAP32[$22>>2]|0; //@line 233 "modules/graphics/opengl/Mesh.cpp"
 $24 = Math_imul($23, $vertindex)|0; //@line 233 "modules/graphics/opengl/Mesh.cpp"
 $25 = ($attribindex|0)==(0); //@line 181 "modules/graphics/opengl/Mesh.cpp"
 $$phi$trans$insert = ((($this)) + 20|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($25) {
  $offset$0$lcssa$i = 0;
 } else {
  $i$02$i = 0;$offset$01$i = 0;
  while(1) {
   $26 = (($$pre) + ($i$02$i<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $27 = HEAP32[$26>>2]|0; //@line 182 "modules/graphics/opengl/Mesh.cpp"
   $28 = (($27) + ($offset$01$i))|0; //@line 182 "modules/graphics/opengl/Mesh.cpp"
   $29 = (($i$02$i) + 1)|0; //@line 181 "modules/graphics/opengl/Mesh.cpp"
   $exitcond$i = ($29|0)==($attribindex|0); //@line 181 "modules/graphics/opengl/Mesh.cpp"
   if ($exitcond$i) {
    $offset$0$lcssa$i = $28;
    break;
   } else {
    $i$02$i = $29;$offset$01$i = $28;
   }
  }
 }
 $30 = (($offset$0$lcssa$i) + ($24))|0; //@line 233 "modules/graphics/opengl/Mesh.cpp"
 $31 = (($$pre) + ($attribindex<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = HEAP32[$31>>2]|0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $33 = ($32>>>0)<($datasize>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $34 = $33 ? $32 : $datasize; //@line 234 "modules/graphics/opengl/Mesh.cpp"
 $35 = ((($this)) + 52|0); //@line 236 "modules/graphics/opengl/Mesh.cpp"
 $36 = HEAP32[$35>>2]|0; //@line 236 "modules/graphics/opengl/Mesh.cpp"
 __ZN4love8graphics6opengl8GLBuffer4bindEv($36); //@line 188 "modules/graphics/opengl/GLBuffer.h"
 $37 = HEAP32[$35>>2]|0; //@line 237 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 $38 = (invoke_ii(4096,($37|0))|0); //@line 237 "modules/graphics/opengl/Mesh.cpp"
 $39 = __THREW__; __THREW__ = 0;
 $40 = $39&1;
 if (!($40)) {
  $41 = (($38) + ($30)|0); //@line 239 "modules/graphics/opengl/Mesh.cpp"
  _memcpy(($41|0),($data|0),($34|0))|0; //@line 239 "modules/graphics/opengl/Mesh.cpp"
  $42 = HEAP32[$35>>2]|0; //@line 241 "modules/graphics/opengl/Mesh.cpp"
  __THREW__ = 0;
  invoke_viii(4097,($42|0),($30|0),($34|0)); //@line 241 "modules/graphics/opengl/Mesh.cpp"
  $43 = __THREW__; __THREW__ = 0;
  $44 = $43&1;
  if (!($44)) {
   __THREW__ = 0;
   invoke_vi(4089,($36|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   $45 = __THREW__; __THREW__ = 0;
   $46 = $45&1;
   if ($46) {
    $47 = ___cxa_find_matching_catch_3(0|0)|0;
    $48 = tempRet0;
    ___clang_call_terminate($47); //@line 196 "modules/graphics/opengl/GLBuffer.h"
    // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
   } else {
    STACKTOP = sp;return; //@line 242 "modules/graphics/opengl/Mesh.cpp"
   }
  }
 }
 $49 = ___cxa_find_matching_catch_2()|0;
 $50 = tempRet0;
 __THREW__ = 0;
 invoke_vi(4089,($36|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
 $51 = __THREW__; __THREW__ = 0;
 $52 = $51&1;
 if ($52) {
  $53 = ___cxa_find_matching_catch_3(0|0)|0;
  $54 = tempRet0;
  ___clang_call_terminate($53); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
 } else {
  $$0 = $49;$$01 = $50;
  ___resumeException($$0|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl4Mesh18getVertexAttributeEjiPvj($this,$vertindex,$attribindex,$data,$datasize) {
 $this = $this|0;
 $vertindex = $vertindex|0;
 $attribindex = $attribindex|0;
 $data = $data|0;
 $datasize = $datasize|0;
 var $$0 = 0, $$01 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i = 0, $i$02$i = 0, $offset$0$lcssa$i = 0, $offset$01$i = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($this)) + 56|0); //@line 246 "modules/graphics/opengl/Mesh.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 246 "modules/graphics/opengl/Mesh.cpp"
 $2 = ($1>>>0)>($vertindex>>>0); //@line 246 "modules/graphics/opengl/Mesh.cpp"
 if (!($2)) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 247 "modules/graphics/opengl/Mesh.cpp"
  $4 = (($vertindex) + 1)|0; //@line 247 "modules/graphics/opengl/Mesh.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer>>2] = $4; //@line 247 "modules/graphics/opengl/Mesh.cpp"
  invoke_viii(3349,($3|0),(437085|0),($vararg_buffer|0)); //@line 247 "modules/graphics/opengl/Mesh.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   ___cxa_throw(($3|0),(96|0),(344|0)); //@line 247 "modules/graphics/opengl/Mesh.cpp"
   // unreachable; //@line 247 "modules/graphics/opengl/Mesh.cpp"
  }
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  ___cxa_free_exception(($3|0)); //@line 247 "modules/graphics/opengl/Mesh.cpp"
  $$0 = $7;$$01 = $8;
  ___resumeException($$0|0);
  // unreachable;
 }
 $9 = ((($this)) + 8|0); //@line 249 "modules/graphics/opengl/Mesh.cpp"
 $10 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = HEAP32[$10>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = HEAP32[$9>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = (($11) - ($12))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = (($13|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = ($14|0)>($attribindex|0); //@line 249 "modules/graphics/opengl/Mesh.cpp"
 if (!($15)) {
  $16 = (___cxa_allocate_exception(16)|0); //@line 250 "modules/graphics/opengl/Mesh.cpp"
  $17 = (($attribindex) + 1)|0; //@line 250 "modules/graphics/opengl/Mesh.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer1>>2] = $17; //@line 250 "modules/graphics/opengl/Mesh.cpp"
  invoke_viii(3349,($16|0),(437111|0),($vararg_buffer1|0)); //@line 250 "modules/graphics/opengl/Mesh.cpp"
  $18 = __THREW__; __THREW__ = 0;
  $19 = $18&1;
  if (!($19)) {
   ___cxa_throw(($16|0),(96|0),(344|0)); //@line 250 "modules/graphics/opengl/Mesh.cpp"
   // unreachable; //@line 250 "modules/graphics/opengl/Mesh.cpp"
  }
  $20 = ___cxa_find_matching_catch_2()|0;
  $21 = tempRet0;
  ___cxa_free_exception(($16|0)); //@line 250 "modules/graphics/opengl/Mesh.cpp"
  $$0 = $20;$$01 = $21;
  ___resumeException($$0|0);
  // unreachable;
 }
 $22 = ((($this)) + 60|0); //@line 252 "modules/graphics/opengl/Mesh.cpp"
 $23 = HEAP32[$22>>2]|0; //@line 252 "modules/graphics/opengl/Mesh.cpp"
 $24 = ($attribindex|0)==(0); //@line 181 "modules/graphics/opengl/Mesh.cpp"
 $$phi$trans$insert = ((($this)) + 20|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($24) {
  $offset$0$lcssa$i = 0;
 } else {
  $i$02$i = 0;$offset$01$i = 0;
  while(1) {
   $25 = (($$pre) + ($i$02$i<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $26 = HEAP32[$25>>2]|0; //@line 182 "modules/graphics/opengl/Mesh.cpp"
   $27 = (($26) + ($offset$01$i))|0; //@line 182 "modules/graphics/opengl/Mesh.cpp"
   $28 = (($i$02$i) + 1)|0; //@line 181 "modules/graphics/opengl/Mesh.cpp"
   $exitcond$i = ($28|0)==($attribindex|0); //@line 181 "modules/graphics/opengl/Mesh.cpp"
   if ($exitcond$i) {
    $offset$0$lcssa$i = $27;
    break;
   } else {
    $i$02$i = $28;$offset$01$i = $27;
   }
  }
 }
 $29 = (($$pre) + ($attribindex<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = HEAP32[$29>>2]|0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $31 = ($30>>>0)<($datasize>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $32 = $31 ? $30 : $datasize; //@line 253 "modules/graphics/opengl/Mesh.cpp"
 $33 = ((($this)) + 52|0); //@line 256 "modules/graphics/opengl/Mesh.cpp"
 $34 = HEAP32[$33>>2]|0; //@line 256 "modules/graphics/opengl/Mesh.cpp"
 __ZN4love8graphics6opengl8GLBuffer4bindEv($34); //@line 188 "modules/graphics/opengl/GLBuffer.h"
 $35 = HEAP32[$33>>2]|0; //@line 257 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 $36 = (invoke_ii(4096,($35|0))|0); //@line 257 "modules/graphics/opengl/Mesh.cpp"
 $37 = __THREW__; __THREW__ = 0;
 $38 = $37&1;
 if ($38) {
  $46 = ___cxa_find_matching_catch_2()|0;
  $47 = tempRet0;
  __THREW__ = 0;
  invoke_vi(4089,($34|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  $48 = __THREW__; __THREW__ = 0;
  $49 = $48&1;
  if ($49) {
   $50 = ___cxa_find_matching_catch_3(0|0)|0;
   $51 = tempRet0;
   ___clang_call_terminate($50); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
  } else {
   $$0 = $46;$$01 = $47;
   ___resumeException($$0|0);
   // unreachable;
  }
 } else {
  $39 = Math_imul($23, $vertindex)|0; //@line 252 "modules/graphics/opengl/Mesh.cpp"
  $40 = (($offset$0$lcssa$i) + ($39))|0; //@line 252 "modules/graphics/opengl/Mesh.cpp"
  $41 = (($36) + ($40)|0); //@line 259 "modules/graphics/opengl/Mesh.cpp"
  _memcpy(($data|0),($41|0),($32|0))|0; //@line 259 "modules/graphics/opengl/Mesh.cpp"
  __THREW__ = 0;
  invoke_vi(4089,($34|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  $42 = __THREW__; __THREW__ = 0;
  $43 = $42&1;
  if ($43) {
   $44 = ___cxa_find_matching_catch_3(0|0)|0;
   $45 = tempRet0;
   ___clang_call_terminate($44); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
  } else {
   STACKTOP = sp;return ($32|0); //@line 262 "modules/graphics/opengl/Mesh.cpp"
  }
 }
 return (0)|0;
}
function __ZNK4love8graphics6opengl4Mesh14getVertexCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 56|0); //@line 266 "modules/graphics/opengl/Mesh.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 266 "modules/graphics/opengl/Mesh.cpp"
 return ($1|0); //@line 266 "modules/graphics/opengl/Mesh.cpp"
}
function __ZNK4love8graphics6opengl4Mesh15getVertexStrideEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 60|0); //@line 271 "modules/graphics/opengl/Mesh.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 271 "modules/graphics/opengl/Mesh.cpp"
 return ($1|0); //@line 271 "modules/graphics/opengl/Mesh.cpp"
}
function __ZNK4love8graphics6opengl4Mesh15getVertexFormatEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 276 "modules/graphics/opengl/Mesh.cpp"
 return ($0|0); //@line 276 "modules/graphics/opengl/Mesh.cpp"
}
function __ZNK4love8graphics6opengl4Mesh16getAttributeInfoEiRi($this,$attribindex,$components) {
 $this = $this|0;
 $attribindex = $attribindex|0;
 $components = $components|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($attribindex|0)<(0); //@line 281 "modules/graphics/opengl/Mesh.cpp"
 if (!($0)) {
  $1 = ((($this)) + 8|0); //@line 281 "modules/graphics/opengl/Mesh.cpp"
  $2 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $6 = (($5|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $7 = ($6|0)>($attribindex|0); //@line 281 "modules/graphics/opengl/Mesh.cpp"
  $8 = $4; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($7) {
   $15 = (((($8) + (($attribindex*20)|0)|0)) + 12|0); //@line 284 "modules/graphics/opengl/Mesh.cpp"
   $16 = HEAP32[$15>>2]|0; //@line 284 "modules/graphics/opengl/Mesh.cpp"
   $17 = (((($8) + (($attribindex*20)|0)|0)) + 16|0); //@line 285 "modules/graphics/opengl/Mesh.cpp"
   $18 = HEAP32[$17>>2]|0; //@line 285 "modules/graphics/opengl/Mesh.cpp"
   HEAP32[$components>>2] = $18; //@line 285 "modules/graphics/opengl/Mesh.cpp"
   STACKTOP = sp;return ($16|0); //@line 287 "modules/graphics/opengl/Mesh.cpp"
  }
 }
 $9 = (___cxa_allocate_exception(16)|0); //@line 282 "modules/graphics/opengl/Mesh.cpp"
 $10 = (($attribindex) + 1)|0; //@line 282 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $10; //@line 282 "modules/graphics/opengl/Mesh.cpp"
 invoke_viii(3349,($9|0),(437111|0),($vararg_buffer|0)); //@line 282 "modules/graphics/opengl/Mesh.cpp"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  ___cxa_free_exception(($9|0)); //@line 282 "modules/graphics/opengl/Mesh.cpp"
  ___resumeException($13|0);
  // unreachable;
 } else {
  ___cxa_throw(($9|0),(96|0),(344|0)); //@line 282 "modules/graphics/opengl/Mesh.cpp"
  // unreachable; //@line 282 "modules/graphics/opengl/Mesh.cpp"
 }
 return (0)|0;
}
function __ZNK4love8graphics6opengl4Mesh17getAttributeIndexERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($this,$name) {
 $this = $this|0;
 $name = $name|0;
 var $$mask = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__lhs_sz$03$i18 = 0, $__lp$04$i17 = 0, $__rp$05$i16 = 0, $i$06 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 292 "modules/graphics/opengl/Mesh.cpp"
 $1 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($2) - ($3))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($4|0)>(0); //@line 292 "modules/graphics/opengl/Mesh.cpp"
 $6 = $3; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($5)) {
  $48 = -1;
  return ($48|0); //@line 299 "modules/graphics/opengl/Mesh.cpp"
 }
 $7 = HEAP8[$name>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $7 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = ($8<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = ((($name)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = HEAP32[$10>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = $7&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = $12 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = $9 ? $13 : $11; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $15 = ((($name)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $16 = HEAP32[$15>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $17 = ((($name)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $18 = $9 ? $17 : $16; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $19 = ($14|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $20 = (($4|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $i$06 = 0;
 L4: while(1) {
  $21 = (($6) + (($i$06*20)|0)|0); //@line 294 "modules/graphics/opengl/Mesh.cpp"
  $22 = HEAP16[$21>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $23 = $22 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $24 = ($23<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $25 = (((($6) + (($i$06*20)|0)|0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $26 = HEAP32[$25>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $$mask = $22 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $27 = $$mask&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $28 = $27 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $29 = $24 ? $28 : $26; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $30 = ($29|0)==($14|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $31 = ($22&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $32 = $31&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  do {
   if ($30) {
    if (!($24)) {
     $42 = (((($6) + (($i$06*20)|0)|0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $43 = HEAP32[$42>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $44 = (_memcmp($43,$18,$14)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $45 = ($44|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($45) {
      $48 = $i$06;
      label = 12;
      break L4;
     } else {
      break;
     }
    }
    if ($19) {
     $48 = $i$06;
     label = 12;
     break L4;
    }
    $33 = HEAP8[$18>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $34 = ($32<<24>>24)==($33<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if ($34) {
     $35 = ((($21)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $__lhs_sz$03$i18 = $28;$__lp$04$i17 = $35;$__rp$05$i16 = $18;
     while(1) {
      $36 = (($__lhs_sz$03$i18) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $37 = ((($__lp$04$i17)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $38 = ($36|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($38) {
       $48 = $i$06;
       label = 12;
       break L4;
      }
      $39 = ((($__rp$05$i16)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $$pre = HEAP8[$37>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $40 = HEAP8[$39>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $41 = ($$pre<<24>>24)==($40<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($41) {
       $__lhs_sz$03$i18 = $36;$__lp$04$i17 = $37;$__rp$05$i16 = $39;
      } else {
       break;
      }
     }
    }
   }
  } while(0);
  $46 = (($i$06) + 1)|0; //@line 292 "modules/graphics/opengl/Mesh.cpp"
  $47 = ($46|0)<($20|0); //@line 292 "modules/graphics/opengl/Mesh.cpp"
  if ($47) {
   $i$06 = $46;
  } else {
   $48 = -1;
   label = 12;
   break;
  }
 }
 if ((label|0) == 12) {
  return ($48|0); //@line 299 "modules/graphics/opengl/Mesh.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl4Mesh19setAttributeEnabledERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEEb($this,$name,$enable) {
 $this = $this|0;
 $name = $name|0;
 $enable = $enable|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 32|0); //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $1 = (__ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl4Mesh17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE4findIS7_EENS_15__hash_iteratorIPNS_11__hash_nodeISD_PvEEEERKT_($0,$name)|0); //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $2 = ($1|0)==(0|0); //@line 189 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($2)) {
  $15 = ((($1)) + 28|0); //@line 308 "modules/graphics/opengl/Mesh.cpp"
  $16 = $enable&1; //@line 308 "modules/graphics/opengl/Mesh.cpp"
  HEAP8[$15>>0] = $16; //@line 308 "modules/graphics/opengl/Mesh.cpp"
  STACKTOP = sp;return; //@line 309 "modules/graphics/opengl/Mesh.cpp"
 }
 $3 = (___cxa_allocate_exception(16)|0); //@line 306 "modules/graphics/opengl/Mesh.cpp"
 $4 = HEAP8[$name>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = $4 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = ($5<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($name)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = HEAP32[$7>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = ((($name)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $6 ? $9 : $8; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $10; //@line 306 "modules/graphics/opengl/Mesh.cpp"
 invoke_viii(3349,($3|0),(437146|0),($vararg_buffer|0)); //@line 306 "modules/graphics/opengl/Mesh.cpp"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  ___cxa_free_exception(($3|0)); //@line 306 "modules/graphics/opengl/Mesh.cpp"
  ___resumeException($13|0);
  // unreachable;
 } else {
  ___cxa_throw(($3|0),(96|0),(344|0)); //@line 306 "modules/graphics/opengl/Mesh.cpp"
  // unreachable; //@line 306 "modules/graphics/opengl/Mesh.cpp"
 }
}
function __ZNK4love8graphics6opengl4Mesh18isAttributeEnabledERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($this,$name) {
 $this = $this|0;
 $name = $name|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 32|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $1 = (__ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl4Mesh17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeISD_PvEEEERKT_($0,$name)|0); //@line 974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $2 = ($1|0)==(0|0); //@line 336 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($2)) {
  $15 = ((($1)) + 28|0); //@line 318 "modules/graphics/opengl/Mesh.cpp"
  $16 = HEAP8[$15>>0]|0; //@line 318 "modules/graphics/opengl/Mesh.cpp"
  $17 = ($16<<24>>24)!=(0); //@line 318 "modules/graphics/opengl/Mesh.cpp"
  STACKTOP = sp;return ($17|0); //@line 318 "modules/graphics/opengl/Mesh.cpp"
 }
 $3 = (___cxa_allocate_exception(16)|0); //@line 316 "modules/graphics/opengl/Mesh.cpp"
 $4 = HEAP8[$name>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = $4 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = ($5<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($name)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = HEAP32[$7>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = ((($name)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $6 ? $9 : $8; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $10; //@line 316 "modules/graphics/opengl/Mesh.cpp"
 invoke_viii(3349,($3|0),(437146|0),($vararg_buffer|0)); //@line 316 "modules/graphics/opengl/Mesh.cpp"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  ___cxa_free_exception(($3|0)); //@line 316 "modules/graphics/opengl/Mesh.cpp"
  ___resumeException($13|0);
  // unreachable;
 } else {
  ___cxa_throw(($3|0),(96|0),(344|0)); //@line 316 "modules/graphics/opengl/Mesh.cpp"
  // unreachable; //@line 316 "modules/graphics/opengl/Mesh.cpp"
 }
 return (0)|0;
}
function __ZNKSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl4Mesh17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE4findIS7_EENS_21__hash_const_iteratorIPNS_11__hash_nodeISD_PvEEEERKT_($this,$__k) {
 $this = $this|0;
 $__k = $__k|0;
 var $$0$lcssa$i$i$i$i = 0, $$03$i$i$i$i = 0, $$lcssa = 0, $$mask$i$i = 0, $$mask$i$i$us = 0, $$pre$i$i = 0, $$pre$i$i$us = 0, $$sroa$0$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__data$0$lcssa$i$i$i$i = 0, $__data$02$i$i$i$i = 0, $__h$0$lcssa$i$i$i$i = 0, $__h$01$i$i$i$i = 0, $__h$1$i$i$i$i = 0, $__h$2$i$i$i$i = 0, $__h$3$i$i$i$i = 0, $__lhs_sz$03$i$i3$i = 0, $__lhs_sz$03$i$i3$i$us = 0, $__lp$04$i$i2$i = 0, $__lp$04$i$i2$i$us = 0;
 var $__nd$0 = 0, $__nd$0$us = 0, $__nd$08 = 0, $__nd$09 = 0, $__nd$09$us = 0, $__rp$05$i$i1$i = 0, $__rp$05$i$i1$i$us = 0, $scevgep$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$__k>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $1 = $0 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = ($1<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ((($__k)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = HEAP32[$3>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($__k)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = $2 ? $5 : $4; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($__k)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = HEAP32[$7>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $0&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $9 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $2 ? $10 : $8; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = ($11>>>0)>(3); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($12) {
  $$03$i$i$i$i = $11;$__data$02$i$i$i$i = $6;$__h$01$i$i$i$i = $11;
  while(1) {
   $13 = HEAPU8[$__data$02$i$i$i$i>>0]|(HEAPU8[$__data$02$i$i$i$i+1>>0]<<8)|(HEAPU8[$__data$02$i$i$i$i+2>>0]<<16)|(HEAPU8[$__data$02$i$i$i$i+3>>0]<<24); //@line 3073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $14 = Math_imul($13, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $15 = $14 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $16 = $15 ^ $14; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $17 = Math_imul($16, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $18 = Math_imul($__h$01$i$i$i$i, 1540483477)|0; //@line 3104 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $19 = $17 ^ $18; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $20 = ((($__data$02$i$i$i$i)) + 4|0); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $21 = (($$03$i$i$i$i) + -4)|0; //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $22 = ($21>>>0)>(3); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if ($22) {
    $$03$i$i$i$i = $21;$__data$02$i$i$i$i = $20;$__h$01$i$i$i$i = $19;
   } else {
    $$lcssa = $19;
    break;
   }
  }
  $23 = (($11) + -4)|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = $23 & -4; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $25 = (($24) + 4)|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (($23) - ($24))|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $scevgep$i$i$i$i = (($6) + ($25)|0); //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $$0$lcssa$i$i$i$i = $26;$__data$0$lcssa$i$i$i$i = $scevgep$i$i$i$i;$__h$0$lcssa$i$i$i$i = $$lcssa;
 } else {
  $$0$lcssa$i$i$i$i = $11;$__data$0$lcssa$i$i$i$i = $6;$__h$0$lcssa$i$i$i$i = $11;
 }
 switch ($$0$lcssa$i$i$i$i|0) {
 case 3:  {
  $27 = ((($__data$0$lcssa$i$i$i$i)) + 2|0); //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $28 = HEAP8[$27>>0]|0; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $29 = $28&255; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $30 = $29 << 16; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $31 = $30 ^ $__h$0$lcssa$i$i$i$i; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$1$i$i$i$i = $31;
  label = 6;
  break;
 }
 case 2:  {
  $__h$1$i$i$i$i = $__h$0$lcssa$i$i$i$i;
  label = 6;
  break;
 }
 case 1:  {
  $__h$2$i$i$i$i = $__h$0$lcssa$i$i$i$i;
  label = 7;
  break;
 }
 default: {
  $__h$3$i$i$i$i = $__h$0$lcssa$i$i$i$i;
 }
 }
 if ((label|0) == 6) {
  $32 = ((($__data$0$lcssa$i$i$i$i)) + 1|0); //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $33 = HEAP8[$32>>0]|0; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $34 = $33&255; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $35 = $34 << 8; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $36 = $35 ^ $__h$1$i$i$i$i; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$2$i$i$i$i = $36;
  label = 7;
 }
 if ((label|0) == 7) {
  $37 = HEAP8[$__data$0$lcssa$i$i$i$i>>0]|0; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $38 = $37&255; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $39 = $38 ^ $__h$2$i$i$i$i; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $40 = Math_imul($39, 1540483477)|0; //@line 3115 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$3$i$i$i$i = $40;
 }
 $41 = $__h$3$i$i$i$i >>> 13; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $42 = $41 ^ $__h$3$i$i$i$i; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $43 = Math_imul($42, 1540483477)|0; //@line 3118 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $44 = $43 >>> 15; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $45 = $44 ^ $43; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $46 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = HEAP32[$46>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $48 = ($47|0)==(0); //@line 2047 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($48) {
  $$sroa$0$0 = 0;
  return ($$sroa$0$0|0); //@line 2068 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $49 = (($47) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $50 = $49 & $47; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $51 = ($50|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($51) {
  $52 = $45 & $49; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $56 = $52;
 } else {
  $53 = (($45>>>0) % ($47>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $56 = $53;
 }
 $54 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $55 = (($54) + ($56<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $57 = HEAP32[$55>>2]|0; //@line 2050 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $58 = ($57|0)==(0|0); //@line 2051 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($58) {
  $$sroa$0$0 = 0;
  return ($$sroa$0$0|0); //@line 2068 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $__nd$08 = HEAP32[$57>>2]|0; //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $59 = ($__nd$08|0)==(0|0); //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($59) {
  $$sroa$0$0 = 0;
  return ($$sroa$0$0|0); //@line 2068 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $60 = ($11|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($51) {
  $__nd$09$us = $__nd$08;
  L27: while(1) {
   $61 = ((($__nd$09$us)) + 4|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $62 = HEAP32[$61>>2]|0; //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $63 = $62 & $49; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $64 = ($63|0)==($56|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($64)) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   }
   $65 = ((($__nd$09$us)) + 8|0); //@line 2057 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $66 = HEAP16[$65>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $67 = $66 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $68 = ($67<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $69 = ((($__nd$09$us)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $70 = HEAP32[$69>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$mask$i$i$us = $66 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $71 = $$mask$i$i$us&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $72 = $71 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $73 = $68 ? $72 : $70; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $74 = ($73|0)==($11|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $75 = ($66&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $76 = $75&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   do {
    if ($74) {
     if (!($68)) {
      $77 = ((($__nd$09$us)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $78 = HEAP32[$77>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $79 = (_memcmp($78,$6,$11)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $80 = ($79|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($80) {
       $$sroa$0$0 = $__nd$09$us;
       label = 35;
       break L27;
      } else {
       break;
      }
     }
     if ($60) {
      $$sroa$0$0 = $__nd$09$us;
      label = 35;
      break L27;
     }
     $81 = HEAP8[$6>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $82 = ($76<<24>>24)==($81<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($82) {
      $83 = ((($65)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $__lhs_sz$03$i$i3$i$us = $72;$__lp$04$i$i2$i$us = $83;$__rp$05$i$i1$i$us = $6;
      while(1) {
       $84 = (($__lhs_sz$03$i$i3$i$us) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $85 = ((($__lp$04$i$i2$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $86 = ($84|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($86) {
        $$sroa$0$0 = $__nd$09$us;
        label = 35;
        break L27;
       }
       $87 = ((($__rp$05$i$i1$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$pre$i$i$us = HEAP8[$85>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $88 = HEAP8[$87>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $89 = ($$pre$i$i$us<<24>>24)==($88<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($89) {
        $__lhs_sz$03$i$i3$i$us = $84;$__lp$04$i$i2$i$us = $85;$__rp$05$i$i1$i$us = $87;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   $__nd$0$us = HEAP32[$__nd$09$us>>2]|0; //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $90 = ($__nd$0$us|0)==(0|0); //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($90) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   } else {
    $__nd$09$us = $__nd$0$us;
   }
  }
  if ((label|0) == 35) {
   return ($$sroa$0$0|0); //@line 2068 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 } else {
  $__nd$09 = $__nd$08;
  L47: while(1) {
   $91 = ((($__nd$09)) + 4|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $92 = HEAP32[$91>>2]|0; //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $93 = (($92>>>0) % ($47>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $94 = ($93|0)==($56|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($94)) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   }
   $95 = ((($__nd$09)) + 8|0); //@line 2057 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $96 = HEAP16[$95>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $97 = $96 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $98 = ($97<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $99 = ((($__nd$09)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $100 = HEAP32[$99>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$mask$i$i = $96 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $101 = $$mask$i$i&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $102 = $101 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $103 = $98 ? $102 : $100; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $104 = ($103|0)==($11|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $105 = ($96&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $106 = $105&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   do {
    if ($104) {
     if (!($98)) {
      $116 = ((($__nd$09)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $117 = HEAP32[$116>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $118 = (_memcmp($117,$6,$11)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $119 = ($118|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($119) {
       $$sroa$0$0 = $__nd$09;
       label = 35;
       break L47;
      } else {
       break;
      }
     }
     if ($60) {
      $$sroa$0$0 = $__nd$09;
      label = 35;
      break L47;
     }
     $107 = HEAP8[$6>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $108 = ($106<<24>>24)==($107<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($108) {
      $109 = ((($95)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $__lhs_sz$03$i$i3$i = $102;$__lp$04$i$i2$i = $109;$__rp$05$i$i1$i = $6;
      while(1) {
       $110 = (($__lhs_sz$03$i$i3$i) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $111 = ((($__lp$04$i$i2$i)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $112 = ($110|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($112) {
        $$sroa$0$0 = $__nd$09;
        label = 35;
        break L47;
       }
       $113 = ((($__rp$05$i$i1$i)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$pre$i$i = HEAP8[$111>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $114 = HEAP8[$113>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $115 = ($$pre$i$i<<24>>24)==($114<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($115) {
        $__lhs_sz$03$i$i3$i = $110;$__lp$04$i$i2$i = $111;$__rp$05$i$i1$i = $113;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   $__nd$0 = HEAP32[$__nd$09>>2]|0; //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $120 = ($__nd$0|0)==(0|0); //@line 2053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($120) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   } else {
    $__nd$09 = $__nd$0;
   }
  }
  if ((label|0) == 35) {
   return ($$sroa$0$0|0); //@line 2068 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl4Mesh15attachAttributeERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEEPS2_($this,$name,$mesh) {
 $this = $this|0;
 $name = $name|0;
 $mesh = $mesh|0;
 var $$0 = 0, $$02 = 0, $$mask$i = 0, $$mask$i$us = 0, $$pre$i = 0, $$pre$phi20Z2D = 0, $$pre$phi24Z2D = 0, $$pre19 = 0, $$pre23 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $11 = 0, $12 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$010 = 0, $__begin$sroa$0$011 = 0, $__lhs_sz$03$i$i9 = 0, $__lp$04$i$i8 = 0;
 var $__rp$05$i$i7 = 0, $extract$t = 0, $i$06$i = 0, $i$06$i$us = 0, $i$06$i6 = 0, $newattrib$sroa$10 = 0, $not$ = 0, $oldattrib$sroa$0$0 = 0, $oldattrib$sroa$81$0$off0 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $newattrib$sroa$10 = sp + 12|0;
 $0 = ($mesh|0)!=($this|0); //@line 323 "modules/graphics/opengl/Mesh.cpp"
 L1: do {
  if ($0) {
   $1 = ((($mesh)) + 40|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $__begin$sroa$0$010 = HEAP32[$1>>2]|0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $2 = ($__begin$sroa$0$010|0)==(0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($2)) {
    $__begin$sroa$0$011 = $__begin$sroa$0$010;
    while(1) {
     $3 = $__begin$sroa$0$011; //@line 189 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $4 = ((($3)) + 20|0); //@line 329 "modules/graphics/opengl/Mesh.cpp"
     $5 = HEAP32[$4>>2]|0; //@line 329 "modules/graphics/opengl/Mesh.cpp"
     $6 = ($5|0)==($mesh|0); //@line 329 "modules/graphics/opengl/Mesh.cpp"
     if (!($6)) {
      break;
     }
     $12 = $__begin$sroa$0$011; //@line 174 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $__begin$sroa$0$0 = HEAP32[$12>>2]|0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $13 = ($__begin$sroa$0$0|0)==(0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     if ($13) {
      break L1;
     } else {
      $__begin$sroa$0$011 = $__begin$sroa$0$0;
     }
    }
    $7 = (___cxa_allocate_exception(16)|0); //@line 330 "modules/graphics/opengl/Mesh.cpp"
    __THREW__ = 0;
    invoke_viii(3349,($7|0),(437205|0),($vararg_buffer|0)); //@line 330 "modules/graphics/opengl/Mesh.cpp"
    $8 = __THREW__; __THREW__ = 0;
    $9 = $8&1;
    if (!($9)) {
     ___cxa_throw(($7|0),(96|0),(344|0)); //@line 330 "modules/graphics/opengl/Mesh.cpp"
     // unreachable; //@line 330 "modules/graphics/opengl/Mesh.cpp"
    }
    $10 = ___cxa_find_matching_catch_2()|0;
    $11 = tempRet0;
    ___cxa_free_exception(($7|0)); //@line 330 "modules/graphics/opengl/Mesh.cpp"
    $$0 = $11;$$02 = $10;
    ___resumeException($$02|0);
    // unreachable;
   }
  }
 } while(0);
 ;HEAP8[$newattrib$sroa$10>>0]=0|0;HEAP8[$newattrib$sroa$10+1>>0]=0|0;HEAP8[$newattrib$sroa$10+2>>0]=0|0; //@line 335 "modules/graphics/opengl/Mesh.cpp"
 $14 = ((($this)) + 32|0); //@line 337 "modules/graphics/opengl/Mesh.cpp"
 $15 = (__ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl4Mesh17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE4findIS7_EENS_15__hash_iteratorIPNS_11__hash_nodeISD_PvEEEERKT_($14,$name)|0); //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $16 = ($15|0)==(0|0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($16) {
  $oldattrib$sroa$0$0 = 0;$oldattrib$sroa$81$0$off0 = 0;
 } else {
  $17 = ((($15)) + 20|0); //@line 339 "modules/graphics/opengl/Mesh.cpp"
  $18 = HEAP32[$17>>2]|0; //@line 339 "modules/graphics/opengl/Mesh.cpp"
  $19 = ((($15)) + 28|0); //@line 339 "modules/graphics/opengl/Mesh.cpp"
  $20 = HEAP8[$19>>0]|0; //@line 339 "modules/graphics/opengl/Mesh.cpp"
  $21 = $20 & 1; //@line 339 "modules/graphics/opengl/Mesh.cpp"
  $extract$t = ($21<<24>>24)!=(0); //@line 339 "modules/graphics/opengl/Mesh.cpp"
  $oldattrib$sroa$0$0 = $18;$oldattrib$sroa$81$0$off0 = $extract$t;
 }
 $not$ = ($oldattrib$sroa$0$0|0)==(0|0); //@line 342 "modules/graphics/opengl/Mesh.cpp"
 $22 = $oldattrib$sroa$81$0$off0 | $not$; //@line 342 "modules/graphics/opengl/Mesh.cpp"
 $23 = $22&1; //@line 342 "modules/graphics/opengl/Mesh.cpp"
 $24 = ((($mesh)) + 8|0); //@line 292 "modules/graphics/opengl/Mesh.cpp"
 $25 = ((($mesh)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $26 = HEAP32[$25>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = HEAP32[$24>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $28 = (($26) - ($27))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $29 = ($28|0)>(0); //@line 292 "modules/graphics/opengl/Mesh.cpp"
 $30 = $27; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 L15: do {
  if ($29) {
   $31 = HEAP8[$name>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $32 = $31 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $33 = ($32<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $34 = ((($name)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $35 = HEAP32[$34>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $36 = $31&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $37 = $36 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $38 = $33 ? $37 : $35; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $39 = ((($name)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $40 = HEAP32[$39>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $41 = ((($name)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $42 = $33 ? $41 : $40; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $43 = ($38|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $44 = (($28|0) / 20)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   L17: do {
    if ($43) {
     $i$06$i$us = 0;
     while(1) {
      $45 = (($30) + (($i$06$i$us*20)|0)|0); //@line 294 "modules/graphics/opengl/Mesh.cpp"
      $46 = HEAP16[$45>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $47 = $46 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $48 = ($47<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $49 = (((($30) + (($i$06$i$us*20)|0)|0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $50 = HEAP32[$49>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $$mask$i$us = $46 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $51 = $$mask$i$us&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $52 = $51 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $53 = $48 ? $52 : $50; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $54 = ($53|0)==(0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($54) {
       $i$06$i6 = $i$06$i$us;
       break L17;
      }
      $55 = (($i$06$i$us) + 1)|0; //@line 292 "modules/graphics/opengl/Mesh.cpp"
      $56 = ($55|0)<($44|0); //@line 292 "modules/graphics/opengl/Mesh.cpp"
      if ($56) {
       $i$06$i$us = $55;
      } else {
       $$pre$phi20Z2D = $39;$$pre$phi24Z2D = $41;
       break L15;
      }
     }
    } else {
     $i$06$i = 0;
     while(1) {
      $57 = (($30) + (($i$06$i*20)|0)|0); //@line 294 "modules/graphics/opengl/Mesh.cpp"
      $58 = HEAP16[$57>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $59 = $58 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $60 = ($59<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $61 = (((($30) + (($i$06$i*20)|0)|0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $62 = HEAP32[$61>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $$mask$i = $58 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $63 = $$mask$i&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $64 = $63 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $65 = $60 ? $64 : $62; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $66 = ($65|0)==($38|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $67 = ($58&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $68 = $67&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      do {
       if ($66) {
        if (!($60)) {
         $78 = (((($30) + (($i$06$i*20)|0)|0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $79 = HEAP32[$78>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $80 = (_memcmp($79,$42,$38)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $81 = ($80|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         if ($81) {
          $i$06$i6 = $i$06$i;
          break L17;
         } else {
          break;
         }
        }
        $69 = HEAP8[$42>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $70 = ($68<<24>>24)==($69<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if ($70) {
         $71 = ((($57)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $__lhs_sz$03$i$i9 = $64;$__lp$04$i$i8 = $71;$__rp$05$i$i7 = $42;
         while(1) {
          $72 = (($__lhs_sz$03$i$i9) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $73 = ((($__lp$04$i$i8)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $74 = ($72|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          if ($74) {
           $i$06$i6 = $i$06$i;
           break L17;
          }
          $75 = ((($__rp$05$i$i7)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $$pre$i = HEAP8[$73>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $76 = HEAP8[$75>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $77 = ($$pre$i<<24>>24)==($76<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          if ($77) {
           $__lhs_sz$03$i$i9 = $72;$__lp$04$i$i8 = $73;$__rp$05$i$i7 = $75;
          } else {
           break;
          }
         }
        }
       }
      } while(0);
      $82 = (($i$06$i) + 1)|0; //@line 292 "modules/graphics/opengl/Mesh.cpp"
      $83 = ($82|0)<($44|0); //@line 292 "modules/graphics/opengl/Mesh.cpp"
      if ($83) {
       $i$06$i = $82;
      } else {
       $$pre$phi20Z2D = $39;$$pre$phi24Z2D = $41;
       break L15;
      }
     }
    }
   } while(0);
   $84 = ($i$06$i6|0)<(0); //@line 345 "modules/graphics/opengl/Mesh.cpp"
   if ($84) {
    $$pre$phi20Z2D = $39;$$pre$phi24Z2D = $41;
   } else {
    if ($0) {
     $95 = HEAP32[$mesh>>2]|0; //@line 349 "modules/graphics/opengl/Mesh.cpp"
     $96 = ((($95)) + 8|0); //@line 349 "modules/graphics/opengl/Mesh.cpp"
     $97 = HEAP32[$96>>2]|0; //@line 349 "modules/graphics/opengl/Mesh.cpp"
     FUNCTION_TABLE_vi[$97 & 8191]($mesh); //@line 349 "modules/graphics/opengl/Mesh.cpp"
    }
    $98 = (__ZNSt3__113unordered_mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl4Mesh17AttachedAttributeENS_4hashIS6_EENS_8equal_toIS6_EENS4_INS_4pairIKS6_SB_EEEEEixERSH_($14,$name)|0); //@line 351 "modules/graphics/opengl/Mesh.cpp"
    HEAP32[$98>>2] = $mesh; //@line 351 "modules/graphics/opengl/Mesh.cpp"
    $99 = ((($98)) + 4|0); //@line 351 "modules/graphics/opengl/Mesh.cpp"
    HEAP32[$99>>2] = $i$06$i6; //@line 351 "modules/graphics/opengl/Mesh.cpp"
    $100 = ((($98)) + 8|0); //@line 351 "modules/graphics/opengl/Mesh.cpp"
    HEAP8[$100>>0] = $23; //@line 351 "modules/graphics/opengl/Mesh.cpp"
    $101 = ((($98)) + 9|0); //@line 351 "modules/graphics/opengl/Mesh.cpp"
    ;HEAP8[$101>>0]=HEAP8[$newattrib$sroa$10>>0]|0;HEAP8[$101+1>>0]=HEAP8[$newattrib$sroa$10+1>>0]|0;HEAP8[$101+2>>0]=HEAP8[$newattrib$sroa$10+2>>0]|0; //@line 351 "modules/graphics/opengl/Mesh.cpp"
    $102 = ($oldattrib$sroa$0$0|0)==($this|0); //@line 353 "modules/graphics/opengl/Mesh.cpp"
    $or$cond = $not$ | $102; //@line 353 "modules/graphics/opengl/Mesh.cpp"
    if ($or$cond) {
     STACKTOP = sp;return; //@line 355 "modules/graphics/opengl/Mesh.cpp"
    }
    $103 = HEAP32[$oldattrib$sroa$0$0>>2]|0; //@line 354 "modules/graphics/opengl/Mesh.cpp"
    $104 = ((($103)) + 12|0); //@line 354 "modules/graphics/opengl/Mesh.cpp"
    $105 = HEAP32[$104>>2]|0; //@line 354 "modules/graphics/opengl/Mesh.cpp"
    FUNCTION_TABLE_vi[$105 & 8191]($oldattrib$sroa$0$0); //@line 354 "modules/graphics/opengl/Mesh.cpp"
    STACKTOP = sp;return; //@line 355 "modules/graphics/opengl/Mesh.cpp"
   }
  } else {
   $$pre19 = ((($name)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$pre23 = ((($name)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$pre$phi20Z2D = $$pre19;$$pre$phi24Z2D = $$pre23;
  }
 } while(0);
 $85 = (___cxa_allocate_exception(16)|0); //@line 346 "modules/graphics/opengl/Mesh.cpp"
 $86 = HEAP8[$name>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $87 = $86 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $88 = ($87<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $89 = HEAP32[$$pre$phi20Z2D>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $90 = $88 ? $$pre$phi24Z2D : $89; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 HEAP32[$vararg_buffer1>>2] = $90; //@line 346 "modules/graphics/opengl/Mesh.cpp"
 invoke_viii(3349,($85|0),(439154|0),($vararg_buffer1|0)); //@line 346 "modules/graphics/opengl/Mesh.cpp"
 $91 = __THREW__; __THREW__ = 0;
 $92 = $91&1;
 if (!($92)) {
  ___cxa_throw(($85|0),(96|0),(344|0)); //@line 346 "modules/graphics/opengl/Mesh.cpp"
  // unreachable; //@line 346 "modules/graphics/opengl/Mesh.cpp"
 }
 $93 = ___cxa_find_matching_catch_2()|0;
 $94 = tempRet0;
 ___cxa_free_exception(($85|0)); //@line 346 "modules/graphics/opengl/Mesh.cpp"
 $$0 = $94;$$02 = $93;
 ___resumeException($$02|0);
 // unreachable;
}
function __ZN4love8graphics6opengl4Mesh13mapVertexDataEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 359 "modules/graphics/opengl/Mesh.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 359 "modules/graphics/opengl/Mesh.cpp"
 __ZN4love8graphics6opengl8GLBuffer4bindEv($1); //@line 188 "modules/graphics/opengl/GLBuffer.h"
 $2 = HEAP32[$0>>2]|0; //@line 360 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 $3 = (invoke_ii(4096,($2|0))|0); //@line 360 "modules/graphics/opengl/Mesh.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  __THREW__ = 0;
  invoke_vi(4089,($1|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if ($13) {
   $14 = ___cxa_find_matching_catch_3(0|0)|0;
   $15 = tempRet0;
   ___clang_call_terminate($14); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
  } else {
   ___resumeException($10|0);
   // unreachable;
  }
 } else {
  __THREW__ = 0;
  invoke_vi(4089,($1|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if ($7) {
   $8 = ___cxa_find_matching_catch_3(0|0)|0;
   $9 = tempRet0;
   ___clang_call_terminate($8); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
  } else {
   return ($3|0); //@line 361 "modules/graphics/opengl/Mesh.cpp"
  }
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl4Mesh15unmapVertexDataEjj($this,$modifiedoffset,$modifiedsize) {
 $this = $this|0;
 $modifiedoffset = $modifiedoffset|0;
 $modifiedsize = $modifiedsize|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 365 "modules/graphics/opengl/Mesh.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 365 "modules/graphics/opengl/Mesh.cpp"
 __ZN4love8graphics6opengl8GLBuffer4bindEv($1); //@line 188 "modules/graphics/opengl/GLBuffer.h"
 $2 = HEAP32[$0>>2]|0; //@line 366 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 invoke_viii(4097,($2|0),($modifiedoffset|0),($modifiedsize|0)); //@line 366 "modules/graphics/opengl/Mesh.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  $5 = HEAP32[$0>>2]|0; //@line 367 "modules/graphics/opengl/Mesh.cpp"
  __THREW__ = 0;
  invoke_vi(4088,($5|0)); //@line 367 "modules/graphics/opengl/Mesh.cpp"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if (!($7)) {
   __THREW__ = 0;
   invoke_vi(4089,($1|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   $8 = __THREW__; __THREW__ = 0;
   $9 = $8&1;
   if ($9) {
    $10 = ___cxa_find_matching_catch_3(0|0)|0;
    $11 = tempRet0;
    ___clang_call_terminate($10); //@line 196 "modules/graphics/opengl/GLBuffer.h"
    // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
   } else {
    return; //@line 368 "modules/graphics/opengl/Mesh.cpp"
   }
  }
 }
 $12 = ___cxa_find_matching_catch_2()|0;
 $13 = tempRet0;
 __THREW__ = 0;
 invoke_vi(4089,($1|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if ($15) {
  $16 = ___cxa_find_matching_catch_3(0|0)|0;
  $17 = tempRet0;
  ___clang_call_terminate($16); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
 } else {
  ___resumeException($12|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl4Mesh5flushEv($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 373 "modules/graphics/opengl/Mesh.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 373 "modules/graphics/opengl/Mesh.cpp"
 __ZN4love8graphics6opengl8GLBuffer4bindEv($1); //@line 188 "modules/graphics/opengl/GLBuffer.h"
 $2 = HEAP32[$0>>2]|0; //@line 374 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 invoke_vi(4088,($2|0)); //@line 374 "modules/graphics/opengl/Mesh.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  __THREW__ = 0;
  invoke_vi(4089,($1|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  $21 = __THREW__; __THREW__ = 0;
  $22 = $21&1;
  if ($22) {
   $23 = ___cxa_find_matching_catch_3(0|0)|0;
   $24 = tempRet0;
   ___clang_call_terminate($23); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
  } else {
   $$0 = $20;$$01 = $19;
   ___resumeException($$01|0);
   // unreachable;
  }
 }
 __THREW__ = 0;
 invoke_vi(4089,($1|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_3(0|0)|0;
  $8 = tempRet0;
  ___clang_call_terminate($7); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
 }
 $9 = ((($this)) + 68|0); //@line 377 "modules/graphics/opengl/Mesh.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 377 "modules/graphics/opengl/Mesh.cpp"
 $11 = ($10|0)==(0|0); //@line 377 "modules/graphics/opengl/Mesh.cpp"
 if ($11) {
  return; //@line 382 "modules/graphics/opengl/Mesh.cpp"
 }
 __ZN4love8graphics6opengl8GLBuffer4bindEv($10); //@line 188 "modules/graphics/opengl/GLBuffer.h"
 $12 = HEAP32[$9>>2]|0; //@line 380 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 invoke_vi(4088,($12|0)); //@line 380 "modules/graphics/opengl/Mesh.cpp"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if ($14) {
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  __THREW__ = 0;
  invoke_vi(4089,($10|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  $27 = __THREW__; __THREW__ = 0;
  $28 = $27&1;
  if ($28) {
   $29 = ___cxa_find_matching_catch_3(0|0)|0;
   $30 = tempRet0;
   ___clang_call_terminate($29); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
  } else {
   $$0 = $26;$$01 = $25;
   ___resumeException($$01|0);
   // unreachable;
  }
 } else {
  __THREW__ = 0;
  invoke_vi(4089,($10|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  $15 = __THREW__; __THREW__ = 0;
  $16 = $15&1;
  if ($16) {
   $17 = ___cxa_find_matching_catch_3(0|0)|0;
   $18 = tempRet0;
   ___clang_call_terminate($17); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
  } else {
   return; //@line 382 "modules/graphics/opengl/Mesh.cpp"
  }
 }
}
function __ZN4love8graphics6opengl4Mesh12setVertexMapERKNSt3__16vectorIjNS3_9allocatorIjEEEE($this,$map) {
 $this = $this|0;
 $map = $map|0;
 var $$ = 0, $$$i = 0, $$0 = 0, $$02 = 0, $$1 = 0, $$13 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index6Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i$02$i = 0, $i$02$i$lcssa = 0, $i$02$i9 = 0, $i$02$i9$lcssa = 0, $or$cond4 = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($this)) + 56|0); //@line 266 "modules/graphics/opengl/Mesh.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 266 "modules/graphics/opengl/Mesh.cpp"
 $2 = ($1>>>0)>(65535); //@line 689 "modules/graphics/opengl/Mesh.cpp"
 $$$i = $2 ? 5125 : 5123; //@line 690 "modules/graphics/opengl/Mesh.cpp"
 $3 = ((($map)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$map>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = (($4) - ($5))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = $6 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $$ = $2 ? 4 : 2; //@line 704 "modules/graphics/opengl/Mesh.cpp"
 $8 = Math_imul($7, $$)|0; //@line 408 "modules/graphics/opengl/Mesh.cpp"
 $9 = ((($this)) + 68|0); //@line 410 "modules/graphics/opengl/Mesh.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 410 "modules/graphics/opengl/Mesh.cpp"
 $11 = ($10|0)==(0|0); //@line 410 "modules/graphics/opengl/Mesh.cpp"
 if ($11) {
  label = 4;
 } else {
  $12 = ((($10)) + 8|0); //@line 76 "modules/graphics/opengl/GLBuffer.h"
  $13 = HEAP32[$12>>2]|0; //@line 76 "modules/graphics/opengl/GLBuffer.h"
  $14 = ($8>>>0)>($13>>>0); //@line 410 "modules/graphics/opengl/Mesh.cpp"
  if ($14) {
   $15 = HEAP32[$10>>2]|0; //@line 412 "modules/graphics/opengl/Mesh.cpp"
   $16 = ((($15)) + 4|0); //@line 412 "modules/graphics/opengl/Mesh.cpp"
   $17 = HEAP32[$16>>2]|0; //@line 412 "modules/graphics/opengl/Mesh.cpp"
   FUNCTION_TABLE_vi[$17 & 8191]($10); //@line 412 "modules/graphics/opengl/Mesh.cpp"
   HEAP32[$9>>2] = 0; //@line 413 "modules/graphics/opengl/Mesh.cpp"
   label = 4;
  } else {
   $35 = $10;
  }
 }
 do {
  if ((label|0) == 4) {
   $18 = ($8|0)==(0); //@line 416 "modules/graphics/opengl/Mesh.cpp"
   if ($18) {
    $35 = 0;
   } else {
    $19 = (__Znwj(40)|0); //@line 417 "modules/graphics/opengl/Mesh.cpp"
    $20 = ((($this)) + 52|0); //@line 417 "modules/graphics/opengl/Mesh.cpp"
    $21 = HEAP32[$20>>2]|0; //@line 417 "modules/graphics/opengl/Mesh.cpp"
    $22 = ((($21)) + 16|0); //@line 96 "modules/graphics/opengl/GLBuffer.h"
    $23 = HEAP32[$22>>2]|0; //@line 96 "modules/graphics/opengl/GLBuffer.h"
    __THREW__ = 0;
    invoke_viiiiii(4042,($19|0),($8|0),(0|0),34963,($23|0),0); //@line 417 "modules/graphics/opengl/Mesh.cpp"
    $24 = __THREW__; __THREW__ = 0;
    $25 = $24&1;
    if (!($25)) {
     HEAP32[$9>>2] = $19; //@line 417 "modules/graphics/opengl/Mesh.cpp"
     $35 = $19;
     break;
    }
    $26 = ___cxa_find_matching_catch_2()|0;
    $27 = tempRet0;
    __ZdlPv($19); //@line 417 "modules/graphics/opengl/Mesh.cpp"
    $$1 = $27;$$13 = $26;
    ___resumeException($$13|0);
    // unreachable;
   }
  }
 } while(0);
 $28 = ((($this)) + 72|0); //@line 419 "modules/graphics/opengl/Mesh.cpp"
 HEAP8[$28>>0] = 1; //@line 419 "modules/graphics/opengl/Mesh.cpp"
 $29 = HEAP32[$3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = HEAP32[$map>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = (($29) - ($30))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = $31 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $33 = ((($this)) + 76|0); //@line 420 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$33>>2] = $32; //@line 420 "modules/graphics/opengl/Mesh.cpp"
 $34 = ($35|0)==(0|0); //@line 422 "modules/graphics/opengl/Mesh.cpp"
 $36 = ($32|0)==(0); //@line 422 "modules/graphics/opengl/Mesh.cpp"
 $or$cond4 = $34 | $36; //@line 422 "modules/graphics/opengl/Mesh.cpp"
 if ($or$cond4) {
  STACKTOP = sp;return; //@line 441 "modules/graphics/opengl/Mesh.cpp"
 }
 __ZN4love8graphics6opengl8GLBuffer4bindEv($35); //@line 188 "modules/graphics/opengl/GLBuffer.h"
 $37 = HEAP32[$9>>2]|0; //@line 426 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 $38 = (invoke_ii(4096,($37|0))|0); //@line 216 "modules/graphics/opengl/GLBuffer.h"
 $39 = __THREW__; __THREW__ = 0;
 $40 = $39&1;
 if ($40) {
  $41 = ___cxa_find_matching_catch_2()|0;
  $42 = tempRet0;
  $$0 = $42;$$02 = $41;
 } else {
  L18: do {
   if ($2) {
    $72 = HEAP32[$3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $73 = HEAP32[$map>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $74 = ($72|0)==($73|0); //@line 392 "modules/graphics/opengl/Mesh.cpp"
    $75 = $73; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($74) {
     label = 29;
    } else {
     $76 = (($72) - ($73))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $77 = $76 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $i$02$i = 0;
     while(1) {
      $78 = (($75) + ($i$02$i<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $79 = HEAP32[$78>>2]|0; //@line 394 "modules/graphics/opengl/Mesh.cpp"
      $80 = ($79>>>0)<($1>>>0); //@line 394 "modules/graphics/opengl/Mesh.cpp"
      if (!($80)) {
       $i$02$i$lcssa = $i$02$i;
       break;
      }
      $91 = (($38) + ($i$02$i<<2)|0); //@line 397 "modules/graphics/opengl/Mesh.cpp"
      HEAP32[$91>>2] = $79; //@line 397 "modules/graphics/opengl/Mesh.cpp"
      $92 = (($i$02$i) + 1)|0; //@line 392 "modules/graphics/opengl/Mesh.cpp"
      $93 = ($92>>>0)<($77>>>0); //@line 392 "modules/graphics/opengl/Mesh.cpp"
      if ($93) {
       $i$02$i = $92;
      } else {
       label = 29;
       break L18;
      }
     }
     $81 = (___cxa_allocate_exception(16)|0); //@line 395 "modules/graphics/opengl/Mesh.cpp"
     $82 = HEAP32[$map>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $83 = (($82) + ($i$02$i$lcssa<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $84 = HEAP32[$83>>2]|0; //@line 395 "modules/graphics/opengl/Mesh.cpp"
     $85 = (($84) + 1)|0; //@line 395 "modules/graphics/opengl/Mesh.cpp"
     __THREW__ = 0;
     HEAP32[$vararg_buffer1>>2] = $85; //@line 395 "modules/graphics/opengl/Mesh.cpp"
     invoke_viii(3349,($81|0),(437264|0),($vararg_buffer1|0)); //@line 395 "modules/graphics/opengl/Mesh.cpp"
     $86 = __THREW__; __THREW__ = 0;
     $87 = $86&1;
     if ($87) {
      $89 = ___cxa_find_matching_catch_2()|0;
      $90 = tempRet0;
      ___cxa_free_exception(($81|0)); //@line 395 "modules/graphics/opengl/Mesh.cpp"
      $eh$lpad$body$index6Z2D = $90;$eh$lpad$body$indexZ2D = $89;
      break;
     } else {
      __THREW__ = 0;
      invoke_viii(3350,($81|0),(96|0),(344|0)); //@line 395 "modules/graphics/opengl/Mesh.cpp"
      $88 = __THREW__; __THREW__ = 0;
      label = 19;
      break;
     }
    }
   } else {
    $43 = HEAP32[$3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $44 = HEAP32[$map>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $45 = ($43|0)==($44|0); //@line 392 "modules/graphics/opengl/Mesh.cpp"
    $46 = $44; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($45) {
     label = 29;
    } else {
     $47 = (($43) - ($44))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $48 = $47 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $i$02$i9 = 0;
     while(1) {
      $49 = (($46) + ($i$02$i9<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $50 = HEAP32[$49>>2]|0; //@line 394 "modules/graphics/opengl/Mesh.cpp"
      $51 = ($50>>>0)<($1>>>0); //@line 394 "modules/graphics/opengl/Mesh.cpp"
      if (!($51)) {
       $i$02$i9$lcssa = $i$02$i9;
       break;
      }
      $62 = $50&65535; //@line 397 "modules/graphics/opengl/Mesh.cpp"
      $63 = (($38) + ($i$02$i9<<1)|0); //@line 397 "modules/graphics/opengl/Mesh.cpp"
      HEAP16[$63>>1] = $62; //@line 397 "modules/graphics/opengl/Mesh.cpp"
      $64 = (($i$02$i9) + 1)|0; //@line 392 "modules/graphics/opengl/Mesh.cpp"
      $65 = ($64>>>0)<($48>>>0); //@line 392 "modules/graphics/opengl/Mesh.cpp"
      if ($65) {
       $i$02$i9 = $64;
      } else {
       label = 29;
       break L18;
      }
     }
     $52 = (___cxa_allocate_exception(16)|0); //@line 395 "modules/graphics/opengl/Mesh.cpp"
     $53 = HEAP32[$map>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $54 = (($53) + ($i$02$i9$lcssa<<2)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $55 = HEAP32[$54>>2]|0; //@line 395 "modules/graphics/opengl/Mesh.cpp"
     $56 = (($55) + 1)|0; //@line 395 "modules/graphics/opengl/Mesh.cpp"
     __THREW__ = 0;
     HEAP32[$vararg_buffer>>2] = $56; //@line 395 "modules/graphics/opengl/Mesh.cpp"
     invoke_viii(3349,($52|0),(437264|0),($vararg_buffer|0)); //@line 395 "modules/graphics/opengl/Mesh.cpp"
     $57 = __THREW__; __THREW__ = 0;
     $58 = $57&1;
     if ($58) {
      $60 = ___cxa_find_matching_catch_2()|0;
      $61 = tempRet0;
      ___cxa_free_exception(($52|0)); //@line 395 "modules/graphics/opengl/Mesh.cpp"
      $eh$lpad$body$index6Z2D = $61;$eh$lpad$body$indexZ2D = $60;
      break;
     } else {
      __THREW__ = 0;
      invoke_viii(3350,($52|0),(96|0),(344|0)); //@line 395 "modules/graphics/opengl/Mesh.cpp"
      $59 = __THREW__; __THREW__ = 0;
      label = 19;
      break;
     }
    }
   }
  } while(0);
  if ((label|0) == 19) {
   $66 = ___cxa_find_matching_catch_2()|0;
   $67 = tempRet0;
   $eh$lpad$body$index6Z2D = $67;$eh$lpad$body$indexZ2D = $66;
  }
  else if ((label|0) == 29) {
   $94 = ((($this)) + 80|0); //@line 440 "modules/graphics/opengl/Mesh.cpp"
   HEAP32[$94>>2] = $$$i; //@line 440 "modules/graphics/opengl/Mesh.cpp"
   __THREW__ = 0;
   invoke_vi(4088,($37|0)); //@line 224 "modules/graphics/opengl/GLBuffer.h"
   $95 = __THREW__; __THREW__ = 0;
   $96 = $95&1;
   if ($96) {
    $97 = ___cxa_find_matching_catch_3(0|0)|0;
    $98 = tempRet0;
    ___clang_call_terminate($97); //@line 224 "modules/graphics/opengl/GLBuffer.h"
    // unreachable; //@line 224 "modules/graphics/opengl/GLBuffer.h"
   }
   __THREW__ = 0;
   invoke_vi(4089,($35|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   $99 = __THREW__; __THREW__ = 0;
   $100 = $99&1;
   if ($100) {
    $101 = ___cxa_find_matching_catch_3(0|0)|0;
    $102 = tempRet0;
    ___clang_call_terminate($101); //@line 196 "modules/graphics/opengl/GLBuffer.h"
    // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
   } else {
    STACKTOP = sp;return; //@line 441 "modules/graphics/opengl/Mesh.cpp"
   }
  }
  __THREW__ = 0;
  invoke_vi(4088,($37|0)); //@line 224 "modules/graphics/opengl/GLBuffer.h"
  $68 = __THREW__; __THREW__ = 0;
  $69 = $68&1;
  if ($69) {
   $70 = ___cxa_find_matching_catch_3(0|0)|0;
   $71 = tempRet0;
   ___clang_call_terminate($70); //@line 224 "modules/graphics/opengl/GLBuffer.h"
   // unreachable; //@line 224 "modules/graphics/opengl/GLBuffer.h"
  } else {
   $$0 = $eh$lpad$body$index6Z2D;$$02 = $eh$lpad$body$indexZ2D;
  }
 }
 __THREW__ = 0;
 invoke_vi(4089,($35|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
 $103 = __THREW__; __THREW__ = 0;
 $104 = $103&1;
 if ($104) {
  $105 = ___cxa_find_matching_catch_3(0|0)|0;
  $106 = tempRet0;
  ___clang_call_terminate($105); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
 } else {
  $$1 = $$0;$$13 = $$02;
  ___resumeException($$13|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl4Mesh12setVertexMapEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 72|0); //@line 445 "modules/graphics/opengl/Mesh.cpp"
 HEAP8[$0>>0] = 0; //@line 445 "modules/graphics/opengl/Mesh.cpp"
 return; //@line 446 "modules/graphics/opengl/Mesh.cpp"
}
function __ZNK4love8graphics6opengl4Mesh12getVertexMapERNSt3__16vectorIjNS3_9allocatorIjEEEE($this,$map) {
 $this = $this|0;
 $map = $map|0;
 var $$0 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $9 = 0, $cond = 0, $exitcond$i = 0, $exitcond$i4 = 0, $i$01$i = 0, $i$01$i3 = 0, $lpad$phi$index = 0, $lpad$phi$index3 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 4|0;
 $1 = sp;
 $2 = ((($this)) + 72|0); //@line 461 "modules/graphics/opengl/Mesh.cpp"
 $3 = HEAP8[$2>>0]|0; //@line 461 "modules/graphics/opengl/Mesh.cpp"
 $4 = ($3<<24>>24)==(0); //@line 461 "modules/graphics/opengl/Mesh.cpp"
 if ($4) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 488 "modules/graphics/opengl/Mesh.cpp"
 }
 $5 = HEAP32[$map>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ((($map)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = HEAP32[$6>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ($7|0)==($5|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = $5; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($8) {
  $$in = $7;
 } else {
  $scevgep$i$i$i = ((($7)) + -4|0);
  $10 = $scevgep$i$i$i;
  $11 = (($10) - ($9))|0;
  $12 = $11 >>> 2;
  $13 = $12 ^ -1;
  $scevgep4$i$i$i = (($7) + ($13<<2)|0);
  HEAP32[$6>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $scevgep4$i$i$i;
 }
 $14 = ((($this)) + 76|0); //@line 465 "modules/graphics/opengl/Mesh.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 465 "modules/graphics/opengl/Mesh.cpp"
 $16 = ((($map)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $17 = HEAP32[$16>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $18 = (($17) - ($9))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $19 = $18 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = ($19>>>0)<($15>>>0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($20) {
  $21 = $$in; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = (($21) - ($9))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = $22 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $24 = ($15|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($24) {
   $28 = 0;
  } else {
   $25 = $15 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $26 = (__Znwj($25)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $28 = $26;
  }
  $27 = (($28) + ($23<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $29 = $27; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $30 = (($28) + ($15<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $31 = $30; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $32 = (0 - ($23))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $33 = (($27) + ($32<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $34 = $33; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  _memcpy(($33|0),($5|0),($22|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$map>>2] = $34; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$6>>2] = $29; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$16>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $35 = ($5|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if (!($35)) {
   __ZdlPv($5); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
 }
 $36 = ((($this)) + 68|0); //@line 467 "modules/graphics/opengl/Mesh.cpp"
 $37 = HEAP32[$36>>2]|0; //@line 467 "modules/graphics/opengl/Mesh.cpp"
 $38 = ($37|0)==(0|0); //@line 467 "modules/graphics/opengl/Mesh.cpp"
 if ($38) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 488 "modules/graphics/opengl/Mesh.cpp"
 }
 $39 = HEAP32[$14>>2]|0; //@line 467 "modules/graphics/opengl/Mesh.cpp"
 $40 = ($39|0)==(0); //@line 467 "modules/graphics/opengl/Mesh.cpp"
 if ($40) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 488 "modules/graphics/opengl/Mesh.cpp"
 }
 __ZN4love8graphics6opengl8GLBuffer4bindEv($37); //@line 188 "modules/graphics/opengl/GLBuffer.h"
 $41 = HEAP32[$36>>2]|0; //@line 473 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 $42 = (invoke_ii(4096,($41|0))|0); //@line 473 "modules/graphics/opengl/Mesh.cpp"
 $43 = __THREW__; __THREW__ = 0;
 $44 = $43&1;
 L20: do {
  if ($44) {
   $52 = ___cxa_find_matching_catch_2()|0;
   $53 = tempRet0;
   $lpad$phi$index = $52;$lpad$phi$index3 = $53;
  } else {
   $45 = ((($this)) + 80|0); //@line 476 "modules/graphics/opengl/Mesh.cpp"
   $46 = HEAP32[$45>>2]|0; //@line 476 "modules/graphics/opengl/Mesh.cpp"
   $cond = ($46|0)==(5123); //@line 476 "modules/graphics/opengl/Mesh.cpp"
   $47 = HEAP32[$14>>2]|0; //@line 479 "modules/graphics/opengl/Mesh.cpp"
   if ($cond) {
    $58 = ($47|0)==(0); //@line 455 "modules/graphics/opengl/Mesh.cpp"
    L25: do {
     if (!($58)) {
      $i$01$i = 0;
      while(1) {
       $59 = (($42) + ($i$01$i<<1)|0); //@line 456 "modules/graphics/opengl/Mesh.cpp"
       $60 = HEAP16[$59>>1]|0; //@line 456 "modules/graphics/opengl/Mesh.cpp"
       $61 = $60&65535; //@line 456 "modules/graphics/opengl/Mesh.cpp"
       HEAP32[$1>>2] = $61; //@line 456 "modules/graphics/opengl/Mesh.cpp"
       $62 = HEAP32[$6>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $63 = HEAP32[$16>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $64 = ($62>>>0)<($63>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($64) {
        HEAP32[$62>>2] = $61; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $65 = ((($62)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$6>>2] = $65; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       } else {
        __THREW__ = 0;
        invoke_vii(3597,($map|0),($1|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $66 = __THREW__; __THREW__ = 0;
        $67 = $66&1;
        if ($67) {
         break;
        }
       }
       $68 = (($i$01$i) + 1)|0; //@line 455 "modules/graphics/opengl/Mesh.cpp"
       $exitcond$i = ($68|0)==($47|0); //@line 455 "modules/graphics/opengl/Mesh.cpp"
       if ($exitcond$i) {
        break L25;
       } else {
        $i$01$i = $68;
       }
      }
      $48 = ___cxa_find_matching_catch_2()|0;
      $49 = tempRet0;
      $lpad$phi$index = $48;$lpad$phi$index3 = $49;
      break L20;
     }
    } while(0);
   } else {
    $69 = ($47|0)==(0); //@line 455 "modules/graphics/opengl/Mesh.cpp"
    L35: do {
     if (!($69)) {
      $i$01$i3 = 0;
      while(1) {
       $70 = (($42) + ($i$01$i3<<2)|0); //@line 456 "modules/graphics/opengl/Mesh.cpp"
       $71 = HEAP32[$70>>2]|0; //@line 456 "modules/graphics/opengl/Mesh.cpp"
       HEAP32[$0>>2] = $71; //@line 456 "modules/graphics/opengl/Mesh.cpp"
       $72 = HEAP32[$6>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $73 = HEAP32[$16>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $74 = ($72>>>0)<($73>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($74) {
        HEAP32[$72>>2] = $71; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $75 = ((($72)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$6>>2] = $75; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       } else {
        __THREW__ = 0;
        invoke_vii(3597,($map|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $76 = __THREW__; __THREW__ = 0;
        $77 = $76&1;
        if ($77) {
         break;
        }
       }
       $78 = (($i$01$i3) + 1)|0; //@line 455 "modules/graphics/opengl/Mesh.cpp"
       $exitcond$i4 = ($78|0)==($47|0); //@line 455 "modules/graphics/opengl/Mesh.cpp"
       if ($exitcond$i4) {
        break L35;
       } else {
        $i$01$i3 = $78;
       }
      }
      $50 = ___cxa_find_matching_catch_2()|0;
      $51 = tempRet0;
      $lpad$phi$index = $50;$lpad$phi$index3 = $51;
      break L20;
     }
    } while(0);
   }
   __THREW__ = 0;
   invoke_vi(4089,($37|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   $79 = __THREW__; __THREW__ = 0;
   $80 = $79&1;
   if ($80) {
    $81 = ___cxa_find_matching_catch_3(0|0)|0;
    $82 = tempRet0;
    ___clang_call_terminate($81); //@line 196 "modules/graphics/opengl/GLBuffer.h"
    // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
   } else {
    $$0 = 1;
    STACKTOP = sp;return ($$0|0); //@line 488 "modules/graphics/opengl/Mesh.cpp"
   }
  }
 } while(0);
 __THREW__ = 0;
 invoke_vi(4089,($37|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
 $54 = __THREW__; __THREW__ = 0;
 $55 = $54&1;
 if ($55) {
  $56 = ___cxa_find_matching_catch_3(0|0)|0;
  $57 = tempRet0;
  ___clang_call_terminate($56); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
 } else {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl4Mesh10setTextureEPNS0_7TextureE($this,$tex) {
 $this = $this|0;
 $tex = $tex|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($tex|0)==(0|0); //@line 132 "./common/Object.h"
 if (!($0)) {
  $1 = HEAP32[$tex>>2]|0; //@line 132 "./common/Object.h"
  $2 = ((($1)) + 8|0); //@line 132 "./common/Object.h"
  $3 = HEAP32[$2>>2]|0; //@line 132 "./common/Object.h"
  FUNCTION_TABLE_vi[$3 & 8191]($tex); //@line 132 "./common/Object.h"
 }
 $4 = ((($this)) + 96|0); //@line 133 "./common/Object.h"
 $5 = HEAP32[$4>>2]|0; //@line 133 "./common/Object.h"
 $6 = ($5|0)==(0|0); //@line 133 "./common/Object.h"
 if ($6) {
  HEAP32[$4>>2] = $tex; //@line 134 "./common/Object.h"
  return; //@line 498 "modules/graphics/opengl/Mesh.cpp"
 }
 $7 = HEAP32[$5>>2]|0; //@line 133 "./common/Object.h"
 $8 = ((($7)) + 12|0); //@line 133 "./common/Object.h"
 $9 = HEAP32[$8>>2]|0; //@line 133 "./common/Object.h"
 FUNCTION_TABLE_vi[$9 & 8191]($5); //@line 133 "./common/Object.h"
 HEAP32[$4>>2] = $tex; //@line 134 "./common/Object.h"
 return; //@line 498 "modules/graphics/opengl/Mesh.cpp"
}
function __ZN4love8graphics6opengl4Mesh10setTextureEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 96|0); //@line 133 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 133 "./common/Object.h"
 $2 = ($1|0)==(0|0); //@line 133 "./common/Object.h"
 if (!($2)) {
  $3 = HEAP32[$1>>2]|0; //@line 133 "./common/Object.h"
  $4 = ((($3)) + 12|0); //@line 133 "./common/Object.h"
  $5 = HEAP32[$4>>2]|0; //@line 133 "./common/Object.h"
  FUNCTION_TABLE_vi[$5 & 8191]($1); //@line 133 "./common/Object.h"
 }
 HEAP32[$0>>2] = 0; //@line 134 "./common/Object.h"
 return; //@line 503 "modules/graphics/opengl/Mesh.cpp"
}
function __ZNK4love8graphics6opengl4Mesh10getTextureEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 96|0); //@line 139 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 139 "./common/Object.h"
 return ($1|0); //@line 507 "modules/graphics/opengl/Mesh.cpp"
}
function __ZN4love8graphics6opengl4Mesh11setDrawModeENS2_8DrawModeE($this,$mode) {
 $this = $this|0;
 $mode = $mode|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 84|0); //@line 512 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$0>>2] = $mode; //@line 512 "modules/graphics/opengl/Mesh.cpp"
 return; //@line 513 "modules/graphics/opengl/Mesh.cpp"
}
function __ZNK4love8graphics6opengl4Mesh11getDrawModeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 84|0); //@line 517 "modules/graphics/opengl/Mesh.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 517 "modules/graphics/opengl/Mesh.cpp"
 return ($1|0); //@line 517 "modules/graphics/opengl/Mesh.cpp"
}
function __ZN4love8graphics6opengl4Mesh12setDrawRangeEii($this,$min,$max) {
 $this = $this|0;
 $min = $min|0;
 $max = $max|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = $max | $min; //@line 522 "modules/graphics/opengl/Mesh.cpp"
 $1 = ($0|0)<(0); //@line 522 "modules/graphics/opengl/Mesh.cpp"
 $2 = ($min|0)>($max|0); //@line 522 "modules/graphics/opengl/Mesh.cpp"
 $or$cond = $2 | $1; //@line 522 "modules/graphics/opengl/Mesh.cpp"
 if (!($or$cond)) {
  $8 = ((($this)) + 88|0); //@line 525 "modules/graphics/opengl/Mesh.cpp"
  HEAP32[$8>>2] = $min; //@line 525 "modules/graphics/opengl/Mesh.cpp"
  $9 = ((($this)) + 92|0); //@line 526 "modules/graphics/opengl/Mesh.cpp"
  HEAP32[$9>>2] = $max; //@line 526 "modules/graphics/opengl/Mesh.cpp"
  STACKTOP = sp;return; //@line 527 "modules/graphics/opengl/Mesh.cpp"
 }
 $3 = (___cxa_allocate_exception(16)|0); //@line 523 "modules/graphics/opengl/Mesh.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($3|0),(437293|0),($vararg_buffer|0)); //@line 523 "modules/graphics/opengl/Mesh.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  ___cxa_free_exception(($3|0)); //@line 523 "modules/graphics/opengl/Mesh.cpp"
  ___resumeException($6|0);
  // unreachable;
 } else {
  ___cxa_throw(($3|0),(96|0),(344|0)); //@line 523 "modules/graphics/opengl/Mesh.cpp"
  // unreachable; //@line 523 "modules/graphics/opengl/Mesh.cpp"
 }
}
function __ZN4love8graphics6opengl4Mesh12setDrawRangeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 92|0); //@line 531 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$0>>2] = -1; //@line 531 "modules/graphics/opengl/Mesh.cpp"
 $1 = ((($this)) + 88|0); //@line 531 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$1>>2] = -1; //@line 531 "modules/graphics/opengl/Mesh.cpp"
 return; //@line 532 "modules/graphics/opengl/Mesh.cpp"
}
function __ZNK4love8graphics6opengl4Mesh12getDrawRangeERiS3_($this,$min,$max) {
 $this = $this|0;
 $min = $min|0;
 $max = $max|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 88|0); //@line 536 "modules/graphics/opengl/Mesh.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 536 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$min>>2] = $1; //@line 536 "modules/graphics/opengl/Mesh.cpp"
 $2 = ((($this)) + 92|0); //@line 537 "modules/graphics/opengl/Mesh.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 537 "modules/graphics/opengl/Mesh.cpp"
 HEAP32[$max>>2] = $3; //@line 537 "modules/graphics/opengl/Mesh.cpp"
 return; //@line 538 "modules/graphics/opengl/Mesh.cpp"
}
function __ZN4love8graphics6opengl4Mesh11getConstantEPKcRNS2_5UsageE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 6)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((633792 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (633792 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(6); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((633792 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 727 "modules/graphics/opengl/Mesh.cpp"
  return ($$$2$i|0); //@line 727 "modules/graphics/opengl/Mesh.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 727 "modules/graphics/opengl/Mesh.cpp"
  return ($$$2$i|0); //@line 727 "modules/graphics/opengl/Mesh.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl4Mesh11getConstantEPKcRNS2_8DrawModeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = $9 & 7; //@line 72 "./common/StringMap.h"
  $11 = (((633876 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (633876 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(8); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((633876 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 737 "modules/graphics/opengl/Mesh.cpp"
  return ($$$2$i|0); //@line 737 "modules/graphics/opengl/Mesh.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 737 "modules/graphics/opengl/Mesh.cpp"
  return ($$$2$i|0); //@line 737 "modules/graphics/opengl/Mesh.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl4Mesh11getConstantENS2_8DrawModeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(3); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (633972 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 742 "modules/graphics/opengl/Mesh.cpp"
}
function __ZN4love8graphics6opengl4Mesh11getConstantEPKcRNS2_8DataTypeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love9StringMapINS_8graphics6opengl4Mesh8DataTypeELj2EE4findEPKcRS4_(633988,$in,$out)|0); //@line 747 "modules/graphics/opengl/Mesh.cpp"
 return ($0|0); //@line 747 "modules/graphics/opengl/Mesh.cpp"
}
function __ZN4love9StringMapINS_8graphics6opengl4Mesh8DataTypeELj2EE4findEPKcRS4_($this,$key,$t) {
 $this = $this|0;
 $key = $key|0;
 $t = $t|0;
 var $$$2 = 0, $$011$i$137 = 0, $$011$i$233 = 0, $$011$i$330 = 0, $$011$i41 = 0, $$0110$i$138 = 0, $$0110$i$234 = 0, $$0110$i$331 = 0, $$0110$i42 = 0, $$02$i = 0, $$13 = 0, $$2 = 0, $$lcssa12 = 0, $$pre14 = 0, $$pre16 = 0, $$pre18 = 0, $$pre20 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $or$cond$i = 0, $or$cond$i$1 = 0, $or$cond$i$136 = 0, $or$cond$i$2 = 0, $or$cond$i$232 = 0, $or$cond$i$3 = 0, $or$cond$i$329 = 0, $or$cond$i40 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 3; //@line 72 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 74 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 74 "./common/StringMap.h"
 if ($12) {
  $$13 = 1;$$2 = 0;
  $$$2 = $$13 & $$2;
  return ($$$2|0); //@line 85 "./common/StringMap.h"
 }
 $13 = (($this) + (($9*12)|0)|0); //@line 77 "./common/StringMap.h"
 $14 = HEAP32[$13>>2]|0; //@line 77 "./common/StringMap.h"
 $15 = HEAP8[$14>>0]|0; //@line 54 "./common/StringMap.h"
 $16 = ($15<<24>>24)==(0); //@line 54 "./common/StringMap.h"
 L8: do {
  if ($16) {
   $26 = $0;
   label = 8;
  } else {
   $17 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
   $18 = ($15<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
   $or$cond$i40 = $17 & $18; //@line 54 "./common/StringMap.h"
   if ($or$cond$i40) {
    $$011$i41 = $key;$$0110$i42 = $14;
    while(1) {
     $19 = ((($$0110$i42)) + 1|0); //@line 59 "./common/StringMap.h"
     $20 = ((($$011$i41)) + 1|0); //@line 60 "./common/StringMap.h"
     $21 = HEAP8[$19>>0]|0; //@line 54 "./common/StringMap.h"
     $22 = ($21<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     $$pre14 = HEAP8[$20>>0]|0; //@line 63 "./common/StringMap.h"
     if ($22) {
      $26 = $$pre14;
      label = 8;
      break L8;
     }
     $23 = ($$pre14<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $24 = ($21<<24>>24)==($$pre14<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i = $23 & $24; //@line 54 "./common/StringMap.h"
     if ($or$cond$i) {
      $$011$i41 = $20;$$0110$i42 = $19;
     } else {
      label = 10;
      break;
     }
    }
   } else {
    label = 10;
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $25 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
  if ($25) {
   $$lcssa12 = $9;
  } else {
   label = 10;
  }
 }
 do {
  if ((label|0) == 10) {
   $29 = (($hash$0$lcssa$i) + 1)|0; //@line 72 "./common/StringMap.h"
   $30 = $29 & 3; //@line 72 "./common/StringMap.h"
   $31 = (((($this) + (($30*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $32 = HEAP8[$31>>0]|0; //@line 74 "./common/StringMap.h"
   $33 = ($32<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($33) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $34 = (($this) + (($30*12)|0)|0); //@line 77 "./common/StringMap.h"
   $35 = HEAP32[$34>>2]|0; //@line 77 "./common/StringMap.h"
   $36 = HEAP8[$35>>0]|0; //@line 54 "./common/StringMap.h"
   $37 = ($36<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L20: do {
    if ($37) {
     $47 = $0;
     label = 16;
    } else {
     $38 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $39 = ($36<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$136 = $38 & $39; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$136) {
      $$011$i$137 = $key;$$0110$i$138 = $35;
      while(1) {
       $40 = ((($$0110$i$138)) + 1|0); //@line 59 "./common/StringMap.h"
       $41 = ((($$011$i$137)) + 1|0); //@line 60 "./common/StringMap.h"
       $42 = HEAP8[$40>>0]|0; //@line 54 "./common/StringMap.h"
       $43 = ($42<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre16 = HEAP8[$41>>0]|0; //@line 63 "./common/StringMap.h"
       if ($43) {
        $47 = $$pre16;
        label = 16;
        break L20;
       }
       $44 = ($$pre16<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $45 = ($42<<24>>24)==($$pre16<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$1 = $44 & $45; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$1) {
        $$011$i$137 = $41;$$0110$i$138 = $40;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 16) {
    $46 = ($47<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($46) {
     $$lcssa12 = $30;
     break;
    }
   }
   $48 = (($hash$0$lcssa$i) + 2)|0; //@line 72 "./common/StringMap.h"
   $49 = $48 & 3; //@line 72 "./common/StringMap.h"
   $50 = (((($this) + (($49*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $51 = HEAP8[$50>>0]|0; //@line 74 "./common/StringMap.h"
   $52 = ($51<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($52) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $53 = (($this) + (($49*12)|0)|0); //@line 77 "./common/StringMap.h"
   $54 = HEAP32[$53>>2]|0; //@line 77 "./common/StringMap.h"
   $55 = HEAP8[$54>>0]|0; //@line 54 "./common/StringMap.h"
   $56 = ($55<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L31: do {
    if ($56) {
     $66 = $0;
     label = 22;
    } else {
     $57 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $58 = ($55<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$232 = $57 & $58; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$232) {
      $$011$i$233 = $key;$$0110$i$234 = $54;
      while(1) {
       $59 = ((($$0110$i$234)) + 1|0); //@line 59 "./common/StringMap.h"
       $60 = ((($$011$i$233)) + 1|0); //@line 60 "./common/StringMap.h"
       $61 = HEAP8[$59>>0]|0; //@line 54 "./common/StringMap.h"
       $62 = ($61<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre18 = HEAP8[$60>>0]|0; //@line 63 "./common/StringMap.h"
       if ($62) {
        $66 = $$pre18;
        label = 22;
        break L31;
       }
       $63 = ($$pre18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $64 = ($61<<24>>24)==($$pre18<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$2 = $63 & $64; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$2) {
        $$011$i$233 = $60;$$0110$i$234 = $59;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 22) {
    $65 = ($66<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($65) {
     $$lcssa12 = $49;
     break;
    }
   }
   $67 = (($hash$0$lcssa$i) + 3)|0; //@line 72 "./common/StringMap.h"
   $68 = $67 & 3; //@line 72 "./common/StringMap.h"
   $69 = (((($this) + (($68*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $70 = HEAP8[$69>>0]|0; //@line 74 "./common/StringMap.h"
   $71 = ($70<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($71) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $72 = (($this) + (($68*12)|0)|0); //@line 77 "./common/StringMap.h"
   $73 = HEAP32[$72>>2]|0; //@line 77 "./common/StringMap.h"
   $74 = HEAP8[$73>>0]|0; //@line 54 "./common/StringMap.h"
   $75 = ($74<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L42: do {
    if ($75) {
     $85 = $0;
    } else {
     $76 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $77 = ($74<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$329 = $76 & $77; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$329) {
      $$011$i$330 = $key;$$0110$i$331 = $73;
     } else {
      $$13 = 0;$$2 = 0;
      $$$2 = $$13 & $$2;
      return ($$$2|0); //@line 85 "./common/StringMap.h"
     }
     while(1) {
      $78 = ((($$0110$i$331)) + 1|0); //@line 59 "./common/StringMap.h"
      $79 = ((($$011$i$330)) + 1|0); //@line 60 "./common/StringMap.h"
      $80 = HEAP8[$78>>0]|0; //@line 54 "./common/StringMap.h"
      $81 = ($80<<24>>24)==(0); //@line 54 "./common/StringMap.h"
      $$pre20 = HEAP8[$79>>0]|0; //@line 63 "./common/StringMap.h"
      if ($81) {
       $85 = $$pre20;
       break L42;
      }
      $82 = ($$pre20<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
      $83 = ($80<<24>>24)==($$pre20<<24>>24); //@line 56 "./common/StringMap.h"
      $or$cond$i$3 = $82 & $83; //@line 54 "./common/StringMap.h"
      if ($or$cond$i$3) {
       $$011$i$330 = $79;$$0110$i$331 = $78;
      } else {
       $$13 = 0;$$2 = 0;
       break;
      }
     }
     $$$2 = $$13 & $$2;
     return ($$$2|0); //@line 85 "./common/StringMap.h"
    }
   } while(0);
   $84 = ($85<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($84) {
    $$lcssa12 = $68;
   } else {
    $$13 = 0;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
  }
 } while(0);
 $27 = (((($this) + (($$lcssa12*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
 $28 = HEAP32[$27>>2]|0; //@line 79 "./common/StringMap.h"
 HEAP32[$t>>2] = $28; //@line 79 "./common/StringMap.h"
 $$13 = 1;$$2 = 1;
 $$$2 = $$13 & $$2;
 return ($$$2|0); //@line 85 "./common/StringMap.h"
}
function __ZN4love8graphics6opengl4Mesh11getConstantENS2_8DataTypeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(1); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (634036 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 752 "modules/graphics/opengl/Mesh.cpp"
}
function __GLOBAL__sub_I_OpenGL_cpp() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl6OpenGLC2Ev(634044); //@line 759 "modules/graphics/opengl/OpenGL.cpp"
 (___cxa_atexit((4099|0),(634044|0),(___dso_handle|0))|0); //@line 759 "modules/graphics/opengl/OpenGL.cpp"
 return;
}
function __ZN4love8graphics6opengl6OpenGLC2Ev($this) {
 $this = $this|0;
 var $$0 = 0, $$0$i = 0, $$01 = 0, $$01$i = 0, $$01$i$i$i = 0, $$01$i$i$i4 = 0, $$lcssa = 0, $$lcssa24 = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i2 = 0, $$pre$i$i6 = 0, $$pre$i3$i = 0, $$pre$i7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__v$sroa$5$0$i = 0, $__v$sroa$5$0$i3 = 0, $scevgep$i$i$i$i = 0, $scevgep$i$i$i$i14 = 0, $scevgep4$i$i$i$i = 0, $scevgep4$i$i$i$i15 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 38|0); //@line 68 "modules/graphics/opengl/OpenGL.cpp"
 HEAP8[$0>>0] = 0; //@line 68 "modules/graphics/opengl/OpenGL.cpp"
 $1 = ((($this)) + 40|0); //@line 69 "modules/graphics/opengl/OpenGL.cpp"
 dest=$this; stop=dest+36|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 67 "modules/graphics/opengl/OpenGL.cpp"
 HEAPF32[$1>>2] = 1.0; //@line 69 "modules/graphics/opengl/OpenGL.cpp"
 $2 = ((($this)) + 44|0); //@line 70 "modules/graphics/opengl/OpenGL.cpp"
 HEAP32[$2>>2] = 0; //@line 70 "modules/graphics/opengl/OpenGL.cpp"
 $3 = ((($this)) + 48|0); //@line 71 "modules/graphics/opengl/OpenGL.cpp"
 HEAP32[$3>>2] = 1; //@line 71 "modules/graphics/opengl/OpenGL.cpp"
 $4 = ((($this)) + 52|0); //@line 72 "modules/graphics/opengl/OpenGL.cpp"
 HEAP32[$4>>2] = 0; //@line 72 "modules/graphics/opengl/OpenGL.cpp"
 $5 = ((($this)) + 56|0); //@line 73 "modules/graphics/opengl/OpenGL.cpp"
 HEAP32[$5>>2] = 1; //@line 73 "modules/graphics/opengl/OpenGL.cpp"
 $6 = ((($this)) + 60|0); //@line 74 "modules/graphics/opengl/OpenGL.cpp"
 HEAP32[$6>>2] = 11; //@line 74 "modules/graphics/opengl/OpenGL.cpp"
 $7 = ((($this)) + 64|0); //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($this)) + 68|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ((($this)) + 128|0); //@line 415 "modules/graphics/opengl/OpenGL.h"
 _memset(($7|0),0,192)|0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 __THREW__ = 0;
 invoke_vi(4100,($9|0)); //@line 415 "modules/graphics/opengl/OpenGL.h"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 do {
  if ($11) {
   $15 = ___cxa_find_matching_catch_2()|0;
   $16 = tempRet0;
   $$0$i = $16;$$01$i = $15;
   label = 5;
  } else {
   $12 = ((($this)) + 192|0); //@line 415 "modules/graphics/opengl/OpenGL.h"
   __THREW__ = 0;
   invoke_vi(4100,($12|0)); //@line 415 "modules/graphics/opengl/OpenGL.h"
   $13 = __THREW__; __THREW__ = 0;
   $14 = $13&1;
   if ($14) {
    $17 = ___cxa_find_matching_catch_2()|0;
    $18 = tempRet0;
    __ZN4love7Matrix4D2Ev($9); //@line 415 "modules/graphics/opengl/OpenGL.h"
    $$0$i = $18;$$01$i = $17;
    label = 5;
    break;
   }
   $28 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $29 = HEAP32[$28>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $30 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $31 = (($29) - ($30))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $32 = $31 >> 6; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $33 = ($32>>>0)<(10); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $34 = $30; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($33) {
    $35 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $36 = HEAP32[$35>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $37 = $36; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __THREW__ = 0;
    $38 = (invoke_ii(3348,640)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $39 = __THREW__; __THREW__ = 0;
    $40 = $39&1;
    if (!($40)) {
     $41 = (($36) - ($30))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $42 = $41 >> 6; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $43 = (($38) + ($42<<6)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $44 = $43; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $45 = ((($38)) + 640|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $46 = $45; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $47 = ($37|0)==($34|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     if ($47) {
      $55 = $44;$57 = $30;$59 = $34;
     } else {
      $$01$i$i$i = $37;$49 = $43;$__v$sroa$5$0$i = $44;
      while(1) {
       $48 = ((($49)) + -64|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $50 = ((($$01$i$i$i)) + -64|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       dest=$48; src=$50; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $51 = $__v$sroa$5$0$i; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $52 = ((($51)) + -64|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $53 = $52; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $54 = ($50|0)==($34|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       if ($54) {
        $$lcssa24 = $53;
        break;
       } else {
        $$01$i$i$i = $50;$49 = $52;$__v$sroa$5$0$i = $53;
       }
      }
      $$pre$i$i2 = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $$pre$i = HEAP32[$35>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $55 = $$lcssa24;$57 = $$pre$i$i2;$59 = $$pre$i;
     }
     HEAP32[$this>>2] = $55; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$35>>2] = $44; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$28>>2] = $46; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $56 = $57; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $58 = ($59|0)==($56|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     if (!($58)) {
      $61 = $59;
      while(1) {
       $60 = ((($61)) + -64|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
       __ZN4love7Matrix4D2Ev($60); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $62 = ($60|0)==($56|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
       if ($62) {
        break;
       } else {
        $61 = $60;
       }
      }
     }
     $63 = ($57|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     if ($63) {
      label = 18;
     } else {
      $64 = $57; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      __ZdlPv($64); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      label = 18;
     }
    }
   } else {
    label = 18;
   }
   if ((label|0) == 18) {
    $65 = ((($this)) + 12|0); //@line 78 "modules/graphics/opengl/OpenGL.cpp"
    $66 = ((($this)) + 20|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $67 = HEAP32[$66>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $68 = HEAP32[$65>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $69 = (($67) - ($68))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $70 = $69 >> 6; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $71 = ($70>>>0)<(2); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $72 = $68; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($71)) {
     return; //@line 79 "modules/graphics/opengl/OpenGL.cpp"
    }
    $73 = ((($this)) + 16|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $74 = HEAP32[$73>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $75 = $74; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __THREW__ = 0;
    $76 = (invoke_ii(3348,128)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $77 = __THREW__; __THREW__ = 0;
    $78 = $77&1;
    if (!($78)) {
     $79 = (($74) - ($68))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $80 = $79 >> 6; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $81 = (($76) + ($80<<6)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $82 = $81; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $83 = ((($76)) + 128|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $84 = $83; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $85 = ($75|0)==($72|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     if ($85) {
      $93 = $82;$95 = $68;$97 = $72;
     } else {
      $$01$i$i$i4 = $75;$87 = $81;$__v$sroa$5$0$i3 = $82;
      while(1) {
       $86 = ((($87)) + -64|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $88 = ((($$01$i$i$i4)) + -64|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       dest=$86; src=$88; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $89 = $__v$sroa$5$0$i3; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $90 = ((($89)) + -64|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $91 = $90; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $92 = ($88|0)==($72|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       if ($92) {
        $$lcssa = $91;
        break;
       } else {
        $$01$i$i$i4 = $88;$87 = $90;$__v$sroa$5$0$i3 = $91;
       }
      }
      $$pre$i$i6 = HEAP32[$65>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $$pre$i7 = HEAP32[$73>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $93 = $$lcssa;$95 = $$pre$i$i6;$97 = $$pre$i7;
     }
     HEAP32[$65>>2] = $93; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$73>>2] = $82; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$66>>2] = $84; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $94 = $95; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $96 = ($97|0)==($94|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     if (!($96)) {
      $99 = $97;
      while(1) {
       $98 = ((($99)) + -64|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
       __ZN4love7Matrix4D2Ev($98); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $100 = ($98|0)==($94|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
       if ($100) {
        break;
       } else {
        $99 = $98;
       }
      }
     }
     $101 = ($95|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     if ($101) {
      return; //@line 79 "modules/graphics/opengl/OpenGL.cpp"
     }
     $102 = $95; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     __ZdlPv($102); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     return; //@line 79 "modules/graphics/opengl/OpenGL.cpp"
    }
   }
   $103 = ___cxa_find_matching_catch_2()|0;
   $104 = tempRet0;
   __ZN4love7Matrix4D2Ev($12); //@line 415 "modules/graphics/opengl/OpenGL.h"
   __ZN4love7Matrix4D2Ev($9); //@line 415 "modules/graphics/opengl/OpenGL.h"
   $105 = HEAP32[$7>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $106 = ($105|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $107 = $105; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($106) {
    $$0 = $104;$$01 = $103;
   } else {
    $108 = HEAP32[$8>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $109 = ($108|0)==($105|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($109)) {
     $scevgep$i$i$i$i14 = ((($108)) + -4|0);
     $110 = $scevgep$i$i$i$i14;
     $111 = (($110) - ($107))|0;
     $112 = $111 >>> 2;
     $113 = $112 ^ -1;
     $scevgep4$i$i$i$i15 = (($108) + ($113<<2)|0);
     HEAP32[$8>>2] = $scevgep4$i$i$i$i15; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($105); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $$0 = $104;$$01 = $103;
   }
  }
 } while(0);
 if ((label|0) == 5) {
  $19 = HEAP32[$7>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = ($19|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = $19; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($20) {
   $$0 = $$0$i;$$01 = $$01$i;
  } else {
   $22 = HEAP32[$8>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $23 = ($22|0)==($19|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($23)) {
    $scevgep$i$i$i$i = ((($22)) + -4|0);
    $24 = $scevgep$i$i$i$i;
    $25 = (($24) - ($21))|0;
    $26 = $25 >>> 2;
    $27 = $26 ^ -1;
    $scevgep4$i$i$i$i = (($22) + ($27<<2)|0);
    HEAP32[$8>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($19); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $$0 = $$0$i;$$01 = $$01$i;
  }
 }
 $114 = ((($this)) + 12|0); //@line 112 "modules/graphics/opengl/OpenGL.h"
 $115 = HEAP32[$114>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $116 = ($115|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($116)) {
  $117 = ((($this)) + 16|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $118 = HEAP32[$117>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $119 = ($118|0)==($115|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($119) {
   $124 = $115;
  } else {
   $121 = $118;
   while(1) {
    $120 = ((($121)) + -64|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$117>>2] = $120; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __ZN4love7Matrix4D2Ev($120); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $122 = HEAP32[$117>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $123 = ($122|0)==($115|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($123) {
     break;
    } else {
     $121 = $122;
    }
   }
   $$pre$i$i = HEAP32[$114>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $124 = $$pre$i$i;
  }
  __ZdlPv($124); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $125 = HEAP32[$this>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $126 = ($125|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($126) {
  ___resumeException($$01|0);
  // unreachable;
 }
 $127 = ((($this)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $128 = HEAP32[$127>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $129 = ($128|0)==($125|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($129) {
  $134 = $125;
 } else {
  $131 = $128;
  while(1) {
   $130 = ((($131)) + -64|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$127>>2] = $130; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZN4love7Matrix4D2Ev($130); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $132 = HEAP32[$127>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $133 = ($132|0)==($125|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($133) {
    break;
   } else {
    $131 = $132;
   }
  }
  $$pre$i3$i = HEAP32[$this>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $134 = $$pre$i3$i;
 }
 __ZdlPv($134); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN4love8graphics6opengl6OpenGLD2Ev($this) {
 $this = $this|0;
 var $$pre$i$i = 0, $$pre$i3$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i$i = 0, $scevgep4$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 192|0); //@line 415 "modules/graphics/opengl/OpenGL.h"
 __ZN4love7Matrix4D2Ev($0); //@line 415 "modules/graphics/opengl/OpenGL.h"
 $1 = ((($this)) + 128|0); //@line 415 "modules/graphics/opengl/OpenGL.h"
 __ZN4love7Matrix4D2Ev($1); //@line 415 "modules/graphics/opengl/OpenGL.h"
 $2 = ((($this)) + 64|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($3|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = $3; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($4)) {
  $6 = ((($this)) + 68|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $7 = HEAP32[$6>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $8 = ($7|0)==($3|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($8)) {
   $scevgep$i$i$i$i = ((($7)) + -4|0);
   $9 = $scevgep$i$i$i$i;
   $10 = (($9) - ($5))|0;
   $11 = $10 >>> 2;
   $12 = $11 ^ -1;
   $scevgep4$i$i$i$i = (($7) + ($12<<2)|0);
   HEAP32[$6>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($3); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $13 = ((($this)) + 12|0); //@line 112 "modules/graphics/opengl/OpenGL.h"
 $14 = HEAP32[$13>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = ($14|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($15)) {
  $16 = ((($this)) + 16|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = HEAP32[$16>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ($17|0)==($14|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($18) {
   $23 = $14;
  } else {
   $20 = $17;
   while(1) {
    $19 = ((($20)) + -64|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$16>>2] = $19; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    __ZN4love7Matrix4D2Ev($19); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $21 = HEAP32[$16>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $22 = ($21|0)==($14|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($22) {
     break;
    } else {
     $20 = $21;
    }
   }
   $$pre$i$i = HEAP32[$13>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $23 = $$pre$i$i;
  }
  __ZdlPv($23); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $24 = HEAP32[$this>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $25 = ($24|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($25) {
  return; //@line 79 "modules/graphics/opengl/OpenGL.h"
 }
 $26 = ((($this)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = HEAP32[$26>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $28 = ($27|0)==($24|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($28) {
  $33 = $24;
 } else {
  $30 = $27;
  while(1) {
   $29 = ((($30)) + -64|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$26>>2] = $29; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZN4love7Matrix4D2Ev($29); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $31 = HEAP32[$26>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $32 = ($31|0)==($24|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($32) {
    break;
   } else {
    $30 = $31;
   }
  }
  $$pre$i3$i = HEAP32[$this>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $33 = $$pre$i3$i;
 }
 __ZdlPv($33); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 79 "modules/graphics/opengl/OpenGL.h"
}
function __ZN4love8graphics6opengl6OpenGL11initContextEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 38|0); //@line 83 "modules/graphics/opengl/OpenGL.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 83 "modules/graphics/opengl/OpenGL.cpp"
 $2 = ($1<<24>>24)==(0); //@line 83 "modules/graphics/opengl/OpenGL.cpp"
 if (!($2)) {
  $$0 = 1;
  return ($$0|0); //@line 107 "modules/graphics/opengl/OpenGL.cpp"
 }
 $3 = (__ZN4glad16gladLoadGLLoaderEPFPvPKcE(4101)|0); //@line 86 "modules/graphics/opengl/OpenGL.cpp"
 if (!($3)) {
  $$0 = 0;
  return ($$0|0); //@line 107 "modules/graphics/opengl/OpenGL.cpp"
 }
 __ZN4love8graphics6opengl6OpenGL19initOpenGLFunctionsEv(0); //@line 89 "modules/graphics/opengl/OpenGL.cpp"
 __ZN4love8graphics6opengl6OpenGL10initVendorEv($this); //@line 90 "modules/graphics/opengl/OpenGL.cpp"
 __ZN4love8graphics6opengl6OpenGL12initMatricesEv($this); //@line 91 "modules/graphics/opengl/OpenGL.cpp"
 $4 = ((($this)) + 36|0); //@line 93 "modules/graphics/opengl/OpenGL.cpp"
 HEAP8[$4>>0] = 0; //@line 93 "modules/graphics/opengl/OpenGL.cpp"
 $5 = ((($this)) + 37|0); //@line 93 "modules/graphics/opengl/OpenGL.cpp"
 HEAP8[$5>>0] = 0; //@line 93 "modules/graphics/opengl/OpenGL.cpp"
 HEAP8[$0>>0] = 1; //@line 104 "modules/graphics/opengl/OpenGL.cpp"
 $$0 = 1;
 return ($$0|0); //@line 107 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6openglL18LOVEGetProcAddressEPKc($name) {
 $name = $name|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GL_GetProcAddress($name)|0); //@line 63 "modules/graphics/opengl/OpenGL.cpp"
 return ($0|0); //@line 63 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL19initOpenGLFunctionsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[1252013]|0; //@line 224 "modules/graphics/opengl/OpenGL.cpp"
 $1 = HEAP8[1252002]|0; //@line 224 "modules/graphics/opengl/OpenGL.cpp"
 $2 = $1 | $0; //@line 224 "modules/graphics/opengl/OpenGL.cpp"
 $3 = HEAP8[1252065]|0; //@line 224 "modules/graphics/opengl/OpenGL.cpp"
 $4 = $2 | $3; //@line 224 "modules/graphics/opengl/OpenGL.cpp"
 $5 = ($4<<24>>24)==(0); //@line 224 "modules/graphics/opengl/OpenGL.cpp"
 if (!($5)) {
  return; //@line 259 "modules/graphics/opengl/OpenGL.cpp"
 }
 $6 = HEAP8[1251994]|0; //@line 226 "modules/graphics/opengl/OpenGL.cpp"
 $7 = ($6<<24>>24)!=(0); //@line 226 "modules/graphics/opengl/OpenGL.cpp"
 $8 = HEAP8[1252185]|0; //@line 226 "modules/graphics/opengl/OpenGL.cpp"
 $9 = ($8<<24>>24)!=(0); //@line 226 "modules/graphics/opengl/OpenGL.cpp"
 $or$cond5 = $7 & $9; //@line 226 "modules/graphics/opengl/OpenGL.cpp"
 if ($or$cond5) {
  $10 = HEAP32[166094]|0; //@line 228 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[165274] = $10; //@line 228 "modules/graphics/opengl/OpenGL.cpp"
  $11 = HEAP32[166095]|0; //@line 229 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[164936] = $11; //@line 229 "modules/graphics/opengl/OpenGL.cpp"
  $12 = HEAP32[166096]|0; //@line 230 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[164676] = $12; //@line 230 "modules/graphics/opengl/OpenGL.cpp"
  $13 = HEAP32[166097]|0; //@line 231 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[164704] = $13; //@line 231 "modules/graphics/opengl/OpenGL.cpp"
  $14 = HEAP32[166098]|0; //@line 232 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[164884] = $14; //@line 232 "modules/graphics/opengl/OpenGL.cpp"
  $15 = HEAP32[166100]|0; //@line 233 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[165202] = $15; //@line 233 "modules/graphics/opengl/OpenGL.cpp"
  $16 = HEAP32[166101]|0; //@line 234 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[164758] = $16; //@line 234 "modules/graphics/opengl/OpenGL.cpp"
  $17 = HEAP32[166102]|0; //@line 235 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[164797] = $17; //@line 235 "modules/graphics/opengl/OpenGL.cpp"
  $18 = HEAP32[166103]|0; //@line 236 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[165498] = $18; //@line 236 "modules/graphics/opengl/OpenGL.cpp"
  $19 = HEAP32[166105]|0; //@line 237 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[164777] = $19; //@line 237 "modules/graphics/opengl/OpenGL.cpp"
  $20 = HEAP32[166107]|0; //@line 238 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[164630] = $20; //@line 238 "modules/graphics/opengl/OpenGL.cpp"
  $21 = HEAP32[166108]|0; //@line 239 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[164778] = $21; //@line 239 "modules/graphics/opengl/OpenGL.cpp"
  $22 = HEAP32[166109]|0; //@line 240 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[165249] = $22; //@line 240 "modules/graphics/opengl/OpenGL.cpp"
 }
 $23 = HEAP8[1252182]|0; //@line 243 "modules/graphics/opengl/OpenGL.cpp"
 $24 = ($23<<24>>24)==(0); //@line 243 "modules/graphics/opengl/OpenGL.cpp"
 do {
  if ($24) {
   $26 = HEAP8[1252344]|0; //@line 245 "modules/graphics/opengl/OpenGL.cpp"
   $27 = ($26<<24>>24)==(0); //@line 245 "modules/graphics/opengl/OpenGL.cpp"
   if (!($27)) {
    $28 = HEAP32[166306]|0; //@line 246 "modules/graphics/opengl/OpenGL.cpp"
    HEAP32[164725] = $28; //@line 246 "modules/graphics/opengl/OpenGL.cpp"
    break;
   }
   $29 = HEAP8[1252431]|0; //@line 247 "modules/graphics/opengl/OpenGL.cpp"
   $30 = ($29<<24>>24)==(0); //@line 247 "modules/graphics/opengl/OpenGL.cpp"
   if (!($30)) {
    $31 = HEAP32[166437]|0; //@line 248 "modules/graphics/opengl/OpenGL.cpp"
    HEAP32[164725] = $31; //@line 248 "modules/graphics/opengl/OpenGL.cpp"
   }
  } else {
   $25 = HEAP32[166091]|0; //@line 244 "modules/graphics/opengl/OpenGL.cpp"
   HEAP32[164725] = $25; //@line 244 "modules/graphics/opengl/OpenGL.cpp"
  }
 } while(0);
 $32 = HEAP8[1252183]|0; //@line 250 "modules/graphics/opengl/OpenGL.cpp"
 $33 = ($32<<24>>24)==(0); //@line 250 "modules/graphics/opengl/OpenGL.cpp"
 if (!($33)) {
  $34 = HEAP32[166092]|0; //@line 251 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[165298] = $34; //@line 251 "modules/graphics/opengl/OpenGL.cpp"
  return; //@line 259 "modules/graphics/opengl/OpenGL.cpp"
 }
 $35 = HEAP8[1252362]|0; //@line 252 "modules/graphics/opengl/OpenGL.cpp"
 $36 = ($35<<24>>24)==(0); //@line 252 "modules/graphics/opengl/OpenGL.cpp"
 if (!($36)) {
  $37 = HEAP32[166328]|0; //@line 253 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[165298] = $37; //@line 253 "modules/graphics/opengl/OpenGL.cpp"
  return; //@line 259 "modules/graphics/opengl/OpenGL.cpp"
 }
 $38 = HEAP8[1252345]|0; //@line 254 "modules/graphics/opengl/OpenGL.cpp"
 $39 = ($38<<24>>24)==(0); //@line 254 "modules/graphics/opengl/OpenGL.cpp"
 if (!($39)) {
  $40 = HEAP32[166307]|0; //@line 255 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[165298] = $40; //@line 255 "modules/graphics/opengl/OpenGL.cpp"
  return; //@line 259 "modules/graphics/opengl/OpenGL.cpp"
 }
 $41 = HEAP8[1252433]|0; //@line 256 "modules/graphics/opengl/OpenGL.cpp"
 $42 = ($41<<24>>24)==(0); //@line 256 "modules/graphics/opengl/OpenGL.cpp"
 if ($42) {
  return; //@line 259 "modules/graphics/opengl/OpenGL.cpp"
 }
 $43 = HEAP32[166441]|0; //@line 257 "modules/graphics/opengl/OpenGL.cpp"
 HEAP32[165298] = $43; //@line 257 "modules/graphics/opengl/OpenGL.cpp"
 return; //@line 259 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL10initVendorEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[165140]|0; //@line 74 "./libraries/glad/gladfuncs.hpp"
 $1 = (FUNCTION_TABLE_ii[$0 & 8191](7936)|0); //@line 74 "./libraries/glad/gladfuncs.hpp"
 $2 = ($1|0)==(0|0); //@line 186 "modules/graphics/opengl/OpenGL.cpp"
 if ($2) {
  $3 = ((($this)) + 60|0); //@line 188 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[$3>>2] = 11; //@line 188 "modules/graphics/opengl/OpenGL.cpp"
  return; //@line 218 "modules/graphics/opengl/OpenGL.cpp"
 }
 $4 = (_strstr($1,437313)|0); //@line 194 "modules/graphics/opengl/OpenGL.cpp"
 $5 = ($4|0)==(0|0); //@line 194 "modules/graphics/opengl/OpenGL.cpp"
 if (!($5)) {
  $6 = ((($this)) + 60|0); //@line 195 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[$6>>2] = 0; //@line 195 "modules/graphics/opengl/OpenGL.cpp"
  return; //@line 218 "modules/graphics/opengl/OpenGL.cpp"
 }
 $7 = (_strstr($1,437330)|0); //@line 196 "modules/graphics/opengl/OpenGL.cpp"
 $8 = ($7|0)==(0|0); //@line 196 "modules/graphics/opengl/OpenGL.cpp"
 if (!($8)) {
  $9 = ((($this)) + 60|0); //@line 197 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[$9>>2] = 1; //@line 197 "modules/graphics/opengl/OpenGL.cpp"
  return; //@line 218 "modules/graphics/opengl/OpenGL.cpp"
 }
 $10 = (_strstr($1,437337)|0); //@line 198 "modules/graphics/opengl/OpenGL.cpp"
 $11 = ($10|0)==(0|0); //@line 198 "modules/graphics/opengl/OpenGL.cpp"
 if (!($11)) {
  $12 = ((($this)) + 60|0); //@line 199 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[$12>>2] = 2; //@line 199 "modules/graphics/opengl/OpenGL.cpp"
  return; //@line 218 "modules/graphics/opengl/OpenGL.cpp"
 }
 $13 = (_strstr($1,437343)|0); //@line 200 "modules/graphics/opengl/OpenGL.cpp"
 $14 = ($13|0)==(0|0); //@line 200 "modules/graphics/opengl/OpenGL.cpp"
 if (!($14)) {
  $15 = ((($this)) + 60|0); //@line 201 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[$15>>2] = 3; //@line 201 "modules/graphics/opengl/OpenGL.cpp"
  return; //@line 218 "modules/graphics/opengl/OpenGL.cpp"
 }
 $16 = (_strstr($1,437348)|0); //@line 202 "modules/graphics/opengl/OpenGL.cpp"
 $17 = ($16|0)==(0|0); //@line 202 "modules/graphics/opengl/OpenGL.cpp"
 if ($17) {
  $18 = (_strstr($1,437363)|0); //@line 202 "modules/graphics/opengl/OpenGL.cpp"
  $19 = ($18|0)==(0|0); //@line 202 "modules/graphics/opengl/OpenGL.cpp"
  if ($19) {
   $21 = (_strstr($1,437374)|0); //@line 204 "modules/graphics/opengl/OpenGL.cpp"
   $22 = ($21|0)==(0|0); //@line 204 "modules/graphics/opengl/OpenGL.cpp"
   if (!($22)) {
    $23 = ((($this)) + 60|0); //@line 205 "modules/graphics/opengl/OpenGL.cpp"
    HEAP32[$23>>2] = 5; //@line 205 "modules/graphics/opengl/OpenGL.cpp"
    return; //@line 218 "modules/graphics/opengl/OpenGL.cpp"
   }
   $24 = (_strstr($1,437384)|0); //@line 206 "modules/graphics/opengl/OpenGL.cpp"
   $25 = ($24|0)==(0|0); //@line 206 "modules/graphics/opengl/OpenGL.cpp"
   if (!($25)) {
    $26 = ((($this)) + 60|0); //@line 207 "modules/graphics/opengl/OpenGL.cpp"
    HEAP32[$26>>2] = 6; //@line 207 "modules/graphics/opengl/OpenGL.cpp"
    return; //@line 218 "modules/graphics/opengl/OpenGL.cpp"
   }
   $27 = (_strstr($1,515143)|0); //@line 208 "modules/graphics/opengl/OpenGL.cpp"
   $28 = ($27|0)==(0|0); //@line 208 "modules/graphics/opengl/OpenGL.cpp"
   if (!($28)) {
    $29 = ((($this)) + 60|0); //@line 209 "modules/graphics/opengl/OpenGL.cpp"
    HEAP32[$29>>2] = 7; //@line 209 "modules/graphics/opengl/OpenGL.cpp"
    return; //@line 218 "modules/graphics/opengl/OpenGL.cpp"
   }
   $30 = (_strstr($1,437396)|0); //@line 210 "modules/graphics/opengl/OpenGL.cpp"
   $31 = ($30|0)==(0|0); //@line 210 "modules/graphics/opengl/OpenGL.cpp"
   if (!($31)) {
    $32 = ((($this)) + 60|0); //@line 211 "modules/graphics/opengl/OpenGL.cpp"
    HEAP32[$32>>2] = 8; //@line 211 "modules/graphics/opengl/OpenGL.cpp"
    return; //@line 218 "modules/graphics/opengl/OpenGL.cpp"
   }
   $33 = (_strstr($1,437405)|0); //@line 212 "modules/graphics/opengl/OpenGL.cpp"
   $34 = ($33|0)==(0|0); //@line 212 "modules/graphics/opengl/OpenGL.cpp"
   if (!($34)) {
    $35 = ((($this)) + 60|0); //@line 213 "modules/graphics/opengl/OpenGL.cpp"
    HEAP32[$35>>2] = 9; //@line 213 "modules/graphics/opengl/OpenGL.cpp"
    return; //@line 218 "modules/graphics/opengl/OpenGL.cpp"
   }
   $36 = (_strstr($1,437414)|0); //@line 214 "modules/graphics/opengl/OpenGL.cpp"
   $37 = ($36|0)==(0|0); //@line 214 "modules/graphics/opengl/OpenGL.cpp"
   $38 = ((($this)) + 60|0); //@line 215 "modules/graphics/opengl/OpenGL.cpp"
   if ($37) {
    HEAP32[$38>>2] = 11; //@line 217 "modules/graphics/opengl/OpenGL.cpp"
    return; //@line 218 "modules/graphics/opengl/OpenGL.cpp"
   } else {
    HEAP32[$38>>2] = 10; //@line 215 "modules/graphics/opengl/OpenGL.cpp"
    return; //@line 218 "modules/graphics/opengl/OpenGL.cpp"
   }
  }
 }
 $20 = ((($this)) + 60|0); //@line 203 "modules/graphics/opengl/OpenGL.cpp"
 HEAP32[$20>>2] = 4; //@line 203 "modules/graphics/opengl/OpenGL.cpp"
 return; //@line 218 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL12initMatricesEv($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 64|0;
 $1 = sp;
 $2 = HEAP32[$this>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($this)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$3>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($4|0)==($2|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($5)) {
  $7 = $4;
  while(1) {
   $6 = ((($7)) + -64|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$3>>2] = $6; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZN4love7Matrix4D2Ev($6); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $8 = HEAP32[$3>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $9 = ($8|0)==($2|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($9) {
    break;
   } else {
    $7 = $8;
   }
  }
 }
 $10 = ((($this)) + 12|0); //@line 297 "modules/graphics/opengl/OpenGL.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ((($this)) + 16|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = HEAP32[$12>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = ($13|0)==($11|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($14)) {
  $16 = $13;
  while(1) {
   $15 = ((($16)) + -64|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$12>>2] = $15; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZN4love7Matrix4D2Ev($15); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $17 = HEAP32[$12>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $18 = ($17|0)==($11|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($18) {
    break;
   } else {
    $16 = $17;
   }
  }
 }
 __ZN4love7Matrix4C2Ev($0); //@line 299 "modules/graphics/opengl/OpenGL.cpp"
 $19 = HEAP32[$3>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $21 = HEAP32[$20>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $22 = ($19>>>0)<($21>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($22) {
  dest=$19; src=$0; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = HEAP32[$3>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $24 = ((($23)) + 64|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$3>>2] = $24; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 } else {
  __THREW__ = 0;
  invoke_vii(4012,($this|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $25 = __THREW__; __THREW__ = 0;
  $26 = $25&1;
  if ($26) {
   $35 = ___cxa_find_matching_catch_2()|0;
   $36 = tempRet0;
   __ZN4love7Matrix4D2Ev($0); //@line 299 "modules/graphics/opengl/OpenGL.cpp"
   $$0 = $36;$$01 = $35;
   ___resumeException($$01|0);
   // unreachable;
  }
 }
 __ZN4love7Matrix4D2Ev($0); //@line 299 "modules/graphics/opengl/OpenGL.cpp"
 __ZN4love7Matrix4C2Ev($1); //@line 300 "modules/graphics/opengl/OpenGL.cpp"
 $27 = HEAP32[$12>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $28 = ((($this)) + 20|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $29 = HEAP32[$28>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = ($27>>>0)<($29>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($30) {
  dest=$27; src=$1; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $31 = HEAP32[$12>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $32 = ((($31)) + 64|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$12>>2] = $32; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  __ZN4love7Matrix4D2Ev($1); //@line 300 "modules/graphics/opengl/OpenGL.cpp"
  STACKTOP = sp;return; //@line 301 "modules/graphics/opengl/OpenGL.cpp"
 }
 __THREW__ = 0;
 invoke_vii(4012,($10|0),($1|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $33 = __THREW__; __THREW__ = 0;
 $34 = $33&1;
 if (!($34)) {
  __ZN4love7Matrix4D2Ev($1); //@line 300 "modules/graphics/opengl/OpenGL.cpp"
  STACKTOP = sp;return; //@line 301 "modules/graphics/opengl/OpenGL.cpp"
 }
 $37 = ___cxa_find_matching_catch_2()|0;
 $38 = tempRet0;
 __ZN4love7Matrix4D2Ev($1); //@line 300 "modules/graphics/opengl/OpenGL.cpp"
 $$0 = $38;$$01 = $37;
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN4love8graphics6opengl6OpenGL12setupContextEv($this) {
 $this = $this|0;
 var $$in = 0, $$neg11 = 0, $$pre = 0, $$pre$i = 0, $$pre12$i = 0, $$pre14 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__new_last2$i$i$i6 = 0, $exitcond$i = 0, $glcolor = 0, $i$012 = 0, $i$02$i = 0, $maxattachments$i = 0;
 var $maxdrawbuffers$i = 0, $maxvertexattribs = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i7 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $maxattachments$i = sp + 28|0;
 $maxdrawbuffers$i = sp + 24|0;
 $glcolor = sp + 8|0;
 $maxvertexattribs = sp + 4|0;
 $0 = sp;
 $1 = ((($this)) + 38|0); //@line 111 "modules/graphics/opengl/OpenGL.cpp"
 $2 = HEAP8[$1>>0]|0; //@line 111 "modules/graphics/opengl/OpenGL.cpp"
 $3 = ($2<<24>>24)==(0); //@line 111 "modules/graphics/opengl/OpenGL.cpp"
 if ($3) {
  STACKTOP = sp;return; //@line 170 "modules/graphics/opengl/OpenGL.cpp"
 }
 $4 = HEAP8[1252239]|0; //@line 264 "modules/graphics/opengl/OpenGL.cpp"
 $5 = ($4<<24>>24)==(0); //@line 264 "modules/graphics/opengl/OpenGL.cpp"
 $6 = ((($this)) + 40|0); //@line 265 "modules/graphics/opengl/OpenGL.cpp"
 if ($5) {
  HEAPF32[$6>>2] = 1.0; //@line 267 "modules/graphics/opengl/OpenGL.cpp"
 } else {
  $7 = HEAP32[165575]|0; //@line 72 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$7 & 8191](34047,$6); //@line 72 "./libraries/glad/gladfuncs.hpp"
 }
 $8 = ((($this)) + 44|0); //@line 269 "modules/graphics/opengl/OpenGL.cpp"
 $9 = HEAP32[165587]|0; //@line 73 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$9 & 8191](3379,$8); //@line 73 "./libraries/glad/gladfuncs.hpp"
 HEAP32[$maxattachments$i>>2] = 1; //@line 271 "modules/graphics/opengl/OpenGL.cpp"
 HEAP32[$maxdrawbuffers$i>>2] = 1; //@line 272 "modules/graphics/opengl/OpenGL.cpp"
 $10 = HEAP8[1252013]|0; //@line 274 "modules/graphics/opengl/OpenGL.cpp"
 $11 = HEAP8[1252000]|0; //@line 274 "modules/graphics/opengl/OpenGL.cpp"
 $12 = $11 | $10; //@line 274 "modules/graphics/opengl/OpenGL.cpp"
 $13 = ($12<<24>>24)==(0); //@line 274 "modules/graphics/opengl/OpenGL.cpp"
 if ($13) {
  $17 = 1;$18 = 1;
 } else {
  $14 = HEAP32[165587]|0; //@line 73 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$14 & 8191](36063,$maxattachments$i); //@line 73 "./libraries/glad/gladfuncs.hpp"
  $15 = HEAP32[165587]|0; //@line 73 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$15 & 8191](34852,$maxdrawbuffers$i); //@line 73 "./libraries/glad/gladfuncs.hpp"
  $$pre$i = HEAP32[$maxdrawbuffers$i>>2]|0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $$pre12$i = HEAP32[$maxattachments$i>>2]|0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $$pre$i;$18 = $$pre12$i;
 }
 $16 = ($17|0)<($18|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $19 = $16 ? $17 : $18; //@line 280 "modules/graphics/opengl/OpenGL.cpp"
 $20 = ((($this)) + 48|0); //@line 280 "modules/graphics/opengl/OpenGL.cpp"
 HEAP32[$20>>2] = $19; //@line 280 "modules/graphics/opengl/OpenGL.cpp"
 $21 = HEAP8[1252013]|0; //@line 282 "modules/graphics/opengl/OpenGL.cpp"
 $22 = HEAP8[1252002]|0; //@line 282 "modules/graphics/opengl/OpenGL.cpp"
 $23 = $22 | $21; //@line 282 "modules/graphics/opengl/OpenGL.cpp"
 $24 = HEAP8[1252065]|0; //@line 282 "modules/graphics/opengl/OpenGL.cpp"
 $25 = $23 | $24; //@line 282 "modules/graphics/opengl/OpenGL.cpp"
 $26 = HEAP8[1252183]|0; //@line 283 "modules/graphics/opengl/OpenGL.cpp"
 $27 = $25 | $26; //@line 282 "modules/graphics/opengl/OpenGL.cpp"
 $28 = HEAP8[1252362]|0; //@line 283 "modules/graphics/opengl/OpenGL.cpp"
 $29 = $27 | $28; //@line 282 "modules/graphics/opengl/OpenGL.cpp"
 $30 = HEAP8[1252345]|0; //@line 284 "modules/graphics/opengl/OpenGL.cpp"
 $31 = $29 | $30; //@line 282 "modules/graphics/opengl/OpenGL.cpp"
 $32 = ($31<<24>>24)==(0); //@line 282 "modules/graphics/opengl/OpenGL.cpp"
 $33 = ((($this)) + 52|0); //@line 286 "modules/graphics/opengl/OpenGL.cpp"
 if ($32) {
  HEAP32[$33>>2] = 0; //@line 289 "modules/graphics/opengl/OpenGL.cpp"
 } else {
  $34 = HEAP32[165587]|0; //@line 73 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$34 & 8191](36183,$33); //@line 73 "./libraries/glad/gladfuncs.hpp"
 }
 $35 = ((($this)) + 56|0); //@line 291 "modules/graphics/opengl/OpenGL.cpp"
 $36 = HEAP32[165587]|0; //@line 73 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$36 & 8191](35661,$35); //@line 73 "./libraries/glad/gladfuncs.hpp"
 ;HEAP32[$glcolor>>2]=HEAP32[25024>>2]|0;HEAP32[$glcolor+4>>2]=HEAP32[25024+4>>2]|0;HEAP32[$glcolor+8>>2]=HEAP32[25024+8>>2]|0;HEAP32[$glcolor+12>>2]=HEAP32[25024+12>>2]|0; //@line 116 "modules/graphics/opengl/OpenGL.cpp"
 $37 = HEAP32[165284]|0; //@line 584 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$37 & 8191](2,$glcolor); //@line 584 "./libraries/glad/gladfuncs.hpp"
 $38 = HEAP32[165284]|0; //@line 584 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$38 & 8191](3,$glcolor); //@line 584 "./libraries/glad/gladfuncs.hpp"
 HEAP32[$maxvertexattribs>>2] = 1; //@line 120 "modules/graphics/opengl/OpenGL.cpp"
 $39 = HEAP32[165587]|0; //@line 73 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$39 & 8191](34921,$maxvertexattribs); //@line 73 "./libraries/glad/gladfuncs.hpp"
 $40 = HEAP32[$maxvertexattribs>>2]|0; //@line 122 "modules/graphics/opengl/OpenGL.cpp"
 $41 = (_bitshift64Shl(1,0,($40|0))|0); //@line 122 "modules/graphics/opengl/OpenGL.cpp"
 $42 = tempRet0; //@line 122 "modules/graphics/opengl/OpenGL.cpp"
 $43 = (_i64Add(($41|0),($42|0),-1,0)|0); //@line 122 "modules/graphics/opengl/OpenGL.cpp"
 $44 = tempRet0; //@line 122 "modules/graphics/opengl/OpenGL.cpp"
 $45 = ((($this)) + 80|0); //@line 122 "modules/graphics/opengl/OpenGL.cpp"
 HEAP32[$45>>2] = $43; //@line 122 "modules/graphics/opengl/OpenGL.cpp"
 $46 = ($43|0)==(0); //@line 395 "modules/graphics/opengl/OpenGL.cpp"
 if (!($46)) {
  $i$02$i = 0;
  while(1) {
   $49 = 1 << $i$02$i; //@line 402 "modules/graphics/opengl/OpenGL.cpp"
   $50 = $49 & $43; //@line 404 "modules/graphics/opengl/OpenGL.cpp"
   $51 = ($50|0)==(0); //@line 404 "modules/graphics/opengl/OpenGL.cpp"
   if (!($51)) {
    $52 = HEAP32[164811]|0; //@line 512 "./libraries/glad/gladfuncs.hpp"
    FUNCTION_TABLE_vi[$52 & 8191]($i$02$i); //@line 512 "./libraries/glad/gladfuncs.hpp"
   }
   $53 = (($i$02$i) + 1)|0; //@line 400 "modules/graphics/opengl/OpenGL.cpp"
   $exitcond$i = ($53|0)==(32); //@line 400 "modules/graphics/opengl/OpenGL.cpp"
   if ($exitcond$i) {
    break;
   } else {
    $i$02$i = $53;
   }
  }
  HEAP32[$45>>2] = 0; //@line 413 "modules/graphics/opengl/OpenGL.cpp"
  $47 = $43 & 4; //@line 419 "modules/graphics/opengl/OpenGL.cpp"
  $48 = ($47|0)==(0); //@line 419 "modules/graphics/opengl/OpenGL.cpp"
  if (!($48)) {
   $54 = HEAP32[165163]|0; //@line 583 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_vidddd[$54 & 8191](2,1.0,1.0,1.0,1.0); //@line 583 "./libraries/glad/gladfuncs.hpp"
  }
 }
 $55 = ((($this)) + 84|0); //@line 126 "modules/graphics/opengl/OpenGL.cpp"
 $56 = HEAP32[165587]|0; //@line 73 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$56 & 8191](2978,$55); //@line 73 "./libraries/glad/gladfuncs.hpp"
 $57 = ((($this)) + 100|0); //@line 130 "modules/graphics/opengl/OpenGL.cpp"
 $58 = HEAP32[165587]|0; //@line 73 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$58 & 8191](3088,$57); //@line 73 "./libraries/glad/gladfuncs.hpp"
 $59 = ((($this)) + 96|0); //@line 131 "modules/graphics/opengl/OpenGL.cpp"
 $60 = HEAP32[$59>>2]|0; //@line 131 "modules/graphics/opengl/OpenGL.cpp"
 $61 = ((($this)) + 104|0); //@line 131 "modules/graphics/opengl/OpenGL.cpp"
 $62 = HEAP32[$61>>2]|0; //@line 131 "modules/graphics/opengl/OpenGL.cpp"
 $63 = ((($this)) + 112|0); //@line 131 "modules/graphics/opengl/OpenGL.cpp"
 $64 = HEAP32[$63>>2]|0; //@line 131 "modules/graphics/opengl/OpenGL.cpp"
 $$neg11 = (($60) - ($62))|0; //@line 131 "modules/graphics/opengl/OpenGL.cpp"
 $65 = (($$neg11) - ($64))|0; //@line 131 "modules/graphics/opengl/OpenGL.cpp"
 HEAP32[$61>>2] = $65; //@line 131 "modules/graphics/opengl/OpenGL.cpp"
 $66 = HEAP8[1251994]|0; //@line 133 "modules/graphics/opengl/OpenGL.cpp"
 $67 = ($66<<24>>24)==(0); //@line 133 "modules/graphics/opengl/OpenGL.cpp"
 $68 = ((($this)) + 116|0); //@line 134 "modules/graphics/opengl/OpenGL.cpp"
 if ($67) {
  HEAPF32[$68>>2] = 1.0; //@line 136 "modules/graphics/opengl/OpenGL.cpp"
 } else {
  $69 = HEAP32[165575]|0; //@line 72 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$69 & 8191](2833,$68); //@line 72 "./libraries/glad/gladfuncs.hpp"
 }
 $70 = HEAP8[1252002]|0; //@line 138 "modules/graphics/opengl/OpenGL.cpp"
 $71 = HEAP8[1252066]|0; //@line 138 "modules/graphics/opengl/OpenGL.cpp"
 $72 = $71 | $70; //@line 138 "modules/graphics/opengl/OpenGL.cpp"
 $73 = HEAP8[1252186]|0; //@line 138 "modules/graphics/opengl/OpenGL.cpp"
 $74 = $72 | $73; //@line 138 "modules/graphics/opengl/OpenGL.cpp"
 $75 = HEAP8[1252225]|0; //@line 139 "modules/graphics/opengl/OpenGL.cpp"
 $76 = $74 | $75; //@line 138 "modules/graphics/opengl/OpenGL.cpp"
 $77 = ($76<<24>>24)==(0); //@line 138 "modules/graphics/opengl/OpenGL.cpp"
 if ($77) {
  $$sink = 0;
 } else {
  $78 = HEAP32[164774]|0; //@line 80 "./libraries/glad/gladfuncs.hpp"
  $79 = (FUNCTION_TABLE_ii[$78 & 8191](36281)|0); //@line 80 "./libraries/glad/gladfuncs.hpp"
  $80 = ($79<<24>>24)==(1); //@line 141 "modules/graphics/opengl/OpenGL.cpp"
  $81 = $80&1; //@line 141 "modules/graphics/opengl/OpenGL.cpp"
  $$sink = $81;
 }
 $82 = ((($this)) + 120|0); //@line 144 "modules/graphics/opengl/OpenGL.cpp"
 HEAP8[$82>>0] = $$sink; //@line 144 "modules/graphics/opengl/OpenGL.cpp"
 $83 = ((($this)) + 64|0); //@line 147 "modules/graphics/opengl/OpenGL.cpp"
 $84 = HEAP32[$83>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $85 = ((($this)) + 68|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $86 = HEAP32[$85>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $87 = ($86|0)==($84|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $88 = $84; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($87) {
  $$in = $86;
 } else {
  $scevgep$i$i$i = ((($86)) + -4|0);
  $89 = $scevgep$i$i$i;
  $90 = (($89) - ($88))|0;
  $91 = $90 >>> 2;
  $92 = $91 ^ -1;
  $scevgep4$i$i$i = (($86) + ($92<<2)|0);
  HEAP32[$85>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $scevgep4$i$i$i;
 }
 $93 = $$in; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $94 = HEAP32[$35>>2]|0; //@line 148 "modules/graphics/opengl/OpenGL.cpp"
 HEAP32[$0>>2] = 0; //@line 148 "modules/graphics/opengl/OpenGL.cpp"
 $95 = (($93) - ($88))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $96 = $95 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $97 = ($94>>>0)>($96>>>0); //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($97) {
  $98 = (($94) - ($96))|0; //@line 2010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  __ZNSt3__16vectorIjNS_9allocatorIjEEE8__appendEjRKj($83,$98,$0); //@line 2010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$85>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre14 = HEAP32[$83>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $108 = $$pre;$109 = $$pre14;
 } else {
  $99 = ($94>>>0)<($96>>>0); //@line 2011 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($99) {
   $100 = (($84) + ($94<<2)|0); //@line 2012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $101 = ($$in|0)==($100|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($101) {
    $108 = $93;$109 = $88;
   } else {
    $__new_last2$i$i$i6 = $100; //@line 818 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $scevgep$i$i$i7 = ((($$in)) + -4|0);
    $102 = $scevgep$i$i$i7;
    $103 = (($102) - ($__new_last2$i$i$i6))|0;
    $104 = $103 >>> 2;
    $105 = $104 ^ -1;
    $scevgep4$i$i$i8 = (($$in) + ($105<<2)|0);
    HEAP32[$85>>2] = $scevgep4$i$i$i8; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $106 = $scevgep4$i$i$i8; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $108 = $106;$109 = $88;
   }
  } else {
   $108 = $93;$109 = $88;
  }
 }
 $107 = (($108) - ($109))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $110 = ($107|0)>(0); //@line 150 "modules/graphics/opengl/OpenGL.cpp"
 if ($110) {
  $i$012 = 0;
  while(1) {
   $117 = (($i$012) + 33984)|0; //@line 152 "modules/graphics/opengl/OpenGL.cpp"
   $118 = HEAP32[165288]|0; //@line 381 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_vi[$118 & 8191]($117); //@line 381 "./libraries/glad/gladfuncs.hpp"
   $119 = HEAP32[164685]|0; //@line 353 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_vii[$119 & 8191](3553,0); //@line 353 "./libraries/glad/gladfuncs.hpp"
   $120 = (($i$012) + 1)|0; //@line 150 "modules/graphics/opengl/OpenGL.cpp"
   $121 = HEAP32[$85>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $122 = HEAP32[$83>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $123 = (($121) - ($122))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $124 = $123 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $125 = ($120|0)<($124|0); //@line 150 "modules/graphics/opengl/OpenGL.cpp"
   if ($125) {
    $i$012 = $120;
   } else {
    break;
   }
  }
 }
 $111 = HEAP32[165288]|0; //@line 381 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$111 & 8191](33984); //@line 381 "./libraries/glad/gladfuncs.hpp"
 $112 = ((($this)) + 76|0); //@line 157 "modules/graphics/opengl/OpenGL.cpp"
 HEAP32[$112>>2] = 0; //@line 157 "modules/graphics/opengl/OpenGL.cpp"
 __ZN4love8graphics6opengl6OpenGL20createDefaultTextureEv($this); //@line 159 "modules/graphics/opengl/OpenGL.cpp"
 $113 = ((($this)) + 128|0); //@line 163 "modules/graphics/opengl/OpenGL.cpp"
 __ZN4love7Matrix414setTranslationEff($113,nan,nan); //@line 163 "modules/graphics/opengl/OpenGL.cpp"
 $114 = ((($this)) + 192|0); //@line 164 "modules/graphics/opengl/OpenGL.cpp"
 __ZN4love7Matrix414setTranslationEff($114,nan,nan); //@line 164 "modules/graphics/opengl/OpenGL.cpp"
 $115 = HEAP8[1251994]|0; //@line 166 "modules/graphics/opengl/OpenGL.cpp"
 $116 = ($115<<24>>24)==(0); //@line 166 "modules/graphics/opengl/OpenGL.cpp"
 if (!($116)) {
  $126 = HEAP32[164981]|0; //@line 329 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$126 & 8191](5888); //@line 329 "./libraries/glad/gladfuncs.hpp"
 }
 HEAP8[$1>>0] = 1; //@line 169 "modules/graphics/opengl/OpenGL.cpp"
 STACKTOP = sp;return; //@line 170 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZNSt3__16vectorIjNS_9allocatorIjEEE8__appendEjRKj($this,$__n,$__x) {
 $this = $this|0;
 $__n = $__n|0;
 $__x = $__x|0;
 var $$0$i = 0, $$0$i2 = 0, $$0$i45 = 0, $$0$i46 = 0, $$in = 0, $$pre = 0, $$pre7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $scevgep$i = 0, $scevgep$i3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 1063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($this)) + 4|0); //@line 1063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($1) - ($3))|0; //@line 1063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = $4 >> 2; //@line 1063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)<($__n>>>0); //@line 1063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = $3; //@line 1010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($6)) {
  $$0$i = $__n;$9 = $7;
  while(1) {
   $8 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$9>>2] = $8; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $10 = ((($9)) + 4|0); //@line 1011 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $11 = (($$0$i) + -1)|0; //@line 1012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $12 = ($11|0)==(0); //@line 1014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($12) {
    break;
   } else {
    $$0$i = $11;$9 = $10;
   }
  }
  $scevgep$i = (($7) + ($__n<<2)|0); //@line 1007 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$2>>2] = $scevgep$i; //@line 1010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  return; //@line 1072 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $13 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = (($3) - ($13))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = $14 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = (($15) + ($__n))|0; //@line 1068 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ($16>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($17) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$0>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre7 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre7;$19 = $$pre;
 } else {
  $$in = $13;$19 = $1;
 }
 $18 = (($19) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = $18 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = ($20>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($21) {
  $25 = $18 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $26 = ($25>>>0)<($16>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $27 = $26 ? $16 : $25; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $28 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $29 = (($28) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $30 = $29 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $31 = ($27|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($31) {
   $$0$i46 = 0;$35 = 0;$36 = $30;$46 = $28;
  } else {
   $$0$i45 = $27;$54 = $28;$55 = $30;
   label = 9;
  }
 } else {
  $22 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = (($22) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $24 = $23 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i45 = 1073741823;$54 = $22;$55 = $24;
  label = 9;
 }
 if ((label|0) == 9) {
  $32 = $$0$i45 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $33 = (__Znwj($32)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i46 = $$0$i45;$35 = $33;$36 = $55;$46 = $54;
 }
 $34 = (($35) + ($36<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $37 = (($35) + ($$0$i46<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $$0$i2 = $__n;$39 = $34;
 while(1) {
  $38 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$39>>2] = $38; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $40 = ((($39)) + 4|0); //@line 237 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $41 = (($$0$i2) + -1)|0; //@line 238 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $42 = ($41|0)==(0); //@line 239 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($42) {
   break;
  } else {
   $$0$i2 = $41;$39 = $40;
  }
 }
 $43 = $37; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $scevgep$i3 = (($34) + ($__n<<2)|0); //@line 234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $44 = $scevgep$i3; //@line 237 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $45 = (($46) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = $45 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $48 = (0 - ($47))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $49 = (($34) + ($48<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $50 = $49; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $51 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($49|0),($51|0),($45|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $50; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$2>>2] = $44; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $43; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $52 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($52) {
  return; //@line 1072 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $53 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($53); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1072 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl6OpenGL20createDefaultTextureEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $pix = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $pix = sp;
 $0 = ((($this)) + 76|0); //@line 310 "modules/graphics/opengl/OpenGL.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 310 "modules/graphics/opengl/OpenGL.cpp"
 $2 = ((($this)) + 64|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($3) + ($1<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$4>>2]|0; //@line 310 "modules/graphics/opengl/OpenGL.cpp"
 $6 = ((($this)) + 124|0); //@line 312 "modules/graphics/opengl/OpenGL.cpp"
 $7 = HEAP32[164922]|0; //@line 355 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$7 & 8191](1,$6); //@line 355 "./libraries/glad/gladfuncs.hpp"
 $8 = HEAP32[$6>>2]|0; //@line 313 "modules/graphics/opengl/OpenGL.cpp"
 $9 = HEAP32[$0>>2]|0; //@line 525 "modules/graphics/opengl/OpenGL.cpp"
 $10 = HEAP32[$2>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = (($10) + ($9<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = HEAP32[$11>>2]|0; //@line 525 "modules/graphics/opengl/OpenGL.cpp"
 $13 = ($12|0)==($8|0); //@line 525 "modules/graphics/opengl/OpenGL.cpp"
 if (!($13)) {
  HEAP32[$11>>2] = $8; //@line 527 "modules/graphics/opengl/OpenGL.cpp"
  $14 = HEAP32[164685]|0; //@line 353 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$14 & 8191](3553,$8); //@line 353 "./libraries/glad/gladfuncs.hpp"
 }
 $15 = HEAP32[165108]|0; //@line 44 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$15 & 8191](3553,10241,9728); //@line 44 "./libraries/glad/gladfuncs.hpp"
 $16 = HEAP32[165108]|0; //@line 44 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$16 & 8191](3553,10240,9728); //@line 44 "./libraries/glad/gladfuncs.hpp"
 $17 = HEAP32[165108]|0; //@line 44 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$17 & 8191](3553,10242,10497); //@line 44 "./libraries/glad/gladfuncs.hpp"
 $18 = HEAP32[165108]|0; //@line 44 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$18 & 8191](3553,10243,10497); //@line 44 "./libraries/glad/gladfuncs.hpp"
 HEAP32[$pix>>2] = -1; //@line 321 "modules/graphics/opengl/OpenGL.cpp"
 $19 = HEAP32[165602]|0; //@line 47 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiiiiiiii[$19 & 8191](3553,0,6408,1,1,0,6408,5121,$pix); //@line 47 "./libraries/glad/gladfuncs.hpp"
 $20 = HEAP32[$0>>2]|0; //@line 525 "modules/graphics/opengl/OpenGL.cpp"
 $21 = HEAP32[$2>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $22 = (($21) + ($20<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $23 = HEAP32[$22>>2]|0; //@line 525 "modules/graphics/opengl/OpenGL.cpp"
 $24 = ($23|0)==($5|0); //@line 525 "modules/graphics/opengl/OpenGL.cpp"
 if ($24) {
  STACKTOP = sp;return; //@line 325 "modules/graphics/opengl/OpenGL.cpp"
 }
 HEAP32[$22>>2] = $5; //@line 527 "modules/graphics/opengl/OpenGL.cpp"
 $25 = HEAP32[164685]|0; //@line 353 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$25 & 8191](3553,$5); //@line 353 "./libraries/glad/gladfuncs.hpp"
 STACKTOP = sp;return; //@line 325 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL21useVertexAttribArraysEj($this,$arraybits) {
 $this = $this|0;
 $arraybits = $arraybits|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$02 = 0, $or$cond = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 80|0); //@line 393 "modules/graphics/opengl/OpenGL.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 393 "modules/graphics/opengl/OpenGL.cpp"
 $2 = $1 ^ $arraybits; //@line 393 "modules/graphics/opengl/OpenGL.cpp"
 $3 = ($1|0)==($arraybits|0); //@line 395 "modules/graphics/opengl/OpenGL.cpp"
 if ($3) {
  return; //@line 421 "modules/graphics/opengl/OpenGL.cpp"
 } else {
  $i$02 = 0;
 }
 while(1) {
  $8 = 1 << $i$02; //@line 402 "modules/graphics/opengl/OpenGL.cpp"
  $9 = $8 & $2; //@line 404 "modules/graphics/opengl/OpenGL.cpp"
  $10 = ($9|0)==(0); //@line 404 "modules/graphics/opengl/OpenGL.cpp"
  do {
   if (!($10)) {
    $11 = $8 & $arraybits; //@line 406 "modules/graphics/opengl/OpenGL.cpp"
    $12 = ($11|0)==(0); //@line 406 "modules/graphics/opengl/OpenGL.cpp"
    if ($12) {
     $14 = HEAP32[164811]|0; //@line 512 "./libraries/glad/gladfuncs.hpp"
     FUNCTION_TABLE_vi[$14 & 8191]($i$02); //@line 512 "./libraries/glad/gladfuncs.hpp"
     break;
    } else {
     $13 = HEAP32[165289]|0; //@line 513 "./libraries/glad/gladfuncs.hpp"
     FUNCTION_TABLE_vi[$13 & 8191]($i$02); //@line 513 "./libraries/glad/gladfuncs.hpp"
     break;
    }
   }
  } while(0);
  $15 = (($i$02) + 1)|0; //@line 400 "modules/graphics/opengl/OpenGL.cpp"
  $exitcond = ($15|0)==(32); //@line 400 "modules/graphics/opengl/OpenGL.cpp"
  if ($exitcond) {
   break;
  } else {
   $i$02 = $15;
  }
 }
 HEAP32[$0>>2] = $arraybits; //@line 413 "modules/graphics/opengl/OpenGL.cpp"
 $4 = $2 & 4; //@line 419 "modules/graphics/opengl/OpenGL.cpp"
 $5 = ($4|0)!=(0); //@line 419 "modules/graphics/opengl/OpenGL.cpp"
 $6 = $arraybits & 4; //@line 419 "modules/graphics/opengl/OpenGL.cpp"
 $7 = ($6|0)==(0); //@line 419 "modules/graphics/opengl/OpenGL.cpp"
 $or$cond = $7 & $5; //@line 419 "modules/graphics/opengl/OpenGL.cpp"
 if (!($or$cond)) {
  return; //@line 421 "modules/graphics/opengl/OpenGL.cpp"
 }
 $16 = HEAP32[165163]|0; //@line 583 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vidddd[$16 & 8191](2,1.0,1.0,1.0,1.0); //@line 583 "./libraries/glad/gladfuncs.hpp"
 return; //@line 421 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL13deInitContextEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 38|0); //@line 174 "modules/graphics/opengl/OpenGL.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 174 "modules/graphics/opengl/OpenGL.cpp"
 $2 = ($1<<24>>24)==(0); //@line 174 "modules/graphics/opengl/OpenGL.cpp"
 if ($2) {
  return; //@line 181 "modules/graphics/opengl/OpenGL.cpp"
 }
 $3 = ((($this)) + 124|0); //@line 177 "modules/graphics/opengl/OpenGL.cpp"
 $4 = HEAP32[165157]|0; //@line 354 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$4 & 8191](1,$3); //@line 354 "./libraries/glad/gladfuncs.hpp"
 HEAP32[$3>>2] = 0; //@line 178 "modules/graphics/opengl/OpenGL.cpp"
 HEAP8[$0>>0] = 0; //@line 180 "modules/graphics/opengl/OpenGL.cpp"
 return; //@line 181 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL11bindTextureEj($this,$texture) {
 $this = $this|0;
 $texture = $texture|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 76|0); //@line 525 "modules/graphics/opengl/OpenGL.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 525 "modules/graphics/opengl/OpenGL.cpp"
 $2 = ((($this)) + 64|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($3) + ($1<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$4>>2]|0; //@line 525 "modules/graphics/opengl/OpenGL.cpp"
 $6 = ($5|0)==($texture|0); //@line 525 "modules/graphics/opengl/OpenGL.cpp"
 if ($6) {
  return; //@line 530 "modules/graphics/opengl/OpenGL.cpp"
 }
 HEAP32[$4>>2] = $texture; //@line 527 "modules/graphics/opengl/OpenGL.cpp"
 $7 = HEAP32[164685]|0; //@line 353 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$7 & 8191](3553,$texture); //@line 353 "./libraries/glad/gladfuncs.hpp"
 return; //@line 530 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL13pushTransformEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -64|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $4 = HEAP32[$3>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($1|0)==($4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($5) {
  __ZNSt3__16vectorIN4love7Matrix4ENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($this,$2); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  return; //@line 330 "modules/graphics/opengl/OpenGL.cpp"
 } else {
  dest=$1; src=$2; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $6 = HEAP32[$0>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $7 = ((($6)) + 64|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$0>>2] = $7; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  return; //@line 330 "modules/graphics/opengl/OpenGL.cpp"
 }
}
function __ZNSt3__16vectorIN4love7Matrix4ENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$01$i$i = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__v$sroa$5$0 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 6; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(67108863); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = $$in; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = HEAP32[$8>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = (($9) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $10 >> 6; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($11>>>0)<(33554431); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $17 = $10 >> 5; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ($17>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $19 = $18 ? $5 : $17; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = (($20) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = $21 >> 6; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = ($19|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $24 = $20; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($23) {
   $$0$i2 = 0;$28 = 0;$29 = $22;$36 = $24;
  } else {
   $$0$i1 = $19;$54 = $24;$55 = $22;
   label = 6;
  }
 } else {
  $13 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = $14 >> 6; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = $13; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $$0$i1 = 67108863;$54 = $16;$55 = $15;
  label = 6;
 }
 if ((label|0) == 6) {
  $25 = $$0$i1 << 6; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (__Znwj($25)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$28 = $26;$29 = $55;$36 = $54;
 }
 $27 = (($28) + ($29<<6)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = $27; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $31 = (($28) + ($$0$i2<<6)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $32 = $31; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 dest=$27; src=$__x; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = ((($27)) + 64|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $34 = $33; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $35 = ($36|0)==($7|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($35) {
  $44 = $30;$46 = $$in;$48 = $36;
 } else {
  $$01$i$i = $36;$38 = $27;$__v$sroa$5$0 = $30;
  while(1) {
   $37 = ((($38)) + -64|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $39 = ((($$01$i$i)) + -64|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   dest=$37; src=$39; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $40 = $__v$sroa$5$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $41 = ((($40)) + -64|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $42 = $41; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $43 = ($39|0)==($7|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if ($43) {
    $$lcssa = $42;
    break;
   } else {
    $$01$i$i = $39;$38 = $41;$__v$sroa$5$0 = $42;
   }
  }
  $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $$pre3 = HEAP32[$0>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $44 = $$lcssa;$46 = $$pre$i;$48 = $$pre3;
 }
 HEAP32[$this>>2] = $44; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $34; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$8>>2] = $32; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $45 = $46; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $47 = ($48|0)==($45|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if (!($47)) {
  $50 = $48;
  while(1) {
   $49 = ((($50)) + -64|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   __ZN4love7Matrix4D2Ev($49); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $51 = ($49|0)==($45|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($51) {
    break;
   } else {
    $50 = $49;
   }
  }
 }
 $52 = ($46|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($52) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $53 = $46; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($53); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl6OpenGL12popTransformEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 1666 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 1666 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -64|0); //@line 1666 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $1;
 while(1) {
  $3 = ((($4)) + -64|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$0>>2] = $3; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  __ZN4love7Matrix4D2Ev($3); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $5 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $6 = ($5|0)==($2|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($6) {
   break;
  } else {
   $4 = $5;
  }
 }
 return; //@line 335 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL12getTransformEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($1)) + -64|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 return ($2|0); //@line 339 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL11prepareDrawEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = HEAP32[158581]|0; //@line 347 "modules/graphics/opengl/OpenGL.cpp"
 $1 = ($0|0)==(0|0); //@line 347 "modules/graphics/opengl/OpenGL.cpp"
 if (!($1)) {
  __ZN4love8graphics6opengl6Shader23checkSetBuiltinUniformsEv($0); //@line 348 "modules/graphics/opengl/OpenGL.cpp"
 }
 $2 = HEAP8[1251994]|0; //@line 352 "modules/graphics/opengl/OpenGL.cpp"
 $3 = ($2<<24>>24)==(0); //@line 352 "modules/graphics/opengl/OpenGL.cpp"
 if ($3) {
  return; //@line 377 "modules/graphics/opengl/OpenGL.cpp"
 }
 $4 = ((($this)) + 16|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$4>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ((($5)) + -64|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ((($this)) + 4|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = HEAP32[$7>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ((($8)) + -64|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = ((($this)) + 128|0); //@line 357 "modules/graphics/opengl/OpenGL.cpp"
 $11 = ((($this)) + 192|0); //@line 358 "modules/graphics/opengl/OpenGL.cpp"
 $12 = (__ZNK4love7Matrix411getElementsEv($6)|0); //@line 361 "modules/graphics/opengl/OpenGL.cpp"
 $13 = (__ZNK4love7Matrix411getElementsEv($10)|0); //@line 361 "modules/graphics/opengl/OpenGL.cpp"
 $14 = (_memcmp($12,$13,64)|0); //@line 361 "modules/graphics/opengl/OpenGL.cpp"
 $15 = ($14|0)==(0); //@line 361 "modules/graphics/opengl/OpenGL.cpp"
 if (!($15)) {
  $16 = HEAP32[164981]|0; //@line 329 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$16 & 8191](5889); //@line 329 "./libraries/glad/gladfuncs.hpp"
  $17 = (__ZNK4love7Matrix411getElementsEv($6)|0); //@line 364 "modules/graphics/opengl/OpenGL.cpp"
  $18 = HEAP32[165611]|0; //@line 327 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$18 & 8191]($17); //@line 327 "./libraries/glad/gladfuncs.hpp"
  $19 = HEAP32[164981]|0; //@line 329 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$19 & 8191](5888); //@line 329 "./libraries/glad/gladfuncs.hpp"
  $20 = HEAP32[$4>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ((($20)) + -64|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  dest=$10; src=$21; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 367 "modules/graphics/opengl/OpenGL.cpp"
 }
 $22 = (__ZNK4love7Matrix411getElementsEv($9)|0); //@line 371 "modules/graphics/opengl/OpenGL.cpp"
 $23 = (__ZNK4love7Matrix411getElementsEv($11)|0); //@line 371 "modules/graphics/opengl/OpenGL.cpp"
 $24 = (_memcmp($22,$23,64)|0); //@line 371 "modules/graphics/opengl/OpenGL.cpp"
 $25 = ($24|0)==(0); //@line 371 "modules/graphics/opengl/OpenGL.cpp"
 if ($25) {
  return; //@line 377 "modules/graphics/opengl/OpenGL.cpp"
 }
 $26 = (__ZNK4love7Matrix411getElementsEv($9)|0); //@line 373 "modules/graphics/opengl/OpenGL.cpp"
 $27 = HEAP32[165611]|0; //@line 327 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$27 & 8191]($26); //@line 327 "./libraries/glad/gladfuncs.hpp"
 $28 = HEAP32[$7>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $29 = ((($28)) + -64|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 dest=$11; src=$29; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 374 "modules/graphics/opengl/OpenGL.cpp"
 return; //@line 377 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL10drawArraysEjii($this,$mode,$first,$count) {
 $this = $this|0;
 $mode = $mode|0;
 $first = $first|0;
 $count = $count|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[164759]|0; //@line 343 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$0 & 8191]($mode,$first,$count); //@line 343 "./libraries/glad/gladfuncs.hpp"
 $1 = ((($this)) + 28|0); //@line 382 "modules/graphics/opengl/OpenGL.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 382 "modules/graphics/opengl/OpenGL.cpp"
 $3 = (($2) + 1)|0; //@line 382 "modules/graphics/opengl/OpenGL.cpp"
 HEAP32[$1>>2] = $3; //@line 382 "modules/graphics/opengl/OpenGL.cpp"
 return; //@line 383 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL12drawElementsEjijPKv($this,$mode,$count,$type,$indices) {
 $this = $this|0;
 $mode = $mode|0;
 $count = $count|0;
 $type = $type|0;
 $indices = $indices|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[165086]|0; //@line 344 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiii[$0 & 8191]($mode,$count,$type,$indices); //@line 344 "./libraries/glad/gladfuncs.hpp"
 $1 = ((($this)) + 28|0); //@line 388 "modules/graphics/opengl/OpenGL.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 388 "modules/graphics/opengl/OpenGL.cpp"
 $3 = (($2) + 1)|0; //@line 388 "modules/graphics/opengl/OpenGL.cpp"
 HEAP32[$1>>2] = $3; //@line 388 "modules/graphics/opengl/OpenGL.cpp"
 return; //@line 389 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL11setViewportERKNS2_8ViewportE($this,$v) {
 $this = $this|0;
 $v = $v|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $sum = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$v>>2]|0; //@line 425 "modules/graphics/opengl/OpenGL.cpp"
 $1 = ((($v)) + 4|0); //@line 425 "modules/graphics/opengl/OpenGL.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 425 "modules/graphics/opengl/OpenGL.cpp"
 $3 = ((($v)) + 8|0); //@line 425 "modules/graphics/opengl/OpenGL.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 425 "modules/graphics/opengl/OpenGL.cpp"
 $5 = ((($v)) + 12|0); //@line 425 "modules/graphics/opengl/OpenGL.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 425 "modules/graphics/opengl/OpenGL.cpp"
 $7 = HEAP32[165094]|0; //@line 82 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiii[$7 & 8191]($0,$2,$4,$6); //@line 82 "./libraries/glad/gladfuncs.hpp"
 $8 = ((($this)) + 84|0); //@line 426 "modules/graphics/opengl/OpenGL.cpp"
 ;HEAP32[$8>>2]=HEAP32[$v>>2]|0;HEAP32[$8+4>>2]=HEAP32[$v+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$v+8>>2]|0;HEAP32[$8+12>>2]=HEAP32[$v+12>>2]|0; //@line 426 "modules/graphics/opengl/OpenGL.cpp"
 $9 = HEAP32[158272]|0; //@line 441 "modules/graphics/opengl/OpenGL.cpp"
 $10 = ($9|0)==(0|0); //@line 441 "modules/graphics/opengl/OpenGL.cpp"
 $11 = ((($this)) + 100|0); //@line 442 "modules/graphics/opengl/OpenGL.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 442 "modules/graphics/opengl/OpenGL.cpp"
 $13 = ((($this)) + 104|0); //@line 447 "modules/graphics/opengl/OpenGL.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 447 "modules/graphics/opengl/OpenGL.cpp"
 $15 = ((($this)) + 112|0); //@line 447 "modules/graphics/opengl/OpenGL.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 447 "modules/graphics/opengl/OpenGL.cpp"
 $17 = ((($this)) + 108|0); //@line 447 "modules/graphics/opengl/OpenGL.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 447 "modules/graphics/opengl/OpenGL.cpp"
 if ($10) {
  $20 = ((($this)) + 96|0); //@line 447 "modules/graphics/opengl/OpenGL.cpp"
  $21 = HEAP32[$20>>2]|0; //@line 447 "modules/graphics/opengl/OpenGL.cpp"
  $sum = (($16) + ($14))|0; //@line 447 "modules/graphics/opengl/OpenGL.cpp"
  $22 = (($21) - ($sum))|0; //@line 447 "modules/graphics/opengl/OpenGL.cpp"
  $23 = HEAP32[165383]|0; //@line 41 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiii[$23 & 8191]($12,$22,$18,$16); //@line 41 "./libraries/glad/gladfuncs.hpp"
  return; //@line 432 "modules/graphics/opengl/OpenGL.cpp"
 } else {
  $19 = HEAP32[165383]|0; //@line 41 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiii[$19 & 8191]($12,$14,$18,$16); //@line 41 "./libraries/glad/gladfuncs.hpp"
  return; //@line 432 "modules/graphics/opengl/OpenGL.cpp"
 }
}
function __ZN4love8graphics6opengl6OpenGL10setScissorERKNS2_8ViewportE($this,$v) {
 $this = $this|0;
 $v = $v|0;
 var $$neg2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[158272]|0; //@line 441 "modules/graphics/opengl/OpenGL.cpp"
 $1 = ($0|0)==(0|0); //@line 441 "modules/graphics/opengl/OpenGL.cpp"
 $2 = HEAP32[$v>>2]|0; //@line 442 "modules/graphics/opengl/OpenGL.cpp"
 $3 = ((($v)) + 4|0); //@line 447 "modules/graphics/opengl/OpenGL.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 447 "modules/graphics/opengl/OpenGL.cpp"
 $5 = ((($v)) + 12|0); //@line 447 "modules/graphics/opengl/OpenGL.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 447 "modules/graphics/opengl/OpenGL.cpp"
 $7 = ((($v)) + 8|0); //@line 447 "modules/graphics/opengl/OpenGL.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 447 "modules/graphics/opengl/OpenGL.cpp"
 if ($1) {
  $10 = ((($this)) + 96|0); //@line 447 "modules/graphics/opengl/OpenGL.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 447 "modules/graphics/opengl/OpenGL.cpp"
  $$neg2 = (($11) - ($4))|0; //@line 447 "modules/graphics/opengl/OpenGL.cpp"
  $12 = (($$neg2) - ($6))|0; //@line 447 "modules/graphics/opengl/OpenGL.cpp"
  $13 = HEAP32[165383]|0; //@line 41 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiii[$13 & 8191]($2,$12,$8,$6); //@line 41 "./libraries/glad/gladfuncs.hpp"
 } else {
  $9 = HEAP32[165383]|0; //@line 41 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiii[$9 & 8191]($2,$4,$8,$6); //@line 41 "./libraries/glad/gladfuncs.hpp"
 }
 $14 = ((($this)) + 100|0); //@line 450 "modules/graphics/opengl/OpenGL.cpp"
 ;HEAP32[$14>>2]=HEAP32[$v>>2]|0;HEAP32[$14+4>>2]=HEAP32[$v+4>>2]|0;HEAP32[$14+8>>2]=HEAP32[$v+8>>2]|0;HEAP32[$14+12>>2]=HEAP32[$v+12>>2]|0; //@line 450 "modules/graphics/opengl/OpenGL.cpp"
 return; //@line 451 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZNK4love8graphics6opengl6OpenGL11getViewportEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 84|0); //@line 436 "modules/graphics/opengl/OpenGL.cpp"
 ;HEAP32[$agg$result>>2]=HEAP32[$0>>2]|0;HEAP32[$agg$result+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$agg$result+8>>2]=HEAP32[$0+8>>2]|0;HEAP32[$agg$result+12>>2]=HEAP32[$0+12>>2]|0; //@line 436 "modules/graphics/opengl/OpenGL.cpp"
 return; //@line 436 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL12setPointSizeEf($this,$size) {
 $this = $this|0;
 $size = +$size;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[1251994]|0; //@line 460 "modules/graphics/opengl/OpenGL.cpp"
 $1 = ($0<<24>>24)==(0); //@line 460 "modules/graphics/opengl/OpenGL.cpp"
 if (!($1)) {
  $2 = HEAP32[164694]|0; //@line 39 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vd[$2 & 8191]($size); //@line 39 "./libraries/glad/gladfuncs.hpp"
 }
 $3 = ((($this)) + 116|0); //@line 463 "modules/graphics/opengl/OpenGL.cpp"
 HEAPF32[$3>>2] = $size; //@line 463 "modules/graphics/opengl/OpenGL.cpp"
 return; //@line 464 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZNK4love8graphics6opengl6OpenGL12getPointSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 116|0); //@line 468 "modules/graphics/opengl/OpenGL.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 468 "modules/graphics/opengl/OpenGL.cpp"
 return (+$1); //@line 468 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL18setFramebufferSRGBEb($this,$enable) {
 $this = $this|0;
 $enable = $enable|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 if ($enable) {
  $0 = HEAP32[164832]|0; //@line 57 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$0 & 8191](36281); //@line 57 "./libraries/glad/gladfuncs.hpp"
 } else {
  $1 = HEAP32[165193]|0; //@line 56 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$1 & 8191](36281); //@line 56 "./libraries/glad/gladfuncs.hpp"
 }
 $2 = ((($this)) + 120|0); //@line 478 "modules/graphics/opengl/OpenGL.cpp"
 $3 = $enable&1; //@line 478 "modules/graphics/opengl/OpenGL.cpp"
 HEAP8[$2>>0] = $3; //@line 478 "modules/graphics/opengl/OpenGL.cpp"
 return; //@line 479 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZNK4love8graphics6opengl6OpenGL18hasFramebufferSRGBEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 120|0); //@line 483 "modules/graphics/opengl/OpenGL.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 483 "modules/graphics/opengl/OpenGL.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 483 "modules/graphics/opengl/OpenGL.cpp"
 return ($2|0); //@line 483 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL15bindFramebufferEjj($this,$target,$framebuffer) {
 $this = $this|0;
 $target = $target|0;
 $framebuffer = $framebuffer|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[165202]|0; //@line 667 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$0 & 8191]($target,$framebuffer); //@line 667 "./libraries/glad/gladfuncs.hpp"
 $1 = ($target|0)==(36160); //@line 490 "modules/graphics/opengl/OpenGL.cpp"
 if (!($1)) {
  return; //@line 492 "modules/graphics/opengl/OpenGL.cpp"
 }
 $2 = ((($this)) + 32|0); //@line 491 "modules/graphics/opengl/OpenGL.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 491 "modules/graphics/opengl/OpenGL.cpp"
 $4 = (($3) + 1)|0; //@line 491 "modules/graphics/opengl/OpenGL.cpp"
 HEAP32[$2>>2] = $4; //@line 491 "modules/graphics/opengl/OpenGL.cpp"
 return; //@line 492 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZNK4love8graphics6opengl6OpenGL13getDefaultFBOEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0; //@line 503 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZNK4love8graphics6opengl6OpenGL17getDefaultTextureEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 124|0); //@line 509 "modules/graphics/opengl/OpenGL.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 509 "modules/graphics/opengl/OpenGL.cpp"
 return ($1|0); //@line 509 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL14setTextureUnitEi($this,$textureunit) {
 $this = $this|0;
 $textureunit = $textureunit|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($textureunit|0)<(0); //@line 514 "modules/graphics/opengl/OpenGL.cpp"
 if (!($0)) {
  $1 = ((($this)) + 64|0); //@line 514 "modules/graphics/opengl/OpenGL.cpp"
  $2 = ((($this)) + 68|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $6 = $5 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $7 = ($6>>>0)>($textureunit>>>0); //@line 514 "modules/graphics/opengl/OpenGL.cpp"
  if ($7) {
   $13 = ((($this)) + 76|0); //@line 517 "modules/graphics/opengl/OpenGL.cpp"
   $14 = HEAP32[$13>>2]|0; //@line 517 "modules/graphics/opengl/OpenGL.cpp"
   $15 = ($14|0)==($textureunit|0); //@line 517 "modules/graphics/opengl/OpenGL.cpp"
   if ($15) {
    HEAP32[$13>>2] = $textureunit; //@line 520 "modules/graphics/opengl/OpenGL.cpp"
    STACKTOP = sp;return; //@line 521 "modules/graphics/opengl/OpenGL.cpp"
   }
   $16 = (($textureunit) + 33984)|0; //@line 518 "modules/graphics/opengl/OpenGL.cpp"
   $17 = HEAP32[165288]|0; //@line 381 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_vi[$17 & 8191]($16); //@line 381 "./libraries/glad/gladfuncs.hpp"
   HEAP32[$13>>2] = $textureunit; //@line 520 "modules/graphics/opengl/OpenGL.cpp"
   STACKTOP = sp;return; //@line 521 "modules/graphics/opengl/OpenGL.cpp"
  }
 }
 $8 = (___cxa_allocate_exception(16)|0); //@line 515 "modules/graphics/opengl/OpenGL.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $textureunit; //@line 515 "modules/graphics/opengl/OpenGL.cpp"
 invoke_viii(3349,($8|0),(437422|0),($vararg_buffer|0)); //@line 515 "modules/graphics/opengl/OpenGL.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $11 = ___cxa_find_matching_catch_2()|0;
  $12 = tempRet0;
  ___cxa_free_exception(($8|0)); //@line 515 "modules/graphics/opengl/OpenGL.cpp"
  ___resumeException($11|0);
  // unreachable;
 } else {
  ___cxa_throw(($8|0),(96|0),(344|0)); //@line 515 "modules/graphics/opengl/OpenGL.cpp"
  // unreachable; //@line 515 "modules/graphics/opengl/OpenGL.cpp"
 }
}
function __ZN4love8graphics6opengl6OpenGL17bindTextureToUnitEjib($this,$texture,$textureunit,$restoreprev) {
 $this = $this|0;
 $texture = $texture|0;
 $textureunit = $textureunit|0;
 $restoreprev = $restoreprev|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($textureunit|0)<(0); //@line 534 "modules/graphics/opengl/OpenGL.cpp"
 if (!($0)) {
  $1 = ((($this)) + 64|0); //@line 534 "modules/graphics/opengl/OpenGL.cpp"
  $2 = ((($this)) + 68|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $6 = $5 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $7 = ($6>>>0)>($textureunit>>>0); //@line 534 "modules/graphics/opengl/OpenGL.cpp"
  $8 = $4; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($7) {
   $14 = (($8) + ($textureunit<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $15 = HEAP32[$14>>2]|0; //@line 537 "modules/graphics/opengl/OpenGL.cpp"
   $16 = ($15|0)==($texture|0); //@line 537 "modules/graphics/opengl/OpenGL.cpp"
   if ($16) {
    STACKTOP = sp;return; //@line 548 "modules/graphics/opengl/OpenGL.cpp"
   }
   $17 = ((($this)) + 76|0); //@line 539 "modules/graphics/opengl/OpenGL.cpp"
   $18 = HEAP32[$17>>2]|0; //@line 539 "modules/graphics/opengl/OpenGL.cpp"
   $19 = ($18|0)==($textureunit|0); //@line 517 "modules/graphics/opengl/OpenGL.cpp"
   if ($19) {
    $23 = $8;
   } else {
    $20 = (($textureunit) + 33984)|0; //@line 518 "modules/graphics/opengl/OpenGL.cpp"
    $21 = HEAP32[165288]|0; //@line 381 "./libraries/glad/gladfuncs.hpp"
    FUNCTION_TABLE_vi[$21 & 8191]($20); //@line 381 "./libraries/glad/gladfuncs.hpp"
    $$pre = HEAP32[$1>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $23 = $$pre;
   }
   HEAP32[$17>>2] = $textureunit; //@line 520 "modules/graphics/opengl/OpenGL.cpp"
   $22 = (($23) + ($textureunit<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$22>>2] = $texture; //@line 542 "modules/graphics/opengl/OpenGL.cpp"
   $24 = HEAP32[164685]|0; //@line 353 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_vii[$24 & 8191](3553,$texture); //@line 353 "./libraries/glad/gladfuncs.hpp"
   if (!($restoreprev)) {
    STACKTOP = sp;return; //@line 548 "modules/graphics/opengl/OpenGL.cpp"
   }
   $25 = ($18|0)<(0); //@line 514 "modules/graphics/opengl/OpenGL.cpp"
   if (!($25)) {
    $26 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $27 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $28 = (($26) - ($27))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $29 = $28 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $30 = ($29>>>0)>($18>>>0); //@line 514 "modules/graphics/opengl/OpenGL.cpp"
    if ($30) {
     $36 = HEAP32[$17>>2]|0; //@line 517 "modules/graphics/opengl/OpenGL.cpp"
     $37 = ($36|0)==($18|0); //@line 517 "modules/graphics/opengl/OpenGL.cpp"
     if (!($37)) {
      $38 = (($18) + 33984)|0; //@line 518 "modules/graphics/opengl/OpenGL.cpp"
      $39 = HEAP32[165288]|0; //@line 381 "./libraries/glad/gladfuncs.hpp"
      FUNCTION_TABLE_vi[$39 & 8191]($38); //@line 381 "./libraries/glad/gladfuncs.hpp"
     }
     HEAP32[$17>>2] = $18; //@line 520 "modules/graphics/opengl/OpenGL.cpp"
     STACKTOP = sp;return; //@line 548 "modules/graphics/opengl/OpenGL.cpp"
    }
   }
   $31 = (___cxa_allocate_exception(16)|0); //@line 515 "modules/graphics/opengl/OpenGL.cpp"
   __THREW__ = 0;
   HEAP32[$vararg_buffer1>>2] = $18; //@line 515 "modules/graphics/opengl/OpenGL.cpp"
   invoke_viii(3349,($31|0),(437422|0),($vararg_buffer1|0)); //@line 515 "modules/graphics/opengl/OpenGL.cpp"
   $32 = __THREW__; __THREW__ = 0;
   $33 = $32&1;
   if ($33) {
    $34 = ___cxa_find_matching_catch_2()|0;
    $35 = tempRet0;
    ___cxa_free_exception(($31|0)); //@line 515 "modules/graphics/opengl/OpenGL.cpp"
    ___resumeException($34|0);
    // unreachable;
   } else {
    ___cxa_throw(($31|0),(96|0),(344|0)); //@line 515 "modules/graphics/opengl/OpenGL.cpp"
    // unreachable; //@line 515 "modules/graphics/opengl/OpenGL.cpp"
   }
  }
 }
 $9 = (___cxa_allocate_exception(16)|0); //@line 535 "modules/graphics/opengl/OpenGL.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($9|0),(437455|0),($vararg_buffer|0)); //@line 535 "modules/graphics/opengl/OpenGL.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $12 = ___cxa_find_matching_catch_2()|0;
  $13 = tempRet0;
  ___cxa_free_exception(($9|0)); //@line 535 "modules/graphics/opengl/OpenGL.cpp"
  ___resumeException($12|0);
  // unreachable;
 } else {
  ___cxa_throw(($9|0),(96|0),(344|0)); //@line 535 "modules/graphics/opengl/OpenGL.cpp"
  // unreachable; //@line 535 "modules/graphics/opengl/OpenGL.cpp"
 }
}
function __ZN4love8graphics6opengl6OpenGL13deleteTextureEj($this,$texture) {
 $this = $this|0;
 $texture = $texture|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 HEAP32[$0>>2] = $texture;
 $1 = ((($this)) + 64|0); //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($this)) + 68|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$3>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($2|0)==($4|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($5) {
  $6 = HEAP32[165157]|0; //@line 354 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vii[$6 & 8191](1,$0); //@line 354 "./libraries/glad/gladfuncs.hpp"
  STACKTOP = sp;return; //@line 561 "modules/graphics/opengl/OpenGL.cpp"
 } else {
  $9 = $texture;$__begin$sroa$0$01 = $2;
 }
 while(1) {
  $7 = HEAP32[$__begin$sroa$0$01>>2]|0; //@line 556 "modules/graphics/opengl/OpenGL.cpp"
  $8 = ($7|0)==($9|0); //@line 556 "modules/graphics/opengl/OpenGL.cpp"
  if ($8) {
   HEAP32[$__begin$sroa$0$01>>2] = 0; //@line 557 "modules/graphics/opengl/OpenGL.cpp"
  }
  $10 = ((($__begin$sroa$0$01)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $11 = ($10|0)==($4|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($11) {
   break;
  }
  $$pre = HEAP32[$0>>2]|0; //@line 556 "modules/graphics/opengl/OpenGL.cpp"
  $9 = $$pre;$__begin$sroa$0$01 = $10;
 }
 $6 = HEAP32[165157]|0; //@line 354 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$6 & 8191](1,$0); //@line 354 "./libraries/glad/gladfuncs.hpp"
 STACKTOP = sp;return; //@line 561 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL16setTextureFilterERNS0_7Texture6FilterE($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $$ = 0, $$7 = 0, $$8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cond = 0, $gmin$0 = 0, $or$cond = 0, $or$cond2 = 0, $or$cond4 = 0, $or$cond6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 8|0); //@line 567 "modules/graphics/opengl/OpenGL.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 567 "modules/graphics/opengl/OpenGL.cpp"
 $2 = ($1|0)==(0); //@line 567 "modules/graphics/opengl/OpenGL.cpp"
 $3 = HEAP32[$f>>2]|0; //@line 569 "modules/graphics/opengl/OpenGL.cpp"
 $4 = ($3|0)==(2); //@line 569 "modules/graphics/opengl/OpenGL.cpp"
 if ($2) {
  $$ = $4 ? 9728 : 9729; //@line 570 "modules/graphics/opengl/OpenGL.cpp"
  $gmin$0 = $$;
 } else {
  $5 = ($1|0)==(2); //@line 576 "modules/graphics/opengl/OpenGL.cpp"
  $or$cond = $5 & $4; //@line 576 "modules/graphics/opengl/OpenGL.cpp"
  if ($or$cond) {
   $gmin$0 = 9984;
  } else {
   $6 = ($1|0)==(1); //@line 578 "modules/graphics/opengl/OpenGL.cpp"
   $or$cond2 = $6 & $4; //@line 578 "modules/graphics/opengl/OpenGL.cpp"
   if ($or$cond2) {
    $gmin$0 = 9986;
   } else {
    $7 = ($3|0)==(1); //@line 580 "modules/graphics/opengl/OpenGL.cpp"
    $or$cond4 = $5 & $7; //@line 580 "modules/graphics/opengl/OpenGL.cpp"
    if ($or$cond4) {
     $gmin$0 = 9985;
    } else {
     $or$cond6 = $6 & $7; //@line 582 "modules/graphics/opengl/OpenGL.cpp"
     $$8 = $or$cond6 ? 9987 : 9729; //@line 582 "modules/graphics/opengl/OpenGL.cpp"
     $gmin$0 = $$8;
    }
   }
  }
 }
 $8 = ((($f)) + 4|0); //@line 588 "modules/graphics/opengl/OpenGL.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 588 "modules/graphics/opengl/OpenGL.cpp"
 $cond = ($9|0)==(2); //@line 588 "modules/graphics/opengl/OpenGL.cpp"
 $$7 = $cond ? 9728 : 9729; //@line 592 "modules/graphics/opengl/OpenGL.cpp"
 $10 = HEAP32[165108]|0; //@line 44 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$10 & 8191](3553,10241,$gmin$0); //@line 44 "./libraries/glad/gladfuncs.hpp"
 $11 = HEAP32[165108]|0; //@line 44 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$11 & 8191](3553,10240,$$7); //@line 44 "./libraries/glad/gladfuncs.hpp"
 $12 = HEAP8[1252239]|0; //@line 602 "modules/graphics/opengl/OpenGL.cpp"
 $13 = ($12<<24>>24)==(0); //@line 602 "modules/graphics/opengl/OpenGL.cpp"
 $14 = ((($f)) + 12|0); //@line 604 "modules/graphics/opengl/OpenGL.cpp"
 if ($13) {
  HEAPF32[$14>>2] = 1.0; //@line 608 "modules/graphics/opengl/OpenGL.cpp"
  return; //@line 609 "modules/graphics/opengl/OpenGL.cpp"
 } else {
  $15 = +HEAPF32[$14>>2]; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $16 = $15 < 1.0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = ((($this)) + 40|0); //@line 604 "modules/graphics/opengl/OpenGL.cpp"
  $18 = +HEAPF32[$17>>2]; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $19 = $16 ? 1.0 : $15; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $20 = $18 < $19; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $21 = $20 ? $18 : $19; //@line 604 "modules/graphics/opengl/OpenGL.cpp"
  HEAPF32[$14>>2] = $21; //@line 604 "modules/graphics/opengl/OpenGL.cpp"
  $22 = HEAP32[165106]|0; //@line 42 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viid[$22 & 8191](3553,34046,$21); //@line 42 "./libraries/glad/gladfuncs.hpp"
  return; //@line 609 "modules/graphics/opengl/OpenGL.cpp"
 }
}
function __ZN4love8graphics6opengl6OpenGL14setTextureWrapERKNS0_7Texture4WrapE($this,$w) {
 $this = $this|0;
 $w = $w|0;
 var $$0$i = 0, $$0$i1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$w>>2]|0; //@line 630 "modules/graphics/opengl/OpenGL.cpp"
 switch ($0|0) {
 case 3:  {
  $$0$i = 33648;
  break;
 }
 case 1:  {
  $$0$i = 33069;
  break;
 }
 case 2:  {
  $$0$i = 10497;
  break;
 }
 default: {
  $$0$i = 33071;
 }
 }
 $1 = HEAP32[165108]|0; //@line 44 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$1 & 8191](3553,10242,$$0$i); //@line 44 "./libraries/glad/gladfuncs.hpp"
 $2 = ((($w)) + 4|0); //@line 631 "modules/graphics/opengl/OpenGL.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 631 "modules/graphics/opengl/OpenGL.cpp"
 switch ($3|0) {
 case 3:  {
  $$0$i1 = 33648;
  break;
 }
 case 1:  {
  $$0$i1 = 33069;
  break;
 }
 case 2:  {
  $$0$i1 = 10497;
  break;
 }
 default: {
  $$0$i1 = 33071;
 }
 }
 $4 = HEAP32[165108]|0; //@line 44 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$4 & 8191](3553,10243,$$0$i1); //@line 44 "./libraries/glad/gladfuncs.hpp"
 return; //@line 632 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZNK4love8graphics6opengl6OpenGL31isClampZeroTextureWrapSupportedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[1251997]|0; //@line 636 "modules/graphics/opengl/OpenGL.cpp"
 $1 = HEAP8[1252230]|0; //@line 636 "modules/graphics/opengl/OpenGL.cpp"
 $2 = $1 | $0; //@line 636 "modules/graphics/opengl/OpenGL.cpp"
 $3 = HEAP8[1252479]|0; //@line 636 "modules/graphics/opengl/OpenGL.cpp"
 $4 = $2 | $3; //@line 636 "modules/graphics/opengl/OpenGL.cpp"
 $5 = ($4<<24>>24)!=(0); //@line 636 "modules/graphics/opengl/OpenGL.cpp"
 return ($5|0); //@line 636 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZNK4love8graphics6opengl6OpenGL17getMaxTextureSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 44|0); //@line 641 "modules/graphics/opengl/OpenGL.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 641 "modules/graphics/opengl/OpenGL.cpp"
 return ($1|0); //@line 641 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZNK4love8graphics6opengl6OpenGL19getMaxRenderTargetsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 646 "modules/graphics/opengl/OpenGL.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 646 "modules/graphics/opengl/OpenGL.cpp"
 return ($1|0); //@line 646 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZNK4love8graphics6opengl6OpenGL25getMaxRenderbufferSamplesEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 651 "modules/graphics/opengl/OpenGL.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 651 "modules/graphics/opengl/OpenGL.cpp"
 return ($1|0); //@line 651 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZNK4love8graphics6opengl6OpenGL18getMaxTextureUnitsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 56|0); //@line 656 "modules/graphics/opengl/OpenGL.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 656 "modules/graphics/opengl/OpenGL.cpp"
 return ($1|0); //@line 656 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL23updateTextureMemorySizeEjj($this,$oldsize,$newsize) {
 $this = $this|0;
 $oldsize = $oldsize|0;
 $newsize = $newsize|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 661 "modules/graphics/opengl/OpenGL.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 661 "modules/graphics/opengl/OpenGL.cpp"
 $2 = (_i64Subtract(($newsize|0),0,($oldsize|0),0)|0); //@line 661 "modules/graphics/opengl/OpenGL.cpp"
 $3 = tempRet0; //@line 661 "modules/graphics/opengl/OpenGL.cpp"
 $4 = (_i64Add(($1|0),0,($2|0),($3|0))|0); //@line 661 "modules/graphics/opengl/OpenGL.cpp"
 $5 = tempRet0; //@line 661 "modules/graphics/opengl/OpenGL.cpp"
 $6 = ($5|0)<(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $7 = $6 ? 0 : $4; //@line 662 "modules/graphics/opengl/OpenGL.cpp"
 $6 ? 0 : $5; //@line 662 "modules/graphics/opengl/OpenGL.cpp"
 HEAP32[$0>>2] = $7; //@line 662 "modules/graphics/opengl/OpenGL.cpp"
 return; //@line 663 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL11errorStringEj($errorcode) {
 $errorcode = $errorcode|0;
 var $$0 = 0, $vararg_buffer = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 670 "modules/graphics/opengl/OpenGL.cpp"
 switch ($errorcode|0) {
 case 0:  {
  $$0 = 437592;
  break;
 }
 case 1280:  {
  $$0 = 437579;
  break;
 }
 case 1281:  {
  $$0 = 522528;
  break;
 }
 case 1282:  {
  $$0 = 437561;
  break;
 }
 case 1285:  {
  $$0 = 437547;
  break;
 }
 case 1286:  {
  $$0 = 437517;
  break;
 }
 case 1287:  {
  $$0 = 437488;
  break;
 }
 default: {
  dest=1251930; stop=dest+64|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 694 "modules/graphics/opengl/OpenGL.cpp"
  HEAP32[$vararg_buffer>>2] = $errorcode; //@line 695 "modules/graphics/opengl/OpenGL.cpp"
  (_sprintf(1251930,437483,$vararg_buffer)|0); //@line 695 "modules/graphics/opengl/OpenGL.cpp"
  $$0 = 1251930;
 }
 }
 STACKTOP = sp;return ($$0|0); //@line 698 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL19debugSeverityStringEj($severity) {
 $severity = $severity|0;
 var $$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 switch ($severity|0) {
 case 37190:  {
  $$0 = 437612;
  break;
 }
 case 37191:  {
  $$0 = 437605;
  break;
 }
 case 37192:  {
  $$0 = 437601;
  break;
 }
 default: {
  $$0 = 442424;
 }
 }
 return ($$0|0); //@line 713 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL17debugSourceStringEj($source) {
 $source = $source|0;
 var $$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 switch ($source|0) {
 case 33350:  {
  $$0 = 437651;
  break;
 }
 case 33351:  {
  $$0 = 437644;
  break;
 }
 case 33352:  {
  $$0 = 437637;
  break;
 }
 case 33353:  {
  $$0 = 437628;
  break;
 }
 case 33354:  {
  $$0 = 437623;
  break;
 }
 case 33355:  {
  $$0 = 437617;
  break;
 }
 default: {
  $$0 = 442424;
 }
 }
 return ($$0|0); //@line 734 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl6OpenGL15debugTypeStringEj($type) {
 $type = $type|0;
 var $$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 switch ($type|0) {
 case 33356:  {
  $$0 = 520240;
  break;
 }
 case 33357:  {
  $$0 = 437698;
  break;
 }
 case 33358:  {
  $$0 = 437679;
  break;
 }
 case 33360:  {
  $$0 = 437667;
  break;
 }
 case 33359:  {
  $$0 = 437655;
  break;
 }
 case 33361:  {
  $$0 = 437617;
  break;
 }
 default: {
  $$0 = 442424;
 }
 }
 return ($$0|0); //@line 755 "modules/graphics/opengl/OpenGL.cpp"
}
function __ZN4love8graphics6opengl14ParticleSystemD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (25048); //@line 56 "modules/graphics/opengl/ParticleSystem.cpp"
 $0 = ((($this)) + 228|0); //@line 57 "modules/graphics/opengl/ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 57 "modules/graphics/opengl/ParticleSystem.cpp"
 $2 = ($1|0)==(0|0); //@line 57 "modules/graphics/opengl/ParticleSystem.cpp"
 if (!($2)) {
  __ZdaPv($1); //@line 57 "modules/graphics/opengl/ParticleSystem.cpp"
 }
 $3 = ((($this)) + 232|0); //@line 58 "modules/graphics/opengl/ParticleSystem.cpp"
 __ZN4love8graphics6opengl11QuadIndicesD2Ev($3); //@line 58 "modules/graphics/opengl/ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystemD2Ev($this); //@line 58 "modules/graphics/opengl/ParticleSystem.cpp"
 return; //@line 58 "modules/graphics/opengl/ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl14ParticleSystemD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (25048); //@line 56 "modules/graphics/opengl/ParticleSystem.cpp"
 $0 = ((($this)) + 228|0); //@line 57 "modules/graphics/opengl/ParticleSystem.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 57 "modules/graphics/opengl/ParticleSystem.cpp"
 $2 = ($1|0)==(0|0); //@line 57 "modules/graphics/opengl/ParticleSystem.cpp"
 if (!($2)) {
  __ZdaPv($1); //@line 57 "modules/graphics/opengl/ParticleSystem.cpp"
 }
 $3 = ((($this)) + 232|0); //@line 58 "modules/graphics/opengl/ParticleSystem.cpp"
 __ZN4love8graphics6opengl11QuadIndicesD2Ev($3); //@line 58 "modules/graphics/opengl/ParticleSystem.cpp"
 __ZN4love8graphics14ParticleSystemD2Ev($this); //@line 58 "modules/graphics/opengl/ParticleSystem.cpp"
 __ZdlPv($this); //@line 56 "modules/graphics/opengl/ParticleSystem.cpp"
 return; //@line 58 "modules/graphics/opengl/ParticleSystem.cpp"
}
function __ZN4love8graphics6opengl14ParticleSystem4drawEfffffffff($this,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0, $105 = 0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0;
 var $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0, $123 = 0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0.0;
 var $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0.0, $144 = 0.0, $145 = 0, $146 = 0, $147 = 0.0, $148 = 0.0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0.0, $152 = 0.0, $153 = 0, $154 = 0, $155 = 0.0, $156 = 0.0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0.0, $214 = 0, $215 = 0.0, $216 = 0, $217 = 0.0, $218 = 0, $219 = 0.0, $22 = 0, $220 = 0.0, $221 = 0.0;
 var $222 = 0, $223 = 0, $224 = 0.0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0, $229 = 0.0, $23 = 0, $230 = 0.0, $231 = 0.0, $232 = 0.0, $233 = 0.0, $234 = 0.0, $235 = 0.0, $236 = 0.0, $237 = 0.0, $238 = 0.0, $239 = 0.0, $24 = 0;
 var $240 = 0.0, $241 = 0, $242 = 0.0, $243 = 0, $244 = 0.0, $245 = 0.0, $246 = 0.0, $247 = 0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0.0, $254 = 0.0, $255 = 0.0, $256 = 0.0, $257 = 0.0, $258 = 0.0;
 var $259 = 0.0, $26 = 0, $260 = 0, $261 = 0, $262 = 0.0, $263 = 0, $264 = 0.0, $265 = 0.0, $266 = 0.0, $267 = 0, $268 = 0.0, $269 = 0.0, $27 = 0, $270 = 0.0, $271 = 0.0, $272 = 0.0, $273 = 0.0, $274 = 0.0, $275 = 0.0, $276 = 0.0;
 var $277 = 0.0, $278 = 0.0, $279 = 0.0, $28 = 0, $280 = 0, $281 = 0, $282 = 0.0, $283 = 0, $284 = 0.0, $285 = 0.0, $286 = 0.0, $287 = 0, $288 = 0.0, $289 = 0.0, $29 = 0, $290 = 0.0, $291 = 0.0, $292 = 0.0, $293 = 0.0, $294 = 0.0;
 var $295 = 0.0, $296 = 0.0, $297 = 0.0, $298 = 0.0, $299 = 0.0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0.0, $304 = 0.0, $305 = 0, $306 = 0, $307 = 0.0, $308 = 0.0, $309 = 0, $31 = 0, $310 = 0, $311 = 0.0;
 var $312 = 0.0, $313 = 0, $314 = 0, $315 = 0.0, $316 = 0.0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0;
 var $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0;
 var $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0;
 var $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0;
 var $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0;
 var $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0, $62 = 0.0, $63 = 0, $64 = 0.0, $65 = 0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0;
 var $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $lpad$phi$index = 0, $lpad$phi$index3 = 0, $p$08 = 0, $p$08$us = 0, $pVerts$07 = 0, $pVerts$07$us = 0, $t = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 40|0;
 $t = sp;
 $1 = (__ZNK4love8graphics14ParticleSystem8getCountEv($this)|0); //@line 89 "modules/graphics/opengl/ParticleSystem.cpp"
 $2 = ($1|0)==(0); //@line 91 "modules/graphics/opengl/ParticleSystem.cpp"
 if ($2) {
  STACKTOP = sp;return; //@line 152 "modules/graphics/opengl/ParticleSystem.cpp"
 }
 $3 = ((($this)) + 24|0); //@line 139 "./common/Object.h"
 $4 = HEAP32[$3>>2]|0; //@line 139 "./common/Object.h"
 $5 = ($4|0)==(0|0); //@line 91 "modules/graphics/opengl/ParticleSystem.cpp"
 if ($5) {
  STACKTOP = sp;return; //@line 152 "modules/graphics/opengl/ParticleSystem.cpp"
 }
 $6 = ((($this)) + 8|0); //@line 91 "modules/graphics/opengl/ParticleSystem.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 91 "modules/graphics/opengl/ParticleSystem.cpp"
 $8 = ($7|0)==(0|0); //@line 91 "modules/graphics/opengl/ParticleSystem.cpp"
 if ($8) {
  STACKTOP = sp;return; //@line 152 "modules/graphics/opengl/ParticleSystem.cpp"
 }
 $9 = ((($this)) + 228|0); //@line 91 "modules/graphics/opengl/ParticleSystem.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 91 "modules/graphics/opengl/ParticleSystem.cpp"
 $11 = ($10|0)==(0|0); //@line 91 "modules/graphics/opengl/ParticleSystem.cpp"
 if ($11) {
  STACKTOP = sp;return; //@line 152 "modules/graphics/opengl/ParticleSystem.cpp"
 }
 __ZN4love8graphics6opengl6OpenGL13pushTransformEv(634044); //@line 125 "modules/graphics/opengl/OpenGL.h"
 __THREW__ = 0;
 $12 = (invoke_ii(4003,(634044|0))|0); //@line 135 "modules/graphics/opengl/OpenGL.h"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 do {
  if ($14) {
   label = 17;
  } else {
   __THREW__ = 0;
   invoke_viddddddddd(4039,($0|0),(+$x),(+$y),(+$angle),(+$sx),(+$sy),(+$ox),(+$oy),(+$kx),(+$ky)); //@line 97 "modules/graphics/opengl/ParticleSystem.cpp"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if ($16) {
    label = 17;
   } else {
    __THREW__ = 0;
    invoke_vii(4004,($12|0),($0|0)); //@line 97 "modules/graphics/opengl/ParticleSystem.cpp"
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if ($18) {
     $206 = ___cxa_find_matching_catch_2()|0;
     $207 = tempRet0;
     __ZN4love7Matrix4D2Ev($0); //@line 97 "modules/graphics/opengl/ParticleSystem.cpp"
     $$0 = $206;$$01 = $207;
     break;
    }
    __ZN4love7Matrix4D2Ev($0); //@line 97 "modules/graphics/opengl/ParticleSystem.cpp"
    $19 = HEAP32[$3>>2]|0; //@line 117 "./common/Object.h"
    $20 = HEAP32[$19>>2]|0; //@line 99 "modules/graphics/opengl/ParticleSystem.cpp"
    $21 = ((($20)) + 48|0); //@line 99 "modules/graphics/opengl/ParticleSystem.cpp"
    $22 = HEAP32[$21>>2]|0; //@line 99 "modules/graphics/opengl/ParticleSystem.cpp"
    __THREW__ = 0;
    $23 = (invoke_ii($22|0,($19|0))|0); //@line 99 "modules/graphics/opengl/ParticleSystem.cpp"
    $24 = __THREW__; __THREW__ = 0;
    $25 = $24&1;
    if ($25) {
     label = 17;
    } else {
     $26 = HEAP32[$9>>2]|0; //@line 100 "modules/graphics/opengl/ParticleSystem.cpp"
     $27 = ((($this)) + 16|0); //@line 101 "modules/graphics/opengl/ParticleSystem.cpp"
     $28 = HEAP32[$27>>2]|0; //@line 101 "modules/graphics/opengl/ParticleSystem.cpp"
     $29 = ((($this)) + 212|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $30 = HEAP32[$29>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $31 = ((($this)) + 216|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $32 = HEAP32[$31>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $33 = ($30|0)==($32|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     __THREW__ = 0;
     invoke_vi(4102,($t|0)); //@line 105 "modules/graphics/opengl/ParticleSystem.cpp"
     $34 = __THREW__; __THREW__ = 0;
     $35 = $34&1;
     if ($35) {
      label = 17;
     } else {
      $36 = ($28|0)==(0|0); //@line 108 "modules/graphics/opengl/ParticleSystem.cpp"
      L21: do {
       if ($36) {
        label = 24;
       } else {
        $37 = ((($this)) + 188|0); //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
        $38 = ((($this)) + 192|0); //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
        $39 = ((($t)) + 12|0); //@line 280 "./common/Matrix.h"
        $40 = ((($t)) + 24|0); //@line 280 "./common/Matrix.h"
        $41 = ((($t)) + 4|0); //@line 281 "./common/Matrix.h"
        $42 = ((($t)) + 16|0); //@line 281 "./common/Matrix.h"
        $43 = ((($t)) + 28|0); //@line 281 "./common/Matrix.h"
        if ($33) {
         $44 = ((($23)) + 4|0); //@line 280 "./common/Matrix.h"
         $45 = ((($23)) + 20|0); //@line 280 "./common/Matrix.h"
         $46 = ((($23)) + 24|0); //@line 280 "./common/Matrix.h"
         $47 = ((($23)) + 40|0); //@line 280 "./common/Matrix.h"
         $48 = ((($23)) + 44|0); //@line 280 "./common/Matrix.h"
         $49 = ((($23)) + 60|0); //@line 280 "./common/Matrix.h"
         $50 = ((($23)) + 64|0); //@line 280 "./common/Matrix.h"
         $51 = ((($23)) + 8|0); //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
         $52 = ((($23)) + 12|0); //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
         $53 = ((($23)) + 28|0); //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
         $54 = ((($23)) + 32|0); //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
         $55 = ((($23)) + 48|0); //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
         $56 = ((($23)) + 52|0); //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
         $57 = ((($23)) + 68|0); //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
         $58 = ((($23)) + 72|0); //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
         $p$08$us = $28;$pVerts$07$us = $26;
         while(1) {
          $59 = ((($p$08$us)) + 16|0); //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $60 = +HEAPF32[$59>>2]; //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $61 = ((($p$08$us)) + 20|0); //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $62 = +HEAPF32[$61>>2]; //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $63 = ((($p$08$us)) + 76|0); //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $64 = +HEAPF32[$63>>2]; //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $65 = ((($p$08$us)) + 60|0); //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $66 = +HEAPF32[$65>>2]; //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $67 = +HEAPF32[$37>>2]; //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $68 = +HEAPF32[$38>>2]; //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          __THREW__ = 0;
          invoke_viddddddddd(4103,($t|0),(+$60),(+$62),(+$64),(+$66),(+$66),(+$67),(+$68),0.0,0.0); //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $69 = __THREW__; __THREW__ = 0;
          $70 = $69&1;
          if ($70) {
           break;
          }
          $71 = +HEAPF32[$t>>2]; //@line 280 "./common/Matrix.h"
          $72 = +HEAPF32[$23>>2]; //@line 280 "./common/Matrix.h"
          $73 = $71 * $72; //@line 280 "./common/Matrix.h"
          $74 = +HEAPF32[$39>>2]; //@line 280 "./common/Matrix.h"
          $75 = +HEAPF32[$44>>2]; //@line 280 "./common/Matrix.h"
          $76 = $74 * $75; //@line 280 "./common/Matrix.h"
          $77 = $73 + $76; //@line 280 "./common/Matrix.h"
          $78 = +HEAPF32[$40>>2]; //@line 280 "./common/Matrix.h"
          $79 = $78 + $77; //@line 280 "./common/Matrix.h"
          $80 = +HEAPF32[$41>>2]; //@line 281 "./common/Matrix.h"
          $81 = $72 * $80; //@line 281 "./common/Matrix.h"
          $82 = +HEAPF32[$42>>2]; //@line 281 "./common/Matrix.h"
          $83 = $75 * $82; //@line 281 "./common/Matrix.h"
          $84 = $81 + $83; //@line 281 "./common/Matrix.h"
          $85 = +HEAPF32[$43>>2]; //@line 281 "./common/Matrix.h"
          $86 = $85 + $84; //@line 281 "./common/Matrix.h"
          HEAPF32[$pVerts$07$us>>2] = $79; //@line 283 "./common/Matrix.h"
          $87 = ((($pVerts$07$us)) + 4|0); //@line 284 "./common/Matrix.h"
          HEAPF32[$87>>2] = $86; //@line 284 "./common/Matrix.h"
          $88 = +HEAPF32[$t>>2]; //@line 280 "./common/Matrix.h"
          $89 = +HEAPF32[$45>>2]; //@line 280 "./common/Matrix.h"
          $90 = $88 * $89; //@line 280 "./common/Matrix.h"
          $91 = +HEAPF32[$39>>2]; //@line 280 "./common/Matrix.h"
          $92 = +HEAPF32[$46>>2]; //@line 280 "./common/Matrix.h"
          $93 = $91 * $92; //@line 280 "./common/Matrix.h"
          $94 = $90 + $93; //@line 280 "./common/Matrix.h"
          $95 = +HEAPF32[$40>>2]; //@line 280 "./common/Matrix.h"
          $96 = $95 + $94; //@line 280 "./common/Matrix.h"
          $97 = +HEAPF32[$41>>2]; //@line 281 "./common/Matrix.h"
          $98 = $89 * $97; //@line 281 "./common/Matrix.h"
          $99 = +HEAPF32[$42>>2]; //@line 281 "./common/Matrix.h"
          $100 = $92 * $99; //@line 281 "./common/Matrix.h"
          $101 = $98 + $100; //@line 281 "./common/Matrix.h"
          $102 = +HEAPF32[$43>>2]; //@line 281 "./common/Matrix.h"
          $103 = $102 + $101; //@line 281 "./common/Matrix.h"
          $104 = ((($pVerts$07$us)) + 20|0); //@line 283 "./common/Matrix.h"
          HEAPF32[$104>>2] = $96; //@line 283 "./common/Matrix.h"
          $105 = ((($pVerts$07$us)) + 24|0); //@line 284 "./common/Matrix.h"
          HEAPF32[$105>>2] = $103; //@line 284 "./common/Matrix.h"
          $106 = +HEAPF32[$t>>2]; //@line 280 "./common/Matrix.h"
          $107 = +HEAPF32[$47>>2]; //@line 280 "./common/Matrix.h"
          $108 = $106 * $107; //@line 280 "./common/Matrix.h"
          $109 = +HEAPF32[$39>>2]; //@line 280 "./common/Matrix.h"
          $110 = +HEAPF32[$48>>2]; //@line 280 "./common/Matrix.h"
          $111 = $109 * $110; //@line 280 "./common/Matrix.h"
          $112 = $108 + $111; //@line 280 "./common/Matrix.h"
          $113 = +HEAPF32[$40>>2]; //@line 280 "./common/Matrix.h"
          $114 = $113 + $112; //@line 280 "./common/Matrix.h"
          $115 = +HEAPF32[$41>>2]; //@line 281 "./common/Matrix.h"
          $116 = $107 * $115; //@line 281 "./common/Matrix.h"
          $117 = +HEAPF32[$42>>2]; //@line 281 "./common/Matrix.h"
          $118 = $110 * $117; //@line 281 "./common/Matrix.h"
          $119 = $116 + $118; //@line 281 "./common/Matrix.h"
          $120 = +HEAPF32[$43>>2]; //@line 281 "./common/Matrix.h"
          $121 = $120 + $119; //@line 281 "./common/Matrix.h"
          $122 = ((($pVerts$07$us)) + 40|0); //@line 283 "./common/Matrix.h"
          HEAPF32[$122>>2] = $114; //@line 283 "./common/Matrix.h"
          $123 = ((($pVerts$07$us)) + 44|0); //@line 284 "./common/Matrix.h"
          HEAPF32[$123>>2] = $121; //@line 284 "./common/Matrix.h"
          $124 = +HEAPF32[$t>>2]; //@line 280 "./common/Matrix.h"
          $125 = +HEAPF32[$49>>2]; //@line 280 "./common/Matrix.h"
          $126 = $124 * $125; //@line 280 "./common/Matrix.h"
          $127 = +HEAPF32[$39>>2]; //@line 280 "./common/Matrix.h"
          $128 = +HEAPF32[$50>>2]; //@line 280 "./common/Matrix.h"
          $129 = $127 * $128; //@line 280 "./common/Matrix.h"
          $130 = $126 + $129; //@line 280 "./common/Matrix.h"
          $131 = +HEAPF32[$40>>2]; //@line 280 "./common/Matrix.h"
          $132 = $131 + $130; //@line 280 "./common/Matrix.h"
          $133 = +HEAPF32[$41>>2]; //@line 281 "./common/Matrix.h"
          $134 = $125 * $133; //@line 281 "./common/Matrix.h"
          $135 = +HEAPF32[$42>>2]; //@line 281 "./common/Matrix.h"
          $136 = $128 * $135; //@line 281 "./common/Matrix.h"
          $137 = $134 + $136; //@line 281 "./common/Matrix.h"
          $138 = +HEAPF32[$43>>2]; //@line 281 "./common/Matrix.h"
          $139 = $138 + $137; //@line 281 "./common/Matrix.h"
          $140 = ((($pVerts$07$us)) + 60|0); //@line 283 "./common/Matrix.h"
          HEAPF32[$140>>2] = $132; //@line 283 "./common/Matrix.h"
          $141 = ((($pVerts$07$us)) + 64|0); //@line 284 "./common/Matrix.h"
          HEAPF32[$141>>2] = $139; //@line 284 "./common/Matrix.h"
          $142 = ((($p$08$us)) + 88|0); //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          $143 = +HEAPF32[$142>>2]; //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          $144 = $143 * 255.0; //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          $145 = (~~(($144))&255); //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          $146 = ((($p$08$us)) + 92|0); //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          $147 = +HEAPF32[$146>>2]; //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          $148 = $147 * 255.0; //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          $149 = (~~(($148))&255); //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          $150 = ((($p$08$us)) + 96|0); //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          $151 = +HEAPF32[$150>>2]; //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          $152 = $151 * 255.0; //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          $153 = (~~(($152))&255); //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          $154 = ((($p$08$us)) + 100|0); //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          $155 = +HEAPF32[$154>>2]; //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          $156 = $155 * 255.0; //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          $157 = (~~(($156))&255); //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          $158 = HEAP32[$51>>2]|0; //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $159 = ((($pVerts$07$us)) + 8|0); //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP32[$159>>2] = $158; //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $160 = HEAP32[$52>>2]|0; //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $161 = ((($pVerts$07$us)) + 12|0); //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP32[$161>>2] = $160; //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $162 = ((($pVerts$07$us)) + 16|0); //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$162>>0] = $145; //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          $163 = ((($pVerts$07$us)) + 17|0); //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$163>>0] = $149; //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          $164 = ((($pVerts$07$us)) + 18|0); //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$164>>0] = $153; //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          $165 = ((($pVerts$07$us)) + 19|0); //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$165>>0] = $157; //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          $166 = HEAP32[$53>>2]|0; //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $167 = ((($pVerts$07$us)) + 28|0); //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP32[$167>>2] = $166; //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $168 = HEAP32[$54>>2]|0; //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $169 = ((($pVerts$07$us)) + 32|0); //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP32[$169>>2] = $168; //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $170 = ((($pVerts$07$us)) + 36|0); //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$170>>0] = $145; //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          $171 = ((($pVerts$07$us)) + 37|0); //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$171>>0] = $149; //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          $172 = ((($pVerts$07$us)) + 38|0); //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$172>>0] = $153; //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          $173 = ((($pVerts$07$us)) + 39|0); //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$173>>0] = $157; //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          $174 = HEAP32[$55>>2]|0; //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $175 = ((($pVerts$07$us)) + 48|0); //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP32[$175>>2] = $174; //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $176 = HEAP32[$56>>2]|0; //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $177 = ((($pVerts$07$us)) + 52|0); //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP32[$177>>2] = $176; //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $178 = ((($pVerts$07$us)) + 56|0); //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$178>>0] = $145; //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          $179 = ((($pVerts$07$us)) + 57|0); //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$179>>0] = $149; //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          $180 = ((($pVerts$07$us)) + 58|0); //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$180>>0] = $153; //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          $181 = ((($pVerts$07$us)) + 59|0); //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$181>>0] = $157; //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          $182 = HEAP32[$57>>2]|0; //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $183 = ((($pVerts$07$us)) + 68|0); //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP32[$183>>2] = $182; //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $184 = HEAP32[$58>>2]|0; //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $185 = ((($pVerts$07$us)) + 72|0); //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP32[$185>>2] = $184; //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $186 = ((($pVerts$07$us)) + 76|0); //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$186>>0] = $145; //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          $187 = ((($pVerts$07$us)) + 77|0); //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$187>>0] = $149; //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          $188 = ((($pVerts$07$us)) + 78|0); //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$188>>0] = $153; //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          $189 = ((($pVerts$07$us)) + 79|0); //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$189>>0] = $157; //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          $190 = ((($pVerts$07$us)) + 80|0); //@line 131 "modules/graphics/opengl/ParticleSystem.cpp"
          $191 = ((($p$08$us)) + 4|0); //@line 132 "modules/graphics/opengl/ParticleSystem.cpp"
          $192 = HEAP32[$191>>2]|0; //@line 132 "modules/graphics/opengl/ParticleSystem.cpp"
          $193 = ($192|0)==(0|0); //@line 108 "modules/graphics/opengl/ParticleSystem.cpp"
          if ($193) {
           label = 24;
           break L21;
          } else {
           $p$08$us = $192;$pVerts$07$us = $190;
          }
         }
         $194 = ___cxa_find_matching_catch_2()|0;
         $195 = tempRet0;
         $lpad$phi$index = $194;$lpad$phi$index3 = $195;
         break;
        } else {
         $p$08 = $28;$pVerts$07 = $26;
         while(1) {
          $196 = ((($p$08)) + 104|0); //@line 111 "modules/graphics/opengl/ParticleSystem.cpp"
          $197 = HEAP32[$196>>2]|0; //@line 111 "modules/graphics/opengl/ParticleSystem.cpp"
          $198 = HEAP32[$29>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $199 = (($198) + ($197<<2)|0); //@line 117 "./common/Object.h"
          $200 = HEAP32[$199>>2]|0; //@line 117 "./common/Object.h"
          __THREW__ = 0;
          $201 = (invoke_ii(4010,($200|0))|0); //@line 111 "modules/graphics/opengl/ParticleSystem.cpp"
          $202 = __THREW__; __THREW__ = 0;
          $203 = $202&1;
          if ($203) {
           break;
          }
          $212 = ((($p$08)) + 16|0); //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $213 = +HEAPF32[$212>>2]; //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $214 = ((($p$08)) + 20|0); //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $215 = +HEAPF32[$214>>2]; //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $216 = ((($p$08)) + 76|0); //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $217 = +HEAPF32[$216>>2]; //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $218 = ((($p$08)) + 60|0); //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $219 = +HEAPF32[$218>>2]; //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $220 = +HEAPF32[$37>>2]; //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $221 = +HEAPF32[$38>>2]; //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          __THREW__ = 0;
          invoke_viddddddddd(4103,($t|0),(+$213),(+$215),(+$217),(+$219),(+$219),(+$220),(+$221),0.0,0.0); //@line 114 "modules/graphics/opengl/ParticleSystem.cpp"
          $222 = __THREW__; __THREW__ = 0;
          $223 = $222&1;
          if ($223) {
           break;
          }
          $224 = +HEAPF32[$t>>2]; //@line 280 "./common/Matrix.h"
          $225 = +HEAPF32[$201>>2]; //@line 280 "./common/Matrix.h"
          $226 = $224 * $225; //@line 280 "./common/Matrix.h"
          $227 = +HEAPF32[$39>>2]; //@line 280 "./common/Matrix.h"
          $228 = ((($201)) + 4|0); //@line 280 "./common/Matrix.h"
          $229 = +HEAPF32[$228>>2]; //@line 280 "./common/Matrix.h"
          $230 = $227 * $229; //@line 280 "./common/Matrix.h"
          $231 = $226 + $230; //@line 280 "./common/Matrix.h"
          $232 = +HEAPF32[$40>>2]; //@line 280 "./common/Matrix.h"
          $233 = $232 + $231; //@line 280 "./common/Matrix.h"
          $234 = +HEAPF32[$41>>2]; //@line 281 "./common/Matrix.h"
          $235 = $225 * $234; //@line 281 "./common/Matrix.h"
          $236 = +HEAPF32[$42>>2]; //@line 281 "./common/Matrix.h"
          $237 = $229 * $236; //@line 281 "./common/Matrix.h"
          $238 = $235 + $237; //@line 281 "./common/Matrix.h"
          $239 = +HEAPF32[$43>>2]; //@line 281 "./common/Matrix.h"
          $240 = $239 + $238; //@line 281 "./common/Matrix.h"
          HEAPF32[$pVerts$07>>2] = $233; //@line 283 "./common/Matrix.h"
          $241 = ((($pVerts$07)) + 4|0); //@line 284 "./common/Matrix.h"
          HEAPF32[$241>>2] = $240; //@line 284 "./common/Matrix.h"
          $242 = +HEAPF32[$t>>2]; //@line 280 "./common/Matrix.h"
          $243 = ((($201)) + 20|0); //@line 280 "./common/Matrix.h"
          $244 = +HEAPF32[$243>>2]; //@line 280 "./common/Matrix.h"
          $245 = $242 * $244; //@line 280 "./common/Matrix.h"
          $246 = +HEAPF32[$39>>2]; //@line 280 "./common/Matrix.h"
          $247 = ((($201)) + 24|0); //@line 280 "./common/Matrix.h"
          $248 = +HEAPF32[$247>>2]; //@line 280 "./common/Matrix.h"
          $249 = $246 * $248; //@line 280 "./common/Matrix.h"
          $250 = $245 + $249; //@line 280 "./common/Matrix.h"
          $251 = +HEAPF32[$40>>2]; //@line 280 "./common/Matrix.h"
          $252 = $251 + $250; //@line 280 "./common/Matrix.h"
          $253 = +HEAPF32[$41>>2]; //@line 281 "./common/Matrix.h"
          $254 = $244 * $253; //@line 281 "./common/Matrix.h"
          $255 = +HEAPF32[$42>>2]; //@line 281 "./common/Matrix.h"
          $256 = $248 * $255; //@line 281 "./common/Matrix.h"
          $257 = $254 + $256; //@line 281 "./common/Matrix.h"
          $258 = +HEAPF32[$43>>2]; //@line 281 "./common/Matrix.h"
          $259 = $258 + $257; //@line 281 "./common/Matrix.h"
          $260 = ((($pVerts$07)) + 20|0); //@line 283 "./common/Matrix.h"
          HEAPF32[$260>>2] = $252; //@line 283 "./common/Matrix.h"
          $261 = ((($pVerts$07)) + 24|0); //@line 284 "./common/Matrix.h"
          HEAPF32[$261>>2] = $259; //@line 284 "./common/Matrix.h"
          $262 = +HEAPF32[$t>>2]; //@line 280 "./common/Matrix.h"
          $263 = ((($201)) + 40|0); //@line 280 "./common/Matrix.h"
          $264 = +HEAPF32[$263>>2]; //@line 280 "./common/Matrix.h"
          $265 = $262 * $264; //@line 280 "./common/Matrix.h"
          $266 = +HEAPF32[$39>>2]; //@line 280 "./common/Matrix.h"
          $267 = ((($201)) + 44|0); //@line 280 "./common/Matrix.h"
          $268 = +HEAPF32[$267>>2]; //@line 280 "./common/Matrix.h"
          $269 = $266 * $268; //@line 280 "./common/Matrix.h"
          $270 = $265 + $269; //@line 280 "./common/Matrix.h"
          $271 = +HEAPF32[$40>>2]; //@line 280 "./common/Matrix.h"
          $272 = $271 + $270; //@line 280 "./common/Matrix.h"
          $273 = +HEAPF32[$41>>2]; //@line 281 "./common/Matrix.h"
          $274 = $264 * $273; //@line 281 "./common/Matrix.h"
          $275 = +HEAPF32[$42>>2]; //@line 281 "./common/Matrix.h"
          $276 = $268 * $275; //@line 281 "./common/Matrix.h"
          $277 = $274 + $276; //@line 281 "./common/Matrix.h"
          $278 = +HEAPF32[$43>>2]; //@line 281 "./common/Matrix.h"
          $279 = $278 + $277; //@line 281 "./common/Matrix.h"
          $280 = ((($pVerts$07)) + 40|0); //@line 283 "./common/Matrix.h"
          HEAPF32[$280>>2] = $272; //@line 283 "./common/Matrix.h"
          $281 = ((($pVerts$07)) + 44|0); //@line 284 "./common/Matrix.h"
          HEAPF32[$281>>2] = $279; //@line 284 "./common/Matrix.h"
          $282 = +HEAPF32[$t>>2]; //@line 280 "./common/Matrix.h"
          $283 = ((($201)) + 60|0); //@line 280 "./common/Matrix.h"
          $284 = +HEAPF32[$283>>2]; //@line 280 "./common/Matrix.h"
          $285 = $282 * $284; //@line 280 "./common/Matrix.h"
          $286 = +HEAPF32[$39>>2]; //@line 280 "./common/Matrix.h"
          $287 = ((($201)) + 64|0); //@line 280 "./common/Matrix.h"
          $288 = +HEAPF32[$287>>2]; //@line 280 "./common/Matrix.h"
          $289 = $286 * $288; //@line 280 "./common/Matrix.h"
          $290 = $285 + $289; //@line 280 "./common/Matrix.h"
          $291 = +HEAPF32[$40>>2]; //@line 280 "./common/Matrix.h"
          $292 = $291 + $290; //@line 280 "./common/Matrix.h"
          $293 = +HEAPF32[$41>>2]; //@line 281 "./common/Matrix.h"
          $294 = $284 * $293; //@line 281 "./common/Matrix.h"
          $295 = +HEAPF32[$42>>2]; //@line 281 "./common/Matrix.h"
          $296 = $288 * $295; //@line 281 "./common/Matrix.h"
          $297 = $294 + $296; //@line 281 "./common/Matrix.h"
          $298 = +HEAPF32[$43>>2]; //@line 281 "./common/Matrix.h"
          $299 = $298 + $297; //@line 281 "./common/Matrix.h"
          $300 = ((($pVerts$07)) + 60|0); //@line 283 "./common/Matrix.h"
          HEAPF32[$300>>2] = $292; //@line 283 "./common/Matrix.h"
          $301 = ((($pVerts$07)) + 64|0); //@line 284 "./common/Matrix.h"
          HEAPF32[$301>>2] = $299; //@line 284 "./common/Matrix.h"
          $302 = ((($p$08)) + 88|0); //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          $303 = +HEAPF32[$302>>2]; //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          $304 = $303 * 255.0; //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          $305 = (~~(($304))&255); //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          $306 = ((($p$08)) + 92|0); //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          $307 = +HEAPF32[$306>>2]; //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          $308 = $307 * 255.0; //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          $309 = (~~(($308))&255); //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          $310 = ((($p$08)) + 96|0); //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          $311 = +HEAPF32[$310>>2]; //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          $312 = $311 * 255.0; //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          $313 = (~~(($312))&255); //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          $314 = ((($p$08)) + 100|0); //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          $315 = +HEAPF32[$314>>2]; //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          $316 = $315 * 255.0; //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          $317 = (~~(($316))&255); //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          $318 = ((($201)) + 8|0); //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $319 = HEAP32[$318>>2]|0; //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $320 = ((($pVerts$07)) + 8|0); //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP32[$320>>2] = $319; //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $321 = ((($201)) + 12|0); //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $322 = HEAP32[$321>>2]|0; //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $323 = ((($pVerts$07)) + 12|0); //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP32[$323>>2] = $322; //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $324 = ((($pVerts$07)) + 16|0); //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$324>>0] = $305; //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          $325 = ((($pVerts$07)) + 17|0); //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$325>>0] = $309; //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          $326 = ((($pVerts$07)) + 18|0); //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$326>>0] = $313; //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          $327 = ((($pVerts$07)) + 19|0); //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$327>>0] = $317; //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          $328 = ((($201)) + 28|0); //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $329 = HEAP32[$328>>2]|0; //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $330 = ((($pVerts$07)) + 28|0); //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP32[$330>>2] = $329; //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $331 = ((($201)) + 32|0); //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $332 = HEAP32[$331>>2]|0; //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $333 = ((($pVerts$07)) + 32|0); //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP32[$333>>2] = $332; //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $334 = ((($pVerts$07)) + 36|0); //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$334>>0] = $305; //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          $335 = ((($pVerts$07)) + 37|0); //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$335>>0] = $309; //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          $336 = ((($pVerts$07)) + 38|0); //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$336>>0] = $313; //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          $337 = ((($pVerts$07)) + 39|0); //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$337>>0] = $317; //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          $338 = ((($201)) + 48|0); //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $339 = HEAP32[$338>>2]|0; //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $340 = ((($pVerts$07)) + 48|0); //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP32[$340>>2] = $339; //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $341 = ((($201)) + 52|0); //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $342 = HEAP32[$341>>2]|0; //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $343 = ((($pVerts$07)) + 52|0); //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP32[$343>>2] = $342; //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $344 = ((($pVerts$07)) + 56|0); //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$344>>0] = $305; //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          $345 = ((($pVerts$07)) + 57|0); //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$345>>0] = $309; //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          $346 = ((($pVerts$07)) + 58|0); //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$346>>0] = $313; //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          $347 = ((($pVerts$07)) + 59|0); //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$347>>0] = $317; //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          $348 = ((($201)) + 68|0); //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $349 = HEAP32[$348>>2]|0; //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $350 = ((($pVerts$07)) + 68|0); //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP32[$350>>2] = $349; //@line 120 "modules/graphics/opengl/ParticleSystem.cpp"
          $351 = ((($201)) + 72|0); //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $352 = HEAP32[$351>>2]|0; //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $353 = ((($pVerts$07)) + 72|0); //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP32[$353>>2] = $352; //@line 121 "modules/graphics/opengl/ParticleSystem.cpp"
          $354 = ((($pVerts$07)) + 76|0); //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$354>>0] = $305; //@line 125 "modules/graphics/opengl/ParticleSystem.cpp"
          $355 = ((($pVerts$07)) + 77|0); //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$355>>0] = $309; //@line 126 "modules/graphics/opengl/ParticleSystem.cpp"
          $356 = ((($pVerts$07)) + 78|0); //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$356>>0] = $313; //@line 127 "modules/graphics/opengl/ParticleSystem.cpp"
          $357 = ((($pVerts$07)) + 79|0); //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          HEAP8[$357>>0] = $317; //@line 128 "modules/graphics/opengl/ParticleSystem.cpp"
          $358 = ((($pVerts$07)) + 80|0); //@line 131 "modules/graphics/opengl/ParticleSystem.cpp"
          $359 = ((($p$08)) + 4|0); //@line 132 "modules/graphics/opengl/ParticleSystem.cpp"
          $360 = HEAP32[$359>>2]|0; //@line 132 "modules/graphics/opengl/ParticleSystem.cpp"
          $361 = ($360|0)==(0|0); //@line 108 "modules/graphics/opengl/ParticleSystem.cpp"
          if ($361) {
           label = 24;
           break L21;
          } else {
           $p$08 = $360;$pVerts$07 = $358;
          }
         }
         $208 = ___cxa_find_matching_catch_2()|0;
         $209 = tempRet0;
         $lpad$phi$index = $208;$lpad$phi$index3 = $209;
         break;
        }
       }
      } while(0);
      if ((label|0) == 24) {
       $362 = HEAP32[$3>>2]|0; //@line 117 "./common/Object.h"
       $363 = HEAP32[$362>>2]|0; //@line 135 "modules/graphics/opengl/ParticleSystem.cpp"
       $364 = ((($363)) + 52|0); //@line 135 "modules/graphics/opengl/ParticleSystem.cpp"
       $365 = HEAP32[$364>>2]|0; //@line 135 "modules/graphics/opengl/ParticleSystem.cpp"
       __THREW__ = 0;
       $366 = (invoke_ii($365|0,($362|0))|0); //@line 135 "modules/graphics/opengl/ParticleSystem.cpp"
       $367 = __THREW__; __THREW__ = 0;
       $368 = $367&1;
       do {
        if (!($368)) {
         $369 = HEAP32[$366>>2]|0; //@line 135 "modules/graphics/opengl/ParticleSystem.cpp"
         __THREW__ = 0;
         invoke_vii(4005,(634044|0),($369|0)); //@line 135 "modules/graphics/opengl/ParticleSystem.cpp"
         $370 = __THREW__; __THREW__ = 0;
         $371 = $370&1;
         if (!($371)) {
          __THREW__ = 0;
          invoke_vi(4007,(634044|0)); //@line 136 "modules/graphics/opengl/ParticleSystem.cpp"
          $372 = __THREW__; __THREW__ = 0;
          $373 = $372&1;
          if (!($373)) {
           __THREW__ = 0;
           invoke_vii(4006,(634044|0),7); //@line 138 "modules/graphics/opengl/ParticleSystem.cpp"
           $374 = __THREW__; __THREW__ = 0;
           $375 = $374&1;
           if (!($375)) {
            $376 = HEAP32[$9>>2]|0; //@line 140 "modules/graphics/opengl/ParticleSystem.cpp"
            $377 = ((($376)) + 16|0); //@line 140 "modules/graphics/opengl/ParticleSystem.cpp"
            $378 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
            __THREW__ = 0;
            invoke_viiiiii($378|0,2,4,5121,1,20,($377|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
            $379 = __THREW__; __THREW__ = 0;
            $380 = $379&1;
            if (!($380)) {
             $381 = HEAP32[$9>>2]|0; //@line 141 "modules/graphics/opengl/ParticleSystem.cpp"
             $382 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
             __THREW__ = 0;
             invoke_viiiiii($382|0,0,2,5126,0,20,($381|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
             $383 = __THREW__; __THREW__ = 0;
             $384 = $383&1;
             if (!($384)) {
              $385 = HEAP32[$9>>2]|0; //@line 142 "modules/graphics/opengl/ParticleSystem.cpp"
              $386 = ((($385)) + 8|0); //@line 142 "modules/graphics/opengl/ParticleSystem.cpp"
              $387 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
              __THREW__ = 0;
              invoke_viiiiii($387|0,1,2,5126,0,20,($386|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
              $388 = __THREW__; __THREW__ = 0;
              $389 = $388&1;
              if (!($389)) {
               $390 = ((($this)) + 232|0); //@line 144 "modules/graphics/opengl/ParticleSystem.cpp"
               __THREW__ = 0;
               $391 = (invoke_iii(4104,($390|0),($1|0))|0); //@line 144 "modules/graphics/opengl/ParticleSystem.cpp"
               $392 = __THREW__; __THREW__ = 0;
               $393 = $392&1;
               if ($393) {
                break;
               }
               $394 = HEAP32[158426]|0; //@line 357 "modules/graphics/opengl/GLBuffer.h"
               __THREW__ = 0;
               $395 = (invoke_iii(4105,($390|0),($394|0))|0); //@line 357 "modules/graphics/opengl/GLBuffer.h"
               $396 = __THREW__; __THREW__ = 0;
               $397 = $396&1;
               if ($397) {
                break;
               }
               __THREW__ = 0;
               $398 = (invoke_iii(4106,($390|0),0)|0); //@line 151 "modules/graphics/opengl/ParticleSystem.cpp"
               $399 = __THREW__; __THREW__ = 0;
               $400 = $399&1;
               if ($400) {
                break;
               }
               __THREW__ = 0;
               invoke_viiiii(4091,(634044|0),4,($391|0),($395|0),($398|0)); //@line 151 "modules/graphics/opengl/ParticleSystem.cpp"
               $401 = __THREW__; __THREW__ = 0;
               $402 = $401&1;
               if ($402) {
                break;
               }
               __ZN4love7Matrix3D2Ev($t); //@line 152 "modules/graphics/opengl/ParticleSystem.cpp"
               __THREW__ = 0;
               invoke_vi(4009,(634044|0)); //@line 130 "modules/graphics/opengl/OpenGL.h"
               $403 = __THREW__; __THREW__ = 0;
               $404 = $403&1;
               if ($404) {
                $405 = ___cxa_find_matching_catch_3(0|0)|0;
                $406 = tempRet0;
                ___clang_call_terminate($405); //@line 130 "modules/graphics/opengl/OpenGL.h"
                // unreachable; //@line 130 "modules/graphics/opengl/OpenGL.h"
               } else {
                STACKTOP = sp;return; //@line 152 "modules/graphics/opengl/ParticleSystem.cpp"
               }
              }
             }
            }
           }
          }
         }
        }
       } while(0);
       $210 = ___cxa_find_matching_catch_2()|0;
       $211 = tempRet0;
       $lpad$phi$index = $210;$lpad$phi$index3 = $211;
      }
      __ZN4love7Matrix3D2Ev($t); //@line 152 "modules/graphics/opengl/ParticleSystem.cpp"
      $$0 = $lpad$phi$index;$$01 = $lpad$phi$index3;
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 17) {
  $204 = ___cxa_find_matching_catch_2()|0;
  $205 = tempRet0;
  $$0 = $204;$$01 = $205;
 }
 __THREW__ = 0;
 invoke_vi(4009,(634044|0)); //@line 130 "modules/graphics/opengl/OpenGL.h"
 $407 = __THREW__; __THREW__ = 0;
 $408 = $407&1;
 if ($408) {
  $409 = ___cxa_find_matching_catch_3(0|0)|0;
  $410 = tempRet0;
  ___clang_call_terminate($409); //@line 130 "modules/graphics/opengl/OpenGL.h"
  // unreachable; //@line 130 "modules/graphics/opengl/OpenGL.h"
 } else {
  ___resumeException($$0|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl14ParticleSystem5cloneEv($this) {
 $this = $this|0;
 var $$0$i = 0, $$01$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(236)|0); //@line 76 "modules/graphics/opengl/ParticleSystem.cpp"
 __THREW__ = 0;
 invoke_vii(4107,($0|0),($this|0)); //@line 48 "modules/graphics/opengl/ParticleSystem.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  $eh$lpad$body$index2Z2D = $17;$eh$lpad$body$indexZ2D = $16;
  __ZdlPv($0); //@line 76 "modules/graphics/opengl/ParticleSystem.cpp"
  ___resumeException($eh$lpad$body$indexZ2D|0);
  // unreachable;
 }
 HEAP32[$0>>2] = (25048); //@line 51 "modules/graphics/opengl/ParticleSystem.cpp"
 $3 = ((($0)) + 228|0); //@line 49 "modules/graphics/opengl/ParticleSystem.cpp"
 HEAP32[$3>>2] = 0; //@line 49 "modules/graphics/opengl/ParticleSystem.cpp"
 $4 = ((($0)) + 232|0); //@line 50 "modules/graphics/opengl/ParticleSystem.cpp"
 $5 = ((($this)) + 232|0); //@line 50 "modules/graphics/opengl/ParticleSystem.cpp"
 __THREW__ = 0;
 invoke_vii(4108,($4|0),($5|0)); //@line 50 "modules/graphics/opengl/ParticleSystem.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 do {
  if ($7) {
   $12 = ___cxa_find_matching_catch_2()|0;
   $13 = tempRet0;
   $$0$i = $13;$$01$i = $12;
  } else {
   $8 = ((($0)) + 36|0); //@line 52 "modules/graphics/opengl/ParticleSystem.cpp"
   $9 = HEAP32[$8>>2]|0; //@line 52 "modules/graphics/opengl/ParticleSystem.cpp"
   __THREW__ = 0;
   invoke_vii(4109,($0|0),($9|0)); //@line 52 "modules/graphics/opengl/ParticleSystem.cpp"
   $10 = __THREW__; __THREW__ = 0;
   $11 = $10&1;
   if ($11) {
    $14 = ___cxa_find_matching_catch_2()|0;
    $15 = tempRet0;
    __ZN4love8graphics6opengl11QuadIndicesD2Ev($4); //@line 53 "modules/graphics/opengl/ParticleSystem.cpp"
    $$0$i = $15;$$01$i = $14;
    break;
   } else {
    return ($0|0); //@line 76 "modules/graphics/opengl/ParticleSystem.cpp"
   }
  }
 } while(0);
 __ZN4love8graphics14ParticleSystemD2Ev($0); //@line 53 "modules/graphics/opengl/ParticleSystem.cpp"
 $eh$lpad$body$index2Z2D = $$0$i;$eh$lpad$body$indexZ2D = $$01$i;
 __ZdlPv($0); //@line 76 "modules/graphics/opengl/ParticleSystem.cpp"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8graphics6opengl14ParticleSystem14createVerticesEj($this,$numparticles) {
 $this = $this|0;
 $numparticles = $numparticles|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$arith = 0, $$overflow = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = $numparticles << 2; //@line 64 "modules/graphics/opengl/ParticleSystem.cpp"
 $$arith = ($0*20)|0; //@line 64 "modules/graphics/opengl/ParticleSystem.cpp"
 $$overflow = ($0>>>0)>(214748364); //@line 64 "modules/graphics/opengl/ParticleSystem.cpp"
 $1 = $$overflow ? -1 : $$arith; //@line 64 "modules/graphics/opengl/ParticleSystem.cpp"
 __THREW__ = 0;
 $2 = (invoke_ii(3492,($1|0))|0); //@line 64 "modules/graphics/opengl/ParticleSystem.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  $5 = ((($this)) + 228|0); //@line 65 "modules/graphics/opengl/ParticleSystem.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 65 "modules/graphics/opengl/ParticleSystem.cpp"
  $7 = ($6|0)==(0|0); //@line 65 "modules/graphics/opengl/ParticleSystem.cpp"
  if ($7) {
   HEAP32[$5>>2] = $2; //@line 66 "modules/graphics/opengl/ParticleSystem.cpp"
   STACKTOP = sp;return; //@line 72 "modules/graphics/opengl/ParticleSystem.cpp"
  }
  __ZdaPv($6); //@line 65 "modules/graphics/opengl/ParticleSystem.cpp"
  HEAP32[$5>>2] = $2; //@line 66 "modules/graphics/opengl/ParticleSystem.cpp"
  STACKTOP = sp;return; //@line 72 "modules/graphics/opengl/ParticleSystem.cpp"
 }
 $8 = ___cxa_find_matching_catch_3(8992|0)|0;
 $9 = tempRet0;
 $10 = (_llvm_eh_typeid_for((8992|0))|0); //@line 67 "modules/graphics/opengl/ParticleSystem.cpp"
 $11 = ($9|0)==($10|0); //@line 67 "modules/graphics/opengl/ParticleSystem.cpp"
 if (!($11)) {
  $$1 = $8;$$12 = $9;
  ___resumeException($$1|0);
  // unreachable;
 }
 (___cxa_begin_catch(($8|0))|0); //@line 67 "modules/graphics/opengl/ParticleSystem.cpp"
 $12 = (___cxa_allocate_exception(16)|0); //@line 70 "modules/graphics/opengl/ParticleSystem.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($12|0),(507610|0),($vararg_buffer|0)); //@line 70 "modules/graphics/opengl/ParticleSystem.cpp"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if ($14) {
  $18 = ___cxa_find_matching_catch_2()|0;
  $19 = tempRet0;
  ___cxa_free_exception(($12|0)); //@line 70 "modules/graphics/opengl/ParticleSystem.cpp"
  $$0 = $18;$$01 = $19;
 } else {
  __THREW__ = 0;
  invoke_viii(3350,($12|0),(96|0),(344|0)); //@line 70 "modules/graphics/opengl/ParticleSystem.cpp"
  $15 = __THREW__; __THREW__ = 0;
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  $$0 = $16;$$01 = $17;
 }
 __THREW__ = 0;
 invoke_v(3378); //@line 71 "modules/graphics/opengl/ParticleSystem.cpp"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 71 "modules/graphics/opengl/ParticleSystem.cpp"
  // unreachable; //@line 71 "modules/graphics/opengl/ParticleSystem.cpp"
 } else {
  $$1 = $$0;$$12 = $$01;
  ___resumeException($$1|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl14ParticleSystem13setBufferSizeEj($this,$size) {
 $this = $this|0;
 $size = $size|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 __ZN4love8graphics14ParticleSystem13setBufferSizeEj($this,$size); //@line 81 "modules/graphics/opengl/ParticleSystem.cpp"
 $1 = ((($this)) + 232|0); //@line 83 "modules/graphics/opengl/ParticleSystem.cpp"
 __ZN4love8graphics6opengl11QuadIndicesC2Ej($0,$size); //@line 83 "modules/graphics/opengl/ParticleSystem.cpp"
 __THREW__ = 0;
 (invoke_iii(4036,($1|0),($0|0))|0); //@line 83 "modules/graphics/opengl/ParticleSystem.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  __ZN4love8graphics6opengl11QuadIndicesD2Ev($0); //@line 83 "modules/graphics/opengl/ParticleSystem.cpp"
  ___resumeException($4|0);
  // unreachable;
 } else {
  __ZN4love8graphics6opengl11QuadIndicesD2Ev($0); //@line 83 "modules/graphics/opengl/ParticleSystem.cpp"
  __ZN4love8graphics6opengl14ParticleSystem14createVerticesEj($this,$size); //@line 84 "modules/graphics/opengl/ParticleSystem.cpp"
  STACKTOP = sp;return; //@line 85 "modules/graphics/opengl/ParticleSystem.cpp"
 }
}
function __ZN4love8graphics6opengl14ParticleSystemC2EPNS0_7TextureEj($this,$texture,$size) {
 $this = $this|0;
 $texture = $texture|0;
 $size = $size|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics14ParticleSystemC2EPNS0_7TextureEj($this,$texture,$size); //@line 40 "modules/graphics/opengl/ParticleSystem.cpp"
 HEAP32[$this>>2] = (25048); //@line 43 "modules/graphics/opengl/ParticleSystem.cpp"
 $0 = ((($this)) + 228|0); //@line 41 "modules/graphics/opengl/ParticleSystem.cpp"
 HEAP32[$0>>2] = 0; //@line 41 "modules/graphics/opengl/ParticleSystem.cpp"
 $1 = ((($this)) + 232|0); //@line 42 "modules/graphics/opengl/ParticleSystem.cpp"
 __THREW__ = 0;
 invoke_vii(4020,($1|0),($size|0)); //@line 42 "modules/graphics/opengl/ParticleSystem.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  $$0 = $7;$$01 = $6;
  __ZN4love8graphics14ParticleSystemD2Ev($this); //@line 45 "modules/graphics/opengl/ParticleSystem.cpp"
  ___resumeException($$01|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_vii(4109,($this|0),($size|0)); //@line 44 "modules/graphics/opengl/ParticleSystem.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  return; //@line 45 "modules/graphics/opengl/ParticleSystem.cpp"
 }
 $8 = ___cxa_find_matching_catch_2()|0;
 $9 = tempRet0;
 __ZN4love8graphics6opengl11QuadIndicesD2Ev($1); //@line 45 "modules/graphics/opengl/ParticleSystem.cpp"
 $$0 = $9;$$01 = $8;
 __ZN4love8graphics14ParticleSystemD2Ev($this); //@line 45 "modules/graphics/opengl/ParticleSystem.cpp"
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN4love8graphics6opengl8PolylineD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (25084); //@line 356 "modules/graphics/opengl/Polyline.cpp"
 $0 = ((($this)) + 4|0); //@line 357 "modules/graphics/opengl/Polyline.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 357 "modules/graphics/opengl/Polyline.cpp"
 $2 = ($1|0)==(0|0); //@line 357 "modules/graphics/opengl/Polyline.cpp"
 if ($2) {
  return; //@line 359 "modules/graphics/opengl/Polyline.cpp"
 }
 __ZdaPv($1); //@line 358 "modules/graphics/opengl/Polyline.cpp"
 return; //@line 359 "modules/graphics/opengl/Polyline.cpp"
}
function __ZN4love8graphics6opengl8PolylineD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (25084); //@line 356 "modules/graphics/opengl/Polyline.cpp"
 $0 = ((($this)) + 4|0); //@line 357 "modules/graphics/opengl/Polyline.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 357 "modules/graphics/opengl/Polyline.cpp"
 $2 = ($1|0)==(0|0); //@line 357 "modules/graphics/opengl/Polyline.cpp"
 if (!($2)) {
  __ZdaPv($1); //@line 358 "modules/graphics/opengl/Polyline.cpp"
 }
 __ZdlPv($this); //@line 356 "modules/graphics/opengl/Polyline.cpp"
 return; //@line 359 "modules/graphics/opengl/Polyline.cpp"
}
function __ZN4love8graphics6opengl8Polyline26calc_overdraw_vertex_countEb($this,$is_looping) {
 $this = $this|0;
 $is_looping = $is_looping|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 271 "modules/graphics/opengl/Polyline.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 271 "modules/graphics/opengl/Polyline.cpp"
 $2 = $1 << 1; //@line 271 "modules/graphics/opengl/Polyline.cpp"
 $3 = $is_looping ? 0 : 2; //@line 271 "modules/graphics/opengl/Polyline.cpp"
 $4 = (($2) + ($3))|0; //@line 271 "modules/graphics/opengl/Polyline.cpp"
 $5 = ((($this)) + 16|0); //@line 271 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$5>>2] = $4; //@line 271 "modules/graphics/opengl/Polyline.cpp"
 return; //@line 272 "modules/graphics/opengl/Polyline.cpp"
}
function __ZN4love8graphics6opengl8Polyline15render_overdrawERKNSt3__16vectorINS_6VectorENS3_9allocatorIS5_EEEEfb($this,$normals,$pixel_size,$is_looping) {
 $this = $this|0;
 $normals = $normals|0;
 $pixel_size = +$pixel_size;
 $is_looping = $is_looping|0;
 var $$lcssa20 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0, $108 = 0.0, $109 = 0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0;
 var $115 = 0.0, $116 = 0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0.0, $127 = 0.0, $128 = 0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0.0, $136 = 0, $137 = 0.0, $138 = 0.0, $139 = 0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0.0, $148 = 0, $149 = 0.0, $15 = 0, $150 = 0.0;
 var $151 = 0.0, $152 = 0.0, $153 = 0.0, $154 = 0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0.0, $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0;
 var $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0.0, $85 = 0, $86 = 0.0, $87 = 0.0, $88 = 0, $89 = 0.0, $9 = 0, $90 = 0.0;
 var $91 = 0, $92 = 0.0, $93 = 0.0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$014 = 0, $i1$013 = 0, $spacer$sroa$0$0$in = 0.0, $spacer$sroa$0$1$in = 0.0, $spacer$sroa$12$0$in = 0.0, $spacer$sroa$12$1$in = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 277 "modules/graphics/opengl/Polyline.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 277 "modules/graphics/opengl/Polyline.cpp"
 $2 = ($1>>>0)>(1); //@line 277 "modules/graphics/opengl/Polyline.cpp"
 if ($2) {
  $3 = ((($this)) + 8|0); //@line 279 "modules/graphics/opengl/Polyline.cpp"
  $4 = ((($this)) + 4|0); //@line 279 "modules/graphics/opengl/Polyline.cpp"
  $43 = 1;$i$014 = 0;
  while(1) {
   $8 = HEAP32[$3>>2]|0; //@line 279 "modules/graphics/opengl/Polyline.cpp"
   $9 = (($8) + ($i$014<<3)|0); //@line 279 "modules/graphics/opengl/Polyline.cpp"
   $10 = HEAP32[$4>>2]|0; //@line 279 "modules/graphics/opengl/Polyline.cpp"
   $11 = (($10) + ($i$014<<3)|0); //@line 279 "modules/graphics/opengl/Polyline.cpp"
   $12 = $11; //@line 279 "modules/graphics/opengl/Polyline.cpp"
   $13 = $12; //@line 279 "modules/graphics/opengl/Polyline.cpp"
   $14 = HEAP32[$13>>2]|0; //@line 279 "modules/graphics/opengl/Polyline.cpp"
   $15 = (($12) + 4)|0; //@line 279 "modules/graphics/opengl/Polyline.cpp"
   $16 = $15; //@line 279 "modules/graphics/opengl/Polyline.cpp"
   $17 = HEAP32[$16>>2]|0; //@line 279 "modules/graphics/opengl/Polyline.cpp"
   $18 = $9; //@line 279 "modules/graphics/opengl/Polyline.cpp"
   $19 = $18; //@line 279 "modules/graphics/opengl/Polyline.cpp"
   HEAP32[$19>>2] = $14; //@line 279 "modules/graphics/opengl/Polyline.cpp"
   $20 = (($18) + 4)|0; //@line 279 "modules/graphics/opengl/Polyline.cpp"
   $21 = $20; //@line 279 "modules/graphics/opengl/Polyline.cpp"
   HEAP32[$21>>2] = $17; //@line 279 "modules/graphics/opengl/Polyline.cpp"
   $22 = HEAP32[$3>>2]|0; //@line 280 "modules/graphics/opengl/Polyline.cpp"
   $23 = HEAP32[$4>>2]|0; //@line 280 "modules/graphics/opengl/Polyline.cpp"
   $24 = HEAP32[$normals>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $25 = (($24) + ($i$014<<3)|0); //@line 190 "./common/Vector.h"
   $26 = +HEAPF32[$25>>2]; //@line 190 "./common/Vector.h"
   $27 = $26 * $26; //@line 190 "./common/Vector.h"
   $28 = (((($24) + ($i$014<<3)|0)) + 4|0); //@line 190 "./common/Vector.h"
   $29 = +HEAPF32[$28>>2]; //@line 190 "./common/Vector.h"
   $30 = $29 * $29; //@line 190 "./common/Vector.h"
   $31 = $27 + $30; //@line 190 "./common/Vector.h"
   $32 = (+Math_sqrt((+$31))); //@line 190 "./common/Vector.h"
   $33 = $pixel_size / $32; //@line 280 "modules/graphics/opengl/Polyline.cpp"
   $34 = $26 * $33; //@line 242 "./common/Vector.h"
   $35 = $29 * $33; //@line 242 "./common/Vector.h"
   $36 = (($23) + ($i$014<<3)|0); //@line 232 "./common/Vector.h"
   $37 = +HEAPF32[$36>>2]; //@line 232 "./common/Vector.h"
   $38 = $37 + $34; //@line 232 "./common/Vector.h"
   $39 = (((($23) + ($i$014<<3)|0)) + 4|0); //@line 232 "./common/Vector.h"
   $40 = +HEAPF32[$39>>2]; //@line 232 "./common/Vector.h"
   $41 = $40 + $35; //@line 232 "./common/Vector.h"
   $42 = (($22) + ($43<<3)|0); //@line 280 "modules/graphics/opengl/Polyline.cpp"
   HEAPF32[$42>>2] = $38; //@line 280 "modules/graphics/opengl/Polyline.cpp"
   $44 = (((($22) + ($43<<3)|0)) + 4|0); //@line 280 "modules/graphics/opengl/Polyline.cpp"
   HEAPF32[$44>>2] = $41; //@line 280 "modules/graphics/opengl/Polyline.cpp"
   $45 = (($i$014) + 2)|0; //@line 277 "modules/graphics/opengl/Polyline.cpp"
   $46 = $45 | 1; //@line 277 "modules/graphics/opengl/Polyline.cpp"
   $47 = HEAP32[$0>>2]|0; //@line 277 "modules/graphics/opengl/Polyline.cpp"
   $48 = ($46>>>0)<($47>>>0); //@line 277 "modules/graphics/opengl/Polyline.cpp"
   if ($48) {
    $43 = $46;$i$014 = $45;
   } else {
    $$lcssa20 = $47;
    break;
   }
  }
  $5 = ($$lcssa20>>>0)>(1); //@line 283 "modules/graphics/opengl/Polyline.cpp"
  if ($5) {
   $6 = ((($this)) + 8|0); //@line 286 "modules/graphics/opengl/Polyline.cpp"
   $7 = ((($this)) + 4|0); //@line 286 "modules/graphics/opengl/Polyline.cpp"
   $50 = $$lcssa20;$i1$013 = 0;
   while(1) {
    $49 = (($50) - ($i1$013))|0; //@line 285 "modules/graphics/opengl/Polyline.cpp"
    $51 = (($49) + -1)|0; //@line 285 "modules/graphics/opengl/Polyline.cpp"
    $52 = (($50) + ($i1$013))|0; //@line 286 "modules/graphics/opengl/Polyline.cpp"
    $53 = HEAP32[$6>>2]|0; //@line 286 "modules/graphics/opengl/Polyline.cpp"
    $54 = (($53) + ($52<<3)|0); //@line 286 "modules/graphics/opengl/Polyline.cpp"
    $55 = HEAP32[$7>>2]|0; //@line 286 "modules/graphics/opengl/Polyline.cpp"
    $56 = (($55) + ($51<<3)|0); //@line 286 "modules/graphics/opengl/Polyline.cpp"
    $57 = $56; //@line 286 "modules/graphics/opengl/Polyline.cpp"
    $58 = $57; //@line 286 "modules/graphics/opengl/Polyline.cpp"
    $59 = HEAP32[$58>>2]|0; //@line 286 "modules/graphics/opengl/Polyline.cpp"
    $60 = (($57) + 4)|0; //@line 286 "modules/graphics/opengl/Polyline.cpp"
    $61 = $60; //@line 286 "modules/graphics/opengl/Polyline.cpp"
    $62 = HEAP32[$61>>2]|0; //@line 286 "modules/graphics/opengl/Polyline.cpp"
    $63 = $54; //@line 286 "modules/graphics/opengl/Polyline.cpp"
    $64 = $63; //@line 286 "modules/graphics/opengl/Polyline.cpp"
    HEAP32[$64>>2] = $59; //@line 286 "modules/graphics/opengl/Polyline.cpp"
    $65 = (($63) + 4)|0; //@line 286 "modules/graphics/opengl/Polyline.cpp"
    $66 = $65; //@line 286 "modules/graphics/opengl/Polyline.cpp"
    HEAP32[$66>>2] = $62; //@line 286 "modules/graphics/opengl/Polyline.cpp"
    $67 = HEAP32[$0>>2]|0; //@line 287 "modules/graphics/opengl/Polyline.cpp"
    $68 = $i1$013 | 1; //@line 287 "modules/graphics/opengl/Polyline.cpp"
    $69 = (($68) + ($67))|0; //@line 287 "modules/graphics/opengl/Polyline.cpp"
    $70 = HEAP32[$6>>2]|0; //@line 287 "modules/graphics/opengl/Polyline.cpp"
    $71 = HEAP32[$7>>2]|0; //@line 287 "modules/graphics/opengl/Polyline.cpp"
    $72 = HEAP32[$normals>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $73 = (($72) + ($i1$013<<3)|0); //@line 190 "./common/Vector.h"
    $74 = +HEAPF32[$73>>2]; //@line 190 "./common/Vector.h"
    $75 = $74 * $74; //@line 190 "./common/Vector.h"
    $76 = (((($72) + ($i1$013<<3)|0)) + 4|0); //@line 190 "./common/Vector.h"
    $77 = +HEAPF32[$76>>2]; //@line 190 "./common/Vector.h"
    $78 = $77 * $77; //@line 190 "./common/Vector.h"
    $79 = $75 + $78; //@line 190 "./common/Vector.h"
    $80 = (+Math_sqrt((+$79))); //@line 190 "./common/Vector.h"
    $81 = $pixel_size / $80; //@line 287 "modules/graphics/opengl/Polyline.cpp"
    $82 = (($72) + ($51<<3)|0); //@line 242 "./common/Vector.h"
    $83 = +HEAPF32[$82>>2]; //@line 242 "./common/Vector.h"
    $84 = $83 * $81; //@line 242 "./common/Vector.h"
    $85 = (((($72) + ($51<<3)|0)) + 4|0); //@line 242 "./common/Vector.h"
    $86 = +HEAPF32[$85>>2]; //@line 242 "./common/Vector.h"
    $87 = $81 * $86; //@line 242 "./common/Vector.h"
    $88 = (($71) + ($51<<3)|0); //@line 232 "./common/Vector.h"
    $89 = +HEAPF32[$88>>2]; //@line 232 "./common/Vector.h"
    $90 = $84 + $89; //@line 232 "./common/Vector.h"
    $91 = (((($71) + ($51<<3)|0)) + 4|0); //@line 232 "./common/Vector.h"
    $92 = +HEAPF32[$91>>2]; //@line 232 "./common/Vector.h"
    $93 = $87 + $92; //@line 232 "./common/Vector.h"
    $94 = (($70) + ($69<<3)|0); //@line 287 "modules/graphics/opengl/Polyline.cpp"
    HEAPF32[$94>>2] = $90; //@line 287 "modules/graphics/opengl/Polyline.cpp"
    $95 = (((($70) + ($69<<3)|0)) + 4|0); //@line 287 "modules/graphics/opengl/Polyline.cpp"
    HEAPF32[$95>>2] = $93; //@line 287 "modules/graphics/opengl/Polyline.cpp"
    $96 = (($i1$013) + 2)|0; //@line 283 "modules/graphics/opengl/Polyline.cpp"
    $97 = $96 | 1; //@line 283 "modules/graphics/opengl/Polyline.cpp"
    $98 = HEAP32[$0>>2]|0; //@line 283 "modules/graphics/opengl/Polyline.cpp"
    $99 = ($97>>>0)<($98>>>0); //@line 283 "modules/graphics/opengl/Polyline.cpp"
    if ($99) {
     $50 = $98;$i1$013 = $96;
    } else {
     $132 = $98;
     break;
    }
   }
  } else {
   $132 = $$lcssa20;
  }
 } else {
  $132 = $1;
 }
 if ($is_looping) {
  return; //@line 316 "modules/graphics/opengl/Polyline.cpp"
 }
 $100 = ((($this)) + 8|0); //@line 300 "modules/graphics/opengl/Polyline.cpp"
 $101 = HEAP32[$100>>2]|0; //@line 313 "modules/graphics/opengl/Polyline.cpp"
 $102 = ((($101)) + 8|0); //@line 237 "./common/Vector.h"
 $103 = +HEAPF32[$102>>2]; //@line 257 "./common/Vector.h"
 $104 = ((($101)) + 24|0); //@line 305 "./common/Vector.h"
 $105 = +HEAPF32[$104>>2]; //@line 305 "./common/Vector.h"
 $106 = $103 - $105; //@line 237 "./common/Vector.h"
 $107 = ((($101)) + 12|0); //@line 237 "./common/Vector.h"
 $108 = +HEAPF32[$107>>2]; //@line 258 "./common/Vector.h"
 $109 = ((($101)) + 28|0); //@line 310 "./common/Vector.h"
 $110 = +HEAPF32[$109>>2]; //@line 310 "./common/Vector.h"
 $111 = $108 - $110; //@line 237 "./common/Vector.h"
 $112 = $106 * $106; //@line 190 "./common/Vector.h"
 $113 = $111 * $111; //@line 190 "./common/Vector.h"
 $114 = $112 + $113; //@line 190 "./common/Vector.h"
 $115 = (+Math_sqrt((+$114))); //@line 190 "./common/Vector.h"
 $116 = $115 > 0.0; //@line 208 "./common/Vector.h"
 if ($116) {
  $117 = $pixel_size / $115; //@line 209 "./common/Vector.h"
  $118 = $106 * $117; //@line 269 "./common/Vector.h"
  $119 = $111 * $117; //@line 270 "./common/Vector.h"
  $spacer$sroa$0$0$in = $118;$spacer$sroa$12$0$in = $119;
 } else {
  $spacer$sroa$0$0$in = $106;$spacer$sroa$12$0$in = $111;
 }
 $120 = $spacer$sroa$0$0$in + $103; //@line 257 "./common/Vector.h"
 HEAPF32[$102>>2] = $120; //@line 257 "./common/Vector.h"
 $121 = $spacer$sroa$12$0$in + $108; //@line 258 "./common/Vector.h"
 HEAPF32[$107>>2] = $121; //@line 258 "./common/Vector.h"
 $122 = ((($this)) + 16|0); //@line 303 "modules/graphics/opengl/Polyline.cpp"
 $123 = HEAP32[$122>>2]|0; //@line 313 "modules/graphics/opengl/Polyline.cpp"
 $124 = (($123) + -3)|0; //@line 303 "modules/graphics/opengl/Polyline.cpp"
 $125 = (($101) + ($124<<3)|0); //@line 257 "./common/Vector.h"
 $126 = +HEAPF32[$125>>2]; //@line 257 "./common/Vector.h"
 $127 = $spacer$sroa$0$0$in + $126; //@line 257 "./common/Vector.h"
 HEAPF32[$125>>2] = $127; //@line 257 "./common/Vector.h"
 $128 = (((($101) + ($124<<3)|0)) + 4|0); //@line 258 "./common/Vector.h"
 $129 = +HEAPF32[$128>>2]; //@line 258 "./common/Vector.h"
 $130 = $spacer$sroa$12$0$in + $129; //@line 258 "./common/Vector.h"
 HEAPF32[$128>>2] = $130; //@line 258 "./common/Vector.h"
 $131 = (($132) + -1)|0; //@line 306 "modules/graphics/opengl/Polyline.cpp"
 $133 = (($132) + -3)|0; //@line 306 "modules/graphics/opengl/Polyline.cpp"
 $134 = (($101) + ($131<<3)|0); //@line 237 "./common/Vector.h"
 $135 = +HEAPF32[$134>>2]; //@line 257 "./common/Vector.h"
 $136 = (($101) + ($133<<3)|0); //@line 305 "./common/Vector.h"
 $137 = +HEAPF32[$136>>2]; //@line 305 "./common/Vector.h"
 $138 = $135 - $137; //@line 237 "./common/Vector.h"
 $139 = (((($101) + ($131<<3)|0)) + 4|0); //@line 237 "./common/Vector.h"
 $140 = +HEAPF32[$139>>2]; //@line 258 "./common/Vector.h"
 $141 = (((($101) + ($133<<3)|0)) + 4|0); //@line 310 "./common/Vector.h"
 $142 = +HEAPF32[$141>>2]; //@line 310 "./common/Vector.h"
 $143 = $140 - $142; //@line 237 "./common/Vector.h"
 $144 = $138 * $138; //@line 190 "./common/Vector.h"
 $145 = $143 * $143; //@line 190 "./common/Vector.h"
 $146 = $144 + $145; //@line 190 "./common/Vector.h"
 $147 = (+Math_sqrt((+$146))); //@line 190 "./common/Vector.h"
 $148 = $147 > 0.0; //@line 208 "./common/Vector.h"
 if ($148) {
  $149 = $pixel_size / $147; //@line 209 "./common/Vector.h"
  $150 = $138 * $149; //@line 269 "./common/Vector.h"
  $151 = $143 * $149; //@line 270 "./common/Vector.h"
  $spacer$sroa$0$1$in = $150;$spacer$sroa$12$1$in = $151;
 } else {
  $spacer$sroa$0$1$in = $138;$spacer$sroa$12$1$in = $143;
 }
 $152 = $spacer$sroa$0$1$in + $135; //@line 257 "./common/Vector.h"
 HEAPF32[$134>>2] = $152; //@line 257 "./common/Vector.h"
 $153 = $spacer$sroa$12$1$in + $140; //@line 258 "./common/Vector.h"
 HEAPF32[$139>>2] = $153; //@line 258 "./common/Vector.h"
 $154 = (($132) + 1)|0; //@line 309 "modules/graphics/opengl/Polyline.cpp"
 $155 = (($101) + ($154<<3)|0); //@line 257 "./common/Vector.h"
 $156 = +HEAPF32[$155>>2]; //@line 257 "./common/Vector.h"
 $157 = $spacer$sroa$0$1$in + $156; //@line 257 "./common/Vector.h"
 HEAPF32[$155>>2] = $157; //@line 257 "./common/Vector.h"
 $158 = (((($101) + ($154<<3)|0)) + 4|0); //@line 258 "./common/Vector.h"
 $159 = +HEAPF32[$158>>2]; //@line 258 "./common/Vector.h"
 $160 = $spacer$sroa$12$1$in + $159; //@line 258 "./common/Vector.h"
 HEAPF32[$158>>2] = $160; //@line 258 "./common/Vector.h"
 $161 = (($123) + -2)|0; //@line 313 "modules/graphics/opengl/Polyline.cpp"
 $162 = (($101) + ($161<<3)|0); //@line 313 "modules/graphics/opengl/Polyline.cpp"
 $163 = $101; //@line 313 "modules/graphics/opengl/Polyline.cpp"
 $164 = $163; //@line 313 "modules/graphics/opengl/Polyline.cpp"
 $165 = HEAP32[$164>>2]|0; //@line 313 "modules/graphics/opengl/Polyline.cpp"
 $166 = (($163) + 4)|0; //@line 313 "modules/graphics/opengl/Polyline.cpp"
 $167 = $166; //@line 313 "modules/graphics/opengl/Polyline.cpp"
 $168 = HEAP32[$167>>2]|0; //@line 313 "modules/graphics/opengl/Polyline.cpp"
 $169 = $162; //@line 313 "modules/graphics/opengl/Polyline.cpp"
 $170 = $169; //@line 313 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$170>>2] = $165; //@line 313 "modules/graphics/opengl/Polyline.cpp"
 $171 = (($169) + 4)|0; //@line 313 "modules/graphics/opengl/Polyline.cpp"
 $172 = $171; //@line 313 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$172>>2] = $168; //@line 313 "modules/graphics/opengl/Polyline.cpp"
 $173 = HEAP32[$122>>2]|0; //@line 314 "modules/graphics/opengl/Polyline.cpp"
 $174 = (($173) + -1)|0; //@line 314 "modules/graphics/opengl/Polyline.cpp"
 $175 = HEAP32[$100>>2]|0; //@line 314 "modules/graphics/opengl/Polyline.cpp"
 $176 = (($175) + ($174<<3)|0); //@line 314 "modules/graphics/opengl/Polyline.cpp"
 $177 = ((($175)) + 8|0); //@line 314 "modules/graphics/opengl/Polyline.cpp"
 $178 = $177; //@line 314 "modules/graphics/opengl/Polyline.cpp"
 $179 = $178; //@line 314 "modules/graphics/opengl/Polyline.cpp"
 $180 = HEAP32[$179>>2]|0; //@line 314 "modules/graphics/opengl/Polyline.cpp"
 $181 = (($178) + 4)|0; //@line 314 "modules/graphics/opengl/Polyline.cpp"
 $182 = $181; //@line 314 "modules/graphics/opengl/Polyline.cpp"
 $183 = HEAP32[$182>>2]|0; //@line 314 "modules/graphics/opengl/Polyline.cpp"
 $184 = $176; //@line 314 "modules/graphics/opengl/Polyline.cpp"
 $185 = $184; //@line 314 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$185>>2] = $180; //@line 314 "modules/graphics/opengl/Polyline.cpp"
 $186 = (($184) + 4)|0; //@line 314 "modules/graphics/opengl/Polyline.cpp"
 $187 = $186; //@line 314 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$187>>2] = $183; //@line 314 "modules/graphics/opengl/Polyline.cpp"
 return; //@line 316 "modules/graphics/opengl/Polyline.cpp"
}
function __ZN4love8graphics6opengl8Polyline16fill_color_arrayEPNS0_6ColorTIhEE($this,$colors) {
 $this = $this|0;
 $colors = $colors|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 442 "modules/graphics/opengl/Polyline.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 442 "modules/graphics/opengl/Polyline.cpp"
 $2 = ($1|0)==(0); //@line 442 "modules/graphics/opengl/Polyline.cpp"
 if ($2) {
  return; //@line 447 "modules/graphics/opengl/Polyline.cpp"
 } else {
  $i$01 = 0;
 }
 while(1) {
  $3 = (($i$01) + 1)|0; //@line 445 "modules/graphics/opengl/Polyline.cpp"
  $4 = $3 & 1; //@line 445 "modules/graphics/opengl/Polyline.cpp"
  $5 = (0 - ($4))|0; //@line 445 "modules/graphics/opengl/Polyline.cpp"
  $6 = (($colors) + ($i$01<<2)|0); //@line 445 "modules/graphics/opengl/Polyline.cpp"
  $7 = $5 << 24; //@line 445 "modules/graphics/opengl/Polyline.cpp"
  $8 = $7 | 16777215; //@line 445 "modules/graphics/opengl/Polyline.cpp"
  HEAP8[$6>>0]=$8&255;HEAP8[$6+1>>0]=($8>>8)&255;HEAP8[$6+2>>0]=($8>>16)&255;HEAP8[$6+3>>0]=$8>>24; //@line 445 "modules/graphics/opengl/Polyline.cpp"
  $9 = HEAP32[$0>>2]|0; //@line 442 "modules/graphics/opengl/Polyline.cpp"
  $10 = ($3>>>0)<($9>>>0); //@line 442 "modules/graphics/opengl/Polyline.cpp"
  if ($10) {
   $i$01 = $3;
  } else {
   break;
  }
 }
 return; //@line 447 "modules/graphics/opengl/Polyline.cpp"
}
function __ZN4love8graphics6opengl16NoneJoinPolylineD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (25084); //@line 356 "modules/graphics/opengl/Polyline.cpp"
 $0 = ((($this)) + 4|0); //@line 357 "modules/graphics/opengl/Polyline.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 357 "modules/graphics/opengl/Polyline.cpp"
 $2 = ($1|0)==(0|0); //@line 357 "modules/graphics/opengl/Polyline.cpp"
 if (!($2)) {
  __ZdaPv($1); //@line 358 "modules/graphics/opengl/Polyline.cpp"
 }
 __ZdlPv($this); //@line 109 "modules/graphics/opengl/Polyline.h"
 return; //@line 109 "modules/graphics/opengl/Polyline.h"
}
function __ZN4love8graphics6opengl16NoneJoinPolyline26calc_overdraw_vertex_countEb($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($this)) + 12|0); //@line 320 "modules/graphics/opengl/Polyline.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 320 "modules/graphics/opengl/Polyline.cpp"
 $3 = $2 << 2; //@line 320 "modules/graphics/opengl/Polyline.cpp"
 $4 = (($3) + -8)|0; //@line 320 "modules/graphics/opengl/Polyline.cpp"
 $5 = ((($this)) + 16|0); //@line 320 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$5>>2] = $4; //@line 320 "modules/graphics/opengl/Polyline.cpp"
 return; //@line 321 "modules/graphics/opengl/Polyline.cpp"
}
function __ZN4love8graphics6opengl16NoneJoinPolyline15render_overdrawERKNSt3__16vectorINS_6VectorENS3_9allocatorIS5_EEEEfb($this,$0,$pixel_size,$1) {
 $this = $this|0;
 $0 = $0|0;
 $pixel_size = +$pixel_size;
 $1 = $1|0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0.0;
 var $118 = 0.0, $119 = 0, $12 = 0.0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0.0, $130 = 0, $131 = 0.0, $132 = 0.0, $133 = 0, $134 = 0.0, $135 = 0.0;
 var $136 = 0.0, $137 = 0.0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0.0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0.0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0.0, $174 = 0.0, $175 = 0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0.0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0.0, $187 = 0.0, $188 = 0, $189 = 0.0, $19 = 0;
 var $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0;
 var $208 = 0, $209 = 0, $21 = 0.0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0.0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0;
 var $226 = 0, $227 = 0, $228 = 0, $229 = 0.0, $23 = 0, $230 = 0.0, $231 = 0, $232 = 0.0, $233 = 0.0, $234 = 0.0, $235 = 0.0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0.0, $240 = 0, $241 = 0, $242 = 0.0, $243 = 0.0;
 var $244 = 0, $245 = 0.0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0, $25 = 0.0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0.0, $260 = 0, $261 = 0;
 var $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0;
 var $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0.0, $63 = 0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0.0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$039 = 0, $s$sroa$0$0 = 0.0, $s$sroa$13$0 = 0.0, $t$sroa$0$0 = 0.0, $t$sroa$13$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($this)) + 12|0); //@line 325 "modules/graphics/opengl/Polyline.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 325 "modules/graphics/opengl/Polyline.cpp"
 $4 = ($3>>>0)>(5); //@line 325 "modules/graphics/opengl/Polyline.cpp"
 if (!($4)) {
  return; //@line 353 "modules/graphics/opengl/Polyline.cpp"
 }
 $5 = ((($this)) + 4|0); //@line 327 "modules/graphics/opengl/Polyline.cpp"
 $6 = ((($this)) + 8|0); //@line 333 "modules/graphics/opengl/Polyline.cpp"
 $11 = 5;$i$039 = 2;
 while(1) {
  $7 = HEAP32[$5>>2]|0; //@line 333 "modules/graphics/opengl/Polyline.cpp"
  $8 = (($7) + ($i$039<<3)|0); //@line 237 "./common/Vector.h"
  $9 = +HEAPF32[$8>>2]; //@line 237 "./common/Vector.h"
  $10 = (($7) + ($11<<3)|0); //@line 305 "./common/Vector.h"
  $12 = +HEAPF32[$10>>2]; //@line 305 "./common/Vector.h"
  $13 = $9 - $12; //@line 237 "./common/Vector.h"
  $14 = (((($7) + ($i$039<<3)|0)) + 4|0); //@line 237 "./common/Vector.h"
  $15 = +HEAPF32[$14>>2]; //@line 237 "./common/Vector.h"
  $16 = (((($7) + ($11<<3)|0)) + 4|0); //@line 310 "./common/Vector.h"
  $17 = +HEAPF32[$16>>2]; //@line 310 "./common/Vector.h"
  $18 = $15 - $17; //@line 237 "./common/Vector.h"
  $19 = $i$039 | 1; //@line 328 "modules/graphics/opengl/Polyline.cpp"
  $20 = (($7) + ($19<<3)|0); //@line 305 "./common/Vector.h"
  $21 = +HEAPF32[$20>>2]; //@line 305 "./common/Vector.h"
  $22 = $9 - $21; //@line 237 "./common/Vector.h"
  $23 = (((($7) + ($19<<3)|0)) + 4|0); //@line 310 "./common/Vector.h"
  $24 = +HEAPF32[$23>>2]; //@line 310 "./common/Vector.h"
  $25 = $15 - $24; //@line 237 "./common/Vector.h"
  $26 = $13 * $13; //@line 190 "./common/Vector.h"
  $27 = $18 * $18; //@line 190 "./common/Vector.h"
  $28 = $26 + $27; //@line 190 "./common/Vector.h"
  $29 = (+Math_sqrt((+$28))); //@line 190 "./common/Vector.h"
  $30 = $29 > 0.0; //@line 208 "./common/Vector.h"
  if ($30) {
   $31 = $pixel_size / $29; //@line 209 "./common/Vector.h"
   $32 = $13 * $31; //@line 269 "./common/Vector.h"
   $33 = $18 * $31; //@line 270 "./common/Vector.h"
   $s$sroa$0$0 = $32;$s$sroa$13$0 = $33;
  } else {
   $s$sroa$0$0 = $13;$s$sroa$13$0 = $18;
  }
  $34 = $22 * $22; //@line 190 "./common/Vector.h"
  $35 = $25 * $25; //@line 190 "./common/Vector.h"
  $36 = $34 + $35; //@line 190 "./common/Vector.h"
  $37 = (+Math_sqrt((+$36))); //@line 190 "./common/Vector.h"
  $38 = $37 > 0.0; //@line 208 "./common/Vector.h"
  if ($38) {
   $39 = $pixel_size / $37; //@line 209 "./common/Vector.h"
   $40 = $22 * $39; //@line 269 "./common/Vector.h"
   $41 = $25 * $39; //@line 270 "./common/Vector.h"
   $t$sroa$0$0 = $40;$t$sroa$13$0 = $41;
  } else {
   $t$sroa$0$0 = $22;$t$sroa$13$0 = $25;
  }
  $42 = $i$039 << 2; //@line 332 "modules/graphics/opengl/Polyline.cpp"
  $43 = (($42) + -8)|0; //@line 332 "modules/graphics/opengl/Polyline.cpp"
  $44 = HEAP32[$6>>2]|0; //@line 333 "modules/graphics/opengl/Polyline.cpp"
  $45 = (($44) + ($43<<3)|0); //@line 333 "modules/graphics/opengl/Polyline.cpp"
  $46 = (($7) + ($i$039<<3)|0); //@line 333 "modules/graphics/opengl/Polyline.cpp"
  $47 = $46; //@line 333 "modules/graphics/opengl/Polyline.cpp"
  $48 = $47; //@line 333 "modules/graphics/opengl/Polyline.cpp"
  $49 = HEAP32[$48>>2]|0; //@line 333 "modules/graphics/opengl/Polyline.cpp"
  $50 = (($47) + 4)|0; //@line 333 "modules/graphics/opengl/Polyline.cpp"
  $51 = $50; //@line 333 "modules/graphics/opengl/Polyline.cpp"
  $52 = HEAP32[$51>>2]|0; //@line 333 "modules/graphics/opengl/Polyline.cpp"
  $53 = $45; //@line 333 "modules/graphics/opengl/Polyline.cpp"
  $54 = $53; //@line 333 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$54>>2] = $49; //@line 333 "modules/graphics/opengl/Polyline.cpp"
  $55 = (($53) + 4)|0; //@line 333 "modules/graphics/opengl/Polyline.cpp"
  $56 = $55; //@line 333 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$56>>2] = $52; //@line 333 "modules/graphics/opengl/Polyline.cpp"
  $57 = (($42) + -7)|0; //@line 334 "modules/graphics/opengl/Polyline.cpp"
  $58 = HEAP32[$6>>2]|0; //@line 334 "modules/graphics/opengl/Polyline.cpp"
  $59 = HEAP32[$5>>2]|0; //@line 334 "modules/graphics/opengl/Polyline.cpp"
  $60 = (($59) + ($i$039<<3)|0); //@line 232 "./common/Vector.h"
  $61 = +HEAPF32[$60>>2]; //@line 232 "./common/Vector.h"
  $62 = $s$sroa$0$0 + $61; //@line 232 "./common/Vector.h"
  $63 = (((($59) + ($i$039<<3)|0)) + 4|0); //@line 232 "./common/Vector.h"
  $64 = +HEAPF32[$63>>2]; //@line 232 "./common/Vector.h"
  $65 = $s$sroa$13$0 + $64; //@line 232 "./common/Vector.h"
  $66 = $t$sroa$0$0 + $62; //@line 232 "./common/Vector.h"
  $67 = $t$sroa$13$0 + $65; //@line 232 "./common/Vector.h"
  $68 = (($58) + ($57<<3)|0); //@line 334 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$68>>2] = $66; //@line 334 "modules/graphics/opengl/Polyline.cpp"
  $69 = (((($58) + ($57<<3)|0)) + 4|0); //@line 334 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$69>>2] = $67; //@line 334 "modules/graphics/opengl/Polyline.cpp"
  $70 = (($42) + -6)|0; //@line 335 "modules/graphics/opengl/Polyline.cpp"
  $71 = HEAP32[$6>>2]|0; //@line 335 "modules/graphics/opengl/Polyline.cpp"
  $72 = HEAP32[$5>>2]|0; //@line 335 "modules/graphics/opengl/Polyline.cpp"
  $73 = (($72) + ($19<<3)|0); //@line 232 "./common/Vector.h"
  $74 = +HEAPF32[$73>>2]; //@line 232 "./common/Vector.h"
  $75 = $s$sroa$0$0 + $74; //@line 232 "./common/Vector.h"
  $76 = (((($72) + ($19<<3)|0)) + 4|0); //@line 232 "./common/Vector.h"
  $77 = +HEAPF32[$76>>2]; //@line 232 "./common/Vector.h"
  $78 = $s$sroa$13$0 + $77; //@line 232 "./common/Vector.h"
  $79 = $75 - $t$sroa$0$0; //@line 237 "./common/Vector.h"
  $80 = $78 - $t$sroa$13$0; //@line 237 "./common/Vector.h"
  $81 = (($71) + ($70<<3)|0); //@line 335 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$81>>2] = $79; //@line 335 "modules/graphics/opengl/Polyline.cpp"
  $82 = (((($71) + ($70<<3)|0)) + 4|0); //@line 335 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$82>>2] = $80; //@line 335 "modules/graphics/opengl/Polyline.cpp"
  $83 = (($42) + -5)|0; //@line 336 "modules/graphics/opengl/Polyline.cpp"
  $84 = HEAP32[$6>>2]|0; //@line 336 "modules/graphics/opengl/Polyline.cpp"
  $85 = (($84) + ($83<<3)|0); //@line 336 "modules/graphics/opengl/Polyline.cpp"
  $86 = HEAP32[$5>>2]|0; //@line 336 "modules/graphics/opengl/Polyline.cpp"
  $87 = (($86) + ($19<<3)|0); //@line 336 "modules/graphics/opengl/Polyline.cpp"
  $88 = $87; //@line 336 "modules/graphics/opengl/Polyline.cpp"
  $89 = $88; //@line 336 "modules/graphics/opengl/Polyline.cpp"
  $90 = HEAP32[$89>>2]|0; //@line 336 "modules/graphics/opengl/Polyline.cpp"
  $91 = (($88) + 4)|0; //@line 336 "modules/graphics/opengl/Polyline.cpp"
  $92 = $91; //@line 336 "modules/graphics/opengl/Polyline.cpp"
  $93 = HEAP32[$92>>2]|0; //@line 336 "modules/graphics/opengl/Polyline.cpp"
  $94 = $85; //@line 336 "modules/graphics/opengl/Polyline.cpp"
  $95 = $94; //@line 336 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$95>>2] = $90; //@line 336 "modules/graphics/opengl/Polyline.cpp"
  $96 = (($94) + 4)|0; //@line 336 "modules/graphics/opengl/Polyline.cpp"
  $97 = $96; //@line 336 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$97>>2] = $93; //@line 336 "modules/graphics/opengl/Polyline.cpp"
  $98 = (($42) + -4)|0; //@line 338 "modules/graphics/opengl/Polyline.cpp"
  $99 = HEAP32[$6>>2]|0; //@line 338 "modules/graphics/opengl/Polyline.cpp"
  $100 = (($99) + ($98<<3)|0); //@line 338 "modules/graphics/opengl/Polyline.cpp"
  $101 = HEAP32[$5>>2]|0; //@line 338 "modules/graphics/opengl/Polyline.cpp"
  $102 = (($101) + ($19<<3)|0); //@line 338 "modules/graphics/opengl/Polyline.cpp"
  $103 = $102; //@line 338 "modules/graphics/opengl/Polyline.cpp"
  $104 = $103; //@line 338 "modules/graphics/opengl/Polyline.cpp"
  $105 = HEAP32[$104>>2]|0; //@line 338 "modules/graphics/opengl/Polyline.cpp"
  $106 = (($103) + 4)|0; //@line 338 "modules/graphics/opengl/Polyline.cpp"
  $107 = $106; //@line 338 "modules/graphics/opengl/Polyline.cpp"
  $108 = HEAP32[$107>>2]|0; //@line 338 "modules/graphics/opengl/Polyline.cpp"
  $109 = $100; //@line 338 "modules/graphics/opengl/Polyline.cpp"
  $110 = $109; //@line 338 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$110>>2] = $105; //@line 338 "modules/graphics/opengl/Polyline.cpp"
  $111 = (($109) + 4)|0; //@line 338 "modules/graphics/opengl/Polyline.cpp"
  $112 = $111; //@line 338 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$112>>2] = $108; //@line 338 "modules/graphics/opengl/Polyline.cpp"
  $113 = (($42) + -3)|0; //@line 339 "modules/graphics/opengl/Polyline.cpp"
  $114 = HEAP32[$6>>2]|0; //@line 339 "modules/graphics/opengl/Polyline.cpp"
  $115 = HEAP32[$5>>2]|0; //@line 339 "modules/graphics/opengl/Polyline.cpp"
  $116 = (($115) + ($19<<3)|0); //@line 232 "./common/Vector.h"
  $117 = +HEAPF32[$116>>2]; //@line 232 "./common/Vector.h"
  $118 = $s$sroa$0$0 + $117; //@line 232 "./common/Vector.h"
  $119 = (((($115) + ($19<<3)|0)) + 4|0); //@line 232 "./common/Vector.h"
  $120 = +HEAPF32[$119>>2]; //@line 232 "./common/Vector.h"
  $121 = $s$sroa$13$0 + $120; //@line 232 "./common/Vector.h"
  $122 = $118 - $t$sroa$0$0; //@line 237 "./common/Vector.h"
  $123 = $121 - $t$sroa$13$0; //@line 237 "./common/Vector.h"
  $124 = (($114) + ($113<<3)|0); //@line 339 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$124>>2] = $122; //@line 339 "modules/graphics/opengl/Polyline.cpp"
  $125 = (((($114) + ($113<<3)|0)) + 4|0); //@line 339 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$125>>2] = $123; //@line 339 "modules/graphics/opengl/Polyline.cpp"
  $126 = (($42) + -2)|0; //@line 340 "modules/graphics/opengl/Polyline.cpp"
  $127 = HEAP32[$6>>2]|0; //@line 340 "modules/graphics/opengl/Polyline.cpp"
  $128 = (($i$039) + 2)|0; //@line 340 "modules/graphics/opengl/Polyline.cpp"
  $129 = HEAP32[$5>>2]|0; //@line 340 "modules/graphics/opengl/Polyline.cpp"
  $130 = (($129) + ($128<<3)|0); //@line 237 "./common/Vector.h"
  $131 = +HEAPF32[$130>>2]; //@line 237 "./common/Vector.h"
  $132 = $131 - $s$sroa$0$0; //@line 237 "./common/Vector.h"
  $133 = (((($129) + ($128<<3)|0)) + 4|0); //@line 237 "./common/Vector.h"
  $134 = +HEAPF32[$133>>2]; //@line 237 "./common/Vector.h"
  $135 = $134 - $s$sroa$13$0; //@line 237 "./common/Vector.h"
  $136 = $132 - $t$sroa$0$0; //@line 237 "./common/Vector.h"
  $137 = $135 - $t$sroa$13$0; //@line 237 "./common/Vector.h"
  $138 = (($127) + ($126<<3)|0); //@line 340 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$138>>2] = $136; //@line 340 "modules/graphics/opengl/Polyline.cpp"
  $139 = (((($127) + ($126<<3)|0)) + 4|0); //@line 340 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$139>>2] = $137; //@line 340 "modules/graphics/opengl/Polyline.cpp"
  $140 = (($42) + -1)|0; //@line 341 "modules/graphics/opengl/Polyline.cpp"
  $141 = HEAP32[$6>>2]|0; //@line 341 "modules/graphics/opengl/Polyline.cpp"
  $142 = (($141) + ($140<<3)|0); //@line 341 "modules/graphics/opengl/Polyline.cpp"
  $143 = HEAP32[$5>>2]|0; //@line 341 "modules/graphics/opengl/Polyline.cpp"
  $144 = (($143) + ($128<<3)|0); //@line 341 "modules/graphics/opengl/Polyline.cpp"
  $145 = $144; //@line 341 "modules/graphics/opengl/Polyline.cpp"
  $146 = $145; //@line 341 "modules/graphics/opengl/Polyline.cpp"
  $147 = HEAP32[$146>>2]|0; //@line 341 "modules/graphics/opengl/Polyline.cpp"
  $148 = (($145) + 4)|0; //@line 341 "modules/graphics/opengl/Polyline.cpp"
  $149 = $148; //@line 341 "modules/graphics/opengl/Polyline.cpp"
  $150 = HEAP32[$149>>2]|0; //@line 341 "modules/graphics/opengl/Polyline.cpp"
  $151 = $142; //@line 341 "modules/graphics/opengl/Polyline.cpp"
  $152 = $151; //@line 341 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$152>>2] = $147; //@line 341 "modules/graphics/opengl/Polyline.cpp"
  $153 = (($151) + 4)|0; //@line 341 "modules/graphics/opengl/Polyline.cpp"
  $154 = $153; //@line 341 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$154>>2] = $150; //@line 341 "modules/graphics/opengl/Polyline.cpp"
  $155 = HEAP32[$6>>2]|0; //@line 343 "modules/graphics/opengl/Polyline.cpp"
  $156 = (($155) + ($42<<3)|0); //@line 343 "modules/graphics/opengl/Polyline.cpp"
  $157 = HEAP32[$5>>2]|0; //@line 343 "modules/graphics/opengl/Polyline.cpp"
  $158 = (($157) + ($128<<3)|0); //@line 343 "modules/graphics/opengl/Polyline.cpp"
  $159 = $158; //@line 343 "modules/graphics/opengl/Polyline.cpp"
  $160 = $159; //@line 343 "modules/graphics/opengl/Polyline.cpp"
  $161 = HEAP32[$160>>2]|0; //@line 343 "modules/graphics/opengl/Polyline.cpp"
  $162 = (($159) + 4)|0; //@line 343 "modules/graphics/opengl/Polyline.cpp"
  $163 = $162; //@line 343 "modules/graphics/opengl/Polyline.cpp"
  $164 = HEAP32[$163>>2]|0; //@line 343 "modules/graphics/opengl/Polyline.cpp"
  $165 = $156; //@line 343 "modules/graphics/opengl/Polyline.cpp"
  $166 = $165; //@line 343 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$166>>2] = $161; //@line 343 "modules/graphics/opengl/Polyline.cpp"
  $167 = (($165) + 4)|0; //@line 343 "modules/graphics/opengl/Polyline.cpp"
  $168 = $167; //@line 343 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$168>>2] = $164; //@line 343 "modules/graphics/opengl/Polyline.cpp"
  $169 = $42 | 1; //@line 344 "modules/graphics/opengl/Polyline.cpp"
  $170 = HEAP32[$6>>2]|0; //@line 344 "modules/graphics/opengl/Polyline.cpp"
  $171 = HEAP32[$5>>2]|0; //@line 344 "modules/graphics/opengl/Polyline.cpp"
  $172 = (($171) + ($128<<3)|0); //@line 237 "./common/Vector.h"
  $173 = +HEAPF32[$172>>2]; //@line 237 "./common/Vector.h"
  $174 = $173 - $s$sroa$0$0; //@line 237 "./common/Vector.h"
  $175 = (((($171) + ($128<<3)|0)) + 4|0); //@line 237 "./common/Vector.h"
  $176 = +HEAPF32[$175>>2]; //@line 237 "./common/Vector.h"
  $177 = $176 - $s$sroa$13$0; //@line 237 "./common/Vector.h"
  $178 = $174 - $t$sroa$0$0; //@line 237 "./common/Vector.h"
  $179 = $177 - $t$sroa$13$0; //@line 237 "./common/Vector.h"
  $180 = (($170) + ($169<<3)|0); //@line 344 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$180>>2] = $178; //@line 344 "modules/graphics/opengl/Polyline.cpp"
  $181 = (((($170) + ($169<<3)|0)) + 4|0); //@line 344 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$181>>2] = $179; //@line 344 "modules/graphics/opengl/Polyline.cpp"
  $182 = $42 | 2; //@line 345 "modules/graphics/opengl/Polyline.cpp"
  $183 = HEAP32[$6>>2]|0; //@line 345 "modules/graphics/opengl/Polyline.cpp"
  $184 = HEAP32[$5>>2]|0; //@line 345 "modules/graphics/opengl/Polyline.cpp"
  $185 = (($184) + ($11<<3)|0); //@line 237 "./common/Vector.h"
  $186 = +HEAPF32[$185>>2]; //@line 237 "./common/Vector.h"
  $187 = $186 - $s$sroa$0$0; //@line 237 "./common/Vector.h"
  $188 = (((($184) + ($11<<3)|0)) + 4|0); //@line 237 "./common/Vector.h"
  $189 = +HEAPF32[$188>>2]; //@line 237 "./common/Vector.h"
  $190 = $189 - $s$sroa$13$0; //@line 237 "./common/Vector.h"
  $191 = $t$sroa$0$0 + $187; //@line 232 "./common/Vector.h"
  $192 = $t$sroa$13$0 + $190; //@line 232 "./common/Vector.h"
  $193 = (($183) + ($182<<3)|0); //@line 345 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$193>>2] = $191; //@line 345 "modules/graphics/opengl/Polyline.cpp"
  $194 = (((($183) + ($182<<3)|0)) + 4|0); //@line 345 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$194>>2] = $192; //@line 345 "modules/graphics/opengl/Polyline.cpp"
  $195 = $42 | 3; //@line 346 "modules/graphics/opengl/Polyline.cpp"
  $196 = HEAP32[$6>>2]|0; //@line 346 "modules/graphics/opengl/Polyline.cpp"
  $197 = (($196) + ($195<<3)|0); //@line 346 "modules/graphics/opengl/Polyline.cpp"
  $198 = HEAP32[$5>>2]|0; //@line 346 "modules/graphics/opengl/Polyline.cpp"
  $199 = (($198) + ($11<<3)|0); //@line 346 "modules/graphics/opengl/Polyline.cpp"
  $200 = $199; //@line 346 "modules/graphics/opengl/Polyline.cpp"
  $201 = $200; //@line 346 "modules/graphics/opengl/Polyline.cpp"
  $202 = HEAP32[$201>>2]|0; //@line 346 "modules/graphics/opengl/Polyline.cpp"
  $203 = (($200) + 4)|0; //@line 346 "modules/graphics/opengl/Polyline.cpp"
  $204 = $203; //@line 346 "modules/graphics/opengl/Polyline.cpp"
  $205 = HEAP32[$204>>2]|0; //@line 346 "modules/graphics/opengl/Polyline.cpp"
  $206 = $197; //@line 346 "modules/graphics/opengl/Polyline.cpp"
  $207 = $206; //@line 346 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$207>>2] = $202; //@line 346 "modules/graphics/opengl/Polyline.cpp"
  $208 = (($206) + 4)|0; //@line 346 "modules/graphics/opengl/Polyline.cpp"
  $209 = $208; //@line 346 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$209>>2] = $205; //@line 346 "modules/graphics/opengl/Polyline.cpp"
  $210 = $42 | 4; //@line 348 "modules/graphics/opengl/Polyline.cpp"
  $211 = HEAP32[$6>>2]|0; //@line 348 "modules/graphics/opengl/Polyline.cpp"
  $212 = (($211) + ($210<<3)|0); //@line 348 "modules/graphics/opengl/Polyline.cpp"
  $213 = HEAP32[$5>>2]|0; //@line 348 "modules/graphics/opengl/Polyline.cpp"
  $214 = (($213) + ($11<<3)|0); //@line 348 "modules/graphics/opengl/Polyline.cpp"
  $215 = $214; //@line 348 "modules/graphics/opengl/Polyline.cpp"
  $216 = $215; //@line 348 "modules/graphics/opengl/Polyline.cpp"
  $217 = HEAP32[$216>>2]|0; //@line 348 "modules/graphics/opengl/Polyline.cpp"
  $218 = (($215) + 4)|0; //@line 348 "modules/graphics/opengl/Polyline.cpp"
  $219 = $218; //@line 348 "modules/graphics/opengl/Polyline.cpp"
  $220 = HEAP32[$219>>2]|0; //@line 348 "modules/graphics/opengl/Polyline.cpp"
  $221 = $212; //@line 348 "modules/graphics/opengl/Polyline.cpp"
  $222 = $221; //@line 348 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$222>>2] = $217; //@line 348 "modules/graphics/opengl/Polyline.cpp"
  $223 = (($221) + 4)|0; //@line 348 "modules/graphics/opengl/Polyline.cpp"
  $224 = $223; //@line 348 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$224>>2] = $220; //@line 348 "modules/graphics/opengl/Polyline.cpp"
  $225 = $42 | 5; //@line 349 "modules/graphics/opengl/Polyline.cpp"
  $226 = HEAP32[$6>>2]|0; //@line 349 "modules/graphics/opengl/Polyline.cpp"
  $227 = HEAP32[$5>>2]|0; //@line 349 "modules/graphics/opengl/Polyline.cpp"
  $228 = (($227) + ($11<<3)|0); //@line 237 "./common/Vector.h"
  $229 = +HEAPF32[$228>>2]; //@line 237 "./common/Vector.h"
  $230 = $229 - $s$sroa$0$0; //@line 237 "./common/Vector.h"
  $231 = (((($227) + ($11<<3)|0)) + 4|0); //@line 237 "./common/Vector.h"
  $232 = +HEAPF32[$231>>2]; //@line 237 "./common/Vector.h"
  $233 = $232 - $s$sroa$13$0; //@line 237 "./common/Vector.h"
  $234 = $t$sroa$0$0 + $230; //@line 232 "./common/Vector.h"
  $235 = $t$sroa$13$0 + $233; //@line 232 "./common/Vector.h"
  $236 = (($226) + ($225<<3)|0); //@line 349 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$236>>2] = $234; //@line 349 "modules/graphics/opengl/Polyline.cpp"
  $237 = (((($226) + ($225<<3)|0)) + 4|0); //@line 349 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$237>>2] = $235; //@line 349 "modules/graphics/opengl/Polyline.cpp"
  $238 = $42 | 6; //@line 350 "modules/graphics/opengl/Polyline.cpp"
  $239 = HEAP32[$6>>2]|0; //@line 350 "modules/graphics/opengl/Polyline.cpp"
  $240 = HEAP32[$5>>2]|0; //@line 350 "modules/graphics/opengl/Polyline.cpp"
  $241 = (($240) + ($i$039<<3)|0); //@line 232 "./common/Vector.h"
  $242 = +HEAPF32[$241>>2]; //@line 232 "./common/Vector.h"
  $243 = $s$sroa$0$0 + $242; //@line 232 "./common/Vector.h"
  $244 = (((($240) + ($i$039<<3)|0)) + 4|0); //@line 232 "./common/Vector.h"
  $245 = +HEAPF32[$244>>2]; //@line 232 "./common/Vector.h"
  $246 = $s$sroa$13$0 + $245; //@line 232 "./common/Vector.h"
  $247 = $t$sroa$0$0 + $243; //@line 232 "./common/Vector.h"
  $248 = $t$sroa$13$0 + $246; //@line 232 "./common/Vector.h"
  $249 = (($239) + ($238<<3)|0); //@line 350 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$249>>2] = $247; //@line 350 "modules/graphics/opengl/Polyline.cpp"
  $250 = (((($239) + ($238<<3)|0)) + 4|0); //@line 350 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$250>>2] = $248; //@line 350 "modules/graphics/opengl/Polyline.cpp"
  $251 = $42 | 7; //@line 351 "modules/graphics/opengl/Polyline.cpp"
  $252 = HEAP32[$6>>2]|0; //@line 351 "modules/graphics/opengl/Polyline.cpp"
  $253 = (($252) + ($251<<3)|0); //@line 351 "modules/graphics/opengl/Polyline.cpp"
  $254 = HEAP32[$5>>2]|0; //@line 351 "modules/graphics/opengl/Polyline.cpp"
  $255 = (($254) + ($i$039<<3)|0); //@line 351 "modules/graphics/opengl/Polyline.cpp"
  $256 = $255; //@line 351 "modules/graphics/opengl/Polyline.cpp"
  $257 = $256; //@line 351 "modules/graphics/opengl/Polyline.cpp"
  $258 = HEAP32[$257>>2]|0; //@line 351 "modules/graphics/opengl/Polyline.cpp"
  $259 = (($256) + 4)|0; //@line 351 "modules/graphics/opengl/Polyline.cpp"
  $260 = $259; //@line 351 "modules/graphics/opengl/Polyline.cpp"
  $261 = HEAP32[$260>>2]|0; //@line 351 "modules/graphics/opengl/Polyline.cpp"
  $262 = $253; //@line 351 "modules/graphics/opengl/Polyline.cpp"
  $263 = $262; //@line 351 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$263>>2] = $258; //@line 351 "modules/graphics/opengl/Polyline.cpp"
  $264 = (($262) + 4)|0; //@line 351 "modules/graphics/opengl/Polyline.cpp"
  $265 = $264; //@line 351 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$265>>2] = $261; //@line 351 "modules/graphics/opengl/Polyline.cpp"
  $266 = (($i$039) + 4)|0; //@line 325 "modules/graphics/opengl/Polyline.cpp"
  $267 = (($i$039) + 7)|0; //@line 325 "modules/graphics/opengl/Polyline.cpp"
  $268 = HEAP32[$2>>2]|0; //@line 325 "modules/graphics/opengl/Polyline.cpp"
  $269 = ($267>>>0)<($268>>>0); //@line 325 "modules/graphics/opengl/Polyline.cpp"
  if ($269) {
   $11 = $267;$i$039 = $266;
  } else {
   break;
  }
 }
 return; //@line 353 "modules/graphics/opengl/Polyline.cpp"
}
function __ZN4love8graphics6opengl16NoneJoinPolyline16fill_color_arrayEPNS0_6ColorTIhEE($this,$colors) {
 $this = $this|0;
 $colors = $colors|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 451 "modules/graphics/opengl/Polyline.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 451 "modules/graphics/opengl/Polyline.cpp"
 $2 = ($1|0)==(0); //@line 451 "modules/graphics/opengl/Polyline.cpp"
 if ($2) {
  return; //@line 456 "modules/graphics/opengl/Polyline.cpp"
 } else {
  $i$01 = 0;
 }
 while(1) {
  $3 = (($i$01) + 1)|0; //@line 454 "modules/graphics/opengl/Polyline.cpp"
  $4 = $3 & 2; //@line 454 "modules/graphics/opengl/Polyline.cpp"
  $5 = ($4>>>0)<(2); //@line 454 "modules/graphics/opengl/Polyline.cpp"
  $6 = (($colors) + ($i$01<<2)|0); //@line 454 "modules/graphics/opengl/Polyline.cpp"
  $7 = $5 ? -1 : 16777215; //@line 454 "modules/graphics/opengl/Polyline.cpp"
  HEAP8[$6>>0]=$7&255;HEAP8[$6+1>>0]=($7>>8)&255;HEAP8[$6+2>>0]=($7>>16)&255;HEAP8[$6+3>>0]=$7>>24; //@line 454 "modules/graphics/opengl/Polyline.cpp"
  $8 = HEAP32[$0>>2]|0; //@line 451 "modules/graphics/opengl/Polyline.cpp"
  $9 = ($3>>>0)<($8>>>0); //@line 451 "modules/graphics/opengl/Polyline.cpp"
  if ($9) {
   $i$01 = $3;
  } else {
   break;
  }
 }
 return; //@line 456 "modules/graphics/opengl/Polyline.cpp"
}
function __ZN4love8graphics6opengl16NoneJoinPolyline10renderEdgeERNSt3__16vectorINS_6VectorENS3_9allocatorIS5_EEEES9_RS5_RfSA_RKS5_SD_f($this,$anchors,$normals,$s,$len_s,$ns,$q,$r,$hw) {
 $this = $this|0;
 $anchors = $anchors|0;
 $normals = $normals|0;
 $s = $s|0;
 $len_s = $len_s|0;
 $ns = $ns|0;
 $q = $q|0;
 $r = $r|0;
 $hw = +$hw;
 var $$pre = 0, $$pre11 = 0, $$pre12 = 0, $$pre13 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0;
 var $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 8|0;
 $1 = sp;
 $2 = ((($anchors)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ((($anchors)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = HEAP32[$4>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($3|0)==($5|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($anchors,$q); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$2>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = $$pre;
 } else {
  $7 = $q; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $8 = $7; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $9 = HEAP32[$8>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $10 = (($7) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $11 = $10; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $12 = HEAP32[$11>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $13 = $3; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $14 = $13; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$14>>2] = $9; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $15 = (($13) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $16 = $15; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$16>>2] = $12; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $17 = HEAP32[$2>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = ((($17)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$2>>2] = $18; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = $18;
 }
 $19 = HEAP32[$4>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = ($21|0)==($19|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($20) {
  __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($anchors,$q); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 } else {
  $22 = $q; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = $22; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = HEAP32[$23>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $25 = (($22) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = $25; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $27 = HEAP32[$26>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $28 = $21; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $29 = $28; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$29>>2] = $24; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $30 = (($28) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $31 = $30; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$31>>2] = $27; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $32 = HEAP32[$2>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $33 = ((($32)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$2>>2] = $33; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $34 = ((($normals)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $35 = HEAP32[$34>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $36 = ((($normals)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = HEAP32[$36>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $38 = ($35|0)==($37|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($38) {
  __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($normals,$ns); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre11 = HEAP32[$34>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $59 = $$pre11;
 } else {
  $39 = $ns; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $40 = $39; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $41 = HEAP32[$40>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $42 = (($39) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $43 = $42; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $44 = HEAP32[$43>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $45 = $35; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $46 = $45; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$46>>2] = $41; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $47 = (($45) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $48 = $47; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$48>>2] = $44; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $49 = HEAP32[$34>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $50 = ((($49)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$34>>2] = $50; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $59 = $50;
 }
 $51 = +HEAPF32[$ns>>2]; //@line 252 "./common/Vector.h"
 $52 = -$51; //@line 252 "./common/Vector.h"
 $53 = ((($ns)) + 4|0); //@line 252 "./common/Vector.h"
 $54 = +HEAPF32[$53>>2]; //@line 252 "./common/Vector.h"
 $55 = -$54; //@line 252 "./common/Vector.h"
 HEAPF32[$0>>2] = $52; //@line 225 "./common/Vector.h"
 $56 = ((($0)) + 4|0); //@line 226 "./common/Vector.h"
 HEAPF32[$56>>2] = $55; //@line 226 "./common/Vector.h"
 $57 = HEAP32[$36>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $58 = ($59>>>0)<($57>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($58) {
  $60 = $0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $61 = $60; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $62 = HEAP32[$61>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $63 = (($60) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $64 = $63; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $65 = HEAP32[$64>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $66 = $59; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $67 = $66; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$67>>2] = $62; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $68 = (($66) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $69 = $68; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$69>>2] = $65; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $70 = HEAP32[$34>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $71 = ((($70)) + 8|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$34>>2] = $71; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 } else {
  __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIS2_EEvOT_($normals,$0); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $72 = +HEAPF32[$r>>2]; //@line 237 "./common/Vector.h"
 $73 = +HEAPF32[$q>>2]; //@line 305 "./common/Vector.h"
 $74 = $72 - $73; //@line 237 "./common/Vector.h"
 $75 = ((($r)) + 4|0); //@line 237 "./common/Vector.h"
 $76 = +HEAPF32[$75>>2]; //@line 237 "./common/Vector.h"
 $77 = ((($q)) + 4|0); //@line 310 "./common/Vector.h"
 $78 = +HEAPF32[$77>>2]; //@line 310 "./common/Vector.h"
 $79 = $76 - $78; //@line 237 "./common/Vector.h"
 HEAPF32[$s>>2] = $74; //@line 123 "modules/graphics/opengl/Polyline.cpp"
 $80 = ((($s)) + 4|0); //@line 123 "modules/graphics/opengl/Polyline.cpp"
 HEAPF32[$80>>2] = $79; //@line 123 "modules/graphics/opengl/Polyline.cpp"
 $81 = $74 * $74; //@line 190 "./common/Vector.h"
 $82 = $79 * $79; //@line 190 "./common/Vector.h"
 $83 = $81 + $82; //@line 190 "./common/Vector.h"
 $84 = (+Math_sqrt((+$83))); //@line 190 "./common/Vector.h"
 HEAPF32[$len_s>>2] = $84; //@line 124 "modules/graphics/opengl/Polyline.cpp"
 $85 = $hw / $84; //@line 125 "modules/graphics/opengl/Polyline.cpp"
 $86 = +HEAPF32[$80>>2]; //@line 200 "./common/Vector.h"
 $87 = $85 * $86; //@line 200 "./common/Vector.h"
 $88 = -$87; //@line 200 "./common/Vector.h"
 $89 = +HEAPF32[$s>>2]; //@line 200 "./common/Vector.h"
 $90 = $85 * $89; //@line 200 "./common/Vector.h"
 HEAPF32[$ns>>2] = $88; //@line 125 "modules/graphics/opengl/Polyline.cpp"
 HEAPF32[$53>>2] = $90; //@line 125 "modules/graphics/opengl/Polyline.cpp"
 $91 = HEAP32[$2>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $92 = HEAP32[$4>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $93 = ($91|0)==($92|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($93) {
  __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($anchors,$q); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre12 = HEAP32[$2>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $108 = $$pre12;
 } else {
  $94 = $q; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $95 = $94; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $96 = HEAP32[$95>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $97 = (($94) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $98 = $97; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $99 = HEAP32[$98>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $100 = $91; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $101 = $100; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$101>>2] = $96; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $102 = (($100) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $103 = $102; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$103>>2] = $99; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $104 = HEAP32[$2>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $105 = ((($104)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$2>>2] = $105; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $108 = $105;
 }
 $106 = HEAP32[$4>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $107 = ($108|0)==($106|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($107) {
  __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($anchors,$q); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 } else {
  $109 = $q; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $110 = $109; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $111 = HEAP32[$110>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $112 = (($109) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $113 = $112; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $114 = HEAP32[$113>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $115 = $108; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $116 = $115; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$116>>2] = $111; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $117 = (($115) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $118 = $117; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$118>>2] = $114; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $119 = HEAP32[$2>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $120 = ((($119)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$2>>2] = $120; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $121 = +HEAPF32[$ns>>2]; //@line 252 "./common/Vector.h"
 $122 = -$121; //@line 252 "./common/Vector.h"
 $123 = +HEAPF32[$53>>2]; //@line 252 "./common/Vector.h"
 $124 = -$123; //@line 252 "./common/Vector.h"
 HEAPF32[$1>>2] = $122; //@line 225 "./common/Vector.h"
 $125 = ((($1)) + 4|0); //@line 226 "./common/Vector.h"
 HEAPF32[$125>>2] = $124; //@line 226 "./common/Vector.h"
 $126 = HEAP32[$34>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $127 = HEAP32[$36>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $128 = ($126>>>0)<($127>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($128) {
  $129 = $1; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $130 = $129; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $131 = HEAP32[$130>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $132 = (($129) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $133 = $132; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $134 = HEAP32[$133>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $135 = $126; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $136 = $135; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$136>>2] = $131; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $137 = (($135) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $138 = $137; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$138>>2] = $134; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $139 = HEAP32[$34>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $140 = ((($139)) + 8|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$34>>2] = $140; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $143 = $140;
 } else {
  __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIS2_EEvOT_($normals,$1); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre13 = HEAP32[$34>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $143 = $$pre13;
 }
 $141 = HEAP32[$36>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $142 = ($143|0)==($141|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($142) {
  __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($normals,$ns); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  STACKTOP = sp;return; //@line 131 "modules/graphics/opengl/Polyline.cpp"
 } else {
  $144 = $ns; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $145 = $144; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $146 = HEAP32[$145>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $147 = (($144) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $148 = $147; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $149 = HEAP32[$148>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $150 = $143; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $151 = $150; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$151>>2] = $146; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $152 = (($150) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $153 = $152; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$153>>2] = $149; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $154 = HEAP32[$34>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $155 = ((($154)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$34>>2] = $155; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  STACKTOP = sp;return; //@line 131 "modules/graphics/opengl/Polyline.cpp"
 }
}
function __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIS2_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(536870911); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 3; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(268435455); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 2; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$42 = $18;
  } else {
   $$0$i1 = $17;$50 = $18;$51 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 536870911;$50 = $12;$51 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 3; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $51;$42 = $50;
 }
 $24 = (($25) + ($26<<3)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = $__x; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $31 = HEAP32[$30>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $32 = (($29) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = $32; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = HEAP32[$33>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = $24; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = $35; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$36>>2] = $31; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = (($35) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $37; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$38>>2] = $34; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $39 = ((($24)) + 8|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $40 = $39; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $41 = (($42) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $43 = $41 >> 3; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $44 = (0 - ($43))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $45 = (($24) + ($44<<3)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $46 = $45; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($45|0),($47|0),($41|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $46; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $40; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $48 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($48) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $49 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($49); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl17MiterJoinPolylineD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (25084); //@line 356 "modules/graphics/opengl/Polyline.cpp"
 $0 = ((($this)) + 4|0); //@line 357 "modules/graphics/opengl/Polyline.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 357 "modules/graphics/opengl/Polyline.cpp"
 $2 = ($1|0)==(0|0); //@line 357 "modules/graphics/opengl/Polyline.cpp"
 if (!($2)) {
  __ZdaPv($1); //@line 358 "modules/graphics/opengl/Polyline.cpp"
 }
 __ZdlPv($this); //@line 147 "modules/graphics/opengl/Polyline.h"
 return; //@line 147 "modules/graphics/opengl/Polyline.h"
}
function __ZN4love8graphics6opengl17MiterJoinPolyline10renderEdgeERNSt3__16vectorINS_6VectorENS3_9allocatorIS5_EEEES9_RS5_RfSA_RKS5_SD_f($this,$anchors,$normals,$s,$len_s,$ns,$q,$r,$hw) {
 $this = $this|0;
 $anchors = $anchors|0;
 $normals = $normals|0;
 $s = $s|0;
 $len_s = $len_s|0;
 $ns = $ns|0;
 $q = $q|0;
 $r = $r|0;
 $hw = +$hw;
 var $$pre = 0, $$pre$phi12Z2D = 0, $$pre$phiZ2D = 0, $$pre10 = 0, $$pre11 = 0, $0 = 0, $1 = 0, $10 = 0.0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0.0, $105 = 0.0, $106 = 0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0.0, $110 = 0.0;
 var $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0, $118 = 0, $119 = 0, $12 = 0.0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0.0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0, $14 = 0.0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0.0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0.0, $60 = 0.0, $61 = 0, $62 = 0.0;
 var $63 = 0.0, $64 = 0.0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0.0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0.0, $84 = 0.0, $85 = 0, $86 = 0.0, $87 = 0.0, $88 = 0, $89 = 0, $9 = 0.0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $d = 0, $fabsf = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 16|0;
 $d = sp + 8|0;
 $1 = sp;
 $2 = +HEAPF32[$r>>2]; //@line 237 "./common/Vector.h"
 $3 = +HEAPF32[$q>>2]; //@line 305 "./common/Vector.h"
 $4 = $2 - $3; //@line 237 "./common/Vector.h"
 $5 = ((($r)) + 4|0); //@line 237 "./common/Vector.h"
 $6 = +HEAPF32[$5>>2]; //@line 237 "./common/Vector.h"
 $7 = ((($q)) + 4|0); //@line 310 "./common/Vector.h"
 $8 = +HEAPF32[$7>>2]; //@line 310 "./common/Vector.h"
 $9 = $6 - $8; //@line 237 "./common/Vector.h"
 $10 = $4 * $4; //@line 190 "./common/Vector.h"
 $11 = $9 * $9; //@line 190 "./common/Vector.h"
 $12 = $10 + $11; //@line 190 "./common/Vector.h"
 $13 = (+Math_sqrt((+$12))); //@line 190 "./common/Vector.h"
 $14 = $hw / $13; //@line 177 "modules/graphics/opengl/Polyline.cpp"
 $15 = $9 * $14; //@line 200 "./common/Vector.h"
 $16 = -$15; //@line 200 "./common/Vector.h"
 $17 = $4 * $14; //@line 200 "./common/Vector.h"
 $18 = ((($anchors)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $19 = HEAP32[$18>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = ((($anchors)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $21 = HEAP32[$20>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $22 = ($19|0)==($21|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($22) {
  __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($anchors,$q); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$18>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $37 = $$pre;
 } else {
  $23 = $q; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = $23; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $25 = HEAP32[$24>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (($23) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $27 = $26; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $28 = HEAP32[$27>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $29 = $19; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $30 = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$30>>2] = $25; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $31 = (($29) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $32 = $31; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$32>>2] = $28; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $33 = HEAP32[$18>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $34 = ((($33)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$18>>2] = $34; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $37 = $34;
 }
 $35 = HEAP32[$20>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $36 = ($37|0)==($35|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($36) {
  __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($anchors,$q); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 } else {
  $38 = $q; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $39 = $38; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $40 = HEAP32[$39>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $41 = (($38) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $42 = $41; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $43 = HEAP32[$42>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $44 = $37; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $45 = $44; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$45>>2] = $40; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $46 = (($44) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $47 = $46; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$47>>2] = $43; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $48 = HEAP32[$18>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $49 = ((($48)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$18>>2] = $49; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $50 = +HEAPF32[$s>>2]; //@line 242 "./common/Vector.h"
 $51 = $9 * $50; //@line 286 "./common/Vector.h"
 $52 = ((($s)) + 4|0); //@line 286 "./common/Vector.h"
 $53 = +HEAPF32[$52>>2]; //@line 242 "./common/Vector.h"
 $54 = $4 * $53; //@line 286 "./common/Vector.h"
 $55 = $51 - $54; //@line 286 "./common/Vector.h"
 $fabsf = (+Math_abs((+$55))); //@line 183 "modules/graphics/opengl/Polyline.cpp"
 $56 = $fabsf; //@line 183 "modules/graphics/opengl/Polyline.cpp"
 $57 = +HEAPF32[$len_s>>2]; //@line 183 "modules/graphics/opengl/Polyline.cpp"
 $58 = $13 * $57; //@line 183 "modules/graphics/opengl/Polyline.cpp"
 $59 = $58; //@line 183 "modules/graphics/opengl/Polyline.cpp"
 $60 = $56 / $59; //@line 183 "modules/graphics/opengl/Polyline.cpp"
 $61 = $60 < 0.05000000074505806; //@line 183 "modules/graphics/opengl/Polyline.cpp"
 if ($61) {
  $62 = $4 * $50; //@line 281 "./common/Vector.h"
  $63 = $9 * $53; //@line 281 "./common/Vector.h"
  $64 = $62 + $63; //@line 281 "./common/Vector.h"
  $65 = $64 > 0.0; //@line 183 "modules/graphics/opengl/Polyline.cpp"
  if ($65) {
   $66 = ((($normals)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $67 = HEAP32[$66>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $68 = ((($normals)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $69 = HEAP32[$68>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $70 = ($67|0)==($69|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($70) {
    __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($normals,$ns); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre11 = HEAP32[$66>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $91 = $$pre11;
   } else {
    $71 = $ns; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $72 = $71; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $73 = HEAP32[$72>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $74 = (($71) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $75 = $74; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $76 = HEAP32[$75>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $77 = $67; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $78 = $77; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$78>>2] = $73; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $79 = (($77) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $80 = $79; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$80>>2] = $76; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $81 = HEAP32[$66>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $82 = ((($81)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$66>>2] = $82; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $91 = $82;
   }
   $83 = +HEAPF32[$ns>>2]; //@line 252 "./common/Vector.h"
   $84 = -$83; //@line 252 "./common/Vector.h"
   $85 = ((($ns)) + 4|0); //@line 252 "./common/Vector.h"
   $86 = +HEAPF32[$85>>2]; //@line 252 "./common/Vector.h"
   $87 = -$86; //@line 252 "./common/Vector.h"
   HEAPF32[$0>>2] = $84; //@line 225 "./common/Vector.h"
   $88 = ((($0)) + 4|0); //@line 226 "./common/Vector.h"
   HEAPF32[$88>>2] = $87; //@line 226 "./common/Vector.h"
   $89 = HEAP32[$68>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $90 = ($91>>>0)<($89>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($90) {
    $92 = $0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $93 = $92; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $94 = HEAP32[$93>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $95 = (($92) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $96 = $95; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $97 = HEAP32[$96>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $98 = $91; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $99 = $98; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$99>>2] = $94; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $100 = (($98) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $101 = $100; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$101>>2] = $97; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $102 = HEAP32[$66>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $103 = ((($102)) + 8|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$66>>2] = $103; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre$phi12Z2D = $85;$$pre$phiZ2D = $ns;
    HEAPF32[$s>>2] = $4; //@line 198 "modules/graphics/opengl/Polyline.cpp"
    HEAPF32[$52>>2] = $9; //@line 198 "modules/graphics/opengl/Polyline.cpp"
    HEAPF32[$$pre$phiZ2D>>2] = $16; //@line 199 "modules/graphics/opengl/Polyline.cpp"
    HEAPF32[$$pre$phi12Z2D>>2] = $17; //@line 199 "modules/graphics/opengl/Polyline.cpp"
    HEAPF32[$len_s>>2] = $13; //@line 200 "modules/graphics/opengl/Polyline.cpp"
    STACKTOP = sp;return; //@line 201 "modules/graphics/opengl/Polyline.cpp"
   } else {
    __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIS2_EEvOT_($normals,$0); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre$phi12Z2D = $85;$$pre$phiZ2D = $ns;
    HEAPF32[$s>>2] = $4; //@line 198 "modules/graphics/opengl/Polyline.cpp"
    HEAPF32[$52>>2] = $9; //@line 198 "modules/graphics/opengl/Polyline.cpp"
    HEAPF32[$$pre$phiZ2D>>2] = $16; //@line 199 "modules/graphics/opengl/Polyline.cpp"
    HEAPF32[$$pre$phi12Z2D>>2] = $17; //@line 199 "modules/graphics/opengl/Polyline.cpp"
    HEAPF32[$len_s>>2] = $13; //@line 200 "modules/graphics/opengl/Polyline.cpp"
    STACKTOP = sp;return; //@line 201 "modules/graphics/opengl/Polyline.cpp"
   }
  }
 }
 $104 = +HEAPF32[$ns>>2]; //@line 305 "./common/Vector.h"
 $105 = $16 - $104; //@line 237 "./common/Vector.h"
 $106 = ((($ns)) + 4|0); //@line 310 "./common/Vector.h"
 $107 = +HEAPF32[$106>>2]; //@line 310 "./common/Vector.h"
 $108 = $17 - $107; //@line 237 "./common/Vector.h"
 $109 = $9 * $105; //@line 286 "./common/Vector.h"
 $110 = $4 * $108; //@line 286 "./common/Vector.h"
 $111 = $109 - $110; //@line 286 "./common/Vector.h"
 $112 = $111 / $55; //@line 192 "modules/graphics/opengl/Polyline.cpp"
 $113 = $50 * $112; //@line 242 "./common/Vector.h"
 $114 = $53 * $112; //@line 242 "./common/Vector.h"
 $115 = $104 + $113; //@line 232 "./common/Vector.h"
 $116 = $107 + $114; //@line 252 "./common/Vector.h"
 HEAPF32[$d>>2] = $115; //@line 225 "./common/Vector.h"
 $117 = ((($d)) + 4|0); //@line 226 "./common/Vector.h"
 HEAPF32[$117>>2] = $116; //@line 226 "./common/Vector.h"
 $118 = ((($normals)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $119 = HEAP32[$118>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $120 = ((($normals)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $121 = HEAP32[$120>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $122 = ($119|0)==($121|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($122) {
  __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($normals,$d); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre10 = HEAP32[$118>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $137 = $115;$142 = $$pre10;
 } else {
  $123 = $d; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $124 = $123; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $125 = HEAP32[$124>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $126 = (($123) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $127 = $126; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $128 = HEAP32[$127>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $129 = $119; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $130 = $129; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$130>>2] = $125; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $131 = (($129) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $132 = $131; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$132>>2] = $128; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $133 = HEAP32[$118>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $134 = ((($133)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$118>>2] = $134; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $135 = (HEAP32[tempDoublePtr>>2]=$125,+HEAPF32[tempDoublePtr>>2]); //@line 1598 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $137 = $135;$142 = $134;
 }
 $136 = -$137; //@line 252 "./common/Vector.h"
 $138 = -$116; //@line 252 "./common/Vector.h"
 HEAPF32[$1>>2] = $136; //@line 225 "./common/Vector.h"
 $139 = ((($1)) + 4|0); //@line 226 "./common/Vector.h"
 HEAPF32[$139>>2] = $138; //@line 226 "./common/Vector.h"
 $140 = HEAP32[$120>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $141 = ($142>>>0)<($140>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($141) {
  $143 = $1; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $144 = $143; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $145 = HEAP32[$144>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $146 = (($143) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $147 = $146; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $148 = HEAP32[$147>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $149 = $142; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $150 = $149; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$150>>2] = $145; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $151 = (($149) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $152 = $151; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$152>>2] = $148; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $153 = HEAP32[$118>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $154 = ((($153)) + 8|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$118>>2] = $154; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 } else {
  __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIS2_EEvOT_($normals,$1); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $$pre$phi12Z2D = $106;$$pre$phiZ2D = $ns;
 HEAPF32[$s>>2] = $4; //@line 198 "modules/graphics/opengl/Polyline.cpp"
 HEAPF32[$52>>2] = $9; //@line 198 "modules/graphics/opengl/Polyline.cpp"
 HEAPF32[$$pre$phiZ2D>>2] = $16; //@line 199 "modules/graphics/opengl/Polyline.cpp"
 HEAPF32[$$pre$phi12Z2D>>2] = $17; //@line 199 "modules/graphics/opengl/Polyline.cpp"
 HEAPF32[$len_s>>2] = $13; //@line 200 "modules/graphics/opengl/Polyline.cpp"
 STACKTOP = sp;return; //@line 201 "modules/graphics/opengl/Polyline.cpp"
}
function __ZN4love8graphics6opengl17BevelJoinPolylineD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (25084); //@line 356 "modules/graphics/opengl/Polyline.cpp"
 $0 = ((($this)) + 4|0); //@line 357 "modules/graphics/opengl/Polyline.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 357 "modules/graphics/opengl/Polyline.cpp"
 $2 = ($1|0)==(0|0); //@line 357 "modules/graphics/opengl/Polyline.cpp"
 if (!($2)) {
  __ZdaPv($1); //@line 358 "modules/graphics/opengl/Polyline.cpp"
 }
 __ZdlPv($this); //@line 166 "modules/graphics/opengl/Polyline.h"
 return; //@line 166 "modules/graphics/opengl/Polyline.h"
}
function __ZN4love8graphics6opengl17BevelJoinPolyline10renderEdgeERNSt3__16vectorINS_6VectorENS3_9allocatorIS5_EEEES9_RS5_RfSA_RKS5_SD_f($this,$anchors,$normals,$s,$len_s,$ns,$q,$r,$hw) {
 $this = $this|0;
 $anchors = $anchors|0;
 $normals = $normals|0;
 $s = $s|0;
 $len_s = $len_s|0;
 $ns = $ns|0;
 $q = $q|0;
 $r = $r|0;
 $hw = +$hw;
 var $$pre = 0, $$pre24 = 0, $$pre25 = 0, $$pre26 = 0, $$pre27 = 0, $$pre28 = 0, $$pre29 = 0, $$pre30 = 0, $$pre31 = 0, $$pre32 = 0, $$pre33 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0.0, $110 = 0.0, $111 = 0, $112 = 0.0, $113 = 0.0, $114 = 0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0.0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0;
 var $124 = 0.0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0.0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0.0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0.0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0.0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0.0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0.0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0.0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0.0, $21 = 0.0, $210 = 0, $211 = 0, $212 = 0, $213 = 0;
 var $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0.0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0.0, $230 = 0, $231 = 0;
 var $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0.0, $240 = 0, $241 = 0.0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0.0;
 var $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0.0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0;
 var $269 = 0, $27 = 0.0, $270 = 0.0, $271 = 0.0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0;
 var $287 = 0, $288 = 0, $289 = 0, $29 = 0.0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0.0, $300 = 0, $301 = 0, $302 = 0, $303 = 0;
 var $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0.0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0;
 var $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0.0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0.0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0.0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $d = 0, $fabsf = 0.0, $n = 0, $nt = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $n = sp + 56|0;
 $0 = sp + 48|0;
 $nt = sp + 40|0;
 $d = sp + 32|0;
 $1 = sp + 24|0;
 $2 = sp + 16|0;
 $3 = sp + 8|0;
 $4 = sp;
 $5 = +HEAPF32[$r>>2]; //@line 237 "./common/Vector.h"
 $6 = +HEAPF32[$q>>2]; //@line 305 "./common/Vector.h"
 $7 = $5 - $6; //@line 237 "./common/Vector.h"
 $8 = ((($r)) + 4|0); //@line 237 "./common/Vector.h"
 $9 = +HEAPF32[$8>>2]; //@line 237 "./common/Vector.h"
 $10 = ((($q)) + 4|0); //@line 310 "./common/Vector.h"
 $11 = +HEAPF32[$10>>2]; //@line 310 "./common/Vector.h"
 $12 = $9 - $11; //@line 237 "./common/Vector.h"
 $13 = $7 * $7; //@line 190 "./common/Vector.h"
 $14 = $12 * $12; //@line 190 "./common/Vector.h"
 $15 = $13 + $14; //@line 190 "./common/Vector.h"
 $16 = (+Math_sqrt((+$15))); //@line 190 "./common/Vector.h"
 $17 = +HEAPF32[$s>>2]; //@line 242 "./common/Vector.h"
 $18 = $12 * $17; //@line 286 "./common/Vector.h"
 $19 = ((($s)) + 4|0); //@line 286 "./common/Vector.h"
 $20 = +HEAPF32[$19>>2]; //@line 242 "./common/Vector.h"
 $21 = $7 * $20; //@line 286 "./common/Vector.h"
 $22 = $18 - $21; //@line 286 "./common/Vector.h"
 $fabsf = (+Math_abs((+$22))); //@line 228 "modules/graphics/opengl/Polyline.cpp"
 $23 = $fabsf; //@line 228 "modules/graphics/opengl/Polyline.cpp"
 $24 = +HEAPF32[$len_s>>2]; //@line 228 "modules/graphics/opengl/Polyline.cpp"
 $25 = $16 * $24; //@line 228 "modules/graphics/opengl/Polyline.cpp"
 $26 = $25; //@line 228 "modules/graphics/opengl/Polyline.cpp"
 $27 = $23 / $26; //@line 228 "modules/graphics/opengl/Polyline.cpp"
 $28 = $27 < 0.05000000074505806; //@line 228 "modules/graphics/opengl/Polyline.cpp"
 if ($28) {
  $29 = $7 * $17; //@line 281 "./common/Vector.h"
  $30 = $12 * $20; //@line 281 "./common/Vector.h"
  $31 = $29 + $30; //@line 281 "./common/Vector.h"
  $32 = $31 > 0.0; //@line 228 "modules/graphics/opengl/Polyline.cpp"
  if ($32) {
   $33 = $hw / $16; //@line 231 "modules/graphics/opengl/Polyline.cpp"
   $34 = $12 * $33; //@line 200 "./common/Vector.h"
   $35 = -$34; //@line 200 "./common/Vector.h"
   $36 = $7 * $33; //@line 252 "./common/Vector.h"
   HEAPF32[$n>>2] = $35; //@line 225 "./common/Vector.h"
   $37 = ((($n)) + 4|0); //@line 226 "./common/Vector.h"
   HEAPF32[$37>>2] = $36; //@line 226 "./common/Vector.h"
   $38 = ((($anchors)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $39 = HEAP32[$38>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $40 = ((($anchors)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $41 = HEAP32[$40>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $42 = ($39|0)==($41|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($42) {
    __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($anchors,$q); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre32 = HEAP32[$38>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $57 = $$pre32;
   } else {
    $43 = $q; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $44 = $43; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $45 = HEAP32[$44>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $46 = (($43) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $47 = $46; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $48 = HEAP32[$47>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $49 = $39; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $50 = $49; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$50>>2] = $45; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $51 = (($49) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $52 = $51; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$52>>2] = $48; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $53 = HEAP32[$38>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $54 = ((($53)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$38>>2] = $54; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $57 = $54;
   }
   $55 = HEAP32[$40>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $56 = ($57|0)==($55|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($56) {
    __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($anchors,$q); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   } else {
    $58 = $q; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $59 = $58; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $60 = HEAP32[$59>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $61 = (($58) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $62 = $61; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $63 = HEAP32[$62>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $64 = $57; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $65 = $64; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$65>>2] = $60; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $66 = (($64) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $67 = $66; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$67>>2] = $63; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $68 = HEAP32[$38>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $69 = ((($68)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$38>>2] = $69; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $70 = ((($normals)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $71 = HEAP32[$70>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $72 = ((($normals)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $73 = HEAP32[$72>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $74 = ($71|0)==($73|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($74) {
    __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($normals,$n); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre33 = HEAP32[$70>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $89 = $35;$94 = $$pre33;
   } else {
    $75 = $n; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $76 = $75; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $77 = HEAP32[$76>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $78 = (($75) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $79 = $78; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $80 = HEAP32[$79>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $81 = $71; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $82 = $81; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$82>>2] = $77; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $83 = (($81) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $84 = $83; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$84>>2] = $80; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $85 = HEAP32[$70>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $86 = ((($85)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$70>>2] = $86; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $87 = (HEAP32[tempDoublePtr>>2]=$77,+HEAPF32[tempDoublePtr>>2]); //@line 1598 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $89 = $87;$94 = $86;
   }
   $88 = -$89; //@line 252 "./common/Vector.h"
   $90 = -$36; //@line 252 "./common/Vector.h"
   HEAPF32[$0>>2] = $88; //@line 225 "./common/Vector.h"
   $91 = ((($0)) + 4|0); //@line 226 "./common/Vector.h"
   HEAPF32[$91>>2] = $90; //@line 226 "./common/Vector.h"
   $92 = HEAP32[$72>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $93 = ($94>>>0)<($92>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($93) {
    $95 = $0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $96 = $95; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $97 = HEAP32[$96>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $98 = (($95) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $99 = $98; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $100 = HEAP32[$99>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $101 = $94; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $102 = $101; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$102>>2] = $97; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $103 = (($101) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $104 = $103; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$104>>2] = $100; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $105 = HEAP32[$70>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $106 = ((($105)) + 8|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$70>>2] = $106; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   } else {
    __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIS2_EEvOT_($normals,$0); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   HEAPF32[$s>>2] = $7; //@line 236 "modules/graphics/opengl/Polyline.cpp"
   HEAPF32[$19>>2] = $12; //@line 236 "modules/graphics/opengl/Polyline.cpp"
   HEAPF32[$len_s>>2] = $16; //@line 237 "modules/graphics/opengl/Polyline.cpp"
   STACKTOP = sp;return; //@line 267 "modules/graphics/opengl/Polyline.cpp"
  }
 }
 $107 = $hw / $16; //@line 242 "modules/graphics/opengl/Polyline.cpp"
 $108 = $12 * $107; //@line 200 "./common/Vector.h"
 $109 = -$108; //@line 252 "./common/Vector.h"
 $110 = $7 * $107; //@line 252 "./common/Vector.h"
 HEAPF32[$nt>>2] = $109; //@line 225 "./common/Vector.h"
 $111 = ((($nt)) + 4|0); //@line 226 "./common/Vector.h"
 HEAPF32[$111>>2] = $110; //@line 226 "./common/Vector.h"
 $112 = +HEAPF32[$ns>>2]; //@line 305 "./common/Vector.h"
 $113 = $109 - $112; //@line 237 "./common/Vector.h"
 $114 = ((($ns)) + 4|0); //@line 310 "./common/Vector.h"
 $115 = +HEAPF32[$114>>2]; //@line 310 "./common/Vector.h"
 $116 = $110 - $115; //@line 237 "./common/Vector.h"
 $117 = $12 * $113; //@line 286 "./common/Vector.h"
 $118 = $7 * $116; //@line 286 "./common/Vector.h"
 $119 = $117 - $118; //@line 286 "./common/Vector.h"
 $120 = $119 / $22; //@line 243 "modules/graphics/opengl/Polyline.cpp"
 $121 = $17 * $120; //@line 242 "./common/Vector.h"
 $122 = $20 * $120; //@line 242 "./common/Vector.h"
 $123 = $112 + $121; //@line 252 "./common/Vector.h"
 $124 = $115 + $122; //@line 252 "./common/Vector.h"
 HEAPF32[$d>>2] = $123; //@line 225 "./common/Vector.h"
 $125 = ((($d)) + 4|0); //@line 226 "./common/Vector.h"
 HEAPF32[$125>>2] = $124; //@line 226 "./common/Vector.h"
 $126 = ((($anchors)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $127 = HEAP32[$126>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $128 = ((($anchors)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $129 = HEAP32[$128>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $130 = ($127|0)==($129|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($130) {
  __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($anchors,$q); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$126>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $145 = $$pre;
 } else {
  $131 = $q; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $132 = $131; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $133 = HEAP32[$132>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $134 = (($131) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $135 = $134; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $136 = HEAP32[$135>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $137 = $127; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $138 = $137; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$138>>2] = $133; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $139 = (($137) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $140 = $139; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$140>>2] = $136; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $141 = HEAP32[$126>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $142 = ((($141)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$126>>2] = $142; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $145 = $142;
 }
 $143 = HEAP32[$128>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $144 = ($145|0)==($143|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($144) {
  __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($anchors,$q); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre24 = HEAP32[$126>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $160 = $$pre24;
 } else {
  $146 = $q; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $147 = $146; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $148 = HEAP32[$147>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $149 = (($146) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $150 = $149; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $151 = HEAP32[$150>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $152 = $145; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $153 = $152; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$153>>2] = $148; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $154 = (($152) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $155 = $154; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$155>>2] = $151; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $156 = HEAP32[$126>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $157 = ((($156)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$126>>2] = $157; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $160 = $157;
 }
 $158 = HEAP32[$128>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $159 = ($160|0)==($158|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($159) {
  __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($anchors,$q); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre25 = HEAP32[$126>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $175 = $$pre25;
 } else {
  $161 = $q; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $162 = $161; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $163 = HEAP32[$162>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $164 = (($161) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $165 = $164; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $166 = HEAP32[$165>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $167 = $160; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $168 = $167; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$168>>2] = $163; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $169 = (($167) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $170 = $169; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$170>>2] = $166; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $171 = HEAP32[$126>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $172 = ((($171)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$126>>2] = $172; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $175 = $172;
 }
 $173 = HEAP32[$128>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $174 = ($175|0)==($173|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($174) {
  __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($anchors,$q); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 } else {
  $176 = $q; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $177 = $176; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $178 = HEAP32[$177>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $179 = (($176) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $180 = $179; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $181 = HEAP32[$180>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $182 = $175; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $183 = $182; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$183>>2] = $178; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $184 = (($182) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $185 = $184; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$185>>2] = $181; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $186 = HEAP32[$126>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $187 = ((($186)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$126>>2] = $187; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $188 = $22 > 0.0; //@line 250 "modules/graphics/opengl/Polyline.cpp"
 $189 = ((($normals)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $190 = HEAP32[$189>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $191 = ((($normals)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $192 = HEAP32[$191>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $193 = ($190|0)==($192|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if ($188) {
   if ($193) {
    __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($normals,$d); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre29 = HEAP32[$189>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $213 = $$pre29;
   } else {
    $194 = $d; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $195 = $194; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $196 = HEAP32[$195>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $197 = (($194) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $198 = $197; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $199 = HEAP32[$198>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $200 = $190; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $201 = $200; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$201>>2] = $196; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $202 = (($200) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $203 = $202; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$203>>2] = $199; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $204 = HEAP32[$189>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $205 = ((($204)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$189>>2] = $205; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $213 = $205;
   }
   $206 = +HEAPF32[$ns>>2]; //@line 252 "./common/Vector.h"
   $207 = -$206; //@line 252 "./common/Vector.h"
   $208 = +HEAPF32[$114>>2]; //@line 252 "./common/Vector.h"
   $209 = -$208; //@line 252 "./common/Vector.h"
   HEAPF32[$1>>2] = $207; //@line 225 "./common/Vector.h"
   $210 = ((($1)) + 4|0); //@line 226 "./common/Vector.h"
   HEAPF32[$210>>2] = $209; //@line 226 "./common/Vector.h"
   $211 = HEAP32[$191>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $212 = ($213>>>0)<($211>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($212) {
    $214 = $1; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $215 = $214; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $216 = HEAP32[$215>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $217 = (($214) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $218 = $217; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $219 = HEAP32[$218>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $220 = $213; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $221 = $220; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$221>>2] = $216; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $222 = (($220) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $223 = $222; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$223>>2] = $219; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $224 = HEAP32[$189>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $225 = ((($224)) + 8|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$189>>2] = $225; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $228 = $225;
   } else {
    __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIS2_EEvOT_($normals,$1); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre30 = HEAP32[$189>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $228 = $$pre30;
   }
   $226 = HEAP32[$191>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $227 = ($228|0)==($226|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($227) {
    __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($normals,$d); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre31 = HEAP32[$189>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $245 = $$pre31;
   } else {
    $229 = $d; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $230 = $229; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $231 = HEAP32[$230>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $232 = (($229) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $233 = $232; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $234 = HEAP32[$233>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $235 = $228; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $236 = $235; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$236>>2] = $231; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $237 = (($235) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $238 = $237; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$238>>2] = $234; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $239 = HEAP32[$189>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $240 = ((($239)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$189>>2] = $240; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $245 = $240;
   }
   $241 = -$110; //@line 252 "./common/Vector.h"
   HEAPF32[$2>>2] = $108; //@line 225 "./common/Vector.h"
   $242 = ((($2)) + 4|0); //@line 226 "./common/Vector.h"
   HEAPF32[$242>>2] = $241; //@line 226 "./common/Vector.h"
   $243 = HEAP32[$191>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $244 = ($245>>>0)<($243>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($244) {
    $246 = $2; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $247 = $246; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $248 = HEAP32[$247>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $249 = (($246) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $250 = $249; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $251 = HEAP32[$250>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $252 = $245; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $253 = $252; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$253>>2] = $248; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $254 = (($252) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $255 = $254; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$255>>2] = $251; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $256 = HEAP32[$189>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $257 = ((($256)) + 8|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$189>>2] = $257; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    break;
   } else {
    __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIS2_EEvOT_($normals,$2); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    break;
   }
  } else {
   if ($193) {
    __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($normals,$ns); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre26 = HEAP32[$189>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $275 = $$pre26;
   } else {
    $258 = $ns; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $259 = $258; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $260 = HEAP32[$259>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $261 = (($258) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $262 = $261; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $263 = HEAP32[$262>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $264 = $190; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $265 = $264; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$265>>2] = $260; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $266 = (($264) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $267 = $266; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$267>>2] = $263; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $268 = HEAP32[$189>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $269 = ((($268)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$189>>2] = $269; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $275 = $269;
   }
   $270 = -$123; //@line 252 "./common/Vector.h"
   $271 = -$124; //@line 252 "./common/Vector.h"
   HEAPF32[$3>>2] = $270; //@line 225 "./common/Vector.h"
   $272 = ((($3)) + 4|0); //@line 226 "./common/Vector.h"
   HEAPF32[$272>>2] = $271; //@line 226 "./common/Vector.h"
   $273 = HEAP32[$191>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $274 = ($275>>>0)<($273>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($274) {
    $276 = $3; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $277 = $276; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $278 = HEAP32[$277>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $279 = (($276) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $280 = $279; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $281 = HEAP32[$280>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $282 = $275; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $283 = $282; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$283>>2] = $278; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $284 = (($282) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $285 = $284; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$285>>2] = $281; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $286 = HEAP32[$189>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $287 = ((($286)) + 8|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$189>>2] = $287; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $290 = $287;
   } else {
    __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIS2_EEvOT_($normals,$3); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre27 = HEAP32[$189>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $290 = $$pre27;
   }
   $288 = HEAP32[$191>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $289 = ($290|0)==($288|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($289) {
    __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($normals,$nt); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre28 = HEAP32[$189>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $306 = $$pre28;
   } else {
    $291 = $nt; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $292 = $291; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $293 = HEAP32[$292>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $294 = (($291) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $295 = $294; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $296 = HEAP32[$295>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $297 = $290; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $298 = $297; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$298>>2] = $293; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $299 = (($297) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $300 = $299; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$300>>2] = $296; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $301 = HEAP32[$189>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $302 = ((($301)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$189>>2] = $302; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $306 = $302;
   }
   HEAPF32[$4>>2] = $270; //@line 225 "./common/Vector.h"
   $303 = ((($4)) + 4|0); //@line 226 "./common/Vector.h"
   HEAPF32[$303>>2] = $271; //@line 226 "./common/Vector.h"
   $304 = HEAP32[$191>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $305 = ($306>>>0)<($304>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($305) {
    $307 = $4; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $308 = $307; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $309 = HEAP32[$308>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $310 = (($307) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $311 = $310; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $312 = HEAP32[$311>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $313 = $306; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $314 = $313; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$314>>2] = $309; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $315 = (($313) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $316 = $315; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$316>>2] = $312; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $317 = HEAP32[$189>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $318 = ((($317)) + 8|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$189>>2] = $318; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    break;
   } else {
    __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIS2_EEvOT_($normals,$4); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    break;
   }
  }
 } while(0);
 HEAPF32[$s>>2] = $7; //@line 264 "modules/graphics/opengl/Polyline.cpp"
 HEAPF32[$19>>2] = $12; //@line 264 "modules/graphics/opengl/Polyline.cpp"
 HEAPF32[$len_s>>2] = $16; //@line 265 "modules/graphics/opengl/Polyline.cpp"
 $319 = $nt; //@line 266 "modules/graphics/opengl/Polyline.cpp"
 $320 = $319; //@line 266 "modules/graphics/opengl/Polyline.cpp"
 $321 = HEAP32[$320>>2]|0; //@line 266 "modules/graphics/opengl/Polyline.cpp"
 $322 = (($319) + 4)|0; //@line 266 "modules/graphics/opengl/Polyline.cpp"
 $323 = $322; //@line 266 "modules/graphics/opengl/Polyline.cpp"
 $324 = HEAP32[$323>>2]|0; //@line 266 "modules/graphics/opengl/Polyline.cpp"
 $325 = $ns; //@line 266 "modules/graphics/opengl/Polyline.cpp"
 $326 = $325; //@line 266 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$326>>2] = $321; //@line 266 "modules/graphics/opengl/Polyline.cpp"
 $327 = (($325) + 4)|0; //@line 266 "modules/graphics/opengl/Polyline.cpp"
 $328 = $327; //@line 266 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$328>>2] = $324; //@line 266 "modules/graphics/opengl/Polyline.cpp"
 STACKTOP = sp;return; //@line 267 "modules/graphics/opengl/Polyline.cpp"
}
function __ZN4love8graphics6opengl8Polyline6renderEPKfjjffb($this,$coords,$count,$size_hint,$halfwidth,$pixel_size,$draw_overdraw) {
 $this = $this|0;
 $coords = $coords|0;
 $count = $count|0;
 $size_hint = $size_hint|0;
 $halfwidth = +$halfwidth;
 $pixel_size = +$pixel_size;
 $draw_overdraw = $draw_overdraw|0;
 var $$ = 0, $$0 = 0.0, $$arith = 0, $$in = 0, $$in29 = 0, $$in30 = 0, $$overflow = 0, $$phi$trans$insert = 0, $$pre = 0.0, $$pre26 = 0, $$pre27 = 0, $$sink = 0.0, $$sink19 = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0.0, $109 = 0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0;
 var $159 = 0.0, $16 = 0, $160 = 0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0;
 var $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0;
 var $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0.0, $203 = 0, $204 = 0.0, $205 = 0.0, $206 = 0, $207 = 0.0, $208 = 0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0;
 var $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0;
 var $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0;
 var $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0, $68 = 0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0.0, $75 = 0, $76 = 0, $77 = 0.0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0.0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0.0, $92 = 0.0, $93 = 0, $94 = 0, $95 = 0.0, $96 = 0.0, $97 = 0, $98 = 0, $99 = 0;
 var $extra_vertices$0 = 0, $i$021 = 0, $i1$020 = 0, $len_s = 0, $ns = 0, $q = 0, $r = 0, $s = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i9 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i10 = 0, $storemerge$in = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $s = sp + 32|0;
 $0 = sp + 24|0;
 $1 = sp + 16|0;
 $len_s = sp + 48|0;
 $ns = sp + 40|0;
 $q = sp + 8|0;
 $r = sp;
 $2 = HEAP8[624072]|0; //@line 42 "modules/graphics/opengl/Polyline.cpp"
 $3 = ($2<<24>>24)==(0); //@line 42 "modules/graphics/opengl/Polyline.cpp"
 if ($3) {
  $4 = (___cxa_guard_acquire((624072|0))|0); //@line 42 "modules/graphics/opengl/Polyline.cpp"
  $5 = ($4|0)==(0); //@line 42 "modules/graphics/opengl/Polyline.cpp"
  if (!($5)) {
   HEAP32[158575] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[(634304)>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[(634308)>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   (___cxa_atexit((4110|0),(634300|0),(___dso_handle|0))|0); //@line 42 "modules/graphics/opengl/Polyline.cpp"
   ___cxa_guard_release((624072|0)); //@line 42 "modules/graphics/opengl/Polyline.cpp"
  }
 }
 $6 = HEAP32[158575]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = HEAP32[(634304)>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ($7|0)==($6|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = $6;
 if ($8) {
  $$in = $7;
 } else {
  $scevgep$i$i$i = ((($7)) + -8|0);
  $10 = $scevgep$i$i$i;
  $11 = (($10) - ($9))|0;
  $12 = $11 >>> 3;
  $13 = $12 ^ -1;
  $scevgep4$i$i$i = (($7) + ($13<<3)|0);
  HEAP32[(634304)>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $scevgep4$i$i$i;
 }
 $14 = HEAP32[(634308)>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = (($14) - ($9))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = $15 >> 3; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ($16>>>0)<($size_hint>>>0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($17) {
  $18 = $$in; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($9))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($size_hint|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $25 = 0;
  } else {
   $22 = $size_hint << 3; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $25 = $23;
  }
  $24 = (($25) + ($20<<3)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $26 = $24; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $27 = (($25) + ($size_hint<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $29 = (0 - ($20))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $30 = (($24) + ($29<<3)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $31 = $30; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  _memcpy(($30|0),($6|0),($19|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[158575] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[(634304)>>2] = $26; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[(634308)>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $32 = ($6|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if (!($32)) {
   __ZdlPv($6); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
 }
 $33 = HEAP8[624080]|0; //@line 46 "modules/graphics/opengl/Polyline.cpp"
 $34 = ($33<<24>>24)==(0); //@line 46 "modules/graphics/opengl/Polyline.cpp"
 if ($34) {
  $35 = (___cxa_guard_acquire((624080|0))|0); //@line 46 "modules/graphics/opengl/Polyline.cpp"
  $36 = ($35|0)==(0); //@line 46 "modules/graphics/opengl/Polyline.cpp"
  if (!($36)) {
   HEAP32[158578] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[(634316)>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[(634320)>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   (___cxa_atexit((4110|0),(634312|0),(___dso_handle|0))|0); //@line 46 "modules/graphics/opengl/Polyline.cpp"
   ___cxa_guard_release((624080|0)); //@line 46 "modules/graphics/opengl/Polyline.cpp"
  }
 }
 $37 = HEAP32[158578]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $38 = HEAP32[(634316)>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $39 = ($38|0)==($37|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $40 = $37;
 if ($39) {
  $$in29 = $38;
 } else {
  $scevgep$i$i$i9 = ((($38)) + -8|0);
  $41 = $scevgep$i$i$i9;
  $42 = (($41) - ($40))|0;
  $43 = $42 >>> 3;
  $44 = $43 ^ -1;
  $scevgep4$i$i$i10 = (($38) + ($44<<3)|0);
  HEAP32[(634316)>>2] = $scevgep4$i$i$i10; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in29 = $scevgep4$i$i$i10;
 }
 $45 = HEAP32[(634320)>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $46 = (($45) - ($40))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $47 = $46 >> 3; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $48 = ($47>>>0)<($size_hint>>>0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($48) {
  $49 = $$in29; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $50 = (($49) - ($40))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $51 = $50 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $52 = ($size_hint|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($52) {
   $56 = 0;
  } else {
   $53 = $size_hint << 3; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $54 = (__Znwj($53)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $56 = $54;
  }
  $55 = (($56) + ($51<<3)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $57 = $55; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $58 = (($56) + ($size_hint<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $59 = $58; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $60 = (0 - ($51))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $61 = (($55) + ($60<<3)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $62 = $61; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  _memcpy(($61|0),($37|0),($50|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[158578] = $62; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[(634316)>>2] = $57; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[(634320)>>2] = $59; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $63 = ($37|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if (!($63)) {
   __ZdlPv($37); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
 }
 $64 = $pixel_size * 0.30000001192092896; //@line 52 "modules/graphics/opengl/Polyline.cpp"
 $65 = $halfwidth - $64; //@line 52 "modules/graphics/opengl/Polyline.cpp"
 $$0 = $draw_overdraw ? $65 : $halfwidth; //@line 51 "modules/graphics/opengl/Polyline.cpp"
 $66 = +HEAPF32[$coords>>2]; //@line 60 "modules/graphics/opengl/Polyline.cpp"
 $67 = (($count) + -2)|0; //@line 55 "modules/graphics/opengl/Polyline.cpp"
 $68 = (($coords) + ($67<<2)|0); //@line 55 "modules/graphics/opengl/Polyline.cpp"
 $69 = +HEAPF32[$68>>2]; //@line 55 "modules/graphics/opengl/Polyline.cpp"
 $70 = $66 == $69; //@line 55 "modules/graphics/opengl/Polyline.cpp"
 if ($70) {
  $73 = ((($coords)) + 4|0); //@line 55 "modules/graphics/opengl/Polyline.cpp"
  $74 = +HEAPF32[$73>>2]; //@line 60 "modules/graphics/opengl/Polyline.cpp"
  $75 = (($count) + -1)|0; //@line 55 "modules/graphics/opengl/Polyline.cpp"
  $76 = (($coords) + ($75<<2)|0); //@line 55 "modules/graphics/opengl/Polyline.cpp"
  $77 = +HEAPF32[$76>>2]; //@line 55 "modules/graphics/opengl/Polyline.cpp"
  $78 = $74 == $77; //@line 55 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$s>>2] = 0.0; //@line 219 "./common/Vector.h"
  $79 = ((($s)) + 4|0); //@line 220 "./common/Vector.h"
  HEAPF32[$79>>2] = 0.0; //@line 220 "./common/Vector.h"
  $80 = (HEAPF32[tempDoublePtr>>2]=$74,HEAP32[tempDoublePtr>>2]|0); //@line 57 "modules/graphics/opengl/Polyline.cpp"
  if ($78) {
   $89 = (($count) + -4)|0; //@line 60 "modules/graphics/opengl/Polyline.cpp"
   $90 = (($coords) + ($89<<2)|0); //@line 60 "modules/graphics/opengl/Polyline.cpp"
   $91 = +HEAPF32[$90>>2]; //@line 60 "modules/graphics/opengl/Polyline.cpp"
   $92 = $66 - $91; //@line 60 "modules/graphics/opengl/Polyline.cpp"
   $93 = (($count) + -3)|0; //@line 60 "modules/graphics/opengl/Polyline.cpp"
   $94 = (($coords) + ($93<<2)|0); //@line 60 "modules/graphics/opengl/Polyline.cpp"
   $95 = +HEAPF32[$94>>2]; //@line 60 "modules/graphics/opengl/Polyline.cpp"
   $96 = $74 - $95; //@line 60 "modules/graphics/opengl/Polyline.cpp"
   HEAPF32[$1>>2] = $92; //@line 225 "./common/Vector.h"
   $97 = ((($1)) + 4|0); //@line 226 "./common/Vector.h"
   HEAPF32[$97>>2] = $96; //@line 226 "./common/Vector.h"
   $109 = $s;$112 = $79;$124 = $80;$180 = 1;$264 = $s;$storemerge$in = $1;
  } else {
   $260 = $80;$261 = $79;$262 = $s;$263 = $s;$87 = $74;
   label = 24;
  }
 } else {
  HEAPF32[$s>>2] = 0.0; //@line 219 "./common/Vector.h"
  $71 = ((($s)) + 4|0); //@line 220 "./common/Vector.h"
  HEAPF32[$71>>2] = 0.0; //@line 220 "./common/Vector.h"
  $$phi$trans$insert = ((($coords)) + 4|0); //@line 58 "modules/graphics/opengl/Polyline.cpp"
  $$pre = +HEAPF32[$$phi$trans$insert>>2]; //@line 58 "modules/graphics/opengl/Polyline.cpp"
  $72 = (HEAPF32[tempDoublePtr>>2]=$$pre,HEAP32[tempDoublePtr>>2]|0); //@line 57 "modules/graphics/opengl/Polyline.cpp"
  $260 = $72;$261 = $71;$262 = $s;$263 = $s;$87 = $$pre;
  label = 24;
 }
 if ((label|0) == 24) {
  $81 = ((($coords)) + 8|0); //@line 58 "modules/graphics/opengl/Polyline.cpp"
  $82 = +HEAPF32[$81>>2]; //@line 58 "modules/graphics/opengl/Polyline.cpp"
  $83 = $82 - $66; //@line 58 "modules/graphics/opengl/Polyline.cpp"
  $84 = ((($coords)) + 12|0); //@line 58 "modules/graphics/opengl/Polyline.cpp"
  $85 = +HEAPF32[$84>>2]; //@line 58 "modules/graphics/opengl/Polyline.cpp"
  $86 = $85 - $87; //@line 58 "modules/graphics/opengl/Polyline.cpp"
  HEAPF32[$0>>2] = $83; //@line 225 "./common/Vector.h"
  $88 = ((($0)) + 4|0); //@line 226 "./common/Vector.h"
  HEAPF32[$88>>2] = $86; //@line 226 "./common/Vector.h"
  $109 = $262;$112 = $261;$124 = $260;$180 = 0;$264 = $263;$storemerge$in = $0;
 }
 $98 = $storemerge$in; //@line 60 "modules/graphics/opengl/Polyline.cpp"
 $99 = $98; //@line 60 "modules/graphics/opengl/Polyline.cpp"
 $100 = HEAP32[$99>>2]|0; //@line 60 "modules/graphics/opengl/Polyline.cpp"
 $101 = (($98) + 4)|0; //@line 60 "modules/graphics/opengl/Polyline.cpp"
 $102 = $101; //@line 60 "modules/graphics/opengl/Polyline.cpp"
 $103 = HEAP32[$102>>2]|0; //@line 60 "modules/graphics/opengl/Polyline.cpp"
 $104 = $s; //@line 58 "modules/graphics/opengl/Polyline.cpp"
 $105 = $104; //@line 58 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$105>>2] = $100; //@line 58 "modules/graphics/opengl/Polyline.cpp"
 $106 = (($104) + 4)|0; //@line 58 "modules/graphics/opengl/Polyline.cpp"
 $107 = $106; //@line 58 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$107>>2] = $103; //@line 58 "modules/graphics/opengl/Polyline.cpp"
 $108 = +HEAPF32[$109>>2]; //@line 190 "./common/Vector.h"
 $110 = $108 * $108; //@line 190 "./common/Vector.h"
 $111 = +HEAPF32[$112>>2]; //@line 190 "./common/Vector.h"
 $113 = $111 * $111; //@line 190 "./common/Vector.h"
 $114 = $110 + $113; //@line 190 "./common/Vector.h"
 $115 = (+Math_sqrt((+$114))); //@line 190 "./common/Vector.h"
 HEAPF32[$len_s>>2] = $115; //@line 62 "modules/graphics/opengl/Polyline.cpp"
 $116 = $$0 / $115; //@line 63 "modules/graphics/opengl/Polyline.cpp"
 $117 = +HEAPF32[$112>>2]; //@line 200 "./common/Vector.h"
 $118 = $116 * $117; //@line 200 "./common/Vector.h"
 $119 = -$118; //@line 200 "./common/Vector.h"
 $120 = +HEAPF32[$109>>2]; //@line 200 "./common/Vector.h"
 $121 = $116 * $120; //@line 200 "./common/Vector.h"
 HEAPF32[$ns>>2] = $119; //@line 225 "./common/Vector.h"
 $122 = ((($ns)) + 4|0); //@line 226 "./common/Vector.h"
 HEAPF32[$122>>2] = $121; //@line 226 "./common/Vector.h"
 HEAPF32[$r>>2] = $66; //@line 225 "./common/Vector.h"
 $123 = ((($r)) + 4|0); //@line 226 "./common/Vector.h"
 HEAP32[$123>>2] = $124; //@line 226 "./common/Vector.h"
 $125 = ($count>>>0)>(3); //@line 66 "modules/graphics/opengl/Polyline.cpp"
 $126 = $r; //@line 68 "modules/graphics/opengl/Polyline.cpp"
 $127 = $126; //@line 68 "modules/graphics/opengl/Polyline.cpp"
 $128 = HEAP32[$127>>2]|0; //@line 68 "modules/graphics/opengl/Polyline.cpp"
 $129 = (($126) + 4)|0; //@line 68 "modules/graphics/opengl/Polyline.cpp"
 $130 = $129; //@line 68 "modules/graphics/opengl/Polyline.cpp"
 $131 = HEAP32[$130>>2]|0; //@line 68 "modules/graphics/opengl/Polyline.cpp"
 $132 = $q; //@line 68 "modules/graphics/opengl/Polyline.cpp"
 $133 = $132; //@line 68 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$133>>2] = $128; //@line 68 "modules/graphics/opengl/Polyline.cpp"
 $134 = (($132) + 4)|0; //@line 68 "modules/graphics/opengl/Polyline.cpp"
 $135 = $134; //@line 68 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$135>>2] = $131; //@line 68 "modules/graphics/opengl/Polyline.cpp"
 if ($125) {
  $136 = ((($r)) + 4|0); //@line 69 "modules/graphics/opengl/Polyline.cpp"
  $141 = 3;$i$021 = 0;
  while(1) {
   $137 = (($i$021) + 2)|0; //@line 69 "modules/graphics/opengl/Polyline.cpp"
   $138 = (($coords) + ($137<<2)|0); //@line 69 "modules/graphics/opengl/Polyline.cpp"
   $139 = HEAP32[$138>>2]|0; //@line 69 "modules/graphics/opengl/Polyline.cpp"
   $140 = (($coords) + ($141<<2)|0); //@line 69 "modules/graphics/opengl/Polyline.cpp"
   $142 = HEAP32[$140>>2]|0; //@line 69 "modules/graphics/opengl/Polyline.cpp"
   HEAP32[$r>>2] = $139; //@line 69 "modules/graphics/opengl/Polyline.cpp"
   HEAP32[$136>>2] = $142; //@line 69 "modules/graphics/opengl/Polyline.cpp"
   $143 = HEAP32[$this>>2]|0; //@line 70 "modules/graphics/opengl/Polyline.cpp"
   $144 = ((($143)) + 20|0); //@line 70 "modules/graphics/opengl/Polyline.cpp"
   $145 = HEAP32[$144>>2]|0; //@line 70 "modules/graphics/opengl/Polyline.cpp"
   FUNCTION_TABLE_viiiiiiiid[$145 & 2047]($this,634300,634312,$s,$len_s,$ns,$q,$r,$$0); //@line 70 "modules/graphics/opengl/Polyline.cpp"
   $146 = (($i$021) + 5)|0; //@line 66 "modules/graphics/opengl/Polyline.cpp"
   $147 = ($146>>>0)<($count>>>0); //@line 66 "modules/graphics/opengl/Polyline.cpp"
   $148 = $r; //@line 68 "modules/graphics/opengl/Polyline.cpp"
   $149 = $148; //@line 68 "modules/graphics/opengl/Polyline.cpp"
   $150 = HEAP32[$149>>2]|0; //@line 68 "modules/graphics/opengl/Polyline.cpp"
   $151 = (($148) + 4)|0; //@line 68 "modules/graphics/opengl/Polyline.cpp"
   $152 = $151; //@line 68 "modules/graphics/opengl/Polyline.cpp"
   $153 = HEAP32[$152>>2]|0; //@line 68 "modules/graphics/opengl/Polyline.cpp"
   $154 = $q; //@line 68 "modules/graphics/opengl/Polyline.cpp"
   $155 = $154; //@line 68 "modules/graphics/opengl/Polyline.cpp"
   HEAP32[$155>>2] = $150; //@line 68 "modules/graphics/opengl/Polyline.cpp"
   $156 = (($154) + 4)|0; //@line 68 "modules/graphics/opengl/Polyline.cpp"
   $157 = $156; //@line 68 "modules/graphics/opengl/Polyline.cpp"
   HEAP32[$157>>2] = $153; //@line 68 "modules/graphics/opengl/Polyline.cpp"
   if ($147) {
    $141 = $146;$i$021 = $137;
   } else {
    $265 = $150;$266 = $153;
    break;
   }
  }
  $$in30 = $265;
 } else {
  $$in30 = $128;
 }
 if ($180) {
  $158 = ((($coords)) + 8|0); //@line 74 "modules/graphics/opengl/Polyline.cpp"
  $159 = +HEAPF32[$158>>2]; //@line 74 "modules/graphics/opengl/Polyline.cpp"
  $160 = ((($coords)) + 12|0); //@line 74 "modules/graphics/opengl/Polyline.cpp"
  $161 = +HEAPF32[$160>>2]; //@line 74 "modules/graphics/opengl/Polyline.cpp"
  $$sink = $161;$$sink19 = $159;
 } else {
  $162 = (HEAP32[tempDoublePtr>>2]=$$in30,+HEAPF32[tempDoublePtr>>2]); //@line 232 "./common/Vector.h"
  $163 = +HEAPF32[$109>>2]; //@line 232 "./common/Vector.h"
  $164 = $162 + $163; //@line 232 "./common/Vector.h"
  $165 = +HEAPF32[$123>>2]; //@line 232 "./common/Vector.h"
  $166 = +HEAPF32[$112>>2]; //@line 232 "./common/Vector.h"
  $167 = $165 + $166; //@line 232 "./common/Vector.h"
  $$sink = $167;$$sink19 = $164;
 }
 HEAPF32[$r>>2] = $$sink19; //@line 74 "modules/graphics/opengl/Polyline.cpp"
 $168 = ((($r)) + 4|0); //@line 74 "modules/graphics/opengl/Polyline.cpp"
 HEAPF32[$168>>2] = $$sink; //@line 74 "modules/graphics/opengl/Polyline.cpp"
 $169 = HEAP32[$this>>2]|0; //@line 75 "modules/graphics/opengl/Polyline.cpp"
 $170 = ((($169)) + 20|0); //@line 75 "modules/graphics/opengl/Polyline.cpp"
 $171 = HEAP32[$170>>2]|0; //@line 75 "modules/graphics/opengl/Polyline.cpp"
 FUNCTION_TABLE_viiiiiiiid[$171 & 2047]($this,634300,634312,$s,$len_s,$ns,$q,$r,$$0); //@line 75 "modules/graphics/opengl/Polyline.cpp"
 $172 = HEAP32[(634316)>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $173 = HEAP32[158578]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $174 = (($172) - ($173))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $175 = $174 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $176 = ((($this)) + 12|0); //@line 77 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$176>>2] = $175; //@line 77 "modules/graphics/opengl/Polyline.cpp"
 if ($draw_overdraw) {
  $177 = HEAP32[$this>>2]|0; //@line 83 "modules/graphics/opengl/Polyline.cpp"
  $178 = ((($177)) + 8|0); //@line 83 "modules/graphics/opengl/Polyline.cpp"
  $179 = HEAP32[$178>>2]|0; //@line 83 "modules/graphics/opengl/Polyline.cpp"
  FUNCTION_TABLE_vii[$179 & 8191]($this,$180); //@line 83 "modules/graphics/opengl/Polyline.cpp"
  $181 = ((($this)) + 20|0); //@line 89 "modules/graphics/opengl/Polyline.cpp"
  $182 = HEAP32[$181>>2]|0; //@line 89 "modules/graphics/opengl/Polyline.cpp"
  $183 = ($182|0)==(5); //@line 89 "modules/graphics/opengl/Polyline.cpp"
  $$ = $183 ? 2 : 0; //@line 89 "modules/graphics/opengl/Polyline.cpp"
  $$pre26 = HEAP32[$176>>2]|0; //@line 94 "modules/graphics/opengl/Polyline.cpp"
  $185 = $$pre26;$extra_vertices$0 = $$;
 } else {
  $185 = $175;$extra_vertices$0 = 0;
 }
 $184 = (($185) + ($extra_vertices$0))|0; //@line 94 "modules/graphics/opengl/Polyline.cpp"
 $186 = ((($this)) + 16|0); //@line 94 "modules/graphics/opengl/Polyline.cpp"
 $187 = HEAP32[$186>>2]|0; //@line 94 "modules/graphics/opengl/Polyline.cpp"
 $188 = (($184) + ($187))|0; //@line 94 "modules/graphics/opengl/Polyline.cpp"
 $$arith = $188<<3; //@line 94 "modules/graphics/opengl/Polyline.cpp"
 $$overflow = ($188>>>0)>(536870911); //@line 94 "modules/graphics/opengl/Polyline.cpp"
 $189 = $$overflow ? -1 : $$arith; //@line 94 "modules/graphics/opengl/Polyline.cpp"
 $190 = (__Znaj($189)|0); //@line 94 "modules/graphics/opengl/Polyline.cpp"
 $191 = ($188|0)==(0); //@line 94 "modules/graphics/opengl/Polyline.cpp"
 if (!($191)) {
  $192 = (($190) + ($188<<3)|0); //@line 94 "modules/graphics/opengl/Polyline.cpp"
  $193 = $190;
  while(1) {
   HEAPF32[$193>>2] = 0.0; //@line 219 "./common/Vector.h"
   $194 = ((($193)) + 4|0); //@line 220 "./common/Vector.h"
   HEAPF32[$194>>2] = 0.0; //@line 220 "./common/Vector.h"
   $195 = ((($193)) + 8|0); //@line 94 "modules/graphics/opengl/Polyline.cpp"
   $196 = ($195|0)==($192|0); //@line 94 "modules/graphics/opengl/Polyline.cpp"
   if ($196) {
    break;
   } else {
    $193 = $195;
   }
  }
 }
 $197 = ((($this)) + 4|0); //@line 94 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$197>>2] = $190; //@line 94 "modules/graphics/opengl/Polyline.cpp"
 $198 = ($185|0)==(0); //@line 96 "modules/graphics/opengl/Polyline.cpp"
 L54: do {
  if ($198) {
   $219 = 0;
  } else {
   $212 = $190;$i1$020 = 0;
   while(1) {
    $199 = HEAP32[158575]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $200 = HEAP32[158578]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $201 = (($199) + ($i1$020<<3)|0); //@line 232 "./common/Vector.h"
    $202 = +HEAPF32[$201>>2]; //@line 232 "./common/Vector.h"
    $203 = (($200) + ($i1$020<<3)|0); //@line 232 "./common/Vector.h"
    $204 = +HEAPF32[$203>>2]; //@line 232 "./common/Vector.h"
    $205 = $202 + $204; //@line 232 "./common/Vector.h"
    $206 = (((($199) + ($i1$020<<3)|0)) + 4|0); //@line 232 "./common/Vector.h"
    $207 = +HEAPF32[$206>>2]; //@line 232 "./common/Vector.h"
    $208 = (((($200) + ($i1$020<<3)|0)) + 4|0); //@line 232 "./common/Vector.h"
    $209 = +HEAPF32[$208>>2]; //@line 232 "./common/Vector.h"
    $210 = $207 + $209; //@line 232 "./common/Vector.h"
    $211 = (($212) + ($i1$020<<3)|0); //@line 97 "modules/graphics/opengl/Polyline.cpp"
    HEAPF32[$211>>2] = $205; //@line 97 "modules/graphics/opengl/Polyline.cpp"
    $213 = (((($212) + ($i1$020<<3)|0)) + 4|0); //@line 97 "modules/graphics/opengl/Polyline.cpp"
    HEAPF32[$213>>2] = $210; //@line 97 "modules/graphics/opengl/Polyline.cpp"
    $214 = (($i1$020) + 1)|0; //@line 96 "modules/graphics/opengl/Polyline.cpp"
    $215 = HEAP32[$176>>2]|0; //@line 96 "modules/graphics/opengl/Polyline.cpp"
    $216 = ($214>>>0)<($215>>>0); //@line 96 "modules/graphics/opengl/Polyline.cpp"
    if (!($216)) {
     $219 = $215;
     break L54;
    }
    $$pre27 = HEAP32[$197>>2]|0; //@line 97 "modules/graphics/opengl/Polyline.cpp"
    $212 = $$pre27;$i1$020 = $214;
   }
  }
 } while(0);
 if ($draw_overdraw) {
  $217 = HEAP32[$197>>2]|0; //@line 101 "modules/graphics/opengl/Polyline.cpp"
  $218 = (($217) + ($219<<3)|0); //@line 101 "modules/graphics/opengl/Polyline.cpp"
  $220 = (($218) + ($extra_vertices$0<<3)|0); //@line 101 "modules/graphics/opengl/Polyline.cpp"
  $221 = ((($this)) + 8|0); //@line 101 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$221>>2] = $220; //@line 101 "modules/graphics/opengl/Polyline.cpp"
  $222 = (($219) + ($extra_vertices$0))|0; //@line 102 "modules/graphics/opengl/Polyline.cpp"
  $223 = ((($this)) + 28|0); //@line 102 "modules/graphics/opengl/Polyline.cpp"
  HEAP32[$223>>2] = $222; //@line 102 "modules/graphics/opengl/Polyline.cpp"
  $224 = HEAP32[$this>>2]|0; //@line 103 "modules/graphics/opengl/Polyline.cpp"
  $225 = ((($224)) + 12|0); //@line 103 "modules/graphics/opengl/Polyline.cpp"
  $226 = HEAP32[$225>>2]|0; //@line 103 "modules/graphics/opengl/Polyline.cpp"
  FUNCTION_TABLE_viidi[$226 & 4095]($this,634312,$pixel_size,$180); //@line 103 "modules/graphics/opengl/Polyline.cpp"
 }
 $227 = ($extra_vertices$0|0)==(0); //@line 107 "modules/graphics/opengl/Polyline.cpp"
 if ($227) {
  STACKTOP = sp;return; //@line 112 "modules/graphics/opengl/Polyline.cpp"
 }
 $228 = HEAP32[$176>>2]|0; //@line 109 "modules/graphics/opengl/Polyline.cpp"
 $229 = HEAP32[$197>>2]|0; //@line 109 "modules/graphics/opengl/Polyline.cpp"
 $230 = (($229) + ($228<<3)|0); //@line 109 "modules/graphics/opengl/Polyline.cpp"
 $231 = (($228) + -1)|0; //@line 109 "modules/graphics/opengl/Polyline.cpp"
 $232 = (($229) + ($231<<3)|0); //@line 109 "modules/graphics/opengl/Polyline.cpp"
 $233 = $232; //@line 109 "modules/graphics/opengl/Polyline.cpp"
 $234 = $233; //@line 109 "modules/graphics/opengl/Polyline.cpp"
 $235 = HEAP32[$234>>2]|0; //@line 109 "modules/graphics/opengl/Polyline.cpp"
 $236 = (($233) + 4)|0; //@line 109 "modules/graphics/opengl/Polyline.cpp"
 $237 = $236; //@line 109 "modules/graphics/opengl/Polyline.cpp"
 $238 = HEAP32[$237>>2]|0; //@line 109 "modules/graphics/opengl/Polyline.cpp"
 $239 = $230; //@line 109 "modules/graphics/opengl/Polyline.cpp"
 $240 = $239; //@line 109 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$240>>2] = $235; //@line 109 "modules/graphics/opengl/Polyline.cpp"
 $241 = (($239) + 4)|0; //@line 109 "modules/graphics/opengl/Polyline.cpp"
 $242 = $241; //@line 109 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$242>>2] = $238; //@line 109 "modules/graphics/opengl/Polyline.cpp"
 $243 = HEAP32[$176>>2]|0; //@line 110 "modules/graphics/opengl/Polyline.cpp"
 $244 = (($243) + 1)|0; //@line 110 "modules/graphics/opengl/Polyline.cpp"
 $245 = HEAP32[$197>>2]|0; //@line 110 "modules/graphics/opengl/Polyline.cpp"
 $246 = (($245) + ($244<<3)|0); //@line 110 "modules/graphics/opengl/Polyline.cpp"
 $247 = ((($this)) + 28|0); //@line 110 "modules/graphics/opengl/Polyline.cpp"
 $248 = HEAP32[$247>>2]|0; //@line 110 "modules/graphics/opengl/Polyline.cpp"
 $249 = (($245) + ($248<<3)|0); //@line 110 "modules/graphics/opengl/Polyline.cpp"
 $250 = $249; //@line 110 "modules/graphics/opengl/Polyline.cpp"
 $251 = $250; //@line 110 "modules/graphics/opengl/Polyline.cpp"
 $252 = HEAP32[$251>>2]|0; //@line 110 "modules/graphics/opengl/Polyline.cpp"
 $253 = (($250) + 4)|0; //@line 110 "modules/graphics/opengl/Polyline.cpp"
 $254 = $253; //@line 110 "modules/graphics/opengl/Polyline.cpp"
 $255 = HEAP32[$254>>2]|0; //@line 110 "modules/graphics/opengl/Polyline.cpp"
 $256 = $246; //@line 110 "modules/graphics/opengl/Polyline.cpp"
 $257 = $256; //@line 110 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$257>>2] = $252; //@line 110 "modules/graphics/opengl/Polyline.cpp"
 $258 = (($256) + 4)|0; //@line 110 "modules/graphics/opengl/Polyline.cpp"
 $259 = $258; //@line 110 "modules/graphics/opengl/Polyline.cpp"
 HEAP32[$259>>2] = $255; //@line 110 "modules/graphics/opengl/Polyline.cpp"
 STACKTOP = sp;return; //@line 112 "modules/graphics/opengl/Polyline.cpp"
}
function __ZNSt3__113__vector_baseIN4love6VectorENS_9allocatorIS2_EEED2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i = 0, $scevgep4$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ($0|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = $0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($1) {
  return; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $3 = ((($this)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$3>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($4|0)==($0|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($5)) {
  $scevgep$i$i = ((($4)) + -8|0);
  $6 = $scevgep$i$i;
  $7 = (($6) - ($2))|0;
  $8 = $7 >>> 3;
  $9 = $8 ^ -1;
  $scevgep4$i$i = (($4) + ($9<<3)|0);
  HEAP32[$3>>2] = $scevgep4$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($0); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl8Polyline4drawEv($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$arith = 0, $$arith2 = 0, $$overflow = 0, $$overflow3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $colors$0 = 0, $enabledattribs$0 = 0, $i$04 = 0, $indices$0 = 0, $or$cond = 0;
 var $total_vertex_count$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 12|0); //@line 368 "modules/graphics/opengl/Polyline.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 368 "modules/graphics/opengl/Polyline.cpp"
 $2 = ((($this)) + 8|0); //@line 369 "modules/graphics/opengl/Polyline.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 369 "modules/graphics/opengl/Polyline.cpp"
 $4 = ($3|0)==(0|0); //@line 369 "modules/graphics/opengl/Polyline.cpp"
 if ($4) {
  $total_vertex_count$0 = $1;
 } else {
  $5 = ((($this)) + 28|0); //@line 370 "modules/graphics/opengl/Polyline.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 370 "modules/graphics/opengl/Polyline.cpp"
  $7 = ((($this)) + 16|0); //@line 370 "modules/graphics/opengl/Polyline.cpp"
  $8 = HEAP32[$7>>2]|0; //@line 370 "modules/graphics/opengl/Polyline.cpp"
  $9 = (($8) + ($6))|0; //@line 370 "modules/graphics/opengl/Polyline.cpp"
  $total_vertex_count$0 = $9;
 }
 $10 = ((($this)) + 24|0); //@line 373 "modules/graphics/opengl/Polyline.cpp"
 $11 = HEAP8[$10>>0]|0; //@line 373 "modules/graphics/opengl/Polyline.cpp"
 $12 = ($11<<24>>24)==(0); //@line 373 "modules/graphics/opengl/Polyline.cpp"
 L4: do {
  if ($12) {
   $indices$0 = 0;
  } else {
   $13 = $total_vertex_count$0 >>> 2; //@line 375 "modules/graphics/opengl/Polyline.cpp"
   $14 = ($13*6)|0; //@line 375 "modules/graphics/opengl/Polyline.cpp"
   $$arith2 = $14<<1; //@line 379 "modules/graphics/opengl/Polyline.cpp"
   $$overflow3 = ($14>>>0)>(2147483647); //@line 379 "modules/graphics/opengl/Polyline.cpp"
   $15 = $$overflow3 ? -1 : $$arith2; //@line 379 "modules/graphics/opengl/Polyline.cpp"
   __THREW__ = 0;
   $16 = (invoke_ii(3492,($15|0))|0); //@line 379 "modules/graphics/opengl/Polyline.cpp"
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if (!($18)) {
    $19 = (($14>>>0) / 6)&-1; //@line 388 "modules/graphics/opengl/Polyline.cpp"
    $20 = ($14>>>0)>(5); //@line 388 "modules/graphics/opengl/Polyline.cpp"
    if ($20) {
     $i$04 = 0;
    } else {
     $indices$0 = $16;
     break;
    }
    while(1) {
     $35 = $i$04 << 2; //@line 391 "modules/graphics/opengl/Polyline.cpp"
     $36 = $35&65535; //@line 391 "modules/graphics/opengl/Polyline.cpp"
     $37 = ($i$04*6)|0; //@line 391 "modules/graphics/opengl/Polyline.cpp"
     $38 = (($16) + ($37<<1)|0); //@line 391 "modules/graphics/opengl/Polyline.cpp"
     HEAP16[$38>>1] = $36; //@line 391 "modules/graphics/opengl/Polyline.cpp"
     $39 = $35 | 1; //@line 392 "modules/graphics/opengl/Polyline.cpp"
     $40 = $39&65535; //@line 392 "modules/graphics/opengl/Polyline.cpp"
     $41 = $37 | 1; //@line 392 "modules/graphics/opengl/Polyline.cpp"
     $42 = (($16) + ($41<<1)|0); //@line 392 "modules/graphics/opengl/Polyline.cpp"
     HEAP16[$42>>1] = $40; //@line 392 "modules/graphics/opengl/Polyline.cpp"
     $43 = $35 | 2; //@line 393 "modules/graphics/opengl/Polyline.cpp"
     $44 = $43&65535; //@line 393 "modules/graphics/opengl/Polyline.cpp"
     $45 = (($37) + 2)|0; //@line 393 "modules/graphics/opengl/Polyline.cpp"
     $46 = (($16) + ($45<<1)|0); //@line 393 "modules/graphics/opengl/Polyline.cpp"
     HEAP16[$46>>1] = $44; //@line 393 "modules/graphics/opengl/Polyline.cpp"
     $47 = (($37) + 3)|0; //@line 396 "modules/graphics/opengl/Polyline.cpp"
     $48 = (($16) + ($47<<1)|0); //@line 396 "modules/graphics/opengl/Polyline.cpp"
     HEAP16[$48>>1] = $36; //@line 396 "modules/graphics/opengl/Polyline.cpp"
     $49 = (($37) + 4)|0; //@line 397 "modules/graphics/opengl/Polyline.cpp"
     $50 = (($16) + ($49<<1)|0); //@line 397 "modules/graphics/opengl/Polyline.cpp"
     HEAP16[$50>>1] = $44; //@line 397 "modules/graphics/opengl/Polyline.cpp"
     $51 = $35 | 3; //@line 398 "modules/graphics/opengl/Polyline.cpp"
     $52 = $51&65535; //@line 398 "modules/graphics/opengl/Polyline.cpp"
     $53 = (($37) + 5)|0; //@line 398 "modules/graphics/opengl/Polyline.cpp"
     $54 = (($16) + ($53<<1)|0); //@line 398 "modules/graphics/opengl/Polyline.cpp"
     HEAP16[$54>>1] = $52; //@line 398 "modules/graphics/opengl/Polyline.cpp"
     $55 = (($i$04) + 1)|0; //@line 388 "modules/graphics/opengl/Polyline.cpp"
     $56 = ($55>>>0)<($19>>>0); //@line 388 "modules/graphics/opengl/Polyline.cpp"
     if ($56) {
      $i$04 = $55;
     } else {
      $indices$0 = $16;
      break L4;
     }
    }
   }
   $21 = ___cxa_find_matching_catch_3(9056|0)|0;
   $22 = tempRet0;
   $23 = (_llvm_eh_typeid_for((9056|0))|0); //@line 380 "modules/graphics/opengl/Polyline.cpp"
   $24 = ($22|0)==($23|0); //@line 380 "modules/graphics/opengl/Polyline.cpp"
   if (!($24)) {
    $$1 = $22;$$12 = $21;
    ___resumeException($$12|0);
    // unreachable;
   }
   (___cxa_begin_catch(($21|0))|0); //@line 380 "modules/graphics/opengl/Polyline.cpp"
   $25 = (___cxa_allocate_exception(16)|0); //@line 383 "modules/graphics/opengl/Polyline.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($25|0),(507610|0),($vararg_buffer|0)); //@line 383 "modules/graphics/opengl/Polyline.cpp"
   $26 = __THREW__; __THREW__ = 0;
   $27 = $26&1;
   if ($27) {
    $31 = ___cxa_find_matching_catch_2()|0;
    $32 = tempRet0;
    ___cxa_free_exception(($25|0)); //@line 383 "modules/graphics/opengl/Polyline.cpp"
    $$0 = $32;$$01 = $31;
   } else {
    __THREW__ = 0;
    invoke_viii(3350,($25|0),(96|0),(344|0)); //@line 383 "modules/graphics/opengl/Polyline.cpp"
    $28 = __THREW__; __THREW__ = 0;
    $29 = ___cxa_find_matching_catch_2()|0;
    $30 = tempRet0;
    $$0 = $30;$$01 = $29;
   }
   __THREW__ = 0;
   invoke_v(3378); //@line 384 "modules/graphics/opengl/Polyline.cpp"
   $33 = __THREW__; __THREW__ = 0;
   $34 = $33&1;
   if ($34) {
    $85 = ___cxa_find_matching_catch_3(0|0)|0;
    $86 = tempRet0;
    ___clang_call_terminate($85); //@line 384 "modules/graphics/opengl/Polyline.cpp"
    // unreachable; //@line 384 "modules/graphics/opengl/Polyline.cpp"
   } else {
    $$1 = $$0;$$12 = $$01;
    ___resumeException($$12|0);
    // unreachable;
   }
  }
 } while(0);
 __ZN4love8graphics6opengl6OpenGL11prepareDrawEv(634044); //@line 402 "modules/graphics/opengl/Polyline.cpp"
 $57 = (__ZNK4love8graphics6opengl6OpenGL17getDefaultTextureEv(634044)|0); //@line 404 "modules/graphics/opengl/Polyline.cpp"
 __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$57); //@line 404 "modules/graphics/opengl/Polyline.cpp"
 $58 = HEAP32[$2>>2]|0; //@line 408 "modules/graphics/opengl/Polyline.cpp"
 $59 = ($58|0)==(0|0); //@line 408 "modules/graphics/opengl/Polyline.cpp"
 if ($59) {
  $colors$0 = 0;$enabledattribs$0 = 1;
 } else {
  $$arith = $total_vertex_count$0<<2; //@line 412 "modules/graphics/opengl/Polyline.cpp"
  $$overflow = ($total_vertex_count$0>>>0)>(1073741823); //@line 412 "modules/graphics/opengl/Polyline.cpp"
  $60 = $$overflow ? -1 : $$arith; //@line 412 "modules/graphics/opengl/Polyline.cpp"
  $61 = (__Znaj($60)|0); //@line 412 "modules/graphics/opengl/Polyline.cpp"
  $62 = ($total_vertex_count$0|0)==(0); //@line 412 "modules/graphics/opengl/Polyline.cpp"
  if (!($62)) {
   $63 = $total_vertex_count$0 << 2; //@line 412 "modules/graphics/opengl/Polyline.cpp"
   _memset(($61|0),0,($63|0))|0; //@line 37 "./modules/graphics/Color.h"
  }
  $64 = ((($this)) + 28|0); //@line 413 "modules/graphics/opengl/Polyline.cpp"
  $65 = HEAP32[$64>>2]|0; //@line 413 "modules/graphics/opengl/Polyline.cpp"
  $66 = $65 << 2; //@line 413 "modules/graphics/opengl/Polyline.cpp"
  _memset(($61|0),-1,($66|0))|0; //@line 413 "modules/graphics/opengl/Polyline.cpp"
  $67 = HEAP32[$this>>2]|0; //@line 414 "modules/graphics/opengl/Polyline.cpp"
  $68 = ((($67)) + 16|0); //@line 414 "modules/graphics/opengl/Polyline.cpp"
  $69 = HEAP32[$68>>2]|0; //@line 414 "modules/graphics/opengl/Polyline.cpp"
  $70 = (($61) + ($65<<2)|0); //@line 414 "modules/graphics/opengl/Polyline.cpp"
  FUNCTION_TABLE_vii[$69 & 8191]($this,$70); //@line 414 "modules/graphics/opengl/Polyline.cpp"
  $71 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiiiii[$71 & 8191](2,4,5121,1,0,$61); //@line 591 "./libraries/glad/gladfuncs.hpp"
  $colors$0 = $61;$enabledattribs$0 = 5;
 }
 __ZN4love8graphics6opengl6OpenGL21useVertexAttribArraysEj(634044,$enabledattribs$0); //@line 421 "modules/graphics/opengl/Polyline.cpp"
 $72 = ((($this)) + 4|0); //@line 423 "modules/graphics/opengl/Polyline.cpp"
 $73 = HEAP32[$72>>2]|0; //@line 423 "modules/graphics/opengl/Polyline.cpp"
 $74 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiiiii[$74 & 8191](0,2,5126,0,0,$73); //@line 591 "./libraries/glad/gladfuncs.hpp"
 $75 = HEAP8[$10>>0]|0; //@line 428 "modules/graphics/opengl/Polyline.cpp"
 $76 = ($75<<24>>24)==(0); //@line 428 "modules/graphics/opengl/Polyline.cpp"
 $77 = ((($this)) + 20|0); //@line 429 "modules/graphics/opengl/Polyline.cpp"
 $78 = HEAP32[$77>>2]|0; //@line 429 "modules/graphics/opengl/Polyline.cpp"
 if ($76) {
  __ZN4love8graphics6opengl6OpenGL10drawArraysEjii(634044,$78,0,$total_vertex_count$0); //@line 431 "modules/graphics/opengl/Polyline.cpp"
 } else {
  $79 = $total_vertex_count$0 >>> 2; //@line 429 "modules/graphics/opengl/Polyline.cpp"
  $80 = ($79*6)|0; //@line 429 "modules/graphics/opengl/Polyline.cpp"
  __ZN4love8graphics6opengl6OpenGL12drawElementsEjijPKv(634044,$78,$80,5123,$indices$0); //@line 429 "modules/graphics/opengl/Polyline.cpp"
 }
 $81 = HEAP32[$2>>2]|0; //@line 433 "modules/graphics/opengl/Polyline.cpp"
 $82 = ($81|0)==(0|0); //@line 433 "modules/graphics/opengl/Polyline.cpp"
 $83 = ($colors$0|0)==(0|0); //@line 434 "modules/graphics/opengl/Polyline.cpp"
 $or$cond = $83 | $82; //@line 433 "modules/graphics/opengl/Polyline.cpp"
 if (!($or$cond)) {
  __ZdaPv($colors$0); //@line 434 "modules/graphics/opengl/Polyline.cpp"
 }
 $84 = ($indices$0|0)==(0|0); //@line 436 "modules/graphics/opengl/Polyline.cpp"
 if ($84) {
  STACKTOP = sp;return; //@line 438 "modules/graphics/opengl/Polyline.cpp"
 }
 __ZdaPv($indices$0); //@line 437 "modules/graphics/opengl/Polyline.cpp"
 STACKTOP = sp;return; //@line 438 "modules/graphics/opengl/Polyline.cpp"
}
function __ZN4love8graphics6opengl6ShaderD2Ev($this) {
 $this = $this|0;
 var $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$01$i$i$i = 0, $$pre$i$i = 0, $$pre2$i$i = 0, $$pre4$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$013 = 0, $i$01$i$i = 0, $lpad$phi$index = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i4 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i5 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (25212); //@line 96 "modules/graphics/opengl/Shader.cpp"
 $0 = ((($this)) + 8|0); //@line 96 "modules/graphics/opengl/Shader.cpp"
 HEAP32[$0>>2] = (25244); //@line 96 "modules/graphics/opengl/Shader.cpp"
 $1 = HEAP32[158581]|0; //@line 97 "modules/graphics/opengl/Shader.cpp"
 $2 = ($1|0)==($this|0); //@line 97 "modules/graphics/opengl/Shader.cpp"
 L1: do {
  if ($2) {
   $3 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
   $4 = ($3|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
   if ($4) {
    $30 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
    __THREW__ = 0;
    invoke_vi($30|0,0); //@line 534 "./libraries/glad/gladfuncs.hpp"
    $31 = __THREW__; __THREW__ = 0;
    $32 = $31&1;
    if ($32) {
     label = 15;
     break;
    }
    HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
    label = 21;
    break;
   }
   $5 = ($3|0)==($this|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
   if ($5) {
    label = 21;
   } else {
    $6 = ((($3)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
    $7 = HEAP32[$6>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
    $8 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
    __THREW__ = 0;
    invoke_vi($8|0,($7|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
    $9 = __THREW__; __THREW__ = 0;
    $10 = $9&1;
    if ($10) {
     label = 15;
    } else {
     HEAP32[158581] = $3; //@line 409 "modules/graphics/opengl/Shader.cpp"
     $11 = ((($3)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
     $12 = ((($3)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $13 = HEAP32[$12>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $14 = HEAP32[$11>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $15 = ($13|0)==($14|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
     L9: do {
      if (!($15)) {
       $125 = $13;$17 = $14;$i$01$i$i = 0;
       while(1) {
        $16 = $17; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $18 = (($16) + ($i$01$i$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $19 = HEAP32[$18>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
        $20 = ($19|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
        $$pre4$i$i = (($i$01$i$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
        if ($20) {
         $24 = $125;$25 = $17;
        } else {
         __THREW__ = 0;
         invoke_viiii(4111,(634044|0),($19|0),($$pre4$i$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
         $21 = __THREW__; __THREW__ = 0;
         $22 = $21&1;
         if ($22) {
          break;
         }
         $$pre$i$i = HEAP32[$12>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $$pre2$i$i = HEAP32[$11>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $24 = $$pre$i$i;$25 = $$pre2$i$i;
        }
        $23 = (($24) - ($25))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $26 = $23 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $27 = ($$pre4$i$i>>>0)<($26>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
        if ($27) {
         $125 = $24;$17 = $25;$i$01$i$i = $$pre4$i$i;
        } else {
         break L9;
        }
       }
       $35 = ___cxa_find_matching_catch_3(0|0)|0;
       $36 = tempRet0;
       $lpad$phi$index = $35;
       break L1;
      }
     } while(0);
     __THREW__ = 0;
     invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
     $28 = __THREW__; __THREW__ = 0;
     $29 = $28&1;
     if ($29) {
      label = 15;
     } else {
      label = 21;
     }
    }
   }
  } else {
   label = 21;
  }
 } while(0);
 L18: do {
  if ((label|0) == 21) {
   $69 = ((($this)) + 152|0); //@line 100 "modules/graphics/opengl/Shader.cpp"
   $70 = HEAP32[$69>>2]|0; //@line 911 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $71 = ((($this)) + 156|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $72 = ($70|0)==($71|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   L20: do {
    if (!($72)) {
     $__begin$sroa$0$013 = $70;
     while(1) {
      $80 = ((($__begin$sroa$0$013)) + 28|0); //@line 101 "modules/graphics/opengl/Shader.cpp"
      $81 = HEAP32[$80>>2]|0; //@line 101 "modules/graphics/opengl/Shader.cpp"
      $82 = HEAP32[$81>>2]|0; //@line 101 "modules/graphics/opengl/Shader.cpp"
      $83 = ((($82)) + 12|0); //@line 101 "modules/graphics/opengl/Shader.cpp"
      $84 = HEAP32[$83>>2]|0; //@line 101 "modules/graphics/opengl/Shader.cpp"
      __THREW__ = 0;
      invoke_vi($84|0,($81|0)); //@line 101 "modules/graphics/opengl/Shader.cpp"
      $85 = __THREW__; __THREW__ = 0;
      $86 = $85&1;
      if ($86) {
       break;
      }
      $87 = ((($__begin$sroa$0$013)) + 4|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $88 = HEAP32[$87>>2]|0; //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $89 = ($88|0)==(0|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      if ($89) {
       $$0$i$i$i = $__begin$sroa$0$013;
       while(1) {
        $92 = ((($$0$i$i$i)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        $93 = HEAP32[$92>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        $94 = HEAP32[$93>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        $95 = ($94|0)==($$0$i$i$i|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        if ($95) {
         $$01$i$i$i = $93;
         break;
        } else {
         $$0$i$i$i = $93;
        }
       }
      } else {
       $$0$i$i$i$i = $88;
       while(1) {
        $90 = HEAP32[$$0$i$i$i$i>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        $91 = ($90|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
        if ($91) {
         $$01$i$i$i = $$0$i$i$i$i;
         break;
        } else {
         $$0$i$i$i$i = $90;
        }
       }
      }
      $96 = ($$01$i$i$i|0)==($71|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      if ($96) {
       break L20;
      } else {
       $__begin$sroa$0$013 = $$01$i$i$i;
      }
     }
     $33 = ___cxa_find_matching_catch_3(0|0)|0;
     $34 = tempRet0;
     $lpad$phi$index = $33;
     break L18;
    }
   } while(0);
   $73 = HEAP32[$71>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ObjectEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE7destroyEPNS_11__tree_nodeISB_PvEE($69,$73); //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $74 = ((($this)) + 160|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$74>>2] = 0; //@line 1478 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$69>>2] = $71; //@line 1479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$71>>2] = 0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $75 = HEAP32[$this>>2]|0; //@line 105 "modules/graphics/opengl/Shader.cpp"
   $76 = ((($75)) + 20|0); //@line 105 "modules/graphics/opengl/Shader.cpp"
   $77 = HEAP32[$76>>2]|0; //@line 105 "modules/graphics/opengl/Shader.cpp"
   __THREW__ = 0;
   invoke_vi($77|0,($this|0)); //@line 105 "modules/graphics/opengl/Shader.cpp"
   $78 = __THREW__; __THREW__ = 0;
   $79 = $78&1;
   if ($79) {
    label = 15;
   } else {
    $97 = ((($this)) + 252|0); //@line 106 "modules/graphics/opengl/Shader.cpp"
    __ZN4love7Matrix4D2Ev($97); //@line 106 "modules/graphics/opengl/Shader.cpp"
    $98 = ((($this)) + 188|0); //@line 106 "modules/graphics/opengl/Shader.cpp"
    __ZN4love7Matrix4D2Ev($98); //@line 106 "modules/graphics/opengl/Shader.cpp"
    $99 = HEAP32[$71>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ObjectEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE7destroyEPNS_11__tree_nodeISB_PvEE($69,$99); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $100 = ((($this)) + 140|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $101 = HEAP32[$100>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $102 = ($101|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $103 = $101; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($102) {
     $111 = ((($this)) + 128|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $112 = ((($this)) + 132|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $113 = HEAP32[$112>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($111,$113); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $114 = ((($this)) + 116|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $115 = ((($this)) + 120|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $116 = HEAP32[$115>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformEEENS_19__map_value_compareIS7_SD_NS_4lessIS7_EELb1EEENS5_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($114,$116); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $117 = ((($this)) + 104|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $118 = ((($this)) + 108|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $119 = HEAP32[$118>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($117,$119); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $120 = ((($this)) + 36|0); //@line 783 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $121 = ((($this)) + 40|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $122 = HEAP32[$121>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     __ZNSt3__16__treeINS_12__value_typeIN4love8graphics6opengl6Shader11ShaderStageENS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEEENS_19__map_value_compareIS6_SD_NS_4lessIS6_EELb1EEENSA_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($120,$122); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $123 = ((($this)) + 24|0); //@line 84 "modules/graphics/opengl/Shader.h"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($123); //@line 84 "modules/graphics/opengl/Shader.h"
     $124 = ((($this)) + 12|0); //@line 84 "modules/graphics/opengl/Shader.h"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($124); //@line 84 "modules/graphics/opengl/Shader.h"
     __ZN4love8graphics8VolatileD2Ev($0); //@line 106 "modules/graphics/opengl/Shader.cpp"
     return; //@line 106 "modules/graphics/opengl/Shader.cpp"
    }
    $104 = ((($this)) + 144|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $105 = HEAP32[$104>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $106 = ($105|0)==($101|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($106)) {
     $scevgep$i$i$i4 = ((($105)) + -4|0);
     $107 = $scevgep$i$i$i4;
     $108 = (($107) - ($103))|0;
     $109 = $108 >>> 2;
     $110 = $109 ^ -1;
     $scevgep4$i$i$i5 = (($105) + ($110<<2)|0);
     HEAP32[$104>>2] = $scevgep4$i$i$i5; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($101); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $111 = ((($this)) + 128|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $112 = ((($this)) + 132|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $113 = HEAP32[$112>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($111,$113); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $114 = ((($this)) + 116|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $115 = ((($this)) + 120|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $116 = HEAP32[$115>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformEEENS_19__map_value_compareIS7_SD_NS_4lessIS7_EELb1EEENS5_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($114,$116); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $117 = ((($this)) + 104|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $118 = ((($this)) + 108|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $119 = HEAP32[$118>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($117,$119); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $120 = ((($this)) + 36|0); //@line 783 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $121 = ((($this)) + 40|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $122 = HEAP32[$121>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    __ZNSt3__16__treeINS_12__value_typeIN4love8graphics6opengl6Shader11ShaderStageENS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEEENS_19__map_value_compareIS6_SD_NS_4lessIS6_EELb1EEENSA_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($120,$122); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $123 = ((($this)) + 24|0); //@line 84 "modules/graphics/opengl/Shader.h"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($123); //@line 84 "modules/graphics/opengl/Shader.h"
    $124 = ((($this)) + 12|0); //@line 84 "modules/graphics/opengl/Shader.h"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($124); //@line 84 "modules/graphics/opengl/Shader.h"
    __ZN4love8graphics8VolatileD2Ev($0); //@line 106 "modules/graphics/opengl/Shader.cpp"
    return; //@line 106 "modules/graphics/opengl/Shader.cpp"
   }
  }
 } while(0);
 if ((label|0) == 15) {
  $37 = ___cxa_find_matching_catch_3(0|0)|0;
  $38 = tempRet0;
  $lpad$phi$index = $37;
 }
 $39 = ((($this)) + 252|0); //@line 106 "modules/graphics/opengl/Shader.cpp"
 __ZN4love7Matrix4D2Ev($39); //@line 106 "modules/graphics/opengl/Shader.cpp"
 $40 = ((($this)) + 188|0); //@line 106 "modules/graphics/opengl/Shader.cpp"
 __ZN4love7Matrix4D2Ev($40); //@line 106 "modules/graphics/opengl/Shader.cpp"
 $41 = ((($this)) + 152|0); //@line 783 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $42 = ((($this)) + 156|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $43 = HEAP32[$42>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ObjectEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE7destroyEPNS_11__tree_nodeISB_PvEE($41,$43); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $44 = ((($this)) + 140|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $45 = HEAP32[$44>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $46 = ($45|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $47 = $45; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($46) {
  $55 = ((($this)) + 128|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $56 = ((($this)) + 132|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $57 = HEAP32[$56>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($55,$57); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $58 = ((($this)) + 116|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $59 = ((($this)) + 120|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $60 = HEAP32[$59>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformEEENS_19__map_value_compareIS7_SD_NS_4lessIS7_EELb1EEENS5_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($58,$60); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $61 = ((($this)) + 104|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $62 = ((($this)) + 108|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $63 = HEAP32[$62>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($61,$63); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $64 = ((($this)) + 36|0); //@line 783 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $65 = ((($this)) + 40|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $66 = HEAP32[$65>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIN4love8graphics6opengl6Shader11ShaderStageENS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEEENS_19__map_value_compareIS6_SD_NS_4lessIS6_EELb1EEENSA_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($64,$66); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $67 = ((($this)) + 24|0); //@line 84 "modules/graphics/opengl/Shader.h"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($67); //@line 84 "modules/graphics/opengl/Shader.h"
  $68 = ((($this)) + 12|0); //@line 84 "modules/graphics/opengl/Shader.h"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($68); //@line 84 "modules/graphics/opengl/Shader.h"
  __ZN4love8graphics8VolatileD2Ev($0); //@line 106 "modules/graphics/opengl/Shader.cpp"
  ___clang_call_terminate($lpad$phi$index); //@line 106 "modules/graphics/opengl/Shader.cpp"
  // unreachable; //@line 106 "modules/graphics/opengl/Shader.cpp"
 }
 $48 = ((($this)) + 144|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $49 = HEAP32[$48>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $50 = ($49|0)==($45|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($50)) {
  $scevgep$i$i$i = ((($49)) + -4|0);
  $51 = $scevgep$i$i$i;
  $52 = (($51) - ($47))|0;
  $53 = $52 >>> 2;
  $54 = $53 ^ -1;
  $scevgep4$i$i$i = (($49) + ($54<<2)|0);
  HEAP32[$48>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($45); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $55 = ((($this)) + 128|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $56 = ((($this)) + 132|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $57 = HEAP32[$56>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($55,$57); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $58 = ((($this)) + 116|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $59 = ((($this)) + 120|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $60 = HEAP32[$59>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformEEENS_19__map_value_compareIS7_SD_NS_4lessIS7_EELb1EEENS5_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($58,$60); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $61 = ((($this)) + 104|0); //@line 1119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $62 = ((($this)) + 108|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $63 = HEAP32[$62>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($61,$63); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $64 = ((($this)) + 36|0); //@line 783 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $65 = ((($this)) + 40|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $66 = HEAP32[$65>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeIN4love8graphics6opengl6Shader11ShaderStageENS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEEENS_19__map_value_compareIS6_SD_NS_4lessIS6_EELb1EEENSA_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($64,$66); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $67 = ((($this)) + 24|0); //@line 84 "modules/graphics/opengl/Shader.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($67); //@line 84 "modules/graphics/opengl/Shader.h"
 $68 = ((($this)) + 12|0); //@line 84 "modules/graphics/opengl/Shader.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($68); //@line 84 "modules/graphics/opengl/Shader.h"
 __ZN4love8graphics8VolatileD2Ev($0); //@line 106 "modules/graphics/opengl/Shader.cpp"
 ___clang_call_terminate($lpad$phi$index); //@line 106 "modules/graphics/opengl/Shader.cpp"
 // unreachable; //@line 106 "modules/graphics/opengl/Shader.cpp"
}
function __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ObjectEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE7destroyEPNS_11__tree_nodeISB_PvEE($this,$__nd) {
 $this = $this|0;
 $__nd = $__nd|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__nd|0)==(0|0); //@line 1440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($0) {
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 } else {
  $1 = HEAP32[$__nd>>2]|0; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ObjectEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE7destroyEPNS_11__tree_nodeISB_PvEE($this,$1); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2 = ((($__nd)) + 4|0); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3 = HEAP32[$2>>2]|0; //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ObjectEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE7destroyEPNS_11__tree_nodeISB_PvEE($this,$3); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4 = ((($__nd)) + 16|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
  __ZdlPv($__nd); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
}
function __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($this,$__nd) {
 $this = $this|0;
 $__nd = $__nd|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__nd|0)==(0|0); //@line 1440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($0) {
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 } else {
  $1 = HEAP32[$__nd>>2]|0; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($this,$1); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2 = ((($__nd)) + 4|0); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3 = HEAP32[$2>>2]|0; //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($this,$3); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4 = ((($__nd)) + 16|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
  __ZdlPv($__nd); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
}
function __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformEEENS_19__map_value_compareIS7_SD_NS_4lessIS7_EELb1EEENS5_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($this,$__nd) {
 $this = $this|0;
 $__nd = $__nd|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__nd|0)==(0|0); //@line 1440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($0) {
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 } else {
  $1 = HEAP32[$__nd>>2]|0; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformEEENS_19__map_value_compareIS7_SD_NS_4lessIS7_EELb1EEENS5_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($this,$1); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2 = ((($__nd)) + 4|0); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3 = HEAP32[$2>>2]|0; //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformEEENS_19__map_value_compareIS7_SD_NS_4lessIS7_EELb1EEENS5_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($this,$3); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4 = ((($__nd)) + 44|0); //@line 214 "modules/graphics/opengl/Shader.h"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 214 "modules/graphics/opengl/Shader.h"
  $5 = ((($__nd)) + 16|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
  __ZdlPv($__nd); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
}
function __ZNSt3__16__treeINS_12__value_typeIN4love8graphics6opengl6Shader11ShaderStageENS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEEENS_19__map_value_compareIS6_SD_NS_4lessIS6_EELb1EEENSA_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($this,$__nd) {
 $this = $this|0;
 $__nd = $__nd|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__nd|0)==(0|0); //@line 1440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($0) {
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 } else {
  $1 = HEAP32[$__nd>>2]|0; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIN4love8graphics6opengl6Shader11ShaderStageENS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEEENS_19__map_value_compareIS6_SD_NS_4lessIS6_EELb1EEENSA_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($this,$1); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2 = ((($__nd)) + 4|0); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3 = HEAP32[$2>>2]|0; //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeIN4love8graphics6opengl6Shader11ShaderStageENS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEEENS_19__map_value_compareIS6_SD_NS_4lessIS6_EELb1EEENSA_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($this,$3); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4 = ((($__nd)) + 20|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
  __ZdlPv($__nd); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
}
function __ZN4love8graphics6opengl6ShaderD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl6ShaderD2Ev($this); //@line 96 "modules/graphics/opengl/Shader.cpp"
 __ZdlPv($this); //@line 96 "modules/graphics/opengl/Shader.cpp"
 return; //@line 106 "modules/graphics/opengl/Shader.cpp"
}
function __ZN4love8graphics6opengl6Shader12loadVolatileEv($this) {
 $this = $this|0;
 var $$ = 0, $$0 = 0, $$02 = 0, $$1 = 0, $$13 = 0, $$byval_copy = 0, $$pre = 0, $$pre$i = 0, $$pre2$i = 0, $$pre4$i = 0, $$sink = 0, $$sink1 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0;
 var $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0;
 var $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0;
 var $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0;
 var $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0;
 var $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$062 = 0, $__begin11$sroa$0$056 = 0, $__begin2$sroa$0$053 = 0, $__begin6$sroa$0$060 = 0, $i$01$i = 0, $lpad$phi$index = 0, $lpad$phi$index14 = 0, $lpad$phi52$index = 0, $lpad$phi52$index4 = 0, $scevgep = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i4 = 0, $scevgep$i$i$i8 = 0;
 var $scevgep4$i$i$i = 0, $scevgep4$i$i$i5 = 0, $scevgep4$i$i$i9 = 0, $shaderids = 0, $status = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $warnings = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 56|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 52|0;
 $1 = sp + 48|0;
 $shaderids = sp + 36|0;
 $2 = sp + 32|0;
 $3 = sp + 28|0;
 $status = sp + 24|0;
 $warnings = sp + 12|0;
 $4 = ((($this)) + 164|0); //@line 234 "modules/graphics/opengl/Shader.cpp"
 HEAP32[$4>>2] = (-1); //@line 234 "modules/graphics/opengl/Shader.cpp"
 $5 = ((($this)) + 168|0); //@line 235 "modules/graphics/opengl/Shader.cpp"
 ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;HEAP32[$5+12>>2]=0|0; //@line 235 "modules/graphics/opengl/Shader.cpp"
 $6 = ((($this)) + 184|0); //@line 237 "modules/graphics/opengl/Shader.cpp"
 HEAPF32[$6>>2] = -1.0; //@line 237 "modules/graphics/opengl/Shader.cpp"
 $7 = ((($this)) + 252|0); //@line 241 "modules/graphics/opengl/Shader.cpp"
 __ZN4love7Matrix414setTranslationEff($7,nan,nan); //@line 241 "modules/graphics/opengl/Shader.cpp"
 $8 = ((($this)) + 188|0); //@line 242 "modules/graphics/opengl/Shader.cpp"
 __ZN4love7Matrix414setTranslationEff($8,nan,nan); //@line 242 "modules/graphics/opengl/Shader.cpp"
 $scevgep = ((($this)) + 316|0); //@line 244 "modules/graphics/opengl/Shader.cpp"
 ;HEAP32[$scevgep>>2]=0|0;HEAP32[$scevgep+4>>2]=0|0;HEAP32[$scevgep+8>>2]=0|0; //@line 245 "modules/graphics/opengl/Shader.cpp"
 $9 = ((($this)) + 140|0); //@line 248 "modules/graphics/opengl/Shader.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ((($this)) + 144|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ($12|0)==($10|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = $10; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($13)) {
  $scevgep$i$i$i4 = ((($12)) + -4|0);
  $15 = $scevgep$i$i$i4;
  $16 = (($15) - ($14))|0;
  $17 = $16 >>> 2;
  $18 = $17 ^ -1;
  $scevgep4$i$i$i5 = (($12) + ($18<<2)|0);
  HEAP32[$11>>2] = $scevgep4$i$i$i5; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 HEAP32[$0>>2] = $14; //@line 1128 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $19 = (__ZNK4love8graphics6opengl6OpenGL18getMaxTextureUnitsEv(634044)|0); //@line 249 "modules/graphics/opengl/Shader.cpp"
 $20 = (($19) + -1)|0; //@line 249 "modules/graphics/opengl/Shader.cpp"
 HEAP32[$1>>2] = 0; //@line 249 "modules/graphics/opengl/Shader.cpp"
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0; //@line 249 "modules/graphics/opengl/Shader.cpp"
 (__ZNSt3__16vectorIjNS_9allocatorIjEEE6insertENS_11__wrap_iterIPKjEEjRS5_($9,$$byval_copy,$20,$1)|0); //@line 249 "modules/graphics/opengl/Shader.cpp"
 HEAP32[$shaderids>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = ((($shaderids)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$21>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $22 = ((($shaderids)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$22>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $23 = HEAP8[1252012]|0; //@line 254 "modules/graphics/opengl/Shader.cpp"
 $24 = ($23<<24>>24)!=(0); //@line 254 "modules/graphics/opengl/Shader.cpp"
 $$ = $24 ? (634360) : 634336; //@line 254 "modules/graphics/opengl/Shader.cpp"
 $25 = ((($this)) + 12|0); //@line 258 "modules/graphics/opengl/Shader.cpp"
 $26 = HEAP8[$25>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $27 = $26 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $28 = ($27<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($28) {
  $31 = $26&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $32 = $31 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $34 = $32;
 } else {
  $29 = ((($this)) + 16|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $30 = HEAP32[$29>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $34 = $30;
 }
 $33 = ($34|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $$sink = $33 ? $$ : $25; //@line 258 "modules/graphics/opengl/Shader.cpp"
 $35 = ((($this)) + 24|0); //@line 259 "modules/graphics/opengl/Shader.cpp"
 $36 = HEAP8[$35>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $37 = $36 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $38 = ($37<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($38) {
  $41 = $36&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $42 = $41 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $44 = $42;
 } else {
  $39 = ((($this)) + 28|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $40 = HEAP32[$39>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $44 = $40;
 }
 $43 = ($44|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $$sink1 = $43 ? $$ : $25; //@line 259 "modules/graphics/opengl/Shader.cpp"
 $45 = ((($$sink1)) + 12|0); //@line 259 "modules/graphics/opengl/Shader.cpp"
 __THREW__ = 0;
 $46 = (invoke_iiii(4113,($this|0),0,($$sink|0))|0); //@line 263 "modules/graphics/opengl/Shader.cpp"
 $47 = __THREW__; __THREW__ = 0;
 $48 = $47&1;
 L12: do {
  if ($48) {
   label = 17;
  } else {
   HEAP32[$2>>2] = $46; //@line 263 "modules/graphics/opengl/Shader.cpp"
   $49 = HEAP32[$21>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $50 = ((($shaderids)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $51 = HEAP32[$50>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $52 = ($49>>>0)<($51>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($52) {
    HEAP32[$49>>2] = $46; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $53 = ((($49)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$21>>2] = $53; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   } else {
    __THREW__ = 0;
    invoke_vii(3597,($shaderids|0),($2|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $54 = __THREW__; __THREW__ = 0;
    $55 = $54&1;
    if ($55) {
     label = 17;
     break;
    }
   }
   __THREW__ = 0;
   $56 = (invoke_iiii(4113,($this|0),1,($45|0))|0); //@line 264 "modules/graphics/opengl/Shader.cpp"
   $57 = __THREW__; __THREW__ = 0;
   $58 = $57&1;
   if ($58) {
    label = 17;
   } else {
    HEAP32[$3>>2] = $56; //@line 264 "modules/graphics/opengl/Shader.cpp"
    $59 = HEAP32[$21>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $60 = HEAP32[$50>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $61 = ($59>>>0)<($60>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($61) {
     HEAP32[$59>>2] = $56; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $62 = ((($59)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$21>>2] = $62; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    } else {
     __THREW__ = 0;
     invoke_vii(3597,($shaderids|0),($3|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $63 = __THREW__; __THREW__ = 0;
     $64 = $63&1;
     if ($64) {
      label = 17;
      break;
     }
    }
    $85 = HEAP32[165365]|0; //@line 507 "./libraries/glad/gladfuncs.hpp"
    __THREW__ = 0;
    $86 = (invoke_i($85|0)|0); //@line 507 "./libraries/glad/gladfuncs.hpp"
    $87 = __THREW__; __THREW__ = 0;
    $88 = $87&1;
    L23: do {
     if ($88) {
      label = 35;
     } else {
      $89 = ((($this)) + 48|0); //@line 273 "modules/graphics/opengl/Shader.cpp"
      HEAP32[$89>>2] = $86; //@line 273 "modules/graphics/opengl/Shader.cpp"
      $90 = ($86|0)==(0); //@line 275 "modules/graphics/opengl/Shader.cpp"
      $91 = HEAP32[$shaderids>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $92 = HEAP32[$21>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $93 = ($91|0)==($92|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($90) {
       L27: do {
        if (!($93)) {
         $__begin2$sroa$0$053 = $91;
         while(1) {
          $111 = HEAP32[$__begin2$sroa$0$053>>2]|0; //@line 277 "modules/graphics/opengl/Shader.cpp"
          $112 = HEAP32[165085]|0; //@line 510 "./libraries/glad/gladfuncs.hpp"
          __THREW__ = 0;
          invoke_vi($112|0,($111|0)); //@line 510 "./libraries/glad/gladfuncs.hpp"
          $113 = __THREW__; __THREW__ = 0;
          $114 = $113&1;
          if ($114) {
           break;
          }
          $115 = ((($__begin2$sroa$0$053)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
          $116 = ($115|0)==($92|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
          if ($116) {
           break L27;
          } else {
           $__begin2$sroa$0$053 = $115;
          }
         }
         $97 = ___cxa_find_matching_catch_2()|0;
         $98 = tempRet0;
         $lpad$phi$index = $97;$lpad$phi$index14 = $98;
         break L23;
        }
       } while(0);
       $94 = (___cxa_allocate_exception(16)|0); //@line 279 "modules/graphics/opengl/Shader.cpp"
       __THREW__ = 0;
       invoke_viii(3349,($94|0),(438124|0),($vararg_buffer|0)); //@line 279 "modules/graphics/opengl/Shader.cpp"
       $95 = __THREW__; __THREW__ = 0;
       $96 = $95&1;
       if ($96) {
        $118 = ___cxa_find_matching_catch_2()|0;
        $119 = tempRet0;
        ___cxa_free_exception(($94|0)); //@line 279 "modules/graphics/opengl/Shader.cpp"
        $$1 = $118;$$13 = $119;
        break L12;
       } else {
        __THREW__ = 0;
        invoke_viii(3350,($94|0),(96|0),(344|0)); //@line 279 "modules/graphics/opengl/Shader.cpp"
        $117 = __THREW__; __THREW__ = 0;
        label = 35;
        break;
       }
      }
      L37: do {
       if (!($93)) {
        $122 = $86;$__begin6$sroa$0$060 = $91;
        while(1) {
         $120 = HEAP32[$__begin6$sroa$0$060>>2]|0; //@line 282 "modules/graphics/opengl/Shader.cpp"
         $121 = HEAP32[165223]|0; //@line 504 "./libraries/glad/gladfuncs.hpp"
         __THREW__ = 0;
         invoke_vii($121|0,($122|0),($120|0)); //@line 504 "./libraries/glad/gladfuncs.hpp"
         $123 = __THREW__; __THREW__ = 0;
         $124 = $123&1;
         if ($124) {
          break;
         }
         $125 = ((($__begin6$sroa$0$060)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
         $126 = ($125|0)==($92|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
         if ($126) {
          break L37;
         }
         $$pre = HEAP32[$89>>2]|0; //@line 283 "modules/graphics/opengl/Shader.cpp"
         $122 = $$pre;$__begin6$sroa$0$060 = $125;
        }
        $107 = ___cxa_find_matching_catch_2()|0;
        $108 = tempRet0;
        $lpad$phi$index = $107;$lpad$phi$index14 = $108;
        break L23;
       }
      } while(0);
      $127 = HEAP32[(634596)>>2]|0; //@line 94 "./common/StringMap.h"
      $128 = ($127|0)==(0|0); //@line 94 "./common/StringMap.h"
      if ($128) {
       label = 95;
      } else {
       $129 = HEAP32[$89>>2]|0; //@line 290 "modules/graphics/opengl/Shader.cpp"
       $130 = HEAP32[165504]|0; //@line 505 "./libraries/glad/gladfuncs.hpp"
       __THREW__ = 0;
       invoke_viii($130|0,($129|0),0,($127|0)); //@line 505 "./libraries/glad/gladfuncs.hpp"
       $131 = __THREW__; __THREW__ = 0;
       $132 = $131&1;
       if (!($132)) {
        label = 95;
       }
      }
      do {
       if ((label|0) == 95) {
        $252 = HEAP32[(634600)>>2]|0; //@line 94 "./common/StringMap.h"
        $253 = ($252|0)==(0|0); //@line 94 "./common/StringMap.h"
        if (!($253)) {
         $254 = HEAP32[$89>>2]|0; //@line 290 "modules/graphics/opengl/Shader.cpp"
         $255 = HEAP32[165504]|0; //@line 505 "./libraries/glad/gladfuncs.hpp"
         __THREW__ = 0;
         invoke_viii($255|0,($254|0),1,($252|0)); //@line 505 "./libraries/glad/gladfuncs.hpp"
         $256 = __THREW__; __THREW__ = 0;
         $257 = $256&1;
         if ($257) {
          break;
         }
        }
        $258 = HEAP32[(634604)>>2]|0; //@line 94 "./common/StringMap.h"
        $259 = ($258|0)==(0|0); //@line 94 "./common/StringMap.h"
        if (!($259)) {
         $260 = HEAP32[$89>>2]|0; //@line 290 "modules/graphics/opengl/Shader.cpp"
         $261 = HEAP32[165504]|0; //@line 505 "./libraries/glad/gladfuncs.hpp"
         __THREW__ = 0;
         invoke_viii($261|0,($260|0),2,($258|0)); //@line 505 "./libraries/glad/gladfuncs.hpp"
         $262 = __THREW__; __THREW__ = 0;
         $263 = $262&1;
         if ($263) {
          break;
         }
        }
        $264 = HEAP32[(634608)>>2]|0; //@line 94 "./common/StringMap.h"
        $265 = ($264|0)==(0|0); //@line 94 "./common/StringMap.h"
        if (!($265)) {
         $266 = HEAP32[$89>>2]|0; //@line 290 "modules/graphics/opengl/Shader.cpp"
         $267 = HEAP32[165504]|0; //@line 505 "./libraries/glad/gladfuncs.hpp"
         __THREW__ = 0;
         invoke_viii($267|0,($266|0),3,($264|0)); //@line 505 "./libraries/glad/gladfuncs.hpp"
         $268 = __THREW__; __THREW__ = 0;
         $269 = $268&1;
         if ($269) {
          break;
         }
        }
        $270 = HEAP32[$89>>2]|0; //@line 293 "modules/graphics/opengl/Shader.cpp"
        $271 = HEAP32[165133]|0; //@line 532 "./libraries/glad/gladfuncs.hpp"
        __THREW__ = 0;
        invoke_vi($271|0,($270|0)); //@line 532 "./libraries/glad/gladfuncs.hpp"
        $272 = __THREW__; __THREW__ = 0;
        $273 = $272&1;
        if ($273) {
         label = 35;
         break L23;
        }
        $133 = HEAP32[$shaderids>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $134 = HEAP32[$21>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $135 = ($133|0)==($134|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        L58: do {
         if (!($135)) {
          $__begin11$sroa$0$056 = $133;
          while(1) {
           $140 = HEAP32[$__begin11$sroa$0$056>>2]|0; //@line 296 "modules/graphics/opengl/Shader.cpp"
           $141 = HEAP32[165085]|0; //@line 510 "./libraries/glad/gladfuncs.hpp"
           __THREW__ = 0;
           invoke_vi($141|0,($140|0)); //@line 510 "./libraries/glad/gladfuncs.hpp"
           $142 = __THREW__; __THREW__ = 0;
           $143 = $142&1;
           if ($143) {
            break;
           }
           $144 = ((($__begin11$sroa$0$056)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
           $145 = ($144|0)==($134|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
           if ($145) {
            break L58;
           } else {
            $__begin11$sroa$0$056 = $144;
           }
          }
          $103 = ___cxa_find_matching_catch_2()|0;
          $104 = tempRet0;
          $lpad$phi$index = $103;$lpad$phi$index14 = $104;
          break L23;
         }
        } while(0);
        $136 = HEAP32[$89>>2]|0; //@line 300 "modules/graphics/opengl/Shader.cpp"
        $137 = HEAP32[165573]|0; //@line 518 "./libraries/glad/gladfuncs.hpp"
        __THREW__ = 0;
        invoke_viii($137|0,($136|0),35714,($status|0)); //@line 518 "./libraries/glad/gladfuncs.hpp"
        $138 = __THREW__; __THREW__ = 0;
        $139 = $138&1;
        if ($139) {
         label = 35;
         break L23;
        }
        $146 = HEAP32[$status>>2]|0; //@line 302 "modules/graphics/opengl/Shader.cpp"
        $147 = ($146|0)==(0); //@line 302 "modules/graphics/opengl/Shader.cpp"
        if ($147) {
         __THREW__ = 0;
         invoke_vii(4114,($warnings|0),($this|0)); //@line 304 "modules/graphics/opengl/Shader.cpp"
         $148 = __THREW__; __THREW__ = 0;
         $149 = $148&1;
         if ($149) {
          label = 35;
          break L23;
         }
         $150 = HEAP32[$89>>2]|0; //@line 305 "modules/graphics/opengl/Shader.cpp"
         $151 = HEAP32[164697]|0; //@line 509 "./libraries/glad/gladfuncs.hpp"
         __THREW__ = 0;
         invoke_vi($151|0,($150|0)); //@line 509 "./libraries/glad/gladfuncs.hpp"
         $152 = __THREW__; __THREW__ = 0;
         $153 = $152&1;
         do {
          if ($153) {
           label = 55;
          } else {
           HEAP32[$89>>2] = 0; //@line 306 "modules/graphics/opengl/Shader.cpp"
           $154 = (___cxa_allocate_exception(16)|0); //@line 307 "modules/graphics/opengl/Shader.cpp"
           $155 = HEAP8[$warnings>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $156 = $155 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $157 = ($156<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $158 = ((($warnings)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $159 = HEAP32[$158>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $160 = ((($warnings)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $161 = $157 ? $160 : $159; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           __THREW__ = 0;
           HEAP32[$vararg_buffer1>>2] = $161; //@line 307 "modules/graphics/opengl/Shader.cpp"
           invoke_viii(3349,($154|0),(438161|0),($vararg_buffer1|0)); //@line 307 "modules/graphics/opengl/Shader.cpp"
           $162 = __THREW__; __THREW__ = 0;
           $163 = $162&1;
           if ($163) {
            $167 = ___cxa_find_matching_catch_2()|0;
            $168 = tempRet0;
            ___cxa_free_exception(($154|0)); //@line 307 "modules/graphics/opengl/Shader.cpp"
            $$0 = $167;$$02 = $168;
            break;
           } else {
            __THREW__ = 0;
            invoke_viii(3350,($154|0),(96|0),(344|0)); //@line 307 "modules/graphics/opengl/Shader.cpp"
            $164 = __THREW__; __THREW__ = 0;
            label = 55;
            break;
           }
          }
         } while(0);
         if ((label|0) == 55) {
          $165 = ___cxa_find_matching_catch_2()|0;
          $166 = tempRet0;
          $$0 = $165;$$02 = $166;
         }
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($warnings); //@line 308 "modules/graphics/opengl/Shader.cpp"
         $$1 = $$0;$$13 = $$02;
         break L12;
        }
        __THREW__ = 0;
        invoke_vi(4115,($this|0)); //@line 311 "modules/graphics/opengl/Shader.cpp"
        $169 = __THREW__; __THREW__ = 0;
        $170 = $169&1;
        if ($170) {
         label = 35;
         break L23;
        }
        $171 = HEAP32[(634596)>>2]|0; //@line 94 "./common/StringMap.h"
        $172 = ($171|0)==(0|0); //@line 94 "./common/StringMap.h"
        do {
         if ($172) {
          $179 = ((($this)) + 88|0); //@line 319 "modules/graphics/opengl/Shader.cpp"
          HEAP32[$179>>2] = -1; //@line 319 "modules/graphics/opengl/Shader.cpp"
          label = 82;
         } else {
          $173 = HEAP32[$89>>2]|0; //@line 317 "modules/graphics/opengl/Shader.cpp"
          $174 = HEAP32[164909]|0; //@line 517 "./libraries/glad/gladfuncs.hpp"
          __THREW__ = 0;
          $175 = (invoke_iii($174|0,($173|0),($171|0))|0); //@line 517 "./libraries/glad/gladfuncs.hpp"
          $176 = __THREW__; __THREW__ = 0;
          $177 = $176&1;
          if ($177) {
           break;
          }
          $178 = ((($this)) + 88|0); //@line 317 "modules/graphics/opengl/Shader.cpp"
          HEAP32[$178>>2] = $175; //@line 317 "modules/graphics/opengl/Shader.cpp"
          label = 82;
         }
        } while(0);
        do {
         if ((label|0) == 82) {
          $223 = HEAP32[(634600)>>2]|0; //@line 94 "./common/StringMap.h"
          $224 = ($223|0)==(0|0); //@line 94 "./common/StringMap.h"
          if ($224) {
           $231 = ((($this)) + 92|0); //@line 319 "modules/graphics/opengl/Shader.cpp"
           HEAP32[$231>>2] = -1; //@line 319 "modules/graphics/opengl/Shader.cpp"
          } else {
           $225 = HEAP32[$89>>2]|0; //@line 317 "modules/graphics/opengl/Shader.cpp"
           $226 = HEAP32[164909]|0; //@line 517 "./libraries/glad/gladfuncs.hpp"
           __THREW__ = 0;
           $227 = (invoke_iii($226|0,($225|0),($223|0))|0); //@line 517 "./libraries/glad/gladfuncs.hpp"
           $228 = __THREW__; __THREW__ = 0;
           $229 = $228&1;
           if ($229) {
            break;
           }
           $230 = ((($this)) + 92|0); //@line 317 "modules/graphics/opengl/Shader.cpp"
           HEAP32[$230>>2] = $227; //@line 317 "modules/graphics/opengl/Shader.cpp"
          }
          $232 = HEAP32[(634604)>>2]|0; //@line 94 "./common/StringMap.h"
          $233 = ($232|0)==(0|0); //@line 94 "./common/StringMap.h"
          if ($233) {
           $240 = ((($this)) + 96|0); //@line 319 "modules/graphics/opengl/Shader.cpp"
           HEAP32[$240>>2] = -1; //@line 319 "modules/graphics/opengl/Shader.cpp"
          } else {
           $234 = HEAP32[$89>>2]|0; //@line 317 "modules/graphics/opengl/Shader.cpp"
           $235 = HEAP32[164909]|0; //@line 517 "./libraries/glad/gladfuncs.hpp"
           __THREW__ = 0;
           $236 = (invoke_iii($235|0,($234|0),($232|0))|0); //@line 517 "./libraries/glad/gladfuncs.hpp"
           $237 = __THREW__; __THREW__ = 0;
           $238 = $237&1;
           if ($238) {
            break;
           }
           $239 = ((($this)) + 96|0); //@line 317 "modules/graphics/opengl/Shader.cpp"
           HEAP32[$239>>2] = $236; //@line 317 "modules/graphics/opengl/Shader.cpp"
          }
          $241 = HEAP32[(634608)>>2]|0; //@line 94 "./common/StringMap.h"
          $242 = ($241|0)==(0|0); //@line 94 "./common/StringMap.h"
          if ($242) {
           $249 = ((($this)) + 100|0); //@line 319 "modules/graphics/opengl/Shader.cpp"
           HEAP32[$249>>2] = -1; //@line 319 "modules/graphics/opengl/Shader.cpp"
          } else {
           $243 = HEAP32[$89>>2]|0; //@line 317 "modules/graphics/opengl/Shader.cpp"
           $244 = HEAP32[164909]|0; //@line 517 "./libraries/glad/gladfuncs.hpp"
           __THREW__ = 0;
           $245 = (invoke_iii($244|0,($243|0),($241|0))|0); //@line 517 "./libraries/glad/gladfuncs.hpp"
           $246 = __THREW__; __THREW__ = 0;
           $247 = $246&1;
           if ($247) {
            break;
           }
           $248 = ((($this)) + 100|0); //@line 317 "modules/graphics/opengl/Shader.cpp"
           HEAP32[$248>>2] = $245; //@line 317 "modules/graphics/opengl/Shader.cpp"
          }
          $250 = HEAP32[158581]|0; //@line 322 "modules/graphics/opengl/Shader.cpp"
          $251 = ($250|0)==($this|0); //@line 322 "modules/graphics/opengl/Shader.cpp"
          if ($251) {
           HEAP32[158581] = 0; //@line 325 "modules/graphics/opengl/Shader.cpp"
           $180 = HEAP32[$89>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
           $181 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
           __THREW__ = 0;
           invoke_vi($181|0,($180|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
           $182 = __THREW__; __THREW__ = 0;
           $183 = $182&1;
           if ($183) {
            label = 35;
            break L23;
           }
           HEAP32[158581] = $this; //@line 409 "modules/graphics/opengl/Shader.cpp"
           $184 = HEAP32[$11>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $185 = HEAP32[$9>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $186 = ($184|0)==($185|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
           L102: do {
            if (!($186)) {
             $190 = $185;$274 = $184;$i$01$i = 0;
             while(1) {
              $189 = $190; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $191 = (($189) + ($i$01$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $192 = HEAP32[$191>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
              $193 = ($192|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
              $$pre4$i = (($i$01$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
              if ($193) {
               $197 = $274;$198 = $190;
              } else {
               __THREW__ = 0;
               invoke_viiii(4111,(634044|0),($192|0),($$pre4$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
               $194 = __THREW__; __THREW__ = 0;
               $195 = $194&1;
               if ($195) {
                break;
               }
               $$pre$i = HEAP32[$11>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
               $$pre2$i = HEAP32[$9>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
               $197 = $$pre$i;$198 = $$pre2$i;
              }
              $196 = (($197) - ($198))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $199 = $196 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $200 = ($$pre4$i>>>0)<($199>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
              if ($200) {
               $190 = $198;$274 = $197;$i$01$i = $$pre4$i;
              } else {
               break L102;
              }
             }
             $99 = ___cxa_find_matching_catch_2()|0;
             $100 = tempRet0;
             $lpad$phi$index = $99;$lpad$phi$index14 = $100;
             break L23;
            }
           } while(0);
           __THREW__ = 0;
           invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
           $187 = __THREW__; __THREW__ = 0;
           $188 = $187&1;
           if ($188) {
            label = 35;
            break L23;
           }
           __THREW__ = 0;
           invoke_vi(4116,($this|0)); //@line 327 "modules/graphics/opengl/Shader.cpp"
           $201 = __THREW__; __THREW__ = 0;
           $202 = $201&1;
           if ($202) {
            label = 35;
            break L23;
           }
          }
          $203 = HEAP32[$shaderids>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $204 = ($203|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $205 = $203; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          if ($204) {
           STACKTOP = sp;return 1; //@line 331 "modules/graphics/opengl/Shader.cpp"
          }
          $206 = HEAP32[$21>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $207 = ($206|0)==($203|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          if (!($207)) {
           $scevgep$i$i$i8 = ((($206)) + -4|0);
           $208 = $scevgep$i$i$i8;
           $209 = (($208) - ($205))|0;
           $210 = $209 >>> 2;
           $211 = $210 ^ -1;
           $scevgep4$i$i$i9 = (($206) + ($211<<2)|0);
           HEAP32[$21>>2] = $scevgep4$i$i$i9; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          }
          __ZdlPv($203); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
          STACKTOP = sp;return 1; //@line 331 "modules/graphics/opengl/Shader.cpp"
         }
        } while(0);
        $101 = ___cxa_find_matching_catch_2()|0;
        $102 = tempRet0;
        $lpad$phi$index = $101;$lpad$phi$index14 = $102;
        break L23;
       }
      } while(0);
      $105 = ___cxa_find_matching_catch_2()|0;
      $106 = tempRet0;
      $lpad$phi$index = $105;$lpad$phi$index14 = $106;
     }
    } while(0);
    if ((label|0) == 35) {
     $109 = ___cxa_find_matching_catch_2()|0;
     $110 = tempRet0;
     $lpad$phi$index = $109;$lpad$phi$index14 = $110;
    }
    $$1 = $lpad$phi$index;$$13 = $lpad$phi$index14;
   }
  }
 } while(0);
 if ((label|0) == 17) {
  $65 = ___cxa_find_matching_catch_3(96|0)|0;
  $66 = tempRet0;
  $67 = (_llvm_eh_typeid_for((96|0))|0); //@line 265 "modules/graphics/opengl/Shader.cpp"
  $68 = ($66|0)==($67|0); //@line 265 "modules/graphics/opengl/Shader.cpp"
  if ($68) {
   (___cxa_begin_catch(($65|0))|0); //@line 265 "modules/graphics/opengl/Shader.cpp"
   $69 = HEAP32[$shaderids>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $70 = HEAP32[$21>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $71 = ($69|0)==($70|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   L128: do {
    if ($71) {
     label = 19;
    } else {
     $__begin$sroa$0$062 = $69;
     while(1) {
      $75 = HEAP32[$__begin$sroa$0$062>>2]|0; //@line 268 "modules/graphics/opengl/Shader.cpp"
      $76 = HEAP32[165085]|0; //@line 510 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_vi($76|0,($75|0)); //@line 510 "./libraries/glad/gladfuncs.hpp"
      $77 = __THREW__; __THREW__ = 0;
      $78 = $77&1;
      if ($78) {
       break;
      }
      $79 = ((($__begin$sroa$0$062)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $80 = ($79|0)==($70|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($80) {
       label = 19;
       break L128;
      } else {
       $__begin$sroa$0$062 = $79;
      }
     }
     $81 = ___cxa_find_matching_catch_2()|0;
     $82 = tempRet0;
     $lpad$phi52$index = $81;$lpad$phi52$index4 = $82;
    }
   } while(0);
   if ((label|0) == 19) {
    __THREW__ = 0;
    invoke_v(3656); //@line 270 "modules/graphics/opengl/Shader.cpp"
    $72 = __THREW__; __THREW__ = 0;
    $73 = ___cxa_find_matching_catch_2()|0;
    $74 = tempRet0;
    $lpad$phi52$index = $73;$lpad$phi52$index4 = $74;
   }
   __THREW__ = 0;
   invoke_v(3378); //@line 271 "modules/graphics/opengl/Shader.cpp"
   $83 = __THREW__; __THREW__ = 0;
   $84 = $83&1;
   if ($84) {
    $221 = ___cxa_find_matching_catch_3(0|0)|0;
    $222 = tempRet0;
    ___clang_call_terminate($221); //@line 271 "modules/graphics/opengl/Shader.cpp"
    // unreachable; //@line 271 "modules/graphics/opengl/Shader.cpp"
   } else {
    $$1 = $lpad$phi52$index;$$13 = $lpad$phi52$index4;
   }
  } else {
   $$1 = $65;$$13 = $66;
  }
 }
 $212 = HEAP32[$shaderids>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $213 = ($212|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $214 = $212; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($213) {
  ___resumeException($$1|0);
  // unreachable;
 }
 $215 = HEAP32[$21>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $216 = ($215|0)==($212|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($216)) {
  $scevgep$i$i$i = ((($215)) + -4|0);
  $217 = $scevgep$i$i$i;
  $218 = (($217) - ($214))|0;
  $219 = $218 >>> 2;
  $220 = $219 ^ -1;
  $scevgep4$i$i$i = (($215) + ($220<<2)|0);
  HEAP32[$21>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($212); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__16vectorIjNS_9allocatorIjEEE6insertENS_11__wrap_iterIPKjEEjRS5_($this,$__position,$__n,$__x) {
 $this = $this|0;
 $__position = $__position|0;
 $__n = $__n|0;
 $__x = $__x|0;
 var $$ = 0, $$0$i = 0, $$0$i17 = 0, $$0$i18 = 0, $$0$i3 = 0, $$012$i$i = 0, $$03$i$i = 0, $$06 = 0, $$pre = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__i$01$i = 0, $__p$0 = 0, $__xr$0 = 0, $scevgep$i = 0, $scevgep$i4 = 0, $scevgep$i5 = 0, $scevgep5$i = 0, $uglygep$i = 0, $uglygep4$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 1849 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = $0; //@line 1849 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$__position>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $3 = (($2) - ($1))|0; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $4 = $3 >> 2; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $5 = (($0) + ($4<<2)|0); //@line 1849 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($__n|0)==(0); //@line 1850 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  $__p$0 = $5;
  return ($__p$0|0); //@line 1882 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $7 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 1852 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ((($this)) + 4|0); //@line 1852 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = HEAP32[$9>>2]|0; //@line 1852 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $10; //@line 1852 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = (($8) - ($11))|0; //@line 1852 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = $12 >> 2; //@line 1852 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = ($13>>>0)<($__n>>>0); //@line 1852 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($14) {
  $52 = (($11) - ($1))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $53 = $52 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $54 = (($53) + ($__n))|0; //@line 1876 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $55 = ($54>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($55) {
   __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$pre = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$pre9 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $56 = $$pre9; //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $58 = $$pre;$59 = $$pre9;$85 = $56;
  } else {
   $58 = $8;$59 = $1;$85 = $0;
  }
  $57 = (($58) - ($59))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $60 = $57 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $61 = ($60>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($61) {
   $65 = $57 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $66 = ($65>>>0)<($54>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $67 = $66 ? $54 : $65; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $68 = $5; //@line 1876 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $69 = (($68) - ($59))|0; //@line 1876 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $70 = $69 >> 2; //@line 1876 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $71 = ($67|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($71) {
    $$0$i18 = 0;$75 = 0;$76 = $70;$87 = $68;
   } else {
    $$0$i17 = $67;$98 = $70;$99 = $68;
    label = 20;
   }
  } else {
   $62 = $5; //@line 1876 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $63 = (($62) - ($59))|0; //@line 1876 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $64 = $63 >> 2; //@line 1876 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$0$i17 = 1073741823;$98 = $64;$99 = $62;
   label = 20;
  }
  if ((label|0) == 20) {
   $72 = $$0$i17 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $73 = (__Znwj($72)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $$0$i18 = $$0$i17;$75 = $73;$76 = $98;$87 = $99;
  }
  $74 = (($75) + ($76<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $77 = (($75) + ($$0$i18<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $$0$i = $__n;$79 = $74;
  while(1) {
   $78 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$79>>2] = $78; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $80 = ((($79)) + 4|0); //@line 237 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $81 = (($$0$i) + -1)|0; //@line 238 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $82 = ($81|0)==(0); //@line 239 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($82) {
    break;
   } else {
    $$0$i = $81;$79 = $80;
   }
  }
  $83 = $77; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $scevgep$i = (($74) + ($__n<<2)|0); //@line 234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $84 = $85; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $86 = (($87) - ($84))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $88 = $86 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $89 = (0 - ($88))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $90 = (($74) + ($89<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $91 = $90; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  _memcpy(($90|0),($85|0),($86|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $92 = HEAP32[$9>>2]|0; //@line 913 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $93 = (($92) - ($87))|0; //@line 1522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $94 = $93 >> 2; //@line 1522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  _memcpy(($scevgep$i|0),($5|0),($93|0))|0; //@line 1523 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $95 = (($scevgep$i) + ($94<<2)|0); //@line 1524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $96 = $95; //@line 1524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$this>>2] = $91; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$9>>2] = $96; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$7>>2] = $83; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $97 = ($85|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($97) {
   $__p$0 = $74;
   return ($__p$0|0); //@line 1882 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($85); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $__p$0 = $74;
  return ($__p$0|0); //@line 1882 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $15 = $5; //@line 1856 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = (($11) - ($15))|0; //@line 1856 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = $16 >> 2; //@line 1856 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $18 = ($17>>>0)<($__n>>>0); //@line 1856 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($18) {
  $19 = (($__n) - ($17))|0; //@line 1858 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i3 = $19;$21 = $10;
  while(1) {
   $20 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$21>>2] = $20; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $22 = ((($21)) + 4|0); //@line 1011 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $23 = (($$0$i3) + -1)|0; //@line 1012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $24 = ($23|0)==(0); //@line 1014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($24) {
    break;
   } else {
    $$0$i3 = $23;$21 = $22;
   }
  }
  $scevgep$i4 = (($10) + ($19<<2)|0); //@line 1007 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$9>>2] = $scevgep$i4; //@line 1010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $25 = ($17|0)==(0); //@line 1862 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($25) {
   $__p$0 = $5;
   return ($__p$0|0); //@line 1882 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  } else {
   $$06 = $17;$28 = $scevgep$i4;
  }
 } else {
  $$06 = $__n;$28 = $10;
 }
 $26 = (($5) + ($__n<<2)|0); //@line 1865 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = $28; //@line 1710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $29 = $26; //@line 1710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = (($27) - ($29))|0; //@line 1710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30 >> 2; //@line 1710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($5) + ($31<<2)|0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $33 = ($32>>>0)<($10>>>0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $34 = (0 - ($31))|0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($33) {
  $scevgep$i5 = (($10) + ($34<<2)|0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $35 = $15 ^ -1; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $uglygep$i = (($scevgep$i5) + ($35)|0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $uglygep4$i = $uglygep$i; //@line 1865 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $36 = $uglygep4$i >>> 2; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $39 = $28;$__i$01$i = $32;
  while(1) {
   $38 = HEAP32[$__i$01$i>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$39>>2] = $38; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $40 = ((($__i$01$i)) + 4|0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $41 = ((($39)) + 4|0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $42 = ($40>>>0)<($10>>>0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($42) {
    $39 = $41;$__i$01$i = $40;
   } else {
    break;
   }
  }
  $37 = (($36) + 1)|0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $scevgep5$i = (($28) + ($37<<2)|0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$9>>2] = $scevgep5$i; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $43 = (($28) + ($34<<2)|0); //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 _memmove(($43|0),($5|0),($30|0))|0; //@line 1929 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $44 = ($5>>>0)>($__x>>>0); //@line 1868 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($44) {
  $__xr$0 = $__x;
 } else {
  $45 = HEAP32[$9>>2]|0; //@line 1868 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $46 = ($45>>>0)>($__x>>>0); //@line 1868 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $47 = (($__x) + ($__n<<2)|0); //@line 1869 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$ = $46 ? $47 : $__x; //@line 1868 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $__xr$0 = $$;
 }
 $$012$i$i = $$06;$$03$i$i = $5;
 while(1) {
  $48 = HEAP32[$__xr$0>>2]|0; //@line 2033 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  HEAP32[$$03$i$i>>2] = $48; //@line 2033 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $49 = ((($$03$i$i)) + 4|0); //@line 2032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $50 = (($$012$i$i) + -1)|0; //@line 2032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $51 = ($50|0)==(0); //@line 2032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if ($51) {
   $__p$0 = $5;
   break;
  } else {
   $$012$i$i = $50;$$03$i$i = $49;
  }
 }
 return ($__p$0|0); //@line 1882 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl6Shader11compileCodeENS2_11ShaderStageERKNSt3__112basic_stringIcNS4_11char_traitsIcEENS4_9allocatorIcEEEE($this,$stage,$code) {
 $this = $this|0;
 $stage = $stage|0;
 $code = $code|0;
 var $$0 = 0, $$0$i7$i = 0, $$0$i7$i11 = 0, $$01 = 0, $$lcssa = 0, $$lcssa54 = 0, $$lcssa57 = 0, $$lcssa58 = 0, $$pre$i$i = 0, $$pre$i$i13 = 0, $$sink$i$i = 0, $$sink$i$i6 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $__nd$0$i$i = 0, $__nd$0$i$i$lcssa = 0, $__nd$0$i$i$lcssa52 = 0, $__nd$0$i$i$lcssa53 = 0, $__nd$0$i$i5 = 0, $__nd$0$i$i5$lcssa = 0, $__nd$0$i$i5$lcssa55 = 0, $__nd$0$i$i5$lcssa56 = 0, $__parent$i = 0, $__parent$i3 = 0, $__r$0$i = 0, $__r$0$i15 = 0, $glstage$0 = 0, $infologlen = 0, $src = 0, $srclen = 0, $status = 0;
 var $typestr$1 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, $vararg_buffer7 = 0, $vararg_ptr10 = 0, $vararg_ptr3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $__parent$i3 = sp + 52|0;
 $__parent$i = sp + 48|0;
 $src = sp + 44|0;
 $srclen = sp + 40|0;
 $infologlen = sp + 36|0;
 $status = sp + 32|0;
 $0 = ($stage>>>0)>(1); //@line 91 "./common/StringMap.h"
 if ($0) {
  label = 3;
 } else {
  $1 = (634492 + ($stage<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   label = 3;
  } else {
   $typestr$1 = $2;
  }
 }
 if ((label|0) == 3) {
  $typestr$1 = 1282466;
 }
 switch ($stage|0) {
 case 0:  {
  $glstage$0 = 35633;
  break;
 }
 case 1:  {
  $glstage$0 = 35632;
  break;
 }
 default: {
  $4 = (___cxa_allocate_exception(16)|0); //@line 125 "modules/graphics/opengl/Shader.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($4|0),(437950|0),($vararg_buffer|0)); //@line 125 "modules/graphics/opengl/Shader.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   ___cxa_throw(($4|0),(96|0),(344|0)); //@line 125 "modules/graphics/opengl/Shader.cpp"
   // unreachable; //@line 125 "modules/graphics/opengl/Shader.cpp"
  }
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  ___cxa_free_exception(($4|0)); //@line 125 "modules/graphics/opengl/Shader.cpp"
  $$0 = $7;$$01 = $8;
  ___resumeException($$0|0);
  // unreachable;
 }
 }
 $9 = HEAP32[164672]|0; //@line 508 "./libraries/glad/gladfuncs.hpp"
 $10 = (FUNCTION_TABLE_ii[$9 & 8191]($glstage$0)|0); //@line 508 "./libraries/glad/gladfuncs.hpp"
 $11 = ($10|0)==(0); //@line 131 "modules/graphics/opengl/Shader.cpp"
 if ($11) {
  $12 = HEAP32[165571]|0; //@line 71 "./libraries/glad/gladfuncs.hpp"
  $13 = (FUNCTION_TABLE_i[$12 & 8191]()|0); //@line 71 "./libraries/glad/gladfuncs.hpp"
  $14 = ($13|0)==(1280); //@line 133 "modules/graphics/opengl/Shader.cpp"
  $15 = (___cxa_allocate_exception(16)|0); //@line 134 "modules/graphics/opengl/Shader.cpp"
  if ($14) {
   __THREW__ = 0;
   HEAP32[$vararg_buffer1>>2] = $typestr$1; //@line 134 "modules/graphics/opengl/Shader.cpp"
   $vararg_ptr3 = ((($vararg_buffer1)) + 4|0); //@line 134 "modules/graphics/opengl/Shader.cpp"
   HEAP32[$vararg_ptr3>>2] = $typestr$1; //@line 134 "modules/graphics/opengl/Shader.cpp"
   invoke_viii(3349,($15|0),(438000|0),($vararg_buffer1|0)); //@line 134 "modules/graphics/opengl/Shader.cpp"
   $16 = __THREW__; __THREW__ = 0;
   $17 = $16&1;
   if (!($17)) {
    ___cxa_throw(($15|0),(96|0),(344|0)); //@line 134 "modules/graphics/opengl/Shader.cpp"
    // unreachable; //@line 134 "modules/graphics/opengl/Shader.cpp"
   }
   $18 = ___cxa_find_matching_catch_2()|0;
   $19 = tempRet0;
   ___cxa_free_exception(($15|0)); //@line 134 "modules/graphics/opengl/Shader.cpp"
   $$0 = $18;$$01 = $19;
   ___resumeException($$0|0);
   // unreachable;
  } else {
   __THREW__ = 0;
   HEAP32[$vararg_buffer4>>2] = $typestr$1; //@line 136 "modules/graphics/opengl/Shader.cpp"
   invoke_viii(3349,($15|0),(438058|0),($vararg_buffer4|0)); //@line 136 "modules/graphics/opengl/Shader.cpp"
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    ___cxa_throw(($15|0),(96|0),(344|0)); //@line 136 "modules/graphics/opengl/Shader.cpp"
    // unreachable; //@line 136 "modules/graphics/opengl/Shader.cpp"
   }
   $22 = ___cxa_find_matching_catch_2()|0;
   $23 = tempRet0;
   ___cxa_free_exception(($15|0)); //@line 136 "modules/graphics/opengl/Shader.cpp"
   $$0 = $22;$$01 = $23;
   ___resumeException($$0|0);
   // unreachable;
  }
 }
 $24 = HEAP8[$code>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $25 = $24 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $26 = ($25<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $27 = ((($code)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $28 = HEAP32[$27>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $29 = ((($code)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $30 = $26 ? $29 : $28; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 HEAP32[$src>>2] = $30; //@line 139 "modules/graphics/opengl/Shader.cpp"
 $31 = ((($code)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $32 = HEAP32[$31>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $33 = $24&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $34 = $33 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $35 = $26 ? $34 : $32; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 HEAP32[$srclen>>2] = $35; //@line 140 "modules/graphics/opengl/Shader.cpp"
 $36 = HEAP32[165539]|0; //@line 533 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiii[$36 & 8191]($10,1,$src,$srclen); //@line 533 "./libraries/glad/gladfuncs.hpp"
 $37 = HEAP32[164646]|0; //@line 506 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$37 & 8191]($10); //@line 506 "./libraries/glad/gladfuncs.hpp"
 $38 = HEAP32[164990]|0; //@line 520 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$38 & 8191]($10,35716,$infologlen); //@line 520 "./libraries/glad/gladfuncs.hpp"
 $39 = HEAP32[$infologlen>>2]|0; //@line 152 "modules/graphics/opengl/Shader.cpp"
 $40 = ($39|0)>(0); //@line 149 "modules/graphics/opengl/Shader.cpp"
 if ($40) {
  $41 = (__Znaj($39)|0); //@line 151 "modules/graphics/opengl/Shader.cpp"
  $42 = HEAP32[164795]|0; //@line 521 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiii[$42 & 8191]($10,$39,0,$41); //@line 521 "./libraries/glad/gladfuncs.hpp"
  $43 = ((($this)) + 36|0); //@line 155 "modules/graphics/opengl/Shader.cpp"
  $44 = ((($this)) + 40|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $45 = HEAP32[$44>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $46 = ($45|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  do {
   if ($46) {
    HEAP32[$__parent$i3>>2] = $44; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $58 = $44; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i11 = $44;$66 = $58;
    label = 28;
   } else {
    $__nd$0$i$i5 = $45;
    while(1) {
     $47 = ((($__nd$0$i$i5)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $48 = HEAP32[$47>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     $49 = ($48|0)>($stage|0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if ($49) {
      $50 = HEAP32[$__nd$0$i$i5>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
      $51 = ($50|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
      if ($51) {
       $$lcssa58 = $__nd$0$i$i5;$__nd$0$i$i5$lcssa56 = $__nd$0$i$i5;
       label = 21;
       break;
      } else {
       $$sink$i$i6 = $50;
      }
     } else {
      $53 = ($48|0)<($stage|0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
      if (!($53)) {
       $__nd$0$i$i5$lcssa = $__nd$0$i$i5;
       label = 27;
       break;
      }
      $54 = ((($__nd$0$i$i5)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
      $55 = HEAP32[$54>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
      $56 = ($55|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
      if ($56) {
       $$lcssa57 = $54;$__nd$0$i$i5$lcssa55 = $__nd$0$i$i5;
       label = 25;
       break;
      } else {
       $$sink$i$i6 = $55;
      }
     }
     $__nd$0$i$i5 = $$sink$i$i6;
    }
    if ((label|0) == 21) {
     HEAP32[$__parent$i3>>2] = $__nd$0$i$i5$lcssa56; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $52 = $__nd$0$i$i5$lcssa56; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $$0$i7$i11 = $$lcssa58;$66 = $52;
     label = 28;
     break;
    }
    else if ((label|0) == 25) {
     HEAP32[$__parent$i3>>2] = $__nd$0$i$i5$lcssa55; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $57 = $__nd$0$i$i5$lcssa55; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $$0$i7$i11 = $$lcssa57;$66 = $57;
     label = 28;
     break;
    }
    else if ((label|0) == 27) {
     HEAP32[$__parent$i3>>2] = $__nd$0$i$i5$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $59 = $__nd$0$i$i5$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $60 = ($__nd$0$i$i5$lcssa|0)==(0|0); //@line 1378 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($60) {
      $$0$i7$i11 = $__parent$i3;$66 = $59;
      label = 28;
      break;
     } else {
      $__r$0$i15 = $__nd$0$i$i5$lcssa;
      break;
     }
    }
   }
  } while(0);
  if ((label|0) == 28) {
   $61 = (__Znwj(32)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $62 = ((($61)) + 16|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   HEAP32[$62>>2] = $stage; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $63 = ((($61)) + 20|0); //@line 1366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   ;HEAP32[$63>>2]=0|0;HEAP32[$63+4>>2]=0|0;HEAP32[$63+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP32[$61>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $64 = ((($61)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$64>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $65 = ((($61)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$65>>2] = $66; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$$0$i7$i11>>2] = $61; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $67 = HEAP32[$43>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $68 = HEAP32[$67>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $69 = ($68|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($69) {
    $73 = $61;
   } else {
    $70 = $68; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP32[$43>>2] = $70; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $$pre$i$i13 = HEAP32[$$0$i7$i11>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $73 = $$pre$i$i13;
   }
   $71 = ((($this)) + 40|0); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $72 = HEAP32[$71>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($72,$73); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $74 = ((($this)) + 44|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $75 = HEAP32[$74>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $76 = (($75) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$74>>2] = $76; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $__r$0$i15 = $61;
  }
  $77 = ((($__r$0$i15)) + 20|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  (__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc($77,$41)|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  __ZdaPv($41); //@line 157 "modules/graphics/opengl/Shader.cpp"
 }
 $78 = HEAP32[164990]|0; //@line 520 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$78 & 8191]($10,35713,$status); //@line 520 "./libraries/glad/gladfuncs.hpp"
 $79 = HEAP32[$status>>2]|0; //@line 163 "modules/graphics/opengl/Shader.cpp"
 $80 = ($79|0)==(0); //@line 163 "modules/graphics/opengl/Shader.cpp"
 if (!($80)) {
  STACKTOP = sp;return ($10|0); //@line 170 "modules/graphics/opengl/Shader.cpp"
 }
 $81 = HEAP32[165085]|0; //@line 510 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$81 & 8191]($10); //@line 510 "./libraries/glad/gladfuncs.hpp"
 $82 = (___cxa_allocate_exception(16)|0); //@line 166 "modules/graphics/opengl/Shader.cpp"
 $83 = ((($this)) + 36|0); //@line 167 "modules/graphics/opengl/Shader.cpp"
 $84 = ((($this)) + 40|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $85 = HEAP32[$84>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $86 = ($85|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 do {
  if ($86) {
   HEAP32[$__parent$i>>2] = $84; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $98 = $84; //@line 1235 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $$0$i7$i = $84;$108 = $98;
   label = 43;
  } else {
   $__nd$0$i$i = $85;
   while(1) {
    $87 = ((($__nd$0$i$i)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $88 = HEAP32[$87>>2]|0; //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    $89 = ($88|0)>($stage|0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
    if ($89) {
     $90 = HEAP32[$__nd$0$i$i>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $91 = ($90|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($91) {
      $$lcssa54 = $__nd$0$i$i;$__nd$0$i$i$lcssa53 = $__nd$0$i$i;
      label = 36;
      break;
     } else {
      $$sink$i$i = $90;
     }
    } else {
     $93 = ($88|0)<($stage|0); //@line 63 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__functional_base"
     if (!($93)) {
      $__nd$0$i$i$lcssa = $__nd$0$i$i;
      label = 42;
      break;
     }
     $94 = ((($__nd$0$i$i)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $95 = HEAP32[$94>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $96 = ($95|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     if ($96) {
      $$lcssa = $94;$__nd$0$i$i$lcssa52 = $__nd$0$i$i;
      label = 40;
      break;
     } else {
      $$sink$i$i = $95;
     }
    }
    $__nd$0$i$i = $$sink$i$i;
   }
   if ((label|0) == 36) {
    HEAP32[$__parent$i>>2] = $__nd$0$i$i$lcssa53; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $92 = $__nd$0$i$i$lcssa53; //@line 1214 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i = $$lcssa54;$108 = $92;
    label = 43;
    break;
   }
   else if ((label|0) == 40) {
    HEAP32[$__parent$i>>2] = $__nd$0$i$i$lcssa52; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $97 = $__nd$0$i$i$lcssa52; //@line 1224 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $$0$i7$i = $$lcssa;$108 = $97;
    label = 43;
    break;
   }
   else if ((label|0) == 42) {
    HEAP32[$__parent$i>>2] = $__nd$0$i$i$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $99 = $__nd$0$i$i$lcssa; //@line 1230 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    $100 = ($__nd$0$i$i$lcssa|0)==(0|0); //@line 1378 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
    if ($100) {
     $$0$i7$i = $__parent$i;$108 = $99;
     label = 43;
     break;
    } else {
     $__r$0$i = $__nd$0$i$i$lcssa;
     label = 47;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 43) {
  __THREW__ = 0;
  $101 = (invoke_ii(3348,32)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $102 = __THREW__; __THREW__ = 0;
  $103 = $102&1;
  if (!($103)) {
   $104 = ((($101)) + 16|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   HEAP32[$104>>2] = $stage; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $105 = ((($101)) + 20|0); //@line 1366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   ;HEAP32[$105>>2]=0|0;HEAP32[$105+4>>2]=0|0;HEAP32[$105+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP32[$101>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $106 = ((($101)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$106>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $107 = ((($101)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$107>>2] = $108; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$$0$i7$i>>2] = $101; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $109 = HEAP32[$83>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $110 = HEAP32[$109>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $111 = ($110|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($111) {
    $115 = $101;
   } else {
    $112 = $110; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    HEAP32[$83>>2] = $112; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $$pre$i$i = HEAP32[$$0$i7$i>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $115 = $$pre$i$i;
   }
   $113 = ((($this)) + 40|0); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $114 = HEAP32[$113>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($114,$115); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $116 = ((($this)) + 44|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $117 = HEAP32[$116>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $118 = (($117) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$116>>2] = $118; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $__r$0$i = $101;
   label = 47;
  }
 }
 if ((label|0) == 47) {
  $119 = ((($__r$0$i)) + 20|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $120 = HEAP8[$119>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $121 = $120 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $122 = ($121<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($122) {
   $125 = ((($119)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $126 = $125;
  } else {
   $123 = ((($__r$0$i)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $124 = HEAP32[$123>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $126 = $124;
  }
  __THREW__ = 0;
  HEAP32[$vararg_buffer7>>2] = $typestr$1; //@line 166 "modules/graphics/opengl/Shader.cpp"
  $vararg_ptr10 = ((($vararg_buffer7)) + 4|0); //@line 166 "modules/graphics/opengl/Shader.cpp"
  HEAP32[$vararg_ptr10>>2] = $126; //@line 166 "modules/graphics/opengl/Shader.cpp"
  invoke_viii(3349,($82|0),(438090|0),($vararg_buffer7|0)); //@line 166 "modules/graphics/opengl/Shader.cpp"
  $127 = __THREW__; __THREW__ = 0;
  $128 = $127&1;
  if (!($128)) {
   ___cxa_throw(($82|0),(96|0),(344|0)); //@line 166 "modules/graphics/opengl/Shader.cpp"
   // unreachable; //@line 166 "modules/graphics/opengl/Shader.cpp"
  }
 }
 $129 = ___cxa_find_matching_catch_2()|0;
 $130 = tempRet0;
 ___cxa_free_exception(($82|0)); //@line 166 "modules/graphics/opengl/Shader.cpp"
 $$0 = $129;$$01 = $130;
 ___resumeException($$0|0);
 // unreachable;
 return (0)|0;
}
function __ZNK4love8graphics6opengl6Shader18getProgramWarningsEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nullpos = 0, $strsize = 0, $warnings = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $strsize = sp + 16|0;
 $nullpos = sp + 12|0;
 $warnings = sp;
 $0 = ((($this)) + 48|0); //@line 370 "modules/graphics/opengl/Shader.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 370 "modules/graphics/opengl/Shader.cpp"
 $2 = HEAP32[165573]|0; //@line 518 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$2 & 8191]($1,35716,$strsize); //@line 518 "./libraries/glad/gladfuncs.hpp"
 $3 = HEAP32[$strsize>>2]|0; //@line 377 "modules/graphics/opengl/Shader.cpp"
 $4 = ($3|0)==(0); //@line 372 "modules/graphics/opengl/Shader.cpp"
 if ($4) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($agg$result,1282466,0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  STACKTOP = sp;return; //@line 385 "modules/graphics/opengl/Shader.cpp"
 } else {
  $5 = ($3|0)>(-1); //@line 375 "modules/graphics/opengl/Shader.cpp"
  $6 = $5 ? $3 : -1; //@line 375 "modules/graphics/opengl/Shader.cpp"
  $7 = (__Znaj($6)|0); //@line 375 "modules/graphics/opengl/Shader.cpp"
  _memset(($7|0),0,($3|0))|0; //@line 377 "modules/graphics/opengl/Shader.cpp"
  $8 = HEAP32[$0>>2]|0; //@line 378 "modules/graphics/opengl/Shader.cpp"
  $9 = HEAP32[165061]|0; //@line 519 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_viiii[$9 & 8191]($8,$3,$nullpos,$7); //@line 519 "./libraries/glad/gladfuncs.hpp"
  $10 = HEAP32[$nullpos>>2]|0; //@line 379 "modules/graphics/opengl/Shader.cpp"
  $11 = (($7) + ($10)|0); //@line 379 "modules/graphics/opengl/Shader.cpp"
  HEAP8[$11>>0] = 0; //@line 379 "modules/graphics/opengl/Shader.cpp"
  $12 = (_strlen($7)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($warnings,$7,$12); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  __ZdaPv($7); //@line 382 "modules/graphics/opengl/Shader.cpp"
  ;HEAP32[$agg$result>>2]=HEAP32[$warnings>>2]|0;HEAP32[$agg$result+4>>2]=HEAP32[$warnings+4>>2]|0;HEAP32[$agg$result+8>>2]=HEAP32[$warnings+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  ;HEAP32[$warnings>>2]=0|0;HEAP32[$warnings+4>>2]=0|0;HEAP32[$warnings+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($warnings); //@line 385 "modules/graphics/opengl/Shader.cpp"
  STACKTOP = sp;return; //@line 385 "modules/graphics/opengl/Shader.cpp"
 }
}
function __ZN4love8graphics6opengl6Shader17mapActiveUniformsEv($this) {
 $this = $this|0;
 var $$0$i$ph = 0, $$011$i$i = 0, $$0110$i$i = 0, $$016$i$i$i = 0, $$016$i$i$i$lcssa = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$lcssa40 = 0, $$pre = 0, $$pre22 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__m1$0$i$i$i = 0, $__m2$0$i$i$i = 0, $activeprogram = 0, $cname = 0, $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $i1$015 = 0, $namelen = 0, $notlhs = 0, $notrhs = 0, $numuniforms = 0, $or$cond$i$i = 0, $or$cond$not = 0, $or$cond28 = 0, $scevgep = 0;
 var $switch$split12D = 0, $switch$split2D = 0, $switch$split42D = 0, $switch$split72D = 0, $u = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 320|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $activeprogram = sp + 48|0;
 $numuniforms = sp + 44|0;
 $cname = sp + 56|0;
 $namelen = sp + 40|0;
 $u = sp + 12|0;
 $0 = sp;
 $scevgep = ((($this)) + 52|0); //@line 176 "modules/graphics/opengl/Shader.cpp"
 dest=$scevgep; stop=dest+36|0; do { HEAP32[dest>>2]=4294967295|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 177 "modules/graphics/opengl/Shader.cpp"
 $1 = ((($this)) + 116|0); //@line 179 "modules/graphics/opengl/Shader.cpp"
 $2 = ((($this)) + 120|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformEEENS_19__map_value_compareIS7_SD_NS_4lessIS7_EELb1EEENS5_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($1,$3); //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $4 = ((($this)) + 124|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$4>>2] = 0; //@line 1478 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$1>>2] = $2; //@line 1479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$2>>2] = 0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$activeprogram>>2] = 0; //@line 181 "modules/graphics/opengl/Shader.cpp"
 $5 = HEAP32[165587]|0; //@line 73 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$5 & 8191](35725,$activeprogram); //@line 73 "./libraries/glad/gladfuncs.hpp"
 $6 = ((($this)) + 48|0); //@line 184 "modules/graphics/opengl/Shader.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 184 "modules/graphics/opengl/Shader.cpp"
 $8 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$8 & 8191]($7); //@line 534 "./libraries/glad/gladfuncs.hpp"
 $9 = HEAP32[$6>>2]|0; //@line 187 "modules/graphics/opengl/Shader.cpp"
 $10 = HEAP32[165573]|0; //@line 518 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viii[$10 & 8191]($9,35718,$numuniforms); //@line 518 "./libraries/glad/gladfuncs.hpp"
 $11 = HEAP32[$numuniforms>>2]|0; //@line 192 "modules/graphics/opengl/Shader.cpp"
 $12 = ($11|0)>(0); //@line 192 "modules/graphics/opengl/Shader.cpp"
 if (!($12)) {
  $20 = HEAP32[$activeprogram>>2]|0; //@line 226 "modules/graphics/opengl/Shader.cpp"
  $21 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$21 & 8191]($20); //@line 534 "./libraries/glad/gladfuncs.hpp"
  STACKTOP = sp;return; //@line 227 "modules/graphics/opengl/Shader.cpp"
 }
 $13 = ((($u)) + 4|0); //@line 195 "modules/graphics/opengl/Shader.cpp"
 $14 = ((($u)) + 8|0); //@line 195 "modules/graphics/opengl/Shader.cpp"
 $15 = ((($u)) + 12|0); //@line 195 "modules/graphics/opengl/Shader.cpp"
 $16 = ((($u)) + 16|0); //@line 195 "modules/graphics/opengl/Shader.cpp"
 $17 = ((($16)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $18 = ((($u)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $19 = ((($u)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $i1$015 = 0;
 L4: while(1) {
  HEAP32[$namelen>>2] = 0; //@line 194 "modules/graphics/opengl/Shader.cpp"
  ;HEAP32[$u>>2]=0|0;HEAP32[$u+4>>2]=0|0;HEAP32[$u+8>>2]=0|0;HEAP32[$u+12>>2]=0|0;HEAP32[$u+16>>2]=0|0;HEAP32[$u+20>>2]=0|0;HEAP32[$u+24>>2]=0|0; //@line 195 "modules/graphics/opengl/Shader.cpp"
  $22 = HEAP32[$6>>2]|0; //@line 197 "modules/graphics/opengl/Shader.cpp"
  $23 = HEAP32[165651]|0; //@line 515 "./libraries/glad/gladfuncs.hpp"
  __THREW__ = 0;
  invoke_viiiiiii($23|0,($22|0),($i1$015|0),256,($namelen|0),($13|0),($14|0),($cname|0)); //@line 515 "./libraries/glad/gladfuncs.hpp"
  $24 = __THREW__; __THREW__ = 0;
  $25 = $24&1;
  if ($25) {
   label = 18;
   break;
  }
  $26 = HEAP32[$namelen>>2]|0; //@line 199 "modules/graphics/opengl/Shader.cpp"
  __THREW__ = 0;
  invoke_viii(3460,($0|0),($cname|0),($26|0)); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $27 = __THREW__; __THREW__ = 0;
  $28 = $27&1;
  if ($28) {
   label = 18;
   break;
  }
  $29 = HEAP8[$16>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $30 = $29 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $31 = ($30<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($31) {
   HEAP8[$17>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP8[$16>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  } else {
   $32 = HEAP32[$18>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP8[$32>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   HEAP32[$19>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  }
  __THREW__ = 0;
  invoke_vii(3368,($16|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $33 = __THREW__; __THREW__ = 0;
  $34 = $33&1;
  if ($34) {
   label = 10;
   break;
  }
  ;HEAP32[$16>>2]=HEAP32[$0>>2]|0;HEAP32[$16+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$16+8>>2]=HEAP32[$0+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 199 "modules/graphics/opengl/Shader.cpp"
  $37 = HEAP32[$6>>2]|0; //@line 200 "modules/graphics/opengl/Shader.cpp"
  $38 = HEAP8[$16>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $39 = $38 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $40 = ($39<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $41 = HEAP32[$18>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $42 = $40 ? $17 : $41; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $43 = HEAP32[165470]|0; //@line 523 "./libraries/glad/gladfuncs.hpp"
  __THREW__ = 0;
  $44 = (invoke_iii($43|0,($37|0),($42|0))|0); //@line 207 "modules/graphics/opengl/Shader.cpp"
  $45 = __THREW__; __THREW__ = 0;
  $46 = $45&1;
  if ($46) {
   label = 18;
   break;
  }
  HEAP32[$u>>2] = $44; //@line 200 "modules/graphics/opengl/Shader.cpp"
  $47 = HEAP32[$14>>2]|0; //@line 201 "modules/graphics/opengl/Shader.cpp"
  $switch$split2D = ($47|0)<(35681);
  L14: do {
   if ($switch$split2D) {
    $switch$split12D = ($47|0)<(35671);
    L16: do {
     if ($switch$split12D) {
      $switch$split42D = ($47|0)<(35666);
      if (!($switch$split42D)) {
       switch ($47|0) {
       case 35666:  {
        label = 13;
        break L14;
        break;
       }
       case 35669: case 35668: case 35667:  {
        $$0$i$ph = 1;
        label = 16;
        break L14;
        break;
       }
       case 35670:  {
        break L16;
        break;
       }
       default: {
        label = 15;
        break L14;
       }
       }
      }
      $switch$split72D = ($47|0)<(35664);
      if ($switch$split72D) {
       switch ($47|0) {
       case 5126:  {
        label = 13;
        break L14;
        break;
       }
       case 5124:  {
        $$0$i$ph = 1;
        label = 16;
        break L14;
        break;
       }
       default: {
        label = 15;
        break L14;
       }
       }
      } else {
       switch ($47|0) {
       case 35665: case 35664:  {
        label = 13;
        break L14;
        break;
       }
       default: {
        label = 15;
        break L14;
       }
       }
      }
     } else {
      switch ($47|0) {
      case 35676: case 35675: case 35674:  {
       label = 13;
       break L14;
       break;
      }
      case 35680: case 35679: case 35678: case 35677:  {
       label = 17;
       break L14;
       break;
      }
      case 35673: case 35672: case 35671:  {
       break;
      }
      default: {
       label = 15;
       break L14;
      }
      }
     }
    } while(0);
    $$0$i$ph = 2;
    label = 16;
   } else {
    switch ($47|0) {
    case 35690: case 35689: case 35688: case 35687: case 35686: case 35685:  {
     label = 13;
     break;
    }
    case 36877: case 36876: case 36293: case 36292: case 36289: case 35682: case 35684: case 35683: case 37131: case 37128: case 36291: case 36288: case 35681:  {
     label = 17;
     break;
    }
    default: {
     label = 15;
    }
    }
   }
  } while(0);
  if ((label|0) == 13) {
   label = 0;
   $$0$i$ph = 0;
   label = 16;
  }
  else if ((label|0) == 15) {
   label = 0;
   $$0$i$ph = 4;
   label = 16;
  }
  else if ((label|0) == 17) {
   label = 0;
   HEAP32[$15>>2] = 3; //@line 201 "modules/graphics/opengl/Shader.cpp"
   $48 = HEAP32[165188]|0; //@line 539 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   invoke_vii($48|0,($44|0),0); //@line 539 "./libraries/glad/gladfuncs.hpp"
   $49 = __THREW__; __THREW__ = 0;
   $50 = $49&1;
   if ($50) {
    label = 18;
    break;
   }
  }
  if ((label|0) == 16) {
   label = 0;
   HEAP32[$15>>2] = $$0$i$ph; //@line 201 "modules/graphics/opengl/Shader.cpp"
  }
  $53 = HEAP8[$16>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $54 = $53 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $55 = ($54<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $56 = HEAP32[$19>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $57 = $53&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $58 = $57 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $59 = $55 ? $58 : $56; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $60 = ($59>>>0)>(3); //@line 210 "modules/graphics/opengl/Shader.cpp"
  L34: do {
   if ($60) {
    $61 = HEAP32[$18>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $62 = $55 ? $17 : $61; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $63 = (($62) + ($59)|0); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $64 = $62; //@line 1466 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $65 = ($59|0)<(3); //@line 1467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    if ($65) {
     $85 = $53;
    } else {
     $66 = ((($63)) + -2|0); //@line 1469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $67 = ($66|0)==($62|0); //@line 1475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     if ($67) {
      $85 = $53;
     } else {
      $$016$i$i$i = $62;
      L37: while(1) {
       $68 = HEAP8[$$016$i$i$i>>0]|0; //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $69 = ($68<<24>>24)==(91); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($69) {
        $__m1$0$i$i$i = $$016$i$i$i;$__m2$0$i$i$i = 438199;
        while(1) {
         $72 = ((($__m2$0$i$i$i)) + 1|0); //@line 1517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $73 = ($72|0)==((438202)|0); //@line 1517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         if ($73) {
          $$016$i$i$i$lcssa = $$016$i$i$i;
          break L37;
         }
         $74 = ((($__m1$0$i$i$i)) + 1|0); //@line 1519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $75 = HEAP8[$74>>0]|0; //@line 1520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $76 = HEAP8[$72>>0]|0; //@line 1520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $77 = ($75<<24>>24)==($76<<24>>24); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         if ($77) {
          $__m1$0$i$i$i = $74;$__m2$0$i$i$i = $72;
         } else {
          break;
         }
        }
       }
       $70 = ((($$016$i$i$i)) + 1|0); //@line 1479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $71 = ($70|0)==($66|0); //@line 1475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       if ($71) {
        $85 = $53;
        break L34;
       } else {
        $$016$i$i$i = $70;
       }
      }
      $78 = $$016$i$i$i$lcssa; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $79 = (($78) - ($64))|0; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $notlhs = ($$016$i$i$i$lcssa|0)!=($63|0);
      $notrhs = ($79|0)!=(-1);
      $or$cond$not = $notrhs & $notlhs;
      $80 = (($59) + -3)|0; //@line 213 "modules/graphics/opengl/Shader.cpp"
      $81 = ($79|0)==($80|0); //@line 213 "modules/graphics/opengl/Shader.cpp"
      $or$cond28 = $or$cond$not & $81;
      if ($or$cond28) {
       __THREW__ = 0;
       (invoke_iiii(4117,($16|0),($79|0),-1)|0); //@line 214 "modules/graphics/opengl/Shader.cpp"
       $82 = __THREW__; __THREW__ = 0;
       $83 = $82&1;
       if ($83) {
        label = 18;
        break L4;
       }
       $$pre = HEAP8[$16>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $85 = $$pre;
      } else {
       $85 = $53;
      }
     }
    }
   } else {
    $85 = $53;
   }
  } while(0);
  $84 = $85 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $86 = ($84<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $87 = HEAP32[$18>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $88 = $86 ? $17 : $87; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $89 = HEAP8[$88>>0]|0; //@line 142 "./common/StringMap.h"
  $90 = ($89<<24>>24)==(0); //@line 142 "./common/StringMap.h"
  if ($90) {
   $hash$0$lcssa$i$i = 5381;
  } else {
   $$02$i$i = $88;$92 = $89;$hash$01$i$i = 5381;
   while(1) {
    $91 = $92 << 24 >> 24; //@line 142 "./common/StringMap.h"
    $93 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
    $94 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
    $95 = (($94) + ($91))|0; //@line 143 "./common/StringMap.h"
    $96 = HEAP8[$93>>0]|0; //@line 142 "./common/StringMap.h"
    $97 = ($96<<24>>24)==(0); //@line 142 "./common/StringMap.h"
    if ($97) {
     $hash$0$lcssa$i$i = $95;
     break;
    } else {
     $$02$i$i = $93;$92 = $96;$hash$01$i$i = $95;
    }
   }
  }
  $i$010$i = 0;
  while(1) {
   $98 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
   $99 = (($98>>>0) % 18)&-1; //@line 72 "./common/StringMap.h"
   $100 = (((634612 + (($99*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $101 = HEAP8[$100>>0]|0; //@line 74 "./common/StringMap.h"
   $102 = ($101<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($102) {
    label = 39;
    break;
   }
   $103 = (634612 + (($99*12)|0)|0); //@line 77 "./common/StringMap.h"
   $104 = HEAP32[$103>>2]|0; //@line 77 "./common/StringMap.h"
   $105 = HEAP8[$104>>0]|0; //@line 54 "./common/StringMap.h"
   $106 = ($105<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L55: do {
    if ($106) {
     $$08$ph$i$i = $88;
     label = 36;
    } else {
     $$011$i$i = $88;$$0110$i$i = $104;$110 = $105;
     while(1) {
      $107 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
      $108 = ($107<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
      $109 = ($110<<24>>24)==($107<<24>>24); //@line 56 "./common/StringMap.h"
      $or$cond$i$i = $108 & $109; //@line 54 "./common/StringMap.h"
      if (!($or$cond$i$i)) {
       break L55;
      }
      $111 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
      $112 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
      $113 = HEAP8[$111>>0]|0; //@line 54 "./common/StringMap.h"
      $114 = ($113<<24>>24)==(0); //@line 54 "./common/StringMap.h"
      if ($114) {
       $$08$ph$i$i = $112;
       label = 36;
       break;
      } else {
       $$011$i$i = $112;$$0110$i$i = $111;$110 = $113;
      }
     }
    }
   } while(0);
   if ((label|0) == 36) {
    label = 0;
    $115 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
    $116 = ($115<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($116) {
     $$lcssa40 = $99;
     label = 38;
     break;
    }
   }
   $117 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
   $118 = ($117>>>0)<(18); //@line 70 "./common/StringMap.h"
   if ($118) {
    $i$010$i = $117;
   } else {
    label = 39;
    break;
   }
  }
  if ((label|0) == 38) {
   label = 0;
   $119 = (((634612 + (($$lcssa40*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
   $120 = HEAP32[$119>>2]|0; //@line 79 "./common/StringMap.h"
   $121 = HEAP32[$u>>2]|0; //@line 220 "modules/graphics/opengl/Shader.cpp"
   $122 = (((($this)) + 52|0) + ($120<<2)|0); //@line 220 "modules/graphics/opengl/Shader.cpp"
   HEAP32[$122>>2] = $121; //@line 220 "modules/graphics/opengl/Shader.cpp"
   $124 = $121;
  }
  else if ((label|0) == 39) {
   label = 0;
   $$pre22 = HEAP32[$u>>2]|0; //@line 222 "modules/graphics/opengl/Shader.cpp"
   $124 = $$pre22;
  }
  $123 = ($124|0)==(-1); //@line 222 "modules/graphics/opengl/Shader.cpp"
  if (!($123)) {
   __THREW__ = 0;
   $125 = (invoke_iii(4118,($1|0),($16|0))|0); //@line 223 "modules/graphics/opengl/Shader.cpp"
   $126 = __THREW__; __THREW__ = 0;
   $127 = $126&1;
   if ($127) {
    label = 18;
    break;
   }
   ;HEAP32[$125>>2]=HEAP32[$u>>2]|0;HEAP32[$125+4>>2]=HEAP32[$u+4>>2]|0;HEAP32[$125+8>>2]=HEAP32[$u+8>>2]|0;HEAP32[$125+12>>2]=HEAP32[$u+12>>2]|0; //@line 214 "modules/graphics/opengl/Shader.h"
   $128 = ((($125)) + 16|0); //@line 214 "modules/graphics/opengl/Shader.h"
   __THREW__ = 0;
   (invoke_iii(3578,($128|0),($16|0))|0); //@line 214 "modules/graphics/opengl/Shader.h"
   $129 = __THREW__; __THREW__ = 0;
   $130 = $129&1;
   if ($130) {
    label = 18;
    break;
   }
  }
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($16); //@line 214 "modules/graphics/opengl/Shader.h"
  $131 = (($i1$015) + 1)|0; //@line 192 "modules/graphics/opengl/Shader.cpp"
  $132 = HEAP32[$numuniforms>>2]|0; //@line 192 "modules/graphics/opengl/Shader.cpp"
  $133 = ($131|0)<($132|0); //@line 192 "modules/graphics/opengl/Shader.cpp"
  if ($133) {
   $i1$015 = $131;
  } else {
   label = 3;
   break;
  }
 }
 if ((label|0) == 3) {
  $20 = HEAP32[$activeprogram>>2]|0; //@line 226 "modules/graphics/opengl/Shader.cpp"
  $21 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$21 & 8191]($20); //@line 534 "./libraries/glad/gladfuncs.hpp"
  STACKTOP = sp;return; //@line 227 "modules/graphics/opengl/Shader.cpp"
 }
 else if ((label|0) == 10) {
  $35 = ___cxa_find_matching_catch_3(0|0)|0;
  $36 = tempRet0;
  ___clang_call_terminate($35); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 else if ((label|0) == 18) {
  $51 = ___cxa_find_matching_catch_2()|0;
  $52 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($16); //@line 214 "modules/graphics/opengl/Shader.h"
  ___resumeException($51|0);
  // unreachable;
 }
}
function __ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformENS_4lessIS6_EENS4_INS_4pairIKS6_SB_EEEEEixERSF_($this,$__k) {
 $this = $this|0;
 $__k = $__k|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__parent = 0, $__r$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__parent = sp;
 $0 = (__ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformENS_4lessIS6_EENS4_INS_4pairIKS6_SB_EEEEE16__find_equal_keyERPNS_16__tree_node_baseIPvEERSF_($this,$__parent,$__k)|0); //@line 1376 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $1 = HEAP32[$0>>2]|0; //@line 1377 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $2 = ($1|0)==(0|0); //@line 1378 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 if (!($2)) {
  $__r$0 = $1;
  $23 = ((($__r$0)) + 28|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  STACKTOP = sp;return ($23|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 $3 = (__Znwj(56)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $4 = ((($3)) + 16|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 __THREW__ = 0;
 invoke_vii(3351,($4|0),($__k|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZdlPv($3); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($7|0);
  // unreachable;
 }
 $9 = ((($3)) + 28|0); //@line 1366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;HEAP32[$9+12>>2]=0|0;HEAP32[$9+16>>2]=0|0;HEAP32[$9+20>>2]=0|0;HEAP32[$9+24>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = HEAP32[$__parent>>2]|0; //@line 1381 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $11 = ((($3)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$11>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $12 = ((($3)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$12>>2] = $10; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$0>>2] = $3; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $13 = HEAP32[$this>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $14 = HEAP32[$13>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $15 = ($14|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($15) {
  $19 = $3;
 } else {
  $16 = $14; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$this>>2] = $16; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre$i = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $19 = $$pre$i;
 }
 $17 = ((($this)) + 4|0); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $18 = HEAP32[$17>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($18,$19); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $20 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $21 = HEAP32[$20>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $22 = (($21) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$20>>2] = $22; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $__r$0 = $3;
 $23 = ((($__r$0)) + 28|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 STACKTOP = sp;return ($23|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
}
function __ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformENS_4lessIS6_EENS4_INS_4pairIKS6_SB_EEEEE16__find_equal_keyERPNS_16__tree_node_baseIPvEERSF_($this,$__parent,$__k) {
 $this = $this|0;
 $__parent = $__parent|0;
 $__k = $__k|0;
 var $$$$i$i$i = 0, $$$$i$i$i1 = 0, $$0 = 0, $$lcssa = 0, $$lcssa19 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__nd$0 = 0, $__nd$0$lcssa = 0, $__nd$0$lcssa17 = 0, $__nd$0$lcssa18 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $2 = ($1|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 if ($2) {
  HEAP32[$__parent>>2] = $0; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $0;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 $3 = HEAP8[$__k>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = $3 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ($4<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = ((($__k)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = HEAP32[$6>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $3&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $8 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $5 ? $9 : $7; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = ((($__k)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = HEAP32[$11>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ((($__k)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = $5 ? $13 : $12; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $__nd$0 = $1;
 while(1) {
  $15 = ((($__nd$0)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $16 = HEAP8[$15>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $17 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $18 = ($17<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $19 = ((($__nd$0)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $20 = HEAP32[$19>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $21 = $16&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $22 = $21 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $23 = $18 ? $22 : $20; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $24 = ((($__nd$0)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $25 = HEAP32[$24>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $26 = ((($15)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $27 = $18 ? $26 : $25; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $28 = ($23>>>0)<($10>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $29 = $28 ? $23 : $10; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $30 = (_memcmp($14,$27,$29)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $31 = ($30|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $32 = ($10>>>0)<($23>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $$$$i$i$i1 = $32 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $33 = $31 ? $$$$i$i$i1 : $30; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $34 = ($33|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($34) {
   $35 = HEAP32[$__nd$0>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $36 = ($35|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   if ($36) {
    $$lcssa19 = $__nd$0;$__nd$0$lcssa18 = $__nd$0;
    label = 5;
    break;
   } else {
    $$sink = $35;
   }
  } else {
   $37 = $32 ? $10 : $23; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $38 = (_memcmp($27,$14,$37)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $39 = ($38|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$$$i$i$i = $28 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $40 = $39 ? $$$$i$i$i : $38; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $41 = ($40|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if (!($41)) {
    $__nd$0$lcssa = $__nd$0;
    label = 10;
    break;
   }
   $42 = ((($__nd$0)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $43 = HEAP32[$42>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $44 = ($43|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   if ($44) {
    $$lcssa = $42;$__nd$0$lcssa17 = $__nd$0;
    label = 9;
    break;
   } else {
    $$sink = $43;
   }
  }
  $__nd$0 = $$sink;
 }
 if ((label|0) == 5) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa18; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $$lcssa19;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 else if ((label|0) == 9) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa17; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $$lcssa;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 else if ((label|0) == 10) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $__parent;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl6Shader23checkSetBuiltinUniformsEv($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$cast = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i31 = 0, $$pre$i21 = 0, $$pre2$i = 0, $$pre2$i$i = 0, $$pre2$i$i32 = 0, $$pre2$i22 = 0, $$pre4$i = 0, $$pre4$i$i = 0, $$pre4$i$i29 = 0, $$pre4$i19 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$01$i = 0, $i$01$i$i = 0, $i$01$i$i28 = 0, $i$01$i18 = 0, $lpad$phi$index = 0, $lpad$phi46$index = 0, $normalmatrix = 0, $tp_matrix = 0, $tpmatrixneedsupdate$0$off0 = 0, dest = 0;
 var label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $normalmatrix = sp + 100|0;
 $0 = sp + 64|0;
 $tp_matrix = sp;
 __ZN4love8graphics6opengl6Shader20checkSetScreenParamsEv($this); //@line 767 "modules/graphics/opengl/Shader.cpp"
 $1 = HEAP8[1252012]|0; //@line 771 "modules/graphics/opengl/Shader.cpp"
 $2 = ($1<<24>>24)==(0); //@line 771 "modules/graphics/opengl/Shader.cpp"
 if ($2) {
  STACKTOP = sp;return; //@line 823 "modules/graphics/opengl/Shader.cpp"
 }
 $3 = (+__ZNK4love8graphics6opengl6OpenGL12getPointSizeEv(634044)); //@line 773 "modules/graphics/opengl/Shader.cpp"
 __ZN4love8graphics6opengl6Shader17checkSetPointSizeEf($this,$3); //@line 773 "modules/graphics/opengl/Shader.cpp"
 $4 = HEAP32[(634048)>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ((($4)) + -64|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[(634060)>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ((($6)) + -64|0); //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = HEAP32[158581]|0; //@line 46 "modules/graphics/opengl/Shader.cpp"
 $$cast = $8; //@line 406 "modules/graphics/opengl/Shader.cpp"
 $9 = ($$cast|0)==($this|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
 if (!($9)) {
  $10 = ((($this)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
  $12 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$12 & 8191]($11); //@line 534 "./libraries/glad/gladfuncs.hpp"
  HEAP32[158581] = $this; //@line 409 "modules/graphics/opengl/Shader.cpp"
 }
 __THREW__ = 0;
 $13 = (invoke_ii(4119,($5|0))|0); //@line 783 "modules/graphics/opengl/Shader.cpp"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 L7: do {
  if ($15) {
   label = 10;
  } else {
   $16 = ((($this)) + 188|0); //@line 783 "modules/graphics/opengl/Shader.cpp"
   __THREW__ = 0;
   $17 = (invoke_ii(4119,($16|0))|0); //@line 783 "modules/graphics/opengl/Shader.cpp"
   $18 = __THREW__; __THREW__ = 0;
   $19 = $18&1;
   if ($19) {
    label = 10;
   } else {
    $20 = (_memcmp($13,$17,64)|0); //@line 783 "modules/graphics/opengl/Shader.cpp"
    $21 = ($20|0)==(0); //@line 783 "modules/graphics/opengl/Shader.cpp"
    if ($21) {
     $tpmatrixneedsupdate$0$off0 = 0;
    } else {
     $22 = ((($this)) + 52|0); //@line 785 "modules/graphics/opengl/Shader.cpp"
     $23 = HEAP32[$22>>2]|0; //@line 785 "modules/graphics/opengl/Shader.cpp"
     $24 = ($23|0)>(-1); //@line 786 "modules/graphics/opengl/Shader.cpp"
     if ($24) {
      __THREW__ = 0;
      $25 = (invoke_ii(4119,($5|0))|0); //@line 787 "modules/graphics/opengl/Shader.cpp"
      $26 = __THREW__; __THREW__ = 0;
      $27 = $26&1;
      if ($27) {
       label = 10;
       break;
      }
      $28 = HEAP32[165649]|0; //@line 553 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_viiii($28|0,($23|0),1,0,($25|0)); //@line 553 "./libraries/glad/gladfuncs.hpp"
      $29 = __THREW__; __THREW__ = 0;
      $30 = $29&1;
      if ($30) {
       label = 10;
       break;
      }
     }
     $33 = ((($this)) + 64|0); //@line 792 "modules/graphics/opengl/Shader.cpp"
     $34 = HEAP32[$33>>2]|0; //@line 792 "modules/graphics/opengl/Shader.cpp"
     $35 = ($34|0)>(-1); //@line 793 "modules/graphics/opengl/Shader.cpp"
     do {
      if ($35) {
       __THREW__ = 0;
       invoke_vii(4120,($0|0),($5|0)); //@line 795 "modules/graphics/opengl/Shader.cpp"
       $36 = __THREW__; __THREW__ = 0;
       $37 = $36&1;
       if ($37) {
        label = 10;
        break L7;
       }
       __THREW__ = 0;
       invoke_vii(4121,($normalmatrix|0),($0|0)); //@line 795 "modules/graphics/opengl/Shader.cpp"
       $38 = __THREW__; __THREW__ = 0;
       $39 = $38&1;
       if ($39) {
        $46 = ___cxa_find_matching_catch_2()|0;
        $47 = tempRet0;
        __ZN4love7Matrix3D2Ev($0); //@line 795 "modules/graphics/opengl/Shader.cpp"
        $$0 = $46;$$01 = $47;
        break L7;
       }
       __ZN4love7Matrix3D2Ev($0); //@line 795 "modules/graphics/opengl/Shader.cpp"
       __THREW__ = 0;
       $40 = (invoke_ii(4122,($normalmatrix|0))|0); //@line 796 "modules/graphics/opengl/Shader.cpp"
       $41 = __THREW__; __THREW__ = 0;
       $42 = $41&1;
       if (!($42)) {
        $43 = HEAP32[165618]|0; //@line 552 "./libraries/glad/gladfuncs.hpp"
        __THREW__ = 0;
        invoke_viiii($43|0,($34|0),1,0,($40|0)); //@line 552 "./libraries/glad/gladfuncs.hpp"
        $44 = __THREW__; __THREW__ = 0;
        $45 = $44&1;
        if (!($45)) {
         __ZN4love7Matrix3D2Ev($normalmatrix); //@line 797 "modules/graphics/opengl/Shader.cpp"
         break;
        }
       }
       $48 = ___cxa_find_matching_catch_2()|0;
       $49 = tempRet0;
       __ZN4love7Matrix3D2Ev($normalmatrix); //@line 797 "modules/graphics/opengl/Shader.cpp"
       $$0 = $48;$$01 = $49;
       break L7;
      }
     } while(0);
     dest=$16; src=$5; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 800 "modules/graphics/opengl/Shader.cpp"
     $tpmatrixneedsupdate$0$off0 = 1;
    }
    __THREW__ = 0;
    $50 = (invoke_ii(4119,($7|0))|0); //@line 803 "modules/graphics/opengl/Shader.cpp"
    $51 = __THREW__; __THREW__ = 0;
    $52 = $51&1;
    if ($52) {
     label = 10;
    } else {
     $53 = ((($this)) + 252|0); //@line 803 "modules/graphics/opengl/Shader.cpp"
     __THREW__ = 0;
     $54 = (invoke_ii(4119,($53|0))|0); //@line 803 "modules/graphics/opengl/Shader.cpp"
     $55 = __THREW__; __THREW__ = 0;
     $56 = $55&1;
     if ($56) {
      label = 10;
     } else {
      $57 = (_memcmp($50,$54,64)|0); //@line 803 "modules/graphics/opengl/Shader.cpp"
      $58 = ($57|0)==(0); //@line 803 "modules/graphics/opengl/Shader.cpp"
      if ($58) {
       if ($tpmatrixneedsupdate$0$off0) {
        label = 28;
       }
      } else {
       $59 = ((($this)) + 56|0); //@line 805 "modules/graphics/opengl/Shader.cpp"
       $60 = HEAP32[$59>>2]|0; //@line 805 "modules/graphics/opengl/Shader.cpp"
       $61 = ($60|0)>(-1); //@line 806 "modules/graphics/opengl/Shader.cpp"
       if ($61) {
        __THREW__ = 0;
        $62 = (invoke_ii(4119,($7|0))|0); //@line 807 "modules/graphics/opengl/Shader.cpp"
        $63 = __THREW__; __THREW__ = 0;
        $64 = $63&1;
        if ($64) {
         label = 10;
         break;
        }
        $65 = HEAP32[165649]|0; //@line 553 "./libraries/glad/gladfuncs.hpp"
        __THREW__ = 0;
        invoke_viiii($65|0,($60|0),1,0,($62|0)); //@line 553 "./libraries/glad/gladfuncs.hpp"
        $66 = __THREW__; __THREW__ = 0;
        $67 = $66&1;
        if ($67) {
         label = 10;
         break;
        }
       }
       dest=$53; src=$7; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 810 "modules/graphics/opengl/Shader.cpp"
       label = 28;
      }
      do {
       if ((label|0) == 28) {
        $68 = ((($this)) + 60|0); //@line 815 "modules/graphics/opengl/Shader.cpp"
        $69 = HEAP32[$68>>2]|0; //@line 815 "modules/graphics/opengl/Shader.cpp"
        $70 = ($69|0)>(-1); //@line 816 "modules/graphics/opengl/Shader.cpp"
        if ($70) {
         __THREW__ = 0;
         invoke_viii(4123,($tp_matrix|0),($7|0),($5|0)); //@line 818 "modules/graphics/opengl/Shader.cpp"
         $71 = __THREW__; __THREW__ = 0;
         $72 = $71&1;
         if ($72) {
          label = 10;
          break L7;
         }
         __THREW__ = 0;
         $73 = (invoke_ii(4119,($tp_matrix|0))|0); //@line 819 "modules/graphics/opengl/Shader.cpp"
         $74 = __THREW__; __THREW__ = 0;
         $75 = $74&1;
         if (!($75)) {
          $76 = HEAP32[165649]|0; //@line 553 "./libraries/glad/gladfuncs.hpp"
          __THREW__ = 0;
          invoke_viiii($76|0,($69|0),1,0,($73|0)); //@line 553 "./libraries/glad/gladfuncs.hpp"
          $77 = __THREW__; __THREW__ = 0;
          $78 = $77&1;
          if (!($78)) {
           __ZN4love7Matrix4D2Ev($tp_matrix); //@line 820 "modules/graphics/opengl/Shader.cpp"
           break;
          }
         }
         $79 = ___cxa_find_matching_catch_2()|0;
         $80 = tempRet0;
         __ZN4love7Matrix4D2Ev($tp_matrix); //@line 820 "modules/graphics/opengl/Shader.cpp"
         $$0 = $79;$$01 = $80;
         break L7;
        }
       }
      } while(0);
      $81 = ($8|0)==(0); //@line 53 "modules/graphics/opengl/Shader.cpp"
      do {
       if ($81) {
        $108 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
        $109 = ($108|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
        $110 = HEAP32[158581]|0; //@line 432 "modules/graphics/opengl/Shader.cpp"
        if ($109) {
         $136 = ($110|0)==(0|0); //@line 438 "modules/graphics/opengl/Shader.cpp"
         if (!($136)) {
          $137 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
          __THREW__ = 0;
          invoke_vi($137|0,0); //@line 534 "./libraries/glad/gladfuncs.hpp"
          $138 = __THREW__; __THREW__ = 0;
          $139 = $138&1;
          if ($139) {
           break;
          }
         }
         HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
         STACKTOP = sp;return; //@line 823 "modules/graphics/opengl/Shader.cpp"
        }
        $111 = ($110|0)==($108|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
        if ($111) {
         STACKTOP = sp;return; //@line 823 "modules/graphics/opengl/Shader.cpp"
        }
        $112 = ((($108)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
        $113 = HEAP32[$112>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
        $114 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
        __THREW__ = 0;
        invoke_vi($114|0,($113|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
        $115 = __THREW__; __THREW__ = 0;
        $116 = $115&1;
        if (!($116)) {
         HEAP32[158581] = $108; //@line 409 "modules/graphics/opengl/Shader.cpp"
         $117 = ((($108)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
         $118 = ((($108)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $119 = HEAP32[$118>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $120 = HEAP32[$117>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $121 = ($119|0)==($120|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
         L59: do {
          if (!($121)) {
           $123 = $120;$212 = $119;$i$01$i$i28 = 0;
           while(1) {
            $122 = $123; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $124 = (($122) + ($i$01$i$i28<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $125 = HEAP32[$124>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
            $126 = ($125|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
            $$pre4$i$i29 = (($i$01$i$i28) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
            if ($126) {
             $130 = $212;$131 = $123;
            } else {
             __THREW__ = 0;
             invoke_viiii(4111,(634044|0),($125|0),($$pre4$i$i29|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
             $127 = __THREW__; __THREW__ = 0;
             $128 = $127&1;
             if ($128) {
              break;
             }
             $$pre$i$i31 = HEAP32[$118>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $$pre2$i$i32 = HEAP32[$117>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $130 = $$pre$i$i31;$131 = $$pre2$i$i32;
            }
            $129 = (($130) - ($131))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $132 = $129 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $133 = ($$pre4$i$i29>>>0)<($132>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
            if ($133) {
             $123 = $131;$212 = $130;$i$01$i$i28 = $$pre4$i$i29;
            } else {
             break L59;
            }
           }
           $140 = ___cxa_find_matching_catch_3(0|0)|0;
           $141 = tempRet0;
           $lpad$phi$index = $140;
           ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
           // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
          }
         } while(0);
         __THREW__ = 0;
         invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
         $134 = __THREW__; __THREW__ = 0;
         $135 = $134&1;
         if (!($135)) {
          STACKTOP = sp;return; //@line 823 "modules/graphics/opengl/Shader.cpp"
         }
        }
       } else {
        $82 = HEAP32[158581]|0; //@line 406 "modules/graphics/opengl/Shader.cpp"
        $83 = ($82|0)==($$cast|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
        if (!($83)) {
         $84 = ((($$cast)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
         $85 = HEAP32[$84>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
         $86 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
         __THREW__ = 0;
         invoke_vi($86|0,($85|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
         $87 = __THREW__; __THREW__ = 0;
         $88 = $87&1;
         if ($88) {
          break;
         }
         HEAP32[158581] = $$cast; //@line 409 "modules/graphics/opengl/Shader.cpp"
        }
        $89 = ((($$cast)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
        $90 = ((($$cast)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $91 = HEAP32[$90>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $92 = HEAP32[$89>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $93 = ($91|0)==($92|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
        L75: do {
         if (!($93)) {
          $211 = $91;$97 = $92;$i$01$i18 = 0;
          while(1) {
           $96 = $97; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $98 = (($96) + ($i$01$i18<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $99 = HEAP32[$98>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
           $100 = ($99|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
           $$pre4$i19 = (($i$01$i18) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
           if ($100) {
            $104 = $211;$105 = $97;
           } else {
            __THREW__ = 0;
            invoke_viiii(4111,(634044|0),($99|0),($$pre4$i19|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
            $101 = __THREW__; __THREW__ = 0;
            $102 = $101&1;
            if ($102) {
             break;
            }
            $$pre$i21 = HEAP32[$90>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $$pre2$i22 = HEAP32[$89>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $104 = $$pre$i21;$105 = $$pre2$i22;
           }
           $103 = (($104) - ($105))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $106 = $103 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $107 = ($$pre4$i19>>>0)<($106>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
           if ($107) {
            $211 = $104;$97 = $105;$i$01$i18 = $$pre4$i19;
           } else {
            break L75;
           }
          }
          $142 = ___cxa_find_matching_catch_3(0|0)|0;
          $143 = tempRet0;
          $lpad$phi$index = $142;
          ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
          // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
         }
        } while(0);
        __THREW__ = 0;
        invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
        $94 = __THREW__; __THREW__ = 0;
        $95 = $94&1;
        if (!($95)) {
         STACKTOP = sp;return; //@line 823 "modules/graphics/opengl/Shader.cpp"
        }
       }
      } while(0);
      $144 = ___cxa_find_matching_catch_3(0|0)|0;
      $145 = tempRet0;
      $lpad$phi$index = $144;
      ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
      // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 10) {
  $31 = ___cxa_find_matching_catch_2()|0;
  $32 = tempRet0;
  $$0 = $31;$$01 = $32;
 }
 $146 = ($8|0)==(0); //@line 53 "modules/graphics/opengl/Shader.cpp"
 do {
  if ($146) {
   $173 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
   $174 = ($173|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
   $175 = HEAP32[158581]|0; //@line 432 "modules/graphics/opengl/Shader.cpp"
   if ($174) {
    $201 = ($175|0)==(0|0); //@line 438 "modules/graphics/opengl/Shader.cpp"
    if (!($201)) {
     $202 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_vi($202|0,0); //@line 534 "./libraries/glad/gladfuncs.hpp"
     $203 = __THREW__; __THREW__ = 0;
     $204 = $203&1;
     if ($204) {
      break;
     }
    }
    HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
    ___resumeException($$0|0);
    // unreachable;
   }
   $176 = ($175|0)==($173|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
   if ($176) {
    ___resumeException($$0|0);
    // unreachable;
   }
   $177 = ((($173)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
   $178 = HEAP32[$177>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
   $179 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   invoke_vi($179|0,($178|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
   $180 = __THREW__; __THREW__ = 0;
   $181 = $180&1;
   if (!($181)) {
    HEAP32[158581] = $173; //@line 409 "modules/graphics/opengl/Shader.cpp"
    $182 = ((($173)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
    $183 = ((($173)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $184 = HEAP32[$183>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $185 = HEAP32[$182>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $186 = ($184|0)==($185|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
    L104: do {
     if (!($186)) {
      $188 = $185;$214 = $184;$i$01$i$i = 0;
      while(1) {
       $187 = $188; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $189 = (($187) + ($i$01$i$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $190 = HEAP32[$189>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
       $191 = ($190|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
       $$pre4$i$i = (($i$01$i$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
       if ($191) {
        $195 = $214;$196 = $188;
       } else {
        __THREW__ = 0;
        invoke_viiii(4111,(634044|0),($190|0),($$pre4$i$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
        $192 = __THREW__; __THREW__ = 0;
        $193 = $192&1;
        if ($193) {
         break;
        }
        $$pre$i$i = HEAP32[$183>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $$pre2$i$i = HEAP32[$182>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $195 = $$pre$i$i;$196 = $$pre2$i$i;
       }
       $194 = (($195) - ($196))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $197 = $194 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $198 = ($$pre4$i$i>>>0)<($197>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
       if ($198) {
        $188 = $196;$214 = $195;$i$01$i$i = $$pre4$i$i;
       } else {
        break L104;
       }
      }
      $205 = ___cxa_find_matching_catch_3(0|0)|0;
      $206 = tempRet0;
      $lpad$phi46$index = $205;
      ___clang_call_terminate($lpad$phi46$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
      // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
     }
    } while(0);
    __THREW__ = 0;
    invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
    $199 = __THREW__; __THREW__ = 0;
    $200 = $199&1;
    if (!($200)) {
     ___resumeException($$0|0);
     // unreachable;
    }
   }
  } else {
   $147 = HEAP32[158581]|0; //@line 406 "modules/graphics/opengl/Shader.cpp"
   $148 = ($147|0)==($$cast|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
   if (!($148)) {
    $149 = ((($$cast)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
    $150 = HEAP32[$149>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
    $151 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
    __THREW__ = 0;
    invoke_vi($151|0,($150|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
    $152 = __THREW__; __THREW__ = 0;
    $153 = $152&1;
    if ($153) {
     break;
    }
    HEAP32[158581] = $$cast; //@line 409 "modules/graphics/opengl/Shader.cpp"
   }
   $154 = ((($$cast)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
   $155 = ((($$cast)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $156 = HEAP32[$155>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $157 = HEAP32[$154>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $158 = ($156|0)==($157|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
   L120: do {
    if (!($158)) {
     $162 = $157;$213 = $156;$i$01$i = 0;
     while(1) {
      $161 = $162; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $163 = (($161) + ($i$01$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $164 = HEAP32[$163>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
      $165 = ($164|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
      $$pre4$i = (($i$01$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
      if ($165) {
       $169 = $213;$170 = $162;
      } else {
       __THREW__ = 0;
       invoke_viiii(4111,(634044|0),($164|0),($$pre4$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
       $166 = __THREW__; __THREW__ = 0;
       $167 = $166&1;
       if ($167) {
        break;
       }
       $$pre$i = HEAP32[$155>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $$pre2$i = HEAP32[$154>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $169 = $$pre$i;$170 = $$pre2$i;
      }
      $168 = (($169) - ($170))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $171 = $168 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $172 = ($$pre4$i>>>0)<($171>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      if ($172) {
       $162 = $170;$213 = $169;$i$01$i = $$pre4$i;
      } else {
       break L120;
      }
     }
     $207 = ___cxa_find_matching_catch_3(0|0)|0;
     $208 = tempRet0;
     $lpad$phi46$index = $207;
     ___clang_call_terminate($lpad$phi46$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
     // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
    }
   } while(0);
   __THREW__ = 0;
   invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
   $159 = __THREW__; __THREW__ = 0;
   $160 = $159&1;
   if (!($160)) {
    ___resumeException($$0|0);
    // unreachable;
   }
  }
 } while(0);
 $209 = ___cxa_find_matching_catch_3(0|0)|0;
 $210 = tempRet0;
 $lpad$phi46$index = $209;
 ___clang_call_terminate($lpad$phi46$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
 // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
}
function __ZN4love8graphics6opengl6Shader20checkSetScreenParamsEv($this) {
 $this = $this|0;
 var $$ = 0.0, $$43 = 0.0, $$cast = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i23 = 0, $$pre$i13 = 0, $$pre2$i = 0, $$pre2$i$i = 0, $$pre2$i$i24 = 0, $$pre2$i14 = 0, $$pre4$i = 0, $$pre4$i$i = 0, $$pre4$i$i21 = 0, $$pre4$i11 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0.0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0.0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$01$i = 0, $i$01$i$i = 0;
 var $i$01$i$i20 = 0, $i$01$i10 = 0, $lpad$phi$index = 0, $lpad$phi38$index = 0, $params = 0, $view = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $view = sp + 16|0;
 $params = sp;
 __ZNK4love8graphics6opengl6OpenGL11getViewportEv($view,634044); //@line 710 "modules/graphics/opengl/Shader.cpp"
 $0 = ((($this)) + 168|0); //@line 712 "modules/graphics/opengl/Shader.cpp"
 $1 = HEAP32[$view>>2]|0; //@line 108 "modules/graphics/opengl/OpenGL.h"
 $2 = HEAP32[$0>>2]|0; //@line 108 "modules/graphics/opengl/OpenGL.h"
 $3 = ($1|0)==($2|0); //@line 108 "modules/graphics/opengl/OpenGL.h"
 if ($3) {
  $4 = ((($view)) + 4|0); //@line 108 "modules/graphics/opengl/OpenGL.h"
  $5 = HEAP32[$4>>2]|0; //@line 108 "modules/graphics/opengl/OpenGL.h"
  $6 = ((($this)) + 172|0); //@line 108 "modules/graphics/opengl/OpenGL.h"
  $7 = HEAP32[$6>>2]|0; //@line 108 "modules/graphics/opengl/OpenGL.h"
  $8 = ($5|0)==($7|0); //@line 108 "modules/graphics/opengl/OpenGL.h"
  if ($8) {
   $9 = ((($view)) + 8|0); //@line 108 "modules/graphics/opengl/OpenGL.h"
   $10 = HEAP32[$9>>2]|0; //@line 108 "modules/graphics/opengl/OpenGL.h"
   $11 = ((($this)) + 176|0); //@line 108 "modules/graphics/opengl/OpenGL.h"
   $12 = HEAP32[$11>>2]|0; //@line 108 "modules/graphics/opengl/OpenGL.h"
   $13 = ($10|0)==($12|0); //@line 108 "modules/graphics/opengl/OpenGL.h"
   if ($13) {
    $14 = ((($view)) + 12|0); //@line 108 "modules/graphics/opengl/OpenGL.h"
    $15 = HEAP32[$14>>2]|0; //@line 108 "modules/graphics/opengl/OpenGL.h"
    $16 = ((($this)) + 180|0); //@line 108 "modules/graphics/opengl/OpenGL.h"
    $17 = HEAP32[$16>>2]|0; //@line 108 "modules/graphics/opengl/OpenGL.h"
    $18 = ($15|0)==($17|0); //@line 108 "modules/graphics/opengl/OpenGL.h"
    if ($18) {
     $19 = ((($this)) + 164|0); //@line 712 "modules/graphics/opengl/Shader.cpp"
     $20 = HEAP32[$19>>2]|0; //@line 712 "modules/graphics/opengl/Shader.cpp"
     $21 = HEAP32[158272]|0; //@line 712 "modules/graphics/opengl/Shader.cpp"
     $22 = ($20|0)==($21|0); //@line 712 "modules/graphics/opengl/Shader.cpp"
     if ($22) {
      STACKTOP = sp;return; //@line 747 "modules/graphics/opengl/Shader.cpp"
     }
    }
   }
  }
 }
 $23 = ((($view)) + 8|0); //@line 719 "modules/graphics/opengl/Shader.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 719 "modules/graphics/opengl/Shader.cpp"
 $25 = (+($24|0)); //@line 719 "modules/graphics/opengl/Shader.cpp"
 HEAPF32[$params>>2] = $25; //@line 718 "modules/graphics/opengl/Shader.cpp"
 $26 = ((($params)) + 4|0); //@line 718 "modules/graphics/opengl/Shader.cpp"
 $27 = ((($view)) + 12|0); //@line 719 "modules/graphics/opengl/Shader.cpp"
 $28 = HEAP32[$27>>2]|0; //@line 734 "modules/graphics/opengl/Shader.cpp"
 $29 = (+($28|0)); //@line 719 "modules/graphics/opengl/Shader.cpp"
 HEAPF32[$26>>2] = $29; //@line 718 "modules/graphics/opengl/Shader.cpp"
 $30 = ((($params)) + 8|0); //@line 718 "modules/graphics/opengl/Shader.cpp"
 HEAPF32[$30>>2] = 0.0; //@line 718 "modules/graphics/opengl/Shader.cpp"
 $31 = ((($params)) + 12|0); //@line 718 "modules/graphics/opengl/Shader.cpp"
 HEAPF32[$31>>2] = 0.0; //@line 718 "modules/graphics/opengl/Shader.cpp"
 $32 = HEAP32[158272]|0; //@line 723 "modules/graphics/opengl/Shader.cpp"
 $33 = ($32|0)==(0|0); //@line 723 "modules/graphics/opengl/Shader.cpp"
 $$ = $33 ? -1.0 : 1.0; //@line 723 "modules/graphics/opengl/Shader.cpp"
 $$43 = $33 ? $29 : 0.0; //@line 723 "modules/graphics/opengl/Shader.cpp"
 HEAPF32[$30>>2] = $$; //@line 733 "modules/graphics/opengl/Shader.cpp"
 HEAPF32[$31>>2] = $$43; //@line 734 "modules/graphics/opengl/Shader.cpp"
 $34 = ((($this)) + 72|0); //@line 737 "modules/graphics/opengl/Shader.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 737 "modules/graphics/opengl/Shader.cpp"
 $36 = ($35|0)>(-1); //@line 739 "modules/graphics/opengl/Shader.cpp"
 L8: do {
  if ($36) {
   $37 = HEAP32[158581]|0; //@line 46 "modules/graphics/opengl/Shader.cpp"
   $$cast = $37; //@line 406 "modules/graphics/opengl/Shader.cpp"
   $38 = ($$cast|0)==($this|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
   if (!($38)) {
    $39 = ((($this)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
    $40 = HEAP32[$39>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
    $41 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
    FUNCTION_TABLE_vi[$41 & 8191]($40); //@line 534 "./libraries/glad/gladfuncs.hpp"
    HEAP32[158581] = $this; //@line 409 "modules/graphics/opengl/Shader.cpp"
   }
   $42 = HEAP32[165650]|0; //@line 546 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   invoke_viii($42|0,($35|0),1,($params|0)); //@line 546 "./libraries/glad/gladfuncs.hpp"
   $43 = __THREW__; __THREW__ = 0;
   $44 = $43&1;
   if (!($44)) {
    $45 = ($37|0)==(0); //@line 53 "modules/graphics/opengl/Shader.cpp"
    do {
     if ($45) {
      $72 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
      $73 = ($72|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
      $74 = HEAP32[158581]|0; //@line 432 "modules/graphics/opengl/Shader.cpp"
      if ($73) {
       $100 = ($74|0)==(0|0); //@line 438 "modules/graphics/opengl/Shader.cpp"
       if (!($100)) {
        $101 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
        __THREW__ = 0;
        invoke_vi($101|0,0); //@line 534 "./libraries/glad/gladfuncs.hpp"
        $102 = __THREW__; __THREW__ = 0;
        $103 = $102&1;
        if ($103) {
         break;
        }
       }
       HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
       break L8;
      }
      $75 = ($74|0)==($72|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
      if ($75) {
       break L8;
      }
      $76 = ((($72)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
      $77 = HEAP32[$76>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
      $78 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_vi($78|0,($77|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
      $79 = __THREW__; __THREW__ = 0;
      $80 = $79&1;
      if (!($80)) {
       HEAP32[158581] = $72; //@line 409 "modules/graphics/opengl/Shader.cpp"
       $81 = ((($72)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
       $82 = ((($72)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $83 = HEAP32[$82>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $84 = HEAP32[$81>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $85 = ($83|0)==($84|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
       L25: do {
        if (!($85)) {
         $180 = $83;$87 = $84;$i$01$i$i20 = 0;
         while(1) {
          $86 = $87; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $88 = (($86) + ($i$01$i$i20<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $89 = HEAP32[$88>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
          $90 = ($89|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
          $$pre4$i$i21 = (($i$01$i$i20) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
          if ($90) {
           $94 = $180;$95 = $87;
          } else {
           __THREW__ = 0;
           invoke_viiii(4111,(634044|0),($89|0),($$pre4$i$i21|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
           $91 = __THREW__; __THREW__ = 0;
           $92 = $91&1;
           if ($92) {
            break;
           }
           $$pre$i$i23 = HEAP32[$82>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $$pre2$i$i24 = HEAP32[$81>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $94 = $$pre$i$i23;$95 = $$pre2$i$i24;
          }
          $93 = (($94) - ($95))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $96 = $93 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $97 = ($$pre4$i$i21>>>0)<($96>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
          if ($97) {
           $180 = $94;$87 = $95;$i$01$i$i20 = $$pre4$i$i21;
          } else {
           break L25;
          }
         }
         $104 = ___cxa_find_matching_catch_3(0|0)|0;
         $105 = tempRet0;
         $lpad$phi$index = $104;
         ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
         // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
        }
       } while(0);
       __THREW__ = 0;
       invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
       $98 = __THREW__; __THREW__ = 0;
       $99 = $98&1;
       if (!($99)) {
        break L8;
       }
      }
     } else {
      $46 = HEAP32[158581]|0; //@line 406 "modules/graphics/opengl/Shader.cpp"
      $47 = ($46|0)==($$cast|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
      if (!($47)) {
       $48 = ((($$cast)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
       $49 = HEAP32[$48>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
       $50 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
       __THREW__ = 0;
       invoke_vi($50|0,($49|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
       $51 = __THREW__; __THREW__ = 0;
       $52 = $51&1;
       if ($52) {
        break;
       }
       HEAP32[158581] = $$cast; //@line 409 "modules/graphics/opengl/Shader.cpp"
      }
      $53 = ((($$cast)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      $54 = ((($$cast)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $55 = HEAP32[$54>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $56 = HEAP32[$53>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $57 = ($55|0)==($56|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      L40: do {
       if (!($57)) {
        $179 = $55;$61 = $56;$i$01$i10 = 0;
        while(1) {
         $60 = $61; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $62 = (($60) + ($i$01$i10<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $63 = HEAP32[$62>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
         $64 = ($63|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
         $$pre4$i11 = (($i$01$i10) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
         if ($64) {
          $68 = $179;$69 = $61;
         } else {
          __THREW__ = 0;
          invoke_viiii(4111,(634044|0),($63|0),($$pre4$i11|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
          $65 = __THREW__; __THREW__ = 0;
          $66 = $65&1;
          if ($66) {
           break;
          }
          $$pre$i13 = HEAP32[$54>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $$pre2$i14 = HEAP32[$53>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $68 = $$pre$i13;$69 = $$pre2$i14;
         }
         $67 = (($68) - ($69))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $70 = $67 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $71 = ($$pre4$i11>>>0)<($70>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
         if ($71) {
          $179 = $68;$61 = $69;$i$01$i10 = $$pre4$i11;
         } else {
          break L40;
         }
        }
        $106 = ___cxa_find_matching_catch_3(0|0)|0;
        $107 = tempRet0;
        $lpad$phi$index = $106;
        ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
        // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
       }
      } while(0);
      __THREW__ = 0;
      invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
      $58 = __THREW__; __THREW__ = 0;
      $59 = $58&1;
      if (!($59)) {
       break L8;
      }
     }
    } while(0);
    $108 = ___cxa_find_matching_catch_3(0|0)|0;
    $109 = tempRet0;
    $lpad$phi$index = $108;
    ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
    // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
   }
   $110 = ___cxa_find_matching_catch_2()|0;
   $111 = tempRet0;
   $112 = ($37|0)==(0); //@line 53 "modules/graphics/opengl/Shader.cpp"
   do {
    if ($112) {
     $139 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
     $140 = ($139|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
     $141 = HEAP32[158581]|0; //@line 432 "modules/graphics/opengl/Shader.cpp"
     if ($140) {
      $167 = ($141|0)==(0|0); //@line 438 "modules/graphics/opengl/Shader.cpp"
      if (!($167)) {
       $168 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
       __THREW__ = 0;
       invoke_vi($168|0,0); //@line 534 "./libraries/glad/gladfuncs.hpp"
       $169 = __THREW__; __THREW__ = 0;
       $170 = $169&1;
       if ($170) {
        break;
       }
      }
      HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
      ___resumeException($110|0);
      // unreachable;
     }
     $142 = ($141|0)==($139|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
     if ($142) {
      ___resumeException($110|0);
      // unreachable;
     }
     $143 = ((($139)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
     $144 = HEAP32[$143>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
     $145 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_vi($145|0,($144|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
     $146 = __THREW__; __THREW__ = 0;
     $147 = $146&1;
     if (!($147)) {
      HEAP32[158581] = $139; //@line 409 "modules/graphics/opengl/Shader.cpp"
      $148 = ((($139)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      $149 = ((($139)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $150 = HEAP32[$149>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $151 = HEAP32[$148>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $152 = ($150|0)==($151|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      L66: do {
       if (!($152)) {
        $154 = $151;$182 = $150;$i$01$i$i = 0;
        while(1) {
         $153 = $154; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $155 = (($153) + ($i$01$i$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $156 = HEAP32[$155>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
         $157 = ($156|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
         $$pre4$i$i = (($i$01$i$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
         if ($157) {
          $161 = $182;$162 = $154;
         } else {
          __THREW__ = 0;
          invoke_viiii(4111,(634044|0),($156|0),($$pre4$i$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
          $158 = __THREW__; __THREW__ = 0;
          $159 = $158&1;
          if ($159) {
           break;
          }
          $$pre$i$i = HEAP32[$149>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $$pre2$i$i = HEAP32[$148>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $161 = $$pre$i$i;$162 = $$pre2$i$i;
         }
         $160 = (($161) - ($162))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $163 = $160 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $164 = ($$pre4$i$i>>>0)<($163>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
         if ($164) {
          $154 = $162;$182 = $161;$i$01$i$i = $$pre4$i$i;
         } else {
          break L66;
         }
        }
        $171 = ___cxa_find_matching_catch_3(0|0)|0;
        $172 = tempRet0;
        $lpad$phi38$index = $171;
        ___clang_call_terminate($lpad$phi38$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
        // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
       }
      } while(0);
      __THREW__ = 0;
      invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
      $165 = __THREW__; __THREW__ = 0;
      $166 = $165&1;
      if (!($166)) {
       ___resumeException($110|0);
       // unreachable;
      }
     }
    } else {
     $113 = HEAP32[158581]|0; //@line 406 "modules/graphics/opengl/Shader.cpp"
     $114 = ($113|0)==($$cast|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
     if (!($114)) {
      $115 = ((($$cast)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
      $116 = HEAP32[$115>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
      $117 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_vi($117|0,($116|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
      $118 = __THREW__; __THREW__ = 0;
      $119 = $118&1;
      if ($119) {
       break;
      }
      HEAP32[158581] = $$cast; //@line 409 "modules/graphics/opengl/Shader.cpp"
     }
     $120 = ((($$cast)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
     $121 = ((($$cast)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $122 = HEAP32[$121>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $123 = HEAP32[$120>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $124 = ($122|0)==($123|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
     L82: do {
      if (!($124)) {
       $128 = $123;$181 = $122;$i$01$i = 0;
       while(1) {
        $127 = $128; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $129 = (($127) + ($i$01$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $130 = HEAP32[$129>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
        $131 = ($130|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
        $$pre4$i = (($i$01$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
        if ($131) {
         $135 = $181;$136 = $128;
        } else {
         __THREW__ = 0;
         invoke_viiii(4111,(634044|0),($130|0),($$pre4$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
         $132 = __THREW__; __THREW__ = 0;
         $133 = $132&1;
         if ($133) {
          break;
         }
         $$pre$i = HEAP32[$121>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $$pre2$i = HEAP32[$120>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $135 = $$pre$i;$136 = $$pre2$i;
        }
        $134 = (($135) - ($136))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $137 = $134 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $138 = ($$pre4$i>>>0)<($137>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
        if ($138) {
         $128 = $136;$181 = $135;$i$01$i = $$pre4$i;
        } else {
         break L82;
        }
       }
       $173 = ___cxa_find_matching_catch_3(0|0)|0;
       $174 = tempRet0;
       $lpad$phi38$index = $173;
       ___clang_call_terminate($lpad$phi38$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
       // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
      }
     } while(0);
     __THREW__ = 0;
     invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
     $125 = __THREW__; __THREW__ = 0;
     $126 = $125&1;
     if (!($126)) {
      ___resumeException($110|0);
      // unreachable;
     }
    }
   } while(0);
   $175 = ___cxa_find_matching_catch_3(0|0)|0;
   $176 = tempRet0;
   $lpad$phi38$index = $175;
   ___clang_call_terminate($lpad$phi38$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
   // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
  }
 } while(0);
 $177 = HEAP32[158272]|0; //@line 745 "modules/graphics/opengl/Shader.cpp"
 $178 = ((($this)) + 164|0); //@line 745 "modules/graphics/opengl/Shader.cpp"
 HEAP32[$178>>2] = $177; //@line 745 "modules/graphics/opengl/Shader.cpp"
 ;HEAP32[$0>>2]=HEAP32[$view>>2]|0;HEAP32[$0+4>>2]=HEAP32[$view+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$view+8>>2]|0;HEAP32[$0+12>>2]=HEAP32[$view+12>>2]|0; //@line 746 "modules/graphics/opengl/Shader.cpp"
 STACKTOP = sp;return; //@line 747 "modules/graphics/opengl/Shader.cpp"
}
function __ZN4love8graphics6opengl6Shader17checkSetPointSizeEf($this,$size) {
 $this = $this|0;
 $size = +$size;
 var $$cast = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i23 = 0, $$pre$i13 = 0, $$pre2$i = 0, $$pre2$i$i = 0, $$pre2$i$i24 = 0, $$pre2$i14 = 0, $$pre4$i = 0, $$pre4$i$i = 0, $$pre4$i$i21 = 0, $$pre4$i11 = 0, $0 = 0, $1 = 0.0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $i$01$i = 0, $i$01$i$i = 0, $i$01$i$i20 = 0, $i$01$i10 = 0, $lpad$phi$index = 0, $lpad$phi38$index = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 184|0); //@line 751 "modules/graphics/opengl/Shader.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 751 "modules/graphics/opengl/Shader.cpp"
 $2 = $1 == $size; //@line 751 "modules/graphics/opengl/Shader.cpp"
 if ($2) {
  return; //@line 763 "modules/graphics/opengl/Shader.cpp"
 }
 $3 = ((($this)) + 68|0); //@line 754 "modules/graphics/opengl/Shader.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 754 "modules/graphics/opengl/Shader.cpp"
 $5 = ($4|0)>(-1); //@line 756 "modules/graphics/opengl/Shader.cpp"
 L4: do {
  if ($5) {
   $6 = HEAP32[158581]|0; //@line 46 "modules/graphics/opengl/Shader.cpp"
   $$cast = $6; //@line 406 "modules/graphics/opengl/Shader.cpp"
   $7 = ($$cast|0)==($this|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
   if (!($7)) {
    $8 = ((($this)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
    $9 = HEAP32[$8>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
    $10 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
    FUNCTION_TABLE_vi[$10 & 8191]($9); //@line 534 "./libraries/glad/gladfuncs.hpp"
    HEAP32[158581] = $this; //@line 409 "modules/graphics/opengl/Shader.cpp"
   }
   $11 = HEAP32[165181]|0; //@line 535 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   invoke_vid($11|0,($4|0),(+$size)); //@line 535 "./libraries/glad/gladfuncs.hpp"
   $12 = __THREW__; __THREW__ = 0;
   $13 = $12&1;
   if (!($13)) {
    $14 = ($6|0)==(0); //@line 53 "modules/graphics/opengl/Shader.cpp"
    do {
     if ($14) {
      $41 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
      $42 = ($41|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
      $43 = HEAP32[158581]|0; //@line 432 "modules/graphics/opengl/Shader.cpp"
      if ($42) {
       $69 = ($43|0)==(0|0); //@line 438 "modules/graphics/opengl/Shader.cpp"
       if (!($69)) {
        $70 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
        __THREW__ = 0;
        invoke_vi($70|0,0); //@line 534 "./libraries/glad/gladfuncs.hpp"
        $71 = __THREW__; __THREW__ = 0;
        $72 = $71&1;
        if ($72) {
         break;
        }
       }
       HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
       break L4;
      }
      $44 = ($43|0)==($41|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
      if ($44) {
       break L4;
      }
      $45 = ((($41)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
      $46 = HEAP32[$45>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
      $47 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_vi($47|0,($46|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
      $48 = __THREW__; __THREW__ = 0;
      $49 = $48&1;
      if (!($49)) {
       HEAP32[158581] = $41; //@line 409 "modules/graphics/opengl/Shader.cpp"
       $50 = ((($41)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
       $51 = ((($41)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $52 = HEAP32[$51>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $53 = HEAP32[$50>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $54 = ($52|0)==($53|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
       L21: do {
        if (!($54)) {
         $147 = $52;$56 = $53;$i$01$i$i = 0;
         while(1) {
          $55 = $56; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $57 = (($55) + ($i$01$i$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $58 = HEAP32[$57>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
          $59 = ($58|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
          $$pre4$i$i = (($i$01$i$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
          if ($59) {
           $63 = $147;$64 = $56;
          } else {
           __THREW__ = 0;
           invoke_viiii(4111,(634044|0),($58|0),($$pre4$i$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
           $60 = __THREW__; __THREW__ = 0;
           $61 = $60&1;
           if ($61) {
            break;
           }
           $$pre$i$i = HEAP32[$51>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $$pre2$i$i = HEAP32[$50>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $63 = $$pre$i$i;$64 = $$pre2$i$i;
          }
          $62 = (($63) - ($64))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $65 = $62 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $66 = ($$pre4$i$i>>>0)<($65>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
          if ($66) {
           $147 = $63;$56 = $64;$i$01$i$i = $$pre4$i$i;
          } else {
           break L21;
          }
         }
         $73 = ___cxa_find_matching_catch_3(0|0)|0;
         $74 = tempRet0;
         $lpad$phi$index = $73;
         ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
         // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
        }
       } while(0);
       __THREW__ = 0;
       invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
       $67 = __THREW__; __THREW__ = 0;
       $68 = $67&1;
       if (!($68)) {
        break L4;
       }
      }
     } else {
      $15 = HEAP32[158581]|0; //@line 406 "modules/graphics/opengl/Shader.cpp"
      $16 = ($15|0)==($$cast|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
      if (!($16)) {
       $17 = ((($$cast)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
       $18 = HEAP32[$17>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
       $19 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
       __THREW__ = 0;
       invoke_vi($19|0,($18|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
       $20 = __THREW__; __THREW__ = 0;
       $21 = $20&1;
       if ($21) {
        break;
       }
       HEAP32[158581] = $$cast; //@line 409 "modules/graphics/opengl/Shader.cpp"
      }
      $22 = ((($$cast)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      $23 = ((($$cast)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $24 = HEAP32[$23>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $25 = HEAP32[$22>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $26 = ($24|0)==($25|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      L36: do {
       if (!($26)) {
        $146 = $24;$30 = $25;$i$01$i = 0;
        while(1) {
         $29 = $30; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $31 = (($29) + ($i$01$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $32 = HEAP32[$31>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
         $33 = ($32|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
         $$pre4$i = (($i$01$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
         if ($33) {
          $37 = $146;$38 = $30;
         } else {
          __THREW__ = 0;
          invoke_viiii(4111,(634044|0),($32|0),($$pre4$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
          $34 = __THREW__; __THREW__ = 0;
          $35 = $34&1;
          if ($35) {
           break;
          }
          $$pre$i = HEAP32[$23>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $$pre2$i = HEAP32[$22>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $37 = $$pre$i;$38 = $$pre2$i;
         }
         $36 = (($37) - ($38))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $39 = $36 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $40 = ($$pre4$i>>>0)<($39>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
         if ($40) {
          $146 = $37;$30 = $38;$i$01$i = $$pre4$i;
         } else {
          break L36;
         }
        }
        $75 = ___cxa_find_matching_catch_3(0|0)|0;
        $76 = tempRet0;
        $lpad$phi$index = $75;
        ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
        // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
       }
      } while(0);
      __THREW__ = 0;
      invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
      $27 = __THREW__; __THREW__ = 0;
      $28 = $27&1;
      if (!($28)) {
       break L4;
      }
     }
    } while(0);
    $77 = ___cxa_find_matching_catch_3(0|0)|0;
    $78 = tempRet0;
    $lpad$phi$index = $77;
    ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
    // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
   }
   $79 = ___cxa_find_matching_catch_2()|0;
   $80 = tempRet0;
   $81 = ($6|0)==(0); //@line 53 "modules/graphics/opengl/Shader.cpp"
   do {
    if ($81) {
     $108 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
     $109 = ($108|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
     $110 = HEAP32[158581]|0; //@line 432 "modules/graphics/opengl/Shader.cpp"
     if ($109) {
      $136 = ($110|0)==(0|0); //@line 438 "modules/graphics/opengl/Shader.cpp"
      if (!($136)) {
       $137 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
       __THREW__ = 0;
       invoke_vi($137|0,0); //@line 534 "./libraries/glad/gladfuncs.hpp"
       $138 = __THREW__; __THREW__ = 0;
       $139 = $138&1;
       if ($139) {
        break;
       }
      }
      HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
      ___resumeException($79|0);
      // unreachable;
     }
     $111 = ($110|0)==($108|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
     if ($111) {
      ___resumeException($79|0);
      // unreachable;
     }
     $112 = ((($108)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
     $113 = HEAP32[$112>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
     $114 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_vi($114|0,($113|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
     $115 = __THREW__; __THREW__ = 0;
     $116 = $115&1;
     if (!($116)) {
      HEAP32[158581] = $108; //@line 409 "modules/graphics/opengl/Shader.cpp"
      $117 = ((($108)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      $118 = ((($108)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $119 = HEAP32[$118>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $120 = HEAP32[$117>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $121 = ($119|0)==($120|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      L62: do {
       if (!($121)) {
        $123 = $120;$149 = $119;$i$01$i$i20 = 0;
        while(1) {
         $122 = $123; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $124 = (($122) + ($i$01$i$i20<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $125 = HEAP32[$124>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
         $126 = ($125|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
         $$pre4$i$i21 = (($i$01$i$i20) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
         if ($126) {
          $130 = $149;$131 = $123;
         } else {
          __THREW__ = 0;
          invoke_viiii(4111,(634044|0),($125|0),($$pre4$i$i21|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
          $127 = __THREW__; __THREW__ = 0;
          $128 = $127&1;
          if ($128) {
           break;
          }
          $$pre$i$i23 = HEAP32[$118>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $$pre2$i$i24 = HEAP32[$117>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $130 = $$pre$i$i23;$131 = $$pre2$i$i24;
         }
         $129 = (($130) - ($131))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $132 = $129 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $133 = ($$pre4$i$i21>>>0)<($132>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
         if ($133) {
          $123 = $131;$149 = $130;$i$01$i$i20 = $$pre4$i$i21;
         } else {
          break L62;
         }
        }
        $140 = ___cxa_find_matching_catch_3(0|0)|0;
        $141 = tempRet0;
        $lpad$phi38$index = $140;
        ___clang_call_terminate($lpad$phi38$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
        // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
       }
      } while(0);
      __THREW__ = 0;
      invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
      $134 = __THREW__; __THREW__ = 0;
      $135 = $134&1;
      if (!($135)) {
       ___resumeException($79|0);
       // unreachable;
      }
     }
    } else {
     $82 = HEAP32[158581]|0; //@line 406 "modules/graphics/opengl/Shader.cpp"
     $83 = ($82|0)==($$cast|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
     if (!($83)) {
      $84 = ((($$cast)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
      $85 = HEAP32[$84>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
      $86 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_vi($86|0,($85|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
      $87 = __THREW__; __THREW__ = 0;
      $88 = $87&1;
      if ($88) {
       break;
      }
      HEAP32[158581] = $$cast; //@line 409 "modules/graphics/opengl/Shader.cpp"
     }
     $89 = ((($$cast)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
     $90 = ((($$cast)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $91 = HEAP32[$90>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $92 = HEAP32[$89>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $93 = ($91|0)==($92|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
     L78: do {
      if (!($93)) {
       $148 = $91;$97 = $92;$i$01$i10 = 0;
       while(1) {
        $96 = $97; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $98 = (($96) + ($i$01$i10<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $99 = HEAP32[$98>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
        $100 = ($99|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
        $$pre4$i11 = (($i$01$i10) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
        if ($100) {
         $104 = $148;$105 = $97;
        } else {
         __THREW__ = 0;
         invoke_viiii(4111,(634044|0),($99|0),($$pre4$i11|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
         $101 = __THREW__; __THREW__ = 0;
         $102 = $101&1;
         if ($102) {
          break;
         }
         $$pre$i13 = HEAP32[$90>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $$pre2$i14 = HEAP32[$89>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $104 = $$pre$i13;$105 = $$pre2$i14;
        }
        $103 = (($104) - ($105))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $106 = $103 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $107 = ($$pre4$i11>>>0)<($106>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
        if ($107) {
         $148 = $104;$97 = $105;$i$01$i10 = $$pre4$i11;
        } else {
         break L78;
        }
       }
       $142 = ___cxa_find_matching_catch_3(0|0)|0;
       $143 = tempRet0;
       $lpad$phi38$index = $142;
       ___clang_call_terminate($lpad$phi38$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
       // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
      }
     } while(0);
     __THREW__ = 0;
     invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
     $94 = __THREW__; __THREW__ = 0;
     $95 = $94&1;
     if (!($95)) {
      ___resumeException($79|0);
      // unreachable;
     }
    }
   } while(0);
   $144 = ___cxa_find_matching_catch_3(0|0)|0;
   $145 = tempRet0;
   $lpad$phi38$index = $144;
   ___clang_call_terminate($lpad$phi38$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
   // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
  }
 } while(0);
 HEAPF32[$0>>2] = $size; //@line 762 "modules/graphics/opengl/Shader.cpp"
 return; //@line 763 "modules/graphics/opengl/Shader.cpp"
}
function __ZN4love8graphics6opengl6Shader14unloadVolatileEv($this) {
 $this = $this|0;
 var $$cast10 = 0, $$lcssa8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $__new_last2$i$i$i1 = 0, $i$011 = 0, $scevgep = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i2 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i3 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = HEAP32[158581]|0; //@line 335 "modules/graphics/opengl/Shader.cpp"
 $2 = ($1|0)==($this|0); //@line 335 "modules/graphics/opengl/Shader.cpp"
 if ($2) {
  $3 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$3 & 8191](0); //@line 534 "./libraries/glad/gladfuncs.hpp"
 }
 $4 = ((($this)) + 48|0); //@line 338 "modules/graphics/opengl/Shader.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 338 "modules/graphics/opengl/Shader.cpp"
 $6 = ($5|0)==(0); //@line 338 "modules/graphics/opengl/Shader.cpp"
 if (!($6)) {
  $7 = HEAP32[164697]|0; //@line 509 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$7 & 8191]($5); //@line 509 "./libraries/glad/gladfuncs.hpp"
  HEAP32[$4>>2] = 0; //@line 341 "modules/graphics/opengl/Shader.cpp"
 }
 $8 = ((($this)) + 140|0); //@line 345 "modules/graphics/opengl/Shader.cpp"
 $9 = ((($this)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = HEAP32[$9>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = HEAP32[$8>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($10|0)==($11|0); //@line 345 "modules/graphics/opengl/Shader.cpp"
 $$cast10 = $11; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = $10; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $$lcssa8 = $10;
 } else {
  $14 = (($10) - ($11))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = $14 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = HEAP32[158608]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $i$011 = 0;
  while(1) {
   $51 = (($$cast10) + ($i$011<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $52 = HEAP32[$51>>2]|0; //@line 347 "modules/graphics/opengl/Shader.cpp"
   $53 = ($52|0)==(0); //@line 347 "modules/graphics/opengl/Shader.cpp"
   if (!($53)) {
    $54 = (($16) + ($i$011<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $55 = HEAP32[$54>>2]|0; //@line 348 "modules/graphics/opengl/Shader.cpp"
    $56 = (($55) + -1)|0; //@line 348 "modules/graphics/opengl/Shader.cpp"
    $57 = ($56|0)<(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $58 = $57 ? 0 : $56; //@line 348 "modules/graphics/opengl/Shader.cpp"
    HEAP32[$54>>2] = $58; //@line 348 "modules/graphics/opengl/Shader.cpp"
   }
   $59 = (($i$011) + 1)|0; //@line 345 "modules/graphics/opengl/Shader.cpp"
   $60 = ($59>>>0)<($15>>>0); //@line 345 "modules/graphics/opengl/Shader.cpp"
   if ($60) {
    $i$011 = $59;
   } else {
    $$lcssa8 = $11;
    break;
   }
  }
 }
 $17 = ($13|0)==($$cast10|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($17)) {
  $scevgep$i$i$i = ((($13)) + -4|0);
  $18 = $scevgep$i$i$i;
  $19 = (($18) - ($$lcssa8))|0;
  $20 = $19 >>> 2;
  $21 = $20 ^ -1;
  $scevgep4$i$i$i = (($13) + ($21<<2)|0);
  HEAP32[$9>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $22 = (__ZNK4love8graphics6opengl6OpenGL18getMaxTextureUnitsEv(634044)|0); //@line 353 "modules/graphics/opengl/Shader.cpp"
 $23 = (($22) + -1)|0; //@line 353 "modules/graphics/opengl/Shader.cpp"
 HEAP32[$0>>2] = 0; //@line 353 "modules/graphics/opengl/Shader.cpp"
 $24 = HEAP32[$9>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $25 = HEAP32[$8>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $26 = (($24) - ($25))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = $26 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $28 = ($23>>>0)>($27>>>0); //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $29 = $25; //@line 2012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = $24; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($28) {
  $31 = (($23) - ($27))|0; //@line 2010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  __ZNSt3__16vectorIjNS_9allocatorIjEEE8__appendEjRKj($8,$31,$0); //@line 2010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 } else {
  $32 = ($23>>>0)<($27>>>0); //@line 2011 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($32) {
   $33 = (($29) + ($23<<2)|0); //@line 2012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $34 = ($30|0)==($33|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($34)) {
    $__new_last2$i$i$i1 = $33; //@line 818 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $scevgep$i$i$i2 = ((($30)) + -4|0);
    $35 = $scevgep$i$i$i2;
    $36 = (($35) - ($__new_last2$i$i$i1))|0;
    $37 = $36 >>> 2;
    $38 = $37 ^ -1;
    $scevgep4$i$i$i3 = (($30) + ($38<<2)|0);
    HEAP32[$9>>2] = $scevgep4$i$i$i3; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
  }
 }
 $39 = ((($this)) + 104|0); //@line 355 "modules/graphics/opengl/Shader.cpp"
 $40 = ((($this)) + 108|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $41 = HEAP32[$40>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($39,$41); //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $42 = ((($this)) + 112|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$42>>2] = 0; //@line 1478 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$39>>2] = $40; //@line 1479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$40>>2] = 0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $43 = ((($this)) + 116|0); //@line 358 "modules/graphics/opengl/Shader.cpp"
 $44 = ((($this)) + 120|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $45 = HEAP32[$44>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformEEENS_19__map_value_compareIS7_SD_NS_4lessIS7_EELb1EEENS5_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($43,$45); //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $46 = ((($this)) + 124|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$46>>2] = 0; //@line 1478 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$43>>2] = $44; //@line 1479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$44>>2] = 0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $scevgep = ((($this)) + 52|0); //@line 361 "modules/graphics/opengl/Shader.cpp"
 dest=$scevgep; stop=dest+36|0; do { HEAP32[dest>>2]=4294967295|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 362 "modules/graphics/opengl/Shader.cpp"
 $47 = ((($this)) + 36|0); //@line 364 "modules/graphics/opengl/Shader.cpp"
 $48 = ((($this)) + 40|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $49 = HEAP32[$48>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeIN4love8graphics6opengl6Shader11ShaderStageENS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEEENS_19__map_value_compareIS6_SD_NS_4lessIS6_EELb1EEENSA_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($47,$49); //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $50 = ((($this)) + 44|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$50>>2] = 0; //@line 1478 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$47>>2] = $48; //@line 1479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$48>>2] = 0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 STACKTOP = sp;return; //@line 365 "modules/graphics/opengl/Shader.cpp"
}
function __ZThn8_N4love8graphics6opengl6ShaderD1Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 95 "modules/graphics/opengl/Shader.cpp"
 __ZN4love8graphics6opengl6ShaderD2Ev($0); //@line 95 "modules/graphics/opengl/Shader.cpp"
 return; //@line 95 "modules/graphics/opengl/Shader.cpp"
}
function __ZThn8_N4love8graphics6opengl6ShaderD0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 95 "modules/graphics/opengl/Shader.cpp"
 __ZN4love8graphics6opengl6ShaderD2Ev($0); //@line 96 "modules/graphics/opengl/Shader.cpp"
 __ZdlPv($0); //@line 96 "modules/graphics/opengl/Shader.cpp"
 return; //@line 95 "modules/graphics/opengl/Shader.cpp"
}
function __ZThn8_N4love8graphics6opengl6Shader12loadVolatileEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 229 "modules/graphics/opengl/Shader.cpp"
 (__ZN4love8graphics6opengl6Shader12loadVolatileEv($0)|0); //@line 229 "modules/graphics/opengl/Shader.cpp"
 return 1; //@line 229 "modules/graphics/opengl/Shader.cpp"
}
function __ZThn8_N4love8graphics6opengl6Shader14unloadVolatileEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 333 "modules/graphics/opengl/Shader.cpp"
 __ZN4love8graphics6opengl6Shader14unloadVolatileEv($0); //@line 333 "modules/graphics/opengl/Shader.cpp"
 return; //@line 333 "modules/graphics/opengl/Shader.cpp"
}
function __GLOBAL__sub_I_Shader_cpp() {
 var $$02$i$i$1$i = 0, $$02$i$i$i = 0, $$lcssa$i$1$i = 0, $$lcssa$i$i = 0, $$lcssa10$i$1$i = 0, $$lcssa10$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $hash$0$lcssa$i$i$1$i = 0, $hash$0$lcssa$i$i$i = 0, $hash$01$i$i$1$i = 0, $hash$01$i$i$i = 0;
 var $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 dest=634336; stop=dest+48|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 (___cxa_atexit((4124|0),(0|0),(___dso_handle|0))|0); //@line 69 "modules/graphics/opengl/Shader.cpp"
 dest=634384; stop=dest+48|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 (___cxa_atexit((4125|0),(0|0),(___dso_handle|0))|0); //@line 70 "modules/graphics/opengl/Shader.cpp"
 HEAP32[158608] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[(634436)>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[(634440)>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 (___cxa_atexit((4126|0),(634432|0),(___dso_handle|0))|0); //@line 72 "modules/graphics/opengl/Shader.cpp"
 HEAP8[(634452)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634464)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634476)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634488)>>0] = 0; //@line 155 "./common/StringMap.h"
 $0 = (634492); //@line 44 "./common/StringMap.h"
 $1 = $0; //@line 44 "./common/StringMap.h"
 HEAP32[$1>>2] = 0; //@line 44 "./common/StringMap.h"
 $2 = (($0) + 4)|0; //@line 44 "./common/StringMap.h"
 $3 = $2; //@line 44 "./common/StringMap.h"
 HEAP32[$3>>2] = 0; //@line 44 "./common/StringMap.h"
 $4 = HEAP32[6315]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(25264)>>2]|0; //@line 49 "./common/StringMap.h"
 $6 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
 $7 = ($6<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($7) {
  $hash$0$lcssa$i$i$i = 5381;
 } else {
  $$02$i$i$i = $4;$9 = $6;$hash$01$i$i$i = 5381;
  while(1) {
   $8 = $9 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $10 = ((($$02$i$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $11 = ($hash$01$i$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $12 = (($11) + ($8))|0; //@line 143 "./common/StringMap.h"
   $13 = HEAP8[$10>>0]|0; //@line 142 "./common/StringMap.h"
   $14 = ($13<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($14) {
    $hash$0$lcssa$i$i$i = $12;
    break;
   } else {
    $$02$i$i$i = $10;$9 = $13;$hash$01$i$i$i = $12;
   }
  }
 }
 $15 = $hash$0$lcssa$i$i$i & 3; //@line 112 "./common/StringMap.h"
 $16 = (((634444 + (($15*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $17 = HEAP8[$16>>0]|0; //@line 114 "./common/StringMap.h"
 $18 = ($17<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($18) {
  $$lcssa$i$i = $15;$$lcssa10$i$i = $16;
  label = 5;
 } else {
  $19 = (($hash$0$lcssa$i$i$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $20 = $19 & 3; //@line 112 "./common/StringMap.h"
  $21 = (((634444 + (($20*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $22 = HEAP8[$21>>0]|0; //@line 114 "./common/StringMap.h"
  $23 = ($22<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($23) {
   $$lcssa$i$i = $20;$$lcssa10$i$i = $21;
   label = 5;
  } else {
   $28 = (($hash$0$lcssa$i$i$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $29 = $28 & 3; //@line 112 "./common/StringMap.h"
   $30 = (((634444 + (($29*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $31 = HEAP8[$30>>0]|0; //@line 114 "./common/StringMap.h"
   $32 = ($31<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($32) {
    $$lcssa$i$i = $29;$$lcssa10$i$i = $30;
    label = 5;
   } else {
    $33 = (($hash$0$lcssa$i$i$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $34 = $33 & 3; //@line 112 "./common/StringMap.h"
    $35 = (((634444 + (($34*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $36 = HEAP8[$35>>0]|0; //@line 114 "./common/StringMap.h"
    $37 = ($36<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($37) {
     $$lcssa$i$i = $34;$$lcssa10$i$i = $35;
     label = 5;
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10$i$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $24 = (634444 + (($$lcssa$i$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$24>>2] = $4; //@line 118 "./common/StringMap.h"
  $25 = (((634444 + (($$lcssa$i$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$25>>2] = $5; //@line 119 "./common/StringMap.h"
 }
 $26 = ($5>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($26) {
  HEAP32[$vararg_buffer>>2] = $4; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $5; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
 } else {
  $27 = (634492 + ($5<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$27>>2] = $4; //@line 132 "./common/StringMap.h"
 }
 $38 = HEAP32[(25268)>>2]|0; //@line 49 "./common/StringMap.h"
 $39 = HEAP32[(25272)>>2]|0; //@line 49 "./common/StringMap.h"
 $40 = HEAP8[$38>>0]|0; //@line 142 "./common/StringMap.h"
 $41 = ($40<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($41) {
  $hash$0$lcssa$i$i$1$i = 5381;
 } else {
  $$02$i$i$1$i = $38;$43 = $40;$hash$01$i$i$1$i = 5381;
  while(1) {
   $42 = $43 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $44 = ((($$02$i$i$1$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $45 = ($hash$01$i$i$1$i*33)|0; //@line 143 "./common/StringMap.h"
   $46 = (($45) + ($42))|0; //@line 143 "./common/StringMap.h"
   $47 = HEAP8[$44>>0]|0; //@line 142 "./common/StringMap.h"
   $48 = ($47<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($48) {
    $hash$0$lcssa$i$i$1$i = $46;
    break;
   } else {
    $$02$i$i$1$i = $44;$43 = $47;$hash$01$i$i$1$i = $46;
   }
  }
 }
 $49 = $hash$0$lcssa$i$i$1$i & 3; //@line 112 "./common/StringMap.h"
 $50 = (((634444 + (($49*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $51 = HEAP8[$50>>0]|0; //@line 114 "./common/StringMap.h"
 $52 = ($51<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($52) {
  $$lcssa$i$1$i = $49;$$lcssa10$i$1$i = $50;
  label = 17;
 } else {
  $53 = (($hash$0$lcssa$i$i$1$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $54 = $53 & 3; //@line 112 "./common/StringMap.h"
  $55 = (((634444 + (($54*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $56 = HEAP8[$55>>0]|0; //@line 114 "./common/StringMap.h"
  $57 = ($56<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($57) {
   $$lcssa$i$1$i = $54;$$lcssa10$i$1$i = $55;
   label = 17;
  } else {
   $58 = (($hash$0$lcssa$i$i$1$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $59 = $58 & 3; //@line 112 "./common/StringMap.h"
   $60 = (((634444 + (($59*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $61 = HEAP8[$60>>0]|0; //@line 114 "./common/StringMap.h"
   $62 = ($61<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($62) {
    $$lcssa$i$1$i = $59;$$lcssa10$i$1$i = $60;
    label = 17;
   } else {
    $63 = (($hash$0$lcssa$i$i$1$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $64 = $63 & 3; //@line 112 "./common/StringMap.h"
    $65 = (((634444 + (($64*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $66 = HEAP8[$65>>0]|0; //@line 114 "./common/StringMap.h"
    $67 = ($66<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($67) {
     $$lcssa$i$1$i = $64;$$lcssa10$i$1$i = $65;
     label = 17;
    }
   }
  }
 }
 if ((label|0) == 17) {
  HEAP8[$$lcssa10$i$1$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $68 = (634444 + (($$lcssa$i$1$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$68>>2] = $38; //@line 118 "./common/StringMap.h"
  $69 = (((634444 + (($$lcssa$i$1$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$69>>2] = $39; //@line 119 "./common/StringMap.h"
 }
 $70 = ($39>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($70) {
  HEAP32[$vararg_buffer2>>2] = $38; //@line 128 "./common/StringMap.h"
  $vararg_ptr5 = ((($vararg_buffer2)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr5>>2] = $39; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer2)|0); //@line 128 "./common/StringMap.h"
 } else {
  $71 = (634492 + ($39<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$71>>2] = $38; //@line 132 "./common/StringMap.h"
 }
 HEAP8[(634872)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634884)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634896)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634908)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634920)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634932)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634944)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634956)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634968)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634980)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(634984)>>2]=0|0;HEAP32[(634984)+4>>2]=0|0;HEAP32[(634984)+8>>2]=0|0;HEAP32[(634984)+12>>2]=0|0;HEAP32[(634984)+16>>2]=0|0; //@line 44 "./common/StringMap.h"
 $72 = HEAP32[6319]|0; //@line 49 "./common/StringMap.h"
 $73 = HEAP32[(25280)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl6Shader11UniformTypeELj5EE3addEPKcS4_(634864,$72,$73)|0); //@line 49 "./common/StringMap.h"
 $74 = HEAP32[(25284)>>2]|0; //@line 49 "./common/StringMap.h"
 $75 = HEAP32[(25288)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl6Shader11UniformTypeELj5EE3addEPKcS4_(634864,$74,$75)|0); //@line 49 "./common/StringMap.h"
 $76 = HEAP32[(25292)>>2]|0; //@line 49 "./common/StringMap.h"
 $77 = HEAP32[(25296)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl6Shader11UniformTypeELj5EE3addEPKcS4_(634864,$76,$77)|0); //@line 49 "./common/StringMap.h"
 $78 = HEAP32[(25300)>>2]|0; //@line 49 "./common/StringMap.h"
 $79 = HEAP32[(25304)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl6Shader11UniformTypeELj5EE3addEPKcS4_(634864,$78,$79)|0); //@line 49 "./common/StringMap.h"
 $80 = HEAP32[(25308)>>2]|0; //@line 49 "./common/StringMap.h"
 $81 = HEAP32[(25312)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl6Shader11UniformTypeELj5EE3addEPKcS4_(634864,$80,$81)|0); //@line 49 "./common/StringMap.h"
 HEAP8[(634508)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634520)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634532)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634544)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634556)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634568)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634580)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(634592)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(634596)>>2]=0|0;HEAP32[(634596)+4>>2]=0|0;HEAP32[(634596)+8>>2]=0|0;HEAP32[(634596)+12>>2]=0|0; //@line 44 "./common/StringMap.h"
 $82 = HEAP32[6329]|0; //@line 49 "./common/StringMap.h"
 $83 = HEAP32[(25320)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl14VertexAttribIDELj4EE3addEPKcS3_(634500,$82,$83)|0); //@line 49 "./common/StringMap.h"
 $84 = HEAP32[(25324)>>2]|0; //@line 49 "./common/StringMap.h"
 $85 = HEAP32[(25328)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl14VertexAttribIDELj4EE3addEPKcS3_(634500,$84,$85)|0); //@line 49 "./common/StringMap.h"
 $86 = HEAP32[(25332)>>2]|0; //@line 49 "./common/StringMap.h"
 $87 = HEAP32[(25336)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl14VertexAttribIDELj4EE3addEPKcS3_(634500,$86,$87)|0); //@line 49 "./common/StringMap.h"
 $88 = HEAP32[(25340)>>2]|0; //@line 49 "./common/StringMap.h"
 $89 = HEAP32[(25344)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8graphics6opengl14VertexAttribIDELj4EE3addEPKcS3_(634500,$88,$89)|0); //@line 49 "./common/StringMap.h"
 __ZN4love9StringMapINS_8graphics6opengl6Shader14BuiltinUniformELj9EEC2EPKNS5_5EntryEj(634612,25348,72); //@line 996 "modules/graphics/opengl/Shader.cpp"
 STACKTOP = sp;return;
}
function ___cxx_global_array_dtor_3998($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((634372)); //@line 84 "modules/graphics/opengl/Shader.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((634360)); //@line 84 "modules/graphics/opengl/Shader.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((634348)); //@line 84 "modules/graphics/opengl/Shader.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev(634336); //@line 84 "modules/graphics/opengl/Shader.h"
 return; //@line 69 "modules/graphics/opengl/Shader.cpp"
}
function ___cxx_global_array_dtor_2($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((634420)); //@line 84 "modules/graphics/opengl/Shader.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((634408)); //@line 84 "modules/graphics/opengl/Shader.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((634396)); //@line 84 "modules/graphics/opengl/Shader.h"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev(634384); //@line 84 "modules/graphics/opengl/Shader.h"
 return; //@line 70 "modules/graphics/opengl/Shader.cpp"
}
function __ZNSt3__113__vector_baseIiNS_9allocatorIiEEED2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i = 0, $scevgep4$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ($0|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = $0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($1) {
  return; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $3 = ((($this)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$3>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($4|0)==($0|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($5)) {
  $scevgep$i$i = ((($4)) + -4|0);
  $6 = $scevgep$i$i;
  $7 = (($6) - ($2))|0;
  $8 = $7 >>> 2;
  $9 = $8 ^ -1;
  $scevgep4$i$i = (($4) + ($9<<2)|0);
  HEAP32[$3>>2] = $scevgep4$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($0); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love9StringMapINS_8graphics6opengl6Shader11UniformTypeELj5EE3addEPKcS4_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $42 = (($hash$0$lcssa$i) + 6)|0; //@line 112 "./common/StringMap.h"
       $43 = (($42>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
       $44 = (((($this) + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
       $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
       $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
       if ($46) {
        $$lcssa = $43;$$lcssa10 = $44;
        label = 5;
       } else {
        $47 = (($hash$0$lcssa$i) + 7)|0; //@line 112 "./common/StringMap.h"
        $48 = (($47>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
        $49 = (((($this) + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
        $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
        $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
        if ($51) {
         $$lcssa = $48;$$lcssa10 = $49;
         label = 5;
        } else {
         $52 = (($hash$0$lcssa$i) + 8)|0; //@line 112 "./common/StringMap.h"
         $53 = (($52>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
         $54 = (((($this) + (($53*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
         $55 = HEAP8[$54>>0]|0; //@line 114 "./common/StringMap.h"
         $56 = ($55<<24>>24)==(0); //@line 114 "./common/StringMap.h"
         if ($56) {
          $$lcssa = $53;$$lcssa10 = $54;
          label = 5;
         } else {
          $57 = (($hash$0$lcssa$i) + 9)|0; //@line 112 "./common/StringMap.h"
          $58 = (($57>>>0) % 10)&-1; //@line 112 "./common/StringMap.h"
          $59 = (((($this) + (($58*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
          $60 = HEAP8[$59>>0]|0; //@line 114 "./common/StringMap.h"
          $61 = ($60<<24>>24)==(0); //@line 114 "./common/StringMap.h"
          if ($61) {
           $$lcssa = $58;$$lcssa10 = $59;
           label = 5;
          } else {
           $inserted$2$off0 = 0;
          }
         }
        }
       }
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(4); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 120|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love9StringMapINS_8graphics6opengl14VertexAttribIDELj4EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0;
 var $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 7; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = $13 & 7; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = $22 & 7; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = $27 & 7; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = $32 & 7; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = $37 & 7; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $42 = (($hash$0$lcssa$i) + 6)|0; //@line 112 "./common/StringMap.h"
       $43 = $42 & 7; //@line 112 "./common/StringMap.h"
       $44 = (((($this) + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
       $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
       $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
       if ($46) {
        $$lcssa = $43;$$lcssa10 = $44;
        label = 5;
       } else {
        $47 = (($hash$0$lcssa$i) + 7)|0; //@line 112 "./common/StringMap.h"
        $48 = $47 & 7; //@line 112 "./common/StringMap.h"
        $49 = (((($this) + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
        $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
        $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
        if ($51) {
         $$lcssa = $48;$$lcssa10 = $49;
         label = 5;
        } else {
         $inserted$2$off0 = 0;
        }
       }
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(3); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 96|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love9StringMapINS_8graphics6opengl6Shader14BuiltinUniformELj9EEC2EPKNS5_5EntryEj($this,$entries,$num) {
 $this = $this|0;
 $entries = $entries|0;
 $num = $num|0;
 var $$02$i$i = 0, $$lcssa26 = 0, $$lcssa28 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$08$i = 0, $i1$05 = 0, $scevgep = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 8|0); //@line 155 "./common/StringMap.h"
 HEAP8[$0>>0] = 0; //@line 155 "./common/StringMap.h"
 $1 = ((($this)) + 20|0); //@line 155 "./common/StringMap.h"
 HEAP8[$1>>0] = 0; //@line 155 "./common/StringMap.h"
 $2 = ((($this)) + 32|0); //@line 155 "./common/StringMap.h"
 HEAP8[$2>>0] = 0; //@line 155 "./common/StringMap.h"
 $3 = ((($this)) + 44|0); //@line 155 "./common/StringMap.h"
 HEAP8[$3>>0] = 0; //@line 155 "./common/StringMap.h"
 $4 = ((($this)) + 56|0); //@line 155 "./common/StringMap.h"
 HEAP8[$4>>0] = 0; //@line 155 "./common/StringMap.h"
 $5 = ((($this)) + 68|0); //@line 155 "./common/StringMap.h"
 HEAP8[$5>>0] = 0; //@line 155 "./common/StringMap.h"
 $6 = ((($this)) + 80|0); //@line 155 "./common/StringMap.h"
 HEAP8[$6>>0] = 0; //@line 155 "./common/StringMap.h"
 $7 = ((($this)) + 92|0); //@line 155 "./common/StringMap.h"
 HEAP8[$7>>0] = 0; //@line 155 "./common/StringMap.h"
 $8 = ((($this)) + 104|0); //@line 155 "./common/StringMap.h"
 HEAP8[$8>>0] = 0; //@line 155 "./common/StringMap.h"
 $9 = ((($this)) + 116|0); //@line 155 "./common/StringMap.h"
 HEAP8[$9>>0] = 0; //@line 155 "./common/StringMap.h"
 $10 = ((($this)) + 128|0); //@line 155 "./common/StringMap.h"
 HEAP8[$10>>0] = 0; //@line 155 "./common/StringMap.h"
 $11 = ((($this)) + 140|0); //@line 155 "./common/StringMap.h"
 HEAP8[$11>>0] = 0; //@line 155 "./common/StringMap.h"
 $12 = ((($this)) + 152|0); //@line 155 "./common/StringMap.h"
 HEAP8[$12>>0] = 0; //@line 155 "./common/StringMap.h"
 $13 = ((($this)) + 164|0); //@line 155 "./common/StringMap.h"
 HEAP8[$13>>0] = 0; //@line 155 "./common/StringMap.h"
 $14 = ((($this)) + 176|0); //@line 155 "./common/StringMap.h"
 HEAP8[$14>>0] = 0; //@line 155 "./common/StringMap.h"
 $15 = ((($this)) + 188|0); //@line 155 "./common/StringMap.h"
 HEAP8[$15>>0] = 0; //@line 155 "./common/StringMap.h"
 $16 = ((($this)) + 200|0); //@line 155 "./common/StringMap.h"
 HEAP8[$16>>0] = 0; //@line 155 "./common/StringMap.h"
 $17 = ((($this)) + 212|0); //@line 155 "./common/StringMap.h"
 HEAP8[$17>>0] = 0; //@line 155 "./common/StringMap.h"
 $scevgep = ((($this)) + 216|0); //@line 44 "./common/StringMap.h"
 dest=$scevgep; stop=dest+36|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 44 "./common/StringMap.h"
 $18 = $num >>> 3; //@line 46 "./common/StringMap.h"
 $19 = ($18|0)==(0); //@line 48 "./common/StringMap.h"
 if ($19) {
  STACKTOP = sp;return; //@line 50 "./common/StringMap.h"
 } else {
  $i1$05 = 0;
 }
 while(1) {
  $20 = (($entries) + ($i1$05<<3)|0); //@line 49 "./common/StringMap.h"
  $21 = HEAP32[$20>>2]|0; //@line 49 "./common/StringMap.h"
  $22 = (((($entries) + ($i1$05<<3)|0)) + 4|0); //@line 49 "./common/StringMap.h"
  $23 = HEAP32[$22>>2]|0; //@line 49 "./common/StringMap.h"
  $24 = HEAP8[$21>>0]|0; //@line 142 "./common/StringMap.h"
  $25 = ($24<<24>>24)==(0); //@line 142 "./common/StringMap.h"
  if ($25) {
   $hash$0$lcssa$i$i = 5381;
  } else {
   $$02$i$i = $21;$27 = $24;$hash$01$i$i = 5381;
   while(1) {
    $26 = $27 << 24 >> 24; //@line 142 "./common/StringMap.h"
    $28 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
    $29 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
    $30 = (($29) + ($26))|0; //@line 143 "./common/StringMap.h"
    $31 = HEAP8[$28>>0]|0; //@line 142 "./common/StringMap.h"
    $32 = ($31<<24>>24)==(0); //@line 142 "./common/StringMap.h"
    if ($32) {
     $hash$0$lcssa$i$i = $30;
     break;
    } else {
     $$02$i$i = $28;$27 = $31;$hash$01$i$i = $30;
    }
   }
  }
  $i$08$i = 0;
  while(1) {
   $35 = (($i$08$i) + ($hash$0$lcssa$i$i))|0; //@line 112 "./common/StringMap.h"
   $36 = (($35>>>0) % 18)&-1; //@line 112 "./common/StringMap.h"
   $37 = (((($this) + (($36*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $38 = HEAP8[$37>>0]|0; //@line 114 "./common/StringMap.h"
   $39 = ($38<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   $34 = (($i$08$i) + 1)|0; //@line 110 "./common/StringMap.h"
   if ($39) {
    $$lcssa26 = $36;$$lcssa28 = $37;
    label = 8;
    break;
   }
   $33 = ($34>>>0)<(18); //@line 110 "./common/StringMap.h"
   if ($33) {
    $i$08$i = $34;
   } else {
    break;
   }
  }
  if ((label|0) == 8) {
   label = 0;
   HEAP8[$$lcssa28>>0] = 1; //@line 117 "./common/StringMap.h"
   $40 = (($this) + (($$lcssa26*12)|0)|0); //@line 118 "./common/StringMap.h"
   HEAP32[$40>>2] = $21; //@line 118 "./common/StringMap.h"
   $41 = (((($this) + (($$lcssa26*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
   HEAP32[$41>>2] = $23; //@line 119 "./common/StringMap.h"
  }
  $42 = ($23>>>0)>(8); //@line 126 "./common/StringMap.h"
  if ($42) {
   HEAP32[$vararg_buffer>>2] = $21; //@line 128 "./common/StringMap.h"
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
   HEAP32[$vararg_ptr1>>2] = $23; //@line 128 "./common/StringMap.h"
   (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  } else {
   $43 = (((($this)) + 216|0) + ($23<<2)|0); //@line 132 "./common/StringMap.h"
   HEAP32[$43>>2] = $21; //@line 132 "./common/StringMap.h"
  }
  $44 = (($i1$05) + 1)|0; //@line 48 "./common/StringMap.h"
  $exitcond = ($44|0)==($18|0); //@line 48 "./common/StringMap.h"
  if ($exitcond) {
   break;
  } else {
   $i1$05 = $44;
  }
 }
 STACKTOP = sp;return; //@line 50 "./common/StringMap.h"
}
function __ZN4love8graphics6opengl6ShaderC2ERKNS2_12ShaderSourceE($this,$source) {
 $this = $this|0;
 $source = $source|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$3 = 0, $$34 = 0, $$4 = 0, $$45 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__new_last2$i$i$i = 0;
 var $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i7 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i8 = 0, $vararg_buffer = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 __ZN4love6ObjectC2Ev($this); //@line 74 "modules/graphics/opengl/Shader.cpp"
 $1 = ((($this)) + 8|0); //@line 83 "modules/graphics/opengl/Shader.cpp"
 __THREW__ = 0;
 invoke_vi(4011,($1|0)); //@line 74 "modules/graphics/opengl/Shader.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $68 = ___cxa_find_matching_catch_2()|0;
  $69 = tempRet0;
  $$4 = $69;$$45 = $68;
  ___resumeException($$45|0);
  // unreachable;
 }
 HEAP32[$this>>2] = (25212); //@line 83 "modules/graphics/opengl/Shader.cpp"
 HEAP32[$1>>2] = (25244); //@line 83 "modules/graphics/opengl/Shader.cpp"
 $4 = ((($this)) + 12|0); //@line 84 "modules/graphics/opengl/Shader.h"
 __THREW__ = 0;
 invoke_vii(3351,($4|0),($source|0)); //@line 84 "modules/graphics/opengl/Shader.h"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 do {
  if ($6) {
   $70 = ___cxa_find_matching_catch_2()|0;
   $71 = tempRet0;
   $eh$lpad$body$index2Z2D = $71;$eh$lpad$body$indexZ2D = $70;
   label = 13;
  } else {
   $7 = ((($this)) + 24|0); //@line 84 "modules/graphics/opengl/Shader.h"
   $8 = ((($source)) + 12|0); //@line 84 "modules/graphics/opengl/Shader.h"
   __THREW__ = 0;
   invoke_vii(3351,($7|0),($8|0)); //@line 84 "modules/graphics/opengl/Shader.h"
   $9 = __THREW__; __THREW__ = 0;
   $10 = $9&1;
   if ($10) {
    $11 = ___cxa_find_matching_catch_2()|0;
    $12 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 84 "modules/graphics/opengl/Shader.h"
    $eh$lpad$body$index2Z2D = $12;$eh$lpad$body$indexZ2D = $11;
    label = 13;
    break;
   }
   $13 = ((($this)) + 36|0); //@line 74 "modules/graphics/opengl/Shader.cpp"
   $14 = ((($this)) + 40|0); //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$14>>2] = 0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $15 = ((($this)) + 44|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$15>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $16 = ((($this)) + 40|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$13>>2] = $16; //@line 1130 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $17 = ((($this)) + 48|0); //@line 76 "modules/graphics/opengl/Shader.cpp"
   $18 = ((($this)) + 104|0); //@line 74 "modules/graphics/opengl/Shader.cpp"
   $19 = ((($this)) + 108|0); //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$19>>2] = 0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $20 = ((($this)) + 112|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$20>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $21 = ((($this)) + 108|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   dest=$17; stop=dest+56|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 77 "modules/graphics/opengl/Shader.cpp"
   HEAP32[$18>>2] = $21; //@line 1130 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $22 = ((($this)) + 116|0); //@line 74 "modules/graphics/opengl/Shader.cpp"
   $23 = ((($this)) + 120|0); //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$23>>2] = 0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $24 = ((($this)) + 124|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$24>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $25 = ((($this)) + 120|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$22>>2] = $25; //@line 1130 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $26 = ((($this)) + 128|0); //@line 74 "modules/graphics/opengl/Shader.cpp"
   $27 = ((($this)) + 132|0); //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$27>>2] = 0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $28 = ((($this)) + 136|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$28>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $29 = ((($this)) + 132|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$26>>2] = $29; //@line 1130 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $30 = ((($this)) + 140|0); //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$30>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $31 = ((($this)) + 144|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$31>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $32 = ((($this)) + 148|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$32>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $33 = ((($this)) + 152|0); //@line 74 "modules/graphics/opengl/Shader.cpp"
   $34 = ((($this)) + 156|0); //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   HEAP32[$34>>2] = 0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $35 = ((($this)) + 160|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$35>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $36 = ((($this)) + 156|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$33>>2] = $36; //@line 1130 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $37 = ((($this)) + 164|0); //@line 79 "modules/graphics/opengl/Shader.cpp"
   HEAP32[$37>>2] = (-1); //@line 79 "modules/graphics/opengl/Shader.cpp"
   $38 = ((($this)) + 168|0); //@line 80 "modules/graphics/opengl/Shader.cpp"
   $39 = ((($this)) + 188|0); //@line 74 "modules/graphics/opengl/Shader.cpp"
   ;HEAP32[$38>>2]=0|0;HEAP32[$38+4>>2]=0|0;HEAP32[$38+8>>2]=0|0;HEAP32[$38+12>>2]=0|0;HEAP32[$38+16>>2]=0|0; //@line 81 "modules/graphics/opengl/Shader.cpp"
   __THREW__ = 0;
   invoke_vi(4100,($39|0)); //@line 74 "modules/graphics/opengl/Shader.cpp"
   $40 = __THREW__; __THREW__ = 0;
   $41 = $40&1;
   if ($41) {
    $72 = ___cxa_find_matching_catch_2()|0;
    $73 = tempRet0;
    $$2 = $73;$$23 = $72;
   } else {
    $42 = ((($this)) + 252|0); //@line 74 "modules/graphics/opengl/Shader.cpp"
    __THREW__ = 0;
    invoke_vi(4100,($42|0)); //@line 74 "modules/graphics/opengl/Shader.cpp"
    $43 = __THREW__; __THREW__ = 0;
    $44 = $43&1;
    if ($44) {
     $74 = ___cxa_find_matching_catch_2()|0;
     $75 = tempRet0;
     $$1 = $75;$$12 = $74;
    } else {
     $45 = ((($this)) + 316|0); //@line 82 "modules/graphics/opengl/Shader.cpp"
     ;HEAP32[$45>>2]=0|0;HEAP32[$45+4>>2]=0|0;HEAP32[$45+8>>2]=0|0;
     $46 = HEAP8[$source>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $47 = $46 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $48 = ($47<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $49 = ((($source)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $50 = HEAP32[$49>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $51 = $46&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $52 = $51 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $53 = $48 ? $52 : $50; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $54 = ($53|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     do {
      if ($54) {
       $55 = HEAP8[$8>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $56 = $55 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $57 = ($56<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $58 = ((($source)) + 16|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $59 = HEAP32[$58>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $60 = $55&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $61 = $60 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $62 = $57 ? $61 : $59; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $63 = ($62|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($63) {
        $64 = (___cxa_allocate_exception(16)|0); //@line 85 "modules/graphics/opengl/Shader.cpp"
        __THREW__ = 0;
        invoke_viii(3349,($64|0),(438455|0),($vararg_buffer|0)); //@line 85 "modules/graphics/opengl/Shader.cpp"
        $65 = __THREW__; __THREW__ = 0;
        $66 = $65&1;
        if ($66) {
         $76 = ___cxa_find_matching_catch_2()|0;
         $77 = tempRet0;
         ___cxa_free_exception(($64|0)); //@line 85 "modules/graphics/opengl/Shader.cpp"
         $$0 = $77;$$01 = $76;
         break;
        } else {
         __THREW__ = 0;
         invoke_viii(3350,($64|0),(96|0),(344|0)); //@line 85 "modules/graphics/opengl/Shader.cpp"
         $67 = __THREW__; __THREW__ = 0;
         label = 17;
         break;
        }
       } else {
        label = 18;
       }
      } else {
       label = 18;
      }
     } while(0);
     L23: do {
      if ((label|0) == 18) {
       $80 = HEAP32[(634436)>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $81 = HEAP32[158608]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       __THREW__ = 0;
       $82 = (invoke_ii(4127,(634044|0))|0); //@line 88 "modules/graphics/opengl/Shader.cpp"
       $83 = __THREW__; __THREW__ = 0;
       $84 = $83&1;
       if ($84) {
        label = 17;
       } else {
        $85 = (($80) - ($81))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $86 = $85 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $87 = (($82) + -1)|0; //@line 88 "modules/graphics/opengl/Shader.cpp"
        $88 = ($86|0)<($87|0); //@line 88 "modules/graphics/opengl/Shader.cpp"
        do {
         if ($88) {
          __THREW__ = 0;
          $89 = (invoke_ii(4127,(634044|0))|0); //@line 89 "modules/graphics/opengl/Shader.cpp"
          $90 = __THREW__; __THREW__ = 0;
          $91 = $90&1;
          if ($91) {
           label = 17;
           break L23;
          }
          $92 = (($89) + -1)|0; //@line 89 "modules/graphics/opengl/Shader.cpp"
          HEAP32[$0>>2] = 0; //@line 89 "modules/graphics/opengl/Shader.cpp"
          $93 = HEAP32[(634436)>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $94 = HEAP32[158608]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $95 = (($93) - ($94))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $96 = $95 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $97 = ($92>>>0)>($96>>>0); //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $98 = $94; //@line 2012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $99 = $93; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          if ($97) {
           $100 = (($92) - ($96))|0; //@line 2010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           __THREW__ = 0;
           invoke_viii(4128,(634432|0),($100|0),($0|0)); //@line 2010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $101 = __THREW__; __THREW__ = 0;
           $102 = $101&1;
           if ($102) {
            label = 17;
            break L23;
           } else {
            break;
           }
          }
          $103 = ($92>>>0)<($96>>>0); //@line 2011 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          if ($103) {
           $104 = (($98) + ($92<<2)|0); //@line 2012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $105 = ($99|0)==($104|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           if (!($105)) {
            $__new_last2$i$i$i = $104; //@line 818 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $scevgep$i$i$i7 = ((($99)) + -4|0);
            $106 = $scevgep$i$i$i7;
            $107 = (($106) - ($__new_last2$i$i$i))|0;
            $108 = $107 >>> 2;
            $109 = $108 ^ -1;
            $scevgep4$i$i$i8 = (($99) + ($109<<2)|0);
            HEAP32[(634436)>>2] = $scevgep4$i$i$i8; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           }
          }
         }
        } while(0);
        $110 = HEAP32[$this>>2]|0; //@line 92 "modules/graphics/opengl/Shader.cpp"
        $111 = ((($110)) + 16|0); //@line 92 "modules/graphics/opengl/Shader.cpp"
        $112 = HEAP32[$111>>2]|0; //@line 92 "modules/graphics/opengl/Shader.cpp"
        __THREW__ = 0;
        (invoke_ii($112|0,($this|0))|0); //@line 92 "modules/graphics/opengl/Shader.cpp"
        $113 = __THREW__; __THREW__ = 0;
        $114 = $113&1;
        if ($114) {
         label = 17;
        } else {
         STACKTOP = sp;return; //@line 93 "modules/graphics/opengl/Shader.cpp"
        }
       }
      }
     } while(0);
     if ((label|0) == 17) {
      $78 = ___cxa_find_matching_catch_2()|0;
      $79 = tempRet0;
      $$0 = $79;$$01 = $78;
     }
     __ZN4love7Matrix4D2Ev($42); //@line 93 "modules/graphics/opengl/Shader.cpp"
     $$1 = $$0;$$12 = $$01;
    }
    __ZN4love7Matrix4D2Ev($39); //@line 93 "modules/graphics/opengl/Shader.cpp"
    $$2 = $$1;$$23 = $$12;
   }
   $115 = HEAP32[$36>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ObjectEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE7destroyEPNS_11__tree_nodeISB_PvEE($33,$115); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $116 = HEAP32[$30>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $117 = ($116|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $118 = $116; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($117)) {
    $119 = HEAP32[$31>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $120 = ($119|0)==($116|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($120)) {
     $scevgep$i$i$i = ((($119)) + -4|0);
     $121 = $scevgep$i$i$i;
     $122 = (($121) - ($118))|0;
     $123 = $122 >>> 2;
     $124 = $123 ^ -1;
     $scevgep4$i$i$i = (($119) + ($124<<2)|0);
     HEAP32[$31>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($116); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $125 = HEAP32[$29>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($26,$125); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $126 = HEAP32[$25>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformEEENS_19__map_value_compareIS7_SD_NS_4lessIS7_EELb1EEENS5_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($22,$126); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $127 = HEAP32[$21>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($18,$127); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $128 = HEAP32[$16>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__16__treeINS_12__value_typeIN4love8graphics6opengl6Shader11ShaderStageENS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEEENS_19__map_value_compareIS6_SD_NS_4lessIS6_EELb1EEENSA_ISD_EEE7destroyEPNS_11__tree_nodeISD_PvEE($13,$128); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7); //@line 84 "modules/graphics/opengl/Shader.h"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 84 "modules/graphics/opengl/Shader.h"
   $$3 = $$2;$$34 = $$23;
  }
 } while(0);
 if ((label|0) == 13) {
  $$3 = $eh$lpad$body$index2Z2D;$$34 = $eh$lpad$body$indexZ2D;
 }
 __ZN4love8graphics8VolatileD2Ev($1); //@line 93 "modules/graphics/opengl/Shader.cpp"
 $$4 = $$3;$$45 = $$34;
 ___resumeException($$45|0);
 // unreachable;
}
function __ZNSt3__16vectorIiNS_9allocatorIiEEE8__appendEjRKi($this,$__n,$__x) {
 $this = $this|0;
 $__n = $__n|0;
 $__x = $__x|0;
 var $$0$i = 0, $$0$i2 = 0, $$0$i45 = 0, $$0$i46 = 0, $$in = 0, $$pre = 0, $$pre7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $scevgep$i = 0, $scevgep$i3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 1063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($this)) + 4|0); //@line 1063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($1) - ($3))|0; //@line 1063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = $4 >> 2; //@line 1063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)<($__n>>>0); //@line 1063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = $3; //@line 1010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($6)) {
  $$0$i = $__n;$9 = $7;
  while(1) {
   $8 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$9>>2] = $8; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $10 = ((($9)) + 4|0); //@line 1011 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $11 = (($$0$i) + -1)|0; //@line 1012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $12 = ($11|0)==(0); //@line 1014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($12) {
    break;
   } else {
    $$0$i = $11;$9 = $10;
   }
  }
  $scevgep$i = (($7) + ($__n<<2)|0); //@line 1007 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$2>>2] = $scevgep$i; //@line 1010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  return; //@line 1072 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $13 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = (($3) - ($13))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = $14 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = (($15) + ($__n))|0; //@line 1068 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ($16>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($17) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$0>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre7 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre7;$19 = $$pre;
 } else {
  $$in = $13;$19 = $1;
 }
 $18 = (($19) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = $18 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = ($20>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($21) {
  $25 = $18 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $26 = ($25>>>0)<($16>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $27 = $26 ? $16 : $25; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $28 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $29 = (($28) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $30 = $29 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $31 = ($27|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($31) {
   $$0$i46 = 0;$35 = 0;$36 = $30;$46 = $28;
  } else {
   $$0$i45 = $27;$54 = $28;$55 = $30;
   label = 9;
  }
 } else {
  $22 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = (($22) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $24 = $23 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i45 = 1073741823;$54 = $22;$55 = $24;
  label = 9;
 }
 if ((label|0) == 9) {
  $32 = $$0$i45 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $33 = (__Znwj($32)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i46 = $$0$i45;$35 = $33;$36 = $55;$46 = $54;
 }
 $34 = (($35) + ($36<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $37 = (($35) + ($$0$i46<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $$0$i2 = $__n;$39 = $34;
 while(1) {
  $38 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$39>>2] = $38; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $40 = ((($39)) + 4|0); //@line 237 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $41 = (($$0$i2) + -1)|0; //@line 238 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $42 = ($41|0)==(0); //@line 239 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($42) {
   break;
  } else {
   $$0$i2 = $41;$39 = $40;
  }
 }
 $43 = $37; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $scevgep$i3 = (($34) + ($__n<<2)|0); //@line 234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $44 = $scevgep$i3; //@line 237 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $45 = (($46) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = $45 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $48 = (0 - ($47))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $49 = (($34) + ($48<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $50 = $49; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $51 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($49|0),($51|0),($45|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $50; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$2>>2] = $44; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $43; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $52 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($52) {
  return; //@line 1072 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $53 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($53); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1072 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl6Shader6detachEv() {
 var $$pre$i = 0, $$pre2$i = 0, $$pre4$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
 $1 = ($0|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
 $2 = HEAP32[158581]|0; //@line 432 "modules/graphics/opengl/Shader.cpp"
 if ($1) {
  $22 = ($2|0)==(0|0); //@line 438 "modules/graphics/opengl/Shader.cpp"
  if (!($22)) {
   $23 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
   FUNCTION_TABLE_vi[$23 & 8191](0); //@line 534 "./libraries/glad/gladfuncs.hpp"
  }
  HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
  return; //@line 442 "modules/graphics/opengl/Shader.cpp"
 }
 $3 = ($2|0)==($0|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
 if ($3) {
  return; //@line 442 "modules/graphics/opengl/Shader.cpp"
 }
 $4 = ((($0)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
 $6 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vi[$6 & 8191]($5); //@line 534 "./libraries/glad/gladfuncs.hpp"
 HEAP32[158581] = $0; //@line 409 "modules/graphics/opengl/Shader.cpp"
 $7 = ((($0)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
 $8 = ((($0)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = HEAP32[$8>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = HEAP32[$7>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($9|0)==($10|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
 if (!($11)) {
  $13 = $10;$24 = $9;$i$01$i = 0;
  while(1) {
   $12 = $13; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $14 = (($12) + ($i$01$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $15 = HEAP32[$14>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
   $16 = ($15|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
   $$pre4$i = (($i$01$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
   if ($16) {
    $18 = $24;$19 = $13;
   } else {
    __ZN4love8graphics6opengl6OpenGL17bindTextureToUnitEjib(634044,$15,$$pre4$i,0); //@line 420 "modules/graphics/opengl/Shader.cpp"
    $$pre$i = HEAP32[$8>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre2$i = HEAP32[$7>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $18 = $$pre$i;$19 = $$pre2$i;
   }
   $17 = (($18) - ($19))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $20 = $17 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $21 = ($$pre4$i>>>0)<($20>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
   if ($21) {
    $13 = $19;$24 = $18;$i$01$i = $$pre4$i;
   } else {
    break;
   }
  }
 }
 __ZN4love8graphics6opengl6OpenGL14setTextureUnitEi(634044,0); //@line 424 "modules/graphics/opengl/Shader.cpp"
 return; //@line 442 "modules/graphics/opengl/Shader.cpp"
}
function __ZN4love8graphics6opengl6Shader6attachEb($this,$temporary) {
 $this = $this|0;
 $temporary = $temporary|0;
 var $$pre = 0, $$pre2 = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[158581]|0; //@line 406 "modules/graphics/opengl/Shader.cpp"
 $1 = ($0|0)==($this|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
 if (!($1)) {
  $2 = ((($this)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
  $4 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$4 & 8191]($3); //@line 534 "./libraries/glad/gladfuncs.hpp"
  HEAP32[158581] = $this; //@line 409 "modules/graphics/opengl/Shader.cpp"
 }
 if ($temporary) {
  return; //@line 426 "modules/graphics/opengl/Shader.cpp"
 }
 $5 = ((($this)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
 $6 = ((($this)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = HEAP32[$6>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = HEAP32[$5>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ($7|0)==($8|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
 if (!($9)) {
  $11 = $8;$20 = $7;$i$01 = 0;
  while(1) {
   $10 = $11; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $12 = (($10) + ($i$01<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $13 = HEAP32[$12>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
   $14 = ($13|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
   $$pre4 = (($i$01) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
   if ($14) {
    $16 = $20;$17 = $11;
   } else {
    __ZN4love8graphics6opengl6OpenGL17bindTextureToUnitEjib(634044,$13,$$pre4,0); //@line 420 "modules/graphics/opengl/Shader.cpp"
    $$pre = HEAP32[$6>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre2 = HEAP32[$5>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $16 = $$pre;$17 = $$pre2;
   }
   $15 = (($16) - ($17))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $18 = $15 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $19 = ($$pre4>>>0)<($18>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
   if ($19) {
    $11 = $17;$20 = $16;$i$01 = $$pre4;
   } else {
    break;
   }
  }
 }
 __ZN4love8graphics6opengl6OpenGL14setTextureUnitEi(634044,0); //@line 424 "modules/graphics/opengl/Shader.cpp"
 return; //@line 426 "modules/graphics/opengl/Shader.cpp"
}
function __ZNK4love8graphics6opengl6Shader11getWarningsEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$0 = 0, $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$01 = 0, $$01$i$i$i = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$3 = 0, $$34 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $11 = 0;
 var $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$09 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 48|0;
 $1 = sp + 36|0;
 $2 = sp + 24|0;
 $3 = sp + 12|0;
 $4 = sp;
 ;HEAP32[$agg$result>>2]=0|0;HEAP32[$agg$result+4>>2]=0|0;HEAP32[$agg$result+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($this)) + 36|0); //@line 856 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $6 = HEAP32[$5>>2]|0; //@line 913 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $7 = ((($this)) + 40|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = ($6|0)==($7|0); //@line 766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 L1: do {
  if ($8) {
   label = 3;
  } else {
   $9 = ((($3)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $10 = ((($3)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $11 = ((($3)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $12 = ((($0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $13 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $14 = ((($0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $__begin$sroa$0$09 = $6;
   while(1) {
    $21 = ((($__begin$sroa$0$09)) + 16|0); //@line 395 "modules/graphics/opengl/Shader.cpp"
    $22 = HEAP32[$21>>2]|0; //@line 395 "modules/graphics/opengl/Shader.cpp"
    $23 = ($22>>>0)>(1); //@line 91 "./common/StringMap.h"
    if (!($23)) {
     $24 = (634492 + ($22<<2)|0); //@line 94 "./common/StringMap.h"
     $25 = HEAP32[$24>>2]|0; //@line 94 "./common/StringMap.h"
     $26 = ($25|0)==(0|0); //@line 94 "./common/StringMap.h"
     if (!($26)) {
      $27 = (_strlen($25)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __THREW__ = 0;
      invoke_viii(3460,($2|0),($25|0),($27|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $28 = __THREW__; __THREW__ = 0;
      $29 = $28&1;
      if ($29) {
       label = 4;
       break;
      }
      __THREW__ = 0;
      invoke_viii(3460,($3|0),(438493|0),9); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $30 = __THREW__; __THREW__ = 0;
      $31 = $30&1;
      if ($31) {
       label = 15;
       break;
      }
      $32 = HEAP8[$3>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $33 = $32 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $34 = ($33<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $35 = HEAP32[$9>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $36 = $34 ? $10 : $35; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $37 = HEAP32[$11>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $38 = $32&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $39 = $38 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $40 = $34 ? $39 : $37; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __THREW__ = 0;
      $41 = (invoke_iiii(3462,($2|0),($36|0),($40|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $42 = __THREW__; __THREW__ = 0;
      $43 = $42&1;
      if ($43) {
       label = 16;
       break;
      }
      ;HEAP32[$1>>2]=HEAP32[$41>>2]|0;HEAP32[$1+4>>2]=HEAP32[$41+4>>2]|0;HEAP32[$1+8>>2]=HEAP32[$41+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      ;HEAP32[$41>>2]=0|0;HEAP32[$41+4>>2]=0|0;HEAP32[$41+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $44 = ((($__begin$sroa$0$09)) + 20|0); //@line 396 "modules/graphics/opengl/Shader.cpp"
      $45 = HEAP8[$44>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $46 = $45 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $47 = ($46<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $48 = ((($__begin$sroa$0$09)) + 28|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $49 = HEAP32[$48>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $50 = ((($44)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $51 = $47 ? $50 : $49; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $52 = ((($__begin$sroa$0$09)) + 24|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $53 = HEAP32[$52>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $54 = $45&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $55 = $54 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $56 = $47 ? $55 : $53; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __THREW__ = 0;
      $57 = (invoke_iiii(3462,($1|0),($51|0),($56|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $58 = __THREW__; __THREW__ = 0;
      $59 = $58&1;
      if ($59) {
       label = 17;
       break;
      }
      ;HEAP32[$0>>2]=HEAP32[$57>>2]|0;HEAP32[$0+4>>2]=HEAP32[$57+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$57+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      ;HEAP32[$57>>2]=0|0;HEAP32[$57+4>>2]=0|0;HEAP32[$57+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $60 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $61 = $60 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $62 = ($61<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $63 = HEAP32[$12>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $64 = $62 ? $13 : $63; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $65 = HEAP32[$14>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $66 = $60&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $67 = $66 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $68 = $62 ? $67 : $65; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __THREW__ = 0;
      (invoke_iiii(3462,($agg$result|0),($64|0),($68|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $69 = __THREW__; __THREW__ = 0;
      $70 = $69&1;
      if ($70) {
       label = 18;
       break;
      }
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 396 "modules/graphics/opengl/Shader.cpp"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 396 "modules/graphics/opengl/Shader.cpp"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 396 "modules/graphics/opengl/Shader.cpp"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 396 "modules/graphics/opengl/Shader.cpp"
     }
    }
    $79 = ((($__begin$sroa$0$09)) + 4|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $80 = HEAP32[$79>>2]|0; //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $81 = ($80|0)==(0|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($81) {
     $$0$i$i$i = $__begin$sroa$0$09;
     while(1) {
      $84 = ((($$0$i$i$i)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $85 = HEAP32[$84>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $86 = HEAP32[$85>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $87 = ($86|0)==($$0$i$i$i|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      if ($87) {
       $$01$i$i$i = $85;
       break;
      } else {
       $$0$i$i$i = $85;
      }
     }
    } else {
     $$0$i$i$i$i = $80;
     while(1) {
      $82 = HEAP32[$$0$i$i$i$i>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      $83 = ($82|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
      if ($83) {
       $$01$i$i$i = $$0$i$i$i$i;
       break;
      } else {
       $$0$i$i$i$i = $82;
      }
     }
    }
    $88 = ($$01$i$i$i|0)==($7|0); //@line 766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($88) {
     label = 3;
     break L1;
    } else {
     $__begin$sroa$0$09 = $$01$i$i$i;
    }
   }
   if ((label|0) == 4) {
    $17 = ___cxa_find_matching_catch_2()|0;
    $18 = tempRet0;
    $lpad$phi$index = $17;$lpad$phi$index2 = $18;
    break;
   }
   else if ((label|0) == 15) {
    $71 = ___cxa_find_matching_catch_2()|0;
    $72 = tempRet0;
    $$2 = $72;$$23 = $71;
   }
   else if ((label|0) == 16) {
    $73 = ___cxa_find_matching_catch_2()|0;
    $74 = tempRet0;
    $$1 = $74;$$12 = $73;
    label = 20;
   }
   else if ((label|0) == 17) {
    $75 = ___cxa_find_matching_catch_2()|0;
    $76 = tempRet0;
    $$0 = $76;$$01 = $75;
    label = 19;
   }
   else if ((label|0) == 18) {
    $77 = ___cxa_find_matching_catch_2()|0;
    $78 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 396 "modules/graphics/opengl/Shader.cpp"
    $$0 = $78;$$01 = $77;
    label = 19;
   }
   if ((label|0) == 19) {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 396 "modules/graphics/opengl/Shader.cpp"
    $$1 = $$0;$$12 = $$01;
    label = 20;
   }
   if ((label|0) == 20) {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 396 "modules/graphics/opengl/Shader.cpp"
    $$2 = $$1;$$23 = $$12;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 396 "modules/graphics/opengl/Shader.cpp"
   $$3 = $$2;$$34 = $$23;
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($agg$result); //@line 402 "modules/graphics/opengl/Shader.cpp"
   ___resumeException($$34|0);
   // unreachable;
  }
 } while(0);
 do {
  if ((label|0) == 3) {
   __THREW__ = 0;
   invoke_vii(4114,($4|0),($this|0)); //@line 399 "modules/graphics/opengl/Shader.cpp"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if ($16) {
    $19 = ___cxa_find_matching_catch_2()|0;
    $20 = tempRet0;
    $lpad$phi$index = $19;$lpad$phi$index2 = $20;
    break;
   }
   $89 = HEAP8[$4>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $90 = $89 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $91 = ($90<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $92 = ((($4)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $93 = HEAP32[$92>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $94 = ((($4)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $95 = $91 ? $94 : $93; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $96 = ((($4)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $97 = HEAP32[$96>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $98 = $89&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $99 = $98 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $100 = $91 ? $99 : $97; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   (invoke_iiii(3462,($agg$result|0),($95|0),($100|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $101 = __THREW__; __THREW__ = 0;
   $102 = $101&1;
   if (!($102)) {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 399 "modules/graphics/opengl/Shader.cpp"
    STACKTOP = sp;return; //@line 402 "modules/graphics/opengl/Shader.cpp"
   }
   $103 = ___cxa_find_matching_catch_2()|0;
   $104 = tempRet0;
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 399 "modules/graphics/opengl/Shader.cpp"
   $$3 = $104;$$34 = $103;
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($agg$result); //@line 402 "modules/graphics/opengl/Shader.cpp"
   ___resumeException($$34|0);
   // unreachable;
  }
 } while(0);
 $$3 = $lpad$phi$index2;$$34 = $lpad$phi$index;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($agg$result); //@line 402 "modules/graphics/opengl/Shader.cpp"
 ___resumeException($$34|0);
 // unreachable;
}
function __ZNKSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformEEENS_19__map_value_compareIS7_SD_NS_4lessIS7_EELb1EEENS5_ISD_EEE4findIS7_EENS_21__tree_const_iteratorISD_PNS_11__tree_nodeISD_PvEEiEERKT_($this,$__v) {
 $this = $this|0;
 $__v = $__v|0;
 var $$$$i$i$i$i = 0, $$$$i$i$i$i$i = 0, $$0$ph$lcssa$i = 0, $$0$ph6$i = 0, $$01$ph5$i = 0, $$013$i = 0, $$013$i$lcssa = 0, $$sroa$0$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 878 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $2 = ($1|0)==(0|0); //@line 2120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if (!($2)) {
  $3 = HEAP8[$__v>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $4 = $3 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $5 = ($4<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $6 = ((($__v)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = HEAP32[$6>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $8 = $3&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = $8 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $10 = $5 ? $9 : $7; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $11 = ((($__v)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $12 = HEAP32[$11>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $13 = ((($__v)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $14 = $5 ? $13 : $12; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $$0$ph6$i = $0;$$01$ph5$i = $1;
  L3: while(1) {
   $$013$i = $$01$ph5$i;
   while(1) {
    $15 = ((($$013$i)) + 16|0); //@line 2122 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $16 = HEAP8[$15>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $17 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $18 = ($17<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $19 = ((($$013$i)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $20 = HEAP32[$19>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $21 = $16&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $22 = $21 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $23 = $18 ? $22 : $20; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $24 = ((($$013$i)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $25 = HEAP32[$24>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $26 = ((($15)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $27 = $18 ? $26 : $25; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $28 = ($10>>>0)<($23>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $29 = $28 ? $10 : $23; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $30 = (_memcmp($27,$14,$29)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $31 = ($30|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $32 = ($23>>>0)<($10>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $$$$i$i$i$i$i = $32 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $33 = $31 ? $$$$i$i$i$i$i : $30; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $34 = ($33|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if (!($34)) {
     $$013$i$lcssa = $$013$i;
     break;
    }
    $37 = ((($$013$i)) + 4|0); //@line 2128 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $38 = HEAP32[$37>>2]|0; //@line 2128 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $39 = ($38|0)==(0|0); //@line 2120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($39) {
     $$0$ph$lcssa$i = $$0$ph6$i;
     break L3;
    } else {
     $$013$i = $38;
    }
   }
   $35 = HEAP32[$$013$i$lcssa>>2]|0; //@line 2125 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $36 = ($35|0)==(0|0); //@line 2120 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($36) {
    $$0$ph$lcssa$i = $$013$i$lcssa;
    break;
   } else {
    $$0$ph6$i = $$013$i$lcssa;$$01$ph5$i = $35;
   }
  }
  $40 = ($$0$ph$lcssa$i|0)==($0|0); //@line 766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if (!($40)) {
   $41 = ((($$0$ph$lcssa$i)) + 16|0); //@line 741 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $42 = HEAP8[$41>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $43 = $42 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $44 = ($43<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $45 = ((($$0$ph$lcssa$i)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $46 = HEAP32[$45>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $47 = $42&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $48 = $47 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $49 = $44 ? $48 : $46; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $50 = ((($$0$ph$lcssa$i)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $51 = HEAP32[$50>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $52 = ((($41)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $53 = $44 ? $52 : $51; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $54 = ($49>>>0)<($10>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $55 = $54 ? $49 : $10; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $56 = (_memcmp($14,$53,$55)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $57 = ($56|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $58 = ($10>>>0)<($49>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$$$i$i$i$i = $58 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $59 = $57 ? $$$$i$i$i$i : $56; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $60 = ($59|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if (!($60)) {
    $$sroa$0$0 = $$0$ph$lcssa$i;
    return ($$sroa$0$0|0); //@line 2043 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   }
  }
 }
 $$sroa$0$0 = $0;
 return ($$sroa$0$0|0); //@line 2043 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
}
function __ZNK4love8graphics6opengl6Shader20checkSetUniformErrorERKNS2_7UniformEiiNS2_11UniformTypeE($this,$u,$size,$count,$sendtype) {
 $this = $this|0;
 $u = $u|0;
 $size = $size|0;
 $count = $count|0;
 $sendtype = $sendtype|0;
 var $$0 = 0, $$0$i = 0, $$04 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond3 = 0, $or$cond5 = 0, $or$cond7 = 0, $or$cond9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer10 = 0, $vararg_buffer4 = 0, $vararg_buffer8 = 0, $vararg_ptr3 = 0, $vararg_ptr7 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer10 = sp + 32|0;
 $vararg_buffer8 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($this)) + 48|0); //@line 457 "modules/graphics/opengl/Shader.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 457 "modules/graphics/opengl/Shader.cpp"
 $2 = ($1|0)==(0); //@line 457 "modules/graphics/opengl/Shader.cpp"
 if ($2) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 458 "modules/graphics/opengl/Shader.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($3|0),(438587|0),($vararg_buffer|0)); //@line 458 "modules/graphics/opengl/Shader.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if (!($5)) {
   ___cxa_throw(($3|0),(96|0),(344|0)); //@line 458 "modules/graphics/opengl/Shader.cpp"
   // unreachable; //@line 458 "modules/graphics/opengl/Shader.cpp"
  }
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  ___cxa_free_exception(($3|0)); //@line 458 "modules/graphics/opengl/Shader.cpp"
  $$0 = $6;$$04 = $7;
  ___resumeException($$0|0);
  // unreachable;
 }
 $8 = ((($u)) + 8|0); //@line 460 "modules/graphics/opengl/Shader.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 460 "modules/graphics/opengl/Shader.cpp"
 switch ($9|0) {
 case 35669: case 35666: case 35676: case 35673:  {
  $$0$i = 4;
  break;
 }
 case 35665: case 35668: case 35675: case 35672:  {
  $$0$i = 3;
  break;
 }
 case 35671: case 35674: case 35664: case 35667:  {
  $$0$i = 2;
  break;
 }
 default: {
  $$0$i = 1;
 }
 }
 $10 = ($$0$i|0)==($size|0); //@line 462 "modules/graphics/opengl/Shader.cpp"
 if (!($10)) {
  $11 = (___cxa_allocate_exception(16)|0); //@line 463 "modules/graphics/opengl/Shader.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer1>>2] = $size; //@line 463 "modules/graphics/opengl/Shader.cpp"
  $vararg_ptr3 = ((($vararg_buffer1)) + 4|0); //@line 463 "modules/graphics/opengl/Shader.cpp"
  HEAP32[$vararg_ptr3>>2] = $$0$i; //@line 463 "modules/graphics/opengl/Shader.cpp"
  invoke_viii(3349,($11|0),(438613|0),($vararg_buffer1|0)); //@line 463 "modules/graphics/opengl/Shader.cpp"
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   ___cxa_throw(($11|0),(96|0),(344|0)); //@line 463 "modules/graphics/opengl/Shader.cpp"
   // unreachable; //@line 463 "modules/graphics/opengl/Shader.cpp"
  }
  $14 = ___cxa_find_matching_catch_2()|0;
  $15 = tempRet0;
  ___cxa_free_exception(($11|0)); //@line 463 "modules/graphics/opengl/Shader.cpp"
  $$0 = $14;$$04 = $15;
  ___resumeException($$0|0);
  // unreachable;
 }
 $16 = ((($u)) + 4|0); //@line 465 "modules/graphics/opengl/Shader.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 465 "modules/graphics/opengl/Shader.cpp"
 $18 = ($17|0)==(1); //@line 465 "modules/graphics/opengl/Shader.cpp"
 $19 = ($count|0)>(1); //@line 465 "modules/graphics/opengl/Shader.cpp"
 $or$cond = $19 & $18; //@line 465 "modules/graphics/opengl/Shader.cpp"
 $20 = ($count|0)<(0); //@line 465 "modules/graphics/opengl/Shader.cpp"
 $or$cond3 = $20 | $or$cond; //@line 465 "modules/graphics/opengl/Shader.cpp"
 if ($or$cond3) {
  $21 = (___cxa_allocate_exception(16)|0); //@line 466 "modules/graphics/opengl/Shader.cpp"
  $22 = HEAP32[$16>>2]|0; //@line 466 "modules/graphics/opengl/Shader.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer4>>2] = $22; //@line 466 "modules/graphics/opengl/Shader.cpp"
  $vararg_ptr7 = ((($vararg_buffer4)) + 4|0); //@line 466 "modules/graphics/opengl/Shader.cpp"
  HEAP32[$vararg_ptr7>>2] = $count; //@line 466 "modules/graphics/opengl/Shader.cpp"
  invoke_viii(3349,($21|0),(438666|0),($vararg_buffer4|0)); //@line 466 "modules/graphics/opengl/Shader.cpp"
  $23 = __THREW__; __THREW__ = 0;
  $24 = $23&1;
  if (!($24)) {
   ___cxa_throw(($21|0),(96|0),(344|0)); //@line 466 "modules/graphics/opengl/Shader.cpp"
   // unreachable; //@line 466 "modules/graphics/opengl/Shader.cpp"
  }
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  ___cxa_free_exception(($21|0)); //@line 466 "modules/graphics/opengl/Shader.cpp"
  $$0 = $25;$$04 = $26;
  ___resumeException($$0|0);
  // unreachable;
 }
 $27 = ((($u)) + 12|0); //@line 468 "modules/graphics/opengl/Shader.cpp"
 $28 = HEAP32[$27>>2]|0; //@line 468 "modules/graphics/opengl/Shader.cpp"
 $29 = ($28|0)!=(3); //@line 468 "modules/graphics/opengl/Shader.cpp"
 $30 = ($28|0)==($sendtype|0); //@line 468 "modules/graphics/opengl/Shader.cpp"
 $or$cond5 = $29 | $30; //@line 468 "modules/graphics/opengl/Shader.cpp"
 if (!($or$cond5)) {
  $31 = (___cxa_allocate_exception(16)|0); //@line 469 "modules/graphics/opengl/Shader.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($31|0),(438714|0),($vararg_buffer8|0)); //@line 469 "modules/graphics/opengl/Shader.cpp"
  $32 = __THREW__; __THREW__ = 0;
  $33 = $32&1;
  if (!($33)) {
   ___cxa_throw(($31|0),(96|0),(344|0)); //@line 469 "modules/graphics/opengl/Shader.cpp"
   // unreachable; //@line 469 "modules/graphics/opengl/Shader.cpp"
  }
  $34 = ___cxa_find_matching_catch_2()|0;
  $35 = tempRet0;
  ___cxa_free_exception(($31|0)); //@line 469 "modules/graphics/opengl/Shader.cpp"
  $$0 = $34;$$04 = $35;
  ___resumeException($$0|0);
  // unreachable;
 }
 $36 = ($sendtype|0)==(0); //@line 471 "modules/graphics/opengl/Shader.cpp"
 $37 = ($28|0)==(1); //@line 471 "modules/graphics/opengl/Shader.cpp"
 $or$cond7 = $36 & $37; //@line 471 "modules/graphics/opengl/Shader.cpp"
 if (!($or$cond7)) {
  $38 = ($sendtype|0)==(1); //@line 471 "modules/graphics/opengl/Shader.cpp"
  $39 = ($28|0)==(0); //@line 471 "modules/graphics/opengl/Shader.cpp"
  $or$cond9 = $38 & $39; //@line 471 "modules/graphics/opengl/Shader.cpp"
  if (!($or$cond9)) {
   STACKTOP = sp;return; //@line 473 "modules/graphics/opengl/Shader.cpp"
  }
 }
 $40 = (___cxa_allocate_exception(16)|0); //@line 472 "modules/graphics/opengl/Shader.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($40|0),(438769|0),($vararg_buffer10|0)); //@line 472 "modules/graphics/opengl/Shader.cpp"
 $41 = __THREW__; __THREW__ = 0;
 $42 = $41&1;
 if (!($42)) {
  ___cxa_throw(($40|0),(96|0),(344|0)); //@line 472 "modules/graphics/opengl/Shader.cpp"
  // unreachable; //@line 472 "modules/graphics/opengl/Shader.cpp"
 }
 $43 = ___cxa_find_matching_catch_2()|0;
 $44 = tempRet0;
 ___cxa_free_exception(($40|0)); //@line 472 "modules/graphics/opengl/Shader.cpp"
 $$0 = $43;$$04 = $44;
 ___resumeException($$0|0);
 // unreachable;
}
function __ZN4love8graphics6opengl6Shader7sendIntERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEEiPKii($this,$name,$size,$vec,$count) {
 $this = $this|0;
 $name = $name|0;
 $size = $size|0;
 $vec = $vec|0;
 $count = $count|0;
 var $$cast = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i28 = 0, $$pre$i18 = 0, $$pre2$i = 0, $$pre2$i$i = 0, $$pre2$i$i29 = 0, $$pre2$i19 = 0, $$pre4$i = 0, $$pre4$i$i = 0, $$pre4$i$i26 = 0, $$pre4$i16 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0;
 var $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i$01$i = 0, $i$01$i$i = 0, $i$01$i$i25 = 0, $i$01$i15 = 0, $lpad$phi$index = 0, $lpad$phi43$index = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP32[158581]|0; //@line 46 "modules/graphics/opengl/Shader.cpp"
 $$cast = $0; //@line 406 "modules/graphics/opengl/Shader.cpp"
 $1 = ($$cast|0)==($this|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
 if (!($1)) {
  $2 = ((($this)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
  $4 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$4 & 8191]($3); //@line 534 "./libraries/glad/gladfuncs.hpp"
  HEAP32[158581] = $this; //@line 409 "modules/graphics/opengl/Shader.cpp"
 }
 $5 = ((($this)) + 116|0); //@line 1092 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $6 = (__ZNKSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformEEENS_19__map_value_compareIS7_SD_NS_4lessIS7_EELb1EEENS5_ISD_EEE4findIS7_EENS_21__tree_const_iteratorISD_PNS_11__tree_nodeISD_PvEEiEERKT_($5,$name)|0); //@line 1092 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $7 = ((($this)) + 120|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = ($6|0)==($7|0); //@line 763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 L4: do {
  if ($8) {
   $9 = (___cxa_allocate_exception(16)|0); //@line 449 "modules/graphics/opengl/Shader.cpp"
   $10 = HEAP8[$name>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $11 = $10 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $12 = ($11<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $13 = ((($name)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $14 = HEAP32[$13>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $15 = ((($name)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $16 = $12 ? $15 : $14; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   HEAP32[$vararg_buffer>>2] = $16; //@line 449 "modules/graphics/opengl/Shader.cpp"
   invoke_viii(3349,($9|0),(438503|0),($vararg_buffer|0)); //@line 449 "modules/graphics/opengl/Shader.cpp"
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if ($18) {
    $20 = ___cxa_find_matching_catch_2()|0;
    $21 = tempRet0;
    ___cxa_free_exception(($9|0)); //@line 449 "modules/graphics/opengl/Shader.cpp"
    $eh$lpad$body$index2Z2D = $21;$eh$lpad$body$indexZ2D = $20;
    break;
   } else {
    __THREW__ = 0;
    invoke_viii(3350,($9|0),(96|0),(344|0)); //@line 449 "modules/graphics/opengl/Shader.cpp"
    $19 = __THREW__; __THREW__ = 0;
    label = 9;
    break;
   }
  } else {
   $22 = ((($6)) + 28|0); //@line 452 "modules/graphics/opengl/Shader.cpp"
   __THREW__ = 0;
   invoke_viiiii(4129,($this|0),($22|0),($size|0),($count|0),1); //@line 480 "modules/graphics/opengl/Shader.cpp"
   $23 = __THREW__; __THREW__ = 0;
   $24 = $23&1;
   if ($24) {
    label = 9;
   } else {
    switch ($size|0) {
    case 4:  {
     $92 = HEAP32[$22>>2]|0; //@line 485 "modules/graphics/opengl/Shader.cpp"
     $93 = HEAP32[164915]|0; //@line 550 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_viii($93|0,($92|0),($count|0),($vec|0)); //@line 550 "./libraries/glad/gladfuncs.hpp"
     $94 = __THREW__; __THREW__ = 0;
     $95 = $94&1;
     if ($95) {
      label = 9;
      break L4;
     }
     break;
    }
    case 3:  {
     $96 = HEAP32[$22>>2]|0; //@line 488 "modules/graphics/opengl/Shader.cpp"
     $97 = HEAP32[165043]|0; //@line 549 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_viii($97|0,($96|0),($count|0),($vec|0)); //@line 549 "./libraries/glad/gladfuncs.hpp"
     $98 = __THREW__; __THREW__ = 0;
     $99 = $98&1;
     if ($99) {
      label = 9;
      break L4;
     }
     break;
    }
    case 2:  {
     $100 = HEAP32[$22>>2]|0; //@line 491 "modules/graphics/opengl/Shader.cpp"
     $101 = HEAP32[164977]|0; //@line 548 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_viii($101|0,($100|0),($count|0),($vec|0)); //@line 548 "./libraries/glad/gladfuncs.hpp"
     $102 = __THREW__; __THREW__ = 0;
     $103 = $102&1;
     if ($103) {
      label = 9;
      break L4;
     }
     break;
    }
    default: {
     $104 = HEAP32[$22>>2]|0; //@line 495 "modules/graphics/opengl/Shader.cpp"
     $105 = HEAP32[165316]|0; //@line 547 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_viii($105|0,($104|0),($count|0),($vec|0)); //@line 547 "./libraries/glad/gladfuncs.hpp"
     $106 = __THREW__; __THREW__ = 0;
     $107 = $106&1;
     if ($107) {
      label = 9;
      break L4;
     }
    }
    }
    $108 = ($0|0)==(0); //@line 53 "modules/graphics/opengl/Shader.cpp"
    do {
     if ($108) {
      $135 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
      $136 = ($135|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
      $137 = HEAP32[158581]|0; //@line 432 "modules/graphics/opengl/Shader.cpp"
      if ($136) {
       $163 = ($137|0)==(0|0); //@line 438 "modules/graphics/opengl/Shader.cpp"
       if (!($163)) {
        $164 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
        __THREW__ = 0;
        invoke_vi($164|0,0); //@line 534 "./libraries/glad/gladfuncs.hpp"
        $165 = __THREW__; __THREW__ = 0;
        $166 = $165&1;
        if ($166) {
         break;
        }
       }
       HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
       STACKTOP = sp;return; //@line 498 "modules/graphics/opengl/Shader.cpp"
      }
      $138 = ($137|0)==($135|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
      if ($138) {
       STACKTOP = sp;return; //@line 498 "modules/graphics/opengl/Shader.cpp"
      }
      $139 = ((($135)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
      $140 = HEAP32[$139>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
      $141 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_vi($141|0,($140|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
      $142 = __THREW__; __THREW__ = 0;
      $143 = $142&1;
      if (!($143)) {
       HEAP32[158581] = $135; //@line 409 "modules/graphics/opengl/Shader.cpp"
       $144 = ((($135)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
       $145 = ((($135)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $146 = HEAP32[$145>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $147 = HEAP32[$144>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $148 = ($146|0)==($147|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
       L30: do {
        if (!($148)) {
         $150 = $147;$176 = $146;$i$01$i$i25 = 0;
         while(1) {
          $149 = $150; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $151 = (($149) + ($i$01$i$i25<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $152 = HEAP32[$151>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
          $153 = ($152|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
          $$pre4$i$i26 = (($i$01$i$i25) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
          if ($153) {
           $157 = $176;$158 = $150;
          } else {
           __THREW__ = 0;
           invoke_viiii(4111,(634044|0),($152|0),($$pre4$i$i26|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
           $154 = __THREW__; __THREW__ = 0;
           $155 = $154&1;
           if ($155) {
            break;
           }
           $$pre$i$i28 = HEAP32[$145>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $$pre2$i$i29 = HEAP32[$144>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $157 = $$pre$i$i28;$158 = $$pre2$i$i29;
          }
          $156 = (($157) - ($158))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $159 = $156 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $160 = ($$pre4$i$i26>>>0)<($159>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
          if ($160) {
           $150 = $158;$176 = $157;$i$01$i$i25 = $$pre4$i$i26;
          } else {
           break L30;
          }
         }
         $167 = ___cxa_find_matching_catch_3(0|0)|0;
         $168 = tempRet0;
         $lpad$phi43$index = $167;
         ___clang_call_terminate($lpad$phi43$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
         // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
        }
       } while(0);
       __THREW__ = 0;
       invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
       $161 = __THREW__; __THREW__ = 0;
       $162 = $161&1;
       if (!($162)) {
        STACKTOP = sp;return; //@line 498 "modules/graphics/opengl/Shader.cpp"
       }
      }
     } else {
      $109 = HEAP32[158581]|0; //@line 406 "modules/graphics/opengl/Shader.cpp"
      $110 = ($109|0)==($$cast|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
      if (!($110)) {
       $111 = ((($$cast)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
       $112 = HEAP32[$111>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
       $113 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
       __THREW__ = 0;
       invoke_vi($113|0,($112|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
       $114 = __THREW__; __THREW__ = 0;
       $115 = $114&1;
       if ($115) {
        break;
       }
       HEAP32[158581] = $$cast; //@line 409 "modules/graphics/opengl/Shader.cpp"
      }
      $116 = ((($$cast)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      $117 = ((($$cast)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $118 = HEAP32[$117>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $119 = HEAP32[$116>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $120 = ($118|0)==($119|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      L46: do {
       if (!($120)) {
        $124 = $119;$175 = $118;$i$01$i15 = 0;
        while(1) {
         $123 = $124; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $125 = (($123) + ($i$01$i15<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $126 = HEAP32[$125>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
         $127 = ($126|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
         $$pre4$i16 = (($i$01$i15) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
         if ($127) {
          $131 = $175;$132 = $124;
         } else {
          __THREW__ = 0;
          invoke_viiii(4111,(634044|0),($126|0),($$pre4$i16|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
          $128 = __THREW__; __THREW__ = 0;
          $129 = $128&1;
          if ($129) {
           break;
          }
          $$pre$i18 = HEAP32[$117>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $$pre2$i19 = HEAP32[$116>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $131 = $$pre$i18;$132 = $$pre2$i19;
         }
         $130 = (($131) - ($132))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $133 = $130 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $134 = ($$pre4$i16>>>0)<($133>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
         if ($134) {
          $124 = $132;$175 = $131;$i$01$i15 = $$pre4$i16;
         } else {
          break L46;
         }
        }
        $169 = ___cxa_find_matching_catch_3(0|0)|0;
        $170 = tempRet0;
        $lpad$phi43$index = $169;
        ___clang_call_terminate($lpad$phi43$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
        // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
       }
      } while(0);
      __THREW__ = 0;
      invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
      $121 = __THREW__; __THREW__ = 0;
      $122 = $121&1;
      if (!($122)) {
       STACKTOP = sp;return; //@line 498 "modules/graphics/opengl/Shader.cpp"
      }
     }
    } while(0);
    $171 = ___cxa_find_matching_catch_3(0|0)|0;
    $172 = tempRet0;
    $lpad$phi43$index = $171;
    ___clang_call_terminate($lpad$phi43$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
    // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
   }
  }
 } while(0);
 if ((label|0) == 9) {
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  $eh$lpad$body$index2Z2D = $26;$eh$lpad$body$indexZ2D = $25;
 }
 $27 = ($0|0)==(0); //@line 53 "modules/graphics/opengl/Shader.cpp"
 do {
  if ($27) {
   $54 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
   $55 = ($54|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
   $56 = HEAP32[158581]|0; //@line 432 "modules/graphics/opengl/Shader.cpp"
   if ($55) {
    $82 = ($56|0)==(0|0); //@line 438 "modules/graphics/opengl/Shader.cpp"
    if (!($82)) {
     $83 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_vi($83|0,0); //@line 534 "./libraries/glad/gladfuncs.hpp"
     $84 = __THREW__; __THREW__ = 0;
     $85 = $84&1;
     if ($85) {
      break;
     }
    }
    HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
   $57 = ($56|0)==($54|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
   if ($57) {
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
   $58 = ((($54)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
   $59 = HEAP32[$58>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
   $60 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   invoke_vi($60|0,($59|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
   $61 = __THREW__; __THREW__ = 0;
   $62 = $61&1;
   if (!($62)) {
    HEAP32[158581] = $54; //@line 409 "modules/graphics/opengl/Shader.cpp"
    $63 = ((($54)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
    $64 = ((($54)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $65 = HEAP32[$64>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $66 = HEAP32[$63>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $67 = ($65|0)==($66|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
    L75: do {
     if (!($67)) {
      $174 = $65;$69 = $66;$i$01$i$i = 0;
      while(1) {
       $68 = $69; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $70 = (($68) + ($i$01$i$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $71 = HEAP32[$70>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
       $72 = ($71|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
       $$pre4$i$i = (($i$01$i$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
       if ($72) {
        $76 = $174;$77 = $69;
       } else {
        __THREW__ = 0;
        invoke_viiii(4111,(634044|0),($71|0),($$pre4$i$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
        $73 = __THREW__; __THREW__ = 0;
        $74 = $73&1;
        if ($74) {
         break;
        }
        $$pre$i$i = HEAP32[$64>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $$pre2$i$i = HEAP32[$63>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $76 = $$pre$i$i;$77 = $$pre2$i$i;
       }
       $75 = (($76) - ($77))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $78 = $75 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $79 = ($$pre4$i$i>>>0)<($78>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
       if ($79) {
        $174 = $76;$69 = $77;$i$01$i$i = $$pre4$i$i;
       } else {
        break L75;
       }
      }
      $86 = ___cxa_find_matching_catch_3(0|0)|0;
      $87 = tempRet0;
      $lpad$phi$index = $86;
      ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
      // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
     }
    } while(0);
    __THREW__ = 0;
    invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
    $80 = __THREW__; __THREW__ = 0;
    $81 = $80&1;
    if (!($81)) {
     ___resumeException($eh$lpad$body$indexZ2D|0);
     // unreachable;
    }
   }
  } else {
   $28 = HEAP32[158581]|0; //@line 406 "modules/graphics/opengl/Shader.cpp"
   $29 = ($28|0)==($$cast|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
   if (!($29)) {
    $30 = ((($$cast)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
    $31 = HEAP32[$30>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
    $32 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
    __THREW__ = 0;
    invoke_vi($32|0,($31|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
    $33 = __THREW__; __THREW__ = 0;
    $34 = $33&1;
    if ($34) {
     break;
    }
    HEAP32[158581] = $$cast; //@line 409 "modules/graphics/opengl/Shader.cpp"
   }
   $35 = ((($$cast)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
   $36 = ((($$cast)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $37 = HEAP32[$36>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $38 = HEAP32[$35>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $39 = ($37|0)==($38|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
   L91: do {
    if (!($39)) {
     $173 = $37;$43 = $38;$i$01$i = 0;
     while(1) {
      $42 = $43; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $44 = (($42) + ($i$01$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $45 = HEAP32[$44>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
      $46 = ($45|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
      $$pre4$i = (($i$01$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
      if ($46) {
       $50 = $173;$51 = $43;
      } else {
       __THREW__ = 0;
       invoke_viiii(4111,(634044|0),($45|0),($$pre4$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
       $47 = __THREW__; __THREW__ = 0;
       $48 = $47&1;
       if ($48) {
        break;
       }
       $$pre$i = HEAP32[$36>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $$pre2$i = HEAP32[$35>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $50 = $$pre$i;$51 = $$pre2$i;
      }
      $49 = (($50) - ($51))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $52 = $49 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $53 = ($$pre4$i>>>0)<($52>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      if ($53) {
       $173 = $50;$43 = $51;$i$01$i = $$pre4$i;
      } else {
       break L91;
      }
     }
     $88 = ___cxa_find_matching_catch_3(0|0)|0;
     $89 = tempRet0;
     $lpad$phi$index = $88;
     ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
     // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
    }
   } while(0);
   __THREW__ = 0;
   invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
   $40 = __THREW__; __THREW__ = 0;
   $41 = $40&1;
   if (!($41)) {
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
  }
 } while(0);
 $90 = ___cxa_find_matching_catch_3(0|0)|0;
 $91 = tempRet0;
 $lpad$phi$index = $90;
 ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
 // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
}
function __ZN4love8graphics6opengl6Shader9sendFloatERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEEiPKfi($this,$name,$size,$vec,$count) {
 $this = $this|0;
 $name = $name|0;
 $size = $size|0;
 $vec = $vec|0;
 $count = $count|0;
 var $$cast = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i28 = 0, $$pre$i18 = 0, $$pre2$i = 0, $$pre2$i$i = 0, $$pre2$i$i29 = 0, $$pre2$i19 = 0, $$pre4$i = 0, $$pre4$i$i = 0, $$pre4$i$i26 = 0, $$pre4$i16 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0;
 var $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i$01$i = 0, $i$01$i$i = 0, $i$01$i$i25 = 0, $i$01$i15 = 0, $lpad$phi$index = 0, $lpad$phi43$index = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP32[158581]|0; //@line 46 "modules/graphics/opengl/Shader.cpp"
 $$cast = $0; //@line 406 "modules/graphics/opengl/Shader.cpp"
 $1 = ($$cast|0)==($this|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
 if (!($1)) {
  $2 = ((($this)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
  $4 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$4 & 8191]($3); //@line 534 "./libraries/glad/gladfuncs.hpp"
  HEAP32[158581] = $this; //@line 409 "modules/graphics/opengl/Shader.cpp"
 }
 $5 = ((($this)) + 116|0); //@line 1092 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $6 = (__ZNKSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformEEENS_19__map_value_compareIS7_SD_NS_4lessIS7_EELb1EEENS5_ISD_EEE4findIS7_EENS_21__tree_const_iteratorISD_PNS_11__tree_nodeISD_PvEEiEERKT_($5,$name)|0); //@line 1092 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $7 = ((($this)) + 120|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = ($6|0)==($7|0); //@line 763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 L4: do {
  if ($8) {
   $9 = (___cxa_allocate_exception(16)|0); //@line 449 "modules/graphics/opengl/Shader.cpp"
   $10 = HEAP8[$name>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $11 = $10 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $12 = ($11<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $13 = ((($name)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $14 = HEAP32[$13>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $15 = ((($name)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $16 = $12 ? $15 : $14; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   HEAP32[$vararg_buffer>>2] = $16; //@line 449 "modules/graphics/opengl/Shader.cpp"
   invoke_viii(3349,($9|0),(438503|0),($vararg_buffer|0)); //@line 449 "modules/graphics/opengl/Shader.cpp"
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if ($18) {
    $20 = ___cxa_find_matching_catch_2()|0;
    $21 = tempRet0;
    ___cxa_free_exception(($9|0)); //@line 449 "modules/graphics/opengl/Shader.cpp"
    $eh$lpad$body$index2Z2D = $21;$eh$lpad$body$indexZ2D = $20;
    break;
   } else {
    __THREW__ = 0;
    invoke_viii(3350,($9|0),(96|0),(344|0)); //@line 449 "modules/graphics/opengl/Shader.cpp"
    $19 = __THREW__; __THREW__ = 0;
    label = 9;
    break;
   }
  } else {
   $22 = ((($6)) + 28|0); //@line 452 "modules/graphics/opengl/Shader.cpp"
   __THREW__ = 0;
   invoke_viiiii(4129,($this|0),($22|0),($size|0),($count|0),0); //@line 505 "modules/graphics/opengl/Shader.cpp"
   $23 = __THREW__; __THREW__ = 0;
   $24 = $23&1;
   if ($24) {
    label = 9;
   } else {
    switch ($size|0) {
    case 4:  {
     $92 = HEAP32[$22>>2]|0; //@line 510 "modules/graphics/opengl/Shader.cpp"
     $93 = HEAP32[165650]|0; //@line 546 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_viii($93|0,($92|0),($count|0),($vec|0)); //@line 546 "./libraries/glad/gladfuncs.hpp"
     $94 = __THREW__; __THREW__ = 0;
     $95 = $94&1;
     if ($95) {
      label = 9;
      break L4;
     }
     break;
    }
    case 3:  {
     $96 = HEAP32[$22>>2]|0; //@line 513 "modules/graphics/opengl/Shader.cpp"
     $97 = HEAP32[165017]|0; //@line 545 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_viii($97|0,($96|0),($count|0),($vec|0)); //@line 545 "./libraries/glad/gladfuncs.hpp"
     $98 = __THREW__; __THREW__ = 0;
     $99 = $98&1;
     if ($99) {
      label = 9;
      break L4;
     }
     break;
    }
    case 2:  {
     $100 = HEAP32[$22>>2]|0; //@line 516 "modules/graphics/opengl/Shader.cpp"
     $101 = HEAP32[165418]|0; //@line 544 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_viii($101|0,($100|0),($count|0),($vec|0)); //@line 544 "./libraries/glad/gladfuncs.hpp"
     $102 = __THREW__; __THREW__ = 0;
     $103 = $102&1;
     if ($103) {
      label = 9;
      break L4;
     }
     break;
    }
    default: {
     $104 = HEAP32[$22>>2]|0; //@line 520 "modules/graphics/opengl/Shader.cpp"
     $105 = HEAP32[165610]|0; //@line 543 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_viii($105|0,($104|0),($count|0),($vec|0)); //@line 543 "./libraries/glad/gladfuncs.hpp"
     $106 = __THREW__; __THREW__ = 0;
     $107 = $106&1;
     if ($107) {
      label = 9;
      break L4;
     }
    }
    }
    $108 = ($0|0)==(0); //@line 53 "modules/graphics/opengl/Shader.cpp"
    do {
     if ($108) {
      $135 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
      $136 = ($135|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
      $137 = HEAP32[158581]|0; //@line 432 "modules/graphics/opengl/Shader.cpp"
      if ($136) {
       $163 = ($137|0)==(0|0); //@line 438 "modules/graphics/opengl/Shader.cpp"
       if (!($163)) {
        $164 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
        __THREW__ = 0;
        invoke_vi($164|0,0); //@line 534 "./libraries/glad/gladfuncs.hpp"
        $165 = __THREW__; __THREW__ = 0;
        $166 = $165&1;
        if ($166) {
         break;
        }
       }
       HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
       STACKTOP = sp;return; //@line 523 "modules/graphics/opengl/Shader.cpp"
      }
      $138 = ($137|0)==($135|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
      if ($138) {
       STACKTOP = sp;return; //@line 523 "modules/graphics/opengl/Shader.cpp"
      }
      $139 = ((($135)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
      $140 = HEAP32[$139>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
      $141 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_vi($141|0,($140|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
      $142 = __THREW__; __THREW__ = 0;
      $143 = $142&1;
      if (!($143)) {
       HEAP32[158581] = $135; //@line 409 "modules/graphics/opengl/Shader.cpp"
       $144 = ((($135)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
       $145 = ((($135)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $146 = HEAP32[$145>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $147 = HEAP32[$144>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $148 = ($146|0)==($147|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
       L30: do {
        if (!($148)) {
         $150 = $147;$176 = $146;$i$01$i$i25 = 0;
         while(1) {
          $149 = $150; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $151 = (($149) + ($i$01$i$i25<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $152 = HEAP32[$151>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
          $153 = ($152|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
          $$pre4$i$i26 = (($i$01$i$i25) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
          if ($153) {
           $157 = $176;$158 = $150;
          } else {
           __THREW__ = 0;
           invoke_viiii(4111,(634044|0),($152|0),($$pre4$i$i26|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
           $154 = __THREW__; __THREW__ = 0;
           $155 = $154&1;
           if ($155) {
            break;
           }
           $$pre$i$i28 = HEAP32[$145>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $$pre2$i$i29 = HEAP32[$144>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $157 = $$pre$i$i28;$158 = $$pre2$i$i29;
          }
          $156 = (($157) - ($158))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $159 = $156 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $160 = ($$pre4$i$i26>>>0)<($159>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
          if ($160) {
           $150 = $158;$176 = $157;$i$01$i$i25 = $$pre4$i$i26;
          } else {
           break L30;
          }
         }
         $167 = ___cxa_find_matching_catch_3(0|0)|0;
         $168 = tempRet0;
         $lpad$phi43$index = $167;
         ___clang_call_terminate($lpad$phi43$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
         // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
        }
       } while(0);
       __THREW__ = 0;
       invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
       $161 = __THREW__; __THREW__ = 0;
       $162 = $161&1;
       if (!($162)) {
        STACKTOP = sp;return; //@line 523 "modules/graphics/opengl/Shader.cpp"
       }
      }
     } else {
      $109 = HEAP32[158581]|0; //@line 406 "modules/graphics/opengl/Shader.cpp"
      $110 = ($109|0)==($$cast|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
      if (!($110)) {
       $111 = ((($$cast)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
       $112 = HEAP32[$111>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
       $113 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
       __THREW__ = 0;
       invoke_vi($113|0,($112|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
       $114 = __THREW__; __THREW__ = 0;
       $115 = $114&1;
       if ($115) {
        break;
       }
       HEAP32[158581] = $$cast; //@line 409 "modules/graphics/opengl/Shader.cpp"
      }
      $116 = ((($$cast)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      $117 = ((($$cast)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $118 = HEAP32[$117>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $119 = HEAP32[$116>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $120 = ($118|0)==($119|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      L46: do {
       if (!($120)) {
        $124 = $119;$175 = $118;$i$01$i15 = 0;
        while(1) {
         $123 = $124; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $125 = (($123) + ($i$01$i15<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $126 = HEAP32[$125>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
         $127 = ($126|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
         $$pre4$i16 = (($i$01$i15) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
         if ($127) {
          $131 = $175;$132 = $124;
         } else {
          __THREW__ = 0;
          invoke_viiii(4111,(634044|0),($126|0),($$pre4$i16|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
          $128 = __THREW__; __THREW__ = 0;
          $129 = $128&1;
          if ($129) {
           break;
          }
          $$pre$i18 = HEAP32[$117>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $$pre2$i19 = HEAP32[$116>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $131 = $$pre$i18;$132 = $$pre2$i19;
         }
         $130 = (($131) - ($132))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $133 = $130 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $134 = ($$pre4$i16>>>0)<($133>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
         if ($134) {
          $124 = $132;$175 = $131;$i$01$i15 = $$pre4$i16;
         } else {
          break L46;
         }
        }
        $169 = ___cxa_find_matching_catch_3(0|0)|0;
        $170 = tempRet0;
        $lpad$phi43$index = $169;
        ___clang_call_terminate($lpad$phi43$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
        // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
       }
      } while(0);
      __THREW__ = 0;
      invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
      $121 = __THREW__; __THREW__ = 0;
      $122 = $121&1;
      if (!($122)) {
       STACKTOP = sp;return; //@line 523 "modules/graphics/opengl/Shader.cpp"
      }
     }
    } while(0);
    $171 = ___cxa_find_matching_catch_3(0|0)|0;
    $172 = tempRet0;
    $lpad$phi43$index = $171;
    ___clang_call_terminate($lpad$phi43$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
    // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
   }
  }
 } while(0);
 if ((label|0) == 9) {
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  $eh$lpad$body$index2Z2D = $26;$eh$lpad$body$indexZ2D = $25;
 }
 $27 = ($0|0)==(0); //@line 53 "modules/graphics/opengl/Shader.cpp"
 do {
  if ($27) {
   $54 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
   $55 = ($54|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
   $56 = HEAP32[158581]|0; //@line 432 "modules/graphics/opengl/Shader.cpp"
   if ($55) {
    $82 = ($56|0)==(0|0); //@line 438 "modules/graphics/opengl/Shader.cpp"
    if (!($82)) {
     $83 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_vi($83|0,0); //@line 534 "./libraries/glad/gladfuncs.hpp"
     $84 = __THREW__; __THREW__ = 0;
     $85 = $84&1;
     if ($85) {
      break;
     }
    }
    HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
   $57 = ($56|0)==($54|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
   if ($57) {
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
   $58 = ((($54)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
   $59 = HEAP32[$58>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
   $60 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   invoke_vi($60|0,($59|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
   $61 = __THREW__; __THREW__ = 0;
   $62 = $61&1;
   if (!($62)) {
    HEAP32[158581] = $54; //@line 409 "modules/graphics/opengl/Shader.cpp"
    $63 = ((($54)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
    $64 = ((($54)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $65 = HEAP32[$64>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $66 = HEAP32[$63>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $67 = ($65|0)==($66|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
    L75: do {
     if (!($67)) {
      $174 = $65;$69 = $66;$i$01$i$i = 0;
      while(1) {
       $68 = $69; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $70 = (($68) + ($i$01$i$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $71 = HEAP32[$70>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
       $72 = ($71|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
       $$pre4$i$i = (($i$01$i$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
       if ($72) {
        $76 = $174;$77 = $69;
       } else {
        __THREW__ = 0;
        invoke_viiii(4111,(634044|0),($71|0),($$pre4$i$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
        $73 = __THREW__; __THREW__ = 0;
        $74 = $73&1;
        if ($74) {
         break;
        }
        $$pre$i$i = HEAP32[$64>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $$pre2$i$i = HEAP32[$63>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $76 = $$pre$i$i;$77 = $$pre2$i$i;
       }
       $75 = (($76) - ($77))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $78 = $75 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $79 = ($$pre4$i$i>>>0)<($78>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
       if ($79) {
        $174 = $76;$69 = $77;$i$01$i$i = $$pre4$i$i;
       } else {
        break L75;
       }
      }
      $86 = ___cxa_find_matching_catch_3(0|0)|0;
      $87 = tempRet0;
      $lpad$phi$index = $86;
      ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
      // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
     }
    } while(0);
    __THREW__ = 0;
    invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
    $80 = __THREW__; __THREW__ = 0;
    $81 = $80&1;
    if (!($81)) {
     ___resumeException($eh$lpad$body$indexZ2D|0);
     // unreachable;
    }
   }
  } else {
   $28 = HEAP32[158581]|0; //@line 406 "modules/graphics/opengl/Shader.cpp"
   $29 = ($28|0)==($$cast|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
   if (!($29)) {
    $30 = ((($$cast)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
    $31 = HEAP32[$30>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
    $32 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
    __THREW__ = 0;
    invoke_vi($32|0,($31|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
    $33 = __THREW__; __THREW__ = 0;
    $34 = $33&1;
    if ($34) {
     break;
    }
    HEAP32[158581] = $$cast; //@line 409 "modules/graphics/opengl/Shader.cpp"
   }
   $35 = ((($$cast)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
   $36 = ((($$cast)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $37 = HEAP32[$36>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $38 = HEAP32[$35>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $39 = ($37|0)==($38|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
   L91: do {
    if (!($39)) {
     $173 = $37;$43 = $38;$i$01$i = 0;
     while(1) {
      $42 = $43; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $44 = (($42) + ($i$01$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $45 = HEAP32[$44>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
      $46 = ($45|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
      $$pre4$i = (($i$01$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
      if ($46) {
       $50 = $173;$51 = $43;
      } else {
       __THREW__ = 0;
       invoke_viiii(4111,(634044|0),($45|0),($$pre4$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
       $47 = __THREW__; __THREW__ = 0;
       $48 = $47&1;
       if ($48) {
        break;
       }
       $$pre$i = HEAP32[$36>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $$pre2$i = HEAP32[$35>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $50 = $$pre$i;$51 = $$pre2$i;
      }
      $49 = (($50) - ($51))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $52 = $49 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $53 = ($$pre4$i>>>0)<($52>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      if ($53) {
       $173 = $50;$43 = $51;$i$01$i = $$pre4$i;
      } else {
       break L91;
      }
     }
     $88 = ___cxa_find_matching_catch_3(0|0)|0;
     $89 = tempRet0;
     $lpad$phi$index = $88;
     ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
     // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
    }
   } while(0);
   __THREW__ = 0;
   invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
   $40 = __THREW__; __THREW__ = 0;
   $41 = $40&1;
   if (!($41)) {
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
  }
 } while(0);
 $90 = ___cxa_find_matching_catch_3(0|0)|0;
 $91 = tempRet0;
 $lpad$phi$index = $90;
 ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
 // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
}
function __ZN4love8graphics6opengl6Shader10sendMatrixERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEEiPKfi($this,$name,$size,$m,$count) {
 $this = $this|0;
 $name = $name|0;
 $size = $size|0;
 $m = $m|0;
 $count = $count|0;
 var $$0 = 0, $$02 = 0, $$cast = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i29 = 0, $$pre$i19 = 0, $$pre2$i = 0, $$pre2$i$i = 0, $$pre2$i$i30 = 0, $$pre2$i20 = 0, $$pre4$i = 0, $$pre4$i$i = 0, $$pre4$i$i27 = 0, $$pre4$i17 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index6Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i$01$i = 0, $i$01$i$i = 0, $i$01$i$i26 = 0;
 var $i$01$i16 = 0, $lpad$phi$index = 0, $lpad$phi44$index = 0, $size$off = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[158581]|0; //@line 46 "modules/graphics/opengl/Shader.cpp"
 $$cast = $0; //@line 406 "modules/graphics/opengl/Shader.cpp"
 $1 = ($$cast|0)==($this|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
 if (!($1)) {
  $2 = ((($this)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
  $4 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$4 & 8191]($3); //@line 534 "./libraries/glad/gladfuncs.hpp"
  HEAP32[158581] = $this; //@line 409 "modules/graphics/opengl/Shader.cpp"
 }
 $size$off = (($size) + -2)|0; //@line 529 "modules/graphics/opengl/Shader.cpp"
 $5 = ($size$off>>>0)>(2); //@line 529 "modules/graphics/opengl/Shader.cpp"
 L4: do {
  if ($5) {
   $6 = (___cxa_allocate_exception(16)|0); //@line 531 "modules/graphics/opengl/Shader.cpp"
   __THREW__ = 0;
   HEAP32[$vararg_buffer>>2] = $size; //@line 531 "modules/graphics/opengl/Shader.cpp"
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 531 "modules/graphics/opengl/Shader.cpp"
   HEAP32[$vararg_ptr1>>2] = $size; //@line 531 "modules/graphics/opengl/Shader.cpp"
   invoke_viii(3349,($6|0),(438807|0),($vararg_buffer|0)); //@line 531 "modules/graphics/opengl/Shader.cpp"
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if ($8) {
    $10 = ___cxa_find_matching_catch_2()|0;
    $11 = tempRet0;
    ___cxa_free_exception(($6|0)); //@line 531 "modules/graphics/opengl/Shader.cpp"
    $$0 = $10;$$02 = $11;
    break;
   } else {
    __THREW__ = 0;
    invoke_viii(3350,($6|0),(96|0),(344|0)); //@line 531 "modules/graphics/opengl/Shader.cpp"
    $9 = __THREW__; __THREW__ = 0;
    label = 7;
    break;
   }
  } else {
   $14 = ((($this)) + 116|0); //@line 1092 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $15 = (__ZNKSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformEEENS_19__map_value_compareIS7_SD_NS_4lessIS7_EELb1EEENS5_ISD_EEE4findIS7_EENS_21__tree_const_iteratorISD_PNS_11__tree_nodeISD_PvEEiEERKT_($14,$name)|0); //@line 1092 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $16 = ((($this)) + 120|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $17 = ($15|0)==($16|0); //@line 763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($17) {
    $18 = (___cxa_allocate_exception(16)|0); //@line 449 "modules/graphics/opengl/Shader.cpp"
    $19 = HEAP8[$name>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $20 = $19 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $21 = ($20<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $22 = ((($name)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $23 = HEAP32[$22>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $24 = ((($name)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $25 = $21 ? $24 : $23; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    __THREW__ = 0;
    HEAP32[$vararg_buffer2>>2] = $25; //@line 449 "modules/graphics/opengl/Shader.cpp"
    invoke_viii(3349,($18|0),(438503|0),($vararg_buffer2|0)); //@line 449 "modules/graphics/opengl/Shader.cpp"
    $26 = __THREW__; __THREW__ = 0;
    $27 = $26&1;
    if ($27) {
     $29 = ___cxa_find_matching_catch_2()|0;
     $30 = tempRet0;
     ___cxa_free_exception(($18|0)); //@line 449 "modules/graphics/opengl/Shader.cpp"
     $eh$lpad$body$index6Z2D = $30;$eh$lpad$body$indexZ2D = $29;
     label = 8;
     break;
    } else {
     __THREW__ = 0;
     invoke_viii(3350,($18|0),(96|0),(344|0)); //@line 449 "modules/graphics/opengl/Shader.cpp"
     $28 = __THREW__; __THREW__ = 0;
     label = 7;
     break;
    }
   }
   $31 = ((($15)) + 28|0); //@line 452 "modules/graphics/opengl/Shader.cpp"
   __THREW__ = 0;
   invoke_viiiii(4129,($this|0),($31|0),($size|0),($count|0),0); //@line 536 "modules/graphics/opengl/Shader.cpp"
   $32 = __THREW__; __THREW__ = 0;
   $33 = $32&1;
   if ($33) {
    label = 7;
   } else {
    switch ($size|0) {
    case 4:  {
     $34 = HEAP32[$31>>2]|0; //@line 541 "modules/graphics/opengl/Shader.cpp"
     $35 = HEAP32[165649]|0; //@line 553 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_viiii($35|0,($34|0),($count|0),0,($m|0)); //@line 553 "./libraries/glad/gladfuncs.hpp"
     $36 = __THREW__; __THREW__ = 0;
     $37 = $36&1;
     if ($37) {
      label = 7;
      break L4;
     }
     break;
    }
    case 3:  {
     $38 = HEAP32[$31>>2]|0; //@line 544 "modules/graphics/opengl/Shader.cpp"
     $39 = HEAP32[165618]|0; //@line 552 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_viiii($39|0,($38|0),($count|0),0,($m|0)); //@line 552 "./libraries/glad/gladfuncs.hpp"
     $40 = __THREW__; __THREW__ = 0;
     $41 = $40&1;
     if ($41) {
      label = 7;
      break L4;
     }
     break;
    }
    default: {
     $42 = HEAP32[$31>>2]|0; //@line 548 "modules/graphics/opengl/Shader.cpp"
     $43 = HEAP32[165563]|0; //@line 551 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_viiii($43|0,($42|0),($count|0),0,($m|0)); //@line 551 "./libraries/glad/gladfuncs.hpp"
     $44 = __THREW__; __THREW__ = 0;
     $45 = $44&1;
     if ($45) {
      label = 7;
      break L4;
     }
    }
    }
    $46 = ($0|0)==(0); //@line 53 "modules/graphics/opengl/Shader.cpp"
    do {
     if ($46) {
      $73 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
      $74 = ($73|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
      $75 = HEAP32[158581]|0; //@line 432 "modules/graphics/opengl/Shader.cpp"
      if ($74) {
       $101 = ($75|0)==(0|0); //@line 438 "modules/graphics/opengl/Shader.cpp"
       if (!($101)) {
        $102 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
        __THREW__ = 0;
        invoke_vi($102|0,0); //@line 534 "./libraries/glad/gladfuncs.hpp"
        $103 = __THREW__; __THREW__ = 0;
        $104 = $103&1;
        if ($104) {
         break;
        }
       }
       HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
       STACKTOP = sp;return; //@line 551 "modules/graphics/opengl/Shader.cpp"
      }
      $76 = ($75|0)==($73|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
      if ($76) {
       STACKTOP = sp;return; //@line 551 "modules/graphics/opengl/Shader.cpp"
      }
      $77 = ((($73)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
      $78 = HEAP32[$77>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
      $79 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_vi($79|0,($78|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
      $80 = __THREW__; __THREW__ = 0;
      $81 = $80&1;
      if (!($81)) {
       HEAP32[158581] = $73; //@line 409 "modules/graphics/opengl/Shader.cpp"
       $82 = ((($73)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
       $83 = ((($73)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $84 = HEAP32[$83>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $85 = HEAP32[$82>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $86 = ($84|0)==($85|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
       L35: do {
        if (!($86)) {
         $177 = $84;$88 = $85;$i$01$i$i = 0;
         while(1) {
          $87 = $88; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $89 = (($87) + ($i$01$i$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $90 = HEAP32[$89>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
          $91 = ($90|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
          $$pre4$i$i = (($i$01$i$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
          if ($91) {
           $95 = $177;$96 = $88;
          } else {
           __THREW__ = 0;
           invoke_viiii(4111,(634044|0),($90|0),($$pre4$i$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
           $92 = __THREW__; __THREW__ = 0;
           $93 = $92&1;
           if ($93) {
            break;
           }
           $$pre$i$i = HEAP32[$83>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $$pre2$i$i = HEAP32[$82>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $95 = $$pre$i$i;$96 = $$pre2$i$i;
          }
          $94 = (($95) - ($96))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $97 = $94 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $98 = ($$pre4$i$i>>>0)<($97>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
          if ($98) {
           $177 = $95;$88 = $96;$i$01$i$i = $$pre4$i$i;
          } else {
           break L35;
          }
         }
         $105 = ___cxa_find_matching_catch_3(0|0)|0;
         $106 = tempRet0;
         $lpad$phi44$index = $105;
         ___clang_call_terminate($lpad$phi44$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
         // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
        }
       } while(0);
       __THREW__ = 0;
       invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
       $99 = __THREW__; __THREW__ = 0;
       $100 = $99&1;
       if (!($100)) {
        STACKTOP = sp;return; //@line 551 "modules/graphics/opengl/Shader.cpp"
       }
      }
     } else {
      $47 = HEAP32[158581]|0; //@line 406 "modules/graphics/opengl/Shader.cpp"
      $48 = ($47|0)==($$cast|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
      if (!($48)) {
       $49 = ((($$cast)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
       $50 = HEAP32[$49>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
       $51 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
       __THREW__ = 0;
       invoke_vi($51|0,($50|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
       $52 = __THREW__; __THREW__ = 0;
       $53 = $52&1;
       if ($53) {
        break;
       }
       HEAP32[158581] = $$cast; //@line 409 "modules/graphics/opengl/Shader.cpp"
      }
      $54 = ((($$cast)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      $55 = ((($$cast)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $56 = HEAP32[$55>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $57 = HEAP32[$54>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $58 = ($56|0)==($57|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      L51: do {
       if (!($58)) {
        $176 = $56;$62 = $57;$i$01$i = 0;
        while(1) {
         $61 = $62; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $63 = (($61) + ($i$01$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $64 = HEAP32[$63>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
         $65 = ($64|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
         $$pre4$i = (($i$01$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
         if ($65) {
          $69 = $176;$70 = $62;
         } else {
          __THREW__ = 0;
          invoke_viiii(4111,(634044|0),($64|0),($$pre4$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
          $66 = __THREW__; __THREW__ = 0;
          $67 = $66&1;
          if ($67) {
           break;
          }
          $$pre$i = HEAP32[$55>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $$pre2$i = HEAP32[$54>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $69 = $$pre$i;$70 = $$pre2$i;
         }
         $68 = (($69) - ($70))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $71 = $68 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $72 = ($$pre4$i>>>0)<($71>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
         if ($72) {
          $176 = $69;$62 = $70;$i$01$i = $$pre4$i;
         } else {
          break L51;
         }
        }
        $107 = ___cxa_find_matching_catch_3(0|0)|0;
        $108 = tempRet0;
        $lpad$phi44$index = $107;
        ___clang_call_terminate($lpad$phi44$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
        // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
       }
      } while(0);
      __THREW__ = 0;
      invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if (!($60)) {
       STACKTOP = sp;return; //@line 551 "modules/graphics/opengl/Shader.cpp"
      }
     }
    } while(0);
    $109 = ___cxa_find_matching_catch_3(0|0)|0;
    $110 = tempRet0;
    $lpad$phi44$index = $109;
    ___clang_call_terminate($lpad$phi44$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
    // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
   }
  }
 } while(0);
 if ((label|0) == 7) {
  $12 = ___cxa_find_matching_catch_2()|0;
  $13 = tempRet0;
  $eh$lpad$body$index6Z2D = $13;$eh$lpad$body$indexZ2D = $12;
  label = 8;
 }
 if ((label|0) == 8) {
  $$0 = $eh$lpad$body$indexZ2D;$$02 = $eh$lpad$body$index6Z2D;
 }
 $111 = ($0|0)==(0); //@line 53 "modules/graphics/opengl/Shader.cpp"
 do {
  if ($111) {
   $138 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
   $139 = ($138|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
   $140 = HEAP32[158581]|0; //@line 432 "modules/graphics/opengl/Shader.cpp"
   if ($139) {
    $166 = ($140|0)==(0|0); //@line 438 "modules/graphics/opengl/Shader.cpp"
    if (!($166)) {
     $167 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_vi($167|0,0); //@line 534 "./libraries/glad/gladfuncs.hpp"
     $168 = __THREW__; __THREW__ = 0;
     $169 = $168&1;
     if ($169) {
      break;
     }
    }
    HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
    ___resumeException($$0|0);
    // unreachable;
   }
   $141 = ($140|0)==($138|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
   if ($141) {
    ___resumeException($$0|0);
    // unreachable;
   }
   $142 = ((($138)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
   $143 = HEAP32[$142>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
   $144 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   invoke_vi($144|0,($143|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
   $145 = __THREW__; __THREW__ = 0;
   $146 = $145&1;
   if (!($146)) {
    HEAP32[158581] = $138; //@line 409 "modules/graphics/opengl/Shader.cpp"
    $147 = ((($138)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
    $148 = ((($138)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $149 = HEAP32[$148>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $150 = HEAP32[$147>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $151 = ($149|0)==($150|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
    L82: do {
     if (!($151)) {
      $153 = $150;$179 = $149;$i$01$i$i26 = 0;
      while(1) {
       $152 = $153; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $154 = (($152) + ($i$01$i$i26<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $155 = HEAP32[$154>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
       $156 = ($155|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
       $$pre4$i$i27 = (($i$01$i$i26) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
       if ($156) {
        $160 = $179;$161 = $153;
       } else {
        __THREW__ = 0;
        invoke_viiii(4111,(634044|0),($155|0),($$pre4$i$i27|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
        $157 = __THREW__; __THREW__ = 0;
        $158 = $157&1;
        if ($158) {
         break;
        }
        $$pre$i$i29 = HEAP32[$148>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $$pre2$i$i30 = HEAP32[$147>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $160 = $$pre$i$i29;$161 = $$pre2$i$i30;
       }
       $159 = (($160) - ($161))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $162 = $159 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $163 = ($$pre4$i$i27>>>0)<($162>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
       if ($163) {
        $153 = $161;$179 = $160;$i$01$i$i26 = $$pre4$i$i27;
       } else {
        break L82;
       }
      }
      $170 = ___cxa_find_matching_catch_3(0|0)|0;
      $171 = tempRet0;
      $lpad$phi$index = $170;
      ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
      // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
     }
    } while(0);
    __THREW__ = 0;
    invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
    $164 = __THREW__; __THREW__ = 0;
    $165 = $164&1;
    if (!($165)) {
     ___resumeException($$0|0);
     // unreachable;
    }
   }
  } else {
   $112 = HEAP32[158581]|0; //@line 406 "modules/graphics/opengl/Shader.cpp"
   $113 = ($112|0)==($$cast|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
   if (!($113)) {
    $114 = ((($$cast)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
    $115 = HEAP32[$114>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
    $116 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
    __THREW__ = 0;
    invoke_vi($116|0,($115|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
    $117 = __THREW__; __THREW__ = 0;
    $118 = $117&1;
    if ($118) {
     break;
    }
    HEAP32[158581] = $$cast; //@line 409 "modules/graphics/opengl/Shader.cpp"
   }
   $119 = ((($$cast)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
   $120 = ((($$cast)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $121 = HEAP32[$120>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $122 = HEAP32[$119>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $123 = ($121|0)==($122|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
   L98: do {
    if (!($123)) {
     $127 = $122;$178 = $121;$i$01$i16 = 0;
     while(1) {
      $126 = $127; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $128 = (($126) + ($i$01$i16<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $129 = HEAP32[$128>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
      $130 = ($129|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
      $$pre4$i17 = (($i$01$i16) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
      if ($130) {
       $134 = $178;$135 = $127;
      } else {
       __THREW__ = 0;
       invoke_viiii(4111,(634044|0),($129|0),($$pre4$i17|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
       $131 = __THREW__; __THREW__ = 0;
       $132 = $131&1;
       if ($132) {
        break;
       }
       $$pre$i19 = HEAP32[$120>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $$pre2$i20 = HEAP32[$119>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $134 = $$pre$i19;$135 = $$pre2$i20;
      }
      $133 = (($134) - ($135))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $136 = $133 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $137 = ($$pre4$i17>>>0)<($136>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      if ($137) {
       $127 = $135;$178 = $134;$i$01$i16 = $$pre4$i17;
      } else {
       break L98;
      }
     }
     $172 = ___cxa_find_matching_catch_3(0|0)|0;
     $173 = tempRet0;
     $lpad$phi$index = $172;
     ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
     // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
    }
   } while(0);
   __THREW__ = 0;
   invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
   $124 = __THREW__; __THREW__ = 0;
   $125 = $124&1;
   if (!($125)) {
    ___resumeException($$0|0);
    // unreachable;
   }
  }
 } while(0);
 $174 = ___cxa_find_matching_catch_3(0|0)|0;
 $175 = tempRet0;
 $lpad$phi$index = $174;
 ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
 // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
}
function __ZN4love8graphics6opengl6Shader11sendTextureERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEEPNS0_7TextureE($this,$name,$texture) {
 $this = $this|0;
 $name = $name|0;
 $texture = $texture|0;
 var $$cast = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i28 = 0, $$pre$i18 = 0, $$pre2$i = 0, $$pre2$i$i = 0, $$pre2$i$i29 = 0, $$pre2$i19 = 0, $$pre4$i = 0, $$pre4$i$i = 0, $$pre4$i$i26 = 0, $$pre4$i16 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0;
 var $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0;
 var $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0;
 var $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i$01$i = 0;
 var $i$01$i$i = 0, $i$01$i$i25 = 0, $i$01$i15 = 0, $lpad$phi$index = 0, $lpad$phi43$index = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP32[$texture>>2]|0; //@line 555 "modules/graphics/opengl/Shader.cpp"
 $1 = ((($0)) + 52|0); //@line 555 "modules/graphics/opengl/Shader.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 555 "modules/graphics/opengl/Shader.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($texture)|0); //@line 555 "modules/graphics/opengl/Shader.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 555 "modules/graphics/opengl/Shader.cpp"
 $5 = HEAP32[158581]|0; //@line 46 "modules/graphics/opengl/Shader.cpp"
 $$cast = $5; //@line 406 "modules/graphics/opengl/Shader.cpp"
 $6 = ($$cast|0)==($this|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
 if (!($6)) {
  $7 = ((($this)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
  $8 = HEAP32[$7>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
  $9 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$9 & 8191]($8); //@line 534 "./libraries/glad/gladfuncs.hpp"
  HEAP32[158581] = $this; //@line 409 "modules/graphics/opengl/Shader.cpp"
 }
 __THREW__ = 0;
 $10 = (invoke_iii(4130,($this|0),($name|0))|0); //@line 559 "modules/graphics/opengl/Shader.cpp"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 do {
  if ($12) {
   label = 13;
  } else {
   $13 = ((($this)) + 116|0); //@line 1092 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $14 = (__ZNKSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformEEENS_19__map_value_compareIS7_SD_NS_4lessIS7_EELb1EEENS5_ISD_EEE4findIS7_EENS_21__tree_const_iteratorISD_PNS_11__tree_nodeISD_PvEEiEERKT_($13,$name)|0); //@line 1092 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $15 = ((($this)) + 120|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $16 = ($14|0)==($15|0); //@line 763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($16) {
    $17 = (___cxa_allocate_exception(16)|0); //@line 449 "modules/graphics/opengl/Shader.cpp"
    $18 = HEAP8[$name>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $19 = $18 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $20 = ($19<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $21 = ((($name)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $22 = HEAP32[$21>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $23 = ((($name)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $24 = $20 ? $23 : $22; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    __THREW__ = 0;
    HEAP32[$vararg_buffer>>2] = $24; //@line 449 "modules/graphics/opengl/Shader.cpp"
    invoke_viii(3349,($17|0),(438503|0),($vararg_buffer|0)); //@line 449 "modules/graphics/opengl/Shader.cpp"
    $25 = __THREW__; __THREW__ = 0;
    $26 = $25&1;
    if ($26) {
     $28 = ___cxa_find_matching_catch_2()|0;
     $29 = tempRet0;
     ___cxa_free_exception(($17|0)); //@line 449 "modules/graphics/opengl/Shader.cpp"
     $eh$lpad$body$index2Z2D = $29;$eh$lpad$body$indexZ2D = $28;
     break;
    } else {
     __THREW__ = 0;
     invoke_viii(3350,($17|0),(96|0),(344|0)); //@line 449 "modules/graphics/opengl/Shader.cpp"
     $27 = __THREW__; __THREW__ = 0;
     label = 13;
     break;
    }
   }
   $30 = ((($14)) + 28|0); //@line 452 "modules/graphics/opengl/Shader.cpp"
   __THREW__ = 0;
   invoke_viiiii(4129,($this|0),($30|0),1,1,3); //@line 562 "modules/graphics/opengl/Shader.cpp"
   $31 = __THREW__; __THREW__ = 0;
   $32 = $31&1;
   if ($32) {
    label = 13;
   } else {
    __THREW__ = 0;
    invoke_viiii(4111,(634044|0),($4|0),($10|0),1); //@line 565 "modules/graphics/opengl/Shader.cpp"
    $33 = __THREW__; __THREW__ = 0;
    $34 = $33&1;
    if ($34) {
     label = 13;
    } else {
     $35 = HEAP32[$30>>2]|0; //@line 567 "modules/graphics/opengl/Shader.cpp"
     $36 = HEAP32[165188]|0; //@line 539 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_vii($36|0,($35|0),($10|0)); //@line 539 "./libraries/glad/gladfuncs.hpp"
     $37 = __THREW__; __THREW__ = 0;
     $38 = $37&1;
     if ($38) {
      label = 13;
     } else {
      $39 = (($10) + -1)|0; //@line 570 "modules/graphics/opengl/Shader.cpp"
      $40 = ((($this)) + 140|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $41 = HEAP32[$40>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $42 = (($41) + ($39<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $43 = HEAP32[$42>>2]|0; //@line 570 "modules/graphics/opengl/Shader.cpp"
      $44 = ($43|0)==(0); //@line 570 "modules/graphics/opengl/Shader.cpp"
      if ($44) {
       $45 = HEAP32[158608]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $46 = (($45) + ($39<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $47 = HEAP32[$46>>2]|0; //@line 571 "modules/graphics/opengl/Shader.cpp"
       $48 = (($47) + 1)|0; //@line 571 "modules/graphics/opengl/Shader.cpp"
       HEAP32[$46>>2] = $48; //@line 571 "modules/graphics/opengl/Shader.cpp"
      }
      HEAP32[$42>>2] = $4; //@line 574 "modules/graphics/opengl/Shader.cpp"
      $116 = HEAP32[$texture>>2]|0; //@line 581 "modules/graphics/opengl/Shader.cpp"
      $117 = ((($116)) + 8|0); //@line 581 "modules/graphics/opengl/Shader.cpp"
      $118 = HEAP32[$117>>2]|0; //@line 581 "modules/graphics/opengl/Shader.cpp"
      __THREW__ = 0;
      invoke_vi($118|0,($texture|0)); //@line 581 "modules/graphics/opengl/Shader.cpp"
      $119 = __THREW__; __THREW__ = 0;
      $120 = $119&1;
      if ($120) {
       label = 13;
      } else {
       $121 = ((($this)) + 152|0); //@line 583 "modules/graphics/opengl/Shader.cpp"
       $122 = (__ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ObjectEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE4findIS7_EENS_15__tree_iteratorISB_PNS_11__tree_nodeISB_PvEEiEERKT_($121,$name)|0); //@line 1090 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
       $123 = ((($this)) + 156|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $124 = ($122|0)==($123|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
       if (!($124)) {
        $125 = ((($122)) + 28|0); //@line 585 "modules/graphics/opengl/Shader.cpp"
        $126 = HEAP32[$125>>2]|0; //@line 585 "modules/graphics/opengl/Shader.cpp"
        $127 = HEAP32[$126>>2]|0; //@line 585 "modules/graphics/opengl/Shader.cpp"
        $128 = ((($127)) + 12|0); //@line 585 "modules/graphics/opengl/Shader.cpp"
        $129 = HEAP32[$128>>2]|0; //@line 585 "modules/graphics/opengl/Shader.cpp"
        __THREW__ = 0;
        invoke_vi($129|0,($126|0)); //@line 585 "modules/graphics/opengl/Shader.cpp"
        $130 = __THREW__; __THREW__ = 0;
        $131 = $130&1;
        if ($131) {
         label = 13;
         break;
        }
       }
       __THREW__ = 0;
       $132 = (invoke_iii(4131,($121|0),($name|0))|0); //@line 587 "modules/graphics/opengl/Shader.cpp"
       $133 = __THREW__; __THREW__ = 0;
       $134 = $133&1;
       if ($134) {
        label = 13;
       } else {
        HEAP32[$132>>2] = $texture; //@line 587 "modules/graphics/opengl/Shader.cpp"
        $135 = ($5|0)==(0); //@line 53 "modules/graphics/opengl/Shader.cpp"
        do {
         if ($135) {
          $162 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
          $163 = ($162|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
          $164 = HEAP32[158581]|0; //@line 432 "modules/graphics/opengl/Shader.cpp"
          if ($163) {
           $190 = ($164|0)==(0|0); //@line 438 "modules/graphics/opengl/Shader.cpp"
           if (!($190)) {
            $191 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
            __THREW__ = 0;
            invoke_vi($191|0,0); //@line 534 "./libraries/glad/gladfuncs.hpp"
            $192 = __THREW__; __THREW__ = 0;
            $193 = $192&1;
            if ($193) {
             break;
            }
           }
           HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
           STACKTOP = sp;return; //@line 577 "modules/graphics/opengl/Shader.cpp"
          }
          $165 = ($164|0)==($162|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
          if ($165) {
           STACKTOP = sp;return; //@line 577 "modules/graphics/opengl/Shader.cpp"
          }
          $166 = ((($162)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
          $167 = HEAP32[$166>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
          $168 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
          __THREW__ = 0;
          invoke_vi($168|0,($167|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
          $169 = __THREW__; __THREW__ = 0;
          $170 = $169&1;
          if (!($170)) {
           HEAP32[158581] = $162; //@line 409 "modules/graphics/opengl/Shader.cpp"
           $171 = ((($162)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
           $172 = ((($162)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $173 = HEAP32[$172>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $174 = HEAP32[$171>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $175 = ($173|0)==($174|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
           L36: do {
            if (!($175)) {
             $177 = $174;$203 = $173;$i$01$i$i25 = 0;
             while(1) {
              $176 = $177; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $178 = (($176) + ($i$01$i$i25<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $179 = HEAP32[$178>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
              $180 = ($179|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
              $$pre4$i$i26 = (($i$01$i$i25) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
              if ($180) {
               $184 = $203;$185 = $177;
              } else {
               __THREW__ = 0;
               invoke_viiii(4111,(634044|0),($179|0),($$pre4$i$i26|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
               $181 = __THREW__; __THREW__ = 0;
               $182 = $181&1;
               if ($182) {
                break;
               }
               $$pre$i$i28 = HEAP32[$172>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
               $$pre2$i$i29 = HEAP32[$171>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
               $184 = $$pre$i$i28;$185 = $$pre2$i$i29;
              }
              $183 = (($184) - ($185))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $186 = $183 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $187 = ($$pre4$i$i26>>>0)<($186>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
              if ($187) {
               $177 = $185;$203 = $184;$i$01$i$i25 = $$pre4$i$i26;
              } else {
               break L36;
              }
             }
             $194 = ___cxa_find_matching_catch_3(0|0)|0;
             $195 = tempRet0;
             $lpad$phi43$index = $194;
             ___clang_call_terminate($lpad$phi43$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
             // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
            }
           } while(0);
           __THREW__ = 0;
           invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
           $188 = __THREW__; __THREW__ = 0;
           $189 = $188&1;
           if (!($189)) {
            STACKTOP = sp;return; //@line 577 "modules/graphics/opengl/Shader.cpp"
           }
          }
         } else {
          $136 = HEAP32[158581]|0; //@line 406 "modules/graphics/opengl/Shader.cpp"
          $137 = ($136|0)==($$cast|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
          if (!($137)) {
           $138 = ((($$cast)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
           $139 = HEAP32[$138>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
           $140 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
           __THREW__ = 0;
           invoke_vi($140|0,($139|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
           $141 = __THREW__; __THREW__ = 0;
           $142 = $141&1;
           if ($142) {
            break;
           }
           HEAP32[158581] = $$cast; //@line 409 "modules/graphics/opengl/Shader.cpp"
          }
          $143 = ((($$cast)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
          $144 = ((($$cast)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $145 = HEAP32[$144>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $146 = HEAP32[$143>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $147 = ($145|0)==($146|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
          L52: do {
           if (!($147)) {
            $151 = $146;$202 = $145;$i$01$i15 = 0;
            while(1) {
             $150 = $151; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $152 = (($150) + ($i$01$i15<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $153 = HEAP32[$152>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
             $154 = ($153|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
             $$pre4$i16 = (($i$01$i15) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
             if ($154) {
              $158 = $202;$159 = $151;
             } else {
              __THREW__ = 0;
              invoke_viiii(4111,(634044|0),($153|0),($$pre4$i16|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
              $155 = __THREW__; __THREW__ = 0;
              $156 = $155&1;
              if ($156) {
               break;
              }
              $$pre$i18 = HEAP32[$144>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $$pre2$i19 = HEAP32[$143>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $158 = $$pre$i18;$159 = $$pre2$i19;
             }
             $157 = (($158) - ($159))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $160 = $157 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $161 = ($$pre4$i16>>>0)<($160>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
             if ($161) {
              $151 = $159;$202 = $158;$i$01$i15 = $$pre4$i16;
             } else {
              break L52;
             }
            }
            $196 = ___cxa_find_matching_catch_3(0|0)|0;
            $197 = tempRet0;
            $lpad$phi43$index = $196;
            ___clang_call_terminate($lpad$phi43$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
            // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
           }
          } while(0);
          __THREW__ = 0;
          invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
          $148 = __THREW__; __THREW__ = 0;
          $149 = $148&1;
          if (!($149)) {
           STACKTOP = sp;return; //@line 577 "modules/graphics/opengl/Shader.cpp"
          }
         }
        } while(0);
        $198 = ___cxa_find_matching_catch_3(0|0)|0;
        $199 = tempRet0;
        $lpad$phi43$index = $198;
        ___clang_call_terminate($lpad$phi43$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
        // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
       }
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 13) {
  $49 = ___cxa_find_matching_catch_2()|0;
  $50 = tempRet0;
  $eh$lpad$body$index2Z2D = $50;$eh$lpad$body$indexZ2D = $49;
 }
 $51 = ($5|0)==(0); //@line 53 "modules/graphics/opengl/Shader.cpp"
 do {
  if ($51) {
   $78 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
   $79 = ($78|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
   $80 = HEAP32[158581]|0; //@line 432 "modules/graphics/opengl/Shader.cpp"
   if ($79) {
    $106 = ($80|0)==(0|0); //@line 438 "modules/graphics/opengl/Shader.cpp"
    if (!($106)) {
     $107 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_vi($107|0,0); //@line 534 "./libraries/glad/gladfuncs.hpp"
     $108 = __THREW__; __THREW__ = 0;
     $109 = $108&1;
     if ($109) {
      break;
     }
    }
    HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
   $81 = ($80|0)==($78|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
   if ($81) {
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
   $82 = ((($78)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
   $83 = HEAP32[$82>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
   $84 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   invoke_vi($84|0,($83|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
   $85 = __THREW__; __THREW__ = 0;
   $86 = $85&1;
   if (!($86)) {
    HEAP32[158581] = $78; //@line 409 "modules/graphics/opengl/Shader.cpp"
    $87 = ((($78)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
    $88 = ((($78)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $89 = HEAP32[$88>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $90 = HEAP32[$87>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $91 = ($89|0)==($90|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
    L81: do {
     if (!($91)) {
      $201 = $89;$93 = $90;$i$01$i$i = 0;
      while(1) {
       $92 = $93; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $94 = (($92) + ($i$01$i$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $95 = HEAP32[$94>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
       $96 = ($95|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
       $$pre4$i$i = (($i$01$i$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
       if ($96) {
        $100 = $201;$101 = $93;
       } else {
        __THREW__ = 0;
        invoke_viiii(4111,(634044|0),($95|0),($$pre4$i$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
        $97 = __THREW__; __THREW__ = 0;
        $98 = $97&1;
        if ($98) {
         break;
        }
        $$pre$i$i = HEAP32[$88>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $$pre2$i$i = HEAP32[$87>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $100 = $$pre$i$i;$101 = $$pre2$i$i;
       }
       $99 = (($100) - ($101))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $102 = $99 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $103 = ($$pre4$i$i>>>0)<($102>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
       if ($103) {
        $201 = $100;$93 = $101;$i$01$i$i = $$pre4$i$i;
       } else {
        break L81;
       }
      }
      $110 = ___cxa_find_matching_catch_3(0|0)|0;
      $111 = tempRet0;
      $lpad$phi$index = $110;
      ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
      // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
     }
    } while(0);
    __THREW__ = 0;
    invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
    $104 = __THREW__; __THREW__ = 0;
    $105 = $104&1;
    if (!($105)) {
     ___resumeException($eh$lpad$body$indexZ2D|0);
     // unreachable;
    }
   }
  } else {
   $52 = HEAP32[158581]|0; //@line 406 "modules/graphics/opengl/Shader.cpp"
   $53 = ($52|0)==($$cast|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
   if (!($53)) {
    $54 = ((($$cast)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
    $55 = HEAP32[$54>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
    $56 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
    __THREW__ = 0;
    invoke_vi($56|0,($55|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
    $57 = __THREW__; __THREW__ = 0;
    $58 = $57&1;
    if ($58) {
     break;
    }
    HEAP32[158581] = $$cast; //@line 409 "modules/graphics/opengl/Shader.cpp"
   }
   $59 = ((($$cast)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
   $60 = ((($$cast)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $61 = HEAP32[$60>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $62 = HEAP32[$59>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $63 = ($61|0)==($62|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
   L97: do {
    if (!($63)) {
     $200 = $61;$67 = $62;$i$01$i = 0;
     while(1) {
      $66 = $67; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $68 = (($66) + ($i$01$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $69 = HEAP32[$68>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
      $70 = ($69|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
      $$pre4$i = (($i$01$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
      if ($70) {
       $74 = $200;$75 = $67;
      } else {
       __THREW__ = 0;
       invoke_viiii(4111,(634044|0),($69|0),($$pre4$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
       $71 = __THREW__; __THREW__ = 0;
       $72 = $71&1;
       if ($72) {
        break;
       }
       $$pre$i = HEAP32[$60>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $$pre2$i = HEAP32[$59>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $74 = $$pre$i;$75 = $$pre2$i;
      }
      $73 = (($74) - ($75))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $76 = $73 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $77 = ($$pre4$i>>>0)<($76>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      if ($77) {
       $200 = $74;$67 = $75;$i$01$i = $$pre4$i;
      } else {
       break L97;
      }
     }
     $112 = ___cxa_find_matching_catch_3(0|0)|0;
     $113 = tempRet0;
     $lpad$phi$index = $112;
     ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
     // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
    }
   } while(0);
   __THREW__ = 0;
   invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
   $64 = __THREW__; __THREW__ = 0;
   $65 = $64&1;
   if (!($65)) {
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
  }
 } while(0);
 $114 = ___cxa_find_matching_catch_3(0|0)|0;
 $115 = tempRet0;
 $lpad$phi$index = $114;
 ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
 // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
}
function __ZN4love8graphics6opengl6Shader14getTextureUnitERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($this,$name) {
 $this = $this|0;
 $name = $name|0;
 var $$0 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $texunit$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 862 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $0 = ((($this)) + 128|0); //@line 592 "modules/graphics/opengl/Shader.cpp"
 $1 = (__ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_15__tree_iteratorIS8_PNS_11__tree_nodeIS8_PvEEiEERKT_($0,$name)|0); //@line 1090 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $2 = ((($this)) + 132|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = ($1|0)==($2|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if (!($3)) {
  $4 = ((($1)) + 28|0); //@line 595 "modules/graphics/opengl/Shader.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 595 "modules/graphics/opengl/Shader.cpp"
  $$0 = $5;
  STACKTOP = sp;return ($$0|0); //@line 621 "modules/graphics/opengl/Shader.cpp"
 }
 $6 = HEAP32[158608]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = HEAP32[(634436)>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ($6|0)==($7|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $9 = $6; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 L5: do {
  if ($8) {
   $16 = $6;
   label = 6;
  } else {
   $11 = $6;
   while(1) {
    $10 = HEAP32[$11>>2]|0; //@line 865 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $12 = ($10|0)==(0); //@line 865 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    if ($12) {
     $16 = $11;
     label = 6;
     break L5;
    }
    $13 = ((($11)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $14 = ($13|0)==($7|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($14) {
     label = 8;
     break;
    } else {
     $11 = $13;
    }
   }
  }
 } while(0);
 if ((label|0) == 6) {
  $15 = ($16|0)==($7|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($15) {
   label = 8;
  } else {
   $17 = $16; //@line 600 "modules/graphics/opengl/Shader.cpp"
   $18 = (($17) - ($9))|0; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $$sink = $18;
  }
 }
 do {
  if ((label|0) == 8) {
   $19 = ((($this)) + 140|0); //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $20 = HEAP32[$19>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $21 = ((($this)) + 144|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $22 = HEAP32[$21>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $23 = ($20|0)==($22|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   $24 = $20; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   L14: do {
    if ($23) {
     $31 = $20;
     label = 11;
    } else {
     $26 = $20;
     while(1) {
      $25 = HEAP32[$26>>2]|0; //@line 865 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $27 = ($25|0)==(0); //@line 865 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if ($27) {
       $31 = $26;
       label = 11;
       break L14;
      }
      $28 = ((($26)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $29 = ($28|0)==($22|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($29) {
       break;
      } else {
       $26 = $28;
      }
     }
    }
   } while(0);
   if ((label|0) == 11) {
    $30 = ($31|0)==($22|0); //@line 1295 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if (!($30)) {
     $37 = $31; //@line 610 "modules/graphics/opengl/Shader.cpp"
     $38 = (($37) - ($24))|0; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $$sink = $38;
     break;
    }
   }
   $32 = (___cxa_allocate_exception(16)|0); //@line 613 "modules/graphics/opengl/Shader.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($32|0),(438875|0),($vararg_buffer|0)); //@line 613 "modules/graphics/opengl/Shader.cpp"
   $33 = __THREW__; __THREW__ = 0;
   $34 = $33&1;
   if ($34) {
    $35 = ___cxa_find_matching_catch_2()|0;
    $36 = tempRet0;
    ___cxa_free_exception(($32|0)); //@line 613 "modules/graphics/opengl/Shader.cpp"
    ___resumeException($35|0);
    // unreachable;
   } else {
    ___cxa_throw(($32|0),(96|0),(344|0)); //@line 613 "modules/graphics/opengl/Shader.cpp"
    // unreachable; //@line 613 "modules/graphics/opengl/Shader.cpp"
   }
  }
 } while(0);
 $39 = $$sink >> 2; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $texunit$0 = (($39) + 1)|0; //@line 605 "modules/graphics/opengl/Shader.cpp"
 $40 = (__ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiNS_4lessIS6_EENS4_INS_4pairIKS6_iEEEEEixERSA_($0,$name)|0); //@line 619 "modules/graphics/opengl/Shader.cpp"
 HEAP32[$40>>2] = $texunit$0; //@line 619 "modules/graphics/opengl/Shader.cpp"
 $$0 = $texunit$0;
 STACKTOP = sp;return ($$0|0); //@line 621 "modules/graphics/opengl/Shader.cpp"
}
function __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_15__tree_iteratorIS8_PNS_11__tree_nodeIS8_PvEEiEERKT_($this,$__v) {
 $this = $this|0;
 $__v = $__v|0;
 var $$$$i$i$i$i = 0, $$$$i$i$i$i$i = 0, $$0$ph$lcssa$i = 0, $$0$ph6$i = 0, $$01$ph5$i = 0, $$013$i = 0, $$013$i$lcssa = 0, $$sroa$0$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $2 = ($1|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if (!($2)) {
  $3 = HEAP8[$__v>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $4 = $3 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $5 = ($4<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $6 = ((($__v)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = HEAP32[$6>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $8 = $3&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = $8 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $10 = $5 ? $9 : $7; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $11 = ((($__v)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $12 = HEAP32[$11>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $13 = ((($__v)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $14 = $5 ? $13 : $12; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $$0$ph6$i = $0;$$01$ph5$i = $1;
  L3: while(1) {
   $$013$i = $$01$ph5$i;
   while(1) {
    $15 = ((($$013$i)) + 16|0); //@line 2102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $16 = HEAP8[$15>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $17 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $18 = ($17<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $19 = ((($$013$i)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $20 = HEAP32[$19>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $21 = $16&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $22 = $21 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $23 = $18 ? $22 : $20; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $24 = ((($$013$i)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $25 = HEAP32[$24>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $26 = ((($15)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $27 = $18 ? $26 : $25; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $28 = ($10>>>0)<($23>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $29 = $28 ? $10 : $23; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $30 = (_memcmp($27,$14,$29)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $31 = ($30|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $32 = ($23>>>0)<($10>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $$$$i$i$i$i$i = $32 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $33 = $31 ? $$$$i$i$i$i$i : $30; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $34 = ($33|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if (!($34)) {
     $$013$i$lcssa = $$013$i;
     break;
    }
    $37 = ((($$013$i)) + 4|0); //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $38 = HEAP32[$37>>2]|0; //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $39 = ($38|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($39) {
     $$0$ph$lcssa$i = $$0$ph6$i;
     break L3;
    } else {
     $$013$i = $38;
    }
   }
   $35 = HEAP32[$$013$i$lcssa>>2]|0; //@line 2105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $36 = ($35|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($36) {
    $$0$ph$lcssa$i = $$013$i$lcssa;
    break;
   } else {
    $$0$ph6$i = $$013$i$lcssa;$$01$ph5$i = $35;
   }
  }
  $40 = ($$0$ph$lcssa$i|0)==($0|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if (!($40)) {
   $41 = ((($$0$ph$lcssa$i)) + 16|0); //@line 650 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $42 = HEAP8[$41>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $43 = $42 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $44 = ($43<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $45 = ((($$0$ph$lcssa$i)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $46 = HEAP32[$45>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $47 = $42&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $48 = $47 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $49 = $44 ? $48 : $46; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $50 = ((($$0$ph$lcssa$i)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $51 = HEAP32[$50>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $52 = ((($41)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $53 = $44 ? $52 : $51; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $54 = ($49>>>0)<($10>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $55 = $54 ? $49 : $10; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $56 = (_memcmp($14,$53,$55)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $57 = ($56|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $58 = ($10>>>0)<($49>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$$$i$i$i$i = $58 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $59 = $57 ? $$$$i$i$i$i : $56; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $60 = ($59|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if (!($60)) {
    $$sroa$0$0 = $$0$ph$lcssa$i;
    return ($$sroa$0$0|0); //@line 2032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   }
  }
 }
 $$sroa$0$0 = $0;
 return ($$sroa$0$0|0); //@line 2032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
}
function __ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiNS_4lessIS6_EENS4_INS_4pairIKS6_iEEEEEixERSA_($this,$__k) {
 $this = $this|0;
 $__k = $__k|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__parent = 0, $__r$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__parent = sp;
 $0 = (__ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiNS_4lessIS6_EENS4_INS_4pairIKS6_iEEEEE16__find_equal_keyERPNS_16__tree_node_baseIPvEERSA_($this,$__parent,$__k)|0); //@line 1376 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $1 = HEAP32[$0>>2]|0; //@line 1377 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $2 = ($1|0)==(0|0); //@line 1378 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 if (!($2)) {
  $__r$0 = $1;
  $23 = ((($__r$0)) + 28|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  STACKTOP = sp;return ($23|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 $3 = (__Znwj(32)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $4 = ((($3)) + 16|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 __THREW__ = 0;
 invoke_vii(3351,($4|0),($__k|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZdlPv($3); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($7|0);
  // unreachable;
 }
 $9 = ((($3)) + 28|0); //@line 1366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$9>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $10 = HEAP32[$__parent>>2]|0; //@line 1381 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $11 = ((($3)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$11>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $12 = ((($3)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$12>>2] = $10; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$0>>2] = $3; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $13 = HEAP32[$this>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $14 = HEAP32[$13>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $15 = ($14|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($15) {
  $19 = $3;
 } else {
  $16 = $14; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$this>>2] = $16; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre$i = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $19 = $$pre$i;
 }
 $17 = ((($this)) + 4|0); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $18 = HEAP32[$17>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($18,$19); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $20 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $21 = HEAP32[$20>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $22 = (($21) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$20>>2] = $22; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $__r$0 = $3;
 $23 = ((($__r$0)) + 28|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 STACKTOP = sp;return ($23|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
}
function __ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiNS_4lessIS6_EENS4_INS_4pairIKS6_iEEEEE16__find_equal_keyERPNS_16__tree_node_baseIPvEERSA_($this,$__parent,$__k) {
 $this = $this|0;
 $__parent = $__parent|0;
 $__k = $__k|0;
 var $$$$i$i$i = 0, $$$$i$i$i1 = 0, $$0 = 0, $$lcssa = 0, $$lcssa19 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__nd$0 = 0, $__nd$0$lcssa = 0, $__nd$0$lcssa17 = 0, $__nd$0$lcssa18 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $2 = ($1|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 if ($2) {
  HEAP32[$__parent>>2] = $0; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $0;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 $3 = HEAP8[$__k>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = $3 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ($4<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = ((($__k)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = HEAP32[$6>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $3&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $8 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $5 ? $9 : $7; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = ((($__k)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = HEAP32[$11>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ((($__k)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = $5 ? $13 : $12; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $__nd$0 = $1;
 while(1) {
  $15 = ((($__nd$0)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $16 = HEAP8[$15>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $17 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $18 = ($17<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $19 = ((($__nd$0)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $20 = HEAP32[$19>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $21 = $16&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $22 = $21 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $23 = $18 ? $22 : $20; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $24 = ((($__nd$0)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $25 = HEAP32[$24>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $26 = ((($15)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $27 = $18 ? $26 : $25; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $28 = ($23>>>0)<($10>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $29 = $28 ? $23 : $10; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $30 = (_memcmp($14,$27,$29)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $31 = ($30|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $32 = ($10>>>0)<($23>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $$$$i$i$i1 = $32 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $33 = $31 ? $$$$i$i$i1 : $30; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $34 = ($33|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($34) {
   $35 = HEAP32[$__nd$0>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $36 = ($35|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   if ($36) {
    $$lcssa19 = $__nd$0;$__nd$0$lcssa18 = $__nd$0;
    label = 5;
    break;
   } else {
    $$sink = $35;
   }
  } else {
   $37 = $32 ? $10 : $23; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $38 = (_memcmp($27,$14,$37)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $39 = ($38|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$$$i$i$i = $28 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $40 = $39 ? $$$$i$i$i : $38; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $41 = ($40|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if (!($41)) {
    $__nd$0$lcssa = $__nd$0;
    label = 10;
    break;
   }
   $42 = ((($__nd$0)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $43 = HEAP32[$42>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $44 = ($43|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   if ($44) {
    $$lcssa = $42;$__nd$0$lcssa17 = $__nd$0;
    label = 9;
    break;
   } else {
    $$sink = $43;
   }
  }
  $__nd$0 = $$sink;
 }
 if ((label|0) == 5) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa18; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $$lcssa19;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 else if ((label|0) == 9) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa17; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $$lcssa;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 else if ((label|0) == 10) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $__parent;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 return (0)|0;
}
function __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ObjectEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE4findIS7_EENS_15__tree_iteratorISB_PNS_11__tree_nodeISB_PvEEiEERKT_($this,$__v) {
 $this = $this|0;
 $__v = $__v|0;
 var $$$$i$i$i$i = 0, $$$$i$i$i$i$i = 0, $$0$ph$lcssa$i = 0, $$0$ph6$i = 0, $$01$ph5$i = 0, $$013$i = 0, $$013$i$lcssa = 0, $$sroa$0$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $2 = ($1|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if (!($2)) {
  $3 = HEAP8[$__v>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $4 = $3 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $5 = ($4<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $6 = ((($__v)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = HEAP32[$6>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $8 = $3&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = $8 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $10 = $5 ? $9 : $7; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $11 = ((($__v)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $12 = HEAP32[$11>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $13 = ((($__v)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $14 = $5 ? $13 : $12; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $$0$ph6$i = $0;$$01$ph5$i = $1;
  L3: while(1) {
   $$013$i = $$01$ph5$i;
   while(1) {
    $15 = ((($$013$i)) + 16|0); //@line 2102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $16 = HEAP8[$15>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $17 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $18 = ($17<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $19 = ((($$013$i)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $20 = HEAP32[$19>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $21 = $16&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $22 = $21 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $23 = $18 ? $22 : $20; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $24 = ((($$013$i)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $25 = HEAP32[$24>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $26 = ((($15)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $27 = $18 ? $26 : $25; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $28 = ($10>>>0)<($23>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $29 = $28 ? $10 : $23; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $30 = (_memcmp($27,$14,$29)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $31 = ($30|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $32 = ($23>>>0)<($10>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $$$$i$i$i$i$i = $32 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $33 = $31 ? $$$$i$i$i$i$i : $30; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $34 = ($33|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if (!($34)) {
     $$013$i$lcssa = $$013$i;
     break;
    }
    $37 = ((($$013$i)) + 4|0); //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $38 = HEAP32[$37>>2]|0; //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $39 = ($38|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($39) {
     $$0$ph$lcssa$i = $$0$ph6$i;
     break L3;
    } else {
     $$013$i = $38;
    }
   }
   $35 = HEAP32[$$013$i$lcssa>>2]|0; //@line 2105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $36 = ($35|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($36) {
    $$0$ph$lcssa$i = $$013$i$lcssa;
    break;
   } else {
    $$0$ph6$i = $$013$i$lcssa;$$01$ph5$i = $35;
   }
  }
  $40 = ($$0$ph$lcssa$i|0)==($0|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if (!($40)) {
   $41 = ((($$0$ph$lcssa$i)) + 16|0); //@line 650 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $42 = HEAP8[$41>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $43 = $42 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $44 = ($43<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $45 = ((($$0$ph$lcssa$i)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $46 = HEAP32[$45>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $47 = $42&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $48 = $47 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $49 = $44 ? $48 : $46; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $50 = ((($$0$ph$lcssa$i)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $51 = HEAP32[$50>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $52 = ((($41)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $53 = $44 ? $52 : $51; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $54 = ($49>>>0)<($10>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $55 = $54 ? $49 : $10; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $56 = (_memcmp($14,$53,$55)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $57 = ($56|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $58 = ($10>>>0)<($49>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$$$i$i$i$i = $58 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $59 = $57 ? $$$$i$i$i$i : $56; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $60 = ($59|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if (!($60)) {
    $$sroa$0$0 = $$0$ph$lcssa$i;
    return ($$sroa$0$0|0); //@line 2032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   }
  }
 }
 $$sroa$0$0 = $0;
 return ($$sroa$0$0|0); //@line 2032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
}
function __ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ObjectENS_4lessIS6_EENS4_INS_4pairIKS6_S9_EEEEEixERSD_($this,$__k) {
 $this = $this|0;
 $__k = $__k|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__parent = 0, $__r$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__parent = sp;
 $0 = (__ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ObjectENS_4lessIS6_EENS4_INS_4pairIKS6_S9_EEEEE16__find_equal_keyERPNS_16__tree_node_baseIPvEERSD_($this,$__parent,$__k)|0); //@line 1376 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $1 = HEAP32[$0>>2]|0; //@line 1377 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $2 = ($1|0)==(0|0); //@line 1378 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 if (!($2)) {
  $__r$0 = $1;
  $23 = ((($__r$0)) + 28|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  STACKTOP = sp;return ($23|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 $3 = (__Znwj(32)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $4 = ((($3)) + 16|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 __THREW__ = 0;
 invoke_vii(3351,($4|0),($__k|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZdlPv($3); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($7|0);
  // unreachable;
 }
 $9 = ((($3)) + 28|0); //@line 1366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$9>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $10 = HEAP32[$__parent>>2]|0; //@line 1381 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $11 = ((($3)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$11>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $12 = ((($3)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$12>>2] = $10; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$0>>2] = $3; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $13 = HEAP32[$this>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $14 = HEAP32[$13>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $15 = ($14|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($15) {
  $19 = $3;
 } else {
  $16 = $14; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$this>>2] = $16; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre$i = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $19 = $$pre$i;
 }
 $17 = ((($this)) + 4|0); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $18 = HEAP32[$17>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($18,$19); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $20 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $21 = HEAP32[$20>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $22 = (($21) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$20>>2] = $22; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $__r$0 = $3;
 $23 = ((($__r$0)) + 28|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 STACKTOP = sp;return ($23|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
}
function __ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ObjectENS_4lessIS6_EENS4_INS_4pairIKS6_S9_EEEEE16__find_equal_keyERPNS_16__tree_node_baseIPvEERSD_($this,$__parent,$__k) {
 $this = $this|0;
 $__parent = $__parent|0;
 $__k = $__k|0;
 var $$$$i$i$i = 0, $$$$i$i$i1 = 0, $$0 = 0, $$lcssa = 0, $$lcssa19 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__nd$0 = 0, $__nd$0$lcssa = 0, $__nd$0$lcssa17 = 0, $__nd$0$lcssa18 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $2 = ($1|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 if ($2) {
  HEAP32[$__parent>>2] = $0; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $0;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 $3 = HEAP8[$__k>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = $3 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ($4<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = ((($__k)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = HEAP32[$6>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $3&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $8 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $5 ? $9 : $7; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = ((($__k)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = HEAP32[$11>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ((($__k)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = $5 ? $13 : $12; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $__nd$0 = $1;
 while(1) {
  $15 = ((($__nd$0)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $16 = HEAP8[$15>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $17 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $18 = ($17<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $19 = ((($__nd$0)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $20 = HEAP32[$19>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $21 = $16&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $22 = $21 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $23 = $18 ? $22 : $20; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $24 = ((($__nd$0)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $25 = HEAP32[$24>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $26 = ((($15)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $27 = $18 ? $26 : $25; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $28 = ($23>>>0)<($10>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $29 = $28 ? $23 : $10; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $30 = (_memcmp($14,$27,$29)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $31 = ($30|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $32 = ($10>>>0)<($23>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $$$$i$i$i1 = $32 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $33 = $31 ? $$$$i$i$i1 : $30; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $34 = ($33|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($34) {
   $35 = HEAP32[$__nd$0>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $36 = ($35|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   if ($36) {
    $$lcssa19 = $__nd$0;$__nd$0$lcssa18 = $__nd$0;
    label = 5;
    break;
   } else {
    $$sink = $35;
   }
  } else {
   $37 = $32 ? $10 : $23; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $38 = (_memcmp($27,$14,$37)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $39 = ($38|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$$$i$i$i = $28 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $40 = $39 ? $$$$i$i$i : $38; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $41 = ($40|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if (!($41)) {
    $__nd$0$lcssa = $__nd$0;
    label = 10;
    break;
   }
   $42 = ((($__nd$0)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $43 = HEAP32[$42>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $44 = ($43|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   if ($44) {
    $$lcssa = $42;$__nd$0$lcssa17 = $__nd$0;
    label = 9;
    break;
   } else {
    $$sink = $43;
   }
  }
  $__nd$0 = $$sink;
 }
 if ((label|0) == 5) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa18; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $$lcssa19;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 else if ((label|0) == 9) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa17; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $$lcssa;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 else if ((label|0) == 10) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $__parent;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl6Shader17getExternVariableERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEERiSC_($this,$name,$components,$count) {
 $this = $this|0;
 $name = $name|0;
 $components = $components|0;
 $count = $count|0;
 var $$0 = 0, $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 116|0); //@line 1090 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $1 = (__ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformEEENS_19__map_value_compareIS7_SD_NS_4lessIS7_EELb1EEENS5_ISD_EEE4findIS7_EENS_15__tree_iteratorISD_PNS_11__tree_nodeISD_PvEEiEERKT_($0,$name)|0); //@line 1090 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $2 = ((($this)) + 120|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = ($1|0)==($2|0); //@line 672 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($3) {
  HEAP32[$components>>2] = 0; //@line 629 "modules/graphics/opengl/Shader.cpp"
  HEAP32[$count>>2] = 0; //@line 630 "modules/graphics/opengl/Shader.cpp"
  $$0 = 4;
  return ($$0|0); //@line 638 "modules/graphics/opengl/Shader.cpp"
 }
 $4 = ((($1)) + 36|0); //@line 634 "modules/graphics/opengl/Shader.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 634 "modules/graphics/opengl/Shader.cpp"
 switch ($5|0) {
 case 35669: case 35666: case 35676: case 35673:  {
  $$0$i = 4;
  break;
 }
 case 35665: case 35668: case 35675: case 35672:  {
  $$0$i = 3;
  break;
 }
 case 35671: case 35674: case 35664: case 35667:  {
  $$0$i = 2;
  break;
 }
 default: {
  $$0$i = 1;
 }
 }
 HEAP32[$components>>2] = $$0$i; //@line 634 "modules/graphics/opengl/Shader.cpp"
 $6 = ((($1)) + 32|0); //@line 635 "modules/graphics/opengl/Shader.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 635 "modules/graphics/opengl/Shader.cpp"
 HEAP32[$count>>2] = $7; //@line 635 "modules/graphics/opengl/Shader.cpp"
 $8 = ((($1)) + 40|0); //@line 637 "modules/graphics/opengl/Shader.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 637 "modules/graphics/opengl/Shader.cpp"
 $$0 = $9;
 return ($$0|0); //@line 638 "modules/graphics/opengl/Shader.cpp"
}
function __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformEEENS_19__map_value_compareIS7_SD_NS_4lessIS7_EELb1EEENS5_ISD_EEE4findIS7_EENS_15__tree_iteratorISD_PNS_11__tree_nodeISD_PvEEiEERKT_($this,$__v) {
 $this = $this|0;
 $__v = $__v|0;
 var $$$$i$i$i$i = 0, $$$$i$i$i$i$i = 0, $$0$ph$lcssa$i = 0, $$0$ph6$i = 0, $$01$ph5$i = 0, $$013$i = 0, $$013$i$lcssa = 0, $$sroa$0$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $2 = ($1|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if (!($2)) {
  $3 = HEAP8[$__v>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $4 = $3 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $5 = ($4<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $6 = ((($__v)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = HEAP32[$6>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $8 = $3&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = $8 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $10 = $5 ? $9 : $7; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $11 = ((($__v)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $12 = HEAP32[$11>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $13 = ((($__v)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $14 = $5 ? $13 : $12; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $$0$ph6$i = $0;$$01$ph5$i = $1;
  L3: while(1) {
   $$013$i = $$01$ph5$i;
   while(1) {
    $15 = ((($$013$i)) + 16|0); //@line 2102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $16 = HEAP8[$15>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $17 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $18 = ($17<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $19 = ((($$013$i)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $20 = HEAP32[$19>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $21 = $16&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $22 = $21 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $23 = $18 ? $22 : $20; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $24 = ((($$013$i)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $25 = HEAP32[$24>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $26 = ((($15)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $27 = $18 ? $26 : $25; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $28 = ($10>>>0)<($23>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $29 = $28 ? $10 : $23; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $30 = (_memcmp($27,$14,$29)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $31 = ($30|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $32 = ($23>>>0)<($10>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $$$$i$i$i$i$i = $32 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $33 = $31 ? $$$$i$i$i$i$i : $30; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $34 = ($33|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if (!($34)) {
     $$013$i$lcssa = $$013$i;
     break;
    }
    $37 = ((($$013$i)) + 4|0); //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $38 = HEAP32[$37>>2]|0; //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $39 = ($38|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($39) {
     $$0$ph$lcssa$i = $$0$ph6$i;
     break L3;
    } else {
     $$013$i = $38;
    }
   }
   $35 = HEAP32[$$013$i$lcssa>>2]|0; //@line 2105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $36 = ($35|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($36) {
    $$0$ph$lcssa$i = $$013$i$lcssa;
    break;
   } else {
    $$0$ph6$i = $$013$i$lcssa;$$01$ph5$i = $35;
   }
  }
  $40 = ($$0$ph$lcssa$i|0)==($0|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if (!($40)) {
   $41 = ((($$0$ph$lcssa$i)) + 16|0); //@line 650 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $42 = HEAP8[$41>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $43 = $42 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $44 = ($43<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $45 = ((($$0$ph$lcssa$i)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $46 = HEAP32[$45>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $47 = $42&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $48 = $47 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $49 = $44 ? $48 : $46; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $50 = ((($$0$ph$lcssa$i)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $51 = HEAP32[$50>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $52 = ((($41)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $53 = $44 ? $52 : $51; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $54 = ($49>>>0)<($10>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $55 = $54 ? $49 : $10; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $56 = (_memcmp($14,$53,$55)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $57 = ($56|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $58 = ($10>>>0)<($49>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$$$i$i$i$i = $58 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $59 = $57 ? $$$$i$i$i$i : $56; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $60 = ($59|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if (!($60)) {
    $$sroa$0$0 = $$0$ph$lcssa$i;
    return ($$sroa$0$0|0); //@line 2032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   }
  }
 }
 $$sroa$0$0 = $0;
 return ($$sroa$0$0|0); //@line 2032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
}
function __ZN4love8graphics6opengl6Shader17getAttribLocationERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($this,$name) {
 $this = $this|0;
 $name = $name|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 104|0); //@line 642 "modules/graphics/opengl/Shader.cpp"
 $1 = (__ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE4findIS7_EENS_15__tree_iteratorIS8_PNS_11__tree_nodeIS8_PvEEiEERKT_($0,$name)|0); //@line 1090 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $2 = ((($this)) + 108|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = ($1|0)==($2|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($3) {
  $6 = ((($this)) + 48|0); //@line 646 "modules/graphics/opengl/Shader.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 646 "modules/graphics/opengl/Shader.cpp"
  $8 = HEAP8[$name>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = $8 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $10 = ($9<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $11 = ((($name)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $12 = HEAP32[$11>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $13 = ((($name)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $14 = $10 ? $13 : $12; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $15 = HEAP32[164909]|0; //@line 517 "./libraries/glad/gladfuncs.hpp"
  $16 = (FUNCTION_TABLE_iii[$15 & 8191]($7,$14)|0); //@line 517 "./libraries/glad/gladfuncs.hpp"
  $17 = (__ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEiNS_4lessIS6_EENS4_INS_4pairIKS6_iEEEEEixERSA_($0,$name)|0); //@line 648 "modules/graphics/opengl/Shader.cpp"
  HEAP32[$17>>2] = $16; //@line 648 "modules/graphics/opengl/Shader.cpp"
  $$0 = $16;
  return ($$0|0); //@line 650 "modules/graphics/opengl/Shader.cpp"
 } else {
  $4 = ((($1)) + 28|0); //@line 644 "modules/graphics/opengl/Shader.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 644 "modules/graphics/opengl/Shader.cpp"
  $$0 = $5;
  return ($$0|0); //@line 650 "modules/graphics/opengl/Shader.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl6Shader16setVideoTexturesEjjj($this,$ytexture,$cbtexture,$crtexture) {
 $this = $this|0;
 $ytexture = $ytexture|0;
 $cbtexture = $cbtexture|0;
 $crtexture = $crtexture|0;
 var $$0 = 0, $$01 = 0, $$cast = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i30 = 0, $$pre$i20 = 0, $$pre$phi62Z2D = 0, $$pre$phi63Z2D = 0, $$pre2$i = 0, $$pre2$i$i = 0, $$pre2$i$i31 = 0, $$pre2$i21 = 0, $$pre4$i = 0, $$pre4$i$i = 0, $$pre4$i$i28 = 0, $$pre4$i18 = 0, $$pre58 = 0, $$pre59 = 0;
 var $$pre60 = 0, $$pre61 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$01$i = 0;
 var $i$01$i$i = 0, $i$01$i$i27 = 0, $i$01$i17 = 0, $lpad$phi$index = 0, $lpad$phi45$index = 0, $lpad$phi52$index = 0, $lpad$phi52$index3 = 0, $or$cond = 0, $or$cond65 = 0, $or$cond66 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = HEAP32[158581]|0; //@line 46 "modules/graphics/opengl/Shader.cpp"
 $$cast = $1; //@line 406 "modules/graphics/opengl/Shader.cpp"
 $2 = ($$cast|0)==($this|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
 if (!($2)) {
  $3 = ((($this)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
  $5 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
  FUNCTION_TABLE_vi[$5 & 8191]($4); //@line 534 "./libraries/glad/gladfuncs.hpp"
  HEAP32[158581] = $this; //@line 409 "modules/graphics/opengl/Shader.cpp"
 }
 $6 = ((($this)) + 316|0); //@line 663 "modules/graphics/opengl/Shader.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 663 "modules/graphics/opengl/Shader.cpp"
 $8 = ($7|0)==(0); //@line 663 "modules/graphics/opengl/Shader.cpp"
 L4: do {
  if ($8) {
   $9 = ((($this)) + 76|0); //@line 666 "modules/graphics/opengl/Shader.cpp"
   $10 = HEAP32[$9>>2]|0; //@line 687 "modules/graphics/opengl/Shader.cpp"
   $11 = ((($this)) + 80|0); //@line 667 "modules/graphics/opengl/Shader.cpp"
   $12 = HEAP32[$11>>2]|0; //@line 687 "modules/graphics/opengl/Shader.cpp"
   $13 = ((($this)) + 84|0); //@line 668 "modules/graphics/opengl/Shader.cpp"
   $14 = HEAP32[$13>>2]|0; //@line 687 "modules/graphics/opengl/Shader.cpp"
   $15 = HEAP32[(634852)>>2]|0; //@line 94 "./common/StringMap.h"
   $16 = HEAP32[(634856)>>2]|0; //@line 94 "./common/StringMap.h"
   $17 = HEAP32[(634860)>>2]|0; //@line 94 "./common/StringMap.h"
   $18 = ((($this)) + 140|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $19 = ($10|0)<(0); //@line 678 "modules/graphics/opengl/Shader.cpp"
   $20 = ($15|0)==(0|0); //@line 678 "modules/graphics/opengl/Shader.cpp"
   $or$cond = $19 | $20; //@line 678 "modules/graphics/opengl/Shader.cpp"
   if ($or$cond) {
    label = 16;
   } else {
    $27 = (_strlen($15)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    __THREW__ = 0;
    invoke_viii(3460,($0|0),($15|0),($27|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $28 = __THREW__; __THREW__ = 0;
    $29 = $28&1;
    if ($29) {
     label = 7;
    } else {
     __THREW__ = 0;
     $30 = (invoke_iii(4130,($this|0),($0|0))|0); //@line 680 "modules/graphics/opengl/Shader.cpp"
     $31 = __THREW__; __THREW__ = 0;
     $32 = $31&1;
     if ($32) {
      label = 14;
     } else {
      HEAP32[$6>>2] = $30; //@line 680 "modules/graphics/opengl/Shader.cpp"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 680 "modules/graphics/opengl/Shader.cpp"
      $33 = HEAP32[$6>>2]|0; //@line 684 "modules/graphics/opengl/Shader.cpp"
      $34 = (($33) + -1)|0; //@line 684 "modules/graphics/opengl/Shader.cpp"
      $35 = HEAP32[$18>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $36 = (($35) + ($34<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $37 = HEAP32[$36>>2]|0; //@line 684 "modules/graphics/opengl/Shader.cpp"
      $38 = ($37|0)==(0); //@line 684 "modules/graphics/opengl/Shader.cpp"
      if ($38) {
       $39 = HEAP32[158608]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $40 = (($39) + ($34<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $41 = HEAP32[$40>>2]|0; //@line 685 "modules/graphics/opengl/Shader.cpp"
       $42 = (($41) + 1)|0; //@line 685 "modules/graphics/opengl/Shader.cpp"
       HEAP32[$40>>2] = $42; //@line 685 "modules/graphics/opengl/Shader.cpp"
       $$pre = HEAP32[$6>>2]|0; //@line 687 "modules/graphics/opengl/Shader.cpp"
       $46 = $$pre;
      } else {
       $46 = $33;
      }
      $45 = HEAP32[165188]|0; //@line 539 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_vii($45|0,($10|0),($46|0)); //@line 539 "./libraries/glad/gladfuncs.hpp"
      $47 = __THREW__; __THREW__ = 0;
      $48 = $47&1;
      if ($48) {
       label = 7;
      } else {
       label = 16;
      }
     }
    }
   }
   do {
    if ((label|0) == 16) {
     $49 = ($12|0)<(0); //@line 678 "modules/graphics/opengl/Shader.cpp"
     $50 = ($16|0)==(0|0); //@line 678 "modules/graphics/opengl/Shader.cpp"
     $or$cond65 = $49 | $50; //@line 678 "modules/graphics/opengl/Shader.cpp"
     if (!($or$cond65)) {
      $209 = (_strlen($16)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __THREW__ = 0;
      invoke_viii(3460,($0|0),($16|0),($209|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $210 = __THREW__; __THREW__ = 0;
      $211 = $210&1;
      if ($211) {
       label = 7;
       break;
      }
      __THREW__ = 0;
      $212 = (invoke_iii(4130,($this|0),($0|0))|0); //@line 680 "modules/graphics/opengl/Shader.cpp"
      $213 = __THREW__; __THREW__ = 0;
      $214 = $213&1;
      if ($214) {
       label = 14;
       break;
      }
      $215 = ((($this)) + 320|0); //@line 680 "modules/graphics/opengl/Shader.cpp"
      HEAP32[$215>>2] = $212; //@line 680 "modules/graphics/opengl/Shader.cpp"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 680 "modules/graphics/opengl/Shader.cpp"
      $216 = HEAP32[$215>>2]|0; //@line 684 "modules/graphics/opengl/Shader.cpp"
      $217 = (($216) + -1)|0; //@line 684 "modules/graphics/opengl/Shader.cpp"
      $218 = HEAP32[$18>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $219 = (($218) + ($217<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $220 = HEAP32[$219>>2]|0; //@line 684 "modules/graphics/opengl/Shader.cpp"
      $221 = ($220|0)==(0); //@line 684 "modules/graphics/opengl/Shader.cpp"
      if ($221) {
       $222 = HEAP32[158608]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $223 = (($222) + ($217<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $224 = HEAP32[$223>>2]|0; //@line 685 "modules/graphics/opengl/Shader.cpp"
       $225 = (($224) + 1)|0; //@line 685 "modules/graphics/opengl/Shader.cpp"
       HEAP32[$223>>2] = $225; //@line 685 "modules/graphics/opengl/Shader.cpp"
       $$pre58 = HEAP32[$215>>2]|0; //@line 687 "modules/graphics/opengl/Shader.cpp"
       $227 = $$pre58;
      } else {
       $227 = $216;
      }
      $226 = HEAP32[165188]|0; //@line 539 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_vii($226|0,($12|0),($227|0)); //@line 539 "./libraries/glad/gladfuncs.hpp"
      $228 = __THREW__; __THREW__ = 0;
      $229 = $228&1;
      if ($229) {
       label = 7;
       break;
      }
     }
     $230 = ($14|0)<(0); //@line 678 "modules/graphics/opengl/Shader.cpp"
     $231 = ($17|0)==(0|0); //@line 678 "modules/graphics/opengl/Shader.cpp"
     $or$cond66 = $230 | $231; //@line 678 "modules/graphics/opengl/Shader.cpp"
     if (!($or$cond66)) {
      $232 = (_strlen($17)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __THREW__ = 0;
      invoke_viii(3460,($0|0),($17|0),($232|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $233 = __THREW__; __THREW__ = 0;
      $234 = $233&1;
      if ($234) {
       label = 7;
       break;
      }
      __THREW__ = 0;
      $235 = (invoke_iii(4130,($this|0),($0|0))|0); //@line 680 "modules/graphics/opengl/Shader.cpp"
      $236 = __THREW__; __THREW__ = 0;
      $237 = $236&1;
      if ($237) {
       label = 14;
       break;
      }
      $238 = ((($this)) + 324|0); //@line 680 "modules/graphics/opengl/Shader.cpp"
      HEAP32[$238>>2] = $235; //@line 680 "modules/graphics/opengl/Shader.cpp"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 680 "modules/graphics/opengl/Shader.cpp"
      $239 = HEAP32[$238>>2]|0; //@line 684 "modules/graphics/opengl/Shader.cpp"
      $240 = (($239) + -1)|0; //@line 684 "modules/graphics/opengl/Shader.cpp"
      $241 = HEAP32[$18>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $242 = (($241) + ($240<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $243 = HEAP32[$242>>2]|0; //@line 684 "modules/graphics/opengl/Shader.cpp"
      $244 = ($243|0)==(0); //@line 684 "modules/graphics/opengl/Shader.cpp"
      if ($244) {
       $245 = HEAP32[158608]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $246 = (($245) + ($240<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $247 = HEAP32[$246>>2]|0; //@line 685 "modules/graphics/opengl/Shader.cpp"
       $248 = (($247) + 1)|0; //@line 685 "modules/graphics/opengl/Shader.cpp"
       HEAP32[$246>>2] = $248; //@line 685 "modules/graphics/opengl/Shader.cpp"
       $$pre59 = HEAP32[$238>>2]|0; //@line 687 "modules/graphics/opengl/Shader.cpp"
       $250 = $$pre59;
      } else {
       $250 = $239;
      }
      $249 = HEAP32[165188]|0; //@line 539 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_vii($249|0,($14|0),($250|0)); //@line 539 "./libraries/glad/gladfuncs.hpp"
      $251 = __THREW__; __THREW__ = 0;
      $252 = $251&1;
      if ($252) {
       label = 7;
       break;
      }
     }
     $$pre60 = HEAP32[$6>>2]|0; //@line 697 "modules/graphics/opengl/Shader.cpp"
     $51 = ($$pre60|0)==(0); //@line 697 "modules/graphics/opengl/Shader.cpp"
     if ($51) {
      $$pre$phi63Z2D = $18;
      label = 19;
      break L4;
     } else {
      $$pre$phi62Z2D = $18;$53 = $$pre60;
      label = 18;
      break L4;
     }
    }
   } while(0);
   if ((label|0) == 7) {
    $23 = ___cxa_find_matching_catch_2()|0;
    $24 = tempRet0;
    $lpad$phi52$index = $23;$lpad$phi52$index3 = $24;
    label = 9;
    break;
   }
   else if ((label|0) == 14) {
    $43 = ___cxa_find_matching_catch_2()|0;
    $44 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 680 "modules/graphics/opengl/Shader.cpp"
    $$0 = $44;$$01 = $43;
    break;
   }
  } else {
   $$pre61 = ((($this)) + 140|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$pre$phi62Z2D = $$pre61;$53 = $7;
   label = 18;
  }
 } while(0);
 if ((label|0) == 18) {
  $52 = (($53) + -1)|0; //@line 700 "modules/graphics/opengl/Shader.cpp"
  $54 = HEAP32[$$pre$phi62Z2D>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $55 = (($54) + ($52<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$55>>2] = $ytexture; //@line 700 "modules/graphics/opengl/Shader.cpp"
  $56 = HEAP32[$6>>2]|0; //@line 701 "modules/graphics/opengl/Shader.cpp"
  __THREW__ = 0;
  invoke_viiii(4111,(634044|0),($ytexture|0),($56|0),0); //@line 701 "modules/graphics/opengl/Shader.cpp"
  $57 = __THREW__; __THREW__ = 0;
  $58 = $57&1;
  if ($58) {
   label = 6;
  } else {
   $$pre$phi63Z2D = $$pre$phi62Z2D;
   label = 19;
  }
 }
 do {
  if ((label|0) == 19) {
   $59 = ((($this)) + 320|0); //@line 697 "modules/graphics/opengl/Shader.cpp"
   $60 = HEAP32[$59>>2]|0; //@line 697 "modules/graphics/opengl/Shader.cpp"
   $61 = ($60|0)==(0); //@line 697 "modules/graphics/opengl/Shader.cpp"
   if (!($61)) {
    $192 = (($60) + -1)|0; //@line 700 "modules/graphics/opengl/Shader.cpp"
    $193 = HEAP32[$$pre$phi63Z2D>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $194 = (($193) + ($192<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$194>>2] = $cbtexture; //@line 700 "modules/graphics/opengl/Shader.cpp"
    $195 = HEAP32[$59>>2]|0; //@line 701 "modules/graphics/opengl/Shader.cpp"
    __THREW__ = 0;
    invoke_viiii(4111,(634044|0),($cbtexture|0),($195|0),0); //@line 701 "modules/graphics/opengl/Shader.cpp"
    $196 = __THREW__; __THREW__ = 0;
    $197 = $196&1;
    if ($197) {
     label = 6;
     break;
    }
   }
   $198 = ((($this)) + 324|0); //@line 697 "modules/graphics/opengl/Shader.cpp"
   $199 = HEAP32[$198>>2]|0; //@line 697 "modules/graphics/opengl/Shader.cpp"
   $200 = ($199|0)==(0); //@line 697 "modules/graphics/opengl/Shader.cpp"
   if (!($200)) {
    $201 = (($199) + -1)|0; //@line 700 "modules/graphics/opengl/Shader.cpp"
    $202 = HEAP32[$$pre$phi63Z2D>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $203 = (($202) + ($201<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$203>>2] = $crtexture; //@line 700 "modules/graphics/opengl/Shader.cpp"
    $204 = HEAP32[$198>>2]|0; //@line 701 "modules/graphics/opengl/Shader.cpp"
    __THREW__ = 0;
    invoke_viiii(4111,(634044|0),($crtexture|0),($204|0),0); //@line 701 "modules/graphics/opengl/Shader.cpp"
    $205 = __THREW__; __THREW__ = 0;
    $206 = $205&1;
    if ($206) {
     label = 6;
     break;
    }
   }
   __THREW__ = 0;
   invoke_vii(4112,(634044|0),0); //@line 705 "modules/graphics/opengl/Shader.cpp"
   $207 = __THREW__; __THREW__ = 0;
   $208 = $207&1;
   if ($208) {
    $25 = ___cxa_find_matching_catch_2()|0;
    $26 = tempRet0;
    $lpad$phi52$index = $25;$lpad$phi52$index3 = $26;
    label = 9;
    break;
   }
   $62 = ($1|0)==(0); //@line 53 "modules/graphics/opengl/Shader.cpp"
   do {
    if ($62) {
     $89 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
     $90 = ($89|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
     $91 = HEAP32[158581]|0; //@line 432 "modules/graphics/opengl/Shader.cpp"
     if ($90) {
      $117 = ($91|0)==(0|0); //@line 438 "modules/graphics/opengl/Shader.cpp"
      if (!($117)) {
       $118 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
       __THREW__ = 0;
       invoke_vi($118|0,0); //@line 534 "./libraries/glad/gladfuncs.hpp"
       $119 = __THREW__; __THREW__ = 0;
       $120 = $119&1;
       if ($120) {
        break;
       }
      }
      HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
      STACKTOP = sp;return; //@line 706 "modules/graphics/opengl/Shader.cpp"
     }
     $92 = ($91|0)==($89|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
     if ($92) {
      STACKTOP = sp;return; //@line 706 "modules/graphics/opengl/Shader.cpp"
     }
     $93 = ((($89)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
     $94 = HEAP32[$93>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
     $95 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_vi($95|0,($94|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
     $96 = __THREW__; __THREW__ = 0;
     $97 = $96&1;
     if (!($97)) {
      HEAP32[158581] = $89; //@line 409 "modules/graphics/opengl/Shader.cpp"
      $98 = ((($89)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      $99 = ((($89)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $100 = HEAP32[$99>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $101 = HEAP32[$98>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $102 = ($100|0)==($101|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      L61: do {
       if (!($102)) {
        $104 = $101;$254 = $100;$i$01$i$i27 = 0;
        while(1) {
         $103 = $104; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $105 = (($103) + ($i$01$i$i27<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $106 = HEAP32[$105>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
         $107 = ($106|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
         $$pre4$i$i28 = (($i$01$i$i27) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
         if ($107) {
          $111 = $254;$112 = $104;
         } else {
          __THREW__ = 0;
          invoke_viiii(4111,(634044|0),($106|0),($$pre4$i$i28|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
          $108 = __THREW__; __THREW__ = 0;
          $109 = $108&1;
          if ($109) {
           break;
          }
          $$pre$i$i30 = HEAP32[$99>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $$pre2$i$i31 = HEAP32[$98>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $111 = $$pre$i$i30;$112 = $$pre2$i$i31;
         }
         $110 = (($111) - ($112))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $113 = $110 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $114 = ($$pre4$i$i28>>>0)<($113>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
         if ($114) {
          $104 = $112;$254 = $111;$i$01$i$i27 = $$pre4$i$i28;
         } else {
          break L61;
         }
        }
        $121 = ___cxa_find_matching_catch_3(0|0)|0;
        $122 = tempRet0;
        $lpad$phi45$index = $121;
        ___clang_call_terminate($lpad$phi45$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
        // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
       }
      } while(0);
      __THREW__ = 0;
      invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
      $115 = __THREW__; __THREW__ = 0;
      $116 = $115&1;
      if (!($116)) {
       STACKTOP = sp;return; //@line 706 "modules/graphics/opengl/Shader.cpp"
      }
     }
    } else {
     $63 = HEAP32[158581]|0; //@line 406 "modules/graphics/opengl/Shader.cpp"
     $64 = ($63|0)==($$cast|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
     if (!($64)) {
      $65 = ((($$cast)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
      $66 = HEAP32[$65>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
      $67 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_vi($67|0,($66|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
      $68 = __THREW__; __THREW__ = 0;
      $69 = $68&1;
      if ($69) {
       break;
      }
      HEAP32[158581] = $$cast; //@line 409 "modules/graphics/opengl/Shader.cpp"
     }
     $70 = ((($$cast)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
     $71 = ((($$cast)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $72 = HEAP32[$71>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $73 = HEAP32[$70>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $74 = ($72|0)==($73|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
     L77: do {
      if (!($74)) {
       $253 = $72;$78 = $73;$i$01$i17 = 0;
       while(1) {
        $77 = $78; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $79 = (($77) + ($i$01$i17<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $80 = HEAP32[$79>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
        $81 = ($80|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
        $$pre4$i18 = (($i$01$i17) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
        if ($81) {
         $85 = $253;$86 = $78;
        } else {
         __THREW__ = 0;
         invoke_viiii(4111,(634044|0),($80|0),($$pre4$i18|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
         $82 = __THREW__; __THREW__ = 0;
         $83 = $82&1;
         if ($83) {
          break;
         }
         $$pre$i20 = HEAP32[$71>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $$pre2$i21 = HEAP32[$70>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $85 = $$pre$i20;$86 = $$pre2$i21;
        }
        $84 = (($85) - ($86))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $87 = $84 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $88 = ($$pre4$i18>>>0)<($87>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
        if ($88) {
         $253 = $85;$78 = $86;$i$01$i17 = $$pre4$i18;
        } else {
         break L77;
        }
       }
       $123 = ___cxa_find_matching_catch_3(0|0)|0;
       $124 = tempRet0;
       $lpad$phi45$index = $123;
       ___clang_call_terminate($lpad$phi45$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
       // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
      }
     } while(0);
     __THREW__ = 0;
     invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
     $75 = __THREW__; __THREW__ = 0;
     $76 = $75&1;
     if (!($76)) {
      STACKTOP = sp;return; //@line 706 "modules/graphics/opengl/Shader.cpp"
     }
    }
   } while(0);
   $125 = ___cxa_find_matching_catch_3(0|0)|0;
   $126 = tempRet0;
   $lpad$phi45$index = $125;
   ___clang_call_terminate($lpad$phi45$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
   // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
  }
 } while(0);
 if ((label|0) == 6) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  $lpad$phi52$index = $21;$lpad$phi52$index3 = $22;
  label = 9;
 }
 if ((label|0) == 9) {
  $$0 = $lpad$phi52$index3;$$01 = $lpad$phi52$index;
 }
 $127 = ($1|0)==(0); //@line 53 "modules/graphics/opengl/Shader.cpp"
 do {
  if ($127) {
   $154 = HEAP32[158582]|0; //@line 430 "modules/graphics/opengl/Shader.cpp"
   $155 = ($154|0)==(0|0); //@line 430 "modules/graphics/opengl/Shader.cpp"
   $156 = HEAP32[158581]|0; //@line 432 "modules/graphics/opengl/Shader.cpp"
   if ($155) {
    $182 = ($156|0)==(0|0); //@line 438 "modules/graphics/opengl/Shader.cpp"
    if (!($182)) {
     $183 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_vi($183|0,0); //@line 534 "./libraries/glad/gladfuncs.hpp"
     $184 = __THREW__; __THREW__ = 0;
     $185 = $184&1;
     if ($185) {
      break;
     }
    }
    HEAP32[158581] = 0; //@line 441 "modules/graphics/opengl/Shader.cpp"
    ___resumeException($$01|0);
    // unreachable;
   }
   $157 = ($156|0)==($154|0); //@line 432 "modules/graphics/opengl/Shader.cpp"
   if ($157) {
    ___resumeException($$01|0);
    // unreachable;
   }
   $158 = ((($154)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
   $159 = HEAP32[$158>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
   $160 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
   __THREW__ = 0;
   invoke_vi($160|0,($159|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
   $161 = __THREW__; __THREW__ = 0;
   $162 = $161&1;
   if (!($162)) {
    HEAP32[158581] = $154; //@line 409 "modules/graphics/opengl/Shader.cpp"
    $163 = ((($154)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
    $164 = ((($154)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $165 = HEAP32[$164>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $166 = HEAP32[$163>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $167 = ($165|0)==($166|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
    L108: do {
     if (!($167)) {
      $169 = $166;$256 = $165;$i$01$i$i = 0;
      while(1) {
       $168 = $169; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $170 = (($168) + ($i$01$i$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $171 = HEAP32[$170>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
       $172 = ($171|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
       $$pre4$i$i = (($i$01$i$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
       if ($172) {
        $176 = $256;$177 = $169;
       } else {
        __THREW__ = 0;
        invoke_viiii(4111,(634044|0),($171|0),($$pre4$i$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
        $173 = __THREW__; __THREW__ = 0;
        $174 = $173&1;
        if ($174) {
         break;
        }
        $$pre$i$i = HEAP32[$164>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $$pre2$i$i = HEAP32[$163>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $176 = $$pre$i$i;$177 = $$pre2$i$i;
       }
       $175 = (($176) - ($177))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $178 = $175 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $179 = ($$pre4$i$i>>>0)<($178>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
       if ($179) {
        $169 = $177;$256 = $176;$i$01$i$i = $$pre4$i$i;
       } else {
        break L108;
       }
      }
      $186 = ___cxa_find_matching_catch_3(0|0)|0;
      $187 = tempRet0;
      $lpad$phi$index = $186;
      ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
      // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
     }
    } while(0);
    __THREW__ = 0;
    invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
    $180 = __THREW__; __THREW__ = 0;
    $181 = $180&1;
    if (!($181)) {
     ___resumeException($$01|0);
     // unreachable;
    }
   }
  } else {
   $128 = HEAP32[158581]|0; //@line 406 "modules/graphics/opengl/Shader.cpp"
   $129 = ($128|0)==($$cast|0); //@line 406 "modules/graphics/opengl/Shader.cpp"
   if (!($129)) {
    $130 = ((($$cast)) + 48|0); //@line 408 "modules/graphics/opengl/Shader.cpp"
    $131 = HEAP32[$130>>2]|0; //@line 408 "modules/graphics/opengl/Shader.cpp"
    $132 = HEAP32[165058]|0; //@line 534 "./libraries/glad/gladfuncs.hpp"
    __THREW__ = 0;
    invoke_vi($132|0,($131|0)); //@line 534 "./libraries/glad/gladfuncs.hpp"
    $133 = __THREW__; __THREW__ = 0;
    $134 = $133&1;
    if ($134) {
     break;
    }
    HEAP32[158581] = $$cast; //@line 409 "modules/graphics/opengl/Shader.cpp"
   }
   $135 = ((($$cast)) + 140|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
   $136 = ((($$cast)) + 144|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $137 = HEAP32[$136>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $138 = HEAP32[$135>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $139 = ($137|0)==($138|0); //@line 417 "modules/graphics/opengl/Shader.cpp"
   L124: do {
    if (!($139)) {
     $143 = $138;$255 = $137;$i$01$i = 0;
     while(1) {
      $142 = $143; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $144 = (($142) + ($i$01$i<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $145 = HEAP32[$144>>2]|0; //@line 419 "modules/graphics/opengl/Shader.cpp"
      $146 = ($145|0)==(0); //@line 419 "modules/graphics/opengl/Shader.cpp"
      $$pre4$i = (($i$01$i) + 1)|0; //@line 417 "modules/graphics/opengl/Shader.cpp"
      if ($146) {
       $150 = $255;$151 = $143;
      } else {
       __THREW__ = 0;
       invoke_viiii(4111,(634044|0),($145|0),($$pre4$i|0),0); //@line 420 "modules/graphics/opengl/Shader.cpp"
       $147 = __THREW__; __THREW__ = 0;
       $148 = $147&1;
       if ($148) {
        break;
       }
       $$pre$i = HEAP32[$136>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $$pre2$i = HEAP32[$135>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $150 = $$pre$i;$151 = $$pre2$i;
      }
      $149 = (($150) - ($151))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $152 = $149 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $153 = ($$pre4$i>>>0)<($152>>>0); //@line 417 "modules/graphics/opengl/Shader.cpp"
      if ($153) {
       $143 = $151;$255 = $150;$i$01$i = $$pre4$i;
      } else {
       break L124;
      }
     }
     $188 = ___cxa_find_matching_catch_3(0|0)|0;
     $189 = tempRet0;
     $lpad$phi$index = $188;
     ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
     // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
    }
   } while(0);
   __THREW__ = 0;
   invoke_vii(4112,(634044|0),0); //@line 424 "modules/graphics/opengl/Shader.cpp"
   $140 = __THREW__; __THREW__ = 0;
   $141 = $140&1;
   if (!($141)) {
    ___resumeException($$01|0);
    // unreachable;
   }
  }
 } while(0);
 $190 = ___cxa_find_matching_catch_3(0|0)|0;
 $191 = tempRet0;
 $lpad$phi$index = $190;
 ___clang_call_terminate($lpad$phi$index); //@line 54 "modules/graphics/opengl/Shader.cpp"
 // unreachable; //@line 54 "modules/graphics/opengl/Shader.cpp"
}
function __ZN4love8graphics6opengl6Shader11getConstantENS2_11UniformTypeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(4); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (634984 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 941 "modules/graphics/opengl/Shader.cpp"
}
function __ZN4love8graphics6opengl6Shader11getConstantEPKcRNS1_14VertexAttribIDE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = $9 & 7; //@line 72 "./common/StringMap.h"
  $11 = (((634500 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (634500 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(8); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((634500 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 946 "modules/graphics/opengl/Shader.cpp"
  return ($$$2$i|0); //@line 946 "modules/graphics/opengl/Shader.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 946 "modules/graphics/opengl/Shader.cpp"
  return ($$$2$i|0); //@line 946 "modules/graphics/opengl/Shader.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl6Shader11getConstantENS1_14VertexAttribIDERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(3); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (634596 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 951 "modules/graphics/opengl/Shader.cpp"
}
function __ZN4love8graphics6opengl11SpriteBatchD2Ev($this) {
 $this = $this|0;
 var $$01$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (25428); //@line 62 "modules/graphics/opengl/SpriteBatch.cpp"
 $0 = ((($this)) + 20|0); //@line 63 "modules/graphics/opengl/SpriteBatch.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 63 "modules/graphics/opengl/SpriteBatch.cpp"
 $2 = ($1|0)==(0|0); //@line 63 "modules/graphics/opengl/SpriteBatch.cpp"
 if (!($2)) {
  __ZdlPv($1); //@line 63 "modules/graphics/opengl/SpriteBatch.cpp"
 }
 $3 = ((($this)) + 24|0); //@line 64 "modules/graphics/opengl/SpriteBatch.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 64 "modules/graphics/opengl/SpriteBatch.cpp"
 $5 = ($4|0)==(0|0); //@line 64 "modules/graphics/opengl/SpriteBatch.cpp"
 if (!($5)) {
  $6 = HEAP32[$4>>2]|0; //@line 64 "modules/graphics/opengl/SpriteBatch.cpp"
  $7 = ((($6)) + 4|0); //@line 64 "modules/graphics/opengl/SpriteBatch.cpp"
  $8 = HEAP32[$7>>2]|0; //@line 64 "modules/graphics/opengl/SpriteBatch.cpp"
  FUNCTION_TABLE_vi[$8 & 8191]($4); //@line 64 "modules/graphics/opengl/SpriteBatch.cpp"
 }
 $9 = ((($this)) + 40|0); //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $10 = HEAP32[$9>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $11 = ($10|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L7: do {
  if (!($11)) {
   $$01$i$i$i = $10;
   while(1) {
    $12 = HEAP32[$$01$i$i$i>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $13 = ((($$01$i$i$i)) + 20|0); //@line 106 "./common/Object.h"
    $14 = HEAP32[$13>>2]|0; //@line 106 "./common/Object.h"
    $15 = ($14|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($15)) {
     $16 = HEAP32[$14>>2]|0; //@line 106 "./common/Object.h"
     $17 = ((($16)) + 12|0); //@line 106 "./common/Object.h"
     $18 = HEAP32[$17>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($18|0,($14|0)); //@line 106 "./common/Object.h"
     $19 = __THREW__; __THREW__ = 0;
     $20 = $19&1;
     if ($20) {
      break;
     }
    }
    $23 = ((($$01$i$i$i)) + 8|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($23); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
    __ZdlPv($$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $24 = ($12|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($24) {
     break L7;
    } else {
     $$01$i$i$i = $12;
    }
   }
   $21 = ___cxa_find_matching_catch_3(0|0)|0;
   $22 = tempRet0;
   ___clang_call_terminate($21); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 } while(0);
 $25 = ((($this)) + 32|0); //@line 2014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $26 = HEAP32[$25>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$25>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $27 = ($26|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($27)) {
  __ZdlPv($26); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $28 = ((($this)) + 28|0); //@line 65 "modules/graphics/opengl/SpriteBatch.cpp"
 __ZN4love8graphics6opengl11QuadIndicesD2Ev($28); //@line 65 "modules/graphics/opengl/SpriteBatch.cpp"
 $29 = ((($this)) + 8|0); //@line 106 "./common/Object.h"
 $30 = HEAP32[$29>>2]|0; //@line 106 "./common/Object.h"
 $31 = ($30|0)==(0|0); //@line 106 "./common/Object.h"
 if ($31) {
  return; //@line 65 "modules/graphics/opengl/SpriteBatch.cpp"
 }
 $32 = HEAP32[$30>>2]|0; //@line 106 "./common/Object.h"
 $33 = ((($32)) + 12|0); //@line 106 "./common/Object.h"
 $34 = HEAP32[$33>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($34|0,($30|0)); //@line 106 "./common/Object.h"
 $35 = __THREW__; __THREW__ = 0;
 $36 = $35&1;
 if ($36) {
  $37 = ___cxa_find_matching_catch_3(0|0)|0;
  $38 = tempRet0;
  ___clang_call_terminate($37); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  return; //@line 65 "modules/graphics/opengl/SpriteBatch.cpp"
 }
}
function __ZN4love8graphics6opengl11SpriteBatchD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl11SpriteBatchD2Ev($this); //@line 62 "modules/graphics/opengl/SpriteBatch.cpp"
 __ZdlPv($this); //@line 62 "modules/graphics/opengl/SpriteBatch.cpp"
 return; //@line 65 "modules/graphics/opengl/SpriteBatch.cpp"
}
function __ZN4love8graphics6opengl11SpriteBatch4drawEfffffffff($this,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $$0 = 0, $$01 = 0, $$lcssa32 = 0, $$lcssa34 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$013 = 0, $__begin$sroa$0$015 = 0, $enabledattribs$0 = 0, $enabledattribs$1$lcssa = 0, $enabledattribs$114 = 0, $enabledattribs$2 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $vararg_buffer = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = ((($this)) + 16|0); //@line 224 "modules/graphics/opengl/SpriteBatch.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 224 "modules/graphics/opengl/SpriteBatch.cpp"
 $3 = ($2|0)==(0); //@line 224 "modules/graphics/opengl/SpriteBatch.cpp"
 if ($3) {
  STACKTOP = sp;return; //@line 276 "modules/graphics/opengl/SpriteBatch.cpp"
 }
 __ZN4love8graphics6opengl6OpenGL13pushTransformEv(634044); //@line 125 "modules/graphics/opengl/OpenGL.h"
 __THREW__ = 0;
 $4 = (invoke_ii(4003,(634044|0))|0); //@line 135 "modules/graphics/opengl/OpenGL.h"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 L4: do {
  if ($6) {
   label = 13;
  } else {
   __THREW__ = 0;
   invoke_viddddddddd(4039,($0|0),(+$x),(+$y),(+$angle),(+$sx),(+$sy),(+$ox),(+$oy),(+$kx),(+$ky)); //@line 230 "modules/graphics/opengl/SpriteBatch.cpp"
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if ($8) {
    label = 13;
   } else {
    __THREW__ = 0;
    invoke_vii(4004,($4|0),($0|0)); //@line 230 "modules/graphics/opengl/SpriteBatch.cpp"
    $9 = __THREW__; __THREW__ = 0;
    $10 = $9&1;
    if ($10) {
     $43 = ___cxa_find_matching_catch_2()|0;
     $44 = tempRet0;
     __ZN4love7Matrix4D2Ev($0); //@line 230 "modules/graphics/opengl/SpriteBatch.cpp"
     $$0 = $43;$$01 = $44;
     break;
    }
    __ZN4love7Matrix4D2Ev($0); //@line 230 "modules/graphics/opengl/SpriteBatch.cpp"
    $11 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
    $12 = HEAP32[$11>>2]|0; //@line 117 "./common/Object.h"
    $13 = HEAP32[$12>>2]|0; //@line 232 "modules/graphics/opengl/SpriteBatch.cpp"
    $14 = ((($13)) + 52|0); //@line 232 "modules/graphics/opengl/SpriteBatch.cpp"
    $15 = HEAP32[$14>>2]|0; //@line 232 "modules/graphics/opengl/SpriteBatch.cpp"
    __THREW__ = 0;
    $16 = (invoke_ii($15|0,($12|0))|0); //@line 232 "modules/graphics/opengl/SpriteBatch.cpp"
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if ($18) {
     label = 13;
    } else {
     $19 = HEAP32[$16>>2]|0; //@line 232 "modules/graphics/opengl/SpriteBatch.cpp"
     __THREW__ = 0;
     invoke_vii(4005,(634044|0),($19|0)); //@line 232 "modules/graphics/opengl/SpriteBatch.cpp"
     $20 = __THREW__; __THREW__ = 0;
     $21 = $20&1;
     if ($21) {
      label = 13;
     } else {
      $22 = ((($this)) + 24|0); //@line 239 "modules/graphics/opengl/SpriteBatch.cpp"
      $23 = HEAP32[$22>>2]|0; //@line 239 "modules/graphics/opengl/SpriteBatch.cpp"
      __THREW__ = 0;
      invoke_vi(4087,($23|0)); //@line 188 "modules/graphics/opengl/GLBuffer.h"
      $24 = __THREW__; __THREW__ = 0;
      $25 = $24&1;
      if ($25) {
       label = 13;
      } else {
       $26 = HEAP32[$22>>2]|0; //@line 242 "modules/graphics/opengl/SpriteBatch.cpp"
       __THREW__ = 0;
       invoke_vi(4088,($26|0)); //@line 242 "modules/graphics/opengl/SpriteBatch.cpp"
       $27 = __THREW__; __THREW__ = 0;
       $28 = $27&1;
       do {
        if (!($28)) {
         $29 = ((($this)) + 20|0); //@line 245 "modules/graphics/opengl/SpriteBatch.cpp"
         $30 = HEAP32[$29>>2]|0; //@line 245 "modules/graphics/opengl/SpriteBatch.cpp"
         $31 = ($30|0)==(0|0); //@line 245 "modules/graphics/opengl/SpriteBatch.cpp"
         if ($31) {
          $enabledattribs$0 = 3;
         } else {
          $32 = HEAP32[$22>>2]|0; //@line 248 "modules/graphics/opengl/SpriteBatch.cpp"
          __THREW__ = 0;
          $33 = (invoke_iii(4090,($32|0),16)|0); //@line 248 "modules/graphics/opengl/SpriteBatch.cpp"
          $34 = __THREW__; __THREW__ = 0;
          $35 = $34&1;
          if ($35) {
           break;
          }
          $36 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
          __THREW__ = 0;
          invoke_viiiiii($36|0,2,4,5121,1,20,($33|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
          $37 = __THREW__; __THREW__ = 0;
          $38 = $37&1;
          if ($38) {
           break;
          } else {
           $enabledattribs$0 = 7;
          }
         }
         $51 = HEAP32[$22>>2]|0; //@line 251 "modules/graphics/opengl/SpriteBatch.cpp"
         __THREW__ = 0;
         $52 = (invoke_iii(4090,($51|0),0)|0); //@line 251 "modules/graphics/opengl/SpriteBatch.cpp"
         $53 = __THREW__; __THREW__ = 0;
         $54 = $53&1;
         if (!($54)) {
          $55 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
          __THREW__ = 0;
          invoke_viiiiii($55|0,0,2,5126,0,20,($52|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
          $56 = __THREW__; __THREW__ = 0;
          $57 = $56&1;
          if (!($57)) {
           $58 = HEAP32[$22>>2]|0; //@line 252 "modules/graphics/opengl/SpriteBatch.cpp"
           __THREW__ = 0;
           $59 = (invoke_iii(4090,($58|0),8)|0); //@line 252 "modules/graphics/opengl/SpriteBatch.cpp"
           $60 = __THREW__; __THREW__ = 0;
           $61 = $60&1;
           if (!($61)) {
            $62 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
            __THREW__ = 0;
            invoke_viiiiii($62|0,1,2,5126,0,20,($59|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
            $63 = __THREW__; __THREW__ = 0;
            $64 = $63&1;
            if (!($64)) {
             __THREW__ = 0;
             invoke_vi(4089,($23|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
             $65 = __THREW__; __THREW__ = 0;
             $66 = $65&1;
             if ($66) {
              $67 = ___cxa_find_matching_catch_3(0|0)|0;
              $68 = tempRet0;
              ___clang_call_terminate($67); //@line 196 "modules/graphics/opengl/GLBuffer.h"
              // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
             }
             $69 = ((($this)) + 40|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
             $__begin$sroa$0$013 = HEAP32[$69>>2]|0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
             $70 = ($__begin$sroa$0$013|0)==(0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
             L26: do {
              if ($70) {
               $enabledattribs$1$lcssa = $enabledattribs$0;
              } else {
               $71 = ((($this)) + 12|0); //@line 193 "modules/graphics/opengl/SpriteBatch.cpp"
               $__begin$sroa$0$015 = $__begin$sroa$0$013;$enabledattribs$114 = $enabledattribs$0;
               while(1) {
                $74 = $__begin$sroa$0$015; //@line 189 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
                $75 = ((($74)) + 8|0); //@line 623 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
                $76 = ((($74)) + 20|0); //@line 139 "./common/Object.h"
                $77 = HEAP32[$76>>2]|0; //@line 139 "./common/Object.h"
                __THREW__ = 0;
                $78 = (invoke_ii(4132,($77|0))|0); //@line 261 "modules/graphics/opengl/SpriteBatch.cpp"
                $79 = __THREW__; __THREW__ = 0;
                $80 = $79&1;
                if ($80) {
                 label = 12;
                 break;
                }
                $81 = HEAP32[$71>>2]|0; //@line 193 "modules/graphics/opengl/SpriteBatch.cpp"
                $82 = $81 << 2; //@line 261 "modules/graphics/opengl/SpriteBatch.cpp"
                $83 = ($78>>>0)<($82>>>0); //@line 261 "modules/graphics/opengl/SpriteBatch.cpp"
                if ($83) {
                 $$lcssa32 = $74;$$lcssa34 = $75;
                 break;
                }
                $97 = ((($74)) + 24|0); //@line 264 "modules/graphics/opengl/SpriteBatch.cpp"
                $98 = HEAP32[$97>>2]|0; //@line 264 "modules/graphics/opengl/SpriteBatch.cpp"
                __THREW__ = 0;
                $99 = (invoke_iiii(4133,($77|0),($98|0),($75|0))|0); //@line 264 "modules/graphics/opengl/SpriteBatch.cpp"
                $100 = __THREW__; __THREW__ = 0;
                $101 = $100&1;
                if ($101) {
                 label = 12;
                 break;
                }
                $102 = ($99|0)>(-1); //@line 266 "modules/graphics/opengl/SpriteBatch.cpp"
                $103 = 1 << $99; //@line 267 "modules/graphics/opengl/SpriteBatch.cpp"
                $104 = $102 ? $103 : 0; //@line 266 "modules/graphics/opengl/SpriteBatch.cpp"
                $enabledattribs$2 = $104 | $enabledattribs$114; //@line 266 "modules/graphics/opengl/SpriteBatch.cpp"
                $105 = $__begin$sroa$0$015; //@line 174 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
                $__begin$sroa$0$0 = HEAP32[$105>>2]|0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
                $106 = ($__begin$sroa$0$0|0)==(0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
                if ($106) {
                 $enabledattribs$1$lcssa = $enabledattribs$2;
                 break L26;
                } else {
                 $__begin$sroa$0$015 = $__begin$sroa$0$0;$enabledattribs$114 = $enabledattribs$2;
                }
               }
               if ((label|0) == 12) {
                $39 = ___cxa_find_matching_catch_2()|0;
                $40 = tempRet0;
                $lpad$phi$index = $39;$lpad$phi$index2 = $40;
                label = 14;
                break L4;
               }
               $84 = (___cxa_allocate_exception(16)|0); //@line 262 "modules/graphics/opengl/SpriteBatch.cpp"
               $85 = HEAP8[$$lcssa34>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $86 = $85 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $87 = ($86<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               if ($87) {
                $90 = ((($$lcssa34)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                $91 = $90;
               } else {
                $88 = ((($$lcssa32)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                $89 = HEAP32[$88>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                $91 = $89;
               }
               __THREW__ = 0;
               HEAP32[$vararg_buffer>>2] = $91; //@line 262 "modules/graphics/opengl/SpriteBatch.cpp"
               invoke_viii(3349,($84|0),(438956|0),($vararg_buffer|0)); //@line 262 "modules/graphics/opengl/SpriteBatch.cpp"
               $92 = __THREW__; __THREW__ = 0;
               $93 = $92&1;
               if ($93) {
                $95 = ___cxa_find_matching_catch_2()|0;
                $96 = tempRet0;
                ___cxa_free_exception(($84|0)); //@line 262 "modules/graphics/opengl/SpriteBatch.cpp"
                $$0 = $95;$$01 = $96;
                break L4;
               } else {
                __THREW__ = 0;
                invoke_viii(3350,($84|0),(96|0),(344|0)); //@line 262 "modules/graphics/opengl/SpriteBatch.cpp"
                $94 = __THREW__; __THREW__ = 0;
                label = 13;
                break L4;
               }
              }
             } while(0);
             __THREW__ = 0;
             invoke_vii(4006,(634044|0),($enabledattribs$1$lcssa|0)); //@line 270 "modules/graphics/opengl/SpriteBatch.cpp"
             $72 = __THREW__; __THREW__ = 0;
             $73 = $72&1;
             if ($73) {
              label = 13;
              break L4;
             }
             __THREW__ = 0;
             invoke_vi(4007,(634044|0)); //@line 272 "modules/graphics/opengl/SpriteBatch.cpp"
             $107 = __THREW__; __THREW__ = 0;
             $108 = $107&1;
             if ($108) {
              label = 13;
              break L4;
             }
             $109 = ((($this)) + 28|0); //@line 274 "modules/graphics/opengl/SpriteBatch.cpp"
             __THREW__ = 0;
             $110 = (invoke_ii(4134,($109|0))|0); //@line 274 "modules/graphics/opengl/SpriteBatch.cpp"
             $111 = __THREW__; __THREW__ = 0;
             $112 = $111&1;
             if ($112) {
              label = 13;
              break L4;
             }
             __THREW__ = 0;
             invoke_vi(4087,($110|0)); //@line 188 "modules/graphics/opengl/GLBuffer.h"
             $113 = __THREW__; __THREW__ = 0;
             $114 = $113&1;
             if ($114) {
              label = 13;
              break L4;
             }
             $115 = HEAP32[$1>>2]|0; //@line 275 "modules/graphics/opengl/SpriteBatch.cpp"
             __THREW__ = 0;
             $116 = (invoke_iii(4104,($109|0),($115|0))|0); //@line 275 "modules/graphics/opengl/SpriteBatch.cpp"
             $117 = __THREW__; __THREW__ = 0;
             $118 = $117&1;
             do {
              if (!($118)) {
               $119 = HEAP32[158426]|0; //@line 357 "modules/graphics/opengl/GLBuffer.h"
               __THREW__ = 0;
               $120 = (invoke_iii(4105,($109|0),($119|0))|0); //@line 357 "modules/graphics/opengl/GLBuffer.h"
               $121 = __THREW__; __THREW__ = 0;
               $122 = $121&1;
               if ($122) {
                break;
               }
               __THREW__ = 0;
               $123 = (invoke_iii(4135,($109|0),0)|0); //@line 275 "modules/graphics/opengl/SpriteBatch.cpp"
               $124 = __THREW__; __THREW__ = 0;
               $125 = $124&1;
               if ($125) {
                break;
               }
               __THREW__ = 0;
               invoke_viiiii(4091,(634044|0),4,($116|0),($120|0),($123|0)); //@line 275 "modules/graphics/opengl/SpriteBatch.cpp"
               $126 = __THREW__; __THREW__ = 0;
               $127 = $126&1;
               if ($127) {
                break;
               }
               __THREW__ = 0;
               invoke_vi(4089,($110|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
               $128 = __THREW__; __THREW__ = 0;
               $129 = $128&1;
               if ($129) {
                $130 = ___cxa_find_matching_catch_3(0|0)|0;
                $131 = tempRet0;
                ___clang_call_terminate($130); //@line 196 "modules/graphics/opengl/GLBuffer.h"
                // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
               }
               __THREW__ = 0;
               invoke_vi(4009,(634044|0)); //@line 130 "modules/graphics/opengl/OpenGL.h"
               $132 = __THREW__; __THREW__ = 0;
               $133 = $132&1;
               if ($133) {
                $134 = ___cxa_find_matching_catch_3(0|0)|0;
                $135 = tempRet0;
                ___clang_call_terminate($134); //@line 130 "modules/graphics/opengl/OpenGL.h"
                // unreachable; //@line 130 "modules/graphics/opengl/OpenGL.h"
               } else {
                STACKTOP = sp;return; //@line 276 "modules/graphics/opengl/SpriteBatch.cpp"
               }
              }
             } while(0);
             $136 = ___cxa_find_matching_catch_2()|0;
             $137 = tempRet0;
             __THREW__ = 0;
             invoke_vi(4089,($110|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
             $138 = __THREW__; __THREW__ = 0;
             $139 = $138&1;
             if (!($139)) {
              $$0 = $136;$$01 = $137;
              break L4;
             }
             $140 = ___cxa_find_matching_catch_3(0|0)|0;
             $141 = tempRet0;
             ___clang_call_terminate($140); //@line 196 "modules/graphics/opengl/GLBuffer.h"
             // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
            }
           }
          }
         }
        }
       } while(0);
       $45 = ___cxa_find_matching_catch_2()|0;
       $46 = tempRet0;
       __THREW__ = 0;
       invoke_vi(4089,($23|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
       $47 = __THREW__; __THREW__ = 0;
       $48 = $47&1;
       if ($48) {
        $49 = ___cxa_find_matching_catch_3(0|0)|0;
        $50 = tempRet0;
        ___clang_call_terminate($49); //@line 196 "modules/graphics/opengl/GLBuffer.h"
        // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
       } else {
        $$0 = $45;$$01 = $46;
       }
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 13) {
  $41 = ___cxa_find_matching_catch_2()|0;
  $42 = tempRet0;
  $lpad$phi$index = $41;$lpad$phi$index2 = $42;
  label = 14;
 }
 if ((label|0) == 14) {
  $$0 = $lpad$phi$index;$$01 = $lpad$phi$index2;
 }
 __THREW__ = 0;
 invoke_vi(4009,(634044|0)); //@line 130 "modules/graphics/opengl/OpenGL.h"
 $142 = __THREW__; __THREW__ = 0;
 $143 = $142&1;
 if ($143) {
  $144 = ___cxa_find_matching_catch_3(0|0)|0;
  $145 = tempRet0;
  ___clang_call_terminate($144); //@line 130 "modules/graphics/opengl/OpenGL.h"
  // unreachable; //@line 130 "modules/graphics/opengl/OpenGL.h"
 } else {
  ___resumeException($$0|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl11SpriteBatchC2EPNS0_7TextureEiNS1_4Mesh5UsageE($this,$texture,$size,$usage) {
 $this = $this|0;
 $texture = $texture|0;
 $size = $size|0;
 $usage = $usage|0;
 var $$0 = 0, $$01 = 0, $$01$i$i$i = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 __ZN4love6ObjectC2Ev($this); //@line 36 "./modules/graphics/Drawable.h"
 HEAP32[$this>>2] = (25428); //@line 51 "modules/graphics/opengl/SpriteBatch.cpp"
 $0 = ((($this)) + 8|0); //@line 93 "./common/Object.h"
 HEAP32[$0>>2] = $texture; //@line 93 "./common/Object.h"
 $1 = ($texture|0)==(0|0); //@line 95 "./common/Object.h"
 if (!($1)) {
  $2 = HEAP32[$texture>>2]|0; //@line 95 "./common/Object.h"
  $3 = ((($2)) + 8|0); //@line 95 "./common/Object.h"
  $4 = HEAP32[$3>>2]|0; //@line 95 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($4|0,($texture|0)); //@line 95 "./common/Object.h"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if ($6) {
   $21 = ___cxa_find_matching_catch_2()|0;
   $22 = tempRet0;
   $$2 = $21;$$23 = $22;
   ___resumeException($$2|0);
   // unreachable;
  }
 }
 $7 = ((($this)) + 12|0); //@line 46 "modules/graphics/opengl/SpriteBatch.cpp"
 HEAP32[$7>>2] = $size; //@line 46 "modules/graphics/opengl/SpriteBatch.cpp"
 $8 = ((($this)) + 16|0); //@line 47 "modules/graphics/opengl/SpriteBatch.cpp"
 HEAP32[$8>>2] = 0; //@line 47 "modules/graphics/opengl/SpriteBatch.cpp"
 $9 = ((($this)) + 20|0); //@line 48 "modules/graphics/opengl/SpriteBatch.cpp"
 HEAP32[$9>>2] = 0; //@line 48 "modules/graphics/opengl/SpriteBatch.cpp"
 $10 = ((($this)) + 24|0); //@line 49 "modules/graphics/opengl/SpriteBatch.cpp"
 HEAP32[$10>>2] = 0; //@line 49 "modules/graphics/opengl/SpriteBatch.cpp"
 $11 = ((($this)) + 28|0); //@line 50 "modules/graphics/opengl/SpriteBatch.cpp"
 __THREW__ = 0;
 invoke_vii(4020,($11|0),($size|0)); //@line 50 "modules/graphics/opengl/SpriteBatch.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if ($13) {
  $23 = ___cxa_find_matching_catch_2()|0;
  $24 = tempRet0;
  $$1 = $23;$$12 = $24;
 } else {
  $14 = ((($this)) + 32|0); //@line 44 "modules/graphics/opengl/SpriteBatch.cpp"
  $15 = ((($this)) + 48|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  ;HEAP32[$14>>2]=0|0;HEAP32[$14+4>>2]=0|0;HEAP32[$14+8>>2]=0|0;HEAP32[$14+12>>2]=0|0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAPF32[$15>>2] = 1.0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $16 = ($size|0)<(1); //@line 52 "modules/graphics/opengl/SpriteBatch.cpp"
  do {
   if ($16) {
    $17 = (___cxa_allocate_exception(16)|0); //@line 53 "modules/graphics/opengl/SpriteBatch.cpp"
    __THREW__ = 0;
    invoke_viii(3349,($17|0),(439031|0),($vararg_buffer|0)); //@line 53 "modules/graphics/opengl/SpriteBatch.cpp"
    $18 = __THREW__; __THREW__ = 0;
    $19 = $18&1;
    if ($19) {
     $25 = ___cxa_find_matching_catch_2()|0;
     $26 = tempRet0;
     ___cxa_free_exception(($17|0)); //@line 53 "modules/graphics/opengl/SpriteBatch.cpp"
     $$0 = $25;$$01 = $26;
     break;
    } else {
     __THREW__ = 0;
     invoke_viii(3350,($17|0),(96|0),(344|0)); //@line 53 "modules/graphics/opengl/SpriteBatch.cpp"
     $20 = __THREW__; __THREW__ = 0;
     label = 10;
     break;
    }
   } else {
    __THREW__ = 0;
    $29 = (invoke_ii(4136,($usage|0))|0); //@line 55 "modules/graphics/opengl/SpriteBatch.cpp"
    $30 = __THREW__; __THREW__ = 0;
    $31 = $30&1;
    if ($31) {
     label = 10;
    } else {
     __THREW__ = 0;
     $32 = (invoke_ii(3348,40)|0); //@line 58 "modules/graphics/opengl/SpriteBatch.cpp"
     $33 = __THREW__; __THREW__ = 0;
     $34 = $33&1;
     if ($34) {
      label = 10;
     } else {
      $35 = ($size*80)|0; //@line 56 "modules/graphics/opengl/SpriteBatch.cpp"
      __THREW__ = 0;
      invoke_viiiiii(4042,($32|0),($35|0),(0|0),34962,($29|0),1); //@line 58 "modules/graphics/opengl/SpriteBatch.cpp"
      $36 = __THREW__; __THREW__ = 0;
      $37 = $36&1;
      if ($37) {
       $38 = ___cxa_find_matching_catch_2()|0;
       $39 = tempRet0;
       __ZdlPv($32); //@line 58 "modules/graphics/opengl/SpriteBatch.cpp"
       $$0 = $38;$$01 = $39;
       break;
      } else {
       HEAP32[$10>>2] = $32; //@line 58 "modules/graphics/opengl/SpriteBatch.cpp"
       STACKTOP = sp;return; //@line 59 "modules/graphics/opengl/SpriteBatch.cpp"
      }
     }
    }
   }
  } while(0);
  if ((label|0) == 10) {
   $27 = ___cxa_find_matching_catch_2()|0;
   $28 = tempRet0;
   $$0 = $27;$$01 = $28;
  }
  $40 = ((($this)) + 40|0); //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $41 = HEAP32[$40>>2]|0; //@line 1283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $42 = ($41|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  L23: do {
   if (!($42)) {
    $$01$i$i$i = $41;
    while(1) {
     $43 = HEAP32[$$01$i$i$i>>2]|0; //@line 1327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $44 = ((($$01$i$i$i)) + 20|0); //@line 106 "./common/Object.h"
     $45 = HEAP32[$44>>2]|0; //@line 106 "./common/Object.h"
     $46 = ($45|0)==(0|0); //@line 106 "./common/Object.h"
     if (!($46)) {
      $47 = HEAP32[$45>>2]|0; //@line 106 "./common/Object.h"
      $48 = ((($47)) + 12|0); //@line 106 "./common/Object.h"
      $49 = HEAP32[$48>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($49|0,($45|0)); //@line 106 "./common/Object.h"
      $50 = __THREW__; __THREW__ = 0;
      $51 = $50&1;
      if ($51) {
       break;
      }
     }
     $54 = ((($$01$i$i$i)) + 8|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($54); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
     __ZdlPv($$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $55 = ($43|0)==(0|0); //@line 1325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     if ($55) {
      break L23;
     } else {
      $$01$i$i$i = $43;
     }
    }
    $52 = ___cxa_find_matching_catch_3(0|0)|0;
    $53 = tempRet0;
    ___clang_call_terminate($52); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
  } while(0);
  $56 = HEAP32[$14>>2]|0; //@line 2836 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$14>>2] = 0; //@line 2837 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $57 = ($56|0)==(0|0); //@line 2838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  if (!($57)) {
   __ZdlPv($56); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  __ZN4love8graphics6opengl11QuadIndicesD2Ev($11); //@line 59 "modules/graphics/opengl/SpriteBatch.cpp"
  $$1 = $$0;$$12 = $$01;
 }
 $58 = HEAP32[$0>>2]|0; //@line 106 "./common/Object.h"
 $59 = ($58|0)==(0|0); //@line 106 "./common/Object.h"
 if ($59) {
  $$2 = $$1;$$23 = $$12;
  ___resumeException($$2|0);
  // unreachable;
 }
 $60 = HEAP32[$58>>2]|0; //@line 106 "./common/Object.h"
 $61 = ((($60)) + 12|0); //@line 106 "./common/Object.h"
 $62 = HEAP32[$61>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($62|0,($58|0)); //@line 106 "./common/Object.h"
 $63 = __THREW__; __THREW__ = 0;
 $64 = $63&1;
 if ($64) {
  $65 = ___cxa_find_matching_catch_3(0|0)|0;
  $66 = tempRet0;
  ___clang_call_terminate($65); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  $$2 = $$1;$$23 = $$12;
  ___resumeException($$2|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl11SpriteBatch3addEfffffffffi($this,$x,$y,$a,$sx,$sy,$ox,$oy,$kx,$ky,$index) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $a = +$a;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 $index = $index|0;
 var $$0 = 0, $$1 = 0, $$old1 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $t = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $t = sp;
 $0 = ($index|0)==(-1); //@line 70 "modules/graphics/opengl/SpriteBatch.cpp"
 do {
  if ($0) {
   $1 = ((($this)) + 16|0); //@line 70 "modules/graphics/opengl/SpriteBatch.cpp"
   $2 = HEAP32[$1>>2]|0; //@line 70 "modules/graphics/opengl/SpriteBatch.cpp"
   $3 = ((($this)) + 12|0); //@line 70 "modules/graphics/opengl/SpriteBatch.cpp"
   $4 = HEAP32[$3>>2]|0; //@line 70 "modules/graphics/opengl/SpriteBatch.cpp"
   $5 = ($2|0)<($4|0); //@line 70 "modules/graphics/opengl/SpriteBatch.cpp"
   if ($5) {
    $7 = $4;
   } else {
    $$1 = -1;
    STACKTOP = sp;return ($$1|0); //@line 82 "modules/graphics/opengl/SpriteBatch.cpp"
   }
  } else {
   $$old1 = ($index|0)<(-1); //@line 70 "modules/graphics/opengl/SpriteBatch.cpp"
   if ($$old1) {
    $$1 = -1;
    STACKTOP = sp;return ($$1|0); //@line 82 "modules/graphics/opengl/SpriteBatch.cpp"
   } else {
    $$phi$trans$insert = ((($this)) + 12|0); //@line 70 "modules/graphics/opengl/SpriteBatch.cpp"
    $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 70 "modules/graphics/opengl/SpriteBatch.cpp"
    $7 = $$pre;
    break;
   }
  }
 } while(0);
 $6 = ($7|0)>($index|0); //@line 70 "modules/graphics/opengl/SpriteBatch.cpp"
 if (!($6)) {
  $$1 = -1;
  STACKTOP = sp;return ($$1|0); //@line 82 "modules/graphics/opengl/SpriteBatch.cpp"
 }
 __ZN4love7Matrix3C2Efffffffff($t,$x,$y,$a,$sx,$sy,$ox,$oy,$kx,$ky); //@line 73 "modules/graphics/opengl/SpriteBatch.cpp"
 $8 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $9 = HEAP32[$8>>2]|0; //@line 117 "./common/Object.h"
 $10 = HEAP32[$9>>2]|0; //@line 75 "modules/graphics/opengl/SpriteBatch.cpp"
 $11 = ((($10)) + 48|0); //@line 75 "modules/graphics/opengl/SpriteBatch.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 75 "modules/graphics/opengl/SpriteBatch.cpp"
 __THREW__ = 0;
 $13 = (invoke_ii($12|0,($9|0))|0); //@line 75 "modules/graphics/opengl/SpriteBatch.cpp"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if ($15) {
  $24 = ___cxa_find_matching_catch_2()|0;
  $25 = tempRet0;
  __ZN4love7Matrix3D2Ev($t); //@line 82 "modules/graphics/opengl/SpriteBatch.cpp"
  ___resumeException($24|0);
  // unreachable;
 }
 if ($0) {
  $16 = ((($this)) + 16|0); //@line 75 "modules/graphics/opengl/SpriteBatch.cpp"
  $17 = HEAP32[$16>>2]|0; //@line 75 "modules/graphics/opengl/SpriteBatch.cpp"
  $18 = $17;
 } else {
  $18 = $index;
 }
 __THREW__ = 0;
 invoke_viiii(4137,($this|0),($13|0),($t|0),($18|0)); //@line 75 "modules/graphics/opengl/SpriteBatch.cpp"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $24 = ___cxa_find_matching_catch_2()|0;
  $25 = tempRet0;
  __ZN4love7Matrix3D2Ev($t); //@line 82 "modules/graphics/opengl/SpriteBatch.cpp"
  ___resumeException($24|0);
  // unreachable;
 }
 if ($0) {
  $21 = ((($this)) + 16|0); //@line 79 "modules/graphics/opengl/SpriteBatch.cpp"
  $22 = HEAP32[$21>>2]|0; //@line 79 "modules/graphics/opengl/SpriteBatch.cpp"
  $23 = (($22) + 1)|0; //@line 79 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP32[$21>>2] = $23; //@line 79 "modules/graphics/opengl/SpriteBatch.cpp"
  $$0 = $22;
 } else {
  $$0 = $index;
 }
 __ZN4love7Matrix3D2Ev($t); //@line 82 "modules/graphics/opengl/SpriteBatch.cpp"
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0); //@line 82 "modules/graphics/opengl/SpriteBatch.cpp"
}
function __ZN4love8graphics6opengl11SpriteBatch4addvEPKNS_6VertexERKNS_7Matrix3Ei($this,$v,$m,$index) {
 $this = $this|0;
 $v = $v|0;
 $m = $m|0;
 $index = $index|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0;
 var $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0;
 var $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0.0, $54 = 0.0;
 var $55 = 0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $sprite = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $sprite = sp;
 ;HEAP32[$sprite>>2]=HEAP32[$v>>2]|0;HEAP32[$sprite+4>>2]=HEAP32[$v+4>>2]|0;HEAP32[$sprite+8>>2]=HEAP32[$v+8>>2]|0;HEAP32[$sprite+12>>2]=HEAP32[$v+12>>2]|0;HEAP32[$sprite+16>>2]=HEAP32[$v+16>>2]|0; //@line 281 "modules/graphics/opengl/SpriteBatch.cpp"
 $0 = ((($sprite)) + 20|0); //@line 281 "modules/graphics/opengl/SpriteBatch.cpp"
 $1 = ((($v)) + 20|0); //@line 281 "modules/graphics/opengl/SpriteBatch.cpp"
 ;HEAP32[$0>>2]=HEAP32[$1>>2]|0;HEAP32[$0+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$1+8>>2]|0;HEAP32[$0+12>>2]=HEAP32[$1+12>>2]|0;HEAP32[$0+16>>2]=HEAP32[$1+16>>2]|0; //@line 281 "modules/graphics/opengl/SpriteBatch.cpp"
 $2 = ((($sprite)) + 40|0); //@line 281 "modules/graphics/opengl/SpriteBatch.cpp"
 $3 = ((($v)) + 40|0); //@line 281 "modules/graphics/opengl/SpriteBatch.cpp"
 ;HEAP32[$2>>2]=HEAP32[$3>>2]|0;HEAP32[$2+4>>2]=HEAP32[$3+4>>2]|0;HEAP32[$2+8>>2]=HEAP32[$3+8>>2]|0;HEAP32[$2+12>>2]=HEAP32[$3+12>>2]|0;HEAP32[$2+16>>2]=HEAP32[$3+16>>2]|0; //@line 281 "modules/graphics/opengl/SpriteBatch.cpp"
 $4 = ((($sprite)) + 60|0); //@line 281 "modules/graphics/opengl/SpriteBatch.cpp"
 $5 = ((($v)) + 60|0); //@line 281 "modules/graphics/opengl/SpriteBatch.cpp"
 ;HEAP32[$4>>2]=HEAP32[$5>>2]|0;HEAP32[$4+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$4+8>>2]=HEAP32[$5+8>>2]|0;HEAP32[$4+12>>2]=HEAP32[$5+12>>2]|0;HEAP32[$4+16>>2]=HEAP32[$5+16>>2]|0; //@line 281 "modules/graphics/opengl/SpriteBatch.cpp"
 $6 = ((($m)) + 12|0); //@line 280 "./common/Matrix.h"
 $7 = ((($m)) + 24|0); //@line 280 "./common/Matrix.h"
 $8 = ((($m)) + 4|0); //@line 281 "./common/Matrix.h"
 $9 = ((($m)) + 16|0); //@line 281 "./common/Matrix.h"
 $10 = ((($m)) + 28|0); //@line 281 "./common/Matrix.h"
 $11 = +HEAPF32[$m>>2]; //@line 280 "./common/Matrix.h"
 $12 = +HEAPF32[$6>>2]; //@line 280 "./common/Matrix.h"
 $13 = +HEAPF32[$7>>2]; //@line 280 "./common/Matrix.h"
 $14 = +HEAPF32[$8>>2]; //@line 281 "./common/Matrix.h"
 $15 = +HEAPF32[$9>>2]; //@line 281 "./common/Matrix.h"
 $16 = +HEAPF32[$10>>2]; //@line 281 "./common/Matrix.h"
 $17 = +HEAPF32[$sprite>>2]; //@line 280 "./common/Matrix.h"
 $18 = $11 * $17; //@line 280 "./common/Matrix.h"
 $19 = ((($sprite)) + 4|0); //@line 280 "./common/Matrix.h"
 $20 = +HEAPF32[$19>>2]; //@line 280 "./common/Matrix.h"
 $21 = $12 * $20; //@line 280 "./common/Matrix.h"
 $22 = $18 + $21; //@line 280 "./common/Matrix.h"
 $23 = $13 + $22; //@line 280 "./common/Matrix.h"
 $24 = $17 * $14; //@line 281 "./common/Matrix.h"
 $25 = $20 * $15; //@line 281 "./common/Matrix.h"
 $26 = $24 + $25; //@line 281 "./common/Matrix.h"
 $27 = $16 + $26; //@line 281 "./common/Matrix.h"
 HEAPF32[$sprite>>2] = $23; //@line 283 "./common/Matrix.h"
 HEAPF32[$19>>2] = $27; //@line 284 "./common/Matrix.h"
 $28 = ((($sprite)) + 20|0); //@line 280 "./common/Matrix.h"
 $29 = +HEAPF32[$28>>2]; //@line 280 "./common/Matrix.h"
 $30 = $11 * $29; //@line 280 "./common/Matrix.h"
 $31 = ((($sprite)) + 24|0); //@line 280 "./common/Matrix.h"
 $32 = +HEAPF32[$31>>2]; //@line 280 "./common/Matrix.h"
 $33 = $12 * $32; //@line 280 "./common/Matrix.h"
 $34 = $30 + $33; //@line 280 "./common/Matrix.h"
 $35 = $13 + $34; //@line 280 "./common/Matrix.h"
 $36 = $29 * $14; //@line 281 "./common/Matrix.h"
 $37 = $32 * $15; //@line 281 "./common/Matrix.h"
 $38 = $36 + $37; //@line 281 "./common/Matrix.h"
 $39 = $16 + $38; //@line 281 "./common/Matrix.h"
 HEAPF32[$28>>2] = $35; //@line 283 "./common/Matrix.h"
 HEAPF32[$31>>2] = $39; //@line 284 "./common/Matrix.h"
 $40 = ((($sprite)) + 40|0); //@line 280 "./common/Matrix.h"
 $41 = +HEAPF32[$40>>2]; //@line 280 "./common/Matrix.h"
 $42 = $11 * $41; //@line 280 "./common/Matrix.h"
 $43 = ((($sprite)) + 44|0); //@line 280 "./common/Matrix.h"
 $44 = +HEAPF32[$43>>2]; //@line 280 "./common/Matrix.h"
 $45 = $12 * $44; //@line 280 "./common/Matrix.h"
 $46 = $42 + $45; //@line 280 "./common/Matrix.h"
 $47 = $13 + $46; //@line 280 "./common/Matrix.h"
 $48 = $41 * $14; //@line 281 "./common/Matrix.h"
 $49 = $44 * $15; //@line 281 "./common/Matrix.h"
 $50 = $48 + $49; //@line 281 "./common/Matrix.h"
 $51 = $16 + $50; //@line 281 "./common/Matrix.h"
 HEAPF32[$40>>2] = $47; //@line 283 "./common/Matrix.h"
 HEAPF32[$43>>2] = $51; //@line 284 "./common/Matrix.h"
 $52 = ((($sprite)) + 60|0); //@line 280 "./common/Matrix.h"
 $53 = +HEAPF32[$52>>2]; //@line 280 "./common/Matrix.h"
 $54 = $11 * $53; //@line 280 "./common/Matrix.h"
 $55 = ((($sprite)) + 64|0); //@line 280 "./common/Matrix.h"
 $56 = +HEAPF32[$55>>2]; //@line 280 "./common/Matrix.h"
 $57 = $12 * $56; //@line 280 "./common/Matrix.h"
 $58 = $54 + $57; //@line 280 "./common/Matrix.h"
 $59 = $13 + $58; //@line 280 "./common/Matrix.h"
 $60 = $53 * $14; //@line 281 "./common/Matrix.h"
 $61 = $56 * $15; //@line 281 "./common/Matrix.h"
 $62 = $60 + $61; //@line 281 "./common/Matrix.h"
 $63 = $16 + $62; //@line 281 "./common/Matrix.h"
 HEAPF32[$52>>2] = $59; //@line 283 "./common/Matrix.h"
 HEAPF32[$55>>2] = $63; //@line 284 "./common/Matrix.h"
 $64 = ((($this)) + 20|0); //@line 286 "modules/graphics/opengl/SpriteBatch.cpp"
 $65 = HEAP32[$64>>2]|0; //@line 286 "modules/graphics/opengl/SpriteBatch.cpp"
 $66 = ($65|0)==(0|0); //@line 286 "modules/graphics/opengl/SpriteBatch.cpp"
 if (!($66)) {
  $67 = HEAP8[$65>>0]|0; //@line 302 "modules/graphics/opengl/SpriteBatch.cpp"
  $68 = ((($65)) + 1|0); //@line 303 "modules/graphics/opengl/SpriteBatch.cpp"
  $69 = HEAP8[$68>>0]|0; //@line 303 "modules/graphics/opengl/SpriteBatch.cpp"
  $70 = ((($65)) + 2|0); //@line 304 "modules/graphics/opengl/SpriteBatch.cpp"
  $71 = HEAP8[$70>>0]|0; //@line 304 "modules/graphics/opengl/SpriteBatch.cpp"
  $72 = ((($65)) + 3|0); //@line 305 "modules/graphics/opengl/SpriteBatch.cpp"
  $73 = HEAP8[$72>>0]|0; //@line 305 "modules/graphics/opengl/SpriteBatch.cpp"
  $74 = ((($sprite)) + 16|0); //@line 302 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$74>>0] = $67; //@line 302 "modules/graphics/opengl/SpriteBatch.cpp"
  $75 = ((($sprite)) + 17|0); //@line 303 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$75>>0] = $69; //@line 303 "modules/graphics/opengl/SpriteBatch.cpp"
  $76 = ((($sprite)) + 18|0); //@line 304 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$76>>0] = $71; //@line 304 "modules/graphics/opengl/SpriteBatch.cpp"
  $77 = ((($sprite)) + 19|0); //@line 305 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$77>>0] = $73; //@line 305 "modules/graphics/opengl/SpriteBatch.cpp"
  $78 = ((($sprite)) + 36|0); //@line 302 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$78>>0] = $67; //@line 302 "modules/graphics/opengl/SpriteBatch.cpp"
  $79 = ((($sprite)) + 37|0); //@line 303 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$79>>0] = $69; //@line 303 "modules/graphics/opengl/SpriteBatch.cpp"
  $80 = ((($sprite)) + 38|0); //@line 304 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$80>>0] = $71; //@line 304 "modules/graphics/opengl/SpriteBatch.cpp"
  $81 = ((($sprite)) + 39|0); //@line 305 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$81>>0] = $73; //@line 305 "modules/graphics/opengl/SpriteBatch.cpp"
  $82 = ((($sprite)) + 56|0); //@line 302 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$82>>0] = $67; //@line 302 "modules/graphics/opengl/SpriteBatch.cpp"
  $83 = ((($sprite)) + 57|0); //@line 303 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$83>>0] = $69; //@line 303 "modules/graphics/opengl/SpriteBatch.cpp"
  $84 = ((($sprite)) + 58|0); //@line 304 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$84>>0] = $71; //@line 304 "modules/graphics/opengl/SpriteBatch.cpp"
  $85 = ((($sprite)) + 59|0); //@line 305 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$85>>0] = $73; //@line 305 "modules/graphics/opengl/SpriteBatch.cpp"
  $86 = ((($sprite)) + 76|0); //@line 302 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$86>>0] = $67; //@line 302 "modules/graphics/opengl/SpriteBatch.cpp"
  $87 = ((($sprite)) + 77|0); //@line 303 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$87>>0] = $69; //@line 303 "modules/graphics/opengl/SpriteBatch.cpp"
  $88 = ((($sprite)) + 78|0); //@line 304 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$88>>0] = $71; //@line 304 "modules/graphics/opengl/SpriteBatch.cpp"
  $89 = ((($sprite)) + 79|0); //@line 305 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$89>>0] = $73; //@line 305 "modules/graphics/opengl/SpriteBatch.cpp"
 }
 $90 = ((($this)) + 24|0); //@line 289 "modules/graphics/opengl/SpriteBatch.cpp"
 $91 = HEAP32[$90>>2]|0; //@line 289 "modules/graphics/opengl/SpriteBatch.cpp"
 __ZN4love8graphics6opengl8GLBuffer4bindEv($91); //@line 188 "modules/graphics/opengl/GLBuffer.h"
 $92 = HEAP32[$90>>2]|0; //@line 293 "modules/graphics/opengl/SpriteBatch.cpp"
 __THREW__ = 0;
 (invoke_ii(4096,($92|0))|0); //@line 293 "modules/graphics/opengl/SpriteBatch.cpp"
 $93 = __THREW__; __THREW__ = 0;
 $94 = $93&1;
 if (!($94)) {
  $95 = HEAP32[$90>>2]|0; //@line 295 "modules/graphics/opengl/SpriteBatch.cpp"
  $96 = ($index*80)|0; //@line 295 "modules/graphics/opengl/SpriteBatch.cpp"
  __THREW__ = 0;
  invoke_viiii(4138,($95|0),($96|0),80,($sprite|0)); //@line 295 "modules/graphics/opengl/SpriteBatch.cpp"
  $97 = __THREW__; __THREW__ = 0;
  $98 = $97&1;
  if (!($98)) {
   __THREW__ = 0;
   invoke_vi(4089,($91|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   $99 = __THREW__; __THREW__ = 0;
   $100 = $99&1;
   if ($100) {
    $101 = ___cxa_find_matching_catch_3(0|0)|0;
    $102 = tempRet0;
    ___clang_call_terminate($101); //@line 196 "modules/graphics/opengl/GLBuffer.h"
    // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
   } else {
    STACKTOP = sp;return; //@line 296 "modules/graphics/opengl/SpriteBatch.cpp"
   }
  }
 }
 $103 = ___cxa_find_matching_catch_2()|0;
 $104 = tempRet0;
 __THREW__ = 0;
 invoke_vi(4089,($91|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
 $105 = __THREW__; __THREW__ = 0;
 $106 = $105&1;
 if ($106) {
  $107 = ___cxa_find_matching_catch_3(0|0)|0;
  $108 = tempRet0;
  ___clang_call_terminate($107); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
 } else {
  ___resumeException($103|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl11SpriteBatch4addqEPNS0_4QuadEfffffffffi($this,$quad,$x,$y,$a,$sx,$sy,$ox,$oy,$kx,$ky,$index) {
 $this = $this|0;
 $quad = $quad|0;
 $x = +$x;
 $y = +$y;
 $a = +$a;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 $index = $index|0;
 var $$0 = 0, $$1 = 0, $$old1 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $t = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $t = sp;
 $0 = ($index|0)==(-1); //@line 87 "modules/graphics/opengl/SpriteBatch.cpp"
 do {
  if ($0) {
   $1 = ((($this)) + 16|0); //@line 87 "modules/graphics/opengl/SpriteBatch.cpp"
   $2 = HEAP32[$1>>2]|0; //@line 87 "modules/graphics/opengl/SpriteBatch.cpp"
   $3 = ((($this)) + 12|0); //@line 87 "modules/graphics/opengl/SpriteBatch.cpp"
   $4 = HEAP32[$3>>2]|0; //@line 87 "modules/graphics/opengl/SpriteBatch.cpp"
   $5 = ($2|0)<($4|0); //@line 87 "modules/graphics/opengl/SpriteBatch.cpp"
   if ($5) {
    $$pre$phiZ2D = $1;$7 = $2;
   } else {
    $$1 = -1;
    STACKTOP = sp;return ($$1|0); //@line 99 "modules/graphics/opengl/SpriteBatch.cpp"
   }
  } else {
   $$old1 = ($index|0)<(-1); //@line 87 "modules/graphics/opengl/SpriteBatch.cpp"
   if ($$old1) {
    $$1 = -1;
    STACKTOP = sp;return ($$1|0); //@line 99 "modules/graphics/opengl/SpriteBatch.cpp"
   } else {
    $$phi$trans$insert = ((($this)) + 16|0); //@line 87 "modules/graphics/opengl/SpriteBatch.cpp"
    $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 87 "modules/graphics/opengl/SpriteBatch.cpp"
    $$pre$phiZ2D = $$phi$trans$insert;$7 = $$pre;
    break;
   }
  }
 } while(0);
 $6 = ($7|0)>($index|0); //@line 87 "modules/graphics/opengl/SpriteBatch.cpp"
 if (!($6)) {
  $$1 = -1;
  STACKTOP = sp;return ($$1|0); //@line 99 "modules/graphics/opengl/SpriteBatch.cpp"
 }
 __ZN4love7Matrix3C2Efffffffff($t,$x,$y,$a,$sx,$sy,$ox,$oy,$kx,$ky); //@line 90 "modules/graphics/opengl/SpriteBatch.cpp"
 __THREW__ = 0;
 $8 = (invoke_ii(4010,($quad|0))|0); //@line 92 "modules/graphics/opengl/SpriteBatch.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  __ZN4love7Matrix3D2Ev($t); //@line 99 "modules/graphics/opengl/SpriteBatch.cpp"
  ___resumeException($17|0);
  // unreachable;
 }
 if ($0) {
  $11 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 92 "modules/graphics/opengl/SpriteBatch.cpp"
  $12 = $11;
 } else {
  $12 = $index;
 }
 __THREW__ = 0;
 invoke_viiii(4137,($this|0),($8|0),($t|0),($12|0)); //@line 92 "modules/graphics/opengl/SpriteBatch.cpp"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if ($14) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  __ZN4love7Matrix3D2Ev($t); //@line 99 "modules/graphics/opengl/SpriteBatch.cpp"
  ___resumeException($17|0);
  // unreachable;
 }
 if ($0) {
  $15 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 96 "modules/graphics/opengl/SpriteBatch.cpp"
  $16 = (($15) + 1)|0; //@line 96 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP32[$$pre$phiZ2D>>2] = $16; //@line 96 "modules/graphics/opengl/SpriteBatch.cpp"
  $$0 = $15;
 } else {
  $$0 = $index;
 }
 __ZN4love7Matrix3D2Ev($t); //@line 99 "modules/graphics/opengl/SpriteBatch.cpp"
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0); //@line 99 "modules/graphics/opengl/SpriteBatch.cpp"
}
function __ZN4love8graphics6opengl11SpriteBatch5clearEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 104 "modules/graphics/opengl/SpriteBatch.cpp"
 HEAP32[$0>>2] = 0; //@line 104 "modules/graphics/opengl/SpriteBatch.cpp"
 return; //@line 105 "modules/graphics/opengl/SpriteBatch.cpp"
}
function __ZN4love8graphics6opengl11SpriteBatch5flushEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 109 "modules/graphics/opengl/SpriteBatch.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 109 "modules/graphics/opengl/SpriteBatch.cpp"
 __ZN4love8graphics6opengl8GLBuffer4bindEv($1); //@line 188 "modules/graphics/opengl/GLBuffer.h"
 $2 = HEAP32[$0>>2]|0; //@line 110 "modules/graphics/opengl/SpriteBatch.cpp"
 __THREW__ = 0;
 invoke_vi(4088,($2|0)); //@line 110 "modules/graphics/opengl/SpriteBatch.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  __THREW__ = 0;
  invoke_vi(4089,($1|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  $11 = __THREW__; __THREW__ = 0;
  $12 = $11&1;
  if ($12) {
   $13 = ___cxa_find_matching_catch_3(0|0)|0;
   $14 = tempRet0;
   ___clang_call_terminate($13); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
  } else {
   ___resumeException($9|0);
   // unreachable;
  }
 } else {
  __THREW__ = 0;
  invoke_vi(4089,($1|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if ($6) {
   $7 = ___cxa_find_matching_catch_3(0|0)|0;
   $8 = tempRet0;
   ___clang_call_terminate($7); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
  } else {
   return; //@line 111 "modules/graphics/opengl/SpriteBatch.cpp"
  }
 }
}
function __ZN4love8graphics6opengl11SpriteBatch10setTextureEPNS0_7TextureE($this,$newtexture) {
 $this = $this|0;
 $newtexture = $newtexture|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($newtexture|0)==(0|0); //@line 132 "./common/Object.h"
 if (!($0)) {
  $1 = HEAP32[$newtexture>>2]|0; //@line 132 "./common/Object.h"
  $2 = ((($1)) + 8|0); //@line 132 "./common/Object.h"
  $3 = HEAP32[$2>>2]|0; //@line 132 "./common/Object.h"
  FUNCTION_TABLE_vi[$3 & 8191]($newtexture); //@line 132 "./common/Object.h"
 }
 $4 = ((($this)) + 8|0); //@line 133 "./common/Object.h"
 $5 = HEAP32[$4>>2]|0; //@line 133 "./common/Object.h"
 $6 = ($5|0)==(0|0); //@line 133 "./common/Object.h"
 if ($6) {
  HEAP32[$4>>2] = $newtexture; //@line 134 "./common/Object.h"
  return; //@line 116 "modules/graphics/opengl/SpriteBatch.cpp"
 }
 $7 = HEAP32[$5>>2]|0; //@line 133 "./common/Object.h"
 $8 = ((($7)) + 12|0); //@line 133 "./common/Object.h"
 $9 = HEAP32[$8>>2]|0; //@line 133 "./common/Object.h"
 FUNCTION_TABLE_vi[$9 & 8191]($5); //@line 133 "./common/Object.h"
 HEAP32[$4>>2] = $newtexture; //@line 134 "./common/Object.h"
 return; //@line 116 "modules/graphics/opengl/SpriteBatch.cpp"
}
function __ZNK4love8graphics6opengl11SpriteBatch10getTextureEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 139 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 139 "./common/Object.h"
 return ($1|0); //@line 120 "modules/graphics/opengl/SpriteBatch.cpp"
}
function __ZN4love8graphics6opengl11SpriteBatch8setColorERKNS0_6ColorTIhEE($this,$color) {
 $this = $this|0;
 $color = $color|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 125 "modules/graphics/opengl/SpriteBatch.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 125 "modules/graphics/opengl/SpriteBatch.cpp"
 $2 = ($1|0)==(0|0); //@line 125 "modules/graphics/opengl/SpriteBatch.cpp"
 if ($2) {
  $3 = (__Znwj(4)|0); //@line 126 "modules/graphics/opengl/SpriteBatch.cpp"
  $4 = HEAPU8[$color>>0]|(HEAPU8[$color+1>>0]<<8)|(HEAPU8[$color+2>>0]<<16)|(HEAPU8[$color+3>>0]<<24); //@line 126 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$3>>0]=$4&255;HEAP8[$3+1>>0]=($4>>8)&255;HEAP8[$3+2>>0]=($4>>16)&255;HEAP8[$3+3>>0]=$4>>24; //@line 126 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP32[$0>>2] = $3; //@line 126 "modules/graphics/opengl/SpriteBatch.cpp"
  return; //@line 129 "modules/graphics/opengl/SpriteBatch.cpp"
 } else {
  $5 = HEAPU8[$color>>0]|(HEAPU8[$color+1>>0]<<8)|(HEAPU8[$color+2>>0]<<16)|(HEAPU8[$color+3>>0]<<24); //@line 128 "modules/graphics/opengl/SpriteBatch.cpp"
  HEAP8[$1>>0]=$5&255;HEAP8[$1+1>>0]=($5>>8)&255;HEAP8[$1+2>>0]=($5>>16)&255;HEAP8[$1+3>>0]=$5>>24; //@line 128 "modules/graphics/opengl/SpriteBatch.cpp"
  return; //@line 129 "modules/graphics/opengl/SpriteBatch.cpp"
 }
}
function __ZN4love8graphics6opengl11SpriteBatch8setColorEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 133 "modules/graphics/opengl/SpriteBatch.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 133 "modules/graphics/opengl/SpriteBatch.cpp"
 $2 = ($1|0)==(0|0); //@line 133 "modules/graphics/opengl/SpriteBatch.cpp"
 if (!($2)) {
  __ZdlPv($1); //@line 133 "modules/graphics/opengl/SpriteBatch.cpp"
 }
 HEAP32[$0>>2] = 0; //@line 134 "modules/graphics/opengl/SpriteBatch.cpp"
 return; //@line 135 "modules/graphics/opengl/SpriteBatch.cpp"
}
function __ZNK4love8graphics6opengl11SpriteBatch8getColorEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 139 "modules/graphics/opengl/SpriteBatch.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 139 "modules/graphics/opengl/SpriteBatch.cpp"
 return ($1|0); //@line 139 "modules/graphics/opengl/SpriteBatch.cpp"
}
function __ZNK4love8graphics6opengl11SpriteBatch8getCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 144 "modules/graphics/opengl/SpriteBatch.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 144 "modules/graphics/opengl/SpriteBatch.cpp"
 return ($1|0); //@line 144 "modules/graphics/opengl/SpriteBatch.cpp"
}
function __ZN4love8graphics6opengl11SpriteBatch13setBufferSizeEi($this,$newsize) {
 $this = $this|0;
 $newsize = $newsize|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $new_array_buf$0 = 0, $new_array_buf$1 = 0, $vararg_buffer = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 $1 = ($newsize|0)<(1); //@line 149 "modules/graphics/opengl/SpriteBatch.cpp"
 if ($1) {
  $2 = (___cxa_allocate_exception(16)|0); //@line 150 "modules/graphics/opengl/SpriteBatch.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($2|0),(439031|0),($vararg_buffer|0)); //@line 150 "modules/graphics/opengl/SpriteBatch.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if (!($4)) {
   ___cxa_throw(($2|0),(96|0),(344|0)); //@line 150 "modules/graphics/opengl/SpriteBatch.cpp"
   // unreachable; //@line 150 "modules/graphics/opengl/SpriteBatch.cpp"
  }
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  ___cxa_free_exception(($2|0)); //@line 150 "modules/graphics/opengl/SpriteBatch.cpp"
  $$2 = $5;$$23 = $6;
  ___resumeException($$2|0);
  // unreachable;
 }
 $7 = ((($this)) + 12|0); //@line 152 "modules/graphics/opengl/SpriteBatch.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 152 "modules/graphics/opengl/SpriteBatch.cpp"
 $9 = ($8|0)==($newsize|0); //@line 152 "modules/graphics/opengl/SpriteBatch.cpp"
 if ($9) {
  STACKTOP = sp;return; //@line 189 "modules/graphics/opengl/SpriteBatch.cpp"
 }
 $10 = ((($this)) + 24|0); //@line 158 "modules/graphics/opengl/SpriteBatch.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 158 "modules/graphics/opengl/SpriteBatch.cpp"
 __ZN4love8graphics6opengl8GLBuffer4bindEv($11); //@line 188 "modules/graphics/opengl/GLBuffer.h"
 $12 = HEAP32[$10>>2]|0; //@line 159 "modules/graphics/opengl/SpriteBatch.cpp"
 __THREW__ = 0;
 $13 = (invoke_ii(4096,($12|0))|0); //@line 159 "modules/graphics/opengl/SpriteBatch.cpp"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if ($15) {
  $53 = ___cxa_find_matching_catch_2()|0;
  $54 = tempRet0;
  __THREW__ = 0;
  invoke_vi(4089,($11|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  $55 = __THREW__; __THREW__ = 0;
  $56 = $55&1;
  if ($56) {
   $57 = ___cxa_find_matching_catch_3(0|0)|0;
   $58 = tempRet0;
   ___clang_call_terminate($57); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
  } else {
   $$2 = $53;$$23 = $54;
   ___resumeException($$2|0);
   // unreachable;
  }
 }
 __THREW__ = 0;
 invoke_vi(4089,($11|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if ($17) {
  $18 = ___cxa_find_matching_catch_3(0|0)|0;
  $19 = tempRet0;
  ___clang_call_terminate($18); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
 }
 __THREW__ = 0;
 $20 = (invoke_ii(3348,40)|0); //@line 167 "modules/graphics/opengl/SpriteBatch.cpp"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 do {
  if ($22) {
   $new_array_buf$0 = 0;
   label = 20;
  } else {
   $23 = ($newsize*80)|0; //@line 162 "modules/graphics/opengl/SpriteBatch.cpp"
   $24 = HEAP32[$10>>2]|0; //@line 167 "modules/graphics/opengl/SpriteBatch.cpp"
   $25 = ((($24)) + 12|0); //@line 86 "modules/graphics/opengl/GLBuffer.h"
   $26 = HEAP32[$25>>2]|0; //@line 86 "modules/graphics/opengl/GLBuffer.h"
   $27 = ((($24)) + 16|0); //@line 96 "modules/graphics/opengl/GLBuffer.h"
   $28 = HEAP32[$27>>2]|0; //@line 96 "modules/graphics/opengl/GLBuffer.h"
   $29 = ((($24)) + 36|0); //@line 167 "modules/graphics/opengl/GLBuffer.h"
   $30 = HEAP32[$29>>2]|0; //@line 167 "modules/graphics/opengl/GLBuffer.h"
   __THREW__ = 0;
   invoke_viiiiii(4042,($20|0),($23|0),(0|0),($26|0),($28|0),($30|0)); //@line 167 "modules/graphics/opengl/SpriteBatch.cpp"
   $31 = __THREW__; __THREW__ = 0;
   $32 = $31&1;
   if ($32) {
    $61 = ___cxa_find_matching_catch_3(96|0)|0;
    $62 = tempRet0;
    __ZdlPv($20); //@line 167 "modules/graphics/opengl/SpriteBatch.cpp"
    $$1 = $61;$$12 = $62;$new_array_buf$1 = 0;
    break;
   }
   __THREW__ = 0;
   invoke_vi(4087,($20|0)); //@line 188 "modules/graphics/opengl/GLBuffer.h"
   $33 = __THREW__; __THREW__ = 0;
   $34 = $33&1;
   if ($34) {
    $new_array_buf$0 = $20;
    label = 20;
   } else {
    __THREW__ = 0;
    $35 = (invoke_ii(4096,($20|0))|0); //@line 171 "modules/graphics/opengl/SpriteBatch.cpp"
    $36 = __THREW__; __THREW__ = 0;
    $37 = $36&1;
    do {
     if ($37) {
      label = 22;
     } else {
      $38 = HEAP32[$7>>2]|0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $39 = ($38|0)<($newsize|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $40 = $39 ? $38 : $newsize; //@line 172 "modules/graphics/opengl/SpriteBatch.cpp"
      $41 = ($40*80)|0; //@line 172 "modules/graphics/opengl/SpriteBatch.cpp"
      _memcpy(($35|0),($13|0),($41|0))|0; //@line 172 "modules/graphics/opengl/SpriteBatch.cpp"
      __THREW__ = 0;
      invoke_vii(4020,($0|0),($newsize|0)); //@line 174 "modules/graphics/opengl/SpriteBatch.cpp"
      $42 = __THREW__; __THREW__ = 0;
      $43 = $42&1;
      if ($43) {
       label = 22;
      } else {
       $44 = ((($this)) + 28|0); //@line 174 "modules/graphics/opengl/SpriteBatch.cpp"
       __THREW__ = 0;
       (invoke_iii(4036,($44|0),($0|0))|0); //@line 174 "modules/graphics/opengl/SpriteBatch.cpp"
       $45 = __THREW__; __THREW__ = 0;
       $46 = $45&1;
       if ($46) {
        $65 = ___cxa_find_matching_catch_3(96|0)|0;
        $66 = tempRet0;
        __ZN4love8graphics6opengl11QuadIndicesD2Ev($0); //@line 174 "modules/graphics/opengl/SpriteBatch.cpp"
        $$0 = $65;$$01 = $66;
        break;
       }
       __ZN4love8graphics6opengl11QuadIndicesD2Ev($0); //@line 174 "modules/graphics/opengl/SpriteBatch.cpp"
       __THREW__ = 0;
       invoke_vi(4089,($20|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
       $47 = __THREW__; __THREW__ = 0;
       $48 = $47&1;
       if ($48) {
        $49 = ___cxa_find_matching_catch_3(0|0)|0;
        $50 = tempRet0;
        ___clang_call_terminate($49); //@line 196 "modules/graphics/opengl/GLBuffer.h"
        // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
       }
       $51 = HEAP32[$10>>2]|0; //@line 183 "modules/graphics/opengl/SpriteBatch.cpp"
       $52 = ($51|0)==(0|0); //@line 183 "modules/graphics/opengl/SpriteBatch.cpp"
       if (!($52)) {
        $82 = HEAP32[$51>>2]|0; //@line 183 "modules/graphics/opengl/SpriteBatch.cpp"
        $83 = ((($82)) + 4|0); //@line 183 "modules/graphics/opengl/SpriteBatch.cpp"
        $84 = HEAP32[$83>>2]|0; //@line 183 "modules/graphics/opengl/SpriteBatch.cpp"
        FUNCTION_TABLE_vi[$84 & 8191]($51); //@line 183 "modules/graphics/opengl/SpriteBatch.cpp"
       }
       HEAP32[$10>>2] = $20; //@line 185 "modules/graphics/opengl/SpriteBatch.cpp"
       HEAP32[$7>>2] = $newsize; //@line 186 "modules/graphics/opengl/SpriteBatch.cpp"
       $85 = ((($this)) + 16|0); //@line 188 "modules/graphics/opengl/SpriteBatch.cpp"
       $86 = HEAP32[$85>>2]|0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $87 = ($86|0)>($newsize|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $88 = $87 ? $newsize : $86; //@line 188 "modules/graphics/opengl/SpriteBatch.cpp"
       HEAP32[$85>>2] = $88; //@line 188 "modules/graphics/opengl/SpriteBatch.cpp"
       STACKTOP = sp;return; //@line 189 "modules/graphics/opengl/SpriteBatch.cpp"
      }
     }
    } while(0);
    if ((label|0) == 22) {
     $63 = ___cxa_find_matching_catch_3(96|0)|0;
     $64 = tempRet0;
     $$0 = $63;$$01 = $64;
    }
    __THREW__ = 0;
    invoke_vi(4089,($20|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
    $67 = __THREW__; __THREW__ = 0;
    $68 = $67&1;
    if ($68) {
     $69 = ___cxa_find_matching_catch_3(0|0)|0;
     $70 = tempRet0;
     ___clang_call_terminate($69); //@line 196 "modules/graphics/opengl/GLBuffer.h"
     // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
    } else {
     $$1 = $$0;$$12 = $$01;$new_array_buf$1 = $20;
    }
   }
  }
 } while(0);
 if ((label|0) == 20) {
  $59 = ___cxa_find_matching_catch_3(96|0)|0;
  $60 = tempRet0;
  $$1 = $59;$$12 = $60;$new_array_buf$1 = $new_array_buf$0;
 }
 $71 = (_llvm_eh_typeid_for((96|0))|0); //@line 175 "modules/graphics/opengl/SpriteBatch.cpp"
 $72 = ($$12|0)==($71|0); //@line 175 "modules/graphics/opengl/SpriteBatch.cpp"
 if (!($72)) {
  $$2 = $$1;$$23 = $$12;
  ___resumeException($$2|0);
  // unreachable;
 }
 (___cxa_begin_catch(($$1|0))|0); //@line 175 "modules/graphics/opengl/SpriteBatch.cpp"
 $73 = ($new_array_buf$1|0)==(0|0); //@line 178 "modules/graphics/opengl/SpriteBatch.cpp"
 if (!($73)) {
  $74 = HEAP32[$new_array_buf$1>>2]|0; //@line 178 "modules/graphics/opengl/SpriteBatch.cpp"
  $75 = ((($74)) + 4|0); //@line 178 "modules/graphics/opengl/SpriteBatch.cpp"
  $76 = HEAP32[$75>>2]|0; //@line 178 "modules/graphics/opengl/SpriteBatch.cpp"
  FUNCTION_TABLE_vi[$76 & 8191]($new_array_buf$1); //@line 178 "modules/graphics/opengl/SpriteBatch.cpp"
 }
 __THREW__ = 0;
 invoke_v(3656); //@line 179 "modules/graphics/opengl/SpriteBatch.cpp"
 $77 = __THREW__; __THREW__ = 0;
 $78 = ___cxa_find_matching_catch_2()|0;
 $79 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 180 "modules/graphics/opengl/SpriteBatch.cpp"
 $80 = __THREW__; __THREW__ = 0;
 $81 = $80&1;
 if ($81) {
  $89 = ___cxa_find_matching_catch_3(0|0)|0;
  $90 = tempRet0;
  ___clang_call_terminate($89); //@line 180 "modules/graphics/opengl/SpriteBatch.cpp"
  // unreachable; //@line 180 "modules/graphics/opengl/SpriteBatch.cpp"
 } else {
  $$2 = $78;$$23 = $79;
  ___resumeException($$2|0);
  // unreachable;
 }
}
function __ZNK4love8graphics6opengl11SpriteBatch13getBufferSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 193 "modules/graphics/opengl/SpriteBatch.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 193 "modules/graphics/opengl/SpriteBatch.cpp"
 return ($1|0); //@line 193 "modules/graphics/opengl/SpriteBatch.cpp"
}
function __ZN4love8graphics6opengl11SpriteBatch15attachAttributeERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEEPNS1_4MeshE($this,$name,$mesh) {
 $this = $this|0;
 $name = $name|0;
 $mesh = $mesh|0;
 var $$0115 = 0, $$011522 = 0, $$016 = 0, $$01621 = 0, $$pre$phi24Z2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $oldattrib$sroa$0$0$ph = 0, $oldattrib$sroa$0$1 = 0, $oldattrib$sroa$0$214 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 __THREW__ = 0;
 $0 = (invoke_ii(4132,($mesh|0))|0); //@line 201 "modules/graphics/opengl/SpriteBatch.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 L1: do {
  if ($2) {
   $oldattrib$sroa$0$0$ph = 0;
   label = 34;
  } else {
   $3 = ((($this)) + 12|0); //@line 193 "modules/graphics/opengl/SpriteBatch.cpp"
   $4 = HEAP32[$3>>2]|0; //@line 193 "modules/graphics/opengl/SpriteBatch.cpp"
   $5 = $4 << 2; //@line 201 "modules/graphics/opengl/SpriteBatch.cpp"
   $6 = ($0>>>0)<($5>>>0); //@line 201 "modules/graphics/opengl/SpriteBatch.cpp"
   if ($6) {
    $7 = (___cxa_allocate_exception(16)|0); //@line 202 "modules/graphics/opengl/SpriteBatch.cpp"
    $8 = HEAP32[$3>>2]|0; //@line 193 "modules/graphics/opengl/SpriteBatch.cpp"
    $9 = $8 << 2; //@line 202 "modules/graphics/opengl/SpriteBatch.cpp"
    __THREW__ = 0;
    HEAP32[$vararg_buffer>>2] = $9; //@line 202 "modules/graphics/opengl/SpriteBatch.cpp"
    invoke_viii(3349,($7|0),(439057|0),($vararg_buffer|0)); //@line 202 "modules/graphics/opengl/SpriteBatch.cpp"
    $10 = __THREW__; __THREW__ = 0;
    $11 = $10&1;
    if (!($11)) {
     __THREW__ = 0;
     invoke_viii(3350,($7|0),(96|0),(344|0)); //@line 202 "modules/graphics/opengl/SpriteBatch.cpp"
     $12 = __THREW__; __THREW__ = 0;
     $oldattrib$sroa$0$0$ph = 0;
     label = 34;
     break;
    }
    $13 = ___cxa_find_matching_catch_2()|0;
    $14 = tempRet0;
    ___cxa_free_exception(($7|0)); //@line 202 "modules/graphics/opengl/SpriteBatch.cpp"
    $$011522 = $13;$$01621 = $14;
    ___resumeException($$011522|0);
    // unreachable;
   }
   $15 = ((($this)) + 32|0); //@line 204 "modules/graphics/opengl/SpriteBatch.cpp"
   $16 = (__ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl11SpriteBatch17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE4findIS7_EENS_15__hash_iteratorIPNS_11__hash_nodeISD_PvEEEERKT_($15,$name)|0); //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
   $17 = ($16|0)==(0|0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($17) {
    $oldattrib$sroa$0$1 = 0;
   } else {
    $18 = ((($16)) + 20|0); //@line 139 "./common/Object.h"
    $19 = HEAP32[$18>>2]|0; //@line 139 "./common/Object.h"
    $20 = ($19|0)==(0|0); //@line 132 "./common/Object.h"
    if ($20) {
     $oldattrib$sroa$0$1 = 0;
    } else {
     $21 = HEAP32[$19>>2]|0; //@line 132 "./common/Object.h"
     $22 = ((($21)) + 8|0); //@line 132 "./common/Object.h"
     $23 = HEAP32[$22>>2]|0; //@line 132 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($23|0,($19|0)); //@line 132 "./common/Object.h"
     $24 = __THREW__; __THREW__ = 0;
     $25 = $24&1;
     if ($25) {
      $oldattrib$sroa$0$0$ph = 0;
      label = 34;
      break;
     } else {
      $oldattrib$sroa$0$1 = $19;
     }
    }
   }
   __THREW__ = 0;
   $26 = (invoke_iii(4139,($mesh|0),($name|0))|0); //@line 208 "modules/graphics/opengl/SpriteBatch.cpp"
   $27 = __THREW__; __THREW__ = 0;
   $28 = $27&1;
   if ($28) {
    $oldattrib$sroa$0$0$ph = $oldattrib$sroa$0$1;
    label = 34;
   } else {
    $29 = ($26|0)<(0); //@line 210 "modules/graphics/opengl/SpriteBatch.cpp"
    if ($29) {
     $30 = (___cxa_allocate_exception(16)|0); //@line 211 "modules/graphics/opengl/SpriteBatch.cpp"
     $31 = HEAP8[$name>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $32 = $31 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $33 = ($32<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $34 = ((($name)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $35 = HEAP32[$34>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $36 = ((($name)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $37 = $33 ? $36 : $35; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     __THREW__ = 0;
     HEAP32[$vararg_buffer1>>2] = $37; //@line 211 "modules/graphics/opengl/SpriteBatch.cpp"
     invoke_viii(3349,($30|0),(439154|0),($vararg_buffer1|0)); //@line 211 "modules/graphics/opengl/SpriteBatch.cpp"
     $38 = __THREW__; __THREW__ = 0;
     $39 = $38&1;
     if ($39) {
      $41 = ___cxa_find_matching_catch_2()|0;
      $42 = tempRet0;
      ___cxa_free_exception(($30|0)); //@line 211 "modules/graphics/opengl/SpriteBatch.cpp"
      $$0115 = $41;$$016 = $42;$oldattrib$sroa$0$214 = $oldattrib$sroa$0$1;
      break;
     } else {
      __THREW__ = 0;
      invoke_viii(3350,($30|0),(96|0),(344|0)); //@line 211 "modules/graphics/opengl/SpriteBatch.cpp"
      $40 = __THREW__; __THREW__ = 0;
      $oldattrib$sroa$0$0$ph = $oldattrib$sroa$0$1;
      label = 34;
      break;
     }
    }
    $43 = ($mesh|0)==(0|0); //@line 95 "./common/Object.h"
    do {
     if (!($43)) {
      $44 = HEAP32[$mesh>>2]|0; //@line 95 "./common/Object.h"
      $45 = ((($44)) + 8|0); //@line 95 "./common/Object.h"
      $46 = HEAP32[$45>>2]|0; //@line 95 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($46|0,($mesh|0)); //@line 95 "./common/Object.h"
      $47 = __THREW__; __THREW__ = 0;
      $48 = $47&1;
      if ($48) {
       $oldattrib$sroa$0$0$ph = $oldattrib$sroa$0$1;
       label = 34;
       break L1;
      }
      $49 = HEAP32[$mesh>>2]|0; //@line 132 "./common/Object.h"
      $50 = ((($49)) + 8|0); //@line 132 "./common/Object.h"
      $51 = HEAP32[$50>>2]|0; //@line 132 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($51|0,($mesh|0)); //@line 132 "./common/Object.h"
      $52 = __THREW__; __THREW__ = 0;
      $53 = $52&1;
      if ($53) {
       $94 = ___cxa_find_matching_catch_2()|0;
       $95 = tempRet0;
       $96 = HEAP32[$mesh>>2]|0; //@line 106 "./common/Object.h"
       $97 = ((($96)) + 12|0); //@line 106 "./common/Object.h"
       $98 = HEAP32[$97>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($98|0,($mesh|0)); //@line 106 "./common/Object.h"
       $99 = __THREW__; __THREW__ = 0;
       $100 = $99&1;
       if (!($100)) {
        $$0115 = $94;$$016 = $95;$oldattrib$sroa$0$214 = $oldattrib$sroa$0$1;
        break L1;
       }
       $101 = ___cxa_find_matching_catch_3(0|0)|0;
       $102 = tempRet0;
       ___clang_call_terminate($101); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      } else {
       $54 = HEAP32[$mesh>>2]|0; //@line 106 "./common/Object.h"
       $55 = ((($54)) + 12|0); //@line 106 "./common/Object.h"
       $56 = HEAP32[$55>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($56|0,($mesh|0)); //@line 106 "./common/Object.h"
       $57 = __THREW__; __THREW__ = 0;
       $58 = $57&1;
       if (!($58)) {
        break;
       }
       $59 = ___cxa_find_matching_catch_3(0|0)|0;
       $60 = tempRet0;
       ___clang_call_terminate($59); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      }
     }
    } while(0);
    __THREW__ = 0;
    $61 = (invoke_iii(4140,($15|0),($name|0))|0); //@line 215 "modules/graphics/opengl/SpriteBatch.cpp"
    $62 = __THREW__; __THREW__ = 0;
    $63 = $62&1;
    do {
     if ($63) {
      label = 35;
     } else {
      if (!($43)) {
       $64 = HEAP32[$mesh>>2]|0; //@line 132 "./common/Object.h"
       $65 = ((($64)) + 8|0); //@line 132 "./common/Object.h"
       $66 = HEAP32[$65>>2]|0; //@line 132 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($66|0,($mesh|0)); //@line 132 "./common/Object.h"
       $67 = __THREW__; __THREW__ = 0;
       $68 = $67&1;
       if ($68) {
        $69 = ___cxa_find_matching_catch_2()|0;
        $70 = tempRet0;
        $$pre$phi24Z2D = $mesh;$$pre$phiZ2D = $mesh;$122 = $70;$123 = $69;
        break;
       }
      }
      $71 = HEAP32[$61>>2]|0; //@line 133 "./common/Object.h"
      $72 = ($71|0)==(0|0); //@line 133 "./common/Object.h"
      if (!($72)) {
       $73 = HEAP32[$71>>2]|0; //@line 133 "./common/Object.h"
       $74 = ((($73)) + 12|0); //@line 133 "./common/Object.h"
       $75 = HEAP32[$74>>2]|0; //@line 133 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($75|0,($71|0)); //@line 133 "./common/Object.h"
       $76 = __THREW__; __THREW__ = 0;
       $77 = $76&1;
       if ($77) {
        label = 35;
        break;
       }
      }
      HEAP32[$61>>2] = $mesh; //@line 134 "./common/Object.h"
      $78 = ((($61)) + 4|0); //@line 115 "modules/graphics/opengl/SpriteBatch.h"
      HEAP32[$78>>2] = $26; //@line 115 "modules/graphics/opengl/SpriteBatch.h"
      if (!($43)) {
       $79 = HEAP32[$mesh>>2]|0; //@line 106 "./common/Object.h"
       $80 = ((($79)) + 12|0); //@line 106 "./common/Object.h"
       $81 = HEAP32[$80>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($81|0,($mesh|0)); //@line 106 "./common/Object.h"
       $82 = __THREW__; __THREW__ = 0;
       $83 = $82&1;
       if ($83) {
        $84 = ___cxa_find_matching_catch_3(0|0)|0;
        $85 = tempRet0;
        ___clang_call_terminate($84); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
      }
      $86 = ($oldattrib$sroa$0$1|0)==(0|0); //@line 106 "./common/Object.h"
      if ($86) {
       STACKTOP = sp;return; //@line 216 "modules/graphics/opengl/SpriteBatch.cpp"
      }
      $87 = HEAP32[$oldattrib$sroa$0$1>>2]|0; //@line 106 "./common/Object.h"
      $88 = ((($87)) + 12|0); //@line 106 "./common/Object.h"
      $89 = HEAP32[$88>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($89|0,($oldattrib$sroa$0$1|0)); //@line 106 "./common/Object.h"
      $90 = __THREW__; __THREW__ = 0;
      $91 = $90&1;
      if ($91) {
       $92 = ___cxa_find_matching_catch_3(0|0)|0;
       $93 = tempRet0;
       ___clang_call_terminate($92); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      } else {
       STACKTOP = sp;return; //@line 216 "modules/graphics/opengl/SpriteBatch.cpp"
      }
     }
    } while(0);
    if ((label|0) == 35) {
     $105 = ___cxa_find_matching_catch_2()|0;
     $106 = tempRet0;
     if ($43) {
      $$0115 = $105;$$016 = $106;$oldattrib$sroa$0$214 = $oldattrib$sroa$0$1;
      break;
     }
     $$pre$phi24Z2D = $mesh;$$pre$phiZ2D = $mesh;$122 = $106;$123 = $105;
    }
    $107 = HEAP32[$$pre$phi24Z2D>>2]|0; //@line 106 "./common/Object.h"
    $108 = ((($107)) + 12|0); //@line 106 "./common/Object.h"
    $109 = HEAP32[$108>>2]|0; //@line 106 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($109|0,($$pre$phiZ2D|0)); //@line 106 "./common/Object.h"
    $110 = __THREW__; __THREW__ = 0;
    $111 = $110&1;
    if ($111) {
     $112 = ___cxa_find_matching_catch_3(0|0)|0;
     $113 = tempRet0;
     ___clang_call_terminate($112); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    } else {
     $$0115 = $123;$$016 = $122;$oldattrib$sroa$0$214 = $oldattrib$sroa$0$1;
    }
   }
  }
 } while(0);
 if ((label|0) == 34) {
  $103 = ___cxa_find_matching_catch_2()|0;
  $104 = tempRet0;
  $$0115 = $103;$$016 = $104;$oldattrib$sroa$0$214 = $oldattrib$sroa$0$0$ph;
 }
 $114 = ($oldattrib$sroa$0$214|0)==(0|0); //@line 106 "./common/Object.h"
 if ($114) {
  $$011522 = $$0115;$$01621 = $$016;
  ___resumeException($$011522|0);
  // unreachable;
 }
 $115 = HEAP32[$oldattrib$sroa$0$214>>2]|0; //@line 106 "./common/Object.h"
 $116 = ((($115)) + 12|0); //@line 106 "./common/Object.h"
 $117 = HEAP32[$116>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($117|0,($oldattrib$sroa$0$214|0)); //@line 106 "./common/Object.h"
 $118 = __THREW__; __THREW__ = 0;
 $119 = $118&1;
 if ($119) {
  $120 = ___cxa_find_matching_catch_3(0|0)|0;
  $121 = tempRet0;
  ___clang_call_terminate($120); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  $$011522 = $$0115;$$01621 = $$016;
  ___resumeException($$011522|0);
  // unreachable;
 }
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl11SpriteBatch17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE4findIS7_EENS_15__hash_iteratorIPNS_11__hash_nodeISD_PvEEEERKT_($this,$__k) {
 $this = $this|0;
 $__k = $__k|0;
 var $$0$lcssa$i$i$i$i = 0, $$03$i$i$i$i = 0, $$lcssa = 0, $$mask$i$i = 0, $$mask$i$i$us = 0, $$pre$i$i = 0, $$pre$i$i$us = 0, $$sroa$0$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__data$0$lcssa$i$i$i$i = 0, $__data$02$i$i$i$i = 0, $__h$0$lcssa$i$i$i$i = 0, $__h$01$i$i$i$i = 0, $__h$1$i$i$i$i = 0, $__h$2$i$i$i$i = 0, $__h$3$i$i$i$i = 0, $__lhs_sz$03$i$i3$i = 0, $__lhs_sz$03$i$i3$i$us = 0, $__lp$04$i$i2$i = 0, $__lp$04$i$i2$i$us = 0;
 var $__nd$0 = 0, $__nd$0$us = 0, $__nd$08 = 0, $__nd$09 = 0, $__nd$09$us = 0, $__rp$05$i$i1$i = 0, $__rp$05$i$i1$i$us = 0, $scevgep$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$__k>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $1 = $0 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = ($1<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ((($__k)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = HEAP32[$3>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($__k)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = $2 ? $5 : $4; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($__k)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = HEAP32[$7>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $0&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $9 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $2 ? $10 : $8; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = ($11>>>0)>(3); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($12) {
  $$03$i$i$i$i = $11;$__data$02$i$i$i$i = $6;$__h$01$i$i$i$i = $11;
  while(1) {
   $13 = HEAPU8[$__data$02$i$i$i$i>>0]|(HEAPU8[$__data$02$i$i$i$i+1>>0]<<8)|(HEAPU8[$__data$02$i$i$i$i+2>>0]<<16)|(HEAPU8[$__data$02$i$i$i$i+3>>0]<<24); //@line 3073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $14 = Math_imul($13, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $15 = $14 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $16 = $15 ^ $14; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $17 = Math_imul($16, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $18 = Math_imul($__h$01$i$i$i$i, 1540483477)|0; //@line 3104 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $19 = $17 ^ $18; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $20 = ((($__data$02$i$i$i$i)) + 4|0); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $21 = (($$03$i$i$i$i) + -4)|0; //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $22 = ($21>>>0)>(3); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if ($22) {
    $$03$i$i$i$i = $21;$__data$02$i$i$i$i = $20;$__h$01$i$i$i$i = $19;
   } else {
    $$lcssa = $19;
    break;
   }
  }
  $23 = (($11) + -4)|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = $23 & -4; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $25 = (($24) + 4)|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (($23) - ($24))|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $scevgep$i$i$i$i = (($6) + ($25)|0); //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $$0$lcssa$i$i$i$i = $26;$__data$0$lcssa$i$i$i$i = $scevgep$i$i$i$i;$__h$0$lcssa$i$i$i$i = $$lcssa;
 } else {
  $$0$lcssa$i$i$i$i = $11;$__data$0$lcssa$i$i$i$i = $6;$__h$0$lcssa$i$i$i$i = $11;
 }
 switch ($$0$lcssa$i$i$i$i|0) {
 case 3:  {
  $27 = ((($__data$0$lcssa$i$i$i$i)) + 2|0); //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $28 = HEAP8[$27>>0]|0; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $29 = $28&255; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $30 = $29 << 16; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $31 = $30 ^ $__h$0$lcssa$i$i$i$i; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$1$i$i$i$i = $31;
  label = 6;
  break;
 }
 case 2:  {
  $__h$1$i$i$i$i = $__h$0$lcssa$i$i$i$i;
  label = 6;
  break;
 }
 case 1:  {
  $__h$2$i$i$i$i = $__h$0$lcssa$i$i$i$i;
  label = 7;
  break;
 }
 default: {
  $__h$3$i$i$i$i = $__h$0$lcssa$i$i$i$i;
 }
 }
 if ((label|0) == 6) {
  $32 = ((($__data$0$lcssa$i$i$i$i)) + 1|0); //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $33 = HEAP8[$32>>0]|0; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $34 = $33&255; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $35 = $34 << 8; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $36 = $35 ^ $__h$1$i$i$i$i; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$2$i$i$i$i = $36;
  label = 7;
 }
 if ((label|0) == 7) {
  $37 = HEAP8[$__data$0$lcssa$i$i$i$i>>0]|0; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $38 = $37&255; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $39 = $38 ^ $__h$2$i$i$i$i; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $40 = Math_imul($39, 1540483477)|0; //@line 3115 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$3$i$i$i$i = $40;
 }
 $41 = $__h$3$i$i$i$i >>> 13; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $42 = $41 ^ $__h$3$i$i$i$i; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $43 = Math_imul($42, 1540483477)|0; //@line 3118 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $44 = $43 >>> 15; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $45 = $44 ^ $43; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $46 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = HEAP32[$46>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $48 = ($47|0)==(0); //@line 2018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($48) {
  $$sroa$0$0 = 0;
  return ($$sroa$0$0|0); //@line 2038 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $49 = (($47) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $50 = $49 & $47; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $51 = ($50|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($51) {
  $52 = $45 & $49; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $56 = $52;
 } else {
  $53 = (($45>>>0) % ($47>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $56 = $53;
 }
 $54 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $55 = (($54) + ($56<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $57 = HEAP32[$55>>2]|0; //@line 2021 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $58 = ($57|0)==(0|0); //@line 2022 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($58) {
  $$sroa$0$0 = 0;
  return ($$sroa$0$0|0); //@line 2038 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $__nd$08 = HEAP32[$57>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $59 = ($__nd$08|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($59) {
  $$sroa$0$0 = 0;
  return ($$sroa$0$0|0); //@line 2038 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $60 = ($11|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($51) {
  $__nd$09$us = $__nd$08;
  L27: while(1) {
   $61 = ((($__nd$09$us)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $62 = HEAP32[$61>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $63 = $62 & $49; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $64 = ($63|0)==($56|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($64)) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   }
   $65 = ((($__nd$09$us)) + 8|0); //@line 2028 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $66 = HEAP16[$65>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $67 = $66 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $68 = ($67<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $69 = ((($__nd$09$us)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $70 = HEAP32[$69>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$mask$i$i$us = $66 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $71 = $$mask$i$i$us&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $72 = $71 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $73 = $68 ? $72 : $70; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $74 = ($73|0)==($11|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $75 = ($66&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $76 = $75&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   do {
    if ($74) {
     if (!($68)) {
      $77 = ((($__nd$09$us)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $78 = HEAP32[$77>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $79 = (_memcmp($78,$6,$11)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $80 = ($79|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($80) {
       $$sroa$0$0 = $__nd$09$us;
       label = 35;
       break L27;
      } else {
       break;
      }
     }
     if ($60) {
      $$sroa$0$0 = $__nd$09$us;
      label = 35;
      break L27;
     }
     $81 = HEAP8[$6>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $82 = ($76<<24>>24)==($81<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($82) {
      $83 = ((($65)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $__lhs_sz$03$i$i3$i$us = $72;$__lp$04$i$i2$i$us = $83;$__rp$05$i$i1$i$us = $6;
      while(1) {
       $84 = (($__lhs_sz$03$i$i3$i$us) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $85 = ((($__lp$04$i$i2$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $86 = ($84|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($86) {
        $$sroa$0$0 = $__nd$09$us;
        label = 35;
        break L27;
       }
       $87 = ((($__rp$05$i$i1$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$pre$i$i$us = HEAP8[$85>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $88 = HEAP8[$87>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $89 = ($$pre$i$i$us<<24>>24)==($88<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($89) {
        $__lhs_sz$03$i$i3$i$us = $84;$__lp$04$i$i2$i$us = $85;$__rp$05$i$i1$i$us = $87;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   $__nd$0$us = HEAP32[$__nd$09$us>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $90 = ($__nd$0$us|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($90) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   } else {
    $__nd$09$us = $__nd$0$us;
   }
  }
  if ((label|0) == 35) {
   return ($$sroa$0$0|0); //@line 2038 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 } else {
  $__nd$09 = $__nd$08;
  L47: while(1) {
   $91 = ((($__nd$09)) + 4|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $92 = HEAP32[$91>>2]|0; //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $93 = (($92>>>0) % ($47>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $94 = ($93|0)==($56|0); //@line 2025 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if (!($94)) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   }
   $95 = ((($__nd$09)) + 8|0); //@line 2028 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $96 = HEAP16[$95>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $97 = $96 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $98 = ($97<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $99 = ((($__nd$09)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $100 = HEAP32[$99>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$mask$i$i = $96 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $101 = $$mask$i$i&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $102 = $101 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $103 = $98 ? $102 : $100; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $104 = ($103|0)==($11|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $105 = ($96&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $106 = $105&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   do {
    if ($104) {
     if (!($98)) {
      $116 = ((($__nd$09)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $117 = HEAP32[$116>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $118 = (_memcmp($117,$6,$11)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $119 = ($118|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($119) {
       $$sroa$0$0 = $__nd$09;
       label = 35;
       break L47;
      } else {
       break;
      }
     }
     if ($60) {
      $$sroa$0$0 = $__nd$09;
      label = 35;
      break L47;
     }
     $107 = HEAP8[$6>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $108 = ($106<<24>>24)==($107<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($108) {
      $109 = ((($95)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $__lhs_sz$03$i$i3$i = $102;$__lp$04$i$i2$i = $109;$__rp$05$i$i1$i = $6;
      while(1) {
       $110 = (($__lhs_sz$03$i$i3$i) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $111 = ((($__lp$04$i$i2$i)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $112 = ($110|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($112) {
        $$sroa$0$0 = $__nd$09;
        label = 35;
        break L47;
       }
       $113 = ((($__rp$05$i$i1$i)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$pre$i$i = HEAP8[$111>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $114 = HEAP8[$113>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $115 = ($$pre$i$i<<24>>24)==($114<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($115) {
        $__lhs_sz$03$i$i3$i = $110;$__lp$04$i$i2$i = $111;$__rp$05$i$i1$i = $113;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   $__nd$0 = HEAP32[$__nd$09>>2]|0; //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $120 = ($__nd$0|0)==(0|0); //@line 2024 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($120) {
    $$sroa$0$0 = 0;
    label = 35;
    break;
   } else {
    $__nd$09 = $__nd$0;
   }
  }
  if ((label|0) == 35) {
   return ($$sroa$0$0|0); //@line 2038 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 }
 return (0)|0;
}
function __ZNSt3__113unordered_mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl11SpriteBatch17AttachedAttributeENS_4hashIS6_EENS_8equal_toIS6_EENS4_INS_4pairIKS6_SB_EEEEEixERSH_($this,$__k) {
 $this = $this|0;
 $__k = $__k|0;
 var $$0 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl11SpriteBatch17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE4findIS7_EENS_15__hash_iteratorIPNS_11__hash_nodeISD_PvEEEERKT_($this,$__k)|0); //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $2 = ($1|0)==(0|0); //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($2)) {
  $$sink = $1;
  $$0 = ((($$sink)) + 20|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
  STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 }
 $3 = (__Znwj(28)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $4 = ((($3)) + 8|0); //@line 1340 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 __THREW__ = 0;
 invoke_vii(3351,($4|0),($__k|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZdlPv($3); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($7|0);
  // unreachable;
 }
 $9 = ((($3)) + 20|0); //@line 1342 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $10 = $9; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $11 = $10; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$11>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $12 = (($10) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $13 = $12; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$13>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$9>>2] = 0; //@line 88 "./common/Object.h"
 __THREW__ = 0;
 invoke_viii(4141,($0|0),($this|0),($3|0)); //@line 1366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if (!($15)) {
  $16 = HEAP32[$0>>2]|0; //@line 315 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
  $$sink = $16;
  $$0 = ((($$sink)) + 20|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
  STACKTOP = sp;return ($$0|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/unordered_map"
 }
 $17 = ___cxa_find_matching_catch_2()|0;
 $18 = tempRet0;
 $19 = HEAP32[$9>>2]|0; //@line 106 "./common/Object.h"
 $20 = ($19|0)==(0|0); //@line 106 "./common/Object.h"
 if ($20) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 1589 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  __ZdlPv($3); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($17|0);
  // unreachable;
 }
 $21 = HEAP32[$19>>2]|0; //@line 106 "./common/Object.h"
 $22 = ((($21)) + 12|0); //@line 106 "./common/Object.h"
 $23 = HEAP32[$22>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($23|0,($19|0)); //@line 106 "./common/Object.h"
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $26 = ___cxa_find_matching_catch_3(0|0)|0;
  $27 = tempRet0;
  ___clang_call_terminate($26); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 1589 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  __ZdlPv($3); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($17|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl11SpriteBatch17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE20__node_insert_uniqueEPNS_11__hash_nodeISD_PvEE($agg$result,$this,$__nd) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $__nd = $__nd|0;
 var $$0$lcssa$i$i$i$i = 0, $$03$i$i$i$i = 0, $$lcssa = 0, $$mask$i$i = 0, $$mask$i$i$us = 0, $$pre$i$i = 0, $$pre$i$i$us = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0.0, $127 = 0.0;
 var $128 = 0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0.0, $142 = 0.0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__bc$0 = 0, $__chash$0 = 0, $__chash$1 = 0, $__data$0$lcssa$i$i$i$i = 0, $__data$02$i$i$i$i = 0, $__h$0$lcssa$i$i$i$i = 0, $__h$01$i$i$i$i = 0, $__h$1$i$i$i$i = 0, $__h$2$i$i$i$i = 0, $__h$3$i$i$i$i = 0, $__inserted$0 = 0, $__lhs_sz$03$i$i3$i = 0, $__lhs_sz$03$i$i3$i$us = 0, $__lp$04$i$i2$i = 0, $__lp$04$i$i2$i$us = 0;
 var $__ndptr$0 = 0, $__ndptr$0$us = 0, $__ndptr$011 = 0, $__ndptr$012 = 0, $__ndptr$012$us = 0, $__ndptr$1 = 0, $__rp$05$i$i1$i = 0, $__rp$05$i$i1$i$us = 0, $or$cond = 0, $scevgep$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($__nd)) + 8|0); //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $1 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = $1 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ($2<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ((($__nd)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = HEAP32[$4>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = $3 ? $6 : $5; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = ((($__nd)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = HEAP32[$8>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $1&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $10 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = $3 ? $11 : $9; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ($12>>>0)>(3); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($13) {
  $$03$i$i$i$i = $12;$__data$02$i$i$i$i = $7;$__h$01$i$i$i$i = $12;
  while(1) {
   $14 = HEAPU8[$__data$02$i$i$i$i>>0]|(HEAPU8[$__data$02$i$i$i$i+1>>0]<<8)|(HEAPU8[$__data$02$i$i$i$i+2>>0]<<16)|(HEAPU8[$__data$02$i$i$i$i+3>>0]<<24); //@line 3073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $15 = Math_imul($14, 1540483477)|0; //@line 3101 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $16 = $15 >>> 24; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $17 = $16 ^ $15; //@line 3102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $18 = Math_imul($17, 1540483477)|0; //@line 3103 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $19 = Math_imul($__h$01$i$i$i$i, 1540483477)|0; //@line 3104 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $20 = $18 ^ $19; //@line 3105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $21 = ((($__data$02$i$i$i$i)) + 4|0); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $22 = (($$03$i$i$i$i) + -4)|0; //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $23 = ($22>>>0)>(3); //@line 3098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if ($23) {
    $$03$i$i$i$i = $22;$__data$02$i$i$i$i = $21;$__h$01$i$i$i$i = $20;
   } else {
    $$lcssa = $20;
    break;
   }
  }
  $24 = (($12) + -4)|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $25 = $24 & -4; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $26 = (($25) + 4)|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $27 = (($24) - ($25))|0; //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $scevgep$i$i$i$i = (($7) + ($26)|0); //@line 3070 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $$0$lcssa$i$i$i$i = $27;$__data$0$lcssa$i$i$i$i = $scevgep$i$i$i$i;$__h$0$lcssa$i$i$i$i = $$lcssa;
 } else {
  $$0$lcssa$i$i$i$i = $12;$__data$0$lcssa$i$i$i$i = $7;$__h$0$lcssa$i$i$i$i = $12;
 }
 switch ($$0$lcssa$i$i$i$i|0) {
 case 3:  {
  $28 = ((($__data$0$lcssa$i$i$i$i)) + 2|0); //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $29 = HEAP8[$28>>0]|0; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $30 = $29&255; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $31 = $30 << 16; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $32 = $31 ^ $__h$0$lcssa$i$i$i$i; //@line 3110 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$1$i$i$i$i = $32;
  label = 6;
  break;
 }
 case 2:  {
  $__h$1$i$i$i$i = $__h$0$lcssa$i$i$i$i;
  label = 6;
  break;
 }
 case 1:  {
  $__h$2$i$i$i$i = $__h$0$lcssa$i$i$i$i;
  label = 7;
  break;
 }
 default: {
  $__h$3$i$i$i$i = $__h$0$lcssa$i$i$i$i;
 }
 }
 if ((label|0) == 6) {
  $33 = ((($__data$0$lcssa$i$i$i$i)) + 1|0); //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $34 = HEAP8[$33>>0]|0; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $35 = $34&255; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $36 = $35 << 8; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $37 = $36 ^ $__h$1$i$i$i$i; //@line 3112 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$2$i$i$i$i = $37;
  label = 7;
 }
 if ((label|0) == 7) {
  $38 = HEAP8[$__data$0$lcssa$i$i$i$i>>0]|0; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $39 = $38&255; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $40 = $39 ^ $__h$2$i$i$i$i; //@line 3114 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $41 = Math_imul($40, 1540483477)|0; //@line 3115 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $__h$3$i$i$i$i = $41;
 }
 $42 = $__h$3$i$i$i$i >>> 13; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $43 = $42 ^ $__h$3$i$i$i$i; //@line 3117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $44 = Math_imul($43, 1540483477)|0; //@line 3118 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $45 = $44 >>> 15; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $46 = $45 ^ $44; //@line 3119 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = ((($__nd)) + 4|0); //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$47>>2] = $46; //@line 1595 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $48 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $49 = HEAP32[$48>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $50 = ($49|0)==(0); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 L13: do {
  if ($50) {
   $__chash$0 = 0;
  } else {
   $51 = (($49) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $52 = $51 & $49; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $53 = ($52|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($53) {
    $54 = $46 & $51; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $58 = $54;
   } else {
    $55 = (($46>>>0) % ($49>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $58 = $55;
   }
   $56 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $57 = (($56) + ($58<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $59 = HEAP32[$57>>2]|0; //@line 1603 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $60 = ($59|0)==(0|0); //@line 1604 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($60) {
    $__chash$0 = $58;
   } else {
    $__ndptr$011 = HEAP32[$59>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $61 = ($__ndptr$011|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($61) {
     $__chash$0 = $58;
    } else {
     $62 = ($12|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($53) {
      $__ndptr$012$us = $__ndptr$011;
      L22: while(1) {
       $63 = ((($__ndptr$012$us)) + 4|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $64 = HEAP32[$63>>2]|0; //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $65 = $64 & $51; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $66 = ($65|0)==($58|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if (!($66)) {
        $__chash$0 = $58;
        break L13;
       }
       $67 = ((($__ndptr$012$us)) + 8|0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $68 = HEAP16[$67>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $69 = $68 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $70 = ($69<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $71 = ((($__ndptr$012$us)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $72 = HEAP32[$71>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$mask$i$i$us = $68 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $73 = $$mask$i$i$us&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $74 = $73 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $75 = $70 ? $74 : $72; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $76 = ($75|0)==($12|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $77 = ($68&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $78 = $77&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       do {
        if ($76) {
         if (!($70)) {
          $79 = ((($__ndptr$012$us)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $80 = HEAP32[$79>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $81 = (_memcmp($80,$7,$12)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $82 = ($81|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          if ($82) {
           $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012$us;
           label = 49;
           break L22;
          } else {
           break;
          }
         }
         if ($62) {
          $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012$us;
          label = 49;
          break L22;
         }
         $83 = HEAP8[$7>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $84 = ($78<<24>>24)==($83<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         if ($84) {
          $85 = ((($67)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $__lhs_sz$03$i$i3$i$us = $74;$__lp$04$i$i2$i$us = $85;$__rp$05$i$i1$i$us = $7;
          while(1) {
           $86 = (($__lhs_sz$03$i$i3$i$us) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $87 = ((($__lp$04$i$i2$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $88 = ($86|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($88) {
            $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012$us;
            label = 49;
            break L22;
           }
           $89 = ((($__rp$05$i$i1$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $$pre$i$i$us = HEAP8[$87>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $90 = HEAP8[$89>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $91 = ($$pre$i$i$us<<24>>24)==($90<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($91) {
            $__lhs_sz$03$i$i3$i$us = $86;$__lp$04$i$i2$i$us = $87;$__rp$05$i$i1$i$us = $89;
           } else {
            break;
           }
          }
         }
        }
       } while(0);
       $__ndptr$0$us = HEAP32[$__ndptr$012$us>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $92 = ($__ndptr$0$us|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($92) {
        $__chash$0 = $58;
        break L13;
       } else {
        $__ndptr$012$us = $__ndptr$0$us;
       }
      }
      if ((label|0) == 49) {
       $176 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       HEAP32[$agg$result>>2] = $176; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       $177 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       HEAP8[$177>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      }
     } else {
      $__ndptr$012 = $__ndptr$011;
      L40: while(1) {
       $93 = ((($__ndptr$012)) + 4|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $94 = HEAP32[$93>>2]|0; //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $95 = (($94>>>0) % ($49>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $96 = ($95|0)==($58|0); //@line 1607 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if (!($96)) {
        $__chash$0 = $58;
        break L13;
       }
       $97 = ((($__ndptr$012)) + 8|0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $98 = HEAP16[$97>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $99 = $98 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $100 = ($99<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $101 = ((($__ndptr$012)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $102 = HEAP32[$101>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $$mask$i$i = $98 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $103 = $$mask$i$i&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $104 = $103 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $105 = $100 ? $104 : $102; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $106 = ($105|0)==($12|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $107 = ($98&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $108 = $107&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       do {
        if ($106) {
         if (!($100)) {
          $118 = ((($__ndptr$012)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $119 = HEAP32[$118>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $120 = (_memcmp($119,$7,$12)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $121 = ($120|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          if ($121) {
           $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012;
           label = 49;
           break L40;
          } else {
           break;
          }
         }
         if ($62) {
          $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012;
          label = 49;
          break L40;
         }
         $109 = HEAP8[$7>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $110 = ($108<<24>>24)==($109<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         if ($110) {
          $111 = ((($97)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $__lhs_sz$03$i$i3$i = $104;$__lp$04$i$i2$i = $111;$__rp$05$i$i1$i = $7;
          while(1) {
           $112 = (($__lhs_sz$03$i$i3$i) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $113 = ((($__lp$04$i$i2$i)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $114 = ($112|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($114) {
            $__inserted$0 = 0;$__ndptr$1 = $__ndptr$012;
            label = 49;
            break L40;
           }
           $115 = ((($__rp$05$i$i1$i)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $$pre$i$i = HEAP8[$113>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $116 = HEAP8[$115>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $117 = ($$pre$i$i<<24>>24)==($116<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($117) {
            $__lhs_sz$03$i$i3$i = $112;$__lp$04$i$i2$i = $113;$__rp$05$i$i1$i = $115;
           } else {
            break;
           }
          }
         }
        }
       } while(0);
       $__ndptr$0 = HEAP32[$__ndptr$012>>2]|0; //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       $122 = ($__ndptr$0|0)==(0|0); //@line 1606 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       if ($122) {
        $__chash$0 = $58;
        break L13;
       } else {
        $__ndptr$012 = $__ndptr$0;
       }
      }
      if ((label|0) == 49) {
       $176 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
       HEAP32[$agg$result>>2] = $176; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       $177 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       HEAP8[$177>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
       return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      }
     }
    }
   }
  }
 } while(0);
 $123 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $124 = HEAP32[$123>>2]|0; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $125 = (($124) + 1)|0; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $126 = (+($125>>>0)); //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $127 = (+($49>>>0)); //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $128 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $129 = +HEAPF32[$128>>2]; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $130 = $127 * $129; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $131 = $126 > $130; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $or$cond = $50 | $131; //@line 1616 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 do {
  if ($or$cond) {
   $132 = $49 << 1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $133 = ($49>>>0)>(2); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($133) {
    $134 = (($49) + -1)|0; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $135 = $134 & $49; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $136 = ($135|0)==(0); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $138 = $136;
   } else {
    $138 = 0;
   }
   $137 = $138&1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $139 = $137 | $132; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $140 = $139 ^ 1; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $141 = $126 / $129; //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $142 = (+Math_ceil((+$141))); //@line 750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/cmath"
   $143 = (~~(($142))>>>0); //@line 1619 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $144 = ($140>>>0)<($143>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $145 = $144 ? $143 : $140; //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl11SpriteBatch17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE6rehashEj($this,$145); //@line 1618 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $146 = HEAP32[$48>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $147 = HEAP32[$47>>2]|0; //@line 1621 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $148 = (($146) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $149 = $148 & $146; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $150 = ($149|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($150) {
    $151 = $148 & $147; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__bc$0 = $146;$__chash$1 = $151;
    break;
   } else {
    $152 = (($147>>>0) % ($146>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__bc$0 = $146;$__chash$1 = $152;
    break;
   }
  } else {
   $__bc$0 = $49;$__chash$1 = $__chash$0;
  }
 } while(0);
 $153 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $154 = (($153) + ($__chash$1<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $155 = HEAP32[$154>>2]|0; //@line 1624 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $156 = ($155|0)==(0|0); //@line 1625 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($156) {
  $157 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $158 = HEAP32[$157>>2]|0; //@line 1628 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$__nd>>2] = $158; //@line 1628 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$157>>2] = $__nd; //@line 1629 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $159 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $160 = (($159) + ($__chash$1<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$160>>2] = $157; //@line 1631 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $161 = HEAP32[$__nd>>2]|0; //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $162 = ($161|0)==(0|0); //@line 1632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if (!($162)) {
   $163 = ((($161)) + 4|0); //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $164 = HEAP32[$163>>2]|0; //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $165 = (($__bc$0) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $166 = $165 & $__bc$0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $167 = ($166|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($167) {
    $168 = $164 & $165; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $172 = $168;
   } else {
    $169 = (($164>>>0) % ($__bc$0>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $172 = $169;
   }
   $170 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $171 = (($170) + ($172<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$171>>2] = $__nd; //@line 1633 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  }
 } else {
  $173 = HEAP32[$155>>2]|0; //@line 1637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$__nd>>2] = $173; //@line 1637 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  HEAP32[$155>>2] = $__nd; //@line 1638 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $174 = HEAP32[$123>>2]|0; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $175 = (($174) + 1)|0; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$123>>2] = $175; //@line 1642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__inserted$0 = 1;$__ndptr$1 = $__nd;
 $176 = $__ndptr$1; //@line 206 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 HEAP32[$agg$result>>2] = $176; //@line 306 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 $177 = ((($agg$result)) + 4|0); //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 HEAP8[$177>>0] = $__inserted$0; //@line 307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 return; //@line 1651 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl11SpriteBatch17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE6rehashEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__n|0)==(1); //@line 1937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($0) {
  $$0 = 2;
 } else {
  $1 = (($__n) + -1)|0; //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $2 = $1 & $__n; //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $3 = ($2|0)==(0); //@line 1939 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($3) {
   $$0 = $__n;
  } else {
   $4 = (__ZNSt3__112__next_primeEj($__n)|0); //@line 1940 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $$0 = $4;
  }
 }
 $5 = ((($this)) + 4|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = HEAP32[$5>>2]|0; //@line 706 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $7 = ($$0>>>0)>($6>>>0); //@line 1942 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($7) {
  __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl11SpriteBatch17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE8__rehashEj($this,$$0); //@line 1943 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $8 = ($$0>>>0)<($6>>>0); //@line 1944 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($8)) {
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $9 = ($6>>>0)>(2); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($9) {
  $10 = (($6) + -1)|0; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $11 = $10 & $6; //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $12 = ($11|0)==(0); //@line 66 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $30 = $12;
 } else {
  $30 = 0;
 }
 $13 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = HEAP32[$13>>2]|0; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $15 = (+($14>>>0)); //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $16 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $17 = +HEAPF32[$16>>2]; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $18 = $15 / $17; //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $19 = (+Math_ceil((+$18))); //@line 750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/cmath"
 $20 = (~~(($19))>>>0); //@line 1949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($30) {
  $21 = (($20) + -1)|0; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $22 = (Math_clz32(($21|0))|0); //@line 785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $23 = (32 - ($22))|0; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $24 = 1 << $23; //@line 80 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $24;
 } else {
  $25 = (__ZNSt3__112__next_primeEj($20)|0); //@line 1950 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $27 = $25;
 }
 $26 = ($$0>>>0)<($27>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $28 = $26 ? $27 : $$0; //@line 1946 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $29 = ($28>>>0)<($6>>>0); //@line 1952 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if (!($29)) {
  return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl11SpriteBatch17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE8__rehashEj($this,$28); //@line 1953 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 return; //@line 1955 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
}
function __ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl11SpriteBatch17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE8__rehashEj($this,$__nbc) {
 $this = $this|0;
 $__nbc = $__nbc|0;
 var $$lcssa = 0, $$lcssa113 = 0, $$mask$i$i = 0, $$pre$i$i$us = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__cp$0 = 0, $__cp$0$in2332 = 0, $__cp$0$in26 = 0, $__cp$018 = 0, $__cp$01824 = 0, $__cp$0182429 = 0, $__cp$0182433 = 0, $__cp$01827 = 0, $__cp$019 = 0, $__cp$019$lcssa = 0, $__cp$019$lcssa111 = 0, $__i$034 = 0, $__lhs_sz$03$i$i3$i$us = 0, $__lp$04$i$i2$i$us = 0, $__np$013 = 0, $__np$013$phi = 0;
 var $__np$013$us = 0, $__np$013$us$phi = 0, $__np$013$us36 = 0, $__np$013$us36$phi = 0, $__np$08 = 0, $__phash$0$ph$ph31 = 0, $__pp$0$ph$ph30 = 0, $__pp$0$ph25 = 0, $__pp$0$ph25$lcssa = 0, $__rp$05$i$i1$i$us = 0, $exitcond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = ($__nbc|0)==(0); //@line 1965 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($1) {
  $118 = HEAP32[$this>>2]|0; //@line 2822 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$this>>2] = 0; //@line 2823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $119 = ($118|0)==(0|0); //@line 2824 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  if (!($119)) {
   __ZdlPv($118); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  }
  HEAP32[$0>>2] = 0; //@line 1967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $2 = $__nbc << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = (__Znwj($2)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $4 = HEAP32[$this>>2]|0; //@line 2822 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $3; //@line 2823 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ($4|0)==(0|0); //@line 2824 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($5)) {
  __ZdlPv($4); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 HEAP32[$0>>2] = $__nbc; //@line 1967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__i$034 = 0;
 while(1) {
  $9 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $10 = (($9) + ($__i$034<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$10>>2] = 0; //@line 1971 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $11 = (($__i$034) + 1)|0; //@line 1970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $exitcond = ($11|0)==($__nbc|0); //@line 1970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($exitcond) {
   break;
  } else {
   $__i$034 = $11;
  }
 }
 $6 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $7 = HEAP32[$6>>2]|0; //@line 1973 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $8 = ($7|0)==(0|0); //@line 1974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($8) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
 $12 = ((($7)) + 4|0); //@line 1976 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $13 = HEAP32[$12>>2]|0; //@line 1976 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $14 = (($__nbc) + -1)|0; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $15 = $14 & $__nbc; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $16 = ($15|0)==(0); //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($16) {
  $17 = $13 & $14; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $21 = $17;
 } else {
  $18 = (($13>>>0) % ($__nbc>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $21 = $18;
 }
 $19 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $20 = (($19) + ($21<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$20>>2] = $6; //@line 1977 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $__cp$0182429 = HEAP32[$7>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 $22 = ($__cp$0182429|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 if ($22) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 } else {
  $__cp$0$in2332 = $7;$__cp$0182433 = $__cp$0182429;$__phash$0$ph$ph31 = $21;$__pp$0$ph$ph30 = $7;
 }
 L23: while(1) {
  $__cp$0$in26 = $__cp$0$in2332;$__cp$01827 = $__cp$0182433;$__pp$0$ph25 = $__pp$0$ph$ph30;
  L25: while(1) {
   $__cp$019 = $__cp$01827;
   while(1) {
    $24 = ((($__cp$019)) + 4|0); //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $25 = HEAP32[$24>>2]|0; //@line 1982 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($16) {
     $26 = $25 & $14; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $29 = $26;
    } else {
     $27 = (($25>>>0) % ($__nbc>>>0))&-1; //@line 73 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
     $29 = $27;
    }
    $28 = ($29|0)==($__phash$0$ph$ph31|0); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($28) {
     $__cp$019$lcssa = $__cp$019;
     break;
    }
    $30 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $31 = (($30) + ($29<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $32 = HEAP32[$31>>2]|0; //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $33 = ($32|0)==(0|0); //@line 1987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($33) {
     $$lcssa = $31;$$lcssa113 = $29;$__cp$019$lcssa111 = $__cp$019;$__pp$0$ph25$lcssa = $__pp$0$ph25;
     break L25;
    }
    $34 = HEAP32[$__cp$019>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $35 = ($34|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    L35: do {
     if ($35) {
      $__np$08 = $__cp$019;
     } else {
      $36 = ((($__cp$019)) + 8|0); //@line 1997 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
      $37 = HEAP16[$36>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $38 = $37 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $39 = ($38<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $40 = ((($__cp$019)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $41 = HEAP32[$40>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $$mask$i$i = $37 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $42 = $$mask$i$i&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $43 = $42 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $44 = $39 ? $43 : $41; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $45 = ($37&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $46 = $45&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $47 = ($44|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $48 = ((($36)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $49 = ((($__cp$019)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($39) {
       $51 = $34;$__np$013$us = $__cp$019;
       while(1) {
        $50 = ((($51)) + 8|0); //@line 1997 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $52 = HEAP8[$50>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $53 = $52 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $54 = ($53<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $55 = ((($51)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $56 = HEAP32[$55>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $57 = $52&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $58 = $57 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $59 = $54 ? $58 : $56; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $60 = ($44|0)==($59|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if (!($60)) {
         $__np$08 = $__np$013$us;
         break L35;
        }
        $61 = ((($51)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $62 = HEAP32[$61>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $63 = ((($50)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $64 = $54 ? $63 : $62; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        L41: do {
         if (!($47)) {
          $65 = HEAP8[$64>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $66 = ($46<<24>>24)==($65<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          if ($66) {
           $__lhs_sz$03$i$i3$i$us = $43;$__lp$04$i$i2$i$us = $48;$__rp$05$i$i1$i$us = $64;
          } else {
           $__np$08 = $__np$013$us;
           break L35;
          }
          while(1) {
           $67 = (($__lhs_sz$03$i$i3$i$us) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $68 = ((($__lp$04$i$i2$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $69 = ($67|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($69) {
            break L41;
           }
           $70 = ((($__rp$05$i$i1$i$us)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $$pre$i$i$us = HEAP8[$68>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $71 = HEAP8[$70>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $72 = ($$pre$i$i$us<<24>>24)==($71<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($72) {
            $__lhs_sz$03$i$i3$i$us = $67;$__lp$04$i$i2$i$us = $68;$__rp$05$i$i1$i$us = $70;
           } else {
            $__np$08 = $__np$013$us;
            break L35;
           }
          }
         }
        } while(0);
        $73 = HEAP32[$51>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $74 = ($73|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        if ($74) {
         $__np$08 = $51;
         break;
        } else {
         $__np$013$us$phi = $51;$51 = $73;$__np$013$us = $__np$013$us$phi;
        }
       }
      } else {
       if ($47) {
        $76 = $34;$__np$013$us36 = $__cp$019;
        while(1) {
         $75 = ((($76)) + 8|0); //@line 1997 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $77 = HEAP8[$75>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $78 = $77 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $79 = ($78<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $80 = ((($76)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $81 = HEAP32[$80>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $82 = $77&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $83 = $82 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $84 = $79 ? $83 : $81; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $85 = ($84|0)==(0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         if (!($85)) {
          $__np$08 = $__np$013$us36;
          break L35;
         }
         $86 = HEAP32[$76>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         $87 = ($86|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
         if ($87) {
          $__np$08 = $76;
          break L35;
         } else {
          $__np$013$us36$phi = $76;$76 = $86;$__np$013$us36 = $__np$013$us36$phi;
         }
        }
       } else {
        $90 = $34;$__np$013 = $__cp$019;
       }
       while(1) {
        $89 = ((($90)) + 8|0); //@line 1997 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $91 = HEAP8[$89>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $92 = $91 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $93 = ($92<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $94 = ((($90)) + 12|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $95 = HEAP32[$94>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $96 = $91&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $97 = $96 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $98 = $93 ? $97 : $95; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $99 = ($44|0)==($98|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if (!($99)) {
         $__np$08 = $__np$013;
         break L35;
        }
        $100 = ((($90)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $101 = HEAP32[$100>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $102 = ((($89)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $103 = $93 ? $102 : $101; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $104 = HEAP32[$49>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $105 = (_memcmp($104,$103,$44)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $106 = ($105|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if (!($106)) {
         $__np$08 = $__np$013;
         break L35;
        }
        $107 = HEAP32[$90>>2]|0; //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        $108 = ($107|0)==(0|0); //@line 1996 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
        if ($108) {
         $__np$08 = $90;
         break;
        } else {
         $__np$013$phi = $90;$90 = $107;$__np$013 = $__np$013$phi;
        }
       }
      }
     }
    } while(0);
    $109 = HEAP32[$__np$08>>2]|0; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    HEAP32[$__pp$0$ph25>>2] = $109; //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $110 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $111 = (($110) + ($29<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $112 = HEAP32[$111>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $113 = HEAP32[$112>>2]|0; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    HEAP32[$__np$08>>2] = $113; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $114 = HEAP32[$this>>2]|0; //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $115 = (($114) + ($29<<2)|0); //@line 2799 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $116 = HEAP32[$115>>2]|0; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    HEAP32[$116>>2] = $__cp$019; //@line 2002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $__cp$0 = HEAP32[$__cp$0$in26>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    $117 = ($__cp$0|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
    if ($117) {
     label = 38;
     break L23;
    } else {
     $__cp$019 = $__cp$0;
    }
   }
   $__cp$018 = HEAP32[$__cp$019$lcssa>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   $23 = ($__cp$018|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
   if ($23) {
    label = 38;
    break L23;
   } else {
    $__cp$0$in26 = $__cp$019$lcssa;$__cp$01827 = $__cp$018;$__pp$0$ph25 = $__cp$019$lcssa;
   }
  }
  HEAP32[$$lcssa>>2] = $__pp$0$ph25$lcssa; //@line 1989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $__cp$01824 = HEAP32[$__cp$019$lcssa111>>2]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  $88 = ($__cp$01824|0)==(0|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
  if ($88) {
   label = 38;
   break;
  } else {
   $__cp$0$in2332 = $__cp$019$lcssa111;$__cp$0182433 = $__cp$01824;$__phash$0$ph$ph31 = $$lcssa113;$__pp$0$ph$ph30 = $__cp$019$lcssa111;
  }
 }
 if ((label|0) == 38) {
  return; //@line 2009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__hash_table"
 }
}
function __ZN4love8graphics6opengl4TextD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (25456); //@line 43 "modules/graphics/opengl/Text.cpp"
 $0 = ((($this)) + 12|0); //@line 44 "modules/graphics/opengl/Text.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 44 "modules/graphics/opengl/Text.cpp"
 $2 = ($1|0)==(0|0); //@line 44 "modules/graphics/opengl/Text.cpp"
 if (!($2)) {
  $3 = HEAP32[$1>>2]|0; //@line 44 "modules/graphics/opengl/Text.cpp"
  $4 = ((($3)) + 4|0); //@line 44 "modules/graphics/opengl/Text.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 44 "modules/graphics/opengl/Text.cpp"
  FUNCTION_TABLE_vi[$5 & 8191]($1); //@line 44 "modules/graphics/opengl/Text.cpp"
 }
 $6 = ((($this)) + 28|0); //@line 45 "modules/graphics/opengl/Text.cpp"
 __ZNSt3__113__vector_baseIN4love8graphics6opengl4Text8TextDataENS_9allocatorIS5_EEED2Ev($6); //@line 45 "modules/graphics/opengl/Text.cpp"
 $7 = ((($this)) + 16|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = HEAP32[$7>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ($8|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $8; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($9)) {
  $11 = ((($this)) + 20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $12 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = ($12|0)==($8|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($13)) {
   $scevgep$i$i$i = ((($12)) + -12|0);
   $14 = $scevgep$i$i$i;
   $15 = (($14) - ($10))|0;
   $16 = (($15>>>0) / 12)&-1;
   $17 = $16 ^ -1;
   $scevgep4$i$i$i = (($12) + (($17*12)|0)|0);
   HEAP32[$11>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($8); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $18 = ((($this)) + 8|0); //@line 106 "./common/Object.h"
 $19 = HEAP32[$18>>2]|0; //@line 106 "./common/Object.h"
 $20 = ($19|0)==(0|0); //@line 106 "./common/Object.h"
 if ($20) {
  return; //@line 45 "modules/graphics/opengl/Text.cpp"
 }
 $21 = HEAP32[$19>>2]|0; //@line 106 "./common/Object.h"
 $22 = ((($21)) + 12|0); //@line 106 "./common/Object.h"
 $23 = HEAP32[$22>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($23|0,($19|0)); //@line 106 "./common/Object.h"
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $26 = ___cxa_find_matching_catch_3(0|0)|0;
  $27 = tempRet0;
  ___clang_call_terminate($26); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  return; //@line 45 "modules/graphics/opengl/Text.cpp"
 }
}
function __ZNSt3__113__vector_baseIN4love8graphics6opengl4Text8TextDataENS_9allocatorIS5_EEED2Ev($this) {
 $this = $this|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__new_last1$i$i$i$i$i$i$i$i$i$i = 0, $scevgep$i$i$i$i$i$i$i$i$i$i = 0, $scevgep$i$i$i1$i$i$i$i$i$i$i = 0, $scevgep4$i$i$i$i$i$i$i$i$i$i = 0, $scevgep4$i$i$i2$i$i$i$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ($0|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($1) {
  return; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $2 = ((($this)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($3|0)==($0|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($4) {
  $30 = $0;
 } else {
  $6 = $3;
  while(1) {
   $5 = ((($6)) + -80|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$2>>2] = $5; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $7 = ((($6)) + -36|0); //@line 70 "modules/graphics/opengl/Text.h"
   __ZN4love7Matrix3D2Ev($7); //@line 70 "modules/graphics/opengl/Text.h"
   $8 = ((($6)) + -68|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $9 = HEAP32[$8>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $10 = ($9|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($10)) {
    $11 = ((($6)) + -64|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $12 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $13 = ($12|0)==($9|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($13)) {
     $__new_last1$i$i$i$i$i$i$i$i$i$i = $9; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $scevgep$i$i$i$i$i$i$i$i$i$i = ((($12)) + -8|0);
     $14 = $scevgep$i$i$i$i$i$i$i$i$i$i;
     $15 = (($14) - ($__new_last1$i$i$i$i$i$i$i$i$i$i))|0;
     $16 = $15 >>> 3;
     $17 = $16 ^ -1;
     $scevgep4$i$i$i$i$i$i$i$i$i$i = (($12) + ($17<<3)|0);
     HEAP32[$11>>2] = $scevgep4$i$i$i$i$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($9); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $18 = HEAP32[$5>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $19 = ($18|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $20 = $18; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($19)) {
    $23 = ((($6)) + -76|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $24 = HEAP32[$23>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $25 = ($24|0)==($18|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($25)) {
     $scevgep$i$i$i1$i$i$i$i$i$i$i = ((($24)) + -4|0);
     $26 = $scevgep$i$i$i1$i$i$i$i$i$i$i;
     $27 = (($26) - ($20))|0;
     $28 = $27 >>> 2;
     $29 = $28 ^ -1;
     $scevgep4$i$i$i2$i$i$i$i$i$i$i = (($24) + ($29<<2)|0);
     HEAP32[$23>>2] = $scevgep4$i$i$i2$i$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($18); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $21 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $22 = ($21|0)==($0|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($22) {
    break;
   } else {
    $6 = $21;
   }
  }
  $$pre = HEAP32[$this>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $30 = $$pre;
 }
 __ZdlPv($30); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8graphics6opengl4TextD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i$i = 0, $scevgep4$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (25456); //@line 43 "modules/graphics/opengl/Text.cpp"
 $0 = ((($this)) + 12|0); //@line 44 "modules/graphics/opengl/Text.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 44 "modules/graphics/opengl/Text.cpp"
 $2 = ($1|0)==(0|0); //@line 44 "modules/graphics/opengl/Text.cpp"
 if (!($2)) {
  $3 = HEAP32[$1>>2]|0; //@line 44 "modules/graphics/opengl/Text.cpp"
  $4 = ((($3)) + 4|0); //@line 44 "modules/graphics/opengl/Text.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 44 "modules/graphics/opengl/Text.cpp"
  FUNCTION_TABLE_vi[$5 & 8191]($1); //@line 44 "modules/graphics/opengl/Text.cpp"
 }
 $6 = ((($this)) + 28|0); //@line 45 "modules/graphics/opengl/Text.cpp"
 __ZNSt3__113__vector_baseIN4love8graphics6opengl4Text8TextDataENS_9allocatorIS5_EEED2Ev($6); //@line 45 "modules/graphics/opengl/Text.cpp"
 $7 = ((($this)) + 16|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = HEAP32[$7>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ($8|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $8; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($9)) {
  $11 = ((($this)) + 20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $12 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = ($12|0)==($8|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($13)) {
   $scevgep$i$i$i$i = ((($12)) + -12|0);
   $14 = $scevgep$i$i$i$i;
   $15 = (($14) - ($10))|0;
   $16 = (($15>>>0) / 12)&-1;
   $17 = $16 ^ -1;
   $scevgep4$i$i$i$i = (($12) + (($17*12)|0)|0);
   HEAP32[$11>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($8); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $18 = ((($this)) + 8|0); //@line 106 "./common/Object.h"
 $19 = HEAP32[$18>>2]|0; //@line 106 "./common/Object.h"
 $20 = ($19|0)==(0|0); //@line 106 "./common/Object.h"
 if ($20) {
  __ZdlPv($this); //@line 43 "modules/graphics/opengl/Text.cpp"
  return; //@line 45 "modules/graphics/opengl/Text.cpp"
 }
 $21 = HEAP32[$19>>2]|0; //@line 106 "./common/Object.h"
 $22 = ((($21)) + 12|0); //@line 106 "./common/Object.h"
 $23 = HEAP32[$22>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($23|0,($19|0)); //@line 106 "./common/Object.h"
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $26 = ___cxa_find_matching_catch_3(0|0)|0;
  $27 = tempRet0;
  ___clang_call_terminate($26); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  __ZdlPv($this); //@line 43 "modules/graphics/opengl/Text.cpp"
  return; //@line 45 "modules/graphics/opengl/Text.cpp"
 }
}
function __ZN4love8graphics6opengl4Text4drawEfffffffff($this,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = ((($this)) + 12|0); //@line 219 "modules/graphics/opengl/Text.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 219 "modules/graphics/opengl/Text.cpp"
 $3 = ($2|0)==(0|0); //@line 219 "modules/graphics/opengl/Text.cpp"
 if ($3) {
  STACKTOP = sp;return; //@line 249 "modules/graphics/opengl/Text.cpp"
 }
 $4 = ((($this)) + 16|0); //@line 219 "modules/graphics/opengl/Text.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ((($this)) + 20|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = HEAP32[$6>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ($5|0)==($7|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($8) {
  STACKTOP = sp;return; //@line 249 "modules/graphics/opengl/Text.cpp"
 }
 $9 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $10 = HEAP32[$9>>2]|0; //@line 117 "./common/Object.h"
 $11 = (__ZNK4love8graphics6opengl4Font17getTextureCacheIDEv($10)|0); //@line 225 "modules/graphics/opengl/Text.cpp"
 $12 = ((($this)) + 44|0); //@line 225 "modules/graphics/opengl/Text.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 225 "modules/graphics/opengl/Text.cpp"
 $14 = ($11|0)==($13|0); //@line 225 "modules/graphics/opengl/Text.cpp"
 if (!($14)) {
  __ZN4love8graphics6opengl4Text18regenerateVerticesEv($this); //@line 226 "modules/graphics/opengl/Text.cpp"
 }
 __ZN4love8graphics6opengl6OpenGL13pushTransformEv(634044); //@line 125 "modules/graphics/opengl/OpenGL.h"
 __THREW__ = 0;
 $15 = (invoke_ii(4003,(634044|0))|0); //@line 135 "modules/graphics/opengl/OpenGL.h"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 do {
  if ($17) {
   label = 23;
  } else {
   __THREW__ = 0;
   invoke_viddddddddd(4039,($0|0),(+$x),(+$y),(+$angle),(+$sx),(+$sy),(+$ox),(+$oy),(+$kx),(+$ky)); //@line 234 "modules/graphics/opengl/Text.cpp"
   $18 = __THREW__; __THREW__ = 0;
   $19 = $18&1;
   if ($19) {
    label = 23;
   } else {
    __THREW__ = 0;
    invoke_vii(4004,($15|0),($0|0)); //@line 234 "modules/graphics/opengl/Text.cpp"
    $20 = __THREW__; __THREW__ = 0;
    $21 = $20&1;
    if ($21) {
     $64 = ___cxa_find_matching_catch_2()|0;
     $65 = tempRet0;
     __ZN4love7Matrix4D2Ev($0); //@line 234 "modules/graphics/opengl/Text.cpp"
     $$0 = $65;$$01 = $64;
     break;
    }
    __ZN4love7Matrix4D2Ev($0); //@line 234 "modules/graphics/opengl/Text.cpp"
    $22 = HEAP32[$1>>2]|0; //@line 237 "modules/graphics/opengl/Text.cpp"
    __THREW__ = 0;
    invoke_vi(4087,($22|0)); //@line 188 "modules/graphics/opengl/GLBuffer.h"
    $23 = __THREW__; __THREW__ = 0;
    $24 = $23&1;
    if ($24) {
     label = 23;
    } else {
     $25 = HEAP32[$1>>2]|0; //@line 238 "modules/graphics/opengl/Text.cpp"
     __THREW__ = 0;
     invoke_vi(4088,($25|0)); //@line 238 "modules/graphics/opengl/Text.cpp"
     $26 = __THREW__; __THREW__ = 0;
     $27 = $26&1;
     if (!($27)) {
      $28 = HEAP32[$1>>2]|0; //@line 241 "modules/graphics/opengl/Text.cpp"
      __THREW__ = 0;
      $29 = (invoke_iii(4090,($28|0),0)|0); //@line 241 "modules/graphics/opengl/Text.cpp"
      $30 = __THREW__; __THREW__ = 0;
      $31 = $30&1;
      if (!($31)) {
       $32 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
       __THREW__ = 0;
       invoke_viiiiii($32|0,0,2,5126,0,16,($29|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
       $33 = __THREW__; __THREW__ = 0;
       $34 = $33&1;
       if (!($34)) {
        $35 = HEAP32[$1>>2]|0; //@line 242 "modules/graphics/opengl/Text.cpp"
        __THREW__ = 0;
        $36 = (invoke_iii(4090,($35|0),8)|0); //@line 242 "modules/graphics/opengl/Text.cpp"
        $37 = __THREW__; __THREW__ = 0;
        $38 = $37&1;
        if (!($38)) {
         $39 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
         __THREW__ = 0;
         invoke_viiiiii($39|0,1,2,5123,1,16,($36|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
         $40 = __THREW__; __THREW__ = 0;
         $41 = $40&1;
         if (!($41)) {
          $42 = HEAP32[$1>>2]|0; //@line 243 "modules/graphics/opengl/Text.cpp"
          __THREW__ = 0;
          $43 = (invoke_iii(4090,($42|0),12)|0); //@line 243 "modules/graphics/opengl/Text.cpp"
          $44 = __THREW__; __THREW__ = 0;
          $45 = $44&1;
          if (!($45)) {
           $46 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
           __THREW__ = 0;
           invoke_viiiiii($46|0,2,4,5121,1,16,($43|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
           $47 = __THREW__; __THREW__ = 0;
           $48 = $47&1;
           if (!($48)) {
            __THREW__ = 0;
            invoke_vi(4089,($22|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
            $49 = __THREW__; __THREW__ = 0;
            $50 = $49&1;
            if ($50) {
             $51 = ___cxa_find_matching_catch_3(0|0)|0;
             $52 = tempRet0;
             ___clang_call_terminate($51); //@line 196 "modules/graphics/opengl/GLBuffer.h"
             // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
            }
            __THREW__ = 0;
            invoke_vii(4006,(634044|0),7); //@line 246 "modules/graphics/opengl/Text.cpp"
            $53 = __THREW__; __THREW__ = 0;
            $54 = $53&1;
            if ($54) {
             label = 23;
             break;
            }
            $55 = HEAP32[$9>>2]|0; //@line 117 "./common/Object.h"
            __THREW__ = 0;
            invoke_viii(4037,($55|0),($4|0),1); //@line 248 "modules/graphics/opengl/Text.cpp"
            $56 = __THREW__; __THREW__ = 0;
            $57 = $56&1;
            if ($57) {
             label = 23;
             break;
            }
            __THREW__ = 0;
            invoke_vi(4009,(634044|0)); //@line 130 "modules/graphics/opengl/OpenGL.h"
            $58 = __THREW__; __THREW__ = 0;
            $59 = $58&1;
            if ($59) {
             $60 = ___cxa_find_matching_catch_3(0|0)|0;
             $61 = tempRet0;
             ___clang_call_terminate($60); //@line 130 "modules/graphics/opengl/OpenGL.h"
             // unreachable; //@line 130 "modules/graphics/opengl/OpenGL.h"
            } else {
             STACKTOP = sp;return; //@line 249 "modules/graphics/opengl/Text.cpp"
            }
           }
          }
         }
        }
       }
      }
     }
     $66 = ___cxa_find_matching_catch_2()|0;
     $67 = tempRet0;
     __THREW__ = 0;
     invoke_vi(4089,($22|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
     $68 = __THREW__; __THREW__ = 0;
     $69 = $68&1;
     if ($69) {
      $70 = ___cxa_find_matching_catch_3(0|0)|0;
      $71 = tempRet0;
      ___clang_call_terminate($70); //@line 196 "modules/graphics/opengl/GLBuffer.h"
      // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
     } else {
      $$0 = $67;$$01 = $66;
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 23) {
  $62 = ___cxa_find_matching_catch_2()|0;
  $63 = tempRet0;
  $$0 = $63;$$01 = $62;
 }
 __THREW__ = 0;
 invoke_vi(4009,(634044|0)); //@line 130 "modules/graphics/opengl/OpenGL.h"
 $72 = __THREW__; __THREW__ = 0;
 $73 = $72&1;
 if ($73) {
  $74 = ___cxa_find_matching_catch_3(0|0)|0;
  $75 = tempRet0;
  ___clang_call_terminate($74); //@line 130 "modules/graphics/opengl/OpenGL.h"
  // unreachable; //@line 130 "modules/graphics/opengl/OpenGL.h"
 } else {
  ___resumeException($$01|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl4Text18regenerateVerticesEv($this) {
 $this = $this|0;
 var $$pre$i = 0, $$pre$i14 = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $9 = 0, $__begin$sroa$0$018 = 0, $__new_last1$i$i$i$i$i$i$i$i$i$i$i = 0, $__new_last1$i$i$i$i$i$i$i$i$i$i$i2 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $scevgep$i$i$i$i$i$i$i$i$i$i$i = 0, $scevgep$i$i$i$i$i$i$i$i$i$i$i3 = 0, $scevgep$i$i$i1$i$i$i$i$i$i$i$i = 0, $scevgep$i$i$i1$i$i$i$i$i$i$i$i9 = 0, $scevgep4$i$i$i$i$i$i$i$i$i$i$i = 0, $scevgep4$i$i$i$i$i$i$i$i$i$i$i4 = 0;
 var $scevgep4$i$i$i2$i$i$i$i$i$i$i$i = 0, $scevgep4$i$i$i2$i$i$i$i$i$i$i$i10 = 0, $textdata = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $textdata = sp;
 $0 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = (__ZNK4love8graphics6opengl4Font17getTextureCacheIDEv($1)|0); //@line 97 "modules/graphics/opengl/Text.cpp"
 $3 = ((($this)) + 44|0); //@line 97 "modules/graphics/opengl/Text.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 97 "modules/graphics/opengl/Text.cpp"
 $5 = ($2|0)==($4|0); //@line 97 "modules/graphics/opengl/Text.cpp"
 if ($5) {
  STACKTOP = sp;return; //@line 108 "modules/graphics/opengl/Text.cpp"
 }
 $6 = ((($this)) + 28|0); //@line 99 "modules/graphics/opengl/Text.cpp"
 __ZNSt3__16vectorIN4love8graphics6opengl4Text8TextDataENS_9allocatorIS5_EEEC2ERKS8_($textdata,$6); //@line 99 "modules/graphics/opengl/Text.cpp"
 __THREW__ = 0;
 invoke_vi(3622,($this|0)); //@line 101 "modules/graphics/opengl/Text.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 L4: do {
  if ($8) {
   label = 6;
  } else {
   $9 = HEAP32[$textdata>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $10 = ((($textdata)) + 4|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $11 = HEAP32[$10>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $12 = ($9|0)==($11|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   L6: do {
    if (!($12)) {
     $__begin$sroa$0$018 = $9;
     while(1) {
      __THREW__ = 0;
      invoke_vii(4143,($this|0),($__begin$sroa$0$018|0)); //@line 104 "modules/graphics/opengl/Text.cpp"
      $52 = __THREW__; __THREW__ = 0;
      $53 = $52&1;
      if ($53) {
       break;
      }
      $54 = ((($__begin$sroa$0$018)) + 80|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $55 = ($54|0)==($11|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($55) {
       break L6;
      } else {
       $__begin$sroa$0$018 = $54;
      }
     }
     $17 = ___cxa_find_matching_catch_2()|0;
     $18 = tempRet0;
     $$pre$phiZ2D = $textdata;$lpad$phi$index = $17;$lpad$phi$index2 = $18;
     break L4;
    }
   } while(0);
   $13 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
   __THREW__ = 0;
   $14 = (invoke_ii(4142,($13|0))|0); //@line 106 "modules/graphics/opengl/Text.cpp"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if ($16) {
    label = 6;
   } else {
    HEAP32[$3>>2] = $14; //@line 106 "modules/graphics/opengl/Text.cpp"
    $56 = HEAP32[$textdata>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $57 = ($56|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($57)) {
     $58 = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $59 = ($58|0)==($56|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($59) {
      $85 = $56;
     } else {
      $61 = $58;
      while(1) {
       $60 = ((($61)) + -80|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$10>>2] = $60; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $62 = ((($61)) + -36|0); //@line 70 "modules/graphics/opengl/Text.h"
       __ZN4love7Matrix3D2Ev($62); //@line 70 "modules/graphics/opengl/Text.h"
       $63 = ((($61)) + -68|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $64 = HEAP32[$63>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $65 = ($64|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($65)) {
        $66 = ((($61)) + -64|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $67 = HEAP32[$66>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $68 = ($67|0)==($64|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($68)) {
         $__new_last1$i$i$i$i$i$i$i$i$i$i$i2 = $64; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $scevgep$i$i$i$i$i$i$i$i$i$i$i3 = ((($67)) + -8|0);
         $69 = $scevgep$i$i$i$i$i$i$i$i$i$i$i3;
         $70 = (($69) - ($__new_last1$i$i$i$i$i$i$i$i$i$i$i2))|0;
         $71 = $70 >>> 3;
         $72 = $71 ^ -1;
         $scevgep4$i$i$i$i$i$i$i$i$i$i$i4 = (($67) + ($72<<3)|0);
         HEAP32[$66>>2] = $scevgep4$i$i$i$i$i$i$i$i$i$i$i4; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
        __ZdlPv($64); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       }
       $73 = HEAP32[$60>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $74 = ($73|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $75 = $73; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($74)) {
        $78 = ((($61)) + -76|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $79 = HEAP32[$78>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $80 = ($79|0)==($73|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($80)) {
         $scevgep$i$i$i1$i$i$i$i$i$i$i$i9 = ((($79)) + -4|0);
         $81 = $scevgep$i$i$i1$i$i$i$i$i$i$i$i9;
         $82 = (($81) - ($75))|0;
         $83 = $82 >>> 2;
         $84 = $83 ^ -1;
         $scevgep4$i$i$i2$i$i$i$i$i$i$i$i10 = (($79) + ($84<<2)|0);
         HEAP32[$78>>2] = $scevgep4$i$i$i2$i$i$i$i$i$i$i$i10; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
        __ZdlPv($73); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       }
       $76 = HEAP32[$10>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $77 = ($76|0)==($56|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($77) {
        break;
       } else {
        $61 = $76;
       }
      }
      $$pre$i14 = HEAP32[$textdata>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $85 = $$pre$i14;
     }
     __ZdlPv($85); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    }
    STACKTOP = sp;return; //@line 108 "modules/graphics/opengl/Text.cpp"
   }
  }
 } while(0);
 if ((label|0) == 6) {
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  $$pre$phiZ2D = $textdata;$lpad$phi$index = $19;$lpad$phi$index2 = $20;
 }
 $21 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $22 = ($21|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($22) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $23 = ((($textdata)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $24 = HEAP32[$23>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $25 = ($24|0)==($21|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($25) {
  $51 = $21;
 } else {
  $27 = $24;
  while(1) {
   $26 = ((($27)) + -80|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$23>>2] = $26; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $28 = ((($27)) + -36|0); //@line 70 "modules/graphics/opengl/Text.h"
   __ZN4love7Matrix3D2Ev($28); //@line 70 "modules/graphics/opengl/Text.h"
   $29 = ((($27)) + -68|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $30 = HEAP32[$29>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $31 = ($30|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($31)) {
    $32 = ((($27)) + -64|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $33 = HEAP32[$32>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $34 = ($33|0)==($30|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($34)) {
     $__new_last1$i$i$i$i$i$i$i$i$i$i$i = $30; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $scevgep$i$i$i$i$i$i$i$i$i$i$i = ((($33)) + -8|0);
     $35 = $scevgep$i$i$i$i$i$i$i$i$i$i$i;
     $36 = (($35) - ($__new_last1$i$i$i$i$i$i$i$i$i$i$i))|0;
     $37 = $36 >>> 3;
     $38 = $37 ^ -1;
     $scevgep4$i$i$i$i$i$i$i$i$i$i$i = (($33) + ($38<<3)|0);
     HEAP32[$32>>2] = $scevgep4$i$i$i$i$i$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($30); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $39 = HEAP32[$26>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $40 = ($39|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $41 = $39; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($40)) {
    $44 = ((($27)) + -76|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $45 = HEAP32[$44>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $46 = ($45|0)==($39|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($46)) {
     $scevgep$i$i$i1$i$i$i$i$i$i$i$i = ((($45)) + -4|0);
     $47 = $scevgep$i$i$i1$i$i$i$i$i$i$i$i;
     $48 = (($47) - ($41))|0;
     $49 = $48 >>> 2;
     $50 = $49 ^ -1;
     $scevgep4$i$i$i2$i$i$i$i$i$i$i$i = (($45) + ($50<<2)|0);
     HEAP32[$44>>2] = $scevgep4$i$i$i2$i$i$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($39); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $42 = HEAP32[$23>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $43 = ($42|0)==($21|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($43) {
    break;
   } else {
    $27 = $42;
   }
  }
  $$pre$i = HEAP32[$textdata>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $51 = $$pre$i;
 }
 __ZdlPv($51); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love8graphics6opengl4Text8TextDataENS_9allocatorIS5_EEEC2ERKS8_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$01$i = 0, $$lcssa13 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index4Z2D = 0, $eh$lpad$body$indexZ2D = 0, $scevgep$i$i$i$i$i$i$i$i$i = 0, $scevgep4$i$i$i$i$i$i$i$i$i = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $0 = ((($this)) + 4|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($this)) + 8|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($__x)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$__x>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = (($5|0) / 80)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ($3|0)==($4|0); //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($7) {
  return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $8 = ($6>>>0)>(53687091); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($8) {
  __THREW__ = 0;
  invoke_vi(3528,($this|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if (!($10)) {
   label = 4;
  }
 } else {
  label = 4;
 }
 if ((label|0) == 4) {
  __THREW__ = 0;
  $11 = (invoke_ii(3348,($5|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   HEAP32[$0>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$this>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $14 = (($11) + (($6*80)|0)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $15 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$15>>2] = $14; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $16 = HEAP32[$__x>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $17 = HEAP32[$2>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $18 = ($16|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($18) {
    return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $$01$i = $16;$19 = $11;
   while(1) {
    __THREW__ = 0;
    invoke_vii(4032,($19|0),($$01$i|0)); //@line 76 "modules/graphics/opengl/Font.h"
    $20 = __THREW__; __THREW__ = 0;
    $21 = $20&1;
    if ($21) {
     label = 14;
     break;
    }
    $22 = ((($19)) + 12|0); //@line 76 "modules/graphics/opengl/Font.h"
    $23 = ((($$01$i)) + 12|0); //@line 76 "modules/graphics/opengl/Font.h"
    __THREW__ = 0;
    invoke_vii(4035,($22|0),($23|0)); //@line 76 "modules/graphics/opengl/Font.h"
    $24 = __THREW__; __THREW__ = 0;
    $25 = $24&1;
    if ($25) {
     $$lcssa13 = $19;
     label = 9;
     break;
    }
    $38 = ((($19)) + 24|0); //@line 70 "modules/graphics/opengl/Text.h"
    $39 = ((($$01$i)) + 24|0); //@line 70 "modules/graphics/opengl/Text.h"
    dest=$38; src=$39; stop=dest+56|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 70 "modules/graphics/opengl/Text.h"
    $40 = HEAP32[$0>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $41 = ((($40)) + 80|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$0>>2] = $41; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $42 = ((($$01$i)) + 80|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $43 = ($42|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($43) {
     label = 17;
     break;
    } else {
     $$01$i = $42;$19 = $41;
    }
   }
   if ((label|0) == 9) {
    $26 = ___cxa_find_matching_catch_2()|0;
    $27 = tempRet0;
    $28 = HEAP32[$$lcssa13>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $29 = ($28|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $30 = $28; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($29) {
     $eh$lpad$body$index4Z2D = $27;$eh$lpad$body$indexZ2D = $26;
     __ZNSt3__113__vector_baseIN4love8graphics6opengl4Text8TextDataENS_9allocatorIS5_EEED2Ev($this); //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     ___resumeException($eh$lpad$body$indexZ2D|0);
     // unreachable;
    }
    $31 = ((($$lcssa13)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $32 = HEAP32[$31>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $33 = ($32|0)==($28|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($33)) {
     $scevgep$i$i$i$i$i$i$i$i$i = ((($32)) + -4|0);
     $34 = $scevgep$i$i$i$i$i$i$i$i$i;
     $35 = (($34) - ($30))|0;
     $36 = $35 >>> 2;
     $37 = $36 ^ -1;
     $scevgep4$i$i$i$i$i$i$i$i$i = (($32) + ($37<<2)|0);
     HEAP32[$31>>2] = $scevgep4$i$i$i$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($28); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $eh$lpad$body$index4Z2D = $27;$eh$lpad$body$indexZ2D = $26;
    __ZNSt3__113__vector_baseIN4love8graphics6opengl4Text8TextDataENS_9allocatorIS5_EEED2Ev($this); //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
   else if ((label|0) == 14) {
    $44 = ___cxa_find_matching_catch_2()|0;
    $45 = tempRet0;
    $eh$lpad$body$index4Z2D = $45;$eh$lpad$body$indexZ2D = $44;
    __ZNSt3__113__vector_baseIN4love8graphics6opengl4Text8TextDataENS_9allocatorIS5_EEED2Ev($this); //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
   else if ((label|0) == 17) {
    return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
  }
 }
 $46 = ___cxa_find_matching_catch_2()|0;
 $47 = tempRet0;
 $eh$lpad$body$index4Z2D = $47;$eh$lpad$body$indexZ2D = $46;
 __ZNSt3__113__vector_baseIN4love8graphics6opengl4Text8TextDataENS_9allocatorIS5_EEED2Ev($this); //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
}
function __ZN4love8graphics6opengl4Text5clearEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__new_last1$i$i$i$i$i$i$i$i$i$i$i = 0, $__new_last2$i$i$i = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i$i$i$i$i$i$i$i$i = 0, $scevgep$i$i$i1$i$i$i$i$i$i$i$i = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i$i$i$i$i$i$i$i$i = 0, $scevgep4$i$i$i2$i$i$i$i$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($this)) + 32|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($3|0)==($1|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($4)) {
  $6 = $3;
  while(1) {
   $5 = ((($6)) + -80|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$2>>2] = $5; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $7 = ((($6)) + -36|0); //@line 70 "modules/graphics/opengl/Text.h"
   __ZN4love7Matrix3D2Ev($7); //@line 70 "modules/graphics/opengl/Text.h"
   $8 = ((($6)) + -68|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $9 = HEAP32[$8>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $10 = ($9|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($10)) {
    $11 = ((($6)) + -64|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $12 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $13 = ($12|0)==($9|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($13)) {
     $__new_last1$i$i$i$i$i$i$i$i$i$i$i = $9; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $scevgep$i$i$i$i$i$i$i$i$i$i$i = ((($12)) + -8|0);
     $14 = $scevgep$i$i$i$i$i$i$i$i$i$i$i;
     $15 = (($14) - ($__new_last1$i$i$i$i$i$i$i$i$i$i$i))|0;
     $16 = $15 >>> 3;
     $17 = $16 ^ -1;
     $scevgep4$i$i$i$i$i$i$i$i$i$i$i = (($12) + ($17<<3)|0);
     HEAP32[$11>>2] = $scevgep4$i$i$i$i$i$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($9); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $18 = HEAP32[$5>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $19 = ($18|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $20 = $18; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($19)) {
    $23 = ((($6)) + -76|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $24 = HEAP32[$23>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $25 = ($24|0)==($18|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($25)) {
     $scevgep$i$i$i1$i$i$i$i$i$i$i$i = ((($24)) + -4|0);
     $26 = $scevgep$i$i$i1$i$i$i$i$i$i$i$i;
     $27 = (($26) - ($20))|0;
     $28 = $27 >>> 2;
     $29 = $28 ^ -1;
     $scevgep4$i$i$i2$i$i$i$i$i$i$i$i = (($24) + ($29<<2)|0);
     HEAP32[$23>>2] = $scevgep4$i$i$i2$i$i$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($18); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $21 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $22 = ($21|0)==($1|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($22) {
    break;
   } else {
    $6 = $21;
   }
  }
 }
 $30 = ((($this)) + 16|0); //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = HEAP32[$30>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = ((($this)) + 20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $33 = HEAP32[$32>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $34 = ($33|0)==($31|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($34) {
  $39 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
  $40 = HEAP32[$39>>2]|0; //@line 117 "./common/Object.h"
  $41 = (__ZNK4love8graphics6opengl4Font17getTextureCacheIDEv($40)|0); //@line 213 "modules/graphics/opengl/Text.cpp"
  $42 = ((($this)) + 44|0); //@line 213 "modules/graphics/opengl/Text.cpp"
  HEAP32[$42>>2] = $41; //@line 213 "modules/graphics/opengl/Text.cpp"
  $43 = ((($this)) + 40|0); //@line 214 "modules/graphics/opengl/Text.cpp"
  HEAP32[$43>>2] = 0; //@line 214 "modules/graphics/opengl/Text.cpp"
  return; //@line 215 "modules/graphics/opengl/Text.cpp"
 }
 $__new_last2$i$i$i = $31; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $scevgep$i$i$i = ((($33)) + -12|0);
 $35 = $scevgep$i$i$i;
 $36 = (($35) - ($__new_last2$i$i$i))|0;
 $37 = (($36>>>0) / 12)&-1;
 $38 = $37 ^ -1;
 $scevgep4$i$i$i = (($33) + (($38*12)|0)|0);
 HEAP32[$32>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $39 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $40 = HEAP32[$39>>2]|0; //@line 117 "./common/Object.h"
 $41 = (__ZNK4love8graphics6opengl4Font17getTextureCacheIDEv($40)|0); //@line 213 "modules/graphics/opengl/Text.cpp"
 $42 = ((($this)) + 44|0); //@line 213 "modules/graphics/opengl/Text.cpp"
 HEAP32[$42>>2] = $41; //@line 213 "modules/graphics/opengl/Text.cpp"
 $43 = ((($this)) + 40|0); //@line 214 "modules/graphics/opengl/Text.cpp"
 HEAP32[$43>>2] = 0; //@line 214 "modules/graphics/opengl/Text.cpp"
 return; //@line 215 "modules/graphics/opengl/Text.cpp"
}
function __ZN4love8graphics6opengl4Text11addTextDataERKNS2_8TextDataE($this,$t) {
 $this = $this|0;
 $t = $t|0;
 var $$byval_copy = 0, $$byval_copy10 = 0, $$byval_copy8 = 0, $$byval_copy9 = 0, $$cast = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0.0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0;
 var $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$038 = 0, $__new_last2$i$i$i = 0, $eh$lpad$body$index3Z2D = 0, $eh$lpad$body$indexZ2D = 0, $exitcond$i = 0;
 var $firstcmd$sroa$0$0 = 0, $i$01$i = 0, $new_commands$sroa$0$0 = 0, $new_commands$sroa$0$1 = 0, $new_commands$sroa$0$2 = 0, $new_commands$sroa$14$2 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i$i$i$i$i$i$i = 0, $scevgep$i$i$i28 = 0, $scevgep$i$i$i9 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i$i$i$i$i$i$i = 0, $scevgep4$i$i$i10 = 0, $scevgep4$i$i$i29 = 0, $text_info = 0, $vertices = 0, $voffset$0 = 0, dest = 0, label = 0, sp = 0;
 var src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy10 = sp + 80|0;
 $$byval_copy9 = sp + 76|0;
 $$byval_copy8 = sp + 72|0;
 $$byval_copy = sp + 64|0;
 $vertices = sp + 52|0;
 $text_info = sp;
 $0 = sp + 40|0;
 $1 = sp + 32|0;
 $2 = sp + 20|0;
 $3 = sp + 16|0;
 $4 = sp + 12|0;
 $5 = sp + 8|0;
 HEAP32[$vertices>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ((($vertices)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$6>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ((($vertices)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$7>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = ((($t)) + 28|0); //@line 118 "modules/graphics/opengl/Text.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 118 "modules/graphics/opengl/Text.cpp"
 $10 = ($9|0)==(4); //@line 118 "modules/graphics/opengl/Text.cpp"
 $11 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $12 = HEAP32[$11>>2]|0; //@line 117 "./common/Object.h"
 if ($10) {
  HEAPF32[$1>>2] = 0.0; //@line 225 "./common/Vector.h"
  $13 = ((($1)) + 4|0); //@line 226 "./common/Vector.h"
  HEAPF32[$13>>2] = 0.0; //@line 226 "./common/Vector.h"
  __THREW__ = 0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;HEAP32[$$byval_copy+4>>2]=HEAP32[$1+4>>2]|0; //@line 119 "modules/graphics/opengl/Text.cpp"
  invoke_viiiidii(4029,($0|0),($12|0),($t|0),($vertices|0),0.0,($$byval_copy|0),($text_info|0)); //@line 119 "modules/graphics/opengl/Text.cpp"
  $14 = __THREW__; __THREW__ = 0;
  $15 = $14&1;
  if ($15) {
   $new_commands$sroa$0$0 = 0;
   label = 4;
  } else {
   $16 = HEAP32[$0>>2]|0; //@line 1343 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $17 = ((($0)) + 4|0); //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $18 = HEAP32[$17>>2]|0; //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $19 = ((($0)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$19>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$17>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$0>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $new_commands$sroa$0$2 = $16;$new_commands$sroa$14$2 = $18;
   label = 14;
  }
 } else {
  $33 = ((($t)) + 24|0); //@line 121 "modules/graphics/opengl/Text.cpp"
  $34 = +HEAPF32[$33>>2]; //@line 121 "modules/graphics/opengl/Text.cpp"
  __THREW__ = 0;
  invoke_viiidiii(4031,($2|0),($12|0),($t|0),(+$34),($9|0),($vertices|0),($text_info|0)); //@line 121 "modules/graphics/opengl/Text.cpp"
  $35 = __THREW__; __THREW__ = 0;
  $36 = $35&1;
  if ($36) {
   $new_commands$sroa$0$0 = 0;
   label = 4;
  } else {
   $37 = HEAP32[$2>>2]|0; //@line 1343 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $38 = ((($2)) + 4|0); //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $39 = HEAP32[$38>>2]|0; //@line 1344 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $40 = ((($2)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$40>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$38>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$2>>2] = 0; //@line 1346 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $new_commands$sroa$0$2 = $37;$new_commands$sroa$14$2 = $39;
   label = 14;
  }
 }
 L6: do {
  if ((label|0) == 14) {
   $41 = ((($t)) + 40|0); //@line 123 "modules/graphics/opengl/Text.cpp"
   $42 = HEAP16[$41>>1]|0; //@line 123 "modules/graphics/opengl/Text.cpp"
   $43 = $42&255; //@line 123 "modules/graphics/opengl/Text.cpp"
   $44 = ($43<<24>>24)==(0); //@line 123 "modules/graphics/opengl/Text.cpp"
   if (!($44)) {
    $45 = HEAP32[$vertices>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $46 = HEAP32[$6>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$cast = $45; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $47 = (($46) - ($$cast))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $48 = $47 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $49 = ($47|0)>(0); //@line 278 "./common/Matrix.h"
    if ($49) {
     $50 = ((($t)) + 44|0); //@line 280 "./common/Matrix.h"
     $51 = ((($t)) + 56|0); //@line 280 "./common/Matrix.h"
     $52 = ((($t)) + 68|0); //@line 280 "./common/Matrix.h"
     $53 = ((($t)) + 48|0); //@line 281 "./common/Matrix.h"
     $54 = ((($t)) + 60|0); //@line 281 "./common/Matrix.h"
     $55 = ((($t)) + 72|0); //@line 281 "./common/Matrix.h"
     $i$01$i = 0;
     while(1) {
      $56 = +HEAPF32[$50>>2]; //@line 280 "./common/Matrix.h"
      $57 = (($45) + ($i$01$i<<4)|0); //@line 280 "./common/Matrix.h"
      $58 = +HEAPF32[$57>>2]; //@line 280 "./common/Matrix.h"
      $59 = $56 * $58; //@line 280 "./common/Matrix.h"
      $60 = +HEAPF32[$51>>2]; //@line 280 "./common/Matrix.h"
      $61 = (((($45) + ($i$01$i<<4)|0)) + 4|0); //@line 280 "./common/Matrix.h"
      $62 = +HEAPF32[$61>>2]; //@line 280 "./common/Matrix.h"
      $63 = $60 * $62; //@line 280 "./common/Matrix.h"
      $64 = $59 + $63; //@line 280 "./common/Matrix.h"
      $65 = +HEAPF32[$52>>2]; //@line 280 "./common/Matrix.h"
      $66 = $65 + $64; //@line 280 "./common/Matrix.h"
      $67 = +HEAPF32[$53>>2]; //@line 281 "./common/Matrix.h"
      $68 = $58 * $67; //@line 281 "./common/Matrix.h"
      $69 = +HEAPF32[$54>>2]; //@line 281 "./common/Matrix.h"
      $70 = $62 * $69; //@line 281 "./common/Matrix.h"
      $71 = $68 + $70; //@line 281 "./common/Matrix.h"
      $72 = +HEAPF32[$55>>2]; //@line 281 "./common/Matrix.h"
      $73 = $72 + $71; //@line 281 "./common/Matrix.h"
      HEAPF32[$57>>2] = $66; //@line 283 "./common/Matrix.h"
      HEAPF32[$61>>2] = $73; //@line 284 "./common/Matrix.h"
      $74 = (($i$01$i) + 1)|0; //@line 278 "./common/Matrix.h"
      $exitcond$i = ($74|0)==($48|0); //@line 278 "./common/Matrix.h"
      if ($exitcond$i) {
       break;
      } else {
       $i$01$i = $74;
      }
     }
    }
   }
   $75 = ((($this)) + 40|0); //@line 126 "modules/graphics/opengl/Text.cpp"
   $76 = HEAP32[$75>>2]|0; //@line 126 "modules/graphics/opengl/Text.cpp"
   $77 = ($42&65535)<(256); //@line 128 "modules/graphics/opengl/Text.cpp"
   if ($77) {
    $78 = ((($this)) + 16|0); //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $79 = HEAP32[$78>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $80 = ((($this)) + 20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $81 = HEAP32[$80>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $82 = ($81|0)==($79|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($82) {
     $voffset$0 = 0;
    } else {
     $__new_last2$i$i$i = $79; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $scevgep$i$i$i28 = ((($81)) + -12|0);
     $83 = $scevgep$i$i$i28;
     $84 = (($83) - ($__new_last2$i$i$i))|0;
     $85 = (($84>>>0) / 12)&-1;
     $86 = $85 ^ -1;
     $scevgep4$i$i$i29 = (($81) + (($86*12)|0)|0);
     HEAP32[$80>>2] = $scevgep4$i$i$i29; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $voffset$0 = 0;
    }
   } else {
    $voffset$0 = $76;
   }
   __THREW__ = 0;
   invoke_viii(4144,($this|0),($vertices|0),($voffset$0|0)); //@line 134 "modules/graphics/opengl/Text.cpp"
   $87 = __THREW__; __THREW__ = 0;
   $88 = $87&1;
   if ($88) {
    $new_commands$sroa$0$0 = $new_commands$sroa$0$2;
    label = 4;
   } else {
    $89 = $new_commands$sroa$0$2; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $90 = $new_commands$sroa$14$2; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $91 = ($89|0)==($90|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($91)) {
     $__begin$sroa$0$038 = $89;
     while(1) {
      $98 = ((($__begin$sroa$0$038)) + 4|0); //@line 140 "modules/graphics/opengl/Text.cpp"
      $99 = HEAP32[$98>>2]|0; //@line 140 "modules/graphics/opengl/Text.cpp"
      $100 = (($99) + ($voffset$0))|0; //@line 140 "modules/graphics/opengl/Text.cpp"
      HEAP32[$98>>2] = $100; //@line 140 "modules/graphics/opengl/Text.cpp"
      $101 = ((($__begin$sroa$0$038)) + 12|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $102 = ($101|0)==($90|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($102) {
       break;
      } else {
       $__begin$sroa$0$038 = $101;
      }
     }
     $92 = ((($this)) + 16|0); //@line 147 "modules/graphics/opengl/Text.cpp"
     $93 = HEAP32[$92>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $94 = ((($this)) + 20|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $95 = HEAP32[$94>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $96 = ($93|0)==($95|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $97 = $95; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($96) {
      $firstcmd$sroa$0$0 = $new_commands$sroa$0$2;
     } else {
      $103 = ((($95)) + -12|0); //@line 149 "modules/graphics/opengl/Text.cpp"
      $104 = HEAP32[$103>>2]|0; //@line 149 "modules/graphics/opengl/Text.cpp"
      $105 = ((($95)) + -4|0); //@line 149 "modules/graphics/opengl/Text.cpp"
      $106 = HEAP32[$105>>2]|0; //@line 149 "modules/graphics/opengl/Text.cpp"
      $107 = HEAP32[$89>>2]|0; //@line 150 "modules/graphics/opengl/Text.cpp"
      $108 = ($104|0)==($107|0); //@line 150 "modules/graphics/opengl/Text.cpp"
      if ($108) {
       $109 = ((($95)) + -8|0); //@line 149 "modules/graphics/opengl/Text.cpp"
       $110 = HEAP32[$109>>2]|0; //@line 149 "modules/graphics/opengl/Text.cpp"
       $111 = (($110) + ($106))|0; //@line 150 "modules/graphics/opengl/Text.cpp"
       $112 = ((($89)) + 4|0); //@line 150 "modules/graphics/opengl/Text.cpp"
       $113 = HEAP32[$112>>2]|0; //@line 150 "modules/graphics/opengl/Text.cpp"
       $114 = ($111|0)==($113|0); //@line 150 "modules/graphics/opengl/Text.cpp"
       if ($114) {
        $115 = ((($89)) + 8|0); //@line 152 "modules/graphics/opengl/Text.cpp"
        $116 = HEAP32[$115>>2]|0; //@line 152 "modules/graphics/opengl/Text.cpp"
        $117 = (($116) + ($106))|0; //@line 152 "modules/graphics/opengl/Text.cpp"
        HEAP32[$105>>2] = $117; //@line 152 "modules/graphics/opengl/Text.cpp"
        $118 = ((($89)) + 12|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        $119 = $118; //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        $firstcmd$sroa$0$0 = $119;
       } else {
        $firstcmd$sroa$0$0 = $new_commands$sroa$0$2;
       }
      } else {
       $firstcmd$sroa$0$0 = $new_commands$sroa$0$2;
      }
     }
     HEAP32[$3>>2] = $97; //@line 1128 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     HEAP32[$4>>2] = $firstcmd$sroa$0$0; //@line 158 "modules/graphics/opengl/Text.cpp"
     HEAP32[$5>>2] = $90; //@line 158 "modules/graphics/opengl/Text.cpp"
     __THREW__ = 0;
     ;HEAP32[$$byval_copy8>>2]=HEAP32[$3>>2]|0; //@line 158 "modules/graphics/opengl/Text.cpp"
     ;HEAP32[$$byval_copy9>>2]=HEAP32[$4>>2]|0; //@line 158 "modules/graphics/opengl/Text.cpp"
     ;HEAP32[$$byval_copy10>>2]=HEAP32[$5>>2]|0; //@line 158 "modules/graphics/opengl/Text.cpp"
     (invoke_iiiii(4030,($92|0),($$byval_copy8|0),($$byval_copy9|0),($$byval_copy10|0))|0); //@line 158 "modules/graphics/opengl/Text.cpp"
     $120 = __THREW__; __THREW__ = 0;
     $121 = $120&1;
     if ($121) {
      $new_commands$sroa$0$0 = $new_commands$sroa$0$2;
      label = 4;
      break;
     }
    }
    $122 = HEAP32[$6>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $123 = HEAP32[$vertices>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $124 = (($122) - ($123))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $125 = $124 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $126 = (($125) + ($voffset$0))|0; //@line 161 "modules/graphics/opengl/Text.cpp"
    HEAP32[$75>>2] = $126; //@line 161 "modules/graphics/opengl/Text.cpp"
    $127 = ((($this)) + 32|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $128 = HEAP32[$127>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $129 = ((($this)) + 36|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $130 = HEAP32[$129>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $131 = ($128|0)==($130|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    do {
     if ($131) {
      $154 = ((($this)) + 28|0); //@line 163 "modules/graphics/opengl/Text.cpp"
      __THREW__ = 0;
      invoke_vii(4145,($154|0),($t|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $155 = __THREW__; __THREW__ = 0;
      $156 = $155&1;
      if ($156) {
       $new_commands$sroa$0$0 = $new_commands$sroa$0$2;
       label = 4;
       break L6;
      }
      $$pre = HEAP32[$127>>2]|0; //@line 682 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $158 = $$pre;
     } else {
      __THREW__ = 0;
      invoke_vii(4032,($128|0),($t|0)); //@line 76 "modules/graphics/opengl/Font.h"
      $132 = __THREW__; __THREW__ = 0;
      $133 = $132&1;
      if ($133) {
       $new_commands$sroa$0$0 = $new_commands$sroa$0$2;
       label = 4;
       break L6;
      }
      $134 = ((($128)) + 12|0); //@line 76 "modules/graphics/opengl/Font.h"
      $135 = ((($t)) + 12|0); //@line 76 "modules/graphics/opengl/Font.h"
      __THREW__ = 0;
      invoke_vii(4035,($134|0),($135|0)); //@line 76 "modules/graphics/opengl/Font.h"
      $136 = __THREW__; __THREW__ = 0;
      $137 = $136&1;
      if (!($137)) {
       $150 = ((($128)) + 24|0); //@line 70 "modules/graphics/opengl/Text.h"
       $151 = ((($t)) + 24|0); //@line 70 "modules/graphics/opengl/Text.h"
       dest=$150; src=$151; stop=dest+56|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 70 "modules/graphics/opengl/Text.h"
       $152 = HEAP32[$127>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $153 = ((($152)) + 80|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$127>>2] = $153; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $158 = $153;
       break;
      }
      $138 = ___cxa_find_matching_catch_2()|0;
      $139 = tempRet0;
      $140 = HEAP32[$128>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $141 = ($140|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $142 = $140; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($141) {
       $eh$lpad$body$index3Z2D = $139;$eh$lpad$body$indexZ2D = $138;$new_commands$sroa$0$1 = $new_commands$sroa$0$2;
       break L6;
      }
      $143 = ((($128)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $144 = HEAP32[$143>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $145 = ($144|0)==($140|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($145)) {
       $scevgep$i$i$i$i$i$i$i$i$i = ((($144)) + -4|0);
       $146 = $scevgep$i$i$i$i$i$i$i$i$i;
       $147 = (($146) - ($142))|0;
       $148 = $147 >>> 2;
       $149 = $148 ^ -1;
       $scevgep4$i$i$i$i$i$i$i$i$i = (($144) + ($149<<2)|0);
       HEAP32[$143>>2] = $scevgep4$i$i$i$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($140); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $eh$lpad$body$index3Z2D = $139;$eh$lpad$body$indexZ2D = $138;$new_commands$sroa$0$1 = $new_commands$sroa$0$2;
      break L6;
     }
    } while(0);
    $157 = ((($158)) + -48|0); //@line 164 "modules/graphics/opengl/Text.cpp"
    $159 = $text_info; //@line 164 "modules/graphics/opengl/Text.cpp"
    $160 = $159; //@line 164 "modules/graphics/opengl/Text.cpp"
    $161 = HEAP32[$160>>2]|0; //@line 164 "modules/graphics/opengl/Text.cpp"
    $162 = (($159) + 4)|0; //@line 164 "modules/graphics/opengl/Text.cpp"
    $163 = $162; //@line 164 "modules/graphics/opengl/Text.cpp"
    $164 = HEAP32[$163>>2]|0; //@line 164 "modules/graphics/opengl/Text.cpp"
    $165 = $157; //@line 164 "modules/graphics/opengl/Text.cpp"
    $166 = $165; //@line 164 "modules/graphics/opengl/Text.cpp"
    HEAP32[$166>>2] = $161; //@line 164 "modules/graphics/opengl/Text.cpp"
    $167 = (($165) + 4)|0; //@line 164 "modules/graphics/opengl/Text.cpp"
    $168 = $167; //@line 164 "modules/graphics/opengl/Text.cpp"
    HEAP32[$168>>2] = $164; //@line 164 "modules/graphics/opengl/Text.cpp"
    $169 = HEAP32[$11>>2]|0; //@line 117 "./common/Object.h"
    __THREW__ = 0;
    $170 = (invoke_ii(4142,($169|0))|0); //@line 167 "modules/graphics/opengl/Text.cpp"
    $171 = __THREW__; __THREW__ = 0;
    $172 = $171&1;
    if ($172) {
     $new_commands$sroa$0$0 = $new_commands$sroa$0$2;
     label = 4;
    } else {
     $173 = ((($this)) + 44|0); //@line 167 "modules/graphics/opengl/Text.cpp"
     $174 = HEAP32[$173>>2]|0; //@line 167 "modules/graphics/opengl/Text.cpp"
     $175 = ($170|0)==($174|0); //@line 167 "modules/graphics/opengl/Text.cpp"
     if (!($175)) {
      __THREW__ = 0;
      invoke_vi(4146,($this|0)); //@line 168 "modules/graphics/opengl/Text.cpp"
      $176 = __THREW__; __THREW__ = 0;
      $177 = $176&1;
      if ($177) {
       $new_commands$sroa$0$0 = $new_commands$sroa$0$2;
       label = 4;
       break;
      }
     }
     $178 = ($new_commands$sroa$0$2|0)==(0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($178)) {
      $179 = $new_commands$sroa$0$2; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      __ZdlPv($179); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $180 = HEAP32[$vertices>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $181 = ($180|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $182 = $180; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($181) {
      STACKTOP = sp;return; //@line 169 "modules/graphics/opengl/Text.cpp"
     }
     $183 = HEAP32[$6>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $184 = ($183|0)==($180|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($184)) {
      $scevgep$i$i$i = ((($183)) + -16|0);
      $185 = $scevgep$i$i$i;
      $186 = (($185) - ($182))|0;
      $187 = $186 >>> 4;
      $188 = $187 ^ -1;
      $scevgep4$i$i$i = (($183) + ($188<<4)|0);
      HEAP32[$6>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __ZdlPv($180); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     STACKTOP = sp;return; //@line 169 "modules/graphics/opengl/Text.cpp"
    }
   }
  }
 } while(0);
 if ((label|0) == 4) {
  $20 = ___cxa_find_matching_catch_2()|0;
  $21 = tempRet0;
  $eh$lpad$body$index3Z2D = $21;$eh$lpad$body$indexZ2D = $20;$new_commands$sroa$0$1 = $new_commands$sroa$0$0;
 }
 $22 = ($new_commands$sroa$0$1|0)==(0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($22)) {
  $23 = $new_commands$sroa$0$1; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  __ZdlPv($23); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $24 = HEAP32[$vertices>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $25 = ($24|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $26 = $24; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($25) {
  ___resumeException($eh$lpad$body$indexZ2D|0);
  // unreachable;
 }
 $27 = HEAP32[$6>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $28 = ($27|0)==($24|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($28)) {
  $scevgep$i$i$i9 = ((($27)) + -16|0);
  $29 = $scevgep$i$i$i9;
  $30 = (($29) - ($26))|0;
  $31 = $30 >>> 4;
  $32 = $31 ^ -1;
  $scevgep4$i$i$i10 = (($27) + ($32<<4)|0);
  HEAP32[$6>>2] = $scevgep4$i$i$i10; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($24); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
}
function __ZN4love8graphics6opengl4Text14uploadVerticesERKNSt3__16vectorINS1_4Font11GlyphVertexENS3_9allocatorIS6_EEEEj($this,$vertices,$vertoffset) {
 $this = $this|0;
 $vertices = $vertices|0;
 $vertoffset = $vertoffset|0;
 var $$0 = 0, $$02 = 0, $$1 = 0, $$13 = 0, $$ph = 0, $$pr9 = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $newsize$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $vertoffset << 4; //@line 49 "modules/graphics/opengl/Text.cpp"
 $1 = ((($vertices)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$vertices>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($2) - ($3))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($2|0)==($3|0); //@line 54 "modules/graphics/opengl/Text.cpp"
 if ($5) {
  return; //@line 91 "modules/graphics/opengl/Text.cpp"
 }
 $6 = ((($this)) + 12|0); //@line 54 "modules/graphics/opengl/Text.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 58 "modules/graphics/opengl/Text.cpp"
 $8 = ($7|0)==(0|0); //@line 54 "modules/graphics/opengl/Text.cpp"
 $9 = (($4) + ($0))|0; //@line 57 "modules/graphics/opengl/Text.cpp"
 if ($8) {
  $10 = (+($9>>>0)); //@line 57 "modules/graphics/opengl/Text.cpp"
  $11 = $10 * 1.5; //@line 57 "modules/graphics/opengl/Text.cpp"
  $12 = (~~(($11))>>>0); //@line 57 "modules/graphics/opengl/Text.cpp"
  $newsize$0 = $12;
  label = 6;
 } else {
  $13 = ((($7)) + 8|0); //@line 76 "modules/graphics/opengl/GLBuffer.h"
  $14 = HEAP32[$13>>2]|0; //@line 76 "modules/graphics/opengl/GLBuffer.h"
  $15 = ($9>>>0)>($14>>>0); //@line 54 "modules/graphics/opengl/Text.cpp"
  if ($15) {
   $16 = (+($9>>>0)); //@line 57 "modules/graphics/opengl/Text.cpp"
   $17 = $16 * 1.5; //@line 57 "modules/graphics/opengl/Text.cpp"
   $18 = (~~(($17))>>>0); //@line 57 "modules/graphics/opengl/Text.cpp"
   $19 = (+($14>>>0)); //@line 59 "modules/graphics/opengl/Text.cpp"
   $20 = $19 * 1.5; //@line 59 "modules/graphics/opengl/Text.cpp"
   $21 = (~~(($20))>>>0); //@line 59 "modules/graphics/opengl/Text.cpp"
   $22 = ($21>>>0)<($18>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $23 = $22 ? $18 : $21; //@line 59 "modules/graphics/opengl/Text.cpp"
   $newsize$0 = $23;
   label = 6;
  } else {
   $$ph = $7;
  }
 }
 if ((label|0) == 6) {
  $24 = (__Znwj(40)|0); //@line 61 "modules/graphics/opengl/Text.cpp"
  __THREW__ = 0;
  invoke_viiiiii(4042,($24|0),($newsize$0|0),(0|0),34962,35048,0); //@line 61 "modules/graphics/opengl/Text.cpp"
  $25 = __THREW__; __THREW__ = 0;
  $26 = $25&1;
  if ($26) {
   $44 = ___cxa_find_matching_catch_2()|0;
   $45 = tempRet0;
   __ZdlPv($24); //@line 61 "modules/graphics/opengl/Text.cpp"
   $$1 = $45;$$13 = $44;
   ___resumeException($$13|0);
   // unreachable;
  }
  $27 = HEAP32[$6>>2]|0; //@line 63 "modules/graphics/opengl/Text.cpp"
  $28 = ($27|0)==(0|0); //@line 63 "modules/graphics/opengl/Text.cpp"
  L14: do {
   if (!($28)) {
    __THREW__ = 0;
    invoke_vi(4087,($27|0)); //@line 188 "modules/graphics/opengl/GLBuffer.h"
    $29 = __THREW__; __THREW__ = 0;
    $30 = $29&1;
    do {
     if ($30) {
      $46 = ___cxa_find_matching_catch_3(96|0)|0;
      $47 = tempRet0;
      $$0 = $47;$$02 = $46;
     } else {
      $31 = HEAP32[$6>>2]|0; //@line 68 "modules/graphics/opengl/Text.cpp"
      __THREW__ = 0;
      $32 = (invoke_ii(4096,($31|0))|0); //@line 68 "modules/graphics/opengl/Text.cpp"
      $33 = __THREW__; __THREW__ = 0;
      $34 = $33&1;
      if ($34) {
       $48 = ___cxa_find_matching_catch_3(96|0)|0;
       $49 = tempRet0;
       __THREW__ = 0;
       invoke_vi(4089,($27|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
       $50 = __THREW__; __THREW__ = 0;
       $51 = $50&1;
       if (!($51)) {
        $$0 = $49;$$02 = $48;
        break;
       }
       $52 = ___cxa_find_matching_catch_3(0|0)|0;
       $53 = tempRet0;
       ___clang_call_terminate($52); //@line 196 "modules/graphics/opengl/GLBuffer.h"
       // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
      }
      __THREW__ = 0;
      invoke_vi(4089,($27|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
      $35 = __THREW__; __THREW__ = 0;
      $36 = $35&1;
      if ($36) {
       $37 = ___cxa_find_matching_catch_3(0|0)|0;
       $38 = tempRet0;
       ___clang_call_terminate($37); //@line 196 "modules/graphics/opengl/GLBuffer.h"
       // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
      }
      __ZN4love8graphics6opengl8GLBuffer4bindEv($24); //@line 188 "modules/graphics/opengl/GLBuffer.h"
      $39 = HEAP32[$6>>2]|0; //@line 77 "modules/graphics/opengl/Text.cpp"
      $40 = ((($39)) + 8|0); //@line 76 "modules/graphics/opengl/GLBuffer.h"
      $41 = HEAP32[$40>>2]|0; //@line 76 "modules/graphics/opengl/GLBuffer.h"
      __THREW__ = 0;
      invoke_viiii(4138,($24|0),0,($41|0),($32|0)); //@line 77 "modules/graphics/opengl/Text.cpp"
      $42 = __THREW__; __THREW__ = 0;
      $43 = $42&1;
      if ($43) {
       $68 = ___cxa_find_matching_catch_2()|0;
       $69 = tempRet0;
       __THREW__ = 0;
       invoke_vi(4089,($24|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
       $70 = __THREW__; __THREW__ = 0;
       $71 = $70&1;
       if ($71) {
        $72 = ___cxa_find_matching_catch_3(0|0)|0;
        $73 = tempRet0;
        ___clang_call_terminate($72); //@line 196 "modules/graphics/opengl/GLBuffer.h"
        // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
       } else {
        $$1 = $69;$$13 = $68;
        ___resumeException($$13|0);
        // unreachable;
       }
      }
      __THREW__ = 0;
      invoke_vi(4089,($24|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
      $64 = __THREW__; __THREW__ = 0;
      $65 = $64&1;
      if ($65) {
       $66 = ___cxa_find_matching_catch_3(0|0)|0;
       $67 = tempRet0;
       ___clang_call_terminate($66); //@line 196 "modules/graphics/opengl/GLBuffer.h"
       // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
      }
      $$pr9 = HEAP32[$6>>2]|0; //@line 80 "modules/graphics/opengl/Text.cpp"
      $74 = ($$pr9|0)==(0|0); //@line 80 "modules/graphics/opengl/Text.cpp"
      if ($74) {
       break L14;
      }
      $75 = HEAP32[$$pr9>>2]|0; //@line 80 "modules/graphics/opengl/Text.cpp"
      $76 = ((($75)) + 4|0); //@line 80 "modules/graphics/opengl/Text.cpp"
      $77 = HEAP32[$76>>2]|0; //@line 80 "modules/graphics/opengl/Text.cpp"
      FUNCTION_TABLE_vi[$77 & 8191]($$pr9); //@line 80 "modules/graphics/opengl/Text.cpp"
      break L14;
     }
    } while(0);
    $54 = (_llvm_eh_typeid_for((96|0))|0); //@line 69 "modules/graphics/opengl/Text.cpp"
    $55 = ($$0|0)==($54|0); //@line 69 "modules/graphics/opengl/Text.cpp"
    if (!($55)) {
     $$1 = $$0;$$13 = $$02;
     ___resumeException($$13|0);
     // unreachable;
    }
    (___cxa_begin_catch(($$02|0))|0); //@line 69 "modules/graphics/opengl/Text.cpp"
    $56 = HEAP32[$24>>2]|0; //@line 72 "modules/graphics/opengl/Text.cpp"
    $57 = ((($56)) + 4|0); //@line 72 "modules/graphics/opengl/Text.cpp"
    $58 = HEAP32[$57>>2]|0; //@line 72 "modules/graphics/opengl/Text.cpp"
    FUNCTION_TABLE_vi[$58 & 8191]($24); //@line 72 "modules/graphics/opengl/Text.cpp"
    __THREW__ = 0;
    invoke_v(3656); //@line 73 "modules/graphics/opengl/Text.cpp"
    $59 = __THREW__; __THREW__ = 0;
    $60 = ___cxa_find_matching_catch_2()|0;
    $61 = tempRet0;
    __THREW__ = 0;
    invoke_v(3378); //@line 74 "modules/graphics/opengl/Text.cpp"
    $62 = __THREW__; __THREW__ = 0;
    $63 = $62&1;
    if ($63) {
     $94 = ___cxa_find_matching_catch_3(0|0)|0;
     $95 = tempRet0;
     ___clang_call_terminate($94); //@line 74 "modules/graphics/opengl/Text.cpp"
     // unreachable; //@line 74 "modules/graphics/opengl/Text.cpp"
    } else {
     $$1 = $61;$$13 = $60;
     ___resumeException($$13|0);
     // unreachable;
    }
   }
  } while(0);
  HEAP32[$6>>2] = $24; //@line 81 "modules/graphics/opengl/Text.cpp"
  $$ph = $24;
 }
 __ZN4love8graphics6opengl8GLBuffer4bindEv($$ph); //@line 188 "modules/graphics/opengl/GLBuffer.h"
 $78 = HEAP32[$6>>2]|0; //@line 87 "modules/graphics/opengl/Text.cpp"
 __THREW__ = 0;
 $79 = (invoke_ii(4096,($78|0))|0); //@line 87 "modules/graphics/opengl/Text.cpp"
 $80 = __THREW__; __THREW__ = 0;
 $81 = $80&1;
 if ($81) {
  $88 = ___cxa_find_matching_catch_2()|0;
  $89 = tempRet0;
  __THREW__ = 0;
  invoke_vi(4089,($$ph|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  $90 = __THREW__; __THREW__ = 0;
  $91 = $90&1;
  if ($91) {
   $92 = ___cxa_find_matching_catch_3(0|0)|0;
   $93 = tempRet0;
   ___clang_call_terminate($92); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
  } else {
   $$1 = $89;$$13 = $88;
   ___resumeException($$13|0);
   // unreachable;
  }
 } else {
  $82 = (($79) + ($0)|0); //@line 88 "modules/graphics/opengl/Text.cpp"
  $83 = HEAP32[$vertices>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  _memcpy(($82|0),($83|0),($4|0))|0; //@line 88 "modules/graphics/opengl/Text.cpp"
  __THREW__ = 0;
  invoke_vi(4089,($$ph|0)); //@line 196 "modules/graphics/opengl/GLBuffer.h"
  $84 = __THREW__; __THREW__ = 0;
  $85 = $84&1;
  if ($85) {
   $86 = ___cxa_find_matching_catch_3(0|0)|0;
   $87 = tempRet0;
   ___clang_call_terminate($86); //@line 196 "modules/graphics/opengl/GLBuffer.h"
   // unreachable; //@line 196 "modules/graphics/opengl/GLBuffer.h"
  } else {
   return; //@line 91 "modules/graphics/opengl/Text.cpp"
  }
 }
}
function __ZNSt3__16vectorIN4love8graphics6opengl4Text8TextDataENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i15 = 0, $$0$i16 = 0, $$01$i$i = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre17 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__new_last1$i$i$i$i$i$i$i$i$i$i$i$i2 = 0, $__v$sroa$6$0 = 0, $eh$lpad$body18$index3Z2D = 0, $eh$lpad$body18$indexZ2D = 0, $eh$lpad$body19$index9Z2D = 0, $eh$lpad$body19$indexZ2D = 0, $scevgep$i$i$i$i$i$i$i$i = 0, $scevgep$i$i$i$i$i$i$i$i$i$i$i$i3 = 0, $scevgep$i$i$i1$i$i$i$i$i$i$i$i$i9 = 0, $scevgep4$i$i$i$i$i$i$i$i = 0, $scevgep4$i$i$i$i$i$i$i$i$i$i$i$i4 = 0, $scevgep4$i$i$i2$i$i$i$i$i$i$i$i$i10 = 0, dest = 0, label = 0;
 var sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($3|0) / 80)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(53687091); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $10 = $$pre;
 } else {
  $10 = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($10))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = (($9|0) / 80)&-1; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ($11>>>0)<(26843545); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($12) {
  $16 = $11 << 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = ($16>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $18 = $17 ? $5 : $16; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = (($19) - ($10))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = (($20|0) / 80)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = ($18|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($22) {
   $$0$i16 = 0;$26 = 0;$27 = $21;
  } else {
   $$0$i15 = $18;$118 = $21;
   label = 6;
  }
 } else {
  $13 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = (($13) - ($10))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = (($14|0) / 80)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i15 = 53687091;$118 = $15;
  label = 6;
 }
 if ((label|0) == 6) {
  $23 = ($$0$i15*80)|0; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $24 = (__Znwj($23)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i16 = $$0$i15;$26 = $24;$27 = $118;
 }
 $25 = (($26) + (($27*80)|0)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $25; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = (($26) + (($$0$i16*80)|0)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $30 = $29; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 __THREW__ = 0;
 invoke_vii(4032,($25|0),($__x|0)); //@line 76 "modules/graphics/opengl/Font.h"
 $31 = __THREW__; __THREW__ = 0;
 $32 = $31&1;
 do {
  if ($32) {
   $115 = ___cxa_find_matching_catch_2()|0;
   $116 = tempRet0;
   $117 = ($26|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($117) {
    $eh$lpad$body19$index9Z2D = $116;$eh$lpad$body19$indexZ2D = $115;
    ___resumeException($eh$lpad$body19$indexZ2D|0);
    // unreachable;
   } else {
    $eh$lpad$body18$index3Z2D = $116;$eh$lpad$body18$indexZ2D = $115;
   }
  } else {
   $33 = (((($26) + (($27*80)|0)|0)) + 12|0); //@line 76 "modules/graphics/opengl/Font.h"
   $34 = ((($__x)) + 12|0); //@line 76 "modules/graphics/opengl/Font.h"
   __THREW__ = 0;
   invoke_vii(4035,($33|0),($34|0)); //@line 76 "modules/graphics/opengl/Font.h"
   $35 = __THREW__; __THREW__ = 0;
   $36 = $35&1;
   if ($36) {
    $37 = ___cxa_find_matching_catch_2()|0;
    $38 = tempRet0;
    $39 = HEAP32[$25>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $40 = ($39|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $41 = $39; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($40) {
     $eh$lpad$body18$index3Z2D = $38;$eh$lpad$body18$indexZ2D = $37;
     break;
    }
    $42 = (((($26) + (($27*80)|0)|0)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $43 = HEAP32[$42>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $44 = ($43|0)==($39|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($44)) {
     $scevgep$i$i$i$i$i$i$i$i = ((($43)) + -4|0);
     $45 = $scevgep$i$i$i$i$i$i$i$i;
     $46 = (($45) - ($41))|0;
     $47 = $46 >>> 2;
     $48 = $47 ^ -1;
     $scevgep4$i$i$i$i$i$i$i$i = (($43) + ($48<<2)|0);
     HEAP32[$42>>2] = $scevgep4$i$i$i$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($39); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $eh$lpad$body18$index3Z2D = $38;$eh$lpad$body18$indexZ2D = $37;
    break;
   }
   $49 = (((($26) + (($27*80)|0)|0)) + 24|0); //@line 70 "modules/graphics/opengl/Text.h"
   $50 = ((($__x)) + 24|0); //@line 70 "modules/graphics/opengl/Text.h"
   dest=$49; src=$50; stop=dest+56|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 70 "modules/graphics/opengl/Text.h"
   $51 = ((($25)) + 80|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $52 = $51; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $53 = HEAP32[$this>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $54 = HEAP32[$0>>2]|0; //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $55 = ($54|0)==($53|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   if ($55) {
    $56 = $53; //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $84 = $28;$86 = $56;$88 = $53;
   } else {
    $$01$i$i = $54;$58 = $25;$__v$sroa$6$0 = $28;
    while(1) {
     $57 = ((($58)) + -80|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $59 = ((($$01$i$i)) + -80|0); //@line 1535 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$57>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $60 = ((($58)) + -76|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$60>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $61 = ((($58)) + -72|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$61>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $62 = HEAP32[$59>>2]|0; //@line 1246 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$57>>2] = $62; //@line 1246 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $63 = ((($$01$i$i)) + -76|0); //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $64 = HEAP32[$63>>2]|0; //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$60>>2] = $64; //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $65 = ((($$01$i$i)) + -72|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $66 = HEAP32[$65>>2]|0; //@line 1248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $67 = ((($58)) + -72|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$67>>2] = $66; //@line 1248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$65>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$63>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$59>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $68 = ((($58)) + -68|0); //@line 76 "modules/graphics/opengl/Font.h"
     $69 = ((($$01$i$i)) + -68|0); //@line 76 "modules/graphics/opengl/Font.h"
     HEAP32[$68>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $70 = ((($58)) + -64|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$70>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $71 = ((($58)) + -60|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$71>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $72 = HEAP32[$69>>2]|0; //@line 1246 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$68>>2] = $72; //@line 1246 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $73 = ((($$01$i$i)) + -64|0); //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $74 = HEAP32[$73>>2]|0; //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$70>>2] = $74; //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $75 = ((($$01$i$i)) + -60|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $76 = HEAP32[$75>>2]|0; //@line 1248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $77 = ((($58)) + -60|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$77>>2] = $76; //@line 1248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$75>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$73>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$69>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $78 = ((($58)) + -56|0); //@line 70 "modules/graphics/opengl/Text.h"
     $79 = ((($$01$i$i)) + -56|0); //@line 70 "modules/graphics/opengl/Text.h"
     dest=$78; src=$79; stop=dest+56|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 70 "modules/graphics/opengl/Text.h"
     $80 = $__v$sroa$6$0; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $81 = ((($80)) + -80|0); //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $82 = $81; //@line 1536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $83 = ($59|0)==($53|0); //@line 1533 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     if ($83) {
      $$lcssa = $82;
      break;
     } else {
      $$01$i$i = $59;$58 = $81;$__v$sroa$6$0 = $82;
     }
    }
    $$pre$i = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $$pre17 = HEAP32[$0>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $84 = $$lcssa;$86 = $$pre$i;$88 = $$pre17;
   }
   HEAP32[$this>>2] = $84; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$0>>2] = $52; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$7>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $85 = $86; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $87 = ($88|0)==($85|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if (!($87)) {
    $90 = $88;
    while(1) {
     $89 = ((($90)) + -80|0); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $91 = ((($90)) + -36|0); //@line 70 "modules/graphics/opengl/Text.h"
     __ZN4love7Matrix3D2Ev($91); //@line 70 "modules/graphics/opengl/Text.h"
     $92 = ((($90)) + -68|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $93 = HEAP32[$92>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $94 = ($93|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($94)) {
      $95 = ((($90)) + -64|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $96 = HEAP32[$95>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $97 = ($96|0)==($93|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($97)) {
       $__new_last1$i$i$i$i$i$i$i$i$i$i$i$i2 = $93; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $scevgep$i$i$i$i$i$i$i$i$i$i$i$i3 = ((($96)) + -8|0);
       $98 = $scevgep$i$i$i$i$i$i$i$i$i$i$i$i3;
       $99 = (($98) - ($__new_last1$i$i$i$i$i$i$i$i$i$i$i$i2))|0;
       $100 = $99 >>> 3;
       $101 = $100 ^ -1;
       $scevgep4$i$i$i$i$i$i$i$i$i$i$i$i4 = (($96) + ($101<<3)|0);
       HEAP32[$95>>2] = $scevgep4$i$i$i$i$i$i$i$i$i$i$i$i4; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($93); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $102 = HEAP32[$89>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $103 = ($102|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $104 = $102; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($103)) {
      $106 = ((($90)) + -76|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $107 = HEAP32[$106>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $108 = ($107|0)==($102|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($108)) {
       $scevgep$i$i$i1$i$i$i$i$i$i$i$i$i9 = ((($107)) + -4|0);
       $109 = $scevgep$i$i$i1$i$i$i$i$i$i$i$i$i9;
       $110 = (($109) - ($104))|0;
       $111 = $110 >>> 2;
       $112 = $111 ^ -1;
       $scevgep4$i$i$i2$i$i$i$i$i$i$i$i$i10 = (($107) + ($112<<2)|0);
       HEAP32[$106>>2] = $scevgep4$i$i$i2$i$i$i$i$i$i$i$i$i10; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($102); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $105 = ($89|0)==($85|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     if ($105) {
      break;
     } else {
      $90 = $89;
     }
    }
   }
   $113 = ($86|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($113) {
    return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $114 = $86; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __ZdlPv($114); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
 } while(0);
 __ZdlPv($26); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $eh$lpad$body19$index9Z2D = $eh$lpad$body18$index3Z2D;$eh$lpad$body19$indexZ2D = $eh$lpad$body18$indexZ2D;
 ___resumeException($eh$lpad$body19$indexZ2D|0);
 // unreachable;
}
function __ZN4love8graphics6opengl4TextC2EPNS1_4FontERKNSt3__16vectorINS3_13ColoredStringENS5_9allocatorIS7_EEEE($this,$font,$text) {
 $this = $this|0;
 $font = $font|0;
 $text = $text|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 36 "./modules/graphics/Drawable.h"
 HEAP32[$this>>2] = (25456); //@line 38 "modules/graphics/opengl/Text.cpp"
 $0 = ((($this)) + 8|0); //@line 93 "./common/Object.h"
 HEAP32[$0>>2] = $font; //@line 93 "./common/Object.h"
 $1 = ($font|0)==(0|0); //@line 95 "./common/Object.h"
 if (!($1)) {
  $2 = HEAP32[$font>>2]|0; //@line 95 "./common/Object.h"
  $3 = ((($2)) + 8|0); //@line 95 "./common/Object.h"
  $4 = HEAP32[$3>>2]|0; //@line 95 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($4|0,($font|0)); //@line 95 "./common/Object.h"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if ($6) {
   $12 = ___cxa_find_matching_catch_2()|0;
   $13 = tempRet0;
   $$0 = $13;$$01 = $12;
   ___resumeException($$01|0);
   // unreachable;
  }
 }
 $7 = ((($this)) + 12|0); //@line 35 "modules/graphics/opengl/Text.cpp"
 $8 = ((($this)) + 20|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ((($this)) + 44|0); //@line 37 "modules/graphics/opengl/Text.cpp"
 ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;HEAP32[$7+12>>2]=0|0;HEAP32[$7+16>>2]=0|0;HEAP32[$7+20>>2]=0|0;HEAP32[$7+24>>2]=0|0;HEAP32[$7+28>>2]=0|0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$9>>2] = -1; //@line 37 "modules/graphics/opengl/Text.cpp"
 __THREW__ = 0;
 invoke_viidi(3619,($this|0),($text|0),-1.0,4); //@line 173 "modules/graphics/opengl/Text.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if (!($11)) {
  return; //@line 40 "modules/graphics/opengl/Text.cpp"
 }
 $14 = ___cxa_find_matching_catch_2()|0;
 $15 = tempRet0;
 $16 = ((($this)) + 16|0); //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ((($this)) + 28|0); //@line 40 "modules/graphics/opengl/Text.cpp"
 __ZNSt3__113__vector_baseIN4love8graphics6opengl4Text8TextDataENS_9allocatorIS5_EEED2Ev($17); //@line 40 "modules/graphics/opengl/Text.cpp"
 $18 = HEAP32[$16>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $19 = ($18|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = $18; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($19)) {
  $21 = HEAP32[$8>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = ($21|0)==($18|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($22)) {
   $scevgep$i$i$i = ((($21)) + -12|0);
   $23 = $scevgep$i$i$i;
   $24 = (($23) - ($20))|0;
   $25 = (($24>>>0) / 12)&-1;
   $26 = $25 ^ -1;
   $scevgep4$i$i$i = (($21) + (($26*12)|0)|0);
   HEAP32[$8>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($18); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $27 = HEAP32[$0>>2]|0; //@line 106 "./common/Object.h"
 $28 = ($27|0)==(0|0); //@line 106 "./common/Object.h"
 if ($28) {
  $$0 = $15;$$01 = $14;
  ___resumeException($$01|0);
  // unreachable;
 }
 $29 = HEAP32[$27>>2]|0; //@line 106 "./common/Object.h"
 $30 = ((($29)) + 12|0); //@line 106 "./common/Object.h"
 $31 = HEAP32[$30>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($31|0,($27|0)); //@line 106 "./common/Object.h"
 $32 = __THREW__; __THREW__ = 0;
 $33 = $32&1;
 if ($33) {
  $34 = ___cxa_find_matching_catch_3(0|0)|0;
  $35 = tempRet0;
  ___clang_call_terminate($34); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  $$0 = $15;$$01 = $14;
  ___resumeException($$01|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl4Text3setERKNSt3__16vectorINS1_4Font13ColoredStringENS3_9allocatorIS6_EEEEfNS5_9AlignModeE($this,$text,$wrap,$align) {
 $this = $this|0;
 $text = $text|0;
 $wrap = +$wrap;
 $align = $align|0;
 var $$0 = 0, $$01 = 0, $$01$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__new_last1$i$i$i$i = 0, $__new_last1$i$i$i$i$i = 0, $__new_last1$i$i$i$i$i6 = 0, $__new_last1$i$i$i$i17 = 0, $__new_last1$i$i$i$i29 = 0, $__new_last1$i$i$i$i41 = 0, $codepoints = 0, $eh$lpad$body$index3Z2D = 0, $eh$lpad$body$indexZ2D = 0;
 var $scevgep$i$i$i$i = 0, $scevgep$i$i$i$i$i = 0, $scevgep$i$i$i$i$i7 = 0, $scevgep$i$i$i$i18 = 0, $scevgep$i$i$i$i2 = 0, $scevgep$i$i$i$i30 = 0, $scevgep$i$i$i$i42 = 0, $scevgep$i$i$i1$i = 0, $scevgep$i$i$i1$i$i = 0, $scevgep$i$i$i1$i$i12 = 0, $scevgep$i$i$i1$i23 = 0, $scevgep$i$i$i1$i35 = 0, $scevgep4$i$i$i$i = 0, $scevgep4$i$i$i$i$i = 0, $scevgep4$i$i$i$i$i8 = 0, $scevgep4$i$i$i$i19 = 0, $scevgep4$i$i$i$i3 = 0, $scevgep4$i$i$i$i31 = 0, $scevgep4$i$i$i$i43 = 0, $scevgep4$i$i$i2$i = 0;
 var $scevgep4$i$i$i2$i$i = 0, $scevgep4$i$i$i2$i$i13 = 0, $scevgep4$i$i$i2$i24 = 0, $scevgep4$i$i$i2$i36 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $codepoints = sp + 80|0;
 $0 = sp;
 $1 = HEAP32[$text>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($text)) + 4|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($1|0)==($3|0); //@line 659 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 do {
  if (!($4)) {
   $5 = $1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $6 = $3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $7 = (($6) - ($5))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $8 = ($7|0)==(16); //@line 178 "modules/graphics/opengl/Text.cpp"
   if ($8) {
    $9 = HEAP8[$1>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $10 = $9 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $11 = ($10<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if ($11) {
     $14 = $9&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $15 = $14 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $17 = $15;
    } else {
     $12 = ((($1)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $13 = HEAP32[$12>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $17 = $13;
    }
    $16 = ($17|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if ($16) {
     break;
    }
   }
   ;HEAP32[$codepoints>>2]=0|0;HEAP32[$codepoints+4>>2]=0|0;HEAP32[$codepoints+8>>2]=0|0;HEAP32[$codepoints+12>>2]=0|0;HEAP32[$codepoints+16>>2]=0|0;HEAP32[$codepoints+20>>2]=0|0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __THREW__ = 0;
   invoke_vii(4038,($text|0),($codepoints|0)); //@line 182 "modules/graphics/opengl/Text.cpp"
   $18 = __THREW__; __THREW__ = 0;
   $19 = $18&1;
   L10: do {
    if ($19) {
     label = 45;
    } else {
     __THREW__ = 0;
     invoke_vii(4032,($0|0),($codepoints|0)); //@line 76 "modules/graphics/opengl/Font.h"
     $20 = __THREW__; __THREW__ = 0;
     $21 = $20&1;
     if ($21) {
      label = 45;
     } else {
      $22 = ((($0)) + 12|0); //@line 76 "modules/graphics/opengl/Font.h"
      $23 = ((($codepoints)) + 12|0); //@line 76 "modules/graphics/opengl/Font.h"
      HEAP32[$22>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $24 = ((($0)) + 16|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$24>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $25 = ((($0)) + 20|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$25>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $26 = ((($codepoints)) + 16|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $27 = HEAP32[$26>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $28 = HEAP32[$23>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $29 = (($27) - ($28))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $30 = $29 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $31 = ($30|0)==(0); //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      L13: do {
       if (!($31)) {
        $32 = ($30>>>0)>(536870911); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($32) {
         __THREW__ = 0;
         invoke_vi(3528,($22|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $33 = __THREW__; __THREW__ = 0;
         $34 = $33&1;
         if (!($34)) {
          label = 13;
         }
        } else {
         label = 13;
        }
        if ((label|0) == 13) {
         __THREW__ = 0;
         $35 = (invoke_ii(3348,($29|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
         $36 = __THREW__; __THREW__ = 0;
         $37 = $36&1;
         if (!($37)) {
          HEAP32[$24>>2] = $35; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$22>>2] = $35; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $38 = (($35) + ($30<<3)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $39 = ((($0)) + 20|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          HEAP32[$39>>2] = $38; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $40 = HEAP32[$23>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $41 = HEAP32[$26>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $42 = ($40|0)==($41|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          if ($42) {
           break;
          }
          $$01$i$i = $40;$50 = $35;
          while(1) {
           $43 = $$01$i$i; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $44 = $43; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $45 = HEAP32[$44>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $46 = (($43) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $47 = $46; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $48 = HEAP32[$47>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $49 = $50; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $51 = $49; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           HEAP32[$51>>2] = $45; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $52 = (($49) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $53 = $52; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           HEAP32[$53>>2] = $48; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $54 = HEAP32[$24>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $55 = ((($54)) + 8|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           HEAP32[$24>>2] = $55; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $56 = ((($$01$i$i)) + 8|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $57 = ($56|0)==($41|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           if ($57) {
            break L13;
           } else {
            $$01$i$i = $56;$50 = $55;
           }
          }
         }
        }
        $58 = ___cxa_find_matching_catch_2()|0;
        $59 = tempRet0;
        $60 = HEAP32[$22>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $61 = ($60|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($61)) {
         $62 = HEAP32[$24>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $63 = ($62|0)==($60|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if (!($63)) {
          $__new_last1$i$i$i$i41 = $60; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $scevgep$i$i$i$i42 = ((($62)) + -8|0);
          $64 = $scevgep$i$i$i$i42;
          $65 = (($64) - ($__new_last1$i$i$i$i41))|0;
          $66 = $65 >>> 3;
          $67 = $66 ^ -1;
          $scevgep4$i$i$i$i43 = (($62) + ($67<<3)|0);
          HEAP32[$24>>2] = $scevgep4$i$i$i$i43; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         }
         __ZdlPv($60); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        }
        $68 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $69 = ($68|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $70 = $68; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($69) {
         $eh$lpad$body$index3Z2D = $59;$eh$lpad$body$indexZ2D = $58;
         label = 46;
         break L10;
        }
        $71 = ((($0)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $72 = HEAP32[$71>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $73 = ($72|0)==($68|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($73)) {
         $scevgep$i$i$i$i2 = ((($72)) + -4|0);
         $74 = $scevgep$i$i$i$i2;
         $75 = (($74) - ($70))|0;
         $76 = $75 >>> 2;
         $77 = $76 ^ -1;
         $scevgep4$i$i$i$i3 = (($72) + ($77<<2)|0);
         HEAP32[$71>>2] = $scevgep4$i$i$i$i3; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
        __ZdlPv($68); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        $eh$lpad$body$index3Z2D = $59;$eh$lpad$body$indexZ2D = $58;
        label = 46;
        break L10;
       }
      } while(0);
      $78 = ((($0)) + 24|0); //@line 184 "modules/graphics/opengl/Text.cpp"
      HEAPF32[$78>>2] = $wrap; //@line 184 "modules/graphics/opengl/Text.cpp"
      $79 = ((($0)) + 28|0); //@line 184 "modules/graphics/opengl/Text.cpp"
      HEAP32[$79>>2] = $align; //@line 184 "modules/graphics/opengl/Text.cpp"
      $80 = ((($0)) + 32|0); //@line 184 "modules/graphics/opengl/Text.cpp"
      $81 = ((($0)) + 44|0); //@line 184 "modules/graphics/opengl/Text.cpp"
      ;HEAP32[$80>>2]=0|0;HEAP32[$80+4>>2]=0|0;HEAP16[$80+8>>1]=0|0; //@line 184 "modules/graphics/opengl/Text.cpp"
      __THREW__ = 0;
      invoke_vi(4102,($81|0)); //@line 184 "modules/graphics/opengl/Text.cpp"
      $82 = __THREW__; __THREW__ = 0;
      $83 = $82&1;
      if ($83) {
       $126 = ___cxa_find_matching_catch_2()|0;
       $127 = tempRet0;
       $128 = ((($0)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $129 = HEAP32[$128>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $130 = ($129|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($130)) {
        $131 = HEAP32[$24>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $132 = ($131|0)==($129|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($132)) {
         $__new_last1$i$i$i$i29 = $129; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $scevgep$i$i$i$i30 = ((($131)) + -8|0);
         $133 = $scevgep$i$i$i$i30;
         $134 = (($133) - ($__new_last1$i$i$i$i29))|0;
         $135 = $134 >>> 3;
         $136 = $135 ^ -1;
         $scevgep4$i$i$i$i31 = (($131) + ($136<<3)|0);
         HEAP32[$24>>2] = $scevgep4$i$i$i$i31; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
        __ZdlPv($129); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       }
       $137 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $138 = ($137|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $139 = $137; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($138) {
        $$0 = $127;$$01 = $126;
        break;
       }
       $140 = ((($0)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $141 = HEAP32[$140>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $142 = ($141|0)==($137|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($142)) {
        $scevgep$i$i$i1$i35 = ((($141)) + -4|0);
        $143 = $scevgep$i$i$i1$i35;
        $144 = (($143) - ($139))|0;
        $145 = $144 >>> 2;
        $146 = $145 ^ -1;
        $scevgep4$i$i$i2$i36 = (($141) + ($146<<2)|0);
        HEAP32[$140>>2] = $scevgep4$i$i$i2$i36; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       __ZdlPv($137); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       $$0 = $127;$$01 = $126;
       break;
      }
      __THREW__ = 0;
      invoke_vii(4143,($this|0),($0|0)); //@line 184 "modules/graphics/opengl/Text.cpp"
      $84 = __THREW__; __THREW__ = 0;
      $85 = $84&1;
      if ($85) {
       $147 = ___cxa_find_matching_catch_2()|0;
       $148 = tempRet0;
       __ZN4love7Matrix3D2Ev($81); //@line 70 "modules/graphics/opengl/Text.h"
       $149 = ((($0)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $150 = HEAP32[$149>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $151 = ($150|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($151)) {
        $152 = HEAP32[$24>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $153 = ($152|0)==($150|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($153)) {
         $__new_last1$i$i$i$i$i = $150; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $scevgep$i$i$i$i$i = ((($152)) + -8|0);
         $154 = $scevgep$i$i$i$i$i;
         $155 = (($154) - ($__new_last1$i$i$i$i$i))|0;
         $156 = $155 >>> 3;
         $157 = $156 ^ -1;
         $scevgep4$i$i$i$i$i = (($152) + ($157<<3)|0);
         HEAP32[$24>>2] = $scevgep4$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
        __ZdlPv($150); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       }
       $158 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $159 = ($158|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $160 = $158; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($159) {
        $$0 = $148;$$01 = $147;
        break;
       }
       $161 = ((($0)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $162 = HEAP32[$161>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $163 = ($162|0)==($158|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($163)) {
        $scevgep$i$i$i1$i$i = ((($162)) + -4|0);
        $164 = $scevgep$i$i$i1$i$i;
        $165 = (($164) - ($160))|0;
        $166 = $165 >>> 2;
        $167 = $166 ^ -1;
        $scevgep4$i$i$i2$i$i = (($162) + ($167<<2)|0);
        HEAP32[$161>>2] = $scevgep4$i$i$i2$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       __ZdlPv($158); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       $$0 = $148;$$01 = $147;
       break;
      }
      __ZN4love7Matrix3D2Ev($81); //@line 70 "modules/graphics/opengl/Text.h"
      $86 = ((($0)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $87 = HEAP32[$86>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $88 = ($87|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($88)) {
       $89 = HEAP32[$24>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $90 = ($89|0)==($87|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($90)) {
        $__new_last1$i$i$i$i$i6 = $87; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $scevgep$i$i$i$i$i7 = ((($89)) + -8|0);
        $91 = $scevgep$i$i$i$i$i7;
        $92 = (($91) - ($__new_last1$i$i$i$i$i6))|0;
        $93 = $92 >>> 3;
        $94 = $93 ^ -1;
        $scevgep4$i$i$i$i$i8 = (($89) + ($94<<3)|0);
        HEAP32[$24>>2] = $scevgep4$i$i$i$i$i8; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       __ZdlPv($87); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      }
      $95 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $96 = ($95|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $97 = $95; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($96)) {
       $98 = ((($0)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $99 = HEAP32[$98>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $100 = ($99|0)==($95|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($100)) {
        $scevgep$i$i$i1$i$i12 = ((($99)) + -4|0);
        $101 = $scevgep$i$i$i1$i$i12;
        $102 = (($101) - ($97))|0;
        $103 = $102 >>> 2;
        $104 = $103 ^ -1;
        $scevgep4$i$i$i2$i$i13 = (($99) + ($104<<2)|0);
        HEAP32[$98>>2] = $scevgep4$i$i$i2$i$i13; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       __ZdlPv($95); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      }
      $105 = ((($codepoints)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $106 = HEAP32[$105>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $107 = ($106|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($107)) {
       $108 = HEAP32[$26>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $109 = ($108|0)==($106|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($109)) {
        $__new_last1$i$i$i$i17 = $106; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $scevgep$i$i$i$i18 = ((($108)) + -8|0);
        $110 = $scevgep$i$i$i$i18;
        $111 = (($110) - ($__new_last1$i$i$i$i17))|0;
        $112 = $111 >>> 3;
        $113 = $112 ^ -1;
        $scevgep4$i$i$i$i19 = (($108) + ($113<<3)|0);
        HEAP32[$26>>2] = $scevgep4$i$i$i$i19; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       __ZdlPv($106); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      }
      $114 = HEAP32[$codepoints>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $115 = ($114|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $116 = $114; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($115)) {
       $117 = ((($codepoints)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $118 = HEAP32[$117>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $119 = ($118|0)==($114|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($119)) {
        $scevgep$i$i$i1$i23 = ((($118)) + -4|0);
        $120 = $scevgep$i$i$i1$i23;
        $121 = (($120) - ($116))|0;
        $122 = $121 >>> 2;
        $123 = $122 ^ -1;
        $scevgep4$i$i$i2$i24 = (($118) + ($123<<2)|0);
        HEAP32[$117>>2] = $scevgep4$i$i$i2$i24; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       __ZdlPv($114); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      }
      STACKTOP = sp;return; //@line 185 "modules/graphics/opengl/Text.cpp"
     }
    }
   } while(0);
   if ((label|0) == 45) {
    $124 = ___cxa_find_matching_catch_2()|0;
    $125 = tempRet0;
    $eh$lpad$body$index3Z2D = $125;$eh$lpad$body$indexZ2D = $124;
    label = 46;
   }
   if ((label|0) == 46) {
    $$0 = $eh$lpad$body$index3Z2D;$$01 = $eh$lpad$body$indexZ2D;
   }
   $168 = ((($codepoints)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $169 = HEAP32[$168>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $170 = ($169|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($170)) {
    $171 = ((($codepoints)) + 16|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $172 = HEAP32[$171>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $173 = ($172|0)==($169|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($173)) {
     $__new_last1$i$i$i$i = $169; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $scevgep$i$i$i$i = ((($172)) + -8|0);
     $174 = $scevgep$i$i$i$i;
     $175 = (($174) - ($__new_last1$i$i$i$i))|0;
     $176 = $175 >>> 3;
     $177 = $176 ^ -1;
     $scevgep4$i$i$i$i = (($172) + ($177<<3)|0);
     HEAP32[$171>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($169); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $178 = HEAP32[$codepoints>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $179 = ($178|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $180 = $178; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($179) {
    ___resumeException($$01|0);
    // unreachable;
   }
   $181 = ((($codepoints)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $182 = HEAP32[$181>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $183 = ($182|0)==($178|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($183)) {
    $scevgep$i$i$i1$i = ((($182)) + -4|0);
    $184 = $scevgep$i$i$i1$i;
    $185 = (($184) - ($180))|0;
    $186 = $185 >>> 2;
    $187 = $186 ^ -1;
    $scevgep4$i$i$i2$i = (($182) + ($187<<2)|0);
    HEAP32[$181>>2] = $scevgep4$i$i$i2$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($178); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($$01|0);
   // unreachable;
  }
 } while(0);
 __ZN4love8graphics6opengl4Text5clearEv($this); //@line 189 "modules/graphics/opengl/Text.cpp"
 STACKTOP = sp;return; //@line 185 "modules/graphics/opengl/Text.cpp"
}
function __ZN4love8graphics6opengl4Text3setERKNSt3__16vectorINS1_4Font13ColoredStringENS3_9allocatorIS6_EEEE($this,$text) {
 $this = $this|0;
 $text = $text|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl4Text3setERKNSt3__16vectorINS1_4Font13ColoredStringENS3_9allocatorIS6_EEEEfNS5_9AlignModeE($this,$text,-1.0,4); //@line 173 "modules/graphics/opengl/Text.cpp"
 return; //@line 173 "modules/graphics/opengl/Text.cpp"
}
function __ZN4love8graphics6opengl4Text3setEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl4Text5clearEv($this); //@line 189 "modules/graphics/opengl/Text.cpp"
 return; //@line 190 "modules/graphics/opengl/Text.cpp"
}
function __ZN4love8graphics6opengl4Text3addERKNSt3__16vectorINS1_4Font13ColoredStringENS3_9allocatorIS6_EEEEfffffffff($this,$text,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $text = $text|0;
 $x = +$x;
 $y = +$y;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love8graphics6opengl4Text4addfERKNSt3__16vectorINS1_4Font13ColoredStringENS3_9allocatorIS6_EEEEfNS5_9AlignModeEfffffffff($this,$text,-1.0,4,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky)|0); //@line 194 "modules/graphics/opengl/Text.cpp"
 return ($0|0); //@line 194 "modules/graphics/opengl/Text.cpp"
}
function __ZN4love8graphics6opengl4Text4addfERKNSt3__16vectorINS1_4Font13ColoredStringENS3_9allocatorIS6_EEEEfNS5_9AlignModeEfffffffff($this,$text,$wrap,$align,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $text = $text|0;
 $wrap = +$wrap;
 $align = $align|0;
 $x = +$x;
 $y = +$y;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $$0 = 0, $$01 = 0, $$01$i$i = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__new_last1$i$i$i$i = 0, $__new_last1$i$i$i$i$i = 0, $__new_last1$i$i$i$i$i19 = 0, $__new_last1$i$i$i$i30 = 0, $__new_last1$i$i$i$i7 = 0, $codepoints = 0, $eh$lpad$body$index3Z2D = 0, $eh$lpad$body$indexZ2D = 0, $m = 0, $scevgep$i$i$i$i = 0, $scevgep$i$i$i$i$i = 0, $scevgep$i$i$i$i$i20 = 0, $scevgep$i$i$i$i3 = 0;
 var $scevgep$i$i$i$i31 = 0, $scevgep$i$i$i$i8 = 0, $scevgep$i$i$i1$i = 0, $scevgep$i$i$i1$i$i = 0, $scevgep$i$i$i1$i$i25 = 0, $scevgep$i$i$i1$i13 = 0, $scevgep4$i$i$i$i = 0, $scevgep4$i$i$i$i$i = 0, $scevgep4$i$i$i$i$i21 = 0, $scevgep4$i$i$i$i32 = 0, $scevgep4$i$i$i$i4 = 0, $scevgep4$i$i$i$i9 = 0, $scevgep4$i$i$i2$i = 0, $scevgep4$i$i$i2$i$i = 0, $scevgep4$i$i$i2$i$i26 = 0, $scevgep4$i$i$i2$i14 = 0, dest = 0, label = 0, sp = 0, src = 0;
 var stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $codepoints = sp + 120|0;
 $m = sp + 80|0;
 $0 = sp;
 ;HEAP32[$codepoints>>2]=0|0;HEAP32[$codepoints+4>>2]=0|0;HEAP32[$codepoints+8>>2]=0|0;HEAP32[$codepoints+12>>2]=0|0;HEAP32[$codepoints+16>>2]=0|0;HEAP32[$codepoints+20>>2]=0|0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 __THREW__ = 0;
 invoke_vii(4038,($text|0),($codepoints|0)); //@line 200 "modules/graphics/opengl/Text.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  label = 37;
 } else {
  __THREW__ = 0;
  invoke_viddddddddd(4147,($m|0),(+$x),(+$y),(+$angle),(+$sx),(+$sy),(+$ox),(+$oy),(+$kx),(+$ky)); //@line 202 "modules/graphics/opengl/Text.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if ($4) {
   label = 37;
  } else {
   __THREW__ = 0;
   invoke_vii(4032,($0|0),($codepoints|0)); //@line 76 "modules/graphics/opengl/Font.h"
   $5 = __THREW__; __THREW__ = 0;
   $6 = $5&1;
   L4: do {
    if ($6) {
     $119 = ___cxa_find_matching_catch_2()|0;
     $120 = tempRet0;
     $eh$lpad$body$index3Z2D = $120;$eh$lpad$body$indexZ2D = $119;
     label = 39;
    } else {
     $7 = ((($0)) + 12|0); //@line 76 "modules/graphics/opengl/Font.h"
     $8 = ((($codepoints)) + 12|0); //@line 76 "modules/graphics/opengl/Font.h"
     HEAP32[$7>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $9 = ((($0)) + 16|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$9>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $10 = ((($0)) + 20|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$10>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $11 = ((($codepoints)) + 16|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $12 = HEAP32[$11>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $13 = HEAP32[$8>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $14 = (($12) - ($13))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $15 = $14 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $16 = ($15|0)==(0); //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     L7: do {
      if (!($16)) {
       $17 = ($15>>>0)>(536870911); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($17) {
        __THREW__ = 0;
        invoke_vi(3528,($7|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $18 = __THREW__; __THREW__ = 0;
        $19 = $18&1;
        if (!($19)) {
         label = 7;
        }
       } else {
        label = 7;
       }
       if ((label|0) == 7) {
        __THREW__ = 0;
        $20 = (invoke_ii(3348,($14|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        $21 = __THREW__; __THREW__ = 0;
        $22 = $21&1;
        if (!($22)) {
         HEAP32[$9>>2] = $20; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$7>>2] = $20; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $23 = (($20) + ($15<<3)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $24 = ((($0)) + 20|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         HEAP32[$24>>2] = $23; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $25 = HEAP32[$8>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $26 = HEAP32[$11>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $27 = ($25|0)==($26|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($27) {
          break;
         }
         $$01$i$i = $25;$35 = $20;
         while(1) {
          $28 = $$01$i$i; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $29 = $28; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $30 = HEAP32[$29>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $31 = (($28) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $32 = $31; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $33 = HEAP32[$32>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $34 = $35; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $36 = $34; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          HEAP32[$36>>2] = $30; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $37 = (($34) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $38 = $37; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          HEAP32[$38>>2] = $33; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $39 = HEAP32[$9>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $40 = ((($39)) + 8|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$9>>2] = $40; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $41 = ((($$01$i$i)) + 8|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $42 = ($41|0)==($26|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          if ($42) {
           break L7;
          } else {
           $$01$i$i = $41;$35 = $40;
          }
         }
        }
       }
       $43 = ___cxa_find_matching_catch_2()|0;
       $44 = tempRet0;
       $45 = HEAP32[$7>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $46 = ($45|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($46)) {
        $47 = HEAP32[$9>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $48 = ($47|0)==($45|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($48)) {
         $__new_last1$i$i$i$i30 = $45; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $scevgep$i$i$i$i31 = ((($47)) + -8|0);
         $49 = $scevgep$i$i$i$i31;
         $50 = (($49) - ($__new_last1$i$i$i$i30))|0;
         $51 = $50 >>> 3;
         $52 = $51 ^ -1;
         $scevgep4$i$i$i$i32 = (($47) + ($52<<3)|0);
         HEAP32[$9>>2] = $scevgep4$i$i$i$i32; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
        __ZdlPv($45); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       }
       $53 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $54 = ($53|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $55 = $53; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($54) {
        $eh$lpad$body$index3Z2D = $44;$eh$lpad$body$indexZ2D = $43;
        label = 39;
        break L4;
       }
       $56 = ((($0)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $57 = HEAP32[$56>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $58 = ($57|0)==($53|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($58)) {
        $scevgep$i$i$i$i3 = ((($57)) + -4|0);
        $59 = $scevgep$i$i$i$i3;
        $60 = (($59) - ($55))|0;
        $61 = $60 >>> 2;
        $62 = $61 ^ -1;
        $scevgep4$i$i$i$i4 = (($57) + ($62<<2)|0);
        HEAP32[$56>>2] = $scevgep4$i$i$i$i4; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       __ZdlPv($53); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       $eh$lpad$body$index3Z2D = $44;$eh$lpad$body$indexZ2D = $43;
       label = 39;
       break L4;
      }
     } while(0);
     $63 = ((($0)) + 24|0); //@line 204 "modules/graphics/opengl/Text.cpp"
     HEAPF32[$63>>2] = $wrap; //@line 204 "modules/graphics/opengl/Text.cpp"
     $64 = ((($0)) + 28|0); //@line 204 "modules/graphics/opengl/Text.cpp"
     HEAP32[$64>>2] = $align; //@line 204 "modules/graphics/opengl/Text.cpp"
     $65 = ((($0)) + 32|0); //@line 204 "modules/graphics/opengl/Text.cpp"
     HEAP32[$65>>2] = 0; //@line 204 "modules/graphics/opengl/Text.cpp"
     $66 = ((($0)) + 36|0); //@line 204 "modules/graphics/opengl/Text.cpp"
     HEAP32[$66>>2] = 0; //@line 204 "modules/graphics/opengl/Text.cpp"
     $67 = ((($0)) + 40|0); //@line 204 "modules/graphics/opengl/Text.cpp"
     HEAP8[$67>>0] = 1; //@line 204 "modules/graphics/opengl/Text.cpp"
     $68 = ((($0)) + 41|0); //@line 204 "modules/graphics/opengl/Text.cpp"
     HEAP8[$68>>0] = 1; //@line 204 "modules/graphics/opengl/Text.cpp"
     $69 = ((($0)) + 44|0); //@line 204 "modules/graphics/opengl/Text.cpp"
     dest=$69; src=$m; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 204 "modules/graphics/opengl/Text.cpp"
     __THREW__ = 0;
     invoke_vii(4143,($this|0),($0|0)); //@line 204 "modules/graphics/opengl/Text.cpp"
     $70 = __THREW__; __THREW__ = 0;
     $71 = $70&1;
     if ($71) {
      $121 = ___cxa_find_matching_catch_2()|0;
      $122 = tempRet0;
      __ZN4love7Matrix3D2Ev($69); //@line 70 "modules/graphics/opengl/Text.h"
      $123 = ((($0)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $124 = HEAP32[$123>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $125 = ($124|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($125)) {
       $126 = HEAP32[$9>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $127 = ($126|0)==($124|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($127)) {
        $__new_last1$i$i$i$i$i19 = $124; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $scevgep$i$i$i$i$i20 = ((($126)) + -8|0);
        $128 = $scevgep$i$i$i$i$i20;
        $129 = (($128) - ($__new_last1$i$i$i$i$i19))|0;
        $130 = $129 >>> 3;
        $131 = $130 ^ -1;
        $scevgep4$i$i$i$i$i21 = (($126) + ($131<<3)|0);
        HEAP32[$9>>2] = $scevgep4$i$i$i$i$i21; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       __ZdlPv($124); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      }
      $132 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $133 = ($132|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $134 = $132; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($133) {
       $$0 = $122;$$01 = $121;
       break;
      }
      $135 = ((($0)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $136 = HEAP32[$135>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $137 = ($136|0)==($132|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($137)) {
       $scevgep$i$i$i1$i$i25 = ((($136)) + -4|0);
       $138 = $scevgep$i$i$i1$i$i25;
       $139 = (($138) - ($134))|0;
       $140 = $139 >>> 2;
       $141 = $140 ^ -1;
       $scevgep4$i$i$i2$i$i26 = (($136) + ($141<<2)|0);
       HEAP32[$135>>2] = $scevgep4$i$i$i2$i$i26; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($132); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $$0 = $122;$$01 = $121;
      break;
     }
     __ZN4love7Matrix3D2Ev($69); //@line 70 "modules/graphics/opengl/Text.h"
     $72 = ((($0)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $73 = HEAP32[$72>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $74 = ($73|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($74)) {
      $75 = HEAP32[$9>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $76 = ($75|0)==($73|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($76)) {
       $__new_last1$i$i$i$i$i = $73; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $scevgep$i$i$i$i$i = ((($75)) + -8|0);
       $77 = $scevgep$i$i$i$i$i;
       $78 = (($77) - ($__new_last1$i$i$i$i$i))|0;
       $79 = $78 >>> 3;
       $80 = $79 ^ -1;
       $scevgep4$i$i$i$i$i = (($75) + ($80<<3)|0);
       HEAP32[$9>>2] = $scevgep4$i$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($73); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $81 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $82 = ($81|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $83 = $81; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($82)) {
      $84 = ((($0)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $85 = HEAP32[$84>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $86 = ($85|0)==($81|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($86)) {
       $scevgep$i$i$i1$i$i = ((($85)) + -4|0);
       $87 = $scevgep$i$i$i1$i$i;
       $88 = (($87) - ($83))|0;
       $89 = $88 >>> 2;
       $90 = $89 ^ -1;
       $scevgep4$i$i$i2$i$i = (($85) + ($90<<2)|0);
       HEAP32[$84>>2] = $scevgep4$i$i$i2$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($81); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $91 = ((($this)) + 28|0); //@line 206 "modules/graphics/opengl/Text.cpp"
     $92 = ((($this)) + 32|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $93 = HEAP32[$92>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $94 = HEAP32[$91>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $95 = (($93) - ($94))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $96 = (($95|0) / 80)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $97 = (($96) + -1)|0; //@line 206 "modules/graphics/opengl/Text.cpp"
     __ZN4love7Matrix3D2Ev($m); //@line 207 "modules/graphics/opengl/Text.cpp"
     $98 = ((($codepoints)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $99 = HEAP32[$98>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $100 = ($99|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($100)) {
      $101 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $102 = ($101|0)==($99|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($102)) {
       $__new_last1$i$i$i$i7 = $99; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $scevgep$i$i$i$i8 = ((($101)) + -8|0);
       $103 = $scevgep$i$i$i$i8;
       $104 = (($103) - ($__new_last1$i$i$i$i7))|0;
       $105 = $104 >>> 3;
       $106 = $105 ^ -1;
       $scevgep4$i$i$i$i9 = (($101) + ($106<<3)|0);
       HEAP32[$11>>2] = $scevgep4$i$i$i$i9; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($99); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $107 = HEAP32[$codepoints>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $108 = ($107|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $109 = $107; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($108) {
      STACKTOP = sp;return ($97|0); //@line 207 "modules/graphics/opengl/Text.cpp"
     }
     $110 = ((($codepoints)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $111 = HEAP32[$110>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $112 = ($111|0)==($107|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($112)) {
      $scevgep$i$i$i1$i13 = ((($111)) + -4|0);
      $113 = $scevgep$i$i$i1$i13;
      $114 = (($113) - ($109))|0;
      $115 = $114 >>> 2;
      $116 = $115 ^ -1;
      $scevgep4$i$i$i2$i14 = (($111) + ($116<<2)|0);
      HEAP32[$110>>2] = $scevgep4$i$i$i2$i14; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __ZdlPv($107); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     STACKTOP = sp;return ($97|0); //@line 207 "modules/graphics/opengl/Text.cpp"
    }
   } while(0);
   if ((label|0) == 39) {
    $$0 = $eh$lpad$body$index3Z2D;$$01 = $eh$lpad$body$indexZ2D;
   }
   __ZN4love7Matrix3D2Ev($m); //@line 207 "modules/graphics/opengl/Text.cpp"
   $$1 = $$0;$$12 = $$01;
  }
 }
 if ((label|0) == 37) {
  $117 = ___cxa_find_matching_catch_2()|0;
  $118 = tempRet0;
  $$1 = $118;$$12 = $117;
 }
 $142 = ((($codepoints)) + 12|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $143 = HEAP32[$142>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $144 = ($143|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($144)) {
  $145 = ((($codepoints)) + 16|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $146 = HEAP32[$145>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $147 = ($146|0)==($143|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($147)) {
   $__new_last1$i$i$i$i = $143; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $scevgep$i$i$i$i = ((($146)) + -8|0);
   $148 = $scevgep$i$i$i$i;
   $149 = (($148) - ($__new_last1$i$i$i$i))|0;
   $150 = $149 >>> 3;
   $151 = $150 ^ -1;
   $scevgep4$i$i$i$i = (($146) + ($151<<3)|0);
   HEAP32[$145>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($143); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $152 = HEAP32[$codepoints>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $153 = ($152|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $154 = $152; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($153) {
  ___resumeException($$12|0);
  // unreachable;
 }
 $155 = ((($codepoints)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $156 = HEAP32[$155>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $157 = ($156|0)==($152|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($157)) {
  $scevgep$i$i$i1$i = ((($156)) + -4|0);
  $158 = $scevgep$i$i$i1$i;
  $159 = (($158) - ($154))|0;
  $160 = $159 >>> 2;
  $161 = $160 ^ -1;
  $scevgep4$i$i$i2$i = (($156) + ($161<<2)|0);
  HEAP32[$155>>2] = $scevgep4$i$i$i2$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($152); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$12|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8graphics6opengl4Text7setFontEPNS1_4FontE($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($f|0)==(0|0); //@line 132 "./common/Object.h"
 if (!($0)) {
  $1 = HEAP32[$f>>2]|0; //@line 132 "./common/Object.h"
  $2 = ((($1)) + 8|0); //@line 132 "./common/Object.h"
  $3 = HEAP32[$2>>2]|0; //@line 132 "./common/Object.h"
  FUNCTION_TABLE_vi[$3 & 8191]($f); //@line 132 "./common/Object.h"
 }
 $4 = ((($this)) + 8|0); //@line 133 "./common/Object.h"
 $5 = HEAP32[$4>>2]|0; //@line 133 "./common/Object.h"
 $6 = ($5|0)==(0|0); //@line 133 "./common/Object.h"
 if (!($6)) {
  $7 = HEAP32[$5>>2]|0; //@line 133 "./common/Object.h"
  $8 = ((($7)) + 12|0); //@line 133 "./common/Object.h"
  $9 = HEAP32[$8>>2]|0; //@line 133 "./common/Object.h"
  FUNCTION_TABLE_vi[$9 & 8191]($5); //@line 133 "./common/Object.h"
 }
 HEAP32[$4>>2] = $f; //@line 134 "./common/Object.h"
 $10 = ((($this)) + 44|0); //@line 257 "modules/graphics/opengl/Text.cpp"
 HEAP32[$10>>2] = -1; //@line 257 "modules/graphics/opengl/Text.cpp"
 __ZN4love8graphics6opengl4Text18regenerateVerticesEv($this); //@line 258 "modules/graphics/opengl/Text.cpp"
 return; //@line 259 "modules/graphics/opengl/Text.cpp"
}
function __ZNK4love8graphics6opengl4Text7getFontEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 139 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 139 "./common/Object.h"
 return ($1|0); //@line 263 "modules/graphics/opengl/Text.cpp"
}
function __ZNK4love8graphics6opengl4Text8getWidthEi($this,$index) {
 $this = $this|0;
 $index = $index|0;
 var $$0 = 0, $$01 = 0, $$in = 0, $$phi$trans$insert = 0, $$phi$trans$insert3 = 0, $$pre = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($index|0)<(0); //@line 268 "modules/graphics/opengl/Text.cpp"
 if ($0) {
  $1 = ((($this)) + 28|0); //@line 269 "modules/graphics/opengl/Text.cpp"
  $2 = ((($this)) + 32|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $6 = (($5|0) / 80)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $7 = (($6) + -1)|0; //@line 269 "modules/graphics/opengl/Text.cpp"
  $8 = ($7|0)<(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $9 = $8 ? 0 : $7; //@line 269 "modules/graphics/opengl/Text.cpp"
  $$0 = $9;$$in = $4;$11 = $3;
 } else {
  $$phi$trans$insert = ((($this)) + 32|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$phi$trans$insert3 = ((($this)) + 28|0); //@line 271 "modules/graphics/opengl/Text.cpp"
  $$pre5 = HEAP32[$$phi$trans$insert3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0 = $index;$$in = $$pre5;$11 = $$pre;
 }
 $10 = (($11) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = (($10|0) / 80)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ($$0|0)<($12|0); //@line 271 "modules/graphics/opengl/Text.cpp"
 if (!($13)) {
  $$01 = 0;
  return ($$01|0); //@line 275 "modules/graphics/opengl/Text.cpp"
 }
 $14 = $$in; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = (((($14) + (($$0*80)|0)|0)) + 32|0); //@line 274 "modules/graphics/opengl/Text.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 274 "modules/graphics/opengl/Text.cpp"
 $$01 = $16;
 return ($$01|0); //@line 275 "modules/graphics/opengl/Text.cpp"
}
function __ZNK4love8graphics6opengl4Text9getHeightEi($this,$index) {
 $this = $this|0;
 $index = $index|0;
 var $$0 = 0, $$01 = 0, $$in = 0, $$phi$trans$insert = 0, $$phi$trans$insert3 = 0, $$pre = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($index|0)<(0); //@line 279 "modules/graphics/opengl/Text.cpp"
 if ($0) {
  $1 = ((($this)) + 28|0); //@line 280 "modules/graphics/opengl/Text.cpp"
  $2 = ((($this)) + 32|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $6 = (($5|0) / 80)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $7 = (($6) + -1)|0; //@line 280 "modules/graphics/opengl/Text.cpp"
  $8 = ($7|0)<(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $9 = $8 ? 0 : $7; //@line 280 "modules/graphics/opengl/Text.cpp"
  $$0 = $9;$$in = $4;$11 = $3;
 } else {
  $$phi$trans$insert = ((($this)) + 32|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$phi$trans$insert3 = ((($this)) + 28|0); //@line 282 "modules/graphics/opengl/Text.cpp"
  $$pre5 = HEAP32[$$phi$trans$insert3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0 = $index;$$in = $$pre5;$11 = $$pre;
 }
 $10 = (($11) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = (($10|0) / 80)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ($$0|0)<($12|0); //@line 282 "modules/graphics/opengl/Text.cpp"
 if (!($13)) {
  $$01 = 0;
  return ($$01|0); //@line 286 "modules/graphics/opengl/Text.cpp"
 }
 $14 = $$in; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = (((($14) + (($$0*80)|0)|0)) + 36|0); //@line 285 "modules/graphics/opengl/Text.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 285 "modules/graphics/opengl/Text.cpp"
 $$01 = $16;
 return ($$01|0); //@line 286 "modules/graphics/opengl/Text.cpp"
}
function __ZN4love8graphics6opengl5VideoD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (25484); //@line 71 "modules/graphics/opengl/Video.cpp"
 $0 = ((($this)) + 8|0); //@line 71 "modules/graphics/opengl/Video.cpp"
 HEAP32[$0>>2] = (25520); //@line 71 "modules/graphics/opengl/Video.cpp"
 $1 = ((($this)) + 20|0); //@line 107 "modules/graphics/opengl/Video.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 107 "modules/graphics/opengl/Video.cpp"
 __THREW__ = 0;
 invoke_vii(4014,(634044|0),($2|0)); //@line 107 "modules/graphics/opengl/Video.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  HEAP32[$1>>2] = 0; //@line 108 "modules/graphics/opengl/Video.cpp"
  $5 = ((($this)) + 24|0); //@line 107 "modules/graphics/opengl/Video.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 107 "modules/graphics/opengl/Video.cpp"
  __THREW__ = 0;
  invoke_vii(4014,(634044|0),($6|0)); //@line 107 "modules/graphics/opengl/Video.cpp"
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if (!($8)) {
   HEAP32[$5>>2] = 0; //@line 108 "modules/graphics/opengl/Video.cpp"
   $9 = ((($this)) + 28|0); //@line 107 "modules/graphics/opengl/Video.cpp"
   $10 = HEAP32[$9>>2]|0; //@line 107 "modules/graphics/opengl/Video.cpp"
   __THREW__ = 0;
   invoke_vii(4014,(634044|0),($10|0)); //@line 107 "modules/graphics/opengl/Video.cpp"
   $11 = __THREW__; __THREW__ = 0;
   $12 = $11&1;
   if (!($12)) {
    HEAP32[$9>>2] = 0; //@line 108 "modules/graphics/opengl/Video.cpp"
    $13 = ((($this)) + 16|0); //@line 106 "./common/Object.h"
    $14 = HEAP32[$13>>2]|0; //@line 106 "./common/Object.h"
    $15 = ($14|0)==(0|0); //@line 106 "./common/Object.h"
    if (!($15)) {
     $16 = HEAP32[$14>>2]|0; //@line 106 "./common/Object.h"
     $17 = ((($16)) + 12|0); //@line 106 "./common/Object.h"
     $18 = HEAP32[$17>>2]|0; //@line 106 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($18|0,($14|0)); //@line 106 "./common/Object.h"
     $19 = __THREW__; __THREW__ = 0;
     $20 = $19&1;
     if ($20) {
      $21 = ___cxa_find_matching_catch_3(0|0)|0;
      $22 = tempRet0;
      ___clang_call_terminate($21); //@line 106 "./common/Object.h"
      // unreachable; //@line 106 "./common/Object.h"
     }
    }
    $23 = ((($this)) + 12|0); //@line 106 "./common/Object.h"
    $24 = HEAP32[$23>>2]|0; //@line 106 "./common/Object.h"
    $25 = ($24|0)==(0|0); //@line 106 "./common/Object.h"
    if ($25) {
     __ZN4love8graphics8VolatileD2Ev($0); //@line 73 "modules/graphics/opengl/Video.cpp"
     return; //@line 73 "modules/graphics/opengl/Video.cpp"
    }
    $26 = HEAP32[$24>>2]|0; //@line 106 "./common/Object.h"
    $27 = ((($26)) + 12|0); //@line 106 "./common/Object.h"
    $28 = HEAP32[$27>>2]|0; //@line 106 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($28|0,($24|0)); //@line 106 "./common/Object.h"
    $29 = __THREW__; __THREW__ = 0;
    $30 = $29&1;
    if ($30) {
     $31 = ___cxa_find_matching_catch_3(0|0)|0;
     $32 = tempRet0;
     ___clang_call_terminate($31); //@line 106 "./common/Object.h"
     // unreachable; //@line 106 "./common/Object.h"
    } else {
     __ZN4love8graphics8VolatileD2Ev($0); //@line 73 "modules/graphics/opengl/Video.cpp"
     return; //@line 73 "modules/graphics/opengl/Video.cpp"
    }
   }
  }
 }
 $33 = ___cxa_find_matching_catch_3(0|0)|0;
 $34 = tempRet0;
 $35 = ((($this)) + 16|0); //@line 106 "./common/Object.h"
 $36 = HEAP32[$35>>2]|0; //@line 106 "./common/Object.h"
 $37 = ($36|0)==(0|0); //@line 106 "./common/Object.h"
 if (!($37)) {
  $38 = HEAP32[$36>>2]|0; //@line 106 "./common/Object.h"
  $39 = ((($38)) + 12|0); //@line 106 "./common/Object.h"
  $40 = HEAP32[$39>>2]|0; //@line 106 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($40|0,($36|0)); //@line 106 "./common/Object.h"
  $41 = __THREW__; __THREW__ = 0;
  $42 = $41&1;
  if ($42) {
   $43 = ___cxa_find_matching_catch_3(0|0)|0;
   $44 = tempRet0;
   ___clang_call_terminate($43); //@line 106 "./common/Object.h"
   // unreachable; //@line 106 "./common/Object.h"
  }
 }
 $45 = ((($this)) + 12|0); //@line 106 "./common/Object.h"
 $46 = HEAP32[$45>>2]|0; //@line 106 "./common/Object.h"
 $47 = ($46|0)==(0|0); //@line 106 "./common/Object.h"
 if ($47) {
  __ZN4love8graphics8VolatileD2Ev($0); //@line 73 "modules/graphics/opengl/Video.cpp"
  ___clang_call_terminate($33); //@line 73 "modules/graphics/opengl/Video.cpp"
  // unreachable; //@line 73 "modules/graphics/opengl/Video.cpp"
 }
 $48 = HEAP32[$46>>2]|0; //@line 106 "./common/Object.h"
 $49 = ((($48)) + 12|0); //@line 106 "./common/Object.h"
 $50 = HEAP32[$49>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($50|0,($46|0)); //@line 106 "./common/Object.h"
 $51 = __THREW__; __THREW__ = 0;
 $52 = $51&1;
 if ($52) {
  $53 = ___cxa_find_matching_catch_3(0|0)|0;
  $54 = tempRet0;
  ___clang_call_terminate($53); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  __ZN4love8graphics8VolatileD2Ev($0); //@line 73 "modules/graphics/opengl/Video.cpp"
  ___clang_call_terminate($33); //@line 73 "modules/graphics/opengl/Video.cpp"
  // unreachable; //@line 73 "modules/graphics/opengl/Video.cpp"
 }
}
function __ZN4love8graphics6opengl5VideoD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8graphics6opengl5VideoD2Ev($this); //@line 71 "modules/graphics/opengl/Video.cpp"
 __ZdlPv($this); //@line 71 "modules/graphics/opengl/Video.cpp"
 return; //@line 73 "modules/graphics/opengl/Video.cpp"
}
function __ZN4love8graphics6opengl5Video4drawEfffffffff($this,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $shader$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 __ZN4love8graphics6opengl5Video6updateEv($this); //@line 119 "modules/graphics/opengl/Video.cpp"
 $1 = HEAP32[158581]|0; //@line 121 "modules/graphics/opengl/Video.cpp"
 $2 = HEAP32[158582]|0; //@line 122 "modules/graphics/opengl/Video.cpp"
 $3 = ($1|0)==($2|0); //@line 122 "modules/graphics/opengl/Video.cpp"
 if ($3) {
  $4 = HEAP32[158583]|0; //@line 126 "modules/graphics/opengl/Video.cpp"
  __ZN4love8graphics6opengl6Shader6attachEb($4,0); //@line 126 "modules/graphics/opengl/Video.cpp"
  $5 = HEAP32[158583]|0; //@line 127 "modules/graphics/opengl/Video.cpp"
  $shader$0 = $5;
 } else {
  $shader$0 = $1;
 }
 $6 = ((($this)) + 20|0); //@line 130 "modules/graphics/opengl/Video.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 130 "modules/graphics/opengl/Video.cpp"
 $8 = ((($this)) + 24|0); //@line 130 "modules/graphics/opengl/Video.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 130 "modules/graphics/opengl/Video.cpp"
 $10 = ((($this)) + 28|0); //@line 130 "modules/graphics/opengl/Video.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 130 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6Shader16setVideoTexturesEjjj($shader$0,$7,$9,$11); //@line 130 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL13pushTransformEv(634044); //@line 125 "modules/graphics/opengl/OpenGL.h"
 __THREW__ = 0;
 $12 = (invoke_ii(4003,(634044|0))|0); //@line 135 "modules/graphics/opengl/OpenGL.h"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 do {
  if ($14) {
   label = 13;
  } else {
   __THREW__ = 0;
   invoke_viddddddddd(4039,($0|0),(+$x),(+$y),(+$angle),(+$sx),(+$sy),(+$ox),(+$oy),(+$kx),(+$ky)); //@line 133 "modules/graphics/opengl/Video.cpp"
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if ($16) {
    label = 13;
   } else {
    __THREW__ = 0;
    invoke_vii(4004,($12|0),($0|0)); //@line 133 "modules/graphics/opengl/Video.cpp"
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if ($18) {
     $38 = ___cxa_find_matching_catch_2()|0;
     $39 = tempRet0;
     __ZN4love7Matrix4D2Ev($0); //@line 133 "modules/graphics/opengl/Video.cpp"
     $$0 = $39;$$01 = $38;
     break;
    }
    __ZN4love7Matrix4D2Ev($0); //@line 133 "modules/graphics/opengl/Video.cpp"
    __THREW__ = 0;
    invoke_vii(4006,(634044|0),3); //@line 135 "modules/graphics/opengl/Video.cpp"
    $19 = __THREW__; __THREW__ = 0;
    $20 = $19&1;
    if ($20) {
     label = 13;
    } else {
     $21 = ((($this)) + 32|0); //@line 137 "modules/graphics/opengl/Video.cpp"
     $22 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
     __THREW__ = 0;
     invoke_viiiiii($22|0,0,2,5126,0,20,($21|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
     $23 = __THREW__; __THREW__ = 0;
     $24 = $23&1;
     if ($24) {
      label = 13;
     } else {
      $25 = ((($this)) + 40|0); //@line 138 "modules/graphics/opengl/Video.cpp"
      $26 = HEAP32[164942]|0; //@line 591 "./libraries/glad/gladfuncs.hpp"
      __THREW__ = 0;
      invoke_viiiiii($26|0,1,2,5126,0,20,($25|0)); //@line 591 "./libraries/glad/gladfuncs.hpp"
      $27 = __THREW__; __THREW__ = 0;
      $28 = $27&1;
      if ($28) {
       label = 13;
      } else {
       __THREW__ = 0;
       invoke_vi(4007,(634044|0)); //@line 140 "modules/graphics/opengl/Video.cpp"
       $29 = __THREW__; __THREW__ = 0;
       $30 = $29&1;
       if ($30) {
        label = 13;
       } else {
        __THREW__ = 0;
        invoke_viiii(4008,(634044|0),5,0,4); //@line 141 "modules/graphics/opengl/Video.cpp"
        $31 = __THREW__; __THREW__ = 0;
        $32 = $31&1;
        if ($32) {
         label = 13;
        } else {
         if ($3) {
          $33 = HEAP32[158582]|0; //@line 145 "modules/graphics/opengl/Video.cpp"
          __THREW__ = 0;
          invoke_vii(4148,($33|0),0); //@line 145 "modules/graphics/opengl/Video.cpp"
          $34 = __THREW__; __THREW__ = 0;
          $35 = $34&1;
          if ($35) {
           label = 13;
           break;
          }
         }
         __THREW__ = 0;
         invoke_vi(4009,(634044|0)); //@line 130 "modules/graphics/opengl/OpenGL.h"
         $40 = __THREW__; __THREW__ = 0;
         $41 = $40&1;
         if ($41) {
          $42 = ___cxa_find_matching_catch_3(0|0)|0;
          $43 = tempRet0;
          ___clang_call_terminate($42); //@line 130 "modules/graphics/opengl/OpenGL.h"
          // unreachable; //@line 130 "modules/graphics/opengl/OpenGL.h"
         } else {
          STACKTOP = sp;return; //@line 146 "modules/graphics/opengl/Video.cpp"
         }
        }
       }
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 13) {
  $36 = ___cxa_find_matching_catch_2()|0;
  $37 = tempRet0;
  $$0 = $37;$$01 = $36;
 }
 __THREW__ = 0;
 invoke_vi(4009,(634044|0)); //@line 130 "modules/graphics/opengl/OpenGL.h"
 $44 = __THREW__; __THREW__ = 0;
 $45 = $44&1;
 if ($45) {
  $46 = ___cxa_find_matching_catch_3(0|0)|0;
  $47 = tempRet0;
  ___clang_call_terminate($46); //@line 130 "modules/graphics/opengl/OpenGL.h"
  // unreachable; //@line 130 "modules/graphics/opengl/OpenGL.h"
 } else {
  ___resumeException($$01|0);
  // unreachable;
 }
}
function __ZN4love8graphics6opengl5Video6updateEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 150 "modules/graphics/opengl/Video.cpp"
 $3 = ((($2)) + 28|0); //@line 150 "modules/graphics/opengl/Video.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 150 "modules/graphics/opengl/Video.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($1)|0); //@line 150 "modules/graphics/opengl/Video.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $7 = HEAP32[$6>>2]|0; //@line 151 "modules/graphics/opengl/Video.cpp"
 $8 = ((($7)) + 16|0); //@line 151 "modules/graphics/opengl/Video.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 151 "modules/graphics/opengl/Video.cpp"
 FUNCTION_TABLE_vi[$9 & 8191]($6); //@line 151 "modules/graphics/opengl/Video.cpp"
 if (!($5)) {
  return; //@line 169 "modules/graphics/opengl/Video.cpp"
 }
 $10 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $11 = HEAP32[$10>>2]|0; //@line 155 "modules/graphics/opengl/Video.cpp"
 $12 = ((($11)) + 20|0); //@line 155 "modules/graphics/opengl/Video.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 155 "modules/graphics/opengl/Video.cpp"
 $14 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0); //@line 155 "modules/graphics/opengl/Video.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 157 "modules/graphics/opengl/Video.cpp"
 $16 = ((($14)) + 12|0); //@line 157 "modules/graphics/opengl/Video.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 157 "modules/graphics/opengl/Video.cpp"
 $18 = ((($14)) + 4|0); //@line 158 "modules/graphics/opengl/Video.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 158 "modules/graphics/opengl/Video.cpp"
 $20 = ((($14)) + 16|0); //@line 158 "modules/graphics/opengl/Video.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 158 "modules/graphics/opengl/Video.cpp"
 $22 = ((($14)) + 8|0); //@line 160 "modules/graphics/opengl/Video.cpp"
 $23 = HEAP32[$22>>2]|0; //@line 160 "modules/graphics/opengl/Video.cpp"
 $24 = ((($14)) + 20|0); //@line 160 "modules/graphics/opengl/Video.cpp"
 $25 = HEAP32[$24>>2]|0; //@line 160 "modules/graphics/opengl/Video.cpp"
 $26 = ((($14)) + 24|0); //@line 160 "modules/graphics/opengl/Video.cpp"
 $27 = HEAP32[$26>>2]|0; //@line 160 "modules/graphics/opengl/Video.cpp"
 $28 = ((($this)) + 20|0); //@line 164 "modules/graphics/opengl/Video.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 164 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$29); //@line 164 "modules/graphics/opengl/Video.cpp"
 $30 = HEAP32[165192]|0; //@line 352 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiiiiiiii[$30 & 8191](3553,0,0,0,$15,$19,6409,5121,$23); //@line 352 "./libraries/glad/gladfuncs.hpp"
 $31 = ((($this)) + 24|0); //@line 164 "modules/graphics/opengl/Video.cpp"
 $32 = HEAP32[$31>>2]|0; //@line 164 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$32); //@line 164 "modules/graphics/opengl/Video.cpp"
 $33 = HEAP32[165192]|0; //@line 352 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiiiiiiii[$33 & 8191](3553,0,0,0,$17,$21,6409,5121,$25); //@line 352 "./libraries/glad/gladfuncs.hpp"
 $34 = ((($this)) + 28|0); //@line 164 "modules/graphics/opengl/Video.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 164 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$35); //@line 164 "modules/graphics/opengl/Video.cpp"
 $36 = HEAP32[165192]|0; //@line 352 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiiiiiiii[$36 & 8191](3553,0,0,0,$17,$21,6409,5121,$27); //@line 352 "./libraries/glad/gladfuncs.hpp"
 return; //@line 169 "modules/graphics/opengl/Video.cpp"
}
function __ZN4love8graphics6opengl5Video12loadVolatileEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $wrap = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $wrap = sp;
 $0 = ((($this)) + 20|0); //@line 77 "modules/graphics/opengl/Video.cpp"
 $1 = HEAP32[164922]|0; //@line 355 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_vii[$1 & 8191](3,$0); //@line 355 "./libraries/glad/gladfuncs.hpp"
 $2 = ((($this)) + 12|0); //@line 117 "./common/Object.h"
 $3 = HEAP32[$2>>2]|0; //@line 117 "./common/Object.h"
 $4 = HEAP32[$3>>2]|0; //@line 80 "modules/graphics/opengl/Video.cpp"
 $5 = ((($4)) + 20|0); //@line 80 "modules/graphics/opengl/Video.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 80 "modules/graphics/opengl/Video.cpp"
 $7 = (FUNCTION_TABLE_ii[$6 & 8191]($3)|0); //@line 80 "modules/graphics/opengl/Video.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 82 "modules/graphics/opengl/Video.cpp"
 $9 = ((($7)) + 12|0); //@line 82 "modules/graphics/opengl/Video.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 82 "modules/graphics/opengl/Video.cpp"
 $11 = ((($7)) + 4|0); //@line 83 "modules/graphics/opengl/Video.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 83 "modules/graphics/opengl/Video.cpp"
 $13 = ((($7)) + 16|0); //@line 83 "modules/graphics/opengl/Video.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 83 "modules/graphics/opengl/Video.cpp"
 $15 = ((($7)) + 8|0); //@line 85 "modules/graphics/opengl/Video.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 85 "modules/graphics/opengl/Video.cpp"
 $17 = ((($7)) + 20|0); //@line 85 "modules/graphics/opengl/Video.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 85 "modules/graphics/opengl/Video.cpp"
 $19 = ((($7)) + 24|0); //@line 85 "modules/graphics/opengl/Video.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 85 "modules/graphics/opengl/Video.cpp"
 HEAP32[$wrap>>2] = 0; //@line 70 "./modules/graphics/Texture.h"
 $21 = ((($wrap)) + 4|0); //@line 71 "./modules/graphics/Texture.h"
 HEAP32[$21>>2] = 0; //@line 71 "./modules/graphics/Texture.h"
 $22 = ((($this)) + 112|0); //@line 93 "modules/graphics/opengl/Video.cpp"
 $23 = HEAP32[$0>>2]|0; //@line 91 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$23); //@line 91 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL16setTextureFilterERNS0_7Texture6FilterE(634044,$22); //@line 93 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL14setTextureWrapERKNS0_7Texture4WrapE(634044,$wrap); //@line 94 "modules/graphics/opengl/Video.cpp"
 $24 = HEAP32[165602]|0; //@line 47 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiiiiiiii[$24 & 8191](3553,0,6409,$8,$12,0,6409,5121,$16); //@line 47 "./libraries/glad/gladfuncs.hpp"
 $25 = ((($this)) + 24|0); //@line 91 "modules/graphics/opengl/Video.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 91 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$26); //@line 91 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL16setTextureFilterERNS0_7Texture6FilterE(634044,$22); //@line 93 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL14setTextureWrapERKNS0_7Texture4WrapE(634044,$wrap); //@line 94 "modules/graphics/opengl/Video.cpp"
 $27 = HEAP32[165602]|0; //@line 47 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiiiiiiii[$27 & 8191](3553,0,6409,$10,$14,0,6409,5121,$18); //@line 47 "./libraries/glad/gladfuncs.hpp"
 $28 = ((($this)) + 28|0); //@line 91 "modules/graphics/opengl/Video.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 91 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$29); //@line 91 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL16setTextureFilterERNS0_7Texture6FilterE(634044,$22); //@line 93 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL14setTextureWrapERKNS0_7Texture4WrapE(634044,$wrap); //@line 94 "modules/graphics/opengl/Video.cpp"
 $30 = HEAP32[165602]|0; //@line 47 "./libraries/glad/gladfuncs.hpp"
 FUNCTION_TABLE_viiiiiiiii[$30 & 8191](3553,0,6409,$10,$14,0,6409,5121,$20); //@line 47 "./libraries/glad/gladfuncs.hpp"
 STACKTOP = sp;return 1; //@line 100 "modules/graphics/opengl/Video.cpp"
}
function __ZN4love8graphics6opengl5Video14unloadVolatileEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 107 "modules/graphics/opengl/Video.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 107 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL13deleteTextureEj(634044,$1); //@line 107 "modules/graphics/opengl/Video.cpp"
 HEAP32[$0>>2] = 0; //@line 108 "modules/graphics/opengl/Video.cpp"
 $2 = ((($this)) + 24|0); //@line 107 "modules/graphics/opengl/Video.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 107 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL13deleteTextureEj(634044,$3); //@line 107 "modules/graphics/opengl/Video.cpp"
 HEAP32[$2>>2] = 0; //@line 108 "modules/graphics/opengl/Video.cpp"
 $4 = ((($this)) + 28|0); //@line 107 "modules/graphics/opengl/Video.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 107 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL13deleteTextureEj(634044,$5); //@line 107 "modules/graphics/opengl/Video.cpp"
 HEAP32[$4>>2] = 0; //@line 108 "modules/graphics/opengl/Video.cpp"
 return; //@line 110 "modules/graphics/opengl/Video.cpp"
}
function __ZThn8_N4love8graphics6opengl5VideoD1Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 70 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl5VideoD2Ev($0); //@line 70 "modules/graphics/opengl/Video.cpp"
 return; //@line 70 "modules/graphics/opengl/Video.cpp"
}
function __ZThn8_N4love8graphics6opengl5VideoD0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 70 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl5VideoD2Ev($0); //@line 71 "modules/graphics/opengl/Video.cpp"
 __ZdlPv($0); //@line 71 "modules/graphics/opengl/Video.cpp"
 return; //@line 70 "modules/graphics/opengl/Video.cpp"
}
function __ZThn8_N4love8graphics6opengl5Video12loadVolatileEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 75 "modules/graphics/opengl/Video.cpp"
 (__ZN4love8graphics6opengl5Video12loadVolatileEv($0)|0); //@line 75 "modules/graphics/opengl/Video.cpp"
 return 1; //@line 75 "modules/graphics/opengl/Video.cpp"
}
function __ZThn8_N4love8graphics6opengl5Video14unloadVolatileEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 103 "modules/graphics/opengl/Video.cpp"
 $1 = ((($0)) + 20|0); //@line 107 "modules/graphics/opengl/Video.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 107 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL13deleteTextureEj(634044,$2); //@line 107 "modules/graphics/opengl/Video.cpp"
 HEAP32[$1>>2] = 0; //@line 108 "modules/graphics/opengl/Video.cpp"
 $3 = ((($0)) + 24|0); //@line 107 "modules/graphics/opengl/Video.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 107 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL13deleteTextureEj(634044,$4); //@line 107 "modules/graphics/opengl/Video.cpp"
 HEAP32[$3>>2] = 0; //@line 108 "modules/graphics/opengl/Video.cpp"
 $5 = ((($0)) + 28|0); //@line 107 "modules/graphics/opengl/Video.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 107 "modules/graphics/opengl/Video.cpp"
 __ZN4love8graphics6opengl6OpenGL13deleteTextureEj(634044,$6); //@line 107 "modules/graphics/opengl/Video.cpp"
 HEAP32[$5>>2] = 0; //@line 108 "modules/graphics/opengl/Video.cpp"
 return; //@line 103 "modules/graphics/opengl/Video.cpp"
}
function __ZN4love8graphics6opengl5VideoC2EPNS_5video11VideoStreamE($this,$stream) {
 $this = $this|0;
 $stream = $stream|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0.0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0.0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0.0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0.0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 36 "./modules/graphics/Drawable.h"
 HEAP32[$this>>2] = (25544); //@line 36 "./modules/graphics/Drawable.h"
 $0 = ((($this)) + 8|0); //@line 36 "modules/graphics/opengl/Video.cpp"
 __THREW__ = 0;
 invoke_vi(4011,($0|0)); //@line 33 "modules/graphics/opengl/Video.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $34 = ___cxa_find_matching_catch_2()|0;
  $35 = tempRet0;
  $$1 = $35;$$12 = $34;
  ___resumeException($$12|0);
  // unreachable;
 }
 HEAP32[$this>>2] = (25484); //@line 36 "modules/graphics/opengl/Video.cpp"
 HEAP32[$0>>2] = (25520); //@line 36 "modules/graphics/opengl/Video.cpp"
 $3 = ((($this)) + 12|0); //@line 93 "./common/Object.h"
 HEAP32[$3>>2] = $stream; //@line 93 "./common/Object.h"
 $4 = ($stream|0)==(0|0); //@line 95 "./common/Object.h"
 if ($4) {
  label = 4;
 } else {
  $5 = HEAP32[$stream>>2]|0; //@line 95 "./common/Object.h"
  $6 = ((($5)) + 8|0); //@line 95 "./common/Object.h"
  $7 = HEAP32[$6>>2]|0; //@line 95 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($7|0,($stream|0)); //@line 95 "./common/Object.h"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if ($9) {
   $36 = ___cxa_find_matching_catch_2()|0;
   $37 = tempRet0;
   $$0 = $37;$$01 = $36;
  } else {
   label = 4;
  }
 }
 if ((label|0) == 4) {
  $10 = ((($this)) + 16|0); //@line 88 "./common/Object.h"
  HEAP32[$10>>2] = 0; //@line 88 "./common/Object.h"
  __THREW__ = 0;
  $11 = (invoke_i(4056)|0); //@line 35 "modules/graphics/opengl/Video.cpp"
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   $14 = ((($this)) + 112|0); //@line 35 "modules/graphics/opengl/Video.cpp"
   ;HEAP32[$14>>2]=HEAP32[$11>>2]|0;HEAP32[$14+4>>2]=HEAP32[$11+4>>2]|0;HEAP32[$14+8>>2]=HEAP32[$11+8>>2]|0;HEAP32[$14+12>>2]=HEAP32[$11+12>>2]|0; //@line 35 "modules/graphics/opengl/Video.cpp"
   $15 = ((($this)) + 120|0); //@line 37 "modules/graphics/opengl/Video.cpp"
   HEAP32[$15>>2] = 0; //@line 37 "modules/graphics/opengl/Video.cpp"
   $16 = HEAP32[$stream>>2]|0; //@line 39 "modules/graphics/opengl/Video.cpp"
   $17 = ((($16)) + 16|0); //@line 39 "modules/graphics/opengl/Video.cpp"
   $18 = HEAP32[$17>>2]|0; //@line 39 "modules/graphics/opengl/Video.cpp"
   __THREW__ = 0;
   invoke_vi($18|0,($stream|0)); //@line 39 "modules/graphics/opengl/Video.cpp"
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   if (!($20)) {
    $21 = ((($this)) + 48|0); //@line 42 "modules/graphics/opengl/Video.cpp"
    $22 = ((($this)) + 68|0); //@line 42 "modules/graphics/opengl/Video.cpp"
    $23 = ((($this)) + 88|0); //@line 42 "modules/graphics/opengl/Video.cpp"
    $24 = ((($this)) + 108|0); //@line 42 "modules/graphics/opengl/Video.cpp"
    $25 = ((($this)) + 32|0); //@line 49 "modules/graphics/opengl/Video.cpp"
    HEAP32[$21>>2] = -1; //@line 42 "modules/graphics/opengl/Video.cpp"
    HEAP32[$22>>2] = -1; //@line 42 "modules/graphics/opengl/Video.cpp"
    HEAP32[$23>>2] = -1; //@line 42 "modules/graphics/opengl/Video.cpp"
    HEAP32[$24>>2] = -1; //@line 42 "modules/graphics/opengl/Video.cpp"
    HEAPF32[$25>>2] = 0.0; //@line 49 "modules/graphics/opengl/Video.cpp"
    $26 = ((($this)) + 36|0); //@line 50 "modules/graphics/opengl/Video.cpp"
    HEAPF32[$26>>2] = 0.0; //@line 50 "modules/graphics/opengl/Video.cpp"
    $27 = ((($this)) + 52|0); //@line 51 "modules/graphics/opengl/Video.cpp"
    HEAPF32[$27>>2] = 0.0; //@line 51 "modules/graphics/opengl/Video.cpp"
    $28 = HEAP32[$stream>>2]|0; //@line 52 "modules/graphics/opengl/Video.cpp"
    $29 = ((($28)) + 36|0); //@line 52 "modules/graphics/opengl/Video.cpp"
    $30 = HEAP32[$29>>2]|0; //@line 52 "modules/graphics/opengl/Video.cpp"
    __THREW__ = 0;
    $31 = (invoke_ii($30|0,($stream|0))|0); //@line 52 "modules/graphics/opengl/Video.cpp"
    $32 = __THREW__; __THREW__ = 0;
    $33 = $32&1;
    if (!($33)) {
     $58 = (+($31|0)); //@line 52 "modules/graphics/opengl/Video.cpp"
     $59 = ((($this)) + 56|0); //@line 52 "modules/graphics/opengl/Video.cpp"
     HEAPF32[$59>>2] = $58; //@line 52 "modules/graphics/opengl/Video.cpp"
     $60 = HEAP32[$stream>>2]|0; //@line 53 "modules/graphics/opengl/Video.cpp"
     $61 = ((($60)) + 32|0); //@line 53 "modules/graphics/opengl/Video.cpp"
     $62 = HEAP32[$61>>2]|0; //@line 53 "modules/graphics/opengl/Video.cpp"
     __THREW__ = 0;
     $63 = (invoke_ii($62|0,($stream|0))|0); //@line 53 "modules/graphics/opengl/Video.cpp"
     $64 = __THREW__; __THREW__ = 0;
     $65 = $64&1;
     if (!($65)) {
      $66 = (+($63|0)); //@line 53 "modules/graphics/opengl/Video.cpp"
      $67 = ((($this)) + 72|0); //@line 53 "modules/graphics/opengl/Video.cpp"
      HEAPF32[$67>>2] = $66; //@line 53 "modules/graphics/opengl/Video.cpp"
      $68 = ((($this)) + 76|0); //@line 54 "modules/graphics/opengl/Video.cpp"
      HEAPF32[$68>>2] = 0.0; //@line 54 "modules/graphics/opengl/Video.cpp"
      $69 = HEAP32[$stream>>2]|0; //@line 55 "modules/graphics/opengl/Video.cpp"
      $70 = ((($69)) + 32|0); //@line 55 "modules/graphics/opengl/Video.cpp"
      $71 = HEAP32[$70>>2]|0; //@line 55 "modules/graphics/opengl/Video.cpp"
      __THREW__ = 0;
      $72 = (invoke_ii($71|0,($stream|0))|0); //@line 55 "modules/graphics/opengl/Video.cpp"
      $73 = __THREW__; __THREW__ = 0;
      $74 = $73&1;
      if (!($74)) {
       $75 = (+($72|0)); //@line 55 "modules/graphics/opengl/Video.cpp"
       $76 = ((($this)) + 92|0); //@line 55 "modules/graphics/opengl/Video.cpp"
       HEAPF32[$76>>2] = $75; //@line 55 "modules/graphics/opengl/Video.cpp"
       $77 = HEAP32[$stream>>2]|0; //@line 56 "modules/graphics/opengl/Video.cpp"
       $78 = ((($77)) + 36|0); //@line 56 "modules/graphics/opengl/Video.cpp"
       $79 = HEAP32[$78>>2]|0; //@line 56 "modules/graphics/opengl/Video.cpp"
       __THREW__ = 0;
       $80 = (invoke_ii($79|0,($stream|0))|0); //@line 56 "modules/graphics/opengl/Video.cpp"
       $81 = __THREW__; __THREW__ = 0;
       $82 = $81&1;
       if (!($82)) {
        $83 = (+($80|0)); //@line 56 "modules/graphics/opengl/Video.cpp"
        $84 = ((($this)) + 96|0); //@line 56 "modules/graphics/opengl/Video.cpp"
        HEAPF32[$84>>2] = $83; //@line 56 "modules/graphics/opengl/Video.cpp"
        $85 = ((($this)) + 40|0); //@line 58 "modules/graphics/opengl/Video.cpp"
        HEAPF32[$85>>2] = 0.0; //@line 58 "modules/graphics/opengl/Video.cpp"
        $86 = ((($this)) + 44|0); //@line 59 "modules/graphics/opengl/Video.cpp"
        HEAPF32[$86>>2] = 0.0; //@line 59 "modules/graphics/opengl/Video.cpp"
        $87 = ((($this)) + 60|0); //@line 60 "modules/graphics/opengl/Video.cpp"
        HEAPF32[$87>>2] = 0.0; //@line 60 "modules/graphics/opengl/Video.cpp"
        $88 = ((($this)) + 64|0); //@line 61 "modules/graphics/opengl/Video.cpp"
        HEAPF32[$88>>2] = 1.0; //@line 61 "modules/graphics/opengl/Video.cpp"
        $89 = ((($this)) + 80|0); //@line 62 "modules/graphics/opengl/Video.cpp"
        HEAPF32[$89>>2] = 1.0; //@line 62 "modules/graphics/opengl/Video.cpp"
        $90 = ((($this)) + 84|0); //@line 63 "modules/graphics/opengl/Video.cpp"
        HEAPF32[$90>>2] = 0.0; //@line 63 "modules/graphics/opengl/Video.cpp"
        $91 = ((($this)) + 100|0); //@line 64 "modules/graphics/opengl/Video.cpp"
        HEAPF32[$91>>2] = 1.0; //@line 64 "modules/graphics/opengl/Video.cpp"
        $92 = ((($this)) + 104|0); //@line 65 "modules/graphics/opengl/Video.cpp"
        HEAPF32[$92>>2] = 1.0; //@line 65 "modules/graphics/opengl/Video.cpp"
        $93 = HEAP32[$this>>2]|0; //@line 67 "modules/graphics/opengl/Video.cpp"
        $94 = ((($93)) + 20|0); //@line 67 "modules/graphics/opengl/Video.cpp"
        $95 = HEAP32[$94>>2]|0; //@line 67 "modules/graphics/opengl/Video.cpp"
        __THREW__ = 0;
        (invoke_ii($95|0,($this|0))|0); //@line 67 "modules/graphics/opengl/Video.cpp"
        $96 = __THREW__; __THREW__ = 0;
        $97 = $96&1;
        if (!($97)) {
         return; //@line 68 "modules/graphics/opengl/Video.cpp"
        }
       }
      }
     }
    }
   }
  }
  $38 = ___cxa_find_matching_catch_2()|0;
  $39 = tempRet0;
  $40 = HEAP32[$10>>2]|0; //@line 106 "./common/Object.h"
  $41 = ($40|0)==(0|0); //@line 106 "./common/Object.h"
  if (!($41)) {
   $42 = HEAP32[$40>>2]|0; //@line 106 "./common/Object.h"
   $43 = ((($42)) + 12|0); //@line 106 "./common/Object.h"
   $44 = HEAP32[$43>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($44|0,($40|0)); //@line 106 "./common/Object.h"
   $45 = __THREW__; __THREW__ = 0;
   $46 = $45&1;
   if ($46) {
    $47 = ___cxa_find_matching_catch_3(0|0)|0;
    $48 = tempRet0;
    ___clang_call_terminate($47); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   }
  }
  $49 = HEAP32[$3>>2]|0; //@line 106 "./common/Object.h"
  $50 = ($49|0)==(0|0); //@line 106 "./common/Object.h"
  if ($50) {
   $$0 = $39;$$01 = $38;
  } else {
   $51 = HEAP32[$49>>2]|0; //@line 106 "./common/Object.h"
   $52 = ((($51)) + 12|0); //@line 106 "./common/Object.h"
   $53 = HEAP32[$52>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($53|0,($49|0)); //@line 106 "./common/Object.h"
   $54 = __THREW__; __THREW__ = 0;
   $55 = $54&1;
   if ($55) {
    $56 = ___cxa_find_matching_catch_3(0|0)|0;
    $57 = tempRet0;
    ___clang_call_terminate($56); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   } else {
    $$0 = $39;$$01 = $38;
   }
  }
 }
 __ZN4love8graphics8VolatileD2Ev($0); //@line 68 "modules/graphics/opengl/Video.cpp"
 $$1 = $$0;$$12 = $$01;
 ___resumeException($$12|0);
 // unreachable;
}
function __ZN4love8graphics8DrawableD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 43 "./modules/graphics/Drawable.h"
 return; //@line 43 "./modules/graphics/Drawable.h"
}
function __ZN4love8graphics6opengl5Video9getStreamEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 127 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 127 "./common/Object.h"
 return ($1|0); //@line 114 "modules/graphics/opengl/Video.cpp"
}
function __ZN4love8graphics6opengl5Video9getSourceEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 127 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 127 "./common/Object.h"
 return ($1|0); //@line 173 "modules/graphics/opengl/Video.cpp"
}
function __ZN4love8graphics6opengl5Video9setSourceEPNS_5audio6SourceE($this,$source) {
 $this = $this|0;
 $source = $source|0;
 var $$index = 0, $$index10 = 0, $$index2 = 0, $$index6 = 0, $$pre$phi4Z2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($source|0)==(0|0); //@line 95 "./common/Object.h"
 if ($0) {
  label = 4;
 } else {
  $1 = HEAP32[$source>>2]|0; //@line 95 "./common/Object.h"
  $2 = ((($1)) + 8|0); //@line 95 "./common/Object.h"
  $3 = HEAP32[$2>>2]|0; //@line 95 "./common/Object.h"
  FUNCTION_TABLE_vi[$3 & 8191]($source); //@line 95 "./common/Object.h"
  $4 = HEAP32[$source>>2]|0; //@line 132 "./common/Object.h"
  $5 = ((($4)) + 8|0); //@line 132 "./common/Object.h"
  $6 = HEAP32[$5>>2]|0; //@line 132 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($6|0,($source|0)); //@line 132 "./common/Object.h"
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if ($8) {
   $9 = ___cxa_find_matching_catch_2()|0;
   $10 = tempRet0;
   $$index = $9;$$index2 = $10;$$pre$phi4Z2D = $source;$$pre$phiZ2D = $source;
  } else {
   label = 4;
  }
 }
 do {
  if ((label|0) == 4) {
   $11 = ((($this)) + 16|0); //@line 133 "./common/Object.h"
   $12 = HEAP32[$11>>2]|0; //@line 133 "./common/Object.h"
   $13 = ($12|0)==(0|0); //@line 133 "./common/Object.h"
   if (!($13)) {
    $14 = HEAP32[$12>>2]|0; //@line 133 "./common/Object.h"
    $15 = ((($14)) + 12|0); //@line 133 "./common/Object.h"
    $16 = HEAP32[$15>>2]|0; //@line 133 "./common/Object.h"
    __THREW__ = 0;
    invoke_vi($16|0,($12|0)); //@line 133 "./common/Object.h"
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if ($18) {
     $26 = ___cxa_find_matching_catch_2()|0;
     $27 = tempRet0;
     if ($0) {
      $$index10 = $27;$$index6 = $26;
      ___resumeException($$index6|0);
      // unreachable;
     } else {
      $$index = $26;$$index2 = $27;$$pre$phi4Z2D = $source;$$pre$phiZ2D = $source;
      break;
     }
    }
   }
   HEAP32[$11>>2] = $source; //@line 134 "./common/Object.h"
   if ($0) {
    return; //@line 179 "modules/graphics/opengl/Video.cpp"
   }
   $19 = HEAP32[$source>>2]|0; //@line 106 "./common/Object.h"
   $20 = ((($19)) + 12|0); //@line 106 "./common/Object.h"
   $21 = HEAP32[$20>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($21|0,($source|0)); //@line 106 "./common/Object.h"
   $22 = __THREW__; __THREW__ = 0;
   $23 = $22&1;
   if ($23) {
    $24 = ___cxa_find_matching_catch_3(0|0)|0;
    $25 = tempRet0;
    ___clang_call_terminate($24); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   } else {
    return; //@line 179 "modules/graphics/opengl/Video.cpp"
   }
  }
 } while(0);
 $28 = HEAP32[$$pre$phi4Z2D>>2]|0; //@line 106 "./common/Object.h"
 $29 = ((($28)) + 12|0); //@line 106 "./common/Object.h"
 $30 = HEAP32[$29>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($30|0,($$pre$phiZ2D|0)); //@line 106 "./common/Object.h"
 $31 = __THREW__; __THREW__ = 0;
 $32 = $31&1;
 if ($32) {
  $33 = ___cxa_find_matching_catch_3(0|0)|0;
  $34 = tempRet0;
  ___clang_call_terminate($33); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  $$index10 = $$index2;$$index6 = $$index;
  ___resumeException($$index6|0);
  // unreachable;
 }
}
function __ZNK4love8graphics6opengl5Video8getWidthEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 183 "modules/graphics/opengl/Video.cpp"
 $3 = ((($2)) + 32|0); //@line 183 "modules/graphics/opengl/Video.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 183 "modules/graphics/opengl/Video.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($1)|0); //@line 183 "modules/graphics/opengl/Video.cpp"
 return ($5|0); //@line 183 "modules/graphics/opengl/Video.cpp"
}
function __ZNK4love8graphics6opengl5Video9getHeightEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 117 "./common/Object.h"
 $1 = HEAP32[$0>>2]|0; //@line 117 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 188 "modules/graphics/opengl/Video.cpp"
 $3 = ((($2)) + 36|0); //@line 188 "modules/graphics/opengl/Video.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 188 "modules/graphics/opengl/Video.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($1)|0); //@line 188 "modules/graphics/opengl/Video.cpp"
 return ($5|0); //@line 188 "modules/graphics/opengl/Video.cpp"
}
function __ZN4love8graphics6opengl5Video9setFilterERKNS0_7Texture6FilterE($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (__ZN4love8graphics7Texture14validateFilterERKNS1_6FilterEb($f,0)|0); //@line 193 "modules/graphics/opengl/Video.cpp"
 if ($0) {
  $6 = ((($this)) + 112|0); //@line 196 "modules/graphics/opengl/Video.cpp"
  ;HEAP32[$6>>2]=HEAP32[$f>>2]|0;HEAP32[$6+4>>2]=HEAP32[$f+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$f+8>>2]|0;HEAP32[$6+12>>2]=HEAP32[$f+12>>2]|0; //@line 196 "modules/graphics/opengl/Video.cpp"
  $7 = ((($this)) + 20|0); //@line 200 "modules/graphics/opengl/Video.cpp"
  $8 = HEAP32[$7>>2]|0; //@line 200 "modules/graphics/opengl/Video.cpp"
  __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$8); //@line 200 "modules/graphics/opengl/Video.cpp"
  __ZN4love8graphics6opengl6OpenGL16setTextureFilterERNS0_7Texture6FilterE(634044,$6); //@line 201 "modules/graphics/opengl/Video.cpp"
  $9 = ((($this)) + 24|0); //@line 200 "modules/graphics/opengl/Video.cpp"
  $10 = HEAP32[$9>>2]|0; //@line 200 "modules/graphics/opengl/Video.cpp"
  __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$10); //@line 200 "modules/graphics/opengl/Video.cpp"
  __ZN4love8graphics6opengl6OpenGL16setTextureFilterERNS0_7Texture6FilterE(634044,$6); //@line 201 "modules/graphics/opengl/Video.cpp"
  $11 = ((($this)) + 28|0); //@line 200 "modules/graphics/opengl/Video.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 200 "modules/graphics/opengl/Video.cpp"
  __ZN4love8graphics6opengl6OpenGL11bindTextureEj(634044,$12); //@line 200 "modules/graphics/opengl/Video.cpp"
  __ZN4love8graphics6opengl6OpenGL16setTextureFilterERNS0_7Texture6FilterE(634044,$6); //@line 201 "modules/graphics/opengl/Video.cpp"
  STACKTOP = sp;return; //@line 203 "modules/graphics/opengl/Video.cpp"
 }
 $1 = (___cxa_allocate_exception(16)|0); //@line 194 "modules/graphics/opengl/Video.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($1|0),(439276|0),($vararg_buffer|0)); //@line 194 "modules/graphics/opengl/Video.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  ___cxa_free_exception(($1|0)); //@line 194 "modules/graphics/opengl/Video.cpp"
  ___resumeException($4|0);
  // unreachable;
 } else {
  ___cxa_throw(($1|0),(96|0),(344|0)); //@line 194 "modules/graphics/opengl/Video.cpp"
  // unreachable; //@line 194 "modules/graphics/opengl/Video.cpp"
 }
}
function __ZNK4love8graphics6opengl5Video9getFilterEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0); //@line 207 "modules/graphics/opengl/Video.cpp"
 return ($0|0); //@line 207 "modules/graphics/opengl/Video.cpp"
}
function __ZN4love8graphics6opengl16luax_checkcanvasEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl6CanvasEEEPT_P9lua_StateiNS_4TypeE($L,$idx,17)|0); //@line 33 "modules/graphics/opengl/wrap_Canvas.cpp"
 return ($0|0); //@line 33 "modules/graphics/opengl/wrap_Canvas.cpp"
}
function __ZN4love14luax_checktypeINS_8graphics6opengl6CanvasEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love8graphics6opengl17w_Canvas_renderToEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $$01$i = 0, $$02$i = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $__begin$sroa$0$014 = 0, $__begin2$sroa$0$013 = 0, $eh$lpad$body$index4Z2D = 0, $eh$lpad$body$indexZ2D = 0, $oldcanvases = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i3 = 0, $scevgep4$i$i$i = 0;
 var $scevgep4$i$i$i4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $oldcanvases = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl6CanvasEEEPT_P9lua_StateiNS_4TypeE($L,1,17)|0); //@line 33 "modules/graphics/opengl/wrap_Canvas.cpp"
 _luaL_checktype($L,2,6); //@line 39 "modules/graphics/opengl/wrap_Canvas.cpp"
 $1 = HEAP32[(625624)>>2]|0; //@line 98 "./common/Module.h"
 $2 = ($1|0)==(0|0); //@line 43 "modules/graphics/opengl/wrap_Canvas.cpp"
 L1: do {
  if (!($2)) {
   __ZNK4love8graphics6opengl8Graphics9getCanvasEv($oldcanvases,$1); //@line 46 "modules/graphics/opengl/wrap_Canvas.cpp"
   $3 = HEAP32[$oldcanvases>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $4 = ((($oldcanvases)) + 4|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $5 = HEAP32[$4>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $6 = ($3|0)==($5|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   L3: do {
    if ($6) {
     label = 3;
    } else {
     $__begin$sroa$0$014 = $3;
     while(1) {
      $33 = HEAP32[$__begin$sroa$0$014>>2]|0; //@line 48 "modules/graphics/opengl/wrap_Canvas.cpp"
      $34 = HEAP32[$33>>2]|0; //@line 49 "modules/graphics/opengl/wrap_Canvas.cpp"
      $35 = ((($34)) + 8|0); //@line 49 "modules/graphics/opengl/wrap_Canvas.cpp"
      $36 = HEAP32[$35>>2]|0; //@line 49 "modules/graphics/opengl/wrap_Canvas.cpp"
      __THREW__ = 0;
      invoke_vi($36|0,($33|0)); //@line 49 "modules/graphics/opengl/wrap_Canvas.cpp"
      $37 = __THREW__; __THREW__ = 0;
      $38 = $37&1;
      if ($38) {
       break;
      }
      $39 = ((($__begin$sroa$0$014)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $40 = ($39|0)==($5|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($40) {
       label = 3;
       break L3;
      } else {
       $__begin$sroa$0$014 = $39;
      }
     }
     $43 = ___cxa_find_matching_catch_2()|0;
     $44 = tempRet0;
     $eh$lpad$body$index4Z2D = $44;$eh$lpad$body$indexZ2D = $43;
    }
   } while(0);
   L8: do {
    if ((label|0) == 3) {
     __THREW__ = 0;
     invoke_vii(4149,($1|0),($0|0)); //@line 51 "modules/graphics/opengl/wrap_Canvas.cpp"
     $7 = __THREW__; __THREW__ = 0;
     $8 = $7&1;
     L10: do {
      if ($8) {
       $9 = ___cxa_find_matching_catch_3(8992|0)|0;
       $10 = tempRet0;
       $11 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
       $12 = ($10|0)==($11|0); //@line 535 "./common/runtime.h"
       do {
        if ($12) {
         $13 = (___cxa_begin_catch(($9|0))|0); //@line 535 "./common/runtime.h"
         $14 = HEAP32[$13>>2]|0; //@line 539 "./common/runtime.h"
         $15 = ((($14)) + 8|0); //@line 539 "./common/runtime.h"
         $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
         $17 = (FUNCTION_TABLE_ii[$16 & 8191]($13)|0); //@line 539 "./common/runtime.h"
         __THREW__ = 0;
         invoke_vii(3380,($L|0),($17|0)); //@line 539 "./common/runtime.h"
         $18 = __THREW__; __THREW__ = 0;
         $19 = $18&1;
         if ($19) {
          $27 = ___cxa_find_matching_catch_2()|0;
          $28 = tempRet0;
          __THREW__ = 0;
          invoke_v(3378); //@line 540 "./common/runtime.h"
          $29 = __THREW__; __THREW__ = 0;
          $30 = $29&1;
          if (!($30)) {
           $$01$i = $28;$$02$i = $27;
           break;
          }
          $31 = ___cxa_find_matching_catch_3(0|0)|0;
          $32 = tempRet0;
          ___clang_call_terminate($31); //@line 540 "./common/runtime.h"
          // unreachable; //@line 540 "./common/runtime.h"
         }
         __THREW__ = 0;
         invoke_v(3378); //@line 540 "./common/runtime.h"
         $20 = __THREW__; __THREW__ = 0;
         $21 = $20&1;
         if ($21) {
          break L10;
         }
         __THREW__ = 0;
         $22 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
         $23 = __THREW__; __THREW__ = 0;
         $24 = $23&1;
         if ($24) {
          break L10;
         }
         __THREW__ = 0;
         HEAP32[$vararg_buffer>>2] = $22; //@line 543 "./common/runtime.h"
         (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 543 "./common/runtime.h"
         $25 = __THREW__; __THREW__ = 0;
         $26 = $25&1;
         if ($26) {
          break L10;
         } else {
          label = 22;
          break L10;
         }
        } else {
         $$01$i = $10;$$02$i = $9;
        }
       } while(0);
       $eh$lpad$body$index4Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
       break L8;
      } else {
       label = 22;
      }
     } while(0);
     do {
      if ((label|0) == 22) {
       __THREW__ = 0;
       invoke_vii(3384,($L|0),2); //@line 53 "modules/graphics/opengl/wrap_Canvas.cpp"
       $56 = __THREW__; __THREW__ = 0;
       $57 = $56&1;
       if (!($57)) {
        __THREW__ = 0;
        $58 = (invoke_iiiii(3552,($L|0),0,0,0)|0); //@line 54 "modules/graphics/opengl/wrap_Canvas.cpp"
        $59 = __THREW__; __THREW__ = 0;
        $60 = $59&1;
        if (!($60)) {
         __THREW__ = 0;
         invoke_vii(3575,($1|0),($oldcanvases|0)); //@line 56 "modules/graphics/opengl/wrap_Canvas.cpp"
         $61 = __THREW__; __THREW__ = 0;
         $62 = $61&1;
         if (!($62)) {
          $63 = HEAP32[$oldcanvases>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $64 = HEAP32[$4>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $65 = ($63|0)==($64|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
          L26: do {
           if (!($65)) {
            $__begin2$sroa$0$013 = $63;
            while(1) {
             $67 = HEAP32[$__begin2$sroa$0$013>>2]|0; //@line 58 "modules/graphics/opengl/wrap_Canvas.cpp"
             $68 = HEAP32[$67>>2]|0; //@line 59 "modules/graphics/opengl/wrap_Canvas.cpp"
             $69 = ((($68)) + 12|0); //@line 59 "modules/graphics/opengl/wrap_Canvas.cpp"
             $70 = HEAP32[$69>>2]|0; //@line 59 "modules/graphics/opengl/wrap_Canvas.cpp"
             __THREW__ = 0;
             invoke_vi($70|0,($67|0)); //@line 59 "modules/graphics/opengl/wrap_Canvas.cpp"
             $71 = __THREW__; __THREW__ = 0;
             $72 = $71&1;
             if ($72) {
              break;
             }
             $73 = ((($__begin2$sroa$0$013)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
             $74 = ($73|0)==($64|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
             if ($74) {
              break L26;
             } else {
              $__begin2$sroa$0$013 = $73;
             }
            }
            $41 = ___cxa_find_matching_catch_2()|0;
            $42 = tempRet0;
            $eh$lpad$body$index4Z2D = $42;$eh$lpad$body$indexZ2D = $41;
            break L8;
           }
          } while(0);
          $66 = ($58|0)==(0); //@line 61 "modules/graphics/opengl/wrap_Canvas.cpp"
          if ($66) {
           $$0 = 1;$$01 = 0;
          } else {
           __THREW__ = 0;
           $75 = (invoke_ii(3553,($L|0))|0); //@line 62 "modules/graphics/opengl/wrap_Canvas.cpp"
           $76 = __THREW__; __THREW__ = 0;
           $77 = $76&1;
           if ($77) {
            break;
           } else {
            $$0 = 0;$$01 = $75;
           }
          }
          $78 = HEAP32[$oldcanvases>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $79 = ($78|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $80 = $78; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          if (!($79)) {
           $81 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $82 = ($81|0)==($78|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           if (!($82)) {
            $scevgep$i$i$i = ((($81)) + -4|0);
            $83 = $scevgep$i$i$i;
            $84 = (($83) - ($80))|0;
            $85 = $84 >>> 2;
            $86 = $85 ^ -1;
            $scevgep4$i$i$i = (($81) + ($86<<2)|0);
            HEAP32[$4>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           }
           __ZdlPv($78); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
          }
          if ($$0) {
           break L1;
          } else {
           $$1 = $$01;
          }
          STACKTOP = sp;return ($$1|0); //@line 66 "modules/graphics/opengl/wrap_Canvas.cpp"
         }
        }
       }
      }
     } while(0);
     $45 = ___cxa_find_matching_catch_2()|0;
     $46 = tempRet0;
     $eh$lpad$body$index4Z2D = $46;$eh$lpad$body$indexZ2D = $45;
    }
   } while(0);
   $47 = HEAP32[$oldcanvases>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $48 = ($47|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $49 = $47; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($48) {
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
   $50 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $51 = ($50|0)==($47|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($51)) {
    $scevgep$i$i$i3 = ((($50)) + -4|0);
    $52 = $scevgep$i$i$i3;
    $53 = (($52) - ($49))|0;
    $54 = $53 >>> 2;
    $55 = $54 ^ -1;
    $scevgep4$i$i$i4 = (($50) + ($55<<2)|0);
    HEAP32[$4>>2] = $scevgep4$i$i$i4; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($47); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($eh$lpad$body$indexZ2D|0);
   // unreachable;
  }
 } while(0);
 $$1 = 0;
 STACKTOP = sp;return ($$1|0); //@line 66 "modules/graphics/opengl/wrap_Canvas.cpp"
}
function __ZN4love8graphics6opengl21w_Canvas_newImageDataEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $img$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 68 "modules/graphics/opengl/wrap_Canvas.cpp"
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl6CanvasEEEPT_P9lua_StateiNS_4TypeE($L,1,17)|0); //@line 33 "modules/graphics/opengl/wrap_Canvas.cpp"
 $1 = (__ZN4love14luax_getmoduleINS_5image5ImageEEEPT_P9lua_StateNS_4TypeE($L,58)|0); //@line 71 "modules/graphics/opengl/wrap_Canvas.cpp"
 $2 = (+_luaL_optnumber($L,2,0.0)); //@line 72 "modules/graphics/opengl/wrap_Canvas.cpp"
 $3 = (~~(($2))); //@line 72 "modules/graphics/opengl/wrap_Canvas.cpp"
 $4 = (+_luaL_optnumber($L,3,0.0)); //@line 73 "modules/graphics/opengl/wrap_Canvas.cpp"
 $5 = (~~(($4))); //@line 73 "modules/graphics/opengl/wrap_Canvas.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 74 "modules/graphics/opengl/wrap_Canvas.cpp"
 $7 = ((($6)) + 24|0); //@line 74 "modules/graphics/opengl/wrap_Canvas.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 74 "modules/graphics/opengl/wrap_Canvas.cpp"
 $9 = (FUNCTION_TABLE_ii[$8 & 8191]($0)|0); //@line 74 "modules/graphics/opengl/wrap_Canvas.cpp"
 $10 = (+($9|0)); //@line 74 "modules/graphics/opengl/wrap_Canvas.cpp"
 $11 = (+_luaL_optnumber($L,4,$10)); //@line 74 "modules/graphics/opengl/wrap_Canvas.cpp"
 $12 = (~~(($11))); //@line 74 "modules/graphics/opengl/wrap_Canvas.cpp"
 $13 = HEAP32[$0>>2]|0; //@line 75 "modules/graphics/opengl/wrap_Canvas.cpp"
 $14 = ((($13)) + 28|0); //@line 75 "modules/graphics/opengl/wrap_Canvas.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 75 "modules/graphics/opengl/wrap_Canvas.cpp"
 $16 = (FUNCTION_TABLE_ii[$15 & 8191]($0)|0); //@line 75 "modules/graphics/opengl/wrap_Canvas.cpp"
 $17 = (+($16|0)); //@line 75 "modules/graphics/opengl/wrap_Canvas.cpp"
 $18 = (+_luaL_optnumber($L,5,$17)); //@line 75 "modules/graphics/opengl/wrap_Canvas.cpp"
 $19 = (~~(($18))); //@line 75 "modules/graphics/opengl/wrap_Canvas.cpp"
 __THREW__ = 0;
 $20 = (invoke_iiiiiii(4150,($0|0),($1|0),($3|0),($5|0),($12|0),($19|0))|0); //@line 78 "modules/graphics/opengl/wrap_Canvas.cpp"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if (!($22)) {
  $img$0 = $20;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,22,$img$0); //@line 80 "modules/graphics/opengl/wrap_Canvas.cpp"
  $41 = HEAP32[$img$0>>2]|0; //@line 81 "modules/graphics/opengl/wrap_Canvas.cpp"
  $42 = ((($41)) + 12|0); //@line 81 "modules/graphics/opengl/wrap_Canvas.cpp"
  $43 = HEAP32[$42>>2]|0; //@line 81 "modules/graphics/opengl/wrap_Canvas.cpp"
  FUNCTION_TABLE_vi[$43 & 8191]($img$0); //@line 81 "modules/graphics/opengl/wrap_Canvas.cpp"
  STACKTOP = sp;return 1; //@line 82 "modules/graphics/opengl/wrap_Canvas.cpp"
 }
 $23 = ___cxa_find_matching_catch_3(8992|0)|0;
 $24 = tempRet0;
 $25 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $26 = ($24|0)==($25|0); //@line 535 "./common/runtime.h"
 if (!($26)) {
  $$01$i = $24;$$02$i = $23;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $27 = (___cxa_begin_catch(($23|0))|0); //@line 535 "./common/runtime.h"
 $28 = HEAP32[$27>>2]|0; //@line 539 "./common/runtime.h"
 $29 = ((($28)) + 8|0); //@line 539 "./common/runtime.h"
 $30 = HEAP32[$29>>2]|0; //@line 539 "./common/runtime.h"
 $31 = (FUNCTION_TABLE_ii[$30 & 8191]($27)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($31|0)); //@line 539 "./common/runtime.h"
 $32 = __THREW__; __THREW__ = 0;
 $33 = $32&1;
 if (!($33)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $34 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $34; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  $img$0 = 0;
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,22,$img$0); //@line 80 "modules/graphics/opengl/wrap_Canvas.cpp"
  $41 = HEAP32[$img$0>>2]|0; //@line 81 "modules/graphics/opengl/wrap_Canvas.cpp"
  $42 = ((($41)) + 12|0); //@line 81 "modules/graphics/opengl/wrap_Canvas.cpp"
  $43 = HEAP32[$42>>2]|0; //@line 81 "modules/graphics/opengl/wrap_Canvas.cpp"
  FUNCTION_TABLE_vi[$43 & 8191]($img$0); //@line 81 "modules/graphics/opengl/wrap_Canvas.cpp"
  STACKTOP = sp;return 1; //@line 82 "modules/graphics/opengl/wrap_Canvas.cpp"
 }
 $35 = ___cxa_find_matching_catch_2()|0;
 $36 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $37 = __THREW__; __THREW__ = 0;
 $38 = $37&1;
 if ($38) {
  $39 = ___cxa_find_matching_catch_3(0|0)|0;
  $40 = tempRet0;
  ___clang_call_terminate($39); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $36;$$02$i = $35;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl18w_Canvas_getFormatEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $str = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $str = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl6CanvasEEEPT_P9lua_StateiNS_4TypeE($L,1,17)|0); //@line 33 "modules/graphics/opengl/wrap_Canvas.cpp"
 $1 = ((($0)) + 144|0); //@line 112 "modules/graphics/opengl/Canvas.h"
 $2 = HEAP32[$1>>2]|0; //@line 112 "modules/graphics/opengl/Canvas.h"
 $3 = (__ZN4love8graphics6opengl6Canvas11getConstantENS2_6FormatERPKc($2,$str)|0); //@line 90 "modules/graphics/opengl/wrap_Canvas.cpp"
 if ($3) {
  $5 = HEAP32[$str>>2]|0; //@line 93 "modules/graphics/opengl/wrap_Canvas.cpp"
  _lua_pushstring($L,$5); //@line 93 "modules/graphics/opengl/wrap_Canvas.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 95 "modules/graphics/opengl/wrap_Canvas.cpp"
 } else {
  $4 = (_luaL_error($L,439300,$vararg_buffer)|0); //@line 91 "modules/graphics/opengl/wrap_Canvas.cpp"
  $$0 = $4;
  STACKTOP = sp;return ($$0|0); //@line 95 "modules/graphics/opengl/wrap_Canvas.cpp"
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl16w_Canvas_getMSAAEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl6CanvasEEEPT_P9lua_StateiNS_4TypeE($L,1,17)|0); //@line 33 "modules/graphics/opengl/wrap_Canvas.cpp"
 $1 = ((($0)) + 168|0); //@line 117 "modules/graphics/opengl/Canvas.h"
 $2 = HEAP32[$1>>2]|0; //@line 117 "modules/graphics/opengl/Canvas.h"
 _lua_pushinteger($L,$2); //@line 100 "modules/graphics/opengl/wrap_Canvas.cpp"
 return 1; //@line 101 "modules/graphics/opengl/wrap_Canvas.cpp"
}
function _luaopen_canvas($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 113 "modules/graphics/opengl/wrap_Canvas.cpp"
 HEAP32[$vararg_buffer>>2] = 15812; //@line 115 "modules/graphics/opengl/wrap_Canvas.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 115 "modules/graphics/opengl/wrap_Canvas.cpp"
 HEAP32[$vararg_ptr1>>2] = 25564; //@line 115 "modules/graphics/opengl/wrap_Canvas.cpp"
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 115 "modules/graphics/opengl/wrap_Canvas.cpp"
 HEAP32[$vararg_ptr2>>2] = 0; //@line 115 "modules/graphics/opengl/wrap_Canvas.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,17,439323,$vararg_buffer)|0); //@line 115 "modules/graphics/opengl/wrap_Canvas.cpp"
 STACKTOP = sp;return ($0|0); //@line 115 "modules/graphics/opengl/wrap_Canvas.cpp"
}
function __ZN4love8graphics6opengl14luax_checkfontEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE($L,$idx,14)|0); //@line 38 "modules/graphics/opengl/wrap_Font.cpp"
 return ($0|0); //@line 38 "modules/graphics/opengl/wrap_Font.cpp"
}
function __ZN4love8graphics6opengl16w_Font_getHeightEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE($L,1,14)|0); //@line 38 "modules/graphics/opengl/wrap_Font.cpp"
 $1 = (+__ZNK4love8graphics6opengl4Font9getHeightEv($0)); //@line 44 "modules/graphics/opengl/wrap_Font.cpp"
 $2 = $1; //@line 44 "modules/graphics/opengl/wrap_Font.cpp"
 _lua_pushnumber($L,$2); //@line 44 "modules/graphics/opengl/wrap_Font.cpp"
 return 1; //@line 45 "modules/graphics/opengl/wrap_Font.cpp"
}
function __ZN4love8graphics6opengl15w_Font_getWidthEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 $1 = (__ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE($L,1,14)|0); //@line 38 "modules/graphics/opengl/wrap_Font.cpp"
 $2 = (_luaL_checklstring($L,2,0)|0); //@line 51 "modules/graphics/opengl/wrap_Font.cpp"
 $3 = (_strlen($2)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viii(3460,($0|0),($2|0),($3|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $13 = ___cxa_find_matching_catch_3(8992|0)|0;
  $14 = tempRet0;
  $eh$lpad$body$i$index2Z2D = $14;$eh$lpad$body$i$indexZ2D = $13;
 } else {
  __THREW__ = 0;
  $6 = (invoke_iii(4151,($1|0),($0|0))|0); //@line 53 "modules/graphics/opengl/wrap_Font.cpp"
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if (!($8)) {
   __THREW__ = 0;
   invoke_vii(3468,($L|0),($6|0)); //@line 53 "modules/graphics/opengl/wrap_Font.cpp"
   $9 = __THREW__; __THREW__ = 0;
   $10 = $9&1;
   if (!($10)) {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 53 "modules/graphics/opengl/wrap_Font.cpp"
    STACKTOP = sp;return 1; //@line 54 "modules/graphics/opengl/wrap_Font.cpp"
   }
  }
  $11 = ___cxa_find_matching_catch_3(8992|0)|0;
  $12 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 53 "modules/graphics/opengl/wrap_Font.cpp"
  $eh$lpad$body$i$index2Z2D = $12;$eh$lpad$body$i$indexZ2D = $11;
 }
 $15 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $16 = ($eh$lpad$body$i$index2Z2D|0)==($15|0); //@line 535 "./common/runtime.h"
 if (!($16)) {
  $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $17 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
 $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
 $19 = ((($18)) + 8|0); //@line 539 "./common/runtime.h"
 $20 = HEAP32[$19>>2]|0; //@line 539 "./common/runtime.h"
 $21 = (FUNCTION_TABLE_ii[$20 & 8191]($17)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($21|0)); //@line 539 "./common/runtime.h"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 if (!($23)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $24 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer>>2] = $24; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
  STACKTOP = sp;return 1; //@line 54 "modules/graphics/opengl/wrap_Font.cpp"
 }
 $25 = ___cxa_find_matching_catch_2()|0;
 $26 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $27 = __THREW__; __THREW__ = 0;
 $28 = $27&1;
 if ($28) {
  $29 = ___cxa_find_matching_catch_3(0|0)|0;
  $30 = tempRet0;
  ___clang_call_terminate($29); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $26;$$02$i = $25;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl14w_Font_getWrapEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01 = 0, $$01$i = 0, $$02$i = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre$i12 = 0, $$pre$i20 = 0, $$pre$i4 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0.0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0.0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$024 = 0, $eh$lpad$body$index3Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i$0 = 0, $lines = 0, $max_width$0$lcssa = 0, $max_width$025 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i13 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i14 = 0, $text = 0, $vararg_buffer = 0, $widths = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $text = sp + 28|0;
 $lines = sp + 16|0;
 $widths = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE($L,1,14)|0); //@line 38 "modules/graphics/opengl/wrap_Font.cpp"
 HEAP32[$text>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($text)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($text)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 invoke_viii(3568,($L|0),2,($text|0)); //@line 62 "modules/graphics/opengl/wrap_Font.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  label = 14;
 } else {
  __THREW__ = 0;
  $5 = (+invoke_dii(3361,($L|0),3)); //@line 64 "modules/graphics/opengl/wrap_Font.cpp"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if ($7) {
   label = 14;
  } else {
   $8 = $5; //@line 64 "modules/graphics/opengl/wrap_Font.cpp"
   HEAP32[$lines>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $9 = ((($lines)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$9>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $10 = ((($lines)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$10>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$widths>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $11 = ((($widths)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$11>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $12 = ((($widths)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$12>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __THREW__ = 0;
   invoke_viidii(4152,($0|0),($text|0),(+$8),($lines|0),($widths|0)); //@line 69 "modules/graphics/opengl/wrap_Font.cpp"
   $13 = __THREW__; __THREW__ = 0;
   $14 = $13&1;
   L4: do {
    if ($14) {
     $15 = ___cxa_find_matching_catch_3(8992|0)|0;
     $16 = tempRet0;
     $17 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
     $18 = ($16|0)==($17|0); //@line 535 "./common/runtime.h"
     do {
      if ($18) {
       $19 = (___cxa_begin_catch(($15|0))|0); //@line 535 "./common/runtime.h"
       $20 = HEAP32[$19>>2]|0; //@line 539 "./common/runtime.h"
       $21 = ((($20)) + 8|0); //@line 539 "./common/runtime.h"
       $22 = HEAP32[$21>>2]|0; //@line 539 "./common/runtime.h"
       $23 = (FUNCTION_TABLE_ii[$22 & 8191]($19)|0); //@line 539 "./common/runtime.h"
       __THREW__ = 0;
       invoke_vii(3380,($L|0),($23|0)); //@line 539 "./common/runtime.h"
       $24 = __THREW__; __THREW__ = 0;
       $25 = $24&1;
       if ($25) {
        $33 = ___cxa_find_matching_catch_2()|0;
        $34 = tempRet0;
        __THREW__ = 0;
        invoke_v(3378); //@line 540 "./common/runtime.h"
        $35 = __THREW__; __THREW__ = 0;
        $36 = $35&1;
        if (!($36)) {
         $$01$i = $34;$$02$i = $33;
         break;
        }
        $37 = ___cxa_find_matching_catch_3(0|0)|0;
        $38 = tempRet0;
        ___clang_call_terminate($37); //@line 540 "./common/runtime.h"
        // unreachable; //@line 540 "./common/runtime.h"
       }
       __THREW__ = 0;
       invoke_v(3378); //@line 540 "./common/runtime.h"
       $26 = __THREW__; __THREW__ = 0;
       $27 = $26&1;
       if ($27) {
        label = 16;
        break L4;
       }
       __THREW__ = 0;
       $28 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
       $29 = __THREW__; __THREW__ = 0;
       $30 = $29&1;
       if ($30) {
        label = 16;
        break L4;
       }
       __THREW__ = 0;
       HEAP32[$vararg_buffer>>2] = $28; //@line 543 "./common/runtime.h"
       (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 543 "./common/runtime.h"
       $31 = __THREW__; __THREW__ = 0;
       $32 = $31&1;
       if ($32) {
        label = 16;
        break L4;
       } else {
        label = 12;
        break L4;
       }
      } else {
       $$01$i = $16;$$02$i = $15;
      }
     } while(0);
     $eh$lpad$body$index3Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
    } else {
     label = 12;
    }
   } while(0);
   do {
    if ((label|0) == 12) {
     $39 = HEAP32[$widths>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $40 = HEAP32[$11>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $41 = ($39|0)==($40|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($41) {
      $max_width$0$lcssa = 0;
     } else {
      $__begin$sroa$0$024 = $39;$max_width$025 = 0;
      while(1) {
       $68 = HEAP32[$__begin$sroa$0$024>>2]|0; //@line 71 "modules/graphics/opengl/wrap_Font.cpp"
       $69 = ($max_width$025|0)<($68|0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $70 = $69 ? $68 : $max_width$025; //@line 72 "modules/graphics/opengl/wrap_Font.cpp"
       $71 = ((($__begin$sroa$0$024)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       $72 = ($71|0)==($40|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       if ($72) {
        $max_width$0$lcssa = $70;
        break;
       } else {
        $__begin$sroa$0$024 = $71;$max_width$025 = $70;
       }
      }
     }
     __THREW__ = 0;
     invoke_vii(3468,($L|0),($max_width$0$lcssa|0)); //@line 74 "modules/graphics/opengl/wrap_Font.cpp"
     $42 = __THREW__; __THREW__ = 0;
     $43 = $42&1;
     if ($43) {
      label = 16;
     } else {
      $73 = HEAP32[$9>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $74 = HEAP32[$lines>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $75 = (($73) - ($74))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $76 = (($75|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      __THREW__ = 0;
      invoke_viii(3393,($L|0),($76|0),0); //@line 75 "modules/graphics/opengl/wrap_Font.cpp"
      $77 = __THREW__; __THREW__ = 0;
      $78 = $77&1;
      if ($78) {
       label = 16;
      } else {
       $i$0 = 0;
       while(1) {
        $79 = HEAP32[$9>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $80 = HEAP32[$lines>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $81 = (($79) - ($80))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $82 = (($81|0) / 12)&-1; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $83 = ($i$0|0)<($82|0); //@line 77 "modules/graphics/opengl/wrap_Font.cpp"
        $84 = $80; //@line 77 "modules/graphics/opengl/wrap_Font.cpp"
        if (!($83)) {
         $$lcssa = $84;
         break;
        }
        $112 = (($84) + (($i$0*12)|0)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $113 = HEAP8[$112>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $114 = $113 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $115 = ($114<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if ($115) {
         $118 = ((($112)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $119 = $118;
        } else {
         $116 = (((($84) + (($i$0*12)|0)|0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $117 = HEAP32[$116>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $119 = $117;
        }
        __THREW__ = 0;
        invoke_vii(3380,($L|0),($119|0)); //@line 79 "modules/graphics/opengl/wrap_Font.cpp"
        $120 = __THREW__; __THREW__ = 0;
        $121 = $120&1;
        if ($121) {
         label = 15;
         break;
        }
        $122 = (($i$0) + 1)|0; //@line 80 "modules/graphics/opengl/wrap_Font.cpp"
        __THREW__ = 0;
        invoke_viii(3395,($L|0),-2,($122|0)); //@line 80 "modules/graphics/opengl/wrap_Font.cpp"
        $123 = __THREW__; __THREW__ = 0;
        $124 = $123&1;
        if ($124) {
         label = 15;
         break;
        } else {
         $i$0 = $122;
        }
       }
       if ((label|0) == 15) {
        $46 = ___cxa_find_matching_catch_2()|0;
        $47 = tempRet0;
        $eh$lpad$body$index3Z2D = $47;$eh$lpad$body$indexZ2D = $46;
        break;
       }
       $85 = HEAP32[$widths>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $86 = ($85|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $87 = $85; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($86) {
        $95 = $$lcssa;
       } else {
        $88 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $89 = ($88|0)==($85|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($89)) {
         $scevgep$i$i$i13 = ((($88)) + -4|0);
         $90 = $scevgep$i$i$i13;
         $91 = (($90) - ($87))|0;
         $92 = $91 >>> 2;
         $93 = $92 ^ -1;
         $scevgep4$i$i$i14 = (($88) + ($93<<2)|0);
         HEAP32[$11>>2] = $scevgep4$i$i$i14; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
        __ZdlPv($85); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        $$pre = HEAP32[$lines>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $95 = $$pre;
       }
       $94 = ($95|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($94)) {
        $96 = HEAP32[$9>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $97 = ($96|0)==($95|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($97) {
         $102 = $95;
        } else {
         $99 = $96;
         while(1) {
          $98 = ((($99)) + -12|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$9>>2] = $98; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($98); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $100 = HEAP32[$9>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $101 = ($100|0)==($95|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          if ($101) {
           break;
          } else {
           $99 = $100;
          }
         }
         $$pre$i20 = HEAP32[$lines>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $102 = $$pre$i20;
        }
        __ZdlPv($102); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       }
       $103 = HEAP32[$text>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $104 = ($103|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($104) {
        STACKTOP = sp;return 2; //@line 84 "modules/graphics/opengl/wrap_Font.cpp"
       }
       $105 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $106 = ($105|0)==($103|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($106) {
        $111 = $103;
       } else {
        $108 = $105;
        while(1) {
         $107 = ((($108)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$1>>2] = $107; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($107); //@line 64 "modules/graphics/opengl/Font.h"
         $109 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $110 = ($109|0)==($103|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($110) {
          break;
         } else {
          $108 = $109;
         }
        }
        $$pre$i4 = HEAP32[$text>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $111 = $$pre$i4;
       }
       __ZdlPv($111); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       STACKTOP = sp;return 2; //@line 84 "modules/graphics/opengl/wrap_Font.cpp"
      }
     }
    }
   } while(0);
   if ((label|0) == 16) {
    $48 = ___cxa_find_matching_catch_2()|0;
    $49 = tempRet0;
    $eh$lpad$body$index3Z2D = $49;$eh$lpad$body$indexZ2D = $48;
   }
   $50 = HEAP32[$widths>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $51 = ($50|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $52 = $50; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($51)) {
    $53 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $54 = ($53|0)==($50|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($54)) {
     $scevgep$i$i$i = ((($53)) + -4|0);
     $55 = $scevgep$i$i$i;
     $56 = (($55) - ($52))|0;
     $57 = $56 >>> 2;
     $58 = $57 ^ -1;
     $scevgep4$i$i$i = (($53) + ($58<<2)|0);
     HEAP32[$11>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($50); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   }
   $59 = HEAP32[$lines>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $60 = ($59|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($60) {
    $$0 = $eh$lpad$body$indexZ2D;$$01 = $eh$lpad$body$index3Z2D;
   } else {
    $61 = HEAP32[$9>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $62 = ($61|0)==($59|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($62) {
     $67 = $59;
    } else {
     $64 = $61;
     while(1) {
      $63 = ((($64)) + -12|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$9>>2] = $63; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($63); //@line 1700 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $65 = HEAP32[$9>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $66 = ($65|0)==($59|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($66) {
       break;
      } else {
       $64 = $65;
      }
     }
     $$pre$i12 = HEAP32[$lines>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $67 = $$pre$i12;
    }
    __ZdlPv($67); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $$0 = $eh$lpad$body$indexZ2D;$$01 = $eh$lpad$body$index3Z2D;
   }
  }
 }
 if ((label|0) == 14) {
  $44 = ___cxa_find_matching_catch_2()|0;
  $45 = tempRet0;
  $$0 = $44;$$01 = $45;
 }
 $125 = HEAP32[$text>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $126 = ($125|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($126) {
  ___resumeException($$0|0);
  // unreachable;
 }
 $127 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $128 = ($127|0)==($125|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($128) {
  $133 = $125;
 } else {
  $130 = $127;
  while(1) {
   $129 = ((($130)) + -16|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$1>>2] = $129; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($129); //@line 64 "modules/graphics/opengl/Font.h"
   $131 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $132 = ($131|0)==($125|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($132) {
    break;
   } else {
    $130 = $131;
   }
  }
  $$pre$i = HEAP32[$text>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $133 = $$pre$i;
 }
 __ZdlPv($133); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($$0|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8graphics6opengl20w_Font_setLineHeightEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE($L,1,14)|0); //@line 38 "modules/graphics/opengl/wrap_Font.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 89 "modules/graphics/opengl/wrap_Font.cpp"
 $2 = $1; //@line 89 "modules/graphics/opengl/wrap_Font.cpp"
 __ZN4love8graphics6opengl4Font13setLineHeightEf($0,$2); //@line 90 "modules/graphics/opengl/wrap_Font.cpp"
 return 0; //@line 91 "modules/graphics/opengl/wrap_Font.cpp"
}
function __ZN4love8graphics6opengl20w_Font_getLineHeightEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE($L,1,14)|0); //@line 38 "modules/graphics/opengl/wrap_Font.cpp"
 $1 = (+__ZNK4love8graphics6opengl4Font13getLineHeightEv($0)); //@line 97 "modules/graphics/opengl/wrap_Font.cpp"
 $2 = $1; //@line 97 "modules/graphics/opengl/wrap_Font.cpp"
 _lua_pushnumber($L,$2); //@line 97 "modules/graphics/opengl/wrap_Font.cpp"
 return 1; //@line 98 "modules/graphics/opengl/wrap_Font.cpp"
}
function __ZN4love8graphics6opengl16w_Font_setFilterEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, $f = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $f = sp + 24|0;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE($L,1,14)|0); //@line 38 "modules/graphics/opengl/wrap_Font.cpp"
 $1 = (__ZN4love8graphics6opengl4Font9getFilterEv($0)|0); //@line 104 "modules/graphics/opengl/wrap_Font.cpp"
 ;HEAP32[$f>>2]=HEAP32[$1>>2]|0;HEAP32[$f+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$f+8>>2]=HEAP32[$1+8>>2]|0;HEAP32[$f+12>>2]=HEAP32[$1+12>>2]|0; //@line 104 "modules/graphics/opengl/wrap_Font.cpp"
 $2 = (_luaL_checklstring($L,2,0)|0); //@line 106 "modules/graphics/opengl/wrap_Font.cpp"
 $3 = (_luaL_optlstring($L,3,$2,0)|0); //@line 107 "modules/graphics/opengl/wrap_Font.cpp"
 $4 = (__ZN4love8graphics7Texture11getConstantEPKcRNS1_10FilterModeE($2,$f)|0); //@line 109 "modules/graphics/opengl/wrap_Font.cpp"
 if (!($4)) {
  HEAP32[$vararg_buffer>>2] = $2; //@line 110 "modules/graphics/opengl/wrap_Font.cpp"
  $5 = (_luaL_error($L,439370,$vararg_buffer)|0); //@line 110 "modules/graphics/opengl/wrap_Font.cpp"
  $$0 = $5;
  STACKTOP = sp;return ($$0|0); //@line 118 "modules/graphics/opengl/wrap_Font.cpp"
 }
 $6 = ((($f)) + 4|0); //@line 111 "modules/graphics/opengl/wrap_Font.cpp"
 $7 = (__ZN4love8graphics7Texture11getConstantEPKcRNS1_10FilterModeE($3,$6)|0); //@line 111 "modules/graphics/opengl/wrap_Font.cpp"
 if (!($7)) {
  HEAP32[$vararg_buffer1>>2] = $3; //@line 112 "modules/graphics/opengl/wrap_Font.cpp"
  $8 = (_luaL_error($L,439370,$vararg_buffer1)|0); //@line 112 "modules/graphics/opengl/wrap_Font.cpp"
  $$0 = $8;
  STACKTOP = sp;return ($$0|0); //@line 118 "modules/graphics/opengl/wrap_Font.cpp"
 }
 $9 = (+_luaL_optnumber($L,4,1.0)); //@line 114 "modules/graphics/opengl/wrap_Font.cpp"
 $10 = $9; //@line 114 "modules/graphics/opengl/wrap_Font.cpp"
 $11 = ((($f)) + 12|0); //@line 114 "modules/graphics/opengl/wrap_Font.cpp"
 HEAPF32[$11>>2] = $10; //@line 114 "modules/graphics/opengl/wrap_Font.cpp"
 __THREW__ = 0;
 invoke_vii(4153,($0|0),($f|0)); //@line 116 "modules/graphics/opengl/wrap_Font.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if (!($13)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 118 "modules/graphics/opengl/wrap_Font.cpp"
 }
 $14 = ___cxa_find_matching_catch_3(8992|0)|0;
 $15 = tempRet0;
 $16 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
 $17 = ($15|0)==($16|0); //@line 535 "./common/runtime.h"
 if (!($17)) {
  $$01$i = $15;$$02$i = $14;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 $18 = (___cxa_begin_catch(($14|0))|0); //@line 535 "./common/runtime.h"
 $19 = HEAP32[$18>>2]|0; //@line 539 "./common/runtime.h"
 $20 = ((($19)) + 8|0); //@line 539 "./common/runtime.h"
 $21 = HEAP32[$20>>2]|0; //@line 539 "./common/runtime.h"
 $22 = (FUNCTION_TABLE_ii[$21 & 8191]($18)|0); //@line 539 "./common/runtime.h"
 __THREW__ = 0;
 invoke_vii(3380,($L|0),($22|0)); //@line 539 "./common/runtime.h"
 $23 = __THREW__; __THREW__ = 0;
 $24 = $23&1;
 if (!($24)) {
  ___cxa_end_catch(); //@line 540 "./common/runtime.h"
  $25 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
  HEAP32[$vararg_buffer4>>2] = $25; //@line 543 "./common/runtime.h"
  (_luaL_error($L,595867,$vararg_buffer4)|0); //@line 543 "./common/runtime.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 118 "modules/graphics/opengl/wrap_Font.cpp"
 }
 $26 = ___cxa_find_matching_catch_2()|0;
 $27 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 540 "./common/runtime.h"
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if ($29) {
  $30 = ___cxa_find_matching_catch_3(0|0)|0;
  $31 = tempRet0;
  ___clang_call_terminate($30); //@line 540 "./common/runtime.h"
  // unreachable; //@line 540 "./common/runtime.h"
 } else {
  $$01$i = $27;$$02$i = $26;
  ___resumeException($$02$i|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love8graphics6opengl16w_Font_getFilterEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0.0, $magstr = 0, $minstr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $minstr = sp + 4|0;
 $magstr = sp;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE($L,1,14)|0); //@line 38 "modules/graphics/opengl/wrap_Font.cpp"
 $1 = (__ZN4love8graphics6opengl4Font9getFilterEv($0)|0); //@line 123 "modules/graphics/opengl/wrap_Font.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 123 "modules/graphics/opengl/wrap_Font.cpp"
 $3 = ((($1)) + 4|0); //@line 123 "modules/graphics/opengl/wrap_Font.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 123 "modules/graphics/opengl/wrap_Font.cpp"
 $5 = ((($1)) + 12|0); //@line 123 "modules/graphics/opengl/wrap_Font.cpp"
 $6 = +HEAPF32[$5>>2]; //@line 123 "modules/graphics/opengl/wrap_Font.cpp"
 (__ZN4love8graphics7Texture11getConstantENS1_10FilterModeERPKc($2,$minstr)|0); //@line 126 "modules/graphics/opengl/wrap_Font.cpp"
 (__ZN4love8graphics7Texture11getConstantENS1_10FilterModeERPKc($4,$magstr)|0); //@line 127 "modules/graphics/opengl/wrap_Font.cpp"
 $7 = HEAP32[$minstr>>2]|0; //@line 128 "modules/graphics/opengl/wrap_Font.cpp"
 _lua_pushstring($L,$7); //@line 128 "modules/graphics/opengl/wrap_Font.cpp"
 $8 = HEAP32[$magstr>>2]|0; //@line 129 "modules/graphics/opengl/wrap_Font.cpp"
 _lua_pushstring($L,$8); //@line 129 "modules/graphics/opengl/wrap_Font.cpp"
 $9 = $6; //@line 130 "modules/graphics/opengl/wrap_Font.cpp"
 _lua_pushnumber($L,$9); //@line 130 "modules/graphics/opengl/wrap_Font.cpp"
 STACKTOP = sp;return 3; //@line 131 "modules/graphics/opengl/wrap_Font.cpp"
}
function __ZN4love8graphics6opengl16w_Font_getAscentEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE($L,1,14)|0); //@line 38 "modules/graphics/opengl/wrap_Font.cpp"
 $1 = (__ZNK4love8graphics6opengl4Font9getAscentEv($0)|0); //@line 137 "modules/graphics/opengl/wrap_Font.cpp"
 $2 = (+($1|0)); //@line 137 "modules/graphics/opengl/wrap_Font.cpp"
 _lua_pushnumber($L,$2); //@line 137 "modules/graphics/opengl/wrap_Font.cpp"
 return 1; //@line 138 "modules/graphics/opengl/wrap_Font.cpp"
}
function __ZN4love8graphics6opengl17w_Font_getDescentEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE($L,1,14)|0); //@line 38 "modules/graphics/opengl/wrap_Font.cpp"
 $1 = (__ZNK4love8graphics6opengl4Font10getDescentEv($0)|0); //@line 144 "modules/graphics/opengl/wrap_Font.cpp"
 $2 = (+($1|0)); //@line 144 "modules/graphics/opengl/wrap_Font.cpp"
 _lua_pushnumber($L,$2); //@line 144 "modules/graphics/opengl/wrap_Font.cpp"
 return 1; //@line 145 "modules/graphics/opengl/wrap_Font.cpp"
}
function __ZN4love8graphics6opengl18w_Font_getBaselineEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE($L,1,14)|0); //@line 38 "modules/graphics/opengl/wrap_Font.cpp"
 $1 = (+__ZNK4love8graphics6opengl4Font11getBaselineEv($0)); //@line 151 "modules/graphics/opengl/wrap_Font.cpp"
 $2 = $1; //@line 151 "modules/graphics/opengl/wrap_Font.cpp"
 _lua_pushnumber($L,$2); //@line 151 "modules/graphics/opengl/wrap_Font.cpp"
 return 1; //@line 152 "modules/graphics/opengl/wrap_Font.cpp"
}
function __ZN4love8graphics6opengl16w_Font_hasGlyphsEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $hasglyph$0 = 0, $hasglyph$0$lcssa = 0, $hasglyph$0$lcssa19 = 0, $hasglyph$020 = 0, $hasglyph$18 = 0, $hasglyph$18$in = 0, $hasglyph$2 = 0, $hasglyph$3 = 0, $i$01$i$i = 0, $vararg_buffer = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 4|0;
 $1 = (__ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE($L,1,14)|0); //@line 38 "modules/graphics/opengl/wrap_Font.cpp"
 $2 = (_lua_gettop($L)|0); //@line 160 "modules/graphics/opengl/wrap_Font.cpp"
 $3 = (($2) + -1)|0; //@line 160 "modules/graphics/opengl/wrap_Font.cpp"
 $4 = ($3|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $5 = $4 ? 1 : $3; //@line 160 "modules/graphics/opengl/wrap_Font.cpp"
 $6 = ($5|0)>(0); //@line 163 "modules/graphics/opengl/wrap_Font.cpp"
 L1: do {
  if ($6) {
   $hasglyph$0 = 0;$i$01$i$i = 2;
   while(1) {
    __THREW__ = 0;
    $9 = (invoke_iii(3358,($L|0),($i$01$i$i|0))|0); //@line 165 "modules/graphics/opengl/wrap_Font.cpp"
    $10 = __THREW__; __THREW__ = 0;
    $11 = $10&1;
    if ($11) {
     $hasglyph$0$lcssa = $hasglyph$0;
     label = 13;
     break;
    }
    $12 = ($9|0)==(4); //@line 165 "modules/graphics/opengl/wrap_Font.cpp"
    if ($12) {
     __THREW__ = 0;
     invoke_viii(3403,($0|0),($L|0),($i$01$i$i|0)); //@line 166 "modules/graphics/opengl/wrap_Font.cpp"
     $13 = __THREW__; __THREW__ = 0;
     $14 = $13&1;
     if ($14) {
      $hasglyph$0$lcssa = $hasglyph$0;
      label = 13;
      break;
     }
     __THREW__ = 0;
     $15 = (invoke_iii(4154,($1|0),($0|0))|0); //@line 166 "modules/graphics/opengl/wrap_Font.cpp"
     $16 = __THREW__; __THREW__ = 0;
     $17 = $16&1;
     if ($17) {
      $hasglyph$0$lcssa19 = $hasglyph$0;
      label = 8;
      break;
     }
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 166 "modules/graphics/opengl/wrap_Font.cpp"
     if ($15) {
      $hasglyph$18$in = $15;
     } else {
      $hasglyph$2 = 0;
      break L1;
     }
    } else {
     __THREW__ = 0;
     $20 = (+invoke_dii(3361,($L|0),($i$01$i$i|0))); //@line 168 "modules/graphics/opengl/wrap_Font.cpp"
     $21 = __THREW__; __THREW__ = 0;
     $22 = $21&1;
     if ($22) {
      $hasglyph$0$lcssa = $hasglyph$0;
      label = 13;
      break;
     }
     $23 = (~~(($20))>>>0); //@line 168 "modules/graphics/opengl/wrap_Font.cpp"
     __THREW__ = 0;
     $24 = (invoke_iii(4155,($1|0),($23|0))|0); //@line 168 "modules/graphics/opengl/wrap_Font.cpp"
     $25 = __THREW__; __THREW__ = 0;
     $26 = $25&1;
     if ($26) {
      $hasglyph$0$lcssa = $hasglyph$0;
      label = 13;
      break;
     }
     if ($24) {
      $hasglyph$18$in = $24;
     } else {
      $hasglyph$2 = 0;
      break L1;
     }
    }
    $hasglyph$18 = $hasglyph$18$in&1; //@line 166 "modules/graphics/opengl/wrap_Font.cpp"
    $7 = (($i$01$i$i) + 1)|0; //@line 163 "modules/graphics/opengl/wrap_Font.cpp"
    $8 = ($i$01$i$i|0)>($5|0); //@line 163 "modules/graphics/opengl/wrap_Font.cpp"
    if ($8) {
     $hasglyph$2 = $hasglyph$18;
     break L1;
    } else {
     $hasglyph$0 = $hasglyph$18;$i$01$i$i = $7;
    }
   }
   if ((label|0) == 8) {
    $18 = ___cxa_find_matching_catch_3(8992|0)|0;
    $19 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 166 "modules/graphics/opengl/wrap_Font.cpp"
    $eh$lpad$body$i$index2Z2D = $19;$eh$lpad$body$i$indexZ2D = $18;$hasglyph$020 = $hasglyph$0$lcssa19;
   }
   else if ((label|0) == 13) {
    $27 = ___cxa_find_matching_catch_3(8992|0)|0;
    $28 = tempRet0;
    $eh$lpad$body$i$index2Z2D = $28;$eh$lpad$body$i$indexZ2D = $27;$hasglyph$020 = $hasglyph$0$lcssa;
   }
   $29 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
   $30 = ($eh$lpad$body$i$index2Z2D|0)==($29|0); //@line 535 "./common/runtime.h"
   if (!($30)) {
    $$01$i = $eh$lpad$body$i$index2Z2D;$$02$i = $eh$lpad$body$i$indexZ2D;
    ___resumeException($$02$i|0);
    // unreachable;
   }
   $31 = (___cxa_begin_catch(($eh$lpad$body$i$indexZ2D|0))|0); //@line 535 "./common/runtime.h"
   $32 = HEAP32[$31>>2]|0; //@line 539 "./common/runtime.h"
   $33 = ((($32)) + 8|0); //@line 539 "./common/runtime.h"
   $34 = HEAP32[$33>>2]|0; //@line 539 "./common/runtime.h"
   $35 = (FUNCTION_TABLE_ii[$34 & 8191]($31)|0); //@line 539 "./common/runtime.h"
   __THREW__ = 0;
   invoke_vii(3380,($L|0),($35|0)); //@line 539 "./common/runtime.h"
   $36 = __THREW__; __THREW__ = 0;
   $37 = $36&1;
   if (!($37)) {
    ___cxa_end_catch(); //@line 540 "./common/runtime.h"
    $38 = (_lua_tolstring($L,-1,0)|0); //@line 543 "./common/runtime.h"
    HEAP32[$vararg_buffer>>2] = $38; //@line 543 "./common/runtime.h"
    (_luaL_error($L,595867,$vararg_buffer)|0); //@line 543 "./common/runtime.h"
    $hasglyph$3 = $hasglyph$020;
    $45 = ($hasglyph$3<<24>>24)!=(0); //@line 175 "modules/graphics/opengl/wrap_Font.cpp"
    __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$45); //@line 175 "modules/graphics/opengl/wrap_Font.cpp"
    STACKTOP = sp;return 1; //@line 176 "modules/graphics/opengl/wrap_Font.cpp"
   }
   $39 = ___cxa_find_matching_catch_2()|0;
   $40 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 540 "./common/runtime.h"
   $41 = __THREW__; __THREW__ = 0;
   $42 = $41&1;
   if ($42) {
    $43 = ___cxa_find_matching_catch_3(0|0)|0;
    $44 = tempRet0;
    ___clang_call_terminate($43); //@line 540 "./common/runtime.h"
    // unreachable; //@line 540 "./common/runtime.h"
   } else {
    $$01$i = $40;$$02$i = $39;
    ___resumeException($$02$i|0);
    // unreachable;
   }
  } else {
   $hasglyph$2 = 0;
  }
 } while(0);
 $hasglyph$3 = $hasglyph$2;
 $45 = ($hasglyph$3<<24>>24)!=(0); //@line 175 "modules/graphics/opengl/wrap_Font.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$45); //@line 175 "modules/graphics/opengl/wrap_Font.cpp"
 STACKTOP = sp;return 1; //@line 176 "modules/graphics/opengl/wrap_Font.cpp"
}
function __ZN4love8graphics6opengl19w_Font_setFallbacksEP9lua_State($L) {
 $L = $L|0;
 var $$01$i = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index3Z2D = 0, $eh$lpad$body$indexZ2D = 0, $fallbacks = 0, $i$0 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i4 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i5 = 0, $vararg_buffer = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $fallbacks = sp + 8|0;
 $0 = sp + 4|0;
 $1 = (__ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE($L,1,14)|0); //@line 38 "modules/graphics/opengl/wrap_Font.cpp"
 HEAP32[$fallbacks>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($fallbacks)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($fallbacks)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$3>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $4 = ((($fallbacks)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $i$0 = 2;
 while(1) {
  __THREW__ = 0;
  $5 = (invoke_ii(3385,($L|0))|0); //@line 184 "modules/graphics/opengl/wrap_Font.cpp"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if ($7) {
   label = 13;
   break;
  }
  $8 = ($i$0|0)>($5|0); //@line 184 "modules/graphics/opengl/wrap_Font.cpp"
  if ($8) {
   label = 4;
   break;
  }
  __THREW__ = 0;
  $48 = (invoke_iiii(4157,($L|0),($i$0|0),14)|0); //@line 38 "modules/graphics/opengl/wrap_Font.cpp"
  $49 = __THREW__; __THREW__ = 0;
  $50 = $49&1;
  if ($50) {
   label = 13;
   break;
  }
  HEAP32[$0>>2] = $48; //@line 185 "modules/graphics/opengl/wrap_Font.cpp"
  $51 = HEAP32[$2>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $52 = HEAP32[$4>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $53 = ($51>>>0)<($52>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($53) {
   $54 = $48; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$51>>2] = $54; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $55 = HEAP32[$2>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $56 = ((($55)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$2>>2] = $56; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  } else {
   __THREW__ = 0;
   invoke_vii(4158,($fallbacks|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $57 = __THREW__; __THREW__ = 0;
   $58 = $57&1;
   if ($58) {
    label = 13;
    break;
   }
  }
  $59 = (($i$0) + 1)|0; //@line 184 "modules/graphics/opengl/wrap_Font.cpp"
  $i$0 = $59;
 }
 L10: do {
  if ((label|0) == 4) {
   __THREW__ = 0;
   invoke_vii(4156,($1|0),($fallbacks|0)); //@line 187 "modules/graphics/opengl/wrap_Font.cpp"
   $9 = __THREW__; __THREW__ = 0;
   $10 = $9&1;
   L13: do {
    if ($10) {
     $11 = ___cxa_find_matching_catch_3(8992|0)|0;
     $12 = tempRet0;
     $13 = (_llvm_eh_typeid_for((8992|0))|0); //@line 535 "./common/runtime.h"
     $14 = ($12|0)==($13|0); //@line 535 "./common/runtime.h"
     do {
      if ($14) {
       $15 = (___cxa_begin_catch(($11|0))|0); //@line 535 "./common/runtime.h"
       $16 = HEAP32[$15>>2]|0; //@line 539 "./common/runtime.h"
       $17 = ((($16)) + 8|0); //@line 539 "./common/runtime.h"
       $18 = HEAP32[$17>>2]|0; //@line 539 "./common/runtime.h"
       $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 539 "./common/runtime.h"
       __THREW__ = 0;
       invoke_vii(3380,($L|0),($19|0)); //@line 539 "./common/runtime.h"
       $20 = __THREW__; __THREW__ = 0;
       $21 = $20&1;
       if ($21) {
        $29 = ___cxa_find_matching_catch_2()|0;
        $30 = tempRet0;
        __THREW__ = 0;
        invoke_v(3378); //@line 540 "./common/runtime.h"
        $31 = __THREW__; __THREW__ = 0;
        $32 = $31&1;
        if (!($32)) {
         $$01$i = $30;$$02$i = $29;
         break;
        }
        $33 = ___cxa_find_matching_catch_3(0|0)|0;
        $34 = tempRet0;
        ___clang_call_terminate($33); //@line 540 "./common/runtime.h"
        // unreachable; //@line 540 "./common/runtime.h"
       }
       __THREW__ = 0;
       invoke_v(3378); //@line 540 "./common/runtime.h"
       $22 = __THREW__; __THREW__ = 0;
       $23 = $22&1;
       if (!($23)) {
        __THREW__ = 0;
        $24 = (invoke_iiii(3392,($L|0),-1,(0|0))|0); //@line 543 "./common/runtime.h"
        $25 = __THREW__; __THREW__ = 0;
        $26 = $25&1;
        if (!($26)) {
         __THREW__ = 0;
         HEAP32[$vararg_buffer>>2] = $24; //@line 543 "./common/runtime.h"
         (invoke_iiii(3389,($L|0),(595867|0),($vararg_buffer|0))|0); //@line 543 "./common/runtime.h"
         $27 = __THREW__; __THREW__ = 0;
         $28 = $27&1;
         if (!($28)) {
          break L13;
         }
        }
       }
       $37 = ___cxa_find_matching_catch_2()|0;
       $38 = tempRet0;
       $eh$lpad$body$index3Z2D = $38;$eh$lpad$body$indexZ2D = $37;
       break L10;
      } else {
       $$01$i = $12;$$02$i = $11;
      }
     } while(0);
     $eh$lpad$body$index3Z2D = $$01$i;$eh$lpad$body$indexZ2D = $$02$i;
     break L10;
    }
   } while(0);
   $60 = HEAP32[$fallbacks>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $61 = ($60|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $62 = $60; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($61) {
    STACKTOP = sp;return 0; //@line 189 "modules/graphics/opengl/wrap_Font.cpp"
   }
   $63 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $64 = ($63|0)==($60|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($64)) {
    $scevgep$i$i$i = ((($63)) + -4|0);
    $65 = $scevgep$i$i$i;
    $66 = (($65) - ($62))|0;
    $67 = $66 >>> 2;
    $68 = $67 ^ -1;
    $scevgep4$i$i$i = (($63) + ($68<<2)|0);
    HEAP32[$2>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($60); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   STACKTOP = sp;return 0; //@line 189 "modules/graphics/opengl/wrap_Font.cpp"
  }
  else if ((label|0) == 13) {
   $35 = ___cxa_find_matching_catch_2()|0;
   $36 = tempRet0;
   $eh$lpad$body$index3Z2D = $36;$eh$lpad$body$indexZ2D = $35;
  }
 } while(0);
 $39 = HEAP32[$fallbacks>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $40 = ($39|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $41 = $39; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($40) {
  ___resumeException($eh$lpad$body$indexZ2D|0);
  // unreachable;
 }
 $42 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $43 = ($42|0)==($39|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($43)) {
  $scevgep$i$i$i4 = ((($42)) + -4|0);
  $44 = $scevgep$i$i$i4;
  $45 = (($44) - ($41))|0;
  $46 = $45 >>> 2;
  $47 = $46 ^ -1;
  $scevgep4$i$i$i5 = (($42) + ($47<<2)|0);
  HEAP32[$2>>2] = $scevgep4$i$i$i5; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($39); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__16vectorIPN4love8graphics6opengl4FontENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function _luaopen_font($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 208 "modules/graphics/opengl/wrap_Font.cpp"
 HEAP32[$vararg_buffer>>2] = 25604; //@line 210 "modules/graphics/opengl/wrap_Font.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 210 "modules/graphics/opengl/wrap_Font.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 210 "modules/graphics/opengl/wrap_Font.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,14,439394,$vararg_buffer)|0); //@line 210 "modules/graphics/opengl/wrap_Font.cpp"
 STACKTOP = sp;return ($0|0); //@line 210 "modules/graphics/opengl/wrap_Font.cpp"
}
function __ZN4love8graphics8VolatileD2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 HEAP32[$this>>2] = (25716); //@line 38 "modules/graphics/Volatile.cpp"
 HEAP32[$0>>2] = $this; //@line 40 "modules/graphics/Volatile.cpp"
 __ZNSt3__14listIPN4love8graphics8VolatileENS_9allocatorIS4_EEE6removeERKS4_(635004,$0); //@line 40 "modules/graphics/Volatile.cpp"
 STACKTOP = sp;return; //@line 41 "modules/graphics/Volatile.cpp"
}
function __ZNSt3__14listIPN4love8graphics8VolatileENS_9allocatorIS4_EEE6removeERKS4_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$cast$i$i$i = 0, $$lcssa = 0, $$lcssa13 = 0, $$lcssa13$us = 0, $$lcssa14 = 0, $$lcssa14$us = 0, $$lcssa37 = 0, $$lcssa38 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__deleted_nodes = 0;
 var $__f$01$i$i = 0, $__i$sroa$0$0$be = 0, $__i$sroa$0$0$be$us = 0, $__i$sroa$0$023 = 0, $__i$sroa$0$023$us = 0, $__j$sroa$0$0$lcssa = 0, $__j$sroa$0$0$lcssa$us = 0, $__j$sroa$0$016 = 0, $__j$sroa$0$016$us = 0, $__r$01$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__deleted_nodes = sp;
 HEAP32[$__deleted_nodes>>2] = $__deleted_nodes; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $0 = ((($__deleted_nodes)) + 4|0); //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$0>>2] = $__deleted_nodes; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $1 = ((($__deleted_nodes)) + 8|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$1>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($this)) + 4|0); //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $3 = HEAP32[$2>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $4 = $3; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $5 = ($4|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($5) {
  STACKTOP = sp;return; //@line 2073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 }
 $6 = ($__deleted_nodes|0)==($this|0); //@line 2017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $7 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($6) {
  $9 = $4;$__i$sroa$0$023$us = $3;
  while(1) {
   $8 = ((($9)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $10 = HEAP32[$8>>2]|0; //@line 2060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $11 = HEAP32[$__x>>2]|0; //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $12 = ($10|0)==($11|0); //@line 2060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $13 = $__i$sroa$0$023$us; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $14 = ((($13)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $15 = HEAP32[$14>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($12) {
    $16 = $15; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $17 = ($16|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    L9: do {
     if ($17) {
      $$lcssa13$us = $this;$$lcssa14$us = 0;$__j$sroa$0$0$lcssa$us = $15;
     } else {
      $19 = $16;$__j$sroa$0$016$us = $15;
      while(1) {
       $18 = ((($19)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $20 = HEAP32[$18>>2]|0; //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $21 = ($20|0)==($10|0); //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       if (!($21)) {
        $$lcssa13$us = $19;$$lcssa14$us = 1;$__j$sroa$0$0$lcssa$us = $__j$sroa$0$016$us;
        break L9;
       }
       $39 = $__j$sroa$0$016$us; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $40 = ((($39)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $41 = HEAP32[$40>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $42 = $41; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $43 = ($42|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       if ($43) {
        $$lcssa13$us = $this;$$lcssa14$us = 0;$__j$sroa$0$0$lcssa$us = $41;
        break;
       } else {
        $19 = $42;$__j$sroa$0$016$us = $41;
       }
      }
     }
    } while(0);
    $22 = ($9|0)==($$lcssa13$us|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if (!($22)) {
     $23 = $__j$sroa$0$0$lcssa$us; //@line 487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $24 = HEAP32[$23>>2]|0; //@line 487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $25 = ((($24)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $26 = HEAP32[$25>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $27 = $__i$sroa$0$023$us; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $28 = HEAP32[$27>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $29 = ((($28)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$29>>2] = $26; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $30 = $28; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $31 = HEAP32[$25>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$31>>2] = $30; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $32 = HEAP32[$__deleted_nodes>>2]|0; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $33 = ((($32)) + 4|0); //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$33>>2] = $9; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $34 = $32; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $35 = $__i$sroa$0$023$us; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$35>>2] = $34; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$__deleted_nodes>>2] = $24; //@line 1075 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$25>>2] = $this; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    }
    if ($$lcssa14$us) {
     $36 = $__j$sroa$0$0$lcssa$us; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $37 = ((($36)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $38 = HEAP32[$37>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $__i$sroa$0$0$be$us = $38;
    } else {
     $__i$sroa$0$0$be$us = $__j$sroa$0$0$lcssa$us;
    }
   } else {
    $__i$sroa$0$0$be$us = $15;
   }
   $44 = $__i$sroa$0$0$be$us; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $45 = ($44|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($45) {
    break;
   } else {
    $9 = $44;$__i$sroa$0$023$us = $__i$sroa$0$0$be$us;
   }
  }
  STACKTOP = sp;return; //@line 2073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 }
 $106 = $__deleted_nodes;$59 = $4;$87 = 0;$__i$sroa$0$023 = $3;
 while(1) {
  $58 = ((($59)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $60 = HEAP32[$58>>2]|0; //@line 2060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $61 = HEAP32[$__x>>2]|0; //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $62 = ($60|0)==($61|0); //@line 2060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $63 = $__i$sroa$0$023; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $64 = ((($63)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $65 = HEAP32[$64>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($62) {
   $66 = $65; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $67 = ($66|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   L25: do {
    if ($67) {
     $$lcssa13 = $this;$$lcssa14 = 0;$__j$sroa$0$0$lcssa = $65;
    } else {
     $69 = $66;$__j$sroa$0$016 = $65;
     while(1) {
      $68 = ((($69)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $70 = HEAP32[$68>>2]|0; //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $71 = ($70|0)==($60|0); //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if (!($71)) {
       $$lcssa13 = $69;$$lcssa14 = 1;$__j$sroa$0$0$lcssa = $__j$sroa$0$016;
       break L25;
      }
      $72 = $__j$sroa$0$016; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $73 = ((($72)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $74 = HEAP32[$73>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $75 = $74; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $76 = ($75|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if ($76) {
       $$lcssa13 = $this;$$lcssa14 = 0;$__j$sroa$0$0$lcssa = $74;
       break;
      } else {
       $69 = $75;$__j$sroa$0$016 = $74;
      }
     }
    }
   } while(0);
   $77 = ($59|0)==($$lcssa13|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($77) {
    $109 = $106;$110 = $87;
   } else {
    $80 = $59;$__r$01$i$i$i = 0;
    while(1) {
     $78 = (($__r$01$i$i$i) + 1)|0; //@line 485 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $79 = ((($80)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $81 = HEAP32[$79>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $$cast$i$i$i = $81; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $82 = ($$cast$i$i$i|0)==($$lcssa13|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $83 = $81; //@line 484 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($82) {
      $$lcssa = $78;
      break;
     } else {
      $80 = $83;$__r$01$i$i$i = $78;
     }
    }
    $84 = HEAP32[$7>>2]|0; //@line 2020 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $85 = (($84) - ($$lcssa))|0; //@line 2020 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$7>>2] = $85; //@line 2020 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $86 = (($87) + ($$lcssa))|0; //@line 2021 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$1>>2] = $86; //@line 2021 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $88 = $__j$sroa$0$0$lcssa; //@line 487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $89 = HEAP32[$88>>2]|0; //@line 487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $90 = ((($89)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $91 = HEAP32[$90>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $92 = $__i$sroa$0$023; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $93 = HEAP32[$92>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $94 = ((($93)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$94>>2] = $91; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $95 = $93; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $96 = HEAP32[$90>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$96>>2] = $95; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $97 = HEAP32[$__deleted_nodes>>2]|0; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $98 = ((($97)) + 4|0); //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$98>>2] = $59; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $99 = $97; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $100 = $__i$sroa$0$023; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$100>>2] = $99; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$__deleted_nodes>>2] = $89; //@line 1075 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$90>>2] = $__deleted_nodes; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $109 = $89;$110 = $86;
   }
   if ($$lcssa14) {
    $101 = $__j$sroa$0$0$lcssa; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $102 = ((($101)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $103 = HEAP32[$102>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $107 = $109;$108 = $110;$__i$sroa$0$0$be = $103;
   } else {
    $107 = $109;$108 = $110;$__i$sroa$0$0$be = $__j$sroa$0$0$lcssa;
   }
  } else {
   $107 = $106;$108 = $87;$__i$sroa$0$0$be = $65;
  }
  $104 = $__i$sroa$0$0$be; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $105 = ($104|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($105) {
   $$lcssa37 = $107;$$lcssa38 = $108;
   break;
  } else {
   $106 = $107;$59 = $104;$87 = $108;$__i$sroa$0$023 = $__i$sroa$0$0$be;
  }
 }
 $46 = ($$lcssa38|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($46) {
  STACKTOP = sp;return; //@line 2073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 }
 $47 = HEAP32[$0>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $48 = ((($$lcssa37)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $49 = HEAP32[$48>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $50 = HEAP32[$47>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $51 = ((($50)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$51>>2] = $49; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $52 = $50; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $53 = HEAP32[$48>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$53>>2] = $52; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$1>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $54 = ($47|0)==($__deleted_nodes|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($54) {
  STACKTOP = sp;return; //@line 2073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 } else {
  $__f$01$i$i = $47;
 }
 while(1) {
  $55 = ((($__f$01$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $56 = HEAP32[$55>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  __ZdlPv($__f$01$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $57 = ($56|0)==($__deleted_nodes|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($57) {
   break;
  } else {
   $__f$01$i$i = $56;
  }
 }
 STACKTOP = sp;return; //@line 2073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
}
function __ZN4love8graphics8VolatileD0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 HEAP32[$this>>2] = (25716); //@line 38 "modules/graphics/Volatile.cpp"
 HEAP32[$0>>2] = $this; //@line 40 "modules/graphics/Volatile.cpp"
 __ZNSt3__14listIPN4love8graphics8VolatileENS_9allocatorIS4_EEE6removeERKS4_(635004,$0); //@line 40 "modules/graphics/Volatile.cpp"
 __ZdlPv($this); //@line 38 "modules/graphics/Volatile.cpp"
 STACKTOP = sp;return; //@line 41 "modules/graphics/Volatile.cpp"
}
function __GLOBAL__sub_I_Volatile_cpp() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[158751] = 635004; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[(635008)>>2] = 635004; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[(635012)>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 (___cxa_atexit((4159|0),(635004|0),(___dso_handle|0))|0); //@line 29 "modules/graphics/Volatile.cpp"
 return;
}
function __ZNSt3__110__list_impIPN4love8graphics8VolatileENS_9allocatorIS4_EEED2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__f$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $2 = ($1|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($2) {
  return; //@line 701 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 }
 $3 = ((($this)) + 4|0); //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $4 = HEAP32[$3>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $5 = HEAP32[$this>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $6 = ((($5)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $7 = HEAP32[$6>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $8 = HEAP32[$4>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $9 = ((($8)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$9>>2] = $7; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $10 = $8; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $11 = HEAP32[$6>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$11>>2] = $10; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$0>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $12 = ($4|0)==($this|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($12) {
  return; //@line 701 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 } else {
  $__f$01$i = $4;
 }
 while(1) {
  $13 = ((($__f$01$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $14 = HEAP32[$13>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  __ZdlPv($__f$01$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $15 = ($14|0)==($this|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($15) {
   break;
  } else {
   $__f$01$i = $14;
  }
 }
 return; //@line 701 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
}
function __ZN4love8graphics8VolatileC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (25716); //@line 32 "modules/graphics/Volatile.cpp"
 $0 = $this; //@line 34 "modules/graphics/Volatile.cpp"
 $1 = (__Znwj(12)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $2 = ((($1)) + 8|0); //@line 1565 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$2>>2] = $0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = ((($1)) + 4|0); //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$3>>2] = 635004; //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $4 = HEAP32[158751]|0; //@line 1098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$1>>2] = $4; //@line 1098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $5 = ((($4)) + 4|0); //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$5>>2] = $1; //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[158751] = $1; //@line 1100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $6 = HEAP32[(635012)>>2]|0; //@line 1567 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $7 = (($6) + 1)|0; //@line 1567 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[(635012)>>2] = $7; //@line 1567 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 return; //@line 35 "modules/graphics/Volatile.cpp"
}
function __ZN4love8graphics8Volatile7loadAllEv() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$01 = 0, $__begin$sroa$0$03 = 0, $success$0$off0$lcssa = 0, $success$0$off02 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $__begin$sroa$0$01 = HEAP32[(635008)>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $0 = $__begin$sroa$0$01; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $1 = ($0|0)==(635004|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($1) {
  $success$0$off0$lcssa = 1;
  return ($success$0$off0$lcssa|0); //@line 50 "modules/graphics/Volatile.cpp"
 } else {
  $3 = $0;$__begin$sroa$0$03 = $__begin$sroa$0$01;$success$0$off02 = 1;
 }
 while(1) {
  $2 = ((($3)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $4 = HEAP32[$2>>2]|0; //@line 47 "modules/graphics/Volatile.cpp"
  if ($success$0$off02) {
   $5 = HEAP32[$4>>2]|0; //@line 48 "modules/graphics/Volatile.cpp"
   $6 = ((($5)) + 8|0); //@line 48 "modules/graphics/Volatile.cpp"
   $7 = HEAP32[$6>>2]|0; //@line 48 "modules/graphics/Volatile.cpp"
   $8 = (FUNCTION_TABLE_ii[$7 & 8191]($4)|0); //@line 48 "modules/graphics/Volatile.cpp"
   $13 = $8;
  } else {
   $13 = 0;
  }
  $9 = $__begin$sroa$0$03; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $10 = ((($9)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $__begin$sroa$0$0 = HEAP32[$10>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $11 = $__begin$sroa$0$0; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $12 = ($11|0)==(635004|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($12) {
   $success$0$off0$lcssa = $13;
   break;
  } else {
   $3 = $11;$__begin$sroa$0$03 = $__begin$sroa$0$0;$success$0$off02 = $13;
  }
 }
 return ($success$0$off0$lcssa|0); //@line 50 "modules/graphics/Volatile.cpp"
}
function __ZN4love8graphics8Volatile9unloadAllEv() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$01 = 0, $__begin$sroa$0$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $__begin$sroa$0$01 = HEAP32[(635008)>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $0 = $__begin$sroa$0$01; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $1 = ($0|0)==(635004|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($1) {
  return; //@line 57 "modules/graphics/Volatile.cpp"
 } else {
  $3 = $0;$__begin$sroa$0$02 = $__begin$sroa$0$01;
 }
 while(1) {
  $2 = ((($3)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $4 = HEAP32[$2>>2]|0; //@line 55 "modules/graphics/Volatile.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 56 "modules/graphics/Volatile.cpp"
  $6 = ((($5)) + 12|0); //@line 56 "modules/graphics/Volatile.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 56 "modules/graphics/Volatile.cpp"
  FUNCTION_TABLE_vi[$7 & 8191]($4); //@line 56 "modules/graphics/Volatile.cpp"
  $8 = $__begin$sroa$0$02; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $9 = ((($8)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $__begin$sroa$0$0 = HEAP32[$9>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $10 = $__begin$sroa$0$0; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $11 = ($10|0)==(635004|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($11) {
   break;
  } else {
   $3 = $10;$__begin$sroa$0$02 = $__begin$sroa$0$0;
  }
 }
 return; //@line 57 "modules/graphics/Volatile.cpp"
}
function __ZN4love5image6magpie5ImageD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$013 = 0;
 var $__begin$sroa$0$014 = 0, $__begin2$sroa$0$0 = 0, $__begin2$sroa$0$010 = 0, $__begin2$sroa$0$011 = 0, $__f$01$i$i = 0, $__f$01$i$i1 = 0, $__f$01$i$i3 = 0, $__f$01$i$i6 = 0, $lpad$phi$index = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (25740); //@line 61 "modules/image/magpie/Image.cpp"
 $0 = ((($this)) + 8|0); //@line 64 "modules/image/magpie/Image.cpp"
 $1 = ((($this)) + 12|0); //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $__begin$sroa$0$013 = HEAP32[$1>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $2 = $__begin$sroa$0$013; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $3 = ($2|0)==($0|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 L1: do {
  if ($3) {
   label = 2;
  } else {
   $45 = $2;$__begin$sroa$0$014 = $__begin$sroa$0$013;
   while(1) {
    $44 = ((($45)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $46 = HEAP32[$44>>2]|0; //@line 64 "modules/image/magpie/Image.cpp"
    $47 = HEAP32[$46>>2]|0; //@line 65 "modules/image/magpie/Image.cpp"
    $48 = ((($47)) + 12|0); //@line 65 "modules/image/magpie/Image.cpp"
    $49 = HEAP32[$48>>2]|0; //@line 65 "modules/image/magpie/Image.cpp"
    __THREW__ = 0;
    invoke_vi($49|0,($46|0)); //@line 65 "modules/image/magpie/Image.cpp"
    $50 = __THREW__; __THREW__ = 0;
    $51 = $50&1;
    if ($51) {
     break;
    }
    $52 = $__begin$sroa$0$014; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $53 = ((($52)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $__begin$sroa$0$0 = HEAP32[$53>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $54 = $__begin$sroa$0$0; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $55 = ($54|0)==($0|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($55) {
     label = 2;
     break L1;
    } else {
     $45 = $54;$__begin$sroa$0$014 = $__begin$sroa$0$0;
    }
   }
   $10 = ___cxa_find_matching_catch_3(0|0)|0;
   $11 = tempRet0;
   $lpad$phi$index = $10;
  }
 } while(0);
 L6: do {
  if ((label|0) == 2) {
   $4 = ((($this)) + 20|0); //@line 67 "modules/image/magpie/Image.cpp"
   $5 = ((($this)) + 24|0); //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $__begin2$sroa$0$010 = HEAP32[$5>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $6 = $__begin2$sroa$0$010; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $7 = ($6|0)==($4|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   L8: do {
    if (!($7)) {
     $87 = $6;$__begin2$sroa$0$011 = $__begin2$sroa$0$010;
     while(1) {
      $86 = ((($87)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $88 = HEAP32[$86>>2]|0; //@line 67 "modules/image/magpie/Image.cpp"
      $89 = HEAP32[$88>>2]|0; //@line 68 "modules/image/magpie/Image.cpp"
      $90 = ((($89)) + 12|0); //@line 68 "modules/image/magpie/Image.cpp"
      $91 = HEAP32[$90>>2]|0; //@line 68 "modules/image/magpie/Image.cpp"
      __THREW__ = 0;
      invoke_vi($91|0,($88|0)); //@line 68 "modules/image/magpie/Image.cpp"
      $92 = __THREW__; __THREW__ = 0;
      $93 = $92&1;
      if ($93) {
       break;
      }
      $94 = $__begin2$sroa$0$011; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $95 = ((($94)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $__begin2$sroa$0$0 = HEAP32[$95>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $96 = $__begin2$sroa$0$0; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $97 = ($96|0)==($4|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if ($97) {
       break L8;
      } else {
       $87 = $96;$__begin2$sroa$0$011 = $__begin2$sroa$0$0;
      }
     }
     $8 = ___cxa_find_matching_catch_3(0|0)|0;
     $9 = tempRet0;
     $lpad$phi$index = $8;
     break L6;
    }
   } while(0);
   $56 = ((($this)) + 28|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $57 = HEAP32[$56>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $58 = ($57|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if (!($58)) {
    $59 = HEAP32[$5>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $60 = HEAP32[$4>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $61 = ((($60)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $62 = HEAP32[$61>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $63 = HEAP32[$59>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $64 = ((($63)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$64>>2] = $62; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $65 = $63; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $66 = HEAP32[$61>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$66>>2] = $65; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$56>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $67 = ($59|0)==($4|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if (!($67)) {
     $__f$01$i$i1 = $59;
     while(1) {
      $68 = ((($__f$01$i$i1)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $69 = HEAP32[$68>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      __ZdlPv($__f$01$i$i1); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $70 = ($69|0)==($4|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if ($70) {
       break;
      } else {
       $__f$01$i$i1 = $69;
      }
     }
    }
   }
   $71 = ((($this)) + 16|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $72 = HEAP32[$71>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $73 = ($72|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($73) {
    __ZN4love6ModuleD2Ev($this); //@line 69 "modules/image/magpie/Image.cpp"
    return; //@line 69 "modules/image/magpie/Image.cpp"
   }
   $74 = HEAP32[$1>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $75 = HEAP32[$0>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $76 = ((($75)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $77 = HEAP32[$76>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $78 = HEAP32[$74>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $79 = ((($78)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   HEAP32[$79>>2] = $77; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $80 = $78; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $81 = HEAP32[$76>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   HEAP32[$81>>2] = $80; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   HEAP32[$71>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $82 = ($74|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($82) {
    __ZN4love6ModuleD2Ev($this); //@line 69 "modules/image/magpie/Image.cpp"
    return; //@line 69 "modules/image/magpie/Image.cpp"
   } else {
    $__f$01$i$i = $74;
   }
   while(1) {
    $83 = ((($__f$01$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $84 = HEAP32[$83>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    __ZdlPv($__f$01$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $85 = ($84|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($85) {
     break;
    } else {
     $__f$01$i$i = $84;
    }
   }
   __ZN4love6ModuleD2Ev($this); //@line 69 "modules/image/magpie/Image.cpp"
   return; //@line 69 "modules/image/magpie/Image.cpp"
  }
 } while(0);
 $12 = ((($this)) + 20|0); //@line 69 "modules/image/magpie/Image.cpp"
 $13 = ((($this)) + 28|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $14 = HEAP32[$13>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $15 = ($14|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if (!($15)) {
  $16 = ((($this)) + 24|0); //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $17 = HEAP32[$16>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $18 = HEAP32[$12>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $19 = ((($18)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $20 = HEAP32[$19>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $21 = HEAP32[$17>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $22 = ((($21)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$22>>2] = $20; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $23 = $21; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $24 = HEAP32[$19>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$24>>2] = $23; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$13>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $25 = ($17|0)==($12|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if (!($25)) {
   $__f$01$i$i3 = $17;
   while(1) {
    $26 = ((($__f$01$i$i3)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $27 = HEAP32[$26>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    __ZdlPv($__f$01$i$i3); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $28 = ($27|0)==($12|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($28) {
     break;
    } else {
     $__f$01$i$i3 = $27;
    }
   }
  }
 }
 $29 = ((($this)) + 16|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = HEAP32[$29>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $31 = ($30|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($31) {
  __ZN4love6ModuleD2Ev($this); //@line 69 "modules/image/magpie/Image.cpp"
  ___clang_call_terminate($lpad$phi$index); //@line 69 "modules/image/magpie/Image.cpp"
  // unreachable; //@line 69 "modules/image/magpie/Image.cpp"
 }
 $32 = HEAP32[$1>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $33 = HEAP32[$0>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $34 = ((($33)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $35 = HEAP32[$34>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $36 = HEAP32[$32>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $37 = ((($36)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$37>>2] = $35; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $38 = $36; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $39 = HEAP32[$34>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$39>>2] = $38; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$29>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $40 = ($32|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($40) {
  __ZN4love6ModuleD2Ev($this); //@line 69 "modules/image/magpie/Image.cpp"
  ___clang_call_terminate($lpad$phi$index); //@line 69 "modules/image/magpie/Image.cpp"
  // unreachable; //@line 69 "modules/image/magpie/Image.cpp"
 } else {
  $__f$01$i$i6 = $32;
 }
 while(1) {
  $41 = ((($__f$01$i$i6)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $42 = HEAP32[$41>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  __ZdlPv($__f$01$i$i6); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $43 = ($42|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($43) {
   break;
  } else {
   $__f$01$i$i6 = $42;
  }
 }
 __ZN4love6ModuleD2Ev($this); //@line 69 "modules/image/magpie/Image.cpp"
 ___clang_call_terminate($lpad$phi$index); //@line 69 "modules/image/magpie/Image.cpp"
 // unreachable; //@line 69 "modules/image/magpie/Image.cpp"
}
function __ZN4love5image6magpie5ImageD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5image6magpie5ImageD2Ev($this); //@line 61 "modules/image/magpie/Image.cpp"
 __ZdlPv($this); //@line 61 "modules/image/magpie/Image.cpp"
 return; //@line 69 "modules/image/magpie/Image.cpp"
}
function __ZNK4love5image5Image13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 5; //@line 50 "./modules/image/Image.h"
}
function __ZNK4love5image6magpie5Image7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (439603|0); //@line 73 "modules/image/magpie/Image.cpp"
}
function __ZN4love5image6magpie5Image12newImageDataEPNS_10filesystem8FileDataE($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$0 = 0, $$01 = 0, $$lcssa = 0, $$lcssa21 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $8 = 0, $9 = 0, $__f$01$i$i = 0, $__f$01$i$i$i = 0, $__f$01$i$i2 = 0, $__i$sroa$0$0$i = 0, $__i$sroa$0$02$i = 0, $__i$sroa$0$03$i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__Znwj(40)|0); //@line 78 "modules/image/magpie/Image.cpp"
 $2 = ((($this)) + 8|0); //@line 78 "modules/image/magpie/Image.cpp"
 HEAP32[$0>>2] = $0; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $3 = ((($0)) + 4|0); //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$3>>2] = $0; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $4 = ((($0)) + 8|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$4>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ((($this)) + 12|0); //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $__i$sroa$0$02$i = HEAP32[$5>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $6 = $__i$sroa$0$02$i; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $7 = ($6|0)==($2|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 L1: do {
  if (!($7)) {
   $26 = $6;$30 = $0;$33 = 0;$__i$sroa$0$03$i = $__i$sroa$0$02$i;
   while(1) {
    __THREW__ = 0;
    $22 = (invoke_ii(3348,12)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $23 = __THREW__; __THREW__ = 0;
    $24 = $23&1;
    if ($24) {
     $$lcssa = $30;$$lcssa21 = $33;
     break;
    }
    $25 = ((($26)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $27 = ((($22)) + 8|0); //@line 1537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $28 = HEAP32[$25>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$27>>2] = $28; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $29 = ((($22)) + 4|0); //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$29>>2] = $0; //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$22>>2] = $30; //@line 1098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $31 = ((($30)) + 4|0); //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$31>>2] = $22; //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$0>>2] = $22; //@line 1100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $32 = (($33) + 1)|0; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$4>>2] = $32; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $34 = $__i$sroa$0$03$i; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $35 = ((($34)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $__i$sroa$0$0$i = HEAP32[$35>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $36 = $__i$sroa$0$0$i; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $37 = ($36|0)==($2|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($37) {
     break L1;
    } else {
     $26 = $36;$30 = $22;$33 = $32;$__i$sroa$0$03$i = $__i$sroa$0$0$i;
    }
   }
   $8 = ___cxa_find_matching_catch_2()|0;
   $9 = tempRet0;
   $10 = ($$lcssa21|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if (!($10)) {
    $11 = HEAP32[$3>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $12 = ((($$lcssa)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $13 = HEAP32[$12>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $14 = HEAP32[$11>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $15 = ((($14)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$15>>2] = $13; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $16 = $14; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $17 = HEAP32[$12>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$17>>2] = $16; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$4>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $18 = ($11|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if (!($18)) {
     $__f$01$i$i$i = $11;
     while(1) {
      $19 = ((($__f$01$i$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $20 = HEAP32[$19>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      __ZdlPv($__f$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $21 = ($20|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if ($21) {
       break;
      } else {
       $__f$01$i$i$i = $20;
      }
     }
    }
   }
   $$0 = $9;$$01 = $8;
   __ZdlPv($1); //@line 78 "modules/image/magpie/Image.cpp"
   ___resumeException($$01|0);
   // unreachable;
  }
 } while(0);
 __THREW__ = 0;
 invoke_viii(4160,($1|0),($0|0),($data|0)); //@line 78 "modules/image/magpie/Image.cpp"
 $38 = __THREW__; __THREW__ = 0;
 $39 = $38&1;
 if ($39) {
  $54 = ___cxa_find_matching_catch_2()|0;
  $55 = tempRet0;
  $56 = HEAP32[$4>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $57 = ($56|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($57) {
   $$0 = $55;$$01 = $54;
   __ZdlPv($1); //@line 78 "modules/image/magpie/Image.cpp"
   ___resumeException($$01|0);
   // unreachable;
  }
  $58 = HEAP32[$3>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $59 = HEAP32[$0>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $60 = ((($59)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $61 = HEAP32[$60>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $62 = HEAP32[$58>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $63 = ((($62)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$63>>2] = $61; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $64 = $62; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $65 = HEAP32[$60>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$65>>2] = $64; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$4>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $66 = ($58|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($66) {
   $$0 = $55;$$01 = $54;
   __ZdlPv($1); //@line 78 "modules/image/magpie/Image.cpp"
   ___resumeException($$01|0);
   // unreachable;
  } else {
   $__f$01$i$i2 = $58;
  }
  while(1) {
   $67 = ((($__f$01$i$i2)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $68 = HEAP32[$67>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   __ZdlPv($__f$01$i$i2); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $69 = ($68|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($69) {
    $$0 = $55;$$01 = $54;
    break;
   } else {
    $__f$01$i$i2 = $68;
   }
  }
  __ZdlPv($1); //@line 78 "modules/image/magpie/Image.cpp"
  ___resumeException($$01|0);
  // unreachable;
 } else {
  $40 = HEAP32[$4>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $41 = ($40|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($41) {
   STACKTOP = sp;return ($1|0); //@line 78 "modules/image/magpie/Image.cpp"
  }
  $42 = HEAP32[$3>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $43 = HEAP32[$0>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $44 = ((($43)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $45 = HEAP32[$44>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $46 = HEAP32[$42>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $47 = ((($46)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$47>>2] = $45; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $48 = $46; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $49 = HEAP32[$44>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$49>>2] = $48; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$4>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $50 = ($42|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($50) {
   STACKTOP = sp;return ($1|0); //@line 78 "modules/image/magpie/Image.cpp"
  } else {
   $__f$01$i$i = $42;
  }
  while(1) {
   $51 = ((($__f$01$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $52 = HEAP32[$51>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   __ZdlPv($__f$01$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $53 = ($52|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($53) {
    break;
   } else {
    $__f$01$i$i = $52;
   }
  }
  STACKTOP = sp;return ($1|0); //@line 78 "modules/image/magpie/Image.cpp"
 }
 return (0)|0;
}
function __ZN4love5image6magpie5Image12newImageDataEii($this,$width,$height) {
 $this = $this|0;
 $width = $width|0;
 $height = $height|0;
 var $$0 = 0, $$01 = 0, $$lcssa = 0, $$lcssa21 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $8 = 0, $9 = 0, $__f$01$i$i = 0, $__f$01$i$i$i = 0, $__f$01$i$i2 = 0, $__i$sroa$0$0$i = 0, $__i$sroa$0$02$i = 0, $__i$sroa$0$03$i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__Znwj(40)|0); //@line 83 "modules/image/magpie/Image.cpp"
 $2 = ((($this)) + 8|0); //@line 83 "modules/image/magpie/Image.cpp"
 HEAP32[$0>>2] = $0; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $3 = ((($0)) + 4|0); //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$3>>2] = $0; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $4 = ((($0)) + 8|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$4>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ((($this)) + 12|0); //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $__i$sroa$0$02$i = HEAP32[$5>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $6 = $__i$sroa$0$02$i; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $7 = ($6|0)==($2|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 L1: do {
  if (!($7)) {
   $26 = $6;$30 = $0;$33 = 0;$__i$sroa$0$03$i = $__i$sroa$0$02$i;
   while(1) {
    __THREW__ = 0;
    $22 = (invoke_ii(3348,12)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $23 = __THREW__; __THREW__ = 0;
    $24 = $23&1;
    if ($24) {
     $$lcssa = $30;$$lcssa21 = $33;
     break;
    }
    $25 = ((($26)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $27 = ((($22)) + 8|0); //@line 1537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $28 = HEAP32[$25>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$27>>2] = $28; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $29 = ((($22)) + 4|0); //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$29>>2] = $0; //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$22>>2] = $30; //@line 1098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $31 = ((($30)) + 4|0); //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$31>>2] = $22; //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$0>>2] = $22; //@line 1100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $32 = (($33) + 1)|0; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$4>>2] = $32; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $34 = $__i$sroa$0$03$i; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $35 = ((($34)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $__i$sroa$0$0$i = HEAP32[$35>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $36 = $__i$sroa$0$0$i; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $37 = ($36|0)==($2|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($37) {
     break L1;
    } else {
     $26 = $36;$30 = $22;$33 = $32;$__i$sroa$0$03$i = $__i$sroa$0$0$i;
    }
   }
   $8 = ___cxa_find_matching_catch_2()|0;
   $9 = tempRet0;
   $10 = ($$lcssa21|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if (!($10)) {
    $11 = HEAP32[$3>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $12 = ((($$lcssa)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $13 = HEAP32[$12>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $14 = HEAP32[$11>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $15 = ((($14)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$15>>2] = $13; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $16 = $14; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $17 = HEAP32[$12>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$17>>2] = $16; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$4>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $18 = ($11|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if (!($18)) {
     $__f$01$i$i$i = $11;
     while(1) {
      $19 = ((($__f$01$i$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $20 = HEAP32[$19>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      __ZdlPv($__f$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $21 = ($20|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if ($21) {
       break;
      } else {
       $__f$01$i$i$i = $20;
      }
     }
    }
   }
   $$0 = $9;$$01 = $8;
   __ZdlPv($1); //@line 83 "modules/image/magpie/Image.cpp"
   ___resumeException($$01|0);
   // unreachable;
  }
 } while(0);
 __THREW__ = 0;
 invoke_viiii(4161,($1|0),($0|0),($width|0),($height|0)); //@line 83 "modules/image/magpie/Image.cpp"
 $38 = __THREW__; __THREW__ = 0;
 $39 = $38&1;
 if ($39) {
  $54 = ___cxa_find_matching_catch_2()|0;
  $55 = tempRet0;
  $56 = HEAP32[$4>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $57 = ($56|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($57) {
   $$0 = $55;$$01 = $54;
   __ZdlPv($1); //@line 83 "modules/image/magpie/Image.cpp"
   ___resumeException($$01|0);
   // unreachable;
  }
  $58 = HEAP32[$3>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $59 = HEAP32[$0>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $60 = ((($59)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $61 = HEAP32[$60>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $62 = HEAP32[$58>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $63 = ((($62)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$63>>2] = $61; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $64 = $62; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $65 = HEAP32[$60>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$65>>2] = $64; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$4>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $66 = ($58|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($66) {
   $$0 = $55;$$01 = $54;
   __ZdlPv($1); //@line 83 "modules/image/magpie/Image.cpp"
   ___resumeException($$01|0);
   // unreachable;
  } else {
   $__f$01$i$i2 = $58;
  }
  while(1) {
   $67 = ((($__f$01$i$i2)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $68 = HEAP32[$67>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   __ZdlPv($__f$01$i$i2); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $69 = ($68|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($69) {
    $$0 = $55;$$01 = $54;
    break;
   } else {
    $__f$01$i$i2 = $68;
   }
  }
  __ZdlPv($1); //@line 83 "modules/image/magpie/Image.cpp"
  ___resumeException($$01|0);
  // unreachable;
 } else {
  $40 = HEAP32[$4>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $41 = ($40|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($41) {
   STACKTOP = sp;return ($1|0); //@line 83 "modules/image/magpie/Image.cpp"
  }
  $42 = HEAP32[$3>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $43 = HEAP32[$0>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $44 = ((($43)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $45 = HEAP32[$44>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $46 = HEAP32[$42>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $47 = ((($46)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$47>>2] = $45; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $48 = $46; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $49 = HEAP32[$44>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$49>>2] = $48; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$4>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $50 = ($42|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($50) {
   STACKTOP = sp;return ($1|0); //@line 83 "modules/image/magpie/Image.cpp"
  } else {
   $__f$01$i$i = $42;
  }
  while(1) {
   $51 = ((($__f$01$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $52 = HEAP32[$51>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   __ZdlPv($__f$01$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $53 = ($52|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($53) {
    break;
   } else {
    $__f$01$i$i = $52;
   }
  }
  STACKTOP = sp;return ($1|0); //@line 83 "modules/image/magpie/Image.cpp"
 }
 return (0)|0;
}
function __ZN4love5image6magpie5Image12newImageDataEiiPvb($this,$width,$height,$data,$own) {
 $this = $this|0;
 $width = $width|0;
 $height = $height|0;
 $data = $data|0;
 $own = $own|0;
 var $$0 = 0, $$01 = 0, $$lcssa = 0, $$lcssa21 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $8 = 0, $9 = 0, $__f$01$i$i = 0, $__f$01$i$i$i = 0, $__f$01$i$i2 = 0, $__i$sroa$0$0$i = 0, $__i$sroa$0$02$i = 0, $__i$sroa$0$03$i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__Znwj(40)|0); //@line 88 "modules/image/magpie/Image.cpp"
 $2 = ((($this)) + 8|0); //@line 88 "modules/image/magpie/Image.cpp"
 HEAP32[$0>>2] = $0; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $3 = ((($0)) + 4|0); //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$3>>2] = $0; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $4 = ((($0)) + 8|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$4>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ((($this)) + 12|0); //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $__i$sroa$0$02$i = HEAP32[$5>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $6 = $__i$sroa$0$02$i; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $7 = ($6|0)==($2|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 L1: do {
  if (!($7)) {
   $26 = $6;$30 = $0;$33 = 0;$__i$sroa$0$03$i = $__i$sroa$0$02$i;
   while(1) {
    __THREW__ = 0;
    $22 = (invoke_ii(3348,12)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $23 = __THREW__; __THREW__ = 0;
    $24 = $23&1;
    if ($24) {
     $$lcssa = $30;$$lcssa21 = $33;
     break;
    }
    $25 = ((($26)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $27 = ((($22)) + 8|0); //@line 1537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $28 = HEAP32[$25>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$27>>2] = $28; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $29 = ((($22)) + 4|0); //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$29>>2] = $0; //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$22>>2] = $30; //@line 1098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $31 = ((($30)) + 4|0); //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$31>>2] = $22; //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$0>>2] = $22; //@line 1100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $32 = (($33) + 1)|0; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$4>>2] = $32; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $34 = $__i$sroa$0$03$i; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $35 = ((($34)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $__i$sroa$0$0$i = HEAP32[$35>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $36 = $__i$sroa$0$0$i; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $37 = ($36|0)==($2|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($37) {
     break L1;
    } else {
     $26 = $36;$30 = $22;$33 = $32;$__i$sroa$0$03$i = $__i$sroa$0$0$i;
    }
   }
   $8 = ___cxa_find_matching_catch_2()|0;
   $9 = tempRet0;
   $10 = ($$lcssa21|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if (!($10)) {
    $11 = HEAP32[$3>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $12 = ((($$lcssa)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $13 = HEAP32[$12>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $14 = HEAP32[$11>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $15 = ((($14)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$15>>2] = $13; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $16 = $14; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $17 = HEAP32[$12>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$17>>2] = $16; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$4>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $18 = ($11|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if (!($18)) {
     $__f$01$i$i$i = $11;
     while(1) {
      $19 = ((($__f$01$i$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $20 = HEAP32[$19>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      __ZdlPv($__f$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $21 = ($20|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if ($21) {
       break;
      } else {
       $__f$01$i$i$i = $20;
      }
     }
    }
   }
   $$0 = $9;$$01 = $8;
   __ZdlPv($1); //@line 88 "modules/image/magpie/Image.cpp"
   ___resumeException($$01|0);
   // unreachable;
  }
 } while(0);
 __THREW__ = 0;
 invoke_viiiiii(4162,($1|0),($0|0),($width|0),($height|0),($data|0),($own|0)); //@line 88 "modules/image/magpie/Image.cpp"
 $38 = __THREW__; __THREW__ = 0;
 $39 = $38&1;
 if ($39) {
  $54 = ___cxa_find_matching_catch_2()|0;
  $55 = tempRet0;
  $56 = HEAP32[$4>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $57 = ($56|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($57) {
   $$0 = $55;$$01 = $54;
   __ZdlPv($1); //@line 88 "modules/image/magpie/Image.cpp"
   ___resumeException($$01|0);
   // unreachable;
  }
  $58 = HEAP32[$3>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $59 = HEAP32[$0>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $60 = ((($59)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $61 = HEAP32[$60>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $62 = HEAP32[$58>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $63 = ((($62)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$63>>2] = $61; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $64 = $62; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $65 = HEAP32[$60>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$65>>2] = $64; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$4>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $66 = ($58|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($66) {
   $$0 = $55;$$01 = $54;
   __ZdlPv($1); //@line 88 "modules/image/magpie/Image.cpp"
   ___resumeException($$01|0);
   // unreachable;
  } else {
   $__f$01$i$i2 = $58;
  }
  while(1) {
   $67 = ((($__f$01$i$i2)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $68 = HEAP32[$67>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   __ZdlPv($__f$01$i$i2); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $69 = ($68|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($69) {
    $$0 = $55;$$01 = $54;
    break;
   } else {
    $__f$01$i$i2 = $68;
   }
  }
  __ZdlPv($1); //@line 88 "modules/image/magpie/Image.cpp"
  ___resumeException($$01|0);
  // unreachable;
 } else {
  $40 = HEAP32[$4>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $41 = ($40|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($41) {
   STACKTOP = sp;return ($1|0); //@line 88 "modules/image/magpie/Image.cpp"
  }
  $42 = HEAP32[$3>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $43 = HEAP32[$0>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $44 = ((($43)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $45 = HEAP32[$44>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $46 = HEAP32[$42>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $47 = ((($46)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$47>>2] = $45; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $48 = $46; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $49 = HEAP32[$44>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$49>>2] = $48; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$4>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $50 = ($42|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($50) {
   STACKTOP = sp;return ($1|0); //@line 88 "modules/image/magpie/Image.cpp"
  } else {
   $__f$01$i$i = $42;
  }
  while(1) {
   $51 = ((($__f$01$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $52 = HEAP32[$51>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   __ZdlPv($__f$01$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $53 = ($52|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($53) {
    break;
   } else {
    $__f$01$i$i = $52;
   }
  }
  STACKTOP = sp;return ($1|0); //@line 88 "modules/image/magpie/Image.cpp"
 }
 return (0)|0;
}
function __ZN4love5image6magpie5Image17newCompressedDataEPNS_10filesystem8FileDataE($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$0 = 0, $$01 = 0, $$lcssa = 0, $$lcssa21 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $8 = 0, $9 = 0, $__f$01$i$i = 0, $__f$01$i$i$i = 0, $__f$01$i$i2 = 0, $__i$sroa$0$0$i = 0, $__i$sroa$0$02$i = 0, $__i$sroa$0$03$i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__Znwj(36)|0); //@line 93 "modules/image/magpie/Image.cpp"
 $2 = ((($this)) + 20|0); //@line 93 "modules/image/magpie/Image.cpp"
 HEAP32[$0>>2] = $0; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $3 = ((($0)) + 4|0); //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$3>>2] = $0; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $4 = ((($0)) + 8|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$4>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ((($this)) + 24|0); //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $__i$sroa$0$02$i = HEAP32[$5>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $6 = $__i$sroa$0$02$i; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $7 = ($6|0)==($2|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 L1: do {
  if (!($7)) {
   $26 = $6;$30 = $0;$33 = 0;$__i$sroa$0$03$i = $__i$sroa$0$02$i;
   while(1) {
    __THREW__ = 0;
    $22 = (invoke_ii(3348,12)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $23 = __THREW__; __THREW__ = 0;
    $24 = $23&1;
    if ($24) {
     $$lcssa = $30;$$lcssa21 = $33;
     break;
    }
    $25 = ((($26)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $27 = ((($22)) + 8|0); //@line 1537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $28 = HEAP32[$25>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$27>>2] = $28; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $29 = ((($22)) + 4|0); //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$29>>2] = $0; //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$22>>2] = $30; //@line 1098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $31 = ((($30)) + 4|0); //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$31>>2] = $22; //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$0>>2] = $22; //@line 1100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $32 = (($33) + 1)|0; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$4>>2] = $32; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $34 = $__i$sroa$0$03$i; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $35 = ((($34)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $__i$sroa$0$0$i = HEAP32[$35>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $36 = $__i$sroa$0$0$i; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $37 = ($36|0)==($2|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($37) {
     break L1;
    } else {
     $26 = $36;$30 = $22;$33 = $32;$__i$sroa$0$03$i = $__i$sroa$0$0$i;
    }
   }
   $8 = ___cxa_find_matching_catch_2()|0;
   $9 = tempRet0;
   $10 = ($$lcssa21|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if (!($10)) {
    $11 = HEAP32[$3>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $12 = ((($$lcssa)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $13 = HEAP32[$12>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $14 = HEAP32[$11>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $15 = ((($14)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$15>>2] = $13; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $16 = $14; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $17 = HEAP32[$12>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$17>>2] = $16; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$4>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $18 = ($11|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if (!($18)) {
     $__f$01$i$i$i = $11;
     while(1) {
      $19 = ((($__f$01$i$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $20 = HEAP32[$19>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      __ZdlPv($__f$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $21 = ($20|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if ($21) {
       break;
      } else {
       $__f$01$i$i$i = $20;
      }
     }
    }
   }
   $$0 = $9;$$01 = $8;
   __ZdlPv($1); //@line 93 "modules/image/magpie/Image.cpp"
   ___resumeException($$01|0);
   // unreachable;
  }
 } while(0);
 __THREW__ = 0;
 invoke_viii(4163,($1|0),($0|0),($data|0)); //@line 93 "modules/image/magpie/Image.cpp"
 $38 = __THREW__; __THREW__ = 0;
 $39 = $38&1;
 if ($39) {
  $54 = ___cxa_find_matching_catch_2()|0;
  $55 = tempRet0;
  $56 = HEAP32[$4>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $57 = ($56|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($57) {
   $$0 = $55;$$01 = $54;
   __ZdlPv($1); //@line 93 "modules/image/magpie/Image.cpp"
   ___resumeException($$01|0);
   // unreachable;
  }
  $58 = HEAP32[$3>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $59 = HEAP32[$0>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $60 = ((($59)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $61 = HEAP32[$60>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $62 = HEAP32[$58>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $63 = ((($62)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$63>>2] = $61; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $64 = $62; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $65 = HEAP32[$60>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$65>>2] = $64; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$4>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $66 = ($58|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($66) {
   $$0 = $55;$$01 = $54;
   __ZdlPv($1); //@line 93 "modules/image/magpie/Image.cpp"
   ___resumeException($$01|0);
   // unreachable;
  } else {
   $__f$01$i$i2 = $58;
  }
  while(1) {
   $67 = ((($__f$01$i$i2)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $68 = HEAP32[$67>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   __ZdlPv($__f$01$i$i2); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $69 = ($68|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($69) {
    $$0 = $55;$$01 = $54;
    break;
   } else {
    $__f$01$i$i2 = $68;
   }
  }
  __ZdlPv($1); //@line 93 "modules/image/magpie/Image.cpp"
  ___resumeException($$01|0);
  // unreachable;
 } else {
  $40 = HEAP32[$4>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $41 = ($40|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($41) {
   STACKTOP = sp;return ($1|0); //@line 93 "modules/image/magpie/Image.cpp"
  }
  $42 = HEAP32[$3>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $43 = HEAP32[$0>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $44 = ((($43)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $45 = HEAP32[$44>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $46 = HEAP32[$42>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $47 = ((($46)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$47>>2] = $45; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $48 = $46; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $49 = HEAP32[$44>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$49>>2] = $48; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$4>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $50 = ($42|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($50) {
   STACKTOP = sp;return ($1|0); //@line 93 "modules/image/magpie/Image.cpp"
  } else {
   $__f$01$i$i = $42;
  }
  while(1) {
   $51 = ((($__f$01$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $52 = HEAP32[$51>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   __ZdlPv($__f$01$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $53 = ($52|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($53) {
    break;
   } else {
    $__f$01$i$i = $52;
   }
  }
  STACKTOP = sp;return ($1|0); //@line 93 "modules/image/magpie/Image.cpp"
 }
 return (0)|0;
}
function __ZN4love5image6magpie5Image12isCompressedEPNS_10filesystem8FileDataE($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$06 = 0, $__begin$sroa$0$07 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 98 "modules/image/magpie/Image.cpp"
 $1 = ((($this)) + 24|0); //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $__begin$sroa$0$06 = HEAP32[$1>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $2 = $__begin$sroa$0$06; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $3 = ($2|0)==($0|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($3) {
  $$12 = 0;
  return ($$12|0); //@line 105 "modules/image/magpie/Image.cpp"
 } else {
  $5 = $2;$__begin$sroa$0$07 = $__begin$sroa$0$06;
 }
 while(1) {
  $4 = ((($5)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $6 = HEAP32[$4>>2]|0; //@line 98 "modules/image/magpie/Image.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 100 "modules/image/magpie/Image.cpp"
  $8 = ((($7)) + 16|0); //@line 100 "modules/image/magpie/Image.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 100 "modules/image/magpie/Image.cpp"
  $10 = (FUNCTION_TABLE_iii[$9 & 8191]($6,$data)|0); //@line 100 "modules/image/magpie/Image.cpp"
  if ($10) {
   $$12 = 1;
   label = 4;
   break;
  }
  $11 = $__begin$sroa$0$07; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $12 = ((($11)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $__begin$sroa$0$0 = HEAP32[$12>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $13 = $__begin$sroa$0$0; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $14 = ($13|0)==($0|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($14) {
   $$12 = 0;
   label = 4;
   break;
  } else {
   $5 = $13;$__begin$sroa$0$07 = $__begin$sroa$0$0;
  }
 }
 if ((label|0) == 4) {
  return ($$12|0); //@line 105 "modules/image/magpie/Image.cpp"
 }
 return (0)|0;
}
function __ZN4love5image6magpie5ImageC2Ev($this) {
 $this = $this|0;
 var $$0 = 0, $$0$lcssa$i$i = 0, $$0$lcssa$i$i19 = 0, $$01 = 0, $$byval_copy = 0, $$byval_copy1 = 0, $$cast$i$i$i = 0, $$cast$i$i$i22 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__f$01$i$i = 0, $__f$01$i$i2 = 0, $__i$sroa$0$0$i$i = 0, $__i$sroa$0$0$i$i$1 = 0, $__i$sroa$0$0$i$i17 = 0, $__i$sroa$0$0$i$i17$1 = 0, $__i$sroa$0$0$i$i17$2 = 0;
 var $__i$sroa$0$0$i$i17$3 = 0, $__i$sroa$0$0$i$i17$4 = 0, $__i$sroa$0$0$lcssa$i$i = 0, $__i$sroa$0$0$lcssa$i$i18 = 0, $__i$sroa$0$06$i$i = 0, $__i$sroa$0$06$i$i11 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 44|0;
 $$byval_copy = sp + 40|0;
 $0 = sp + 36|0;
 $1 = sp + 32|0;
 $2 = sp + 24|0;
 $3 = sp;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./common/Module.h"
 HEAP32[$this>>2] = (25740); //@line 45 "modules/image/magpie/Image.cpp"
 $4 = ((($this)) + 8|0); //@line 44 "modules/image/magpie/Image.cpp"
 HEAP32[$4>>2] = $4; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $5 = ((($this)) + 12|0); //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$5>>2] = $4; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $6 = ((($this)) + 16|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$6>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $7 = ((($this)) + 20|0); //@line 44 "modules/image/magpie/Image.cpp"
 HEAP32[$7>>2] = $7; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $8 = ((($this)) + 24|0); //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$8>>2] = $7; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $9 = ((($this)) + 28|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$9>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 $10 = (invoke_ii(3348,8)|0); //@line 47 "modules/image/magpie/Image.cpp"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 do {
  if ($12) {
   label = 28;
  } else {
   __THREW__ = 0;
   invoke_vi(4164,($10|0)); //@line 37 "modules/image/magpie/PNGHandler.h"
   $13 = __THREW__; __THREW__ = 0;
   $14 = $13&1;
   if ($14) {
    $114 = ___cxa_find_matching_catch_2()|0;
    $115 = tempRet0;
    __ZdlPv($10); //@line 47 "modules/image/magpie/Image.cpp"
    $$0 = $115;$$01 = $114;
    break;
   }
   HEAP32[$10>>2] = (25892); //@line 37 "modules/image/magpie/PNGHandler.h"
   HEAP32[$2>>2] = $10; //@line 46 "modules/image/magpie/Image.cpp"
   $15 = ((($2)) + 4|0); //@line 46 "modules/image/magpie/Image.cpp"
   __THREW__ = 0;
   $16 = (invoke_ii(3348,8)|0); //@line 48 "modules/image/magpie/Image.cpp"
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if ($18) {
    label = 28;
   } else {
    __THREW__ = 0;
    invoke_vi(4164,($16|0)); //@line 40 "modules/image/magpie/STBHandler.h"
    $19 = __THREW__; __THREW__ = 0;
    $20 = $19&1;
    if ($20) {
     $116 = ___cxa_find_matching_catch_2()|0;
     $117 = tempRet0;
     __ZdlPv($16); //@line 48 "modules/image/magpie/Image.cpp"
     $$0 = $117;$$01 = $116;
     break;
    }
    HEAP32[$16>>2] = (25968); //@line 40 "modules/image/magpie/STBHandler.h"
    HEAP32[$15>>2] = $16; //@line 46 "modules/image/magpie/Image.cpp"
    $21 = ((($2)) + 8|0); //@line 93 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/initializer_list"
    $22 = $4; //@line 1329 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $__i$sroa$0$06$i$i = HEAP32[$5>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $23 = $__i$sroa$0$06$i$i; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $24 = ($23|0)==($4|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($24) {
     $$0$lcssa$i$i = $2;$__i$sroa$0$0$lcssa$i$i = $__i$sroa$0$06$i$i;
    } else {
     $25 = HEAP32[$2>>2]|0; //@line 1331 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $26 = ((($23)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$26>>2] = $25; //@line 1331 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $27 = ((($2)) + 4|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $28 = $__i$sroa$0$06$i$i; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $29 = ((($28)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $__i$sroa$0$0$i$i = HEAP32[$29>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $30 = $__i$sroa$0$0$i$i; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $31 = ($30|0)==($4|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     if ($31) {
      $$0$lcssa$i$i = $27;$__i$sroa$0$0$lcssa$i$i = $__i$sroa$0$0$i$i;
     } else {
      $181 = HEAP32[$27>>2]|0; //@line 1331 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $182 = ((($30)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      HEAP32[$182>>2] = $181; //@line 1331 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $183 = $__i$sroa$0$0$i$i; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $184 = ((($183)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $__i$sroa$0$0$i$i$1 = HEAP32[$184>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $$0$lcssa$i$i = $21;$__i$sroa$0$0$lcssa$i$i = $__i$sroa$0$0$i$i$1;
     }
    }
    $32 = $__i$sroa$0$0$lcssa$i$i; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $33 = ($32|0)==($4|0); //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($33) {
     HEAP32[$1>>2] = $22; //@line 414 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     __THREW__ = 0;
     ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0; //@line 1333 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     (invoke_iiiiii(4165,($4|0),($$byval_copy|0),($$0$lcssa$i$i|0),($21|0),(0|0))|0); //@line 1333 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $34 = __THREW__; __THREW__ = 0;
     $35 = $34&1;
     if ($35) {
      label = 28;
      break;
     }
    } else {
     $36 = HEAP32[$4>>2]|0; //@line 1760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $37 = ((($36)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $38 = HEAP32[$37>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $39 = $__i$sroa$0$0$lcssa$i$i; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $40 = HEAP32[$39>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $41 = ((($40)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$41>>2] = $38; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $42 = $40; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $43 = HEAP32[$37>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$43>>2] = $42; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $45 = $32;$49 = $32;
     while(1) {
      $44 = ((($45)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $46 = HEAP32[$44>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $47 = HEAP32[$6>>2]|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $48 = (($47) + -1)|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      HEAP32[$6>>2] = $48; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      __ZdlPv($49); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $$cast$i$i$i = $46; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $50 = ($$cast$i$i$i|0)==($4|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $51 = $46; //@line 1761 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if ($50) {
       break;
      } else {
       $45 = $51;$49 = $$cast$i$i$i;
      }
     }
    }
    __THREW__ = 0;
    $52 = (invoke_ii(3348,8)|0); //@line 52 "modules/image/magpie/Image.cpp"
    $53 = __THREW__; __THREW__ = 0;
    $54 = $53&1;
    if ($54) {
     label = 28;
    } else {
     __THREW__ = 0;
     invoke_vi(3431,($52|0)); //@line 44 "modules/image/magpie/CompressedFormatHandler.h"
     $55 = __THREW__; __THREW__ = 0;
     $56 = $55&1;
     if ($56) {
      $118 = ___cxa_find_matching_catch_2()|0;
      $119 = tempRet0;
      __ZdlPv($52); //@line 52 "modules/image/magpie/Image.cpp"
      $$0 = $119;$$01 = $118;
      break;
     }
     HEAP32[$52>>2] = (35216); //@line 43 "modules/image/magpie/ddsHandler.h"
     HEAP32[$3>>2] = $52; //@line 51 "modules/image/magpie/Image.cpp"
     $57 = ((($3)) + 4|0); //@line 51 "modules/image/magpie/Image.cpp"
     $58 = $52; //@line 1331 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     __THREW__ = 0;
     $59 = (invoke_ii(3348,8)|0); //@line 53 "modules/image/magpie/Image.cpp"
     $60 = __THREW__; __THREW__ = 0;
     $61 = $60&1;
     if ($61) {
      label = 28;
     } else {
      __THREW__ = 0;
      invoke_vi(3431,($59|0)); //@line 44 "modules/image/magpie/CompressedFormatHandler.h"
      $62 = __THREW__; __THREW__ = 0;
      $63 = $62&1;
      if ($63) {
       $120 = ___cxa_find_matching_catch_2()|0;
       $121 = tempRet0;
       __ZdlPv($59); //@line 53 "modules/image/magpie/Image.cpp"
       $$0 = $121;$$01 = $120;
       break;
      }
      HEAP32[$59>>2] = (25936); //@line 35 "modules/image/magpie/PVRHandler.h"
      HEAP32[$57>>2] = $59; //@line 51 "modules/image/magpie/Image.cpp"
      $64 = ((($3)) + 8|0); //@line 51 "modules/image/magpie/Image.cpp"
      __THREW__ = 0;
      $65 = (invoke_ii(3348,8)|0); //@line 54 "modules/image/magpie/Image.cpp"
      $66 = __THREW__; __THREW__ = 0;
      $67 = $66&1;
      if ($67) {
       label = 28;
      } else {
       __THREW__ = 0;
       invoke_vi(3431,($65|0)); //@line 44 "modules/image/magpie/CompressedFormatHandler.h"
       $68 = __THREW__; __THREW__ = 0;
       $69 = $68&1;
       if ($69) {
        $122 = ___cxa_find_matching_catch_2()|0;
        $123 = tempRet0;
        __ZdlPv($65); //@line 54 "modules/image/magpie/Image.cpp"
        $$0 = $123;$$01 = $122;
        break;
       }
       HEAP32[$65>>2] = (25828); //@line 37 "modules/image/magpie/KTXHandler.h"
       HEAP32[$64>>2] = $65; //@line 51 "modules/image/magpie/Image.cpp"
       $70 = ((($3)) + 12|0); //@line 51 "modules/image/magpie/Image.cpp"
       __THREW__ = 0;
       $71 = (invoke_ii(3348,8)|0); //@line 55 "modules/image/magpie/Image.cpp"
       $72 = __THREW__; __THREW__ = 0;
       $73 = $72&1;
       if ($73) {
        label = 28;
       } else {
        __THREW__ = 0;
        invoke_vi(3431,($71|0)); //@line 44 "modules/image/magpie/CompressedFormatHandler.h"
        $74 = __THREW__; __THREW__ = 0;
        $75 = $74&1;
        if ($75) {
         $124 = ___cxa_find_matching_catch_2()|0;
         $125 = tempRet0;
         __ZdlPv($71); //@line 55 "modules/image/magpie/Image.cpp"
         $$0 = $125;$$01 = $124;
         break;
        }
        HEAP32[$71>>2] = (25860); //@line 37 "modules/image/magpie/PKMHandler.h"
        HEAP32[$70>>2] = $71; //@line 51 "modules/image/magpie/Image.cpp"
        $76 = ((($3)) + 16|0); //@line 51 "modules/image/magpie/Image.cpp"
        __THREW__ = 0;
        $77 = (invoke_ii(3348,8)|0); //@line 56 "modules/image/magpie/Image.cpp"
        $78 = __THREW__; __THREW__ = 0;
        $79 = $78&1;
        if ($79) {
         label = 28;
        } else {
         __THREW__ = 0;
         invoke_vi(3431,($77|0)); //@line 44 "modules/image/magpie/CompressedFormatHandler.h"
         $80 = __THREW__; __THREW__ = 0;
         $81 = $80&1;
         if ($81) {
          $126 = ___cxa_find_matching_catch_2()|0;
          $127 = tempRet0;
          __ZdlPv($77); //@line 56 "modules/image/magpie/Image.cpp"
          $$0 = $127;$$01 = $126;
          break;
         }
         HEAP32[$77>>2] = (35152); //@line 38 "modules/image/magpie/ASTCHandler.h"
         HEAP32[$76>>2] = $77; //@line 51 "modules/image/magpie/Image.cpp"
         $82 = ((($3)) + 20|0); //@line 93 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/initializer_list"
         $83 = $7; //@line 1329 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
         $__i$sroa$0$06$i$i11 = HEAP32[$8>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
         $84 = $__i$sroa$0$06$i$i11; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
         $85 = ($84|0)==($7|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
         do {
          if ($85) {
           $$0$lcssa$i$i19 = $3;$__i$sroa$0$0$lcssa$i$i18 = $__i$sroa$0$06$i$i11;
          } else {
           $86 = ((($84)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
           HEAP32[$86>>2] = $58; //@line 1331 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
           $87 = ((($3)) + 4|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
           $88 = $__i$sroa$0$06$i$i11; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
           $89 = ((($88)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
           $__i$sroa$0$0$i$i17 = HEAP32[$89>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
           $90 = $__i$sroa$0$0$i$i17; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
           $91 = ($90|0)==($7|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
           if ($91) {
            $$0$lcssa$i$i19 = $87;$__i$sroa$0$0$lcssa$i$i18 = $__i$sroa$0$0$i$i17;
           } else {
            $156 = HEAP32[$87>>2]|0; //@line 1331 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $157 = ((($90)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            HEAP32[$157>>2] = $156; //@line 1331 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $158 = ((($3)) + 8|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $159 = $__i$sroa$0$0$i$i17; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $160 = ((($159)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $__i$sroa$0$0$i$i17$1 = HEAP32[$160>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $161 = $__i$sroa$0$0$i$i17$1; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $162 = ($161|0)==($7|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            if ($162) {
             $$0$lcssa$i$i19 = $158;$__i$sroa$0$0$lcssa$i$i18 = $__i$sroa$0$0$i$i17$1;
             break;
            }
            $163 = HEAP32[$158>>2]|0; //@line 1331 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $164 = ((($161)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            HEAP32[$164>>2] = $163; //@line 1331 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $165 = ((($3)) + 12|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $166 = $__i$sroa$0$0$i$i17$1; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $167 = ((($166)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $__i$sroa$0$0$i$i17$2 = HEAP32[$167>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $168 = $__i$sroa$0$0$i$i17$2; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $169 = ($168|0)==($7|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            if ($169) {
             $$0$lcssa$i$i19 = $165;$__i$sroa$0$0$lcssa$i$i18 = $__i$sroa$0$0$i$i17$2;
             break;
            }
            $170 = HEAP32[$165>>2]|0; //@line 1331 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $171 = ((($168)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            HEAP32[$171>>2] = $170; //@line 1331 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $172 = ((($3)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $173 = $__i$sroa$0$0$i$i17$2; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $174 = ((($173)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $__i$sroa$0$0$i$i17$3 = HEAP32[$174>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $175 = $__i$sroa$0$0$i$i17$3; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $176 = ($175|0)==($7|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            if ($176) {
             $$0$lcssa$i$i19 = $172;$__i$sroa$0$0$lcssa$i$i18 = $__i$sroa$0$0$i$i17$3;
             break;
            }
            $177 = HEAP32[$172>>2]|0; //@line 1331 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $178 = ((($175)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            HEAP32[$178>>2] = $177; //@line 1331 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $179 = $__i$sroa$0$0$i$i17$3; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $180 = ((($179)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $__i$sroa$0$0$i$i17$4 = HEAP32[$180>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
            $$0$lcssa$i$i19 = $82;$__i$sroa$0$0$lcssa$i$i18 = $__i$sroa$0$0$i$i17$4;
           }
          }
         } while(0);
         $92 = $__i$sroa$0$0$lcssa$i$i18; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
         $93 = ($92|0)==($7|0); //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
         if ($93) {
          HEAP32[$0>>2] = $83; //@line 414 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
          __THREW__ = 0;
          ;HEAP32[$$byval_copy1>>2]=HEAP32[$0>>2]|0; //@line 1333 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
          (invoke_iiiiii(4166,($7|0),($$byval_copy1|0),($$0$lcssa$i$i19|0),($82|0),(0|0))|0); //@line 1333 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
          $94 = __THREW__; __THREW__ = 0;
          $95 = $94&1;
          if ($95) {
           label = 28;
           break;
          }
          STACKTOP = sp;return; //@line 58 "modules/image/magpie/Image.cpp"
         }
         $96 = HEAP32[$7>>2]|0; //@line 1760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
         $97 = ((($96)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
         $98 = HEAP32[$97>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
         $99 = $__i$sroa$0$0$lcssa$i$i18; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
         $100 = HEAP32[$99>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
         $101 = ((($100)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
         HEAP32[$101>>2] = $98; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
         $102 = $100; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
         $103 = HEAP32[$97>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
         HEAP32[$103>>2] = $102; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
         $105 = $92;$109 = $92;
         while(1) {
          $104 = ((($105)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
          $106 = HEAP32[$104>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
          $107 = HEAP32[$9>>2]|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
          $108 = (($107) + -1)|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
          HEAP32[$9>>2] = $108; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
          __ZdlPv($109); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
          $$cast$i$i$i22 = $106; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
          $110 = ($$cast$i$i$i22|0)==($7|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
          $111 = $106; //@line 1761 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
          if ($110) {
           break;
          } else {
           $105 = $111;$109 = $$cast$i$i$i22;
          }
         }
         STACKTOP = sp;return; //@line 58 "modules/image/magpie/Image.cpp"
        }
       }
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 28) {
  $112 = ___cxa_find_matching_catch_2()|0;
  $113 = tempRet0;
  $$0 = $113;$$01 = $112;
 }
 $128 = HEAP32[$9>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $129 = ($128|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if (!($129)) {
  $130 = HEAP32[$8>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $131 = HEAP32[$7>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $132 = ((($131)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $133 = HEAP32[$132>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $134 = HEAP32[$130>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $135 = ((($134)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$135>>2] = $133; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $136 = $134; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $137 = HEAP32[$132>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$137>>2] = $136; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$9>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $138 = ($130|0)==($7|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if (!($138)) {
   $__f$01$i$i2 = $130;
   while(1) {
    $139 = ((($__f$01$i$i2)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $140 = HEAP32[$139>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    __ZdlPv($__f$01$i$i2); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $141 = ($140|0)==($7|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($141) {
     break;
    } else {
     $__f$01$i$i2 = $140;
    }
   }
  }
 }
 $142 = HEAP32[$6>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $143 = ($142|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($143) {
  __ZN4love6ModuleD2Ev($this); //@line 58 "modules/image/magpie/Image.cpp"
  ___resumeException($$01|0);
  // unreachable;
 }
 $144 = HEAP32[$5>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $145 = HEAP32[$4>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $146 = ((($145)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $147 = HEAP32[$146>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $148 = HEAP32[$144>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $149 = ((($148)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$149>>2] = $147; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $150 = $148; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $151 = HEAP32[$146>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$151>>2] = $150; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$6>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $152 = ($144|0)==($4|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($152) {
  __ZN4love6ModuleD2Ev($this); //@line 58 "modules/image/magpie/Image.cpp"
  ___resumeException($$01|0);
  // unreachable;
 } else {
  $__f$01$i$i = $144;
 }
 while(1) {
  $153 = ((($__f$01$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $154 = HEAP32[$153>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  __ZdlPv($__f$01$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $155 = ($154|0)==($4|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($155) {
   break;
  } else {
   $__f$01$i$i = $154;
  }
 }
 __ZN4love6ModuleD2Ev($this); //@line 58 "modules/image/magpie/Image.cpp"
 ___resumeException($$01|0);
 // unreachable;
}
function __ZNSt3__14listIPN4love5image6magpie13FormatHandlerENS_9allocatorIS5_EEE6insertIPKS5_EENS_15__list_iteratorIS5_PvEENS_21__list_const_iteratorIS5_SD_EET_SH_PNS_9enable_ifIXsr19__is_input_iteratorISH_EE5valueEvE4typeE($this,$__p,$__f,$__l,$0) {
 $this = $this|0;
 $__p = $__p|0;
 $__f = $__f|0;
 $__l = $__l|0;
 $0 = $0|0;
 var $$01 = 0, $$0110 = 0, $$0113 = 0, $$sroa$07$0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__ds$0$lcssa = 0, $__ds$012 = 0, $__e$sroa$0$0$lcssa = 0, $__e$sroa$0$011 = 0, $__e$sroa$0$011$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$__p>>2]|0; //@line 1461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $2 = ($__f|0)==($__l|0); //@line 1463 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $3 = $1; //@line 1511 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($2) {
  $$sroa$07$0 = $1;
  $42 = $$sroa$07$0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  return ($42|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 }
 $4 = (__Znwj(12)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 HEAP32[$4>>2] = 0; //@line 1469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $5 = ((($4)) + 8|0); //@line 1470 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $6 = HEAP32[$__f>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$5>>2] = $6; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $7 = $4; //@line 254 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $$0110 = ((($__f)) + 4|0); //@line 1483 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $8 = ($$0110|0)==($__l|0); //@line 1483 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 L4: do {
  if ($8) {
   $__ds$0$lcssa = 1;$__e$sroa$0$0$lcssa = $7;
  } else {
   $$0113 = $$0110;$__ds$012 = 1;$__e$sroa$0$011 = $7;
   while(1) {
    __THREW__ = 0;
    $9 = (invoke_ii(3348,12)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $10 = __THREW__; __THREW__ = 0;
    $11 = $10&1;
    if ($11) {
     $__e$sroa$0$011$lcssa = $__e$sroa$0$011;
     break;
    }
    $12 = ((($9)) + 8|0); //@line 1486 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $13 = HEAP32[$$0113>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$12>>2] = $13; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $14 = $__e$sroa$0$011; //@line 1487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $15 = ((($14)) + 4|0); //@line 1487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$15>>2] = $9; //@line 1487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$9>>2] = $__e$sroa$0$011; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $16 = $9; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $17 = (($__ds$012) + 1)|0; //@line 1483 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $$01 = ((($$0113)) + 4|0); //@line 1483 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $18 = ($$01|0)==($__l|0); //@line 1483 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($18) {
     $__ds$0$lcssa = $17;$__e$sroa$0$0$lcssa = $16;
     break L4;
    } else {
     $$0113 = $$01;$__ds$012 = $17;$__e$sroa$0$011 = $16;
    }
   }
   $19 = ___cxa_find_matching_catch_3(0|0)|0;
   $20 = tempRet0;
   (___cxa_begin_catch(($19|0))|0); //@line 1492 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $21 = $__e$sroa$0$011$lcssa; //@line 1498 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $22 = HEAP32[$21>>2]|0; //@line 1498 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $23 = $__e$sroa$0$011$lcssa; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __ZdlPv($23); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $24 = ($22|0)==(0|0); //@line 1500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if (!($24)) {
    $26 = $22;
    while(1) {
     $25 = HEAP32[$26>>2]|0; //@line 1498 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     __ZdlPv($26); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $27 = ($25|0)==(0|0); //@line 1500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     if ($27) {
      break;
     } else {
      $26 = $25;
     }
    }
   }
   __THREW__ = 0;
   invoke_v(3656); //@line 1508 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $28 = __THREW__; __THREW__ = 0;
   $29 = ___cxa_find_matching_catch_2()|0;
   $30 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 1509 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $31 = __THREW__; __THREW__ = 0;
   $32 = $31&1;
   if ($32) {
    $43 = ___cxa_find_matching_catch_3(0|0)|0;
    $44 = tempRet0;
    ___clang_call_terminate($43); //@line 1509 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    // unreachable; //@line 1509 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   } else {
    ___resumeException($29|0);
    // unreachable;
   }
  }
 } while(0);
 $33 = $__e$sroa$0$0$lcssa; //@line 1511 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $34 = $1; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $35 = HEAP32[$34>>2]|0; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $36 = ((($35)) + 4|0); //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$36>>2] = $4; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $37 = $35; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$4>>2] = $37; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$3>>2] = $33; //@line 1075 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $38 = ((($33)) + 4|0); //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$38>>2] = $3; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $39 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $40 = HEAP32[$39>>2]|0; //@line 1512 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $41 = (($40) + ($__ds$0$lcssa))|0; //@line 1512 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$39>>2] = $41; //@line 1512 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $$sroa$07$0 = $7;
 $42 = $$sroa$07$0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 return ($42|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
}
function __ZNSt3__14listIPN4love5image6magpie23CompressedFormatHandlerENS_9allocatorIS5_EEE6insertIPKS5_EENS_15__list_iteratorIS5_PvEENS_21__list_const_iteratorIS5_SD_EET_SH_PNS_9enable_ifIXsr19__is_input_iteratorISH_EE5valueEvE4typeE($this,$__p,$__f,$__l,$0) {
 $this = $this|0;
 $__p = $__p|0;
 $__f = $__f|0;
 $__l = $__l|0;
 $0 = $0|0;
 var $$01 = 0, $$0110 = 0, $$0113 = 0, $$sroa$07$0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__ds$0$lcssa = 0, $__ds$012 = 0, $__e$sroa$0$0$lcssa = 0, $__e$sroa$0$011 = 0, $__e$sroa$0$011$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$__p>>2]|0; //@line 1461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $2 = ($__f|0)==($__l|0); //@line 1463 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $3 = $1; //@line 1511 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($2) {
  $$sroa$07$0 = $1;
  $42 = $$sroa$07$0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  return ($42|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 }
 $4 = (__Znwj(12)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 HEAP32[$4>>2] = 0; //@line 1469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $5 = ((($4)) + 8|0); //@line 1470 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $6 = HEAP32[$__f>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$5>>2] = $6; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $7 = $4; //@line 254 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $$0110 = ((($__f)) + 4|0); //@line 1483 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $8 = ($$0110|0)==($__l|0); //@line 1483 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 L4: do {
  if ($8) {
   $__ds$0$lcssa = 1;$__e$sroa$0$0$lcssa = $7;
  } else {
   $$0113 = $$0110;$__ds$012 = 1;$__e$sroa$0$011 = $7;
   while(1) {
    __THREW__ = 0;
    $9 = (invoke_ii(3348,12)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $10 = __THREW__; __THREW__ = 0;
    $11 = $10&1;
    if ($11) {
     $__e$sroa$0$011$lcssa = $__e$sroa$0$011;
     break;
    }
    $12 = ((($9)) + 8|0); //@line 1486 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $13 = HEAP32[$$0113>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$12>>2] = $13; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $14 = $__e$sroa$0$011; //@line 1487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $15 = ((($14)) + 4|0); //@line 1487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$15>>2] = $9; //@line 1487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$9>>2] = $__e$sroa$0$011; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $16 = $9; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $17 = (($__ds$012) + 1)|0; //@line 1483 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $$01 = ((($$0113)) + 4|0); //@line 1483 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $18 = ($$01|0)==($__l|0); //@line 1483 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($18) {
     $__ds$0$lcssa = $17;$__e$sroa$0$0$lcssa = $16;
     break L4;
    } else {
     $$0113 = $$01;$__ds$012 = $17;$__e$sroa$0$011 = $16;
    }
   }
   $19 = ___cxa_find_matching_catch_3(0|0)|0;
   $20 = tempRet0;
   (___cxa_begin_catch(($19|0))|0); //@line 1492 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $21 = $__e$sroa$0$011$lcssa; //@line 1498 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $22 = HEAP32[$21>>2]|0; //@line 1498 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $23 = $__e$sroa$0$011$lcssa; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __ZdlPv($23); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $24 = ($22|0)==(0|0); //@line 1500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if (!($24)) {
    $26 = $22;
    while(1) {
     $25 = HEAP32[$26>>2]|0; //@line 1498 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     __ZdlPv($26); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $27 = ($25|0)==(0|0); //@line 1500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     if ($27) {
      break;
     } else {
      $26 = $25;
     }
    }
   }
   __THREW__ = 0;
   invoke_v(3656); //@line 1508 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $28 = __THREW__; __THREW__ = 0;
   $29 = ___cxa_find_matching_catch_2()|0;
   $30 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378); //@line 1509 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $31 = __THREW__; __THREW__ = 0;
   $32 = $31&1;
   if ($32) {
    $43 = ___cxa_find_matching_catch_3(0|0)|0;
    $44 = tempRet0;
    ___clang_call_terminate($43); //@line 1509 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    // unreachable; //@line 1509 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   } else {
    ___resumeException($29|0);
    // unreachable;
   }
  }
 } while(0);
 $33 = $__e$sroa$0$0$lcssa; //@line 1511 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $34 = $1; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $35 = HEAP32[$34>>2]|0; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $36 = ((($35)) + 4|0); //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$36>>2] = $4; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $37 = $35; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$4>>2] = $37; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$3>>2] = $33; //@line 1075 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $38 = ((($33)) + 4|0); //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$38>>2] = $3; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $39 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $40 = HEAP32[$39>>2]|0; //@line 1512 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $41 = (($40) + ($__ds$0$lcssa))|0; //@line 1512 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$39>>2] = $41; //@line 1512 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $$sroa$07$0 = $7;
 $42 = $$sroa$07$0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 return ($42|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
}
function __ZN4love5image6magpie9ImageDataD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$05 = 0, $__begin$sroa$0$06 = 0, $__f$01$i$i = 0, $__f$01$i$i1 = 0, $lpad$phi$index = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (25792); //@line 74 "modules/image/magpie/ImageData.cpp"
 $0 = ((($this)) + 36|0); //@line 75 "modules/image/magpie/ImageData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 75 "modules/image/magpie/ImageData.cpp"
 $2 = ($1|0)==(0|0); //@line 75 "modules/image/magpie/ImageData.cpp"
 if ($2) {
  $31 = ((($this)) + 16|0); //@line 78 "modules/image/magpie/ImageData.cpp"
  $32 = HEAP32[$31>>2]|0; //@line 78 "modules/image/magpie/ImageData.cpp"
  $33 = ($32|0)==(0|0); //@line 78 "modules/image/magpie/ImageData.cpp"
  if ($33) {
   label = 11;
  } else {
   __ZdaPv($32); //@line 78 "modules/image/magpie/ImageData.cpp"
   label = 11;
  }
 } else {
  $3 = HEAP32[$1>>2]|0; //@line 76 "modules/image/magpie/ImageData.cpp"
  $4 = ((($3)) + 32|0); //@line 76 "modules/image/magpie/ImageData.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 76 "modules/image/magpie/ImageData.cpp"
  $6 = ((($this)) + 16|0); //@line 76 "modules/image/magpie/ImageData.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 76 "modules/image/magpie/ImageData.cpp"
  __THREW__ = 0;
  invoke_vii($5|0,($1|0),($7|0)); //@line 76 "modules/image/magpie/ImageData.cpp"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if ($9) {
   $12 = ___cxa_find_matching_catch_3(0|0)|0;
   $13 = tempRet0;
   $lpad$phi$index = $12;
  } else {
   label = 11;
  }
 }
 L6: do {
  if ((label|0) == 11) {
   $34 = ((($this)) + 24|0); //@line 80 "modules/image/magpie/ImageData.cpp"
   $35 = ((($this)) + 28|0); //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $__begin$sroa$0$05 = HEAP32[$35>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $36 = $__begin$sroa$0$05; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $37 = ($36|0)==($34|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   L8: do {
    if (!($37)) {
     $54 = $36;$__begin$sroa$0$06 = $__begin$sroa$0$05;
     while(1) {
      $53 = ((($54)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $55 = HEAP32[$53>>2]|0; //@line 80 "modules/image/magpie/ImageData.cpp"
      $56 = HEAP32[$55>>2]|0; //@line 81 "modules/image/magpie/ImageData.cpp"
      $57 = ((($56)) + 12|0); //@line 81 "modules/image/magpie/ImageData.cpp"
      $58 = HEAP32[$57>>2]|0; //@line 81 "modules/image/magpie/ImageData.cpp"
      __THREW__ = 0;
      invoke_vi($58|0,($55|0)); //@line 81 "modules/image/magpie/ImageData.cpp"
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if ($60) {
       break;
      }
      $61 = $__begin$sroa$0$06; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $62 = ((($61)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $__begin$sroa$0$0 = HEAP32[$62>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $63 = $__begin$sroa$0$0; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $64 = ($63|0)==($34|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if ($64) {
       break L8;
      } else {
       $54 = $63;$__begin$sroa$0$06 = $__begin$sroa$0$0;
      }
     }
     $10 = ___cxa_find_matching_catch_3(0|0)|0;
     $11 = tempRet0;
     $lpad$phi$index = $10;
     break L6;
    }
   } while(0);
   $38 = ((($this)) + 32|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $39 = HEAP32[$38>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $40 = ($39|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($40) {
    __ZN4love5image9ImageDataD2Ev($this); //@line 82 "modules/image/magpie/ImageData.cpp"
    return; //@line 82 "modules/image/magpie/ImageData.cpp"
   }
   $41 = HEAP32[$35>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $42 = HEAP32[$34>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $43 = ((($42)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $44 = HEAP32[$43>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $45 = HEAP32[$41>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $46 = ((($45)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   HEAP32[$46>>2] = $44; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $47 = $45; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $48 = HEAP32[$43>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   HEAP32[$48>>2] = $47; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   HEAP32[$38>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $49 = ($41|0)==($34|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($49) {
    __ZN4love5image9ImageDataD2Ev($this); //@line 82 "modules/image/magpie/ImageData.cpp"
    return; //@line 82 "modules/image/magpie/ImageData.cpp"
   } else {
    $__f$01$i$i1 = $41;
   }
   while(1) {
    $50 = ((($__f$01$i$i1)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $51 = HEAP32[$50>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    __ZdlPv($__f$01$i$i1); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $52 = ($51|0)==($34|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($52) {
     break;
    } else {
     $__f$01$i$i1 = $51;
    }
   }
   __ZN4love5image9ImageDataD2Ev($this); //@line 82 "modules/image/magpie/ImageData.cpp"
   return; //@line 82 "modules/image/magpie/ImageData.cpp"
  }
 } while(0);
 $14 = ((($this)) + 24|0); //@line 82 "modules/image/magpie/ImageData.cpp"
 $15 = ((($this)) + 32|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $16 = HEAP32[$15>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $17 = ($16|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($17) {
  __ZN4love5image9ImageDataD2Ev($this); //@line 82 "modules/image/magpie/ImageData.cpp"
  ___clang_call_terminate($lpad$phi$index); //@line 82 "modules/image/magpie/ImageData.cpp"
  // unreachable; //@line 82 "modules/image/magpie/ImageData.cpp"
 }
 $18 = ((($this)) + 28|0); //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $19 = HEAP32[$18>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $20 = HEAP32[$14>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $21 = ((($20)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $22 = HEAP32[$21>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $23 = HEAP32[$19>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $24 = ((($23)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$24>>2] = $22; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $25 = $23; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $26 = HEAP32[$21>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$26>>2] = $25; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$15>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $27 = ($19|0)==($14|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($27) {
  __ZN4love5image9ImageDataD2Ev($this); //@line 82 "modules/image/magpie/ImageData.cpp"
  ___clang_call_terminate($lpad$phi$index); //@line 82 "modules/image/magpie/ImageData.cpp"
  // unreachable; //@line 82 "modules/image/magpie/ImageData.cpp"
 } else {
  $__f$01$i$i = $19;
 }
 while(1) {
  $28 = ((($__f$01$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $29 = HEAP32[$28>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  __ZdlPv($__f$01$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $30 = ($29|0)==($14|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($30) {
   break;
  } else {
   $__f$01$i$i = $29;
  }
 }
 __ZN4love5image9ImageDataD2Ev($this); //@line 82 "modules/image/magpie/ImageData.cpp"
 ___clang_call_terminate($lpad$phi$index); //@line 82 "modules/image/magpie/ImageData.cpp"
 // unreachable; //@line 82 "modules/image/magpie/ImageData.cpp"
}
function __ZN4love5image6magpie9ImageDataD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5image6magpie9ImageDataD2Ev($this); //@line 74 "modules/image/magpie/ImageData.cpp"
 __ZdlPv($this); //@line 74 "modules/image/magpie/ImageData.cpp"
 return; //@line 82 "modules/image/magpie/ImageData.cpp"
}
function __ZN4love5image6magpie9ImageData6encodeENS0_9ImageData13EncodedFormatEPKc($this,$format,$filename) {
 $this = $this|0;
 $format = $format|0;
 $filename = $filename|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $9 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$07 = 0, $__begin$sroa$0$09 = 0, $cond = 0, $encoder$2 = 0, $fname = 0, $lock = 0, $rawimage = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp + 8|0;
 $rawimage = sp + 32|0;
 $lock = sp + 28|0;
 $0 = sp;
 $fname = sp + 24|0;
 $1 = sp + 12|0;
 ;HEAP32[$rawimage>>2]=0|0;HEAP32[$rawimage+4>>2]=0|0;HEAP32[$rawimage+8>>2]=0|0; //@line 48 "modules/image/magpie/FormatHandler.h"
 $2 = ((($this)) + 8|0); //@line 153 "modules/image/magpie/ImageData.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 153 "modules/image/magpie/ImageData.cpp"
 HEAP32[$rawimage>>2] = $3; //@line 153 "modules/image/magpie/ImageData.cpp"
 $4 = ((($this)) + 12|0); //@line 154 "modules/image/magpie/ImageData.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 154 "modules/image/magpie/ImageData.cpp"
 $6 = ((($rawimage)) + 4|0); //@line 154 "modules/image/magpie/ImageData.cpp"
 HEAP32[$6>>2] = $5; //@line 154 "modules/image/magpie/ImageData.cpp"
 $7 = $5 << 2; //@line 155 "modules/image/magpie/ImageData.cpp"
 $8 = Math_imul($7, $3)|0; //@line 155 "modules/image/magpie/ImageData.cpp"
 $9 = ((($rawimage)) + 8|0); //@line 155 "modules/image/magpie/ImageData.cpp"
 HEAP32[$9>>2] = $8; //@line 155 "modules/image/magpie/ImageData.cpp"
 $10 = ((($this)) + 16|0); //@line 156 "modules/image/magpie/ImageData.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 156 "modules/image/magpie/ImageData.cpp"
 $12 = ((($rawimage)) + 12|0); //@line 156 "modules/image/magpie/ImageData.cpp"
 HEAP32[$12>>2] = $11; //@line 156 "modules/image/magpie/ImageData.cpp"
 $13 = ((($this)) + 24|0); //@line 158 "modules/image/magpie/ImageData.cpp"
 $14 = ((($this)) + 28|0); //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $__begin$sroa$0$07 = HEAP32[$14>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $15 = $__begin$sroa$0$07; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $16 = ($15|0)==($13|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if (!($16)) {
  $18 = $15;$__begin$sroa$0$09 = $__begin$sroa$0$07;
  while(1) {
   $17 = ((($18)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $19 = HEAP32[$17>>2]|0; //@line 158 "modules/image/magpie/ImageData.cpp"
   $20 = HEAP32[$19>>2]|0; //@line 160 "modules/image/magpie/ImageData.cpp"
   $21 = ((($20)) + 20|0); //@line 160 "modules/image/magpie/ImageData.cpp"
   $22 = HEAP32[$21>>2]|0; //@line 160 "modules/image/magpie/ImageData.cpp"
   $23 = (FUNCTION_TABLE_iii[$22 & 8191]($19,$format)|0); //@line 160 "modules/image/magpie/ImageData.cpp"
   if ($23) {
    $encoder$2 = $19;
    break;
   }
   $24 = $__begin$sroa$0$09; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $25 = ((($24)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $__begin$sroa$0$0 = HEAP32[$25>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $26 = $__begin$sroa$0$0; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $27 = ($26|0)==($13|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($27) {
    $encoder$2 = 0;
    break;
   } else {
    $18 = $26;$__begin$sroa$0$09 = $__begin$sroa$0$0;
   }
  }
  $cond = ($encoder$2|0)==(0|0); //@line 167 "modules/image/magpie/ImageData.cpp"
  if (!($cond)) {
   $28 = ((($this)) + 20|0); //@line 169 "modules/image/magpie/ImageData.cpp"
   $29 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($28)|0); //@line 169 "modules/image/magpie/ImageData.cpp"
   __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$29); //@line 169 "modules/image/magpie/ImageData.cpp"
   $30 = HEAP32[$encoder$2>>2]|0; //@line 170 "modules/image/magpie/ImageData.cpp"
   $31 = ((($30)) + 28|0); //@line 170 "modules/image/magpie/ImageData.cpp"
   $32 = HEAP32[$31>>2]|0; //@line 170 "modules/image/magpie/ImageData.cpp"
   __THREW__ = 0;
   invoke_viiii($32|0,($0|0),($encoder$2|0),($rawimage|0),($format|0)); //@line 170 "modules/image/magpie/ImageData.cpp"
   $33 = __THREW__; __THREW__ = 0;
   $34 = $33&1;
   if ($34) {
    $35 = ___cxa_find_matching_catch_2()|0;
    $36 = tempRet0;
    __ZN4love6thread4LockD2Ev($lock); //@line 171 "modules/image/magpie/ImageData.cpp"
    $$2 = $36;$$23 = $35;
    ___resumeException($$23|0);
    // unreachable;
   }
   $37 = HEAP32[$0>>2]|0; //@line 170 "modules/image/magpie/ImageData.cpp"
   $38 = ((($0)) + 4|0); //@line 170 "modules/image/magpie/ImageData.cpp"
   $39 = HEAP32[$38>>2]|0; //@line 170 "modules/image/magpie/ImageData.cpp"
   __ZN4love6thread4LockD2Ev($lock); //@line 171 "modules/image/magpie/ImageData.cpp"
   $40 = ($39|0)==(0|0); //@line 173 "modules/image/magpie/ImageData.cpp"
   if (!($40)) {
    __THREW__ = 0;
    $47 = (invoke_ii(3348,48)|0); //@line 184 "modules/image/magpie/ImageData.cpp"
    $48 = __THREW__; __THREW__ = 0;
    $49 = $48&1;
    if ($49) {
     $62 = ___cxa_find_matching_catch_3(96|0)|0;
     $63 = tempRet0;
     $$1 = $63;$$12 = $62;
    } else {
     $50 = (_strlen($filename)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     __THREW__ = 0;
     invoke_viii(3460,($1|0),($filename|0),($50|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $51 = __THREW__; __THREW__ = 0;
     $52 = $51&1;
     do {
      if ($52) {
       $64 = ___cxa_find_matching_catch_3(96|0)|0;
       $65 = tempRet0;
       $$0 = $65;$$01 = $64;
      } else {
       __THREW__ = 0;
       invoke_viiii(3972,($47|0),($37|0),0,($1|0)); //@line 184 "modules/image/magpie/ImageData.cpp"
       $53 = __THREW__; __THREW__ = 0;
       $54 = $53&1;
       if ($54) {
        $66 = ___cxa_find_matching_catch_3(96|0)|0;
        $67 = tempRet0;
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 184 "modules/image/magpie/ImageData.cpp"
        $$0 = $67;$$01 = $66;
        break;
       } else {
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 184 "modules/image/magpie/ImageData.cpp"
        $55 = HEAP32[$47>>2]|0; //@line 192 "modules/image/magpie/ImageData.cpp"
        $56 = ((($55)) + 16|0); //@line 192 "modules/image/magpie/ImageData.cpp"
        $57 = HEAP32[$56>>2]|0; //@line 192 "modules/image/magpie/ImageData.cpp"
        $58 = (FUNCTION_TABLE_ii[$57 & 8191]($47)|0); //@line 192 "modules/image/magpie/ImageData.cpp"
        _memcpy(($58|0),($39|0),($37|0))|0; //@line 192 "modules/image/magpie/ImageData.cpp"
        $59 = HEAP32[$encoder$2>>2]|0; //@line 193 "modules/image/magpie/ImageData.cpp"
        $60 = ((($59)) + 32|0); //@line 193 "modules/image/magpie/ImageData.cpp"
        $61 = HEAP32[$60>>2]|0; //@line 193 "modules/image/magpie/ImageData.cpp"
        FUNCTION_TABLE_vii[$61 & 8191]($encoder$2,$39); //@line 193 "modules/image/magpie/ImageData.cpp"
        STACKTOP = sp;return ($47|0); //@line 195 "modules/image/magpie/ImageData.cpp"
       }
      }
     } while(0);
     __ZdlPv($47); //@line 184 "modules/image/magpie/ImageData.cpp"
     $$1 = $$0;$$12 = $$01;
    }
    $68 = (_llvm_eh_typeid_for((96|0))|0); //@line 185 "modules/image/magpie/ImageData.cpp"
    $69 = ($$1|0)==($68|0); //@line 185 "modules/image/magpie/ImageData.cpp"
    if (!($69)) {
     $$2 = $$1;$$23 = $$12;
     ___resumeException($$23|0);
     // unreachable;
    }
    (___cxa_begin_catch(($$12|0))|0); //@line 185 "modules/image/magpie/ImageData.cpp"
    $70 = HEAP32[$encoder$2>>2]|0; //@line 188 "modules/image/magpie/ImageData.cpp"
    $71 = ((($70)) + 32|0); //@line 188 "modules/image/magpie/ImageData.cpp"
    $72 = HEAP32[$71>>2]|0; //@line 188 "modules/image/magpie/ImageData.cpp"
    __THREW__ = 0;
    invoke_vii($72|0,($encoder$2|0),($39|0)); //@line 188 "modules/image/magpie/ImageData.cpp"
    $73 = __THREW__; __THREW__ = 0;
    $74 = $73&1;
    if (!($74)) {
     __THREW__ = 0;
     invoke_v(3656); //@line 189 "modules/image/magpie/ImageData.cpp"
     $75 = __THREW__; __THREW__ = 0;
    }
    $76 = ___cxa_find_matching_catch_2()|0;
    $77 = tempRet0;
    __THREW__ = 0;
    invoke_v(3378); //@line 190 "modules/image/magpie/ImageData.cpp"
    $78 = __THREW__; __THREW__ = 0;
    $79 = $78&1;
    if ($79) {
     $80 = ___cxa_find_matching_catch_3(0|0)|0;
     $81 = tempRet0;
     ___clang_call_terminate($80); //@line 190 "modules/image/magpie/ImageData.cpp"
     // unreachable; //@line 190 "modules/image/magpie/ImageData.cpp"
    } else {
     $$2 = $77;$$23 = $76;
     ___resumeException($$23|0);
     // unreachable;
    }
   }
  }
 }
 HEAP32[$fname>>2] = 442424; //@line 175 "modules/image/magpie/ImageData.cpp"
 (__ZN4love5image9ImageData11getConstantENS1_13EncodedFormatERPKc($format,$fname)|0); //@line 176 "modules/image/magpie/ImageData.cpp"
 $41 = (___cxa_allocate_exception(16)|0); //@line 177 "modules/image/magpie/ImageData.cpp"
 $42 = HEAP32[$fname>>2]|0; //@line 177 "modules/image/magpie/ImageData.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $42; //@line 177 "modules/image/magpie/ImageData.cpp"
 invoke_viii(3349,($41|0),(439652|0),($vararg_buffer|0)); //@line 177 "modules/image/magpie/ImageData.cpp"
 $43 = __THREW__; __THREW__ = 0;
 $44 = $43&1;
 if (!($44)) {
  ___cxa_throw(($41|0),(96|0),(344|0)); //@line 177 "modules/image/magpie/ImageData.cpp"
  // unreachable; //@line 177 "modules/image/magpie/ImageData.cpp"
 }
 $45 = ___cxa_find_matching_catch_2()|0;
 $46 = tempRet0;
 ___cxa_free_exception(($41|0)); //@line 177 "modules/image/magpie/ImageData.cpp"
 $$2 = $46;$$23 = $45;
 ___resumeException($$23|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love5image6magpie9ImageDataC2ENSt3__14listIPNS1_13FormatHandlerENS3_9allocatorIS6_EEEEPNS_10filesystem8FileDataE($this,$formats,$data) {
 $this = $this|0;
 $formats = $formats|0;
 $data = $data|0;
 var $$0 = 0, $$01 = 0, $$lcssa = 0, $$lcssa23 = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$06 = 0, $__begin$sroa$0$06$pre = 0;
 var $__begin$sroa$0$07 = 0, $__f$01$i$i = 0, $__f$01$i$i$i = 0, $__i$sroa$0$0$i = 0, $__i$sroa$0$02$i = 0, $__i$sroa$0$03$i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5image9ImageDataC2Ev($this); //@line 31 "modules/image/magpie/ImageData.cpp"
 HEAP32[$this>>2] = (25792); //@line 34 "modules/image/magpie/ImageData.cpp"
 $0 = ((($this)) + 24|0); //@line 32 "modules/image/magpie/ImageData.cpp"
 HEAP32[$0>>2] = $0; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $1 = ((($this)) + 28|0); //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$1>>2] = $0; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $2 = ((($this)) + 32|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$2>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = ((($formats)) + 4|0); //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $__i$sroa$0$02$i = HEAP32[$3>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $4 = $__i$sroa$0$02$i; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $5 = ($4|0)==($formats|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 do {
  if ($5) {
   $6 = $0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $$pre$phiZ2D = $0;$__begin$sroa$0$06 = $6;
  } else {
   $25 = $4;$29 = $0;$32 = 0;$__i$sroa$0$03$i = $__i$sroa$0$02$i;
   while(1) {
    __THREW__ = 0;
    $21 = (invoke_ii(3348,12)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $22 = __THREW__; __THREW__ = 0;
    $23 = $22&1;
    if ($23) {
     $$lcssa = $29;$$lcssa23 = $32;
     break;
    }
    $24 = ((($25)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $26 = ((($21)) + 8|0); //@line 1537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $27 = HEAP32[$24>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$26>>2] = $27; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $28 = ((($21)) + 4|0); //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$28>>2] = $0; //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$21>>2] = $29; //@line 1098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $30 = ((($29)) + 4|0); //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$30>>2] = $21; //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$0>>2] = $21; //@line 1100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $31 = (($32) + 1)|0; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$2>>2] = $31; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $33 = $__i$sroa$0$03$i; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $34 = ((($33)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $__i$sroa$0$0$i = HEAP32[$34>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $35 = $__i$sroa$0$0$i; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $36 = ($35|0)==($formats|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($36) {
     label = 9;
     break;
    } else {
     $25 = $35;$29 = $21;$32 = $31;$__i$sroa$0$03$i = $__i$sroa$0$0$i;
    }
   }
   if ((label|0) == 9) {
    $__begin$sroa$0$06$pre = HEAP32[$1>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $$pre$phiZ2D = $0;$__begin$sroa$0$06 = $__begin$sroa$0$06$pre;
    break;
   }
   $7 = ___cxa_find_matching_catch_2()|0;
   $8 = tempRet0;
   $9 = ($$lcssa23|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if (!($9)) {
    $10 = HEAP32[$1>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $11 = ((($$lcssa)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $12 = HEAP32[$11>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $13 = HEAP32[$10>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $14 = ((($13)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$14>>2] = $12; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $15 = $13; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $16 = HEAP32[$11>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$16>>2] = $15; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$2>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $17 = ($10|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if (!($17)) {
     $__f$01$i$i$i = $10;
     while(1) {
      $18 = ((($__f$01$i$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $19 = HEAP32[$18>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      __ZdlPv($__f$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $20 = ($19|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if ($20) {
       break;
      } else {
       $__f$01$i$i$i = $19;
      }
     }
    }
   }
   $$0 = $7;$$01 = $8;
   __ZN4love5image9ImageDataD2Ev($this); //@line 39 "modules/image/magpie/ImageData.cpp"
   ___resumeException($$0|0);
   // unreachable;
  }
 } while(0);
 $37 = ((($this)) + 36|0); //@line 33 "modules/image/magpie/ImageData.cpp"
 HEAP32[$37>>2] = 0; //@line 33 "modules/image/magpie/ImageData.cpp"
 $38 = $__begin$sroa$0$06; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $39 = ($38|0)==($$pre$phiZ2D|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 L17: do {
  if ($39) {
   label = 11;
  } else {
   $61 = $38;$__begin$sroa$0$07 = $__begin$sroa$0$06;
   while(1) {
    $60 = ((($61)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $62 = HEAP32[$60>>2]|0; //@line 35 "modules/image/magpie/ImageData.cpp"
    $63 = HEAP32[$62>>2]|0; //@line 36 "modules/image/magpie/ImageData.cpp"
    $64 = ((($63)) + 8|0); //@line 36 "modules/image/magpie/ImageData.cpp"
    $65 = HEAP32[$64>>2]|0; //@line 36 "modules/image/magpie/ImageData.cpp"
    __THREW__ = 0;
    invoke_vi($65|0,($62|0)); //@line 36 "modules/image/magpie/ImageData.cpp"
    $66 = __THREW__; __THREW__ = 0;
    $67 = $66&1;
    if ($67) {
     break;
    }
    $68 = $__begin$sroa$0$07; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $69 = ((($68)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $__begin$sroa$0$0 = HEAP32[$69>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $70 = $__begin$sroa$0$0; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $71 = ($70|0)==($$pre$phiZ2D|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($71) {
     label = 11;
     break L17;
    } else {
     $61 = $70;$__begin$sroa$0$07 = $__begin$sroa$0$0;
    }
   }
   $42 = ___cxa_find_matching_catch_2()|0;
   $43 = tempRet0;
   $lpad$phi$index = $42;$lpad$phi$index2 = $43;
  }
 } while(0);
 do {
  if ((label|0) == 11) {
   __THREW__ = 0;
   invoke_vii(4167,($this|0),($data|0)); //@line 38 "modules/image/magpie/ImageData.cpp"
   $40 = __THREW__; __THREW__ = 0;
   $41 = $40&1;
   if ($41) {
    $44 = ___cxa_find_matching_catch_2()|0;
    $45 = tempRet0;
    $lpad$phi$index = $44;$lpad$phi$index2 = $45;
    break;
   } else {
    return; //@line 39 "modules/image/magpie/ImageData.cpp"
   }
  }
 } while(0);
 $46 = HEAP32[$2>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $47 = ($46|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($47) {
  $$0 = $lpad$phi$index;$$01 = $lpad$phi$index2;
  __ZN4love5image9ImageDataD2Ev($this); //@line 39 "modules/image/magpie/ImageData.cpp"
  ___resumeException($$0|0);
  // unreachable;
 }
 $48 = HEAP32[$1>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $49 = HEAP32[$0>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $50 = ((($49)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $51 = HEAP32[$50>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $52 = HEAP32[$48>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $53 = ((($52)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$53>>2] = $51; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $54 = $52; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $55 = HEAP32[$50>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$55>>2] = $54; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$2>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $56 = ($48|0)==($$pre$phiZ2D|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($56) {
  $$0 = $lpad$phi$index;$$01 = $lpad$phi$index2;
  __ZN4love5image9ImageDataD2Ev($this); //@line 39 "modules/image/magpie/ImageData.cpp"
  ___resumeException($$0|0);
  // unreachable;
 } else {
  $__f$01$i$i = $48;
 }
 while(1) {
  $57 = ((($__f$01$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $58 = HEAP32[$57>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  __ZdlPv($__f$01$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $59 = ($58|0)==($$pre$phiZ2D|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($59) {
   $$0 = $lpad$phi$index;$$01 = $lpad$phi$index2;
   break;
  } else {
   $__f$01$i$i = $58;
  }
 }
 __ZN4love5image9ImageDataD2Ev($this); //@line 39 "modules/image/magpie/ImageData.cpp"
 ___resumeException($$0|0);
 // unreachable;
}
function __ZN4love5image6magpie9ImageData6decodeEPNS_10filesystem8FileDataE($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$0 = 0, $$01 = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$04 = 0, $__begin$sroa$0$06 = 0, $decoder$2 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 16|0;
 $1 = ((($this)) + 24|0); //@line 106 "modules/image/magpie/ImageData.cpp"
 $2 = ((($this)) + 28|0); //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $__begin$sroa$0$04 = HEAP32[$2>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $3 = $__begin$sroa$0$04; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $4 = ($3|0)==($1|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if (!($4)) {
  $6 = $3;$__begin$sroa$0$06 = $__begin$sroa$0$04;
  while(1) {
   $5 = ((($6)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $7 = HEAP32[$5>>2]|0; //@line 106 "modules/image/magpie/ImageData.cpp"
   $8 = HEAP32[$7>>2]|0; //@line 108 "modules/image/magpie/ImageData.cpp"
   $9 = ((($8)) + 16|0); //@line 108 "modules/image/magpie/ImageData.cpp"
   $10 = HEAP32[$9>>2]|0; //@line 108 "modules/image/magpie/ImageData.cpp"
   $11 = (FUNCTION_TABLE_iii[$10 & 8191]($7,$data)|0); //@line 108 "modules/image/magpie/ImageData.cpp"
   if ($11) {
    $decoder$2 = $7;
    break;
   }
   $12 = $__begin$sroa$0$06; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $13 = ((($12)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $__begin$sroa$0$0 = HEAP32[$13>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $14 = $__begin$sroa$0$0; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $15 = ($14|0)==($1|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($15) {
    $decoder$2 = 0;
    break;
   } else {
    $6 = $14;$__begin$sroa$0$06 = $__begin$sroa$0$0;
   }
  }
  $16 = ($decoder$2|0)==(0|0); //@line 115 "modules/image/magpie/ImageData.cpp"
  if (!($16)) {
   $17 = HEAP32[$decoder$2>>2]|0; //@line 116 "modules/image/magpie/ImageData.cpp"
   $18 = ((($17)) + 24|0); //@line 116 "modules/image/magpie/ImageData.cpp"
   $19 = HEAP32[$18>>2]|0; //@line 116 "modules/image/magpie/ImageData.cpp"
   FUNCTION_TABLE_viii[$19 & 8191]($0,$decoder$2,$data); //@line 116 "modules/image/magpie/ImageData.cpp"
   $20 = HEAP32[$0>>2]|0; //@line 116 "modules/image/magpie/ImageData.cpp"
   $21 = ((($0)) + 4|0); //@line 116 "modules/image/magpie/ImageData.cpp"
   $22 = HEAP32[$21>>2]|0; //@line 116 "modules/image/magpie/ImageData.cpp"
   $23 = ((($0)) + 12|0); //@line 116 "modules/image/magpie/ImageData.cpp"
   $24 = HEAP32[$23>>2]|0; //@line 116 "modules/image/magpie/ImageData.cpp"
   $25 = $24; //@line 118 "modules/image/magpie/ImageData.cpp"
   $26 = ($24|0)==(0); //@line 118 "modules/image/magpie/ImageData.cpp"
   if (!($26)) {
    $40 = ((($0)) + 8|0); //@line 116 "modules/image/magpie/ImageData.cpp"
    $41 = HEAP32[$40>>2]|0; //@line 116 "modules/image/magpie/ImageData.cpp"
    $42 = $22 << 2; //@line 125 "modules/image/magpie/ImageData.cpp"
    $43 = Math_imul($42, $20)|0; //@line 125 "modules/image/magpie/ImageData.cpp"
    $44 = ($41|0)==($43|0); //@line 125 "modules/image/magpie/ImageData.cpp"
    $45 = ((($this)) + 36|0); //@line 127 "modules/image/magpie/ImageData.cpp"
    $46 = HEAP32[$45>>2]|0; //@line 127 "modules/image/magpie/ImageData.cpp"
    $47 = ($46|0)!=(0|0); //@line 127 "modules/image/magpie/ImageData.cpp"
    if ($44) {
     if ($47) {
      $56 = HEAP32[$46>>2]|0; //@line 136 "modules/image/magpie/ImageData.cpp"
      $57 = ((($56)) + 32|0); //@line 136 "modules/image/magpie/ImageData.cpp"
      $58 = HEAP32[$57>>2]|0; //@line 136 "modules/image/magpie/ImageData.cpp"
      $59 = ((($this)) + 16|0); //@line 136 "modules/image/magpie/ImageData.cpp"
      $60 = HEAP32[$59>>2]|0; //@line 136 "modules/image/magpie/ImageData.cpp"
      FUNCTION_TABLE_vii[$58 & 8191]($46,$60); //@line 136 "modules/image/magpie/ImageData.cpp"
      $$pre$phiZ2D = $59;
      $64 = ((($this)) + 8|0); //@line 140 "modules/image/magpie/ImageData.cpp"
      HEAP32[$64>>2] = $20; //@line 140 "modules/image/magpie/ImageData.cpp"
      $65 = ((($this)) + 12|0); //@line 141 "modules/image/magpie/ImageData.cpp"
      HEAP32[$65>>2] = $22; //@line 141 "modules/image/magpie/ImageData.cpp"
      HEAP32[$$pre$phiZ2D>>2] = $24; //@line 142 "modules/image/magpie/ImageData.cpp"
      HEAP32[$45>>2] = $decoder$2; //@line 144 "modules/image/magpie/ImageData.cpp"
      STACKTOP = sp;return; //@line 145 "modules/image/magpie/ImageData.cpp"
     }
     $61 = ((($this)) + 16|0); //@line 138 "modules/image/magpie/ImageData.cpp"
     $62 = HEAP32[$61>>2]|0; //@line 138 "modules/image/magpie/ImageData.cpp"
     $63 = ($62|0)==(0|0); //@line 138 "modules/image/magpie/ImageData.cpp"
     if ($63) {
      $$pre$phiZ2D = $61;
      $64 = ((($this)) + 8|0); //@line 140 "modules/image/magpie/ImageData.cpp"
      HEAP32[$64>>2] = $20; //@line 140 "modules/image/magpie/ImageData.cpp"
      $65 = ((($this)) + 12|0); //@line 141 "modules/image/magpie/ImageData.cpp"
      HEAP32[$65>>2] = $22; //@line 141 "modules/image/magpie/ImageData.cpp"
      HEAP32[$$pre$phiZ2D>>2] = $24; //@line 142 "modules/image/magpie/ImageData.cpp"
      HEAP32[$45>>2] = $decoder$2; //@line 144 "modules/image/magpie/ImageData.cpp"
      STACKTOP = sp;return; //@line 145 "modules/image/magpie/ImageData.cpp"
     }
     __ZdaPv($62); //@line 138 "modules/image/magpie/ImageData.cpp"
     $$pre$phiZ2D = $61;
     $64 = ((($this)) + 8|0); //@line 140 "modules/image/magpie/ImageData.cpp"
     HEAP32[$64>>2] = $20; //@line 140 "modules/image/magpie/ImageData.cpp"
     $65 = ((($this)) + 12|0); //@line 141 "modules/image/magpie/ImageData.cpp"
     HEAP32[$65>>2] = $22; //@line 141 "modules/image/magpie/ImageData.cpp"
     HEAP32[$$pre$phiZ2D>>2] = $24; //@line 142 "modules/image/magpie/ImageData.cpp"
     HEAP32[$45>>2] = $decoder$2; //@line 144 "modules/image/magpie/ImageData.cpp"
     STACKTOP = sp;return; //@line 145 "modules/image/magpie/ImageData.cpp"
    } else {
     if ($47) {
      $48 = HEAP32[$46>>2]|0; //@line 128 "modules/image/magpie/ImageData.cpp"
      $49 = ((($48)) + 32|0); //@line 128 "modules/image/magpie/ImageData.cpp"
      $50 = HEAP32[$49>>2]|0; //@line 128 "modules/image/magpie/ImageData.cpp"
      FUNCTION_TABLE_vii[$50 & 8191]($46,$25); //@line 128 "modules/image/magpie/ImageData.cpp"
     } else {
      __ZdaPv($25); //@line 130 "modules/image/magpie/ImageData.cpp"
     }
     $51 = (___cxa_allocate_exception(16)|0); //@line 131 "modules/image/magpie/ImageData.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($51|0),(439758|0),($vararg_buffer1|0)); //@line 131 "modules/image/magpie/ImageData.cpp"
     $52 = __THREW__; __THREW__ = 0;
     $53 = $52&1;
     if (!($53)) {
      ___cxa_throw(($51|0),(96|0),(344|0)); //@line 131 "modules/image/magpie/ImageData.cpp"
      // unreachable; //@line 131 "modules/image/magpie/ImageData.cpp"
     }
     $54 = ___cxa_find_matching_catch_2()|0;
     $55 = tempRet0;
     ___cxa_free_exception(($51|0)); //@line 131 "modules/image/magpie/ImageData.cpp"
     $$0 = $55;$$01 = $54;
     ___resumeException($$01|0);
     // unreachable;
    }
   }
  }
 }
 $27 = (__ZNK4love10filesystem8FileData11getFilenameEv($data)|0); //@line 120 "modules/image/magpie/ImageData.cpp"
 $28 = (___cxa_allocate_exception(16)|0); //@line 121 "modules/image/magpie/ImageData.cpp"
 $29 = HEAP8[$27>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $30 = $29 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $31 = ($30<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $32 = ((($27)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $33 = HEAP32[$32>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $34 = ((($27)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $35 = $31 ? $34 : $33; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $35; //@line 121 "modules/image/magpie/ImageData.cpp"
 invoke_viii(3349,($28|0),(439693|0),($vararg_buffer|0)); //@line 121 "modules/image/magpie/ImageData.cpp"
 $36 = __THREW__; __THREW__ = 0;
 $37 = $36&1;
 if (!($37)) {
  ___cxa_throw(($28|0),(96|0),(344|0)); //@line 121 "modules/image/magpie/ImageData.cpp"
  // unreachable; //@line 121 "modules/image/magpie/ImageData.cpp"
 }
 $38 = ___cxa_find_matching_catch_2()|0;
 $39 = tempRet0;
 ___cxa_free_exception(($28|0)); //@line 121 "modules/image/magpie/ImageData.cpp"
 $$0 = $39;$$01 = $38;
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN4love5image6magpie9ImageDataC2ENSt3__14listIPNS1_13FormatHandlerENS3_9allocatorIS6_EEEEii($this,$formats,$width,$height) {
 $this = $this|0;
 $formats = $formats|0;
 $width = $width|0;
 $height = $height|0;
 var $$0 = 0, $$01 = 0, $$lcssa = 0, $$lcssa23 = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$06 = 0, $__begin$sroa$0$06$pre = 0, $__begin$sroa$0$07 = 0, $__f$01$i$i = 0, $__f$01$i$i$i = 0, $__i$sroa$0$0$i = 0, $__i$sroa$0$02$i = 0, $__i$sroa$0$03$i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5image9ImageDataC2Ev($this); //@line 41 "modules/image/magpie/ImageData.cpp"
 HEAP32[$this>>2] = (25792); //@line 44 "modules/image/magpie/ImageData.cpp"
 $0 = ((($this)) + 24|0); //@line 42 "modules/image/magpie/ImageData.cpp"
 HEAP32[$0>>2] = $0; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $1 = ((($this)) + 28|0); //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$1>>2] = $0; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $2 = ((($this)) + 32|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$2>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = ((($formats)) + 4|0); //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $__i$sroa$0$02$i = HEAP32[$3>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $4 = $__i$sroa$0$02$i; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $5 = ($4|0)==($formats|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 do {
  if ($5) {
   $6 = $0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $$pre$phiZ2D = $0;$__begin$sroa$0$06 = $6;
  } else {
   $25 = $4;$29 = $0;$32 = 0;$__i$sroa$0$03$i = $__i$sroa$0$02$i;
   while(1) {
    __THREW__ = 0;
    $21 = (invoke_ii(3348,12)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $22 = __THREW__; __THREW__ = 0;
    $23 = $22&1;
    if ($23) {
     $$lcssa = $29;$$lcssa23 = $32;
     break;
    }
    $24 = ((($25)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $26 = ((($21)) + 8|0); //@line 1537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $27 = HEAP32[$24>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$26>>2] = $27; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $28 = ((($21)) + 4|0); //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$28>>2] = $0; //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$21>>2] = $29; //@line 1098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $30 = ((($29)) + 4|0); //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$30>>2] = $21; //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$0>>2] = $21; //@line 1100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $31 = (($32) + 1)|0; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$2>>2] = $31; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $33 = $__i$sroa$0$03$i; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $34 = ((($33)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $__i$sroa$0$0$i = HEAP32[$34>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $35 = $__i$sroa$0$0$i; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $36 = ($35|0)==($formats|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($36) {
     label = 9;
     break;
    } else {
     $25 = $35;$29 = $21;$32 = $31;$__i$sroa$0$03$i = $__i$sroa$0$0$i;
    }
   }
   if ((label|0) == 9) {
    $__begin$sroa$0$06$pre = HEAP32[$1>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $$pre$phiZ2D = $0;$__begin$sroa$0$06 = $__begin$sroa$0$06$pre;
    break;
   }
   $7 = ___cxa_find_matching_catch_2()|0;
   $8 = tempRet0;
   $9 = ($$lcssa23|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if (!($9)) {
    $10 = HEAP32[$1>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $11 = ((($$lcssa)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $12 = HEAP32[$11>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $13 = HEAP32[$10>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $14 = ((($13)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$14>>2] = $12; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $15 = $13; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $16 = HEAP32[$11>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$16>>2] = $15; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$2>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $17 = ($10|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if (!($17)) {
     $__f$01$i$i$i = $10;
     while(1) {
      $18 = ((($__f$01$i$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $19 = HEAP32[$18>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      __ZdlPv($__f$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $20 = ($19|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if ($20) {
       break;
      } else {
       $__f$01$i$i$i = $19;
      }
     }
    }
   }
   $$0 = $7;$$01 = $8;
   __ZN4love5image9ImageDataD2Ev($this); //@line 55 "modules/image/magpie/ImageData.cpp"
   ___resumeException($$0|0);
   // unreachable;
  }
 } while(0);
 $37 = ((($this)) + 36|0); //@line 43 "modules/image/magpie/ImageData.cpp"
 HEAP32[$37>>2] = 0; //@line 43 "modules/image/magpie/ImageData.cpp"
 $38 = $__begin$sroa$0$06; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $39 = ($38|0)==($$pre$phiZ2D|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 L17: do {
  if ($39) {
   label = 11;
  } else {
   $63 = $38;$__begin$sroa$0$07 = $__begin$sroa$0$06;
   while(1) {
    $62 = ((($63)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $64 = HEAP32[$62>>2]|0; //@line 45 "modules/image/magpie/ImageData.cpp"
    $65 = HEAP32[$64>>2]|0; //@line 46 "modules/image/magpie/ImageData.cpp"
    $66 = ((($65)) + 8|0); //@line 46 "modules/image/magpie/ImageData.cpp"
    $67 = HEAP32[$66>>2]|0; //@line 46 "modules/image/magpie/ImageData.cpp"
    __THREW__ = 0;
    invoke_vi($67|0,($64|0)); //@line 46 "modules/image/magpie/ImageData.cpp"
    $68 = __THREW__; __THREW__ = 0;
    $69 = $68&1;
    if ($69) {
     break;
    }
    $70 = $__begin$sroa$0$07; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $71 = ((($70)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $__begin$sroa$0$0 = HEAP32[$71>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $72 = $__begin$sroa$0$0; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $73 = ($72|0)==($$pre$phiZ2D|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($73) {
     label = 11;
     break L17;
    } else {
     $63 = $72;$__begin$sroa$0$07 = $__begin$sroa$0$0;
    }
   }
   $44 = ___cxa_find_matching_catch_2()|0;
   $45 = tempRet0;
   $lpad$phi$index = $44;$lpad$phi$index2 = $45;
  }
 } while(0);
 do {
  if ((label|0) == 11) {
   $40 = ((($this)) + 8|0); //@line 48 "modules/image/magpie/ImageData.cpp"
   HEAP32[$40>>2] = $width; //@line 48 "modules/image/magpie/ImageData.cpp"
   $41 = ((($this)) + 12|0); //@line 49 "modules/image/magpie/ImageData.cpp"
   HEAP32[$41>>2] = $height; //@line 49 "modules/image/magpie/ImageData.cpp"
   __THREW__ = 0;
   invoke_viiii(4168,($this|0),($width|0),($height|0),(0|0)); //@line 51 "modules/image/magpie/ImageData.cpp"
   $42 = __THREW__; __THREW__ = 0;
   $43 = $42&1;
   if ($43) {
    $46 = ___cxa_find_matching_catch_2()|0;
    $47 = tempRet0;
    $lpad$phi$index = $46;$lpad$phi$index2 = $47;
    break;
   } else {
    $74 = ((($this)) + 16|0); //@line 54 "modules/image/magpie/ImageData.cpp"
    $75 = HEAP32[$74>>2]|0; //@line 54 "modules/image/magpie/ImageData.cpp"
    $76 = $width << 2; //@line 54 "modules/image/magpie/ImageData.cpp"
    $77 = Math_imul($76, $height)|0; //@line 54 "modules/image/magpie/ImageData.cpp"
    _memset(($75|0),0,($77|0))|0; //@line 54 "modules/image/magpie/ImageData.cpp"
    return; //@line 55 "modules/image/magpie/ImageData.cpp"
   }
  }
 } while(0);
 $48 = HEAP32[$2>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $49 = ($48|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($49) {
  $$0 = $lpad$phi$index;$$01 = $lpad$phi$index2;
  __ZN4love5image9ImageDataD2Ev($this); //@line 55 "modules/image/magpie/ImageData.cpp"
  ___resumeException($$0|0);
  // unreachable;
 }
 $50 = HEAP32[$1>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $51 = HEAP32[$0>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $52 = ((($51)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $53 = HEAP32[$52>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $54 = HEAP32[$50>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $55 = ((($54)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$55>>2] = $53; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $56 = $54; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $57 = HEAP32[$52>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$57>>2] = $56; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$2>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $58 = ($50|0)==($$pre$phiZ2D|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($58) {
  $$0 = $lpad$phi$index;$$01 = $lpad$phi$index2;
  __ZN4love5image9ImageDataD2Ev($this); //@line 55 "modules/image/magpie/ImageData.cpp"
  ___resumeException($$0|0);
  // unreachable;
 } else {
  $__f$01$i$i = $50;
 }
 while(1) {
  $59 = ((($__f$01$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $60 = HEAP32[$59>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  __ZdlPv($__f$01$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $61 = ($60|0)==($$pre$phiZ2D|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($61) {
   $$0 = $lpad$phi$index;$$01 = $lpad$phi$index2;
   break;
  } else {
   $__f$01$i$i = $60;
  }
 }
 __ZN4love5image9ImageDataD2Ev($this); //@line 55 "modules/image/magpie/ImageData.cpp"
 ___resumeException($$0|0);
 // unreachable;
}
function __ZN4love5image6magpie9ImageData6createEiiPv($this,$width,$height,$data) {
 $this = $this|0;
 $width = $width|0;
 $height = $height|0;
 $data = $data|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = $width << 2; //@line 88 "modules/image/magpie/ImageData.cpp"
 $1 = Math_imul($0, $height)|0; //@line 88 "modules/image/magpie/ImageData.cpp"
 __THREW__ = 0;
 $2 = (invoke_ii(3492,($1|0))|0); //@line 88 "modules/image/magpie/ImageData.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  $5 = ((($this)) + 16|0); //@line 88 "modules/image/magpie/ImageData.cpp"
  HEAP32[$5>>2] = $2; //@line 88 "modules/image/magpie/ImageData.cpp"
  $6 = ($data|0)==(0|0); //@line 95 "modules/image/magpie/ImageData.cpp"
  if ($6) {
   $21 = ((($this)) + 36|0); //@line 98 "modules/image/magpie/ImageData.cpp"
   HEAP32[$21>>2] = 0; //@line 98 "modules/image/magpie/ImageData.cpp"
   STACKTOP = sp;return; //@line 99 "modules/image/magpie/ImageData.cpp"
  }
  _memcpy(($2|0),($data|0),($1|0))|0; //@line 96 "modules/image/magpie/ImageData.cpp"
  $21 = ((($this)) + 36|0); //@line 98 "modules/image/magpie/ImageData.cpp"
  HEAP32[$21>>2] = 0; //@line 98 "modules/image/magpie/ImageData.cpp"
  STACKTOP = sp;return; //@line 99 "modules/image/magpie/ImageData.cpp"
 }
 $7 = ___cxa_find_matching_catch_3(9056|0)|0;
 $8 = tempRet0;
 $9 = (_llvm_eh_typeid_for((9056|0))|0); //@line 89 "modules/image/magpie/ImageData.cpp"
 $10 = ($8|0)==($9|0); //@line 89 "modules/image/magpie/ImageData.cpp"
 if (!($10)) {
  $$1 = $7;$$12 = $8;
  ___resumeException($$1|0);
  // unreachable;
 }
 (___cxa_begin_catch(($7|0))|0); //@line 89 "modules/image/magpie/ImageData.cpp"
 $11 = (___cxa_allocate_exception(16)|0); //@line 92 "modules/image/magpie/ImageData.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($11|0),(596704|0),($vararg_buffer|0)); //@line 92 "modules/image/magpie/ImageData.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if ($13) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  ___cxa_free_exception(($11|0)); //@line 92 "modules/image/magpie/ImageData.cpp"
  $$0 = $17;$$01 = $18;
 } else {
  __THREW__ = 0;
  invoke_viii(3350,($11|0),(96|0),(344|0)); //@line 92 "modules/image/magpie/ImageData.cpp"
  $14 = __THREW__; __THREW__ = 0;
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  $$0 = $15;$$01 = $16;
 }
 __THREW__ = 0;
 invoke_v(3378); //@line 93 "modules/image/magpie/ImageData.cpp"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $22 = ___cxa_find_matching_catch_3(0|0)|0;
  $23 = tempRet0;
  ___clang_call_terminate($22); //@line 93 "modules/image/magpie/ImageData.cpp"
  // unreachable; //@line 93 "modules/image/magpie/ImageData.cpp"
 } else {
  $$1 = $$0;$$12 = $$01;
  ___resumeException($$1|0);
  // unreachable;
 }
}
function __ZN4love5image6magpie9ImageDataC2ENSt3__14listIPNS1_13FormatHandlerENS3_9allocatorIS6_EEEEiiPvb($this,$formats,$width,$height,$data,$own) {
 $this = $this|0;
 $formats = $formats|0;
 $width = $width|0;
 $height = $height|0;
 $data = $data|0;
 $own = $own|0;
 var $$0 = 0, $$01 = 0, $$lcssa = 0, $$lcssa23 = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $8 = 0, $9 = 0;
 var $__begin$sroa$0$0 = 0, $__begin$sroa$0$06 = 0, $__begin$sroa$0$06$pre = 0, $__begin$sroa$0$07 = 0, $__f$01$i$i = 0, $__f$01$i$i$i = 0, $__i$sroa$0$0$i = 0, $__i$sroa$0$02$i = 0, $__i$sroa$0$03$i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5image9ImageDataC2Ev($this); //@line 57 "modules/image/magpie/ImageData.cpp"
 HEAP32[$this>>2] = (25792); //@line 60 "modules/image/magpie/ImageData.cpp"
 $0 = ((($this)) + 24|0); //@line 58 "modules/image/magpie/ImageData.cpp"
 HEAP32[$0>>2] = $0; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $1 = ((($this)) + 28|0); //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$1>>2] = $0; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $2 = ((($this)) + 32|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$2>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = ((($formats)) + 4|0); //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $__i$sroa$0$02$i = HEAP32[$3>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $4 = $__i$sroa$0$02$i; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $5 = ($4|0)==($formats|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 do {
  if ($5) {
   $6 = $0; //@line 1198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $$pre$phiZ2D = $0;$__begin$sroa$0$06 = $6;
  } else {
   $25 = $4;$29 = $0;$32 = 0;$__i$sroa$0$03$i = $__i$sroa$0$02$i;
   while(1) {
    __THREW__ = 0;
    $21 = (invoke_ii(3348,12)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $22 = __THREW__; __THREW__ = 0;
    $23 = $22&1;
    if ($23) {
     $$lcssa = $29;$$lcssa23 = $32;
     break;
    }
    $24 = ((($25)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $26 = ((($21)) + 8|0); //@line 1537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $27 = HEAP32[$24>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$26>>2] = $27; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $28 = ((($21)) + 4|0); //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$28>>2] = $0; //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$21>>2] = $29; //@line 1098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $30 = ((($29)) + 4|0); //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$30>>2] = $21; //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$0>>2] = $21; //@line 1100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $31 = (($32) + 1)|0; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$2>>2] = $31; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $33 = $__i$sroa$0$03$i; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $34 = ((($33)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $__i$sroa$0$0$i = HEAP32[$34>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $35 = $__i$sroa$0$0$i; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $36 = ($35|0)==($formats|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($36) {
     label = 9;
     break;
    } else {
     $25 = $35;$29 = $21;$32 = $31;$__i$sroa$0$03$i = $__i$sroa$0$0$i;
    }
   }
   if ((label|0) == 9) {
    $__begin$sroa$0$06$pre = HEAP32[$1>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $$pre$phiZ2D = $0;$__begin$sroa$0$06 = $__begin$sroa$0$06$pre;
    break;
   }
   $7 = ___cxa_find_matching_catch_2()|0;
   $8 = tempRet0;
   $9 = ($$lcssa23|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if (!($9)) {
    $10 = HEAP32[$1>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $11 = ((($$lcssa)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $12 = HEAP32[$11>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $13 = HEAP32[$10>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $14 = ((($13)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$14>>2] = $12; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $15 = $13; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $16 = HEAP32[$11>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$16>>2] = $15; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$2>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $17 = ($10|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if (!($17)) {
     $__f$01$i$i$i = $10;
     while(1) {
      $18 = ((($__f$01$i$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $19 = HEAP32[$18>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      __ZdlPv($__f$01$i$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $20 = ($19|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if ($20) {
       break;
      } else {
       $__f$01$i$i$i = $19;
      }
     }
    }
   }
   $$0 = $7;$$01 = $8;
   __ZN4love5image9ImageDataD2Ev($this); //@line 71 "modules/image/magpie/ImageData.cpp"
   ___resumeException($$0|0);
   // unreachable;
  }
 } while(0);
 $37 = ((($this)) + 36|0); //@line 59 "modules/image/magpie/ImageData.cpp"
 HEAP32[$37>>2] = 0; //@line 59 "modules/image/magpie/ImageData.cpp"
 $38 = $__begin$sroa$0$06; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $39 = ($38|0)==($$pre$phiZ2D|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 L17: do {
  if ($39) {
   label = 11;
  } else {
   $61 = $38;$__begin$sroa$0$07 = $__begin$sroa$0$06;
   while(1) {
    $60 = ((($61)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $62 = HEAP32[$60>>2]|0; //@line 61 "modules/image/magpie/ImageData.cpp"
    $63 = HEAP32[$62>>2]|0; //@line 62 "modules/image/magpie/ImageData.cpp"
    $64 = ((($63)) + 8|0); //@line 62 "modules/image/magpie/ImageData.cpp"
    $65 = HEAP32[$64>>2]|0; //@line 62 "modules/image/magpie/ImageData.cpp"
    __THREW__ = 0;
    invoke_vi($65|0,($62|0)); //@line 62 "modules/image/magpie/ImageData.cpp"
    $66 = __THREW__; __THREW__ = 0;
    $67 = $66&1;
    if ($67) {
     break;
    }
    $68 = $__begin$sroa$0$07; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $69 = ((($68)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $__begin$sroa$0$0 = HEAP32[$69>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $70 = $__begin$sroa$0$0; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $71 = ($70|0)==($$pre$phiZ2D|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($71) {
     label = 11;
     break L17;
    } else {
     $61 = $70;$__begin$sroa$0$07 = $__begin$sroa$0$0;
    }
   }
   $42 = ___cxa_find_matching_catch_2()|0;
   $43 = tempRet0;
   $lpad$phi$index = $42;$lpad$phi$index2 = $43;
  }
 } while(0);
 do {
  if ((label|0) == 11) {
   $40 = ((($this)) + 8|0); //@line 64 "modules/image/magpie/ImageData.cpp"
   HEAP32[$40>>2] = $width; //@line 64 "modules/image/magpie/ImageData.cpp"
   $41 = ((($this)) + 12|0); //@line 65 "modules/image/magpie/ImageData.cpp"
   HEAP32[$41>>2] = $height; //@line 65 "modules/image/magpie/ImageData.cpp"
   if ($own) {
    $72 = ((($this)) + 16|0); //@line 68 "modules/image/magpie/ImageData.cpp"
    HEAP32[$72>>2] = $data; //@line 68 "modules/image/magpie/ImageData.cpp"
    return; //@line 71 "modules/image/magpie/ImageData.cpp"
   }
   __THREW__ = 0;
   invoke_viiii(4168,($this|0),($width|0),($height|0),($data|0)); //@line 70 "modules/image/magpie/ImageData.cpp"
   $73 = __THREW__; __THREW__ = 0;
   $74 = $73&1;
   if ($74) {
    $44 = ___cxa_find_matching_catch_2()|0;
    $45 = tempRet0;
    $lpad$phi$index = $44;$lpad$phi$index2 = $45;
    break;
   } else {
    return; //@line 71 "modules/image/magpie/ImageData.cpp"
   }
  }
 } while(0);
 $46 = HEAP32[$2>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $47 = ($46|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($47) {
  $$0 = $lpad$phi$index;$$01 = $lpad$phi$index2;
  __ZN4love5image9ImageDataD2Ev($this); //@line 71 "modules/image/magpie/ImageData.cpp"
  ___resumeException($$0|0);
  // unreachable;
 }
 $48 = HEAP32[$1>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $49 = HEAP32[$0>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $50 = ((($49)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $51 = HEAP32[$50>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $52 = HEAP32[$48>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $53 = ((($52)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$53>>2] = $51; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $54 = $52; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $55 = HEAP32[$50>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$55>>2] = $54; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$2>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $56 = ($48|0)==($$pre$phiZ2D|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($56) {
  $$0 = $lpad$phi$index;$$01 = $lpad$phi$index2;
  __ZN4love5image9ImageDataD2Ev($this); //@line 71 "modules/image/magpie/ImageData.cpp"
  ___resumeException($$0|0);
  // unreachable;
 } else {
  $__f$01$i$i = $48;
 }
 while(1) {
  $57 = ((($__f$01$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $58 = HEAP32[$57>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  __ZdlPv($__f$01$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $59 = ($58|0)==($$pre$phiZ2D|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($59) {
   $$0 = $lpad$phi$index;$$01 = $lpad$phi$index2;
   break;
  } else {
   $__f$01$i$i = $58;
  }
 }
 __ZN4love5image9ImageDataD2Ev($this); //@line 71 "modules/image/magpie/ImageData.cpp"
 ___resumeException($$0|0);
 // unreachable;
}
function __ZN4love5image6magpie10KTXHandlerD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 41 "modules/image/magpie/KTXHandler.h"
 return; //@line 41 "modules/image/magpie/KTXHandler.h"
}
function __ZN4love5image6magpie10KTXHandler8canParseEPKNS_10filesystem8FileDataE($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $switch$split2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$data>>2]|0; //@line 234 "modules/image/magpie/KTXHandler.cpp"
 $1 = ((($0)) + 20|0); //@line 234 "modules/image/magpie/KTXHandler.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 234 "modules/image/magpie/KTXHandler.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($data)|0); //@line 234 "modules/image/magpie/KTXHandler.cpp"
 $4 = ($3>>>0)<(64); //@line 234 "modules/image/magpie/KTXHandler.cpp"
 L1: do {
  if ($4) {
   $$1 = 0;
  } else {
   $5 = HEAP32[$data>>2]|0; //@line 237 "modules/image/magpie/KTXHandler.cpp"
   $6 = ((($5)) + 16|0); //@line 237 "modules/image/magpie/KTXHandler.cpp"
   $7 = HEAP32[$6>>2]|0; //@line 237 "modules/image/magpie/KTXHandler.cpp"
   $8 = (FUNCTION_TABLE_ii[$7 & 8191]($data)|0); //@line 237 "modules/image/magpie/KTXHandler.cpp"
   $9 = (_memcmp($8,439862,12)|0); //@line 240 "modules/image/magpie/KTXHandler.cpp"
   $10 = ($9|0)==(0); //@line 240 "modules/image/magpie/KTXHandler.cpp"
   if ($10) {
    $11 = ((($8)) + 12|0); //@line 243 "modules/image/magpie/KTXHandler.cpp"
    $12 = HEAP32[$11>>2]|0; //@line 243 "modules/image/magpie/KTXHandler.cpp"
    $switch$split2D = ($12|0)<(67305985);
    if ($switch$split2D) {
     switch ($12|0) {
     case 16909060:  {
      break;
     }
     default: {
      $$1 = 0;
      break L1;
     }
     }
    } else {
     switch ($12|0) {
     case 67305985:  {
      break;
     }
     default: {
      $$1 = 0;
      break L1;
     }
     }
    }
    $$1 = 1;
   } else {
    $$1 = 0;
   }
  }
 } while(0);
 return ($$1|0); //@line 247 "modules/image/magpie/KTXHandler.cpp"
}
function __ZN4love5image6magpie10KTXHandler5parseEPNS_10filesystem8FileDataERNSt3__16vectorINS0_19CompressedImageData8SubImageENS6_9allocatorIS9_EEEERjRNS8_6FormatERb($this,$filedata,$images,$dataSize,$format,$sRGB) {
 $this = $this|0;
 $filedata = $filedata|0;
 $images = $images|0;
 $dataSize = $dataSize|0;
 $format = $format|0;
 $sRGB = $sRGB|0;
 var $$0 = 0, $$0$i$ph = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$phi$trans$insert = 0, $$phi$trans$insert22 = 0, $$pre = 0, $$pre$phiZ2D = 0, $$pre23 = 0, $$pre24 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dataoffset$08 = 0, $dataoffset$08$us = 0, $fileoffset$010 = 0, $fileoffset$010$us = 0, $fileoffset$17 = 0, $fileoffset$17$us = 0, $header = 0, $i1$012 = 0, $i1$012$us = 0;
 var $i2$09 = 0, $i2$09$us = 0, $isSRGB$0 = 0, $isSRGB$1 = 0, $isSRGB$10 = 0, $isSRGB$11 = 0, $isSRGB$12 = 0, $isSRGB$13 = 0, $isSRGB$14$ph = 0, $isSRGB$2 = 0, $isSRGB$3 = 0, $isSRGB$4 = 0, $isSRGB$5 = 0, $isSRGB$6 = 0, $isSRGB$7 = 0, $isSRGB$8 = 0, $isSRGB$9 = 0, $mip = 0, $totalsize$0$lcssa = 0, $totalsize$011 = 0;
 var $totalsize$011$us = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer11 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, $vararg_buffer9 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer11 = sp + 48|0;
 $vararg_buffer9 = sp + 40|0;
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $header = sp + 72|0;
 $mip = sp + 56|0;
 $0 = HEAP32[$this>>2]|0; //@line 251 "modules/image/magpie/KTXHandler.cpp"
 $1 = ((($0)) + 16|0); //@line 251 "modules/image/magpie/KTXHandler.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 251 "modules/image/magpie/KTXHandler.cpp"
 $3 = (FUNCTION_TABLE_iii[$2 & 8191]($this,$filedata)|0); //@line 251 "modules/image/magpie/KTXHandler.cpp"
 if (!($3)) {
  $4 = (___cxa_allocate_exception(16)|0); //@line 252 "modules/image/magpie/KTXHandler.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($4|0),(439874|0),($vararg_buffer|0)); //@line 252 "modules/image/magpie/KTXHandler.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   ___cxa_throw(($4|0),(96|0),(344|0)); //@line 252 "modules/image/magpie/KTXHandler.cpp"
   // unreachable; //@line 252 "modules/image/magpie/KTXHandler.cpp"
  }
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  ___cxa_free_exception(($4|0)); //@line 252 "modules/image/magpie/KTXHandler.cpp"
  $$1 = $7;$$12 = $8;
  ___resumeException($$1|0);
  // unreachable;
 }
 $9 = HEAP32[$filedata>>2]|0; //@line 254 "modules/image/magpie/KTXHandler.cpp"
 $10 = ((($9)) + 16|0); //@line 254 "modules/image/magpie/KTXHandler.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 254 "modules/image/magpie/KTXHandler.cpp"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($filedata)|0); //@line 254 "modules/image/magpie/KTXHandler.cpp"
 dest=$header; src=$12; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 254 "modules/image/magpie/KTXHandler.cpp"
 $13 = ((($header)) + 12|0); //@line 256 "modules/image/magpie/KTXHandler.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 324 "modules/image/magpie/KTXHandler.cpp"
 $15 = ($14|0)==(16909060); //@line 256 "modules/image/magpie/KTXHandler.cpp"
 if ($15) {
  $16 = ((($header)) + 16|0); //@line 258 "modules/image/magpie/KTXHandler.cpp"
  $17 = HEAP32[$16>>2]|0; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $18 = (_llvm_bswap_i32(($17|0))|0); //@line 57 "./common/int.h"
  HEAP32[$16>>2] = $18; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $19 = ((($16)) + 4|0); //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $20 = HEAP32[$19>>2]|0; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $21 = (_llvm_bswap_i32(($20|0))|0); //@line 57 "./common/int.h"
  HEAP32[$19>>2] = $21; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $22 = ((($16)) + 8|0); //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $23 = HEAP32[$22>>2]|0; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $24 = (_llvm_bswap_i32(($23|0))|0); //@line 57 "./common/int.h"
  HEAP32[$22>>2] = $24; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $25 = ((($16)) + 12|0); //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $26 = HEAP32[$25>>2]|0; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $27 = (_llvm_bswap_i32(($26|0))|0); //@line 57 "./common/int.h"
  HEAP32[$25>>2] = $27; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $28 = ((($16)) + 16|0); //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $29 = HEAP32[$28>>2]|0; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $30 = (_llvm_bswap_i32(($29|0))|0); //@line 57 "./common/int.h"
  HEAP32[$28>>2] = $30; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $31 = ((($16)) + 20|0); //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $32 = HEAP32[$31>>2]|0; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $33 = (_llvm_bswap_i32(($32|0))|0); //@line 57 "./common/int.h"
  HEAP32[$31>>2] = $33; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $34 = ((($16)) + 24|0); //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $35 = HEAP32[$34>>2]|0; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $36 = (_llvm_bswap_i32(($35|0))|0); //@line 57 "./common/int.h"
  HEAP32[$34>>2] = $36; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $37 = ((($16)) + 28|0); //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $38 = HEAP32[$37>>2]|0; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $39 = (_llvm_bswap_i32(($38|0))|0); //@line 57 "./common/int.h"
  HEAP32[$37>>2] = $39; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $40 = ((($16)) + 32|0); //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $41 = HEAP32[$40>>2]|0; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $42 = (_llvm_bswap_i32(($41|0))|0); //@line 57 "./common/int.h"
  HEAP32[$40>>2] = $42; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $43 = ((($16)) + 36|0); //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $44 = HEAP32[$43>>2]|0; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $45 = (_llvm_bswap_i32(($44|0))|0); //@line 57 "./common/int.h"
  HEAP32[$43>>2] = $45; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $46 = ((($16)) + 40|0); //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $47 = HEAP32[$46>>2]|0; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $48 = (_llvm_bswap_i32(($47|0))|0); //@line 57 "./common/int.h"
  HEAP32[$46>>2] = $48; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $49 = ((($16)) + 44|0); //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $50 = HEAP32[$49>>2]|0; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $51 = (_llvm_bswap_i32(($50|0))|0); //@line 57 "./common/int.h"
  HEAP32[$49>>2] = $51; //@line 260 "modules/image/magpie/KTXHandler.cpp"
  $$pre24 = ((($header)) + 56|0); //@line 263 "modules/image/magpie/KTXHandler.cpp"
  $$pre$phiZ2D = $$pre24;$53 = $48;$55 = $27;
 } else {
  $$phi$trans$insert = ((($header)) + 56|0); //@line 263 "modules/image/magpie/KTXHandler.cpp"
  $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $$phi$trans$insert22 = ((($header)) + 28|0); //@line 266 "modules/image/magpie/KTXHandler.cpp"
  $$pre23 = HEAP32[$$phi$trans$insert22>>2]|0; //@line 266 "modules/image/magpie/KTXHandler.cpp"
  $$pre$phiZ2D = $$phi$trans$insert;$53 = $$pre;$55 = $$pre23;
 }
 $52 = ($53|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $54 = $52 ? 1 : $53; //@line 320 "modules/image/magpie/KTXHandler.cpp"
 HEAP32[$$pre$phiZ2D>>2] = $54; //@line 263 "modules/image/magpie/KTXHandler.cpp"
 do {
  switch ($55|0) {
  case 36196:  {
   $$0$i$ph = 15;$isSRGB$14$ph = 0;
   break;
  }
  case 37488:  {
   $$0$i$ph = 19;$isSRGB$14$ph = 0;
   break;
  }
  case 37489:  {
   $$0$i$ph = 20;$isSRGB$14$ph = 0;
   break;
  }
  case 37490:  {
   $$0$i$ph = 21;$isSRGB$14$ph = 0;
   break;
  }
  case 37491:  {
   $$0$i$ph = 22;$isSRGB$14$ph = 0;
   break;
  }
  case 37492:  {
   $$0$i$ph = 16;$isSRGB$14$ph = 0;
   break;
  }
  case 37493:  {
   $$0$i$ph = 16;$isSRGB$14$ph = 1;
   break;
  }
  case 37494:  {
   $$0$i$ph = 18;$isSRGB$14$ph = 0;
   break;
  }
  case 37495:  {
   $$0$i$ph = 18;$isSRGB$14$ph = 1;
   break;
  }
  case 37496:  {
   $$0$i$ph = 17;$isSRGB$14$ph = 0;
   break;
  }
  case 37497:  {
   $$0$i$ph = 17;$isSRGB$14$ph = 1;
   break;
  }
  case 35840:  {
   $$0$i$ph = 12;$isSRGB$14$ph = 0;
   break;
  }
  case 35841:  {
   $$0$i$ph = 11;$isSRGB$14$ph = 0;
   break;
  }
  case 35842:  {
   $$0$i$ph = 14;$isSRGB$14$ph = 0;
   break;
  }
  case 35843:  {
   $$0$i$ph = 13;$isSRGB$14$ph = 0;
   break;
  }
  case 33776:  {
   $$0$i$ph = 1;$isSRGB$14$ph = 0;
   break;
  }
  case 33778:  {
   $$0$i$ph = 2;$isSRGB$14$ph = 0;
   break;
  }
  case 33779:  {
   $$0$i$ph = 3;$isSRGB$14$ph = 0;
   break;
  }
  case 37840:  {
   $isSRGB$0 = 1;
   label = 27;
   break;
  }
  case 37808:  {
   $isSRGB$0 = 0;
   label = 27;
   break;
  }
  case 37841:  {
   $isSRGB$1 = 1;
   label = 29;
   break;
  }
  case 37809:  {
   $isSRGB$1 = 0;
   label = 29;
   break;
  }
  case 37842:  {
   $isSRGB$2 = 1;
   label = 31;
   break;
  }
  case 37810:  {
   $isSRGB$2 = 0;
   label = 31;
   break;
  }
  case 37843:  {
   $isSRGB$3 = 1;
   label = 33;
   break;
  }
  case 37811:  {
   $isSRGB$3 = 0;
   label = 33;
   break;
  }
  case 37844:  {
   $isSRGB$4 = 1;
   label = 35;
   break;
  }
  case 37812:  {
   $isSRGB$4 = 0;
   label = 35;
   break;
  }
  case 37845:  {
   $isSRGB$5 = 1;
   label = 37;
   break;
  }
  case 37813:  {
   $isSRGB$5 = 0;
   label = 37;
   break;
  }
  case 37846:  {
   $isSRGB$6 = 1;
   label = 39;
   break;
  }
  case 37814:  {
   $isSRGB$6 = 0;
   label = 39;
   break;
  }
  case 37847:  {
   $isSRGB$7 = 1;
   label = 41;
   break;
  }
  case 37815:  {
   $isSRGB$7 = 0;
   label = 41;
   break;
  }
  case 37848:  {
   $isSRGB$8 = 1;
   label = 43;
   break;
  }
  case 37816:  {
   $isSRGB$8 = 0;
   label = 43;
   break;
  }
  case 37849:  {
   $isSRGB$9 = 1;
   label = 45;
   break;
  }
  case 37817:  {
   $isSRGB$9 = 0;
   label = 45;
   break;
  }
  case 37850:  {
   $isSRGB$10 = 1;
   label = 47;
   break;
  }
  case 37818:  {
   $isSRGB$10 = 0;
   label = 47;
   break;
  }
  case 37851:  {
   $isSRGB$11 = 1;
   label = 49;
   break;
  }
  case 37819:  {
   $isSRGB$11 = 0;
   label = 49;
   break;
  }
  case 37852:  {
   $isSRGB$12 = 1;
   label = 51;
   break;
  }
  case 37820:  {
   $isSRGB$12 = 0;
   label = 51;
   break;
  }
  case 37853:  {
   $isSRGB$13 = 1;
   label = 53;
   break;
  }
  case 37821:  {
   $isSRGB$13 = 0;
   label = 53;
   break;
  }
  default: {
   $56 = (___cxa_allocate_exception(16)|0); //@line 269 "modules/image/magpie/KTXHandler.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($56|0),(439925|0),($vararg_buffer1|0)); //@line 269 "modules/image/magpie/KTXHandler.cpp"
   $57 = __THREW__; __THREW__ = 0;
   $58 = $57&1;
   if (!($58)) {
    ___cxa_throw(($56|0),(96|0),(344|0)); //@line 269 "modules/image/magpie/KTXHandler.cpp"
    // unreachable; //@line 269 "modules/image/magpie/KTXHandler.cpp"
   }
   $59 = ___cxa_find_matching_catch_2()|0;
   $60 = tempRet0;
   ___cxa_free_exception(($56|0)); //@line 269 "modules/image/magpie/KTXHandler.cpp"
   $$1 = $59;$$12 = $60;
   ___resumeException($$1|0);
   // unreachable;
  }
  }
 } while(0);
 switch (label|0) {
  case 27: {
   $$0$i$ph = 23;$isSRGB$14$ph = $isSRGB$0;
   break;
  }
  case 29: {
   $$0$i$ph = 24;$isSRGB$14$ph = $isSRGB$1;
   break;
  }
  case 31: {
   $$0$i$ph = 25;$isSRGB$14$ph = $isSRGB$2;
   break;
  }
  case 33: {
   $$0$i$ph = 26;$isSRGB$14$ph = $isSRGB$3;
   break;
  }
  case 35: {
   $$0$i$ph = 27;$isSRGB$14$ph = $isSRGB$4;
   break;
  }
  case 37: {
   $$0$i$ph = 28;$isSRGB$14$ph = $isSRGB$5;
   break;
  }
  case 39: {
   $$0$i$ph = 29;$isSRGB$14$ph = $isSRGB$6;
   break;
  }
  case 41: {
   $$0$i$ph = 30;$isSRGB$14$ph = $isSRGB$7;
   break;
  }
  case 43: {
   $$0$i$ph = 31;$isSRGB$14$ph = $isSRGB$8;
   break;
  }
  case 45: {
   $$0$i$ph = 32;$isSRGB$14$ph = $isSRGB$9;
   break;
  }
  case 47: {
   $$0$i$ph = 33;$isSRGB$14$ph = $isSRGB$10;
   break;
  }
  case 49: {
   $$0$i$ph = 34;$isSRGB$14$ph = $isSRGB$11;
   break;
  }
  case 51: {
   $$0$i$ph = 35;$isSRGB$14$ph = $isSRGB$12;
   break;
  }
  case 53: {
   $$0$i$ph = 36;$isSRGB$14$ph = $isSRGB$13;
   break;
  }
 }
 $61 = ((($header)) + 48|0); //@line 271 "modules/image/magpie/KTXHandler.cpp"
 $62 = HEAP32[$61>>2]|0; //@line 271 "modules/image/magpie/KTXHandler.cpp"
 $63 = ($62|0)==(0); //@line 271 "modules/image/magpie/KTXHandler.cpp"
 if (!($63)) {
  $64 = (___cxa_allocate_exception(16)|0); //@line 272 "modules/image/magpie/KTXHandler.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($64|0),(439963|0),($vararg_buffer3|0)); //@line 272 "modules/image/magpie/KTXHandler.cpp"
  $65 = __THREW__; __THREW__ = 0;
  $66 = $65&1;
  if (!($66)) {
   ___cxa_throw(($64|0),(96|0),(344|0)); //@line 272 "modules/image/magpie/KTXHandler.cpp"
   // unreachable; //@line 272 "modules/image/magpie/KTXHandler.cpp"
  }
  $67 = ___cxa_find_matching_catch_2()|0;
  $68 = tempRet0;
  ___cxa_free_exception(($64|0)); //@line 272 "modules/image/magpie/KTXHandler.cpp"
  $$1 = $67;$$12 = $68;
  ___resumeException($$1|0);
  // unreachable;
 }
 $69 = ((($header)) + 44|0); //@line 274 "modules/image/magpie/KTXHandler.cpp"
 $70 = HEAP32[$69>>2]|0; //@line 274 "modules/image/magpie/KTXHandler.cpp"
 $71 = ($70>>>0)>(1); //@line 274 "modules/image/magpie/KTXHandler.cpp"
 if ($71) {
  $72 = (___cxa_allocate_exception(16)|0); //@line 275 "modules/image/magpie/KTXHandler.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($72|0),(440010|0),($vararg_buffer5|0)); //@line 275 "modules/image/magpie/KTXHandler.cpp"
  $73 = __THREW__; __THREW__ = 0;
  $74 = $73&1;
  if (!($74)) {
   ___cxa_throw(($72|0),(96|0),(344|0)); //@line 275 "modules/image/magpie/KTXHandler.cpp"
   // unreachable; //@line 275 "modules/image/magpie/KTXHandler.cpp"
  }
  $75 = ___cxa_find_matching_catch_2()|0;
  $76 = tempRet0;
  ___cxa_free_exception(($72|0)); //@line 275 "modules/image/magpie/KTXHandler.cpp"
  $$1 = $75;$$12 = $76;
  ___resumeException($$1|0);
  // unreachable;
 }
 $77 = ((($header)) + 52|0); //@line 277 "modules/image/magpie/KTXHandler.cpp"
 $78 = HEAP32[$77>>2]|0; //@line 277 "modules/image/magpie/KTXHandler.cpp"
 $79 = ($78>>>0)>(1); //@line 277 "modules/image/magpie/KTXHandler.cpp"
 if ($79) {
  $80 = (___cxa_allocate_exception(16)|0); //@line 278 "modules/image/magpie/KTXHandler.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($80|0),(440054|0),($vararg_buffer7|0)); //@line 278 "modules/image/magpie/KTXHandler.cpp"
  $81 = __THREW__; __THREW__ = 0;
  $82 = $81&1;
  if (!($82)) {
   ___cxa_throw(($80|0),(96|0),(344|0)); //@line 278 "modules/image/magpie/KTXHandler.cpp"
   // unreachable; //@line 278 "modules/image/magpie/KTXHandler.cpp"
  }
  $83 = ___cxa_find_matching_catch_2()|0;
  $84 = tempRet0;
  ___cxa_free_exception(($80|0)); //@line 278 "modules/image/magpie/KTXHandler.cpp"
  $$1 = $83;$$12 = $84;
  ___resumeException($$1|0);
  // unreachable;
 }
 $85 = ((($header)) + 60|0); //@line 280 "modules/image/magpie/KTXHandler.cpp"
 $86 = HEAP32[$85>>2]|0; //@line 316 "modules/image/magpie/KTXHandler.cpp"
 $87 = (($86) + 64)|0; //@line 280 "modules/image/magpie/KTXHandler.cpp"
 $88 = HEAP32[$filedata>>2]|0; //@line 281 "modules/image/magpie/KTXHandler.cpp"
 $89 = ((($88)) + 16|0); //@line 281 "modules/image/magpie/KTXHandler.cpp"
 $90 = HEAP32[$89>>2]|0; //@line 281 "modules/image/magpie/KTXHandler.cpp"
 $91 = (FUNCTION_TABLE_ii[$90 & 8191]($filedata)|0); //@line 281 "modules/image/magpie/KTXHandler.cpp"
 $92 = ($54|0)>(0); //@line 285 "modules/image/magpie/KTXHandler.cpp"
 L86: do {
  if ($92) {
   L88: do {
    if ($15) {
     $fileoffset$010$us = $87;$i1$012$us = 0;$totalsize$011$us = 0;
     while(1) {
      $93 = (($fileoffset$010$us) + 4)|0; //@line 287 "modules/image/magpie/KTXHandler.cpp"
      $94 = HEAP32[$filedata>>2]|0; //@line 287 "modules/image/magpie/KTXHandler.cpp"
      $95 = ((($94)) + 20|0); //@line 287 "modules/image/magpie/KTXHandler.cpp"
      $96 = HEAP32[$95>>2]|0; //@line 287 "modules/image/magpie/KTXHandler.cpp"
      $97 = (FUNCTION_TABLE_ii[$96 & 8191]($filedata)|0); //@line 287 "modules/image/magpie/KTXHandler.cpp"
      $98 = ($93>>>0)>($97>>>0); //@line 287 "modules/image/magpie/KTXHandler.cpp"
      if ($98) {
       break L88;
      }
      $99 = (($91) + ($fileoffset$010$us)|0); //@line 290 "modules/image/magpie/KTXHandler.cpp"
      $100 = HEAP32[$99>>2]|0; //@line 290 "modules/image/magpie/KTXHandler.cpp"
      $101 = (_llvm_bswap_i32(($100|0))|0); //@line 57 "./common/int.h"
      $102 = (($101) + 3)|0; //@line 299 "modules/image/magpie/KTXHandler.cpp"
      $103 = $102 & -4; //@line 299 "modules/image/magpie/KTXHandler.cpp"
      $104 = (($103) + ($totalsize$011$us))|0; //@line 301 "modules/image/magpie/KTXHandler.cpp"
      $105 = (($103) + ($93))|0; //@line 302 "modules/image/magpie/KTXHandler.cpp"
      $106 = (($i1$012$us) + 1)|0; //@line 285 "modules/image/magpie/KTXHandler.cpp"
      $107 = ($106|0)<($54|0); //@line 285 "modules/image/magpie/KTXHandler.cpp"
      if ($107) {
       $fileoffset$010$us = $105;$i1$012$us = $106;$totalsize$011$us = $104;
      } else {
       $totalsize$0$lcssa = $104;
       break L86;
      }
     }
    } else {
     $fileoffset$010 = $87;$i1$012 = 0;$totalsize$011 = 0;
     while(1) {
      $111 = (($fileoffset$010) + 4)|0; //@line 287 "modules/image/magpie/KTXHandler.cpp"
      $112 = HEAP32[$filedata>>2]|0; //@line 287 "modules/image/magpie/KTXHandler.cpp"
      $113 = ((($112)) + 20|0); //@line 287 "modules/image/magpie/KTXHandler.cpp"
      $114 = HEAP32[$113>>2]|0; //@line 287 "modules/image/magpie/KTXHandler.cpp"
      $115 = (FUNCTION_TABLE_ii[$114 & 8191]($filedata)|0); //@line 287 "modules/image/magpie/KTXHandler.cpp"
      $116 = ($111>>>0)>($115>>>0); //@line 287 "modules/image/magpie/KTXHandler.cpp"
      if ($116) {
       break L88;
      }
      $122 = (($91) + ($fileoffset$010)|0); //@line 290 "modules/image/magpie/KTXHandler.cpp"
      $123 = HEAP32[$122>>2]|0; //@line 290 "modules/image/magpie/KTXHandler.cpp"
      $124 = (($123) + 3)|0; //@line 299 "modules/image/magpie/KTXHandler.cpp"
      $125 = $124 & -4; //@line 299 "modules/image/magpie/KTXHandler.cpp"
      $126 = (($125) + ($totalsize$011))|0; //@line 301 "modules/image/magpie/KTXHandler.cpp"
      $127 = (($125) + ($111))|0; //@line 302 "modules/image/magpie/KTXHandler.cpp"
      $128 = (($i1$012) + 1)|0; //@line 285 "modules/image/magpie/KTXHandler.cpp"
      $129 = ($128|0)<($54|0); //@line 285 "modules/image/magpie/KTXHandler.cpp"
      if ($129) {
       $fileoffset$010 = $127;$i1$012 = $128;$totalsize$011 = $126;
      } else {
       $totalsize$0$lcssa = $126;
       break L86;
      }
     }
    }
   } while(0);
   $117 = (___cxa_allocate_exception(16)|0); //@line 288 "modules/image/magpie/KTXHandler.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($117|0),(440103|0),($vararg_buffer9|0)); //@line 288 "modules/image/magpie/KTXHandler.cpp"
   $118 = __THREW__; __THREW__ = 0;
   $119 = $118&1;
   if (!($119)) {
    ___cxa_throw(($117|0),(96|0),(344|0)); //@line 288 "modules/image/magpie/KTXHandler.cpp"
    // unreachable; //@line 288 "modules/image/magpie/KTXHandler.cpp"
   }
   $120 = ___cxa_find_matching_catch_2()|0;
   $121 = tempRet0;
   ___cxa_free_exception(($117|0)); //@line 288 "modules/image/magpie/KTXHandler.cpp"
   $$1 = $120;$$12 = $121;
   ___resumeException($$1|0);
   // unreachable;
  } else {
   $totalsize$0$lcssa = 0;
  }
 } while(0);
 __THREW__ = 0;
 $108 = (invoke_ii(3492,($totalsize$0$lcssa|0))|0); //@line 308 "modules/image/magpie/KTXHandler.cpp"
 $109 = __THREW__; __THREW__ = 0;
 $110 = $109&1;
 if ($110) {
  $162 = ___cxa_find_matching_catch_3(9056|0)|0;
  $163 = tempRet0;
  $164 = (_llvm_eh_typeid_for((9056|0))|0); //@line 309 "modules/image/magpie/KTXHandler.cpp"
  $165 = ($163|0)==($164|0); //@line 309 "modules/image/magpie/KTXHandler.cpp"
  if (!($165)) {
   $$1 = $162;$$12 = $163;
   ___resumeException($$1|0);
   // unreachable;
  }
  (___cxa_begin_catch(($162|0))|0); //@line 309 "modules/image/magpie/KTXHandler.cpp"
  $166 = (___cxa_allocate_exception(16)|0); //@line 312 "modules/image/magpie/KTXHandler.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($166|0),(507610|0),($vararg_buffer11|0)); //@line 312 "modules/image/magpie/KTXHandler.cpp"
  $167 = __THREW__; __THREW__ = 0;
  $168 = $167&1;
  if ($168) {
   $172 = ___cxa_find_matching_catch_2()|0;
   $173 = tempRet0;
   ___cxa_free_exception(($166|0)); //@line 312 "modules/image/magpie/KTXHandler.cpp"
   $$0 = $172;$$01 = $173;
  } else {
   __THREW__ = 0;
   invoke_viii(3350,($166|0),(96|0),(344|0)); //@line 312 "modules/image/magpie/KTXHandler.cpp"
   $169 = __THREW__; __THREW__ = 0;
   $170 = ___cxa_find_matching_catch_2()|0;
   $171 = tempRet0;
   $$0 = $170;$$01 = $171;
  }
  __THREW__ = 0;
  invoke_v(3378); //@line 313 "modules/image/magpie/KTXHandler.cpp"
  $174 = __THREW__; __THREW__ = 0;
  $175 = $174&1;
  if ($175) {
   $198 = ___cxa_find_matching_catch_3(0|0)|0;
   $199 = tempRet0;
   ___clang_call_terminate($198); //@line 313 "modules/image/magpie/KTXHandler.cpp"
   // unreachable; //@line 313 "modules/image/magpie/KTXHandler.cpp"
  } else {
   $$1 = $$0;$$12 = $$01;
   ___resumeException($$1|0);
   // unreachable;
  }
 }
 if (!($92)) {
  HEAP32[$dataSize>>2] = $totalsize$0$lcssa; //@line 345 "modules/image/magpie/KTXHandler.cpp"
  HEAP32[$format>>2] = $$0$i$ph; //@line 346 "modules/image/magpie/KTXHandler.cpp"
  HEAP8[$sRGB>>0] = $isSRGB$14$ph; //@line 347 "modules/image/magpie/KTXHandler.cpp"
  STACKTOP = sp;return ($108|0); //@line 349 "modules/image/magpie/KTXHandler.cpp"
 }
 $130 = ((($header)) + 36|0); //@line 332 "modules/image/magpie/KTXHandler.cpp"
 $131 = HEAP32[$130>>2]|0; //@line 332 "modules/image/magpie/KTXHandler.cpp"
 $132 = ((($header)) + 40|0); //@line 333 "modules/image/magpie/KTXHandler.cpp"
 $133 = HEAP32[$132>>2]|0; //@line 333 "modules/image/magpie/KTXHandler.cpp"
 $134 = ((($mip)) + 4|0); //@line 333 "modules/image/magpie/KTXHandler.cpp"
 $135 = ((($mip)) + 8|0); //@line 334 "modules/image/magpie/KTXHandler.cpp"
 $136 = ((($mip)) + 12|0); //@line 337 "modules/image/magpie/KTXHandler.cpp"
 $137 = ((($images)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $138 = ((($images)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($15) {
  $dataoffset$08$us = 0;$fileoffset$17$us = $87;$i2$09$us = 0;
  while(1) {
   $139 = (($91) + ($fileoffset$17$us)|0); //@line 322 "modules/image/magpie/KTXHandler.cpp"
   $140 = HEAP32[$139>>2]|0; //@line 322 "modules/image/magpie/KTXHandler.cpp"
   $141 = (_llvm_bswap_i32(($140|0))|0); //@line 57 "./common/int.h"
   $142 = (($fileoffset$17$us) + 4)|0; //@line 327 "modules/image/magpie/KTXHandler.cpp"
   $143 = (($141) + 3)|0; //@line 329 "modules/image/magpie/KTXHandler.cpp"
   $144 = $143 & -4; //@line 329 "modules/image/magpie/KTXHandler.cpp"
   $145 = $131 >>> $i2$09$us; //@line 332 "modules/image/magpie/KTXHandler.cpp"
   $146 = ($145|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $147 = $146 ? 1 : $145; //@line 332 "modules/image/magpie/KTXHandler.cpp"
   HEAP32[$mip>>2] = $147; //@line 332 "modules/image/magpie/KTXHandler.cpp"
   $148 = $133 >>> $i2$09$us; //@line 333 "modules/image/magpie/KTXHandler.cpp"
   $149 = ($148|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $150 = $149 ? 1 : $148; //@line 333 "modules/image/magpie/KTXHandler.cpp"
   HEAP32[$134>>2] = $150; //@line 333 "modules/image/magpie/KTXHandler.cpp"
   HEAP32[$135>>2] = $141; //@line 334 "modules/image/magpie/KTXHandler.cpp"
   $151 = (($108) + ($dataoffset$08$us)|0); //@line 336 "modules/image/magpie/KTXHandler.cpp"
   $152 = (($91) + ($142)|0); //@line 336 "modules/image/magpie/KTXHandler.cpp"
   _memcpy(($151|0),($152|0),($141|0))|0; //@line 336 "modules/image/magpie/KTXHandler.cpp"
   HEAP32[$136>>2] = $151; //@line 337 "modules/image/magpie/KTXHandler.cpp"
   $153 = (($144) + ($142))|0; //@line 339 "modules/image/magpie/KTXHandler.cpp"
   $154 = (($144) + ($dataoffset$08$us))|0; //@line 340 "modules/image/magpie/KTXHandler.cpp"
   $155 = HEAP32[$137>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $156 = HEAP32[$138>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $157 = ($155|0)==($156|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($157) {
    __ZNSt3__16vectorIN4love5image19CompressedImageData8SubImageENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_($images,$mip); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   } else {
    ;HEAP32[$155>>2]=HEAP32[$mip>>2]|0;HEAP32[$155+4>>2]=HEAP32[$mip+4>>2]|0;HEAP32[$155+8>>2]=HEAP32[$mip+8>>2]|0;HEAP32[$155+12>>2]=HEAP32[$mip+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $158 = HEAP32[$137>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $159 = ((($158)) + 16|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$137>>2] = $159; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $160 = (($i2$09$us) + 1)|0; //@line 320 "modules/image/magpie/KTXHandler.cpp"
   $161 = ($160|0)<($54|0); //@line 320 "modules/image/magpie/KTXHandler.cpp"
   if ($161) {
    $dataoffset$08$us = $154;$fileoffset$17$us = $153;$i2$09$us = $160;
   } else {
    break;
   }
  }
  HEAP32[$dataSize>>2] = $totalsize$0$lcssa; //@line 345 "modules/image/magpie/KTXHandler.cpp"
  HEAP32[$format>>2] = $$0$i$ph; //@line 346 "modules/image/magpie/KTXHandler.cpp"
  HEAP8[$sRGB>>0] = $isSRGB$14$ph; //@line 347 "modules/image/magpie/KTXHandler.cpp"
  STACKTOP = sp;return ($108|0); //@line 349 "modules/image/magpie/KTXHandler.cpp"
 } else {
  $dataoffset$08 = 0;$fileoffset$17 = $87;$i2$09 = 0;
  while(1) {
   $176 = (($91) + ($fileoffset$17)|0); //@line 322 "modules/image/magpie/KTXHandler.cpp"
   $177 = HEAP32[$176>>2]|0; //@line 322 "modules/image/magpie/KTXHandler.cpp"
   $178 = (($fileoffset$17) + 4)|0; //@line 327 "modules/image/magpie/KTXHandler.cpp"
   $179 = (($177) + 3)|0; //@line 329 "modules/image/magpie/KTXHandler.cpp"
   $180 = $179 & -4; //@line 329 "modules/image/magpie/KTXHandler.cpp"
   $181 = $131 >>> $i2$09; //@line 332 "modules/image/magpie/KTXHandler.cpp"
   $182 = ($181|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $183 = $182 ? 1 : $181; //@line 332 "modules/image/magpie/KTXHandler.cpp"
   HEAP32[$mip>>2] = $183; //@line 332 "modules/image/magpie/KTXHandler.cpp"
   $184 = $133 >>> $i2$09; //@line 333 "modules/image/magpie/KTXHandler.cpp"
   $185 = ($184|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $186 = $185 ? 1 : $184; //@line 333 "modules/image/magpie/KTXHandler.cpp"
   HEAP32[$134>>2] = $186; //@line 333 "modules/image/magpie/KTXHandler.cpp"
   HEAP32[$135>>2] = $177; //@line 334 "modules/image/magpie/KTXHandler.cpp"
   $187 = (($108) + ($dataoffset$08)|0); //@line 336 "modules/image/magpie/KTXHandler.cpp"
   $188 = (($91) + ($178)|0); //@line 336 "modules/image/magpie/KTXHandler.cpp"
   _memcpy(($187|0),($188|0),($177|0))|0; //@line 336 "modules/image/magpie/KTXHandler.cpp"
   HEAP32[$136>>2] = $187; //@line 337 "modules/image/magpie/KTXHandler.cpp"
   $189 = (($180) + ($178))|0; //@line 339 "modules/image/magpie/KTXHandler.cpp"
   $190 = (($180) + ($dataoffset$08))|0; //@line 340 "modules/image/magpie/KTXHandler.cpp"
   $191 = HEAP32[$137>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $192 = HEAP32[$138>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $193 = ($191|0)==($192|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($193) {
    __ZNSt3__16vectorIN4love5image19CompressedImageData8SubImageENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_($images,$mip); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   } else {
    ;HEAP32[$191>>2]=HEAP32[$mip>>2]|0;HEAP32[$191+4>>2]=HEAP32[$mip+4>>2]|0;HEAP32[$191+8>>2]=HEAP32[$mip+8>>2]|0;HEAP32[$191+12>>2]=HEAP32[$mip+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $194 = HEAP32[$137>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $195 = ((($194)) + 16|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$137>>2] = $195; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $196 = (($i2$09) + 1)|0; //@line 320 "modules/image/magpie/KTXHandler.cpp"
   $197 = ($196|0)<($54|0); //@line 320 "modules/image/magpie/KTXHandler.cpp"
   if ($197) {
    $dataoffset$08 = $190;$fileoffset$17 = $189;$i2$09 = $196;
   } else {
    break;
   }
  }
  HEAP32[$dataSize>>2] = $totalsize$0$lcssa; //@line 345 "modules/image/magpie/KTXHandler.cpp"
  HEAP32[$format>>2] = $$0$i$ph; //@line 346 "modules/image/magpie/KTXHandler.cpp"
  HEAP8[$sRGB>>0] = $isSRGB$14$ph; //@line 347 "modules/image/magpie/KTXHandler.cpp"
  STACKTOP = sp;return ($108|0); //@line 349 "modules/image/magpie/KTXHandler.cpp"
 }
 return (0)|0;
}
function __ZNSt3__16vectorIN4love5image19CompressedImageData8SubImageENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(268435455); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 4; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(134217727); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 3; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$32 = $18;
  } else {
   $$0$i1 = $17;$40 = $18;$41 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 268435455;$40 = $12;$41 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 4; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $41;$32 = $40;
 }
 $24 = (($25) + ($26<<4)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<4)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 ;HEAP32[$24>>2]=HEAP32[$__x>>2]|0;HEAP32[$24+4>>2]=HEAP32[$__x+4>>2]|0;HEAP32[$24+8>>2]=HEAP32[$__x+8>>2]|0;HEAP32[$24+12>>2]=HEAP32[$__x+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $29 = ((($24)) + 16|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = $29; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = (($32) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = $31 >> 4; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = (0 - ($33))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (($24) + ($34<<4)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = $35; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($35|0),($37|0),($31|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $36; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $38 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($38) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $39 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($39); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love5image6magpie10PKMHandlerD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 41 "modules/image/magpie/PKMHandler.h"
 return; //@line 41 "modules/image/magpie/PKMHandler.h"
}
function __ZN4love5image6magpie10PKMHandler8canParseEPKNS_10filesystem8FileDataE($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$mask = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $switch = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$data>>2]|0; //@line 101 "modules/image/magpie/PKMHandler.cpp"
 $1 = ((($0)) + 20|0); //@line 101 "modules/image/magpie/PKMHandler.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 101 "modules/image/magpie/PKMHandler.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($data)|0); //@line 101 "modules/image/magpie/PKMHandler.cpp"
 $4 = ($3>>>0)<(17); //@line 101 "modules/image/magpie/PKMHandler.cpp"
 if ($4) {
  return 0; //@line 114 "modules/image/magpie/PKMHandler.cpp"
 }
 $5 = HEAP32[$data>>2]|0; //@line 104 "modules/image/magpie/PKMHandler.cpp"
 $6 = ((($5)) + 16|0); //@line 104 "modules/image/magpie/PKMHandler.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 104 "modules/image/magpie/PKMHandler.cpp"
 $8 = (FUNCTION_TABLE_ii[$7 & 8191]($data)|0); //@line 104 "modules/image/magpie/PKMHandler.cpp"
 $9 = (_memcmp($8,440178,4)|0); //@line 106 "modules/image/magpie/PKMHandler.cpp"
 $10 = ($9|0)==(0); //@line 106 "modules/image/magpie/PKMHandler.cpp"
 if (!($10)) {
  return 0; //@line 114 "modules/image/magpie/PKMHandler.cpp"
 }
 $11 = ((($8)) + 4|0); //@line 110 "modules/image/magpie/PKMHandler.cpp"
 $12 = HEAP16[$11>>1]|0; //@line 110 "modules/image/magpie/PKMHandler.cpp"
 $13 = $12&255; //@line 110 "modules/image/magpie/PKMHandler.cpp"
 $$off = (($13) + -49)<<24>>24; //@line 110 "modules/image/magpie/PKMHandler.cpp"
 $switch = ($$off&255)<(2); //@line 110 "modules/image/magpie/PKMHandler.cpp"
 if ($switch) {
  $$mask = $12 & -256; //@line 110 "modules/image/magpie/PKMHandler.cpp"
  $14 = ($$mask<<16>>16)==(12288); //@line 110 "modules/image/magpie/PKMHandler.cpp"
  return ($14|0); //@line 110 "modules/image/magpie/PKMHandler.cpp"
 } else {
  return 0; //@line 114 "modules/image/magpie/PKMHandler.cpp"
 }
 return (0)|0;
}
function __ZN4love5image6magpie10PKMHandler5parseEPNS_10filesystem8FileDataERNSt3__16vectorINS0_19CompressedImageData8SubImageENS6_9allocatorIS9_EEEERjRNS8_6FormatERb($this,$filedata,$images,$dataSize,$format,$sRGB) {
 $this = $this|0;
 $filedata = $filedata|0;
 $images = $images|0;
 $dataSize = $dataSize|0;
 $format = $format|0;
 $sRGB = $sRGB|0;
 var $$0 = 0, $$0$i$ph = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$masked = 0, $$masked4 = 0, $$masked5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $8 = 0, $9 = 0, $mip = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $mip = sp + 24|0;
 $0 = HEAP32[$this>>2]|0; //@line 118 "modules/image/magpie/PKMHandler.cpp"
 $1 = ((($0)) + 16|0); //@line 118 "modules/image/magpie/PKMHandler.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 118 "modules/image/magpie/PKMHandler.cpp"
 $3 = (FUNCTION_TABLE_iii[$2 & 8191]($this,$filedata)|0); //@line 118 "modules/image/magpie/PKMHandler.cpp"
 if (!($3)) {
  $4 = (___cxa_allocate_exception(16)|0); //@line 119 "modules/image/magpie/PKMHandler.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($4|0),(440182|0),($vararg_buffer|0)); //@line 119 "modules/image/magpie/PKMHandler.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   ___cxa_throw(($4|0),(96|0),(344|0)); //@line 119 "modules/image/magpie/PKMHandler.cpp"
   // unreachable; //@line 119 "modules/image/magpie/PKMHandler.cpp"
  }
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  ___cxa_free_exception(($4|0)); //@line 119 "modules/image/magpie/PKMHandler.cpp"
  $$1 = $7;$$12 = $8;
  ___resumeException($$1|0);
  // unreachable;
 }
 $9 = HEAP32[$filedata>>2]|0; //@line 121 "modules/image/magpie/PKMHandler.cpp"
 $10 = ((($9)) + 16|0); //@line 121 "modules/image/magpie/PKMHandler.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 121 "modules/image/magpie/PKMHandler.cpp"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($filedata)|0); //@line 121 "modules/image/magpie/PKMHandler.cpp"
 $13 = ((($12)) + 6|0); //@line 121 "modules/image/magpie/PKMHandler.cpp"
 $14 = HEAP16[$13>>1]|0; //@line 121 "modules/image/magpie/PKMHandler.cpp"
 $15 = ((($12)) + 12|0); //@line 121 "modules/image/magpie/PKMHandler.cpp"
 $16 = HEAP16[$15>>1]|0; //@line 121 "modules/image/magpie/PKMHandler.cpp"
 $17 = ((($12)) + 14|0); //@line 121 "modules/image/magpie/PKMHandler.cpp"
 $18 = HEAP16[$17>>1]|0; //@line 121 "modules/image/magpie/PKMHandler.cpp"
 $19 = $14&65535; //@line 50 "./common/int.h"
 $20 = $19 >>> 8; //@line 50 "./common/int.h"
 $21 = $19 << 8; //@line 50 "./common/int.h"
 $22 = $16&65535; //@line 50 "./common/int.h"
 $23 = $22 >>> 8; //@line 50 "./common/int.h"
 $24 = $22 << 8; //@line 50 "./common/int.h"
 $25 = $18&65535; //@line 50 "./common/int.h"
 $26 = $25 >>> 8; //@line 50 "./common/int.h"
 $27 = $25 << 8; //@line 50 "./common/int.h"
 $$masked = $21 & 65280; //@line 73 "modules/image/magpie/PKMHandler.cpp"
 $28 = $$masked | $20; //@line 73 "modules/image/magpie/PKMHandler.cpp"
 switch ($28|0) {
 case 0:  {
  $$0$i$ph = 15;
  break;
 }
 case 1:  {
  $$0$i$ph = 16;
  break;
 }
 case 3: case 2:  {
  $$0$i$ph = 17;
  break;
 }
 case 4:  {
  $$0$i$ph = 18;
  break;
 }
 case 5:  {
  $$0$i$ph = 19;
  break;
 }
 case 6:  {
  $$0$i$ph = 21;
  break;
 }
 case 7:  {
  $$0$i$ph = 20;
  break;
 }
 case 8:  {
  $$0$i$ph = 22;
  break;
 }
 default: {
  $29 = (___cxa_allocate_exception(16)|0); //@line 132 "modules/image/magpie/PKMHandler.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($29|0),(440233|0),($vararg_buffer1|0)); //@line 132 "modules/image/magpie/PKMHandler.cpp"
  $30 = __THREW__; __THREW__ = 0;
  $31 = $30&1;
  if (!($31)) {
   ___cxa_throw(($29|0),(96|0),(344|0)); //@line 132 "modules/image/magpie/PKMHandler.cpp"
   // unreachable; //@line 132 "modules/image/magpie/PKMHandler.cpp"
  }
  $32 = ___cxa_find_matching_catch_2()|0;
  $33 = tempRet0;
  ___cxa_free_exception(($29|0)); //@line 132 "modules/image/magpie/PKMHandler.cpp"
  $$1 = $32;$$12 = $33;
  ___resumeException($$1|0);
  // unreachable;
 }
 }
 $34 = HEAP32[$filedata>>2]|0; //@line 135 "modules/image/magpie/PKMHandler.cpp"
 $35 = ((($34)) + 20|0); //@line 135 "modules/image/magpie/PKMHandler.cpp"
 $36 = HEAP32[$35>>2]|0; //@line 135 "modules/image/magpie/PKMHandler.cpp"
 $37 = (FUNCTION_TABLE_ii[$36 & 8191]($filedata)|0); //@line 135 "modules/image/magpie/PKMHandler.cpp"
 $38 = (($37) + -16)|0; //@line 135 "modules/image/magpie/PKMHandler.cpp"
 __THREW__ = 0;
 $39 = (invoke_ii(3492,($38|0))|0); //@line 140 "modules/image/magpie/PKMHandler.cpp"
 $40 = __THREW__; __THREW__ = 0;
 $41 = $40&1;
 if (!($41)) {
  $42 = HEAP32[$filedata>>2]|0; //@line 148 "modules/image/magpie/PKMHandler.cpp"
  $43 = ((($42)) + 16|0); //@line 148 "modules/image/magpie/PKMHandler.cpp"
  $44 = HEAP32[$43>>2]|0; //@line 148 "modules/image/magpie/PKMHandler.cpp"
  $45 = (FUNCTION_TABLE_ii[$44 & 8191]($filedata)|0); //@line 148 "modules/image/magpie/PKMHandler.cpp"
  $46 = ((($45)) + 16|0); //@line 148 "modules/image/magpie/PKMHandler.cpp"
  _memcpy(($39|0),($46|0),($38|0))|0; //@line 148 "modules/image/magpie/PKMHandler.cpp"
  $$masked4 = $24 & 65280; //@line 154 "modules/image/magpie/PKMHandler.cpp"
  $47 = $$masked4 | $23; //@line 154 "modules/image/magpie/PKMHandler.cpp"
  HEAP32[$mip>>2] = $47; //@line 154 "modules/image/magpie/PKMHandler.cpp"
  $$masked5 = $27 & 65280; //@line 155 "modules/image/magpie/PKMHandler.cpp"
  $48 = $$masked5 | $26; //@line 155 "modules/image/magpie/PKMHandler.cpp"
  $49 = ((($mip)) + 4|0); //@line 155 "modules/image/magpie/PKMHandler.cpp"
  HEAP32[$49>>2] = $48; //@line 155 "modules/image/magpie/PKMHandler.cpp"
  $50 = ((($mip)) + 8|0); //@line 157 "modules/image/magpie/PKMHandler.cpp"
  HEAP32[$50>>2] = $38; //@line 157 "modules/image/magpie/PKMHandler.cpp"
  $51 = ((($mip)) + 12|0); //@line 158 "modules/image/magpie/PKMHandler.cpp"
  HEAP32[$51>>2] = $39; //@line 158 "modules/image/magpie/PKMHandler.cpp"
  $52 = ((($images)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $53 = HEAP32[$52>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $54 = ((($images)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $55 = HEAP32[$54>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $56 = ($53|0)==($55|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($56) {
   __ZNSt3__16vectorIN4love5image19CompressedImageData8SubImageENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_($images,$mip); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$dataSize>>2] = $38; //@line 162 "modules/image/magpie/PKMHandler.cpp"
   HEAP32[$format>>2] = $$0$i$ph; //@line 163 "modules/image/magpie/PKMHandler.cpp"
   HEAP8[$sRGB>>0] = 0; //@line 164 "modules/image/magpie/PKMHandler.cpp"
   STACKTOP = sp;return ($39|0); //@line 166 "modules/image/magpie/PKMHandler.cpp"
  } else {
   ;HEAP32[$53>>2]=HEAP32[$mip>>2]|0;HEAP32[$53+4>>2]=HEAP32[$mip+4>>2]|0;HEAP32[$53+8>>2]=HEAP32[$mip+8>>2]|0;HEAP32[$53+12>>2]=HEAP32[$mip+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $57 = HEAP32[$52>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $58 = ((($57)) + 16|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$52>>2] = $58; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$dataSize>>2] = $38; //@line 162 "modules/image/magpie/PKMHandler.cpp"
   HEAP32[$format>>2] = $$0$i$ph; //@line 163 "modules/image/magpie/PKMHandler.cpp"
   HEAP8[$sRGB>>0] = 0; //@line 164 "modules/image/magpie/PKMHandler.cpp"
   STACKTOP = sp;return ($39|0); //@line 166 "modules/image/magpie/PKMHandler.cpp"
  }
 }
 $59 = ___cxa_find_matching_catch_3(9056|0)|0;
 $60 = tempRet0;
 $61 = (_llvm_eh_typeid_for((9056|0))|0); //@line 141 "modules/image/magpie/PKMHandler.cpp"
 $62 = ($60|0)==($61|0); //@line 141 "modules/image/magpie/PKMHandler.cpp"
 if (!($62)) {
  $$1 = $59;$$12 = $60;
  ___resumeException($$1|0);
  // unreachable;
 }
 (___cxa_begin_catch(($59|0))|0); //@line 141 "modules/image/magpie/PKMHandler.cpp"
 $63 = (___cxa_allocate_exception(16)|0); //@line 144 "modules/image/magpie/PKMHandler.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($63|0),(507610|0),($vararg_buffer3|0)); //@line 144 "modules/image/magpie/PKMHandler.cpp"
 $64 = __THREW__; __THREW__ = 0;
 $65 = $64&1;
 if ($65) {
  $69 = ___cxa_find_matching_catch_2()|0;
  $70 = tempRet0;
  ___cxa_free_exception(($63|0)); //@line 144 "modules/image/magpie/PKMHandler.cpp"
  $$0 = $69;$$01 = $70;
 } else {
  __THREW__ = 0;
  invoke_viii(3350,($63|0),(96|0),(344|0)); //@line 144 "modules/image/magpie/PKMHandler.cpp"
  $66 = __THREW__; __THREW__ = 0;
  $67 = ___cxa_find_matching_catch_2()|0;
  $68 = tempRet0;
  $$0 = $67;$$01 = $68;
 }
 __THREW__ = 0;
 invoke_v(3378); //@line 145 "modules/image/magpie/PKMHandler.cpp"
 $71 = __THREW__; __THREW__ = 0;
 $72 = $71&1;
 if ($72) {
  $73 = ___cxa_find_matching_catch_3(0|0)|0;
  $74 = tempRet0;
  ___clang_call_terminate($73); //@line 145 "modules/image/magpie/PKMHandler.cpp"
  // unreachable; //@line 145 "modules/image/magpie/PKMHandler.cpp"
 } else {
  $$1 = $$0;$$12 = $$01;
  ___resumeException($$1|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5image6magpie10PNGHandlerD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5image6magpie13FormatHandlerD2Ev($this); //@line 37 "modules/image/magpie/PNGHandler.h"
 __ZdlPv($this); //@line 37 "modules/image/magpie/PNGHandler.h"
 return; //@line 37 "modules/image/magpie/PNGHandler.h"
}
function __ZN4love5image6magpie10PNGHandler9canDecodeEPNS_10filesystem8FileDataE($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $height = 0;
 var $or$cond = 0, $state = 0, $width = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $width = sp + 300|0;
 $height = sp + 296|0;
 $state = sp;
 HEAP32[$width>>2] = 0; //@line 133 "modules/image/magpie/PNGHandler.cpp"
 HEAP32[$height>>2] = 0; //@line 133 "modules/image/magpie/PNGHandler.cpp"
 $0 = HEAP32[$data>>2]|0; //@line 134 "modules/image/magpie/PNGHandler.cpp"
 $1 = ((($0)) + 16|0); //@line 134 "modules/image/magpie/PNGHandler.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 134 "modules/image/magpie/PNGHandler.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($data)|0); //@line 134 "modules/image/magpie/PNGHandler.cpp"
 $4 = HEAP32[$data>>2]|0; //@line 135 "modules/image/magpie/PNGHandler.cpp"
 $5 = ((($4)) + 20|0); //@line 135 "modules/image/magpie/PNGHandler.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 135 "modules/image/magpie/PNGHandler.cpp"
 $7 = (FUNCTION_TABLE_ii[$6 & 8191]($data)|0); //@line 135 "modules/image/magpie/PNGHandler.cpp"
 __ZN7lodepng5StateC2Ev($state); //@line 137 "modules/image/magpie/PNGHandler.cpp"
 __THREW__ = 0;
 $8 = (invoke_iiiiii(4169,($width|0),($height|0),($state|0),($3|0),($7|0))|0); //@line 138 "modules/image/magpie/PNGHandler.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  __ZN7lodepng5StateD2Ev($state); //@line 141 "modules/image/magpie/PNGHandler.cpp"
  ___resumeException($17|0);
  // unreachable;
 }
 $11 = ($8|0)==(0); //@line 140 "modules/image/magpie/PNGHandler.cpp"
 $12 = HEAP32[$width>>2]|0; //@line 140 "modules/image/magpie/PNGHandler.cpp"
 $13 = ($12|0)!=(0); //@line 140 "modules/image/magpie/PNGHandler.cpp"
 $or$cond = $11 & $13; //@line 140 "modules/image/magpie/PNGHandler.cpp"
 if (!($or$cond)) {
  $16 = 0;
  __ZN7lodepng5StateD2Ev($state); //@line 141 "modules/image/magpie/PNGHandler.cpp"
  STACKTOP = sp;return ($16|0); //@line 141 "modules/image/magpie/PNGHandler.cpp"
 }
 $14 = HEAP32[$height>>2]|0; //@line 140 "modules/image/magpie/PNGHandler.cpp"
 $15 = ($14|0)!=(0); //@line 140 "modules/image/magpie/PNGHandler.cpp"
 $16 = $15;
 __ZN7lodepng5StateD2Ev($state); //@line 141 "modules/image/magpie/PNGHandler.cpp"
 STACKTOP = sp;return ($16|0); //@line 141 "modules/image/magpie/PNGHandler.cpp"
}
function __ZN4love5image6magpie10PNGHandler9canEncodeENS0_9ImageData13EncodedFormatE($this,$format) {
 $this = $this|0;
 $format = $format|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($format|0)==(1); //@line 145 "modules/image/magpie/PNGHandler.cpp"
 return ($0|0); //@line 145 "modules/image/magpie/PNGHandler.cpp"
}
function __ZN4love5image6magpie10PNGHandler6decodeEPNS_10filesystem8FileDataE($agg$result,$this,$fdata) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $fdata = $fdata|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $height = 0, $state = 0, $vararg_buffer = 0, $width = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 320|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $width = sp + 308|0;
 $height = sp + 304|0;
 $state = sp + 8|0;
 HEAP32[$width>>2] = 0; //@line 150 "modules/image/magpie/PNGHandler.cpp"
 HEAP32[$height>>2] = 0; //@line 150 "modules/image/magpie/PNGHandler.cpp"
 $0 = HEAP32[$fdata>>2]|0; //@line 151 "modules/image/magpie/PNGHandler.cpp"
 $1 = ((($0)) + 16|0); //@line 151 "modules/image/magpie/PNGHandler.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 151 "modules/image/magpie/PNGHandler.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($fdata)|0); //@line 151 "modules/image/magpie/PNGHandler.cpp"
 $4 = HEAP32[$fdata>>2]|0; //@line 152 "modules/image/magpie/PNGHandler.cpp"
 $5 = ((($4)) + 20|0); //@line 152 "modules/image/magpie/PNGHandler.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 152 "modules/image/magpie/PNGHandler.cpp"
 $7 = (FUNCTION_TABLE_ii[$6 & 8191]($fdata)|0); //@line 152 "modules/image/magpie/PNGHandler.cpp"
 ;HEAP32[$agg$result>>2]=0|0;HEAP32[$agg$result+4>>2]=0|0;HEAP32[$agg$result+8>>2]=0|0;HEAP32[$agg$result+12>>2]=0|0; //@line 48 "modules/image/magpie/FormatHandler.h"
 __ZN7lodepng5StateC2Ev($state); //@line 156 "modules/image/magpie/PNGHandler.cpp"
 $8 = ((($state)) + 100|0); //@line 158 "modules/image/magpie/PNGHandler.cpp"
 HEAP32[$8>>2] = 6; //@line 158 "modules/image/magpie/PNGHandler.cpp"
 $9 = ((($state)) + 104|0); //@line 159 "modules/image/magpie/PNGHandler.cpp"
 HEAP32[$9>>2] = 8; //@line 159 "modules/image/magpie/PNGHandler.cpp"
 $10 = ((($state)) + 8|0); //@line 161 "modules/image/magpie/PNGHandler.cpp"
 HEAP32[$10>>2] = 4170; //@line 161 "modules/image/magpie/PNGHandler.cpp"
 $11 = ((($agg$result)) + 12|0); //@line 163 "modules/image/magpie/PNGHandler.cpp"
 __THREW__ = 0;
 $12 = (invoke_iiiiiii(4171,($11|0),($width|0),($height|0),($state|0),($3|0),($7|0))|0); //@line 163 "modules/image/magpie/PNGHandler.cpp"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 do {
  if (!($14)) {
   $15 = ($12|0)==(0); //@line 166 "modules/image/magpie/PNGHandler.cpp"
   if ($15) {
    $27 = HEAP32[$width>>2]|0; //@line 172 "modules/image/magpie/PNGHandler.cpp"
    HEAP32[$agg$result>>2] = $27; //@line 172 "modules/image/magpie/PNGHandler.cpp"
    $28 = HEAP32[$height>>2]|0; //@line 173 "modules/image/magpie/PNGHandler.cpp"
    $29 = ((($agg$result)) + 4|0); //@line 173 "modules/image/magpie/PNGHandler.cpp"
    HEAP32[$29>>2] = $28; //@line 173 "modules/image/magpie/PNGHandler.cpp"
    $30 = $27 << 2; //@line 174 "modules/image/magpie/PNGHandler.cpp"
    $31 = Math_imul($30, $28)|0; //@line 174 "modules/image/magpie/PNGHandler.cpp"
    $32 = ((($agg$result)) + 8|0); //@line 174 "modules/image/magpie/PNGHandler.cpp"
    HEAP32[$32>>2] = $31; //@line 174 "modules/image/magpie/PNGHandler.cpp"
    __ZN7lodepng5StateD2Ev($state); //@line 177 "modules/image/magpie/PNGHandler.cpp"
    STACKTOP = sp;return; //@line 177 "modules/image/magpie/PNGHandler.cpp"
   }
   __THREW__ = 0;
   $16 = (invoke_ii(4172,($12|0))|0); //@line 168 "modules/image/magpie/PNGHandler.cpp"
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if (!($18)) {
    $19 = (___cxa_allocate_exception(16)|0); //@line 169 "modules/image/magpie/PNGHandler.cpp"
    __THREW__ = 0;
    HEAP32[$vararg_buffer>>2] = $16; //@line 169 "modules/image/magpie/PNGHandler.cpp"
    invoke_viii(3349,($19|0),(440320|0),($vararg_buffer|0)); //@line 169 "modules/image/magpie/PNGHandler.cpp"
    $20 = __THREW__; __THREW__ = 0;
    $21 = $20&1;
    if (!($21)) {
     __THREW__ = 0;
     invoke_viii(3350,($19|0),(96|0),(344|0)); //@line 169 "modules/image/magpie/PNGHandler.cpp"
     $22 = __THREW__; __THREW__ = 0;
     break;
    }
    $25 = ___cxa_find_matching_catch_2()|0;
    $26 = tempRet0;
    ___cxa_free_exception(($19|0)); //@line 169 "modules/image/magpie/PNGHandler.cpp"
    $$0 = $25;$$01 = $26;
    __ZN7lodepng5StateD2Ev($state); //@line 177 "modules/image/magpie/PNGHandler.cpp"
    ___resumeException($$0|0);
    // unreachable;
   }
  }
 } while(0);
 $23 = ___cxa_find_matching_catch_2()|0;
 $24 = tempRet0;
 $$0 = $23;$$01 = $24;
 __ZN7lodepng5StateD2Ev($state); //@line 177 "modules/image/magpie/PNGHandler.cpp"
 ___resumeException($$0|0);
 // unreachable;
}
function __ZN4love5image6magpieL14zlibDecompressEPPhPjPKhjPK25LodePNGDecompressSettings($out,$outsize,$in,$insize,$0) {
 $out = $out|0;
 $outsize = $outsize|0;
 $in = $in|0;
 $insize = $insize|0;
 $0 = $0|0;
 var $$0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $outdata$0 = 0, $outdata$1 = 0, $outdata$1$lcssa11 = 0, $outdata$1$lcssa12 = 0, $outdatasize = 0, $sizemultiplier$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $outdatasize = sp;
 HEAP32[$outdatasize>>2] = $insize; //@line 52 "modules/image/magpie/PNGHandler.cpp"
 $1 = ($out|0)!=(0|0); //@line 54 "modules/image/magpie/PNGHandler.cpp"
 if ($1) {
  $2 = HEAP32[$out>>2]|0; //@line 54 "modules/image/magpie/PNGHandler.cpp"
  $outdata$0 = $2;$sizemultiplier$0 = 0;
 } else {
  $outdata$0 = 0;$sizemultiplier$0 = 0;
 }
 L3: while(1) {
  $3 = (($sizemultiplier$0) + 1)|0; //@line 59 "modules/image/magpie/PNGHandler.cpp"
  $4 = $insize << $3; //@line 59 "modules/image/magpie/PNGHandler.cpp"
  HEAP32[$outdatasize>>2] = $4; //@line 59 "modules/image/magpie/PNGHandler.cpp"
  $5 = ($outdata$0|0)==(0|0); //@line 65 "modules/image/magpie/PNGHandler.cpp"
  if ($5) {
   $7 = (_malloc($4)|0); //@line 68 "modules/image/magpie/PNGHandler.cpp"
   $outdata$1 = $7;
  } else {
   $6 = (_realloc($outdata$0,$4)|0); //@line 66 "modules/image/magpie/PNGHandler.cpp"
   $outdata$1 = $6;
  }
  $8 = ($outdata$1|0)==(0|0); //@line 70 "modules/image/magpie/PNGHandler.cpp"
  if ($8) {
   $$0 = 83;
   label = 14;
   break;
  }
  $9 = (_uncompress($outdata$1,$outdatasize,$in,$insize)|0); //@line 74 "modules/image/magpie/PNGHandler.cpp"
  switch ($9|0) {
  case 0:  {
   $outdata$1$lcssa12 = $outdata$1;
   label = 10;
   break L3;
   break;
  }
  case -5:  {
   break;
  }
  default: {
   $outdata$1$lcssa11 = $outdata$1;
   label = 9;
   break L3;
  }
  }
  _free($outdata$1); //@line 81 "modules/image/magpie/PNGHandler.cpp"
  $outdata$0 = 0;$sizemultiplier$0 = $3;
 }
 if ((label|0) == 9) {
  _free($outdata$1$lcssa11); //@line 87 "modules/image/magpie/PNGHandler.cpp"
  $$0 = 10000;
  STACKTOP = sp;return ($$0|0); //@line 98 "modules/image/magpie/PNGHandler.cpp"
 }
 else if ((label|0) == 10) {
  if ($1) {
   HEAP32[$out>>2] = $outdata$1$lcssa12; //@line 92 "modules/image/magpie/PNGHandler.cpp"
  }
  $10 = ($outsize|0)==(0|0); //@line 94 "modules/image/magpie/PNGHandler.cpp"
  if ($10) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0); //@line 98 "modules/image/magpie/PNGHandler.cpp"
  }
  $11 = HEAP32[$outdatasize>>2]|0; //@line 95 "modules/image/magpie/PNGHandler.cpp"
  HEAP32[$outsize>>2] = $11; //@line 95 "modules/image/magpie/PNGHandler.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 98 "modules/image/magpie/PNGHandler.cpp"
 }
 else if ((label|0) == 14) {
  STACKTOP = sp;return ($$0|0); //@line 98 "modules/image/magpie/PNGHandler.cpp"
 }
 return (0)|0;
}
function __ZN4love5image6magpie10PNGHandler6encodeERKNS1_13FormatHandler12DecodedImageENS0_9ImageData13EncodedFormatE($agg$result,$this,$img,$format) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $img = $img|0;
 $format = $format|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $state = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 320|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $state = sp + 16|0;
 $0 = ($format|0)==(1); //@line 181 "modules/image/magpie/PNGHandler.cpp"
 if (!($0)) {
  $1 = (___cxa_allocate_exception(16)|0); //@line 182 "modules/image/magpie/PNGHandler.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($1|0),(440352|0),($vararg_buffer|0)); //@line 182 "modules/image/magpie/PNGHandler.cpp"
  $2 = __THREW__; __THREW__ = 0;
  $3 = $2&1;
  if (!($3)) {
   ___cxa_throw(($1|0),(96|0),(344|0)); //@line 182 "modules/image/magpie/PNGHandler.cpp"
   // unreachable; //@line 182 "modules/image/magpie/PNGHandler.cpp"
  }
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  ___cxa_free_exception(($1|0)); //@line 182 "modules/image/magpie/PNGHandler.cpp"
  $$1 = $4;$$12 = $5;
  ___resumeException($$1|0);
  // unreachable;
 }
 HEAP32[$agg$result>>2] = 0; //@line 56 "modules/image/magpie/FormatHandler.h"
 $6 = ((($agg$result)) + 4|0); //@line 57 "modules/image/magpie/FormatHandler.h"
 HEAP32[$6>>2] = 0; //@line 57 "modules/image/magpie/FormatHandler.h"
 __ZN7lodepng5StateC2Ev($state); //@line 186 "modules/image/magpie/PNGHandler.cpp"
 $7 = ((($state)) + 100|0); //@line 188 "modules/image/magpie/PNGHandler.cpp"
 HEAP32[$7>>2] = 6; //@line 188 "modules/image/magpie/PNGHandler.cpp"
 $8 = ((($state)) + 104|0); //@line 189 "modules/image/magpie/PNGHandler.cpp"
 HEAP32[$8>>2] = 8; //@line 189 "modules/image/magpie/PNGHandler.cpp"
 $9 = ((($state)) + 144|0); //@line 192 "modules/image/magpie/PNGHandler.cpp"
 HEAP32[$9>>2] = 6; //@line 192 "modules/image/magpie/PNGHandler.cpp"
 $10 = ((($state)) + 148|0); //@line 193 "modules/image/magpie/PNGHandler.cpp"
 HEAP32[$10>>2] = 8; //@line 193 "modules/image/magpie/PNGHandler.cpp"
 $11 = ((($state)) + 60|0); //@line 195 "modules/image/magpie/PNGHandler.cpp"
 HEAP32[$11>>2] = 4173; //@line 195 "modules/image/magpie/PNGHandler.cpp"
 $12 = ((($img)) + 12|0); //@line 198 "modules/image/magpie/PNGHandler.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 198 "modules/image/magpie/PNGHandler.cpp"
 $14 = HEAP32[$img>>2]|0; //@line 198 "modules/image/magpie/PNGHandler.cpp"
 $15 = ((($img)) + 4|0); //@line 198 "modules/image/magpie/PNGHandler.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 198 "modules/image/magpie/PNGHandler.cpp"
 __THREW__ = 0;
 $17 = (invoke_iiiiiii(4174,($6|0),($agg$result|0),($13|0),($14|0),($16|0),($state|0))|0); //@line 197 "modules/image/magpie/PNGHandler.cpp"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 do {
  if ($19) {
   label = 10;
  } else {
   $20 = ($17|0)==(0); //@line 200 "modules/image/magpie/PNGHandler.cpp"
   if ($20) {
    __ZN7lodepng5StateD2Ev($state); //@line 207 "modules/image/magpie/PNGHandler.cpp"
    STACKTOP = sp;return; //@line 207 "modules/image/magpie/PNGHandler.cpp"
   }
   __THREW__ = 0;
   $21 = (invoke_ii(4172,($17|0))|0); //@line 202 "modules/image/magpie/PNGHandler.cpp"
   $22 = __THREW__; __THREW__ = 0;
   $23 = $22&1;
   if ($23) {
    label = 10;
   } else {
    $24 = (___cxa_allocate_exception(16)|0); //@line 203 "modules/image/magpie/PNGHandler.cpp"
    __THREW__ = 0;
    HEAP32[$vararg_buffer1>>2] = $21; //@line 203 "modules/image/magpie/PNGHandler.cpp"
    invoke_viii(3349,($24|0),(440397|0),($vararg_buffer1|0)); //@line 203 "modules/image/magpie/PNGHandler.cpp"
    $25 = __THREW__; __THREW__ = 0;
    $26 = $25&1;
    if ($26) {
     $30 = ___cxa_find_matching_catch_2()|0;
     $31 = tempRet0;
     ___cxa_free_exception(($24|0)); //@line 203 "modules/image/magpie/PNGHandler.cpp"
     $$0 = $30;$$01 = $31;
     break;
    } else {
     __THREW__ = 0;
     invoke_viii(3350,($24|0),(96|0),(344|0)); //@line 203 "modules/image/magpie/PNGHandler.cpp"
     $27 = __THREW__; __THREW__ = 0;
     label = 10;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 10) {
  $28 = ___cxa_find_matching_catch_2()|0;
  $29 = tempRet0;
  $$0 = $28;$$01 = $29;
 }
 __ZN7lodepng5StateD2Ev($state); //@line 207 "modules/image/magpie/PNGHandler.cpp"
 $$1 = $$0;$$12 = $$01;
 ___resumeException($$1|0);
 // unreachable;
}
function __ZN4love5image6magpieL12zlibCompressEPPhPjPKhjPK23LodePNGCompressSettings($out,$outsize,$in,$insize,$0) {
 $out = $out|0;
 $outsize = $outsize|0;
 $in = $in|0;
 $insize = $insize|0;
 $0 = $0|0;
 var $$1 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $outdatasize = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $outdatasize = sp;
 $1 = (_compressBound($insize)|0); //@line 105 "modules/image/magpie/PNGHandler.cpp"
 HEAP32[$outdatasize>>2] = $1; //@line 105 "modules/image/magpie/PNGHandler.cpp"
 $2 = (_malloc($1)|0); //@line 108 "modules/image/magpie/PNGHandler.cpp"
 $3 = ($2|0)==(0|0); //@line 110 "modules/image/magpie/PNGHandler.cpp"
 do {
  if ($3) {
   $$1 = 83;
  } else {
   $4 = (_compress($2,$outdatasize,$in,$insize)|0); //@line 114 "modules/image/magpie/PNGHandler.cpp"
   $5 = ($4|0)==(0); //@line 116 "modules/image/magpie/PNGHandler.cpp"
   if (!($5)) {
    _free($2); //@line 118 "modules/image/magpie/PNGHandler.cpp"
    $$1 = 10000;
    break;
   }
   $6 = ($out|0)==(0|0); //@line 122 "modules/image/magpie/PNGHandler.cpp"
   if (!($6)) {
    HEAP32[$out>>2] = $2; //@line 123 "modules/image/magpie/PNGHandler.cpp"
   }
   $7 = ($outsize|0)==(0|0); //@line 125 "modules/image/magpie/PNGHandler.cpp"
   if ($7) {
    $$1 = 0;
   } else {
    $8 = HEAP32[$outdatasize>>2]|0; //@line 126 "modules/image/magpie/PNGHandler.cpp"
    HEAP32[$outsize>>2] = $8; //@line 126 "modules/image/magpie/PNGHandler.cpp"
    $$1 = 0;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$1|0); //@line 129 "modules/image/magpie/PNGHandler.cpp"
}
function __ZN4love5image6magpie10PNGHandler4freeEPh($this,$mem) {
 $this = $this|0;
 $mem = $mem|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mem|0)==(0|0); //@line 212 "modules/image/magpie/PNGHandler.cpp"
 if (!($0)) {
  _free($mem); //@line 213 "modules/image/magpie/PNGHandler.cpp"
 }
 return; //@line 214 "modules/image/magpie/PNGHandler.cpp"
}
function __ZN4love5image6magpie10PVRHandlerD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 39 "modules/image/magpie/PVRHandler.h"
 return; //@line 39 "modules/image/magpie/PVRHandler.h"
}
function __ZN4love5image6magpie10PVRHandler8canParseEPKNS_10filesystem8FileDataE($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $switch$split12D = 0, $switch$split2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$data>>2]|0; //@line 458 "modules/image/magpie/PVRHandler.cpp"
 $1 = ((($0)) + 20|0); //@line 458 "modules/image/magpie/PVRHandler.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 458 "modules/image/magpie/PVRHandler.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($data)|0); //@line 458 "modules/image/magpie/PVRHandler.cpp"
 $4 = ($3>>>0)<(52); //@line 458 "modules/image/magpie/PVRHandler.cpp"
 L1: do {
  if ($4) {
   $$2 = 0;
  } else {
   $5 = HEAP32[$data>>2]|0; //@line 458 "modules/image/magpie/PVRHandler.cpp"
   $6 = ((($5)) + 20|0); //@line 458 "modules/image/magpie/PVRHandler.cpp"
   $7 = HEAP32[$6>>2]|0; //@line 458 "modules/image/magpie/PVRHandler.cpp"
   $8 = (FUNCTION_TABLE_ii[$7 & 8191]($data)|0); //@line 458 "modules/image/magpie/PVRHandler.cpp"
   $9 = ($8>>>0)<(52); //@line 458 "modules/image/magpie/PVRHandler.cpp"
   if ($9) {
    $$2 = 0;
   } else {
    $10 = HEAP32[$data>>2]|0; //@line 461 "modules/image/magpie/PVRHandler.cpp"
    $11 = ((($10)) + 16|0); //@line 461 "modules/image/magpie/PVRHandler.cpp"
    $12 = HEAP32[$11>>2]|0; //@line 461 "modules/image/magpie/PVRHandler.cpp"
    $13 = (FUNCTION_TABLE_ii[$12 & 8191]($data)|0); //@line 461 "modules/image/magpie/PVRHandler.cpp"
    $14 = HEAP32[$13>>2]|0; //@line 464 "modules/image/magpie/PVRHandler.cpp"
    $switch$split2D = ($14|0)<(1347834371);
    if ($switch$split2D) {
     switch ($14|0) {
     case 55727696:  {
      $$2 = 1;
      break L1;
      break;
     }
     default: {
     }
     }
    } else {
     switch ($14|0) {
     case 1347834371:  {
      $$2 = 1;
      break L1;
      break;
     }
     default: {
     }
     }
    }
    $15 = HEAP32[$data>>2]|0; //@line 468 "modules/image/magpie/PVRHandler.cpp"
    $16 = ((($15)) + 16|0); //@line 468 "modules/image/magpie/PVRHandler.cpp"
    $17 = HEAP32[$16>>2]|0; //@line 468 "modules/image/magpie/PVRHandler.cpp"
    $18 = (FUNCTION_TABLE_ii[$17 & 8191]($data)|0); //@line 468 "modules/image/magpie/PVRHandler.cpp"
    $19 = ((($18)) + 44|0); //@line 471 "modules/image/magpie/PVRHandler.cpp"
    $20 = HEAP32[$19>>2]|0; //@line 471 "modules/image/magpie/PVRHandler.cpp"
    $switch$split12D = ($20|0)<(1347834401);
    if ($switch$split12D) {
     switch ($20|0) {
     case 559044176:  {
      $$2 = 1;
      break L1;
      break;
     }
     default: {
     }
     }
    } else {
     switch ($20|0) {
     case 1347834401:  {
      $$2 = 1;
      break L1;
      break;
     }
     default: {
     }
     }
    }
    $$2 = 0;
   }
  }
 } while(0);
 return ($$2|0); //@line 475 "modules/image/magpie/PVRHandler.cpp"
}
function __ZN4love5image6magpie10PVRHandler5parseEPNS_10filesystem8FileDataERNSt3__16vectorINS0_19CompressedImageData8SubImageENS6_9allocatorIS9_EEEERjRNS8_6FormatERb($this,$filedata,$images,$dataSize,$format,$sRGB) {
 $this = $this|0;
 $filedata = $filedata|0;
 $images = $images|0;
 $dataSize = $dataSize|0;
 $format = $format|0;
 $sRGB = $sRGB|0;
 var $$0 = 0, $$0$i$ph = 0, $$02 = 0, $$1 = 0, $$13 = 0, $$clear = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0;
 var $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0;
 var $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0;
 var $235 = 0, $236 = 0, $237 = 0, $238 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $curoffset$010 = 0, $header3 = 0, $i$013 = 0, $i1$011 = 0, $mip = 0, $snorm$0$off0$i = 0, $switch$cast$i = 0, $switch$cast$i$clear = 0, $switch$downshift$i = 0, $switch$masked$i = 0, $switch$split2D = 0, $switch$tableidx$i = 0, $totalsize$0$lcssa = 0;
 var $totalsize$014 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 32|0; //@line 152 "modules/image/magpie/PVRHandler.cpp"
 $vararg_buffer5 = sp + 24|0; //@line 152 "modules/image/magpie/PVRHandler.cpp"
 $vararg_buffer3 = sp + 16|0; //@line 152 "modules/image/magpie/PVRHandler.cpp"
 $vararg_buffer1 = sp + 8|0; //@line 152 "modules/image/magpie/PVRHandler.cpp"
 $vararg_buffer = sp; //@line 152 "modules/image/magpie/PVRHandler.cpp"
 $header3 = sp + 56|0;
 $mip = sp + 40|0;
 $0 = HEAP32[$this>>2]|0; //@line 479 "modules/image/magpie/PVRHandler.cpp"
 $1 = ((($0)) + 16|0); //@line 479 "modules/image/magpie/PVRHandler.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 479 "modules/image/magpie/PVRHandler.cpp"
 $3 = (FUNCTION_TABLE_iii[$2 & 8191]($this,$filedata)|0); //@line 479 "modules/image/magpie/PVRHandler.cpp"
 if (!($3)) {
  $4 = (___cxa_allocate_exception(16)|0); //@line 480 "modules/image/magpie/PVRHandler.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($4|0),(440462|0),($vararg_buffer|0)); //@line 480 "modules/image/magpie/PVRHandler.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   ___cxa_throw(($4|0),(96|0),(344|0)); //@line 480 "modules/image/magpie/PVRHandler.cpp"
   // unreachable; //@line 480 "modules/image/magpie/PVRHandler.cpp"
  }
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  ___cxa_free_exception(($4|0)); //@line 480 "modules/image/magpie/PVRHandler.cpp"
  $$1 = $7;$$13 = $8;
  ___resumeException($$1|0);
  // unreachable;
 }
 $9 = HEAP32[$filedata>>2]|0; //@line 482 "modules/image/magpie/PVRHandler.cpp"
 $10 = ((($9)) + 16|0); //@line 482 "modules/image/magpie/PVRHandler.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 482 "modules/image/magpie/PVRHandler.cpp"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($filedata)|0); //@line 482 "modules/image/magpie/PVRHandler.cpp"
 dest=$header3; src=$12; stop=dest+52|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 482 "modules/image/magpie/PVRHandler.cpp"
 $13 = HEAP32[$header3>>2]|0; //@line 485 "modules/image/magpie/PVRHandler.cpp"
 $switch$split2D = ($13|0)<(1347834371);
 L8: do {
  if ($switch$split2D) {
   switch ($13|0) {
   case 55727696:  {
    label = 18;
    break;
   }
   default: {
    label = 6;
   }
   }
  } else {
   switch ($13|0) {
   case 1347834371:  {
    break;
   }
   default: {
    label = 6;
    break L8;
   }
   }
   HEAP32[$header3>>2] = 55727696; //@line 491 "modules/image/magpie/PVRHandler.cpp"
   $94 = ((($header3)) + 4|0); //@line 492 "modules/image/magpie/PVRHandler.cpp"
   $95 = HEAP32[$94>>2]|0; //@line 492 "modules/image/magpie/PVRHandler.cpp"
   $96 = (_llvm_bswap_i32(($95|0))|0); //@line 57 "./common/int.h"
   HEAP32[$94>>2] = $96; //@line 492 "modules/image/magpie/PVRHandler.cpp"
   $97 = ((($header3)) + 8|0); //@line 493 "modules/image/magpie/PVRHandler.cpp"
   $98 = $97; //@line 493 "modules/image/magpie/PVRHandler.cpp"
   $99 = $98; //@line 493 "modules/image/magpie/PVRHandler.cpp"
   $100 = HEAP32[$99>>2]|0; //@line 493 "modules/image/magpie/PVRHandler.cpp"
   $101 = (($98) + 4)|0; //@line 493 "modules/image/magpie/PVRHandler.cpp"
   $102 = $101; //@line 493 "modules/image/magpie/PVRHandler.cpp"
   $103 = HEAP32[$102>>2]|0; //@line 493 "modules/image/magpie/PVRHandler.cpp"
   $104 = (_llvm_bswap_i64(($100|0),($103|0))|0); //@line 66 "./common/int.h"
   $105 = tempRet0; //@line 66 "./common/int.h"
   $106 = $97; //@line 493 "modules/image/magpie/PVRHandler.cpp"
   $107 = $106; //@line 493 "modules/image/magpie/PVRHandler.cpp"
   HEAP32[$107>>2] = $104; //@line 493 "modules/image/magpie/PVRHandler.cpp"
   $108 = (($106) + 4)|0; //@line 493 "modules/image/magpie/PVRHandler.cpp"
   $109 = $108; //@line 493 "modules/image/magpie/PVRHandler.cpp"
   HEAP32[$109>>2] = $105; //@line 493 "modules/image/magpie/PVRHandler.cpp"
   $110 = ((($header3)) + 16|0); //@line 494 "modules/image/magpie/PVRHandler.cpp"
   $111 = HEAP32[$110>>2]|0; //@line 494 "modules/image/magpie/PVRHandler.cpp"
   $112 = (_llvm_bswap_i32(($111|0))|0); //@line 57 "./common/int.h"
   HEAP32[$110>>2] = $112; //@line 494 "modules/image/magpie/PVRHandler.cpp"
   $113 = ((($header3)) + 20|0); //@line 495 "modules/image/magpie/PVRHandler.cpp"
   $114 = HEAP32[$113>>2]|0; //@line 495 "modules/image/magpie/PVRHandler.cpp"
   $115 = (_llvm_bswap_i32(($114|0))|0); //@line 57 "./common/int.h"
   HEAP32[$113>>2] = $115; //@line 495 "modules/image/magpie/PVRHandler.cpp"
   $116 = ((($header3)) + 24|0); //@line 496 "modules/image/magpie/PVRHandler.cpp"
   $117 = HEAP32[$116>>2]|0; //@line 496 "modules/image/magpie/PVRHandler.cpp"
   $118 = (_llvm_bswap_i32(($117|0))|0); //@line 57 "./common/int.h"
   HEAP32[$116>>2] = $118; //@line 496 "modules/image/magpie/PVRHandler.cpp"
   $119 = ((($header3)) + 28|0); //@line 497 "modules/image/magpie/PVRHandler.cpp"
   $120 = HEAP32[$119>>2]|0; //@line 497 "modules/image/magpie/PVRHandler.cpp"
   $121 = (_llvm_bswap_i32(($120|0))|0); //@line 57 "./common/int.h"
   HEAP32[$119>>2] = $121; //@line 497 "modules/image/magpie/PVRHandler.cpp"
   $122 = ((($header3)) + 32|0); //@line 498 "modules/image/magpie/PVRHandler.cpp"
   $123 = HEAP32[$122>>2]|0; //@line 498 "modules/image/magpie/PVRHandler.cpp"
   $124 = (_llvm_bswap_i32(($123|0))|0); //@line 57 "./common/int.h"
   HEAP32[$122>>2] = $124; //@line 498 "modules/image/magpie/PVRHandler.cpp"
   $125 = ((($header3)) + 40|0); //@line 499 "modules/image/magpie/PVRHandler.cpp"
   $126 = HEAP32[$125>>2]|0; //@line 499 "modules/image/magpie/PVRHandler.cpp"
   $127 = (_llvm_bswap_i32(($126|0))|0); //@line 57 "./common/int.h"
   HEAP32[$125>>2] = $127; //@line 499 "modules/image/magpie/PVRHandler.cpp"
   $128 = ((($header3)) + 44|0); //@line 500 "modules/image/magpie/PVRHandler.cpp"
   $129 = HEAP32[$128>>2]|0; //@line 500 "modules/image/magpie/PVRHandler.cpp"
   $130 = (_llvm_bswap_i32(($129|0))|0); //@line 57 "./common/int.h"
   HEAP32[$128>>2] = $130; //@line 500 "modules/image/magpie/PVRHandler.cpp"
   $131 = ((($header3)) + 48|0); //@line 501 "modules/image/magpie/PVRHandler.cpp"
   $132 = HEAP32[$131>>2]|0; //@line 501 "modules/image/magpie/PVRHandler.cpp"
   $133 = (_llvm_bswap_i32(($132|0))|0); //@line 57 "./common/int.h"
   HEAP32[$131>>2] = $133; //@line 501 "modules/image/magpie/PVRHandler.cpp"
   $135 = $124;
  }
 } while(0);
 L12: do {
  if ((label|0) == 6) {
   $14 = HEAP32[$filedata>>2]|0; //@line 486 "modules/image/magpie/PVRHandler.cpp"
   $15 = ((($14)) + 16|0); //@line 486 "modules/image/magpie/PVRHandler.cpp"
   $16 = HEAP32[$15>>2]|0; //@line 486 "modules/image/magpie/PVRHandler.cpp"
   $17 = (FUNCTION_TABLE_ii[$16 & 8191]($filedata)|0); //@line 486 "modules/image/magpie/PVRHandler.cpp"
   $18 = ((($17)) + 4|0); //@line 486 "modules/image/magpie/PVRHandler.cpp"
   $19 = HEAP32[$18>>2]|0; //@line 486 "modules/image/magpie/PVRHandler.cpp"
   $20 = ((($17)) + 8|0); //@line 486 "modules/image/magpie/PVRHandler.cpp"
   $21 = HEAP32[$20>>2]|0; //@line 486 "modules/image/magpie/PVRHandler.cpp"
   $22 = ((($17)) + 12|0); //@line 486 "modules/image/magpie/PVRHandler.cpp"
   $23 = HEAP32[$22>>2]|0; //@line 486 "modules/image/magpie/PVRHandler.cpp"
   $24 = ((($17)) + 16|0); //@line 486 "modules/image/magpie/PVRHandler.cpp"
   $25 = HEAP32[$24>>2]|0; //@line 486 "modules/image/magpie/PVRHandler.cpp"
   $26 = ((($17)) + 44|0); //@line 486 "modules/image/magpie/PVRHandler.cpp"
   $27 = HEAP32[$26>>2]|0; //@line 486 "modules/image/magpie/PVRHandler.cpp"
   $28 = ((($17)) + 48|0); //@line 486 "modules/image/magpie/PVRHandler.cpp"
   $29 = HEAP32[$28>>2]|0; //@line 486 "modules/image/magpie/PVRHandler.cpp"
   $30 = ($27|0)==(1347834401); //@line 155 "modules/image/magpie/PVRHandler.cpp"
   if ($30) {
    $31 = (_llvm_bswap_i32(($19|0))|0); //@line 57 "./common/int.h"
    $32 = (_llvm_bswap_i32(($21|0))|0); //@line 57 "./common/int.h"
    $33 = (_llvm_bswap_i32(($23|0))|0); //@line 57 "./common/int.h"
    $34 = (_llvm_bswap_i32(($25|0))|0); //@line 57 "./common/int.h"
    $35 = (_llvm_bswap_i32(($29|0))|0); //@line 57 "./common/int.h"
    $37 = $31;$39 = $32;$42 = $35;$45 = $33;$48 = $34;
   } else {
    $37 = $19;$39 = $21;$42 = $29;$45 = $23;$48 = $25;
   }
   dest=$header3; stop=dest+44|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 163 "modules/image/magpie/PVRHandler.cpp"
   HEAP32[$header3>>2] = 55727696; //@line 165 "modules/image/magpie/PVRHandler.cpp"
   $36 = ((($header3)) + 24|0); //@line 166 "modules/image/magpie/PVRHandler.cpp"
   HEAP32[$36>>2] = $37; //@line 166 "modules/image/magpie/PVRHandler.cpp"
   $38 = ((($header3)) + 28|0); //@line 167 "modules/image/magpie/PVRHandler.cpp"
   HEAP32[$38>>2] = $39; //@line 167 "modules/image/magpie/PVRHandler.cpp"
   $40 = ((($header3)) + 32|0); //@line 168 "modules/image/magpie/PVRHandler.cpp"
   HEAP32[$40>>2] = 1; //@line 168 "modules/image/magpie/PVRHandler.cpp"
   $41 = ((($header3)) + 36|0); //@line 169 "modules/image/magpie/PVRHandler.cpp"
   HEAP32[$41>>2] = $42; //@line 169 "modules/image/magpie/PVRHandler.cpp"
   $43 = ((($header3)) + 40|0); //@line 170 "modules/image/magpie/PVRHandler.cpp"
   HEAP32[$43>>2] = 1; //@line 170 "modules/image/magpie/PVRHandler.cpp"
   $44 = ((($header3)) + 44|0); //@line 171 "modules/image/magpie/PVRHandler.cpp"
   HEAP32[$44>>2] = $45; //@line 171 "modules/image/magpie/PVRHandler.cpp"
   $46 = ((($header3)) + 48|0); //@line 172 "modules/image/magpie/PVRHandler.cpp"
   HEAP32[$46>>2] = 0; //@line 172 "modules/image/magpie/PVRHandler.cpp"
   $47 = $48 & 255; //@line 174 "modules/image/magpie/PVRHandler.cpp"
   switch ($47|0) {
   case 24:  {
    $49 = ((($header3)) + 8|0); //@line 177 "modules/image/magpie/PVRHandler.cpp"
    $50 = $49; //@line 177 "modules/image/magpie/PVRHandler.cpp"
    $51 = $50; //@line 177 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$51>>2] = 1; //@line 177 "modules/image/magpie/PVRHandler.cpp"
    $52 = (($50) + 4)|0; //@line 177 "modules/image/magpie/PVRHandler.cpp"
    $53 = $52; //@line 177 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$53>>2] = 0; //@line 177 "modules/image/magpie/PVRHandler.cpp"
    label = 18;
    break L12;
    break;
   }
   case 25:  {
    $54 = ((($header3)) + 8|0); //@line 180 "modules/image/magpie/PVRHandler.cpp"
    $55 = $54; //@line 180 "modules/image/magpie/PVRHandler.cpp"
    $56 = $55; //@line 180 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$56>>2] = 3; //@line 180 "modules/image/magpie/PVRHandler.cpp"
    $57 = (($55) + 4)|0; //@line 180 "modules/image/magpie/PVRHandler.cpp"
    $58 = $57; //@line 180 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$58>>2] = 0; //@line 180 "modules/image/magpie/PVRHandler.cpp"
    label = 18;
    break L12;
    break;
   }
   case 28:  {
    $59 = ((($header3)) + 8|0); //@line 183 "modules/image/magpie/PVRHandler.cpp"
    $60 = $59; //@line 183 "modules/image/magpie/PVRHandler.cpp"
    $61 = $60; //@line 183 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$61>>2] = 4; //@line 183 "modules/image/magpie/PVRHandler.cpp"
    $62 = (($60) + 4)|0; //@line 183 "modules/image/magpie/PVRHandler.cpp"
    $63 = $62; //@line 183 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$63>>2] = 0; //@line 183 "modules/image/magpie/PVRHandler.cpp"
    label = 18;
    break L12;
    break;
   }
   case 29:  {
    $64 = ((($header3)) + 8|0); //@line 186 "modules/image/magpie/PVRHandler.cpp"
    $65 = $64; //@line 186 "modules/image/magpie/PVRHandler.cpp"
    $66 = $65; //@line 186 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$66>>2] = 5; //@line 186 "modules/image/magpie/PVRHandler.cpp"
    $67 = (($65) + 4)|0; //@line 186 "modules/image/magpie/PVRHandler.cpp"
    $68 = $67; //@line 186 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$68>>2] = 0; //@line 186 "modules/image/magpie/PVRHandler.cpp"
    label = 18;
    break L12;
    break;
   }
   case 32:  {
    $69 = ((($header3)) + 8|0); //@line 189 "modules/image/magpie/PVRHandler.cpp"
    $70 = $69; //@line 189 "modules/image/magpie/PVRHandler.cpp"
    $71 = $70; //@line 189 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$71>>2] = 7; //@line 189 "modules/image/magpie/PVRHandler.cpp"
    $72 = (($70) + 4)|0; //@line 189 "modules/image/magpie/PVRHandler.cpp"
    $73 = $72; //@line 189 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$73>>2] = 0; //@line 189 "modules/image/magpie/PVRHandler.cpp"
    label = 18;
    break L12;
    break;
   }
   case 34:  {
    $74 = ((($header3)) + 8|0); //@line 192 "modules/image/magpie/PVRHandler.cpp"
    $75 = $74; //@line 192 "modules/image/magpie/PVRHandler.cpp"
    $76 = $75; //@line 192 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$76>>2] = 9; //@line 192 "modules/image/magpie/PVRHandler.cpp"
    $77 = (($75) + 4)|0; //@line 192 "modules/image/magpie/PVRHandler.cpp"
    $78 = $77; //@line 192 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$78>>2] = 0; //@line 192 "modules/image/magpie/PVRHandler.cpp"
    label = 18;
    break L12;
    break;
   }
   case 36:  {
    $79 = ((($header3)) + 8|0); //@line 195 "modules/image/magpie/PVRHandler.cpp"
    $80 = $79; //@line 195 "modules/image/magpie/PVRHandler.cpp"
    $81 = $80; //@line 195 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$81>>2] = 11; //@line 195 "modules/image/magpie/PVRHandler.cpp"
    $82 = (($80) + 4)|0; //@line 195 "modules/image/magpie/PVRHandler.cpp"
    $83 = $82; //@line 195 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$83>>2] = 0; //@line 195 "modules/image/magpie/PVRHandler.cpp"
    label = 18;
    break L12;
    break;
   }
   case 54:  {
    $84 = ((($header3)) + 8|0); //@line 198 "modules/image/magpie/PVRHandler.cpp"
    $85 = $84; //@line 198 "modules/image/magpie/PVRHandler.cpp"
    $86 = $85; //@line 198 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$86>>2] = 6; //@line 198 "modules/image/magpie/PVRHandler.cpp"
    $87 = (($85) + 4)|0; //@line 198 "modules/image/magpie/PVRHandler.cpp"
    $88 = $87; //@line 198 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$88>>2] = 0; //@line 198 "modules/image/magpie/PVRHandler.cpp"
    label = 18;
    break L12;
    break;
   }
   default: {
    $89 = ((($header3)) + 8|0); //@line 201 "modules/image/magpie/PVRHandler.cpp"
    $90 = $89; //@line 201 "modules/image/magpie/PVRHandler.cpp"
    $91 = $90; //@line 201 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$91>>2] = 127; //@line 201 "modules/image/magpie/PVRHandler.cpp"
    $92 = (($90) + 4)|0; //@line 201 "modules/image/magpie/PVRHandler.cpp"
    $93 = $92; //@line 201 "modules/image/magpie/PVRHandler.cpp"
    HEAP32[$93>>2] = 0; //@line 201 "modules/image/magpie/PVRHandler.cpp"
    label = 18;
    break L12;
   }
   }
  }
 } while(0);
 if ((label|0) == 18) {
  $$phi$trans$insert = ((($header3)) + 32|0); //@line 504 "modules/image/magpie/PVRHandler.cpp"
  $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 504 "modules/image/magpie/PVRHandler.cpp"
  $135 = $$pre;
 }
 $134 = ($135>>>0)>(1); //@line 504 "modules/image/magpie/PVRHandler.cpp"
 if ($134) {
  $136 = (___cxa_allocate_exception(16)|0); //@line 505 "modules/image/magpie/PVRHandler.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($136|0),(440513|0),($vararg_buffer1|0)); //@line 505 "modules/image/magpie/PVRHandler.cpp"
  $137 = __THREW__; __THREW__ = 0;
  $138 = $137&1;
  if (!($138)) {
   ___cxa_throw(($136|0),(96|0),(344|0)); //@line 505 "modules/image/magpie/PVRHandler.cpp"
   // unreachable; //@line 505 "modules/image/magpie/PVRHandler.cpp"
  }
  $139 = ___cxa_find_matching_catch_2()|0;
  $140 = tempRet0;
  ___cxa_free_exception(($136|0)); //@line 505 "modules/image/magpie/PVRHandler.cpp"
  $$1 = $139;$$13 = $140;
  ___resumeException($$1|0);
  // unreachable;
 }
 $141 = ((($header3)) + 8|0); //@line 507 "modules/image/magpie/PVRHandler.cpp"
 $142 = $141; //@line 507 "modules/image/magpie/PVRHandler.cpp"
 $143 = $142; //@line 507 "modules/image/magpie/PVRHandler.cpp"
 $144 = HEAP32[$143>>2]|0; //@line 507 "modules/image/magpie/PVRHandler.cpp"
 $145 = (($142) + 4)|0; //@line 507 "modules/image/magpie/PVRHandler.cpp"
 $146 = $145; //@line 507 "modules/image/magpie/PVRHandler.cpp"
 $147 = HEAP32[$146>>2]|0; //@line 507 "modules/image/magpie/PVRHandler.cpp"
 $148 = ((($header3)) + 20|0); //@line 508 "modules/image/magpie/PVRHandler.cpp"
 $149 = HEAP32[$148>>2]|0; //@line 508 "modules/image/magpie/PVRHandler.cpp"
 $switch$tableidx$i = (($149) + -1)|0; //@line 210 "modules/image/magpie/PVRHandler.cpp"
 $150 = ($switch$tableidx$i>>>0)<(9); //@line 210 "modules/image/magpie/PVRHandler.cpp"
 if ($150) {
  $switch$cast$i = $switch$tableidx$i&65535; //@line 210 "modules/image/magpie/PVRHandler.cpp"
  $switch$cast$i$clear = $switch$cast$i & 511; //@line 210 "modules/image/magpie/PVRHandler.cpp"
  $switch$downshift$i = (273&65535) >>> $switch$cast$i$clear;
  $151 = $switch$downshift$i & 1; //@line 210 "modules/image/magpie/PVRHandler.cpp"
  $$clear = $151 & 511; //@line 210 "modules/image/magpie/PVRHandler.cpp"
  $switch$masked$i = ($$clear<<16>>16)!=(0); //@line 210 "modules/image/magpie/PVRHandler.cpp"
  $snorm$0$off0$i = $switch$masked$i;
 } else {
  $snorm$0$off0$i = 0;
 }
 do {
  switch ($144|0) {
  case 0:  {
   $$0$i$ph = 11;
   break;
  }
  case 1:  {
   $$0$i$ph = 13;
   break;
  }
  case 2:  {
   $$0$i$ph = 12;
   break;
  }
  case 3:  {
   $$0$i$ph = 14;
   break;
  }
  case 6:  {
   $$0$i$ph = 15;
   break;
  }
  case 7:  {
   $$0$i$ph = 1;
   break;
  }
  case 9:  {
   $$0$i$ph = 2;
   break;
  }
  case 11:  {
   $$0$i$ph = 3;
   break;
  }
  case 12:  {
   $152 = $snorm$0$off0$i ? 5 : 4; //@line 240 "modules/image/magpie/PVRHandler.cpp"
   $$0$i$ph = $152;
   break;
  }
  case 13:  {
   $153 = $snorm$0$off0$i ? 7 : 6; //@line 242 "modules/image/magpie/PVRHandler.cpp"
   $$0$i$ph = $153;
   break;
  }
  case 14:  {
   $154 = $snorm$0$off0$i ? 9 : 8; //@line 244 "modules/image/magpie/PVRHandler.cpp"
   $$0$i$ph = $154;
   break;
  }
  case 15:  {
   $$0$i$ph = 10;
   break;
  }
  case 22:  {
   $$0$i$ph = 16;
   break;
  }
  case 23:  {
   $$0$i$ph = 17;
   break;
  }
  case 24:  {
   $$0$i$ph = 18;
   break;
  }
  case 25:  {
   $155 = $snorm$0$off0$i ? 20 : 19; //@line 254 "modules/image/magpie/PVRHandler.cpp"
   $$0$i$ph = $155;
   break;
  }
  case 26:  {
   $156 = $snorm$0$off0$i ? 22 : 21; //@line 256 "modules/image/magpie/PVRHandler.cpp"
   $$0$i$ph = $156;
   break;
  }
  case 27:  {
   $$0$i$ph = 23;
   break;
  }
  case 28:  {
   $$0$i$ph = 24;
   break;
  }
  case 29:  {
   $$0$i$ph = 25;
   break;
  }
  case 30:  {
   $$0$i$ph = 26;
   break;
  }
  case 31:  {
   $$0$i$ph = 27;
   break;
  }
  case 32:  {
   $$0$i$ph = 28;
   break;
  }
  case 33:  {
   $$0$i$ph = 29;
   break;
  }
  case 34:  {
   $$0$i$ph = 30;
   break;
  }
  case 35:  {
   $$0$i$ph = 31;
   break;
  }
  case 36:  {
   $$0$i$ph = 32;
   break;
  }
  case 37:  {
   $$0$i$ph = 33;
   break;
  }
  case 38:  {
   $$0$i$ph = 34;
   break;
  }
  case 39:  {
   $$0$i$ph = 35;
   break;
  }
  case 40:  {
   $$0$i$ph = 36;
   break;
  }
  default: {
   $160 = (___cxa_allocate_exception(16)|0); //@line 513 "modules/image/magpie/PVRHandler.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($160|0),(440571|0),($vararg_buffer3|0)); //@line 513 "modules/image/magpie/PVRHandler.cpp"
   $161 = __THREW__; __THREW__ = 0;
   $162 = $161&1;
   if (!($162)) {
    ___cxa_throw(($160|0),(96|0),(344|0)); //@line 513 "modules/image/magpie/PVRHandler.cpp"
    // unreachable; //@line 513 "modules/image/magpie/PVRHandler.cpp"
   }
   $163 = ___cxa_find_matching_catch_2()|0;
   $164 = tempRet0;
   ___cxa_free_exception(($160|0)); //@line 513 "modules/image/magpie/PVRHandler.cpp"
   $$1 = $163;$$13 = $164;
   ___resumeException($$1|0);
   // unreachable;
  }
  }
 } while(0);
 $157 = ((($header3)) + 44|0); //@line 519 "modules/image/magpie/PVRHandler.cpp"
 $158 = HEAP32[$157>>2]|0; //@line 540 "modules/image/magpie/PVRHandler.cpp"
 $159 = ($158|0)>(0); //@line 519 "modules/image/magpie/PVRHandler.cpp"
 if ($159) {
  $i$013 = 0;$totalsize$014 = 0;
  while(1) {
   $174 = (__ZN4love5image6magpie12_GLOBAL__N_115getMipLevelSizeERKNS2_14PVRTexHeaderV3Ei($header3,$i$013)|0); //@line 520 "modules/image/magpie/PVRHandler.cpp"
   $175 = (($174) + ($totalsize$014))|0; //@line 520 "modules/image/magpie/PVRHandler.cpp"
   $176 = (($i$013) + 1)|0; //@line 519 "modules/image/magpie/PVRHandler.cpp"
   $177 = ($176|0)<($158|0); //@line 519 "modules/image/magpie/PVRHandler.cpp"
   if ($177) {
    $i$013 = $176;$totalsize$014 = $175;
   } else {
    $totalsize$0$lcssa = $175;
    break;
   }
  }
 } else {
  $totalsize$0$lcssa = 0;
 }
 $165 = ((($header3)) + 48|0); //@line 522 "modules/image/magpie/PVRHandler.cpp"
 $166 = HEAP32[$165>>2]|0; //@line 522 "modules/image/magpie/PVRHandler.cpp"
 $167 = (($166) + 52)|0; //@line 522 "modules/image/magpie/PVRHandler.cpp"
 $168 = HEAP32[$filedata>>2]|0; //@line 525 "modules/image/magpie/PVRHandler.cpp"
 $169 = ((($168)) + 20|0); //@line 525 "modules/image/magpie/PVRHandler.cpp"
 $170 = HEAP32[$169>>2]|0; //@line 525 "modules/image/magpie/PVRHandler.cpp"
 $171 = (FUNCTION_TABLE_ii[$170 & 8191]($filedata)|0); //@line 525 "modules/image/magpie/PVRHandler.cpp"
 $172 = (($167) + ($totalsize$0$lcssa))|0; //@line 525 "modules/image/magpie/PVRHandler.cpp"
 $173 = ($171>>>0)<($172>>>0); //@line 525 "modules/image/magpie/PVRHandler.cpp"
 if ($173) {
  $178 = (___cxa_allocate_exception(16)|0); //@line 526 "modules/image/magpie/PVRHandler.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($178|0),(440623|0),($vararg_buffer5|0)); //@line 526 "modules/image/magpie/PVRHandler.cpp"
  $179 = __THREW__; __THREW__ = 0;
  $180 = $179&1;
  if (!($180)) {
   ___cxa_throw(($178|0),(96|0),(344|0)); //@line 526 "modules/image/magpie/PVRHandler.cpp"
   // unreachable; //@line 526 "modules/image/magpie/PVRHandler.cpp"
  }
  $181 = ___cxa_find_matching_catch_2()|0;
  $182 = tempRet0;
  ___cxa_free_exception(($178|0)); //@line 526 "modules/image/magpie/PVRHandler.cpp"
  $$1 = $181;$$13 = $182;
  ___resumeException($$1|0);
  // unreachable;
 }
 __THREW__ = 0;
 $183 = (invoke_ii(3492,($totalsize$0$lcssa|0))|0); //@line 530 "modules/image/magpie/PVRHandler.cpp"
 $184 = __THREW__; __THREW__ = 0;
 $185 = $184&1;
 if ($185) {
  $198 = ___cxa_find_matching_catch_3(9056|0)|0;
  $199 = tempRet0;
  $200 = (_llvm_eh_typeid_for((9056|0))|0); //@line 531 "modules/image/magpie/PVRHandler.cpp"
  $201 = ($199|0)==($200|0); //@line 531 "modules/image/magpie/PVRHandler.cpp"
  if (!($201)) {
   $$1 = $198;$$13 = $199;
   ___resumeException($$1|0);
   // unreachable;
  }
  (___cxa_begin_catch(($198|0))|0); //@line 531 "modules/image/magpie/PVRHandler.cpp"
  $202 = (___cxa_allocate_exception(16)|0); //@line 534 "modules/image/magpie/PVRHandler.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($202|0),(507610|0),($vararg_buffer7|0)); //@line 534 "modules/image/magpie/PVRHandler.cpp"
  $203 = __THREW__; __THREW__ = 0;
  $204 = $203&1;
  if ($204) {
   $208 = ___cxa_find_matching_catch_2()|0;
   $209 = tempRet0;
   ___cxa_free_exception(($202|0)); //@line 534 "modules/image/magpie/PVRHandler.cpp"
   $$0 = $208;$$02 = $209;
  } else {
   __THREW__ = 0;
   invoke_viii(3350,($202|0),(96|0),(344|0)); //@line 534 "modules/image/magpie/PVRHandler.cpp"
   $205 = __THREW__; __THREW__ = 0;
   $206 = ___cxa_find_matching_catch_2()|0;
   $207 = tempRet0;
   $$0 = $206;$$02 = $207;
  }
  __THREW__ = 0;
  invoke_v(3378); //@line 535 "modules/image/magpie/PVRHandler.cpp"
  $210 = __THREW__; __THREW__ = 0;
  $211 = $210&1;
  if ($211) {
   $237 = ___cxa_find_matching_catch_3(0|0)|0;
   $238 = tempRet0;
   ___clang_call_terminate($237); //@line 535 "modules/image/magpie/PVRHandler.cpp"
   // unreachable; //@line 535 "modules/image/magpie/PVRHandler.cpp"
  } else {
   $$1 = $$0;$$13 = $$02;
   ___resumeException($$1|0);
   // unreachable;
  }
 }
 $186 = HEAP32[$filedata>>2]|0; //@line 538 "modules/image/magpie/PVRHandler.cpp"
 $187 = ((($186)) + 16|0); //@line 538 "modules/image/magpie/PVRHandler.cpp"
 $188 = HEAP32[$187>>2]|0; //@line 538 "modules/image/magpie/PVRHandler.cpp"
 $189 = (FUNCTION_TABLE_ii[$188 & 8191]($filedata)|0); //@line 538 "modules/image/magpie/PVRHandler.cpp"
 $190 = (($189) + ($167)|0); //@line 538 "modules/image/magpie/PVRHandler.cpp"
 if (!($159)) {
  HEAP32[$dataSize>>2] = $totalsize$0$lcssa; //@line 560 "modules/image/magpie/PVRHandler.cpp"
  HEAP32[$format>>2] = $$0$i$ph; //@line 561 "modules/image/magpie/PVRHandler.cpp"
  $233 = ((($header3)) + 16|0); //@line 562 "modules/image/magpie/PVRHandler.cpp"
  $234 = HEAP32[$233>>2]|0; //@line 562 "modules/image/magpie/PVRHandler.cpp"
  $235 = ($234|0)==(1); //@line 562 "modules/image/magpie/PVRHandler.cpp"
  $236 = $235&1; //@line 562 "modules/image/magpie/PVRHandler.cpp"
  HEAP8[$sRGB>>0] = $236; //@line 562 "modules/image/magpie/PVRHandler.cpp"
  STACKTOP = sp;return ($183|0); //@line 564 "modules/image/magpie/PVRHandler.cpp"
 }
 $191 = ((($header3)) + 28|0); //@line 548 "modules/image/magpie/PVRHandler.cpp"
 $192 = ((($header3)) + 24|0); //@line 549 "modules/image/magpie/PVRHandler.cpp"
 $193 = ((($mip)) + 4|0); //@line 549 "modules/image/magpie/PVRHandler.cpp"
 $194 = ((($mip)) + 8|0); //@line 550 "modules/image/magpie/PVRHandler.cpp"
 $195 = ((($mip)) + 12|0); //@line 553 "modules/image/magpie/PVRHandler.cpp"
 $196 = ((($images)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $197 = ((($images)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $curoffset$010 = 0;$i1$011 = 0;
 while(1) {
  $212 = (__ZN4love5image6magpie12_GLOBAL__N_115getMipLevelSizeERKNS2_14PVRTexHeaderV3Ei($header3,$i1$011)|0); //@line 542 "modules/image/magpie/PVRHandler.cpp"
  $213 = (($212) + ($curoffset$010))|0; //@line 544 "modules/image/magpie/PVRHandler.cpp"
  $214 = ($213>>>0)>($totalsize$0$lcssa>>>0); //@line 544 "modules/image/magpie/PVRHandler.cpp"
  if ($214) {
   label = 79;
   break;
  }
  $215 = HEAP32[$191>>2]|0; //@line 548 "modules/image/magpie/PVRHandler.cpp"
  $216 = $215 >> $i1$011; //@line 548 "modules/image/magpie/PVRHandler.cpp"
  $217 = ($216|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $218 = $217 ? 1 : $216; //@line 548 "modules/image/magpie/PVRHandler.cpp"
  HEAP32[$mip>>2] = $218; //@line 548 "modules/image/magpie/PVRHandler.cpp"
  $219 = HEAP32[$192>>2]|0; //@line 549 "modules/image/magpie/PVRHandler.cpp"
  $220 = $219 >> $i1$011; //@line 549 "modules/image/magpie/PVRHandler.cpp"
  $221 = ($220|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $222 = $221 ? 1 : $220; //@line 549 "modules/image/magpie/PVRHandler.cpp"
  HEAP32[$193>>2] = $222; //@line 549 "modules/image/magpie/PVRHandler.cpp"
  HEAP32[$194>>2] = $212; //@line 550 "modules/image/magpie/PVRHandler.cpp"
  $223 = (($183) + ($curoffset$010)|0); //@line 552 "modules/image/magpie/PVRHandler.cpp"
  $224 = (($190) + ($curoffset$010)|0); //@line 552 "modules/image/magpie/PVRHandler.cpp"
  _memcpy(($223|0),($224|0),($212|0))|0; //@line 552 "modules/image/magpie/PVRHandler.cpp"
  HEAP32[$195>>2] = $223; //@line 553 "modules/image/magpie/PVRHandler.cpp"
  $225 = HEAP32[$196>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $226 = HEAP32[$197>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $227 = ($225|0)==($226|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($227) {
   __ZNSt3__16vectorIN4love5image19CompressedImageData8SubImageENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_($images,$mip); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  } else {
   ;HEAP32[$225>>2]=HEAP32[$mip>>2]|0;HEAP32[$225+4>>2]=HEAP32[$mip+4>>2]|0;HEAP32[$225+8>>2]=HEAP32[$mip+8>>2]|0;HEAP32[$225+12>>2]=HEAP32[$mip+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $228 = HEAP32[$196>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $229 = ((($228)) + 16|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$196>>2] = $229; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  $230 = (($i1$011) + 1)|0; //@line 540 "modules/image/magpie/PVRHandler.cpp"
  $231 = HEAP32[$157>>2]|0; //@line 540 "modules/image/magpie/PVRHandler.cpp"
  $232 = ($230|0)<($231|0); //@line 540 "modules/image/magpie/PVRHandler.cpp"
  if ($232) {
   $curoffset$010 = $213;$i1$011 = $230;
  } else {
   label = 79;
   break;
  }
 }
 if ((label|0) == 79) {
  HEAP32[$dataSize>>2] = $totalsize$0$lcssa; //@line 560 "modules/image/magpie/PVRHandler.cpp"
  HEAP32[$format>>2] = $$0$i$ph; //@line 561 "modules/image/magpie/PVRHandler.cpp"
  $233 = ((($header3)) + 16|0); //@line 562 "modules/image/magpie/PVRHandler.cpp"
  $234 = HEAP32[$233>>2]|0; //@line 562 "modules/image/magpie/PVRHandler.cpp"
  $235 = ($234|0)==(1); //@line 562 "modules/image/magpie/PVRHandler.cpp"
  $236 = $235&1; //@line 562 "modules/image/magpie/PVRHandler.cpp"
  HEAP8[$sRGB>>0] = $236; //@line 562 "modules/image/magpie/PVRHandler.cpp"
  STACKTOP = sp;return ($183|0); //@line 564 "modules/image/magpie/PVRHandler.cpp"
 }
 return (0)|0;
}
function __ZN4love5image6magpie12_GLOBAL__N_115getMipLevelSizeERKNS2_14PVRTexHeaderV3Ei($header,$miplevel) {
 $header = $header|0;
 $miplevel = $miplevel|0;
 var $$0 = 0, $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, $cond = 0, $cond1 = 0, $cond10 = 0, $cond11 = 0, $cond12 = 0, $cond13 = 0, $cond14 = 0, $cond15 = 0, $cond16 = 0, $cond17 = 0, $cond18 = 0, $cond19 = 0, $cond2 = 0, $cond20 = 0;
 var $cond21 = 0, $cond22 = 0, $cond23 = 0, $cond24 = 0, $cond25 = 0, $cond26 = 0, $cond27 = 0, $cond28 = 0, $cond29 = 0, $cond3 = 0, $cond30 = 0, $cond31 = 0, $cond32 = 0, $cond33 = 0, $cond34 = 0, $cond4 = 0, $cond5 = 0, $cond6 = 0, $cond7 = 0, $cond8 = 0;
 var $cond9 = 0, $smallestheight$0 = 0, $smallestwidth$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($header)) + 8|0); //@line 436 "modules/image/magpie/PVRHandler.cpp"
 $1 = $0; //@line 447 "modules/image/magpie/PVRHandler.cpp"
 $2 = $1; //@line 447 "modules/image/magpie/PVRHandler.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 447 "modules/image/magpie/PVRHandler.cpp"
 $4 = (($1) + 4)|0; //@line 447 "modules/image/magpie/PVRHandler.cpp"
 $5 = $4; //@line 447 "modules/image/magpie/PVRHandler.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 447 "modules/image/magpie/PVRHandler.cpp"
 do {
  switch ($3|0) {
  case 0:  {
   $cond34 = ($6|0)==(0);
   if ($cond34) {
    $smallestheight$0 = 8;$smallestwidth$0 = 16;
   } else {
    label = 55;
   }
   break;
  }
  case 1:  {
   $cond33 = ($6|0)==(0);
   if ($cond33) {
    $smallestheight$0 = 8;$smallestwidth$0 = 16;
   } else {
    label = 55;
   }
   break;
  }
  case 2:  {
   $cond32 = ($6|0)==(0);
   if ($cond32) {
    label = 37;
   } else {
    label = 55;
   }
   break;
  }
  case 3:  {
   $cond31 = ($6|0)==(0);
   if ($cond31) {
    label = 37;
   } else {
    label = 55;
   }
   break;
  }
  case 4:  {
   $cond30 = ($6|0)==(0);
   if ($cond30) {
    $smallestheight$0 = 4;$smallestwidth$0 = 8;
   } else {
    label = 55;
   }
   break;
  }
  case 5:  {
   $cond29 = ($6|0)==(0);
   if ($cond29) {
    $smallestheight$0 = 4;$smallestwidth$0 = 4;
   } else {
    label = 55;
   }
   break;
  }
  case 6:  {
   $cond28 = ($6|0)==(0);
   if ($cond28) {
    label = 40;
   } else {
    label = 55;
   }
   break;
  }
  case 7:  {
   $cond27 = ($6|0)==(0);
   if ($cond27) {
    label = 40;
   } else {
    label = 55;
   }
   break;
  }
  case 8:  {
   $cond26 = ($6|0)==(0);
   if ($cond26) {
    label = 40;
   } else {
    label = 55;
   }
   break;
  }
  case 9:  {
   $cond25 = ($6|0)==(0);
   if ($cond25) {
    label = 40;
   } else {
    label = 55;
   }
   break;
  }
  case 10:  {
   $cond24 = ($6|0)==(0);
   if ($cond24) {
    label = 40;
   } else {
    label = 55;
   }
   break;
  }
  case 11:  {
   $cond23 = ($6|0)==(0);
   if ($cond23) {
    label = 40;
   } else {
    label = 55;
   }
   break;
  }
  case 12:  {
   $cond22 = ($6|0)==(0);
   if ($cond22) {
    label = 40;
   } else {
    label = 55;
   }
   break;
  }
  case 13:  {
   $cond21 = ($6|0)==(0);
   if ($cond21) {
    label = 40;
   } else {
    label = 55;
   }
   break;
  }
  case 14:  {
   $cond20 = ($6|0)==(0);
   if ($cond20) {
    label = 40;
   } else {
    label = 55;
   }
   break;
  }
  case 15:  {
   $cond19 = ($6|0)==(0);
   if ($cond19) {
    label = 40;
   } else {
    label = 55;
   }
   break;
  }
  case 22:  {
   $cond18 = ($6|0)==(0);
   if ($cond18) {
    label = 40;
   } else {
    label = 55;
   }
   break;
  }
  case 23:  {
   $cond17 = ($6|0)==(0);
   if ($cond17) {
    label = 40;
   } else {
    label = 55;
   }
   break;
  }
  case 24:  {
   $cond16 = ($6|0)==(0);
   if ($cond16) {
    label = 40;
   } else {
    label = 55;
   }
   break;
  }
  case 25:  {
   $cond15 = ($6|0)==(0);
   if ($cond15) {
    label = 40;
   } else {
    label = 55;
   }
   break;
  }
  case 26:  {
   $cond14 = ($6|0)==(0);
   if ($cond14) {
    label = 40;
   } else {
    label = 55;
   }
   break;
  }
  case 27:  {
   $cond13 = ($6|0)==(0);
   if ($cond13) {
    $smallestheight$0 = 4;$smallestwidth$0 = 4;
   } else {
    label = 55;
   }
   break;
  }
  case 28:  {
   $cond12 = ($6|0)==(0);
   if ($cond12) {
    $smallestheight$0 = 4;$smallestwidth$0 = 5;
   } else {
    label = 55;
   }
   break;
  }
  case 29:  {
   $cond11 = ($6|0)==(0);
   if ($cond11) {
    $smallestheight$0 = 5;$smallestwidth$0 = 5;
   } else {
    label = 55;
   }
   break;
  }
  case 30:  {
   $cond10 = ($6|0)==(0);
   if ($cond10) {
    $smallestheight$0 = 5;$smallestwidth$0 = 6;
   } else {
    label = 55;
   }
   break;
  }
  case 31:  {
   $cond9 = ($6|0)==(0);
   if ($cond9) {
    $smallestheight$0 = 6;$smallestwidth$0 = 6;
   } else {
    label = 55;
   }
   break;
  }
  case 32:  {
   $cond8 = ($6|0)==(0);
   if ($cond8) {
    $smallestheight$0 = 5;$smallestwidth$0 = 8;
   } else {
    label = 55;
   }
   break;
  }
  case 33:  {
   $cond7 = ($6|0)==(0);
   if ($cond7) {
    $smallestheight$0 = 6;$smallestwidth$0 = 8;
   } else {
    label = 55;
   }
   break;
  }
  case 34:  {
   $cond6 = ($6|0)==(0);
   if ($cond6) {
    $smallestheight$0 = 8;$smallestwidth$0 = 8;
   } else {
    label = 55;
   }
   break;
  }
  case 35:  {
   $cond5 = ($6|0)==(0);
   if ($cond5) {
    $smallestheight$0 = 5;$smallestwidth$0 = 10;
   } else {
    label = 55;
   }
   break;
  }
  case 36:  {
   $cond4 = ($6|0)==(0);
   if ($cond4) {
    $smallestheight$0 = 6;$smallestwidth$0 = 10;
   } else {
    label = 55;
   }
   break;
  }
  case 37:  {
   $cond3 = ($6|0)==(0);
   if ($cond3) {
    $smallestheight$0 = 8;$smallestwidth$0 = 10;
   } else {
    label = 55;
   }
   break;
  }
  case 38:  {
   $cond2 = ($6|0)==(0);
   if ($cond2) {
    $smallestheight$0 = 10;$smallestwidth$0 = 10;
   } else {
    label = 55;
   }
   break;
  }
  case 39:  {
   $cond1 = ($6|0)==(0);
   if ($cond1) {
    $smallestheight$0 = 10;$smallestwidth$0 = 12;
   } else {
    label = 55;
   }
   break;
  }
  case 40:  {
   $cond = ($6|0)==(0);
   if ($cond) {
    $smallestheight$0 = 12;$smallestwidth$0 = 12;
   } else {
    label = 55;
   }
   break;
  }
  default: {
   label = 55;
  }
  }
 } while(0);
 if ((label|0) == 37) {
  $smallestheight$0 = 8;$smallestwidth$0 = 8;
 }
 else if ((label|0) == 40) {
  $smallestheight$0 = 4;$smallestwidth$0 = 4;
 }
 else if ((label|0) == 55) {
  $smallestheight$0 = 1;$smallestwidth$0 = 1;
 }
 $7 = ((($header)) + 28|0); //@line 438 "modules/image/magpie/PVRHandler.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 438 "modules/image/magpie/PVRHandler.cpp"
 $9 = $8 >> $miplevel; //@line 438 "modules/image/magpie/PVRHandler.cpp"
 $10 = ($9|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $11 = $10 ? 1 : $9; //@line 438 "modules/image/magpie/PVRHandler.cpp"
 $12 = ((($header)) + 24|0); //@line 439 "modules/image/magpie/PVRHandler.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 439 "modules/image/magpie/PVRHandler.cpp"
 $14 = $13 >> $miplevel; //@line 439 "modules/image/magpie/PVRHandler.cpp"
 $15 = ($14|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $16 = $15 ? 1 : $14; //@line 439 "modules/image/magpie/PVRHandler.cpp"
 $17 = ((($header)) + 32|0); //@line 440 "modules/image/magpie/PVRHandler.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 440 "modules/image/magpie/PVRHandler.cpp"
 $19 = $18 >> $miplevel; //@line 440 "modules/image/magpie/PVRHandler.cpp"
 $20 = ($19|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $21 = $20 ? 1 : $19; //@line 440 "modules/image/magpie/PVRHandler.cpp"
 $22 = (($smallestwidth$0) + -1)|0; //@line 443 "modules/image/magpie/PVRHandler.cpp"
 $23 = (($22) + ($11))|0; //@line 443 "modules/image/magpie/PVRHandler.cpp"
 $24 = (($23|0) / ($smallestwidth$0|0))&-1; //@line 443 "modules/image/magpie/PVRHandler.cpp"
 $25 = (($smallestheight$0) + -1)|0; //@line 444 "modules/image/magpie/PVRHandler.cpp"
 $26 = (($25) + ($16))|0; //@line 444 "modules/image/magpie/PVRHandler.cpp"
 $27 = (($26|0) / ($smallestheight$0|0))&-1; //@line 444 "modules/image/magpie/PVRHandler.cpp"
 $28 = (_i64Add(($3|0),($6|0),-27,-1)|0); //@line 447 "modules/image/magpie/PVRHandler.cpp"
 $29 = tempRet0; //@line 447 "modules/image/magpie/PVRHandler.cpp"
 $30 = ($29>>>0)<(0); //@line 447 "modules/image/magpie/PVRHandler.cpp"
 $31 = ($28>>>0)<(14); //@line 447 "modules/image/magpie/PVRHandler.cpp"
 $32 = ($29|0)==(0); //@line 447 "modules/image/magpie/PVRHandler.cpp"
 $33 = $32 & $31; //@line 447 "modules/image/magpie/PVRHandler.cpp"
 $34 = $30 | $33; //@line 447 "modules/image/magpie/PVRHandler.cpp"
 if ($34) {
  $35 = $24 << 4; //@line 448 "modules/image/magpie/PVRHandler.cpp"
  $36 = Math_imul($35, $27)|0; //@line 448 "modules/image/magpie/PVRHandler.cpp"
  $37 = Math_imul($36, $21)|0; //@line 448 "modules/image/magpie/PVRHandler.cpp"
  $$0 = $37;
  return ($$0|0); //@line 451 "modules/image/magpie/PVRHandler.cpp"
 }
 $38 = (_bitshift64Lshr(($3|0),($6|0),56)|0); //@line 450 "modules/image/magpie/PVRHandler.cpp"
 $39 = tempRet0; //@line 450 "modules/image/magpie/PVRHandler.cpp"
 $40 = (_bitshift64Shl(($38|0),($39|0),56)|0); //@line 293 "modules/image/magpie/PVRHandler.cpp"
 $41 = tempRet0; //@line 293 "modules/image/magpie/PVRHandler.cpp"
 $42 = $6 & 16777215; //@line 293 "modules/image/magpie/PVRHandler.cpp"
 $43 = $40 | $3; //@line 293 "modules/image/magpie/PVRHandler.cpp"
 $44 = $41 | $42; //@line 293 "modules/image/magpie/PVRHandler.cpp"
 $45 = ($3|0)==($43|0); //@line 293 "modules/image/magpie/PVRHandler.cpp"
 $46 = (0)==($44|0); //@line 293 "modules/image/magpie/PVRHandler.cpp"
 $47 = $45 & $46; //@line 293 "modules/image/magpie/PVRHandler.cpp"
 L62: do {
  if ($47) {
   switch ($3|0) {
   case 4: case 1: case 0:  {
    $$0$i = 2;
    break L62;
    break;
   }
   case 25: case 24: case 22: case 12: case 7: case 6: case 5: case 3: case 2:  {
    $$0$i = 4;
    break L62;
    break;
   }
   case 26: case 23: case 15: case 14: case 13: case 11: case 10: case 9: case 8:  {
    $$0$i = 8;
    break L62;
    break;
   }
   default: {
    $$0$i = 0;
    break L62;
   }
   }
  } else {
   $48 = (_bitshift64Lshr(($3|0),($6|0),40)|0); //@line 450 "modules/image/magpie/PVRHandler.cpp"
   $49 = tempRet0; //@line 450 "modules/image/magpie/PVRHandler.cpp"
   $50 = (_bitshift64Lshr(($3|0),($6|0),48)|0); //@line 450 "modules/image/magpie/PVRHandler.cpp"
   $51 = tempRet0; //@line 450 "modules/image/magpie/PVRHandler.cpp"
   $52 = $6 & 255; //@line 296 "modules/image/magpie/PVRHandler.cpp"
   $53 = $48 & 255; //@line 296 "modules/image/magpie/PVRHandler.cpp"
   $54 = $50 & 255; //@line 296 "modules/image/magpie/PVRHandler.cpp"
   $55 = (($53) + ($38))|0; //@line 296 "modules/image/magpie/PVRHandler.cpp"
   $56 = (($55) + ($52))|0; //@line 296 "modules/image/magpie/PVRHandler.cpp"
   $57 = (($56) + ($54))|0; //@line 296 "modules/image/magpie/PVRHandler.cpp"
   $$0$i = $57;
  }
 } while(0);
 $58 = Math_imul($smallestwidth$0, $smallestheight$0)|0; //@line 443 "modules/image/magpie/PVRHandler.cpp"
 $59 = Math_imul($58, $24)|0; //@line 450 "modules/image/magpie/PVRHandler.cpp"
 $60 = Math_imul($59, $27)|0; //@line 444 "modules/image/magpie/PVRHandler.cpp"
 $61 = Math_imul($60, $21)|0; //@line 450 "modules/image/magpie/PVRHandler.cpp"
 $62 = Math_imul($61, $$0$i)|0; //@line 450 "modules/image/magpie/PVRHandler.cpp"
 $63 = (($62|0) / 8)&-1; //@line 450 "modules/image/magpie/PVRHandler.cpp"
 $$0 = $63;
 return ($$0|0); //@line 451 "modules/image/magpie/PVRHandler.cpp"
}
function __ZN4love5image6magpie10STBHandlerD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love5image6magpie13FormatHandlerD2Ev($this); //@line 40 "modules/image/magpie/STBHandler.h"
 __ZdlPv($this); //@line 40 "modules/image/magpie/STBHandler.h"
 return; //@line 40 "modules/image/magpie/STBHandler.h"
}
function __ZN4love5image6magpie10STBHandler9canDecodeEPNS_10filesystem8FileDataE($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $comp = 0, $h = 0, $or$cond = 0, $s$i = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 208|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $s$i = sp + 16|0;
 $w = sp + 8|0;
 $h = sp + 4|0;
 $comp = sp;
 HEAP32[$w>>2] = 0; //@line 52 "modules/image/magpie/STBHandler.cpp"
 HEAP32[$h>>2] = 0; //@line 53 "modules/image/magpie/STBHandler.cpp"
 HEAP32[$comp>>2] = 0; //@line 54 "modules/image/magpie/STBHandler.cpp"
 $0 = HEAP32[$data>>2]|0; //@line 56 "modules/image/magpie/STBHandler.cpp"
 $1 = ((($0)) + 16|0); //@line 56 "modules/image/magpie/STBHandler.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 56 "modules/image/magpie/STBHandler.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($data)|0); //@line 56 "modules/image/magpie/STBHandler.cpp"
 $4 = HEAP32[$data>>2]|0; //@line 57 "modules/image/magpie/STBHandler.cpp"
 $5 = ((($4)) + 20|0); //@line 57 "modules/image/magpie/STBHandler.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 57 "modules/image/magpie/STBHandler.cpp"
 $7 = (FUNCTION_TABLE_ii[$6 & 8191]($data)|0); //@line 57 "modules/image/magpie/STBHandler.cpp"
 $8 = ((($s$i)) + 16|0); //@line 770 "./libraries/stb/stb_image.h"
 HEAP32[$8>>2] = 0; //@line 770 "./libraries/stb/stb_image.h"
 $9 = ((($s$i)) + 32|0); //@line 771 "./libraries/stb/stb_image.h"
 HEAP32[$9>>2] = 0; //@line 771 "./libraries/stb/stb_image.h"
 $10 = ((($s$i)) + 176|0); //@line 772 "./libraries/stb/stb_image.h"
 HEAP32[$10>>2] = $3; //@line 772 "./libraries/stb/stb_image.h"
 $11 = ((($s$i)) + 168|0); //@line 772 "./libraries/stb/stb_image.h"
 HEAP32[$11>>2] = $3; //@line 772 "./libraries/stb/stb_image.h"
 $12 = (($3) + ($7)|0); //@line 773 "./libraries/stb/stb_image.h"
 $13 = ((($s$i)) + 180|0); //@line 773 "./libraries/stb/stb_image.h"
 HEAP32[$13>>2] = $12; //@line 773 "./libraries/stb/stb_image.h"
 $14 = ((($s$i)) + 172|0); //@line 773 "./libraries/stb/stb_image.h"
 HEAP32[$14>>2] = $12; //@line 773 "./libraries/stb/stb_image.h"
 $15 = (__ZL15stbi__info_mainP13stbi__contextPiS1_S1_($s$i,$w,$h,$comp)|0); //@line 6345 "./libraries/stb/stb_image.h"
 $16 = ($15|0)==(1); //@line 59 "modules/image/magpie/STBHandler.cpp"
 $17 = HEAP32[$w>>2]|0; //@line 59 "modules/image/magpie/STBHandler.cpp"
 $18 = ($17|0)>(0); //@line 59 "modules/image/magpie/STBHandler.cpp"
 $or$cond = $16 & $18; //@line 59 "modules/image/magpie/STBHandler.cpp"
 if (!($or$cond)) {
  $21 = 0;
  STACKTOP = sp;return ($21|0); //@line 59 "modules/image/magpie/STBHandler.cpp"
 }
 $19 = HEAP32[$h>>2]|0; //@line 59 "modules/image/magpie/STBHandler.cpp"
 $20 = ($19|0)>(0); //@line 59 "modules/image/magpie/STBHandler.cpp"
 $21 = $20;
 STACKTOP = sp;return ($21|0); //@line 59 "modules/image/magpie/STBHandler.cpp"
}
function __ZL15stbi__info_mainP13stbi__contextPiS1_S1_($s,$x,$y,$comp) {
 $s = $s|0;
 $x = $x|0;
 $y = $y|0;
 $comp = $comp|0;
 var $$0 = 0, $$0$i$i$i = 0, $$0$i$i$i16 = 0, $$0$i$i1 = 0, $$0$i$i22$i = 0, $$0$i$i36$i = 0, $$0$i$i46$i = 0, $$0$i$i50$i = 0, $$0$i13$i = 0, $$0$i23$i = 0, $$0$i30$i = 0, $$0$i37$i = 0, $$0$i6$i$i = 0, $$0$i6$i$i23 = 0, $$0$i6$i29$i = 0, $$0$i6$i43$i = 0, $$0$i6$i53$i = 0, $$0$i6$i57$i = 0, $$in = 0, $$in26 = 0;
 var $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i11 = 0, $$pre$i$i17$i = 0, $$pre$i$i31$i = 0, $$pre$i$i41$i = 0, $$pre$i$i45$i = 0, $$pre$i1$i$i = 0, $$pre$i1$i$i18 = 0, $$pre$i1$i24$i = 0, $$pre$i1$i38$i = 0, $$pre$i1$i48$i = 0, $$pre$i1$i52$i = 0, $$pre$i15$i = 0, $$pre$i18$i = 0, $$pre$i25$i = 0, $$pre$i32$i = 0, $$pre$i39$i = 0, $$pre$i7 = 0;
 var $$pre$i8$i = 0, $$pre$phi$i$i$i13Z2D = 0, $$pre$phi$i$i$iZ2D = 0, $$pre$phi$i$i19$iZ2D = 0, $$pre$phi$i$i33$iZ2D = 0, $$pre$phi$i$i3Z2D = 0, $$pre$phi$i$i43$iZ2D = 0, $$pre$phi$i$i47$iZ2D = 0, $$pre$phi$i$iZ2D = 0, $$pre$phi$i10$iZ2D = 0, $$pre$phi$i20$iZ2D = 0, $$pre$phi$i27$iZ2D = 0, $$pre$phi$i3$i$i20Z2D = 0, $$pre$phi$i3$i$iZ2D = 0, $$pre$phi$i3$i26$iZ2D = 0, $$pre$phi$i3$i40$iZ2D = 0, $$pre$phi$i3$i50$iZ2D = 0, $$pre$phi$i3$i54$iZ2D = 0, $$pre$phi$i34$iZ2D = 0, $$pre1$i$i = 0;
 var $$pre1$i$i$i = 0, $$pre1$i$i$i12 = 0, $$pre1$i$i18$i = 0, $$pre1$i$i2 = 0, $$pre1$i$i32$i = 0, $$pre1$i$i42$i = 0, $$pre1$i$i46$i = 0, $$pre1$i19$i = 0, $$pre1$i2$i$i = 0, $$pre1$i2$i$i19 = 0, $$pre1$i2$i25$i = 0, $$pre1$i2$i39$i = 0, $$pre1$i2$i49$i = 0, $$pre1$i2$i53$i = 0, $$pre1$i26$i = 0, $$pre1$i33$i = 0, $$pre1$i9$i = 0, $$pre56$i = 0, $$pre60$i = 0, $$sink$i$i$i = 0;
 var $$sink$i$i$i$i = 0, $$sink$i$i$i$i14 = 0, $$sink$i$i$i20$i = 0, $$sink$i$i$i34$i = 0, $$sink$i$i$i4 = 0, $$sink$i$i$i44$i = 0, $$sink$i$i$i48$i = 0, $$sink$i$i11$i = 0, $$sink$i$i21$i = 0, $$sink$i$i28$i = 0, $$sink$i$i35$i = 0, $$sink$i$i4$i$i = 0, $$sink$i$i4$i$i21 = 0, $$sink$i$i4$i27$i = 0, $$sink$i$i4$i41$i = 0, $$sink$i$i4$i51$i = 0, $$sink$i$i4$i55$i = 0, $0 = 0, $1 = 0, $10 = 0;
 var $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0;
 var $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0;
 var $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0;
 var $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0;
 var $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0;
 var $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0;
 var $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0;
 var $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0;
 var $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0;
 var $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0;
 var $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0;
 var $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0;
 var $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0;
 var $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0;
 var $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0;
 var $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0;
 var $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0;
 var $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0;
 var $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0;
 var $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0;
 var $461 = 0, $462 = 0, $463 = 0, $464 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $div$i = 0, $j$i = 0, $storemerge$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 18192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $j$i = sp;
 HEAP32[$j$i>>2] = $s; //@line 3443 "./libraries/stb/stb_image.h"
 $0 = (__ZL24stbi__decode_jpeg_headerP10stbi__jpegi($j$i,2)|0); //@line 3430 "./libraries/stb/stb_image.h"
 $1 = ($0|0)==(0); //@line 3430 "./libraries/stb/stb_image.h"
 if (!($1)) {
  $2 = ($x|0)==(0|0); //@line 3434 "./libraries/stb/stb_image.h"
  if (!($2)) {
   $3 = HEAP32[$j$i>>2]|0; //@line 3434 "./libraries/stb/stb_image.h"
   $4 = HEAP32[$3>>2]|0; //@line 3434 "./libraries/stb/stb_image.h"
   HEAP32[$x>>2] = $4; //@line 3434 "./libraries/stb/stb_image.h"
  }
  $5 = ($y|0)==(0|0); //@line 3435 "./libraries/stb/stb_image.h"
  if (!($5)) {
   $6 = HEAP32[$j$i>>2]|0; //@line 3435 "./libraries/stb/stb_image.h"
   $7 = ((($6)) + 4|0); //@line 3435 "./libraries/stb/stb_image.h"
   $8 = HEAP32[$7>>2]|0; //@line 3435 "./libraries/stb/stb_image.h"
   HEAP32[$y>>2] = $8; //@line 3435 "./libraries/stb/stb_image.h"
  }
  $9 = ($comp|0)==(0|0); //@line 3436 "./libraries/stb/stb_image.h"
  if (!($9)) {
   $10 = HEAP32[$j$i>>2]|0; //@line 3436 "./libraries/stb/stb_image.h"
   $11 = ((($10)) + 8|0); //@line 3436 "./libraries/stb/stb_image.h"
   $12 = HEAP32[$11>>2]|0; //@line 3436 "./libraries/stb/stb_image.h"
   HEAP32[$comp>>2] = $12; //@line 3436 "./libraries/stb/stb_image.h"
  }
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 6316 "./libraries/stb/stb_image.h"
 }
 $13 = HEAP32[$j$i>>2]|0; //@line 3431 "./libraries/stb/stb_image.h"
 $14 = ((($13)) + 176|0); //@line 826 "./libraries/stb/stb_image.h"
 $15 = HEAP32[$14>>2]|0; //@line 826 "./libraries/stb/stb_image.h"
 $16 = ((($13)) + 168|0); //@line 826 "./libraries/stb/stb_image.h"
 HEAP32[$16>>2] = $15; //@line 826 "./libraries/stb/stb_image.h"
 $17 = ((($13)) + 180|0); //@line 827 "./libraries/stb/stb_image.h"
 $18 = HEAP32[$17>>2]|0; //@line 827 "./libraries/stb/stb_image.h"
 $19 = ((($13)) + 172|0); //@line 827 "./libraries/stb/stb_image.h"
 HEAP32[$19>>2] = $18; //@line 827 "./libraries/stb/stb_image.h"
 $20 = ((($s)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
 $21 = HEAP32[$20>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
 $22 = ((($s)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
 $23 = HEAP32[$22>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
 $24 = ($21>>>0)<($23>>>0); //@line 1229 "./libraries/stb/stb_image.h"
 if ($24) {
  $25 = ((($21)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
  HEAP32[$20>>2] = $25; //@line 1230 "./libraries/stb/stb_image.h"
  $26 = HEAP8[$21>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
  $$0$i$i1 = $26;$43 = $25;$44 = $23;
  label = 16;
 } else {
  $27 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
  $28 = HEAP32[$27>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
  $29 = ($28|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
  if ($29) {
   label = 25;
  } else {
   $30 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
   $31 = HEAP32[$30>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $32 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
   $33 = HEAP32[$32>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $34 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
   $35 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
   $36 = HEAP32[$35>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $37 = (FUNCTION_TABLE_iiii[$31 & 8191]($33,$34,$36)|0); //@line 1213 "./libraries/stb/stb_image.h"
   $38 = ($37|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
   if ($38) {
    HEAP32[$27>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
    $39 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
    HEAP8[$34>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
    $$pre$phi$i$iZ2D = $39;$$sink$i$i$i = $39;$444 = 0;
   } else {
    $40 = (((($s)) + 40|0) + ($37)|0); //@line 1223 "./libraries/stb/stb_image.h"
    $$pre$i$i = HEAP8[$34>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
    $$pre1$i$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
    $$pre$phi$i$iZ2D = $$pre1$i$i;$$sink$i$i$i = $40;$444 = $$pre$i$i;
   }
   HEAP32[$22>>2] = $$sink$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
   HEAP32[$20>>2] = $$pre$phi$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
   $$0$i$i1 = $444;$43 = $$pre$phi$i$iZ2D;$44 = $$sink$i$i$i;
   label = 16;
  }
 }
 L22: do {
  if ((label|0) == 16) {
   $41 = ($$0$i$i1<<24>>24)==(66); //@line 6053 "./libraries/stb/stb_image.h"
   if ($41) {
    $42 = ($43>>>0)<($44>>>0); //@line 1229 "./libraries/stb/stb_image.h"
    if ($42) {
     $45 = ((($43)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
     HEAP32[$20>>2] = $45; //@line 1230 "./libraries/stb/stb_image.h"
     $46 = HEAP8[$43>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
     $$0$i13$i = $46;$$pre$i15$i = $45;$71 = $44;
    } else {
     $47 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
     $48 = HEAP32[$47>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
     $49 = ($48|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
     if ($49) {
      label = 25;
      break;
     }
     $50 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
     $51 = HEAP32[$50>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $52 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $53 = HEAP32[$52>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $54 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $55 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $56 = HEAP32[$55>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $57 = (FUNCTION_TABLE_iiii[$51 & 8191]($53,$54,$56)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $58 = ($57|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($58) {
      HEAP32[$47>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $59 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$54>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i10$iZ2D = $59;$$sink$i$i11$i = $59;$445 = 0;
     } else {
      $60 = (((($s)) + 40|0) + ($57)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre$i8$i = HEAP8[$54>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
      $$pre1$i9$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i10$iZ2D = $$pre1$i9$i;$$sink$i$i11$i = $60;$445 = $$pre$i8$i;
     }
     HEAP32[$22>>2] = $$sink$i$i11$i; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$20>>2] = $$pre$phi$i10$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
     $$0$i13$i = $445;$$pre$i15$i = $$pre$phi$i10$iZ2D;$71 = $$sink$i$i11$i;
    }
    $61 = $$pre$i15$i; //@line 1257 "./libraries/stb/stb_image.h"
    $62 = ($$0$i13$i<<24>>24)==(77); //@line 6053 "./libraries/stb/stb_image.h"
    if ($62) {
     $67 = ((($s)) + 16|0); //@line 1256 "./libraries/stb/stb_image.h"
     $68 = HEAP32[$67>>2]|0; //@line 1256 "./libraries/stb/stb_image.h"
     $69 = ($68|0)==(0|0); //@line 1256 "./libraries/stb/stb_image.h"
     if ($69) {
      label = 29;
     } else {
      $70 = $71; //@line 1257 "./libraries/stb/stb_image.h"
      $72 = (($70) - ($61))|0; //@line 1257 "./libraries/stb/stb_image.h"
      $73 = ($72|0)<(12); //@line 1258 "./libraries/stb/stb_image.h"
      if ($73) {
       HEAP32[$20>>2] = $71; //@line 1259 "./libraries/stb/stb_image.h"
       $74 = ((($s)) + 20|0); //@line 1260 "./libraries/stb/stb_image.h"
       $75 = HEAP32[$74>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
       $76 = ((($s)) + 28|0); //@line 1260 "./libraries/stb/stb_image.h"
       $77 = HEAP32[$76>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
       $78 = (12 - ($72))|0; //@line 1260 "./libraries/stb/stb_image.h"
       FUNCTION_TABLE_vii[$75 & 8191]($77,$78); //@line 1260 "./libraries/stb/stb_image.h"
      } else {
       label = 29;
      }
     }
     if ((label|0) == 29) {
      $79 = ((($$pre$i15$i)) + 12|0); //@line 1264 "./libraries/stb/stb_image.h"
      HEAP32[$20>>2] = $79; //@line 1264 "./libraries/stb/stb_image.h"
     }
     $80 = (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 6058 "./libraries/stb/stb_image.h"
     switch ($80|0) {
     case 12:  {
      $85 = HEAP32[$20>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
      $86 = HEAP32[$22>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
      $87 = ($85>>>0)<($86>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      if ($87) {
       $88 = ((($85)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
       HEAP32[$20>>2] = $88; //@line 1230 "./libraries/stb/stb_image.h"
       $89 = HEAP8[$85>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
       $$0$i$i22$i = $89;$105 = $88;$106 = $86;
      } else {
       $90 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
       $91 = HEAP32[$90>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
       $92 = ($91|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
       if ($92) {
        $$0$i$i22$i = 0;$105 = $85;$106 = $86;
       } else {
        $93 = HEAP32[$67>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $94 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $95 = HEAP32[$94>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $96 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $97 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $98 = HEAP32[$97>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $99 = (FUNCTION_TABLE_iiii[$93 & 8191]($95,$96,$98)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $100 = ($99|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($100) {
         HEAP32[$90>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $101 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$96>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i19$iZ2D = $101;$$sink$i$i$i20$i = $101;$446 = 0;
        } else {
         $102 = (((($s)) + 40|0) + ($99)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i$i17$i = HEAP8[$96>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i$i18$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i19$iZ2D = $$pre1$i$i18$i;$$sink$i$i$i20$i = $102;$446 = $$pre$i$i17$i;
        }
        HEAP32[$22>>2] = $$sink$i$i$i20$i; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$20>>2] = $$pre$phi$i$i19$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i$i22$i = $446;$105 = $$pre$phi$i$i19$iZ2D;$106 = $$sink$i$i$i20$i;
       }
      }
      $103 = $$0$i$i22$i&255; //@line 1308 "./libraries/stb/stb_image.h"
      $104 = ($105>>>0)<($106>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      if ($104) {
       $107 = ((($105)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
       HEAP32[$20>>2] = $107; //@line 1230 "./libraries/stb/stb_image.h"
       $108 = HEAP8[$105>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
       $$0$i6$i29$i = $108;$126 = $107;$127 = $106;
      } else {
       $109 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
       $110 = HEAP32[$109>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
       $111 = ($110|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
       if ($111) {
        $$0$i6$i29$i = 0;$126 = $105;$127 = $106;
       } else {
        $112 = HEAP32[$67>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $113 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $114 = HEAP32[$113>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $115 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $116 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $117 = HEAP32[$116>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $118 = (FUNCTION_TABLE_iiii[$112 & 8191]($114,$115,$117)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $119 = ($118|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($119) {
         HEAP32[$109>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $120 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$115>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i26$iZ2D = $120;$$sink$i$i4$i27$i = $120;$447 = 0;
        } else {
         $121 = (((($s)) + 40|0) + ($118)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i1$i24$i = HEAP8[$115>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i2$i25$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i26$iZ2D = $$pre1$i2$i25$i;$$sink$i$i4$i27$i = $121;$447 = $$pre$i1$i24$i;
        }
        HEAP32[$22>>2] = $$sink$i$i4$i27$i; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$20>>2] = $$pre$phi$i3$i26$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i6$i29$i = $447;$126 = $$pre$phi$i3$i26$iZ2D;$127 = $$sink$i$i4$i27$i;
       }
      }
      $122 = $$0$i6$i29$i&255; //@line 1309 "./libraries/stb/stb_image.h"
      $123 = $122 << 8; //@line 1309 "./libraries/stb/stb_image.h"
      $124 = $123 | $103; //@line 1309 "./libraries/stb/stb_image.h"
      HEAP32[$x>>2] = $124; //@line 6064 "./libraries/stb/stb_image.h"
      $125 = ($126>>>0)<($127>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      if ($125) {
       $128 = ((($126)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
       HEAP32[$20>>2] = $128; //@line 1230 "./libraries/stb/stb_image.h"
       $129 = HEAP8[$126>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
       $$0$i$i36$i = $129;$145 = $128;$146 = $127;
      } else {
       $130 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
       $131 = HEAP32[$130>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
       $132 = ($131|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
       if ($132) {
        $$0$i$i36$i = 0;$145 = $126;$146 = $127;
       } else {
        $133 = HEAP32[$67>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $134 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $135 = HEAP32[$134>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $136 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $137 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $138 = HEAP32[$137>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $139 = (FUNCTION_TABLE_iiii[$133 & 8191]($135,$136,$138)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $140 = ($139|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($140) {
         HEAP32[$130>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $141 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$136>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i33$iZ2D = $141;$$sink$i$i$i34$i = $141;$448 = 0;
        } else {
         $142 = (((($s)) + 40|0) + ($139)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i$i31$i = HEAP8[$136>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i$i32$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i33$iZ2D = $$pre1$i$i32$i;$$sink$i$i$i34$i = $142;$448 = $$pre$i$i31$i;
        }
        HEAP32[$22>>2] = $$sink$i$i$i34$i; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$20>>2] = $$pre$phi$i$i33$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i$i36$i = $448;$145 = $$pre$phi$i$i33$iZ2D;$146 = $$sink$i$i$i34$i;
       }
      }
      $143 = $$0$i$i36$i&255; //@line 1308 "./libraries/stb/stb_image.h"
      $144 = ($145>>>0)<($146>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      if ($144) {
       $147 = ((($145)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
       HEAP32[$20>>2] = $147; //@line 1230 "./libraries/stb/stb_image.h"
       $148 = HEAP8[$145>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
       $$0$i6$i43$i = $148;$449 = $146;$450 = $147;
      } else {
       $149 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
       $150 = HEAP32[$149>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
       $151 = ($150|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
       if ($151) {
        $$0$i6$i43$i = 0;$449 = $146;$450 = $145;
       } else {
        $152 = HEAP32[$67>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $153 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $154 = HEAP32[$153>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $155 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $156 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $157 = HEAP32[$156>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $158 = (FUNCTION_TABLE_iiii[$152 & 8191]($154,$155,$157)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $159 = ($158|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($159) {
         HEAP32[$149>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $160 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$155>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i40$iZ2D = $160;$$sink$i$i4$i41$i = $160;$451 = 0;
        } else {
         $161 = (((($s)) + 40|0) + ($158)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i1$i38$i = HEAP8[$155>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i2$i39$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i40$iZ2D = $$pre1$i2$i39$i;$$sink$i$i4$i41$i = $161;$451 = $$pre$i1$i38$i;
        }
        HEAP32[$22>>2] = $$sink$i$i4$i41$i; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$20>>2] = $$pre$phi$i3$i40$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i6$i43$i = $451;$449 = $$sink$i$i4$i41$i;$450 = $$pre$phi$i3$i40$iZ2D;
       }
      }
      $162 = $$0$i6$i43$i&255; //@line 1309 "./libraries/stb/stb_image.h"
      $163 = $162 << 8; //@line 1309 "./libraries/stb/stb_image.h"
      $164 = $163 | $143; //@line 1309 "./libraries/stb/stb_image.h"
      $168 = $450;$169 = $449;$storemerge$i = $164;
      break;
     }
     case 124: case 108: case 56: case 40:  {
      $165 = (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 6067 "./libraries/stb/stb_image.h"
      HEAP32[$x>>2] = $165; //@line 6067 "./libraries/stb/stb_image.h"
      $166 = (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 6068 "./libraries/stb/stb_image.h"
      $$pre$i = HEAP32[$20>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
      $$pre60$i = HEAP32[$22>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
      $168 = $$pre$i;$169 = $$pre60$i;$storemerge$i = $166;
      break;
     }
     default: {
      $81 = ((($s)) + 176|0); //@line 826 "./libraries/stb/stb_image.h"
      $82 = HEAP32[$81>>2]|0; //@line 826 "./libraries/stb/stb_image.h"
      HEAP32[$20>>2] = $82; //@line 826 "./libraries/stb/stb_image.h"
      $83 = ((($s)) + 180|0); //@line 827 "./libraries/stb/stb_image.h"
      $84 = HEAP32[$83>>2]|0; //@line 827 "./libraries/stb/stb_image.h"
      HEAP32[$22>>2] = $84; //@line 827 "./libraries/stb/stb_image.h"
      $$in = $84;$$in26 = $82;
      break L22;
     }
     }
     HEAP32[$y>>2] = $storemerge$i; //@line 6065 "./libraries/stb/stb_image.h"
     $167 = ($168>>>0)<($169>>>0); //@line 1229 "./libraries/stb/stb_image.h"
     if ($167) {
      $170 = ((($168)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
      HEAP32[$20>>2] = $170; //@line 1230 "./libraries/stb/stb_image.h"
      $171 = HEAP8[$168>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
      $$0$i$i50$i = $171;$187 = $170;$188 = $169;
     } else {
      $172 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
      $173 = HEAP32[$172>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
      $174 = ($173|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
      if ($174) {
       $$0$i$i50$i = 0;$187 = $168;$188 = $169;
      } else {
       $175 = HEAP32[$67>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $176 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
       $177 = HEAP32[$176>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $178 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
       $179 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
       $180 = HEAP32[$179>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $181 = (FUNCTION_TABLE_iiii[$175 & 8191]($177,$178,$180)|0); //@line 1213 "./libraries/stb/stb_image.h"
       $182 = ($181|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
       if ($182) {
        HEAP32[$172>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
        $183 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
        HEAP8[$178>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
        $$pre$phi$i$i47$iZ2D = $183;$$sink$i$i$i48$i = $183;$452 = 0;
       } else {
        $184 = (((($s)) + 40|0) + ($181)|0); //@line 1223 "./libraries/stb/stb_image.h"
        $$pre$i$i45$i = HEAP8[$178>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
        $$pre1$i$i46$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
        $$pre$phi$i$i47$iZ2D = $$pre1$i$i46$i;$$sink$i$i$i48$i = $184;$452 = $$pre$i$i45$i;
       }
       HEAP32[$22>>2] = $$sink$i$i$i48$i; //@line 1223 "./libraries/stb/stb_image.h"
       HEAP32[$20>>2] = $$pre$phi$i$i47$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
       $$0$i$i50$i = $452;$187 = $$pre$phi$i$i47$iZ2D;$188 = $$sink$i$i$i48$i;
      }
     }
     $185 = $$0$i$i50$i&255; //@line 1308 "./libraries/stb/stb_image.h"
     $186 = ($187>>>0)<($188>>>0); //@line 1229 "./libraries/stb/stb_image.h"
     if ($186) {
      $189 = ((($187)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
      HEAP32[$20>>2] = $189; //@line 1230 "./libraries/stb/stb_image.h"
      $190 = HEAP8[$187>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
      $$0$i6$i57$i = $190;$213 = $189;$214 = $188;
     } else {
      $191 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
      $192 = HEAP32[$191>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
      $193 = ($192|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
      if ($193) {
       $$0$i6$i57$i = 0;$213 = $187;$214 = $188;
      } else {
       $194 = HEAP32[$67>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $195 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
       $196 = HEAP32[$195>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $197 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
       $198 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
       $199 = HEAP32[$198>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $200 = (FUNCTION_TABLE_iiii[$194 & 8191]($196,$197,$199)|0); //@line 1213 "./libraries/stb/stb_image.h"
       $201 = ($200|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
       if ($201) {
        HEAP32[$191>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
        $202 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
        HEAP8[$197>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
        $$pre$phi$i3$i54$iZ2D = $202;$$sink$i$i4$i55$i = $202;$453 = 0;
       } else {
        $203 = (((($s)) + 40|0) + ($200)|0); //@line 1223 "./libraries/stb/stb_image.h"
        $$pre$i1$i52$i = HEAP8[$197>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
        $$pre1$i2$i53$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
        $$pre$phi$i3$i54$iZ2D = $$pre1$i2$i53$i;$$sink$i$i4$i55$i = $203;$453 = $$pre$i1$i52$i;
       }
       HEAP32[$22>>2] = $$sink$i$i4$i55$i; //@line 1223 "./libraries/stb/stb_image.h"
       HEAP32[$20>>2] = $$pre$phi$i3$i54$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
       $$0$i6$i57$i = $453;$213 = $$pre$phi$i3$i54$iZ2D;$214 = $$sink$i$i4$i55$i;
      }
     }
     $204 = $$0$i6$i57$i&255; //@line 1309 "./libraries/stb/stb_image.h"
     $205 = $204 << 8; //@line 1309 "./libraries/stb/stb_image.h"
     $206 = $205 | $185; //@line 1309 "./libraries/stb/stb_image.h"
     $207 = ($206|0)==(1); //@line 6070 "./libraries/stb/stb_image.h"
     if (!($207)) {
      $208 = ((($s)) + 176|0); //@line 826 "./libraries/stb/stb_image.h"
      $209 = HEAP32[$208>>2]|0; //@line 826 "./libraries/stb/stb_image.h"
      HEAP32[$20>>2] = $209; //@line 826 "./libraries/stb/stb_image.h"
      $210 = ((($s)) + 180|0); //@line 827 "./libraries/stb/stb_image.h"
      $211 = HEAP32[$210>>2]|0; //@line 827 "./libraries/stb/stb_image.h"
      HEAP32[$22>>2] = $211; //@line 827 "./libraries/stb/stb_image.h"
      $$in = $211;$$in26 = $209;
      break;
     }
     $212 = ($213>>>0)<($214>>>0); //@line 1229 "./libraries/stb/stb_image.h"
     if ($212) {
      $215 = ((($213)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
      HEAP32[$20>>2] = $215; //@line 1230 "./libraries/stb/stb_image.h"
      $216 = HEAP8[$213>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
      $$0$i$i$i = $216;$232 = $215;$233 = $214;
     } else {
      $217 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
      $218 = HEAP32[$217>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
      $219 = ($218|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
      if ($219) {
       $$0$i$i$i = 0;$232 = $213;$233 = $214;
      } else {
       $220 = HEAP32[$67>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $221 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
       $222 = HEAP32[$221>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $223 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
       $224 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
       $225 = HEAP32[$224>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $226 = (FUNCTION_TABLE_iiii[$220 & 8191]($222,$223,$225)|0); //@line 1213 "./libraries/stb/stb_image.h"
       $227 = ($226|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
       if ($227) {
        HEAP32[$217>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
        $228 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
        HEAP8[$223>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
        $$pre$phi$i$i$iZ2D = $228;$$sink$i$i$i$i = $228;$454 = 0;
       } else {
        $229 = (((($s)) + 40|0) + ($226)|0); //@line 1223 "./libraries/stb/stb_image.h"
        $$pre$i$i$i = HEAP8[$223>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
        $$pre1$i$i$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
        $$pre$phi$i$i$iZ2D = $$pre1$i$i$i;$$sink$i$i$i$i = $229;$454 = $$pre$i$i$i;
       }
       HEAP32[$22>>2] = $$sink$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
       HEAP32[$20>>2] = $$pre$phi$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
       $$0$i$i$i = $454;$232 = $$pre$phi$i$i$iZ2D;$233 = $$sink$i$i$i$i;
      }
     }
     $230 = $$0$i$i$i&255; //@line 1308 "./libraries/stb/stb_image.h"
     $231 = ($232>>>0)<($233>>>0); //@line 1229 "./libraries/stb/stb_image.h"
     if ($231) {
      $234 = ((($232)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
      HEAP32[$20>>2] = $234; //@line 1230 "./libraries/stb/stb_image.h"
      $235 = HEAP8[$232>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
      $$0$i6$i$i = $235;
     } else {
      $236 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
      $237 = HEAP32[$236>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
      $238 = ($237|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
      if ($238) {
       $$0$i6$i$i = 0;
      } else {
       $239 = HEAP32[$67>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $240 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
       $241 = HEAP32[$240>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $242 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
       $243 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
       $244 = HEAP32[$243>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $245 = (FUNCTION_TABLE_iiii[$239 & 8191]($241,$242,$244)|0); //@line 1213 "./libraries/stb/stb_image.h"
       $246 = ($245|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
       if ($246) {
        HEAP32[$236>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
        $247 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
        HEAP8[$242>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
        $$pre$phi$i3$i$iZ2D = $247;$$sink$i$i4$i$i = $247;$455 = 0;
       } else {
        $248 = (((($s)) + 40|0) + ($245)|0); //@line 1223 "./libraries/stb/stb_image.h"
        $$pre$i1$i$i = HEAP8[$242>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
        $$pre1$i2$i$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
        $$pre$phi$i3$i$iZ2D = $$pre1$i2$i$i;$$sink$i$i4$i$i = $248;$455 = $$pre$i1$i$i;
       }
       HEAP32[$22>>2] = $$sink$i$i4$i$i; //@line 1223 "./libraries/stb/stb_image.h"
       HEAP32[$20>>2] = $$pre$phi$i3$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
       $$0$i6$i$i = $455;
      }
     }
     $249 = $$0$i6$i$i&255; //@line 1309 "./libraries/stb/stb_image.h"
     $250 = $249 << 8; //@line 1309 "./libraries/stb/stb_image.h"
     $251 = $250 | $230; //@line 1309 "./libraries/stb/stb_image.h"
     $252 = $251 >>> 3; //@line 6074 "./libraries/stb/stb_image.h"
     HEAP32[$comp>>2] = $252; //@line 6074 "./libraries/stb/stb_image.h"
     $$0 = 1;
     STACKTOP = sp;return ($$0|0); //@line 6316 "./libraries/stb/stb_image.h"
    } else {
     label = 25;
    }
   } else {
    label = 25;
   }
  }
 } while(0);
 if ((label|0) == 25) {
  $63 = ((($s)) + 176|0); //@line 826 "./libraries/stb/stb_image.h"
  $64 = HEAP32[$63>>2]|0; //@line 826 "./libraries/stb/stb_image.h"
  HEAP32[$20>>2] = $64; //@line 826 "./libraries/stb/stb_image.h"
  $65 = ((($s)) + 180|0); //@line 827 "./libraries/stb/stb_image.h"
  $66 = HEAP32[$65>>2]|0; //@line 827 "./libraries/stb/stb_image.h"
  HEAP32[$22>>2] = $66; //@line 827 "./libraries/stb/stb_image.h"
  $$in = $66;$$in26 = $64;
 }
 $253 = $$in26; //@line 1229 "./libraries/stb/stb_image.h"
 $254 = $$in; //@line 1229 "./libraries/stb/stb_image.h"
 $255 = ($253>>>0)<($254>>>0); //@line 1229 "./libraries/stb/stb_image.h"
 if ($255) {
  $256 = ((($253)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
  HEAP32[$20>>2] = $256; //@line 1230 "./libraries/stb/stb_image.h"
  $272 = $256;$273 = $254;
 } else {
  $257 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
  $258 = HEAP32[$257>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
  $259 = ($258|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
  if ($259) {
   $272 = $253;$273 = $254;
  } else {
   $260 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
   $261 = HEAP32[$260>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $262 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
   $263 = HEAP32[$262>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $264 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
   $265 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
   $266 = HEAP32[$265>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $267 = (FUNCTION_TABLE_iiii[$261 & 8191]($263,$264,$266)|0); //@line 1213 "./libraries/stb/stb_image.h"
   $268 = ($267|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
   if ($268) {
    HEAP32[$257>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
    $269 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
    HEAP8[$264>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
    $$pre$phi$i$i3Z2D = $269;$$sink$i$i$i4 = $269;
   } else {
    $270 = (((($s)) + 40|0) + ($267)|0); //@line 1223 "./libraries/stb/stb_image.h"
    $$pre1$i$i2 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
    $$pre$phi$i$i3Z2D = $$pre1$i$i2;$$sink$i$i$i4 = $270;
   }
   HEAP32[$22>>2] = $$sink$i$i$i4; //@line 1223 "./libraries/stb/stb_image.h"
   HEAP32[$20>>2] = $$pre$phi$i$i3Z2D; //@line 1233 "./libraries/stb/stb_image.h"
   $272 = $$pre$phi$i$i3Z2D;$273 = $$sink$i$i$i4;
  }
 }
 $271 = ($272>>>0)<($273>>>0); //@line 1229 "./libraries/stb/stb_image.h"
 if ($271) {
  $274 = ((($272)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
  HEAP32[$20>>2] = $274; //@line 1230 "./libraries/stb/stb_image.h"
  $275 = HEAP8[$272>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
  $$0$i23$i = $275;$456 = $273;$457 = $274;
  label = 107;
 } else {
  $276 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
  $277 = HEAP32[$276>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
  $278 = ($277|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
  if ($278) {
   $296 = $272;$297 = $273;
   label = 109;
  } else {
   $279 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
   $280 = HEAP32[$279>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $281 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
   $282 = HEAP32[$281>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $283 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
   $284 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
   $285 = HEAP32[$284>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $286 = (FUNCTION_TABLE_iiii[$280 & 8191]($282,$283,$285)|0); //@line 1213 "./libraries/stb/stb_image.h"
   $287 = ($286|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
   if ($287) {
    HEAP32[$276>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
    $288 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
    HEAP8[$283>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
    $$pre$phi$i20$iZ2D = $288;$$sink$i$i21$i = $288;$458 = 0;
   } else {
    $289 = (((($s)) + 40|0) + ($286)|0); //@line 1223 "./libraries/stb/stb_image.h"
    $$pre$i18$i = HEAP8[$283>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
    $$pre1$i19$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
    $$pre$phi$i20$iZ2D = $$pre1$i19$i;$$sink$i$i21$i = $289;$458 = $$pre$i18$i;
   }
   HEAP32[$22>>2] = $$sink$i$i21$i; //@line 1223 "./libraries/stb/stb_image.h"
   HEAP32[$20>>2] = $$pre$phi$i20$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
   $$0$i23$i = $458;$456 = $$sink$i$i21$i;$457 = $$pre$phi$i20$iZ2D;
   label = 107;
  }
 }
 if ((label|0) == 107) {
  $290 = ($$0$i23$i&255)>(1); //@line 4818 "./libraries/stb/stb_image.h"
  if ($290) {
   $291 = ((($s)) + 176|0); //@line 826 "./libraries/stb/stb_image.h"
   $292 = HEAP32[$291>>2]|0; //@line 826 "./libraries/stb/stb_image.h"
   HEAP32[$20>>2] = $292; //@line 826 "./libraries/stb/stb_image.h"
   $293 = ((($s)) + 180|0); //@line 827 "./libraries/stb/stb_image.h"
   $294 = HEAP32[$293>>2]|0; //@line 827 "./libraries/stb/stb_image.h"
   HEAP32[$22>>2] = $294; //@line 827 "./libraries/stb/stb_image.h"
  } else {
   $296 = $457;$297 = $456;
   label = 109;
  }
 }
 L145: do {
  if ((label|0) == 109) {
   $295 = ($296>>>0)<($297>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($295) {
    $298 = ((($296)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$20>>2] = $298; //@line 1230 "./libraries/stb/stb_image.h"
    $299 = HEAP8[$296>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
    $$0$i37$i = $299;$$pre$i39$i = $298;$319 = $297;
   } else {
    $300 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $301 = HEAP32[$300>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $302 = ($301|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if ($302) {
     break;
    }
    $303 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
    $304 = HEAP32[$303>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $305 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
    $306 = HEAP32[$305>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $307 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
    $308 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
    $309 = HEAP32[$308>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $310 = (FUNCTION_TABLE_iiii[$304 & 8191]($306,$307,$309)|0); //@line 1213 "./libraries/stb/stb_image.h"
    $311 = ($310|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
    if ($311) {
     HEAP32[$300>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
     $312 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
     HEAP8[$307>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
     $$pre$phi$i34$iZ2D = $312;$$sink$i$i35$i = $312;$459 = 0;
    } else {
     $313 = (((($s)) + 40|0) + ($310)|0); //@line 1223 "./libraries/stb/stb_image.h"
     $$pre$i32$i = HEAP8[$307>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
     $$pre1$i33$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
     $$pre$phi$i34$iZ2D = $$pre1$i33$i;$$sink$i$i35$i = $313;$459 = $$pre$i32$i;
    }
    HEAP32[$22>>2] = $$sink$i$i35$i; //@line 1223 "./libraries/stb/stb_image.h"
    HEAP32[$20>>2] = $$pre$phi$i34$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
    $$0$i37$i = $459;$$pre$i39$i = $$pre$phi$i34$iZ2D;$319 = $$sink$i$i35$i;
   }
   $314 = $$pre$i39$i; //@line 1257 "./libraries/stb/stb_image.h"
   switch ($$0$i37$i<<24>>24) {
   case 1: case 2: case 3: case 9: case 10: case 11:  {
    break;
   }
   default: {
    break L145;
   }
   }
   $315 = ((($s)) + 16|0); //@line 1256 "./libraries/stb/stb_image.h"
   $316 = HEAP32[$315>>2]|0; //@line 1256 "./libraries/stb/stb_image.h"
   $317 = ($316|0)==(0|0); //@line 1256 "./libraries/stb/stb_image.h"
   if ($317) {
    label = 120;
   } else {
    $318 = $319; //@line 1257 "./libraries/stb/stb_image.h"
    $320 = (($318) - ($314))|0; //@line 1257 "./libraries/stb/stb_image.h"
    $321 = ($320|0)<(9); //@line 1258 "./libraries/stb/stb_image.h"
    if ($321) {
     HEAP32[$20>>2] = $319; //@line 1259 "./libraries/stb/stb_image.h"
     $322 = ((($s)) + 20|0); //@line 1260 "./libraries/stb/stb_image.h"
     $323 = HEAP32[$322>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
     $324 = ((($s)) + 28|0); //@line 1260 "./libraries/stb/stb_image.h"
     $325 = HEAP32[$324>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
     $326 = (9 - ($320))|0; //@line 1260 "./libraries/stb/stb_image.h"
     FUNCTION_TABLE_vii[$323 & 8191]($325,$326); //@line 1260 "./libraries/stb/stb_image.h"
     $$pre$i7 = HEAP32[$20>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
     $$pre56$i = HEAP32[$22>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
     $329 = $$pre$i7;$330 = $$pre56$i;
    } else {
     label = 120;
    }
   }
   if ((label|0) == 120) {
    $327 = ((($$pre$i39$i)) + 9|0); //@line 1264 "./libraries/stb/stb_image.h"
    HEAP32[$20>>2] = $327; //@line 1264 "./libraries/stb/stb_image.h"
    $329 = $327;$330 = $319;
   }
   $328 = ($329>>>0)<($330>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($328) {
    $331 = ((($329)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$20>>2] = $331; //@line 1230 "./libraries/stb/stb_image.h"
    $332 = HEAP8[$329>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
    $$0$i$i$i16 = $332;$348 = $331;$349 = $330;
   } else {
    $333 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $334 = HEAP32[$333>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $335 = ($334|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if ($335) {
     $$0$i$i$i16 = 0;$348 = $329;$349 = $330;
    } else {
     $336 = HEAP32[$315>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $337 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $338 = HEAP32[$337>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $339 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $340 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $341 = HEAP32[$340>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $342 = (FUNCTION_TABLE_iiii[$336 & 8191]($338,$339,$341)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $343 = ($342|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($343) {
      HEAP32[$333>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $344 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$339>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i$i$i13Z2D = $344;$$sink$i$i$i$i14 = $344;$460 = 0;
     } else {
      $345 = (((($s)) + 40|0) + ($342)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre$i$i$i11 = HEAP8[$339>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
      $$pre1$i$i$i12 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i$i$i13Z2D = $$pre1$i$i$i12;$$sink$i$i$i$i14 = $345;$460 = $$pre$i$i$i11;
     }
     HEAP32[$22>>2] = $$sink$i$i$i$i14; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$20>>2] = $$pre$phi$i$i$i13Z2D; //@line 1233 "./libraries/stb/stb_image.h"
     $$0$i$i$i16 = $460;$348 = $$pre$phi$i$i$i13Z2D;$349 = $$sink$i$i$i$i14;
    }
   }
   $346 = $$0$i$i$i16&255; //@line 1308 "./libraries/stb/stb_image.h"
   $347 = ($348>>>0)<($349>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($347) {
    $350 = ((($348)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$20>>2] = $350; //@line 1230 "./libraries/stb/stb_image.h"
    $351 = HEAP8[$348>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
    $$0$i6$i$i23 = $351;$374 = $350;$375 = $349;
   } else {
    $352 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $353 = HEAP32[$352>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $354 = ($353|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if ($354) {
     $$0$i6$i$i23 = 0;$374 = $348;$375 = $349;
    } else {
     $355 = HEAP32[$315>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $356 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $357 = HEAP32[$356>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $358 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $359 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $360 = HEAP32[$359>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $361 = (FUNCTION_TABLE_iiii[$355 & 8191]($357,$358,$360)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $362 = ($361|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($362) {
      HEAP32[$352>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $363 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$358>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i3$i$i20Z2D = $363;$$sink$i$i4$i$i21 = $363;$461 = 0;
     } else {
      $364 = (((($s)) + 40|0) + ($361)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre$i1$i$i18 = HEAP8[$358>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
      $$pre1$i2$i$i19 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i3$i$i20Z2D = $$pre1$i2$i$i19;$$sink$i$i4$i$i21 = $364;$461 = $$pre$i1$i$i18;
     }
     HEAP32[$22>>2] = $$sink$i$i4$i$i21; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$20>>2] = $$pre$phi$i3$i$i20Z2D; //@line 1233 "./libraries/stb/stb_image.h"
     $$0$i6$i$i23 = $461;$374 = $$pre$phi$i3$i$i20Z2D;$375 = $$sink$i$i4$i$i21;
    }
   }
   $365 = $$0$i6$i$i23&255; //@line 1309 "./libraries/stb/stb_image.h"
   $366 = $365 << 8; //@line 1309 "./libraries/stb/stb_image.h"
   $367 = $366 | $346; //@line 1309 "./libraries/stb/stb_image.h"
   $368 = ($367|0)==(0); //@line 4827 "./libraries/stb/stb_image.h"
   if ($368) {
    $369 = ((($s)) + 176|0); //@line 826 "./libraries/stb/stb_image.h"
    $370 = HEAP32[$369>>2]|0; //@line 826 "./libraries/stb/stb_image.h"
    HEAP32[$20>>2] = $370; //@line 826 "./libraries/stb/stb_image.h"
    $371 = ((($s)) + 180|0); //@line 827 "./libraries/stb/stb_image.h"
    $372 = HEAP32[$371>>2]|0; //@line 827 "./libraries/stb/stb_image.h"
    HEAP32[$22>>2] = $372; //@line 827 "./libraries/stb/stb_image.h"
    break;
   }
   $373 = ($374>>>0)<($375>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($373) {
    $376 = ((($374)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$20>>2] = $376; //@line 1230 "./libraries/stb/stb_image.h"
    $377 = HEAP8[$374>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
    $$0$i$i46$i = $377;$393 = $376;$394 = $375;
   } else {
    $378 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $379 = HEAP32[$378>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $380 = ($379|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if ($380) {
     $$0$i$i46$i = 0;$393 = $374;$394 = $375;
    } else {
     $381 = HEAP32[$315>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $382 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $383 = HEAP32[$382>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $384 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $385 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $386 = HEAP32[$385>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $387 = (FUNCTION_TABLE_iiii[$381 & 8191]($383,$384,$386)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $388 = ($387|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($388) {
      HEAP32[$378>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $389 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$384>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i$i43$iZ2D = $389;$$sink$i$i$i44$i = $389;$462 = 0;
     } else {
      $390 = (((($s)) + 40|0) + ($387)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre$i$i41$i = HEAP8[$384>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
      $$pre1$i$i42$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i$i43$iZ2D = $$pre1$i$i42$i;$$sink$i$i$i44$i = $390;$462 = $$pre$i$i41$i;
     }
     HEAP32[$22>>2] = $$sink$i$i$i44$i; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$20>>2] = $$pre$phi$i$i43$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
     $$0$i$i46$i = $462;$393 = $$pre$phi$i$i43$iZ2D;$394 = $$sink$i$i$i44$i;
    }
   }
   $391 = $$0$i$i46$i&255; //@line 1308 "./libraries/stb/stb_image.h"
   $392 = ($393>>>0)<($394>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($392) {
    $395 = ((($393)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$20>>2] = $395; //@line 1230 "./libraries/stb/stb_image.h"
    $396 = HEAP8[$393>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
    $$0$i6$i53$i = $396;$419 = $395;$420 = $394;
   } else {
    $397 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $398 = HEAP32[$397>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $399 = ($398|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if ($399) {
     $$0$i6$i53$i = 0;$419 = $393;$420 = $394;
    } else {
     $400 = HEAP32[$315>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $401 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $402 = HEAP32[$401>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $403 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $404 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $405 = HEAP32[$404>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $406 = (FUNCTION_TABLE_iiii[$400 & 8191]($402,$403,$405)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $407 = ($406|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($407) {
      HEAP32[$397>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $408 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$403>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i3$i50$iZ2D = $408;$$sink$i$i4$i51$i = $408;$463 = 0;
     } else {
      $409 = (((($s)) + 40|0) + ($406)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre$i1$i48$i = HEAP8[$403>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
      $$pre1$i2$i49$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i3$i50$iZ2D = $$pre1$i2$i49$i;$$sink$i$i4$i51$i = $409;$463 = $$pre$i1$i48$i;
     }
     HEAP32[$22>>2] = $$sink$i$i4$i51$i; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$20>>2] = $$pre$phi$i3$i50$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
     $$0$i6$i53$i = $463;$419 = $$pre$phi$i3$i50$iZ2D;$420 = $$sink$i$i4$i51$i;
    }
   }
   $410 = $$0$i6$i53$i&255; //@line 1309 "./libraries/stb/stb_image.h"
   $411 = $410 << 8; //@line 1309 "./libraries/stb/stb_image.h"
   $412 = $411 | $391; //@line 1309 "./libraries/stb/stb_image.h"
   $413 = ($412|0)==(0); //@line 4832 "./libraries/stb/stb_image.h"
   if ($413) {
    $414 = ((($s)) + 176|0); //@line 826 "./libraries/stb/stb_image.h"
    $415 = HEAP32[$414>>2]|0; //@line 826 "./libraries/stb/stb_image.h"
    HEAP32[$20>>2] = $415; //@line 826 "./libraries/stb/stb_image.h"
    $416 = ((($s)) + 180|0); //@line 827 "./libraries/stb/stb_image.h"
    $417 = HEAP32[$416>>2]|0; //@line 827 "./libraries/stb/stb_image.h"
    HEAP32[$22>>2] = $417; //@line 827 "./libraries/stb/stb_image.h"
    break;
   }
   $418 = ($419>>>0)<($420>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($418) {
    $421 = ((($419)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$20>>2] = $421; //@line 1230 "./libraries/stb/stb_image.h"
    $422 = HEAP8[$419>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
    $$0$i30$i = $422;
    label = 160;
   } else {
    $423 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $424 = HEAP32[$423>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $425 = ($424|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if (!($425)) {
     $426 = HEAP32[$315>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $427 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $428 = HEAP32[$427>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $429 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $430 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $431 = HEAP32[$430>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $432 = (FUNCTION_TABLE_iiii[$426 & 8191]($428,$429,$431)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $433 = ($432|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($433) {
      HEAP32[$423>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $434 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$429>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i27$iZ2D = $434;$$sink$i$i28$i = $434;$464 = 0;
     } else {
      $435 = (((($s)) + 40|0) + ($432)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre$i25$i = HEAP8[$429>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
      $$pre1$i26$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i27$iZ2D = $$pre1$i26$i;$$sink$i$i28$i = $435;$464 = $$pre$i25$i;
     }
     HEAP32[$22>>2] = $$sink$i$i28$i; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$20>>2] = $$pre$phi$i27$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
     $$0$i30$i = $464;
     label = 160;
    }
   }
   L213: do {
    if ((label|0) == 160) {
     switch ($$0$i30$i<<24>>24) {
     case 8: case 16: case 24: case 32:  {
      break;
     }
     default: {
      break L213;
     }
     }
     $440 = ($x|0)==(0|0); //@line 4843 "./libraries/stb/stb_image.h"
     if (!($440)) {
      HEAP32[$x>>2] = $367; //@line 4843 "./libraries/stb/stb_image.h"
     }
     $441 = ($y|0)==(0|0); //@line 4844 "./libraries/stb/stb_image.h"
     if (!($441)) {
      HEAP32[$y>>2] = $412; //@line 4844 "./libraries/stb/stb_image.h"
     }
     $442 = ($comp|0)==(0|0); //@line 4845 "./libraries/stb/stb_image.h"
     if ($442) {
      $$0 = 1;
      STACKTOP = sp;return ($$0|0); //@line 6316 "./libraries/stb/stb_image.h"
     }
     $div$i = ($$0$i30$i&255) >>> 3; //@line 4845 "./libraries/stb/stb_image.h"
     $443 = $div$i&255; //@line 4845 "./libraries/stb/stb_image.h"
     HEAP32[$comp>>2] = $443; //@line 4845 "./libraries/stb/stb_image.h"
     $$0 = 1;
     STACKTOP = sp;return ($$0|0); //@line 6316 "./libraries/stb/stb_image.h"
    }
   } while(0);
   $436 = ((($s)) + 176|0); //@line 826 "./libraries/stb/stb_image.h"
   $437 = HEAP32[$436>>2]|0; //@line 826 "./libraries/stb/stb_image.h"
   HEAP32[$20>>2] = $437; //@line 826 "./libraries/stb/stb_image.h"
   $438 = ((($s)) + 180|0); //@line 827 "./libraries/stb/stb_image.h"
   $439 = HEAP32[$438>>2]|0; //@line 827 "./libraries/stb/stb_image.h"
   HEAP32[$22>>2] = $439; //@line 827 "./libraries/stb/stb_image.h"
  }
 } while(0);
 HEAP32[158754] = 441011; //@line 894 "./libraries/stb/stb_image.h"
 $$0 = 0;
 STACKTOP = sp;return ($$0|0); //@line 6316 "./libraries/stb/stb_image.h"
}
function __ZL24stbi__decode_jpeg_headerP10stbi__jpegi($z,$scan) {
 $z = $z|0;
 $scan = $scan|0;
 var $$ = 0, $$0 = 0, $$0$i$i = 0, $$0$i$i$i = 0, $$0$i$i23$i = 0, $$0$i$i26 = 0, $$0$i$i44 = 0, $$0$i$i61 = 0, $$0$i$i9 = 0, $$0$i$i9$i = 0, $$0$i12 = 0, $$0$i19 = 0, $$0$i3 = 0, $$0$i36 = 0, $$0$i37$i = 0, $$0$i44$i = 0, $$0$i51$i = 0, $$0$i54 = 0, $$0$i58$i = 0, $$0$i6$i$i = 0;
 var $$0$i6$i16$i = 0, $$0$i6$i30$i = 0, $$75 = 0, $$h_max$0$i = 0, $$lcssa103 = 0, $$lcssa72 = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i$i = 0, $$pre$i$i18$i = 0, $$pre$i$i21 = 0, $$pre$i$i39 = 0, $$pre$i$i4 = 0, $$pre$i$i4$i = 0, $$pre$i$i56 = 0, $$pre$i1$i$i = 0, $$pre$i1$i11$i = 0, $$pre$i1$i25$i = 0, $$pre$i14 = 0, $$pre$i31 = 0;
 var $$pre$i32$i = 0, $$pre$i39$i = 0, $$pre$i46$i = 0, $$pre$i49 = 0, $$pre$i53$i = 0, $$pre$i63 = 0, $$pre$phi$i$i$iZ2D = 0, $$pre$phi$i$i20$iZ2D = 0, $$pre$phi$i$i23Z2D = 0, $$pre$phi$i$i41Z2D = 0, $$pre$phi$i$i58Z2D = 0, $$pre$phi$i$i6$iZ2D = 0, $$pre$phi$i$i6Z2D = 0, $$pre$phi$i$iZ2D = 0, $$pre$phi$i16Z2D = 0, $$pre$phi$i3$i$iZ2D = 0, $$pre$phi$i3$i13$iZ2D = 0, $$pre$phi$i3$i27$iZ2D = 0, $$pre$phi$i33Z2D = 0, $$pre$phi$i34$iZ2D = 0;
 var $$pre$phi$i51Z2D = 0, $$pre$phi$iZ2D = 0, $$pre1$i = 0, $$pre1$i$i = 0, $$pre1$i$i$i = 0, $$pre1$i$i19$i = 0, $$pre1$i$i22 = 0, $$pre1$i$i40 = 0, $$pre1$i$i5 = 0, $$pre1$i$i5$i = 0, $$pre1$i$i57 = 0, $$pre1$i15 = 0, $$pre1$i2$i$i = 0, $$pre1$i2$i12$i = 0, $$pre1$i2$i26$i = 0, $$pre1$i32 = 0, $$pre1$i33$i = 0, $$pre1$i50 = 0, $$pre90$i = 0, $$pre91$i = 0;
 var $$sink$be$ph$sink = 0, $$sink$i = 0, $$sink$i$i = 0, $$sink$i$i$i = 0, $$sink$i$i$i$i = 0, $$sink$i$i$i21$i = 0, $$sink$i$i$i24 = 0, $$sink$i$i$i42 = 0, $$sink$i$i$i59 = 0, $$sink$i$i$i7 = 0, $$sink$i$i$i7$i = 0, $$sink$i$i17 = 0, $$sink$i$i34 = 0, $$sink$i$i35$i = 0, $$sink$i$i4$i$i = 0, $$sink$i$i4$i14$i = 0, $$sink$i$i4$i28$i = 0, $$sink$i$i42$i = 0, $$sink$i$i49$i = 0, $$sink$i$i52 = 0;
 var $$sink$i$i56$i = 0, $$sink88$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0;
 var $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0;
 var $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0;
 var $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0;
 var $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0;
 var $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0;
 var $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0;
 var $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0;
 var $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0;
 var $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0;
 var $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0;
 var $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0;
 var $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0;
 var $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0;
 var $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0;
 var $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0;
 var $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0;
 var $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0;
 var $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i = 0, $h_max$0$lcssa$i = 0, $h_max$073$i = 0, $i$082$i = 0, $i$178$i = 0, $i$272$i = 0, $i$369$i = 0, $i$369$i$lcssa = 0, $i$468$i = 0, $i$468$in$i = 0, $m$076 = 0, $or$cond$i = 0, $or$cond2$i = 0, $v_max$0$lcssa$i = 0, $v_max$074$i = 0, $v_max$1$i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($z)) + 18116|0); //@line 2796 "./libraries/stb/stb_image.h"
 HEAP8[$0>>0] = -1; //@line 2796 "./libraries/stb/stb_image.h"
 $1 = HEAP32[$z>>2]|0; //@line 2423 "./libraries/stb/stb_image.h"
 $2 = ((($1)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
 $3 = HEAP32[$2>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
 $4 = ((($1)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
 $5 = HEAP32[$4>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
 $6 = ($3>>>0)<($5>>>0); //@line 1229 "./libraries/stb/stb_image.h"
 if ($6) {
  $7 = ((($3)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
  HEAP32[$2>>2] = $7; //@line 1230 "./libraries/stb/stb_image.h"
  $8 = HEAP8[$3>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
  $$0$i$i = $8;
  label = 8;
 } else {
  $9 = ((($1)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
  $10 = HEAP32[$9>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
  $11 = ($10|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
  if (!($11)) {
   $12 = ((($1)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
   $13 = HEAP32[$12>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $14 = ((($1)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
   $15 = HEAP32[$14>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $16 = ((($1)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
   $17 = ((($1)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
   $18 = HEAP32[$17>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $19 = (FUNCTION_TABLE_iiii[$13 & 8191]($15,$16,$18)|0); //@line 1213 "./libraries/stb/stb_image.h"
   $20 = ($19|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
   if ($20) {
    HEAP32[$9>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
    $21 = ((($1)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
    HEAP8[$16>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
    $$pre$phi$i$iZ2D = $21;$$sink$i$i$i = $21;$557 = 0;
   } else {
    $22 = (((($1)) + 40|0) + ($19)|0); //@line 1223 "./libraries/stb/stb_image.h"
    $$pre$i$i = HEAP8[$16>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
    $$pre1$i$i = ((($1)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
    $$pre$phi$i$iZ2D = $$pre1$i$i;$$sink$i$i$i = $22;$557 = $$pre$i$i;
   }
   HEAP32[$4>>2] = $$sink$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
   HEAP32[$2>>2] = $$pre$phi$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
   $$0$i$i = $557;
   label = 8;
  }
 }
 L9: do {
  if ((label|0) == 8) {
   $23 = ($$0$i$i<<24>>24)==(-1); //@line 2424 "./libraries/stb/stb_image.h"
   if ($23) {
    L11: while(1) {
     $24 = HEAP32[$z>>2]|0; //@line 2426 "./libraries/stb/stb_image.h"
     $25 = ((($24)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
     $26 = HEAP32[$25>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
     $27 = ((($24)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
     $28 = HEAP32[$27>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
     $29 = ($26>>>0)<($28>>>0); //@line 1229 "./libraries/stb/stb_image.h"
     if ($29) {
      $30 = ((($26)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
      HEAP32[$25>>2] = $30; //@line 1230 "./libraries/stb/stb_image.h"
      $31 = HEAP8[$26>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
      $$0$i3 = $31;
     } else {
      $32 = ((($24)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
      $33 = HEAP32[$32>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
      $34 = ($33|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
      if ($34) {
       break L9;
      }
      $35 = ((($24)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
      $36 = HEAP32[$35>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $37 = ((($24)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
      $38 = HEAP32[$37>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $39 = ((($24)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
      $40 = ((($24)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
      $41 = HEAP32[$40>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $42 = (FUNCTION_TABLE_iiii[$36 & 8191]($38,$39,$41)|0); //@line 1213 "./libraries/stb/stb_image.h"
      $43 = ($42|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
      if ($43) {
       HEAP32[$32>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
       $44 = ((($24)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
       HEAP8[$39>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
       $$pre$phi$iZ2D = $44;$$sink$i$i = $44;$558 = 0;
      } else {
       $45 = (((($24)) + 40|0) + ($42)|0); //@line 1223 "./libraries/stb/stb_image.h"
       $$pre$i = HEAP8[$39>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
       $$pre1$i = ((($24)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
       $$pre$phi$iZ2D = $$pre1$i;$$sink$i$i = $45;$558 = $$pre$i;
      }
      HEAP32[$27>>2] = $$sink$i$i; //@line 1223 "./libraries/stb/stb_image.h"
      HEAP32[$25>>2] = $$pre$phi$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
      $$0$i3 = $558;
     }
     switch ($$0$i3<<24>>24) {
     case -1:  {
      break;
     }
     case -40:  {
      break L11;
      break;
     }
     default: {
      break L9;
     }
     }
    }
    $46 = ($scan|0)==(1); //@line 2799 "./libraries/stb/stb_image.h"
    if ($46) {
     $$0 = 1;
     return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
    }
    $47 = HEAP8[$0>>0]|0; //@line 2422 "./libraries/stb/stb_image.h"
    $48 = ($47<<24>>24)==(-1); //@line 2422 "./libraries/stb/stb_image.h"
    L26: do {
     if ($48) {
      $49 = HEAP32[$z>>2]|0; //@line 2423 "./libraries/stb/stb_image.h"
      $50 = ((($49)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
      $51 = HEAP32[$50>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
      $52 = ((($49)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
      $53 = HEAP32[$52>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
      $54 = ($51>>>0)<($53>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      if ($54) {
       $55 = ((($51)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
       HEAP32[$50>>2] = $55; //@line 1230 "./libraries/stb/stb_image.h"
       $56 = HEAP8[$51>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
       $$0$i$i9 = $56;
      } else {
       $57 = ((($49)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
       $58 = HEAP32[$57>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
       $59 = ($58|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
       if ($59) {
        $$0$i12 = -1;
        break;
       }
       $60 = ((($49)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
       $61 = HEAP32[$60>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $62 = ((($49)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
       $63 = HEAP32[$62>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $64 = ((($49)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
       $65 = ((($49)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
       $66 = HEAP32[$65>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $67 = (FUNCTION_TABLE_iiii[$61 & 8191]($63,$64,$66)|0); //@line 1213 "./libraries/stb/stb_image.h"
       $68 = ($67|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
       if ($68) {
        HEAP32[$57>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
        $69 = ((($49)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
        HEAP8[$64>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
        $$pre$phi$i$i6Z2D = $69;$$sink$i$i$i7 = $69;$559 = 0;
       } else {
        $70 = (((($49)) + 40|0) + ($67)|0); //@line 1223 "./libraries/stb/stb_image.h"
        $$pre$i$i4 = HEAP8[$64>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
        $$pre1$i$i5 = ((($49)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
        $$pre$phi$i$i6Z2D = $$pre1$i$i5;$$sink$i$i$i7 = $70;$559 = $$pre$i$i4;
       }
       HEAP32[$52>>2] = $$sink$i$i$i7; //@line 1223 "./libraries/stb/stb_image.h"
       HEAP32[$50>>2] = $$pre$phi$i$i6Z2D; //@line 1233 "./libraries/stb/stb_image.h"
       $$0$i$i9 = $559;
      }
      $71 = ($$0$i$i9<<24>>24)==(-1); //@line 2424 "./libraries/stb/stb_image.h"
      if ($71) {
       while(1) {
        $72 = HEAP32[$z>>2]|0; //@line 2426 "./libraries/stb/stb_image.h"
        $73 = ((($72)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
        $74 = HEAP32[$73>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
        $75 = ((($72)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
        $76 = HEAP32[$75>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
        $77 = ($74>>>0)<($76>>>0); //@line 1229 "./libraries/stb/stb_image.h"
        if ($77) {
         $78 = ((($74)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
         HEAP32[$73>>2] = $78; //@line 1230 "./libraries/stb/stb_image.h"
         $79 = HEAP8[$74>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
         $$0$i19 = $79;
        } else {
         $80 = ((($72)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
         $81 = HEAP32[$80>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
         $82 = ($81|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
         if ($82) {
          $$0$i12 = 0;
          break L26;
         }
         $83 = ((($72)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
         $84 = HEAP32[$83>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $85 = ((($72)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
         $86 = HEAP32[$85>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $87 = ((($72)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
         $88 = ((($72)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
         $89 = HEAP32[$88>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $90 = (FUNCTION_TABLE_iiii[$84 & 8191]($86,$87,$89)|0); //@line 1213 "./libraries/stb/stb_image.h"
         $91 = ($90|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
         if ($91) {
          HEAP32[$80>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
          $92 = ((($72)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
          HEAP8[$87>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
          $$pre$phi$i16Z2D = $92;$$sink$i$i17 = $92;$560 = 0;
         } else {
          $93 = (((($72)) + 40|0) + ($90)|0); //@line 1223 "./libraries/stb/stb_image.h"
          $$pre$i14 = HEAP8[$87>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
          $$pre1$i15 = ((($72)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
          $$pre$phi$i16Z2D = $$pre1$i15;$$sink$i$i17 = $93;$560 = $$pre$i14;
         }
         HEAP32[$75>>2] = $$sink$i$i17; //@line 1223 "./libraries/stb/stb_image.h"
         HEAP32[$73>>2] = $$pre$phi$i16Z2D; //@line 1233 "./libraries/stb/stb_image.h"
         $$0$i19 = $560;
        }
        $94 = ($$0$i19<<24>>24)==(-1); //@line 2425 "./libraries/stb/stb_image.h"
        if (!($94)) {
         $$0$i12 = $$0$i19;
         break;
        }
       }
      } else {
       $$0$i12 = -1;
      }
     } else {
      HEAP8[$0>>0] = -1; //@line 2422 "./libraries/stb/stb_image.h"
      $$0$i12 = $47;
     }
    } while(0);
    $95 = $$0$i12&255; //@line 2800 "./libraries/stb/stb_image.h"
    $96 = $95 | 1; //@line 2801 "./libraries/stb/stb_image.h"
    $97 = ($96|0)==(193); //@line 2801 "./libraries/stb/stb_image.h"
    $98 = ($$0$i12<<24>>24)==(-62); //@line 2801 "./libraries/stb/stb_image.h"
    $$75 = $98 | $97; //@line 2801 "./libraries/stb/stb_image.h"
    L50: do {
     if ($$75) {
      $$lcssa72 = $98;
     } else {
      $m$076 = $95;
      L51: while(1) {
       $103 = (__ZL20stbi__process_markerP10stbi__jpegi($z,$m$076)|0); //@line 2802 "./libraries/stb/stb_image.h"
       $104 = ($103|0)==(0); //@line 2802 "./libraries/stb/stb_image.h"
       if ($104) {
        $$0 = 0;
        label = 200;
        break;
       }
       $105 = HEAP8[$0>>0]|0; //@line 2422 "./libraries/stb/stb_image.h"
       $106 = ($105<<24>>24)==(-1); //@line 2422 "./libraries/stb/stb_image.h"
       L54: do {
        if ($106) {
         $107 = HEAP32[$z>>2]|0; //@line 2423 "./libraries/stb/stb_image.h"
         $108 = ((($107)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
         $109 = HEAP32[$108>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $110 = ((($107)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
         $111 = HEAP32[$110>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $112 = ($109>>>0)<($111>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         if ($112) {
          $113 = ((($109)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
          HEAP32[$108>>2] = $113; //@line 1230 "./libraries/stb/stb_image.h"
          $114 = HEAP8[$109>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
          $$0$i$i26 = $114;
          label = 49;
         } else {
          $115 = ((($107)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
          $116 = HEAP32[$115>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
          $117 = ($116|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
          if (!($117)) {
           $118 = ((($107)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $119 = HEAP32[$118>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $120 = ((($107)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $121 = HEAP32[$120>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $122 = ((($107)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $123 = ((($107)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $124 = HEAP32[$123>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $125 = (FUNCTION_TABLE_iiii[$119 & 8191]($121,$122,$124)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $126 = ($125|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($126) {
            HEAP32[$115>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $127 = ((($107)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$122>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i$i23Z2D = $127;$$sink$i$i$i24 = $127;$561 = 0;
           } else {
            $128 = (((($107)) + 40|0) + ($125)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre$i$i21 = HEAP8[$122>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
            $$pre1$i$i22 = ((($107)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i$i23Z2D = $$pre1$i$i22;$$sink$i$i$i24 = $128;$561 = $$pre$i$i21;
           }
           HEAP32[$110>>2] = $$sink$i$i$i24; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$108>>2] = $$pre$phi$i$i23Z2D; //@line 1233 "./libraries/stb/stb_image.h"
           $$0$i$i26 = $561;
           label = 49;
          }
         }
         if ((label|0) == 49) {
          label = 0;
          $129 = ($$0$i$i26<<24>>24)==(-1); //@line 2424 "./libraries/stb/stb_image.h"
          if ($129) {
           while(1) {
            $130 = HEAP32[$z>>2]|0; //@line 2426 "./libraries/stb/stb_image.h"
            $131 = ((($130)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
            $132 = HEAP32[$131>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
            $133 = ((($130)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
            $134 = HEAP32[$133>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
            $135 = ($132>>>0)<($134>>>0); //@line 1229 "./libraries/stb/stb_image.h"
            if ($135) {
             $136 = ((($132)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
             HEAP32[$131>>2] = $136; //@line 1230 "./libraries/stb/stb_image.h"
             $137 = HEAP8[$132>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
             $$0$i36 = $137;
            } else {
             $138 = ((($130)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
             $139 = HEAP32[$138>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
             $140 = ($139|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
             if ($140) {
              $$sink$be$ph$sink = 0;
              break L54;
             }
             $141 = ((($130)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
             $142 = HEAP32[$141>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
             $143 = ((($130)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
             $144 = HEAP32[$143>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
             $145 = ((($130)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
             $146 = ((($130)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
             $147 = HEAP32[$146>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
             $148 = (FUNCTION_TABLE_iiii[$142 & 8191]($144,$145,$147)|0); //@line 1213 "./libraries/stb/stb_image.h"
             $149 = ($148|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
             if ($149) {
              HEAP32[$138>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
              $150 = ((($130)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
              HEAP8[$145>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
              $$pre$phi$i33Z2D = $150;$$sink$i$i34 = $150;$562 = 0;
             } else {
              $151 = (((($130)) + 40|0) + ($148)|0); //@line 1223 "./libraries/stb/stb_image.h"
              $$pre$i31 = HEAP8[$145>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
              $$pre1$i32 = ((($130)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
              $$pre$phi$i33Z2D = $$pre1$i32;$$sink$i$i34 = $151;$562 = $$pre$i31;
             }
             HEAP32[$133>>2] = $$sink$i$i34; //@line 1223 "./libraries/stb/stb_image.h"
             HEAP32[$131>>2] = $$pre$phi$i33Z2D; //@line 1233 "./libraries/stb/stb_image.h"
             $$0$i36 = $562;
            }
            $152 = ($$0$i36<<24>>24)==(-1); //@line 2425 "./libraries/stb/stb_image.h"
            if (!($152)) {
             $$sink$be$ph$sink = $$0$i36;
             break L54;
            }
           }
          }
         }
         while(1) {
          $153 = HEAP32[$z>>2]|0; //@line 2806 "./libraries/stb/stb_image.h"
          $154 = ((($153)) + 16|0); //@line 1240 "./libraries/stb/stb_image.h"
          $155 = HEAP32[$154>>2]|0; //@line 1240 "./libraries/stb/stb_image.h"
          $156 = ($155|0)==(0|0); //@line 1240 "./libraries/stb/stb_image.h"
          if ($156) {
           label = 61;
          } else {
           $157 = ((($153)) + 24|0); //@line 1241 "./libraries/stb/stb_image.h"
           $158 = HEAP32[$157>>2]|0; //@line 1241 "./libraries/stb/stb_image.h"
           $159 = ((($153)) + 28|0); //@line 1241 "./libraries/stb/stb_image.h"
           $160 = HEAP32[$159>>2]|0; //@line 1241 "./libraries/stb/stb_image.h"
           $161 = (FUNCTION_TABLE_ii[$158 & 8191]($160)|0); //@line 1241 "./libraries/stb/stb_image.h"
           $162 = ($161|0)==(0); //@line 1241 "./libraries/stb/stb_image.h"
           if (!($162)) {
            $163 = ((($153)) + 32|0); //@line 1244 "./libraries/stb/stb_image.h"
            $164 = HEAP32[$163>>2]|0; //@line 1244 "./libraries/stb/stb_image.h"
            $165 = ($164|0)==(0); //@line 1244 "./libraries/stb/stb_image.h"
            if ($165) {
             break L51;
            } else {
             label = 61;
            }
           }
          }
          if ((label|0) == 61) {
           label = 0;
           $166 = ((($153)) + 168|0); //@line 1247 "./libraries/stb/stb_image.h"
           $167 = HEAP32[$166>>2]|0; //@line 1247 "./libraries/stb/stb_image.h"
           $168 = ((($153)) + 172|0); //@line 1247 "./libraries/stb/stb_image.h"
           $169 = HEAP32[$168>>2]|0; //@line 1247 "./libraries/stb/stb_image.h"
           $170 = ($167>>>0)<($169>>>0); //@line 1247 "./libraries/stb/stb_image.h"
           if (!($170)) {
            break L51;
           }
          }
          $171 = HEAP8[$0>>0]|0; //@line 2422 "./libraries/stb/stb_image.h"
          $172 = ($171<<24>>24)==(-1); //@line 2422 "./libraries/stb/stb_image.h"
          if (!($172)) {
           $$lcssa103 = $171;
           label = 64;
           break;
          }
          $173 = HEAP32[$z>>2]|0; //@line 2423 "./libraries/stb/stb_image.h"
          $174 = ((($173)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
          $175 = HEAP32[$174>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $176 = ((($173)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
          $177 = HEAP32[$176>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $178 = ($175>>>0)<($177>>>0); //@line 1229 "./libraries/stb/stb_image.h"
          if ($178) {
           $179 = ((($175)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
           HEAP32[$174>>2] = $179; //@line 1230 "./libraries/stb/stb_image.h"
           $180 = HEAP8[$175>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
           $$0$i$i44 = $180;
          } else {
           $181 = ((($173)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $182 = HEAP32[$181>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
           $183 = ($182|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
           if ($183) {
            continue;
           }
           $184 = ((($173)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $185 = HEAP32[$184>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $186 = ((($173)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $187 = HEAP32[$186>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $188 = ((($173)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $189 = ((($173)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $190 = HEAP32[$189>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $191 = (FUNCTION_TABLE_iiii[$185 & 8191]($187,$188,$190)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $192 = ($191|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($192) {
            HEAP32[$181>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $193 = ((($173)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$188>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i$i41Z2D = $193;$$sink$i$i$i42 = $193;$563 = 0;
           } else {
            $194 = (((($173)) + 40|0) + ($191)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre$i$i39 = HEAP8[$188>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
            $$pre1$i$i40 = ((($173)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i$i41Z2D = $$pre1$i$i40;$$sink$i$i$i42 = $194;$563 = $$pre$i$i39;
           }
           HEAP32[$176>>2] = $$sink$i$i$i42; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$174>>2] = $$pre$phi$i$i41Z2D; //@line 1233 "./libraries/stb/stb_image.h"
           $$0$i$i44 = $563;
          }
          $195 = ($$0$i$i44<<24>>24)==(-1); //@line 2424 "./libraries/stb/stb_image.h"
          if ($195) {
           break;
          }
         }
         if ((label|0) == 64) {
          label = 0;
          HEAP8[$0>>0] = -1; //@line 2422 "./libraries/stb/stb_image.h"
          $$sink$be$ph$sink = $$lcssa103;
          break;
         }
         while(1) {
          $196 = HEAP32[$z>>2]|0; //@line 2426 "./libraries/stb/stb_image.h"
          $197 = ((($196)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
          $198 = HEAP32[$197>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $199 = ((($196)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
          $200 = HEAP32[$199>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $201 = ($198>>>0)<($200>>>0); //@line 1229 "./libraries/stb/stb_image.h"
          if ($201) {
           $202 = ((($198)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
           HEAP32[$197>>2] = $202; //@line 1230 "./libraries/stb/stb_image.h"
           $203 = HEAP8[$198>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
           $$0$i54 = $203;
          } else {
           $204 = ((($196)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $205 = HEAP32[$204>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
           $206 = ($205|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
           if ($206) {
            $$sink$be$ph$sink = 0;
            break L54;
           }
           $207 = ((($196)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $208 = HEAP32[$207>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $209 = ((($196)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $210 = HEAP32[$209>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $211 = ((($196)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $212 = ((($196)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $213 = HEAP32[$212>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $214 = (FUNCTION_TABLE_iiii[$208 & 8191]($210,$211,$213)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $215 = ($214|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($215) {
            HEAP32[$204>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $216 = ((($196)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$211>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i51Z2D = $216;$$sink$i$i52 = $216;$564 = 0;
           } else {
            $217 = (((($196)) + 40|0) + ($214)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre$i49 = HEAP8[$211>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
            $$pre1$i50 = ((($196)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i51Z2D = $$pre1$i50;$$sink$i$i52 = $217;$564 = $$pre$i49;
           }
           HEAP32[$199>>2] = $$sink$i$i52; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$197>>2] = $$pre$phi$i51Z2D; //@line 1233 "./libraries/stb/stb_image.h"
           $$0$i54 = $564;
          }
          $218 = ($$0$i54<<24>>24)==(-1); //@line 2425 "./libraries/stb/stb_image.h"
          if (!($218)) {
           $$sink$be$ph$sink = $$0$i54;
           break;
          }
         }
        } else {
         HEAP8[$0>>0] = -1; //@line 2422 "./libraries/stb/stb_image.h"
         $$sink$be$ph$sink = $105;
        }
       } while(0);
       $99 = $$sink$be$ph$sink&255; //@line 2803 "./libraries/stb/stb_image.h"
       $100 = $99 | 1; //@line 2801 "./libraries/stb/stb_image.h"
       $101 = ($100|0)==(193); //@line 2801 "./libraries/stb/stb_image.h"
       $102 = ($$sink$be$ph$sink<<24>>24)==(-62); //@line 2801 "./libraries/stb/stb_image.h"
       $$ = $102 | $101; //@line 2801 "./libraries/stb/stb_image.h"
       if ($$) {
        $$lcssa72 = $102;
        break L50;
       } else {
        $m$076 = $99;
       }
      }
      if ((label|0) == 200) {
       return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
      }
      HEAP32[158754] = 440881; //@line 894 "./libraries/stb/stb_image.h"
      $$0 = 0;
      return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
     }
    } while(0);
    $219 = $$lcssa72&1; //@line 2810 "./libraries/stb/stb_image.h"
    $220 = ((($z)) + 18124|0); //@line 2810 "./libraries/stb/stb_image.h"
    HEAP32[$220>>2] = $219; //@line 2810 "./libraries/stb/stb_image.h"
    $221 = HEAP32[$z>>2]|0; //@line 2704 "./libraries/stb/stb_image.h"
    $222 = ((($221)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
    $223 = HEAP32[$222>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
    $224 = ((($221)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
    $225 = HEAP32[$224>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
    $226 = ($223>>>0)<($225>>>0); //@line 1229 "./libraries/stb/stb_image.h"
    if ($226) {
     $227 = ((($223)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
     HEAP32[$222>>2] = $227; //@line 1230 "./libraries/stb/stb_image.h"
     $228 = HEAP8[$223>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
     $$0$i$i$i = $228;$246 = $227;$247 = $225;
    } else {
     $229 = ((($221)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
     $230 = HEAP32[$229>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
     $231 = ($230|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
     if ($231) {
      $$0$i$i$i = 0;$246 = $223;$247 = $225;
     } else {
      $232 = ((($221)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
      $233 = HEAP32[$232>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $234 = ((($221)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
      $235 = HEAP32[$234>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $236 = ((($221)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
      $237 = ((($221)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
      $238 = HEAP32[$237>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $239 = (FUNCTION_TABLE_iiii[$233 & 8191]($235,$236,$238)|0); //@line 1213 "./libraries/stb/stb_image.h"
      $240 = ($239|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
      if ($240) {
       HEAP32[$229>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
       $241 = ((($221)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
       HEAP8[$236>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
       $$pre$phi$i$i$iZ2D = $241;$$sink$i$i$i$i = $241;$565 = 0;
      } else {
       $242 = (((($221)) + 40|0) + ($239)|0); //@line 1223 "./libraries/stb/stb_image.h"
       $$pre$i$i$i = HEAP8[$236>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
       $$pre1$i$i$i = ((($221)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
       $$pre$phi$i$i$iZ2D = $$pre1$i$i$i;$$sink$i$i$i$i = $242;$565 = $$pre$i$i$i;
      }
      HEAP32[$224>>2] = $$sink$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
      HEAP32[$222>>2] = $$pre$phi$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
      $$0$i$i$i = $565;$246 = $$pre$phi$i$i$iZ2D;$247 = $$sink$i$i$i$i;
     }
    }
    $243 = $$0$i$i$i&255; //@line 1293 "./libraries/stb/stb_image.h"
    $244 = $243 << 8; //@line 1294 "./libraries/stb/stb_image.h"
    $245 = ($246>>>0)<($247>>>0); //@line 1229 "./libraries/stb/stb_image.h"
    if ($245) {
     $248 = ((($246)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
     HEAP32[$222>>2] = $248; //@line 1230 "./libraries/stb/stb_image.h"
     $249 = HEAP8[$246>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
     $$0$i6$i$i = $249;$268 = $248;$269 = $247;
    } else {
     $250 = ((($221)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
     $251 = HEAP32[$250>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
     $252 = ($251|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
     if ($252) {
      $$0$i6$i$i = 0;$268 = $246;$269 = $247;
     } else {
      $253 = ((($221)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
      $254 = HEAP32[$253>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $255 = ((($221)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
      $256 = HEAP32[$255>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $257 = ((($221)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
      $258 = ((($221)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
      $259 = HEAP32[$258>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $260 = (FUNCTION_TABLE_iiii[$254 & 8191]($256,$257,$259)|0); //@line 1213 "./libraries/stb/stb_image.h"
      $261 = ($260|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
      if ($261) {
       HEAP32[$250>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
       $262 = ((($221)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
       HEAP8[$257>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
       $$pre$phi$i3$i$iZ2D = $262;$$sink$i$i4$i$i = $262;$566 = 0;
      } else {
       $263 = (((($221)) + 40|0) + ($260)|0); //@line 1223 "./libraries/stb/stb_image.h"
       $$pre$i1$i$i = HEAP8[$257>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
       $$pre1$i2$i$i = ((($221)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
       $$pre$phi$i3$i$iZ2D = $$pre1$i2$i$i;$$sink$i$i4$i$i = $263;$566 = $$pre$i1$i$i;
      }
      HEAP32[$224>>2] = $$sink$i$i4$i$i; //@line 1223 "./libraries/stb/stb_image.h"
      HEAP32[$222>>2] = $$pre$phi$i3$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
      $$0$i6$i$i = $566;$268 = $$pre$phi$i3$i$iZ2D;$269 = $$sink$i$i4$i$i;
     }
    }
    $264 = $$0$i6$i$i&255; //@line 1294 "./libraries/stb/stb_image.h"
    $265 = $264 | $244; //@line 1294 "./libraries/stb/stb_image.h"
    $266 = ($265>>>0)<(11); //@line 2706 "./libraries/stb/stb_image.h"
    if ($266) {
     HEAP32[158754] = 440888; //@line 894 "./libraries/stb/stb_image.h"
     $$0 = 0;
     return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
    }
    $267 = ($268>>>0)<($269>>>0); //@line 1229 "./libraries/stb/stb_image.h"
    if ($267) {
     $270 = ((($268)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
     HEAP32[$222>>2] = $270; //@line 1230 "./libraries/stb/stb_image.h"
     $271 = HEAP8[$268>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
     $$0$i$i61 = $271;$288 = $270;$289 = $269;
     label = 104;
    } else {
     $272 = ((($221)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
     $273 = HEAP32[$272>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
     $274 = ($273|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
     if (!($274)) {
      $275 = ((($221)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
      $276 = HEAP32[$275>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $277 = ((($221)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
      $278 = HEAP32[$277>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $279 = ((($221)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
      $280 = ((($221)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
      $281 = HEAP32[$280>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $282 = (FUNCTION_TABLE_iiii[$276 & 8191]($278,$279,$281)|0); //@line 1213 "./libraries/stb/stb_image.h"
      $283 = ($282|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
      if ($283) {
       HEAP32[$272>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
       $284 = ((($221)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
       HEAP8[$279>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
       $$pre$phi$i$i58Z2D = $284;$$sink$i$i$i59 = $284;$567 = 0;
      } else {
       $285 = (((($221)) + 40|0) + ($282)|0); //@line 1223 "./libraries/stb/stb_image.h"
       $$pre$i$i56 = HEAP8[$279>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
       $$pre1$i$i57 = ((($221)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
       $$pre$phi$i$i58Z2D = $$pre1$i$i57;$$sink$i$i$i59 = $285;$567 = $$pre$i$i56;
      }
      HEAP32[$224>>2] = $$sink$i$i$i59; //@line 1223 "./libraries/stb/stb_image.h"
      HEAP32[$222>>2] = $$pre$phi$i$i58Z2D; //@line 1233 "./libraries/stb/stb_image.h"
      $$0$i$i61 = $567;$288 = $$pre$phi$i$i58Z2D;$289 = $$sink$i$i$i59;
      label = 104;
     }
    }
    if ((label|0) == 104) {
     $286 = ($$0$i$i61<<24>>24)==(8); //@line 2707 "./libraries/stb/stb_image.h"
     if ($286) {
      $287 = ($288>>>0)<($289>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      if ($287) {
       $290 = ((($288)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
       HEAP32[$222>>2] = $290; //@line 1230 "./libraries/stb/stb_image.h"
       $291 = HEAP8[$288>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
       $$0$i$i9$i = $291;$309 = $290;$310 = $289;
      } else {
       $292 = ((($221)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
       $293 = HEAP32[$292>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
       $294 = ($293|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
       if ($294) {
        $$0$i$i9$i = 0;$309 = $288;$310 = $289;
       } else {
        $295 = ((($221)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $296 = HEAP32[$295>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $297 = ((($221)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $298 = HEAP32[$297>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $299 = ((($221)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $300 = ((($221)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $301 = HEAP32[$300>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $302 = (FUNCTION_TABLE_iiii[$296 & 8191]($298,$299,$301)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $303 = ($302|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($303) {
         HEAP32[$292>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $304 = ((($221)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$299>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i6$iZ2D = $304;$$sink$i$i$i7$i = $304;$568 = 0;
        } else {
         $305 = (((($221)) + 40|0) + ($302)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i$i4$i = HEAP8[$299>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i$i5$i = ((($221)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i6$iZ2D = $$pre1$i$i5$i;$$sink$i$i$i7$i = $305;$568 = $$pre$i$i4$i;
        }
        HEAP32[$224>>2] = $$sink$i$i$i7$i; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$222>>2] = $$pre$phi$i$i6$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i$i9$i = $568;$309 = $$pre$phi$i$i6$iZ2D;$310 = $$sink$i$i$i7$i;
       }
      }
      $306 = $$0$i$i9$i&255; //@line 1293 "./libraries/stb/stb_image.h"
      $307 = $306 << 8; //@line 1294 "./libraries/stb/stb_image.h"
      $308 = ($309>>>0)<($310>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      if ($308) {
       $311 = ((($309)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
       HEAP32[$222>>2] = $311; //@line 1230 "./libraries/stb/stb_image.h"
       $312 = HEAP8[$309>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
       $$0$i6$i16$i = $312;$332 = $311;$333 = $310;
      } else {
       $313 = ((($221)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
       $314 = HEAP32[$313>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
       $315 = ($314|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
       if ($315) {
        $$0$i6$i16$i = 0;$332 = $309;$333 = $310;
       } else {
        $316 = ((($221)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $317 = HEAP32[$316>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $318 = ((($221)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $319 = HEAP32[$318>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $320 = ((($221)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $321 = ((($221)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $322 = HEAP32[$321>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $323 = (FUNCTION_TABLE_iiii[$317 & 8191]($319,$320,$322)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $324 = ($323|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($324) {
         HEAP32[$313>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $325 = ((($221)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$320>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i13$iZ2D = $325;$$sink$i$i4$i14$i = $325;$569 = 0;
        } else {
         $326 = (((($221)) + 40|0) + ($323)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i1$i11$i = HEAP8[$320>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i2$i12$i = ((($221)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i13$iZ2D = $$pre1$i2$i12$i;$$sink$i$i4$i14$i = $326;$569 = $$pre$i1$i11$i;
        }
        HEAP32[$224>>2] = $$sink$i$i4$i14$i; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$222>>2] = $$pre$phi$i3$i13$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i6$i16$i = $569;$332 = $$pre$phi$i3$i13$iZ2D;$333 = $$sink$i$i4$i14$i;
       }
      }
      $327 = $$0$i6$i16$i&255; //@line 1294 "./libraries/stb/stb_image.h"
      $328 = $327 | $307; //@line 1294 "./libraries/stb/stb_image.h"
      $329 = ((($221)) + 4|0); //@line 2708 "./libraries/stb/stb_image.h"
      HEAP32[$329>>2] = $328; //@line 2708 "./libraries/stb/stb_image.h"
      $330 = ($328|0)==(0); //@line 2708 "./libraries/stb/stb_image.h"
      if ($330) {
       HEAP32[158754] = 440911; //@line 894 "./libraries/stb/stb_image.h"
       $$0 = 0;
       return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
      }
      $331 = ($332>>>0)<($333>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      if ($331) {
       $334 = ((($332)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
       HEAP32[$222>>2] = $334; //@line 1230 "./libraries/stb/stb_image.h"
       $335 = HEAP8[$332>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
       $$0$i$i23$i = $335;$353 = $334;$354 = $333;
      } else {
       $336 = ((($221)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
       $337 = HEAP32[$336>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
       $338 = ($337|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
       if ($338) {
        $$0$i$i23$i = 0;$353 = $332;$354 = $333;
       } else {
        $339 = ((($221)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $340 = HEAP32[$339>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $341 = ((($221)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $342 = HEAP32[$341>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $343 = ((($221)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $344 = ((($221)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $345 = HEAP32[$344>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $346 = (FUNCTION_TABLE_iiii[$340 & 8191]($342,$343,$345)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $347 = ($346|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($347) {
         HEAP32[$336>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $348 = ((($221)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$343>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i20$iZ2D = $348;$$sink$i$i$i21$i = $348;$570 = 0;
        } else {
         $349 = (((($221)) + 40|0) + ($346)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i$i18$i = HEAP8[$343>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i$i19$i = ((($221)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i20$iZ2D = $$pre1$i$i19$i;$$sink$i$i$i21$i = $349;$570 = $$pre$i$i18$i;
        }
        HEAP32[$224>>2] = $$sink$i$i$i21$i; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$222>>2] = $$pre$phi$i$i20$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i$i23$i = $570;$353 = $$pre$phi$i$i20$iZ2D;$354 = $$sink$i$i$i21$i;
       }
      }
      $350 = $$0$i$i23$i&255; //@line 1293 "./libraries/stb/stb_image.h"
      $351 = $350 << 8; //@line 1294 "./libraries/stb/stb_image.h"
      $352 = ($353>>>0)<($354>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      if ($352) {
       $355 = ((($353)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
       HEAP32[$222>>2] = $355; //@line 1230 "./libraries/stb/stb_image.h"
       $356 = HEAP8[$353>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
       $$0$i6$i30$i = $356;$375 = $355;$376 = $354;
      } else {
       $357 = ((($221)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
       $358 = HEAP32[$357>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
       $359 = ($358|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
       if ($359) {
        $$0$i6$i30$i = 0;$375 = $353;$376 = $354;
       } else {
        $360 = ((($221)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $361 = HEAP32[$360>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $362 = ((($221)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $363 = HEAP32[$362>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $364 = ((($221)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $365 = ((($221)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $366 = HEAP32[$365>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $367 = (FUNCTION_TABLE_iiii[$361 & 8191]($363,$364,$366)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $368 = ($367|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($368) {
         HEAP32[$357>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $369 = ((($221)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$364>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i27$iZ2D = $369;$$sink$i$i4$i28$i = $369;$571 = 0;
        } else {
         $370 = (((($221)) + 40|0) + ($367)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i1$i25$i = HEAP8[$364>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i2$i26$i = ((($221)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i27$iZ2D = $$pre1$i2$i26$i;$$sink$i$i4$i28$i = $370;$571 = $$pre$i1$i25$i;
        }
        HEAP32[$224>>2] = $$sink$i$i4$i28$i; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$222>>2] = $$pre$phi$i3$i27$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i6$i30$i = $571;$375 = $$pre$phi$i3$i27$iZ2D;$376 = $$sink$i$i4$i28$i;
       }
      }
      $371 = $$0$i6$i30$i&255; //@line 1294 "./libraries/stb/stb_image.h"
      $372 = $371 | $351; //@line 1294 "./libraries/stb/stb_image.h"
      HEAP32[$221>>2] = $372; //@line 2709 "./libraries/stb/stb_image.h"
      $373 = ($372|0)==(0); //@line 2709 "./libraries/stb/stb_image.h"
      if ($373) {
       HEAP32[158754] = 440928; //@line 894 "./libraries/stb/stb_image.h"
       $$0 = 0;
       return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
      }
      $374 = ($375>>>0)<($376>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($374) {
        $377 = ((($375)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$222>>2] = $377; //@line 1230 "./libraries/stb/stb_image.h"
        $378 = HEAP8[$375>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i37$i = $378;$572 = $376;$573 = $377;
        label = 145;
       } else {
        $379 = ((($221)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $380 = HEAP32[$379>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $381 = ($380|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($381) {
         break;
        }
        $382 = ((($221)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $383 = HEAP32[$382>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $384 = ((($221)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $385 = HEAP32[$384>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $386 = ((($221)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $387 = ((($221)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $388 = HEAP32[$387>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $389 = (FUNCTION_TABLE_iiii[$383 & 8191]($385,$386,$388)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $390 = ($389|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($390) {
         HEAP32[$379>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $391 = ((($221)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$386>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i34$iZ2D = $391;$$sink$i$i35$i = $391;$574 = 0;
        } else {
         $392 = (((($221)) + 40|0) + ($389)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i32$i = HEAP8[$386>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i33$i = ((($221)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i34$iZ2D = $$pre1$i33$i;$$sink$i$i35$i = $392;$574 = $$pre$i32$i;
        }
        HEAP32[$224>>2] = $$sink$i$i35$i; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$222>>2] = $$pre$phi$i34$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i37$i = $574;$572 = $$sink$i$i35$i;$573 = $$pre$phi$i34$iZ2D;
        label = 145;
       }
      } while(0);
      L200: do {
       if ((label|0) == 145) {
        $393 = $$0$i37$i&255; //@line 2720 "./libraries/stb/stb_image.h"
        switch ($$0$i37$i<<24>>24) {
        case 1: case 3:  {
         break;
        }
        default: {
         break L200;
        }
        }
        $394 = ((($221)) + 8|0); //@line 2712 "./libraries/stb/stb_image.h"
        HEAP32[$394>>2] = $393; //@line 2712 "./libraries/stb/stb_image.h"
        $i$082$i = 0;
        while(1) {
         $395 = (((((($z)) + 17820|0) + (($i$082$i*72)|0)|0)) + 44|0); //@line 2714 "./libraries/stb/stb_image.h"
         HEAP32[$395>>2] = 0; //@line 2714 "./libraries/stb/stb_image.h"
         $396 = (((((($z)) + 17820|0) + (($i$082$i*72)|0)|0)) + 56|0); //@line 2715 "./libraries/stb/stb_image.h"
         HEAP32[$396>>2] = 0; //@line 2715 "./libraries/stb/stb_image.h"
         $397 = (($i$082$i) + 1)|0; //@line 2713 "./libraries/stb/stb_image.h"
         $exitcond$i = ($397|0)==($393|0); //@line 2713 "./libraries/stb/stb_image.h"
         if ($exitcond$i) {
          break;
         } else {
          $i$082$i = $397;
         }
        }
        $398 = ($393*3)|0; //@line 2718 "./libraries/stb/stb_image.h"
        $399 = (($398) + 8)|0; //@line 2718 "./libraries/stb/stb_image.h"
        $400 = ($265|0)==($399|0); //@line 2718 "./libraries/stb/stb_image.h"
        if (!($400)) {
         HEAP32[158754] = 440888; //@line 894 "./libraries/stb/stb_image.h"
         $$0 = 0;
         return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
        }
        $401 = ($$0$i37$i<<24>>24)==(0); //@line 2720 "./libraries/stb/stb_image.h"
        L210: do {
         if ($401) {
          $$pre$i63 = $393;
         } else {
          $402 = ((($221)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
          $403 = ((($221)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
          $404 = ((($221)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
          $405 = ((($221)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
          $406 = ((($221)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
          $407 = ((($221)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
          $409 = $573;$410 = $572;$i$178$i = 0;
          while(1) {
           $408 = ($409>>>0)<($410>>>0); //@line 1229 "./libraries/stb/stb_image.h"
           do {
            if ($408) {
             $411 = ((($409)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
             HEAP32[$222>>2] = $411; //@line 1230 "./libraries/stb/stb_image.h"
             $412 = HEAP8[$409>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
             $$0$i58$i = $412;$427 = $411;$428 = $410;
            } else {
             $413 = HEAP32[$402>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
             $414 = ($413|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
             if ($414) {
              $$0$i58$i = 0;$427 = $409;$428 = $410;
              break;
             }
             $415 = HEAP32[$403>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
             $416 = HEAP32[$404>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
             $417 = HEAP32[$406>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
             $418 = (FUNCTION_TABLE_iiii[$415 & 8191]($416,$405,$417)|0); //@line 1213 "./libraries/stb/stb_image.h"
             $419 = ($418|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
             if ($419) {
              HEAP32[$402>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
              HEAP8[$405>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
              $$sink$i$i56$i = $407;$575 = 0;
             } else {
              $420 = (((($221)) + 40|0) + ($418)|0); //@line 1223 "./libraries/stb/stb_image.h"
              $$pre$i53$i = HEAP8[$405>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
              $$sink$i$i56$i = $420;$575 = $$pre$i53$i;
             }
             HEAP32[$224>>2] = $$sink$i$i56$i; //@line 1223 "./libraries/stb/stb_image.h"
             HEAP32[$222>>2] = $407; //@line 1233 "./libraries/stb/stb_image.h"
             $$0$i58$i = $575;$427 = $407;$428 = $$sink$i$i56$i;
            }
           } while(0);
           $421 = $$0$i58$i&255; //@line 2721 "./libraries/stb/stb_image.h"
           $422 = (((($z)) + 17820|0) + (($i$178$i*72)|0)|0); //@line 2721 "./libraries/stb/stb_image.h"
           HEAP32[$422>>2] = $421; //@line 2721 "./libraries/stb/stb_image.h"
           $423 = (($i$178$i) + 1)|0; //@line 2722 "./libraries/stb/stb_image.h"
           $424 = ($421|0)==($423|0); //@line 2722 "./libraries/stb/stb_image.h"
           $425 = ($421|0)==($i$178$i|0); //@line 2723 "./libraries/stb/stb_image.h"
           $or$cond$i = $424 | $425; //@line 2722 "./libraries/stb/stb_image.h"
           if (!($or$cond$i)) {
            label = 161;
            break;
           }
           $426 = ($427>>>0)<($428>>>0); //@line 1229 "./libraries/stb/stb_image.h"
           do {
            if ($426) {
             $429 = ((($427)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
             HEAP32[$222>>2] = $429; //@line 1230 "./libraries/stb/stb_image.h"
             $430 = HEAP8[$427>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
             $$0$i51$i = $430;$449 = $429;$450 = $428;
            } else {
             $431 = HEAP32[$402>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
             $432 = ($431|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
             if ($432) {
              $$0$i51$i = 0;$449 = $427;$450 = $428;
              break;
             }
             $433 = HEAP32[$403>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
             $434 = HEAP32[$404>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
             $435 = HEAP32[$406>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
             $436 = (FUNCTION_TABLE_iiii[$433 & 8191]($434,$405,$435)|0); //@line 1213 "./libraries/stb/stb_image.h"
             $437 = ($436|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
             if ($437) {
              HEAP32[$402>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
              HEAP8[$405>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
              $$sink$i$i49$i = $407;$576 = 0;
             } else {
              $438 = (((($221)) + 40|0) + ($436)|0); //@line 1223 "./libraries/stb/stb_image.h"
              $$pre$i46$i = HEAP8[$405>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
              $$sink$i$i49$i = $438;$576 = $$pre$i46$i;
             }
             HEAP32[$224>>2] = $$sink$i$i49$i; //@line 1223 "./libraries/stb/stb_image.h"
             HEAP32[$222>>2] = $407; //@line 1233 "./libraries/stb/stb_image.h"
             $$0$i51$i = $576;$449 = $407;$450 = $$sink$i$i49$i;
            }
           } while(0);
           $439 = $$0$i51$i&255; //@line 2725 "./libraries/stb/stb_image.h"
           $440 = $439 >>> 4; //@line 2726 "./libraries/stb/stb_image.h"
           $441 = (((((($z)) + 17820|0) + (($i$178$i*72)|0)|0)) + 4|0); //@line 2726 "./libraries/stb/stb_image.h"
           HEAP32[$441>>2] = $440; //@line 2726 "./libraries/stb/stb_image.h"
           $442 = ($440|0)==(0); //@line 2726 "./libraries/stb/stb_image.h"
           $443 = ($$0$i51$i&255)>(79); //@line 2726 "./libraries/stb/stb_image.h"
           $or$cond2$i = $443 | $442; //@line 2726 "./libraries/stb/stb_image.h"
           if ($or$cond2$i) {
            label = 170;
            break;
           }
           $444 = $439 & 15; //@line 2727 "./libraries/stb/stb_image.h"
           $445 = (((((($z)) + 17820|0) + (($i$178$i*72)|0)|0)) + 8|0); //@line 2727 "./libraries/stb/stb_image.h"
           HEAP32[$445>>2] = $444; //@line 2727 "./libraries/stb/stb_image.h"
           $446 = (($444) + -1)|0; //@line 2727 "./libraries/stb/stb_image.h"
           $447 = ($446>>>0)>(3); //@line 2727 "./libraries/stb/stb_image.h"
           if ($447) {
            label = 172;
            break;
           }
           $448 = ($449>>>0)<($450>>>0); //@line 1229 "./libraries/stb/stb_image.h"
           do {
            if ($448) {
             $451 = ((($449)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
             HEAP32[$222>>2] = $451; //@line 1230 "./libraries/stb/stb_image.h"
             $452 = HEAP8[$449>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
             $$0$i44$i = $452;$577 = $450;$578 = $451;
             label = 181;
            } else {
             $453 = HEAP32[$402>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
             $454 = ($453|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
             if ($454) {
              $455 = (((((($z)) + 17820|0) + (($i$178$i*72)|0)|0)) + 12|0); //@line 2728 "./libraries/stb/stb_image.h"
              HEAP32[$455>>2] = 0; //@line 2728 "./libraries/stb/stb_image.h"
              $579 = $450;$580 = $449;
              break;
             }
             $456 = HEAP32[$403>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
             $457 = HEAP32[$404>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
             $458 = HEAP32[$406>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
             $459 = (FUNCTION_TABLE_iiii[$456 & 8191]($457,$405,$458)|0); //@line 1213 "./libraries/stb/stb_image.h"
             $460 = ($459|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
             if ($460) {
              HEAP32[$402>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
              HEAP8[$405>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
              $$sink$i$i42$i = $407;$581 = 0;
             } else {
              $461 = (((($221)) + 40|0) + ($459)|0); //@line 1223 "./libraries/stb/stb_image.h"
              $$pre$i39$i = HEAP8[$405>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
              $$sink$i$i42$i = $461;$581 = $$pre$i39$i;
             }
             HEAP32[$224>>2] = $$sink$i$i42$i; //@line 1223 "./libraries/stb/stb_image.h"
             HEAP32[$222>>2] = $407; //@line 1233 "./libraries/stb/stb_image.h"
             $$0$i44$i = $581;$577 = $$sink$i$i42$i;$578 = $407;
             label = 181;
            }
           } while(0);
           if ((label|0) == 181) {
            label = 0;
            $462 = $$0$i44$i&255; //@line 2728 "./libraries/stb/stb_image.h"
            $463 = (((((($z)) + 17820|0) + (($i$178$i*72)|0)|0)) + 12|0); //@line 2728 "./libraries/stb/stb_image.h"
            HEAP32[$463>>2] = $462; //@line 2728 "./libraries/stb/stb_image.h"
            $464 = ($$0$i44$i&255)>(3); //@line 2728 "./libraries/stb/stb_image.h"
            if ($464) {
             label = 183;
             break;
            } else {
             $579 = $577;$580 = $578;
            }
           }
           $465 = HEAP32[$394>>2]|0; //@line 2720 "./libraries/stb/stb_image.h"
           $466 = ($423|0)<($465|0); //@line 2720 "./libraries/stb/stb_image.h"
           if ($466) {
            $409 = $580;$410 = $579;$i$178$i = $423;
           } else {
            $$pre$i63 = $465;
            break L210;
           }
          }
          if ((label|0) == 161) {
           HEAP32[158754] = 440956; //@line 894 "./libraries/stb/stb_image.h"
           $$0 = 0;
           return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
          }
          else if ((label|0) == 170) {
           HEAP32[158754] = 440973; //@line 894 "./libraries/stb/stb_image.h"
           $$0 = 0;
           return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
          }
          else if ((label|0) == 172) {
           HEAP32[158754] = 440979; //@line 894 "./libraries/stb/stb_image.h"
           $$0 = 0;
           return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
          }
          else if ((label|0) == 183) {
           HEAP32[158754] = 440985; //@line 894 "./libraries/stb/stb_image.h"
           $$0 = 0;
           return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
          }
         }
        } while(0);
        $467 = ($scan|0)==(0); //@line 2731 "./libraries/stb/stb_image.h"
        if (!($467)) {
         $$0 = 1;
         return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
        }
        $468 = HEAP32[$221>>2]|0; //@line 2745 "./libraries/stb/stb_image.h"
        $469 = (1073741824 / ($468>>>0))&-1; //@line 2733 "./libraries/stb/stb_image.h"
        $470 = (($469>>>0) / ($$pre$i63>>>0))&-1; //@line 2733 "./libraries/stb/stb_image.h"
        $471 = HEAP32[$329>>2]|0; //@line 2746 "./libraries/stb/stb_image.h"
        $472 = ($470>>>0)<($471>>>0); //@line 2733 "./libraries/stb/stb_image.h"
        if ($472) {
         HEAP32[158754] = 440992; //@line 894 "./libraries/stb/stb_image.h"
         $$0 = 0;
         return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
        }
        $473 = ($$pre$i63|0)>(0); //@line 2735 "./libraries/stb/stb_image.h"
        if ($473) {
         $474 = HEAP32[$394>>2]|0; //@line 2735 "./libraries/stb/stb_image.h"
         $h_max$073$i = 1;$i$272$i = 0;$v_max$074$i = 1;
         while(1) {
          $475 = (((((($z)) + 17820|0) + (($i$272$i*72)|0)|0)) + 4|0); //@line 2736 "./libraries/stb/stb_image.h"
          $476 = HEAP32[$475>>2]|0; //@line 2736 "./libraries/stb/stb_image.h"
          $477 = ($476|0)>($h_max$073$i|0); //@line 2736 "./libraries/stb/stb_image.h"
          $$h_max$0$i = $477 ? $476 : $h_max$073$i; //@line 2736 "./libraries/stb/stb_image.h"
          $478 = (((((($z)) + 17820|0) + (($i$272$i*72)|0)|0)) + 8|0); //@line 2737 "./libraries/stb/stb_image.h"
          $479 = HEAP32[$478>>2]|0; //@line 2737 "./libraries/stb/stb_image.h"
          $480 = ($479|0)>($v_max$074$i|0); //@line 2737 "./libraries/stb/stb_image.h"
          $v_max$1$i = $480 ? $479 : $v_max$074$i; //@line 2737 "./libraries/stb/stb_image.h"
          $481 = (($i$272$i) + 1)|0; //@line 2735 "./libraries/stb/stb_image.h"
          $482 = ($481|0)<($474|0); //@line 2735 "./libraries/stb/stb_image.h"
          if ($482) {
           $h_max$073$i = $$h_max$0$i;$i$272$i = $481;$v_max$074$i = $v_max$1$i;
          } else {
           $498 = $474;$h_max$0$lcssa$i = $$h_max$0$i;$v_max$0$lcssa$i = $v_max$1$i;
           break;
          }
         }
        } else {
         $498 = $$pre$i63;$h_max$0$lcssa$i = 1;$v_max$0$lcssa$i = 1;
        }
        $483 = ((($z)) + 17796|0); //@line 2741 "./libraries/stb/stb_image.h"
        HEAP32[$483>>2] = $h_max$0$lcssa$i; //@line 2741 "./libraries/stb/stb_image.h"
        $484 = ((($z)) + 17800|0); //@line 2742 "./libraries/stb/stb_image.h"
        HEAP32[$484>>2] = $v_max$0$lcssa$i; //@line 2742 "./libraries/stb/stb_image.h"
        $485 = $h_max$0$lcssa$i << 3; //@line 2743 "./libraries/stb/stb_image.h"
        $486 = ((($z)) + 17812|0); //@line 2743 "./libraries/stb/stb_image.h"
        HEAP32[$486>>2] = $485; //@line 2743 "./libraries/stb/stb_image.h"
        $487 = $v_max$0$lcssa$i << 3; //@line 2744 "./libraries/stb/stb_image.h"
        $488 = ((($z)) + 17816|0); //@line 2744 "./libraries/stb/stb_image.h"
        HEAP32[$488>>2] = $487; //@line 2744 "./libraries/stb/stb_image.h"
        $489 = (($468) + -1)|0; //@line 2745 "./libraries/stb/stb_image.h"
        $490 = (($489) + ($485))|0; //@line 2745 "./libraries/stb/stb_image.h"
        $491 = (($490>>>0) / ($485>>>0))&-1; //@line 2745 "./libraries/stb/stb_image.h"
        $492 = ((($z)) + 17804|0); //@line 2745 "./libraries/stb/stb_image.h"
        HEAP32[$492>>2] = $491; //@line 2745 "./libraries/stb/stb_image.h"
        $493 = (($471) + -1)|0; //@line 2746 "./libraries/stb/stb_image.h"
        $494 = (($493) + ($487))|0; //@line 2746 "./libraries/stb/stb_image.h"
        $495 = (($494>>>0) / ($487>>>0))&-1; //@line 2746 "./libraries/stb/stb_image.h"
        $496 = ((($z)) + 17808|0); //@line 2746 "./libraries/stb/stb_image.h"
        HEAP32[$496>>2] = $495; //@line 2746 "./libraries/stb/stb_image.h"
        $497 = ($498|0)>(0); //@line 2748 "./libraries/stb/stb_image.h"
        if (!($497)) {
         $$0 = 1;
         return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
        }
        $499 = (($h_max$0$lcssa$i) + -1)|0; //@line 2750 "./libraries/stb/stb_image.h"
        $500 = (($v_max$0$lcssa$i) + -1)|0; //@line 2751 "./libraries/stb/stb_image.h"
        $504 = $468;$511 = $471;$i$369$i = 0;
        while(1) {
         $501 = (((((($z)) + 17820|0) + (($i$369$i*72)|0)|0)) + 4|0); //@line 2750 "./libraries/stb/stb_image.h"
         $502 = HEAP32[$501>>2]|0; //@line 2750 "./libraries/stb/stb_image.h"
         $503 = Math_imul($502, $504)|0; //@line 2750 "./libraries/stb/stb_image.h"
         $505 = (($499) + ($503))|0; //@line 2750 "./libraries/stb/stb_image.h"
         $506 = (($505>>>0) / ($h_max$0$lcssa$i>>>0))&-1; //@line 2750 "./libraries/stb/stb_image.h"
         $507 = (((((($z)) + 17820|0) + (($i$369$i*72)|0)|0)) + 28|0); //@line 2750 "./libraries/stb/stb_image.h"
         HEAP32[$507>>2] = $506; //@line 2750 "./libraries/stb/stb_image.h"
         $508 = (((((($z)) + 17820|0) + (($i$369$i*72)|0)|0)) + 8|0); //@line 2751 "./libraries/stb/stb_image.h"
         $509 = HEAP32[$508>>2]|0; //@line 2751 "./libraries/stb/stb_image.h"
         $510 = Math_imul($509, $511)|0; //@line 2751 "./libraries/stb/stb_image.h"
         $512 = (($500) + ($510))|0; //@line 2751 "./libraries/stb/stb_image.h"
         $513 = (($512>>>0) / ($v_max$0$lcssa$i>>>0))&-1; //@line 2751 "./libraries/stb/stb_image.h"
         $514 = (((((($z)) + 17820|0) + (($i$369$i*72)|0)|0)) + 32|0); //@line 2751 "./libraries/stb/stb_image.h"
         HEAP32[$514>>2] = $513; //@line 2751 "./libraries/stb/stb_image.h"
         $515 = HEAP32[$492>>2]|0; //@line 2756 "./libraries/stb/stb_image.h"
         $516 = $515 << 3; //@line 2756 "./libraries/stb/stb_image.h"
         $517 = Math_imul($516, $502)|0; //@line 2771 "./libraries/stb/stb_image.h"
         $518 = (((((($z)) + 17820|0) + (($i$369$i*72)|0)|0)) + 36|0); //@line 2756 "./libraries/stb/stb_image.h"
         HEAP32[$518>>2] = $517; //@line 2756 "./libraries/stb/stb_image.h"
         $519 = HEAP32[$496>>2]|0; //@line 2757 "./libraries/stb/stb_image.h"
         $520 = $519 << 3; //@line 2757 "./libraries/stb/stb_image.h"
         $521 = Math_imul($520, $509)|0; //@line 2772 "./libraries/stb/stb_image.h"
         $522 = (((((($z)) + 17820|0) + (($i$369$i*72)|0)|0)) + 40|0); //@line 2757 "./libraries/stb/stb_image.h"
         HEAP32[$522>>2] = $521; //@line 2757 "./libraries/stb/stb_image.h"
         $523 = Math_imul($521, $517)|0; //@line 2758 "./libraries/stb/stb_image.h"
         $524 = $523 | 15; //@line 2758 "./libraries/stb/stb_image.h"
         $525 = (_malloc($524)|0); //@line 900 "./libraries/stb/stb_image.h"
         $526 = (((((($z)) + 17820|0) + (($i$369$i*72)|0)|0)) + 48|0); //@line 2758 "./libraries/stb/stb_image.h"
         HEAP32[$526>>2] = $525; //@line 2758 "./libraries/stb/stb_image.h"
         $527 = ($525|0)==(0|0); //@line 2760 "./libraries/stb/stb_image.h"
         if ($527) {
          $i$369$i$lcssa = $i$369$i;
          break;
         }
         $532 = $525; //@line 2768 "./libraries/stb/stb_image.h"
         $533 = (($532) + 15)|0; //@line 2768 "./libraries/stb/stb_image.h"
         $534 = $533 & -16; //@line 2768 "./libraries/stb/stb_image.h"
         $535 = $534; //@line 2768 "./libraries/stb/stb_image.h"
         $536 = (((((($z)) + 17820|0) + (($i$369$i*72)|0)|0)) + 44|0); //@line 2768 "./libraries/stb/stb_image.h"
         HEAP32[$536>>2] = $535; //@line 2768 "./libraries/stb/stb_image.h"
         $537 = (((((($z)) + 17820|0) + (($i$369$i*72)|0)|0)) + 56|0); //@line 2769 "./libraries/stb/stb_image.h"
         HEAP32[$537>>2] = 0; //@line 2769 "./libraries/stb/stb_image.h"
         $538 = HEAP32[$220>>2]|0; //@line 2770 "./libraries/stb/stb_image.h"
         $539 = ($538|0)==(0); //@line 2770 "./libraries/stb/stb_image.h"
         if ($539) {
          $$sink$i = 0;$$sink88$i = 0;
         } else {
          $540 = $517 >> 3; //@line 2771 "./libraries/stb/stb_image.h"
          $541 = (((((($z)) + 17820|0) + (($i$369$i*72)|0)|0)) + 64|0); //@line 2771 "./libraries/stb/stb_image.h"
          HEAP32[$541>>2] = $540; //@line 2771 "./libraries/stb/stb_image.h"
          $542 = $521 >> 3; //@line 2772 "./libraries/stb/stb_image.h"
          $543 = (((((($z)) + 17820|0) + (($i$369$i*72)|0)|0)) + 68|0); //@line 2772 "./libraries/stb/stb_image.h"
          HEAP32[$543>>2] = $542; //@line 2772 "./libraries/stb/stb_image.h"
          $544 = $517 << 4; //@line 2773 "./libraries/stb/stb_image.h"
          $545 = Math_imul($544, $542)|0; //@line 2773 "./libraries/stb/stb_image.h"
          $546 = $545 | 15; //@line 2773 "./libraries/stb/stb_image.h"
          $547 = (_malloc($546)|0); //@line 2773 "./libraries/stb/stb_image.h"
          $548 = $547; //@line 2774 "./libraries/stb/stb_image.h"
          $549 = (($548) + 15)|0; //@line 2774 "./libraries/stb/stb_image.h"
          $550 = $549 & -16; //@line 2774 "./libraries/stb/stb_image.h"
          $551 = $550; //@line 2774 "./libraries/stb/stb_image.h"
          $$sink$i = $551;$$sink88$i = $547;
         }
         $552 = (((((($z)) + 17820|0) + (($i$369$i*72)|0)|0)) + 52|0); //@line 2773 "./libraries/stb/stb_image.h"
         HEAP32[$552>>2] = $$sink88$i; //@line 2773 "./libraries/stb/stb_image.h"
         $553 = (((((($z)) + 17820|0) + (($i$369$i*72)|0)|0)) + 60|0); //@line 2774 "./libraries/stb/stb_image.h"
         HEAP32[$553>>2] = $$sink$i; //@line 2774 "./libraries/stb/stb_image.h"
         $554 = (($i$369$i) + 1)|0; //@line 2748 "./libraries/stb/stb_image.h"
         $555 = HEAP32[$394>>2]|0; //@line 2748 "./libraries/stb/stb_image.h"
         $556 = ($554|0)<($555|0); //@line 2748 "./libraries/stb/stb_image.h"
         if (!($556)) {
          $$0 = 1;
          label = 200;
          break;
         }
         $$pre90$i = HEAP32[$221>>2]|0; //@line 2750 "./libraries/stb/stb_image.h"
         $$pre91$i = HEAP32[$329>>2]|0; //@line 2751 "./libraries/stb/stb_image.h"
         $504 = $$pre90$i;$511 = $$pre91$i;$i$369$i = $554;
        }
        if ((label|0) == 200) {
         return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
        }
        $528 = ($i$369$i$lcssa|0)>(0); //@line 2761 "./libraries/stb/stb_image.h"
        if ($528) {
         $i$468$in$i = $i$369$i$lcssa;
         while(1) {
          $i$468$i = (($i$468$in$i) + -1)|0; //@line 2761 "./libraries/stb/stb_image.h"
          $529 = (((((($z)) + 17820|0) + (($i$468$i*72)|0)|0)) + 48|0); //@line 2762 "./libraries/stb/stb_image.h"
          $530 = HEAP32[$529>>2]|0; //@line 2762 "./libraries/stb/stb_image.h"
          _free($530); //@line 2762 "./libraries/stb/stb_image.h"
          HEAP32[$529>>2] = 0; //@line 2763 "./libraries/stb/stb_image.h"
          $531 = ($i$468$in$i|0)>(1); //@line 2761 "./libraries/stb/stb_image.h"
          if ($531) {
           $i$468$in$i = $i$468$i;
          } else {
           break;
          }
         }
        }
        HEAP32[158754] = 441002; //@line 894 "./libraries/stb/stb_image.h"
        $$0 = 0;
        return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
       }
      } while(0);
      HEAP32[158754] = 440936; //@line 894 "./libraries/stb/stb_image.h"
      $$0 = 0;
      return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
     }
    }
    HEAP32[158754] = 440900; //@line 894 "./libraries/stb/stb_image.h"
    $$0 = 0;
    return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
   }
  }
 } while(0);
 HEAP32[158754] = 440708; //@line 894 "./libraries/stb/stb_image.h"
 $$0 = 0;
 return ($$0|0); //@line 2813 "./libraries/stb/stb_image.h"
}
function __ZL20stbi__process_markerP10stbi__jpegi($z,$m) {
 $z = $z|0;
 $m = $m|0;
 var $$0$i = 0, $$0$i$i = 0, $$0$i$i14 = 0, $$0$i$i28 = 0, $$0$i$i51 = 0, $$0$i$i86 = 0, $$0$i44 = 0, $$0$i6$i = 0, $$0$i6$i21 = 0, $$0$i6$i35 = 0, $$0$i6$i58 = 0, $$0$i6$i93 = 0, $$0$i65 = 0, $$0$i72 = 0, $$0$i79 = 0, $$4 = 0, $$4$shrunk = 0, $$in = 0, $$lcssa131 = 0, $$mask = 0;
 var $$mask97 = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i23 = 0, $$pre$i$i46 = 0, $$pre$i$i81 = 0, $$pre$i$i9 = 0, $$pre$i1$i = 0, $$pre$i1$i16 = 0, $$pre$i1$i30 = 0, $$pre$i1$i53 = 0, $$pre$i1$i88 = 0, $$pre$i37 = 0, $$pre$i39 = 0, $$pre$i60 = 0, $$pre$i67 = 0, $$pre$i74 = 0, $$pre$phi$i$i11Z2D = 0, $$pre$phi$i$i25Z2D = 0, $$pre$phi$i$i48Z2D = 0;
 var $$pre$phi$i$i83Z2D = 0, $$pre$phi$i$iZ2D = 0, $$pre$phi$i3$i18Z2D = 0, $$pre$phi$i3$i32Z2D = 0, $$pre$phi$i3$i55Z2D = 0, $$pre$phi$i3$i90Z2D = 0, $$pre$phi$i3$iZ2D = 0, $$pre$phi$i38Z2D = 0, $$pre$phi$i41Z2D = 0, $$pre$phi$i62Z2D = 0, $$pre$phi$i69Z2D = 0, $$pre$phi$i76Z2D = 0, $$pre1$i = 0, $$pre1$i$i = 0, $$pre1$i$i10 = 0, $$pre1$i$i24 = 0, $$pre1$i$i47 = 0, $$pre1$i$i82 = 0, $$pre1$i2$i = 0, $$pre1$i2$i17 = 0;
 var $$pre1$i2$i31 = 0, $$pre1$i2$i54 = 0, $$pre1$i2$i89 = 0, $$pre1$i40 = 0, $$pre1$i61 = 0, $$pre1$i68 = 0, $$pre1$i75 = 0, $$sink$i$i = 0, $$sink$i$i$i = 0, $$sink$i$i$i12 = 0, $$sink$i$i$i26 = 0, $$sink$i$i$i49 = 0, $$sink$i$i$i84 = 0, $$sink$i$i4$i = 0, $$sink$i$i4$i19 = 0, $$sink$i$i4$i33 = 0, $$sink$i$i4$i56 = 0, $$sink$i$i4$i91 = 0, $$sink$i$i42 = 0, $$sink$i$i63 = 0;
 var $$sink$i$i70 = 0, $$sink$i$i77 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0;
 var $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0;
 var $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0;
 var $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0;
 var $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0;
 var $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0;
 var $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0;
 var $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0;
 var $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0;
 var $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0;
 var $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0;
 var $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $L$0$lcssa = 0, $L$0106 = 0, $L$2$lcssa = 0, $L$2113 = 0, $exitcond = 0, $exitcond$i = 0, $exitcond119 = 0, $exitcond120 = 0, $i$0105 = 0, $i$02$i = 0, $i1$0107 = 0;
 var $i1$1109 = 0, $k$0$i = 0, $k$0$off$i = 0, $n$0108 = 0, $or$cond = 0, $or$cond6 = 0, $sizes = 0, $v$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $sizes = sp;
 switch ($m|0) {
 case 255:  {
  HEAP32[158754] = 440715; //@line 894 "./libraries/stb/stb_image.h"
  $$4$shrunk = 0;
  $$4 = $$4$shrunk&1;
  STACKTOP = sp;return ($$4|0); //@line 2660 "./libraries/stb/stb_image.h"
  break;
 }
 case 221:  {
  $0 = HEAP32[$z>>2]|0; //@line 2607 "./libraries/stb/stb_image.h"
  $1 = ((($0)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
  $2 = HEAP32[$1>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
  $3 = ((($0)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
  $4 = HEAP32[$3>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
  $5 = ($2>>>0)<($4>>>0); //@line 1229 "./libraries/stb/stb_image.h"
  if ($5) {
   $6 = ((($2)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
   HEAP32[$1>>2] = $6; //@line 1230 "./libraries/stb/stb_image.h"
   $7 = HEAP8[$2>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
   $$0$i$i = $7;$25 = $6;$26 = $4;
  } else {
   $8 = ((($0)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
   $9 = HEAP32[$8>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
   $10 = ($9|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
   if ($10) {
    $$0$i$i = 0;$25 = $2;$26 = $4;
   } else {
    $11 = ((($0)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
    $12 = HEAP32[$11>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $13 = ((($0)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
    $14 = HEAP32[$13>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $15 = ((($0)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
    $16 = ((($0)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
    $17 = HEAP32[$16>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $18 = (FUNCTION_TABLE_iiii[$12 & 8191]($14,$15,$17)|0); //@line 1213 "./libraries/stb/stb_image.h"
    $19 = ($18|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
    if ($19) {
     HEAP32[$8>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
     $20 = ((($0)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
     HEAP8[$15>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
     $$pre$phi$i$iZ2D = $20;$$sink$i$i$i = $20;$431 = 0;
    } else {
     $21 = (((($0)) + 40|0) + ($18)|0); //@line 1223 "./libraries/stb/stb_image.h"
     $$pre$i$i = HEAP8[$15>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
     $$pre1$i$i = ((($0)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
     $$pre$phi$i$iZ2D = $$pre1$i$i;$$sink$i$i$i = $21;$431 = $$pre$i$i;
    }
    HEAP32[$3>>2] = $$sink$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
    HEAP32[$1>>2] = $$pre$phi$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
    $$0$i$i = $431;$25 = $$pre$phi$i$iZ2D;$26 = $$sink$i$i$i;
   }
  }
  $22 = $$0$i$i&255; //@line 1293 "./libraries/stb/stb_image.h"
  $23 = $22 << 8; //@line 1294 "./libraries/stb/stb_image.h"
  $24 = ($25>>>0)<($26>>>0); //@line 1229 "./libraries/stb/stb_image.h"
  if ($24) {
   $27 = ((($25)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
   HEAP32[$1>>2] = $27; //@line 1230 "./libraries/stb/stb_image.h"
   $28 = HEAP8[$25>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
   $$0$i6$i = $28;
  } else {
   $29 = ((($0)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
   $30 = HEAP32[$29>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
   $31 = ($30|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
   if ($31) {
    $$0$i6$i = 0;
   } else {
    $32 = ((($0)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
    $33 = HEAP32[$32>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $34 = ((($0)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
    $35 = HEAP32[$34>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $36 = ((($0)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
    $37 = ((($0)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
    $38 = HEAP32[$37>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $39 = (FUNCTION_TABLE_iiii[$33 & 8191]($35,$36,$38)|0); //@line 1213 "./libraries/stb/stb_image.h"
    $40 = ($39|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
    if ($40) {
     HEAP32[$29>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
     $41 = ((($0)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
     HEAP8[$36>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
     $$pre$phi$i3$iZ2D = $41;$$sink$i$i4$i = $41;$432 = 0;
    } else {
     $42 = (((($0)) + 40|0) + ($39)|0); //@line 1223 "./libraries/stb/stb_image.h"
     $$pre$i1$i = HEAP8[$36>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
     $$pre1$i2$i = ((($0)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
     $$pre$phi$i3$iZ2D = $$pre1$i2$i;$$sink$i$i4$i = $42;$432 = $$pre$i1$i;
    }
    HEAP32[$3>>2] = $$sink$i$i4$i; //@line 1223 "./libraries/stb/stb_image.h"
    HEAP32[$1>>2] = $$pre$phi$i3$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
    $$0$i6$i = $432;
   }
  }
  $43 = $$0$i6$i&255; //@line 1294 "./libraries/stb/stb_image.h"
  $44 = $43 | $23; //@line 1294 "./libraries/stb/stb_image.h"
  $45 = ($44|0)==(4); //@line 2607 "./libraries/stb/stb_image.h"
  if (!($45)) {
   HEAP32[158754] = 440731; //@line 894 "./libraries/stb/stb_image.h"
   $$4$shrunk = 0;
   $$4 = $$4$shrunk&1;
   STACKTOP = sp;return ($$4|0); //@line 2660 "./libraries/stb/stb_image.h"
  }
  $46 = HEAP32[$z>>2]|0; //@line 2608 "./libraries/stb/stb_image.h"
  $47 = ((($46)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
  $48 = HEAP32[$47>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
  $49 = ((($46)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
  $50 = HEAP32[$49>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
  $51 = ($48>>>0)<($50>>>0); //@line 1229 "./libraries/stb/stb_image.h"
  if ($51) {
   $52 = ((($48)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
   HEAP32[$47>>2] = $52; //@line 1230 "./libraries/stb/stb_image.h"
   $53 = HEAP8[$48>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
   $$0$i$i14 = $53;$71 = $52;$72 = $50;
  } else {
   $54 = ((($46)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
   $55 = HEAP32[$54>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
   $56 = ($55|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
   if ($56) {
    $$0$i$i14 = 0;$71 = $48;$72 = $50;
   } else {
    $57 = ((($46)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
    $58 = HEAP32[$57>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $59 = ((($46)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
    $60 = HEAP32[$59>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $61 = ((($46)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
    $62 = ((($46)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
    $63 = HEAP32[$62>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $64 = (FUNCTION_TABLE_iiii[$58 & 8191]($60,$61,$63)|0); //@line 1213 "./libraries/stb/stb_image.h"
    $65 = ($64|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
    if ($65) {
     HEAP32[$54>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
     $66 = ((($46)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
     HEAP8[$61>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
     $$pre$phi$i$i11Z2D = $66;$$sink$i$i$i12 = $66;$433 = 0;
    } else {
     $67 = (((($46)) + 40|0) + ($64)|0); //@line 1223 "./libraries/stb/stb_image.h"
     $$pre$i$i9 = HEAP8[$61>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
     $$pre1$i$i10 = ((($46)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
     $$pre$phi$i$i11Z2D = $$pre1$i$i10;$$sink$i$i$i12 = $67;$433 = $$pre$i$i9;
    }
    HEAP32[$49>>2] = $$sink$i$i$i12; //@line 1223 "./libraries/stb/stb_image.h"
    HEAP32[$47>>2] = $$pre$phi$i$i11Z2D; //@line 1233 "./libraries/stb/stb_image.h"
    $$0$i$i14 = $433;$71 = $$pre$phi$i$i11Z2D;$72 = $$sink$i$i$i12;
   }
  }
  $68 = $$0$i$i14&255; //@line 1293 "./libraries/stb/stb_image.h"
  $69 = $68 << 8; //@line 1294 "./libraries/stb/stb_image.h"
  $70 = ($71>>>0)<($72>>>0); //@line 1229 "./libraries/stb/stb_image.h"
  if ($70) {
   $73 = ((($71)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
   HEAP32[$47>>2] = $73; //@line 1230 "./libraries/stb/stb_image.h"
   $74 = HEAP8[$71>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
   $$0$i6$i21 = $74;
  } else {
   $75 = ((($46)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
   $76 = HEAP32[$75>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
   $77 = ($76|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
   if ($77) {
    $$0$i6$i21 = 0;
   } else {
    $78 = ((($46)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
    $79 = HEAP32[$78>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $80 = ((($46)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
    $81 = HEAP32[$80>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $82 = ((($46)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
    $83 = ((($46)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
    $84 = HEAP32[$83>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $85 = (FUNCTION_TABLE_iiii[$79 & 8191]($81,$82,$84)|0); //@line 1213 "./libraries/stb/stb_image.h"
    $86 = ($85|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
    if ($86) {
     HEAP32[$75>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
     $87 = ((($46)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
     HEAP8[$82>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
     $$pre$phi$i3$i18Z2D = $87;$$sink$i$i4$i19 = $87;$434 = 0;
    } else {
     $88 = (((($46)) + 40|0) + ($85)|0); //@line 1223 "./libraries/stb/stb_image.h"
     $$pre$i1$i16 = HEAP8[$82>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
     $$pre1$i2$i17 = ((($46)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
     $$pre$phi$i3$i18Z2D = $$pre1$i2$i17;$$sink$i$i4$i19 = $88;$434 = $$pre$i1$i16;
    }
    HEAP32[$49>>2] = $$sink$i$i4$i19; //@line 1223 "./libraries/stb/stb_image.h"
    HEAP32[$47>>2] = $$pre$phi$i3$i18Z2D; //@line 1233 "./libraries/stb/stb_image.h"
    $$0$i6$i21 = $434;
   }
  }
  $89 = $$0$i6$i21&255; //@line 1294 "./libraries/stb/stb_image.h"
  $90 = $89 | $69; //@line 1294 "./libraries/stb/stb_image.h"
  $91 = ((($z)) + 18168|0); //@line 2608 "./libraries/stb/stb_image.h"
  HEAP32[$91>>2] = $90; //@line 2608 "./libraries/stb/stb_image.h"
  $$4$shrunk = 1;
  $$4 = $$4$shrunk&1;
  STACKTOP = sp;return ($$4|0); //@line 2660 "./libraries/stb/stb_image.h"
  break;
 }
 case 219:  {
  $92 = HEAP32[$z>>2]|0; //@line 2612 "./libraries/stb/stb_image.h"
  $93 = ((($92)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
  $94 = HEAP32[$93>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
  $95 = ((($92)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
  $96 = HEAP32[$95>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
  $97 = ($94>>>0)<($96>>>0); //@line 1229 "./libraries/stb/stb_image.h"
  if ($97) {
   $98 = ((($94)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
   HEAP32[$93>>2] = $98; //@line 1230 "./libraries/stb/stb_image.h"
   $99 = HEAP8[$94>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
   $$0$i$i28 = $99;$117 = $98;$118 = $96;
  } else {
   $100 = ((($92)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
   $101 = HEAP32[$100>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
   $102 = ($101|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
   if ($102) {
    $$0$i$i28 = 0;$117 = $94;$118 = $96;
   } else {
    $103 = ((($92)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
    $104 = HEAP32[$103>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $105 = ((($92)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
    $106 = HEAP32[$105>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $107 = ((($92)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
    $108 = ((($92)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
    $109 = HEAP32[$108>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $110 = (FUNCTION_TABLE_iiii[$104 & 8191]($106,$107,$109)|0); //@line 1213 "./libraries/stb/stb_image.h"
    $111 = ($110|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
    if ($111) {
     HEAP32[$100>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
     $112 = ((($92)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
     HEAP8[$107>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
     $$pre$phi$i$i25Z2D = $112;$$sink$i$i$i26 = $112;$435 = 0;
    } else {
     $113 = (((($92)) + 40|0) + ($110)|0); //@line 1223 "./libraries/stb/stb_image.h"
     $$pre$i$i23 = HEAP8[$107>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
     $$pre1$i$i24 = ((($92)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
     $$pre$phi$i$i25Z2D = $$pre1$i$i24;$$sink$i$i$i26 = $113;$435 = $$pre$i$i23;
    }
    HEAP32[$95>>2] = $$sink$i$i$i26; //@line 1223 "./libraries/stb/stb_image.h"
    HEAP32[$93>>2] = $$pre$phi$i$i25Z2D; //@line 1233 "./libraries/stb/stb_image.h"
    $$0$i$i28 = $435;$117 = $$pre$phi$i$i25Z2D;$118 = $$sink$i$i$i26;
   }
  }
  $114 = $$0$i$i28&255; //@line 1293 "./libraries/stb/stb_image.h"
  $115 = $114 << 8; //@line 1294 "./libraries/stb/stb_image.h"
  $116 = ($117>>>0)<($118>>>0); //@line 1229 "./libraries/stb/stb_image.h"
  if ($116) {
   $119 = ((($117)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
   HEAP32[$93>>2] = $119; //@line 1230 "./libraries/stb/stb_image.h"
   $120 = HEAP8[$117>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
   $$0$i6$i35 = $120;
  } else {
   $121 = ((($92)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
   $122 = HEAP32[$121>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
   $123 = ($122|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
   if ($123) {
    $$0$i6$i35 = 0;
   } else {
    $124 = ((($92)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
    $125 = HEAP32[$124>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $126 = ((($92)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
    $127 = HEAP32[$126>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $128 = ((($92)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
    $129 = ((($92)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
    $130 = HEAP32[$129>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $131 = (FUNCTION_TABLE_iiii[$125 & 8191]($127,$128,$130)|0); //@line 1213 "./libraries/stb/stb_image.h"
    $132 = ($131|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
    if ($132) {
     HEAP32[$121>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
     $133 = ((($92)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
     HEAP8[$128>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
     $$pre$phi$i3$i32Z2D = $133;$$sink$i$i4$i33 = $133;$436 = 0;
    } else {
     $134 = (((($92)) + 40|0) + ($131)|0); //@line 1223 "./libraries/stb/stb_image.h"
     $$pre$i1$i30 = HEAP8[$128>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
     $$pre1$i2$i31 = ((($92)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
     $$pre$phi$i3$i32Z2D = $$pre1$i2$i31;$$sink$i$i4$i33 = $134;$436 = $$pre$i1$i30;
    }
    HEAP32[$95>>2] = $$sink$i$i4$i33; //@line 1223 "./libraries/stb/stb_image.h"
    HEAP32[$93>>2] = $$pre$phi$i3$i32Z2D; //@line 1233 "./libraries/stb/stb_image.h"
    $$0$i6$i35 = $436;
   }
  }
  $135 = $$0$i6$i35&255; //@line 1294 "./libraries/stb/stb_image.h"
  $136 = $135 | $115; //@line 1294 "./libraries/stb/stb_image.h"
  $137 = (($136) + -2)|0; //@line 2612 "./libraries/stb/stb_image.h"
  $138 = ($136>>>0)>(2); //@line 2613 "./libraries/stb/stb_image.h"
  L97: do {
   if ($138) {
    $L$0106 = $137;
    while(1) {
     $139 = HEAP32[$z>>2]|0; //@line 2614 "./libraries/stb/stb_image.h"
     $140 = ((($139)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
     $141 = HEAP32[$140>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
     $142 = ((($139)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
     $143 = HEAP32[$142>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
     $144 = ($141>>>0)<($143>>>0); //@line 1229 "./libraries/stb/stb_image.h"
     if ($144) {
      $145 = ((($141)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
      HEAP32[$140>>2] = $145; //@line 1230 "./libraries/stb/stb_image.h"
      $146 = HEAP8[$141>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
      $$0$i = $146;
     } else {
      $147 = ((($139)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
      $148 = HEAP32[$147>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
      $149 = ($148|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
      if ($149) {
       $$0$i = 0;
      } else {
       $150 = ((($139)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
       $151 = HEAP32[$150>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $152 = ((($139)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
       $153 = HEAP32[$152>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $154 = ((($139)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
       $155 = ((($139)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
       $156 = HEAP32[$155>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $157 = (FUNCTION_TABLE_iiii[$151 & 8191]($153,$154,$156)|0); //@line 1213 "./libraries/stb/stb_image.h"
       $158 = ($157|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
       if ($158) {
        HEAP32[$147>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
        $159 = ((($139)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
        HEAP8[$154>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
        $$pre$phi$i38Z2D = $159;$$sink$i$i = $159;$437 = 0;
       } else {
        $160 = (((($139)) + 40|0) + ($157)|0); //@line 1223 "./libraries/stb/stb_image.h"
        $$pre$i37 = HEAP8[$154>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
        $$pre1$i = ((($139)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
        $$pre$phi$i38Z2D = $$pre1$i;$$sink$i$i = $160;$437 = $$pre$i37;
       }
       HEAP32[$142>>2] = $$sink$i$i; //@line 1223 "./libraries/stb/stb_image.h"
       HEAP32[$140>>2] = $$pre$phi$i38Z2D; //@line 1233 "./libraries/stb/stb_image.h"
       $$0$i = $437;
      }
     }
     $161 = $$0$i&255; //@line 2614 "./libraries/stb/stb_image.h"
     $162 = $161 & 15; //@line 2616 "./libraries/stb/stb_image.h"
     $$mask = $161 & 240; //@line 2617 "./libraries/stb/stb_image.h"
     $163 = ($$mask|0)==(0); //@line 2617 "./libraries/stb/stb_image.h"
     if (!($163)) {
      label = 57;
      break;
     }
     $164 = ($162>>>0)>(3); //@line 2618 "./libraries/stb/stb_image.h"
     if ($164) {
      label = 59;
      break;
     } else {
      $i$0105 = 0;
     }
     while(1) {
      $165 = HEAP32[$z>>2]|0; //@line 2620 "./libraries/stb/stb_image.h"
      $166 = ((($165)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
      $167 = HEAP32[$166>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
      $168 = ((($165)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
      $169 = HEAP32[$168>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
      $170 = ($167>>>0)<($169>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      if ($170) {
       $171 = ((($167)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
       HEAP32[$166>>2] = $171; //@line 1230 "./libraries/stb/stb_image.h"
       $172 = HEAP8[$167>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
       $$0$i44 = $172;
      } else {
       $173 = ((($165)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
       $174 = HEAP32[$173>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
       $175 = ($174|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
       if ($175) {
        $$0$i44 = 0;
       } else {
        $176 = ((($165)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $177 = HEAP32[$176>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $178 = ((($165)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $179 = HEAP32[$178>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $180 = ((($165)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $181 = ((($165)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $182 = HEAP32[$181>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $183 = (FUNCTION_TABLE_iiii[$177 & 8191]($179,$180,$182)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $184 = ($183|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($184) {
         HEAP32[$173>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $185 = ((($165)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$180>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i41Z2D = $185;$$sink$i$i42 = $185;$438 = 0;
        } else {
         $186 = (((($165)) + 40|0) + ($183)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i39 = HEAP8[$180>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i40 = ((($165)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i41Z2D = $$pre1$i40;$$sink$i$i42 = $186;$438 = $$pre$i39;
        }
        HEAP32[$168>>2] = $$sink$i$i42; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$166>>2] = $$pre$phi$i41Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i44 = $438;
       }
      }
      $187 = (440770 + ($i$0105)|0); //@line 2620 "./libraries/stb/stb_image.h"
      $188 = HEAP8[$187>>0]|0; //@line 2620 "./libraries/stb/stb_image.h"
      $189 = $188&255; //@line 2620 "./libraries/stb/stb_image.h"
      $190 = ((((($z)) + 13444|0) + ($162<<6)|0) + ($189)|0); //@line 2620 "./libraries/stb/stb_image.h"
      HEAP8[$190>>0] = $$0$i44; //@line 2620 "./libraries/stb/stb_image.h"
      $191 = (($i$0105) + 1)|0; //@line 2619 "./libraries/stb/stb_image.h"
      $exitcond = ($191|0)==(64); //@line 2619 "./libraries/stb/stb_image.h"
      if ($exitcond) {
       break;
      } else {
       $i$0105 = $191;
      }
     }
     $192 = (($L$0106) + -65)|0; //@line 2621 "./libraries/stb/stb_image.h"
     $193 = ($L$0106|0)>(65); //@line 2613 "./libraries/stb/stb_image.h"
     if ($193) {
      $L$0106 = $192;
     } else {
      $L$0$lcssa = $192;
      break L97;
     }
    }
    if ((label|0) == 57) {
     HEAP32[158754] = 440743; //@line 894 "./libraries/stb/stb_image.h"
     $$4$shrunk = 0;
     $$4 = $$4$shrunk&1;
     STACKTOP = sp;return ($$4|0); //@line 2660 "./libraries/stb/stb_image.h"
    }
    else if ((label|0) == 59) {
     HEAP32[158754] = 440756; //@line 894 "./libraries/stb/stb_image.h"
     $$4$shrunk = 0;
     $$4 = $$4$shrunk&1;
     STACKTOP = sp;return ($$4|0); //@line 2660 "./libraries/stb/stb_image.h"
    }
   } else {
    $L$0$lcssa = $137;
   }
  } while(0);
  $194 = ($L$0$lcssa|0)==(0); //@line 2623 "./libraries/stb/stb_image.h"
  $$4$shrunk = $194;
  $$4 = $$4$shrunk&1;
  STACKTOP = sp;return ($$4|0); //@line 2660 "./libraries/stb/stb_image.h"
  break;
 }
 case 196:  {
  $195 = HEAP32[$z>>2]|0; //@line 2626 "./libraries/stb/stb_image.h"
  $196 = ((($195)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
  $197 = HEAP32[$196>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
  $198 = ((($195)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
  $199 = HEAP32[$198>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
  $200 = ($197>>>0)<($199>>>0); //@line 1229 "./libraries/stb/stb_image.h"
  if ($200) {
   $201 = ((($197)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
   HEAP32[$196>>2] = $201; //@line 1230 "./libraries/stb/stb_image.h"
   $202 = HEAP8[$197>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
   $$0$i$i51 = $202;$220 = $201;$221 = $199;
  } else {
   $203 = ((($195)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
   $204 = HEAP32[$203>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
   $205 = ($204|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
   if ($205) {
    $$0$i$i51 = 0;$220 = $197;$221 = $199;
   } else {
    $206 = ((($195)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
    $207 = HEAP32[$206>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $208 = ((($195)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
    $209 = HEAP32[$208>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $210 = ((($195)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
    $211 = ((($195)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
    $212 = HEAP32[$211>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $213 = (FUNCTION_TABLE_iiii[$207 & 8191]($209,$210,$212)|0); //@line 1213 "./libraries/stb/stb_image.h"
    $214 = ($213|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
    if ($214) {
     HEAP32[$203>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
     $215 = ((($195)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
     HEAP8[$210>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
     $$pre$phi$i$i48Z2D = $215;$$sink$i$i$i49 = $215;$439 = 0;
    } else {
     $216 = (((($195)) + 40|0) + ($213)|0); //@line 1223 "./libraries/stb/stb_image.h"
     $$pre$i$i46 = HEAP8[$210>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
     $$pre1$i$i47 = ((($195)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
     $$pre$phi$i$i48Z2D = $$pre1$i$i47;$$sink$i$i$i49 = $216;$439 = $$pre$i$i46;
    }
    HEAP32[$198>>2] = $$sink$i$i$i49; //@line 1223 "./libraries/stb/stb_image.h"
    HEAP32[$196>>2] = $$pre$phi$i$i48Z2D; //@line 1233 "./libraries/stb/stb_image.h"
    $$0$i$i51 = $439;$220 = $$pre$phi$i$i48Z2D;$221 = $$sink$i$i$i49;
   }
  }
  $217 = $$0$i$i51&255; //@line 1293 "./libraries/stb/stb_image.h"
  $218 = $217 << 8; //@line 1294 "./libraries/stb/stb_image.h"
  $219 = ($220>>>0)<($221>>>0); //@line 1229 "./libraries/stb/stb_image.h"
  if ($219) {
   $222 = ((($220)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
   HEAP32[$196>>2] = $222; //@line 1230 "./libraries/stb/stb_image.h"
   $223 = HEAP8[$220>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
   $$0$i6$i58 = $223;
  } else {
   $224 = ((($195)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
   $225 = HEAP32[$224>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
   $226 = ($225|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
   if ($226) {
    $$0$i6$i58 = 0;
   } else {
    $227 = ((($195)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
    $228 = HEAP32[$227>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $229 = ((($195)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
    $230 = HEAP32[$229>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $231 = ((($195)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
    $232 = ((($195)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
    $233 = HEAP32[$232>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $234 = (FUNCTION_TABLE_iiii[$228 & 8191]($230,$231,$233)|0); //@line 1213 "./libraries/stb/stb_image.h"
    $235 = ($234|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
    if ($235) {
     HEAP32[$224>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
     $236 = ((($195)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
     HEAP8[$231>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
     $$pre$phi$i3$i55Z2D = $236;$$sink$i$i4$i56 = $236;$440 = 0;
    } else {
     $237 = (((($195)) + 40|0) + ($234)|0); //@line 1223 "./libraries/stb/stb_image.h"
     $$pre$i1$i53 = HEAP8[$231>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
     $$pre1$i2$i54 = ((($195)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
     $$pre$phi$i3$i55Z2D = $$pre1$i2$i54;$$sink$i$i4$i56 = $237;$440 = $$pre$i1$i53;
    }
    HEAP32[$198>>2] = $$sink$i$i4$i56; //@line 1223 "./libraries/stb/stb_image.h"
    HEAP32[$196>>2] = $$pre$phi$i3$i55Z2D; //@line 1233 "./libraries/stb/stb_image.h"
    $$0$i6$i58 = $440;
   }
  }
  $238 = $$0$i6$i58&255; //@line 1294 "./libraries/stb/stb_image.h"
  $239 = $238 | $218; //@line 1294 "./libraries/stb/stb_image.h"
  $240 = (($239) + -2)|0; //@line 2626 "./libraries/stb/stb_image.h"
  $241 = ($239>>>0)>(2); //@line 2627 "./libraries/stb/stb_image.h"
  L148: do {
   if ($241) {
    $L$2113 = $240;
    while(1) {
     $242 = HEAP32[$z>>2]|0; //@line 2630 "./libraries/stb/stb_image.h"
     $243 = ((($242)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
     $244 = HEAP32[$243>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
     $245 = ((($242)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
     $246 = HEAP32[$245>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
     $247 = ($244>>>0)<($246>>>0); //@line 1229 "./libraries/stb/stb_image.h"
     if ($247) {
      $248 = ((($244)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
      HEAP32[$243>>2] = $248; //@line 1230 "./libraries/stb/stb_image.h"
      $249 = HEAP8[$244>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
      $$0$i65 = $249;
     } else {
      $250 = ((($242)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
      $251 = HEAP32[$250>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
      $252 = ($251|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
      if ($252) {
       $$0$i65 = 0;
      } else {
       $253 = ((($242)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
       $254 = HEAP32[$253>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $255 = ((($242)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
       $256 = HEAP32[$255>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $257 = ((($242)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
       $258 = ((($242)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
       $259 = HEAP32[$258>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $260 = (FUNCTION_TABLE_iiii[$254 & 8191]($256,$257,$259)|0); //@line 1213 "./libraries/stb/stb_image.h"
       $261 = ($260|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
       if ($261) {
        HEAP32[$250>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
        $262 = ((($242)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
        HEAP8[$257>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
        $$pre$phi$i62Z2D = $262;$$sink$i$i63 = $262;$441 = 0;
       } else {
        $263 = (((($242)) + 40|0) + ($260)|0); //@line 1223 "./libraries/stb/stb_image.h"
        $$pre$i60 = HEAP8[$257>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
        $$pre1$i61 = ((($242)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
        $$pre$phi$i62Z2D = $$pre1$i61;$$sink$i$i63 = $263;$441 = $$pre$i60;
       }
       HEAP32[$245>>2] = $$sink$i$i63; //@line 1223 "./libraries/stb/stb_image.h"
       HEAP32[$243>>2] = $$pre$phi$i62Z2D; //@line 1233 "./libraries/stb/stb_image.h"
       $$0$i65 = $441;
      }
     }
     $264 = $$0$i65&255; //@line 2630 "./libraries/stb/stb_image.h"
     $265 = $264 & 15; //@line 2632 "./libraries/stb/stb_image.h"
     $266 = ($$0$i65&255)>(31); //@line 2633 "./libraries/stb/stb_image.h"
     $267 = ($265>>>0)>(3); //@line 2633 "./libraries/stb/stb_image.h"
     $or$cond = $266 | $267; //@line 2633 "./libraries/stb/stb_image.h"
     if ($or$cond) {
      label = 94;
      break;
     } else {
      $i1$0107 = 0;$n$0108 = 0;
     }
     while(1) {
      $268 = HEAP32[$z>>2]|0; //@line 2635 "./libraries/stb/stb_image.h"
      $269 = ((($268)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
      $270 = HEAP32[$269>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
      $271 = ((($268)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
      $272 = HEAP32[$271>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
      $273 = ($270>>>0)<($272>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      if ($273) {
       $274 = ((($270)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
       HEAP32[$269>>2] = $274; //@line 1230 "./libraries/stb/stb_image.h"
       $275 = HEAP8[$270>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
       $$0$i72 = $275;
      } else {
       $276 = ((($268)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
       $277 = HEAP32[$276>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
       $278 = ($277|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
       if ($278) {
        $$0$i72 = 0;
       } else {
        $279 = ((($268)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $280 = HEAP32[$279>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $281 = ((($268)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $282 = HEAP32[$281>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $283 = ((($268)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $284 = ((($268)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $285 = HEAP32[$284>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $286 = (FUNCTION_TABLE_iiii[$280 & 8191]($282,$283,$285)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $287 = ($286|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($287) {
         HEAP32[$276>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $288 = ((($268)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$283>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i69Z2D = $288;$$sink$i$i70 = $288;$442 = 0;
        } else {
         $289 = (((($268)) + 40|0) + ($286)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i67 = HEAP8[$283>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i68 = ((($268)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i69Z2D = $$pre1$i68;$$sink$i$i70 = $289;$442 = $$pre$i67;
        }
        HEAP32[$271>>2] = $$sink$i$i70; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$269>>2] = $$pre$phi$i69Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i72 = $442;
       }
      }
      $290 = $$0$i72&255; //@line 2635 "./libraries/stb/stb_image.h"
      $291 = (($sizes) + ($i1$0107<<2)|0); //@line 2635 "./libraries/stb/stb_image.h"
      HEAP32[$291>>2] = $290; //@line 2635 "./libraries/stb/stb_image.h"
      $292 = (($290) + ($n$0108))|0; //@line 2636 "./libraries/stb/stb_image.h"
      $293 = (($i1$0107) + 1)|0; //@line 2634 "./libraries/stb/stb_image.h"
      $exitcond119 = ($293|0)==(16); //@line 2634 "./libraries/stb/stb_image.h"
      if ($exitcond119) {
       $$lcssa131 = $292;
       break;
      } else {
       $i1$0107 = $293;$n$0108 = $292;
      }
     }
     $294 = (($L$2113) + -17)|0; //@line 2638 "./libraries/stb/stb_image.h"
     $$mask97 = $264 & 240; //@line 2639 "./libraries/stb/stb_image.h"
     $295 = ($$mask97|0)==(0); //@line 2639 "./libraries/stb/stb_image.h"
     if ($295) {
      $296 = (((($z)) + 4|0) + (($265*1680)|0)|0); //@line 2640 "./libraries/stb/stb_image.h"
      $297 = (__ZL19stbi__build_huffmanP13stbi__huffmanPi($296,$sizes)|0); //@line 2640 "./libraries/stb/stb_image.h"
      $298 = ($297|0)==(0); //@line 2640 "./libraries/stb/stb_image.h"
      if ($298) {
       break;
      }
      $299 = (((((($z)) + 4|0) + (($265*1680)|0)|0)) + 1024|0); //@line 2641 "./libraries/stb/stb_image.h"
      $v$0 = $299;
     } else {
      $300 = (((($z)) + 6724|0) + (($265*1680)|0)|0); //@line 2643 "./libraries/stb/stb_image.h"
      $301 = (__ZL19stbi__build_huffmanP13stbi__huffmanPi($300,$sizes)|0); //@line 2643 "./libraries/stb/stb_image.h"
      $302 = ($301|0)==(0); //@line 2643 "./libraries/stb/stb_image.h"
      if ($302) {
       break;
      }
      $303 = (((((($z)) + 6724|0) + (($265*1680)|0)|0)) + 1024|0); //@line 2644 "./libraries/stb/stb_image.h"
      $v$0 = $303;
     }
     $304 = ($$lcssa131|0)>(0); //@line 2646 "./libraries/stb/stb_image.h"
     if ($304) {
      $i1$1109 = 0;
      while(1) {
       $305 = HEAP32[$z>>2]|0; //@line 2647 "./libraries/stb/stb_image.h"
       $306 = ((($305)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
       $307 = HEAP32[$306>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
       $308 = ((($305)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
       $309 = HEAP32[$308>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
       $310 = ($307>>>0)<($309>>>0); //@line 1229 "./libraries/stb/stb_image.h"
       if ($310) {
        $311 = ((($307)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$306>>2] = $311; //@line 1230 "./libraries/stb/stb_image.h"
        $312 = HEAP8[$307>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i79 = $312;
       } else {
        $313 = ((($305)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $314 = HEAP32[$313>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $315 = ($314|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($315) {
         $$0$i79 = 0;
        } else {
         $316 = ((($305)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
         $317 = HEAP32[$316>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $318 = ((($305)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
         $319 = HEAP32[$318>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $320 = ((($305)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
         $321 = ((($305)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
         $322 = HEAP32[$321>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $323 = (FUNCTION_TABLE_iiii[$317 & 8191]($319,$320,$322)|0); //@line 1213 "./libraries/stb/stb_image.h"
         $324 = ($323|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
         if ($324) {
          HEAP32[$313>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
          $325 = ((($305)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
          HEAP8[$320>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
          $$pre$phi$i76Z2D = $325;$$sink$i$i77 = $325;$443 = 0;
         } else {
          $326 = (((($305)) + 40|0) + ($323)|0); //@line 1223 "./libraries/stb/stb_image.h"
          $$pre$i74 = HEAP8[$320>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
          $$pre1$i75 = ((($305)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
          $$pre$phi$i76Z2D = $$pre1$i75;$$sink$i$i77 = $326;$443 = $$pre$i74;
         }
         HEAP32[$308>>2] = $$sink$i$i77; //@line 1223 "./libraries/stb/stb_image.h"
         HEAP32[$306>>2] = $$pre$phi$i76Z2D; //@line 1233 "./libraries/stb/stb_image.h"
         $$0$i79 = $443;
        }
       }
       $327 = (($v$0) + ($i1$1109)|0); //@line 2647 "./libraries/stb/stb_image.h"
       HEAP8[$327>>0] = $$0$i79; //@line 2647 "./libraries/stb/stb_image.h"
       $328 = (($i1$1109) + 1)|0; //@line 2646 "./libraries/stb/stb_image.h"
       $exitcond120 = ($328|0)==($$lcssa131|0); //@line 2646 "./libraries/stb/stb_image.h"
       if ($exitcond120) {
        break;
       } else {
        $i1$1109 = $328;
       }
      }
     }
     if (!($295)) {
      $i$02$i = 0;
      while(1) {
       $329 = ((((($z)) + 6724|0) + (($265*1680)|0)|0) + ($i$02$i)|0); //@line 1566 "./libraries/stb/stb_image.h"
       $330 = HEAP8[$329>>0]|0; //@line 1566 "./libraries/stb/stb_image.h"
       $331 = ((((($z)) + 13700|0) + ($265<<10)|0) + ($i$02$i<<1)|0); //@line 1567 "./libraries/stb/stb_image.h"
       HEAP16[$331>>1] = 0; //@line 1567 "./libraries/stb/stb_image.h"
       $332 = $330&255; //@line 1568 "./libraries/stb/stb_image.h"
       $333 = ($330<<24>>24)==(-1); //@line 1568 "./libraries/stb/stb_image.h"
       if (!($333)) {
        $334 = ((((((($z)) + 6724|0) + (($265*1680)|0)|0)) + 1024|0) + ($332)|0); //@line 1569 "./libraries/stb/stb_image.h"
        $335 = HEAP8[$334>>0]|0; //@line 1569 "./libraries/stb/stb_image.h"
        $336 = $335&255; //@line 1569 "./libraries/stb/stb_image.h"
        $337 = $336 & 240; //@line 1581 "./libraries/stb/stb_image.h"
        $338 = $336 & 15; //@line 1571 "./libraries/stb/stb_image.h"
        $339 = ((((((($z)) + 6724|0) + (($265*1680)|0)|0)) + 1280|0) + ($332)|0); //@line 1572 "./libraries/stb/stb_image.h"
        $340 = HEAP8[$339>>0]|0; //@line 1572 "./libraries/stb/stb_image.h"
        $341 = $340&255; //@line 1572 "./libraries/stb/stb_image.h"
        $342 = ($338|0)==(0); //@line 1574 "./libraries/stb/stb_image.h"
        if (!($342)) {
         $343 = (($341) + ($338))|0; //@line 1574 "./libraries/stb/stb_image.h"
         $344 = ($343>>>0)<(10); //@line 1574 "./libraries/stb/stb_image.h"
         if ($344) {
          $345 = $i$02$i << $341; //@line 1576 "./libraries/stb/stb_image.h"
          $346 = $345 & 511; //@line 1576 "./libraries/stb/stb_image.h"
          $347 = (9 - ($338))|0; //@line 1576 "./libraries/stb/stb_image.h"
          $348 = $346 >>> $347; //@line 1576 "./libraries/stb/stb_image.h"
          $349 = (($338) + -1)|0; //@line 1577 "./libraries/stb/stb_image.h"
          $350 = 1 << $349; //@line 1577 "./libraries/stb/stb_image.h"
          $351 = ($348|0)<($350|0); //@line 1578 "./libraries/stb/stb_image.h"
          if ($351) {
           $352 = -1 << $338; //@line 1578 "./libraries/stb/stb_image.h"
           $353 = (($352) + 1)|0; //@line 1578 "./libraries/stb/stb_image.h"
           $354 = (($353) + ($348))|0; //@line 1578 "./libraries/stb/stb_image.h"
           $k$0$i = $354;
          } else {
           $k$0$i = $348;
          }
          $k$0$off$i = (($k$0$i) + 128)|0; //@line 1580 "./libraries/stb/stb_image.h"
          $355 = ($k$0$off$i>>>0)<(256); //@line 1580 "./libraries/stb/stb_image.h"
          if ($355) {
           $356 = $k$0$i << 8; //@line 1581 "./libraries/stb/stb_image.h"
           $357 = $356 | $337; //@line 1581 "./libraries/stb/stb_image.h"
           $358 = (($357) + ($343))|0; //@line 1581 "./libraries/stb/stb_image.h"
           $359 = $358&65535; //@line 1581 "./libraries/stb/stb_image.h"
           HEAP16[$331>>1] = $359; //@line 1581 "./libraries/stb/stb_image.h"
          }
         }
        }
       }
       $360 = (($i$02$i) + 1)|0; //@line 1565 "./libraries/stb/stb_image.h"
       $exitcond$i = ($360|0)==(512); //@line 1565 "./libraries/stb/stb_image.h"
       if ($exitcond$i) {
        break;
       } else {
        $i$02$i = $360;
       }
      }
     }
     $361 = (($294) - ($$lcssa131))|0; //@line 2650 "./libraries/stb/stb_image.h"
     $362 = ($361|0)>(0); //@line 2627 "./libraries/stb/stb_image.h"
     if ($362) {
      $L$2113 = $361;
     } else {
      $L$2$lcssa = $361;
      break L148;
     }
    }
    if ((label|0) == 94) {
     HEAP32[158754] = 440849; //@line 894 "./libraries/stb/stb_image.h"
    }
    $$4$shrunk = 0;
    $$4 = $$4$shrunk&1;
    STACKTOP = sp;return ($$4|0); //@line 2660 "./libraries/stb/stb_image.h"
   } else {
    $L$2$lcssa = $240;
   }
  } while(0);
  $363 = ($L$2$lcssa|0)==(0); //@line 2652 "./libraries/stb/stb_image.h"
  $$4$shrunk = $363;
  $$4 = $$4$shrunk&1;
  STACKTOP = sp;return ($$4|0); //@line 2660 "./libraries/stb/stb_image.h"
  break;
 }
 default: {
  $364 = $m & -16; //@line 2655 "./libraries/stb/stb_image.h"
  $365 = ($364|0)==(224); //@line 2655 "./libraries/stb/stb_image.h"
  $366 = ($m|0)==(254); //@line 2655 "./libraries/stb/stb_image.h"
  $or$cond6 = $366 | $365; //@line 2655 "./libraries/stb/stb_image.h"
  if (!($or$cond6)) {
   $$4$shrunk = 0;
   $$4 = $$4$shrunk&1;
   STACKTOP = sp;return ($$4|0); //@line 2660 "./libraries/stb/stb_image.h"
  }
  $367 = HEAP32[$z>>2]|0; //@line 2656 "./libraries/stb/stb_image.h"
  $368 = ((($367)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
  $369 = HEAP32[$368>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
  $370 = ((($367)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
  $371 = HEAP32[$370>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
  $372 = ($369>>>0)<($371>>>0); //@line 1229 "./libraries/stb/stb_image.h"
  $373 = $371; //@line 1229 "./libraries/stb/stb_image.h"
  if ($372) {
   $374 = ((($369)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
   HEAP32[$368>>2] = $374; //@line 1230 "./libraries/stb/stb_image.h"
   $375 = HEAP8[$369>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
   $$0$i$i86 = $375;$$in = $374;$394 = $371;$444 = $373;
  } else {
   $376 = ((($367)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
   $377 = HEAP32[$376>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
   $378 = ($377|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
   if ($378) {
    $$0$i$i86 = 0;$$in = $369;$394 = $371;$444 = $373;
   } else {
    $379 = ((($367)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
    $380 = HEAP32[$379>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $381 = ((($367)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
    $382 = HEAP32[$381>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $383 = ((($367)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
    $384 = ((($367)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
    $385 = HEAP32[$384>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $386 = (FUNCTION_TABLE_iiii[$380 & 8191]($382,$383,$385)|0); //@line 1213 "./libraries/stb/stb_image.h"
    $387 = ($386|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
    if ($387) {
     HEAP32[$376>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
     $388 = ((($367)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
     HEAP8[$383>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
     $$pre$phi$i$i83Z2D = $388;$$sink$i$i$i84 = $388;$445 = 0;
    } else {
     $389 = (((($367)) + 40|0) + ($386)|0); //@line 1223 "./libraries/stb/stb_image.h"
     $$pre$i$i81 = HEAP8[$383>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
     $$pre1$i$i82 = ((($367)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
     $$pre$phi$i$i83Z2D = $$pre1$i$i82;$$sink$i$i$i84 = $389;$445 = $$pre$i$i81;
    }
    HEAP32[$370>>2] = $$sink$i$i$i84; //@line 1223 "./libraries/stb/stb_image.h"
    HEAP32[$368>>2] = $$pre$phi$i$i83Z2D; //@line 1233 "./libraries/stb/stb_image.h"
    $390 = $$sink$i$i$i84; //@line 1233 "./libraries/stb/stb_image.h"
    $$0$i$i86 = $445;$$in = $$pre$phi$i$i83Z2D;$394 = $$sink$i$i$i84;$444 = $390;
   }
  }
  $391 = $$0$i$i86&255; //@line 1293 "./libraries/stb/stb_image.h"
  $392 = $391 << 8; //@line 1294 "./libraries/stb/stb_image.h"
  $393 = ($$in>>>0)<($394>>>0); //@line 1229 "./libraries/stb/stb_image.h"
  if ($393) {
   $395 = ((($$in)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
   HEAP32[$368>>2] = $395; //@line 1230 "./libraries/stb/stb_image.h"
   $396 = HEAP8[$$in>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
   $$0$i6$i93 = $396;$$pre$i = $395;$417 = $444;$422 = $394;
  } else {
   $397 = ((($367)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
   $398 = HEAP32[$397>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
   $399 = ($398|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
   if ($399) {
    $$0$i6$i93 = 0;$$pre$i = $$in;$417 = $444;$422 = $394;
   } else {
    $400 = ((($367)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
    $401 = HEAP32[$400>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $402 = ((($367)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
    $403 = HEAP32[$402>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $404 = ((($367)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
    $405 = ((($367)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
    $406 = HEAP32[$405>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $407 = (FUNCTION_TABLE_iiii[$401 & 8191]($403,$404,$406)|0); //@line 1213 "./libraries/stb/stb_image.h"
    $408 = ($407|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
    if ($408) {
     HEAP32[$397>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
     $409 = ((($367)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
     HEAP8[$404>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
     $$pre$phi$i3$i90Z2D = $409;$$sink$i$i4$i91 = $409;$446 = 0;
    } else {
     $410 = (((($367)) + 40|0) + ($407)|0); //@line 1223 "./libraries/stb/stb_image.h"
     $$pre$i1$i88 = HEAP8[$404>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
     $$pre1$i2$i89 = ((($367)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
     $$pre$phi$i3$i90Z2D = $$pre1$i2$i89;$$sink$i$i4$i91 = $410;$446 = $$pre$i1$i88;
    }
    HEAP32[$370>>2] = $$sink$i$i4$i91; //@line 1223 "./libraries/stb/stb_image.h"
    HEAP32[$368>>2] = $$pre$phi$i3$i90Z2D; //@line 1233 "./libraries/stb/stb_image.h"
    $411 = $$sink$i$i4$i91; //@line 1233 "./libraries/stb/stb_image.h"
    $$0$i6$i93 = $446;$$pre$i = $$pre$phi$i3$i90Z2D;$417 = $411;$422 = $$sink$i$i4$i91;
   }
  }
  $412 = $$pre$i; //@line 1257 "./libraries/stb/stb_image.h"
  $413 = $$0$i6$i93&255; //@line 1294 "./libraries/stb/stb_image.h"
  $414 = $413 | $392; //@line 1294 "./libraries/stb/stb_image.h"
  $415 = (($414) + -2)|0; //@line 2656 "./libraries/stb/stb_image.h"
  $416 = ($414>>>0)<(2); //@line 1252 "./libraries/stb/stb_image.h"
  if ($416) {
   HEAP32[$368>>2] = $417; //@line 1253 "./libraries/stb/stb_image.h"
   $$4$shrunk = 1;
   $$4 = $$4$shrunk&1;
   STACKTOP = sp;return ($$4|0); //@line 2660 "./libraries/stb/stb_image.h"
  }
  $418 = ((($367)) + 16|0); //@line 1256 "./libraries/stb/stb_image.h"
  $419 = HEAP32[$418>>2]|0; //@line 1256 "./libraries/stb/stb_image.h"
  $420 = ($419|0)==(0|0); //@line 1256 "./libraries/stb/stb_image.h"
  if (!($420)) {
   $421 = $422; //@line 1257 "./libraries/stb/stb_image.h"
   $423 = (($421) - ($412))|0; //@line 1257 "./libraries/stb/stb_image.h"
   $424 = ($415|0)>($423|0); //@line 1258 "./libraries/stb/stb_image.h"
   if ($424) {
    HEAP32[$368>>2] = $422; //@line 1259 "./libraries/stb/stb_image.h"
    $425 = ((($367)) + 20|0); //@line 1260 "./libraries/stb/stb_image.h"
    $426 = HEAP32[$425>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
    $427 = ((($367)) + 28|0); //@line 1260 "./libraries/stb/stb_image.h"
    $428 = HEAP32[$427>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
    $429 = (($415) - ($423))|0; //@line 1260 "./libraries/stb/stb_image.h"
    FUNCTION_TABLE_vii[$426 & 8191]($428,$429); //@line 1260 "./libraries/stb/stb_image.h"
    $$4$shrunk = 1;
    $$4 = $$4$shrunk&1;
    STACKTOP = sp;return ($$4|0); //@line 2660 "./libraries/stb/stb_image.h"
   }
  }
  $430 = (($$pre$i) + ($415)|0); //@line 1264 "./libraries/stb/stb_image.h"
  HEAP32[$368>>2] = $430; //@line 1264 "./libraries/stb/stb_image.h"
  $$4$shrunk = 1;
  $$4 = $$4$shrunk&1;
  STACKTOP = sp;return ($$4|0); //@line 2660 "./libraries/stb/stb_image.h"
 }
 }
 return (0)|0;
}
function __ZL19stbi__build_huffmanP13stbi__huffmanPi($h,$count) {
 $h = $h|0;
 $count = $count|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa36 = 0, $$lcssa37 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $code$014 = 0, $code$110 = 0, $code$110$lcssa = 0, $code$2 = 0, $exitcond = 0, $exitcond26 = 0, $i$022 = 0;
 var $i$17 = 0, $j$017 = 0, $j$115 = 0, $k$021 = 0, $k$1$lcssa = 0, $k$1$lcssa$lcssa = 0, $k$116 = 0, $k$213 = 0, $k$39 = 0, $k$4 = 0, $k$4$lcssa = 0, $scevgep = 0, $smax = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $i$022 = 0;$k$021 = 0;
 while(1) {
  $1 = (($count) + ($i$022<<2)|0); //@line 1524 "./libraries/stb/stb_image.h"
  $2 = HEAP32[$1>>2]|0; //@line 1524 "./libraries/stb/stb_image.h"
  $3 = ($2|0)>(0); //@line 1524 "./libraries/stb/stb_image.h"
  $0 = (($i$022) + 1)|0; //@line 1525 "./libraries/stb/stb_image.h"
  if ($3) {
   $4 = $0&255; //@line 1525 "./libraries/stb/stb_image.h"
   $j$017 = 0;$k$116 = $k$021;
   while(1) {
    $5 = (($k$116) + 1)|0; //@line 1525 "./libraries/stb/stb_image.h"
    $6 = (((($h)) + 1280|0) + ($k$116)|0); //@line 1525 "./libraries/stb/stb_image.h"
    HEAP8[$6>>0] = $4; //@line 1525 "./libraries/stb/stb_image.h"
    $7 = (($j$017) + 1)|0; //@line 1524 "./libraries/stb/stb_image.h"
    $8 = HEAP32[$1>>2]|0; //@line 1524 "./libraries/stb/stb_image.h"
    $9 = ($7|0)<($8|0); //@line 1524 "./libraries/stb/stb_image.h"
    if ($9) {
     $j$017 = $7;$k$116 = $5;
    } else {
     $k$1$lcssa = $5;
     break;
    }
   }
  } else {
   $k$1$lcssa = $k$021;
  }
  $exitcond26 = ($0|0)==(16); //@line 1523 "./libraries/stb/stb_image.h"
  if ($exitcond26) {
   $k$1$lcssa$lcssa = $k$1$lcssa;
   break;
  } else {
   $i$022 = $0;$k$021 = $k$1$lcssa;
  }
 }
 $10 = (((($h)) + 1280|0) + ($k$1$lcssa$lcssa)|0); //@line 1526 "./libraries/stb/stb_image.h"
 HEAP8[$10>>0] = 0; //@line 1526 "./libraries/stb/stb_image.h"
 $code$014 = 0;$j$115 = 1;$k$213 = 0;
 while(1) {
  $11 = (($k$213) - ($code$014))|0; //@line 1533 "./libraries/stb/stb_image.h"
  $12 = (((($h)) + 1612|0) + ($j$115<<2)|0); //@line 1533 "./libraries/stb/stb_image.h"
  HEAP32[$12>>2] = $11; //@line 1533 "./libraries/stb/stb_image.h"
  $13 = (((($h)) + 1280|0) + ($k$213)|0); //@line 1534 "./libraries/stb/stb_image.h"
  $14 = HEAP8[$13>>0]|0; //@line 1535 "./libraries/stb/stb_image.h"
  $15 = $14&255; //@line 1534 "./libraries/stb/stb_image.h"
  $16 = ($15|0)==($j$115|0); //@line 1534 "./libraries/stb/stb_image.h"
  if ($16) {
   $code$110 = $code$014;$k$39 = $k$213;
   while(1) {
    $17 = (($code$110) + 1)|0; //@line 1536 "./libraries/stb/stb_image.h"
    $18 = $code$110&65535; //@line 1536 "./libraries/stb/stb_image.h"
    $19 = (($k$39) + 1)|0; //@line 1536 "./libraries/stb/stb_image.h"
    $20 = (((($h)) + 512|0) + ($k$39<<1)|0); //@line 1536 "./libraries/stb/stb_image.h"
    HEAP16[$20>>1] = $18; //@line 1536 "./libraries/stb/stb_image.h"
    $21 = (((($h)) + 1280|0) + ($19)|0); //@line 1535 "./libraries/stb/stb_image.h"
    $22 = HEAP8[$21>>0]|0; //@line 1535 "./libraries/stb/stb_image.h"
    $23 = $22&255; //@line 1535 "./libraries/stb/stb_image.h"
    $24 = ($23|0)==($j$115|0); //@line 1535 "./libraries/stb/stb_image.h"
    if ($24) {
     $code$110 = $17;$k$39 = $19;
    } else {
     $$lcssa = $17;$$lcssa36 = $19;$code$110$lcssa = $code$110;
     break;
    }
   }
   $25 = 1 << $j$115; //@line 1537 "./libraries/stb/stb_image.h"
   $26 = ($code$110$lcssa|0)<($25|0); //@line 1537 "./libraries/stb/stb_image.h"
   if ($26) {
    $code$2 = $$lcssa;$k$4 = $$lcssa36;
   } else {
    label = 10;
    break;
   }
  } else {
   $code$2 = $code$014;$k$4 = $k$213;
  }
  $27 = (16 - ($j$115))|0; //@line 1540 "./libraries/stb/stb_image.h"
  $28 = $code$2 << $27; //@line 1540 "./libraries/stb/stb_image.h"
  $29 = (((($h)) + 1540|0) + ($j$115<<2)|0); //@line 1540 "./libraries/stb/stb_image.h"
  HEAP32[$29>>2] = $28; //@line 1540 "./libraries/stb/stb_image.h"
  $30 = $code$2 << 1; //@line 1541 "./libraries/stb/stb_image.h"
  $31 = (($j$115) + 1)|0; //@line 1531 "./libraries/stb/stb_image.h"
  $32 = ($31|0)<(17); //@line 1531 "./libraries/stb/stb_image.h"
  if ($32) {
   $code$014 = $30;$j$115 = $31;$k$213 = $k$4;
  } else {
   $$lcssa37 = $31;$k$4$lcssa = $k$4;
   break;
  }
 }
 if ((label|0) == 10) {
  HEAP32[158754] = 440864; //@line 894 "./libraries/stb/stb_image.h"
  $$0 = 0;
  return ($$0|0); //@line 1558 "./libraries/stb/stb_image.h"
 }
 $33 = (((($h)) + 1540|0) + ($$lcssa37<<2)|0); //@line 1543 "./libraries/stb/stb_image.h"
 HEAP32[$33>>2] = -1; //@line 1543 "./libraries/stb/stb_image.h"
 _memset(($h|0),-1,512)|0; //@line 1546 "./libraries/stb/stb_image.h"
 $34 = ($k$4$lcssa|0)>(0); //@line 1547 "./libraries/stb/stb_image.h"
 if ($34) {
  $i$17 = 0;
 } else {
  $$0 = 1;
  return ($$0|0); //@line 1558 "./libraries/stb/stb_image.h"
 }
 while(1) {
  $35 = (((($h)) + 1280|0) + ($i$17)|0); //@line 1548 "./libraries/stb/stb_image.h"
  $36 = HEAP8[$35>>0]|0; //@line 1548 "./libraries/stb/stb_image.h"
  $37 = ($36&255)<(10); //@line 1549 "./libraries/stb/stb_image.h"
  if ($37) {
   $38 = $36&255; //@line 1548 "./libraries/stb/stb_image.h"
   $39 = (9 - ($38))|0; //@line 1550 "./libraries/stb/stb_image.h"
   $40 = 1 << $39; //@line 1551 "./libraries/stb/stb_image.h"
   $41 = ($39|0)==(31); //@line 1552 "./libraries/stb/stb_image.h"
   if (!($41)) {
    $42 = (((($h)) + 512|0) + ($i$17<<1)|0); //@line 1550 "./libraries/stb/stb_image.h"
    $43 = HEAP16[$42>>1]|0; //@line 1550 "./libraries/stb/stb_image.h"
    $44 = $43&65535; //@line 1550 "./libraries/stb/stb_image.h"
    $45 = $44 << $39; //@line 1550 "./libraries/stb/stb_image.h"
    $46 = $i$17&255; //@line 1553 "./libraries/stb/stb_image.h"
    $scevgep = (($h) + ($45)|0); //@line 1552 "./libraries/stb/stb_image.h"
    $47 = ($40|0)>(1); //@line 1552 "./libraries/stb/stb_image.h"
    $smax = $47 ? $40 : 1; //@line 1552 "./libraries/stb/stb_image.h"
    _memset(($scevgep|0),($46|0),($smax|0))|0; //@line 1553 "./libraries/stb/stb_image.h"
   }
  }
  $48 = (($i$17) + 1)|0; //@line 1547 "./libraries/stb/stb_image.h"
  $exitcond = ($48|0)==($k$4$lcssa|0); //@line 1547 "./libraries/stb/stb_image.h"
  if ($exitcond) {
   $$0 = 1;
   break;
  } else {
   $i$17 = $48;
  }
 }
 return ($$0|0); //@line 1558 "./libraries/stb/stb_image.h"
}
function __ZL13stbi__get32leP13stbi__context($s) {
 $s = $s|0;
 var $$0$i$i = 0, $$0$i$i6 = 0, $$0$i6$i = 0, $$0$i6$i13 = 0, $$pre$i$i = 0, $$pre$i$i1 = 0, $$pre$i1$i = 0, $$pre$i1$i8 = 0, $$pre$phi$i$i3Z2D = 0, $$pre$phi$i$iZ2D = 0, $$pre$phi$i3$i10Z2D = 0, $$pre$phi$i3$iZ2D = 0, $$pre1$i$i = 0, $$pre1$i$i2 = 0, $$pre1$i2$i = 0, $$pre1$i2$i9 = 0, $$sink$i$i$i = 0, $$sink$i$i$i4 = 0, $$sink$i$i4$i = 0, $$sink$i$i4$i11 = 0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($s)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
 $1 = HEAP32[$0>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
 $2 = ((($s)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
 $3 = HEAP32[$2>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
 $4 = ($1>>>0)<($3>>>0); //@line 1229 "./libraries/stb/stb_image.h"
 if ($4) {
  $5 = ((($1)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
  HEAP32[$0>>2] = $5; //@line 1230 "./libraries/stb/stb_image.h"
  $6 = HEAP8[$1>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
  $$0$i$i = $6;$23 = $5;$24 = $3;
 } else {
  $7 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
  $8 = HEAP32[$7>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
  $9 = ($8|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
  if ($9) {
   $$0$i$i = 0;$23 = $1;$24 = $3;
  } else {
   $10 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
   $11 = HEAP32[$10>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $12 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
   $13 = HEAP32[$12>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $14 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
   $15 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
   $16 = HEAP32[$15>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $17 = (FUNCTION_TABLE_iiii[$11 & 8191]($13,$14,$16)|0); //@line 1213 "./libraries/stb/stb_image.h"
   $18 = ($17|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
   if ($18) {
    HEAP32[$7>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
    $19 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
    HEAP8[$14>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
    $$pre$phi$i$iZ2D = $19;$$sink$i$i$i = $19;$88 = 0;
   } else {
    $20 = (((($s)) + 40|0) + ($17)|0); //@line 1223 "./libraries/stb/stb_image.h"
    $$pre$i$i = HEAP8[$14>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
    $$pre1$i$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
    $$pre$phi$i$iZ2D = $$pre1$i$i;$$sink$i$i$i = $20;$88 = $$pre$i$i;
   }
   HEAP32[$2>>2] = $$sink$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
   HEAP32[$0>>2] = $$pre$phi$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
   $$0$i$i = $88;$23 = $$pre$phi$i$iZ2D;$24 = $$sink$i$i$i;
  }
 }
 $21 = $$0$i$i&255; //@line 1308 "./libraries/stb/stb_image.h"
 $22 = ($23>>>0)<($24>>>0); //@line 1229 "./libraries/stb/stb_image.h"
 if ($22) {
  $25 = ((($23)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
  HEAP32[$0>>2] = $25; //@line 1230 "./libraries/stb/stb_image.h"
  $26 = HEAP8[$23>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
  $$0$i6$i = $26;$45 = $25;$46 = $24;
 } else {
  $27 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
  $28 = HEAP32[$27>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
  $29 = ($28|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
  if ($29) {
   $$0$i6$i = 0;$45 = $23;$46 = $24;
  } else {
   $30 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
   $31 = HEAP32[$30>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $32 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
   $33 = HEAP32[$32>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $34 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
   $35 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
   $36 = HEAP32[$35>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $37 = (FUNCTION_TABLE_iiii[$31 & 8191]($33,$34,$36)|0); //@line 1213 "./libraries/stb/stb_image.h"
   $38 = ($37|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
   if ($38) {
    HEAP32[$27>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
    $39 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
    HEAP8[$34>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
    $$pre$phi$i3$iZ2D = $39;$$sink$i$i4$i = $39;$89 = 0;
   } else {
    $40 = (((($s)) + 40|0) + ($37)|0); //@line 1223 "./libraries/stb/stb_image.h"
    $$pre$i1$i = HEAP8[$34>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
    $$pre1$i2$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
    $$pre$phi$i3$iZ2D = $$pre1$i2$i;$$sink$i$i4$i = $40;$89 = $$pre$i1$i;
   }
   HEAP32[$2>>2] = $$sink$i$i4$i; //@line 1223 "./libraries/stb/stb_image.h"
   HEAP32[$0>>2] = $$pre$phi$i3$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
   $$0$i6$i = $89;$45 = $$pre$phi$i3$iZ2D;$46 = $$sink$i$i4$i;
  }
 }
 $41 = $$0$i6$i&255; //@line 1309 "./libraries/stb/stb_image.h"
 $42 = $41 << 8; //@line 1309 "./libraries/stb/stb_image.h"
 $43 = $42 | $21; //@line 1309 "./libraries/stb/stb_image.h"
 $44 = ($45>>>0)<($46>>>0); //@line 1229 "./libraries/stb/stb_image.h"
 if ($44) {
  $47 = ((($45)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
  HEAP32[$0>>2] = $47; //@line 1230 "./libraries/stb/stb_image.h"
  $48 = HEAP8[$45>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
  $$0$i$i6 = $48;$65 = $47;$66 = $46;
 } else {
  $49 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
  $50 = HEAP32[$49>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
  $51 = ($50|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
  if ($51) {
   $$0$i$i6 = 0;$65 = $45;$66 = $46;
  } else {
   $52 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
   $53 = HEAP32[$52>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $54 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
   $55 = HEAP32[$54>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $56 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
   $57 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
   $58 = HEAP32[$57>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $59 = (FUNCTION_TABLE_iiii[$53 & 8191]($55,$56,$58)|0); //@line 1213 "./libraries/stb/stb_image.h"
   $60 = ($59|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
   if ($60) {
    HEAP32[$49>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
    $61 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
    HEAP8[$56>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
    $$pre$phi$i$i3Z2D = $61;$$sink$i$i$i4 = $61;$90 = 0;
   } else {
    $62 = (((($s)) + 40|0) + ($59)|0); //@line 1223 "./libraries/stb/stb_image.h"
    $$pre$i$i1 = HEAP8[$56>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
    $$pre1$i$i2 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
    $$pre$phi$i$i3Z2D = $$pre1$i$i2;$$sink$i$i$i4 = $62;$90 = $$pre$i$i1;
   }
   HEAP32[$2>>2] = $$sink$i$i$i4; //@line 1223 "./libraries/stb/stb_image.h"
   HEAP32[$0>>2] = $$pre$phi$i$i3Z2D; //@line 1233 "./libraries/stb/stb_image.h"
   $$0$i$i6 = $90;$65 = $$pre$phi$i$i3Z2D;$66 = $$sink$i$i$i4;
  }
 }
 $63 = $$0$i$i6&255; //@line 1308 "./libraries/stb/stb_image.h"
 $64 = ($65>>>0)<($66>>>0); //@line 1229 "./libraries/stb/stb_image.h"
 if ($64) {
  $67 = ((($65)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
  HEAP32[$0>>2] = $67; //@line 1230 "./libraries/stb/stb_image.h"
  $68 = HEAP8[$65>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
  $$0$i6$i13 = $68;
  $83 = $$0$i6$i13&255; //@line 1309 "./libraries/stb/stb_image.h"
  $84 = $83 << 8; //@line 1309 "./libraries/stb/stb_image.h"
  $85 = $84 | $63; //@line 1309 "./libraries/stb/stb_image.h"
  $86 = $85 << 16; //@line 1317 "./libraries/stb/stb_image.h"
  $87 = $43 | $86; //@line 1317 "./libraries/stb/stb_image.h"
  return ($87|0); //@line 1317 "./libraries/stb/stb_image.h"
 }
 $69 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
 $70 = HEAP32[$69>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
 $71 = ($70|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
 if ($71) {
  $$0$i6$i13 = 0;
  $83 = $$0$i6$i13&255; //@line 1309 "./libraries/stb/stb_image.h"
  $84 = $83 << 8; //@line 1309 "./libraries/stb/stb_image.h"
  $85 = $84 | $63; //@line 1309 "./libraries/stb/stb_image.h"
  $86 = $85 << 16; //@line 1317 "./libraries/stb/stb_image.h"
  $87 = $43 | $86; //@line 1317 "./libraries/stb/stb_image.h"
  return ($87|0); //@line 1317 "./libraries/stb/stb_image.h"
 }
 $72 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
 $73 = HEAP32[$72>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
 $74 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
 $75 = HEAP32[$74>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
 $76 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
 $77 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
 $78 = HEAP32[$77>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
 $79 = (FUNCTION_TABLE_iiii[$73 & 8191]($75,$76,$78)|0); //@line 1213 "./libraries/stb/stb_image.h"
 $80 = ($79|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
 if ($80) {
  HEAP32[$69>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
  $81 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
  HEAP8[$76>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
  $$pre$phi$i3$i10Z2D = $81;$$sink$i$i4$i11 = $81;$91 = 0;
 } else {
  $82 = (((($s)) + 40|0) + ($79)|0); //@line 1223 "./libraries/stb/stb_image.h"
  $$pre$i1$i8 = HEAP8[$76>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
  $$pre1$i2$i9 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
  $$pre$phi$i3$i10Z2D = $$pre1$i2$i9;$$sink$i$i4$i11 = $82;$91 = $$pre$i1$i8;
 }
 HEAP32[$2>>2] = $$sink$i$i4$i11; //@line 1223 "./libraries/stb/stb_image.h"
 HEAP32[$0>>2] = $$pre$phi$i3$i10Z2D; //@line 1233 "./libraries/stb/stb_image.h"
 $$0$i6$i13 = $91;
 $83 = $$0$i6$i13&255; //@line 1309 "./libraries/stb/stb_image.h"
 $84 = $83 << 8; //@line 1309 "./libraries/stb/stb_image.h"
 $85 = $84 | $63; //@line 1309 "./libraries/stb/stb_image.h"
 $86 = $85 << 16; //@line 1317 "./libraries/stb/stb_image.h"
 $87 = $43 | $86; //@line 1317 "./libraries/stb/stb_image.h"
 return ($87|0); //@line 1317 "./libraries/stb/stb_image.h"
}
function __ZN4love5image6magpie10STBHandler9canEncodeENS0_9ImageData13EncodedFormatE($this,$format) {
 $this = $this|0;
 $format = $format|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($format|0)==(0); //@line 64 "modules/image/magpie/STBHandler.cpp"
 return ($0|0); //@line 64 "modules/image/magpie/STBHandler.cpp"
}
function __ZN4love5image6magpie10STBHandler6decodeEPNS_10filesystem8FileDataE($agg$result,$this,$data) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $data = $data|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $col$04$us$us$i$us$i = 0, $comp = 0, $exitcond20$i$us$i = 0, $exitcond21$i$us$i = 0, $or$cond$i$i = 0, $row$07$us$i$us$i = 0, $s$i = 0;
 var $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $s$i = sp + 8|0;
 $comp = sp + 4|0;
 ;HEAP32[$agg$result>>2]=0|0;HEAP32[$agg$result+4>>2]=0|0;HEAP32[$agg$result+8>>2]=0|0;HEAP32[$agg$result+12>>2]=0|0; //@line 48 "modules/image/magpie/FormatHandler.h"
 HEAP32[$comp>>2] = 0; //@line 71 "modules/image/magpie/STBHandler.cpp"
 $0 = HEAP32[$data>>2]|0; //@line 72 "modules/image/magpie/STBHandler.cpp"
 $1 = ((($0)) + 16|0); //@line 72 "modules/image/magpie/STBHandler.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 72 "modules/image/magpie/STBHandler.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($data)|0); //@line 72 "modules/image/magpie/STBHandler.cpp"
 $4 = HEAP32[$data>>2]|0; //@line 73 "modules/image/magpie/STBHandler.cpp"
 $5 = ((($4)) + 20|0); //@line 73 "modules/image/magpie/STBHandler.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 73 "modules/image/magpie/STBHandler.cpp"
 $7 = (FUNCTION_TABLE_ii[$6 & 8191]($data)|0); //@line 73 "modules/image/magpie/STBHandler.cpp"
 $8 = ((($agg$result)) + 4|0); //@line 74 "modules/image/magpie/STBHandler.cpp"
 $9 = ((($s$i)) + 16|0); //@line 770 "./libraries/stb/stb_image.h"
 HEAP32[$9>>2] = 0; //@line 770 "./libraries/stb/stb_image.h"
 $10 = ((($s$i)) + 32|0); //@line 771 "./libraries/stb/stb_image.h"
 HEAP32[$10>>2] = 0; //@line 771 "./libraries/stb/stb_image.h"
 $11 = ((($s$i)) + 176|0); //@line 772 "./libraries/stb/stb_image.h"
 HEAP32[$11>>2] = $3; //@line 772 "./libraries/stb/stb_image.h"
 $12 = ((($s$i)) + 168|0); //@line 772 "./libraries/stb/stb_image.h"
 HEAP32[$12>>2] = $3; //@line 772 "./libraries/stb/stb_image.h"
 $13 = (($3) + ($7)|0); //@line 773 "./libraries/stb/stb_image.h"
 $14 = ((($s$i)) + 180|0); //@line 773 "./libraries/stb/stb_image.h"
 HEAP32[$14>>2] = $13; //@line 773 "./libraries/stb/stb_image.h"
 $15 = ((($s$i)) + 172|0); //@line 773 "./libraries/stb/stb_image.h"
 HEAP32[$15>>2] = $13; //@line 773 "./libraries/stb/stb_image.h"
 $16 = (__ZL15stbi__load_mainP13stbi__contextPiS1_S1_i($s$i,$agg$result,$8,$comp,4)|0); //@line 980 "./libraries/stb/stb_image.h"
 $17 = HEAP32[158755]|0; //@line 982 "./libraries/stb/stb_image.h"
 $18 = ($17|0)!=(0); //@line 982 "./libraries/stb/stb_image.h"
 $19 = ($16|0)!=(0|0); //@line 982 "./libraries/stb/stb_image.h"
 $or$cond$i$i = $19 & $18; //@line 982 "./libraries/stb/stb_image.h"
 if ($or$cond$i$i) {
  $20 = HEAP32[$agg$result>>2]|0; //@line 983 "./libraries/stb/stb_image.h"
  $21 = HEAP32[$8>>2]|0; //@line 983 "./libraries/stb/stb_image.h"
  $22 = $21 >> 1; //@line 989 "./libraries/stb/stb_image.h"
  $23 = ($22|0)>(0); //@line 989 "./libraries/stb/stb_image.h"
  if ($23) {
   $24 = ($20|0)>(0); //@line 990 "./libraries/stb/stb_image.h"
   $25 = (($21) + -1)|0; //@line 993 "./libraries/stb/stb_image.h"
   if ($24) {
    $row$07$us$i$us$i = 0;
    while(1) {
     $26 = Math_imul($row$07$us$i$us$i, $20)|0; //@line 992 "./libraries/stb/stb_image.h"
     $27 = (($25) - ($row$07$us$i$us$i))|0; //@line 993 "./libraries/stb/stb_image.h"
     $28 = Math_imul($27, $20)|0; //@line 993 "./libraries/stb/stb_image.h"
     $col$04$us$us$i$us$i = 0;
     while(1) {
      $29 = (($col$04$us$us$i$us$i) + ($26))|0; //@line 992 "./libraries/stb/stb_image.h"
      $30 = $29 << 2; //@line 992 "./libraries/stb/stb_image.h"
      $31 = (($col$04$us$us$i$us$i) + ($28))|0; //@line 993 "./libraries/stb/stb_image.h"
      $32 = $31 << 2; //@line 993 "./libraries/stb/stb_image.h"
      $33 = (($16) + ($30)|0); //@line 992 "./libraries/stb/stb_image.h"
      $34 = HEAP8[$33>>0]|0; //@line 992 "./libraries/stb/stb_image.h"
      $35 = (($16) + ($32)|0); //@line 993 "./libraries/stb/stb_image.h"
      $36 = HEAP8[$35>>0]|0; //@line 993 "./libraries/stb/stb_image.h"
      HEAP8[$33>>0] = $36; //@line 993 "./libraries/stb/stb_image.h"
      HEAP8[$35>>0] = $34; //@line 994 "./libraries/stb/stb_image.h"
      $37 = $30 | 1; //@line 992 "./libraries/stb/stb_image.h"
      $38 = (($16) + ($37)|0); //@line 992 "./libraries/stb/stb_image.h"
      $39 = HEAP8[$38>>0]|0; //@line 992 "./libraries/stb/stb_image.h"
      $40 = $32 | 1; //@line 993 "./libraries/stb/stb_image.h"
      $41 = (($16) + ($40)|0); //@line 993 "./libraries/stb/stb_image.h"
      $42 = HEAP8[$41>>0]|0; //@line 993 "./libraries/stb/stb_image.h"
      HEAP8[$38>>0] = $42; //@line 993 "./libraries/stb/stb_image.h"
      HEAP8[$41>>0] = $39; //@line 994 "./libraries/stb/stb_image.h"
      $43 = $30 | 2; //@line 992 "./libraries/stb/stb_image.h"
      $44 = (($16) + ($43)|0); //@line 992 "./libraries/stb/stb_image.h"
      $45 = HEAP8[$44>>0]|0; //@line 992 "./libraries/stb/stb_image.h"
      $46 = $32 | 2; //@line 993 "./libraries/stb/stb_image.h"
      $47 = (($16) + ($46)|0); //@line 993 "./libraries/stb/stb_image.h"
      $48 = HEAP8[$47>>0]|0; //@line 993 "./libraries/stb/stb_image.h"
      HEAP8[$44>>0] = $48; //@line 993 "./libraries/stb/stb_image.h"
      HEAP8[$47>>0] = $45; //@line 994 "./libraries/stb/stb_image.h"
      $49 = $30 | 3; //@line 992 "./libraries/stb/stb_image.h"
      $50 = (($16) + ($49)|0); //@line 992 "./libraries/stb/stb_image.h"
      $51 = HEAP8[$50>>0]|0; //@line 992 "./libraries/stb/stb_image.h"
      $52 = $32 | 3; //@line 993 "./libraries/stb/stb_image.h"
      $53 = (($16) + ($52)|0); //@line 993 "./libraries/stb/stb_image.h"
      $54 = HEAP8[$53>>0]|0; //@line 993 "./libraries/stb/stb_image.h"
      HEAP8[$50>>0] = $54; //@line 993 "./libraries/stb/stb_image.h"
      HEAP8[$53>>0] = $51; //@line 994 "./libraries/stb/stb_image.h"
      $55 = (($col$04$us$us$i$us$i) + 1)|0; //@line 990 "./libraries/stb/stb_image.h"
      $exitcond20$i$us$i = ($55|0)==($20|0); //@line 990 "./libraries/stb/stb_image.h"
      if ($exitcond20$i$us$i) {
       break;
      } else {
       $col$04$us$us$i$us$i = $55;
      }
     }
     $56 = (($row$07$us$i$us$i) + 1)|0; //@line 989 "./libraries/stb/stb_image.h"
     $exitcond21$i$us$i = ($56|0)==($22|0); //@line 989 "./libraries/stb/stb_image.h"
     if ($exitcond21$i$us$i) {
      break;
     } else {
      $row$07$us$i$us$i = $56;
     }
    }
    $57 = ((($agg$result)) + 12|0); //@line 72 "modules/image/magpie/STBHandler.cpp"
    HEAP32[$57>>2] = $16; //@line 72 "modules/image/magpie/STBHandler.cpp"
    $61 = $20;
    label = 10;
   } else {
    label = 8;
   }
  } else {
   label = 8;
  }
 } else {
  label = 8;
 }
 if ((label|0) == 8) {
  $58 = ((($agg$result)) + 12|0); //@line 72 "modules/image/magpie/STBHandler.cpp"
  HEAP32[$58>>2] = $16; //@line 72 "modules/image/magpie/STBHandler.cpp"
  $59 = ($16|0)==(0|0); //@line 77 "modules/image/magpie/STBHandler.cpp"
  if (!($59)) {
   $$pre = HEAP32[$agg$result>>2]|0; //@line 77 "modules/image/magpie/STBHandler.cpp"
   $61 = $$pre;
   label = 10;
  }
 }
 if ((label|0) == 10) {
  $60 = ($61|0)<(1); //@line 77 "modules/image/magpie/STBHandler.cpp"
  if (!($60)) {
   $62 = HEAP32[$8>>2]|0; //@line 77 "modules/image/magpie/STBHandler.cpp"
   $63 = ($62|0)<(1); //@line 77 "modules/image/magpie/STBHandler.cpp"
   if (!($63)) {
    $69 = $61 << 2; //@line 80 "modules/image/magpie/STBHandler.cpp"
    $70 = Math_imul($69, $62)|0; //@line 80 "modules/image/magpie/STBHandler.cpp"
    $71 = ((($agg$result)) + 8|0); //@line 80 "modules/image/magpie/STBHandler.cpp"
    HEAP32[$71>>2] = $70; //@line 80 "modules/image/magpie/STBHandler.cpp"
    STACKTOP = sp;return; //@line 83 "modules/image/magpie/STBHandler.cpp"
   }
  }
 }
 $64 = (___cxa_allocate_exception(16)|0); //@line 78 "modules/image/magpie/STBHandler.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($64|0),(441512|0),($vararg_buffer|0)); //@line 78 "modules/image/magpie/STBHandler.cpp"
 $65 = __THREW__; __THREW__ = 0;
 $66 = $65&1;
 if ($66) {
  $67 = ___cxa_find_matching_catch_2()|0;
  $68 = tempRet0;
  ___cxa_free_exception(($64|0)); //@line 78 "modules/image/magpie/STBHandler.cpp"
  ___resumeException($67|0);
  // unreachable;
 } else {
  ___cxa_throw(($64|0),(96|0),(344|0)); //@line 78 "modules/image/magpie/STBHandler.cpp"
  // unreachable; //@line 78 "modules/image/magpie/STBHandler.cpp"
 }
}
function __ZL15stbi__load_mainP13stbi__contextPiS1_S1_i($s,$x,$y,$comp,$req_comp) {
 $s = $s|0;
 $x = $x|0;
 $y = $y|0;
 $comp = $comp|0;
 $req_comp = $req_comp|0;
 var $$$i = 0, $$$i$i = 0, $$$i$i$i$i$i = 0, $$$i$i28 = 0, $$$i150 = 0, $$$i20$i$i$i$i = 0, $$$i218$i = 0, $$$i236$i = 0, $$$i261$i = 0, $$$i30$i$i$i$i = 0, $$$i47$i$i$i$i = 0, $$$i57$i$i$i$i = 0, $$$i68$i$i$i$i = 0, $$0 = 0, $$0$i$i = 0, $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i$i = 0, $$0$i$i$i$i$i$i = 0, $$0$i$i$i$i$i$i$i = 0;
 var $$0$i$i$i$i$i$i$i$lcssa = 0, $$0$i$i$i25$i$i$i = 0, $$0$i$i$i72 = 0, $$0$i$i100$i = 0, $$0$i$i100$i129 = 0, $$0$i$i109$i = 0, $$0$i$i18 = 0, $$0$i$i208$i = 0, $$0$i$i33$i$i$i = 0, $$0$i$i33$i$i$i$lcssa = 0, $$0$i$i37$i = 0, $$0$i$i51 = 0, $$0$i$i55$i$i$i = 0, $$0$i$i57$i$i$i$i$i = 0, $$0$i$i57$i$i$i$i$i$lcssa = 0, $$0$i$i67$i = 0, $$0$i$i8$i$i$i = 0, $$0$i$i81$i = 0, $$0$i$i86$i = 0, $$0$i$i86$i116 = 0;
 var $$0$i$i95$i = 0, $$0$i$i97$i$i$i$i$i = 0, $$0$i$i97$i$i$i$i$i$lcssa = 0, $$0$i108$i$i$i$i$i = 0, $$0$i11$i$i = 0, $$0$i114$i = 0, $$0$i114$i142 = 0, $$0$i115$i$i$i$i$i = 0, $$0$i115$i$i$i$i$i$lcssa = 0, $$0$i121$i = 0, $$0$i122$i$i$i$i$i = 0, $$0$i123$i = 0, $$0$i130$i = 0, $$0$i137$i = 0, $$0$i14$i$i$i$i = 0, $$0$i148$i = 0, $$0$i153 = 0, $$0$i153$i = 0, $$0$i153$i$us = 0, $$0$i155$i = 0;
 var $$0$i160$i = 0, $$0$i160$i$us = 0, $$0$i162$i = 0, $$0$i180$i = 0, $$0$i182$i = 0, $$0$i189$i = 0, $$0$i196$i = 0, $$0$i2$i$i$i = 0, $$0$i20$i$i$i$i$i = 0, $$0$i20$i$i$i$i$i$lcssa = 0, $$0$i21$i$i$i$i = 0, $$0$i23$i = 0, $$0$i23$i58 = 0, $$0$i233$i = 0, $$0$i251$i = 0, $$0$i258$i = 0, $$0$i27$i$i$i$i$i = 0, $$0$i276$i = 0, $$0$i28$i$i$i$i = 0, $$0$i28$i19$i$i$i = 0;
 var $$0$i28$i19$i$i$i$lcssa = 0, $$0$i3$i$i = 0, $$0$i30$i = 0, $$0$i30$i65 = 0, $$0$i35$i$i$i$i = 0, $$0$i36$i$i$i$i = 0, $$0$i39$i$i$i$i$i = 0, $$0$i41$i$i$i = 0, $$0$i45$i$i$i$i = 0, $$0$i45$i$i$i$i$lcssa = 0, $$0$i46$i = 0, $$0$i48$i$i$i = 0, $$0$i48$i$i$i$i$i = 0, $$0$i48$i$i$i$i$i$lcssa = 0, $$0$i51$i = 0, $$0$i53$i$i$i$i = 0, $$0$i6$i$i = 0, $$0$i6$i$i$i$i$i = 0, $$0$i6$i102$i = 0, $$0$i6$i107$i = 0;
 var $$0$i6$i107$i136 = 0, $$0$i6$i116$i = 0, $$0$i6$i215$i = 0, $$0$i6$i44$i = 0, $$0$i6$i74$i = 0, $$0$i6$i88$i = 0, $$0$i6$i93$i = 0, $$0$i6$i93$i123 = 0, $$0$i66$i$i$i = 0, $$0$i66$i$i$i$i = 0, $$0$i66$i$i$i$i$lcssa = 0, $$0$i68$i$i$i$i$i = 0, $$0$i74$i$i$i$i = 0, $$0$i81$i$i$i$i = 0, $$0$i81$i$i$i$i$i = 0, $$0$i88$i$i$i$i$i = 0, $$0$i88$i$i$i$i$i$lcssa = 0, $$01$i269$i = 0, $$1$i$i = 0, $$1$i219$i = 0;
 var $$1$i237$i = 0, $$1$i262$i = 0, $$11$i$i$i$i = 0, $$16$i = 0, $$2$i = 0, $$2$i$i = 0, $$2$i$i29 = 0, $$2$i221$i = 0, $$2$i239$i = 0, $$2$i264$i = 0, $$26$i$i$i$i = 0, $$3$i$i = 0, $$3$i$i$i$i = 0, $$3$i$i30 = 0, $$3$i223$i = 0, $$3$i241$i = 0, $$3$i266$i = 0, $$34$i = 0, $$35$i = 0, $$36$i = 0;
 var $$37$i = 0, $$cast = 0, $$cast156 = 0, $$in$i = 0, $$in$i$i = 0, $$in215$i = 0, $$in216$i = 0, $$in217$i = 0, $$in218$i = 0, $$in219$i = 0, $$in220$i = 0, $$in221$i = 0, $$in222$i = 0, $$in223$i = 0, $$in224$i = 0, $$in225$i = 0, $$in226$i = 0, $$in227$i = 0, $$in228$i = 0, $$in229$i = 0;
 var $$in230$i = 0, $$in5$i$i = 0, $$lcssa$i$i = 0, $$lcssa486 = 0, $$lcssa492 = 0, $$lcssa493 = 0, $$lcssa498 = 0, $$lcssa498$lcssa = 0, $$lcssa502 = 0, $$lcssa502$lcssa = 0, $$lcssa505 = 0, $$lcssa505$lcssa = 0, $$lcssa510 = 0, $$lcssa514 = 0, $$lcssa515 = 0, $$n$3$i$i = 0, $$n$3$i225$i = 0, $$n$3$i243$i = 0, $$n$3$i268$i = 0, $$off$i = 0;
 var $$ph$i = 0, $$phi$trans$insert$i = 0, $$phi$trans$insert$i$i$i$i = 0, $$phi$trans$insert50$i$i$i$i = 0, $$pr$i$i$i = 0, $$pr76$i$i$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i$i = 0, $$pre$i$i$i$i$i = 0, $$pre$i$i$i$i$i$i = 0, $$pre$i$i$i$i$i$i$i = 0, $$pre$i$i$i$i1 = 0, $$pre$i$i$i27$i$i$i = 0, $$pre$i$i$i67 = 0, $$pre$i$i$i7 = 0, $$pre$i$i104$i = 0, $$pre$i$i203$i = 0;
 var $$pre$i$i28$i$i$i = 0, $$pre$i$i3$i$i$i = 0, $$pre$i$i32$i = 0, $$pre$i$i46 = 0, $$pre$i$i50$i$i$i = 0, $$pre$i$i52$i$i$i$i$i = 0, $$pre$i$i62$i = 0, $$pre$i$i76$i = 0, $$pre$i$i81$i = 0, $$pre$i$i81$i111 = 0, $$pre$i$i90$i = 0, $$pre$i$i92$i$i$i$i$i = 0, $$pre$i$i95$i = 0, $$pre$i$i95$i124 = 0, $$pre$i1$i$i = 0, $$pre$i1$i$i$i$i$i = 0, $$pre$i1$i102$i = 0, $$pre$i1$i102$i131 = 0, $$pre$i1$i111$i = 0, $$pre$i1$i210$i = 0;
 var $$pre$i1$i39$i = 0, $$pre$i1$i69$i = 0, $$pre$i1$i83$i = 0, $$pre$i1$i88$i = 0, $$pre$i1$i88$i118 = 0, $$pre$i1$i97$i = 0, $$pre$i103$i$i$i$i$i = 0, $$pre$i109$i = 0, $$pre$i109$i137 = 0, $$pre$i11$i$i$i = 0, $$pre$i110$i$i$i$i$i = 0, $$pre$i116$i = 0, $$pre$i117$i$i$i$i$i = 0, $$pre$i118$i = 0, $$pre$i12$i$i$i$i$i = 0, $$pre$i123$i = 0, $$pre$i125$i = 0, $$pre$i126$i = 0, $$pre$i126$us$i = 0, $$pre$i129$i = 0;
 var $$pre$i132$i = 0, $$pre$i135$i = 0, $$pre$i143$i = 0, $$pre$i146$i = 0, $$pre$i148$i = 0, $$pre$i148$i$us = 0, $$pre$i15$i$i$i$i$i = 0, $$pre$i150$i = 0, $$pre$i155$i = 0, $$pre$i155$i$us = 0, $$pre$i157$i = 0, $$pre$i16$i$i$i$i = 0, $$pre$i163$i = 0, $$pre$i169$i = 0, $$pre$i175$i = 0, $$pre$i18$i = 0, $$pre$i18$i53 = 0, $$pre$i22$i$i$i = 0, $$pre$i22$i$i$i$i$i = 0, $$pre$i228$i = 0;
 var $$pre$i23$i$i$i$i = 0, $$pre$i23$i14$i$i$i = 0, $$pre$i24 = 0, $$pre$i246$i = 0, $$pre$i25$i = 0, $$pre$i25$i60 = 0, $$pre$i253$i = 0, $$pre$i271$i = 0, $$pre$i30$i$i$i$i = 0, $$pre$i31$i$i$i$i = 0, $$pre$i31$i$i$i$i$i = 0, $$pre$i34$i$i$i$i$i = 0, $$pre$i36$i$i$i = 0, $$pre$i40$i$i$i$i = 0, $$pre$i41$i = 0, $$pre$i43$i$i$i = 0, $$pre$i43$i$i$i$i$i = 0, $$pre$i46$i = 0, $$pre$i48$i$i$i$i = 0, $$pre$i55$i$i$i$i = 0;
 var $$pre$i59$i$i$i$i$i = 0, $$pre$i6$i$i = 0, $$pre$i61$i$i$i = 0, $$pre$i61$i$i$i$i = 0, $$pre$i63$i$i$i$i$i = 0, $$pre$i69$i$i$i = 0, $$pre$i69$i$i$i$i = 0, $$pre$i72$i$i$i$i$i = 0, $$pre$i76$i$i$i$i = 0, $$pre$i76$i$i$i$i$i = 0, $$pre$i83$i$i$i$i$i = 0, $$pre$i9$i$i$i$i = 0, $$pre$i99$i$i$i$i$i = 0, $$pre$phi$i$i$i$i$i$i$iZ2D = 0, $$pre$phi$i$i$i$i$i$iZ2D = 0, $$pre$phi$i$i$i$i$iZ2D = 0, $$pre$phi$i$i$i$i12Z2D = 0, $$pre$phi$i$i$i$i3Z2D = 0, $$pre$phi$i$i$i$iZ2D = 0, $$pre$phi$i$i$i20Z2D = 0;
 var $$pre$phi$i$i$i38Z2D = 0, $$pre$phi$i$i$i69Z2D = 0, $$pre$phi$i$i$iZ2D = 0, $$pre$phi$i$i106$iZ2D = 0, $$pre$phi$i$i15$i$iZ2D = 0, $$pre$phi$i$i30$i$i$iZ2D = 0, $$pre$phi$i$i33Z2D = 0, $$pre$phi$i$i34$i80Z2D = 0, $$pre$phi$i$i34$iZ2D = 0, $$pre$phi$i$i48Z2D = 0, $$pre$phi$i$i5$i$i$iZ2D = 0, $$pre$phi$i$i50$iZ2D = 0, $$pre$phi$i$i52$i$i$iZ2D = 0, $$pre$phi$i$i54$i$i$i$i$iZ2D = 0, $$pre$phi$i$i55$i94Z2D = 0, $$pre$phi$i$i55$iZ2D = 0, $$pre$phi$i$i64$iZ2D = 0, $$pre$phi$i$i69$i103Z2D = 0, $$pre$phi$i$i69$iZ2D = 0, $$pre$phi$i$i78$iZ2D = 0;
 var $$pre$phi$i$i83$i113Z2D = 0, $$pre$phi$i$i83$iZ2D = 0, $$pre$phi$i$i92$iZ2D = 0, $$pre$phi$i$i94$i$i$i$i$iZ2D = 0, $$pre$phi$i$i97$i126Z2D = 0, $$pre$phi$i$i97$iZ2D = 0, $$pre$phi$i$iZ2D = 0, $$pre$phi$i105$i$i$i$i$iZ2D = 0, $$pre$phi$i11$i$i$i$iZ2D = 0, $$pre$phi$i111$i139Z2D = 0, $$pre$phi$i111$iZ2D = 0, $$pre$phi$i112$i$i$i$i$iZ2D = 0, $$pre$phi$i118$iZ2D = 0, $$pre$phi$i119$i$i$i$i$iZ2D = 0, $$pre$phi$i14$i$i$i$i$iZ2D = 0, $$pre$phi$i17$i$i$i$i$iZ2D = 0, $$pre$phi$i18$i$i$i$iZ2D = 0, $$pre$phi$i20$i55Z2D = 0, $$pre$phi$i20$iZ2D = 0, $$pre$phi$i24$i$i$i$i$iZ2D = 0;
 var $$pre$phi$i25$i$i$i$iZ2D = 0, $$pre$phi$i25$i16$i$i$iZ2D = 0, $$pre$phi$i27$i62Z2D = 0, $$pre$phi$i27$iZ2D = 0, $$pre$phi$i3$i$i$i$i$iZ2D = 0, $$pre$phi$i3$i$i$iZ2D = 0, $$pre$phi$i3$i$i43Z2D = 0, $$pre$phi$i3$i$i75Z2D = 0, $$pre$phi$i3$i$iZ2D = 0, $$pre$phi$i3$i104$i133Z2D = 0, $$pre$phi$i3$i104$iZ2D = 0, $$pre$phi$i3$i113$iZ2D = 0, $$pre$phi$i3$i22$i$iZ2D = 0, $$pre$phi$i3$i41$i85Z2D = 0, $$pre$phi$i3$i41$iZ2D = 0, $$pre$phi$i3$i57$iZ2D = 0, $$pre$phi$i3$i62$i99Z2D = 0, $$pre$phi$i3$i62$iZ2D = 0, $$pre$phi$i3$i71$iZ2D = 0, $$pre$phi$i3$i76$i108Z2D = 0;
 var $$pre$phi$i3$i76$iZ2D = 0, $$pre$phi$i3$i85$iZ2D = 0, $$pre$phi$i3$i90$i120Z2D = 0, $$pre$phi$i3$i90$iZ2D = 0, $$pre$phi$i3$i99$iZ2D = 0, $$pre$phi$i32$i$i$i$iZ2D = 0, $$pre$phi$i33$i$i$i$iZ2D = 0, $$pre$phi$i36$i$i$i$i$iZ2D = 0, $$pre$phi$i38$i$i$iZ2D = 0, $$pre$phi$i42$i$i$i$iZ2D = 0, $$pre$phi$i43$iZ2D = 0, $$pre$phi$i45$i$i$i$i$iZ2D = 0, $$pre$phi$i45$i$i$iZ2D = 0, $$pre$phi$i48$i89Z2D = 0, $$pre$phi$i48$iZ2D = 0, $$pre$phi$i50$i$i$i$iZ2D = 0, $$pre$phi$i63$i$i$i$iZ2D = 0, $$pre$phi$i63$i$i$iZ2D = 0, $$pre$phi$i65$i$i$i$i$iZ2D = 0, $$pre$phi$i71$i$i$i$iZ2D = 0;
 var $$pre$phi$i78$i$i$i$i$iZ2D = 0, $$pre$phi$i78$i$i$i$iZ2D = 0, $$pre$phi$i8$i$iZ2D = 0, $$pre$phi$i85$i$i$i$i$iZ2D = 0, $$pre1$i$i = 0, $$pre1$i$i$i = 0, $$pre1$i$i$i$i = 0, $$pre1$i$i$i$i$i = 0, $$pre1$i$i$i$i$i$i = 0, $$pre1$i$i$i$i$i$i$i = 0, $$pre1$i$i$i$i11 = 0, $$pre1$i$i$i$i2 = 0, $$pre1$i$i$i19 = 0, $$pre1$i$i$i37 = 0, $$pre1$i$i$i68 = 0, $$pre1$i$i105$i = 0, $$pre1$i$i14$i$i = 0, $$pre1$i$i29$i$i$i = 0, $$pre1$i$i32 = 0, $$pre1$i$i33$i = 0;
 var $$pre1$i$i33$i79 = 0, $$pre1$i$i4$i$i$i = 0, $$pre1$i$i47 = 0, $$pre1$i$i49$i = 0, $$pre1$i$i51$i$i$i = 0, $$pre1$i$i53$i$i$i$i$i = 0, $$pre1$i$i54$i = 0, $$pre1$i$i54$i93 = 0, $$pre1$i$i63$i = 0, $$pre1$i$i68$i = 0, $$pre1$i$i68$i102 = 0, $$pre1$i$i77$i = 0, $$pre1$i$i82$i = 0, $$pre1$i$i82$i112 = 0, $$pre1$i$i91$i = 0, $$pre1$i$i93$i$i$i$i$i = 0, $$pre1$i$i96$i = 0, $$pre1$i$i96$i125 = 0, $$pre1$i10$i$i$i$i = 0, $$pre1$i104$i$i$i$i$i = 0;
 var $$pre1$i110$i = 0, $$pre1$i110$i138 = 0, $$pre1$i111$i$i$i$i$i = 0, $$pre1$i117$i = 0, $$pre1$i118$i$i$i$i$i = 0, $$pre1$i13$i$i$i$i$i = 0, $$pre1$i16$i$i$i$i$i = 0, $$pre1$i17$i$i$i$i = 0, $$pre1$i19$i = 0, $$pre1$i19$i54 = 0, $$pre1$i2$i$i = 0, $$pre1$i2$i$i$i = 0, $$pre1$i2$i$i$i$i$i = 0, $$pre1$i2$i$i42 = 0, $$pre1$i2$i$i74 = 0, $$pre1$i2$i103$i = 0, $$pre1$i2$i103$i132 = 0, $$pre1$i2$i112$i = 0, $$pre1$i2$i21$i$i = 0, $$pre1$i2$i40$i = 0;
 var $$pre1$i2$i40$i84 = 0, $$pre1$i2$i56$i = 0, $$pre1$i2$i61$i = 0, $$pre1$i2$i61$i98 = 0, $$pre1$i2$i70$i = 0, $$pre1$i2$i75$i = 0, $$pre1$i2$i75$i107 = 0, $$pre1$i2$i84$i = 0, $$pre1$i2$i89$i = 0, $$pre1$i2$i89$i119 = 0, $$pre1$i2$i98$i = 0, $$pre1$i23$i$i$i$i$i = 0, $$pre1$i24$i$i$i$i = 0, $$pre1$i24$i15$i$i$i = 0, $$pre1$i26$i = 0, $$pre1$i26$i61 = 0, $$pre1$i31$i$i$i$i = 0, $$pre1$i32$i$i$i$i = 0, $$pre1$i35$i$i$i$i$i = 0, $$pre1$i37$i$i$i = 0;
 var $$pre1$i41$i$i$i$i = 0, $$pre1$i42$i = 0, $$pre1$i44$i$i$i = 0, $$pre1$i44$i$i$i$i$i = 0, $$pre1$i47$i = 0, $$pre1$i47$i88 = 0, $$pre1$i49$i$i$i$i = 0, $$pre1$i62$i$i$i = 0, $$pre1$i62$i$i$i$i = 0, $$pre1$i64$i$i$i$i$i = 0, $$pre1$i7$i$i = 0, $$pre1$i70$i$i$i$i = 0, $$pre1$i77$i$i$i$i = 0, $$pre1$i77$i$i$i$i$i = 0, $$pre1$i84$i$i$i$i$i = 0, $$pre125$i$i = 0, $$pre127$i$i = 0, $$pre128$i$i = 0, $$pre200$i$i$i$i = 0, $$pre201$i$i$i$i = 0;
 var $$pre203$i$i$i$i = 0, $$pre204$i$i$i$i = 0, $$pre205$i$i$i$i = 0, $$pre206$i$i$i$i = 0, $$pre207$i$i$i$i = 0, $$pre208$i$i$i$i = 0, $$pre210$i$i$i$i = 0, $$pre211$i$i$i$i = 0, $$pre212$i$i$i$i = 0, $$pre281 = 0, $$pre281$pre = 0, $$pre282 = 0, $$pre283 = 0, $$pre283$pre = 0, $$pre3$i$i = 0, $$pre3$i$us$i = 0, $$pre3$i137$i = 0, $$pre345$i = 0, $$pre346$i = 0, $$pre347$i = 0;
 var $$pre348$i = 0, $$pre349$i = 0, $$pre350$i = 0, $$pre351$i = 0, $$pre352$i = 0, $$pre47$i$i$i$i = 0, $$pre48$i$i$i$i = 0, $$pre49$i$i$i$i = 0, $$pre51$i$i$i$i = 0, $$pre52$i$i$i$i = 0, $$pre93$i = 0, $$pre94$i = 0, $$req_comp$i$i = 0, $$sink$i$i$i = 0, $$sink$i$i$i$i = 0, $$sink$i$i$i$i$i = 0, $$sink$i$i$i$i$i$i = 0, $$sink$i$i$i$i$i$i$i = 0, $$sink$i$i$i$i$i$i$i$i = 0, $$sink$i$i$i$i$i13 = 0;
 var $$sink$i$i$i$i$i4 = 0, $$sink$i$i$i$i21 = 0, $$sink$i$i$i$i23$i$i$i = 0, $$sink$i$i$i$i39 = 0, $$sink$i$i$i$i70 = 0, $$sink$i$i$i$i8 = 0, $$sink$i$i$i107$i = 0, $$sink$i$i$i16$i$i = 0, $$sink$i$i$i17 = 0, $$sink$i$i$i206$i = 0, $$sink$i$i$i31$i$i$i = 0, $$sink$i$i$i34 = 0, $$sink$i$i$i35$i = 0, $$sink$i$i$i35$i81 = 0, $$sink$i$i$i49 = 0, $$sink$i$i$i51$i = 0, $$sink$i$i$i53$i$i$i = 0, $$sink$i$i$i55$i$i$i$i$i = 0, $$sink$i$i$i56$i = 0, $$sink$i$i$i56$i95 = 0;
 var $$sink$i$i$i6$i$i$i = 0, $$sink$i$i$i65$i = 0, $$sink$i$i$i70$i = 0, $$sink$i$i$i70$i104 = 0, $$sink$i$i$i79$i = 0, $$sink$i$i$i84$i = 0, $$sink$i$i$i84$i114 = 0, $$sink$i$i$i93$i = 0, $$sink$i$i$i95$i$i$i$i$i = 0, $$sink$i$i$i98$i = 0, $$sink$i$i$i98$i127 = 0, $$sink$i$i106$i$i$i$i$i = 0, $$sink$i$i112$i = 0, $$sink$i$i112$i140 = 0, $$sink$i$i113$i$i$i$i$i = 0, $$sink$i$i119$i = 0, $$sink$i$i12$i$i$i$i = 0, $$sink$i$i120$i$i$i$i$i = 0, $$sink$i$i121$i = 0, $$sink$i$i128$i = 0;
 var $$sink$i$i135$i = 0, $$sink$i$i142$i = 0, $$sink$i$i146$i = 0, $$sink$i$i151$i = 0, $$sink$i$i151$i$us = 0, $$sink$i$i153$i = 0, $$sink$i$i158$i = 0, $$sink$i$i158$i$us = 0, $$sink$i$i160$i = 0, $$sink$i$i18$i$i$i$i$i = 0, $$sink$i$i19$i$i$i$i = 0, $$sink$i$i21$i = 0, $$sink$i$i21$i56 = 0, $$sink$i$i231$i = 0, $$sink$i$i249$i = 0, $$sink$i$i25$i$i$i$i$i = 0, $$sink$i$i256$i = 0, $$sink$i$i26$i$i$i$i = 0, $$sink$i$i26$i17$i$i$i = 0, $$sink$i$i274$i = 0;
 var $$sink$i$i28$i = 0, $$sink$i$i28$i63 = 0, $$sink$i$i33$i$i$i$i = 0, $$sink$i$i34$i$i$i$i = 0, $$sink$i$i37$i$i$i$i$i = 0, $$sink$i$i39$i$i$i = 0, $$sink$i$i4$i$i = 0, $$sink$i$i4$i$i$i = 0, $$sink$i$i4$i$i$i$i$i = 0, $$sink$i$i4$i$i44 = 0, $$sink$i$i4$i$i76 = 0, $$sink$i$i4$i100$i = 0, $$sink$i$i4$i105$i = 0, $$sink$i$i4$i105$i134 = 0, $$sink$i$i4$i114$i = 0, $$sink$i$i4$i213$i = 0, $$sink$i$i4$i23$i$i = 0, $$sink$i$i4$i42$i = 0, $$sink$i$i4$i42$i86 = 0, $$sink$i$i4$i58$i = 0;
 var $$sink$i$i4$i63$i = 0, $$sink$i$i4$i63$i100 = 0, $$sink$i$i4$i72$i = 0, $$sink$i$i4$i77$i = 0, $$sink$i$i4$i77$i109 = 0, $$sink$i$i4$i86$i = 0, $$sink$i$i4$i91$i = 0, $$sink$i$i4$i91$i121 = 0, $$sink$i$i43$i$i$i$i = 0, $$sink$i$i44$i = 0, $$sink$i$i46$i$i$i = 0, $$sink$i$i46$i$i$i$i$i = 0, $$sink$i$i49$i = 0, $$sink$i$i49$i90 = 0, $$sink$i$i51$i$i$i$i = 0, $$sink$i$i64$i$i$i = 0, $$sink$i$i64$i$i$i$i = 0, $$sink$i$i66$i$i$i$i$i = 0, $$sink$i$i72$i$i$i$i = 0, $$sink$i$i79$i$i$i$i = 0;
 var $$sink$i$i79$i$i$i$i$i = 0, $$sink$i$i86$i$i$i$i$i = 0, $$sink$i$i9$i$i = 0, $$sink$ph88$i$i$i = 0, $$z$i$i = 0, $$z$i217$i = 0, $$z$i235$i = 0, $$z$i260$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0;
 var $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0;
 var $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0;
 var $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0;
 var $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0;
 var $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0;
 var $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0;
 var $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0;
 var $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0;
 var $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0;
 var $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0;
 var $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0;
 var $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0;
 var $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0;
 var $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0;
 var $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0, $1267 = 0, $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0;
 var $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0, $1285 = 0, $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0;
 var $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0, $1302 = 0, $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0;
 var $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0, $1320 = 0, $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0;
 var $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0, $1339 = 0, $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0;
 var $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0, $1357 = 0, $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0;
 var $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0, $1375 = 0, $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0;
 var $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0, $1393 = 0, $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0;
 var $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0, $1410 = 0, $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0;
 var $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0, $1429 = 0, $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0;
 var $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0, $1447 = 0, $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0;
 var $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0, $1465 = 0, $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0;
 var $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0, $1483 = 0, $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0;
 var $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0, $1500 = 0, $1501 = 0, $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0, $1511 = 0;
 var $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0, $1518 = 0, $1519 = 0, $152 = 0, $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0, $153 = 0;
 var $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0, $1536 = 0, $1537 = 0, $1538 = 0, $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0, $1548 = 0;
 var $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0, $1554 = 0, $1555 = 0, $1556 = 0, $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $1565 = 0, $1566 = 0;
 var $1567 = 0, $1568 = 0, $1569 = 0, $157 = 0, $1570 = 0, $1571 = 0, $1572 = 0, $1573 = 0, $1574 = 0, $1575 = 0, $1576 = 0, $1577 = 0, $1578 = 0, $1579 = 0, $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0, $1583 = 0, $1584 = 0;
 var $1585 = 0, $1586 = 0, $1587 = 0, $1588 = 0, $1589 = 0, $159 = 0, $1590 = 0, $1591 = 0, $1592 = 0, $1593 = 0, $1594 = 0, $1595 = 0, $1596 = 0, $1597 = 0, $1598 = 0, $1599 = 0, $16 = 0, $160 = 0, $1600 = 0, $1601 = 0;
 var $1602 = 0, $1603 = 0, $1604 = 0, $1605 = 0, $1606 = 0, $1607 = 0, $1608 = 0, $1609 = 0, $161 = 0, $1610 = 0, $1611 = 0, $1612 = 0, $1613 = 0, $1614 = 0, $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0, $1619 = 0, $162 = 0;
 var $1620 = 0, $1621 = 0, $1622 = 0, $1623 = 0, $1624 = 0, $1625 = 0, $1626 = 0, $1627 = 0, $1628 = 0, $1629 = 0, $163 = 0, $1630 = 0, $1631 = 0, $1632 = 0, $1633 = 0, $1634 = 0, $1635 = 0, $1636 = 0, $1637 = 0, $1638 = 0;
 var $1639 = 0, $164 = 0, $1640 = 0, $1641 = 0, $1642 = 0, $1643 = 0, $1644 = 0, $1645 = 0, $1646 = 0, $1647 = 0, $1648 = 0, $1649 = 0, $165 = 0, $1650 = 0, $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0, $1655 = 0, $1656 = 0;
 var $1657 = 0, $1658 = 0, $1659 = 0, $166 = 0, $1660 = 0, $1661 = 0, $1662 = 0, $1663 = 0, $1664 = 0, $1665 = 0, $1666 = 0, $1667 = 0, $1668 = 0, $1669 = 0, $167 = 0, $1670 = 0, $1671 = 0, $1672 = 0, $1673 = 0, $1674 = 0;
 var $1675 = 0, $1676 = 0, $1677 = 0, $1678 = 0, $1679 = 0, $168 = 0, $1680 = 0, $1681 = 0, $1682 = 0, $1683 = 0, $1684 = 0, $1685 = 0, $1686 = 0, $1687 = 0, $1688 = 0, $1689 = 0, $169 = 0, $1690 = 0, $1691 = 0, $1692 = 0;
 var $1693 = 0, $1694 = 0, $1695 = 0, $1696 = 0, $1697 = 0, $1698 = 0, $1699 = 0, $17 = 0, $170 = 0, $1700 = 0, $1701 = 0, $1702 = 0, $1703 = 0, $1704 = 0, $1705 = 0, $1706 = 0, $1707 = 0, $1708 = 0, $1709 = 0, $171 = 0;
 var $1710 = 0, $1711 = 0, $1712 = 0, $1713 = 0, $1714 = 0, $1715 = 0, $1716 = 0, $1717 = 0, $1718 = 0, $1719 = 0, $172 = 0, $1720 = 0, $1721 = 0, $1722 = 0, $1723 = 0, $1724 = 0, $1725 = 0, $1726 = 0, $1727 = 0, $1728 = 0;
 var $1729 = 0, $173 = 0, $1730 = 0, $1731 = 0, $1732 = 0, $1733 = 0, $1734 = 0, $1735 = 0, $1736 = 0, $1737 = 0, $1738 = 0, $1739 = 0, $174 = 0, $1740 = 0, $1741 = 0, $1742 = 0, $1743 = 0, $1744 = 0, $1745 = 0, $1746 = 0;
 var $1747 = 0, $1748 = 0, $1749 = 0, $175 = 0, $1750 = 0, $1751 = 0, $1752 = 0, $1753 = 0, $1754 = 0, $1755 = 0, $1756 = 0, $1757 = 0, $1758 = 0, $1759 = 0, $176 = 0, $1760 = 0, $1761 = 0, $1762 = 0, $1763 = 0, $1764 = 0;
 var $1765 = 0, $1766 = 0, $1767 = 0, $1768 = 0, $1769 = 0, $177 = 0, $1770 = 0, $1771 = 0, $1772 = 0, $1773 = 0, $1774 = 0, $1775 = 0, $1776 = 0, $1777 = 0, $1778 = 0, $1779 = 0, $178 = 0, $1780 = 0, $1781 = 0, $1782 = 0;
 var $1783 = 0, $1784 = 0, $1785 = 0, $1786 = 0, $1787 = 0, $1788 = 0, $1789 = 0, $179 = 0, $1790 = 0, $1791 = 0, $1792 = 0, $1793 = 0, $1794 = 0, $1795 = 0, $1796 = 0, $1797 = 0, $1798 = 0, $1799 = 0, $18 = 0, $180 = 0;
 var $1800 = 0, $1801 = 0, $1802 = 0, $1803 = 0, $1804 = 0, $1805 = 0, $1806 = 0, $1807 = 0, $1808 = 0, $1809 = 0, $181 = 0, $1810 = 0, $1811 = 0, $1812 = 0, $1813 = 0, $1814 = 0, $1815 = 0, $1816 = 0, $1817 = 0, $1818 = 0;
 var $1819 = 0, $182 = 0, $1820 = 0, $1821 = 0, $1822 = 0, $1823 = 0, $1824 = 0, $1825 = 0, $1826 = 0, $1827 = 0, $1828 = 0, $1829 = 0, $183 = 0, $1830 = 0, $1831 = 0, $1832 = 0, $1833 = 0, $1834 = 0, $1835 = 0, $1836 = 0;
 var $1837 = 0, $1838 = 0, $1839 = 0, $184 = 0, $1840 = 0, $1841 = 0, $1842 = 0, $1843 = 0, $1844 = 0, $1845 = 0, $1846 = 0, $1847 = 0, $1848 = 0, $1849 = 0, $185 = 0, $1850 = 0, $1851 = 0, $1852 = 0, $1853 = 0, $1854 = 0;
 var $1855 = 0, $1856 = 0, $1857 = 0, $1858 = 0, $1859 = 0, $186 = 0, $1860 = 0, $1861 = 0, $1862 = 0, $1863 = 0, $1864 = 0, $1865 = 0, $1866 = 0, $1867 = 0, $1868 = 0, $1869 = 0, $187 = 0, $1870 = 0, $1871 = 0, $1872 = 0;
 var $1873 = 0, $1874 = 0, $1875 = 0, $1876 = 0, $1877 = 0, $1878 = 0, $1879 = 0, $188 = 0, $1880 = 0, $1881 = 0, $1882 = 0, $1883 = 0, $1884 = 0, $1885 = 0, $1886 = 0, $1887 = 0, $1888 = 0, $1889 = 0, $189 = 0, $1890 = 0;
 var $1891 = 0, $1892 = 0, $1893 = 0, $1894 = 0, $1895 = 0, $1896 = 0, $1897 = 0, $1898 = 0, $1899 = 0, $19 = 0, $190 = 0, $1900 = 0, $1901 = 0, $1902 = 0, $1903 = 0, $1904 = 0, $1905 = 0, $1906 = 0, $1907 = 0, $1908 = 0;
 var $1909 = 0, $191 = 0, $1910 = 0, $1911 = 0, $1912 = 0, $1913 = 0, $1914 = 0, $1915 = 0, $1916 = 0, $1917 = 0, $1918 = 0, $1919 = 0, $192 = 0, $1920 = 0, $1921 = 0, $1922 = 0, $1923 = 0, $1924 = 0, $1925 = 0, $1926 = 0;
 var $1927 = 0, $1928 = 0, $1929 = 0, $193 = 0, $1930 = 0, $1931 = 0, $1932 = 0, $1933 = 0, $1934 = 0, $1935 = 0, $1936 = 0, $1937 = 0, $1938 = 0, $1939 = 0, $194 = 0, $1940 = 0, $1941 = 0, $1942 = 0, $1943 = 0, $1944 = 0;
 var $1945 = 0, $1946 = 0, $1947 = 0, $1948 = 0, $1949 = 0, $195 = 0, $1950 = 0, $1951 = 0, $1952 = 0, $1953 = 0, $1954 = 0, $1955 = 0, $1956 = 0, $1957 = 0, $1958 = 0, $1959 = 0, $196 = 0, $1960 = 0, $1961 = 0, $1962 = 0;
 var $1963 = 0, $1964 = 0, $1965 = 0, $1966 = 0, $1967 = 0, $1968 = 0, $1969 = 0, $197 = 0, $1970 = 0, $1971 = 0, $1972 = 0, $1973 = 0, $1974 = 0, $1975 = 0, $1976 = 0, $1977 = 0, $1978 = 0, $1979 = 0, $198 = 0, $1980 = 0;
 var $1981 = 0, $1982 = 0, $1983 = 0, $1984 = 0, $1985 = 0, $1986 = 0, $1987 = 0, $1988 = 0, $1989 = 0, $199 = 0, $1990 = 0, $1991 = 0, $1992 = 0, $1993 = 0, $1994 = 0, $1995 = 0, $1996 = 0, $1997 = 0, $1998 = 0, $1999 = 0;
 var $2 = 0, $20 = 0, $200 = 0, $2000 = 0, $2001 = 0, $2002 = 0, $2003 = 0, $2004 = 0, $2005 = 0, $2006 = 0, $2007 = 0, $2008 = 0, $2009 = 0, $201 = 0, $2010 = 0, $2011 = 0, $2012 = 0, $2013 = 0, $2014 = 0, $2015 = 0;
 var $2016 = 0, $2017 = 0, $2018 = 0, $2019 = 0, $202 = 0, $2020 = 0, $2021 = 0, $2022 = 0, $2023 = 0, $2024 = 0, $2025 = 0, $2026 = 0, $2027 = 0, $2028 = 0, $2029 = 0, $203 = 0, $2030 = 0, $2031 = 0, $2032 = 0, $2033 = 0;
 var $2034 = 0, $2035 = 0, $2036 = 0, $2037 = 0, $2038 = 0, $2039 = 0, $204 = 0, $2040 = 0, $2041 = 0, $2042 = 0, $2043 = 0, $2044 = 0, $2045 = 0, $2046 = 0, $2047 = 0, $2048 = 0, $2049 = 0, $205 = 0, $2050 = 0, $2051 = 0;
 var $2052 = 0, $2053 = 0, $2054 = 0, $2055 = 0, $2056 = 0, $2057 = 0, $2058 = 0, $2059 = 0, $206 = 0, $2060 = 0, $2061 = 0, $2062 = 0, $2063 = 0, $2064 = 0, $2065 = 0, $2066 = 0, $2067 = 0, $2068 = 0, $2069 = 0, $207 = 0;
 var $2070 = 0, $2071 = 0, $2072 = 0, $2073 = 0, $2074 = 0, $2075 = 0, $2076 = 0, $2077 = 0, $2078 = 0, $2079 = 0, $208 = 0, $2080 = 0, $2081 = 0, $2082 = 0, $2083 = 0, $2084 = 0, $2085 = 0, $2086 = 0, $2087 = 0, $2088 = 0;
 var $2089 = 0, $209 = 0, $2090 = 0, $2091 = 0, $2092 = 0, $2093 = 0, $2094 = 0, $2095 = 0, $2096 = 0, $2097 = 0, $2098 = 0, $2099 = 0, $21 = 0, $210 = 0, $2100 = 0, $2101 = 0, $2102 = 0, $2103 = 0, $2104 = 0, $2105 = 0;
 var $2106 = 0, $2107 = 0, $2108 = 0, $2109 = 0, $211 = 0, $2110 = 0, $2111 = 0, $2112 = 0, $2113 = 0, $2114 = 0, $2115 = 0, $2116 = 0, $2117 = 0, $2118 = 0, $2119 = 0, $212 = 0, $2120 = 0, $2121 = 0, $2122 = 0, $2123 = 0;
 var $2124 = 0, $2125 = 0, $2126 = 0, $2127 = 0, $2128 = 0, $2129 = 0, $213 = 0, $2130 = 0, $2131 = 0, $2132 = 0, $2133 = 0, $2134 = 0, $2135 = 0, $2136 = 0, $2137 = 0, $2138 = 0, $2139 = 0, $214 = 0, $2140 = 0, $2141 = 0;
 var $2142 = 0, $2143 = 0, $2144 = 0, $2145 = 0, $2146 = 0, $2147 = 0, $2148 = 0, $2149 = 0, $215 = 0, $2150 = 0, $2151 = 0, $2152 = 0, $2153 = 0, $2154 = 0, $2155 = 0, $2156 = 0, $2157 = 0, $2158 = 0, $2159 = 0, $216 = 0;
 var $2160 = 0, $2161 = 0, $2162 = 0, $2163 = 0, $2164 = 0, $2165 = 0, $2166 = 0, $2167 = 0, $2168 = 0, $2169 = 0, $217 = 0, $2170 = 0, $2171 = 0, $2172 = 0, $2173 = 0, $2174 = 0, $2175 = 0, $2176 = 0, $2177 = 0, $2178 = 0;
 var $2179 = 0, $218 = 0, $2180 = 0, $2181 = 0, $2182 = 0, $2183 = 0, $2184 = 0, $2185 = 0, $2186 = 0, $2187 = 0, $2188 = 0, $2189 = 0, $219 = 0, $2190 = 0, $2191 = 0, $2192 = 0, $2193 = 0, $2194 = 0, $2195 = 0, $2196 = 0;
 var $2197 = 0, $2198 = 0, $2199 = 0, $22 = 0, $220 = 0, $2200 = 0, $2201 = 0, $2202 = 0, $2203 = 0, $2204 = 0, $2205 = 0, $2206 = 0, $2207 = 0, $2208 = 0, $2209 = 0, $221 = 0, $2210 = 0, $2211 = 0, $2212 = 0, $2213 = 0;
 var $2214 = 0, $2215 = 0, $2216 = 0, $2217 = 0, $2218 = 0, $2219 = 0, $222 = 0, $2220 = 0, $2221 = 0, $2222 = 0, $2223 = 0, $2224 = 0, $2225 = 0, $2226 = 0, $2227 = 0, $2228 = 0, $2229 = 0, $223 = 0, $2230 = 0, $2231 = 0;
 var $2232 = 0, $2233 = 0, $2234 = 0, $2235 = 0, $2236 = 0, $2237 = 0, $2238 = 0, $2239 = 0, $224 = 0, $2240 = 0, $2241 = 0, $2242 = 0, $2243 = 0, $2244 = 0, $2245 = 0, $2246 = 0, $2247 = 0, $2248 = 0, $2249 = 0, $225 = 0;
 var $2250 = 0, $2251 = 0, $2252 = 0, $2253 = 0, $2254 = 0, $2255 = 0, $2256 = 0, $2257 = 0, $2258 = 0, $2259 = 0, $226 = 0, $2260 = 0, $2261 = 0, $2262 = 0, $2263 = 0, $2264 = 0, $2265 = 0, $2266 = 0, $2267 = 0, $2268 = 0;
 var $2269 = 0, $227 = 0, $2270 = 0, $2271 = 0, $2272 = 0, $2273 = 0, $2274 = 0, $2275 = 0, $2276 = 0, $2277 = 0, $2278 = 0, $2279 = 0, $228 = 0, $2280 = 0, $2281 = 0, $2282 = 0, $2283 = 0, $2284 = 0, $2285 = 0, $2286 = 0;
 var $2287 = 0, $2288 = 0, $2289 = 0, $229 = 0, $2290 = 0, $2291 = 0, $2292 = 0, $2293 = 0, $2294 = 0, $2295 = 0, $2296 = 0, $2297 = 0, $2298 = 0, $2299 = 0, $23 = 0, $230 = 0, $2300 = 0, $2301 = 0, $2302 = 0, $2303 = 0;
 var $2304 = 0, $2305 = 0, $2306 = 0, $2307 = 0, $2308 = 0, $2309 = 0, $231 = 0, $2310 = 0, $2311 = 0, $2312 = 0, $2313 = 0, $2314 = 0, $2315 = 0, $2316 = 0, $2317 = 0, $2318 = 0, $2319 = 0, $232 = 0, $2320 = 0, $2321 = 0;
 var $2322 = 0, $2323 = 0, $2324 = 0, $2325 = 0, $2326 = 0, $2327 = 0, $2328 = 0, $2329 = 0, $233 = 0, $2330 = 0, $2331 = 0, $2332 = 0, $2333 = 0, $2334 = 0, $2335 = 0, $2336 = 0, $2337 = 0, $2338 = 0, $2339 = 0, $234 = 0;
 var $2340 = 0, $2341 = 0, $2342 = 0, $2343 = 0, $2344 = 0, $2345 = 0, $2346 = 0, $2347 = 0, $2348 = 0, $2349 = 0, $235 = 0, $2350 = 0, $2351 = 0, $2352 = 0, $2353 = 0, $2354 = 0, $2355 = 0, $2356 = 0, $2357 = 0, $2358 = 0;
 var $2359 = 0, $236 = 0, $2360 = 0, $2361 = 0, $2362 = 0, $2363 = 0, $2364 = 0, $2365 = 0, $2366 = 0, $2367 = 0, $2368 = 0, $2369 = 0, $237 = 0, $2370 = 0, $2371 = 0, $2372 = 0, $2373 = 0, $2374 = 0, $2375 = 0, $2376 = 0;
 var $2377 = 0, $2378 = 0, $2379 = 0, $238 = 0, $2380 = 0, $2381 = 0, $2382 = 0, $2383 = 0, $2384 = 0, $2385 = 0, $2386 = 0, $2387 = 0, $2388 = 0, $2389 = 0, $239 = 0, $2390 = 0, $2391 = 0, $2392 = 0, $2393 = 0, $2394 = 0;
 var $2395 = 0, $2396 = 0, $2397 = 0, $2398 = 0, $2399 = 0, $24 = 0, $240 = 0, $2400 = 0, $2401 = 0, $2402 = 0, $2403 = 0, $2404 = 0, $2405 = 0, $2406 = 0, $2407 = 0, $2408 = 0, $2409 = 0, $241 = 0, $2410 = 0, $2411 = 0;
 var $2412 = 0, $2413 = 0, $2414 = 0, $2415 = 0, $2416 = 0, $2417 = 0, $2418 = 0, $2419 = 0, $242 = 0, $2420 = 0, $2421 = 0, $2422 = 0, $2423 = 0, $2424 = 0, $2425 = 0, $2426 = 0, $2427 = 0, $2428 = 0, $2429 = 0, $243 = 0;
 var $2430 = 0, $2431 = 0, $2432 = 0, $2433 = 0, $2434 = 0, $2435 = 0, $2436 = 0, $2437 = 0, $2438 = 0, $2439 = 0, $244 = 0, $2440 = 0, $2441 = 0, $2442 = 0, $2443 = 0, $2444 = 0, $2445 = 0, $2446 = 0, $2447 = 0, $2448 = 0;
 var $2449 = 0, $245 = 0, $2450 = 0, $2451 = 0, $2452 = 0, $2453 = 0, $2454 = 0, $2455 = 0, $2456 = 0, $2457 = 0, $2458 = 0, $2459 = 0, $246 = 0, $2460 = 0, $2461 = 0, $2462 = 0, $2463 = 0, $2464 = 0, $2465 = 0, $2466 = 0;
 var $2467 = 0, $2468 = 0, $2469 = 0, $247 = 0, $2470 = 0, $2471 = 0, $2472 = 0, $2473 = 0, $2474 = 0, $2475 = 0, $2476 = 0, $2477 = 0, $2478 = 0, $2479 = 0, $248 = 0, $2480 = 0, $2481 = 0, $2482 = 0, $2483 = 0, $2484 = 0;
 var $2485 = 0, $2486 = 0, $2487 = 0, $2488 = 0, $2489 = 0, $249 = 0, $2490 = 0, $2491 = 0, $2492 = 0, $2493 = 0, $2494 = 0, $2495 = 0, $2496 = 0, $2497 = 0, $2498 = 0, $2499 = 0, $25 = 0, $250 = 0, $2500 = 0, $2501 = 0;
 var $2502 = 0, $2503 = 0, $2504 = 0, $2505 = 0, $2506 = 0, $2507 = 0, $2508 = 0, $2509 = 0, $251 = 0, $2510 = 0, $2511 = 0, $2512 = 0, $2513 = 0, $2514 = 0, $2515 = 0, $2516 = 0, $2517 = 0, $2518 = 0, $2519 = 0, $252 = 0;
 var $2520 = 0, $2521 = 0, $2522 = 0, $2523 = 0, $2524 = 0, $2525 = 0, $2526 = 0, $2527 = 0, $2528 = 0, $2529 = 0, $253 = 0, $2530 = 0, $2531 = 0, $2532 = 0, $2533 = 0, $2534 = 0, $2535 = 0, $2536 = 0, $2537 = 0, $2538 = 0;
 var $2539 = 0, $254 = 0, $2540 = 0, $2541 = 0, $2542 = 0, $2543 = 0, $2544 = 0, $2545 = 0, $2546 = 0, $2547 = 0, $2548 = 0, $2549 = 0, $255 = 0, $2550 = 0, $2551 = 0, $2552 = 0, $2553 = 0, $2554 = 0, $2555 = 0, $2556 = 0;
 var $2557 = 0, $2558 = 0, $2559 = 0, $256 = 0, $2560 = 0, $2561 = 0, $2562 = 0, $2563 = 0, $2564 = 0, $2565 = 0, $2566 = 0, $2567 = 0, $2568 = 0, $2569 = 0, $257 = 0, $2570 = 0, $2571 = 0, $2572 = 0, $2573 = 0, $2574 = 0;
 var $2575 = 0, $2576 = 0, $2577 = 0, $2578 = 0, $2579 = 0, $258 = 0, $2580 = 0, $2581 = 0, $2582 = 0, $2583 = 0, $2584 = 0, $2585 = 0, $2586 = 0, $2587 = 0, $2588 = 0, $2589 = 0, $259 = 0, $2590 = 0, $2591 = 0, $2592 = 0;
 var $2593 = 0, $2594 = 0, $2595 = 0, $2596 = 0, $2597 = 0, $2598 = 0, $2599 = 0, $26 = 0, $260 = 0, $2600 = 0, $2601 = 0, $2602 = 0, $2603 = 0, $2604 = 0, $2605 = 0, $2606 = 0, $2607 = 0, $2608 = 0, $2609 = 0, $261 = 0;
 var $2610 = 0, $2611 = 0, $2612 = 0, $2613 = 0, $2614 = 0, $2615 = 0, $2616 = 0, $2617 = 0, $2618 = 0, $2619 = 0, $262 = 0, $2620 = 0, $2621 = 0, $2622 = 0, $2623 = 0, $2624 = 0, $2625 = 0, $2626 = 0, $2627 = 0, $2628 = 0;
 var $2629 = 0, $263 = 0, $2630 = 0, $2631 = 0, $2632 = 0, $2633 = 0, $2634 = 0, $2635 = 0, $2636 = 0, $2637 = 0, $2638 = 0, $2639 = 0, $264 = 0, $2640 = 0, $2641 = 0, $2642 = 0, $2643 = 0, $2644 = 0, $2645 = 0, $2646 = 0;
 var $2647 = 0, $2648 = 0, $2649 = 0, $265 = 0, $2650 = 0, $2651 = 0, $2652 = 0, $2653 = 0, $2654 = 0, $2655 = 0, $2656 = 0, $2657 = 0, $2658 = 0, $2659 = 0, $266 = 0, $2660 = 0, $2661 = 0, $2662 = 0, $2663 = 0, $2664 = 0;
 var $2665 = 0, $2666 = 0, $2667 = 0, $2668 = 0, $2669 = 0, $267 = 0, $2670 = 0, $2671 = 0, $2672 = 0, $2673 = 0, $2674 = 0, $2675 = 0, $2676 = 0, $2677 = 0, $2678 = 0, $2679 = 0, $268 = 0, $2680 = 0, $2681 = 0, $2682 = 0;
 var $2683 = 0, $2684 = 0, $2685 = 0, $2686 = 0, $2687 = 0, $2688 = 0, $2689 = 0, $269 = 0, $2690 = 0, $2691 = 0, $2692 = 0, $2693 = 0, $2694 = 0, $2695 = 0, $2696 = 0, $2697 = 0, $2698 = 0, $2699 = 0, $27 = 0, $270 = 0;
 var $2700 = 0, $2701 = 0, $2702 = 0, $2703 = 0, $2704 = 0, $2705 = 0, $2706 = 0, $2707 = 0, $2708 = 0, $2709 = 0, $271 = 0, $2710 = 0, $2711 = 0, $2712 = 0, $2713 = 0, $2714 = 0, $2715 = 0, $2716 = 0, $2717 = 0, $2718 = 0;
 var $2719 = 0, $272 = 0, $2720 = 0, $2721 = 0, $2722 = 0, $2723 = 0, $2724 = 0, $2725 = 0, $2726 = 0, $2727 = 0, $2728 = 0, $2729 = 0, $273 = 0, $2730 = 0, $2731 = 0, $2732 = 0, $2733 = 0, $2734 = 0, $2735 = 0, $2736 = 0;
 var $2737 = 0, $2738 = 0, $2739 = 0, $274 = 0, $2740 = 0, $2741 = 0, $2742 = 0, $2743 = 0, $2744 = 0, $2745 = 0, $2746 = 0, $2747 = 0, $2748 = 0, $2749 = 0, $275 = 0, $2750 = 0, $2751 = 0, $2752 = 0, $2753 = 0, $2754 = 0;
 var $2755 = 0, $2756 = 0, $2757 = 0, $2758 = 0, $2759 = 0, $276 = 0, $2760 = 0, $2761 = 0, $2762 = 0, $2763 = 0, $2764 = 0, $2765 = 0, $2766 = 0, $2767 = 0, $2768 = 0, $2769 = 0, $277 = 0, $2770 = 0, $2771 = 0, $2772 = 0;
 var $2773 = 0, $2774 = 0, $2775 = 0, $2776 = 0, $2777 = 0, $2778 = 0, $2779 = 0, $278 = 0, $2780 = 0, $2781 = 0, $2782 = 0, $2783 = 0, $2784 = 0, $2785 = 0, $2786 = 0, $2787 = 0, $2788 = 0, $2789 = 0, $279 = 0, $2790 = 0;
 var $2791 = 0, $2792 = 0, $2793 = 0, $2794 = 0, $2795 = 0, $2796 = 0, $2797 = 0, $2798 = 0, $2799 = 0, $28 = 0, $280 = 0, $2800 = 0, $2801 = 0, $2802 = 0, $2803 = 0, $2804 = 0, $2805 = 0, $2806 = 0, $2807 = 0, $2808 = 0;
 var $2809 = 0, $281 = 0, $2810 = 0, $2811 = 0, $2812 = 0, $2813 = 0, $2814 = 0, $2815 = 0, $2816 = 0, $2817 = 0, $2818 = 0, $2819 = 0, $282 = 0, $2820 = 0, $2821 = 0, $2822 = 0, $2823 = 0, $2824 = 0, $2825 = 0, $2826 = 0;
 var $2827 = 0, $2828 = 0, $2829 = 0, $283 = 0, $2830 = 0, $2831 = 0, $2832 = 0, $2833 = 0, $2834 = 0, $2835 = 0, $2836 = 0, $2837 = 0, $2838 = 0, $2839 = 0, $284 = 0, $2840 = 0, $2841 = 0, $2842 = 0, $2843 = 0, $2844 = 0;
 var $2845 = 0, $2846 = 0, $2847 = 0, $2848 = 0, $2849 = 0, $285 = 0, $2850 = 0, $2851 = 0, $2852 = 0, $2853 = 0, $2854 = 0, $2855 = 0, $2856 = 0, $2857 = 0, $2858 = 0, $2859 = 0, $286 = 0, $2860 = 0, $2861 = 0, $2862 = 0;
 var $2863 = 0, $2864 = 0, $2865 = 0, $2866 = 0, $2867 = 0, $2868 = 0, $2869 = 0, $287 = 0, $2870 = 0, $2871 = 0, $2872 = 0, $2873 = 0, $2874 = 0, $2875 = 0, $2876 = 0, $2877 = 0, $2878 = 0, $2879 = 0, $288 = 0, $2880 = 0;
 var $2881 = 0, $2882 = 0, $2883 = 0, $2884 = 0, $2885 = 0, $2886 = 0, $2887 = 0, $2888 = 0, $2889 = 0, $289 = 0, $2890 = 0, $2891 = 0, $2892 = 0, $2893 = 0, $2894 = 0, $2895 = 0, $2896 = 0, $2897 = 0, $2898 = 0, $2899 = 0;
 var $29 = 0, $290 = 0, $2900 = 0, $2901 = 0, $2902 = 0, $2903 = 0, $2904 = 0, $2905 = 0, $2906 = 0, $2907 = 0, $2908 = 0, $2909 = 0, $291 = 0, $2910 = 0, $2911 = 0, $2912 = 0, $2913 = 0, $2914 = 0, $2915 = 0, $2916 = 0;
 var $2917 = 0, $2918 = 0, $2919 = 0, $292 = 0, $2920 = 0, $2921 = 0, $2922 = 0, $2923 = 0, $2924 = 0, $2925 = 0, $2926 = 0, $2927 = 0, $2928 = 0, $2929 = 0, $293 = 0, $2930 = 0, $2931 = 0, $2932 = 0, $2933 = 0, $2934 = 0;
 var $2935 = 0, $2936 = 0, $2937 = 0, $2938 = 0, $2939 = 0, $294 = 0, $2940 = 0, $2941 = 0, $2942 = 0, $2943 = 0, $2944 = 0, $2945 = 0, $2946 = 0, $2947 = 0, $2948 = 0, $2949 = 0, $295 = 0, $2950 = 0, $2951 = 0, $2952 = 0;
 var $2953 = 0, $2954 = 0, $2955 = 0, $2956 = 0, $2957 = 0, $2958 = 0, $2959 = 0, $296 = 0, $2960 = 0, $2961 = 0, $2962 = 0, $2963 = 0, $2964 = 0, $2965 = 0, $2966 = 0, $2967 = 0, $2968 = 0, $2969 = 0, $297 = 0, $2970 = 0;
 var $2971 = 0, $2972 = 0, $2973 = 0, $2974 = 0, $2975 = 0, $2976 = 0, $2977 = 0, $2978 = 0, $2979 = 0, $298 = 0, $2980 = 0, $2981 = 0, $2982 = 0, $2983 = 0, $2984 = 0, $2985 = 0, $2986 = 0, $2987 = 0, $2988 = 0, $2989 = 0;
 var $299 = 0, $2990 = 0, $2991 = 0, $2992 = 0, $2993 = 0, $2994 = 0, $2995 = 0, $2996 = 0, $2997 = 0, $2998 = 0, $2999 = 0, $3 = 0, $30 = 0, $300 = 0, $3000 = 0, $3001 = 0, $3002 = 0, $3003 = 0, $3004 = 0, $3005 = 0;
 var $3006 = 0, $3007 = 0, $3008 = 0, $3009 = 0, $301 = 0, $3010 = 0, $3011 = 0, $3012 = 0, $3013 = 0, $3014 = 0, $3015 = 0, $3016 = 0, $3017 = 0, $3018 = 0, $3019 = 0, $302 = 0, $3020 = 0, $3021 = 0, $3022 = 0, $3023 = 0;
 var $3024 = 0, $3025 = 0, $3026 = 0, $3027 = 0, $3028 = 0, $3029 = 0, $303 = 0, $3030 = 0, $3031 = 0, $3032 = 0, $3033 = 0, $3034 = 0, $3035 = 0, $3036 = 0, $3037 = 0, $3038 = 0, $3039 = 0, $304 = 0, $3040 = 0, $3041 = 0;
 var $3042 = 0, $3043 = 0, $3044 = 0, $3045 = 0, $3046 = 0, $3047 = 0, $3048 = 0, $3049 = 0, $305 = 0, $3050 = 0, $3051 = 0, $3052 = 0, $3053 = 0, $3054 = 0, $3055 = 0, $3056 = 0, $3057 = 0, $3058 = 0, $3059 = 0, $306 = 0;
 var $3060 = 0, $3061 = 0, $3062 = 0, $3063 = 0, $3064 = 0, $3065 = 0, $3066 = 0, $3067 = 0, $3068 = 0, $3069 = 0, $307 = 0, $3070 = 0, $3071 = 0, $3072 = 0, $3073 = 0, $3074 = 0, $3075 = 0, $3076 = 0, $3077 = 0, $3078 = 0;
 var $3079 = 0, $308 = 0, $3080 = 0, $3081 = 0, $3082 = 0, $3083 = 0, $3084 = 0, $3085 = 0, $3086 = 0, $3087 = 0, $3088 = 0, $3089 = 0, $309 = 0, $3090 = 0, $3091 = 0, $3092 = 0, $3093 = 0, $3094 = 0, $3095 = 0, $3096 = 0;
 var $3097 = 0, $3098 = 0, $3099 = 0, $31 = 0, $310 = 0, $3100 = 0, $3101 = 0, $3102 = 0, $3103 = 0, $3104 = 0, $3105 = 0, $3106 = 0, $3107 = 0, $3108 = 0, $3109 = 0, $311 = 0, $3110 = 0, $3111 = 0, $3112 = 0, $3113 = 0;
 var $3114 = 0, $3115 = 0, $3116 = 0, $3117 = 0, $3118 = 0, $3119 = 0, $312 = 0, $3120 = 0, $3121 = 0, $3122 = 0, $3123 = 0, $3124 = 0, $3125 = 0, $3126 = 0, $3127 = 0, $3128 = 0, $3129 = 0, $313 = 0, $3130 = 0, $3131 = 0;
 var $3132 = 0, $3133 = 0, $3134 = 0, $3135 = 0, $3136 = 0, $3137 = 0, $3138 = 0, $3139 = 0, $314 = 0, $3140 = 0, $3141 = 0, $3142 = 0, $3143 = 0, $3144 = 0, $3145 = 0, $3146 = 0, $3147 = 0, $3148 = 0, $3149 = 0, $315 = 0;
 var $3150 = 0, $3151 = 0, $3152 = 0, $3153 = 0, $3154 = 0, $3155 = 0, $3156 = 0, $3157 = 0, $3158 = 0, $3159 = 0, $316 = 0, $3160 = 0, $3161 = 0, $3162 = 0, $3163 = 0, $3164 = 0, $3165 = 0, $3166 = 0, $3167 = 0, $3168 = 0;
 var $3169 = 0, $317 = 0, $3170 = 0, $3171 = 0, $3172 = 0, $3173 = 0, $3174 = 0, $3175 = 0, $3176 = 0, $3177 = 0, $3178 = 0, $3179 = 0, $318 = 0, $3180 = 0, $3181 = 0, $3182 = 0, $3183 = 0, $3184 = 0, $3185 = 0, $3186 = 0;
 var $3187 = 0, $3188 = 0, $3189 = 0, $319 = 0, $3190 = 0, $3191 = 0, $3192 = 0, $3193 = 0, $3194 = 0, $3195 = 0, $3196 = 0, $3197 = 0, $3198 = 0, $3199 = 0, $32 = 0, $320 = 0, $3200 = 0, $3201 = 0, $3202 = 0, $3203 = 0;
 var $3204 = 0, $3205 = 0, $3206 = 0, $3207 = 0, $3208 = 0, $3209 = 0, $321 = 0, $3210 = 0, $3211 = 0, $3212 = 0, $3213 = 0, $3214 = 0, $3215 = 0, $3216 = 0, $3217 = 0, $3218 = 0, $3219 = 0, $322 = 0, $3220 = 0, $3221 = 0;
 var $3222 = 0, $3223 = 0, $3224 = 0, $3225 = 0, $3226 = 0, $3227 = 0, $3228 = 0, $3229 = 0, $323 = 0, $3230 = 0, $3231 = 0, $3232 = 0, $3233 = 0, $3234 = 0, $3235 = 0, $3236 = 0, $3237 = 0, $3238 = 0, $3239 = 0, $324 = 0;
 var $3240 = 0, $3241 = 0, $3242 = 0, $3243 = 0, $3244 = 0, $3245 = 0, $3246 = 0, $3247 = 0, $3248 = 0, $3249 = 0, $325 = 0, $3250 = 0, $3251 = 0, $3252 = 0, $3253 = 0, $3254 = 0, $3255 = 0, $3256 = 0, $3257 = 0, $3258 = 0;
 var $3259 = 0, $326 = 0, $3260 = 0, $3261 = 0, $3262 = 0, $3263 = 0, $3264 = 0, $3265 = 0, $3266 = 0, $3267 = 0, $3268 = 0, $3269 = 0, $327 = 0, $3270 = 0, $3271 = 0, $3272 = 0, $3273 = 0, $3274 = 0, $3275 = 0, $3276 = 0;
 var $3277 = 0, $3278 = 0, $3279 = 0, $328 = 0, $3280 = 0, $3281 = 0, $3282 = 0, $3283 = 0, $3284 = 0, $3285 = 0, $3286 = 0, $3287 = 0, $3288 = 0, $3289 = 0, $329 = 0, $3290 = 0, $3291 = 0, $3292 = 0, $3293 = 0, $3294 = 0;
 var $3295 = 0, $3296 = 0, $3297 = 0, $3298 = 0, $3299 = 0, $33 = 0, $330 = 0, $3300 = 0, $3301 = 0, $3302 = 0, $3303 = 0, $3304 = 0, $3305 = 0, $3306 = 0, $3307 = 0, $3308 = 0, $3309 = 0, $331 = 0, $3310 = 0, $3311 = 0;
 var $3312 = 0, $3313 = 0, $3314 = 0, $3315 = 0, $3316 = 0, $3317 = 0, $3318 = 0, $3319 = 0, $332 = 0, $3320 = 0, $3321 = 0, $3322 = 0, $3323 = 0, $3324 = 0, $3325 = 0, $3326 = 0, $3327 = 0, $3328 = 0, $3329 = 0, $333 = 0;
 var $3330 = 0, $3331 = 0, $3332 = 0, $3333 = 0, $3334 = 0, $3335 = 0, $3336 = 0, $3337 = 0, $3338 = 0, $3339 = 0, $334 = 0, $3340 = 0, $3341 = 0, $3342 = 0, $3343 = 0, $3344 = 0, $3345 = 0, $3346 = 0, $3347 = 0, $3348 = 0;
 var $3349 = 0, $335 = 0, $3350 = 0, $3351 = 0, $3352 = 0, $3353 = 0, $3354 = 0, $3355 = 0, $3356 = 0, $3357 = 0, $3358 = 0, $3359 = 0, $336 = 0, $3360 = 0, $3361 = 0, $3362 = 0, $3363 = 0, $3364 = 0, $3365 = 0, $3366 = 0;
 var $3367 = 0, $3368 = 0, $3369 = 0, $337 = 0, $3370 = 0, $3371 = 0, $3372 = 0, $3373 = 0, $3374 = 0, $3375 = 0, $3376 = 0, $3377 = 0, $3378 = 0, $3379 = 0, $338 = 0, $3380 = 0, $3381 = 0, $3382 = 0, $3383 = 0, $3384 = 0;
 var $3385 = 0, $3386 = 0, $3387 = 0, $3388 = 0, $3389 = 0, $339 = 0, $3390 = 0, $3391 = 0, $3392 = 0, $3393 = 0, $3394 = 0, $3395 = 0, $3396 = 0, $3397 = 0, $3398 = 0, $3399 = 0, $34 = 0, $340 = 0, $3400 = 0, $3401 = 0;
 var $3402 = 0, $3403 = 0, $3404 = 0, $3405 = 0, $3406 = 0, $3407 = 0, $3408 = 0, $3409 = 0, $341 = 0, $3410 = 0, $3411 = 0, $3412 = 0, $3413 = 0, $3414 = 0, $3415 = 0, $3416 = 0, $3417 = 0, $3418 = 0, $3419 = 0, $342 = 0;
 var $3420 = 0, $3421 = 0, $3422 = 0, $3423 = 0, $3424 = 0, $3425 = 0, $3426 = 0, $3427 = 0, $3428 = 0, $3429 = 0, $343 = 0, $3430 = 0, $3431 = 0, $3432 = 0, $3433 = 0, $3434 = 0, $3435 = 0, $3436 = 0, $3437 = 0, $3438 = 0;
 var $3439 = 0, $344 = 0, $3440 = 0, $3441 = 0, $3442 = 0, $3443 = 0, $3444 = 0, $3445 = 0, $3446 = 0, $3447 = 0, $3448 = 0, $3449 = 0, $345 = 0, $3450 = 0, $3451 = 0, $3452 = 0, $3453 = 0, $3454 = 0, $3455 = 0, $3456 = 0;
 var $3457 = 0, $3458 = 0, $3459 = 0, $346 = 0, $3460 = 0, $3461 = 0, $3462 = 0, $3463 = 0, $3464 = 0, $3465 = 0, $3466 = 0, $3467 = 0, $3468 = 0, $3469 = 0, $347 = 0, $3470 = 0, $3471 = 0, $3472 = 0, $3473 = 0, $3474 = 0;
 var $3475 = 0, $3476 = 0, $3477 = 0, $3478 = 0, $3479 = 0, $348 = 0, $3480 = 0, $3481 = 0, $3482 = 0, $3483 = 0, $3484 = 0, $3485 = 0, $3486 = 0, $3487 = 0, $3488 = 0, $3489 = 0, $349 = 0, $3490 = 0, $3491 = 0, $3492 = 0;
 var $3493 = 0, $3494 = 0, $3495 = 0, $3496 = 0, $3497 = 0, $3498 = 0, $3499 = 0, $35 = 0, $350 = 0, $3500 = 0, $3501 = 0, $3502 = 0, $3503 = 0, $3504 = 0, $3505 = 0, $3506 = 0, $3507 = 0, $3508 = 0, $3509 = 0, $351 = 0;
 var $3510 = 0, $3511 = 0, $3512 = 0, $3513 = 0, $3514 = 0, $3515 = 0, $3516 = 0, $3517 = 0, $3518 = 0, $3519 = 0, $352 = 0, $3520 = 0, $3521 = 0, $3522 = 0, $3523 = 0, $3524 = 0, $3525 = 0, $3526 = 0, $3527 = 0, $3528 = 0;
 var $3529 = 0, $353 = 0, $3530 = 0, $3531 = 0, $3532 = 0, $3533 = 0, $3534 = 0, $3535 = 0, $3536 = 0, $3537 = 0, $3538 = 0, $3539 = 0, $354 = 0, $3540 = 0, $3541 = 0, $3542 = 0, $3543 = 0, $3544 = 0, $3545 = 0, $3546 = 0;
 var $3547 = 0, $3548 = 0, $3549 = 0, $355 = 0, $3550 = 0, $3551 = 0, $3552 = 0, $3553 = 0, $3554 = 0, $3555 = 0, $3556 = 0, $3557 = 0, $3558 = 0, $3559 = 0, $356 = 0, $3560 = 0, $3561 = 0, $3562 = 0, $3563 = 0, $3564 = 0;
 var $3565 = 0, $3566 = 0, $3567 = 0, $3568 = 0, $3569 = 0, $357 = 0, $3570 = 0, $3571 = 0, $3572 = 0, $3573 = 0, $3574 = 0, $3575 = 0, $3576 = 0, $3577 = 0, $3578 = 0, $3579 = 0, $358 = 0, $3580 = 0, $3581 = 0, $3582 = 0;
 var $3583 = 0, $3584 = 0, $3585 = 0, $3586 = 0, $3587 = 0, $3588 = 0, $3589 = 0, $359 = 0, $3590 = 0, $3591 = 0, $3592 = 0, $3593 = 0, $3594 = 0, $3595 = 0, $3596 = 0, $3597 = 0, $3598 = 0, $3599 = 0, $36 = 0, $360 = 0;
 var $3600 = 0, $3601 = 0, $3602 = 0, $3603 = 0, $3604 = 0, $3605 = 0, $3606 = 0, $3607 = 0, $3608 = 0, $3609 = 0, $361 = 0, $3610 = 0, $3611 = 0, $3612 = 0, $3613 = 0, $3614 = 0, $3615 = 0, $3616 = 0, $3617 = 0, $3618 = 0;
 var $3619 = 0, $362 = 0, $3620 = 0, $3621 = 0, $3622 = 0, $3623 = 0, $3624 = 0, $3625 = 0, $3626 = 0, $3627 = 0, $3628 = 0, $3629 = 0, $363 = 0, $3630 = 0, $3631 = 0, $3632 = 0, $3633 = 0, $3634 = 0, $3635 = 0, $3636 = 0;
 var $3637 = 0, $3638 = 0, $3639 = 0, $364 = 0, $3640 = 0, $3641 = 0, $3642 = 0, $3643 = 0, $3644 = 0, $3645 = 0, $3646 = 0, $3647 = 0, $3648 = 0, $3649 = 0, $365 = 0, $3650 = 0, $3651 = 0, $3652 = 0, $3653 = 0, $3654 = 0;
 var $3655 = 0, $3656 = 0, $3657 = 0, $3658 = 0, $3659 = 0, $366 = 0, $3660 = 0, $3661 = 0, $3662 = 0, $3663 = 0, $3664 = 0, $3665 = 0, $3666 = 0, $3667 = 0, $3668 = 0, $3669 = 0, $367 = 0, $3670 = 0, $3671 = 0, $3672 = 0;
 var $3673 = 0, $3674 = 0, $3675 = 0, $3676 = 0, $3677 = 0, $3678 = 0, $3679 = 0, $368 = 0, $3680 = 0, $3681 = 0, $3682 = 0, $3683 = 0, $3684 = 0, $3685 = 0, $3686 = 0, $3687 = 0, $3688 = 0, $3689 = 0, $369 = 0, $3690 = 0;
 var $3691 = 0, $3692 = 0, $3693 = 0, $3694 = 0, $3695 = 0, $3696 = 0, $3697 = 0, $3698 = 0, $3699 = 0, $37 = 0, $370 = 0, $3700 = 0, $3701 = 0, $3702 = 0, $3703 = 0, $3704 = 0, $3705 = 0, $3706 = 0, $3707 = 0, $3708 = 0;
 var $3709 = 0, $371 = 0, $3710 = 0, $3711 = 0, $3712 = 0, $3713 = 0, $3714 = 0, $3715 = 0, $3716 = 0, $3717 = 0, $3718 = 0, $3719 = 0, $372 = 0, $3720 = 0, $3721 = 0, $3722 = 0, $3723 = 0, $3724 = 0, $3725 = 0, $3726 = 0;
 var $3727 = 0, $3728 = 0, $3729 = 0, $373 = 0, $3730 = 0, $3731 = 0, $3732 = 0, $3733 = 0, $3734 = 0, $3735 = 0, $3736 = 0, $3737 = 0, $3738 = 0, $3739 = 0, $374 = 0, $3740 = 0, $3741 = 0, $3742 = 0, $3743 = 0, $3744 = 0;
 var $3745 = 0, $3746 = 0, $3747 = 0, $3748 = 0, $3749 = 0, $375 = 0, $3750 = 0, $3751 = 0, $3752 = 0, $3753 = 0, $3754 = 0, $3755 = 0, $3756 = 0, $3757 = 0, $3758 = 0, $3759 = 0, $376 = 0, $3760 = 0, $3761 = 0, $3762 = 0;
 var $3763 = 0, $3764 = 0, $3765 = 0, $3766 = 0, $3767 = 0, $3768 = 0, $3769 = 0, $377 = 0, $3770 = 0, $3771 = 0, $3772 = 0, $3773 = 0, $3774 = 0, $3775 = 0, $3776 = 0, $3777 = 0, $3778 = 0, $3779 = 0, $378 = 0, $3780 = 0;
 var $3781 = 0, $3782 = 0, $3783 = 0, $3784 = 0, $3785 = 0, $3786 = 0, $3787 = 0, $3788 = 0, $3789 = 0, $379 = 0, $3790 = 0, $3791 = 0, $3792 = 0, $3793 = 0, $3794 = 0, $3795 = 0, $3796 = 0, $3797 = 0, $3798 = 0, $3799 = 0;
 var $38 = 0, $380 = 0, $3800 = 0, $3801 = 0, $3802 = 0, $3803 = 0, $3804 = 0, $3805 = 0, $3806 = 0, $3807 = 0, $3808 = 0, $3809 = 0, $381 = 0, $3810 = 0, $3811 = 0, $3812 = 0, $3813 = 0, $3814 = 0, $3815 = 0, $3816 = 0;
 var $3817 = 0, $3818 = 0, $3819 = 0, $382 = 0, $3820 = 0, $3821 = 0, $3822 = 0, $3823 = 0, $3824 = 0, $3825 = 0, $3826 = 0, $3827 = 0, $3828 = 0, $3829 = 0, $383 = 0, $3830 = 0, $3831 = 0, $3832 = 0, $3833 = 0, $3834 = 0;
 var $3835 = 0, $3836 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0;
 var $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0;
 var $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0;
 var $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0;
 var $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0;
 var $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0;
 var $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0;
 var $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0;
 var $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0;
 var $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0;
 var $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0;
 var $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0;
 var $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0;
 var $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0;
 var $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0;
 var $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0;
 var $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0;
 var $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0;
 var $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0;
 var $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0;
 var $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0;
 var $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0;
 var $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0;
 var $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0;
 var $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0;
 var $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0;
 var $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0;
 var $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0;
 var $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0;
 var $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0;
 var $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0;
 var $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0;
 var $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0;
 var $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0;
 var $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $RLE_count$0194$i = 0, $RLE_count$1165$i = 0, $RLE_count$1166$i = 0, $RLE_repeating$0195$i = 0, $RLE_repeating$1167$i = 0, $RLE_repeating$1168$i = 0, $acount$0$i = 0, $all_a$1$i = 0, $all_a$2331$i = 0, $all_a$3325$i = 0, $all_a$4319$i = 0, $all_a$5$i = 0, $all_a$7$i = 0, $ashift$0$i = 0, $bcount$0$i = 0;
 var $bshift$0$i = 0, $coutput$i$i = 0, $data$i$i$i$i = 0, $data3$i$i$i$i = 0, $div$i = 0, $div14$i = 0, $easy$0287$i = 0, $exitcond = 0, $exitcond$i = 0, $exitcond$i$i = 0, $exitcond$i$us$i$i$i$i = 0, $exitcond10$i$i$i$i = 0, $exitcond11$i$i$i$i = 0, $exitcond207$i = 0, $exitcond208$i = 0, $exitcond280 = 0, $gcount$0$i = 0, $gshift$0$i = 0, $i$0$i$i$i$i = 0, $i$01$i$i$i = 0;
 var $i$01$i$us$i$i$i$i = 0, $i$01$i12$i$i = 0, $i$01$i6$i$i = 0, $i$01$i9$i$i = 0, $i$0133$i$i$i$i = 0, $i$02$us$i$i$i$i = 0, $i$0200$i = 0, $i$0200$us$i = 0, $i$069$i$i = 0, $i$1192$i = 0, $i$1315$i = 0, $i$167$i$i = 0, $i$2178$us$i = 0, $i$2307$i = 0, $i$2307$i$us = 0, $i$264$i$i = 0, $i$3176$i = 0, $i$3324$i = 0, $i$4318$i = 0, $i$5304$i = 0;
 var $i$6301$us$i = 0, $i1$0145$i$i$i$i = 0, $i17$0164$i$i$i$i = 0, $i8$0150$i$i$i$i = 0, $index1$0179$us$i = 0, $index2$0180$us$i = 0, $ispos$i = 0, $j$0134$i$i$i$i = 0, $j$03$us$i$i$i$i = 0, $j$0313$i = 0, $j$072$i$i = 0, $j$1184$i = 0, $j$1330$i = 0, $j$2302$us$i = 0, $j$3182$us$i = 0, $j$i = 0, $j18$0167$i$i$i$i = 0, $j2$0148$i$i$i$i = 0, $j9$0153$i$i$i$i = 0, $k$0$i$i$i$i$i = 0;
 var $k$0142$i$i$i$i = 0, $k$075$i$i = 0, $k$163$i$i = 0, $k$4$ph$i$i$i$i$i = 0, $k$5165$i$i$i$i$i = 0, $k$6$i$i$i$i$i = 0, $k$7$ph163$i$i$i$i$i = 0, $k$7160$i$i$i$i$i = 0, $k$7160$i$i$i$i$i$lcssa = 0, $k$9$ph$i$i$i$i$i = 0, $k19$0161$i$i$i$i = 0, $ma$0$i = 0, $ma$1$i = 0, $mb$0$i = 0, $mb$1$i = 0, $mg$0$i = 0, $mg$1$i = 0, $mr$0$i = 0, $mr$1$i = 0, $n$06$i$i$i$i = 0;
 var $n$1$i$i = 0, $n$1$i220$i = 0, $n$1$i238$i = 0, $n$1$i263$i = 0, $n$2$i$i = 0, $n$2$i222$i = 0, $n$2$i240$i = 0, $n$2$i265$i = 0, $n$3$i$i = 0, $n$3$i224$i = 0, $n$3$i242$i = 0, $n$3$i267$i = 0, $neg$i = 0, $or$cond$i = 0, $or$cond$i144 = 0, $or$cond$i25 = 0, $or$cond13$i = 0, $or$cond17$i = 0, $or$cond18$i = 0, $or$cond20$i = 0;
 var $or$cond22$i = 0, $or$cond233$i = 0, $or$cond24$i = 0, $or$cond26$i = 0, $or$cond28$i = 0, $or$cond3$i$i = 0, $or$cond30$i = 0, $or$cond40$i = 0, $or$cond6$i$i$i$i = 0, $or$cond7$i$i$i$i = 0, $out$0$i = 0, $out$070$i$i = 0, $out$165$i$i = 0, $pal$i = 0, $phitmp = 0, $psize$0$i = 0, $r3$0$ph$i$i$i$i$i = 0, $r3$0$ph141162$i$i$i$i$i = 0, $raw_data$i = 0, $rcount$0$i = 0;
 var $read_next_pixel$0196$i = 0, $req_comp$$i = 0, $res_comp$i$i = 0, $result$0$lcssa$i186$i = 0, $result$0$lcssa$i193$i = 0, $result$0$lcssa$i200$i = 0, $result$01$i$i = 0, $result$01$i184$i = 0, $result$01$i191$i = 0, $result$01$i198$i = 0, $rshift$0$i = 0, $s4$0$ph$i$i$i$i$i = 0, $scevgep$i = 0, $scevgep211$i = 0, $sext$i$i$i$i$i = 0, $sext11$i$i$i$i$i = 0, $smax206$i = 0, $smax212$i = 0, $tga_comp$0$i = 0, $tga_image_type$0$off$i = 0;
 var $tga_palette$0$i = 0, $tga_pixel$0177$i = 0, $v$0$i = 0, $v$0$i$us = 0, $v2$0$i = 0, $v2$0$i$us = 0, $vararg_buffer = 0, $which$0$lcssa$i$i$i$i = 0, $which$041$i$i$i$i = 0, $width$0$i = 0, $x$0136$i$i$i$i = 0, $x20$0155$i$i$i$i = 0, $y$0139$i$i$i$i = 0, $y21$0158$i$i$i$i = 0, $z$02$i$i = 0, $z$02$i183$i = 0, $z$02$i190$i = 0, $z$02$i197$i = 0, $z$0314$i = 0, $z$1308$i = 0;
 var $z$1308$i$us = 0, $z$5$i = 0, $z2$0332$i = 0, $z2$1326$i = 0, $z2$2$i = 0, $z2$3320$i = 0, $z2$4$i = 0, $z2$5$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 19632|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $raw_data$i = sp + 19616|0;
 $pal$i = sp + 18592|0;
 $data$i$i$i$i = sp + 18464|0;
 $data3$i$i$i$i = sp + 18336|0;
 $coutput$i$i = sp + 18320|0;
 $res_comp$i$i = sp + 18192|0;
 $j$i = sp + 4|0;
 $0 = ((($s)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
 $1 = HEAP32[$0>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
 $2 = ((($s)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
 $3 = HEAP32[$2>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
 $4 = ($1>>>0)<($3>>>0); //@line 1229 "./libraries/stb/stb_image.h"
 if ($4) {
  $5 = ((($1)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
  HEAP32[$0>>2] = $5; //@line 1230 "./libraries/stb/stb_image.h"
  $6 = HEAP8[$1>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
  $$0$i$i$i$i = $6;$3701 = $3;$3702 = $5;
  label = 8;
 } else {
  $7 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
  $8 = HEAP32[$7>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
  $9 = ($8|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
  if ($9) {
   label = 18;
  } else {
   $10 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
   $11 = HEAP32[$10>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $12 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
   $13 = HEAP32[$12>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $14 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
   $15 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
   $16 = HEAP32[$15>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $17 = (FUNCTION_TABLE_iiii[$11 & 8191]($13,$14,$16)|0); //@line 1213 "./libraries/stb/stb_image.h"
   $18 = ($17|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
   if ($18) {
    HEAP32[$7>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
    $19 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
    HEAP8[$14>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
    $$pre$phi$i$i$i$iZ2D = $19;$$sink$i$i$i$i$i = $19;$3703 = 0;
   } else {
    $20 = (((($s)) + 40|0) + ($17)|0); //@line 1223 "./libraries/stb/stb_image.h"
    $$pre$i$i$i$i = HEAP8[$14>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
    $$pre1$i$i$i$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
    $$pre$phi$i$i$i$iZ2D = $$pre1$i$i$i$i;$$sink$i$i$i$i$i = $20;$3703 = $$pre$i$i$i$i;
   }
   HEAP32[$2>>2] = $$sink$i$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
   HEAP32[$0>>2] = $$pre$phi$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
   $$0$i$i$i$i = $3703;$3701 = $$sink$i$i$i$i$i;$3702 = $$pre$phi$i$i$i$iZ2D;
   label = 8;
  }
 }
 L9: do {
  if ((label|0) == 8) {
   $21 = ($$0$i$i$i$i<<24>>24)==(-1); //@line 2424 "./libraries/stb/stb_image.h"
   if ($21) {
    $22 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $23 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
    $24 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
    $25 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
    $26 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
    $27 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
    $29 = $3702;$30 = $3701;
    while(1) {
     $28 = ($29>>>0)<($30>>>0); //@line 1229 "./libraries/stb/stb_image.h"
     if ($28) {
      $31 = ((($29)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
      HEAP32[$0>>2] = $31; //@line 1230 "./libraries/stb/stb_image.h"
      $32 = HEAP8[$29>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
      $$0$i3$i$i = $32;$3704 = $30;$3705 = $31;
     } else {
      $33 = HEAP32[$22>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
      $34 = ($33|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
      if ($34) {
       label = 18;
       break L9;
      }
      $35 = HEAP32[$23>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $36 = HEAP32[$24>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $37 = HEAP32[$26>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $38 = (FUNCTION_TABLE_iiii[$35 & 8191]($36,$25,$37)|0); //@line 1213 "./libraries/stb/stb_image.h"
      $39 = ($38|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
      if ($39) {
       HEAP32[$22>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
       HEAP8[$25>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
       $$sink$i$i$i$i = $27;$3706 = 0;
      } else {
       $40 = (((($s)) + 40|0) + ($38)|0); //@line 1223 "./libraries/stb/stb_image.h"
       $$pre$i$i$i = HEAP8[$25>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
       $$sink$i$i$i$i = $40;$3706 = $$pre$i$i$i;
      }
      HEAP32[$2>>2] = $$sink$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
      HEAP32[$0>>2] = $27; //@line 1233 "./libraries/stb/stb_image.h"
      $$0$i3$i$i = $3706;$3704 = $$sink$i$i$i$i;$3705 = $27;
     }
     switch ($$0$i3$i$i<<24>>24) {
     case -40:  {
      $$0$i$i = 0;
      break L9;
      break;
     }
     case -1:  {
      $29 = $3705;$30 = $3704;
      break;
     }
     default: {
      label = 18;
      break L9;
     }
     }
    }
   } else {
    label = 18;
   }
  }
 } while(0);
 if ((label|0) == 18) {
  HEAP32[158754] = 440708; //@line 894 "./libraries/stb/stb_image.h"
  $$0$i$i = 1;
 }
 $41 = ((($s)) + 176|0); //@line 826 "./libraries/stb/stb_image.h"
 $42 = HEAP32[$41>>2]|0; //@line 826 "./libraries/stb/stb_image.h"
 HEAP32[$0>>2] = $42; //@line 826 "./libraries/stb/stb_image.h"
 $43 = ((($s)) + 180|0); //@line 827 "./libraries/stb/stb_image.h"
 $44 = HEAP32[$43>>2]|0; //@line 827 "./libraries/stb/stb_image.h"
 HEAP32[$2>>2] = $44; //@line 827 "./libraries/stb/stb_image.h"
 $45 = $42; //@line 1229 "./libraries/stb/stb_image.h"
 $46 = $44; //@line 1229 "./libraries/stb/stb_image.h"
 if (!($$0$i$i)) {
  HEAP32[$j$i>>2] = $s; //@line 3412 "./libraries/stb/stb_image.h"
  $47 = ((($j$i)) + 18176|0); //@line 3257 "./libraries/stb/stb_image.h"
  HEAP32[$47>>2] = 4175; //@line 3257 "./libraries/stb/stb_image.h"
  $48 = ((($j$i)) + 18180|0); //@line 3258 "./libraries/stb/stb_image.h"
  HEAP32[$48>>2] = 4176; //@line 3258 "./libraries/stb/stb_image.h"
  $49 = ((($j$i)) + 18184|0); //@line 3259 "./libraries/stb/stb_image.h"
  HEAP32[$49>>2] = 4177; //@line 3259 "./libraries/stb/stb_image.h"
  $50 = ((($s)) + 8|0); //@line 3315 "./libraries/stb/stb_image.h"
  HEAP32[$50>>2] = 0; //@line 3315 "./libraries/stb/stb_image.h"
  $51 = ($req_comp>>>0)>(4); //@line 3318 "./libraries/stb/stb_image.h"
  L28: do {
   if ($51) {
    HEAP32[158754] = 441030; //@line 894 "./libraries/stb/stb_image.h"
    $$3$i$i = 0;
   } else {
    $52 = ((($j$i)) + 17868|0); //@line 2820 "./libraries/stb/stb_image.h"
    HEAP32[$52>>2] = 0; //@line 2820 "./libraries/stb/stb_image.h"
    $53 = ((($j$i)) + 17872|0); //@line 2821 "./libraries/stb/stb_image.h"
    HEAP32[$53>>2] = 0; //@line 2821 "./libraries/stb/stb_image.h"
    $54 = ((($j$i)) + 17940|0); //@line 2820 "./libraries/stb/stb_image.h"
    HEAP32[$54>>2] = 0; //@line 2820 "./libraries/stb/stb_image.h"
    $55 = ((($j$i)) + 17944|0); //@line 2821 "./libraries/stb/stb_image.h"
    HEAP32[$55>>2] = 0; //@line 2821 "./libraries/stb/stb_image.h"
    $56 = ((($j$i)) + 18012|0); //@line 2820 "./libraries/stb/stb_image.h"
    HEAP32[$56>>2] = 0; //@line 2820 "./libraries/stb/stb_image.h"
    $57 = ((($j$i)) + 18016|0); //@line 2821 "./libraries/stb/stb_image.h"
    HEAP32[$57>>2] = 0; //@line 2821 "./libraries/stb/stb_image.h"
    $58 = ((($j$i)) + 18084|0); //@line 2820 "./libraries/stb/stb_image.h"
    HEAP32[$58>>2] = 0; //@line 2820 "./libraries/stb/stb_image.h"
    $59 = ((($j$i)) + 18088|0); //@line 2821 "./libraries/stb/stb_image.h"
    HEAP32[$59>>2] = 0; //@line 2821 "./libraries/stb/stb_image.h"
    $60 = ((($j$i)) + 18168|0); //@line 2823 "./libraries/stb/stb_image.h"
    HEAP32[$60>>2] = 0; //@line 2823 "./libraries/stb/stb_image.h"
    $61 = (__ZL24stbi__decode_jpeg_headerP10stbi__jpegi($j$i,0)|0); //@line 2824 "./libraries/stb/stb_image.h"
    $62 = ($61|0)==(0); //@line 2824 "./libraries/stb/stb_image.h"
    L31: do {
     if (!($62)) {
      $63 = ((($j$i)) + 18116|0); //@line 2422 "./libraries/stb/stb_image.h"
      $64 = HEAP8[$63>>0]|0; //@line 2422 "./libraries/stb/stb_image.h"
      $65 = ($64<<24>>24)==(-1); //@line 2422 "./libraries/stb/stb_image.h"
      L33: do {
       if ($65) {
        $66 = HEAP32[$j$i>>2]|0; //@line 2423 "./libraries/stb/stb_image.h"
        $67 = ((($66)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
        $68 = HEAP32[$67>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
        $69 = ((($66)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
        $70 = HEAP32[$69>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
        $71 = ($68>>>0)<($70>>>0); //@line 1229 "./libraries/stb/stb_image.h"
        if ($71) {
         $72 = ((($68)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
         HEAP32[$67>>2] = $72; //@line 1230 "./libraries/stb/stb_image.h"
         $73 = HEAP8[$68>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
         $$0$i$i$i$i$i = $73;
        } else {
         $74 = ((($66)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
         $75 = HEAP32[$74>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
         $76 = ($75|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
         if ($76) {
          $$sink$ph88$i$i$i = -1;
          break;
         }
         $77 = ((($66)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
         $78 = HEAP32[$77>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $79 = ((($66)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
         $80 = HEAP32[$79>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $81 = ((($66)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
         $82 = ((($66)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
         $83 = HEAP32[$82>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $84 = (FUNCTION_TABLE_iiii[$78 & 8191]($80,$81,$83)|0); //@line 1213 "./libraries/stb/stb_image.h"
         $85 = ($84|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
         if ($85) {
          HEAP32[$74>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
          $86 = ((($66)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
          HEAP8[$81>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
          $$pre$phi$i$i$i$i$iZ2D = $86;$$sink$i$i$i$i$i$i = $86;$3707 = 0;
         } else {
          $87 = (((($66)) + 40|0) + ($84)|0); //@line 1223 "./libraries/stb/stb_image.h"
          $$pre$i$i$i$i$i = HEAP8[$81>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
          $$pre1$i$i$i$i$i = ((($66)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
          $$pre$phi$i$i$i$i$iZ2D = $$pre1$i$i$i$i$i;$$sink$i$i$i$i$i$i = $87;$3707 = $$pre$i$i$i$i$i;
         }
         HEAP32[$69>>2] = $$sink$i$i$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
         HEAP32[$67>>2] = $$pre$phi$i$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
         $$0$i$i$i$i$i = $3707;
        }
        $88 = ($$0$i$i$i$i$i<<24>>24)==(-1); //@line 2424 "./libraries/stb/stb_image.h"
        if ($88) {
         while(1) {
          $89 = HEAP32[$j$i>>2]|0; //@line 2426 "./libraries/stb/stb_image.h"
          $90 = ((($89)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
          $91 = HEAP32[$90>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $92 = ((($89)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
          $93 = HEAP32[$92>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $94 = ($91>>>0)<($93>>>0); //@line 1229 "./libraries/stb/stb_image.h"
          if ($94) {
           $95 = ((($91)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
           HEAP32[$90>>2] = $95; //@line 1230 "./libraries/stb/stb_image.h"
           $96 = HEAP8[$91>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
           $$0$i2$i$i$i = $96;
          } else {
           $97 = ((($89)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $98 = HEAP32[$97>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
           $99 = ($98|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
           if ($99) {
            $$sink$ph88$i$i$i = 0;
            break L33;
           }
           $100 = ((($89)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $101 = HEAP32[$100>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $102 = ((($89)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $103 = HEAP32[$102>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $104 = ((($89)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $105 = ((($89)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $106 = HEAP32[$105>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $107 = (FUNCTION_TABLE_iiii[$101 & 8191]($103,$104,$106)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $108 = ($107|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($108) {
            HEAP32[$97>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $109 = ((($89)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$104>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i$i$i$i3Z2D = $109;$$sink$i$i$i$i$i4 = $109;$3708 = 0;
           } else {
            $110 = (((($89)) + 40|0) + ($107)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre$i$i$i$i1 = HEAP8[$104>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
            $$pre1$i$i$i$i2 = ((($89)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i$i$i$i3Z2D = $$pre1$i$i$i$i2;$$sink$i$i$i$i$i4 = $110;$3708 = $$pre$i$i$i$i1;
           }
           HEAP32[$92>>2] = $$sink$i$i$i$i$i4; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$90>>2] = $$pre$phi$i$i$i$i3Z2D; //@line 1233 "./libraries/stb/stb_image.h"
           $$0$i2$i$i$i = $3708;
          }
          $111 = ($$0$i2$i$i$i<<24>>24)==(-1); //@line 2425 "./libraries/stb/stb_image.h"
          if (!($111)) {
           $$sink$ph88$i$i$i = $$0$i2$i$i$i;
           break;
          }
         }
        } else {
         $$sink$ph88$i$i$i = -1;
        }
       } else {
        HEAP8[$63>>0] = -1; //@line 2422 "./libraries/stb/stb_image.h"
        $$sink$ph88$i$i$i = $64;
       }
      } while(0);
      $112 = ((($j$i)) + 18148|0); //@line 2667 "./libraries/stb/stb_image.h"
      $113 = ((($j$i)) + 18128|0); //@line 2684 "./libraries/stb/stb_image.h"
      $114 = ((($j$i)) + 18132|0); //@line 2685 "./libraries/stb/stb_image.h"
      $115 = ((($j$i)) + 18136|0); //@line 2687 "./libraries/stb/stb_image.h"
      $116 = ((($j$i)) + 18140|0); //@line 2688 "./libraries/stb/stb_image.h"
      $117 = ((($j$i)) + 18124|0); //@line 2689 "./libraries/stb/stb_image.h"
      $118 = ((($j$i)) + 18112|0); //@line 2438 "./libraries/stb/stb_image.h"
      $119 = ((($j$i)) + 18108|0); //@line 2439 "./libraries/stb/stb_image.h"
      $120 = ((($j$i)) + 18120|0); //@line 2440 "./libraries/stb/stb_image.h"
      $121 = ((($j$i)) + 17988|0); //@line 2441 "./libraries/stb/stb_image.h"
      $122 = ((($j$i)) + 17916|0); //@line 2441 "./libraries/stb/stb_image.h"
      $123 = ((($j$i)) + 17844|0); //@line 2441 "./libraries/stb/stb_image.h"
      $124 = ((($j$i)) + 18172|0); //@line 2443 "./libraries/stb/stb_image.h"
      $125 = ((($j$i)) + 18144|0); //@line 2444 "./libraries/stb/stb_image.h"
      $126 = ((($j$i)) + 18152|0); //@line 2456 "./libraries/stb/stb_image.h"
      $127 = ((($j$i)) + 17808|0); //@line 2482 "./libraries/stb/stb_image.h"
      $128 = ((($j$i)) + 17804|0); //@line 2483 "./libraries/stb/stb_image.h"
      $$sink$i$i$i = $$sink$ph88$i$i$i;
      L57: while(1) {
       $129 = $$sink$i$i$i&255; //@line 2825 "./libraries/stb/stb_image.h"
       L59: do {
        switch ($129|0) {
        case 217:  {
         label = 492;
         break L57;
         break;
        }
        case 218:  {
         $130 = HEAP32[$j$i>>2]|0; //@line 2666 "./libraries/stb/stb_image.h"
         $131 = ((($130)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
         $132 = HEAP32[$131>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $133 = ((($130)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
         $134 = HEAP32[$133>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $135 = ($132>>>0)<($134>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         if ($135) {
          $136 = ((($132)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
          HEAP32[$131>>2] = $136; //@line 1230 "./libraries/stb/stb_image.h"
          $137 = HEAP8[$132>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
          $$0$i$i$i$i$i$i = $137;$155 = $136;$156 = $134;
         } else {
          $138 = ((($130)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
          $139 = HEAP32[$138>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
          $140 = ($139|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
          if ($140) {
           $$0$i$i$i$i$i$i = 0;$155 = $132;$156 = $134;
          } else {
           $141 = ((($130)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $142 = HEAP32[$141>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $143 = ((($130)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $144 = HEAP32[$143>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $145 = ((($130)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $146 = ((($130)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $147 = HEAP32[$146>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $148 = (FUNCTION_TABLE_iiii[$142 & 8191]($144,$145,$147)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $149 = ($148|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($149) {
            HEAP32[$138>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $150 = ((($130)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$145>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i$i$i$i$i$iZ2D = $150;$$sink$i$i$i$i$i$i$i = $150;$3709 = 0;
           } else {
            $151 = (((($130)) + 40|0) + ($148)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre$i$i$i$i$i$i = HEAP8[$145>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
            $$pre1$i$i$i$i$i$i = ((($130)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i$i$i$i$i$iZ2D = $$pre1$i$i$i$i$i$i;$$sink$i$i$i$i$i$i$i = $151;$3709 = $$pre$i$i$i$i$i$i;
           }
           HEAP32[$133>>2] = $$sink$i$i$i$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$131>>2] = $$pre$phi$i$i$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
           $$0$i$i$i$i$i$i = $3709;$155 = $$pre$phi$i$i$i$i$i$iZ2D;$156 = $$sink$i$i$i$i$i$i$i;
          }
         }
         $152 = $$0$i$i$i$i$i$i&255; //@line 1293 "./libraries/stb/stb_image.h"
         $153 = $152 << 8; //@line 1294 "./libraries/stb/stb_image.h"
         $154 = ($155>>>0)<($156>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         if ($154) {
          $157 = ((($155)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
          HEAP32[$131>>2] = $157; //@line 1230 "./libraries/stb/stb_image.h"
          $158 = HEAP8[$155>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
          $$0$i6$i$i$i$i$i = $158;
         } else {
          $159 = ((($130)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
          $160 = HEAP32[$159>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
          $161 = ($160|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
          if ($161) {
           $$0$i6$i$i$i$i$i = 0;
          } else {
           $162 = ((($130)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $163 = HEAP32[$162>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $164 = ((($130)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $165 = HEAP32[$164>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $166 = ((($130)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $167 = ((($130)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $168 = HEAP32[$167>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $169 = (FUNCTION_TABLE_iiii[$163 & 8191]($165,$166,$168)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $170 = ($169|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($170) {
            HEAP32[$159>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $171 = ((($130)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$166>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i3$i$i$i$i$iZ2D = $171;$$sink$i$i4$i$i$i$i$i = $171;$3710 = 0;
           } else {
            $172 = (((($130)) + 40|0) + ($169)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre$i1$i$i$i$i$i = HEAP8[$166>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
            $$pre1$i2$i$i$i$i$i = ((($130)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i3$i$i$i$i$iZ2D = $$pre1$i2$i$i$i$i$i;$$sink$i$i4$i$i$i$i$i = $172;$3710 = $$pre$i1$i$i$i$i$i;
           }
           HEAP32[$133>>2] = $$sink$i$i4$i$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$131>>2] = $$pre$phi$i3$i$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
           $$0$i6$i$i$i$i$i = $3710;
          }
         }
         $173 = $$0$i6$i$i$i$i$i&255; //@line 1294 "./libraries/stb/stb_image.h"
         $174 = $173 | $153; //@line 1294 "./libraries/stb/stb_image.h"
         $175 = HEAP32[$j$i>>2]|0; //@line 2667 "./libraries/stb/stb_image.h"
         $176 = ((($175)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
         $177 = HEAP32[$176>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $178 = ((($175)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
         $179 = HEAP32[$178>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $180 = ($177>>>0)<($179>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         if ($180) {
          $181 = ((($177)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
          HEAP32[$176>>2] = $181; //@line 1230 "./libraries/stb/stb_image.h"
          $182 = HEAP8[$177>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
          $$0$i$i8$i$i$i = $182;
         } else {
          $183 = ((($175)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
          $184 = HEAP32[$183>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
          $185 = ($184|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
          if ($185) {
           label = 60;
           break L57;
          }
          $186 = ((($175)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
          $187 = HEAP32[$186>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
          $188 = ((($175)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
          $189 = HEAP32[$188>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
          $190 = ((($175)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
          $191 = ((($175)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
          $192 = HEAP32[$191>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
          $193 = (FUNCTION_TABLE_iiii[$187 & 8191]($189,$190,$192)|0); //@line 1213 "./libraries/stb/stb_image.h"
          $194 = ($193|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
          if ($194) {
           HEAP32[$183>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
           $195 = ((($175)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
           HEAP8[$190>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
           $$pre$phi$i$i5$i$i$iZ2D = $195;$$sink$i$i$i6$i$i$i = $195;$3711 = 0;
          } else {
           $196 = (((($175)) + 40|0) + ($193)|0); //@line 1223 "./libraries/stb/stb_image.h"
           $$pre$i$i3$i$i$i = HEAP8[$190>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
           $$pre1$i$i4$i$i$i = ((($175)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
           $$pre$phi$i$i5$i$i$iZ2D = $$pre1$i$i4$i$i$i;$$sink$i$i$i6$i$i$i = $196;$3711 = $$pre$i$i3$i$i$i;
          }
          HEAP32[$178>>2] = $$sink$i$i$i6$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
          HEAP32[$176>>2] = $$pre$phi$i$i5$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
          $$0$i$i8$i$i$i = $3711;
         }
         $197 = $$0$i$i8$i$i$i&255; //@line 2667 "./libraries/stb/stb_image.h"
         HEAP32[$112>>2] = $197; //@line 2667 "./libraries/stb/stb_image.h"
         $198 = (($$0$i$i8$i$i$i) + -1)<<24>>24; //@line 2668 "./libraries/stb/stb_image.h"
         $199 = ($198&255)>(3); //@line 2668 "./libraries/stb/stb_image.h"
         if ($199) {
          break L57;
         }
         $200 = HEAP32[$j$i>>2]|0; //@line 2668 "./libraries/stb/stb_image.h"
         $201 = ((($200)) + 8|0); //@line 2668 "./libraries/stb/stb_image.h"
         $202 = HEAP32[$201>>2]|0; //@line 2668 "./libraries/stb/stb_image.h"
         $203 = ($197|0)>($202|0); //@line 2668 "./libraries/stb/stb_image.h"
         if ($203) {
          break L57;
         }
         $204 = $197 << 1; //@line 2669 "./libraries/stb/stb_image.h"
         $205 = (($204) + 6)|0; //@line 2669 "./libraries/stb/stb_image.h"
         $206 = ($174|0)==($205|0); //@line 2669 "./libraries/stb/stb_image.h"
         if ($206) {
          $208 = $197;$210 = $200;$i$0$i$i$i$i = 0;
         } else {
          label = 69;
          break L57;
         }
         while(1) {
          $207 = ($i$0$i$i$i$i|0)<($208|0); //@line 2670 "./libraries/stb/stb_image.h"
          $209 = ((($210)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
          $211 = HEAP32[$209>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $212 = ((($210)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
          $213 = HEAP32[$212>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $214 = ($211>>>0)<($213>>>0); //@line 1229 "./libraries/stb/stb_image.h"
          if ($214) {
           $215 = ((($211)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
           HEAP32[$209>>2] = $215; //@line 1230 "./libraries/stb/stb_image.h"
           $216 = HEAP8[$211>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
           $$0$i21$i$i$i$i = $216;
          } else {
           $217 = ((($210)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $218 = HEAP32[$217>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
           $219 = ($218|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
           if ($219) {
            $$0$i21$i$i$i$i = 0;
           } else {
            $220 = ((($210)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
            $221 = HEAP32[$220>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $222 = ((($210)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
            $223 = HEAP32[$222>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $224 = ((($210)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
            $225 = ((($210)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
            $226 = HEAP32[$225>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $227 = (FUNCTION_TABLE_iiii[$221 & 8191]($223,$224,$226)|0); //@line 1213 "./libraries/stb/stb_image.h"
            $228 = ($227|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
            if ($228) {
             HEAP32[$217>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
             $229 = ((($210)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
             HEAP8[$224>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
             $$pre$phi$i18$i$i$i$iZ2D = $229;$$sink$i$i19$i$i$i$i = $229;$3712 = 0;
            } else {
             $230 = (((($210)) + 40|0) + ($227)|0); //@line 1223 "./libraries/stb/stb_image.h"
             $$pre$i16$i$i$i$i = HEAP8[$224>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
             $$pre1$i17$i$i$i$i = ((($210)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
             $$pre$phi$i18$i$i$i$iZ2D = $$pre1$i17$i$i$i$i;$$sink$i$i19$i$i$i$i = $230;$3712 = $$pre$i16$i$i$i$i;
            }
            HEAP32[$212>>2] = $$sink$i$i19$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
            HEAP32[$209>>2] = $$pre$phi$i18$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
            $$0$i21$i$i$i$i = $3712;
           }
          }
          $231 = $$0$i21$i$i$i$i&255; //@line 2671 "./libraries/stb/stb_image.h"
          if (!($207)) {
           $$lcssa486 = $231;
           break;
          }
          $232 = HEAP32[$j$i>>2]|0; //@line 2672 "./libraries/stb/stb_image.h"
          $233 = ((($232)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
          $234 = HEAP32[$233>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $235 = ((($232)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
          $236 = HEAP32[$235>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $237 = ($234>>>0)<($236>>>0); //@line 1229 "./libraries/stb/stb_image.h"
          do {
           if ($237) {
            $238 = ((($234)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
            HEAP32[$233>>2] = $238; //@line 1230 "./libraries/stb/stb_image.h"
            $239 = HEAP8[$234>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
            $$0$i28$i$i$i$i = $239;$256 = $232;
           } else {
            $240 = ((($232)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
            $241 = HEAP32[$240>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
            $242 = ($241|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
            if ($242) {
             $$0$i28$i$i$i$i = 0;$256 = $232;
             break;
            }
            $243 = ((($232)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
            $244 = HEAP32[$243>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $245 = ((($232)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
            $246 = HEAP32[$245>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $247 = ((($232)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
            $248 = ((($232)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
            $249 = HEAP32[$248>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $250 = (FUNCTION_TABLE_iiii[$244 & 8191]($246,$247,$249)|0); //@line 1213 "./libraries/stb/stb_image.h"
            $251 = ($250|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
            if ($251) {
             HEAP32[$240>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
             $252 = ((($232)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
             HEAP8[$247>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
             $$pre$phi$i25$i$i$i$iZ2D = $252;$$sink$i$i26$i$i$i$i = $252;$3713 = 0;
            } else {
             $253 = (((($232)) + 40|0) + ($250)|0); //@line 1223 "./libraries/stb/stb_image.h"
             $$pre$i23$i$i$i$i = HEAP8[$247>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
             $$pre1$i24$i$i$i$i = ((($232)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
             $$pre$phi$i25$i$i$i$iZ2D = $$pre1$i24$i$i$i$i;$$sink$i$i26$i$i$i$i = $253;$3713 = $$pre$i23$i$i$i$i;
            }
            HEAP32[$235>>2] = $$sink$i$i26$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
            HEAP32[$233>>2] = $$pre$phi$i25$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
            $$pre52$i$i$i$i = HEAP32[$j$i>>2]|0; //@line 2673 "./libraries/stb/stb_image.h"
            $$0$i28$i$i$i$i = $3713;$256 = $$pre52$i$i$i$i;
           }
          } while(0);
          $254 = $$0$i28$i$i$i$i&255; //@line 2672 "./libraries/stb/stb_image.h"
          $255 = ((($256)) + 8|0); //@line 2673 "./libraries/stb/stb_image.h"
          $257 = HEAP32[$255>>2]|0; //@line 2676 "./libraries/stb/stb_image.h"
          $258 = ($257|0)>(0); //@line 2673 "./libraries/stb/stb_image.h"
          L112: do {
           if ($258) {
            $which$041$i$i$i$i = 0;
            while(1) {
             $259 = (((($j$i)) + 17820|0) + (($which$041$i$i$i$i*72)|0)|0); //@line 2674 "./libraries/stb/stb_image.h"
             $260 = HEAP32[$259>>2]|0; //@line 2674 "./libraries/stb/stb_image.h"
             $261 = ($260|0)==($231|0); //@line 2674 "./libraries/stb/stb_image.h"
             if ($261) {
              $which$0$lcssa$i$i$i$i = $which$041$i$i$i$i;
              break L112;
             }
             $262 = (($which$041$i$i$i$i) + 1)|0; //@line 2673 "./libraries/stb/stb_image.h"
             $263 = ($262|0)<($257|0); //@line 2673 "./libraries/stb/stb_image.h"
             if ($263) {
              $which$041$i$i$i$i = $262;
             } else {
              $which$0$lcssa$i$i$i$i = $262;
              break;
             }
            }
           } else {
            $which$0$lcssa$i$i$i$i = 0;
           }
          } while(0);
          $264 = ($which$0$lcssa$i$i$i$i|0)==($257|0); //@line 2676 "./libraries/stb/stb_image.h"
          if ($264) {
           break L31;
          }
          $265 = $254 >>> 4; //@line 2677 "./libraries/stb/stb_image.h"
          $266 = (((((($j$i)) + 17820|0) + (($which$0$lcssa$i$i$i$i*72)|0)|0)) + 16|0); //@line 2677 "./libraries/stb/stb_image.h"
          HEAP32[$266>>2] = $265; //@line 2677 "./libraries/stb/stb_image.h"
          $267 = ($$0$i28$i$i$i$i&255)>(63); //@line 2677 "./libraries/stb/stb_image.h"
          if ($267) {
           label = 90;
           break L57;
          }
          $268 = $254 & 15; //@line 2678 "./libraries/stb/stb_image.h"
          $269 = (((((($j$i)) + 17820|0) + (($which$0$lcssa$i$i$i$i*72)|0)|0)) + 20|0); //@line 2678 "./libraries/stb/stb_image.h"
          HEAP32[$269>>2] = $268; //@line 2678 "./libraries/stb/stb_image.h"
          $270 = ($268>>>0)>(3); //@line 2678 "./libraries/stb/stb_image.h"
          if ($270) {
           label = 92;
           break L57;
          }
          $271 = (((($j$i)) + 18152|0) + ($i$0$i$i$i$i<<2)|0); //@line 2679 "./libraries/stb/stb_image.h"
          HEAP32[$271>>2] = $which$0$lcssa$i$i$i$i; //@line 2679 "./libraries/stb/stb_image.h"
          $272 = (($i$0$i$i$i$i) + 1)|0; //@line 2670 "./libraries/stb/stb_image.h"
          $$pre$i11$i$i$i = HEAP32[$112>>2]|0; //@line 2670 "./libraries/stb/stb_image.h"
          $$pre47$i$i$i$i = HEAP32[$j$i>>2]|0; //@line 2671 "./libraries/stb/stb_image.h"
          $208 = $$pre$i11$i$i$i;$210 = $$pre47$i$i$i$i;$i$0$i$i$i$i = $272;
         }
         HEAP32[$113>>2] = $$lcssa486; //@line 2684 "./libraries/stb/stb_image.h"
         $273 = HEAP32[$j$i>>2]|0; //@line 2685 "./libraries/stb/stb_image.h"
         $274 = ((($273)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
         $275 = HEAP32[$274>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $276 = ((($273)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
         $277 = HEAP32[$276>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $278 = ($275>>>0)<($277>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         if ($278) {
          $279 = ((($275)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
          HEAP32[$274>>2] = $279; //@line 1230 "./libraries/stb/stb_image.h"
          $280 = HEAP8[$275>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
          $$0$i35$i$i$i$i = $280;$297 = $273;$300 = $279;$301 = $277;
         } else {
          $281 = ((($273)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
          $282 = HEAP32[$281>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
          $283 = ($282|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
          if ($283) {
           $$0$i35$i$i$i$i = 0;$297 = $273;$300 = $275;$301 = $277;
          } else {
           $284 = ((($273)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $285 = HEAP32[$284>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $286 = ((($273)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $287 = HEAP32[$286>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $288 = ((($273)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $289 = ((($273)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $290 = HEAP32[$289>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $291 = (FUNCTION_TABLE_iiii[$285 & 8191]($287,$288,$290)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $292 = ($291|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($292) {
            HEAP32[$281>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $293 = ((($273)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$288>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i32$i$i$i$iZ2D = $293;$$sink$i$i33$i$i$i$i = $293;$3714 = 0;
           } else {
            $294 = (((($273)) + 40|0) + ($291)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre$i30$i$i$i$i = HEAP8[$288>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
            $$pre1$i31$i$i$i$i = ((($273)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i32$i$i$i$iZ2D = $$pre1$i31$i$i$i$i;$$sink$i$i33$i$i$i$i = $294;$3714 = $$pre$i30$i$i$i$i;
           }
           HEAP32[$276>>2] = $$sink$i$i33$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$274>>2] = $$pre$phi$i32$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
           $$pre48$i$i$i$i = HEAP32[$j$i>>2]|0; //@line 2686 "./libraries/stb/stb_image.h"
           $$phi$trans$insert$i$i$i$i = ((($$pre48$i$i$i$i)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
           $$pre49$i$i$i$i = HEAP32[$$phi$trans$insert$i$i$i$i>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
           $$phi$trans$insert50$i$i$i$i = ((($$pre48$i$i$i$i)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
           $$pre51$i$i$i$i = HEAP32[$$phi$trans$insert50$i$i$i$i>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
           $$0$i35$i$i$i$i = $3714;$297 = $$pre48$i$i$i$i;$300 = $$pre49$i$i$i$i;$301 = $$pre51$i$i$i$i;
          }
         }
         $295 = $$0$i35$i$i$i$i&255; //@line 2685 "./libraries/stb/stb_image.h"
         HEAP32[$114>>2] = $295; //@line 2685 "./libraries/stb/stb_image.h"
         $296 = ((($297)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
         $298 = ((($297)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
         $299 = ($300>>>0)<($301>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         if ($299) {
          $302 = ((($300)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
          HEAP32[$296>>2] = $302; //@line 1230 "./libraries/stb/stb_image.h"
          $303 = HEAP8[$300>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
          $$0$i14$i$i$i$i = $303;
         } else {
          $304 = ((($297)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
          $305 = HEAP32[$304>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
          $306 = ($305|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
          if ($306) {
           $$0$i14$i$i$i$i = 0;
          } else {
           $307 = ((($297)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $308 = HEAP32[$307>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $309 = ((($297)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $310 = HEAP32[$309>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $311 = ((($297)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $312 = ((($297)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $313 = HEAP32[$312>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $314 = (FUNCTION_TABLE_iiii[$308 & 8191]($310,$311,$313)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $315 = ($314|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($315) {
            HEAP32[$304>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $316 = ((($297)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$311>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i11$i$i$i$iZ2D = $316;$$sink$i$i12$i$i$i$i = $316;$3715 = 0;
           } else {
            $317 = (((($297)) + 40|0) + ($314)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre$i9$i$i$i$i = HEAP8[$311>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
            $$pre1$i10$i$i$i$i = ((($297)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i11$i$i$i$iZ2D = $$pre1$i10$i$i$i$i;$$sink$i$i12$i$i$i$i = $317;$3715 = $$pre$i9$i$i$i$i;
           }
           HEAP32[$298>>2] = $$sink$i$i12$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$296>>2] = $$pre$phi$i11$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
           $$0$i14$i$i$i$i = $3715;
          }
         }
         $318 = $$0$i14$i$i$i$i&255; //@line 2686 "./libraries/stb/stb_image.h"
         $319 = $318 >>> 4; //@line 2694 "./libraries/stb/stb_image.h"
         HEAP32[$115>>2] = $319; //@line 2687 "./libraries/stb/stb_image.h"
         $320 = $318 & 15; //@line 2688 "./libraries/stb/stb_image.h"
         HEAP32[$116>>2] = $320; //@line 2688 "./libraries/stb/stb_image.h"
         $321 = HEAP32[$117>>2]|0; //@line 2452 "./libraries/stb/stb_image.h"
         $322 = ($321|0)==(0); //@line 2689 "./libraries/stb/stb_image.h"
         $323 = HEAP32[$113>>2]|0; //@line 2690 "./libraries/stb/stb_image.h"
         if ($322) {
          $330 = ($323|0)==(0); //@line 2693 "./libraries/stb/stb_image.h"
          if (!($330)) {
           label = 114;
           break L57;
          }
          $331 = $319 | $320; //@line 2694 "./libraries/stb/stb_image.h"
          $332 = ($331|0)==(0); //@line 2694 "./libraries/stb/stb_image.h"
          if (!($332)) {
           label = 116;
           break L57;
          }
          HEAP32[$114>>2] = 63; //@line 2695 "./libraries/stb/stb_image.h"
         } else {
          $324 = ($323|0)>(63); //@line 2690 "./libraries/stb/stb_image.h"
          if ($324) {
           label = 112;
           break L57;
          }
          $325 = HEAP32[$114>>2]|0; //@line 2690 "./libraries/stb/stb_image.h"
          $326 = ($325|0)>(63); //@line 2690 "./libraries/stb/stb_image.h"
          $327 = ($323|0)>($325|0); //@line 2690 "./libraries/stb/stb_image.h"
          $or$cond6$i$i$i$i = $326 | $327; //@line 2690 "./libraries/stb/stb_image.h"
          if ($or$cond6$i$i$i$i) {
           label = 112;
           break L57;
          }
          $328 = ($$0$i14$i$i$i$i&255)>(223); //@line 2690 "./libraries/stb/stb_image.h"
          $329 = ($320>>>0)>(13); //@line 2690 "./libraries/stb/stb_image.h"
          $or$cond7$i$i$i$i = $328 | $329; //@line 2690 "./libraries/stb/stb_image.h"
          if ($or$cond7$i$i$i$i) {
           label = 112;
           break L57;
          }
         }
         HEAP32[$118>>2] = 0; //@line 2438 "./libraries/stb/stb_image.h"
         HEAP32[$119>>2] = 0; //@line 2439 "./libraries/stb/stb_image.h"
         HEAP32[$120>>2] = 0; //@line 2440 "./libraries/stb/stb_image.h"
         HEAP32[$121>>2] = 0; //@line 2441 "./libraries/stb/stb_image.h"
         HEAP32[$122>>2] = 0; //@line 2441 "./libraries/stb/stb_image.h"
         HEAP32[$123>>2] = 0; //@line 2441 "./libraries/stb/stb_image.h"
         HEAP8[$63>>0] = -1; //@line 2442 "./libraries/stb/stb_image.h"
         $333 = HEAP32[$60>>2]|0; //@line 2443 "./libraries/stb/stb_image.h"
         $334 = ($333|0)==(0); //@line 2443 "./libraries/stb/stb_image.h"
         $$$i$i$i$i$i = $334 ? 2147483647 : $333; //@line 2443 "./libraries/stb/stb_image.h"
         HEAP32[$124>>2] = $$$i$i$i$i$i; //@line 2443 "./libraries/stb/stb_image.h"
         HEAP32[$125>>2] = 0; //@line 2444 "./libraries/stb/stb_image.h"
         $335 = HEAP32[$112>>2]|0; //@line 2511 "./libraries/stb/stb_image.h"
         $336 = ($335|0)==(1); //@line 2511 "./libraries/stb/stb_image.h"
         L147: do {
          if ($322) {
           if ($336) {
            $337 = HEAP32[$126>>2]|0; //@line 2456 "./libraries/stb/stb_image.h"
            $338 = (((((($j$i)) + 17820|0) + (($337*72)|0)|0)) + 28|0); //@line 2461 "./libraries/stb/stb_image.h"
            $339 = HEAP32[$338>>2]|0; //@line 2461 "./libraries/stb/stb_image.h"
            $340 = (($339) + 7)|0; //@line 2461 "./libraries/stb/stb_image.h"
            $341 = $340 >> 3; //@line 2461 "./libraries/stb/stb_image.h"
            $342 = (((((($j$i)) + 17820|0) + (($337*72)|0)|0)) + 32|0); //@line 2462 "./libraries/stb/stb_image.h"
            $343 = HEAP32[$342>>2]|0; //@line 2462 "./libraries/stb/stb_image.h"
            $344 = (($343) + 7)|0; //@line 2462 "./libraries/stb/stb_image.h"
            $345 = $344 >> 3; //@line 2462 "./libraries/stb/stb_image.h"
            $346 = ($345|0)>(0); //@line 2463 "./libraries/stb/stb_image.h"
            L151: do {
             if ($346) {
              $347 = ($341|0)>(0); //@line 2464 "./libraries/stb/stb_image.h"
              $348 = (((((($j$i)) + 17820|0) + (($337*72)|0)|0)) + 20|0); //@line 2465 "./libraries/stb/stb_image.h"
              $349 = (((((($j$i)) + 17820|0) + (($337*72)|0)|0)) + 16|0); //@line 2466 "./libraries/stb/stb_image.h"
              $350 = (((((($j$i)) + 17820|0) + (($337*72)|0)|0)) + 12|0); //@line 2466 "./libraries/stb/stb_image.h"
              $351 = (((((($j$i)) + 17820|0) + (($337*72)|0)|0)) + 44|0); //@line 2467 "./libraries/stb/stb_image.h"
              $352 = (((((($j$i)) + 17820|0) + (($337*72)|0)|0)) + 36|0); //@line 2467 "./libraries/stb/stb_image.h"
              $j$0134$i$i$i$i = 0;
              while(1) {
               if ($347) {
                $353 = $j$0134$i$i$i$i << 3; //@line 2467 "./libraries/stb/stb_image.h"
                $i$0133$i$i$i$i = 0;
                while(1) {
                 $354 = HEAP32[$348>>2]|0; //@line 2465 "./libraries/stb/stb_image.h"
                 $355 = HEAP32[$349>>2]|0; //@line 2466 "./libraries/stb/stb_image.h"
                 $356 = (((($j$i)) + 4|0) + (($355*1680)|0)|0); //@line 2466 "./libraries/stb/stb_image.h"
                 $357 = (((($j$i)) + 6724|0) + (($354*1680)|0)|0); //@line 2466 "./libraries/stb/stb_image.h"
                 $358 = (((($j$i)) + 13700|0) + ($354<<10)|0); //@line 2466 "./libraries/stb/stb_image.h"
                 $359 = HEAP32[$350>>2]|0; //@line 2466 "./libraries/stb/stb_image.h"
                 $360 = (((($j$i)) + 13444|0) + ($359<<6)|0); //@line 2466 "./libraries/stb/stb_image.h"
                 $361 = (__ZL23stbi__jpeg_decode_blockP10stbi__jpegPsP13stbi__huffmanS3_S1_iPh($j$i,$data$i$i$i$i,$356,$357,$358,$337,$360)|0); //@line 2466 "./libraries/stb/stb_image.h"
                 $362 = ($361|0)==(0); //@line 2466 "./libraries/stb/stb_image.h"
                 if ($362) {
                  $$3$i$i$i$i = 0;
                  break L151;
                 }
                 $363 = HEAP32[$47>>2]|0; //@line 2467 "./libraries/stb/stb_image.h"
                 $364 = HEAP32[$351>>2]|0; //@line 2467 "./libraries/stb/stb_image.h"
                 $365 = HEAP32[$352>>2]|0; //@line 2467 "./libraries/stb/stb_image.h"
                 $366 = Math_imul($353, $365)|0; //@line 2467 "./libraries/stb/stb_image.h"
                 $367 = (($364) + ($366)|0); //@line 2467 "./libraries/stb/stb_image.h"
                 $368 = $i$0133$i$i$i$i << 3; //@line 2467 "./libraries/stb/stb_image.h"
                 $369 = (($367) + ($368)|0); //@line 2467 "./libraries/stb/stb_image.h"
                 FUNCTION_TABLE_viii[$363 & 8191]($369,$365,$data$i$i$i$i); //@line 2467 "./libraries/stb/stb_image.h"
                 $370 = HEAP32[$124>>2]|0; //@line 2469 "./libraries/stb/stb_image.h"
                 $371 = (($370) + -1)|0; //@line 2469 "./libraries/stb/stb_image.h"
                 HEAP32[$124>>2] = $371; //@line 2469 "./libraries/stb/stb_image.h"
                 $372 = ($370|0)<(2); //@line 2469 "./libraries/stb/stb_image.h"
                 if ($372) {
                  $373 = HEAP32[$118>>2]|0; //@line 2470 "./libraries/stb/stb_image.h"
                  $374 = ($373|0)<(24); //@line 2470 "./libraries/stb/stb_image.h"
                  L162: do {
                   if ($374) {
                    L163: while(1) {
                     $375 = HEAP32[$120>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                     $376 = ($375|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
                     do {
                      if ($376) {
                       $377 = HEAP32[$j$i>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                       $378 = ((($377)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                       $379 = HEAP32[$378>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                       $380 = ((($377)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                       $381 = HEAP32[$380>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                       $382 = ($379>>>0)<($381>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                       if ($382) {
                        $383 = ((($379)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                        HEAP32[$378>>2] = $383; //@line 1230 "./libraries/stb/stb_image.h"
                        $384 = HEAP8[$379>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                        $$0$i36$i$i$i$i = $384;
                       } else {
                        $385 = ((($377)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                        $386 = HEAP32[$385>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                        $387 = ($386|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                        if ($387) {
                         $427 = 0;
                         break;
                        }
                        $388 = ((($377)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $389 = HEAP32[$388>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $390 = ((($377)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $391 = HEAP32[$390>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $392 = ((($377)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $393 = ((($377)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $394 = HEAP32[$393>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $395 = (FUNCTION_TABLE_iiii[$389 & 8191]($391,$392,$394)|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $396 = ($395|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                        if ($396) {
                         HEAP32[$385>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                         $397 = ((($377)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                         HEAP8[$392>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                         $$pre$phi$i33$i$i$i$iZ2D = $397;$$sink$i$i34$i$i$i$i = $397;$3716 = 0;
                        } else {
                         $398 = (((($377)) + 40|0) + ($395)|0); //@line 1223 "./libraries/stb/stb_image.h"
                         $$pre$i31$i$i$i$i = HEAP8[$392>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                         $$pre1$i32$i$i$i$i = ((($377)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                         $$pre$phi$i33$i$i$i$iZ2D = $$pre1$i32$i$i$i$i;$$sink$i$i34$i$i$i$i = $398;$3716 = $$pre$i31$i$i$i$i;
                        }
                        HEAP32[$380>>2] = $$sink$i$i34$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                        HEAP32[$378>>2] = $$pre$phi$i33$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                        $$0$i36$i$i$i$i = $3716;
                       }
                       $399 = $$0$i36$i$i$i$i&255; //@line 1590 "./libraries/stb/stb_image.h"
                       $400 = ($$0$i36$i$i$i$i<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
                       if (!($400)) {
                        $427 = $399;
                        break;
                       }
                       $401 = HEAP32[$j$i>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
                       $402 = ((($401)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                       $403 = HEAP32[$402>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                       $404 = ((($401)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                       $405 = HEAP32[$404>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                       $406 = ($403>>>0)<($405>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                       if ($406) {
                        $407 = ((($403)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                        HEAP32[$402>>2] = $407; //@line 1230 "./libraries/stb/stb_image.h"
                        $408 = HEAP8[$403>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                        $$0$i28$i19$i$i$i = $408;
                       } else {
                        $409 = ((($401)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                        $410 = HEAP32[$409>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                        $411 = ($410|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                        if ($411) {
                         $427 = 255;
                         break;
                        }
                        $412 = ((($401)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $413 = HEAP32[$412>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $414 = ((($401)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $415 = HEAP32[$414>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $416 = ((($401)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $417 = ((($401)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $418 = HEAP32[$417>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $419 = (FUNCTION_TABLE_iiii[$413 & 8191]($415,$416,$418)|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $420 = ($419|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                        if ($420) {
                         HEAP32[$409>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                         $421 = ((($401)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                         HEAP8[$416>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                         $$pre$phi$i25$i16$i$i$iZ2D = $421;$$sink$i$i26$i17$i$i$i = $421;$3717 = 0;
                        } else {
                         $422 = (((($401)) + 40|0) + ($419)|0); //@line 1223 "./libraries/stb/stb_image.h"
                         $$pre$i23$i14$i$i$i = HEAP8[$416>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                         $$pre1$i24$i15$i$i$i = ((($401)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                         $$pre$phi$i25$i16$i$i$iZ2D = $$pre1$i24$i15$i$i$i;$$sink$i$i26$i17$i$i$i = $422;$3717 = $$pre$i23$i14$i$i$i;
                        }
                        HEAP32[$404>>2] = $$sink$i$i26$i17$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                        HEAP32[$402>>2] = $$pre$phi$i25$i16$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                        $$0$i28$i19$i$i$i = $3717;
                       }
                       $423 = ($$0$i28$i19$i$i$i<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
                       if ($423) {
                        $427 = 255;
                       } else {
                        $$0$i28$i19$i$i$i$lcssa = $$0$i28$i19$i$i$i;
                        break L163;
                       }
                      } else {
                       $427 = 0;
                      }
                     } while(0);
                     $424 = HEAP32[$118>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                     $425 = (24 - ($424))|0; //@line 1599 "./libraries/stb/stb_image.h"
                     $426 = $427 << $425; //@line 1599 "./libraries/stb/stb_image.h"
                     $428 = HEAP32[$119>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                     $429 = $426 | $428; //@line 1599 "./libraries/stb/stb_image.h"
                     HEAP32[$119>>2] = $429; //@line 1599 "./libraries/stb/stb_image.h"
                     $430 = (($424) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
                     HEAP32[$118>>2] = $430; //@line 1600 "./libraries/stb/stb_image.h"
                     $431 = ($430|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
                     if (!($431)) {
                      break L162;
                     }
                    }
                    HEAP8[$63>>0] = $$0$i28$i19$i$i$i$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
                    HEAP32[$120>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
                   }
                  } while(0);
                  $432 = HEAP8[$63>>0]|0; //@line 2473 "./libraries/stb/stb_image.h"
                  $433 = $432 & -8; //@line 2473 "./libraries/stb/stb_image.h"
                  $434 = ($433<<24>>24)==(-48); //@line 2473 "./libraries/stb/stb_image.h"
                  if (!($434)) {
                   $$3$i$i$i$i = 1;
                   break L151;
                  }
                  HEAP32[$118>>2] = 0; //@line 2438 "./libraries/stb/stb_image.h"
                  HEAP32[$119>>2] = 0; //@line 2439 "./libraries/stb/stb_image.h"
                  HEAP32[$120>>2] = 0; //@line 2440 "./libraries/stb/stb_image.h"
                  HEAP32[$121>>2] = 0; //@line 2441 "./libraries/stb/stb_image.h"
                  HEAP32[$122>>2] = 0; //@line 2441 "./libraries/stb/stb_image.h"
                  HEAP32[$123>>2] = 0; //@line 2441 "./libraries/stb/stb_image.h"
                  HEAP8[$63>>0] = -1; //@line 2442 "./libraries/stb/stb_image.h"
                  $435 = HEAP32[$60>>2]|0; //@line 2443 "./libraries/stb/stb_image.h"
                  $436 = ($435|0)==(0); //@line 2443 "./libraries/stb/stb_image.h"
                  $$$i30$i$i$i$i = $436 ? 2147483647 : $435; //@line 2443 "./libraries/stb/stb_image.h"
                  HEAP32[$124>>2] = $$$i30$i$i$i$i; //@line 2443 "./libraries/stb/stb_image.h"
                  HEAP32[$125>>2] = 0; //@line 2444 "./libraries/stb/stb_image.h"
                 }
                 $437 = (($i$0133$i$i$i$i) + 1)|0; //@line 2464 "./libraries/stb/stb_image.h"
                 $438 = ($437|0)<($341|0); //@line 2464 "./libraries/stb/stb_image.h"
                 if ($438) {
                  $i$0133$i$i$i$i = $437;
                 } else {
                  break;
                 }
                }
               }
               $439 = (($j$0134$i$i$i$i) + 1)|0; //@line 2463 "./libraries/stb/stb_image.h"
               $440 = ($439|0)<($345|0); //@line 2463 "./libraries/stb/stb_image.h"
               if ($440) {
                $j$0134$i$i$i$i = $439;
               } else {
                $$3$i$i$i$i = 1;
                break;
               }
              }
             } else {
              $$3$i$i$i$i = 1;
             }
            } while(0);
            $$26$i$i$i$i = $$3$i$i$i$i;
           } else {
            $441 = HEAP32[$127>>2]|0; //@line 2482 "./libraries/stb/stb_image.h"
            $442 = ($441|0)>(0); //@line 2482 "./libraries/stb/stb_image.h"
            L194: do {
             if ($442) {
              $$pre206$i$i$i$i = HEAP32[$128>>2]|0; //@line 2483 "./libraries/stb/stb_image.h"
              $3718 = $441;$3719 = $$$i$i$i$i$i;$444 = $$pre206$i$i$i$i;$j2$0148$i$i$i$i = 0;
              while(1) {
               $443 = ($444|0)>(0); //@line 2483 "./libraries/stb/stb_image.h"
               if ($443) {
                $3720 = $3719;$i1$0145$i$i$i$i = 0;
                while(1) {
                 $445 = HEAP32[$112>>2]|0; //@line 2485 "./libraries/stb/stb_image.h"
                 $446 = ($445|0)>(0); //@line 2485 "./libraries/stb/stb_image.h"
                 if ($446) {
                  $3723 = $445;$k$0142$i$i$i$i = 0;
                  while(1) {
                   $447 = (((($j$i)) + 18152|0) + ($k$0142$i$i$i$i<<2)|0); //@line 2486 "./libraries/stb/stb_image.h"
                   $448 = HEAP32[$447>>2]|0; //@line 2486 "./libraries/stb/stb_image.h"
                   $449 = (((((($j$i)) + 17820|0) + (($448*72)|0)|0)) + 8|0); //@line 2489 "./libraries/stb/stb_image.h"
                   $450 = HEAP32[$449>>2]|0; //@line 2489 "./libraries/stb/stb_image.h"
                   $451 = ($450|0)>(0); //@line 2489 "./libraries/stb/stb_image.h"
                   if ($451) {
                    $452 = (((((($j$i)) + 17820|0) + (($448*72)|0)|0)) + 4|0); //@line 2490 "./libraries/stb/stb_image.h"
                    $453 = (((((($j$i)) + 17820|0) + (($448*72)|0)|0)) + 20|0); //@line 2493 "./libraries/stb/stb_image.h"
                    $454 = (((((($j$i)) + 17820|0) + (($448*72)|0)|0)) + 16|0); //@line 2494 "./libraries/stb/stb_image.h"
                    $455 = (((((($j$i)) + 17820|0) + (($448*72)|0)|0)) + 12|0); //@line 2494 "./libraries/stb/stb_image.h"
                    $456 = (((((($j$i)) + 17820|0) + (($448*72)|0)|0)) + 44|0); //@line 2495 "./libraries/stb/stb_image.h"
                    $457 = (((((($j$i)) + 17820|0) + (($448*72)|0)|0)) + 36|0); //@line 2495 "./libraries/stb/stb_image.h"
                    $$pre207$i$i$i$i = HEAP32[$452>>2]|0; //@line 2490 "./libraries/stb/stb_image.h"
                    $3724 = $450;$459 = $$pre207$i$i$i$i;$y$0139$i$i$i$i = 0;
                    while(1) {
                     $458 = ($459|0)>(0); //@line 2490 "./libraries/stb/stb_image.h"
                     if ($458) {
                      $470 = $3724;$474 = $459;$x$0136$i$i$i$i = 0;
                      while(1) {
                       $460 = HEAP32[$453>>2]|0; //@line 2493 "./libraries/stb/stb_image.h"
                       $461 = HEAP32[$454>>2]|0; //@line 2494 "./libraries/stb/stb_image.h"
                       $462 = (((($j$i)) + 4|0) + (($461*1680)|0)|0); //@line 2494 "./libraries/stb/stb_image.h"
                       $463 = (((($j$i)) + 6724|0) + (($460*1680)|0)|0); //@line 2494 "./libraries/stb/stb_image.h"
                       $464 = (((($j$i)) + 13700|0) + ($460<<10)|0); //@line 2494 "./libraries/stb/stb_image.h"
                       $465 = HEAP32[$455>>2]|0; //@line 2494 "./libraries/stb/stb_image.h"
                       $466 = (((($j$i)) + 13444|0) + ($465<<6)|0); //@line 2494 "./libraries/stb/stb_image.h"
                       $467 = (__ZL23stbi__jpeg_decode_blockP10stbi__jpegPsP13stbi__huffmanS3_S1_iPh($j$i,$data3$i$i$i$i,$462,$463,$464,$448,$466)|0); //@line 2494 "./libraries/stb/stb_image.h"
                       $468 = ($467|0)==(0); //@line 2494 "./libraries/stb/stb_image.h"
                       if ($468) {
                        $$11$i$i$i$i = 0;
                        break L194;
                       }
                       $469 = Math_imul($470, $j2$0148$i$i$i$i)|0; //@line 2492 "./libraries/stb/stb_image.h"
                       $471 = (($469) + ($y$0139$i$i$i$i))|0; //@line 2492 "./libraries/stb/stb_image.h"
                       $472 = $471 << 3; //@line 2492 "./libraries/stb/stb_image.h"
                       $473 = Math_imul($474, $i1$0145$i$i$i$i)|0; //@line 2491 "./libraries/stb/stb_image.h"
                       $475 = (($473) + ($x$0136$i$i$i$i))|0; //@line 2491 "./libraries/stb/stb_image.h"
                       $476 = $475 << 3; //@line 2491 "./libraries/stb/stb_image.h"
                       $477 = HEAP32[$47>>2]|0; //@line 2495 "./libraries/stb/stb_image.h"
                       $478 = HEAP32[$456>>2]|0; //@line 2495 "./libraries/stb/stb_image.h"
                       $479 = HEAP32[$457>>2]|0; //@line 2495 "./libraries/stb/stb_image.h"
                       $480 = Math_imul($472, $479)|0; //@line 2495 "./libraries/stb/stb_image.h"
                       $481 = (($478) + ($480)|0); //@line 2495 "./libraries/stb/stb_image.h"
                       $482 = (($481) + ($476)|0); //@line 2495 "./libraries/stb/stb_image.h"
                       FUNCTION_TABLE_viii[$477 & 8191]($482,$479,$data3$i$i$i$i); //@line 2495 "./libraries/stb/stb_image.h"
                       $483 = (($x$0136$i$i$i$i) + 1)|0; //@line 2490 "./libraries/stb/stb_image.h"
                       $484 = HEAP32[$452>>2]|0; //@line 2490 "./libraries/stb/stb_image.h"
                       $485 = ($483|0)<($484|0); //@line 2490 "./libraries/stb/stb_image.h"
                       $$pre208$i$i$i$i = HEAP32[$449>>2]|0; //@line 2492 "./libraries/stb/stb_image.h"
                       if ($485) {
                        $470 = $$pre208$i$i$i$i;$474 = $484;$x$0136$i$i$i$i = $483;
                       } else {
                        $3725 = $484;$488 = $$pre208$i$i$i$i;
                        break;
                       }
                      }
                     } else {
                      $3725 = $459;$488 = $3724;
                     }
                     $486 = (($y$0139$i$i$i$i) + 1)|0; //@line 2489 "./libraries/stb/stb_image.h"
                     $487 = ($486|0)<($488|0); //@line 2489 "./libraries/stb/stb_image.h"
                     if ($487) {
                      $3724 = $488;$459 = $3725;$y$0139$i$i$i$i = $486;
                     } else {
                      break;
                     }
                    }
                    $$pre210$i$i$i$i = HEAP32[$112>>2]|0; //@line 2485 "./libraries/stb/stb_image.h"
                    $491 = $$pre210$i$i$i$i;
                   } else {
                    $491 = $3723;
                   }
                   $489 = (($k$0142$i$i$i$i) + 1)|0; //@line 2485 "./libraries/stb/stb_image.h"
                   $490 = ($489|0)<($491|0); //@line 2485 "./libraries/stb/stb_image.h"
                   if ($490) {
                    $3723 = $491;$k$0142$i$i$i$i = $489;
                   } else {
                    break;
                   }
                  }
                  $$pre211$i$i$i$i = HEAP32[$124>>2]|0; //@line 2501 "./libraries/stb/stb_image.h"
                  $493 = $$pre211$i$i$i$i;
                 } else {
                  $493 = $3720;
                 }
                 $492 = (($493) + -1)|0; //@line 2501 "./libraries/stb/stb_image.h"
                 HEAP32[$124>>2] = $492; //@line 2501 "./libraries/stb/stb_image.h"
                 $494 = ($493|0)<(2); //@line 2501 "./libraries/stb/stb_image.h"
                 if ($494) {
                  $495 = HEAP32[$118>>2]|0; //@line 2502 "./libraries/stb/stb_image.h"
                  $496 = ($495|0)<(24); //@line 2502 "./libraries/stb/stb_image.h"
                  L219: do {
                   if ($496) {
                    L220: while(1) {
                     $497 = HEAP32[$120>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                     $498 = ($497|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
                     do {
                      if ($498) {
                       $499 = HEAP32[$j$i>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                       $500 = ((($499)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                       $501 = HEAP32[$500>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                       $502 = ((($499)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                       $503 = HEAP32[$502>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                       $504 = ($501>>>0)<($503>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                       if ($504) {
                        $505 = ((($501)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                        HEAP32[$500>>2] = $505; //@line 1230 "./libraries/stb/stb_image.h"
                        $506 = HEAP8[$501>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                        $$0$i53$i$i$i$i = $506;
                       } else {
                        $507 = ((($499)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                        $508 = HEAP32[$507>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                        $509 = ($508|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                        if ($509) {
                         $549 = 0;
                         break;
                        }
                        $510 = ((($499)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $511 = HEAP32[$510>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $512 = ((($499)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $513 = HEAP32[$512>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $514 = ((($499)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $515 = ((($499)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $516 = HEAP32[$515>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $517 = (FUNCTION_TABLE_iiii[$511 & 8191]($513,$514,$516)|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $518 = ($517|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                        if ($518) {
                         HEAP32[$507>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                         $519 = ((($499)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                         HEAP8[$514>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                         $$pre$phi$i50$i$i$i$iZ2D = $519;$$sink$i$i51$i$i$i$i = $519;$3727 = 0;
                        } else {
                         $520 = (((($499)) + 40|0) + ($517)|0); //@line 1223 "./libraries/stb/stb_image.h"
                         $$pre$i48$i$i$i$i = HEAP8[$514>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                         $$pre1$i49$i$i$i$i = ((($499)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                         $$pre$phi$i50$i$i$i$iZ2D = $$pre1$i49$i$i$i$i;$$sink$i$i51$i$i$i$i = $520;$3727 = $$pre$i48$i$i$i$i;
                        }
                        HEAP32[$502>>2] = $$sink$i$i51$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                        HEAP32[$500>>2] = $$pre$phi$i50$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                        $$0$i53$i$i$i$i = $3727;
                       }
                       $521 = $$0$i53$i$i$i$i&255; //@line 1590 "./libraries/stb/stb_image.h"
                       $522 = ($$0$i53$i$i$i$i<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
                       if (!($522)) {
                        $549 = $521;
                        break;
                       }
                       $523 = HEAP32[$j$i>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
                       $524 = ((($523)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                       $525 = HEAP32[$524>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                       $526 = ((($523)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                       $527 = HEAP32[$526>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                       $528 = ($525>>>0)<($527>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                       if ($528) {
                        $529 = ((($525)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                        HEAP32[$524>>2] = $529; //@line 1230 "./libraries/stb/stb_image.h"
                        $530 = HEAP8[$525>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                        $$0$i45$i$i$i$i = $530;
                       } else {
                        $531 = ((($523)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                        $532 = HEAP32[$531>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                        $533 = ($532|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                        if ($533) {
                         $549 = 255;
                         break;
                        }
                        $534 = ((($523)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $535 = HEAP32[$534>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $536 = ((($523)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $537 = HEAP32[$536>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $538 = ((($523)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $539 = ((($523)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $540 = HEAP32[$539>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $541 = (FUNCTION_TABLE_iiii[$535 & 8191]($537,$538,$540)|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $542 = ($541|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                        if ($542) {
                         HEAP32[$531>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                         $543 = ((($523)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                         HEAP8[$538>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                         $$pre$phi$i42$i$i$i$iZ2D = $543;$$sink$i$i43$i$i$i$i = $543;$3728 = 0;
                        } else {
                         $544 = (((($523)) + 40|0) + ($541)|0); //@line 1223 "./libraries/stb/stb_image.h"
                         $$pre$i40$i$i$i$i = HEAP8[$538>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                         $$pre1$i41$i$i$i$i = ((($523)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                         $$pre$phi$i42$i$i$i$iZ2D = $$pre1$i41$i$i$i$i;$$sink$i$i43$i$i$i$i = $544;$3728 = $$pre$i40$i$i$i$i;
                        }
                        HEAP32[$526>>2] = $$sink$i$i43$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                        HEAP32[$524>>2] = $$pre$phi$i42$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                        $$0$i45$i$i$i$i = $3728;
                       }
                       $545 = ($$0$i45$i$i$i$i<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
                       if ($545) {
                        $549 = 255;
                       } else {
                        $$0$i45$i$i$i$i$lcssa = $$0$i45$i$i$i$i;
                        break L220;
                       }
                      } else {
                       $549 = 0;
                      }
                     } while(0);
                     $546 = HEAP32[$118>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                     $547 = (24 - ($546))|0; //@line 1599 "./libraries/stb/stb_image.h"
                     $548 = $549 << $547; //@line 1599 "./libraries/stb/stb_image.h"
                     $550 = HEAP32[$119>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                     $551 = $548 | $550; //@line 1599 "./libraries/stb/stb_image.h"
                     HEAP32[$119>>2] = $551; //@line 1599 "./libraries/stb/stb_image.h"
                     $552 = (($546) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
                     HEAP32[$118>>2] = $552; //@line 1600 "./libraries/stb/stb_image.h"
                     $553 = ($552|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
                     if (!($553)) {
                      break L219;
                     }
                    }
                    HEAP8[$63>>0] = $$0$i45$i$i$i$i$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
                    HEAP32[$120>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
                   }
                  } while(0);
                  $554 = HEAP8[$63>>0]|0; //@line 2503 "./libraries/stb/stb_image.h"
                  $555 = $554 & -8; //@line 2503 "./libraries/stb/stb_image.h"
                  $556 = ($555<<24>>24)==(-48); //@line 2503 "./libraries/stb/stb_image.h"
                  if (!($556)) {
                   $$11$i$i$i$i = 1;
                   break L194;
                  }
                  HEAP32[$118>>2] = 0; //@line 2438 "./libraries/stb/stb_image.h"
                  HEAP32[$119>>2] = 0; //@line 2439 "./libraries/stb/stb_image.h"
                  HEAP32[$120>>2] = 0; //@line 2440 "./libraries/stb/stb_image.h"
                  HEAP32[$121>>2] = 0; //@line 2441 "./libraries/stb/stb_image.h"
                  HEAP32[$122>>2] = 0; //@line 2441 "./libraries/stb/stb_image.h"
                  HEAP32[$123>>2] = 0; //@line 2441 "./libraries/stb/stb_image.h"
                  HEAP8[$63>>0] = -1; //@line 2442 "./libraries/stb/stb_image.h"
                  $557 = HEAP32[$60>>2]|0; //@line 2443 "./libraries/stb/stb_image.h"
                  $558 = ($557|0)==(0); //@line 2443 "./libraries/stb/stb_image.h"
                  $$$i47$i$i$i$i = $558 ? 2147483647 : $557; //@line 2443 "./libraries/stb/stb_image.h"
                  HEAP32[$124>>2] = $$$i47$i$i$i$i; //@line 2443 "./libraries/stb/stb_image.h"
                  HEAP32[$125>>2] = 0; //@line 2444 "./libraries/stb/stb_image.h"
                  $3726 = $$$i47$i$i$i$i;
                 } else {
                  $3726 = $492;
                 }
                 $559 = (($i1$0145$i$i$i$i) + 1)|0; //@line 2483 "./libraries/stb/stb_image.h"
                 $560 = HEAP32[$128>>2]|0; //@line 2483 "./libraries/stb/stb_image.h"
                 $561 = ($559|0)<($560|0); //@line 2483 "./libraries/stb/stb_image.h"
                 if ($561) {
                  $3720 = $3726;$i1$0145$i$i$i$i = $559;
                 } else {
                  $$lcssa514 = $3726;$$lcssa515 = $560;
                  break;
                 }
                }
                $$pre212$i$i$i$i = HEAP32[$127>>2]|0; //@line 2482 "./libraries/stb/stb_image.h"
                $3721 = $$lcssa514;$3722 = $$lcssa515;$564 = $$pre212$i$i$i$i;
               } else {
                $3721 = $3719;$3722 = $444;$564 = $3718;
               }
               $562 = (($j2$0148$i$i$i$i) + 1)|0; //@line 2482 "./libraries/stb/stb_image.h"
               $563 = ($562|0)<($564|0); //@line 2482 "./libraries/stb/stb_image.h"
               if ($563) {
                $3718 = $564;$3719 = $3721;$444 = $3722;$j2$0148$i$i$i$i = $562;
               } else {
                $$11$i$i$i$i = 1;
                break;
               }
              }
             } else {
              $$11$i$i$i$i = 1;
             }
            } while(0);
            $$26$i$i$i$i = $$11$i$i$i$i;
           }
           $1363 = ($$26$i$i$i$i|0)==(0); //@line 2829 "./libraries/stb/stb_image.h"
           if ($1363) {
            break L31;
           } else {
            label = 449;
           }
          } else {
           if (!($336)) {
            $565 = HEAP32[$127>>2]|0; //@line 2542 "./libraries/stb/stb_image.h"
            $566 = ($565|0)>(0); //@line 2542 "./libraries/stb/stb_image.h"
            if (!($566)) {
             label = 449;
             break;
            }
            $$pre$i22$i$i$i = HEAP32[$128>>2]|0; //@line 2543 "./libraries/stb/stb_image.h"
            $1251 = $$pre$i22$i$i$i;$3729 = $565;$3730 = $$$i$i$i$i$i;$j18$0167$i$i$i$i = 0;
            while(1) {
             $1250 = ($1251|0)>(0); //@line 2543 "./libraries/stb/stb_image.h"
             if ($1250) {
              $3749 = $3730;$i17$0164$i$i$i$i = 0;
              while(1) {
               $1252 = HEAP32[$112>>2]|0; //@line 2545 "./libraries/stb/stb_image.h"
               $1253 = ($1252|0)>(0); //@line 2545 "./libraries/stb/stb_image.h"
               if ($1253) {
                $3752 = $1252;$k19$0161$i$i$i$i = 0;
                while(1) {
                 $1254 = (((($j$i)) + 18152|0) + ($k19$0161$i$i$i$i<<2)|0); //@line 2546 "./libraries/stb/stb_image.h"
                 $1255 = HEAP32[$1254>>2]|0; //@line 2546 "./libraries/stb/stb_image.h"
                 $1256 = (((((($j$i)) + 17820|0) + (($1255*72)|0)|0)) + 8|0); //@line 2549 "./libraries/stb/stb_image.h"
                 $1257 = HEAP32[$1256>>2]|0; //@line 2549 "./libraries/stb/stb_image.h"
                 $1258 = ($1257|0)>(0); //@line 2549 "./libraries/stb/stb_image.h"
                 if ($1258) {
                  $1259 = (((((($j$i)) + 17820|0) + (($1255*72)|0)|0)) + 4|0); //@line 2550 "./libraries/stb/stb_image.h"
                  $1260 = (((((($j$i)) + 17820|0) + (($1255*72)|0)|0)) + 60|0); //@line 2553 "./libraries/stb/stb_image.h"
                  $1261 = (((((($j$i)) + 17820|0) + (($1255*72)|0)|0)) + 64|0); //@line 2553 "./libraries/stb/stb_image.h"
                  $1262 = (((((($j$i)) + 17820|0) + (($1255*72)|0)|0)) + 16|0); //@line 2554 "./libraries/stb/stb_image.h"
                  $$pre200$i$i$i$i = HEAP32[$1259>>2]|0; //@line 2550 "./libraries/stb/stb_image.h"
                  $1264 = $$pre200$i$i$i$i;$3753 = $1257;$y21$0158$i$i$i$i = 0;
                  while(1) {
                   $1263 = ($1264|0)>(0); //@line 2550 "./libraries/stb/stb_image.h"
                   if ($1263) {
                    $1269 = $1264;$1272 = $3753;$x20$0155$i$i$i$i = 0;
                    while(1) {
                     $1268 = Math_imul($1269, $i17$0164$i$i$i$i)|0; //@line 2551 "./libraries/stb/stb_image.h"
                     $1270 = (($1268) + ($x20$0155$i$i$i$i))|0; //@line 2551 "./libraries/stb/stb_image.h"
                     $1271 = Math_imul($1272, $j18$0167$i$i$i$i)|0; //@line 2552 "./libraries/stb/stb_image.h"
                     $1273 = (($1271) + ($y21$0158$i$i$i$i))|0; //@line 2552 "./libraries/stb/stb_image.h"
                     $1274 = HEAP32[$1260>>2]|0; //@line 2553 "./libraries/stb/stb_image.h"
                     $1275 = HEAP32[$1261>>2]|0; //@line 2553 "./libraries/stb/stb_image.h"
                     $1276 = Math_imul($1275, $1273)|0; //@line 2553 "./libraries/stb/stb_image.h"
                     $1277 = (($1270) + ($1276))|0; //@line 2553 "./libraries/stb/stb_image.h"
                     $1278 = $1277 << 6; //@line 2553 "./libraries/stb/stb_image.h"
                     $1279 = (($1274) + ($1278<<1)|0); //@line 2553 "./libraries/stb/stb_image.h"
                     $1280 = HEAP32[$1262>>2]|0; //@line 2554 "./libraries/stb/stb_image.h"
                     $1281 = (((($j$i)) + 4|0) + (($1280*1680)|0)|0); //@line 2554 "./libraries/stb/stb_image.h"
                     $1282 = (__ZL31stbi__jpeg_decode_block_prog_dcP10stbi__jpegPsP13stbi__huffmani($j$i,$1279,$1281,$1255)|0); //@line 2554 "./libraries/stb/stb_image.h"
                     $1283 = ($1282|0)==(0); //@line 2554 "./libraries/stb/stb_image.h"
                     $1267 = (($x20$0155$i$i$i$i) + 1)|0; //@line 2550 "./libraries/stb/stb_image.h"
                     if ($1283) {
                      break L31;
                     }
                     $1265 = HEAP32[$1259>>2]|0; //@line 2550 "./libraries/stb/stb_image.h"
                     $1266 = ($1267|0)<($1265|0); //@line 2550 "./libraries/stb/stb_image.h"
                     $$pre201$i$i$i$i = HEAP32[$1256>>2]|0; //@line 2552 "./libraries/stb/stb_image.h"
                     if ($1266) {
                      $1269 = $1265;$1272 = $$pre201$i$i$i$i;$x20$0155$i$i$i$i = $1267;
                     } else {
                      $1286 = $$pre201$i$i$i$i;$3754 = $1265;
                      break;
                     }
                    }
                   } else {
                    $1286 = $3753;$3754 = $1264;
                   }
                   $1284 = (($y21$0158$i$i$i$i) + 1)|0; //@line 2549 "./libraries/stb/stb_image.h"
                   $1285 = ($1284|0)<($1286|0); //@line 2549 "./libraries/stb/stb_image.h"
                   if ($1285) {
                    $1264 = $3754;$3753 = $1286;$y21$0158$i$i$i$i = $1284;
                   } else {
                    break;
                   }
                  }
                  $$pre203$i$i$i$i = HEAP32[$112>>2]|0; //@line 2545 "./libraries/stb/stb_image.h"
                  $1289 = $$pre203$i$i$i$i;
                 } else {
                  $1289 = $3752;
                 }
                 $1287 = (($k19$0161$i$i$i$i) + 1)|0; //@line 2545 "./libraries/stb/stb_image.h"
                 $1288 = ($1287|0)<($1289|0); //@line 2545 "./libraries/stb/stb_image.h"
                 if ($1288) {
                  $3752 = $1289;$k19$0161$i$i$i$i = $1287;
                 } else {
                  break;
                 }
                }
                $$pre204$i$i$i$i = HEAP32[$124>>2]|0; //@line 2561 "./libraries/stb/stb_image.h"
                $1291 = $$pre204$i$i$i$i;
               } else {
                $1291 = $3749;
               }
               $1290 = (($1291) + -1)|0; //@line 2561 "./libraries/stb/stb_image.h"
               HEAP32[$124>>2] = $1290; //@line 2561 "./libraries/stb/stb_image.h"
               $1292 = ($1291|0)<(2); //@line 2561 "./libraries/stb/stb_image.h"
               if ($1292) {
                $1293 = HEAP32[$118>>2]|0; //@line 2562 "./libraries/stb/stb_image.h"
                $1294 = ($1293|0)<(24); //@line 2562 "./libraries/stb/stb_image.h"
                L279: do {
                 if ($1294) {
                  L280: while(1) {
                   $1295 = HEAP32[$120>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                   $1296 = ($1295|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
                   do {
                    if ($1296) {
                     $1297 = HEAP32[$j$i>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                     $1298 = ((($1297)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                     $1299 = HEAP32[$1298>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                     $1300 = ((($1297)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                     $1301 = HEAP32[$1300>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                     $1302 = ($1299>>>0)<($1301>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                     if ($1302) {
                      $1303 = ((($1299)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                      HEAP32[$1298>>2] = $1303; //@line 1230 "./libraries/stb/stb_image.h"
                      $1304 = HEAP8[$1299>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                      $$0$i81$i$i$i$i = $1304;
                     } else {
                      $1305 = ((($1297)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                      $1306 = HEAP32[$1305>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                      $1307 = ($1306|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                      if ($1307) {
                       $1347 = 0;
                       break;
                      }
                      $1308 = ((($1297)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                      $1309 = HEAP32[$1308>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                      $1310 = ((($1297)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                      $1311 = HEAP32[$1310>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                      $1312 = ((($1297)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                      $1313 = ((($1297)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                      $1314 = HEAP32[$1313>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                      $1315 = (FUNCTION_TABLE_iiii[$1309 & 8191]($1311,$1312,$1314)|0); //@line 1213 "./libraries/stb/stb_image.h"
                      $1316 = ($1315|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                      if ($1316) {
                       HEAP32[$1305>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                       $1317 = ((($1297)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                       HEAP8[$1312>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                       $$pre$phi$i78$i$i$i$iZ2D = $1317;$$sink$i$i79$i$i$i$i = $1317;$3756 = 0;
                      } else {
                       $1318 = (((($1297)) + 40|0) + ($1315)|0); //@line 1223 "./libraries/stb/stb_image.h"
                       $$pre$i76$i$i$i$i = HEAP8[$1312>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                       $$pre1$i77$i$i$i$i = ((($1297)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                       $$pre$phi$i78$i$i$i$iZ2D = $$pre1$i77$i$i$i$i;$$sink$i$i79$i$i$i$i = $1318;$3756 = $$pre$i76$i$i$i$i;
                      }
                      HEAP32[$1300>>2] = $$sink$i$i79$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                      HEAP32[$1298>>2] = $$pre$phi$i78$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                      $$0$i81$i$i$i$i = $3756;
                     }
                     $1319 = $$0$i81$i$i$i$i&255; //@line 1590 "./libraries/stb/stb_image.h"
                     $1320 = ($$0$i81$i$i$i$i<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
                     if (!($1320)) {
                      $1347 = $1319;
                      break;
                     }
                     $1321 = HEAP32[$j$i>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
                     $1322 = ((($1321)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                     $1323 = HEAP32[$1322>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                     $1324 = ((($1321)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                     $1325 = HEAP32[$1324>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                     $1326 = ($1323>>>0)<($1325>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                     if ($1326) {
                      $1327 = ((($1323)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                      HEAP32[$1322>>2] = $1327; //@line 1230 "./libraries/stb/stb_image.h"
                      $1328 = HEAP8[$1323>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                      $$0$i$i33$i$i$i = $1328;
                     } else {
                      $1329 = ((($1321)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                      $1330 = HEAP32[$1329>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                      $1331 = ($1330|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                      if ($1331) {
                       $1347 = 255;
                       break;
                      }
                      $1332 = ((($1321)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                      $1333 = HEAP32[$1332>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                      $1334 = ((($1321)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                      $1335 = HEAP32[$1334>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                      $1336 = ((($1321)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                      $1337 = ((($1321)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                      $1338 = HEAP32[$1337>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                      $1339 = (FUNCTION_TABLE_iiii[$1333 & 8191]($1335,$1336,$1338)|0); //@line 1213 "./libraries/stb/stb_image.h"
                      $1340 = ($1339|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                      if ($1340) {
                       HEAP32[$1329>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                       $1341 = ((($1321)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                       HEAP8[$1336>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                       $$pre$phi$i$i30$i$i$iZ2D = $1341;$$sink$i$i$i31$i$i$i = $1341;$3757 = 0;
                      } else {
                       $1342 = (((($1321)) + 40|0) + ($1339)|0); //@line 1223 "./libraries/stb/stb_image.h"
                       $$pre$i$i28$i$i$i = HEAP8[$1336>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                       $$pre1$i$i29$i$i$i = ((($1321)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                       $$pre$phi$i$i30$i$i$iZ2D = $$pre1$i$i29$i$i$i;$$sink$i$i$i31$i$i$i = $1342;$3757 = $$pre$i$i28$i$i$i;
                      }
                      HEAP32[$1324>>2] = $$sink$i$i$i31$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                      HEAP32[$1322>>2] = $$pre$phi$i$i30$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                      $$0$i$i33$i$i$i = $3757;
                     }
                     $1343 = ($$0$i$i33$i$i$i<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
                     if ($1343) {
                      $1347 = 255;
                     } else {
                      $$0$i$i33$i$i$i$lcssa = $$0$i$i33$i$i$i;
                      break L280;
                     }
                    } else {
                     $1347 = 0;
                    }
                   } while(0);
                   $1344 = HEAP32[$118>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                   $1345 = (24 - ($1344))|0; //@line 1599 "./libraries/stb/stb_image.h"
                   $1346 = $1347 << $1345; //@line 1599 "./libraries/stb/stb_image.h"
                   $1348 = HEAP32[$119>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                   $1349 = $1346 | $1348; //@line 1599 "./libraries/stb/stb_image.h"
                   HEAP32[$119>>2] = $1349; //@line 1599 "./libraries/stb/stb_image.h"
                   $1350 = (($1344) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
                   HEAP32[$118>>2] = $1350; //@line 1600 "./libraries/stb/stb_image.h"
                   $1351 = ($1350|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
                   if (!($1351)) {
                    break L279;
                   }
                  }
                  HEAP8[$63>>0] = $$0$i$i33$i$i$i$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
                  HEAP32[$120>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
                 }
                } while(0);
                $1352 = HEAP8[$63>>0]|0; //@line 2563 "./libraries/stb/stb_image.h"
                $1353 = $1352 & -8; //@line 2563 "./libraries/stb/stb_image.h"
                $1354 = ($1353<<24>>24)==(-48); //@line 2563 "./libraries/stb/stb_image.h"
                if (!($1354)) {
                 $1365 = $1352;
                 break L147;
                }
                HEAP32[$118>>2] = 0; //@line 2438 "./libraries/stb/stb_image.h"
                HEAP32[$119>>2] = 0; //@line 2439 "./libraries/stb/stb_image.h"
                HEAP32[$120>>2] = 0; //@line 2440 "./libraries/stb/stb_image.h"
                HEAP32[$121>>2] = 0; //@line 2441 "./libraries/stb/stb_image.h"
                HEAP32[$122>>2] = 0; //@line 2441 "./libraries/stb/stb_image.h"
                HEAP32[$123>>2] = 0; //@line 2441 "./libraries/stb/stb_image.h"
                HEAP8[$63>>0] = -1; //@line 2442 "./libraries/stb/stb_image.h"
                $1355 = HEAP32[$60>>2]|0; //@line 2443 "./libraries/stb/stb_image.h"
                $1356 = ($1355|0)==(0); //@line 2443 "./libraries/stb/stb_image.h"
                $$$i20$i$i$i$i = $1356 ? 2147483647 : $1355; //@line 2443 "./libraries/stb/stb_image.h"
                HEAP32[$124>>2] = $$$i20$i$i$i$i; //@line 2443 "./libraries/stb/stb_image.h"
                HEAP32[$125>>2] = 0; //@line 2444 "./libraries/stb/stb_image.h"
                $3755 = $$$i20$i$i$i$i;
               } else {
                $3755 = $1290;
               }
               $1357 = (($i17$0164$i$i$i$i) + 1)|0; //@line 2543 "./libraries/stb/stb_image.h"
               $1358 = HEAP32[$128>>2]|0; //@line 2543 "./libraries/stb/stb_image.h"
               $1359 = ($1357|0)<($1358|0); //@line 2543 "./libraries/stb/stb_image.h"
               if ($1359) {
                $3749 = $3755;$i17$0164$i$i$i$i = $1357;
               } else {
                $$lcssa492 = $3755;$$lcssa493 = $1358;
                break;
               }
              }
              $$pre205$i$i$i$i = HEAP32[$127>>2]|0; //@line 2542 "./libraries/stb/stb_image.h"
              $1362 = $$pre205$i$i$i$i;$3750 = $$lcssa492;$3751 = $$lcssa493;
             } else {
              $1362 = $3729;$3750 = $3730;$3751 = $1251;
             }
             $1360 = (($j18$0167$i$i$i$i) + 1)|0; //@line 2542 "./libraries/stb/stb_image.h"
             $1361 = ($1360|0)<($1362|0); //@line 2542 "./libraries/stb/stb_image.h"
             if ($1361) {
              $1251 = $3751;$3729 = $1362;$3730 = $3750;$j18$0167$i$i$i$i = $1360;
             } else {
              label = 449;
              break L147;
             }
            }
           }
           $567 = HEAP32[$126>>2]|0; //@line 2513 "./libraries/stb/stb_image.h"
           $568 = (((((($j$i)) + 17820|0) + (($567*72)|0)|0)) + 28|0); //@line 2518 "./libraries/stb/stb_image.h"
           $569 = HEAP32[$568>>2]|0; //@line 2518 "./libraries/stb/stb_image.h"
           $570 = (($569) + 7)|0; //@line 2518 "./libraries/stb/stb_image.h"
           $571 = $570 >> 3; //@line 2518 "./libraries/stb/stb_image.h"
           $572 = (((((($j$i)) + 17820|0) + (($567*72)|0)|0)) + 32|0); //@line 2519 "./libraries/stb/stb_image.h"
           $573 = HEAP32[$572>>2]|0; //@line 2519 "./libraries/stb/stb_image.h"
           $574 = (($573) + 7)|0; //@line 2519 "./libraries/stb/stb_image.h"
           $575 = $574 >> 3; //@line 2519 "./libraries/stb/stb_image.h"
           $576 = ($575|0)>(0); //@line 2520 "./libraries/stb/stb_image.h"
           if (!($576)) {
            label = 449;
            break;
           }
           $577 = ($571|0)>(0); //@line 2521 "./libraries/stb/stb_image.h"
           $578 = (((((($j$i)) + 17820|0) + (($567*72)|0)|0)) + 60|0); //@line 2522 "./libraries/stb/stb_image.h"
           $579 = (((((($j$i)) + 17820|0) + (($567*72)|0)|0)) + 64|0); //@line 2522 "./libraries/stb/stb_image.h"
           $580 = (((((($j$i)) + 17820|0) + (($567*72)|0)|0)) + 16|0); //@line 2524 "./libraries/stb/stb_image.h"
           $581 = (((((($j$i)) + 17820|0) + (($567*72)|0)|0)) + 20|0); //@line 2527 "./libraries/stb/stb_image.h"
           $j9$0153$i$i$i$i = 0;
           while(1) {
            if ($577) {
             $i8$0150$i$i$i$i = 0;
             while(1) {
              $582 = HEAP32[$578>>2]|0; //@line 2522 "./libraries/stb/stb_image.h"
              $583 = HEAP32[$579>>2]|0; //@line 2522 "./libraries/stb/stb_image.h"
              $584 = Math_imul($583, $j9$0153$i$i$i$i)|0; //@line 2522 "./libraries/stb/stb_image.h"
              $585 = (($584) + ($i8$0150$i$i$i$i))|0; //@line 2522 "./libraries/stb/stb_image.h"
              $586 = $585 << 6; //@line 2522 "./libraries/stb/stb_image.h"
              $587 = (($582) + ($586<<1)|0); //@line 2522 "./libraries/stb/stb_image.h"
              $588 = HEAP32[$113>>2]|0; //@line 2523 "./libraries/stb/stb_image.h"
              $589 = ($588|0)==(0); //@line 2523 "./libraries/stb/stb_image.h"
              L317: do {
               if ($589) {
                $590 = HEAP32[$580>>2]|0; //@line 2524 "./libraries/stb/stb_image.h"
                $591 = (((($j$i)) + 4|0) + (($590*1680)|0)|0); //@line 2524 "./libraries/stb/stb_image.h"
                $592 = (__ZL31stbi__jpeg_decode_block_prog_dcP10stbi__jpegPsP13stbi__huffmani($j$i,$587,$591,$567)|0); //@line 2524 "./libraries/stb/stb_image.h"
                $593 = ($592|0)==(0); //@line 2524 "./libraries/stb/stb_image.h"
                if ($593) {
                 break L31;
                }
               } else {
                $594 = HEAP32[$581>>2]|0; //@line 2527 "./libraries/stb/stb_image.h"
                $595 = (((($j$i)) + 6724|0) + (($594*1680)|0)|0); //@line 2528 "./libraries/stb/stb_image.h"
                $596 = HEAP32[$115>>2]|0; //@line 1801 "./libraries/stb/stb_image.h"
                $597 = ($596|0)==(0); //@line 1801 "./libraries/stb/stb_image.h"
                $598 = HEAP32[$116>>2]|0; //@line 1802 "./libraries/stb/stb_image.h"
                if ($597) {
                 $599 = HEAP32[$125>>2]|0; //@line 1804 "./libraries/stb/stb_image.h"
                 $600 = ($599|0)==(0); //@line 1804 "./libraries/stb/stb_image.h"
                 if ($600) {
                  $k$0$i$i$i$i$i = $588;
                 } else {
                  $601 = (($599) + -1)|0; //@line 1805 "./libraries/stb/stb_image.h"
                  HEAP32[$125>>2] = $601; //@line 1805 "./libraries/stb/stb_image.h"
                  break;
                 }
                 L324: while(1) {
                  $602 = HEAP32[$118>>2]|0; //@line 1813 "./libraries/stb/stb_image.h"
                  $603 = ($602|0)<(16); //@line 1813 "./libraries/stb/stb_image.h"
                  L326: do {
                   if ($603) {
                    L327: while(1) {
                     $604 = HEAP32[$120>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                     $605 = ($604|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
                     do {
                      if ($605) {
                       $606 = HEAP32[$j$i>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                       $607 = ((($606)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                       $608 = HEAP32[$607>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                       $609 = ((($606)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                       $610 = HEAP32[$609>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                       $611 = ($608>>>0)<($610>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                       if ($611) {
                        $612 = ((($608)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                        HEAP32[$607>>2] = $612; //@line 1230 "./libraries/stb/stb_image.h"
                        $613 = HEAP8[$608>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                        $$0$i27$i$i$i$i$i = $613;
                       } else {
                        $614 = ((($606)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                        $615 = HEAP32[$614>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                        $616 = ($615|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                        if ($616) {
                         $656 = 0;
                         break;
                        }
                        $617 = ((($606)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $618 = HEAP32[$617>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $619 = ((($606)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $620 = HEAP32[$619>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $621 = ((($606)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $622 = ((($606)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $623 = HEAP32[$622>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $624 = (FUNCTION_TABLE_iiii[$618 & 8191]($620,$621,$623)|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $625 = ($624|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                        if ($625) {
                         HEAP32[$614>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                         $626 = ((($606)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                         HEAP8[$621>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                         $$pre$phi$i24$i$i$i$i$iZ2D = $626;$$sink$i$i25$i$i$i$i$i = $626;$3731 = 0;
                        } else {
                         $627 = (((($606)) + 40|0) + ($624)|0); //@line 1223 "./libraries/stb/stb_image.h"
                         $$pre$i22$i$i$i$i$i = HEAP8[$621>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                         $$pre1$i23$i$i$i$i$i = ((($606)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                         $$pre$phi$i24$i$i$i$i$iZ2D = $$pre1$i23$i$i$i$i$i;$$sink$i$i25$i$i$i$i$i = $627;$3731 = $$pre$i22$i$i$i$i$i;
                        }
                        HEAP32[$609>>2] = $$sink$i$i25$i$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                        HEAP32[$607>>2] = $$pre$phi$i24$i$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                        $$0$i27$i$i$i$i$i = $3731;
                       }
                       $628 = $$0$i27$i$i$i$i$i&255; //@line 1590 "./libraries/stb/stb_image.h"
                       $629 = ($$0$i27$i$i$i$i$i<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
                       if (!($629)) {
                        $656 = $628;
                        break;
                       }
                       $630 = HEAP32[$j$i>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
                       $631 = ((($630)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                       $632 = HEAP32[$631>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                       $633 = ((($630)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                       $634 = HEAP32[$633>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                       $635 = ($632>>>0)<($634>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                       if ($635) {
                        $636 = ((($632)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                        HEAP32[$631>>2] = $636; //@line 1230 "./libraries/stb/stb_image.h"
                        $637 = HEAP8[$632>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                        $$0$i20$i$i$i$i$i = $637;
                       } else {
                        $638 = ((($630)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                        $639 = HEAP32[$638>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                        $640 = ($639|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                        if ($640) {
                         $656 = 255;
                         break;
                        }
                        $641 = ((($630)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $642 = HEAP32[$641>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $643 = ((($630)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $644 = HEAP32[$643>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $645 = ((($630)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $646 = ((($630)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $647 = HEAP32[$646>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $648 = (FUNCTION_TABLE_iiii[$642 & 8191]($644,$645,$647)|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $649 = ($648|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                        if ($649) {
                         HEAP32[$638>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                         $650 = ((($630)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                         HEAP8[$645>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                         $$pre$phi$i17$i$i$i$i$iZ2D = $650;$$sink$i$i18$i$i$i$i$i = $650;$3732 = 0;
                        } else {
                         $651 = (((($630)) + 40|0) + ($648)|0); //@line 1223 "./libraries/stb/stb_image.h"
                         $$pre$i15$i$i$i$i$i = HEAP8[$645>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                         $$pre1$i16$i$i$i$i$i = ((($630)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                         $$pre$phi$i17$i$i$i$i$iZ2D = $$pre1$i16$i$i$i$i$i;$$sink$i$i18$i$i$i$i$i = $651;$3732 = $$pre$i15$i$i$i$i$i;
                        }
                        HEAP32[$633>>2] = $$sink$i$i18$i$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                        HEAP32[$631>>2] = $$pre$phi$i17$i$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                        $$0$i20$i$i$i$i$i = $3732;
                       }
                       $652 = ($$0$i20$i$i$i$i$i<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
                       if ($652) {
                        $656 = 255;
                       } else {
                        $$0$i20$i$i$i$i$i$lcssa = $$0$i20$i$i$i$i$i;
                        break L327;
                       }
                      } else {
                       $656 = 0;
                      }
                     } while(0);
                     $653 = HEAP32[$118>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                     $654 = (24 - ($653))|0; //@line 1599 "./libraries/stb/stb_image.h"
                     $655 = $656 << $654; //@line 1599 "./libraries/stb/stb_image.h"
                     $657 = HEAP32[$119>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                     $658 = $655 | $657; //@line 1599 "./libraries/stb/stb_image.h"
                     HEAP32[$119>>2] = $658; //@line 1599 "./libraries/stb/stb_image.h"
                     $659 = (($653) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
                     HEAP32[$118>>2] = $659; //@line 1600 "./libraries/stb/stb_image.h"
                     $660 = ($659|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
                     if (!($660)) {
                      break L326;
                     }
                    }
                    HEAP8[$63>>0] = $$0$i20$i$i$i$i$i$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
                    HEAP32[$120>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
                   }
                  } while(0);
                  $661 = HEAP32[$119>>2]|0; //@line 1814 "./libraries/stb/stb_image.h"
                  $662 = $661 >>> 23; //@line 1814 "./libraries/stb/stb_image.h"
                  $663 = ((((($j$i)) + 13700|0) + ($594<<10)|0) + ($662<<1)|0); //@line 1815 "./libraries/stb/stb_image.h"
                  $664 = HEAP16[$663>>1]|0; //@line 1815 "./libraries/stb/stb_image.h"
                  $665 = $664 << 16 >> 16; //@line 1815 "./libraries/stb/stb_image.h"
                  $666 = ($664<<16>>16)==(0); //@line 1816 "./libraries/stb/stb_image.h"
                  do {
                   if ($666) {
                    $682 = (__ZL22stbi__jpeg_huff_decodeP10stbi__jpegP13stbi__huffman($j$i,$595)|0); //@line 1824 "./libraries/stb/stb_image.h"
                    $683 = ($682|0)<(0); //@line 1825 "./libraries/stb/stb_image.h"
                    if ($683) {
                     label = 225;
                     break L57;
                    }
                    $684 = $682 & 15; //@line 1826 "./libraries/stb/stb_image.h"
                    $685 = $682 >> 4; //@line 1827 "./libraries/stb/stb_image.h"
                    $686 = ($684|0)==(0); //@line 1828 "./libraries/stb/stb_image.h"
                    if ($686) {
                     $687 = ($685|0)<(15); //@line 1829 "./libraries/stb/stb_image.h"
                     if ($687) {
                      $$lcssa510 = $685;
                      break L324;
                     }
                     $763 = (($k$0$i$i$i$i$i) + 16)|0; //@line 1836 "./libraries/stb/stb_image.h"
                     $k$4$ph$i$i$i$i$i = $763;
                     break;
                    }
                    $764 = (($685) + ($k$0$i$i$i$i$i))|0; //@line 1838 "./libraries/stb/stb_image.h"
                    $765 = (($764) + 1)|0; //@line 1839 "./libraries/stb/stb_image.h"
                    $766 = (440770 + ($764)|0); //@line 1839 "./libraries/stb/stb_image.h"
                    $767 = HEAP8[$766>>0]|0; //@line 1839 "./libraries/stb/stb_image.h"
                    $768 = $767&255; //@line 1839 "./libraries/stb/stb_image.h"
                    $769 = HEAP32[$118>>2]|0; //@line 1666 "./libraries/stb/stb_image.h"
                    $770 = ($769|0)<($684|0); //@line 1666 "./libraries/stb/stb_image.h"
                    L360: do {
                     if ($770) {
                      L361: while(1) {
                       $771 = HEAP32[$120>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                       $772 = ($771|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
                       do {
                        if ($772) {
                         $773 = HEAP32[$j$i>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                         $774 = ((($773)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                         $775 = HEAP32[$774>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                         $776 = ((($773)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                         $777 = HEAP32[$776>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                         $778 = ($775>>>0)<($777>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                         if ($778) {
                          $779 = ((($775)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                          HEAP32[$774>>2] = $779; //@line 1230 "./libraries/stb/stb_image.h"
                          $780 = HEAP8[$775>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                          $$0$i122$i$i$i$i$i = $780;
                         } else {
                          $781 = ((($773)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                          $782 = HEAP32[$781>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                          $783 = ($782|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                          if ($783) {
                           $823 = 0;
                           break;
                          }
                          $784 = ((($773)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $785 = HEAP32[$784>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                          $786 = ((($773)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $787 = HEAP32[$786>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                          $788 = ((($773)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $789 = ((($773)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $790 = HEAP32[$789>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                          $791 = (FUNCTION_TABLE_iiii[$785 & 8191]($787,$788,$790)|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $792 = ($791|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                          if ($792) {
                           HEAP32[$781>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                           $793 = ((($773)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                           HEAP8[$788>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                           $$pre$phi$i119$i$i$i$i$iZ2D = $793;$$sink$i$i120$i$i$i$i$i = $793;$3735 = 0;
                          } else {
                           $794 = (((($773)) + 40|0) + ($791)|0); //@line 1223 "./libraries/stb/stb_image.h"
                           $$pre$i117$i$i$i$i$i = HEAP8[$788>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                           $$pre1$i118$i$i$i$i$i = ((($773)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                           $$pre$phi$i119$i$i$i$i$iZ2D = $$pre1$i118$i$i$i$i$i;$$sink$i$i120$i$i$i$i$i = $794;$3735 = $$pre$i117$i$i$i$i$i;
                          }
                          HEAP32[$776>>2] = $$sink$i$i120$i$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                          HEAP32[$774>>2] = $$pre$phi$i119$i$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                          $$0$i122$i$i$i$i$i = $3735;
                         }
                         $795 = $$0$i122$i$i$i$i$i&255; //@line 1590 "./libraries/stb/stb_image.h"
                         $796 = ($$0$i122$i$i$i$i$i<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
                         if (!($796)) {
                          $823 = $795;
                          break;
                         }
                         $797 = HEAP32[$j$i>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
                         $798 = ((($797)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                         $799 = HEAP32[$798>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                         $800 = ((($797)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                         $801 = HEAP32[$800>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                         $802 = ($799>>>0)<($801>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                         if ($802) {
                          $803 = ((($799)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                          HEAP32[$798>>2] = $803; //@line 1230 "./libraries/stb/stb_image.h"
                          $804 = HEAP8[$799>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                          $$0$i48$i$i$i$i$i = $804;
                         } else {
                          $805 = ((($797)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                          $806 = HEAP32[$805>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                          $807 = ($806|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                          if ($807) {
                           $823 = 255;
                           break;
                          }
                          $808 = ((($797)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $809 = HEAP32[$808>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                          $810 = ((($797)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $811 = HEAP32[$810>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                          $812 = ((($797)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $813 = ((($797)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $814 = HEAP32[$813>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                          $815 = (FUNCTION_TABLE_iiii[$809 & 8191]($811,$812,$814)|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $816 = ($815|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                          if ($816) {
                           HEAP32[$805>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                           $817 = ((($797)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                           HEAP8[$812>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                           $$pre$phi$i45$i$i$i$i$iZ2D = $817;$$sink$i$i46$i$i$i$i$i = $817;$3736 = 0;
                          } else {
                           $818 = (((($797)) + 40|0) + ($815)|0); //@line 1223 "./libraries/stb/stb_image.h"
                           $$pre$i43$i$i$i$i$i = HEAP8[$812>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                           $$pre1$i44$i$i$i$i$i = ((($797)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                           $$pre$phi$i45$i$i$i$i$iZ2D = $$pre1$i44$i$i$i$i$i;$$sink$i$i46$i$i$i$i$i = $818;$3736 = $$pre$i43$i$i$i$i$i;
                          }
                          HEAP32[$800>>2] = $$sink$i$i46$i$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                          HEAP32[$798>>2] = $$pre$phi$i45$i$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                          $$0$i48$i$i$i$i$i = $3736;
                         }
                         $819 = ($$0$i48$i$i$i$i$i<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
                         if ($819) {
                          $823 = 255;
                         } else {
                          $$0$i48$i$i$i$i$i$lcssa = $$0$i48$i$i$i$i$i;
                          break L361;
                         }
                        } else {
                         $823 = 0;
                        }
                       } while(0);
                       $820 = HEAP32[$118>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                       $821 = (24 - ($820))|0; //@line 1599 "./libraries/stb/stb_image.h"
                       $822 = $823 << $821; //@line 1599 "./libraries/stb/stb_image.h"
                       $824 = HEAP32[$119>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                       $825 = $822 | $824; //@line 1599 "./libraries/stb/stb_image.h"
                       HEAP32[$119>>2] = $825; //@line 1599 "./libraries/stb/stb_image.h"
                       $826 = (($820) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
                       HEAP32[$118>>2] = $826; //@line 1600 "./libraries/stb/stb_image.h"
                       $827 = ($826|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
                       if (!($827)) {
                        $840 = $826;
                        break L360;
                       }
                      }
                      HEAP8[$63>>0] = $$0$i48$i$i$i$i$i$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
                      HEAP32[$120>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
                      $$pre$i55$i$i$i$i = HEAP32[$118>>2]|0; //@line 1673 "./libraries/stb/stb_image.h"
                      $840 = $$pre$i55$i$i$i$i;
                     } else {
                      $840 = $769;
                     }
                    } while(0);
                    $828 = HEAP32[$119>>2]|0; //@line 1668 "./libraries/stb/stb_image.h"
                    $829 = $828 << $684; //@line 1669 "./libraries/stb/stb_image.h"
                    $830 = (32 - ($684))|0; //@line 1669 "./libraries/stb/stb_image.h"
                    $831 = $828 >>> $830; //@line 1669 "./libraries/stb/stb_image.h"
                    $832 = $829 | $831; //@line 1669 "./libraries/stb/stb_image.h"
                    $833 = $828 >> 31; //@line 1668 "./libraries/stb/stb_image.h"
                    $834 = (67056 + ($684<<2)|0); //@line 1671 "./libraries/stb/stb_image.h"
                    $835 = HEAP32[$834>>2]|0; //@line 1671 "./libraries/stb/stb_image.h"
                    $836 = $835 ^ -1; //@line 1671 "./libraries/stb/stb_image.h"
                    $837 = $832 & $836; //@line 1671 "./libraries/stb/stb_image.h"
                    HEAP32[$119>>2] = $837; //@line 1671 "./libraries/stb/stb_image.h"
                    $838 = $832 & $835; //@line 1672 "./libraries/stb/stb_image.h"
                    $839 = (($840) - ($684))|0; //@line 1673 "./libraries/stb/stb_image.h"
                    HEAP32[$118>>2] = $839; //@line 1673 "./libraries/stb/stb_image.h"
                    $841 = (26004 + ($684<<2)|0); //@line 1674 "./libraries/stb/stb_image.h"
                    $842 = HEAP32[$841>>2]|0; //@line 1674 "./libraries/stb/stb_image.h"
                    $843 = $833 ^ -1; //@line 1674 "./libraries/stb/stb_image.h"
                    $844 = $842 & $843; //@line 1674 "./libraries/stb/stb_image.h"
                    $845 = (($844) + ($838))|0; //@line 1674 "./libraries/stb/stb_image.h"
                    $846 = $845 << $598; //@line 1840 "./libraries/stb/stb_image.h"
                    $847 = $846&65535; //@line 1840 "./libraries/stb/stb_image.h"
                    $848 = (($587) + ($768<<1)|0); //@line 1840 "./libraries/stb/stb_image.h"
                    HEAP16[$848>>1] = $847; //@line 1840 "./libraries/stb/stb_image.h"
                    $k$4$ph$i$i$i$i$i = $765;
                   } else {
                    $667 = $665 >>> 4; //@line 1817 "./libraries/stb/stb_image.h"
                    $668 = $667 & 15; //@line 1817 "./libraries/stb/stb_image.h"
                    $669 = (($668) + ($k$0$i$i$i$i$i))|0; //@line 1817 "./libraries/stb/stb_image.h"
                    $670 = $665 & 15; //@line 1818 "./libraries/stb/stb_image.h"
                    $671 = $661 << $670; //@line 1819 "./libraries/stb/stb_image.h"
                    HEAP32[$119>>2] = $671; //@line 1819 "./libraries/stb/stb_image.h"
                    $672 = HEAP32[$118>>2]|0; //@line 1820 "./libraries/stb/stb_image.h"
                    $673 = (($672) - ($670))|0; //@line 1820 "./libraries/stb/stb_image.h"
                    HEAP32[$118>>2] = $673; //@line 1820 "./libraries/stb/stb_image.h"
                    $674 = (($669) + 1)|0; //@line 1821 "./libraries/stb/stb_image.h"
                    $675 = (440770 + ($669)|0); //@line 1821 "./libraries/stb/stb_image.h"
                    $676 = HEAP8[$675>>0]|0; //@line 1821 "./libraries/stb/stb_image.h"
                    $677 = $676&255; //@line 1821 "./libraries/stb/stb_image.h"
                    $678 = $665 >> 8; //@line 1822 "./libraries/stb/stb_image.h"
                    $679 = $678 << $598; //@line 1822 "./libraries/stb/stb_image.h"
                    $680 = $679&65535; //@line 1822 "./libraries/stb/stb_image.h"
                    $681 = (($587) + ($677<<1)|0); //@line 1822 "./libraries/stb/stb_image.h"
                    HEAP16[$681>>1] = $680; //@line 1822 "./libraries/stb/stb_image.h"
                    $k$4$ph$i$i$i$i$i = $674;
                   }
                  } while(0);
                  $851 = HEAP32[$114>>2]|0; //@line 1843 "./libraries/stb/stb_image.h"
                  $852 = ($k$4$ph$i$i$i$i$i|0)>($851|0); //@line 1843 "./libraries/stb/stb_image.h"
                  if ($852) {
                   break L317;
                  } else {
                   $k$0$i$i$i$i$i = $k$4$ph$i$i$i$i$i;
                  }
                 }
                 $688 = 1 << $$lcssa510; //@line 1830 "./libraries/stb/stb_image.h"
                 HEAP32[$125>>2] = $688; //@line 1830 "./libraries/stb/stb_image.h"
                 $689 = ($$lcssa510|0)==(0); //@line 1831 "./libraries/stb/stb_image.h"
                 if ($689) {
                  $850 = $688;
                 } else {
                  $690 = HEAP32[$118>>2]|0; //@line 1681 "./libraries/stb/stb_image.h"
                  $691 = ($690|0)<($$lcssa510|0); //@line 1681 "./libraries/stb/stb_image.h"
                  L392: do {
                   if ($691) {
                    L393: while(1) {
                     $692 = HEAP32[$120>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                     $693 = ($692|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
                     do {
                      if ($693) {
                       $694 = HEAP32[$j$i>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                       $695 = ((($694)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                       $696 = HEAP32[$695>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                       $697 = ((($694)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                       $698 = HEAP32[$697>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                       $699 = ($696>>>0)<($698>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                       if ($699) {
                        $700 = ((($696)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                        HEAP32[$695>>2] = $700; //@line 1230 "./libraries/stb/stb_image.h"
                        $701 = HEAP8[$696>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                        $$0$i39$i$i$i$i$i = $701;
                       } else {
                        $702 = ((($694)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                        $703 = HEAP32[$702>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                        $704 = ($703|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                        if ($704) {
                         $744 = 0;
                         break;
                        }
                        $705 = ((($694)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $706 = HEAP32[$705>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $707 = ((($694)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $708 = HEAP32[$707>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $709 = ((($694)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $710 = ((($694)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $711 = HEAP32[$710>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $712 = (FUNCTION_TABLE_iiii[$706 & 8191]($708,$709,$711)|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $713 = ($712|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                        if ($713) {
                         HEAP32[$702>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                         $714 = ((($694)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                         HEAP8[$709>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                         $$pre$phi$i36$i$i$i$i$iZ2D = $714;$$sink$i$i37$i$i$i$i$i = $714;$3733 = 0;
                        } else {
                         $715 = (((($694)) + 40|0) + ($712)|0); //@line 1223 "./libraries/stb/stb_image.h"
                         $$pre$i34$i$i$i$i$i = HEAP8[$709>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                         $$pre1$i35$i$i$i$i$i = ((($694)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                         $$pre$phi$i36$i$i$i$i$iZ2D = $$pre1$i35$i$i$i$i$i;$$sink$i$i37$i$i$i$i$i = $715;$3733 = $$pre$i34$i$i$i$i$i;
                        }
                        HEAP32[$697>>2] = $$sink$i$i37$i$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                        HEAP32[$695>>2] = $$pre$phi$i36$i$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                        $$0$i39$i$i$i$i$i = $3733;
                       }
                       $716 = $$0$i39$i$i$i$i$i&255; //@line 1590 "./libraries/stb/stb_image.h"
                       $717 = ($$0$i39$i$i$i$i$i<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
                       if (!($717)) {
                        $744 = $716;
                        break;
                       }
                       $718 = HEAP32[$j$i>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
                       $719 = ((($718)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                       $720 = HEAP32[$719>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                       $721 = ((($718)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                       $722 = HEAP32[$721>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                       $723 = ($720>>>0)<($722>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                       if ($723) {
                        $724 = ((($720)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                        HEAP32[$719>>2] = $724; //@line 1230 "./libraries/stb/stb_image.h"
                        $725 = HEAP8[$720>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                        $$0$i115$i$i$i$i$i = $725;
                       } else {
                        $726 = ((($718)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                        $727 = HEAP32[$726>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                        $728 = ($727|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                        if ($728) {
                         $744 = 255;
                         break;
                        }
                        $729 = ((($718)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $730 = HEAP32[$729>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $731 = ((($718)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $732 = HEAP32[$731>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $733 = ((($718)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $734 = ((($718)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $735 = HEAP32[$734>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                        $736 = (FUNCTION_TABLE_iiii[$730 & 8191]($732,$733,$735)|0); //@line 1213 "./libraries/stb/stb_image.h"
                        $737 = ($736|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                        if ($737) {
                         HEAP32[$726>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                         $738 = ((($718)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                         HEAP8[$733>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                         $$pre$phi$i112$i$i$i$i$iZ2D = $738;$$sink$i$i113$i$i$i$i$i = $738;$3734 = 0;
                        } else {
                         $739 = (((($718)) + 40|0) + ($736)|0); //@line 1223 "./libraries/stb/stb_image.h"
                         $$pre$i110$i$i$i$i$i = HEAP8[$733>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                         $$pre1$i111$i$i$i$i$i = ((($718)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                         $$pre$phi$i112$i$i$i$i$iZ2D = $$pre1$i111$i$i$i$i$i;$$sink$i$i113$i$i$i$i$i = $739;$3734 = $$pre$i110$i$i$i$i$i;
                        }
                        HEAP32[$721>>2] = $$sink$i$i113$i$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                        HEAP32[$719>>2] = $$pre$phi$i112$i$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                        $$0$i115$i$i$i$i$i = $3734;
                       }
                       $740 = ($$0$i115$i$i$i$i$i<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
                       if ($740) {
                        $744 = 255;
                       } else {
                        $$0$i115$i$i$i$i$i$lcssa = $$0$i115$i$i$i$i$i;
                        break L393;
                       }
                      } else {
                       $744 = 0;
                      }
                     } while(0);
                     $741 = HEAP32[$118>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                     $742 = (24 - ($741))|0; //@line 1599 "./libraries/stb/stb_image.h"
                     $743 = $744 << $742; //@line 1599 "./libraries/stb/stb_image.h"
                     $745 = HEAP32[$119>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                     $746 = $743 | $745; //@line 1599 "./libraries/stb/stb_image.h"
                     HEAP32[$119>>2] = $746; //@line 1599 "./libraries/stb/stb_image.h"
                     $747 = (($741) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
                     HEAP32[$118>>2] = $747; //@line 1600 "./libraries/stb/stb_image.h"
                     $748 = ($747|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
                     if (!($748)) {
                      $760 = $747;
                      break L392;
                     }
                    }
                    HEAP8[$63>>0] = $$0$i115$i$i$i$i$i$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
                    HEAP32[$120>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
                    $$pre$i31$i$i$i$i$i = HEAP32[$118>>2]|0; //@line 1685 "./libraries/stb/stb_image.h"
                    $760 = $$pre$i31$i$i$i$i$i;
                   } else {
                    $760 = $690;
                   }
                  } while(0);
                  $749 = HEAP32[$119>>2]|0; //@line 1682 "./libraries/stb/stb_image.h"
                  $750 = $749 << $$lcssa510; //@line 1682 "./libraries/stb/stb_image.h"
                  $751 = (32 - ($$lcssa510))|0; //@line 1682 "./libraries/stb/stb_image.h"
                  $752 = $749 >>> $751; //@line 1682 "./libraries/stb/stb_image.h"
                  $753 = $750 | $752; //@line 1682 "./libraries/stb/stb_image.h"
                  $754 = (67056 + ($$lcssa510<<2)|0); //@line 1683 "./libraries/stb/stb_image.h"
                  $755 = HEAP32[$754>>2]|0; //@line 1683 "./libraries/stb/stb_image.h"
                  $756 = $755 ^ -1; //@line 1683 "./libraries/stb/stb_image.h"
                  $757 = $753 & $756; //@line 1683 "./libraries/stb/stb_image.h"
                  HEAP32[$119>>2] = $757; //@line 1683 "./libraries/stb/stb_image.h"
                  $758 = $753 & $755; //@line 1684 "./libraries/stb/stb_image.h"
                  $759 = (($760) - ($$lcssa510))|0; //@line 1685 "./libraries/stb/stb_image.h"
                  HEAP32[$118>>2] = $759; //@line 1685 "./libraries/stb/stb_image.h"
                  $761 = HEAP32[$125>>2]|0; //@line 1832 "./libraries/stb/stb_image.h"
                  $762 = (($761) + ($758))|0; //@line 1832 "./libraries/stb/stb_image.h"
                  HEAP32[$125>>2] = $762; //@line 1832 "./libraries/stb/stb_image.h"
                  $850 = $762;
                 }
                 $849 = (($850) + -1)|0; //@line 1833 "./libraries/stb/stb_image.h"
                 HEAP32[$125>>2] = $849; //@line 1833 "./libraries/stb/stb_image.h"
                 break;
                }
                $853 = 1 << $598; //@line 1847 "./libraries/stb/stb_image.h"
                $854 = HEAP32[$125>>2]|0; //@line 1849 "./libraries/stb/stb_image.h"
                $855 = ($854|0)==(0); //@line 1849 "./libraries/stb/stb_image.h"
                if (!($855)) {
                 $858 = (($854) + -1)|0; //@line 1850 "./libraries/stb/stb_image.h"
                 HEAP32[$125>>2] = $858; //@line 1850 "./libraries/stb/stb_image.h"
                 $859 = HEAP32[$114>>2]|0; //@line 1851 "./libraries/stb/stb_image.h"
                 $860 = ($588|0)>($859|0); //@line 1851 "./libraries/stb/stb_image.h"
                 if ($860) {
                  break;
                 }
                 $sext11$i$i$i$i$i = $853 << 16; //@line 1855 "./libraries/stb/stb_image.h"
                 $861 = $sext11$i$i$i$i$i >> 16; //@line 1855 "./libraries/stb/stb_image.h"
                 $k$5165$i$i$i$i$i = $588;
                 while(1) {
                  $862 = (440770 + ($k$5165$i$i$i$i$i)|0); //@line 1852 "./libraries/stb/stb_image.h"
                  $863 = HEAP8[$862>>0]|0; //@line 1852 "./libraries/stb/stb_image.h"
                  $864 = $863&255; //@line 1852 "./libraries/stb/stb_image.h"
                  $865 = (($587) + ($864<<1)|0); //@line 1852 "./libraries/stb/stb_image.h"
                  $866 = HEAP16[$865>>1]|0; //@line 1853 "./libraries/stb/stb_image.h"
                  $867 = ($866<<16>>16)==(0); //@line 1853 "./libraries/stb/stb_image.h"
                  do {
                   if (!($867)) {
                    $868 = HEAP32[$118>>2]|0; //@line 1692 "./libraries/stb/stb_image.h"
                    $869 = ($868|0)<(1); //@line 1692 "./libraries/stb/stb_image.h"
                    L428: do {
                     if ($869) {
                      L429: while(1) {
                       $870 = HEAP32[$120>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                       $871 = ($870|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
                       do {
                        if ($871) {
                         $872 = HEAP32[$j$i>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                         $873 = ((($872)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                         $874 = HEAP32[$873>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                         $875 = ((($872)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                         $876 = HEAP32[$875>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                         $877 = ($874>>>0)<($876>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                         if ($877) {
                          $878 = ((($874)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                          HEAP32[$873>>2] = $878; //@line 1230 "./libraries/stb/stb_image.h"
                          $879 = HEAP8[$874>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                          $$0$i68$i$i$i$i$i = $879;
                         } else {
                          $880 = ((($872)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                          $881 = HEAP32[$880>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                          $882 = ($881|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                          if ($882) {
                           $922 = 0;
                           break;
                          }
                          $883 = ((($872)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $884 = HEAP32[$883>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                          $885 = ((($872)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $886 = HEAP32[$885>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                          $887 = ((($872)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $888 = ((($872)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $889 = HEAP32[$888>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                          $890 = (FUNCTION_TABLE_iiii[$884 & 8191]($886,$887,$889)|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $891 = ($890|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                          if ($891) {
                           HEAP32[$880>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                           $892 = ((($872)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                           HEAP8[$887>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                           $$pre$phi$i65$i$i$i$i$iZ2D = $892;$$sink$i$i66$i$i$i$i$i = $892;$3737 = 0;
                          } else {
                           $893 = (((($872)) + 40|0) + ($890)|0); //@line 1223 "./libraries/stb/stb_image.h"
                           $$pre$i63$i$i$i$i$i = HEAP8[$887>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                           $$pre1$i64$i$i$i$i$i = ((($872)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                           $$pre$phi$i65$i$i$i$i$iZ2D = $$pre1$i64$i$i$i$i$i;$$sink$i$i66$i$i$i$i$i = $893;$3737 = $$pre$i63$i$i$i$i$i;
                          }
                          HEAP32[$875>>2] = $$sink$i$i66$i$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                          HEAP32[$873>>2] = $$pre$phi$i65$i$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                          $$0$i68$i$i$i$i$i = $3737;
                         }
                         $894 = $$0$i68$i$i$i$i$i&255; //@line 1590 "./libraries/stb/stb_image.h"
                         $895 = ($$0$i68$i$i$i$i$i<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
                         if (!($895)) {
                          $922 = $894;
                          break;
                         }
                         $896 = HEAP32[$j$i>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
                         $897 = ((($896)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                         $898 = HEAP32[$897>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                         $899 = ((($896)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                         $900 = HEAP32[$899>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                         $901 = ($898>>>0)<($900>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                         if ($901) {
                          $902 = ((($898)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                          HEAP32[$897>>2] = $902; //@line 1230 "./libraries/stb/stb_image.h"
                          $903 = HEAP8[$898>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                          $$0$i$i57$i$i$i$i$i = $903;
                         } else {
                          $904 = ((($896)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                          $905 = HEAP32[$904>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                          $906 = ($905|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                          if ($906) {
                           $922 = 255;
                           break;
                          }
                          $907 = ((($896)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $908 = HEAP32[$907>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                          $909 = ((($896)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $910 = HEAP32[$909>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                          $911 = ((($896)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $912 = ((($896)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $913 = HEAP32[$912>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                          $914 = (FUNCTION_TABLE_iiii[$908 & 8191]($910,$911,$913)|0); //@line 1213 "./libraries/stb/stb_image.h"
                          $915 = ($914|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                          if ($915) {
                           HEAP32[$904>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                           $916 = ((($896)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                           HEAP8[$911>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                           $$pre$phi$i$i54$i$i$i$i$iZ2D = $916;$$sink$i$i$i55$i$i$i$i$i = $916;$3738 = 0;
                          } else {
                           $917 = (((($896)) + 40|0) + ($914)|0); //@line 1223 "./libraries/stb/stb_image.h"
                           $$pre$i$i52$i$i$i$i$i = HEAP8[$911>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                           $$pre1$i$i53$i$i$i$i$i = ((($896)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                           $$pre$phi$i$i54$i$i$i$i$iZ2D = $$pre1$i$i53$i$i$i$i$i;$$sink$i$i$i55$i$i$i$i$i = $917;$3738 = $$pre$i$i52$i$i$i$i$i;
                          }
                          HEAP32[$899>>2] = $$sink$i$i$i55$i$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                          HEAP32[$897>>2] = $$pre$phi$i$i54$i$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                          $$0$i$i57$i$i$i$i$i = $3738;
                         }
                         $918 = ($$0$i$i57$i$i$i$i$i<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
                         if ($918) {
                          $922 = 255;
                         } else {
                          $$0$i$i57$i$i$i$i$i$lcssa = $$0$i$i57$i$i$i$i$i;
                          break L429;
                         }
                        } else {
                         $922 = 0;
                        }
                       } while(0);
                       $919 = HEAP32[$118>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                       $920 = (24 - ($919))|0; //@line 1599 "./libraries/stb/stb_image.h"
                       $921 = $922 << $920; //@line 1599 "./libraries/stb/stb_image.h"
                       $923 = HEAP32[$119>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                       $924 = $921 | $923; //@line 1599 "./libraries/stb/stb_image.h"
                       HEAP32[$119>>2] = $924; //@line 1599 "./libraries/stb/stb_image.h"
                       $925 = (($919) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
                       HEAP32[$118>>2] = $925; //@line 1600 "./libraries/stb/stb_image.h"
                       $926 = ($925|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
                       if (!($926)) {
                        $930 = $925;
                        break L428;
                       }
                      }
                      HEAP8[$63>>0] = $$0$i$i57$i$i$i$i$i$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
                      HEAP32[$120>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
                      $$pre$i59$i$i$i$i$i = HEAP32[$118>>2]|0; //@line 1695 "./libraries/stb/stb_image.h"
                      $930 = $$pre$i59$i$i$i$i$i;
                     } else {
                      $930 = $868;
                     }
                    } while(0);
                    $927 = HEAP32[$119>>2]|0; //@line 1693 "./libraries/stb/stb_image.h"
                    $928 = $927 << 1; //@line 1694 "./libraries/stb/stb_image.h"
                    HEAP32[$119>>2] = $928; //@line 1694 "./libraries/stb/stb_image.h"
                    $929 = (($930) + -1)|0; //@line 1695 "./libraries/stb/stb_image.h"
                    HEAP32[$118>>2] = $929; //@line 1695 "./libraries/stb/stb_image.h"
                    $931 = ($927|0)>(-1); //@line 1854 "./libraries/stb/stb_image.h"
                    if ($931) {
                     break;
                    }
                    $932 = HEAP16[$865>>1]|0; //@line 1855 "./libraries/stb/stb_image.h"
                    $933 = $932 << 16 >> 16; //@line 1855 "./libraries/stb/stb_image.h"
                    $934 = $933 & $861; //@line 1855 "./libraries/stb/stb_image.h"
                    $935 = ($934|0)==(0); //@line 1855 "./libraries/stb/stb_image.h"
                    if (!($935)) {
                     break;
                    }
                    $936 = ($932<<16>>16)>(0); //@line 1856 "./libraries/stb/stb_image.h"
                    if ($936) {
                     $937 = (($933) + ($861))|0; //@line 1857 "./libraries/stb/stb_image.h"
                     $938 = $937&65535; //@line 1857 "./libraries/stb/stb_image.h"
                     HEAP16[$865>>1] = $938; //@line 1857 "./libraries/stb/stb_image.h"
                     break;
                    } else {
                     $939 = (($933) - ($861))|0; //@line 1859 "./libraries/stb/stb_image.h"
                     $940 = $939&65535; //@line 1859 "./libraries/stb/stb_image.h"
                     HEAP16[$865>>1] = $940; //@line 1859 "./libraries/stb/stb_image.h"
                     break;
                    }
                   }
                  } while(0);
                  $941 = (($k$5165$i$i$i$i$i) + 1)|0; //@line 1851 "./libraries/stb/stb_image.h"
                  $942 = HEAP32[$114>>2]|0; //@line 1851 "./libraries/stb/stb_image.h"
                  $943 = ($k$5165$i$i$i$i$i|0)<($942|0); //@line 1851 "./libraries/stb/stb_image.h"
                  if ($943) {
                   $k$5165$i$i$i$i$i = $941;
                  } else {
                   break L317;
                  }
                 }
                }
                $sext$i$i$i$i$i = $853 << 16; //@line 1885 "./libraries/stb/stb_image.h"
                $856 = $sext$i$i$i$i$i >> 16; //@line 1885 "./libraries/stb/stb_image.h"
                $857 = (0 - ($856))|0; //@line 1887 "./libraries/stb/stb_image.h"
                $k$6$i$i$i$i$i = $588;
                while(1) {
                 $944 = (__ZL22stbi__jpeg_huff_decodeP10stbi__jpegP13stbi__huffman($j$i,$595)|0); //@line 1866 "./libraries/stb/stb_image.h"
                 $945 = ($944|0)<(0); //@line 1867 "./libraries/stb/stb_image.h"
                 if ($945) {
                  label = 306;
                  break L57;
                 }
                 $946 = $944 & 15; //@line 1868 "./libraries/stb/stb_image.h"
                 $947 = $944 >> 4; //@line 1869 "./libraries/stb/stb_image.h"
                 L465: do {
                  switch ($946|0) {
                  case 0:  {
                   $948 = ($947|0)<(15); //@line 1871 "./libraries/stb/stb_image.h"
                   if (!($948)) {
                    $r3$0$ph$i$i$i$i$i = $947;$s4$0$ph$i$i$i$i$i = 0;
                    break L465;
                   }
                   $949 = 1 << $947; //@line 1872 "./libraries/stb/stb_image.h"
                   $950 = (($949) + -1)|0; //@line 1872 "./libraries/stb/stb_image.h"
                   HEAP32[$125>>2] = $950; //@line 1872 "./libraries/stb/stb_image.h"
                   $951 = ($947|0)==(0); //@line 1873 "./libraries/stb/stb_image.h"
                   if ($951) {
                    $r3$0$ph$i$i$i$i$i = 64;$s4$0$ph$i$i$i$i$i = 0;
                    break L465;
                   }
                   $952 = HEAP32[$118>>2]|0; //@line 1681 "./libraries/stb/stb_image.h"
                   $953 = ($952|0)<($947|0); //@line 1681 "./libraries/stb/stb_image.h"
                   L469: do {
                    if ($953) {
                     L470: while(1) {
                      $954 = HEAP32[$120>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                      $955 = ($954|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
                      do {
                       if ($955) {
                        $956 = HEAP32[$j$i>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                        $957 = ((($956)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                        $958 = HEAP32[$957>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                        $959 = ((($956)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                        $960 = HEAP32[$959>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                        $961 = ($958>>>0)<($960>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                        if ($961) {
                         $962 = ((($958)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                         HEAP32[$957>>2] = $962; //@line 1230 "./libraries/stb/stb_image.h"
                         $963 = HEAP8[$958>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                         $$0$i81$i$i$i$i$i = $963;
                        } else {
                         $964 = ((($956)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                         $965 = HEAP32[$964>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                         $966 = ($965|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                         if ($966) {
                          $1006 = 0;
                          break;
                         }
                         $967 = ((($956)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $968 = HEAP32[$967>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                         $969 = ((($956)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $970 = HEAP32[$969>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                         $971 = ((($956)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $972 = ((($956)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $973 = HEAP32[$972>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                         $974 = (FUNCTION_TABLE_iiii[$968 & 8191]($970,$971,$973)|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $975 = ($974|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                         if ($975) {
                          HEAP32[$964>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                          $976 = ((($956)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                          HEAP8[$971>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                          $$pre$phi$i78$i$i$i$i$iZ2D = $976;$$sink$i$i79$i$i$i$i$i = $976;$3739 = 0;
                         } else {
                          $977 = (((($956)) + 40|0) + ($974)|0); //@line 1223 "./libraries/stb/stb_image.h"
                          $$pre$i76$i$i$i$i$i = HEAP8[$971>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                          $$pre1$i77$i$i$i$i$i = ((($956)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                          $$pre$phi$i78$i$i$i$i$iZ2D = $$pre1$i77$i$i$i$i$i;$$sink$i$i79$i$i$i$i$i = $977;$3739 = $$pre$i76$i$i$i$i$i;
                         }
                         HEAP32[$959>>2] = $$sink$i$i79$i$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                         HEAP32[$957>>2] = $$pre$phi$i78$i$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                         $$0$i81$i$i$i$i$i = $3739;
                        }
                        $978 = $$0$i81$i$i$i$i$i&255; //@line 1590 "./libraries/stb/stb_image.h"
                        $979 = ($$0$i81$i$i$i$i$i<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
                        if (!($979)) {
                         $1006 = $978;
                         break;
                        }
                        $980 = HEAP32[$j$i>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
                        $981 = ((($980)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                        $982 = HEAP32[$981>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                        $983 = ((($980)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                        $984 = HEAP32[$983>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                        $985 = ($982>>>0)<($984>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                        if ($985) {
                         $986 = ((($982)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                         HEAP32[$981>>2] = $986; //@line 1230 "./libraries/stb/stb_image.h"
                         $987 = HEAP8[$982>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                         $$0$i88$i$i$i$i$i = $987;
                        } else {
                         $988 = ((($980)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                         $989 = HEAP32[$988>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                         $990 = ($989|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                         if ($990) {
                          $1006 = 255;
                          break;
                         }
                         $991 = ((($980)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $992 = HEAP32[$991>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                         $993 = ((($980)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $994 = HEAP32[$993>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                         $995 = ((($980)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $996 = ((($980)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $997 = HEAP32[$996>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                         $998 = (FUNCTION_TABLE_iiii[$992 & 8191]($994,$995,$997)|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $999 = ($998|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                         if ($999) {
                          HEAP32[$988>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                          $1000 = ((($980)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                          HEAP8[$995>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                          $$pre$phi$i85$i$i$i$i$iZ2D = $1000;$$sink$i$i86$i$i$i$i$i = $1000;$3740 = 0;
                         } else {
                          $1001 = (((($980)) + 40|0) + ($998)|0); //@line 1223 "./libraries/stb/stb_image.h"
                          $$pre$i83$i$i$i$i$i = HEAP8[$995>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                          $$pre1$i84$i$i$i$i$i = ((($980)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                          $$pre$phi$i85$i$i$i$i$iZ2D = $$pre1$i84$i$i$i$i$i;$$sink$i$i86$i$i$i$i$i = $1001;$3740 = $$pre$i83$i$i$i$i$i;
                         }
                         HEAP32[$983>>2] = $$sink$i$i86$i$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                         HEAP32[$981>>2] = $$pre$phi$i85$i$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                         $$0$i88$i$i$i$i$i = $3740;
                        }
                        $1002 = ($$0$i88$i$i$i$i$i<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
                        if ($1002) {
                         $1006 = 255;
                        } else {
                         $$0$i88$i$i$i$i$i$lcssa = $$0$i88$i$i$i$i$i;
                         break L470;
                        }
                       } else {
                        $1006 = 0;
                       }
                      } while(0);
                      $1003 = HEAP32[$118>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                      $1004 = (24 - ($1003))|0; //@line 1599 "./libraries/stb/stb_image.h"
                      $1005 = $1006 << $1004; //@line 1599 "./libraries/stb/stb_image.h"
                      $1007 = HEAP32[$119>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                      $1008 = $1005 | $1007; //@line 1599 "./libraries/stb/stb_image.h"
                      HEAP32[$119>>2] = $1008; //@line 1599 "./libraries/stb/stb_image.h"
                      $1009 = (($1003) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
                      HEAP32[$118>>2] = $1009; //@line 1600 "./libraries/stb/stb_image.h"
                      $1010 = ($1009|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
                      if (!($1010)) {
                       $1022 = $1009;
                       break L469;
                      }
                     }
                     HEAP8[$63>>0] = $$0$i88$i$i$i$i$i$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
                     HEAP32[$120>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
                     $$pre$i72$i$i$i$i$i = HEAP32[$118>>2]|0; //@line 1685 "./libraries/stb/stb_image.h"
                     $1022 = $$pre$i72$i$i$i$i$i;
                    } else {
                     $1022 = $952;
                    }
                   } while(0);
                   $1011 = HEAP32[$119>>2]|0; //@line 1682 "./libraries/stb/stb_image.h"
                   $1012 = $1011 << $947; //@line 1682 "./libraries/stb/stb_image.h"
                   $1013 = (32 - ($947))|0; //@line 1682 "./libraries/stb/stb_image.h"
                   $1014 = $1011 >>> $1013; //@line 1682 "./libraries/stb/stb_image.h"
                   $1015 = $1012 | $1014; //@line 1682 "./libraries/stb/stb_image.h"
                   $1016 = (67056 + ($947<<2)|0); //@line 1683 "./libraries/stb/stb_image.h"
                   $1017 = HEAP32[$1016>>2]|0; //@line 1683 "./libraries/stb/stb_image.h"
                   $1018 = $1017 ^ -1; //@line 1683 "./libraries/stb/stb_image.h"
                   $1019 = $1015 & $1018; //@line 1683 "./libraries/stb/stb_image.h"
                   HEAP32[$119>>2] = $1019; //@line 1683 "./libraries/stb/stb_image.h"
                   $1020 = $1015 & $1017; //@line 1684 "./libraries/stb/stb_image.h"
                   $1021 = (($1022) - ($947))|0; //@line 1685 "./libraries/stb/stb_image.h"
                   HEAP32[$118>>2] = $1021; //@line 1685 "./libraries/stb/stb_image.h"
                   $1023 = HEAP32[$125>>2]|0; //@line 1874 "./libraries/stb/stb_image.h"
                   $1024 = (($1023) + ($1020))|0; //@line 1874 "./libraries/stb/stb_image.h"
                   HEAP32[$125>>2] = $1024; //@line 1874 "./libraries/stb/stb_image.h"
                   $r3$0$ph$i$i$i$i$i = 64;$s4$0$ph$i$i$i$i$i = 0;
                   break;
                  }
                  case 1:  {
                   $1025 = HEAP32[$118>>2]|0; //@line 1692 "./libraries/stb/stb_image.h"
                   $1026 = ($1025|0)<(1); //@line 1692 "./libraries/stb/stb_image.h"
                   L497: do {
                    if ($1026) {
                     L498: while(1) {
                      $1027 = HEAP32[$120>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                      $1028 = ($1027|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
                      do {
                       if ($1028) {
                        $1029 = HEAP32[$j$i>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                        $1030 = ((($1029)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                        $1031 = HEAP32[$1030>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                        $1032 = ((($1029)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                        $1033 = HEAP32[$1032>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                        $1034 = ($1031>>>0)<($1033>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                        if ($1034) {
                         $1035 = ((($1031)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                         HEAP32[$1030>>2] = $1035; //@line 1230 "./libraries/stb/stb_image.h"
                         $1036 = HEAP8[$1031>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                         $$0$i108$i$i$i$i$i = $1036;
                        } else {
                         $1037 = ((($1029)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                         $1038 = HEAP32[$1037>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                         $1039 = ($1038|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                         if ($1039) {
                          $1079 = 0;
                          break;
                         }
                         $1040 = ((($1029)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $1041 = HEAP32[$1040>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                         $1042 = ((($1029)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $1043 = HEAP32[$1042>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                         $1044 = ((($1029)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $1045 = ((($1029)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $1046 = HEAP32[$1045>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                         $1047 = (FUNCTION_TABLE_iiii[$1041 & 8191]($1043,$1044,$1046)|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $1048 = ($1047|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                         if ($1048) {
                          HEAP32[$1037>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                          $1049 = ((($1029)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                          HEAP8[$1044>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                          $$pre$phi$i105$i$i$i$i$iZ2D = $1049;$$sink$i$i106$i$i$i$i$i = $1049;$3741 = 0;
                         } else {
                          $1050 = (((($1029)) + 40|0) + ($1047)|0); //@line 1223 "./libraries/stb/stb_image.h"
                          $$pre$i103$i$i$i$i$i = HEAP8[$1044>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                          $$pre1$i104$i$i$i$i$i = ((($1029)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                          $$pre$phi$i105$i$i$i$i$iZ2D = $$pre1$i104$i$i$i$i$i;$$sink$i$i106$i$i$i$i$i = $1050;$3741 = $$pre$i103$i$i$i$i$i;
                         }
                         HEAP32[$1032>>2] = $$sink$i$i106$i$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                         HEAP32[$1030>>2] = $$pre$phi$i105$i$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                         $$0$i108$i$i$i$i$i = $3741;
                        }
                        $1051 = $$0$i108$i$i$i$i$i&255; //@line 1590 "./libraries/stb/stb_image.h"
                        $1052 = ($$0$i108$i$i$i$i$i<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
                        if (!($1052)) {
                         $1079 = $1051;
                         break;
                        }
                        $1053 = HEAP32[$j$i>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
                        $1054 = ((($1053)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                        $1055 = HEAP32[$1054>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                        $1056 = ((($1053)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                        $1057 = HEAP32[$1056>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                        $1058 = ($1055>>>0)<($1057>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                        if ($1058) {
                         $1059 = ((($1055)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                         HEAP32[$1054>>2] = $1059; //@line 1230 "./libraries/stb/stb_image.h"
                         $1060 = HEAP8[$1055>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                         $$0$i$i97$i$i$i$i$i = $1060;
                        } else {
                         $1061 = ((($1053)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                         $1062 = HEAP32[$1061>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                         $1063 = ($1062|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                         if ($1063) {
                          $1079 = 255;
                          break;
                         }
                         $1064 = ((($1053)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $1065 = HEAP32[$1064>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                         $1066 = ((($1053)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $1067 = HEAP32[$1066>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                         $1068 = ((($1053)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $1069 = ((($1053)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $1070 = HEAP32[$1069>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                         $1071 = (FUNCTION_TABLE_iiii[$1065 & 8191]($1067,$1068,$1070)|0); //@line 1213 "./libraries/stb/stb_image.h"
                         $1072 = ($1071|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                         if ($1072) {
                          HEAP32[$1061>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                          $1073 = ((($1053)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                          HEAP8[$1068>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                          $$pre$phi$i$i94$i$i$i$i$iZ2D = $1073;$$sink$i$i$i95$i$i$i$i$i = $1073;$3742 = 0;
                         } else {
                          $1074 = (((($1053)) + 40|0) + ($1071)|0); //@line 1223 "./libraries/stb/stb_image.h"
                          $$pre$i$i92$i$i$i$i$i = HEAP8[$1068>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                          $$pre1$i$i93$i$i$i$i$i = ((($1053)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                          $$pre$phi$i$i94$i$i$i$i$iZ2D = $$pre1$i$i93$i$i$i$i$i;$$sink$i$i$i95$i$i$i$i$i = $1074;$3742 = $$pre$i$i92$i$i$i$i$i;
                         }
                         HEAP32[$1056>>2] = $$sink$i$i$i95$i$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                         HEAP32[$1054>>2] = $$pre$phi$i$i94$i$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                         $$0$i$i97$i$i$i$i$i = $3742;
                        }
                        $1075 = ($$0$i$i97$i$i$i$i$i<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
                        if ($1075) {
                         $1079 = 255;
                        } else {
                         $$0$i$i97$i$i$i$i$i$lcssa = $$0$i$i97$i$i$i$i$i;
                         break L498;
                        }
                       } else {
                        $1079 = 0;
                       }
                      } while(0);
                      $1076 = HEAP32[$118>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                      $1077 = (24 - ($1076))|0; //@line 1599 "./libraries/stb/stb_image.h"
                      $1078 = $1079 << $1077; //@line 1599 "./libraries/stb/stb_image.h"
                      $1080 = HEAP32[$119>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                      $1081 = $1078 | $1080; //@line 1599 "./libraries/stb/stb_image.h"
                      HEAP32[$119>>2] = $1081; //@line 1599 "./libraries/stb/stb_image.h"
                      $1082 = (($1076) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
                      HEAP32[$118>>2] = $1082; //@line 1600 "./libraries/stb/stb_image.h"
                      $1083 = ($1082|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
                      if (!($1083)) {
                       $1087 = $1082;
                       break L497;
                      }
                     }
                     HEAP8[$63>>0] = $$0$i$i97$i$i$i$i$i$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
                     HEAP32[$120>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
                     $$pre$i99$i$i$i$i$i = HEAP32[$118>>2]|0; //@line 1695 "./libraries/stb/stb_image.h"
                     $1087 = $$pre$i99$i$i$i$i$i;
                    } else {
                     $1087 = $1025;
                    }
                   } while(0);
                   $1084 = HEAP32[$119>>2]|0; //@line 1693 "./libraries/stb/stb_image.h"
                   $1085 = $1084 << 1; //@line 1694 "./libraries/stb/stb_image.h"
                   HEAP32[$119>>2] = $1085; //@line 1694 "./libraries/stb/stb_image.h"
                   $1086 = (($1087) + -1)|0; //@line 1695 "./libraries/stb/stb_image.h"
                   HEAP32[$118>>2] = $1086; //@line 1695 "./libraries/stb/stb_image.h"
                   $1088 = ($1084|0)>(-1); //@line 1884 "./libraries/stb/stb_image.h"
                   $$$i57$i$i$i$i = $1088 ? $857 : $856; //@line 1884 "./libraries/stb/stb_image.h"
                   $r3$0$ph$i$i$i$i$i = $947;$s4$0$ph$i$i$i$i$i = $$$i57$i$i$i$i;
                   break;
                  }
                  default: {
                   label = 331;
                   break L57;
                  }
                  }
                 } while(0);
                 $1089 = HEAP32[$114>>2]|0; //@line 1891 "./libraries/stb/stb_image.h"
                 $1090 = ($k$6$i$i$i$i$i|0)>($1089|0); //@line 1891 "./libraries/stb/stb_image.h"
                 L525: do {
                  if ($1090) {
                   $1178 = $1089;$k$9$ph$i$i$i$i$i = $k$6$i$i$i$i$i;
                  } else {
                   $3743 = $1089;$k$7$ph163$i$i$i$i$i = $k$6$i$i$i$i$i;$r3$0$ph141162$i$i$i$i$i = $r3$0$ph$i$i$i$i$i;
                   while(1) {
                    $3744 = $3743;$k$7160$i$i$i$i$i = $k$7$ph163$i$i$i$i$i;
                    while(1) {
                     $1093 = (($k$7160$i$i$i$i$i) + 1)|0; //@line 1892 "./libraries/stb/stb_image.h"
                     $1094 = (440770 + ($k$7160$i$i$i$i$i)|0); //@line 1892 "./libraries/stb/stb_image.h"
                     $1095 = HEAP8[$1094>>0]|0; //@line 1892 "./libraries/stb/stb_image.h"
                     $1096 = $1095&255; //@line 1892 "./libraries/stb/stb_image.h"
                     $1097 = (($587) + ($1096<<1)|0); //@line 1892 "./libraries/stb/stb_image.h"
                     $1098 = HEAP16[$1097>>1]|0; //@line 1893 "./libraries/stb/stb_image.h"
                     $1099 = ($1098<<16>>16)==(0); //@line 1893 "./libraries/stb/stb_image.h"
                     if ($1099) {
                      $$lcssa498 = $3744;$$lcssa502 = $1093;$$lcssa505 = $1097;$k$7160$i$i$i$i$i$lcssa = $k$7160$i$i$i$i$i;
                      break;
                     }
                     $1100 = HEAP32[$118>>2]|0; //@line 1692 "./libraries/stb/stb_image.h"
                     $1101 = ($1100|0)<(1); //@line 1692 "./libraries/stb/stb_image.h"
                     L531: do {
                      if ($1101) {
                       L532: while(1) {
                        $1102 = HEAP32[$120>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                        $1103 = ($1102|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
                        do {
                         if ($1103) {
                          $1104 = HEAP32[$j$i>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                          $1105 = ((($1104)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                          $1106 = HEAP32[$1105>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                          $1107 = ((($1104)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                          $1108 = HEAP32[$1107>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                          $1109 = ($1106>>>0)<($1108>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                          if ($1109) {
                           $1110 = ((($1106)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                           HEAP32[$1105>>2] = $1110; //@line 1230 "./libraries/stb/stb_image.h"
                           $1111 = HEAP8[$1106>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                           $$0$i$i$i25$i$i$i = $1111;
                          } else {
                           $1112 = ((($1104)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                           $1113 = HEAP32[$1112>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                           $1114 = ($1113|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                           if ($1114) {
                            $1154 = 0;
                            break;
                           }
                           $1115 = ((($1104)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                           $1116 = HEAP32[$1115>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                           $1117 = ((($1104)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                           $1118 = HEAP32[$1117>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                           $1119 = ((($1104)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                           $1120 = ((($1104)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                           $1121 = HEAP32[$1120>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                           $1122 = (FUNCTION_TABLE_iiii[$1116 & 8191]($1118,$1119,$1121)|0); //@line 1213 "./libraries/stb/stb_image.h"
                           $1123 = ($1122|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                           if ($1123) {
                            HEAP32[$1112>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                            $1124 = ((($1104)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                            HEAP8[$1119>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                            $$pre$phi$i14$i$i$i$i$iZ2D = $1124;$$sink$i$i$i$i23$i$i$i = $1124;$3745 = 0;
                           } else {
                            $1125 = (((($1104)) + 40|0) + ($1122)|0); //@line 1223 "./libraries/stb/stb_image.h"
                            $$pre$i12$i$i$i$i$i = HEAP8[$1119>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                            $$pre1$i13$i$i$i$i$i = ((($1104)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                            $$pre$phi$i14$i$i$i$i$iZ2D = $$pre1$i13$i$i$i$i$i;$$sink$i$i$i$i23$i$i$i = $1125;$3745 = $$pre$i12$i$i$i$i$i;
                           }
                           HEAP32[$1107>>2] = $$sink$i$i$i$i23$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                           HEAP32[$1105>>2] = $$pre$phi$i14$i$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                           $$0$i$i$i25$i$i$i = $3745;
                          }
                          $1126 = $$0$i$i$i25$i$i$i&255; //@line 1590 "./libraries/stb/stb_image.h"
                          $1127 = ($$0$i$i$i25$i$i$i<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
                          if (!($1127)) {
                           $1154 = $1126;
                           break;
                          }
                          $1128 = HEAP32[$j$i>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
                          $1129 = ((($1128)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                          $1130 = HEAP32[$1129>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                          $1131 = ((($1128)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                          $1132 = HEAP32[$1131>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                          $1133 = ($1130>>>0)<($1132>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                          if ($1133) {
                           $1134 = ((($1130)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                           HEAP32[$1129>>2] = $1134; //@line 1230 "./libraries/stb/stb_image.h"
                           $1135 = HEAP8[$1130>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                           $$0$i$i$i$i$i$i$i = $1135;
                          } else {
                           $1136 = ((($1128)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                           $1137 = HEAP32[$1136>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                           $1138 = ($1137|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                           if ($1138) {
                            $1154 = 255;
                            break;
                           }
                           $1139 = ((($1128)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                           $1140 = HEAP32[$1139>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                           $1141 = ((($1128)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                           $1142 = HEAP32[$1141>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                           $1143 = ((($1128)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                           $1144 = ((($1128)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                           $1145 = HEAP32[$1144>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                           $1146 = (FUNCTION_TABLE_iiii[$1140 & 8191]($1142,$1143,$1145)|0); //@line 1213 "./libraries/stb/stb_image.h"
                           $1147 = ($1146|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                           if ($1147) {
                            HEAP32[$1136>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                            $1148 = ((($1128)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                            HEAP8[$1143>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                            $$pre$phi$i$i$i$i$i$i$iZ2D = $1148;$$sink$i$i$i$i$i$i$i$i = $1148;$3746 = 0;
                           } else {
                            $1149 = (((($1128)) + 40|0) + ($1146)|0); //@line 1223 "./libraries/stb/stb_image.h"
                            $$pre$i$i$i$i$i$i$i = HEAP8[$1143>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                            $$pre1$i$i$i$i$i$i$i = ((($1128)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                            $$pre$phi$i$i$i$i$i$i$iZ2D = $$pre1$i$i$i$i$i$i$i;$$sink$i$i$i$i$i$i$i$i = $1149;$3746 = $$pre$i$i$i$i$i$i$i;
                           }
                           HEAP32[$1131>>2] = $$sink$i$i$i$i$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                           HEAP32[$1129>>2] = $$pre$phi$i$i$i$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                           $$0$i$i$i$i$i$i$i = $3746;
                          }
                          $1150 = ($$0$i$i$i$i$i$i$i<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
                          if ($1150) {
                           $1154 = 255;
                          } else {
                           $$0$i$i$i$i$i$i$i$lcssa = $$0$i$i$i$i$i$i$i;
                           break L532;
                          }
                         } else {
                          $1154 = 0;
                         }
                        } while(0);
                        $1151 = HEAP32[$118>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                        $1152 = (24 - ($1151))|0; //@line 1599 "./libraries/stb/stb_image.h"
                        $1153 = $1154 << $1152; //@line 1599 "./libraries/stb/stb_image.h"
                        $1155 = HEAP32[$119>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                        $1156 = $1153 | $1155; //@line 1599 "./libraries/stb/stb_image.h"
                        HEAP32[$119>>2] = $1156; //@line 1599 "./libraries/stb/stb_image.h"
                        $1157 = (($1151) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
                        HEAP32[$118>>2] = $1157; //@line 1600 "./libraries/stb/stb_image.h"
                        $1158 = ($1157|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
                        if (!($1158)) {
                         $1162 = $1157;
                         break L531;
                        }
                       }
                       HEAP8[$63>>0] = $$0$i$i$i$i$i$i$i$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
                       HEAP32[$120>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
                       $$pre$i$i$i27$i$i$i = HEAP32[$118>>2]|0; //@line 1695 "./libraries/stb/stb_image.h"
                       $1162 = $$pre$i$i$i27$i$i$i;
                      } else {
                       $1162 = $1100;
                      }
                     } while(0);
                     $1159 = HEAP32[$119>>2]|0; //@line 1693 "./libraries/stb/stb_image.h"
                     $1160 = $1159 << 1; //@line 1694 "./libraries/stb/stb_image.h"
                     HEAP32[$119>>2] = $1160; //@line 1694 "./libraries/stb/stb_image.h"
                     $1161 = (($1162) + -1)|0; //@line 1695 "./libraries/stb/stb_image.h"
                     HEAP32[$118>>2] = $1161; //@line 1695 "./libraries/stb/stb_image.h"
                     $1163 = ($1159|0)>(-1); //@line 1894 "./libraries/stb/stb_image.h"
                     do {
                      if (!($1163)) {
                       $1166 = HEAP16[$1097>>1]|0; //@line 1895 "./libraries/stb/stb_image.h"
                       $1167 = $1166 << 16 >> 16; //@line 1895 "./libraries/stb/stb_image.h"
                       $1168 = $1167 & $856; //@line 1895 "./libraries/stb/stb_image.h"
                       $1169 = ($1168|0)==(0); //@line 1895 "./libraries/stb/stb_image.h"
                       if (!($1169)) {
                        break;
                       }
                       $1170 = ($1166<<16>>16)>(0); //@line 1896 "./libraries/stb/stb_image.h"
                       if ($1170) {
                        $1171 = (($1167) + ($856))|0; //@line 1897 "./libraries/stb/stb_image.h"
                        $1172 = $1171&65535; //@line 1897 "./libraries/stb/stb_image.h"
                        HEAP16[$1097>>1] = $1172; //@line 1897 "./libraries/stb/stb_image.h"
                        break;
                       } else {
                        $1173 = (($1167) - ($856))|0; //@line 1899 "./libraries/stb/stb_image.h"
                        $1174 = $1173&65535; //@line 1899 "./libraries/stb/stb_image.h"
                        HEAP16[$1097>>1] = $1174; //@line 1899 "./libraries/stb/stb_image.h"
                        break;
                       }
                      }
                     } while(0);
                     $1164 = HEAP32[$114>>2]|0; //@line 1891 "./libraries/stb/stb_image.h"
                     $1165 = ($k$7160$i$i$i$i$i|0)<($1164|0); //@line 1891 "./libraries/stb/stb_image.h"
                     if ($1165) {
                      $3744 = $1164;$k$7160$i$i$i$i$i = $1093;
                     } else {
                      $1178 = $1164;$k$9$ph$i$i$i$i$i = $1093;
                      break L525;
                     }
                    }
                    $1175 = ($r3$0$ph141162$i$i$i$i$i|0)==(0); //@line 1902 "./libraries/stb/stb_image.h"
                    if ($1175) {
                     $$lcssa498$lcssa = $$lcssa498;$$lcssa502$lcssa = $$lcssa502;$$lcssa505$lcssa = $$lcssa505;
                     break;
                    }
                    $1091 = (($r3$0$ph141162$i$i$i$i$i) + -1)|0; //@line 1906 "./libraries/stb/stb_image.h"
                    $1092 = ($k$7160$i$i$i$i$i$lcssa|0)<($$lcssa498|0); //@line 1891 "./libraries/stb/stb_image.h"
                    if ($1092) {
                     $3743 = $$lcssa498;$k$7$ph163$i$i$i$i$i = $$lcssa502;$r3$0$ph141162$i$i$i$i$i = $1091;
                    } else {
                     $1178 = $$lcssa498;$k$9$ph$i$i$i$i$i = $$lcssa502;
                     break L525;
                    }
                   }
                   $1176 = $s4$0$ph$i$i$i$i$i&65535; //@line 1903 "./libraries/stb/stb_image.h"
                   HEAP16[$$lcssa505$lcssa>>1] = $1176; //@line 1903 "./libraries/stb/stb_image.h"
                   $1178 = $$lcssa498$lcssa;$k$9$ph$i$i$i$i$i = $$lcssa502$lcssa;
                  }
                 } while(0);
                 $1177 = ($k$9$ph$i$i$i$i$i|0)>($1178|0); //@line 1909 "./libraries/stb/stb_image.h"
                 if ($1177) {
                  break L317;
                 } else {
                  $k$6$i$i$i$i$i = $k$9$ph$i$i$i$i$i;
                 }
                }
               }
              } while(0);
              $1179 = HEAP32[$124>>2]|0; //@line 2532 "./libraries/stb/stb_image.h"
              $1180 = (($1179) + -1)|0; //@line 2532 "./libraries/stb/stb_image.h"
              HEAP32[$124>>2] = $1180; //@line 2532 "./libraries/stb/stb_image.h"
              $1181 = ($1179|0)<(2); //@line 2532 "./libraries/stb/stb_image.h"
              if ($1181) {
               $1182 = HEAP32[$118>>2]|0; //@line 2533 "./libraries/stb/stb_image.h"
               $1183 = ($1182|0)<(24); //@line 2533 "./libraries/stb/stb_image.h"
               L572: do {
                if ($1183) {
                 L573: while(1) {
                  $1184 = HEAP32[$120>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                  $1185 = ($1184|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
                  do {
                   if ($1185) {
                    $1186 = HEAP32[$j$i>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
                    $1187 = ((($1186)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                    $1188 = HEAP32[$1187>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                    $1189 = ((($1186)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                    $1190 = HEAP32[$1189>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                    $1191 = ($1188>>>0)<($1190>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                    if ($1191) {
                     $1192 = ((($1188)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                     HEAP32[$1187>>2] = $1192; //@line 1230 "./libraries/stb/stb_image.h"
                     $1193 = HEAP8[$1188>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                     $$0$i74$i$i$i$i = $1193;
                    } else {
                     $1194 = ((($1186)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                     $1195 = HEAP32[$1194>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                     $1196 = ($1195|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                     if ($1196) {
                      $1236 = 0;
                      break;
                     }
                     $1197 = ((($1186)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                     $1198 = HEAP32[$1197>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                     $1199 = ((($1186)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                     $1200 = HEAP32[$1199>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                     $1201 = ((($1186)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                     $1202 = ((($1186)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                     $1203 = HEAP32[$1202>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                     $1204 = (FUNCTION_TABLE_iiii[$1198 & 8191]($1200,$1201,$1203)|0); //@line 1213 "./libraries/stb/stb_image.h"
                     $1205 = ($1204|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                     if ($1205) {
                      HEAP32[$1194>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                      $1206 = ((($1186)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                      HEAP8[$1201>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                      $$pre$phi$i71$i$i$i$iZ2D = $1206;$$sink$i$i72$i$i$i$i = $1206;$3747 = 0;
                     } else {
                      $1207 = (((($1186)) + 40|0) + ($1204)|0); //@line 1223 "./libraries/stb/stb_image.h"
                      $$pre$i69$i$i$i$i = HEAP8[$1201>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                      $$pre1$i70$i$i$i$i = ((($1186)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                      $$pre$phi$i71$i$i$i$iZ2D = $$pre1$i70$i$i$i$i;$$sink$i$i72$i$i$i$i = $1207;$3747 = $$pre$i69$i$i$i$i;
                     }
                     HEAP32[$1189>>2] = $$sink$i$i72$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                     HEAP32[$1187>>2] = $$pre$phi$i71$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                     $$0$i74$i$i$i$i = $3747;
                    }
                    $1208 = $$0$i74$i$i$i$i&255; //@line 1590 "./libraries/stb/stb_image.h"
                    $1209 = ($$0$i74$i$i$i$i<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
                    if (!($1209)) {
                     $1236 = $1208;
                     break;
                    }
                    $1210 = HEAP32[$j$i>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
                    $1211 = ((($1210)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
                    $1212 = HEAP32[$1211>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                    $1213 = ((($1210)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
                    $1214 = HEAP32[$1213>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                    $1215 = ($1212>>>0)<($1214>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                    if ($1215) {
                     $1216 = ((($1212)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                     HEAP32[$1211>>2] = $1216; //@line 1230 "./libraries/stb/stb_image.h"
                     $1217 = HEAP8[$1212>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                     $$0$i66$i$i$i$i = $1217;
                    } else {
                     $1218 = ((($1210)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
                     $1219 = HEAP32[$1218>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                     $1220 = ($1219|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                     if ($1220) {
                      $1236 = 255;
                      break;
                     }
                     $1221 = ((($1210)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
                     $1222 = HEAP32[$1221>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                     $1223 = ((($1210)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
                     $1224 = HEAP32[$1223>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                     $1225 = ((($1210)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
                     $1226 = ((($1210)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
                     $1227 = HEAP32[$1226>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                     $1228 = (FUNCTION_TABLE_iiii[$1222 & 8191]($1224,$1225,$1227)|0); //@line 1213 "./libraries/stb/stb_image.h"
                     $1229 = ($1228|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                     if ($1229) {
                      HEAP32[$1218>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                      $1230 = ((($1210)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
                      HEAP8[$1225>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                      $$pre$phi$i63$i$i$i$iZ2D = $1230;$$sink$i$i64$i$i$i$i = $1230;$3748 = 0;
                     } else {
                      $1231 = (((($1210)) + 40|0) + ($1228)|0); //@line 1223 "./libraries/stb/stb_image.h"
                      $$pre$i61$i$i$i$i = HEAP8[$1225>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                      $$pre1$i62$i$i$i$i = ((($1210)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
                      $$pre$phi$i63$i$i$i$iZ2D = $$pre1$i62$i$i$i$i;$$sink$i$i64$i$i$i$i = $1231;$3748 = $$pre$i61$i$i$i$i;
                     }
                     HEAP32[$1213>>2] = $$sink$i$i64$i$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
                     HEAP32[$1211>>2] = $$pre$phi$i63$i$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
                     $$0$i66$i$i$i$i = $3748;
                    }
                    $1232 = ($$0$i66$i$i$i$i<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
                    if ($1232) {
                     $1236 = 255;
                    } else {
                     $$0$i66$i$i$i$i$lcssa = $$0$i66$i$i$i$i;
                     break L573;
                    }
                   } else {
                    $1236 = 0;
                   }
                  } while(0);
                  $1233 = HEAP32[$118>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                  $1234 = (24 - ($1233))|0; //@line 1599 "./libraries/stb/stb_image.h"
                  $1235 = $1236 << $1234; //@line 1599 "./libraries/stb/stb_image.h"
                  $1237 = HEAP32[$119>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
                  $1238 = $1235 | $1237; //@line 1599 "./libraries/stb/stb_image.h"
                  HEAP32[$119>>2] = $1238; //@line 1599 "./libraries/stb/stb_image.h"
                  $1239 = (($1233) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
                  HEAP32[$118>>2] = $1239; //@line 1600 "./libraries/stb/stb_image.h"
                  $1240 = ($1239|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
                  if (!($1240)) {
                   break L572;
                  }
                 }
                 HEAP8[$63>>0] = $$0$i66$i$i$i$i$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
                 HEAP32[$120>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
                }
               } while(0);
               $1241 = HEAP8[$63>>0]|0; //@line 2534 "./libraries/stb/stb_image.h"
               $1242 = $1241 & -8; //@line 2534 "./libraries/stb/stb_image.h"
               $1243 = ($1242<<24>>24)==(-48); //@line 2534 "./libraries/stb/stb_image.h"
               if (!($1243)) {
                $1365 = $1241;
                break L147;
               }
               HEAP32[$118>>2] = 0; //@line 2438 "./libraries/stb/stb_image.h"
               HEAP32[$119>>2] = 0; //@line 2439 "./libraries/stb/stb_image.h"
               HEAP32[$120>>2] = 0; //@line 2440 "./libraries/stb/stb_image.h"
               HEAP32[$121>>2] = 0; //@line 2441 "./libraries/stb/stb_image.h"
               HEAP32[$122>>2] = 0; //@line 2441 "./libraries/stb/stb_image.h"
               HEAP32[$123>>2] = 0; //@line 2441 "./libraries/stb/stb_image.h"
               HEAP8[$63>>0] = -1; //@line 2442 "./libraries/stb/stb_image.h"
               $1244 = HEAP32[$60>>2]|0; //@line 2443 "./libraries/stb/stb_image.h"
               $1245 = ($1244|0)==(0); //@line 2443 "./libraries/stb/stb_image.h"
               $$$i68$i$i$i$i = $1245 ? 2147483647 : $1244; //@line 2443 "./libraries/stb/stb_image.h"
               HEAP32[$124>>2] = $$$i68$i$i$i$i; //@line 2443 "./libraries/stb/stb_image.h"
               HEAP32[$125>>2] = 0; //@line 2444 "./libraries/stb/stb_image.h"
              }
              $1246 = (($i8$0150$i$i$i$i) + 1)|0; //@line 2521 "./libraries/stb/stb_image.h"
              $1247 = ($1246|0)<($571|0); //@line 2521 "./libraries/stb/stb_image.h"
              if ($1247) {
               $i8$0150$i$i$i$i = $1246;
              } else {
               break;
              }
             }
            }
            $1248 = (($j9$0153$i$i$i$i) + 1)|0; //@line 2520 "./libraries/stb/stb_image.h"
            $1249 = ($1248|0)<($575|0); //@line 2520 "./libraries/stb/stb_image.h"
            if ($1249) {
             $j9$0153$i$i$i$i = $1248;
            } else {
             label = 449;
             break;
            }
           }
          }
         } while(0);
         if ((label|0) == 449) {
          label = 0;
          $$pr76$i$i$i = HEAP8[$63>>0]|0; //@line 2830 "./libraries/stb/stb_image.h"
          $1365 = $$pr76$i$i$i;
         }
         $1364 = ($1365<<24>>24)==(-1); //@line 2830 "./libraries/stb/stb_image.h"
         if (!($1364)) {
          $3758 = $1365;
          break L59;
         }
         L605: while(1) {
          $1366 = HEAP32[$j$i>>2]|0; //@line 2832 "./libraries/stb/stb_image.h"
          $1367 = ((($1366)) + 16|0); //@line 1240 "./libraries/stb/stb_image.h"
          $1368 = HEAP32[$1367>>2]|0; //@line 1240 "./libraries/stb/stb_image.h"
          $1369 = ($1368|0)==(0|0); //@line 1240 "./libraries/stb/stb_image.h"
          do {
           if ($1369) {
            label = 454;
           } else {
            $1370 = ((($1366)) + 24|0); //@line 1241 "./libraries/stb/stb_image.h"
            $1371 = HEAP32[$1370>>2]|0; //@line 1241 "./libraries/stb/stb_image.h"
            $1372 = ((($1366)) + 28|0); //@line 1241 "./libraries/stb/stb_image.h"
            $1373 = HEAP32[$1372>>2]|0; //@line 1241 "./libraries/stb/stb_image.h"
            $1374 = (FUNCTION_TABLE_ii[$1371 & 8191]($1373)|0); //@line 1241 "./libraries/stb/stb_image.h"
            $1375 = ($1374|0)==(0); //@line 1241 "./libraries/stb/stb_image.h"
            if ($1375) {
             break;
            }
            $1376 = ((($1366)) + 32|0); //@line 1244 "./libraries/stb/stb_image.h"
            $1377 = HEAP32[$1376>>2]|0; //@line 1244 "./libraries/stb/stb_image.h"
            $1378 = ($1377|0)==(0); //@line 1244 "./libraries/stb/stb_image.h"
            if ($1378) {
             label = 473;
             break L59;
            } else {
             label = 454;
            }
           }
          } while(0);
          if ((label|0) == 454) {
           label = 0;
           $1379 = ((($1366)) + 168|0); //@line 1247 "./libraries/stb/stb_image.h"
           $1380 = HEAP32[$1379>>2]|0; //@line 1247 "./libraries/stb/stb_image.h"
           $1381 = ((($1366)) + 172|0); //@line 1247 "./libraries/stb/stb_image.h"
           $1382 = HEAP32[$1381>>2]|0; //@line 1247 "./libraries/stb/stb_image.h"
           $1383 = ($1380>>>0)<($1382>>>0); //@line 1247 "./libraries/stb/stb_image.h"
           if (!($1383)) {
            label = 473;
            break L59;
           }
          }
          $1384 = HEAP32[$j$i>>2]|0; //@line 2833 "./libraries/stb/stb_image.h"
          $1385 = ((($1384)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
          $1386 = HEAP32[$1385>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $1387 = ((($1384)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
          $1388 = HEAP32[$1387>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $1389 = ($1386>>>0)<($1388>>>0); //@line 1229 "./libraries/stb/stb_image.h"
          if ($1389) {
           $1390 = ((($1386)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
           HEAP32[$1385>>2] = $1390; //@line 1230 "./libraries/stb/stb_image.h"
           $1391 = HEAP8[$1386>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
           $$0$i41$i$i$i = $1391;
          } else {
           $1392 = ((($1384)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $1393 = HEAP32[$1392>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
           $1394 = ($1393|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
           if ($1394) {
            continue;
           }
           $1395 = ((($1384)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1396 = HEAP32[$1395>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1397 = ((($1384)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1398 = HEAP32[$1397>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1399 = ((($1384)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1400 = ((($1384)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1401 = HEAP32[$1400>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1402 = (FUNCTION_TABLE_iiii[$1396 & 8191]($1398,$1399,$1401)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1403 = ($1402|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($1403) {
            HEAP32[$1392>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $1404 = ((($1384)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$1399>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i38$i$i$iZ2D = $1404;$$sink$i$i39$i$i$i = $1404;$3759 = 0;
           } else {
            $1405 = (((($1384)) + 40|0) + ($1402)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre$i36$i$i$i = HEAP8[$1399>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
            $$pre1$i37$i$i$i = ((($1384)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i38$i$i$iZ2D = $$pre1$i37$i$i$i;$$sink$i$i39$i$i$i = $1405;$3759 = $$pre$i36$i$i$i;
           }
           HEAP32[$1387>>2] = $$sink$i$i39$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$1385>>2] = $$pre$phi$i38$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
           $$0$i41$i$i$i = $3759;
          }
          switch ($$0$i41$i$i$i<<24>>24) {
          case 0:  {
           break;
          }
          case -1:  {
           break L605;
           break;
          }
          default: {
           label = 470;
           break L57;
          }
          }
         }
         $1406 = HEAP32[$j$i>>2]|0; //@line 2835 "./libraries/stb/stb_image.h"
         $1407 = ((($1406)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
         $1408 = HEAP32[$1407>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $1409 = ((($1406)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
         $1410 = HEAP32[$1409>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $1411 = ($1408>>>0)<($1410>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         do {
          if ($1411) {
           $1412 = ((($1408)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
           HEAP32[$1407>>2] = $1412; //@line 1230 "./libraries/stb/stb_image.h"
           $1413 = HEAP8[$1408>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
           $$0$i48$i$i$i = $1413;
          } else {
           $1414 = ((($1406)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $1415 = HEAP32[$1414>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
           $1416 = ($1415|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
           if ($1416) {
            $$0$i48$i$i$i = 0;
            break;
           }
           $1417 = ((($1406)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1418 = HEAP32[$1417>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1419 = ((($1406)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1420 = HEAP32[$1419>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1421 = ((($1406)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1422 = ((($1406)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1423 = HEAP32[$1422>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1424 = (FUNCTION_TABLE_iiii[$1418 & 8191]($1420,$1421,$1423)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1425 = ($1424|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($1425) {
            HEAP32[$1414>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $1426 = ((($1406)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$1421>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i45$i$i$iZ2D = $1426;$$sink$i$i46$i$i$i = $1426;$3760 = 0;
           } else {
            $1427 = (((($1406)) + 40|0) + ($1424)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre$i43$i$i$i = HEAP8[$1421>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
            $$pre1$i44$i$i$i = ((($1406)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i45$i$i$iZ2D = $$pre1$i44$i$i$i;$$sink$i$i46$i$i$i = $1427;$3760 = $$pre$i43$i$i$i;
           }
           HEAP32[$1409>>2] = $$sink$i$i46$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$1407>>2] = $$pre$phi$i45$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
           $$0$i48$i$i$i = $3760;
          }
         } while(0);
         HEAP8[$63>>0] = $$0$i48$i$i$i; //@line 2835 "./libraries/stb/stb_image.h"
         $1431 = $$0$i48$i$i$i;
         label = 474;
         break;
        }
        default: {
         $1428 = (__ZL20stbi__process_markerP10stbi__jpegi($j$i,$129)|0); //@line 2844 "./libraries/stb/stb_image.h"
         $1429 = ($1428|0)==(0); //@line 2844 "./libraries/stb/stb_image.h"
         if ($1429) {
          break L31;
         } else {
          label = 473;
         }
        }
        }
       } while(0);
       if ((label|0) == 473) {
        label = 0;
        $$pr$i$i$i = HEAP8[$63>>0]|0; //@line 2422 "./libraries/stb/stb_image.h"
        $1431 = $$pr$i$i$i;
        label = 474;
       }
       if ((label|0) == 474) {
        label = 0;
        $1430 = ($1431<<24>>24)==(-1); //@line 2422 "./libraries/stb/stb_image.h"
        if ($1430) {
         $1432 = HEAP32[$j$i>>2]|0; //@line 2423 "./libraries/stb/stb_image.h"
         $1433 = ((($1432)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
         $1434 = HEAP32[$1433>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $1435 = ((($1432)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
         $1436 = HEAP32[$1435>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $1437 = ($1434>>>0)<($1436>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         if ($1437) {
          $1438 = ((($1434)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
          HEAP32[$1433>>2] = $1438; //@line 1230 "./libraries/stb/stb_image.h"
          $1439 = HEAP8[$1434>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
          $$0$i$i55$i$i$i = $1439;
         } else {
          $1440 = ((($1432)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
          $1441 = HEAP32[$1440>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
          $1442 = ($1441|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
          if ($1442) {
           $$sink$i$i$i = -1;
           continue;
          }
          $1443 = ((($1432)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
          $1444 = HEAP32[$1443>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
          $1445 = ((($1432)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
          $1446 = HEAP32[$1445>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
          $1447 = ((($1432)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
          $1448 = ((($1432)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
          $1449 = HEAP32[$1448>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
          $1450 = (FUNCTION_TABLE_iiii[$1444 & 8191]($1446,$1447,$1449)|0); //@line 1213 "./libraries/stb/stb_image.h"
          $1451 = ($1450|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
          if ($1451) {
           HEAP32[$1440>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
           $1452 = ((($1432)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
           HEAP8[$1447>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
           $$pre$phi$i$i52$i$i$iZ2D = $1452;$$sink$i$i$i53$i$i$i = $1452;$3761 = 0;
          } else {
           $1453 = (((($1432)) + 40|0) + ($1450)|0); //@line 1223 "./libraries/stb/stb_image.h"
           $$pre$i$i50$i$i$i = HEAP8[$1447>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
           $$pre1$i$i51$i$i$i = ((($1432)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
           $$pre$phi$i$i52$i$i$iZ2D = $$pre1$i$i51$i$i$i;$$sink$i$i$i53$i$i$i = $1453;$3761 = $$pre$i$i50$i$i$i;
          }
          HEAP32[$1435>>2] = $$sink$i$i$i53$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
          HEAP32[$1433>>2] = $$pre$phi$i$i52$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
          $$0$i$i55$i$i$i = $3761;
         }
         $1454 = ($$0$i$i55$i$i$i<<24>>24)==(-1); //@line 2424 "./libraries/stb/stb_image.h"
         if (!($1454)) {
          $$sink$i$i$i = -1;
          continue;
         }
         while(1) {
          $1455 = HEAP32[$j$i>>2]|0; //@line 2426 "./libraries/stb/stb_image.h"
          $1456 = ((($1455)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
          $1457 = HEAP32[$1456>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $1458 = ((($1455)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
          $1459 = HEAP32[$1458>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $1460 = ($1457>>>0)<($1459>>>0); //@line 1229 "./libraries/stb/stb_image.h"
          if ($1460) {
           $1461 = ((($1457)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
           HEAP32[$1456>>2] = $1461; //@line 1230 "./libraries/stb/stb_image.h"
           $1462 = HEAP8[$1457>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
           $$0$i66$i$i$i = $1462;
          } else {
           $1463 = ((($1455)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $1464 = HEAP32[$1463>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
           $1465 = ($1464|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
           if ($1465) {
            $$sink$i$i$i = 0;
            continue L57;
           }
           $1466 = ((($1455)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1467 = HEAP32[$1466>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1468 = ((($1455)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1469 = HEAP32[$1468>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1470 = ((($1455)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1471 = ((($1455)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1472 = HEAP32[$1471>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1473 = (FUNCTION_TABLE_iiii[$1467 & 8191]($1469,$1470,$1472)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1474 = ($1473|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($1474) {
            HEAP32[$1463>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $1475 = ((($1455)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$1470>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i63$i$i$iZ2D = $1475;$$sink$i$i64$i$i$i = $1475;$3762 = 0;
           } else {
            $1476 = (((($1455)) + 40|0) + ($1473)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre$i61$i$i$i = HEAP8[$1470>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
            $$pre1$i62$i$i$i = ((($1455)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i63$i$i$iZ2D = $$pre1$i62$i$i$i;$$sink$i$i64$i$i$i = $1476;$3762 = $$pre$i61$i$i$i;
           }
           HEAP32[$1458>>2] = $$sink$i$i64$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$1456>>2] = $$pre$phi$i63$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
           $$0$i66$i$i$i = $3762;
          }
          $1477 = ($$0$i66$i$i$i<<24>>24)==(-1); //@line 2425 "./libraries/stb/stb_image.h"
          if (!($1477)) {
           $$sink$i$i$i = $$0$i66$i$i$i;
           continue L57;
          }
         }
        } else {
         $3758 = $1431;
        }
       }
       HEAP8[$63>>0] = -1; //@line 2422 "./libraries/stb/stb_image.h"
       $$sink$i$i$i = $3758;
      }
      switch (label|0) {
       case 60: {
        HEAP32[$112>>2] = 0; //@line 2667 "./libraries/stb/stb_image.h"
        break;
       }
       case 69: {
        HEAP32[158754] = 441067; //@line 894 "./libraries/stb/stb_image.h"
        break L31;
        break;
       }
       case 90: {
        HEAP32[158754] = 441079; //@line 894 "./libraries/stb/stb_image.h"
        break L31;
        break;
       }
       case 92: {
        HEAP32[158754] = 441091; //@line 894 "./libraries/stb/stb_image.h"
        break L31;
        break;
       }
       case 112: {
        HEAP32[158754] = 441103; //@line 894 "./libraries/stb/stb_image.h"
        break L31;
        break;
       }
       case 114: {
        HEAP32[158754] = 441103; //@line 894 "./libraries/stb/stb_image.h"
        break L31;
        break;
       }
       case 116: {
        HEAP32[158754] = 441103; //@line 894 "./libraries/stb/stb_image.h"
        break L31;
        break;
       }
       case 225: {
        HEAP32[158754] = 441250; //@line 894 "./libraries/stb/stb_image.h"
        break L31;
        break;
       }
       case 306: {
        HEAP32[158754] = 441250; //@line 894 "./libraries/stb/stb_image.h"
        break L31;
        break;
       }
       case 331: {
        HEAP32[158754] = 441250; //@line 894 "./libraries/stb/stb_image.h"
        break L31;
        break;
       }
       case 470: {
        HEAP32[158754] = 441352; //@line 894 "./libraries/stb/stb_image.h"
        break L31;
        break;
       }
       case 492: {
        $1478 = HEAP32[$117>>2]|0; //@line 2848 "./libraries/stb/stb_image.h"
        $1479 = ($1478|0)!=(0); //@line 2848 "./libraries/stb/stb_image.h"
        $$pre$i = HEAP32[$j$i>>2]|0; //@line 3324 "./libraries/stb/stb_image.h"
        $$phi$trans$insert$i = ((($$pre$i)) + 8|0); //@line 3324 "./libraries/stb/stb_image.h"
        $$pre93$i = HEAP32[$$phi$trans$insert$i>>2]|0; //@line 3324 "./libraries/stb/stb_image.h"
        $1480 = ($$pre93$i|0)>(0); //@line 2585 "./libraries/stb/stb_image.h"
        $or$cond$i = $1479 & $1480; //@line 2848 "./libraries/stb/stb_image.h"
        if ($or$cond$i) {
         $3763 = $$pre$i;$n$06$i$i$i$i = 0;
         while(1) {
          $1481 = (((((($j$i)) + 17820|0) + (($n$06$i$i$i$i*72)|0)|0)) + 28|0); //@line 2586 "./libraries/stb/stb_image.h"
          $1482 = HEAP32[$1481>>2]|0; //@line 2586 "./libraries/stb/stb_image.h"
          $1483 = (($1482) + 7)|0; //@line 2586 "./libraries/stb/stb_image.h"
          $1484 = $1483 >> 3; //@line 2586 "./libraries/stb/stb_image.h"
          $1485 = (((((($j$i)) + 17820|0) + (($n$06$i$i$i$i*72)|0)|0)) + 32|0); //@line 2587 "./libraries/stb/stb_image.h"
          $1486 = HEAP32[$1485>>2]|0; //@line 2587 "./libraries/stb/stb_image.h"
          $1487 = (($1486) + 7)|0; //@line 2587 "./libraries/stb/stb_image.h"
          $1488 = $1487 >> 3; //@line 2587 "./libraries/stb/stb_image.h"
          $1489 = ($1488|0)>(0); //@line 2588 "./libraries/stb/stb_image.h"
          if ($1489) {
           $1490 = ($1484|0)>(0); //@line 2589 "./libraries/stb/stb_image.h"
           $1491 = (((((($j$i)) + 17820|0) + (($n$06$i$i$i$i*72)|0)|0)) + 60|0); //@line 2590 "./libraries/stb/stb_image.h"
           $1492 = (((((($j$i)) + 17820|0) + (($n$06$i$i$i$i*72)|0)|0)) + 64|0); //@line 2590 "./libraries/stb/stb_image.h"
           $1493 = (((((($j$i)) + 17820|0) + (($n$06$i$i$i$i*72)|0)|0)) + 12|0); //@line 2591 "./libraries/stb/stb_image.h"
           $1494 = (((((($j$i)) + 17820|0) + (($n$06$i$i$i$i*72)|0)|0)) + 44|0); //@line 2592 "./libraries/stb/stb_image.h"
           $1495 = (((((($j$i)) + 17820|0) + (($n$06$i$i$i$i*72)|0)|0)) + 36|0); //@line 2592 "./libraries/stb/stb_image.h"
           if ($1490) {
            $j$03$us$i$i$i$i = 0;
            while(1) {
             $1516 = $j$03$us$i$i$i$i << 3; //@line 2592 "./libraries/stb/stb_image.h"
             $i$02$us$i$i$i$i = 0;
             while(1) {
              $1496 = HEAP32[$1491>>2]|0; //@line 2590 "./libraries/stb/stb_image.h"
              $1497 = HEAP32[$1492>>2]|0; //@line 2590 "./libraries/stb/stb_image.h"
              $1498 = Math_imul($1497, $j$03$us$i$i$i$i)|0; //@line 2590 "./libraries/stb/stb_image.h"
              $1499 = (($1498) + ($i$02$us$i$i$i$i))|0; //@line 2590 "./libraries/stb/stb_image.h"
              $1500 = $1499 << 6; //@line 2590 "./libraries/stb/stb_image.h"
              $1501 = (($1496) + ($1500<<1)|0); //@line 2590 "./libraries/stb/stb_image.h"
              $1502 = HEAP32[$1493>>2]|0; //@line 2591 "./libraries/stb/stb_image.h"
              $i$01$i$us$i$i$i$i = 0;
              while(1) {
               $1503 = ((((($j$i)) + 13444|0) + ($1502<<6)|0) + ($i$01$i$us$i$i$i$i)|0); //@line 2577 "./libraries/stb/stb_image.h"
               $1504 = HEAP8[$1503>>0]|0; //@line 2577 "./libraries/stb/stb_image.h"
               $1505 = $1504&255; //@line 2577 "./libraries/stb/stb_image.h"
               $1506 = (($1501) + ($i$01$i$us$i$i$i$i<<1)|0); //@line 2577 "./libraries/stb/stb_image.h"
               $1507 = HEAP16[$1506>>1]|0; //@line 2577 "./libraries/stb/stb_image.h"
               $1508 = $1507 << 16 >> 16; //@line 2577 "./libraries/stb/stb_image.h"
               $1509 = Math_imul($1508, $1505)|0; //@line 2577 "./libraries/stb/stb_image.h"
               $1510 = $1509&65535; //@line 2577 "./libraries/stb/stb_image.h"
               HEAP16[$1506>>1] = $1510; //@line 2577 "./libraries/stb/stb_image.h"
               $1511 = (($i$01$i$us$i$i$i$i) + 1)|0; //@line 2576 "./libraries/stb/stb_image.h"
               $exitcond$i$us$i$i$i$i = ($1511|0)==(64); //@line 2576 "./libraries/stb/stb_image.h"
               if ($exitcond$i$us$i$i$i$i) {
                break;
               } else {
                $i$01$i$us$i$i$i$i = $1511;
               }
              }
              $1512 = HEAP32[$47>>2]|0; //@line 2592 "./libraries/stb/stb_image.h"
              $1513 = HEAP32[$1494>>2]|0; //@line 2592 "./libraries/stb/stb_image.h"
              $1514 = HEAP32[$1495>>2]|0; //@line 2592 "./libraries/stb/stb_image.h"
              $1515 = Math_imul($1514, $1516)|0; //@line 2592 "./libraries/stb/stb_image.h"
              $1517 = (($1513) + ($1515)|0); //@line 2592 "./libraries/stb/stb_image.h"
              $1518 = $i$02$us$i$i$i$i << 3; //@line 2592 "./libraries/stb/stb_image.h"
              $1519 = (($1517) + ($1518)|0); //@line 2592 "./libraries/stb/stb_image.h"
              FUNCTION_TABLE_viii[$1512 & 8191]($1519,$1514,$1501); //@line 2592 "./libraries/stb/stb_image.h"
              $1520 = (($i$02$us$i$i$i$i) + 1)|0; //@line 2589 "./libraries/stb/stb_image.h"
              $exitcond10$i$i$i$i = ($1520|0)==($1484|0); //@line 2589 "./libraries/stb/stb_image.h"
              if ($exitcond10$i$i$i$i) {
               break;
              } else {
               $i$02$us$i$i$i$i = $1520;
              }
             }
             $1521 = (($j$03$us$i$i$i$i) + 1)|0; //@line 2588 "./libraries/stb/stb_image.h"
             $exitcond11$i$i$i$i = ($1521|0)==($1488|0); //@line 2588 "./libraries/stb/stb_image.h"
             if ($exitcond11$i$i$i$i) {
              break;
             } else {
              $j$03$us$i$i$i$i = $1521;
             }
            }
            $$pre$i69$i$i$i = HEAP32[$j$i>>2]|0; //@line 2585 "./libraries/stb/stb_image.h"
            $1524 = $$pre$i69$i$i$i;
           } else {
            $1524 = $3763;
           }
          } else {
           $1524 = $3763;
          }
          $1522 = (($n$06$i$i$i$i) + 1)|0; //@line 2585 "./libraries/stb/stb_image.h"
          $1523 = ((($1524)) + 8|0); //@line 2585 "./libraries/stb/stb_image.h"
          $1525 = HEAP32[$1523>>2]|0; //@line 2585 "./libraries/stb/stb_image.h"
          $1526 = ($1522|0)<($1525|0); //@line 2585 "./libraries/stb/stb_image.h"
          if ($1526) {
           $3763 = $1524;$n$06$i$i$i$i = $1522;
          } else {
           $1549 = $1524;$1550 = $1525;
           break;
          }
         }
        } else {
         $1549 = $$pre$i;$1550 = $$pre93$i;
        }
        $1547 = ($req_comp|0)==(0); //@line 3324 "./libraries/stb/stb_image.h"
        $1548 = ((($1549)) + 8|0); //@line 3324 "./libraries/stb/stb_image.h"
        $$req_comp$i$i = $1547 ? $1550 : $req_comp; //@line 3324 "./libraries/stb/stb_image.h"
        $1551 = ($1550|0)==(3); //@line 3326 "./libraries/stb/stb_image.h"
        $1552 = ($$req_comp$i$i|0)<(3); //@line 3326 "./libraries/stb/stb_image.h"
        $or$cond3$i$i = $1551 & $1552; //@line 3326 "./libraries/stb/stb_image.h"
        $$$i$i = $or$cond3$i$i ? 1 : $1550; //@line 3327 "./libraries/stb/stb_image.h"
        $1553 = ($$$i$i|0)>(0); //@line 3340 "./libraries/stb/stb_image.h"
        L676: do {
         if ($1553) {
          $1554 = ((($j$i)) + 17796|0); //@line 3348 "./libraries/stb/stb_image.h"
          $1555 = ((($j$i)) + 17800|0); //@line 3349 "./libraries/stb/stb_image.h"
          $$pre125$i$i = HEAP32[$1549>>2]|0; //@line 3345 "./libraries/stb/stb_image.h"
          $1558 = $$pre125$i$i;$k$075$i$i = 0;
          while(1) {
           $1556 = (($res_comp$i$i) + ($k$075$i$i<<5)|0); //@line 3341 "./libraries/stb/stb_image.h"
           $1557 = (($1558) + 3)|0; //@line 3345 "./libraries/stb/stb_image.h"
           $1559 = (_malloc($1557)|0); //@line 900 "./libraries/stb/stb_image.h"
           $1560 = (((((($j$i)) + 17820|0) + (($k$075$i$i*72)|0)|0)) + 56|0); //@line 3345 "./libraries/stb/stb_image.h"
           HEAP32[$1560>>2] = $1559; //@line 3345 "./libraries/stb/stb_image.h"
           $1561 = ($1559|0)==(0|0); //@line 3346 "./libraries/stb/stb_image.h"
           if ($1561) {
            break;
           }
           $1580 = HEAP32[$1554>>2]|0; //@line 3348 "./libraries/stb/stb_image.h"
           $1581 = (((((($j$i)) + 17820|0) + (($k$075$i$i*72)|0)|0)) + 4|0); //@line 3348 "./libraries/stb/stb_image.h"
           $1582 = HEAP32[$1581>>2]|0; //@line 3348 "./libraries/stb/stb_image.h"
           $1583 = (($1580|0) / ($1582|0))&-1; //@line 3358 "./libraries/stb/stb_image.h"
           $1584 = (((($res_comp$i$i) + ($k$075$i$i<<5)|0)) + 12|0); //@line 3348 "./libraries/stb/stb_image.h"
           HEAP32[$1584>>2] = $1583; //@line 3348 "./libraries/stb/stb_image.h"
           $1585 = HEAP32[$1555>>2]|0; //@line 3349 "./libraries/stb/stb_image.h"
           $1586 = (((((($j$i)) + 17820|0) + (($k$075$i$i*72)|0)|0)) + 8|0); //@line 3349 "./libraries/stb/stb_image.h"
           $1587 = HEAP32[$1586>>2]|0; //@line 3349 "./libraries/stb/stb_image.h"
           $1588 = (($1585|0) / ($1587|0))&-1; //@line 3358 "./libraries/stb/stb_image.h"
           $1589 = (((($res_comp$i$i) + ($k$075$i$i<<5)|0)) + 16|0); //@line 3349 "./libraries/stb/stb_image.h"
           HEAP32[$1589>>2] = $1588; //@line 3349 "./libraries/stb/stb_image.h"
           $1590 = $1588 >> 1; //@line 3350 "./libraries/stb/stb_image.h"
           $1591 = (((($res_comp$i$i) + ($k$075$i$i<<5)|0)) + 24|0); //@line 3350 "./libraries/stb/stb_image.h"
           HEAP32[$1591>>2] = $1590; //@line 3350 "./libraries/stb/stb_image.h"
           $1592 = HEAP32[$1549>>2]|0; //@line 3351 "./libraries/stb/stb_image.h"
           $1593 = (($1583) + -1)|0; //@line 3351 "./libraries/stb/stb_image.h"
           $1594 = (($1593) + ($1592))|0; //@line 3351 "./libraries/stb/stb_image.h"
           $1595 = (($1594>>>0) / ($1583>>>0))&-1; //@line 3351 "./libraries/stb/stb_image.h"
           $1596 = (((($res_comp$i$i) + ($k$075$i$i<<5)|0)) + 20|0); //@line 3351 "./libraries/stb/stb_image.h"
           HEAP32[$1596>>2] = $1595; //@line 3351 "./libraries/stb/stb_image.h"
           $1597 = (((($res_comp$i$i) + ($k$075$i$i<<5)|0)) + 28|0); //@line 3352 "./libraries/stb/stb_image.h"
           HEAP32[$1597>>2] = 0; //@line 3352 "./libraries/stb/stb_image.h"
           $1598 = (((((($j$i)) + 17820|0) + (($k$075$i$i*72)|0)|0)) + 44|0); //@line 3353 "./libraries/stb/stb_image.h"
           $1599 = HEAP32[$1598>>2]|0; //@line 3353 "./libraries/stb/stb_image.h"
           $1600 = (((($res_comp$i$i) + ($k$075$i$i<<5)|0)) + 8|0); //@line 3353 "./libraries/stb/stb_image.h"
           HEAP32[$1600>>2] = $1599; //@line 3353 "./libraries/stb/stb_image.h"
           $1601 = (((($res_comp$i$i) + ($k$075$i$i<<5)|0)) + 4|0); //@line 3353 "./libraries/stb/stb_image.h"
           HEAP32[$1601>>2] = $1599; //@line 3353 "./libraries/stb/stb_image.h"
           L681: do {
            switch ($1583|0) {
            case 1:  {
             switch ($1588|0) {
             case 1:  {
              HEAP32[$1556>>2] = 4178; //@line 3355 "./libraries/stb/stb_image.h"
              break L681;
              break;
             }
             case 2:  {
              HEAP32[$1556>>2] = 4179; //@line 3356 "./libraries/stb/stb_image.h"
              break L681;
              break;
             }
             default: {
              label = 530;
              break L681;
             }
             }
             break;
            }
            case 2:  {
             switch ($1588|0) {
             case 1:  {
              HEAP32[$1556>>2] = 4180; //@line 3357 "./libraries/stb/stb_image.h"
              break L681;
              break;
             }
             case 2:  {
              $1602 = HEAP32[$49>>2]|0; //@line 3358 "./libraries/stb/stb_image.h"
              HEAP32[$1556>>2] = $1602; //@line 3358 "./libraries/stb/stb_image.h"
              break L681;
              break;
             }
             default: {
              label = 530;
              break L681;
             }
             }
             break;
            }
            default: {
             label = 530;
            }
            }
           } while(0);
           if ((label|0) == 530) {
            label = 0;
            HEAP32[$1556>>2] = 4181; //@line 3359 "./libraries/stb/stb_image.h"
           }
           $1603 = (($k$075$i$i) + 1)|0; //@line 3340 "./libraries/stb/stb_image.h"
           $1604 = ($1603|0)<($$$i$i|0); //@line 3340 "./libraries/stb/stb_image.h"
           if ($1604) {
            $1558 = $1592;$k$075$i$i = $1603;
           } else {
            $1606 = $1592;
            label = 532;
            break L676;
           }
          }
          $1562 = HEAP32[$1548>>2]|0; //@line 3284 "./libraries/stb/stb_image.h"
          $1563 = ($1562|0)>(0); //@line 3284 "./libraries/stb/stb_image.h"
          if ($1563) {
           $i$01$i9$i$i = 0;
           while(1) {
            $1564 = (((((($j$i)) + 17820|0) + (($i$01$i9$i$i*72)|0)|0)) + 48|0); //@line 3285 "./libraries/stb/stb_image.h"
            $1565 = HEAP32[$1564>>2]|0; //@line 3285 "./libraries/stb/stb_image.h"
            $1566 = ($1565|0)==(0|0); //@line 3285 "./libraries/stb/stb_image.h"
            if (!($1566)) {
             _free($1565); //@line 3286 "./libraries/stb/stb_image.h"
             HEAP32[$1564>>2] = 0; //@line 3287 "./libraries/stb/stb_image.h"
             $1567 = (((((($j$i)) + 17820|0) + (($i$01$i9$i$i*72)|0)|0)) + 44|0); //@line 3288 "./libraries/stb/stb_image.h"
             HEAP32[$1567>>2] = 0; //@line 3288 "./libraries/stb/stb_image.h"
            }
            $1568 = (((((($j$i)) + 17820|0) + (($i$01$i9$i$i*72)|0)|0)) + 52|0); //@line 3290 "./libraries/stb/stb_image.h"
            $1569 = HEAP32[$1568>>2]|0; //@line 3290 "./libraries/stb/stb_image.h"
            $1570 = ($1569|0)==(0|0); //@line 3290 "./libraries/stb/stb_image.h"
            if (!($1570)) {
             _free($1569); //@line 3291 "./libraries/stb/stb_image.h"
             HEAP32[$1568>>2] = 0; //@line 3292 "./libraries/stb/stb_image.h"
             $1571 = (((((($j$i)) + 17820|0) + (($i$01$i9$i$i*72)|0)|0)) + 60|0); //@line 3293 "./libraries/stb/stb_image.h"
             HEAP32[$1571>>2] = 0; //@line 3293 "./libraries/stb/stb_image.h"
            }
            $1572 = (((((($j$i)) + 17820|0) + (($i$01$i9$i$i*72)|0)|0)) + 56|0); //@line 3295 "./libraries/stb/stb_image.h"
            $1573 = HEAP32[$1572>>2]|0; //@line 3295 "./libraries/stb/stb_image.h"
            $1574 = ($1573|0)==(0|0); //@line 3295 "./libraries/stb/stb_image.h"
            if (!($1574)) {
             _free($1573); //@line 3296 "./libraries/stb/stb_image.h"
             HEAP32[$1572>>2] = 0; //@line 3297 "./libraries/stb/stb_image.h"
            }
            $1575 = (($i$01$i9$i$i) + 1)|0; //@line 3284 "./libraries/stb/stb_image.h"
            $1576 = HEAP32[$j$i>>2]|0; //@line 3284 "./libraries/stb/stb_image.h"
            $1577 = ((($1576)) + 8|0); //@line 3284 "./libraries/stb/stb_image.h"
            $1578 = HEAP32[$1577>>2]|0; //@line 3284 "./libraries/stb/stb_image.h"
            $1579 = ($1575|0)<($1578|0); //@line 3284 "./libraries/stb/stb_image.h"
            if ($1579) {
             $i$01$i9$i$i = $1575;
            } else {
             break;
            }
           }
          }
          HEAP32[158754] = 441002; //@line 894 "./libraries/stb/stb_image.h"
          $$2$i$i = 0;
         } else {
          $$pre94$i = HEAP32[$1549>>2]|0; //@line 3363 "./libraries/stb/stb_image.h"
          $1606 = $$pre94$i;
          label = 532;
         }
        } while(0);
        do {
         if ((label|0) == 532) {
          $1605 = Math_imul($1606, $$req_comp$i$i)|0; //@line 3363 "./libraries/stb/stb_image.h"
          $1607 = ((($1549)) + 4|0); //@line 3363 "./libraries/stb/stb_image.h"
          $1608 = HEAP32[$1607>>2]|0; //@line 3363 "./libraries/stb/stb_image.h"
          $1609 = Math_imul($1605, $1608)|0; //@line 3363 "./libraries/stb/stb_image.h"
          $1610 = (($1609) + 1)|0; //@line 3363 "./libraries/stb/stb_image.h"
          $1611 = (_malloc($1610)|0); //@line 900 "./libraries/stb/stb_image.h"
          $1612 = ($1611|0)==(0|0); //@line 3364 "./libraries/stb/stb_image.h"
          if ($1612) {
           $1619 = HEAP32[$1548>>2]|0; //@line 3284 "./libraries/stb/stb_image.h"
           $1620 = ($1619|0)>(0); //@line 3284 "./libraries/stb/stb_image.h"
           if ($1620) {
            $i$01$i12$i$i = 0;
            while(1) {
             $1621 = (((((($j$i)) + 17820|0) + (($i$01$i12$i$i*72)|0)|0)) + 48|0); //@line 3285 "./libraries/stb/stb_image.h"
             $1622 = HEAP32[$1621>>2]|0; //@line 3285 "./libraries/stb/stb_image.h"
             $1623 = ($1622|0)==(0|0); //@line 3285 "./libraries/stb/stb_image.h"
             if (!($1623)) {
              _free($1622); //@line 3286 "./libraries/stb/stb_image.h"
              HEAP32[$1621>>2] = 0; //@line 3287 "./libraries/stb/stb_image.h"
              $1624 = (((((($j$i)) + 17820|0) + (($i$01$i12$i$i*72)|0)|0)) + 44|0); //@line 3288 "./libraries/stb/stb_image.h"
              HEAP32[$1624>>2] = 0; //@line 3288 "./libraries/stb/stb_image.h"
             }
             $1625 = (((((($j$i)) + 17820|0) + (($i$01$i12$i$i*72)|0)|0)) + 52|0); //@line 3290 "./libraries/stb/stb_image.h"
             $1626 = HEAP32[$1625>>2]|0; //@line 3290 "./libraries/stb/stb_image.h"
             $1627 = ($1626|0)==(0|0); //@line 3290 "./libraries/stb/stb_image.h"
             if (!($1627)) {
              _free($1626); //@line 3291 "./libraries/stb/stb_image.h"
              HEAP32[$1625>>2] = 0; //@line 3292 "./libraries/stb/stb_image.h"
              $1628 = (((((($j$i)) + 17820|0) + (($i$01$i12$i$i*72)|0)|0)) + 60|0); //@line 3293 "./libraries/stb/stb_image.h"
              HEAP32[$1628>>2] = 0; //@line 3293 "./libraries/stb/stb_image.h"
             }
             $1629 = (((((($j$i)) + 17820|0) + (($i$01$i12$i$i*72)|0)|0)) + 56|0); //@line 3295 "./libraries/stb/stb_image.h"
             $1630 = HEAP32[$1629>>2]|0; //@line 3295 "./libraries/stb/stb_image.h"
             $1631 = ($1630|0)==(0|0); //@line 3295 "./libraries/stb/stb_image.h"
             if (!($1631)) {
              _free($1630); //@line 3296 "./libraries/stb/stb_image.h"
              HEAP32[$1629>>2] = 0; //@line 3297 "./libraries/stb/stb_image.h"
             }
             $1632 = (($i$01$i12$i$i) + 1)|0; //@line 3284 "./libraries/stb/stb_image.h"
             $1633 = HEAP32[$j$i>>2]|0; //@line 3284 "./libraries/stb/stb_image.h"
             $1634 = ((($1633)) + 8|0); //@line 3284 "./libraries/stb/stb_image.h"
             $1635 = HEAP32[$1634>>2]|0; //@line 3284 "./libraries/stb/stb_image.h"
             $1636 = ($1632|0)<($1635|0); //@line 3284 "./libraries/stb/stb_image.h"
             if ($1636) {
              $i$01$i12$i$i = $1632;
             } else {
              break;
             }
            }
           }
           HEAP32[158754] = 441002; //@line 894 "./libraries/stb/stb_image.h"
           $$2$i$i = 0;
           break;
          }
          $1613 = HEAP32[$1607>>2]|0; //@line 3367 "./libraries/stb/stb_image.h"
          $1614 = ($1613|0)==(0); //@line 3367 "./libraries/stb/stb_image.h"
          if ($1614) {
           $$lcssa$i$i = $1549;
          } else {
           $1615 = ($$req_comp$i$i|0)>(2); //@line 3383 "./libraries/stb/stb_image.h"
           $1616 = ((($coutput$i$i)) + 4|0); //@line 3386 "./libraries/stb/stb_image.h"
           $1617 = ((($coutput$i$i)) + 8|0); //@line 3386 "./libraries/stb/stb_image.h"
           $1618 = ($$req_comp$i$i|0)==(1); //@line 3395 "./libraries/stb/stb_image.h"
           $1638 = $1549;$j$072$i$i = 0;
           while(1) {
            $1637 = HEAP32[$1638>>2]|0; //@line 3368 "./libraries/stb/stb_image.h"
            $1639 = Math_imul($j$072$i$i, $$req_comp$i$i)|0; //@line 3368 "./libraries/stb/stb_image.h"
            $1640 = Math_imul($1639, $1637)|0; //@line 3368 "./libraries/stb/stb_image.h"
            $1641 = (($1611) + ($1640)|0); //@line 3368 "./libraries/stb/stb_image.h"
            if ($1553) {
             $k$163$i$i = 0;
             while(1) {
              $1642 = (((($res_comp$i$i) + ($k$163$i$i<<5)|0)) + 24|0); //@line 3371 "./libraries/stb/stb_image.h"
              $1643 = HEAP32[$1642>>2]|0; //@line 3371 "./libraries/stb/stb_image.h"
              $1644 = (((($res_comp$i$i) + ($k$163$i$i<<5)|0)) + 16|0); //@line 3371 "./libraries/stb/stb_image.h"
              $1645 = HEAP32[$1644>>2]|0; //@line 3371 "./libraries/stb/stb_image.h"
              $1646 = $1645 >> 1; //@line 3371 "./libraries/stb/stb_image.h"
              $1647 = ($1643|0)>=($1646|0); //@line 3371 "./libraries/stb/stb_image.h"
              $1648 = (($res_comp$i$i) + ($k$163$i$i<<5)|0); //@line 3372 "./libraries/stb/stb_image.h"
              $1649 = HEAP32[$1648>>2]|0; //@line 3372 "./libraries/stb/stb_image.h"
              $1650 = (((((($j$i)) + 17820|0) + (($k$163$i$i*72)|0)|0)) + 56|0); //@line 3372 "./libraries/stb/stb_image.h"
              $1651 = HEAP32[$1650>>2]|0; //@line 3372 "./libraries/stb/stb_image.h"
              $1652 = (((($res_comp$i$i) + ($k$163$i$i<<5)|0)) + 8|0); //@line 3373 "./libraries/stb/stb_image.h"
              $1653 = (((($res_comp$i$i) + ($k$163$i$i<<5)|0)) + 4|0); //@line 3373 "./libraries/stb/stb_image.h"
              $$in$i$i = $1647 ? $1652 : $1653; //@line 3373 "./libraries/stb/stb_image.h"
              $1654 = HEAP32[$$in$i$i>>2]|0; //@line 3373 "./libraries/stb/stb_image.h"
              $$in5$i$i = $1647 ? $1653 : $1652; //@line 3374 "./libraries/stb/stb_image.h"
              $1655 = HEAP32[$$in5$i$i>>2]|0; //@line 3374 "./libraries/stb/stb_image.h"
              $1656 = (((($res_comp$i$i) + ($k$163$i$i<<5)|0)) + 20|0); //@line 3375 "./libraries/stb/stb_image.h"
              $1657 = HEAP32[$1656>>2]|0; //@line 3375 "./libraries/stb/stb_image.h"
              $1658 = (((($res_comp$i$i) + ($k$163$i$i<<5)|0)) + 12|0); //@line 3375 "./libraries/stb/stb_image.h"
              $1659 = HEAP32[$1658>>2]|0; //@line 3375 "./libraries/stb/stb_image.h"
              $1660 = (FUNCTION_TABLE_iiiiii[$1649 & 8191]($1651,$1654,$1655,$1657,$1659)|0); //@line 3372 "./libraries/stb/stb_image.h"
              $1661 = (($coutput$i$i) + ($k$163$i$i<<2)|0); //@line 3372 "./libraries/stb/stb_image.h"
              HEAP32[$1661>>2] = $1660; //@line 3372 "./libraries/stb/stb_image.h"
              $1662 = (($1643) + 1)|0; //@line 3376 "./libraries/stb/stb_image.h"
              HEAP32[$1642>>2] = $1662; //@line 3376 "./libraries/stb/stb_image.h"
              $1663 = ($1662|0)<($1645|0); //@line 3376 "./libraries/stb/stb_image.h"
              do {
               if (!($1663)) {
                HEAP32[$1642>>2] = 0; //@line 3377 "./libraries/stb/stb_image.h"
                $1664 = HEAP32[$1652>>2]|0; //@line 3378 "./libraries/stb/stb_image.h"
                HEAP32[$1653>>2] = $1664; //@line 3378 "./libraries/stb/stb_image.h"
                $1665 = (((($res_comp$i$i) + ($k$163$i$i<<5)|0)) + 28|0); //@line 3379 "./libraries/stb/stb_image.h"
                $1666 = HEAP32[$1665>>2]|0; //@line 3379 "./libraries/stb/stb_image.h"
                $1667 = (($1666) + 1)|0; //@line 3379 "./libraries/stb/stb_image.h"
                HEAP32[$1665>>2] = $1667; //@line 3379 "./libraries/stb/stb_image.h"
                $1668 = (((((($j$i)) + 17820|0) + (($k$163$i$i*72)|0)|0)) + 32|0); //@line 3379 "./libraries/stb/stb_image.h"
                $1669 = HEAP32[$1668>>2]|0; //@line 3379 "./libraries/stb/stb_image.h"
                $1670 = ($1667|0)<($1669|0); //@line 3379 "./libraries/stb/stb_image.h"
                if (!($1670)) {
                 break;
                }
                $1671 = $1664; //@line 3380 "./libraries/stb/stb_image.h"
                $1672 = (((((($j$i)) + 17820|0) + (($k$163$i$i*72)|0)|0)) + 36|0); //@line 3380 "./libraries/stb/stb_image.h"
                $1673 = HEAP32[$1672>>2]|0; //@line 3380 "./libraries/stb/stb_image.h"
                $1674 = (($1671) + ($1673)|0); //@line 3380 "./libraries/stb/stb_image.h"
                HEAP32[$1652>>2] = $1674; //@line 3380 "./libraries/stb/stb_image.h"
               }
              } while(0);
              $1675 = (($k$163$i$i) + 1)|0; //@line 3369 "./libraries/stb/stb_image.h"
              $exitcond$i$i = ($1675|0)==($$$i$i|0); //@line 3369 "./libraries/stb/stb_image.h"
              if ($exitcond$i$i) {
               break;
              } else {
               $k$163$i$i = $1675;
              }
             }
             $$pre127$i$i = HEAP32[$j$i>>2]|0; //@line 3385 "./libraries/stb/stb_image.h"
             $1678 = $$pre127$i$i;
            } else {
             $1678 = $1638;
            }
            $1676 = HEAP32[$coutput$i$i>>2]|0; //@line 3384 "./libraries/stb/stb_image.h"
            do {
             if ($1615) {
              $1677 = ((($1678)) + 8|0); //@line 3385 "./libraries/stb/stb_image.h"
              $1679 = HEAP32[$1677>>2]|0; //@line 3385 "./libraries/stb/stb_image.h"
              $1680 = ($1679|0)==(3); //@line 3385 "./libraries/stb/stb_image.h"
              if ($1680) {
               $1683 = HEAP32[$48>>2]|0; //@line 3386 "./libraries/stb/stb_image.h"
               $1684 = HEAP32[$1616>>2]|0; //@line 3386 "./libraries/stb/stb_image.h"
               $1685 = HEAP32[$1617>>2]|0; //@line 3386 "./libraries/stb/stb_image.h"
               $1686 = HEAP32[$1678>>2]|0; //@line 3386 "./libraries/stb/stb_image.h"
               FUNCTION_TABLE_viiiiii[$1683 & 8191]($1641,$1676,$1684,$1685,$1686,$$req_comp$i$i); //@line 3386 "./libraries/stb/stb_image.h"
               $$pre128$i$i = HEAP32[$j$i>>2]|0; //@line 3367 "./libraries/stb/stb_image.h"
               $1715 = $$pre128$i$i;
               break;
              }
              $1681 = HEAP32[$1678>>2]|0; //@line 3388 "./libraries/stb/stb_image.h"
              $1682 = ($1681|0)==(0); //@line 3388 "./libraries/stb/stb_image.h"
              if ($1682) {
               $1715 = $1678;
               break;
              } else {
               $i$069$i$i = 0;$out$070$i$i = $1641;
              }
              while(1) {
               $1687 = (($1676) + ($i$069$i$i)|0); //@line 3389 "./libraries/stb/stb_image.h"
               $1688 = HEAP8[$1687>>0]|0; //@line 3389 "./libraries/stb/stb_image.h"
               $1689 = ((($out$070$i$i)) + 2|0); //@line 3389 "./libraries/stb/stb_image.h"
               HEAP8[$1689>>0] = $1688; //@line 3389 "./libraries/stb/stb_image.h"
               $1690 = ((($out$070$i$i)) + 1|0); //@line 3389 "./libraries/stb/stb_image.h"
               HEAP8[$1690>>0] = $1688; //@line 3389 "./libraries/stb/stb_image.h"
               HEAP8[$out$070$i$i>>0] = $1688; //@line 3389 "./libraries/stb/stb_image.h"
               $1691 = ((($out$070$i$i)) + 3|0); //@line 3390 "./libraries/stb/stb_image.h"
               HEAP8[$1691>>0] = -1; //@line 3390 "./libraries/stb/stb_image.h"
               $1692 = (($out$070$i$i) + ($$req_comp$i$i)|0); //@line 3391 "./libraries/stb/stb_image.h"
               $1693 = (($i$069$i$i) + 1)|0; //@line 3388 "./libraries/stb/stb_image.h"
               $1694 = HEAP32[$j$i>>2]|0; //@line 3388 "./libraries/stb/stb_image.h"
               $1695 = HEAP32[$1694>>2]|0; //@line 3388 "./libraries/stb/stb_image.h"
               $1696 = ($1693>>>0)<($1695>>>0); //@line 3388 "./libraries/stb/stb_image.h"
               if ($1696) {
                $i$069$i$i = $1693;$out$070$i$i = $1692;
               } else {
                $1715 = $1694;
                break;
               }
              }
             } else {
              $1697 = HEAP32[$1678>>2]|0; //@line 3396 "./libraries/stb/stb_image.h"
              $1698 = ($1697|0)!=(0); //@line 3396 "./libraries/stb/stb_image.h"
              if ($1618) {
               if ($1698) {
                $i$167$i$i = 0;
               } else {
                $1715 = $1678;
                break;
               }
               while(1) {
                $1699 = (($1676) + ($i$167$i$i)|0); //@line 3396 "./libraries/stb/stb_image.h"
                $1700 = HEAP8[$1699>>0]|0; //@line 3396 "./libraries/stb/stb_image.h"
                $1701 = (($1641) + ($i$167$i$i)|0); //@line 3396 "./libraries/stb/stb_image.h"
                HEAP8[$1701>>0] = $1700; //@line 3396 "./libraries/stb/stb_image.h"
                $1702 = (($i$167$i$i) + 1)|0; //@line 3396 "./libraries/stb/stb_image.h"
                $1703 = HEAP32[$1678>>2]|0; //@line 3396 "./libraries/stb/stb_image.h"
                $1704 = ($1702>>>0)<($1703>>>0); //@line 3396 "./libraries/stb/stb_image.h"
                if ($1704) {
                 $i$167$i$i = $1702;
                } else {
                 $1715 = $1678;
                 break;
                }
               }
              } else {
               if ($1698) {
                $i$264$i$i = 0;$out$165$i$i = $1641;
               } else {
                $1715 = $1678;
                break;
               }
               while(1) {
                $1705 = (($1676) + ($i$264$i$i)|0); //@line 3398 "./libraries/stb/stb_image.h"
                $1706 = HEAP8[$1705>>0]|0; //@line 3398 "./libraries/stb/stb_image.h"
                $1707 = ((($out$165$i$i)) + 1|0); //@line 3398 "./libraries/stb/stb_image.h"
                HEAP8[$out$165$i$i>>0] = $1706; //@line 3398 "./libraries/stb/stb_image.h"
                $1708 = ((($out$165$i$i)) + 2|0); //@line 3398 "./libraries/stb/stb_image.h"
                HEAP8[$1707>>0] = -1; //@line 3398 "./libraries/stb/stb_image.h"
                $1709 = (($i$264$i$i) + 1)|0; //@line 3398 "./libraries/stb/stb_image.h"
                $1710 = HEAP32[$j$i>>2]|0; //@line 3398 "./libraries/stb/stb_image.h"
                $1711 = HEAP32[$1710>>2]|0; //@line 3398 "./libraries/stb/stb_image.h"
                $1712 = ($1709>>>0)<($1711>>>0); //@line 3398 "./libraries/stb/stb_image.h"
                if ($1712) {
                 $i$264$i$i = $1709;$out$165$i$i = $1708;
                } else {
                 $1715 = $1710;
                 break;
                }
               }
              }
             }
            } while(0);
            $1713 = (($j$072$i$i) + 1)|0; //@line 3367 "./libraries/stb/stb_image.h"
            $1714 = ((($1715)) + 4|0); //@line 3367 "./libraries/stb/stb_image.h"
            $1716 = HEAP32[$1714>>2]|0; //@line 3367 "./libraries/stb/stb_image.h"
            $1717 = ($1713>>>0)<($1716>>>0); //@line 3367 "./libraries/stb/stb_image.h"
            if ($1717) {
             $1638 = $1715;$j$072$i$i = $1713;
            } else {
             $$lcssa$i$i = $1715;
             break;
            }
           }
          }
          $1718 = ((($$lcssa$i$i)) + 8|0); //@line 3284 "./libraries/stb/stb_image.h"
          $1719 = HEAP32[$1718>>2]|0; //@line 3284 "./libraries/stb/stb_image.h"
          $1720 = ($1719|0)>(0); //@line 3284 "./libraries/stb/stb_image.h"
          if ($1720) {
           $i$01$i$i$i = 0;
           while(1) {
            $1721 = (((((($j$i)) + 17820|0) + (($i$01$i$i$i*72)|0)|0)) + 48|0); //@line 3285 "./libraries/stb/stb_image.h"
            $1722 = HEAP32[$1721>>2]|0; //@line 3285 "./libraries/stb/stb_image.h"
            $1723 = ($1722|0)==(0|0); //@line 3285 "./libraries/stb/stb_image.h"
            if (!($1723)) {
             _free($1722); //@line 3286 "./libraries/stb/stb_image.h"
             HEAP32[$1721>>2] = 0; //@line 3287 "./libraries/stb/stb_image.h"
             $1724 = (((((($j$i)) + 17820|0) + (($i$01$i$i$i*72)|0)|0)) + 44|0); //@line 3288 "./libraries/stb/stb_image.h"
             HEAP32[$1724>>2] = 0; //@line 3288 "./libraries/stb/stb_image.h"
            }
            $1725 = (((((($j$i)) + 17820|0) + (($i$01$i$i$i*72)|0)|0)) + 52|0); //@line 3290 "./libraries/stb/stb_image.h"
            $1726 = HEAP32[$1725>>2]|0; //@line 3290 "./libraries/stb/stb_image.h"
            $1727 = ($1726|0)==(0|0); //@line 3290 "./libraries/stb/stb_image.h"
            if (!($1727)) {
             _free($1726); //@line 3291 "./libraries/stb/stb_image.h"
             HEAP32[$1725>>2] = 0; //@line 3292 "./libraries/stb/stb_image.h"
             $1728 = (((((($j$i)) + 17820|0) + (($i$01$i$i$i*72)|0)|0)) + 60|0); //@line 3293 "./libraries/stb/stb_image.h"
             HEAP32[$1728>>2] = 0; //@line 3293 "./libraries/stb/stb_image.h"
            }
            $1729 = (((((($j$i)) + 17820|0) + (($i$01$i$i$i*72)|0)|0)) + 56|0); //@line 3295 "./libraries/stb/stb_image.h"
            $1730 = HEAP32[$1729>>2]|0; //@line 3295 "./libraries/stb/stb_image.h"
            $1731 = ($1730|0)==(0|0); //@line 3295 "./libraries/stb/stb_image.h"
            if (!($1731)) {
             _free($1730); //@line 3296 "./libraries/stb/stb_image.h"
             HEAP32[$1729>>2] = 0; //@line 3297 "./libraries/stb/stb_image.h"
            }
            $1732 = (($i$01$i$i$i) + 1)|0; //@line 3284 "./libraries/stb/stb_image.h"
            $1733 = HEAP32[$j$i>>2]|0; //@line 3284 "./libraries/stb/stb_image.h"
            $1734 = ((($1733)) + 8|0); //@line 3284 "./libraries/stb/stb_image.h"
            $1735 = HEAP32[$1734>>2]|0; //@line 3284 "./libraries/stb/stb_image.h"
            $1736 = ($1732|0)<($1735|0); //@line 3284 "./libraries/stb/stb_image.h"
            if ($1736) {
             $i$01$i$i$i = $1732;
            } else {
             $1738 = $1733;
             break;
            }
           }
          } else {
           $1738 = $$lcssa$i$i;
          }
          $1737 = HEAP32[$1738>>2]|0; //@line 3402 "./libraries/stb/stb_image.h"
          HEAP32[$x>>2] = $1737; //@line 3402 "./libraries/stb/stb_image.h"
          $1739 = ((($1738)) + 4|0); //@line 3403 "./libraries/stb/stb_image.h"
          $1740 = HEAP32[$1739>>2]|0; //@line 3403 "./libraries/stb/stb_image.h"
          HEAP32[$y>>2] = $1740; //@line 3403 "./libraries/stb/stb_image.h"
          $1741 = ($comp|0)==(0|0); //@line 3404 "./libraries/stb/stb_image.h"
          if ($1741) {
           $$2$i$i = $1611;
          } else {
           $1742 = ((($1738)) + 8|0); //@line 3404 "./libraries/stb/stb_image.h"
           $1743 = HEAP32[$1742>>2]|0; //@line 3404 "./libraries/stb/stb_image.h"
           HEAP32[$comp>>2] = $1743; //@line 3404 "./libraries/stb/stb_image.h"
           $$2$i$i = $1611;
          }
         }
        } while(0);
        $$3$i$i = $$2$i$i;
        break L28;
        break;
       }
      }
      HEAP32[158754] = 441043; //@line 894 "./libraries/stb/stb_image.h"
     }
    } while(0);
    $1527 = HEAP32[$j$i>>2]|0; //@line 3284 "./libraries/stb/stb_image.h"
    $1528 = ((($1527)) + 8|0); //@line 3284 "./libraries/stb/stb_image.h"
    $1529 = HEAP32[$1528>>2]|0; //@line 3284 "./libraries/stb/stb_image.h"
    $1530 = ($1529|0)>(0); //@line 3284 "./libraries/stb/stb_image.h"
    if ($1530) {
     $i$01$i6$i$i = 0;
     while(1) {
      $1531 = (((((($j$i)) + 17820|0) + (($i$01$i6$i$i*72)|0)|0)) + 48|0); //@line 3285 "./libraries/stb/stb_image.h"
      $1532 = HEAP32[$1531>>2]|0; //@line 3285 "./libraries/stb/stb_image.h"
      $1533 = ($1532|0)==(0|0); //@line 3285 "./libraries/stb/stb_image.h"
      if (!($1533)) {
       _free($1532); //@line 3286 "./libraries/stb/stb_image.h"
       HEAP32[$1531>>2] = 0; //@line 3287 "./libraries/stb/stb_image.h"
       $1534 = (((((($j$i)) + 17820|0) + (($i$01$i6$i$i*72)|0)|0)) + 44|0); //@line 3288 "./libraries/stb/stb_image.h"
       HEAP32[$1534>>2] = 0; //@line 3288 "./libraries/stb/stb_image.h"
      }
      $1535 = (((((($j$i)) + 17820|0) + (($i$01$i6$i$i*72)|0)|0)) + 52|0); //@line 3290 "./libraries/stb/stb_image.h"
      $1536 = HEAP32[$1535>>2]|0; //@line 3290 "./libraries/stb/stb_image.h"
      $1537 = ($1536|0)==(0|0); //@line 3290 "./libraries/stb/stb_image.h"
      if (!($1537)) {
       _free($1536); //@line 3291 "./libraries/stb/stb_image.h"
       HEAP32[$1535>>2] = 0; //@line 3292 "./libraries/stb/stb_image.h"
       $1538 = (((((($j$i)) + 17820|0) + (($i$01$i6$i$i*72)|0)|0)) + 60|0); //@line 3293 "./libraries/stb/stb_image.h"
       HEAP32[$1538>>2] = 0; //@line 3293 "./libraries/stb/stb_image.h"
      }
      $1539 = (((((($j$i)) + 17820|0) + (($i$01$i6$i$i*72)|0)|0)) + 56|0); //@line 3295 "./libraries/stb/stb_image.h"
      $1540 = HEAP32[$1539>>2]|0; //@line 3295 "./libraries/stb/stb_image.h"
      $1541 = ($1540|0)==(0|0); //@line 3295 "./libraries/stb/stb_image.h"
      if (!($1541)) {
       _free($1540); //@line 3296 "./libraries/stb/stb_image.h"
       HEAP32[$1539>>2] = 0; //@line 3297 "./libraries/stb/stb_image.h"
      }
      $1542 = (($i$01$i6$i$i) + 1)|0; //@line 3284 "./libraries/stb/stb_image.h"
      $1543 = HEAP32[$j$i>>2]|0; //@line 3284 "./libraries/stb/stb_image.h"
      $1544 = ((($1543)) + 8|0); //@line 3284 "./libraries/stb/stb_image.h"
      $1545 = HEAP32[$1544>>2]|0; //@line 3284 "./libraries/stb/stb_image.h"
      $1546 = ($1542|0)<($1545|0); //@line 3284 "./libraries/stb/stb_image.h"
      if ($1546) {
       $i$01$i6$i$i = $1542;
      } else {
       $$3$i$i = 0;
       break;
      }
     }
    } else {
     $$3$i$i = 0;
    }
   }
  } while(0);
  $$0 = $$3$i$i;
  STACKTOP = sp;return ($$0|0); //@line 976 "./libraries/stb/stb_image.h"
 }
 $1744 = ($45>>>0)<($46>>>0); //@line 1229 "./libraries/stb/stb_image.h"
 if ($1744) {
  $1745 = ((($45)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
  HEAP32[$0>>2] = $1745; //@line 1230 "./libraries/stb/stb_image.h"
  $1746 = HEAP8[$45>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
  $$0$i$i$i = $1746;$1763 = $1745;$1764 = $46;
  label = 580;
 } else {
  $1747 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
  $1748 = HEAP32[$1747>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
  $1749 = ($1748|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
  if ($1749) {
   label = 617;
  } else {
   $1750 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
   $1751 = HEAP32[$1750>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $1752 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
   $1753 = HEAP32[$1752>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $1754 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
   $1755 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
   $1756 = HEAP32[$1755>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $1757 = (FUNCTION_TABLE_iiii[$1751 & 8191]($1753,$1754,$1756)|0); //@line 1213 "./libraries/stb/stb_image.h"
   $1758 = ($1757|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
   if ($1758) {
    HEAP32[$1747>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
    $1759 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
    HEAP8[$1754>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
    $$pre$phi$i$i$iZ2D = $1759;$$sink$i$i$i$i8 = $1759;$3764 = 0;
   } else {
    $1760 = (((($s)) + 40|0) + ($1757)|0); //@line 1223 "./libraries/stb/stb_image.h"
    $$pre$i$i$i7 = HEAP8[$1754>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
    $$pre1$i$i$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
    $$pre$phi$i$i$iZ2D = $$pre1$i$i$i;$$sink$i$i$i$i8 = $1760;$3764 = $$pre$i$i$i7;
   }
   HEAP32[$2>>2] = $$sink$i$i$i$i8; //@line 1223 "./libraries/stb/stb_image.h"
   HEAP32[$0>>2] = $$pre$phi$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
   $$0$i$i$i = $3764;$1763 = $$pre$phi$i$i$iZ2D;$1764 = $$sink$i$i$i$i8;
   label = 580;
  }
 }
 L806: do {
  if ((label|0) == 580) {
   $1761 = ($$0$i$i$i<<24>>24)==(66); //@line 4540 "./libraries/stb/stb_image.h"
   if ($1761) {
    $1762 = ($1763>>>0)<($1764>>>0); //@line 1229 "./libraries/stb/stb_image.h"
    if ($1762) {
     $1765 = ((($1763)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
     HEAP32[$0>>2] = $1765; //@line 1230 "./libraries/stb/stb_image.h"
     $1766 = HEAP8[$1763>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
     $$0$i11$i$i = $1766;
    } else {
     $1767 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
     $1768 = HEAP32[$1767>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
     $1769 = ($1768|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
     if ($1769) {
      label = 617;
      break;
     }
     $1770 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
     $1771 = HEAP32[$1770>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $1772 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $1773 = HEAP32[$1772>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $1774 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $1775 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $1776 = HEAP32[$1775>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $1777 = (FUNCTION_TABLE_iiii[$1771 & 8191]($1773,$1774,$1776)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $1778 = ($1777|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($1778) {
      HEAP32[$1767>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $1779 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$1774>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i8$i$iZ2D = $1779;$$sink$i$i9$i$i = $1779;$3765 = 0;
     } else {
      $1780 = (((($s)) + 40|0) + ($1777)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre$i6$i$i = HEAP8[$1774>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
      $$pre1$i7$i$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i8$i$iZ2D = $$pre1$i7$i$i;$$sink$i$i9$i$i = $1780;$3765 = $$pre$i6$i$i;
     }
     HEAP32[$2>>2] = $$sink$i$i9$i$i; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$0>>2] = $$pre$phi$i8$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
     $$0$i11$i$i = $3765;
    }
    $1781 = ($$0$i11$i$i<<24>>24)==(77); //@line 4541 "./libraries/stb/stb_image.h"
    if ($1781) {
     (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4542 "./libraries/stb/stb_image.h"
     $1782 = HEAP32[$0>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
     $1783 = HEAP32[$2>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
     $1784 = ($1782>>>0)<($1783>>>0); //@line 1229 "./libraries/stb/stb_image.h"
     if ($1784) {
      $1785 = ((($1782)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
      HEAP32[$0>>2] = $1785; //@line 1230 "./libraries/stb/stb_image.h"
      $1801 = $1785;$1802 = $1783;
     } else {
      $1786 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
      $1787 = HEAP32[$1786>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
      $1788 = ($1787|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
      if ($1788) {
       $1801 = $1782;$1802 = $1783;
      } else {
       $1789 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1790 = HEAP32[$1789>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $1791 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1792 = HEAP32[$1791>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $1793 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1794 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1795 = HEAP32[$1794>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $1796 = (FUNCTION_TABLE_iiii[$1790 & 8191]($1792,$1793,$1795)|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1797 = ($1796|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
       if ($1797) {
        HEAP32[$1786>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
        $1798 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
        HEAP8[$1793>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
        $$pre$phi$i$i$i$i12Z2D = $1798;$$sink$i$i$i$i$i13 = $1798;
       } else {
        $1799 = (((($s)) + 40|0) + ($1796)|0); //@line 1223 "./libraries/stb/stb_image.h"
        $$pre1$i$i$i$i11 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
        $$pre$phi$i$i$i$i12Z2D = $$pre1$i$i$i$i11;$$sink$i$i$i$i$i13 = $1799;
       }
       HEAP32[$2>>2] = $$sink$i$i$i$i$i13; //@line 1223 "./libraries/stb/stb_image.h"
       HEAP32[$0>>2] = $$pre$phi$i$i$i$i12Z2D; //@line 1233 "./libraries/stb/stb_image.h"
       $1801 = $$pre$phi$i$i$i$i12Z2D;$1802 = $$sink$i$i$i$i$i13;
      }
     }
     $1800 = ($1801>>>0)<($1802>>>0); //@line 1229 "./libraries/stb/stb_image.h"
     if ($1800) {
      $1803 = ((($1801)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
      HEAP32[$0>>2] = $1803; //@line 1230 "./libraries/stb/stb_image.h"
      $1819 = $1803;$1820 = $1802;
     } else {
      $1804 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
      $1805 = HEAP32[$1804>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
      $1806 = ($1805|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
      if ($1806) {
       $1819 = $1801;$1820 = $1802;
      } else {
       $1807 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1808 = HEAP32[$1807>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $1809 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1810 = HEAP32[$1809>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $1811 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1812 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1813 = HEAP32[$1812>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $1814 = (FUNCTION_TABLE_iiii[$1808 & 8191]($1810,$1811,$1813)|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1815 = ($1814|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
       if ($1815) {
        HEAP32[$1804>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
        $1816 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
        HEAP8[$1811>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
        $$pre$phi$i3$i$i$iZ2D = $1816;$$sink$i$i4$i$i$i = $1816;
       } else {
        $1817 = (((($s)) + 40|0) + ($1814)|0); //@line 1223 "./libraries/stb/stb_image.h"
        $$pre1$i2$i$i$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
        $$pre$phi$i3$i$i$iZ2D = $$pre1$i2$i$i$i;$$sink$i$i4$i$i$i = $1817;
       }
       HEAP32[$2>>2] = $$sink$i$i4$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
       HEAP32[$0>>2] = $$pre$phi$i3$i$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
       $1819 = $$pre$phi$i3$i$i$iZ2D;$1820 = $$sink$i$i4$i$i$i;
      }
     }
     $1818 = ($1819>>>0)<($1820>>>0); //@line 1229 "./libraries/stb/stb_image.h"
     if ($1818) {
      $1821 = ((($1819)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
      HEAP32[$0>>2] = $1821; //@line 1230 "./libraries/stb/stb_image.h"
      $1837 = $1821;$1838 = $1820;
     } else {
      $1822 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
      $1823 = HEAP32[$1822>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
      $1824 = ($1823|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
      if ($1824) {
       $1837 = $1819;$1838 = $1820;
      } else {
       $1825 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1826 = HEAP32[$1825>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $1827 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1828 = HEAP32[$1827>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $1829 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1830 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1831 = HEAP32[$1830>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $1832 = (FUNCTION_TABLE_iiii[$1826 & 8191]($1828,$1829,$1831)|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1833 = ($1832|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
       if ($1833) {
        HEAP32[$1822>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
        $1834 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
        HEAP8[$1829>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
        $$pre$phi$i$i15$i$iZ2D = $1834;$$sink$i$i$i16$i$i = $1834;
       } else {
        $1835 = (((($s)) + 40|0) + ($1832)|0); //@line 1223 "./libraries/stb/stb_image.h"
        $$pre1$i$i14$i$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
        $$pre$phi$i$i15$i$iZ2D = $$pre1$i$i14$i$i;$$sink$i$i$i16$i$i = $1835;
       }
       HEAP32[$2>>2] = $$sink$i$i$i16$i$i; //@line 1223 "./libraries/stb/stb_image.h"
       HEAP32[$0>>2] = $$pre$phi$i$i15$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
       $1837 = $$pre$phi$i$i15$i$iZ2D;$1838 = $$sink$i$i$i16$i$i;
      }
     }
     $1836 = ($1837>>>0)<($1838>>>0); //@line 1229 "./libraries/stb/stb_image.h"
     if ($1836) {
      $1839 = ((($1837)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
      HEAP32[$0>>2] = $1839; //@line 1230 "./libraries/stb/stb_image.h"
     } else {
      $1840 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
      $1841 = HEAP32[$1840>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
      $1842 = ($1841|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
      if (!($1842)) {
       $1843 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1844 = HEAP32[$1843>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $1845 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1846 = HEAP32[$1845>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $1847 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1848 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1849 = HEAP32[$1848>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $1850 = (FUNCTION_TABLE_iiii[$1844 & 8191]($1846,$1847,$1849)|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1851 = ($1850|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
       if ($1851) {
        HEAP32[$1840>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
        $1852 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
        HEAP8[$1847>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
        $$pre$phi$i3$i22$i$iZ2D = $1852;$$sink$i$i4$i23$i$i = $1852;
       } else {
        $1853 = (((($s)) + 40|0) + ($1850)|0); //@line 1223 "./libraries/stb/stb_image.h"
        $$pre1$i2$i21$i$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
        $$pre$phi$i3$i22$i$iZ2D = $$pre1$i2$i21$i$i;$$sink$i$i4$i23$i$i = $1853;
       }
       HEAP32[$2>>2] = $$sink$i$i4$i23$i$i; //@line 1223 "./libraries/stb/stb_image.h"
       HEAP32[$0>>2] = $$pre$phi$i3$i22$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
      }
     }
     (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4545 "./libraries/stb/stb_image.h"
     $1858 = (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4546 "./libraries/stb/stb_image.h"
     $1859 = ($1858|0)==(12); //@line 4547 "./libraries/stb/stb_image.h"
     $1860 = HEAP32[$41>>2]|0; //@line 826 "./libraries/stb/stb_image.h"
     HEAP32[$0>>2] = $1860; //@line 826 "./libraries/stb/stb_image.h"
     $1861 = HEAP32[$43>>2]|0; //@line 827 "./libraries/stb/stb_image.h"
     HEAP32[$2>>2] = $1861; //@line 827 "./libraries/stb/stb_image.h"
     $1862 = $1860; //@line 947 "./libraries/stb/stb_image.h"
     $1863 = $1861; //@line 947 "./libraries/stb/stb_image.h"
     if (!($1859)) {
      $1864 = $1858 | 16; //@line 4547 "./libraries/stb/stb_image.h"
      switch ($1864|0) {
      case 56: case 124:  {
       break;
      }
      default: {
       $2824 = $1862;$2825 = $1863;
       break L806;
      }
      }
     }
     $1865 = ($1862>>>0)<($1863>>>0); //@line 1229 "./libraries/stb/stb_image.h"
     if ($1865) {
      $1866 = ((($1862)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
      HEAP32[$0>>2] = $1866; //@line 1230 "./libraries/stb/stb_image.h"
      $1867 = HEAP8[$1862>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
      $$0$i$i18 = $1867;$1884 = $1866;$1885 = $1863;
      label = 627;
     } else {
      $1868 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
      $1869 = HEAP32[$1868>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
      $1870 = ($1869|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
      if ($1870) {
       label = 636;
      } else {
       $1871 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1872 = HEAP32[$1871>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $1873 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1874 = HEAP32[$1873>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $1875 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1876 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1877 = HEAP32[$1876>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $1878 = (FUNCTION_TABLE_iiii[$1872 & 8191]($1874,$1875,$1877)|0); //@line 1213 "./libraries/stb/stb_image.h"
       $1879 = ($1878|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
       if ($1879) {
        HEAP32[$1868>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
        $1880 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
        HEAP8[$1875>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
        $$pre$phi$i$iZ2D = $1880;$$sink$i$i$i17 = $1880;$3766 = 0;
       } else {
        $1881 = (((($s)) + 40|0) + ($1878)|0); //@line 1223 "./libraries/stb/stb_image.h"
        $$pre$i$i = HEAP8[$1875>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
        $$pre1$i$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
        $$pre$phi$i$iZ2D = $$pre1$i$i;$$sink$i$i$i17 = $1881;$3766 = $$pre$i$i;
       }
       HEAP32[$2>>2] = $$sink$i$i$i17; //@line 1223 "./libraries/stb/stb_image.h"
       HEAP32[$0>>2] = $$pre$phi$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
       $$0$i$i18 = $3766;$1884 = $$pre$phi$i$iZ2D;$1885 = $$sink$i$i$i17;
       label = 627;
      }
     }
     L866: do {
      if ((label|0) == 627) {
       $1882 = ($$0$i$i18<<24>>24)==(66); //@line 4606 "./libraries/stb/stb_image.h"
       if ($1882) {
        $1883 = ($1884>>>0)<($1885>>>0); //@line 1229 "./libraries/stb/stb_image.h"
        if ($1883) {
         $1886 = ((($1884)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
         HEAP32[$0>>2] = $1886; //@line 1230 "./libraries/stb/stb_image.h"
         $1887 = HEAP8[$1884>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
         $$0$i46$i = $1887;
        } else {
         $1888 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
         $1889 = HEAP32[$1888>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
         $1890 = ($1889|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
         if ($1890) {
          label = 636;
          break;
         }
         $1891 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
         $1892 = HEAP32[$1891>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $1893 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
         $1894 = HEAP32[$1893>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $1895 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
         $1896 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
         $1897 = HEAP32[$1896>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $1898 = (FUNCTION_TABLE_iiii[$1892 & 8191]($1894,$1895,$1897)|0); //@line 1213 "./libraries/stb/stb_image.h"
         $1899 = ($1898|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
         if ($1899) {
          HEAP32[$1888>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
          $1900 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
          HEAP8[$1895>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
          $$pre$phi$i43$iZ2D = $1900;$$sink$i$i44$i = $1900;$3767 = 0;
         } else {
          $1901 = (((($s)) + 40|0) + ($1898)|0); //@line 1223 "./libraries/stb/stb_image.h"
          $$pre$i41$i = HEAP8[$1895>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
          $$pre1$i42$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
          $$pre$phi$i43$iZ2D = $$pre1$i42$i;$$sink$i$i44$i = $1901;$3767 = $$pre$i41$i;
         }
         HEAP32[$2>>2] = $$sink$i$i44$i; //@line 1223 "./libraries/stb/stb_image.h"
         HEAP32[$0>>2] = $$pre$phi$i43$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
         $$0$i46$i = $3767;
        }
        $1902 = ($$0$i46$i<<24>>24)==(77); //@line 4606 "./libraries/stb/stb_image.h"
        if ($1902) {
         (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4607 "./libraries/stb/stb_image.h"
         $1903 = HEAP32[$0>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $1904 = HEAP32[$2>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $1905 = ($1903>>>0)<($1904>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         if ($1905) {
          $1906 = ((($1903)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
          HEAP32[$0>>2] = $1906; //@line 1230 "./libraries/stb/stb_image.h"
          $1922 = $1906;$1923 = $1904;
         } else {
          $1907 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
          $1908 = HEAP32[$1907>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
          $1909 = ($1908|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
          if ($1909) {
           $1922 = $1903;$1923 = $1904;
          } else {
           $1910 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1911 = HEAP32[$1910>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1912 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1913 = HEAP32[$1912>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1914 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1915 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1916 = HEAP32[$1915>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1917 = (FUNCTION_TABLE_iiii[$1911 & 8191]($1913,$1914,$1916)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1918 = ($1917|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($1918) {
            HEAP32[$1907>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $1919 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$1914>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i$i$i20Z2D = $1919;$$sink$i$i$i$i21 = $1919;
           } else {
            $1920 = (((($s)) + 40|0) + ($1917)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre1$i$i$i19 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i$i$i20Z2D = $$pre1$i$i$i19;$$sink$i$i$i$i21 = $1920;
           }
           HEAP32[$2>>2] = $$sink$i$i$i$i21; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$0>>2] = $$pre$phi$i$i$i20Z2D; //@line 1233 "./libraries/stb/stb_image.h"
           $1922 = $$pre$phi$i$i$i20Z2D;$1923 = $$sink$i$i$i$i21;
          }
         }
         $1921 = ($1922>>>0)<($1923>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         do {
          if ($1921) {
           $1924 = ((($1922)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
           HEAP32[$0>>2] = $1924; //@line 1230 "./libraries/stb/stb_image.h"
           $1940 = $1924;$1941 = $1923;
          } else {
           $1925 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $1926 = HEAP32[$1925>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
           $1927 = ($1926|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
           if ($1927) {
            $1940 = $1922;$1941 = $1923;
            break;
           }
           $1928 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1929 = HEAP32[$1928>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1930 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1931 = HEAP32[$1930>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1932 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1933 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1934 = HEAP32[$1933>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1935 = (FUNCTION_TABLE_iiii[$1929 & 8191]($1931,$1932,$1934)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1936 = ($1935|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($1936) {
            HEAP32[$1925>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $1937 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$1932>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i3$i$iZ2D = $1937;$$sink$i$i4$i$i = $1937;
           } else {
            $1938 = (((($s)) + 40|0) + ($1935)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre1$i2$i$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i3$i$iZ2D = $$pre1$i2$i$i;$$sink$i$i4$i$i = $1938;
           }
           HEAP32[$2>>2] = $$sink$i$i4$i$i; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$0>>2] = $$pre$phi$i3$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
           $1940 = $$pre$phi$i3$i$iZ2D;$1941 = $$sink$i$i4$i$i;
          }
         } while(0);
         $1939 = ($1940>>>0)<($1941>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         do {
          if ($1939) {
           $1942 = ((($1940)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
           HEAP32[$0>>2] = $1942; //@line 1230 "./libraries/stb/stb_image.h"
           $1958 = $1942;$1959 = $1941;
          } else {
           $1943 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $1944 = HEAP32[$1943>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
           $1945 = ($1944|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
           if ($1945) {
            $1958 = $1940;$1959 = $1941;
            break;
           }
           $1946 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1947 = HEAP32[$1946>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1948 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1949 = HEAP32[$1948>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1950 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1951 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1952 = HEAP32[$1951>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1953 = (FUNCTION_TABLE_iiii[$1947 & 8191]($1949,$1950,$1952)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1954 = ($1953|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($1954) {
            HEAP32[$1943>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $1955 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$1950>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i$i50$iZ2D = $1955;$$sink$i$i$i51$i = $1955;
           } else {
            $1956 = (((($s)) + 40|0) + ($1953)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre1$i$i49$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i$i50$iZ2D = $$pre1$i$i49$i;$$sink$i$i$i51$i = $1956;
           }
           HEAP32[$2>>2] = $$sink$i$i$i51$i; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$0>>2] = $$pre$phi$i$i50$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
           $1958 = $$pre$phi$i$i50$iZ2D;$1959 = $$sink$i$i$i51$i;
          }
         } while(0);
         $1957 = ($1958>>>0)<($1959>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         do {
          if ($1957) {
           $1960 = ((($1958)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
           HEAP32[$0>>2] = $1960; //@line 1230 "./libraries/stb/stb_image.h"
          } else {
           $1961 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $1962 = HEAP32[$1961>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
           $1963 = ($1962|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
           if ($1963) {
            break;
           }
           $1964 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1965 = HEAP32[$1964>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1966 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1967 = HEAP32[$1966>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1968 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1969 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1970 = HEAP32[$1969>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $1971 = (FUNCTION_TABLE_iiii[$1965 & 8191]($1967,$1968,$1970)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $1972 = ($1971|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($1972) {
            HEAP32[$1961>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $1973 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$1968>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i3$i57$iZ2D = $1973;$$sink$i$i4$i58$i = $1973;
           } else {
            $1974 = (((($s)) + 40|0) + ($1971)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre1$i2$i56$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i3$i57$iZ2D = $$pre1$i2$i56$i;$$sink$i$i4$i58$i = $1974;
           }
           HEAP32[$2>>2] = $$sink$i$i4$i58$i; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$0>>2] = $$pre$phi$i3$i57$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
          }
         } while(0);
         $1975 = (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4610 "./libraries/stb/stb_image.h"
         $1976 = (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4611 "./libraries/stb/stb_image.h"
         $1977 = ($1976|0)==(12); //@line 4612 "./libraries/stb/stb_image.h"
         switch ($1976|0) {
         case 12: case 40: case 56: case 108: case 124:  {
          break;
         }
         default: {
          HEAP32[158754] = 441379; //@line 894 "./libraries/stb/stb_image.h"
          $$2$i = 0;
          break L866;
         }
         }
         if ($1977) {
          $1978 = HEAP32[$0>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $1979 = HEAP32[$2>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $1980 = ($1978>>>0)<($1979>>>0); //@line 1229 "./libraries/stb/stb_image.h"
          do {
           if ($1980) {
            $1981 = ((($1978)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
            HEAP32[$0>>2] = $1981; //@line 1230 "./libraries/stb/stb_image.h"
            $1982 = HEAP8[$1978>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
            $$0$i$i67$i = $1982;$1999 = $1981;$2000 = $1979;
           } else {
            $1983 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
            $1984 = HEAP32[$1983>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
            $1985 = ($1984|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
            if ($1985) {
             $$0$i$i67$i = 0;$1999 = $1978;$2000 = $1979;
             break;
            }
            $1986 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
            $1987 = HEAP32[$1986>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $1988 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
            $1989 = HEAP32[$1988>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $1990 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
            $1991 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
            $1992 = HEAP32[$1991>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $1993 = (FUNCTION_TABLE_iiii[$1987 & 8191]($1989,$1990,$1992)|0); //@line 1213 "./libraries/stb/stb_image.h"
            $1994 = ($1993|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
            if ($1994) {
             HEAP32[$1983>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
             $1995 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
             HEAP8[$1990>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
             $$pre$phi$i$i64$iZ2D = $1995;$$sink$i$i$i65$i = $1995;$3768 = 0;
            } else {
             $1996 = (((($s)) + 40|0) + ($1993)|0); //@line 1223 "./libraries/stb/stb_image.h"
             $$pre$i$i62$i = HEAP8[$1990>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
             $$pre1$i$i63$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
             $$pre$phi$i$i64$iZ2D = $$pre1$i$i63$i;$$sink$i$i$i65$i = $1996;$3768 = $$pre$i$i62$i;
            }
            HEAP32[$2>>2] = $$sink$i$i$i65$i; //@line 1223 "./libraries/stb/stb_image.h"
            HEAP32[$0>>2] = $$pre$phi$i$i64$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
            $$0$i$i67$i = $3768;$1999 = $$pre$phi$i$i64$iZ2D;$2000 = $$sink$i$i$i65$i;
           }
          } while(0);
          $1997 = $$0$i$i67$i&255; //@line 1308 "./libraries/stb/stb_image.h"
          $1998 = ($1999>>>0)<($2000>>>0); //@line 1229 "./libraries/stb/stb_image.h"
          do {
           if ($1998) {
            $2001 = ((($1999)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
            HEAP32[$0>>2] = $2001; //@line 1230 "./libraries/stb/stb_image.h"
            $2002 = HEAP8[$1999>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
            $$0$i6$i74$i = $2002;$2021 = $2001;$2022 = $2000;
           } else {
            $2003 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
            $2004 = HEAP32[$2003>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
            $2005 = ($2004|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
            if ($2005) {
             $$0$i6$i74$i = 0;$2021 = $1999;$2022 = $2000;
             break;
            }
            $2006 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2007 = HEAP32[$2006>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $2008 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2009 = HEAP32[$2008>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $2010 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2011 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2012 = HEAP32[$2011>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $2013 = (FUNCTION_TABLE_iiii[$2007 & 8191]($2009,$2010,$2012)|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2014 = ($2013|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
            if ($2014) {
             HEAP32[$2003>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
             $2015 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
             HEAP8[$2010>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
             $$pre$phi$i3$i71$iZ2D = $2015;$$sink$i$i4$i72$i = $2015;$3769 = 0;
            } else {
             $2016 = (((($s)) + 40|0) + ($2013)|0); //@line 1223 "./libraries/stb/stb_image.h"
             $$pre$i1$i69$i = HEAP8[$2010>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
             $$pre1$i2$i70$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
             $$pre$phi$i3$i71$iZ2D = $$pre1$i2$i70$i;$$sink$i$i4$i72$i = $2016;$3769 = $$pre$i1$i69$i;
            }
            HEAP32[$2>>2] = $$sink$i$i4$i72$i; //@line 1223 "./libraries/stb/stb_image.h"
            HEAP32[$0>>2] = $$pre$phi$i3$i71$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
            $$0$i6$i74$i = $3769;$2021 = $$pre$phi$i3$i71$iZ2D;$2022 = $$sink$i$i4$i72$i;
           }
          } while(0);
          $2017 = $$0$i6$i74$i&255; //@line 1309 "./libraries/stb/stb_image.h"
          $2018 = $2017 << 8; //@line 1309 "./libraries/stb/stb_image.h"
          $2019 = $2018 | $1997; //@line 1309 "./libraries/stb/stb_image.h"
          HEAP32[$s>>2] = $2019; //@line 4614 "./libraries/stb/stb_image.h"
          $2020 = ($2021>>>0)<($2022>>>0); //@line 1229 "./libraries/stb/stb_image.h"
          do {
           if ($2020) {
            $2023 = ((($2021)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
            HEAP32[$0>>2] = $2023; //@line 1230 "./libraries/stb/stb_image.h"
            $2024 = HEAP8[$2021>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
            $$0$i$i81$i = $2024;$2041 = $2023;$2042 = $2022;
           } else {
            $2025 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
            $2026 = HEAP32[$2025>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
            $2027 = ($2026|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
            if ($2027) {
             $$0$i$i81$i = 0;$2041 = $2021;$2042 = $2022;
             break;
            }
            $2028 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2029 = HEAP32[$2028>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $2030 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2031 = HEAP32[$2030>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $2032 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2033 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2034 = HEAP32[$2033>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $2035 = (FUNCTION_TABLE_iiii[$2029 & 8191]($2031,$2032,$2034)|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2036 = ($2035|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
            if ($2036) {
             HEAP32[$2025>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
             $2037 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
             HEAP8[$2032>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
             $$pre$phi$i$i78$iZ2D = $2037;$$sink$i$i$i79$i = $2037;$3770 = 0;
            } else {
             $2038 = (((($s)) + 40|0) + ($2035)|0); //@line 1223 "./libraries/stb/stb_image.h"
             $$pre$i$i76$i = HEAP8[$2032>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
             $$pre1$i$i77$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
             $$pre$phi$i$i78$iZ2D = $$pre1$i$i77$i;$$sink$i$i$i79$i = $2038;$3770 = $$pre$i$i76$i;
            }
            HEAP32[$2>>2] = $$sink$i$i$i79$i; //@line 1223 "./libraries/stb/stb_image.h"
            HEAP32[$0>>2] = $$pre$phi$i$i78$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
            $$0$i$i81$i = $3770;$2041 = $$pre$phi$i$i78$iZ2D;$2042 = $$sink$i$i$i79$i;
           }
          } while(0);
          $2039 = $$0$i$i81$i&255; //@line 1308 "./libraries/stb/stb_image.h"
          $2040 = ($2041>>>0)<($2042>>>0); //@line 1229 "./libraries/stb/stb_image.h"
          do {
           if ($2040) {
            $2043 = ((($2041)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
            HEAP32[$0>>2] = $2043; //@line 1230 "./libraries/stb/stb_image.h"
            $2044 = HEAP8[$2041>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
            $$0$i6$i88$i = $2044;$3771 = $2042;$3772 = $2043;
           } else {
            $2045 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
            $2046 = HEAP32[$2045>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
            $2047 = ($2046|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
            if ($2047) {
             $$0$i6$i88$i = 0;$3771 = $2042;$3772 = $2041;
             break;
            }
            $2048 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2049 = HEAP32[$2048>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $2050 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2051 = HEAP32[$2050>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $2052 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2053 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2054 = HEAP32[$2053>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
            $2055 = (FUNCTION_TABLE_iiii[$2049 & 8191]($2051,$2052,$2054)|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2056 = ($2055|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
            if ($2056) {
             HEAP32[$2045>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
             $2057 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
             HEAP8[$2052>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
             $$pre$phi$i3$i85$iZ2D = $2057;$$sink$i$i4$i86$i = $2057;$3773 = 0;
            } else {
             $2058 = (((($s)) + 40|0) + ($2055)|0); //@line 1223 "./libraries/stb/stb_image.h"
             $$pre$i1$i83$i = HEAP8[$2052>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
             $$pre1$i2$i84$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
             $$pre$phi$i3$i85$iZ2D = $$pre1$i2$i84$i;$$sink$i$i4$i86$i = $2058;$3773 = $$pre$i1$i83$i;
            }
            HEAP32[$2>>2] = $$sink$i$i4$i86$i; //@line 1223 "./libraries/stb/stb_image.h"
            HEAP32[$0>>2] = $$pre$phi$i3$i85$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
            $$0$i6$i88$i = $3773;$3771 = $$sink$i$i4$i86$i;$3772 = $$pre$phi$i3$i85$iZ2D;
           }
          } while(0);
          $2059 = $$0$i6$i88$i&255; //@line 1309 "./libraries/stb/stb_image.h"
          $2060 = $2059 << 8; //@line 1309 "./libraries/stb/stb_image.h"
          $2061 = $2060 | $2039; //@line 1309 "./libraries/stb/stb_image.h"
          $2062 = ((($s)) + 4|0); //@line 4615 "./libraries/stb/stb_image.h"
          HEAP32[$2062>>2] = $2061; //@line 4615 "./libraries/stb/stb_image.h"
          $2067 = $3772;$2068 = $3771;
         } else {
          $2063 = (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4617 "./libraries/stb/stb_image.h"
          HEAP32[$s>>2] = $2063; //@line 4617 "./libraries/stb/stb_image.h"
          $2064 = (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4618 "./libraries/stb/stb_image.h"
          $2065 = ((($s)) + 4|0); //@line 4618 "./libraries/stb/stb_image.h"
          HEAP32[$2065>>2] = $2064; //@line 4618 "./libraries/stb/stb_image.h"
          $$pre$i24 = HEAP32[$0>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $$pre345$i = HEAP32[$2>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $2067 = $$pre$i24;$2068 = $$pre345$i;
         }
         $2066 = ($2067>>>0)<($2068>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         do {
          if ($2066) {
           $2069 = ((($2067)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
           HEAP32[$0>>2] = $2069; //@line 1230 "./libraries/stb/stb_image.h"
           $2070 = HEAP8[$2067>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
           $$0$i$i95$i = $2070;$2087 = $2069;$2088 = $2068;
          } else {
           $2071 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $2072 = HEAP32[$2071>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
           $2073 = ($2072|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
           if ($2073) {
            $$0$i$i95$i = 0;$2087 = $2067;$2088 = $2068;
            break;
           }
           $2074 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2075 = HEAP32[$2074>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $2076 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2077 = HEAP32[$2076>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $2078 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2079 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2080 = HEAP32[$2079>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $2081 = (FUNCTION_TABLE_iiii[$2075 & 8191]($2077,$2078,$2080)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2082 = ($2081|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($2082) {
            HEAP32[$2071>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $2083 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$2078>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i$i92$iZ2D = $2083;$$sink$i$i$i93$i = $2083;$3774 = 0;
           } else {
            $2084 = (((($s)) + 40|0) + ($2081)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre$i$i90$i = HEAP8[$2078>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
            $$pre1$i$i91$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i$i92$iZ2D = $$pre1$i$i91$i;$$sink$i$i$i93$i = $2084;$3774 = $$pre$i$i90$i;
           }
           HEAP32[$2>>2] = $$sink$i$i$i93$i; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$0>>2] = $$pre$phi$i$i92$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
           $$0$i$i95$i = $3774;$2087 = $$pre$phi$i$i92$iZ2D;$2088 = $$sink$i$i$i93$i;
          }
         } while(0);
         $2085 = $$0$i$i95$i&255; //@line 1308 "./libraries/stb/stb_image.h"
         $2086 = ($2087>>>0)<($2088>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         do {
          if ($2086) {
           $2089 = ((($2087)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
           HEAP32[$0>>2] = $2089; //@line 1230 "./libraries/stb/stb_image.h"
           $2090 = HEAP8[$2087>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
           $$0$i6$i102$i = $2090;$2110 = $2089;$2111 = $2088;
          } else {
           $2091 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $2092 = HEAP32[$2091>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
           $2093 = ($2092|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
           if ($2093) {
            $$0$i6$i102$i = 0;$2110 = $2087;$2111 = $2088;
            break;
           }
           $2094 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2095 = HEAP32[$2094>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $2096 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2097 = HEAP32[$2096>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $2098 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2099 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2100 = HEAP32[$2099>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $2101 = (FUNCTION_TABLE_iiii[$2095 & 8191]($2097,$2098,$2100)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2102 = ($2101|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($2102) {
            HEAP32[$2091>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $2103 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$2098>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i3$i99$iZ2D = $2103;$$sink$i$i4$i100$i = $2103;$3775 = 0;
           } else {
            $2104 = (((($s)) + 40|0) + ($2101)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre$i1$i97$i = HEAP8[$2098>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
            $$pre1$i2$i98$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i3$i99$iZ2D = $$pre1$i2$i98$i;$$sink$i$i4$i100$i = $2104;$3775 = $$pre$i1$i97$i;
           }
           HEAP32[$2>>2] = $$sink$i$i4$i100$i; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$0>>2] = $$pre$phi$i3$i99$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
           $$0$i6$i102$i = $3775;$2110 = $$pre$phi$i3$i99$iZ2D;$2111 = $$sink$i$i4$i100$i;
          }
         } while(0);
         $2105 = $$0$i6$i102$i&255; //@line 1309 "./libraries/stb/stb_image.h"
         $2106 = $2105 << 8; //@line 1309 "./libraries/stb/stb_image.h"
         $2107 = $2106 | $2085; //@line 1309 "./libraries/stb/stb_image.h"
         $2108 = ($2107|0)==(1); //@line 4620 "./libraries/stb/stb_image.h"
         if (!($2108)) {
          HEAP32[158754] = 441391; //@line 894 "./libraries/stb/stb_image.h"
          $$2$i = 0;
          break;
         }
         $2109 = ($2110>>>0)<($2111>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         do {
          if ($2109) {
           $2112 = ((($2110)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
           HEAP32[$0>>2] = $2112; //@line 1230 "./libraries/stb/stb_image.h"
           $2113 = HEAP8[$2110>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
           $$0$i$i109$i = $2113;$2130 = $2112;$2131 = $2111;
          } else {
           $2114 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $2115 = HEAP32[$2114>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
           $2116 = ($2115|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
           if ($2116) {
            $$0$i$i109$i = 0;$2130 = $2110;$2131 = $2111;
            break;
           }
           $2117 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2118 = HEAP32[$2117>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $2119 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2120 = HEAP32[$2119>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $2121 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2122 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2123 = HEAP32[$2122>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $2124 = (FUNCTION_TABLE_iiii[$2118 & 8191]($2120,$2121,$2123)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2125 = ($2124|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($2125) {
            HEAP32[$2114>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $2126 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$2121>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i$i106$iZ2D = $2126;$$sink$i$i$i107$i = $2126;$3776 = 0;
           } else {
            $2127 = (((($s)) + 40|0) + ($2124)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre$i$i104$i = HEAP8[$2121>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
            $$pre1$i$i105$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i$i106$iZ2D = $$pre1$i$i105$i;$$sink$i$i$i107$i = $2127;$3776 = $$pre$i$i104$i;
           }
           HEAP32[$2>>2] = $$sink$i$i$i107$i; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$0>>2] = $$pre$phi$i$i106$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
           $$0$i$i109$i = $3776;$2130 = $$pre$phi$i$i106$iZ2D;$2131 = $$sink$i$i$i107$i;
          }
         } while(0);
         $2128 = $$0$i$i109$i&255; //@line 1308 "./libraries/stb/stb_image.h"
         $2129 = ($2130>>>0)<($2131>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         do {
          if ($2129) {
           $2132 = ((($2130)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
           HEAP32[$0>>2] = $2132; //@line 1230 "./libraries/stb/stb_image.h"
           $2133 = HEAP8[$2130>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
           $$0$i6$i116$i = $2133;
          } else {
           $2134 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $2135 = HEAP32[$2134>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
           $2136 = ($2135|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
           if ($2136) {
            $$0$i6$i116$i = 0;
            break;
           }
           $2137 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2138 = HEAP32[$2137>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $2139 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2140 = HEAP32[$2139>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $2141 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2142 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2143 = HEAP32[$2142>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $2144 = (FUNCTION_TABLE_iiii[$2138 & 8191]($2140,$2141,$2143)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2145 = ($2144|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($2145) {
            HEAP32[$2134>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $2146 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$2141>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i3$i113$iZ2D = $2146;$$sink$i$i4$i114$i = $2146;$3777 = 0;
           } else {
            $2147 = (((($s)) + 40|0) + ($2144)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre$i1$i111$i = HEAP8[$2141>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
            $$pre1$i2$i112$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i3$i113$iZ2D = $$pre1$i2$i112$i;$$sink$i$i4$i114$i = $2147;$3777 = $$pre$i1$i111$i;
           }
           HEAP32[$2>>2] = $$sink$i$i4$i114$i; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$0>>2] = $$pre$phi$i3$i113$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
           $$0$i6$i116$i = $3777;
          }
         } while(0);
         $2148 = $$0$i6$i116$i&255; //@line 1309 "./libraries/stb/stb_image.h"
         $2149 = $2148 << 8; //@line 1309 "./libraries/stb/stb_image.h"
         $2150 = $2149 | $2128; //@line 1309 "./libraries/stb/stb_image.h"
         $2151 = ($2150|0)==(1); //@line 4622 "./libraries/stb/stb_image.h"
         if ($2151) {
          HEAP32[158754] = 441399; //@line 894 "./libraries/stb/stb_image.h"
          $$2$i = 0;
          break;
         }
         $2152 = ((($s)) + 4|0); //@line 4623 "./libraries/stb/stb_image.h"
         $2153 = HEAP32[$2152>>2]|0; //@line 4623 "./libraries/stb/stb_image.h"
         $2154 = ($2153|0)>(0); //@line 4623 "./libraries/stb/stb_image.h"
         $ispos$i = ($2153|0)>(-1); //@line 4624 "./libraries/stb/stb_image.h"
         $neg$i = (0 - ($2153))|0; //@line 4624 "./libraries/stb/stb_image.h"
         $2155 = $ispos$i ? $2153 : $neg$i; //@line 4624 "./libraries/stb/stb_image.h"
         HEAP32[$2152>>2] = $2155; //@line 4624 "./libraries/stb/stb_image.h"
         L1000: do {
          if ($1977) {
           $2156 = ($2150>>>0)<(24); //@line 4626 "./libraries/stb/stb_image.h"
           if (!($2156)) {
            $all_a$1$i = 255;$ma$1$i = 0;$mb$1$i = 0;$mg$1$i = 0;$mr$1$i = 0;$psize$0$i = 0;
            break;
           }
           $2157 = (($1975) + -38)|0; //@line 4627 "./libraries/stb/stb_image.h"
           $2158 = (($2157|0) / 3)&-1; //@line 4627 "./libraries/stb/stb_image.h"
           $all_a$1$i = 255;$ma$1$i = 0;$mb$1$i = 0;$mg$1$i = 0;$mr$1$i = 0;$psize$0$i = $2158;
          } else {
           $2159 = (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4629 "./libraries/stb/stb_image.h"
           $$off$i = (($2159) + -1)|0; //@line 4630 "./libraries/stb/stb_image.h"
           $2160 = ($$off$i>>>0)<(2); //@line 4630 "./libraries/stb/stb_image.h"
           if ($2160) {
            HEAP32[158754] = 441410; //@line 894 "./libraries/stb/stb_image.h"
            $$2$i = 0;
            break L866;
           }
           (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4631 "./libraries/stb/stb_image.h"
           (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4632 "./libraries/stb/stb_image.h"
           (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4633 "./libraries/stb/stb_image.h"
           (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4634 "./libraries/stb/stb_image.h"
           (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4635 "./libraries/stb/stb_image.h"
           $2161 = $1976 | 16; //@line 4636 "./libraries/stb/stb_image.h"
           L1007: do {
            switch ($2161|0) {
            case 56:  {
             $2162 = ($1976|0)==(56); //@line 4636 "./libraries/stb/stb_image.h"
             if ($2162) {
              (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4638 "./libraries/stb/stb_image.h"
              (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4639 "./libraries/stb/stb_image.h"
              (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4640 "./libraries/stb/stb_image.h"
              (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4641 "./libraries/stb/stb_image.h"
             }
             $2163 = ($2150|0)==(32); //@line 4643 "./libraries/stb/stb_image.h"
             switch ($2150|0) {
             case 16: case 32:  {
              break;
             }
             default: {
              $ma$0$i = 0;$mb$0$i = 0;$mg$0$i = 0;$mr$0$i = 0;
              break L1007;
             }
             }
             switch ($2159|0) {
             case 0:  {
              $$$i = $2163 ? 0 : 255; //@line 4652 "./libraries/stb/stb_image.h"
              $$34$i = $2163 ? -16777216 : 0; //@line 4652 "./libraries/stb/stb_image.h"
              $$35$i = $2163 ? 255 : 31; //@line 4652 "./libraries/stb/stb_image.h"
              $$36$i = $2163 ? 65280 : 992; //@line 4652 "./libraries/stb/stb_image.h"
              $$37$i = $2163 ? 16711680 : 31744; //@line 4652 "./libraries/stb/stb_image.h"
              $all_a$1$i = $$$i;$ma$1$i = $$34$i;$mb$1$i = $$35$i;$mg$1$i = $$36$i;$mr$1$i = $$37$i;$psize$0$i = 0;
              break L1000;
              break;
             }
             case 3:  {
              $2164 = (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4658 "./libraries/stb/stb_image.h"
              $2165 = (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4659 "./libraries/stb/stb_image.h"
              $2166 = (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4660 "./libraries/stb/stb_image.h"
              $2167 = ($2164|0)==($2165|0); //@line 4662 "./libraries/stb/stb_image.h"
              $2168 = ($2165|0)==($2166|0); //@line 4662 "./libraries/stb/stb_image.h"
              $or$cond$i25 = $2167 & $2168; //@line 4662 "./libraries/stb/stb_image.h"
              if (!($or$cond$i25)) {
               $all_a$1$i = 255;$ma$1$i = 0;$mb$1$i = $2166;$mg$1$i = $2165;$mr$1$i = $2164;$psize$0$i = 0;
               break L1000;
              }
              HEAP32[158754] = 441391; //@line 894 "./libraries/stb/stb_image.h"
              $$2$i = 0;
              break L866;
              break;
             }
             default: {
              HEAP32[158754] = 441391; //@line 894 "./libraries/stb/stb_image.h"
              $$2$i = 0;
              break L866;
             }
             }
             break;
            }
            case 124:  {
             $2174 = ($1976|0)==(124); //@line 4670 "./libraries/stb/stb_image.h"
             $2175 = (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4671 "./libraries/stb/stb_image.h"
             $2176 = (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4672 "./libraries/stb/stb_image.h"
             $2177 = (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4673 "./libraries/stb/stb_image.h"
             $2178 = (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4674 "./libraries/stb/stb_image.h"
             (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4675 "./libraries/stb/stb_image.h"
             (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4677 "./libraries/stb/stb_image.h"
             (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4677 "./libraries/stb/stb_image.h"
             (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4677 "./libraries/stb/stb_image.h"
             (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4677 "./libraries/stb/stb_image.h"
             (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4677 "./libraries/stb/stb_image.h"
             (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4677 "./libraries/stb/stb_image.h"
             (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4677 "./libraries/stb/stb_image.h"
             (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4677 "./libraries/stb/stb_image.h"
             (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4677 "./libraries/stb/stb_image.h"
             (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4677 "./libraries/stb/stb_image.h"
             (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4677 "./libraries/stb/stb_image.h"
             (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4677 "./libraries/stb/stb_image.h"
             if (!($2174)) {
              $ma$0$i = $2178;$mb$0$i = $2177;$mg$0$i = $2176;$mr$0$i = $2175;
              break L1007;
             }
             (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4679 "./libraries/stb/stb_image.h"
             (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4680 "./libraries/stb/stb_image.h"
             (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4681 "./libraries/stb/stb_image.h"
             (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4682 "./libraries/stb/stb_image.h"
             $ma$0$i = $2178;$mb$0$i = $2177;$mg$0$i = $2176;$mr$0$i = $2175;
             break;
            }
            default: {
             $2169 = (___cxa_allocate_exception(16)|0); //@line 27 "modules/image/magpie/STBHandler.cpp"
             __THREW__ = 0;
             HEAP32[$vararg_buffer>>2] = 441418; //@line 27 "modules/image/magpie/STBHandler.cpp"
             invoke_viii(3349,($2169|0),(441193|0),($vararg_buffer|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
             $2170 = __THREW__; __THREW__ = 0;
             $2171 = $2170&1;
             if ($2171) {
              $2172 = ___cxa_find_matching_catch_2()|0;
              $2173 = tempRet0;
              ___cxa_free_exception(($2169|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
              ___resumeException($2172|0);
              // unreachable;
             } else {
              ___cxa_throw(($2169|0),(96|0),(344|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
              // unreachable; //@line 27 "modules/image/magpie/STBHandler.cpp"
             }
            }
            }
           } while(0);
           $2179 = ($2150>>>0)<(16); //@line 4685 "./libraries/stb/stb_image.h"
           if (!($2179)) {
            $all_a$1$i = 255;$ma$1$i = $ma$0$i;$mb$1$i = $mb$0$i;$mg$1$i = $mg$0$i;$mr$1$i = $mr$0$i;$psize$0$i = 0;
            break;
           }
           $2180 = (($1975) + -14)|0; //@line 4686 "./libraries/stb/stb_image.h"
           $2181 = (($2180) - ($1976))|0; //@line 4686 "./libraries/stb/stb_image.h"
           $2182 = $2181 >> 2; //@line 4686 "./libraries/stb/stb_image.h"
           $all_a$1$i = 255;$ma$1$i = $ma$0$i;$mb$1$i = $mb$0$i;$mg$1$i = $mg$0$i;$mr$1$i = $mr$0$i;$psize$0$i = $2182;
          }
         } while(0);
         $2183 = ($ma$1$i|0)!=(0); //@line 4688 "./libraries/stb/stb_image.h"
         $2184 = $2183 ? 4 : 3; //@line 4688 "./libraries/stb/stb_image.h"
         $2185 = ((($s)) + 8|0); //@line 4688 "./libraries/stb/stb_image.h"
         HEAP32[$2185>>2] = $2184; //@line 4688 "./libraries/stb/stb_image.h"
         $2186 = ($req_comp|0)==(0); //@line 4689 "./libraries/stb/stb_image.h"
         $2187 = ($req_comp|0)>(2); //@line 4689 "./libraries/stb/stb_image.h"
         $req_comp$$i = $2187 ? $req_comp : $2184; //@line 4690 "./libraries/stb/stb_image.h"
         $2188 = HEAP32[$s>>2]|0; //@line 4693 "./libraries/stb/stb_image.h"
         $2189 = Math_imul($2188, $req_comp$$i)|0; //@line 4693 "./libraries/stb/stb_image.h"
         $2190 = HEAP32[$2152>>2]|0; //@line 4693 "./libraries/stb/stb_image.h"
         $2191 = Math_imul($2189, $2190)|0; //@line 4693 "./libraries/stb/stb_image.h"
         $2192 = (_malloc($2191)|0); //@line 900 "./libraries/stb/stb_image.h"
         $2193 = ($2192|0)==(0|0); //@line 4694 "./libraries/stb/stb_image.h"
         if ($2193) {
          HEAP32[158754] = 441002; //@line 894 "./libraries/stb/stb_image.h"
          $$2$i = 0;
          break;
         }
         $2194 = ($2150>>>0)<(16); //@line 4695 "./libraries/stb/stb_image.h"
         L1030: do {
          if ($2194) {
           $2195 = ($psize$0$i|0)==(0); //@line 4697 "./libraries/stb/stb_image.h"
           $2196 = ($psize$0$i|0)>(256); //@line 4697 "./libraries/stb/stb_image.h"
           $or$cond18$i = $2195 | $2196; //@line 4697 "./libraries/stb/stb_image.h"
           if ($or$cond18$i) {
            _free($2192); //@line 4697 "./libraries/stb/stb_image.h"
            HEAP32[158754] = 441443; //@line 894 "./libraries/stb/stb_image.h"
            $$2$i = 0;
            break L866;
           }
           $2197 = ($psize$0$i|0)>(0); //@line 4698 "./libraries/stb/stb_image.h"
           if ($2197) {
            $2198 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
            $2199 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2200 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2201 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2202 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
            $2203 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            $$pre348$i = HEAP32[$0>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
            $$pre349$i = HEAP32[$2>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
            $2205 = $$pre348$i;$2206 = $$pre349$i;$i$1315$i = 0;
            while(1) {
             $2204 = ($2205>>>0)<($2206>>>0); //@line 1229 "./libraries/stb/stb_image.h"
             do {
              if ($2204) {
               $2207 = ((($2205)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
               HEAP32[$0>>2] = $2207; //@line 1230 "./libraries/stb/stb_image.h"
               $2208 = HEAP8[$2205>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
               $$0$i123$i = $2208;$2219 = $2207;$2220 = $2206;
              } else {
               $2209 = HEAP32[$2198>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
               $2210 = ($2209|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
               if ($2210) {
                $$0$i123$i = 0;$2219 = $2205;$2220 = $2206;
                break;
               }
               $2211 = HEAP32[$2199>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
               $2212 = HEAP32[$2200>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
               $2213 = HEAP32[$2202>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
               $2214 = (FUNCTION_TABLE_iiii[$2211 & 8191]($2212,$2201,$2213)|0); //@line 1213 "./libraries/stb/stb_image.h"
               $2215 = ($2214|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
               if ($2215) {
                HEAP32[$2198>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                HEAP8[$2201>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                $$sink$i$i121$i = $2203;$3778 = 0;
               } else {
                $2216 = (((($s)) + 40|0) + ($2214)|0); //@line 1223 "./libraries/stb/stb_image.h"
                $$pre$i118$i = HEAP8[$2201>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                $$sink$i$i121$i = $2216;$3778 = $$pre$i118$i;
               }
               HEAP32[$2>>2] = $$sink$i$i121$i; //@line 1223 "./libraries/stb/stb_image.h"
               HEAP32[$0>>2] = $2203; //@line 1233 "./libraries/stb/stb_image.h"
               $$0$i123$i = $3778;$2219 = $2203;$2220 = $$sink$i$i121$i;
              }
             } while(0);
             $2217 = (((($pal$i) + ($i$1315$i<<2)|0)) + 2|0); //@line 4699 "./libraries/stb/stb_image.h"
             HEAP8[$2217>>0] = $$0$i123$i; //@line 4699 "./libraries/stb/stb_image.h"
             $2218 = ($2219>>>0)<($2220>>>0); //@line 1229 "./libraries/stb/stb_image.h"
             do {
              if ($2218) {
               $2221 = ((($2219)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
               HEAP32[$0>>2] = $2221; //@line 1230 "./libraries/stb/stb_image.h"
               $2222 = HEAP8[$2219>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
               $$0$i130$i = $2222;$2233 = $2221;$2234 = $2220;
              } else {
               $2223 = HEAP32[$2198>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
               $2224 = ($2223|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
               if ($2224) {
                $$0$i130$i = 0;$2233 = $2219;$2234 = $2220;
                break;
               }
               $2225 = HEAP32[$2199>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
               $2226 = HEAP32[$2200>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
               $2227 = HEAP32[$2202>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
               $2228 = (FUNCTION_TABLE_iiii[$2225 & 8191]($2226,$2201,$2227)|0); //@line 1213 "./libraries/stb/stb_image.h"
               $2229 = ($2228|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
               if ($2229) {
                HEAP32[$2198>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                HEAP8[$2201>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                $$sink$i$i128$i = $2203;$3779 = 0;
               } else {
                $2230 = (((($s)) + 40|0) + ($2228)|0); //@line 1223 "./libraries/stb/stb_image.h"
                $$pre$i125$i = HEAP8[$2201>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                $$sink$i$i128$i = $2230;$3779 = $$pre$i125$i;
               }
               HEAP32[$2>>2] = $$sink$i$i128$i; //@line 1223 "./libraries/stb/stb_image.h"
               HEAP32[$0>>2] = $2203; //@line 1233 "./libraries/stb/stb_image.h"
               $$0$i130$i = $3779;$2233 = $2203;$2234 = $$sink$i$i128$i;
              }
             } while(0);
             $2231 = (((($pal$i) + ($i$1315$i<<2)|0)) + 1|0); //@line 4700 "./libraries/stb/stb_image.h"
             HEAP8[$2231>>0] = $$0$i130$i; //@line 4700 "./libraries/stb/stb_image.h"
             $2232 = ($2233>>>0)<($2234>>>0); //@line 1229 "./libraries/stb/stb_image.h"
             do {
              if ($2232) {
               $2235 = ((($2233)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
               HEAP32[$0>>2] = $2235; //@line 1230 "./libraries/stb/stb_image.h"
               $2236 = HEAP8[$2233>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
               $$0$i137$i = $2236;$2247 = $2235;$2248 = $2234;
              } else {
               $2237 = HEAP32[$2198>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
               $2238 = ($2237|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
               if ($2238) {
                $$0$i137$i = 0;$2247 = $2233;$2248 = $2234;
                break;
               }
               $2239 = HEAP32[$2199>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
               $2240 = HEAP32[$2200>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
               $2241 = HEAP32[$2202>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
               $2242 = (FUNCTION_TABLE_iiii[$2239 & 8191]($2240,$2201,$2241)|0); //@line 1213 "./libraries/stb/stb_image.h"
               $2243 = ($2242|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
               if ($2243) {
                HEAP32[$2198>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                HEAP8[$2201>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                $$sink$i$i135$i = $2203;$3780 = 0;
               } else {
                $2244 = (((($s)) + 40|0) + ($2242)|0); //@line 1223 "./libraries/stb/stb_image.h"
                $$pre$i132$i = HEAP8[$2201>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                $$sink$i$i135$i = $2244;$3780 = $$pre$i132$i;
               }
               HEAP32[$2>>2] = $$sink$i$i135$i; //@line 1223 "./libraries/stb/stb_image.h"
               HEAP32[$0>>2] = $2203; //@line 1233 "./libraries/stb/stb_image.h"
               $$0$i137$i = $3780;$2247 = $2203;$2248 = $$sink$i$i135$i;
              }
             } while(0);
             $2245 = (($pal$i) + ($i$1315$i<<2)|0); //@line 4701 "./libraries/stb/stb_image.h"
             HEAP8[$2245>>0] = $$0$i137$i; //@line 4701 "./libraries/stb/stb_image.h"
             do {
              if ($1977) {
               $3781 = $2248;$3782 = $2247;
              } else {
               $2246 = ($2247>>>0)<($2248>>>0); //@line 1229 "./libraries/stb/stb_image.h"
               if ($2246) {
                $2249 = ((($2247)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                HEAP32[$0>>2] = $2249; //@line 1230 "./libraries/stb/stb_image.h"
                $3781 = $2248;$3782 = $2249;
                break;
               }
               $2250 = HEAP32[$2198>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
               $2251 = ($2250|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
               if ($2251) {
                $3781 = $2248;$3782 = $2247;
                break;
               }
               $2252 = HEAP32[$2199>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
               $2253 = HEAP32[$2200>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
               $2254 = HEAP32[$2202>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
               $2255 = (FUNCTION_TABLE_iiii[$2252 & 8191]($2253,$2201,$2254)|0); //@line 1213 "./libraries/stb/stb_image.h"
               $2256 = ($2255|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
               if ($2256) {
                HEAP32[$2198>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                HEAP8[$2201>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                $$sink$i$i142$i = $2203;
               } else {
                $2257 = (((($s)) + 40|0) + ($2255)|0); //@line 1223 "./libraries/stb/stb_image.h"
                $$sink$i$i142$i = $2257;
               }
               HEAP32[$2>>2] = $$sink$i$i142$i; //@line 1223 "./libraries/stb/stb_image.h"
               HEAP32[$0>>2] = $2203; //@line 1233 "./libraries/stb/stb_image.h"
               $3781 = $$sink$i$i142$i;$3782 = $2203;
              }
             } while(0);
             $2258 = (((($pal$i) + ($i$1315$i<<2)|0)) + 3|0); //@line 4703 "./libraries/stb/stb_image.h"
             HEAP8[$2258>>0] = -1; //@line 4703 "./libraries/stb/stb_image.h"
             $2259 = (($i$1315$i) + 1)|0; //@line 4698 "./libraries/stb/stb_image.h"
             $exitcond$i = ($2259|0)==($psize$0$i|0); //@line 4698 "./libraries/stb/stb_image.h"
             if ($exitcond$i) {
              break;
             } else {
              $2205 = $3782;$2206 = $3781;$i$1315$i = $2259;
             }
            }
           }
           $2260 = (($1975) + -14)|0; //@line 4705 "./libraries/stb/stb_image.h"
           $2261 = (($2260) - ($1976))|0; //@line 4705 "./libraries/stb/stb_image.h"
           $2262 = $1977 ? 3 : 4; //@line 4705 "./libraries/stb/stb_image.h"
           $2263 = Math_imul($psize$0$i, $2262)|0; //@line 4705 "./libraries/stb/stb_image.h"
           $2264 = (($2261) - ($2263))|0; //@line 4705 "./libraries/stb/stb_image.h"
           $2265 = ($2264|0)<(0); //@line 1252 "./libraries/stb/stb_image.h"
           L1078: do {
            if ($2265) {
             $2266 = HEAP32[$2>>2]|0; //@line 1253 "./libraries/stb/stb_image.h"
             HEAP32[$0>>2] = $2266; //@line 1253 "./libraries/stb/stb_image.h"
            } else {
             $2267 = ((($s)) + 16|0); //@line 1256 "./libraries/stb/stb_image.h"
             $2268 = HEAP32[$2267>>2]|0; //@line 1256 "./libraries/stb/stb_image.h"
             $2269 = ($2268|0)==(0|0); //@line 1256 "./libraries/stb/stb_image.h"
             do {
              if ($2269) {
               $$pre$i146$i = HEAP32[$0>>2]|0; //@line 1264 "./libraries/stb/stb_image.h"
               $2282 = $$pre$i146$i;
              } else {
               $2270 = HEAP32[$2>>2]|0; //@line 1257 "./libraries/stb/stb_image.h"
               $2271 = HEAP32[$0>>2]|0; //@line 1257 "./libraries/stb/stb_image.h"
               $2272 = $2270; //@line 1257 "./libraries/stb/stb_image.h"
               $2273 = (($2272) - ($2271))|0; //@line 1257 "./libraries/stb/stb_image.h"
               $2274 = ($2264|0)>($2273|0); //@line 1258 "./libraries/stb/stb_image.h"
               $2275 = $2271; //@line 1258 "./libraries/stb/stb_image.h"
               if (!($2274)) {
                $2282 = $2275;
                break;
               }
               HEAP32[$0>>2] = $2270; //@line 1259 "./libraries/stb/stb_image.h"
               $2276 = ((($s)) + 20|0); //@line 1260 "./libraries/stb/stb_image.h"
               $2277 = HEAP32[$2276>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
               $2278 = ((($s)) + 28|0); //@line 1260 "./libraries/stb/stb_image.h"
               $2279 = HEAP32[$2278>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
               $2280 = (($2264) - ($2273))|0; //@line 1260 "./libraries/stb/stb_image.h"
               FUNCTION_TABLE_vii[$2277 & 8191]($2279,$2280); //@line 1260 "./libraries/stb/stb_image.h"
               break L1078;
              }
             } while(0);
             $2281 = (($2282) + ($2264)|0); //@line 1264 "./libraries/stb/stb_image.h"
             HEAP32[$0>>2] = $2281; //@line 1264 "./libraries/stb/stb_image.h"
            }
           } while(0);
           $2283 = ($2150|0)==(4); //@line 4706 "./libraries/stb/stb_image.h"
           do {
            if ($2283) {
             $2284 = HEAP32[$s>>2]|0; //@line 4706 "./libraries/stb/stb_image.h"
             $2285 = (($2284) + 1)|0; //@line 4706 "./libraries/stb/stb_image.h"
             $2286 = $2285 >>> 1; //@line 4706 "./libraries/stb/stb_image.h"
             $3783 = $2284;$width$0$i = $2286;
            } else {
             $2287 = ($2150|0)==(8); //@line 4707 "./libraries/stb/stb_image.h"
             if ($2287) {
              $2288 = HEAP32[$s>>2]|0; //@line 4707 "./libraries/stb/stb_image.h"
              $3783 = $2288;$width$0$i = $2288;
              break;
             } else {
              _free($2192); //@line 4708 "./libraries/stb/stb_image.h"
              HEAP32[158754] = 441451; //@line 894 "./libraries/stb/stb_image.h"
              $$2$i = 0;
              break L866;
             }
            }
           } while(0);
           $2289 = (0 - ($width$0$i))|0; //@line 4709 "./libraries/stb/stb_image.h"
           $2290 = $2289 & 3; //@line 4709 "./libraries/stb/stb_image.h"
           $2291 = HEAP32[$2152>>2]|0; //@line 4710 "./libraries/stb/stb_image.h"
           $2292 = ($2291|0)>(0); //@line 4710 "./libraries/stb/stb_image.h"
           if (!($2292)) {
            $2791 = $2291;$all_a$7$i = $all_a$1$i;
            break;
           }
           $2293 = ((($s)) + 16|0); //@line 1256 "./libraries/stb/stb_image.h"
           $2294 = ((($s)) + 20|0); //@line 1260 "./libraries/stb/stb_image.h"
           $2295 = ((($s)) + 28|0); //@line 1260 "./libraries/stb/stb_image.h"
           $2296 = ($req_comp$$i|0)==(4); //@line 4720 "./libraries/stb/stb_image.h"
           $2297 = ($2150|0)==(8); //@line 4722 "./libraries/stb/stb_image.h"
           $2298 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $2299 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2300 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2301 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
           $2303 = $3783;$j$0313$i = 0;$z$0314$i = 0;
           while(1) {
            $2302 = ($2303|0)>(0); //@line 4711 "./libraries/stb/stb_image.h"
            L1097: do {
             if ($2302) {
              $$pre351$i = HEAP32[$0>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
              $$pre352$i = HEAP32[$2>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
              if ($2296) {
               $2305 = $$pre351$i;$2306 = $$pre352$i;$3784 = $2303;$i$2307$i$us = 0;$z$1308$i$us = $z$0314$i;
               while(1) {
                $2304 = ($2305>>>0)<($2306>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                do {
                 if ($2304) {
                  $2315 = ((($2305)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                  HEAP32[$0>>2] = $2315; //@line 1230 "./libraries/stb/stb_image.h"
                  $2316 = HEAP8[$2305>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                  $$0$i153$i$us = $2316;$2335 = $3784;$2338 = $2315;$2339 = $2306;
                 } else {
                  $2307 = HEAP32[$2298>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                  $2308 = ($2307|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                  if ($2308) {
                   $$0$i153$i$us = 0;$2335 = $3784;$2338 = $2305;$2339 = $2306;
                   break;
                  }
                  $2309 = HEAP32[$2293>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $2310 = HEAP32[$2295>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $2311 = HEAP32[$2300>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $2312 = (FUNCTION_TABLE_iiii[$2309 & 8191]($2310,$2299,$2311)|0); //@line 1213 "./libraries/stb/stb_image.h"
                  $2313 = ($2312|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                  if ($2313) {
                   HEAP32[$2298>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                   HEAP8[$2299>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                   $$sink$i$i151$i$us = $2301;$3786 = 0;
                  } else {
                   $2314 = (((($s)) + 40|0) + ($2312)|0); //@line 1223 "./libraries/stb/stb_image.h"
                   $$pre$i148$i$us = HEAP8[$2299>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                   $$sink$i$i151$i$us = $2314;$3786 = $$pre$i148$i$us;
                  }
                  HEAP32[$2>>2] = $$sink$i$i151$i$us; //@line 1223 "./libraries/stb/stb_image.h"
                  HEAP32[$0>>2] = $2301; //@line 1233 "./libraries/stb/stb_image.h"
                  $$pre282 = HEAP32[$s>>2]|0; //@line 4721 "./libraries/stb/stb_image.h"
                  $$0$i153$i$us = $3786;$2335 = $$pre282;$2338 = $2301;$2339 = $$sink$i$i151$i$us;
                 }
                } while(0);
                $2317 = $$0$i153$i$us&255; //@line 4712 "./libraries/stb/stb_image.h"
                $2318 = $2317 >>> 4; //@line 4715 "./libraries/stb/stb_image.h"
                $v$0$i$us = $2283 ? $2318 : $2317; //@line 4713 "./libraries/stb/stb_image.h"
                $2319 = (($pal$i) + ($v$0$i$us<<2)|0); //@line 4717 "./libraries/stb/stb_image.h"
                $2320 = HEAP8[$2319>>0]|0; //@line 4717 "./libraries/stb/stb_image.h"
                $2321 = (($z$1308$i$us) + 1)|0; //@line 4717 "./libraries/stb/stb_image.h"
                $2322 = (($2192) + ($z$1308$i$us)|0); //@line 4717 "./libraries/stb/stb_image.h"
                HEAP8[$2322>>0] = $2320; //@line 4717 "./libraries/stb/stb_image.h"
                $2323 = (((($pal$i) + ($v$0$i$us<<2)|0)) + 1|0); //@line 4718 "./libraries/stb/stb_image.h"
                $2324 = HEAP8[$2323>>0]|0; //@line 4718 "./libraries/stb/stb_image.h"
                $2325 = (($z$1308$i$us) + 2)|0; //@line 4718 "./libraries/stb/stb_image.h"
                $2326 = (($2192) + ($2321)|0); //@line 4718 "./libraries/stb/stb_image.h"
                HEAP8[$2326>>0] = $2324; //@line 4718 "./libraries/stb/stb_image.h"
                $2327 = (((($pal$i) + ($v$0$i$us<<2)|0)) + 2|0); //@line 4719 "./libraries/stb/stb_image.h"
                $2328 = HEAP8[$2327>>0]|0; //@line 4719 "./libraries/stb/stb_image.h"
                $2329 = (($z$1308$i$us) + 3)|0; //@line 4719 "./libraries/stb/stb_image.h"
                $2330 = (($2192) + ($2325)|0); //@line 4719 "./libraries/stb/stb_image.h"
                HEAP8[$2330>>0] = $2328; //@line 4719 "./libraries/stb/stb_image.h"
                $2331 = (($z$1308$i$us) + 4)|0; //@line 4720 "./libraries/stb/stb_image.h"
                $2332 = (($2192) + ($2329)|0); //@line 4720 "./libraries/stb/stb_image.h"
                HEAP8[$2332>>0] = -1; //@line 4720 "./libraries/stb/stb_image.h"
                $2333 = $i$2307$i$us | 1; //@line 4721 "./libraries/stb/stb_image.h"
                $2334 = ($2333|0)==($2335|0); //@line 4721 "./libraries/stb/stb_image.h"
                if ($2334) {
                 $z$5$i = $2331;
                 break L1097;
                }
                $2336 = $2317 & 15; //@line 4714 "./libraries/stb/stb_image.h"
                $v2$0$i$us = $2283 ? $2336 : 0; //@line 4713 "./libraries/stb/stb_image.h"
                if ($2297) {
                 $2337 = ($2338>>>0)<($2339>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                 do {
                  if ($2337) {
                   $2348 = ((($2338)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                   HEAP32[$0>>2] = $2348; //@line 1230 "./libraries/stb/stb_image.h"
                   $2349 = HEAP8[$2338>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                   $$0$i160$i$us = $2349;$$pre283 = $2335;$3789 = $2339;$3790 = $2348;
                  } else {
                   $2340 = HEAP32[$2298>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                   $2341 = ($2340|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                   if ($2341) {
                    $$0$i160$i$us = 0;$$pre283 = $2335;$3789 = $2339;$3790 = $2338;
                    break;
                   }
                   $2342 = HEAP32[$2293>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                   $2343 = HEAP32[$2295>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                   $2344 = HEAP32[$2300>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                   $2345 = (FUNCTION_TABLE_iiii[$2342 & 8191]($2343,$2299,$2344)|0); //@line 1213 "./libraries/stb/stb_image.h"
                   $2346 = ($2345|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                   if ($2346) {
                    HEAP32[$2298>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                    HEAP8[$2299>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                    $$sink$i$i158$i$us = $2301;$3791 = 0;
                   } else {
                    $2347 = (((($s)) + 40|0) + ($2345)|0); //@line 1223 "./libraries/stb/stb_image.h"
                    $$pre$i155$i$us = HEAP8[$2299>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                    $$sink$i$i158$i$us = $2347;$3791 = $$pre$i155$i$us;
                   }
                   HEAP32[$2>>2] = $$sink$i$i158$i$us; //@line 1223 "./libraries/stb/stb_image.h"
                   HEAP32[$0>>2] = $2301; //@line 1233 "./libraries/stb/stb_image.h"
                   $$pre283$pre = HEAP32[$s>>2]|0; //@line 4711 "./libraries/stb/stb_image.h"
                   $$0$i160$i$us = $3791;$$pre283 = $$pre283$pre;$3789 = $$sink$i$i158$i$us;$3790 = $2301;
                  }
                 } while(0);
                 $2350 = $$0$i160$i$us&255; //@line 4722 "./libraries/stb/stb_image.h"
                 $2352 = $2350;$2368 = $$pre283;$3787 = $3789;$3788 = $3790;
                } else {
                 $2352 = $v2$0$i$us;$2368 = $2335;$3787 = $2339;$3788 = $2338;
                }
                $2351 = (($pal$i) + ($2352<<2)|0); //@line 4723 "./libraries/stb/stb_image.h"
                $2353 = HEAP8[$2351>>0]|0; //@line 4723 "./libraries/stb/stb_image.h"
                $2354 = (($z$1308$i$us) + 5)|0; //@line 4723 "./libraries/stb/stb_image.h"
                $2355 = (($2192) + ($2331)|0); //@line 4723 "./libraries/stb/stb_image.h"
                HEAP8[$2355>>0] = $2353; //@line 4723 "./libraries/stb/stb_image.h"
                $2356 = (((($pal$i) + ($2352<<2)|0)) + 1|0); //@line 4724 "./libraries/stb/stb_image.h"
                $2357 = HEAP8[$2356>>0]|0; //@line 4724 "./libraries/stb/stb_image.h"
                $2358 = (($z$1308$i$us) + 6)|0; //@line 4724 "./libraries/stb/stb_image.h"
                $2359 = (($2192) + ($2354)|0); //@line 4724 "./libraries/stb/stb_image.h"
                HEAP8[$2359>>0] = $2357; //@line 4724 "./libraries/stb/stb_image.h"
                $2360 = (((($pal$i) + ($2352<<2)|0)) + 2|0); //@line 4725 "./libraries/stb/stb_image.h"
                $2361 = HEAP8[$2360>>0]|0; //@line 4725 "./libraries/stb/stb_image.h"
                $2362 = (($z$1308$i$us) + 7)|0; //@line 4725 "./libraries/stb/stb_image.h"
                $2363 = (($2192) + ($2358)|0); //@line 4725 "./libraries/stb/stb_image.h"
                HEAP8[$2363>>0] = $2361; //@line 4725 "./libraries/stb/stb_image.h"
                $2364 = (($z$1308$i$us) + 8)|0; //@line 4726 "./libraries/stb/stb_image.h"
                $2365 = (($2192) + ($2362)|0); //@line 4726 "./libraries/stb/stb_image.h"
                HEAP8[$2365>>0] = -1; //@line 4726 "./libraries/stb/stb_image.h"
                $2366 = (($i$2307$i$us) + 2)|0; //@line 4711 "./libraries/stb/stb_image.h"
                $2367 = ($2366|0)<($2368|0); //@line 4711 "./libraries/stb/stb_image.h"
                if ($2367) {
                 $2305 = $3788;$2306 = $3787;$3784 = $2368;$i$2307$i$us = $2366;$z$1308$i$us = $2364;
                } else {
                 $z$5$i = $2364;
                 break;
                }
               }
              } else {
               $2370 = $$pre351$i;$2371 = $$pre352$i;$3785 = $2303;$i$2307$i = 0;$z$1308$i = $z$0314$i;
               while(1) {
                $2369 = ($2370>>>0)<($2371>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                do {
                 if ($2369) {
                  $2372 = ((($2370)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                  HEAP32[$0>>2] = $2372; //@line 1230 "./libraries/stb/stb_image.h"
                  $2373 = HEAP8[$2370>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                  $$0$i153$i = $2373;$2398 = $3785;$2401 = $2372;$2402 = $2371;
                 } else {
                  $2374 = HEAP32[$2298>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                  $2375 = ($2374|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                  if ($2375) {
                   $$0$i153$i = 0;$2398 = $3785;$2401 = $2370;$2402 = $2371;
                   break;
                  }
                  $2376 = HEAP32[$2293>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $2377 = HEAP32[$2295>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $2378 = HEAP32[$2300>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $2379 = (FUNCTION_TABLE_iiii[$2376 & 8191]($2377,$2299,$2378)|0); //@line 1213 "./libraries/stb/stb_image.h"
                  $2380 = ($2379|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                  if ($2380) {
                   HEAP32[$2298>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                   HEAP8[$2299>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                   $$sink$i$i151$i = $2301;$3792 = 0;
                  } else {
                   $2381 = (((($s)) + 40|0) + ($2379)|0); //@line 1223 "./libraries/stb/stb_image.h"
                   $$pre$i148$i = HEAP8[$2299>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                   $$sink$i$i151$i = $2381;$3792 = $$pre$i148$i;
                  }
                  HEAP32[$2>>2] = $$sink$i$i151$i; //@line 1223 "./libraries/stb/stb_image.h"
                  HEAP32[$0>>2] = $2301; //@line 1233 "./libraries/stb/stb_image.h"
                  $$pre = HEAP32[$s>>2]|0; //@line 4721 "./libraries/stb/stb_image.h"
                  $$0$i153$i = $3792;$2398 = $$pre;$2401 = $2301;$2402 = $$sink$i$i151$i;
                 }
                } while(0);
                $2382 = $$0$i153$i&255; //@line 4712 "./libraries/stb/stb_image.h"
                $2383 = $2382 >>> 4; //@line 4715 "./libraries/stb/stb_image.h"
                $v$0$i = $2283 ? $2383 : $2382; //@line 4713 "./libraries/stb/stb_image.h"
                $2384 = (($pal$i) + ($v$0$i<<2)|0); //@line 4717 "./libraries/stb/stb_image.h"
                $2385 = HEAP8[$2384>>0]|0; //@line 4717 "./libraries/stb/stb_image.h"
                $2386 = (($z$1308$i) + 1)|0; //@line 4717 "./libraries/stb/stb_image.h"
                $2387 = (($2192) + ($z$1308$i)|0); //@line 4717 "./libraries/stb/stb_image.h"
                HEAP8[$2387>>0] = $2385; //@line 4717 "./libraries/stb/stb_image.h"
                $2388 = (((($pal$i) + ($v$0$i<<2)|0)) + 1|0); //@line 4718 "./libraries/stb/stb_image.h"
                $2389 = HEAP8[$2388>>0]|0; //@line 4718 "./libraries/stb/stb_image.h"
                $2390 = (($z$1308$i) + 2)|0; //@line 4718 "./libraries/stb/stb_image.h"
                $2391 = (($2192) + ($2386)|0); //@line 4718 "./libraries/stb/stb_image.h"
                HEAP8[$2391>>0] = $2389; //@line 4718 "./libraries/stb/stb_image.h"
                $2392 = (((($pal$i) + ($v$0$i<<2)|0)) + 2|0); //@line 4719 "./libraries/stb/stb_image.h"
                $2393 = HEAP8[$2392>>0]|0; //@line 4719 "./libraries/stb/stb_image.h"
                $2394 = (($z$1308$i) + 3)|0; //@line 4719 "./libraries/stb/stb_image.h"
                $2395 = (($2192) + ($2390)|0); //@line 4719 "./libraries/stb/stb_image.h"
                HEAP8[$2395>>0] = $2393; //@line 4719 "./libraries/stb/stb_image.h"
                $2396 = $i$2307$i | 1; //@line 4721 "./libraries/stb/stb_image.h"
                $2397 = ($2396|0)==($2398|0); //@line 4721 "./libraries/stb/stb_image.h"
                if ($2397) {
                 $z$5$i = $2394;
                 break L1097;
                }
                $2399 = $2382 & 15; //@line 4714 "./libraries/stb/stb_image.h"
                $v2$0$i = $2283 ? $2399 : 0; //@line 4713 "./libraries/stb/stb_image.h"
                if ($2297) {
                 $2400 = ($2401>>>0)<($2402>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                 do {
                  if ($2400) {
                   $2403 = ((($2401)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                   HEAP32[$0>>2] = $2403; //@line 1230 "./libraries/stb/stb_image.h"
                   $2404 = HEAP8[$2401>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                   $$0$i160$i = $2404;$$pre281 = $2398;$3795 = $2402;$3796 = $2403;
                  } else {
                   $2405 = HEAP32[$2298>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                   $2406 = ($2405|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                   if ($2406) {
                    $$0$i160$i = 0;$$pre281 = $2398;$3795 = $2402;$3796 = $2401;
                    break;
                   }
                   $2407 = HEAP32[$2293>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                   $2408 = HEAP32[$2295>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                   $2409 = HEAP32[$2300>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                   $2410 = (FUNCTION_TABLE_iiii[$2407 & 8191]($2408,$2299,$2409)|0); //@line 1213 "./libraries/stb/stb_image.h"
                   $2411 = ($2410|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                   if ($2411) {
                    HEAP32[$2298>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                    HEAP8[$2299>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                    $$sink$i$i158$i = $2301;$3797 = 0;
                   } else {
                    $2412 = (((($s)) + 40|0) + ($2410)|0); //@line 1223 "./libraries/stb/stb_image.h"
                    $$pre$i155$i = HEAP8[$2299>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                    $$sink$i$i158$i = $2412;$3797 = $$pre$i155$i;
                   }
                   HEAP32[$2>>2] = $$sink$i$i158$i; //@line 1223 "./libraries/stb/stb_image.h"
                   HEAP32[$0>>2] = $2301; //@line 1233 "./libraries/stb/stb_image.h"
                   $$pre281$pre = HEAP32[$s>>2]|0; //@line 4711 "./libraries/stb/stb_image.h"
                   $$0$i160$i = $3797;$$pre281 = $$pre281$pre;$3795 = $$sink$i$i158$i;$3796 = $2301;
                  }
                 } while(0);
                 $2413 = $$0$i160$i&255; //@line 4722 "./libraries/stb/stb_image.h"
                 $2415 = $2413;$2429 = $$pre281;$3793 = $3795;$3794 = $3796;
                } else {
                 $2415 = $v2$0$i;$2429 = $2398;$3793 = $2402;$3794 = $2401;
                }
                $2414 = (($pal$i) + ($2415<<2)|0); //@line 4723 "./libraries/stb/stb_image.h"
                $2416 = HEAP8[$2414>>0]|0; //@line 4723 "./libraries/stb/stb_image.h"
                $2417 = (($z$1308$i) + 4)|0; //@line 4723 "./libraries/stb/stb_image.h"
                $2418 = (($2192) + ($2394)|0); //@line 4723 "./libraries/stb/stb_image.h"
                HEAP8[$2418>>0] = $2416; //@line 4723 "./libraries/stb/stb_image.h"
                $2419 = (((($pal$i) + ($2415<<2)|0)) + 1|0); //@line 4724 "./libraries/stb/stb_image.h"
                $2420 = HEAP8[$2419>>0]|0; //@line 4724 "./libraries/stb/stb_image.h"
                $2421 = (($z$1308$i) + 5)|0; //@line 4724 "./libraries/stb/stb_image.h"
                $2422 = (($2192) + ($2417)|0); //@line 4724 "./libraries/stb/stb_image.h"
                HEAP8[$2422>>0] = $2420; //@line 4724 "./libraries/stb/stb_image.h"
                $2423 = (((($pal$i) + ($2415<<2)|0)) + 2|0); //@line 4725 "./libraries/stb/stb_image.h"
                $2424 = HEAP8[$2423>>0]|0; //@line 4725 "./libraries/stb/stb_image.h"
                $2425 = (($z$1308$i) + 6)|0; //@line 4725 "./libraries/stb/stb_image.h"
                $2426 = (($2192) + ($2421)|0); //@line 4725 "./libraries/stb/stb_image.h"
                HEAP8[$2426>>0] = $2424; //@line 4725 "./libraries/stb/stb_image.h"
                $2427 = (($i$2307$i) + 2)|0; //@line 4711 "./libraries/stb/stb_image.h"
                $2428 = ($2427|0)<($2429|0); //@line 4711 "./libraries/stb/stb_image.h"
                if ($2428) {
                 $2370 = $3794;$2371 = $3793;$3785 = $2429;$i$2307$i = $2427;$z$1308$i = $2425;
                } else {
                 $z$5$i = $2425;
                 break;
                }
               }
              }
             } else {
              $z$5$i = $z$0314$i;
             }
            } while(0);
            $2430 = HEAP32[$2293>>2]|0; //@line 1256 "./libraries/stb/stb_image.h"
            $2431 = ($2430|0)==(0|0); //@line 1256 "./libraries/stb/stb_image.h"
            do {
             if ($2431) {
              $$pre$i163$i = HEAP32[$0>>2]|0; //@line 1264 "./libraries/stb/stb_image.h"
              $2442 = $$pre$i163$i;
              label = 844;
             } else {
              $2432 = HEAP32[$2>>2]|0; //@line 1257 "./libraries/stb/stb_image.h"
              $2433 = HEAP32[$0>>2]|0; //@line 1257 "./libraries/stb/stb_image.h"
              $2434 = $2432; //@line 1257 "./libraries/stb/stb_image.h"
              $2435 = (($2434) - ($2433))|0; //@line 1257 "./libraries/stb/stb_image.h"
              $2436 = ($2290|0)>($2435|0); //@line 1258 "./libraries/stb/stb_image.h"
              $2437 = $2433; //@line 1258 "./libraries/stb/stb_image.h"
              if (!($2436)) {
               $2442 = $2437;
               label = 844;
               break;
              }
              HEAP32[$0>>2] = $2432; //@line 1259 "./libraries/stb/stb_image.h"
              $2438 = HEAP32[$2294>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
              $2439 = HEAP32[$2295>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
              $2440 = (($2290) - ($2435))|0; //@line 1260 "./libraries/stb/stb_image.h"
              FUNCTION_TABLE_vii[$2438 & 8191]($2439,$2440); //@line 1260 "./libraries/stb/stb_image.h"
             }
            } while(0);
            if ((label|0) == 844) {
             label = 0;
             $2441 = (($2442) + ($2290)|0); //@line 1264 "./libraries/stb/stb_image.h"
             HEAP32[$0>>2] = $2441; //@line 1264 "./libraries/stb/stb_image.h"
            }
            $2443 = (($j$0313$i) + 1)|0; //@line 4710 "./libraries/stb/stb_image.h"
            $2444 = HEAP32[$2152>>2]|0; //@line 4710 "./libraries/stb/stb_image.h"
            $2445 = ($2443|0)<($2444|0); //@line 4710 "./libraries/stb/stb_image.h"
            if (!($2445)) {
             $2791 = $2444;$all_a$7$i = $all_a$1$i;
             break L1030;
            }
            $$pre350$i = HEAP32[$s>>2]|0; //@line 4711 "./libraries/stb/stb_image.h"
            $2303 = $$pre350$i;$j$0313$i = $2443;$z$0314$i = $z$5$i;
           }
          } else {
           $2446 = (($1975) + -14)|0; //@line 4734 "./libraries/stb/stb_image.h"
           $2447 = (($2446) - ($1976))|0; //@line 4734 "./libraries/stb/stb_image.h"
           $2448 = ($2447|0)<(0); //@line 1252 "./libraries/stb/stb_image.h"
           L1158: do {
            if ($2448) {
             $2449 = HEAP32[$2>>2]|0; //@line 1253 "./libraries/stb/stb_image.h"
             HEAP32[$0>>2] = $2449; //@line 1253 "./libraries/stb/stb_image.h"
            } else {
             $2450 = ((($s)) + 16|0); //@line 1256 "./libraries/stb/stb_image.h"
             $2451 = HEAP32[$2450>>2]|0; //@line 1256 "./libraries/stb/stb_image.h"
             $2452 = ($2451|0)==(0|0); //@line 1256 "./libraries/stb/stb_image.h"
             do {
              if ($2452) {
               $$pre$i169$i = HEAP32[$0>>2]|0; //@line 1264 "./libraries/stb/stb_image.h"
               $2465 = $$pre$i169$i;
              } else {
               $2453 = HEAP32[$2>>2]|0; //@line 1257 "./libraries/stb/stb_image.h"
               $2454 = HEAP32[$0>>2]|0; //@line 1257 "./libraries/stb/stb_image.h"
               $2455 = $2453; //@line 1257 "./libraries/stb/stb_image.h"
               $2456 = (($2455) - ($2454))|0; //@line 1257 "./libraries/stb/stb_image.h"
               $2457 = ($2447|0)>($2456|0); //@line 1258 "./libraries/stb/stb_image.h"
               $2458 = $2454; //@line 1258 "./libraries/stb/stb_image.h"
               if (!($2457)) {
                $2465 = $2458;
                break;
               }
               HEAP32[$0>>2] = $2453; //@line 1259 "./libraries/stb/stb_image.h"
               $2459 = ((($s)) + 20|0); //@line 1260 "./libraries/stb/stb_image.h"
               $2460 = HEAP32[$2459>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
               $2461 = ((($s)) + 28|0); //@line 1260 "./libraries/stb/stb_image.h"
               $2462 = HEAP32[$2461>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
               $2463 = (($2447) - ($2456))|0; //@line 1260 "./libraries/stb/stb_image.h"
               FUNCTION_TABLE_vii[$2460 & 8191]($2462,$2463); //@line 1260 "./libraries/stb/stb_image.h"
               break L1158;
              }
             } while(0);
             $2464 = (($2465) + ($2447)|0); //@line 1264 "./libraries/stb/stb_image.h"
             HEAP32[$0>>2] = $2464; //@line 1264 "./libraries/stb/stb_image.h"
            }
           } while(0);
           switch ($2150|0) {
           case 24:  {
            $2470 = HEAP32[$s>>2]|0; //@line 4735 "./libraries/stb/stb_image.h"
            $2471 = Math_imul($2470, -3)|0; //@line 4738 "./libraries/stb/stb_image.h"
            $2472 = $2471 & 3; //@line 4738 "./libraries/stb/stb_image.h"
            $2777 = $2472;$3798 = 1;$acount$0$i = 0;$ashift$0$i = 0;$bcount$0$i = 0;$bshift$0$i = 0;$easy$0287$i = 1;$gcount$0$i = 0;$gshift$0$i = 0;$rcount$0$i = 0;$rshift$0$i = 0;
            break;
           }
           case 16:  {
            $2466 = HEAP32[$s>>2]|0; //@line 4736 "./libraries/stb/stb_image.h"
            $2467 = $2466 << 1; //@line 4736 "./libraries/stb/stb_image.h"
            $2468 = (0 - ($2467))|0; //@line 4738 "./libraries/stb/stb_image.h"
            $2469 = $2468 & 2; //@line 4738 "./libraries/stb/stb_image.h"
            $$ph$i = $2469;
            label = 858;
            break;
           }
           case 32:  {
            $2473 = ($mb$1$i|0)==(255); //@line 4742 "./libraries/stb/stb_image.h"
            $2474 = ($mg$1$i|0)==(65280); //@line 4742 "./libraries/stb/stb_image.h"
            $or$cond20$i = $2473 & $2474; //@line 4742 "./libraries/stb/stb_image.h"
            $2475 = ($mr$1$i|0)==(16711680); //@line 4742 "./libraries/stb/stb_image.h"
            $or$cond22$i = $or$cond20$i & $2475; //@line 4742 "./libraries/stb/stb_image.h"
            $2476 = ($ma$1$i|0)==(-16777216); //@line 4742 "./libraries/stb/stb_image.h"
            $or$cond24$i = $2476 & $or$cond22$i; //@line 4742 "./libraries/stb/stb_image.h"
            if ($or$cond24$i) {
             $2777 = 0;$3798 = 1;$acount$0$i = 0;$ashift$0$i = 0;$bcount$0$i = 0;$bshift$0$i = 0;$easy$0287$i = 2;$gcount$0$i = 0;$gshift$0$i = 0;$rcount$0$i = 0;$rshift$0$i = 0;
            } else {
             $$ph$i = 0;
             label = 858;
            }
            break;
           }
           default: {
            $$ph$i = 0;
            label = 858;
           }
           }
           if ((label|0) == 858) {
            $2477 = ($mr$1$i|0)!=(0); //@line 4746 "./libraries/stb/stb_image.h"
            $2478 = ($mg$1$i|0)!=(0); //@line 4746 "./libraries/stb/stb_image.h"
            $or$cond26$i = $2478 & $2477; //@line 4746 "./libraries/stb/stb_image.h"
            $2479 = ($mb$1$i|0)!=(0); //@line 4746 "./libraries/stb/stb_image.h"
            $or$cond28$i = $2479 & $or$cond26$i; //@line 4746 "./libraries/stb/stb_image.h"
            if (!($or$cond28$i)) {
             _free($2192); //@line 4746 "./libraries/stb/stb_image.h"
             HEAP32[158754] = 441459; //@line 894 "./libraries/stb/stb_image.h"
             $$2$i = 0;
             break L866;
            }
            $2480 = ($mr$1$i>>>0)>(65535); //@line 4564 "./libraries/stb/stb_image.h"
            $2481 = $mr$1$i >>> 16; //@line 4564 "./libraries/stb/stb_image.h"
            $$z$i$i = $2480 ? $2481 : $mr$1$i; //@line 4564 "./libraries/stb/stb_image.h"
            $$$i$i28 = $2480 ? 16 : 0; //@line 4564 "./libraries/stb/stb_image.h"
            $2482 = ($$z$i$i>>>0)>(255); //@line 4565 "./libraries/stb/stb_image.h"
            $2483 = $$$i$i28 | 8; //@line 4565 "./libraries/stb/stb_image.h"
            $2484 = $$z$i$i >>> 8; //@line 4565 "./libraries/stb/stb_image.h"
            $$1$i$i = $2482 ? $2484 : $$z$i$i; //@line 4565 "./libraries/stb/stb_image.h"
            $n$1$i$i = $2482 ? $2483 : $$$i$i28; //@line 4565 "./libraries/stb/stb_image.h"
            $2485 = ($$1$i$i>>>0)>(15); //@line 4566 "./libraries/stb/stb_image.h"
            $2486 = $n$1$i$i | 4; //@line 4566 "./libraries/stb/stb_image.h"
            $2487 = $$1$i$i >>> 4; //@line 4566 "./libraries/stb/stb_image.h"
            $$2$i$i29 = $2485 ? $2487 : $$1$i$i; //@line 4566 "./libraries/stb/stb_image.h"
            $n$2$i$i = $2485 ? $2486 : $n$1$i$i; //@line 4566 "./libraries/stb/stb_image.h"
            $2488 = ($$2$i$i29>>>0)>(3); //@line 4567 "./libraries/stb/stb_image.h"
            $2489 = $n$2$i$i | 2; //@line 4567 "./libraries/stb/stb_image.h"
            $2490 = $$2$i$i29 >>> 2; //@line 4567 "./libraries/stb/stb_image.h"
            $$3$i$i30 = $2488 ? $2490 : $$2$i$i29; //@line 4567 "./libraries/stb/stb_image.h"
            $n$3$i$i = $2488 ? $2489 : $n$2$i$i; //@line 4567 "./libraries/stb/stb_image.h"
            $2491 = ($$3$i$i30>>>0)>(1); //@line 4568 "./libraries/stb/stb_image.h"
            $2492 = $2491&1; //@line 4568 "./libraries/stb/stb_image.h"
            $$n$3$i$i = (($n$3$i$i) + -7)|0; //@line 4568 "./libraries/stb/stb_image.h"
            $2493 = (($$n$3$i$i) + ($2492))|0; //@line 4748 "./libraries/stb/stb_image.h"
            $2494 = $mr$1$i & 1431655765; //@line 4574 "./libraries/stb/stb_image.h"
            $2495 = $mr$1$i >>> 1; //@line 4574 "./libraries/stb/stb_image.h"
            $2496 = $2495 & 1431655765; //@line 4574 "./libraries/stb/stb_image.h"
            $2497 = (($2496) + ($2494))|0; //@line 4574 "./libraries/stb/stb_image.h"
            $2498 = $2497 & 858993459; //@line 4575 "./libraries/stb/stb_image.h"
            $2499 = $2497 >>> 2; //@line 4575 "./libraries/stb/stb_image.h"
            $2500 = $2499 & 858993459; //@line 4575 "./libraries/stb/stb_image.h"
            $2501 = (($2500) + ($2498))|0; //@line 4575 "./libraries/stb/stb_image.h"
            $2502 = $2501 >>> 4; //@line 4576 "./libraries/stb/stb_image.h"
            $2503 = (($2502) + ($2501))|0; //@line 4576 "./libraries/stb/stb_image.h"
            $2504 = $2503 & 252645135; //@line 4576 "./libraries/stb/stb_image.h"
            $2505 = $2504 >>> 8; //@line 4577 "./libraries/stb/stb_image.h"
            $2506 = (($2505) + ($2504))|0; //@line 4577 "./libraries/stb/stb_image.h"
            $2507 = $2506 >>> 16; //@line 4578 "./libraries/stb/stb_image.h"
            $2508 = (($2507) + ($2506))|0; //@line 4578 "./libraries/stb/stb_image.h"
            $2509 = $2508 & 63; //@line 4579 "./libraries/stb/stb_image.h"
            $2510 = ($mg$1$i>>>0)>(65535); //@line 4564 "./libraries/stb/stb_image.h"
            $2511 = $mg$1$i >>> 16; //@line 4564 "./libraries/stb/stb_image.h"
            $$z$i217$i = $2510 ? $2511 : $mg$1$i; //@line 4564 "./libraries/stb/stb_image.h"
            $$$i218$i = $2510 ? 16 : 0; //@line 4564 "./libraries/stb/stb_image.h"
            $2512 = ($$z$i217$i>>>0)>(255); //@line 4565 "./libraries/stb/stb_image.h"
            $2513 = $$$i218$i | 8; //@line 4565 "./libraries/stb/stb_image.h"
            $2514 = $$z$i217$i >>> 8; //@line 4565 "./libraries/stb/stb_image.h"
            $$1$i219$i = $2512 ? $2514 : $$z$i217$i; //@line 4565 "./libraries/stb/stb_image.h"
            $n$1$i220$i = $2512 ? $2513 : $$$i218$i; //@line 4565 "./libraries/stb/stb_image.h"
            $2515 = ($$1$i219$i>>>0)>(15); //@line 4566 "./libraries/stb/stb_image.h"
            $2516 = $n$1$i220$i | 4; //@line 4566 "./libraries/stb/stb_image.h"
            $2517 = $$1$i219$i >>> 4; //@line 4566 "./libraries/stb/stb_image.h"
            $$2$i221$i = $2515 ? $2517 : $$1$i219$i; //@line 4566 "./libraries/stb/stb_image.h"
            $n$2$i222$i = $2515 ? $2516 : $n$1$i220$i; //@line 4566 "./libraries/stb/stb_image.h"
            $2518 = ($$2$i221$i>>>0)>(3); //@line 4567 "./libraries/stb/stb_image.h"
            $2519 = $n$2$i222$i | 2; //@line 4567 "./libraries/stb/stb_image.h"
            $2520 = $$2$i221$i >>> 2; //@line 4567 "./libraries/stb/stb_image.h"
            $$3$i223$i = $2518 ? $2520 : $$2$i221$i; //@line 4567 "./libraries/stb/stb_image.h"
            $n$3$i224$i = $2518 ? $2519 : $n$2$i222$i; //@line 4567 "./libraries/stb/stb_image.h"
            $2521 = ($$3$i223$i>>>0)>(1); //@line 4568 "./libraries/stb/stb_image.h"
            $2522 = $2521&1; //@line 4568 "./libraries/stb/stb_image.h"
            $$n$3$i225$i = (($n$3$i224$i) + -7)|0; //@line 4568 "./libraries/stb/stb_image.h"
            $2523 = (($$n$3$i225$i) + ($2522))|0; //@line 4749 "./libraries/stb/stb_image.h"
            $2524 = $mg$1$i & 1431655765; //@line 4574 "./libraries/stb/stb_image.h"
            $2525 = $mg$1$i >>> 1; //@line 4574 "./libraries/stb/stb_image.h"
            $2526 = $2525 & 1431655765; //@line 4574 "./libraries/stb/stb_image.h"
            $2527 = (($2526) + ($2524))|0; //@line 4574 "./libraries/stb/stb_image.h"
            $2528 = $2527 & 858993459; //@line 4575 "./libraries/stb/stb_image.h"
            $2529 = $2527 >>> 2; //@line 4575 "./libraries/stb/stb_image.h"
            $2530 = $2529 & 858993459; //@line 4575 "./libraries/stb/stb_image.h"
            $2531 = (($2530) + ($2528))|0; //@line 4575 "./libraries/stb/stb_image.h"
            $2532 = $2531 >>> 4; //@line 4576 "./libraries/stb/stb_image.h"
            $2533 = (($2532) + ($2531))|0; //@line 4576 "./libraries/stb/stb_image.h"
            $2534 = $2533 & 252645135; //@line 4576 "./libraries/stb/stb_image.h"
            $2535 = $2534 >>> 8; //@line 4577 "./libraries/stb/stb_image.h"
            $2536 = (($2535) + ($2534))|0; //@line 4577 "./libraries/stb/stb_image.h"
            $2537 = $2536 >>> 16; //@line 4578 "./libraries/stb/stb_image.h"
            $2538 = (($2537) + ($2536))|0; //@line 4578 "./libraries/stb/stb_image.h"
            $2539 = $2538 & 63; //@line 4579 "./libraries/stb/stb_image.h"
            $2540 = ($mb$1$i>>>0)>(65535); //@line 4564 "./libraries/stb/stb_image.h"
            $2541 = $mb$1$i >>> 16; //@line 4564 "./libraries/stb/stb_image.h"
            $$z$i235$i = $2540 ? $2541 : $mb$1$i; //@line 4564 "./libraries/stb/stb_image.h"
            $$$i236$i = $2540 ? 16 : 0; //@line 4564 "./libraries/stb/stb_image.h"
            $2542 = ($$z$i235$i>>>0)>(255); //@line 4565 "./libraries/stb/stb_image.h"
            $2543 = $$$i236$i | 8; //@line 4565 "./libraries/stb/stb_image.h"
            $2544 = $$z$i235$i >>> 8; //@line 4565 "./libraries/stb/stb_image.h"
            $$1$i237$i = $2542 ? $2544 : $$z$i235$i; //@line 4565 "./libraries/stb/stb_image.h"
            $n$1$i238$i = $2542 ? $2543 : $$$i236$i; //@line 4565 "./libraries/stb/stb_image.h"
            $2545 = ($$1$i237$i>>>0)>(15); //@line 4566 "./libraries/stb/stb_image.h"
            $2546 = $n$1$i238$i | 4; //@line 4566 "./libraries/stb/stb_image.h"
            $2547 = $$1$i237$i >>> 4; //@line 4566 "./libraries/stb/stb_image.h"
            $$2$i239$i = $2545 ? $2547 : $$1$i237$i; //@line 4566 "./libraries/stb/stb_image.h"
            $n$2$i240$i = $2545 ? $2546 : $n$1$i238$i; //@line 4566 "./libraries/stb/stb_image.h"
            $2548 = ($$2$i239$i>>>0)>(3); //@line 4567 "./libraries/stb/stb_image.h"
            $2549 = $n$2$i240$i | 2; //@line 4567 "./libraries/stb/stb_image.h"
            $2550 = $$2$i239$i >>> 2; //@line 4567 "./libraries/stb/stb_image.h"
            $$3$i241$i = $2548 ? $2550 : $$2$i239$i; //@line 4567 "./libraries/stb/stb_image.h"
            $n$3$i242$i = $2548 ? $2549 : $n$2$i240$i; //@line 4567 "./libraries/stb/stb_image.h"
            $2551 = ($$3$i241$i>>>0)>(1); //@line 4568 "./libraries/stb/stb_image.h"
            $2552 = $2551&1; //@line 4568 "./libraries/stb/stb_image.h"
            $$n$3$i243$i = (($n$3$i242$i) + -7)|0; //@line 4568 "./libraries/stb/stb_image.h"
            $2553 = (($$n$3$i243$i) + ($2552))|0; //@line 4750 "./libraries/stb/stb_image.h"
            $2554 = $mb$1$i & 1431655765; //@line 4574 "./libraries/stb/stb_image.h"
            $2555 = $mb$1$i >>> 1; //@line 4574 "./libraries/stb/stb_image.h"
            $2556 = $2555 & 1431655765; //@line 4574 "./libraries/stb/stb_image.h"
            $2557 = (($2556) + ($2554))|0; //@line 4574 "./libraries/stb/stb_image.h"
            $2558 = $2557 & 858993459; //@line 4575 "./libraries/stb/stb_image.h"
            $2559 = $2557 >>> 2; //@line 4575 "./libraries/stb/stb_image.h"
            $2560 = $2559 & 858993459; //@line 4575 "./libraries/stb/stb_image.h"
            $2561 = (($2560) + ($2558))|0; //@line 4575 "./libraries/stb/stb_image.h"
            $2562 = $2561 >>> 4; //@line 4576 "./libraries/stb/stb_image.h"
            $2563 = (($2562) + ($2561))|0; //@line 4576 "./libraries/stb/stb_image.h"
            $2564 = $2563 & 252645135; //@line 4576 "./libraries/stb/stb_image.h"
            $2565 = $2564 >>> 8; //@line 4577 "./libraries/stb/stb_image.h"
            $2566 = (($2565) + ($2564))|0; //@line 4577 "./libraries/stb/stb_image.h"
            $2567 = $2566 >>> 16; //@line 4578 "./libraries/stb/stb_image.h"
            $2568 = (($2567) + ($2566))|0; //@line 4578 "./libraries/stb/stb_image.h"
            $2569 = $2568 & 63; //@line 4579 "./libraries/stb/stb_image.h"
            $2570 = ($ma$1$i|0)==(0); //@line 4563 "./libraries/stb/stb_image.h"
            if ($2570) {
             $$01$i269$i = -8;
            } else {
             $2571 = ($ma$1$i>>>0)>(65535); //@line 4564 "./libraries/stb/stb_image.h"
             $2572 = $ma$1$i >>> 16; //@line 4564 "./libraries/stb/stb_image.h"
             $$z$i260$i = $2571 ? $2572 : $ma$1$i; //@line 4564 "./libraries/stb/stb_image.h"
             $$$i261$i = $2571 ? 16 : 0; //@line 4564 "./libraries/stb/stb_image.h"
             $2573 = ($$z$i260$i>>>0)>(255); //@line 4565 "./libraries/stb/stb_image.h"
             $2574 = $$$i261$i | 8; //@line 4565 "./libraries/stb/stb_image.h"
             $2575 = $$z$i260$i >>> 8; //@line 4565 "./libraries/stb/stb_image.h"
             $$1$i262$i = $2573 ? $2575 : $$z$i260$i; //@line 4565 "./libraries/stb/stb_image.h"
             $n$1$i263$i = $2573 ? $2574 : $$$i261$i; //@line 4565 "./libraries/stb/stb_image.h"
             $2576 = ($$1$i262$i>>>0)>(15); //@line 4566 "./libraries/stb/stb_image.h"
             $2577 = $n$1$i263$i | 4; //@line 4566 "./libraries/stb/stb_image.h"
             $2578 = $$1$i262$i >>> 4; //@line 4566 "./libraries/stb/stb_image.h"
             $$2$i264$i = $2576 ? $2578 : $$1$i262$i; //@line 4566 "./libraries/stb/stb_image.h"
             $n$2$i265$i = $2576 ? $2577 : $n$1$i263$i; //@line 4566 "./libraries/stb/stb_image.h"
             $2579 = ($$2$i264$i>>>0)>(3); //@line 4567 "./libraries/stb/stb_image.h"
             $2580 = $n$2$i265$i | 2; //@line 4567 "./libraries/stb/stb_image.h"
             $2581 = $$2$i264$i >>> 2; //@line 4567 "./libraries/stb/stb_image.h"
             $$3$i266$i = $2579 ? $2581 : $$2$i264$i; //@line 4567 "./libraries/stb/stb_image.h"
             $n$3$i267$i = $2579 ? $2580 : $n$2$i265$i; //@line 4567 "./libraries/stb/stb_image.h"
             $2582 = ($$3$i266$i>>>0)>(1); //@line 4568 "./libraries/stb/stb_image.h"
             $2583 = $2582&1; //@line 4568 "./libraries/stb/stb_image.h"
             $$n$3$i268$i = (($2583) + ($n$3$i267$i))|0; //@line 4568 "./libraries/stb/stb_image.h"
             $phitmp = (($$n$3$i268$i) + -7)|0; //@line 4569 "./libraries/stb/stb_image.h"
             $$01$i269$i = $phitmp;
            }
            $2584 = $ma$1$i & 1431655765; //@line 4574 "./libraries/stb/stb_image.h"
            $2585 = $ma$1$i >>> 1; //@line 4574 "./libraries/stb/stb_image.h"
            $2586 = $2585 & 1431655765; //@line 4574 "./libraries/stb/stb_image.h"
            $2587 = (($2586) + ($2584))|0; //@line 4574 "./libraries/stb/stb_image.h"
            $2588 = $2587 & 858993459; //@line 4575 "./libraries/stb/stb_image.h"
            $2589 = $2587 >>> 2; //@line 4575 "./libraries/stb/stb_image.h"
            $2590 = $2589 & 858993459; //@line 4575 "./libraries/stb/stb_image.h"
            $2591 = (($2590) + ($2588))|0; //@line 4575 "./libraries/stb/stb_image.h"
            $2592 = $2591 >>> 4; //@line 4576 "./libraries/stb/stb_image.h"
            $2593 = (($2592) + ($2591))|0; //@line 4576 "./libraries/stb/stb_image.h"
            $2594 = $2593 & 252645135; //@line 4576 "./libraries/stb/stb_image.h"
            $2595 = $2594 >>> 8; //@line 4577 "./libraries/stb/stb_image.h"
            $2596 = (($2595) + ($2594))|0; //@line 4577 "./libraries/stb/stb_image.h"
            $2597 = $2596 >>> 16; //@line 4578 "./libraries/stb/stb_image.h"
            $2598 = (($2597) + ($2596))|0; //@line 4578 "./libraries/stb/stb_image.h"
            $2599 = $2598 & 63; //@line 4579 "./libraries/stb/stb_image.h"
            $2777 = $$ph$i;$3798 = 0;$acount$0$i = $2599;$ashift$0$i = $$01$i269$i;$bcount$0$i = $2569;$bshift$0$i = $2553;$easy$0287$i = 0;$gcount$0$i = $2539;$gshift$0$i = $2523;$rcount$0$i = $2509;$rshift$0$i = $2493;
           }
           $2600 = HEAP32[$2152>>2]|0; //@line 4753 "./libraries/stb/stb_image.h"
           $2601 = ($2600|0)>(0); //@line 4753 "./libraries/stb/stb_image.h"
           if (!($2601)) {
            $2791 = $2600;$all_a$7$i = $all_a$1$i;
            break;
           }
           $2602 = ((($s)) + 16|0); //@line 1256 "./libraries/stb/stb_image.h"
           $2603 = ((($s)) + 20|0); //@line 1260 "./libraries/stb/stb_image.h"
           $2604 = ((($s)) + 28|0); //@line 1260 "./libraries/stb/stb_image.h"
           $2605 = ($easy$0287$i|0)==(2); //@line 4761 "./libraries/stb/stb_image.h"
           $2606 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $2607 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2608 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $2609 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
           $2610 = ($req_comp$$i|0)==(4); //@line 4763 "./libraries/stb/stb_image.h"
           $2611 = ($2150|0)==(16); //@line 4767 "./libraries/stb/stb_image.h"
           $2612 = ($rshift$0$i|0)<(0); //@line 4587 "./libraries/stb/stb_image.h"
           $2613 = (0 - ($rshift$0$i))|0; //@line 4587 "./libraries/stb/stb_image.h"
           $2614 = ($rcount$0$i>>>0)<(8); //@line 4592 "./libraries/stb/stb_image.h"
           $2615 = ($gshift$0$i|0)<(0); //@line 4587 "./libraries/stb/stb_image.h"
           $2616 = (0 - ($gshift$0$i))|0; //@line 4587 "./libraries/stb/stb_image.h"
           $2617 = ($gcount$0$i>>>0)<(8); //@line 4592 "./libraries/stb/stb_image.h"
           $2618 = ($bshift$0$i|0)<(0); //@line 4587 "./libraries/stb/stb_image.h"
           $2619 = (0 - ($bshift$0$i))|0; //@line 4587 "./libraries/stb/stb_image.h"
           $2620 = ($bcount$0$i>>>0)<(8); //@line 4592 "./libraries/stb/stb_image.h"
           $2621 = ($ashift$0$i|0)<(0); //@line 4587 "./libraries/stb/stb_image.h"
           $2622 = (0 - ($ashift$0$i))|0; //@line 4587 "./libraries/stb/stb_image.h"
           $2623 = ($acount$0$i>>>0)<(8); //@line 4592 "./libraries/stb/stb_image.h"
           $all_a$2331$i = $all_a$1$i;$j$1330$i = 0;$z2$0332$i = 0;
           while(1) {
            $2624 = HEAP32[$s>>2]|0; //@line 4755 "./libraries/stb/stb_image.h"
            $2625 = ($2624|0)>(0); //@line 4755 "./libraries/stb/stb_image.h"
            do {
             if ($3798) {
              if (!($2625)) {
               $all_a$5$i = $all_a$2331$i;$z2$5$i = $z2$0332$i;
               break;
              }
              $$pre346$i = HEAP32[$0>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
              $$pre347$i = HEAP32[$2>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
              $2627 = $$pre346$i;$2628 = $$pre347$i;$all_a$3325$i = $all_a$2331$i;$i$3324$i = 0;$z2$1326$i = $z2$0332$i;
              while(1) {
               $2626 = ($2627>>>0)<($2628>>>0); //@line 1229 "./libraries/stb/stb_image.h"
               do {
                if ($2626) {
                 $2629 = ((($2627)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                 HEAP32[$0>>2] = $2629; //@line 1230 "./libraries/stb/stb_image.h"
                 $2630 = HEAP8[$2627>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                 $$0$i276$i = $2630;$2642 = $2629;$2643 = $2628;
                } else {
                 $2631 = HEAP32[$2606>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                 $2632 = ($2631|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                 if ($2632) {
                  $$0$i276$i = 0;$2642 = $2627;$2643 = $2628;
                  break;
                 }
                 $2633 = HEAP32[$2602>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                 $2634 = HEAP32[$2604>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                 $2635 = HEAP32[$2608>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                 $2636 = (FUNCTION_TABLE_iiii[$2633 & 8191]($2634,$2607,$2635)|0); //@line 1213 "./libraries/stb/stb_image.h"
                 $2637 = ($2636|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                 if ($2637) {
                  HEAP32[$2606>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                  HEAP8[$2607>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                  $$sink$i$i274$i = $2609;$3799 = 0;
                 } else {
                  $2638 = (((($s)) + 40|0) + ($2636)|0); //@line 1223 "./libraries/stb/stb_image.h"
                  $$pre$i271$i = HEAP8[$2607>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                  $$sink$i$i274$i = $2638;$3799 = $$pre$i271$i;
                 }
                 HEAP32[$2>>2] = $$sink$i$i274$i; //@line 1223 "./libraries/stb/stb_image.h"
                 HEAP32[$0>>2] = $2609; //@line 1233 "./libraries/stb/stb_image.h"
                 $$0$i276$i = $3799;$2642 = $2609;$2643 = $$sink$i$i274$i;
                }
               } while(0);
               $2639 = (($z2$1326$i) + 2)|0; //@line 4757 "./libraries/stb/stb_image.h"
               $2640 = (($2192) + ($2639)|0); //@line 4757 "./libraries/stb/stb_image.h"
               HEAP8[$2640>>0] = $$0$i276$i; //@line 4757 "./libraries/stb/stb_image.h"
               $2641 = ($2642>>>0)<($2643>>>0); //@line 1229 "./libraries/stb/stb_image.h"
               do {
                if ($2641) {
                 $2644 = ((($2642)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                 HEAP32[$0>>2] = $2644; //@line 1230 "./libraries/stb/stb_image.h"
                 $2645 = HEAP8[$2642>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                 $$0$i258$i = $2645;$2657 = $2644;$2658 = $2643;
                } else {
                 $2646 = HEAP32[$2606>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                 $2647 = ($2646|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                 if ($2647) {
                  $$0$i258$i = 0;$2657 = $2642;$2658 = $2643;
                  break;
                 }
                 $2648 = HEAP32[$2602>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                 $2649 = HEAP32[$2604>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                 $2650 = HEAP32[$2608>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                 $2651 = (FUNCTION_TABLE_iiii[$2648 & 8191]($2649,$2607,$2650)|0); //@line 1213 "./libraries/stb/stb_image.h"
                 $2652 = ($2651|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                 if ($2652) {
                  HEAP32[$2606>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                  HEAP8[$2607>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                  $$sink$i$i256$i = $2609;$3800 = 0;
                 } else {
                  $2653 = (((($s)) + 40|0) + ($2651)|0); //@line 1223 "./libraries/stb/stb_image.h"
                  $$pre$i253$i = HEAP8[$2607>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                  $$sink$i$i256$i = $2653;$3800 = $$pre$i253$i;
                 }
                 HEAP32[$2>>2] = $$sink$i$i256$i; //@line 1223 "./libraries/stb/stb_image.h"
                 HEAP32[$0>>2] = $2609; //@line 1233 "./libraries/stb/stb_image.h"
                 $$0$i258$i = $3800;$2657 = $2609;$2658 = $$sink$i$i256$i;
                }
               } while(0);
               $2654 = (($z2$1326$i) + 1)|0; //@line 4758 "./libraries/stb/stb_image.h"
               $2655 = (($2192) + ($2654)|0); //@line 4758 "./libraries/stb/stb_image.h"
               HEAP8[$2655>>0] = $$0$i258$i; //@line 4758 "./libraries/stb/stb_image.h"
               $2656 = ($2657>>>0)<($2658>>>0); //@line 1229 "./libraries/stb/stb_image.h"
               do {
                if ($2656) {
                 $2659 = ((($2657)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                 HEAP32[$0>>2] = $2659; //@line 1230 "./libraries/stb/stb_image.h"
                 $2660 = HEAP8[$2657>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                 $$0$i251$i = $2660;$2672 = $2659;$2673 = $2658;
                } else {
                 $2661 = HEAP32[$2606>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                 $2662 = ($2661|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                 if ($2662) {
                  $$0$i251$i = 0;$2672 = $2657;$2673 = $2658;
                  break;
                 }
                 $2663 = HEAP32[$2602>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                 $2664 = HEAP32[$2604>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                 $2665 = HEAP32[$2608>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                 $2666 = (FUNCTION_TABLE_iiii[$2663 & 8191]($2664,$2607,$2665)|0); //@line 1213 "./libraries/stb/stb_image.h"
                 $2667 = ($2666|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                 if ($2667) {
                  HEAP32[$2606>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                  HEAP8[$2607>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                  $$sink$i$i249$i = $2609;$3801 = 0;
                 } else {
                  $2668 = (((($s)) + 40|0) + ($2666)|0); //@line 1223 "./libraries/stb/stb_image.h"
                  $$pre$i246$i = HEAP8[$2607>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                  $$sink$i$i249$i = $2668;$3801 = $$pre$i246$i;
                 }
                 HEAP32[$2>>2] = $$sink$i$i249$i; //@line 1223 "./libraries/stb/stb_image.h"
                 HEAP32[$0>>2] = $2609; //@line 1233 "./libraries/stb/stb_image.h"
                 $$0$i251$i = $3801;$2672 = $2609;$2673 = $$sink$i$i249$i;
                }
               } while(0);
               $2669 = (($2192) + ($z2$1326$i)|0); //@line 4759 "./libraries/stb/stb_image.h"
               HEAP8[$2669>>0] = $$0$i251$i; //@line 4759 "./libraries/stb/stb_image.h"
               $2670 = (($z2$1326$i) + 3)|0; //@line 4760 "./libraries/stb/stb_image.h"
               if ($2605) {
                $2671 = ($2672>>>0)<($2673>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                do {
                 if ($2671) {
                  $2674 = ((($2672)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                  HEAP32[$0>>2] = $2674; //@line 1230 "./libraries/stb/stb_image.h"
                  $2675 = HEAP8[$2672>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                  $$0$i233$i = $2675;$3804 = $2673;$3805 = $2674;
                 } else {
                  $2676 = HEAP32[$2606>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                  $2677 = ($2676|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                  if ($2677) {
                   $$0$i233$i = 0;$3804 = $2673;$3805 = $2672;
                   break;
                  }
                  $2678 = HEAP32[$2602>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $2679 = HEAP32[$2604>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $2680 = HEAP32[$2608>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $2681 = (FUNCTION_TABLE_iiii[$2678 & 8191]($2679,$2607,$2680)|0); //@line 1213 "./libraries/stb/stb_image.h"
                  $2682 = ($2681|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                  if ($2682) {
                   HEAP32[$2606>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                   HEAP8[$2607>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                   $$sink$i$i231$i = $2609;$3806 = 0;
                  } else {
                   $2683 = (((($s)) + 40|0) + ($2681)|0); //@line 1223 "./libraries/stb/stb_image.h"
                   $$pre$i228$i = HEAP8[$2607>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                   $$sink$i$i231$i = $2683;$3806 = $$pre$i228$i;
                  }
                  HEAP32[$2>>2] = $$sink$i$i231$i; //@line 1223 "./libraries/stb/stb_image.h"
                  HEAP32[$0>>2] = $2609; //@line 1233 "./libraries/stb/stb_image.h"
                  $$0$i233$i = $3806;$3804 = $$sink$i$i231$i;$3805 = $2609;
                 }
                } while(0);
                $2684 = $$0$i233$i&255; //@line 4761 "./libraries/stb/stb_image.h"
                $2686 = $2684;$3802 = $3804;$3803 = $3805;
               } else {
                $2686 = 255;$3802 = $2673;$3803 = $2672;
               }
               $2685 = $2686 | $all_a$3325$i; //@line 4762 "./libraries/stb/stb_image.h"
               if ($2610) {
                $2687 = $2686&255; //@line 4761 "./libraries/stb/stb_image.h"
                $2688 = (($z2$1326$i) + 4)|0; //@line 4763 "./libraries/stb/stb_image.h"
                $2689 = (($2192) + ($2670)|0); //@line 4763 "./libraries/stb/stb_image.h"
                HEAP8[$2689>>0] = $2687; //@line 4763 "./libraries/stb/stb_image.h"
                $z2$2$i = $2688;
               } else {
                $z2$2$i = $2670;
               }
               $2690 = (($i$3324$i) + 1)|0; //@line 4755 "./libraries/stb/stb_image.h"
               $2691 = HEAP32[$s>>2]|0; //@line 4755 "./libraries/stb/stb_image.h"
               $2692 = ($2690|0)<($2691|0); //@line 4755 "./libraries/stb/stb_image.h"
               if ($2692) {
                $2627 = $3803;$2628 = $3802;$all_a$3325$i = $2685;$i$3324$i = $2690;$z2$1326$i = $z2$2$i;
               } else {
                $all_a$5$i = $2685;$z2$5$i = $z2$2$i;
                break;
               }
              }
             } else {
              if ($2625) {
               $all_a$4319$i = $all_a$2331$i;$i$4318$i = 0;$z2$3320$i = $z2$0332$i;
              } else {
               $all_a$5$i = $all_a$2331$i;$z2$5$i = $z2$0332$i;
               break;
              }
              while(1) {
               if ($2611) {
                $2693 = HEAP32[$0>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                $2694 = HEAP32[$2>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
                $2695 = ($2693>>>0)<($2694>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                do {
                 if ($2695) {
                  $2696 = ((($2693)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                  HEAP32[$0>>2] = $2696; //@line 1230 "./libraries/stb/stb_image.h"
                  $2697 = HEAP8[$2693>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                  $$0$i$i208$i = $2697;$2708 = $2696;$2709 = $2694;
                 } else {
                  $2698 = HEAP32[$2606>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                  $2699 = ($2698|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                  if ($2699) {
                   $$0$i$i208$i = 0;$2708 = $2693;$2709 = $2694;
                   break;
                  }
                  $2700 = HEAP32[$2602>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $2701 = HEAP32[$2604>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $2702 = HEAP32[$2608>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $2703 = (FUNCTION_TABLE_iiii[$2700 & 8191]($2701,$2607,$2702)|0); //@line 1213 "./libraries/stb/stb_image.h"
                  $2704 = ($2703|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                  if ($2704) {
                   HEAP32[$2606>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                   HEAP8[$2607>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                   $$sink$i$i$i206$i = $2609;$3807 = 0;
                  } else {
                   $2705 = (((($s)) + 40|0) + ($2703)|0); //@line 1223 "./libraries/stb/stb_image.h"
                   $$pre$i$i203$i = HEAP8[$2607>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                   $$sink$i$i$i206$i = $2705;$3807 = $$pre$i$i203$i;
                  }
                  HEAP32[$2>>2] = $$sink$i$i$i206$i; //@line 1223 "./libraries/stb/stb_image.h"
                  HEAP32[$0>>2] = $2609; //@line 1233 "./libraries/stb/stb_image.h"
                  $$0$i$i208$i = $3807;$2708 = $2609;$2709 = $$sink$i$i$i206$i;
                 }
                } while(0);
                $2706 = $$0$i$i208$i&255; //@line 1308 "./libraries/stb/stb_image.h"
                $2707 = ($2708>>>0)<($2709>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                do {
                 if ($2707) {
                  $2710 = ((($2708)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                  HEAP32[$0>>2] = $2710; //@line 1230 "./libraries/stb/stb_image.h"
                  $2711 = HEAP8[$2708>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                  $$0$i6$i215$i = $2711;
                 } else {
                  $2712 = HEAP32[$2606>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                  $2713 = ($2712|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                  if ($2713) {
                   $$0$i6$i215$i = 0;
                   break;
                  }
                  $2714 = HEAP32[$2602>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $2715 = HEAP32[$2604>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $2716 = HEAP32[$2608>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $2717 = (FUNCTION_TABLE_iiii[$2714 & 8191]($2715,$2607,$2716)|0); //@line 1213 "./libraries/stb/stb_image.h"
                  $2718 = ($2717|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                  if ($2718) {
                   HEAP32[$2606>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                   HEAP8[$2607>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                   $$sink$i$i4$i213$i = $2609;$3808 = 0;
                  } else {
                   $2719 = (((($s)) + 40|0) + ($2717)|0); //@line 1223 "./libraries/stb/stb_image.h"
                   $$pre$i1$i210$i = HEAP8[$2607>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                   $$sink$i$i4$i213$i = $2719;$3808 = $$pre$i1$i210$i;
                  }
                  HEAP32[$2>>2] = $$sink$i$i4$i213$i; //@line 1223 "./libraries/stb/stb_image.h"
                  HEAP32[$0>>2] = $2609; //@line 1233 "./libraries/stb/stb_image.h"
                  $$0$i6$i215$i = $3808;
                 }
                } while(0);
                $2720 = $$0$i6$i215$i&255; //@line 1309 "./libraries/stb/stb_image.h"
                $2721 = $2720 << 8; //@line 1309 "./libraries/stb/stb_image.h"
                $2722 = $2721 | $2706; //@line 1309 "./libraries/stb/stb_image.h"
                $2725 = $2722;
               } else {
                $2723 = (__ZL13stbi__get32leP13stbi__context($s)|0); //@line 4767 "./libraries/stb/stb_image.h"
                $2725 = $2723;
               }
               $2724 = $2725 & $mr$1$i; //@line 4769 "./libraries/stb/stb_image.h"
               $2726 = $2724 << $2613; //@line 4587 "./libraries/stb/stb_image.h"
               $2727 = $2724 >> $rshift$0$i; //@line 4588 "./libraries/stb/stb_image.h"
               $$0$i196$i = $2612 ? $2726 : $2727; //@line 4587 "./libraries/stb/stb_image.h"
               if ($2614) {
                $result$01$i198$i = $$0$i196$i;$z$02$i197$i = $rcount$0$i;
                while(1) {
                 $2728 = $$0$i196$i >> $z$02$i197$i; //@line 4593 "./libraries/stb/stb_image.h"
                 $2729 = (($2728) + ($result$01$i198$i))|0; //@line 4593 "./libraries/stb/stb_image.h"
                 $2730 = (($z$02$i197$i) + ($rcount$0$i))|0; //@line 4594 "./libraries/stb/stb_image.h"
                 $2731 = ($2730|0)<(8); //@line 4592 "./libraries/stb/stb_image.h"
                 if ($2731) {
                  $result$01$i198$i = $2729;$z$02$i197$i = $2730;
                 } else {
                  $result$0$lcssa$i200$i = $2729;
                  break;
                 }
                }
               } else {
                $result$0$lcssa$i200$i = $$0$i196$i;
               }
               $2732 = $result$0$lcssa$i200$i&255; //@line 4769 "./libraries/stb/stb_image.h"
               $2733 = (($z2$3320$i) + 1)|0; //@line 4769 "./libraries/stb/stb_image.h"
               $2734 = (($2192) + ($z2$3320$i)|0); //@line 4769 "./libraries/stb/stb_image.h"
               HEAP8[$2734>>0] = $2732; //@line 4769 "./libraries/stb/stb_image.h"
               $2735 = $2725 & $mg$1$i; //@line 4770 "./libraries/stb/stb_image.h"
               $2736 = $2735 << $2616; //@line 4587 "./libraries/stb/stb_image.h"
               $2737 = $2735 >> $gshift$0$i; //@line 4588 "./libraries/stb/stb_image.h"
               $$0$i189$i = $2615 ? $2736 : $2737; //@line 4587 "./libraries/stb/stb_image.h"
               if ($2617) {
                $result$01$i191$i = $$0$i189$i;$z$02$i190$i = $gcount$0$i;
                while(1) {
                 $2738 = $$0$i189$i >> $z$02$i190$i; //@line 4593 "./libraries/stb/stb_image.h"
                 $2739 = (($2738) + ($result$01$i191$i))|0; //@line 4593 "./libraries/stb/stb_image.h"
                 $2740 = (($z$02$i190$i) + ($gcount$0$i))|0; //@line 4594 "./libraries/stb/stb_image.h"
                 $2741 = ($2740|0)<(8); //@line 4592 "./libraries/stb/stb_image.h"
                 if ($2741) {
                  $result$01$i191$i = $2739;$z$02$i190$i = $2740;
                 } else {
                  $result$0$lcssa$i193$i = $2739;
                  break;
                 }
                }
               } else {
                $result$0$lcssa$i193$i = $$0$i189$i;
               }
               $2742 = $result$0$lcssa$i193$i&255; //@line 4770 "./libraries/stb/stb_image.h"
               $2743 = (($z2$3320$i) + 2)|0; //@line 4770 "./libraries/stb/stb_image.h"
               $2744 = (($2192) + ($2733)|0); //@line 4770 "./libraries/stb/stb_image.h"
               HEAP8[$2744>>0] = $2742; //@line 4770 "./libraries/stb/stb_image.h"
               $2745 = $2725 & $mb$1$i; //@line 4771 "./libraries/stb/stb_image.h"
               $2746 = $2745 << $2619; //@line 4587 "./libraries/stb/stb_image.h"
               $2747 = $2745 >> $bshift$0$i; //@line 4588 "./libraries/stb/stb_image.h"
               $$0$i182$i = $2618 ? $2746 : $2747; //@line 4587 "./libraries/stb/stb_image.h"
               if ($2620) {
                $result$01$i184$i = $$0$i182$i;$z$02$i183$i = $bcount$0$i;
                while(1) {
                 $2748 = $$0$i182$i >> $z$02$i183$i; //@line 4593 "./libraries/stb/stb_image.h"
                 $2749 = (($2748) + ($result$01$i184$i))|0; //@line 4593 "./libraries/stb/stb_image.h"
                 $2750 = (($z$02$i183$i) + ($bcount$0$i))|0; //@line 4594 "./libraries/stb/stb_image.h"
                 $2751 = ($2750|0)<(8); //@line 4592 "./libraries/stb/stb_image.h"
                 if ($2751) {
                  $result$01$i184$i = $2749;$z$02$i183$i = $2750;
                 } else {
                  $result$0$lcssa$i186$i = $2749;
                  break;
                 }
                }
               } else {
                $result$0$lcssa$i186$i = $$0$i182$i;
               }
               $2752 = $result$0$lcssa$i186$i&255; //@line 4771 "./libraries/stb/stb_image.h"
               $2753 = (($z2$3320$i) + 3)|0; //@line 4771 "./libraries/stb/stb_image.h"
               $2754 = (($2192) + ($2743)|0); //@line 4771 "./libraries/stb/stb_image.h"
               HEAP8[$2754>>0] = $2752; //@line 4771 "./libraries/stb/stb_image.h"
               do {
                if ($2183) {
                 $2755 = $2725 & $ma$1$i; //@line 4772 "./libraries/stb/stb_image.h"
                 $2756 = $2755 << $2622; //@line 4587 "./libraries/stb/stb_image.h"
                 $2757 = $2755 >> $ashift$0$i; //@line 4588 "./libraries/stb/stb_image.h"
                 $$0$i180$i = $2621 ? $2756 : $2757; //@line 4587 "./libraries/stb/stb_image.h"
                 if ($2623) {
                  $result$01$i$i = $$0$i180$i;$z$02$i$i = $acount$0$i;
                 } else {
                  $2763 = $$0$i180$i;
                  break;
                 }
                 while(1) {
                  $2758 = $$0$i180$i >> $z$02$i$i; //@line 4593 "./libraries/stb/stb_image.h"
                  $2759 = (($2758) + ($result$01$i$i))|0; //@line 4593 "./libraries/stb/stb_image.h"
                  $2760 = (($z$02$i$i) + ($acount$0$i))|0; //@line 4594 "./libraries/stb/stb_image.h"
                  $2761 = ($2760|0)<(8); //@line 4592 "./libraries/stb/stb_image.h"
                  if ($2761) {
                   $result$01$i$i = $2759;$z$02$i$i = $2760;
                  } else {
                   $2763 = $2759;
                   break;
                  }
                 }
                } else {
                 $2763 = 255;
                }
               } while(0);
               $2762 = $2763 | $all_a$4319$i; //@line 4773 "./libraries/stb/stb_image.h"
               if ($2610) {
                $2764 = $2763&255; //@line 4774 "./libraries/stb/stb_image.h"
                $2765 = (($z2$3320$i) + 4)|0; //@line 4774 "./libraries/stb/stb_image.h"
                $2766 = (($2192) + ($2753)|0); //@line 4774 "./libraries/stb/stb_image.h"
                HEAP8[$2766>>0] = $2764; //@line 4774 "./libraries/stb/stb_image.h"
                $z2$4$i = $2765;
               } else {
                $z2$4$i = $2753;
               }
               $2767 = (($i$4318$i) + 1)|0; //@line 4766 "./libraries/stb/stb_image.h"
               $2768 = HEAP32[$s>>2]|0; //@line 4766 "./libraries/stb/stb_image.h"
               $2769 = ($2767|0)<($2768|0); //@line 4766 "./libraries/stb/stb_image.h"
               if ($2769) {
                $all_a$4319$i = $2762;$i$4318$i = $2767;$z2$3320$i = $z2$4$i;
               } else {
                $all_a$5$i = $2762;$z2$5$i = $z2$4$i;
                break;
               }
              }
             }
            } while(0);
            $2770 = HEAP32[$2602>>2]|0; //@line 1256 "./libraries/stb/stb_image.h"
            $2771 = ($2770|0)==(0|0); //@line 1256 "./libraries/stb/stb_image.h"
            do {
             if ($2771) {
              $$pre$i175$i = HEAP32[$0>>2]|0; //@line 1264 "./libraries/stb/stb_image.h"
              $2783 = $$pre$i175$i;
              label = 934;
             } else {
              $2772 = HEAP32[$2>>2]|0; //@line 1257 "./libraries/stb/stb_image.h"
              $2773 = HEAP32[$0>>2]|0; //@line 1257 "./libraries/stb/stb_image.h"
              $2774 = $2772; //@line 1257 "./libraries/stb/stb_image.h"
              $2775 = (($2774) - ($2773))|0; //@line 1257 "./libraries/stb/stb_image.h"
              $2776 = ($2777|0)>($2775|0); //@line 1258 "./libraries/stb/stb_image.h"
              $2778 = $2773; //@line 1258 "./libraries/stb/stb_image.h"
              if (!($2776)) {
               $2783 = $2778;
               label = 934;
               break;
              }
              HEAP32[$0>>2] = $2772; //@line 1259 "./libraries/stb/stb_image.h"
              $2779 = HEAP32[$2603>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
              $2780 = HEAP32[$2604>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
              $2781 = (($2777) - ($2775))|0; //@line 1260 "./libraries/stb/stb_image.h"
              FUNCTION_TABLE_vii[$2779 & 8191]($2780,$2781); //@line 1260 "./libraries/stb/stb_image.h"
             }
            } while(0);
            if ((label|0) == 934) {
             label = 0;
             $2782 = (($2783) + ($2777)|0); //@line 1264 "./libraries/stb/stb_image.h"
             HEAP32[$0>>2] = $2782; //@line 1264 "./libraries/stb/stb_image.h"
            }
            $2784 = (($j$1330$i) + 1)|0; //@line 4753 "./libraries/stb/stb_image.h"
            $2785 = HEAP32[$2152>>2]|0; //@line 4753 "./libraries/stb/stb_image.h"
            $2786 = ($2784|0)<($2785|0); //@line 4753 "./libraries/stb/stb_image.h"
            if ($2786) {
             $all_a$2331$i = $all_a$5$i;$j$1330$i = $2784;$z2$0332$i = $z2$5$i;
            } else {
             $2791 = $2785;$all_a$7$i = $all_a$5$i;
             break;
            }
           }
          }
         } while(0);
         $2787 = ($req_comp$$i|0)==(4); //@line 4782 "./libraries/stb/stb_image.h"
         $2788 = ($all_a$7$i|0)==(0); //@line 4782 "./libraries/stb/stb_image.h"
         $or$cond30$i = $2787 & $2788; //@line 4782 "./libraries/stb/stb_image.h"
         do {
          if ($or$cond30$i) {
           $2789 = HEAP32[$s>>2]|0; //@line 4783 "./libraries/stb/stb_image.h"
           $2790 = $2791 << 2; //@line 4783 "./libraries/stb/stb_image.h"
           $2792 = Math_imul($2790, $2789)|0; //@line 4783 "./libraries/stb/stb_image.h"
           $2793 = (($2792) + -1)|0; //@line 4783 "./libraries/stb/stb_image.h"
           $2794 = ($2793|0)>(-1); //@line 4783 "./libraries/stb/stb_image.h"
           if ($2794) {
            $i$5304$i = $2793;
           } else {
            break;
           }
           while(1) {
            $2795 = (($2192) + ($i$5304$i)|0); //@line 4784 "./libraries/stb/stb_image.h"
            HEAP8[$2795>>0] = -1; //@line 4784 "./libraries/stb/stb_image.h"
            $2796 = (($i$5304$i) + -4)|0; //@line 4783 "./libraries/stb/stb_image.h"
            $2797 = ($2796|0)>(-1); //@line 4783 "./libraries/stb/stb_image.h"
            if ($2797) {
             $i$5304$i = $2796;
            } else {
             break;
            }
           }
          }
         } while(0);
         do {
          if ($2154) {
           $2798 = $2791 >> 1; //@line 4788 "./libraries/stb/stb_image.h"
           $2799 = ($2798|0)>(0); //@line 4788 "./libraries/stb/stb_image.h"
           if (!($2799)) {
            break;
           }
           $2800 = HEAP32[$s>>2]|0; //@line 4789 "./libraries/stb/stb_image.h"
           $2801 = Math_imul($2800, $req_comp$$i)|0; //@line 4791 "./libraries/stb/stb_image.h"
           $2802 = ($2801|0)>(0); //@line 4791 "./libraries/stb/stb_image.h"
           if ($2802) {
            $j$2302$us$i = 0;
           } else {
            break;
           }
           while(1) {
            $2803 = Math_imul($2801, $j$2302$us$i)|0; //@line 4789 "./libraries/stb/stb_image.h"
            $2804 = (($2192) + ($2803)|0); //@line 4789 "./libraries/stb/stb_image.h"
            $2805 = $j$2302$us$i ^ -1; //@line 4790 "./libraries/stb/stb_image.h"
            $2806 = (($2791) + ($2805))|0; //@line 4790 "./libraries/stb/stb_image.h"
            $2807 = Math_imul($2801, $2806)|0; //@line 4790 "./libraries/stb/stb_image.h"
            $2808 = (($2192) + ($2807)|0); //@line 4790 "./libraries/stb/stb_image.h"
            $i$6301$us$i = 0;
            while(1) {
             $2809 = (($2804) + ($i$6301$us$i)|0); //@line 4792 "./libraries/stb/stb_image.h"
             $2810 = HEAP8[$2809>>0]|0; //@line 4792 "./libraries/stb/stb_image.h"
             $2811 = (($2808) + ($i$6301$us$i)|0); //@line 4792 "./libraries/stb/stb_image.h"
             $2812 = HEAP8[$2811>>0]|0; //@line 4792 "./libraries/stb/stb_image.h"
             HEAP8[$2809>>0] = $2812; //@line 4792 "./libraries/stb/stb_image.h"
             HEAP8[$2811>>0] = $2810; //@line 4792 "./libraries/stb/stb_image.h"
             $2813 = (($i$6301$us$i) + 1)|0; //@line 4791 "./libraries/stb/stb_image.h"
             $exitcond = ($2813|0)==($2801|0); //@line 4791 "./libraries/stb/stb_image.h"
             if ($exitcond) {
              break;
             } else {
              $i$6301$us$i = $2813;
             }
            }
            $2814 = (($j$2302$us$i) + 1)|0; //@line 4788 "./libraries/stb/stb_image.h"
            $exitcond280 = ($2814|0)==($2798|0); //@line 4788 "./libraries/stb/stb_image.h"
            if ($exitcond280) {
             break;
            } else {
             $j$2302$us$i = $2814;
            }
           }
          }
         } while(0);
         $2815 = ($req_comp$$i|0)==($req_comp|0); //@line 4797 "./libraries/stb/stb_image.h"
         $or$cond40$i = $2186 | $2815; //@line 4797 "./libraries/stb/stb_image.h"
         if ($or$cond40$i) {
          $out$0$i = $2192;
         } else {
          $2816 = HEAP32[$s>>2]|0; //@line 4798 "./libraries/stb/stb_image.h"
          $2817 = (__ZL20stbi__convert_formatPhiijj($2192,$req_comp$$i,$req_comp,$2816,$2791)|0); //@line 4798 "./libraries/stb/stb_image.h"
          $2818 = ($2817|0)==(0|0); //@line 4799 "./libraries/stb/stb_image.h"
          if ($2818) {
           $$2$i = 0;
           break;
          } else {
           $out$0$i = $2817;
          }
         }
         $2819 = HEAP32[$s>>2]|0; //@line 4802 "./libraries/stb/stb_image.h"
         HEAP32[$x>>2] = $2819; //@line 4802 "./libraries/stb/stb_image.h"
         $2820 = HEAP32[$2152>>2]|0; //@line 4803 "./libraries/stb/stb_image.h"
         HEAP32[$y>>2] = $2820; //@line 4803 "./libraries/stb/stb_image.h"
         $2821 = ($comp|0)==(0|0); //@line 4804 "./libraries/stb/stb_image.h"
         if ($2821) {
          $$2$i = $out$0$i;
          break;
         }
         $2822 = HEAP32[$2185>>2]|0; //@line 4804 "./libraries/stb/stb_image.h"
         HEAP32[$comp>>2] = $2822; //@line 4804 "./libraries/stb/stb_image.h"
         $$2$i = $out$0$i;
        } else {
         label = 636;
        }
       } else {
        label = 636;
       }
      }
     } while(0);
     if ((label|0) == 636) {
      HEAP32[158754] = 441371; //@line 894 "./libraries/stb/stb_image.h"
      $$2$i = 0;
     }
     $$0 = $$2$i;
     STACKTOP = sp;return ($$0|0); //@line 976 "./libraries/stb/stb_image.h"
    } else {
     label = 617;
    }
   } else {
    label = 617;
   }
  }
 } while(0);
 if ((label|0) == 617) {
  $1854 = HEAP32[$41>>2]|0; //@line 826 "./libraries/stb/stb_image.h"
  HEAP32[$0>>2] = $1854; //@line 826 "./libraries/stb/stb_image.h"
  $1855 = HEAP32[$43>>2]|0; //@line 827 "./libraries/stb/stb_image.h"
  HEAP32[$2>>2] = $1855; //@line 827 "./libraries/stb/stb_image.h"
  $1856 = $1854; //@line 947 "./libraries/stb/stb_image.h"
  $1857 = $1855; //@line 947 "./libraries/stb/stb_image.h"
  $2824 = $1856;$2825 = $1857;
 }
 $2823 = ($2824>>>0)<($2825>>>0); //@line 1229 "./libraries/stb/stb_image.h"
 if ($2823) {
  $2826 = ((($2824)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
  HEAP32[$0>>2] = $2826; //@line 1230 "./libraries/stb/stb_image.h"
  $2842 = $2826;$2843 = $2825;
 } else {
  $2827 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
  $2828 = HEAP32[$2827>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
  $2829 = ($2828|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
  if ($2829) {
   $2842 = $2824;$2843 = $2825;
  } else {
   $2830 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
   $2831 = HEAP32[$2830>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $2832 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
   $2833 = HEAP32[$2832>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $2834 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
   $2835 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
   $2836 = HEAP32[$2835>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $2837 = (FUNCTION_TABLE_iiii[$2831 & 8191]($2833,$2834,$2836)|0); //@line 1213 "./libraries/stb/stb_image.h"
   $2838 = ($2837|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
   if ($2838) {
    HEAP32[$2827>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
    $2839 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
    HEAP8[$2834>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
    $$pre$phi$i$i33Z2D = $2839;$$sink$i$i$i34 = $2839;
   } else {
    $2840 = (((($s)) + 40|0) + ($2837)|0); //@line 1223 "./libraries/stb/stb_image.h"
    $$pre1$i$i32 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
    $$pre$phi$i$i33Z2D = $$pre1$i$i32;$$sink$i$i$i34 = $2840;
   }
   HEAP32[$2>>2] = $$sink$i$i$i34; //@line 1223 "./libraries/stb/stb_image.h"
   HEAP32[$0>>2] = $$pre$phi$i$i33Z2D; //@line 1233 "./libraries/stb/stb_image.h"
   $2842 = $$pre$phi$i$i33Z2D;$2843 = $$sink$i$i$i34;
  }
 }
 $2841 = ($2842>>>0)<($2843>>>0); //@line 1229 "./libraries/stb/stb_image.h"
 if ($2841) {
  $2844 = ((($2842)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
  HEAP32[$0>>2] = $2844; //@line 1230 "./libraries/stb/stb_image.h"
  $2845 = HEAP8[$2842>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
  $$0$i23$i = $2845;$3809 = $2843;$3810 = $2844;
  label = 965;
 } else {
  $2846 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
  $2847 = HEAP32[$2846>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
  $2848 = ($2847|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
  if ($2848) {
   $2862 = $2842;$2863 = $2843;
   label = 966;
  } else {
   $2849 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
   $2850 = HEAP32[$2849>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $2851 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
   $2852 = HEAP32[$2851>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $2853 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
   $2854 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
   $2855 = HEAP32[$2854>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
   $2856 = (FUNCTION_TABLE_iiii[$2850 & 8191]($2852,$2853,$2855)|0); //@line 1213 "./libraries/stb/stb_image.h"
   $2857 = ($2856|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
   if ($2857) {
    HEAP32[$2846>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
    $2858 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
    HEAP8[$2853>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
    $$pre$phi$i20$iZ2D = $2858;$$sink$i$i21$i = $2858;$3811 = 0;
   } else {
    $2859 = (((($s)) + 40|0) + ($2856)|0); //@line 1223 "./libraries/stb/stb_image.h"
    $$pre$i18$i = HEAP8[$2853>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
    $$pre1$i19$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
    $$pre$phi$i20$iZ2D = $$pre1$i19$i;$$sink$i$i21$i = $2859;$3811 = $$pre$i18$i;
   }
   HEAP32[$2>>2] = $$sink$i$i21$i; //@line 1223 "./libraries/stb/stb_image.h"
   HEAP32[$0>>2] = $$pre$phi$i20$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
   $$0$i23$i = $3811;$3809 = $$sink$i$i21$i;$3810 = $$pre$phi$i20$iZ2D;
   label = 965;
  }
 }
 if ((label|0) == 965) {
  $2860 = ($$0$i23$i&255)>(1); //@line 4855 "./libraries/stb/stb_image.h"
  if (!($2860)) {
   $2862 = $3810;$2863 = $3809;
   label = 966;
  }
 }
 L1328: do {
  if ((label|0) == 966) {
   $2861 = ($2862>>>0)<($2863>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($2861) {
    $2864 = ((($2862)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $2864; //@line 1230 "./libraries/stb/stb_image.h"
    $2865 = HEAP8[$2862>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
    $$0$i30$i = $2865;$2881 = $2864;$2882 = $2863;
   } else {
    $2866 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $2867 = HEAP32[$2866>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $2868 = ($2867|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if ($2868) {
     break;
    }
    $2869 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
    $2870 = HEAP32[$2869>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $2871 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
    $2872 = HEAP32[$2871>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $2873 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
    $2874 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
    $2875 = HEAP32[$2874>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
    $2876 = (FUNCTION_TABLE_iiii[$2870 & 8191]($2872,$2873,$2875)|0); //@line 1213 "./libraries/stb/stb_image.h"
    $2877 = ($2876|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
    if ($2877) {
     HEAP32[$2866>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
     $2878 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
     HEAP8[$2873>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
     $$pre$phi$i27$iZ2D = $2878;$$sink$i$i28$i = $2878;$3812 = 0;
    } else {
     $2879 = (((($s)) + 40|0) + ($2876)|0); //@line 1223 "./libraries/stb/stb_image.h"
     $$pre$i25$i = HEAP8[$2873>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
     $$pre1$i26$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
     $$pre$phi$i27$iZ2D = $$pre1$i26$i;$$sink$i$i28$i = $2879;$3812 = $$pre$i25$i;
    }
    HEAP32[$2>>2] = $$sink$i$i28$i; //@line 1223 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $$pre$phi$i27$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
    $$0$i30$i = $3812;$2881 = $$pre$phi$i27$iZ2D;$2882 = $$sink$i$i28$i;
   }
   switch ($$0$i30$i<<24>>24) {
   case 1: case 2: case 3: case 9: case 10: case 11:  {
    break;
   }
   default: {
    break L1328;
   }
   }
   $2880 = ($2881>>>0)<($2882>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($2880) {
    $2883 = ((($2881)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $2883; //@line 1230 "./libraries/stb/stb_image.h"
    $2899 = $2883;$2900 = $2882;
   } else {
    $2884 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $2885 = HEAP32[$2884>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $2886 = ($2885|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if ($2886) {
     $2899 = $2881;$2900 = $2882;
    } else {
     $2887 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2888 = HEAP32[$2887>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2889 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2890 = HEAP32[$2889>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2891 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2892 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2893 = HEAP32[$2892>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2894 = (FUNCTION_TABLE_iiii[$2888 & 8191]($2890,$2891,$2893)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2895 = ($2894|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($2895) {
      HEAP32[$2884>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $2896 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$2891>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i$i$i38Z2D = $2896;$$sink$i$i$i$i39 = $2896;
     } else {
      $2897 = (((($s)) + 40|0) + ($2894)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre1$i$i$i37 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i$i$i38Z2D = $$pre1$i$i$i37;$$sink$i$i$i$i39 = $2897;
     }
     HEAP32[$2>>2] = $$sink$i$i$i$i39; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$0>>2] = $$pre$phi$i$i$i38Z2D; //@line 1233 "./libraries/stb/stb_image.h"
     $2899 = $$pre$phi$i$i$i38Z2D;$2900 = $$sink$i$i$i$i39;
    }
   }
   $2898 = ($2899>>>0)<($2900>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($2898) {
    $2901 = ((($2899)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $2901; //@line 1230 "./libraries/stb/stb_image.h"
    $2917 = $2901;$2918 = $2900;
   } else {
    $2902 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $2903 = HEAP32[$2902>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $2904 = ($2903|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if ($2904) {
     $2917 = $2899;$2918 = $2900;
    } else {
     $2905 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2906 = HEAP32[$2905>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2907 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2908 = HEAP32[$2907>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2909 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2910 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2911 = HEAP32[$2910>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2912 = (FUNCTION_TABLE_iiii[$2906 & 8191]($2908,$2909,$2911)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2913 = ($2912|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($2913) {
      HEAP32[$2902>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $2914 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$2909>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i3$i$i43Z2D = $2914;$$sink$i$i4$i$i44 = $2914;
     } else {
      $2915 = (((($s)) + 40|0) + ($2912)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre1$i2$i$i42 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i3$i$i43Z2D = $$pre1$i2$i$i42;$$sink$i$i4$i$i44 = $2915;
     }
     HEAP32[$2>>2] = $$sink$i$i4$i$i44; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$0>>2] = $$pre$phi$i3$i$i43Z2D; //@line 1233 "./libraries/stb/stb_image.h"
     $2917 = $$pre$phi$i3$i$i43Z2D;$2918 = $$sink$i$i4$i$i44;
    }
   }
   $2916 = ($2917>>>0)<($2918>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($2916) {
    $2919 = ((($2917)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $2919; //@line 1230 "./libraries/stb/stb_image.h"
    $2935 = $2919;$2936 = $2918;
   } else {
    $2920 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $2921 = HEAP32[$2920>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $2922 = ($2921|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if ($2922) {
     $2935 = $2917;$2936 = $2918;
    } else {
     $2923 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2924 = HEAP32[$2923>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2925 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2926 = HEAP32[$2925>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2927 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2928 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2929 = HEAP32[$2928>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2930 = (FUNCTION_TABLE_iiii[$2924 & 8191]($2926,$2927,$2929)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2931 = ($2930|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($2931) {
      HEAP32[$2920>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $2932 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$2927>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i$i34$iZ2D = $2932;$$sink$i$i$i35$i = $2932;
     } else {
      $2933 = (((($s)) + 40|0) + ($2930)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre1$i$i33$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i$i34$iZ2D = $$pre1$i$i33$i;$$sink$i$i$i35$i = $2933;
     }
     HEAP32[$2>>2] = $$sink$i$i$i35$i; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$0>>2] = $$pre$phi$i$i34$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
     $2935 = $$pre$phi$i$i34$iZ2D;$2936 = $$sink$i$i$i35$i;
    }
   }
   $2934 = ($2935>>>0)<($2936>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($2934) {
    $2937 = ((($2935)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $2937; //@line 1230 "./libraries/stb/stb_image.h"
    $2953 = $2937;$2954 = $2936;
   } else {
    $2938 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $2939 = HEAP32[$2938>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $2940 = ($2939|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if ($2940) {
     $2953 = $2935;$2954 = $2936;
    } else {
     $2941 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2942 = HEAP32[$2941>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2943 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2944 = HEAP32[$2943>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2945 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2946 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2947 = HEAP32[$2946>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2948 = (FUNCTION_TABLE_iiii[$2942 & 8191]($2944,$2945,$2947)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2949 = ($2948|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($2949) {
      HEAP32[$2938>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $2950 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$2945>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i3$i41$iZ2D = $2950;$$sink$i$i4$i42$i = $2950;
     } else {
      $2951 = (((($s)) + 40|0) + ($2948)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre1$i2$i40$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i3$i41$iZ2D = $$pre1$i2$i40$i;$$sink$i$i4$i42$i = $2951;
     }
     HEAP32[$2>>2] = $$sink$i$i4$i42$i; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$0>>2] = $$pre$phi$i3$i41$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
     $2953 = $$pre$phi$i3$i41$iZ2D;$2954 = $$sink$i$i4$i42$i;
    }
   }
   $2952 = ($2953>>>0)<($2954>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($2952) {
    $2955 = ((($2953)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $2955; //@line 1230 "./libraries/stb/stb_image.h"
    $2971 = $2955;$2972 = $2954;
   } else {
    $2956 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $2957 = HEAP32[$2956>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $2958 = ($2957|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if ($2958) {
     $2971 = $2953;$2972 = $2954;
    } else {
     $2959 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2960 = HEAP32[$2959>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2961 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2962 = HEAP32[$2961>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2963 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2964 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2965 = HEAP32[$2964>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2966 = (FUNCTION_TABLE_iiii[$2960 & 8191]($2962,$2963,$2965)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2967 = ($2966|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($2967) {
      HEAP32[$2956>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $2968 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$2963>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i48$iZ2D = $2968;$$sink$i$i49$i = $2968;
     } else {
      $2969 = (((($s)) + 40|0) + ($2966)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre1$i47$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i48$iZ2D = $$pre1$i47$i;$$sink$i$i49$i = $2969;
     }
     HEAP32[$2>>2] = $$sink$i$i49$i; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$0>>2] = $$pre$phi$i48$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
     $2971 = $$pre$phi$i48$iZ2D;$2972 = $$sink$i$i49$i;
    }
   }
   $2970 = ($2971>>>0)<($2972>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($2970) {
    $2973 = ((($2971)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $2973; //@line 1230 "./libraries/stb/stb_image.h"
    $2989 = $2973;$2990 = $2972;
   } else {
    $2974 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $2975 = HEAP32[$2974>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $2976 = ($2975|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if ($2976) {
     $2989 = $2971;$2990 = $2972;
    } else {
     $2977 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2978 = HEAP32[$2977>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2979 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2980 = HEAP32[$2979>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2981 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2982 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2983 = HEAP32[$2982>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2984 = (FUNCTION_TABLE_iiii[$2978 & 8191]($2980,$2981,$2983)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2985 = ($2984|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($2985) {
      HEAP32[$2974>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $2986 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$2981>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i$i55$iZ2D = $2986;$$sink$i$i$i56$i = $2986;
     } else {
      $2987 = (((($s)) + 40|0) + ($2984)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre1$i$i54$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i$i55$iZ2D = $$pre1$i$i54$i;$$sink$i$i$i56$i = $2987;
     }
     HEAP32[$2>>2] = $$sink$i$i$i56$i; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$0>>2] = $$pre$phi$i$i55$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
     $2989 = $$pre$phi$i$i55$iZ2D;$2990 = $$sink$i$i$i56$i;
    }
   }
   $2988 = ($2989>>>0)<($2990>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($2988) {
    $2991 = ((($2989)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $2991; //@line 1230 "./libraries/stb/stb_image.h"
    $3007 = $2991;$3008 = $2990;
   } else {
    $2992 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $2993 = HEAP32[$2992>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $2994 = ($2993|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if ($2994) {
     $3007 = $2989;$3008 = $2990;
    } else {
     $2995 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2996 = HEAP32[$2995>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2997 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $2998 = HEAP32[$2997>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $2999 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3000 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3001 = HEAP32[$3000>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $3002 = (FUNCTION_TABLE_iiii[$2996 & 8191]($2998,$2999,$3001)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3003 = ($3002|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($3003) {
      HEAP32[$2992>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $3004 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$2999>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i3$i62$iZ2D = $3004;$$sink$i$i4$i63$i = $3004;
     } else {
      $3005 = (((($s)) + 40|0) + ($3002)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre1$i2$i61$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i3$i62$iZ2D = $$pre1$i2$i61$i;$$sink$i$i4$i63$i = $3005;
     }
     HEAP32[$2>>2] = $$sink$i$i4$i63$i; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$0>>2] = $$pre$phi$i3$i62$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
     $3007 = $$pre$phi$i3$i62$iZ2D;$3008 = $$sink$i$i4$i63$i;
    }
   }
   $3006 = ($3007>>>0)<($3008>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($3006) {
    $3009 = ((($3007)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $3009; //@line 1230 "./libraries/stb/stb_image.h"
    $3025 = $3009;$3026 = $3008;
   } else {
    $3010 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $3011 = HEAP32[$3010>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $3012 = ($3011|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if ($3012) {
     $3025 = $3007;$3026 = $3008;
    } else {
     $3013 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3014 = HEAP32[$3013>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $3015 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3016 = HEAP32[$3015>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $3017 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3018 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3019 = HEAP32[$3018>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $3020 = (FUNCTION_TABLE_iiii[$3014 & 8191]($3016,$3017,$3019)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3021 = ($3020|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($3021) {
      HEAP32[$3010>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $3022 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$3017>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i$i69$iZ2D = $3022;$$sink$i$i$i70$i = $3022;
     } else {
      $3023 = (((($s)) + 40|0) + ($3020)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre1$i$i68$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i$i69$iZ2D = $$pre1$i$i68$i;$$sink$i$i$i70$i = $3023;
     }
     HEAP32[$2>>2] = $$sink$i$i$i70$i; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$0>>2] = $$pre$phi$i$i69$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
     $3025 = $$pre$phi$i$i69$iZ2D;$3026 = $$sink$i$i$i70$i;
    }
   }
   $3024 = ($3025>>>0)<($3026>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($3024) {
    $3027 = ((($3025)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $3027; //@line 1230 "./libraries/stb/stb_image.h"
    $3043 = $3027;$3044 = $3026;
   } else {
    $3028 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $3029 = HEAP32[$3028>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $3030 = ($3029|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if ($3030) {
     $3043 = $3025;$3044 = $3026;
    } else {
     $3031 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3032 = HEAP32[$3031>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $3033 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3034 = HEAP32[$3033>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $3035 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3036 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3037 = HEAP32[$3036>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $3038 = (FUNCTION_TABLE_iiii[$3032 & 8191]($3034,$3035,$3037)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3039 = ($3038|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($3039) {
      HEAP32[$3028>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $3040 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$3035>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i3$i76$iZ2D = $3040;$$sink$i$i4$i77$i = $3040;
     } else {
      $3041 = (((($s)) + 40|0) + ($3038)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre1$i2$i75$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i3$i76$iZ2D = $$pre1$i2$i75$i;$$sink$i$i4$i77$i = $3041;
     }
     HEAP32[$2>>2] = $$sink$i$i4$i77$i; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$0>>2] = $$pre$phi$i3$i76$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
     $3043 = $$pre$phi$i3$i76$iZ2D;$3044 = $$sink$i$i4$i77$i;
    }
   }
   $3042 = ($3043>>>0)<($3044>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($3042) {
    $3045 = ((($3043)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $3045; //@line 1230 "./libraries/stb/stb_image.h"
    $3046 = HEAP8[$3043>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
    $$0$i$i86$i = $3046;$3064 = $3045;$3065 = $3044;
   } else {
    $3047 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $3048 = HEAP32[$3047>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $3049 = ($3048|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if ($3049) {
     $$0$i$i86$i = 0;$3064 = $3043;$3065 = $3044;
    } else {
     $3050 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3051 = HEAP32[$3050>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $3052 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3053 = HEAP32[$3052>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $3054 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3055 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3056 = HEAP32[$3055>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $3057 = (FUNCTION_TABLE_iiii[$3051 & 8191]($3053,$3054,$3056)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3058 = ($3057|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($3058) {
      HEAP32[$3047>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $3059 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$3054>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i$i83$iZ2D = $3059;$$sink$i$i$i84$i = $3059;$3813 = 0;
     } else {
      $3060 = (((($s)) + 40|0) + ($3057)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre$i$i81$i = HEAP8[$3054>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
      $$pre1$i$i82$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i$i83$iZ2D = $$pre1$i$i82$i;$$sink$i$i$i84$i = $3060;$3813 = $$pre$i$i81$i;
     }
     HEAP32[$2>>2] = $$sink$i$i$i84$i; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$0>>2] = $$pre$phi$i$i83$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
     $$0$i$i86$i = $3813;$3064 = $$pre$phi$i$i83$iZ2D;$3065 = $$sink$i$i$i84$i;
    }
   }
   $3061 = $$0$i$i86$i&255; //@line 1293 "./libraries/stb/stb_image.h"
   $3062 = $3061 << 8; //@line 1294 "./libraries/stb/stb_image.h"
   $3063 = ($3064>>>0)<($3065>>>0); //@line 1229 "./libraries/stb/stb_image.h"
   if ($3063) {
    $3066 = ((($3064)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $3066; //@line 1230 "./libraries/stb/stb_image.h"
    $3067 = HEAP8[$3064>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
    $$0$i6$i93$i = $3067;$3086 = $3066;$3087 = $3065;
   } else {
    $3068 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
    $3069 = HEAP32[$3068>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
    $3070 = ($3069|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
    if ($3070) {
     $$0$i6$i93$i = 0;$3086 = $3064;$3087 = $3065;
    } else {
     $3071 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3072 = HEAP32[$3071>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $3073 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3074 = HEAP32[$3073>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $3075 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3076 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3077 = HEAP32[$3076>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
     $3078 = (FUNCTION_TABLE_iiii[$3072 & 8191]($3074,$3075,$3077)|0); //@line 1213 "./libraries/stb/stb_image.h"
     $3079 = ($3078|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
     if ($3079) {
      HEAP32[$3068>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
      $3080 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
      HEAP8[$3075>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
      $$pre$phi$i3$i90$iZ2D = $3080;$$sink$i$i4$i91$i = $3080;$3814 = 0;
     } else {
      $3081 = (((($s)) + 40|0) + ($3078)|0); //@line 1223 "./libraries/stb/stb_image.h"
      $$pre$i1$i88$i = HEAP8[$3075>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
      $$pre1$i2$i89$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
      $$pre$phi$i3$i90$iZ2D = $$pre1$i2$i89$i;$$sink$i$i4$i91$i = $3081;$3814 = $$pre$i1$i88$i;
     }
     HEAP32[$2>>2] = $$sink$i$i4$i91$i; //@line 1223 "./libraries/stb/stb_image.h"
     HEAP32[$0>>2] = $$pre$phi$i3$i90$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
     $$0$i6$i93$i = $3814;$3086 = $$pre$phi$i3$i90$iZ2D;$3087 = $$sink$i$i4$i91$i;
    }
   }
   $3082 = $$0$i6$i93$i&255; //@line 1294 "./libraries/stb/stb_image.h"
   $3083 = $3082 | $3062; //@line 1294 "./libraries/stb/stb_image.h"
   $3084 = ($3083|0)==(0); //@line 4863 "./libraries/stb/stb_image.h"
   if (!($3084)) {
    $3085 = ($3086>>>0)<($3087>>>0); //@line 1229 "./libraries/stb/stb_image.h"
    do {
     if ($3085) {
      $3088 = ((($3086)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
      HEAP32[$0>>2] = $3088; //@line 1230 "./libraries/stb/stb_image.h"
      $3089 = HEAP8[$3086>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
      $$0$i$i100$i = $3089;$3107 = $3088;$3108 = $3087;
     } else {
      $3090 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
      $3091 = HEAP32[$3090>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
      $3092 = ($3091|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
      if ($3092) {
       $$0$i$i100$i = 0;$3107 = $3086;$3108 = $3087;
       break;
      }
      $3093 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
      $3094 = HEAP32[$3093>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $3095 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
      $3096 = HEAP32[$3095>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $3097 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
      $3098 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
      $3099 = HEAP32[$3098>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $3100 = (FUNCTION_TABLE_iiii[$3094 & 8191]($3096,$3097,$3099)|0); //@line 1213 "./libraries/stb/stb_image.h"
      $3101 = ($3100|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
      if ($3101) {
       HEAP32[$3090>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
       $3102 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
       HEAP8[$3097>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
       $$pre$phi$i$i97$iZ2D = $3102;$$sink$i$i$i98$i = $3102;$3815 = 0;
      } else {
       $3103 = (((($s)) + 40|0) + ($3100)|0); //@line 1223 "./libraries/stb/stb_image.h"
       $$pre$i$i95$i = HEAP8[$3097>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
       $$pre1$i$i96$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
       $$pre$phi$i$i97$iZ2D = $$pre1$i$i96$i;$$sink$i$i$i98$i = $3103;$3815 = $$pre$i$i95$i;
      }
      HEAP32[$2>>2] = $$sink$i$i$i98$i; //@line 1223 "./libraries/stb/stb_image.h"
      HEAP32[$0>>2] = $$pre$phi$i$i97$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
      $$0$i$i100$i = $3815;$3107 = $$pre$phi$i$i97$iZ2D;$3108 = $$sink$i$i$i98$i;
     }
    } while(0);
    $3104 = $$0$i$i100$i&255; //@line 1293 "./libraries/stb/stb_image.h"
    $3105 = $3104 << 8; //@line 1294 "./libraries/stb/stb_image.h"
    $3106 = ($3107>>>0)<($3108>>>0); //@line 1229 "./libraries/stb/stb_image.h"
    do {
     if ($3106) {
      $3109 = ((($3107)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
      HEAP32[$0>>2] = $3109; //@line 1230 "./libraries/stb/stb_image.h"
      $3110 = HEAP8[$3107>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
      $$0$i6$i107$i = $3110;$3129 = $3109;$3130 = $3108;
     } else {
      $3111 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
      $3112 = HEAP32[$3111>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
      $3113 = ($3112|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
      if ($3113) {
       $$0$i6$i107$i = 0;$3129 = $3107;$3130 = $3108;
       break;
      }
      $3114 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
      $3115 = HEAP32[$3114>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $3116 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
      $3117 = HEAP32[$3116>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $3118 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
      $3119 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
      $3120 = HEAP32[$3119>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $3121 = (FUNCTION_TABLE_iiii[$3115 & 8191]($3117,$3118,$3120)|0); //@line 1213 "./libraries/stb/stb_image.h"
      $3122 = ($3121|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
      if ($3122) {
       HEAP32[$3111>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
       $3123 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
       HEAP8[$3118>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
       $$pre$phi$i3$i104$iZ2D = $3123;$$sink$i$i4$i105$i = $3123;$3816 = 0;
      } else {
       $3124 = (((($s)) + 40|0) + ($3121)|0); //@line 1223 "./libraries/stb/stb_image.h"
       $$pre$i1$i102$i = HEAP8[$3118>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
       $$pre1$i2$i103$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
       $$pre$phi$i3$i104$iZ2D = $$pre1$i2$i103$i;$$sink$i$i4$i105$i = $3124;$3816 = $$pre$i1$i102$i;
      }
      HEAP32[$2>>2] = $$sink$i$i4$i105$i; //@line 1223 "./libraries/stb/stb_image.h"
      HEAP32[$0>>2] = $$pre$phi$i3$i104$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
      $$0$i6$i107$i = $3816;$3129 = $$pre$phi$i3$i104$iZ2D;$3130 = $$sink$i$i4$i105$i;
     }
    } while(0);
    $3125 = $$0$i6$i107$i&255; //@line 1294 "./libraries/stb/stb_image.h"
    $3126 = $3125 | $3105; //@line 1294 "./libraries/stb/stb_image.h"
    $3127 = ($3126|0)==(0); //@line 4864 "./libraries/stb/stb_image.h"
    if ($3127) {
     break;
    }
    $3128 = ($3129>>>0)<($3130>>>0); //@line 1229 "./libraries/stb/stb_image.h"
    do {
     if ($3128) {
      $3131 = ((($3129)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
      HEAP32[$0>>2] = $3131; //@line 1230 "./libraries/stb/stb_image.h"
      $3132 = HEAP8[$3129>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
      $$0$i114$i = $3132;
      label = 1074;
     } else {
      $3133 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
      $3134 = HEAP32[$3133>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
      $3135 = ($3134|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
      if ($3135) {
       break;
      }
      $3136 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
      $3137 = HEAP32[$3136>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $3138 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
      $3139 = HEAP32[$3138>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $3140 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
      $3141 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
      $3142 = HEAP32[$3141>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
      $3143 = (FUNCTION_TABLE_iiii[$3137 & 8191]($3139,$3140,$3142)|0); //@line 1213 "./libraries/stb/stb_image.h"
      $3144 = ($3143|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
      if ($3144) {
       HEAP32[$3133>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
       $3145 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
       HEAP8[$3140>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
       $$pre$phi$i111$iZ2D = $3145;$$sink$i$i112$i = $3145;$3817 = 0;
      } else {
       $3146 = (((($s)) + 40|0) + ($3143)|0); //@line 1223 "./libraries/stb/stb_image.h"
       $$pre$i109$i = HEAP8[$3140>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
       $$pre1$i110$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
       $$pre$phi$i111$iZ2D = $$pre1$i110$i;$$sink$i$i112$i = $3146;$3817 = $$pre$i109$i;
      }
      HEAP32[$2>>2] = $$sink$i$i112$i; //@line 1223 "./libraries/stb/stb_image.h"
      HEAP32[$0>>2] = $$pre$phi$i111$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
      $$0$i114$i = $3817;
      label = 1074;
     }
    } while(0);
    L1467: do {
     if ((label|0) == 1074) {
      switch ($$0$i114$i<<24>>24) {
      case 8: case 16: case 24: case 32:  {
       break;
      }
      default: {
       break L1467;
      }
      }
      $3149 = HEAP32[$41>>2]|0; //@line 826 "./libraries/stb/stb_image.h"
      HEAP32[$0>>2] = $3149; //@line 826 "./libraries/stb/stb_image.h"
      $3150 = HEAP32[$43>>2]|0; //@line 827 "./libraries/stb/stb_image.h"
      HEAP32[$2>>2] = $3150; //@line 827 "./libraries/stb/stb_image.h"
      $$cast = $3149; //@line 1229 "./libraries/stb/stb_image.h"
      $$cast156 = $3150; //@line 1229 "./libraries/stb/stb_image.h"
      $3151 = ($$cast>>>0)<($$cast156>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3151) {
        $3152 = ((($$cast)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3152; //@line 1230 "./libraries/stb/stb_image.h"
        $3153 = HEAP8[$$cast>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i$i51 = $3153;$$in$i = $3152;$3170 = $$cast156;
       } else {
        $3154 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3155 = HEAP32[$3154>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3156 = ($3155|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3156) {
         $$0$i$i51 = 0;$$in$i = $$cast;$3170 = $$cast156;
         break;
        }
        $3157 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3158 = HEAP32[$3157>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3159 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3160 = HEAP32[$3159>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3161 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3162 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3163 = HEAP32[$3162>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3164 = (FUNCTION_TABLE_iiii[$3158 & 8191]($3160,$3161,$3163)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3165 = ($3164|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3165) {
         HEAP32[$3154>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3166 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3161>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i48Z2D = $3166;$$sink$i$i$i49 = $3166;$3818 = 0;
        } else {
         $3167 = (((($s)) + 40|0) + ($3164)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i$i46 = HEAP8[$3161>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i$i47 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i48Z2D = $$pre1$i$i47;$$sink$i$i$i49 = $3167;$3818 = $$pre$i$i46;
        }
        HEAP32[$2>>2] = $$sink$i$i$i49; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i$i48Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i$i51 = $3818;$$in$i = $$pre$phi$i$i48Z2D;$3170 = $$sink$i$i$i49;
       }
      } while(0);
      $3168 = $$0$i$i51&255; //@line 4877 "./libraries/stb/stb_image.h"
      $3169 = ($$in$i>>>0)<($3170>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3169) {
        $3171 = ((($$in$i)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3171; //@line 1230 "./libraries/stb/stb_image.h"
        $3172 = HEAP8[$$in$i>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i23$i58 = $3172;$$in215$i = $3171;$3188 = $3170;
       } else {
        $3173 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3174 = HEAP32[$3173>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3175 = ($3174|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3175) {
         $$0$i23$i58 = 0;$$in215$i = $$in$i;$3188 = $3170;
         break;
        }
        $3176 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3177 = HEAP32[$3176>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3178 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3179 = HEAP32[$3178>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3180 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3181 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3182 = HEAP32[$3181>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3183 = (FUNCTION_TABLE_iiii[$3177 & 8191]($3179,$3180,$3182)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3184 = ($3183|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3184) {
         HEAP32[$3173>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3185 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3180>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i20$i55Z2D = $3185;$$sink$i$i21$i56 = $3185;$3819 = 0;
        } else {
         $3186 = (((($s)) + 40|0) + ($3183)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i18$i53 = HEAP8[$3180>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i19$i54 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i20$i55Z2D = $$pre1$i19$i54;$$sink$i$i21$i56 = $3186;$3819 = $$pre$i18$i53;
        }
        HEAP32[$2>>2] = $$sink$i$i21$i56; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i20$i55Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i23$i58 = $3819;$$in215$i = $$pre$phi$i20$i55Z2D;$3188 = $$sink$i$i21$i56;
       }
      } while(0);
      $3187 = ($$in215$i>>>0)<($3188>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3187) {
        $3189 = ((($$in215$i)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3189; //@line 1230 "./libraries/stb/stb_image.h"
        $3190 = HEAP8[$$in215$i>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i30$i65 = $3190;$$in216$i = $3189;$3207 = $3188;
       } else {
        $3191 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3192 = HEAP32[$3191>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3193 = ($3192|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3193) {
         $$0$i30$i65 = 0;$$in216$i = $$in215$i;$3207 = $3188;
         break;
        }
        $3194 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3195 = HEAP32[$3194>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3196 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3197 = HEAP32[$3196>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3198 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3199 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3200 = HEAP32[$3199>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3201 = (FUNCTION_TABLE_iiii[$3195 & 8191]($3197,$3198,$3200)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3202 = ($3201|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3202) {
         HEAP32[$3191>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3203 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3198>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i27$i62Z2D = $3203;$$sink$i$i28$i63 = $3203;$3820 = 0;
        } else {
         $3204 = (((($s)) + 40|0) + ($3201)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i25$i60 = HEAP8[$3198>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i26$i61 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i27$i62Z2D = $$pre1$i26$i61;$$sink$i$i28$i63 = $3204;$3820 = $$pre$i25$i60;
        }
        HEAP32[$2>>2] = $$sink$i$i28$i63; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i27$i62Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i30$i65 = $3820;$$in216$i = $$pre$phi$i27$i62Z2D;$3207 = $$sink$i$i28$i63;
       }
      } while(0);
      $3205 = $$0$i30$i65&255; //@line 4879 "./libraries/stb/stb_image.h"
      $3206 = ($$in216$i>>>0)<($3207>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3206) {
        $3208 = ((($$in216$i)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3208; //@line 1230 "./libraries/stb/stb_image.h"
        $3209 = HEAP8[$$in216$i>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i$i$i72 = $3209;$$in217$i = $3208;$3226 = $3207;
       } else {
        $3210 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3211 = HEAP32[$3210>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3212 = ($3211|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3212) {
         $$0$i$i$i72 = 0;$$in217$i = $$in216$i;$3226 = $3207;
         break;
        }
        $3213 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3214 = HEAP32[$3213>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3215 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3216 = HEAP32[$3215>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3217 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3218 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3219 = HEAP32[$3218>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3220 = (FUNCTION_TABLE_iiii[$3214 & 8191]($3216,$3217,$3219)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3221 = ($3220|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3221) {
         HEAP32[$3210>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3222 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3217>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i$i69Z2D = $3222;$$sink$i$i$i$i70 = $3222;$3821 = 0;
        } else {
         $3223 = (((($s)) + 40|0) + ($3220)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i$i$i67 = HEAP8[$3217>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i$i$i68 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i$i69Z2D = $$pre1$i$i$i68;$$sink$i$i$i$i70 = $3223;$3821 = $$pre$i$i$i67;
        }
        HEAP32[$2>>2] = $$sink$i$i$i$i70; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i$i$i69Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i$i$i72 = $3821;$$in217$i = $$pre$phi$i$i$i69Z2D;$3226 = $$sink$i$i$i$i70;
       }
      } while(0);
      $3224 = $$0$i$i$i72&255; //@line 1308 "./libraries/stb/stb_image.h"
      $3225 = ($$in217$i>>>0)<($3226>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3225) {
        $3227 = ((($$in217$i)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3227; //@line 1230 "./libraries/stb/stb_image.h"
        $3228 = HEAP8[$$in217$i>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i6$i$i = $3228;$$in218$i = $3227;$3247 = $3226;
       } else {
        $3229 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3230 = HEAP32[$3229>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3231 = ($3230|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3231) {
         $$0$i6$i$i = 0;$$in218$i = $$in217$i;$3247 = $3226;
         break;
        }
        $3232 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3233 = HEAP32[$3232>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3234 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3235 = HEAP32[$3234>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3236 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3237 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3238 = HEAP32[$3237>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3239 = (FUNCTION_TABLE_iiii[$3233 & 8191]($3235,$3236,$3238)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3240 = ($3239|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3240) {
         HEAP32[$3229>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3241 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3236>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i$i75Z2D = $3241;$$sink$i$i4$i$i76 = $3241;$3822 = 0;
        } else {
         $3242 = (((($s)) + 40|0) + ($3239)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i1$i$i = HEAP8[$3236>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i2$i$i74 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i$i75Z2D = $$pre1$i2$i$i74;$$sink$i$i4$i$i76 = $3242;$3822 = $$pre$i1$i$i;
        }
        HEAP32[$2>>2] = $$sink$i$i4$i$i76; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i3$i$i75Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i6$i$i = $3822;$$in218$i = $$pre$phi$i3$i$i75Z2D;$3247 = $$sink$i$i4$i$i76;
       }
      } while(0);
      $3243 = $$0$i6$i$i&255; //@line 1309 "./libraries/stb/stb_image.h"
      $3244 = $3243 << 8; //@line 1309 "./libraries/stb/stb_image.h"
      $3245 = $3244 | $3224; //@line 1309 "./libraries/stb/stb_image.h"
      $3246 = ($$in218$i>>>0)<($3247>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3246) {
        $3248 = ((($$in218$i)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3248; //@line 1230 "./libraries/stb/stb_image.h"
        $3249 = HEAP8[$$in218$i>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i$i37$i = $3249;$$in219$i = $3248;$3266 = $3247;
       } else {
        $3250 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3251 = HEAP32[$3250>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3252 = ($3251|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3252) {
         $$0$i$i37$i = 0;$$in219$i = $$in218$i;$3266 = $3247;
         break;
        }
        $3253 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3254 = HEAP32[$3253>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3255 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3256 = HEAP32[$3255>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3257 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3258 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3259 = HEAP32[$3258>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3260 = (FUNCTION_TABLE_iiii[$3254 & 8191]($3256,$3257,$3259)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3261 = ($3260|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3261) {
         HEAP32[$3250>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3262 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3257>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i34$i80Z2D = $3262;$$sink$i$i$i35$i81 = $3262;$3823 = 0;
        } else {
         $3263 = (((($s)) + 40|0) + ($3260)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i$i32$i = HEAP8[$3257>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i$i33$i79 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i34$i80Z2D = $$pre1$i$i33$i79;$$sink$i$i$i35$i81 = $3263;$3823 = $$pre$i$i32$i;
        }
        HEAP32[$2>>2] = $$sink$i$i$i35$i81; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i$i34$i80Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i$i37$i = $3823;$$in219$i = $$pre$phi$i$i34$i80Z2D;$3266 = $$sink$i$i$i35$i81;
       }
      } while(0);
      $3264 = $$0$i$i37$i&255; //@line 1308 "./libraries/stb/stb_image.h"
      $3265 = ($$in219$i>>>0)<($3266>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3265) {
        $3267 = ((($$in219$i)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3267; //@line 1230 "./libraries/stb/stb_image.h"
        $3268 = HEAP8[$$in219$i>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i6$i44$i = $3268;$$in220$i = $3267;$3287 = $3266;
       } else {
        $3269 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3270 = HEAP32[$3269>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3271 = ($3270|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3271) {
         $$0$i6$i44$i = 0;$$in220$i = $$in219$i;$3287 = $3266;
         break;
        }
        $3272 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3273 = HEAP32[$3272>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3274 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3275 = HEAP32[$3274>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3276 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3277 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3278 = HEAP32[$3277>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3279 = (FUNCTION_TABLE_iiii[$3273 & 8191]($3275,$3276,$3278)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3280 = ($3279|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3280) {
         HEAP32[$3269>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3281 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3276>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i41$i85Z2D = $3281;$$sink$i$i4$i42$i86 = $3281;$3824 = 0;
        } else {
         $3282 = (((($s)) + 40|0) + ($3279)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i1$i39$i = HEAP8[$3276>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i2$i40$i84 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i41$i85Z2D = $$pre1$i2$i40$i84;$$sink$i$i4$i42$i86 = $3282;$3824 = $$pre$i1$i39$i;
        }
        HEAP32[$2>>2] = $$sink$i$i4$i42$i86; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i3$i41$i85Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i6$i44$i = $3824;$$in220$i = $$pre$phi$i3$i41$i85Z2D;$3287 = $$sink$i$i4$i42$i86;
       }
      } while(0);
      $3283 = $$0$i6$i44$i&255; //@line 1309 "./libraries/stb/stb_image.h"
      $3284 = $3283 << 8; //@line 1309 "./libraries/stb/stb_image.h"
      $3285 = $3284 | $3264; //@line 1309 "./libraries/stb/stb_image.h"
      $3286 = ($$in220$i>>>0)<($3287>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3286) {
        $3288 = ((($$in220$i)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3288; //@line 1230 "./libraries/stb/stb_image.h"
        $3289 = HEAP8[$$in220$i>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i51$i = $3289;$$in221$i = $3288;$3306 = $3287;
       } else {
        $3290 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3291 = HEAP32[$3290>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3292 = ($3291|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3292) {
         $$0$i51$i = 0;$$in221$i = $$in220$i;$3306 = $3287;
         break;
        }
        $3293 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3294 = HEAP32[$3293>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3295 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3296 = HEAP32[$3295>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3297 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3298 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3299 = HEAP32[$3298>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3300 = (FUNCTION_TABLE_iiii[$3294 & 8191]($3296,$3297,$3299)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3301 = ($3300|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3301) {
         HEAP32[$3290>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3302 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3297>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i48$i89Z2D = $3302;$$sink$i$i49$i90 = $3302;$3825 = 0;
        } else {
         $3303 = (((($s)) + 40|0) + ($3300)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i46$i = HEAP8[$3297>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i47$i88 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i48$i89Z2D = $$pre1$i47$i88;$$sink$i$i49$i90 = $3303;$3825 = $$pre$i46$i;
        }
        HEAP32[$2>>2] = $$sink$i$i49$i90; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i48$i89Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i51$i = $3825;$$in221$i = $$pre$phi$i48$i89Z2D;$3306 = $$sink$i$i49$i90;
       }
      } while(0);
      $3304 = $$0$i51$i&255; //@line 4883 "./libraries/stb/stb_image.h"
      $3305 = ($$in221$i>>>0)<($3306>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3305) {
        $3307 = ((($$in221$i)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3307; //@line 1230 "./libraries/stb/stb_image.h"
        $$in222$i = $3307;$3323 = $3306;
       } else {
        $3308 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3309 = HEAP32[$3308>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3310 = ($3309|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3310) {
         $$in222$i = $$in221$i;$3323 = $3306;
         break;
        }
        $3311 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3312 = HEAP32[$3311>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3313 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3314 = HEAP32[$3313>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3315 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3316 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3317 = HEAP32[$3316>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3318 = (FUNCTION_TABLE_iiii[$3312 & 8191]($3314,$3315,$3317)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3319 = ($3318|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3319) {
         HEAP32[$3308>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3320 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3315>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i55$i94Z2D = $3320;$$sink$i$i$i56$i95 = $3320;
        } else {
         $3321 = (((($s)) + 40|0) + ($3318)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre1$i$i54$i93 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i55$i94Z2D = $$pre1$i$i54$i93;$$sink$i$i$i56$i95 = $3321;
        }
        HEAP32[$2>>2] = $$sink$i$i$i56$i95; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i$i55$i94Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$in222$i = $$pre$phi$i$i55$i94Z2D;$3323 = $$sink$i$i$i56$i95;
       }
      } while(0);
      $3322 = ($$in222$i>>>0)<($3323>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3322) {
        $3324 = ((($$in222$i)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3324; //@line 1230 "./libraries/stb/stb_image.h"
        $$in223$i = $3324;$3340 = $3323;
       } else {
        $3325 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3326 = HEAP32[$3325>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3327 = ($3326|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3327) {
         $$in223$i = $$in222$i;$3340 = $3323;
         break;
        }
        $3328 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3329 = HEAP32[$3328>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3330 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3331 = HEAP32[$3330>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3332 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3333 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3334 = HEAP32[$3333>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3335 = (FUNCTION_TABLE_iiii[$3329 & 8191]($3331,$3332,$3334)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3336 = ($3335|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3336) {
         HEAP32[$3325>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3337 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3332>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i62$i99Z2D = $3337;$$sink$i$i4$i63$i100 = $3337;
        } else {
         $3338 = (((($s)) + 40|0) + ($3335)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre1$i2$i61$i98 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i62$i99Z2D = $$pre1$i2$i61$i98;$$sink$i$i4$i63$i100 = $3338;
        }
        HEAP32[$2>>2] = $$sink$i$i4$i63$i100; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i3$i62$i99Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$in223$i = $$pre$phi$i3$i62$i99Z2D;$3340 = $$sink$i$i4$i63$i100;
       }
      } while(0);
      $3339 = ($$in223$i>>>0)<($3340>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3339) {
        $3341 = ((($$in223$i)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3341; //@line 1230 "./libraries/stb/stb_image.h"
        $$in224$i = $3341;$3357 = $3340;
       } else {
        $3342 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3343 = HEAP32[$3342>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3344 = ($3343|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3344) {
         $$in224$i = $$in223$i;$3357 = $3340;
         break;
        }
        $3345 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3346 = HEAP32[$3345>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3347 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3348 = HEAP32[$3347>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3349 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3350 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3351 = HEAP32[$3350>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3352 = (FUNCTION_TABLE_iiii[$3346 & 8191]($3348,$3349,$3351)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3353 = ($3352|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3353) {
         HEAP32[$3342>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3354 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3349>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i69$i103Z2D = $3354;$$sink$i$i$i70$i104 = $3354;
        } else {
         $3355 = (((($s)) + 40|0) + ($3352)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre1$i$i68$i102 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i69$i103Z2D = $$pre1$i$i68$i102;$$sink$i$i$i70$i104 = $3355;
        }
        HEAP32[$2>>2] = $$sink$i$i$i70$i104; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i$i69$i103Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$in224$i = $$pre$phi$i$i69$i103Z2D;$3357 = $$sink$i$i$i70$i104;
       }
      } while(0);
      $3356 = ($$in224$i>>>0)<($3357>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3356) {
        $3358 = ((($$in224$i)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3358; //@line 1230 "./libraries/stb/stb_image.h"
        $$in225$i = $3358;$3374 = $3357;
       } else {
        $3359 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3360 = HEAP32[$3359>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3361 = ($3360|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3361) {
         $$in225$i = $$in224$i;$3374 = $3357;
         break;
        }
        $3362 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3363 = HEAP32[$3362>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3364 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3365 = HEAP32[$3364>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3366 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3367 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3368 = HEAP32[$3367>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3369 = (FUNCTION_TABLE_iiii[$3363 & 8191]($3365,$3366,$3368)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3370 = ($3369|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3370) {
         HEAP32[$3359>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3371 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3366>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i76$i108Z2D = $3371;$$sink$i$i4$i77$i109 = $3371;
        } else {
         $3372 = (((($s)) + 40|0) + ($3369)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre1$i2$i75$i107 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i76$i108Z2D = $$pre1$i2$i75$i107;$$sink$i$i4$i77$i109 = $3372;
        }
        HEAP32[$2>>2] = $$sink$i$i4$i77$i109; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i3$i76$i108Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$in225$i = $$pre$phi$i3$i76$i108Z2D;$3374 = $$sink$i$i4$i77$i109;
       }
      } while(0);
      $3373 = ($$in225$i>>>0)<($3374>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3373) {
        $3375 = ((($$in225$i)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3375; //@line 1230 "./libraries/stb/stb_image.h"
        $3376 = HEAP8[$$in225$i>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i$i86$i116 = $3376;$$in226$i = $3375;$3393 = $3374;
       } else {
        $3377 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3378 = HEAP32[$3377>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3379 = ($3378|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3379) {
         $$0$i$i86$i116 = 0;$$in226$i = $$in225$i;$3393 = $3374;
         break;
        }
        $3380 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3381 = HEAP32[$3380>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3382 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3383 = HEAP32[$3382>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3384 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3385 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3386 = HEAP32[$3385>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3387 = (FUNCTION_TABLE_iiii[$3381 & 8191]($3383,$3384,$3386)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3388 = ($3387|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3388) {
         HEAP32[$3377>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3389 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3384>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i83$i113Z2D = $3389;$$sink$i$i$i84$i114 = $3389;$3826 = 0;
        } else {
         $3390 = (((($s)) + 40|0) + ($3387)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i$i81$i111 = HEAP8[$3384>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i$i82$i112 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i83$i113Z2D = $$pre1$i$i82$i112;$$sink$i$i$i84$i114 = $3390;$3826 = $$pre$i$i81$i111;
        }
        HEAP32[$2>>2] = $$sink$i$i$i84$i114; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i$i83$i113Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i$i86$i116 = $3826;$$in226$i = $$pre$phi$i$i83$i113Z2D;$3393 = $$sink$i$i$i84$i114;
       }
      } while(0);
      $3391 = $$0$i$i86$i116&255; //@line 1308 "./libraries/stb/stb_image.h"
      $3392 = ($$in226$i>>>0)<($3393>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3392) {
        $3394 = ((($$in226$i)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3394; //@line 1230 "./libraries/stb/stb_image.h"
        $3395 = HEAP8[$$in226$i>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i6$i93$i123 = $3395;$$in227$i = $3394;$3414 = $3393;
       } else {
        $3396 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3397 = HEAP32[$3396>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3398 = ($3397|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3398) {
         $$0$i6$i93$i123 = 0;$$in227$i = $$in226$i;$3414 = $3393;
         break;
        }
        $3399 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3400 = HEAP32[$3399>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3401 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3402 = HEAP32[$3401>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3403 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3404 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3405 = HEAP32[$3404>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3406 = (FUNCTION_TABLE_iiii[$3400 & 8191]($3402,$3403,$3405)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3407 = ($3406|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3407) {
         HEAP32[$3396>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3408 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3403>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i90$i120Z2D = $3408;$$sink$i$i4$i91$i121 = $3408;$3827 = 0;
        } else {
         $3409 = (((($s)) + 40|0) + ($3406)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i1$i88$i118 = HEAP8[$3403>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i2$i89$i119 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i90$i120Z2D = $$pre1$i2$i89$i119;$$sink$i$i4$i91$i121 = $3409;$3827 = $$pre$i1$i88$i118;
        }
        HEAP32[$2>>2] = $$sink$i$i4$i91$i121; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i3$i90$i120Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i6$i93$i123 = $3827;$$in227$i = $$pre$phi$i3$i90$i120Z2D;$3414 = $$sink$i$i4$i91$i121;
       }
      } while(0);
      $3410 = $$0$i6$i93$i123&255; //@line 1309 "./libraries/stb/stb_image.h"
      $3411 = $3410 << 8; //@line 1309 "./libraries/stb/stb_image.h"
      $3412 = $3411 | $3391; //@line 1309 "./libraries/stb/stb_image.h"
      $3413 = ($$in227$i>>>0)<($3414>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3413) {
        $3415 = ((($$in227$i)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3415; //@line 1230 "./libraries/stb/stb_image.h"
        $3416 = HEAP8[$$in227$i>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i$i100$i129 = $3416;$$in228$i = $3415;$3433 = $3414;
       } else {
        $3417 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3418 = HEAP32[$3417>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3419 = ($3418|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3419) {
         $$0$i$i100$i129 = 0;$$in228$i = $$in227$i;$3433 = $3414;
         break;
        }
        $3420 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3421 = HEAP32[$3420>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3422 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3423 = HEAP32[$3422>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3424 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3425 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3426 = HEAP32[$3425>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3427 = (FUNCTION_TABLE_iiii[$3421 & 8191]($3423,$3424,$3426)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3428 = ($3427|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3428) {
         HEAP32[$3417>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3429 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3424>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i97$i126Z2D = $3429;$$sink$i$i$i98$i127 = $3429;$3828 = 0;
        } else {
         $3430 = (((($s)) + 40|0) + ($3427)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i$i95$i124 = HEAP8[$3424>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i$i96$i125 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i$i97$i126Z2D = $$pre1$i$i96$i125;$$sink$i$i$i98$i127 = $3430;$3828 = $$pre$i$i95$i124;
        }
        HEAP32[$2>>2] = $$sink$i$i$i98$i127; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i$i97$i126Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i$i100$i129 = $3828;$$in228$i = $$pre$phi$i$i97$i126Z2D;$3433 = $$sink$i$i$i98$i127;
       }
      } while(0);
      $3431 = $$0$i$i100$i129&255; //@line 1308 "./libraries/stb/stb_image.h"
      $3432 = ($$in228$i>>>0)<($3433>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3432) {
        $3434 = ((($$in228$i)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3434; //@line 1230 "./libraries/stb/stb_image.h"
        $3435 = HEAP8[$$in228$i>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i6$i107$i136 = $3435;$$in229$i = $3434;$3454 = $3433;
       } else {
        $3436 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3437 = HEAP32[$3436>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3438 = ($3437|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3438) {
         $$0$i6$i107$i136 = 0;$$in229$i = $$in228$i;$3454 = $3433;
         break;
        }
        $3439 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3440 = HEAP32[$3439>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3441 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3442 = HEAP32[$3441>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3443 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3444 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3445 = HEAP32[$3444>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3446 = (FUNCTION_TABLE_iiii[$3440 & 8191]($3442,$3443,$3445)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3447 = ($3446|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3447) {
         HEAP32[$3436>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3448 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3443>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i104$i133Z2D = $3448;$$sink$i$i4$i105$i134 = $3448;$3829 = 0;
        } else {
         $3449 = (((($s)) + 40|0) + ($3446)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i1$i102$i131 = HEAP8[$3443>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i2$i103$i132 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i3$i104$i133Z2D = $$pre1$i2$i103$i132;$$sink$i$i4$i105$i134 = $3449;$3829 = $$pre$i1$i102$i131;
        }
        HEAP32[$2>>2] = $$sink$i$i4$i105$i134; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i3$i104$i133Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i6$i107$i136 = $3829;$$in229$i = $$pre$phi$i3$i104$i133Z2D;$3454 = $$sink$i$i4$i105$i134;
       }
      } while(0);
      $3450 = $$0$i6$i107$i136&255; //@line 1309 "./libraries/stb/stb_image.h"
      $3451 = $3450 << 8; //@line 1309 "./libraries/stb/stb_image.h"
      $3452 = $3451 | $3431; //@line 1309 "./libraries/stb/stb_image.h"
      $3453 = ($$in229$i>>>0)<($3454>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3453) {
        $3455 = ((($$in229$i)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3455; //@line 1230 "./libraries/stb/stb_image.h"
        $3456 = HEAP8[$$in229$i>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i114$i142 = $3456;$$in230$i = $3455;$3473 = $3454;
       } else {
        $3457 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3458 = HEAP32[$3457>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3459 = ($3458|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3459) {
         $$0$i114$i142 = 0;$$in230$i = $$in229$i;$3473 = $3454;
         break;
        }
        $3460 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3461 = HEAP32[$3460>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3462 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3463 = HEAP32[$3462>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3464 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3465 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3466 = HEAP32[$3465>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3467 = (FUNCTION_TABLE_iiii[$3461 & 8191]($3463,$3464,$3466)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3468 = ($3467|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3468) {
         HEAP32[$3457>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3469 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3464>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i111$i139Z2D = $3469;$$sink$i$i112$i140 = $3469;$3830 = 0;
        } else {
         $3470 = (((($s)) + 40|0) + ($3467)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i109$i137 = HEAP8[$3464>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i110$i138 = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i111$i139Z2D = $$pre1$i110$i138;$$sink$i$i112$i140 = $3470;$3830 = $$pre$i109$i137;
        }
        HEAP32[$2>>2] = $$sink$i$i112$i140; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i111$i139Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i114$i142 = $3830;$$in230$i = $$pre$phi$i111$i139Z2D;$3473 = $$sink$i$i112$i140;
       }
      } while(0);
      $div$i = ($$0$i114$i142&255) >>> 3; //@line 4889 "./libraries/stb/stb_image.h"
      $3471 = $div$i&255; //@line 4889 "./libraries/stb/stb_image.h"
      $3472 = ($$in230$i>>>0)<($3473>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      do {
       if ($3472) {
        $3474 = ((($$in230$i)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $3474; //@line 1230 "./libraries/stb/stb_image.h"
        $3475 = HEAP8[$$in230$i>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i121$i = $3475;$$pre$i123$i = $3474;$3511 = $3473;
       } else {
        $3476 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $3477 = HEAP32[$3476>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $3478 = ($3477|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($3478) {
         $$0$i121$i = 0;$$pre$i123$i = $$in230$i;$3511 = $3473;
         break;
        }
        $3479 = ((($s)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3480 = HEAP32[$3479>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3481 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3482 = HEAP32[$3481>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3483 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3484 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3485 = HEAP32[$3484>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $3486 = (FUNCTION_TABLE_iiii[$3480 & 8191]($3482,$3483,$3485)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $3487 = ($3486|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($3487) {
         HEAP32[$3476>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $3488 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$3483>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i118$iZ2D = $3488;$$sink$i$i119$i = $3488;$3831 = 0;
        } else {
         $3489 = (((($s)) + 40|0) + ($3486)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i116$i = HEAP8[$3483>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i117$i = ((($s)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i118$iZ2D = $$pre1$i117$i;$$sink$i$i119$i = $3489;$3831 = $$pre$i116$i;
        }
        HEAP32[$2>>2] = $$sink$i$i119$i; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$0>>2] = $$pre$phi$i118$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i121$i = $3831;$$pre$i123$i = $$pre$phi$i118$iZ2D;$3511 = $$sink$i$i119$i;
       }
      } while(0);
      $3490 = $$pre$i123$i; //@line 1257 "./libraries/stb/stb_image.h"
      $3491 = $$0$i121$i&255; //@line 4890 "./libraries/stb/stb_image.h"
      $3492 = ($$0$i30$i65&255)>(7); //@line 4901 "./libraries/stb/stb_image.h"
      $3493 = (($3205) + -8)|0; //@line 4903 "./libraries/stb/stb_image.h"
      $$16$i = $3492 ? $3493 : $3205; //@line 4901 "./libraries/stb/stb_image.h"
      $3494 = $3491 >>> 5; //@line 4907 "./libraries/stb/stb_image.h"
      $3495 = $3494 & 1; //@line 4907 "./libraries/stb/stb_image.h"
      $3496 = ($3412|0)==(0); //@line 4911 "./libraries/stb/stb_image.h"
      $3497 = ($3452|0)==(0); //@line 4911 "./libraries/stb/stb_image.h"
      $or$cond$i144 = $3496 | $3497; //@line 4911 "./libraries/stb/stb_image.h"
      $tga_image_type$0$off$i = (($$16$i) + -1)|0; //@line 4911 "./libraries/stb/stb_image.h"
      $3498 = ($tga_image_type$0$off$i>>>0)>(2); //@line 4911 "./libraries/stb/stb_image.h"
      $3499 = $3498 | $or$cond$i144; //@line 4911 "./libraries/stb/stb_image.h"
      L1632: do {
       if ($3499) {
        $$0$i153 = 0;
       } else {
        switch ($$0$i114$i142<<24>>24) {
        case 8: case 16: case 24: case 32:  {
         break;
        }
        default: {
         $$0$i153 = 0;
         break L1632;
        }
        }
        $3500 = ($$0$i23$i58<<24>>24)!=(0); //@line 4921 "./libraries/stb/stb_image.h"
        $div14$i = ($$0$i51$i&255) >>> 3; //@line 4923 "./libraries/stb/stb_image.h"
        $3501 = $div14$i&255; //@line 4923 "./libraries/stb/stb_image.h"
        $tga_comp$0$i = $3500 ? $3501 : $3471; //@line 4921 "./libraries/stb/stb_image.h"
        HEAP32[$x>>2] = $3412; //@line 4927 "./libraries/stb/stb_image.h"
        HEAP32[$y>>2] = $3452; //@line 4928 "./libraries/stb/stb_image.h"
        $3502 = ($comp|0)==(0|0); //@line 4929 "./libraries/stb/stb_image.h"
        if (!($3502)) {
         HEAP32[$comp>>2] = $tga_comp$0$i; //@line 4929 "./libraries/stb/stb_image.h"
        }
        $3503 = Math_imul($3452, $3412)|0; //@line 4931 "./libraries/stb/stb_image.h"
        $3504 = Math_imul($tga_comp$0$i, $3503)|0; //@line 4931 "./libraries/stb/stb_image.h"
        $3505 = (_malloc($3504)|0); //@line 900 "./libraries/stb/stb_image.h"
        $3506 = ($3505|0)==(0|0); //@line 4932 "./libraries/stb/stb_image.h"
        if ($3506) {
         HEAP32[158754] = 441002; //@line 894 "./libraries/stb/stb_image.h"
         $$0$i153 = 0;
         break;
        }
        $3507 = ((($s)) + 16|0); //@line 1256 "./libraries/stb/stb_image.h"
        $3508 = HEAP32[$3507>>2]|0; //@line 1256 "./libraries/stb/stb_image.h"
        $3509 = ($3508|0)==(0|0); //@line 1256 "./libraries/stb/stb_image.h"
        do {
         if ($3509) {
          label = 1211;
         } else {
          $3510 = $3511; //@line 1257 "./libraries/stb/stb_image.h"
          $3512 = (($3510) - ($3490))|0; //@line 1257 "./libraries/stb/stb_image.h"
          $3513 = ($3168|0)>($3512|0); //@line 1258 "./libraries/stb/stb_image.h"
          if (!($3513)) {
           label = 1211;
           break;
          }
          HEAP32[$0>>2] = $3511; //@line 1259 "./libraries/stb/stb_image.h"
          $3514 = ((($s)) + 20|0); //@line 1260 "./libraries/stb/stb_image.h"
          $3515 = HEAP32[$3514>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
          $3516 = ((($s)) + 28|0); //@line 1260 "./libraries/stb/stb_image.h"
          $3517 = HEAP32[$3516>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
          $3518 = (($3168) - ($3512))|0; //@line 1260 "./libraries/stb/stb_image.h"
          FUNCTION_TABLE_vii[$3515 & 8191]($3517,$3518); //@line 1260 "./libraries/stb/stb_image.h"
         }
        } while(0);
        if ((label|0) == 1211) {
         $3519 = (($$pre$i123$i) + ($3168)|0); //@line 1264 "./libraries/stb/stb_image.h"
         HEAP32[$0>>2] = $3519; //@line 1264 "./libraries/stb/stb_image.h"
        }
        $or$cond13$i = $3500 | $3492; //@line 4937 "./libraries/stb/stb_image.h"
        do {
         if ($or$cond13$i) {
          L1649: do {
           if ($3500) {
            $3565 = HEAP32[$3507>>2]|0; //@line 1256 "./libraries/stb/stb_image.h"
            $3566 = ($3565|0)==(0|0); //@line 1256 "./libraries/stb/stb_image.h"
            do {
             if ($3566) {
              $$pre$i129$i = HEAP32[$0>>2]|0; //@line 1264 "./libraries/stb/stb_image.h"
              $3579 = $$pre$i129$i;
              label = 1233;
             } else {
              $3567 = HEAP32[$2>>2]|0; //@line 1257 "./libraries/stb/stb_image.h"
              $3568 = HEAP32[$0>>2]|0; //@line 1257 "./libraries/stb/stb_image.h"
              $3569 = $3567; //@line 1257 "./libraries/stb/stb_image.h"
              $3570 = (($3569) - ($3568))|0; //@line 1257 "./libraries/stb/stb_image.h"
              $3571 = ($3245|0)>($3570|0); //@line 1258 "./libraries/stb/stb_image.h"
              $3572 = $3568; //@line 1258 "./libraries/stb/stb_image.h"
              if (!($3571)) {
               $3579 = $3572;
               label = 1233;
               break;
              }
              HEAP32[$0>>2] = $3567; //@line 1259 "./libraries/stb/stb_image.h"
              $3573 = ((($s)) + 20|0); //@line 1260 "./libraries/stb/stb_image.h"
              $3574 = HEAP32[$3573>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
              $3575 = ((($s)) + 28|0); //@line 1260 "./libraries/stb/stb_image.h"
              $3576 = HEAP32[$3575>>2]|0; //@line 1260 "./libraries/stb/stb_image.h"
              $3577 = (($3245) - ($3570))|0; //@line 1260 "./libraries/stb/stb_image.h"
              FUNCTION_TABLE_vii[$3574 & 8191]($3576,$3577); //@line 1260 "./libraries/stb/stb_image.h"
             }
            } while(0);
            if ((label|0) == 1233) {
             $3578 = (($3579) + ($3245)|0); //@line 1264 "./libraries/stb/stb_image.h"
             HEAP32[$0>>2] = $3578; //@line 1264 "./libraries/stb/stb_image.h"
            }
            $3580 = Math_imul($3304, $3285)|0; //@line 4950 "./libraries/stb/stb_image.h"
            $3581 = $3580 >>> 3; //@line 4950 "./libraries/stb/stb_image.h"
            $3582 = (_malloc($3581)|0); //@line 900 "./libraries/stb/stb_image.h"
            $3583 = ($3582|0)==(0|0); //@line 4951 "./libraries/stb/stb_image.h"
            if ($3583) {
             _free($3505); //@line 4952 "./libraries/stb/stb_image.h"
             HEAP32[158754] = 441002; //@line 894 "./libraries/stb/stb_image.h"
             $$0$i153 = 0;
             break L1632;
            }
            $3584 = HEAP32[$3507>>2]|0; //@line 1276 "./libraries/stb/stb_image.h"
            $3585 = ($3584|0)==(0|0); //@line 1269 "./libraries/stb/stb_image.h"
            do {
             if ($3585) {
              $$pre$i135$i = HEAP32[$0>>2]|0; //@line 1283 "./libraries/stb/stb_image.h"
              $$pre3$i137$i = HEAP32[$2>>2]|0; //@line 1283 "./libraries/stb/stb_image.h"
              $3593 = $$pre$i135$i;$3595 = $$pre3$i137$i;
              label = 1239;
             } else {
              $3586 = HEAP32[$2>>2]|0; //@line 1270 "./libraries/stb/stb_image.h"
              $3587 = HEAP32[$0>>2]|0; //@line 1270 "./libraries/stb/stb_image.h"
              $3588 = $3587; //@line 1270 "./libraries/stb/stb_image.h"
              $3589 = (($3586) - ($3588))|0; //@line 1270 "./libraries/stb/stb_image.h"
              $3590 = ($3581|0)>($3589|0); //@line 1271 "./libraries/stb/stb_image.h"
              $3591 = $3586; //@line 1271 "./libraries/stb/stb_image.h"
              if (!($3590)) {
               $3593 = $3587;$3595 = $3591;
               label = 1239;
               break;
              }
              _memcpy(($3582|0),($3587|0),($3589|0))|0; //@line 1274 "./libraries/stb/stb_image.h"
              $3596 = ((($s)) + 28|0); //@line 1276 "./libraries/stb/stb_image.h"
              $3597 = HEAP32[$3596>>2]|0; //@line 1276 "./libraries/stb/stb_image.h"
              $3598 = (($3582) + ($3589)|0); //@line 1276 "./libraries/stb/stb_image.h"
              $3599 = (($3581) - ($3589))|0; //@line 1276 "./libraries/stb/stb_image.h"
              $3600 = (FUNCTION_TABLE_iiii[$3584 & 8191]($3597,$3598,$3599)|0); //@line 1276 "./libraries/stb/stb_image.h"
              $3601 = ($3600|0)==($3599|0); //@line 1277 "./libraries/stb/stb_image.h"
              $3602 = HEAP32[$2>>2]|0; //@line 1278 "./libraries/stb/stb_image.h"
              HEAP32[$0>>2] = $3602; //@line 1278 "./libraries/stb/stb_image.h"
              if ($3601) {
               $tga_palette$0$i = $3582;
               break L1649;
              }
             }
            } while(0);
            do {
             if ((label|0) == 1239) {
              $3592 = (($3593) + ($3581)|0); //@line 1283 "./libraries/stb/stb_image.h"
              $3594 = ($3592>>>0)>($3595>>>0); //@line 1283 "./libraries/stb/stb_image.h"
              if ($3594) {
               break;
              }
              _memcpy(($3582|0),($3593|0),($3581|0))|0; //@line 1284 "./libraries/stb/stb_image.h"
              HEAP32[$0>>2] = $3592; //@line 1285 "./libraries/stb/stb_image.h"
              $tga_palette$0$i = $3582;
              break L1649;
             }
            } while(0);
            _free($3505); //@line 4956 "./libraries/stb/stb_image.h"
            _free($3582); //@line 4957 "./libraries/stb/stb_image.h"
            HEAP32[158754] = 441500; //@line 894 "./libraries/stb/stb_image.h"
            $$0$i153 = 0;
            break L1632;
           } else {
            $tga_palette$0$i = 0;
           }
          } while(0);
          $3603 = ($3503|0)==(0); //@line 4962 "./libraries/stb/stb_image.h"
          if (!($3603)) {
           $3604 = ((($s)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $3605 = ((($s)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $3606 = ((($s)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $3607 = ((($s)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $3608 = ((($s)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
           $3609 = ($tga_comp$0$i|0)==(0); //@line 5013 "./libraries/stb/stb_image.h"
           $3610 = $$0$i114$i142&255; //@line 4962 "./libraries/stb/stb_image.h"
           $3611 = (($3610) + -1)|0; //@line 4962 "./libraries/stb/stb_image.h"
           $3612 = $3611 >>> 3; //@line 4962 "./libraries/stb/stb_image.h"
           $3613 = (($3612) + 1)|0; //@line 4962 "./libraries/stb/stb_image.h"
           $3614 = ($tga_comp$0$i>>>0)>(1); //@line 4962 "./libraries/stb/stb_image.h"
           $smax212$i = $3614 ? $tga_comp$0$i : 1; //@line 4962 "./libraries/stb/stb_image.h"
           $RLE_count$0194$i = 0;$RLE_repeating$0195$i = 0;$i$1192$i = 0;$read_next_pixel$0196$i = 1;
           while(1) {
            $3615 = Math_imul($i$1192$i, $tga_comp$0$i)|0; //@line 4965 "./libraries/stb/stb_image.h"
            $scevgep211$i = (($3505) + ($3615)|0); //@line 4965 "./libraries/stb/stb_image.h"
            do {
             if ($3492) {
              $3616 = ($RLE_count$0194$i|0)==(0); //@line 4967 "./libraries/stb/stb_image.h"
              if (!($3616)) {
               $3634 = ($RLE_repeating$0195$i|0)==(0); //@line 4974 "./libraries/stb/stb_image.h"
               if ($3634) {
                $RLE_count$1165$i = $RLE_count$0194$i;$RLE_repeating$1167$i = 0;
                label = 1257;
                break;
               }
               $3635 = ($read_next_pixel$0196$i|0)==(0); //@line 4983 "./libraries/stb/stb_image.h"
               if ($3635) {
                $RLE_count$1166$i = $RLE_count$0194$i;$RLE_repeating$1168$i = $RLE_repeating$0195$i;
                break;
               } else {
                $RLE_count$1165$i = $RLE_count$0194$i;$RLE_repeating$1167$i = $RLE_repeating$0195$i;
                label = 1257;
                break;
               }
              }
              $3617 = HEAP32[$0>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
              $3618 = HEAP32[$2>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
              $3619 = ($3617>>>0)<($3618>>>0); //@line 1229 "./libraries/stb/stb_image.h"
              do {
               if ($3619) {
                $3620 = ((($3617)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                HEAP32[$0>>2] = $3620; //@line 1230 "./libraries/stb/stb_image.h"
                $3621 = HEAP8[$3617>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                $$0$i148$i = $3621;
               } else {
                $3622 = HEAP32[$3604>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                $3623 = ($3622|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                if ($3623) {
                 $$0$i148$i = 0;
                 break;
                }
                $3624 = HEAP32[$3507>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                $3625 = HEAP32[$3605>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                $3626 = HEAP32[$3607>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                $3627 = (FUNCTION_TABLE_iiii[$3624 & 8191]($3625,$3606,$3626)|0); //@line 1213 "./libraries/stb/stb_image.h"
                $3628 = ($3627|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                if ($3628) {
                 HEAP32[$3604>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                 HEAP8[$3606>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                 $$sink$i$i146$i = $3608;$3832 = 0;
                } else {
                 $3629 = (((($s)) + 40|0) + ($3627)|0); //@line 1223 "./libraries/stb/stb_image.h"
                 $$pre$i143$i = HEAP8[$3606>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                 $$sink$i$i146$i = $3629;$3832 = $$pre$i143$i;
                }
                HEAP32[$2>>2] = $$sink$i$i146$i; //@line 1223 "./libraries/stb/stb_image.h"
                HEAP32[$0>>2] = $3608; //@line 1233 "./libraries/stb/stb_image.h"
                $$0$i148$i = $3832;
               }
              } while(0);
              $3630 = $$0$i148$i&255; //@line 4970 "./libraries/stb/stb_image.h"
              $3631 = $3630 & 127; //@line 4971 "./libraries/stb/stb_image.h"
              $3632 = (($3631) + 1)|0; //@line 4971 "./libraries/stb/stb_image.h"
              $3633 = $3630 >>> 7; //@line 4972 "./libraries/stb/stb_image.h"
              $RLE_count$1165$i = $3632;$RLE_repeating$1167$i = $3633;
              label = 1257;
             } else {
              $RLE_count$1165$i = $RLE_count$0194$i;$RLE_repeating$1167$i = $RLE_repeating$0195$i;
              label = 1257;
             }
            } while(0);
            do {
             if ((label|0) == 1257) {
              label = 0;
              $3636 = HEAP32[$0>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
              $3637 = HEAP32[$2>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
              if ($3500) {
               $3638 = ($3636>>>0)<($3637>>>0); //@line 1229 "./libraries/stb/stb_image.h"
               do {
                if ($3638) {
                 $3639 = ((($3636)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                 HEAP32[$0>>2] = $3639; //@line 1230 "./libraries/stb/stb_image.h"
                 $3640 = HEAP8[$3636>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                 $$0$i155$i = $3640;
                } else {
                 $3641 = HEAP32[$3604>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                 $3642 = ($3641|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                 if ($3642) {
                  $$0$i155$i = 0;
                  break;
                 }
                 $3643 = HEAP32[$3507>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                 $3644 = HEAP32[$3605>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                 $3645 = HEAP32[$3607>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                 $3646 = (FUNCTION_TABLE_iiii[$3643 & 8191]($3644,$3606,$3645)|0); //@line 1213 "./libraries/stb/stb_image.h"
                 $3647 = ($3646|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                 if ($3647) {
                  HEAP32[$3604>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                  HEAP8[$3606>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                  $$sink$i$i153$i = $3608;$3833 = 0;
                 } else {
                  $3648 = (((($s)) + 40|0) + ($3646)|0); //@line 1223 "./libraries/stb/stb_image.h"
                  $$pre$i150$i = HEAP8[$3606>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                  $$sink$i$i153$i = $3648;$3833 = $$pre$i150$i;
                 }
                 HEAP32[$2>>2] = $$sink$i$i153$i; //@line 1223 "./libraries/stb/stb_image.h"
                 HEAP32[$0>>2] = $3608; //@line 1233 "./libraries/stb/stb_image.h"
                 $$0$i155$i = $3833;
                }
               } while(0);
               $3649 = $$0$i155$i&255; //@line 4989 "./libraries/stb/stb_image.h"
               $3650 = ($3649>>>0)>=($3285>>>0); //@line 4990 "./libraries/stb/stb_image.h"
               $$$i150 = $3650 ? 0 : $3649; //@line 4990 "./libraries/stb/stb_image.h"
               $3651 = Math_imul($$$i150, $3471)|0; //@line 4998 "./libraries/stb/stb_image.h"
               $scevgep$i = (($tga_palette$0$i) + ($3651)|0); //@line 4998 "./libraries/stb/stb_image.h"
               _memcpy(($raw_data$i|0),($scevgep$i|0),($3613|0))|0; //@line 4998 "./libraries/stb/stb_image.h"
               $RLE_count$1166$i = $RLE_count$1165$i;$RLE_repeating$1168$i = $RLE_repeating$1167$i;
               break;
              } else {
               $3653 = $3636;$3654 = $3637;$j$1184$i = 0;
               while(1) {
                $3652 = ($3653>>>0)<($3654>>>0); //@line 1229 "./libraries/stb/stb_image.h"
                do {
                 if ($3652) {
                  $3655 = ((($3653)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
                  HEAP32[$0>>2] = $3655; //@line 1230 "./libraries/stb/stb_image.h"
                  $3656 = HEAP8[$3653>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
                  $$0$i162$i = $3656;$3834 = $3654;$3835 = $3655;
                 } else {
                  $3657 = HEAP32[$3604>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
                  $3658 = ($3657|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
                  if ($3658) {
                   $$0$i162$i = 0;$3834 = $3654;$3835 = $3653;
                   break;
                  }
                  $3659 = HEAP32[$3507>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $3660 = HEAP32[$3605>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $3661 = HEAP32[$3607>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
                  $3662 = (FUNCTION_TABLE_iiii[$3659 & 8191]($3660,$3606,$3661)|0); //@line 1213 "./libraries/stb/stb_image.h"
                  $3663 = ($3662|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
                  if ($3663) {
                   HEAP32[$3604>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
                   HEAP8[$3606>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
                   $$sink$i$i160$i = $3608;$3836 = 0;
                  } else {
                   $3664 = (((($s)) + 40|0) + ($3662)|0); //@line 1223 "./libraries/stb/stb_image.h"
                   $$pre$i157$i = HEAP8[$3606>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
                   $$sink$i$i160$i = $3664;$3836 = $$pre$i157$i;
                  }
                  HEAP32[$2>>2] = $$sink$i$i160$i; //@line 1223 "./libraries/stb/stb_image.h"
                  HEAP32[$0>>2] = $3608; //@line 1233 "./libraries/stb/stb_image.h"
                  $$0$i162$i = $3836;$3834 = $$sink$i$i160$i;$3835 = $3608;
                 }
                } while(0);
                $3665 = (($raw_data$i) + ($j$1184$i)|0); //@line 5005 "./libraries/stb/stb_image.h"
                HEAP8[$3665>>0] = $$0$i162$i; //@line 5005 "./libraries/stb/stb_image.h"
                $3666 = (($j$1184$i) + 1)|0; //@line 5003 "./libraries/stb/stb_image.h"
                $exitcond208$i = ($3666|0)==($3613|0); //@line 5003 "./libraries/stb/stb_image.h"
                if ($exitcond208$i) {
                 $RLE_count$1166$i = $RLE_count$1165$i;$RLE_repeating$1168$i = $RLE_repeating$1167$i;
                 break;
                } else {
                 $3653 = $3835;$3654 = $3834;$j$1184$i = $3666;
                }
               }
              }
             }
            } while(0);
            if (!($3609)) {
             _memcpy(($scevgep211$i|0),($raw_data$i|0),($smax212$i|0))|0; //@line 5014 "./libraries/stb/stb_image.h"
            }
            $3667 = (($RLE_count$1166$i) + -1)|0; //@line 5017 "./libraries/stb/stb_image.h"
            $3668 = (($i$1192$i) + 1)|0; //@line 4962 "./libraries/stb/stb_image.h"
            $3669 = ($3668|0)<($3503|0); //@line 4962 "./libraries/stb/stb_image.h"
            if ($3669) {
             $RLE_count$0194$i = $3667;$RLE_repeating$0195$i = $RLE_repeating$1168$i;$i$1192$i = $3668;$read_next_pixel$0196$i = 0;
            } else {
             break;
            }
           }
          }
          $3670 = ($3495|0)==(0); //@line 5020 "./libraries/stb/stb_image.h"
          do {
           if ($3670) {
            $3671 = Math_imul($tga_comp$0$i, $3412)|0; //@line 5024 "./libraries/stb/stb_image.h"
            $3672 = (($3452) + -1)|0; //@line 5025 "./libraries/stb/stb_image.h"
            $3673 = ($3671|0)==(0); //@line 5026 "./libraries/stb/stb_image.h"
            if ($3673) {
             break;
            }
            $3674 = ($3452>>>0)>(2); //@line 5024 "./libraries/stb/stb_image.h"
            $smax206$i = $3674 ? $3452 : 2; //@line 5024 "./libraries/stb/stb_image.h"
            $3675 = (($smax206$i) + -1)|0; //@line 5024 "./libraries/stb/stb_image.h"
            $3676 = $3675 >>> 1; //@line 5024 "./libraries/stb/stb_image.h"
            $j$3182$us$i = 0;
            while(1) {
             $3677 = Math_imul($j$3182$us$i, $3671)|0; //@line 5024 "./libraries/stb/stb_image.h"
             $3678 = (($3672) - ($j$3182$us$i))|0; //@line 5025 "./libraries/stb/stb_image.h"
             $3679 = Math_imul($3678, $3671)|0; //@line 5025 "./libraries/stb/stb_image.h"
             $i$2178$us$i = $3671;$index1$0179$us$i = $3677;$index2$0180$us$i = $3679;
             while(1) {
              $3680 = (($3505) + ($index1$0179$us$i)|0); //@line 5028 "./libraries/stb/stb_image.h"
              $3681 = HEAP8[$3680>>0]|0; //@line 5028 "./libraries/stb/stb_image.h"
              $3682 = (($3505) + ($index2$0180$us$i)|0); //@line 5029 "./libraries/stb/stb_image.h"
              $3683 = HEAP8[$3682>>0]|0; //@line 5029 "./libraries/stb/stb_image.h"
              HEAP8[$3680>>0] = $3683; //@line 5029 "./libraries/stb/stb_image.h"
              HEAP8[$3682>>0] = $3681; //@line 5030 "./libraries/stb/stb_image.h"
              $3684 = (($index1$0179$us$i) + 1)|0; //@line 5031 "./libraries/stb/stb_image.h"
              $3685 = (($index2$0180$us$i) + 1)|0; //@line 5032 "./libraries/stb/stb_image.h"
              $3686 = (($i$2178$us$i) + -1)|0; //@line 5026 "./libraries/stb/stb_image.h"
              $3687 = ($i$2178$us$i|0)>(1); //@line 5026 "./libraries/stb/stb_image.h"
              if ($3687) {
               $i$2178$us$i = $3686;$index1$0179$us$i = $3684;$index2$0180$us$i = $3685;
              } else {
               break;
              }
             }
             $3688 = (($j$3182$us$i) + 1)|0; //@line 5022 "./libraries/stb/stb_image.h"
             $exitcond207$i = ($j$3182$us$i|0)==($3676|0); //@line 5022 "./libraries/stb/stb_image.h"
             if ($exitcond207$i) {
              break;
             } else {
              $j$3182$us$i = $3688;
             }
            }
           }
          } while(0);
          $3689 = ($tga_palette$0$i|0)==(0|0); //@line 5037 "./libraries/stb/stb_image.h"
          if ($3689) {
           break;
          }
          _free($tga_palette$0$i); //@line 5039 "./libraries/stb/stb_image.h"
         } else {
          $3520 = ($3495|0)==(0); //@line 4939 "./libraries/stb/stb_image.h"
          $3521 = (($3452) + -1)|0; //@line 4939 "./libraries/stb/stb_image.h"
          $3522 = Math_imul($tga_comp$0$i, $3412)|0; //@line 4940 "./libraries/stb/stb_image.h"
          $3523 = ((($s)) + 28|0); //@line 1276 "./libraries/stb/stb_image.h"
          if ($3520) {
           $i$0200$us$i = 0;
           while(1) {
            $3524 = (($3521) - ($i$0200$us$i))|0; //@line 4939 "./libraries/stb/stb_image.h"
            $3525 = Math_imul($3524, $3522)|0; //@line 4940 "./libraries/stb/stb_image.h"
            $3526 = (($3505) + ($3525)|0); //@line 4940 "./libraries/stb/stb_image.h"
            $3527 = HEAP32[$3507>>2]|0; //@line 1276 "./libraries/stb/stb_image.h"
            $3528 = ($3527|0)==(0|0); //@line 1269 "./libraries/stb/stb_image.h"
            do {
             if ($3528) {
              $$pre$i126$us$i = HEAP32[$0>>2]|0; //@line 1283 "./libraries/stb/stb_image.h"
              $$pre3$i$us$i = HEAP32[$2>>2]|0; //@line 1283 "./libraries/stb/stb_image.h"
              $3540 = $$pre$i126$us$i;$3542 = $$pre3$i$us$i;
              label = 1218;
             } else {
              $3529 = HEAP32[$2>>2]|0; //@line 1270 "./libraries/stb/stb_image.h"
              $3530 = HEAP32[$0>>2]|0; //@line 1270 "./libraries/stb/stb_image.h"
              $3531 = $3530; //@line 1270 "./libraries/stb/stb_image.h"
              $3532 = (($3529) - ($3531))|0; //@line 1270 "./libraries/stb/stb_image.h"
              $3533 = ($3522|0)>($3532|0); //@line 1271 "./libraries/stb/stb_image.h"
              $3534 = $3529; //@line 1271 "./libraries/stb/stb_image.h"
              if (!($3533)) {
               $3540 = $3530;$3542 = $3534;
               label = 1218;
               break;
              }
              _memcpy(($3526|0),($3530|0),($3532|0))|0; //@line 1274 "./libraries/stb/stb_image.h"
              $3535 = HEAP32[$3523>>2]|0; //@line 1276 "./libraries/stb/stb_image.h"
              $3536 = (($3526) + ($3532)|0); //@line 1276 "./libraries/stb/stb_image.h"
              $3537 = (($3522) - ($3532))|0; //@line 1276 "./libraries/stb/stb_image.h"
              (FUNCTION_TABLE_iiii[$3527 & 8191]($3535,$3536,$3537)|0); //@line 1276 "./libraries/stb/stb_image.h"
              $3538 = HEAP32[$2>>2]|0; //@line 1278 "./libraries/stb/stb_image.h"
              HEAP32[$0>>2] = $3538; //@line 1278 "./libraries/stb/stb_image.h"
             }
            } while(0);
            do {
             if ((label|0) == 1218) {
              label = 0;
              $3539 = (($3540) + ($3522)|0); //@line 1283 "./libraries/stb/stb_image.h"
              $3541 = ($3539>>>0)>($3542>>>0); //@line 1283 "./libraries/stb/stb_image.h"
              if ($3541) {
               break;
              }
              _memcpy(($3526|0),($3540|0),($3522|0))|0; //@line 1284 "./libraries/stb/stb_image.h"
              HEAP32[$0>>2] = $3539; //@line 1285 "./libraries/stb/stb_image.h"
             }
            } while(0);
            $3543 = (($i$0200$us$i) + 1)|0; //@line 4938 "./libraries/stb/stb_image.h"
            $3544 = ($3452|0)>($3543|0); //@line 4938 "./libraries/stb/stb_image.h"
            if ($3544) {
             $i$0200$us$i = $3543;
            } else {
             break;
            }
           }
          } else {
           $i$0200$i = 0;
           while(1) {
            $3545 = Math_imul($i$0200$i, $3522)|0; //@line 4940 "./libraries/stb/stb_image.h"
            $3546 = (($3505) + ($3545)|0); //@line 4940 "./libraries/stb/stb_image.h"
            $3547 = HEAP32[$3507>>2]|0; //@line 1276 "./libraries/stb/stb_image.h"
            $3548 = ($3547|0)==(0|0); //@line 1269 "./libraries/stb/stb_image.h"
            do {
             if ($3548) {
              $$pre$i126$i = HEAP32[$0>>2]|0; //@line 1283 "./libraries/stb/stb_image.h"
              $$pre3$i$i = HEAP32[$2>>2]|0; //@line 1283 "./libraries/stb/stb_image.h"
              $3560 = $$pre$i126$i;$3562 = $$pre3$i$i;
              label = 1225;
             } else {
              $3549 = HEAP32[$2>>2]|0; //@line 1270 "./libraries/stb/stb_image.h"
              $3550 = HEAP32[$0>>2]|0; //@line 1270 "./libraries/stb/stb_image.h"
              $3551 = $3550; //@line 1270 "./libraries/stb/stb_image.h"
              $3552 = (($3549) - ($3551))|0; //@line 1270 "./libraries/stb/stb_image.h"
              $3553 = ($3522|0)>($3552|0); //@line 1271 "./libraries/stb/stb_image.h"
              $3554 = $3549; //@line 1271 "./libraries/stb/stb_image.h"
              if (!($3553)) {
               $3560 = $3550;$3562 = $3554;
               label = 1225;
               break;
              }
              _memcpy(($3546|0),($3550|0),($3552|0))|0; //@line 1274 "./libraries/stb/stb_image.h"
              $3555 = HEAP32[$3523>>2]|0; //@line 1276 "./libraries/stb/stb_image.h"
              $3556 = (($3546) + ($3552)|0); //@line 1276 "./libraries/stb/stb_image.h"
              $3557 = (($3522) - ($3552))|0; //@line 1276 "./libraries/stb/stb_image.h"
              (FUNCTION_TABLE_iiii[$3547 & 8191]($3555,$3556,$3557)|0); //@line 1276 "./libraries/stb/stb_image.h"
              $3558 = HEAP32[$2>>2]|0; //@line 1278 "./libraries/stb/stb_image.h"
              HEAP32[$0>>2] = $3558; //@line 1278 "./libraries/stb/stb_image.h"
             }
            } while(0);
            do {
             if ((label|0) == 1225) {
              label = 0;
              $3559 = (($3560) + ($3522)|0); //@line 1283 "./libraries/stb/stb_image.h"
              $3561 = ($3559>>>0)>($3562>>>0); //@line 1283 "./libraries/stb/stb_image.h"
              if ($3561) {
               break;
              }
              _memcpy(($3546|0),($3560|0),($3522|0))|0; //@line 1284 "./libraries/stb/stb_image.h"
              HEAP32[$0>>2] = $3559; //@line 1285 "./libraries/stb/stb_image.h"
             }
            } while(0);
            $3563 = (($i$0200$i) + 1)|0; //@line 4938 "./libraries/stb/stb_image.h"
            $3564 = ($3452|0)>($3563|0); //@line 4938 "./libraries/stb/stb_image.h"
            if ($3564) {
             $i$0200$i = $3563;
            } else {
             break;
            }
           }
          }
         }
        } while(0);
        $3690 = ($tga_comp$0$i>>>0)<(3); //@line 5044 "./libraries/stb/stb_image.h"
        $3691 = ($3503|0)==(0); //@line 5047 "./libraries/stb/stb_image.h"
        $or$cond233$i = $3691 | $3690; //@line 5044 "./libraries/stb/stb_image.h"
        if (!($or$cond233$i)) {
         $i$3176$i = 0;$tga_pixel$0177$i = $3505;
         while(1) {
          $3692 = HEAP8[$tga_pixel$0177$i>>0]|0; //@line 5049 "./libraries/stb/stb_image.h"
          $3693 = ((($tga_pixel$0177$i)) + 2|0); //@line 5050 "./libraries/stb/stb_image.h"
          $3694 = HEAP8[$3693>>0]|0; //@line 5050 "./libraries/stb/stb_image.h"
          HEAP8[$tga_pixel$0177$i>>0] = $3694; //@line 5050 "./libraries/stb/stb_image.h"
          HEAP8[$3693>>0] = $3692; //@line 5051 "./libraries/stb/stb_image.h"
          $3695 = (($tga_pixel$0177$i) + ($tga_comp$0$i)|0); //@line 5052 "./libraries/stb/stb_image.h"
          $3696 = (($i$3176$i) + 1)|0; //@line 5047 "./libraries/stb/stb_image.h"
          $3697 = ($3696|0)<($3503|0); //@line 5047 "./libraries/stb/stb_image.h"
          if ($3697) {
           $i$3176$i = $3696;$tga_pixel$0177$i = $3695;
          } else {
           break;
          }
         }
        }
        $3698 = ($req_comp|0)==(0); //@line 5057 "./libraries/stb/stb_image.h"
        $3699 = ($tga_comp$0$i|0)==($req_comp|0); //@line 5057 "./libraries/stb/stb_image.h"
        $or$cond17$i = $3698 | $3699; //@line 5057 "./libraries/stb/stb_image.h"
        if ($or$cond17$i) {
         $$0$i153 = $3505;
         break;
        }
        $3700 = (__ZL20stbi__convert_formatPhiijj($3505,$tga_comp$0$i,$req_comp,$3412,$3452)|0); //@line 5058 "./libraries/stb/stb_image.h"
        $$0$i153 = $3700;
       }
      } while(0);
      $$0 = $$0$i153;
      STACKTOP = sp;return ($$0|0); //@line 976 "./libraries/stb/stb_image.h"
     }
    } while(0);
    $3147 = HEAP32[$41>>2]|0; //@line 826 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $3147; //@line 826 "./libraries/stb/stb_image.h"
    $3148 = HEAP32[$43>>2]|0; //@line 827 "./libraries/stb/stb_image.h"
    HEAP32[$2>>2] = $3148; //@line 827 "./libraries/stb/stb_image.h"
   }
  }
 } while(0);
 HEAP32[158754] = 441011; //@line 894 "./libraries/stb/stb_image.h"
 $$0 = 0;
 STACKTOP = sp;return ($$0|0); //@line 976 "./libraries/stb/stb_image.h"
}
function __ZL16stbi__idct_blockPhiPs($out,$out_stride,$data) {
 $out = $out|0;
 $out_stride = $out_stride|0;
 $data = $data|0;
 var $$not$i = 0, $$not$i11 = 0, $$not$i14 = 0, $$not$i17 = 0, $$not$i2 = 0, $$not$i20 = 0, $$not$i5 = 0, $$not$i8 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0;
 var $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0;
 var $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $d$025 = 0, $exitcond = 0, $exitcond30 = 0, $i$029 = 0, $i$124 = 0, $o$022 = 0, $v$027 = 0, $v$123 = 0, $val = 0, $x$lobit$i = 0, $x$lobit$i1 = 0, $x$lobit$i10 = 0, $x$lobit$i13 = 0, $x$lobit$i16 = 0, $x$lobit$i19 = 0, $x$lobit$i4 = 0, $x$lobit$i7 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $val = sp;
 $d$025 = $data;$i$029 = 0;$v$027 = $val;
 while(1) {
  $0 = ((($d$025)) + 16|0); //@line 1976 "./libraries/stb/stb_image.h"
  $1 = HEAP16[$0>>1]|0; //@line 1985 "./libraries/stb/stb_image.h"
  $2 = ((($d$025)) + 32|0); //@line 1976 "./libraries/stb/stb_image.h"
  $3 = HEAP16[$2>>1]|0; //@line 1976 "./libraries/stb/stb_image.h"
  $4 = $1 | $3; //@line 1976 "./libraries/stb/stb_image.h"
  $5 = ($4<<16>>16)==(0); //@line 1976 "./libraries/stb/stb_image.h"
  if ($5) {
   $6 = ((($d$025)) + 48|0); //@line 1976 "./libraries/stb/stb_image.h"
   $7 = HEAP16[$6>>1]|0; //@line 1976 "./libraries/stb/stb_image.h"
   $8 = ($7<<16>>16)==(0); //@line 1976 "./libraries/stb/stb_image.h"
   if ($8) {
    $9 = ((($d$025)) + 64|0); //@line 1976 "./libraries/stb/stb_image.h"
    $10 = HEAP16[$9>>1]|0; //@line 1976 "./libraries/stb/stb_image.h"
    $11 = ($10<<16>>16)==(0); //@line 1976 "./libraries/stb/stb_image.h"
    if ($11) {
     $12 = ((($d$025)) + 80|0); //@line 1977 "./libraries/stb/stb_image.h"
     $13 = HEAP16[$12>>1]|0; //@line 1977 "./libraries/stb/stb_image.h"
     $14 = ($13<<16>>16)==(0); //@line 1977 "./libraries/stb/stb_image.h"
     if ($14) {
      $15 = ((($d$025)) + 96|0); //@line 1977 "./libraries/stb/stb_image.h"
      $16 = HEAP16[$15>>1]|0; //@line 1977 "./libraries/stb/stb_image.h"
      $17 = ($16<<16>>16)==(0); //@line 1977 "./libraries/stb/stb_image.h"
      if ($17) {
       $18 = ((($d$025)) + 112|0); //@line 1977 "./libraries/stb/stb_image.h"
       $19 = HEAP16[$18>>1]|0; //@line 1977 "./libraries/stb/stb_image.h"
       $20 = ($19<<16>>16)==(0); //@line 1977 "./libraries/stb/stb_image.h"
       if ($20) {
        $21 = HEAP16[$d$025>>1]|0; //@line 1982 "./libraries/stb/stb_image.h"
        $22 = $21 << 16 >> 16; //@line 1982 "./libraries/stb/stb_image.h"
        $23 = $22 << 2; //@line 1982 "./libraries/stb/stb_image.h"
        $24 = ((($v$027)) + 224|0); //@line 1983 "./libraries/stb/stb_image.h"
        HEAP32[$24>>2] = $23; //@line 1983 "./libraries/stb/stb_image.h"
        $25 = ((($v$027)) + 192|0); //@line 1983 "./libraries/stb/stb_image.h"
        HEAP32[$25>>2] = $23; //@line 1983 "./libraries/stb/stb_image.h"
        $26 = ((($v$027)) + 160|0); //@line 1983 "./libraries/stb/stb_image.h"
        HEAP32[$26>>2] = $23; //@line 1983 "./libraries/stb/stb_image.h"
        $27 = ((($v$027)) + 128|0); //@line 1983 "./libraries/stb/stb_image.h"
        HEAP32[$27>>2] = $23; //@line 1983 "./libraries/stb/stb_image.h"
        $28 = ((($v$027)) + 96|0); //@line 1983 "./libraries/stb/stb_image.h"
        HEAP32[$28>>2] = $23; //@line 1983 "./libraries/stb/stb_image.h"
        $29 = ((($v$027)) + 64|0); //@line 1983 "./libraries/stb/stb_image.h"
        HEAP32[$29>>2] = $23; //@line 1983 "./libraries/stb/stb_image.h"
        $30 = ((($v$027)) + 32|0); //@line 1983 "./libraries/stb/stb_image.h"
        HEAP32[$30>>2] = $23; //@line 1983 "./libraries/stb/stb_image.h"
        HEAP32[$v$027>>2] = $23; //@line 1983 "./libraries/stb/stb_image.h"
       } else {
        $32 = 0;
        label = 9;
       }
      } else {
       $32 = 0;
       label = 9;
      }
     } else {
      $32 = 0;
      label = 9;
     }
    } else {
     $32 = 0;
     label = 9;
    }
   } else {
    $32 = 0;
    label = 9;
   }
  } else {
   $32 = $3;
   label = 9;
  }
  if ((label|0) == 9) {
   label = 0;
   $31 = $32 << 16 >> 16; //@line 1985 "./libraries/stb/stb_image.h"
   $33 = ((($d$025)) + 96|0); //@line 1985 "./libraries/stb/stb_image.h"
   $34 = HEAP16[$33>>1]|0; //@line 1985 "./libraries/stb/stb_image.h"
   $35 = $34 << 16 >> 16; //@line 1985 "./libraries/stb/stb_image.h"
   $36 = (($35) + ($31))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $37 = ($36*2217)|0; //@line 1985 "./libraries/stb/stb_image.h"
   $38 = Math_imul($35, -7567)|0; //@line 1985 "./libraries/stb/stb_image.h"
   $39 = (($37) + ($38))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $40 = ($31*3135)|0; //@line 1985 "./libraries/stb/stb_image.h"
   $41 = (($37) + ($40))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $42 = HEAP16[$d$025>>1]|0; //@line 1985 "./libraries/stb/stb_image.h"
   $43 = $42 << 16 >> 16; //@line 1985 "./libraries/stb/stb_image.h"
   $44 = ((($d$025)) + 64|0); //@line 1985 "./libraries/stb/stb_image.h"
   $45 = HEAP16[$44>>1]|0; //@line 1985 "./libraries/stb/stb_image.h"
   $46 = $45 << 16 >> 16; //@line 1985 "./libraries/stb/stb_image.h"
   $47 = (($46) + ($43))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $48 = $47 << 12; //@line 1985 "./libraries/stb/stb_image.h"
   $49 = (($43) - ($46))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $50 = $49 << 12; //@line 1985 "./libraries/stb/stb_image.h"
   $51 = (($48) - ($41))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $52 = (($50) - ($39))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $53 = ((($d$025)) + 112|0); //@line 1985 "./libraries/stb/stb_image.h"
   $54 = HEAP16[$53>>1]|0; //@line 1985 "./libraries/stb/stb_image.h"
   $55 = $54 << 16 >> 16; //@line 1985 "./libraries/stb/stb_image.h"
   $56 = ((($d$025)) + 80|0); //@line 1985 "./libraries/stb/stb_image.h"
   $57 = HEAP16[$56>>1]|0; //@line 1985 "./libraries/stb/stb_image.h"
   $58 = $57 << 16 >> 16; //@line 1985 "./libraries/stb/stb_image.h"
   $59 = ((($d$025)) + 48|0); //@line 1985 "./libraries/stb/stb_image.h"
   $60 = HEAP16[$59>>1]|0; //@line 1985 "./libraries/stb/stb_image.h"
   $61 = $60 << 16 >> 16; //@line 1985 "./libraries/stb/stb_image.h"
   $62 = $1 << 16 >> 16; //@line 1985 "./libraries/stb/stb_image.h"
   $63 = (($61) + ($55))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $64 = (($62) + ($58))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $65 = (($62) + ($55))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $66 = (($61) + ($58))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $67 = (($64) + ($63))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $68 = ($67*4816)|0; //@line 1985 "./libraries/stb/stb_image.h"
   $69 = ($55*1223)|0; //@line 1985 "./libraries/stb/stb_image.h"
   $70 = ($58*8410)|0; //@line 1985 "./libraries/stb/stb_image.h"
   $71 = ($61*12586)|0; //@line 1985 "./libraries/stb/stb_image.h"
   $72 = ($62*6149)|0; //@line 1985 "./libraries/stb/stb_image.h"
   $73 = Math_imul($65, -3685)|0; //@line 1985 "./libraries/stb/stb_image.h"
   $74 = (($68) + ($73))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $75 = Math_imul($66, -10497)|0; //@line 1985 "./libraries/stb/stb_image.h"
   $76 = (($68) + ($75))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $77 = Math_imul($63, -8034)|0; //@line 1985 "./libraries/stb/stb_image.h"
   $78 = Math_imul($64, -1597)|0; //@line 1985 "./libraries/stb/stb_image.h"
   $79 = (($78) + ($72))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $80 = (($79) + ($74))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $81 = (($77) + ($71))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $82 = (($81) + ($76))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $83 = (($78) + ($70))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $84 = (($83) + ($76))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $85 = (($77) + ($69))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $86 = (($85) + ($74))|0; //@line 1985 "./libraries/stb/stb_image.h"
   $87 = (($41) + 512)|0; //@line 1985 "./libraries/stb/stb_image.h"
   $88 = (($87) + ($48))|0; //@line 1988 "./libraries/stb/stb_image.h"
   $89 = (($39) + 512)|0; //@line 1985 "./libraries/stb/stb_image.h"
   $90 = (($89) + ($50))|0; //@line 1988 "./libraries/stb/stb_image.h"
   $91 = (($52) + 512)|0; //@line 1988 "./libraries/stb/stb_image.h"
   $92 = (($51) + 512)|0; //@line 1988 "./libraries/stb/stb_image.h"
   $93 = (($80) + ($88))|0; //@line 1989 "./libraries/stb/stb_image.h"
   $94 = $93 >> 10; //@line 1989 "./libraries/stb/stb_image.h"
   HEAP32[$v$027>>2] = $94; //@line 1989 "./libraries/stb/stb_image.h"
   $95 = (($88) - ($80))|0; //@line 1990 "./libraries/stb/stb_image.h"
   $96 = $95 >> 10; //@line 1990 "./libraries/stb/stb_image.h"
   $97 = ((($v$027)) + 224|0); //@line 1990 "./libraries/stb/stb_image.h"
   HEAP32[$97>>2] = $96; //@line 1990 "./libraries/stb/stb_image.h"
   $98 = (($82) + ($90))|0; //@line 1991 "./libraries/stb/stb_image.h"
   $99 = $98 >> 10; //@line 1991 "./libraries/stb/stb_image.h"
   $100 = ((($v$027)) + 32|0); //@line 1991 "./libraries/stb/stb_image.h"
   HEAP32[$100>>2] = $99; //@line 1991 "./libraries/stb/stb_image.h"
   $101 = (($90) - ($82))|0; //@line 1992 "./libraries/stb/stb_image.h"
   $102 = $101 >> 10; //@line 1992 "./libraries/stb/stb_image.h"
   $103 = ((($v$027)) + 192|0); //@line 1992 "./libraries/stb/stb_image.h"
   HEAP32[$103>>2] = $102; //@line 1992 "./libraries/stb/stb_image.h"
   $104 = (($84) + ($91))|0; //@line 1993 "./libraries/stb/stb_image.h"
   $105 = $104 >> 10; //@line 1993 "./libraries/stb/stb_image.h"
   $106 = ((($v$027)) + 64|0); //@line 1993 "./libraries/stb/stb_image.h"
   HEAP32[$106>>2] = $105; //@line 1993 "./libraries/stb/stb_image.h"
   $107 = (($91) - ($84))|0; //@line 1994 "./libraries/stb/stb_image.h"
   $108 = $107 >> 10; //@line 1994 "./libraries/stb/stb_image.h"
   $109 = ((($v$027)) + 160|0); //@line 1994 "./libraries/stb/stb_image.h"
   HEAP32[$109>>2] = $108; //@line 1994 "./libraries/stb/stb_image.h"
   $110 = (($86) + ($92))|0; //@line 1995 "./libraries/stb/stb_image.h"
   $111 = $110 >> 10; //@line 1995 "./libraries/stb/stb_image.h"
   $112 = ((($v$027)) + 96|0); //@line 1995 "./libraries/stb/stb_image.h"
   HEAP32[$112>>2] = $111; //@line 1995 "./libraries/stb/stb_image.h"
   $113 = (($92) - ($86))|0; //@line 1996 "./libraries/stb/stb_image.h"
   $114 = $113 >> 10; //@line 1996 "./libraries/stb/stb_image.h"
   $115 = ((($v$027)) + 128|0); //@line 1996 "./libraries/stb/stb_image.h"
   HEAP32[$115>>2] = $114; //@line 1996 "./libraries/stb/stb_image.h"
  }
  $116 = (($i$029) + 1)|0; //@line 1974 "./libraries/stb/stb_image.h"
  $117 = ((($d$025)) + 2|0); //@line 1974 "./libraries/stb/stb_image.h"
  $118 = ((($v$027)) + 4|0); //@line 1974 "./libraries/stb/stb_image.h"
  $exitcond30 = ($116|0)==(8); //@line 1974 "./libraries/stb/stb_image.h"
  if ($exitcond30) {
   $i$124 = 0;$o$022 = $out;$v$123 = $val;
   break;
  } else {
   $d$025 = $117;$i$029 = $116;$v$027 = $118;
  }
 }
 while(1) {
  $119 = ((($v$123)) + 8|0); //@line 2002 "./libraries/stb/stb_image.h"
  $120 = HEAP32[$119>>2]|0; //@line 2002 "./libraries/stb/stb_image.h"
  $121 = ((($v$123)) + 24|0); //@line 2002 "./libraries/stb/stb_image.h"
  $122 = HEAP32[$121>>2]|0; //@line 2002 "./libraries/stb/stb_image.h"
  $123 = (($122) + ($120))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $124 = ($123*2217)|0; //@line 2002 "./libraries/stb/stb_image.h"
  $125 = Math_imul($122, -7567)|0; //@line 2002 "./libraries/stb/stb_image.h"
  $126 = (($124) + ($125))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $127 = ($120*3135)|0; //@line 2002 "./libraries/stb/stb_image.h"
  $128 = (($124) + ($127))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $129 = HEAP32[$v$123>>2]|0; //@line 2002 "./libraries/stb/stb_image.h"
  $130 = ((($v$123)) + 16|0); //@line 2002 "./libraries/stb/stb_image.h"
  $131 = HEAP32[$130>>2]|0; //@line 2002 "./libraries/stb/stb_image.h"
  $132 = (($131) + ($129))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $133 = $132 << 12; //@line 2002 "./libraries/stb/stb_image.h"
  $134 = (($129) - ($131))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $135 = $134 << 12; //@line 2002 "./libraries/stb/stb_image.h"
  $136 = (($133) - ($128))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $137 = (($135) - ($126))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $138 = ((($v$123)) + 28|0); //@line 2002 "./libraries/stb/stb_image.h"
  $139 = HEAP32[$138>>2]|0; //@line 2002 "./libraries/stb/stb_image.h"
  $140 = ((($v$123)) + 20|0); //@line 2002 "./libraries/stb/stb_image.h"
  $141 = HEAP32[$140>>2]|0; //@line 2002 "./libraries/stb/stb_image.h"
  $142 = ((($v$123)) + 12|0); //@line 2002 "./libraries/stb/stb_image.h"
  $143 = HEAP32[$142>>2]|0; //@line 2002 "./libraries/stb/stb_image.h"
  $144 = ((($v$123)) + 4|0); //@line 2002 "./libraries/stb/stb_image.h"
  $145 = HEAP32[$144>>2]|0; //@line 2002 "./libraries/stb/stb_image.h"
  $146 = (($143) + ($139))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $147 = (($145) + ($141))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $148 = (($145) + ($139))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $149 = (($143) + ($141))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $150 = (($147) + ($146))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $151 = ($150*4816)|0; //@line 2002 "./libraries/stb/stb_image.h"
  $152 = ($139*1223)|0; //@line 2002 "./libraries/stb/stb_image.h"
  $153 = ($141*8410)|0; //@line 2002 "./libraries/stb/stb_image.h"
  $154 = ($143*12586)|0; //@line 2002 "./libraries/stb/stb_image.h"
  $155 = ($145*6149)|0; //@line 2002 "./libraries/stb/stb_image.h"
  $156 = Math_imul($148, -3685)|0; //@line 2002 "./libraries/stb/stb_image.h"
  $157 = (($151) + ($156))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $158 = Math_imul($149, -10497)|0; //@line 2002 "./libraries/stb/stb_image.h"
  $159 = (($151) + ($158))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $160 = Math_imul($146, -8034)|0; //@line 2002 "./libraries/stb/stb_image.h"
  $161 = Math_imul($147, -1597)|0; //@line 2002 "./libraries/stb/stb_image.h"
  $162 = (($161) + ($155))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $163 = (($162) + ($157))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $164 = (($160) + ($154))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $165 = (($164) + ($159))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $166 = (($161) + ($153))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $167 = (($166) + ($159))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $168 = (($160) + ($152))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $169 = (($168) + ($157))|0; //@line 2002 "./libraries/stb/stb_image.h"
  $170 = (($128) + 16842752)|0; //@line 2002 "./libraries/stb/stb_image.h"
  $171 = (($170) + ($133))|0; //@line 2009 "./libraries/stb/stb_image.h"
  $172 = (($126) + 16842752)|0; //@line 2002 "./libraries/stb/stb_image.h"
  $173 = (($172) + ($135))|0; //@line 2010 "./libraries/stb/stb_image.h"
  $174 = (($137) + 16842752)|0; //@line 2011 "./libraries/stb/stb_image.h"
  $175 = (($136) + 16842752)|0; //@line 2012 "./libraries/stb/stb_image.h"
  $176 = (($163) + ($171))|0; //@line 2015 "./libraries/stb/stb_image.h"
  $177 = $176 >> 17; //@line 2015 "./libraries/stb/stb_image.h"
  $178 = ($177>>>0)>(255); //@line 1919 "./libraries/stb/stb_image.h"
  if ($178) {
   $x$lobit$i = $176 >> 31; //@line 1920 "./libraries/stb/stb_image.h"
   $179 = $x$lobit$i&255; //@line 1920 "./libraries/stb/stb_image.h"
   $$not$i = $179 ^ -1; //@line 1920 "./libraries/stb/stb_image.h"
   $181 = $$not$i;
  } else {
   $180 = $177&255; //@line 1923 "./libraries/stb/stb_image.h"
   $181 = $180;
  }
  HEAP8[$o$022>>0] = $181; //@line 2015 "./libraries/stb/stb_image.h"
  $182 = (($171) - ($163))|0; //@line 2016 "./libraries/stb/stb_image.h"
  $183 = $182 >> 17; //@line 2016 "./libraries/stb/stb_image.h"
  $184 = ($183>>>0)>(255); //@line 1919 "./libraries/stb/stb_image.h"
  if ($184) {
   $x$lobit$i19 = $182 >> 31; //@line 1920 "./libraries/stb/stb_image.h"
   $185 = $x$lobit$i19&255; //@line 1920 "./libraries/stb/stb_image.h"
   $$not$i20 = $185 ^ -1; //@line 1920 "./libraries/stb/stb_image.h"
   $188 = $$not$i20;
  } else {
   $186 = $183&255; //@line 1923 "./libraries/stb/stb_image.h"
   $188 = $186;
  }
  $187 = ((($o$022)) + 7|0); //@line 2016 "./libraries/stb/stb_image.h"
  HEAP8[$187>>0] = $188; //@line 2016 "./libraries/stb/stb_image.h"
  $189 = (($165) + ($173))|0; //@line 2017 "./libraries/stb/stb_image.h"
  $190 = $189 >> 17; //@line 2017 "./libraries/stb/stb_image.h"
  $191 = ($190>>>0)>(255); //@line 1919 "./libraries/stb/stb_image.h"
  if ($191) {
   $x$lobit$i16 = $189 >> 31; //@line 1920 "./libraries/stb/stb_image.h"
   $192 = $x$lobit$i16&255; //@line 1920 "./libraries/stb/stb_image.h"
   $$not$i17 = $192 ^ -1; //@line 1920 "./libraries/stb/stb_image.h"
   $195 = $$not$i17;
  } else {
   $193 = $190&255; //@line 1923 "./libraries/stb/stb_image.h"
   $195 = $193;
  }
  $194 = ((($o$022)) + 1|0); //@line 2017 "./libraries/stb/stb_image.h"
  HEAP8[$194>>0] = $195; //@line 2017 "./libraries/stb/stb_image.h"
  $196 = (($173) - ($165))|0; //@line 2018 "./libraries/stb/stb_image.h"
  $197 = $196 >> 17; //@line 2018 "./libraries/stb/stb_image.h"
  $198 = ($197>>>0)>(255); //@line 1919 "./libraries/stb/stb_image.h"
  if ($198) {
   $x$lobit$i13 = $196 >> 31; //@line 1920 "./libraries/stb/stb_image.h"
   $199 = $x$lobit$i13&255; //@line 1920 "./libraries/stb/stb_image.h"
   $$not$i14 = $199 ^ -1; //@line 1920 "./libraries/stb/stb_image.h"
   $202 = $$not$i14;
  } else {
   $200 = $197&255; //@line 1923 "./libraries/stb/stb_image.h"
   $202 = $200;
  }
  $201 = ((($o$022)) + 6|0); //@line 2018 "./libraries/stb/stb_image.h"
  HEAP8[$201>>0] = $202; //@line 2018 "./libraries/stb/stb_image.h"
  $203 = (($167) + ($174))|0; //@line 2019 "./libraries/stb/stb_image.h"
  $204 = $203 >> 17; //@line 2019 "./libraries/stb/stb_image.h"
  $205 = ($204>>>0)>(255); //@line 1919 "./libraries/stb/stb_image.h"
  if ($205) {
   $x$lobit$i10 = $203 >> 31; //@line 1920 "./libraries/stb/stb_image.h"
   $206 = $x$lobit$i10&255; //@line 1920 "./libraries/stb/stb_image.h"
   $$not$i11 = $206 ^ -1; //@line 1920 "./libraries/stb/stb_image.h"
   $209 = $$not$i11;
  } else {
   $207 = $204&255; //@line 1923 "./libraries/stb/stb_image.h"
   $209 = $207;
  }
  $208 = ((($o$022)) + 2|0); //@line 2019 "./libraries/stb/stb_image.h"
  HEAP8[$208>>0] = $209; //@line 2019 "./libraries/stb/stb_image.h"
  $210 = (($174) - ($167))|0; //@line 2020 "./libraries/stb/stb_image.h"
  $211 = $210 >> 17; //@line 2020 "./libraries/stb/stb_image.h"
  $212 = ($211>>>0)>(255); //@line 1919 "./libraries/stb/stb_image.h"
  if ($212) {
   $x$lobit$i7 = $210 >> 31; //@line 1920 "./libraries/stb/stb_image.h"
   $213 = $x$lobit$i7&255; //@line 1920 "./libraries/stb/stb_image.h"
   $$not$i8 = $213 ^ -1; //@line 1920 "./libraries/stb/stb_image.h"
   $216 = $$not$i8;
  } else {
   $214 = $211&255; //@line 1923 "./libraries/stb/stb_image.h"
   $216 = $214;
  }
  $215 = ((($o$022)) + 5|0); //@line 2020 "./libraries/stb/stb_image.h"
  HEAP8[$215>>0] = $216; //@line 2020 "./libraries/stb/stb_image.h"
  $217 = (($169) + ($175))|0; //@line 2021 "./libraries/stb/stb_image.h"
  $218 = $217 >> 17; //@line 2021 "./libraries/stb/stb_image.h"
  $219 = ($218>>>0)>(255); //@line 1919 "./libraries/stb/stb_image.h"
  if ($219) {
   $x$lobit$i4 = $217 >> 31; //@line 1920 "./libraries/stb/stb_image.h"
   $220 = $x$lobit$i4&255; //@line 1920 "./libraries/stb/stb_image.h"
   $$not$i5 = $220 ^ -1; //@line 1920 "./libraries/stb/stb_image.h"
   $223 = $$not$i5;
  } else {
   $221 = $218&255; //@line 1923 "./libraries/stb/stb_image.h"
   $223 = $221;
  }
  $222 = ((($o$022)) + 3|0); //@line 2021 "./libraries/stb/stb_image.h"
  HEAP8[$222>>0] = $223; //@line 2021 "./libraries/stb/stb_image.h"
  $224 = (($175) - ($169))|0; //@line 2022 "./libraries/stb/stb_image.h"
  $225 = $224 >> 17; //@line 2022 "./libraries/stb/stb_image.h"
  $226 = ($225>>>0)>(255); //@line 1919 "./libraries/stb/stb_image.h"
  if ($226) {
   $x$lobit$i1 = $224 >> 31; //@line 1920 "./libraries/stb/stb_image.h"
   $227 = $x$lobit$i1&255; //@line 1920 "./libraries/stb/stb_image.h"
   $$not$i2 = $227 ^ -1; //@line 1920 "./libraries/stb/stb_image.h"
   $230 = $$not$i2;
  } else {
   $228 = $225&255; //@line 1923 "./libraries/stb/stb_image.h"
   $230 = $228;
  }
  $229 = ((($o$022)) + 4|0); //@line 2022 "./libraries/stb/stb_image.h"
  HEAP8[$229>>0] = $230; //@line 2022 "./libraries/stb/stb_image.h"
  $231 = (($i$124) + 1)|0; //@line 2000 "./libraries/stb/stb_image.h"
  $232 = ((($v$123)) + 32|0); //@line 2000 "./libraries/stb/stb_image.h"
  $233 = (($o$022) + ($out_stride)|0); //@line 2000 "./libraries/stb/stb_image.h"
  $exitcond = ($231|0)==(8); //@line 2000 "./libraries/stb/stb_image.h"
  if ($exitcond) {
   break;
  } else {
   $i$124 = $231;$o$022 = $233;$v$123 = $232;
  }
 }
 STACKTOP = sp;return; //@line 2024 "./libraries/stb/stb_image.h"
}
function __ZL22stbi__YCbCr_to_RGB_rowPhPKhS1_S1_ii($out,$y,$pcb,$pcr,$count,$step) {
 $out = $out|0;
 $y = $y|0;
 $pcb = $pcb|0;
 $pcr = $pcr|0;
 $count = $count|0;
 $step = $step|0;
 var $$04 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b$0 = 0, $exitcond = 0, $g$0 = 0, $i$03 = 0, $r$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($count|0)>(0); //@line 3096 "./libraries/stb/stb_image.h"
 if ($0) {
  $$04 = $out;$i$03 = 0;
 } else {
  return; //@line 3116 "./libraries/stb/stb_image.h"
 }
 while(1) {
  $1 = (($y) + ($i$03)|0); //@line 3097 "./libraries/stb/stb_image.h"
  $2 = HEAP8[$1>>0]|0; //@line 3097 "./libraries/stb/stb_image.h"
  $3 = $2&255; //@line 3097 "./libraries/stb/stb_image.h"
  $4 = $3 << 20; //@line 3097 "./libraries/stb/stb_image.h"
  $5 = $4 | 524288; //@line 3097 "./libraries/stb/stb_image.h"
  $6 = (($pcr) + ($i$03)|0); //@line 3099 "./libraries/stb/stb_image.h"
  $7 = HEAP8[$6>>0]|0; //@line 3099 "./libraries/stb/stb_image.h"
  $8 = $7&255; //@line 3099 "./libraries/stb/stb_image.h"
  $9 = (($8) + -128)|0; //@line 3099 "./libraries/stb/stb_image.h"
  $10 = (($pcb) + ($i$03)|0); //@line 3100 "./libraries/stb/stb_image.h"
  $11 = HEAP8[$10>>0]|0; //@line 3100 "./libraries/stb/stb_image.h"
  $12 = $11&255; //@line 3100 "./libraries/stb/stb_image.h"
  $13 = (($12) + -128)|0; //@line 3100 "./libraries/stb/stb_image.h"
  $14 = Math_imul($9, 1470208)|0; //@line 3101 "./libraries/stb/stb_image.h"
  $15 = (($14) + ($5))|0; //@line 3101 "./libraries/stb/stb_image.h"
  $16 = Math_imul($9, -748800)|0; //@line 3102 "./libraries/stb/stb_image.h"
  $17 = (($16) + ($5))|0; //@line 3102 "./libraries/stb/stb_image.h"
  $18 = Math_imul($13, -360960)|0; //@line 3102 "./libraries/stb/stb_image.h"
  $19 = $18 & -65536; //@line 3102 "./libraries/stb/stb_image.h"
  $20 = (($17) + ($19))|0; //@line 3102 "./libraries/stb/stb_image.h"
  $21 = Math_imul($13, 1858048)|0; //@line 3103 "./libraries/stb/stb_image.h"
  $22 = (($21) + ($5))|0; //@line 3103 "./libraries/stb/stb_image.h"
  $23 = $15 >> 20; //@line 3104 "./libraries/stb/stb_image.h"
  $24 = $20 >> 20; //@line 3105 "./libraries/stb/stb_image.h"
  $25 = $22 >> 20; //@line 3106 "./libraries/stb/stb_image.h"
  $26 = ($23>>>0)>(255); //@line 3107 "./libraries/stb/stb_image.h"
  $27 = $15 >>> 31; //@line 3107 "./libraries/stb/stb_image.h"
  $28 = (($27) + 255)|0; //@line 3107 "./libraries/stb/stb_image.h"
  $r$0 = $26 ? $28 : $23; //@line 3107 "./libraries/stb/stb_image.h"
  $29 = ($24>>>0)>(255); //@line 3108 "./libraries/stb/stb_image.h"
  $30 = $20 >>> 31; //@line 3108 "./libraries/stb/stb_image.h"
  $31 = (($30) + 255)|0; //@line 3108 "./libraries/stb/stb_image.h"
  $g$0 = $29 ? $31 : $24; //@line 3108 "./libraries/stb/stb_image.h"
  $32 = ($25>>>0)>(255); //@line 3109 "./libraries/stb/stb_image.h"
  $33 = $22 >>> 31; //@line 3109 "./libraries/stb/stb_image.h"
  $34 = (($33) + 255)|0; //@line 3109 "./libraries/stb/stb_image.h"
  $b$0 = $32 ? $34 : $25; //@line 3109 "./libraries/stb/stb_image.h"
  $35 = $r$0&255; //@line 3110 "./libraries/stb/stb_image.h"
  HEAP8[$$04>>0] = $35; //@line 3110 "./libraries/stb/stb_image.h"
  $36 = $g$0&255; //@line 3111 "./libraries/stb/stb_image.h"
  $37 = ((($$04)) + 1|0); //@line 3111 "./libraries/stb/stb_image.h"
  HEAP8[$37>>0] = $36; //@line 3111 "./libraries/stb/stb_image.h"
  $38 = $b$0&255; //@line 3112 "./libraries/stb/stb_image.h"
  $39 = ((($$04)) + 2|0); //@line 3112 "./libraries/stb/stb_image.h"
  HEAP8[$39>>0] = $38; //@line 3112 "./libraries/stb/stb_image.h"
  $40 = ((($$04)) + 3|0); //@line 3113 "./libraries/stb/stb_image.h"
  HEAP8[$40>>0] = -1; //@line 3113 "./libraries/stb/stb_image.h"
  $41 = (($$04) + ($step)|0); //@line 3114 "./libraries/stb/stb_image.h"
  $42 = (($i$03) + 1)|0; //@line 3096 "./libraries/stb/stb_image.h"
  $exitcond = ($42|0)==($count|0); //@line 3096 "./libraries/stb/stb_image.h"
  if ($exitcond) {
   break;
  } else {
   $$04 = $41;$i$03 = $42;
  }
 }
 return; //@line 3116 "./libraries/stb/stb_image.h"
}
function __ZL23stbi__resample_row_hv_2PhS_S_ii($out,$in_near,$in_far,$w,$hs) {
 $out = $out|0;
 $in_near = $in_near|0;
 $in_far = $in_far|0;
 $w = $w|0;
 $hs = $hs|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $exitcond = 0, $i$01 = 0, $t1$0$lcssa = 0, $t1$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($w|0)==(1); //@line 2913 "./libraries/stb/stb_image.h"
 $1 = HEAP8[$in_near>>0]|0; //@line 2914 "./libraries/stb/stb_image.h"
 $2 = $1&255; //@line 2914 "./libraries/stb/stb_image.h"
 $3 = ($2*3)|0; //@line 2914 "./libraries/stb/stb_image.h"
 $4 = HEAP8[$in_far>>0]|0; //@line 2914 "./libraries/stb/stb_image.h"
 $5 = $4&255; //@line 2914 "./libraries/stb/stb_image.h"
 $6 = (($3) + ($5))|0; //@line 2914 "./libraries/stb/stb_image.h"
 $7 = (($6) + 2)|0; //@line 2914 "./libraries/stb/stb_image.h"
 $8 = $7 >>> 2; //@line 2914 "./libraries/stb/stb_image.h"
 $9 = $8&255; //@line 2914 "./libraries/stb/stb_image.h"
 if ($0) {
  $10 = ((($out)) + 1|0); //@line 2914 "./libraries/stb/stb_image.h"
  HEAP8[$10>>0] = $9; //@line 2914 "./libraries/stb/stb_image.h"
  HEAP8[$out>>0] = $9; //@line 2914 "./libraries/stb/stb_image.h"
  return ($out|0); //@line 2931 "./libraries/stb/stb_image.h"
 }
 HEAP8[$out>>0] = $9; //@line 2919 "./libraries/stb/stb_image.h"
 $11 = ($w|0)>(1); //@line 2920 "./libraries/stb/stb_image.h"
 if ($11) {
  $i$01 = 1;$t1$02 = $6;
  while(1) {
   $12 = (($in_near) + ($i$01)|0); //@line 2922 "./libraries/stb/stb_image.h"
   $13 = HEAP8[$12>>0]|0; //@line 2922 "./libraries/stb/stb_image.h"
   $14 = $13&255; //@line 2922 "./libraries/stb/stb_image.h"
   $15 = ($14*3)|0; //@line 2922 "./libraries/stb/stb_image.h"
   $16 = (($in_far) + ($i$01)|0); //@line 2922 "./libraries/stb/stb_image.h"
   $17 = HEAP8[$16>>0]|0; //@line 2922 "./libraries/stb/stb_image.h"
   $18 = $17&255; //@line 2922 "./libraries/stb/stb_image.h"
   $19 = (($15) + ($18))|0; //@line 2922 "./libraries/stb/stb_image.h"
   $20 = ($t1$02*3)|0; //@line 2923 "./libraries/stb/stb_image.h"
   $21 = (($20) + 8)|0; //@line 2923 "./libraries/stb/stb_image.h"
   $22 = (($21) + ($19))|0; //@line 2923 "./libraries/stb/stb_image.h"
   $23 = $22 >>> 4; //@line 2923 "./libraries/stb/stb_image.h"
   $24 = $23&255; //@line 2923 "./libraries/stb/stb_image.h"
   $25 = $i$01 << 1; //@line 2923 "./libraries/stb/stb_image.h"
   $26 = (($25) + -1)|0; //@line 2923 "./libraries/stb/stb_image.h"
   $27 = (($out) + ($26)|0); //@line 2923 "./libraries/stb/stb_image.h"
   HEAP8[$27>>0] = $24; //@line 2923 "./libraries/stb/stb_image.h"
   $28 = ($19*3)|0; //@line 2924 "./libraries/stb/stb_image.h"
   $29 = (($t1$02) + 8)|0; //@line 2924 "./libraries/stb/stb_image.h"
   $30 = (($29) + ($28))|0; //@line 2924 "./libraries/stb/stb_image.h"
   $31 = $30 >>> 4; //@line 2924 "./libraries/stb/stb_image.h"
   $32 = $31&255; //@line 2924 "./libraries/stb/stb_image.h"
   $33 = (($out) + ($25)|0); //@line 2924 "./libraries/stb/stb_image.h"
   HEAP8[$33>>0] = $32; //@line 2924 "./libraries/stb/stb_image.h"
   $34 = (($i$01) + 1)|0; //@line 2920 "./libraries/stb/stb_image.h"
   $exitcond = ($34|0)==($w|0); //@line 2920 "./libraries/stb/stb_image.h"
   if ($exitcond) {
    $t1$0$lcssa = $19;
    break;
   } else {
    $i$01 = $34;$t1$02 = $19;
   }
  }
 } else {
  $t1$0$lcssa = $6;
 }
 $35 = (($t1$0$lcssa) + 2)|0; //@line 2926 "./libraries/stb/stb_image.h"
 $36 = $35 >>> 2; //@line 2926 "./libraries/stb/stb_image.h"
 $37 = $36&255; //@line 2926 "./libraries/stb/stb_image.h"
 $38 = $w << 1; //@line 2926 "./libraries/stb/stb_image.h"
 $39 = (($38) + -1)|0; //@line 2926 "./libraries/stb/stb_image.h"
 $40 = (($out) + ($39)|0); //@line 2926 "./libraries/stb/stb_image.h"
 HEAP8[$40>>0] = $37; //@line 2926 "./libraries/stb/stb_image.h"
 return ($out|0); //@line 2931 "./libraries/stb/stb_image.h"
}
function __ZL23stbi__jpeg_decode_blockP10stbi__jpegPsP13stbi__huffmanS3_S1_iPh($j,$data,$hdc,$hac,$fac,$b,$dequant) {
 $j = $j|0;
 $data = $data|0;
 $hdc = $hdc|0;
 $hac = $hac|0;
 $fac = $fac|0;
 $b = $b|0;
 $dequant = $dequant|0;
 var $$0$i = 0, $$0$i$lcssa = 0, $$0$i17 = 0, $$0$i17$lcssa = 0, $$0$i24 = 0, $$0$i34 = 0, $$0$i34$lcssa = 0, $$0$i43 = 0, $$0$i43$lcssa = 0, $$0$i50 = 0, $$0$i57 = 0, $$0$i8 = 0, $$4 = 0, $$pre = 0, $$pre$i = 0, $$pre$i12 = 0, $$pre$i19 = 0, $$pre$i29 = 0, $$pre$i3 = 0, $$pre$i38 = 0;
 var $$pre$i45 = 0, $$pre$i52 = 0, $$pre$phi$i14Z2D = 0, $$pre$phi$i21Z2D = 0, $$pre$phi$i31Z2D = 0, $$pre$phi$i40Z2D = 0, $$pre$phi$i47Z2D = 0, $$pre$phi$i54Z2D = 0, $$pre$phi$i5Z2D = 0, $$pre$phi$iZ2D = 0, $$pre$phi67Z2D = 0, $$pre$phiZ2D = 0, $$pre1$i = 0, $$pre1$i13 = 0, $$pre1$i20 = 0, $$pre1$i30 = 0, $$pre1$i39 = 0, $$pre1$i4 = 0, $$pre1$i46 = 0, $$pre1$i53 = 0;
 var $$pre65 = 0, $$pre66 = 0, $$sink$i$i = 0, $$sink$i$i15 = 0, $$sink$i$i22 = 0, $$sink$i$i32 = 0, $$sink$i$i41 = 0, $$sink$i$i48 = 0, $$sink$i$i55 = 0, $$sink$i$i6 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0;
 var $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0;
 var $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0;
 var $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0;
 var $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0;
 var $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0;
 var $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0;
 var $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0;
 var $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0;
 var $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $k$0 = 0;
 var $k$4$ph = 0, $vararg_buffer = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 1717 "./libraries/stb/stb_image.h"
 $0 = ((($j)) + 18112|0); //@line 1722 "./libraries/stb/stb_image.h"
 $1 = HEAP32[$0>>2]|0; //@line 1722 "./libraries/stb/stb_image.h"
 $2 = ($1|0)<(16); //@line 1722 "./libraries/stb/stb_image.h"
 L1: do {
  if ($2) {
   $3 = ((($j)) + 18120|0); //@line 1590 "./libraries/stb/stb_image.h"
   $4 = ((($j)) + 18108|0); //@line 1599 "./libraries/stb/stb_image.h"
   L3: while(1) {
    $5 = HEAP32[$3>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
    $6 = ($5|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
    do {
     if ($6) {
      $7 = HEAP32[$j>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
      $8 = ((($7)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
      $9 = HEAP32[$8>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
      $10 = ((($7)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
      $11 = HEAP32[$10>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
      $12 = ($9>>>0)<($11>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      if ($12) {
       $13 = ((($9)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
       HEAP32[$8>>2] = $13; //@line 1230 "./libraries/stb/stb_image.h"
       $14 = HEAP8[$9>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
       $$0$i8 = $14;
      } else {
       $15 = ((($7)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
       $16 = HEAP32[$15>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
       $17 = ($16|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
       if ($17) {
        $58 = 0;
        break;
       }
       $18 = ((($7)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
       $19 = HEAP32[$18>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $20 = ((($7)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
       $21 = HEAP32[$20>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $22 = ((($7)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
       $23 = ((($7)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
       $24 = HEAP32[$23>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $25 = (FUNCTION_TABLE_iiii[$19 & 8191]($21,$22,$24)|0); //@line 1213 "./libraries/stb/stb_image.h"
       $26 = ($25|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
       if ($26) {
        HEAP32[$15>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
        $27 = ((($7)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
        HEAP8[$22>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
        $$pre$phi$i5Z2D = $27;$$sink$i$i6 = $27;$341 = 0;
       } else {
        $28 = (((($7)) + 40|0) + ($25)|0); //@line 1223 "./libraries/stb/stb_image.h"
        $$pre$i3 = HEAP8[$22>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
        $$pre1$i4 = ((($7)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
        $$pre$phi$i5Z2D = $$pre1$i4;$$sink$i$i6 = $28;$341 = $$pre$i3;
       }
       HEAP32[$10>>2] = $$sink$i$i6; //@line 1223 "./libraries/stb/stb_image.h"
       HEAP32[$8>>2] = $$pre$phi$i5Z2D; //@line 1233 "./libraries/stb/stb_image.h"
       $$0$i8 = $341;
      }
      $29 = $$0$i8&255; //@line 1590 "./libraries/stb/stb_image.h"
      $30 = ($$0$i8<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
      if ($30) {
       $31 = HEAP32[$j>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
       $32 = ((($31)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
       $33 = HEAP32[$32>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
       $34 = ((($31)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
       $35 = HEAP32[$34>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
       $36 = ($33>>>0)<($35>>>0); //@line 1229 "./libraries/stb/stb_image.h"
       if ($36) {
        $37 = ((($33)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$32>>2] = $37; //@line 1230 "./libraries/stb/stb_image.h"
        $38 = HEAP8[$33>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i = $38;
       } else {
        $39 = ((($31)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $40 = HEAP32[$39>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $41 = ($40|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($41) {
         $58 = 255;
         break;
        }
        $42 = ((($31)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $43 = HEAP32[$42>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $44 = ((($31)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $45 = HEAP32[$44>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $46 = ((($31)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $47 = ((($31)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $48 = HEAP32[$47>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $49 = (FUNCTION_TABLE_iiii[$43 & 8191]($45,$46,$48)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $50 = ($49|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($50) {
         HEAP32[$39>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $51 = ((($31)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$46>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$iZ2D = $51;$$sink$i$i = $51;$342 = 0;
        } else {
         $52 = (((($31)) + 40|0) + ($49)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i = HEAP8[$46>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i = ((($31)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$iZ2D = $$pre1$i;$$sink$i$i = $52;$342 = $$pre$i;
        }
        HEAP32[$34>>2] = $$sink$i$i; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$32>>2] = $$pre$phi$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i = $342;
       }
       $53 = ($$0$i<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
       if ($53) {
        $58 = 255;
       } else {
        $$0$i$lcssa = $$0$i;
        break L3;
       }
      } else {
       $58 = $29;
      }
     } else {
      $58 = 0;
     }
    } while(0);
    $55 = HEAP32[$0>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
    $56 = (24 - ($55))|0; //@line 1599 "./libraries/stb/stb_image.h"
    $57 = $58 << $56; //@line 1599 "./libraries/stb/stb_image.h"
    $59 = HEAP32[$4>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
    $60 = $57 | $59; //@line 1599 "./libraries/stb/stb_image.h"
    HEAP32[$4>>2] = $60; //@line 1599 "./libraries/stb/stb_image.h"
    $61 = (($55) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $61; //@line 1600 "./libraries/stb/stb_image.h"
    $62 = ($61|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
    if (!($62)) {
     break L1;
    }
   }
   $54 = ((($j)) + 18116|0); //@line 1594 "./libraries/stb/stb_image.h"
   HEAP8[$54>>0] = $$0$i$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
   HEAP32[$3>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
  }
 } while(0);
 $63 = (__ZL22stbi__jpeg_huff_decodeP10stbi__jpegP13stbi__huffman($j,$hdc)|0); //@line 1723 "./libraries/stb/stb_image.h"
 $64 = ($63|0)<(0); //@line 1724 "./libraries/stb/stb_image.h"
 if ($64) {
  HEAP32[158754] = 441250; //@line 894 "./libraries/stb/stb_image.h"
  $$4 = 0;
  STACKTOP = sp;return ($$4|0); //@line 1767 "./libraries/stb/stb_image.h"
 }
 dest=$data; stop=dest+128|0; do { HEAP16[dest>>1]=0|0; dest=dest+2|0; } while ((dest|0) < (stop|0)); //@line 1727 "./libraries/stb/stb_image.h"
 $65 = ($63|0)==(0); //@line 1729 "./libraries/stb/stb_image.h"
 do {
  if ($65) {
   $$pre65 = ((($j)) + 18108|0); //@line 1599 "./libraries/stb/stb_image.h"
   $$pre$phiZ2D = $$pre65;$155 = 0;
  } else {
   $66 = HEAP32[$0>>2]|0; //@line 1666 "./libraries/stb/stb_image.h"
   $67 = ($66|0)<($63|0); //@line 1666 "./libraries/stb/stb_image.h"
   L36: do {
    if ($67) {
     $68 = ((($j)) + 18120|0); //@line 1590 "./libraries/stb/stb_image.h"
     $69 = ((($j)) + 18108|0); //@line 1599 "./libraries/stb/stb_image.h"
     L38: while(1) {
      $70 = HEAP32[$68>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
      $71 = ($70|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
      do {
       if ($71) {
        $72 = HEAP32[$j>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
        $73 = ((($72)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
        $74 = HEAP32[$73>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
        $75 = ((($72)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
        $76 = HEAP32[$75>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
        $77 = ($74>>>0)<($76>>>0); //@line 1229 "./libraries/stb/stb_image.h"
        if ($77) {
         $78 = ((($74)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
         HEAP32[$73>>2] = $78; //@line 1230 "./libraries/stb/stb_image.h"
         $79 = HEAP8[$74>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
         $$0$i57 = $79;
        } else {
         $80 = ((($72)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
         $81 = HEAP32[$80>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
         $82 = ($81|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
         if ($82) {
          $123 = 0;
          break;
         }
         $83 = ((($72)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
         $84 = HEAP32[$83>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $85 = ((($72)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
         $86 = HEAP32[$85>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $87 = ((($72)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
         $88 = ((($72)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
         $89 = HEAP32[$88>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $90 = (FUNCTION_TABLE_iiii[$84 & 8191]($86,$87,$89)|0); //@line 1213 "./libraries/stb/stb_image.h"
         $91 = ($90|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
         if ($91) {
          HEAP32[$80>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
          $92 = ((($72)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
          HEAP8[$87>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
          $$pre$phi$i54Z2D = $92;$$sink$i$i55 = $92;$343 = 0;
         } else {
          $93 = (((($72)) + 40|0) + ($90)|0); //@line 1223 "./libraries/stb/stb_image.h"
          $$pre$i52 = HEAP8[$87>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
          $$pre1$i53 = ((($72)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
          $$pre$phi$i54Z2D = $$pre1$i53;$$sink$i$i55 = $93;$343 = $$pre$i52;
         }
         HEAP32[$75>>2] = $$sink$i$i55; //@line 1223 "./libraries/stb/stb_image.h"
         HEAP32[$73>>2] = $$pre$phi$i54Z2D; //@line 1233 "./libraries/stb/stb_image.h"
         $$0$i57 = $343;
        }
        $94 = $$0$i57&255; //@line 1590 "./libraries/stb/stb_image.h"
        $95 = ($$0$i57<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
        if ($95) {
         $96 = HEAP32[$j>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
         $97 = ((($96)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
         $98 = HEAP32[$97>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $99 = ((($96)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
         $100 = HEAP32[$99>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $101 = ($98>>>0)<($100>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         if ($101) {
          $102 = ((($98)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
          HEAP32[$97>>2] = $102; //@line 1230 "./libraries/stb/stb_image.h"
          $103 = HEAP8[$98>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
          $$0$i43 = $103;
         } else {
          $104 = ((($96)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
          $105 = HEAP32[$104>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
          $106 = ($105|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
          if ($106) {
           $123 = 255;
           break;
          }
          $107 = ((($96)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
          $108 = HEAP32[$107>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
          $109 = ((($96)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
          $110 = HEAP32[$109>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
          $111 = ((($96)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
          $112 = ((($96)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
          $113 = HEAP32[$112>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
          $114 = (FUNCTION_TABLE_iiii[$108 & 8191]($110,$111,$113)|0); //@line 1213 "./libraries/stb/stb_image.h"
          $115 = ($114|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
          if ($115) {
           HEAP32[$104>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
           $116 = ((($96)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
           HEAP8[$111>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
           $$pre$phi$i40Z2D = $116;$$sink$i$i41 = $116;$344 = 0;
          } else {
           $117 = (((($96)) + 40|0) + ($114)|0); //@line 1223 "./libraries/stb/stb_image.h"
           $$pre$i38 = HEAP8[$111>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
           $$pre1$i39 = ((($96)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
           $$pre$phi$i40Z2D = $$pre1$i39;$$sink$i$i41 = $117;$344 = $$pre$i38;
          }
          HEAP32[$99>>2] = $$sink$i$i41; //@line 1223 "./libraries/stb/stb_image.h"
          HEAP32[$97>>2] = $$pre$phi$i40Z2D; //@line 1233 "./libraries/stb/stb_image.h"
          $$0$i43 = $344;
         }
         $118 = ($$0$i43<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
         if ($118) {
          $123 = 255;
         } else {
          $$0$i43$lcssa = $$0$i43;
          break L38;
         }
        } else {
         $123 = $94;
        }
       } else {
        $123 = 0;
       }
      } while(0);
      $120 = HEAP32[$0>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
      $121 = (24 - ($120))|0; //@line 1599 "./libraries/stb/stb_image.h"
      $122 = $123 << $121; //@line 1599 "./libraries/stb/stb_image.h"
      $124 = HEAP32[$69>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
      $125 = $122 | $124; //@line 1599 "./libraries/stb/stb_image.h"
      HEAP32[$69>>2] = $125; //@line 1599 "./libraries/stb/stb_image.h"
      $126 = (($120) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
      HEAP32[$0>>2] = $126; //@line 1600 "./libraries/stb/stb_image.h"
      $127 = ($126|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
      if (!($127)) {
       $$pre$phi67Z2D = $69;
       break L36;
      }
     }
     $119 = ((($j)) + 18116|0); //@line 1594 "./libraries/stb/stb_image.h"
     HEAP8[$119>>0] = $$0$i43$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
     HEAP32[$68>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
     $$pre$phi67Z2D = $69;
    } else {
     $$pre66 = ((($j)) + 18108|0); //@line 1668 "./libraries/stb/stb_image.h"
     $$pre$phi67Z2D = $$pre66;
    }
   } while(0);
   $128 = HEAP32[$$pre$phi67Z2D>>2]|0; //@line 1668 "./libraries/stb/stb_image.h"
   $129 = $128 << $63; //@line 1669 "./libraries/stb/stb_image.h"
   $130 = (32 - ($63))|0; //@line 1669 "./libraries/stb/stb_image.h"
   $131 = $128 >>> $130; //@line 1669 "./libraries/stb/stb_image.h"
   $132 = $129 | $131; //@line 1669 "./libraries/stb/stb_image.h"
   $133 = ($63>>>0)<(17); //@line 1670 "./libraries/stb/stb_image.h"
   if ($133) {
    $139 = $128 >> 31; //@line 1668 "./libraries/stb/stb_image.h"
    $140 = (67056 + ($63<<2)|0); //@line 1671 "./libraries/stb/stb_image.h"
    $141 = HEAP32[$140>>2]|0; //@line 1671 "./libraries/stb/stb_image.h"
    $142 = $141 ^ -1; //@line 1671 "./libraries/stb/stb_image.h"
    $143 = $132 & $142; //@line 1671 "./libraries/stb/stb_image.h"
    HEAP32[$$pre$phi67Z2D>>2] = $143; //@line 1671 "./libraries/stb/stb_image.h"
    $144 = $141 & $132; //@line 1672 "./libraries/stb/stb_image.h"
    $145 = HEAP32[$0>>2]|0; //@line 1673 "./libraries/stb/stb_image.h"
    $146 = (($145) - ($63))|0; //@line 1673 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $146; //@line 1673 "./libraries/stb/stb_image.h"
    $147 = (26004 + ($63<<2)|0); //@line 1674 "./libraries/stb/stb_image.h"
    $148 = HEAP32[$147>>2]|0; //@line 1674 "./libraries/stb/stb_image.h"
    $149 = $139 ^ -1; //@line 1674 "./libraries/stb/stb_image.h"
    $150 = $148 & $149; //@line 1674 "./libraries/stb/stb_image.h"
    $151 = (($150) + ($144))|0; //@line 1674 "./libraries/stb/stb_image.h"
    $$pre$phiZ2D = $$pre$phi67Z2D;$155 = $151;
    break;
   }
   $134 = (___cxa_allocate_exception(16)|0); //@line 27 "modules/image/magpie/STBHandler.cpp"
   __THREW__ = 0;
   HEAP32[$vararg_buffer>>2] = 441267; //@line 27 "modules/image/magpie/STBHandler.cpp"
   invoke_viii(3349,($134|0),(441193|0),($vararg_buffer|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
   $135 = __THREW__; __THREW__ = 0;
   $136 = $135&1;
   if ($136) {
    $137 = ___cxa_find_matching_catch_2()|0;
    $138 = tempRet0;
    ___cxa_free_exception(($134|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
    ___resumeException($137|0);
    // unreachable;
   } else {
    ___cxa_throw(($134|0),(96|0),(344|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
    // unreachable; //@line 27 "modules/image/magpie/STBHandler.cpp"
   }
  }
 } while(0);
 $152 = (((((($j)) + 17820|0) + (($b*72)|0)|0)) + 24|0); //@line 1730 "./libraries/stb/stb_image.h"
 $153 = HEAP32[$152>>2]|0; //@line 1730 "./libraries/stb/stb_image.h"
 $154 = (($153) + ($155))|0; //@line 1730 "./libraries/stb/stb_image.h"
 HEAP32[$152>>2] = $154; //@line 1731 "./libraries/stb/stb_image.h"
 $156 = HEAP8[$dequant>>0]|0; //@line 1732 "./libraries/stb/stb_image.h"
 $157 = $156&255; //@line 1732 "./libraries/stb/stb_image.h"
 $158 = Math_imul($157, $154)|0; //@line 1732 "./libraries/stb/stb_image.h"
 $159 = $158&65535; //@line 1732 "./libraries/stb/stb_image.h"
 HEAP16[$data>>1] = $159; //@line 1732 "./libraries/stb/stb_image.h"
 $160 = ((($j)) + 18120|0); //@line 1590 "./libraries/stb/stb_image.h"
 $161 = ((($j)) + 18116|0); //@line 1594 "./libraries/stb/stb_image.h"
 $k$0 = 1;
 L72: while(1) {
  $162 = HEAP32[$0>>2]|0; //@line 1739 "./libraries/stb/stb_image.h"
  $163 = ($162|0)<(16); //@line 1739 "./libraries/stb/stb_image.h"
  L74: do {
   if ($163) {
    L75: while(1) {
     $164 = HEAP32[$160>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
     $165 = ($164|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
     do {
      if ($165) {
       $166 = HEAP32[$j>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
       $167 = ((($166)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
       $168 = HEAP32[$167>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
       $169 = ((($166)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
       $170 = HEAP32[$169>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
       $171 = ($168>>>0)<($170>>>0); //@line 1229 "./libraries/stb/stb_image.h"
       if ($171) {
        $172 = ((($168)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$167>>2] = $172; //@line 1230 "./libraries/stb/stb_image.h"
        $173 = HEAP8[$168>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i24 = $173;
       } else {
        $174 = ((($166)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $175 = HEAP32[$174>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $176 = ($175|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($176) {
         $216 = 0;
         break;
        }
        $177 = ((($166)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $178 = HEAP32[$177>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $179 = ((($166)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $180 = HEAP32[$179>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $181 = ((($166)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $182 = ((($166)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $183 = HEAP32[$182>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $184 = (FUNCTION_TABLE_iiii[$178 & 8191]($180,$181,$183)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $185 = ($184|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($185) {
         HEAP32[$174>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $186 = ((($166)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$181>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i21Z2D = $186;$$sink$i$i22 = $186;$345 = 0;
        } else {
         $187 = (((($166)) + 40|0) + ($184)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i19 = HEAP8[$181>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i20 = ((($166)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i21Z2D = $$pre1$i20;$$sink$i$i22 = $187;$345 = $$pre$i19;
        }
        HEAP32[$169>>2] = $$sink$i$i22; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$167>>2] = $$pre$phi$i21Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i24 = $345;
       }
       $188 = $$0$i24&255; //@line 1590 "./libraries/stb/stb_image.h"
       $189 = ($$0$i24<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
       if ($189) {
        $190 = HEAP32[$j>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
        $191 = ((($190)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
        $192 = HEAP32[$191>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
        $193 = ((($190)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
        $194 = HEAP32[$193>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
        $195 = ($192>>>0)<($194>>>0); //@line 1229 "./libraries/stb/stb_image.h"
        if ($195) {
         $196 = ((($192)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
         HEAP32[$191>>2] = $196; //@line 1230 "./libraries/stb/stb_image.h"
         $197 = HEAP8[$192>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
         $$0$i17 = $197;
        } else {
         $198 = ((($190)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
         $199 = HEAP32[$198>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
         $200 = ($199|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
         if ($200) {
          $216 = 255;
          break;
         }
         $201 = ((($190)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
         $202 = HEAP32[$201>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $203 = ((($190)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
         $204 = HEAP32[$203>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $205 = ((($190)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
         $206 = ((($190)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
         $207 = HEAP32[$206>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $208 = (FUNCTION_TABLE_iiii[$202 & 8191]($204,$205,$207)|0); //@line 1213 "./libraries/stb/stb_image.h"
         $209 = ($208|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
         if ($209) {
          HEAP32[$198>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
          $210 = ((($190)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
          HEAP8[$205>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
          $$pre$phi$i14Z2D = $210;$$sink$i$i15 = $210;$346 = 0;
         } else {
          $211 = (((($190)) + 40|0) + ($208)|0); //@line 1223 "./libraries/stb/stb_image.h"
          $$pre$i12 = HEAP8[$205>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
          $$pre1$i13 = ((($190)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
          $$pre$phi$i14Z2D = $$pre1$i13;$$sink$i$i15 = $211;$346 = $$pre$i12;
         }
         HEAP32[$193>>2] = $$sink$i$i15; //@line 1223 "./libraries/stb/stb_image.h"
         HEAP32[$191>>2] = $$pre$phi$i14Z2D; //@line 1233 "./libraries/stb/stb_image.h"
         $$0$i17 = $346;
        }
        $212 = ($$0$i17<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
        if ($212) {
         $216 = 255;
        } else {
         $$0$i17$lcssa = $$0$i17;
         break L75;
        }
       } else {
        $216 = $188;
       }
      } else {
       $216 = 0;
      }
     } while(0);
     $213 = HEAP32[$0>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
     $214 = (24 - ($213))|0; //@line 1599 "./libraries/stb/stb_image.h"
     $215 = $216 << $214; //@line 1599 "./libraries/stb/stb_image.h"
     $217 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
     $218 = $215 | $217; //@line 1599 "./libraries/stb/stb_image.h"
     HEAP32[$$pre$phiZ2D>>2] = $218; //@line 1599 "./libraries/stb/stb_image.h"
     $219 = (($213) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
     HEAP32[$0>>2] = $219; //@line 1600 "./libraries/stb/stb_image.h"
     $220 = ($219|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
     if (!($220)) {
      break L74;
     }
    }
    HEAP8[$161>>0] = $$0$i17$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
    HEAP32[$160>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
   }
  } while(0);
  $221 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 1740 "./libraries/stb/stb_image.h"
  $222 = $221 >>> 23; //@line 1740 "./libraries/stb/stb_image.h"
  $223 = (($fac) + ($222<<1)|0); //@line 1741 "./libraries/stb/stb_image.h"
  $224 = HEAP16[$223>>1]|0; //@line 1741 "./libraries/stb/stb_image.h"
  $225 = $224 << 16 >> 16; //@line 1741 "./libraries/stb/stb_image.h"
  $226 = ($224<<16>>16)==(0); //@line 1742 "./libraries/stb/stb_image.h"
  do {
   if ($226) {
    $245 = (__ZL22stbi__jpeg_huff_decodeP10stbi__jpegP13stbi__huffman($j,$hac)|0); //@line 1751 "./libraries/stb/stb_image.h"
    $246 = ($245|0)<(0); //@line 1752 "./libraries/stb/stb_image.h"
    if ($246) {
     label = 77;
     break L72;
    }
    $247 = $245 & 15; //@line 1753 "./libraries/stb/stb_image.h"
    $248 = ($247|0)==(0); //@line 1755 "./libraries/stb/stb_image.h"
    if ($248) {
     $249 = ($245|0)==(240); //@line 1756 "./libraries/stb/stb_image.h"
     $250 = (($k$0) + 16)|0; //@line 1757 "./libraries/stb/stb_image.h"
     if ($249) {
      $k$4$ph = $250;
      break;
     } else {
      $$4 = 1;
      label = 102;
      break L72;
     }
    }
    $251 = $245 >> 4; //@line 1754 "./libraries/stb/stb_image.h"
    $252 = (($251) + ($k$0))|0; //@line 1759 "./libraries/stb/stb_image.h"
    $253 = (($252) + 1)|0; //@line 1761 "./libraries/stb/stb_image.h"
    $254 = (440770 + ($252)|0); //@line 1761 "./libraries/stb/stb_image.h"
    $255 = HEAP8[$254>>0]|0; //@line 1761 "./libraries/stb/stb_image.h"
    $256 = $255&255; //@line 1761 "./libraries/stb/stb_image.h"
    $257 = HEAP32[$0>>2]|0; //@line 1666 "./libraries/stb/stb_image.h"
    $258 = ($257|0)<($247|0); //@line 1666 "./libraries/stb/stb_image.h"
    L107: do {
     if ($258) {
      L108: while(1) {
       $259 = HEAP32[$160>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
       $260 = ($259|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
       do {
        if ($260) {
         $261 = HEAP32[$j>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
         $262 = ((($261)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
         $263 = HEAP32[$262>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $264 = ((($261)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
         $265 = HEAP32[$264>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $266 = ($263>>>0)<($265>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         if ($266) {
          $267 = ((($263)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
          HEAP32[$262>>2] = $267; //@line 1230 "./libraries/stb/stb_image.h"
          $268 = HEAP8[$263>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
          $$0$i50 = $268;
         } else {
          $269 = ((($261)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
          $270 = HEAP32[$269>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
          $271 = ($270|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
          if ($271) {
           $311 = 0;
           break;
          }
          $272 = ((($261)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
          $273 = HEAP32[$272>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
          $274 = ((($261)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
          $275 = HEAP32[$274>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
          $276 = ((($261)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
          $277 = ((($261)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
          $278 = HEAP32[$277>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
          $279 = (FUNCTION_TABLE_iiii[$273 & 8191]($275,$276,$278)|0); //@line 1213 "./libraries/stb/stb_image.h"
          $280 = ($279|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
          if ($280) {
           HEAP32[$269>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
           $281 = ((($261)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
           HEAP8[$276>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
           $$pre$phi$i47Z2D = $281;$$sink$i$i48 = $281;$347 = 0;
          } else {
           $282 = (((($261)) + 40|0) + ($279)|0); //@line 1223 "./libraries/stb/stb_image.h"
           $$pre$i45 = HEAP8[$276>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
           $$pre1$i46 = ((($261)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
           $$pre$phi$i47Z2D = $$pre1$i46;$$sink$i$i48 = $282;$347 = $$pre$i45;
          }
          HEAP32[$264>>2] = $$sink$i$i48; //@line 1223 "./libraries/stb/stb_image.h"
          HEAP32[$262>>2] = $$pre$phi$i47Z2D; //@line 1233 "./libraries/stb/stb_image.h"
          $$0$i50 = $347;
         }
         $283 = $$0$i50&255; //@line 1590 "./libraries/stb/stb_image.h"
         $284 = ($$0$i50<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
         if ($284) {
          $285 = HEAP32[$j>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
          $286 = ((($285)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
          $287 = HEAP32[$286>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $288 = ((($285)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
          $289 = HEAP32[$288>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
          $290 = ($287>>>0)<($289>>>0); //@line 1229 "./libraries/stb/stb_image.h"
          if ($290) {
           $291 = ((($287)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
           HEAP32[$286>>2] = $291; //@line 1230 "./libraries/stb/stb_image.h"
           $292 = HEAP8[$287>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
           $$0$i34 = $292;
          } else {
           $293 = ((($285)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
           $294 = HEAP32[$293>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
           $295 = ($294|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
           if ($295) {
            $311 = 255;
            break;
           }
           $296 = ((($285)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
           $297 = HEAP32[$296>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $298 = ((($285)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
           $299 = HEAP32[$298>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $300 = ((($285)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
           $301 = ((($285)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
           $302 = HEAP32[$301>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
           $303 = (FUNCTION_TABLE_iiii[$297 & 8191]($299,$300,$302)|0); //@line 1213 "./libraries/stb/stb_image.h"
           $304 = ($303|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
           if ($304) {
            HEAP32[$293>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
            $305 = ((($285)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
            HEAP8[$300>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
            $$pre$phi$i31Z2D = $305;$$sink$i$i32 = $305;$348 = 0;
           } else {
            $306 = (((($285)) + 40|0) + ($303)|0); //@line 1223 "./libraries/stb/stb_image.h"
            $$pre$i29 = HEAP8[$300>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
            $$pre1$i30 = ((($285)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
            $$pre$phi$i31Z2D = $$pre1$i30;$$sink$i$i32 = $306;$348 = $$pre$i29;
           }
           HEAP32[$288>>2] = $$sink$i$i32; //@line 1223 "./libraries/stb/stb_image.h"
           HEAP32[$286>>2] = $$pre$phi$i31Z2D; //@line 1233 "./libraries/stb/stb_image.h"
           $$0$i34 = $348;
          }
          $307 = ($$0$i34<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
          if ($307) {
           $311 = 255;
          } else {
           $$0$i34$lcssa = $$0$i34;
           break L108;
          }
         } else {
          $311 = $283;
         }
        } else {
         $311 = 0;
        }
       } while(0);
       $308 = HEAP32[$0>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
       $309 = (24 - ($308))|0; //@line 1599 "./libraries/stb/stb_image.h"
       $310 = $311 << $309; //@line 1599 "./libraries/stb/stb_image.h"
       $312 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
       $313 = $310 | $312; //@line 1599 "./libraries/stb/stb_image.h"
       HEAP32[$$pre$phiZ2D>>2] = $313; //@line 1599 "./libraries/stb/stb_image.h"
       $314 = (($308) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
       HEAP32[$0>>2] = $314; //@line 1600 "./libraries/stb/stb_image.h"
       $315 = ($314|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
       if (!($315)) {
        $328 = $314;
        break L107;
       }
      }
      HEAP8[$161>>0] = $$0$i34$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
      HEAP32[$160>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
      $$pre = HEAP32[$0>>2]|0; //@line 1673 "./libraries/stb/stb_image.h"
      $328 = $$pre;
     } else {
      $328 = $257;
     }
    } while(0);
    $316 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 1668 "./libraries/stb/stb_image.h"
    $317 = $316 << $247; //@line 1669 "./libraries/stb/stb_image.h"
    $318 = (32 - ($247))|0; //@line 1669 "./libraries/stb/stb_image.h"
    $319 = $316 >>> $318; //@line 1669 "./libraries/stb/stb_image.h"
    $320 = $317 | $319; //@line 1669 "./libraries/stb/stb_image.h"
    $321 = $316 >> 31; //@line 1668 "./libraries/stb/stb_image.h"
    $322 = (67056 + ($247<<2)|0); //@line 1671 "./libraries/stb/stb_image.h"
    $323 = HEAP32[$322>>2]|0; //@line 1671 "./libraries/stb/stb_image.h"
    $324 = $323 ^ -1; //@line 1671 "./libraries/stb/stb_image.h"
    $325 = $320 & $324; //@line 1671 "./libraries/stb/stb_image.h"
    HEAP32[$$pre$phiZ2D>>2] = $325; //@line 1671 "./libraries/stb/stb_image.h"
    $326 = $320 & $323; //@line 1672 "./libraries/stb/stb_image.h"
    $327 = (($328) - ($247))|0; //@line 1673 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $327; //@line 1673 "./libraries/stb/stb_image.h"
    $329 = (26004 + ($247<<2)|0); //@line 1674 "./libraries/stb/stb_image.h"
    $330 = HEAP32[$329>>2]|0; //@line 1674 "./libraries/stb/stb_image.h"
    $331 = $321 ^ -1; //@line 1674 "./libraries/stb/stb_image.h"
    $332 = $330 & $331; //@line 1674 "./libraries/stb/stb_image.h"
    $333 = (($332) + ($326))|0; //@line 1674 "./libraries/stb/stb_image.h"
    $334 = (($dequant) + ($256)|0); //@line 1762 "./libraries/stb/stb_image.h"
    $335 = HEAP8[$334>>0]|0; //@line 1762 "./libraries/stb/stb_image.h"
    $336 = $335&255; //@line 1762 "./libraries/stb/stb_image.h"
    $337 = Math_imul($333, $336)|0; //@line 1762 "./libraries/stb/stb_image.h"
    $338 = $337&65535; //@line 1762 "./libraries/stb/stb_image.h"
    $339 = (($data) + ($256<<1)|0); //@line 1762 "./libraries/stb/stb_image.h"
    HEAP16[$339>>1] = $338; //@line 1762 "./libraries/stb/stb_image.h"
    $k$4$ph = $253;
   } else {
    $227 = $225 >>> 4; //@line 1743 "./libraries/stb/stb_image.h"
    $228 = $227 & 15; //@line 1743 "./libraries/stb/stb_image.h"
    $229 = (($228) + ($k$0))|0; //@line 1743 "./libraries/stb/stb_image.h"
    $230 = $225 & 15; //@line 1744 "./libraries/stb/stb_image.h"
    $231 = $221 << $230; //@line 1745 "./libraries/stb/stb_image.h"
    HEAP32[$$pre$phiZ2D>>2] = $231; //@line 1745 "./libraries/stb/stb_image.h"
    $232 = HEAP32[$0>>2]|0; //@line 1746 "./libraries/stb/stb_image.h"
    $233 = (($232) - ($230))|0; //@line 1746 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $233; //@line 1746 "./libraries/stb/stb_image.h"
    $234 = (($229) + 1)|0; //@line 1748 "./libraries/stb/stb_image.h"
    $235 = (440770 + ($229)|0); //@line 1748 "./libraries/stb/stb_image.h"
    $236 = HEAP8[$235>>0]|0; //@line 1748 "./libraries/stb/stb_image.h"
    $237 = $236&255; //@line 1748 "./libraries/stb/stb_image.h"
    $238 = $225 >> 8; //@line 1749 "./libraries/stb/stb_image.h"
    $239 = (($dequant) + ($237)|0); //@line 1749 "./libraries/stb/stb_image.h"
    $240 = HEAP8[$239>>0]|0; //@line 1749 "./libraries/stb/stb_image.h"
    $241 = $240&255; //@line 1749 "./libraries/stb/stb_image.h"
    $242 = Math_imul($241, $238)|0; //@line 1749 "./libraries/stb/stb_image.h"
    $243 = $242&65535; //@line 1749 "./libraries/stb/stb_image.h"
    $244 = (($data) + ($237<<1)|0); //@line 1749 "./libraries/stb/stb_image.h"
    HEAP16[$244>>1] = $243; //@line 1749 "./libraries/stb/stb_image.h"
    $k$4$ph = $234;
   }
  } while(0);
  $340 = ($k$4$ph|0)<(64); //@line 1765 "./libraries/stb/stb_image.h"
  if ($340) {
   $k$0 = $k$4$ph;
  } else {
   $$4 = 1;
   label = 102;
   break;
  }
 }
 if ((label|0) == 77) {
  HEAP32[158754] = 441250; //@line 894 "./libraries/stb/stb_image.h"
  $$4 = 0;
  STACKTOP = sp;return ($$4|0); //@line 1767 "./libraries/stb/stb_image.h"
 }
 else if ((label|0) == 102) {
  STACKTOP = sp;return ($$4|0); //@line 1767 "./libraries/stb/stb_image.h"
 }
 return (0)|0;
}
function __ZL22stbi__jpeg_huff_decodeP10stbi__jpegP13stbi__huffman($j,$h) {
 $j = $j|0;
 $h = $h|0;
 var $$0$i = 0, $$0$i$lcssa = 0, $$0$i6 = 0, $$1 = 0, $$pre = 0, $$pre$i = 0, $$pre$i1 = 0, $$pre$phi$i3Z2D = 0, $$pre$phi$iZ2D = 0, $$pre$phiZ2D = 0, $$pre1$i = 0, $$pre1$i2 = 0, $$sink$i$i = 0, $$sink$i$i4 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $k$0 = 0, $k$0$lcssa = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 1608 "./libraries/stb/stb_image.h"
 $0 = ((($j)) + 18112|0); //@line 1613 "./libraries/stb/stb_image.h"
 $1 = HEAP32[$0>>2]|0; //@line 1613 "./libraries/stb/stb_image.h"
 $2 = ($1|0)<(16); //@line 1613 "./libraries/stb/stb_image.h"
 L1: do {
  if ($2) {
   $3 = ((($j)) + 18120|0); //@line 1590 "./libraries/stb/stb_image.h"
   $4 = ((($j)) + 18108|0); //@line 1599 "./libraries/stb/stb_image.h"
   L3: while(1) {
    $5 = HEAP32[$3>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
    $6 = ($5|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
    do {
     if ($6) {
      $7 = HEAP32[$j>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
      $8 = ((($7)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
      $9 = HEAP32[$8>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
      $10 = ((($7)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
      $11 = HEAP32[$10>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
      $12 = ($9>>>0)<($11>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      if ($12) {
       $13 = ((($9)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
       HEAP32[$8>>2] = $13; //@line 1230 "./libraries/stb/stb_image.h"
       $14 = HEAP8[$9>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
       $$0$i6 = $14;
      } else {
       $15 = ((($7)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
       $16 = HEAP32[$15>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
       $17 = ($16|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
       if ($17) {
        $58 = 0;
        break;
       }
       $18 = ((($7)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
       $19 = HEAP32[$18>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $20 = ((($7)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
       $21 = HEAP32[$20>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $22 = ((($7)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
       $23 = ((($7)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
       $24 = HEAP32[$23>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $25 = (FUNCTION_TABLE_iiii[$19 & 8191]($21,$22,$24)|0); //@line 1213 "./libraries/stb/stb_image.h"
       $26 = ($25|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
       if ($26) {
        HEAP32[$15>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
        $27 = ((($7)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
        HEAP8[$22>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
        $$pre$phi$i3Z2D = $27;$$sink$i$i4 = $27;$118 = 0;
       } else {
        $28 = (((($7)) + 40|0) + ($25)|0); //@line 1223 "./libraries/stb/stb_image.h"
        $$pre$i1 = HEAP8[$22>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
        $$pre1$i2 = ((($7)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
        $$pre$phi$i3Z2D = $$pre1$i2;$$sink$i$i4 = $28;$118 = $$pre$i1;
       }
       HEAP32[$10>>2] = $$sink$i$i4; //@line 1223 "./libraries/stb/stb_image.h"
       HEAP32[$8>>2] = $$pre$phi$i3Z2D; //@line 1233 "./libraries/stb/stb_image.h"
       $$0$i6 = $118;
      }
      $29 = $$0$i6&255; //@line 1590 "./libraries/stb/stb_image.h"
      $30 = ($$0$i6<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
      if ($30) {
       $31 = HEAP32[$j>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
       $32 = ((($31)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
       $33 = HEAP32[$32>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
       $34 = ((($31)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
       $35 = HEAP32[$34>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
       $36 = ($33>>>0)<($35>>>0); //@line 1229 "./libraries/stb/stb_image.h"
       if ($36) {
        $37 = ((($33)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$32>>2] = $37; //@line 1230 "./libraries/stb/stb_image.h"
        $38 = HEAP8[$33>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i = $38;
       } else {
        $39 = ((($31)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $40 = HEAP32[$39>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $41 = ($40|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($41) {
         $58 = 255;
         break;
        }
        $42 = ((($31)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $43 = HEAP32[$42>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $44 = ((($31)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $45 = HEAP32[$44>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $46 = ((($31)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $47 = ((($31)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $48 = HEAP32[$47>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $49 = (FUNCTION_TABLE_iiii[$43 & 8191]($45,$46,$48)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $50 = ($49|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($50) {
         HEAP32[$39>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $51 = ((($31)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$46>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$iZ2D = $51;$$sink$i$i = $51;$119 = 0;
        } else {
         $52 = (((($31)) + 40|0) + ($49)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i = HEAP8[$46>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i = ((($31)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$iZ2D = $$pre1$i;$$sink$i$i = $52;$119 = $$pre$i;
        }
        HEAP32[$34>>2] = $$sink$i$i; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$32>>2] = $$pre$phi$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i = $119;
       }
       $53 = ($$0$i<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
       if ($53) {
        $58 = 255;
       } else {
        $$0$i$lcssa = $$0$i;
        break L3;
       }
      } else {
       $58 = $29;
      }
     } else {
      $58 = 0;
     }
    } while(0);
    $55 = HEAP32[$0>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
    $56 = (24 - ($55))|0; //@line 1599 "./libraries/stb/stb_image.h"
    $57 = $58 << $56; //@line 1599 "./libraries/stb/stb_image.h"
    $59 = HEAP32[$4>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
    $60 = $57 | $59; //@line 1599 "./libraries/stb/stb_image.h"
    HEAP32[$4>>2] = $60; //@line 1599 "./libraries/stb/stb_image.h"
    $61 = (($55) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
    HEAP32[$0>>2] = $61; //@line 1600 "./libraries/stb/stb_image.h"
    $62 = ($61|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
    if (!($62)) {
     $$pre$phiZ2D = $4;
     break L1;
    }
   }
   $54 = ((($j)) + 18116|0); //@line 1594 "./libraries/stb/stb_image.h"
   HEAP8[$54>>0] = $$0$i$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
   HEAP32[$3>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
   $$pre$phiZ2D = $4;
  } else {
   $$pre = ((($j)) + 18108|0); //@line 1617 "./libraries/stb/stb_image.h"
   $$pre$phiZ2D = $$pre;
  }
 } while(0);
 $63 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 1653 "./libraries/stb/stb_image.h"
 $64 = $63 >>> 23; //@line 1617 "./libraries/stb/stb_image.h"
 $65 = (($h) + ($64)|0); //@line 1618 "./libraries/stb/stb_image.h"
 $66 = HEAP8[$65>>0]|0; //@line 1618 "./libraries/stb/stb_image.h"
 $67 = $66&255; //@line 1618 "./libraries/stb/stb_image.h"
 $68 = ($66<<24>>24)==(-1); //@line 1619 "./libraries/stb/stb_image.h"
 if (!($68)) {
  $69 = (((($h)) + 1280|0) + ($67)|0); //@line 1620 "./libraries/stb/stb_image.h"
  $70 = HEAP8[$69>>0]|0; //@line 1620 "./libraries/stb/stb_image.h"
  $71 = $70&255; //@line 1620 "./libraries/stb/stb_image.h"
  $72 = HEAP32[$0>>2]|0; //@line 1624 "./libraries/stb/stb_image.h"
  $73 = ($72|0)<($71|0); //@line 1621 "./libraries/stb/stb_image.h"
  if ($73) {
   $$1 = -1;
   STACKTOP = sp;return ($$1|0); //@line 1655 "./libraries/stb/stb_image.h"
  }
  $74 = $63 << $71; //@line 1623 "./libraries/stb/stb_image.h"
  HEAP32[$$pre$phiZ2D>>2] = $74; //@line 1623 "./libraries/stb/stb_image.h"
  $75 = (($72) - ($71))|0; //@line 1624 "./libraries/stb/stb_image.h"
  HEAP32[$0>>2] = $75; //@line 1624 "./libraries/stb/stb_image.h"
  $76 = (((($h)) + 1024|0) + ($67)|0); //@line 1625 "./libraries/stb/stb_image.h"
  $77 = HEAP8[$76>>0]|0; //@line 1625 "./libraries/stb/stb_image.h"
  $78 = $77&255; //@line 1625 "./libraries/stb/stb_image.h"
  $$1 = $78;
  STACKTOP = sp;return ($$1|0); //@line 1655 "./libraries/stb/stb_image.h"
 }
 $79 = $63 >>> 16; //@line 1634 "./libraries/stb/stb_image.h"
 $k$0 = 10;
 while(1) {
  $80 = (((($h)) + 1540|0) + ($k$0<<2)|0); //@line 1636 "./libraries/stb/stb_image.h"
  $81 = HEAP32[$80>>2]|0; //@line 1636 "./libraries/stb/stb_image.h"
  $82 = ($79>>>0)<($81>>>0); //@line 1636 "./libraries/stb/stb_image.h"
  $83 = (($k$0) + 1)|0; //@line 1635 "./libraries/stb/stb_image.h"
  if ($82) {
   $k$0$lcssa = $k$0;
   break;
  } else {
   $k$0 = $83;
  }
 }
 $84 = ($k$0$lcssa|0)==(17); //@line 1638 "./libraries/stb/stb_image.h"
 $85 = HEAP32[$0>>2]|0; //@line 1640 "./libraries/stb/stb_image.h"
 if ($84) {
  $86 = (($85) + -16)|0; //@line 1640 "./libraries/stb/stb_image.h"
  HEAP32[$0>>2] = $86; //@line 1640 "./libraries/stb/stb_image.h"
  $$1 = -1;
  STACKTOP = sp;return ($$1|0); //@line 1655 "./libraries/stb/stb_image.h"
 }
 $87 = ($85|0)<($k$0$lcssa|0); //@line 1644 "./libraries/stb/stb_image.h"
 if ($87) {
  $$1 = -1;
  STACKTOP = sp;return ($$1|0); //@line 1655 "./libraries/stb/stb_image.h"
 }
 $88 = (32 - ($k$0$lcssa))|0; //@line 1648 "./libraries/stb/stb_image.h"
 $89 = $63 >>> $88; //@line 1648 "./libraries/stb/stb_image.h"
 $90 = (67056 + ($k$0$lcssa<<2)|0); //@line 1648 "./libraries/stb/stb_image.h"
 $91 = HEAP32[$90>>2]|0; //@line 1648 "./libraries/stb/stb_image.h"
 $92 = $89 & $91; //@line 1648 "./libraries/stb/stb_image.h"
 $93 = (((($h)) + 1612|0) + ($k$0$lcssa<<2)|0); //@line 1648 "./libraries/stb/stb_image.h"
 $94 = HEAP32[$93>>2]|0; //@line 1648 "./libraries/stb/stb_image.h"
 $95 = (($92) + ($94))|0; //@line 1648 "./libraries/stb/stb_image.h"
 $96 = (((($h)) + 1280|0) + ($95)|0); //@line 1649 "./libraries/stb/stb_image.h"
 $97 = HEAP8[$96>>0]|0; //@line 1649 "./libraries/stb/stb_image.h"
 $98 = $97&255; //@line 1649 "./libraries/stb/stb_image.h"
 $99 = (32 - ($98))|0; //@line 1649 "./libraries/stb/stb_image.h"
 $100 = $63 >>> $99; //@line 1649 "./libraries/stb/stb_image.h"
 $101 = (67056 + ($98<<2)|0); //@line 1649 "./libraries/stb/stb_image.h"
 $102 = HEAP32[$101>>2]|0; //@line 1649 "./libraries/stb/stb_image.h"
 $103 = $100 & $102; //@line 1649 "./libraries/stb/stb_image.h"
 $104 = (((($h)) + 512|0) + ($95<<1)|0); //@line 1649 "./libraries/stb/stb_image.h"
 $105 = HEAP16[$104>>1]|0; //@line 1649 "./libraries/stb/stb_image.h"
 $106 = $105&65535; //@line 1649 "./libraries/stb/stb_image.h"
 $107 = ($103|0)==($106|0); //@line 1649 "./libraries/stb/stb_image.h"
 if ($107) {
  $113 = (($85) - ($k$0$lcssa))|0; //@line 1652 "./libraries/stb/stb_image.h"
  HEAP32[$0>>2] = $113; //@line 1652 "./libraries/stb/stb_image.h"
  $114 = $63 << $k$0$lcssa; //@line 1653 "./libraries/stb/stb_image.h"
  HEAP32[$$pre$phiZ2D>>2] = $114; //@line 1653 "./libraries/stb/stb_image.h"
  $115 = (((($h)) + 1024|0) + ($95)|0); //@line 1654 "./libraries/stb/stb_image.h"
  $116 = HEAP8[$115>>0]|0; //@line 1654 "./libraries/stb/stb_image.h"
  $117 = $116&255; //@line 1654 "./libraries/stb/stb_image.h"
  $$1 = $117;
  STACKTOP = sp;return ($$1|0); //@line 1655 "./libraries/stb/stb_image.h"
 }
 $108 = (___cxa_allocate_exception(16)|0); //@line 27 "modules/image/magpie/STBHandler.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = 441111; //@line 27 "modules/image/magpie/STBHandler.cpp"
 invoke_viii(3349,($108|0),(441193|0),($vararg_buffer|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
 $109 = __THREW__; __THREW__ = 0;
 $110 = $109&1;
 if ($110) {
  $111 = ___cxa_find_matching_catch_2()|0;
  $112 = tempRet0;
  ___cxa_free_exception(($108|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
  ___resumeException($111|0);
  // unreachable;
 } else {
  ___cxa_throw(($108|0),(96|0),(344|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
  // unreachable; //@line 27 "modules/image/magpie/STBHandler.cpp"
 }
 return (0)|0;
}
function __ZL31stbi__jpeg_decode_block_prog_dcP10stbi__jpegPsP13stbi__huffmani($j,$data,$hdc,$b) {
 $j = $j|0;
 $data = $data|0;
 $hdc = $hdc|0;
 $b = $b|0;
 var $$0 = 0, $$0$i = 0, $$0$i$i = 0, $$0$i$i$lcssa = 0, $$0$i16 = 0, $$0$i25 = 0, $$0$i25$lcssa = 0, $$0$i32 = 0, $$0$i9 = 0, $$0$i9$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i1 = 0, $$pre$i11 = 0, $$pre$i20 = 0, $$pre$i27 = 0, $$pre$i4 = 0, $$pre$phi$i$iZ2D = 0, $$pre$phi$i13Z2D = 0;
 var $$pre$phi$i22Z2D = 0, $$pre$phi$i29Z2D = 0, $$pre$phi$i3Z2D = 0, $$pre$phi$i6Z2D = 0, $$pre$phi$iZ2D = 0, $$pre$phiZ2D = 0, $$pre1$i = 0, $$pre1$i$i = 0, $$pre1$i12 = 0, $$pre1$i2 = 0, $$pre1$i21 = 0, $$pre1$i28 = 0, $$pre1$i5 = 0, $$sink$i$i = 0, $$sink$i$i$i = 0, $$sink$i$i14 = 0, $$sink$i$i23 = 0, $$sink$i$i30 = 0, $$sink$i$i7 = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $sext = 0, $vararg_buffer = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 1769 "./libraries/stb/stb_image.h"
 $0 = ((($j)) + 18132|0); //@line 1773 "./libraries/stb/stb_image.h"
 $1 = HEAP32[$0>>2]|0; //@line 1773 "./libraries/stb/stb_image.h"
 $2 = ($1|0)==(0); //@line 1773 "./libraries/stb/stb_image.h"
 if (!($2)) {
  HEAP32[158754] = 441330; //@line 894 "./libraries/stb/stb_image.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 1792 "./libraries/stb/stb_image.h"
 }
 $3 = ((($j)) + 18112|0); //@line 1775 "./libraries/stb/stb_image.h"
 $4 = HEAP32[$3>>2]|0; //@line 1775 "./libraries/stb/stb_image.h"
 $5 = ($4|0)<(16); //@line 1775 "./libraries/stb/stb_image.h"
 L5: do {
  if ($5) {
   $6 = ((($j)) + 18120|0); //@line 1590 "./libraries/stb/stb_image.h"
   $7 = ((($j)) + 18108|0); //@line 1599 "./libraries/stb/stb_image.h"
   L7: while(1) {
    $8 = HEAP32[$6>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
    $9 = ($8|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
    do {
     if ($9) {
      $10 = HEAP32[$j>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
      $11 = ((($10)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
      $12 = HEAP32[$11>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
      $13 = ((($10)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
      $14 = HEAP32[$13>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
      $15 = ($12>>>0)<($14>>>0); //@line 1229 "./libraries/stb/stb_image.h"
      if ($15) {
       $16 = ((($12)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
       HEAP32[$11>>2] = $16; //@line 1230 "./libraries/stb/stb_image.h"
       $17 = HEAP8[$12>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
       $$0$i16 = $17;
      } else {
       $18 = ((($10)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
       $19 = HEAP32[$18>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
       $20 = ($19|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
       if ($20) {
        $61 = 0;
        break;
       }
       $21 = ((($10)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
       $22 = HEAP32[$21>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $23 = ((($10)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
       $24 = HEAP32[$23>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $25 = ((($10)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
       $26 = ((($10)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
       $27 = HEAP32[$26>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
       $28 = (FUNCTION_TABLE_iiii[$22 & 8191]($24,$25,$27)|0); //@line 1213 "./libraries/stb/stb_image.h"
       $29 = ($28|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
       if ($29) {
        HEAP32[$18>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
        $30 = ((($10)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
        HEAP8[$25>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
        $$pre$phi$i13Z2D = $30;$$sink$i$i14 = $30;$239 = 0;
       } else {
        $31 = (((($10)) + 40|0) + ($28)|0); //@line 1223 "./libraries/stb/stb_image.h"
        $$pre$i11 = HEAP8[$25>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
        $$pre1$i12 = ((($10)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
        $$pre$phi$i13Z2D = $$pre1$i12;$$sink$i$i14 = $31;$239 = $$pre$i11;
       }
       HEAP32[$13>>2] = $$sink$i$i14; //@line 1223 "./libraries/stb/stb_image.h"
       HEAP32[$11>>2] = $$pre$phi$i13Z2D; //@line 1233 "./libraries/stb/stb_image.h"
       $$0$i16 = $239;
      }
      $32 = $$0$i16&255; //@line 1590 "./libraries/stb/stb_image.h"
      $33 = ($$0$i16<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
      if ($33) {
       $34 = HEAP32[$j>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
       $35 = ((($34)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
       $36 = HEAP32[$35>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
       $37 = ((($34)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
       $38 = HEAP32[$37>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
       $39 = ($36>>>0)<($38>>>0); //@line 1229 "./libraries/stb/stb_image.h"
       if ($39) {
        $40 = ((($36)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$35>>2] = $40; //@line 1230 "./libraries/stb/stb_image.h"
        $41 = HEAP8[$36>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i9 = $41;
       } else {
        $42 = ((($34)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $43 = HEAP32[$42>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $44 = ($43|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($44) {
         $61 = 255;
         break;
        }
        $45 = ((($34)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $46 = HEAP32[$45>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $47 = ((($34)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $48 = HEAP32[$47>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $49 = ((($34)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $50 = ((($34)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $51 = HEAP32[$50>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $52 = (FUNCTION_TABLE_iiii[$46 & 8191]($48,$49,$51)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $53 = ($52|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($53) {
         HEAP32[$42>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $54 = ((($34)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$49>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i6Z2D = $54;$$sink$i$i7 = $54;$240 = 0;
        } else {
         $55 = (((($34)) + 40|0) + ($52)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i4 = HEAP8[$49>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i5 = ((($34)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i6Z2D = $$pre1$i5;$$sink$i$i7 = $55;$240 = $$pre$i4;
        }
        HEAP32[$37>>2] = $$sink$i$i7; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$35>>2] = $$pre$phi$i6Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i9 = $240;
       }
       $56 = ($$0$i9<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
       if ($56) {
        $61 = 255;
       } else {
        $$0$i9$lcssa = $$0$i9;
        break L7;
       }
      } else {
       $61 = $32;
      }
     } else {
      $61 = 0;
     }
    } while(0);
    $58 = HEAP32[$3>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
    $59 = (24 - ($58))|0; //@line 1599 "./libraries/stb/stb_image.h"
    $60 = $61 << $59; //@line 1599 "./libraries/stb/stb_image.h"
    $62 = HEAP32[$7>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
    $63 = $60 | $62; //@line 1599 "./libraries/stb/stb_image.h"
    HEAP32[$7>>2] = $63; //@line 1599 "./libraries/stb/stb_image.h"
    $64 = (($58) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
    HEAP32[$3>>2] = $64; //@line 1600 "./libraries/stb/stb_image.h"
    $65 = ($64|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
    if (!($65)) {
     break L5;
    }
   }
   $57 = ((($j)) + 18116|0); //@line 1594 "./libraries/stb/stb_image.h"
   HEAP8[$57>>0] = $$0$i9$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
   HEAP32[$6>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
  }
 } while(0);
 $66 = ((($j)) + 18136|0); //@line 1777 "./libraries/stb/stb_image.h"
 $67 = HEAP32[$66>>2]|0; //@line 1777 "./libraries/stb/stb_image.h"
 $68 = ($67|0)==(0); //@line 1777 "./libraries/stb/stb_image.h"
 if (!($68)) {
  $165 = HEAP32[$3>>2]|0; //@line 1692 "./libraries/stb/stb_image.h"
  $166 = ($165|0)<(1); //@line 1692 "./libraries/stb/stb_image.h"
  L35: do {
   if ($166) {
    $167 = ((($j)) + 18120|0); //@line 1590 "./libraries/stb/stb_image.h"
    $168 = ((($j)) + 18108|0); //@line 1599 "./libraries/stb/stb_image.h"
    L37: while(1) {
     $169 = HEAP32[$167>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
     $170 = ($169|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
     do {
      if ($170) {
       $171 = HEAP32[$j>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
       $172 = ((($171)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
       $173 = HEAP32[$172>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
       $174 = ((($171)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
       $175 = HEAP32[$174>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
       $176 = ($173>>>0)<($175>>>0); //@line 1229 "./libraries/stb/stb_image.h"
       if ($176) {
        $177 = ((($173)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
        HEAP32[$172>>2] = $177; //@line 1230 "./libraries/stb/stb_image.h"
        $178 = HEAP8[$173>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
        $$0$i = $178;
       } else {
        $179 = ((($171)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
        $180 = HEAP32[$179>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
        $181 = ($180|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
        if ($181) {
         $222 = 0;
         break;
        }
        $182 = ((($171)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
        $183 = HEAP32[$182>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $184 = ((($171)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
        $185 = HEAP32[$184>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $186 = ((($171)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
        $187 = ((($171)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
        $188 = HEAP32[$187>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
        $189 = (FUNCTION_TABLE_iiii[$183 & 8191]($185,$186,$188)|0); //@line 1213 "./libraries/stb/stb_image.h"
        $190 = ($189|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
        if ($190) {
         HEAP32[$179>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
         $191 = ((($171)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
         HEAP8[$186>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
         $$pre$phi$i3Z2D = $191;$$sink$i$i = $191;$243 = 0;
        } else {
         $192 = (((($171)) + 40|0) + ($189)|0); //@line 1223 "./libraries/stb/stb_image.h"
         $$pre$i1 = HEAP8[$186>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
         $$pre1$i2 = ((($171)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
         $$pre$phi$i3Z2D = $$pre1$i2;$$sink$i$i = $192;$243 = $$pre$i1;
        }
        HEAP32[$174>>2] = $$sink$i$i; //@line 1223 "./libraries/stb/stb_image.h"
        HEAP32[$172>>2] = $$pre$phi$i3Z2D; //@line 1233 "./libraries/stb/stb_image.h"
        $$0$i = $243;
       }
       $193 = $$0$i&255; //@line 1590 "./libraries/stb/stb_image.h"
       $194 = ($$0$i<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
       if ($194) {
        $195 = HEAP32[$j>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
        $196 = ((($195)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
        $197 = HEAP32[$196>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
        $198 = ((($195)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
        $199 = HEAP32[$198>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
        $200 = ($197>>>0)<($199>>>0); //@line 1229 "./libraries/stb/stb_image.h"
        if ($200) {
         $201 = ((($197)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
         HEAP32[$196>>2] = $201; //@line 1230 "./libraries/stb/stb_image.h"
         $202 = HEAP8[$197>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
         $$0$i$i = $202;
        } else {
         $203 = ((($195)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
         $204 = HEAP32[$203>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
         $205 = ($204|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
         if ($205) {
          $222 = 255;
          break;
         }
         $206 = ((($195)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
         $207 = HEAP32[$206>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $208 = ((($195)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
         $209 = HEAP32[$208>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $210 = ((($195)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
         $211 = ((($195)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
         $212 = HEAP32[$211>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $213 = (FUNCTION_TABLE_iiii[$207 & 8191]($209,$210,$212)|0); //@line 1213 "./libraries/stb/stb_image.h"
         $214 = ($213|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
         if ($214) {
          HEAP32[$203>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
          $215 = ((($195)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
          HEAP8[$210>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
          $$pre$phi$i$iZ2D = $215;$$sink$i$i$i = $215;$244 = 0;
         } else {
          $216 = (((($195)) + 40|0) + ($213)|0); //@line 1223 "./libraries/stb/stb_image.h"
          $$pre$i$i = HEAP8[$210>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
          $$pre1$i$i = ((($195)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
          $$pre$phi$i$iZ2D = $$pre1$i$i;$$sink$i$i$i = $216;$244 = $$pre$i$i;
         }
         HEAP32[$198>>2] = $$sink$i$i$i; //@line 1223 "./libraries/stb/stb_image.h"
         HEAP32[$196>>2] = $$pre$phi$i$iZ2D; //@line 1233 "./libraries/stb/stb_image.h"
         $$0$i$i = $244;
        }
        $217 = ($$0$i$i<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
        if ($217) {
         $222 = 255;
        } else {
         $$0$i$i$lcssa = $$0$i$i;
         break L37;
        }
       } else {
        $222 = $193;
       }
      } else {
       $222 = 0;
      }
     } while(0);
     $219 = HEAP32[$3>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
     $220 = (24 - ($219))|0; //@line 1599 "./libraries/stb/stb_image.h"
     $221 = $222 << $220; //@line 1599 "./libraries/stb/stb_image.h"
     $223 = HEAP32[$168>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
     $224 = $221 | $223; //@line 1599 "./libraries/stb/stb_image.h"
     HEAP32[$168>>2] = $224; //@line 1599 "./libraries/stb/stb_image.h"
     $225 = (($219) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
     HEAP32[$3>>2] = $225; //@line 1600 "./libraries/stb/stb_image.h"
     $226 = ($225|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
     if (!($226)) {
      $$pre$phi$iZ2D = $168;$230 = $225;
      break L35;
     }
    }
    $218 = ((($j)) + 18116|0); //@line 1594 "./libraries/stb/stb_image.h"
    HEAP8[$218>>0] = $$0$i$i$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
    HEAP32[$167>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
    $$pre$i = HEAP32[$3>>2]|0; //@line 1695 "./libraries/stb/stb_image.h"
    $$pre$phi$iZ2D = $168;$230 = $$pre$i;
   } else {
    $$pre1$i = ((($j)) + 18108|0); //@line 1693 "./libraries/stb/stb_image.h"
    $$pre$phi$iZ2D = $$pre1$i;$230 = $165;
   }
  } while(0);
  $227 = HEAP32[$$pre$phi$iZ2D>>2]|0; //@line 1693 "./libraries/stb/stb_image.h"
  $228 = $227 << 1; //@line 1694 "./libraries/stb/stb_image.h"
  HEAP32[$$pre$phi$iZ2D>>2] = $228; //@line 1694 "./libraries/stb/stb_image.h"
  $229 = (($230) + -1)|0; //@line 1695 "./libraries/stb/stb_image.h"
  HEAP32[$3>>2] = $229; //@line 1695 "./libraries/stb/stb_image.h"
  $231 = ($227|0)>(-1); //@line 1788 "./libraries/stb/stb_image.h"
  if ($231) {
   $$0 = 1;
   STACKTOP = sp;return ($$0|0); //@line 1792 "./libraries/stb/stb_image.h"
  }
  $232 = ((($j)) + 18140|0); //@line 1789 "./libraries/stb/stb_image.h"
  $233 = HEAP32[$232>>2]|0; //@line 1789 "./libraries/stb/stb_image.h"
  $sext = 65536 << $233; //@line 1789 "./libraries/stb/stb_image.h"
  $234 = $sext >>> 16; //@line 1789 "./libraries/stb/stb_image.h"
  $235 = HEAP16[$data>>1]|0; //@line 1789 "./libraries/stb/stb_image.h"
  $236 = $235&65535; //@line 1789 "./libraries/stb/stb_image.h"
  $237 = (($236) + ($234))|0; //@line 1789 "./libraries/stb/stb_image.h"
  $238 = $237&65535; //@line 1789 "./libraries/stb/stb_image.h"
  HEAP16[$data>>1] = $238; //@line 1789 "./libraries/stb/stb_image.h"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 1792 "./libraries/stb/stb_image.h"
 }
 dest=$data; stop=dest+128|0; do { HEAP16[dest>>1]=0|0; dest=dest+2|0; } while ((dest|0) < (stop|0)); //@line 1779 "./libraries/stb/stb_image.h"
 $69 = (__ZL22stbi__jpeg_huff_decodeP10stbi__jpegP13stbi__huffman($j,$hdc)|0); //@line 1780 "./libraries/stb/stb_image.h"
 $70 = ($69|0)==(0); //@line 1781 "./libraries/stb/stb_image.h"
 do {
  if ($70) {
   $160 = 0;
  } else {
   $71 = HEAP32[$3>>2]|0; //@line 1666 "./libraries/stb/stb_image.h"
   $72 = ($71|0)<($69|0); //@line 1666 "./libraries/stb/stb_image.h"
   L71: do {
    if ($72) {
     $73 = ((($j)) + 18120|0); //@line 1590 "./libraries/stb/stb_image.h"
     $74 = ((($j)) + 18108|0); //@line 1599 "./libraries/stb/stb_image.h"
     L73: while(1) {
      $75 = HEAP32[$73>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
      $76 = ($75|0)==(0); //@line 1590 "./libraries/stb/stb_image.h"
      do {
       if ($76) {
        $77 = HEAP32[$j>>2]|0; //@line 1590 "./libraries/stb/stb_image.h"
        $78 = ((($77)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
        $79 = HEAP32[$78>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
        $80 = ((($77)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
        $81 = HEAP32[$80>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
        $82 = ($79>>>0)<($81>>>0); //@line 1229 "./libraries/stb/stb_image.h"
        if ($82) {
         $83 = ((($79)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
         HEAP32[$78>>2] = $83; //@line 1230 "./libraries/stb/stb_image.h"
         $84 = HEAP8[$79>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
         $$0$i32 = $84;
        } else {
         $85 = ((($77)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
         $86 = HEAP32[$85>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
         $87 = ($86|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
         if ($87) {
          $128 = 0;
          break;
         }
         $88 = ((($77)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
         $89 = HEAP32[$88>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $90 = ((($77)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
         $91 = HEAP32[$90>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $92 = ((($77)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
         $93 = ((($77)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
         $94 = HEAP32[$93>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
         $95 = (FUNCTION_TABLE_iiii[$89 & 8191]($91,$92,$94)|0); //@line 1213 "./libraries/stb/stb_image.h"
         $96 = ($95|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
         if ($96) {
          HEAP32[$85>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
          $97 = ((($77)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
          HEAP8[$92>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
          $$pre$phi$i29Z2D = $97;$$sink$i$i30 = $97;$241 = 0;
         } else {
          $98 = (((($77)) + 40|0) + ($95)|0); //@line 1223 "./libraries/stb/stb_image.h"
          $$pre$i27 = HEAP8[$92>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
          $$pre1$i28 = ((($77)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
          $$pre$phi$i29Z2D = $$pre1$i28;$$sink$i$i30 = $98;$241 = $$pre$i27;
         }
         HEAP32[$80>>2] = $$sink$i$i30; //@line 1223 "./libraries/stb/stb_image.h"
         HEAP32[$78>>2] = $$pre$phi$i29Z2D; //@line 1233 "./libraries/stb/stb_image.h"
         $$0$i32 = $241;
        }
        $99 = $$0$i32&255; //@line 1590 "./libraries/stb/stb_image.h"
        $100 = ($$0$i32<<24>>24)==(-1); //@line 1591 "./libraries/stb/stb_image.h"
        if ($100) {
         $101 = HEAP32[$j>>2]|0; //@line 1592 "./libraries/stb/stb_image.h"
         $102 = ((($101)) + 168|0); //@line 1229 "./libraries/stb/stb_image.h"
         $103 = HEAP32[$102>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $104 = ((($101)) + 172|0); //@line 1229 "./libraries/stb/stb_image.h"
         $105 = HEAP32[$104>>2]|0; //@line 1229 "./libraries/stb/stb_image.h"
         $106 = ($103>>>0)<($105>>>0); //@line 1229 "./libraries/stb/stb_image.h"
         if ($106) {
          $107 = ((($103)) + 1|0); //@line 1230 "./libraries/stb/stb_image.h"
          HEAP32[$102>>2] = $107; //@line 1230 "./libraries/stb/stb_image.h"
          $108 = HEAP8[$103>>0]|0; //@line 1230 "./libraries/stb/stb_image.h"
          $$0$i25 = $108;
         } else {
          $109 = ((($101)) + 32|0); //@line 1231 "./libraries/stb/stb_image.h"
          $110 = HEAP32[$109>>2]|0; //@line 1231 "./libraries/stb/stb_image.h"
          $111 = ($110|0)==(0); //@line 1231 "./libraries/stb/stb_image.h"
          if ($111) {
           $128 = 255;
           break;
          }
          $112 = ((($101)) + 16|0); //@line 1213 "./libraries/stb/stb_image.h"
          $113 = HEAP32[$112>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
          $114 = ((($101)) + 28|0); //@line 1213 "./libraries/stb/stb_image.h"
          $115 = HEAP32[$114>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
          $116 = ((($101)) + 40|0); //@line 1213 "./libraries/stb/stb_image.h"
          $117 = ((($101)) + 36|0); //@line 1213 "./libraries/stb/stb_image.h"
          $118 = HEAP32[$117>>2]|0; //@line 1213 "./libraries/stb/stb_image.h"
          $119 = (FUNCTION_TABLE_iiii[$113 & 8191]($115,$116,$118)|0); //@line 1213 "./libraries/stb/stb_image.h"
          $120 = ($119|0)==(0); //@line 1214 "./libraries/stb/stb_image.h"
          if ($120) {
           HEAP32[$109>>2] = 0; //@line 1217 "./libraries/stb/stb_image.h"
           $121 = ((($101)) + 41|0); //@line 1219 "./libraries/stb/stb_image.h"
           HEAP8[$116>>0] = 0; //@line 1220 "./libraries/stb/stb_image.h"
           $$pre$phi$i22Z2D = $121;$$sink$i$i23 = $121;$242 = 0;
          } else {
           $122 = (((($101)) + 40|0) + ($119)|0); //@line 1223 "./libraries/stb/stb_image.h"
           $$pre$i20 = HEAP8[$116>>0]|0; //@line 1233 "./libraries/stb/stb_image.h"
           $$pre1$i21 = ((($101)) + 41|0); //@line 1233 "./libraries/stb/stb_image.h"
           $$pre$phi$i22Z2D = $$pre1$i21;$$sink$i$i23 = $122;$242 = $$pre$i20;
          }
          HEAP32[$104>>2] = $$sink$i$i23; //@line 1223 "./libraries/stb/stb_image.h"
          HEAP32[$102>>2] = $$pre$phi$i22Z2D; //@line 1233 "./libraries/stb/stb_image.h"
          $$0$i25 = $242;
         }
         $123 = ($$0$i25<<24>>24)==(0); //@line 1593 "./libraries/stb/stb_image.h"
         if ($123) {
          $128 = 255;
         } else {
          $$0$i25$lcssa = $$0$i25;
          break L73;
         }
        } else {
         $128 = $99;
        }
       } else {
        $128 = 0;
       }
      } while(0);
      $125 = HEAP32[$3>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
      $126 = (24 - ($125))|0; //@line 1599 "./libraries/stb/stb_image.h"
      $127 = $128 << $126; //@line 1599 "./libraries/stb/stb_image.h"
      $129 = HEAP32[$74>>2]|0; //@line 1599 "./libraries/stb/stb_image.h"
      $130 = $127 | $129; //@line 1599 "./libraries/stb/stb_image.h"
      HEAP32[$74>>2] = $130; //@line 1599 "./libraries/stb/stb_image.h"
      $131 = (($125) + 8)|0; //@line 1600 "./libraries/stb/stb_image.h"
      HEAP32[$3>>2] = $131; //@line 1600 "./libraries/stb/stb_image.h"
      $132 = ($131|0)<(25); //@line 1601 "./libraries/stb/stb_image.h"
      if (!($132)) {
       $$pre$phiZ2D = $74;
       break L71;
      }
     }
     $124 = ((($j)) + 18116|0); //@line 1594 "./libraries/stb/stb_image.h"
     HEAP8[$124>>0] = $$0$i25$lcssa; //@line 1594 "./libraries/stb/stb_image.h"
     HEAP32[$73>>2] = 1; //@line 1595 "./libraries/stb/stb_image.h"
     $$pre$phiZ2D = $74;
    } else {
     $$pre = ((($j)) + 18108|0); //@line 1668 "./libraries/stb/stb_image.h"
     $$pre$phiZ2D = $$pre;
    }
   } while(0);
   $133 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 1668 "./libraries/stb/stb_image.h"
   $134 = $133 << $69; //@line 1669 "./libraries/stb/stb_image.h"
   $135 = (32 - ($69))|0; //@line 1669 "./libraries/stb/stb_image.h"
   $136 = $133 >>> $135; //@line 1669 "./libraries/stb/stb_image.h"
   $137 = $134 | $136; //@line 1669 "./libraries/stb/stb_image.h"
   $138 = ($69>>>0)<(17); //@line 1670 "./libraries/stb/stb_image.h"
   if ($138) {
    $144 = $133 >> 31; //@line 1668 "./libraries/stb/stb_image.h"
    $145 = (67056 + ($69<<2)|0); //@line 1671 "./libraries/stb/stb_image.h"
    $146 = HEAP32[$145>>2]|0; //@line 1671 "./libraries/stb/stb_image.h"
    $147 = $146 ^ -1; //@line 1671 "./libraries/stb/stb_image.h"
    $148 = $137 & $147; //@line 1671 "./libraries/stb/stb_image.h"
    HEAP32[$$pre$phiZ2D>>2] = $148; //@line 1671 "./libraries/stb/stb_image.h"
    $149 = $146 & $137; //@line 1672 "./libraries/stb/stb_image.h"
    $150 = HEAP32[$3>>2]|0; //@line 1673 "./libraries/stb/stb_image.h"
    $151 = (($150) - ($69))|0; //@line 1673 "./libraries/stb/stb_image.h"
    HEAP32[$3>>2] = $151; //@line 1673 "./libraries/stb/stb_image.h"
    $152 = (26004 + ($69<<2)|0); //@line 1674 "./libraries/stb/stb_image.h"
    $153 = HEAP32[$152>>2]|0; //@line 1674 "./libraries/stb/stb_image.h"
    $154 = $144 ^ -1; //@line 1674 "./libraries/stb/stb_image.h"
    $155 = $153 & $154; //@line 1674 "./libraries/stb/stb_image.h"
    $156 = (($155) + ($149))|0; //@line 1674 "./libraries/stb/stb_image.h"
    $160 = $156;
    break;
   }
   $139 = (___cxa_allocate_exception(16)|0); //@line 27 "modules/image/magpie/STBHandler.cpp"
   __THREW__ = 0;
   HEAP32[$vararg_buffer>>2] = 441267; //@line 27 "modules/image/magpie/STBHandler.cpp"
   invoke_viii(3349,($139|0),(441193|0),($vararg_buffer|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
   $140 = __THREW__; __THREW__ = 0;
   $141 = $140&1;
   if ($141) {
    $142 = ___cxa_find_matching_catch_2()|0;
    $143 = tempRet0;
    ___cxa_free_exception(($139|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
    ___resumeException($142|0);
    // unreachable;
   } else {
    ___cxa_throw(($139|0),(96|0),(344|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
    // unreachable; //@line 27 "modules/image/magpie/STBHandler.cpp"
   }
  }
 } while(0);
 $157 = (((((($j)) + 17820|0) + (($b*72)|0)|0)) + 24|0); //@line 1783 "./libraries/stb/stb_image.h"
 $158 = HEAP32[$157>>2]|0; //@line 1783 "./libraries/stb/stb_image.h"
 $159 = (($158) + ($160))|0; //@line 1783 "./libraries/stb/stb_image.h"
 HEAP32[$157>>2] = $159; //@line 1784 "./libraries/stb/stb_image.h"
 $161 = ((($j)) + 18140|0); //@line 1785 "./libraries/stb/stb_image.h"
 $162 = HEAP32[$161>>2]|0; //@line 1785 "./libraries/stb/stb_image.h"
 $163 = $159 << $162; //@line 1785 "./libraries/stb/stb_image.h"
 $164 = $163&65535; //@line 1785 "./libraries/stb/stb_image.h"
 HEAP16[$data>>1] = $164; //@line 1785 "./libraries/stb/stb_image.h"
 $$0 = 1;
 STACKTOP = sp;return ($$0|0); //@line 1792 "./libraries/stb/stb_image.h"
}
function __ZL14resample_row_1PhS_S_ii($out,$in_near,$in_far,$w,$hs) {
 $out = $out|0;
 $in_near = $in_near|0;
 $in_far = $in_far|0;
 $w = $w|0;
 $hs = $hs|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ($in_near|0); //@line 2866 "./libraries/stb/stb_image.h"
}
function __ZL22stbi__resample_row_v_2PhS_S_ii($out,$in_near,$in_far,$w,$hs) {
 $out = $out|0;
 $in_near = $in_near|0;
 $in_far = $in_far|0;
 $w = $w|0;
 $hs = $hs|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($w|0)>(0); //@line 2874 "./libraries/stb/stb_image.h"
 if ($0) {
  $i$01 = 0;
 } else {
  return ($out|0); //@line 2876 "./libraries/stb/stb_image.h"
 }
 while(1) {
  $1 = (($in_near) + ($i$01)|0); //@line 2875 "./libraries/stb/stb_image.h"
  $2 = HEAP8[$1>>0]|0; //@line 2875 "./libraries/stb/stb_image.h"
  $3 = $2&255; //@line 2875 "./libraries/stb/stb_image.h"
  $4 = ($3*3)|0; //@line 2875 "./libraries/stb/stb_image.h"
  $5 = (($in_far) + ($i$01)|0); //@line 2875 "./libraries/stb/stb_image.h"
  $6 = HEAP8[$5>>0]|0; //@line 2875 "./libraries/stb/stb_image.h"
  $7 = $6&255; //@line 2875 "./libraries/stb/stb_image.h"
  $8 = (($7) + 2)|0; //@line 2875 "./libraries/stb/stb_image.h"
  $9 = (($8) + ($4))|0; //@line 2875 "./libraries/stb/stb_image.h"
  $10 = $9 >>> 2; //@line 2875 "./libraries/stb/stb_image.h"
  $11 = $10&255; //@line 2875 "./libraries/stb/stb_image.h"
  $12 = (($out) + ($i$01)|0); //@line 2875 "./libraries/stb/stb_image.h"
  HEAP8[$12>>0] = $11; //@line 2875 "./libraries/stb/stb_image.h"
  $13 = (($i$01) + 1)|0; //@line 2874 "./libraries/stb/stb_image.h"
  $exitcond = ($13|0)==($w|0); //@line 2874 "./libraries/stb/stb_image.h"
  if ($exitcond) {
   break;
  } else {
   $i$01 = $13;
  }
 }
 return ($out|0); //@line 2876 "./libraries/stb/stb_image.h"
}
function __ZL22stbi__resample_row_h_2PhS_S_ii($out,$in_near,$in_far,$w,$hs) {
 $out = $out|0;
 $in_near = $in_near|0;
 $in_far = $in_far|0;
 $w = $w|0;
 $hs = $hs|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$0$lcssa = 0, $i$01 = 0, $phitmp = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ($w|0)==(1); //@line 2885 "./libraries/stb/stb_image.h"
 $1 = HEAP8[$in_near>>0]|0; //@line 2887 "./libraries/stb/stb_image.h"
 if ($0) {
  $2 = ((($out)) + 1|0); //@line 2887 "./libraries/stb/stb_image.h"
  HEAP8[$2>>0] = $1; //@line 2887 "./libraries/stb/stb_image.h"
  HEAP8[$out>>0] = $1; //@line 2887 "./libraries/stb/stb_image.h"
  return ($out|0); //@line 2905 "./libraries/stb/stb_image.h"
 }
 HEAP8[$out>>0] = $1; //@line 2891 "./libraries/stb/stb_image.h"
 $3 = HEAP8[$in_near>>0]|0; //@line 2892 "./libraries/stb/stb_image.h"
 $4 = $3&255; //@line 2892 "./libraries/stb/stb_image.h"
 $5 = ($4*3)|0; //@line 2892 "./libraries/stb/stb_image.h"
 $6 = ((($in_near)) + 1|0); //@line 2892 "./libraries/stb/stb_image.h"
 $7 = HEAP8[$6>>0]|0; //@line 2892 "./libraries/stb/stb_image.h"
 $8 = $7&255; //@line 2892 "./libraries/stb/stb_image.h"
 $9 = (($8) + 2)|0; //@line 2892 "./libraries/stb/stb_image.h"
 $10 = (($9) + ($5))|0; //@line 2892 "./libraries/stb/stb_image.h"
 $11 = $10 >>> 2; //@line 2892 "./libraries/stb/stb_image.h"
 $12 = $11&255; //@line 2892 "./libraries/stb/stb_image.h"
 $13 = ((($out)) + 1|0); //@line 2892 "./libraries/stb/stb_image.h"
 HEAP8[$13>>0] = $12; //@line 2892 "./libraries/stb/stb_image.h"
 $14 = (($w) + -1)|0; //@line 2893 "./libraries/stb/stb_image.h"
 $15 = ($14|0)>(1); //@line 2893 "./libraries/stb/stb_image.h"
 if ($15) {
  $i$01 = 1;
  while(1) {
   $16 = (($in_near) + ($i$01)|0); //@line 2894 "./libraries/stb/stb_image.h"
   $17 = HEAP8[$16>>0]|0; //@line 2894 "./libraries/stb/stb_image.h"
   $18 = $17&255; //@line 2894 "./libraries/stb/stb_image.h"
   $19 = ($18*3)|0; //@line 2894 "./libraries/stb/stb_image.h"
   $20 = (($19) + 2)|0; //@line 2894 "./libraries/stb/stb_image.h"
   $21 = (($i$01) + -1)|0; //@line 2895 "./libraries/stb/stb_image.h"
   $22 = (($in_near) + ($21)|0); //@line 2895 "./libraries/stb/stb_image.h"
   $23 = HEAP8[$22>>0]|0; //@line 2895 "./libraries/stb/stb_image.h"
   $24 = $23&255; //@line 2895 "./libraries/stb/stb_image.h"
   $25 = (($20) + ($24))|0; //@line 2895 "./libraries/stb/stb_image.h"
   $26 = $25 >>> 2; //@line 2895 "./libraries/stb/stb_image.h"
   $27 = $26&255; //@line 2895 "./libraries/stb/stb_image.h"
   $28 = $i$01 << 1; //@line 2895 "./libraries/stb/stb_image.h"
   $29 = (($out) + ($28)|0); //@line 2895 "./libraries/stb/stb_image.h"
   HEAP8[$29>>0] = $27; //@line 2895 "./libraries/stb/stb_image.h"
   $30 = (($i$01) + 1)|0; //@line 2896 "./libraries/stb/stb_image.h"
   $31 = (($in_near) + ($30)|0); //@line 2896 "./libraries/stb/stb_image.h"
   $32 = HEAP8[$31>>0]|0; //@line 2896 "./libraries/stb/stb_image.h"
   $33 = $32&255; //@line 2896 "./libraries/stb/stb_image.h"
   $34 = (($33) + ($20))|0; //@line 2896 "./libraries/stb/stb_image.h"
   $35 = $34 >>> 2; //@line 2896 "./libraries/stb/stb_image.h"
   $36 = $35&255; //@line 2896 "./libraries/stb/stb_image.h"
   $37 = $28 | 1; //@line 2896 "./libraries/stb/stb_image.h"
   $38 = (($out) + ($37)|0); //@line 2896 "./libraries/stb/stb_image.h"
   HEAP8[$38>>0] = $36; //@line 2896 "./libraries/stb/stb_image.h"
   $exitcond = ($30|0)==($14|0); //@line 2893 "./libraries/stb/stb_image.h"
   if ($exitcond) {
    break;
   } else {
    $i$01 = $30;
   }
  }
  $phitmp = $14 << 1; //@line 2898 "./libraries/stb/stb_image.h"
  $i$0$lcssa = $phitmp;
 } else {
  $i$0$lcssa = 2;
 }
 $39 = (($w) + -2)|0; //@line 2898 "./libraries/stb/stb_image.h"
 $40 = (($in_near) + ($39)|0); //@line 2898 "./libraries/stb/stb_image.h"
 $41 = HEAP8[$40>>0]|0; //@line 2898 "./libraries/stb/stb_image.h"
 $42 = $41&255; //@line 2898 "./libraries/stb/stb_image.h"
 $43 = ($42*3)|0; //@line 2898 "./libraries/stb/stb_image.h"
 $44 = (($in_near) + ($14)|0); //@line 2898 "./libraries/stb/stb_image.h"
 $45 = HEAP8[$44>>0]|0; //@line 2898 "./libraries/stb/stb_image.h"
 $46 = $45&255; //@line 2898 "./libraries/stb/stb_image.h"
 $47 = (($46) + 2)|0; //@line 2898 "./libraries/stb/stb_image.h"
 $48 = (($47) + ($43))|0; //@line 2898 "./libraries/stb/stb_image.h"
 $49 = $48 >>> 2; //@line 2898 "./libraries/stb/stb_image.h"
 $50 = $49&255; //@line 2898 "./libraries/stb/stb_image.h"
 $51 = (($out) + ($i$0$lcssa)|0); //@line 2898 "./libraries/stb/stb_image.h"
 HEAP8[$51>>0] = $50; //@line 2898 "./libraries/stb/stb_image.h"
 $52 = HEAP8[$44>>0]|0; //@line 2899 "./libraries/stb/stb_image.h"
 $53 = $i$0$lcssa | 1; //@line 2899 "./libraries/stb/stb_image.h"
 $54 = (($out) + ($53)|0); //@line 2899 "./libraries/stb/stb_image.h"
 HEAP8[$54>>0] = $52; //@line 2899 "./libraries/stb/stb_image.h"
 return ($out|0); //@line 2905 "./libraries/stb/stb_image.h"
}
function __ZL26stbi__resample_row_genericPhS_S_ii($out,$in_near,$in_far,$w,$hs) {
 $out = $out|0;
 $in_near = $in_near|0;
 $in_far = $in_far|0;
 $w = $w|0;
 $hs = $hs|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $exitcond = 0, $exitcond6 = 0, $i$02$us = 0, $j$01$us = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($w|0)>(0); //@line 3055 "./libraries/stb/stb_image.h"
 $1 = ($hs|0)>(0); //@line 3056 "./libraries/stb/stb_image.h"
 $or$cond = $0 & $1; //@line 3055 "./libraries/stb/stb_image.h"
 if ($or$cond) {
  $i$02$us = 0;
 } else {
  return ($out|0); //@line 3058 "./libraries/stb/stb_image.h"
 }
 while(1) {
  $3 = (($in_near) + ($i$02$us)|0); //@line 3057 "./libraries/stb/stb_image.h"
  $5 = Math_imul($i$02$us, $hs)|0; //@line 3057 "./libraries/stb/stb_image.h"
  $j$01$us = 0;
  while(1) {
   $2 = HEAP8[$3>>0]|0; //@line 3057 "./libraries/stb/stb_image.h"
   $4 = (($j$01$us) + ($5))|0; //@line 3057 "./libraries/stb/stb_image.h"
   $6 = (($out) + ($4)|0); //@line 3057 "./libraries/stb/stb_image.h"
   HEAP8[$6>>0] = $2; //@line 3057 "./libraries/stb/stb_image.h"
   $7 = (($j$01$us) + 1)|0; //@line 3056 "./libraries/stb/stb_image.h"
   $exitcond = ($7|0)==($hs|0); //@line 3056 "./libraries/stb/stb_image.h"
   if ($exitcond) {
    break;
   } else {
    $j$01$us = $7;
   }
  }
  $8 = (($i$02$us) + 1)|0; //@line 3055 "./libraries/stb/stb_image.h"
  $exitcond6 = ($8|0)==($w|0); //@line 3055 "./libraries/stb/stb_image.h"
  if ($exitcond6) {
   break;
  } else {
   $i$02$us = $8;
  }
 }
 return ($out|0); //@line 3058 "./libraries/stb/stb_image.h"
}
function __ZL20stbi__convert_formatPhiijj($data,$img_n,$req_comp,$x,$y) {
 $data = $data|0;
 $img_n = $img_n|0;
 $req_comp = $req_comp|0;
 $x = $x|0;
 $y = $y|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $dest$081 = 0, $dest$1031 = 0, $dest$1127 = 0, $dest$176 = 0, $dest$271 = 0, $dest$366 = 0, $dest$461 = 0, $dest$556 = 0, $dest$651 = 0, $dest$746 = 0, $dest$841 = 0, $dest$936 = 0, $i$0 = 0, $i$079 = 0, $i$082 = 0, $i$1 = 0, $i$10 = 0, $i$1032 = 0;
 var $i$11 = 0, $i$1128 = 0, $i$177 = 0, $i$2 = 0, $i$272 = 0, $i$3 = 0, $i$367 = 0, $i$4 = 0, $i$462 = 0, $i$5 = 0, $i$557 = 0, $i$6 = 0, $i$652 = 0, $i$7 = 0, $i$747 = 0, $i$8 = 0, $i$842 = 0, $i$9 = 0, $i$937 = 0, $j$084 = 0;
 var $req_comp$off = 0, $src$080 = 0, $src$1030 = 0, $src$1126 = 0, $src$175 = 0, $src$270 = 0, $src$365 = 0, $src$460 = 0, $src$555 = 0, $src$650 = 0, $src$745 = 0, $src$840 = 0, $src$935 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 1340 "./libraries/stb/stb_image.h"
 $vararg_buffer = sp; //@line 1340 "./libraries/stb/stb_image.h"
 $0 = ($req_comp|0)==($img_n|0); //@line 1345 "./libraries/stb/stb_image.h"
 if ($0) {
  $$0 = $data;
  STACKTOP = sp;return ($$0|0); //@line 1382 "./libraries/stb/stb_image.h"
 }
 $req_comp$off = (($req_comp) + -1)|0; //@line 1346 "./libraries/stb/stb_image.h"
 $1 = ($req_comp$off>>>0)<(4); //@line 1346 "./libraries/stb/stb_image.h"
 if (!($1)) {
  $2 = (___cxa_allocate_exception(16)|0); //@line 27 "modules/image/magpie/STBHandler.cpp"
  __THREW__ = 0;
  HEAP32[$vararg_buffer>>2] = 441469; //@line 27 "modules/image/magpie/STBHandler.cpp"
  invoke_viii(3349,($2|0),(441193|0),($vararg_buffer|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if ($4) {
   $5 = ___cxa_find_matching_catch_2()|0;
   $6 = tempRet0;
   ___cxa_free_exception(($2|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
   ___resumeException($5|0);
   // unreachable;
  } else {
   ___cxa_throw(($2|0),(96|0),(344|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
   // unreachable; //@line 27 "modules/image/magpie/STBHandler.cpp"
  }
 }
 $7 = Math_imul($x, $req_comp)|0; //@line 1348 "./libraries/stb/stb_image.h"
 $8 = Math_imul($7, $y)|0; //@line 1348 "./libraries/stb/stb_image.h"
 $9 = (_malloc($8)|0); //@line 900 "./libraries/stb/stb_image.h"
 $10 = ($9|0)==(0|0); //@line 1349 "./libraries/stb/stb_image.h"
 if ($10) {
  _free($data); //@line 1350 "./libraries/stb/stb_image.h"
  HEAP32[158754] = 441002; //@line 894 "./libraries/stb/stb_image.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 1382 "./libraries/stb/stb_image.h"
 }
 $11 = ($y|0)>(0); //@line 1354 "./libraries/stb/stb_image.h"
 L14: do {
  if ($11) {
   $12 = $img_n << 3; //@line 1362 "./libraries/stb/stb_image.h"
   $13 = (($12) + ($req_comp))|0; //@line 1362 "./libraries/stb/stb_image.h"
   $i$079 = (($x) + -1)|0; //@line 1363 "./libraries/stb/stb_image.h"
   $14 = ($i$079|0)>(-1); //@line 1363 "./libraries/stb/stb_image.h"
   $j$084 = 0;
   L16: while(1) {
    $15 = Math_imul($j$084, $x)|0; //@line 1355 "./libraries/stb/stb_image.h"
    $16 = Math_imul($15, $img_n)|0; //@line 1355 "./libraries/stb/stb_image.h"
    $17 = (($data) + ($16)|0); //@line 1355 "./libraries/stb/stb_image.h"
    $18 = Math_imul($15, $req_comp)|0; //@line 1356 "./libraries/stb/stb_image.h"
    $19 = (($9) + ($18)|0); //@line 1356 "./libraries/stb/stb_image.h"
    do {
     switch ($13|0) {
     case 10:  {
      if ($14) {
       $dest$081 = $19;$i$082 = $i$079;$src$080 = $17;
       while(1) {
        $20 = HEAP8[$src$080>>0]|0; //@line 1363 "./libraries/stb/stb_image.h"
        HEAP8[$dest$081>>0] = $20; //@line 1363 "./libraries/stb/stb_image.h"
        $21 = ((($dest$081)) + 1|0); //@line 1363 "./libraries/stb/stb_image.h"
        HEAP8[$21>>0] = -1; //@line 1363 "./libraries/stb/stb_image.h"
        $22 = ((($src$080)) + 1|0); //@line 1363 "./libraries/stb/stb_image.h"
        $23 = ((($dest$081)) + 2|0); //@line 1363 "./libraries/stb/stb_image.h"
        $i$0 = (($i$082) + -1)|0; //@line 1363 "./libraries/stb/stb_image.h"
        $24 = ($i$0|0)>(-1); //@line 1363 "./libraries/stb/stb_image.h"
        if ($24) {
         $dest$081 = $23;$i$082 = $i$0;$src$080 = $22;
        } else {
         break;
        }
       }
      }
      break;
     }
     case 11:  {
      if ($14) {
       $dest$176 = $19;$i$177 = $i$079;$src$175 = $17;
       while(1) {
        $25 = HEAP8[$src$175>>0]|0; //@line 1364 "./libraries/stb/stb_image.h"
        $26 = ((($dest$176)) + 2|0); //@line 1364 "./libraries/stb/stb_image.h"
        HEAP8[$26>>0] = $25; //@line 1364 "./libraries/stb/stb_image.h"
        $27 = ((($dest$176)) + 1|0); //@line 1364 "./libraries/stb/stb_image.h"
        HEAP8[$27>>0] = $25; //@line 1364 "./libraries/stb/stb_image.h"
        HEAP8[$dest$176>>0] = $25; //@line 1364 "./libraries/stb/stb_image.h"
        $28 = ((($src$175)) + 1|0); //@line 1364 "./libraries/stb/stb_image.h"
        $29 = ((($dest$176)) + 3|0); //@line 1364 "./libraries/stb/stb_image.h"
        $i$1 = (($i$177) + -1)|0; //@line 1364 "./libraries/stb/stb_image.h"
        $30 = ($i$1|0)>(-1); //@line 1364 "./libraries/stb/stb_image.h"
        if ($30) {
         $dest$176 = $29;$i$177 = $i$1;$src$175 = $28;
        } else {
         break;
        }
       }
      }
      break;
     }
     case 12:  {
      if ($14) {
       $dest$271 = $19;$i$272 = $i$079;$src$270 = $17;
       while(1) {
        $31 = HEAP8[$src$270>>0]|0; //@line 1365 "./libraries/stb/stb_image.h"
        $32 = ((($dest$271)) + 2|0); //@line 1365 "./libraries/stb/stb_image.h"
        HEAP8[$32>>0] = $31; //@line 1365 "./libraries/stb/stb_image.h"
        $33 = ((($dest$271)) + 1|0); //@line 1365 "./libraries/stb/stb_image.h"
        HEAP8[$33>>0] = $31; //@line 1365 "./libraries/stb/stb_image.h"
        HEAP8[$dest$271>>0] = $31; //@line 1365 "./libraries/stb/stb_image.h"
        $34 = ((($dest$271)) + 3|0); //@line 1365 "./libraries/stb/stb_image.h"
        HEAP8[$34>>0] = -1; //@line 1365 "./libraries/stb/stb_image.h"
        $35 = ((($src$270)) + 1|0); //@line 1365 "./libraries/stb/stb_image.h"
        $36 = ((($dest$271)) + 4|0); //@line 1365 "./libraries/stb/stb_image.h"
        $i$2 = (($i$272) + -1)|0; //@line 1365 "./libraries/stb/stb_image.h"
        $37 = ($i$2|0)>(-1); //@line 1365 "./libraries/stb/stb_image.h"
        if ($37) {
         $dest$271 = $36;$i$272 = $i$2;$src$270 = $35;
        } else {
         break;
        }
       }
      }
      break;
     }
     case 17:  {
      if ($14) {
       $dest$366 = $19;$i$367 = $i$079;$src$365 = $17;
       while(1) {
        $38 = HEAP8[$src$365>>0]|0; //@line 1366 "./libraries/stb/stb_image.h"
        HEAP8[$dest$366>>0] = $38; //@line 1366 "./libraries/stb/stb_image.h"
        $39 = ((($src$365)) + 2|0); //@line 1366 "./libraries/stb/stb_image.h"
        $40 = ((($dest$366)) + 1|0); //@line 1366 "./libraries/stb/stb_image.h"
        $i$3 = (($i$367) + -1)|0; //@line 1366 "./libraries/stb/stb_image.h"
        $41 = ($i$3|0)>(-1); //@line 1366 "./libraries/stb/stb_image.h"
        if ($41) {
         $dest$366 = $40;$i$367 = $i$3;$src$365 = $39;
        } else {
         break;
        }
       }
      }
      break;
     }
     case 19:  {
      if ($14) {
       $dest$461 = $19;$i$462 = $i$079;$src$460 = $17;
       while(1) {
        $42 = HEAP8[$src$460>>0]|0; //@line 1367 "./libraries/stb/stb_image.h"
        $43 = ((($dest$461)) + 2|0); //@line 1367 "./libraries/stb/stb_image.h"
        HEAP8[$43>>0] = $42; //@line 1367 "./libraries/stb/stb_image.h"
        $44 = ((($dest$461)) + 1|0); //@line 1367 "./libraries/stb/stb_image.h"
        HEAP8[$44>>0] = $42; //@line 1367 "./libraries/stb/stb_image.h"
        HEAP8[$dest$461>>0] = $42; //@line 1367 "./libraries/stb/stb_image.h"
        $45 = ((($src$460)) + 2|0); //@line 1367 "./libraries/stb/stb_image.h"
        $46 = ((($dest$461)) + 3|0); //@line 1367 "./libraries/stb/stb_image.h"
        $i$4 = (($i$462) + -1)|0; //@line 1367 "./libraries/stb/stb_image.h"
        $47 = ($i$4|0)>(-1); //@line 1367 "./libraries/stb/stb_image.h"
        if ($47) {
         $dest$461 = $46;$i$462 = $i$4;$src$460 = $45;
        } else {
         break;
        }
       }
      }
      break;
     }
     case 20:  {
      if ($14) {
       $dest$556 = $19;$i$557 = $i$079;$src$555 = $17;
       while(1) {
        $48 = HEAP8[$src$555>>0]|0; //@line 1368 "./libraries/stb/stb_image.h"
        $49 = ((($dest$556)) + 2|0); //@line 1368 "./libraries/stb/stb_image.h"
        HEAP8[$49>>0] = $48; //@line 1368 "./libraries/stb/stb_image.h"
        $50 = ((($dest$556)) + 1|0); //@line 1368 "./libraries/stb/stb_image.h"
        HEAP8[$50>>0] = $48; //@line 1368 "./libraries/stb/stb_image.h"
        HEAP8[$dest$556>>0] = $48; //@line 1368 "./libraries/stb/stb_image.h"
        $51 = ((($src$555)) + 1|0); //@line 1368 "./libraries/stb/stb_image.h"
        $52 = HEAP8[$51>>0]|0; //@line 1368 "./libraries/stb/stb_image.h"
        $53 = ((($dest$556)) + 3|0); //@line 1368 "./libraries/stb/stb_image.h"
        HEAP8[$53>>0] = $52; //@line 1368 "./libraries/stb/stb_image.h"
        $54 = ((($src$555)) + 2|0); //@line 1368 "./libraries/stb/stb_image.h"
        $55 = ((($dest$556)) + 4|0); //@line 1368 "./libraries/stb/stb_image.h"
        $i$5 = (($i$557) + -1)|0; //@line 1368 "./libraries/stb/stb_image.h"
        $56 = ($i$5|0)>(-1); //@line 1368 "./libraries/stb/stb_image.h"
        if ($56) {
         $dest$556 = $55;$i$557 = $i$5;$src$555 = $54;
        } else {
         break;
        }
       }
      }
      break;
     }
     case 28:  {
      if ($14) {
       $dest$651 = $19;$i$652 = $i$079;$src$650 = $17;
       while(1) {
        $57 = HEAP8[$src$650>>0]|0; //@line 1369 "./libraries/stb/stb_image.h"
        HEAP8[$dest$651>>0] = $57; //@line 1369 "./libraries/stb/stb_image.h"
        $58 = ((($src$650)) + 1|0); //@line 1369 "./libraries/stb/stb_image.h"
        $59 = HEAP8[$58>>0]|0; //@line 1369 "./libraries/stb/stb_image.h"
        $60 = ((($dest$651)) + 1|0); //@line 1369 "./libraries/stb/stb_image.h"
        HEAP8[$60>>0] = $59; //@line 1369 "./libraries/stb/stb_image.h"
        $61 = ((($src$650)) + 2|0); //@line 1369 "./libraries/stb/stb_image.h"
        $62 = HEAP8[$61>>0]|0; //@line 1369 "./libraries/stb/stb_image.h"
        $63 = ((($dest$651)) + 2|0); //@line 1369 "./libraries/stb/stb_image.h"
        HEAP8[$63>>0] = $62; //@line 1369 "./libraries/stb/stb_image.h"
        $64 = ((($dest$651)) + 3|0); //@line 1369 "./libraries/stb/stb_image.h"
        HEAP8[$64>>0] = -1; //@line 1369 "./libraries/stb/stb_image.h"
        $65 = ((($src$650)) + 3|0); //@line 1369 "./libraries/stb/stb_image.h"
        $66 = ((($dest$651)) + 4|0); //@line 1369 "./libraries/stb/stb_image.h"
        $i$6 = (($i$652) + -1)|0; //@line 1369 "./libraries/stb/stb_image.h"
        $67 = ($i$6|0)>(-1); //@line 1369 "./libraries/stb/stb_image.h"
        if ($67) {
         $dest$651 = $66;$i$652 = $i$6;$src$650 = $65;
        } else {
         break;
        }
       }
      }
      break;
     }
     case 25:  {
      if ($14) {
       $dest$746 = $19;$i$747 = $i$079;$src$745 = $17;
       while(1) {
        $68 = HEAP8[$src$745>>0]|0; //@line 1370 "./libraries/stb/stb_image.h"
        $69 = $68&255; //@line 1370 "./libraries/stb/stb_image.h"
        $70 = ((($src$745)) + 1|0); //@line 1370 "./libraries/stb/stb_image.h"
        $71 = HEAP8[$70>>0]|0; //@line 1370 "./libraries/stb/stb_image.h"
        $72 = $71&255; //@line 1370 "./libraries/stb/stb_image.h"
        $73 = ((($src$745)) + 2|0); //@line 1370 "./libraries/stb/stb_image.h"
        $74 = HEAP8[$73>>0]|0; //@line 1370 "./libraries/stb/stb_image.h"
        $75 = $74&255; //@line 1370 "./libraries/stb/stb_image.h"
        $76 = ($69*77)|0; //@line 1337 "./libraries/stb/stb_image.h"
        $77 = ($72*150)|0; //@line 1337 "./libraries/stb/stb_image.h"
        $78 = (($77) + ($76))|0; //@line 1337 "./libraries/stb/stb_image.h"
        $79 = ($75*29)|0; //@line 1337 "./libraries/stb/stb_image.h"
        $80 = (($78) + ($79))|0; //@line 1337 "./libraries/stb/stb_image.h"
        $81 = $80 >>> 8; //@line 1337 "./libraries/stb/stb_image.h"
        $82 = $81&255; //@line 1337 "./libraries/stb/stb_image.h"
        HEAP8[$dest$746>>0] = $82; //@line 1370 "./libraries/stb/stb_image.h"
        $83 = ((($src$745)) + 3|0); //@line 1370 "./libraries/stb/stb_image.h"
        $84 = ((($dest$746)) + 1|0); //@line 1370 "./libraries/stb/stb_image.h"
        $i$7 = (($i$747) + -1)|0; //@line 1370 "./libraries/stb/stb_image.h"
        $85 = ($i$7|0)>(-1); //@line 1370 "./libraries/stb/stb_image.h"
        if ($85) {
         $dest$746 = $84;$i$747 = $i$7;$src$745 = $83;
        } else {
         break;
        }
       }
      }
      break;
     }
     case 26:  {
      if ($14) {
       $dest$841 = $19;$i$842 = $i$079;$src$840 = $17;
       while(1) {
        $86 = HEAP8[$src$840>>0]|0; //@line 1371 "./libraries/stb/stb_image.h"
        $87 = $86&255; //@line 1371 "./libraries/stb/stb_image.h"
        $88 = ((($src$840)) + 1|0); //@line 1371 "./libraries/stb/stb_image.h"
        $89 = HEAP8[$88>>0]|0; //@line 1371 "./libraries/stb/stb_image.h"
        $90 = $89&255; //@line 1371 "./libraries/stb/stb_image.h"
        $91 = ((($src$840)) + 2|0); //@line 1371 "./libraries/stb/stb_image.h"
        $92 = HEAP8[$91>>0]|0; //@line 1371 "./libraries/stb/stb_image.h"
        $93 = $92&255; //@line 1371 "./libraries/stb/stb_image.h"
        $94 = ($87*77)|0; //@line 1337 "./libraries/stb/stb_image.h"
        $95 = ($90*150)|0; //@line 1337 "./libraries/stb/stb_image.h"
        $96 = (($95) + ($94))|0; //@line 1337 "./libraries/stb/stb_image.h"
        $97 = ($93*29)|0; //@line 1337 "./libraries/stb/stb_image.h"
        $98 = (($96) + ($97))|0; //@line 1337 "./libraries/stb/stb_image.h"
        $99 = $98 >>> 8; //@line 1337 "./libraries/stb/stb_image.h"
        $100 = $99&255; //@line 1337 "./libraries/stb/stb_image.h"
        HEAP8[$dest$841>>0] = $100; //@line 1371 "./libraries/stb/stb_image.h"
        $101 = ((($dest$841)) + 1|0); //@line 1371 "./libraries/stb/stb_image.h"
        HEAP8[$101>>0] = -1; //@line 1371 "./libraries/stb/stb_image.h"
        $102 = ((($src$840)) + 3|0); //@line 1371 "./libraries/stb/stb_image.h"
        $103 = ((($dest$841)) + 2|0); //@line 1371 "./libraries/stb/stb_image.h"
        $i$8 = (($i$842) + -1)|0; //@line 1371 "./libraries/stb/stb_image.h"
        $104 = ($i$8|0)>(-1); //@line 1371 "./libraries/stb/stb_image.h"
        if ($104) {
         $dest$841 = $103;$i$842 = $i$8;$src$840 = $102;
        } else {
         break;
        }
       }
      }
      break;
     }
     case 33:  {
      if ($14) {
       $dest$936 = $19;$i$937 = $i$079;$src$935 = $17;
       while(1) {
        $105 = HEAP8[$src$935>>0]|0; //@line 1372 "./libraries/stb/stb_image.h"
        $106 = $105&255; //@line 1372 "./libraries/stb/stb_image.h"
        $107 = ((($src$935)) + 1|0); //@line 1372 "./libraries/stb/stb_image.h"
        $108 = HEAP8[$107>>0]|0; //@line 1372 "./libraries/stb/stb_image.h"
        $109 = $108&255; //@line 1372 "./libraries/stb/stb_image.h"
        $110 = ((($src$935)) + 2|0); //@line 1372 "./libraries/stb/stb_image.h"
        $111 = HEAP8[$110>>0]|0; //@line 1372 "./libraries/stb/stb_image.h"
        $112 = $111&255; //@line 1372 "./libraries/stb/stb_image.h"
        $113 = ($106*77)|0; //@line 1337 "./libraries/stb/stb_image.h"
        $114 = ($109*150)|0; //@line 1337 "./libraries/stb/stb_image.h"
        $115 = (($114) + ($113))|0; //@line 1337 "./libraries/stb/stb_image.h"
        $116 = ($112*29)|0; //@line 1337 "./libraries/stb/stb_image.h"
        $117 = (($115) + ($116))|0; //@line 1337 "./libraries/stb/stb_image.h"
        $118 = $117 >>> 8; //@line 1337 "./libraries/stb/stb_image.h"
        $119 = $118&255; //@line 1337 "./libraries/stb/stb_image.h"
        HEAP8[$dest$936>>0] = $119; //@line 1372 "./libraries/stb/stb_image.h"
        $120 = ((($src$935)) + 4|0); //@line 1372 "./libraries/stb/stb_image.h"
        $121 = ((($dest$936)) + 1|0); //@line 1372 "./libraries/stb/stb_image.h"
        $i$9 = (($i$937) + -1)|0; //@line 1372 "./libraries/stb/stb_image.h"
        $122 = ($i$9|0)>(-1); //@line 1372 "./libraries/stb/stb_image.h"
        if ($122) {
         $dest$936 = $121;$i$937 = $i$9;$src$935 = $120;
        } else {
         break;
        }
       }
      }
      break;
     }
     case 34:  {
      if ($14) {
       $dest$1031 = $19;$i$1032 = $i$079;$src$1030 = $17;
       while(1) {
        $123 = HEAP8[$src$1030>>0]|0; //@line 1373 "./libraries/stb/stb_image.h"
        $124 = $123&255; //@line 1373 "./libraries/stb/stb_image.h"
        $125 = ((($src$1030)) + 1|0); //@line 1373 "./libraries/stb/stb_image.h"
        $126 = HEAP8[$125>>0]|0; //@line 1373 "./libraries/stb/stb_image.h"
        $127 = $126&255; //@line 1373 "./libraries/stb/stb_image.h"
        $128 = ((($src$1030)) + 2|0); //@line 1373 "./libraries/stb/stb_image.h"
        $129 = HEAP8[$128>>0]|0; //@line 1373 "./libraries/stb/stb_image.h"
        $130 = $129&255; //@line 1373 "./libraries/stb/stb_image.h"
        $131 = ($124*77)|0; //@line 1337 "./libraries/stb/stb_image.h"
        $132 = ($127*150)|0; //@line 1337 "./libraries/stb/stb_image.h"
        $133 = (($132) + ($131))|0; //@line 1337 "./libraries/stb/stb_image.h"
        $134 = ($130*29)|0; //@line 1337 "./libraries/stb/stb_image.h"
        $135 = (($133) + ($134))|0; //@line 1337 "./libraries/stb/stb_image.h"
        $136 = $135 >>> 8; //@line 1337 "./libraries/stb/stb_image.h"
        $137 = $136&255; //@line 1337 "./libraries/stb/stb_image.h"
        HEAP8[$dest$1031>>0] = $137; //@line 1373 "./libraries/stb/stb_image.h"
        $138 = ((($src$1030)) + 3|0); //@line 1373 "./libraries/stb/stb_image.h"
        $139 = HEAP8[$138>>0]|0; //@line 1373 "./libraries/stb/stb_image.h"
        $140 = ((($dest$1031)) + 1|0); //@line 1373 "./libraries/stb/stb_image.h"
        HEAP8[$140>>0] = $139; //@line 1373 "./libraries/stb/stb_image.h"
        $141 = ((($src$1030)) + 4|0); //@line 1373 "./libraries/stb/stb_image.h"
        $142 = ((($dest$1031)) + 2|0); //@line 1373 "./libraries/stb/stb_image.h"
        $i$10 = (($i$1032) + -1)|0; //@line 1373 "./libraries/stb/stb_image.h"
        $143 = ($i$10|0)>(-1); //@line 1373 "./libraries/stb/stb_image.h"
        if ($143) {
         $dest$1031 = $142;$i$1032 = $i$10;$src$1030 = $141;
        } else {
         break;
        }
       }
      }
      break;
     }
     case 35:  {
      if ($14) {
       $dest$1127 = $19;$i$1128 = $i$079;$src$1126 = $17;
       while(1) {
        $144 = HEAP8[$src$1126>>0]|0; //@line 1374 "./libraries/stb/stb_image.h"
        HEAP8[$dest$1127>>0] = $144; //@line 1374 "./libraries/stb/stb_image.h"
        $145 = ((($src$1126)) + 1|0); //@line 1374 "./libraries/stb/stb_image.h"
        $146 = HEAP8[$145>>0]|0; //@line 1374 "./libraries/stb/stb_image.h"
        $147 = ((($dest$1127)) + 1|0); //@line 1374 "./libraries/stb/stb_image.h"
        HEAP8[$147>>0] = $146; //@line 1374 "./libraries/stb/stb_image.h"
        $148 = ((($src$1126)) + 2|0); //@line 1374 "./libraries/stb/stb_image.h"
        $149 = HEAP8[$148>>0]|0; //@line 1374 "./libraries/stb/stb_image.h"
        $150 = ((($dest$1127)) + 2|0); //@line 1374 "./libraries/stb/stb_image.h"
        HEAP8[$150>>0] = $149; //@line 1374 "./libraries/stb/stb_image.h"
        $151 = ((($src$1126)) + 4|0); //@line 1374 "./libraries/stb/stb_image.h"
        $152 = ((($dest$1127)) + 3|0); //@line 1374 "./libraries/stb/stb_image.h"
        $i$11 = (($i$1128) + -1)|0; //@line 1374 "./libraries/stb/stb_image.h"
        $153 = ($i$11|0)>(-1); //@line 1374 "./libraries/stb/stb_image.h"
        if ($153) {
         $dest$1127 = $152;$i$1128 = $i$11;$src$1126 = $151;
        } else {
         break;
        }
       }
      }
      break;
     }
     default: {
      break L16;
     }
     }
    } while(0);
    $159 = (($j$084) + 1)|0; //@line 1354 "./libraries/stb/stb_image.h"
    $160 = ($159|0)<($y|0); //@line 1354 "./libraries/stb/stb_image.h"
    if ($160) {
     $j$084 = $159;
    } else {
     break L14;
    }
   }
   $154 = (___cxa_allocate_exception(16)|0); //@line 27 "modules/image/magpie/STBHandler.cpp"
   __THREW__ = 0;
   HEAP32[$vararg_buffer1>>2] = 604546; //@line 27 "modules/image/magpie/STBHandler.cpp"
   invoke_viii(3349,($154|0),(441193|0),($vararg_buffer1|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
   $155 = __THREW__; __THREW__ = 0;
   $156 = $155&1;
   if ($156) {
    $157 = ___cxa_find_matching_catch_2()|0;
    $158 = tempRet0;
    ___cxa_free_exception(($154|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
    ___resumeException($157|0);
    // unreachable;
   } else {
    ___cxa_throw(($154|0),(96|0),(344|0)); //@line 27 "modules/image/magpie/STBHandler.cpp"
    // unreachable; //@line 27 "modules/image/magpie/STBHandler.cpp"
   }
  }
 } while(0);
 _free($data); //@line 1380 "./libraries/stb/stb_image.h"
 $$0 = $9;
 STACKTOP = sp;return ($$0|0); //@line 1382 "./libraries/stb/stb_image.h"
}
function __ZN4love5image6magpie10STBHandler6encodeERKNS1_13FormatHandler12DecodedImageENS0_9ImageData13EncodedFormatE($agg$result,$this,$img,$format) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $img = $img|0;
 $format = $format|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0;
 var $x$02$us = 0, $y$03$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[$this>>2]|0; //@line 87 "modules/image/magpie/STBHandler.cpp"
 $1 = ((($0)) + 20|0); //@line 87 "modules/image/magpie/STBHandler.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 87 "modules/image/magpie/STBHandler.cpp"
 $3 = (FUNCTION_TABLE_iii[$2 & 8191]($this,$format)|0); //@line 87 "modules/image/magpie/STBHandler.cpp"
 if (!($3)) {
  $4 = (___cxa_allocate_exception(16)|0); //@line 88 "modules/image/magpie/STBHandler.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($4|0),(441551|0),($vararg_buffer|0)); //@line 88 "modules/image/magpie/STBHandler.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   ___cxa_throw(($4|0),(96|0),(344|0)); //@line 88 "modules/image/magpie/STBHandler.cpp"
   // unreachable; //@line 88 "modules/image/magpie/STBHandler.cpp"
  }
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  ___cxa_free_exception(($4|0)); //@line 88 "modules/image/magpie/STBHandler.cpp"
  $$0 = $7;$$01 = $8;
  ___resumeException($$0|0);
  // unreachable;
 }
 $9 = ((($agg$result)) + 4|0); //@line 57 "modules/image/magpie/FormatHandler.h"
 $10 = HEAP32[$img>>2]|0; //@line 134 "modules/image/magpie/STBHandler.cpp"
 $11 = ((($img)) + 4|0); //@line 98 "modules/image/magpie/STBHandler.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 132 "modules/image/magpie/STBHandler.cpp"
 $13 = $10 << 2; //@line 98 "modules/image/magpie/STBHandler.cpp"
 $14 = Math_imul($13, $12)|0; //@line 98 "modules/image/magpie/STBHandler.cpp"
 $15 = (($14) + 18)|0; //@line 98 "modules/image/magpie/STBHandler.cpp"
 HEAP32[$agg$result>>2] = $15; //@line 98 "modules/image/magpie/STBHandler.cpp"
 $16 = (_malloc($15)|0); //@line 131 "modules/image/magpie/STBHandler.cpp"
 HEAP32[$9>>2] = $16; //@line 105 "modules/image/magpie/STBHandler.cpp"
 $17 = ($16|0)==(0|0); //@line 107 "modules/image/magpie/STBHandler.cpp"
 if ($17) {
  $18 = (___cxa_allocate_exception(16)|0); //@line 108 "modules/image/magpie/STBHandler.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($18|0),(507610|0),($vararg_buffer1|0)); //@line 108 "modules/image/magpie/STBHandler.cpp"
  $19 = __THREW__; __THREW__ = 0;
  $20 = $19&1;
  if (!($20)) {
   ___cxa_throw(($18|0),(96|0),(344|0)); //@line 108 "modules/image/magpie/STBHandler.cpp"
   // unreachable; //@line 108 "modules/image/magpie/STBHandler.cpp"
  }
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  ___cxa_free_exception(($18|0)); //@line 108 "modules/image/magpie/STBHandler.cpp"
  $$0 = $21;$$01 = $22;
  ___resumeException($$0|0);
  // unreachable;
 }
 HEAP8[$16>>0] = 0; //@line 111 "modules/image/magpie/STBHandler.cpp"
 $23 = ((($16)) + 1|0); //@line 112 "modules/image/magpie/STBHandler.cpp"
 HEAP8[$23>>0] = 0; //@line 112 "modules/image/magpie/STBHandler.cpp"
 $24 = ((($16)) + 2|0); //@line 113 "modules/image/magpie/STBHandler.cpp"
 HEAP8[$24>>0] = 2; //@line 113 "modules/image/magpie/STBHandler.cpp"
 $25 = ((($16)) + 3|0); //@line 114 "modules/image/magpie/STBHandler.cpp"
 $26 = ((($16)) + 7|0); //@line 116 "modules/image/magpie/STBHandler.cpp"
 HEAP8[$25>>0]=0&255;HEAP8[$25+1>>0]=(0>>8)&255;HEAP8[$25+2>>0]=(0>>16)&255;HEAP8[$25+3>>0]=0>>24; //@line 114 "modules/image/magpie/STBHandler.cpp"
 HEAP8[$26>>0] = 32; //@line 116 "modules/image/magpie/STBHandler.cpp"
 $27 = ((($16)) + 8|0); //@line 117 "modules/image/magpie/STBHandler.cpp"
 $28 = $10&255; //@line 120 "modules/image/magpie/STBHandler.cpp"
 $29 = ((($16)) + 12|0); //@line 120 "modules/image/magpie/STBHandler.cpp"
 HEAP8[$27>>0]=0&255;HEAP8[$27+1>>0]=(0>>8)&255;HEAP8[$27+2>>0]=(0>>16)&255;HEAP8[$27+3>>0]=0>>24; //@line 117 "modules/image/magpie/STBHandler.cpp"
 HEAP8[$29>>0] = $28; //@line 120 "modules/image/magpie/STBHandler.cpp"
 $30 = $10 >>> 8; //@line 121 "modules/image/magpie/STBHandler.cpp"
 $31 = $30&255; //@line 121 "modules/image/magpie/STBHandler.cpp"
 $32 = ((($16)) + 13|0); //@line 121 "modules/image/magpie/STBHandler.cpp"
 HEAP8[$32>>0] = $31; //@line 121 "modules/image/magpie/STBHandler.cpp"
 $33 = $12&255; //@line 122 "modules/image/magpie/STBHandler.cpp"
 $34 = ((($16)) + 14|0); //@line 122 "modules/image/magpie/STBHandler.cpp"
 HEAP8[$34>>0] = $33; //@line 122 "modules/image/magpie/STBHandler.cpp"
 $35 = $12 >>> 8; //@line 123 "modules/image/magpie/STBHandler.cpp"
 $36 = $35&255; //@line 123 "modules/image/magpie/STBHandler.cpp"
 $37 = ((($16)) + 15|0); //@line 123 "modules/image/magpie/STBHandler.cpp"
 HEAP8[$37>>0] = $36; //@line 123 "modules/image/magpie/STBHandler.cpp"
 $38 = ((($16)) + 16|0); //@line 124 "modules/image/magpie/STBHandler.cpp"
 HEAP8[$38>>0] = 32; //@line 124 "modules/image/magpie/STBHandler.cpp"
 $39 = ((($16)) + 17|0); //@line 125 "modules/image/magpie/STBHandler.cpp"
 HEAP8[$39>>0] = 32; //@line 125 "modules/image/magpie/STBHandler.cpp"
 $40 = ((($16)) + 18|0); //@line 128 "modules/image/magpie/STBHandler.cpp"
 $41 = ((($img)) + 12|0); //@line 128 "modules/image/magpie/STBHandler.cpp"
 $42 = HEAP32[$41>>2]|0; //@line 128 "modules/image/magpie/STBHandler.cpp"
 _memcpy(($40|0),($42|0),($14|0))|0; //@line 128 "modules/image/magpie/STBHandler.cpp"
 $43 = ($12|0)>(0); //@line 132 "modules/image/magpie/STBHandler.cpp"
 $44 = ($10|0)>(0); //@line 134 "modules/image/magpie/STBHandler.cpp"
 $or$cond = $43 & $44; //@line 132 "modules/image/magpie/STBHandler.cpp"
 if ($or$cond) {
  $y$03$us = 0;
 } else {
  STACKTOP = sp;return; //@line 144 "modules/image/magpie/STBHandler.cpp"
 }
 while(1) {
  $46 = Math_imul($10, $y$03$us)|0; //@line 139 "modules/image/magpie/STBHandler.cpp"
  $x$02$us = 0;
  while(1) {
   $45 = (($46) + ($x$02$us))|0; //@line 136 "modules/image/magpie/STBHandler.cpp"
   $47 = (($40) + ($45<<2)|0); //@line 136 "modules/image/magpie/STBHandler.cpp"
   $48 = HEAP8[$47>>0]|0; //@line 136 "modules/image/magpie/STBHandler.cpp"
   $49 = (((($40) + ($45<<2)|0)) + 2|0); //@line 137 "modules/image/magpie/STBHandler.cpp"
   $50 = HEAP8[$49>>0]|0; //@line 137 "modules/image/magpie/STBHandler.cpp"
   HEAP8[$47>>0] = $50; //@line 138 "modules/image/magpie/STBHandler.cpp"
   HEAP8[$49>>0] = $48; //@line 139 "modules/image/magpie/STBHandler.cpp"
   $51 = (($x$02$us) + 1)|0; //@line 134 "modules/image/magpie/STBHandler.cpp"
   $exitcond = ($51|0)==($10|0); //@line 134 "modules/image/magpie/STBHandler.cpp"
   if ($exitcond) {
    break;
   } else {
    $x$02$us = $51;
   }
  }
  $52 = (($y$03$us) + 1)|0; //@line 132 "modules/image/magpie/STBHandler.cpp"
  $53 = ($52|0)<($12|0); //@line 132 "modules/image/magpie/STBHandler.cpp"
  if ($53) {
   $y$03$us = $52;
  } else {
   break;
  }
 }
 STACKTOP = sp;return; //@line 144 "modules/image/magpie/STBHandler.cpp"
}
function __ZN4love5image6magpie10STBHandler4freeEPh($this,$mem) {
 $this = $this|0;
 $mem = $mem|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _free($mem); //@line 920 "./libraries/stb/stb_image.h"
 return; //@line 151 "modules/image/magpie/STBHandler.cpp"
}
function __ZN4love5image9ImageDataD2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (26076); //@line 36 "modules/image/ImageData.cpp"
 $0 = ((($this)) + 20|0); //@line 37 "modules/image/ImageData.cpp"
 __ZN4love6thread8MutexRefD2Ev($0); //@line 37 "modules/image/ImageData.cpp"
 return; //@line 37 "modules/image/ImageData.cpp"
}
function __ZN4love5image9ImageDataD0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (26076); //@line 36 "modules/image/ImageData.cpp"
 $0 = ((($this)) + 20|0); //@line 37 "modules/image/ImageData.cpp"
 __ZN4love6thread8MutexRefD2Ev($0); //@line 37 "modules/image/ImageData.cpp"
 __ZdlPv($this); //@line 36 "modules/image/ImageData.cpp"
 return; //@line 37 "modules/image/ImageData.cpp"
}
function __ZNK4love5image9ImageData7getDataEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 46 "modules/image/ImageData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 46 "modules/image/ImageData.cpp"
 return ($1|0); //@line 46 "modules/image/ImageData.cpp"
}
function __ZNK4love5image9ImageData7getSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 56 "modules/image/ImageData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 56 "modules/image/ImageData.cpp"
 $2 = ((($this)) + 12|0); //@line 61 "modules/image/ImageData.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 61 "modules/image/ImageData.cpp"
 $4 = $1 << 2; //@line 41 "modules/image/ImageData.cpp"
 $5 = Math_imul($4, $3)|0; //@line 41 "modules/image/ImageData.cpp"
 return ($5|0); //@line 41 "modules/image/ImageData.cpp"
}
function __GLOBAL__sub_I_ImageData_cpp() {
 var $$02$i$i$1$i = 0, $$02$i$i$i = 0, $$lcssa$i$1$i = 0, $$lcssa$i$i = 0, $$lcssa10$i$1$i = 0, $$lcssa10$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i$i$1$i = 0, $hash$0$lcssa$i$i$i = 0;
 var $hash$01$i$i$1$i = 0, $hash$01$i$i$i = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 HEAP8[(635032)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(635044)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(635056)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(635068)>>0] = 0; //@line 155 "./common/StringMap.h"
 $0 = (635072); //@line 44 "./common/StringMap.h"
 $1 = $0; //@line 44 "./common/StringMap.h"
 HEAP32[$1>>2] = 0; //@line 44 "./common/StringMap.h"
 $2 = (($0) + 4)|0; //@line 44 "./common/StringMap.h"
 $3 = $2; //@line 44 "./common/StringMap.h"
 HEAP32[$3>>2] = 0; //@line 44 "./common/StringMap.h"
 $4 = HEAP32[6526]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(26108)>>2]|0; //@line 49 "./common/StringMap.h"
 $6 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
 $7 = ($6<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($7) {
  $hash$0$lcssa$i$i$i = 5381;
 } else {
  $$02$i$i$i = $4;$9 = $6;$hash$01$i$i$i = 5381;
  while(1) {
   $8 = $9 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $10 = ((($$02$i$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $11 = ($hash$01$i$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $12 = (($11) + ($8))|0; //@line 143 "./common/StringMap.h"
   $13 = HEAP8[$10>>0]|0; //@line 142 "./common/StringMap.h"
   $14 = ($13<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($14) {
    $hash$0$lcssa$i$i$i = $12;
    break;
   } else {
    $$02$i$i$i = $10;$9 = $13;$hash$01$i$i$i = $12;
   }
  }
 }
 $15 = $hash$0$lcssa$i$i$i & 3; //@line 112 "./common/StringMap.h"
 $16 = (((635024 + (($15*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $17 = HEAP8[$16>>0]|0; //@line 114 "./common/StringMap.h"
 $18 = ($17<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($18) {
  $$lcssa$i$i = $15;$$lcssa10$i$i = $16;
  label = 5;
 } else {
  $19 = (($hash$0$lcssa$i$i$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $20 = $19 & 3; //@line 112 "./common/StringMap.h"
  $21 = (((635024 + (($20*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $22 = HEAP8[$21>>0]|0; //@line 114 "./common/StringMap.h"
  $23 = ($22<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($23) {
   $$lcssa$i$i = $20;$$lcssa10$i$i = $21;
   label = 5;
  } else {
   $28 = (($hash$0$lcssa$i$i$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $29 = $28 & 3; //@line 112 "./common/StringMap.h"
   $30 = (((635024 + (($29*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $31 = HEAP8[$30>>0]|0; //@line 114 "./common/StringMap.h"
   $32 = ($31<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($32) {
    $$lcssa$i$i = $29;$$lcssa10$i$i = $30;
    label = 5;
   } else {
    $33 = (($hash$0$lcssa$i$i$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $34 = $33 & 3; //@line 112 "./common/StringMap.h"
    $35 = (((635024 + (($34*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $36 = HEAP8[$35>>0]|0; //@line 114 "./common/StringMap.h"
    $37 = ($36<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($37) {
     $$lcssa$i$i = $34;$$lcssa10$i$i = $35;
     label = 5;
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10$i$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $24 = (635024 + (($$lcssa$i$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$24>>2] = $4; //@line 118 "./common/StringMap.h"
  $25 = (((635024 + (($$lcssa$i$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$25>>2] = $5; //@line 119 "./common/StringMap.h"
 }
 $26 = ($5>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($26) {
  HEAP32[$vararg_buffer>>2] = $4; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $5; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
 } else {
  $27 = (635072 + ($5<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$27>>2] = $4; //@line 132 "./common/StringMap.h"
 }
 $38 = HEAP32[(26112)>>2]|0; //@line 49 "./common/StringMap.h"
 $39 = HEAP32[(26116)>>2]|0; //@line 49 "./common/StringMap.h"
 $40 = HEAP8[$38>>0]|0; //@line 142 "./common/StringMap.h"
 $41 = ($40<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($41) {
  $hash$0$lcssa$i$i$1$i = 5381;
 } else {
  $$02$i$i$1$i = $38;$43 = $40;$hash$01$i$i$1$i = 5381;
  while(1) {
   $42 = $43 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $44 = ((($$02$i$i$1$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $45 = ($hash$01$i$i$1$i*33)|0; //@line 143 "./common/StringMap.h"
   $46 = (($45) + ($42))|0; //@line 143 "./common/StringMap.h"
   $47 = HEAP8[$44>>0]|0; //@line 142 "./common/StringMap.h"
   $48 = ($47<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($48) {
    $hash$0$lcssa$i$i$1$i = $46;
    break;
   } else {
    $$02$i$i$1$i = $44;$43 = $47;$hash$01$i$i$1$i = $46;
   }
  }
 }
 $49 = $hash$0$lcssa$i$i$1$i & 3; //@line 112 "./common/StringMap.h"
 $50 = (((635024 + (($49*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $51 = HEAP8[$50>>0]|0; //@line 114 "./common/StringMap.h"
 $52 = ($51<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($52) {
  $$lcssa$i$1$i = $49;$$lcssa10$i$1$i = $50;
  label = 17;
 } else {
  $53 = (($hash$0$lcssa$i$i$1$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $54 = $53 & 3; //@line 112 "./common/StringMap.h"
  $55 = (((635024 + (($54*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $56 = HEAP8[$55>>0]|0; //@line 114 "./common/StringMap.h"
  $57 = ($56<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($57) {
   $$lcssa$i$1$i = $54;$$lcssa10$i$1$i = $55;
   label = 17;
  } else {
   $58 = (($hash$0$lcssa$i$i$1$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $59 = $58 & 3; //@line 112 "./common/StringMap.h"
   $60 = (((635024 + (($59*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $61 = HEAP8[$60>>0]|0; //@line 114 "./common/StringMap.h"
   $62 = ($61<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($62) {
    $$lcssa$i$1$i = $59;$$lcssa10$i$1$i = $60;
    label = 17;
   } else {
    $63 = (($hash$0$lcssa$i$i$1$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $64 = $63 & 3; //@line 112 "./common/StringMap.h"
    $65 = (((635024 + (($64*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $66 = HEAP8[$65>>0]|0; //@line 114 "./common/StringMap.h"
    $67 = ($66<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($67) {
     $$lcssa$i$1$i = $64;$$lcssa10$i$1$i = $65;
     label = 17;
    }
   }
  }
 }
 if ((label|0) == 17) {
  HEAP8[$$lcssa10$i$1$i>>0] = 1; //@line 117 "./common/StringMap.h"
  $68 = (635024 + (($$lcssa$i$1$i*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$68>>2] = $38; //@line 118 "./common/StringMap.h"
  $69 = (((635024 + (($$lcssa$i$1$i*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$69>>2] = $39; //@line 119 "./common/StringMap.h"
 }
 $70 = ($39>>>0)>(1); //@line 126 "./common/StringMap.h"
 if ($70) {
  HEAP32[$vararg_buffer2>>2] = $38; //@line 128 "./common/StringMap.h"
  $vararg_ptr5 = ((($vararg_buffer2)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr5>>2] = $39; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer2)|0); //@line 128 "./common/StringMap.h"
  STACKTOP = sp;return;
 } else {
  $71 = (635072 + ($39<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$71>>2] = $38; //@line 132 "./common/StringMap.h"
  STACKTOP = sp;return;
 }
}
function __ZN4love5image9ImageDataC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 37 "./common/Data.h"
 HEAP32[$this>>2] = (26076); //@line 32 "modules/image/ImageData.cpp"
 $0 = ((($this)) + 16|0); //@line 31 "modules/image/ImageData.cpp"
 HEAP32[$0>>2] = 0; //@line 31 "modules/image/ImageData.cpp"
 $1 = ((($this)) + 20|0); //@line 30 "modules/image/ImageData.cpp"
 __THREW__ = 0;
 invoke_vi(3788,($1|0)); //@line 30 "modules/image/ImageData.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  ___resumeException($4|0);
  // unreachable;
 } else {
  return; //@line 33 "modules/image/ImageData.cpp"
 }
}
function __ZNK4love5image9ImageData8getWidthEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 56 "modules/image/ImageData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 56 "modules/image/ImageData.cpp"
 return ($1|0); //@line 56 "modules/image/ImageData.cpp"
}
function __ZNK4love5image9ImageData9getHeightEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 61 "modules/image/ImageData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 61 "modules/image/ImageData.cpp"
 return ($1|0); //@line 61 "modules/image/ImageData.cpp"
}
function __ZNK4love5image9ImageData6insideEii($this,$x,$y) {
 $this = $this|0;
 $x = $x|0;
 $y = $y|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($x|0)>(-1); //@line 51 "modules/image/ImageData.cpp"
 if (!($0)) {
  $8 = 0;
  return ($8|0); //@line 51 "modules/image/ImageData.cpp"
 }
 $1 = ((($this)) + 8|0); //@line 56 "modules/image/ImageData.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 56 "modules/image/ImageData.cpp"
 $3 = ($2|0)>($x|0); //@line 51 "modules/image/ImageData.cpp"
 $4 = ($y|0)>(-1); //@line 51 "modules/image/ImageData.cpp"
 $or$cond = $4 & $3; //@line 51 "modules/image/ImageData.cpp"
 if (!($or$cond)) {
  $8 = 0;
  return ($8|0); //@line 51 "modules/image/ImageData.cpp"
 }
 $5 = ((($this)) + 12|0); //@line 61 "modules/image/ImageData.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 61 "modules/image/ImageData.cpp"
 $7 = ($6|0)>($y|0); //@line 51 "modules/image/ImageData.cpp"
 $8 = $7;
 return ($8|0); //@line 51 "modules/image/ImageData.cpp"
}
function __ZN4love5image9ImageData8setPixelEiiNS0_5pixelE($this,$x,$y,$c) {
 $this = $this|0;
 $x = $x|0;
 $y = $y|0;
 $c = $c|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lock = 0, $or$cond$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $lock = sp + 4|0;
 $0 = ($x|0)>(-1); //@line 51 "modules/image/ImageData.cpp"
 if ($0) {
  $1 = ((($this)) + 8|0); //@line 56 "modules/image/ImageData.cpp"
  $2 = HEAP32[$1>>2]|0; //@line 56 "modules/image/ImageData.cpp"
  $3 = ($2|0)>($x|0); //@line 51 "modules/image/ImageData.cpp"
  $4 = ($y|0)>(-1); //@line 51 "modules/image/ImageData.cpp"
  $or$cond$i = $4 & $3; //@line 51 "modules/image/ImageData.cpp"
  if ($or$cond$i) {
   $5 = ((($this)) + 12|0); //@line 61 "modules/image/ImageData.cpp"
   $6 = HEAP32[$5>>2]|0; //@line 61 "modules/image/ImageData.cpp"
   $7 = ($6|0)>($y|0); //@line 51 "modules/image/ImageData.cpp"
   if ($7) {
    $13 = ((($this)) + 20|0); //@line 69 "modules/image/ImageData.cpp"
    $14 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($13)|0); //@line 69 "modules/image/ImageData.cpp"
    __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$14); //@line 69 "modules/image/ImageData.cpp"
    $15 = HEAP32[$this>>2]|0; //@line 71 "modules/image/ImageData.cpp"
    $16 = ((($15)) + 16|0); //@line 71 "modules/image/ImageData.cpp"
    $17 = HEAP32[$16>>2]|0; //@line 71 "modules/image/ImageData.cpp"
    __THREW__ = 0;
    $18 = (invoke_ii($17|0,($this|0))|0); //@line 71 "modules/image/ImageData.cpp"
    $19 = __THREW__; __THREW__ = 0;
    $20 = $19&1;
    if (!($20)) {
     $21 = HEAP32[$1>>2]|0; //@line 72 "modules/image/ImageData.cpp"
     $22 = Math_imul($21, $y)|0; //@line 72 "modules/image/ImageData.cpp"
     $23 = (($22) + ($x))|0; //@line 72 "modules/image/ImageData.cpp"
     $24 = (($18) + ($23<<2)|0); //@line 72 "modules/image/ImageData.cpp"
     $25 = HEAPU8[$c>>0]|(HEAPU8[$c+1>>0]<<8)|(HEAPU8[$c+2>>0]<<16)|(HEAPU8[$c+3>>0]<<24); //@line 72 "modules/image/ImageData.cpp"
     HEAP8[$24>>0]=$25&255;HEAP8[$24+1>>0]=($25>>8)&255;HEAP8[$24+2>>0]=($25>>16)&255;HEAP8[$24+3>>0]=$25>>24; //@line 72 "modules/image/ImageData.cpp"
     __ZN4love6thread4LockD2Ev($lock); //@line 73 "modules/image/ImageData.cpp"
     STACKTOP = sp;return; //@line 73 "modules/image/ImageData.cpp"
    }
    $26 = ___cxa_find_matching_catch_2()|0;
    $27 = tempRet0;
    __ZN4love6thread4LockD2Ev($lock); //@line 73 "modules/image/ImageData.cpp"
    $$0 = $26;$$01 = $27;
    ___resumeException($$0|0);
    // unreachable;
   }
  }
 }
 $8 = (___cxa_allocate_exception(16)|0); //@line 67 "modules/image/ImageData.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($8|0),(441599|0),($vararg_buffer|0)); //@line 67 "modules/image/ImageData.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if (!($10)) {
  ___cxa_throw(($8|0),(96|0),(344|0)); //@line 67 "modules/image/ImageData.cpp"
  // unreachable; //@line 67 "modules/image/ImageData.cpp"
 }
 $11 = ___cxa_find_matching_catch_2()|0;
 $12 = tempRet0;
 ___cxa_free_exception(($8|0)); //@line 67 "modules/image/ImageData.cpp"
 $$0 = $11;$$01 = $12;
 ___resumeException($$0|0);
 // unreachable;
}
function __ZN4love5image9ImageData14setPixelUnsafeEiiNS0_5pixelE($this,$x,$y,$c) {
 $this = $this|0;
 $x = $x|0;
 $y = $y|0;
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 77 "modules/image/ImageData.cpp"
 $1 = ((($0)) + 16|0); //@line 77 "modules/image/ImageData.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 77 "modules/image/ImageData.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 77 "modules/image/ImageData.cpp"
 $4 = ((($this)) + 8|0); //@line 78 "modules/image/ImageData.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 78 "modules/image/ImageData.cpp"
 $6 = Math_imul($5, $y)|0; //@line 78 "modules/image/ImageData.cpp"
 $7 = (($6) + ($x))|0; //@line 78 "modules/image/ImageData.cpp"
 $8 = (($3) + ($7<<2)|0); //@line 78 "modules/image/ImageData.cpp"
 $9 = HEAPU8[$c>>0]|(HEAPU8[$c+1>>0]<<8)|(HEAPU8[$c+2>>0]<<16)|(HEAPU8[$c+3>>0]<<24); //@line 78 "modules/image/ImageData.cpp"
 HEAP8[$8>>0]=$9&255;HEAP8[$8+1>>0]=($9>>8)&255;HEAP8[$8+2>>0]=($9>>16)&255;HEAP8[$8+3>>0]=$9>>24; //@line 78 "modules/image/ImageData.cpp"
 return; //@line 79 "modules/image/ImageData.cpp"
}
function __ZNK4love5image9ImageData8getPixelEii($agg$result,$this,$x,$y) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $x = $x|0;
 $y = $y|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lock = 0, $or$cond$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $lock = sp + 4|0;
 $0 = ($x|0)>(-1); //@line 51 "modules/image/ImageData.cpp"
 if ($0) {
  $1 = ((($this)) + 8|0); //@line 56 "modules/image/ImageData.cpp"
  $2 = HEAP32[$1>>2]|0; //@line 56 "modules/image/ImageData.cpp"
  $3 = ($2|0)>($x|0); //@line 51 "modules/image/ImageData.cpp"
  $4 = ($y|0)>(-1); //@line 51 "modules/image/ImageData.cpp"
  $or$cond$i = $4 & $3; //@line 51 "modules/image/ImageData.cpp"
  if ($or$cond$i) {
   $5 = ((($this)) + 12|0); //@line 61 "modules/image/ImageData.cpp"
   $6 = HEAP32[$5>>2]|0; //@line 61 "modules/image/ImageData.cpp"
   $7 = ($6|0)>($y|0); //@line 51 "modules/image/ImageData.cpp"
   if ($7) {
    $13 = ((($this)) + 20|0); //@line 86 "modules/image/ImageData.cpp"
    $14 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($13)|0); //@line 86 "modules/image/ImageData.cpp"
    __ZN4love6thread4LockC2EPNS0_5MutexE($lock,$14); //@line 86 "modules/image/ImageData.cpp"
    $15 = HEAP32[$this>>2]|0; //@line 88 "modules/image/ImageData.cpp"
    $16 = ((($15)) + 16|0); //@line 88 "modules/image/ImageData.cpp"
    $17 = HEAP32[$16>>2]|0; //@line 88 "modules/image/ImageData.cpp"
    __THREW__ = 0;
    $18 = (invoke_ii($17|0,($this|0))|0); //@line 88 "modules/image/ImageData.cpp"
    $19 = __THREW__; __THREW__ = 0;
    $20 = $19&1;
    if (!($20)) {
     $21 = HEAP32[$1>>2]|0; //@line 89 "modules/image/ImageData.cpp"
     $22 = Math_imul($21, $y)|0; //@line 89 "modules/image/ImageData.cpp"
     $23 = (($22) + ($x))|0; //@line 89 "modules/image/ImageData.cpp"
     $24 = (($18) + ($23<<2)|0); //@line 89 "modules/image/ImageData.cpp"
     $25 = HEAPU8[$24>>0]|(HEAPU8[$24+1>>0]<<8)|(HEAPU8[$24+2>>0]<<16)|(HEAPU8[$24+3>>0]<<24); //@line 89 "modules/image/ImageData.cpp"
     HEAP8[$agg$result>>0]=$25&255;HEAP8[$agg$result+1>>0]=($25>>8)&255;HEAP8[$agg$result+2>>0]=($25>>16)&255;HEAP8[$agg$result+3>>0]=$25>>24; //@line 89 "modules/image/ImageData.cpp"
     __ZN4love6thread4LockD2Ev($lock); //@line 90 "modules/image/ImageData.cpp"
     STACKTOP = sp;return; //@line 90 "modules/image/ImageData.cpp"
    }
    $26 = ___cxa_find_matching_catch_2()|0;
    $27 = tempRet0;
    __ZN4love6thread4LockD2Ev($lock); //@line 90 "modules/image/ImageData.cpp"
    $$0 = $26;$$01 = $27;
    ___resumeException($$0|0);
    // unreachable;
   }
  }
 }
 $8 = (___cxa_allocate_exception(16)|0); //@line 84 "modules/image/ImageData.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($8|0),(441634|0),($vararg_buffer|0)); //@line 84 "modules/image/ImageData.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if (!($10)) {
  ___cxa_throw(($8|0),(96|0),(344|0)); //@line 84 "modules/image/ImageData.cpp"
  // unreachable; //@line 84 "modules/image/ImageData.cpp"
 }
 $11 = ___cxa_find_matching_catch_2()|0;
 $12 = tempRet0;
 ___cxa_free_exception(($8|0)); //@line 84 "modules/image/ImageData.cpp"
 $$0 = $11;$$01 = $12;
 ___resumeException($$0|0);
 // unreachable;
}
function __ZN4love5image9ImageData5pasteEPS1_iiiiii($this,$src,$dx,$dy,$sx,$sy,$sw,$sh) {
 $this = $this|0;
 $src = $src|0;
 $dx = $dx|0;
 $dy = $dy|0;
 $sx = $sx|0;
 $sy = $sy|0;
 $sw = $sw|0;
 $sh = $sh|0;
 var $$$$18 = 0, $$$18 = 0, $$01 = 0, $$011 = 0, $$03 = 0, $$04 = 0, $$05 = 0, $$1 = 0, $$112 = 0, $$114 = 0, $$12 = 0, $$16 = 0, $$18 = 0, $$2 = 0, $$3 = 0, $$dx = 0, $$pre = 0, $$pre20 = 0, $$sw = 0, $$sx = 0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $lock1 = 0, $lock2 = 0, $or$cond = 0, $or$cond17 = 0, $or$cond19 = 0, $or$cond21 = 0, $or$cond22 = 0, $or$cond23 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lock2 = sp + 4|0;
 $lock1 = sp;
 $0 = ((($src)) + 20|0); //@line 100 "modules/image/ImageData.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 100 "modules/image/ImageData.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($lock2,$1); //@line 100 "modules/image/ImageData.cpp"
 $2 = ((($this)) + 20|0); //@line 101 "modules/image/ImageData.cpp"
 __THREW__ = 0;
 $3 = (invoke_ii(3782,($2|0))|0); //@line 101 "modules/image/ImageData.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  __THREW__ = 0;
  invoke_vii(3783,($lock1|0),($3|0)); //@line 101 "modules/image/ImageData.cpp"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if (!($7)) {
   $8 = HEAP32[$src>>2]|0; //@line 103 "modules/image/ImageData.cpp"
   $9 = ((($8)) + 16|0); //@line 103 "modules/image/ImageData.cpp"
   $10 = HEAP32[$9>>2]|0; //@line 103 "modules/image/ImageData.cpp"
   __THREW__ = 0;
   $11 = (invoke_ii($10|0,($src|0))|0); //@line 103 "modules/image/ImageData.cpp"
   $12 = __THREW__; __THREW__ = 0;
   $13 = $12&1;
   if (!($13)) {
    $14 = HEAP32[$this>>2]|0; //@line 104 "modules/image/ImageData.cpp"
    $15 = ((($14)) + 16|0); //@line 104 "modules/image/ImageData.cpp"
    $16 = HEAP32[$15>>2]|0; //@line 104 "modules/image/ImageData.cpp"
    __THREW__ = 0;
    $17 = (invoke_ii($16|0,($this|0))|0); //@line 104 "modules/image/ImageData.cpp"
    $18 = __THREW__; __THREW__ = 0;
    $19 = $18&1;
    if (!($19)) {
     $20 = ((($src)) + 8|0); //@line 56 "modules/image/ImageData.cpp"
     $21 = HEAP32[$20>>2]|0; //@line 56 "modules/image/ImageData.cpp"
     $22 = ($21|0)<=($sx|0); //@line 107 "modules/image/ImageData.cpp"
     $23 = (($sw) + ($sx))|0; //@line 107 "modules/image/ImageData.cpp"
     $24 = ($23|0)<(0); //@line 107 "modules/image/ImageData.cpp"
     $or$cond = $24 | $22; //@line 107 "modules/image/ImageData.cpp"
     do {
      if (!($or$cond)) {
       $25 = ((($src)) + 12|0); //@line 61 "modules/image/ImageData.cpp"
       $26 = HEAP32[$25>>2]|0; //@line 61 "modules/image/ImageData.cpp"
       $27 = ($26|0)<=($sy|0); //@line 107 "modules/image/ImageData.cpp"
       $28 = (($sh) + ($sy))|0; //@line 107 "modules/image/ImageData.cpp"
       $29 = ($28|0)<(0); //@line 107 "modules/image/ImageData.cpp"
       $or$cond17 = $29 | $27; //@line 107 "modules/image/ImageData.cpp"
       if (!($or$cond17)) {
        $30 = ((($this)) + 8|0); //@line 56 "modules/image/ImageData.cpp"
        $31 = HEAP32[$30>>2]|0; //@line 56 "modules/image/ImageData.cpp"
        $32 = ($31|0)>($dx|0); //@line 108 "modules/image/ImageData.cpp"
        if ($32) {
         $33 = (($sw) + ($dx))|0; //@line 108 "modules/image/ImageData.cpp"
         $34 = ($33|0)<(0); //@line 108 "modules/image/ImageData.cpp"
         if (!($34)) {
          $35 = ((($this)) + 12|0); //@line 61 "modules/image/ImageData.cpp"
          $36 = HEAP32[$35>>2]|0; //@line 61 "modules/image/ImageData.cpp"
          $37 = ($36|0)>($dy|0); //@line 108 "modules/image/ImageData.cpp"
          if ($37) {
           $38 = (($sh) + ($dy))|0; //@line 108 "modules/image/ImageData.cpp"
           $39 = ($38|0)<(0); //@line 108 "modules/image/ImageData.cpp"
           if (!($39)) {
            $44 = ($dx|0)<(0); //@line 112 "modules/image/ImageData.cpp"
            $45 = $44 ? $dx : 0; //@line 112 "modules/image/ImageData.cpp"
            $$sx = (($sx) - ($45))|0; //@line 112 "modules/image/ImageData.cpp"
            $$sw = $44 ? $33 : $sw; //@line 112 "modules/image/ImageData.cpp"
            $$dx = $44 ? 0 : $dx; //@line 112 "modules/image/ImageData.cpp"
            $46 = ($dy|0)<(0); //@line 118 "modules/image/ImageData.cpp"
            $47 = $46 ? $dy : 0; //@line 118 "modules/image/ImageData.cpp"
            $$011 = (($sy) - ($47))|0; //@line 118 "modules/image/ImageData.cpp"
            $$05 = $46 ? $38 : $sh; //@line 118 "modules/image/ImageData.cpp"
            $$01 = $46 ? 0 : $dy; //@line 118 "modules/image/ImageData.cpp"
            $48 = ($$sx|0)<(0); //@line 124 "modules/image/ImageData.cpp"
            $$114 = $48 ? 0 : $$sx; //@line 124 "modules/image/ImageData.cpp"
            $49 = $48 ? $$sx : 0; //@line 124 "modules/image/ImageData.cpp"
            $$18 = (($$sw) + ($49))|0; //@line 124 "modules/image/ImageData.cpp"
            $$1 = (($$dx) - ($49))|0; //@line 124 "modules/image/ImageData.cpp"
            $50 = ($$011|0)<(0); //@line 130 "modules/image/ImageData.cpp"
            $$112 = $50 ? 0 : $$011; //@line 130 "modules/image/ImageData.cpp"
            $51 = $50 ? $$011 : 0; //@line 130 "modules/image/ImageData.cpp"
            $$16 = (($$05) + ($51))|0; //@line 130 "modules/image/ImageData.cpp"
            $$12 = (($$01) - ($51))|0; //@line 130 "modules/image/ImageData.cpp"
            $52 = (($sw) + ($dx))|0; //@line 137 "modules/image/ImageData.cpp"
            $53 = $44 ? $33 : $52; //@line 137 "modules/image/ImageData.cpp"
            $54 = ($53|0)>($31|0); //@line 137 "modules/image/ImageData.cpp"
            $55 = (($31) - ($$1))|0; //@line 138 "modules/image/ImageData.cpp"
            $$$18 = $54 ? $55 : $$18; //@line 137 "modules/image/ImageData.cpp"
            $56 = (($sh) + ($dy))|0; //@line 140 "modules/image/ImageData.cpp"
            $57 = $46 ? $38 : $56; //@line 140 "modules/image/ImageData.cpp"
            $58 = ($57|0)>($36|0); //@line 140 "modules/image/ImageData.cpp"
            $59 = (($36) - ($$12))|0; //@line 141 "modules/image/ImageData.cpp"
            $$2 = $58 ? $59 : $$16; //@line 140 "modules/image/ImageData.cpp"
            $60 = (($$$18) + ($$114))|0; //@line 143 "modules/image/ImageData.cpp"
            $61 = ($60|0)>($21|0); //@line 143 "modules/image/ImageData.cpp"
            $62 = (($21) - ($$114))|0; //@line 144 "modules/image/ImageData.cpp"
            $$$$18 = $61 ? $62 : $$$18; //@line 143 "modules/image/ImageData.cpp"
            $63 = (($$2) + ($$112))|0; //@line 146 "modules/image/ImageData.cpp"
            $64 = ($63|0)>($26|0); //@line 146 "modules/image/ImageData.cpp"
            $65 = (($26) - ($$112))|0; //@line 147 "modules/image/ImageData.cpp"
            $$3 = $64 ? $65 : $$2; //@line 146 "modules/image/ImageData.cpp"
            $66 = ($$$$18|0)==($31|0); //@line 150 "modules/image/ImageData.cpp"
            $67 = ($31|0)==($21|0); //@line 150 "modules/image/ImageData.cpp"
            $or$cond21 = $66 & $67; //@line 150 "modules/image/ImageData.cpp"
            $68 = ($$3|0)==($36|0); //@line 151 "modules/image/ImageData.cpp"
            $or$cond22 = $or$cond21 & $68; //@line 150 "modules/image/ImageData.cpp"
            $69 = ($36|0)==($26|0); //@line 151 "modules/image/ImageData.cpp"
            $or$cond23 = $or$cond22 & $69; //@line 150 "modules/image/ImageData.cpp"
            if ($or$cond23) {
             $70 = $21 << 2; //@line 153 "modules/image/ImageData.cpp"
             $71 = Math_imul($70, $26)|0; //@line 153 "modules/image/ImageData.cpp"
             _memcpy(($17|0),($11|0),($71|0))|0; //@line 153 "modules/image/ImageData.cpp"
             break;
            }
            $72 = ($$$$18|0)>(0); //@line 155 "modules/image/ImageData.cpp"
            $73 = ($$3|0)>(0); //@line 158 "modules/image/ImageData.cpp"
            $or$cond19 = $72 & $73; //@line 155 "modules/image/ImageData.cpp"
            if ($or$cond19) {
             $74 = (($17) + ($$1<<2)|0); //@line 159 "modules/image/ImageData.cpp"
             $75 = (($11) + ($$114<<2)|0); //@line 159 "modules/image/ImageData.cpp"
             $76 = $$$$18 << 2; //@line 159 "modules/image/ImageData.cpp"
             $77 = Math_imul($31, $$12)|0; //@line 159 "modules/image/ImageData.cpp"
             $78 = (($74) + ($77<<2)|0); //@line 159 "modules/image/ImageData.cpp"
             $79 = Math_imul($21, $$112)|0; //@line 159 "modules/image/ImageData.cpp"
             $80 = (($75) + ($79<<2)|0); //@line 159 "modules/image/ImageData.cpp"
             _memcpy(($78|0),($80|0),($76|0))|0; //@line 159 "modules/image/ImageData.cpp"
             $81 = ($$3|0)>(1); //@line 158 "modules/image/ImageData.cpp"
             if ($81) {
              $83 = 1;
              while(1) {
               $$pre = HEAP32[$30>>2]|0; //@line 56 "modules/image/ImageData.cpp"
               $$pre20 = HEAP32[$20>>2]|0; //@line 56 "modules/image/ImageData.cpp"
               $82 = (($83) + ($$12))|0; //@line 159 "modules/image/ImageData.cpp"
               $84 = Math_imul($$pre, $82)|0; //@line 159 "modules/image/ImageData.cpp"
               $85 = (($74) + ($84<<2)|0); //@line 159 "modules/image/ImageData.cpp"
               $86 = (($83) + ($$112))|0; //@line 159 "modules/image/ImageData.cpp"
               $87 = Math_imul($$pre20, $86)|0; //@line 159 "modules/image/ImageData.cpp"
               $88 = (($75) + ($87<<2)|0); //@line 159 "modules/image/ImageData.cpp"
               _memcpy(($85|0),($88|0),($76|0))|0; //@line 159 "modules/image/ImageData.cpp"
               $89 = (($83) + 1)|0; //@line 158 "modules/image/ImageData.cpp"
               $90 = ($89|0)<($$3|0); //@line 158 "modules/image/ImageData.cpp"
               if ($90) {
                $83 = $89;
               } else {
                break;
               }
              }
             }
            }
           }
          }
         }
        }
       }
      }
     } while(0);
     __ZN4love6thread4LockD2Ev($lock1); //@line 161 "modules/image/ImageData.cpp"
     __ZN4love6thread4LockD2Ev($lock2); //@line 161 "modules/image/ImageData.cpp"
     STACKTOP = sp;return; //@line 161 "modules/image/ImageData.cpp"
    }
   }
   $42 = ___cxa_find_matching_catch_2()|0;
   $43 = tempRet0;
   __ZN4love6thread4LockD2Ev($lock1); //@line 161 "modules/image/ImageData.cpp"
   $$03 = $43;$$04 = $42;
   __ZN4love6thread4LockD2Ev($lock2); //@line 161 "modules/image/ImageData.cpp"
   ___resumeException($$04|0);
   // unreachable;
  }
 }
 $40 = ___cxa_find_matching_catch_2()|0;
 $41 = tempRet0;
 $$03 = $41;$$04 = $40;
 __ZN4love6thread4LockD2Ev($lock2); //@line 161 "modules/image/ImageData.cpp"
 ___resumeException($$04|0);
 // unreachable;
}
function __ZNK4love5image9ImageData8getMutexEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 165 "modules/image/ImageData.cpp"
 $1 = (__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv($0)|0); //@line 165 "modules/image/ImageData.cpp"
 return ($1|0); //@line 165 "modules/image/ImageData.cpp"
}
function __ZN4love5image9ImageData11getConstantEPKcRNS1_13EncodedFormatE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love9StringMapINS_5image9ImageData13EncodedFormatELj2EE4findEPKcRS3_(635024,$in,$out)|0); //@line 170 "modules/image/ImageData.cpp"
 return ($0|0); //@line 170 "modules/image/ImageData.cpp"
}
function __ZN4love9StringMapINS_5image9ImageData13EncodedFormatELj2EE4findEPKcRS3_($this,$key,$t) {
 $this = $this|0;
 $key = $key|0;
 $t = $t|0;
 var $$$2 = 0, $$011$i$137 = 0, $$011$i$233 = 0, $$011$i$330 = 0, $$011$i41 = 0, $$0110$i$138 = 0, $$0110$i$234 = 0, $$0110$i$331 = 0, $$0110$i42 = 0, $$02$i = 0, $$13 = 0, $$2 = 0, $$lcssa12 = 0, $$pre14 = 0, $$pre16 = 0, $$pre18 = 0, $$pre20 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $or$cond$i = 0, $or$cond$i$1 = 0, $or$cond$i$136 = 0, $or$cond$i$2 = 0, $or$cond$i$232 = 0, $or$cond$i$3 = 0, $or$cond$i$329 = 0, $or$cond$i40 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = $hash$0$lcssa$i & 3; //@line 72 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 74 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 74 "./common/StringMap.h"
 if ($12) {
  $$13 = 1;$$2 = 0;
  $$$2 = $$13 & $$2;
  return ($$$2|0); //@line 85 "./common/StringMap.h"
 }
 $13 = (($this) + (($9*12)|0)|0); //@line 77 "./common/StringMap.h"
 $14 = HEAP32[$13>>2]|0; //@line 77 "./common/StringMap.h"
 $15 = HEAP8[$14>>0]|0; //@line 54 "./common/StringMap.h"
 $16 = ($15<<24>>24)==(0); //@line 54 "./common/StringMap.h"
 L8: do {
  if ($16) {
   $26 = $0;
   label = 8;
  } else {
   $17 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
   $18 = ($15<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
   $or$cond$i40 = $17 & $18; //@line 54 "./common/StringMap.h"
   if ($or$cond$i40) {
    $$011$i41 = $key;$$0110$i42 = $14;
    while(1) {
     $19 = ((($$0110$i42)) + 1|0); //@line 59 "./common/StringMap.h"
     $20 = ((($$011$i41)) + 1|0); //@line 60 "./common/StringMap.h"
     $21 = HEAP8[$19>>0]|0; //@line 54 "./common/StringMap.h"
     $22 = ($21<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     $$pre14 = HEAP8[$20>>0]|0; //@line 63 "./common/StringMap.h"
     if ($22) {
      $26 = $$pre14;
      label = 8;
      break L8;
     }
     $23 = ($$pre14<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $24 = ($21<<24>>24)==($$pre14<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i = $23 & $24; //@line 54 "./common/StringMap.h"
     if ($or$cond$i) {
      $$011$i41 = $20;$$0110$i42 = $19;
     } else {
      label = 10;
      break;
     }
    }
   } else {
    label = 10;
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $25 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
  if ($25) {
   $$lcssa12 = $9;
  } else {
   label = 10;
  }
 }
 do {
  if ((label|0) == 10) {
   $29 = (($hash$0$lcssa$i) + 1)|0; //@line 72 "./common/StringMap.h"
   $30 = $29 & 3; //@line 72 "./common/StringMap.h"
   $31 = (((($this) + (($30*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $32 = HEAP8[$31>>0]|0; //@line 74 "./common/StringMap.h"
   $33 = ($32<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($33) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $34 = (($this) + (($30*12)|0)|0); //@line 77 "./common/StringMap.h"
   $35 = HEAP32[$34>>2]|0; //@line 77 "./common/StringMap.h"
   $36 = HEAP8[$35>>0]|0; //@line 54 "./common/StringMap.h"
   $37 = ($36<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L20: do {
    if ($37) {
     $47 = $0;
     label = 16;
    } else {
     $38 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $39 = ($36<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$136 = $38 & $39; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$136) {
      $$011$i$137 = $key;$$0110$i$138 = $35;
      while(1) {
       $40 = ((($$0110$i$138)) + 1|0); //@line 59 "./common/StringMap.h"
       $41 = ((($$011$i$137)) + 1|0); //@line 60 "./common/StringMap.h"
       $42 = HEAP8[$40>>0]|0; //@line 54 "./common/StringMap.h"
       $43 = ($42<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre16 = HEAP8[$41>>0]|0; //@line 63 "./common/StringMap.h"
       if ($43) {
        $47 = $$pre16;
        label = 16;
        break L20;
       }
       $44 = ($$pre16<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $45 = ($42<<24>>24)==($$pre16<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$1 = $44 & $45; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$1) {
        $$011$i$137 = $41;$$0110$i$138 = $40;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 16) {
    $46 = ($47<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($46) {
     $$lcssa12 = $30;
     break;
    }
   }
   $48 = (($hash$0$lcssa$i) + 2)|0; //@line 72 "./common/StringMap.h"
   $49 = $48 & 3; //@line 72 "./common/StringMap.h"
   $50 = (((($this) + (($49*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $51 = HEAP8[$50>>0]|0; //@line 74 "./common/StringMap.h"
   $52 = ($51<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($52) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $53 = (($this) + (($49*12)|0)|0); //@line 77 "./common/StringMap.h"
   $54 = HEAP32[$53>>2]|0; //@line 77 "./common/StringMap.h"
   $55 = HEAP8[$54>>0]|0; //@line 54 "./common/StringMap.h"
   $56 = ($55<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L31: do {
    if ($56) {
     $66 = $0;
     label = 22;
    } else {
     $57 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $58 = ($55<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$232 = $57 & $58; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$232) {
      $$011$i$233 = $key;$$0110$i$234 = $54;
      while(1) {
       $59 = ((($$0110$i$234)) + 1|0); //@line 59 "./common/StringMap.h"
       $60 = ((($$011$i$233)) + 1|0); //@line 60 "./common/StringMap.h"
       $61 = HEAP8[$59>>0]|0; //@line 54 "./common/StringMap.h"
       $62 = ($61<<24>>24)==(0); //@line 54 "./common/StringMap.h"
       $$pre18 = HEAP8[$60>>0]|0; //@line 63 "./common/StringMap.h"
       if ($62) {
        $66 = $$pre18;
        label = 22;
        break L31;
       }
       $63 = ($$pre18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
       $64 = ($61<<24>>24)==($$pre18<<24>>24); //@line 56 "./common/StringMap.h"
       $or$cond$i$2 = $63 & $64; //@line 54 "./common/StringMap.h"
       if ($or$cond$i$2) {
        $$011$i$233 = $60;$$0110$i$234 = $59;
       } else {
        break;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 22) {
    $65 = ($66<<24>>24)==(0); //@line 63 "./common/StringMap.h"
    if ($65) {
     $$lcssa12 = $49;
     break;
    }
   }
   $67 = (($hash$0$lcssa$i) + 3)|0; //@line 72 "./common/StringMap.h"
   $68 = $67 & 3; //@line 72 "./common/StringMap.h"
   $69 = (((($this) + (($68*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
   $70 = HEAP8[$69>>0]|0; //@line 74 "./common/StringMap.h"
   $71 = ($70<<24>>24)==(0); //@line 74 "./common/StringMap.h"
   if ($71) {
    $$13 = 1;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
   $72 = (($this) + (($68*12)|0)|0); //@line 77 "./common/StringMap.h"
   $73 = HEAP32[$72>>2]|0; //@line 77 "./common/StringMap.h"
   $74 = HEAP8[$73>>0]|0; //@line 54 "./common/StringMap.h"
   $75 = ($74<<24>>24)==(0); //@line 54 "./common/StringMap.h"
   L42: do {
    if ($75) {
     $85 = $0;
    } else {
     $76 = ($0<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $77 = ($74<<24>>24)==($0<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$329 = $76 & $77; //@line 54 "./common/StringMap.h"
     if ($or$cond$i$329) {
      $$011$i$330 = $key;$$0110$i$331 = $73;
     } else {
      $$13 = 0;$$2 = 0;
      $$$2 = $$13 & $$2;
      return ($$$2|0); //@line 85 "./common/StringMap.h"
     }
     while(1) {
      $78 = ((($$0110$i$331)) + 1|0); //@line 59 "./common/StringMap.h"
      $79 = ((($$011$i$330)) + 1|0); //@line 60 "./common/StringMap.h"
      $80 = HEAP8[$78>>0]|0; //@line 54 "./common/StringMap.h"
      $81 = ($80<<24>>24)==(0); //@line 54 "./common/StringMap.h"
      $$pre20 = HEAP8[$79>>0]|0; //@line 63 "./common/StringMap.h"
      if ($81) {
       $85 = $$pre20;
       break L42;
      }
      $82 = ($$pre20<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
      $83 = ($80<<24>>24)==($$pre20<<24>>24); //@line 56 "./common/StringMap.h"
      $or$cond$i$3 = $82 & $83; //@line 54 "./common/StringMap.h"
      if ($or$cond$i$3) {
       $$011$i$330 = $79;$$0110$i$331 = $78;
      } else {
       $$13 = 0;$$2 = 0;
       break;
      }
     }
     $$$2 = $$13 & $$2;
     return ($$$2|0); //@line 85 "./common/StringMap.h"
    }
   } while(0);
   $84 = ($85<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($84) {
    $$lcssa12 = $68;
   } else {
    $$13 = 0;$$2 = 0;
    $$$2 = $$13 & $$2;
    return ($$$2|0); //@line 85 "./common/StringMap.h"
   }
  }
 } while(0);
 $27 = (((($this) + (($$lcssa12*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
 $28 = HEAP32[$27>>2]|0; //@line 79 "./common/StringMap.h"
 HEAP32[$t>>2] = $28; //@line 79 "./common/StringMap.h"
 $$13 = 1;$$2 = 1;
 $$$2 = $$13 & $$2;
 return ($$$2|0); //@line 85 "./common/StringMap.h"
}
function __ZN4love5image9ImageData11getConstantENS1_13EncodedFormatERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(1); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (635072 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 175 "modules/image/ImageData.cpp"
}
function __ZN4love8joystick3sdl8JoystickD2Ev($this) {
 $this = $this|0;
 var $$sroa$4$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$sroa$4$i = sp;
 HEAP32[$this>>2] = (26128); //@line 63 "modules/joystick/sdl/Joystick.cpp"
 $0 = ((($this)) + 16|0); //@line 102 "modules/joystick/sdl/Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 102 "modules/joystick/sdl/Joystick.cpp"
 $2 = ($1|0)==(0|0); //@line 102 "modules/joystick/sdl/Joystick.cpp"
 if ($2) {
  label = 3;
 } else {
  __THREW__ = 0;
  invoke_vi(4182,($1|0)); //@line 103 "modules/joystick/sdl/Joystick.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if (!($4)) {
   label = 3;
  }
 }
 do {
  if ((label|0) == 3) {
   $5 = ((($this)) + 12|0); //@line 105 "modules/joystick/sdl/Joystick.cpp"
   $6 = HEAP32[$5>>2]|0; //@line 105 "modules/joystick/sdl/Joystick.cpp"
   $7 = ($6|0)==(0|0); //@line 105 "modules/joystick/sdl/Joystick.cpp"
   if (!($7)) {
    __THREW__ = 0;
    invoke_vi(4183,($6|0)); //@line 106 "modules/joystick/sdl/Joystick.cpp"
    $8 = __THREW__; __THREW__ = 0;
    $9 = $8&1;
    if ($9) {
     break;
    }
   }
   $10 = ((($this)) + 8|0); //@line 108 "modules/joystick/sdl/Joystick.cpp"
   $11 = HEAP32[$10>>2]|0; //@line 108 "modules/joystick/sdl/Joystick.cpp"
   $12 = ($11|0)==(0|0); //@line 108 "modules/joystick/sdl/Joystick.cpp"
   if (!($12)) {
    __THREW__ = 0;
    invoke_vi(4184,($11|0)); //@line 109 "modules/joystick/sdl/Joystick.cpp"
    $13 = __THREW__; __THREW__ = 0;
    $14 = $13&1;
    if ($14) {
     break;
    }
   }
   HEAP32[$10>>2] = 0; //@line 111 "modules/joystick/sdl/Joystick.cpp"
   HEAP32[$5>>2] = 0; //@line 112 "modules/joystick/sdl/Joystick.cpp"
   HEAP32[$0>>2] = 0; //@line 113 "modules/joystick/sdl/Joystick.cpp"
   $15 = ((($this)) + 20|0); //@line 114 "modules/joystick/sdl/Joystick.cpp"
   HEAP32[$15>>2] = -1; //@line 114 "modules/joystick/sdl/Joystick.cpp"
   $16 = ((($this)) + 52|0); //@line 115 "modules/joystick/sdl/Joystick.cpp"
   HEAPF32[$16>>2] = 0.0; //@line 115 "modules/joystick/sdl/Joystick.cpp"
   $17 = ((($this)) + 56|0); //@line 115 "modules/joystick/sdl/Joystick.cpp"
   HEAPF32[$17>>2] = 0.0; //@line 115 "modules/joystick/sdl/Joystick.cpp"
   $18 = ((($this)) + 60|0); //@line 115 "modules/joystick/sdl/Joystick.cpp"
   HEAP16[$18>>1] = 0; //@line 115 "modules/joystick/sdl/Joystick.cpp"
   $19 = ((($this)) + 62|0); //@line 115 "modules/joystick/sdl/Joystick.cpp"
   dest=$19; src=$$sroa$4$i; stop=dest+66|0; do { HEAP16[dest>>1]=HEAP16[src>>1]|0; dest=dest+2|0; src=src+2|0; } while ((dest|0) < (stop|0)); //@line 115 "modules/joystick/sdl/Joystick.cpp"
   $20 = ((($this)) + 128|0); //@line 115 "modules/joystick/sdl/Joystick.cpp"
   $21 = $20; //@line 115 "modules/joystick/sdl/Joystick.cpp"
   $22 = $21; //@line 115 "modules/joystick/sdl/Joystick.cpp"
   HEAP32[$22>>2] = 0; //@line 115 "modules/joystick/sdl/Joystick.cpp"
   $23 = (($21) + 4)|0; //@line 115 "modules/joystick/sdl/Joystick.cpp"
   $24 = $23; //@line 115 "modules/joystick/sdl/Joystick.cpp"
   HEAP32[$24>>2] = 0; //@line 115 "modules/joystick/sdl/Joystick.cpp"
   $25 = ((($this)) + 136|0); //@line 115 "modules/joystick/sdl/Joystick.cpp"
   HEAP32[$25>>2] = -1; //@line 115 "modules/joystick/sdl/Joystick.cpp"
   $26 = ((($this)) + 140|0); //@line 115 "modules/joystick/sdl/Joystick.cpp"
   HEAP32[$26>>2] = -1; //@line 115 "modules/joystick/sdl/Joystick.cpp"
   $27 = ((($this)) + 40|0); //@line 65 "modules/joystick/sdl/Joystick.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($27); //@line 65 "modules/joystick/sdl/Joystick.cpp"
   $28 = ((($this)) + 24|0); //@line 65 "modules/joystick/sdl/Joystick.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($28); //@line 65 "modules/joystick/sdl/Joystick.cpp"
   STACKTOP = sp;return; //@line 65 "modules/joystick/sdl/Joystick.cpp"
  }
 } while(0);
 $29 = ___cxa_find_matching_catch_3(0|0)|0;
 $30 = tempRet0;
 $31 = ((($this)) + 40|0); //@line 65 "modules/joystick/sdl/Joystick.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($31); //@line 65 "modules/joystick/sdl/Joystick.cpp"
 $32 = ((($this)) + 24|0); //@line 65 "modules/joystick/sdl/Joystick.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($32); //@line 65 "modules/joystick/sdl/Joystick.cpp"
 ___clang_call_terminate($29); //@line 65 "modules/joystick/sdl/Joystick.cpp"
 // unreachable; //@line 65 "modules/joystick/sdl/Joystick.cpp"
}
function __ZN4love8joystick3sdl8JoystickD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8joystick3sdl8JoystickD2Ev($this); //@line 63 "modules/joystick/sdl/Joystick.cpp"
 __ZdlPv($this); //@line 63 "modules/joystick/sdl/Joystick.cpp"
 return; //@line 65 "modules/joystick/sdl/Joystick.cpp"
}
function __ZN4love8joystick3sdl8Joystick4openEi($this,$deviceindex) {
 $this = $this|0;
 $deviceindex = $deviceindex|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cstr = 0;
 var $joyname$02 = 0, $sdlguid = 0, $sdlguid$byval_copy = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $sdlguid$byval_copy = sp + 72|0;
 $cstr = sp + 32|0;
 $sdlguid = sp + 16|0;
 $0 = sp;
 $1 = HEAP32[$this>>2]|0; //@line 69 "modules/joystick/sdl/Joystick.cpp"
 $2 = ((($1)) + 20|0); //@line 69 "modules/joystick/sdl/Joystick.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 69 "modules/joystick/sdl/Joystick.cpp"
 FUNCTION_TABLE_vi[$3 & 8191]($this); //@line 69 "modules/joystick/sdl/Joystick.cpp"
 $4 = (_SDL_JoystickOpen($deviceindex)|0); //@line 71 "modules/joystick/sdl/Joystick.cpp"
 $5 = ((($this)) + 8|0); //@line 71 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$5>>2] = $4; //@line 71 "modules/joystick/sdl/Joystick.cpp"
 $6 = ($4|0)==(0|0); //@line 73 "modules/joystick/sdl/Joystick.cpp"
 if ($6) {
  $35 = HEAP32[$this>>2]|0; //@line 97 "modules/joystick/sdl/Joystick.cpp"
  $36 = ((($35)) + 24|0); //@line 97 "modules/joystick/sdl/Joystick.cpp"
  $37 = HEAP32[$36>>2]|0; //@line 97 "modules/joystick/sdl/Joystick.cpp"
  $38 = (FUNCTION_TABLE_ii[$37 & 8191]($this)|0); //@line 97 "modules/joystick/sdl/Joystick.cpp"
  STACKTOP = sp;return ($38|0); //@line 97 "modules/joystick/sdl/Joystick.cpp"
 }
 $7 = (_SDL_JoystickInstanceID($4)|0); //@line 75 "modules/joystick/sdl/Joystick.cpp"
 $8 = ((($this)) + 20|0); //@line 75 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$8>>2] = $7; //@line 75 "modules/joystick/sdl/Joystick.cpp"
 $9 = HEAP32[$5>>2]|0; //@line 80 "modules/joystick/sdl/Joystick.cpp"
 _SDL_JoystickGetGUID($sdlguid,$9); //@line 80 "modules/joystick/sdl/Joystick.cpp"
 dest=$sdlguid$byval_copy; src=$sdlguid; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 81 "modules/joystick/sdl/Joystick.cpp"
 _SDL_JoystickGetGUIDString($sdlguid$byval_copy,$cstr,33); //@line 81 "modules/joystick/sdl/Joystick.cpp"
 $10 = ((($this)) + 24|0); //@line 83 "modules/joystick/sdl/Joystick.cpp"
 $11 = (_strlen($cstr)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($0,$cstr,$11); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = HEAP8[$10>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = $12 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = ($13<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($14) {
  $18 = ((($10)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$18>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$10>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 } else {
  $15 = ((($this)) + 32|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $16 = HEAP32[$15>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$16>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $17 = ((($this)) + 28|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$17>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 __THREW__ = 0;
 invoke_vii(3368,($10|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $21 = ___cxa_find_matching_catch_3(0|0)|0;
  $22 = tempRet0;
  ___clang_call_terminate($21); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 ;HEAP32[$10>>2]=HEAP32[$0>>2]|0;HEAP32[$10+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$10+8>>2]=HEAP32[$0+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 83 "modules/joystick/sdl/Joystick.cpp"
 $23 = HEAP32[$this>>2]|0; //@line 86 "modules/joystick/sdl/Joystick.cpp"
 $24 = ((($23)) + 60|0); //@line 86 "modules/joystick/sdl/Joystick.cpp"
 $25 = HEAP32[$24>>2]|0; //@line 86 "modules/joystick/sdl/Joystick.cpp"
 (FUNCTION_TABLE_iii[$25 & 8191]($this,$deviceindex)|0); //@line 86 "modules/joystick/sdl/Joystick.cpp"
 $26 = HEAP32[$5>>2]|0; //@line 89 "modules/joystick/sdl/Joystick.cpp"
 $27 = (_SDL_JoystickName($26)|0); //@line 89 "modules/joystick/sdl/Joystick.cpp"
 $28 = ($27|0)==(0|0); //@line 90 "modules/joystick/sdl/Joystick.cpp"
 if ($28) {
  $29 = ((($this)) + 12|0); //@line 90 "modules/joystick/sdl/Joystick.cpp"
  $30 = HEAP32[$29>>2]|0; //@line 90 "modules/joystick/sdl/Joystick.cpp"
  $31 = ($30|0)==(0|0); //@line 90 "modules/joystick/sdl/Joystick.cpp"
  if (!($31)) {
   $32 = (_SDL_GameControllerName($30)|0); //@line 91 "modules/joystick/sdl/Joystick.cpp"
   $33 = ($32|0)==(0|0); //@line 93 "modules/joystick/sdl/Joystick.cpp"
   if (!($33)) {
    $joyname$02 = $32;
    label = 10;
   }
  }
 } else {
  $joyname$02 = $27;
  label = 10;
 }
 if ((label|0) == 10) {
  $34 = ((($this)) + 40|0); //@line 94 "modules/joystick/sdl/Joystick.cpp"
  (__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc($34,$joyname$02)|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 $35 = HEAP32[$this>>2]|0; //@line 97 "modules/joystick/sdl/Joystick.cpp"
 $36 = ((($35)) + 24|0); //@line 97 "modules/joystick/sdl/Joystick.cpp"
 $37 = HEAP32[$36>>2]|0; //@line 97 "modules/joystick/sdl/Joystick.cpp"
 $38 = (FUNCTION_TABLE_ii[$37 & 8191]($this)|0); //@line 97 "modules/joystick/sdl/Joystick.cpp"
 STACKTOP = sp;return ($38|0); //@line 97 "modules/joystick/sdl/Joystick.cpp"
}
function __ZN4love8joystick3sdl8Joystick5closeEv($this) {
 $this = $this|0;
 var $$sroa$4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$sroa$4 = sp;
 $0 = ((($this)) + 16|0); //@line 102 "modules/joystick/sdl/Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 102 "modules/joystick/sdl/Joystick.cpp"
 $2 = ($1|0)==(0|0); //@line 102 "modules/joystick/sdl/Joystick.cpp"
 if (!($2)) {
  _SDL_HapticClose($1); //@line 103 "modules/joystick/sdl/Joystick.cpp"
 }
 $3 = ((($this)) + 12|0); //@line 105 "modules/joystick/sdl/Joystick.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 105 "modules/joystick/sdl/Joystick.cpp"
 $5 = ($4|0)==(0|0); //@line 105 "modules/joystick/sdl/Joystick.cpp"
 if (!($5)) {
  _SDL_GameControllerClose($4); //@line 106 "modules/joystick/sdl/Joystick.cpp"
 }
 $6 = ((($this)) + 8|0); //@line 108 "modules/joystick/sdl/Joystick.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 108 "modules/joystick/sdl/Joystick.cpp"
 $8 = ($7|0)==(0|0); //@line 108 "modules/joystick/sdl/Joystick.cpp"
 if (!($8)) {
  _SDL_JoystickClose($7); //@line 109 "modules/joystick/sdl/Joystick.cpp"
 }
 HEAP32[$6>>2] = 0; //@line 111 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$3>>2] = 0; //@line 112 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$0>>2] = 0; //@line 113 "modules/joystick/sdl/Joystick.cpp"
 $9 = ((($this)) + 20|0); //@line 114 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$9>>2] = -1; //@line 114 "modules/joystick/sdl/Joystick.cpp"
 $10 = ((($this)) + 52|0); //@line 115 "modules/joystick/sdl/Joystick.cpp"
 HEAPF32[$10>>2] = 0.0; //@line 115 "modules/joystick/sdl/Joystick.cpp"
 $11 = ((($this)) + 56|0); //@line 115 "modules/joystick/sdl/Joystick.cpp"
 HEAPF32[$11>>2] = 0.0; //@line 115 "modules/joystick/sdl/Joystick.cpp"
 $12 = ((($this)) + 60|0); //@line 115 "modules/joystick/sdl/Joystick.cpp"
 HEAP16[$12>>1] = 0; //@line 115 "modules/joystick/sdl/Joystick.cpp"
 $13 = ((($this)) + 62|0); //@line 115 "modules/joystick/sdl/Joystick.cpp"
 dest=$13; src=$$sroa$4; stop=dest+66|0; do { HEAP16[dest>>1]=HEAP16[src>>1]|0; dest=dest+2|0; src=src+2|0; } while ((dest|0) < (stop|0)); //@line 115 "modules/joystick/sdl/Joystick.cpp"
 $14 = ((($this)) + 128|0); //@line 115 "modules/joystick/sdl/Joystick.cpp"
 $15 = $14; //@line 115 "modules/joystick/sdl/Joystick.cpp"
 $16 = $15; //@line 115 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$16>>2] = 0; //@line 115 "modules/joystick/sdl/Joystick.cpp"
 $17 = (($15) + 4)|0; //@line 115 "modules/joystick/sdl/Joystick.cpp"
 $18 = $17; //@line 115 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$18>>2] = 0; //@line 115 "modules/joystick/sdl/Joystick.cpp"
 $19 = ((($this)) + 136|0); //@line 115 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$19>>2] = -1; //@line 115 "modules/joystick/sdl/Joystick.cpp"
 $20 = ((($this)) + 140|0); //@line 115 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$20>>2] = -1; //@line 115 "modules/joystick/sdl/Joystick.cpp"
 STACKTOP = sp;return; //@line 116 "modules/joystick/sdl/Joystick.cpp"
}
function __ZNK4love8joystick3sdl8Joystick11isConnectedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 120 "modules/joystick/sdl/Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 120 "modules/joystick/sdl/Joystick.cpp"
 $2 = ($1|0)==(0|0); //@line 120 "modules/joystick/sdl/Joystick.cpp"
 if ($2) {
  $5 = 0;
  return ($5|0); //@line 120 "modules/joystick/sdl/Joystick.cpp"
 }
 $3 = (_SDL_JoystickGetAttached($1)|0); //@line 120 "modules/joystick/sdl/Joystick.cpp"
 $4 = ($3|0)!=(0); //@line 120 "modules/joystick/sdl/Joystick.cpp"
 $5 = $4;
 return ($5|0); //@line 120 "modules/joystick/sdl/Joystick.cpp"
}
function __ZNK4love8joystick3sdl8Joystick7getNameEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0); //@line 125 "modules/joystick/sdl/Joystick.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = $1 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ($2<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($3) {
  $6 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = $6;
  return ($7|0); //@line 125 "modules/joystick/sdl/Joystick.cpp"
 } else {
  $4 = ((($this)) + 48|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $5 = HEAP32[$4>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = $5;
  return ($7|0); //@line 125 "modules/joystick/sdl/Joystick.cpp"
 }
 return (0)|0;
}
function __ZNK4love8joystick3sdl8Joystick12getAxisCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 130 "modules/joystick/sdl/Joystick.cpp"
 $1 = ((($0)) + 24|0); //@line 130 "modules/joystick/sdl/Joystick.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 130 "modules/joystick/sdl/Joystick.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 130 "modules/joystick/sdl/Joystick.cpp"
 if (!($3)) {
  $7 = 0;
  return ($7|0); //@line 130 "modules/joystick/sdl/Joystick.cpp"
 }
 $4 = ((($this)) + 8|0); //@line 130 "modules/joystick/sdl/Joystick.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 130 "modules/joystick/sdl/Joystick.cpp"
 $6 = (_SDL_JoystickNumAxes($5)|0); //@line 130 "modules/joystick/sdl/Joystick.cpp"
 $7 = $6;
 return ($7|0); //@line 130 "modules/joystick/sdl/Joystick.cpp"
}
function __ZNK4love8joystick3sdl8Joystick14getButtonCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 135 "modules/joystick/sdl/Joystick.cpp"
 $1 = ((($0)) + 24|0); //@line 135 "modules/joystick/sdl/Joystick.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 135 "modules/joystick/sdl/Joystick.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 135 "modules/joystick/sdl/Joystick.cpp"
 if (!($3)) {
  $7 = 0;
  return ($7|0); //@line 135 "modules/joystick/sdl/Joystick.cpp"
 }
 $4 = ((($this)) + 8|0); //@line 135 "modules/joystick/sdl/Joystick.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 135 "modules/joystick/sdl/Joystick.cpp"
 $6 = (_SDL_JoystickNumButtons($5)|0); //@line 135 "modules/joystick/sdl/Joystick.cpp"
 $7 = $6;
 return ($7|0); //@line 135 "modules/joystick/sdl/Joystick.cpp"
}
function __ZNK4love8joystick3sdl8Joystick11getHatCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 140 "modules/joystick/sdl/Joystick.cpp"
 $1 = ((($0)) + 24|0); //@line 140 "modules/joystick/sdl/Joystick.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 140 "modules/joystick/sdl/Joystick.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 140 "modules/joystick/sdl/Joystick.cpp"
 if (!($3)) {
  $7 = 0;
  return ($7|0); //@line 140 "modules/joystick/sdl/Joystick.cpp"
 }
 $4 = ((($this)) + 8|0); //@line 140 "modules/joystick/sdl/Joystick.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 140 "modules/joystick/sdl/Joystick.cpp"
 $6 = (_SDL_JoystickNumHats($5)|0); //@line 140 "modules/joystick/sdl/Joystick.cpp"
 $7 = $6;
 return ($7|0); //@line 140 "modules/joystick/sdl/Joystick.cpp"
}
function __ZNK4love8joystick3sdl8Joystick7getAxisEi($this,$axisindex) {
 $this = $this|0;
 $axisindex = $axisindex|0;
 var $$0 = 0.0, $$not = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 145 "modules/joystick/sdl/Joystick.cpp"
 $1 = ((($0)) + 24|0); //@line 145 "modules/joystick/sdl/Joystick.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 145 "modules/joystick/sdl/Joystick.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 145 "modules/joystick/sdl/Joystick.cpp"
 $$not = $3 ^ 1; //@line 145 "modules/joystick/sdl/Joystick.cpp"
 $4 = ($axisindex|0)<(0); //@line 145 "modules/joystick/sdl/Joystick.cpp"
 $or$cond = $4 | $$not; //@line 145 "modules/joystick/sdl/Joystick.cpp"
 if ($or$cond) {
  $$0 = 0.0;
  return (+$$0); //@line 149 "modules/joystick/sdl/Joystick.cpp"
 }
 $5 = HEAP32[$this>>2]|0; //@line 145 "modules/joystick/sdl/Joystick.cpp"
 $6 = ((($5)) + 32|0); //@line 145 "modules/joystick/sdl/Joystick.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 145 "modules/joystick/sdl/Joystick.cpp"
 $8 = (FUNCTION_TABLE_ii[$7 & 8191]($this)|0); //@line 145 "modules/joystick/sdl/Joystick.cpp"
 $9 = ($8|0)>($axisindex|0); //@line 145 "modules/joystick/sdl/Joystick.cpp"
 if (!($9)) {
  $$0 = 0.0;
  return (+$$0); //@line 149 "modules/joystick/sdl/Joystick.cpp"
 }
 $10 = ((($this)) + 8|0); //@line 148 "modules/joystick/sdl/Joystick.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 148 "modules/joystick/sdl/Joystick.cpp"
 $12 = (_SDL_JoystickGetAxis($11,$axisindex)|0); //@line 148 "modules/joystick/sdl/Joystick.cpp"
 $13 = (+($12<<16>>16)); //@line 148 "modules/joystick/sdl/Joystick.cpp"
 $14 = $13 * 3.0517578125E-5; //@line 148 "modules/joystick/sdl/Joystick.cpp"
 $15 = (+__ZN4love8joystick8Joystick8clampvalEf($14)); //@line 148 "modules/joystick/sdl/Joystick.cpp"
 $$0 = $15;
 return (+$$0); //@line 149 "modules/joystick/sdl/Joystick.cpp"
}
function __ZNK4love8joystick3sdl8Joystick7getAxesEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$not = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$04 = 0, $or$cond = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 HEAP32[$agg$result>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($agg$result)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($agg$result)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$this>>2]|0; //@line 154 "modules/joystick/sdl/Joystick.cpp"
 $4 = ((($3)) + 32|0); //@line 154 "modules/joystick/sdl/Joystick.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 154 "modules/joystick/sdl/Joystick.cpp"
 $6 = (FUNCTION_TABLE_ii[$5 & 8191]($this)|0); //@line 154 "modules/joystick/sdl/Joystick.cpp"
 $7 = HEAP32[$this>>2]|0; //@line 156 "modules/joystick/sdl/Joystick.cpp"
 $8 = ((($7)) + 24|0); //@line 156 "modules/joystick/sdl/Joystick.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 156 "modules/joystick/sdl/Joystick.cpp"
 $10 = (FUNCTION_TABLE_ii[$9 & 8191]($this)|0); //@line 156 "modules/joystick/sdl/Joystick.cpp"
 $$not = $10 ^ 1; //@line 156 "modules/joystick/sdl/Joystick.cpp"
 $11 = ($6|0)<(1); //@line 156 "modules/joystick/sdl/Joystick.cpp"
 $or$cond = $11 | $$not; //@line 156 "modules/joystick/sdl/Joystick.cpp"
 if ($or$cond) {
  STACKTOP = sp;return; //@line 165 "modules/joystick/sdl/Joystick.cpp"
 }
 $22 = ((($agg$result)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $23 = ($6|0)==(0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($23) {
  STACKTOP = sp;return; //@line 165 "modules/joystick/sdl/Joystick.cpp"
 }
 $24 = $6 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $25 = (__Znwj($24)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $26 = $25; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($6<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 HEAP32[$agg$result>>2] = $26; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$1>>2] = $26; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$22>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $29 = ($6|0)>(0); //@line 161 "modules/joystick/sdl/Joystick.cpp"
 if (!($29)) {
  STACKTOP = sp;return; //@line 165 "modules/joystick/sdl/Joystick.cpp"
 }
 $30 = ((($this)) + 8|0); //@line 162 "modules/joystick/sdl/Joystick.cpp"
 $i$04 = 0;
 while(1) {
  $31 = HEAP32[$30>>2]|0; //@line 162 "modules/joystick/sdl/Joystick.cpp"
  __THREW__ = 0;
  $32 = (invoke_iii(4185,($31|0),($i$04|0))|0); //@line 162 "modules/joystick/sdl/Joystick.cpp"
  $33 = __THREW__; __THREW__ = 0;
  $34 = $33&1;
  if ($34) {
   break;
  }
  $35 = (+($32<<16>>16)); //@line 162 "modules/joystick/sdl/Joystick.cpp"
  $36 = $35 * 3.0517578125E-5; //@line 162 "modules/joystick/sdl/Joystick.cpp"
  __THREW__ = 0;
  $37 = (+invoke_dd(3950,(+$36))); //@line 162 "modules/joystick/sdl/Joystick.cpp"
  $38 = __THREW__; __THREW__ = 0;
  $39 = $38&1;
  if ($39) {
   break;
  }
  HEAPF32[$0>>2] = $37; //@line 162 "modules/joystick/sdl/Joystick.cpp"
  $40 = HEAP32[$1>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $41 = HEAP32[$22>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $42 = ($40>>>0)<($41>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($42) {
   HEAPF32[$40>>2] = $37; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $43 = ((($40)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$1>>2] = $43; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  } else {
   __THREW__ = 0;
   invoke_vii(3625,($agg$result|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $44 = __THREW__; __THREW__ = 0;
   $45 = $44&1;
   if ($45) {
    break;
   }
  }
  $46 = (($i$04) + 1)|0; //@line 161 "modules/joystick/sdl/Joystick.cpp"
  $47 = ($46|0)<($6|0); //@line 161 "modules/joystick/sdl/Joystick.cpp"
  if ($47) {
   $i$04 = $46;
  } else {
   label = 16;
   break;
  }
 }
 if ((label|0) == 16) {
  STACKTOP = sp;return; //@line 165 "modules/joystick/sdl/Joystick.cpp"
 }
 $12 = ___cxa_find_matching_catch_2()|0;
 $13 = tempRet0;
 $$pre = HEAP32[$agg$result>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = ($$pre|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = $$pre; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($14) {
  ___resumeException($12|0);
  // unreachable;
 }
 $16 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ($16|0)==($$pre|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($17)) {
  $scevgep$i$i$i = ((($16)) + -4|0);
  $18 = $scevgep$i$i$i;
  $19 = (($18) - ($15))|0;
  $20 = $19 >>> 2;
  $21 = $20 ^ -1;
  $scevgep4$i$i$i = (($16) + ($21<<2)|0);
  HEAP32[$1>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($$pre); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($12|0);
 // unreachable;
}
function __ZNK4love8joystick3sdl8Joystick6getHatEi($this,$hatindex) {
 $this = $this|0;
 $hatindex = $hatindex|0;
 var $$not = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $h$1 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 171 "modules/joystick/sdl/Joystick.cpp"
 $1 = ((($0)) + 24|0); //@line 171 "modules/joystick/sdl/Joystick.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 171 "modules/joystick/sdl/Joystick.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 171 "modules/joystick/sdl/Joystick.cpp"
 $$not = $3 ^ 1; //@line 171 "modules/joystick/sdl/Joystick.cpp"
 $4 = ($hatindex|0)<(0); //@line 171 "modules/joystick/sdl/Joystick.cpp"
 $or$cond = $4 | $$not; //@line 171 "modules/joystick/sdl/Joystick.cpp"
 if ($or$cond) {
  $h$1 = 0;
  return ($h$1|0); //@line 177 "modules/joystick/sdl/Joystick.cpp"
 }
 $5 = HEAP32[$this>>2]|0; //@line 171 "modules/joystick/sdl/Joystick.cpp"
 $6 = ((($5)) + 40|0); //@line 171 "modules/joystick/sdl/Joystick.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 171 "modules/joystick/sdl/Joystick.cpp"
 $8 = (FUNCTION_TABLE_ii[$7 & 8191]($this)|0); //@line 171 "modules/joystick/sdl/Joystick.cpp"
 $9 = ($8|0)>($hatindex|0); //@line 171 "modules/joystick/sdl/Joystick.cpp"
 if (!($9)) {
  $h$1 = 0;
  return ($h$1|0); //@line 177 "modules/joystick/sdl/Joystick.cpp"
 }
 $10 = ((($this)) + 8|0); //@line 174 "modules/joystick/sdl/Joystick.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 174 "modules/joystick/sdl/Joystick.cpp"
 $12 = (_SDL_JoystickGetHat($11,$hatindex)|0); //@line 174 "modules/joystick/sdl/Joystick.cpp"
 $13 = ($12&255)<(16); //@line 75 "./common/EnumMap.h"
 if (!($13)) {
  $h$1 = 0;
  return ($h$1|0); //@line 177 "modules/joystick/sdl/Joystick.cpp"
 }
 $14 = $12&255; //@line 75 "./common/EnumMap.h"
 $15 = (((635080 + ($14<<3)|0)) + 4|0); //@line 75 "./common/EnumMap.h"
 $16 = HEAP8[$15>>0]|0; //@line 75 "./common/EnumMap.h"
 $17 = ($16<<24>>24)==(0); //@line 75 "./common/EnumMap.h"
 if ($17) {
  $h$1 = 0;
  return ($h$1|0); //@line 177 "modules/joystick/sdl/Joystick.cpp"
 }
 $18 = (635080 + ($14<<3)|0); //@line 77 "./common/EnumMap.h"
 $19 = HEAP32[$18>>2]|0; //@line 77 "./common/EnumMap.h"
 $h$1 = $19;
 return ($h$1|0); //@line 177 "modules/joystick/sdl/Joystick.cpp"
}
function __ZNK4love8joystick3sdl8Joystick6isDownERKNSt3__16vectorIiNS3_9allocatorIiEEEE($this,$buttonlist) {
 $this = $this|0;
 $buttonlist = $buttonlist|0;
 var $$4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $__begin$sroa$0$06 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 181 "modules/joystick/sdl/Joystick.cpp"
 $1 = ((($0)) + 24|0); //@line 181 "modules/joystick/sdl/Joystick.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 181 "modules/joystick/sdl/Joystick.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 181 "modules/joystick/sdl/Joystick.cpp"
 if (!($3)) {
  $$4 = 0;
  return ($$4|0); //@line 196 "modules/joystick/sdl/Joystick.cpp"
 }
 $4 = HEAP32[$this>>2]|0; //@line 184 "modules/joystick/sdl/Joystick.cpp"
 $5 = ((($4)) + 36|0); //@line 184 "modules/joystick/sdl/Joystick.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 184 "modules/joystick/sdl/Joystick.cpp"
 $7 = (FUNCTION_TABLE_ii[$6 & 8191]($this)|0); //@line 184 "modules/joystick/sdl/Joystick.cpp"
 $8 = HEAP32[$buttonlist>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ((($buttonlist)) + 4|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = HEAP32[$9>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ((($this)) + 8|0); //@line 191 "modules/joystick/sdl/Joystick.cpp"
 $12 = ($8|0)==($10|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($12) {
  $$4 = 0;
  return ($$4|0); //@line 196 "modules/joystick/sdl/Joystick.cpp"
 } else {
  $__begin$sroa$0$06 = $8;
 }
 while(1) {
  $13 = HEAP32[$__begin$sroa$0$06>>2]|0; //@line 186 "modules/joystick/sdl/Joystick.cpp"
  $14 = ($13|0)>(-1); //@line 188 "modules/joystick/sdl/Joystick.cpp"
  $15 = ($13|0)<($7|0); //@line 188 "modules/joystick/sdl/Joystick.cpp"
  $or$cond = $14 & $15; //@line 188 "modules/joystick/sdl/Joystick.cpp"
  if ($or$cond) {
   $16 = HEAP32[$11>>2]|0; //@line 191 "modules/joystick/sdl/Joystick.cpp"
   $17 = (_SDL_JoystickGetButton($16,$13)|0); //@line 191 "modules/joystick/sdl/Joystick.cpp"
   $18 = ($17<<24>>24)==(1); //@line 191 "modules/joystick/sdl/Joystick.cpp"
   if ($18) {
    $$4 = 1;
    label = 6;
    break;
   }
  }
  $19 = ((($__begin$sroa$0$06)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $20 = ($19|0)==($10|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($20) {
   $$4 = 0;
   label = 6;
   break;
  } else {
   $__begin$sroa$0$06 = $19;
  }
 }
 if ((label|0) == 6) {
  return ($$4|0); //@line 196 "modules/joystick/sdl/Joystick.cpp"
 }
 return (0)|0;
}
function __ZN4love8joystick3sdl8Joystick11openGamepadEi($this,$deviceindex) {
 $this = $this|0;
 $deviceindex = $deviceindex|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_IsGameController($deviceindex)|0); //@line 200 "modules/joystick/sdl/Joystick.cpp"
 $1 = ($0|0)==(0); //@line 200 "modules/joystick/sdl/Joystick.cpp"
 if ($1) {
  $$0 = 0;
  return ($$0|0); //@line 211 "modules/joystick/sdl/Joystick.cpp"
 }
 $2 = HEAP32[$this>>2]|0; //@line 203 "modules/joystick/sdl/Joystick.cpp"
 $3 = ((($2)) + 64|0); //@line 203 "modules/joystick/sdl/Joystick.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 203 "modules/joystick/sdl/Joystick.cpp"
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($this)|0); //@line 203 "modules/joystick/sdl/Joystick.cpp"
 $6 = ((($this)) + 12|0); //@line 205 "modules/joystick/sdl/Joystick.cpp"
 if ($5) {
  $7 = HEAP32[$6>>2]|0; //@line 205 "modules/joystick/sdl/Joystick.cpp"
  _SDL_GameControllerClose($7); //@line 205 "modules/joystick/sdl/Joystick.cpp"
  HEAP32[$6>>2] = 0; //@line 206 "modules/joystick/sdl/Joystick.cpp"
 }
 $8 = (_SDL_GameControllerOpen($deviceindex)|0); //@line 209 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$6>>2] = $8; //@line 209 "modules/joystick/sdl/Joystick.cpp"
 $9 = HEAP32[$this>>2]|0; //@line 210 "modules/joystick/sdl/Joystick.cpp"
 $10 = ((($9)) + 64|0); //@line 210 "modules/joystick/sdl/Joystick.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 210 "modules/joystick/sdl/Joystick.cpp"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($this)|0); //@line 210 "modules/joystick/sdl/Joystick.cpp"
 $$0 = $12;
 return ($$0|0); //@line 211 "modules/joystick/sdl/Joystick.cpp"
}
function __ZNK4love8joystick3sdl8Joystick9isGamepadEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 215 "modules/joystick/sdl/Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 215 "modules/joystick/sdl/Joystick.cpp"
 $2 = ($1|0)!=(0|0); //@line 215 "modules/joystick/sdl/Joystick.cpp"
 return ($2|0); //@line 215 "modules/joystick/sdl/Joystick.cpp"
}
function __ZNK4love8joystick3sdl8Joystick14getGamepadAxisENS0_8Joystick11GamepadAxisE($this,$axis) {
 $this = $this|0;
 $axis = $axis|0;
 var $$1 = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 220 "modules/joystick/sdl/Joystick.cpp"
 $1 = ((($0)) + 24|0); //@line 220 "modules/joystick/sdl/Joystick.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 220 "modules/joystick/sdl/Joystick.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 220 "modules/joystick/sdl/Joystick.cpp"
 if (!($3)) {
  $$1 = 0.0;
  return (+$$1); //@line 230 "modules/joystick/sdl/Joystick.cpp"
 }
 $4 = HEAP32[$this>>2]|0; //@line 220 "modules/joystick/sdl/Joystick.cpp"
 $5 = ((($4)) + 64|0); //@line 220 "modules/joystick/sdl/Joystick.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 220 "modules/joystick/sdl/Joystick.cpp"
 $7 = (FUNCTION_TABLE_ii[$6 & 8191]($this)|0); //@line 220 "modules/joystick/sdl/Joystick.cpp"
 $8 = ($axis>>>0)<(7); //@line 64 "./common/EnumMap.h"
 $or$cond = $8 & $7; //@line 220 "modules/joystick/sdl/Joystick.cpp"
 if (!($or$cond)) {
  $$1 = 0.0;
  return (+$$1); //@line 230 "modules/joystick/sdl/Joystick.cpp"
 }
 $9 = (((635392 + ($axis<<3)|0)) + 4|0); //@line 64 "./common/EnumMap.h"
 $10 = HEAP8[$9>>0]|0; //@line 64 "./common/EnumMap.h"
 $11 = ($10<<24>>24)==(0); //@line 64 "./common/EnumMap.h"
 if ($11) {
  $$1 = 0.0;
  return (+$$1); //@line 230 "modules/joystick/sdl/Joystick.cpp"
 }
 $12 = (635392 + ($axis<<3)|0); //@line 66 "./common/EnumMap.h"
 $13 = HEAP32[$12>>2]|0; //@line 66 "./common/EnumMap.h"
 $14 = ((($this)) + 12|0); //@line 227 "modules/joystick/sdl/Joystick.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 227 "modules/joystick/sdl/Joystick.cpp"
 $16 = (_SDL_GameControllerGetAxis($15,$13)|0); //@line 227 "modules/joystick/sdl/Joystick.cpp"
 $17 = (+($16<<16>>16)); //@line 229 "modules/joystick/sdl/Joystick.cpp"
 $18 = $17 * 3.0517578125E-5; //@line 229 "modules/joystick/sdl/Joystick.cpp"
 $19 = (+__ZN4love8joystick8Joystick8clampvalEf($18)); //@line 229 "modules/joystick/sdl/Joystick.cpp"
 $$1 = $19;
 return (+$$1); //@line 230 "modules/joystick/sdl/Joystick.cpp"
}
function __ZNK4love8joystick3sdl8Joystick13isGamepadDownERKNSt3__16vectorINS0_8Joystick13GamepadButtonENS3_9allocatorIS6_EEEE($this,$blist) {
 $this = $this|0;
 $blist = $blist|0;
 var $$4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$08 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 234 "modules/joystick/sdl/Joystick.cpp"
 $1 = ((($0)) + 24|0); //@line 234 "modules/joystick/sdl/Joystick.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 234 "modules/joystick/sdl/Joystick.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 234 "modules/joystick/sdl/Joystick.cpp"
 if (!($3)) {
  $$4 = 0;
  return ($$4|0); //@line 249 "modules/joystick/sdl/Joystick.cpp"
 }
 $4 = HEAP32[$this>>2]|0; //@line 234 "modules/joystick/sdl/Joystick.cpp"
 $5 = ((($4)) + 64|0); //@line 234 "modules/joystick/sdl/Joystick.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 234 "modules/joystick/sdl/Joystick.cpp"
 $7 = (FUNCTION_TABLE_ii[$6 & 8191]($this)|0); //@line 234 "modules/joystick/sdl/Joystick.cpp"
 if (!($7)) {
  $$4 = 0;
  return ($$4|0); //@line 249 "modules/joystick/sdl/Joystick.cpp"
 }
 $8 = HEAP32[$blist>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ((($blist)) + 4|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = HEAP32[$9>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($8|0)==($10|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($11) {
  $$4 = 0;
  return ($$4|0); //@line 249 "modules/joystick/sdl/Joystick.cpp"
 }
 $12 = ((($this)) + 12|0); //@line 244 "modules/joystick/sdl/Joystick.cpp"
 $__begin$sroa$0$08 = $8;
 while(1) {
  $13 = HEAP32[$__begin$sroa$0$08>>2]|0; //@line 239 "modules/joystick/sdl/Joystick.cpp"
  $14 = ($13>>>0)<(16); //@line 64 "./common/EnumMap.h"
  if ($14) {
   $15 = (((635576 + ($13<<3)|0)) + 4|0); //@line 64 "./common/EnumMap.h"
   $16 = HEAP8[$15>>0]|0; //@line 64 "./common/EnumMap.h"
   $17 = ($16<<24>>24)==(0); //@line 64 "./common/EnumMap.h"
   if (!($17)) {
    $18 = (635576 + ($13<<3)|0); //@line 66 "./common/EnumMap.h"
    $19 = HEAP32[$18>>2]|0; //@line 66 "./common/EnumMap.h"
    $20 = HEAP32[$12>>2]|0; //@line 244 "modules/joystick/sdl/Joystick.cpp"
    $21 = (_SDL_GameControllerGetButton($20,$19)|0); //@line 244 "modules/joystick/sdl/Joystick.cpp"
    $22 = ($21<<24>>24)==(1); //@line 244 "modules/joystick/sdl/Joystick.cpp"
    if ($22) {
     $$4 = 1;
     label = 9;
     break;
    }
   }
  }
  $23 = ((($__begin$sroa$0$08)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $24 = ($23|0)==($10|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($24) {
   $$4 = 0;
   label = 9;
   break;
  } else {
   $__begin$sroa$0$08 = $23;
  }
 }
 if ((label|0) == 9) {
  return ($$4|0); //@line 249 "modules/joystick/sdl/Joystick.cpp"
 }
 return (0)|0;
}
function __ZNK4love8joystick3sdl8Joystick9getHandleEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 253 "modules/joystick/sdl/Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 253 "modules/joystick/sdl/Joystick.cpp"
 return ($1|0); //@line 253 "modules/joystick/sdl/Joystick.cpp"
}
function __ZNK4love8joystick3sdl8Joystick7getGUIDEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 259 "modules/joystick/sdl/Joystick.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_($agg$result,$0); //@line 259 "modules/joystick/sdl/Joystick.cpp"
 return; //@line 259 "modules/joystick/sdl/Joystick.cpp"
}
function __ZNK4love8joystick3sdl8Joystick13getInstanceIDEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 264 "modules/joystick/sdl/Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 264 "modules/joystick/sdl/Joystick.cpp"
 return ($1|0); //@line 264 "modules/joystick/sdl/Joystick.cpp"
}
function __ZNK4love8joystick3sdl8Joystick5getIDEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 36|0); //@line 269 "modules/joystick/sdl/Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 269 "modules/joystick/sdl/Joystick.cpp"
 return ($1|0); //@line 269 "modules/joystick/sdl/Joystick.cpp"
}
function __ZN4love8joystick3sdl8Joystick20isVibrationSupportedEv($this) {
 $this = $this|0;
 var $$1 = 0, $$not = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $not$ = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love8joystick3sdl8Joystick17checkCreateHapticEv($this)|0); //@line 297 "modules/joystick/sdl/Joystick.cpp"
 if (!($0)) {
  $$1 = 0;
  return ($$1|0); //@line 314 "modules/joystick/sdl/Joystick.cpp"
 }
 $1 = ((($this)) + 16|0); //@line 300 "modules/joystick/sdl/Joystick.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 300 "modules/joystick/sdl/Joystick.cpp"
 $3 = (_SDL_HapticQuery($2)|0); //@line 300 "modules/joystick/sdl/Joystick.cpp"
 $4 = $3 & 4; //@line 302 "modules/joystick/sdl/Joystick.cpp"
 $5 = ($4|0)==(0); //@line 302 "modules/joystick/sdl/Joystick.cpp"
 if (!($5)) {
  $$1 = 1;
  return ($$1|0); //@line 314 "modules/joystick/sdl/Joystick.cpp"
 }
 $6 = HEAP32[$this>>2]|0; //@line 306 "modules/joystick/sdl/Joystick.cpp"
 $7 = ((($6)) + 64|0); //@line 306 "modules/joystick/sdl/Joystick.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 306 "modules/joystick/sdl/Joystick.cpp"
 $9 = (FUNCTION_TABLE_ii[$8 & 8191]($this)|0); //@line 306 "modules/joystick/sdl/Joystick.cpp"
 $$not = $9 ^ 1; //@line 306 "modules/joystick/sdl/Joystick.cpp"
 $10 = $3 & 2048; //@line 306 "modules/joystick/sdl/Joystick.cpp"
 $11 = ($10|0)==(0); //@line 306 "modules/joystick/sdl/Joystick.cpp"
 $or$cond = $11 | $$not; //@line 306 "modules/joystick/sdl/Joystick.cpp"
 if (!($or$cond)) {
  $$1 = 1;
  return ($$1|0); //@line 314 "modules/joystick/sdl/Joystick.cpp"
 }
 $12 = $3 & 2; //@line 310 "modules/joystick/sdl/Joystick.cpp"
 $not$ = ($12|0)!=(0); //@line 313 "modules/joystick/sdl/Joystick.cpp"
 $$1 = $not$;
 return ($$1|0); //@line 314 "modules/joystick/sdl/Joystick.cpp"
}
function __ZN4love8joystick3sdl8Joystick17checkCreateHapticEv($this) {
 $this = $this|0;
 var $$0 = 0, $$pr = 0, $$sroa$4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$sroa$4 = sp;
 $0 = HEAP32[$this>>2]|0; //@line 274 "modules/joystick/sdl/Joystick.cpp"
 $1 = ((($0)) + 24|0); //@line 274 "modules/joystick/sdl/Joystick.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 274 "modules/joystick/sdl/Joystick.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0); //@line 274 "modules/joystick/sdl/Joystick.cpp"
 if (!($3)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 293 "modules/joystick/sdl/Joystick.cpp"
 }
 $4 = (_SDL_WasInit(4096)|0); //@line 277 "modules/joystick/sdl/Joystick.cpp"
 $5 = ($4|0)==(0); //@line 277 "modules/joystick/sdl/Joystick.cpp"
 if ($5) {
  $6 = (_SDL_InitSubSystem(4096)|0); //@line 277 "modules/joystick/sdl/Joystick.cpp"
  $7 = ($6|0)<(0); //@line 277 "modules/joystick/sdl/Joystick.cpp"
  if ($7) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0); //@line 293 "modules/joystick/sdl/Joystick.cpp"
  }
 }
 $8 = ((($this)) + 16|0); //@line 280 "modules/joystick/sdl/Joystick.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 280 "modules/joystick/sdl/Joystick.cpp"
 $10 = ($9|0)==(0|0); //@line 280 "modules/joystick/sdl/Joystick.cpp"
 if (!($10)) {
  $11 = (_SDL_HapticIndex($9)|0); //@line 280 "modules/joystick/sdl/Joystick.cpp"
  $12 = ($11|0)==(-1); //@line 280 "modules/joystick/sdl/Joystick.cpp"
  if (!($12)) {
   $$0 = 1;
   STACKTOP = sp;return ($$0|0); //@line 293 "modules/joystick/sdl/Joystick.cpp"
  }
  $$pr = HEAP32[$8>>2]|0; //@line 283 "modules/joystick/sdl/Joystick.cpp"
  $13 = ($$pr|0)==(0|0); //@line 283 "modules/joystick/sdl/Joystick.cpp"
  if (!($13)) {
   _SDL_HapticClose($$pr); //@line 285 "modules/joystick/sdl/Joystick.cpp"
   HEAP32[$8>>2] = 0; //@line 286 "modules/joystick/sdl/Joystick.cpp"
  }
 }
 $14 = ((($this)) + 8|0); //@line 289 "modules/joystick/sdl/Joystick.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 289 "modules/joystick/sdl/Joystick.cpp"
 $16 = (_SDL_HapticOpenFromJoystick($15)|0); //@line 289 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$8>>2] = $16; //@line 289 "modules/joystick/sdl/Joystick.cpp"
 $17 = ((($this)) + 52|0); //@line 290 "modules/joystick/sdl/Joystick.cpp"
 HEAPF32[$17>>2] = 0.0; //@line 290 "modules/joystick/sdl/Joystick.cpp"
 $18 = ((($this)) + 56|0); //@line 290 "modules/joystick/sdl/Joystick.cpp"
 HEAPF32[$18>>2] = 0.0; //@line 290 "modules/joystick/sdl/Joystick.cpp"
 $19 = ((($this)) + 60|0); //@line 290 "modules/joystick/sdl/Joystick.cpp"
 HEAP16[$19>>1] = 0; //@line 290 "modules/joystick/sdl/Joystick.cpp"
 $20 = ((($this)) + 62|0); //@line 290 "modules/joystick/sdl/Joystick.cpp"
 dest=$20; src=$$sroa$4; stop=dest+66|0; do { HEAP16[dest>>1]=HEAP16[src>>1]|0; dest=dest+2|0; src=src+2|0; } while ((dest|0) < (stop|0)); //@line 290 "modules/joystick/sdl/Joystick.cpp"
 $21 = ((($this)) + 128|0); //@line 290 "modules/joystick/sdl/Joystick.cpp"
 $22 = $21; //@line 290 "modules/joystick/sdl/Joystick.cpp"
 $23 = $22; //@line 290 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$23>>2] = 0; //@line 290 "modules/joystick/sdl/Joystick.cpp"
 $24 = (($22) + 4)|0; //@line 290 "modules/joystick/sdl/Joystick.cpp"
 $25 = $24; //@line 290 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$25>>2] = 0; //@line 290 "modules/joystick/sdl/Joystick.cpp"
 $26 = ((($this)) + 136|0); //@line 290 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$26>>2] = -1; //@line 290 "modules/joystick/sdl/Joystick.cpp"
 $27 = ((($this)) + 140|0); //@line 290 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$27>>2] = -1; //@line 290 "modules/joystick/sdl/Joystick.cpp"
 $28 = ($16|0)!=(0|0); //@line 292 "modules/joystick/sdl/Joystick.cpp"
 $$0 = $28;
 STACKTOP = sp;return ($$0|0); //@line 293 "modules/joystick/sdl/Joystick.cpp"
}
function __ZN4love8joystick3sdl8Joystick12setVibrationEfff($this,$left,$right,$duration) {
 $this = $this|0;
 $left = +$left;
 $right = +$right;
 $duration = +$duration;
 var $$0 = 0, $$sroa$013 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0.0, $35 = 0, $36 = 0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0.0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $9 = 0.0, $length$0 = 0, $or$cond = 0, $or$cond12 = 0, $or$cond3 = 0, $success$1$off0 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$sroa$013 = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 HEAPF32[$$sroa$013>>2] = $left;
 HEAPF32[$0>>2] = 0.0; //@line 341 "modules/joystick/sdl/Joystick.cpp"
 $2 = $left < 0.0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $3 = $2 ? $0 : $$sroa$013; //@line 2648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 HEAPF32[$1>>2] = 1.0; //@line 341 "modules/joystick/sdl/Joystick.cpp"
 $4 = +HEAPF32[$3>>2]; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $5 = $4 > 1.0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $6 = $5 ? $1 : $3; //@line 2575 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $7 = HEAP32[$6>>2]|0; //@line 414 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$$sroa$013>>2] = $7; //@line 341 "modules/joystick/sdl/Joystick.cpp"
 $8 = $right < 0.0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $9 = $8 ? 0.0 : $right; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $10 = $9 > 1.0; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $11 = $10 ? 1.0 : $9; //@line 342 "modules/joystick/sdl/Joystick.cpp"
 $12 = (HEAP32[tempDoublePtr>>2]=$7,+HEAPF32[tempDoublePtr>>2]); //@line 344 "modules/joystick/sdl/Joystick.cpp"
 $13 = $12 == 0.0; //@line 344 "modules/joystick/sdl/Joystick.cpp"
 $14 = $11 == 0.0; //@line 344 "modules/joystick/sdl/Joystick.cpp"
 $or$cond = $14 & $13; //@line 344 "modules/joystick/sdl/Joystick.cpp"
 if ($or$cond) {
  $15 = HEAP32[$this>>2]|0; //@line 345 "modules/joystick/sdl/Joystick.cpp"
  $16 = ((($15)) + 100|0); //@line 345 "modules/joystick/sdl/Joystick.cpp"
  $17 = HEAP32[$16>>2]|0; //@line 345 "modules/joystick/sdl/Joystick.cpp"
  $18 = (FUNCTION_TABLE_ii[$17 & 8191]($this)|0); //@line 345 "modules/joystick/sdl/Joystick.cpp"
  $$0 = $18;
  STACKTOP = sp;return ($$0|0); //@line 429 "modules/joystick/sdl/Joystick.cpp"
 }
 $19 = (__ZN4love8joystick3sdl8Joystick17checkCreateHapticEv($this)|0); //@line 347 "modules/joystick/sdl/Joystick.cpp"
 if (!($19)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 429 "modules/joystick/sdl/Joystick.cpp"
 }
 $20 = !($duration >= 0.0); //@line 351 "modules/joystick/sdl/Joystick.cpp"
 if ($20) {
  $length$0 = -1;
 } else {
  $21 = $duration > 4294967.5; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $22 = $21 ? 4294967.5 : $duration; //@line 354 "modules/joystick/sdl/Joystick.cpp"
  $23 = $22 * 1000.0; //@line 354 "modules/joystick/sdl/Joystick.cpp"
  $24 = (~~(($23))>>>0); //@line 354 "modules/joystick/sdl/Joystick.cpp"
  $length$0 = $24;
 }
 $25 = ((($this)) + 16|0); //@line 358 "modules/joystick/sdl/Joystick.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 358 "modules/joystick/sdl/Joystick.cpp"
 $27 = (_SDL_HapticQuery($26)|0); //@line 358 "modules/joystick/sdl/Joystick.cpp"
 $28 = HEAP32[$25>>2]|0; //@line 359 "modules/joystick/sdl/Joystick.cpp"
 $29 = (_SDL_HapticNumAxes($28)|0); //@line 359 "modules/joystick/sdl/Joystick.cpp"
 $30 = $27 & 4; //@line 361 "modules/joystick/sdl/Joystick.cpp"
 $31 = ($30|0)==(0); //@line 361 "modules/joystick/sdl/Joystick.cpp"
 if ($31) {
  label = 8;
 } else {
  $32 = ((($this)) + 60|0); //@line 363 "modules/joystick/sdl/Joystick.cpp"
  dest=$32; stop=dest+68|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 363 "modules/joystick/sdl/Joystick.cpp"
  HEAP16[$32>>1] = 4; //@line 364 "modules/joystick/sdl/Joystick.cpp"
  $33 = ((($this)) + 64|0); //@line 366 "modules/joystick/sdl/Joystick.cpp"
  HEAP32[$33>>2] = $length$0; //@line 366 "modules/joystick/sdl/Joystick.cpp"
  $34 = $12 * 65535.0; //@line 367 "modules/joystick/sdl/Joystick.cpp"
  $35 = (~~(($34))&65535); //@line 367 "modules/joystick/sdl/Joystick.cpp"
  $36 = ((($this)) + 68|0); //@line 367 "modules/joystick/sdl/Joystick.cpp"
  HEAP16[$36>>1] = $35; //@line 367 "modules/joystick/sdl/Joystick.cpp"
  $37 = $11 * 65535.0; //@line 368 "modules/joystick/sdl/Joystick.cpp"
  $38 = (~~(($37))&65535); //@line 368 "modules/joystick/sdl/Joystick.cpp"
  $39 = ((($32)) + 10|0); //@line 368 "modules/joystick/sdl/Joystick.cpp"
  HEAP16[$39>>1] = $38; //@line 368 "modules/joystick/sdl/Joystick.cpp"
  $40 = (__ZN4love8joystick3sdl8Joystick18runVibrationEffectEv($this)|0); //@line 370 "modules/joystick/sdl/Joystick.cpp"
  if (!($40)) {
   label = 8;
  }
 }
 do {
  if ((label|0) == 8) {
   $41 = HEAP32[$this>>2]|0; //@line 375 "modules/joystick/sdl/Joystick.cpp"
   $42 = ((($41)) + 64|0); //@line 375 "modules/joystick/sdl/Joystick.cpp"
   $43 = HEAP32[$42>>2]|0; //@line 375 "modules/joystick/sdl/Joystick.cpp"
   $44 = (FUNCTION_TABLE_ii[$43 & 8191]($this)|0); //@line 375 "modules/joystick/sdl/Joystick.cpp"
   if ($44) {
    $45 = $27 & 2048; //@line 375 "modules/joystick/sdl/Joystick.cpp"
    $46 = ($45|0)!=(0); //@line 375 "modules/joystick/sdl/Joystick.cpp"
    $47 = ($29|0)==(2); //@line 375 "modules/joystick/sdl/Joystick.cpp"
    $or$cond3 = $46 & $47; //@line 375 "modules/joystick/sdl/Joystick.cpp"
    if ($or$cond3) {
     $48 = $12 * 32767.0; //@line 381 "modules/joystick/sdl/Joystick.cpp"
     $49 = (~~(($48))&65535); //@line 381 "modules/joystick/sdl/Joystick.cpp"
     $50 = ((($this)) + 132|0); //@line 381 "modules/joystick/sdl/Joystick.cpp"
     HEAP16[$50>>1] = $49; //@line 381 "modules/joystick/sdl/Joystick.cpp"
     $51 = ((($this)) + 128|0); //@line 381 "modules/joystick/sdl/Joystick.cpp"
     HEAP16[$51>>1] = $49; //@line 381 "modules/joystick/sdl/Joystick.cpp"
     $52 = $11 * 32767.0; //@line 382 "modules/joystick/sdl/Joystick.cpp"
     $53 = (~~(($52))&65535); //@line 382 "modules/joystick/sdl/Joystick.cpp"
     $54 = ((($this)) + 134|0); //@line 382 "modules/joystick/sdl/Joystick.cpp"
     HEAP16[$54>>1] = $53; //@line 382 "modules/joystick/sdl/Joystick.cpp"
     $55 = ((($this)) + 130|0); //@line 382 "modules/joystick/sdl/Joystick.cpp"
     HEAP16[$55>>1] = $53; //@line 382 "modules/joystick/sdl/Joystick.cpp"
     $56 = ((($this)) + 60|0); //@line 384 "modules/joystick/sdl/Joystick.cpp"
     dest=$56; stop=dest+68|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 384 "modules/joystick/sdl/Joystick.cpp"
     HEAP16[$56>>1] = 2048; //@line 385 "modules/joystick/sdl/Joystick.cpp"
     $57 = ((($this)) + 80|0); //@line 387 "modules/joystick/sdl/Joystick.cpp"
     HEAP32[$57>>2] = $length$0; //@line 387 "modules/joystick/sdl/Joystick.cpp"
     $58 = ((($this)) + 90|0); //@line 388 "modules/joystick/sdl/Joystick.cpp"
     HEAP8[$58>>0] = 2; //@line 388 "modules/joystick/sdl/Joystick.cpp"
     $59 = ((($this)) + 92|0); //@line 389 "modules/joystick/sdl/Joystick.cpp"
     HEAP16[$59>>1] = 10; //@line 389 "modules/joystick/sdl/Joystick.cpp"
     $60 = ((($this)) + 94|0); //@line 390 "modules/joystick/sdl/Joystick.cpp"
     HEAP16[$60>>1] = 2; //@line 390 "modules/joystick/sdl/Joystick.cpp"
     $61 = ((($this)) + 96|0); //@line 391 "modules/joystick/sdl/Joystick.cpp"
     HEAP32[$61>>2] = $51; //@line 391 "modules/joystick/sdl/Joystick.cpp"
     $62 = (__ZN4love8joystick3sdl8Joystick18runVibrationEffectEv($this)|0); //@line 393 "modules/joystick/sdl/Joystick.cpp"
     $success$1$off0 = $62;
    } else {
     $success$1$off0 = 0;
    }
   } else {
    $success$1$off0 = 0;
   }
   $63 = $27 & 2; //@line 398 "modules/joystick/sdl/Joystick.cpp"
   $64 = ($63|0)==(0); //@line 398 "modules/joystick/sdl/Joystick.cpp"
   $or$cond12 = $64 | $success$1$off0; //@line 398 "modules/joystick/sdl/Joystick.cpp"
   if ($or$cond12) {
    if ($success$1$off0) {
     break;
    }
   } else {
    $65 = ((($this)) + 60|0); //@line 400 "modules/joystick/sdl/Joystick.cpp"
    dest=$65; stop=dest+68|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 400 "modules/joystick/sdl/Joystick.cpp"
    HEAP16[$65>>1] = 2; //@line 401 "modules/joystick/sdl/Joystick.cpp"
    $66 = ((($this)) + 80|0); //@line 403 "modules/joystick/sdl/Joystick.cpp"
    HEAP32[$66>>2] = $length$0; //@line 403 "modules/joystick/sdl/Joystick.cpp"
    $67 = ((($this)) + 90|0); //@line 404 "modules/joystick/sdl/Joystick.cpp"
    HEAP16[$67>>1] = 10; //@line 404 "modules/joystick/sdl/Joystick.cpp"
    $68 = $12 < $11; //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $69 = $68 ? $11 : $12; //@line 406 "modules/joystick/sdl/Joystick.cpp"
    $70 = $69 * 32767.0; //@line 407 "modules/joystick/sdl/Joystick.cpp"
    $71 = (~~(($70))); //@line 407 "modules/joystick/sdl/Joystick.cpp"
    $72 = ((($this)) + 92|0); //@line 407 "modules/joystick/sdl/Joystick.cpp"
    HEAP16[$72>>1] = $71; //@line 407 "modules/joystick/sdl/Joystick.cpp"
    $73 = (__ZN4love8joystick3sdl8Joystick18runVibrationEffectEv($this)|0); //@line 409 "modules/joystick/sdl/Joystick.cpp"
    if ($73) {
     break;
    }
   }
   $81 = ((($this)) + 56|0); //@line 424 "modules/joystick/sdl/Joystick.cpp"
   HEAPF32[$81>>2] = 0.0; //@line 424 "modules/joystick/sdl/Joystick.cpp"
   $82 = ((($this)) + 52|0); //@line 424 "modules/joystick/sdl/Joystick.cpp"
   HEAPF32[$82>>2] = 0.0; //@line 424 "modules/joystick/sdl/Joystick.cpp"
   $83 = ((($this)) + 140|0); //@line 425 "modules/joystick/sdl/Joystick.cpp"
   HEAP32[$83>>2] = -1; //@line 425 "modules/joystick/sdl/Joystick.cpp"
   $$0 = 0;
   STACKTOP = sp;return ($$0|0); //@line 429 "modules/joystick/sdl/Joystick.cpp"
  }
 } while(0);
 $74 = ((($this)) + 52|0); //@line 414 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$74>>2] = $7; //@line 414 "modules/joystick/sdl/Joystick.cpp"
 $75 = ((($this)) + 56|0); //@line 415 "modules/joystick/sdl/Joystick.cpp"
 HEAPF32[$75>>2] = $11; //@line 415 "modules/joystick/sdl/Joystick.cpp"
 $76 = ($length$0|0)==(-1); //@line 417 "modules/joystick/sdl/Joystick.cpp"
 if ($76) {
  $77 = ((($this)) + 140|0); //@line 418 "modules/joystick/sdl/Joystick.cpp"
  HEAP32[$77>>2] = -1; //@line 418 "modules/joystick/sdl/Joystick.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 429 "modules/joystick/sdl/Joystick.cpp"
 } else {
  $78 = (_SDL_GetTicks()|0); //@line 420 "modules/joystick/sdl/Joystick.cpp"
  $79 = (($78) + ($length$0))|0; //@line 420 "modules/joystick/sdl/Joystick.cpp"
  $80 = ((($this)) + 140|0); //@line 420 "modules/joystick/sdl/Joystick.cpp"
  HEAP32[$80>>2] = $79; //@line 420 "modules/joystick/sdl/Joystick.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 429 "modules/joystick/sdl/Joystick.cpp"
 }
 return (0)|0;
}
function __ZN4love8joystick3sdl8Joystick18runVibrationEffectEv($this) {
 $this = $this|0;
 var $$0 = 0, $$pre = 0, $$pre$phi2Z2D = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 136|0); //@line 318 "modules/joystick/sdl/Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 318 "modules/joystick/sdl/Joystick.cpp"
 $2 = ($1|0)==(-1); //@line 318 "modules/joystick/sdl/Joystick.cpp"
 $$pre = ((($this)) + 16|0); //@line 331 "modules/joystick/sdl/Joystick.cpp"
 if ($2) {
  $$pre1 = ((($this)) + 60|0); //@line 331 "modules/joystick/sdl/Joystick.cpp"
  $$pre$phi2Z2D = $$pre1;
 } else {
  $3 = HEAP32[$$pre>>2]|0; //@line 320 "modules/joystick/sdl/Joystick.cpp"
  $4 = ((($this)) + 60|0); //@line 320 "modules/joystick/sdl/Joystick.cpp"
  $5 = (_SDL_HapticUpdateEffect($3,$1,$4)|0); //@line 320 "modules/joystick/sdl/Joystick.cpp"
  $6 = ($5|0)==(0); //@line 320 "modules/joystick/sdl/Joystick.cpp"
  if ($6) {
   $7 = HEAP32[$$pre>>2]|0; //@line 322 "modules/joystick/sdl/Joystick.cpp"
   $8 = HEAP32[$0>>2]|0; //@line 322 "modules/joystick/sdl/Joystick.cpp"
   $9 = (_SDL_HapticRunEffect($7,$8,1)|0); //@line 322 "modules/joystick/sdl/Joystick.cpp"
   $10 = ($9|0)==(0); //@line 322 "modules/joystick/sdl/Joystick.cpp"
   if ($10) {
    $$0 = 1;
    return ($$0|0); //@line 337 "modules/joystick/sdl/Joystick.cpp"
   }
  }
  $11 = HEAP32[$$pre>>2]|0; //@line 327 "modules/joystick/sdl/Joystick.cpp"
  $12 = HEAP32[$0>>2]|0; //@line 327 "modules/joystick/sdl/Joystick.cpp"
  _SDL_HapticDestroyEffect($11,$12); //@line 327 "modules/joystick/sdl/Joystick.cpp"
  HEAP32[$0>>2] = -1; //@line 328 "modules/joystick/sdl/Joystick.cpp"
  $$pre$phi2Z2D = $4;
 }
 $13 = HEAP32[$$pre>>2]|0; //@line 331 "modules/joystick/sdl/Joystick.cpp"
 $14 = (_SDL_HapticNewEffect($13,$$pre$phi2Z2D)|0); //@line 331 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$0>>2] = $14; //@line 331 "modules/joystick/sdl/Joystick.cpp"
 $15 = ($14|0)==(-1); //@line 333 "modules/joystick/sdl/Joystick.cpp"
 if (!($15)) {
  $16 = HEAP32[$$pre>>2]|0; //@line 333 "modules/joystick/sdl/Joystick.cpp"
  $17 = (_SDL_HapticRunEffect($16,$14,1)|0); //@line 333 "modules/joystick/sdl/Joystick.cpp"
  $18 = ($17|0)==(0); //@line 333 "modules/joystick/sdl/Joystick.cpp"
  if ($18) {
   $$0 = 1;
   return ($$0|0); //@line 337 "modules/joystick/sdl/Joystick.cpp"
  }
 }
 $$0 = 0;
 return ($$0|0); //@line 337 "modules/joystick/sdl/Joystick.cpp"
}
function __ZN4love8joystick3sdl8Joystick12setVibrationEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $success$0$off01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_WasInit(4096)|0); //@line 435 "modules/joystick/sdl/Joystick.cpp"
 $1 = ($0|0)==(0); //@line 435 "modules/joystick/sdl/Joystick.cpp"
 if (!($1)) {
  $2 = ((($this)) + 16|0); //@line 435 "modules/joystick/sdl/Joystick.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 435 "modules/joystick/sdl/Joystick.cpp"
  $4 = ($3|0)==(0|0); //@line 435 "modules/joystick/sdl/Joystick.cpp"
  if (!($4)) {
   $5 = (_SDL_HapticIndex($3)|0); //@line 435 "modules/joystick/sdl/Joystick.cpp"
   $6 = ($5|0)==(-1); //@line 435 "modules/joystick/sdl/Joystick.cpp"
   if (!($6)) {
    $7 = HEAP32[$2>>2]|0; //@line 436 "modules/joystick/sdl/Joystick.cpp"
    $8 = ((($this)) + 136|0); //@line 436 "modules/joystick/sdl/Joystick.cpp"
    $9 = HEAP32[$8>>2]|0; //@line 436 "modules/joystick/sdl/Joystick.cpp"
    $10 = (_SDL_HapticStopEffect($7,$9)|0); //@line 436 "modules/joystick/sdl/Joystick.cpp"
    $11 = ($10|0)==(0); //@line 436 "modules/joystick/sdl/Joystick.cpp"
    if (!($11)) {
     $success$0$off01 = 0;
     return ($success$0$off01|0); //@line 441 "modules/joystick/sdl/Joystick.cpp"
    }
   }
  }
 }
 $12 = ((($this)) + 56|0); //@line 439 "modules/joystick/sdl/Joystick.cpp"
 HEAPF32[$12>>2] = 0.0; //@line 439 "modules/joystick/sdl/Joystick.cpp"
 $13 = ((($this)) + 52|0); //@line 439 "modules/joystick/sdl/Joystick.cpp"
 HEAPF32[$13>>2] = 0.0; //@line 439 "modules/joystick/sdl/Joystick.cpp"
 $success$0$off01 = 1;
 return ($success$0$off01|0); //@line 441 "modules/joystick/sdl/Joystick.cpp"
}
function __ZN4love8joystick3sdl8Joystick12getVibrationERfS3_($this,$left,$right) {
 $this = $this|0;
 $left = $left|0;
 $right = $right|0;
 var $$pre = 0, $$pre$phi4Z2D = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 446 "modules/joystick/sdl/Joystick.cpp"
 $1 = ((($this)) + 140|0); //@line 446 "modules/joystick/sdl/Joystick.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 446 "modules/joystick/sdl/Joystick.cpp"
 $3 = ($2|0)==(-1); //@line 446 "modules/joystick/sdl/Joystick.cpp"
 if (!($3)) {
  $4 = (_SDL_GetTicks()|0); //@line 451 "modules/joystick/sdl/Joystick.cpp"
  $5 = (($2) - ($4))|0; //@line 451 "modules/joystick/sdl/Joystick.cpp"
  $6 = ($5|0)<(1); //@line 451 "modules/joystick/sdl/Joystick.cpp"
  if ($6) {
   $7 = HEAP32[$this>>2]|0; //@line 453 "modules/joystick/sdl/Joystick.cpp"
   $8 = ((($7)) + 100|0); //@line 453 "modules/joystick/sdl/Joystick.cpp"
   $9 = HEAP32[$8>>2]|0; //@line 453 "modules/joystick/sdl/Joystick.cpp"
   (FUNCTION_TABLE_ii[$9 & 8191]($this)|0); //@line 453 "modules/joystick/sdl/Joystick.cpp"
   HEAP32[$1>>2] = -1; //@line 454 "modules/joystick/sdl/Joystick.cpp"
  }
 }
 $10 = ((($this)) + 136|0); //@line 459 "modules/joystick/sdl/Joystick.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 459 "modules/joystick/sdl/Joystick.cpp"
 $12 = ((($this)) + 16|0); //@line 460 "modules/joystick/sdl/Joystick.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 460 "modules/joystick/sdl/Joystick.cpp"
 $14 = ($13|0)==(0|0); //@line 460 "modules/joystick/sdl/Joystick.cpp"
 $15 = ($11|0)==(-1); //@line 460 "modules/joystick/sdl/Joystick.cpp"
 $or$cond = $15 | $14; //@line 460 "modules/joystick/sdl/Joystick.cpp"
 if (!($or$cond)) {
  $16 = (_SDL_HapticGetEffectStatus($13,$11)|0); //@line 460 "modules/joystick/sdl/Joystick.cpp"
  $17 = ($16|0)==(1); //@line 460 "modules/joystick/sdl/Joystick.cpp"
  if ($17) {
   $$pre = HEAP32[$0>>2]|0; //@line 463 "modules/joystick/sdl/Joystick.cpp"
   $$pre3 = ((($this)) + 56|0); //@line 464 "modules/joystick/sdl/Joystick.cpp"
   $$pre$phi4Z2D = $$pre3;$19 = $$pre;
   HEAP32[$left>>2] = $19; //@line 463 "modules/joystick/sdl/Joystick.cpp"
   $20 = HEAP32[$$pre$phi4Z2D>>2]|0; //@line 464 "modules/joystick/sdl/Joystick.cpp"
   HEAP32[$right>>2] = $20; //@line 464 "modules/joystick/sdl/Joystick.cpp"
   return; //@line 465 "modules/joystick/sdl/Joystick.cpp"
  }
 }
 $18 = ((($this)) + 56|0); //@line 461 "modules/joystick/sdl/Joystick.cpp"
 HEAPF32[$18>>2] = 0.0; //@line 461 "modules/joystick/sdl/Joystick.cpp"
 HEAPF32[$0>>2] = 0.0; //@line 461 "modules/joystick/sdl/Joystick.cpp"
 $$pre$phi4Z2D = $18;$19 = 0;
 HEAP32[$left>>2] = $19; //@line 463 "modules/joystick/sdl/Joystick.cpp"
 $20 = HEAP32[$$pre$phi4Z2D>>2]|0; //@line 464 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$right>>2] = $20; //@line 464 "modules/joystick/sdl/Joystick.cpp"
 return; //@line 465 "modules/joystick/sdl/Joystick.cpp"
}
function __GLOBAL__sub_I_Joystick_cpp() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $8 = 0, $9 = 0, $exitcond$i$i = 0, $exitcond$i$i3 = 0, $i$01$i$i = 0, $i$01$i$i1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[(635084)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635092)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635100)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635108)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635116)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635124)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635132)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635140)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635148)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635156)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635164)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635172)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635180)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635188)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635196)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635204)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635212)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635220)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635228)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635236)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635244)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635252)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635260)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635268)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635276)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635284)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635292)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635300)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635308)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635316)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635324)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635332)>>0] = 0; //@line 90 "./common/EnumMap.h"
 $i$01$i$i = 0;
 while(1) {
  $0 = (26236 + ($i$01$i$i<<3)|0); //@line 46 "./common/EnumMap.h"
  $1 = HEAP32[$0>>2]|0; //@line 46 "./common/EnumMap.h"
  $2 = (((26236 + ($i$01$i$i<<3)|0)) + 4|0); //@line 47 "./common/EnumMap.h"
  $3 = HEAP8[$2>>0]|0; //@line 47 "./common/EnumMap.h"
  $4 = $3&255; //@line 47 "./common/EnumMap.h"
  $5 = ($1>>>0)<(16); //@line 49 "./common/EnumMap.h"
  if ($5) {
   $6 = (635208 + ($1<<3)|0); //@line 51 "./common/EnumMap.h"
   HEAP32[$6>>2] = $4; //@line 51 "./common/EnumMap.h"
   $7 = (((635208 + ($1<<3)|0)) + 4|0); //@line 52 "./common/EnumMap.h"
   HEAP8[$7>>0] = 1; //@line 52 "./common/EnumMap.h"
  }
  $8 = ($3&255)<(16); //@line 54 "./common/EnumMap.h"
  if ($8) {
   $9 = (635080 + ($4<<3)|0); //@line 56 "./common/EnumMap.h"
   HEAP32[$9>>2] = $1; //@line 56 "./common/EnumMap.h"
   $10 = (((635080 + ($4<<3)|0)) + 4|0); //@line 57 "./common/EnumMap.h"
   HEAP8[$10>>0] = 1; //@line 57 "./common/EnumMap.h"
  }
  $11 = (($i$01$i$i) + 1)|0; //@line 44 "./common/EnumMap.h"
  $exitcond$i$i = ($11|0)==(9); //@line 44 "./common/EnumMap.h"
  if ($exitcond$i$i) {
   break;
  } else {
   $i$01$i$i = $11;
  }
 }
 HEAP8[(635340)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635348)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635356)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635364)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635372)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635380)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635388)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635396)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635404)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635412)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635420)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635428)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635436)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635444)>>0] = 0; //@line 90 "./common/EnumMap.h"
 $12 = HEAP32[6577]|0; //@line 46 "./common/EnumMap.h"
 $13 = HEAP32[(26312)>>2]|0; //@line 47 "./common/EnumMap.h"
 $14 = ($12>>>0)<(7); //@line 49 "./common/EnumMap.h"
 if ($14) {
  $15 = (635392 + ($12<<3)|0); //@line 51 "./common/EnumMap.h"
  HEAP32[$15>>2] = $13; //@line 51 "./common/EnumMap.h"
  $16 = (((635392 + ($12<<3)|0)) + 4|0); //@line 52 "./common/EnumMap.h"
  HEAP8[$16>>0] = 1; //@line 52 "./common/EnumMap.h"
 }
 $17 = ($13>>>0)<(7); //@line 54 "./common/EnumMap.h"
 if ($17) {
  $18 = (635336 + ($13<<3)|0); //@line 56 "./common/EnumMap.h"
  HEAP32[$18>>2] = $12; //@line 56 "./common/EnumMap.h"
  $19 = (((635336 + ($13<<3)|0)) + 4|0); //@line 57 "./common/EnumMap.h"
  HEAP8[$19>>0] = 1; //@line 57 "./common/EnumMap.h"
 }
 $20 = HEAP32[(26316)>>2]|0; //@line 46 "./common/EnumMap.h"
 $21 = HEAP32[(26320)>>2]|0; //@line 47 "./common/EnumMap.h"
 $22 = ($20>>>0)<(7); //@line 49 "./common/EnumMap.h"
 if ($22) {
  $23 = (635392 + ($20<<3)|0); //@line 51 "./common/EnumMap.h"
  HEAP32[$23>>2] = $21; //@line 51 "./common/EnumMap.h"
  $24 = (((635392 + ($20<<3)|0)) + 4|0); //@line 52 "./common/EnumMap.h"
  HEAP8[$24>>0] = 1; //@line 52 "./common/EnumMap.h"
 }
 $25 = ($21>>>0)<(7); //@line 54 "./common/EnumMap.h"
 if ($25) {
  $26 = (635336 + ($21<<3)|0); //@line 56 "./common/EnumMap.h"
  HEAP32[$26>>2] = $20; //@line 56 "./common/EnumMap.h"
  $27 = (((635336 + ($21<<3)|0)) + 4|0); //@line 57 "./common/EnumMap.h"
  HEAP8[$27>>0] = 1; //@line 57 "./common/EnumMap.h"
 }
 $28 = HEAP32[(26324)>>2]|0; //@line 46 "./common/EnumMap.h"
 $29 = HEAP32[(26328)>>2]|0; //@line 47 "./common/EnumMap.h"
 $30 = ($28>>>0)<(7); //@line 49 "./common/EnumMap.h"
 if ($30) {
  $31 = (635392 + ($28<<3)|0); //@line 51 "./common/EnumMap.h"
  HEAP32[$31>>2] = $29; //@line 51 "./common/EnumMap.h"
  $32 = (((635392 + ($28<<3)|0)) + 4|0); //@line 52 "./common/EnumMap.h"
  HEAP8[$32>>0] = 1; //@line 52 "./common/EnumMap.h"
 }
 $33 = ($29>>>0)<(7); //@line 54 "./common/EnumMap.h"
 if ($33) {
  $34 = (635336 + ($29<<3)|0); //@line 56 "./common/EnumMap.h"
  HEAP32[$34>>2] = $28; //@line 56 "./common/EnumMap.h"
  $35 = (((635336 + ($29<<3)|0)) + 4|0); //@line 57 "./common/EnumMap.h"
  HEAP8[$35>>0] = 1; //@line 57 "./common/EnumMap.h"
 }
 $36 = HEAP32[(26332)>>2]|0; //@line 46 "./common/EnumMap.h"
 $37 = HEAP32[(26336)>>2]|0; //@line 47 "./common/EnumMap.h"
 $38 = ($36>>>0)<(7); //@line 49 "./common/EnumMap.h"
 if ($38) {
  $39 = (635392 + ($36<<3)|0); //@line 51 "./common/EnumMap.h"
  HEAP32[$39>>2] = $37; //@line 51 "./common/EnumMap.h"
  $40 = (((635392 + ($36<<3)|0)) + 4|0); //@line 52 "./common/EnumMap.h"
  HEAP8[$40>>0] = 1; //@line 52 "./common/EnumMap.h"
 }
 $41 = ($37>>>0)<(7); //@line 54 "./common/EnumMap.h"
 if ($41) {
  $42 = (635336 + ($37<<3)|0); //@line 56 "./common/EnumMap.h"
  HEAP32[$42>>2] = $36; //@line 56 "./common/EnumMap.h"
  $43 = (((635336 + ($37<<3)|0)) + 4|0); //@line 57 "./common/EnumMap.h"
  HEAP8[$43>>0] = 1; //@line 57 "./common/EnumMap.h"
 }
 $44 = HEAP32[(26340)>>2]|0; //@line 46 "./common/EnumMap.h"
 $45 = HEAP32[(26344)>>2]|0; //@line 47 "./common/EnumMap.h"
 $46 = ($44>>>0)<(7); //@line 49 "./common/EnumMap.h"
 if ($46) {
  $47 = (635392 + ($44<<3)|0); //@line 51 "./common/EnumMap.h"
  HEAP32[$47>>2] = $45; //@line 51 "./common/EnumMap.h"
  $48 = (((635392 + ($44<<3)|0)) + 4|0); //@line 52 "./common/EnumMap.h"
  HEAP8[$48>>0] = 1; //@line 52 "./common/EnumMap.h"
 }
 $49 = ($45>>>0)<(7); //@line 54 "./common/EnumMap.h"
 if ($49) {
  $50 = (635336 + ($45<<3)|0); //@line 56 "./common/EnumMap.h"
  HEAP32[$50>>2] = $44; //@line 56 "./common/EnumMap.h"
  $51 = (((635336 + ($45<<3)|0)) + 4|0); //@line 57 "./common/EnumMap.h"
  HEAP8[$51>>0] = 1; //@line 57 "./common/EnumMap.h"
 }
 $52 = HEAP32[(26348)>>2]|0; //@line 46 "./common/EnumMap.h"
 $53 = HEAP32[(26352)>>2]|0; //@line 47 "./common/EnumMap.h"
 $54 = ($52>>>0)<(7); //@line 49 "./common/EnumMap.h"
 if ($54) {
  $55 = (635392 + ($52<<3)|0); //@line 51 "./common/EnumMap.h"
  HEAP32[$55>>2] = $53; //@line 51 "./common/EnumMap.h"
  $56 = (((635392 + ($52<<3)|0)) + 4|0); //@line 52 "./common/EnumMap.h"
  HEAP8[$56>>0] = 1; //@line 52 "./common/EnumMap.h"
 }
 $57 = ($53>>>0)<(7); //@line 54 "./common/EnumMap.h"
 if ($57) {
  $58 = (635336 + ($53<<3)|0); //@line 56 "./common/EnumMap.h"
  HEAP32[$58>>2] = $52; //@line 56 "./common/EnumMap.h"
  $59 = (((635336 + ($53<<3)|0)) + 4|0); //@line 57 "./common/EnumMap.h"
  HEAP8[$59>>0] = 1; //@line 57 "./common/EnumMap.h"
 }
 HEAP8[(635452)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635460)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635468)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635476)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635484)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635492)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635500)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635508)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635516)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635524)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635532)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635540)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635548)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635556)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635564)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635572)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635580)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635588)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635596)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635604)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635612)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635620)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635628)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635636)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635644)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635652)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635660)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635668)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635676)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635684)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635692)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(635700)>>0] = 0; //@line 90 "./common/EnumMap.h"
 $i$01$i$i1 = 0;
 while(1) {
  $60 = (26356 + ($i$01$i$i1<<3)|0); //@line 46 "./common/EnumMap.h"
  $61 = HEAP32[$60>>2]|0; //@line 46 "./common/EnumMap.h"
  $62 = (((26356 + ($i$01$i$i1<<3)|0)) + 4|0); //@line 47 "./common/EnumMap.h"
  $63 = HEAP32[$62>>2]|0; //@line 47 "./common/EnumMap.h"
  $64 = ($61>>>0)<(16); //@line 49 "./common/EnumMap.h"
  if ($64) {
   $65 = (635576 + ($61<<3)|0); //@line 51 "./common/EnumMap.h"
   HEAP32[$65>>2] = $63; //@line 51 "./common/EnumMap.h"
   $66 = (((635576 + ($61<<3)|0)) + 4|0); //@line 52 "./common/EnumMap.h"
   HEAP8[$66>>0] = 1; //@line 52 "./common/EnumMap.h"
  }
  $67 = ($63>>>0)<(16); //@line 54 "./common/EnumMap.h"
  if ($67) {
   $68 = (635448 + ($63<<3)|0); //@line 56 "./common/EnumMap.h"
   HEAP32[$68>>2] = $61; //@line 56 "./common/EnumMap.h"
   $69 = (((635448 + ($63<<3)|0)) + 4|0); //@line 57 "./common/EnumMap.h"
   HEAP8[$69>>0] = 1; //@line 57 "./common/EnumMap.h"
  }
  $70 = (($i$01$i$i1) + 1)|0; //@line 44 "./common/EnumMap.h"
  $exitcond$i$i3 = ($70|0)==(15); //@line 44 "./common/EnumMap.h"
  if ($exitcond$i$i3) {
   break;
  } else {
   $i$01$i$i1 = $70;
  }
 }
 return;
}
function __ZN4love8joystick3sdl8JoystickC2Ei($this,$id) {
 $this = $this|0;
 $id = $id|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 37 "./modules/joystick/Joystick.h"
 HEAP32[$this>>2] = (26128); //@line 48 "modules/joystick/sdl/Joystick.cpp"
 $0 = ((($this)) + 8|0); //@line 42 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$0>>2] = 0; //@line 42 "modules/joystick/sdl/Joystick.cpp"
 $1 = ((($this)) + 12|0); //@line 43 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$1>>2] = 0; //@line 43 "modules/joystick/sdl/Joystick.cpp"
 $2 = ((($this)) + 16|0); //@line 44 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$2>>2] = 0; //@line 44 "modules/joystick/sdl/Joystick.cpp"
 $3 = ((($this)) + 20|0); //@line 45 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$3>>2] = -1; //@line 45 "modules/joystick/sdl/Joystick.cpp"
 $4 = ((($this)) + 24|0); //@line 41 "modules/joystick/sdl/Joystick.cpp"
 ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($this)) + 36|0); //@line 46 "modules/joystick/sdl/Joystick.cpp"
 HEAP32[$5>>2] = $id; //@line 46 "modules/joystick/sdl/Joystick.cpp"
 $6 = ((($this)) + 40|0); //@line 41 "modules/joystick/sdl/Joystick.cpp"
 $7 = ((($this)) + 128|0); //@line 117 "modules/joystick/sdl/Joystick.h"
 $8 = $7;
 $9 = $8;
 HEAP32[$9>>2] = 0;
 $10 = (($8) + 4)|0;
 $11 = $10;
 HEAP32[$11>>2] = 0;
 $12 = ((($this)) + 136|0); //@line 113 "modules/joystick/sdl/Joystick.h"
 ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;HEAP32[$6+12>>2]=0|0;HEAP32[$6+16>>2]=0|0;HEAP16[$6+20>>1]=0|0; //@line 109 "modules/joystick/sdl/Joystick.h"
 HEAP32[$12>>2] = -1; //@line 113 "modules/joystick/sdl/Joystick.h"
 $13 = ((($this)) + 140|0); //@line 114 "modules/joystick/sdl/Joystick.h"
 HEAP32[$13>>2] = -1; //@line 114 "modules/joystick/sdl/Joystick.h"
 return; //@line 49 "modules/joystick/sdl/Joystick.cpp"
}
function __ZN4love8joystick3sdl8Joystick11getConstantEhRNS0_8Joystick3HatE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in&255)<(16); //@line 75 "./common/EnumMap.h"
 if ($0) {
  $1 = $in&255; //@line 75 "./common/EnumMap.h"
  $2 = (((635080 + ($1<<3)|0)) + 4|0); //@line 75 "./common/EnumMap.h"
  $3 = HEAP8[$2>>0]|0; //@line 75 "./common/EnumMap.h"
  $4 = ($3<<24>>24)==(0); //@line 75 "./common/EnumMap.h"
  if ($4) {
   $$0$i = 0;
  } else {
   $5 = (635080 + ($1<<3)|0); //@line 77 "./common/EnumMap.h"
   $6 = HEAP32[$5>>2]|0; //@line 77 "./common/EnumMap.h"
   HEAP32[$out>>2] = $6; //@line 77 "./common/EnumMap.h"
   $$0$i = 1;
  }
 } else {
  $$0$i = 0;
 }
 return ($$0$i|0); //@line 469 "modules/joystick/sdl/Joystick.cpp"
}
function __ZN4love8joystick3sdl8Joystick11getConstantENS0_8Joystick11GamepadAxisER22SDL_GameControllerAxis($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)<(7); //@line 64 "./common/EnumMap.h"
 if ($0) {
  $1 = (((635392 + ($in<<3)|0)) + 4|0); //@line 64 "./common/EnumMap.h"
  $2 = HEAP8[$1>>0]|0; //@line 64 "./common/EnumMap.h"
  $3 = ($2<<24>>24)==(0); //@line 64 "./common/EnumMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   $4 = (635392 + ($in<<3)|0); //@line 66 "./common/EnumMap.h"
   $5 = HEAP32[$4>>2]|0; //@line 66 "./common/EnumMap.h"
   HEAP32[$out>>2] = $5; //@line 66 "./common/EnumMap.h"
   $$0$i = 1;
  }
 } else {
  $$0$i = 0;
 }
 return ($$0$i|0); //@line 484 "modules/joystick/sdl/Joystick.cpp"
}
function __ZN4love8joystick3sdl8Joystick11getConstantENS0_8Joystick13GamepadButtonER24SDL_GameControllerButton($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)<(16); //@line 64 "./common/EnumMap.h"
 if ($0) {
  $1 = (((635576 + ($in<<3)|0)) + 4|0); //@line 64 "./common/EnumMap.h"
  $2 = HEAP8[$1>>0]|0; //@line 64 "./common/EnumMap.h"
  $3 = ($2<<24>>24)==(0); //@line 64 "./common/EnumMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   $4 = (635576 + ($in<<3)|0); //@line 66 "./common/EnumMap.h"
   $5 = HEAP32[$4>>2]|0; //@line 66 "./common/EnumMap.h"
   HEAP32[$out>>2] = $5; //@line 66 "./common/EnumMap.h"
   $$0$i = 1;
  }
 } else {
  $$0$i = 0;
 }
 return ($$0$i|0); //@line 494 "modules/joystick/sdl/Joystick.cpp"
}
function __ZN4love8joystick3sdl8Joystick11getConstantENS0_8Joystick3HatERh($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)<(16); //@line 64 "./common/EnumMap.h"
 if ($0) {
  $1 = (((635208 + ($in<<3)|0)) + 4|0); //@line 64 "./common/EnumMap.h"
  $2 = HEAP8[$1>>0]|0; //@line 64 "./common/EnumMap.h"
  $3 = ($2<<24>>24)==(0); //@line 64 "./common/EnumMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   $4 = (635208 + ($in<<3)|0); //@line 66 "./common/EnumMap.h"
   $5 = HEAP32[$4>>2]|0; //@line 66 "./common/EnumMap.h"
   $6 = $5&255; //@line 66 "./common/EnumMap.h"
   HEAP8[$out>>0] = $6; //@line 66 "./common/EnumMap.h"
   $$0$i = 1;
  }
 } else {
  $$0$i = 0;
 }
 return ($$0$i|0); //@line 474 "modules/joystick/sdl/Joystick.cpp"
}
function __ZN4love8joystick3sdl8Joystick11getConstantE22SDL_GameControllerAxisRNS0_8Joystick11GamepadAxisE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)<(7); //@line 75 "./common/EnumMap.h"
 if ($0) {
  $1 = (((635336 + ($in<<3)|0)) + 4|0); //@line 75 "./common/EnumMap.h"
  $2 = HEAP8[$1>>0]|0; //@line 75 "./common/EnumMap.h"
  $3 = ($2<<24>>24)==(0); //@line 75 "./common/EnumMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   $4 = (635336 + ($in<<3)|0); //@line 77 "./common/EnumMap.h"
   $5 = HEAP32[$4>>2]|0; //@line 77 "./common/EnumMap.h"
   HEAP32[$out>>2] = $5; //@line 77 "./common/EnumMap.h"
   $$0$i = 1;
  }
 } else {
  $$0$i = 0;
 }
 return ($$0$i|0); //@line 479 "modules/joystick/sdl/Joystick.cpp"
}
function __ZN4love8joystick3sdl8Joystick11getConstantE24SDL_GameControllerButtonRNS0_8Joystick13GamepadButtonE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)<(16); //@line 75 "./common/EnumMap.h"
 if ($0) {
  $1 = (((635448 + ($in<<3)|0)) + 4|0); //@line 75 "./common/EnumMap.h"
  $2 = HEAP8[$1>>0]|0; //@line 75 "./common/EnumMap.h"
  $3 = ($2<<24>>24)==(0); //@line 75 "./common/EnumMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   $4 = (635448 + ($in<<3)|0); //@line 77 "./common/EnumMap.h"
   $5 = HEAP32[$4>>2]|0; //@line 77 "./common/EnumMap.h"
   HEAP32[$out>>2] = $5; //@line 77 "./common/EnumMap.h"
   $$0$i = 1;
  }
 } else {
  $$0$i = 0;
 }
 return ($$0$i|0); //@line 489 "modules/joystick/sdl/Joystick.cpp"
}
function __ZN4love8joystick3sdl14JoystickModuleD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$010 = 0, $__begin$sroa$0$011 = 0, $__f$01$i$i = 0, $__f$01$i$i1 = 0, $lpad$phi$index = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i4 = 0, $scevgep4$i$i$i = 0;
 var $scevgep4$i$i$i5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (26484); //@line 58 "modules/joystick/sdl/JoystickModule.cpp"
 $0 = ((($this)) + 20|0); //@line 60 "modules/joystick/sdl/JoystickModule.cpp"
 $1 = ((($this)) + 24|0); //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $__begin$sroa$0$010 = HEAP32[$1>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $2 = $__begin$sroa$0$010; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $3 = ($2|0)==($0|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 L1: do {
  if ($3) {
   label = 2;
  } else {
   $41 = $2;$__begin$sroa$0$011 = $__begin$sroa$0$010;
   while(1) {
    $40 = ((($41)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $42 = HEAP32[$40>>2]|0; //@line 60 "modules/joystick/sdl/JoystickModule.cpp"
    $43 = HEAP32[$42>>2]|0; //@line 62 "modules/joystick/sdl/JoystickModule.cpp"
    $44 = ((($43)) + 20|0); //@line 62 "modules/joystick/sdl/JoystickModule.cpp"
    $45 = HEAP32[$44>>2]|0; //@line 62 "modules/joystick/sdl/JoystickModule.cpp"
    __THREW__ = 0;
    invoke_vi($45|0,($42|0)); //@line 62 "modules/joystick/sdl/JoystickModule.cpp"
    $46 = __THREW__; __THREW__ = 0;
    $47 = $46&1;
    if ($47) {
     break;
    }
    $48 = HEAP32[$42>>2]|0; //@line 63 "modules/joystick/sdl/JoystickModule.cpp"
    $49 = ((($48)) + 12|0); //@line 63 "modules/joystick/sdl/JoystickModule.cpp"
    $50 = HEAP32[$49>>2]|0; //@line 63 "modules/joystick/sdl/JoystickModule.cpp"
    __THREW__ = 0;
    invoke_vi($50|0,($42|0)); //@line 63 "modules/joystick/sdl/JoystickModule.cpp"
    $51 = __THREW__; __THREW__ = 0;
    $52 = $51&1;
    if ($52) {
     break;
    }
    $53 = $__begin$sroa$0$011; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $54 = ((($53)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $__begin$sroa$0$0 = HEAP32[$54>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $55 = $__begin$sroa$0$0; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $56 = ($55|0)==($0|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($56) {
     label = 2;
     break L1;
    } else {
     $41 = $55;$__begin$sroa$0$011 = $__begin$sroa$0$0;
    }
   }
   $7 = ___cxa_find_matching_catch_3(0|0)|0;
   $8 = tempRet0;
   $lpad$phi$index = $7;
  }
 } while(0);
 if ((label|0) == 2) {
  __THREW__ = 0;
  $4 = (invoke_ii(4186,4096)|0); //@line 66 "modules/joystick/sdl/JoystickModule.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  do {
   if (!($6)) {
    $57 = ($4|0)==(0); //@line 66 "modules/joystick/sdl/JoystickModule.cpp"
    if (!($57)) {
     __THREW__ = 0;
     invoke_vi(3800,4096); //@line 67 "modules/joystick/sdl/JoystickModule.cpp"
     $58 = __THREW__; __THREW__ = 0;
     $59 = $58&1;
     if ($59) {
      break;
     }
    }
    __THREW__ = 0;
    invoke_vi(3800,8704); //@line 69 "modules/joystick/sdl/JoystickModule.cpp"
    $60 = __THREW__; __THREW__ = 0;
    $61 = $60&1;
    if (!($61)) {
     $62 = ((($this)) + 32|0); //@line 783 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
     $63 = ((($this)) + 36|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $64 = HEAP32[$63>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEbEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($62,$64); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $65 = ((($this)) + 28|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $66 = HEAP32[$65>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $67 = ($66|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     if (!($67)) {
      $68 = HEAP32[$1>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $69 = HEAP32[$0>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $70 = ((($69)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $71 = HEAP32[$70>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $72 = HEAP32[$68>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $73 = ((($72)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      HEAP32[$73>>2] = $71; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $74 = $72; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $75 = HEAP32[$70>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      HEAP32[$75>>2] = $74; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      HEAP32[$65>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $76 = ($68|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if (!($76)) {
       $__f$01$i$i = $68;
       while(1) {
        $77 = ((($__f$01$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
        $78 = HEAP32[$77>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
        __ZdlPv($__f$01$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        $79 = ($78|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
        if ($79) {
         break;
        } else {
         $__f$01$i$i = $78;
        }
       }
      }
     }
     $80 = ((($this)) + 8|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $81 = HEAP32[$80>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $82 = ($81|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $83 = $81; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($82) {
      __ZN4love6ModuleD2Ev($this); //@line 70 "modules/joystick/sdl/JoystickModule.cpp"
      return; //@line 70 "modules/joystick/sdl/JoystickModule.cpp"
     }
     $84 = ((($this)) + 12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $85 = HEAP32[$84>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $86 = ($85|0)==($81|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($86)) {
      $scevgep$i$i$i = ((($85)) + -4|0);
      $87 = $scevgep$i$i$i;
      $88 = (($87) - ($83))|0;
      $89 = $88 >>> 2;
      $90 = $89 ^ -1;
      $scevgep4$i$i$i = (($85) + ($90<<2)|0);
      HEAP32[$84>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     __ZdlPv($81); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     __ZN4love6ModuleD2Ev($this); //@line 70 "modules/joystick/sdl/JoystickModule.cpp"
     return; //@line 70 "modules/joystick/sdl/JoystickModule.cpp"
    }
   }
  } while(0);
  $9 = ___cxa_find_matching_catch_3(0|0)|0;
  $10 = tempRet0;
  $lpad$phi$index = $9;
 }
 $11 = ((($this)) + 32|0); //@line 783 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $12 = ((($this)) + 36|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $13 = HEAP32[$12>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEbEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($11,$13); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $14 = ((($this)) + 28|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $15 = HEAP32[$14>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $16 = ($15|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if (!($16)) {
  $17 = HEAP32[$1>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $18 = HEAP32[$0>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $19 = ((($18)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $20 = HEAP32[$19>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $21 = HEAP32[$17>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $22 = ((($21)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$22>>2] = $20; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $23 = $21; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $24 = HEAP32[$19>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$24>>2] = $23; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$14>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $25 = ($17|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if (!($25)) {
   $__f$01$i$i1 = $17;
   while(1) {
    $26 = ((($__f$01$i$i1)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $27 = HEAP32[$26>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    __ZdlPv($__f$01$i$i1); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $28 = ($27|0)==($0|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($28) {
     break;
    } else {
     $__f$01$i$i1 = $27;
    }
   }
  }
 }
 $29 = ((($this)) + 8|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = HEAP32[$29>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = ($30|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = $30; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($31) {
  __ZN4love6ModuleD2Ev($this); //@line 70 "modules/joystick/sdl/JoystickModule.cpp"
  ___clang_call_terminate($lpad$phi$index); //@line 70 "modules/joystick/sdl/JoystickModule.cpp"
  // unreachable; //@line 70 "modules/joystick/sdl/JoystickModule.cpp"
 }
 $33 = ((($this)) + 12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $34 = HEAP32[$33>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $35 = ($34|0)==($30|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($35)) {
  $scevgep$i$i$i4 = ((($34)) + -4|0);
  $36 = $scevgep$i$i$i4;
  $37 = (($36) - ($32))|0;
  $38 = $37 >>> 2;
  $39 = $38 ^ -1;
  $scevgep4$i$i$i5 = (($34) + ($39<<2)|0);
  HEAP32[$33>>2] = $scevgep4$i$i$i5; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($30); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 __ZN4love6ModuleD2Ev($this); //@line 70 "modules/joystick/sdl/JoystickModule.cpp"
 ___clang_call_terminate($lpad$phi$index); //@line 70 "modules/joystick/sdl/JoystickModule.cpp"
 // unreachable; //@line 70 "modules/joystick/sdl/JoystickModule.cpp"
}
function __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEbEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($this,$__nd) {
 $this = $this|0;
 $__nd = $__nd|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__nd|0)==(0|0); //@line 1440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($0) {
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 } else {
  $1 = HEAP32[$__nd>>2]|0; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEbEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($this,$1); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2 = ((($__nd)) + 4|0); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3 = HEAP32[$2>>2]|0; //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEbEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($this,$3); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4 = ((($__nd)) + 16|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
  __ZdlPv($__nd); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
}
function __ZN4love8joystick3sdl14JoystickModuleD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love8joystick3sdl14JoystickModuleD2Ev($this); //@line 58 "modules/joystick/sdl/JoystickModule.cpp"
 __ZdlPv($this); //@line 58 "modules/joystick/sdl/JoystickModule.cpp"
 return; //@line 70 "modules/joystick/sdl/JoystickModule.cpp"
}
function __ZNK4love8joystick14JoystickModule13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 6; //@line 40 "./modules/joystick/JoystickModule.h"
}
function __ZNK4love8joystick3sdl14JoystickModule7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (441795|0); //@line 74 "modules/joystick/sdl/JoystickModule.cpp"
}
function __ZN4love8joystick3sdl14JoystickModule11addJoystickEi($this,$deviceindex) {
 $this = $this|0;
 $deviceindex = $deviceindex|0;
 var $$04 = 0, $$08 = 0, $$3 = 0, $$4 = 0, $$lcssa70 = 0, $$lcssa72 = 0, $$mask = 0, $$pre = 0, $$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$032 = 0, $__begin$sroa$0$033 = 0, $__begin2$sroa$0$030 = 0, $__lhs_sz$03$i61 = 0, $__lp$04$i60 = 0;
 var $__parent$i = 0, $__r$0$i = 0, $__rp$05$i59 = 0, $guid$i = 0, $guid$i$byval_copy = 0, $guidstr = 0, $guidstr$i = 0, $joystick = 0, $lpad$phi$index = 0, $lpad$phi$index3 = 0, $reused$2$off042 = 0, $reused$2$off043 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $guid$i$byval_copy = sp + 104|0;
 $__parent$i = sp + 40|0;
 $guidstr$i = sp + 64|0;
 $guid$i = sp + 48|0;
 $guidstr = sp + 28|0;
 $joystick = sp + 24|0;
 $0 = sp + 12|0;
 $1 = sp;
 $2 = ($deviceindex|0)<(0); //@line 115 "modules/joystick/sdl/JoystickModule.cpp"
 if ($2) {
  $$4 = 0;
  STACKTOP = sp;return ($$4|0); //@line 169 "modules/joystick/sdl/JoystickModule.cpp"
 }
 $3 = (_SDL_NumJoysticks()|0); //@line 115 "modules/joystick/sdl/JoystickModule.cpp"
 $4 = ($3|0)>($deviceindex|0); //@line 115 "modules/joystick/sdl/JoystickModule.cpp"
 if (!($4)) {
  $$4 = 0;
  STACKTOP = sp;return ($$4|0); //@line 169 "modules/joystick/sdl/JoystickModule.cpp"
 }
 $5 = (_SDL_NumJoysticks()|0); //@line 463 "modules/joystick/sdl/JoystickModule.cpp"
 $6 = ($5|0)>($deviceindex|0); //@line 463 "modules/joystick/sdl/JoystickModule.cpp"
 if ($6) {
  dest=$guidstr$i; stop=dest+33|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 467 "modules/joystick/sdl/JoystickModule.cpp"
  _SDL_JoystickGetDeviceGUID($guid$i,$deviceindex); //@line 470 "modules/joystick/sdl/JoystickModule.cpp"
  dest=$guid$i$byval_copy; src=$guid$i; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 471 "modules/joystick/sdl/JoystickModule.cpp"
  _SDL_JoystickGetGUIDString($guid$i$byval_copy,$guidstr$i,33); //@line 471 "modules/joystick/sdl/JoystickModule.cpp"
  $7 = (_strlen($guidstr$i)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($guidstr,$guidstr$i,$7); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($guidstr,1282466,0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 HEAP32[$joystick>>2] = 0; //@line 119 "modules/joystick/sdl/JoystickModule.cpp"
 $8 = ((($this)) + 20|0); //@line 122 "modules/joystick/sdl/JoystickModule.cpp"
 $9 = ((($this)) + 24|0); //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $__begin$sroa$0$032 = HEAP32[$9>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $10 = $__begin$sroa$0$032; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $11 = ($10|0)==($8|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 L11: do {
  if ($11) {
   $reused$2$off043 = 0;
   label = 26;
  } else {
   $12 = ((($0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $13 = ((($guidstr)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $14 = ((($guidstr)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $15 = ((($guidstr)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $16 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $17 = ((($0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $25 = $10;$__begin$sroa$0$033 = $__begin$sroa$0$032;
   L13: while(1) {
    $24 = ((($25)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $26 = HEAP32[$24>>2]|0; //@line 122 "modules/joystick/sdl/JoystickModule.cpp"
    $27 = HEAP32[$26>>2]|0; //@line 125 "modules/joystick/sdl/JoystickModule.cpp"
    $28 = ((($27)) + 24|0); //@line 125 "modules/joystick/sdl/JoystickModule.cpp"
    $29 = HEAP32[$28>>2]|0; //@line 125 "modules/joystick/sdl/JoystickModule.cpp"
    __THREW__ = 0;
    $30 = (invoke_ii($29|0,($26|0))|0); //@line 125 "modules/joystick/sdl/JoystickModule.cpp"
    $31 = __THREW__; __THREW__ = 0;
    $32 = $31&1;
    if ($32) {
     label = 9;
     break;
    }
    do {
     if (!($30)) {
      $33 = HEAP32[$26>>2]|0; //@line 125 "modules/joystick/sdl/JoystickModule.cpp"
      $34 = ((($33)) + 80|0); //@line 125 "modules/joystick/sdl/JoystickModule.cpp"
      $35 = HEAP32[$34>>2]|0; //@line 125 "modules/joystick/sdl/JoystickModule.cpp"
      __THREW__ = 0;
      invoke_vii($35|0,($0|0),($26|0)); //@line 125 "modules/joystick/sdl/JoystickModule.cpp"
      $36 = __THREW__; __THREW__ = 0;
      $37 = $36&1;
      if ($37) {
       label = 9;
       break L13;
      }
      $38 = HEAP16[$0>>1]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $39 = $38 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $40 = ($39<<16>>16)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $41 = HEAP32[$12>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $$mask = $38 & 254; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $42 = $$mask&65535; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $43 = $42 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $44 = $40 ? $43 : $41; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $45 = HEAP8[$guidstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $46 = $45 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $47 = ($46<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $48 = HEAP32[$13>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $49 = $45&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $50 = $49 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $51 = $47 ? $50 : $48; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $52 = ($44|0)==($51|0); //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $53 = ($38&65535) >>> 8; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $54 = $53&255; //@line 3760 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($52) {
       $55 = HEAP32[$14>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $56 = $47 ? $15 : $55; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if (!($40)) {
        $66 = HEAP32[$17>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $67 = (_memcmp($66,$56,$44)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $68 = ($67|0)==(0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 125 "modules/joystick/sdl/JoystickModule.cpp"
        if ($68) {
         $$lcssa72 = $26;
         label = 23;
         break L13;
        } else {
         break;
        }
       }
       $57 = ($44|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($57) {
        $74 = $26;
        label = 25;
        break L13;
       }
       $58 = HEAP8[$56>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $59 = ($54<<24>>24)==($58<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($59) {
        $__lhs_sz$03$i61 = $43;$__lp$04$i60 = $16;$__rp$05$i59 = $56;
        while(1) {
         $60 = (($__lhs_sz$03$i61) + -1)|0; //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $61 = ((($__lp$04$i60)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $62 = ($60|0)==(0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         if ($62) {
          $74 = $26;
          label = 25;
          break L13;
         }
         $63 = ((($__rp$05$i59)) + 1|0); //@line 3766 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $$pre = HEAP8[$61>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $64 = HEAP8[$63>>0]|0; //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $65 = ($$pre<<24>>24)==($64<<24>>24); //@line 3767 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         if ($65) {
          $__lhs_sz$03$i61 = $60;$__lp$04$i60 = $61;$__rp$05$i59 = $63;
         } else {
          break;
         }
        }
       }
      }
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 125 "modules/joystick/sdl/JoystickModule.cpp"
     }
    } while(0);
    $70 = $__begin$sroa$0$033; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $71 = ((($70)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $__begin$sroa$0$0 = HEAP32[$71>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $72 = $__begin$sroa$0$0; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $73 = ($72|0)==($8|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($73) {
     $reused$2$off043 = 0;
     label = 26;
     break L11;
    } else {
     $25 = $72;$__begin$sroa$0$033 = $__begin$sroa$0$0;
    }
   }
   if ((label|0) == 9) {
    $20 = ___cxa_find_matching_catch_2()|0;
    $21 = tempRet0;
    $lpad$phi$index = $20;$lpad$phi$index3 = $21;
    break;
   }
   else if ((label|0) == 23) {
    HEAP32[$joystick>>2] = $$lcssa72; //@line 127 "modules/joystick/sdl/JoystickModule.cpp"
    $69 = $$lcssa72;
    $135 = $$lcssa72;$180 = $69;$99 = $$lcssa72;$reused$2$off042 = 1;
    label = 31;
    break;
   }
   else if ((label|0) == 25) {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 125 "modules/joystick/sdl/JoystickModule.cpp"
    HEAP32[$joystick>>2] = $74; //@line 127 "modules/joystick/sdl/JoystickModule.cpp"
    $75 = $74;
    $76 = ($74|0)==(0|0); //@line 133 "modules/joystick/sdl/JoystickModule.cpp"
    if ($76) {
     $reused$2$off043 = 1;
     label = 26;
     break;
    } else {
     $135 = $74;$180 = $75;$99 = $74;$reused$2$off042 = 1;
     label = 31;
     break;
    }
   }
  }
 } while(0);
 do {
  if ((label|0) == 26) {
   __THREW__ = 0;
   $77 = (invoke_ii(3348,144)|0); //@line 135 "modules/joystick/sdl/JoystickModule.cpp"
   $78 = __THREW__; __THREW__ = 0;
   $79 = $78&1;
   if ($79) {
    label = 10;
   } else {
    $80 = ((($this)) + 28|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $81 = HEAP32[$80>>2]|0; //@line 893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    __THREW__ = 0;
    invoke_vii(4187,($77|0),($81|0)); //@line 135 "modules/joystick/sdl/JoystickModule.cpp"
    $82 = __THREW__; __THREW__ = 0;
    $83 = $82&1;
    if ($83) {
     $94 = ___cxa_find_matching_catch_2()|0;
     $95 = tempRet0;
     __ZdlPv($77); //@line 135 "modules/joystick/sdl/JoystickModule.cpp"
     $$04 = $94;$$08 = $95;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($guidstr); //@line 169 "modules/joystick/sdl/JoystickModule.cpp"
     ___resumeException($$04|0);
     // unreachable;
    } else {
     HEAP32[$joystick>>2] = $77; //@line 135 "modules/joystick/sdl/JoystickModule.cpp"
     $84 = $77; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     __THREW__ = 0;
     $85 = (invoke_ii(3348,12)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $86 = __THREW__; __THREW__ = 0;
     $87 = $86&1;
     if ($87) {
      label = 10;
      break;
     }
     $88 = ((($85)) + 8|0); //@line 1537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$88>>2] = $84; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $89 = ((($85)) + 4|0); //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$89>>2] = $8; //@line 1097 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $90 = HEAP32[$8>>2]|0; //@line 1098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$85>>2] = $90; //@line 1098 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $91 = ((($90)) + 4|0); //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$91>>2] = $85; //@line 1099 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$8>>2] = $85; //@line 1100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $92 = HEAP32[$80>>2]|0; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $93 = (($92) + 1)|0; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$80>>2] = $93; //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $135 = $77;$180 = $84;$99 = $77;$reused$2$off042 = $reused$2$off043;
     label = 31;
     break;
    }
   }
  }
 } while(0);
 L42: do {
  if ((label|0) == 31) {
   $96 = HEAP32[$this>>2]|0; //@line 140 "modules/joystick/sdl/JoystickModule.cpp"
   $97 = ((($96)) + 28|0); //@line 140 "modules/joystick/sdl/JoystickModule.cpp"
   $98 = HEAP32[$97>>2]|0; //@line 140 "modules/joystick/sdl/JoystickModule.cpp"
   __THREW__ = 0;
   invoke_vii($98|0,($this|0),($99|0)); //@line 140 "modules/joystick/sdl/JoystickModule.cpp"
   $100 = __THREW__; __THREW__ = 0;
   $101 = $100&1;
   if ($101) {
    label = 10;
   } else {
    $102 = HEAP32[$99>>2]|0; //@line 142 "modules/joystick/sdl/JoystickModule.cpp"
    $103 = ((($102)) + 16|0); //@line 142 "modules/joystick/sdl/JoystickModule.cpp"
    $104 = HEAP32[$103>>2]|0; //@line 142 "modules/joystick/sdl/JoystickModule.cpp"
    __THREW__ = 0;
    $105 = (invoke_iii($104|0,($99|0),($deviceindex|0))|0); //@line 142 "modules/joystick/sdl/JoystickModule.cpp"
    $106 = __THREW__; __THREW__ = 0;
    $107 = $106&1;
    if ($107) {
     label = 10;
    } else {
     L46: do {
      if ($105) {
       $108 = ((($this)) + 8|0); //@line 147 "modules/joystick/sdl/JoystickModule.cpp"
       $109 = HEAP32[$108>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $110 = ((($this)) + 12|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $111 = HEAP32[$110>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $112 = ($109|0)==($111|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
       L48: do {
        if (!($112)) {
         $__begin2$sroa$0$030 = $109;
         while(1) {
          $115 = HEAP32[$__begin2$sroa$0$030>>2]|0; //@line 147 "modules/joystick/sdl/JoystickModule.cpp"
          $116 = HEAP32[$99>>2]|0; //@line 149 "modules/joystick/sdl/JoystickModule.cpp"
          $117 = ((($116)) + 76|0); //@line 149 "modules/joystick/sdl/JoystickModule.cpp"
          $118 = HEAP32[$117>>2]|0; //@line 149 "modules/joystick/sdl/JoystickModule.cpp"
          __THREW__ = 0;
          $119 = (invoke_ii($118|0,($99|0))|0); //@line 149 "modules/joystick/sdl/JoystickModule.cpp"
          $120 = __THREW__; __THREW__ = 0;
          $121 = $120&1;
          if ($121) {
           label = 8;
           break;
          }
          $122 = HEAP32[$115>>2]|0; //@line 149 "modules/joystick/sdl/JoystickModule.cpp"
          $123 = ((($122)) + 76|0); //@line 149 "modules/joystick/sdl/JoystickModule.cpp"
          $124 = HEAP32[$123>>2]|0; //@line 149 "modules/joystick/sdl/JoystickModule.cpp"
          __THREW__ = 0;
          $125 = (invoke_ii($124|0,($115|0))|0); //@line 149 "modules/joystick/sdl/JoystickModule.cpp"
          $126 = __THREW__; __THREW__ = 0;
          $127 = $126&1;
          if ($127) {
           label = 8;
           break;
          }
          $128 = ($119|0)==($125|0); //@line 149 "modules/joystick/sdl/JoystickModule.cpp"
          $114 = ((($__begin2$sroa$0$030)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
          if ($128) {
           $$lcssa70 = $115;
           break;
          }
          $113 = ($114|0)==($111|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
          if ($113) {
           break L48;
          } else {
           $__begin2$sroa$0$030 = $114;
          }
         }
         if ((label|0) == 8) {
          $18 = ___cxa_find_matching_catch_2()|0;
          $19 = tempRet0;
          $lpad$phi$index = $18;$lpad$phi$index3 = $19;
          break L42;
         }
         $129 = HEAP32[$99>>2]|0; //@line 151 "modules/joystick/sdl/JoystickModule.cpp"
         $130 = ((($129)) + 20|0); //@line 151 "modules/joystick/sdl/JoystickModule.cpp"
         $131 = HEAP32[$130>>2]|0; //@line 151 "modules/joystick/sdl/JoystickModule.cpp"
         __THREW__ = 0;
         invoke_vi($131|0,($99|0)); //@line 151 "modules/joystick/sdl/JoystickModule.cpp"
         $132 = __THREW__; __THREW__ = 0;
         $133 = $132&1;
         if ($133) {
          label = 10;
          break L42;
         }
         if ($reused$2$off042) {
          $$3 = $$lcssa70;
          break L46;
         }
         __ZNSt3__14listIPN4love8joystick8JoystickENS_9allocatorIS4_EEE6removeERKS4_($8,$joystick); //@line 156 "modules/joystick/sdl/JoystickModule.cpp"
         $134 = HEAP32[$135>>2]|0; //@line 157 "modules/joystick/sdl/JoystickModule.cpp"
         $136 = ((($134)) + 12|0); //@line 157 "modules/joystick/sdl/JoystickModule.cpp"
         $137 = HEAP32[$136>>2]|0; //@line 157 "modules/joystick/sdl/JoystickModule.cpp"
         __THREW__ = 0;
         invoke_vi($137|0,($135|0)); //@line 157 "modules/joystick/sdl/JoystickModule.cpp"
         $138 = __THREW__; __THREW__ = 0;
         $139 = $138&1;
         if ($139) {
          label = 10;
          break L42;
         } else {
          $$3 = $$lcssa70;
          break L46;
         }
        }
       } while(0);
       $140 = HEAP32[$99>>2]|0; //@line 164 "modules/joystick/sdl/JoystickModule.cpp"
       $141 = ((($140)) + 64|0); //@line 164 "modules/joystick/sdl/JoystickModule.cpp"
       $142 = HEAP32[$141>>2]|0; //@line 164 "modules/joystick/sdl/JoystickModule.cpp"
       __THREW__ = 0;
       $143 = (invoke_ii($142|0,($99|0))|0); //@line 164 "modules/joystick/sdl/JoystickModule.cpp"
       $144 = __THREW__; __THREW__ = 0;
       $145 = $144&1;
       if ($145) {
        label = 10;
        break L42;
       }
       if ($143) {
        $146 = HEAP32[$99>>2]|0; //@line 165 "modules/joystick/sdl/JoystickModule.cpp"
        $147 = ((($146)) + 80|0); //@line 165 "modules/joystick/sdl/JoystickModule.cpp"
        $148 = HEAP32[$147>>2]|0; //@line 165 "modules/joystick/sdl/JoystickModule.cpp"
        __THREW__ = 0;
        invoke_vii($148|0,($1|0),($99|0)); //@line 165 "modules/joystick/sdl/JoystickModule.cpp"
        $149 = __THREW__; __THREW__ = 0;
        $150 = $149&1;
        if ($150) {
         label = 10;
         break L42;
        }
        $151 = ((($this)) + 32|0); //@line 165 "modules/joystick/sdl/JoystickModule.cpp"
        $152 = (__ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEbNS_4lessIS6_EENS4_INS_4pairIKS6_bEEEEE16__find_equal_keyERPNS_16__tree_node_baseIPvEERSA_($151,$__parent$i,$1)|0); //@line 1394 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
        $153 = HEAP32[$152>>2]|0; //@line 1395 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
        $154 = ($153|0)==(0|0); //@line 1396 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
        if ($154) {
         __THREW__ = 0;
         $155 = (invoke_ii(3348,32)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
         $156 = __THREW__; __THREW__ = 0;
         $157 = $156&1;
         if ($157) {
          $174 = ___cxa_find_matching_catch_2()|0;
          $175 = tempRet0;
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 165 "modules/joystick/sdl/JoystickModule.cpp"
          $$04 = $174;$$08 = $175;
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($guidstr); //@line 169 "modules/joystick/sdl/JoystickModule.cpp"
          ___resumeException($$04|0);
          // unreachable;
         }
         $158 = ((($155)) + 16|0); //@line 1330 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
         ;HEAP32[$158>>2]=HEAP32[$1>>2]|0;HEAP32[$158+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$158+8>>2]=HEAP32[$1+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $159 = ((($155)) + 28|0); //@line 1332 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
         HEAP8[$159>>0] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $160 = HEAP32[$__parent$i>>2]|0; //@line 1399 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
         HEAP32[$155>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
         $161 = ((($155)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
         HEAP32[$161>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
         $162 = ((($155)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
         HEAP32[$162>>2] = $160; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
         HEAP32[$152>>2] = $155; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
         $163 = HEAP32[$151>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
         $164 = HEAP32[$163>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
         $165 = ($164|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
         if ($165) {
          $169 = $155;
         } else {
          $166 = $164; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
          HEAP32[$151>>2] = $166; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
          $$pre$i$i = HEAP32[$152>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
          $169 = $$pre$i$i;
         }
         $167 = ((($this)) + 36|0); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
         $168 = HEAP32[$167>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
         __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($168,$169); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
         $170 = ((($this)) + 40|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $171 = HEAP32[$170>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
         $172 = (($171) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
         HEAP32[$170>>2] = $172; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
         $__r$0$i = $155;
        } else {
         $__r$0$i = $153;
        }
        $173 = ((($__r$0$i)) + 28|0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
        HEAP8[$173>>0] = 1; //@line 165 "modules/joystick/sdl/JoystickModule.cpp"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 165 "modules/joystick/sdl/JoystickModule.cpp"
       }
       $176 = HEAP32[$110>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $177 = ((($this)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $178 = HEAP32[$177>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $179 = ($176|0)==($178|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($179) {
        __THREW__ = 0;
        invoke_vii(4188,($108|0),($joystick|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $183 = __THREW__; __THREW__ = 0;
        $184 = $183&1;
        if ($184) {
         label = 10;
         break L42;
        } else {
         $$3 = $99;
         break;
        }
       } else {
        HEAP32[$176>>2] = $180; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $181 = HEAP32[$110>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $182 = ((($181)) + 4|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$110>>2] = $182; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $$3 = $99;
        break;
       }
      } else {
       $$3 = 0;
      }
     } while(0);
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($guidstr); //@line 169 "modules/joystick/sdl/JoystickModule.cpp"
     $$4 = $$3;
     STACKTOP = sp;return ($$4|0); //@line 169 "modules/joystick/sdl/JoystickModule.cpp"
    }
   }
  }
 } while(0);
 if ((label|0) == 10) {
  $22 = ___cxa_find_matching_catch_2()|0;
  $23 = tempRet0;
  $lpad$phi$index = $22;$lpad$phi$index3 = $23;
 }
 $$04 = $lpad$phi$index;$$08 = $lpad$phi$index3;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($guidstr); //@line 169 "modules/joystick/sdl/JoystickModule.cpp"
 ___resumeException($$04|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__14listIPN4love8joystick8JoystickENS_9allocatorIS4_EEE6removeERKS4_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$cast$i$i$i = 0, $$lcssa = 0, $$lcssa13 = 0, $$lcssa13$us = 0, $$lcssa14 = 0, $$lcssa14$us = 0, $$lcssa37 = 0, $$lcssa38 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__deleted_nodes = 0;
 var $__f$01$i$i = 0, $__i$sroa$0$0$be = 0, $__i$sroa$0$0$be$us = 0, $__i$sroa$0$023 = 0, $__i$sroa$0$023$us = 0, $__j$sroa$0$0$lcssa = 0, $__j$sroa$0$0$lcssa$us = 0, $__j$sroa$0$016 = 0, $__j$sroa$0$016$us = 0, $__r$01$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__deleted_nodes = sp;
 HEAP32[$__deleted_nodes>>2] = $__deleted_nodes; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $0 = ((($__deleted_nodes)) + 4|0); //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$0>>2] = $__deleted_nodes; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $1 = ((($__deleted_nodes)) + 8|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$1>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($this)) + 4|0); //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $3 = HEAP32[$2>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $4 = $3; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $5 = ($4|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($5) {
  STACKTOP = sp;return; //@line 2073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 }
 $6 = ($__deleted_nodes|0)==($this|0); //@line 2017 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $7 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if ($6) {
  $9 = $4;$__i$sroa$0$023$us = $3;
  while(1) {
   $8 = ((($9)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $10 = HEAP32[$8>>2]|0; //@line 2060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $11 = HEAP32[$__x>>2]|0; //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $12 = ($10|0)==($11|0); //@line 2060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $13 = $__i$sroa$0$023$us; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $14 = ((($13)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $15 = HEAP32[$14>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($12) {
    $16 = $15; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $17 = ($16|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    L9: do {
     if ($17) {
      $$lcssa13$us = $this;$$lcssa14$us = 0;$__j$sroa$0$0$lcssa$us = $15;
     } else {
      $19 = $16;$__j$sroa$0$016$us = $15;
      while(1) {
       $18 = ((($19)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $20 = HEAP32[$18>>2]|0; //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $21 = ($20|0)==($10|0); //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       if (!($21)) {
        $$lcssa13$us = $19;$$lcssa14$us = 1;$__j$sroa$0$0$lcssa$us = $__j$sroa$0$016$us;
        break L9;
       }
       $39 = $__j$sroa$0$016$us; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $40 = ((($39)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $41 = HEAP32[$40>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $42 = $41; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       $43 = ($42|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
       if ($43) {
        $$lcssa13$us = $this;$$lcssa14$us = 0;$__j$sroa$0$0$lcssa$us = $41;
        break;
       } else {
        $19 = $42;$__j$sroa$0$016$us = $41;
       }
      }
     }
    } while(0);
    $22 = ($9|0)==($$lcssa13$us|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if (!($22)) {
     $23 = $__j$sroa$0$0$lcssa$us; //@line 487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $24 = HEAP32[$23>>2]|0; //@line 487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $25 = ((($24)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $26 = HEAP32[$25>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $27 = $__i$sroa$0$023$us; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $28 = HEAP32[$27>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $29 = ((($28)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$29>>2] = $26; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $30 = $28; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $31 = HEAP32[$25>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$31>>2] = $30; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $32 = HEAP32[$__deleted_nodes>>2]|0; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $33 = ((($32)) + 4|0); //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$33>>2] = $9; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $34 = $32; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $35 = $__i$sroa$0$023$us; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$35>>2] = $34; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$__deleted_nodes>>2] = $24; //@line 1075 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     HEAP32[$25>>2] = $this; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    }
    if ($$lcssa14$us) {
     $36 = $__j$sroa$0$0$lcssa$us; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $37 = ((($36)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $38 = HEAP32[$37>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $__i$sroa$0$0$be$us = $38;
    } else {
     $__i$sroa$0$0$be$us = $__j$sroa$0$0$lcssa$us;
    }
   } else {
    $__i$sroa$0$0$be$us = $15;
   }
   $44 = $__i$sroa$0$0$be$us; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $45 = ($44|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($45) {
    break;
   } else {
    $9 = $44;$__i$sroa$0$023$us = $__i$sroa$0$0$be$us;
   }
  }
  STACKTOP = sp;return; //@line 2073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 }
 $106 = $__deleted_nodes;$59 = $4;$87 = 0;$__i$sroa$0$023 = $3;
 while(1) {
  $58 = ((($59)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $60 = HEAP32[$58>>2]|0; //@line 2060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $61 = HEAP32[$__x>>2]|0; //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $62 = ($60|0)==($61|0); //@line 2060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $63 = $__i$sroa$0$023; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $64 = ((($63)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $65 = HEAP32[$64>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($62) {
   $66 = $65; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   $67 = ($66|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   L25: do {
    if ($67) {
     $$lcssa13 = $this;$$lcssa14 = 0;$__j$sroa$0$0$lcssa = $65;
    } else {
     $69 = $66;$__j$sroa$0$016 = $65;
     while(1) {
      $68 = ((($69)) + 8|0); //@line 455 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $70 = HEAP32[$68>>2]|0; //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $71 = ($70|0)==($60|0); //@line 2063 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if (!($71)) {
       $$lcssa13 = $69;$$lcssa14 = 1;$__j$sroa$0$0$lcssa = $__j$sroa$0$016;
       break L25;
      }
      $72 = $__j$sroa$0$016; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $73 = ((($72)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $74 = HEAP32[$73>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $75 = $74; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      $76 = ($75|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
      if ($76) {
       $$lcssa13 = $this;$$lcssa14 = 0;$__j$sroa$0$0$lcssa = $74;
       break;
      } else {
       $69 = $75;$__j$sroa$0$016 = $74;
      }
     }
    }
   } while(0);
   $77 = ($59|0)==($$lcssa13|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
   if ($77) {
    $109 = $106;$110 = $87;
   } else {
    $80 = $59;$__r$01$i$i$i = 0;
    while(1) {
     $78 = (($__r$01$i$i$i) + 1)|0; //@line 485 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $79 = ((($80)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $81 = HEAP32[$79>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $$cast$i$i$i = $81; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $82 = ($$cast$i$i$i|0)==($$lcssa13|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
     $83 = $81; //@line 484 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($82) {
      $$lcssa = $78;
      break;
     } else {
      $80 = $83;$__r$01$i$i$i = $78;
     }
    }
    $84 = HEAP32[$7>>2]|0; //@line 2020 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $85 = (($84) - ($$lcssa))|0; //@line 2020 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$7>>2] = $85; //@line 2020 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $86 = (($87) + ($$lcssa))|0; //@line 2021 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$1>>2] = $86; //@line 2021 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $88 = $__j$sroa$0$0$lcssa; //@line 487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $89 = HEAP32[$88>>2]|0; //@line 487 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $90 = ((($89)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $91 = HEAP32[$90>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $92 = $__i$sroa$0$023; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $93 = HEAP32[$92>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $94 = ((($93)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$94>>2] = $91; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $95 = $93; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $96 = HEAP32[$90>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$96>>2] = $95; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $97 = HEAP32[$__deleted_nodes>>2]|0; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $98 = ((($97)) + 4|0); //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$98>>2] = $59; //@line 1073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $99 = $97; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $100 = $__i$sroa$0$023; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$100>>2] = $99; //@line 1074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$__deleted_nodes>>2] = $89; //@line 1075 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    HEAP32[$90>>2] = $__deleted_nodes; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $109 = $89;$110 = $86;
   }
   if ($$lcssa14) {
    $101 = $__j$sroa$0$0$lcssa; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $102 = ((($101)) + 4|0); //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $103 = HEAP32[$102>>2]|0; //@line 474 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $107 = $109;$108 = $110;$__i$sroa$0$0$be = $103;
   } else {
    $107 = $109;$108 = $110;$__i$sroa$0$0$be = $__j$sroa$0$0$lcssa;
   }
  } else {
   $107 = $106;$108 = $87;$__i$sroa$0$0$be = $65;
  }
  $104 = $__i$sroa$0$0$be; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $105 = ($104|0)==($this|0); //@line 500 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($105) {
   $$lcssa37 = $107;$$lcssa38 = $108;
   break;
  } else {
   $106 = $107;$59 = $104;$87 = $108;$__i$sroa$0$023 = $__i$sroa$0$0$be;
  }
 }
 $46 = ($$lcssa38|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($46) {
  STACKTOP = sp;return; //@line 2073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 }
 $47 = HEAP32[$0>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $48 = ((($$lcssa37)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $49 = HEAP32[$48>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $50 = HEAP32[$47>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $51 = ((($50)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$51>>2] = $49; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $52 = $50; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $53 = HEAP32[$48>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$53>>2] = $52; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$1>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $54 = ($47|0)==($__deleted_nodes|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if ($54) {
  STACKTOP = sp;return; //@line 2073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 } else {
  $__f$01$i$i = $47;
 }
 while(1) {
  $55 = ((($__f$01$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $56 = HEAP32[$55>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  __ZdlPv($__f$01$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $57 = ($56|0)==($__deleted_nodes|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if ($57) {
   break;
  } else {
   $__f$01$i$i = $56;
  }
 }
 STACKTOP = sp;return; //@line 2073 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
}
function __ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEbNS_4lessIS6_EENS4_INS_4pairIKS6_bEEEEE16__find_equal_keyERPNS_16__tree_node_baseIPvEERSA_($this,$__parent,$__k) {
 $this = $this|0;
 $__parent = $__parent|0;
 $__k = $__k|0;
 var $$$$i$i$i = 0, $$$$i$i$i1 = 0, $$0 = 0, $$lcssa = 0, $$lcssa19 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__nd$0 = 0, $__nd$0$lcssa = 0, $__nd$0$lcssa17 = 0, $__nd$0$lcssa18 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $2 = ($1|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 if ($2) {
  HEAP32[$__parent>>2] = $0; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $0;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 $3 = HEAP8[$__k>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = $3 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ($4<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = ((($__k)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = HEAP32[$6>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $3&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $8 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $5 ? $9 : $7; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = ((($__k)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = HEAP32[$11>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ((($__k)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = $5 ? $13 : $12; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $__nd$0 = $1;
 while(1) {
  $15 = ((($__nd$0)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $16 = HEAP8[$15>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $17 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $18 = ($17<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $19 = ((($__nd$0)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $20 = HEAP32[$19>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $21 = $16&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $22 = $21 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $23 = $18 ? $22 : $20; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $24 = ((($__nd$0)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $25 = HEAP32[$24>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $26 = ((($15)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $27 = $18 ? $26 : $25; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $28 = ($23>>>0)<($10>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $29 = $28 ? $23 : $10; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $30 = (_memcmp($14,$27,$29)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $31 = ($30|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $32 = ($10>>>0)<($23>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $$$$i$i$i1 = $32 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $33 = $31 ? $$$$i$i$i1 : $30; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $34 = ($33|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($34) {
   $35 = HEAP32[$__nd$0>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $36 = ($35|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   if ($36) {
    $$lcssa19 = $__nd$0;$__nd$0$lcssa18 = $__nd$0;
    label = 5;
    break;
   } else {
    $$sink = $35;
   }
  } else {
   $37 = $32 ? $10 : $23; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $38 = (_memcmp($27,$14,$37)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $39 = ($38|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$$$i$i$i = $28 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $40 = $39 ? $$$$i$i$i : $38; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $41 = ($40|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if (!($41)) {
    $__nd$0$lcssa = $__nd$0;
    label = 10;
    break;
   }
   $42 = ((($__nd$0)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $43 = HEAP32[$42>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $44 = ($43|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   if ($44) {
    $$lcssa = $42;$__nd$0$lcssa17 = $__nd$0;
    label = 9;
    break;
   } else {
    $$sink = $43;
   }
  }
  $__nd$0 = $$sink;
 }
 if ((label|0) == 5) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa18; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $$lcssa19;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 else if ((label|0) == 9) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa17; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $$lcssa;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 else if ((label|0) == 10) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $__parent;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 return (0)|0;
}
function __ZNSt3__16vectorIPN4love8joystick8JoystickENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8joystick3sdl14JoystickModule14removeJoystickEPNS0_8JoystickE($this,$joystick) {
 $this = $this|0;
 $joystick = $joystick|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__new_last2$i$i$i = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($joystick|0)==(0|0); //@line 173 "modules/joystick/sdl/JoystickModule.cpp"
 if ($0) {
  return; //@line 183 "modules/joystick/sdl/JoystickModule.cpp"
 }
 $1 = ((($this)) + 8|0); //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($this)) + 12|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$3>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($2|0)==($4|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 L4: do {
  if ($5) {
   $12 = $2;
  } else {
   $7 = $2;
   while(1) {
    $6 = HEAP32[$7>>2]|0; //@line 865 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $8 = ($6|0)==($joystick|0); //@line 865 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    if ($8) {
     $12 = $7;
     break L4;
    }
    $9 = ((($7)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $10 = ($9|0)==($4|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($10) {
     break;
    } else {
     $7 = $9;
    }
   }
   return; //@line 183 "modules/joystick/sdl/JoystickModule.cpp"
  }
 } while(0);
 $11 = ($12|0)==($4|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($11) {
  return; //@line 183 "modules/joystick/sdl/JoystickModule.cpp"
 }
 $13 = HEAP32[$12>>2]|0; //@line 180 "modules/joystick/sdl/JoystickModule.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 180 "modules/joystick/sdl/JoystickModule.cpp"
 $15 = ((($14)) + 20|0); //@line 180 "modules/joystick/sdl/JoystickModule.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 180 "modules/joystick/sdl/JoystickModule.cpp"
 FUNCTION_TABLE_vi[$16 & 8191]($13); //@line 180 "modules/joystick/sdl/JoystickModule.cpp"
 $17 = ((($12)) + 4|0); //@line 1684 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $18 = HEAP32[$3>>2]|0; //@line 1684 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $19 = $17; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $20 = (($18) - ($19))|0; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $21 = $20 >> 2; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 _memmove(($12|0),($17|0),($20|0))|0; //@line 1893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $22 = (($12) + ($21<<2)|0); //@line 1894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $23 = HEAP32[$3>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $24 = ($23|0)==($22|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($24) {
  return; //@line 183 "modules/joystick/sdl/JoystickModule.cpp"
 }
 $__new_last2$i$i$i = $22; //@line 818 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $scevgep$i$i$i = ((($23)) + -4|0);
 $25 = $scevgep$i$i$i;
 $26 = (($25) - ($__new_last2$i$i$i))|0;
 $27 = $26 >>> 2;
 $28 = $27 ^ -1;
 $scevgep4$i$i$i = (($23) + ($28<<2)|0);
 HEAP32[$3>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 return; //@line 183 "modules/joystick/sdl/JoystickModule.cpp"
}
function __ZN4love8joystick3sdl14JoystickModule17getJoystickFromIDEi($this,$instanceid) {
 $this = $this|0;
 $instanceid = $instanceid|0;
 var $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($this)) + 12|0); //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($1|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($4) {
  $$12 = 0;
  return ($$12|0); //@line 111 "modules/joystick/sdl/JoystickModule.cpp"
 } else {
  $__begin$sroa$0$03 = $1;
 }
 while(1) {
  $7 = HEAP32[$__begin$sroa$0$03>>2]|0; //@line 104 "modules/joystick/sdl/JoystickModule.cpp"
  $8 = HEAP32[$7>>2]|0; //@line 106 "modules/joystick/sdl/JoystickModule.cpp"
  $9 = ((($8)) + 84|0); //@line 106 "modules/joystick/sdl/JoystickModule.cpp"
  $10 = HEAP32[$9>>2]|0; //@line 106 "modules/joystick/sdl/JoystickModule.cpp"
  $11 = (FUNCTION_TABLE_ii[$10 & 8191]($7)|0); //@line 106 "modules/joystick/sdl/JoystickModule.cpp"
  $12 = ($11|0)==($instanceid|0); //@line 106 "modules/joystick/sdl/JoystickModule.cpp"
  $6 = ((($__begin$sroa$0$03)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($12) {
   $$12 = $7;
   label = 4;
   break;
  }
  $5 = ($6|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($5) {
   $$12 = 0;
   label = 4;
   break;
  } else {
   $__begin$sroa$0$03 = $6;
  }
 }
 if ((label|0) == 4) {
  return ($$12|0); //@line 111 "modules/joystick/sdl/JoystickModule.cpp"
 }
 return (0)|0;
}
function __ZN4love8joystick3sdl14JoystickModule11getJoystickEi($this,$joyindex) {
 $this = $this|0;
 $joyindex = $joyindex|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($joyindex|0)<(0); //@line 79 "modules/joystick/sdl/JoystickModule.cpp"
 if ($0) {
  $$0 = 0;
  return ($$0|0); //@line 83 "modules/joystick/sdl/JoystickModule.cpp"
 }
 $1 = ((($this)) + 8|0); //@line 79 "modules/joystick/sdl/JoystickModule.cpp"
 $2 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = $5 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ($6>>>0)>($joyindex>>>0); //@line 79 "modules/joystick/sdl/JoystickModule.cpp"
 if (!($7)) {
  $$0 = 0;
  return ($$0|0); //@line 83 "modules/joystick/sdl/JoystickModule.cpp"
 }
 $8 = $4; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) + ($joyindex<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = HEAP32[$9>>2]|0; //@line 82 "modules/joystick/sdl/JoystickModule.cpp"
 $$0 = $10;
 return ($$0|0); //@line 83 "modules/joystick/sdl/JoystickModule.cpp"
}
function __ZN4love8joystick3sdl14JoystickModule8getIndexEPKNS0_8JoystickE($this,$joystick) {
 $this = $this|0;
 $joystick = $joystick|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 87 "modules/joystick/sdl/JoystickModule.cpp"
 $1 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($2) - ($3))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($4|0)>(0); //@line 87 "modules/joystick/sdl/JoystickModule.cpp"
 $6 = $3; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($5)) {
  $13 = -1;
  return ($13|0); //@line 95 "modules/joystick/sdl/JoystickModule.cpp"
 }
 $7 = $4 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $i$02 = 0;
 while(1) {
  $8 = (($6) + ($i$02<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $9 = HEAP32[$8>>2]|0; //@line 89 "modules/joystick/sdl/JoystickModule.cpp"
  $10 = ($9|0)==($joystick|0); //@line 89 "modules/joystick/sdl/JoystickModule.cpp"
  if ($10) {
   $13 = $i$02;
   label = 5;
   break;
  }
  $11 = (($i$02) + 1)|0; //@line 87 "modules/joystick/sdl/JoystickModule.cpp"
  $12 = ($11|0)<($7|0); //@line 87 "modules/joystick/sdl/JoystickModule.cpp"
  if ($12) {
   $i$02 = $11;
  } else {
   $13 = -1;
   label = 5;
   break;
  }
 }
 if ((label|0) == 5) {
  return ($13|0); //@line 95 "modules/joystick/sdl/JoystickModule.cpp"
 }
 return (0)|0;
}
function __ZNK4love8joystick3sdl14JoystickModule16getJoystickCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 99 "modules/joystick/sdl/JoystickModule.cpp"
 $1 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($2) - ($3))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = $4 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 return ($5|0); //@line 99 "modules/joystick/sdl/JoystickModule.cpp"
}
function __ZN4love8joystick3sdl14JoystickModule17setGamepadMappingERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEENS0_8Joystick12GamepadInputENSC_13JoystickInputE($this,$guid,$gpinput,$joyinput) {
 $this = $this|0;
 $guid = $guid|0;
 $gpinput = $gpinput|0;
 $joyinput = $joyinput|0;
 var $$0 = 0, $$0$i = 0, $$01 = 0, $$01$i = 0, $$016$i$i$i = 0, $$016$i$i$i$lcssa = 0, $$1 = 0, $$1$i$i = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$3 = 0, $$34 = 0, $$4 = 0, $$45 = 0, $$5 = 0, $$56 = 0, $$6 = 0, $$67 = 0, $$7 = 0;
 var $$78 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0;
 var $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0;
 var $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0;
 var $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0;
 var $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0;
 var $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0;
 var $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__m1$0$i$i$i = 0;
 var $__m2$0$i$i$i = 0, $cond = 0, $eh$lpad$body$i$index11Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $eh$lpad$body$index5Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body18$index16Z2D = 0, $eh$lpad$body18$indexZ2D = 0, $endpos$0 = 0, $gpinputname = 0, $insertstr = 0, $joyinputstr = 0, $joyinputstream = 0, $mapstr = 0, $sdlguid = 0, $sdlguid$byval_copy = 0, $sdlhat = 0, $tmpcast$byval_copy = 0, $vararg_buffer = 0, $vararg_buffer1 = 0;
 var dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 352|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tmpcast$byval_copy = sp + 304|0;
 $sdlguid$byval_copy = sp + 336|0;
 $vararg_buffer1 = sp + 16|0;
 $vararg_buffer = sp + 8|0;
 $0 = sp + 288|0;
 $sdlguid = sp + 320|0;
 $mapstr = sp + 276|0;
 $1 = sp + 264|0;
 $joyinputstream = sp + 120|0;
 $sdlhat = sp + 312|0;
 $joyinputstr = sp + 104|0;
 $gpinputname = sp + 92|0;
 $2 = sp;
 $insertstr = sp + 80|0;
 $3 = sp + 68|0;
 $4 = sp + 56|0;
 $5 = sp + 44|0;
 $6 = sp + 32|0;
 $7 = sp + 20|0;
 $8 = HEAP8[$guid>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $8 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = ($9<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = ((($guid)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = HEAP32[$11>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = $8&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = $13 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $15 = $10 ? $14 : $12; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $16 = ($15|0)==(32); //@line 188 "modules/joystick/sdl/JoystickModule.cpp"
 if (!($16)) {
  $17 = (___cxa_allocate_exception(16)|0); //@line 189 "modules/joystick/sdl/JoystickModule.cpp"
  $18 = HEAP8[$guid>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $19 = $18 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $20 = ($19<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $21 = ((($guid)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $22 = HEAP32[$21>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $23 = ((($guid)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $24 = $20 ? $23 : $22; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  __THREW__ = 0;
  HEAP32[$vararg_buffer>>2] = $24; //@line 189 "modules/joystick/sdl/JoystickModule.cpp"
  invoke_viii(3349,($17|0),(441813|0),($vararg_buffer|0)); //@line 189 "modules/joystick/sdl/JoystickModule.cpp"
  $25 = __THREW__; __THREW__ = 0;
  $26 = $25&1;
  if (!($26)) {
   ___cxa_throw(($17|0),(96|0),(344|0)); //@line 189 "modules/joystick/sdl/JoystickModule.cpp"
   // unreachable; //@line 189 "modules/joystick/sdl/JoystickModule.cpp"
  }
  $27 = ___cxa_find_matching_catch_2()|0;
  $28 = tempRet0;
  ___cxa_free_exception(($17|0)); //@line 189 "modules/joystick/sdl/JoystickModule.cpp"
  $$7 = $27;$$78 = $28;
  ___resumeException($$7|0);
  // unreachable;
 }
 $29 = ((($guid)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $30 = HEAP32[$29>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $31 = ((($guid)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $32 = $10 ? $31 : $30; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 _SDL_JoystickGetGUIDFromString($sdlguid,$32); //@line 191 "modules/joystick/sdl/JoystickModule.cpp"
 ;HEAP32[$mapstr>>2]=0|0;HEAP32[$mapstr+4>>2]=0|0;HEAP32[$mapstr+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 dest=$sdlguid$byval_copy; src=$sdlguid; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 194 "modules/joystick/sdl/JoystickModule.cpp"
 $33 = (invoke_ii(4189,($sdlguid$byval_copy|0))|0); //@line 194 "modules/joystick/sdl/JoystickModule.cpp"
 $34 = __THREW__; __THREW__ = 0;
 $35 = $34&1;
 L8: do {
  if ($35) {
   label = 9;
  } else {
   $36 = ($33|0)==(0|0); //@line 195 "modules/joystick/sdl/JoystickModule.cpp"
   do {
    if ($36) {
     ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $43 = HEAP8[$guid>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $44 = $43 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $45 = ($44<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $46 = HEAP32[$11>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $47 = $43&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $48 = $47 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $49 = $45 ? $48 : $46; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $50 = HEAP32[$29>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $51 = $45 ? $31 : $50; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $52 = (($49) + 12)|0; //@line 3980 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     __THREW__ = 0;
     invoke_viiii(3514,($1|0),($51|0),($49|0),($52|0)); //@line 3980 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $53 = __THREW__; __THREW__ = 0;
     $54 = $53&1;
     if (!($54)) {
      __THREW__ = 0;
      (invoke_iiii(3462,($1|0),(441839|0),12)|0); //@line 3981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $55 = __THREW__; __THREW__ = 0;
      $56 = $55&1;
      if (!($56)) {
       $59 = HEAP8[$mapstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $60 = $59 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $61 = ($60<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if ($61) {
        $65 = ((($mapstr)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP8[$65>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP8[$mapstr>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       } else {
        $62 = ((($mapstr)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $63 = HEAP32[$62>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP8[$63>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $64 = ((($mapstr)) + 4|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        HEAP32[$64>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       }
       __THREW__ = 0;
       invoke_vii(3368,($mapstr|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $66 = __THREW__; __THREW__ = 0;
       $67 = $66&1;
       if ($67) {
        $68 = ___cxa_find_matching_catch_3(0|0)|0;
        $69 = tempRet0;
        ___clang_call_terminate($68); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       } else {
        ;HEAP32[$mapstr>>2]=HEAP32[$1>>2]|0;HEAP32[$mapstr+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$mapstr+8>>2]=HEAP32[$1+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 203 "modules/joystick/sdl/JoystickModule.cpp"
        break;
       }
      }
     }
     $57 = ___cxa_find_matching_catch_2()|0;
     $58 = tempRet0;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 3983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $eh$lpad$body$index5Z2D = $58;$eh$lpad$body$indexZ2D = $57;
     label = 10;
     break L8;
    } else {
     __THREW__ = 0;
     (invoke_iii(3360,($mapstr|0),($33|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $37 = __THREW__; __THREW__ = 0;
     $38 = $37&1;
     if ($38) {
      label = 9;
      break L8;
     }
     __THREW__ = 0;
     invoke_vi(3775,($33|0)); //@line 198 "modules/joystick/sdl/JoystickModule.cpp"
     $39 = __THREW__; __THREW__ = 0;
     $40 = $39&1;
     if ($40) {
      label = 9;
      break L8;
     }
    }
   } while(0);
   $70 = ((($joyinputstream)) + 64|0); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $71 = ((($joyinputstream)) + 8|0); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   HEAP32[$71>>2] = (9320); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $72 = ((($joyinputstream)) + 12|0); //@line 897 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   HEAP32[$joyinputstream>>2] = ((9524)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
   HEAP32[$70>>2] = ((9544)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
   $73 = ((($joyinputstream)) + 4|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
   HEAP32[$73>>2] = 0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
   $74 = ((($joyinputstream)) + 64|0); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
   __THREW__ = 0;
   invoke_vii(3362,($74|0),($72|0)); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
   $75 = __THREW__; __THREW__ = 0;
   $76 = $75&1;
   if ($76) {
    $87 = ___cxa_find_matching_catch_2()|0;
    $88 = tempRet0;
    $$0$i = $88;$$01$i = $87;
   } else {
    $77 = ((($joyinputstream)) + 136|0); //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    HEAP32[$77>>2] = 0; //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    $78 = ((($joyinputstream)) + 140|0); //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    HEAP32[$78>>2] = -1; //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
    HEAP32[$joyinputstream>>2] = (9300); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    HEAP32[$70>>2] = (9340); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    HEAP32[$71>>2] = (9320); //@line 899 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    __THREW__ = 0;
    invoke_vi(3363,($72|0)); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $79 = __THREW__; __THREW__ = 0;
    $80 = $79&1;
    do {
     if ($80) {
      $89 = ___cxa_find_matching_catch_2()|0;
      $90 = tempRet0;
      $eh$lpad$body$i$index11Z2D = $90;$eh$lpad$body$i$indexZ2D = $89;
     } else {
      HEAP32[$72>>2] = (9356); //@line 243 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
      $81 = ((($joyinputstream)) + 44|0); //@line 210 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
      $82 = ((($joyinputstream)) + 60|0); //@line 242 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
      ;HEAP32[$81>>2]=0|0;HEAP32[$81+4>>2]=0|0;HEAP32[$81+8>>2]=0|0;HEAP32[$81+12>>2]=0|0; //@line 241 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
      HEAP32[$82>>2] = 24; //@line 242 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
      ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      __THREW__ = 0;
      invoke_vii(3364,($72|0),($0|0)); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
      $83 = __THREW__; __THREW__ = 0;
      $84 = $83&1;
      if ($84) {
       $85 = ___cxa_find_matching_catch_2()|0;
       $86 = tempRet0;
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($81); //@line 245 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
       __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($72); //@line 245 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
       $eh$lpad$body$i$index11Z2D = $86;$eh$lpad$body$i$indexZ2D = $85;
       break;
      }
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 244 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
      $91 = HEAP32[$joyinput>>2]|0; //@line 210 "modules/joystick/sdl/JoystickModule.cpp"
      switch ($91|0) {
      case 0:  {
       $92 = ((($joyinput)) + 4|0); //@line 213 "modules/joystick/sdl/JoystickModule.cpp"
       $93 = HEAP32[$92>>2]|0; //@line 214 "modules/joystick/sdl/JoystickModule.cpp"
       $94 = ($93|0)>(-1); //@line 213 "modules/joystick/sdl/JoystickModule.cpp"
       if ($94) {
        $95 = ((($joyinputstream)) + 8|0); //@line 214 "modules/joystick/sdl/JoystickModule.cpp"
        __THREW__ = 0;
        $96 = (invoke_iiii(3366,($95|0),(610849|0),1)|0); //@line 894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
        $97 = __THREW__; __THREW__ = 0;
        $98 = $97&1;
        if ($98) {
         label = 32;
        } else {
         __THREW__ = 0;
         (invoke_iii(3365,($96|0),($93|0))|0); //@line 214 "modules/joystick/sdl/JoystickModule.cpp"
         $99 = __THREW__; __THREW__ = 0;
         $100 = $99&1;
         if ($100) {
          label = 32;
         } else {
          label = 43;
         }
        }
       } else {
        label = 43;
       }
       break;
      }
      case 1:  {
       $103 = ((($joyinput)) + 4|0); //@line 217 "modules/joystick/sdl/JoystickModule.cpp"
       $104 = HEAP32[$103>>2]|0; //@line 218 "modules/joystick/sdl/JoystickModule.cpp"
       $105 = ($104|0)>(-1); //@line 217 "modules/joystick/sdl/JoystickModule.cpp"
       if ($105) {
        $106 = ((($joyinputstream)) + 8|0); //@line 218 "modules/joystick/sdl/JoystickModule.cpp"
        __THREW__ = 0;
        $107 = (invoke_iiii(3366,($106|0),(610851|0),1)|0); //@line 894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
        $108 = __THREW__; __THREW__ = 0;
        $109 = $108&1;
        if ($109) {
         label = 32;
        } else {
         __THREW__ = 0;
         (invoke_iii(3365,($107|0),($104|0))|0); //@line 218 "modules/joystick/sdl/JoystickModule.cpp"
         $110 = __THREW__; __THREW__ = 0;
         $111 = $110&1;
         if ($111) {
          label = 32;
         } else {
          label = 43;
         }
        }
       } else {
        label = 43;
       }
       break;
      }
      case 2:  {
       $112 = ((($joyinput)) + 4|0); //@line 221 "modules/joystick/sdl/JoystickModule.cpp"
       $113 = HEAP32[$112>>2]|0; //@line 222 "modules/joystick/sdl/JoystickModule.cpp"
       $114 = ($113|0)>(-1); //@line 221 "modules/joystick/sdl/JoystickModule.cpp"
       if ($114) {
        $115 = ((($joyinput)) + 8|0); //@line 221 "modules/joystick/sdl/JoystickModule.cpp"
        $116 = HEAP32[$115>>2]|0; //@line 221 "modules/joystick/sdl/JoystickModule.cpp"
        __THREW__ = 0;
        $117 = (invoke_iii(4190,($116|0),($sdlhat|0))|0); //@line 221 "modules/joystick/sdl/JoystickModule.cpp"
        $118 = __THREW__; __THREW__ = 0;
        $119 = $118&1;
        if ($119) {
         label = 32;
        } else {
         if ($117) {
          $120 = ((($joyinputstream)) + 8|0); //@line 222 "modules/joystick/sdl/JoystickModule.cpp"
          __THREW__ = 0;
          $121 = (invoke_iiii(3366,($120|0),(612359|0),1)|0); //@line 894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
          $122 = __THREW__; __THREW__ = 0;
          $123 = $122&1;
          if ($123) {
           label = 32;
          } else {
           __THREW__ = 0;
           $124 = (invoke_iii(3365,($121|0),($113|0))|0); //@line 222 "modules/joystick/sdl/JoystickModule.cpp"
           $125 = __THREW__; __THREW__ = 0;
           $126 = $125&1;
           if ($126) {
            label = 32;
           } else {
            __THREW__ = 0;
            $127 = (invoke_iiii(3366,($124|0),(619449|0),1)|0); //@line 894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ostream"
            $128 = __THREW__; __THREW__ = 0;
            $129 = $128&1;
            if ($129) {
             label = 32;
            } else {
             $130 = HEAP8[$sdlhat>>0]|0; //@line 222 "modules/joystick/sdl/JoystickModule.cpp"
             $131 = $130&255; //@line 222 "modules/joystick/sdl/JoystickModule.cpp"
             __THREW__ = 0;
             (invoke_iii(3365,($127|0),($131|0))|0); //@line 222 "modules/joystick/sdl/JoystickModule.cpp"
             $132 = __THREW__; __THREW__ = 0;
             $133 = $132&1;
             if ($133) {
              label = 32;
             } else {
              label = 43;
             }
            }
           }
          }
         } else {
          label = 43;
         }
        }
       } else {
        label = 43;
       }
       break;
      }
      default: {
       label = 43;
      }
      }
      if ((label|0) == 43) {
       __THREW__ = 0;
       invoke_vii(3367,($joyinputstr|0),($72|0)); //@line 964 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
       $134 = __THREW__; __THREW__ = 0;
       $135 = $134&1;
       if ($135) {
        label = 32;
       } else {
        $136 = HEAP8[$joyinputstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $137 = $136 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $138 = ($137<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $139 = ((($joyinputstr)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $140 = HEAP32[$139>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $141 = $136&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $142 = $141 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $143 = $138 ? $142 : $140; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $144 = ($143|0)==(0); //@line 230 "modules/joystick/sdl/JoystickModule.cpp"
        do {
         if ($144) {
          $145 = (___cxa_allocate_exception(16)|0); //@line 231 "modules/joystick/sdl/JoystickModule.cpp"
          __THREW__ = 0;
          invoke_viii(3349,($145|0),(441852|0),($vararg_buffer1|0)); //@line 231 "modules/joystick/sdl/JoystickModule.cpp"
          $146 = __THREW__; __THREW__ = 0;
          $147 = $146&1;
          if ($147) {
           $149 = ___cxa_find_matching_catch_2()|0;
           $150 = tempRet0;
           ___cxa_free_exception(($145|0)); //@line 231 "modules/joystick/sdl/JoystickModule.cpp"
           $$4 = $149;$$45 = $150;
           break;
          } else {
           __THREW__ = 0;
           invoke_viii(3350,($145|0),(96|0),(344|0)); //@line 231 "modules/joystick/sdl/JoystickModule.cpp"
           $148 = __THREW__; __THREW__ = 0;
           label = 48;
           break;
          }
         } else {
          $153 = $gpinput; //@line 234 "modules/joystick/sdl/JoystickModule.cpp"
          $154 = $153; //@line 234 "modules/joystick/sdl/JoystickModule.cpp"
          $155 = HEAP32[$154>>2]|0; //@line 234 "modules/joystick/sdl/JoystickModule.cpp"
          $156 = (($153) + 4)|0; //@line 234 "modules/joystick/sdl/JoystickModule.cpp"
          $157 = $156; //@line 234 "modules/joystick/sdl/JoystickModule.cpp"
          $158 = HEAP32[$157>>2]|0; //@line 234 "modules/joystick/sdl/JoystickModule.cpp"
          $159 = $2; //@line 234 "modules/joystick/sdl/JoystickModule.cpp"
          $160 = $159; //@line 234 "modules/joystick/sdl/JoystickModule.cpp"
          HEAP32[$160>>2] = $155; //@line 234 "modules/joystick/sdl/JoystickModule.cpp"
          $161 = (($159) + 4)|0; //@line 234 "modules/joystick/sdl/JoystickModule.cpp"
          $162 = $161; //@line 234 "modules/joystick/sdl/JoystickModule.cpp"
          HEAP32[$162>>2] = $158; //@line 234 "modules/joystick/sdl/JoystickModule.cpp"
          __THREW__ = 0;
          ;HEAP32[$tmpcast$byval_copy>>2]=HEAP32[$2>>2]|0;HEAP32[$tmpcast$byval_copy+4>>2]=HEAP32[$2+4>>2]|0; //@line 234 "modules/joystick/sdl/JoystickModule.cpp"
          invoke_viii(4191,($gpinputname|0),0,($tmpcast$byval_copy|0)); //@line 234 "modules/joystick/sdl/JoystickModule.cpp"
          $163 = __THREW__; __THREW__ = 0;
          $164 = $163&1;
          if ($164) {
           label = 48;
          } else {
           __THREW__ = 0;
           invoke_viii(4192,0,($mapstr|0),($joyinputstr|0)); //@line 238 "modules/joystick/sdl/JoystickModule.cpp"
           $165 = __THREW__; __THREW__ = 0;
           $166 = $165&1;
           L59: do {
            if ($166) {
             $278 = ___cxa_find_matching_catch_2()|0;
             $279 = tempRet0;
             $eh$lpad$body18$index16Z2D = $279;$eh$lpad$body18$indexZ2D = $278;
             label = 74;
            } else {
             ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $167 = HEAP8[$gpinputname>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $168 = $167 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $169 = ($168<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $170 = ((($gpinputname)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $171 = HEAP32[$170>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $172 = $167&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $173 = $172 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $174 = $169 ? $173 : $171; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $175 = ((($gpinputname)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $176 = HEAP32[$175>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $177 = ((($gpinputname)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $178 = $169 ? $177 : $176; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $179 = (($174) + 1)|0; //@line 3980 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             __THREW__ = 0;
             invoke_viiii(3514,($4|0),($178|0),($174|0),($179|0)); //@line 3980 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $180 = __THREW__; __THREW__ = 0;
             $181 = $180&1;
             if (!($181)) {
              __THREW__ = 0;
              (invoke_iiii(3462,($4|0),(442485|0),1)|0); //@line 3981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              $182 = __THREW__; __THREW__ = 0;
              $183 = $182&1;
              if (!($183)) {
               $186 = HEAP8[$joyinputstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $187 = $186 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $188 = ($187<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $189 = ((($joyinputstr)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $190 = HEAP32[$189>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $191 = ((($joyinputstr)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $192 = $188 ? $191 : $190; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $193 = HEAP32[$139>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $194 = $186&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $195 = $194 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $196 = $188 ? $195 : $193; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               __THREW__ = 0;
               $197 = (invoke_iiii(3462,($4|0),($192|0),($196|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $198 = __THREW__; __THREW__ = 0;
               $199 = $198&1;
               do {
                if ($199) {
                 $280 = ___cxa_find_matching_catch_2()|0;
                 $281 = tempRet0;
                 $$0 = $280;$$01 = $281;
                } else {
                 ;HEAP32[$3>>2]=HEAP32[$197>>2]|0;HEAP32[$3+4>>2]=HEAP32[$197+4>>2]|0;HEAP32[$3+8>>2]=HEAP32[$197+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                 ;HEAP32[$197>>2]=0|0;HEAP32[$197+4>>2]=0|0;HEAP32[$197+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                 __THREW__ = 0;
                 $200 = (invoke_iii(3806,($3|0),(442475|0))|0); //@line 4044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                 $201 = __THREW__; __THREW__ = 0;
                 $202 = $201&1;
                 if ($202) {
                  $282 = ___cxa_find_matching_catch_2()|0;
                  $283 = tempRet0;
                  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 241 "modules/joystick/sdl/JoystickModule.cpp"
                  $$0 = $282;$$01 = $283;
                  break;
                 }
                 ;HEAP32[$insertstr>>2]=HEAP32[$200>>2]|0;HEAP32[$insertstr+4>>2]=HEAP32[$200+4>>2]|0;HEAP32[$insertstr+8>>2]=HEAP32[$200+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                 ;HEAP32[$200>>2]=0|0;HEAP32[$200+4>>2]=0|0;HEAP32[$200+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 241 "modules/joystick/sdl/JoystickModule.cpp"
                 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 241 "modules/joystick/sdl/JoystickModule.cpp"
                 __THREW__ = 0;
                 invoke_viii(3460,($7|0),(441911|0),2); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                 $203 = __THREW__; __THREW__ = 0;
                 $204 = $203&1;
                 L71: do {
                  if ($204) {
                   label = 78;
                  } else {
                   $205 = HEAP8[$gpinputname>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                   $206 = $205 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                   $207 = ($206<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                   $208 = HEAP32[$175>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                   $209 = $207 ? $177 : $208; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                   $210 = HEAP32[$170>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                   $211 = $205&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                   $212 = $211 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                   $213 = $207 ? $212 : $210; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                   __THREW__ = 0;
                   $214 = (invoke_iiii(3462,($7|0),($209|0),($213|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                   $215 = __THREW__; __THREW__ = 0;
                   $216 = $215&1;
                   do {
                    if ($216) {
                     $286 = ___cxa_find_matching_catch_2()|0;
                     $287 = tempRet0;
                     $$1 = $286;$$12 = $287;
                    } else {
                     ;HEAP32[$6>>2]=HEAP32[$214>>2]|0;HEAP32[$6+4>>2]=HEAP32[$214+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$214+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     ;HEAP32[$214>>2]=0|0;HEAP32[$214+4>>2]=0|0;HEAP32[$214+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     __THREW__ = 0;
                     $217 = (invoke_iii(3806,($6|0),(442485|0))|0); //@line 4044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $218 = __THREW__; __THREW__ = 0;
                     $219 = $218&1;
                     if ($219) {
                      $288 = ___cxa_find_matching_catch_2()|0;
                      $289 = tempRet0;
                      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 244 "modules/joystick/sdl/JoystickModule.cpp"
                      $$1 = $288;$$12 = $289;
                      break;
                     }
                     ;HEAP32[$5>>2]=HEAP32[$217>>2]|0;HEAP32[$5+4>>2]=HEAP32[$217+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$217+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     ;HEAP32[$217>>2]=0|0;HEAP32[$217+4>>2]=0|0;HEAP32[$217+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $220 = HEAP8[$mapstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $221 = $220 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $222 = ($221<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $223 = ((($mapstr)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $224 = HEAP32[$223>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $225 = ((($mapstr)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $226 = HEAP32[$225>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $227 = ((($mapstr)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $228 = $220&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $229 = $228 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $230 = $222 ? $227 : $224; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $231 = $222 ? $229 : $226; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $232 = HEAP8[$5>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $233 = $232 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $234 = ($233<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $235 = ((($5)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $236 = HEAP32[$235>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $237 = ((($5)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $238 = $234 ? $237 : $236; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $239 = ((($5)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $240 = HEAP32[$239>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $241 = $232&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $242 = $241 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $243 = $234 ? $242 : $240; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $244 = ($231>>>0)<($243>>>0); //@line 962 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     L79: do {
                      if ($244) {
                       $$1$i$i = -1;
                      } else {
                       $245 = ($243|0)==(0); //@line 964 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                       if ($245) {
                        $$1$i$i = 0;
                        break;
                       }
                       $246 = (($230) + ($231)|0); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                       $247 = (($238) + ($243)|0); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                       $248 = $230; //@line 1466 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                       $249 = ($231|0)<($243|0); //@line 1467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                       L82: do {
                        if (!($249)) {
                         $250 = (1 - ($243))|0; //@line 1469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                         $251 = (($246) + ($250)|0); //@line 1469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                         $252 = ($251|0)==($230|0); //@line 1475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                         if ($252) {
                          break;
                         }
                         $253 = HEAP8[$238>>0]|0; //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                         $$016$i$i$i = $230;
                         L85: while(1) {
                          $254 = HEAP8[$$016$i$i$i>>0]|0; //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                          $255 = ($254<<24>>24)==($253<<24>>24); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                          if ($255) {
                           $__m1$0$i$i$i = $$016$i$i$i;$__m2$0$i$i$i = $238;
                           while(1) {
                            $258 = ((($__m2$0$i$i$i)) + 1|0); //@line 1517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                            $259 = ($258|0)==($247|0); //@line 1517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                            if ($259) {
                             $$016$i$i$i$lcssa = $$016$i$i$i;
                             break L85;
                            }
                            $260 = ((($__m1$0$i$i$i)) + 1|0); //@line 1519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                            $261 = HEAP8[$260>>0]|0; //@line 1520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                            $262 = HEAP8[$258>>0]|0; //@line 1520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                            $263 = ($261<<24>>24)==($262<<24>>24); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                            if ($263) {
                             $__m1$0$i$i$i = $260;$__m2$0$i$i$i = $258;
                            } else {
                             break;
                            }
                           }
                          }
                          $256 = ((($$016$i$i$i)) + 1|0); //@line 1479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                          $257 = ($256|0)==($251|0); //@line 1475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                          if ($257) {
                           break L82;
                          } else {
                           $$016$i$i$i = $256;
                          }
                         }
                         $264 = ($$016$i$i$i$lcssa|0)==($246|0); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                         $265 = $$016$i$i$i$lcssa; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                         $266 = (($265) - ($248))|0; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                         if (!($264)) {
                          $$1$i$i = $266;
                          break L79;
                         }
                        }
                       } while(0);
                       $$1$i$i = -1;
                      }
                     } while(0);
                     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5); //@line 244 "modules/joystick/sdl/JoystickModule.cpp"
                     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 244 "modules/joystick/sdl/JoystickModule.cpp"
                     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7); //@line 244 "modules/joystick/sdl/JoystickModule.cpp"
                     $267 = ($$1$i$i|0)==(-1); //@line 245 "modules/joystick/sdl/JoystickModule.cpp"
                     if ($267) {
                      $306 = HEAP8[$insertstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $307 = $306 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $308 = ($307<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $309 = ((($insertstr)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $310 = HEAP32[$309>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $311 = ((($insertstr)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $312 = $308 ? $311 : $310; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $313 = ((($insertstr)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $314 = HEAP32[$313>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $315 = $306&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $316 = $315 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $317 = $308 ? $316 : $314; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      __THREW__ = 0;
                      (invoke_iiii(3462,($mapstr|0),($312|0),($317|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $318 = __THREW__; __THREW__ = 0;
                      $319 = $318&1;
                      if ($319) {
                       label = 78;
                       break L71;
                      }
                     } else {
                      $268 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE4findEcj($mapstr,44,$$1$i$i)|0); //@line 3501 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $269 = ($268|0)==(-1); //@line 249 "modules/joystick/sdl/JoystickModule.cpp"
                      if ($269) {
                       $270 = HEAP8[$mapstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                       $271 = $270 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                       $272 = ($271<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                       $273 = HEAP32[$225>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                       $274 = $270&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                       $275 = $274 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                       $276 = $272 ? $275 : $273; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                       $277 = (($276) + -1)|0; //@line 250 "modules/joystick/sdl/JoystickModule.cpp"
                       $endpos$0 = $277;
                      } else {
                       $endpos$0 = $268;
                      }
                      $290 = (1 - ($$1$i$i))|0; //@line 252 "modules/joystick/sdl/JoystickModule.cpp"
                      $291 = (($290) + ($endpos$0))|0; //@line 252 "modules/joystick/sdl/JoystickModule.cpp"
                      $292 = HEAP8[$insertstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $293 = $292 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $294 = ($293<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $295 = ((($insertstr)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $296 = HEAP32[$295>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $297 = ((($insertstr)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $298 = $294 ? $297 : $296; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $299 = ((($insertstr)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $300 = HEAP32[$299>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $301 = $292&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $302 = $301 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $303 = $294 ? $302 : $300; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      __THREW__ = 0;
                      (invoke_iiiiii(3519,($mapstr|0),($$1$i$i|0),($291|0),($298|0),($303|0))|0); //@line 2997 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                      $304 = __THREW__; __THREW__ = 0;
                      $305 = $304&1;
                      if ($305) {
                       label = 78;
                       break L71;
                      }
                     }
                     $320 = HEAP8[$mapstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $321 = $320 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $322 = ($321<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $323 = HEAP32[$223>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     $324 = $322 ? $227 : $323; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                     __THREW__ = 0;
                     $325 = (invoke_ii(4193,($324|0))|0); //@line 261 "modules/joystick/sdl/JoystickModule.cpp"
                     $326 = __THREW__; __THREW__ = 0;
                     $327 = $326&1;
                     if ($327) {
                      label = 78;
                      break L71;
                     }
                     $cond = ($325|0)==(-1); //@line 263 "modules/joystick/sdl/JoystickModule.cpp"
                     if ($cond) {
                      $335 = ($325|0)>(-1); //@line 271 "modules/joystick/sdl/JoystickModule.cpp"
                      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($insertstr); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
                      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($gpinputname); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
                      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($joyinputstr); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
                      HEAP32[$joyinputstream>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                      HEAP32[$70>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                      $336 = ((($joyinputstream)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                      HEAP32[$336>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                      HEAP32[$72>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($81); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                      __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($72); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                      __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($70); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($mapstr); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
                      STACKTOP = sp;return ($335|0); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
                     }
                     $328 = ((($this)) + 32|0); //@line 264 "modules/joystick/sdl/JoystickModule.cpp"
                     __THREW__ = 0;
                     $329 = (invoke_iii(4194,($328|0),($guid|0))|0); //@line 264 "modules/joystick/sdl/JoystickModule.cpp"
                     $330 = __THREW__; __THREW__ = 0;
                     $331 = $330&1;
                     if ($331) {
                      label = 78;
                      break L71;
                     }
                     HEAP8[$329>>0] = 1; //@line 264 "modules/joystick/sdl/JoystickModule.cpp"
                     $332 = ($325|0)==(1); //@line 268 "modules/joystick/sdl/JoystickModule.cpp"
                     if (!($332)) {
                      $335 = ($325|0)>(-1); //@line 271 "modules/joystick/sdl/JoystickModule.cpp"
                      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($insertstr); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
                      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($gpinputname); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
                      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($joyinputstr); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
                      HEAP32[$joyinputstream>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                      HEAP32[$70>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                      $336 = ((($joyinputstream)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                      HEAP32[$336>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                      HEAP32[$72>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($81); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                      __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($72); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                      __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($70); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($mapstr); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
                      STACKTOP = sp;return ($335|0); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
                     }
                     __THREW__ = 0;
                     invoke_vii(4195,($this|0),($guid|0)); //@line 269 "modules/joystick/sdl/JoystickModule.cpp"
                     $333 = __THREW__; __THREW__ = 0;
                     $334 = $333&1;
                     if ($334) {
                      label = 78;
                      break L71;
                     }
                     $335 = ($325|0)>(-1); //@line 271 "modules/joystick/sdl/JoystickModule.cpp"
                     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($insertstr); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
                     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($gpinputname); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
                     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($joyinputstr); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
                     HEAP32[$joyinputstream>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                     HEAP32[$70>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                     $336 = ((($joyinputstream)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                     HEAP32[$336>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                     HEAP32[$72>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($81); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                     __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($72); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                     __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($70); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
                     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($mapstr); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
                     STACKTOP = sp;return ($335|0); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
                    }
                   } while(0);
                   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7); //@line 244 "modules/joystick/sdl/JoystickModule.cpp"
                   $$2 = $$1;$$23 = $$12;
                  }
                 } while(0);
                 if ((label|0) == 78) {
                  $284 = ___cxa_find_matching_catch_2()|0;
                  $285 = tempRet0;
                  $$2 = $284;$$23 = $285;
                 }
                 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($insertstr); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
                 $$3 = $$2;$$34 = $$23;
                 break L59;
                }
               } while(0);
               __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 241 "modules/joystick/sdl/JoystickModule.cpp"
               $$3 = $$0;$$34 = $$01;
               break;
              }
             }
             $184 = ___cxa_find_matching_catch_2()|0;
             $185 = tempRet0;
             __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 3983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $eh$lpad$body18$index16Z2D = $185;$eh$lpad$body18$indexZ2D = $184;
             label = 74;
            }
           } while(0);
           if ((label|0) == 74) {
            $$3 = $eh$lpad$body18$indexZ2D;$$34 = $eh$lpad$body18$index16Z2D;
           }
           __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($gpinputname); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
           $$4 = $$3;$$45 = $$34;
          }
         }
        } while(0);
        if ((label|0) == 48) {
         $151 = ___cxa_find_matching_catch_2()|0;
         $152 = tempRet0;
         $$4 = $151;$$45 = $152;
        }
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($joyinputstr); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
        $$5 = $$4;$$56 = $$45;
       }
      }
      if ((label|0) == 32) {
       $101 = ___cxa_find_matching_catch_2()|0;
       $102 = tempRet0;
       $$5 = $101;$$56 = $102;
      }
      HEAP32[$joyinputstream>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      HEAP32[$70>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      $337 = ((($joyinputstream)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      HEAP32[$337>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      HEAP32[$72>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($81); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($72); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($70); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      $$6 = $$5;$$67 = $$56;
      break L8;
     }
    } while(0);
    $$0$i = $eh$lpad$body$i$index11Z2D;$$01$i = $eh$lpad$body$i$indexZ2D;
   }
   __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($70); //@line 900 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $eh$lpad$body$index5Z2D = $$0$i;$eh$lpad$body$indexZ2D = $$01$i;
   label = 10;
  }
 } while(0);
 if ((label|0) == 9) {
  $41 = ___cxa_find_matching_catch_2()|0;
  $42 = tempRet0;
  $eh$lpad$body$index5Z2D = $42;$eh$lpad$body$indexZ2D = $41;
  label = 10;
 }
 if ((label|0) == 10) {
  $$6 = $eh$lpad$body$indexZ2D;$$67 = $eh$lpad$body$index5Z2D;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($mapstr); //@line 272 "modules/joystick/sdl/JoystickModule.cpp"
 $$7 = $$6;$$78 = $$67;
 ___resumeException($$7|0);
 // unreachable;
 return (0)|0;
}
function __ZNK4love8joystick3sdl14JoystickModule22stringFromGamepadInputENS0_8Joystick12GamepadInputE($agg$result,$this,$gpinput) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $gpinput = $gpinput|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $gpinputname$0 = 0, $sdlaxis = 0;
 var $sdlbutton = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $sdlaxis = sp + 8|0;
 $sdlbutton = sp + 4|0;
 $0 = HEAP32[$gpinput>>2]|0; //@line 329 "modules/joystick/sdl/JoystickModule.cpp"
 switch ($0|0) {
 case 0:  {
  $1 = ((($gpinput)) + 4|0); //@line 332 "modules/joystick/sdl/JoystickModule.cpp"
  $2 = HEAP32[$1>>2]|0; //@line 332 "modules/joystick/sdl/JoystickModule.cpp"
  $3 = (__ZN4love8joystick3sdl8Joystick11getConstantENS0_8Joystick11GamepadAxisER22SDL_GameControllerAxis($2,$sdlaxis)|0); //@line 332 "modules/joystick/sdl/JoystickModule.cpp"
  if ($3) {
   $4 = HEAP32[$sdlaxis>>2]|0; //@line 333 "modules/joystick/sdl/JoystickModule.cpp"
   $5 = (_SDL_GameControllerGetStringForAxis($4)|0); //@line 333 "modules/joystick/sdl/JoystickModule.cpp"
   $gpinputname$0 = $5;
   label = 6;
  }
  break;
 }
 case 1:  {
  $6 = ((($gpinput)) + 4|0); //@line 336 "modules/joystick/sdl/JoystickModule.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 336 "modules/joystick/sdl/JoystickModule.cpp"
  $8 = (__ZN4love8joystick3sdl8Joystick11getConstantENS0_8Joystick13GamepadButtonER24SDL_GameControllerButton($7,$sdlbutton)|0); //@line 336 "modules/joystick/sdl/JoystickModule.cpp"
  if ($8) {
   $9 = HEAP32[$sdlbutton>>2]|0; //@line 337 "modules/joystick/sdl/JoystickModule.cpp"
   $10 = (_SDL_GameControllerGetStringForButton($9)|0); //@line 337 "modules/joystick/sdl/JoystickModule.cpp"
   $gpinputname$0 = $10;
   label = 6;
  }
  break;
 }
 default: {
 }
 }
 if ((label|0) == 6) {
  $11 = ($gpinputname$0|0)==(0|0); //@line 343 "modules/joystick/sdl/JoystickModule.cpp"
  if (!($11)) {
   $17 = (_strlen($gpinputname$0)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($agg$result,$gpinputname$0,$17); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   STACKTOP = sp;return; //@line 347 "modules/joystick/sdl/JoystickModule.cpp"
  }
 }
 $12 = (___cxa_allocate_exception(16)|0); //@line 344 "modules/joystick/sdl/JoystickModule.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($12|0),(441882|0),($vararg_buffer|0)); //@line 344 "modules/joystick/sdl/JoystickModule.cpp"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if ($14) {
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  ___cxa_free_exception(($12|0)); //@line 344 "modules/joystick/sdl/JoystickModule.cpp"
  ___resumeException($15|0);
  // unreachable;
 } else {
  ___cxa_throw(($12|0),(96|0),(344|0)); //@line 344 "modules/joystick/sdl/JoystickModule.cpp"
  // unreachable; //@line 344 "modules/joystick/sdl/JoystickModule.cpp"
 }
}
function __ZNK4love8joystick3sdl14JoystickModule23removeBindFromMapStringERNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEERKS9_($this,$mapstr,$joybindstr) {
 $this = $this|0;
 $mapstr = $mapstr|0;
 $joybindstr = $joybindstr|0;
 var $$0$i$i8 = 0, $$016$i$i$i = 0, $$016$i$i$i$lcssa = 0, $$4$i$i$i7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $__l1$04$i$i$i = 0, $__m1$0$i$i$i = 0, $__m1$0$i$i$i4 = 0, $__m2$0$i$i$i = 0, $__m2$0$i$i$i5 = 0, $bindend$0 = 0, $joybindpos$013 = 0, $or$cond = 0, $or$cond$i$i = 0, $or$cond4$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $1 = HEAP8[$joybindstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $2 = $1 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ($2<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ((($joybindstr)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = HEAP32[$4>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = $1&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = $6 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $3 ? $7 : $5; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = ((($joybindstr)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = HEAP32[$9>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = ((($joybindstr)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = $3 ? $11 : $10; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = (($8) + 1)|0; //@line 3980 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 invoke_viiii(3514,($0|0),($12|0),($8|0),($13|0)); //@line 3980 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if ($15) {
  $18 = ___cxa_find_matching_catch_2()|0;
  $19 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 3983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  ___resumeException($18|0);
  // unreachable;
 }
 __THREW__ = 0;
 (invoke_iiii(3462,($0|0),(442475|0),1)|0); //@line 3981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if ($17) {
  $18 = ___cxa_find_matching_catch_2()|0;
  $19 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 3983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  ___resumeException($18|0);
  // unreachable;
 }
 $20 = HEAP8[$mapstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $21 = $20 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $22 = ($21<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $23 = ((($mapstr)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $24 = HEAP32[$23>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $25 = ((($mapstr)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $26 = HEAP32[$25>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $27 = ((($mapstr)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $28 = $20&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $29 = $28 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $30 = $22 ? $27 : $24; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $31 = $22 ? $29 : $26; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $32 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $33 = $32 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $34 = ($33<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $35 = ((($0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $36 = HEAP32[$35>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $37 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $38 = $34 ? $37 : $36; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $39 = ((($0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $40 = HEAP32[$39>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $41 = $32&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $42 = $41 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $43 = $34 ? $42 : $40; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $44 = ($31>>>0)<($43>>>0); //@line 962 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 L7: do {
  if ($44) {
   label = 15;
  } else {
   $45 = ($43|0)==(0); //@line 964 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if ($45) {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 397 "modules/joystick/sdl/JoystickModule.cpp"
    $joybindpos$013 = 0;
    break;
   }
   $46 = (($30) + ($31)|0); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $47 = (($38) + ($43)|0); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $48 = $30; //@line 1466 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $49 = ($31|0)<($43|0); //@line 1467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   if ($49) {
    label = 15;
   } else {
    $50 = (1 - ($43))|0; //@line 1469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $51 = (($46) + ($50)|0); //@line 1469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $52 = ($51|0)==($30|0); //@line 1475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    if ($52) {
     label = 15;
    } else {
     $53 = HEAP8[$38>>0]|0; //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $$016$i$i$i = $30;
     L14: while(1) {
      $54 = HEAP8[$$016$i$i$i>>0]|0; //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $55 = ($54<<24>>24)==($53<<24>>24); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($55) {
       $__m1$0$i$i$i = $$016$i$i$i;$__m2$0$i$i$i = $38;
       while(1) {
        $58 = ((($__m2$0$i$i$i)) + 1|0); //@line 1517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $59 = ($58|0)==($47|0); //@line 1517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        if ($59) {
         $$016$i$i$i$lcssa = $$016$i$i$i;
         break L14;
        }
        $60 = ((($__m1$0$i$i$i)) + 1|0); //@line 1519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $61 = HEAP8[$60>>0]|0; //@line 1520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $62 = HEAP8[$58>>0]|0; //@line 1520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $63 = ($61<<24>>24)==($62<<24>>24); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if ($63) {
         $__m1$0$i$i$i = $60;$__m2$0$i$i$i = $58;
        } else {
         break;
        }
       }
      }
      $56 = ((($$016$i$i$i)) + 1|0); //@line 1479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $57 = ($56|0)==($51|0); //@line 1475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if ($57) {
       label = 15;
       break L7;
      } else {
       $$016$i$i$i = $56;
      }
     }
     $64 = ($$016$i$i$i$lcssa|0)==($46|0); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $65 = $$016$i$i$i$lcssa; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $66 = (($65) - ($48))|0; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($64) {
      label = 15;
     } else {
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 397 "modules/joystick/sdl/JoystickModule.cpp"
      $67 = ($66|0)==(-1); //@line 398 "modules/joystick/sdl/JoystickModule.cpp"
      if ($67) {
       label = 17;
      } else {
       $joybindpos$013 = $66;
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 15) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 397 "modules/joystick/sdl/JoystickModule.cpp"
  label = 17;
 }
 if ((label|0) == 17) {
  $68 = HEAP8[$mapstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $69 = $68 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $70 = ($69<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $71 = HEAP32[$23>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $72 = HEAP32[$25>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $73 = $68&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $74 = $73 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $75 = $70 ? $27 : $71; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $76 = $70 ? $74 : $72; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $77 = HEAP8[$joybindstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $78 = $77 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $79 = ($78<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $80 = HEAP32[$9>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $81 = $79 ? $11 : $80; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $82 = HEAP32[$4>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $83 = $77&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $84 = $83 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $85 = $79 ? $84 : $82; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $86 = (($75) + ($76)|0); //@line 1008 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $87 = ($85|0)==(0); //@line 987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $88 = ($76|0)<($85|0); //@line 990 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $or$cond4$i$i = $87 | $88; //@line 987 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  L27: do {
   if ($or$cond4$i$i) {
    $$4$i$i$i7 = $86;
   } else {
    $89 = (($81) + ($85)|0); //@line 1008 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $90 = (($85) + -1)|0; //@line 992 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $91 = (($75) + ($90)|0); //@line 992 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $92 = ((($89)) + -1|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $93 = ($90|0)==($76|0); //@line 1000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    if ($93) {
     $$4$i$i$i7 = $86;
    } else {
     $94 = HEAP8[$92>>0]|0; //@line 1002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $__l1$04$i$i$i = $86;
     while(1) {
      $95 = ((($__l1$04$i$i$i)) + -1|0); //@line 1002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $96 = HEAP8[$95>>0]|0; //@line 1002 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $97 = ($96<<24>>24)==($94<<24>>24); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      if ($97) {
       $__m1$0$i$i$i4 = $95;$__m2$0$i$i$i5 = $92;
       while(1) {
        $99 = ($__m2$0$i$i$i5|0)==($81|0); //@line 1009 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        if ($99) {
         $$4$i$i$i7 = $__m1$0$i$i$i4;
         break L27;
        }
        $100 = ((($__m1$0$i$i$i4)) + -1|0); //@line 1012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $101 = HEAP8[$100>>0]|0; //@line 1012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $102 = ((($__m2$0$i$i$i5)) + -1|0); //@line 1012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $103 = HEAP8[$102>>0]|0; //@line 1012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $104 = ($101<<24>>24)==($103<<24>>24); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        if ($104) {
         $__m1$0$i$i$i4 = $100;$__m2$0$i$i$i5 = $102;
        } else {
         break;
        }
       }
      }
      $98 = ($91|0)==($95|0); //@line 1000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if ($98) {
       $$4$i$i$i7 = $86;
       break;
      } else {
       $__l1$04$i$i$i = $95;
      }
     }
    }
   }
  } while(0);
  $105 = ($85|0)!=(0); //@line 1010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $106 = ($$4$i$i$i7|0)==($86|0); //@line 1010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $or$cond$i$i = $105 & $106; //@line 1010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $107 = $$4$i$i$i7; //@line 1012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $108 = $75; //@line 1012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $109 = (($107) - ($108))|0; //@line 1012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $$0$i$i8 = $or$cond$i$i ? -1 : $109; //@line 1010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $110 = (($76) - ($85))|0; //@line 401 "modules/joystick/sdl/JoystickModule.cpp"
  $111 = ($$0$i$i8|0)!=($110|0); //@line 401 "modules/joystick/sdl/JoystickModule.cpp"
  $112 = ($$0$i$i8|0)==(-1); //@line 405 "modules/joystick/sdl/JoystickModule.cpp"
  $or$cond = $112 | $111; //@line 401 "modules/joystick/sdl/JoystickModule.cpp"
  if ($or$cond) {
   STACKTOP = sp;return; //@line 423 "modules/joystick/sdl/JoystickModule.cpp"
  } else {
   $joybindpos$013 = $$0$i$i8;
  }
 }
 $113 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE5rfindEcj($mapstr,44,$joybindpos$013)|0); //@line 410 "modules/joystick/sdl/JoystickModule.cpp"
 $114 = ($113|0)==(-1); //@line 411 "modules/joystick/sdl/JoystickModule.cpp"
 if ($114) {
  STACKTOP = sp;return; //@line 423 "modules/joystick/sdl/JoystickModule.cpp"
 }
 $115 = HEAP8[$mapstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $116 = $115 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $117 = ($116<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $118 = HEAP32[$25>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $119 = $115&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $120 = $119 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $121 = $117 ? $120 : $118; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $122 = (($121) + -1)|0; //@line 411 "modules/joystick/sdl/JoystickModule.cpp"
 $123 = ($113>>>0)<($122>>>0); //@line 411 "modules/joystick/sdl/JoystickModule.cpp"
 if (!($123)) {
  STACKTOP = sp;return; //@line 423 "modules/joystick/sdl/JoystickModule.cpp"
 }
 $124 = (($113) + 1)|0; //@line 414 "modules/joystick/sdl/JoystickModule.cpp"
 $125 = (($113) + 2)|0; //@line 416 "modules/joystick/sdl/JoystickModule.cpp"
 $126 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE4findEcj($mapstr,44,$125)|0); //@line 416 "modules/joystick/sdl/JoystickModule.cpp"
 $127 = ($126|0)==(-1); //@line 417 "modules/joystick/sdl/JoystickModule.cpp"
 if ($127) {
  $128 = HEAP8[$mapstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $129 = $128 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $130 = ($129<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $131 = HEAP32[$25>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $132 = $128&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $133 = $132 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $134 = $130 ? $133 : $131; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $135 = (($134) + -1)|0; //@line 418 "modules/joystick/sdl/JoystickModule.cpp"
  $bindend$0 = $135;
 } else {
  $bindend$0 = $126;
 }
 $136 = (($bindend$0) - ($113))|0; //@line 421 "modules/joystick/sdl/JoystickModule.cpp"
 (__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7replaceEjjPKc($mapstr,$124,$136,1282466)|0); //@line 421 "modules/joystick/sdl/JoystickModule.cpp"
 STACKTOP = sp;return; //@line 423 "modules/joystick/sdl/JoystickModule.cpp"
}
function __ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEbNS_4lessIS6_EENS4_INS_4pairIKS6_bEEEEEixERSA_($this,$__k) {
 $this = $this|0;
 $__k = $__k|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__parent = 0, $__r$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__parent = sp;
 $0 = (__ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEbNS_4lessIS6_EENS4_INS_4pairIKS6_bEEEEE16__find_equal_keyERPNS_16__tree_node_baseIPvEERSA_($this,$__parent,$__k)|0); //@line 1376 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $1 = HEAP32[$0>>2]|0; //@line 1377 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $2 = ($1|0)==(0|0); //@line 1378 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 if (!($2)) {
  $__r$0 = $1;
  $23 = ((($__r$0)) + 28|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  STACKTOP = sp;return ($23|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 $3 = (__Znwj(32)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $4 = ((($3)) + 16|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 __THREW__ = 0;
 invoke_vii(3351,($4|0),($__k|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZdlPv($3); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($7|0);
  // unreachable;
 }
 $9 = ((($3)) + 28|0); //@line 1366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP8[$9>>0] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $10 = HEAP32[$__parent>>2]|0; //@line 1381 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $11 = ((($3)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$11>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $12 = ((($3)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$12>>2] = $10; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$0>>2] = $3; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $13 = HEAP32[$this>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $14 = HEAP32[$13>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $15 = ($14|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($15) {
  $19 = $3;
 } else {
  $16 = $14; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$this>>2] = $16; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre$i = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $19 = $$pre$i;
 }
 $17 = ((($this)) + 4|0); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $18 = HEAP32[$17>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($18,$19); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $20 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $21 = HEAP32[$20>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $22 = (($21) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$20>>2] = $22; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $__r$0 = $3;
 $23 = ((($__r$0)) + 28|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 STACKTOP = sp;return ($23|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
}
function __ZNK4love8joystick3sdl14JoystickModule13checkGamepadsERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($this,$guid) {
 $this = $this|0;
 $guid = $guid|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $__begin$sroa$0$011 = 0, $d_index$012 = 0, $guid$i = 0, $guid$i$byval_copy = 0, $guidstr$i = 0;
 var dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $guid$i$byval_copy = sp + 80|0;
 $guidstr$i = sp + 40|0;
 $guid$i = sp + 24|0;
 $0 = sp + 12|0;
 $1 = sp;
 $2 = (_SDL_NumJoysticks()|0); //@line 432 "modules/joystick/sdl/JoystickModule.cpp"
 $3 = ($2|0)>(0); //@line 432 "modules/joystick/sdl/JoystickModule.cpp"
 if (!($3)) {
  STACKTOP = sp;return; //@line 459 "modules/joystick/sdl/JoystickModule.cpp"
 }
 $4 = ((($guid)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = ((($guid)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ((($guid)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = ((($0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = ((($this)) + 8|0); //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ((($this)) + 12|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = ((($1)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ((($1)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = ((($1)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $d_index$012 = 0;
 while(1) {
  $15 = (_SDL_IsGameController($d_index$012)|0); //@line 434 "modules/joystick/sdl/JoystickModule.cpp"
  $16 = ($15|0)==(0); //@line 434 "modules/joystick/sdl/JoystickModule.cpp"
  if (!($16)) {
   $17 = (_SDL_NumJoysticks()|0); //@line 463 "modules/joystick/sdl/JoystickModule.cpp"
   $18 = ($17|0)>($d_index$012|0); //@line 463 "modules/joystick/sdl/JoystickModule.cpp"
   if ($18) {
    dest=$guidstr$i; stop=dest+33|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 467 "modules/joystick/sdl/JoystickModule.cpp"
    _SDL_JoystickGetDeviceGUID($guid$i,$d_index$012); //@line 470 "modules/joystick/sdl/JoystickModule.cpp"
    dest=$guid$i$byval_copy; src=$guid$i; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 471 "modules/joystick/sdl/JoystickModule.cpp"
    _SDL_JoystickGetGUIDString($guid$i$byval_copy,$guidstr$i,33); //@line 471 "modules/joystick/sdl/JoystickModule.cpp"
    $19 = (_strlen($guidstr$i)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($0,$guidstr$i,$19); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   } else {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($0,1282466,0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   }
   $20 = HEAP8[$guid>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $21 = $20 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $22 = ($21<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $23 = HEAP32[$4>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $24 = $20&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $25 = $24 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $26 = $22 ? $25 : $23; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $27 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $28 = $27 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $29 = ($28<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $30 = HEAP32[$5>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $31 = $27&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $32 = $31 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $33 = $29 ? $32 : $30; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $34 = HEAP32[$6>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $35 = $22 ? $7 : $34; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $36 = HEAP32[$8>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $37 = $29 ? $9 : $36; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $38 = ($26>>>0)>($33>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $39 = $38 ? $33 : $26; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $40 = (_memcmp($35,$37,$39)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $41 = ($40|0)!=(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 437 "modules/joystick/sdl/JoystickModule.cpp"
   $42 = ($26|0)!=($33|0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $43 = $42 | $41; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if (!($43)) {
    $44 = HEAP32[$10>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $45 = HEAP32[$11>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $46 = ($44|0)==($45|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if (!($46)) {
     $__begin$sroa$0$011 = $44;
     while(1) {
      $47 = HEAP32[$__begin$sroa$0$011>>2]|0; //@line 440 "modules/joystick/sdl/JoystickModule.cpp"
      $48 = HEAP32[$47>>2]|0; //@line 442 "modules/joystick/sdl/JoystickModule.cpp"
      $49 = ((($48)) + 64|0); //@line 442 "modules/joystick/sdl/JoystickModule.cpp"
      $50 = HEAP32[$49>>2]|0; //@line 442 "modules/joystick/sdl/JoystickModule.cpp"
      $51 = (FUNCTION_TABLE_ii[$50 & 8191]($47)|0); //@line 442 "modules/joystick/sdl/JoystickModule.cpp"
      if (!($51)) {
       $52 = HEAP32[$47>>2]|0; //@line 442 "modules/joystick/sdl/JoystickModule.cpp"
       $53 = ((($52)) + 80|0); //@line 442 "modules/joystick/sdl/JoystickModule.cpp"
       $54 = HEAP32[$53>>2]|0; //@line 442 "modules/joystick/sdl/JoystickModule.cpp"
       FUNCTION_TABLE_vii[$54 & 8191]($1,$47); //@line 442 "modules/joystick/sdl/JoystickModule.cpp"
       $55 = HEAP8[$guid>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $56 = $55 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $57 = ($56<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $58 = HEAP32[$4>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $59 = $55&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $60 = $59 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $61 = $57 ? $60 : $58; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $62 = HEAP8[$1>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $63 = $62 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $64 = ($63<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $65 = HEAP32[$12>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $66 = $62&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $67 = $66 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $68 = $64 ? $67 : $65; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $69 = HEAP32[$6>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $70 = $57 ? $7 : $69; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $71 = HEAP32[$13>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $72 = $64 ? $14 : $71; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $73 = ($61>>>0)>($68>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $74 = $73 ? $68 : $61; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $75 = (_memcmp($70,$72,$74)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $76 = ($75|0)!=(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 442 "modules/joystick/sdl/JoystickModule.cpp"
       $77 = ($61|0)!=($68|0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $78 = $77 | $76; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       if (!($78)) {
        $79 = (_SDL_GameControllerOpen($d_index$012)|0); //@line 447 "modules/joystick/sdl/JoystickModule.cpp"
        $80 = ($79|0)==(0|0); //@line 448 "modules/joystick/sdl/JoystickModule.cpp"
        if (!($80)) {
         $81 = (_SDL_GameControllerGetJoystick($79)|0); //@line 451 "modules/joystick/sdl/JoystickModule.cpp"
         $82 = HEAP32[$47>>2]|0; //@line 452 "modules/joystick/sdl/JoystickModule.cpp"
         $83 = ((($82)) + 76|0); //@line 452 "modules/joystick/sdl/JoystickModule.cpp"
         $84 = HEAP32[$83>>2]|0; //@line 452 "modules/joystick/sdl/JoystickModule.cpp"
         $85 = (FUNCTION_TABLE_ii[$84 & 8191]($47)|0); //@line 452 "modules/joystick/sdl/JoystickModule.cpp"
         $86 = ($81|0)==($85|0); //@line 452 "modules/joystick/sdl/JoystickModule.cpp"
         if ($86) {
          $87 = HEAP32[$47>>2]|0; //@line 453 "modules/joystick/sdl/JoystickModule.cpp"
          $88 = ((($87)) + 60|0); //@line 453 "modules/joystick/sdl/JoystickModule.cpp"
          $89 = HEAP32[$88>>2]|0; //@line 453 "modules/joystick/sdl/JoystickModule.cpp"
          (FUNCTION_TABLE_iii[$89 & 8191]($47,$d_index$012)|0); //@line 453 "modules/joystick/sdl/JoystickModule.cpp"
         }
         _SDL_GameControllerClose($79); //@line 456 "modules/joystick/sdl/JoystickModule.cpp"
        }
       }
      }
      $90 = ((($__begin$sroa$0$011)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $91 = ($90|0)==($45|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($91) {
       break;
      } else {
       $__begin$sroa$0$011 = $90;
      }
     }
    }
   }
  }
  $92 = (($d_index$012) + 1)|0; //@line 432 "modules/joystick/sdl/JoystickModule.cpp"
  $93 = (_SDL_NumJoysticks()|0); //@line 432 "modules/joystick/sdl/JoystickModule.cpp"
  $94 = ($92|0)<($93|0); //@line 432 "modules/joystick/sdl/JoystickModule.cpp"
  if ($94) {
   $d_index$012 = $92;
  } else {
   break;
  }
 }
 STACKTOP = sp;return; //@line 459 "modules/joystick/sdl/JoystickModule.cpp"
}
function __ZN4love8joystick3sdl14JoystickModule17getGamepadMappingERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEENS0_8Joystick12GamepadInputE($agg$result,$this,$guid,$gpinput) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $guid = $guid|0;
 $gpinput = $gpinput|0;
 var $$0 = 0, $$01 = 0, $$016$i$i$i = 0, $$016$i$i$i$lcssa = 0, $$1 = 0, $$1$i$i = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$3 = 0, $$34 = 0, $$endpos$0 = 0, $$pre11 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0;
 var $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $__m1$0$i$i$i = 0, $__m2$0$i$i$i = 0, $endpos$0 = 0, $gpbindname = 0, $jbindstr = 0, $mapstr = 0, $sdlguid = 0, $sdlguid$byval_copy = 0, $tmpcast$byval_copy = 0, $vararg_buffer = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tmpcast$byval_copy = sp + 96|0;
 $sdlguid$byval_copy = sp + 120|0;
 $vararg_buffer = sp + 8|0;
 $sdlguid = sp + 104|0;
 $mapstr = sp + 84|0;
 $gpbindname = sp + 72|0;
 $0 = sp;
 $1 = sp + 60|0;
 $2 = sp + 48|0;
 $3 = sp + 36|0;
 $jbindstr = sp + 24|0;
 $4 = sp + 12|0;
 $5 = HEAP8[$guid>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = $5 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = ($6<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = ((($guid)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = HEAP32[$8>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $5&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = $10 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = $7 ? $11 : $9; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ($12|0)==(32); //@line 277 "modules/joystick/sdl/JoystickModule.cpp"
 if (!($13)) {
  $14 = (___cxa_allocate_exception(16)|0); //@line 278 "modules/joystick/sdl/JoystickModule.cpp"
  $15 = HEAP8[$guid>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $16 = $15 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $17 = ($16<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $18 = ((($guid)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $19 = HEAP32[$18>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $20 = ((($guid)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $21 = $17 ? $20 : $19; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  __THREW__ = 0;
  HEAP32[$vararg_buffer>>2] = $21; //@line 278 "modules/joystick/sdl/JoystickModule.cpp"
  invoke_viii(3349,($14|0),(441813|0),($vararg_buffer|0)); //@line 278 "modules/joystick/sdl/JoystickModule.cpp"
  $22 = __THREW__; __THREW__ = 0;
  $23 = $22&1;
  if (!($23)) {
   ___cxa_throw(($14|0),(96|0),(344|0)); //@line 278 "modules/joystick/sdl/JoystickModule.cpp"
   // unreachable; //@line 278 "modules/joystick/sdl/JoystickModule.cpp"
  }
  $24 = ___cxa_find_matching_catch_2()|0;
  $25 = tempRet0;
  ___cxa_free_exception(($14|0)); //@line 278 "modules/joystick/sdl/JoystickModule.cpp"
  $$3 = $24;$$34 = $25;
  ___resumeException($$3|0);
  // unreachable;
 }
 HEAP32[$agg$result>>2] = 3; //@line 281 "modules/joystick/sdl/JoystickModule.cpp"
 $26 = ((($guid)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $27 = HEAP32[$26>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $28 = ((($guid)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $29 = $7 ? $28 : $27; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 _SDL_JoystickGetGUIDFromString($sdlguid,$29); //@line 283 "modules/joystick/sdl/JoystickModule.cpp"
 ;HEAP32[$mapstr>>2]=0|0;HEAP32[$mapstr+4>>2]=0|0;HEAP32[$mapstr+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 dest=$sdlguid$byval_copy; src=$sdlguid; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 287 "modules/joystick/sdl/JoystickModule.cpp"
 $30 = (invoke_ii(4189,($sdlguid$byval_copy|0))|0); //@line 287 "modules/joystick/sdl/JoystickModule.cpp"
 $31 = __THREW__; __THREW__ = 0;
 $32 = $31&1;
 if ($32) {
  label = 7;
 } else {
  $33 = ($30|0)==(0|0); //@line 288 "modules/joystick/sdl/JoystickModule.cpp"
  if ($33) {
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($mapstr); //@line 320 "modules/joystick/sdl/JoystickModule.cpp"
   STACKTOP = sp;return; //@line 320 "modules/joystick/sdl/JoystickModule.cpp"
  }
  __THREW__ = 0;
  (invoke_iii(3360,($mapstr|0),($30|0))|0); //@line 1370 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $36 = __THREW__; __THREW__ = 0;
  $37 = $36&1;
  if ($37) {
   label = 7;
  } else {
   __THREW__ = 0;
   invoke_vi(3775,($30|0)); //@line 292 "modules/joystick/sdl/JoystickModule.cpp"
   $38 = __THREW__; __THREW__ = 0;
   $39 = $38&1;
   if ($39) {
    label = 7;
   } else {
    $40 = $gpinput; //@line 294 "modules/joystick/sdl/JoystickModule.cpp"
    $41 = $40; //@line 294 "modules/joystick/sdl/JoystickModule.cpp"
    $42 = HEAP32[$41>>2]|0; //@line 294 "modules/joystick/sdl/JoystickModule.cpp"
    $43 = (($40) + 4)|0; //@line 294 "modules/joystick/sdl/JoystickModule.cpp"
    $44 = $43; //@line 294 "modules/joystick/sdl/JoystickModule.cpp"
    $45 = HEAP32[$44>>2]|0; //@line 294 "modules/joystick/sdl/JoystickModule.cpp"
    $46 = $0; //@line 294 "modules/joystick/sdl/JoystickModule.cpp"
    $47 = $46; //@line 294 "modules/joystick/sdl/JoystickModule.cpp"
    HEAP32[$47>>2] = $42; //@line 294 "modules/joystick/sdl/JoystickModule.cpp"
    $48 = (($46) + 4)|0; //@line 294 "modules/joystick/sdl/JoystickModule.cpp"
    $49 = $48; //@line 294 "modules/joystick/sdl/JoystickModule.cpp"
    HEAP32[$49>>2] = $45; //@line 294 "modules/joystick/sdl/JoystickModule.cpp"
    __THREW__ = 0;
    ;HEAP32[$tmpcast$byval_copy>>2]=HEAP32[$0>>2]|0;HEAP32[$tmpcast$byval_copy+4>>2]=HEAP32[$0+4>>2]|0; //@line 294 "modules/joystick/sdl/JoystickModule.cpp"
    invoke_viii(4191,($gpbindname|0),0,($tmpcast$byval_copy|0)); //@line 294 "modules/joystick/sdl/JoystickModule.cpp"
    $50 = __THREW__; __THREW__ = 0;
    $51 = $50&1;
    if ($51) {
     label = 7;
    } else {
     __THREW__ = 0;
     invoke_viii(3460,($3|0),(442475|0),1); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $52 = __THREW__; __THREW__ = 0;
     $53 = $52&1;
     L16: do {
      if ($53) {
       label = 26;
      } else {
       $54 = HEAP8[$gpbindname>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $55 = $54 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $56 = ($55<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $57 = ((($gpbindname)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $58 = HEAP32[$57>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $59 = ((($gpbindname)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $60 = $56 ? $59 : $58; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $61 = ((($gpbindname)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $62 = HEAP32[$61>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $63 = $54&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $64 = $63 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $65 = $56 ? $64 : $62; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       __THREW__ = 0;
       $66 = (invoke_iiii(3462,($3|0),($60|0),($65|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $67 = __THREW__; __THREW__ = 0;
       $68 = $67&1;
       do {
        if ($68) {
         $122 = ___cxa_find_matching_catch_2()|0;
         $123 = tempRet0;
         $$0 = $122;$$01 = $123;
        } else {
         ;HEAP32[$2>>2]=HEAP32[$66>>2]|0;HEAP32[$2+4>>2]=HEAP32[$66+4>>2]|0;HEAP32[$2+8>>2]=HEAP32[$66+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         ;HEAP32[$66>>2]=0|0;HEAP32[$66+4>>2]=0|0;HEAP32[$66+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         __THREW__ = 0;
         $69 = (invoke_iii(3806,($2|0),(442485|0))|0); //@line 4044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $70 = __THREW__; __THREW__ = 0;
         $71 = $70&1;
         if ($71) {
          $124 = ___cxa_find_matching_catch_2()|0;
          $125 = tempRet0;
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 296 "modules/joystick/sdl/JoystickModule.cpp"
          $$0 = $124;$$01 = $125;
          break;
         }
         ;HEAP32[$1>>2]=HEAP32[$69>>2]|0;HEAP32[$1+4>>2]=HEAP32[$69+4>>2]|0;HEAP32[$1+8>>2]=HEAP32[$69+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         ;HEAP32[$69>>2]=0|0;HEAP32[$69+4>>2]=0|0;HEAP32[$69+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $72 = HEAP8[$mapstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $73 = $72 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $74 = ($73<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $75 = ((($mapstr)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $76 = HEAP32[$75>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $77 = ((($mapstr)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $78 = HEAP32[$77>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $79 = ((($mapstr)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $80 = $72&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $81 = $80 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $82 = $74 ? $79 : $76; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $83 = $74 ? $81 : $78; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $84 = HEAP8[$1>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $85 = $84 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $86 = ($85<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $87 = ((($1)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $88 = HEAP32[$87>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $89 = ((($1)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $90 = $86 ? $89 : $88; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $91 = ((($1)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $92 = HEAP32[$91>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $93 = $84&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $94 = $93 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $95 = $86 ? $94 : $92; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $96 = ($83>>>0)<($95>>>0); //@line 962 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         L24: do {
          if ($96) {
           $$1$i$i = -1;
          } else {
           $97 = ($95|0)==(0); //@line 964 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($97) {
            $$1$i$i = 0;
           } else {
            $98 = (($82) + ($83)|0); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $99 = (($90) + ($95)|0); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $100 = $82; //@line 1466 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
            $101 = ($83|0)<($95|0); //@line 1467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
            L27: do {
             if (!($101)) {
              $102 = (1 - ($95))|0; //@line 1469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
              $103 = (($98) + ($102)|0); //@line 1469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
              $104 = ($103|0)==($82|0); //@line 1475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
              if (!($104)) {
               $105 = HEAP8[$90>>0]|0; //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
               $$016$i$i$i = $82;
               L30: while(1) {
                $106 = HEAP8[$$016$i$i$i>>0]|0; //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                $107 = ($106<<24>>24)==($105<<24>>24); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                if ($107) {
                 $__m1$0$i$i$i = $$016$i$i$i;$__m2$0$i$i$i = $90;
                 while(1) {
                  $110 = ((($__m2$0$i$i$i)) + 1|0); //@line 1517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                  $111 = ($110|0)==($99|0); //@line 1517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                  if ($111) {
                   $$016$i$i$i$lcssa = $$016$i$i$i;
                   break L30;
                  }
                  $112 = ((($__m1$0$i$i$i)) + 1|0); //@line 1519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                  $113 = HEAP8[$112>>0]|0; //@line 1520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                  $114 = HEAP8[$110>>0]|0; //@line 1520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                  $115 = ($113<<24>>24)==($114<<24>>24); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                  if ($115) {
                   $__m1$0$i$i$i = $112;$__m2$0$i$i$i = $110;
                  } else {
                   break;
                  }
                 }
                }
                $108 = ((($$016$i$i$i)) + 1|0); //@line 1479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                $109 = ($108|0)==($103|0); //@line 1475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
                if ($109) {
                 break L27;
                } else {
                 $$016$i$i$i = $108;
                }
               }
               $116 = ($$016$i$i$i$lcssa|0)==($98|0); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $117 = $$016$i$i$i$lcssa; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $118 = (($117) - ($100))|0; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               if (!($116)) {
                $$1$i$i = $118;
                break L24;
               }
              }
             }
            } while(0);
            $$1$i$i = -1;
           }
          }
         } while(0);
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 296 "modules/joystick/sdl/JoystickModule.cpp"
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 296 "modules/joystick/sdl/JoystickModule.cpp"
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 296 "modules/joystick/sdl/JoystickModule.cpp"
         $119 = ($$1$i$i|0)==(-1); //@line 297 "modules/joystick/sdl/JoystickModule.cpp"
         do {
          if (!($119)) {
           $126 = (($$1$i$i) + 1)|0; //@line 300 "modules/joystick/sdl/JoystickModule.cpp"
           $127 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE4findEcj($mapstr,44,$126)|0); //@line 3501 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $128 = ($127|0)==(-1); //@line 301 "modules/joystick/sdl/JoystickModule.cpp"
           $129 = HEAP8[$mapstr>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           if ($128) {
            $130 = $129 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $131 = ($130<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $132 = HEAP32[$77>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $133 = $129&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $134 = $133 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $135 = $131 ? $134 : $132; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $136 = (($135) + -1)|0; //@line 304 "modules/joystick/sdl/JoystickModule.cpp"
            $142 = $132;$endpos$0 = $136;
           } else {
            $$pre11 = HEAP32[$77>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $142 = $$pre11;$endpos$0 = $127;
           }
           $137 = $129 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $138 = ($137<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $139 = $129&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $140 = $139 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $141 = $138 ? $140 : $142; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $143 = ($endpos$0>>>0)<($141>>>0); //@line 307 "modules/joystick/sdl/JoystickModule.cpp"
           if ($143) {
            $144 = HEAP32[$75>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $145 = $138 ? $79 : $144; //@line 1755 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $146 = (($145) + ($endpos$0)|0); //@line 3271 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $147 = HEAP8[$146>>0]|0; //@line 311 "modules/joystick/sdl/JoystickModule.cpp"
            $148 = ($147<<24>>24)==(44); //@line 311 "modules/joystick/sdl/JoystickModule.cpp"
            $149 = $148 << 31 >> 31; //@line 311 "modules/joystick/sdl/JoystickModule.cpp"
            $150 = HEAP8[$gpbindname>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $151 = $150 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $152 = ($151<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $153 = HEAP32[$61>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $154 = $150&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $155 = $154 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $156 = $152 ? $155 : $153; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $157 = (($$1$i$i) + 2)|0; //@line 315 "modules/joystick/sdl/JoystickModule.cpp"
            $158 = (($157) + ($156))|0; //@line 315 "modules/joystick/sdl/JoystickModule.cpp"
            $$endpos$0 = (($endpos$0) + 1)|0; //@line 311 "modules/joystick/sdl/JoystickModule.cpp"
            $159 = (($$endpos$0) + ($149))|0; //@line 316 "modules/joystick/sdl/JoystickModule.cpp"
            $160 = (($159) - ($158))|0; //@line 316 "modules/joystick/sdl/JoystickModule.cpp"
            __THREW__ = 0;
            invoke_viiiii(3511,($jbindstr|0),($mapstr|0),($158|0),($160|0),($mapstr|0)); //@line 3345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $161 = __THREW__; __THREW__ = 0;
            $162 = $161&1;
            if ($162) {
             label = 26;
             break L16;
            }
            __THREW__ = 0;
            invoke_viii(4196,($4|0),0,($jbindstr|0)); //@line 318 "modules/joystick/sdl/JoystickModule.cpp"
            $163 = __THREW__; __THREW__ = 0;
            $164 = $163&1;
            if ($164) {
             $165 = ___cxa_find_matching_catch_2()|0;
             $166 = tempRet0;
             __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($jbindstr); //@line 320 "modules/joystick/sdl/JoystickModule.cpp"
             $$1 = $165;$$12 = $166;
             break L16;
            } else {
             ;HEAP32[$agg$result>>2]=HEAP32[$4>>2]|0;HEAP32[$agg$result+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$agg$result+8>>2]=HEAP32[$4+8>>2]|0; //@line 318 "modules/joystick/sdl/JoystickModule.cpp"
             __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($jbindstr); //@line 320 "modules/joystick/sdl/JoystickModule.cpp"
             break;
            }
           }
          }
         } while(0);
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($gpbindname); //@line 320 "modules/joystick/sdl/JoystickModule.cpp"
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($mapstr); //@line 320 "modules/joystick/sdl/JoystickModule.cpp"
         STACKTOP = sp;return; //@line 320 "modules/joystick/sdl/JoystickModule.cpp"
        }
       } while(0);
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 296 "modules/joystick/sdl/JoystickModule.cpp"
       $$1 = $$0;$$12 = $$01;
      }
     } while(0);
     if ((label|0) == 26) {
      $120 = ___cxa_find_matching_catch_2()|0;
      $121 = tempRet0;
      $$1 = $120;$$12 = $121;
     }
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($gpbindname); //@line 320 "modules/joystick/sdl/JoystickModule.cpp"
     $$2 = $$1;$$23 = $$12;
    }
   }
  }
 }
 if ((label|0) == 7) {
  $34 = ___cxa_find_matching_catch_2()|0;
  $35 = tempRet0;
  $$2 = $34;$$23 = $35;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($mapstr); //@line 320 "modules/joystick/sdl/JoystickModule.cpp"
 $$3 = $$2;$$34 = $$23;
 ___resumeException($$3|0);
 // unreachable;
}
function __ZNK4love8joystick3sdl14JoystickModule23JoystickInputFromStringERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($agg$result,$this,$str) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $str = $str|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $bindvalues = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $bindvalues = sp + 24|0;
 $0 = sp + 12|0;
 $1 = sp;
 HEAP32[$agg$result>>2] = 3; //@line 352 "modules/joystick/sdl/JoystickModule.cpp"
 $2 = HEAP8[$str>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = $2 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = ($3<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ((($str)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = HEAP32[$5>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = $2&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $7 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $4 ? $8 : $6; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = ($9>>>0)<(2); //@line 355 "modules/joystick/sdl/JoystickModule.cpp"
 if ($10) {
  STACKTOP = sp;return; //@line 392 "modules/joystick/sdl/JoystickModule.cpp"
 }
 $11 = ((($str)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = HEAP32[$11>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ((($str)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = $4 ? $13 : $12; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $15 = HEAP8[$14>>0]|0; //@line 359 "modules/joystick/sdl/JoystickModule.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_jjRKS4_($bindvalues,$str,1,-1,$str); //@line 3345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $16 = $15 << 24 >> 24; //@line 363 "modules/joystick/sdl/JoystickModule.cpp"
 switch ($16|0) {
 case 97:  {
  HEAP32[$agg$result>>2] = 0; //@line 366 "modules/joystick/sdl/JoystickModule.cpp"
  $17 = HEAP8[$bindvalues>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $18 = $17 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $19 = ($18<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $20 = ((($bindvalues)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $21 = HEAP32[$20>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $22 = ((($bindvalues)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $23 = $19 ? $22 : $21; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $24 = (_strtol($23,0,10)|0); //@line 367 "modules/joystick/sdl/JoystickModule.cpp"
  $25 = ((($agg$result)) + 4|0); //@line 367 "modules/joystick/sdl/JoystickModule.cpp"
  HEAP32[$25>>2] = $24; //@line 367 "modules/joystick/sdl/JoystickModule.cpp"
  break;
 }
 case 98:  {
  HEAP32[$agg$result>>2] = 1; //@line 370 "modules/joystick/sdl/JoystickModule.cpp"
  $28 = HEAP8[$bindvalues>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $29 = $28 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $30 = ($29<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $31 = ((($bindvalues)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $32 = HEAP32[$31>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $33 = ((($bindvalues)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $34 = $30 ? $33 : $32; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $35 = (_strtol($34,0,10)|0); //@line 371 "modules/joystick/sdl/JoystickModule.cpp"
  $36 = ((($agg$result)) + 4|0); //@line 371 "modules/joystick/sdl/JoystickModule.cpp"
  HEAP32[$36>>2] = $35; //@line 371 "modules/joystick/sdl/JoystickModule.cpp"
  break;
 }
 case 104:  {
  $37 = HEAP8[$bindvalues>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $38 = $37 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $39 = ($38<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $40 = ((($bindvalues)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $41 = HEAP32[$40>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $42 = $37&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $43 = $42 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $44 = $39 ? $43 : $41; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $45 = ($44>>>0)<(3); //@line 375 "modules/joystick/sdl/JoystickModule.cpp"
  if (!($45)) {
   HEAP32[$agg$result>>2] = 2; //@line 377 "modules/joystick/sdl/JoystickModule.cpp"
   __THREW__ = 0;
   invoke_viiiii(3511,($0|0),($bindvalues|0),0,1,($bindvalues|0)); //@line 3345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $46 = __THREW__; __THREW__ = 0;
   $47 = $46&1;
   if ($47) {
    $26 = ___cxa_find_matching_catch_2()|0;
    $27 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($bindvalues); //@line 392 "modules/joystick/sdl/JoystickModule.cpp"
    ___resumeException($26|0);
    // unreachable;
   }
   $48 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $49 = $48 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $50 = ($49<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $51 = ((($0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $52 = HEAP32[$51>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $53 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $54 = $50 ? $53 : $52; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $55 = (_strtol($54,0,10)|0); //@line 378 "modules/joystick/sdl/JoystickModule.cpp"
   $56 = ((($agg$result)) + 4|0); //@line 378 "modules/joystick/sdl/JoystickModule.cpp"
   HEAP32[$56>>2] = $55; //@line 378 "modules/joystick/sdl/JoystickModule.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 378 "modules/joystick/sdl/JoystickModule.cpp"
   __THREW__ = 0;
   invoke_viiiii(3511,($1|0),($bindvalues|0),2,-1,($bindvalues|0)); //@line 3345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $57 = __THREW__; __THREW__ = 0;
   $58 = $57&1;
   if ($58) {
    $26 = ___cxa_find_matching_catch_2()|0;
    $27 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($bindvalues); //@line 392 "modules/joystick/sdl/JoystickModule.cpp"
    ___resumeException($26|0);
    // unreachable;
   }
   $59 = HEAP8[$1>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $60 = $59 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $61 = ($60<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $62 = ((($1)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $63 = HEAP32[$62>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $64 = ((($1)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $65 = $61 ? $64 : $63; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $66 = (_strtol($65,0,10)|0); //@line 379 "modules/joystick/sdl/JoystickModule.cpp"
   $67 = $66&255; //@line 379 "modules/joystick/sdl/JoystickModule.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 379 "modules/joystick/sdl/JoystickModule.cpp"
   $68 = ((($agg$result)) + 8|0); //@line 380 "modules/joystick/sdl/JoystickModule.cpp"
   __THREW__ = 0;
   $69 = (invoke_iii(3951,($67|0),($68|0))|0); //@line 380 "modules/joystick/sdl/JoystickModule.cpp"
   $70 = __THREW__; __THREW__ = 0;
   $71 = $70&1;
   if ($71) {
    $26 = ___cxa_find_matching_catch_2()|0;
    $27 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($bindvalues); //@line 392 "modules/joystick/sdl/JoystickModule.cpp"
    ___resumeException($26|0);
    // unreachable;
   }
   if (!($69)) {
    HEAP32[$agg$result>>2] = 3; //@line 383 "modules/joystick/sdl/JoystickModule.cpp"
   }
  }
  break;
 }
 default: {
 }
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($bindvalues); //@line 392 "modules/joystick/sdl/JoystickModule.cpp"
 STACKTOP = sp;return; //@line 392 "modules/joystick/sdl/JoystickModule.cpp"
}
function __ZN4love8joystick3sdl14JoystickModule19loadGamepadMappingsERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($this,$mappings) {
 $this = $this|0;
 $mappings = $mappings|0;
 var $$0$i = 0, $$01 = 0, $$01$i = 0, $$016$i$i$i = 0, $$016$i$i$i$lcssa = 0, $$02 = 0, $$pre = 0, $$pre31 = 0, $$pre32 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__m1$0$i$i$i = 0, $__m2$0$i$i$i = 0, $eh$lpad$body$i$index2Z2D = 0, $eh$lpad$body$i$indexZ2D = 0, $eh$lpad$body$index9Z2D = 0, $eh$lpad$body$indexZ2D = 0, $guid = 0, $mapping = 0, $or$cond = 0, $platform = 0, $ss = 0, $success$0$off0$ph = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 184|0;
 $ss = sp + 40|0;
 $mapping = sp + 28|0;
 $platform = sp + 16|0;
 $guid = sp + 4|0;
 $1 = ((($ss)) + 64|0); //@line 908 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $2 = ((($ss)) + 8|0); //@line 908 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$2>>2] = (9320); //@line 908 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $3 = ((($ss)) + 12|0); //@line 906 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$ss>>2] = ((9524)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 HEAP32[$1>>2] = ((9544)); //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $4 = ((($ss)) + 4|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 HEAP32[$4>>2] = 0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
 $5 = ((($ss)) + 64|0); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 __THREW__ = 0;
 invoke_vii(3362,($5|0),($3|0)); //@line 668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $18 = ___cxa_find_matching_catch_2()|0;
  $19 = tempRet0;
  $$0$i = $19;$$01$i = $18;
  __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($1); //@line 909 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
  ___resumeException($$01$i|0);
  // unreachable;
 }
 $8 = ((($ss)) + 136|0); //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$8>>2] = 0; //@line 669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 $9 = ((($ss)) + 140|0); //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$9>>2] = -1; //@line 670 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
 HEAP32[$ss>>2] = (9300); //@line 908 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$1>>2] = (9340); //@line 908 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 HEAP32[$2>>2] = (9320); //@line 908 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 __THREW__ = 0;
 invoke_vi(3363,($3|0)); //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 do {
  if ($11) {
   $20 = ___cxa_find_matching_catch_2()|0;
   $21 = tempRet0;
   $eh$lpad$body$i$index2Z2D = $21;$eh$lpad$body$i$indexZ2D = $20;
  } else {
   HEAP32[$3>>2] = (9356); //@line 253 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $12 = ((($ss)) + 44|0); //@line 211 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $13 = ((($ss)) + 60|0); //@line 252 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   ;HEAP32[$12>>2]=0|0;HEAP32[$12+4>>2]=0|0;HEAP32[$12+8>>2]=0|0;HEAP32[$12+12>>2]=0|0; //@line 251 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   HEAP32[$13>>2] = 24; //@line 252 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   __THREW__ = 0;
   invoke_vii(3364,($3|0),($mappings|0)); //@line 254 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
   $14 = __THREW__; __THREW__ = 0;
   $15 = $14&1;
   if ($15) {
    $16 = ___cxa_find_matching_catch_2()|0;
    $17 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12); //@line 255 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($3); //@line 255 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
    $eh$lpad$body$i$index2Z2D = $17;$eh$lpad$body$i$indexZ2D = $16;
    break;
   }
   ;HEAP32[$mapping>>2]=0|0;HEAP32[$mapping+4>>2]=0|0;HEAP32[$mapping+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $22 = ((($mapping)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $23 = ((($mapping)) + 8|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $24 = ((($mapping)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $25 = ((($this)) + 32|0); //@line 520 "modules/joystick/sdl/JoystickModule.cpp"
   $success$0$off0$ph = 0;
   L11: while(1) {
    L13: while(1) {
     $26 = HEAP32[$ss>>2]|0; //@line 1652 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
     $27 = ((($26)) + -12|0); //@line 1652 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
     $28 = HEAP32[$27>>2]|0; //@line 1652 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
     $29 = (($ss) + ($28)|0); //@line 1652 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
     __THREW__ = 0;
     $30 = (invoke_ii(3373,($29|0))|0); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     $31 = __THREW__; __THREW__ = 0;
     $32 = $31&1;
     if ($32) {
      label = 18;
      break L11;
     }
     HEAP32[$0>>2] = $30; //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     __THREW__ = 0;
     $33 = (invoke_iii(3374,($0|0),(1207576|0))|0); //@line 181 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
     $34 = __THREW__; __THREW__ = 0;
     $35 = $34&1;
     if ($35) {
      label = 14;
      break L11;
     }
     $36 = HEAP32[$33>>2]|0; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
     $37 = ((($36)) + 28|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
     $38 = HEAP32[$37>>2]|0; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
     __THREW__ = 0;
     $39 = (invoke_iii($38|0,($33|0),10)|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__locale"
     $40 = __THREW__; __THREW__ = 0;
     $41 = $40&1;
     if ($41) {
      label = 14;
      break L11;
     }
     __ZNSt3__16localeD2Ev($0); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     __THREW__ = 0;
     $44 = (invoke_iiii(3516,($ss|0),($mapping|0),($39|0))|0); //@line 1652 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/istream"
     $45 = __THREW__; __THREW__ = 0;
     $46 = $45&1;
     if ($46) {
      label = 18;
      break L11;
     }
     $47 = HEAP32[$44>>2]|0; //@line 487 "modules/joystick/sdl/JoystickModule.cpp"
     $48 = ((($47)) + -12|0); //@line 487 "modules/joystick/sdl/JoystickModule.cpp"
     $49 = HEAP32[$48>>2]|0; //@line 487 "modules/joystick/sdl/JoystickModule.cpp"
     $50 = (($44) + ($49)|0); //@line 487 "modules/joystick/sdl/JoystickModule.cpp"
     $51 = ((($50)) + 16|0); //@line 545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     $52 = HEAP32[$51>>2]|0; //@line 545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     $53 = $52 & 5; //@line 545 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     $54 = ($53|0)==(0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     if (!($54)) {
      label = 45;
      break L11;
     }
     $55 = HEAP8[$mapping>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $56 = $55 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $57 = ($56<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $58 = HEAP32[$22>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $59 = $55&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $60 = $59 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $61 = $57 ? $60 : $58; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $62 = ($61|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($62) {
      continue;
     }
     $67 = HEAP32[$23>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $68 = $57 ? $24 : $67; //@line 1755 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $69 = HEAP8[$68>>0]|0; //@line 493 "modules/joystick/sdl/JoystickModule.cpp"
     $70 = ($69<<24>>24)==(35); //@line 493 "modules/joystick/sdl/JoystickModule.cpp"
     if ($70) {
      continue;
     }
     $71 = ($61>>>0)<(9); //@line 962 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     L22: do {
      if ($71) {
       $108 = $55;$111 = $67;
      } else {
       $72 = (($68) + ($61)|0); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $73 = $68; //@line 1466 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $74 = ($61|0)<(9); //@line 1467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       if ($74) {
        $108 = $55;$111 = $67;
       } else {
        $75 = ((($72)) + -8|0); //@line 1469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $76 = ($75|0)==($68|0); //@line 1475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        if ($76) {
         $108 = $55;$111 = $67;
        } else {
         $$016$i$i$i = $68;$78 = $69;
         L25: while(1) {
          $77 = ($78<<24>>24)==(112); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          if ($77) {
           $__m1$0$i$i$i = $$016$i$i$i;$__m2$0$i$i$i = 441914;
           while(1) {
            $81 = ((($__m2$0$i$i$i)) + 1|0); //@line 1517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
            $82 = ($81|0)==((441923)|0); //@line 1517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
            if ($82) {
             $$016$i$i$i$lcssa = $$016$i$i$i;
             break L25;
            }
            $83 = ((($__m1$0$i$i$i)) + 1|0); //@line 1519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
            $84 = HEAP8[$83>>0]|0; //@line 1520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
            $85 = HEAP8[$81>>0]|0; //@line 1520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
            $86 = ($84<<24>>24)==($85<<24>>24); //@line 634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            if ($86) {
             $__m1$0$i$i$i = $83;$__m2$0$i$i$i = $81;
            } else {
             break;
            }
           }
          }
          $79 = ((($$016$i$i$i)) + 1|0); //@line 1479 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
          $80 = ($79|0)==($75|0); //@line 1475 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
          if ($80) {
           $108 = $55;$111 = $67;
           break L22;
          }
          $$pre = HEAP8[$79>>0]|0; //@line 1477 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
          $$016$i$i$i = $79;$78 = $$pre;
         }
         $87 = ($$016$i$i$i$lcssa|0)==($72|0); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $88 = $$016$i$i$i$lcssa; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $89 = (($88) - ($73))|0; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $90 = ($89|0)==(-1); //@line 498 "modules/joystick/sdl/JoystickModule.cpp"
         $or$cond = $87 | $90;
         if ($or$cond) {
          $108 = $55;$111 = $67;
         } else {
          $91 = (($89) + 9)|0; //@line 500 "modules/joystick/sdl/JoystickModule.cpp"
          $92 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE4findEcj($mapping,44,$91)|0); //@line 3501 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $93 = (($92) - ($91))|0; //@line 503 "modules/joystick/sdl/JoystickModule.cpp"
          __THREW__ = 0;
          invoke_viiiii(3511,($platform|0),($mapping|0),($91|0),($93|0),($mapping|0)); //@line 3345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $94 = __THREW__; __THREW__ = 0;
          $95 = $94&1;
          if ($95) {
           label = 18;
           break L11;
          }
          __THREW__ = 0;
          $96 = (invoke_i(4197)|0); //@line 505 "modules/joystick/sdl/JoystickModule.cpp"
          $97 = __THREW__; __THREW__ = 0;
          $98 = $97&1;
          if ($98) {
           label = 34;
           break L11;
          }
          $99 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7compareEPKc($platform,$96)|0); //@line 505 "modules/joystick/sdl/JoystickModule.cpp"
          $100 = ($99|0)==(0); //@line 505 "modules/joystick/sdl/JoystickModule.cpp"
          if (!($100)) {
           label = 37;
           break L13;
          }
          $103 = (1 - ($89))|0; //@line 513 "modules/joystick/sdl/JoystickModule.cpp"
          $104 = (($103) + ($92))|0; //@line 513 "modules/joystick/sdl/JoystickModule.cpp"
          __THREW__ = 0;
          (invoke_iiii(4117,($mapping|0),($89|0),($104|0))|0); //@line 513 "modules/joystick/sdl/JoystickModule.cpp"
          $105 = __THREW__; __THREW__ = 0;
          $106 = $105&1;
          if ($106) {
           label = 34;
           break L11;
          }
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($platform); //@line 514 "modules/joystick/sdl/JoystickModule.cpp"
          $$pre31 = HEAP8[$mapping>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $$pre32 = HEAP32[$23>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $108 = $$pre31;$111 = $$pre32;
         }
        }
       }
      }
     } while(0);
     $107 = $108 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $109 = ($107<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $110 = $109 ? $24 : $111; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     __THREW__ = 0;
     $112 = (invoke_ii(4193,($110|0))|0); //@line 516 "modules/joystick/sdl/JoystickModule.cpp"
     $113 = __THREW__; __THREW__ = 0;
     $114 = $113&1;
     if ($114) {
      label = 18;
      break L11;
     }
     $115 = ($112|0)==(-1); //@line 516 "modules/joystick/sdl/JoystickModule.cpp"
     if (!($115)) {
      break;
     }
    }
    if ((label|0) == 37) {
     label = 0;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($platform); //@line 514 "modules/joystick/sdl/JoystickModule.cpp"
     $success$0$off0$ph = 1;
     continue;
    }
    $116 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE4findEcj($mapping,44,0)|0); //@line 3501 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    __THREW__ = 0;
    invoke_viiiii(3511,($guid|0),($mapping|0),0,($116|0),($mapping|0)); //@line 3345 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $117 = __THREW__; __THREW__ = 0;
    $118 = $117&1;
    if ($118) {
     label = 19;
     break;
    }
    __THREW__ = 0;
    $119 = (invoke_iii(4194,($25|0),($guid|0))|0); //@line 520 "modules/joystick/sdl/JoystickModule.cpp"
    $120 = __THREW__; __THREW__ = 0;
    $121 = $120&1;
    if ($121) {
     label = 44;
     break;
    }
    HEAP8[$119>>0] = 1; //@line 520 "modules/joystick/sdl/JoystickModule.cpp"
    __THREW__ = 0;
    invoke_vii(4195,($this|0),($guid|0)); //@line 524 "modules/joystick/sdl/JoystickModule.cpp"
    $122 = __THREW__; __THREW__ = 0;
    $123 = $122&1;
    if ($123) {
     label = 44;
     break;
    }
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($guid); //@line 525 "modules/joystick/sdl/JoystickModule.cpp"
    $success$0$off0$ph = 1;
   }
   do {
    if ((label|0) == 14) {
     $42 = ___cxa_find_matching_catch_2()|0;
     $43 = tempRet0;
     __ZNSt3__16localeD2Ev($0); //@line 734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/ios"
     $eh$lpad$body$index9Z2D = $43;$eh$lpad$body$indexZ2D = $42;
     label = 20;
    }
    else if ((label|0) == 18) {
     $63 = ___cxa_find_matching_catch_2()|0;
     $64 = tempRet0;
     $eh$lpad$body$index9Z2D = $64;$eh$lpad$body$indexZ2D = $63;
     label = 20;
    }
    else if ((label|0) == 19) {
     $65 = ___cxa_find_matching_catch_2()|0;
     $66 = tempRet0;
     $eh$lpad$body$index9Z2D = $66;$eh$lpad$body$indexZ2D = $65;
     label = 20;
    }
    else if ((label|0) == 34) {
     $101 = ___cxa_find_matching_catch_2()|0;
     $102 = tempRet0;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($platform); //@line 514 "modules/joystick/sdl/JoystickModule.cpp"
     $$01 = $102;$$02 = $101;
    }
    else if ((label|0) == 44) {
     $124 = ___cxa_find_matching_catch_2()|0;
     $125 = tempRet0;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($guid); //@line 525 "modules/joystick/sdl/JoystickModule.cpp"
     $$01 = $125;$$02 = $124;
    }
    else if ((label|0) == 45) {
     if ($success$0$off0$ph) {
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($mapping); //@line 530 "modules/joystick/sdl/JoystickModule.cpp"
      HEAP32[$ss>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      HEAP32[$1>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      $134 = ((($ss)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      HEAP32[$134>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      HEAP32[$3>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($3); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($1); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
      STACKTOP = sp;return; //@line 530 "modules/joystick/sdl/JoystickModule.cpp"
     }
     $126 = (___cxa_allocate_exception(16)|0); //@line 529 "modules/joystick/sdl/JoystickModule.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($126|0),(441924|0),($vararg_buffer|0)); //@line 529 "modules/joystick/sdl/JoystickModule.cpp"
     $127 = __THREW__; __THREW__ = 0;
     $128 = $127&1;
     if ($128) {
      $132 = ___cxa_find_matching_catch_2()|0;
      $133 = tempRet0;
      ___cxa_free_exception(($126|0)); //@line 529 "modules/joystick/sdl/JoystickModule.cpp"
      $$01 = $133;$$02 = $132;
      break;
     } else {
      __THREW__ = 0;
      invoke_viii(3350,($126|0),(96|0),(344|0)); //@line 529 "modules/joystick/sdl/JoystickModule.cpp"
      $129 = __THREW__; __THREW__ = 0;
      $130 = ___cxa_find_matching_catch_2()|0;
      $131 = tempRet0;
      $eh$lpad$body$index9Z2D = $131;$eh$lpad$body$indexZ2D = $130;
      label = 20;
      break;
     }
    }
   } while(0);
   if ((label|0) == 20) {
    $$01 = $eh$lpad$body$index9Z2D;$$02 = $eh$lpad$body$indexZ2D;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($mapping); //@line 530 "modules/joystick/sdl/JoystickModule.cpp"
   HEAP32[$ss>>2] = ((9300)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$1>>2] = ((9340)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   $135 = ((($ss)) + 8|0); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$135>>2] = ((9320)); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   HEAP32[$3>>2] = (9356); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($3); //@line 117 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($1); //@line 126 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iosfwd"
   ___resumeException($$02|0);
   // unreachable;
  }
 } while(0);
 $$0$i = $eh$lpad$body$i$index2Z2D;$$01$i = $eh$lpad$body$i$indexZ2D;
 __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($1); //@line 909 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/sstream"
 ___resumeException($$01$i|0);
 // unreachable;
}
function __ZN4love8joystick3sdl14JoystickModule19saveGamepadMappingsEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$0 = 0, $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$01$i$i$i = 0, $$03 = 0, $$1 = 0, $$14 = 0, $$2 = 0, $$25 = 0, $$3 = 0, $$36 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $__begin$sroa$0$010 = 0, $mapping = 0, $sdlguid = 0, $sdlguid$byval_copy = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $sdlguid$byval_copy = sp + 64|0;
 $sdlguid = sp + 48|0;
 $mapping = sp + 36|0;
 $0 = sp + 24|0;
 $1 = sp + 12|0;
 $2 = sp;
 ;HEAP32[$agg$result>>2]=0|0;HEAP32[$agg$result+4>>2]=0|0;HEAP32[$agg$result+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $3 = ((($this)) + 32|0); //@line 854 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $4 = HEAP32[$3>>2]|0; //@line 911 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $5 = ((($this)) + 36|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = ($4|0)==($5|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($6) {
  STACKTOP = sp;return; //@line 556 "modules/joystick/sdl/JoystickModule.cpp"
 }
 $7 = ((($mapping)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = ((($0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = ((($0)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = ((($0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = ((($mapping)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = ((($mapping)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $__begin$sroa$0$010 = $4;
 while(1) {
  $15 = ((($__begin$sroa$0$010)) + 16|0); //@line 538 "modules/joystick/sdl/JoystickModule.cpp"
  $16 = HEAP8[$15>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $17 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $18 = ($17<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($18) {
   $21 = ((($15)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $22 = $21;
  } else {
   $19 = ((($__begin$sroa$0$010)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $20 = HEAP32[$19>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $22 = $20;
  }
  __THREW__ = 0;
  invoke_vii(4198,($sdlguid|0),($22|0)); //@line 538 "modules/joystick/sdl/JoystickModule.cpp"
  $23 = __THREW__; __THREW__ = 0;
  $24 = $23&1;
  if ($24) {
   label = 4;
   break;
  }
  __THREW__ = 0;
  dest=$sdlguid$byval_copy; src=$sdlguid; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 540 "modules/joystick/sdl/JoystickModule.cpp"
  $25 = (invoke_ii(4189,($sdlguid$byval_copy|0))|0); //@line 540 "modules/joystick/sdl/JoystickModule.cpp"
  $26 = __THREW__; __THREW__ = 0;
  $27 = $26&1;
  if ($27) {
   label = 4;
   break;
  }
  $28 = ($25|0)==(0|0); //@line 541 "modules/joystick/sdl/JoystickModule.cpp"
  if ($28) {
  } else {
   $29 = (_strlen($25)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   invoke_viii(3460,($mapping|0),($25|0),($29|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $30 = __THREW__; __THREW__ = 0;
   $31 = $30&1;
   if ($31) {
    label = 4;
    break;
   }
   __THREW__ = 0;
   invoke_vi(3775,($25|0)); //@line 545 "modules/joystick/sdl/JoystickModule.cpp"
   $32 = __THREW__; __THREW__ = 0;
   $33 = $32&1;
   if ($33) {
    label = 15;
    break;
   }
   $34 = (__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE5rfindEcj($mapping,44,-1)|0); //@line 3544 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $35 = HEAP8[$mapping>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $36 = $35 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $37 = ($36<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $38 = HEAP32[$7>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $39 = $35&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $40 = $39 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $41 = $37 ? $40 : $38; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $42 = (($41) + -1)|0; //@line 547 "modules/joystick/sdl/JoystickModule.cpp"
   $43 = ($34|0)==($42|0); //@line 547 "modules/joystick/sdl/JoystickModule.cpp"
   if (!($43)) {
    __THREW__ = 0;
    (invoke_iii(3806,($mapping|0),(442475|0))|0); //@line 1454 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $44 = __THREW__; __THREW__ = 0;
    $45 = $44&1;
    if ($45) {
     label = 15;
     break;
    }
   }
   __THREW__ = 0;
   $48 = (invoke_i(4197)|0); //@line 551 "modules/joystick/sdl/JoystickModule.cpp"
   $49 = __THREW__; __THREW__ = 0;
   $50 = $49&1;
   if ($50) {
    label = 15;
    break;
   }
   $51 = (_strlen($48)|0); //@line 640 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   invoke_viii(3460,($2|0),($48|0),($51|0)); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $52 = __THREW__; __THREW__ = 0;
   $53 = $52&1;
   if ($53) {
    label = 15;
    break;
   }
   __THREW__ = 0;
   $54 = (invoke_iiii(3640,($2|0),0,(441914|0))|0); //@line 4027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $55 = __THREW__; __THREW__ = 0;
   $56 = $55&1;
   if ($56) {
    label = 29;
    break;
   }
   ;HEAP32[$1>>2]=HEAP32[$54>>2]|0;HEAP32[$1+4>>2]=HEAP32[$54+4>>2]|0;HEAP32[$1+8>>2]=HEAP32[$54+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   ;HEAP32[$54>>2]=0|0;HEAP32[$54+4>>2]=0|0;HEAP32[$54+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   $57 = (invoke_iii(3806,($1|0),(441950|0))|0); //@line 4044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $58 = __THREW__; __THREW__ = 0;
   $59 = $58&1;
   if ($59) {
    label = 30;
    break;
   }
   ;HEAP32[$0>>2]=HEAP32[$57>>2]|0;HEAP32[$0+4>>2]=HEAP32[$57+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$57+8>>2]|0; //@line 2074 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   ;HEAP32[$57>>2]=0|0;HEAP32[$57+4>>2]=0|0;HEAP32[$57+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $60 = HEAP8[$0>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $61 = $60 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $62 = ($61<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $63 = HEAP32[$8>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $64 = $62 ? $9 : $63; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $65 = HEAP32[$10>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $66 = $60&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $67 = $66 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $68 = $62 ? $67 : $65; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   (invoke_iiii(3462,($mapping|0),($64|0),($68|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $69 = __THREW__; __THREW__ = 0;
   $70 = $69&1;
   if ($70) {
    label = 31;
    break;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 551 "modules/joystick/sdl/JoystickModule.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 551 "modules/joystick/sdl/JoystickModule.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 551 "modules/joystick/sdl/JoystickModule.cpp"
   $71 = HEAP8[$mapping>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $72 = $71 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $73 = ($72<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $74 = HEAP32[$11>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $75 = $73 ? $12 : $74; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $76 = HEAP32[$7>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $77 = $71&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $78 = $77 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $79 = $73 ? $78 : $76; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   __THREW__ = 0;
   (invoke_iiii(3462,($agg$result|0),($75|0),($79|0))|0); //@line 2653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $80 = __THREW__; __THREW__ = 0;
   $81 = $80&1;
   if ($81) {
    label = 15;
    break;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($mapping); //@line 553 "modules/joystick/sdl/JoystickModule.cpp"
  }
  $82 = ((($__begin$sroa$0$010)) + 4|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $83 = HEAP32[$82>>2]|0; //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $84 = ($83|0)==(0|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($84) {
   $$0$i$i$i = $__begin$sroa$0$010;
   while(1) {
    $87 = ((($$0$i$i$i)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $88 = HEAP32[$87>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $89 = HEAP32[$88>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $90 = ($89|0)==($$0$i$i$i|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($90) {
     $$01$i$i$i = $88;
     break;
    } else {
     $$0$i$i$i = $88;
    }
   }
  } else {
   $$0$i$i$i$i = $83;
   while(1) {
    $85 = HEAP32[$$0$i$i$i$i>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $86 = ($85|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($86) {
     $$01$i$i$i = $$0$i$i$i$i;
     break;
    } else {
     $$0$i$i$i$i = $85;
    }
   }
  }
  $91 = ($$01$i$i$i|0)==($5|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if ($91) {
   label = 3;
   break;
  } else {
   $__begin$sroa$0$010 = $$01$i$i$i;
  }
 }
 if ((label|0) == 3) {
  STACKTOP = sp;return; //@line 556 "modules/joystick/sdl/JoystickModule.cpp"
 }
 else if ((label|0) == 4) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  $$3 = $13;$$36 = $14;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($agg$result); //@line 556 "modules/joystick/sdl/JoystickModule.cpp"
  ___resumeException($$3|0);
  // unreachable;
 }
 else if ((label|0) == 15) {
  $46 = ___cxa_find_matching_catch_2()|0;
  $47 = tempRet0;
  $$2 = $46;$$25 = $47;
 }
 else if ((label|0) == 29) {
  $92 = ___cxa_find_matching_catch_2()|0;
  $93 = tempRet0;
  $$1 = $92;$$14 = $93;
  label = 33;
 }
 else if ((label|0) == 30) {
  $94 = ___cxa_find_matching_catch_2()|0;
  $95 = tempRet0;
  $$0 = $94;$$03 = $95;
  label = 32;
 }
 else if ((label|0) == 31) {
  $96 = ___cxa_find_matching_catch_2()|0;
  $97 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 551 "modules/joystick/sdl/JoystickModule.cpp"
  $$0 = $96;$$03 = $97;
  label = 32;
 }
 if ((label|0) == 32) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 551 "modules/joystick/sdl/JoystickModule.cpp"
  $$1 = $$0;$$14 = $$03;
  label = 33;
 }
 if ((label|0) == 33) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 551 "modules/joystick/sdl/JoystickModule.cpp"
  $$2 = $$1;$$25 = $$14;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($mapping); //@line 553 "modules/joystick/sdl/JoystickModule.cpp"
 $$3 = $$2;$$36 = $$25;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($agg$result); //@line 556 "modules/joystick/sdl/JoystickModule.cpp"
 ___resumeException($$3|0);
 // unreachable;
}
function __ZN4love8joystick3sdl14JoystickModuleC2Ev($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $__f$01$i$i = 0, $i$0 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./common/Module.h"
 HEAP32[$this>>2] = (26484); //@line 43 "modules/joystick/sdl/JoystickModule.cpp"
 $0 = ((($this)) + 8|0); //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($this)) + 12|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($this)) + 16|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = ((($this)) + 20|0); //@line 42 "modules/joystick/sdl/JoystickModule.cpp"
 HEAP32[$3>>2] = $3; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $4 = ((($this)) + 24|0); //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 HEAP32[$4>>2] = $3; //@line 213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $5 = ((($this)) + 28|0); //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$5>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $6 = ((($this)) + 32|0); //@line 42 "modules/joystick/sdl/JoystickModule.cpp"
 $7 = ((($this)) + 36|0); //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$7>>2] = 0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $8 = ((($this)) + 40|0); //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$8>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = ((($this)) + 36|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$6>>2] = $9; //@line 1130 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __THREW__ = 0;
 $10 = (invoke_ii(3813,8704)|0); //@line 44 "modules/joystick/sdl/JoystickModule.cpp"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 do {
  if ($12) {
   label = 8;
  } else {
   $13 = ($10|0)<(0); //@line 44 "modules/joystick/sdl/JoystickModule.cpp"
   if ($13) {
    $14 = (___cxa_allocate_exception(16)|0); //@line 45 "modules/joystick/sdl/JoystickModule.cpp"
    __THREW__ = 0;
    $15 = (invoke_i(3814)|0); //@line 45 "modules/joystick/sdl/JoystickModule.cpp"
    $16 = __THREW__; __THREW__ = 0;
    $17 = $16&1;
    if (!($17)) {
     __THREW__ = 0;
     HEAP32[$vararg_buffer>>2] = $15; //@line 45 "modules/joystick/sdl/JoystickModule.cpp"
     invoke_viii(3349,($14|0),(441953|0),($vararg_buffer|0)); //@line 45 "modules/joystick/sdl/JoystickModule.cpp"
     $18 = __THREW__; __THREW__ = 0;
     $19 = $18&1;
     if (!($19)) {
      __THREW__ = 0;
      invoke_viii(3350,($14|0),(96|0),(344|0)); //@line 45 "modules/joystick/sdl/JoystickModule.cpp"
      $20 = __THREW__; __THREW__ = 0;
      label = 8;
      break;
     }
    }
    $25 = ___cxa_find_matching_catch_2()|0;
    $26 = tempRet0;
    ___cxa_free_exception(($14|0)); //@line 45 "modules/joystick/sdl/JoystickModule.cpp"
    $$0 = $26;$$01 = $25;
    break;
   }
   $i$0 = 0;
   while(1) {
    __THREW__ = 0;
    $27 = (invoke_i(4199)|0); //@line 48 "modules/joystick/sdl/JoystickModule.cpp"
    $28 = __THREW__; __THREW__ = 0;
    $29 = $28&1;
    if ($29) {
     label = 7;
     break;
    }
    $30 = ($i$0|0)<($27|0); //@line 48 "modules/joystick/sdl/JoystickModule.cpp"
    if (!($30)) {
     break;
    }
    $33 = HEAP32[$this>>2]|0; //@line 49 "modules/joystick/sdl/JoystickModule.cpp"
    $34 = ((($33)) + 24|0); //@line 49 "modules/joystick/sdl/JoystickModule.cpp"
    $35 = HEAP32[$34>>2]|0; //@line 49 "modules/joystick/sdl/JoystickModule.cpp"
    __THREW__ = 0;
    (invoke_iii($35|0,($this|0),($i$0|0))|0); //@line 49 "modules/joystick/sdl/JoystickModule.cpp"
    $36 = __THREW__; __THREW__ = 0;
    $37 = $36&1;
    if ($37) {
     label = 7;
     break;
    }
    $38 = (($i$0) + 1)|0; //@line 48 "modules/joystick/sdl/JoystickModule.cpp"
    $i$0 = $38;
   }
   if ((label|0) == 7) {
    $21 = ___cxa_find_matching_catch_2()|0;
    $22 = tempRet0;
    $lpad$phi$index = $21;$lpad$phi$index2 = $22;
    label = 9;
    break;
   }
   __THREW__ = 0;
   (invoke_ii(4200,1)|0); //@line 53 "modules/joystick/sdl/JoystickModule.cpp"
   $31 = __THREW__; __THREW__ = 0;
   $32 = $31&1;
   if ($32) {
    label = 8;
   } else {
    __THREW__ = 0;
    (invoke_ii(4201,1)|0); //@line 54 "modules/joystick/sdl/JoystickModule.cpp"
    $39 = __THREW__; __THREW__ = 0;
    $40 = $39&1;
    if ($40) {
     label = 8;
    } else {
     STACKTOP = sp;return; //@line 55 "modules/joystick/sdl/JoystickModule.cpp"
    }
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $23 = ___cxa_find_matching_catch_2()|0;
  $24 = tempRet0;
  $lpad$phi$index = $23;$lpad$phi$index2 = $24;
  label = 9;
 }
 if ((label|0) == 9) {
  $$0 = $lpad$phi$index2;$$01 = $lpad$phi$index;
 }
 $41 = HEAP32[$9>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEbEENS_19__map_value_compareIS7_S8_NS_4lessIS7_EELb1EEENS5_IS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($6,$41); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $42 = HEAP32[$5>>2]|0; //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $43 = ($42|0)==(0); //@line 564 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 if (!($43)) {
  $44 = HEAP32[$4>>2]|0; //@line 710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $45 = HEAP32[$3>>2]|0; //@line 713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $46 = ((($45)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $47 = HEAP32[$46>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $48 = HEAP32[$44>>2]|0; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $49 = ((($48)) + 4|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$49>>2] = $47; //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $50 = $48; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $51 = HEAP32[$46>>2]|0; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$51>>2] = $50; //@line 676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  HEAP32[$5>>2] = 0; //@line 714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  $52 = ($44|0)==($3|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
  if (!($52)) {
   $__f$01$i$i = $44;
   while(1) {
    $53 = ((($__f$01$i$i)) + 4|0); //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $54 = HEAP32[$53>>2]|0; //@line 718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    __ZdlPv($__f$01$i$i); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $55 = ($54|0)==($3|0); //@line 715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($55) {
     break;
    } else {
     $__f$01$i$i = $54;
    }
   }
  }
 }
 $56 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $57 = ($56|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $58 = $56; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($57) {
  __ZN4love6ModuleD2Ev($this); //@line 55 "modules/joystick/sdl/JoystickModule.cpp"
  ___resumeException($$01|0);
  // unreachable;
 }
 $59 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $60 = ($59|0)==($56|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($60)) {
  $scevgep$i$i$i = ((($59)) + -4|0);
  $61 = $scevgep$i$i$i;
  $62 = (($61) - ($58))|0;
  $63 = $62 >>> 2;
  $64 = $63 ^ -1;
  $scevgep4$i$i$i = (($59) + ($64<<2)|0);
  HEAP32[$1>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($56); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 __ZN4love6ModuleD2Ev($this); //@line 55 "modules/joystick/sdl/JoystickModule.cpp"
 ___resumeException($$01|0);
 // unreachable;
}
function __GLOBAL__sub_I_Joystick_cpp_4336() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love9StringMapINS_8joystick8Joystick3HatELj16EEC2EPKNS4_5EntryEj(635704,26548,72); //@line 96 "modules/joystick/Joystick.cpp"
 HEAP8[(636160)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636172)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636184)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636196)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636208)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636220)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636232)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636244)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636256)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636268)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636280)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636292)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636304)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636316)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(636320)>>2]=0|0;HEAP32[(636320)+4>>2]=0|0;HEAP32[(636320)+8>>2]=0|0;HEAP32[(636320)+12>>2]=0|0;HEAP32[(636320)+16>>2]=0|0;HEAP32[(636320)+20>>2]=0|0;HEAP32[(636320)+24>>2]=0|0; //@line 44 "./common/StringMap.h"
 $0 = HEAP32[6655]|0; //@line 49 "./common/StringMap.h"
 $1 = HEAP32[(26624)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8joystick8Joystick11GamepadAxisELj7EE3addEPKcS3_(636152,$0,$1)|0); //@line 49 "./common/StringMap.h"
 $2 = HEAP32[(26628)>>2]|0; //@line 49 "./common/StringMap.h"
 $3 = HEAP32[(26632)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8joystick8Joystick11GamepadAxisELj7EE3addEPKcS3_(636152,$2,$3)|0); //@line 49 "./common/StringMap.h"
 $4 = HEAP32[(26636)>>2]|0; //@line 49 "./common/StringMap.h"
 $5 = HEAP32[(26640)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8joystick8Joystick11GamepadAxisELj7EE3addEPKcS3_(636152,$4,$5)|0); //@line 49 "./common/StringMap.h"
 $6 = HEAP32[(26644)>>2]|0; //@line 49 "./common/StringMap.h"
 $7 = HEAP32[(26648)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8joystick8Joystick11GamepadAxisELj7EE3addEPKcS3_(636152,$6,$7)|0); //@line 49 "./common/StringMap.h"
 $8 = HEAP32[(26652)>>2]|0; //@line 49 "./common/StringMap.h"
 $9 = HEAP32[(26656)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8joystick8Joystick11GamepadAxisELj7EE3addEPKcS3_(636152,$8,$9)|0); //@line 49 "./common/StringMap.h"
 $10 = HEAP32[(26660)>>2]|0; //@line 49 "./common/StringMap.h"
 $11 = HEAP32[(26664)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8joystick8Joystick11GamepadAxisELj7EE3addEPKcS3_(636152,$10,$11)|0); //@line 49 "./common/StringMap.h"
 __ZN4love9StringMapINS_8joystick8Joystick13GamepadButtonELj16EEC2EPKNS4_5EntryEj(636348,26668,120); //@line 129 "modules/joystick/Joystick.cpp"
 HEAP8[(636804)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636816)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636828)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636840)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636852)>>0] = 0; //@line 155 "./common/StringMap.h"
 HEAP8[(636864)>>0] = 0; //@line 155 "./common/StringMap.h"
 ;HEAP32[(636868)>>2]=0|0;HEAP32[(636868)+4>>2]=0|0;HEAP32[(636868)+8>>2]=0|0; //@line 44 "./common/StringMap.h"
 $12 = HEAP32[6697]|0; //@line 49 "./common/StringMap.h"
 $13 = HEAP32[(26792)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8joystick8Joystick9InputTypeELj3EE3addEPKcS3_(636796,$12,$13)|0); //@line 49 "./common/StringMap.h"
 $14 = HEAP32[(26796)>>2]|0; //@line 49 "./common/StringMap.h"
 $15 = HEAP32[(26800)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8joystick8Joystick9InputTypeELj3EE3addEPKcS3_(636796,$14,$15)|0); //@line 49 "./common/StringMap.h"
 $16 = HEAP32[(26804)>>2]|0; //@line 49 "./common/StringMap.h"
 $17 = HEAP32[(26808)>>2]|0; //@line 49 "./common/StringMap.h"
 (__ZN4love9StringMapINS_8joystick8Joystick9InputTypeELj3EE3addEPKcS3_(636796,$16,$17)|0); //@line 49 "./common/StringMap.h"
 return;
}
function __ZN4love9StringMapINS_8joystick8Joystick3HatELj16EEC2EPKNS4_5EntryEj($this,$entries,$num) {
 $this = $this|0;
 $entries = $entries|0;
 $num = $num|0;
 var $$02$i$i = 0, $$lcssa26 = 0, $$lcssa28 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $exitcond = 0, $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$08$i = 0, $i1$05 = 0, $scevgep = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 8|0); //@line 155 "./common/StringMap.h"
 HEAP8[$0>>0] = 0; //@line 155 "./common/StringMap.h"
 $1 = ((($this)) + 20|0); //@line 155 "./common/StringMap.h"
 HEAP8[$1>>0] = 0; //@line 155 "./common/StringMap.h"
 $2 = ((($this)) + 32|0); //@line 155 "./common/StringMap.h"
 HEAP8[$2>>0] = 0; //@line 155 "./common/StringMap.h"
 $3 = ((($this)) + 44|0); //@line 155 "./common/StringMap.h"
 HEAP8[$3>>0] = 0; //@line 155 "./common/StringMap.h"
 $4 = ((($this)) + 56|0); //@line 155 "./common/StringMap.h"
 HEAP8[$4>>0] = 0; //@line 155 "./common/StringMap.h"
 $5 = ((($this)) + 68|0); //@line 155 "./common/StringMap.h"
 HEAP8[$5>>0] = 0; //@line 155 "./common/StringMap.h"
 $6 = ((($this)) + 80|0); //@line 155 "./common/StringMap.h"
 HEAP8[$6>>0] = 0; //@line 155 "./common/StringMap.h"
 $7 = ((($this)) + 92|0); //@line 155 "./common/StringMap.h"
 HEAP8[$7>>0] = 0; //@line 155 "./common/StringMap.h"
 $8 = ((($this)) + 104|0); //@line 155 "./common/StringMap.h"
 HEAP8[$8>>0] = 0; //@line 155 "./common/StringMap.h"
 $9 = ((($this)) + 116|0); //@line 155 "./common/StringMap.h"
 HEAP8[$9>>0] = 0; //@line 155 "./common/StringMap.h"
 $10 = ((($this)) + 128|0); //@line 155 "./common/StringMap.h"
 HEAP8[$10>>0] = 0; //@line 155 "./common/StringMap.h"
 $11 = ((($this)) + 140|0); //@line 155 "./common/StringMap.h"
 HEAP8[$11>>0] = 0; //@line 155 "./common/StringMap.h"
 $12 = ((($this)) + 152|0); //@line 155 "./common/StringMap.h"
 HEAP8[$12>>0] = 0; //@line 155 "./common/StringMap.h"
 $13 = ((($this)) + 164|0); //@line 155 "./common/StringMap.h"
 HEAP8[$13>>0] = 0; //@line 155 "./common/StringMap.h"
 $14 = ((($this)) + 176|0); //@line 155 "./common/StringMap.h"
 HEAP8[$14>>0] = 0; //@line 155 "./common/StringMap.h"
 $15 = ((($this)) + 188|0); //@line 155 "./common/StringMap.h"
 HEAP8[$15>>0] = 0; //@line 155 "./common/StringMap.h"
 $16 = ((($this)) + 200|0); //@line 155 "./common/StringMap.h"
 HEAP8[$16>>0] = 0; //@line 155 "./common/StringMap.h"
 $17 = ((($this)) + 212|0); //@line 155 "./common/StringMap.h"
 HEAP8[$17>>0] = 0; //@line 155 "./common/StringMap.h"
 $18 = ((($this)) + 224|0); //@line 155 "./common/StringMap.h"
 HEAP8[$18>>0] = 0; //@line 155 "./common/StringMap.h"
 $19 = ((($this)) + 236|0); //@line 155 "./common/StringMap.h"
 HEAP8[$19>>0] = 0; //@line 155 "./common/StringMap.h"
 $20 = ((($this)) + 248|0); //@line 155 "./common/StringMap.h"
 HEAP8[$20>>0] = 0; //@line 155 "./common/StringMap.h"
 $21 = ((($this)) + 260|0); //@line 155 "./common/StringMap.h"
 HEAP8[$21>>0] = 0; //@line 155 "./common/StringMap.h"
 $22 = ((($this)) + 272|0); //@line 155 "./common/StringMap.h"
 HEAP8[$22>>0] = 0; //@line 155 "./common/StringMap.h"
 $23 = ((($this)) + 284|0); //@line 155 "./common/StringMap.h"
 HEAP8[$23>>0] = 0; //@line 155 "./common/StringMap.h"
 $24 = ((($this)) + 296|0); //@line 155 "./common/StringMap.h"
 HEAP8[$24>>0] = 0; //@line 155 "./common/StringMap.h"
 $25 = ((($this)) + 308|0); //@line 155 "./common/StringMap.h"
 HEAP8[$25>>0] = 0; //@line 155 "./common/StringMap.h"
 $26 = ((($this)) + 320|0); //@line 155 "./common/StringMap.h"
 HEAP8[$26>>0] = 0; //@line 155 "./common/StringMap.h"
 $27 = ((($this)) + 332|0); //@line 155 "./common/StringMap.h"
 HEAP8[$27>>0] = 0; //@line 155 "./common/StringMap.h"
 $28 = ((($this)) + 344|0); //@line 155 "./common/StringMap.h"
 HEAP8[$28>>0] = 0; //@line 155 "./common/StringMap.h"
 $29 = ((($this)) + 356|0); //@line 155 "./common/StringMap.h"
 HEAP8[$29>>0] = 0; //@line 155 "./common/StringMap.h"
 $30 = ((($this)) + 368|0); //@line 155 "./common/StringMap.h"
 HEAP8[$30>>0] = 0; //@line 155 "./common/StringMap.h"
 $31 = ((($this)) + 380|0); //@line 155 "./common/StringMap.h"
 HEAP8[$31>>0] = 0; //@line 155 "./common/StringMap.h"
 $scevgep = ((($this)) + 384|0); //@line 44 "./common/StringMap.h"
 dest=$scevgep; stop=dest+64|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 44 "./common/StringMap.h"
 $32 = $num >>> 3; //@line 46 "./common/StringMap.h"
 $33 = ($32|0)==(0); //@line 48 "./common/StringMap.h"
 if ($33) {
  STACKTOP = sp;return; //@line 50 "./common/StringMap.h"
 } else {
  $i1$05 = 0;
 }
 while(1) {
  $34 = (($entries) + ($i1$05<<3)|0); //@line 49 "./common/StringMap.h"
  $35 = HEAP32[$34>>2]|0; //@line 49 "./common/StringMap.h"
  $36 = (((($entries) + ($i1$05<<3)|0)) + 4|0); //@line 49 "./common/StringMap.h"
  $37 = HEAP32[$36>>2]|0; //@line 49 "./common/StringMap.h"
  $38 = HEAP8[$35>>0]|0; //@line 142 "./common/StringMap.h"
  $39 = ($38<<24>>24)==(0); //@line 142 "./common/StringMap.h"
  if ($39) {
   $hash$0$lcssa$i$i = 5381;
  } else {
   $$02$i$i = $35;$41 = $38;$hash$01$i$i = 5381;
   while(1) {
    $40 = $41 << 24 >> 24; //@line 142 "./common/StringMap.h"
    $42 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
    $43 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
    $44 = (($43) + ($40))|0; //@line 143 "./common/StringMap.h"
    $45 = HEAP8[$42>>0]|0; //@line 142 "./common/StringMap.h"
    $46 = ($45<<24>>24)==(0); //@line 142 "./common/StringMap.h"
    if ($46) {
     $hash$0$lcssa$i$i = $44;
     break;
    } else {
     $$02$i$i = $42;$41 = $45;$hash$01$i$i = $44;
    }
   }
  }
  $i$08$i = 0;
  while(1) {
   $49 = (($i$08$i) + ($hash$0$lcssa$i$i))|0; //@line 112 "./common/StringMap.h"
   $50 = $49 & 31; //@line 112 "./common/StringMap.h"
   $51 = (((($this) + (($50*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $52 = HEAP8[$51>>0]|0; //@line 114 "./common/StringMap.h"
   $53 = ($52<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   $48 = (($i$08$i) + 1)|0; //@line 110 "./common/StringMap.h"
   if ($53) {
    $$lcssa26 = $50;$$lcssa28 = $51;
    label = 8;
    break;
   }
   $47 = ($48>>>0)<(32); //@line 110 "./common/StringMap.h"
   if ($47) {
    $i$08$i = $48;
   } else {
    break;
   }
  }
  if ((label|0) == 8) {
   label = 0;
   HEAP8[$$lcssa28>>0] = 1; //@line 117 "./common/StringMap.h"
   $54 = (($this) + (($$lcssa26*12)|0)|0); //@line 118 "./common/StringMap.h"
   HEAP32[$54>>2] = $35; //@line 118 "./common/StringMap.h"
   $55 = (((($this) + (($$lcssa26*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
   HEAP32[$55>>2] = $37; //@line 119 "./common/StringMap.h"
  }
  $56 = ($37>>>0)>(15); //@line 126 "./common/StringMap.h"
  if ($56) {
   HEAP32[$vararg_buffer>>2] = $35; //@line 128 "./common/StringMap.h"
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
   HEAP32[$vararg_ptr1>>2] = $37; //@line 128 "./common/StringMap.h"
   (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  } else {
   $57 = (((($this)) + 384|0) + ($37<<2)|0); //@line 132 "./common/StringMap.h"
   HEAP32[$57>>2] = $35; //@line 132 "./common/StringMap.h"
  }
  $58 = (($i1$05) + 1)|0; //@line 48 "./common/StringMap.h"
  $exitcond = ($58|0)==($32|0); //@line 48 "./common/StringMap.h"
  if ($exitcond) {
   break;
  } else {
   $i1$05 = $58;
  }
 }
 STACKTOP = sp;return; //@line 50 "./common/StringMap.h"
}
function __ZN4love9StringMapINS_8joystick8Joystick11GamepadAxisELj7EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $42 = (($hash$0$lcssa$i) + 6)|0; //@line 112 "./common/StringMap.h"
       $43 = (($42>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
       $44 = (((($this) + (($43*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
       $45 = HEAP8[$44>>0]|0; //@line 114 "./common/StringMap.h"
       $46 = ($45<<24>>24)==(0); //@line 114 "./common/StringMap.h"
       if ($46) {
        $$lcssa = $43;$$lcssa10 = $44;
        label = 5;
       } else {
        $47 = (($hash$0$lcssa$i) + 7)|0; //@line 112 "./common/StringMap.h"
        $48 = (($47>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
        $49 = (((($this) + (($48*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
        $50 = HEAP8[$49>>0]|0; //@line 114 "./common/StringMap.h"
        $51 = ($50<<24>>24)==(0); //@line 114 "./common/StringMap.h"
        if ($51) {
         $$lcssa = $48;$$lcssa10 = $49;
         label = 5;
        } else {
         $52 = (($hash$0$lcssa$i) + 8)|0; //@line 112 "./common/StringMap.h"
         $53 = (($52>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
         $54 = (((($this) + (($53*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
         $55 = HEAP8[$54>>0]|0; //@line 114 "./common/StringMap.h"
         $56 = ($55<<24>>24)==(0); //@line 114 "./common/StringMap.h"
         if ($56) {
          $$lcssa = $53;$$lcssa10 = $54;
          label = 5;
         } else {
          $57 = (($hash$0$lcssa$i) + 9)|0; //@line 112 "./common/StringMap.h"
          $58 = (($57>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
          $59 = (((($this) + (($58*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
          $60 = HEAP8[$59>>0]|0; //@line 114 "./common/StringMap.h"
          $61 = ($60<<24>>24)==(0); //@line 114 "./common/StringMap.h"
          if ($61) {
           $$lcssa = $58;$$lcssa10 = $59;
           label = 5;
          } else {
           $62 = (($hash$0$lcssa$i) + 10)|0; //@line 112 "./common/StringMap.h"
           $63 = (($62>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
           $64 = (((($this) + (($63*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
           $65 = HEAP8[$64>>0]|0; //@line 114 "./common/StringMap.h"
           $66 = ($65<<24>>24)==(0); //@line 114 "./common/StringMap.h"
           if ($66) {
            $$lcssa = $63;$$lcssa10 = $64;
            label = 5;
           } else {
            $67 = (($hash$0$lcssa$i) + 11)|0; //@line 112 "./common/StringMap.h"
            $68 = (($67>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
            $69 = (((($this) + (($68*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
            $70 = HEAP8[$69>>0]|0; //@line 114 "./common/StringMap.h"
            $71 = ($70<<24>>24)==(0); //@line 114 "./common/StringMap.h"
            if ($71) {
             $$lcssa = $68;$$lcssa10 = $69;
             label = 5;
            } else {
             $72 = (($hash$0$lcssa$i) + 12)|0; //@line 112 "./common/StringMap.h"
             $73 = (($72>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
             $74 = (((($this) + (($73*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
             $75 = HEAP8[$74>>0]|0; //@line 114 "./common/StringMap.h"
             $76 = ($75<<24>>24)==(0); //@line 114 "./common/StringMap.h"
             if ($76) {
              $$lcssa = $73;$$lcssa10 = $74;
              label = 5;
             } else {
              $77 = (($hash$0$lcssa$i) + 13)|0; //@line 112 "./common/StringMap.h"
              $78 = (($77>>>0) % 14)&-1; //@line 112 "./common/StringMap.h"
              $79 = (((($this) + (($78*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
              $80 = HEAP8[$79>>0]|0; //@line 114 "./common/StringMap.h"
              $81 = ($80<<24>>24)==(0); //@line 114 "./common/StringMap.h"
              if ($81) {
               $$lcssa = $78;$$lcssa10 = $79;
               label = 5;
              } else {
               $inserted$2$off0 = 0;
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(6); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 168|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love9StringMapINS_8joystick8Joystick13GamepadButtonELj16EEC2EPKNS4_5EntryEj($this,$entries,$num) {
 $this = $this|0;
 $entries = $entries|0;
 $num = $num|0;
 var $$02$i$i = 0, $$lcssa26 = 0, $$lcssa28 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $exitcond = 0, $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$08$i = 0, $i1$05 = 0, $scevgep = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 8|0); //@line 155 "./common/StringMap.h"
 HEAP8[$0>>0] = 0; //@line 155 "./common/StringMap.h"
 $1 = ((($this)) + 20|0); //@line 155 "./common/StringMap.h"
 HEAP8[$1>>0] = 0; //@line 155 "./common/StringMap.h"
 $2 = ((($this)) + 32|0); //@line 155 "./common/StringMap.h"
 HEAP8[$2>>0] = 0; //@line 155 "./common/StringMap.h"
 $3 = ((($this)) + 44|0); //@line 155 "./common/StringMap.h"
 HEAP8[$3>>0] = 0; //@line 155 "./common/StringMap.h"
 $4 = ((($this)) + 56|0); //@line 155 "./common/StringMap.h"
 HEAP8[$4>>0] = 0; //@line 155 "./common/StringMap.h"
 $5 = ((($this)) + 68|0); //@line 155 "./common/StringMap.h"
 HEAP8[$5>>0] = 0; //@line 155 "./common/StringMap.h"
 $6 = ((($this)) + 80|0); //@line 155 "./common/StringMap.h"
 HEAP8[$6>>0] = 0; //@line 155 "./common/StringMap.h"
 $7 = ((($this)) + 92|0); //@line 155 "./common/StringMap.h"
 HEAP8[$7>>0] = 0; //@line 155 "./common/StringMap.h"
 $8 = ((($this)) + 104|0); //@line 155 "./common/StringMap.h"
 HEAP8[$8>>0] = 0; //@line 155 "./common/StringMap.h"
 $9 = ((($this)) + 116|0); //@line 155 "./common/StringMap.h"
 HEAP8[$9>>0] = 0; //@line 155 "./common/StringMap.h"
 $10 = ((($this)) + 128|0); //@line 155 "./common/StringMap.h"
 HEAP8[$10>>0] = 0; //@line 155 "./common/StringMap.h"
 $11 = ((($this)) + 140|0); //@line 155 "./common/StringMap.h"
 HEAP8[$11>>0] = 0; //@line 155 "./common/StringMap.h"
 $12 = ((($this)) + 152|0); //@line 155 "./common/StringMap.h"
 HEAP8[$12>>0] = 0; //@line 155 "./common/StringMap.h"
 $13 = ((($this)) + 164|0); //@line 155 "./common/StringMap.h"
 HEAP8[$13>>0] = 0; //@line 155 "./common/StringMap.h"
 $14 = ((($this)) + 176|0); //@line 155 "./common/StringMap.h"
 HEAP8[$14>>0] = 0; //@line 155 "./common/StringMap.h"
 $15 = ((($this)) + 188|0); //@line 155 "./common/StringMap.h"
 HEAP8[$15>>0] = 0; //@line 155 "./common/StringMap.h"
 $16 = ((($this)) + 200|0); //@line 155 "./common/StringMap.h"
 HEAP8[$16>>0] = 0; //@line 155 "./common/StringMap.h"
 $17 = ((($this)) + 212|0); //@line 155 "./common/StringMap.h"
 HEAP8[$17>>0] = 0; //@line 155 "./common/StringMap.h"
 $18 = ((($this)) + 224|0); //@line 155 "./common/StringMap.h"
 HEAP8[$18>>0] = 0; //@line 155 "./common/StringMap.h"
 $19 = ((($this)) + 236|0); //@line 155 "./common/StringMap.h"
 HEAP8[$19>>0] = 0; //@line 155 "./common/StringMap.h"
 $20 = ((($this)) + 248|0); //@line 155 "./common/StringMap.h"
 HEAP8[$20>>0] = 0; //@line 155 "./common/StringMap.h"
 $21 = ((($this)) + 260|0); //@line 155 "./common/StringMap.h"
 HEAP8[$21>>0] = 0; //@line 155 "./common/StringMap.h"
 $22 = ((($this)) + 272|0); //@line 155 "./common/StringMap.h"
 HEAP8[$22>>0] = 0; //@line 155 "./common/StringMap.h"
 $23 = ((($this)) + 284|0); //@line 155 "./common/StringMap.h"
 HEAP8[$23>>0] = 0; //@line 155 "./common/StringMap.h"
 $24 = ((($this)) + 296|0); //@line 155 "./common/StringMap.h"
 HEAP8[$24>>0] = 0; //@line 155 "./common/StringMap.h"
 $25 = ((($this)) + 308|0); //@line 155 "./common/StringMap.h"
 HEAP8[$25>>0] = 0; //@line 155 "./common/StringMap.h"
 $26 = ((($this)) + 320|0); //@line 155 "./common/StringMap.h"
 HEAP8[$26>>0] = 0; //@line 155 "./common/StringMap.h"
 $27 = ((($this)) + 332|0); //@line 155 "./common/StringMap.h"
 HEAP8[$27>>0] = 0; //@line 155 "./common/StringMap.h"
 $28 = ((($this)) + 344|0); //@line 155 "./common/StringMap.h"
 HEAP8[$28>>0] = 0; //@line 155 "./common/StringMap.h"
 $29 = ((($this)) + 356|0); //@line 155 "./common/StringMap.h"
 HEAP8[$29>>0] = 0; //@line 155 "./common/StringMap.h"
 $30 = ((($this)) + 368|0); //@line 155 "./common/StringMap.h"
 HEAP8[$30>>0] = 0; //@line 155 "./common/StringMap.h"
 $31 = ((($this)) + 380|0); //@line 155 "./common/StringMap.h"
 HEAP8[$31>>0] = 0; //@line 155 "./common/StringMap.h"
 $scevgep = ((($this)) + 384|0); //@line 44 "./common/StringMap.h"
 dest=$scevgep; stop=dest+64|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 44 "./common/StringMap.h"
 $32 = $num >>> 3; //@line 46 "./common/StringMap.h"
 $33 = ($32|0)==(0); //@line 48 "./common/StringMap.h"
 if ($33) {
  STACKTOP = sp;return; //@line 50 "./common/StringMap.h"
 } else {
  $i1$05 = 0;
 }
 while(1) {
  $34 = (($entries) + ($i1$05<<3)|0); //@line 49 "./common/StringMap.h"
  $35 = HEAP32[$34>>2]|0; //@line 49 "./common/StringMap.h"
  $36 = (((($entries) + ($i1$05<<3)|0)) + 4|0); //@line 49 "./common/StringMap.h"
  $37 = HEAP32[$36>>2]|0; //@line 49 "./common/StringMap.h"
  $38 = HEAP8[$35>>0]|0; //@line 142 "./common/StringMap.h"
  $39 = ($38<<24>>24)==(0); //@line 142 "./common/StringMap.h"
  if ($39) {
   $hash$0$lcssa$i$i = 5381;
  } else {
   $$02$i$i = $35;$41 = $38;$hash$01$i$i = 5381;
   while(1) {
    $40 = $41 << 24 >> 24; //@line 142 "./common/StringMap.h"
    $42 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
    $43 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
    $44 = (($43) + ($40))|0; //@line 143 "./common/StringMap.h"
    $45 = HEAP8[$42>>0]|0; //@line 142 "./common/StringMap.h"
    $46 = ($45<<24>>24)==(0); //@line 142 "./common/StringMap.h"
    if ($46) {
     $hash$0$lcssa$i$i = $44;
     break;
    } else {
     $$02$i$i = $42;$41 = $45;$hash$01$i$i = $44;
    }
   }
  }
  $i$08$i = 0;
  while(1) {
   $49 = (($i$08$i) + ($hash$0$lcssa$i$i))|0; //@line 112 "./common/StringMap.h"
   $50 = $49 & 31; //@line 112 "./common/StringMap.h"
   $51 = (((($this) + (($50*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $52 = HEAP8[$51>>0]|0; //@line 114 "./common/StringMap.h"
   $53 = ($52<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   $48 = (($i$08$i) + 1)|0; //@line 110 "./common/StringMap.h"
   if ($53) {
    $$lcssa26 = $50;$$lcssa28 = $51;
    label = 8;
    break;
   }
   $47 = ($48>>>0)<(32); //@line 110 "./common/StringMap.h"
   if ($47) {
    $i$08$i = $48;
   } else {
    break;
   }
  }
  if ((label|0) == 8) {
   label = 0;
   HEAP8[$$lcssa28>>0] = 1; //@line 117 "./common/StringMap.h"
   $54 = (($this) + (($$lcssa26*12)|0)|0); //@line 118 "./common/StringMap.h"
   HEAP32[$54>>2] = $35; //@line 118 "./common/StringMap.h"
   $55 = (((($this) + (($$lcssa26*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
   HEAP32[$55>>2] = $37; //@line 119 "./common/StringMap.h"
  }
  $56 = ($37>>>0)>(15); //@line 126 "./common/StringMap.h"
  if ($56) {
   HEAP32[$vararg_buffer>>2] = $35; //@line 128 "./common/StringMap.h"
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
   HEAP32[$vararg_ptr1>>2] = $37; //@line 128 "./common/StringMap.h"
   (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  } else {
   $57 = (((($this)) + 384|0) + ($37<<2)|0); //@line 132 "./common/StringMap.h"
   HEAP32[$57>>2] = $35; //@line 132 "./common/StringMap.h"
  }
  $58 = (($i1$05) + 1)|0; //@line 48 "./common/StringMap.h"
  $exitcond = ($58|0)==($32|0); //@line 48 "./common/StringMap.h"
  if ($exitcond) {
   break;
  } else {
   $i1$05 = $58;
  }
 }
 STACKTOP = sp;return; //@line 50 "./common/StringMap.h"
}
function __ZN4love9StringMapINS_8joystick8Joystick9InputTypeELj3EE3addEPKcS3_($this,$key,$value) {
 $this = $this|0;
 $key = $key|0;
 $value = $value|0;
 var $$0 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hash$0$lcssa$i = 0, $hash$01$i = 0, $inserted$2$off0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP8[$key>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i = 5381;
 } else {
  $$02$i = $key;$3 = $0;$hash$01$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i = $6;
    break;
   } else {
    $$02$i = $4;$3 = $7;$hash$01$i = $6;
   }
  }
 }
 $9 = (($hash$0$lcssa$i>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
 $10 = (((($this) + (($9*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
 $11 = HEAP8[$10>>0]|0; //@line 114 "./common/StringMap.h"
 $12 = ($11<<24>>24)==(0); //@line 114 "./common/StringMap.h"
 if ($12) {
  $$lcssa = $9;$$lcssa10 = $10;
  label = 5;
 } else {
  $13 = (($hash$0$lcssa$i) + 1)|0; //@line 112 "./common/StringMap.h"
  $14 = (($13>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
  $15 = (((($this) + (($14*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 114 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 114 "./common/StringMap.h"
  if ($17) {
   $$lcssa = $14;$$lcssa10 = $15;
   label = 5;
  } else {
   $22 = (($hash$0$lcssa$i) + 2)|0; //@line 112 "./common/StringMap.h"
   $23 = (($22>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
   $24 = (((($this) + (($23*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $25 = HEAP8[$24>>0]|0; //@line 114 "./common/StringMap.h"
   $26 = ($25<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   if ($26) {
    $$lcssa = $23;$$lcssa10 = $24;
    label = 5;
   } else {
    $27 = (($hash$0$lcssa$i) + 3)|0; //@line 112 "./common/StringMap.h"
    $28 = (($27>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
    $29 = (((($this) + (($28*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
    $30 = HEAP8[$29>>0]|0; //@line 114 "./common/StringMap.h"
    $31 = ($30<<24>>24)==(0); //@line 114 "./common/StringMap.h"
    if ($31) {
     $$lcssa = $28;$$lcssa10 = $29;
     label = 5;
    } else {
     $32 = (($hash$0$lcssa$i) + 4)|0; //@line 112 "./common/StringMap.h"
     $33 = (($32>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
     $34 = (((($this) + (($33*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
     $35 = HEAP8[$34>>0]|0; //@line 114 "./common/StringMap.h"
     $36 = ($35<<24>>24)==(0); //@line 114 "./common/StringMap.h"
     if ($36) {
      $$lcssa = $33;$$lcssa10 = $34;
      label = 5;
     } else {
      $37 = (($hash$0$lcssa$i) + 5)|0; //@line 112 "./common/StringMap.h"
      $38 = (($37>>>0) % 6)&-1; //@line 112 "./common/StringMap.h"
      $39 = (((($this) + (($38*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
      $40 = HEAP8[$39>>0]|0; //@line 114 "./common/StringMap.h"
      $41 = ($40<<24>>24)==(0); //@line 114 "./common/StringMap.h"
      if ($41) {
       $$lcssa = $38;$$lcssa10 = $39;
       label = 5;
      } else {
       $inserted$2$off0 = 0;
      }
     }
    }
   }
  }
 }
 if ((label|0) == 5) {
  HEAP8[$$lcssa10>>0] = 1; //@line 117 "./common/StringMap.h"
  $18 = (($this) + (($$lcssa*12)|0)|0); //@line 118 "./common/StringMap.h"
  HEAP32[$18>>2] = $key; //@line 118 "./common/StringMap.h"
  $19 = (((($this) + (($$lcssa*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
  HEAP32[$19>>2] = $value; //@line 119 "./common/StringMap.h"
  $inserted$2$off0 = 1;
 }
 $20 = ($value>>>0)>(2); //@line 126 "./common/StringMap.h"
 if ($20) {
  HEAP32[$vararg_buffer>>2] = $key; //@line 128 "./common/StringMap.h"
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
  HEAP32[$vararg_ptr1>>2] = $value; //@line 128 "./common/StringMap.h"
  (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 } else {
  $21 = (((($this)) + 72|0) + ($value<<2)|0); //@line 132 "./common/StringMap.h"
  HEAP32[$21>>2] = $key; //@line 132 "./common/StringMap.h"
  $$0 = $inserted$2$off0;
  STACKTOP = sp;return ($$0|0); //@line 135 "./common/StringMap.h"
 }
 return (0)|0;
}
function __ZN4love8joystick8Joystick8clampvalEf($x) {
 $x = +$x;
 var $$0 = 0.0, $$x = 0.0, $0 = 0.0, $1 = 0.0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+Math_abs((+$x))); //@line 34 "modules/joystick/Joystick.cpp"
 $1 = $0; //@line 34 "modules/joystick/Joystick.cpp"
 $2 = $1 < 0.01; //@line 34 "modules/joystick/Joystick.cpp"
 if ($2) {
  $$0 = 0.0;
 } else {
  $3 = $x < -0.99000000953674316; //@line 37 "modules/joystick/Joystick.cpp"
  if ($3) {
   $$0 = -1.0;
  } else {
   $4 = $x > 0.99000000953674316; //@line 38 "modules/joystick/Joystick.cpp"
   $$x = $4 ? 1.0 : $x; //@line 38 "modules/joystick/Joystick.cpp"
   $$0 = $$x;
  }
 }
 return (+$$0); //@line 41 "modules/joystick/Joystick.cpp"
}
function __ZN4love8joystick8Joystick11getConstantEPKcRNS1_3HatE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = $9 & 31; //@line 72 "./common/StringMap.h"
  $11 = (((635704 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (635704 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(32); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((635704 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 45 "modules/joystick/Joystick.cpp"
  return ($$$2$i|0); //@line 45 "modules/joystick/Joystick.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 45 "modules/joystick/Joystick.cpp"
  return ($$$2$i|0); //@line 45 "modules/joystick/Joystick.cpp"
 }
 return (0)|0;
}
function __ZN4love8joystick8Joystick11getConstantENS1_3HatERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(15); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (636088 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 50 "modules/joystick/Joystick.cpp"
}
function __ZN4love8joystick8Joystick11getConstantEPKcRNS1_11GamepadAxisE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 14)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((636152 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (636152 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(14); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((636152 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 55 "modules/joystick/Joystick.cpp"
  return ($$$2$i|0); //@line 55 "modules/joystick/Joystick.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 55 "modules/joystick/Joystick.cpp"
  return ($$$2$i|0); //@line 55 "modules/joystick/Joystick.cpp"
 }
 return (0)|0;
}
function __ZN4love8joystick8Joystick11getConstantENS1_11GamepadAxisERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(6); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (636320 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 60 "modules/joystick/Joystick.cpp"
}
function __ZN4love8joystick8Joystick11getConstantEPKcRNS1_13GamepadButtonE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = $9 & 31; //@line 72 "./common/StringMap.h"
  $11 = (((636348 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (636348 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(32); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((636348 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 65 "modules/joystick/Joystick.cpp"
  return ($$$2$i|0); //@line 65 "modules/joystick/Joystick.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 65 "modules/joystick/Joystick.cpp"
  return ($$$2$i|0); //@line 65 "modules/joystick/Joystick.cpp"
 }
 return (0)|0;
}
function __ZN4love8joystick8Joystick11getConstantENS1_13GamepadButtonERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(15); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (636732 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 70 "modules/joystick/Joystick.cpp"
}
function __ZN4love8joystick8Joystick11getConstantEPKcRNS1_9InputTypeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 6)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((636796 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (636796 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(6); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((636796 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 75 "modules/joystick/Joystick.cpp"
  return ($$$2$i|0); //@line 75 "modules/joystick/Joystick.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 75 "modules/joystick/Joystick.cpp"
  return ($$$2$i|0); //@line 75 "modules/joystick/Joystick.cpp"
 }
 return (0)|0;
}
function __ZN4love8joystick8Joystick11getConstantENS1_9InputTypeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(2); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (636868 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 80 "modules/joystick/Joystick.cpp"
}
function __ZN4love8joystick18luax_checkjoystickEP9lua_Statei($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,$idx,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 return ($0|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
}
function __ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,$idx,$type) {
 $L = $L|0;
 $idx = $idx|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name = 0, $name1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 4|0;
 $name1 = sp;
 $0 = (_lua_type($L,$idx)|0); //@line 439 "./common/runtime.h"
 $1 = ($0|0)==(7); //@line 439 "./common/runtime.h"
 if (!($1)) {
  HEAP32[$name>>2] = 442055; //@line 441 "./common/runtime.h"
  (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name)|0); //@line 442 "./common/runtime.h"
  $2 = HEAP32[$name>>2]|0; //@line 443 "./common/runtime.h"
  (_luax_typerror($L,$idx,$2)|0); //@line 443 "./common/runtime.h"
 }
 $3 = (_lua_touserdata($L,$idx)|0); //@line 446 "./common/runtime.h"
 $4 = HEAP32[$3>>2]|0; //@line 448 "./common/runtime.h"
 $5 = HEAP32[156504]|0; //@line 448 "./common/runtime.h"
 $6 = $type >>> 5; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $7 = ((($5) + ($4<<3)|0) + ($6<<2)|0); //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $8 = $type & 31; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $9 = 1 << $8; //@line 177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $10 = HEAP32[$7>>2]|0; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = $10 & $9; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
  $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
  STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
 }
 HEAP32[$name1>>2] = 442055; //@line 450 "./common/runtime.h"
 (__ZN4love11getTypeNameENS_4TypeERPKc($type,$name1)|0); //@line 451 "./common/runtime.h"
 $13 = HEAP32[$name1>>2]|0; //@line 452 "./common/runtime.h"
 (_luax_typerror($L,$idx,$13)|0); //@line 452 "./common/runtime.h"
 $14 = ((($3)) + 4|0); //@line 455 "./common/runtime.h"
 $15 = HEAP32[$14>>2]|0; //@line 455 "./common/runtime.h"
 STACKTOP = sp;return ($15|0); //@line 455 "./common/runtime.h"
}
function __ZN4love8joystick22w_Joystick_isConnectedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,1,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 40 "modules/joystick/wrap_Joystick.cpp"
 $2 = ((($1)) + 24|0); //@line 40 "modules/joystick/wrap_Joystick.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 40 "modules/joystick/wrap_Joystick.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 40 "modules/joystick/wrap_Joystick.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 40 "modules/joystick/wrap_Joystick.cpp"
 return 1; //@line 41 "modules/joystick/wrap_Joystick.cpp"
}
function __ZN4love8joystick18w_Joystick_getNameEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,1,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 47 "modules/joystick/wrap_Joystick.cpp"
 $2 = ((($1)) + 28|0); //@line 47 "modules/joystick/wrap_Joystick.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 47 "modules/joystick/wrap_Joystick.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 47 "modules/joystick/wrap_Joystick.cpp"
 _lua_pushstring($L,$4); //@line 47 "modules/joystick/wrap_Joystick.cpp"
 return 1; //@line 48 "modules/joystick/wrap_Joystick.cpp"
}
function __ZN4love8joystick16w_Joystick_getIDEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,1,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 56 "modules/joystick/wrap_Joystick.cpp"
 $2 = ((($1)) + 88|0); //@line 56 "modules/joystick/wrap_Joystick.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 56 "modules/joystick/wrap_Joystick.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 56 "modules/joystick/wrap_Joystick.cpp"
 $5 = (($4) + 1)|0; //@line 56 "modules/joystick/wrap_Joystick.cpp"
 _lua_pushinteger($L,$5); //@line 56 "modules/joystick/wrap_Joystick.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 58 "modules/joystick/wrap_Joystick.cpp"
 $7 = ((($6)) + 84|0); //@line 58 "modules/joystick/wrap_Joystick.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 58 "modules/joystick/wrap_Joystick.cpp"
 $9 = (FUNCTION_TABLE_ii[$8 & 8191]($0)|0); //@line 58 "modules/joystick/wrap_Joystick.cpp"
 $10 = ($9|0)>(-1); //@line 59 "modules/joystick/wrap_Joystick.cpp"
 if ($10) {
  $11 = (($9) + 1)|0; //@line 60 "modules/joystick/wrap_Joystick.cpp"
  _lua_pushinteger($L,$11); //@line 60 "modules/joystick/wrap_Joystick.cpp"
  return 2; //@line 64 "modules/joystick/wrap_Joystick.cpp"
 } else {
  _lua_pushnil($L); //@line 62 "modules/joystick/wrap_Joystick.cpp"
  return 2; //@line 64 "modules/joystick/wrap_Joystick.cpp"
 }
 return (0)|0;
}
function __ZN4love8joystick18w_Joystick_getGUIDEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,1,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 70 "modules/joystick/wrap_Joystick.cpp"
 $3 = ((($2)) + 80|0); //@line 70 "modules/joystick/wrap_Joystick.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 70 "modules/joystick/wrap_Joystick.cpp"
 FUNCTION_TABLE_vii[$4 & 8191]($0,$1); //@line 70 "modules/joystick/wrap_Joystick.cpp"
 __THREW__ = 0;
 invoke_vii(3405,($L|0),($0|0)); //@line 70 "modules/joystick/wrap_Joystick.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 70 "modules/joystick/wrap_Joystick.cpp"
  ___resumeException($7|0);
  // unreachable;
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 70 "modules/joystick/wrap_Joystick.cpp"
  STACKTOP = sp;return 1; //@line 71 "modules/joystick/wrap_Joystick.cpp"
 }
 return (0)|0;
}
function __ZN4love8joystick23w_Joystick_getAxisCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,1,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 77 "modules/joystick/wrap_Joystick.cpp"
 $2 = ((($1)) + 32|0); //@line 77 "modules/joystick/wrap_Joystick.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 77 "modules/joystick/wrap_Joystick.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 77 "modules/joystick/wrap_Joystick.cpp"
 _lua_pushinteger($L,$4); //@line 77 "modules/joystick/wrap_Joystick.cpp"
 return 1; //@line 78 "modules/joystick/wrap_Joystick.cpp"
}
function __ZN4love8joystick25w_Joystick_getButtonCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,1,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 84 "modules/joystick/wrap_Joystick.cpp"
 $2 = ((($1)) + 36|0); //@line 84 "modules/joystick/wrap_Joystick.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 84 "modules/joystick/wrap_Joystick.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 84 "modules/joystick/wrap_Joystick.cpp"
 _lua_pushinteger($L,$4); //@line 84 "modules/joystick/wrap_Joystick.cpp"
 return 1; //@line 85 "modules/joystick/wrap_Joystick.cpp"
}
function __ZN4love8joystick22w_Joystick_getHatCountEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,1,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 91 "modules/joystick/wrap_Joystick.cpp"
 $2 = ((($1)) + 40|0); //@line 91 "modules/joystick/wrap_Joystick.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 91 "modules/joystick/wrap_Joystick.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 91 "modules/joystick/wrap_Joystick.cpp"
 _lua_pushinteger($L,$4); //@line 91 "modules/joystick/wrap_Joystick.cpp"
 return 1; //@line 92 "modules/joystick/wrap_Joystick.cpp"
}
function __ZN4love8joystick18w_Joystick_getAxisEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,1,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 98 "modules/joystick/wrap_Joystick.cpp"
 $2 = (~~(($1))); //@line 98 "modules/joystick/wrap_Joystick.cpp"
 $3 = (($2) + -1)|0; //@line 98 "modules/joystick/wrap_Joystick.cpp"
 $4 = HEAP32[$0>>2]|0; //@line 99 "modules/joystick/wrap_Joystick.cpp"
 $5 = ((($4)) + 44|0); //@line 99 "modules/joystick/wrap_Joystick.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 99 "modules/joystick/wrap_Joystick.cpp"
 $7 = (+FUNCTION_TABLE_dii[$6 & 8191]($0,$3)); //@line 99 "modules/joystick/wrap_Joystick.cpp"
 $8 = $7; //@line 99 "modules/joystick/wrap_Joystick.cpp"
 _lua_pushnumber($L,$8); //@line 99 "modules/joystick/wrap_Joystick.cpp"
 return 1; //@line 100 "modules/joystick/wrap_Joystick.cpp"
}
function __ZN4love8joystick18w_Joystick_getAxesEP9lua_State($L) {
 $L = $L|0;
 var $$cast = 0, $$cast11 = 0, $$in = 0, $$lcssa = 0, $$lcssa7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $axes = 0, $i$09 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i1 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $axes = sp;
 $0 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,1,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 106 "modules/joystick/wrap_Joystick.cpp"
 $2 = ((($1)) + 48|0); //@line 106 "modules/joystick/wrap_Joystick.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 106 "modules/joystick/wrap_Joystick.cpp"
 FUNCTION_TABLE_vii[$3 & 8191]($axes,$0); //@line 106 "modules/joystick/wrap_Joystick.cpp"
 $4 = ((($axes)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$4>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$axes>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ($5|0)==($6|0); //@line 108 "modules/joystick/wrap_Joystick.cpp"
 $$cast = $6; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 L1: do {
  if ($7) {
   $$in = $5;$$lcssa = 0;$$lcssa7 = $$cast;
  } else {
   $17 = $$cast;$i$09 = 0;
   while(1) {
    $16 = (($17) + ($i$09<<2)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $18 = +HEAPF32[$16>>2]; //@line 109 "modules/joystick/wrap_Joystick.cpp"
    $19 = $18; //@line 109 "modules/joystick/wrap_Joystick.cpp"
    __THREW__ = 0;
    invoke_vid(3394,($L|0),(+$19)); //@line 109 "modules/joystick/wrap_Joystick.cpp"
    $20 = __THREW__; __THREW__ = 0;
    $21 = $20&1;
    if ($21) {
     break;
    }
    $22 = (($i$09) + 1)|0; //@line 108 "modules/joystick/wrap_Joystick.cpp"
    $23 = HEAP32[$4>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $24 = HEAP32[$axes>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $25 = (($23) - ($24))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $26 = $25 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $27 = ($22>>>0)<($26>>>0); //@line 108 "modules/joystick/wrap_Joystick.cpp"
    $$cast11 = $24; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($27) {
     $17 = $$cast11;$i$09 = $22;
    } else {
     $$in = $23;$$lcssa = $26;$$lcssa7 = $$cast11;
     break L1;
    }
   }
   $28 = ___cxa_find_matching_catch_2()|0;
   $29 = tempRet0;
   $30 = HEAP32[$axes>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $31 = ($30|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $32 = $30; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($31) {
    ___resumeException($28|0);
    // unreachable;
   }
   $33 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $34 = ($33|0)==($30|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($34)) {
    $scevgep$i$i$i = ((($33)) + -4|0);
    $35 = $scevgep$i$i$i;
    $36 = (($35) - ($32))|0;
    $37 = $36 >>> 2;
    $38 = $37 ^ -1;
    $scevgep4$i$i$i = (($33) + ($38<<2)|0);
    HEAP32[$4>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($30); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($28|0);
   // unreachable;
  }
 } while(0);
 $8 = $$in; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ($$lcssa7|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $$lcssa7; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($9) {
  STACKTOP = sp;return ($$lcssa|0); //@line 112 "modules/joystick/wrap_Joystick.cpp"
 }
 $11 = ($8|0)==($$lcssa7|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($11)) {
  $scevgep$i$i$i1 = ((($8)) + -4|0);
  $12 = $scevgep$i$i$i1;
  $13 = (($12) - ($10))|0;
  $14 = $13 >>> 2;
  $15 = $14 ^ -1;
  $scevgep4$i$i$i2 = (($8) + ($15<<2)|0);
  HEAP32[$4>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($$lcssa7); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 STACKTOP = sp;return ($$lcssa|0); //@line 112 "modules/joystick/wrap_Joystick.cpp"
}
function __ZN4love8joystick17w_Joystick_getHatEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $direction = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $direction = sp;
 $0 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,1,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 $1 = (+_luaL_checknumber($L,2)); //@line 117 "modules/joystick/wrap_Joystick.cpp"
 $2 = (~~(($1))); //@line 117 "modules/joystick/wrap_Joystick.cpp"
 $3 = (($2) + -1)|0; //@line 117 "modules/joystick/wrap_Joystick.cpp"
 $4 = HEAP32[$0>>2]|0; //@line 119 "modules/joystick/wrap_Joystick.cpp"
 $5 = ((($4)) + 52|0); //@line 119 "modules/joystick/wrap_Joystick.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 119 "modules/joystick/wrap_Joystick.cpp"
 $7 = (FUNCTION_TABLE_iii[$6 & 8191]($0,$3)|0); //@line 119 "modules/joystick/wrap_Joystick.cpp"
 HEAP32[$direction>>2] = 1282466; //@line 121 "modules/joystick/wrap_Joystick.cpp"
 (__ZN4love8joystick8Joystick11getConstantENS1_3HatERPKc($7,$direction)|0); //@line 122 "modules/joystick/wrap_Joystick.cpp"
 $8 = HEAP32[$direction>>2]|0; //@line 124 "modules/joystick/wrap_Joystick.cpp"
 _lua_pushstring($L,$8); //@line 124 "modules/joystick/wrap_Joystick.cpp"
 STACKTOP = sp;return 1; //@line 125 "modules/joystick/wrap_Joystick.cpp"
}
function __ZN4love8joystick17w_Joystick_isDownEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buttons = 0, $i$0 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i1 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i2 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $buttons = sp + 4|0;
 $0 = sp;
 $1 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,1,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 (_luaL_checkinteger($L,2)|0); //@line 132 "modules/joystick/wrap_Joystick.cpp"
 HEAP32[$buttons>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($buttons)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($buttons)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$3>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $4 = ((($buttons)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $i$0 = 2;
 while(1) {
  __THREW__ = 0;
  $5 = (invoke_ii(3385,($L|0))|0); //@line 135 "modules/joystick/wrap_Joystick.cpp"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if ($7) {
   label = 5;
   break;
  }
  $8 = ($i$0|0)>($5|0); //@line 135 "modules/joystick/wrap_Joystick.cpp"
  if ($8) {
   label = 4;
   break;
  }
  __THREW__ = 0;
  $28 = (+invoke_dii(3361,($L|0),($i$0|0))); //@line 136 "modules/joystick/wrap_Joystick.cpp"
  $29 = __THREW__; __THREW__ = 0;
  $30 = $29&1;
  if ($30) {
   label = 5;
   break;
  }
  $31 = (~~(($28))); //@line 136 "modules/joystick/wrap_Joystick.cpp"
  $32 = (($31) + -1)|0; //@line 136 "modules/joystick/wrap_Joystick.cpp"
  HEAP32[$0>>2] = $32; //@line 136 "modules/joystick/wrap_Joystick.cpp"
  $33 = HEAP32[$2>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $34 = HEAP32[$4>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $35 = ($33>>>0)<($34>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($35) {
   HEAP32[$33>>2] = $32; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $36 = ((($33)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$2>>2] = $36; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  } else {
   __THREW__ = 0;
   invoke_vii(3406,($buttons|0),($0|0)); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $37 = __THREW__; __THREW__ = 0;
   $38 = $37&1;
   if ($38) {
    label = 5;
    break;
   }
  }
  $39 = (($i$0) + 1)|0; //@line 135 "modules/joystick/wrap_Joystick.cpp"
  $i$0 = $39;
 }
 if ((label|0) == 4) {
  $9 = HEAP32[$1>>2]|0; //@line 138 "modules/joystick/wrap_Joystick.cpp"
  $10 = ((($9)) + 56|0); //@line 138 "modules/joystick/wrap_Joystick.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 138 "modules/joystick/wrap_Joystick.cpp"
  __THREW__ = 0;
  $12 = (invoke_iii($11|0,($1|0),($buttons|0))|0); //@line 138 "modules/joystick/wrap_Joystick.cpp"
  $13 = __THREW__; __THREW__ = 0;
  $14 = $13&1;
  if (!($14)) {
   __THREW__ = 0;
   invoke_vii(3397,($L|0),($12|0)); //@line 138 "modules/joystick/wrap_Joystick.cpp"
   $40 = __THREW__; __THREW__ = 0;
   $41 = $40&1;
   if (!($41)) {
    $42 = HEAP32[$buttons>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $43 = ($42|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $44 = $42; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($43) {
     STACKTOP = sp;return 1; //@line 140 "modules/joystick/wrap_Joystick.cpp"
    }
    $45 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $46 = ($45|0)==($42|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($46)) {
     $scevgep$i$i$i = ((($45)) + -4|0);
     $47 = $scevgep$i$i$i;
     $48 = (($47) - ($44))|0;
     $49 = $48 >>> 2;
     $50 = $49 ^ -1;
     $scevgep4$i$i$i = (($45) + ($50<<2)|0);
     HEAP32[$2>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($42); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    STACKTOP = sp;return 1; //@line 140 "modules/joystick/wrap_Joystick.cpp"
   }
  }
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  $lpad$phi$index = $17;$lpad$phi$index2 = $18;
 }
 else if ((label|0) == 5) {
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  $lpad$phi$index = $15;$lpad$phi$index2 = $16;
 }
 $19 = HEAP32[$buttons>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = ($19|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = $19; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($20) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $22 = HEAP32[$2>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $23 = ($22|0)==($19|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($23)) {
  $scevgep$i$i$i1 = ((($22)) + -4|0);
  $24 = $scevgep$i$i$i1;
  $25 = (($24) - ($21))|0;
  $26 = $25 >>> 2;
  $27 = $26 ^ -1;
  $scevgep4$i$i$i2 = (($22) + ($27<<2)|0);
  HEAP32[$2>>2] = $scevgep4$i$i$i2; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($19); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love8joystick20w_Joystick_isGamepadEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,1,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 145 "modules/joystick/wrap_Joystick.cpp"
 $2 = ((($1)) + 64|0); //@line 145 "modules/joystick/wrap_Joystick.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 145 "modules/joystick/wrap_Joystick.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 145 "modules/joystick/wrap_Joystick.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 145 "modules/joystick/wrap_Joystick.cpp"
 return 1; //@line 146 "modules/joystick/wrap_Joystick.cpp"
}
function __ZN4love8joystick25w_Joystick_getGamepadAxisEP9lua_State($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0.0, $axis = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $axis = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,1,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 $1 = (_luaL_checklstring($L,2,0)|0); //@line 153 "modules/joystick/wrap_Joystick.cpp"
 $2 = (__ZN4love8joystick8Joystick11getConstantEPKcRNS1_11GamepadAxisE($1,$axis)|0); //@line 156 "modules/joystick/wrap_Joystick.cpp"
 if ($2) {
  $4 = HEAP32[$0>>2]|0; //@line 159 "modules/joystick/wrap_Joystick.cpp"
  $5 = ((($4)) + 68|0); //@line 159 "modules/joystick/wrap_Joystick.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 159 "modules/joystick/wrap_Joystick.cpp"
  $7 = HEAP32[$axis>>2]|0; //@line 159 "modules/joystick/wrap_Joystick.cpp"
  $8 = (+FUNCTION_TABLE_dii[$6 & 8191]($0,$7)); //@line 159 "modules/joystick/wrap_Joystick.cpp"
  $9 = $8; //@line 159 "modules/joystick/wrap_Joystick.cpp"
  _lua_pushnumber($L,$9); //@line 159 "modules/joystick/wrap_Joystick.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 161 "modules/joystick/wrap_Joystick.cpp"
 } else {
  HEAP32[$vararg_buffer>>2] = $1; //@line 157 "modules/joystick/wrap_Joystick.cpp"
  $3 = (_luaL_error($L,442063,$vararg_buffer)|0); //@line 157 "modules/joystick/wrap_Joystick.cpp"
  $$0 = $3;
  STACKTOP = sp;return ($$0|0); //@line 161 "modules/joystick/wrap_Joystick.cpp"
 }
 return (0)|0;
}
function __ZN4love8joystick24w_Joystick_isGamepadDownEP9lua_State($L) {
 $L = $L|0;
 var $$3 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $button = 0, $buttons = 0, $i$0 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i4 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i5 = 0;
 var $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $buttons = sp + 8|0;
 $button = sp + 4|0;
 $0 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,1,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 HEAP32[$buttons>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($buttons)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($buttons)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __THREW__ = 0;
 $3 = (invoke_ii(3385,($L|0))|0); //@line 168 "modules/joystick/wrap_Joystick.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 do {
  if ($5) {
   label = 10;
  } else {
   $6 = (($3) + -1)|0; //@line 168 "modules/joystick/wrap_Joystick.cpp"
   $7 = ((($buttons)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $8 = ($6|0)==(0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($8)) {
    $9 = $6 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    __THREW__ = 0;
    $10 = (invoke_ii(3348,($9|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $11 = __THREW__; __THREW__ = 0;
    $12 = $11&1;
    if ($12) {
     label = 10;
     break;
    }
    $13 = $10; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $14 = (($10) + ($6<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $15 = $14; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    HEAP32[$buttons>>2] = $13; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$1>>2] = $13; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$7>>2] = $15; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   }
   __THREW__ = 0;
   (invoke_iiii(3359,($L|0),2,(0|0))|0); //@line 170 "modules/joystick/wrap_Joystick.cpp"
   $16 = __THREW__; __THREW__ = 0;
   $17 = $16&1;
   if ($17) {
    label = 10;
   } else {
    $i$0 = 2;
    while(1) {
     __THREW__ = 0;
     $18 = (invoke_ii(3385,($L|0))|0); //@line 172 "modules/joystick/wrap_Joystick.cpp"
     $19 = __THREW__; __THREW__ = 0;
     $20 = $19&1;
     if ($20) {
      label = 9;
      break;
     }
     $21 = ($i$0|0)>($18|0); //@line 172 "modules/joystick/wrap_Joystick.cpp"
     if ($21) {
      label = 25;
      break;
     }
     __THREW__ = 0;
     $35 = (invoke_iiii(3359,($L|0),($i$0|0),(0|0))|0); //@line 174 "modules/joystick/wrap_Joystick.cpp"
     $36 = __THREW__; __THREW__ = 0;
     $37 = $36&1;
     if ($37) {
      label = 9;
      break;
     }
     __THREW__ = 0;
     $38 = (invoke_iii(3644,($35|0),($button|0))|0); //@line 177 "modules/joystick/wrap_Joystick.cpp"
     $39 = __THREW__; __THREW__ = 0;
     $40 = $39&1;
     if ($40) {
      label = 9;
      break;
     }
     if (!($38)) {
      $$lcssa = $35;
      label = 19;
      break;
     }
     $44 = HEAP32[$1>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $45 = HEAP32[$7>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $46 = ($44|0)==($45|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($46) {
      __THREW__ = 0;
      invoke_vii(4202,($buttons|0),($button|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $49 = __THREW__; __THREW__ = 0;
      $50 = $49&1;
      if ($50) {
       label = 9;
       break;
      }
     } else {
      $47 = HEAP32[$button>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$44>>2] = $47; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $48 = ((($44)) + 4|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$1>>2] = $48; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     }
     $51 = (($i$0) + 1)|0; //@line 172 "modules/joystick/wrap_Joystick.cpp"
     $i$0 = $51;
    }
    if ((label|0) == 9) {
     $22 = ___cxa_find_matching_catch_2()|0;
     $23 = tempRet0;
     $lpad$phi$index = $22;$lpad$phi$index2 = $23;
     break;
    }
    else if ((label|0) == 19) {
     __THREW__ = 0;
     HEAP32[$vararg_buffer>>2] = $$lcssa; //@line 178 "modules/joystick/wrap_Joystick.cpp"
     $41 = (invoke_iiii(3389,($L|0),(442088|0),($vararg_buffer|0))|0); //@line 178 "modules/joystick/wrap_Joystick.cpp"
     $42 = __THREW__; __THREW__ = 0;
     $43 = $42&1;
     if ($43) {
      label = 10;
      break;
     }
     $$3 = $41;
    }
    else if ((label|0) == 25) {
     $52 = HEAP32[$0>>2]|0; //@line 183 "modules/joystick/wrap_Joystick.cpp"
     $53 = ((($52)) + 72|0); //@line 183 "modules/joystick/wrap_Joystick.cpp"
     $54 = HEAP32[$53>>2]|0; //@line 183 "modules/joystick/wrap_Joystick.cpp"
     __THREW__ = 0;
     $55 = (invoke_iii($54|0,($0|0),($buttons|0))|0); //@line 183 "modules/joystick/wrap_Joystick.cpp"
     $56 = __THREW__; __THREW__ = 0;
     $57 = $56&1;
     if ($57) {
      label = 10;
      break;
     }
     __THREW__ = 0;
     invoke_vii(3397,($L|0),($55|0)); //@line 183 "modules/joystick/wrap_Joystick.cpp"
     $58 = __THREW__; __THREW__ = 0;
     $59 = $58&1;
     if ($59) {
      label = 10;
      break;
     } else {
      $$3 = 1;
     }
    }
    $60 = HEAP32[$buttons>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $61 = ($60|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $62 = $60; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($61) {
     STACKTOP = sp;return ($$3|0); //@line 185 "modules/joystick/wrap_Joystick.cpp"
    }
    $63 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $64 = ($63|0)==($60|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if (!($64)) {
     $scevgep$i$i$i = ((($63)) + -4|0);
     $65 = $scevgep$i$i$i;
     $66 = (($65) - ($62))|0;
     $67 = $66 >>> 2;
     $68 = $67 ^ -1;
     $scevgep4$i$i$i = (($63) + ($68<<2)|0);
     HEAP32[$1>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    }
    __ZdlPv($60); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    STACKTOP = sp;return ($$3|0); //@line 185 "modules/joystick/wrap_Joystick.cpp"
   }
  }
 } while(0);
 if ((label|0) == 10) {
  $24 = ___cxa_find_matching_catch_2()|0;
  $25 = tempRet0;
  $lpad$phi$index = $24;$lpad$phi$index2 = $25;
 }
 $26 = HEAP32[$buttons>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $27 = ($26|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $28 = $26; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($27) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $29 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = ($29|0)==($26|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($30)) {
  $scevgep$i$i$i4 = ((($29)) + -4|0);
  $31 = $scevgep$i$i$i4;
  $32 = (($31) - ($28))|0;
  $33 = $32 >>> 2;
  $34 = $33 ^ -1;
  $scevgep4$i$i$i5 = (($29) + ($34<<2)|0);
  HEAP32[$1>>2] = $scevgep4$i$i$i5; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($26); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__16vectorIN4love8joystick8Joystick13GamepadButtonENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love8joystick31w_Joystick_isVibrationSupportedEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,1,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 190 "modules/joystick/wrap_Joystick.cpp"
 $2 = ((($1)) + 92|0); //@line 190 "modules/joystick/wrap_Joystick.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 190 "modules/joystick/wrap_Joystick.cpp"
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0); //@line 190 "modules/joystick/wrap_Joystick.cpp"
 __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$4); //@line 190 "modules/joystick/wrap_Joystick.cpp"
 return 1; //@line 191 "modules/joystick/wrap_Joystick.cpp"
}
function __ZN4love8joystick23w_Joystick_setVibrationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $success$0$in = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,1,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 $1 = (_lua_type($L,2)|0); //@line 199 "modules/joystick/wrap_Joystick.cpp"
 $2 = ($1|0)<(1); //@line 199 "modules/joystick/wrap_Joystick.cpp"
 if ($2) {
  $3 = HEAP32[$0>>2]|0; //@line 202 "modules/joystick/wrap_Joystick.cpp"
  $4 = ((($3)) + 100|0); //@line 202 "modules/joystick/wrap_Joystick.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 202 "modules/joystick/wrap_Joystick.cpp"
  $6 = (FUNCTION_TABLE_ii[$5 & 8191]($0)|0); //@line 202 "modules/joystick/wrap_Joystick.cpp"
  $success$0$in = $6;
  __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$success$0$in); //@line 212 "modules/joystick/wrap_Joystick.cpp"
  return 1; //@line 213 "modules/joystick/wrap_Joystick.cpp"
 } else {
  $7 = (+_luaL_checknumber($L,2)); //@line 206 "modules/joystick/wrap_Joystick.cpp"
  $8 = $7; //@line 206 "modules/joystick/wrap_Joystick.cpp"
  $9 = $8; //@line 207 "modules/joystick/wrap_Joystick.cpp"
  $10 = (+_luaL_optnumber($L,3,$9)); //@line 207 "modules/joystick/wrap_Joystick.cpp"
  $11 = $10; //@line 207 "modules/joystick/wrap_Joystick.cpp"
  $12 = (+_luaL_optnumber($L,4,-1.0)); //@line 208 "modules/joystick/wrap_Joystick.cpp"
  $13 = $12; //@line 208 "modules/joystick/wrap_Joystick.cpp"
  $14 = HEAP32[$0>>2]|0; //@line 209 "modules/joystick/wrap_Joystick.cpp"
  $15 = ((($14)) + 96|0); //@line 209 "modules/joystick/wrap_Joystick.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 209 "modules/joystick/wrap_Joystick.cpp"
  $17 = (FUNCTION_TABLE_iiddd[$16 & 4095]($0,$8,$11,$13)|0); //@line 209 "modules/joystick/wrap_Joystick.cpp"
  $success$0$in = $17;
  __ZN4love16luax_pushbooleanEP9lua_Stateb($L,$success$0$in); //@line 212 "modules/joystick/wrap_Joystick.cpp"
  return 1; //@line 213 "modules/joystick/wrap_Joystick.cpp"
 }
 return (0)|0;
}
function __ZN4love8joystick23w_Joystick_getVibrationEP9lua_State($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $left = 0, $right = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $left = sp + 4|0;
 $right = sp;
 $0 = (__ZN4love14luax_checktypeINS_8joystick8JoystickEEEPT_P9lua_StateiNS_4TypeE($L,1,24)|0); //@line 34 "modules/joystick/wrap_Joystick.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 220 "modules/joystick/wrap_Joystick.cpp"
 $2 = ((($1)) + 104|0); //@line 220 "modules/joystick/wrap_Joystick.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 220 "modules/joystick/wrap_Joystick.cpp"
 FUNCTION_TABLE_viii[$3 & 8191]($0,$left,$right); //@line 220 "modules/joystick/wrap_Joystick.cpp"
 $4 = +HEAPF32[$left>>2]; //@line 221 "modules/joystick/wrap_Joystick.cpp"
 $5 = $4; //@line 221 "modules/joystick/wrap_Joystick.cpp"
 _lua_pushnumber($L,$5); //@line 221 "modules/joystick/wrap_Joystick.cpp"
 $6 = +HEAPF32[$right>>2]; //@line 222 "modules/joystick/wrap_Joystick.cpp"
 $7 = $6; //@line 222 "modules/joystick/wrap_Joystick.cpp"
 _lua_pushnumber($L,$7); //@line 222 "modules/joystick/wrap_Joystick.cpp"
 STACKTOP = sp;return 2; //@line 223 "modules/joystick/wrap_Joystick.cpp"
}
function _luaopen_joystick($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 256 "modules/joystick/wrap_Joystick.cpp"
 HEAP32[$vararg_buffer>>2] = 26812; //@line 258 "modules/joystick/wrap_Joystick.cpp"
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 258 "modules/joystick/wrap_Joystick.cpp"
 HEAP32[$vararg_ptr1>>2] = 0; //@line 258 "modules/joystick/wrap_Joystick.cpp"
 $0 = (__ZN4love18luax_register_typeEP9lua_StateNS_4TypeEPKcz($L,24,442115,$vararg_buffer)|0); //@line 258 "modules/joystick/wrap_Joystick.cpp"
 STACKTOP = sp;return ($0|0); //@line 258 "modules/joystick/wrap_Joystick.cpp"
}
function __ZN4love8keyboard3sdl8KeyboardD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ModuleD2Ev($this); //@line 38 "modules/keyboard/sdl/Keyboard.h"
 __ZdlPv($this); //@line 38 "modules/keyboard/sdl/Keyboard.h"
 return; //@line 38 "modules/keyboard/sdl/Keyboard.h"
}
function __ZNK4love8keyboard8Keyboard13getModuleTypeEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 7; //@line 525 "./modules/keyboard/Keyboard.h"
}
function __ZNK4love8keyboard3sdl8Keyboard7getNameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (442406|0); //@line 38 "modules/keyboard/sdl/Keyboard.cpp"
}
function __ZN4love8keyboard3sdl8Keyboard12setKeyRepeatEb($this,$enable) {
 $this = $this|0;
 $enable = $enable|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 43 "modules/keyboard/sdl/Keyboard.cpp"
 $1 = $enable&1; //@line 43 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP8[$0>>0] = $1; //@line 43 "modules/keyboard/sdl/Keyboard.cpp"
 return; //@line 44 "modules/keyboard/sdl/Keyboard.cpp"
}
function __ZNK4love8keyboard3sdl8Keyboard12hasKeyRepeatEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 48 "modules/keyboard/sdl/Keyboard.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 48 "modules/keyboard/sdl/Keyboard.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 48 "modules/keyboard/sdl/Keyboard.cpp"
 return ($2|0); //@line 48 "modules/keyboard/sdl/Keyboard.cpp"
}
function __ZNK4love8keyboard3sdl8Keyboard6isDownERKNSt3__16vectorINS0_8Keyboard3KeyENS3_9allocatorIS6_EEEE($this,$keylist) {
 $this = $this|0;
 $keylist = $keylist|0;
 var $$25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$07 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GetKeyboardState(0)|0); //@line 53 "modules/keyboard/sdl/Keyboard.cpp"
 $1 = HEAP32[$keylist>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($keylist)) + 4|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($1|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($4) {
  $$25 = 0;
  return ($$25|0); //@line 64 "modules/keyboard/sdl/Keyboard.cpp"
 } else {
  $__begin$sroa$0$07 = $1;
 }
 while(1) {
  $7 = HEAP32[$__begin$sroa$0$07>>2]|0; //@line 55 "modules/keyboard/sdl/Keyboard.cpp"
  $8 = HEAP32[159220]|0; //@line 57 "modules/keyboard/sdl/Keyboard.cpp"
  $9 = (($8) + ($7<<2)|0); //@line 57 "modules/keyboard/sdl/Keyboard.cpp"
  $10 = HEAP32[$9>>2]|0; //@line 57 "modules/keyboard/sdl/Keyboard.cpp"
  $11 = (_SDL_GetScancodeFromKey($10)|0); //@line 57 "modules/keyboard/sdl/Keyboard.cpp"
  $12 = (($0) + ($11)|0); //@line 59 "modules/keyboard/sdl/Keyboard.cpp"
  $13 = HEAP8[$12>>0]|0; //@line 59 "modules/keyboard/sdl/Keyboard.cpp"
  $14 = ($13<<24>>24)==(0); //@line 59 "modules/keyboard/sdl/Keyboard.cpp"
  $6 = ((($__begin$sroa$0$07)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if (!($14)) {
   $$25 = 1;
   label = 4;
   break;
  }
  $5 = ($6|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($5) {
   $$25 = 0;
   label = 4;
   break;
  } else {
   $__begin$sroa$0$07 = $6;
  }
 }
 if ((label|0) == 4) {
  return ($$25|0); //@line 64 "modules/keyboard/sdl/Keyboard.cpp"
 }
 return (0)|0;
}
function __ZNK4love8keyboard3sdl8Keyboard14isScancodeDownERKNSt3__16vectorINS0_8Keyboard8ScancodeENS3_9allocatorIS6_EEEE($this,$scancodelist) {
 $this = $this|0;
 $scancodelist = $scancodelist|0;
 var $$25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$07 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GetKeyboardState(0)|0); //@line 68 "modules/keyboard/sdl/Keyboard.cpp"
 $1 = HEAP32[$scancodelist>>2]|0; //@line 1480 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($scancodelist)) + 4|0); //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($1|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 if ($4) {
  $$25 = 0;
  return ($$25|0); //@line 79 "modules/keyboard/sdl/Keyboard.cpp"
 } else {
  $__begin$sroa$0$07 = $1;
 }
 while(1) {
  $5 = HEAP32[$__begin$sroa$0$07>>2]|0; //@line 70 "modules/keyboard/sdl/Keyboard.cpp"
  $6 = ($5>>>0)<(512); //@line 64 "./common/EnumMap.h"
  if ($6) {
   $7 = (((640980 + ($5<<3)|0)) + 4|0); //@line 64 "./common/EnumMap.h"
   $8 = HEAP8[$7>>0]|0; //@line 64 "./common/EnumMap.h"
   $9 = ($8<<24>>24)==(0); //@line 64 "./common/EnumMap.h"
   if (!($9)) {
    $10 = (640980 + ($5<<3)|0); //@line 66 "./common/EnumMap.h"
    $11 = HEAP32[$10>>2]|0; //@line 66 "./common/EnumMap.h"
    $12 = (($0) + ($11)|0); //@line 74 "modules/keyboard/sdl/Keyboard.cpp"
    $13 = HEAP8[$12>>0]|0; //@line 74 "modules/keyboard/sdl/Keyboard.cpp"
    $14 = ($13<<24>>24)==(0); //@line 74 "modules/keyboard/sdl/Keyboard.cpp"
    if (!($14)) {
     $$25 = 1;
     label = 6;
     break;
    }
   }
  }
  $15 = ((($__begin$sroa$0$07)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $16 = ($15|0)==($3|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  if ($16) {
   $$25 = 0;
   label = 6;
   break;
  } else {
   $__begin$sroa$0$07 = $15;
  }
 }
 if ((label|0) == 6) {
  return ($$25|0); //@line 79 "modules/keyboard/sdl/Keyboard.cpp"
 }
 return (0)|0;
}
function __ZNK4love8keyboard3sdl8Keyboard18getKeyFromScancodeENS0_8Keyboard8ScancodeE($this,$scancode) {
 $this = $this|0;
 $scancode = $scancode|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$02 = 0, $sdlscancode$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($scancode>>>0)<(512); //@line 64 "./common/EnumMap.h"
 if ($0) {
  $1 = (((640980 + ($scancode<<3)|0)) + 4|0); //@line 64 "./common/EnumMap.h"
  $2 = HEAP8[$1>>0]|0; //@line 64 "./common/EnumMap.h"
  $3 = ($2<<24>>24)==(0); //@line 64 "./common/EnumMap.h"
  if ($3) {
   $sdlscancode$0 = 0;
  } else {
   $4 = (640980 + ($scancode<<3)|0); //@line 66 "./common/EnumMap.h"
   $5 = HEAP32[$4>>2]|0; //@line 66 "./common/EnumMap.h"
   $sdlscancode$0 = $5;
  }
 } else {
  $sdlscancode$0 = 0;
 }
 $6 = (_SDL_GetKeyFromScancode($sdlscancode$0)|0); //@line 86 "modules/keyboard/sdl/Keyboard.cpp"
 $7 = HEAP32[159220]|0; //@line 90 "modules/keyboard/sdl/Keyboard.cpp"
 $i$02 = 0;
 while(1) {
  $8 = (($7) + ($i$02<<2)|0); //@line 90 "modules/keyboard/sdl/Keyboard.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 90 "modules/keyboard/sdl/Keyboard.cpp"
  $10 = ($9|0)==($6|0); //@line 90 "modules/keyboard/sdl/Keyboard.cpp"
  if ($10) {
   $13 = $i$02;
   label = 7;
   break;
  }
  $11 = (($i$02) + 1)|0; //@line 88 "modules/keyboard/sdl/Keyboard.cpp"
  $12 = ($11|0)<(191); //@line 88 "modules/keyboard/sdl/Keyboard.cpp"
  if ($12) {
   $i$02 = $11;
  } else {
   $13 = 0;
   label = 7;
   break;
  }
 }
 if ((label|0) == 7) {
  return ($13|0); //@line 95 "modules/keyboard/sdl/Keyboard.cpp"
 }
 return (0)|0;
}
function __ZNK4love8keyboard3sdl8Keyboard18getScancodeFromKeyENS0_8Keyboard3KeyE($this,$key) {
 $this = $this|0;
 $key = $key|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scancode$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($key|0)==(191); //@line 101 "modules/keyboard/sdl/Keyboard.cpp"
 if ($0) {
  $scancode$1 = 0;
 } else {
  $1 = HEAP32[159220]|0; //@line 103 "modules/keyboard/sdl/Keyboard.cpp"
  $2 = (($1) + ($key<<2)|0); //@line 103 "modules/keyboard/sdl/Keyboard.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 103 "modules/keyboard/sdl/Keyboard.cpp"
  $4 = (_SDL_GetScancodeFromKey($3)|0); //@line 105 "modules/keyboard/sdl/Keyboard.cpp"
  $5 = ($4>>>0)<(512); //@line 75 "./common/EnumMap.h"
  if ($5) {
   $6 = (((636884 + ($4<<3)|0)) + 4|0); //@line 75 "./common/EnumMap.h"
   $7 = HEAP8[$6>>0]|0; //@line 75 "./common/EnumMap.h"
   $8 = ($7<<24>>24)==(0); //@line 75 "./common/EnumMap.h"
   if ($8) {
    $scancode$1 = 0;
   } else {
    $9 = (636884 + ($4<<3)|0); //@line 77 "./common/EnumMap.h"
    $10 = HEAP32[$9>>2]|0; //@line 77 "./common/EnumMap.h"
    $scancode$1 = $10;
   }
  } else {
   $scancode$1 = 0;
  }
 }
 return ($scancode$1|0); //@line 109 "modules/keyboard/sdl/Keyboard.cpp"
}
function __ZN4love8keyboard3sdl8Keyboard12setTextInputEb($this,$enable) {
 $this = $this|0;
 $enable = $enable|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 if ($enable) {
  _SDL_StartTextInput(); //@line 115 "modules/keyboard/sdl/Keyboard.cpp"
 } else {
  _SDL_StopTextInput(); //@line 117 "modules/keyboard/sdl/Keyboard.cpp"
 }
 return; //@line 118 "modules/keyboard/sdl/Keyboard.cpp"
}
function __ZN4love8keyboard3sdl8Keyboard12setTextInputEbdddd($this,$enable,$x,$y,$w,$h) {
 $this = $this|0;
 $enable = $enable|0;
 $x = +$x;
 $y = +$y;
 $w = +$w;
 $h = +$h;
 var $$pre = 0.0, $$pre1 = 0.0, $$pre2 = 0.0, $$pre3 = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0.0;
 var $23 = 0, $24 = 0, $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $rect = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 24|0;
 $1 = sp + 16|0;
 $2 = sp + 8|0;
 $3 = sp;
 $rect = sp + 32|0;
 HEAPF64[$0>>3] = $x;
 HEAPF64[$1>>3] = $y;
 HEAPF64[$2>>3] = $w;
 HEAPF64[$3>>3] = $h;
 $4 = HEAP32[(625672)>>2]|0; //@line 98 "./common/Module.h"
 $5 = ($4|0)==(0|0); //@line 125 "modules/keyboard/sdl/Keyboard.cpp"
 if ($5) {
  $13 = $x;$16 = $y;$19 = $w;$22 = $h;
 } else {
  $6 = HEAP32[$4>>2]|0; //@line 127 "modules/keyboard/sdl/Keyboard.cpp"
  $7 = ((($6)) + 148|0); //@line 127 "modules/keyboard/sdl/Keyboard.cpp"
  $8 = HEAP32[$7>>2]|0; //@line 127 "modules/keyboard/sdl/Keyboard.cpp"
  FUNCTION_TABLE_viii[$8 & 8191]($4,$0,$1); //@line 127 "modules/keyboard/sdl/Keyboard.cpp"
  $9 = HEAP32[$4>>2]|0; //@line 128 "modules/keyboard/sdl/Keyboard.cpp"
  $10 = ((($9)) + 148|0); //@line 128 "modules/keyboard/sdl/Keyboard.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 128 "modules/keyboard/sdl/Keyboard.cpp"
  FUNCTION_TABLE_viii[$11 & 8191]($4,$2,$3); //@line 128 "modules/keyboard/sdl/Keyboard.cpp"
  $$pre = +HEAPF64[$0>>3]; //@line 131 "modules/keyboard/sdl/Keyboard.cpp"
  $$pre1 = +HEAPF64[$1>>3]; //@line 131 "modules/keyboard/sdl/Keyboard.cpp"
  $$pre2 = +HEAPF64[$2>>3]; //@line 131 "modules/keyboard/sdl/Keyboard.cpp"
  $$pre3 = +HEAPF64[$3>>3]; //@line 131 "modules/keyboard/sdl/Keyboard.cpp"
  $13 = $$pre;$16 = $$pre1;$19 = $$pre2;$22 = $$pre3;
 }
 $12 = (~~(($13))); //@line 131 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[$rect>>2] = $12; //@line 131 "modules/keyboard/sdl/Keyboard.cpp"
 $14 = ((($rect)) + 4|0); //@line 131 "modules/keyboard/sdl/Keyboard.cpp"
 $15 = (~~(($16))); //@line 131 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[$14>>2] = $15; //@line 131 "modules/keyboard/sdl/Keyboard.cpp"
 $17 = ((($rect)) + 8|0); //@line 131 "modules/keyboard/sdl/Keyboard.cpp"
 $18 = (~~(($19))); //@line 131 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[$17>>2] = $18; //@line 131 "modules/keyboard/sdl/Keyboard.cpp"
 $20 = ((($rect)) + 12|0); //@line 131 "modules/keyboard/sdl/Keyboard.cpp"
 $21 = (~~(($22))); //@line 131 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[$20>>2] = $21; //@line 131 "modules/keyboard/sdl/Keyboard.cpp"
 _SDL_SetTextInputRect($rect); //@line 132 "modules/keyboard/sdl/Keyboard.cpp"
 $23 = HEAP32[$this>>2]|0; //@line 134 "modules/keyboard/sdl/Keyboard.cpp"
 $24 = ((($23)) + 48|0); //@line 134 "modules/keyboard/sdl/Keyboard.cpp"
 $25 = HEAP32[$24>>2]|0; //@line 134 "modules/keyboard/sdl/Keyboard.cpp"
 FUNCTION_TABLE_vii[$25 & 8191]($this,$enable); //@line 134 "modules/keyboard/sdl/Keyboard.cpp"
 STACKTOP = sp;return; //@line 135 "modules/keyboard/sdl/Keyboard.cpp"
}
function __ZNK4love8keyboard3sdl8Keyboard12hasTextInputEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_IsTextInputActive()|0); //@line 139 "modules/keyboard/sdl/Keyboard.cpp"
 $1 = ($0|0)!=(0); //@line 139 "modules/keyboard/sdl/Keyboard.cpp"
 return ($1|0); //@line 139 "modules/keyboard/sdl/Keyboard.cpp"
}
function __ZNK4love8keyboard3sdl8Keyboard17hasScreenKeyboardEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_HasScreenKeyboardSupport()|0); //@line 144 "modules/keyboard/sdl/Keyboard.cpp"
 $1 = ($0|0)!=(0); //@line 144 "modules/keyboard/sdl/Keyboard.cpp"
 return ($1|0); //@line 144 "modules/keyboard/sdl/Keyboard.cpp"
}
function __GLOBAL__sub_I_Keyboard_cpp() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i = 0;
 var $i$01$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 (__ZN4love8keyboard3sdl8Keyboard12createKeyMapEv()|0); //@line 370 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[159220] = 645076; //@line 370 "modules/keyboard/sdl/Keyboard.cpp"
 $1 = 636884;
 while(1) {
  $0 = ((($1)) + 4|0); //@line 90 "./common/EnumMap.h"
  HEAP8[$0>>0] = 0; //@line 90 "./common/EnumMap.h"
  $2 = ((($1)) + 8|0); //@line 40 "./common/EnumMap.h"
  $3 = ($2|0)==((640980)|0); //@line 40 "./common/EnumMap.h"
  if ($3) {
   $5 = (640980);
   break;
  } else {
   $1 = $2;
  }
 }
 while(1) {
  $4 = ((($5)) + 4|0); //@line 90 "./common/EnumMap.h"
  HEAP8[$4>>0] = 0; //@line 90 "./common/EnumMap.h"
  $6 = ((($5)) + 8|0); //@line 40 "./common/EnumMap.h"
  $7 = ($6|0)==((645076)|0); //@line 40 "./common/EnumMap.h"
  if ($7) {
   $i$01$i$i = 0;
   break;
  } else {
   $5 = $6;
  }
 }
 while(1) {
  $8 = (27044 + ($i$01$i$i<<3)|0); //@line 46 "./common/EnumMap.h"
  $9 = HEAP32[$8>>2]|0; //@line 46 "./common/EnumMap.h"
  $10 = (((27044 + ($i$01$i$i<<3)|0)) + 4|0); //@line 47 "./common/EnumMap.h"
  $11 = HEAP32[$10>>2]|0; //@line 47 "./common/EnumMap.h"
  $12 = ($9>>>0)<(512); //@line 49 "./common/EnumMap.h"
  if ($12) {
   $13 = (640980 + ($9<<3)|0); //@line 51 "./common/EnumMap.h"
   HEAP32[$13>>2] = $11; //@line 51 "./common/EnumMap.h"
   $14 = (((640980 + ($9<<3)|0)) + 4|0); //@line 52 "./common/EnumMap.h"
   HEAP8[$14>>0] = 1; //@line 52 "./common/EnumMap.h"
  }
  $15 = ($11>>>0)<(512); //@line 54 "./common/EnumMap.h"
  if ($15) {
   $16 = (636884 + ($11<<3)|0); //@line 56 "./common/EnumMap.h"
   HEAP32[$16>>2] = $9; //@line 56 "./common/EnumMap.h"
   $17 = (((636884 + ($11<<3)|0)) + 4|0); //@line 57 "./common/EnumMap.h"
   HEAP8[$17>>0] = 1; //@line 57 "./common/EnumMap.h"
  }
  $18 = (($i$01$i$i) + 1)|0; //@line 44 "./common/EnumMap.h"
  $exitcond$i$i = ($18|0)==(241); //@line 44 "./common/EnumMap.h"
  if ($exitcond$i$i) {
   break;
  } else {
   $i$01$i$i = $18;
  }
 }
 return;
}
function __ZN4love8keyboard3sdl8Keyboard12createKeyMapEv() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[161269] = 0; //@line 162 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645080)>>2] = 13; //@line 164 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645084)>>2] = 27; //@line 165 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645088)>>2] = 8; //@line 166 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645092)>>2] = 9; //@line 167 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645096)>>2] = 32; //@line 168 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645100)>>2] = 33; //@line 169 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645104)>>2] = 34; //@line 170 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645108)>>2] = 35; //@line 171 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645112)>>2] = 37; //@line 172 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645116)>>2] = 36; //@line 173 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645120)>>2] = 38; //@line 174 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645124)>>2] = 39; //@line 175 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645128)>>2] = 40; //@line 176 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645132)>>2] = 41; //@line 177 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645136)>>2] = 42; //@line 178 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645140)>>2] = 43; //@line 179 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645144)>>2] = 44; //@line 180 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645148)>>2] = 45; //@line 181 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645152)>>2] = 46; //@line 182 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645156)>>2] = 47; //@line 183 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645160)>>2] = 48; //@line 184 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645164)>>2] = 49; //@line 185 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645168)>>2] = 50; //@line 186 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645172)>>2] = 51; //@line 187 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645176)>>2] = 52; //@line 188 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645180)>>2] = 53; //@line 189 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645184)>>2] = 54; //@line 190 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645188)>>2] = 55; //@line 191 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645192)>>2] = 56; //@line 192 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645196)>>2] = 57; //@line 193 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645200)>>2] = 58; //@line 194 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645204)>>2] = 59; //@line 195 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645208)>>2] = 60; //@line 196 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645212)>>2] = 61; //@line 197 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645216)>>2] = 62; //@line 198 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645220)>>2] = 63; //@line 199 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645224)>>2] = 64; //@line 200 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645228)>>2] = 91; //@line 202 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645232)>>2] = 92; //@line 203 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645236)>>2] = 93; //@line 204 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645240)>>2] = 94; //@line 205 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645244)>>2] = 95; //@line 206 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645248)>>2] = 96; //@line 207 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645252)>>2] = 97; //@line 208 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645256)>>2] = 98; //@line 209 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645260)>>2] = 99; //@line 210 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645264)>>2] = 100; //@line 211 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645268)>>2] = 101; //@line 212 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645272)>>2] = 102; //@line 213 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645276)>>2] = 103; //@line 214 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645280)>>2] = 104; //@line 215 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645284)>>2] = 105; //@line 216 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645288)>>2] = 106; //@line 217 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645292)>>2] = 107; //@line 218 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645296)>>2] = 108; //@line 219 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645300)>>2] = 109; //@line 220 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645304)>>2] = 110; //@line 221 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645308)>>2] = 111; //@line 222 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645312)>>2] = 112; //@line 223 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645316)>>2] = 113; //@line 224 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645320)>>2] = 114; //@line 225 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645324)>>2] = 115; //@line 226 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645328)>>2] = 116; //@line 227 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645332)>>2] = 117; //@line 228 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645336)>>2] = 118; //@line 229 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645340)>>2] = 119; //@line 230 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645344)>>2] = 120; //@line 231 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645348)>>2] = 121; //@line 232 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645352)>>2] = 122; //@line 233 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645356)>>2] = 1073741881; //@line 235 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645360)>>2] = 1073741882; //@line 237 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645364)>>2] = 1073741883; //@line 238 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645368)>>2] = 1073741884; //@line 239 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645372)>>2] = 1073741885; //@line 240 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645376)>>2] = 1073741886; //@line 241 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645380)>>2] = 1073741887; //@line 242 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645384)>>2] = 1073741888; //@line 243 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645388)>>2] = 1073741889; //@line 244 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645392)>>2] = 1073741890; //@line 245 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645396)>>2] = 1073741891; //@line 246 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645400)>>2] = 1073741892; //@line 247 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645404)>>2] = 1073741893; //@line 248 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645408)>>2] = 1073741894; //@line 250 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645412)>>2] = 1073741895; //@line 251 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645416)>>2] = 1073741896; //@line 252 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645420)>>2] = 1073741897; //@line 253 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645424)>>2] = 1073741898; //@line 254 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645428)>>2] = 1073741899; //@line 255 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645432)>>2] = 127; //@line 256 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645436)>>2] = 1073741901; //@line 257 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645440)>>2] = 1073741902; //@line 258 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645444)>>2] = 1073741903; //@line 259 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645448)>>2] = 1073741904; //@line 260 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645452)>>2] = 1073741905; //@line 261 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645456)>>2] = 1073741906; //@line 262 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645460)>>2] = 1073741907; //@line 264 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645464)>>2] = 1073741908; //@line 265 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645468)>>2] = 1073741909; //@line 266 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645472)>>2] = 1073741910; //@line 267 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645476)>>2] = 1073741911; //@line 268 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645480)>>2] = 1073741912; //@line 269 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645520)>>2] = 1073741922; //@line 270 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645484)>>2] = 1073741913; //@line 271 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645488)>>2] = 1073741914; //@line 272 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645492)>>2] = 1073741915; //@line 273 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645496)>>2] = 1073741916; //@line 274 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645500)>>2] = 1073741917; //@line 275 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645504)>>2] = 1073741918; //@line 276 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645508)>>2] = 1073741919; //@line 277 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645512)>>2] = 1073741920; //@line 278 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645516)>>2] = 1073741921; //@line 279 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645524)>>2] = 1073741923; //@line 280 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645528)>>2] = 1073741957; //@line 281 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645532)>>2] = 1073741927; //@line 282 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645536)>>2] = 1073741925; //@line 284 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645540)>>2] = 1073741926; //@line 285 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645544)>>2] = 1073741928; //@line 286 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645548)>>2] = 1073741929; //@line 287 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645552)>>2] = 1073741930; //@line 288 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645556)>>2] = 1073741931; //@line 289 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645560)>>2] = 1073741932; //@line 290 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645564)>>2] = 1073741933; //@line 291 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645568)>>2] = 1073741934; //@line 292 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645572)>>2] = 1073741935; //@line 293 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645576)>>2] = 1073741936; //@line 294 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645580)>>2] = 1073741937; //@line 295 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645584)>>2] = 1073741938; //@line 296 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645588)>>2] = 1073741939; //@line 297 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645592)>>2] = 1073741940; //@line 298 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645596)>>2] = 1073741941; //@line 299 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645600)>>2] = 1073741942; //@line 300 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645604)>>2] = 1073741943; //@line 301 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645608)>>2] = 1073741944; //@line 302 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645612)>>2] = 1073741945; //@line 303 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645616)>>2] = 1073741946; //@line 304 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645620)>>2] = 1073741947; //@line 305 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645624)>>2] = 1073741948; //@line 306 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645628)>>2] = 1073741949; //@line 307 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645632)>>2] = 1073741950; //@line 308 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645636)>>2] = 1073741951; //@line 309 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645640)>>2] = 1073741952; //@line 310 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645644)>>2] = 1073741953; //@line 311 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645648)>>2] = 1073741977; //@line 313 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645652)>>2] = 1073741978; //@line 314 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645656)>>2] = 1073741979; //@line 315 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645660)>>2] = 1073741980; //@line 316 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645664)>>2] = 1073741981; //@line 317 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645668)>>2] = 1073741982; //@line 318 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645672)>>2] = 1073741983; //@line 319 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645676)>>2] = 1073741984; //@line 320 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645680)>>2] = 1073741985; //@line 321 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645684)>>2] = 1073741986; //@line 322 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645688)>>2] = 1073742002; //@line 324 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645692)>>2] = 1073742003; //@line 325 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645696)>>2] = 1073742004; //@line 326 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645700)>>2] = 1073742005; //@line 327 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645704)>>2] = 1073742048; //@line 329 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645708)>>2] = 1073742049; //@line 330 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645712)>>2] = 1073742050; //@line 331 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645716)>>2] = 1073742051; //@line 332 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645720)>>2] = 1073742052; //@line 333 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645724)>>2] = 1073742053; //@line 334 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645728)>>2] = 1073742054; //@line 335 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645732)>>2] = 1073742055; //@line 336 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645736)>>2] = 1073742081; //@line 338 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645740)>>2] = 1073742082; //@line 340 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645744)>>2] = 1073742083; //@line 341 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645748)>>2] = 1073742084; //@line 342 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645752)>>2] = 1073742085; //@line 343 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645756)>>2] = 1073742086; //@line 344 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645760)>>2] = 1073742087; //@line 345 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645764)>>2] = 1073742088; //@line 346 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645768)>>2] = 1073742089; //@line 347 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645772)>>2] = 1073742090; //@line 348 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645776)>>2] = 1073742091; //@line 349 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645780)>>2] = 1073742092; //@line 350 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645784)>>2] = 1073742093; //@line 351 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645788)>>2] = 1073742094; //@line 352 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645792)>>2] = 1073742095; //@line 353 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645796)>>2] = 1073742096; //@line 354 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645800)>>2] = 1073742097; //@line 355 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645804)>>2] = 1073742098; //@line 356 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645808)>>2] = 1073742099; //@line 358 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645812)>>2] = 1073742100; //@line 359 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645816)>>2] = 1073742101; //@line 360 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645820)>>2] = 1073742102; //@line 361 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645824)>>2] = 1073742103; //@line 362 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645828)>>2] = 1073742104; //@line 363 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645832)>>2] = 1073742105; //@line 364 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP32[(645836)>>2] = 1073742106; //@line 365 "modules/keyboard/sdl/Keyboard.cpp"
 return (645076|0); //@line 367 "modules/keyboard/sdl/Keyboard.cpp"
}
function __ZN4love8keyboard3sdl8KeyboardC2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./common/Module.h"
 HEAP32[$this>>2] = (26980); //@line 33 "modules/keyboard/sdl/Keyboard.cpp"
 $0 = ((($this)) + 8|0); //@line 32 "modules/keyboard/sdl/Keyboard.cpp"
 HEAP8[$0>>0] = 0; //@line 32 "modules/keyboard/sdl/Keyboard.cpp"
 return; //@line 34 "modules/keyboard/sdl/Keyboard.cpp"
}
function __ZN4love8keyboard3sdl8Keyboard11getConstantE12SDL_ScancodeRNS0_8Keyboard8ScancodeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)<(512); //@line 75 "./common/EnumMap.h"
 if ($0) {
  $1 = (((636884 + ($in<<3)|0)) + 4|0); //@line 75 "./common/EnumMap.h"
  $2 = HEAP8[$1>>0]|0; //@line 75 "./common/EnumMap.h"
  $3 = ($2<<24>>24)==(0); //@line 75 "./common/EnumMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   $4 = (636884 + ($in<<3)|0); //@line 77 "./common/EnumMap.h"
   $5 = HEAP32[$4>>2]|0; //@line 77 "./common/EnumMap.h"
   HEAP32[$out>>2] = $5; //@line 77 "./common/EnumMap.h"
   $$0$i = 1;
  }
 } else {
  $$0$i = 0;
 }
 return ($$0$i|0); //@line 154 "modules/keyboard/sdl/Keyboard.cpp"
}
function __GLOBAL__sub_I_Keyboard_cpp_4545() {
 var $$02$i$i$i$i = 0, $$02$i$i$i$i4 = 0, $$lcssa56 = 0, $$lcssa58 = 0, $$lcssa61 = 0, $$lcssa63 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i = 0, $exitcond$i$i10 = 0, $hash$0$lcssa$i$i$i$i = 0, $hash$0$lcssa$i$i$i$i7 = 0, $hash$01$i$i$i$i = 0, $hash$01$i$i$i$i5 = 0, $i$08$i$i$i = 0, $i$08$i$i$i8 = 0, $i1$05$i$i = 0, $i1$05$i$i2 = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $1 = 645840;
 while(1) {
  $0 = ((($1)) + 8|0); //@line 155 "./common/StringMap.h"
  HEAP8[$0>>0] = 0; //@line 155 "./common/StringMap.h"
  $2 = ((($1)) + 12|0); //@line 40 "./common/StringMap.h"
  $3 = ($2|0)==((650424)|0); //@line 40 "./common/StringMap.h"
  if ($3) {
   break;
  } else {
   $1 = $2;
  }
 }
 _memset(((650424)|0),0,764)|0; //@line 44 "./common/StringMap.h"
 $i1$05$i$i = 0;
 while(1) {
  $4 = (28972 + ($i1$05$i$i<<3)|0); //@line 49 "./common/StringMap.h"
  $5 = HEAP32[$4>>2]|0; //@line 49 "./common/StringMap.h"
  $6 = (((28972 + ($i1$05$i$i<<3)|0)) + 4|0); //@line 49 "./common/StringMap.h"
  $7 = HEAP32[$6>>2]|0; //@line 49 "./common/StringMap.h"
  $8 = HEAP8[$5>>0]|0; //@line 142 "./common/StringMap.h"
  $9 = ($8<<24>>24)==(0); //@line 142 "./common/StringMap.h"
  if ($9) {
   $hash$0$lcssa$i$i$i$i = 5381;
  } else {
   $$02$i$i$i$i = $5;$11 = $8;$hash$01$i$i$i$i = 5381;
   while(1) {
    $10 = $11 << 24 >> 24; //@line 142 "./common/StringMap.h"
    $12 = ((($$02$i$i$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
    $13 = ($hash$01$i$i$i$i*33)|0; //@line 143 "./common/StringMap.h"
    $14 = (($13) + ($10))|0; //@line 143 "./common/StringMap.h"
    $15 = HEAP8[$12>>0]|0; //@line 142 "./common/StringMap.h"
    $16 = ($15<<24>>24)==(0); //@line 142 "./common/StringMap.h"
    if ($16) {
     $hash$0$lcssa$i$i$i$i = $14;
     break;
    } else {
     $$02$i$i$i$i = $12;$11 = $15;$hash$01$i$i$i$i = $14;
    }
   }
  }
  $i$08$i$i$i = 0;
  while(1) {
   $19 = (($i$08$i$i$i) + ($hash$0$lcssa$i$i$i$i))|0; //@line 112 "./common/StringMap.h"
   $20 = (($19>>>0) % 382)&-1; //@line 112 "./common/StringMap.h"
   $21 = (((645840 + (($20*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $22 = HEAP8[$21>>0]|0; //@line 114 "./common/StringMap.h"
   $23 = ($22<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   $18 = (($i$08$i$i$i) + 1)|0; //@line 110 "./common/StringMap.h"
   if ($23) {
    $$lcssa61 = $20;$$lcssa63 = $21;
    label = 9;
    break;
   }
   $17 = ($18>>>0)<(382); //@line 110 "./common/StringMap.h"
   if ($17) {
    $i$08$i$i$i = $18;
   } else {
    break;
   }
  }
  if ((label|0) == 9) {
   label = 0;
   HEAP8[$$lcssa63>>0] = 1; //@line 117 "./common/StringMap.h"
   $24 = (645840 + (($$lcssa61*12)|0)|0); //@line 118 "./common/StringMap.h"
   HEAP32[$24>>2] = $5; //@line 118 "./common/StringMap.h"
   $25 = (((645840 + (($$lcssa61*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
   HEAP32[$25>>2] = $7; //@line 119 "./common/StringMap.h"
  }
  $26 = ($7>>>0)>(190); //@line 126 "./common/StringMap.h"
  if ($26) {
   HEAP32[$vararg_buffer>>2] = $5; //@line 128 "./common/StringMap.h"
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 128 "./common/StringMap.h"
   HEAP32[$vararg_ptr1>>2] = $7; //@line 128 "./common/StringMap.h"
   (_printf(443858,$vararg_buffer)|0); //@line 128 "./common/StringMap.h"
  } else {
   $27 = (650424 + ($7<<2)|0); //@line 132 "./common/StringMap.h"
   HEAP32[$27>>2] = $5; //@line 132 "./common/StringMap.h"
  }
  $28 = (($i1$05$i$i) + 1)|0; //@line 48 "./common/StringMap.h"
  $exitcond$i$i = ($28|0)==(191); //@line 48 "./common/StringMap.h"
  if ($exitcond$i$i) {
   $30 = 651188;
   break;
  } else {
   $i1$05$i$i = $28;
  }
 }
 while(1) {
  $29 = ((($30)) + 8|0); //@line 155 "./common/StringMap.h"
  HEAP8[$29>>0] = 0; //@line 155 "./common/StringMap.h"
  $31 = ((($30)) + 12|0); //@line 40 "./common/StringMap.h"
  $32 = ($31|0)==((656972)|0); //@line 40 "./common/StringMap.h"
  if ($32) {
   break;
  } else {
   $30 = $31;
  }
 }
 _memset(((656972)|0),0,964)|0; //@line 44 "./common/StringMap.h"
 $i1$05$i$i2 = 0;
 while(1) {
  $33 = (30500 + ($i1$05$i$i2<<3)|0); //@line 49 "./common/StringMap.h"
  $34 = HEAP32[$33>>2]|0; //@line 49 "./common/StringMap.h"
  $35 = (((30500 + ($i1$05$i$i2<<3)|0)) + 4|0); //@line 49 "./common/StringMap.h"
  $36 = HEAP32[$35>>2]|0; //@line 49 "./common/StringMap.h"
  $37 = HEAP8[$34>>0]|0; //@line 142 "./common/StringMap.h"
  $38 = ($37<<24>>24)==(0); //@line 142 "./common/StringMap.h"
  if ($38) {
   $hash$0$lcssa$i$i$i$i7 = 5381;
  } else {
   $$02$i$i$i$i4 = $34;$40 = $37;$hash$01$i$i$i$i5 = 5381;
   while(1) {
    $39 = $40 << 24 >> 24; //@line 142 "./common/StringMap.h"
    $41 = ((($$02$i$i$i$i4)) + 1|0); //@line 142 "./common/StringMap.h"
    $42 = ($hash$01$i$i$i$i5*33)|0; //@line 143 "./common/StringMap.h"
    $43 = (($42) + ($39))|0; //@line 143 "./common/StringMap.h"
    $44 = HEAP8[$41>>0]|0; //@line 142 "./common/StringMap.h"
    $45 = ($44<<24>>24)==(0); //@line 142 "./common/StringMap.h"
    if ($45) {
     $hash$0$lcssa$i$i$i$i7 = $43;
     break;
    } else {
     $$02$i$i$i$i4 = $41;$40 = $44;$hash$01$i$i$i$i5 = $43;
    }
   }
  }
  $i$08$i$i$i8 = 0;
  while(1) {
   $48 = (($i$08$i$i$i8) + ($hash$0$lcssa$i$i$i$i7))|0; //@line 112 "./common/StringMap.h"
   $49 = (($48>>>0) % 482)&-1; //@line 112 "./common/StringMap.h"
   $50 = (((651188 + (($49*12)|0)|0)) + 8|0); //@line 114 "./common/StringMap.h"
   $51 = HEAP8[$50>>0]|0; //@line 114 "./common/StringMap.h"
   $52 = ($51<<24>>24)==(0); //@line 114 "./common/StringMap.h"
   $47 = (($i$08$i$i$i8) + 1)|0; //@line 110 "./common/StringMap.h"
   if ($52) {
    $$lcssa56 = $49;$$lcssa58 = $50;
    label = 21;
    break;
   }
   $46 = ($47>>>0)<(482); //@line 110 "./common/StringMap.h"
   if ($46) {
    $i$08$i$i$i8 = $47;
   } else {
    break;
   }
  }
  if ((label|0) == 21) {
   label = 0;
   HEAP8[$$lcssa58>>0] = 1; //@line 117 "./common/StringMap.h"
   $53 = (651188 + (($$lcssa56*12)|0)|0); //@line 118 "./common/StringMap.h"
   HEAP32[$53>>2] = $34; //@line 118 "./common/StringMap.h"
   $54 = (((651188 + (($$lcssa56*12)|0)|0)) + 4|0); //@line 119 "./common/StringMap.h"
   HEAP32[$54>>2] = $36; //@line 119 "./common/StringMap.h"
  }
  $55 = ($36>>>0)>(240); //@line 126 "./common/StringMap.h"
  if ($55) {
   HEAP32[$vararg_buffer2>>2] = $34; //@line 128 "./common/StringMap.h"
   $vararg_ptr5 = ((($vararg_buffer2)) + 4|0); //@line 128 "./common/StringMap.h"
   HEAP32[$vararg_ptr5>>2] = $36; //@line 128 "./common/StringMap.h"
   (_printf(443858,$vararg_buffer2)|0); //@line 128 "./common/StringMap.h"
  } else {
   $56 = (656972 + ($36<<2)|0); //@line 132 "./common/StringMap.h"
   HEAP32[$56>>2] = $34; //@line 132 "./common/StringMap.h"
  }
  $57 = (($i1$05$i$i2) + 1)|0; //@line 48 "./common/StringMap.h"
  $exitcond$i$i10 = ($57|0)==(241); //@line 48 "./common/StringMap.h"
  if ($exitcond$i$i10) {
   break;
  } else {
   $i1$05$i$i2 = $57;
  }
 }
 STACKTOP = sp;return;
}
function __ZN4love8keyboard8Keyboard11getConstantEPKcRNS1_3KeyE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 382)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((645840 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (645840 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(382); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((645840 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 32 "modules/keyboard/Keyboard.cpp"
  return ($$$2$i|0); //@line 32 "modules/keyboard/Keyboard.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 32 "modules/keyboard/Keyboard.cpp"
  return ($$$2$i|0); //@line 32 "modules/keyboard/Keyboard.cpp"
 }
 return (0)|0;
}
function __ZN4love8keyboard8Keyboard11getConstantENS1_3KeyERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(190); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (650424 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 37 "modules/keyboard/Keyboard.cpp"
}
function __ZN4love8keyboard8Keyboard11getConstantEPKcRNS1_8ScancodeE($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$$2$i = 0, $$011$i$i = 0, $$0110$i$i = 0, $$02$i$i = 0, $$08$ph$i$i = 0, $$13$i = 0, $$2$i = 0, $$lcssa17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $hash$0$lcssa$i$i = 0, $hash$01$i$i = 0, $i$010$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$in>>0]|0; //@line 142 "./common/StringMap.h"
 $1 = ($0<<24>>24)==(0); //@line 142 "./common/StringMap.h"
 if ($1) {
  $hash$0$lcssa$i$i = 5381;
 } else {
  $$02$i$i = $in;$3 = $0;$hash$01$i$i = 5381;
  while(1) {
   $2 = $3 << 24 >> 24; //@line 142 "./common/StringMap.h"
   $4 = ((($$02$i$i)) + 1|0); //@line 142 "./common/StringMap.h"
   $5 = ($hash$01$i$i*33)|0; //@line 143 "./common/StringMap.h"
   $6 = (($5) + ($2))|0; //@line 143 "./common/StringMap.h"
   $7 = HEAP8[$4>>0]|0; //@line 142 "./common/StringMap.h"
   $8 = ($7<<24>>24)==(0); //@line 142 "./common/StringMap.h"
   if ($8) {
    $hash$0$lcssa$i$i = $6;
    break;
   } else {
    $$02$i$i = $4;$3 = $7;$hash$01$i$i = $6;
   }
  }
 }
 $i$010$i = 0;
 while(1) {
  $9 = (($i$010$i) + ($hash$0$lcssa$i$i))|0; //@line 72 "./common/StringMap.h"
  $10 = (($9>>>0) % 482)&-1; //@line 72 "./common/StringMap.h"
  $11 = (((651188 + (($10*12)|0)|0)) + 8|0); //@line 74 "./common/StringMap.h"
  $12 = HEAP8[$11>>0]|0; //@line 74 "./common/StringMap.h"
  $13 = ($12<<24>>24)==(0); //@line 74 "./common/StringMap.h"
  if ($13) {
   $$13$i = 1;$$2$i = 0;
   label = 11;
   break;
  }
  $14 = (651188 + (($10*12)|0)|0); //@line 77 "./common/StringMap.h"
  $15 = HEAP32[$14>>2]|0; //@line 77 "./common/StringMap.h"
  $16 = HEAP8[$15>>0]|0; //@line 54 "./common/StringMap.h"
  $17 = ($16<<24>>24)==(0); //@line 54 "./common/StringMap.h"
  L8: do {
   if ($17) {
    $$08$ph$i$i = $in;
    label = 8;
   } else {
    $$011$i$i = $in;$$0110$i$i = $15;$21 = $16;
    while(1) {
     $18 = HEAP8[$$011$i$i>>0]|0; //@line 56 "./common/StringMap.h"
     $19 = ($18<<24>>24)!=(0); //@line 54 "./common/StringMap.h"
     $20 = ($21<<24>>24)==($18<<24>>24); //@line 56 "./common/StringMap.h"
     $or$cond$i$i = $19 & $20; //@line 54 "./common/StringMap.h"
     if (!($or$cond$i$i)) {
      break L8;
     }
     $22 = ((($$0110$i$i)) + 1|0); //@line 59 "./common/StringMap.h"
     $23 = ((($$011$i$i)) + 1|0); //@line 60 "./common/StringMap.h"
     $24 = HEAP8[$22>>0]|0; //@line 54 "./common/StringMap.h"
     $25 = ($24<<24>>24)==(0); //@line 54 "./common/StringMap.h"
     if ($25) {
      $$08$ph$i$i = $23;
      label = 8;
      break;
     } else {
      $$011$i$i = $23;$$0110$i$i = $22;$21 = $24;
     }
    }
   }
  } while(0);
  if ((label|0) == 8) {
   label = 0;
   $26 = HEAP8[$$08$ph$i$i>>0]|0; //@line 63 "./common/StringMap.h"
   $27 = ($26<<24>>24)==(0); //@line 63 "./common/StringMap.h"
   if ($27) {
    $$lcssa17 = $10;
    label = 9;
    break;
   }
  }
  $30 = (($i$010$i) + 1)|0; //@line 70 "./common/StringMap.h"
  $31 = ($30>>>0)<(482); //@line 70 "./common/StringMap.h"
  if ($31) {
   $i$010$i = $30;
  } else {
   $$13$i = 0;$$2$i = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = (((651188 + (($$lcssa17*12)|0)|0)) + 4|0); //@line 79 "./common/StringMap.h"
  $29 = HEAP32[$28>>2]|0; //@line 79 "./common/StringMap.h"
  HEAP32[$out>>2] = $29; //@line 79 "./common/StringMap.h"
  $$13$i = 1;$$2$i = 1;
  $$$2$i = $$13$i & $$2$i; //@line 42 "modules/keyboard/Keyboard.cpp"
  return ($$$2$i|0); //@line 42 "modules/keyboard/Keyboard.cpp"
 }
 else if ((label|0) == 11) {
  $$$2$i = $$13$i & $$2$i; //@line 42 "modules/keyboard/Keyboard.cpp"
  return ($$$2$i|0); //@line 42 "modules/keyboard/Keyboard.cpp"
 }
 return (0)|0;
}
function __ZN4love8keyboard8Keyboard11getConstantENS1_8ScancodeERPKc($in,$out) {
 $in = $in|0;
 $out = $out|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in>>>0)>(240); //@line 91 "./common/StringMap.h"
 if ($0) {
  $$0$i = 0;
 } else {
  $1 = (656972 + ($in<<2)|0); //@line 94 "./common/StringMap.h"
  $2 = HEAP32[$1>>2]|0; //@line 94 "./common/StringMap.h"
  $3 = ($2|0)==(0|0); //@line 94 "./common/StringMap.h"
  if ($3) {
   $$0$i = 0;
  } else {
   HEAP32[$out>>2] = $2; //@line 96 "./common/StringMap.h"
   $$0$i = 1;
  }
 }
 return ($$0$i|0); //@line 47 "modules/keyboard/Keyboard.cpp"
}
function __ZN4love4math11BezierCurveC2ERKNSt3__16vectorINS_6VectorENS2_9allocatorIS4_EEEE($this,$pts) {
 $this = $this|0;
 $pts = $pts|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 86 "modules/math/BezierCurve.cpp"
 HEAP32[$this>>2] = (32436); //@line 88 "modules/math/BezierCurve.cpp"
 $0 = ((($this)) + 8|0); //@line 87 "modules/math/BezierCurve.cpp"
 __THREW__ = 0;
 invoke_vii(4203,($0|0),($pts|0)); //@line 87 "modules/math/BezierCurve.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  ___resumeException($3|0);
  // unreachable;
 } else {
  return; //@line 89 "modules/math/BezierCurve.cpp"
 }
}
function __ZN4love4math11BezierCurveD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (32436); //@line 34 "modules/math/BezierCurve.h"
 $0 = ((($this)) + 8|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ($1|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = $1; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($2) {
  return; //@line 34 "modules/math/BezierCurve.h"
 }
 $4 = ((($this)) + 12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5|0)==($1|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($6)) {
  $scevgep$i$i$i = ((($5)) + -8|0);
  $7 = $scevgep$i$i$i;
  $8 = (($7) - ($3))|0;
  $9 = $8 >>> 3;
  $10 = $9 ^ -1;
  $scevgep4$i$i$i = (($5) + ($10<<3)|0);
  HEAP32[$4>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($1); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 34 "modules/math/BezierCurve.h"
}
function __ZN4love4math11BezierCurveD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i$i = 0, $scevgep4$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (32436); //@line 34 "modules/math/BezierCurve.h"
 $0 = ((($this)) + 8|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ($1|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = $1; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($2) {
  __ZdlPv($this); //@line 34 "modules/math/BezierCurve.h"
  return; //@line 34 "modules/math/BezierCurve.h"
 }
 $4 = ((($this)) + 12|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = HEAP32[$4>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5|0)==($1|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($6)) {
  $scevgep$i$i$i$i = ((($5)) + -8|0);
  $7 = $scevgep$i$i$i$i;
  $8 = (($7) - ($3))|0;
  $9 = $8 >>> 3;
  $10 = $9 ^ -1;
  $scevgep4$i$i$i$i = (($5) + ($10<<3)|0);
  HEAP32[$4>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($1); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 __ZdlPv($this); //@line 34 "modules/math/BezierCurve.h"
 return; //@line 34 "modules/math/BezierCurve.h"
}
function __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEEC2ERKS5_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$01$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $0 = ((($this)) + 4|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($this)) + 8|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($__x)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$__x>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = $5 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = ($6|0)==(0); //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($7) {
  return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $8 = ($6>>>0)>(536870911); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($8) {
  __THREW__ = 0;
  invoke_vi(3528,($this|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if (!($10)) {
   label = 4;
  }
 } else {
  label = 4;
 }
 if ((label|0) == 4) {
  __THREW__ = 0;
  $11 = (invoke_ii(3348,($5|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   HEAP32[$0>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$this>>2] = $11; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $14 = (($11) + ($6<<3)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $15 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$15>>2] = $14; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $16 = HEAP32[$__x>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $17 = HEAP32[$2>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $18 = ($16|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($18) {
    return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $$01$i = $16;$26 = $11;
   while(1) {
    $19 = $$01$i; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $20 = $19; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $21 = HEAP32[$20>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $22 = (($19) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $23 = $22; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $24 = HEAP32[$23>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $25 = $26; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $27 = $25; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$27>>2] = $21; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $28 = (($25) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $29 = $28; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$29>>2] = $24; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $30 = HEAP32[$0>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $31 = ((($30)) + 8|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$0>>2] = $31; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $32 = ((($$01$i)) + 8|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $33 = ($32|0)==($17|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($33) {
     break;
    } else {
     $$01$i = $32;$26 = $31;
    }
   }
   return; //@line 1217 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
 }
 $34 = ___cxa_find_matching_catch_2()|0;
 $35 = tempRet0;
 $36 = HEAP32[$this>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $37 = ($36|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $38 = $36; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($37) {
  ___resumeException($34|0);
  // unreachable;
 }
 $39 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $40 = ($39|0)==($36|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($40)) {
  $scevgep$i$i$i = ((($39)) + -8|0);
  $41 = $scevgep$i$i$i;
  $42 = (($41) - ($38))|0;
  $43 = $42 >>> 3;
  $44 = $43 ^ -1;
  $scevgep4$i$i$i = (($39) + ($44<<3)|0);
  HEAP32[$0>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($36); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($34|0);
 // unreachable;
}
function __ZNK4love4math11BezierCurve13getDerivativeEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$0 = 0, $$0$i$i = 0, $$01 = 0, $$01$i$i = 0, $$pre = 0, $$pre20 = 0, $$pre22 = 0, $$pre23 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0.0, $112 = 0, $113 = 0.0, $114 = 0.0, $115 = 0, $116 = 0.0, $117 = 0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0.0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $forward_differences = 0, $i$018 = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i$i = 0, $scevgep$i$i$i$i8 = 0, $scevgep$i$i$i2 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i$i = 0, $scevgep4$i$i$i$i9 = 0, $scevgep4$i$i$i3 = 0;
 var $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $forward_differences = sp + 4|0;
 $0 = ((($this)) + 8|0); //@line 48 "modules/math/BezierCurve.h"
 $1 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($2) - ($3))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($4|0)==(8); //@line 94 "modules/math/BezierCurve.cpp"
 if ($5) {
  $6 = (___cxa_allocate_exception(16)|0); //@line 95 "modules/math/BezierCurve.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($6|0),(443920|0),($vararg_buffer|0)); //@line 95 "modules/math/BezierCurve.cpp"
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if (!($8)) {
   ___cxa_throw(($6|0),(96|0),(344|0)); //@line 95 "modules/math/BezierCurve.cpp"
   // unreachable; //@line 95 "modules/math/BezierCurve.cpp"
  }
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  ___cxa_free_exception(($6|0)); //@line 95 "modules/math/BezierCurve.cpp"
  $$0 = $9;$$01 = $10;
  ___resumeException($$0|0);
  // unreachable;
 }
 $11 = $4 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = (($11) + -1)|0; //@line 98 "modules/math/BezierCurve.cpp"
 HEAP32[$forward_differences>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ((($forward_differences)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$13>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = ((($forward_differences)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$14>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $15 = ($12|0)==(0); //@line 1080 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 L8: do {
  if ($15) {
   $53 = $13;$55 = $forward_differences;
  } else {
   $16 = ($12>>>0)>(536870911); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($16) {
    __THREW__ = 0;
    invoke_vi(3528,($forward_differences|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if (!($18)) {
     label = 9;
    }
   } else {
    label = 9;
   }
   if ((label|0) == 9) {
    $19 = $12 << 3; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    __THREW__ = 0;
    $20 = (invoke_ii(3348,($19|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $21 = __THREW__; __THREW__ = 0;
    $22 = $21&1;
    if (!($22)) {
     HEAP32[$13>>2] = $20; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$forward_differences>>2] = $20; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $23 = (($20) + ($12<<3)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $24 = ((($forward_differences)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$24>>2] = $23; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $$0$i$i = $12;$25 = $20;
     while(1) {
      HEAPF32[$25>>2] = 0.0; //@line 219 "./common/Vector.h"
      $26 = ((($25)) + 4|0); //@line 220 "./common/Vector.h"
      HEAPF32[$26>>2] = 0.0; //@line 220 "./common/Vector.h"
      $27 = ((($25)) + 8|0); //@line 989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $28 = (($$0$i$i) + -1)|0; //@line 990 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $29 = ($28|0)==(0); //@line 992 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($29) {
       break;
      } else {
       $$0$i$i = $28;$25 = $27;
      }
     }
     HEAP32[$13>>2] = $23; //@line 988 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $$pre = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $$pre20 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $41 = (($$pre) - ($$pre20))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $42 = $41 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $43 = (($42) + -1)|0; //@line 48 "modules/math/BezierCurve.h"
     $44 = (+($43>>>0)); //@line 99 "modules/math/BezierCurve.cpp"
     $45 = ($23|0)==($20|0); //@line 100 "modules/math/BezierCurve.cpp"
     if ($45) {
      $53 = $13;$55 = $forward_differences;
      break;
     }
     $46 = $$pre20; //@line 94 "modules/math/BezierCurve.cpp"
     $110 = $46;$123 = $20;$i$018 = 0;
     while(1) {
      $108 = (($i$018) + 1)|0; //@line 101 "modules/math/BezierCurve.cpp"
      $109 = (($110) + ($108<<3)|0); //@line 237 "./common/Vector.h"
      $111 = +HEAPF32[$109>>2]; //@line 237 "./common/Vector.h"
      $112 = (($110) + ($i$018<<3)|0); //@line 305 "./common/Vector.h"
      $113 = +HEAPF32[$112>>2]; //@line 305 "./common/Vector.h"
      $114 = $111 - $113; //@line 237 "./common/Vector.h"
      $115 = (((($110) + ($108<<3)|0)) + 4|0); //@line 237 "./common/Vector.h"
      $116 = +HEAPF32[$115>>2]; //@line 237 "./common/Vector.h"
      $117 = (((($110) + ($i$018<<3)|0)) + 4|0); //@line 310 "./common/Vector.h"
      $118 = +HEAPF32[$117>>2]; //@line 310 "./common/Vector.h"
      $119 = $116 - $118; //@line 237 "./common/Vector.h"
      $120 = $44 * $114; //@line 242 "./common/Vector.h"
      $121 = $44 * $119; //@line 242 "./common/Vector.h"
      $122 = (($123) + ($i$018<<3)|0); //@line 101 "modules/math/BezierCurve.cpp"
      HEAPF32[$122>>2] = $120; //@line 101 "modules/math/BezierCurve.cpp"
      $124 = (((($123) + ($i$018<<3)|0)) + 4|0); //@line 101 "modules/math/BezierCurve.cpp"
      HEAPF32[$124>>2] = $121; //@line 101 "modules/math/BezierCurve.cpp"
      $125 = HEAP32[$13>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $126 = HEAP32[$forward_differences>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $127 = (($125) - ($126))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $128 = $127 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $129 = ($108>>>0)<($128>>>0); //@line 100 "modules/math/BezierCurve.cpp"
      if (!($129)) {
       $53 = $13;$55 = $forward_differences;
       break L8;
      }
      $130 = $126; //@line 100 "modules/math/BezierCurve.cpp"
      $$pre22 = HEAP32[$0>>2]|0; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $110 = $$pre22;$123 = $130;$i$018 = $108;
     }
    }
   }
   $30 = ___cxa_find_matching_catch_2()|0;
   $31 = tempRet0;
   $32 = HEAP32[$forward_differences>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $33 = ($32|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $34 = $32; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($33) {
    ___resumeException($30|0);
    // unreachable;
   }
   $35 = HEAP32[$13>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $36 = ($35|0)==($32|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($36)) {
    $scevgep$i$i$i$i = ((($35)) + -8|0);
    $37 = $scevgep$i$i$i$i;
    $38 = (($37) - ($34))|0;
    $39 = $38 >>> 3;
    $40 = $39 ^ -1;
    $scevgep4$i$i$i$i = (($35) + ($40<<3)|0);
    HEAP32[$13>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($32); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($30|0);
   // unreachable;
  }
 } while(0);
 __THREW__ = 0;
 invoke_vi(3431,($agg$result|0)); //@line 86 "modules/math/BezierCurve.cpp"
 $47 = __THREW__; __THREW__ = 0;
 $48 = $47&1;
 L32: do {
  if ($48) {
   $97 = ___cxa_find_matching_catch_2()|0;
   $98 = tempRet0;
   $eh$lpad$body$index2Z2D = $98;$eh$lpad$body$indexZ2D = $97;
  } else {
   HEAP32[$agg$result>>2] = (32436); //@line 88 "modules/math/BezierCurve.cpp"
   $49 = ((($agg$result)) + 8|0); //@line 87 "modules/math/BezierCurve.cpp"
   HEAP32[$49>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $50 = ((($agg$result)) + 12|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$50>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $51 = ((($agg$result)) + 16|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$51>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $52 = HEAP32[$53>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $54 = HEAP32[$55>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $56 = (($52) - ($54))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $57 = $56 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $58 = ($57|0)==(0); //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $59 = $54; //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   do {
    if ($58) {
     $132 = $59;
    } else {
     $60 = ($57>>>0)>(536870911); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($60) {
      __THREW__ = 0;
      invoke_vi(3528,($49|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $61 = __THREW__; __THREW__ = 0;
      $62 = $61&1;
      if (!($62)) {
       label = 23;
      }
     } else {
      label = 23;
     }
     if ((label|0) == 23) {
      __THREW__ = 0;
      $63 = (invoke_ii(3348,($56|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $64 = __THREW__; __THREW__ = 0;
      $65 = $64&1;
      if (!($65)) {
       HEAP32[$50>>2] = $63; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$49>>2] = $63; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $66 = (($63) + ($57<<3)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $67 = ((($agg$result)) + 16|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       HEAP32[$67>>2] = $66; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $68 = HEAP32[$forward_differences>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $69 = HEAP32[$13>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $70 = ($68|0)==($69|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($70) {
        $132 = $68;
        break;
       }
       $$01$i$i = $68;$78 = $63;
       while(1) {
        $71 = $$01$i$i; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $72 = $71; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $73 = HEAP32[$72>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $74 = (($71) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $75 = $74; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $76 = HEAP32[$75>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $77 = $78; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $79 = $77; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        HEAP32[$79>>2] = $73; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $80 = (($77) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $81 = $80; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        HEAP32[$81>>2] = $76; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $82 = HEAP32[$50>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $83 = ((($82)) + 8|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$50>>2] = $83; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $84 = ((($$01$i$i)) + 8|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $85 = ($84|0)==($69|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($85) {
         break;
        } else {
         $$01$i$i = $84;$78 = $83;
        }
       }
       $$pre23 = HEAP32[$forward_differences>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $132 = $$pre23;
       break;
      }
     }
     $86 = ___cxa_find_matching_catch_2()|0;
     $87 = tempRet0;
     $88 = HEAP32[$49>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $89 = ($88|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $90 = $88; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($89)) {
      $91 = HEAP32[$50>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $92 = ($91|0)==($88|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($92)) {
       $scevgep$i$i$i$i8 = ((($91)) + -8|0);
       $93 = $scevgep$i$i$i$i8;
       $94 = (($93) - ($90))|0;
       $95 = $94 >>> 3;
       $96 = $95 ^ -1;
       $scevgep4$i$i$i$i9 = (($91) + ($96<<3)|0);
       HEAP32[$50>>2] = $scevgep4$i$i$i$i9; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($88); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $eh$lpad$body$index2Z2D = $87;$eh$lpad$body$indexZ2D = $86;
     break L32;
    }
   } while(0);
   $131 = ($132|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $133 = $132; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($131) {
    STACKTOP = sp;return; //@line 104 "modules/math/BezierCurve.cpp"
   }
   $134 = HEAP32[$13>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $135 = ($134|0)==($132|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($135)) {
    $scevgep$i$i$i = ((($134)) + -8|0);
    $136 = $scevgep$i$i$i;
    $137 = (($136) - ($133))|0;
    $138 = $137 >>> 3;
    $139 = $138 ^ -1;
    $scevgep4$i$i$i = (($134) + ($139<<3)|0);
    HEAP32[$13>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($132); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   STACKTOP = sp;return; //@line 104 "modules/math/BezierCurve.cpp"
  }
 } while(0);
 $99 = HEAP32[$forward_differences>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $100 = ($99|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $101 = $99; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($100) {
  $$0 = $eh$lpad$body$indexZ2D;$$01 = $eh$lpad$body$index2Z2D;
  ___resumeException($$0|0);
  // unreachable;
 }
 $102 = HEAP32[$13>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $103 = ($102|0)==($99|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($103)) {
  $scevgep$i$i$i2 = ((($102)) + -8|0);
  $104 = $scevgep$i$i$i2;
  $105 = (($104) - ($101))|0;
  $106 = $105 >>> 3;
  $107 = $106 ^ -1;
  $scevgep4$i$i$i3 = (($102) + ($107<<3)|0);
  HEAP32[$13>>2] = $scevgep4$i$i$i3; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($99); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $$0 = $eh$lpad$body$indexZ2D;$$01 = $eh$lpad$body$index2Z2D;
 ___resumeException($$0|0);
 // unreachable;
}
function __ZNK4love4math11BezierCurve15getControlPointEi($this,$i) {
 $this = $this|0;
 $i = $i|0;
 var $$0$lcssa = 0, $$01 = 0, $$1 = 0, $$1$lcssa = 0, $$in = 0, $$phi$trans$insert = 0, $$phi$trans$insert3 = 0, $$pre = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($i|0)<(0); //@line 108 "modules/math/BezierCurve.cpp"
 if ($0) {
  $1 = ((($this)) + 8|0); //@line 109 "modules/math/BezierCurve.cpp"
  $2 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $6 = $5 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$01 = $i;
  while(1) {
   $11 = (($6) + ($$01))|0; //@line 109 "modules/math/BezierCurve.cpp"
   $12 = ($11|0)<(0); //@line 108 "modules/math/BezierCurve.cpp"
   if ($12) {
    $$01 = $11;
   } else {
    $$0$lcssa = $11;$$in = $4;$9 = $3;
    break;
   }
  }
 } else {
  $$phi$trans$insert = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$phi$trans$insert3 = ((($this)) + 8|0); //@line 111 "modules/math/BezierCurve.cpp"
  $$pre5 = HEAP32[$$phi$trans$insert3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$lcssa = $i;$$in = $$pre5;$9 = $$pre;
 }
 $7 = $$in; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = (($9) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $8 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $$1 = $$0$lcssa;
 while(1) {
  $13 = ($$1>>>0)<($10>>>0); //@line 111 "modules/math/BezierCurve.cpp"
  $14 = (($$1) - ($10))|0; //@line 112 "modules/math/BezierCurve.cpp"
  if ($13) {
   $$1$lcssa = $$1;
   break;
  } else {
   $$1 = $14;
  }
 }
 $15 = (($7) + ($$1$lcssa<<3)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 return ($15|0); //@line 114 "modules/math/BezierCurve.cpp"
}
function __ZN4love4math11BezierCurve15setControlPointEiRKNS_6VectorE($this,$i,$point) {
 $this = $this|0;
 $i = $i|0;
 $point = $point|0;
 var $$0$lcssa = 0, $$01 = 0, $$1 = 0, $$1$lcssa = 0, $$in = 0, $$phi$trans$insert = 0, $$phi$trans$insert3 = 0, $$pre = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($i|0)<(0); //@line 119 "modules/math/BezierCurve.cpp"
 if ($0) {
  $1 = ((($this)) + 8|0); //@line 120 "modules/math/BezierCurve.cpp"
  $2 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $6 = $5 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$01 = $i;
  while(1) {
   $11 = (($6) + ($$01))|0; //@line 120 "modules/math/BezierCurve.cpp"
   $12 = ($11|0)<(0); //@line 119 "modules/math/BezierCurve.cpp"
   if ($12) {
    $$01 = $11;
   } else {
    $$0$lcssa = $11;$$in = $4;$9 = $3;
    break;
   }
  }
 } else {
  $$phi$trans$insert = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$phi$trans$insert3 = ((($this)) + 8|0); //@line 122 "modules/math/BezierCurve.cpp"
  $$pre5 = HEAP32[$$phi$trans$insert3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$lcssa = $i;$$in = $$pre5;$9 = $$pre;
 }
 $7 = $$in; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = (($9) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $8 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $$1 = $$0$lcssa;
 while(1) {
  $13 = ($$1>>>0)<($10>>>0); //@line 122 "modules/math/BezierCurve.cpp"
  $14 = (($$1) - ($10))|0; //@line 123 "modules/math/BezierCurve.cpp"
  if ($13) {
   $$1$lcssa = $$1;
   break;
  } else {
   $$1 = $14;
  }
 }
 $15 = (($7) + ($$1$lcssa<<3)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = $point; //@line 125 "modules/math/BezierCurve.cpp"
 $17 = $16; //@line 125 "modules/math/BezierCurve.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 125 "modules/math/BezierCurve.cpp"
 $19 = (($16) + 4)|0; //@line 125 "modules/math/BezierCurve.cpp"
 $20 = $19; //@line 125 "modules/math/BezierCurve.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 125 "modules/math/BezierCurve.cpp"
 $22 = $15; //@line 125 "modules/math/BezierCurve.cpp"
 $23 = $22; //@line 125 "modules/math/BezierCurve.cpp"
 HEAP32[$23>>2] = $18; //@line 125 "modules/math/BezierCurve.cpp"
 $24 = (($22) + 4)|0; //@line 125 "modules/math/BezierCurve.cpp"
 $25 = $24; //@line 125 "modules/math/BezierCurve.cpp"
 HEAP32[$25>>2] = $21; //@line 125 "modules/math/BezierCurve.cpp"
 return; //@line 126 "modules/math/BezierCurve.cpp"
}
function __ZN4love4math11BezierCurve18insertControlPointERKNS_6VectorEi($this,$point,$i) {
 $this = $this|0;
 $point = $point|0;
 $i = $i|0;
 var $$0$lcssa = 0, $$02 = 0, $$1 = 0, $$1$lcssa = 0, $$byval_copy = 0, $$in = 0, $$phi$trans$insert = 0, $$phi$trans$insert4 = 0, $$pre = 0, $$pre$phiZ2D = 0, $$pre6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = ($i|0)<(0); //@line 130 "modules/math/BezierCurve.cpp"
 if ($1) {
  $2 = ((($this)) + 8|0); //@line 131 "modules/math/BezierCurve.cpp"
  $3 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = HEAP32[$3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $5 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $6 = (($4) - ($5))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $7 = $6 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$02 = $i;
  while(1) {
   $12 = (($7) + ($$02))|0; //@line 131 "modules/math/BezierCurve.cpp"
   $13 = ($12|0)<(0); //@line 130 "modules/math/BezierCurve.cpp"
   if ($13) {
    $$02 = $12;
   } else {
    $$0$lcssa = $12;$$in = $5;$$pre$phiZ2D = $2;$10 = $4;
    break;
   }
  }
 } else {
  $$phi$trans$insert = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$phi$trans$insert4 = ((($this)) + 8|0); //@line 133 "modules/math/BezierCurve.cpp"
  $$pre6 = HEAP32[$$phi$trans$insert4>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$lcssa = $i;$$in = $$pre6;$$pre$phiZ2D = $$phi$trans$insert4;$10 = $$pre;
 }
 $8 = $$in; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($10) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $9 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $$1 = $$0$lcssa;
 while(1) {
  $14 = ($$1>>>0)>($11>>>0); //@line 133 "modules/math/BezierCurve.cpp"
  $15 = (($$1) - ($11))|0; //@line 134 "modules/math/BezierCurve.cpp"
  if ($14) {
   $$1 = $15;
  } else {
   $$1$lcssa = $$1;
   break;
  }
 }
 $16 = (($8) + ($$1$lcssa<<3)|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $17 = $16; //@line 136 "modules/math/BezierCurve.cpp"
 HEAP32[$0>>2] = $17; //@line 1128 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0; //@line 136 "modules/math/BezierCurve.cpp"
 (__ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE6insertENS_11__wrap_iterIPKS2_EERS7_($$pre$phiZ2D,$$byval_copy,$point)|0); //@line 136 "modules/math/BezierCurve.cpp"
 STACKTOP = sp;return; //@line 137 "modules/math/BezierCurve.cpp"
}
function __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE6insertENS_11__wrap_iterIPKS2_EERS7_($this,$__position,$__x) {
 $this = $this|0;
 $__position = $__position|0;
 $__x = $__x|0;
 var $$ = 0, $$0$i8 = 0, $$0$i9 = 0, $$in$i7 = 0, $$pre = 0, $$pre$phiZ2D = 0, $$pre11 = 0, $$pre12 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__i$01$i = 0, $__p$0 = 0, $__v$sroa$21$0 = 0, $__v$sroa$37$0 = 0, $__xr$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 1727 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = $0; //@line 1727 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$__position>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $3 = (($2) - ($1))|0; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $4 = $3 >> 3; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $5 = (($0) + ($4<<3)|0); //@line 1727 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ((($this)) + 4|0); //@line 1728 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = HEAP32[$6>>2]|0; //@line 1728 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $9 = HEAP32[$8>>2]|0; //@line 1728 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = ($7>>>0)<($9>>>0); //@line 1728 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = $7; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = $9; //@line 1728 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($10) {
  $13 = ($5|0)==($7|0); //@line 1731 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($13) {
   $14 = $__x; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $15 = $14; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $16 = HEAP32[$15>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $17 = (($14) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $18 = $17; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $19 = HEAP32[$18>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $20 = $5; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $21 = $20; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$21>>2] = $16; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $22 = (($20) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $23 = $22; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$23>>2] = $19; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $24 = HEAP32[$6>>2]|0; //@line 1735 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $25 = ((($24)) + 8|0); //@line 1735 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$6>>2] = $25; //@line 1735 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $__p$0 = $5;
   return ($__p$0|0); //@line 1755 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  $26 = ((($5)) + 8|0); //@line 1739 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $27 = $26; //@line 1710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $28 = (($11) - ($27))|0; //@line 1710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $29 = $28 >> 3; //@line 1710 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $30 = (($5) + ($29<<3)|0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $31 = ($30>>>0)<($7>>>0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($31) {
   $$in$i7 = $7;$__i$01$i = $30;
   while(1) {
    $32 = $__i$01$i; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $33 = $32; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $34 = HEAP32[$33>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $35 = (($32) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $36 = $35; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $37 = HEAP32[$36>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $38 = $$in$i7; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $39 = $38; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$39>>2] = $34; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $40 = (($38) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $41 = $40; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$41>>2] = $37; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $42 = ((($__i$01$i)) + 8|0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $43 = HEAP32[$6>>2]|0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $44 = ((($43)) + 8|0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$6>>2] = $44; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $45 = ($42>>>0)<($7>>>0); //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($45) {
     $$in$i7 = $44;$__i$01$i = $42;
    } else {
     break;
    }
   }
  }
  $46 = (0 - ($29))|0; //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $47 = (($7) + ($46<<3)|0); //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  _memmove(($47|0),($5|0),($28|0))|0; //@line 1929 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $48 = ($5>>>0)>($__x>>>0); //@line 1741 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($48) {
   $__xr$0 = $__x;
  } else {
   $49 = HEAP32[$6>>2]|0; //@line 1741 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $50 = ($49>>>0)>($__x>>>0); //@line 1741 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $51 = ((($__x)) + 8|0); //@line 1742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$ = $50 ? $51 : $__x; //@line 1741 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $__xr$0 = $$;
  }
  $52 = $__xr$0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $53 = $52; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $54 = HEAP32[$53>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $55 = (($52) + 4)|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $56 = $55; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $57 = HEAP32[$56>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $58 = $5; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $59 = $58; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$59>>2] = $54; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $60 = (($58) + 4)|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $61 = $60; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$61>>2] = $57; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $__p$0 = $5;
  return ($__p$0|0); //@line 1755 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $62 = (($11) - ($1))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $63 = $62 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $64 = (($63) + 1)|0; //@line 1750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $65 = ($64>>>0)>(536870911); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($65) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$8>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre11 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $66 = $$pre11; //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre$phiZ2D = $8;$140 = $66;$68 = $$pre;$69 = $$pre11;
 } else {
  $$pre$phiZ2D = $8;$140 = $0;$68 = $12;$69 = $1;
 }
 $67 = (($68) - ($69))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $70 = $67 >> 3; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $71 = ($70>>>0)<(268435455); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($71) {
  $75 = $67 >> 2; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $76 = ($75>>>0)<($64>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $77 = $76 ? $64 : $75; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $78 = $5; //@line 1750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $79 = (($78) - ($69))|0; //@line 1750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $80 = $79 >> 3; //@line 1750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $81 = ($77|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($81) {
   $$0$i9 = 0;$125 = $78;$85 = 0;$86 = $80;
  } else {
   $$0$i8 = $77;$141 = $80;$142 = $78;
   label = 15;
  }
 } else {
  $72 = $5; //@line 1750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $73 = (($72) - ($69))|0; //@line 1750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $74 = $73 >> 3; //@line 1750 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i8 = 536870911;$141 = $74;$142 = $72;
  label = 15;
 }
 if ((label|0) == 15) {
  $82 = $$0$i8 << 3; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $83 = (__Znwj($82)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i9 = $$0$i8;$125 = $142;$85 = $83;$86 = $141;
 }
 $84 = (($85) + ($86<<3)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $87 = $84; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $88 = (($85) + ($$0$i9<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $89 = $88; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $90 = ($86|0)==($$0$i9|0); //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 do {
  if ($90) {
   $91 = ($86|0)>(0); //@line 550 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($91) {
    $92 = (($86) + 1)|0; //@line 553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $93 = (($92|0) / -2)&-1; //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $94 = (($84) + ($93<<3)|0); //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $95 = $94; //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $116 = $94;$123 = $140;$__v$sroa$21$0 = $95;$__v$sroa$37$0 = $89;
    break;
   }
   $96 = $86 << 1; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $97 = ($86|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $98 = $97 ? 1 : $96; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $99 = $98 << 3; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   __THREW__ = 0;
   $100 = (invoke_ii(3348,($99|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $101 = __THREW__; __THREW__ = 0;
   $102 = $101&1;
   if (!($102)) {
    $103 = $98 >>> 2; //@line 560 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $104 = (($100) + ($103<<3)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $105 = $104; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $106 = (($100) + ($98<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $107 = $106; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $108 = ($85|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($108) {
     $116 = $104;$123 = $140;$__v$sroa$21$0 = $105;$__v$sroa$37$0 = $107;
     break;
    }
    __ZdlPv($85); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $$pre12 = HEAP32[$this>>2]|0; //@line 912 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $116 = $104;$123 = $$pre12;$__v$sroa$21$0 = $105;$__v$sroa$37$0 = $107;
    break;
   }
   $137 = ___cxa_find_matching_catch_2()|0;
   $138 = tempRet0;
   $139 = ($85|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($139) {
    ___resumeException($137|0);
    // unreachable;
   }
   __ZdlPv($85); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   ___resumeException($137|0);
   // unreachable;
  } else {
   $116 = $84;$123 = $140;$__v$sroa$21$0 = $87;$__v$sroa$37$0 = $89;
  }
 } while(0);
 $109 = $__x; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $110 = $109; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $111 = HEAP32[$110>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $112 = (($109) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $113 = $112; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $114 = HEAP32[$113>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $115 = $116; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $117 = $115; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$117>>2] = $111; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $118 = (($115) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $119 = $118; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$119>>2] = $114; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $120 = $__v$sroa$21$0; //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $121 = ((($120)) + 8|0); //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $122 = $123; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $124 = (($125) - ($122))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $126 = $124 >> 3; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $127 = (0 - ($126))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $128 = (($120) + ($127<<3)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $129 = $128; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($128|0),($123|0),($124|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $130 = HEAP32[$6>>2]|0; //@line 913 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $131 = (($130) - ($125))|0; //@line 1522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $132 = $131 >> 3; //@line 1522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($121|0),($5|0),($131|0))|0; //@line 1523 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $133 = (($121) + ($132<<3)|0); //@line 1524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $134 = $133; //@line 1524 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $135 = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$this>>2] = $129; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$6>>2] = $134; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$$pre$phiZ2D>>2] = $__v$sroa$37$0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $136 = ($135|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($136) {
  $__p$0 = $120;
  return ($__p$0|0); //@line 1755 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($135); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $__p$0 = $120;
 return ($__p$0|0); //@line 1755 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love4math11BezierCurve18removeControlPointEi($this,$i) {
 $this = $this|0;
 $i = $i|0;
 var $$0$lcssa = 0, $$06 = 0, $$1 = 0, $$1$lcssa = 0, $$in = 0, $$phi$trans$insert = 0, $$phi$trans$insert8 = 0, $$pre = 0, $$pre$phi11Z2D = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__new_last2$i$i$i = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($i|0)<(0); //@line 141 "modules/math/BezierCurve.cpp"
 if ($0) {
  $1 = ((($this)) + 8|0); //@line 142 "modules/math/BezierCurve.cpp"
  $2 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $3 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $4 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $5 = (($3) - ($4))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $6 = $5 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$06 = $i;
  while(1) {
   $11 = (($6) + ($$06))|0; //@line 142 "modules/math/BezierCurve.cpp"
   $12 = ($11|0)<(0); //@line 141 "modules/math/BezierCurve.cpp"
   if ($12) {
    $$06 = $11;
   } else {
    $$0$lcssa = $11;$$in = $4;$$pre$phi11Z2D = $2;$9 = $3;
    break;
   }
  }
 } else {
  $$phi$trans$insert = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$phi$trans$insert8 = ((($this)) + 8|0); //@line 144 "modules/math/BezierCurve.cpp"
  $$pre10 = HEAP32[$$phi$trans$insert8>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$lcssa = $i;$$in = $$pre10;$$pre$phi11Z2D = $$phi$trans$insert;$9 = $$pre;
 }
 $7 = $$in; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = (($9) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $8 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $$1 = $$0$lcssa;
 while(1) {
  $13 = ($$1>>>0)<($10>>>0); //@line 144 "modules/math/BezierCurve.cpp"
  $14 = (($$1) - ($10))|0; //@line 145 "modules/math/BezierCurve.cpp"
  if ($13) {
   $$1$lcssa = $$1;
   break;
  } else {
   $$1 = $14;
  }
 }
 $15 = (($7) + ($$1$lcssa<<3)|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $16 = ((($15)) + 8|0); //@line 1684 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = $16; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $18 = (($9) - ($17))|0; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $19 = $18 >> 3; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 _memmove(($15|0),($16|0),($18|0))|0; //@line 1893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $20 = (($15) + ($19<<3)|0); //@line 1894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $21 = HEAP32[$$pre$phi11Z2D>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $22 = ($21|0)==($20|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($22) {
  return; //@line 148 "modules/math/BezierCurve.cpp"
 }
 $__new_last2$i$i$i = $20; //@line 818 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $scevgep$i$i$i = ((($21)) + -8|0);
 $23 = $scevgep$i$i$i;
 $24 = (($23) - ($__new_last2$i$i$i))|0;
 $25 = $24 >>> 3;
 $26 = $25 ^ -1;
 $scevgep4$i$i$i = (($21) + ($26<<3)|0);
 HEAP32[$$pre$phi11Z2D>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 return; //@line 148 "modules/math/BezierCurve.cpp"
}
function __ZN4love4math11BezierCurve9translateERKNS_6VectorE($this,$t) {
 $this = $this|0;
 $t = $t|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, $i$01 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 152 "modules/math/BezierCurve.cpp"
 $1 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($2|0)==($3|0); //@line 152 "modules/math/BezierCurve.cpp"
 $5 = $3; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($4) {
  return; //@line 154 "modules/math/BezierCurve.cpp"
 }
 $6 = ((($t)) + 4|0); //@line 310 "./common/Vector.h"
 $7 = (($2) - ($3))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = $7 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $i$01 = 0;
 while(1) {
  $9 = +HEAPF32[$t>>2]; //@line 305 "./common/Vector.h"
  $10 = (($5) + ($i$01<<3)|0); //@line 257 "./common/Vector.h"
  $11 = +HEAPF32[$10>>2]; //@line 257 "./common/Vector.h"
  $12 = $9 + $11; //@line 257 "./common/Vector.h"
  HEAPF32[$10>>2] = $12; //@line 257 "./common/Vector.h"
  $13 = +HEAPF32[$6>>2]; //@line 310 "./common/Vector.h"
  $14 = (((($5) + ($i$01<<3)|0)) + 4|0); //@line 258 "./common/Vector.h"
  $15 = +HEAPF32[$14>>2]; //@line 258 "./common/Vector.h"
  $16 = $13 + $15; //@line 258 "./common/Vector.h"
  HEAPF32[$14>>2] = $16; //@line 258 "./common/Vector.h"
  $17 = (($i$01) + 1)|0; //@line 152 "modules/math/BezierCurve.cpp"
  $18 = ($17>>>0)<($8>>>0); //@line 152 "modules/math/BezierCurve.cpp"
  if ($18) {
   $i$01 = $17;
  } else {
   break;
  }
 }
 return; //@line 154 "modules/math/BezierCurve.cpp"
}
function __ZN4love4math11BezierCurve6rotateEdRKNS_6VectorE($this,$phi,$center) {
 $this = $this|0;
 $phi = +$phi;
 $center = $center|0;
 var $0 = 0.0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0.0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0.0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+Math_cos((+$phi))); //@line 158 "modules/math/BezierCurve.cpp"
 $1 = $0; //@line 158 "modules/math/BezierCurve.cpp"
 $2 = (+Math_sin((+$phi))); //@line 158 "modules/math/BezierCurve.cpp"
 $3 = $2; //@line 158 "modules/math/BezierCurve.cpp"
 $4 = ((($this)) + 8|0); //@line 159 "modules/math/BezierCurve.cpp"
 $5 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$5>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = HEAP32[$4>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ($6|0)==($7|0); //@line 159 "modules/math/BezierCurve.cpp"
 $9 = $7; //@line 159 "modules/math/BezierCurve.cpp"
 if ($8) {
  return; //@line 165 "modules/math/BezierCurve.cpp"
 }
 $10 = ((($center)) + 4|0); //@line 310 "./common/Vector.h"
 $11 = (($6) - ($7))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = $11 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $i$01 = 0;
 while(1) {
  $13 = (($9) + ($i$01<<3)|0); //@line 237 "./common/Vector.h"
  $14 = +HEAPF32[$13>>2]; //@line 237 "./common/Vector.h"
  $15 = +HEAPF32[$center>>2]; //@line 305 "./common/Vector.h"
  $16 = $14 - $15; //@line 237 "./common/Vector.h"
  $17 = (((($9) + ($i$01<<3)|0)) + 4|0); //@line 237 "./common/Vector.h"
  $18 = +HEAPF32[$17>>2]; //@line 237 "./common/Vector.h"
  $19 = +HEAPF32[$10>>2]; //@line 310 "./common/Vector.h"
  $20 = $18 - $19; //@line 237 "./common/Vector.h"
  $21 = $1 * $16; //@line 162 "modules/math/BezierCurve.cpp"
  $22 = $3 * $20; //@line 162 "modules/math/BezierCurve.cpp"
  $23 = $21 - $22; //@line 162 "modules/math/BezierCurve.cpp"
  $24 = $15 + $23; //@line 162 "modules/math/BezierCurve.cpp"
  HEAPF32[$13>>2] = $24; //@line 162 "modules/math/BezierCurve.cpp"
  $25 = $3 * $16; //@line 163 "modules/math/BezierCurve.cpp"
  $26 = $1 * $20; //@line 163 "modules/math/BezierCurve.cpp"
  $27 = $25 + $26; //@line 163 "modules/math/BezierCurve.cpp"
  $28 = $19 + $27; //@line 163 "modules/math/BezierCurve.cpp"
  HEAPF32[$17>>2] = $28; //@line 163 "modules/math/BezierCurve.cpp"
  $29 = (($i$01) + 1)|0; //@line 159 "modules/math/BezierCurve.cpp"
  $30 = ($29>>>0)<($12>>>0); //@line 159 "modules/math/BezierCurve.cpp"
  if ($30) {
   $i$01 = $29;
  } else {
   break;
  }
 }
 return; //@line 165 "modules/math/BezierCurve.cpp"
}
function __ZN4love4math11BezierCurve5scaleEdRKNS_6VectorE($this,$s,$center) {
 $this = $this|0;
 $s = +$s;
 $center = $center|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0.0, $i$05 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 169 "modules/math/BezierCurve.cpp"
 $1 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ($2|0)==($3|0); //@line 169 "modules/math/BezierCurve.cpp"
 if ($4) {
  return; //@line 171 "modules/math/BezierCurve.cpp"
 }
 $5 = ((($center)) + 4|0); //@line 310 "./common/Vector.h"
 $6 = $s; //@line 170 "modules/math/BezierCurve.cpp"
 $$in = $3;$i$05 = 0;
 while(1) {
  $7 = $$in; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $8 = (($7) + ($i$05<<3)|0); //@line 237 "./common/Vector.h"
  $9 = +HEAPF32[$8>>2]; //@line 237 "./common/Vector.h"
  $10 = +HEAPF32[$center>>2]; //@line 305 "./common/Vector.h"
  $11 = $9 - $10; //@line 237 "./common/Vector.h"
  $12 = (((($7) + ($i$05<<3)|0)) + 4|0); //@line 237 "./common/Vector.h"
  $13 = +HEAPF32[$12>>2]; //@line 237 "./common/Vector.h"
  $14 = +HEAPF32[$5>>2]; //@line 310 "./common/Vector.h"
  $15 = $13 - $14; //@line 237 "./common/Vector.h"
  $16 = $6 * $11; //@line 242 "./common/Vector.h"
  $17 = $6 * $15; //@line 242 "./common/Vector.h"
  $18 = $10 + $16; //@line 232 "./common/Vector.h"
  $19 = $14 + $17; //@line 232 "./common/Vector.h"
  HEAPF32[$8>>2] = $18; //@line 170 "modules/math/BezierCurve.cpp"
  HEAPF32[$12>>2] = $19; //@line 170 "modules/math/BezierCurve.cpp"
  $20 = (($i$05) + 1)|0; //@line 169 "modules/math/BezierCurve.cpp"
  $21 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $22 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = (($21) - ($22))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $24 = $23 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $25 = ($20>>>0)<($24>>>0); //@line 169 "modules/math/BezierCurve.cpp"
  if ($25) {
   $$in = $22;$i$05 = $20;
  } else {
   break;
  }
 }
 return; //@line 171 "modules/math/BezierCurve.cpp"
}
function __ZNK4love4math11BezierCurve8evaluateEd($agg$result,$this,$t) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $t = +$t;
 var $$0 = 0, $$02 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0.0;
 var $25 = 0.0, $26 = 0.0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0.0;
 var $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0.0, $65 = 0, $66 = 0, $67 = 0.0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $9 = 0, $i$07 = 0, $or$cond = 0, $points = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, $step$08 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $points = sp + 12|0;
 $0 = $t < 0.0; //@line 175 "modules/math/BezierCurve.cpp"
 $1 = $t > 1.0; //@line 175 "modules/math/BezierCurve.cpp"
 $or$cond = $0 | $1; //@line 175 "modules/math/BezierCurve.cpp"
 if ($or$cond) {
  $2 = (___cxa_allocate_exception(16)|0); //@line 176 "modules/math/BezierCurve.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($2|0),(443957|0),($vararg_buffer|0)); //@line 176 "modules/math/BezierCurve.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if (!($4)) {
   ___cxa_throw(($2|0),(96|0),(344|0)); //@line 176 "modules/math/BezierCurve.cpp"
   // unreachable; //@line 176 "modules/math/BezierCurve.cpp"
  }
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  ___cxa_free_exception(($2|0)); //@line 176 "modules/math/BezierCurve.cpp"
  $$0 = $6;$$02 = $5;
  ___resumeException($$02|0);
  // unreachable;
 }
 $7 = ((($this)) + 8|0); //@line 177 "modules/math/BezierCurve.cpp"
 $8 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = HEAP32[$8>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = HEAP32[$7>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = (($9) - ($10))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $12 = $11 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $13 = ($12>>>0)<(2); //@line 177 "modules/math/BezierCurve.cpp"
 if ($13) {
  $14 = (___cxa_allocate_exception(16)|0); //@line 178 "modules/math/BezierCurve.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($14|0),(444011|0),($vararg_buffer1|0)); //@line 178 "modules/math/BezierCurve.cpp"
  $15 = __THREW__; __THREW__ = 0;
  $16 = $15&1;
  if (!($16)) {
   ___cxa_throw(($14|0),(96|0),(344|0)); //@line 178 "modules/math/BezierCurve.cpp"
   // unreachable; //@line 178 "modules/math/BezierCurve.cpp"
  }
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  ___cxa_free_exception(($14|0)); //@line 178 "modules/math/BezierCurve.cpp"
  $$0 = $18;$$02 = $17;
  ___resumeException($$02|0);
  // unreachable;
 }
 __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEEC2ERKS5_($points,$7); //@line 181 "modules/math/BezierCurve.cpp"
 $19 = HEAP32[$8>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = HEAP32[$7>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = (($19) - ($20))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $22 = $21 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $23 = ($22>>>0)>(1); //@line 182 "modules/math/BezierCurve.cpp"
 if ($23) {
  $24 = 1.0 - $t; //@line 184 "modules/math/BezierCurve.cpp"
  $25 = $24; //@line 184 "modules/math/BezierCurve.cpp"
  $26 = $t; //@line 184 "modules/math/BezierCurve.cpp"
  $28 = $19;$29 = $20;$step$08 = 1;
  while(1) {
   $27 = (($28) - ($29))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $30 = $27 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $31 = ($30|0)==($step$08|0); //@line 183 "modules/math/BezierCurve.cpp"
   if ($31) {
    $55 = $28;$56 = $29;
   } else {
    $32 = HEAP32[$points>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $i$07 = 0;
    while(1) {
     $59 = (($32) + ($i$07<<3)|0); //@line 242 "./common/Vector.h"
     $60 = +HEAPF32[$59>>2]; //@line 242 "./common/Vector.h"
     $61 = $25 * $60; //@line 242 "./common/Vector.h"
     $62 = (((($32) + ($i$07<<3)|0)) + 4|0); //@line 242 "./common/Vector.h"
     $63 = +HEAPF32[$62>>2]; //@line 242 "./common/Vector.h"
     $64 = $25 * $63; //@line 242 "./common/Vector.h"
     $65 = (($i$07) + 1)|0; //@line 184 "modules/math/BezierCurve.cpp"
     $66 = (($32) + ($65<<3)|0); //@line 242 "./common/Vector.h"
     $67 = +HEAPF32[$66>>2]; //@line 242 "./common/Vector.h"
     $68 = $26 * $67; //@line 242 "./common/Vector.h"
     $69 = (((($32) + ($65<<3)|0)) + 4|0); //@line 242 "./common/Vector.h"
     $70 = +HEAPF32[$69>>2]; //@line 242 "./common/Vector.h"
     $71 = $26 * $70; //@line 242 "./common/Vector.h"
     $72 = $61 + $68; //@line 232 "./common/Vector.h"
     $73 = $64 + $71; //@line 232 "./common/Vector.h"
     HEAPF32[$59>>2] = $72; //@line 184 "modules/math/BezierCurve.cpp"
     HEAPF32[$62>>2] = $73; //@line 184 "modules/math/BezierCurve.cpp"
     $74 = HEAP32[$8>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $75 = HEAP32[$7>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $76 = (($74) - ($75))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $77 = $76 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $78 = (($77) - ($step$08))|0; //@line 183 "modules/math/BezierCurve.cpp"
     $79 = ($65>>>0)<($78>>>0); //@line 183 "modules/math/BezierCurve.cpp"
     if ($79) {
      $i$07 = $65;
     } else {
      $55 = $74;$56 = $75;
      break;
     }
    }
   }
   $53 = (($step$08) + 1)|0; //@line 182 "modules/math/BezierCurve.cpp"
   $54 = (($55) - ($56))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $57 = $54 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $58 = ($53>>>0)<($57>>>0); //@line 182 "modules/math/BezierCurve.cpp"
   if ($58) {
    $28 = $55;$29 = $56;$step$08 = $53;
   } else {
    break;
   }
  }
 }
 $33 = HEAP32[$points>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $34 = $33; //@line 186 "modules/math/BezierCurve.cpp"
 $35 = $34; //@line 186 "modules/math/BezierCurve.cpp"
 $36 = HEAP32[$35>>2]|0; //@line 186 "modules/math/BezierCurve.cpp"
 $37 = (($34) + 4)|0; //@line 186 "modules/math/BezierCurve.cpp"
 $38 = $37; //@line 186 "modules/math/BezierCurve.cpp"
 $39 = HEAP32[$38>>2]|0; //@line 186 "modules/math/BezierCurve.cpp"
 $40 = $agg$result; //@line 186 "modules/math/BezierCurve.cpp"
 $41 = $40; //@line 186 "modules/math/BezierCurve.cpp"
 HEAP32[$41>>2] = $36; //@line 186 "modules/math/BezierCurve.cpp"
 $42 = (($40) + 4)|0; //@line 186 "modules/math/BezierCurve.cpp"
 $43 = $42; //@line 186 "modules/math/BezierCurve.cpp"
 HEAP32[$43>>2] = $39; //@line 186 "modules/math/BezierCurve.cpp"
 $44 = ($33|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $45 = $33; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($44) {
  STACKTOP = sp;return; //@line 187 "modules/math/BezierCurve.cpp"
 }
 $46 = ((($points)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $47 = HEAP32[$46>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $48 = ($47|0)==($33|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($48)) {
  $scevgep$i$i$i = ((($47)) + -8|0);
  $49 = $scevgep$i$i$i;
  $50 = (($49) - ($45))|0;
  $51 = $50 >>> 3;
  $52 = $51 ^ -1;
  $scevgep4$i$i$i = (($47) + ($52<<3)|0);
  HEAP32[$46>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($33); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 STACKTOP = sp;return; //@line 187 "modules/math/BezierCurve.cpp"
}
function __ZNK4love4math11BezierCurve10getSegmentEdd($this,$t1,$t2) {
 $this = $this|0;
 $t1 = +$t1;
 $t2 = +$t2;
 var $$0 = 0, $$01$i$i = 0, $$02 = 0, $$1 = 0, $$13 = 0, $$cast$lcssa = 0, $$cast$lcssa97 = 0, $$cast67 = 0, $$cast679599 = 0, $$cast69 = 0, $$lcssa = 0, $$lcssa58 = 0, $$lcssa60 = 0, $$phi$trans$insert88 = 0, $$phi$trans$insert92 = 0, $$pre$phiZ2D = 0, $$pre83 = 0, $$pre84 = 0, $$pre85 = 0, $$pre87 = 0.0;
 var $$pre89 = 0.0, $$pre91 = 0.0, $$pre93 = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0.0, $116 = 0.0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0.0, $169 = 0, $17 = 0, $170 = 0.0, $171 = 0.0, $172 = 0, $173 = 0.0, $174 = 0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0;
 var $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0;
 var $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0;
 var $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0;
 var $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0.0, $96 = 0, $97 = 0.0, $98 = 0.0, $99 = 0, $eh$lpad$body$index11Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i$065 = 0, $i2$062 = 0, $left = 0, $lpad$phi$index = 0, $lpad$phi$index5 = 0, $or$cond = 0, $points = 0, $right = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i$i = 0, $scevgep$i$i$i17 = 0;
 var $scevgep$i$i$i25 = 0, $scevgep$i$i$i30 = 0, $scevgep$i$i$i4 = 0, $scevgep$i$i$i9 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i$i = 0, $scevgep4$i$i$i10 = 0, $scevgep4$i$i$i18 = 0, $scevgep4$i$i$i26 = 0, $scevgep4$i$i$i31 = 0, $scevgep4$i$i$i5 = 0, $step$068 = 0, $step1$063 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0; //@line 2335 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $vararg_buffer = sp; //@line 2335 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $points = sp + 36|0;
 $left = sp + 24|0;
 $right = sp + 12|0;
 $0 = $t1 < 0.0; //@line 191 "modules/math/BezierCurve.cpp"
 $1 = $t2 > 1.0; //@line 191 "modules/math/BezierCurve.cpp"
 $or$cond = $0 | $1; //@line 191 "modules/math/BezierCurve.cpp"
 if ($or$cond) {
  $2 = (___cxa_allocate_exception(16)|0); //@line 192 "modules/math/BezierCurve.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($2|0),(444060|0),($vararg_buffer|0)); //@line 192 "modules/math/BezierCurve.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if (!($4)) {
   ___cxa_throw(($2|0),(96|0),(344|0)); //@line 192 "modules/math/BezierCurve.cpp"
   // unreachable; //@line 192 "modules/math/BezierCurve.cpp"
  }
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  ___cxa_free_exception(($2|0)); //@line 192 "modules/math/BezierCurve.cpp"
  $$1 = $5;$$13 = $6;
  ___resumeException($$1|0);
  // unreachable;
 }
 $7 = !($t2 <= $t1); //@line 193 "modules/math/BezierCurve.cpp"
 if (!($7)) {
  $8 = (___cxa_allocate_exception(16)|0); //@line 194 "modules/math/BezierCurve.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($8|0),(444112|0),($vararg_buffer1|0)); //@line 194 "modules/math/BezierCurve.cpp"
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if (!($10)) {
   ___cxa_throw(($8|0),(96|0),(344|0)); //@line 194 "modules/math/BezierCurve.cpp"
   // unreachable; //@line 194 "modules/math/BezierCurve.cpp"
  }
  $11 = ___cxa_find_matching_catch_2()|0;
  $12 = tempRet0;
  ___cxa_free_exception(($8|0)); //@line 194 "modules/math/BezierCurve.cpp"
  $$1 = $11;$$13 = $12;
  ___resumeException($$1|0);
  // unreachable;
 }
 $13 = ((($this)) + 8|0); //@line 198 "modules/math/BezierCurve.cpp"
 __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEEC2ERKS5_($points,$13); //@line 198 "modules/math/BezierCurve.cpp"
 HEAP32[$left>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = ((($left)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$14>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = ((($left)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$15>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$right>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = ((($right)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$16>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ((($right)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$17>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $18 = ((($points)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $19 = HEAP32[$18>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = HEAP32[$points>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = (($19) - ($20))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $22 = $21 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $23 = ((($left)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $24 = ($22|0)==(0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 L15: do {
  if ($24) {
   $25 = ((($right)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $$cast679599 = $20; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$cast$lcssa97 = $$cast679599;$320 = $right;$321 = $25;$322 = $points;
   label = 17;
  } else {
   __THREW__ = 0;
   $26 = (invoke_ii(3348,($21|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $27 = __THREW__; __THREW__ = 0;
   $28 = $27&1;
   if ($28) {
    label = 21;
   } else {
    $29 = $26; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $30 = (($26) + ($22<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $31 = $30; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    HEAP32[$left>>2] = $29; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$14>>2] = $29; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$23>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $32 = ((($right)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    __THREW__ = 0;
    $33 = (invoke_ii(3348,($21|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $34 = __THREW__; __THREW__ = 0;
    $35 = $34&1;
    if ($35) {
     label = 21;
    } else {
     $36 = $33; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $37 = (($33) + ($22<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $38 = $37; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     HEAP32[$right>>2] = $36; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$16>>2] = $36; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$32>>2] = $38; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $39 = ($22>>>0)>(1); //@line 204 "modules/math/BezierCurve.cpp"
     $$cast67 = $20; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $40 = ($26|0)==($30|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     L20: do {
      if ($39) {
       $41 = $t2; //@line 208 "modules/math/BezierCurve.cpp"
       $$cast69 = $$cast67;$323 = $40;$65 = $20;$73 = $26;$step$068 = 1;
       while(1) {
        if ($323) {
         __THREW__ = 0;
         invoke_vii(3383,($left|0),($$cast69|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $79 = __THREW__; __THREW__ = 0;
         $80 = $79&1;
         if ($80) {
          break;
         }
        } else {
         $64 = $65; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $66 = $64; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $67 = $66; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $68 = HEAP32[$67>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $69 = (($66) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $70 = $69; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $71 = HEAP32[$70>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $72 = $73; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $74 = $72; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         HEAP32[$74>>2] = $68; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $75 = (($72) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $76 = $75; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         HEAP32[$76>>2] = $71; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $77 = HEAP32[$14>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $78 = ((($77)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$14>>2] = $78; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
        $81 = HEAP32[$18>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $82 = HEAP32[$points>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $83 = (($81) - ($82))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $84 = $83 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $85 = ($84|0)==($step$068|0); //@line 207 "modules/math/BezierCurve.cpp"
        $86 = $82; //@line 207 "modules/math/BezierCurve.cpp"
        if (!($85)) {
         $$pre87 = +HEAPF32[$86>>2]; //@line 305 "./common/Vector.h"
         $$phi$trans$insert88 = ((($86)) + 4|0); //@line 310 "./common/Vector.h"
         $$pre89 = +HEAPF32[$$phi$trans$insert88>>2]; //@line 310 "./common/Vector.h"
         $87 = (($84) - ($step$068))|0; //@line 207 "modules/math/BezierCurve.cpp"
         $103 = $$pre89;$98 = $$pre87;$i$065 = 0;
         while(1) {
          $93 = (($i$065) + 1)|0; //@line 208 "modules/math/BezierCurve.cpp"
          $94 = (($86) + ($93<<3)|0); //@line 237 "./common/Vector.h"
          $95 = +HEAPF32[$94>>2]; //@line 237 "./common/Vector.h"
          $96 = (($86) + ($i$065<<3)|0); //@line 305 "./common/Vector.h"
          $97 = $95 - $98; //@line 237 "./common/Vector.h"
          $99 = (((($86) + ($93<<3)|0)) + 4|0); //@line 237 "./common/Vector.h"
          $100 = +HEAPF32[$99>>2]; //@line 237 "./common/Vector.h"
          $101 = (((($86) + ($i$065<<3)|0)) + 4|0); //@line 310 "./common/Vector.h"
          $102 = $100 - $103; //@line 237 "./common/Vector.h"
          $104 = $41 * $97; //@line 242 "./common/Vector.h"
          $105 = $41 * $102; //@line 242 "./common/Vector.h"
          $106 = $98 + $104; //@line 257 "./common/Vector.h"
          HEAPF32[$96>>2] = $106; //@line 257 "./common/Vector.h"
          $107 = $105 + $103; //@line 258 "./common/Vector.h"
          HEAPF32[$101>>2] = $107; //@line 258 "./common/Vector.h"
          $108 = ($93>>>0)<($87>>>0); //@line 207 "modules/math/BezierCurve.cpp"
          if ($108) {
           $103 = $100;$98 = $95;$i$065 = $93;
          } else {
           break;
          }
         }
        }
        $88 = (($step$068) + 1)|0; //@line 204 "modules/math/BezierCurve.cpp"
        $89 = ($88>>>0)<($84>>>0); //@line 204 "modules/math/BezierCurve.cpp"
        $90 = HEAP32[$14>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $91 = HEAP32[$23>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $92 = ($90|0)==($91|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($89) {
         $$cast69 = $86;$323 = $92;$65 = $82;$73 = $90;$step$068 = $88;
        } else {
         $$cast$lcssa = $86;$$lcssa = $82;$$lcssa58 = $90;$$lcssa60 = $92;
         break L20;
        }
       }
       $60 = ___cxa_find_matching_catch_2()|0;
       $61 = tempRet0;
       $lpad$phi$index = $60;$lpad$phi$index5 = $61;
       label = 22;
       break L15;
      } else {
       $$cast$lcssa = $$cast67;$$lcssa = $20;$$lcssa58 = $26;$$lcssa60 = $40;
      }
     } while(0);
     if ($$lcssa60) {
      $$cast$lcssa97 = $$cast$lcssa;$320 = $right;$321 = $32;$322 = $points;
      label = 17;
     } else {
      $42 = $$lcssa; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $43 = $42; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $44 = $43; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $45 = HEAP32[$44>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $46 = (($43) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $47 = $46; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $48 = HEAP32[$47>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $49 = $$lcssa58; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $50 = $49; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$50>>2] = $45; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $51 = (($49) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $52 = $51; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$52>>2] = $48; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $53 = HEAP32[$14>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $54 = ((($53)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$14>>2] = $54; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $55 = $54; //@line 1598 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$pre$phiZ2D = $14;$111 = $55;$119 = $32;$222 = $right;$282 = $points;
      label = 30;
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 17) {
  __THREW__ = 0;
  invoke_vii(3383,($left|0),($$cast$lcssa97|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $56 = __THREW__; __THREW__ = 0;
  $57 = $56&1;
  if ($57) {
   label = 21;
  } else {
   $$pre83 = HEAP32[$14>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$pre$phiZ2D = $14;$111 = $$pre83;$119 = $321;$222 = $320;$282 = $322;
   label = 30;
  }
 }
 L39: do {
  if ((label|0) == 30) {
   $109 = HEAP32[$left>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $110 = (($111) - ($109))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $112 = $110 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $113 = ($112>>>0)>(1); //@line 214 "modules/math/BezierCurve.cpp"
   $114 = $109; //@line 214 "modules/math/BezierCurve.cpp"
   L41: do {
    if ($113) {
     $115 = $t1 / $t2; //@line 213 "modules/math/BezierCurve.cpp"
     $116 = $115; //@line 218 "modules/math/BezierCurve.cpp"
     $137 = $112;$139 = $114;$step1$063 = 1;
     while(1) {
      $136 = (($137) - ($step1$063))|0; //@line 216 "modules/math/BezierCurve.cpp"
      $138 = (($139) + ($136<<3)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $140 = HEAP32[$16>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $141 = HEAP32[$119>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $142 = ($140|0)==($141|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($142) {
       __THREW__ = 0;
       invoke_vii(3383,($right|0),($138|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $155 = __THREW__; __THREW__ = 0;
       $156 = $155&1;
       if ($156) {
        break;
       }
      } else {
       $143 = $138; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $144 = $143; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $145 = HEAP32[$144>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $146 = (($143) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $147 = $146; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $148 = HEAP32[$147>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $149 = $140; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $150 = $149; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       HEAP32[$150>>2] = $145; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $151 = (($149) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $152 = $151; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       HEAP32[$152>>2] = $148; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $153 = HEAP32[$16>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $154 = ((($153)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$16>>2] = $154; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      $157 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $158 = HEAP32[$left>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $159 = (($157) - ($158))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $160 = $159 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $161 = ($160|0)==($step1$063|0); //@line 217 "modules/math/BezierCurve.cpp"
      $162 = $158; //@line 217 "modules/math/BezierCurve.cpp"
      if (!($161)) {
       $$pre91 = +HEAPF32[$162>>2]; //@line 305 "./common/Vector.h"
       $$phi$trans$insert92 = ((($162)) + 4|0); //@line 310 "./common/Vector.h"
       $$pre93 = +HEAPF32[$$phi$trans$insert92>>2]; //@line 310 "./common/Vector.h"
       $163 = (($160) - ($step1$063))|0; //@line 217 "modules/math/BezierCurve.cpp"
       $171 = $$pre91;$176 = $$pre93;$i2$062 = 0;
       while(1) {
        $166 = (($i2$062) + 1)|0; //@line 218 "modules/math/BezierCurve.cpp"
        $167 = (($162) + ($166<<3)|0); //@line 237 "./common/Vector.h"
        $168 = +HEAPF32[$167>>2]; //@line 237 "./common/Vector.h"
        $169 = (($162) + ($i2$062<<3)|0); //@line 305 "./common/Vector.h"
        $170 = $168 - $171; //@line 237 "./common/Vector.h"
        $172 = (((($162) + ($166<<3)|0)) + 4|0); //@line 237 "./common/Vector.h"
        $173 = +HEAPF32[$172>>2]; //@line 237 "./common/Vector.h"
        $174 = (((($162) + ($i2$062<<3)|0)) + 4|0); //@line 310 "./common/Vector.h"
        $175 = $173 - $176; //@line 237 "./common/Vector.h"
        $177 = $116 * $170; //@line 242 "./common/Vector.h"
        $178 = $116 * $175; //@line 242 "./common/Vector.h"
        $179 = $171 + $177; //@line 257 "./common/Vector.h"
        HEAPF32[$169>>2] = $179; //@line 257 "./common/Vector.h"
        $180 = $178 + $176; //@line 258 "./common/Vector.h"
        HEAPF32[$174>>2] = $180; //@line 258 "./common/Vector.h"
        $181 = ($166>>>0)<($163>>>0); //@line 217 "modules/math/BezierCurve.cpp"
        if ($181) {
         $171 = $168;$176 = $173;$i2$062 = $166;
        } else {
         break;
        }
       }
      }
      $164 = (($step1$063) + 1)|0; //@line 214 "modules/math/BezierCurve.cpp"
      $165 = ($160>>>0)>($164>>>0); //@line 214 "modules/math/BezierCurve.cpp"
      if ($165) {
       $137 = $160;$139 = $162;$step1$063 = $164;
      } else {
       $122 = $162;
       break L41;
      }
     }
     $58 = ___cxa_find_matching_catch_2()|0;
     $59 = tempRet0;
     $lpad$phi$index = $58;$lpad$phi$index5 = $59;
     label = 22;
     break L39;
    } else {
     $122 = $114;
    }
   } while(0);
   $117 = HEAP32[$16>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $118 = HEAP32[$119>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $120 = ($117|0)==($118|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($120) {
    __THREW__ = 0;
    invoke_vii(3383,($right|0),($122|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $134 = __THREW__; __THREW__ = 0;
    $135 = $134&1;
    if ($135) {
     label = 21;
     break;
    }
    $$pre84 = HEAP32[$16>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $184 = $$pre84;
   } else {
    $121 = $122; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $123 = $121; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $124 = HEAP32[$123>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $125 = (($121) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $126 = $125; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $127 = HEAP32[$126>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $128 = $117; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $129 = $128; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$129>>2] = $124; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $130 = (($128) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $131 = $130; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$131>>2] = $127; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $132 = HEAP32[$16>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $133 = ((($132)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$16>>2] = $133; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $184 = $133;
   }
   $182 = HEAP32[$right>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $183 = ($182|0)==($184|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if (!($183)) {
    $185 = ((($184)) + -8|0); //@line 1190 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    $186 = ($182>>>0)<($185>>>0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($186) {
     $188 = $182;$195 = $185;
     while(1) {
      $187 = $188; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $189 = $187; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $190 = HEAP32[$189>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $191 = (($187) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $192 = $191; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $193 = HEAP32[$192>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $194 = $195; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $196 = $194; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $197 = HEAP32[$196>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $198 = (($194) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $199 = $198; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $200 = HEAP32[$199>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $201 = $188; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $202 = $201; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$202>>2] = $197; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $203 = (($201) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $204 = $203; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$204>>2] = $200; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $205 = $195; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $206 = $205; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$206>>2] = $190; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $207 = (($205) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $208 = $207; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$208>>2] = $193; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $209 = ((($188)) + 8|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $210 = ((($195)) + -8|0); //@line 1190 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $211 = ($209>>>0)<($210>>>0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($211) {
       $188 = $209;$195 = $210;
      } else {
       break;
      }
     }
    }
   }
   __THREW__ = 0;
   $212 = (invoke_ii(3348,20)|0); //@line 224 "modules/math/BezierCurve.cpp"
   $213 = __THREW__; __THREW__ = 0;
   $214 = $213&1;
   if ($214) {
    label = 21;
   } else {
    __THREW__ = 0;
    invoke_vi(3431,($212|0)); //@line 86 "modules/math/BezierCurve.cpp"
    $215 = __THREW__; __THREW__ = 0;
    $216 = $215&1;
    L67: do {
     if ($216) {
      $291 = ___cxa_find_matching_catch_2()|0;
      $292 = tempRet0;
      $eh$lpad$body$index11Z2D = $292;$eh$lpad$body$indexZ2D = $291;
     } else {
      HEAP32[$212>>2] = (32436); //@line 88 "modules/math/BezierCurve.cpp"
      $217 = ((($212)) + 8|0); //@line 87 "modules/math/BezierCurve.cpp"
      HEAP32[$217>>2] = 0; //@line 460 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $218 = ((($212)) + 12|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$218>>2] = 0; //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $219 = ((($212)) + 16|0); //@line 462 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$219>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $220 = HEAP32[$16>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $221 = HEAP32[$222>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $223 = (($220) - ($221))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $224 = $223 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $225 = ($224|0)==(0); //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $226 = $221; //@line 1212 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      do {
       if ($225) {
        $264 = $226;
       } else {
        $227 = ($224>>>0)>(536870911); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($227) {
         __THREW__ = 0;
         invoke_vi(3528,($217|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $228 = __THREW__; __THREW__ = 0;
         $229 = $228&1;
         if (!($229)) {
          label = 51;
         }
        } else {
         label = 51;
        }
        if ((label|0) == 51) {
         __THREW__ = 0;
         $230 = (invoke_ii(3348,($223|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
         $231 = __THREW__; __THREW__ = 0;
         $232 = $231&1;
         if (!($232)) {
          HEAP32[$218>>2] = $230; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$217>>2] = $230; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $233 = (($230) + ($224<<3)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$219>>2] = $233; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $234 = HEAP32[$right>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $235 = HEAP32[$16>>2]|0; //@line 1215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $236 = ($234|0)==($235|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          if ($236) {
           $264 = $234;
           break;
          }
          $$01$i$i = $234;$244 = $230;
          while(1) {
           $237 = $$01$i$i; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $238 = $237; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $239 = HEAP32[$238>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $240 = (($237) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $241 = $240; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $242 = HEAP32[$241>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $243 = $244; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $245 = $243; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           HEAP32[$245>>2] = $239; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $246 = (($243) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $247 = $246; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           HEAP32[$247>>2] = $242; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
           $248 = HEAP32[$218>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $249 = ((($248)) + 8|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           HEAP32[$218>>2] = $249; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $250 = ((($$01$i$i)) + 8|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $251 = ($250|0)==($235|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           if ($251) {
            break;
           } else {
            $$01$i$i = $250;$244 = $249;
           }
          }
          $$pre85 = HEAP32[$right>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $264 = $$pre85;
          break;
         }
        }
        $252 = ___cxa_find_matching_catch_2()|0;
        $253 = tempRet0;
        $254 = HEAP32[$217>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $255 = ($254|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $256 = $254; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if (!($255)) {
         $257 = HEAP32[$218>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $258 = ($257|0)==($254|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if (!($258)) {
          $scevgep$i$i$i$i = ((($257)) + -8|0);
          $259 = $scevgep$i$i$i$i;
          $260 = (($259) - ($256))|0;
          $261 = $260 >>> 3;
          $262 = $261 ^ -1;
          $scevgep4$i$i$i$i = (($257) + ($262<<3)|0);
          HEAP32[$218>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         }
         __ZdlPv($254); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        }
        $eh$lpad$body$index11Z2D = $253;$eh$lpad$body$indexZ2D = $252;
        break L67;
       }
      } while(0);
      $263 = ($264|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $265 = $264; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($263)) {
       $266 = HEAP32[$16>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $267 = ($266|0)==($264|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($267)) {
        $scevgep$i$i$i30 = ((($266)) + -8|0);
        $268 = $scevgep$i$i$i30;
        $269 = (($268) - ($265))|0;
        $270 = $269 >>> 3;
        $271 = $270 ^ -1;
        $scevgep4$i$i$i31 = (($266) + ($271<<3)|0);
        HEAP32[$16>>2] = $scevgep4$i$i$i31; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       __ZdlPv($264); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      }
      $272 = HEAP32[$left>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $273 = ($272|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $274 = $272; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($273)) {
       $275 = HEAP32[$14>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $276 = ($275|0)==($272|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if (!($276)) {
        $scevgep$i$i$i25 = ((($275)) + -8|0);
        $277 = $scevgep$i$i$i25;
        $278 = (($277) - ($274))|0;
        $279 = $278 >>> 3;
        $280 = $279 ^ -1;
        $scevgep4$i$i$i26 = (($275) + ($280<<3)|0);
        HEAP32[$14>>2] = $scevgep4$i$i$i26; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       }
       __ZdlPv($272); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      }
      $281 = HEAP32[$282>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $283 = ($281|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $284 = $281; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($283) {
       STACKTOP = sp;return ($212|0); //@line 225 "modules/math/BezierCurve.cpp"
      }
      $285 = HEAP32[$18>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $286 = ($285|0)==($281|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($286)) {
       $scevgep$i$i$i17 = ((($285)) + -8|0);
       $287 = $scevgep$i$i$i17;
       $288 = (($287) - ($284))|0;
       $289 = $288 >>> 3;
       $290 = $289 ^ -1;
       $scevgep4$i$i$i18 = (($285) + ($290<<3)|0);
       HEAP32[$18>>2] = $scevgep4$i$i$i18; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($281); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      STACKTOP = sp;return ($212|0); //@line 225 "modules/math/BezierCurve.cpp"
     }
    } while(0);
    __ZdlPv($212); //@line 224 "modules/math/BezierCurve.cpp"
    $$0 = $eh$lpad$body$indexZ2D;$$02 = $eh$lpad$body$index11Z2D;
   }
  }
 } while(0);
 if ((label|0) == 21) {
  $62 = ___cxa_find_matching_catch_2()|0;
  $63 = tempRet0;
  $lpad$phi$index = $62;$lpad$phi$index5 = $63;
  label = 22;
 }
 if ((label|0) == 22) {
  $$0 = $lpad$phi$index;$$02 = $lpad$phi$index5;
 }
 $293 = HEAP32[$right>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $294 = ($293|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $295 = $293; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($294)) {
  $296 = HEAP32[$16>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $297 = ($296|0)==($293|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($297)) {
   $scevgep$i$i$i9 = ((($296)) + -8|0);
   $298 = $scevgep$i$i$i9;
   $299 = (($298) - ($295))|0;
   $300 = $299 >>> 3;
   $301 = $300 ^ -1;
   $scevgep4$i$i$i10 = (($296) + ($301<<3)|0);
   HEAP32[$16>>2] = $scevgep4$i$i$i10; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($293); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $302 = HEAP32[$left>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $303 = ($302|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $304 = $302; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($303)) {
  $305 = HEAP32[$14>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $306 = ($305|0)==($302|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($306)) {
   $scevgep$i$i$i4 = ((($305)) + -8|0);
   $307 = $scevgep$i$i$i4;
   $308 = (($307) - ($304))|0;
   $309 = $308 >>> 3;
   $310 = $309 ^ -1;
   $scevgep4$i$i$i5 = (($305) + ($310<<3)|0);
   HEAP32[$14>>2] = $scevgep4$i$i$i5; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($302); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $311 = HEAP32[$points>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $312 = ($311|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $313 = $311; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($312) {
  $$1 = $$0;$$13 = $$02;
  ___resumeException($$1|0);
  // unreachable;
 }
 $314 = HEAP32[$18>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $315 = ($314|0)==($311|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($315)) {
  $scevgep$i$i$i = ((($314)) + -8|0);
  $316 = $scevgep$i$i$i;
  $317 = (($316) - ($313))|0;
  $318 = $317 >>> 3;
  $319 = $318 ^ -1;
  $scevgep4$i$i$i = (($314) + ($319<<3)|0);
  HEAP32[$18>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($311); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $$1 = $$0;$$13 = $$02;
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNK4love4math11BezierCurve6renderEi($agg$result,$this,$accuracy) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $accuracy = $accuracy|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 8|0); //@line 229 "modules/math/BezierCurve.cpp"
 $1 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($2) - ($3))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = $4 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)<(2); //@line 229 "modules/math/BezierCurve.cpp"
 if ($6) {
  $7 = (___cxa_allocate_exception(16)|0); //@line 230 "modules/math/BezierCurve.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($7|0),(444011|0),($vararg_buffer|0)); //@line 230 "modules/math/BezierCurve.cpp"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if (!($9)) {
   ___cxa_throw(($7|0),(96|0),(344|0)); //@line 230 "modules/math/BezierCurve.cpp"
   // unreachable; //@line 230 "modules/math/BezierCurve.cpp"
  }
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  ___cxa_free_exception(($7|0)); //@line 230 "modules/math/BezierCurve.cpp"
  $$0 = $11;$$01 = $10;
  ___resumeException($$01|0);
  // unreachable;
 }
 __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEEC2ERKS5_($agg$result,$0); //@line 231 "modules/math/BezierCurve.cpp"
 __THREW__ = 0;
 invoke_vii(4204,($agg$result|0),($accuracy|0)); //@line 232 "modules/math/BezierCurve.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if (!($13)) {
  STACKTOP = sp;return; //@line 234 "modules/math/BezierCurve.cpp"
 }
 $14 = ___cxa_find_matching_catch_2()|0;
 $15 = tempRet0;
 $16 = HEAP32[$agg$result>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ($16|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $18 = $16; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($17) {
  $$0 = $15;$$01 = $14;
  ___resumeException($$01|0);
  // unreachable;
 }
 $19 = ((($agg$result)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = HEAP32[$19>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = ($20|0)==($16|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($21)) {
  $scevgep$i$i$i = ((($20)) + -8|0);
  $22 = $scevgep$i$i$i;
  $23 = (($22) - ($18))|0;
  $24 = $23 >>> 3;
  $25 = $24 ^ -1;
  $scevgep4$i$i$i = (($20) + ($25<<3)|0);
  HEAP32[$19>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($16); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $$0 = $15;$$01 = $14;
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN12_GLOBAL__N_19subdivideERNSt3__16vectorIN4love6VectorENS0_9allocatorIS3_EEEEi($points,$k) {
 $points = $points|0;
 $k = $k|0;
 var $$cast = 0, $$cast$lcssa = 0, $$cast$lcssa66 = 0, $$cast32 = 0, $$cast46 = 0, $$cast466468 = 0, $$cast48 = 0, $$in = 0, $$in63 = 0, $$lcssa = 0, $$lcssa37 = 0, $$lcssa39 = 0, $$pre = 0, $$pre60 = 0, $$pre61 = 0, $$pre62 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0.0, $123 = 0, $124 = 0.0, $125 = 0.0, $126 = 0, $127 = 0.0, $128 = 0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0.0, $132 = 0.0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0;
 var $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0;
 var $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__new_last2$i$i$i = 0, $i$044 = 0, $i1$042 = 0, $i2$041 = 0, $left = 0;
 var $lpad$phi$index = 0, $lpad$phi$index2 = 0, $right = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i10 = 0, $scevgep$i$i$i15 = 0, $scevgep$i$i$i22 = 0, $scevgep$i$i$i4 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i11 = 0, $scevgep4$i$i$i16 = 0, $scevgep4$i$i$i23 = 0, $scevgep4$i$i$i5 = 0, $step$047 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $left = sp + 12|0;
 $right = sp;
 $0 = ($k|0)<(1); //@line 38 "modules/math/BezierCurve.cpp"
 if ($0) {
  STACKTOP = sp;return; //@line 77 "modules/math/BezierCurve.cpp"
 }
 HEAP32[$left>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($left)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($left)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$2>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$right>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = ((($right)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$3>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = ((($right)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$4>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = ((($points)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$5>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = HEAP32[$points>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = (($6) - ($7))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = $8 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = ((($left)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $11 = ($9|0)==(0); //@line 1539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 L4: do {
  if ($11) {
   $12 = ((($right)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $$cast466468 = $7; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$cast$lcssa66 = $$cast466468;$272 = $12;$273 = $right;$274 = $points;
   label = 9;
  } else {
   __THREW__ = 0;
   $13 = (invoke_ii(3348,($8|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $14 = __THREW__; __THREW__ = 0;
   $15 = $14&1;
   if ($15) {
    label = 11;
   } else {
    $16 = $13; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $17 = (($13) + ($9<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $18 = $17; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    HEAP32[$left>>2] = $16; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$1>>2] = $16; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$10>>2] = $18; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $19 = ((($right)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    __THREW__ = 0;
    $20 = (invoke_ii(3348,($8|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $21 = __THREW__; __THREW__ = 0;
    $22 = $21&1;
    if ($22) {
     label = 11;
    } else {
     $23 = $20; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $24 = (($20) + ($9<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $25 = $24; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     HEAP32[$right>>2] = $23; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$3>>2] = $23; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$19>>2] = $25; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $26 = ($9>>>0)>(1); //@line 57 "modules/math/BezierCurve.cpp"
     $$cast46 = $7; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $27 = ($13|0)==($17|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     L9: do {
      if ($26) {
       $$cast48 = $$cast46;$275 = $27;$66 = $7;$74 = $13;$step$047 = 1;
       while(1) {
        if ($275) {
         __THREW__ = 0;
         invoke_vii(3383,($left|0),($$cast48|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $80 = __THREW__; __THREW__ = 0;
         $81 = $80&1;
         if ($81) {
          break;
         }
        } else {
         $65 = $66; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $67 = $65; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $68 = $67; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $69 = HEAP32[$68>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $70 = (($67) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $71 = $70; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $72 = HEAP32[$71>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $73 = $74; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $75 = $73; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         HEAP32[$75>>2] = $69; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $76 = (($73) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $77 = $76; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         HEAP32[$77>>2] = $72; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $78 = HEAP32[$1>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $79 = ((($78)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$1>>2] = $79; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
        $82 = HEAP32[$5>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $83 = HEAP32[$points>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $84 = (($82) - ($83))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $85 = $84 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $86 = (($85) - ($step$047))|0; //@line 60 "modules/math/BezierCurve.cpp"
        $$cast32 = $83; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $87 = (($$cast32) + ($86<<3)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $88 = HEAP32[$3>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $89 = HEAP32[$19>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $90 = ($88|0)==($89|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($90) {
         __THREW__ = 0;
         invoke_vii(3383,($right|0),($87|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $103 = __THREW__; __THREW__ = 0;
         $104 = $103&1;
         if ($104) {
          break;
         }
        } else {
         $91 = $87; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $92 = $91; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $93 = HEAP32[$92>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $94 = (($91) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $95 = $94; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $96 = HEAP32[$95>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $97 = $88; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $98 = $97; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         HEAP32[$98>>2] = $93; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $99 = (($97) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $100 = $99; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         HEAP32[$100>>2] = $96; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $101 = HEAP32[$3>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $102 = ((($101)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$3>>2] = $102; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        }
        $105 = HEAP32[$5>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $106 = HEAP32[$points>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $107 = (($105) - ($106))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $108 = $107 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $109 = ($108|0)==($step$047|0); //@line 61 "modules/math/BezierCurve.cpp"
        if ($109) {
         $112 = $105;$113 = $106;
        } else {
         $$in63 = $106;$i$044 = 0;
         while(1) {
          $119 = $$in63; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $120 = (($i$044) + 1)|0; //@line 62 "modules/math/BezierCurve.cpp"
          $121 = (($119) + ($i$044<<3)|0); //@line 232 "./common/Vector.h"
          $122 = +HEAPF32[$121>>2]; //@line 232 "./common/Vector.h"
          $123 = (($119) + ($120<<3)|0); //@line 232 "./common/Vector.h"
          $124 = +HEAPF32[$123>>2]; //@line 232 "./common/Vector.h"
          $125 = $122 + $124; //@line 232 "./common/Vector.h"
          $126 = (((($119) + ($i$044<<3)|0)) + 4|0); //@line 232 "./common/Vector.h"
          $127 = +HEAPF32[$126>>2]; //@line 232 "./common/Vector.h"
          $128 = (((($119) + ($120<<3)|0)) + 4|0); //@line 232 "./common/Vector.h"
          $129 = +HEAPF32[$128>>2]; //@line 232 "./common/Vector.h"
          $130 = $127 + $129; //@line 232 "./common/Vector.h"
          $131 = $125 * 0.5; //@line 242 "./common/Vector.h"
          $132 = $130 * 0.5; //@line 242 "./common/Vector.h"
          HEAPF32[$121>>2] = $131; //@line 62 "modules/math/BezierCurve.cpp"
          HEAPF32[$126>>2] = $132; //@line 62 "modules/math/BezierCurve.cpp"
          $133 = HEAP32[$5>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $134 = HEAP32[$points>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $135 = (($133) - ($134))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $136 = $135 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $137 = (($136) - ($step$047))|0; //@line 61 "modules/math/BezierCurve.cpp"
          $138 = ($120>>>0)<($137>>>0); //@line 61 "modules/math/BezierCurve.cpp"
          if ($138) {
           $$in63 = $134;$i$044 = $120;
          } else {
           $112 = $133;$113 = $134;
           break;
          }
         }
        }
        $110 = (($step$047) + 1)|0; //@line 57 "modules/math/BezierCurve.cpp"
        $111 = (($112) - ($113))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $114 = $111 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $115 = ($110>>>0)<($114>>>0); //@line 57 "modules/math/BezierCurve.cpp"
        $$cast = $113; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $116 = HEAP32[$1>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $117 = HEAP32[$10>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $118 = ($116|0)==($117|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        if ($115) {
         $$cast48 = $$cast;$275 = $118;$66 = $113;$74 = $116;$step$047 = $110;
        } else {
         $$cast$lcssa = $$cast;$$lcssa = $113;$$lcssa37 = $116;$$lcssa39 = $118;
         break L9;
        }
       }
       $43 = ___cxa_find_matching_catch_2()|0;
       $44 = tempRet0;
       $lpad$phi$index = $43;$lpad$phi$index2 = $44;
       break L4;
      } else {
       $$cast$lcssa = $$cast46;$$lcssa = $7;$$lcssa37 = $13;$$lcssa39 = $27;
      }
     } while(0);
     if ($$lcssa39) {
      $$cast$lcssa66 = $$cast$lcssa;$272 = $19;$273 = $right;$274 = $points;
      label = 9;
     } else {
      $28 = $$lcssa; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $29 = $28; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $30 = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $31 = HEAP32[$30>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $32 = (($29) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $33 = $32; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $34 = HEAP32[$33>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $35 = $$lcssa37; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $36 = $35; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$36>>2] = $31; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $37 = (($35) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $38 = $37; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$38>>2] = $34; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $39 = HEAP32[$1>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $40 = ((($39)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$1>>2] = $40; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $140 = $points;$143 = $19;$170 = $right;
      label = 30;
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 9) {
  __THREW__ = 0;
  invoke_vii(3383,($left|0),($$cast$lcssa66|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $41 = __THREW__; __THREW__ = 0;
  $42 = $41&1;
  if ($42) {
   label = 11;
  } else {
   $140 = $274;$143 = $272;$170 = $273;
   label = 30;
  }
 }
 do {
  if ((label|0) == 30) {
   $139 = HEAP32[$140>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $141 = HEAP32[$3>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $142 = HEAP32[$143>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $144 = ($141|0)==($142|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($144) {
    __THREW__ = 0;
    invoke_vii(3383,($right|0),($139|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $157 = __THREW__; __THREW__ = 0;
    $158 = $157&1;
    if ($158) {
     label = 11;
     break;
    }
   } else {
    $145 = $139; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $146 = $145; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $147 = HEAP32[$146>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $148 = (($145) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $149 = $148; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $150 = HEAP32[$149>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $151 = $141; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $152 = $151; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$152>>2] = $147; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $153 = (($151) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $154 = $153; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$154>>2] = $150; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $155 = HEAP32[$3>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $156 = ((($155)) + 8|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$3>>2] = $156; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $159 = (($k) + -1)|0; //@line 68 "modules/math/BezierCurve.cpp"
   __THREW__ = 0;
   invoke_vii(4204,($left|0),($159|0)); //@line 68 "modules/math/BezierCurve.cpp"
   $160 = __THREW__; __THREW__ = 0;
   $161 = $160&1;
   if ($161) {
    label = 11;
   } else {
    __THREW__ = 0;
    invoke_vii(4204,($right|0),($159|0)); //@line 69 "modules/math/BezierCurve.cpp"
    $162 = __THREW__; __THREW__ = 0;
    $163 = $162&1;
    if ($163) {
     label = 11;
    } else {
     $164 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $165 = HEAP32[$left>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $166 = (($164) - ($165))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $167 = $166 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $168 = HEAP32[$3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $169 = HEAP32[$170>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $171 = (($168) - ($169))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $172 = $171 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $173 = (($167) + -1)|0; //@line 72 "modules/math/BezierCurve.cpp"
     $174 = (($173) + ($172))|0; //@line 72 "modules/math/BezierCurve.cpp"
     $175 = HEAP32[$5>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $176 = HEAP32[$points>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $177 = (($175) - ($176))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $178 = $177 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $179 = ($174>>>0)>($178>>>0); //@line 1998 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $180 = $176; //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $181 = $175; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $182 = $165; //@line 1998 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($179) {
      $183 = (($174) - ($178))|0; //@line 1999 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      __THREW__ = 0;
      invoke_vii(4205,($points|0),($183|0)); //@line 1999 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $184 = __THREW__; __THREW__ = 0;
      $185 = $184&1;
      if ($185) {
       label = 11;
       break;
      }
      $$pre = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$pre60 = HEAP32[$left>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $186 = $$pre60; //@line 1999 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $195 = $$pre;$196 = $$pre60;$276 = $186;
     } else {
      $187 = ($174>>>0)<($178>>>0); //@line 2000 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($187) {
       $188 = (($180) + ($174<<3)|0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $189 = ($181|0)==($188|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($189) {
        $195 = $164;$196 = $165;$276 = $182;
       } else {
        $__new_last2$i$i$i = $188; //@line 818 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $scevgep$i$i$i22 = ((($181)) + -8|0);
        $190 = $scevgep$i$i$i22;
        $191 = (($190) - ($__new_last2$i$i$i))|0;
        $192 = $191 >>> 3;
        $193 = $192 ^ -1;
        $scevgep4$i$i$i23 = (($181) + ($193<<3)|0);
        HEAP32[$5>>2] = $scevgep4$i$i$i23; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $195 = $164;$196 = $165;$276 = $182;
       }
      } else {
       $195 = $164;$196 = $165;$276 = $182;
      }
     }
     $194 = ($195|0)==($196|0); //@line 73 "modules/math/BezierCurve.cpp"
     if ($194) {
      $277 = $196;$278 = $196;
     } else {
      $206 = $276;$i1$042 = 0;
      while(1) {
       $203 = HEAP32[$140>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $204 = (($203) + ($i1$042<<3)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $205 = (($206) + ($i1$042<<3)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $207 = $205; //@line 74 "modules/math/BezierCurve.cpp"
       $208 = $207; //@line 74 "modules/math/BezierCurve.cpp"
       $209 = HEAP32[$208>>2]|0; //@line 74 "modules/math/BezierCurve.cpp"
       $210 = (($207) + 4)|0; //@line 74 "modules/math/BezierCurve.cpp"
       $211 = $210; //@line 74 "modules/math/BezierCurve.cpp"
       $212 = HEAP32[$211>>2]|0; //@line 74 "modules/math/BezierCurve.cpp"
       $213 = $204; //@line 74 "modules/math/BezierCurve.cpp"
       $214 = $213; //@line 74 "modules/math/BezierCurve.cpp"
       HEAP32[$214>>2] = $209; //@line 74 "modules/math/BezierCurve.cpp"
       $215 = (($213) + 4)|0; //@line 74 "modules/math/BezierCurve.cpp"
       $216 = $215; //@line 74 "modules/math/BezierCurve.cpp"
       HEAP32[$216>>2] = $212; //@line 74 "modules/math/BezierCurve.cpp"
       $217 = (($i1$042) + 1)|0; //@line 73 "modules/math/BezierCurve.cpp"
       $218 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $219 = HEAP32[$left>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $220 = (($218) - ($219))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $221 = $220 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $222 = ($217>>>0)<($221>>>0); //@line 73 "modules/math/BezierCurve.cpp"
       $223 = $219; //@line 73 "modules/math/BezierCurve.cpp"
       if ($222) {
        $206 = $223;$i1$042 = $217;
       } else {
        $277 = $219;$278 = $218;
        break;
       }
      }
     }
     $197 = HEAP32[$3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $198 = HEAP32[$170>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $199 = (($197) - ($198))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $200 = $199 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $201 = ($200>>>0)>(1); //@line 75 "modules/math/BezierCurve.cpp"
     $202 = $198; //@line 75 "modules/math/BezierCurve.cpp"
     L48: do {
      if ($201) {
       $244 = $278;$245 = $277;$252 = $200;$254 = $202;$i2$041 = 1;
       while(1) {
        $242 = (($i2$041) + -1)|0; //@line 76 "modules/math/BezierCurve.cpp"
        $243 = (($244) - ($245))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $246 = $243 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $247 = (($242) + ($246))|0; //@line 76 "modules/math/BezierCurve.cpp"
        $248 = HEAP32[$140>>2]|0; //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $249 = (($248) + ($247<<3)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $250 = $i2$041 ^ -1; //@line 76 "modules/math/BezierCurve.cpp"
        $251 = (($252) + ($250))|0; //@line 76 "modules/math/BezierCurve.cpp"
        $253 = (($254) + ($251<<3)|0); //@line 1505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $255 = $253; //@line 76 "modules/math/BezierCurve.cpp"
        $256 = $255; //@line 76 "modules/math/BezierCurve.cpp"
        $257 = HEAP32[$256>>2]|0; //@line 76 "modules/math/BezierCurve.cpp"
        $258 = (($255) + 4)|0; //@line 76 "modules/math/BezierCurve.cpp"
        $259 = $258; //@line 76 "modules/math/BezierCurve.cpp"
        $260 = HEAP32[$259>>2]|0; //@line 76 "modules/math/BezierCurve.cpp"
        $261 = $249; //@line 76 "modules/math/BezierCurve.cpp"
        $262 = $261; //@line 76 "modules/math/BezierCurve.cpp"
        HEAP32[$262>>2] = $257; //@line 76 "modules/math/BezierCurve.cpp"
        $263 = (($261) + 4)|0; //@line 76 "modules/math/BezierCurve.cpp"
        $264 = $263; //@line 76 "modules/math/BezierCurve.cpp"
        HEAP32[$264>>2] = $260; //@line 76 "modules/math/BezierCurve.cpp"
        $265 = (($i2$041) + 1)|0; //@line 75 "modules/math/BezierCurve.cpp"
        $266 = HEAP32[$3>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $267 = HEAP32[$170>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $268 = (($266) - ($267))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $269 = $268 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $270 = ($269>>>0)>($265>>>0); //@line 75 "modules/math/BezierCurve.cpp"
        $271 = $267; //@line 75 "modules/math/BezierCurve.cpp"
        if (!($270)) {
         $$in = $266;$226 = $271;
         break L48;
        }
        $$pre61 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $$pre62 = HEAP32[$left>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $244 = $$pre61;$245 = $$pre62;$252 = $269;$254 = $271;$i2$041 = $265;
       }
      } else {
       $$in = $197;$226 = $202;
      }
     } while(0);
     $224 = $$in; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $225 = ($226|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $227 = $226; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($225)) {
      $228 = ($224|0)==($226|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($228)) {
       $scevgep$i$i$i4 = ((($224)) + -8|0);
       $229 = $scevgep$i$i$i4;
       $230 = (($229) - ($227))|0;
       $231 = $230 >>> 3;
       $232 = $231 ^ -1;
       $scevgep4$i$i$i5 = (($224) + ($232<<3)|0);
       HEAP32[$3>>2] = $scevgep4$i$i$i5; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($226); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     $233 = HEAP32[$left>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $234 = ($233|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $235 = $233; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if (!($234)) {
      $236 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $237 = ($236|0)==($233|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($237)) {
       $scevgep$i$i$i = ((($236)) + -8|0);
       $238 = $scevgep$i$i$i;
       $239 = (($238) - ($235))|0;
       $240 = $239 >>> 3;
       $241 = $240 ^ -1;
       $scevgep4$i$i$i = (($236) + ($241<<3)|0);
       HEAP32[$1>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($233); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     }
     STACKTOP = sp;return; //@line 77 "modules/math/BezierCurve.cpp"
    }
   }
  }
 } while(0);
 if ((label|0) == 11) {
  $45 = ___cxa_find_matching_catch_2()|0;
  $46 = tempRet0;
  $lpad$phi$index = $45;$lpad$phi$index2 = $46;
 }
 $47 = HEAP32[$right>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $48 = ($47|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $49 = $47; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($48)) {
  $50 = HEAP32[$3>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $51 = ($50|0)==($47|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($51)) {
   $scevgep$i$i$i10 = ((($50)) + -8|0);
   $52 = $scevgep$i$i$i10;
   $53 = (($52) - ($49))|0;
   $54 = $53 >>> 3;
   $55 = $54 ^ -1;
   $scevgep4$i$i$i11 = (($50) + ($55<<3)|0);
   HEAP32[$3>>2] = $scevgep4$i$i$i11; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($47); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 }
 $56 = HEAP32[$left>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $57 = ($56|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $58 = $56; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($57) {
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $59 = HEAP32[$1>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $60 = ($59|0)==($56|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($60)) {
  $scevgep$i$i$i15 = ((($59)) + -8|0);
  $61 = $scevgep$i$i$i15;
  $62 = (($61) - ($58))|0;
  $63 = $62 >>> 3;
  $64 = $63 ^ -1;
  $scevgep4$i$i$i16 = (($59) + ($64<<3)|0);
  HEAP32[$1>>2] = $scevgep4$i$i$i16; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($56); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE8__appendEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$0$i = 0, $$0$i2 = 0, $$0$i45 = 0, $$0$i46 = 0, $$in = 0, $$pre = 0, $$pre7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $scevgep$i = 0, $scevgep$i3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = ((($this)) + 4|0); //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$2>>2]|0; //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($1) - ($3))|0; //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = $4 >> 3; //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)<($__n>>>0); //@line 1044 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = $3; //@line 988 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($6)) {
  $$0$i = $__n;$8 = $7;
  while(1) {
   HEAPF32[$8>>2] = 0.0; //@line 219 "./common/Vector.h"
   $9 = ((($8)) + 4|0); //@line 220 "./common/Vector.h"
   HEAPF32[$9>>2] = 0.0; //@line 220 "./common/Vector.h"
   $10 = ((($8)) + 8|0); //@line 989 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $11 = (($$0$i) + -1)|0; //@line 990 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $12 = ($11|0)==(0); //@line 992 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($12) {
    break;
   } else {
    $$0$i = $11;$8 = $10;
   }
  }
  $scevgep$i = (($7) + ($__n<<3)|0); //@line 985 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$2>>2] = $scevgep$i; //@line 988 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  return; //@line 1053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $13 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = (($3) - ($13))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = $14 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = (($15) + ($__n))|0; //@line 1049 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ($16>>>0)>(536870911); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($17) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$0>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre7 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre7;$19 = $$pre;
 } else {
  $$in = $13;$19 = $1;
 }
 $18 = (($19) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $20 = $18 >> 3; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $21 = ($20>>>0)<(268435455); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($21) {
  $25 = $18 >> 2; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $26 = ($25>>>0)<($16>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $27 = $26 ? $16 : $25; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $28 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $29 = (($28) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $30 = $29 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $31 = ($27|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($31) {
   $$0$i46 = 0;$35 = 0;$36 = $30;$46 = $28;
  } else {
   $$0$i45 = $27;$54 = $28;$55 = $30;
   label = 9;
  }
 } else {
  $22 = HEAP32[$2>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $23 = (($22) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $24 = $23 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i45 = 536870911;$54 = $22;$55 = $24;
  label = 9;
 }
 if ((label|0) == 9) {
  $32 = $$0$i45 << 3; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $33 = (__Znwj($32)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i46 = $$0$i45;$35 = $33;$36 = $55;$46 = $54;
 }
 $34 = (($35) + ($36<<3)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $37 = (($35) + ($$0$i46<<3)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $$0$i2 = $__n;$38 = $34;
 while(1) {
  HEAPF32[$38>>2] = 0.0; //@line 219 "./common/Vector.h"
  $39 = ((($38)) + 4|0); //@line 220 "./common/Vector.h"
  HEAPF32[$39>>2] = 0.0; //@line 220 "./common/Vector.h"
  $40 = ((($38)) + 8|0); //@line 218 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $41 = (($$0$i2) + -1)|0; //@line 219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $42 = ($41|0)==(0); //@line 220 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($42) {
   break;
  } else {
   $$0$i2 = $41;$38 = $40;
  }
 }
 $43 = $37; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $scevgep$i3 = (($34) + ($__n<<3)|0); //@line 215 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $44 = $scevgep$i3; //@line 218 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $45 = (($46) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = $45 >> 3; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $48 = (0 - ($47))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $49 = (($34) + ($48<<3)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $50 = $49; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $51 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($49|0),($51|0),($45|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $50; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$2>>2] = $44; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $43; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $52 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($52) {
  return; //@line 1053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $53 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($53); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1053 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNK4love4math11BezierCurve13renderSegmentEddj($agg$result,$this,$start,$end,$accuracy) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $start = +$start;
 $end = +$end;
 $accuracy = $accuracy|0;
 var $$0 = 0, $$01 = 0, $$idx1618 = 0, $$phi$trans$insert19 = 0, $$pre = 0, $$pre$phi23Z2D = 0, $$pre$phi24Z2D = 0, $$pre$phi26Z2D = 0, $$pre$phiZ2D = 0, $$pre21 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0.0;
 var $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $eh$lpad$body$index3Z2D = 0, $eh$lpad$body$indexZ2D = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i$i = 0, $scevgep$i$i$i2 = 0, $scevgep$i$i$i6 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i$i = 0, $scevgep4$i$i$i3 = 0, $scevgep4$i$i$i7 = 0, $vararg_buffer = 0, $vertices = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 544 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $vertices = sp + 4|0;
 $0 = ((($this)) + 8|0); //@line 238 "modules/math/BezierCurve.cpp"
 $1 = ((($this)) + 12|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$1>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = (($2) - ($3))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = $4 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)<(2); //@line 238 "modules/math/BezierCurve.cpp"
 if ($6) {
  $7 = (___cxa_allocate_exception(16)|0); //@line 239 "modules/math/BezierCurve.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($7|0),(444011|0),($vararg_buffer|0)); //@line 239 "modules/math/BezierCurve.cpp"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if (!($9)) {
   ___cxa_throw(($7|0),(96|0),(344|0)); //@line 239 "modules/math/BezierCurve.cpp"
   // unreachable; //@line 239 "modules/math/BezierCurve.cpp"
  }
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  ___cxa_free_exception(($7|0)); //@line 239 "modules/math/BezierCurve.cpp"
  $$0 = $10;$$01 = $11;
  ___resumeException($$0|0);
  // unreachable;
 }
 __ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEEC2ERKS5_($vertices,$0); //@line 240 "modules/math/BezierCurve.cpp"
 __THREW__ = 0;
 invoke_vii(4204,($vertices|0),($accuracy|0)); //@line 241 "modules/math/BezierCurve.cpp"
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 L8: do {
  if ($13) {
   $26 = ___cxa_find_matching_catch_2()|0;
   $27 = tempRet0;
   $$pre$phi26Z2D = $vertices;$eh$lpad$body$index3Z2D = $27;$eh$lpad$body$indexZ2D = $26;
  } else {
   $14 = $start == $end; //@line 242 "modules/math/BezierCurve.cpp"
   L11: do {
    if ($14) {
     $15 = HEAP32[$vertices>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $16 = ((($vertices)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $17 = HEAP32[$16>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $18 = ($17|0)==($15|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $19 = $15; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $20 = $17; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($18) {
      $$pre$phi23Z2D = $16;$$pre$phiZ2D = $vertices;$95 = $19;$96 = $20;
      label = 27;
     } else {
      $scevgep$i$i$i2 = ((($17)) + -8|0);
      $21 = $scevgep$i$i$i2;
      $22 = (($21) - ($19))|0;
      $23 = $22 >>> 3;
      $24 = $23 ^ -1;
      $scevgep4$i$i$i3 = (($17) + ($24<<3)|0);
      HEAP32[$16>>2] = $scevgep4$i$i$i3; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $25 = $scevgep4$i$i$i3; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$pre$phi23Z2D = $16;$$pre$phiZ2D = $vertices;$95 = $19;$96 = $25;
      label = 27;
     }
    } else {
     $38 = $start < $end; //@line 246 "modules/math/BezierCurve.cpp"
     if (!($38)) {
      $$pre = HEAP32[$vertices>>2]|0; //@line 1246 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$phi$trans$insert19 = ((($vertices)) + 4|0); //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$pre21 = HEAP32[$$phi$trans$insert19>>2]|0; //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $$pre$phi23Z2D = $$phi$trans$insert19;$$pre$phiZ2D = $vertices;$95 = $$pre;$96 = $$pre21;
      label = 27;
      break;
     }
     $39 = ((($vertices)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $40 = HEAP32[$39>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $41 = HEAP32[$vertices>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $42 = (($40) - ($41))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $43 = $42 >> 3; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $44 = (+($43>>>0)); //@line 248 "modules/math/BezierCurve.cpp"
     $45 = $44 * $start; //@line 248 "modules/math/BezierCurve.cpp"
     $46 = (~~(($45))>>>0); //@line 248 "modules/math/BezierCurve.cpp"
     $47 = $44 * $end; //@line 249 "modules/math/BezierCurve.cpp"
     $48 = $47 + 0.5; //@line 249 "modules/math/BezierCurve.cpp"
     $49 = (~~(($48))>>>0); //@line 249 "modules/math/BezierCurve.cpp"
     $50 = $41; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $51 = (($50) + ($46<<3)|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $52 = (($50) + ($49<<3)|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     HEAP32[$agg$result>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $53 = ((($agg$result)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$53>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $54 = ((($agg$result)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$54>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $$idx1618 = (($49) - ($46))|0; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $55 = $$idx1618 << 3; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $56 = $55 >> 3; //@line 1383 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $57 = ($56|0)==(0); //@line 1177 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     if ($57) {
      $$pre$phi24Z2D = $vertices;
     } else {
      $58 = ($56>>>0)>(536870911); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($58) {
       __THREW__ = 0;
       invoke_vi(3528,($agg$result|0)); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $59 = __THREW__; __THREW__ = 0;
       $60 = $59&1;
       if (!($60)) {
        label = 19;
       }
      } else {
       label = 19;
      }
      if ((label|0) == 19) {
       __THREW__ = 0;
       $61 = (invoke_ii(3348,($55|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
       $62 = __THREW__; __THREW__ = 0;
       $63 = $62&1;
       if (!($63)) {
        HEAP32[$53>>2] = $61; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        HEAP32[$agg$result>>2] = $61; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $64 = (($61) + ($56<<3)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $65 = ((($agg$result)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        HEAP32[$65>>2] = $64; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $66 = ($49|0)==($46|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        if ($66) {
         $$pre$phi24Z2D = $vertices;
         break;
        }
        $68 = $51;$75 = $61;
        while(1) {
         $67 = $68; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $69 = $67; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $70 = HEAP32[$69>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $71 = (($67) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $72 = $71; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $73 = HEAP32[$72>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $74 = $75; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $76 = $74; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         HEAP32[$76>>2] = $70; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $77 = (($74) + 4)|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $78 = $77; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         HEAP32[$78>>2] = $73; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         $79 = HEAP32[$53>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $80 = ((($79)) + 8|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$53>>2] = $80; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $81 = ((($68)) + 8|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
         $82 = ($81|0)==($52|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
         if ($82) {
          $$pre$phi24Z2D = $vertices;
          break L11;
         } else {
          $68 = $81;$75 = $80;
         }
        }
       }
      }
      $83 = ___cxa_find_matching_catch_2()|0;
      $84 = tempRet0;
      $85 = HEAP32[$agg$result>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $86 = ($85|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $87 = $85; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($86) {
       $$pre$phi26Z2D = $vertices;$eh$lpad$body$index3Z2D = $84;$eh$lpad$body$indexZ2D = $83;
       break L8;
      }
      $88 = HEAP32[$53>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $89 = ($88|0)==($85|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if (!($89)) {
       $scevgep$i$i$i$i = ((($88)) + -8|0);
       $90 = $scevgep$i$i$i$i;
       $91 = (($90) - ($87))|0;
       $92 = $91 >>> 3;
       $93 = $92 ^ -1;
       $scevgep4$i$i$i$i = (($88) + ($93<<3)|0);
       HEAP32[$53>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      __ZdlPv($85); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $$pre$phi26Z2D = $vertices;$eh$lpad$body$index3Z2D = $84;$eh$lpad$body$indexZ2D = $83;
      break L8;
     }
    }
   } while(0);
   if ((label|0) == 27) {
    $94 = ((($agg$result)) + 4|0); //@line 461 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$agg$result>>2] = $95; //@line 1246 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$94>>2] = $96; //@line 1247 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $97 = ((($vertices)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $98 = HEAP32[$97>>2]|0; //@line 1248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $99 = ((($agg$result)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    HEAP32[$99>>2] = $98; //@line 1248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$97>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$$pre$phi23Z2D>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    HEAP32[$$pre$phiZ2D>>2] = 0; //@line 1249 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $$pre$phi24Z2D = $$pre$phiZ2D;
   }
   $100 = HEAP32[$$pre$phi24Z2D>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $101 = ($100|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $102 = $100; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($101) {
    STACKTOP = sp;return; //@line 259 "modules/math/BezierCurve.cpp"
   }
   $103 = ((($vertices)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $104 = HEAP32[$103>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $105 = ($104|0)==($100|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if (!($105)) {
    $scevgep$i$i$i = ((($104)) + -8|0);
    $106 = $scevgep$i$i$i;
    $107 = (($106) - ($102))|0;
    $108 = $107 >>> 3;
    $109 = $108 ^ -1;
    $scevgep4$i$i$i = (($104) + ($109<<3)|0);
    HEAP32[$103>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   __ZdlPv($100); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   STACKTOP = sp;return; //@line 259 "modules/math/BezierCurve.cpp"
  }
 } while(0);
 $28 = HEAP32[$$pre$phi26Z2D>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $29 = ($28|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = $28; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($29) {
  $$0 = $eh$lpad$body$indexZ2D;$$01 = $eh$lpad$body$index3Z2D;
  ___resumeException($$0|0);
  // unreachable;
 }
 $31 = ((($vertices)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = HEAP32[$31>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $33 = ($32|0)==($28|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($33)) {
  $scevgep$i$i$i6 = ((($32)) + -8|0);
  $34 = $scevgep$i$i$i6;
  $35 = (($34) - ($30))|0;
  $36 = $35 >>> 3;
  $37 = $36 ^ -1;
  $scevgep4$i$i$i7 = (($32) + ($37<<3)|0);
  HEAP32[$31>>2] = $scevgep4$i$i$i7; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($28); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $$0 = $eh$lpad$body$indexZ2D;$$01 = $eh$lpad$body$index3Z2D;
 ___resumeException($$0|0);
 // unreachable;
}
function __ZN4love5mouse3sdl6CursorD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (32460); //@line 82 "modules/mouse/sdl/Cursor.cpp"
 $0 = ((($this)) + 8|0); //@line 83 "modules/mouse/sdl/Cursor.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 83 "modules/mouse/sdl/Cursor.cpp"
 $2 = ($1|0)==(0|0); //@line 83 "modules/mouse/sdl/Cursor.cpp"
 if ($2) {
  __ZN4love5mouse6CursorD2Ev($this); //@line 85 "modules/mouse/sdl/Cursor.cpp"
  return; //@line 85 "modules/mouse/sdl/Cursor.cpp"
 }
 __THREW__ = 0;
 invoke_vi(4206,($1|0)); //@line 84 "modules/mouse/sdl/Cursor.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $5 = ___cxa_find_matching_catch_3(0|0)|0;
  $6 = tempRet0;
  __ZN4love5mouse6CursorD2Ev($this); //@line 85 "modules/mouse/sdl/Cursor.cpp"
  ___clang_call_terminate($5); //@line 85 "modules/mouse/sdl/Cursor.cpp"
  // unreachable; //@line 85 "modules/mouse/sdl/Cursor.cpp"
 } else {
  __ZN4love5mouse6CursorD2Ev($this); //@line 85 "modules/mouse/sdl/Cursor.cpp"
  return; //@line 85 "modules/mouse/sdl/Cursor.cpp"
 }
}
function __ZN4love5mouse3sdl6CursorD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (32460); //@line 82 "modules/mouse/sdl/Cursor.cpp"
 $0 = ((($this)) + 8|0); //@line 83 "modules/mouse/sdl/Cursor.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 83 "modules/mouse/sdl/Cursor.cpp"
 $2 = ($1|0)==(0|0); //@line 83 "modules/mouse/sdl/Cursor.cpp"
 if (!($2)) {
  __THREW__ = 0;
  invoke_vi(4206,($1|0)); //@line 84 "modules/mouse/sdl/Cursor.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if ($4) {
   $5 = ___cxa_find_matching_catch_3(0|0)|0;
   $6 = tempRet0;
   __ZN4love5mouse6CursorD2Ev($this); //@line 85 "modules/mouse/sdl/Cursor.cpp"
   ___clang_call_terminate($5); //@line 85 "modules/mouse/sdl/Cursor.cpp"
   // unreachable; //@line 85 "modules/mouse/sdl/Cursor.cpp"
  }
 }
 __ZN4love5mouse6CursorD2Ev($this); //@line 85 "modules/mouse/sdl/Cursor.cpp"
 __ZdlPv($this); //@line 82 "modules/mouse/sdl/Cursor.cpp"
 return; //@line 85 "modules/mouse/sdl/Cursor.cpp"
}
function __ZNK4love5mouse3sdl6Cursor9getHandleEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 89 "modules/mouse/sdl/Cursor.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 89 "modules/mouse/sdl/Cursor.cpp"
 return ($1|0); //@line 89 "modules/mouse/sdl/Cursor.cpp"
}
function __ZNK4love5mouse3sdl6Cursor7getTypeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 94 "modules/mouse/sdl/Cursor.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 94 "modules/mouse/sdl/Cursor.cpp"
 return ($1|0); //@line 94 "modules/mouse/sdl/Cursor.cpp"
}
function __ZNK4love5mouse3sdl6Cursor13getSystemTypeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 99 "modules/mouse/sdl/Cursor.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 99 "modules/mouse/sdl/Cursor.cpp"
 return ($1|0); //@line 99 "modules/mouse/sdl/Cursor.cpp"
}
function __GLOBAL__sub_I_Cursor_cpp_4586() {
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i = 0, $i$01$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[(657940)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(657948)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(657956)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(657964)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(657972)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(657980)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(657988)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(657996)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(658004)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(658012)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(658020)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(658028)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(658036)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(658044)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(658052)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(658060)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(658068)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(658076)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(658084)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(658092)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(658100)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(658108)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(658116)>>0] = 0; //@line 90 "./common/EnumMap.h"
 HEAP8[(658124)>>0] = 0; //@line 90 "./common/EnumMap.h"
 $i$01$i$i = 0;
 while(1) {
  $0 = (32488 + ($i$01$i$i<<3)|0); //@line 46 "./common/EnumMap.h"
  $1 = HEAP32[$0>>2]|0; //@line 46 "./common/EnumMap.h"
  $2 = (((32488 + ($i$01$i$i<<3)|0)) + 4|0); //@line 47 "./common/EnumMap.h"
  $3 = HEAP32[$2>>2]|0; //@line 47 "./common/EnumMap.h"
  $4 = ($1>>>0)<(12); //@line 49 "./common/EnumMap.h"
  if ($4) {
   $5 = (658032 + ($1<<3)|0); //@line 51 "./common/EnumMap.h"
   HEAP32[$5>>2] = $3; //@line 51 "./common/EnumMap.h"
   $6 = (((658032 + ($1<<3)|0)) + 4|0); //@line 52 "./common/EnumMap.h"
   HEAP8[$6>>0] = 1; //@line 52 "./common/EnumMap.h"
  }
  $7 = ($3>>>0)<(12); //@line 54 "./common/EnumMap.h"
  if ($7) {
   $8 = (657936 + ($3<<3)|0); //@line 56 "./common/EnumMap.h"
   HEAP32[$8>>2] = $1; //@line 56 "./common/EnumMap.h"
   $9 = (((657936 + ($3<<3)|0)) + 4|0); //@line 57 "./common/EnumMap.h"
   HEAP8[$9>>0] = 1; //@line 57 "./common/EnumMap.h"
  }
  $10 = (($i$01$i$i) + 1)|0; //@line 44 "./common/EnumMap.h"
  $exitcond$i$i = ($10|0)==(12); //@line 44 "./common/EnumMap.h"
  if ($exitcond$i$i) {
   break;
  } else {
   $i$01$i$i = $10;
  }
 }
 return;
}
function __ZN4love5mouse3sdl6CursorC2EPNS_5image9ImageDataEii($this,$data,$hotx,$hoty) {
 $this = $this|0;
 $data = $data|0;
 $hotx = $hotx|0;
 $hoty = $hoty|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 __ZN4love6ObjectC2Ev($this); //@line 34 "./modules/mouse/Cursor.h"
 HEAP32[$this>>2] = (32460); //@line 36 "modules/mouse/sdl/Cursor.cpp"
 $0 = ((($this)) + 8|0); //@line 33 "modules/mouse/sdl/Cursor.cpp"
 HEAP32[$0>>2] = 0; //@line 33 "modules/mouse/sdl/Cursor.cpp"
 $1 = ((($this)) + 12|0); //@line 34 "modules/mouse/sdl/Cursor.cpp"
 HEAP32[$1>>2] = 1; //@line 34 "modules/mouse/sdl/Cursor.cpp"
 $2 = ((($this)) + 16|0); //@line 35 "modules/mouse/sdl/Cursor.cpp"
 HEAP32[$2>>2] = 12; //@line 35 "modules/mouse/sdl/Cursor.cpp"
 __THREW__ = 0;
 $3 = (invoke_ii(3984,($data|0))|0); //@line 50 "modules/mouse/sdl/Cursor.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 do {
  if (!($5)) {
   __THREW__ = 0;
   $6 = (invoke_ii(4081,($data|0))|0); //@line 51 "modules/mouse/sdl/Cursor.cpp"
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if (!($8)) {
    $9 = HEAP32[$data>>2]|0; //@line 54 "modules/mouse/sdl/Cursor.cpp"
    $10 = ((($9)) + 16|0); //@line 54 "modules/mouse/sdl/Cursor.cpp"
    $11 = HEAP32[$10>>2]|0; //@line 54 "modules/mouse/sdl/Cursor.cpp"
    __THREW__ = 0;
    $12 = (invoke_ii($11|0,($data|0))|0); //@line 54 "modules/mouse/sdl/Cursor.cpp"
    $13 = __THREW__; __THREW__ = 0;
    $14 = $13&1;
    if (!($14)) {
     $15 = $3 << 2; //@line 52 "modules/mouse/sdl/Cursor.cpp"
     __THREW__ = 0;
     $16 = (invoke_iiiiiiiiii(3810,($12|0),($3|0),($6|0),32,($15|0),255,65280,16711680,-16777216)|0); //@line 54 "modules/mouse/sdl/Cursor.cpp"
     $17 = __THREW__; __THREW__ = 0;
     $18 = $17&1;
     if (!($18)) {
      $19 = ($16|0)==(0|0); //@line 55 "modules/mouse/sdl/Cursor.cpp"
      if ($19) {
       $20 = (___cxa_allocate_exception(16)|0); //@line 56 "modules/mouse/sdl/Cursor.cpp"
       __THREW__ = 0;
       invoke_viii(3349,($20|0),(444192|0),($vararg_buffer|0)); //@line 56 "modules/mouse/sdl/Cursor.cpp"
       $21 = __THREW__; __THREW__ = 0;
       $22 = $21&1;
       if (!($22)) {
        __THREW__ = 0;
        invoke_viii(3350,($20|0),(96|0),(344|0)); //@line 56 "modules/mouse/sdl/Cursor.cpp"
        $23 = __THREW__; __THREW__ = 0;
        break;
       }
       $26 = ___cxa_find_matching_catch_2()|0;
       $27 = tempRet0;
       ___cxa_free_exception(($20|0)); //@line 56 "modules/mouse/sdl/Cursor.cpp"
       $$0 = $26;$$01 = $27;
       __ZN4love5mouse6CursorD2Ev($this); //@line 63 "modules/mouse/sdl/Cursor.cpp"
       ___resumeException($$0|0);
       // unreachable;
      }
      __THREW__ = 0;
      $28 = (invoke_iiii(4207,($16|0),($hotx|0),($hoty|0))|0); //@line 58 "modules/mouse/sdl/Cursor.cpp"
      $29 = __THREW__; __THREW__ = 0;
      $30 = $29&1;
      if (!($30)) {
       HEAP32[$0>>2] = $28; //@line 58 "modules/mouse/sdl/Cursor.cpp"
       __THREW__ = 0;
       invoke_vi(4208,($16|0)); //@line 59 "modules/mouse/sdl/Cursor.cpp"
       $31 = __THREW__; __THREW__ = 0;
       $32 = $31&1;
       if (!($32)) {
        $33 = HEAP32[$0>>2]|0; //@line 61 "modules/mouse/sdl/Cursor.cpp"
        $34 = ($33|0)==(0|0); //@line 61 "modules/mouse/sdl/Cursor.cpp"
        if (!($34)) {
         STACKTOP = sp;return; //@line 63 "modules/mouse/sdl/Cursor.cpp"
        }
        $35 = (___cxa_allocate_exception(16)|0); //@line 62 "modules/mouse/sdl/Cursor.cpp"
        __THREW__ = 0;
        $36 = (invoke_i(3814)|0); //@line 62 "modules/mouse/sdl/Cursor.cpp"
        $37 = __THREW__; __THREW__ = 0;
        $38 = $37&1;
        if (!($38)) {
         __THREW__ = 0;
         HEAP32[$vararg_buffer1>>2] = $36; //@line 62 "modules/mouse/sdl/Cursor.cpp"
         invoke_viii(3349,($35|0),(444229|0),($vararg_buffer1|0)); //@line 62 "modules/mouse/sdl/Cursor.cpp"
         $39 = __THREW__; __THREW__ = 0;
         $40 = $39&1;
         if (!($40)) {
          __THREW__ = 0;
          invoke_viii(3350,($35|0),(96|0),(344|0)); //@line 62 "modules/mouse/sdl/Cursor.cpp"
          $41 = __THREW__; __THREW__ = 0;
          break;
         }
        }
        $42 = ___cxa_find_matching_catch_2()|0;
        $43 = tempRet0;
        ___cxa_free_exception(($35|0)); //@line 62 "modules/mouse/sdl/Cursor.cpp"
        $$0 = $42;$$01 = $43;
        __ZN4love5mouse6CursorD2Ev($this); //@line 63 "modules/mouse/sdl/Cursor.cpp"
        ___resumeException($$0|0);
        // unreachable;
       }
      }
     }
    }
   }
  }
 } while(0);
 $24 = ___cxa_find_matching_catch_2()|0;
 $25 = tempRet0;
 $$0 = $24;$$01 = $25;
 __ZN4love5mouse6CursorD2Ev($this); //@line 63 "modules/mouse/sdl/Cursor.cpp"
 ___resumeException($$0|0);
 // unreachable;
}
function __ZN4love5mouse3sdl6CursorC2ENS0_6Cursor12SystemCursorE($this,$cursortype) {
 $this = $this|0;
 $cursortype = $cursortype|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 __ZN4love6ObjectC2Ev($this); //@line 34 "./modules/mouse/Cursor.h"
 HEAP32[$this>>2] = (32460); //@line 69 "modules/mouse/sdl/Cursor.cpp"
 $0 = ((($this)) + 8|0); //@line 66 "modules/mouse/sdl/Cursor.cpp"
 HEAP32[$0>>2] = 0; //@line 66 "modules/mouse/sdl/Cursor.cpp"
 $1 = ((($this)) + 12|0); //@line 67 "modules/mouse/sdl/Cursor.cpp"
 HEAP32[$1>>2] = 0; //@line 67 "modules/mouse/sdl/Cursor.cpp"
 $2 = ((($this)) + 16|0); //@line 68 "modules/mouse/sdl/Cursor.cpp"
 HEAP32[$2>>2] = $cursortype; //@line 68 "modules/mouse/sdl/Cursor.cpp"
 $3 = ($cursortype>>>0)<(12); //@line 64 "./common/EnumMap.h"
 do {
  if ($3) {
   $4 = (((658032 + ($cursortype<<3)|0)) + 4|0); //@line 64 "./common/EnumMap.h"
   $5 = HEAP8[$4>>0]|0; //@line 64 "./common/EnumMap.h"
   $6 = ($5<<24>>24)==(0); //@line 64 "./common/EnumMap.h"
   if ($6) {
    label = 6;
   } else {
    $7 = (658032 + ($cursortype<<3)|0); //@line 66 "./common/EnumMap.h"
    $8 = HEAP32[$7>>2]|0; //@line 66 "./common/EnumMap.h"
    __THREW__ = 0;
    $9 = (invoke_ii(4209,($8|0))|0); //@line 73 "modules/mouse/sdl/Cursor.cpp"
    $10 = __THREW__; __THREW__ = 0;
    $11 = $10&1;
    if (!($11)) {
     HEAP32[$0>>2] = $9; //@line 73 "modules/mouse/sdl/Cursor.cpp"
     $12 = ($9|0)==(0|0); //@line 77 "modules/mouse/sdl/Cursor.cpp"
     if (!($12)) {
      STACKTOP = sp;return; //@line 79 "modules/mouse/sdl/Cursor.cpp"
     }
     $21 = (___cxa_allocate_exception(16)|0); //@line 78 "modules/mouse/sdl/Cursor.cpp"
     __THREW__ = 0;
     $22 = (invoke_i(3814)|0); //@line 78 "modules/mouse/sdl/Cursor.cpp"
     $23 = __THREW__; __THREW__ = 0;
     $24 = $23&1;
     if (!($24)) {
      __THREW__ = 0;
      HEAP32[$vararg_buffer1>>2] = $22; //@line 78 "modules/mouse/sdl/Cursor.cpp"
      invoke_viii(3349,($21|0),(444297|0),($vararg_buffer1|0)); //@line 78 "modules/mouse/sdl/Cursor.cpp"
      $25 = __THREW__; __THREW__ = 0;
      $26 = $25&1;
      if (!($26)) {
       __THREW__ = 0;
       invoke_viii(3350,($21|0),(96|0),(344|0)); //@line 78 "modules/mouse/sdl/Cursor.cpp"
       $27 = __THREW__; __THREW__ = 0;
       break;
      }
     }
     $28 = ___cxa_find_matching_catch_2()|0;
     $29 = tempRet0;
     ___cxa_free_exception(($21|0)); //@line 78 "modules/mouse/sdl/Cursor.cpp"
     $$0 = $29;$$01 = $28;
     __ZN4love5mouse6CursorD2Ev($this); //@line 79 "modules/mouse/sdl/Cursor.cpp"
     ___resumeException($$01|0);
     // unreachable;
    }
   }
  } else {
   label = 6;
  }
 } while(0);
 do {
  if ((label|0) == 6) {
   $15 = (___cxa_allocate_exception(16)|0); //@line 75 "modules/mouse/sdl/Cursor.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($15|0),(444254|0),($vararg_buffer|0)); //@line 75 "modules/mouse/sdl/Cursor.cpp"
   $16 = __THREW__; __THREW__ = 0;
   $17 = $16&1;
   if (!($17)) {
    __THREW__ = 0;
    invoke_viii(3350,($15|0),(96|0),(344|0)); //@line 75 "modules/mouse/sdl/Cursor.cpp"
    $18 = __THREW__; __THREW__ = 0;
    break;
   }
   $19 = ___cxa_find_matching_catch_2()|0;
   $20 = tempRet0;
   ___cxa_free_exception(($15|0)); //@line 75 "modules/mouse/sdl/Cursor.cpp"
   $$0 = $20;$$01 = $19;
   __ZN4love5mouse6CursorD2Ev($this); //@line 79 "modules/mouse/sdl/Cursor.cpp"
   ___resumeException($$01|0);
   // unreachable;
  }
 } while(0);
 $13 = ___cxa_find_matching_catch_2()|0;
 $14 = tempRet0;
 $$0 = $14;$$01 = $13;
 __ZN4love5mouse6CursorD2Ev($this); //@line 79 "modules/mouse/sdl/Cursor.cpp"
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN4love7physics5box2d4BodyD2Ev($this) {
 $this = $this|0;
 var $$pr$pre = 0, $$pr3 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (32592); //@line 69 "modules/physics/box2d/Body.cpp"
 $0 = ((($this)) + 16|0); //@line 70 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 70 "modules/physics/box2d/Body.cpp"
 $2 = ($1|0)==(0|0); //@line 70 "modules/physics/box2d/Body.cpp"
 do {
  if (!($2)) {
   $3 = HEAP32[$1>>2]|0; //@line 71 "modules/physics/box2d/Body.cpp"
   $4 = ($3|0)==(0|0); //@line 71 "modules/physics/box2d/Body.cpp"
   if ($4) {
    $$pr3 = $1;
   } else {
    $5 = HEAP32[$3>>2]|0; //@line 71 "modules/physics/box2d/Body.cpp"
    $6 = ((($5)) + 4|0); //@line 71 "modules/physics/box2d/Body.cpp"
    $7 = HEAP32[$6>>2]|0; //@line 71 "modules/physics/box2d/Body.cpp"
    FUNCTION_TABLE_vi[$7 & 8191]($3); //@line 71 "modules/physics/box2d/Body.cpp"
    $$pr$pre = HEAP32[$0>>2]|0; //@line 72 "modules/physics/box2d/Body.cpp"
    $8 = ($$pr$pre|0)==(0|0); //@line 72 "modules/physics/box2d/Body.cpp"
    if ($8) {
     break;
    } else {
     $$pr3 = $$pr$pre;
    }
   }
   __ZdlPv($$pr3); //@line 72 "modules/physics/box2d/Body.cpp"
  }
 } while(0);
 __ZN4love7physics4BodyD2Ev($this); //@line 73 "modules/physics/box2d/Body.cpp"
 return; //@line 73 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4BodyD0Ev($this) {
 $this = $this|0;
 var $$pr$pre$i = 0, $$pr3$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (32592); //@line 69 "modules/physics/box2d/Body.cpp"
 $0 = ((($this)) + 16|0); //@line 70 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 70 "modules/physics/box2d/Body.cpp"
 $2 = ($1|0)==(0|0); //@line 70 "modules/physics/box2d/Body.cpp"
 do {
  if (!($2)) {
   $3 = HEAP32[$1>>2]|0; //@line 71 "modules/physics/box2d/Body.cpp"
   $4 = ($3|0)==(0|0); //@line 71 "modules/physics/box2d/Body.cpp"
   if ($4) {
    $$pr3$i = $1;
   } else {
    $5 = HEAP32[$3>>2]|0; //@line 71 "modules/physics/box2d/Body.cpp"
    $6 = ((($5)) + 4|0); //@line 71 "modules/physics/box2d/Body.cpp"
    $7 = HEAP32[$6>>2]|0; //@line 71 "modules/physics/box2d/Body.cpp"
    FUNCTION_TABLE_vi[$7 & 8191]($3); //@line 71 "modules/physics/box2d/Body.cpp"
    $$pr$pre$i = HEAP32[$0>>2]|0; //@line 72 "modules/physics/box2d/Body.cpp"
    $8 = ($$pr$pre$i|0)==(0|0); //@line 72 "modules/physics/box2d/Body.cpp"
    if ($8) {
     break;
    } else {
     $$pr3$i = $$pr$pre$i;
    }
   }
   __ZdlPv($$pr3$i); //@line 72 "modules/physics/box2d/Body.cpp"
  }
 } while(0);
 __ZN4love7physics4BodyD2Ev($this); //@line 73 "modules/physics/box2d/Body.cpp"
 __ZdlPv($this); //@line 69 "modules/physics/box2d/Body.cpp"
 return; //@line 73 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4BodyC2EPNS1_5WorldE6b2Vec2NS0_4Body4TypeE($this,$world,$p,$type) {
 $this = $this|0;
 $world = $world|0;
 $p = $p|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $def = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $def = sp + 8|0;
 $0 = sp;
 __ZN4love6ObjectC2Ev($this); //@line 33 "./modules/physics/Body.h"
 HEAP32[$this>>2] = (32592); //@line 44 "modules/physics/box2d/Body.cpp"
 $1 = ((($this)) + 12|0); //@line 42 "modules/physics/box2d/Body.cpp"
 HEAP32[$1>>2] = $world; //@line 42 "modules/physics/box2d/Body.cpp"
 $2 = ((($this)) + 16|0); //@line 43 "modules/physics/box2d/Body.cpp"
 HEAP32[$2>>2] = 0; //@line 43 "modules/physics/box2d/Body.cpp"
 __THREW__ = 0;
 $3 = (invoke_ii(3348,4)|0); //@line 45 "modules/physics/box2d/Body.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 L1: do {
  if (!($5)) {
   HEAP32[$2>>2] = $3; //@line 45 "modules/physics/box2d/Body.cpp"
   HEAP32[$3>>2] = 0; //@line 46 "modules/physics/box2d/Body.cpp"
   $6 = ((($def)) + 44|0); //@line 56 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAP32[$6>>2] = 0; //@line 56 "./libraries/Box2D/Dynamics/b2Body.h"
   $7 = ((($def)) + 4|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
   $8 = ((($def)) + 36|0); //@line 63 "./libraries/Box2D/Dynamics/b2Body.h"
   ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;HEAP32[$7+12>>2]=0|0;HEAP32[$7+16>>2]=0|0;HEAP32[$7+20>>2]=0|0;HEAP32[$7+24>>2]=0|0;HEAP32[$7+28>>2]=0|0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
   HEAP8[$8>>0] = 1; //@line 63 "./libraries/Box2D/Dynamics/b2Body.h"
   $9 = ((($def)) + 37|0); //@line 64 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAP8[$9>>0] = 1; //@line 64 "./libraries/Box2D/Dynamics/b2Body.h"
   $10 = ((($def)) + 38|0); //@line 65 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAP8[$10>>0] = 0; //@line 65 "./libraries/Box2D/Dynamics/b2Body.h"
   $11 = ((($def)) + 39|0); //@line 66 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAP8[$11>>0] = 0; //@line 66 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAP32[$def>>2] = 0; //@line 67 "./libraries/Box2D/Dynamics/b2Body.h"
   $12 = ((($def)) + 40|0); //@line 68 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAP8[$12>>0] = 1; //@line 68 "./libraries/Box2D/Dynamics/b2Body.h"
   $13 = ((($def)) + 48|0); //@line 69 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAPF32[$13>>2] = 1.0; //@line 69 "./libraries/Box2D/Dynamics/b2Body.h"
   __THREW__ = 0;
   invoke_vii(3695,($0|0),($p|0)); //@line 48 "modules/physics/box2d/Body.cpp"
   $14 = __THREW__; __THREW__ = 0;
   $15 = $14&1;
   if (!($15)) {
    $16 = ((($def)) + 4|0); //@line 48 "modules/physics/box2d/Body.cpp"
    $17 = $0; //@line 48 "modules/physics/box2d/Body.cpp"
    $18 = $17; //@line 48 "modules/physics/box2d/Body.cpp"
    $19 = HEAP32[$18>>2]|0; //@line 48 "modules/physics/box2d/Body.cpp"
    $20 = (($17) + 4)|0; //@line 48 "modules/physics/box2d/Body.cpp"
    $21 = $20; //@line 48 "modules/physics/box2d/Body.cpp"
    $22 = HEAP32[$21>>2]|0; //@line 48 "modules/physics/box2d/Body.cpp"
    $23 = $16; //@line 48 "modules/physics/box2d/Body.cpp"
    $24 = $23; //@line 48 "modules/physics/box2d/Body.cpp"
    HEAP32[$24>>2] = $19; //@line 48 "modules/physics/box2d/Body.cpp"
    $25 = (($23) + 4)|0; //@line 48 "modules/physics/box2d/Body.cpp"
    $26 = $25; //@line 48 "modules/physics/box2d/Body.cpp"
    HEAP32[$26>>2] = $22; //@line 48 "modules/physics/box2d/Body.cpp"
    $27 = HEAP32[$2>>2]|0; //@line 49 "modules/physics/box2d/Body.cpp"
    HEAP32[$6>>2] = $27; //@line 49 "modules/physics/box2d/Body.cpp"
    $28 = ((($world)) + 20|0); //@line 50 "modules/physics/box2d/Body.cpp"
    $29 = HEAP32[$28>>2]|0; //@line 50 "modules/physics/box2d/Body.cpp"
    __THREW__ = 0;
    $30 = (invoke_iii(3712,($29|0),($def|0))|0); //@line 50 "modules/physics/box2d/Body.cpp"
    $31 = __THREW__; __THREW__ = 0;
    $32 = $31&1;
    if (!($32)) {
     $33 = ((($this)) + 8|0); //@line 50 "modules/physics/box2d/Body.cpp"
     HEAP32[$33>>2] = $30; //@line 50 "modules/physics/box2d/Body.cpp"
     $34 = HEAP32[$this>>2]|0; //@line 52 "modules/physics/box2d/Body.cpp"
     $35 = ((($34)) + 8|0); //@line 52 "modules/physics/box2d/Body.cpp"
     $36 = HEAP32[$35>>2]|0; //@line 52 "modules/physics/box2d/Body.cpp"
     __THREW__ = 0;
     invoke_vi($36|0,($this|0)); //@line 52 "modules/physics/box2d/Body.cpp"
     $37 = __THREW__; __THREW__ = 0;
     $38 = $37&1;
     if (!($38)) {
      switch ($type|0) {
      case 1:  {
       $39 = HEAP32[$33>>2]|0; //@line 292 "modules/physics/box2d/Body.cpp"
       __THREW__ = 0;
       invoke_vii(4210,($39|0),0); //@line 292 "modules/physics/box2d/Body.cpp"
       $40 = __THREW__; __THREW__ = 0;
       $41 = $40&1;
       if ($41) {
        break L1;
       }
       break;
      }
      case 2:  {
       $42 = HEAP32[$33>>2]|0; //@line 295 "modules/physics/box2d/Body.cpp"
       __THREW__ = 0;
       invoke_vii(4210,($42|0),2); //@line 295 "modules/physics/box2d/Body.cpp"
       $43 = __THREW__; __THREW__ = 0;
       $44 = $43&1;
       if ($44) {
        break L1;
       }
       break;
      }
      case 3:  {
       $45 = HEAP32[$33>>2]|0; //@line 298 "modules/physics/box2d/Body.cpp"
       __THREW__ = 0;
       invoke_vii(4210,($45|0),1); //@line 298 "modules/physics/box2d/Body.cpp"
       $46 = __THREW__; __THREW__ = 0;
       $47 = $46&1;
       if ($47) {
        break L1;
       }
       break;
      }
      default: {
      }
      }
      $48 = HEAP32[$33>>2]|0; //@line 54 "modules/physics/box2d/Body.cpp"
      __THREW__ = 0;
      invoke_vii(3703,($48|0),($this|0)); //@line 54 "modules/physics/box2d/Body.cpp"
      $49 = __THREW__; __THREW__ = 0;
      $50 = $49&1;
      if (!($50)) {
       STACKTOP = sp;return; //@line 55 "modules/physics/box2d/Body.cpp"
      }
     }
    }
   }
  }
 } while(0);
 $51 = ___cxa_find_matching_catch_2()|0;
 $52 = tempRet0;
 __ZN4love7physics4BodyD2Ev($this); //@line 55 "modules/physics/box2d/Body.cpp"
 ___resumeException($51|0);
 // unreachable;
}
function __ZN4love7physics5box2d4Body7setTypeENS0_4Body4TypeE($this,$type) {
 $this = $this|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 switch ($type|0) {
 case 1:  {
  $0 = ((($this)) + 8|0); //@line 292 "modules/physics/box2d/Body.cpp"
  $1 = HEAP32[$0>>2]|0; //@line 292 "modules/physics/box2d/Body.cpp"
  __ZN6b2Body7SetTypeE10b2BodyType($1,0); //@line 292 "modules/physics/box2d/Body.cpp"
  break;
 }
 case 2:  {
  $2 = ((($this)) + 8|0); //@line 295 "modules/physics/box2d/Body.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 295 "modules/physics/box2d/Body.cpp"
  __ZN6b2Body7SetTypeE10b2BodyType($3,2); //@line 295 "modules/physics/box2d/Body.cpp"
  break;
 }
 case 3:  {
  $4 = ((($this)) + 8|0); //@line 298 "modules/physics/box2d/Body.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 298 "modules/physics/box2d/Body.cpp"
  __ZN6b2Body7SetTypeE10b2BodyType($5,1); //@line 298 "modules/physics/box2d/Body.cpp"
  break;
 }
 default: {
 }
 }
 return; //@line 303 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body4getXEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 77 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 77 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 12|0); //@line 77 "modules/physics/box2d/Body.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 77 "modules/physics/box2d/Body.cpp"
 $4 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($3)); //@line 77 "modules/physics/box2d/Body.cpp"
 return (+$4); //@line 77 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body4getYEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 82 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 82 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 16|0); //@line 82 "modules/physics/box2d/Body.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 82 "modules/physics/box2d/Body.cpp"
 $4 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($3)); //@line 82 "modules/physics/box2d/Body.cpp"
 return (+$4); //@line 82 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body11getPositionERfS3_($this,$x_o,$y_o) {
 $this = $this|0;
 $x_o = $x_o|0;
 $y_o = $y_o|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp;
 $0 = ((($this)) + 8|0); //@line 87 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 87 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 12|0); //@line 481 "./libraries/Box2D/Dynamics/b2Body.h"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($v,$2); //@line 87 "modules/physics/box2d/Body.cpp"
 $3 = HEAP32[$v>>2]|0; //@line 88 "modules/physics/box2d/Body.cpp"
 HEAP32[$x_o>>2] = $3; //@line 88 "modules/physics/box2d/Body.cpp"
 $4 = ((($v)) + 4|0); //@line 89 "modules/physics/box2d/Body.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 89 "modules/physics/box2d/Body.cpp"
 HEAP32[$y_o>>2] = $5; //@line 89 "modules/physics/box2d/Body.cpp"
 STACKTOP = sp;return; //@line 90 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body17getLinearVelocityERfS3_($this,$x_o,$y_o) {
 $this = $this|0;
 $x_o = $x_o|0;
 $y_o = $y_o|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp;
 $0 = ((($this)) + 8|0); //@line 94 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 94 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 64|0); //@line 516 "./libraries/Box2D/Dynamics/b2Body.h"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($v,$2); //@line 94 "modules/physics/box2d/Body.cpp"
 $3 = HEAP32[$v>>2]|0; //@line 95 "modules/physics/box2d/Body.cpp"
 HEAP32[$x_o>>2] = $3; //@line 95 "modules/physics/box2d/Body.cpp"
 $4 = ((($v)) + 4|0); //@line 96 "modules/physics/box2d/Body.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 96 "modules/physics/box2d/Body.cpp"
 HEAP32[$y_o>>2] = $5; //@line 96 "modules/physics/box2d/Body.cpp"
 STACKTOP = sp;return; //@line 97 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body8getAngleEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 101 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 101 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 56|0); //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = +HEAPF32[$2>>2]; //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 return (+$3); //@line 101 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body14getWorldCenterERfS3_($this,$x_o,$y_o) {
 $this = $this|0;
 $x_o = $x_o|0;
 $y_o = $y_o|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp;
 $0 = ((($this)) + 8|0); //@line 106 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 106 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 44|0); //@line 491 "./libraries/Box2D/Dynamics/b2Body.h"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($v,$2); //@line 106 "modules/physics/box2d/Body.cpp"
 $3 = HEAP32[$v>>2]|0; //@line 107 "modules/physics/box2d/Body.cpp"
 HEAP32[$x_o>>2] = $3; //@line 107 "modules/physics/box2d/Body.cpp"
 $4 = ((($v)) + 4|0); //@line 108 "modules/physics/box2d/Body.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 108 "modules/physics/box2d/Body.cpp"
 HEAP32[$y_o>>2] = $5; //@line 108 "modules/physics/box2d/Body.cpp"
 STACKTOP = sp;return; //@line 109 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body14getLocalCenterERfS3_($this,$x_o,$y_o) {
 $this = $this|0;
 $x_o = $x_o|0;
 $y_o = $y_o|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp;
 $0 = ((($this)) + 8|0); //@line 113 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 113 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 28|0); //@line 496 "./libraries/Box2D/Dynamics/b2Body.h"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($v,$2); //@line 113 "modules/physics/box2d/Body.cpp"
 $3 = HEAP32[$v>>2]|0; //@line 114 "modules/physics/box2d/Body.cpp"
 HEAP32[$x_o>>2] = $3; //@line 114 "modules/physics/box2d/Body.cpp"
 $4 = ((($v)) + 4|0); //@line 115 "modules/physics/box2d/Body.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 115 "modules/physics/box2d/Body.cpp"
 HEAP32[$y_o>>2] = $5; //@line 115 "modules/physics/box2d/Body.cpp"
 STACKTOP = sp;return; //@line 116 "modules/physics/box2d/Body.cpp"
}
function __ZNK4love7physics5box2d4Body18getAngularVelocityEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 120 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 120 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 72|0); //@line 536 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = +HEAPF32[$2>>2]; //@line 536 "./libraries/Box2D/Dynamics/b2Body.h"
 return (+$3); //@line 120 "modules/physics/box2d/Body.cpp"
}
function __ZNK4love7physics5box2d4Body7getMassEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 125 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 125 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 116|0); //@line 541 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = +HEAPF32[$2>>2]; //@line 541 "./libraries/Box2D/Dynamics/b2Body.h"
 return (+$3); //@line 125 "modules/physics/box2d/Body.cpp"
}
function __ZNK4love7physics5box2d4Body10getInertiaEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $2 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 130 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 130 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 124|0); //@line 546 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = +HEAPF32[$2>>2]; //@line 546 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = ((($1)) + 116|0); //@line 546 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = +HEAPF32[$4>>2]; //@line 546 "./libraries/Box2D/Dynamics/b2Body.h"
 $6 = ((($1)) + 28|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $8 = $7 * $7; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($1)) + 32|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $11 = $10 * $10; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $12 = $8 + $11; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $13 = $5 * $12; //@line 546 "./libraries/Box2D/Dynamics/b2Body.h"
 $14 = $3 + $13; //@line 546 "./libraries/Box2D/Dynamics/b2Body.h"
 $15 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($14)); //@line 130 "modules/physics/box2d/Body.cpp"
 $16 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($15)); //@line 130 "modules/physics/box2d/Body.cpp"
 return (+$16); //@line 130 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body11getMassDataEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0.0, $9 = 0.0, $center = 0;
 var $data = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $data = sp + 8|0;
 $center = sp;
 $0 = ((($this)) + 8|0); //@line 136 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 136 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 116|0); //@line 551 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP32[$2>>2]|0; //@line 551 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP32[$data>>2] = $3; //@line 551 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = ((($1)) + 124|0); //@line 552 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = +HEAPF32[$4>>2]; //@line 552 "./libraries/Box2D/Dynamics/b2Body.h"
 $6 = (HEAP32[tempDoublePtr>>2]=$3,+HEAPF32[tempDoublePtr>>2]); //@line 552 "./libraries/Box2D/Dynamics/b2Body.h"
 $7 = ((($1)) + 28|0); //@line 552 "./libraries/Box2D/Dynamics/b2Body.h"
 $8 = +HEAPF32[$7>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $9 = $8 * $8; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($1)) + 32|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $11 = +HEAPF32[$10>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $12 = $11 * $11; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $13 = $9 + $12; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $14 = $6 * $13; //@line 552 "./libraries/Box2D/Dynamics/b2Body.h"
 $15 = $5 + $14; //@line 552 "./libraries/Box2D/Dynamics/b2Body.h"
 $16 = ((($data)) + 12|0); //@line 552 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$16>>2] = $15; //@line 552 "./libraries/Box2D/Dynamics/b2Body.h"
 $17 = ((($data)) + 4|0); //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $18 = $7; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = $18; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = HEAP32[$19>>2]|0; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $21 = (($18) + 4)|0; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $22 = $21; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $23 = HEAP32[$22>>2]|0; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $24 = $17; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $25 = $24; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP32[$25>>2] = $20; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $26 = (($24) + 4)|0; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $27 = $26; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP32[$27>>2] = $23; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($center,$17); //@line 137 "modules/physics/box2d/Body.cpp"
 $28 = +HEAPF32[$center>>2]; //@line 138 "modules/physics/box2d/Body.cpp"
 $29 = $28; //@line 138 "modules/physics/box2d/Body.cpp"
 _lua_pushnumber($L,$29); //@line 138 "modules/physics/box2d/Body.cpp"
 $30 = ((($center)) + 4|0); //@line 139 "modules/physics/box2d/Body.cpp"
 $31 = +HEAPF32[$30>>2]; //@line 139 "modules/physics/box2d/Body.cpp"
 $32 = $31; //@line 139 "modules/physics/box2d/Body.cpp"
 _lua_pushnumber($L,$32); //@line 139 "modules/physics/box2d/Body.cpp"
 $33 = +HEAPF32[$data>>2]; //@line 140 "modules/physics/box2d/Body.cpp"
 $34 = $33; //@line 140 "modules/physics/box2d/Body.cpp"
 _lua_pushnumber($L,$34); //@line 140 "modules/physics/box2d/Body.cpp"
 $35 = +HEAPF32[$16>>2]; //@line 141 "modules/physics/box2d/Body.cpp"
 $36 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($35)); //@line 141 "modules/physics/box2d/Body.cpp"
 $37 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($36)); //@line 141 "modules/physics/box2d/Body.cpp"
 $38 = $37; //@line 141 "modules/physics/box2d/Body.cpp"
 _lua_pushnumber($L,$38); //@line 141 "modules/physics/box2d/Body.cpp"
 STACKTOP = sp;return 4; //@line 142 "modules/physics/box2d/Body.cpp"
}
function __ZNK4love7physics5box2d4Body17getAngularDampingEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 147 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 147 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 136|0); //@line 598 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = +HEAPF32[$2>>2]; //@line 598 "./libraries/Box2D/Dynamics/b2Body.h"
 return (+$3); //@line 147 "modules/physics/box2d/Body.cpp"
}
function __ZNK4love7physics5box2d4Body16getLinearDampingEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 152 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 152 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 132|0); //@line 588 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = +HEAPF32[$2>>2]; //@line 588 "./libraries/Box2D/Dynamics/b2Body.h"
 return (+$3); //@line 152 "modules/physics/box2d/Body.cpp"
}
function __ZNK4love7physics5box2d4Body15getGravityScaleEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 157 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 157 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 140|0); //@line 608 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = +HEAPF32[$2>>2]; //@line 608 "./libraries/Box2D/Dynamics/b2Body.h"
 return (+$3); //@line 157 "modules/physics/box2d/Body.cpp"
}
function __ZNK4love7physics5box2d4Body7getTypeEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 162 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 162 "modules/physics/box2d/Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 471 "./libraries/Box2D/Dynamics/b2Body.h"
 switch ($2|0) {
 case 0:  {
  $$0 = 1;
  break;
 }
 case 2:  {
  $$0 = 2;
  break;
 }
 case 1:  {
  $$0 = 3;
  break;
 }
 default: {
  $$0 = 0;
 }
 }
 return ($$0|0); //@line 177 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body18applyLinearImpulseEffb($this,$jx,$jy,$wake) {
 $this = $this|0;
 $jx = +$jx;
 $jy = +$jy;
 $wake = $wake|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0.0;
 var $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0, $44 = 0.0;
 var $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0.0, $54 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 8|0;
 $1 = sp;
 $2 = ((($this)) + 8|0); //@line 181 "modules/physics/box2d/Body.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 181 "modules/physics/box2d/Body.cpp"
 HEAPF32[$1>>2] = $jx; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$4>>2] = $jy; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($0,$1); //@line 181 "modules/physics/box2d/Body.cpp"
 $5 = HEAP32[$2>>2]|0; //@line 181 "modules/physics/box2d/Body.cpp"
 $6 = HEAP32[$3>>2]|0; //@line 797 "./libraries/Box2D/Dynamics/b2Body.h"
 $7 = ($6|0)==(2); //@line 797 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($7)) {
  STACKTOP = sp;return; //@line 182 "modules/physics/box2d/Body.cpp"
 }
 $8 = ((($3)) + 4|0); //@line 802 "./libraries/Box2D/Dynamics/b2Body.h"
 $9 = HEAP16[$8>>1]|0; //@line 802 "./libraries/Box2D/Dynamics/b2Body.h"
 $10 = $9 & 2; //@line 802 "./libraries/Box2D/Dynamics/b2Body.h"
 $11 = ($10<<16>>16)==(0); //@line 802 "./libraries/Box2D/Dynamics/b2Body.h"
 $or$cond$i = $11 & $wake; //@line 802 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($or$cond$i) {
  $12 = $9&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $13 = $12 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $14 = ($13|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  if ($14) {
   $15 = $12 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $16 = $15&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAP16[$8>>1] = $16; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $17 = ((($3)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAPF32[$17>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   $19 = $16;
  } else {
   $19 = $9;
  }
 } else {
  $19 = $9;
 }
 $18 = $19 & 2; //@line 808 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = ($18<<16>>16)==(0); //@line 808 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($20) {
  STACKTOP = sp;return; //@line 182 "modules/physics/box2d/Body.cpp"
 }
 $21 = ((($3)) + 120|0); //@line 810 "./libraries/Box2D/Dynamics/b2Body.h"
 $22 = +HEAPF32[$21>>2]; //@line 810 "./libraries/Box2D/Dynamics/b2Body.h"
 $23 = +HEAPF32[$0>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $24 = $22 * $23; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $25 = ((($0)) + 4|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $26 = +HEAPF32[$25>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $27 = $22 * $26; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $28 = ((($3)) + 64|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $29 = +HEAPF32[$28>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $30 = $24 + $29; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$28>>2] = $30; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $31 = ((($3)) + 68|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $32 = +HEAPF32[$31>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $33 = $27 + $32; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$31>>2] = $33; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $34 = ((($3)) + 128|0); //@line 811 "./libraries/Box2D/Dynamics/b2Body.h"
 $35 = +HEAPF32[$34>>2]; //@line 811 "./libraries/Box2D/Dynamics/b2Body.h"
 $36 = ((($5)) + 44|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $37 = +HEAPF32[$36>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $38 = ((($3)) + 44|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $39 = +HEAPF32[$38>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $40 = $37 - $39; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $41 = ((($5)) + 48|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $42 = +HEAPF32[$41>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $43 = ((($3)) + 48|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $44 = +HEAPF32[$43>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $45 = $42 - $44; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $46 = +HEAPF32[$25>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $47 = $40 * $46; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $48 = +HEAPF32[$0>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $49 = $45 * $48; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $50 = $47 - $49; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $51 = $35 * $50; //@line 811 "./libraries/Box2D/Dynamics/b2Body.h"
 $52 = ((($3)) + 72|0); //@line 811 "./libraries/Box2D/Dynamics/b2Body.h"
 $53 = +HEAPF32[$52>>2]; //@line 811 "./libraries/Box2D/Dynamics/b2Body.h"
 $54 = $53 + $51; //@line 811 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$52>>2] = $54; //@line 811 "./libraries/Box2D/Dynamics/b2Body.h"
 STACKTOP = sp;return; //@line 182 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body18applyLinearImpulseEffffb($this,$jx,$jy,$rx,$ry,$wake) {
 $this = $this|0;
 $jx = +$jx;
 $jy = +$jy;
 $rx = +$rx;
 $ry = +$ry;
 $wake = $wake|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0.0;
 var $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0.0, $44 = 0;
 var $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 24|0;
 $1 = sp + 16|0;
 $2 = sp + 8|0;
 $3 = sp;
 $4 = ((($this)) + 8|0); //@line 186 "modules/physics/box2d/Body.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 186 "modules/physics/box2d/Body.cpp"
 HEAPF32[$1>>2] = $jx; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$6>>2] = $jy; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($0,$1); //@line 186 "modules/physics/box2d/Body.cpp"
 HEAPF32[$3>>2] = $rx; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($3)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$7>>2] = $ry; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($2,$3); //@line 186 "modules/physics/box2d/Body.cpp"
 $8 = HEAP32[$5>>2]|0; //@line 797 "./libraries/Box2D/Dynamics/b2Body.h"
 $9 = ($8|0)==(2); //@line 797 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($9)) {
  STACKTOP = sp;return; //@line 187 "modules/physics/box2d/Body.cpp"
 }
 $10 = ((($5)) + 4|0); //@line 802 "./libraries/Box2D/Dynamics/b2Body.h"
 $11 = HEAP16[$10>>1]|0; //@line 802 "./libraries/Box2D/Dynamics/b2Body.h"
 $12 = $11 & 2; //@line 802 "./libraries/Box2D/Dynamics/b2Body.h"
 $13 = ($12<<16>>16)==(0); //@line 802 "./libraries/Box2D/Dynamics/b2Body.h"
 $or$cond$i = $13 & $wake; //@line 802 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($or$cond$i) {
  $14 = $11&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $15 = $14 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $16 = ($15|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  if ($16) {
   $17 = $14 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $18 = $17&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAP16[$10>>1] = $18; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $19 = ((($5)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAPF32[$19>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   $21 = $18;
  } else {
   $21 = $11;
  }
 } else {
  $21 = $11;
 }
 $20 = $21 & 2; //@line 808 "./libraries/Box2D/Dynamics/b2Body.h"
 $22 = ($20<<16>>16)==(0); //@line 808 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($22) {
  STACKTOP = sp;return; //@line 187 "modules/physics/box2d/Body.cpp"
 }
 $23 = ((($5)) + 120|0); //@line 810 "./libraries/Box2D/Dynamics/b2Body.h"
 $24 = +HEAPF32[$23>>2]; //@line 810 "./libraries/Box2D/Dynamics/b2Body.h"
 $25 = +HEAPF32[$0>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $26 = $24 * $25; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($0)) + 4|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $28 = +HEAPF32[$27>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $29 = $24 * $28; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $30 = ((($5)) + 64|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $31 = +HEAPF32[$30>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $32 = $26 + $31; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$30>>2] = $32; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $33 = ((($5)) + 68|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $34 = +HEAPF32[$33>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $35 = $29 + $34; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$33>>2] = $35; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $36 = ((($5)) + 128|0); //@line 811 "./libraries/Box2D/Dynamics/b2Body.h"
 $37 = +HEAPF32[$36>>2]; //@line 811 "./libraries/Box2D/Dynamics/b2Body.h"
 $38 = +HEAPF32[$2>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $39 = ((($5)) + 44|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $40 = +HEAPF32[$39>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $41 = $38 - $40; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $42 = ((($2)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $43 = +HEAPF32[$42>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $44 = ((($5)) + 48|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $45 = +HEAPF32[$44>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $46 = $43 - $45; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $47 = +HEAPF32[$27>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $48 = $41 * $47; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $49 = +HEAPF32[$0>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $50 = $46 * $49; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $51 = $48 - $50; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $52 = $37 * $51; //@line 811 "./libraries/Box2D/Dynamics/b2Body.h"
 $53 = ((($5)) + 72|0); //@line 811 "./libraries/Box2D/Dynamics/b2Body.h"
 $54 = +HEAPF32[$53>>2]; //@line 811 "./libraries/Box2D/Dynamics/b2Body.h"
 $55 = $54 + $52; //@line 811 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$53>>2] = $55; //@line 811 "./libraries/Box2D/Dynamics/b2Body.h"
 STACKTOP = sp;return; //@line 187 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body19applyAngularImpulseEfb($this,$impulse,$wake) {
 $this = $this|0;
 $impulse = +$impulse;
 $wake = $wake|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0.0, $3 = 0.0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 192 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 192 "modules/physics/box2d/Body.cpp"
 $2 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($impulse)); //@line 192 "modules/physics/box2d/Body.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($2)); //@line 192 "modules/physics/box2d/Body.cpp"
 $4 = HEAP32[$1>>2]|0; //@line 817 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = ($4|0)==(2); //@line 817 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($5)) {
  return; //@line 193 "modules/physics/box2d/Body.cpp"
 }
 $6 = ((($1)) + 4|0); //@line 822 "./libraries/Box2D/Dynamics/b2Body.h"
 $7 = HEAP16[$6>>1]|0; //@line 822 "./libraries/Box2D/Dynamics/b2Body.h"
 $8 = $7 & 2; //@line 822 "./libraries/Box2D/Dynamics/b2Body.h"
 $9 = ($8<<16>>16)==(0); //@line 822 "./libraries/Box2D/Dynamics/b2Body.h"
 $or$cond$i = $9 & $wake; //@line 822 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($or$cond$i) {
  $10 = $7&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $11 = $10 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $12 = ($11|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  if ($12) {
   $13 = $10 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $14 = $13&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAP16[$6>>1] = $14; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $15 = ((($1)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAPF32[$15>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   $17 = $14;
  } else {
   $17 = $7;
  }
 } else {
  $17 = $7;
 }
 $16 = $17 & 2; //@line 828 "./libraries/Box2D/Dynamics/b2Body.h"
 $18 = ($16<<16>>16)==(0); //@line 828 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($18) {
  return; //@line 193 "modules/physics/box2d/Body.cpp"
 }
 $19 = ((($1)) + 128|0); //@line 830 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = +HEAPF32[$19>>2]; //@line 830 "./libraries/Box2D/Dynamics/b2Body.h"
 $21 = $3 * $20; //@line 830 "./libraries/Box2D/Dynamics/b2Body.h"
 $22 = ((($1)) + 72|0); //@line 830 "./libraries/Box2D/Dynamics/b2Body.h"
 $23 = +HEAPF32[$22>>2]; //@line 830 "./libraries/Box2D/Dynamics/b2Body.h"
 $24 = $23 + $21; //@line 830 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$22>>2] = $24; //@line 830 "./libraries/Box2D/Dynamics/b2Body.h"
 return; //@line 193 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body11applyTorqueEfb($this,$t,$wake) {
 $this = $this|0;
 $t = +$t;
 $wake = $wake|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0.0, $21 = 0.0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 198 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 198 "modules/physics/box2d/Body.cpp"
 $2 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($t)); //@line 198 "modules/physics/box2d/Body.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($2)); //@line 198 "modules/physics/box2d/Body.cpp"
 $4 = HEAP32[$1>>2]|0; //@line 778 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = ($4|0)==(2); //@line 778 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($5)) {
  return; //@line 199 "modules/physics/box2d/Body.cpp"
 }
 $6 = ((($1)) + 4|0); //@line 783 "./libraries/Box2D/Dynamics/b2Body.h"
 $7 = HEAP16[$6>>1]|0; //@line 783 "./libraries/Box2D/Dynamics/b2Body.h"
 $8 = $7 & 2; //@line 783 "./libraries/Box2D/Dynamics/b2Body.h"
 $9 = ($8<<16>>16)==(0); //@line 783 "./libraries/Box2D/Dynamics/b2Body.h"
 $or$cond$i = $9 & $wake; //@line 783 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($or$cond$i) {
  $10 = $7&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $11 = $10 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $12 = ($11|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  if ($12) {
   $13 = $10 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $14 = $13&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAP16[$6>>1] = $14; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $15 = ((($1)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAPF32[$15>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   $17 = $14;
  } else {
   $17 = $7;
  }
 } else {
  $17 = $7;
 }
 $16 = $17 & 2; //@line 789 "./libraries/Box2D/Dynamics/b2Body.h"
 $18 = ($16<<16>>16)==(0); //@line 789 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($18) {
  return; //@line 199 "modules/physics/box2d/Body.cpp"
 }
 $19 = ((($1)) + 84|0); //@line 791 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = +HEAPF32[$19>>2]; //@line 791 "./libraries/Box2D/Dynamics/b2Body.h"
 $21 = $3 + $20; //@line 791 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$19>>2] = $21; //@line 791 "./libraries/Box2D/Dynamics/b2Body.h"
 return; //@line 199 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body10applyForceEffffb($this,$fx,$fy,$rx,$ry,$wake) {
 $this = $this|0;
 $fx = +$fx;
 $fy = +$fy;
 $rx = +$rx;
 $ry = +$ry;
 $wake = $wake|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0;
 var $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0;
 var $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 24|0;
 $1 = sp + 16|0;
 $2 = sp + 8|0;
 $3 = sp;
 $4 = ((($this)) + 8|0); //@line 203 "modules/physics/box2d/Body.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 203 "modules/physics/box2d/Body.cpp"
 HEAPF32[$1>>2] = $fx; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$6>>2] = $fy; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($0,$1); //@line 203 "modules/physics/box2d/Body.cpp"
 HEAPF32[$3>>2] = $rx; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($3)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$7>>2] = $ry; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($2,$3); //@line 203 "modules/physics/box2d/Body.cpp"
 $8 = HEAP32[$5>>2]|0; //@line 739 "./libraries/Box2D/Dynamics/b2Body.h"
 $9 = ($8|0)==(2); //@line 739 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($9)) {
  STACKTOP = sp;return; //@line 204 "modules/physics/box2d/Body.cpp"
 }
 $10 = ((($5)) + 4|0); //@line 744 "./libraries/Box2D/Dynamics/b2Body.h"
 $11 = HEAP16[$10>>1]|0; //@line 744 "./libraries/Box2D/Dynamics/b2Body.h"
 $12 = $11 & 2; //@line 744 "./libraries/Box2D/Dynamics/b2Body.h"
 $13 = ($12<<16>>16)==(0); //@line 744 "./libraries/Box2D/Dynamics/b2Body.h"
 $or$cond$i = $13 & $wake; //@line 744 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($or$cond$i) {
  $14 = $11&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $15 = $14 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $16 = ($15|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  if ($16) {
   $17 = $14 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $18 = $17&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAP16[$10>>1] = $18; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $19 = ((($5)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAPF32[$19>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   $21 = $18;
  } else {
   $21 = $11;
  }
 } else {
  $21 = $11;
 }
 $20 = $21 & 2; //@line 750 "./libraries/Box2D/Dynamics/b2Body.h"
 $22 = ($20<<16>>16)==(0); //@line 750 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($22) {
  STACKTOP = sp;return; //@line 204 "modules/physics/box2d/Body.cpp"
 }
 $23 = +HEAPF32[$0>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($5)) + 76|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $26 = $23 + $25; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$24>>2] = $26; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($0)) + 4|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $28 = +HEAPF32[$27>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $29 = ((($5)) + 80|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $30 = +HEAPF32[$29>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $31 = $28 + $30; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$29>>2] = $31; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $32 = +HEAPF32[$2>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $33 = ((($5)) + 44|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $34 = +HEAPF32[$33>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $35 = $32 - $34; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $36 = ((($2)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $37 = +HEAPF32[$36>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $38 = ((($5)) + 48|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $39 = +HEAPF32[$38>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $40 = $37 - $39; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $41 = +HEAPF32[$27>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $42 = $35 * $41; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $43 = +HEAPF32[$0>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $44 = $40 * $43; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $45 = $42 - $44; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $46 = ((($5)) + 84|0); //@line 753 "./libraries/Box2D/Dynamics/b2Body.h"
 $47 = +HEAPF32[$46>>2]; //@line 753 "./libraries/Box2D/Dynamics/b2Body.h"
 $48 = $47 + $45; //@line 753 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$46>>2] = $48; //@line 753 "./libraries/Box2D/Dynamics/b2Body.h"
 STACKTOP = sp;return; //@line 204 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body10applyForceEffb($this,$fx,$fy,$wake) {
 $this = $this|0;
 $fx = +$fx;
 $fy = +$fy;
 $wake = $wake|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 8|0;
 $1 = sp;
 $2 = ((($this)) + 8|0); //@line 208 "modules/physics/box2d/Body.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 208 "modules/physics/box2d/Body.cpp"
 HEAPF32[$1>>2] = $fx; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$4>>2] = $fy; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($0,$1); //@line 208 "modules/physics/box2d/Body.cpp"
 $5 = HEAP32[$3>>2]|0; //@line 759 "./libraries/Box2D/Dynamics/b2Body.h"
 $6 = ($5|0)==(2); //@line 759 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($6)) {
  STACKTOP = sp;return; //@line 209 "modules/physics/box2d/Body.cpp"
 }
 $7 = ((($3)) + 4|0); //@line 764 "./libraries/Box2D/Dynamics/b2Body.h"
 $8 = HEAP16[$7>>1]|0; //@line 764 "./libraries/Box2D/Dynamics/b2Body.h"
 $9 = $8 & 2; //@line 764 "./libraries/Box2D/Dynamics/b2Body.h"
 $10 = ($9<<16>>16)==(0); //@line 764 "./libraries/Box2D/Dynamics/b2Body.h"
 $or$cond$i = $10 & $wake; //@line 764 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($or$cond$i) {
  $11 = $8&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $12 = $11 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $13 = ($12|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  if ($13) {
   $14 = $11 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $15 = $14&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAP16[$7>>1] = $15; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $16 = ((($3)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAPF32[$16>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   $18 = $15;
  } else {
   $18 = $8;
  }
 } else {
  $18 = $8;
 }
 $17 = $18 & 2; //@line 770 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = ($17<<16>>16)==(0); //@line 770 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($19) {
  STACKTOP = sp;return; //@line 209 "modules/physics/box2d/Body.cpp"
 }
 $20 = +HEAPF32[$0>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($3)) + 76|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $22 = +HEAPF32[$21>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $23 = $20 + $22; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$21>>2] = $23; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($0)) + 4|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $26 = ((($3)) + 80|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $27 = +HEAPF32[$26>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $28 = $25 + $27; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$26>>2] = $28; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 STACKTOP = sp;return; //@line 209 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body4setXEf($this,$x) {
 $this = $this|0;
 $x = +$x;
 var $0 = 0, $1 = 0, $10 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 8|0;
 $1 = sp;
 $2 = ((($this)) + 8|0); //@line 213 "modules/physics/box2d/Body.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 213 "modules/physics/box2d/Body.cpp"
 $4 = ((($3)) + 16|0); //@line 82 "modules/physics/box2d/Body.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 82 "modules/physics/box2d/Body.cpp"
 $6 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($5)); //@line 82 "modules/physics/box2d/Body.cpp"
 HEAPF32[$1>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$7>>2] = $6; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($0,$1); //@line 213 "modules/physics/box2d/Body.cpp"
 $8 = HEAP32[$2>>2]|0; //@line 101 "modules/physics/box2d/Body.cpp"
 $9 = ((($8)) + 56|0); //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $10 = +HEAPF32[$9>>2]; //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 __ZN6b2Body12SetTransformERK6b2Vec2f($3,$0,$10); //@line 213 "modules/physics/box2d/Body.cpp"
 STACKTOP = sp;return; //@line 214 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body4setYEf($this,$y) {
 $this = $this|0;
 $y = +$y;
 var $0 = 0, $1 = 0, $10 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 8|0;
 $1 = sp;
 $2 = ((($this)) + 8|0); //@line 218 "modules/physics/box2d/Body.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 218 "modules/physics/box2d/Body.cpp"
 $4 = ((($3)) + 12|0); //@line 77 "modules/physics/box2d/Body.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 77 "modules/physics/box2d/Body.cpp"
 $6 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($5)); //@line 77 "modules/physics/box2d/Body.cpp"
 HEAPF32[$1>>2] = $6; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$7>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($0,$1); //@line 218 "modules/physics/box2d/Body.cpp"
 $8 = HEAP32[$2>>2]|0; //@line 101 "modules/physics/box2d/Body.cpp"
 $9 = ((($8)) + 56|0); //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $10 = +HEAPF32[$9>>2]; //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 __ZN6b2Body12SetTransformERK6b2Vec2f($3,$0,$10); //@line 218 "modules/physics/box2d/Body.cpp"
 STACKTOP = sp;return; //@line 219 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body17setLinearVelocityEff($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = sp + 8|0;
 $2 = ((($this)) + 8|0); //@line 223 "modules/physics/box2d/Body.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 223 "modules/physics/box2d/Body.cpp"
 HEAPF32[$1>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$4>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($0,$1); //@line 223 "modules/physics/box2d/Body.cpp"
 $5 = HEAP32[$3>>2]|0; //@line 501 "./libraries/Box2D/Dynamics/b2Body.h"
 $6 = ($5|0)==(0); //@line 501 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($6) {
  STACKTOP = sp;return; //@line 224 "modules/physics/box2d/Body.cpp"
 }
 $7 = +HEAPF32[$0>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $8 = $7 * $7; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($0)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $11 = $10 * $10; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $12 = $8 + $11; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $13 = $12 > 0.0; //@line 506 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($13) {
  $14 = ((($3)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $15 = HEAP16[$14>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $16 = $15&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $17 = $16 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $18 = ($17|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  if ($18) {
   $19 = $16 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $20 = $19&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAP16[$14>>1] = $20; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $21 = ((($3)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAPF32[$21>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  }
 }
 $22 = ((($3)) + 64|0); //@line 511 "./libraries/Box2D/Dynamics/b2Body.h"
 $23 = $0; //@line 511 "./libraries/Box2D/Dynamics/b2Body.h"
 $24 = $23; //@line 511 "./libraries/Box2D/Dynamics/b2Body.h"
 $25 = HEAP32[$24>>2]|0; //@line 511 "./libraries/Box2D/Dynamics/b2Body.h"
 $26 = (($23) + 4)|0; //@line 511 "./libraries/Box2D/Dynamics/b2Body.h"
 $27 = $26; //@line 511 "./libraries/Box2D/Dynamics/b2Body.h"
 $28 = HEAP32[$27>>2]|0; //@line 511 "./libraries/Box2D/Dynamics/b2Body.h"
 $29 = $22; //@line 511 "./libraries/Box2D/Dynamics/b2Body.h"
 $30 = $29; //@line 511 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP32[$30>>2] = $25; //@line 511 "./libraries/Box2D/Dynamics/b2Body.h"
 $31 = (($29) + 4)|0; //@line 511 "./libraries/Box2D/Dynamics/b2Body.h"
 $32 = $31; //@line 511 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP32[$32>>2] = $28; //@line 511 "./libraries/Box2D/Dynamics/b2Body.h"
 STACKTOP = sp;return; //@line 224 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body8setAngleEf($this,$d) {
 $this = $this|0;
 $d = +$d;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 228 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 228 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 12|0); //@line 481 "./libraries/Box2D/Dynamics/b2Body.h"
 __ZN6b2Body12SetTransformERK6b2Vec2f($1,$2,$d); //@line 228 "modules/physics/box2d/Body.cpp"
 return; //@line 229 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body18setAngularVelocityEf($this,$r) {
 $this = $this|0;
 $r = +$r;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 233 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 233 "modules/physics/box2d/Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 521 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = ($2|0)==(0); //@line 521 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($3) {
  return; //@line 234 "modules/physics/box2d/Body.cpp"
 }
 $4 = $r * $r; //@line 526 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = $4 > 0.0; //@line 526 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($5) {
  $6 = ((($1)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $7 = HEAP16[$6>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $8 = $7&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $9 = $8 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  $10 = ($9|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
  if ($10) {
   $11 = $8 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $12 = $11&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAP16[$6>>1] = $12; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
   $13 = ((($1)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAPF32[$13>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  }
 }
 $14 = ((($1)) + 72|0); //@line 531 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$14>>2] = $r; //@line 531 "./libraries/Box2D/Dynamics/b2Body.h"
 return; //@line 234 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body11setPositionEff($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 8|0;
 $1 = sp;
 $2 = ((($this)) + 8|0); //@line 238 "modules/physics/box2d/Body.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 238 "modules/physics/box2d/Body.cpp"
 HEAPF32[$1>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$4>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($0,$1); //@line 238 "modules/physics/box2d/Body.cpp"
 $5 = HEAP32[$2>>2]|0; //@line 238 "modules/physics/box2d/Body.cpp"
 $6 = ((($5)) + 56|0); //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 $7 = +HEAPF32[$6>>2]; //@line 486 "./libraries/Box2D/Dynamics/b2Body.h"
 __ZN6b2Body12SetTransformERK6b2Vec2f($3,$0,$7); //@line 238 "modules/physics/box2d/Body.cpp"
 STACKTOP = sp;return; //@line 239 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body17setAngularDampingEf($this,$d) {
 $this = $this|0;
 $d = +$d;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 243 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 243 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 136|0); //@line 603 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$2>>2] = $d; //@line 603 "./libraries/Box2D/Dynamics/b2Body.h"
 return; //@line 244 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body16setLinearDampingEf($this,$d) {
 $this = $this|0;
 $d = +$d;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 248 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 248 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 132|0); //@line 593 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$2>>2] = $d; //@line 593 "./libraries/Box2D/Dynamics/b2Body.h"
 return; //@line 249 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body13resetMassDataEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 253 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 253 "modules/physics/box2d/Body.cpp"
 __ZN6b2Body13ResetMassDataEv($1); //@line 253 "modules/physics/box2d/Body.cpp"
 return; //@line 254 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body11setMassDataEffff($this,$x,$y,$m,$i) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $m = +$m;
 $i = +$i;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $massData = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $massData = sp + 16|0;
 $0 = sp;
 $1 = sp + 8|0;
 $2 = ((($massData)) + 4|0); //@line 259 "modules/physics/box2d/Body.cpp"
 HEAPF32[$1>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$3>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($0,$1); //@line 259 "modules/physics/box2d/Body.cpp"
 $4 = $0; //@line 259 "modules/physics/box2d/Body.cpp"
 $5 = $4; //@line 259 "modules/physics/box2d/Body.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 259 "modules/physics/box2d/Body.cpp"
 $7 = (($4) + 4)|0; //@line 259 "modules/physics/box2d/Body.cpp"
 $8 = $7; //@line 259 "modules/physics/box2d/Body.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 259 "modules/physics/box2d/Body.cpp"
 $10 = $2; //@line 259 "modules/physics/box2d/Body.cpp"
 $11 = $10; //@line 259 "modules/physics/box2d/Body.cpp"
 HEAP32[$11>>2] = $6; //@line 259 "modules/physics/box2d/Body.cpp"
 $12 = (($10) + 4)|0; //@line 259 "modules/physics/box2d/Body.cpp"
 $13 = $12; //@line 259 "modules/physics/box2d/Body.cpp"
 HEAP32[$13>>2] = $9; //@line 259 "modules/physics/box2d/Body.cpp"
 HEAPF32[$massData>>2] = $m; //@line 260 "modules/physics/box2d/Body.cpp"
 $14 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($i)); //@line 261 "modules/physics/box2d/Body.cpp"
 $15 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($14)); //@line 261 "modules/physics/box2d/Body.cpp"
 $16 = ((($massData)) + 12|0); //@line 261 "modules/physics/box2d/Body.cpp"
 HEAPF32[$16>>2] = $15; //@line 261 "modules/physics/box2d/Body.cpp"
 $17 = ((($this)) + 8|0); //@line 262 "modules/physics/box2d/Body.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 262 "modules/physics/box2d/Body.cpp"
 __ZN6b2Body11SetMassDataEPK10b2MassData($18,$massData); //@line 262 "modules/physics/box2d/Body.cpp"
 STACKTOP = sp;return; //@line 263 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body7setMassEf($this,$m) {
 $this = $this|0;
 $m = +$m;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0.0, $9 = 0.0, $data = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $data = sp;
 $0 = ((($this)) + 8|0); //@line 268 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 268 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 116|0); //@line 551 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP32[$2>>2]|0; //@line 551 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP32[$data>>2] = $3; //@line 551 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = ((($1)) + 124|0); //@line 552 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = +HEAPF32[$4>>2]; //@line 552 "./libraries/Box2D/Dynamics/b2Body.h"
 $6 = (HEAP32[tempDoublePtr>>2]=$3,+HEAPF32[tempDoublePtr>>2]); //@line 552 "./libraries/Box2D/Dynamics/b2Body.h"
 $7 = ((($1)) + 28|0); //@line 552 "./libraries/Box2D/Dynamics/b2Body.h"
 $8 = +HEAPF32[$7>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $9 = $8 * $8; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($1)) + 32|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $11 = +HEAPF32[$10>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $12 = $11 * $11; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $13 = $9 + $12; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $14 = $6 * $13; //@line 552 "./libraries/Box2D/Dynamics/b2Body.h"
 $15 = $5 + $14; //@line 552 "./libraries/Box2D/Dynamics/b2Body.h"
 $16 = ((($data)) + 12|0); //@line 552 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$16>>2] = $15; //@line 552 "./libraries/Box2D/Dynamics/b2Body.h"
 $17 = ((($data)) + 4|0); //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $18 = $7; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = $18; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = HEAP32[$19>>2]|0; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $21 = (($18) + 4)|0; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $22 = $21; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $23 = HEAP32[$22>>2]|0; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $24 = $17; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $25 = $24; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP32[$25>>2] = $20; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $26 = (($24) + 4)|0; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 $27 = $26; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP32[$27>>2] = $23; //@line 553 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$data>>2] = $m; //@line 269 "modules/physics/box2d/Body.cpp"
 __ZN6b2Body11SetMassDataEPK10b2MassData($1,$data); //@line 270 "modules/physics/box2d/Body.cpp"
 STACKTOP = sp;return; //@line 271 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body10setInertiaEf($this,$i) {
 $this = $this|0;
 $i = +$i;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $massData = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $massData = sp;
 $0 = ((($massData)) + 4|0); //@line 276 "modules/physics/box2d/Body.cpp"
 $1 = ((($this)) + 8|0); //@line 276 "modules/physics/box2d/Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 276 "modules/physics/box2d/Body.cpp"
 $3 = ((($2)) + 28|0); //@line 496 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3; //@line 276 "modules/physics/box2d/Body.cpp"
 $5 = $4; //@line 276 "modules/physics/box2d/Body.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 276 "modules/physics/box2d/Body.cpp"
 $7 = (($4) + 4)|0; //@line 276 "modules/physics/box2d/Body.cpp"
 $8 = $7; //@line 276 "modules/physics/box2d/Body.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 276 "modules/physics/box2d/Body.cpp"
 $10 = $0; //@line 276 "modules/physics/box2d/Body.cpp"
 $11 = $10; //@line 276 "modules/physics/box2d/Body.cpp"
 HEAP32[$11>>2] = $6; //@line 276 "modules/physics/box2d/Body.cpp"
 $12 = (($10) + 4)|0; //@line 276 "modules/physics/box2d/Body.cpp"
 $13 = $12; //@line 276 "modules/physics/box2d/Body.cpp"
 HEAP32[$13>>2] = $9; //@line 276 "modules/physics/box2d/Body.cpp"
 $14 = ((($2)) + 116|0); //@line 541 "./libraries/Box2D/Dynamics/b2Body.h"
 $15 = HEAP32[$14>>2]|0; //@line 541 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP32[$massData>>2] = $15; //@line 277 "modules/physics/box2d/Body.cpp"
 $16 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($i)); //@line 278 "modules/physics/box2d/Body.cpp"
 $17 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($16)); //@line 278 "modules/physics/box2d/Body.cpp"
 $18 = ((($massData)) + 12|0); //@line 278 "modules/physics/box2d/Body.cpp"
 HEAPF32[$18>>2] = $17; //@line 278 "modules/physics/box2d/Body.cpp"
 $19 = HEAP32[$1>>2]|0; //@line 279 "modules/physics/box2d/Body.cpp"
 __ZN6b2Body11SetMassDataEPK10b2MassData($19,$massData); //@line 279 "modules/physics/box2d/Body.cpp"
 STACKTOP = sp;return; //@line 280 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body15setGravityScaleEf($this,$scale) {
 $this = $this|0;
 $scale = +$scale;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 284 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 284 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 140|0); //@line 613 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$2>>2] = $scale; //@line 613 "./libraries/Box2D/Dynamics/b2Body.h"
 return; //@line 285 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body13getWorldPointEffRfS3_($this,$x,$y,$x_o,$y_o) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $x_o = $x_o|0;
 $y_o = $y_o|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp + 24|0;
 $0 = sp + 16|0;
 $1 = sp + 8|0;
 $2 = sp;
 $3 = ((($this)) + 8|0); //@line 307 "modules/physics/box2d/Body.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 307 "modules/physics/box2d/Body.cpp"
 HEAPF32[$2>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($2)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$5>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($1,$2); //@line 307 "modules/physics/box2d/Body.cpp"
 $6 = ((($4)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$1>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = $7 * $8; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($4)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = +HEAPF32[$10>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = ((($1)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = +HEAPF32[$12>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = $11 * $13; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $9 - $14; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = ((($4)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $17 = +HEAPF32[$16>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $18 = $17 + $15; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $19 = $8 * $11; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = $7 * $13; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $19 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($4)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $23 = +HEAPF32[$22>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $24 = $21 + $23; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$0>>2] = $18; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $25 = ((($0)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$25>>2] = $24; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($v,$0); //@line 307 "modules/physics/box2d/Body.cpp"
 $26 = HEAP32[$v>>2]|0; //@line 308 "modules/physics/box2d/Body.cpp"
 HEAP32[$x_o>>2] = $26; //@line 308 "modules/physics/box2d/Body.cpp"
 $27 = ((($v)) + 4|0); //@line 309 "modules/physics/box2d/Body.cpp"
 $28 = HEAP32[$27>>2]|0; //@line 309 "modules/physics/box2d/Body.cpp"
 HEAP32[$y_o>>2] = $28; //@line 309 "modules/physics/box2d/Body.cpp"
 STACKTOP = sp;return; //@line 310 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body14getWorldVectorEffRfS3_($this,$x,$y,$x_o,$y_o) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $x_o = $x_o|0;
 $y_o = $y_o|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0.0, $8 = 0.0, $9 = 0.0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp + 24|0;
 $0 = sp + 16|0;
 $1 = sp + 8|0;
 $2 = sp;
 $3 = ((($this)) + 8|0); //@line 314 "modules/physics/box2d/Body.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 314 "modules/physics/box2d/Body.cpp"
 HEAPF32[$2>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($2)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$5>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($1,$2); //@line 314 "modules/physics/box2d/Body.cpp"
 $6 = ((($4)) + 24|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$1>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $9 = $7 * $8; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($4)) + 20|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $11 = +HEAPF32[$10>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $12 = ((($1)) + 4|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $13 = +HEAPF32[$12>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $14 = $11 * $13; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $15 = $9 - $14; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $16 = $8 * $11; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $17 = $7 * $13; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$0>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($0)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$19>>2] = $18; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($v,$0); //@line 314 "modules/physics/box2d/Body.cpp"
 $20 = HEAP32[$v>>2]|0; //@line 315 "modules/physics/box2d/Body.cpp"
 HEAP32[$x_o>>2] = $20; //@line 315 "modules/physics/box2d/Body.cpp"
 $21 = ((($v)) + 4|0); //@line 316 "modules/physics/box2d/Body.cpp"
 $22 = HEAP32[$21>>2]|0; //@line 316 "modules/physics/box2d/Body.cpp"
 HEAP32[$y_o>>2] = $22; //@line 316 "modules/physics/box2d/Body.cpp"
 STACKTOP = sp;return; //@line 317 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body14getWorldPointsEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0.0;
 var $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $i$01 = 0, $point = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $point = sp + 24|0;
 $0 = sp + 16|0;
 $1 = sp + 8|0;
 $2 = sp;
 $3 = (_lua_gettop($L)|0); //@line 321 "modules/physics/box2d/Body.cpp"
 $4 = (($3|0) / 2)&-1; //@line 322 "modules/physics/box2d/Body.cpp"
 (__ZN4love16luax_assert_argcEP9lua_Statei($L,2)|0); //@line 324 "modules/physics/box2d/Body.cpp"
 $5 = ($3|0)>(1); //@line 326 "modules/physics/box2d/Body.cpp"
 if (!($5)) {
  STACKTOP = sp;return ($3|0); //@line 340 "modules/physics/box2d/Body.cpp"
 }
 $6 = ((($this)) + 8|0); //@line 334 "modules/physics/box2d/Body.cpp"
 $7 = ((($2)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($1)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($0)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($point)) + 4|0); //@line 337 "modules/physics/box2d/Body.cpp"
 $i$01 = 0;
 while(1) {
  $11 = (+_lua_tonumber($L,1)); //@line 328 "modules/physics/box2d/Body.cpp"
  $12 = $11; //@line 328 "modules/physics/box2d/Body.cpp"
  $13 = (+_lua_tonumber($L,2)); //@line 329 "modules/physics/box2d/Body.cpp"
  $14 = $13; //@line 329 "modules/physics/box2d/Body.cpp"
  _lua_remove($L,1); //@line 331 "modules/physics/box2d/Body.cpp"
  _lua_remove($L,1); //@line 332 "modules/physics/box2d/Body.cpp"
  $15 = HEAP32[$6>>2]|0; //@line 334 "modules/physics/box2d/Body.cpp"
  HEAPF32[$2>>2] = $12; //@line 59 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$7>>2] = $14; //@line 59 "./libraries/Box2D/Common/b2Math.h"
  __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($1,$2); //@line 334 "modules/physics/box2d/Body.cpp"
  $16 = ((($15)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $17 = +HEAPF32[$16>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $18 = +HEAPF32[$1>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $19 = $17 * $18; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $20 = ((($15)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $21 = +HEAPF32[$20>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $22 = +HEAPF32[$8>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $23 = $21 * $22; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $24 = $19 - $23; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $25 = ((($15)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $26 = +HEAPF32[$25>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $27 = $26 + $24; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $28 = $18 * $21; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $29 = $17 * $22; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $30 = $28 + $29; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $31 = ((($15)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $32 = +HEAPF32[$31>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $33 = $30 + $32; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$0>>2] = $27; //@line 59 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$9>>2] = $33; //@line 59 "./libraries/Box2D/Common/b2Math.h"
  __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($point,$0); //@line 334 "modules/physics/box2d/Body.cpp"
  $34 = +HEAPF32[$point>>2]; //@line 336 "modules/physics/box2d/Body.cpp"
  $35 = $34; //@line 336 "modules/physics/box2d/Body.cpp"
  _lua_pushnumber($L,$35); //@line 336 "modules/physics/box2d/Body.cpp"
  $36 = +HEAPF32[$10>>2]; //@line 337 "modules/physics/box2d/Body.cpp"
  $37 = $36; //@line 337 "modules/physics/box2d/Body.cpp"
  _lua_pushnumber($L,$37); //@line 337 "modules/physics/box2d/Body.cpp"
  $38 = (($i$01) + 1)|0; //@line 326 "modules/physics/box2d/Body.cpp"
  $39 = ($38|0)<($4|0); //@line 326 "modules/physics/box2d/Body.cpp"
  if ($39) {
   $i$01 = $38;
  } else {
   break;
  }
 }
 STACKTOP = sp;return ($3|0); //@line 340 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body13getLocalPointEffRfS3_($this,$x,$y,$x_o,$y_o) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $x_o = $x_o|0;
 $y_o = $y_o|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0.0, $9 = 0.0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp + 24|0;
 $0 = sp + 16|0;
 $1 = sp + 8|0;
 $2 = sp;
 $3 = ((($this)) + 8|0); //@line 345 "modules/physics/box2d/Body.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 345 "modules/physics/box2d/Body.cpp"
 HEAPF32[$2>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($2)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$5>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($1,$2); //@line 345 "modules/physics/box2d/Body.cpp"
 $6 = +HEAPF32[$1>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($4)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $9 = $6 - $8; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($1)) + 4|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $11 = +HEAPF32[$10>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $12 = ((($4)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $13 = +HEAPF32[$12>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $14 = $11 - $13; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $15 = ((($4)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $16 = +HEAPF32[$15>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $17 = $9 * $16; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $18 = ((($4)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $19 = +HEAPF32[$18>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $20 = $14 * $19; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $21 = $17 + $20; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $22 = $9 * $19; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $23 = $16 * $14; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $24 = $23 - $22; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$0>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $25 = ((($0)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$25>>2] = $24; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($v,$0); //@line 345 "modules/physics/box2d/Body.cpp"
 $26 = HEAP32[$v>>2]|0; //@line 346 "modules/physics/box2d/Body.cpp"
 HEAP32[$x_o>>2] = $26; //@line 346 "modules/physics/box2d/Body.cpp"
 $27 = ((($v)) + 4|0); //@line 347 "modules/physics/box2d/Body.cpp"
 $28 = HEAP32[$27>>2]|0; //@line 347 "modules/physics/box2d/Body.cpp"
 HEAP32[$y_o>>2] = $28; //@line 347 "modules/physics/box2d/Body.cpp"
 STACKTOP = sp;return; //@line 348 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body14getLocalVectorEffRfS3_($this,$x,$y,$x_o,$y_o) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $x_o = $x_o|0;
 $y_o = $y_o|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0.0, $8 = 0.0, $9 = 0.0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp + 24|0;
 $0 = sp + 16|0;
 $1 = sp + 8|0;
 $2 = sp;
 $3 = ((($this)) + 8|0); //@line 352 "modules/physics/box2d/Body.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 352 "modules/physics/box2d/Body.cpp"
 HEAPF32[$2>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($2)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$5>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($1,$2); //@line 352 "modules/physics/box2d/Body.cpp"
 $6 = ((($4)) + 24|0); //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$1>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $9 = $7 * $8; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($4)) + 20|0); //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $11 = +HEAPF32[$10>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $12 = ((($1)) + 4|0); //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $13 = +HEAPF32[$12>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $14 = $11 * $13; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $15 = $9 + $14; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $16 = $8 * $11; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $17 = $7 * $13; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $18 = $17 - $16; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$0>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($0)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$19>>2] = $18; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($v,$0); //@line 352 "modules/physics/box2d/Body.cpp"
 $20 = HEAP32[$v>>2]|0; //@line 353 "modules/physics/box2d/Body.cpp"
 HEAP32[$x_o>>2] = $20; //@line 353 "modules/physics/box2d/Body.cpp"
 $21 = ((($v)) + 4|0); //@line 354 "modules/physics/box2d/Body.cpp"
 $22 = HEAP32[$21>>2]|0; //@line 354 "modules/physics/box2d/Body.cpp"
 HEAP32[$y_o>>2] = $22; //@line 354 "modules/physics/box2d/Body.cpp"
 STACKTOP = sp;return; //@line 355 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body31getLinearVelocityFromWorldPointEffRfS3_($this,$x,$y,$x_o,$y_o) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $x_o = $x_o|0;
 $y_o = $y_o|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp + 24|0;
 $0 = sp + 16|0;
 $1 = sp + 8|0;
 $2 = sp;
 $3 = ((($this)) + 8|0); //@line 359 "modules/physics/box2d/Body.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 359 "modules/physics/box2d/Body.cpp"
 HEAPF32[$2>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($2)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$5>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($1,$2); //@line 359 "modules/physics/box2d/Body.cpp"
 $6 = ((($4)) + 72|0); //@line 578 "./libraries/Box2D/Dynamics/b2Body.h"
 $7 = +HEAPF32[$6>>2]; //@line 578 "./libraries/Box2D/Dynamics/b2Body.h"
 $8 = +HEAPF32[$1>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($4)) + 44|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 - $10; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $12 = ((($1)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $13 = +HEAPF32[$12>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $14 = ((($4)) + 48|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $15 = +HEAPF32[$14>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $16 = $13 - $15; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $17 = $7 * $16; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $18 = $7 * $11; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($4)) + 64|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $21 = $20 - $17; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($4)) + 68|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $23 = +HEAPF32[$22>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $24 = $18 + $23; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$0>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $25 = ((($0)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$25>>2] = $24; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($v,$0); //@line 359 "modules/physics/box2d/Body.cpp"
 $26 = HEAP32[$v>>2]|0; //@line 360 "modules/physics/box2d/Body.cpp"
 HEAP32[$x_o>>2] = $26; //@line 360 "modules/physics/box2d/Body.cpp"
 $27 = ((($v)) + 4|0); //@line 361 "modules/physics/box2d/Body.cpp"
 $28 = HEAP32[$27>>2]|0; //@line 361 "modules/physics/box2d/Body.cpp"
 HEAP32[$y_o>>2] = $28; //@line 361 "modules/physics/box2d/Body.cpp"
 STACKTOP = sp;return; //@line 362 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body31getLinearVelocityFromLocalPointEffRfS3_($this,$x,$y,$x_o,$y_o) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $x_o = $x_o|0;
 $y_o = $y_o|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0.0;
 var $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $5 = 0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp + 24|0;
 $0 = sp + 16|0;
 $1 = sp + 8|0;
 $2 = sp;
 $3 = ((($this)) + 8|0); //@line 366 "modules/physics/box2d/Body.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 366 "modules/physics/box2d/Body.cpp"
 HEAPF32[$2>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($2)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$5>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($1,$2); //@line 366 "modules/physics/box2d/Body.cpp"
 $6 = ((($4)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$1>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = $7 * $8; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($4)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = +HEAPF32[$10>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = ((($1)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = +HEAPF32[$12>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = $11 * $13; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $9 - $14; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = ((($4)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $17 = +HEAPF32[$16>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $18 = $17 + $15; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $19 = $8 * $11; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = $7 * $13; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $19 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($4)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $23 = +HEAPF32[$22>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $24 = $21 + $23; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $25 = ((($4)) + 72|0); //@line 578 "./libraries/Box2D/Dynamics/b2Body.h"
 $26 = +HEAPF32[$25>>2]; //@line 578 "./libraries/Box2D/Dynamics/b2Body.h"
 $27 = ((($4)) + 44|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $28 = +HEAPF32[$27>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $29 = $18 - $28; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $30 = ((($4)) + 48|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $31 = +HEAPF32[$30>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $32 = $24 - $31; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $33 = $26 * $32; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $34 = $26 * $29; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $35 = ((($4)) + 64|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $36 = +HEAPF32[$35>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $37 = $36 - $33; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $38 = ((($4)) + 68|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $39 = +HEAPF32[$38>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $40 = $34 + $39; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$0>>2] = $37; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $41 = ((($0)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$41>>2] = $40; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($v,$0); //@line 366 "modules/physics/box2d/Body.cpp"
 $42 = HEAP32[$v>>2]|0; //@line 367 "modules/physics/box2d/Body.cpp"
 HEAP32[$x_o>>2] = $42; //@line 367 "modules/physics/box2d/Body.cpp"
 $43 = ((($v)) + 4|0); //@line 368 "modules/physics/box2d/Body.cpp"
 $44 = HEAP32[$43>>2]|0; //@line 368 "modules/physics/box2d/Body.cpp"
 HEAP32[$y_o>>2] = $44; //@line 368 "modules/physics/box2d/Body.cpp"
 STACKTOP = sp;return; //@line 369 "modules/physics/box2d/Body.cpp"
}
function __ZNK4love7physics5box2d4Body8isBulletEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 373 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 373 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 4|0); //@line 630 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 630 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3 & 8; //@line 630 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = ($4<<16>>16)!=(0); //@line 630 "./libraries/Box2D/Dynamics/b2Body.h"
 return ($5|0); //@line 373 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body9setBulletEb($this,$bullet) {
 $this = $this|0;
 $bullet = $bullet|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $storemerge$i = 0, $storemerge$in$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 378 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 378 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 4|0); //@line 620 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 620 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3&65535; //@line 620 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = $4 & 65527; //@line 624 "./libraries/Box2D/Dynamics/b2Body.h"
 $6 = $4 | 8; //@line 620 "./libraries/Box2D/Dynamics/b2Body.h"
 $storemerge$in$i = $bullet ? $6 : $5; //@line 618 "./libraries/Box2D/Dynamics/b2Body.h"
 $storemerge$i = $storemerge$in$i&65535; //@line 624 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP16[$2>>1] = $storemerge$i; //@line 620 "./libraries/Box2D/Dynamics/b2Body.h"
 return; //@line 378 "modules/physics/box2d/Body.cpp"
}
function __ZNK4love7physics5box2d4Body8isActiveEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 383 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 383 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 4|0); //@line 661 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 661 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3 & 32; //@line 661 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = ($4<<16>>16)!=(0); //@line 661 "./libraries/Box2D/Dynamics/b2Body.h"
 return ($5|0); //@line 383 "modules/physics/box2d/Body.cpp"
}
function __ZNK4love7physics5box2d4Body7isAwakeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 388 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 388 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 4|0); //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3 & 2; //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = ($4<<16>>16)!=(0); //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
 return ($5|0); //@line 388 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body18setSleepingAllowedEb($this,$allow) {
 $this = $this|0;
 $allow = $allow|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 393 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 393 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 4|0); //@line 673 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 673 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3&65535; //@line 673 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($allow) {
  $5 = $4 | 4; //@line 673 "./libraries/Box2D/Dynamics/b2Body.h"
  $6 = $5&65535; //@line 673 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$2>>1] = $6; //@line 673 "./libraries/Box2D/Dynamics/b2Body.h"
  return; //@line 394 "modules/physics/box2d/Body.cpp"
 }
 $7 = $4 & 65531; //@line 677 "./libraries/Box2D/Dynamics/b2Body.h"
 $8 = $7&65535; //@line 677 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP16[$2>>1] = $8; //@line 677 "./libraries/Box2D/Dynamics/b2Body.h"
 $9 = $4 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $10 = ($9|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($10)) {
  return; //@line 394 "modules/physics/box2d/Body.cpp"
 }
 $11 = $7 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $12 = $11&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP16[$2>>1] = $12; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $13 = ((($1)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$13>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 return; //@line 394 "modules/physics/box2d/Body.cpp"
}
function __ZNK4love7physics5box2d4Body17isSleepingAllowedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 398 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 398 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 4|0); //@line 684 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 684 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3 & 4; //@line 684 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = ($4<<16>>16)!=(0); //@line 684 "./libraries/Box2D/Dynamics/b2Body.h"
 return ($5|0); //@line 398 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body9setActiveEb($this,$active) {
 $this = $this|0;
 $active = $active|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 403 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 403 "modules/physics/box2d/Body.cpp"
 __ZN6b2Body9SetActiveEb($1,$active); //@line 403 "modules/physics/box2d/Body.cpp"
 return; //@line 404 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body8setAwakeEb($this,$awake) {
 $this = $this|0;
 $awake = $awake|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 408 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 408 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($awake)) {
  $10 = $4 & 65533; //@line 645 "./libraries/Box2D/Dynamics/b2Body.h"
  $11 = $10&65535; //@line 645 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$2>>1] = $11; //@line 645 "./libraries/Box2D/Dynamics/b2Body.h"
  $12 = ((($1)) + 144|0); //@line 646 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$12>>2] = 0.0; //@line 646 "./libraries/Box2D/Dynamics/b2Body.h"
  $13 = ((($1)) + 64|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
  ;HEAP32[$13>>2]=0|0;HEAP32[$13+4>>2]=0|0;HEAP32[$13+8>>2]=0|0;HEAP32[$13+12>>2]=0|0;HEAP32[$13+16>>2]=0|0;HEAP32[$13+20>>2]=0|0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  return; //@line 409 "modules/physics/box2d/Body.cpp"
 }
 $5 = $4 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $6 = ($5|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if (!($6)) {
  return; //@line 409 "modules/physics/box2d/Body.cpp"
 }
 $7 = $4 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $8 = $7&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAP16[$2>>1] = $8; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
 $9 = ((($1)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 HEAPF32[$9>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 return; //@line 409 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d4Body16setFixedRotationEb($this,$fixed) {
 $this = $this|0;
 $fixed = $fixed|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 413 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 413 "modules/physics/box2d/Body.cpp"
 __ZN6b2Body16SetFixedRotationEb($1,$fixed); //@line 413 "modules/physics/box2d/Body.cpp"
 return; //@line 414 "modules/physics/box2d/Body.cpp"
}
function __ZNK4love7physics5box2d4Body15isFixedRotationEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 418 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 418 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 4|0); //@line 666 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP16[$2>>1]|0; //@line 666 "./libraries/Box2D/Dynamics/b2Body.h"
 $4 = $3 & 16; //@line 666 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = ($4<<16>>16)!=(0); //@line 666 "./libraries/Box2D/Dynamics/b2Body.h"
 return ($5|0); //@line 418 "modules/physics/box2d/Body.cpp"
}
function __ZNK4love7physics5box2d4Body8getWorldEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 423 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 423 "modules/physics/box2d/Body.cpp"
 return ($1|0); //@line 423 "modules/physics/box2d/Body.cpp"
}
function __ZNK4love7physics5box2d4Body14getFixtureListEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $$old1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f$0 = 0, $i$0 = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _lua_createtable($L,0,0); //@line 428 "modules/physics/box2d/Body.cpp"
 $0 = ((($this)) + 8|0); //@line 429 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 429 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 100|0); //@line 689 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP32[$2>>2]|0; //@line 689 "./libraries/Box2D/Dynamics/b2Body.h"
 $$old1 = ($3|0)==(0|0); //@line 433 "modules/physics/box2d/Body.cpp"
 if ($$old1) {
  STACKTOP = sp;return 1; //@line 443 "modules/physics/box2d/Body.cpp"
 } else {
  $f$0 = $3;$i$0 = 1;
 }
 while(1) {
  $4 = (__ZN4love8Memoizer4findEPv($f$0)|0); //@line 435 "modules/physics/box2d/Body.cpp"
  $5 = ($4|0)==(0|0); //@line 436 "modules/physics/box2d/Body.cpp"
  if ($5) {
   break;
  }
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,35,$4); //@line 438 "modules/physics/box2d/Body.cpp"
  _lua_rawseti($L,-2,$i$0); //@line 439 "modules/physics/box2d/Body.cpp"
  $11 = (($i$0) + 1)|0; //@line 440 "modules/physics/box2d/Body.cpp"
  $12 = ((($f$0)) + 4|0); //@line 285 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $13 = HEAP32[$12>>2]|0; //@line 285 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $14 = ($13|0)==(0|0); //@line 442 "modules/physics/box2d/Body.cpp"
  if ($14) {
   label = 7;
   break;
  } else {
   $f$0 = $13;$i$0 = $11;
  }
 }
 if ((label|0) == 7) {
  STACKTOP = sp;return 1; //@line 443 "modules/physics/box2d/Body.cpp"
 }
 $6 = (___cxa_allocate_exception(16)|0); //@line 437 "modules/physics/box2d/Body.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($6|0),(444616|0),($vararg_buffer|0)); //@line 437 "modules/physics/box2d/Body.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  ___cxa_free_exception(($6|0)); //@line 437 "modules/physics/box2d/Body.cpp"
  ___resumeException($9|0);
  // unreachable;
 } else {
  ___cxa_throw(($6|0),(96|0),(344|0)); //@line 437 "modules/physics/box2d/Body.cpp"
  // unreachable; //@line 437 "modules/physics/box2d/Body.cpp"
 }
 return (0)|0;
}
function __ZNK4love7physics5box2d4Body12getJointListEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $$old1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0, $je$0 = 0;
 var $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _lua_createtable($L,0,0); //@line 448 "modules/physics/box2d/Body.cpp"
 $0 = ((($this)) + 8|0); //@line 449 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 449 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 108|0); //@line 699 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP32[$2>>2]|0; //@line 699 "./libraries/Box2D/Dynamics/b2Body.h"
 $$old1 = ($3|0)==(0|0); //@line 454 "modules/physics/box2d/Body.cpp"
 if ($$old1) {
  STACKTOP = sp;return 1; //@line 467 "modules/physics/box2d/Body.cpp"
 } else {
  $i$0 = 1;$je$0 = $3;
 }
 while(1) {
  $4 = ((($je$0)) + 4|0); //@line 457 "modules/physics/box2d/Body.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 457 "modules/physics/box2d/Body.cpp"
  $6 = (__ZN4love8Memoizer4findEPv($5)|0); //@line 457 "modules/physics/box2d/Body.cpp"
  $7 = ($6|0)==(0|0); //@line 458 "modules/physics/box2d/Body.cpp"
  if ($7) {
   break;
  }
  __ZN4love7physics5box2d14luax_pushjointEP9lua_StatePNS1_5JointE($L,$6); //@line 461 "modules/physics/box2d/Body.cpp"
  _lua_rawseti($L,-2,$i$0); //@line 462 "modules/physics/box2d/Body.cpp"
  $13 = (($i$0) + 1)|0; //@line 463 "modules/physics/box2d/Body.cpp"
  $14 = ((($je$0)) + 12|0); //@line 465 "modules/physics/box2d/Body.cpp"
  $15 = HEAP32[$14>>2]|0; //@line 465 "modules/physics/box2d/Body.cpp"
  $16 = ($15|0)==(0|0); //@line 465 "modules/physics/box2d/Body.cpp"
  if ($16) {
   label = 7;
   break;
  } else {
   $i$0 = $13;$je$0 = $15;
  }
 }
 if ((label|0) == 7) {
  STACKTOP = sp;return 1; //@line 467 "modules/physics/box2d/Body.cpp"
 }
 $8 = (___cxa_allocate_exception(16)|0); //@line 459 "modules/physics/box2d/Body.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($8|0),(444962|0),($vararg_buffer|0)); //@line 459 "modules/physics/box2d/Body.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $11 = ___cxa_find_matching_catch_2()|0;
  $12 = tempRet0;
  ___cxa_free_exception(($8|0)); //@line 459 "modules/physics/box2d/Body.cpp"
  ___resumeException($11|0);
  // unreachable;
 } else {
  ___cxa_throw(($8|0),(96|0),(344|0)); //@line 459 "modules/physics/box2d/Body.cpp"
  // unreachable; //@line 459 "modules/physics/box2d/Body.cpp"
 }
 return (0)|0;
}
function __ZNK4love7physics5box2d4Body14getContactListEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $$lcssa = 0, $$old1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ce$0 = 0, $contact$0 = 0, $i$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _lua_createtable($L,0,0); //@line 472 "modules/physics/box2d/Body.cpp"
 $0 = ((($this)) + 8|0); //@line 473 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 473 "modules/physics/box2d/Body.cpp"
 $2 = ((($1)) + 112|0); //@line 709 "./libraries/Box2D/Dynamics/b2Body.h"
 $3 = HEAP32[$2>>2]|0; //@line 709 "./libraries/Box2D/Dynamics/b2Body.h"
 $$old1 = ($3|0)==(0|0); //@line 477 "modules/physics/box2d/Body.cpp"
 if ($$old1) {
  return 1; //@line 492 "modules/physics/box2d/Body.cpp"
 } else {
  $ce$0 = $3;$i$0 = 1;
 }
 while(1) {
  $4 = ((($ce$0)) + 4|0); //@line 480 "modules/physics/box2d/Body.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 480 "modules/physics/box2d/Body.cpp"
  $6 = (__ZN4love8Memoizer4findEPv($5)|0); //@line 480 "modules/physics/box2d/Body.cpp"
  $7 = ($6|0)==(0|0); //@line 481 "modules/physics/box2d/Body.cpp"
  if ($7) {
   $8 = (__Znwj(12)|0); //@line 482 "modules/physics/box2d/Body.cpp"
   $9 = HEAP32[$4>>2]|0; //@line 482 "modules/physics/box2d/Body.cpp"
   __THREW__ = 0;
   invoke_vii(3706,($8|0),($9|0)); //@line 482 "modules/physics/box2d/Body.cpp"
   $10 = __THREW__; __THREW__ = 0;
   $11 = $10&1;
   if ($11) {
    $$lcssa = $8;
    label = 4;
    break;
   } else {
    $contact$0 = $8;
   }
  } else {
   $14 = HEAP32[$6>>2]|0; //@line 484 "modules/physics/box2d/Body.cpp"
   $15 = ((($14)) + 8|0); //@line 484 "modules/physics/box2d/Body.cpp"
   $16 = HEAP32[$15>>2]|0; //@line 484 "modules/physics/box2d/Body.cpp"
   FUNCTION_TABLE_vi[$16 & 8191]($6); //@line 484 "modules/physics/box2d/Body.cpp"
   $contact$0 = $6;
  }
  __ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE($L,33,$contact$0); //@line 486 "modules/physics/box2d/Body.cpp"
  $17 = HEAP32[$contact$0>>2]|0; //@line 487 "modules/physics/box2d/Body.cpp"
  $18 = ((($17)) + 12|0); //@line 487 "modules/physics/box2d/Body.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 487 "modules/physics/box2d/Body.cpp"
  FUNCTION_TABLE_vi[$19 & 8191]($contact$0); //@line 487 "modules/physics/box2d/Body.cpp"
  _lua_rawseti($L,-2,$i$0); //@line 488 "modules/physics/box2d/Body.cpp"
  $20 = (($i$0) + 1)|0; //@line 489 "modules/physics/box2d/Body.cpp"
  $21 = ((($ce$0)) + 12|0); //@line 491 "modules/physics/box2d/Body.cpp"
  $22 = HEAP32[$21>>2]|0; //@line 491 "modules/physics/box2d/Body.cpp"
  $23 = ($22|0)==(0|0); //@line 491 "modules/physics/box2d/Body.cpp"
  if ($23) {
   label = 7;
   break;
  } else {
   $ce$0 = $22;$i$0 = $20;
  }
 }
 if ((label|0) == 4) {
  $12 = ___cxa_find_matching_catch_2()|0;
  $13 = tempRet0;
  __ZdlPv($$lcssa); //@line 482 "modules/physics/box2d/Body.cpp"
  ___resumeException($12|0);
  // unreachable;
 }
 else if ((label|0) == 7) {
  return 1; //@line 492 "modules/physics/box2d/Body.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d4Body7destroyEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = ((($this)) + 12|0); //@line 512 "modules/physics/box2d/Body.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 512 "modules/physics/box2d/Body.cpp"
 $3 = ((($2)) + 20|0); //@line 512 "modules/physics/box2d/Body.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 512 "modules/physics/box2d/Body.cpp"
 $5 = ((($4)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $6 = HEAP32[$5>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $7 = $6 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $8 = ($7|0)==(0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 if ($8) {
  $22 = ((($this)) + 8|0); //@line 520 "modules/physics/box2d/Body.cpp"
  $23 = HEAP32[$22>>2]|0; //@line 520 "modules/physics/box2d/Body.cpp"
  __ZN7b2World11DestroyBodyEP6b2Body($4,$23); //@line 520 "modules/physics/box2d/Body.cpp"
  $24 = HEAP32[$22>>2]|0; //@line 521 "modules/physics/box2d/Body.cpp"
  __ZN4love8Memoizer6removeEPv($24); //@line 521 "modules/physics/box2d/Body.cpp"
  HEAP32[$22>>2] = 0; //@line 522 "modules/physics/box2d/Body.cpp"
  $25 = HEAP32[$this>>2]|0; //@line 525 "modules/physics/box2d/Body.cpp"
  $26 = ((($25)) + 12|0); //@line 525 "modules/physics/box2d/Body.cpp"
  $27 = HEAP32[$26>>2]|0; //@line 525 "modules/physics/box2d/Body.cpp"
  FUNCTION_TABLE_vi[$27 & 8191]($this); //@line 525 "modules/physics/box2d/Body.cpp"
  STACKTOP = sp;return; //@line 526 "modules/physics/box2d/Body.cpp"
 }
 $9 = HEAP32[$this>>2]|0; //@line 515 "modules/physics/box2d/Body.cpp"
 $10 = ((($9)) + 8|0); //@line 515 "modules/physics/box2d/Body.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 515 "modules/physics/box2d/Body.cpp"
 FUNCTION_TABLE_vi[$11 & 8191]($this); //@line 515 "modules/physics/box2d/Body.cpp"
 $12 = HEAP32[$1>>2]|0; //@line 516 "modules/physics/box2d/Body.cpp"
 HEAP32[$0>>2] = $this; //@line 516 "modules/physics/box2d/Body.cpp"
 $13 = ((($12)) + 32|0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $14 = HEAP32[$13>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = ((($12)) + 36|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $16 = HEAP32[$15>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ($14>>>0)<($16>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($17) {
  $18 = $this; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  HEAP32[$14>>2] = $18; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $19 = HEAP32[$13>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = ((($19)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$13>>2] = $20; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  STACKTOP = sp;return; //@line 526 "modules/physics/box2d/Body.cpp"
 } else {
  $21 = ((($12)) + 28|0); //@line 516 "modules/physics/box2d/Body.cpp"
  __ZNSt3__16vectorIPN4love7physics5box2d4BodyENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_($21,$0); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  STACKTOP = sp;return; //@line 526 "modules/physics/box2d/Body.cpp"
 }
}
function __ZNSt3__16vectorIPN4love7physics5box2d4BodyENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love7physics5box2d4Body11setUserDataEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 (__ZN4love16luax_assert_argcEP9lua_Stateii($L,1,1)|0); //@line 530 "modules/physics/box2d/Body.cpp"
 $0 = ((($this)) + 16|0); //@line 532 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 532 "modules/physics/box2d/Body.cpp"
 $2 = ($1|0)==(0|0); //@line 532 "modules/physics/box2d/Body.cpp"
 if ($2) {
  $3 = (__Znwj(4)|0); //@line 534 "modules/physics/box2d/Body.cpp"
  HEAP32[$3>>2] = 0; //@line 51 "./modules/physics/box2d/Body.h"
  HEAP32[$0>>2] = $3; //@line 534 "modules/physics/box2d/Body.cpp"
  $4 = ((($this)) + 8|0); //@line 535 "modules/physics/box2d/Body.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 535 "modules/physics/box2d/Body.cpp"
  $6 = ((($5)) + 148|0); //@line 729 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP32[$6>>2] = $3; //@line 729 "./libraries/Box2D/Dynamics/b2Body.h"
  $8 = $3;
 } else {
  $8 = $1;
 }
 $7 = HEAP32[$8>>2]|0; //@line 538 "modules/physics/box2d/Body.cpp"
 $9 = ($7|0)==(0|0); //@line 538 "modules/physics/box2d/Body.cpp"
 if (!($9)) {
  $10 = HEAP32[$7>>2]|0; //@line 538 "modules/physics/box2d/Body.cpp"
  $11 = ((($10)) + 4|0); //@line 538 "modules/physics/box2d/Body.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 538 "modules/physics/box2d/Body.cpp"
  FUNCTION_TABLE_vi[$12 & 8191]($7); //@line 538 "modules/physics/box2d/Body.cpp"
 }
 $13 = (__Znwj(12)|0); //@line 539 "modules/physics/box2d/Body.cpp"
 __THREW__ = 0;
 invoke_vii(3493,($13|0),($L|0)); //@line 539 "modules/physics/box2d/Body.cpp"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if ($15) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  __ZdlPv($13); //@line 539 "modules/physics/box2d/Body.cpp"
  ___resumeException($17|0);
  // unreachable;
 } else {
  $16 = HEAP32[$0>>2]|0; //@line 539 "modules/physics/box2d/Body.cpp"
  HEAP32[$16>>2] = $13; //@line 539 "modules/physics/box2d/Body.cpp"
  return 0; //@line 541 "modules/physics/box2d/Body.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d4Body11getUserDataEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 546 "modules/physics/box2d/Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 546 "modules/physics/box2d/Body.cpp"
 $2 = ($1|0)==(0|0); //@line 546 "modules/physics/box2d/Body.cpp"
 if (!($2)) {
  $3 = HEAP32[$1>>2]|0; //@line 546 "modules/physics/box2d/Body.cpp"
  $4 = ($3|0)==(0|0); //@line 546 "modules/physics/box2d/Body.cpp"
  if (!($4)) {
   __ZN4love9Reference4pushEP9lua_State($3,$L); //@line 547 "modules/physics/box2d/Body.cpp"
   return 1; //@line 551 "modules/physics/box2d/Body.cpp"
  }
 }
 _lua_pushnil($L); //@line 549 "modules/physics/box2d/Body.cpp"
 return 1; //@line 551 "modules/physics/box2d/Body.cpp"
}
function __ZN4love7physics5box2d10ChainShapeD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5ShapeD2Ev($this); //@line 44 "modules/physics/box2d/ChainShape.cpp"
 return; //@line 44 "modules/physics/box2d/ChainShape.cpp"
}
function __ZN4love7physics5box2d10ChainShapeD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5ShapeD2Ev($this); //@line 44 "modules/physics/box2d/ChainShape.cpp"
 __ZdlPv($this); //@line 43 "modules/physics/box2d/ChainShape.cpp"
 return; //@line 44 "modules/physics/box2d/ChainShape.cpp"
}
function __ZN4love7physics5box2d10ChainShapeC2EP12b2ChainShapebb($this,$c,$loop,$own) {
 $this = $this|0;
 $c = $c|0;
 $loop = $loop|0;
 $own = $own|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5ShapeC2EP7b2Shapeb($this,$c,$own); //@line 38 "modules/physics/box2d/ChainShape.cpp"
 HEAP32[$this>>2] = (32616); //@line 39 "modules/physics/box2d/ChainShape.cpp"
 $0 = ((($this)) + 13|0); //@line 38 "modules/physics/box2d/ChainShape.cpp"
 $1 = $loop&1; //@line 38 "modules/physics/box2d/ChainShape.cpp"
 HEAP8[$0>>0] = $1; //@line 38 "modules/physics/box2d/ChainShape.cpp"
 return; //@line 40 "modules/physics/box2d/ChainShape.cpp"
}
function __ZN4love7physics5box2d10ChainShape13setNextVertexEff($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $v = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $v = sp + 16|0;
 $0 = sp + 8|0;
 $1 = ((($this)) + 13|0); //@line 48 "modules/physics/box2d/ChainShape.cpp"
 $2 = HEAP8[$1>>0]|0; //@line 48 "modules/physics/box2d/ChainShape.cpp"
 $3 = ($2<<24>>24)==(0); //@line 48 "modules/physics/box2d/ChainShape.cpp"
 if ($3) {
  HEAPF32[$v>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
  $9 = ((($v)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$9>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
  $10 = ((($this)) + 8|0); //@line 54 "modules/physics/box2d/ChainShape.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 54 "modules/physics/box2d/ChainShape.cpp"
  __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($0,$v); //@line 55 "modules/physics/box2d/ChainShape.cpp"
  __ZN12b2ChainShape13SetNextVertexERK6b2Vec2($11,$0); //@line 55 "modules/physics/box2d/ChainShape.cpp"
  STACKTOP = sp;return; //@line 56 "modules/physics/box2d/ChainShape.cpp"
 }
 $4 = (___cxa_allocate_exception(16)|0); //@line 50 "modules/physics/box2d/ChainShape.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($4|0),(444390|0),($vararg_buffer|0)); //@line 50 "modules/physics/box2d/ChainShape.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  ___cxa_free_exception(($4|0)); //@line 50 "modules/physics/box2d/ChainShape.cpp"
  ___resumeException($7|0);
  // unreachable;
 } else {
  ___cxa_throw(($4|0),(96|0),(344|0)); //@line 50 "modules/physics/box2d/ChainShape.cpp"
  // unreachable; //@line 50 "modules/physics/box2d/ChainShape.cpp"
 }
}
function __ZN4love7physics5box2d10ChainShape17setPreviousVertexEff($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $v = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $v = sp + 16|0;
 $0 = sp + 8|0;
 $1 = ((($this)) + 13|0); //@line 60 "modules/physics/box2d/ChainShape.cpp"
 $2 = HEAP8[$1>>0]|0; //@line 60 "modules/physics/box2d/ChainShape.cpp"
 $3 = ($2<<24>>24)==(0); //@line 60 "modules/physics/box2d/ChainShape.cpp"
 if ($3) {
  HEAPF32[$v>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
  $9 = ((($v)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$9>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
  $10 = ((($this)) + 8|0); //@line 66 "modules/physics/box2d/ChainShape.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 66 "modules/physics/box2d/ChainShape.cpp"
  __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($0,$v); //@line 67 "modules/physics/box2d/ChainShape.cpp"
  __ZN12b2ChainShape13SetNextVertexERK6b2Vec2($11,$0); //@line 67 "modules/physics/box2d/ChainShape.cpp"
  STACKTOP = sp;return; //@line 68 "modules/physics/box2d/ChainShape.cpp"
 }
 $4 = (___cxa_allocate_exception(16)|0); //@line 62 "modules/physics/box2d/ChainShape.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($4|0),(444451|0),($vararg_buffer|0)); //@line 62 "modules/physics/box2d/ChainShape.cpp"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  ___cxa_free_exception(($4|0)); //@line 62 "modules/physics/box2d/ChainShape.cpp"
  ___resumeException($7|0);
  // unreachable;
 } else {
  ___cxa_throw(($4|0),(96|0),(344|0)); //@line 62 "modules/physics/box2d/ChainShape.cpp"
  // unreachable; //@line 62 "modules/physics/box2d/ChainShape.cpp"
 }
}
function __ZNK4love7physics5box2d10ChainShape12getChildEdgeEi($this,$index) {
 $this = $this|0;
 $index = $index|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 72 "modules/physics/box2d/ChainShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 72 "modules/physics/box2d/ChainShape.cpp"
 $2 = (__Znwj(48)|0); //@line 73 "modules/physics/box2d/ChainShape.cpp"
 HEAP32[$2>>2] = (20456); //@line 63 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $3 = ((($2)) + 4|0); //@line 64 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 HEAP32[$3>>2] = 1; //@line 64 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $4 = ((($2)) + 8|0); //@line 65 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 HEAPF32[$4>>2] = 0.0099999997764825821; //@line 65 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $5 = ((($2)) + 28|0); //@line 66 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;HEAP32[$5+12>>2]=0|0;HEAP16[$5+16>>1]=0|0; //@line 67 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 __THREW__ = 0;
 invoke_viii(4211,($1|0),($2|0),($index|0)); //@line 77 "modules/physics/box2d/ChainShape.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  $20 = (__Znwj(16)|0); //@line 85 "modules/physics/box2d/ChainShape.cpp"
  __THREW__ = 0;
  invoke_viii(3675,($20|0),($2|0),1); //@line 85 "modules/physics/box2d/ChainShape.cpp"
  $21 = __THREW__; __THREW__ = 0;
  $22 = $21&1;
  if (!($22)) {
   return ($20|0); //@line 85 "modules/physics/box2d/ChainShape.cpp"
  }
  $23 = ___cxa_find_matching_catch_2()|0;
  $24 = tempRet0;
  __ZdlPv($20); //@line 85 "modules/physics/box2d/ChainShape.cpp"
  $$0 = $23;$$01 = $24;
  ___resumeException($$0|0);
  // unreachable;
 }
 $8 = ___cxa_find_matching_catch_3(96|0)|0;
 $9 = tempRet0;
 $10 = (_llvm_eh_typeid_for((96|0))|0); //@line 78 "modules/physics/box2d/ChainShape.cpp"
 $11 = ($9|0)==($10|0); //@line 78 "modules/physics/box2d/ChainShape.cpp"
 if (!($11)) {
  $$0 = $8;$$01 = $9;
  ___resumeException($$0|0);
  // unreachable;
 }
 (___cxa_begin_catch(($8|0))|0); //@line 78 "modules/physics/box2d/ChainShape.cpp"
 $12 = HEAP32[$2>>2]|0; //@line 81 "modules/physics/box2d/ChainShape.cpp"
 $13 = ((($12)) + 4|0); //@line 81 "modules/physics/box2d/ChainShape.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 81 "modules/physics/box2d/ChainShape.cpp"
 FUNCTION_TABLE_vi[$14 & 8191]($2); //@line 81 "modules/physics/box2d/ChainShape.cpp"
 __THREW__ = 0;
 invoke_v(3656); //@line 82 "modules/physics/box2d/ChainShape.cpp"
 $15 = __THREW__; __THREW__ = 0;
 $16 = ___cxa_find_matching_catch_2()|0;
 $17 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378); //@line 83 "modules/physics/box2d/ChainShape.cpp"
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $25 = ___cxa_find_matching_catch_3(0|0)|0;
  $26 = tempRet0;
  ___clang_call_terminate($25); //@line 83 "modules/physics/box2d/ChainShape.cpp"
  // unreachable; //@line 83 "modules/physics/box2d/ChainShape.cpp"
 } else {
  $$0 = $16;$$01 = $17;
  ___resumeException($$0|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZNK4love7physics5box2d10ChainShape14getVertexCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 90 "modules/physics/box2d/ChainShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 90 "modules/physics/box2d/ChainShape.cpp"
 $2 = ((($1)) + 16|0); //@line 91 "modules/physics/box2d/ChainShape.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 91 "modules/physics/box2d/ChainShape.cpp"
 return ($3|0); //@line 91 "modules/physics/box2d/ChainShape.cpp"
}
function __ZNK4love7physics5box2d10ChainShape8getPointEi($agg$result,$this,$index) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $index = $index|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 8|0); //@line 96 "modules/physics/box2d/ChainShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 96 "modules/physics/box2d/ChainShape.cpp"
 $2 = ($index|0)<(0); //@line 97 "modules/physics/box2d/ChainShape.cpp"
 if (!($2)) {
  $3 = ((($1)) + 16|0); //@line 97 "modules/physics/box2d/ChainShape.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 97 "modules/physics/box2d/ChainShape.cpp"
  $5 = ($4|0)>($index|0); //@line 97 "modules/physics/box2d/ChainShape.cpp"
  if ($5) {
   $11 = ((($1)) + 12|0); //@line 99 "modules/physics/box2d/ChainShape.cpp"
   $12 = HEAP32[$11>>2]|0; //@line 99 "modules/physics/box2d/ChainShape.cpp"
   $13 = (($12) + ($index<<3)|0); //@line 99 "modules/physics/box2d/ChainShape.cpp"
   __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($agg$result,$13); //@line 100 "modules/physics/box2d/ChainShape.cpp"
   STACKTOP = sp;return; //@line 101 "modules/physics/box2d/ChainShape.cpp"
  }
 }
 $6 = (___cxa_allocate_exception(16)|0); //@line 98 "modules/physics/box2d/ChainShape.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($6|0),(444516|0),($vararg_buffer|0)); //@line 98 "modules/physics/box2d/ChainShape.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  ___cxa_free_exception(($6|0)); //@line 98 "modules/physics/box2d/ChainShape.cpp"
  ___resumeException($9|0);
  // unreachable;
 } else {
  ___cxa_throw(($6|0),(96|0),(344|0)); //@line 98 "modules/physics/box2d/ChainShape.cpp"
  // unreachable; //@line 98 "modules/physics/box2d/ChainShape.cpp"
 }
}
function __ZNK4love7physics5box2d10ChainShape9getPointsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 105 "modules/physics/box2d/ChainShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 105 "modules/physics/box2d/ChainShape.cpp"
 $2 = ((($1)) + 12|0); //@line 106 "modules/physics/box2d/ChainShape.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 106 "modules/physics/box2d/ChainShape.cpp"
 return ($3|0); //@line 106 "modules/physics/box2d/ChainShape.cpp"
}
function __ZN4love7physics5box2d11CircleShapeD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5ShapeD2Ev($this); //@line 44 "modules/physics/box2d/CircleShape.cpp"
 return; //@line 44 "modules/physics/box2d/CircleShape.cpp"
}
function __ZN4love7physics5box2d11CircleShapeD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5ShapeD2Ev($this); //@line 44 "modules/physics/box2d/CircleShape.cpp"
 __ZdlPv($this); //@line 43 "modules/physics/box2d/CircleShape.cpp"
 return; //@line 44 "modules/physics/box2d/CircleShape.cpp"
}
function __ZN4love7physics5box2d11CircleShapeC2EP13b2CircleShapeb($this,$c,$own) {
 $this = $this|0;
 $c = $c|0;
 $own = $own|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5ShapeC2EP7b2Shapeb($this,$c,$own); //@line 38 "modules/physics/box2d/CircleShape.cpp"
 HEAP32[$this>>2] = (32640); //@line 39 "modules/physics/box2d/CircleShape.cpp"
 return; //@line 40 "modules/physics/box2d/CircleShape.cpp"
}
function __ZNK4love7physics5box2d11CircleShape9getRadiusEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 48 "modules/physics/box2d/CircleShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 48 "modules/physics/box2d/CircleShape.cpp"
 $2 = ((($1)) + 8|0); //@line 48 "modules/physics/box2d/CircleShape.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 48 "modules/physics/box2d/CircleShape.cpp"
 $4 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($3)); //@line 48 "modules/physics/box2d/CircleShape.cpp"
 return (+$4); //@line 48 "modules/physics/box2d/CircleShape.cpp"
}
function __ZN4love7physics5box2d11CircleShape9setRadiusEf($this,$r) {
 $this = $this|0;
 $r = +$r;
 var $0 = 0.0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($r)); //@line 53 "modules/physics/box2d/CircleShape.cpp"
 $1 = ((($this)) + 8|0); //@line 53 "modules/physics/box2d/CircleShape.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 53 "modules/physics/box2d/CircleShape.cpp"
 $3 = ((($2)) + 8|0); //@line 53 "modules/physics/box2d/CircleShape.cpp"
 HEAPF32[$3>>2] = $0; //@line 53 "modules/physics/box2d/CircleShape.cpp"
 return; //@line 54 "modules/physics/box2d/CircleShape.cpp"
}
function __ZNK4love7physics5box2d11CircleShape8getPointERfS3_($this,$x_o,$y_o) {
 $this = $this|0;
 $x_o = $x_o|0;
 $y_o = $y_o|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 58 "modules/physics/box2d/CircleShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 58 "modules/physics/box2d/CircleShape.cpp"
 $2 = ((($1)) + 12|0); //@line 59 "modules/physics/box2d/CircleShape.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 59 "modules/physics/box2d/CircleShape.cpp"
 $4 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($3)); //@line 59 "modules/physics/box2d/CircleShape.cpp"
 HEAPF32[$x_o>>2] = $4; //@line 59 "modules/physics/box2d/CircleShape.cpp"
 $5 = ((($1)) + 16|0); //@line 60 "modules/physics/box2d/CircleShape.cpp"
 $6 = +HEAPF32[$5>>2]; //@line 60 "modules/physics/box2d/CircleShape.cpp"
 $7 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($6)); //@line 60 "modules/physics/box2d/CircleShape.cpp"
 HEAPF32[$y_o>>2] = $7; //@line 60 "modules/physics/box2d/CircleShape.cpp"
 return; //@line 61 "modules/physics/box2d/CircleShape.cpp"
}
function __ZN4love7physics5box2d11CircleShape8setPointEff($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = sp + 8|0;
 $2 = ((($this)) + 8|0); //@line 65 "modules/physics/box2d/CircleShape.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 65 "modules/physics/box2d/CircleShape.cpp"
 $4 = ((($3)) + 12|0); //@line 66 "modules/physics/box2d/CircleShape.cpp"
 HEAPF32[$1>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$5>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($0,$1); //@line 66 "modules/physics/box2d/CircleShape.cpp"
 $6 = $0; //@line 66 "modules/physics/box2d/CircleShape.cpp"
 $7 = $6; //@line 66 "modules/physics/box2d/CircleShape.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 66 "modules/physics/box2d/CircleShape.cpp"
 $9 = (($6) + 4)|0; //@line 66 "modules/physics/box2d/CircleShape.cpp"
 $10 = $9; //@line 66 "modules/physics/box2d/CircleShape.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 66 "modules/physics/box2d/CircleShape.cpp"
 $12 = $4; //@line 66 "modules/physics/box2d/CircleShape.cpp"
 $13 = $12; //@line 66 "modules/physics/box2d/CircleShape.cpp"
 HEAP32[$13>>2] = $8; //@line 66 "modules/physics/box2d/CircleShape.cpp"
 $14 = (($12) + 4)|0; //@line 66 "modules/physics/box2d/CircleShape.cpp"
 $15 = $14; //@line 66 "modules/physics/box2d/CircleShape.cpp"
 HEAP32[$15>>2] = $11; //@line 66 "modules/physics/box2d/CircleShape.cpp"
 STACKTOP = sp;return; //@line 67 "modules/physics/box2d/CircleShape.cpp"
}
function __ZN4love7physics5box2d7ContactD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (32664); //@line 41 "modules/physics/box2d/Contact.cpp"
 $0 = ((($this)) + 8|0); //@line 47 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 47 "modules/physics/box2d/Contact.cpp"
 $2 = ($1|0)==(0|0); //@line 47 "modules/physics/box2d/Contact.cpp"
 if ($2) {
  return; //@line 43 "modules/physics/box2d/Contact.cpp"
 }
 __THREW__ = 0;
 invoke_vi(3702,($1|0)); //@line 49 "modules/physics/box2d/Contact.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $5 = ___cxa_find_matching_catch_3(0|0)|0;
  $6 = tempRet0;
  ___clang_call_terminate($5); //@line 43 "modules/physics/box2d/Contact.cpp"
  // unreachable; //@line 43 "modules/physics/box2d/Contact.cpp"
 }
 HEAP32[$0>>2] = 0; //@line 50 "modules/physics/box2d/Contact.cpp"
 return; //@line 43 "modules/physics/box2d/Contact.cpp"
}
function __ZN4love7physics5box2d7ContactD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (32664); //@line 41 "modules/physics/box2d/Contact.cpp"
 $0 = ((($this)) + 8|0); //@line 47 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 47 "modules/physics/box2d/Contact.cpp"
 $2 = ($1|0)==(0|0); //@line 47 "modules/physics/box2d/Contact.cpp"
 if ($2) {
  __ZdlPv($this); //@line 41 "modules/physics/box2d/Contact.cpp"
  return; //@line 43 "modules/physics/box2d/Contact.cpp"
 }
 __THREW__ = 0;
 invoke_vi(3702,($1|0)); //@line 49 "modules/physics/box2d/Contact.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $5 = ___cxa_find_matching_catch_3(0|0)|0;
  $6 = tempRet0;
  ___clang_call_terminate($5); //@line 43 "modules/physics/box2d/Contact.cpp"
  // unreachable; //@line 43 "modules/physics/box2d/Contact.cpp"
 }
 HEAP32[$0>>2] = 0; //@line 50 "modules/physics/box2d/Contact.cpp"
 __ZdlPv($this); //@line 41 "modules/physics/box2d/Contact.cpp"
 return; //@line 43 "modules/physics/box2d/Contact.cpp"
}
function __ZN4love7physics5box2d7ContactC2EP9b2Contact($this,$contact) {
 $this = $this|0;
 $contact = $contact|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 34 "modules/physics/box2d/Contact.cpp"
 HEAP32[$this>>2] = (32664); //@line 36 "modules/physics/box2d/Contact.cpp"
 $0 = ((($this)) + 8|0); //@line 35 "modules/physics/box2d/Contact.cpp"
 HEAP32[$0>>2] = $contact; //@line 35 "modules/physics/box2d/Contact.cpp"
 __THREW__ = 0;
 invoke_vii(3703,($contact|0),($this|0)); //@line 37 "modules/physics/box2d/Contact.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  ___resumeException($3|0);
  // unreachable;
 } else {
  return; //@line 38 "modules/physics/box2d/Contact.cpp"
 }
}
function __ZN4love7physics5box2d7Contact10invalidateEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 47 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 47 "modules/physics/box2d/Contact.cpp"
 $2 = ($1|0)==(0|0); //@line 47 "modules/physics/box2d/Contact.cpp"
 if ($2) {
  return; //@line 52 "modules/physics/box2d/Contact.cpp"
 }
 __ZN4love8Memoizer6removeEPv($1); //@line 49 "modules/physics/box2d/Contact.cpp"
 HEAP32[$0>>2] = 0; //@line 50 "modules/physics/box2d/Contact.cpp"
 return; //@line 52 "modules/physics/box2d/Contact.cpp"
}
function __ZN4love7physics5box2d7Contact7isValidEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 56 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 56 "modules/physics/box2d/Contact.cpp"
 $2 = ($1|0)!=(0|0); //@line 56 "modules/physics/box2d/Contact.cpp"
 return ($2|0); //@line 56 "modules/physics/box2d/Contact.cpp"
}
function __ZN4love7physics5box2d7Contact12getPositionsEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, $manifold = 0, $position = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $manifold = sp + 8|0;
 $position = sp;
 (__ZN4love16luax_assert_argcEP9lua_Stateii($L,1,1)|0); //@line 61 "modules/physics/box2d/Contact.cpp"
 $0 = ((($this)) + 8|0); //@line 63 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 63 "modules/physics/box2d/Contact.cpp"
 $2 = ((($1)) + 48|0); //@line 234 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $3 = HEAP32[$2>>2]|0; //@line 234 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $4 = ((($3)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $5 = HEAP32[$4>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $6 = ((($1)) + 52|0); //@line 235 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $7 = HEAP32[$6>>2]|0; //@line 235 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $8 = ((($7)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $9 = HEAP32[$8>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $10 = ((($3)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $11 = HEAP32[$10>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $12 = ((($7)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $13 = HEAP32[$12>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $14 = ((($1)) + 64|0); //@line 239 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $15 = ((($5)) + 12|0); //@line 476 "./libraries/Box2D/Dynamics/b2Body.h"
 $16 = ((($11)) + 8|0); //@line 239 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $17 = +HEAPF32[$16>>2]; //@line 239 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $18 = ((($9)) + 12|0); //@line 476 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = ((($13)) + 8|0); //@line 239 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $20 = +HEAPF32[$19>>2]; //@line 239 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 __ZN15b2WorldManifold10InitializeEPK10b2ManifoldRK11b2TransformfS5_f($manifold,$14,$15,$17,$18,$20); //@line 239 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $21 = HEAP32[$0>>2]|0; //@line 64 "modules/physics/box2d/Contact.cpp"
 $22 = ((($21)) + 124|0); //@line 64 "modules/physics/box2d/Contact.cpp"
 $23 = HEAP32[$22>>2]|0; //@line 64 "modules/physics/box2d/Contact.cpp"
 $24 = ($23|0)>(0); //@line 65 "modules/physics/box2d/Contact.cpp"
 if (!($24)) {
  $26 = $23 << 1; //@line 71 "modules/physics/box2d/Contact.cpp"
  STACKTOP = sp;return ($26|0); //@line 71 "modules/physics/box2d/Contact.cpp"
 }
 $25 = ((($position)) + 4|0); //@line 69 "modules/physics/box2d/Contact.cpp"
 $i$01 = 0;
 while(1) {
  $27 = (((($manifold)) + 8|0) + ($i$01<<3)|0); //@line 67 "modules/physics/box2d/Contact.cpp"
  __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($position,$27); //@line 67 "modules/physics/box2d/Contact.cpp"
  $28 = +HEAPF32[$position>>2]; //@line 68 "modules/physics/box2d/Contact.cpp"
  $29 = $28; //@line 68 "modules/physics/box2d/Contact.cpp"
  _lua_pushnumber($L,$29); //@line 68 "modules/physics/box2d/Contact.cpp"
  $30 = +HEAPF32[$25>>2]; //@line 69 "modules/physics/box2d/Contact.cpp"
  $31 = $30; //@line 69 "modules/physics/box2d/Contact.cpp"
  _lua_pushnumber($L,$31); //@line 69 "modules/physics/box2d/Contact.cpp"
  $32 = (($i$01) + 1)|0; //@line 65 "modules/physics/box2d/Contact.cpp"
  $exitcond = ($32|0)==($23|0); //@line 65 "modules/physics/box2d/Contact.cpp"
  if ($exitcond) {
   break;
  } else {
   $i$01 = $32;
  }
 }
 $26 = $23 << 1; //@line 71 "modules/physics/box2d/Contact.cpp"
 STACKTOP = sp;return ($26|0); //@line 71 "modules/physics/box2d/Contact.cpp"
}
function __ZN4love7physics5box2d7Contact9getNormalEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $manifold = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $manifold = sp;
 (__ZN4love16luax_assert_argcEP9lua_Stateii($L,1,1)|0); //@line 76 "modules/physics/box2d/Contact.cpp"
 $0 = ((($this)) + 8|0); //@line 78 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 78 "modules/physics/box2d/Contact.cpp"
 $2 = ((($1)) + 48|0); //@line 234 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $3 = HEAP32[$2>>2]|0; //@line 234 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $4 = ((($3)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $5 = HEAP32[$4>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $6 = ((($1)) + 52|0); //@line 235 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $7 = HEAP32[$6>>2]|0; //@line 235 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $8 = ((($7)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $9 = HEAP32[$8>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $10 = ((($3)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $11 = HEAP32[$10>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $12 = ((($7)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $13 = HEAP32[$12>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $14 = ((($1)) + 64|0); //@line 239 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $15 = ((($5)) + 12|0); //@line 476 "./libraries/Box2D/Dynamics/b2Body.h"
 $16 = ((($11)) + 8|0); //@line 239 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $17 = +HEAPF32[$16>>2]; //@line 239 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $18 = ((($9)) + 12|0); //@line 476 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = ((($13)) + 8|0); //@line 239 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $20 = +HEAPF32[$19>>2]; //@line 239 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 __ZN15b2WorldManifold10InitializeEPK10b2ManifoldRK11b2TransformfS5_f($manifold,$14,$15,$17,$18,$20); //@line 239 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $21 = +HEAPF32[$manifold>>2]; //@line 79 "modules/physics/box2d/Contact.cpp"
 $22 = $21; //@line 79 "modules/physics/box2d/Contact.cpp"
 _lua_pushnumber($L,$22); //@line 79 "modules/physics/box2d/Contact.cpp"
 $23 = ((($manifold)) + 4|0); //@line 80 "modules/physics/box2d/Contact.cpp"
 $24 = +HEAPF32[$23>>2]; //@line 80 "modules/physics/box2d/Contact.cpp"
 $25 = $24; //@line 80 "modules/physics/box2d/Contact.cpp"
 _lua_pushnumber($L,$25); //@line 80 "modules/physics/box2d/Contact.cpp"
 STACKTOP = sp;return 2; //@line 81 "modules/physics/box2d/Contact.cpp"
}
function __ZNK4love7physics5box2d7Contact11getFrictionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 86 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 86 "modules/physics/box2d/Contact.cpp"
 $2 = ((($1)) + 136|0); //@line 316 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $3 = +HEAPF32[$2>>2]; //@line 316 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 return (+$3); //@line 86 "modules/physics/box2d/Contact.cpp"
}
function __ZNK4love7physics5box2d7Contact14getRestitutionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 91 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 91 "modules/physics/box2d/Contact.cpp"
 $2 = ((($1)) + 140|0); //@line 331 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $3 = +HEAPF32[$2>>2]; //@line 331 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 return (+$3); //@line 91 "modules/physics/box2d/Contact.cpp"
}
function __ZNK4love7physics5box2d7Contact9isEnabledEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 96 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 96 "modules/physics/box2d/Contact.cpp"
 $2 = ((($1)) + 4|0); //@line 256 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $3 = HEAP32[$2>>2]|0; //@line 256 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $4 = $3 & 4; //@line 256 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $5 = ($4|0)!=(0); //@line 256 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 return ($5|0); //@line 96 "modules/physics/box2d/Contact.cpp"
}
function __ZNK4love7physics5box2d7Contact10isTouchingEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 101 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 101 "modules/physics/box2d/Contact.cpp"
 $2 = ((($1)) + 4|0); //@line 261 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $3 = HEAP32[$2>>2]|0; //@line 261 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $4 = $3 & 2; //@line 261 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $5 = ($4|0)!=(0); //@line 261 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 return ($5|0); //@line 101 "modules/physics/box2d/Contact.cpp"
}
function __ZN4love7physics5box2d7Contact11setFrictionEf($this,$friction) {
 $this = $this|0;
 $friction = +$friction;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 106 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 106 "modules/physics/box2d/Contact.cpp"
 $2 = ((($1)) + 136|0); //@line 311 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 HEAPF32[$2>>2] = $friction; //@line 311 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 return; //@line 107 "modules/physics/box2d/Contact.cpp"
}
function __ZN4love7physics5box2d7Contact14setRestitutionEf($this,$restitution) {
 $this = $this|0;
 $restitution = +$restitution;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 111 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 111 "modules/physics/box2d/Contact.cpp"
 $2 = ((($1)) + 140|0); //@line 326 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 HEAPF32[$2>>2] = $restitution; //@line 326 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 return; //@line 112 "modules/physics/box2d/Contact.cpp"
}
function __ZN4love7physics5box2d7Contact10setEnabledEb($this,$enabled) {
 $this = $this|0;
 $enabled = $enabled|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $storemerge$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 116 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 116 "modules/physics/box2d/Contact.cpp"
 $2 = ((($1)) + 4|0); //@line 246 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $3 = HEAP32[$2>>2]|0; //@line 246 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $4 = $3 & -5; //@line 250 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $5 = $3 | 4; //@line 246 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $storemerge$i = $enabled ? $5 : $4; //@line 244 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 HEAP32[$2>>2] = $storemerge$i; //@line 246 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 return; //@line 117 "modules/physics/box2d/Contact.cpp"
}
function __ZN4love7physics5box2d7Contact13resetFrictionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 121 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 121 "modules/physics/box2d/Contact.cpp"
 $2 = ((($1)) + 48|0); //@line 321 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $3 = HEAP32[$2>>2]|0; //@line 321 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $4 = ((($3)) + 16|0); //@line 321 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $5 = +HEAPF32[$4>>2]; //@line 321 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $6 = ((($1)) + 52|0); //@line 321 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $7 = HEAP32[$6>>2]|0; //@line 321 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $8 = ((($7)) + 16|0); //@line 321 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $9 = +HEAPF32[$8>>2]; //@line 321 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $10 = $5 * $9; //@line 39 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $11 = (+Math_sqrt((+$10))); //@line 39 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $12 = ((($1)) + 136|0); //@line 321 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 HEAPF32[$12>>2] = $11; //@line 321 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 return; //@line 122 "modules/physics/box2d/Contact.cpp"
}
function __ZN4love7physics5box2d7Contact16resetRestitutionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 126 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 126 "modules/physics/box2d/Contact.cpp"
 $2 = ((($1)) + 48|0); //@line 336 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $3 = HEAP32[$2>>2]|0; //@line 336 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $4 = ((($3)) + 20|0); //@line 336 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $5 = +HEAPF32[$4>>2]; //@line 336 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $6 = ((($1)) + 52|0); //@line 336 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $7 = HEAP32[$6>>2]|0; //@line 336 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $8 = ((($7)) + 20|0); //@line 336 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $9 = +HEAPF32[$8>>2]; //@line 336 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $10 = $5 > $9; //@line 46 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $11 = $10 ? $5 : $9; //@line 46 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $12 = ((($1)) + 140|0); //@line 336 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 HEAPF32[$12>>2] = $11; //@line 336 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 return; //@line 127 "modules/physics/box2d/Contact.cpp"
}
function __ZN4love7physics5box2d7Contact15setTangentSpeedEf($this,$speed) {
 $this = $this|0;
 $speed = +$speed;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 131 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 131 "modules/physics/box2d/Contact.cpp"
 $2 = ((($1)) + 144|0); //@line 341 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 HEAPF32[$2>>2] = $speed; //@line 341 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 return; //@line 132 "modules/physics/box2d/Contact.cpp"
}
function __ZNK4love7physics5box2d7Contact15getTangentSpeedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 136 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 136 "modules/physics/box2d/Contact.cpp"
 $2 = ((($1)) + 144|0); //@line 346 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $3 = +HEAPF32[$2>>2]; //@line 346 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 return (+$3); //@line 136 "modules/physics/box2d/Contact.cpp"
}
function __ZN4love7physics5box2d7Contact11getChildrenERiS3_($this,$childA,$childB) {
 $this = $this|0;
 $childA = $childA|0;
 $childB = $childB|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 141 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 141 "modules/physics/box2d/Contact.cpp"
 $2 = ((($1)) + 56|0); //@line 291 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $3 = HEAP32[$2>>2]|0; //@line 291 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 HEAP32[$childA>>2] = $3; //@line 141 "modules/physics/box2d/Contact.cpp"
 $4 = ((($1)) + 60|0); //@line 301 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $5 = HEAP32[$4>>2]|0; //@line 301 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 HEAP32[$childB>>2] = $5; //@line 142 "modules/physics/box2d/Contact.cpp"
 return; //@line 143 "modules/physics/box2d/Contact.cpp"
}
function __ZN4love7physics5box2d7Contact11getFixturesERPNS1_7FixtureES5_($this,$fixtureA,$fixtureB) {
 $this = $this|0;
 $fixtureA = $fixtureA|0;
 $fixtureB = $fixtureB|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 8|0); //@line 147 "modules/physics/box2d/Contact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 147 "modules/physics/box2d/Contact.cpp"
 $2 = ((($1)) + 48|0); //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $3 = HEAP32[$2>>2]|0; //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $4 = (__ZN4love8Memoizer4findEPv($3)|0); //@line 147 "modules/physics/box2d/Contact.cpp"
 HEAP32[$fixtureA>>2] = $4; //@line 147 "modules/physics/box2d/Contact.cpp"
 $5 = HEAP32[$0>>2]|0; //@line 148 "modules/physics/box2d/Contact.cpp"
 $6 = ((($5)) + 52|0); //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $7 = HEAP32[$6>>2]|0; //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $8 = (__ZN4love8Memoizer4findEPv($7)|0); //@line 148 "modules/physics/box2d/Contact.cpp"
 HEAP32[$fixtureB>>2] = $8; //@line 148 "modules/physics/box2d/Contact.cpp"
 $9 = HEAP32[$fixtureA>>2]|0; //@line 150 "modules/physics/box2d/Contact.cpp"
 $10 = ($9|0)==(0|0); //@line 150 "modules/physics/box2d/Contact.cpp"
 $11 = ($8|0)==(0|0); //@line 150 "modules/physics/box2d/Contact.cpp"
 $or$cond = $11 | $10; //@line 150 "modules/physics/box2d/Contact.cpp"
 if (!($or$cond)) {
  STACKTOP = sp;return; //@line 152 "modules/physics/box2d/Contact.cpp"
 }
 $12 = (___cxa_allocate_exception(16)|0); //@line 151 "modules/physics/box2d/Contact.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($12|0),(444616|0),($vararg_buffer|0)); //@line 151 "modules/physics/box2d/Contact.cpp"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if ($14) {
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  ___cxa_free_exception(($12|0)); //@line 151 "modules/physics/box2d/Contact.cpp"
  ___resumeException($15|0);
  // unreachable;
 } else {
  ___cxa_throw(($12|0),(96|0),(344|0)); //@line 151 "modules/physics/box2d/Contact.cpp"
  // unreachable; //@line 151 "modules/physics/box2d/Contact.cpp"
 }
}
function __ZN4love7physics5box2d13DistanceJointD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 47 "modules/physics/box2d/DistanceJoint.cpp"
 return; //@line 47 "modules/physics/box2d/DistanceJoint.cpp"
}
function __ZN4love7physics5box2d13DistanceJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 47 "modules/physics/box2d/DistanceJoint.cpp"
 __ZdlPv($this); //@line 46 "modules/physics/box2d/DistanceJoint.cpp"
 return; //@line 47 "modules/physics/box2d/DistanceJoint.cpp"
}
function __ZN4love7physics5box2d13DistanceJointC2EPNS1_4BodyES4_ffffb($this,$body1,$body2,$x1,$y1,$x2,$y2,$collideConnected) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 $x1 = +$x1;
 $y1 = +$y1;
 $x2 = +$x2;
 $y2 = +$y2;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $def = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $def = sp + 32|0;
 $0 = sp + 24|0;
 $1 = sp + 16|0;
 $2 = sp + 8|0;
 $3 = sp;
 __ZN4love7physics5box2d5JointC2EPNS1_4BodyES4_($this,$body1,$body2); //@line 36 "modules/physics/box2d/DistanceJoint.cpp"
 HEAP32[$this>>2] = (32688); //@line 38 "modules/physics/box2d/DistanceJoint.cpp"
 $4 = ((($this)) + 28|0); //@line 37 "modules/physics/box2d/DistanceJoint.cpp"
 HEAP32[$4>>2] = 0; //@line 37 "modules/physics/box2d/DistanceJoint.cpp"
 ;HEAP32[$def>>2]=0|0;HEAP32[$def+4>>2]=0|0;HEAP32[$def+8>>2]=0|0;HEAP32[$def+12>>2]=0|0;HEAP8[$def+16>>0]=0|0; //@line 79 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 HEAP32[$def>>2] = 3; //@line 34 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 $5 = ((($def)) + 20|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($def)) + 36|0); //@line 37 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;HEAP32[$5+12>>2]=0|0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$6>>2] = 1.0; //@line 37 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 $7 = ((($def)) + 40|0); //@line 38 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 HEAPF32[$7>>2] = 0.0; //@line 38 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 $8 = ((($def)) + 44|0); //@line 39 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 HEAPF32[$8>>2] = 0.0; //@line 39 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 $9 = ((($body1)) + 8|0); //@line 40 "modules/physics/box2d/DistanceJoint.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 40 "modules/physics/box2d/DistanceJoint.cpp"
 $11 = ((($body2)) + 8|0); //@line 40 "modules/physics/box2d/DistanceJoint.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 40 "modules/physics/box2d/DistanceJoint.cpp"
 HEAPF32[$1>>2] = $x1; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$13>>2] = $y1; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 invoke_vii(3695,($0|0),($1|0)); //@line 40 "modules/physics/box2d/DistanceJoint.cpp"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if ($15) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 43 "modules/physics/box2d/DistanceJoint.cpp"
  ___resumeException($26|0);
  // unreachable;
 }
 HEAPF32[$3>>2] = $x2; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $16 = ((($3)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$16>>2] = $y2; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 invoke_vii(3695,($2|0),($3|0)); //@line 40 "modules/physics/box2d/DistanceJoint.cpp"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if ($18) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 43 "modules/physics/box2d/DistanceJoint.cpp"
  ___resumeException($26|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_viiiii(4212,($def|0),($10|0),($12|0),($0|0),($2|0)); //@line 40 "modules/physics/box2d/DistanceJoint.cpp"
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 43 "modules/physics/box2d/DistanceJoint.cpp"
  ___resumeException($26|0);
  // unreachable;
 }
 $21 = ((($def)) + 16|0); //@line 41 "modules/physics/box2d/DistanceJoint.cpp"
 $22 = $collideConnected&1; //@line 41 "modules/physics/box2d/DistanceJoint.cpp"
 HEAP8[$21>>0] = $22; //@line 41 "modules/physics/box2d/DistanceJoint.cpp"
 __THREW__ = 0;
 $23 = (invoke_iii(3697,($this|0),($def|0))|0); //@line 42 "modules/physics/box2d/DistanceJoint.cpp"
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 43 "modules/physics/box2d/DistanceJoint.cpp"
  ___resumeException($26|0);
  // unreachable;
 } else {
  HEAP32[$4>>2] = $23; //@line 42 "modules/physics/box2d/DistanceJoint.cpp"
  STACKTOP = sp;return; //@line 43 "modules/physics/box2d/DistanceJoint.cpp"
 }
}
function __ZN4love7physics5box2d13DistanceJoint9setLengthEf($this,$length) {
 $this = $this|0;
 $length = +$length;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 51 "modules/physics/box2d/DistanceJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 51 "modules/physics/box2d/DistanceJoint.cpp"
 $2 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($length)); //@line 51 "modules/physics/box2d/DistanceJoint.cpp"
 $3 = ((($1)) + 104|0); //@line 141 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 HEAPF32[$3>>2] = $2; //@line 141 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 return; //@line 52 "modules/physics/box2d/DistanceJoint.cpp"
}
function __ZNK4love7physics5box2d13DistanceJoint9getLengthEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 56 "modules/physics/box2d/DistanceJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 56 "modules/physics/box2d/DistanceJoint.cpp"
 $2 = ((($1)) + 104|0); //@line 146 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 $3 = +HEAPF32[$2>>2]; //@line 146 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 $4 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($3)); //@line 56 "modules/physics/box2d/DistanceJoint.cpp"
 return (+$4); //@line 56 "modules/physics/box2d/DistanceJoint.cpp"
}
function __ZN4love7physics5box2d13DistanceJoint12setFrequencyEf($this,$hz) {
 $this = $this|0;
 $hz = +$hz;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 61 "modules/physics/box2d/DistanceJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 61 "modules/physics/box2d/DistanceJoint.cpp"
 $2 = ((($1)) + 68|0); //@line 151 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 HEAPF32[$2>>2] = $hz; //@line 151 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 return; //@line 62 "modules/physics/box2d/DistanceJoint.cpp"
}
function __ZNK4love7physics5box2d13DistanceJoint12getFrequencyEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 66 "modules/physics/box2d/DistanceJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 66 "modules/physics/box2d/DistanceJoint.cpp"
 $2 = ((($1)) + 68|0); //@line 156 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 $3 = +HEAPF32[$2>>2]; //@line 156 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 return (+$3); //@line 66 "modules/physics/box2d/DistanceJoint.cpp"
}
function __ZN4love7physics5box2d13DistanceJoint15setDampingRatioEf($this,$d) {
 $this = $this|0;
 $d = +$d;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 71 "modules/physics/box2d/DistanceJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 71 "modules/physics/box2d/DistanceJoint.cpp"
 $2 = ((($1)) + 72|0); //@line 161 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 HEAPF32[$2>>2] = $d; //@line 161 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 return; //@line 72 "modules/physics/box2d/DistanceJoint.cpp"
}
function __ZNK4love7physics5box2d13DistanceJoint15getDampingRatioEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 76 "modules/physics/box2d/DistanceJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 76 "modules/physics/box2d/DistanceJoint.cpp"
 $2 = ((($1)) + 72|0); //@line 166 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 $3 = +HEAPF32[$2>>2]; //@line 166 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 return (+$3); //@line 76 "modules/physics/box2d/DistanceJoint.cpp"
}
function __ZN4love7physics5box2d9EdgeShapeD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5ShapeD2Ev($this); //@line 44 "modules/physics/box2d/EdgeShape.cpp"
 return; //@line 44 "modules/physics/box2d/EdgeShape.cpp"
}
function __ZN4love7physics5box2d9EdgeShapeD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5ShapeD2Ev($this); //@line 44 "modules/physics/box2d/EdgeShape.cpp"
 __ZdlPv($this); //@line 43 "modules/physics/box2d/EdgeShape.cpp"
 return; //@line 44 "modules/physics/box2d/EdgeShape.cpp"
}
function __ZN4love7physics5box2d9EdgeShapeC2EP11b2EdgeShapeb($this,$e,$own) {
 $this = $this|0;
 $e = $e|0;
 $own = $own|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5ShapeC2EP7b2Shapeb($this,$e,$own); //@line 38 "modules/physics/box2d/EdgeShape.cpp"
 HEAP32[$this>>2] = (32712); //@line 39 "modules/physics/box2d/EdgeShape.cpp"
 return; //@line 40 "modules/physics/box2d/EdgeShape.cpp"
}
function __ZN4love7physics5box2d9EdgeShape9getPointsEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $v1 = 0, $v2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v1 = sp + 8|0;
 $v2 = sp;
 $0 = ((($this)) + 8|0); //@line 48 "modules/physics/box2d/EdgeShape.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 48 "modules/physics/box2d/EdgeShape.cpp"
 $2 = ((($1)) + 12|0); //@line 49 "modules/physics/box2d/EdgeShape.cpp"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($v1,$2); //@line 49 "modules/physics/box2d/EdgeShape.cpp"
 $3 = ((($1)) + 20|0); //@line 50 "modules/physics/box2d/EdgeShape.cpp"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($v2,$3); //@line 50 "modules/physics/box2d/EdgeShape.cpp"
 $4 = +HEAPF32[$v1>>2]; //@line 51 "modules/physics/box2d/EdgeShape.cpp"
 $5 = $4; //@line 51 "modules/physics/box2d/EdgeShape.cpp"
 _lua_pushnumber($L,$5); //@line 51 "modules/physics/box2d/EdgeShape.cpp"
 $6 = ((($v1)) + 4|0); //@line 52 "modules/physics/box2d/EdgeShape.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 52 "modules/physics/box2d/EdgeShape.cpp"
 $8 = $7; //@line 52 "modules/physics/box2d/EdgeShape.cpp"
 _lua_pushnumber($L,$8); //@line 52 "modules/physics/box2d/EdgeShape.cpp"
 $9 = +HEAPF32[$v2>>2]; //@line 53 "modules/physics/box2d/EdgeShape.cpp"
 $10 = $9; //@line 53 "modules/physics/box2d/EdgeShape.cpp"
 _lua_pushnumber($L,$10); //@line 53 "modules/physics/box2d/EdgeShape.cpp"
 $11 = ((($v2)) + 4|0); //@line 54 "modules/physics/box2d/EdgeShape.cpp"
 $12 = +HEAPF32[$11>>2]; //@line 54 "modules/physics/box2d/EdgeShape.cpp"
 $13 = $12; //@line 54 "modules/physics/box2d/EdgeShape.cpp"
 _lua_pushnumber($L,$13); //@line 54 "modules/physics/box2d/EdgeShape.cpp"
 STACKTOP = sp;return 4; //@line 55 "modules/physics/box2d/EdgeShape.cpp"
}
function __ZN4love7physics5box2d7FixtureD2Ev($this) {
 $this = $this|0;
 var $$pr$pre = 0, $$pr3 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (32736); //@line 67 "modules/physics/box2d/Fixture.cpp"
 $0 = ((($this)) + 12|0); //@line 68 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 68 "modules/physics/box2d/Fixture.cpp"
 $2 = ($1|0)==(0|0); //@line 68 "modules/physics/box2d/Fixture.cpp"
 if ($2) {
  return; //@line 72 "modules/physics/box2d/Fixture.cpp"
 }
 $3 = HEAP32[$1>>2]|0; //@line 69 "modules/physics/box2d/Fixture.cpp"
 $4 = ($3|0)==(0|0); //@line 69 "modules/physics/box2d/Fixture.cpp"
 if ($4) {
  $$pr3 = $1;
 } else {
  $5 = HEAP32[$3>>2]|0; //@line 69 "modules/physics/box2d/Fixture.cpp"
  $6 = ((($5)) + 4|0); //@line 69 "modules/physics/box2d/Fixture.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 69 "modules/physics/box2d/Fixture.cpp"
  FUNCTION_TABLE_vi[$7 & 8191]($3); //@line 69 "modules/physics/box2d/Fixture.cpp"
  $$pr$pre = HEAP32[$0>>2]|0; //@line 71 "modules/physics/box2d/Fixture.cpp"
  $8 = ($$pr$pre|0)==(0|0); //@line 71 "modules/physics/box2d/Fixture.cpp"
  if ($8) {
   return; //@line 72 "modules/physics/box2d/Fixture.cpp"
  } else {
   $$pr3 = $$pr$pre;
  }
 }
 __ZdlPv($$pr3); //@line 71 "modules/physics/box2d/Fixture.cpp"
 return; //@line 72 "modules/physics/box2d/Fixture.cpp"
}
function __ZN4love7physics5box2d7FixtureD0Ev($this) {
 $this = $this|0;
 var $$pr$pre$i = 0, $$pr3$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (32736); //@line 67 "modules/physics/box2d/Fixture.cpp"
 $0 = ((($this)) + 12|0); //@line 68 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 68 "modules/physics/box2d/Fixture.cpp"
 $2 = ($1|0)==(0|0); //@line 68 "modules/physics/box2d/Fixture.cpp"
 do {
  if (!($2)) {
   $3 = HEAP32[$1>>2]|0; //@line 69 "modules/physics/box2d/Fixture.cpp"
   $4 = ($3|0)==(0|0); //@line 69 "modules/physics/box2d/Fixture.cpp"
   if ($4) {
    $$pr3$i = $1;
   } else {
    $5 = HEAP32[$3>>2]|0; //@line 69 "modules/physics/box2d/Fixture.cpp"
    $6 = ((($5)) + 4|0); //@line 69 "modules/physics/box2d/Fixture.cpp"
    $7 = HEAP32[$6>>2]|0; //@line 69 "modules/physics/box2d/Fixture.cpp"
    FUNCTION_TABLE_vi[$7 & 8191]($3); //@line 69 "modules/physics/box2d/Fixture.cpp"
    $$pr$pre$i = HEAP32[$0>>2]|0; //@line 71 "modules/physics/box2d/Fixture.cpp"
    $8 = ($$pr$pre$i|0)==(0|0); //@line 71 "modules/physics/box2d/Fixture.cpp"
    if ($8) {
     break;
    } else {
     $$pr3$i = $$pr$pre$i;
    }
   }
   __ZdlPv($$pr3$i); //@line 71 "modules/physics/box2d/Fixture.cpp"
  }
 } while(0);
 __ZdlPv($this); //@line 67 "modules/physics/box2d/Fixture.cpp"
 return; //@line 72 "modules/physics/box2d/Fixture.cpp"
}
function __ZN4love7physics5box2d7FixtureC2EPNS1_4BodyEPNS1_5ShapeEf($this,$body,$shape,$density) {
 $this = $this|0;
 $body = $body|0;
 $shape = $shape|0;
 $density = +$density;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $def = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $def = sp;
 __ZN4love6ObjectC2Ev($this); //@line 40 "modules/physics/box2d/Fixture.cpp"
 HEAP32[$this>>2] = (32736); //@line 43 "modules/physics/box2d/Fixture.cpp"
 $0 = ((($this)) + 8|0); //@line 41 "modules/physics/box2d/Fixture.cpp"
 HEAP32[$0>>2] = $body; //@line 41 "modules/physics/box2d/Fixture.cpp"
 $1 = ((($this)) + 16|0); //@line 42 "modules/physics/box2d/Fixture.cpp"
 HEAP32[$1>>2] = 0; //@line 42 "modules/physics/box2d/Fixture.cpp"
 __THREW__ = 0;
 $2 = (invoke_ii(3348,4)|0); //@line 44 "modules/physics/box2d/Fixture.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $30 = ___cxa_find_matching_catch_2()|0;
  $31 = tempRet0;
  ___resumeException($30|0);
  // unreachable;
 }
 $5 = ((($this)) + 12|0); //@line 44 "modules/physics/box2d/Fixture.cpp"
 HEAP32[$5>>2] = $2; //@line 44 "modules/physics/box2d/Fixture.cpp"
 HEAP32[$2>>2] = 0; //@line 45 "modules/physics/box2d/Fixture.cpp"
 $6 = ((($def)) + 22|0); //@line 36 "./libraries/Box2D/Dynamics/b2Fixture.h"
 HEAP16[$6>>1] = 1; //@line 36 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $7 = ((($def)) + 24|0); //@line 37 "./libraries/Box2D/Dynamics/b2Fixture.h"
 HEAP16[$7>>1] = -1; //@line 37 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $8 = ((($def)) + 26|0); //@line 38 "./libraries/Box2D/Dynamics/b2Fixture.h"
 HEAP16[$8>>1] = 0; //@line 38 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $9 = ((($def)) + 4|0); //@line 62 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $10 = ((($def)) + 8|0); //@line 63 "./libraries/Box2D/Dynamics/b2Fixture.h"
 HEAPF32[$10>>2] = 0.20000000298023224; //@line 63 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $11 = ((($def)) + 12|0); //@line 64 "./libraries/Box2D/Dynamics/b2Fixture.h"
 HEAPF32[$11>>2] = 0.0; //@line 64 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $12 = ((($def)) + 16|0); //@line 65 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $13 = ((($def)) + 20|0); //@line 66 "./libraries/Box2D/Dynamics/b2Fixture.h"
 HEAP8[$13>>0] = 0; //@line 66 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $14 = ((($shape)) + 8|0); //@line 47 "modules/physics/box2d/Fixture.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 47 "modules/physics/box2d/Fixture.cpp"
 HEAP32[$def>>2] = $15; //@line 47 "modules/physics/box2d/Fixture.cpp"
 $16 = $2; //@line 48 "modules/physics/box2d/Fixture.cpp"
 HEAP32[$9>>2] = $16; //@line 48 "modules/physics/box2d/Fixture.cpp"
 HEAPF32[$12>>2] = $density; //@line 49 "modules/physics/box2d/Fixture.cpp"
 $17 = ((($body)) + 8|0); //@line 50 "modules/physics/box2d/Fixture.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 50 "modules/physics/box2d/Fixture.cpp"
 __THREW__ = 0;
 $19 = (invoke_iii(4213,($18|0),($def|0))|0); //@line 50 "modules/physics/box2d/Fixture.cpp"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $30 = ___cxa_find_matching_catch_2()|0;
  $31 = tempRet0;
  ___resumeException($30|0);
  // unreachable;
 }
 HEAP32[$1>>2] = $19; //@line 50 "modules/physics/box2d/Fixture.cpp"
 $22 = HEAP32[$this>>2]|0; //@line 51 "modules/physics/box2d/Fixture.cpp"
 $23 = ((($22)) + 8|0); //@line 51 "modules/physics/box2d/Fixture.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 51 "modules/physics/box2d/Fixture.cpp"
 __THREW__ = 0;
 invoke_vi($24|0,($this|0)); //@line 51 "modules/physics/box2d/Fixture.cpp"
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if ($26) {
  $30 = ___cxa_find_matching_catch_2()|0;
  $31 = tempRet0;
  ___resumeException($30|0);
  // unreachable;
 }
 $27 = HEAP32[$1>>2]|0; //@line 52 "modules/physics/box2d/Fixture.cpp"
 __THREW__ = 0;
 invoke_vii(3703,($27|0),($this|0)); //@line 52 "modules/physics/box2d/Fixture.cpp"
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if ($29) {
  $30 = ___cxa_find_matching_catch_2()|0;
  $31 = tempRet0;
  ___resumeException($30|0);
  // unreachable;
 } else {
  STACKTOP = sp;return; //@line 53 "modules/physics/box2d/Fixture.cpp"
 }
}
function __ZNK4love7physics5box2d7Fixture7getTypeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = ((($this)) + 16|0); //@line 76 "modules/physics/box2d/Fixture.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 76 "modules/physics/box2d/Fixture.cpp"
 $3 = ((($2)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $4 = HEAP32[$3>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 __ZN4love7physics5box2d5ShapeC2EP7b2Shapeb($0,$4,0); //@line 76 "modules/physics/box2d/Fixture.cpp"
 __THREW__ = 0;
 $5 = (invoke_ii(4214,($0|0))|0); //@line 76 "modules/physics/box2d/Fixture.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $8 = ___cxa_find_matching_catch_2()|0;
  $9 = tempRet0;
  __ZN4love7physics5box2d5ShapeD2Ev($0); //@line 76 "modules/physics/box2d/Fixture.cpp"
  ___resumeException($8|0);
  // unreachable;
 } else {
  __ZN4love7physics5box2d5ShapeD2Ev($0); //@line 76 "modules/physics/box2d/Fixture.cpp"
  STACKTOP = sp;return ($5|0); //@line 76 "modules/physics/box2d/Fixture.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Fixture11setFrictionEf($this,$friction) {
 $this = $this|0;
 $friction = +$friction;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 81 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 81 "modules/physics/box2d/Fixture.cpp"
 $2 = ((($1)) + 16|0); //@line 311 "./libraries/Box2D/Dynamics/b2Fixture.h"
 HEAPF32[$2>>2] = $friction; //@line 311 "./libraries/Box2D/Dynamics/b2Fixture.h"
 return; //@line 82 "modules/physics/box2d/Fixture.cpp"
}
function __ZN4love7physics5box2d7Fixture14setRestitutionEf($this,$restitution) {
 $this = $this|0;
 $restitution = +$restitution;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 86 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 86 "modules/physics/box2d/Fixture.cpp"
 $2 = ((($1)) + 20|0); //@line 321 "./libraries/Box2D/Dynamics/b2Fixture.h"
 HEAPF32[$2>>2] = $restitution; //@line 321 "./libraries/Box2D/Dynamics/b2Fixture.h"
 return; //@line 87 "modules/physics/box2d/Fixture.cpp"
}
function __ZN4love7physics5box2d7Fixture10setDensityEf($this,$density) {
 $this = $this|0;
 $density = +$density;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 91 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 91 "modules/physics/box2d/Fixture.cpp"
 $2 = (HEAPF32[tempDoublePtr>>2]=$density,HEAP32[tempDoublePtr>>2]|0); //@line 295 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $3 = $2 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $4 = ($3|0)!=(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $5 = $density >= 0.0; //@line 295 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $6 = $5 & $4; //@line 295 "./libraries/Box2D/Dynamics/b2Fixture.h"
 __Z10loveAssertbPKc($6,444747); //@line 295 "./libraries/Box2D/Dynamics/b2Fixture.h"
 HEAPF32[$1>>2] = $density; //@line 296 "./libraries/Box2D/Dynamics/b2Fixture.h"
 return; //@line 92 "modules/physics/box2d/Fixture.cpp"
}
function __ZN4love7physics5box2d7Fixture9setSensorEb($this,$sensor) {
 $this = $this|0;
 $sensor = $sensor|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 96 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 96 "modules/physics/box2d/Fixture.cpp"
 __ZN9b2Fixture9SetSensorEb($1,$sensor); //@line 96 "modules/physics/box2d/Fixture.cpp"
 return; //@line 97 "modules/physics/box2d/Fixture.cpp"
}
function __ZNK4love7physics5box2d7Fixture11getFrictionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 101 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 101 "modules/physics/box2d/Fixture.cpp"
 $2 = ((($1)) + 16|0); //@line 306 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $3 = +HEAPF32[$2>>2]; //@line 306 "./libraries/Box2D/Dynamics/b2Fixture.h"
 return (+$3); //@line 101 "modules/physics/box2d/Fixture.cpp"
}
function __ZNK4love7physics5box2d7Fixture14getRestitutionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 106 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 106 "modules/physics/box2d/Fixture.cpp"
 $2 = ((($1)) + 20|0); //@line 316 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $3 = +HEAPF32[$2>>2]; //@line 316 "./libraries/Box2D/Dynamics/b2Fixture.h"
 return (+$3); //@line 106 "modules/physics/box2d/Fixture.cpp"
}
function __ZNK4love7physics5box2d7Fixture10getDensityEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 111 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 111 "modules/physics/box2d/Fixture.cpp"
 $2 = +HEAPF32[$1>>2]; //@line 301 "./libraries/Box2D/Dynamics/b2Fixture.h"
 return (+$2); //@line 111 "modules/physics/box2d/Fixture.cpp"
}
function __ZNK4love7physics5box2d7Fixture8isSensorEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 116 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 116 "modules/physics/box2d/Fixture.cpp"
 $2 = ((($1)) + 38|0); //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $3 = HEAP8[$2>>0]|0; //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $4 = ($3<<24>>24)!=(0); //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
 return ($4|0); //@line 116 "modules/physics/box2d/Fixture.cpp"
}
function __ZNK4love7physics5box2d7Fixture7getBodyEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 121 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 121 "modules/physics/box2d/Fixture.cpp"
 return ($1|0); //@line 121 "modules/physics/box2d/Fixture.cpp"
}
function __ZNK4love7physics5box2d7Fixture8getShapeEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 126 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 129 "modules/physics/box2d/Fixture.cpp"
 $2 = ((($1)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $3 = HEAP32[$2>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $4 = ($3|0)==(0|0); //@line 126 "modules/physics/box2d/Fixture.cpp"
 if ($4) {
  $$0 = 0;
  return ($$0|0); //@line 130 "modules/physics/box2d/Fixture.cpp"
 }
 $5 = (__Znwj(16)|0); //@line 129 "modules/physics/box2d/Fixture.cpp"
 $6 = HEAP32[$2>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 __THREW__ = 0;
 invoke_viii(4215,($5|0),($6|0),0); //@line 129 "modules/physics/box2d/Fixture.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  __ZdlPv($5); //@line 129 "modules/physics/box2d/Fixture.cpp"
  ___resumeException($9|0);
  // unreachable;
 } else {
  $$0 = $5;
  return ($$0|0); //@line 130 "modules/physics/box2d/Fixture.cpp"
 }
 return (0)|0;
}
function __ZNK4love7physics5box2d7Fixture7isValidEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 134 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 134 "modules/physics/box2d/Fixture.cpp"
 $2 = ($1|0)!=(0|0); //@line 134 "modules/physics/box2d/Fixture.cpp"
 return ($2|0); //@line 134 "modules/physics/box2d/Fixture.cpp"
}
function __ZN4love7physics5box2d7Fixture13setFilterDataEPi($this,$v) {
 $this = $this|0;
 $v = $v|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f = sp;
 $0 = ((($f)) + 2|0); //@line 37 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $1 = ((($f)) + 4|0); //@line 38 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $2 = HEAP32[$v>>2]|0; //@line 140 "modules/physics/box2d/Fixture.cpp"
 $3 = $2&65535; //@line 140 "modules/physics/box2d/Fixture.cpp"
 HEAP16[$f>>1] = $3; //@line 140 "modules/physics/box2d/Fixture.cpp"
 $4 = ((($v)) + 4|0); //@line 141 "modules/physics/box2d/Fixture.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 141 "modules/physics/box2d/Fixture.cpp"
 $6 = $5&65535; //@line 141 "modules/physics/box2d/Fixture.cpp"
 HEAP16[$0>>1] = $6; //@line 141 "modules/physics/box2d/Fixture.cpp"
 $7 = ((($v)) + 8|0); //@line 142 "modules/physics/box2d/Fixture.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 142 "modules/physics/box2d/Fixture.cpp"
 $9 = $8&65535; //@line 142 "modules/physics/box2d/Fixture.cpp"
 HEAP16[$1>>1] = $9; //@line 142 "modules/physics/box2d/Fixture.cpp"
 $10 = ((($this)) + 16|0); //@line 143 "modules/physics/box2d/Fixture.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 143 "modules/physics/box2d/Fixture.cpp"
 __ZN9b2Fixture13SetFilterDataERK8b2Filter($11,$f); //@line 143 "modules/physics/box2d/Fixture.cpp"
 STACKTOP = sp;return; //@line 144 "modules/physics/box2d/Fixture.cpp"
}
function __ZN4love7physics5box2d7Fixture13getFilterDataEPi($this,$v) {
 $this = $this|0;
 $v = $v|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 148 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 148 "modules/physics/box2d/Fixture.cpp"
 $2 = ((($1)) + 32|0); //@line 148 "modules/physics/box2d/Fixture.cpp"
 $3 = HEAP16[$2>>1]|0; //@line 148 "modules/physics/box2d/Fixture.cpp"
 $4 = ((($1)) + 34|0); //@line 148 "modules/physics/box2d/Fixture.cpp"
 $5 = HEAP16[$4>>1]|0; //@line 148 "modules/physics/box2d/Fixture.cpp"
 $6 = ((($1)) + 36|0); //@line 148 "modules/physics/box2d/Fixture.cpp"
 $7 = HEAP16[$6>>1]|0; //@line 148 "modules/physics/box2d/Fixture.cpp"
 $8 = $3&65535; //@line 149 "modules/physics/box2d/Fixture.cpp"
 HEAP32[$v>>2] = $8; //@line 149 "modules/physics/box2d/Fixture.cpp"
 $9 = $5&65535; //@line 150 "modules/physics/box2d/Fixture.cpp"
 $10 = ((($v)) + 4|0); //@line 150 "modules/physics/box2d/Fixture.cpp"
 HEAP32[$10>>2] = $9; //@line 150 "modules/physics/box2d/Fixture.cpp"
 $11 = $7 << 16 >> 16; //@line 151 "modules/physics/box2d/Fixture.cpp"
 $12 = ((($v)) + 8|0); //@line 151 "modules/physics/box2d/Fixture.cpp"
 HEAP32[$12>>2] = $11; //@line 151 "modules/physics/box2d/Fixture.cpp"
 return; //@line 152 "modules/physics/box2d/Fixture.cpp"
}
function __ZN4love7physics5box2d7Fixture11setCategoryEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $f = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f = sp;
 $0 = ((($this)) + 16|0); //@line 156 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 156 "modules/physics/box2d/Fixture.cpp"
 $2 = ((($1)) + 32|0); //@line 260 "./libraries/Box2D/Dynamics/b2Fixture.h"
 ;HEAP16[$f>>1]=HEAP16[$2>>1]|0;HEAP16[$f+2>>1]=HEAP16[$2+2>>1]|0;HEAP16[$f+4>>1]=HEAP16[$2+4>>1]|0; //@line 156 "modules/physics/box2d/Fixture.cpp"
 $3 = (__ZN4love7physics5box2d7Fixture7getBitsEP9lua_State(0,$L)|0); //@line 157 "modules/physics/box2d/Fixture.cpp"
 HEAP16[$f>>1] = $3; //@line 157 "modules/physics/box2d/Fixture.cpp"
 $4 = HEAP32[$0>>2]|0; //@line 158 "modules/physics/box2d/Fixture.cpp"
 __ZN9b2Fixture13SetFilterDataERK8b2Filter($4,$f); //@line 158 "modules/physics/box2d/Fixture.cpp"
 STACKTOP = sp;return 0; //@line 159 "modules/physics/box2d/Fixture.cpp"
}
function __ZN4love7physics5box2d7Fixture7getBitsEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b$sroa$0$0$lcssa = 0, $b$sroa$0$02 = 0, $i$03 = 0, $phitmp = 0, $vararg_buffer = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_lua_gettop($L)|0); //@line 196 "modules/physics/box2d/Fixture.cpp"
 $1 = ($0|0)<(1); //@line 201 "modules/physics/box2d/Fixture.cpp"
 if ($1) {
  $b$sroa$0$0$lcssa = 0;
  STACKTOP = sp;return ($b$sroa$0$0$lcssa|0); //@line 209 "modules/physics/box2d/Fixture.cpp"
 } else {
  $b$sroa$0$02 = 0;$i$03 = 1;
 }
 while(1) {
  $2 = (_lua_tointeger($L,$i$03)|0); //@line 203 "modules/physics/box2d/Fixture.cpp"
  $3 = (($2) + -1)|0; //@line 203 "modules/physics/box2d/Fixture.cpp"
  $4 = ($3>>>0)>(15); //@line 204 "modules/physics/box2d/Fixture.cpp"
  if ($4) {
   break;
  }
  $10 = 1 << $3; //@line 457 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
  $11 = $10 | $b$sroa$0$02; //@line 60 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
  $12 = (($i$03) + 1)|0; //@line 201 "modules/physics/box2d/Fixture.cpp"
  $13 = ($i$03|0)<($0|0); //@line 201 "modules/physics/box2d/Fixture.cpp"
  if ($13) {
   $b$sroa$0$02 = $11;$i$03 = $12;
  } else {
   $$lcssa = $11;
   label = 2;
   break;
  }
 }
 if ((label|0) == 2) {
  $phitmp = $$lcssa&65535; //@line 199 "modules/physics/box2d/Fixture.cpp"
  $b$sroa$0$0$lcssa = $phitmp;
  STACKTOP = sp;return ($b$sroa$0$0$lcssa|0); //@line 209 "modules/physics/box2d/Fixture.cpp"
 }
 (_luaL_error($L,444785,$vararg_buffer)|0); //@line 205 "modules/physics/box2d/Fixture.cpp"
 $5 = (___cxa_allocate_exception(8)|0); //@line 838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 __THREW__ = 0;
 invoke_vii(3996,($5|0),(444815|0)); //@line 133 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/stdexcept"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $8 = ___cxa_find_matching_catch_2()|0;
  $9 = tempRet0;
  ___cxa_free_exception(($5|0)); //@line 838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
  ___resumeException($8|0);
  // unreachable;
 } else {
  HEAP32[$5>>2] = (157912); //@line 133 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/stdexcept"
  ___cxa_throw(($5|0),(9120|0),(3328|0)); //@line 838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
  // unreachable; //@line 838 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Fixture7setMaskEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $f = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f = sp;
 $0 = ((($this)) + 16|0); //@line 164 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 164 "modules/physics/box2d/Fixture.cpp"
 $2 = ((($1)) + 32|0); //@line 260 "./libraries/Box2D/Dynamics/b2Fixture.h"
 ;HEAP16[$f>>1]=HEAP16[$2>>1]|0;HEAP16[$f+2>>1]=HEAP16[$2+2>>1]|0;HEAP16[$f+4>>1]=HEAP16[$2+4>>1]|0; //@line 164 "modules/physics/box2d/Fixture.cpp"
 $3 = (__ZN4love7physics5box2d7Fixture7getBitsEP9lua_State(0,$L)|0); //@line 165 "modules/physics/box2d/Fixture.cpp"
 $4 = $3&65535; //@line 165 "modules/physics/box2d/Fixture.cpp"
 $5 = $4 ^ 65535; //@line 165 "modules/physics/box2d/Fixture.cpp"
 $6 = $5&65535; //@line 165 "modules/physics/box2d/Fixture.cpp"
 $7 = ((($f)) + 2|0); //@line 165 "modules/physics/box2d/Fixture.cpp"
 HEAP16[$7>>1] = $6; //@line 165 "modules/physics/box2d/Fixture.cpp"
 $8 = HEAP32[$0>>2]|0; //@line 166 "modules/physics/box2d/Fixture.cpp"
 __ZN9b2Fixture13SetFilterDataERK8b2Filter($8,$f); //@line 166 "modules/physics/box2d/Fixture.cpp"
 STACKTOP = sp;return 0; //@line 167 "modules/physics/box2d/Fixture.cpp"
}
function __ZN4love7physics5box2d7Fixture13setGroupIndexEi($this,$index) {
 $this = $this|0;
 $index = $index|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $f = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f = sp;
 $0 = ((($this)) + 16|0); //@line 172 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 172 "modules/physics/box2d/Fixture.cpp"
 $2 = ((($1)) + 32|0); //@line 260 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $3 = HEAPU16[$2>>1]|(HEAPU16[$2+2>>1]<<16); //@line 172 "modules/physics/box2d/Fixture.cpp"
 HEAP32[$f>>2] = $3; //@line 172 "modules/physics/box2d/Fixture.cpp"
 $4 = $index&65535; //@line 173 "modules/physics/box2d/Fixture.cpp"
 $5 = ((($f)) + 4|0); //@line 173 "modules/physics/box2d/Fixture.cpp"
 HEAP16[$5>>1] = $4; //@line 173 "modules/physics/box2d/Fixture.cpp"
 __ZN9b2Fixture13SetFilterDataERK8b2Filter($1,$f); //@line 174 "modules/physics/box2d/Fixture.cpp"
 STACKTOP = sp;return; //@line 175 "modules/physics/box2d/Fixture.cpp"
}
function __ZNK4love7physics5box2d7Fixture13getGroupIndexEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 179 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 179 "modules/physics/box2d/Fixture.cpp"
 $2 = ((($1)) + 36|0); //@line 179 "modules/physics/box2d/Fixture.cpp"
 $3 = HEAP16[$2>>1]|0; //@line 179 "modules/physics/box2d/Fixture.cpp"
 $4 = $3 << 16 >> 16; //@line 180 "modules/physics/box2d/Fixture.cpp"
 return ($4|0); //@line 180 "modules/physics/box2d/Fixture.cpp"
}
function __ZN4love7physics5box2d7Fixture11getCategoryEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 185 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 185 "modules/physics/box2d/Fixture.cpp"
 $2 = ((($1)) + 32|0); //@line 185 "modules/physics/box2d/Fixture.cpp"
 $3 = HEAP16[$2>>1]|0; //@line 185 "modules/physics/box2d/Fixture.cpp"
 $4 = $3&65535; //@line 492 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/bitset"
 $5 = $4 & 1; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $6 = ($5|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($6)) {
  _lua_pushinteger($L,1); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $7 = $4 & 2; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $8 = ($7|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($8)) {
  _lua_pushinteger($L,2); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $9 = $4 & 4; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $10 = ($9|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($10)) {
  _lua_pushinteger($L,3); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $11 = $4 & 8; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $12 = ($11|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($12)) {
  _lua_pushinteger($L,4); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $13 = $4 & 16; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $14 = ($13|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($14)) {
  _lua_pushinteger($L,5); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $15 = $4 & 32; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $16 = ($15|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($16)) {
  _lua_pushinteger($L,6); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $17 = $4 & 64; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $18 = ($17|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($18)) {
  _lua_pushinteger($L,7); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $19 = $4 & 128; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $20 = ($19|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($20)) {
  _lua_pushinteger($L,8); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $21 = $4 & 256; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $22 = ($21|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($22)) {
  _lua_pushinteger($L,9); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $23 = $4 & 512; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $24 = ($23|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($24)) {
  _lua_pushinteger($L,10); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $25 = $4 & 1024; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $26 = ($25|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($26)) {
  _lua_pushinteger($L,11); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $27 = $4 & 2048; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $28 = ($27|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($28)) {
  _lua_pushinteger($L,12); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $29 = $4 & 4096; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $30 = ($29|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($30)) {
  _lua_pushinteger($L,13); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $31 = $4 & 8192; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $32 = ($31|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($32)) {
  _lua_pushinteger($L,14); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $33 = $4 & 16384; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $34 = ($33|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($34)) {
  _lua_pushinteger($L,15); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $35 = $4 & 32768; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $36 = ($35|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if ($36) {
  $37 = (_llvm_ctpop_i32(($4|0))|0); //@line 802 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  return ($37|0); //@line 185 "modules/physics/box2d/Fixture.cpp"
 }
 _lua_pushinteger($L,16); //@line 220 "modules/physics/box2d/Fixture.cpp"
 $37 = (_llvm_ctpop_i32(($4|0))|0); //@line 802 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 return ($37|0); //@line 185 "modules/physics/box2d/Fixture.cpp"
}
function __ZN4love7physics5box2d7Fixture7getMaskEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 190 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 190 "modules/physics/box2d/Fixture.cpp"
 $2 = ((($1)) + 34|0); //@line 190 "modules/physics/box2d/Fixture.cpp"
 $3 = HEAP16[$2>>1]|0; //@line 190 "modules/physics/box2d/Fixture.cpp"
 $4 = $3&65535; //@line 190 "modules/physics/box2d/Fixture.cpp"
 $5 = $4 ^ 65535; //@line 190 "modules/physics/box2d/Fixture.cpp"
 $6 = $5 & 1; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $7 = ($6|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($7)) {
  _lua_pushinteger($L,1); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $8 = $5 & 2; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $9 = ($8|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($9)) {
  _lua_pushinteger($L,2); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $10 = $5 & 4; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $11 = ($10|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($11)) {
  _lua_pushinteger($L,3); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $12 = $5 & 8; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $13 = ($12|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($13)) {
  _lua_pushinteger($L,4); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $14 = $5 & 16; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $15 = ($14|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($15)) {
  _lua_pushinteger($L,5); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $16 = $5 & 32; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $17 = ($16|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($17)) {
  _lua_pushinteger($L,6); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $18 = $5 & 64; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $19 = ($18|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($19)) {
  _lua_pushinteger($L,7); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $20 = $5 & 128; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $21 = ($20|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($21)) {
  _lua_pushinteger($L,8); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $22 = $5 & 256; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $23 = ($22|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($23)) {
  _lua_pushinteger($L,9); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $24 = $5 & 512; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $25 = ($24|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($25)) {
  _lua_pushinteger($L,10); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $26 = $5 & 1024; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $27 = ($26|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($27)) {
  _lua_pushinteger($L,11); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $28 = $5 & 2048; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $29 = ($28|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($29)) {
  _lua_pushinteger($L,12); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $30 = $5 & 4096; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $31 = ($30|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($31)) {
  _lua_pushinteger($L,13); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $32 = $5 & 8192; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $33 = ($32|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($33)) {
  _lua_pushinteger($L,14); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $34 = $5 & 16384; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $35 = ($34|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if (!($35)) {
  _lua_pushinteger($L,15); //@line 220 "modules/physics/box2d/Fixture.cpp"
 }
 $36 = $5 & 32768; //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 $37 = ($36|0)==(0); //@line 145 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__bit_reference"
 if ($37) {
  $38 = (_llvm_ctpop_i32(($5|0))|0); //@line 802 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  return ($38|0); //@line 190 "modules/physics/box2d/Fixture.cpp"
 }
 _lua_pushinteger($L,16); //@line 220 "modules/physics/box2d/Fixture.cpp"
 $38 = (_llvm_ctpop_i32(($5|0))|0); //@line 802 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 return ($38|0); //@line 190 "modules/physics/box2d/Fixture.cpp"
}
function __ZN4love7physics5box2d7Fixture11setUserDataEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 (__ZN4love16luax_assert_argcEP9lua_Stateii($L,1,1)|0); //@line 228 "modules/physics/box2d/Fixture.cpp"
 $0 = ((($this)) + 12|0); //@line 230 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 230 "modules/physics/box2d/Fixture.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 230 "modules/physics/box2d/Fixture.cpp"
 $3 = ($2|0)==(0|0); //@line 230 "modules/physics/box2d/Fixture.cpp"
 if (!($3)) {
  $4 = HEAP32[$2>>2]|0; //@line 230 "modules/physics/box2d/Fixture.cpp"
  $5 = ((($4)) + 4|0); //@line 230 "modules/physics/box2d/Fixture.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 230 "modules/physics/box2d/Fixture.cpp"
  FUNCTION_TABLE_vi[$6 & 8191]($2); //@line 230 "modules/physics/box2d/Fixture.cpp"
 }
 $7 = (__Znwj(12)|0); //@line 231 "modules/physics/box2d/Fixture.cpp"
 __THREW__ = 0;
 invoke_vii(3493,($7|0),($L|0)); //@line 231 "modules/physics/box2d/Fixture.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $11 = ___cxa_find_matching_catch_2()|0;
  $12 = tempRet0;
  __ZdlPv($7); //@line 231 "modules/physics/box2d/Fixture.cpp"
  ___resumeException($11|0);
  // unreachable;
 } else {
  $10 = HEAP32[$0>>2]|0; //@line 231 "modules/physics/box2d/Fixture.cpp"
  HEAP32[$10>>2] = $7; //@line 231 "modules/physics/box2d/Fixture.cpp"
  return 0; //@line 233 "modules/physics/box2d/Fixture.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d7Fixture11getUserDataEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 238 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 238 "modules/physics/box2d/Fixture.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 238 "modules/physics/box2d/Fixture.cpp"
 $3 = ($2|0)==(0|0); //@line 238 "modules/physics/box2d/Fixture.cpp"
 if ($3) {
  _lua_pushnil($L); //@line 241 "modules/physics/box2d/Fixture.cpp"
  return 1; //@line 243 "modules/physics/box2d/Fixture.cpp"
 } else {
  __ZN4love9Reference4pushEP9lua_State($2,$L); //@line 239 "modules/physics/box2d/Fixture.cpp"
  return 1; //@line 243 "modules/physics/box2d/Fixture.cpp"
 }
 return (0)|0;
}
function __ZNK4love7physics5box2d7Fixture9testPointEff($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 8|0;
 $1 = sp;
 $2 = ((($this)) + 16|0); //@line 248 "modules/physics/box2d/Fixture.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 248 "modules/physics/box2d/Fixture.cpp"
 HEAPF32[$1>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$4>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($0,$1); //@line 248 "modules/physics/box2d/Fixture.cpp"
 $5 = ((($3)) + 12|0); //@line 326 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $6 = HEAP32[$5>>2]|0; //@line 326 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $7 = HEAP32[$6>>2]|0; //@line 326 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $8 = ((($7)) + 16|0); //@line 326 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $9 = HEAP32[$8>>2]|0; //@line 326 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $10 = ((($3)) + 8|0); //@line 326 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $11 = HEAP32[$10>>2]|0; //@line 326 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $12 = ((($11)) + 12|0); //@line 476 "./libraries/Box2D/Dynamics/b2Body.h"
 $13 = (FUNCTION_TABLE_iiii[$9 & 8191]($6,$12,$0)|0); //@line 326 "./libraries/Box2D/Dynamics/b2Fixture.h"
 STACKTOP = sp;return ($13|0); //@line 248 "modules/physics/box2d/Fixture.cpp"
}
function __ZNK4love7physics5box2d7Fixture7rayCastEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $$0 = 0, $0 = 0.0, $1 = 0.0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0;
 var $9 = 0.0, $input = 0, $output = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $input = sp + 12|0;
 $output = sp;
 $0 = (+_luaL_checknumber($L,1)); //@line 253 "modules/physics/box2d/Fixture.cpp"
 $1 = $0; //@line 253 "modules/physics/box2d/Fixture.cpp"
 $2 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($1)); //@line 253 "modules/physics/box2d/Fixture.cpp"
 $3 = (+_luaL_checknumber($L,2)); //@line 254 "modules/physics/box2d/Fixture.cpp"
 $4 = $3; //@line 254 "modules/physics/box2d/Fixture.cpp"
 $5 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($4)); //@line 254 "modules/physics/box2d/Fixture.cpp"
 $6 = (+_luaL_checknumber($L,3)); //@line 255 "modules/physics/box2d/Fixture.cpp"
 $7 = $6; //@line 255 "modules/physics/box2d/Fixture.cpp"
 $8 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($7)); //@line 255 "modules/physics/box2d/Fixture.cpp"
 $9 = (+_luaL_checknumber($L,4)); //@line 256 "modules/physics/box2d/Fixture.cpp"
 $10 = $9; //@line 256 "modules/physics/box2d/Fixture.cpp"
 $11 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($10)); //@line 256 "modules/physics/box2d/Fixture.cpp"
 $12 = (+_luaL_checknumber($L,5)); //@line 257 "modules/physics/box2d/Fixture.cpp"
 $13 = $12; //@line 257 "modules/physics/box2d/Fixture.cpp"
 $14 = (+_luaL_optnumber($L,6,1.0)); //@line 258 "modules/physics/box2d/Fixture.cpp"
 $15 = (~~(($14))); //@line 258 "modules/physics/box2d/Fixture.cpp"
 $16 = (($15) + -1)|0; //@line 258 "modules/physics/box2d/Fixture.cpp"
 HEAPF32[$input>>2] = $2; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $17 = ((($input)) + 4|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$17>>2] = $5; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $18 = ((($input)) + 8|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$18>>2] = $8; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($input)) + 12|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$19>>2] = $11; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $20 = ((($input)) + 16|0); //@line 262 "modules/physics/box2d/Fixture.cpp"
 HEAPF32[$20>>2] = $13; //@line 262 "modules/physics/box2d/Fixture.cpp"
 $21 = ((($this)) + 16|0); //@line 264 "modules/physics/box2d/Fixture.cpp"
 $22 = HEAP32[$21>>2]|0; //@line 264 "modules/physics/box2d/Fixture.cpp"
 $23 = ((($22)) + 12|0); //@line 331 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $24 = HEAP32[$23>>2]|0; //@line 331 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $25 = HEAP32[$24>>2]|0; //@line 331 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $26 = ((($25)) + 20|0); //@line 331 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $27 = HEAP32[$26>>2]|0; //@line 331 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $28 = ((($22)) + 8|0); //@line 331 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $29 = HEAP32[$28>>2]|0; //@line 331 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $30 = ((($29)) + 12|0); //@line 476 "./libraries/Box2D/Dynamics/b2Body.h"
 $31 = (FUNCTION_TABLE_iiiiii[$27 & 8191]($24,$output,$input,$30,$16)|0); //@line 331 "./libraries/Box2D/Dynamics/b2Fixture.h"
 if (!($31)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 270 "modules/physics/box2d/Fixture.cpp"
 }
 $32 = +HEAPF32[$output>>2]; //@line 266 "modules/physics/box2d/Fixture.cpp"
 $33 = $32; //@line 266 "modules/physics/box2d/Fixture.cpp"
 _lua_pushnumber($L,$33); //@line 266 "modules/physics/box2d/Fixture.cpp"
 $34 = ((($output)) + 4|0); //@line 267 "modules/physics/box2d/Fixture.cpp"
 $35 = +HEAPF32[$34>>2]; //@line 267 "modules/physics/box2d/Fixture.cpp"
 $36 = $35; //@line 267 "modules/physics/box2d/Fixture.cpp"
 _lua_pushnumber($L,$36); //@line 267 "modules/physics/box2d/Fixture.cpp"
 $37 = ((($output)) + 8|0); //@line 268 "modules/physics/box2d/Fixture.cpp"
 $38 = +HEAPF32[$37>>2]; //@line 268 "modules/physics/box2d/Fixture.cpp"
 $39 = $38; //@line 268 "modules/physics/box2d/Fixture.cpp"
 _lua_pushnumber($L,$39); //@line 268 "modules/physics/box2d/Fixture.cpp"
 $$0 = 3;
 STACKTOP = sp;return ($$0|0); //@line 270 "modules/physics/box2d/Fixture.cpp"
}
function __ZNK4love7physics5box2d7Fixture14getBoundingBoxEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0.0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $box = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $box = sp + 16|0;
 $0 = sp;
 $1 = (+_luaL_optnumber($L,1,1.0)); //@line 274 "modules/physics/box2d/Fixture.cpp"
 $2 = (~~(($1))); //@line 274 "modules/physics/box2d/Fixture.cpp"
 $3 = (($2) + -1)|0; //@line 274 "modules/physics/box2d/Fixture.cpp"
 $4 = ((($this)) + 16|0); //@line 275 "modules/physics/box2d/Fixture.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 275 "modules/physics/box2d/Fixture.cpp"
 $6 = ($2|0)>(0); //@line 341 "./libraries/Box2D/Dynamics/b2Fixture.h"
 if ($6) {
  $7 = ((($5)) + 28|0); //@line 341 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $8 = HEAP32[$7>>2]|0; //@line 341 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $9 = ($8|0)>=($2|0); //@line 341 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $10 = $9;
 } else {
  $10 = 0;
 }
 __Z10loveAssertbPKc($10,444848); //@line 341 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $11 = ((($5)) + 24|0); //@line 342 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $12 = HEAP32[$11>>2]|0; //@line 342 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $13 = (($12) + (($3*28)|0)|0); //@line 342 "./libraries/Box2D/Dynamics/b2Fixture.h"
 ;HEAP32[$box>>2]=HEAP32[$13>>2]|0;HEAP32[$box+4>>2]=HEAP32[$13+4>>2]|0;HEAP32[$box+8>>2]=HEAP32[$13+8>>2]|0;HEAP32[$box+12>>2]=HEAP32[$13+12>>2]|0; //@line 275 "modules/physics/box2d/Fixture.cpp"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2AABB($0,$box); //@line 276 "modules/physics/box2d/Fixture.cpp"
 ;HEAP32[$box>>2]=HEAP32[$0>>2]|0;HEAP32[$box+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$box+8>>2]=HEAP32[$0+8>>2]|0;HEAP32[$box+12>>2]=HEAP32[$0+12>>2]|0; //@line 276 "modules/physics/box2d/Fixture.cpp"
 $14 = +HEAPF32[$box>>2]; //@line 277 "modules/physics/box2d/Fixture.cpp"
 $15 = $14; //@line 277 "modules/physics/box2d/Fixture.cpp"
 _lua_pushnumber($L,$15); //@line 277 "modules/physics/box2d/Fixture.cpp"
 $16 = ((($box)) + 4|0); //@line 278 "modules/physics/box2d/Fixture.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 278 "modules/physics/box2d/Fixture.cpp"
 $18 = $17; //@line 278 "modules/physics/box2d/Fixture.cpp"
 _lua_pushnumber($L,$18); //@line 278 "modules/physics/box2d/Fixture.cpp"
 $19 = ((($box)) + 8|0); //@line 279 "modules/physics/box2d/Fixture.cpp"
 $20 = +HEAPF32[$19>>2]; //@line 279 "modules/physics/box2d/Fixture.cpp"
 $21 = $20; //@line 279 "modules/physics/box2d/Fixture.cpp"
 _lua_pushnumber($L,$21); //@line 279 "modules/physics/box2d/Fixture.cpp"
 $22 = ((($box)) + 12|0); //@line 280 "modules/physics/box2d/Fixture.cpp"
 $23 = +HEAPF32[$22>>2]; //@line 280 "modules/physics/box2d/Fixture.cpp"
 $24 = $23; //@line 280 "modules/physics/box2d/Fixture.cpp"
 _lua_pushnumber($L,$24); //@line 280 "modules/physics/box2d/Fixture.cpp"
 STACKTOP = sp;return 4; //@line 281 "modules/physics/box2d/Fixture.cpp"
}
function __ZNK4love7physics5box2d7Fixture11getMassDataEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0.0, $center = 0;
 var $data = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $data = sp + 8|0;
 $center = sp;
 $0 = ((($this)) + 16|0); //@line 287 "modules/physics/box2d/Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 287 "modules/physics/box2d/Fixture.cpp"
 $2 = ((($1)) + 12|0); //@line 336 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $3 = HEAP32[$2>>2]|0; //@line 336 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $4 = HEAP32[$3>>2]|0; //@line 336 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $5 = ((($4)) + 28|0); //@line 336 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $6 = HEAP32[$5>>2]|0; //@line 336 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $7 = +HEAPF32[$1>>2]; //@line 336 "./libraries/Box2D/Dynamics/b2Fixture.h"
 FUNCTION_TABLE_viid[$6 & 8191]($3,$data,$7); //@line 336 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $8 = ((($data)) + 4|0); //@line 288 "modules/physics/box2d/Fixture.cpp"
 __ZN4love7physics5box2d7Physics7scaleUpERK6b2Vec2($center,$8); //@line 288 "modules/physics/box2d/Fixture.cpp"
 $9 = +HEAPF32[$center>>2]; //@line 289 "modules/physics/box2d/Fixture.cpp"
 $10 = $9; //@line 289 "modules/physics/box2d/Fixture.cpp"
 _lua_pushnumber($L,$10); //@line 289 "modules/physics/box2d/Fixture.cpp"
 $11 = ((($center)) + 4|0); //@line 290 "modules/physics/box2d/Fixture.cpp"
 $12 = +HEAPF32[$11>>2]; //@line 290 "modules/physics/box2d/Fixture.cpp"
 $13 = $12; //@line 290 "modules/physics/box2d/Fixture.cpp"
 _lua_pushnumber($L,$13); //@line 290 "modules/physics/box2d/Fixture.cpp"
 $14 = +HEAPF32[$data>>2]; //@line 291 "modules/physics/box2d/Fixture.cpp"
 $15 = $14; //@line 291 "modules/physics/box2d/Fixture.cpp"
 _lua_pushnumber($L,$15); //@line 291 "modules/physics/box2d/Fixture.cpp"
 $16 = ((($data)) + 12|0); //@line 292 "modules/physics/box2d/Fixture.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 292 "modules/physics/box2d/Fixture.cpp"
 $18 = $17; //@line 292 "modules/physics/box2d/Fixture.cpp"
 _lua_pushnumber($L,$18); //@line 292 "modules/physics/box2d/Fixture.cpp"
 STACKTOP = sp;return 4; //@line 293 "modules/physics/box2d/Fixture.cpp"
}
function __ZN4love7physics5box2d7Fixture7destroyEb($this,$implicit) {
 $this = $this|0;
 $implicit = $implicit|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = ((($this)) + 8|0); //@line 298 "modules/physics/box2d/Fixture.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 298 "modules/physics/box2d/Fixture.cpp"
 $3 = ((($2)) + 12|0); //@line 298 "modules/physics/box2d/Fixture.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 298 "modules/physics/box2d/Fixture.cpp"
 $5 = ((($4)) + 20|0); //@line 298 "modules/physics/box2d/Fixture.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 298 "modules/physics/box2d/Fixture.cpp"
 $7 = ((($6)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $8 = HEAP32[$7>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $9 = $8 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $10 = ($9|0)==(0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 if ($10) {
  $$pre = ((($this)) + 16|0); //@line 308 "modules/physics/box2d/Fixture.cpp"
  if (!($implicit)) {
   $26 = HEAP32[$$pre>>2]|0; //@line 306 "modules/physics/box2d/Fixture.cpp"
   $27 = ($26|0)==(0|0); //@line 306 "modules/physics/box2d/Fixture.cpp"
   if (!($27)) {
    $28 = ((($2)) + 8|0); //@line 307 "modules/physics/box2d/Fixture.cpp"
    $29 = HEAP32[$28>>2]|0; //@line 307 "modules/physics/box2d/Fixture.cpp"
    __ZN6b2Body14DestroyFixtureEP9b2Fixture($29,$26); //@line 307 "modules/physics/box2d/Fixture.cpp"
   }
  }
  $30 = HEAP32[$$pre>>2]|0; //@line 308 "modules/physics/box2d/Fixture.cpp"
  __ZN4love8Memoizer6removeEPv($30); //@line 308 "modules/physics/box2d/Fixture.cpp"
  HEAP32[$$pre>>2] = 0; //@line 309 "modules/physics/box2d/Fixture.cpp"
  $31 = HEAP32[$this>>2]|0; //@line 312 "modules/physics/box2d/Fixture.cpp"
  $32 = ((($31)) + 12|0); //@line 312 "modules/physics/box2d/Fixture.cpp"
  $33 = HEAP32[$32>>2]|0; //@line 312 "modules/physics/box2d/Fixture.cpp"
  FUNCTION_TABLE_vi[$33 & 8191]($this); //@line 312 "modules/physics/box2d/Fixture.cpp"
  STACKTOP = sp;return; //@line 313 "modules/physics/box2d/Fixture.cpp"
 } else {
  $11 = HEAP32[$this>>2]|0; //@line 301 "modules/physics/box2d/Fixture.cpp"
  $12 = ((($11)) + 8|0); //@line 301 "modules/physics/box2d/Fixture.cpp"
  $13 = HEAP32[$12>>2]|0; //@line 301 "modules/physics/box2d/Fixture.cpp"
  FUNCTION_TABLE_vi[$13 & 8191]($this); //@line 301 "modules/physics/box2d/Fixture.cpp"
  $14 = HEAP32[$1>>2]|0; //@line 302 "modules/physics/box2d/Fixture.cpp"
  $15 = ((($14)) + 12|0); //@line 302 "modules/physics/box2d/Fixture.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 302 "modules/physics/box2d/Fixture.cpp"
  HEAP32[$0>>2] = $this; //@line 302 "modules/physics/box2d/Fixture.cpp"
  $17 = ((($16)) + 44|0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$17>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = ((($16)) + 48|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $20 = HEAP32[$19>>2]|0; //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($18>>>0)<($20>>>0); //@line 1610 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($21) {
   $22 = $this; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   HEAP32[$18>>2] = $22; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $23 = HEAP32[$17>>2]|0; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $24 = ((($23)) + 4|0); //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$17>>2] = $24; //@line 1617 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   STACKTOP = sp;return; //@line 313 "modules/physics/box2d/Fixture.cpp"
  } else {
   $25 = ((($16)) + 40|0); //@line 302 "modules/physics/box2d/Fixture.cpp"
   __ZNSt3__16vectorIPN4love7physics5box2d7FixtureENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_($25,$0); //@line 1620 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   STACKTOP = sp;return; //@line 313 "modules/physics/box2d/Fixture.cpp"
  }
 }
}
function __ZNSt3__16vectorIPN4love7physics5box2d7FixtureENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(1073741823); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 2; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(536870911); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 1; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$33 = $18;
  } else {
   $$0$i1 = $17;$41 = $18;$42 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 2; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 1073741823;$41 = $12;$42 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $42;$33 = $41;
 }
 $24 = (($25) + ($26<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$24>>2] = $29; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $30 = ((($24)) + 4|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = $30; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $32 = (($33) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = $32 >> 2; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (0 - ($34))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = (($24) + ($35<<2)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $36; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $38 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($36|0),($38|0),($32|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $39 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($39) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $40 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($40); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN4love7physics5box2d13FrictionJointD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 50 "modules/physics/box2d/FrictionJoint.cpp"
 return; //@line 50 "modules/physics/box2d/FrictionJoint.cpp"
}
function __ZN4love7physics5box2d13FrictionJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 50 "modules/physics/box2d/FrictionJoint.cpp"
 __ZdlPv($this); //@line 49 "modules/physics/box2d/FrictionJoint.cpp"
 return; //@line 50 "modules/physics/box2d/FrictionJoint.cpp"
}
function __ZN4love7physics5box2d13FrictionJointC2EPNS1_4BodyES4_ffffb($this,$body1,$body2,$xA,$yA,$xB,$yB,$collideConnected) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 $xA = +$xA;
 $yA = +$yA;
 $xB = +$xB;
 $yB = +$yB;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0, $26 = 0.0;
 var $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $def = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $def = sp + 32|0;
 $0 = sp + 24|0;
 $1 = sp + 16|0;
 $2 = sp + 8|0;
 $3 = sp;
 __ZN4love7physics5box2d5JointC2EPNS1_4BodyES4_($this,$body1,$body2); //@line 38 "modules/physics/box2d/FrictionJoint.cpp"
 HEAP32[$this>>2] = (32760); //@line 40 "modules/physics/box2d/FrictionJoint.cpp"
 $4 = ((($this)) + 28|0); //@line 39 "modules/physics/box2d/FrictionJoint.cpp"
 HEAP32[$4>>2] = 0; //@line 39 "modules/physics/box2d/FrictionJoint.cpp"
 ;HEAP32[$def>>2]=0|0;HEAP32[$def+4>>2]=0|0;HEAP32[$def+8>>2]=0|0;HEAP32[$def+12>>2]=0|0;HEAP8[$def+16>>0]=0|0; //@line 79 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 HEAP32[$def>>2] = 9; //@line 29 "./libraries/Box2D/Dynamics/Joints/b2FrictionJoint.h"
 $5 = ((($def)) + 20|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;HEAP32[$5+12>>2]=0|0;HEAP32[$5+16>>2]=0|0;HEAP32[$5+20>>2]=0|0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($body1)) + 8|0); //@line 42 "modules/physics/box2d/FrictionJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 42 "modules/physics/box2d/FrictionJoint.cpp"
 $8 = ((($body2)) + 8|0); //@line 42 "modules/physics/box2d/FrictionJoint.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 42 "modules/physics/box2d/FrictionJoint.cpp"
 HEAPF32[$1>>2] = $xA; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$10>>2] = $yA; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 invoke_vii(3695,($0|0),($1|0)); //@line 42 "modules/physics/box2d/FrictionJoint.cpp"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $45 = ___cxa_find_matching_catch_2()|0;
  $46 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 46 "modules/physics/box2d/FrictionJoint.cpp"
  ___resumeException($45|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_viiii(4216,($def|0),($7|0),($9|0),($0|0)); //@line 42 "modules/physics/box2d/FrictionJoint.cpp"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if ($14) {
  $45 = ___cxa_find_matching_catch_2()|0;
  $46 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 46 "modules/physics/box2d/FrictionJoint.cpp"
  ___resumeException($45|0);
  // unreachable;
 }
 $15 = HEAP32[$8>>2]|0; //@line 43 "modules/physics/box2d/FrictionJoint.cpp"
 HEAPF32[$3>>2] = $xB; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $16 = ((($3)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$16>>2] = $yB; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 invoke_vii(3695,($2|0),($3|0)); //@line 43 "modules/physics/box2d/FrictionJoint.cpp"
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if ($18) {
  $45 = ___cxa_find_matching_catch_2()|0;
  $46 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 46 "modules/physics/box2d/FrictionJoint.cpp"
  ___resumeException($45|0);
  // unreachable;
 }
 $19 = +HEAPF32[$2>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $20 = ((($15)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $21 = +HEAPF32[$20>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $22 = $19 - $21; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $23 = ((($2)) + 4|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $24 = +HEAPF32[$23>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $25 = ((($15)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $26 = +HEAPF32[$25>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $27 = $24 - $26; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $28 = ((($15)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $29 = +HEAPF32[$28>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $30 = $22 * $29; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $31 = ((($15)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $32 = +HEAPF32[$31>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $33 = $27 * $32; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $34 = $30 + $33; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $35 = $22 * $32; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $36 = $29 * $27; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $37 = $36 - $35; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $38 = ((($def)) + 28|0); //@line 43 "modules/physics/box2d/FrictionJoint.cpp"
 HEAPF32[$38>>2] = $34; //@line 43 "modules/physics/box2d/FrictionJoint.cpp"
 $39 = ((($def)) + 32|0); //@line 43 "modules/physics/box2d/FrictionJoint.cpp"
 HEAPF32[$39>>2] = $37; //@line 43 "modules/physics/box2d/FrictionJoint.cpp"
 $40 = ((($def)) + 16|0); //@line 44 "modules/physics/box2d/FrictionJoint.cpp"
 $41 = $collideConnected&1; //@line 44 "modules/physics/box2d/FrictionJoint.cpp"
 HEAP8[$40>>0] = $41; //@line 44 "modules/physics/box2d/FrictionJoint.cpp"
 __THREW__ = 0;
 $42 = (invoke_iii(3697,($this|0),($def|0))|0); //@line 45 "modules/physics/box2d/FrictionJoint.cpp"
 $43 = __THREW__; __THREW__ = 0;
 $44 = $43&1;
 if ($44) {
  $45 = ___cxa_find_matching_catch_2()|0;
  $46 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 46 "modules/physics/box2d/FrictionJoint.cpp"
  ___resumeException($45|0);
  // unreachable;
 } else {
  HEAP32[$4>>2] = $42; //@line 45 "modules/physics/box2d/FrictionJoint.cpp"
  STACKTOP = sp;return; //@line 46 "modules/physics/box2d/FrictionJoint.cpp"
 }
}
function __ZN4love7physics5box2d13FrictionJoint11setMaxForceEf($this,$force) {
 $this = $this|0;
 $force = +$force;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 54 "modules/physics/box2d/FrictionJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 54 "modules/physics/box2d/FrictionJoint.cpp"
 $2 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($force)); //@line 54 "modules/physics/box2d/FrictionJoint.cpp"
 __ZN15b2FrictionJoint11SetMaxForceEf($1,$2); //@line 54 "modules/physics/box2d/FrictionJoint.cpp"
 return; //@line 55 "modules/physics/box2d/FrictionJoint.cpp"
}
function __ZNK4love7physics5box2d13FrictionJoint11getMaxForceEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 59 "modules/physics/box2d/FrictionJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 59 "modules/physics/box2d/FrictionJoint.cpp"
 $2 = (+__ZNK15b2FrictionJoint11GetMaxForceEv($1)); //@line 59 "modules/physics/box2d/FrictionJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 59 "modules/physics/box2d/FrictionJoint.cpp"
 return (+$3); //@line 59 "modules/physics/box2d/FrictionJoint.cpp"
}
function __ZN4love7physics5box2d13FrictionJoint12setMaxTorqueEf($this,$torque) {
 $this = $this|0;
 $torque = +$torque;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 64 "modules/physics/box2d/FrictionJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 64 "modules/physics/box2d/FrictionJoint.cpp"
 $2 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($torque)); //@line 64 "modules/physics/box2d/FrictionJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($2)); //@line 64 "modules/physics/box2d/FrictionJoint.cpp"
 __ZN15b2FrictionJoint12SetMaxTorqueEf($1,$3); //@line 64 "modules/physics/box2d/FrictionJoint.cpp"
 return; //@line 65 "modules/physics/box2d/FrictionJoint.cpp"
}
function __ZNK4love7physics5box2d13FrictionJoint12getMaxTorqueEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 69 "modules/physics/box2d/FrictionJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 69 "modules/physics/box2d/FrictionJoint.cpp"
 $2 = (+__ZNK15b2FrictionJoint12GetMaxTorqueEv($1)); //@line 69 "modules/physics/box2d/FrictionJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 69 "modules/physics/box2d/FrictionJoint.cpp"
 $4 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($3)); //@line 69 "modules/physics/box2d/FrictionJoint.cpp"
 return (+$4); //@line 69 "modules/physics/box2d/FrictionJoint.cpp"
}
function __ZN4love7physics5box2d9GearJointD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 52 "modules/physics/box2d/GearJoint.cpp"
 return; //@line 52 "modules/physics/box2d/GearJoint.cpp"
}
function __ZN4love7physics5box2d9GearJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 52 "modules/physics/box2d/GearJoint.cpp"
 __ZdlPv($this); //@line 51 "modules/physics/box2d/GearJoint.cpp"
 return; //@line 52 "modules/physics/box2d/GearJoint.cpp"
}
function __ZN4love7physics5box2d9GearJointC2EPNS1_5JointES4_fb($this,$joint1,$joint2,$ratio,$collideConnected) {
 $this = $this|0;
 $joint1 = $joint1|0;
 $joint2 = $joint2|0;
 $ratio = +$ratio;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $def = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $def = sp;
 $0 = ((($joint1)) + 20|0); //@line 36 "modules/physics/box2d/GearJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 36 "modules/physics/box2d/GearJoint.cpp"
 $2 = ((($joint2)) + 20|0); //@line 36 "modules/physics/box2d/GearJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 36 "modules/physics/box2d/GearJoint.cpp"
 __ZN4love7physics5box2d5JointC2EPNS1_4BodyES4_($this,$1,$3); //@line 36 "modules/physics/box2d/GearJoint.cpp"
 HEAP32[$this>>2] = (32784); //@line 38 "modules/physics/box2d/GearJoint.cpp"
 $4 = ((($this)) + 28|0); //@line 37 "modules/physics/box2d/GearJoint.cpp"
 HEAP32[$4>>2] = 0; //@line 37 "modules/physics/box2d/GearJoint.cpp"
 ;HEAP32[$def>>2]=0|0;HEAP32[$def+4>>2]=0|0;HEAP32[$def+8>>2]=0|0;HEAP32[$def+12>>2]=0|0; //@line 79 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 HEAP32[$def>>2] = 6; //@line 30 "./libraries/Box2D/Dynamics/Joints/b2GearJoint.h"
 $5 = ((($def)) + 20|0); //@line 31 "./libraries/Box2D/Dynamics/Joints/b2GearJoint.h"
 $6 = ((($def)) + 24|0); //@line 32 "./libraries/Box2D/Dynamics/Joints/b2GearJoint.h"
 $7 = ((($def)) + 28|0); //@line 33 "./libraries/Box2D/Dynamics/Joints/b2GearJoint.h"
 $8 = ((($joint1)) + 24|0); //@line 40 "modules/physics/box2d/GearJoint.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 40 "modules/physics/box2d/GearJoint.cpp"
 HEAP32[$5>>2] = $9; //@line 40 "modules/physics/box2d/GearJoint.cpp"
 $10 = ((($joint2)) + 24|0); //@line 41 "modules/physics/box2d/GearJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 41 "modules/physics/box2d/GearJoint.cpp"
 HEAP32[$6>>2] = $11; //@line 41 "modules/physics/box2d/GearJoint.cpp"
 $12 = HEAP32[$0>>2]|0; //@line 42 "modules/physics/box2d/GearJoint.cpp"
 $13 = ((($12)) + 8|0); //@line 42 "modules/physics/box2d/GearJoint.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 42 "modules/physics/box2d/GearJoint.cpp"
 $15 = ((($def)) + 8|0); //@line 42 "modules/physics/box2d/GearJoint.cpp"
 HEAP32[$15>>2] = $14; //@line 42 "modules/physics/box2d/GearJoint.cpp"
 $16 = HEAP32[$2>>2]|0; //@line 43 "modules/physics/box2d/GearJoint.cpp"
 $17 = ((($16)) + 8|0); //@line 43 "modules/physics/box2d/GearJoint.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 43 "modules/physics/box2d/GearJoint.cpp"
 $19 = ((($def)) + 12|0); //@line 43 "modules/physics/box2d/GearJoint.cpp"
 HEAP32[$19>>2] = $18; //@line 43 "modules/physics/box2d/GearJoint.cpp"
 HEAPF32[$7>>2] = $ratio; //@line 44 "modules/physics/box2d/GearJoint.cpp"
 $20 = ((($def)) + 16|0); //@line 45 "modules/physics/box2d/GearJoint.cpp"
 $21 = $collideConnected&1; //@line 45 "modules/physics/box2d/GearJoint.cpp"
 HEAP8[$20>>0] = $21; //@line 45 "modules/physics/box2d/GearJoint.cpp"
 __THREW__ = 0;
 $22 = (invoke_iii(3697,($this|0),($def|0))|0); //@line 47 "modules/physics/box2d/GearJoint.cpp"
 $23 = __THREW__; __THREW__ = 0;
 $24 = $23&1;
 if ($24) {
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 48 "modules/physics/box2d/GearJoint.cpp"
  ___resumeException($25|0);
  // unreachable;
 } else {
  HEAP32[$4>>2] = $22; //@line 47 "modules/physics/box2d/GearJoint.cpp"
  STACKTOP = sp;return; //@line 48 "modules/physics/box2d/GearJoint.cpp"
 }
}
function __ZN4love7physics5box2d9GearJoint8setRatioEf($this,$ratio) {
 $this = $this|0;
 $ratio = +$ratio;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 56 "modules/physics/box2d/GearJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 56 "modules/physics/box2d/GearJoint.cpp"
 __ZN11b2GearJoint8SetRatioEf($1,$ratio); //@line 56 "modules/physics/box2d/GearJoint.cpp"
 return; //@line 57 "modules/physics/box2d/GearJoint.cpp"
}
function __ZNK4love7physics5box2d9GearJoint8getRatioEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 61 "modules/physics/box2d/GearJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 61 "modules/physics/box2d/GearJoint.cpp"
 $2 = (+__ZNK11b2GearJoint8GetRatioEv($1)); //@line 61 "modules/physics/box2d/GearJoint.cpp"
 return (+$2); //@line 61 "modules/physics/box2d/GearJoint.cpp"
}
function __ZNK4love7physics5box2d9GearJoint9getJointAEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 28|0); //@line 66 "modules/physics/box2d/GearJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 66 "modules/physics/box2d/GearJoint.cpp"
 $2 = ((($1)) + 68|0); //@line 66 "./libraries/Box2D/Dynamics/Joints/b2GearJoint.h"
 $3 = HEAP32[$2>>2]|0; //@line 66 "./libraries/Box2D/Dynamics/Joints/b2GearJoint.h"
 $4 = ($3|0)==(0|0); //@line 67 "modules/physics/box2d/GearJoint.cpp"
 if ($4) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 75 "modules/physics/box2d/GearJoint.cpp"
 }
 $5 = (__ZN4love8Memoizer4findEPv($3)|0); //@line 70 "modules/physics/box2d/GearJoint.cpp"
 $6 = ($5|0)==(0|0); //@line 71 "modules/physics/box2d/GearJoint.cpp"
 if (!($6)) {
  $$0 = $5;
  STACKTOP = sp;return ($$0|0); //@line 75 "modules/physics/box2d/GearJoint.cpp"
 }
 $7 = (___cxa_allocate_exception(16)|0); //@line 72 "modules/physics/box2d/GearJoint.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($7|0),(444962|0),($vararg_buffer|0)); //@line 72 "modules/physics/box2d/GearJoint.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  ___cxa_free_exception(($7|0)); //@line 72 "modules/physics/box2d/GearJoint.cpp"
  ___resumeException($10|0);
  // unreachable;
 } else {
  ___cxa_throw(($7|0),(96|0),(344|0)); //@line 72 "modules/physics/box2d/GearJoint.cpp"
  // unreachable; //@line 72 "modules/physics/box2d/GearJoint.cpp"
 }
 return (0)|0;
}
function __ZNK4love7physics5box2d9GearJoint9getJointBEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 28|0); //@line 79 "modules/physics/box2d/GearJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 79 "modules/physics/box2d/GearJoint.cpp"
 $2 = ((($1)) + 72|0); //@line 69 "./libraries/Box2D/Dynamics/Joints/b2GearJoint.h"
 $3 = HEAP32[$2>>2]|0; //@line 69 "./libraries/Box2D/Dynamics/Joints/b2GearJoint.h"
 $4 = ($3|0)==(0|0); //@line 80 "modules/physics/box2d/GearJoint.cpp"
 if ($4) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 88 "modules/physics/box2d/GearJoint.cpp"
 }
 $5 = (__ZN4love8Memoizer4findEPv($3)|0); //@line 83 "modules/physics/box2d/GearJoint.cpp"
 $6 = ($5|0)==(0|0); //@line 84 "modules/physics/box2d/GearJoint.cpp"
 if (!($6)) {
  $$0 = $5;
  STACKTOP = sp;return ($$0|0); //@line 88 "modules/physics/box2d/GearJoint.cpp"
 }
 $7 = (___cxa_allocate_exception(16)|0); //@line 85 "modules/physics/box2d/GearJoint.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($7|0),(444962|0),($vararg_buffer|0)); //@line 85 "modules/physics/box2d/GearJoint.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  ___cxa_free_exception(($7|0)); //@line 85 "modules/physics/box2d/GearJoint.cpp"
  ___resumeException($10|0);
  // unreachable;
 } else {
  ___cxa_throw(($7|0),(96|0),(344|0)); //@line 85 "modules/physics/box2d/GearJoint.cpp"
  // unreachable; //@line 85 "modules/physics/box2d/GearJoint.cpp"
 }
 return (0)|0;
}
function __ZN4love7physics5box2d10MotorJointD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 60 "modules/physics/box2d/MotorJoint.cpp"
 return; //@line 60 "modules/physics/box2d/MotorJoint.cpp"
}
function __ZN4love7physics5box2d10MotorJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 60 "modules/physics/box2d/MotorJoint.cpp"
 __ZdlPv($this); //@line 59 "modules/physics/box2d/MotorJoint.cpp"
 return; //@line 60 "modules/physics/box2d/MotorJoint.cpp"
}
function __ZN4love7physics5box2d10MotorJointC2EPNS1_4BodyES4_($this,$body1,$body2) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $def = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $def = sp;
 __ZN4love7physics5box2d5JointC2EPNS1_4BodyES4_($this,$body1,$body2); //@line 36 "modules/physics/box2d/MotorJoint.cpp"
 HEAP32[$this>>2] = (32808); //@line 38 "modules/physics/box2d/MotorJoint.cpp"
 $0 = ((($this)) + 28|0); //@line 37 "modules/physics/box2d/MotorJoint.cpp"
 HEAP32[$0>>2] = 0; //@line 37 "modules/physics/box2d/MotorJoint.cpp"
 ;HEAP32[$def>>2]=0|0;HEAP32[$def+4>>2]=0|0;HEAP32[$def+8>>2]=0|0;HEAP32[$def+12>>2]=0|0;HEAP8[$def+16>>0]=0|0; //@line 79 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 HEAP32[$def>>2] = 11; //@line 29 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 $1 = ((($def)) + 20|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$1>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $2 = ((($def)) + 24|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$2>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($def)) + 28|0); //@line 31 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 HEAPF32[$3>>2] = 0.0; //@line 31 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 $4 = ((($def)) + 32|0); //@line 32 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 HEAPF32[$4>>2] = 1.0; //@line 32 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 $5 = ((($def)) + 36|0); //@line 33 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 HEAPF32[$5>>2] = 1.0; //@line 33 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 $6 = ((($def)) + 40|0); //@line 34 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 HEAPF32[$6>>2] = 0.30000001192092896; //@line 34 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 $7 = ((($body1)) + 8|0); //@line 41 "modules/physics/box2d/MotorJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 41 "modules/physics/box2d/MotorJoint.cpp"
 $9 = ((($body2)) + 8|0); //@line 41 "modules/physics/box2d/MotorJoint.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 41 "modules/physics/box2d/MotorJoint.cpp"
 __THREW__ = 0;
 invoke_viii(4217,($def|0),($8|0),($10|0)); //@line 41 "modules/physics/box2d/MotorJoint.cpp"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 43 "modules/physics/box2d/MotorJoint.cpp"
  ___resumeException($16|0);
  // unreachable;
 }
 __THREW__ = 0;
 $13 = (invoke_iii(3697,($this|0),($def|0))|0); //@line 42 "modules/physics/box2d/MotorJoint.cpp"
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if ($15) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 43 "modules/physics/box2d/MotorJoint.cpp"
  ___resumeException($16|0);
  // unreachable;
 } else {
  HEAP32[$0>>2] = $13; //@line 42 "modules/physics/box2d/MotorJoint.cpp"
  STACKTOP = sp;return; //@line 43 "modules/physics/box2d/MotorJoint.cpp"
 }
}
function __ZN4love7physics5box2d10MotorJointC2EPNS1_4BodyES4_fb($this,$body1,$body2,$correctionFactor,$collideConnected) {
 $this = $this|0;
 $body1 = $body1|0;
 $body2 = $body2|0;
 $correctionFactor = +$correctionFactor;
 $collideConnected = $collideConnected|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $def = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $def = sp;
 __ZN4love7physics5box2d5JointC2EPNS1_4BodyES4_($this,$body1,$body2); //@line 46 "modules/physics/box2d/MotorJoint.cpp"
 HEAP32[$this>>2] = (32808); //@line 48 "modules/physics/box2d/MotorJoint.cpp"
 $0 = ((($this)) + 28|0); //@line 47 "modules/physics/box2d/MotorJoint.cpp"
 HEAP32[$0>>2] = 0; //@line 47 "modules/physics/box2d/MotorJoint.cpp"
 ;HEAP32[$def>>2]=0|0;HEAP32[$def+4>>2]=0|0;HEAP32[$def+8>>2]=0|0;HEAP32[$def+12>>2]=0|0;HEAP8[$def+16>>0]=0|0; //@line 79 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 HEAP32[$def>>2] = 11; //@line 29 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 $1 = ((($def)) + 20|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$1>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $2 = ((($def)) + 24|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$2>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($def)) + 28|0); //@line 31 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 HEAPF32[$3>>2] = 0.0; //@line 31 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 $4 = ((($def)) + 32|0); //@line 32 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 HEAPF32[$4>>2] = 1.0; //@line 32 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 $5 = ((($def)) + 36|0); //@line 33 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 HEAPF32[$5>>2] = 1.0; //@line 33 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 $6 = ((($def)) + 40|0); //@line 34 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 HEAPF32[$6>>2] = 0.30000001192092896; //@line 34 "./libraries/Box2D/Dynamics/Joints/b2MotorJoint.h"
 $7 = ((($body1)) + 8|0); //@line 51 "modules/physics/box2d/MotorJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 51 "modules/physics/box2d/MotorJoint.cpp"
 $9 = ((($body2)) + 8|0); //@line 51 "modules/physics/box2d/MotorJoint.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 51 "modules/physics/box2d/MotorJoint.cpp"
 __THREW__ = 0;
 invoke_viii(4217,($def|0),($8|0),($10|0)); //@line 51 "modules/physics/box2d/MotorJoint.cpp"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $18 = ___cxa_find_matching_catch_2()|0;
  $19 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 56 "modules/physics/box2d/MotorJoint.cpp"
  ___resumeException($18|0);
  // unreachable;
 }
 HEAPF32[$6>>2] = $correctionFactor; //@line 52 "modules/physics/box2d/MotorJoint.cpp"
 $13 = ((($def)) + 16|0); //@line 53 "modules/physics/box2d/MotorJoint.cpp"
 $14 = $collideConnected&1; //@line 53 "modules/physics/box2d/MotorJoint.cpp"
 HEAP8[$13>>0] = $14; //@line 53 "modules/physics/box2d/MotorJoint.cpp"
 __THREW__ = 0;
 $15 = (invoke_iii(3697,($this|0),($def|0))|0); //@line 55 "modules/physics/box2d/MotorJoint.cpp"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if ($17) {
  $18 = ___cxa_find_matching_catch_2()|0;
  $19 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 56 "modules/physics/box2d/MotorJoint.cpp"
  ___resumeException($18|0);
  // unreachable;
 } else {
  HEAP32[$0>>2] = $15; //@line 55 "modules/physics/box2d/MotorJoint.cpp"
  STACKTOP = sp;return; //@line 56 "modules/physics/box2d/MotorJoint.cpp"
 }
}
function __ZN4love7physics5box2d10MotorJoint15setLinearOffsetEff($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 8|0;
 $1 = sp;
 $2 = ((($this)) + 28|0); //@line 64 "modules/physics/box2d/MotorJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 64 "modules/physics/box2d/MotorJoint.cpp"
 HEAPF32[$1>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$4>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($0,$1); //@line 64 "modules/physics/box2d/MotorJoint.cpp"
 __ZN12b2MotorJoint15SetLinearOffsetERK6b2Vec2($3,$0); //@line 64 "modules/physics/box2d/MotorJoint.cpp"
 STACKTOP = sp;return; //@line 65 "modules/physics/box2d/MotorJoint.cpp"
}
function __ZNK4love7physics5box2d10MotorJoint15getLinearOffsetEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 69 "modules/physics/box2d/MotorJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 69 "modules/physics/box2d/MotorJoint.cpp"
 $2 = (__ZNK12b2MotorJoint15GetLinearOffsetEv($1)|0); //@line 69 "modules/physics/box2d/MotorJoint.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 69 "modules/physics/box2d/MotorJoint.cpp"
 $4 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($3)); //@line 69 "modules/physics/box2d/MotorJoint.cpp"
 $5 = $4; //@line 69 "modules/physics/box2d/MotorJoint.cpp"
 _lua_pushnumber($L,$5); //@line 69 "modules/physics/box2d/MotorJoint.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 70 "modules/physics/box2d/MotorJoint.cpp"
 $7 = (__ZNK12b2MotorJoint15GetLinearOffsetEv($6)|0); //@line 70 "modules/physics/box2d/MotorJoint.cpp"
 $8 = ((($7)) + 4|0); //@line 70 "modules/physics/box2d/MotorJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 70 "modules/physics/box2d/MotorJoint.cpp"
 $10 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($9)); //@line 70 "modules/physics/box2d/MotorJoint.cpp"
 $11 = $10; //@line 70 "modules/physics/box2d/MotorJoint.cpp"
 _lua_pushnumber($L,$11); //@line 70 "modules/physics/box2d/MotorJoint.cpp"
 return 2; //@line 71 "modules/physics/box2d/MotorJoint.cpp"
}
function __ZN4love7physics5box2d10MotorJoint16setAngularOffsetEf($this,$angularOffset) {
 $this = $this|0;
 $angularOffset = +$angularOffset;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 76 "modules/physics/box2d/MotorJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 76 "modules/physics/box2d/MotorJoint.cpp"
 __ZN12b2MotorJoint16SetAngularOffsetEf($1,$angularOffset); //@line 76 "modules/physics/box2d/MotorJoint.cpp"
 return; //@line 77 "modules/physics/box2d/MotorJoint.cpp"
}
function __ZNK4love7physics5box2d10MotorJoint16getAngularOffsetEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 81 "modules/physics/box2d/MotorJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 81 "modules/physics/box2d/MotorJoint.cpp"
 $2 = (+__ZNK12b2MotorJoint16GetAngularOffsetEv($1)); //@line 81 "modules/physics/box2d/MotorJoint.cpp"
 return (+$2); //@line 81 "modules/physics/box2d/MotorJoint.cpp"
}
function __ZN4love7physics5box2d10MotorJoint11setMaxForceEf($this,$force) {
 $this = $this|0;
 $force = +$force;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 86 "modules/physics/box2d/MotorJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 86 "modules/physics/box2d/MotorJoint.cpp"
 $2 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($force)); //@line 86 "modules/physics/box2d/MotorJoint.cpp"
 __ZN12b2MotorJoint11SetMaxForceEf($1,$2); //@line 86 "modules/physics/box2d/MotorJoint.cpp"
 return; //@line 87 "modules/physics/box2d/MotorJoint.cpp"
}
function __ZNK4love7physics5box2d10MotorJoint11getMaxForceEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 91 "modules/physics/box2d/MotorJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 91 "modules/physics/box2d/MotorJoint.cpp"
 $2 = (+__ZNK12b2MotorJoint11GetMaxForceEv($1)); //@line 91 "modules/physics/box2d/MotorJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 91 "modules/physics/box2d/MotorJoint.cpp"
 return (+$3); //@line 91 "modules/physics/box2d/MotorJoint.cpp"
}
function __ZN4love7physics5box2d10MotorJoint12setMaxTorqueEf($this,$torque) {
 $this = $this|0;
 $torque = +$torque;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 96 "modules/physics/box2d/MotorJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 96 "modules/physics/box2d/MotorJoint.cpp"
 $2 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($torque)); //@line 96 "modules/physics/box2d/MotorJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($2)); //@line 96 "modules/physics/box2d/MotorJoint.cpp"
 __ZN12b2MotorJoint12SetMaxTorqueEf($1,$3); //@line 96 "modules/physics/box2d/MotorJoint.cpp"
 return; //@line 97 "modules/physics/box2d/MotorJoint.cpp"
}
function __ZNK4love7physics5box2d10MotorJoint12getMaxTorqueEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 101 "modules/physics/box2d/MotorJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 101 "modules/physics/box2d/MotorJoint.cpp"
 $2 = (+__ZNK12b2MotorJoint12GetMaxTorqueEv($1)); //@line 101 "modules/physics/box2d/MotorJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 101 "modules/physics/box2d/MotorJoint.cpp"
 $4 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($3)); //@line 101 "modules/physics/box2d/MotorJoint.cpp"
 return (+$4); //@line 101 "modules/physics/box2d/MotorJoint.cpp"
}
function __ZN4love7physics5box2d10MotorJoint19setCorrectionFactorEf($this,$factor) {
 $this = $this|0;
 $factor = +$factor;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 106 "modules/physics/box2d/MotorJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 106 "modules/physics/box2d/MotorJoint.cpp"
 __ZN12b2MotorJoint19SetCorrectionFactorEf($1,$factor); //@line 106 "modules/physics/box2d/MotorJoint.cpp"
 return; //@line 107 "modules/physics/box2d/MotorJoint.cpp"
}
function __ZNK4love7physics5box2d10MotorJoint19getCorrectionFactorEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 111 "modules/physics/box2d/MotorJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 111 "modules/physics/box2d/MotorJoint.cpp"
 $2 = (+__ZNK12b2MotorJoint19GetCorrectionFactorEv($1)); //@line 111 "modules/physics/box2d/MotorJoint.cpp"
 return (+$2); //@line 111 "modules/physics/box2d/MotorJoint.cpp"
}
function __ZN4love7physics5box2d10MouseJointD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 50 "modules/physics/box2d/MouseJoint.cpp"
 return; //@line 50 "modules/physics/box2d/MouseJoint.cpp"
}
function __ZN4love7physics5box2d10MouseJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love7physics5box2d5JointD2Ev($this); //@line 50 "modules/physics/box2d/MouseJoint.cpp"
 __ZdlPv($this); //@line 49 "modules/physics/box2d/MouseJoint.cpp"
 return; //@line 50 "modules/physics/box2d/MouseJoint.cpp"
}
function __ZN4love7physics5box2d10MouseJointC2EPNS1_4BodyEff($this,$body1,$x,$y) {
 $this = $this|0;
 $body1 = $body1|0;
 $x = +$x;
 $y = +$y;
 var $$cast = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $def = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $def = sp + 16|0;
 $0 = sp;
 $1 = sp + 8|0;
 __ZN4love7physics5box2d5JointC2EPNS1_4BodyE($this,$body1); //@line 36 "modules/physics/box2d/MouseJoint.cpp"
 HEAP32[$this>>2] = (32832); //@line 38 "modules/physics/box2d/MouseJoint.cpp"
 $2 = ((($this)) + 28|0); //@line 37 "modules/physics/box2d/MouseJoint.cpp"
 HEAP32[$2>>2] = 0; //@line 37 "modules/physics/box2d/MouseJoint.cpp"
 ;HEAP32[$def>>2]=0|0;HEAP32[$def+4>>2]=0|0;HEAP32[$def+8>>2]=0|0;HEAP32[$def+12>>2]=0|0;HEAP8[$def+16>>0]=0|0; //@line 79 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 HEAP32[$def>>2] = 5; //@line 30 "./libraries/Box2D/Dynamics/Joints/b2MouseJoint.h"
 $3 = ((($def)) + 20|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$3>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($def)) + 24|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$4>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($def)) + 28|0); //@line 32 "./libraries/Box2D/Dynamics/Joints/b2MouseJoint.h"
 HEAPF32[$5>>2] = 0.0; //@line 32 "./libraries/Box2D/Dynamics/Joints/b2MouseJoint.h"
 $6 = ((($def)) + 32|0); //@line 33 "./libraries/Box2D/Dynamics/Joints/b2MouseJoint.h"
 HEAPF32[$6>>2] = 5.0; //@line 33 "./libraries/Box2D/Dynamics/Joints/b2MouseJoint.h"
 $7 = ((($def)) + 36|0); //@line 34 "./libraries/Box2D/Dynamics/Joints/b2MouseJoint.h"
 HEAPF32[$7>>2] = 0.69999998807907104; //@line 34 "./libraries/Box2D/Dynamics/Joints/b2MouseJoint.h"
 $8 = ((($body1)) + 12|0); //@line 41 "modules/physics/box2d/MouseJoint.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 41 "modules/physics/box2d/MouseJoint.cpp"
 __THREW__ = 0;
 $10 = (invoke_ii(4218,($9|0))|0); //@line 41 "modules/physics/box2d/MouseJoint.cpp"
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $37 = ___cxa_find_matching_catch_2()|0;
  $38 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 46 "modules/physics/box2d/MouseJoint.cpp"
  ___resumeException($37|0);
  // unreachable;
 }
 $13 = ((($def)) + 8|0); //@line 41 "modules/physics/box2d/MouseJoint.cpp"
 HEAP32[$13>>2] = $10; //@line 41 "modules/physics/box2d/MouseJoint.cpp"
 $14 = ((($body1)) + 8|0); //@line 42 "modules/physics/box2d/MouseJoint.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 42 "modules/physics/box2d/MouseJoint.cpp"
 $16 = ((($def)) + 12|0); //@line 42 "modules/physics/box2d/MouseJoint.cpp"
 HEAP32[$16>>2] = $15; //@line 42 "modules/physics/box2d/MouseJoint.cpp"
 $$cast = $15; //@line 43 "modules/physics/box2d/MouseJoint.cpp"
 $17 = ((($$cast)) + 116|0); //@line 541 "./libraries/Box2D/Dynamics/b2Body.h"
 $18 = +HEAPF32[$17>>2]; //@line 541 "./libraries/Box2D/Dynamics/b2Body.h"
 $19 = $18 * 1000.0; //@line 43 "modules/physics/box2d/MouseJoint.cpp"
 HEAPF32[$5>>2] = $19; //@line 43 "modules/physics/box2d/MouseJoint.cpp"
 HEAPF32[$1>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $20 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$20>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __THREW__ = 0;
 invoke_vii(3695,($0|0),($1|0)); //@line 44 "modules/physics/box2d/MouseJoint.cpp"
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $37 = ___cxa_find_matching_catch_2()|0;
  $38 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 46 "modules/physics/box2d/MouseJoint.cpp"
  ___resumeException($37|0);
  // unreachable;
 }
 $23 = ((($def)) + 20|0); //@line 44 "modules/physics/box2d/MouseJoint.cpp"
 $24 = $0; //@line 44 "modules/physics/box2d/MouseJoint.cpp"
 $25 = $24; //@line 44 "modules/physics/box2d/MouseJoint.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 44 "modules/physics/box2d/MouseJoint.cpp"
 $27 = (($24) + 4)|0; //@line 44 "modules/physics/box2d/MouseJoint.cpp"
 $28 = $27; //@line 44 "modules/physics/box2d/MouseJoint.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 44 "modules/physics/box2d/MouseJoint.cpp"
 $30 = $23; //@line 44 "modules/physics/box2d/MouseJoint.cpp"
 $31 = $30; //@line 44 "modules/physics/box2d/MouseJoint.cpp"
 HEAP32[$31>>2] = $26; //@line 44 "modules/physics/box2d/MouseJoint.cpp"
 $32 = (($30) + 4)|0; //@line 44 "modules/physics/box2d/MouseJoint.cpp"
 $33 = $32; //@line 44 "modules/physics/box2d/MouseJoint.cpp"
 HEAP32[$33>>2] = $29; //@line 44 "modules/physics/box2d/MouseJoint.cpp"
 __THREW__ = 0;
 $34 = (invoke_iii(3697,($this|0),($def|0))|0); //@line 45 "modules/physics/box2d/MouseJoint.cpp"
 $35 = __THREW__; __THREW__ = 0;
 $36 = $35&1;
 if ($36) {
  $37 = ___cxa_find_matching_catch_2()|0;
  $38 = tempRet0;
  __ZN4love7physics5box2d5JointD2Ev($this); //@line 46 "modules/physics/box2d/MouseJoint.cpp"
  ___resumeException($37|0);
  // unreachable;
 } else {
  HEAP32[$2>>2] = $34; //@line 45 "modules/physics/box2d/MouseJoint.cpp"
  STACKTOP = sp;return; //@line 46 "modules/physics/box2d/MouseJoint.cpp"
 }
}
function __ZN4love7physics5box2d10MouseJoint9setTargetEff($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 8|0;
 $1 = sp;
 $2 = ((($this)) + 28|0); //@line 54 "modules/physics/box2d/MouseJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 54 "modules/physics/box2d/MouseJoint.cpp"
 HEAPF32[$1>>2] = $x; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($1)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$4>>2] = $y; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 __ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2($0,$1); //@line 54 "modules/physics/box2d/MouseJoint.cpp"
 __ZN12b2MouseJoint9SetTargetERK6b2Vec2($3,$0); //@line 54 "modules/physics/box2d/MouseJoint.cpp"
 STACKTOP = sp;return; //@line 55 "modules/physics/box2d/MouseJoint.cpp"
}
function __ZN4love7physics5box2d10MouseJoint9getTargetEP9lua_State($this,$L) {
 $this = $this|0;
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 59 "modules/physics/box2d/MouseJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 59 "modules/physics/box2d/MouseJoint.cpp"
 $2 = (__ZNK12b2MouseJoint9GetTargetEv($1)|0); //@line 59 "modules/physics/box2d/MouseJoint.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 59 "modules/physics/box2d/MouseJoint.cpp"
 $4 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($3)); //@line 59 "modules/physics/box2d/MouseJoint.cpp"
 $5 = $4; //@line 59 "modules/physics/box2d/MouseJoint.cpp"
 _lua_pushnumber($L,$5); //@line 59 "modules/physics/box2d/MouseJoint.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 60 "modules/physics/box2d/MouseJoint.cpp"
 $7 = (__ZNK12b2MouseJoint9GetTargetEv($6)|0); //@line 60 "modules/physics/box2d/MouseJoint.cpp"
 $8 = ((($7)) + 4|0); //@line 60 "modules/physics/box2d/MouseJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 60 "modules/physics/box2d/MouseJoint.cpp"
 $10 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($9)); //@line 60 "modules/physics/box2d/MouseJoint.cpp"
 $11 = $10; //@line 60 "modules/physics/box2d/MouseJoint.cpp"
 _lua_pushnumber($L,$11); //@line 60 "modules/physics/box2d/MouseJoint.cpp"
 return 2; //@line 61 "modules/physics/box2d/MouseJoint.cpp"
}
function __ZN4love7physics5box2d10MouseJoint11setMaxForceEf($this,$force) {
 $this = $this|0;
 $force = +$force;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 66 "modules/physics/box2d/MouseJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 66 "modules/physics/box2d/MouseJoint.cpp"
 $2 = (+__ZN4love7physics5box2d7Physics9scaleDownEf($force)); //@line 66 "modules/physics/box2d/MouseJoint.cpp"
 __ZN12b2MouseJoint11SetMaxForceEf($1,$2); //@line 66 "modules/physics/box2d/MouseJoint.cpp"
 return; //@line 67 "modules/physics/box2d/MouseJoint.cpp"
}
function __ZNK4love7physics5box2d10MouseJoint11getMaxForceEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 71 "modules/physics/box2d/MouseJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 71 "modules/physics/box2d/MouseJoint.cpp"
 $2 = (+__ZNK12b2MouseJoint11GetMaxForceEv($1)); //@line 71 "modules/physics/box2d/MouseJoint.cpp"
 $3 = (+__ZN4love7physics5box2d7Physics7scaleUpEf($2)); //@line 71 "modules/physics/box2d/MouseJoint.cpp"
 return (+$3); //@line 71 "modules/physics/box2d/MouseJoint.cpp"
}
function __ZN4love7physics5box2d10MouseJoint12setFrequencyEf($this,$hz) {
 $this = $this|0;
 $hz = +$hz;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 76 "modules/physics/box2d/MouseJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 76 "modules/physics/box2d/MouseJoint.cpp"
 __ZN12b2MouseJoint12SetFrequencyEf($1,$hz); //@line 76 "modules/physics/box2d/MouseJoint.cpp"
 return; //@line 77 "modules/physics/box2d/MouseJoint.cpp"
}
function __ZNK4love7physics5box2d10MouseJoint12getFrequencyEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 81 "modules/physics/box2d/MouseJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 81 "modules/physics/box2d/MouseJoint.cpp"
 $2 = (+__ZNK12b2MouseJoint12GetFrequencyEv($1)); //@line 81 "modules/physics/box2d/MouseJoint.cpp"
 return (+$2); //@line 81 "modules/physics/box2d/MouseJoint.cpp"
}
function __ZN4love7physics5box2d10MouseJoint15setDampingRatioEf($this,$d) {
 $this = $this|0;
 $d = +$d;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 86 "modules/physics/box2d/MouseJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 86 "modules/physics/box2d/MouseJoint.cpp"
 __ZN12b2MouseJoint15SetDampingRatioEf($1,$d); //@line 86 "modules/physics/box2d/MouseJoint.cpp"
 return; //@line 87 "modules/physics/box2d/MouseJoint.cpp"
}
function __ZNK4love7physics5box2d10MouseJoint15getDampingRatioEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 91 "modules/physics/box2d/MouseJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 91 "modules/physics/box2d/MouseJoint.cpp"
 $2 = (+__ZNK12b2MouseJoint15GetDampingRatioEv($1)); //@line 91 "modules/physics/box2d/MouseJoint.cpp"
 return (+$2); //@line 91 "modules/physics/box2d/MouseJoint.cpp"
}
function __ZN4love5sound7lullaby7DecoderD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (32856); //@line 44 "modules/sound/lullaby/Decoder.cpp"
 $0 = ((($this)) + 32|0); //@line 45 "modules/sound/lullaby/Decoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 45 "modules/sound/lullaby/Decoder.cpp"
 $2 = ($1|0)==(0|0); //@line 45 "modules/sound/lullaby/Decoder.cpp"
 if (!($2)) {
  __ZdaPv($1); //@line 46 "modules/sound/lullaby/Decoder.cpp"
 }
 $3 = ((($this)) + 12|0); //@line 47 "modules/sound/lullaby/Decoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 47 "modules/sound/lullaby/Decoder.cpp"
 $4 = ((($this)) + 8|0); //@line 106 "./common/Object.h"
 $5 = HEAP32[$4>>2]|0; //@line 106 "./common/Object.h"
 $6 = ($5|0)==(0|0); //@line 106 "./common/Object.h"
 if ($6) {
  return; //@line 47 "modules/sound/lullaby/Decoder.cpp"
 }
 $7 = HEAP32[$5>>2]|0; //@line 106 "./common/Object.h"
 $8 = ((($7)) + 12|0); //@line 106 "./common/Object.h"
 $9 = HEAP32[$8>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($9|0,($5|0)); //@line 106 "./common/Object.h"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $12 = ___cxa_find_matching_catch_3(0|0)|0;
  $13 = tempRet0;
  ___clang_call_terminate($12); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  return; //@line 47 "modules/sound/lullaby/Decoder.cpp"
 }
}
function __ZN4love5sound7lullaby7DecoderD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (32856); //@line 44 "modules/sound/lullaby/Decoder.cpp"
 $0 = ((($this)) + 32|0); //@line 45 "modules/sound/lullaby/Decoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 45 "modules/sound/lullaby/Decoder.cpp"
 $2 = ($1|0)==(0|0); //@line 45 "modules/sound/lullaby/Decoder.cpp"
 if (!($2)) {
  __ZdaPv($1); //@line 46 "modules/sound/lullaby/Decoder.cpp"
 }
 $3 = ((($this)) + 12|0); //@line 47 "modules/sound/lullaby/Decoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3); //@line 47 "modules/sound/lullaby/Decoder.cpp"
 $4 = ((($this)) + 8|0); //@line 106 "./common/Object.h"
 $5 = HEAP32[$4>>2]|0; //@line 106 "./common/Object.h"
 $6 = ($5|0)==(0|0); //@line 106 "./common/Object.h"
 if ($6) {
  __ZdlPv($this); //@line 44 "modules/sound/lullaby/Decoder.cpp"
  return; //@line 47 "modules/sound/lullaby/Decoder.cpp"
 }
 $7 = HEAP32[$5>>2]|0; //@line 106 "./common/Object.h"
 $8 = ((($7)) + 12|0); //@line 106 "./common/Object.h"
 $9 = HEAP32[$8>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($9|0,($5|0)); //@line 106 "./common/Object.h"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $12 = ___cxa_find_matching_catch_3(0|0)|0;
  $13 = tempRet0;
  ___clang_call_terminate($12); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  __ZdlPv($this); //@line 44 "modules/sound/lullaby/Decoder.cpp"
  return; //@line 47 "modules/sound/lullaby/Decoder.cpp"
 }
}
function __ZNK4love5sound7lullaby7Decoder7getSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 56 "modules/sound/lullaby/Decoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 56 "modules/sound/lullaby/Decoder.cpp"
 return ($1|0); //@line 56 "modules/sound/lullaby/Decoder.cpp"
}
function __ZNK4love5sound7lullaby7Decoder9getBufferEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 51 "modules/sound/lullaby/Decoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 51 "modules/sound/lullaby/Decoder.cpp"
 return ($1|0); //@line 51 "modules/sound/lullaby/Decoder.cpp"
}
function __ZN4love5sound7lullaby7Decoder10isFinishedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 36|0); //@line 66 "modules/sound/lullaby/Decoder.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 66 "modules/sound/lullaby/Decoder.cpp"
 $2 = ($1<<24>>24)!=(0); //@line 66 "modules/sound/lullaby/Decoder.cpp"
 return ($2|0); //@line 66 "modules/sound/lullaby/Decoder.cpp"
}
function __ZNK4love5sound7lullaby7Decoder13getSampleRateEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 61 "modules/sound/lullaby/Decoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 61 "modules/sound/lullaby/Decoder.cpp"
 return ($1|0); //@line 61 "modules/sound/lullaby/Decoder.cpp"
}
function __ZN4love5sound7lullaby7DecoderC2EPNS_4DataERKNSt3__112basic_stringIcNS5_11char_traitsIcEENS5_9allocatorIcEEEEi($this,$data,$ext,$bufferSize) {
 $this = $this|0;
 $data = $data|0;
 $ext = $ext|0;
 $bufferSize = $bufferSize|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 35 "./modules/sound/Decoder.h"
 HEAP32[$this>>2] = (32856); //@line 39 "modules/sound/lullaby/Decoder.cpp"
 $0 = ((($this)) + 8|0); //@line 93 "./common/Object.h"
 HEAP32[$0>>2] = $data; //@line 93 "./common/Object.h"
 $1 = ($data|0)==(0|0); //@line 95 "./common/Object.h"
 if (!($1)) {
  $2 = HEAP32[$data>>2]|0; //@line 95 "./common/Object.h"
  $3 = ((($2)) + 8|0); //@line 95 "./common/Object.h"
  $4 = HEAP32[$3>>2]|0; //@line 95 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($4|0,($data|0)); //@line 95 "./common/Object.h"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if ($6) {
   $19 = ___cxa_find_matching_catch_2()|0;
   $20 = tempRet0;
   $$1 = $20;$$12 = $19;
   ___resumeException($$12|0);
   // unreachable;
  }
 }
 $7 = ((($this)) + 12|0); //@line 34 "modules/sound/lullaby/Decoder.cpp"
 __THREW__ = 0;
 invoke_vii(3351,($7|0),($ext|0)); //@line 34 "modules/sound/lullaby/Decoder.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 do {
  if ($9) {
   $21 = ___cxa_find_matching_catch_2()|0;
   $22 = tempRet0;
   $$0 = $22;$$01 = $21;
  } else {
   $10 = ((($this)) + 24|0); //@line 35 "modules/sound/lullaby/Decoder.cpp"
   HEAP32[$10>>2] = $bufferSize; //@line 35 "modules/sound/lullaby/Decoder.cpp"
   $11 = ((($this)) + 28|0); //@line 36 "modules/sound/lullaby/Decoder.cpp"
   HEAP32[$11>>2] = 44100; //@line 36 "modules/sound/lullaby/Decoder.cpp"
   $12 = ((($this)) + 32|0); //@line 37 "modules/sound/lullaby/Decoder.cpp"
   HEAP32[$12>>2] = 0; //@line 37 "modules/sound/lullaby/Decoder.cpp"
   $13 = ((($this)) + 36|0); //@line 38 "modules/sound/lullaby/Decoder.cpp"
   HEAP8[$13>>0] = 0; //@line 38 "modules/sound/lullaby/Decoder.cpp"
   $14 = ($bufferSize|0)>(-1); //@line 40 "modules/sound/lullaby/Decoder.cpp"
   $15 = $14 ? $bufferSize : -1; //@line 40 "modules/sound/lullaby/Decoder.cpp"
   __THREW__ = 0;
   $16 = (invoke_ii(3492,($15|0))|0); //@line 40 "modules/sound/lullaby/Decoder.cpp"
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if ($18) {
    $23 = ___cxa_find_matching_catch_2()|0;
    $24 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7); //@line 41 "modules/sound/lullaby/Decoder.cpp"
    $$0 = $24;$$01 = $23;
    break;
   } else {
    HEAP32[$12>>2] = $16; //@line 40 "modules/sound/lullaby/Decoder.cpp"
    return; //@line 41 "modules/sound/lullaby/Decoder.cpp"
   }
  }
 } while(0);
 $25 = HEAP32[$0>>2]|0; //@line 106 "./common/Object.h"
 $26 = ($25|0)==(0|0); //@line 106 "./common/Object.h"
 if ($26) {
  $$1 = $$0;$$12 = $$01;
  ___resumeException($$12|0);
  // unreachable;
 }
 $27 = HEAP32[$25>>2]|0; //@line 106 "./common/Object.h"
 $28 = ((($27)) + 12|0); //@line 106 "./common/Object.h"
 $29 = HEAP32[$28>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($29|0,($25|0)); //@line 106 "./common/Object.h"
 $30 = __THREW__; __THREW__ = 0;
 $31 = $30&1;
 if ($31) {
  $32 = ___cxa_find_matching_catch_3(0|0)|0;
  $33 = tempRet0;
  ___clang_call_terminate($32); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  $$1 = $$0;$$12 = $$01;
  ___resumeException($$12|0);
  // unreachable;
 }
}
function __ZN4love5sound7lullaby14ModPlugDecoderD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (32928); //@line 74 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $0 = ((($this)) + 40|0); //@line 75 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 75 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $2 = ($1|0)==(0|0); //@line 75 "modules/sound/lullaby/ModPlugDecoder.cpp"
 if ($2) {
  __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 77 "modules/sound/lullaby/ModPlugDecoder.cpp"
  return; //@line 77 "modules/sound/lullaby/ModPlugDecoder.cpp"
 }
 __THREW__ = 0;
 invoke_vi(4219,($1|0)); //@line 76 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $5 = ___cxa_find_matching_catch_3(0|0)|0;
  $6 = tempRet0;
  __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 77 "modules/sound/lullaby/ModPlugDecoder.cpp"
  ___clang_call_terminate($5); //@line 77 "modules/sound/lullaby/ModPlugDecoder.cpp"
  // unreachable; //@line 77 "modules/sound/lullaby/ModPlugDecoder.cpp"
 } else {
  __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 77 "modules/sound/lullaby/ModPlugDecoder.cpp"
  return; //@line 77 "modules/sound/lullaby/ModPlugDecoder.cpp"
 }
}
function __ZN4love5sound7lullaby14ModPlugDecoderD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (32928); //@line 74 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $0 = ((($this)) + 40|0); //@line 75 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 75 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $2 = ($1|0)==(0|0); //@line 75 "modules/sound/lullaby/ModPlugDecoder.cpp"
 if (!($2)) {
  __THREW__ = 0;
  invoke_vi(4219,($1|0)); //@line 76 "modules/sound/lullaby/ModPlugDecoder.cpp"
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if ($4) {
   $5 = ___cxa_find_matching_catch_3(0|0)|0;
   $6 = tempRet0;
   __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 77 "modules/sound/lullaby/ModPlugDecoder.cpp"
   ___clang_call_terminate($5); //@line 77 "modules/sound/lullaby/ModPlugDecoder.cpp"
   // unreachable; //@line 77 "modules/sound/lullaby/ModPlugDecoder.cpp"
  }
 }
 __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 77 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZdlPv($this); //@line 74 "modules/sound/lullaby/ModPlugDecoder.cpp"
 return; //@line 77 "modules/sound/lullaby/ModPlugDecoder.cpp"
}
function __ZN4love5sound7lullaby14ModPlugDecoder5cloneEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(112)|0); //@line 101 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $1 = ((($this)) + 8|0); //@line 139 "./common/Object.h"
 $2 = HEAP32[$1>>2]|0; //@line 139 "./common/Object.h"
 $3 = ((($this)) + 12|0); //@line 101 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $4 = ((($this)) + 24|0); //@line 101 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 101 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __THREW__ = 0;
 invoke_viiii(3750,($0|0),($2|0),($3|0),($5|0)); //@line 101 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $8 = ___cxa_find_matching_catch_2()|0;
  $9 = tempRet0;
  __ZdlPv($0); //@line 101 "modules/sound/lullaby/ModPlugDecoder.cpp"
  ___resumeException($8|0);
  // unreachable;
 } else {
  return ($0|0); //@line 101 "modules/sound/lullaby/ModPlugDecoder.cpp"
 }
 return (0)|0;
}
function __ZN4love5sound7lullaby14ModPlugDecoderC2EPNS_4DataERKNSt3__112basic_stringIcNS5_11char_traitsIcEENS5_9allocatorIcEEEEi($this,$data,$ext,$bufferSize) {
 $this = $this|0;
 $data = $data|0;
 $ext = $ext|0;
 $bufferSize = $bufferSize|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 __ZN4love5sound7lullaby7DecoderC2EPNS_4DataERKNSt3__112basic_stringIcNS5_11char_traitsIcEENS5_9allocatorIcEEEEi($this,$data,$ext,$bufferSize); //@line 35 "modules/sound/lullaby/ModPlugDecoder.cpp"
 HEAP32[$this>>2] = (32928); //@line 38 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $0 = ((($this)) + 40|0); //@line 36 "modules/sound/lullaby/ModPlugDecoder.cpp"
 HEAP32[$0>>2] = 0; //@line 36 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $1 = ((($this)) + 104|0); //@line 37 "modules/sound/lullaby/ModPlugDecoder.cpp"
 HEAPF64[$1>>3] = -2.0; //@line 37 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $2 = ((($this)) + 44|0); //@line 41 "modules/sound/lullaby/ModPlugDecoder.cpp"
 HEAP32[$2>>2] = 3; //@line 41 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $3 = ((($this)) + 48|0); //@line 42 "modules/sound/lullaby/ModPlugDecoder.cpp"
 HEAP32[$3>>2] = 2; //@line 42 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $4 = ((($this)) + 52|0); //@line 43 "modules/sound/lullaby/ModPlugDecoder.cpp"
 HEAP32[$4>>2] = 16; //@line 43 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $5 = ((($this)) + 28|0); //@line 44 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 44 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $7 = ((($this)) + 56|0); //@line 44 "modules/sound/lullaby/ModPlugDecoder.cpp"
 HEAP32[$7>>2] = $6; //@line 44 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $8 = ((($this)) + 60|0); //@line 45 "modules/sound/lullaby/ModPlugDecoder.cpp"
 HEAP32[$8>>2] = 1; //@line 45 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $9 = ((($this)) + 64|0); //@line 51 "modules/sound/lullaby/ModPlugDecoder.cpp"
 HEAP32[$9>>2] = 128; //@line 51 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $10 = ((($this)) + 68|0); //@line 52 "modules/sound/lullaby/ModPlugDecoder.cpp"
 HEAP32[$10>>2] = 32; //@line 52 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $11 = ((($this)) + 72|0); //@line 53 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $12 = ((($this)) + 96|0); //@line 59 "modules/sound/lullaby/ModPlugDecoder.cpp"
 ;HEAP32[$11>>2]=0|0;HEAP32[$11+4>>2]=0|0;HEAP32[$11+8>>2]=0|0;HEAP32[$11+12>>2]=0|0;HEAP32[$11+16>>2]=0|0;HEAP32[$11+20>>2]=0|0; //@line 54 "modules/sound/lullaby/ModPlugDecoder.cpp"
 HEAP32[$12>>2] = -1; //@line 59 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __THREW__ = 0;
 invoke_vi(4220,($2|0)); //@line 61 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 do {
  if (!($14)) {
   $15 = HEAP32[$data>>2]|0; //@line 64 "modules/sound/lullaby/ModPlugDecoder.cpp"
   $16 = ((($15)) + 16|0); //@line 64 "modules/sound/lullaby/ModPlugDecoder.cpp"
   $17 = HEAP32[$16>>2]|0; //@line 64 "modules/sound/lullaby/ModPlugDecoder.cpp"
   __THREW__ = 0;
   $18 = (invoke_ii($17|0,($data|0))|0); //@line 64 "modules/sound/lullaby/ModPlugDecoder.cpp"
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   if (!($20)) {
    $21 = HEAP32[$data>>2]|0; //@line 64 "modules/sound/lullaby/ModPlugDecoder.cpp"
    $22 = ((($21)) + 20|0); //@line 64 "modules/sound/lullaby/ModPlugDecoder.cpp"
    $23 = HEAP32[$22>>2]|0; //@line 64 "modules/sound/lullaby/ModPlugDecoder.cpp"
    __THREW__ = 0;
    $24 = (invoke_ii($23|0,($data|0))|0); //@line 64 "modules/sound/lullaby/ModPlugDecoder.cpp"
    $25 = __THREW__; __THREW__ = 0;
    $26 = $25&1;
    if (!($26)) {
     __THREW__ = 0;
     $27 = (invoke_iii(4221,($18|0),($24|0))|0); //@line 64 "modules/sound/lullaby/ModPlugDecoder.cpp"
     $28 = __THREW__; __THREW__ = 0;
     $29 = $28&1;
     if (!($29)) {
      HEAP32[$0>>2] = $27; //@line 64 "modules/sound/lullaby/ModPlugDecoder.cpp"
      $30 = ($27|0)==(0|0); //@line 66 "modules/sound/lullaby/ModPlugDecoder.cpp"
      if (!($30)) {
       __THREW__ = 0;
       invoke_vii(4222,($27|0),128); //@line 70 "modules/sound/lullaby/ModPlugDecoder.cpp"
       $39 = __THREW__; __THREW__ = 0;
       $40 = $39&1;
       if ($40) {
        break;
       }
       STACKTOP = sp;return; //@line 71 "modules/sound/lullaby/ModPlugDecoder.cpp"
      }
      $31 = (___cxa_allocate_exception(16)|0); //@line 67 "modules/sound/lullaby/ModPlugDecoder.cpp"
      __THREW__ = 0;
      invoke_viii(3349,($31|0),(445150|0),($vararg_buffer|0)); //@line 67 "modules/sound/lullaby/ModPlugDecoder.cpp"
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if (!($33)) {
       __THREW__ = 0;
       invoke_viii(3350,($31|0),(96|0),(344|0)); //@line 67 "modules/sound/lullaby/ModPlugDecoder.cpp"
       $34 = __THREW__; __THREW__ = 0;
       break;
      }
      $37 = ___cxa_find_matching_catch_2()|0;
      $38 = tempRet0;
      ___cxa_free_exception(($31|0)); //@line 67 "modules/sound/lullaby/ModPlugDecoder.cpp"
      $$0 = $38;$$01 = $37;
      __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 71 "modules/sound/lullaby/ModPlugDecoder.cpp"
      ___resumeException($$01|0);
      // unreachable;
     }
    }
   }
  }
 } while(0);
 $35 = ___cxa_find_matching_catch_2()|0;
 $36 = tempRet0;
 $$0 = $36;$$01 = $35;
 __ZN4love5sound7lullaby7DecoderD2Ev($this); //@line 71 "modules/sound/lullaby/ModPlugDecoder.cpp"
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN4love5sound7lullaby14ModPlugDecoder6decodeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0); //@line 106 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 106 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $2 = ((($this)) + 32|0); //@line 106 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 106 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $4 = ((($this)) + 24|0); //@line 106 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 106 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $6 = (_ModPlug_Read($1,$3,$5)|0); //@line 106 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $7 = ($6|0)==(0); //@line 108 "modules/sound/lullaby/ModPlugDecoder.cpp"
 if (!($7)) {
  return ($6|0); //@line 111 "modules/sound/lullaby/ModPlugDecoder.cpp"
 }
 $8 = ((($this)) + 36|0); //@line 109 "modules/sound/lullaby/ModPlugDecoder.cpp"
 HEAP8[$8>>0] = 1; //@line 109 "modules/sound/lullaby/ModPlugDecoder.cpp"
 return ($6|0); //@line 111 "modules/sound/lullaby/ModPlugDecoder.cpp"
}
function __ZN4love5sound7lullaby14ModPlugDecoder4seekEf($this,$s) {
 $this = $this|0;
 $s = +$s;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0); //@line 116 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 116 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $2 = $s * 1000.0; //@line 116 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $3 = (~~(($2))); //@line 116 "modules/sound/lullaby/ModPlugDecoder.cpp"
 _ModPlug_Seek($1,$3); //@line 116 "modules/sound/lullaby/ModPlugDecoder.cpp"
 return 1; //@line 117 "modules/sound/lullaby/ModPlugDecoder.cpp"
}
function __ZN4love5sound7lullaby14ModPlugDecoder6rewindEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0); //@line 123 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 123 "modules/sound/lullaby/ModPlugDecoder.cpp"
 _ModPlug_Unload($1); //@line 123 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $2 = ((($this)) + 8|0); //@line 117 "./common/Object.h"
 $3 = HEAP32[$2>>2]|0; //@line 117 "./common/Object.h"
 $4 = HEAP32[$3>>2]|0; //@line 124 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $5 = ((($4)) + 16|0); //@line 124 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 124 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $7 = (FUNCTION_TABLE_ii[$6 & 8191]($3)|0); //@line 124 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $8 = HEAP32[$2>>2]|0; //@line 117 "./common/Object.h"
 $9 = HEAP32[$8>>2]|0; //@line 124 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $10 = ((($9)) + 20|0); //@line 124 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 124 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($8)|0); //@line 124 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $13 = (_ModPlug_Load($7,$12)|0); //@line 124 "modules/sound/lullaby/ModPlugDecoder.cpp"
 HEAP32[$0>>2] = $13; //@line 124 "modules/sound/lullaby/ModPlugDecoder.cpp"
 _ModPlug_SetMasterVolume($13,128); //@line 125 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $14 = ((($this)) + 36|0); //@line 126 "modules/sound/lullaby/ModPlugDecoder.cpp"
 HEAP8[$14>>0] = 0; //@line 126 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $15 = HEAP32[$0>>2]|0; //@line 127 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $16 = ($15|0)!=(0|0); //@line 127 "modules/sound/lullaby/ModPlugDecoder.cpp"
 return ($16|0); //@line 127 "modules/sound/lullaby/ModPlugDecoder.cpp"
}
function __ZN4love5sound7lullaby14ModPlugDecoder10isSeekableEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1; //@line 132 "modules/sound/lullaby/ModPlugDecoder.cpp"
}
function __ZNK4love5sound7lullaby14ModPlugDecoder11getChannelsEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 2; //@line 137 "modules/sound/lullaby/ModPlugDecoder.cpp"
}
function __ZNK4love5sound7lullaby14ModPlugDecoder11getBitDepthEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 16; //@line 142 "modules/sound/lullaby/ModPlugDecoder.cpp"
}
function __ZN4love5sound7lullaby14ModPlugDecoder11getDurationEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $storemerge = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 104|0); //@line 148 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $1 = +HEAPF64[$0>>3]; //@line 148 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $2 = $1 == -2.0; //@line 148 "modules/sound/lullaby/ModPlugDecoder.cpp"
 if (!($2)) {
  $9 = $1;
  return (+$9); //@line 158 "modules/sound/lullaby/ModPlugDecoder.cpp"
 }
 $3 = ((($this)) + 40|0); //@line 150 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 150 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $5 = (_ModPlug_GetLength($4)|0); //@line 150 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $6 = ($5|0)<(0); //@line 152 "modules/sound/lullaby/ModPlugDecoder.cpp"
 if ($6) {
  $storemerge = -1.0;
 } else {
  $7 = (+($5|0)); //@line 155 "modules/sound/lullaby/ModPlugDecoder.cpp"
  $8 = $7 / 1000.0; //@line 155 "modules/sound/lullaby/ModPlugDecoder.cpp"
  $storemerge = $8;
 }
 HEAPF64[$0>>3] = $storemerge; //@line 153 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $9 = $storemerge;
 return (+$9); //@line 158 "modules/sound/lullaby/ModPlugDecoder.cpp"
}
function __ZN4love5sound7lullaby14ModPlugDecoder7acceptsERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($ext) {
 $ext = $ext|0;
 var $$02$ph = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $i$0 = 0, $lpad$phi28$index = 0, $lpad$phi28$index2 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[624088]|0; //@line 81 "modules/sound/lullaby/ModPlugDecoder.cpp"
 $1 = ($0<<24>>24)==(0); //@line 81 "modules/sound/lullaby/ModPlugDecoder.cpp"
 L1: do {
  if ($1) {
   $2 = (___cxa_guard_acquire((624088|0))|0); //@line 81 "modules/sound/lullaby/ModPlugDecoder.cpp"
   $3 = ($2|0)==(0); //@line 81 "modules/sound/lullaby/ModPlugDecoder.cpp"
   if (!($3)) {
    __THREW__ = 0;
    invoke_viii(3460,(658128|0),(445184|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $4 = __THREW__; __THREW__ = 0;
    $5 = $4&1;
    if ($5) {
     $78 = ___cxa_find_matching_catch_2()|0;
     $79 = tempRet0;
     $lpad$phi28$index = $78;$lpad$phi28$index2 = $79;
     ___cxa_guard_abort((624088|0)); //@line 81 "modules/sound/lullaby/ModPlugDecoder.cpp"
     ___resumeException($lpad$phi28$index|0);
     // unreachable;
    }
    __THREW__ = 0;
    invoke_viii(3460,((658140)|0),(445188|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $6 = __THREW__; __THREW__ = 0;
    $7 = $6&1;
    do {
     if ($7) {
      $$02$ph = (658140);
     } else {
      __THREW__ = 0;
      invoke_viii(3460,((658152)|0),(445192|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $8 = __THREW__; __THREW__ = 0;
      $9 = $8&1;
      if ($9) {
       $$02$ph = (658152);
      } else {
       __THREW__ = 0;
       invoke_viii(3460,((658164)|0),(445196|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
       $10 = __THREW__; __THREW__ = 0;
       $11 = $10&1;
       if ($11) {
        $$02$ph = (658164);
       } else {
        __THREW__ = 0;
        invoke_viii(3460,((658176)|0),(445200|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
        $12 = __THREW__; __THREW__ = 0;
        $13 = $12&1;
        if ($13) {
         $$02$ph = (658176);
        } else {
         __THREW__ = 0;
         invoke_viii(3460,((658188)|0),(445204|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
         $14 = __THREW__; __THREW__ = 0;
         $15 = $14&1;
         if ($15) {
          $$02$ph = (658188);
         } else {
          __THREW__ = 0;
          invoke_viii(3460,((658200)|0),(445208|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
          $16 = __THREW__; __THREW__ = 0;
          $17 = $16&1;
          if ($17) {
           $$02$ph = (658200);
          } else {
           __THREW__ = 0;
           invoke_viii(3460,((658212)|0),(445212|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $18 = __THREW__; __THREW__ = 0;
           $19 = $18&1;
           if ($19) {
            $$02$ph = (658212);
           } else {
            __THREW__ = 0;
            invoke_viii(3460,((658224)|0),(445216|0),2); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
            $20 = __THREW__; __THREW__ = 0;
            $21 = $20&1;
            if ($21) {
             $$02$ph = (658224);
            } else {
             __THREW__ = 0;
             invoke_viii(3460,((658236)|0),(445219|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
             $22 = __THREW__; __THREW__ = 0;
             $23 = $22&1;
             if ($23) {
              $$02$ph = (658236);
             } else {
              __THREW__ = 0;
              invoke_viii(3460,((658248)|0),(445223|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
              $24 = __THREW__; __THREW__ = 0;
              $25 = $24&1;
              if ($25) {
               $$02$ph = (658248);
              } else {
               __THREW__ = 0;
               invoke_viii(3460,((658260)|0),(445227|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
               $26 = __THREW__; __THREW__ = 0;
               $27 = $26&1;
               if ($27) {
                $$02$ph = (658260);
               } else {
                __THREW__ = 0;
                invoke_viii(3460,((658272)|0),(445231|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                $28 = __THREW__; __THREW__ = 0;
                $29 = $28&1;
                if ($29) {
                 $$02$ph = (658272);
                } else {
                 __THREW__ = 0;
                 invoke_viii(3460,((658284)|0),(522049|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                 $30 = __THREW__; __THREW__ = 0;
                 $31 = $30&1;
                 if ($31) {
                  $$02$ph = (658284);
                 } else {
                  __THREW__ = 0;
                  invoke_viii(3460,((658296)|0),(445235|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                  $32 = __THREW__; __THREW__ = 0;
                  $33 = $32&1;
                  if ($33) {
                   $$02$ph = (658296);
                  } else {
                   __THREW__ = 0;
                   invoke_viii(3460,((658308)|0),(445239|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                   $34 = __THREW__; __THREW__ = 0;
                   $35 = $34&1;
                   if ($35) {
                    $$02$ph = (658308);
                   } else {
                    __THREW__ = 0;
                    invoke_viii(3460,((658320)|0),(445243|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                    $36 = __THREW__; __THREW__ = 0;
                    $37 = $36&1;
                    if ($37) {
                     $$02$ph = (658320);
                     break;
                    }
                    __THREW__ = 0;
                    invoke_viii(3460,((658332)|0),(445247|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                    $38 = __THREW__; __THREW__ = 0;
                    $39 = $38&1;
                    if ($39) {
                     $$02$ph = (658332);
                     break;
                    }
                    __THREW__ = 0;
                    invoke_viii(3460,((658344)|0),(445251|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                    $40 = __THREW__; __THREW__ = 0;
                    $41 = $40&1;
                    if ($41) {
                     $$02$ph = (658344);
                     break;
                    }
                    __THREW__ = 0;
                    invoke_viii(3460,((658356)|0),(445255|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                    $42 = __THREW__; __THREW__ = 0;
                    $43 = $42&1;
                    if ($43) {
                     $$02$ph = (658356);
                     break;
                    }
                    __THREW__ = 0;
                    invoke_viii(3460,((658368)|0),(445259|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                    $44 = __THREW__; __THREW__ = 0;
                    $45 = $44&1;
                    if ($45) {
                     $$02$ph = (658368);
                     break;
                    }
                    __THREW__ = 0;
                    invoke_viii(3460,((658380)|0),(445263|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                    $46 = __THREW__; __THREW__ = 0;
                    $47 = $46&1;
                    if ($47) {
                     $$02$ph = (658380);
                     break;
                    }
                    __THREW__ = 0;
                    invoke_viii(3460,((658392)|0),(445267|0),3); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                    $48 = __THREW__; __THREW__ = 0;
                    $49 = $48&1;
                    if ($49) {
                     $$02$ph = (658392);
                     break;
                    }
                    __THREW__ = 0;
                    invoke_viii(3460,((658404)|0),(445271|0),2); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                    $50 = __THREW__; __THREW__ = 0;
                    $51 = $50&1;
                    if ($51) {
                     $$02$ph = (658404);
                     break;
                    }
                    __THREW__ = 0;
                    invoke_viii(3460,((658416)|0),(1282466|0),0); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
                    $52 = __THREW__; __THREW__ = 0;
                    $53 = $52&1;
                    if ($53) {
                     $$02$ph = (658416);
                     break;
                    }
                    (___cxa_atexit((4223|0),(0|0),(___dso_handle|0))|0); //@line 81 "modules/sound/lullaby/ModPlugDecoder.cpp"
                    ___cxa_guard_release((624088|0)); //@line 81 "modules/sound/lullaby/ModPlugDecoder.cpp"
                    break L1;
                   }
                  }
                 }
                }
               }
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    } while(0);
    $76 = ___cxa_find_matching_catch_2()|0;
    $77 = tempRet0;
    $81 = $$02$ph;
    while(1) {
     $80 = ((($81)) + -12|0); //@line 82 "modules/sound/lullaby/ModPlugDecoder.cpp"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($80); //@line 82 "modules/sound/lullaby/ModPlugDecoder.cpp"
     $82 = ($80|0)==(658128|0); //@line 82 "modules/sound/lullaby/ModPlugDecoder.cpp"
     if ($82) {
      $lpad$phi28$index = $76;$lpad$phi28$index2 = $77;
      break;
     } else {
      $81 = $80;
     }
    }
    ___cxa_guard_abort((624088|0)); //@line 81 "modules/sound/lullaby/ModPlugDecoder.cpp"
    ___resumeException($lpad$phi28$index|0);
    // unreachable;
   }
  }
 } while(0);
 $54 = HEAP8[$ext>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $55 = $54 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $56 = ($55<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $57 = ((($ext)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $58 = HEAP32[$57>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $59 = $54&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $60 = $59 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $61 = $56 ? $60 : $58; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $62 = ((($ext)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $63 = HEAP32[$62>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $64 = ((($ext)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $65 = $56 ? $64 : $63; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $i$0 = 0;
 while(1) {
  $66 = (658128 + (($i$0*12)|0)|0); //@line 90 "modules/sound/lullaby/ModPlugDecoder.cpp"
  $67 = HEAP8[$66>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $68 = $67 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $69 = ($68<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($69) {
   $72 = $67&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $73 = $72 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $75 = $73;
  } else {
   $70 = (((658128 + (($i$0*12)|0)|0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $71 = HEAP32[$70>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $75 = $71;
  }
  $74 = ($75|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($74) {
   $99 = 0;
   label = 44;
   break;
  }
  if ($69) {
   $87 = $67&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $88 = $87 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $89 = ((($66)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $91 = $88;$93 = $89;
  } else {
   $83 = (((658128 + (($i$0*12)|0)|0)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $84 = HEAP32[$83>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $85 = (((658128 + (($i$0*12)|0)|0)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $86 = HEAP32[$85>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $91 = $84;$93 = $86;
  }
  $90 = ($61>>>0)<($91>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $92 = $90 ? $61 : $91; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $94 = (_memcmp($93,$65,$92)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $95 = ($94|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($95) {
   $96 = ($91>>>0)>=($61>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $97 = $90 ^ 1; //@line 92 "modules/sound/lullaby/ModPlugDecoder.cpp"
   $or$cond = $96 & $97;
   if ($or$cond) {
    $99 = 1;
    label = 44;
    break;
   }
  }
  $98 = (($i$0) + 1)|0; //@line 90 "modules/sound/lullaby/ModPlugDecoder.cpp"
  $i$0 = $98;
 }
 if ((label|0) == 44) {
  return ($99|0); //@line 97 "modules/sound/lullaby/ModPlugDecoder.cpp"
 }
 return (0)|0;
}
function ___cxx_global_array_dtor_4964($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658416)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658404)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658392)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658380)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658368)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658356)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658344)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658332)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658320)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658308)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658296)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658284)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658272)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658260)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658248)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658236)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658224)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658212)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658200)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658188)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658176)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658164)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658152)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((658140)); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev(658128); //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
 return; //@line 87 "modules/sound/lullaby/ModPlugDecoder.cpp"
}
function __ZN4love6thread3sdl5MutexD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (33000); //@line 37 "modules/thread/sdl/threads.cpp"
 $0 = ((($this)) + 4|0); //@line 38 "modules/thread/sdl/threads.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 38 "modules/thread/sdl/threads.cpp"
 __THREW__ = 0;
 invoke_vi(4224,($1|0)); //@line 38 "modules/thread/sdl/threads.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  ___clang_call_terminate($4); //@line 39 "modules/thread/sdl/threads.cpp"
  // unreachable; //@line 39 "modules/thread/sdl/threads.cpp"
 } else {
  return; //@line 39 "modules/thread/sdl/threads.cpp"
 }
}
function __ZN4love6thread3sdl5MutexD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (33000); //@line 37 "modules/thread/sdl/threads.cpp"
 $0 = ((($this)) + 4|0); //@line 38 "modules/thread/sdl/threads.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 38 "modules/thread/sdl/threads.cpp"
 __THREW__ = 0;
 invoke_vi(4224,($1|0)); //@line 38 "modules/thread/sdl/threads.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  ___clang_call_terminate($4); //@line 39 "modules/thread/sdl/threads.cpp"
  // unreachable; //@line 39 "modules/thread/sdl/threads.cpp"
 } else {
  __ZdlPv($this); //@line 37 "modules/thread/sdl/threads.cpp"
  return; //@line 39 "modules/thread/sdl/threads.cpp"
 }
}
function __ZN4love6thread3sdl5Mutex4lockEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 44 "modules/thread/sdl/threads.cpp"
}
function __ZN4love6thread3sdl5Mutex6unlockEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 49 "modules/thread/sdl/threads.cpp"
}
function __ZN4love6thread3sdl11ConditionalD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (33024); //@line 57 "modules/thread/sdl/threads.cpp"
 $0 = ((($this)) + 4|0); //@line 58 "modules/thread/sdl/threads.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 58 "modules/thread/sdl/threads.cpp"
 __THREW__ = 0;
 invoke_vi(4225,($1|0)); //@line 58 "modules/thread/sdl/threads.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  ___clang_call_terminate($4); //@line 59 "modules/thread/sdl/threads.cpp"
  // unreachable; //@line 59 "modules/thread/sdl/threads.cpp"
 } else {
  return; //@line 59 "modules/thread/sdl/threads.cpp"
 }
}
function __ZN4love6thread3sdl11ConditionalD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (33024); //@line 57 "modules/thread/sdl/threads.cpp"
 $0 = ((($this)) + 4|0); //@line 58 "modules/thread/sdl/threads.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 58 "modules/thread/sdl/threads.cpp"
 __THREW__ = 0;
 invoke_vi(4225,($1|0)); //@line 58 "modules/thread/sdl/threads.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  ___clang_call_terminate($4); //@line 59 "modules/thread/sdl/threads.cpp"
  // unreachable; //@line 59 "modules/thread/sdl/threads.cpp"
 } else {
  __ZdlPv($this); //@line 57 "modules/thread/sdl/threads.cpp"
  return; //@line 59 "modules/thread/sdl/threads.cpp"
 }
}
function __ZN4love6thread3sdl11Conditional6signalEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 63 "modules/thread/sdl/threads.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 63 "modules/thread/sdl/threads.cpp"
 (_SDL_CondSignal($1)|0); //@line 63 "modules/thread/sdl/threads.cpp"
 return; //@line 64 "modules/thread/sdl/threads.cpp"
}
function __ZN4love6thread3sdl11Conditional9broadcastEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 68 "modules/thread/sdl/threads.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 68 "modules/thread/sdl/threads.cpp"
 (_SDL_CondBroadcast($1)|0); //@line 68 "modules/thread/sdl/threads.cpp"
 return; //@line 69 "modules/thread/sdl/threads.cpp"
}
function __ZN4love6thread3sdl11Conditional4waitEPNS0_5MutexEi($this,$_mutex,$timeout) {
 $this = $this|0;
 $_mutex = $_mutex|0;
 $timeout = $timeout|0;
 var $$0 = 0, $$0$in = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($timeout|0)<(0); //@line 77 "modules/thread/sdl/threads.cpp"
 $1 = ((($this)) + 4|0); //@line 78 "modules/thread/sdl/threads.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 78 "modules/thread/sdl/threads.cpp"
 $3 = ((($_mutex)) + 4|0); //@line 78 "modules/thread/sdl/threads.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 78 "modules/thread/sdl/threads.cpp"
 if ($0) {
  $5 = (_SDL_CondWait($2,$4)|0); //@line 78 "modules/thread/sdl/threads.cpp"
  $$0$in = $5;
 } else {
  $6 = (_SDL_CondWaitTimeout($2,$4,$timeout)|0); //@line 80 "modules/thread/sdl/threads.cpp"
  $$0$in = $6;
 }
 $$0 = ($$0$in|0)==(0); //@line 78 "modules/thread/sdl/threads.cpp"
 return ($$0|0); //@line 81 "modules/thread/sdl/threads.cpp"
}
function __ZN4love6thread3sdl5MutexC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (33000); //@line 32 "modules/thread/sdl/threads.cpp"
 $0 = (_SDL_CreateMutex()|0); //@line 33 "modules/thread/sdl/threads.cpp"
 $1 = ((($this)) + 4|0); //@line 33 "modules/thread/sdl/threads.cpp"
 HEAP32[$1>>2] = $0; //@line 33 "modules/thread/sdl/threads.cpp"
 return; //@line 34 "modules/thread/sdl/threads.cpp"
}
function __ZN4love6thread8newMutexEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(8)|0); //@line 92 "modules/thread/sdl/threads.cpp"
 HEAP32[$0>>2] = (33000); //@line 32 "modules/thread/sdl/threads.cpp"
 __THREW__ = 0;
 $1 = (invoke_i(4226)|0); //@line 33 "modules/thread/sdl/threads.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  __ZdlPv($0); //@line 92 "modules/thread/sdl/threads.cpp"
  ___resumeException($5|0);
  // unreachable;
 } else {
  $4 = ((($0)) + 4|0); //@line 33 "modules/thread/sdl/threads.cpp"
  HEAP32[$4>>2] = $1; //@line 33 "modules/thread/sdl/threads.cpp"
  return ($0|0); //@line 92 "modules/thread/sdl/threads.cpp"
 }
 return (0)|0;
}
function __ZN4love6thread14newConditionalEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(8)|0); //@line 97 "modules/thread/sdl/threads.cpp"
 HEAP32[$0>>2] = (33024); //@line 52 "modules/thread/sdl/threads.cpp"
 __THREW__ = 0;
 $1 = (invoke_i(4227)|0); //@line 53 "modules/thread/sdl/threads.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  __ZdlPv($0); //@line 97 "modules/thread/sdl/threads.cpp"
  ___resumeException($5|0);
  // unreachable;
 } else {
  $4 = ((($0)) + 4|0); //@line 53 "modules/thread/sdl/threads.cpp"
  HEAP32[$4>>2] = $1; //@line 53 "modules/thread/sdl/threads.cpp"
  return ($0|0); //@line 97 "modules/thread/sdl/threads.cpp"
 }
 return (0)|0;
}
function __ZN4love6thread9newThreadEPNS0_10ThreadableE($t) {
 $t = $t|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(24)|0); //@line 102 "modules/thread/sdl/threads.cpp"
 __THREW__ = 0;
 invoke_vii(4228,($0|0),($t|0)); //@line 102 "modules/thread/sdl/threads.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZdlPv($0); //@line 102 "modules/thread/sdl/threads.cpp"
  ___resumeException($3|0);
  // unreachable;
 } else {
  return ($0|0); //@line 102 "modules/thread/sdl/threads.cpp"
 }
 return (0)|0;
}
function __ZN4love6thread7ChannelD2Ev($this) {
 $this = $this|0;
 var $$0$i$i$i$i$i = 0, $$0$i$i$i$i$i$i = 0, $$01$i$i$i$i$i = 0, $$pre = 0, $$pre4 = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (33052); //@line 90 "modules/thread/Channel.cpp"
 $0 = ((($this)) + 16|0); //@line 91 "modules/thread/Channel.cpp"
 $1 = ((($this)) + 36|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = HEAP32[$1>>2]|0; //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $3 = ($2|0)==(0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 L1: do {
  if (!($3)) {
   $4 = ((($this)) + 20|0); //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $5 = ((($this)) + 32|0); //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $$pre = HEAP32[$5>>2]|0; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $8 = $$pre;
   while(1) {
    $6 = HEAP32[$4>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $7 = $8 >>> 10; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $9 = (($6) + ($7<<2)|0); //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $10 = HEAP32[$9>>2]|0; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $11 = $8 & 1023; //@line 1734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $12 = (($10) + ($11<<2)|0); //@line 1734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $13 = HEAP32[$12>>2]|0; //@line 93 "modules/thread/Channel.cpp"
    $14 = HEAP32[$13>>2]|0; //@line 93 "modules/thread/Channel.cpp"
    $15 = ((($14)) + 12|0); //@line 93 "modules/thread/Channel.cpp"
    $16 = HEAP32[$15>>2]|0; //@line 93 "modules/thread/Channel.cpp"
    __THREW__ = 0;
    invoke_vi($16|0,($13|0)); //@line 93 "modules/thread/Channel.cpp"
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if ($18) {
     break;
    }
    $19 = HEAP32[$1>>2]|0; //@line 2538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $20 = (($19) + -1)|0; //@line 2538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    HEAP32[$1>>2] = $20; //@line 2538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $21 = HEAP32[$5>>2]|0; //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $22 = (($21) + 1)|0; //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    HEAP32[$5>>2] = $22; //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $23 = ($22>>>0)>(2047); //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    if ($23) {
     $26 = HEAP32[$4>>2]|0; //@line 90 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $27 = HEAP32[$26>>2]|0; //@line 2541 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
     __ZdlPv($27); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $28 = HEAP32[$4>>2]|0; //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $29 = ((($28)) + 4|0); //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     HEAP32[$4>>2] = $29; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $30 = HEAP32[$5>>2]|0; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
     $31 = (($30) + -1024)|0; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
     HEAP32[$5>>2] = $31; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
     $$pre4 = HEAP32[$1>>2]|0; //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
     $25 = $$pre4;$68 = $31;
    } else {
     $25 = $20;$68 = $22;
    }
    $24 = ($25|0)==(0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    if ($24) {
     break L1;
    } else {
     $8 = $68;
    }
   }
   $32 = ___cxa_find_matching_catch_3(0|0)|0;
   $33 = tempRet0;
   $34 = ((($this)) + 44|0); //@line 102 "modules/thread/Channel.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34); //@line 102 "modules/thread/Channel.cpp"
   __ZNSt3__15queueIPN4love7VariantENS_5dequeIS3_NS_9allocatorIS3_EEEEED2Ev($0); //@line 102 "modules/thread/Channel.cpp"
   ___clang_call_terminate($32); //@line 102 "modules/thread/Channel.cpp"
   // unreachable; //@line 102 "modules/thread/Channel.cpp"
  }
 } while(0);
 $35 = ((($this)) + 8|0); //@line 97 "modules/thread/Channel.cpp"
 $36 = HEAP32[$35>>2]|0; //@line 97 "modules/thread/Channel.cpp"
 $37 = ($36|0)==(0|0); //@line 97 "modules/thread/Channel.cpp"
 if (!($37)) {
  $38 = HEAP32[$36>>2]|0; //@line 97 "modules/thread/Channel.cpp"
  $39 = ((($38)) + 4|0); //@line 97 "modules/thread/Channel.cpp"
  $40 = HEAP32[$39>>2]|0; //@line 97 "modules/thread/Channel.cpp"
  FUNCTION_TABLE_vi[$40 & 8191]($36); //@line 97 "modules/thread/Channel.cpp"
 }
 $41 = ((($this)) + 12|0); //@line 98 "modules/thread/Channel.cpp"
 $42 = HEAP32[$41>>2]|0; //@line 98 "modules/thread/Channel.cpp"
 $43 = ($42|0)==(0|0); //@line 98 "modules/thread/Channel.cpp"
 if (!($43)) {
  $44 = HEAP32[$42>>2]|0; //@line 98 "modules/thread/Channel.cpp"
  $45 = ((($44)) + 4|0); //@line 98 "modules/thread/Channel.cpp"
  $46 = HEAP32[$45>>2]|0; //@line 98 "modules/thread/Channel.cpp"
  FUNCTION_TABLE_vi[$46 & 8191]($42); //@line 98 "modules/thread/Channel.cpp"
 }
 $47 = ((($this)) + 40|0); //@line 100 "modules/thread/Channel.cpp"
 $48 = HEAP8[$47>>0]|0; //@line 100 "modules/thread/Channel.cpp"
 $49 = ($48<<24>>24)==(0); //@line 100 "modules/thread/Channel.cpp"
 $$pre5 = ((($this)) + 44|0); //@line 102 "modules/thread/Channel.cpp"
 if ($49) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre5); //@line 102 "modules/thread/Channel.cpp"
  __ZNSt3__15queueIPN4love7VariantENS_5dequeIS3_NS_9allocatorIS3_EEEEED2Ev($0); //@line 102 "modules/thread/Channel.cpp"
  return; //@line 102 "modules/thread/Channel.cpp"
 }
 $50 = (__ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6thread7ChannelEEENS_19__map_value_compareIS7_SC_NS_4lessIS7_EELb1EEENS5_ISC_EEE4findIS7_EENS_15__tree_iteratorISC_PNS_11__tree_nodeISC_PvEEiEERKT_(658428,$$pre5)|0); //@line 2004 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $51 = ($50|0)==((658432)|0); //@line 672 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($51) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre5); //@line 102 "modules/thread/Channel.cpp"
  __ZNSt3__15queueIPN4love7VariantENS_5dequeIS3_NS_9allocatorIS3_EEEEED2Ev($0); //@line 102 "modules/thread/Channel.cpp"
  return; //@line 102 "modules/thread/Channel.cpp"
 }
 $52 = ((($50)) + 4|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $53 = HEAP32[$52>>2]|0; //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $54 = ($53|0)==(0|0); //@line 157 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($54) {
  $$0$i$i$i$i$i = $50;
  while(1) {
   $57 = ((($$0$i$i$i$i$i)) + 8|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $58 = HEAP32[$57>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $59 = HEAP32[$58>>2]|0; //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $60 = ($59|0)==($$0$i$i$i$i$i|0); //@line 68 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($60) {
    $$01$i$i$i$i$i = $58;
    break;
   } else {
    $$0$i$i$i$i$i = $58;
   }
  }
 } else {
  $$0$i$i$i$i$i$i = $53;
  while(1) {
   $55 = HEAP32[$$0$i$i$i$i$i$i>>2]|0; //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $56 = ($55|0)==(0|0); //@line 134 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($56) {
    $$01$i$i$i$i$i = $$0$i$i$i$i$i$i;
    break;
   } else {
    $$0$i$i$i$i$i$i = $55;
   }
  }
 }
 $61 = HEAP32[164607]|0; //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $62 = ($61|0)==($50|0); //@line 1979 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($62) {
  $63 = $$01$i$i$i$i$i; //@line 656 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[164607] = $63; //@line 1980 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
 $64 = HEAP32[(658436)>>2]|0; //@line 1981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $65 = (($64) + -1)|0; //@line 1981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[(658436)>>2] = $65; //@line 1981 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $66 = HEAP32[(658432)>>2]|0; //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__113__tree_removeIPNS_16__tree_node_baseIPvEEEEvT_S5_($66,$50); //@line 1983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $67 = ((($50)) + 16|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($67); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
 __ZdlPv($50); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($$pre5); //@line 102 "modules/thread/Channel.cpp"
 __ZNSt3__15queueIPN4love7VariantENS_5dequeIS3_NS_9allocatorIS3_EEEEED2Ev($0); //@line 102 "modules/thread/Channel.cpp"
 return; //@line 102 "modules/thread/Channel.cpp"
}
function __ZNSt3__15queueIPN4love7VariantENS_5dequeIS3_NS_9allocatorIS3_EEEEED2Ev($this) {
 $this = $this|0;
 var $$cast$i = 0, $$in$i = 0, $$lcssa$i = 0, $$pre$i = 0, $$pre2$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $__i$01$i = 0, $__i$sroa$0$0$ph$i = 0, $__i$sroa$5$0$i = 0, $__i$sroa$5$0$ph$i = 0, $__new_last2$i$i$i$i$i = 0, $phitmp$i = 0, $scevgep$i$i$i$i$i = 0, $scevgep4$i$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $1 = HEAP32[$0>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $2 = ((($this)) + 16|0); //@line 1060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $3 = HEAP32[$2>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $4 = $3 >>> 10; //@line 1060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $5 = (($1) + ($4<<2)|0); //@line 1060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $6 = ((($this)) + 8|0); //@line 85 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $7 = HEAP32[$6>>2]|0; //@line 85 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $8 = ($7|0)==($1|0); //@line 85 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $9 = $7; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $10 = $1; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($8) {
  $11 = ((($this)) + 20|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $25 = 0;$26 = $11;$57 = 0;
 } else {
  $12 = HEAP32[$5>>2]|0; //@line 1061 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $13 = $3 & 1023; //@line 1061 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $14 = (($12) + ($13<<2)|0); //@line 1061 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $phitmp$i = $14; //@line 1061 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $15 = ((($this)) + 20|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $16 = HEAP32[$15>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $17 = (($16) + ($3))|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $18 = $17 >>> 10; //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $19 = (($1) + ($18<<2)|0); //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $20 = HEAP32[$19>>2]|0; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $21 = $17 & 1023; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $22 = (($20) + ($21<<2)|0); //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $25 = $22;$26 = $15;$57 = $phitmp$i;
 }
 $__i$sroa$0$0$ph$i = $5;$__i$sroa$5$0$ph$i = $57;
 L5: while(1) {
  $__i$sroa$5$0$i = $__i$sroa$5$0$ph$i;
  while(1) {
   $23 = $__i$sroa$5$0$i; //@line 390 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $24 = ($23|0)==($25|0); //@line 394 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   if ($24) {
    break L5;
   }
   $30 = ((($23)) + 4|0); //@line 299 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $31 = $30; //@line 299 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $32 = HEAP32[$__i$sroa$0$0$ph$i>>2]|0; //@line 299 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $33 = (($31) - ($32))|0; //@line 299 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $34 = ($33|0)==(4096); //@line 299 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   if ($34) {
    break;
   } else {
    $__i$sroa$5$0$i = $31;
   }
  }
  $35 = ((($__i$sroa$0$0$ph$i)) + 4|0); //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $36 = HEAP32[$35>>2]|0; //@line 302 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $__i$sroa$0$0$ph$i = $35;$__i$sroa$5$0$ph$i = $36;
 }
 HEAP32[$26>>2] = 0; //@line 1164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $27 = (($9) - ($10))|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27 >> 2; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $29 = ($28>>>0)>(2); //@line 1165 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($29) {
  $$in$i = $1;
  while(1) {
   $37 = HEAP32[$$in$i>>2]|0; //@line 1167 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   __ZdlPv($37); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $38 = HEAP32[$0>>2]|0; //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $39 = ((($38)) + 4|0); //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   HEAP32[$0>>2] = $39; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $40 = HEAP32[$6>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $$cast$i = $39; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $41 = (($40) - ($$cast$i))|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $42 = $41 >> 2; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $43 = ($42>>>0)>(2); //@line 1165 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   if ($43) {
    $$in$i = $39;
   } else {
    $$lcssa$i = $42;
    break;
   }
  }
 } else {
  $$lcssa$i = $28;
 }
 switch ($$lcssa$i|0) {
 case 1:  {
  HEAP32[$2>>2] = 512; //@line 1173 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  break;
 }
 case 2:  {
  HEAP32[$2>>2] = 1024; //@line 1176 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  break;
 }
 default: {
 }
 }
 $44 = HEAP32[$0>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $45 = HEAP32[$6>>2]|0; //@line 78 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $46 = ($44|0)==($45|0); //@line 1107 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if (!($46)) {
  $__i$01$i = $44;
  while(1) {
   $47 = HEAP32[$__i$01$i>>2]|0; //@line 1108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   __ZdlPv($47); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $48 = ((($__i$01$i)) + 4|0); //@line 1107 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $49 = ($48|0)==($45|0); //@line 1107 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   if ($49) {
    break;
   } else {
    $__i$01$i = $48;
   }
  }
  $$pre$i = HEAP32[$0>>2]|0; //@line 83 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $$pre2$i = HEAP32[$6>>2]|0; //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $50 = ($$pre2$i|0)==($$pre$i|0); //@line 309 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if (!($50)) {
   $__new_last2$i$i$i$i$i = $$pre$i; //@line 136 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $scevgep$i$i$i$i$i = ((($$pre2$i)) + -4|0);
   $51 = $scevgep$i$i$i$i$i;
   $52 = (($51) - ($__new_last2$i$i$i$i$i))|0;
   $53 = $52 >>> 2;
   $54 = $53 ^ -1;
   $scevgep4$i$i$i$i$i = (($$pre2$i) + ($54<<2)|0);
   HEAP32[$6>>2] = $scevgep4$i$i$i$i$i; //@line 310 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  }
 }
 $55 = HEAP32[$this>>2]|0; //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $56 = ($55|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($56) {
  return; //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/queue"
 }
 __ZdlPv($55); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 193 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/queue"
}
function __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6thread7ChannelEEENS_19__map_value_compareIS7_SC_NS_4lessIS7_EELb1EEENS5_ISC_EEE4findIS7_EENS_15__tree_iteratorISC_PNS_11__tree_nodeISC_PvEEiEERKT_($this,$__v) {
 $this = $this|0;
 $__v = $__v|0;
 var $$$$i$i$i$i = 0, $$$$i$i$i$i$i = 0, $$0$ph$lcssa$i = 0, $$0$ph6$i = 0, $$01$ph5$i = 0, $$013$i = 0, $$013$i$lcssa = 0, $$sroa$0$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $2 = ($1|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if (!($2)) {
  $3 = HEAP8[$__v>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $4 = $3 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $5 = ($4<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $6 = ((($__v)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = HEAP32[$6>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $8 = $3&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = $8 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $10 = $5 ? $9 : $7; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $11 = ((($__v)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $12 = HEAP32[$11>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $13 = ((($__v)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $14 = $5 ? $13 : $12; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $$0$ph6$i = $0;$$01$ph5$i = $1;
  L3: while(1) {
   $$013$i = $$01$ph5$i;
   while(1) {
    $15 = ((($$013$i)) + 16|0); //@line 2102 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $16 = HEAP8[$15>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $17 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $18 = ($17<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $19 = ((($$013$i)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $20 = HEAP32[$19>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $21 = $16&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $22 = $21 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $23 = $18 ? $22 : $20; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $24 = ((($$013$i)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $25 = HEAP32[$24>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $26 = ((($15)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $27 = $18 ? $26 : $25; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $28 = ($10>>>0)<($23>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $29 = $28 ? $10 : $23; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $30 = (_memcmp($27,$14,$29)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $31 = ($30|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $32 = ($23>>>0)<($10>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $$$$i$i$i$i$i = $32 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $33 = $31 ? $$$$i$i$i$i$i : $30; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $34 = ($33|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if (!($34)) {
     $$013$i$lcssa = $$013$i;
     break;
    }
    $37 = ((($$013$i)) + 4|0); //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $38 = HEAP32[$37>>2]|0; //@line 2108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $39 = ($38|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($39) {
     $$0$ph$lcssa$i = $$0$ph6$i;
     break L3;
    } else {
     $$013$i = $38;
    }
   }
   $35 = HEAP32[$$013$i$lcssa>>2]|0; //@line 2105 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $36 = ($35|0)==(0|0); //@line 2100 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   if ($36) {
    $$0$ph$lcssa$i = $$013$i$lcssa;
    break;
   } else {
    $$0$ph6$i = $$013$i$lcssa;$$01$ph5$i = $35;
   }
  }
  $40 = ($$0$ph$lcssa$i|0)==($0|0); //@line 675 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  if (!($40)) {
   $41 = ((($$0$ph$lcssa$i)) + 16|0); //@line 650 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   $42 = HEAP8[$41>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $43 = $42 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $44 = ($43<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $45 = ((($$0$ph$lcssa$i)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $46 = HEAP32[$45>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $47 = $42&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $48 = $47 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $49 = $44 ? $48 : $46; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $50 = ((($$0$ph$lcssa$i)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $51 = HEAP32[$50>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $52 = ((($41)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $53 = $44 ? $52 : $51; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $54 = ($49>>>0)<($10>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $55 = $54 ? $49 : $10; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $56 = (_memcmp($14,$53,$55)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $57 = ($56|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $58 = ($10>>>0)<($49>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$$$i$i$i$i = $58 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $59 = $57 ? $$$$i$i$i$i : $56; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $60 = ($59|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if (!($60)) {
    $$sroa$0$0 = $$0$ph$lcssa$i;
    return ($$sroa$0$0|0); //@line 2032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
   }
  }
 }
 $$sroa$0$0 = $0;
 return ($$sroa$0$0|0); //@line 2032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
}
function __ZN4love6thread7ChannelD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6thread7ChannelD2Ev($this); //@line 90 "modules/thread/Channel.cpp"
 __ZdlPv($this); //@line 90 "modules/thread/Channel.cpp"
 return; //@line 102 "modules/thread/Channel.cpp"
}
function __ZN4love6thread7Channel6retainEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 __ZN4love6thread9EmptyLockC2Ev($l); //@line 218 "modules/thread/Channel.cpp"
 $0 = ((($this)) + 40|0); //@line 219 "modules/thread/Channel.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 219 "modules/thread/Channel.cpp"
 $2 = ($1<<24>>24)==(0); //@line 219 "modules/thread/Channel.cpp"
 if (!($2)) {
  $3 = HEAP32[164610]|0; //@line 220 "modules/thread/Channel.cpp"
  __THREW__ = 0;
  invoke_vii(4229,($l|0),($3|0)); //@line 220 "modules/thread/Channel.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $6 = ___cxa_find_matching_catch_2()|0;
   $7 = tempRet0;
   __ZN4love6thread9EmptyLockD2Ev($l); //@line 223 "modules/thread/Channel.cpp"
   ___resumeException($6|0);
   // unreachable;
  }
 }
 __THREW__ = 0;
 invoke_vi(3,($this|0)); //@line 222 "modules/thread/Channel.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  __ZN4love6thread9EmptyLockD2Ev($l); //@line 223 "modules/thread/Channel.cpp"
  ___resumeException($6|0);
  // unreachable;
 } else {
  __ZN4love6thread9EmptyLockD2Ev($l); //@line 223 "modules/thread/Channel.cpp"
  STACKTOP = sp;return; //@line 223 "modules/thread/Channel.cpp"
 }
}
function __ZN4love6thread7Channel7releaseEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 __ZN4love6thread9EmptyLockC2Ev($l); //@line 227 "modules/thread/Channel.cpp"
 $0 = ((($this)) + 40|0); //@line 228 "modules/thread/Channel.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 228 "modules/thread/Channel.cpp"
 $2 = ($1<<24>>24)==(0); //@line 228 "modules/thread/Channel.cpp"
 if (!($2)) {
  $3 = HEAP32[164610]|0; //@line 229 "modules/thread/Channel.cpp"
  __THREW__ = 0;
  invoke_vii(4229,($l|0),($3|0)); //@line 229 "modules/thread/Channel.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $6 = ___cxa_find_matching_catch_2()|0;
   $7 = tempRet0;
   __ZN4love6thread9EmptyLockD2Ev($l); //@line 232 "modules/thread/Channel.cpp"
   ___resumeException($6|0);
   // unreachable;
  }
 }
 __THREW__ = 0;
 invoke_vi(4,($this|0)); //@line 231 "modules/thread/Channel.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  __ZN4love6thread9EmptyLockD2Ev($l); //@line 232 "modules/thread/Channel.cpp"
  ___resumeException($6|0);
  // unreachable;
 } else {
  __ZN4love6thread9EmptyLockD2Ev($l); //@line 232 "modules/thread/Channel.cpp"
  STACKTOP = sp;return; //@line 232 "modules/thread/Channel.cpp"
 }
}
function __GLOBAL__sub_I_Channel_cpp() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(658432)>>2] = 0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[(658436)>>2] = 0; //@line 2144 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[164607] = (658432); //@line 1130 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 (___cxa_atexit((4230|0),(658428|0),(___dso_handle|0))|0); //@line 53 "modules/thread/Channel.cpp"
 return;
}
function __ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6thread7ChannelENS_4lessIS6_EENS4_INS_4pairIKS6_SA_EEEEED2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6thread7ChannelEEENS_19__map_value_compareIS7_SC_NS_4lessIS7_EELb1EEENS5_ISC_EEE7destroyEPNS_11__tree_nodeISC_PvEE($this,$1); //@line 1433 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 return; //@line 783 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
}
function __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6thread7ChannelEEENS_19__map_value_compareIS7_SC_NS_4lessIS7_EELb1EEENS5_ISC_EEE7destroyEPNS_11__tree_nodeISC_PvEE($this,$__nd) {
 $this = $this|0;
 $__nd = $__nd|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__nd|0)==(0|0); //@line 1440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($0) {
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 } else {
  $1 = HEAP32[$__nd>>2]|0; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6thread7ChannelEEENS_19__map_value_compareIS7_SC_NS_4lessIS7_EELb1EEENS5_ISC_EEE7destroyEPNS_11__tree_nodeISC_PvEE($this,$1); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $2 = ((($__nd)) + 4|0); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $3 = HEAP32[$2>>2]|0; //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  __ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6thread7ChannelEEENS_19__map_value_compareIS7_SC_NS_4lessIS7_EELb1EEENS5_ISC_EEE7destroyEPNS_11__tree_nodeISC_PvEE($this,$3); //@line 1443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $4 = ((($__nd)) + 16|0); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4); //@line 248 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/utility"
  __ZdlPv($__nd); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  return; //@line 1448 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 }
}
function __ZN4love6thread7Channel10getChannelERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE($name) {
 $name = $name|0;
 var $$$$i$i$i$i$i$i = 0, $$$$i$i$i$i1$i$i = 0, $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $__rt$0$i$i = 0, $__rt$0$in$be$i$i = 0, $__rt$02$i$i = 0, $__rt$03$i$i = 0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = HEAP32[164610]|0; //@line 58 "modules/thread/Channel.cpp"
 $1 = ($0|0)==(0|0); //@line 58 "modules/thread/Channel.cpp"
 if ($1) {
  $2 = (__ZN4love6thread8newMutexEv()|0); //@line 59 "modules/thread/Channel.cpp"
  HEAP32[164610] = $2; //@line 59 "modules/thread/Channel.cpp"
  $3 = $2;
 } else {
  $3 = $0;
 }
 __ZN4love6thread4LockC2EPNS0_5MutexE($l,$3); //@line 61 "modules/thread/Channel.cpp"
 $__rt$02$i$i = HEAP32[(658432)>>2]|0; //@line 878 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $4 = ($__rt$02$i$i|0)==(0|0); //@line 2052 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 L4: do {
  if ($4) {
   label = 10;
  } else {
   $5 = HEAP8[$name>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $6 = $5 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $7 = ($6<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $8 = ((($name)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $9 = HEAP32[$8>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $10 = $5&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $11 = $10 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $12 = $7 ? $11 : $9; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $13 = ((($name)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $14 = HEAP32[$13>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $15 = ((($name)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $16 = $7 ? $15 : $14; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $__rt$03$i$i = $__rt$02$i$i;
   while(1) {
    $17 = ((($__rt$03$i$i)) + 16|0); //@line 2054 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $18 = HEAP8[$17>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $19 = $18 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $20 = ($19<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $21 = ((($__rt$03$i$i)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $22 = HEAP32[$21>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $23 = $18&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $24 = $23 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $25 = $20 ? $24 : $22; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $26 = ((($__rt$03$i$i)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $27 = HEAP32[$26>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $28 = ((($17)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $29 = $20 ? $28 : $27; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $30 = ($25>>>0)<($12>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $31 = $30 ? $25 : $12; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $32 = (_memcmp($16,$29,$31)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $33 = ($32|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $34 = ($12>>>0)<($25>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $$$$i$i$i$i1$i$i = $34 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $35 = $33 ? $$$$i$i$i$i1$i$i : $32; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $36 = ($35|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if ($36) {
     $__rt$0$in$be$i$i = $__rt$03$i$i;
    } else {
     $37 = $34 ? $12 : $25; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $38 = (_memcmp($29,$16,$37)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $39 = ($38|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $$$$i$i$i$i$i$i = $30 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $40 = $39 ? $$$$i$i$i$i$i$i : $38; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $41 = ($40|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if (!($41)) {
      break;
     }
     $42 = ((($__rt$03$i$i)) + 4|0); //@line 2060 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
     $__rt$0$in$be$i$i = $42;
    }
    $__rt$0$i$i = HEAP32[$__rt$0$in$be$i$i>>2]|0; //@line 878 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    $43 = ($__rt$0$i$i|0)==(0|0); //@line 2052 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
    if ($43) {
     label = 10;
     break L4;
    } else {
     $__rt$03$i$i = $__rt$0$i$i;
    }
   }
   __THREW__ = 0;
   $56 = (invoke_iii(4232,(658428|0),($name|0))|0); //@line 65 "modules/thread/Channel.cpp"
   $57 = __THREW__; __THREW__ = 0;
   $58 = $57&1;
   if (!($58)) {
    $59 = HEAP32[$56>>2]|0; //@line 65 "modules/thread/Channel.cpp"
    $60 = HEAP32[$59>>2]|0; //@line 65 "modules/thread/Channel.cpp"
    $61 = ((($60)) + 8|0); //@line 65 "modules/thread/Channel.cpp"
    $62 = HEAP32[$61>>2]|0; //@line 65 "modules/thread/Channel.cpp"
    __THREW__ = 0;
    invoke_vi($62|0,($59|0)); //@line 65 "modules/thread/Channel.cpp"
    $63 = __THREW__; __THREW__ = 0;
    $64 = $63&1;
    if (!($64)) {
     label = 18;
    }
   }
  }
 } while(0);
 do {
  if ((label|0) == 10) {
   __THREW__ = 0;
   $44 = (invoke_ii(3348,64)|0); //@line 63 "modules/thread/Channel.cpp"
   $45 = __THREW__; __THREW__ = 0;
   $46 = $45&1;
   if (!($46)) {
    __THREW__ = 0;
    invoke_vii(4231,($44|0),($name|0)); //@line 63 "modules/thread/Channel.cpp"
    $47 = __THREW__; __THREW__ = 0;
    $48 = $47&1;
    if ($48) {
     $54 = ___cxa_find_matching_catch_2()|0;
     $55 = tempRet0;
     __ZdlPv($44); //@line 63 "modules/thread/Channel.cpp"
     $$0 = $55;$$01 = $54;
     __ZN4love6thread4LockD2Ev($l); //@line 68 "modules/thread/Channel.cpp"
     ___resumeException($$01|0);
     // unreachable;
    } else {
     __THREW__ = 0;
     $49 = (invoke_iii(4232,(658428|0),($name|0))|0); //@line 63 "modules/thread/Channel.cpp"
     $50 = __THREW__; __THREW__ = 0;
     $51 = $50&1;
     if ($51) {
      break;
     }
     HEAP32[$49>>2] = $44; //@line 63 "modules/thread/Channel.cpp"
     label = 18;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 18) {
  __THREW__ = 0;
  $65 = (invoke_iii(4232,(658428|0),($name|0))|0); //@line 67 "modules/thread/Channel.cpp"
  $66 = __THREW__; __THREW__ = 0;
  $67 = $66&1;
  if (!($67)) {
   $68 = HEAP32[$65>>2]|0; //@line 67 "modules/thread/Channel.cpp"
   __ZN4love6thread4LockD2Ev($l); //@line 68 "modules/thread/Channel.cpp"
   STACKTOP = sp;return ($68|0); //@line 68 "modules/thread/Channel.cpp"
  }
 }
 $52 = ___cxa_find_matching_catch_2()|0;
 $53 = tempRet0;
 $$0 = $53;$$01 = $52;
 __ZN4love6thread4LockD2Ev($l); //@line 68 "modules/thread/Channel.cpp"
 ___resumeException($$01|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love6thread7ChannelC2ERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE($this,$name) {
 $this = $this|0;
 $name = $name|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 79 "modules/thread/Channel.cpp"
 HEAP32[$this>>2] = (33052); //@line 84 "modules/thread/Channel.cpp"
 $0 = ((($this)) + 16|0); //@line 79 "modules/thread/Channel.cpp"
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;HEAP32[$0+12>>2]=0|0;HEAP32[$0+16>>2]=0|0;HEAP32[$0+20>>2]=0|0; //@line 1094 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $1 = ((($this)) + 40|0); //@line 80 "modules/thread/Channel.cpp"
 HEAP8[$1>>0] = 1; //@line 80 "modules/thread/Channel.cpp"
 $2 = ((($this)) + 44|0); //@line 81 "modules/thread/Channel.cpp"
 __THREW__ = 0;
 invoke_vii(3351,($2|0),($name|0)); //@line 81 "modules/thread/Channel.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  $$0 = $16;$$01 = $15;
  __ZNSt3__15queueIPN4love7VariantENS_5dequeIS3_NS_9allocatorIS3_EEEEED2Ev($0); //@line 87 "modules/thread/Channel.cpp"
  ___resumeException($$01|0);
  // unreachable;
 }
 $5 = ((($this)) + 56|0); //@line 82 "modules/thread/Channel.cpp"
 HEAP32[$5>>2] = 0; //@line 82 "modules/thread/Channel.cpp"
 $6 = ((($this)) + 60|0); //@line 83 "modules/thread/Channel.cpp"
 HEAP32[$6>>2] = 0; //@line 83 "modules/thread/Channel.cpp"
 __THREW__ = 0;
 $7 = (invoke_i(4233)|0); //@line 85 "modules/thread/Channel.cpp"
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if (!($9)) {
  $10 = ((($this)) + 8|0); //@line 85 "modules/thread/Channel.cpp"
  HEAP32[$10>>2] = $7; //@line 85 "modules/thread/Channel.cpp"
  __THREW__ = 0;
  $11 = (invoke_i(4234)|0); //@line 86 "modules/thread/Channel.cpp"
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   $14 = ((($this)) + 12|0); //@line 86 "modules/thread/Channel.cpp"
   HEAP32[$14>>2] = $11; //@line 86 "modules/thread/Channel.cpp"
   return; //@line 87 "modules/thread/Channel.cpp"
  }
 }
 $17 = ___cxa_find_matching_catch_2()|0;
 $18 = tempRet0;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2); //@line 87 "modules/thread/Channel.cpp"
 $$0 = $18;$$01 = $17;
 __ZNSt3__15queueIPN4love7VariantENS_5dequeIS3_NS_9allocatorIS3_EEEEED2Ev($0); //@line 87 "modules/thread/Channel.cpp"
 ___resumeException($$01|0);
 // unreachable;
}
function __ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6thread7ChannelENS_4lessIS6_EENS4_INS_4pairIKS6_SA_EEEEEixERSE_($this,$__k) {
 $this = $this|0;
 $__k = $__k|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__parent = 0, $__r$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__parent = sp;
 $0 = (__ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6thread7ChannelENS_4lessIS6_EENS4_INS_4pairIKS6_SA_EEEEE16__find_equal_keyERPNS_16__tree_node_baseIPvEERSE_($this,$__parent,$__k)|0); //@line 1376 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $1 = HEAP32[$0>>2]|0; //@line 1377 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 $2 = ($1|0)==(0|0); //@line 1378 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 if (!($2)) {
  $__r$0 = $1;
  $23 = ((($__r$0)) + 28|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  STACKTOP = sp;return ($23|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 $3 = (__Znwj(32)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $4 = ((($3)) + 16|0); //@line 1364 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 __THREW__ = 0;
 invoke_vii(3351,($4|0),($__k|0)); //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZdlPv($3); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($7|0);
  // unreachable;
 }
 $9 = ((($3)) + 28|0); //@line 1366 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$9>>2] = 0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $10 = HEAP32[$__parent>>2]|0; //@line 1381 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 HEAP32[$3>>2] = 0; //@line 1711 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $11 = ((($3)) + 4|0); //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$11>>2] = 0; //@line 1712 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $12 = ((($3)) + 8|0); //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$12>>2] = $10; //@line 1713 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$0>>2] = $3; //@line 1714 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $13 = HEAP32[$this>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $14 = HEAP32[$13>>2]|0; //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $15 = ($14|0)==(0|0); //@line 1715 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 if ($15) {
  $19 = $3;
 } else {
  $16 = $14; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  HEAP32[$this>>2] = $16; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $$pre$i = HEAP32[$0>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
  $19 = $$pre$i;
 }
 $17 = ((($this)) + 4|0); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $18 = HEAP32[$17>>2]|0; //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 __ZNSt3__127__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($18,$19); //@line 1717 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $20 = ((($this)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $21 = HEAP32[$20>>2]|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $22 = (($21) + 1)|0; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 HEAP32[$20>>2] = $22; //@line 1718 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $__r$0 = $3;
 $23 = ((($__r$0)) + 28|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 STACKTOP = sp;return ($23|0); //@line 1384 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
}
function __ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6thread7ChannelENS_4lessIS6_EENS4_INS_4pairIKS6_SA_EEEEE16__find_equal_keyERPNS_16__tree_node_baseIPvEERSE_($this,$__parent,$__k) {
 $this = $this|0;
 $__parent = $__parent|0;
 $__k = $__k|0;
 var $$$$i$i$i = 0, $$$$i$i$i1 = 0, $$0 = 0, $$lcssa = 0, $$lcssa19 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__nd$0 = 0, $__nd$0$lcssa = 0, $__nd$0$lcssa17 = 0, $__nd$0$lcssa18 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $1 = HEAP32[$0>>2]|0; //@line 875 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__tree"
 $2 = ($1|0)==(0|0); //@line 1203 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 if ($2) {
  HEAP32[$__parent>>2] = $0; //@line 1234 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $0;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 $3 = HEAP8[$__k>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = $3 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ($4<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $6 = ((($__k)) + 4|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = HEAP32[$6>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = $3&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $9 = $8 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $10 = $5 ? $9 : $7; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $11 = ((($__k)) + 8|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $12 = HEAP32[$11>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $13 = ((($__k)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $14 = $5 ? $13 : $12; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $__nd$0 = $1;
 while(1) {
  $15 = ((($__nd$0)) + 16|0); //@line 1207 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $16 = HEAP8[$15>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $17 = $16 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $18 = ($17<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $19 = ((($__nd$0)) + 20|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $20 = HEAP32[$19>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $21 = $16&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $22 = $21 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $23 = $18 ? $22 : $20; //@line 1431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $24 = ((($__nd$0)) + 24|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $25 = HEAP32[$24>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $26 = ((($15)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $27 = $18 ? $26 : $25; //@line 1758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $28 = ($23>>>0)<($10>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $29 = $28 ? $23 : $10; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $30 = (_memcmp($14,$27,$29)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $31 = ($30|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $32 = ($10>>>0)<($23>>>0); //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $$$$i$i$i1 = $32 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $33 = $31 ? $$$$i$i$i1 : $30; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $34 = ($33|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($34) {
   $35 = HEAP32[$__nd$0>>2]|0; //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $36 = ($35|0)==(0|0); //@line 1209 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   if ($36) {
    $$lcssa19 = $__nd$0;$__nd$0$lcssa18 = $__nd$0;
    label = 5;
    break;
   } else {
    $$sink = $35;
   }
  } else {
   $37 = $32 ? $10 : $23; //@line 3645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $38 = (_memcmp($27,$14,$37)|0); //@line 639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $39 = ($38|0)==(0); //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $$$$i$i$i = $28 ? -2147483648 : 0; //@line 3648 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $40 = $39 ? $$$$i$i$i : $38; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $41 = ($40|0)<(0); //@line 3827 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if (!($41)) {
    $__nd$0$lcssa = $__nd$0;
    label = 10;
    break;
   }
   $42 = ((($__nd$0)) + 4|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $43 = HEAP32[$42>>2]|0; //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   $44 = ($43|0)==(0|0); //@line 1219 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
   if ($44) {
    $$lcssa = $42;$__nd$0$lcssa17 = $__nd$0;
    label = 9;
    break;
   } else {
    $$sink = $43;
   }
  }
  $__nd$0 = $$sink;
 }
 if ((label|0) == 5) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa18; //@line 1213 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $$lcssa19;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 else if ((label|0) == 9) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa17; //@line 1223 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $$lcssa;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 else if ((label|0) == 10) {
  HEAP32[$__parent>>2] = $__nd$0$lcssa; //@line 1229 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
  $$0 = $__parent;
  return ($$0|0); //@line 1236 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/map"
 }
 return (0)|0;
}
function __ZN4love6thread7ChannelC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 70 "modules/thread/Channel.cpp"
 HEAP32[$this>>2] = (33052); //@line 74 "modules/thread/Channel.cpp"
 $0 = ((($this)) + 16|0); //@line 70 "modules/thread/Channel.cpp"
 $1 = ((($this)) + 44|0); //@line 70 "modules/thread/Channel.cpp"
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;HEAP32[$0+12>>2]=0|0;HEAP32[$0+16>>2]=0|0;HEAP32[$0+20>>2]=0|0;HEAP8[$0+24>>0]=0|0; //@line 71 "modules/thread/Channel.cpp"
 ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0;HEAP32[$1+12>>2]=0|0;HEAP32[$1+16>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 __THREW__ = 0;
 $2 = (invoke_i(4233)|0); //@line 75 "modules/thread/Channel.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  $5 = ((($this)) + 8|0); //@line 75 "modules/thread/Channel.cpp"
  HEAP32[$5>>2] = $2; //@line 75 "modules/thread/Channel.cpp"
  __THREW__ = 0;
  $6 = (invoke_i(4234)|0); //@line 76 "modules/thread/Channel.cpp"
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if (!($8)) {
   $9 = ((($this)) + 12|0); //@line 76 "modules/thread/Channel.cpp"
   HEAP32[$9>>2] = $6; //@line 76 "modules/thread/Channel.cpp"
   return; //@line 77 "modules/thread/Channel.cpp"
  }
 }
 $10 = ___cxa_find_matching_catch_2()|0;
 $11 = tempRet0;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 77 "modules/thread/Channel.cpp"
 __ZNSt3__15queueIPN4love7VariantENS_5dequeIS3_NS_9allocatorIS3_EEEEED2Ev($0); //@line 77 "modules/thread/Channel.cpp"
 ___resumeException($10|0);
 // unreachable;
}
function __ZN4love6thread7Channel4pushEPNS_7VariantE($this,$var) {
 $this = $this|0;
 $var = $var|0;
 var $$0 = 0, $$pre = 0, $$pre$i$i = 0, $$pre1$i$i = 0, $$pre2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $l = 0, $sum$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = $var;
 $1 = ($var|0)==(0|0); //@line 106 "modules/thread/Channel.cpp"
 if ($1) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 121 "modules/thread/Channel.cpp"
 }
 $2 = ((($this)) + 8|0); //@line 109 "modules/thread/Channel.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 109 "modules/thread/Channel.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($l,$3); //@line 109 "modules/thread/Channel.cpp"
 $4 = HEAP32[$var>>2]|0; //@line 110 "modules/thread/Channel.cpp"
 $5 = ((($4)) + 8|0); //@line 110 "modules/thread/Channel.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 110 "modules/thread/Channel.cpp"
 __THREW__ = 0;
 invoke_vi($6|0,($var|0)); //@line 110 "modules/thread/Channel.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  __ZN4love6thread4LockD2Ev($l); //@line 121 "modules/thread/Channel.cpp"
  ___resumeException($19|0);
  // unreachable;
 }
 $9 = ((($this)) + 40|0); //@line 114 "modules/thread/Channel.cpp"
 $10 = HEAP8[$9>>0]|0; //@line 114 "modules/thread/Channel.cpp"
 $11 = ($10<<24>>24)==(0); //@line 114 "modules/thread/Channel.cpp"
 $$pre = ((($this)) + 36|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 if (!($11)) {
  $12 = HEAP32[$$pre>>2]|0; //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $13 = ($12|0)==(0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  if ($13) {
   $14 = HEAP32[$this>>2]|0; //@line 115 "modules/thread/Channel.cpp"
   $15 = ((($14)) + 8|0); //@line 115 "modules/thread/Channel.cpp"
   $16 = HEAP32[$15>>2]|0; //@line 115 "modules/thread/Channel.cpp"
   __THREW__ = 0;
   invoke_vi($16|0,($this|0)); //@line 115 "modules/thread/Channel.cpp"
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if ($18) {
    $19 = ___cxa_find_matching_catch_2()|0;
    $20 = tempRet0;
    __ZN4love6thread4LockD2Ev($l); //@line 121 "modules/thread/Channel.cpp"
    ___resumeException($19|0);
    // unreachable;
   }
  }
 }
 $21 = ((($this)) + 24|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $22 = HEAP32[$21>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $23 = ((($this)) + 20|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $24 = HEAP32[$23>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $25 = (($22) - ($24))|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $26 = ($22|0)==($24|0); //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $27 = $25 << 8; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $28 = (($27) + -1)|0; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $29 = $26 ? 0 : $28; //@line 1369 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $30 = ((($this)) + 32|0); //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $31 = HEAP32[$30>>2]|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $32 = HEAP32[$$pre>>2]|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $sum$i$i$i = (($32) + ($31))|0; //@line 1379 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $33 = ($29|0)==($sum$i$i$i|0); //@line 1769 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $34 = $24; //@line 1769 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 do {
  if ($33) {
   $35 = ((($this)) + 16|0); //@line 286 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/queue"
   __THREW__ = 0;
   invoke_vi(4235,($35|0)); //@line 1770 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $36 = __THREW__; __THREW__ = 0;
   $37 = $36&1;
   if ($37) {
    $19 = ___cxa_find_matching_catch_2()|0;
    $20 = tempRet0;
    __ZN4love6thread4LockD2Ev($l); //@line 121 "modules/thread/Channel.cpp"
    ___resumeException($19|0);
    // unreachable;
   } else {
    $$pre$i$i = HEAP32[$$pre>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $$pre1$i$i = HEAP32[$30>>2]|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
    $$pre2$i$i = HEAP32[$23>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $39 = $$pre$i$i;$40 = $$pre1$i$i;$43 = $$pre2$i$i;
    break;
   }
  } else {
   $39 = $32;$40 = $31;$43 = $34;
  }
 } while(0);
 $38 = (($39) + ($40))|0; //@line 1076 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $41 = $38 >>> 10; //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $42 = (($43) + ($41<<2)|0); //@line 1077 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $44 = HEAP32[$42>>2]|0; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $45 = $38 & 1023; //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $46 = (($44) + ($45<<2)|0); //@line 1078 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 HEAP32[$46>>2] = $0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $47 = (($39) + 1)|0; //@line 1773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 HEAP32[$$pre>>2] = $47; //@line 1773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $48 = ((($this)) + 12|0); //@line 118 "modules/thread/Channel.cpp"
 $49 = HEAP32[$48>>2]|0; //@line 118 "modules/thread/Channel.cpp"
 $50 = HEAP32[$49>>2]|0; //@line 118 "modules/thread/Channel.cpp"
 $51 = ((($50)) + 12|0); //@line 118 "modules/thread/Channel.cpp"
 $52 = HEAP32[$51>>2]|0; //@line 118 "modules/thread/Channel.cpp"
 __THREW__ = 0;
 invoke_vi($52|0,($49|0)); //@line 118 "modules/thread/Channel.cpp"
 $53 = __THREW__; __THREW__ = 0;
 $54 = $53&1;
 if ($54) {
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  __ZN4love6thread4LockD2Ev($l); //@line 121 "modules/thread/Channel.cpp"
  ___resumeException($19|0);
  // unreachable;
 }
 $55 = ((($this)) + 56|0); //@line 120 "modules/thread/Channel.cpp"
 $56 = HEAP32[$55>>2]|0; //@line 120 "modules/thread/Channel.cpp"
 $57 = (($56) + 1)|0; //@line 120 "modules/thread/Channel.cpp"
 HEAP32[$55>>2] = $57; //@line 120 "modules/thread/Channel.cpp"
 __ZN4love6thread4LockD2Ev($l); //@line 121 "modules/thread/Channel.cpp"
 $$0 = $57;
 STACKTOP = sp;return ($$0|0); //@line 121 "modules/thread/Channel.cpp"
}
function __ZNSt3__15dequeIPN4love7VariantENS_9allocatorIS3_EEE19__add_back_capacityEv($this) {
 $this = $this|0;
 var $$0161 = 0, $$0162 = 0, $$063 = 0, $$064 = 0, $$in$i = 0, $$in$i30 = 0, $$in$i4 = 0, $$lcssa = 0, $$lcssa77 = 0, $$pre = 0, $$pre$i32 = 0, $$pre$i6 = 0, $$pre4$i10 = 0, $$pre4$i37 = 0, $$pre59 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $__buf$sroa$0$0 = 0, $__buf$sroa$0$1$lcssa = 0, $__buf$sroa$0$147 = 0, $__buf$sroa$0$147$lcssa = 0, $__buf$sroa$0$2 = 0, $__buf$sroa$0$360 = 0, $__buf$sroa$13$0 = 0, $__buf$sroa$13$0$in = 0, $__buf$sroa$13$1$lcssa = 0, $__buf$sroa$13$148 = 0, $__buf$sroa$13$2 = 0, $__buf$sroa$28$1$lcssa = 0, $__buf$sroa$28$149 = 0, $__buf$sroa$28$2 = 0, $__buf$sroa$46$0 = 0, $__buf$sroa$46$1$lcssa = 0, $__buf$sroa$46$150 = 0, $__buf$sroa$46$2 = 0, $__i$051 = 0, $__t$sroa$10$0$i = 0;
 var $__t$sroa$10$0$i29 = 0, $__t$sroa$10$0$i3 = 0, $__t$sroa$10$1$i = 0, $__t$sroa$10$1$i34 = 0, $__t$sroa$10$1$i8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 4|0;
 $1 = sp;
 $2 = ((($this)) + 16|0); //@line 1374 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $3 = HEAP32[$2>>2]|0; //@line 1374 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $4 = ($3>>>0)>(1023); //@line 2390 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($4) {
  $5 = (($3) + -1024)|0; //@line 2392 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  HEAP32[$2>>2] = $5; //@line 2392 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $6 = ((($this)) + 4|0); //@line 90 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $7 = HEAP32[$6>>2]|0; //@line 90 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $8 = HEAP32[$7>>2]|0; //@line 2393 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $9 = ((($7)) + 4|0); //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  HEAP32[$6>>2] = $9; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $10 = ((($this)) + 8|0); //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $11 = HEAP32[$10>>2]|0; //@line 562 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $12 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $13 = HEAP32[$12>>2]|0; //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $14 = ($11|0)==($13|0); //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $15 = $13; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $16 = $11; //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  do {
   if ($14) {
    $17 = HEAP32[$this>>2]|0; //@line 550 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $18 = ($9>>>0)>($17>>>0); //@line 550 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $19 = $17; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($18) {
     $20 = $9; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $21 = (($20) - ($19))|0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $22 = $21 >> 2; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $23 = (($22) + 1)|0; //@line 553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $24 = (($23|0) / -2)&-1; //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $25 = (($9) + ($24<<2)|0); //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $26 = (($16) - ($20))|0; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $27 = $26 >> 2; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     _memmove(($25|0),($9|0),($26|0))|0; //@line 1893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $28 = (($25) + ($27<<2)|0); //@line 1894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     HEAP32[$10>>2] = $28; //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $29 = HEAP32[$6>>2]|0; //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $30 = (($29) + ($24<<2)|0); //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     HEAP32[$6>>2] = $30; //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $54 = $28;
     break;
    }
    $31 = (($15) - ($19))|0; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $32 = $31 >> 1; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $33 = ($32|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $34 = $33 ? 1 : $32; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $35 = $34 >>> 2; //@line 560 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $36 = $34 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $37 = (__Znwj($36)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $38 = $37; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $39 = (($37) + ($35<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $40 = $39; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $41 = (($37) + ($34<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $42 = $41; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $43 = ($9|0)==($11|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($43) {
     $52 = $17;$__t$sroa$10$1$i8 = $40;
    } else {
     $$in$i4 = $39;$45 = $9;$__t$sroa$10$0$i3 = $40;
     while(1) {
      $44 = HEAP32[$45>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$$in$i4>>2] = $44; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $46 = $__t$sroa$10$0$i3; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $47 = ((($46)) + 4|0); //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $48 = $47; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $49 = ((($45)) + 4|0); //@line 954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $50 = ($49|0)==($11|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($50) {
       $$lcssa = $48;
       break;
      } else {
       $$in$i4 = $47;$45 = $49;$__t$sroa$10$0$i3 = $48;
      }
     }
     $$pre$i6 = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $52 = $$pre$i6;$__t$sroa$10$1$i8 = $$lcssa;
    }
    HEAP32[$this>>2] = $38; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$6>>2] = $40; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$10>>2] = $__t$sroa$10$1$i8; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$12>>2] = $42; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $51 = ($52|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $53 = $__t$sroa$10$1$i8; //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($51) {
     $54 = $53;
    } else {
     __ZdlPv($52); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $$pre4$i10 = HEAP32[$10>>2]|0; //@line 569 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $54 = $$pre4$i10;
    }
   } else {
    $54 = $11;
   }
  } while(0);
  HEAP32[$54>>2] = $8; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $55 = HEAP32[$10>>2]|0; //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $56 = ((($55)) + 4|0); //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  HEAP32[$10>>2] = $56; //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  STACKTOP = sp;return; //@line 2441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 }
 $57 = ((($this)) + 8|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $58 = HEAP32[$57>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $59 = ((($this)) + 4|0); //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $60 = HEAP32[$59>>2]|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $61 = (($58) - ($60))|0; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $62 = $61 >> 2; //@line 84 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $63 = ((($this)) + 12|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $64 = HEAP32[$63>>2]|0; //@line 86 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $65 = HEAP32[$this>>2]|0; //@line 86 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $66 = (($64) - ($65))|0; //@line 86 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $67 = $66 >> 2; //@line 86 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $68 = ($62>>>0)<($67>>>0); //@line 2398 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $69 = $58; //@line 2398 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $70 = $60; //@line 2398 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($68) {
  $71 = ($64|0)==($58|0); //@line 2402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $72 = (__Znwj(4096)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  if (!($71)) {
   HEAP32[$0>>2] = $72; //@line 2403 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   __ZNSt3__114__split_bufferIPPN4love7VariantENS_9allocatorIS4_EEE9push_backEOS4_($this,$0); //@line 2403 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   STACKTOP = sp;return; //@line 2441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  }
  HEAP32[$1>>2] = $72; //@line 2406 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  __ZNSt3__114__split_bufferIPPN4love7VariantENS_9allocatorIS4_EEE10push_frontEOS4_($this,$1); //@line 2406 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $73 = HEAP32[$59>>2]|0; //@line 90 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $74 = HEAP32[$73>>2]|0; //@line 2408 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $75 = ((($73)) + 4|0); //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  HEAP32[$59>>2] = $75; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $76 = HEAP32[$57>>2]|0; //@line 562 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $77 = HEAP32[$63>>2]|0; //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $78 = ($76|0)==($77|0); //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $79 = $77; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $80 = $76; //@line 548 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  do {
   if ($78) {
    $81 = HEAP32[$this>>2]|0; //@line 550 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $82 = ($75>>>0)>($81>>>0); //@line 550 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $83 = $81; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($82) {
     $84 = $75; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $85 = (($84) - ($83))|0; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $86 = $85 >> 2; //@line 552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $87 = (($86) + 1)|0; //@line 553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $88 = (($87|0) / -2)&-1; //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $89 = (($75) + ($88<<2)|0); //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $90 = (($80) - ($84))|0; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $91 = $90 >> 2; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     _memmove(($89|0),($75|0),($90|0))|0; //@line 1893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $92 = (($89) + ($91<<2)|0); //@line 1894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     HEAP32[$57>>2] = $92; //@line 554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $93 = HEAP32[$59>>2]|0; //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $94 = (($93) + ($88<<2)|0); //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     HEAP32[$59>>2] = $94; //@line 555 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $118 = $92;
     break;
    }
    $95 = (($79) - ($83))|0; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $96 = $95 >> 1; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $97 = ($96|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $98 = $97 ? 1 : $96; //@line 559 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $99 = $98 >>> 2; //@line 560 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $100 = $98 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $101 = (__Znwj($100)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $102 = $101; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $103 = (($101) + ($99<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $104 = $103; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $105 = (($101) + ($98<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $106 = $105; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $107 = ($75|0)==($76|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
    if ($107) {
     $116 = $81;$__t$sroa$10$1$i34 = $104;
    } else {
     $$in$i30 = $103;$109 = $75;$__t$sroa$10$0$i29 = $104;
     while(1) {
      $108 = HEAP32[$109>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$$in$i30>>2] = $108; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $110 = $__t$sroa$10$0$i29; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $111 = ((($110)) + 4|0); //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $112 = $111; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $113 = ((($109)) + 4|0); //@line 954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      $114 = ($113|0)==($76|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
      if ($114) {
       $$lcssa77 = $112;
       break;
      } else {
       $$in$i30 = $111;$109 = $113;$__t$sroa$10$0$i29 = $112;
      }
     }
     $$pre$i32 = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $116 = $$pre$i32;$__t$sroa$10$1$i34 = $$lcssa77;
    }
    HEAP32[$this>>2] = $102; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$59>>2] = $104; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$57>>2] = $__t$sroa$10$1$i34; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$63>>2] = $106; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $115 = ($116|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $117 = $__t$sroa$10$1$i34; //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    if ($115) {
     $118 = $117;
    } else {
     __ZdlPv($116); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $$pre4$i37 = HEAP32[$57>>2]|0; //@line 569 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $118 = $$pre4$i37;
    }
   } else {
    $118 = $76;
   }
  } while(0);
  HEAP32[$118>>2] = $74; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $119 = HEAP32[$57>>2]|0; //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $120 = ((($119)) + 4|0); //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  HEAP32[$57>>2] = $120; //@line 570 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  STACKTOP = sp;return; //@line 2441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 }
 $121 = $66 >> 1; //@line 2417 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $122 = ($121|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $123 = $122 ? 1 : $121; //@line 2417 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $124 = $123 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $125 = (__Znwj($124)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $126 = $125; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $127 = (($125) + ($62<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $128 = (($125) + ($123<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $129 = $128; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 __THREW__ = 0;
 $130 = (invoke_ii(3348,4096)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $131 = __THREW__; __THREW__ = 0;
 $132 = $131&1;
 L37: do {
  if ($132) {
   label = 29;
  } else {
   $133 = $130; //@line 2424 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $134 = ($62|0)==($123|0); //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   do {
    if ($134) {
     $135 = ($61|0)>(0); //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     if ($135) {
      $136 = (($62) + 1)|0; //@line 584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $137 = (($136|0) / -2)&-1; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $138 = (($127) + ($137<<2)|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $163 = $69;$164 = $70;$__buf$sroa$0$0 = $126;$__buf$sroa$13$0$in = $138;$__buf$sroa$46$0 = $129;
      break;
     }
     $139 = $61 >> 1; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $140 = ($139|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $141 = $140 ? 1 : $139; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $142 = $141 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     __THREW__ = 0;
     $143 = (invoke_ii(3348,($142|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $144 = __THREW__; __THREW__ = 0;
     $145 = $144&1;
     if ($145) {
      label = 29;
      break L37;
     }
     $146 = $141 >>> 2; //@line 591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $147 = (($143) + ($146<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $148 = (($143) + ($141<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $149 = $148; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $150 = $143; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     __ZdlPv($125); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
     $$pre = HEAP32[$57>>2]|0; //@line 78 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $$pre59 = HEAP32[$59>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $163 = $$pre;$164 = $$pre59;$__buf$sroa$0$0 = $150;$__buf$sroa$13$0$in = $147;$__buf$sroa$46$0 = $149;
    } else {
     $163 = $69;$164 = $70;$__buf$sroa$0$0 = $126;$__buf$sroa$13$0$in = $127;$__buf$sroa$46$0 = $129;
    }
   } while(0);
   $__buf$sroa$13$0 = $__buf$sroa$13$0$in; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   HEAP32[$__buf$sroa$13$0$in>>2] = $133; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $160 = ((($__buf$sroa$13$0$in)) + 4|0); //@line 602 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $161 = $160; //@line 602 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $162 = ($163|0)==($164|0); //@line 2434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   L46: do {
    if ($162) {
     $__buf$sroa$0$1$lcssa = $__buf$sroa$0$0;$__buf$sroa$13$1$lcssa = $__buf$sroa$13$0;$__buf$sroa$28$1$lcssa = $161;$__buf$sroa$46$1$lcssa = $__buf$sroa$46$0;
    } else {
     $__buf$sroa$0$147 = $__buf$sroa$0$0;$__buf$sroa$13$148 = $__buf$sroa$13$0;$__buf$sroa$28$149 = $161;$__buf$sroa$46$150 = $__buf$sroa$46$0;$__i$051 = $163;
     L47: while(1) {
      $167 = ((($__i$051)) + -4|0); //@line 2435 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
      $168 = $__buf$sroa$13$148; //@line 497 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $169 = $__buf$sroa$0$147; //@line 484 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $170 = ($168|0)==($169|0); //@line 484 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      do {
       if ($170) {
        $171 = $__buf$sroa$28$149; //@line 498 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $172 = $__buf$sroa$46$150; //@line 486 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $173 = ($171>>>0)<($172>>>0); //@line 486 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        if ($173) {
         $174 = (($__buf$sroa$46$150) - ($__buf$sroa$28$149))|0; //@line 488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $175 = $174 >> 2; //@line 488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $176 = (($175) + 1)|0; //@line 489 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $177 = (($176|0) / 2)&-1; //@line 489 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $178 = (($171) + ($177<<2)|0); //@line 490 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $179 = (($__buf$sroa$28$149) - ($__buf$sroa$13$148))|0; //@line 1927 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $180 = $179 >> 2; //@line 1927 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $181 = (0 - ($180))|0; //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $182 = (($178) + ($181<<2)|0); //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $183 = $__buf$sroa$13$148; //@line 1929 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         _memmove(($182|0),($183|0),($179|0))|0; //@line 1929 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $184 = $182; //@line 490 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $185 = $178; //@line 491 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
         $212 = $182;$__buf$sroa$0$2 = $__buf$sroa$0$147;$__buf$sroa$13$2 = $184;$__buf$sroa$28$2 = $185;$__buf$sroa$46$2 = $__buf$sroa$46$150;
         break;
        }
        $186 = (($__buf$sroa$46$150) - ($__buf$sroa$0$147))|0; //@line 495 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $187 = $186 >> 1; //@line 495 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $188 = ($187|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $189 = $188 ? 1 : $187; //@line 495 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $190 = $189 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        __THREW__ = 0;
        $191 = (invoke_ii(3348,($190|0))|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        $192 = __THREW__; __THREW__ = 0;
        $193 = $192&1;
        if ($193) {
         $__buf$sroa$0$147$lcssa = $__buf$sroa$0$147;
         break L47;
        }
        $194 = (($189) + 3)|0; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $195 = $194 >>> 2; //@line 496 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $196 = $191; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $197 = (($191) + ($195<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $198 = $197; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $199 = (($191) + ($189<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $200 = $199; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        $201 = ($168|0)==($171|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
        if ($201) {
         $__t$sroa$10$1$i = $198;
        } else {
         $$in$i = $197;$203 = $168;$__t$sroa$10$0$i = $198;
         while(1) {
          $202 = HEAP32[$203>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          HEAP32[$$in$i>>2] = $202; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $204 = $__t$sroa$10$0$i; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
          $205 = ((($204)) + 4|0); //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
          $206 = $205; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
          $207 = ((($203)) + 4|0); //@line 954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
          $208 = ($207|0)==($171|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
          if ($208) {
           $__t$sroa$10$1$i = $206;
           break;
          } else {
           $$in$i = $205;$203 = $207;$__t$sroa$10$0$i = $206;
          }
         }
        }
        $209 = ($__buf$sroa$13$148|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
        if ($209) {
         $212 = $197;$__buf$sroa$0$2 = $196;$__buf$sroa$13$2 = $198;$__buf$sroa$28$2 = $__t$sroa$10$1$i;$__buf$sroa$46$2 = $200;
        } else {
         $210 = $__buf$sroa$13$148; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
         __ZdlPv($210); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
         $212 = $197;$__buf$sroa$0$2 = $196;$__buf$sroa$13$2 = $198;$__buf$sroa$28$2 = $__t$sroa$10$1$i;$__buf$sroa$46$2 = $200;
        }
       } else {
        $212 = $168;$__buf$sroa$0$2 = $__buf$sroa$0$147;$__buf$sroa$13$2 = $__buf$sroa$13$148;$__buf$sroa$28$2 = $__buf$sroa$28$149;$__buf$sroa$46$2 = $__buf$sroa$46$150;
       }
      } while(0);
      $211 = ((($212)) + -4|0); //@line 505 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $213 = HEAP32[$167>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      HEAP32[$211>>2] = $213; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      $214 = $__buf$sroa$13$2; //@line 506 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $215 = ((($214)) + -4|0); //@line 506 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $216 = $215; //@line 506 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $217 = HEAP32[$59>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
      $218 = ($167|0)==($217|0); //@line 2434 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
      if ($218) {
       $__buf$sroa$0$1$lcssa = $__buf$sroa$0$2;$__buf$sroa$13$1$lcssa = $216;$__buf$sroa$28$1$lcssa = $__buf$sroa$28$2;$__buf$sroa$46$1$lcssa = $__buf$sroa$46$2;
       break L46;
      } else {
       $__buf$sroa$0$147 = $__buf$sroa$0$2;$__buf$sroa$13$148 = $216;$__buf$sroa$28$149 = $__buf$sroa$28$2;$__buf$sroa$46$150 = $__buf$sroa$46$2;$__i$051 = $167;
      }
     }
     $219 = ___cxa_find_matching_catch_2()|0;
     $220 = tempRet0;
     $221 = ($__buf$sroa$0$147$lcssa|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     if ($221) {
      $$0162 = $219;$$064 = $220;
     } else {
      $$0161 = $219;$$063 = $220;$__buf$sroa$0$360 = $__buf$sroa$0$147$lcssa;
      break L37;
     }
     ___resumeException($$0162|0);
     // unreachable;
    }
   } while(0);
   $165 = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$this>>2] = $__buf$sroa$0$1$lcssa; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$59>>2] = $__buf$sroa$13$1$lcssa; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$57>>2] = $__buf$sroa$28$1$lcssa; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$63>>2] = $__buf$sroa$46$1$lcssa; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $166 = ($165|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($166) {
    STACKTOP = sp;return; //@line 2441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   }
   __ZdlPv($165); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   STACKTOP = sp;return; //@line 2441 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  }
 } while(0);
 if ((label|0) == 29) {
  $151 = ___cxa_find_matching_catch_3(0|0)|0;
  $152 = tempRet0;
  (___cxa_begin_catch(($151|0))|0); //@line 2426 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $153 = ((($127)) + -4|0); //@line 92 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $154 = HEAP32[$153>>2]|0; //@line 2429 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  __ZdlPv($154); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  __THREW__ = 0;
  invoke_v(3656); //@line 2430 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $155 = __THREW__; __THREW__ = 0;
  $156 = ___cxa_find_matching_catch_2()|0;
  $157 = tempRet0;
  __THREW__ = 0;
  invoke_v(3378); //@line 2431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $158 = __THREW__; __THREW__ = 0;
  $159 = $158&1;
  if ($159) {
   $223 = ___cxa_find_matching_catch_3(0|0)|0;
   $224 = tempRet0;
   ___clang_call_terminate($223); //@line 2431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   // unreachable; //@line 2431 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  } else {
   $$0161 = $156;$$063 = $157;$__buf$sroa$0$360 = $126;
  }
 }
 $222 = $__buf$sroa$0$360; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($222); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 $$0162 = $$0161;$$064 = $$063;
 ___resumeException($$0162|0);
 // unreachable;
}
function __ZNSt3__114__split_bufferIPPN4love7VariantENS_9allocatorIS4_EEE9push_backEOS4_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__t$sroa$10$0 = 0, $__t$sroa$10$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $1 = HEAP32[$0>>2]|0; //@line 593 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $2 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $4 = ($1|0)==($3|0); //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $5 = $3; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $6 = $1; //@line 579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 do {
  if ($4) {
   $7 = ((($this)) + 4|0); //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $8 = HEAP32[$7>>2]|0; //@line 592 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $9 = HEAP32[$this>>2]|0; //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $10 = ($8>>>0)>($9>>>0); //@line 581 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $11 = $9; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($10) {
    $12 = $8; //@line 583 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $13 = (($12) - ($11))|0; //@line 583 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $14 = $13 >> 2; //@line 583 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $15 = (($14) + 1)|0; //@line 584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $16 = (($15|0) / -2)&-1; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $17 = (($8) + ($16<<2)|0); //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $18 = (($6) - ($12))|0; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $19 = $18 >> 2; //@line 1892 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    _memmove(($17|0),($8|0),($18|0))|0; //@line 1893 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $20 = (($17) + ($19<<2)|0); //@line 1894 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    HEAP32[$0>>2] = $20; //@line 585 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $21 = HEAP32[$7>>2]|0; //@line 586 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $22 = (($21) + ($16<<2)|0); //@line 586 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    HEAP32[$7>>2] = $22; //@line 586 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $47 = $20;
    break;
   }
   $23 = (($5) - ($11))|0; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $24 = $23 >> 1; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $25 = ($24|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $26 = $25 ? 1 : $24; //@line 590 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $27 = $26 >>> 2; //@line 591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $28 = $26 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $29 = (__Znwj($28)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $30 = $29; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $31 = (($29) + ($27<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $32 = $31; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $33 = (($29) + ($26<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $34 = $33; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $35 = ($8|0)==($1|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($35) {
    $44 = $9;$__t$sroa$10$1 = $32;
   } else {
    $$in = $31;$37 = $8;$__t$sroa$10$0 = $32;
    while(1) {
     $36 = HEAP32[$37>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$$in>>2] = $36; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $38 = $__t$sroa$10$0; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $39 = ((($38)) + 4|0); //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $40 = $39; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $41 = ((($37)) + 4|0); //@line 954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $42 = ($41|0)==($1|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($42) {
      $$lcssa = $40;
      break;
     } else {
      $$in = $39;$37 = $41;$__t$sroa$10$0 = $40;
     }
    }
    $$pre = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $44 = $$pre;$__t$sroa$10$1 = $$lcssa;
   }
   HEAP32[$this>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$7>>2] = $32; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$0>>2] = $__t$sroa$10$1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$2>>2] = $34; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $43 = ($44|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $45 = $__t$sroa$10$1; //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($43) {
    $47 = $45;
   } else {
    __ZdlPv($44); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $$pre4 = HEAP32[$0>>2]|0; //@line 600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $47 = $$pre4;
   }
  } else {
   $47 = $1;
  }
 } while(0);
 $46 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$47>>2] = $46; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $48 = HEAP32[$0>>2]|0; //@line 602 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $49 = ((($48)) + 4|0); //@line 602 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 HEAP32[$0>>2] = $49; //@line 602 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 return; //@line 603 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
}
function __ZNSt3__114__split_bufferIPPN4love7VariantENS_9allocatorIS4_EEE10push_frontEOS4_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$cast = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__t$sroa$10$0 = 0, $__t$sroa$10$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 515 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $1 = HEAP32[$0>>2]|0; //@line 528 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $2 = HEAP32[$this>>2]|0; //@line 515 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $3 = ($1|0)==($2|0); //@line 515 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $4 = $2; //@line 526 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 do {
  if ($3) {
   $5 = ((($this)) + 8|0); //@line 517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $6 = HEAP32[$5>>2]|0; //@line 529 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $7 = ((($this)) + 12|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $8 = HEAP32[$7>>2]|0; //@line 517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $9 = ($6>>>0)<($8>>>0); //@line 517 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $$cast = $8; //@line 519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($9) {
    $10 = $6; //@line 519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $11 = (($$cast) - ($10))|0; //@line 519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $12 = $11 >> 2; //@line 519 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $13 = (($12) + 1)|0; //@line 520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $14 = (($13|0) / 2)&-1; //@line 520 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $15 = (($6) + ($14<<2)|0); //@line 521 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $16 = $1; //@line 1927 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $17 = (($10) - ($16))|0; //@line 1927 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $18 = $17 >> 2; //@line 1927 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $19 = (0 - ($18))|0; //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $20 = (($15) + ($19<<2)|0); //@line 1928 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    _memmove(($20|0),($1|0),($17|0))|0; //@line 1929 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    HEAP32[$0>>2] = $20; //@line 521 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $21 = HEAP32[$5>>2]|0; //@line 522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $22 = (($21) + ($14<<2)|0); //@line 522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    HEAP32[$5>>2] = $22; //@line 522 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $47 = $20;
    break;
   }
   $23 = (($$cast) - ($4))|0; //@line 526 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $24 = $23 >> 1; //@line 526 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $25 = ($24|0)==(0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $26 = $25 ? 1 : $24; //@line 526 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $27 = (($26) + 3)|0; //@line 527 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $28 = $27 >>> 2; //@line 527 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $29 = $26 << 2; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $30 = (__Znwj($29)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $31 = $30; //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $32 = (($30) + ($28<<2)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $33 = $32; //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $34 = (($30) + ($26<<2)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $35 = $34; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $36 = ($1|0)==($6|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
   if ($36) {
    $45 = $1;$__t$sroa$10$1 = $33;
   } else {
    $$in = $32;$38 = $1;$__t$sroa$10$0 = $33;
    while(1) {
     $37 = HEAP32[$38>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     HEAP32[$$in>>2] = $37; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $39 = $__t$sroa$10$0; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $40 = ((($39)) + 4|0); //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $41 = $40; //@line 283 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
     $42 = ((($38)) + 4|0); //@line 954 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     $43 = ($42|0)==($6|0); //@line 995 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
     if ($43) {
      $$lcssa = $41;
      break;
     } else {
      $$in = $40;$38 = $42;$__t$sroa$10$0 = $41;
     }
    }
    $$pre = HEAP32[$this>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $45 = $$pre;$__t$sroa$10$1 = $$lcssa;
   }
   HEAP32[$this>>2] = $31; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$0>>2] = $33; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$5>>2] = $__t$sroa$10$1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$7>>2] = $35; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $44 = ($45|0)==(0|0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   if ($44) {
    $47 = $32;
   } else {
    __ZdlPv($45); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
    $$pre4 = HEAP32[$0>>2]|0; //@line 536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
    $47 = $$pre4;
   }
  } else {
   $47 = $1;
  }
 } while(0);
 $46 = ((($47)) + -4|0); //@line 536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $48 = HEAP32[$__x>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$46>>2] = $48; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $49 = HEAP32[$0>>2]|0; //@line 538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $50 = ((($49)) + -4|0); //@line 538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 HEAP32[$0>>2] = $50; //@line 538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 return; //@line 539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
}
function __ZN4love6thread7Channel6supplyEPNS_7VariantE($this,$var) {
 $this = $this|0;
 $var = $var|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $l = 0, $lpad$phi$index = 0, $lpad$phi$index3 = 0, $notlhs$i = 0, $notrhs$i = 0, $or$cond = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = ($var|0)==(0|0); //@line 125 "modules/thread/Channel.cpp"
 if ($0) {
  STACKTOP = sp;return; //@line 133 "modules/thread/Channel.cpp"
 }
 $1 = ((($this)) + 8|0); //@line 128 "modules/thread/Channel.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 128 "modules/thread/Channel.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($l,$2); //@line 128 "modules/thread/Channel.cpp"
 __THREW__ = 0;
 $3 = (invoke_iii(4236,($this|0),($var|0))|0); //@line 129 "modules/thread/Channel.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $31 = ___cxa_find_matching_catch_2()|0;
  $32 = tempRet0;
  $lpad$phi$index = $31;$lpad$phi$index3 = $32;
  __ZN4love6thread4LockD2Ev($l); //@line 133 "modules/thread/Channel.cpp"
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $6 = ((($this)) + 60|0); //@line 131 "modules/thread/Channel.cpp"
 $7 = ((($this)) + 12|0); //@line 132 "modules/thread/Channel.cpp"
 $notlhs$i = ($3|0)>(-1); //@line 45 "modules/thread/Channel.cpp"
 L8: do {
  if ($notlhs$i) {
   while(1) {
    $8 = HEAP32[$6>>2]|0; //@line 131 "modules/thread/Channel.cpp"
    $9 = ($3>>>0)>($8>>>0); //@line 36 "modules/thread/Channel.cpp"
    if (!($9)) {
     break L8;
    }
    $10 = HEAP32[$7>>2]|0; //@line 132 "modules/thread/Channel.cpp"
    $11 = HEAP32[$10>>2]|0; //@line 132 "modules/thread/Channel.cpp"
    $12 = ((($11)) + 16|0); //@line 132 "modules/thread/Channel.cpp"
    $13 = HEAP32[$12>>2]|0; //@line 132 "modules/thread/Channel.cpp"
    $14 = HEAP32[$1>>2]|0; //@line 132 "modules/thread/Channel.cpp"
    __THREW__ = 0;
    (invoke_iiii($13|0,($10|0),($14|0),-1)|0); //@line 132 "modules/thread/Channel.cpp"
    $15 = __THREW__; __THREW__ = 0;
    $16 = $15&1;
    if ($16) {
     break;
    }
   }
   $17 = ___cxa_find_matching_catch_2()|0;
   $18 = tempRet0;
   $lpad$phi$index = $17;$lpad$phi$index3 = $18;
   __ZN4love6thread4LockD2Ev($l); //@line 133 "modules/thread/Channel.cpp"
   ___resumeException($lpad$phi$index|0);
   // unreachable;
  } else {
   while(1) {
    $19 = HEAP32[$6>>2]|0; //@line 131 "modules/thread/Channel.cpp"
    $20 = ($3>>>0)>($19>>>0); //@line 36 "modules/thread/Channel.cpp"
    if (!($20)) {
     $21 = ($3|0)==($19|0); //@line 38 "modules/thread/Channel.cpp"
     $notrhs$i = ($19|0)<(1); //@line 45 "modules/thread/Channel.cpp"
     $or$cond = $21 | $notrhs$i; //@line 38 "modules/thread/Channel.cpp"
     if ($or$cond) {
      break L8;
     }
    }
    $22 = HEAP32[$7>>2]|0; //@line 132 "modules/thread/Channel.cpp"
    $23 = HEAP32[$22>>2]|0; //@line 132 "modules/thread/Channel.cpp"
    $24 = ((($23)) + 16|0); //@line 132 "modules/thread/Channel.cpp"
    $25 = HEAP32[$24>>2]|0; //@line 132 "modules/thread/Channel.cpp"
    $26 = HEAP32[$1>>2]|0; //@line 132 "modules/thread/Channel.cpp"
    __THREW__ = 0;
    (invoke_iiii($25|0,($22|0),($26|0),-1)|0); //@line 132 "modules/thread/Channel.cpp"
    $27 = __THREW__; __THREW__ = 0;
    $28 = $27&1;
    if ($28) {
     break;
    }
   }
   $29 = ___cxa_find_matching_catch_2()|0;
   $30 = tempRet0;
   $lpad$phi$index = $29;$lpad$phi$index3 = $30;
   __ZN4love6thread4LockD2Ev($l); //@line 133 "modules/thread/Channel.cpp"
   ___resumeException($lpad$phi$index|0);
   // unreachable;
  }
 } while(0);
 __ZN4love6thread4LockD2Ev($l); //@line 133 "modules/thread/Channel.cpp"
 STACKTOP = sp;return; //@line 133 "modules/thread/Channel.cpp"
}
function __ZN4love6thread7Channel3popEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = ((($this)) + 8|0); //@line 137 "modules/thread/Channel.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 137 "modules/thread/Channel.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($l,$1); //@line 137 "modules/thread/Channel.cpp"
 $2 = ((($this)) + 36|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $4 = ($3|0)==(0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($4) {
  $$0 = 0;
  __ZN4love6thread4LockD2Ev($l); //@line 153 "modules/thread/Channel.cpp"
  STACKTOP = sp;return ($$0|0); //@line 153 "modules/thread/Channel.cpp"
 }
 $7 = ((($this)) + 20|0); //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $8 = HEAP32[$7>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $9 = ((($this)) + 32|0); //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $10 = HEAP32[$9>>2]|0; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $11 = $10 >>> 10; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $12 = (($8) + ($11<<2)|0); //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $13 = HEAP32[$12>>2]|0; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $14 = $10 & 1023; //@line 1734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $15 = (($13) + ($14<<2)|0); //@line 1734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $16 = HEAP32[$15>>2]|0; //@line 141 "modules/thread/Channel.cpp"
 $17 = (($3) + -1)|0; //@line 2538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 HEAP32[$2>>2] = $17; //@line 2538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $18 = (($10) + 1)|0; //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 HEAP32[$9>>2] = $18; //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $19 = ($18>>>0)>(2047); //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($19) {
  $20 = HEAP32[$8>>2]|0; //@line 2541 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  __ZdlPv($20); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $21 = HEAP32[$7>>2]|0; //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $22 = ((($21)) + 4|0); //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  HEAP32[$7>>2] = $22; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $23 = HEAP32[$9>>2]|0; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $24 = (($23) + -1024)|0; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  HEAP32[$9>>2] = $24; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 }
 $25 = ((($this)) + 60|0); //@line 144 "modules/thread/Channel.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 144 "modules/thread/Channel.cpp"
 $27 = (($26) + 1)|0; //@line 144 "modules/thread/Channel.cpp"
 HEAP32[$25>>2] = $27; //@line 144 "modules/thread/Channel.cpp"
 $28 = ((($this)) + 12|0); //@line 145 "modules/thread/Channel.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 145 "modules/thread/Channel.cpp"
 $30 = HEAP32[$29>>2]|0; //@line 145 "modules/thread/Channel.cpp"
 $31 = ((($30)) + 12|0); //@line 145 "modules/thread/Channel.cpp"
 $32 = HEAP32[$31>>2]|0; //@line 145 "modules/thread/Channel.cpp"
 __THREW__ = 0;
 invoke_vi($32|0,($29|0)); //@line 145 "modules/thread/Channel.cpp"
 $33 = __THREW__; __THREW__ = 0;
 $34 = $33&1;
 if ($34) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  __ZN4love6thread4LockD2Ev($l); //@line 153 "modules/thread/Channel.cpp"
  ___resumeException($5|0);
  // unreachable;
 }
 $35 = ((($this)) + 40|0); //@line 149 "modules/thread/Channel.cpp"
 $36 = HEAP8[$35>>0]|0; //@line 149 "modules/thread/Channel.cpp"
 $37 = ($36<<24>>24)==(0); //@line 149 "modules/thread/Channel.cpp"
 if ($37) {
  $$0 = $16;
  __ZN4love6thread4LockD2Ev($l); //@line 153 "modules/thread/Channel.cpp"
  STACKTOP = sp;return ($$0|0); //@line 153 "modules/thread/Channel.cpp"
 }
 $38 = HEAP32[$2>>2]|0; //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $39 = ($38|0)==(0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if (!($39)) {
  $$0 = $16;
  __ZN4love6thread4LockD2Ev($l); //@line 153 "modules/thread/Channel.cpp"
  STACKTOP = sp;return ($$0|0); //@line 153 "modules/thread/Channel.cpp"
 }
 $40 = HEAP32[$this>>2]|0; //@line 150 "modules/thread/Channel.cpp"
 $41 = ((($40)) + 12|0); //@line 150 "modules/thread/Channel.cpp"
 $42 = HEAP32[$41>>2]|0; //@line 150 "modules/thread/Channel.cpp"
 __THREW__ = 0;
 invoke_vi($42|0,($this|0)); //@line 150 "modules/thread/Channel.cpp"
 $43 = __THREW__; __THREW__ = 0;
 $44 = $43&1;
 if ($44) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  __ZN4love6thread4LockD2Ev($l); //@line 153 "modules/thread/Channel.cpp"
  ___resumeException($5|0);
  // unreachable;
 } else {
  $$0 = $16;
  __ZN4love6thread4LockD2Ev($l); //@line 153 "modules/thread/Channel.cpp"
  STACKTOP = sp;return ($$0|0); //@line 153 "modules/thread/Channel.cpp"
 }
 return (0)|0;
}
function __ZN4love6thread7Channel6demandEv($this) {
 $this = $this|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = ((($this)) + 8|0); //@line 158 "modules/thread/Channel.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 158 "modules/thread/Channel.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($l,$1); //@line 158 "modules/thread/Channel.cpp"
 $2 = ((($this)) + 12|0); //@line 160 "modules/thread/Channel.cpp"
 while(1) {
  __THREW__ = 0;
  $3 = (invoke_ii(4237,($this|0))|0); //@line 159 "modules/thread/Channel.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   label = 5;
   break;
  }
  $6 = ($3|0)==(0|0); //@line 159 "modules/thread/Channel.cpp"
  if (!($6)) {
   $$lcssa = $3;
   label = 6;
   break;
  }
  $7 = HEAP32[$2>>2]|0; //@line 160 "modules/thread/Channel.cpp"
  $8 = HEAP32[$7>>2]|0; //@line 160 "modules/thread/Channel.cpp"
  $9 = ((($8)) + 16|0); //@line 160 "modules/thread/Channel.cpp"
  $10 = HEAP32[$9>>2]|0; //@line 160 "modules/thread/Channel.cpp"
  $11 = HEAP32[$0>>2]|0; //@line 160 "modules/thread/Channel.cpp"
  __THREW__ = 0;
  (invoke_iiii($10|0,($7|0),($11|0),-1)|0); //@line 160 "modules/thread/Channel.cpp"
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if ($13) {
   label = 5;
   break;
  }
 }
 if ((label|0) == 5) {
  $14 = ___cxa_find_matching_catch_2()|0;
  $15 = tempRet0;
  __ZN4love6thread4LockD2Ev($l); //@line 163 "modules/thread/Channel.cpp"
  ___resumeException($14|0);
  // unreachable;
 }
 else if ((label|0) == 6) {
  __ZN4love6thread4LockD2Ev($l); //@line 163 "modules/thread/Channel.cpp"
  STACKTOP = sp;return ($$lcssa|0); //@line 163 "modules/thread/Channel.cpp"
 }
 return (0)|0;
}
function __ZN4love6thread7Channel4peekEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = ((($this)) + 8|0); //@line 167 "modules/thread/Channel.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 167 "modules/thread/Channel.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($l,$1); //@line 167 "modules/thread/Channel.cpp"
 $2 = ((($this)) + 36|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $4 = ($3|0)==(0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($4) {
  $$0 = 0;
  __ZN4love6thread4LockD2Ev($l); //@line 174 "modules/thread/Channel.cpp"
  STACKTOP = sp;return ($$0|0); //@line 174 "modules/thread/Channel.cpp"
 }
 $7 = ((($this)) + 20|0); //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $8 = HEAP32[$7>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $9 = ((($this)) + 32|0); //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $10 = HEAP32[$9>>2]|0; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $11 = $10 >>> 10; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $12 = (($8) + ($11<<2)|0); //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $13 = HEAP32[$12>>2]|0; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $14 = $10 & 1023; //@line 1734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $15 = (($13) + ($14<<2)|0); //@line 1734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $16 = HEAP32[$15>>2]|0; //@line 171 "modules/thread/Channel.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 172 "modules/thread/Channel.cpp"
 $18 = ((($17)) + 8|0); //@line 172 "modules/thread/Channel.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 172 "modules/thread/Channel.cpp"
 __THREW__ = 0;
 invoke_vi($19|0,($16|0)); //@line 172 "modules/thread/Channel.cpp"
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  __ZN4love6thread4LockD2Ev($l); //@line 174 "modules/thread/Channel.cpp"
  ___resumeException($5|0);
  // unreachable;
 } else {
  $$0 = $16;
  __ZN4love6thread4LockD2Ev($l); //@line 174 "modules/thread/Channel.cpp"
  STACKTOP = sp;return ($$0|0); //@line 174 "modules/thread/Channel.cpp"
 }
 return (0)|0;
}
function __ZN4love6thread7Channel8getCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = ((($this)) + 8|0); //@line 178 "modules/thread/Channel.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 178 "modules/thread/Channel.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($l,$1); //@line 178 "modules/thread/Channel.cpp"
 $2 = ((($this)) + 36|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 1299 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 __ZN4love6thread4LockD2Ev($l); //@line 180 "modules/thread/Channel.cpp"
 STACKTOP = sp;return ($3|0); //@line 180 "modules/thread/Channel.cpp"
}
function __ZN4love6thread7Channel5clearEv($this) {
 $this = $this|0;
 var $$pr = 0, $$pr$pre = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $l = 0, $lpad$phi$index = 0;
 var $lpad$phi$index2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = ((($this)) + 8|0); //@line 184 "modules/thread/Channel.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 184 "modules/thread/Channel.cpp"
 __ZN4love6thread4LockC2EPNS0_5MutexE($l,$1); //@line 184 "modules/thread/Channel.cpp"
 $2 = ((($this)) + 36|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $3 = HEAP32[$2>>2]|0; //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $4 = ($3|0)==(0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 if ($4) {
  __ZN4love6thread4LockD2Ev($l); //@line 204 "modules/thread/Channel.cpp"
  STACKTOP = sp;return; //@line 204 "modules/thread/Channel.cpp"
 }
 $5 = ((($this)) + 20|0); //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $6 = ((($this)) + 32|0); //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $$pre = HEAP32[$6>>2]|0; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
 $13 = $$pre;
 while(1) {
  $11 = HEAP32[$5>>2]|0; //@line 76 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  $12 = $13 >>> 10; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $14 = (($11) + ($12<<2)|0); //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $15 = HEAP32[$14>>2]|0; //@line 1733 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $16 = $13 & 1023; //@line 1734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $17 = (($15) + ($16<<2)|0); //@line 1734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $18 = HEAP32[$17>>2]|0; //@line 192 "modules/thread/Channel.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 192 "modules/thread/Channel.cpp"
  $20 = ((($19)) + 12|0); //@line 192 "modules/thread/Channel.cpp"
  $21 = HEAP32[$20>>2]|0; //@line 192 "modules/thread/Channel.cpp"
  __THREW__ = 0;
  invoke_vi($21|0,($18|0)); //@line 192 "modules/thread/Channel.cpp"
  $22 = __THREW__; __THREW__ = 0;
  $23 = $22&1;
  if ($23) {
   label = 3;
   break;
  }
  $24 = HEAP32[$2>>2]|0; //@line 2538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $25 = (($24) + -1)|0; //@line 2538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  HEAP32[$2>>2] = $25; //@line 2538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $26 = HEAP32[$6>>2]|0; //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $27 = (($26) + 1)|0; //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  HEAP32[$6>>2] = $27; //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  $28 = ($27>>>0)>(2047); //@line 2539 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  if ($28) {
   $29 = HEAP32[$5>>2]|0; //@line 90 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $30 = HEAP32[$29>>2]|0; //@line 2541 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   __ZdlPv($30); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
   $31 = HEAP32[$5>>2]|0; //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $32 = ((($31)) + 4|0); //@line 108 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   HEAP32[$5>>2] = $32; //@line 301 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
   $33 = HEAP32[$6>>2]|0; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $34 = (($33) + -1024)|0; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   HEAP32[$6>>2] = $34; //@line 2543 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $$pr$pre = HEAP32[$2>>2]|0; //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
   $$pr = $$pr$pre;$54 = $34;
  } else {
   $$pr = $25;$54 = $27;
  }
  $35 = ($$pr|0)==(0); //@line 1307 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/deque"
  if ($35) {
   break;
  } else {
   $13 = $54;
  }
 }
 if ((label|0) == 3) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  $lpad$phi$index = $7;$lpad$phi$index2 = $8;
  __ZN4love6thread4LockD2Ev($l); //@line 204 "modules/thread/Channel.cpp"
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $36 = ((($this)) + 56|0); //@line 197 "modules/thread/Channel.cpp"
 $37 = HEAP32[$36>>2]|0; //@line 197 "modules/thread/Channel.cpp"
 $38 = ((($this)) + 60|0); //@line 197 "modules/thread/Channel.cpp"
 HEAP32[$38>>2] = $37; //@line 197 "modules/thread/Channel.cpp"
 $39 = ((($this)) + 12|0); //@line 198 "modules/thread/Channel.cpp"
 $40 = HEAP32[$39>>2]|0; //@line 198 "modules/thread/Channel.cpp"
 $41 = HEAP32[$40>>2]|0; //@line 198 "modules/thread/Channel.cpp"
 $42 = ((($41)) + 12|0); //@line 198 "modules/thread/Channel.cpp"
 $43 = HEAP32[$42>>2]|0; //@line 198 "modules/thread/Channel.cpp"
 __THREW__ = 0;
 invoke_vi($43|0,($40|0)); //@line 198 "modules/thread/Channel.cpp"
 $44 = __THREW__; __THREW__ = 0;
 $45 = $44&1;
 if (!($45)) {
  $46 = ((($this)) + 40|0); //@line 202 "modules/thread/Channel.cpp"
  $47 = HEAP8[$46>>0]|0; //@line 202 "modules/thread/Channel.cpp"
  $48 = ($47<<24>>24)==(0); //@line 202 "modules/thread/Channel.cpp"
  if ($48) {
   __ZN4love6thread4LockD2Ev($l); //@line 204 "modules/thread/Channel.cpp"
   STACKTOP = sp;return; //@line 204 "modules/thread/Channel.cpp"
  }
  $49 = HEAP32[$this>>2]|0; //@line 203 "modules/thread/Channel.cpp"
  $50 = ((($49)) + 12|0); //@line 203 "modules/thread/Channel.cpp"
  $51 = HEAP32[$50>>2]|0; //@line 203 "modules/thread/Channel.cpp"
  __THREW__ = 0;
  invoke_vi($51|0,($this|0)); //@line 203 "modules/thread/Channel.cpp"
  $52 = __THREW__; __THREW__ = 0;
  $53 = $52&1;
  if (!($53)) {
   __ZN4love6thread4LockD2Ev($l); //@line 204 "modules/thread/Channel.cpp"
   STACKTOP = sp;return; //@line 204 "modules/thread/Channel.cpp"
  }
 }
 $9 = ___cxa_find_matching_catch_2()|0;
 $10 = tempRet0;
 $lpad$phi$index = $9;$lpad$phi$index2 = $10;
 __ZN4love6thread4LockD2Ev($l); //@line 204 "modules/thread/Channel.cpp"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZN4love6thread7Channel9lockMutexEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 208 "modules/thread/Channel.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 208 "modules/thread/Channel.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 208 "modules/thread/Channel.cpp"
 $3 = ((($2)) + 8|0); //@line 208 "modules/thread/Channel.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 208 "modules/thread/Channel.cpp"
 FUNCTION_TABLE_vi[$4 & 8191]($1); //@line 208 "modules/thread/Channel.cpp"
 return; //@line 209 "modules/thread/Channel.cpp"
}
function __ZN4love6thread7Channel11unlockMutexEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 213 "modules/thread/Channel.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 213 "modules/thread/Channel.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 213 "modules/thread/Channel.cpp"
 $3 = ((($2)) + 12|0); //@line 213 "modules/thread/Channel.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 213 "modules/thread/Channel.cpp"
 FUNCTION_TABLE_vi[$4 & 8191]($1); //@line 213 "modules/thread/Channel.cpp"
 return; //@line 214 "modules/thread/Channel.cpp"
}
function __ZN4love6thread9LuaThreadD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $i$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (33076); //@line 43 "modules/thread/LuaThread.cpp"
 $0 = ((($this)) + 8|0); //@line 43 "modules/thread/LuaThread.cpp"
 HEAP32[$0>>2] = (33104); //@line 43 "modules/thread/LuaThread.cpp"
 $1 = ((($this)) + 60|0); //@line 46 "modules/thread/LuaThread.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 46 "modules/thread/LuaThread.cpp"
 $3 = ($2|0)>(0); //@line 46 "modules/thread/LuaThread.cpp"
 L1: do {
  if ($3) {
   $4 = ((($this)) + 56|0); //@line 47 "modules/thread/LuaThread.cpp"
   $i$02 = 0;
   while(1) {
    $17 = HEAP32[$4>>2]|0; //@line 47 "modules/thread/LuaThread.cpp"
    $18 = (($17) + ($i$02<<2)|0); //@line 47 "modules/thread/LuaThread.cpp"
    $19 = HEAP32[$18>>2]|0; //@line 47 "modules/thread/LuaThread.cpp"
    $20 = HEAP32[$19>>2]|0; //@line 47 "modules/thread/LuaThread.cpp"
    $21 = ((($20)) + 12|0); //@line 47 "modules/thread/LuaThread.cpp"
    $22 = HEAP32[$21>>2]|0; //@line 47 "modules/thread/LuaThread.cpp"
    __THREW__ = 0;
    invoke_vi($22|0,($19|0)); //@line 47 "modules/thread/LuaThread.cpp"
    $23 = __THREW__; __THREW__ = 0;
    $24 = $23&1;
    if ($24) {
     break;
    }
    $25 = (($i$02) + 1)|0; //@line 46 "modules/thread/LuaThread.cpp"
    $26 = HEAP32[$1>>2]|0; //@line 46 "modules/thread/LuaThread.cpp"
    $27 = ($25|0)<($26|0); //@line 46 "modules/thread/LuaThread.cpp"
    if ($27) {
     $i$02 = $25;
    } else {
     break L1;
    }
   }
   $28 = ___cxa_find_matching_catch_3(0|0)|0;
   $29 = tempRet0;
   $30 = ((($this)) + 44|0); //@line 48 "modules/thread/LuaThread.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($30); //@line 48 "modules/thread/LuaThread.cpp"
   $31 = ((($this)) + 32|0); //@line 48 "modules/thread/LuaThread.cpp"
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($31); //@line 48 "modules/thread/LuaThread.cpp"
   $32 = ((($this)) + 28|0); //@line 106 "./common/Object.h"
   $33 = HEAP32[$32>>2]|0; //@line 106 "./common/Object.h"
   $34 = ($33|0)==(0|0); //@line 106 "./common/Object.h"
   if ($34) {
    __ZN4love6thread10ThreadableD2Ev($0); //@line 48 "modules/thread/LuaThread.cpp"
    ___clang_call_terminate($28); //@line 48 "modules/thread/LuaThread.cpp"
    // unreachable; //@line 48 "modules/thread/LuaThread.cpp"
   }
   $35 = HEAP32[$33>>2]|0; //@line 106 "./common/Object.h"
   $36 = ((($35)) + 12|0); //@line 106 "./common/Object.h"
   $37 = HEAP32[$36>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($37|0,($33|0)); //@line 106 "./common/Object.h"
   $38 = __THREW__; __THREW__ = 0;
   $39 = $38&1;
   if ($39) {
    $40 = ___cxa_find_matching_catch_3(0|0)|0;
    $41 = tempRet0;
    ___clang_call_terminate($40); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   } else {
    __ZN4love6thread10ThreadableD2Ev($0); //@line 48 "modules/thread/LuaThread.cpp"
    ___clang_call_terminate($28); //@line 48 "modules/thread/LuaThread.cpp"
    // unreachable; //@line 48 "modules/thread/LuaThread.cpp"
   }
  }
 } while(0);
 $5 = ((($this)) + 44|0); //@line 48 "modules/thread/LuaThread.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5); //@line 48 "modules/thread/LuaThread.cpp"
 $6 = ((($this)) + 32|0); //@line 48 "modules/thread/LuaThread.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6); //@line 48 "modules/thread/LuaThread.cpp"
 $7 = ((($this)) + 28|0); //@line 106 "./common/Object.h"
 $8 = HEAP32[$7>>2]|0; //@line 106 "./common/Object.h"
 $9 = ($8|0)==(0|0); //@line 106 "./common/Object.h"
 if ($9) {
  __ZN4love6thread10ThreadableD2Ev($0); //@line 48 "modules/thread/LuaThread.cpp"
  return; //@line 48 "modules/thread/LuaThread.cpp"
 }
 $10 = HEAP32[$8>>2]|0; //@line 106 "./common/Object.h"
 $11 = ((($10)) + 12|0); //@line 106 "./common/Object.h"
 $12 = HEAP32[$11>>2]|0; //@line 106 "./common/Object.h"
 __THREW__ = 0;
 invoke_vi($12|0,($8|0)); //@line 106 "./common/Object.h"
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if ($14) {
  $15 = ___cxa_find_matching_catch_3(0|0)|0;
  $16 = tempRet0;
  ___clang_call_terminate($15); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 } else {
  __ZN4love6thread10ThreadableD2Ev($0); //@line 48 "modules/thread/LuaThread.cpp"
  return; //@line 48 "modules/thread/LuaThread.cpp"
 }
}
function __ZN4love6thread9LuaThreadD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6thread9LuaThreadD2Ev($this); //@line 43 "modules/thread/LuaThread.cpp"
 __ZdlPv($this); //@line 43 "modules/thread/LuaThread.cpp"
 return; //@line 48 "modules/thread/LuaThread.cpp"
}
function __ZN4love6thread9LuaThread14threadFunctionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $i$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 12|0;
 $1 = sp;
 $2 = HEAP32[$this>>2]|0; //@line 52 "modules/thread/LuaThread.cpp"
 $3 = ((($2)) + 8|0); //@line 52 "modules/thread/LuaThread.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 52 "modules/thread/LuaThread.cpp"
 FUNCTION_TABLE_vi[$4 & 8191]($this); //@line 52 "modules/thread/LuaThread.cpp"
 $5 = ((($this)) + 44|0); //@line 53 "modules/thread/LuaThread.cpp"
 $6 = HEAP8[$5>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $7 = $6 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $8 = ($7<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($8) {
  $12 = ((($5)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$12>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$5>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 } else {
  $9 = ((($this)) + 52|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $10 = HEAP32[$9>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP8[$10>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $11 = ((($this)) + 48|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  HEAP32[$11>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 }
 $13 = (_luaL_newstate()|0); //@line 55 "modules/thread/LuaThread.cpp"
 _luaL_openlibs($13); //@line 56 "modules/thread/LuaThread.cpp"
 (__ZN4love12luax_preloadEP9lua_StatePFiS1_EPKc($13,3346,445428)|0); //@line 59 "modules/thread/LuaThread.cpp"
 (__ZN4love12luax_requireEP9lua_StatePKc($13,445428)|0); //@line 60 "modules/thread/LuaThread.cpp"
 _lua_settop($13,-2); //@line 61 "modules/thread/LuaThread.cpp"
 (__ZN4love12luax_requireEP9lua_StatePKc($13,445433)|0); //@line 64 "modules/thread/LuaThread.cpp"
 _lua_settop($13,-2); //@line 65 "modules/thread/LuaThread.cpp"
 (__ZN4love12luax_requireEP9lua_StatePKc($13,445445)|0); //@line 70 "modules/thread/LuaThread.cpp"
 _lua_settop($13,-2); //@line 71 "modules/thread/LuaThread.cpp"
 $14 = ((($this)) + 28|0); //@line 117 "./common/Object.h"
 $15 = HEAP32[$14>>2]|0; //@line 117 "./common/Object.h"
 $16 = HEAP32[$15>>2]|0; //@line 73 "modules/thread/LuaThread.cpp"
 $17 = ((($16)) + 16|0); //@line 73 "modules/thread/LuaThread.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 73 "modules/thread/LuaThread.cpp"
 $19 = (FUNCTION_TABLE_ii[$18 & 8191]($15)|0); //@line 73 "modules/thread/LuaThread.cpp"
 $20 = HEAP32[$14>>2]|0; //@line 117 "./common/Object.h"
 $21 = HEAP32[$20>>2]|0; //@line 73 "modules/thread/LuaThread.cpp"
 $22 = ((($21)) + 20|0); //@line 73 "modules/thread/LuaThread.cpp"
 $23 = HEAP32[$22>>2]|0; //@line 73 "modules/thread/LuaThread.cpp"
 $24 = (FUNCTION_TABLE_ii[$23 & 8191]($20)|0); //@line 73 "modules/thread/LuaThread.cpp"
 $25 = ((($this)) + 32|0); //@line 73 "modules/thread/LuaThread.cpp"
 $26 = HEAP8[$25>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $27 = $26 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $28 = ($27<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($28) {
  $31 = ((($25)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $32 = $31;
 } else {
  $29 = ((($this)) + 40|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $30 = HEAP32[$29>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $32 = $30;
 }
 $33 = (_luaL_loadbuffer($13,$19,$24,$32)|0); //@line 73 "modules/thread/LuaThread.cpp"
 $34 = ($33|0)==(0); //@line 73 "modules/thread/LuaThread.cpp"
 do {
  if ($34) {
   $46 = ((($this)) + 60|0); //@line 77 "modules/thread/LuaThread.cpp"
   $47 = HEAP32[$46>>2]|0; //@line 77 "modules/thread/LuaThread.cpp"
   $48 = ($47|0)>(0); //@line 78 "modules/thread/LuaThread.cpp"
   $49 = ((($this)) + 56|0); //@line 80 "modules/thread/LuaThread.cpp"
   if ($48) {
    $i$03 = 0;
    while(1) {
     $52 = HEAP32[$49>>2]|0; //@line 80 "modules/thread/LuaThread.cpp"
     $53 = (($52) + ($i$03<<2)|0); //@line 80 "modules/thread/LuaThread.cpp"
     $54 = HEAP32[$53>>2]|0; //@line 80 "modules/thread/LuaThread.cpp"
     __ZN4love7Variant5toLuaEP9lua_State($54,$13); //@line 80 "modules/thread/LuaThread.cpp"
     $55 = HEAP32[$49>>2]|0; //@line 81 "modules/thread/LuaThread.cpp"
     $56 = (($55) + ($i$03<<2)|0); //@line 81 "modules/thread/LuaThread.cpp"
     $57 = HEAP32[$56>>2]|0; //@line 81 "modules/thread/LuaThread.cpp"
     $58 = HEAP32[$57>>2]|0; //@line 81 "modules/thread/LuaThread.cpp"
     $59 = ((($58)) + 12|0); //@line 81 "modules/thread/LuaThread.cpp"
     $60 = HEAP32[$59>>2]|0; //@line 81 "modules/thread/LuaThread.cpp"
     FUNCTION_TABLE_vi[$60 & 8191]($57); //@line 81 "modules/thread/LuaThread.cpp"
     $61 = (($i$03) + 1)|0; //@line 78 "modules/thread/LuaThread.cpp"
     $62 = HEAP32[$46>>2]|0; //@line 78 "modules/thread/LuaThread.cpp"
     $63 = ($61|0)<($62|0); //@line 78 "modules/thread/LuaThread.cpp"
     if ($63) {
      $i$03 = $61;
     } else {
      break;
     }
    }
   }
   HEAP32[$46>>2] = 0; //@line 85 "modules/thread/LuaThread.cpp"
   HEAP32[$49>>2] = 0; //@line 86 "modules/thread/LuaThread.cpp"
   $50 = (_lua_pcall($13,$47,0,0)|0); //@line 88 "modules/thread/LuaThread.cpp"
   $51 = ($50|0)==(0); //@line 88 "modules/thread/LuaThread.cpp"
   if (!($51)) {
    __ZN4love13luax_tostringEP9lua_Statei($1,$13,-1); //@line 89 "modules/thread/LuaThread.cpp"
    $64 = HEAP8[$5>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $65 = $64 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $66 = ($65<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    if ($66) {
     $70 = ((($5)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP8[$70>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP8[$5>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    } else {
     $67 = ((($this)) + 52|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $68 = HEAP32[$67>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP8[$68>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $69 = ((($this)) + 48|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     HEAP32[$69>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    }
    __THREW__ = 0;
    invoke_vii(3368,($5|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $71 = __THREW__; __THREW__ = 0;
    $72 = $71&1;
    if ($72) {
     $73 = ___cxa_find_matching_catch_3(0|0)|0;
     $74 = tempRet0;
     ___clang_call_terminate($73); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    } else {
     ;HEAP32[$5>>2]=HEAP32[$1>>2]|0;HEAP32[$5+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$1+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 89 "modules/thread/LuaThread.cpp"
     break;
    }
   }
  } else {
   __ZN4love13luax_tostringEP9lua_Statei($0,$13,-1); //@line 74 "modules/thread/LuaThread.cpp"
   $35 = HEAP8[$5>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $36 = $35 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $37 = ($36<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if ($37) {
    $41 = ((($5)) + 1|0); //@line 1749 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$41>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$5>>0] = 0; //@line 1708 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   } else {
    $38 = ((($this)) + 52|0); //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $39 = HEAP32[$38>>2]|0; //@line 1743 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP8[$39>>0] = 0; //@line 632 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $40 = ((($this)) + 48|0); //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    HEAP32[$40>>2] = 0; //@line 1723 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   }
   __THREW__ = 0;
   invoke_vii(3368,($5|0),0); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $42 = __THREW__; __THREW__ = 0;
   $43 = $42&1;
   if ($43) {
    $44 = ___cxa_find_matching_catch_3(0|0)|0;
    $45 = tempRet0;
    ___clang_call_terminate($44); //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    // unreachable; //@line 1442 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   } else {
    ;HEAP32[$5>>2]=HEAP32[$0>>2]|0;HEAP32[$5+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$0+8>>2]|0; //@line 2440 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0; //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0); //@line 74 "modules/thread/LuaThread.cpp"
    break;
   }
  }
 } while(0);
 _lua_close($13); //@line 92 "modules/thread/LuaThread.cpp"
 $75 = HEAP8[$5>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $76 = $75 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $77 = ($76<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($77) {
  $80 = $75&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $81 = $80 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $83 = $81;
 } else {
  $78 = ((($this)) + 48|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $79 = HEAP32[$78>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $83 = $79;
 }
 $82 = ($83|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($82) {
  $84 = HEAP32[$this>>2]|0; //@line 97 "modules/thread/LuaThread.cpp"
  $85 = ((($84)) + 12|0); //@line 97 "modules/thread/LuaThread.cpp"
  $86 = HEAP32[$85>>2]|0; //@line 97 "modules/thread/LuaThread.cpp"
  FUNCTION_TABLE_vi[$86 & 8191]($this); //@line 97 "modules/thread/LuaThread.cpp"
  STACKTOP = sp;return; //@line 98 "modules/thread/LuaThread.cpp"
 }
 __ZN4love6thread9LuaThread7onErrorEv($this); //@line 95 "modules/thread/LuaThread.cpp"
 $84 = HEAP32[$this>>2]|0; //@line 97 "modules/thread/LuaThread.cpp"
 $85 = ((($84)) + 12|0); //@line 97 "modules/thread/LuaThread.cpp"
 $86 = HEAP32[$85>>2]|0; //@line 97 "modules/thread/LuaThread.cpp"
 FUNCTION_TABLE_vi[$86 & 8191]($this); //@line 97 "modules/thread/LuaThread.cpp"
 STACKTOP = sp;return; //@line 98 "modules/thread/LuaThread.cpp"
}
function __ZN4love6thread9LuaThread7onErrorEv($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$03 = 0, $$1 = 0, $$12 = 0, $$14 = 0, $$2 = 0, $$25 = 0, $$3 = 0, $$36 = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i$i = 0, $$pre$i$i$i9 = 0, $$pre$i12 = 0, $$pre2$i$i = 0, $$pre2$i$i$1 = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__begin$sroa$0$021 = 0, $lpad$phi$i23$index = 0, $lpad$phi$i23$index3 = 0, $lpad$phi$index = 0;
 var $lpad$phi$index9 = 0, $p = 0, $vargs = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $p = sp + 40|0;
 $vargs = sp + 24|0;
 $0 = sp + 16|0;
 $1 = sp;
 $2 = ((($this)) + 44|0); //@line 118 "modules/thread/LuaThread.cpp"
 $3 = HEAP8[$2>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $4 = $3 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $5 = ($4<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($5) {
  $8 = $3&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $9 = $8 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $11 = $9;
 } else {
  $6 = ((($this)) + 48|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $7 = HEAP32[$6>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $11 = $7;
 }
 $10 = ($11|0)==(0); //@line 1445 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($10) {
  STACKTOP = sp;return; //@line 141 "modules/thread/LuaThread.cpp"
 }
 $12 = HEAP32[(625612)>>2]|0; //@line 98 "./common/Module.h"
 $13 = ($12|0)==(0|0); //@line 122 "modules/thread/LuaThread.cpp"
 if ($13) {
  STACKTOP = sp;return; //@line 141 "modules/thread/LuaThread.cpp"
 }
 HEAP32[$p>>2] = 53; //@line 126 "modules/thread/LuaThread.cpp"
 $14 = ((($p)) + 4|0); //@line 127 "modules/thread/LuaThread.cpp"
 HEAP32[$14>>2] = $this; //@line 127 "modules/thread/LuaThread.cpp"
 __THREW__ = 0;
 $15 = (invoke_ii(3348,32)|0); //@line 130 "modules/thread/LuaThread.cpp"
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 do {
  if ($17) {
   $$0 = $0;
   label = 37;
  } else {
   __THREW__ = 0;
   invoke_viii(3949,($15|0),53,($p|0)); //@line 130 "modules/thread/LuaThread.cpp"
   $18 = __THREW__; __THREW__ = 0;
   $19 = $18&1;
   if ($19) {
    $110 = ___cxa_find_matching_catch_2()|0;
    $111 = tempRet0;
    __ZdlPv($15); //@line 130 "modules/thread/LuaThread.cpp"
    $$01 = $110;$$03 = $111;$$1 = $0;
    break;
   }
   HEAP32[$0>>2] = $15; //@line 93 "./common/Object.h"
   $20 = HEAP32[$15>>2]|0; //@line 95 "./common/Object.h"
   $21 = ((($20)) + 8|0); //@line 95 "./common/Object.h"
   $22 = HEAP32[$21>>2]|0; //@line 95 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($22|0,($15|0)); //@line 95 "./common/Object.h"
   $23 = __THREW__; __THREW__ = 0;
   $24 = $23&1;
   if ($24) {
    $$0 = $0;
    label = 37;
   } else {
    $25 = ((($0)) + 4|0); //@line 129 "modules/thread/LuaThread.cpp"
    __THREW__ = 0;
    $26 = (invoke_ii(3348,32)|0); //@line 131 "modules/thread/LuaThread.cpp"
    $27 = __THREW__; __THREW__ = 0;
    $28 = $27&1;
    if ($28) {
     $$0 = $25;
     label = 37;
    } else {
     $29 = HEAP8[$2>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $30 = $29 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     $31 = ($30<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
     if ($31) {
      $36 = ((($2)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $37 = $29&255; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $38 = $37 >>> 1; //@line 1716 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $39 = $36;$40 = $38;
     } else {
      $32 = ((($this)) + 52|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $33 = HEAP32[$32>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $34 = ((($this)) + 48|0); //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $35 = HEAP32[$34>>2]|0; //@line 1726 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
      $39 = $33;$40 = $35;
     }
     __THREW__ = 0;
     invoke_viii(3940,($26|0),($39|0),($40|0)); //@line 131 "modules/thread/LuaThread.cpp"
     $41 = __THREW__; __THREW__ = 0;
     $42 = $41&1;
     if ($42) {
      $112 = ___cxa_find_matching_catch_2()|0;
      $113 = tempRet0;
      __ZdlPv($26); //@line 131 "modules/thread/LuaThread.cpp"
      $$01 = $112;$$03 = $113;$$1 = $25;
      break;
     }
     HEAP32[$25>>2] = $26; //@line 93 "./common/Object.h"
     $43 = HEAP32[$26>>2]|0; //@line 95 "./common/Object.h"
     $44 = ((($43)) + 8|0); //@line 95 "./common/Object.h"
     $45 = HEAP32[$44>>2]|0; //@line 95 "./common/Object.h"
     __THREW__ = 0;
     invoke_vi($45|0,($26|0)); //@line 95 "./common/Object.h"
     $46 = __THREW__; __THREW__ = 0;
     $47 = $46&1;
     if ($47) {
      $$0 = $25;
      label = 37;
     } else {
      HEAP32[$vargs>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $48 = ((($vargs)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$48>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $49 = ((($vargs)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      HEAP32[$49>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
      __THREW__ = 0;
      $50 = (invoke_ii(3348,8)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
      $51 = __THREW__; __THREW__ = 0;
      $52 = $51&1;
      if ($52) {
       $67 = ___cxa_find_matching_catch_2()|0;
       $68 = tempRet0;
       $lpad$phi$i23$index = $67;$lpad$phi$i23$index3 = $68;
      } else {
       HEAP32[$48>>2] = $50; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$vargs>>2] = $50; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $53 = ((($50)) + 8|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $54 = ((($vargs)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       HEAP32[$54>>2] = $53; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $55 = HEAP32[$0>>2]|0; //@line 139 "./common/Object.h"
       HEAP32[$50>>2] = $55; //@line 99 "./common/Object.h"
       $56 = ($55|0)==(0|0); //@line 101 "./common/Object.h"
       if ($56) {
        $63 = $50;
        label = 19;
       } else {
        $57 = HEAP32[$55>>2]|0; //@line 101 "./common/Object.h"
        $58 = ((($57)) + 8|0); //@line 101 "./common/Object.h"
        $59 = HEAP32[$58>>2]|0; //@line 101 "./common/Object.h"
        __THREW__ = 0;
        invoke_vi($59|0,($55|0)); //@line 101 "./common/Object.h"
        $60 = __THREW__; __THREW__ = 0;
        $61 = $60&1;
        if (!($61)) {
         $$pre2$i$i = HEAP32[$48>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $63 = $$pre2$i$i;
         label = 19;
        }
       }
       do {
        if ((label|0) == 19) {
         $62 = ((($63)) + 4|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$48>>2] = $62; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $64 = ((($0)) + 4|0); //@line 139 "./common/Object.h"
         $65 = HEAP32[$64>>2]|0; //@line 139 "./common/Object.h"
         HEAP32[$62>>2] = $65; //@line 99 "./common/Object.h"
         $66 = ($65|0)==(0|0); //@line 101 "./common/Object.h"
         if ($66) {
          $214 = $62;
         } else {
          $208 = HEAP32[$65>>2]|0; //@line 101 "./common/Object.h"
          $209 = ((($208)) + 8|0); //@line 101 "./common/Object.h"
          $210 = HEAP32[$209>>2]|0; //@line 101 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($210|0,($65|0)); //@line 101 "./common/Object.h"
          $211 = __THREW__; __THREW__ = 0;
          $212 = $211&1;
          if ($212) {
           break;
          }
          $$pre2$i$i$1 = HEAP32[$48>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $214 = $$pre2$i$i$1;
         }
         $213 = ((($214)) + 4|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         HEAP32[$48>>2] = $213; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $215 = ((($0)) + 4|0); //@line 106 "./common/Object.h"
         $89 = HEAP32[$215>>2]|0; //@line 106 "./common/Object.h"
         $216 = ($89|0)==(0|0); //@line 106 "./common/Object.h"
         if (!($216)) {
          $88 = HEAP32[$89>>2]|0; //@line 106 "./common/Object.h"
          $90 = ((($88)) + 12|0); //@line 106 "./common/Object.h"
          $91 = HEAP32[$90>>2]|0; //@line 106 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($91|0,($89|0)); //@line 106 "./common/Object.h"
          $92 = __THREW__; __THREW__ = 0;
          $93 = $92&1;
          if ($93) {
           $94 = ___cxa_find_matching_catch_3(0|0)|0;
           $95 = tempRet0;
           ___clang_call_terminate($94); //@line 106 "./common/Object.h"
           // unreachable; //@line 106 "./common/Object.h"
          }
         }
         $96 = HEAP32[$0>>2]|0; //@line 106 "./common/Object.h"
         $97 = ($96|0)==(0|0); //@line 106 "./common/Object.h"
         if (!($97)) {
          $196 = HEAP32[$96>>2]|0; //@line 106 "./common/Object.h"
          $197 = ((($196)) + 12|0); //@line 106 "./common/Object.h"
          $198 = HEAP32[$197>>2]|0; //@line 106 "./common/Object.h"
          __THREW__ = 0;
          invoke_vi($198|0,($96|0)); //@line 106 "./common/Object.h"
          $199 = __THREW__; __THREW__ = 0;
          $200 = $199&1;
          if ($200) {
           $94 = ___cxa_find_matching_catch_3(0|0)|0;
           $95 = tempRet0;
           ___clang_call_terminate($94); //@line 106 "./common/Object.h"
           // unreachable; //@line 106 "./common/Object.h"
          }
         }
         __THREW__ = 0;
         $100 = (invoke_ii(3348,32)|0); //@line 134 "modules/thread/LuaThread.cpp"
         $201 = __THREW__; __THREW__ = 0;
         $202 = $201&1;
         L46: do {
          if ($202) {
           label = 50;
          } else {
           __THREW__ = 0;
           invoke_viii(3460,($1|0),(445461|0),11); //@line 2001 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
           $98 = __THREW__; __THREW__ = 0;
           $99 = $98&1;
           do {
            if ($99) {
             $143 = ___cxa_find_matching_catch_2()|0;
             $144 = tempRet0;
             $$12 = $143;$$14 = $144;
            } else {
             __THREW__ = 0;
             invoke_viii(3508,($100|0),($1|0),($vargs|0)); //@line 134 "modules/thread/LuaThread.cpp"
             $101 = __THREW__; __THREW__ = 0;
             $102 = $101&1;
             if ($102) {
              $145 = ___cxa_find_matching_catch_2()|0;
              $146 = tempRet0;
              __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 134 "modules/thread/LuaThread.cpp"
              $$12 = $145;$$14 = $146;
              break;
             }
             __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1); //@line 134 "modules/thread/LuaThread.cpp"
             $103 = HEAP32[$vargs>>2]|0; //@line 1472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $104 = HEAP32[$48>>2]|0; //@line 1488 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $105 = ($103|0)==($104|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
             L54: do {
              if (!($105)) {
               $__begin$sroa$0$021 = $103;
               while(1) {
                $147 = HEAP32[$__begin$sroa$0$021>>2]|0; //@line 117 "./common/Object.h"
                $148 = HEAP32[$147>>2]|0; //@line 137 "modules/thread/LuaThread.cpp"
                $149 = ((($148)) + 12|0); //@line 137 "modules/thread/LuaThread.cpp"
                $150 = HEAP32[$149>>2]|0; //@line 137 "modules/thread/LuaThread.cpp"
                __THREW__ = 0;
                invoke_vi($150|0,($147|0)); //@line 137 "modules/thread/LuaThread.cpp"
                $151 = __THREW__; __THREW__ = 0;
                $152 = $151&1;
                if ($152) {
                 break;
                }
                $153 = ((($__begin$sroa$0$021)) + 4|0); //@line 1179 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
                $154 = ($153|0)==($104|0); //@line 1347 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
                if ($154) {
                 break L54;
                } else {
                 $__begin$sroa$0$021 = $153;
                }
               }
               $139 = ___cxa_find_matching_catch_2()|0;
               $140 = tempRet0;
               $lpad$phi$index = $139;$lpad$phi$index9 = $140;
               label = 51;
               break L46;
              }
             } while(0);
             __THREW__ = 0;
             invoke_vii(3509,($12|0),($100|0)); //@line 139 "modules/thread/LuaThread.cpp"
             $106 = __THREW__; __THREW__ = 0;
             $107 = $106&1;
             if ($107) {
              label = 50;
              break L46;
             }
             $155 = HEAP32[$100>>2]|0; //@line 140 "modules/thread/LuaThread.cpp"
             $156 = ((($155)) + 12|0); //@line 140 "modules/thread/LuaThread.cpp"
             $157 = HEAP32[$156>>2]|0; //@line 140 "modules/thread/LuaThread.cpp"
             __THREW__ = 0;
             invoke_vi($157|0,($100|0)); //@line 140 "modules/thread/LuaThread.cpp"
             $158 = __THREW__; __THREW__ = 0;
             $159 = $158&1;
             if ($159) {
              label = 50;
              break L46;
             }
             $160 = HEAP32[$vargs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $161 = ($160|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             if (!($161)) {
              $162 = HEAP32[$48>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              $163 = ($162|0)==($160|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
              do {
               if ($163) {
                $177 = $160;
               } else {
                $165 = $162;
                while(1) {
                 $164 = ((($165)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
                 HEAP32[$48>>2] = $164; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
                 $166 = HEAP32[$164>>2]|0; //@line 106 "./common/Object.h"
                 $167 = ($166|0)==(0|0); //@line 106 "./common/Object.h"
                 if ($167) {
                  $169 = $164;
                 } else {
                  $170 = HEAP32[$166>>2]|0; //@line 106 "./common/Object.h"
                  $171 = ((($170)) + 12|0); //@line 106 "./common/Object.h"
                  $172 = HEAP32[$171>>2]|0; //@line 106 "./common/Object.h"
                  __THREW__ = 0;
                  invoke_vi($172|0,($166|0)); //@line 106 "./common/Object.h"
                  $173 = __THREW__; __THREW__ = 0;
                  $174 = $173&1;
                  if ($174) {
                   label = 64;
                   break;
                  }
                  $$pre$i$i$i9 = HEAP32[$48>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
                  $169 = $$pre$i$i$i9;
                 }
                 $168 = ($169|0)==($160|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
                 if ($168) {
                  label = 65;
                  break;
                 } else {
                  $165 = $169;
                 }
                }
                if ((label|0) == 64) {
                 $175 = ___cxa_find_matching_catch_3(0|0)|0;
                 $176 = tempRet0;
                 ___clang_call_terminate($175); //@line 106 "./common/Object.h"
                 // unreachable; //@line 106 "./common/Object.h"
                }
                else if ((label|0) == 65) {
                 $$pre$i12 = HEAP32[$vargs>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
                 $177 = $$pre$i12;
                 break;
                }
               }
              } while(0);
              __ZdlPv($177); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
             }
             STACKTOP = sp;return; //@line 141 "modules/thread/LuaThread.cpp"
            }
           } while(0);
           __ZdlPv($100); //@line 134 "modules/thread/LuaThread.cpp"
           $$2 = $$12;$$25 = $$14;
          }
         } while(0);
         if ((label|0) == 50) {
          $141 = ___cxa_find_matching_catch_2()|0;
          $142 = tempRet0;
          $lpad$phi$index = $141;$lpad$phi$index9 = $142;
          label = 51;
         }
         if ((label|0) == 51) {
          $$2 = $lpad$phi$index;$$25 = $lpad$phi$index9;
         }
         $178 = HEAP32[$vargs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $179 = ($178|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($179) {
          $$3 = $$2;$$36 = $$25;
          ___resumeException($$3|0);
          // unreachable;
         }
         $180 = HEAP32[$48>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $181 = ($180|0)==($178|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         do {
          if ($181) {
           $195 = $178;
          } else {
           $183 = $180;
           while(1) {
            $182 = ((($183)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            HEAP32[$48>>2] = $182; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $184 = HEAP32[$182>>2]|0; //@line 106 "./common/Object.h"
            $185 = ($184|0)==(0|0); //@line 106 "./common/Object.h"
            if ($185) {
             $187 = $182;
            } else {
             $188 = HEAP32[$184>>2]|0; //@line 106 "./common/Object.h"
             $189 = ((($188)) + 12|0); //@line 106 "./common/Object.h"
             $190 = HEAP32[$189>>2]|0; //@line 106 "./common/Object.h"
             __THREW__ = 0;
             invoke_vi($190|0,($184|0)); //@line 106 "./common/Object.h"
             $191 = __THREW__; __THREW__ = 0;
             $192 = $191&1;
             if ($192) {
              label = 75;
              break;
             }
             $$pre$i$i$i = HEAP32[$48>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
             $187 = $$pre$i$i$i;
            }
            $186 = ($187|0)==($178|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            if ($186) {
             label = 76;
             break;
            } else {
             $183 = $187;
            }
           }
           if ((label|0) == 75) {
            $193 = ___cxa_find_matching_catch_3(0|0)|0;
            $194 = tempRet0;
            ___clang_call_terminate($193); //@line 106 "./common/Object.h"
            // unreachable; //@line 106 "./common/Object.h"
           }
           else if ((label|0) == 76) {
            $$pre$i = HEAP32[$vargs>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $195 = $$pre$i;
            break;
           }
          }
         } while(0);
         __ZdlPv($195); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
         $$3 = $$2;$$36 = $$25;
         ___resumeException($$3|0);
         // unreachable;
        }
       } while(0);
       $69 = ___cxa_find_matching_catch_2()|0;
       $70 = tempRet0;
       $$pre = HEAP32[$vargs>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $71 = ($$pre|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       if ($71) {
        $lpad$phi$i23$index = $69;$lpad$phi$i23$index3 = $70;
       } else {
        $72 = HEAP32[$48>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        $73 = ($72|0)==($$pre|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
        do {
         if ($73) {
          $87 = $$pre;
         } else {
          $75 = $72;
          while(1) {
           $74 = ((($75)) + -4|0); //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           HEAP32[$48>>2] = $74; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $76 = HEAP32[$74>>2]|0; //@line 106 "./common/Object.h"
           $77 = ($76|0)==(0|0); //@line 106 "./common/Object.h"
           if ($77) {
            $79 = $74;
           } else {
            $80 = HEAP32[$76>>2]|0; //@line 106 "./common/Object.h"
            $81 = ((($80)) + 12|0); //@line 106 "./common/Object.h"
            $82 = HEAP32[$81>>2]|0; //@line 106 "./common/Object.h"
            __THREW__ = 0;
            invoke_vi($82|0,($76|0)); //@line 106 "./common/Object.h"
            $83 = __THREW__; __THREW__ = 0;
            $84 = $83&1;
            if ($84) {
             label = 27;
             break;
            }
            $$pre$i$i$i$i = HEAP32[$48>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
            $79 = $$pre$i$i$i$i;
           }
           $78 = ($79|0)==($$pre|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           if ($78) {
            label = 28;
            break;
           } else {
            $75 = $79;
           }
          }
          if ((label|0) == 27) {
           $85 = ___cxa_find_matching_catch_3(0|0)|0;
           $86 = tempRet0;
           ___clang_call_terminate($85); //@line 106 "./common/Object.h"
           // unreachable; //@line 106 "./common/Object.h"
          }
          else if ((label|0) == 28) {
           $$pre$i$i = HEAP32[$vargs>>2]|0; //@line 472 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
           $87 = $$pre$i$i;
           break;
          }
         }
        } while(0);
        __ZdlPv($87); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
        $lpad$phi$i23$index = $69;$lpad$phi$i23$index3 = $70;
       }
      }
      $127 = ((($0)) + 4|0); //@line 106 "./common/Object.h"
      $128 = HEAP32[$127>>2]|0; //@line 106 "./common/Object.h"
      $129 = ($128|0)==(0|0); //@line 106 "./common/Object.h"
      if (!($129)) {
       $130 = HEAP32[$128>>2]|0; //@line 106 "./common/Object.h"
       $131 = ((($130)) + 12|0); //@line 106 "./common/Object.h"
       $132 = HEAP32[$131>>2]|0; //@line 106 "./common/Object.h"
       __THREW__ = 0;
       invoke_vi($132|0,($128|0)); //@line 106 "./common/Object.h"
       $133 = __THREW__; __THREW__ = 0;
       $134 = $133&1;
       if ($134) {
        $135 = ___cxa_find_matching_catch_3(0|0)|0;
        $136 = tempRet0;
        ___clang_call_terminate($135); //@line 106 "./common/Object.h"
        // unreachable; //@line 106 "./common/Object.h"
       }
      }
      $137 = HEAP32[$0>>2]|0; //@line 106 "./common/Object.h"
      $138 = ($137|0)==(0|0); //@line 106 "./common/Object.h"
      if ($138) {
       $$3 = $lpad$phi$i23$index;$$36 = $lpad$phi$i23$index3;
       ___resumeException($$3|0);
       // unreachable;
      }
      $203 = HEAP32[$137>>2]|0; //@line 106 "./common/Object.h"
      $204 = ((($203)) + 12|0); //@line 106 "./common/Object.h"
      $205 = HEAP32[$204>>2]|0; //@line 106 "./common/Object.h"
      __THREW__ = 0;
      invoke_vi($205|0,($137|0)); //@line 106 "./common/Object.h"
      $206 = __THREW__; __THREW__ = 0;
      $207 = $206&1;
      if ($207) {
       $135 = ___cxa_find_matching_catch_3(0|0)|0;
       $136 = tempRet0;
       ___clang_call_terminate($135); //@line 106 "./common/Object.h"
       // unreachable; //@line 106 "./common/Object.h"
      } else {
       $$3 = $lpad$phi$i23$index;$$36 = $lpad$phi$i23$index3;
       ___resumeException($$3|0);
       // unreachable;
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 37) {
  $108 = ___cxa_find_matching_catch_2()|0;
  $109 = tempRet0;
  $$01 = $108;$$03 = $109;$$1 = $$0;
 }
 $114 = ($0|0)==($$1|0); //@line 129 "modules/thread/LuaThread.cpp"
 if ($114) {
  $$3 = $$01;$$36 = $$03;
  ___resumeException($$3|0);
  // unreachable;
 } else {
  $116 = $$1;
 }
 while(1) {
  $115 = ((($116)) + -4|0); //@line 129 "modules/thread/LuaThread.cpp"
  $117 = HEAP32[$115>>2]|0; //@line 106 "./common/Object.h"
  $118 = ($117|0)==(0|0); //@line 106 "./common/Object.h"
  if (!($118)) {
   $119 = HEAP32[$117>>2]|0; //@line 106 "./common/Object.h"
   $120 = ((($119)) + 12|0); //@line 106 "./common/Object.h"
   $121 = HEAP32[$120>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($121|0,($117|0)); //@line 106 "./common/Object.h"
   $122 = __THREW__; __THREW__ = 0;
   $123 = $122&1;
   if ($123) {
    label = 43;
    break;
   }
  }
  $126 = ($115|0)==($0|0); //@line 129 "modules/thread/LuaThread.cpp"
  if ($126) {
   $$3 = $$01;$$36 = $$03;
   label = 78;
   break;
  } else {
   $116 = $115;
  }
 }
 if ((label|0) == 43) {
  $124 = ___cxa_find_matching_catch_3(0|0)|0;
  $125 = tempRet0;
  ___clang_call_terminate($124); //@line 106 "./common/Object.h"
  // unreachable; //@line 106 "./common/Object.h"
 }
 else if ((label|0) == 78) {
  ___resumeException($$3|0);
  // unreachable;
 }
}
function __ZThn8_N4love6thread9LuaThreadD1Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 42 "modules/thread/LuaThread.cpp"
 __ZN4love6thread9LuaThreadD2Ev($0); //@line 42 "modules/thread/LuaThread.cpp"
 return; //@line 42 "modules/thread/LuaThread.cpp"
}
function __ZThn8_N4love6thread9LuaThreadD0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 42 "modules/thread/LuaThread.cpp"
 __ZN4love6thread9LuaThreadD2Ev($0); //@line 43 "modules/thread/LuaThread.cpp"
 __ZdlPv($0); //@line 43 "modules/thread/LuaThread.cpp"
 return; //@line 42 "modules/thread/LuaThread.cpp"
}
function __ZThn8_N4love6thread9LuaThread14threadFunctionEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0); //@line 50 "modules/thread/LuaThread.cpp"
 __ZN4love6thread9LuaThread14threadFunctionEv($0); //@line 50 "modules/thread/LuaThread.cpp"
 return; //@line 50 "modules/thread/LuaThread.cpp"
}
function __ZN4love6thread9LuaThreadC2ERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEEPNS_4DataE($this,$name,$code) {
 $this = $this|0;
 $name = $name|0;
 $code = $code|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 33 "modules/thread/LuaThread.cpp"
 $0 = ((($this)) + 8|0); //@line 38 "modules/thread/LuaThread.cpp"
 __THREW__ = 0;
 invoke_vi(3917,($0|0)); //@line 33 "modules/thread/LuaThread.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  $$2 = $18;$$23 = $17;
  ___resumeException($$23|0);
  // unreachable;
 }
 HEAP32[$this>>2] = (33076); //@line 38 "modules/thread/LuaThread.cpp"
 HEAP32[$0>>2] = (33104); //@line 38 "modules/thread/LuaThread.cpp"
 $3 = ((($this)) + 28|0); //@line 93 "./common/Object.h"
 HEAP32[$3>>2] = $code; //@line 93 "./common/Object.h"
 $4 = ($code|0)==(0|0); //@line 95 "./common/Object.h"
 if ($4) {
  label = 4;
 } else {
  $5 = HEAP32[$code>>2]|0; //@line 95 "./common/Object.h"
  $6 = ((($5)) + 8|0); //@line 95 "./common/Object.h"
  $7 = HEAP32[$6>>2]|0; //@line 95 "./common/Object.h"
  __THREW__ = 0;
  invoke_vi($7|0,($code|0)); //@line 95 "./common/Object.h"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if ($9) {
   $19 = ___cxa_find_matching_catch_2()|0;
   $20 = tempRet0;
   $$1 = $20;$$12 = $19;
  } else {
   label = 4;
  }
 }
 if ((label|0) == 4) {
  $10 = ((($this)) + 32|0); //@line 35 "modules/thread/LuaThread.cpp"
  __THREW__ = 0;
  invoke_vii(3351,($10|0),($name|0)); //@line 35 "modules/thread/LuaThread.cpp"
  $11 = __THREW__; __THREW__ = 0;
  $12 = $11&1;
  do {
   if ($12) {
    $21 = ___cxa_find_matching_catch_2()|0;
    $22 = tempRet0;
    $$0 = $22;$$01 = $21;
   } else {
    $13 = ((($this)) + 44|0); //@line 33 "modules/thread/LuaThread.cpp"
    $14 = ((($this)) + 16|0); //@line 39 "modules/thread/LuaThread.cpp"
    ;HEAP32[$13>>2]=0|0;HEAP32[$13+4>>2]=0|0;HEAP32[$13+8>>2]=0|0;HEAP32[$13+12>>2]=0|0;HEAP32[$13+16>>2]=0|0; //@line 36 "modules/thread/LuaThread.cpp"
    __THREW__ = 0;
    (invoke_iii(3578,($14|0),($name|0))|0); //@line 39 "modules/thread/LuaThread.cpp"
    $15 = __THREW__; __THREW__ = 0;
    $16 = $15&1;
    if ($16) {
     $23 = ___cxa_find_matching_catch_2()|0;
     $24 = tempRet0;
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13); //@line 40 "modules/thread/LuaThread.cpp"
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10); //@line 40 "modules/thread/LuaThread.cpp"
     $$0 = $24;$$01 = $23;
     break;
    } else {
     return; //@line 40 "modules/thread/LuaThread.cpp"
    }
   }
  } while(0);
  $25 = HEAP32[$3>>2]|0; //@line 106 "./common/Object.h"
  $26 = ($25|0)==(0|0); //@line 106 "./common/Object.h"
  if ($26) {
   $$1 = $$0;$$12 = $$01;
  } else {
   $27 = HEAP32[$25>>2]|0; //@line 106 "./common/Object.h"
   $28 = ((($27)) + 12|0); //@line 106 "./common/Object.h"
   $29 = HEAP32[$28>>2]|0; //@line 106 "./common/Object.h"
   __THREW__ = 0;
   invoke_vi($29|0,($25|0)); //@line 106 "./common/Object.h"
   $30 = __THREW__; __THREW__ = 0;
   $31 = $30&1;
   if ($31) {
    $32 = ___cxa_find_matching_catch_3(0|0)|0;
    $33 = tempRet0;
    ___clang_call_terminate($32); //@line 106 "./common/Object.h"
    // unreachable; //@line 106 "./common/Object.h"
   } else {
    $$1 = $$0;$$12 = $$01;
   }
  }
 }
 __ZN4love6thread10ThreadableD2Ev($0); //@line 40 "modules/thread/LuaThread.cpp"
 $$2 = $$1;$$23 = $$12;
 ___resumeException($$23|0);
 // unreachable;
}
function __ZN4love6thread9LuaThread5startEPPNS_7VariantEi($this,$args,$nargs) {
 $this = $this|0;
 $args = $args|0;
 $nargs = $nargs|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 60|0); //@line 102 "modules/thread/LuaThread.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 102 "modules/thread/LuaThread.cpp"
 $2 = ($1|0)>(0); //@line 102 "modules/thread/LuaThread.cpp"
 $3 = ((($this)) + 56|0); //@line 103 "modules/thread/LuaThread.cpp"
 if ($2) {
  $i$02 = 0;
  while(1) {
   $6 = HEAP32[$3>>2]|0; //@line 103 "modules/thread/LuaThread.cpp"
   $7 = (($6) + ($i$02<<2)|0); //@line 103 "modules/thread/LuaThread.cpp"
   $8 = HEAP32[$7>>2]|0; //@line 103 "modules/thread/LuaThread.cpp"
   $9 = HEAP32[$8>>2]|0; //@line 103 "modules/thread/LuaThread.cpp"
   $10 = ((($9)) + 12|0); //@line 103 "modules/thread/LuaThread.cpp"
   $11 = HEAP32[$10>>2]|0; //@line 103 "modules/thread/LuaThread.cpp"
   FUNCTION_TABLE_vi[$11 & 8191]($8); //@line 103 "modules/thread/LuaThread.cpp"
   $12 = (($i$02) + 1)|0; //@line 102 "modules/thread/LuaThread.cpp"
   $13 = HEAP32[$0>>2]|0; //@line 102 "modules/thread/LuaThread.cpp"
   $14 = ($12|0)<($13|0); //@line 102 "modules/thread/LuaThread.cpp"
   if ($14) {
    $i$02 = $12;
   } else {
    break;
   }
  }
 }
 HEAP32[$3>>2] = $args; //@line 105 "modules/thread/LuaThread.cpp"
 HEAP32[$0>>2] = $nargs; //@line 106 "modules/thread/LuaThread.cpp"
 $4 = ((($this)) + 8|0); //@line 108 "modules/thread/LuaThread.cpp"
 $5 = (__ZN4love6thread10Threadable5startEv($4)|0); //@line 108 "modules/thread/LuaThread.cpp"
 return ($5|0); //@line 108 "modules/thread/LuaThread.cpp"
}
function __ZNK4love6thread9LuaThread8getErrorEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 44|0); //@line 113 "modules/thread/LuaThread.cpp"
 return ($0|0); //@line 113 "modules/thread/LuaThread.cpp"
}
function __ZN12b2BroadPhaseC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN13b2DynamicTreeC2Ev($this); //@line 22 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $0 = ((($this)) + 28|0); //@line 24 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 HEAP32[$0>>2] = 0; //@line 24 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $1 = ((($this)) + 48|0); //@line 26 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 HEAP32[$1>>2] = 16; //@line 26 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $2 = ((($this)) + 52|0); //@line 27 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 HEAP32[$2>>2] = 0; //@line 27 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 __THREW__ = 0;
 $3 = (invoke_ii(3831,128)|0); //@line 28 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  __ZN13b2DynamicTreeD2Ev($this); //@line 33 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  ___resumeException($13|0);
  // unreachable;
 }
 $6 = ((($this)) + 44|0); //@line 28 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 HEAP32[$6>>2] = $3; //@line 28 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $7 = ((($this)) + 36|0); //@line 30 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 HEAP32[$7>>2] = 16; //@line 30 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $8 = ((($this)) + 40|0); //@line 31 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 HEAP32[$8>>2] = 0; //@line 31 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 __THREW__ = 0;
 $9 = (invoke_ii(3831,64)|0); //@line 32 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  __ZN13b2DynamicTreeD2Ev($this); //@line 33 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  ___resumeException($13|0);
  // unreachable;
 } else {
  $12 = ((($this)) + 32|0); //@line 32 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  HEAP32[$12>>2] = $9; //@line 32 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  return; //@line 33 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 }
}
function __ZN12b2BroadPhaseD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 37 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 37 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 __THREW__ = 0;
 invoke_vi(3816,($1|0)); //@line 37 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  $4 = ((($this)) + 44|0); //@line 38 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 38 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  __THREW__ = 0;
  invoke_vi(3816,($5|0)); //@line 38 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if (!($7)) {
   __ZN13b2DynamicTreeD2Ev($this); //@line 39 "libraries/Box2D/Collision/b2BroadPhase.cpp"
   return; //@line 39 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  }
 }
 $8 = ___cxa_find_matching_catch_3(0|0)|0;
 $9 = tempRet0;
 __ZN13b2DynamicTreeD2Ev($this); //@line 39 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 ___clang_call_terminate($8); //@line 39 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 // unreachable; //@line 39 "libraries/Box2D/Collision/b2BroadPhase.cpp"
}
function __ZN12b2BroadPhase11CreateProxyERK6b2AABBPv($this,$aabb,$userData) {
 $this = $this|0;
 $aabb = $aabb|0;
 $userData = $userData|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN13b2DynamicTree11CreateProxyERK6b2AABBPv($this,$aabb,$userData)|0); //@line 43 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $1 = ((($this)) + 28|0); //@line 44 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 44 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $3 = (($2) + 1)|0; //@line 44 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 HEAP32[$1>>2] = $3; //@line 44 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $4 = ((($this)) + 40|0); //@line 72 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 72 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $6 = ((($this)) + 36|0); //@line 72 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 72 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $8 = ($5|0)==($7|0); //@line 72 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $9 = ((($this)) + 32|0); //@line 74 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 if ($8) {
  $10 = HEAP32[$9>>2]|0; //@line 74 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $11 = $5 << 1; //@line 75 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  HEAP32[$6>>2] = $11; //@line 75 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $12 = $5 << 3; //@line 76 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $13 = (__Z7b2Alloci($12)|0); //@line 76 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  HEAP32[$9>>2] = $13; //@line 76 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $14 = HEAP32[$4>>2]|0; //@line 77 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $15 = $14 << 2; //@line 77 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  _memcpy(($13|0),($10|0),($15|0))|0; //@line 77 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  __Z6b2FreePv($10); //@line 78 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $$pre$i = HEAP32[$4>>2]|0; //@line 81 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $18 = $$pre$i;
 } else {
  $18 = $5;
 }
 $16 = HEAP32[$9>>2]|0; //@line 81 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $17 = (($16) + ($18<<2)|0); //@line 81 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 HEAP32[$17>>2] = $0; //@line 81 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $19 = HEAP32[$4>>2]|0; //@line 82 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $20 = (($19) + 1)|0; //@line 82 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 HEAP32[$4>>2] = $20; //@line 82 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 return ($0|0); //@line 46 "libraries/Box2D/Collision/b2BroadPhase.cpp"
}
function __ZN12b2BroadPhase12DestroyProxyEi($this,$proxyId) {
 $this = $this|0;
 $proxyId = $proxyId|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0); //@line 87 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 87 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $2 = ($1|0)>(0); //@line 87 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 if ($2) {
  $3 = ((($this)) + 32|0); //@line 89 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 89 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $14 = $1;$i$01$i = 0;
  while(1) {
   $5 = (($4) + ($i$01$i<<2)|0); //@line 89 "libraries/Box2D/Collision/b2BroadPhase.cpp"
   $6 = HEAP32[$5>>2]|0; //@line 89 "libraries/Box2D/Collision/b2BroadPhase.cpp"
   $7 = ($6|0)==($proxyId|0); //@line 89 "libraries/Box2D/Collision/b2BroadPhase.cpp"
   if ($7) {
    HEAP32[$5>>2] = -1; //@line 91 "libraries/Box2D/Collision/b2BroadPhase.cpp"
    $$pre$i = HEAP32[$0>>2]|0; //@line 87 "libraries/Box2D/Collision/b2BroadPhase.cpp"
    $10 = $$pre$i;
   } else {
    $10 = $14;
   }
   $8 = (($i$01$i) + 1)|0; //@line 87 "libraries/Box2D/Collision/b2BroadPhase.cpp"
   $9 = ($8|0)<($10|0); //@line 87 "libraries/Box2D/Collision/b2BroadPhase.cpp"
   if ($9) {
    $14 = $10;$i$01$i = $8;
   } else {
    break;
   }
  }
 }
 $11 = ((($this)) + 28|0); //@line 52 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 52 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $13 = (($12) + -1)|0; //@line 52 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 HEAP32[$11>>2] = $13; //@line 52 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 __ZN13b2DynamicTree12DestroyProxyEi($this,$proxyId); //@line 53 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 return; //@line 54 "libraries/Box2D/Collision/b2BroadPhase.cpp"
}
function __ZN12b2BroadPhase9MoveProxyEiRK6b2AABBRK6b2Vec2($this,$proxyId,$aabb,$displacement) {
 $this = $this|0;
 $proxyId = $proxyId|0;
 $aabb = $aabb|0;
 $displacement = $displacement|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = (__ZN13b2DynamicTree9MoveProxyEiRK6b2AABBRK6b2Vec2($this,$proxyId,$aabb,$displacement)|0); //@line 58 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 if (!($0)) {
  return; //@line 63 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 }
 $1 = ((($this)) + 40|0); //@line 72 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 72 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $3 = ((($this)) + 36|0); //@line 72 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 72 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $5 = ($2|0)==($4|0); //@line 72 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $6 = ((($this)) + 32|0); //@line 74 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 if ($5) {
  $7 = HEAP32[$6>>2]|0; //@line 74 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $8 = $2 << 1; //@line 75 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  HEAP32[$3>>2] = $8; //@line 75 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $9 = $2 << 3; //@line 76 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $10 = (__Z7b2Alloci($9)|0); //@line 76 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  HEAP32[$6>>2] = $10; //@line 76 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $11 = HEAP32[$1>>2]|0; //@line 77 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $12 = $11 << 2; //@line 77 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  _memcpy(($10|0),($7|0),($12|0))|0; //@line 77 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  __Z6b2FreePv($7); //@line 78 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $$pre$i = HEAP32[$1>>2]|0; //@line 81 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $15 = $$pre$i;
 } else {
  $15 = $2;
 }
 $13 = HEAP32[$6>>2]|0; //@line 81 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $14 = (($13) + ($15<<2)|0); //@line 81 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 HEAP32[$14>>2] = $proxyId; //@line 81 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $16 = HEAP32[$1>>2]|0; //@line 82 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $17 = (($16) + 1)|0; //@line 82 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 HEAP32[$1>>2] = $17; //@line 82 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 return; //@line 63 "libraries/Box2D/Collision/b2BroadPhase.cpp"
}
function __ZN12b2BroadPhase10TouchProxyEi($this,$proxyId) {
 $this = $this|0;
 $proxyId = $proxyId|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0); //@line 72 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 72 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $2 = ((($this)) + 36|0); //@line 72 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 72 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $4 = ($1|0)==($3|0); //@line 72 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $5 = ((($this)) + 32|0); //@line 74 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 if ($4) {
  $6 = HEAP32[$5>>2]|0; //@line 74 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $7 = $1 << 1; //@line 75 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  HEAP32[$2>>2] = $7; //@line 75 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $8 = $1 << 3; //@line 76 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $9 = (__Z7b2Alloci($8)|0); //@line 76 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  HEAP32[$5>>2] = $9; //@line 76 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $10 = HEAP32[$0>>2]|0; //@line 77 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $11 = $10 << 2; //@line 77 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  _memcpy(($9|0),($6|0),($11|0))|0; //@line 77 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  __Z6b2FreePv($6); //@line 78 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $$pre$i = HEAP32[$0>>2]|0; //@line 81 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $14 = $$pre$i;
 } else {
  $14 = $1;
 }
 $12 = HEAP32[$5>>2]|0; //@line 81 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $13 = (($12) + ($14<<2)|0); //@line 81 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 HEAP32[$13>>2] = $proxyId; //@line 81 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $15 = HEAP32[$0>>2]|0; //@line 82 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $16 = (($15) + 1)|0; //@line 82 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 HEAP32[$0>>2] = $16; //@line 82 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 return; //@line 68 "libraries/Box2D/Collision/b2BroadPhase.cpp"
}
function __ZN12b2BroadPhase13QueryCallbackEi($this,$proxyId) {
 $this = $this|0;
 $proxyId = $proxyId|0;
 var $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 56|0); //@line 100 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 100 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $2 = ($1|0)==($proxyId|0); //@line 100 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 if ($2) {
  return 1; //@line 120 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 }
 $3 = ((($this)) + 52|0); //@line 106 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 106 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $5 = ((($this)) + 48|0); //@line 106 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 106 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $7 = ($4|0)==($6|0); //@line 106 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $8 = ((($this)) + 44|0); //@line 108 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 if ($7) {
  $9 = HEAP32[$8>>2]|0; //@line 108 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $10 = $4 << 1; //@line 109 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  HEAP32[$5>>2] = $10; //@line 109 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $11 = $4 << 4; //@line 110 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $12 = (__Z7b2Alloci($11)|0); //@line 110 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  HEAP32[$8>>2] = $12; //@line 110 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $13 = HEAP32[$3>>2]|0; //@line 111 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $14 = $13 << 3; //@line 111 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  _memcpy(($12|0),($9|0),($14|0))|0; //@line 111 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  __Z6b2FreePv($9); //@line 112 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $$pre = HEAP32[$0>>2]|0; //@line 115 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $$pre1 = HEAP32[$3>>2]|0; //@line 115 "libraries/Box2D/Collision/b2BroadPhase.cpp"
  $16 = $$pre;$20 = $$pre1;
 } else {
  $16 = $1;$20 = $4;
 }
 $15 = ($16|0)>($proxyId|0); //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $17 = $15 ? $proxyId : $16; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $18 = HEAP32[$8>>2]|0; //@line 115 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $19 = (($18) + ($20<<3)|0); //@line 115 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 HEAP32[$19>>2] = $17; //@line 115 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $21 = ($16|0)<($proxyId|0); //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $22 = $21 ? $proxyId : $16; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $23 = (((($18) + ($20<<3)|0)) + 4|0); //@line 116 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 HEAP32[$23>>2] = $22; //@line 116 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 $24 = (($20) + 1)|0; //@line 117 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 HEAP32[$3>>2] = $24; //@line 117 "libraries/Box2D/Collision/b2BroadPhase.cpp"
 return 1; //@line 120 "libraries/Box2D/Collision/b2BroadPhase.cpp"
}
function __Z16b2CollideCirclesP10b2ManifoldPK13b2CircleShapeRK11b2TransformS3_S6_($manifold,$circleA,$xfA,$circleB,$xfB) {
 $manifold = $manifold|0;
 $circleA = $circleA|0;
 $xfA = $xfA|0;
 $circleB = $circleB|0;
 $xfB = $xfB|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0.0;
 var $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0.0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0.0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0;
 var $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0.0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0.0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($manifold)) + 60|0); //@line 28 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 HEAP32[$0>>2] = 0; //@line 28 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $1 = ((($circleA)) + 12|0); //@line 30 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $2 = ((($xfA)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$1>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = $3 * $4; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($xfA)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($circleA)) + 16|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = +HEAPF32[$8>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = $7 * $9; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $5 - $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$xfA>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = $12 + $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = $4 * $7; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $15 = $3 * $9; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $16 = $14 + $15; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = ((($xfA)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = +HEAPF32[$17>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = $16 + $18; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = ((($circleB)) + 12|0); //@line 31 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $21 = ((($xfB)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $22 = +HEAPF32[$21>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $23 = +HEAPF32[$20>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $24 = $22 * $23; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $25 = ((($xfB)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $26 = +HEAPF32[$25>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($circleB)) + 16|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $28 = +HEAPF32[$27>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $29 = $26 * $28; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $30 = $24 - $29; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $31 = +HEAPF32[$xfB>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $32 = $31 + $30; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $33 = $23 * $26; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $34 = $22 * $28; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $35 = $33 + $34; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $36 = ((($xfB)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $37 = +HEAPF32[$36>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $38 = $35 + $37; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $39 = $32 - $13; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $40 = $38 - $19; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $41 = $39 * $39; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $42 = $40 * $40; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $43 = $41 + $42; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $44 = ((($circleA)) + 8|0); //@line 35 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $45 = +HEAPF32[$44>>2]; //@line 35 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $46 = ((($circleB)) + 8|0); //@line 35 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $47 = +HEAPF32[$46>>2]; //@line 35 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $48 = $45 + $47; //@line 36 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $49 = $48 * $48; //@line 37 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $50 = $43 > $49; //@line 37 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 if ($50) {
  return; //@line 49 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 }
 $51 = ((($manifold)) + 56|0); //@line 42 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 HEAP32[$51>>2] = 0; //@line 42 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $52 = ((($manifold)) + 48|0); //@line 43 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $53 = $1; //@line 43 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $54 = $53; //@line 43 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $55 = HEAP32[$54>>2]|0; //@line 43 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $56 = (($53) + 4)|0; //@line 43 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $57 = $56; //@line 43 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $58 = HEAP32[$57>>2]|0; //@line 43 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $59 = $52; //@line 43 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $60 = $59; //@line 43 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 HEAP32[$60>>2] = $55; //@line 43 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $61 = (($59) + 4)|0; //@line 43 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $62 = $61; //@line 43 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 HEAP32[$62>>2] = $58; //@line 43 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $63 = ((($manifold)) + 40|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$63>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $64 = ((($manifold)) + 44|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$64>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAP32[$0>>2] = 1; //@line 45 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $65 = $20; //@line 47 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $66 = $65; //@line 47 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $67 = HEAP32[$66>>2]|0; //@line 47 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $68 = (($65) + 4)|0; //@line 47 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $69 = $68; //@line 47 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $70 = HEAP32[$69>>2]|0; //@line 47 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $71 = $manifold; //@line 47 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $72 = $71; //@line 47 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 HEAP32[$72>>2] = $67; //@line 47 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $73 = (($71) + 4)|0; //@line 47 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $74 = $73; //@line 47 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 HEAP32[$74>>2] = $70; //@line 47 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $75 = ((($manifold)) + 16|0); //@line 48 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 HEAP32[$75>>2] = 0; //@line 48 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 return; //@line 49 "libraries/Box2D/Collision/b2CollideCircle.cpp"
}
function __Z25b2CollidePolygonAndCircleP10b2ManifoldPK14b2PolygonShapeRK11b2TransformPK13b2CircleShapeS6_($manifold,$polygonA,$xfA,$circleB,$xfB) {
 $manifold = $manifold|0;
 $polygonA = $polygonA|0;
 $xfA = $xfA|0;
 $circleB = $circleB|0;
 $xfB = $xfB|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0.0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0;
 var $116 = 0.0, $117 = 0.0, $118 = 0, $119 = 0.0, $12 = 0.0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0.0, $128 = 0, $129 = 0.0, $13 = 0.0, $130 = 0.0, $131 = 0.0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0.0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0.0, $146 = 0.0, $147 = 0.0, $148 = 0.0, $149 = 0.0, $15 = 0.0, $150 = 0, $151 = 0.0;
 var $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0.0, $16 = 0.0, $160 = 0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0.0, $180 = 0.0, $181 = 0.0, $182 = 0.0, $183 = 0, $184 = 0.0, $185 = 0.0, $186 = 0, $187 = 0.0, $188 = 0.0;
 var $189 = 0.0, $19 = 0.0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0.0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0.0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $22 = 0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0.0, $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0.0;
 var $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0.0, $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0.0;
 var $49 = 0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0.0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0.0;
 var $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0.0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$038 = 0, $normalIndex$0$lcssa = 0, $normalIndex$040 = 0, $normalIndex$1 = 0;
 var $normalIndex$1$lcssa = 0, $phitmp = 0, $separation$0$lcssa = 0, $separation$039 = 0.0, $separation$1 = 0.0, $separation$1$lcssa = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($manifold)) + 60|0); //@line 56 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 HEAP32[$0>>2] = 0; //@line 56 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $1 = ((($circleB)) + 12|0); //@line 59 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $2 = ((($xfB)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$1>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = $3 * $4; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($xfB)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($circleB)) + 16|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = +HEAPF32[$8>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = $7 * $9; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $5 - $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$xfB>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = $12 + $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = $4 * $7; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $15 = $3 * $9; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $16 = $14 + $15; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = ((($xfB)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = +HEAPF32[$17>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = $16 + $18; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$xfA>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $21 = $13 - $20; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($xfA)) + 4|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $23 = +HEAPF32[$22>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $24 = $19 - $23; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $25 = ((($xfA)) + 12|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $26 = +HEAPF32[$25>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $27 = $21 * $26; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $28 = ((($xfA)) + 8|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $29 = +HEAPF32[$28>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $30 = $24 * $29; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $31 = $27 + $30; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $32 = $21 * $29; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $33 = $26 * $24; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $34 = $33 - $32; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $35 = ((($polygonA)) + 8|0); //@line 65 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $36 = +HEAPF32[$35>>2]; //@line 65 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $37 = ((($circleB)) + 8|0); //@line 65 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $38 = +HEAPF32[$37>>2]; //@line 65 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $39 = $36 + $38; //@line 65 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $40 = ((($polygonA)) + 148|0); //@line 66 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $41 = HEAP32[$40>>2]|0; //@line 66 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $42 = ($41|0)>(0); //@line 70 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 do {
  if ($42) {
   $i$038 = 0;$normalIndex$040 = 0;$separation$039 = -3.4028234663852886E+38;
   while(1) {
    $43 = (((($polygonA)) + 20|0) + ($i$038<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $44 = +HEAPF32[$43>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $45 = $31 - $44; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $46 = (((((($polygonA)) + 20|0) + ($i$038<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $47 = +HEAPF32[$46>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $48 = $34 - $47; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $49 = (((($polygonA)) + 84|0) + ($i$038<<3)|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $50 = +HEAPF32[$49>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $51 = $45 * $50; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $52 = (((((($polygonA)) + 84|0) + ($i$038<<3)|0)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $53 = +HEAPF32[$52>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $54 = $48 * $53; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $55 = $51 + $54; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $56 = $55 > $39; //@line 74 "libraries/Box2D/Collision/b2CollideCircle.cpp"
    if ($56) {
     label = 19;
     break;
    }
    $57 = $55 > $separation$039; //@line 80 "libraries/Box2D/Collision/b2CollideCircle.cpp"
    $separation$1 = $57 ? $55 : $separation$039; //@line 80 "libraries/Box2D/Collision/b2CollideCircle.cpp"
    $normalIndex$1 = $57 ? $i$038 : $normalIndex$040; //@line 80 "libraries/Box2D/Collision/b2CollideCircle.cpp"
    $58 = (($i$038) + 1)|0; //@line 70 "libraries/Box2D/Collision/b2CollideCircle.cpp"
    $59 = ($58|0)<($41|0); //@line 70 "libraries/Box2D/Collision/b2CollideCircle.cpp"
    if ($59) {
     $i$038 = $58;$normalIndex$040 = $normalIndex$1;$separation$039 = $separation$1;
    } else {
     $normalIndex$1$lcssa = $normalIndex$1;$separation$1$lcssa = $separation$1;
     label = 4;
     break;
    }
   }
   if ((label|0) == 4) {
    $phitmp = $separation$1$lcssa < 1.1920928955078125E-7; //@line 89 "libraries/Box2D/Collision/b2CollideCircle.cpp"
    $normalIndex$0$lcssa = $normalIndex$1$lcssa;$separation$0$lcssa = $phitmp;
    break;
   }
   else if ((label|0) == 19) {
    return; //@line 154 "libraries/Box2D/Collision/b2CollideCircle.cpp"
   }
  } else {
   $normalIndex$0$lcssa = 0;$separation$0$lcssa = 1;
  }
 } while(0);
 $60 = (($normalIndex$0$lcssa) + 1)|0; //@line 89 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $61 = ($60|0)<($41|0); //@line 89 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $62 = $61 ? $60 : 0; //@line 89 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $63 = (((($polygonA)) + 20|0) + ($normalIndex$0$lcssa<<3)|0); //@line 90 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $64 = HEAP32[$63>>2]|0; //@line 90 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $65 = (((((($polygonA)) + 20|0) + ($normalIndex$0$lcssa<<3)|0)) + 4|0); //@line 90 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $66 = HEAP32[$65>>2]|0; //@line 90 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $67 = (((($polygonA)) + 20|0) + ($62<<3)|0); //@line 91 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $68 = HEAP32[$67>>2]|0; //@line 91 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $69 = (((((($polygonA)) + 20|0) + ($62<<3)|0)) + 4|0); //@line 91 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $70 = HEAP32[$69>>2]|0; //@line 91 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 if ($separation$0$lcssa) {
  HEAP32[$0>>2] = 1; //@line 96 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $71 = ((($manifold)) + 56|0); //@line 97 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$71>>2] = 1; //@line 97 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $72 = ((($manifold)) + 40|0); //@line 98 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $73 = (((($polygonA)) + 84|0) + ($normalIndex$0$lcssa<<3)|0); //@line 98 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $74 = $73; //@line 98 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $75 = $74; //@line 98 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $76 = HEAP32[$75>>2]|0; //@line 98 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $77 = (($74) + 4)|0; //@line 98 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $78 = $77; //@line 98 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $79 = HEAP32[$78>>2]|0; //@line 98 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $80 = $72; //@line 98 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $81 = $80; //@line 98 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$81>>2] = $76; //@line 98 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $82 = (($80) + 4)|0; //@line 98 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $83 = $82; //@line 98 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$83>>2] = $79; //@line 98 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $84 = (HEAP32[tempDoublePtr>>2]=$64,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $85 = (HEAP32[tempDoublePtr>>2]=$68,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $86 = $84 + $85; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $87 = (HEAP32[tempDoublePtr>>2]=$66,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $88 = (HEAP32[tempDoublePtr>>2]=$70,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $89 = $87 + $88; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $90 = $86 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $91 = $89 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $92 = ((($manifold)) + 48|0); //@line 99 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAPF32[$92>>2] = $90; //@line 99 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $93 = ((($manifold)) + 52|0); //@line 99 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAPF32[$93>>2] = $91; //@line 99 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $94 = $1; //@line 100 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $95 = $94; //@line 100 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $96 = HEAP32[$95>>2]|0; //@line 100 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $97 = (($94) + 4)|0; //@line 100 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $98 = $97; //@line 100 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $99 = HEAP32[$98>>2]|0; //@line 100 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $100 = $manifold; //@line 100 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $101 = $100; //@line 100 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$101>>2] = $96; //@line 100 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $102 = (($100) + 4)|0; //@line 100 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $103 = $102; //@line 100 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$103>>2] = $99; //@line 100 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $104 = ((($manifold)) + 16|0); //@line 101 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$104>>2] = 0; //@line 101 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  return; //@line 154 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 }
 $105 = (HEAP32[tempDoublePtr>>2]=$64,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $106 = $31 - $105; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $107 = (HEAP32[tempDoublePtr>>2]=$66,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $108 = $34 - $107; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $109 = (HEAP32[tempDoublePtr>>2]=$68,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $110 = $109 - $105; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $111 = (HEAP32[tempDoublePtr>>2]=$70,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $112 = $111 - $107; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $113 = $106 * $110; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $114 = $108 * $112; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $115 = $113 + $114; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $116 = $31 - $109; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $117 = $34 - $111; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $118 = !($115 <= 0.0); //@line 108 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 if (!($118)) {
  $119 = $106 * $106; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $120 = $108 * $108; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $121 = $119 + $120; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $122 = $39 * $39; //@line 110 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $123 = $121 > $122; //@line 110 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  if ($123) {
   return; //@line 154 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  }
  HEAP32[$0>>2] = 1; //@line 115 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $124 = ((($manifold)) + 56|0); //@line 116 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$124>>2] = 1; //@line 116 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $125 = ((($manifold)) + 40|0); //@line 117 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAPF32[$125>>2] = $106; //@line 117 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $126 = ((($manifold)) + 44|0); //@line 117 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAPF32[$126>>2] = $108; //@line 117 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $127 = (+Math_sqrt((+$121))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $128 = $127 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
  if (!($128)) {
   $129 = 1.0 / $127; //@line 121 "./libraries/Box2D/Common/b2Math.h"
   $130 = $106 * $129; //@line 122 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$125>>2] = $130; //@line 122 "./libraries/Box2D/Common/b2Math.h"
   $131 = $108 * $129; //@line 123 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$126>>2] = $131; //@line 123 "./libraries/Box2D/Common/b2Math.h"
  }
  $132 = ((($manifold)) + 48|0); //@line 119 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$132>>2] = $64; //@line 119 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $133 = ((($manifold)) + 52|0); //@line 119 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$133>>2] = $66; //@line 119 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $134 = $1; //@line 120 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $135 = $134; //@line 120 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $136 = HEAP32[$135>>2]|0; //@line 120 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $137 = (($134) + 4)|0; //@line 120 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $138 = $137; //@line 120 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $139 = HEAP32[$138>>2]|0; //@line 120 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $140 = $manifold; //@line 120 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $141 = $140; //@line 120 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$141>>2] = $136; //@line 120 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $142 = (($140) + 4)|0; //@line 120 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $143 = $142; //@line 120 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$143>>2] = $139; //@line 120 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $144 = ((($manifold)) + 16|0); //@line 121 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$144>>2] = 0; //@line 121 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  return; //@line 154 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 }
 $145 = $105 - $109; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $146 = $116 * $145; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $147 = $107 - $111; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $148 = $117 * $147; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $149 = $146 + $148; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $150 = !($149 <= 0.0); //@line 123 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 if ($150) {
  $177 = $105 + $109; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $178 = $107 + $111; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $179 = $177 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $180 = $178 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $181 = $31 - $179; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $182 = $34 - $180; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $183 = (((($polygonA)) + 84|0) + ($normalIndex$0$lcssa<<3)|0); //@line 141 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $184 = +HEAPF32[$183>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $185 = $181 * $184; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $186 = (((((($polygonA)) + 84|0) + ($normalIndex$0$lcssa<<3)|0)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $187 = +HEAPF32[$186>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $188 = $182 * $187; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $189 = $185 + $188; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $190 = $189 > $39; //@line 142 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  if ($190) {
   return; //@line 154 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  }
  HEAP32[$0>>2] = 1; //@line 147 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $191 = ((($manifold)) + 56|0); //@line 148 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$191>>2] = 1; //@line 148 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $192 = ((($manifold)) + 40|0); //@line 149 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $193 = $183; //@line 149 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $194 = $193; //@line 149 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $195 = HEAP32[$194>>2]|0; //@line 149 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $196 = (($193) + 4)|0; //@line 149 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $197 = $196; //@line 149 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $198 = HEAP32[$197>>2]|0; //@line 149 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $199 = $192; //@line 149 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $200 = $199; //@line 149 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$200>>2] = $195; //@line 149 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $201 = (($199) + 4)|0; //@line 149 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $202 = $201; //@line 149 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$202>>2] = $198; //@line 149 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $203 = ((($manifold)) + 48|0); //@line 150 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAPF32[$203>>2] = $179; //@line 150 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $204 = ((($manifold)) + 52|0); //@line 150 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAPF32[$204>>2] = $180; //@line 150 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $205 = $1; //@line 151 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $206 = $205; //@line 151 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $207 = HEAP32[$206>>2]|0; //@line 151 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $208 = (($205) + 4)|0; //@line 151 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $209 = $208; //@line 151 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $210 = HEAP32[$209>>2]|0; //@line 151 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $211 = $manifold; //@line 151 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $212 = $211; //@line 151 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$212>>2] = $207; //@line 151 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $213 = (($211) + 4)|0; //@line 151 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $214 = $213; //@line 151 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$214>>2] = $210; //@line 151 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  $215 = ((($manifold)) + 16|0); //@line 152 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  HEAP32[$215>>2] = 0; //@line 152 "libraries/Box2D/Collision/b2CollideCircle.cpp"
  return; //@line 154 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 }
 $151 = $116 * $116; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $152 = $117 * $117; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $153 = $151 + $152; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $154 = $39 * $39; //@line 125 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $155 = $153 > $154; //@line 125 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 if ($155) {
  return; //@line 154 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 }
 HEAP32[$0>>2] = 1; //@line 130 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $156 = ((($manifold)) + 56|0); //@line 131 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 HEAP32[$156>>2] = 1; //@line 131 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $157 = ((($manifold)) + 40|0); //@line 132 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 HEAPF32[$157>>2] = $116; //@line 132 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $158 = ((($manifold)) + 44|0); //@line 132 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 HEAPF32[$158>>2] = $117; //@line 132 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $159 = (+Math_sqrt((+$153))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $160 = $159 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
 if (!($160)) {
  $161 = 1.0 / $159; //@line 121 "./libraries/Box2D/Common/b2Math.h"
  $162 = $116 * $161; //@line 122 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$157>>2] = $162; //@line 122 "./libraries/Box2D/Common/b2Math.h"
  $163 = $117 * $161; //@line 123 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$158>>2] = $163; //@line 123 "./libraries/Box2D/Common/b2Math.h"
 }
 $164 = ((($manifold)) + 48|0); //@line 134 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 HEAP32[$164>>2] = $68; //@line 134 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $165 = ((($manifold)) + 52|0); //@line 134 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 HEAP32[$165>>2] = $70; //@line 134 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $166 = $1; //@line 135 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $167 = $166; //@line 135 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $168 = HEAP32[$167>>2]|0; //@line 135 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $169 = (($166) + 4)|0; //@line 135 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $170 = $169; //@line 135 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $171 = HEAP32[$170>>2]|0; //@line 135 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $172 = $manifold; //@line 135 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $173 = $172; //@line 135 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 HEAP32[$173>>2] = $168; //@line 135 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $174 = (($172) + 4)|0; //@line 135 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $175 = $174; //@line 135 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 HEAP32[$175>>2] = $171; //@line 135 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 $176 = ((($manifold)) + 16|0); //@line 136 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 HEAP32[$176>>2] = 0; //@line 136 "libraries/Box2D/Collision/b2CollideCircle.cpp"
 return; //@line 154 "libraries/Box2D/Collision/b2CollideCircle.cpp"
}
function __Z22b2CollideEdgeAndCircleP10b2ManifoldPK11b2EdgeShapeRK11b2TransformPK13b2CircleShapeS6_($manifold,$edgeA,$xfA,$circleB,$xfB) {
 $manifold = $manifold|0;
 $edgeA = $edgeA|0;
 $xfA = $xfA|0;
 $circleB = $circleB|0;
 $xfB = $xfB|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0.0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0.0;
 var $116 = 0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0.0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0.0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0.0, $140 = 0, $141 = 0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0, $146 = 0.0, $147 = 0.0, $148 = 0.0, $149 = 0.0, $15 = 0.0, $150 = 0.0, $151 = 0.0;
 var $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0.0, $160 = 0.0, $161 = 0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0, $167 = 0.0, $168 = 0.0, $169 = 0.0, $17 = 0;
 var $170 = 0.0, $171 = 0.0, $172 = 0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0.0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0.0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0.0, $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0.0, $30 = 0.0;
 var $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0;
 var $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0, $62 = 0.0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0.0, $67 = 0.0;
 var $68 = 0.0, $69 = 0, $7 = 0.0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0.0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $n$sroa$0$0$in = 0.0, $n$sroa$0$1$in = 0.0, $n$sroa$9$0$in = 0.0, $n$sroa$9$1$in = 0.0, $not$ = 0;
 var $not$27 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($manifold)) + 60|0); //@line 31 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$0>>2] = 0; //@line 31 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $1 = ((($circleB)) + 12|0); //@line 34 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $2 = ((($xfB)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$1>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = $3 * $4; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($xfB)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($circleB)) + 16|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = +HEAPF32[$8>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = $7 * $9; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $5 - $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$xfB>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = $12 + $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = $4 * $7; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $15 = $3 * $9; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $16 = $14 + $15; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = ((($xfB)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = +HEAPF32[$17>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = $16 + $18; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$xfA>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $21 = $13 - $20; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($xfA)) + 4|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $23 = +HEAPF32[$22>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $24 = $19 - $23; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $25 = ((($xfA)) + 12|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $26 = +HEAPF32[$25>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $27 = $21 * $26; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $28 = ((($xfA)) + 8|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $29 = +HEAPF32[$28>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $30 = $24 * $29; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $31 = $27 + $30; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $32 = $21 * $29; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $33 = $26 * $24; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $34 = $33 - $32; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $35 = ((($edgeA)) + 12|0); //@line 36 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $36 = HEAP32[$35>>2]|0; //@line 36 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $37 = ((($edgeA)) + 16|0); //@line 36 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $38 = HEAP32[$37>>2]|0; //@line 36 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $39 = ((($edgeA)) + 20|0); //@line 36 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $40 = HEAP32[$39>>2]|0; //@line 36 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $41 = ((($edgeA)) + 24|0); //@line 36 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $42 = HEAP32[$41>>2]|0; //@line 36 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $43 = (HEAP32[tempDoublePtr>>2]=$40,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $44 = (HEAP32[tempDoublePtr>>2]=$36,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $45 = $43 - $44; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $46 = (HEAP32[tempDoublePtr>>2]=$42,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $47 = (HEAP32[tempDoublePtr>>2]=$38,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $48 = $46 - $47; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $49 = $43 - $31; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $50 = $46 - $34; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $51 = $45 * $49; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $52 = $48 * $50; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $53 = $51 + $52; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $54 = $31 - $44; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $55 = $34 - $47; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $56 = $54 * $45; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $57 = $55 * $48; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $58 = $56 + $57; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $59 = ((($edgeA)) + 8|0); //@line 43 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $60 = +HEAPF32[$59>>2]; //@line 43 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $61 = ((($circleB)) + 8|0); //@line 43 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $62 = +HEAPF32[$61>>2]; //@line 43 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $63 = $60 + $62; //@line 43 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $64 = !($58 <= 0.0); //@line 50 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 if (!($64)) {
  $65 = $54 * $54; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $66 = $55 * $55; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $67 = $65 + $66; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $68 = $63 * $63; //@line 55 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $69 = $67 > $68; //@line 55 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  if ($69) {
   return; //@line 152 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  }
  $70 = ((($edgeA)) + 44|0); //@line 61 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $71 = HEAP8[$70>>0]|0; //@line 61 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $72 = ($71<<24>>24)==(0); //@line 61 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  if (!($72)) {
   $73 = ((($edgeA)) + 28|0); //@line 63 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $74 = +HEAPF32[$73>>2]; //@line 63 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $75 = ((($edgeA)) + 32|0); //@line 63 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $76 = +HEAPF32[$75>>2]; //@line 63 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $77 = $44 - $74; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $78 = $47 - $76; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $79 = $44 - $31; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $80 = $47 - $34; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $81 = $79 * $77; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $82 = $80 * $78; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $83 = $81 + $82; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $not$ = !($83 > 0.0); //@line 71 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   if (!($not$)) {
    return; //@line 152 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   }
  }
  HEAP32[$0>>2] = 1; //@line 77 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $84 = ((($manifold)) + 56|0); //@line 78 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP32[$84>>2] = 0; //@line 78 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $85 = ((($manifold)) + 40|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$85>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $86 = ((($manifold)) + 44|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$86>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $87 = ((($manifold)) + 48|0); //@line 80 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP32[$87>>2] = $36; //@line 80 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $88 = ((($manifold)) + 52|0); //@line 80 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP32[$88>>2] = $38; //@line 80 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $89 = ((($manifold)) + 16|0); //@line 81 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP32[$89>>2] = 0; //@line 81 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP8[$89>>0] = 0; //@line 82 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $90 = ((($89)) + 1|0); //@line 82 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP8[$90>>0] = 0; //@line 82 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $91 = ((($89)) + 2|0); //@line 82 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP8[$91>>0] = 0; //@line 82 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $92 = ((($89)) + 3|0); //@line 82 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP8[$92>>0] = 0; //@line 82 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $93 = $1; //@line 83 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $94 = $93; //@line 83 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $95 = HEAP32[$94>>2]|0; //@line 83 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $96 = (($93) + 4)|0; //@line 83 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $97 = $96; //@line 83 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $98 = HEAP32[$97>>2]|0; //@line 83 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $99 = $manifold; //@line 83 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $100 = $99; //@line 83 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP32[$100>>2] = $95; //@line 83 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $101 = (($99) + 4)|0; //@line 83 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $102 = $101; //@line 83 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP32[$102>>2] = $98; //@line 83 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  return; //@line 152 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 }
 $103 = !($53 <= 0.0); //@line 88 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 if ($103) {
  $142 = $45 * $45; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $143 = $48 * $48; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $144 = $142 + $143; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $145 = $144 > 0.0; //@line 127 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  __Z10loveAssertbPKc($145,445473); //@line 127 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $146 = 1.0 / $144; //@line 128 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $147 = $44 * $53; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $148 = $47 * $53; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $149 = $43 * $58; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $150 = $46 * $58; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $151 = $147 + $149; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $152 = $148 + $150; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $153 = $146 * $151; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $154 = $146 * $152; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $155 = $31 - $153; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $156 = $34 - $154; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $157 = $155 * $155; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $158 = $156 * $156; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $159 = $157 + $158; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $160 = $63 * $63; //@line 131 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $161 = $159 > $160; //@line 131 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  if ($161) {
   return; //@line 152 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  }
  $162 = -$48; //@line 136 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $163 = $54 * $162; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $164 = $55 * $45; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $165 = $164 + $163; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $166 = $165 < 0.0; //@line 137 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $167 = -$45; //@line 139 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $n$sroa$0$0$in = $166 ? $48 : $162; //@line 137 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $n$sroa$9$0$in = $166 ? $167 : $45; //@line 137 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $168 = $n$sroa$0$0$in * $n$sroa$0$0$in; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $169 = $n$sroa$9$0$in * $n$sroa$9$0$in; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $170 = $168 + $169; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $171 = (+Math_sqrt((+$170))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $172 = $171 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
  if ($172) {
   $n$sroa$0$1$in = $n$sroa$0$0$in;$n$sroa$9$1$in = $n$sroa$9$0$in;
  } else {
   $173 = 1.0 / $171; //@line 121 "./libraries/Box2D/Common/b2Math.h"
   $174 = $n$sroa$0$0$in * $173; //@line 122 "./libraries/Box2D/Common/b2Math.h"
   $175 = $n$sroa$9$0$in * $173; //@line 123 "./libraries/Box2D/Common/b2Math.h"
   $n$sroa$0$1$in = $174;$n$sroa$9$1$in = $175;
  }
  HEAP32[$0>>2] = 1; //@line 145 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $176 = ((($manifold)) + 56|0); //@line 146 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP32[$176>>2] = 1; //@line 146 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $177 = ((($manifold)) + 40|0); //@line 147 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAPF32[$177>>2] = $n$sroa$0$1$in; //@line 147 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $178 = ((($manifold)) + 44|0); //@line 147 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAPF32[$178>>2] = $n$sroa$9$1$in; //@line 147 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $179 = ((($manifold)) + 48|0); //@line 148 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP32[$179>>2] = $36; //@line 148 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $180 = ((($manifold)) + 52|0); //@line 148 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP32[$180>>2] = $38; //@line 148 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $181 = ((($manifold)) + 16|0); //@line 149 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP32[$181>>2] = 0; //@line 149 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP8[$181>>0] = 0; //@line 150 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $182 = ((($181)) + 1|0); //@line 150 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP8[$182>>0] = 0; //@line 150 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $183 = ((($181)) + 2|0); //@line 150 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP8[$183>>0] = 1; //@line 150 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $184 = ((($181)) + 3|0); //@line 150 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP8[$184>>0] = 0; //@line 150 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $185 = $1; //@line 151 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $186 = $185; //@line 151 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $187 = HEAP32[$186>>2]|0; //@line 151 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $188 = (($185) + 4)|0; //@line 151 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $189 = $188; //@line 151 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $190 = HEAP32[$189>>2]|0; //@line 151 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $191 = $manifold; //@line 151 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $192 = $191; //@line 151 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP32[$192>>2] = $187; //@line 151 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $193 = (($191) + 4)|0; //@line 151 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $194 = $193; //@line 151 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP32[$194>>2] = $190; //@line 151 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  return; //@line 152 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 }
 $104 = $31 - $43; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $105 = $34 - $46; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $106 = $104 * $104; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $107 = $105 * $105; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $108 = $106 + $107; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $109 = $63 * $63; //@line 93 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $110 = $108 > $109; //@line 93 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 if ($110) {
  return; //@line 152 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 }
 $111 = ((($edgeA)) + 45|0); //@line 99 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $112 = HEAP8[$111>>0]|0; //@line 99 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $113 = ($112<<24>>24)==(0); //@line 99 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 if (!($113)) {
  $114 = ((($edgeA)) + 36|0); //@line 101 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $115 = +HEAPF32[$114>>2]; //@line 101 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $116 = ((($edgeA)) + 40|0); //@line 101 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $117 = +HEAPF32[$116>>2]; //@line 101 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $118 = $115 - $43; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $119 = $117 - $46; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $120 = $104 * $118; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $121 = $105 * $119; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $122 = $120 + $121; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $not$27 = !($122 > 0.0); //@line 109 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  if (!($not$27)) {
   return; //@line 152 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  }
 }
 HEAP32[$0>>2] = 1; //@line 115 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $123 = ((($manifold)) + 56|0); //@line 116 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$123>>2] = 0; //@line 116 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $124 = ((($manifold)) + 40|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$124>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $125 = ((($manifold)) + 44|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$125>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $126 = ((($manifold)) + 48|0); //@line 118 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$126>>2] = $40; //@line 118 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $127 = ((($manifold)) + 52|0); //@line 118 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$127>>2] = $42; //@line 118 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $128 = ((($manifold)) + 16|0); //@line 119 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$128>>2] = 0; //@line 119 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP8[$128>>0] = 1; //@line 120 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $129 = ((($128)) + 1|0); //@line 120 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP8[$129>>0] = 0; //@line 120 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $130 = ((($128)) + 2|0); //@line 120 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP8[$130>>0] = 0; //@line 120 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $131 = ((($128)) + 3|0); //@line 120 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP8[$131>>0] = 0; //@line 120 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $132 = $1; //@line 121 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $133 = $132; //@line 121 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $134 = HEAP32[$133>>2]|0; //@line 121 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $135 = (($132) + 4)|0; //@line 121 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $136 = $135; //@line 121 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $137 = HEAP32[$136>>2]|0; //@line 121 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $138 = $manifold; //@line 121 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $139 = $138; //@line 121 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$139>>2] = $134; //@line 121 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $140 = (($138) + 4)|0; //@line 121 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $141 = $140; //@line 121 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$141>>2] = $137; //@line 121 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 return; //@line 152 "libraries/Box2D/Collision/b2CollideEdge.cpp"
}
function __ZN12b2EPCollider7CollideEP10b2ManifoldPK11b2EdgeShapeRK11b2TransformPK14b2PolygonShapeS7_($this,$manifold,$edgeA,$xfA,$polygonB,$xfB) {
 $this = $this|0;
 $manifold = $manifold|0;
 $edgeA = $edgeA|0;
 $xfA = $xfA|0;
 $polygonB = $polygonB|0;
 $xfB = $xfB|0;
 var $$ = 0, $$38 = 0, $$cast = 0.0, $$in = 0, $$in165 = 0, $$in166 = 0, $$in167 = 0, $$pre = 0, $$pre$phi163Z2D = 0, $$pre$phi164Z2D = 0, $$pre$phiZ2D = 0, $$pre158 = 0.0, $$pre159 = 0.0, $$pre160 = 0.0, $$pre161 = 0.0, $$pre162 = 0, $0 = 0, $1 = 0.0, $10 = 0.0, $100 = 0.0;
 var $1000 = 0, $1001 = 0, $1002 = 0.0, $1003 = 0.0, $1004 = 0, $1005 = 0.0, $1006 = 0.0, $1007 = 0.0, $1008 = 0.0, $1009 = 0.0, $101 = 0, $1010 = 0, $1011 = 0.0, $1012 = 0.0, $1013 = 0.0, $1014 = 0.0, $1015 = 0.0, $1016 = 0.0, $1017 = 0.0, $1018 = 0.0;
 var $1019 = 0.0, $102 = 0.0, $1020 = 0.0, $1021 = 0.0, $1022 = 0.0, $1023 = 0.0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0.0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0;
 var $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0.0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0.0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0;
 var $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $106 = 0.0, $107 = 0.0, $108 = 0, $109 = 0.0, $11 = 0.0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0, $117 = 0.0, $118 = 0.0, $119 = 0, $12 = 0.0;
 var $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0.0, $130 = 0.0, $131 = 0, $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0, $137 = 0, $138 = 0.0;
 var $139 = 0.0, $14 = 0.0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0.0, $148 = 0.0, $149 = 0.0, $15 = 0.0, $150 = 0.0, $151 = 0.0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0.0, $156 = 0.0;
 var $157 = 0.0, $158 = 0, $159 = 0.0, $16 = 0.0, $160 = 0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0.0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0.0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0.0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0.0;
 var $210 = 0.0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0.0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0;
 var $229 = 0, $23 = 0.0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0.0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0;
 var $247 = 0, $248 = 0, $249 = 0, $25 = 0.0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0.0, $26 = 0.0, $260 = 0, $261 = 0.0, $262 = 0, $263 = 0, $264 = 0;
 var $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0.0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0;
 var $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0.0, $30 = 0;
 var $300 = 0, $301 = 0, $302 = 0, $303 = 0.0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0.0, $309 = 0.0, $31 = 0, $310 = 0, $311 = 0.0, $312 = 0.0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0;
 var $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0.0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0;
 var $337 = 0, $338 = 0, $339 = 0, $34 = 0.0, $340 = 0, $341 = 0.0, $342 = 0, $343 = 0.0, $344 = 0, $345 = 0, $346 = 0, $347 = 0.0, $348 = 0.0, $349 = 0, $35 = 0, $350 = 0.0, $351 = 0.0, $352 = 0, $353 = 0, $354 = 0;
 var $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0.0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0.0, $370 = 0, $371 = 0, $372 = 0;
 var $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0.0, $380 = 0, $381 = 0, $382 = 0.0, $383 = 0, $384 = 0, $385 = 0.0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0.0, $390 = 0;
 var $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0.0, $399 = 0, $4 = 0.0, $40 = 0.0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0;
 var $409 = 0, $41 = 0.0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0.0, $420 = 0.0, $421 = 0, $422 = 0, $423 = 0.0, $424 = 0, $425 = 0, $426 = 0;
 var $427 = 0, $428 = 0, $429 = 0, $43 = 0.0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0.0, $438 = 0.0, $439 = 0, $44 = 0, $440 = 0.0, $441 = 0.0, $442 = 0, $443 = 0, $444 = 0;
 var $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0.0, $461 = 0, $462 = 0;
 var $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0.0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0;
 var $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0;
 var $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0.0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0.0, $513 = 0, $514 = 0, $515 = 0.0, $516 = 0.0;
 var $517 = 0, $518 = 0.0, $519 = 0.0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0;
 var $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0.0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0.0;
 var $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0;
 var $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0.0, $581 = 0.0, $582 = 0, $583 = 0.0, $584 = 0, $585 = 0.0, $586 = 0, $587 = 0.0, $588 = 0.0, $589 = 0;
 var $59 = 0, $590 = 0.0, $591 = 0.0, $592 = 0.0, $593 = 0.0, $594 = 0.0, $595 = 0, $596 = 0.0, $597 = 0.0, $598 = 0, $599 = 0.0, $6 = 0.0, $60 = 0, $600 = 0, $601 = 0.0, $602 = 0.0, $603 = 0.0, $604 = 0, $605 = 0.0, $606 = 0.0;
 var $607 = 0.0, $608 = 0.0, $609 = 0.0, $61 = 0, $610 = 0.0, $611 = 0.0, $612 = 0.0, $613 = 0.0, $614 = 0.0, $615 = 0, $616 = 0, $617 = 0.0, $618 = 0, $619 = 0.0, $62 = 0, $620 = 0.0, $621 = 0.0, $622 = 0, $623 = 0.0, $624 = 0.0;
 var $625 = 0.0, $626 = 0.0, $627 = 0.0, $628 = 0.0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0.0, $639 = 0.0, $64 = 0, $640 = 0, $641 = 0.0, $642 = 0.0;
 var $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0.0, $654 = 0, $655 = 0.0, $656 = 0.0, $657 = 0.0, $658 = 0, $659 = 0.0, $66 = 0, $660 = 0.0;
 var $661 = 0.0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0;
 var $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0;
 var $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0.0, $702 = 0.0, $703 = 0, $704 = 0.0, $705 = 0.0, $706 = 0.0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0;
 var $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0;
 var $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0.0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0;
 var $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0.0, $769 = 0.0;
 var $77 = 0, $770 = 0.0, $771 = 0.0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0;
 var $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0.0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0;
 var $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0;
 var $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0;
 var $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0.0, $847 = 0.0, $848 = 0.0, $849 = 0.0, $85 = 0, $850 = 0.0, $851 = 0, $852 = 0, $853 = 0, $854 = 0.0, $855 = 0.0, $856 = 0, $857 = 0, $858 = 0, $859 = 0.0;
 var $86 = 0, $860 = 0, $861 = 0, $862 = 0.0, $863 = 0, $864 = 0.0, $865 = 0.0, $866 = 0, $867 = 0.0, $868 = 0.0, $869 = 0.0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0.0, $876 = 0, $877 = 0;
 var $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0;
 var $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0.0, $90 = 0, $900 = 0, $901 = 0, $902 = 0.0, $903 = 0.0, $904 = 0.0, $905 = 0.0, $906 = 0.0, $907 = 0.0, $908 = 0, $909 = 0.0, $91 = 0, $910 = 0.0, $911 = 0.0, $912 = 0.0;
 var $913 = 0.0, $914 = 0.0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0;
 var $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0.0, $941 = 0.0, $942 = 0.0, $943 = 0.0, $944 = 0, $945 = 0.0, $946 = 0.0, $947 = 0.0, $948 = 0.0, $949 = 0.0;
 var $95 = 0, $950 = 0.0, $951 = 0, $952 = 0.0, $953 = 0.0, $954 = 0.0, $955 = 0.0, $956 = 0.0, $957 = 0.0, $958 = 0.0, $959 = 0.0, $96 = 0, $960 = 0.0, $961 = 0.0, $962 = 0.0, $963 = 0.0, $964 = 0, $965 = 0, $966 = 0, $967 = 0;
 var $968 = 0, $969 = 0.0, $97 = 0, $970 = 0.0, $971 = 0, $972 = 0.0, $973 = 0.0, $974 = 0.0, $975 = 0.0, $976 = 0.0, $977 = 0, $978 = 0.0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0;
 var $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0.0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $bestIndex$0$lcssa = 0, $bestIndex$0127 = 0, $bestIndex$1 = 0, $bestValue$0128 = 0.0, $bestValue$1 = 0.0;
 var $bitcast = 0.0, $bitcast143 = 0.0, $bitcast145 = 0.0, $bitcast149 = 0.0, $bitcast151 = 0.0, $bitcast153 = 0.0, $bitcast155 = 0.0, $bitcast157 = 0.0, $brmerge37$demorgan = 0, $clipPoints1 = 0, $clipPoints2 = 0, $convex1$0$off0 = 0, $edge0$sroa$0$0 = 0.0, $edge0$sroa$7$0 = 0.0, $edge1$sroa$0$0 = 0.0, $edge2$sroa$0$0 = 0.0, $edgeAxis$sroa$9$0 = 0.0, $edgeAxis$sroa$9$1 = 0.0, $edgeAxis$sroa$9$1$lcssa = 0.0, $exitcond137 = 0;
 var $i$01$i = 0, $i$0131 = 0, $i1$0129 = 0, $ie = 0, $offset0$0 = 0.0, $or$cond = 0, $or$cond3 = 0, $phitmp = 0, $pointCount$0$lcssa = 0, $pointCount$1 = 0, $pointCount$1$us = 0, $polygonAxis = 0, $primaryAxis$sroa$0$0173 = 0, $rf = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $polygonAxis = sp + 128|0;
 $ie = sp + 24|0;
 $rf = sp + 72|0;
 $clipPoints1 = sp + 48|0;
 $clipPoints2 = sp;
 $0 = ((($xfA)) + 12|0); //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $1 = +HEAPF32[$0>>2]; //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $2 = ((($xfB)) + 8|0); //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $4 = $1 * $3; //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($xfA)) + 8|0); //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($xfB)) + 12|0); //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $9 = $6 * $8; //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $10 = $4 - $9; //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $11 = $1 * $8; //@line 561 "./libraries/Box2D/Common/b2Math.h"
 $12 = $3 * $6; //@line 561 "./libraries/Box2D/Common/b2Math.h"
 $13 = $12 + $11; //@line 561 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$xfB>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $15 = +HEAPF32[$xfA>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $16 = $14 - $15; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $17 = ((($xfB)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $18 = +HEAPF32[$17>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($xfA)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 - $20; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $22 = $1 * $16; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $23 = $6 * $21; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $24 = $22 + $23; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $25 = $6 * $16; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $26 = $1 * $21; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $27 = $26 - $25; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $28 = ((($this)) + 132|0); //@line 233 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAPF32[$28>>2] = $24; //@line 233 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $29 = ((($this)) + 136|0); //@line 233 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAPF32[$29>>2] = $27; //@line 233 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $30 = ((($this)) + 140|0); //@line 233 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAPF32[$30>>2] = $10; //@line 233 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $31 = ((($this)) + 144|0); //@line 233 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAPF32[$31>>2] = $13; //@line 233 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $32 = ((($polygonB)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $33 = +HEAPF32[$32>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $34 = $13 * $33; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $35 = ((($polygonB)) + 16|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $36 = +HEAPF32[$35>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $37 = $10 * $36; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $38 = $34 - $37; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $39 = $24 + $38; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $40 = $33 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $41 = $13 * $36; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $42 = $40 + $41; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $43 = $42 + $27; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $44 = ((($this)) + 148|0); //@line 235 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAPF32[$44>>2] = $39; //@line 235 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $45 = ((($this)) + 152|0); //@line 235 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAPF32[$45>>2] = $43; //@line 235 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $46 = ((($this)) + 156|0); //@line 237 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $47 = ((($edgeA)) + 28|0); //@line 237 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $48 = $47; //@line 237 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $49 = $48; //@line 237 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $50 = HEAP32[$49>>2]|0; //@line 237 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $51 = (($48) + 4)|0; //@line 237 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $52 = $51; //@line 237 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $53 = HEAP32[$52>>2]|0; //@line 237 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $54 = $46; //@line 237 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $55 = $54; //@line 237 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$55>>2] = $50; //@line 237 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $56 = (($54) + 4)|0; //@line 237 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $57 = $56; //@line 237 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$57>>2] = $53; //@line 237 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $58 = ((($this)) + 164|0); //@line 238 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $59 = ((($edgeA)) + 12|0); //@line 238 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $60 = $59; //@line 238 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $61 = $60; //@line 238 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $62 = HEAP32[$61>>2]|0; //@line 238 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $63 = (($60) + 4)|0; //@line 238 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $64 = $63; //@line 238 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $65 = HEAP32[$64>>2]|0; //@line 238 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $66 = $58; //@line 238 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $67 = $66; //@line 238 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$67>>2] = $62; //@line 238 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $68 = (($66) + 4)|0; //@line 238 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $69 = $68; //@line 238 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$69>>2] = $65; //@line 238 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $70 = ((($this)) + 172|0); //@line 239 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $71 = ((($edgeA)) + 20|0); //@line 239 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $72 = $71; //@line 239 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $73 = $72; //@line 239 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $74 = HEAP32[$73>>2]|0; //@line 239 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $75 = (($72) + 4)|0; //@line 239 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $76 = $75; //@line 239 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $77 = HEAP32[$76>>2]|0; //@line 239 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $78 = $70; //@line 239 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $79 = $78; //@line 239 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$79>>2] = $74; //@line 239 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $80 = (($78) + 4)|0; //@line 239 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $81 = $80; //@line 239 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$81>>2] = $77; //@line 239 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $82 = ((($this)) + 180|0); //@line 240 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $83 = ((($edgeA)) + 36|0); //@line 240 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $84 = $83; //@line 240 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $85 = $84; //@line 240 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $86 = HEAP32[$85>>2]|0; //@line 240 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $87 = (($84) + 4)|0; //@line 240 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $88 = $87; //@line 240 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $89 = HEAP32[$88>>2]|0; //@line 240 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $90 = $82; //@line 240 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $91 = $90; //@line 240 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$91>>2] = $86; //@line 240 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $92 = (($90) + 4)|0; //@line 240 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $93 = $92; //@line 240 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$93>>2] = $89; //@line 240 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $94 = ((($edgeA)) + 44|0); //@line 242 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $95 = HEAP16[$94>>1]|0; //@line 242 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $96 = $95&255; //@line 242 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $97 = ($96<<24>>24)!=(0); //@line 242 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $98 = ($95&65535)>(255); //@line 243 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $bitcast = (HEAP32[tempDoublePtr>>2]=$74,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $bitcast143 = (HEAP32[tempDoublePtr>>2]=$62,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $99 = $bitcast - $bitcast143; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $100 = (HEAP32[tempDoublePtr>>2]=$77,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $101 = ((($this)) + 168|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $102 = (HEAP32[tempDoublePtr>>2]=$65,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $103 = $100 - $102; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $104 = $99 * $99; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $105 = $103 * $103; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $106 = $104 + $105; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $107 = (+Math_sqrt((+$106))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $108 = $107 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
 $109 = (HEAP32[tempDoublePtr>>2]=$50,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $110 = (HEAP32[tempDoublePtr>>2]=$53,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $111 = (HEAP32[tempDoublePtr>>2]=$86,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $112 = (HEAP32[tempDoublePtr>>2]=$89,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 if ($108) {
  $118 = $103;$edge1$sroa$0$0 = $99;
 } else {
  $113 = 1.0 / $107; //@line 121 "./libraries/Box2D/Common/b2Math.h"
  $114 = $99 * $113; //@line 122 "./libraries/Box2D/Common/b2Math.h"
  $115 = $103 * $113; //@line 123 "./libraries/Box2D/Common/b2Math.h"
  $118 = $115;$edge1$sroa$0$0 = $114;
 }
 $116 = ((($this)) + 196|0); //@line 247 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $117 = -$edge1$sroa$0$0; //@line 68 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$116>>2] = $118; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $119 = ((($this)) + 200|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$119>>2] = $117; //@line 65 "./libraries/Box2D/Common/b2Math.h"
 $120 = $39 - $bitcast143; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $121 = $43 - $102; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $122 = $120 * $118; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $123 = $121 * $117; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $124 = $122 + $123; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 if ($97) {
  $125 = $bitcast143 - $109; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $126 = $102 - $110; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $127 = $125 * $125; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $128 = $126 * $126; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $129 = $127 + $128; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $130 = (+Math_sqrt((+$129))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $131 = $130 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
  if ($131) {
   $edge0$sroa$0$0 = $125;$edge0$sroa$7$0 = $126;
  } else {
   $132 = 1.0 / $130; //@line 121 "./libraries/Box2D/Common/b2Math.h"
   $133 = $125 * $132; //@line 122 "./libraries/Box2D/Common/b2Math.h"
   $134 = $126 * $132; //@line 123 "./libraries/Box2D/Common/b2Math.h"
   $edge0$sroa$0$0 = $133;$edge0$sroa$7$0 = $134;
  }
  $135 = -$edge0$sroa$0$0; //@line 257 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $136 = ((($this)) + 188|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$136>>2] = $edge0$sroa$7$0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $137 = ((($this)) + 192|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$137>>2] = $135; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $138 = $118 * $edge0$sroa$0$0; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $139 = $edge1$sroa$0$0 * $edge0$sroa$7$0; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $140 = $138 - $139; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $141 = $140 >= 0.0; //@line 258 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $142 = $39 - $109; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $143 = $43 - $110; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $144 = $142 * $edge0$sroa$7$0; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $145 = $143 * $135; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $146 = $144 + $145; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $convex1$0$off0 = $141;$offset0$0 = $146;
 } else {
  $convex1$0$off0 = 0;$offset0$0 = 0.0;
 }
 do {
  if ($98) {
   $147 = $111 - $bitcast; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $148 = $112 - $100; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $149 = $147 * $147; //@line 103 "./libraries/Box2D/Common/b2Math.h"
   $150 = $148 * $148; //@line 103 "./libraries/Box2D/Common/b2Math.h"
   $151 = $149 + $150; //@line 103 "./libraries/Box2D/Common/b2Math.h"
   $152 = (+Math_sqrt((+$151))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
   $153 = $152 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
   if ($153) {
    $159 = $148;$edge2$sroa$0$0 = $147;
   } else {
    $154 = 1.0 / $152; //@line 121 "./libraries/Box2D/Common/b2Math.h"
    $155 = $147 * $154; //@line 122 "./libraries/Box2D/Common/b2Math.h"
    $156 = $148 * $154; //@line 123 "./libraries/Box2D/Common/b2Math.h"
    $159 = $156;$edge2$sroa$0$0 = $155;
   }
   $157 = -$edge2$sroa$0$0; //@line 68 "./libraries/Box2D/Common/b2Math.h"
   $158 = ((($this)) + 204|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$158>>2] = $159; //@line 65 "./libraries/Box2D/Common/b2Math.h"
   $160 = ((($this)) + 208|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$160>>2] = $157; //@line 65 "./libraries/Box2D/Common/b2Math.h"
   $161 = $edge1$sroa$0$0 * $159; //@line 414 "./libraries/Box2D/Common/b2Math.h"
   $162 = $118 * $edge2$sroa$0$0; //@line 414 "./libraries/Box2D/Common/b2Math.h"
   $163 = $161 - $162; //@line 414 "./libraries/Box2D/Common/b2Math.h"
   $164 = $163 > 0.0; //@line 268 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $165 = $39 - $bitcast; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $166 = $43 - $100; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $167 = $165 * $159; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $168 = $166 * $157; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $169 = $167 + $168; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   if (!($97)) {
    $444 = $124 >= 0.0; //@line 379 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $445 = $169 >= 0.0; //@line 379 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    if ($164) {
     $446 = $444 | $445; //@line 379 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $447 = ((($this)) + 248|0); //@line 379 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $448 = $446&1; //@line 379 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP8[$447>>0] = $448; //@line 379 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $449 = ((($this)) + 212|0); //@line 382 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     if ($446) {
      $450 = $116; //@line 382 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $451 = $450; //@line 382 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $452 = HEAP32[$451>>2]|0; //@line 382 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $453 = (($450) + 4)|0; //@line 382 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $454 = $453; //@line 382 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $455 = HEAP32[$454>>2]|0; //@line 382 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $456 = $449; //@line 382 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $457 = $456; //@line 382 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$457>>2] = $452; //@line 382 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $458 = (($456) + 4)|0; //@line 382 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $459 = $458; //@line 382 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$459>>2] = $455; //@line 382 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $bitcast151 = (HEAP32[tempDoublePtr>>2]=$452,+HEAPF32[tempDoublePtr>>2]); //@line 68 "./libraries/Box2D/Common/b2Math.h"
      $460 = -$bitcast151; //@line 68 "./libraries/Box2D/Common/b2Math.h"
      $461 = ((($this)) + 228|0); //@line 383 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$461>>2] = $460; //@line 383 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $462 = ((($this)) + 232|0); //@line 383 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$462>>2] = $edge1$sroa$0$0; //@line 383 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $463 = ((($this)) + 236|0); //@line 384 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $464 = ((($this)) + 204|0); //@line 384 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $465 = $464; //@line 384 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $466 = $465; //@line 384 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $467 = HEAP32[$466>>2]|0; //@line 384 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $468 = (($465) + 4)|0; //@line 384 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $469 = $468; //@line 384 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $470 = HEAP32[$469>>2]|0; //@line 384 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $471 = $463; //@line 384 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $472 = $471; //@line 384 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$472>>2] = $467; //@line 384 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $473 = (($471) + 4)|0; //@line 384 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $474 = $473; //@line 384 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$474>>2] = $470; //@line 384 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      break;
     } else {
      $475 = -$118; //@line 68 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$449>>2] = $475; //@line 388 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $476 = ((($this)) + 216|0); //@line 388 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$476>>2] = $edge1$sroa$0$0; //@line 388 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $477 = ((($this)) + 228|0); //@line 389 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$477>>2] = $475; //@line 389 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $478 = ((($this)) + 232|0); //@line 389 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$478>>2] = $edge1$sroa$0$0; //@line 389 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $479 = ((($this)) + 236|0); //@line 390 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $480 = $116; //@line 390 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $481 = $480; //@line 390 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $482 = HEAP32[$481>>2]|0; //@line 390 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $483 = (($480) + 4)|0; //@line 390 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $484 = $483; //@line 390 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $485 = HEAP32[$484>>2]|0; //@line 390 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $486 = $479; //@line 390 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $487 = $486; //@line 390 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$487>>2] = $482; //@line 390 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $488 = (($486) + 4)|0; //@line 390 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $489 = $488; //@line 390 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$489>>2] = $485; //@line 390 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      break;
     }
    } else {
     $490 = $444 & $445; //@line 395 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $491 = ((($this)) + 248|0); //@line 395 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $492 = $490&1; //@line 395 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP8[$491>>0] = $492; //@line 395 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $493 = ((($this)) + 212|0); //@line 398 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     if ($490) {
      $494 = $116; //@line 398 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $495 = $494; //@line 398 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $496 = HEAP32[$495>>2]|0; //@line 398 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $497 = (($494) + 4)|0; //@line 398 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $498 = $497; //@line 398 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $499 = HEAP32[$498>>2]|0; //@line 398 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $500 = $493; //@line 398 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $501 = $500; //@line 398 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$501>>2] = $496; //@line 398 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $502 = (($500) + 4)|0; //@line 398 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $503 = $502; //@line 398 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$503>>2] = $499; //@line 398 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $bitcast149 = (HEAP32[tempDoublePtr>>2]=$496,+HEAPF32[tempDoublePtr>>2]); //@line 68 "./libraries/Box2D/Common/b2Math.h"
      $504 = -$bitcast149; //@line 68 "./libraries/Box2D/Common/b2Math.h"
      $505 = ((($this)) + 228|0); //@line 399 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$505>>2] = $504; //@line 399 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $506 = ((($this)) + 232|0); //@line 399 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$506>>2] = $edge1$sroa$0$0; //@line 399 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $507 = ((($this)) + 236|0); //@line 400 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $508 = $507; //@line 400 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $509 = $508; //@line 400 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$509>>2] = $496; //@line 400 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $510 = (($508) + 4)|0; //@line 400 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $511 = $510; //@line 400 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$511>>2] = $499; //@line 400 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      break;
     } else {
      $512 = -$118; //@line 68 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$493>>2] = $512; //@line 404 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $513 = ((($this)) + 216|0); //@line 404 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$513>>2] = $edge1$sroa$0$0; //@line 404 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $514 = ((($this)) + 204|0); //@line 68 "./libraries/Box2D/Common/b2Math.h"
      $515 = +HEAPF32[$514>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
      $516 = -$515; //@line 68 "./libraries/Box2D/Common/b2Math.h"
      $517 = ((($this)) + 208|0); //@line 68 "./libraries/Box2D/Common/b2Math.h"
      $518 = +HEAPF32[$517>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
      $519 = -$518; //@line 68 "./libraries/Box2D/Common/b2Math.h"
      $520 = ((($this)) + 228|0); //@line 405 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$520>>2] = $516; //@line 405 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $521 = ((($this)) + 232|0); //@line 405 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$521>>2] = $519; //@line 405 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $522 = ((($this)) + 236|0); //@line 406 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $523 = $116; //@line 406 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $524 = $523; //@line 406 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $525 = HEAP32[$524>>2]|0; //@line 406 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $526 = (($523) + 4)|0; //@line 406 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $527 = $526; //@line 406 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $528 = HEAP32[$527>>2]|0; //@line 406 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $529 = $522; //@line 406 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $530 = $529; //@line 406 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$530>>2] = $525; //@line 406 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $531 = (($529) + 4)|0; //@line 406 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $532 = $531; //@line 406 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$532>>2] = $528; //@line 406 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      break;
     }
    }
   }
   $brmerge37$demorgan = $convex1$0$off0 & $164; //@line 275 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   if ($brmerge37$demorgan) {
    $170 = $offset0$0 >= 0.0; //@line 277 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $171 = $124 >= 0.0; //@line 277 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $or$cond = $171 | $170; //@line 277 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $172 = $169 >= 0.0; //@line 277 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $$ = $or$cond | $172; //@line 277 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $173 = ((($this)) + 248|0); //@line 277 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $174 = $$&1; //@line 277 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP8[$173>>0] = $174; //@line 277 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $175 = ((($this)) + 212|0); //@line 280 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    if ($$) {
     $176 = $116; //@line 280 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $177 = $176; //@line 280 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $178 = HEAP32[$177>>2]|0; //@line 280 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $179 = (($176) + 4)|0; //@line 280 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $180 = $179; //@line 280 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $181 = HEAP32[$180>>2]|0; //@line 280 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $182 = $175; //@line 280 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $183 = $182; //@line 280 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$183>>2] = $178; //@line 280 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $184 = (($182) + 4)|0; //@line 280 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $185 = $184; //@line 280 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$185>>2] = $181; //@line 280 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $186 = ((($this)) + 228|0); //@line 281 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $187 = ((($this)) + 188|0); //@line 281 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $188 = $187; //@line 281 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $189 = $188; //@line 281 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $190 = HEAP32[$189>>2]|0; //@line 281 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $191 = (($188) + 4)|0; //@line 281 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $192 = $191; //@line 281 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $193 = HEAP32[$192>>2]|0; //@line 281 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $194 = $186; //@line 281 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $195 = $194; //@line 281 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$195>>2] = $190; //@line 281 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $196 = (($194) + 4)|0; //@line 281 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $197 = $196; //@line 281 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$197>>2] = $193; //@line 281 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $198 = ((($this)) + 236|0); //@line 282 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $199 = ((($this)) + 204|0); //@line 282 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $200 = $199; //@line 282 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $201 = $200; //@line 282 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $202 = HEAP32[$201>>2]|0; //@line 282 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $203 = (($200) + 4)|0; //@line 282 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $204 = $203; //@line 282 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $205 = HEAP32[$204>>2]|0; //@line 282 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $206 = $198; //@line 282 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $207 = $206; //@line 282 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$207>>2] = $202; //@line 282 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $208 = (($206) + 4)|0; //@line 282 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $209 = $208; //@line 282 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$209>>2] = $205; //@line 282 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     break;
    } else {
     $210 = -$118; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     HEAPF32[$175>>2] = $210; //@line 286 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $211 = ((($this)) + 216|0); //@line 286 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$211>>2] = $edge1$sroa$0$0; //@line 286 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $212 = ((($this)) + 228|0); //@line 287 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$212>>2] = $210; //@line 287 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $213 = ((($this)) + 232|0); //@line 287 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$213>>2] = $edge1$sroa$0$0; //@line 287 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $214 = ((($this)) + 236|0); //@line 288 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$214>>2] = $210; //@line 288 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $215 = ((($this)) + 240|0); //@line 288 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$215>>2] = $edge1$sroa$0$0; //@line 288 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     break;
    }
   }
   if ($convex1$0$off0) {
    $216 = !($offset0$0 >= 0.0); //@line 293 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    if ($216) {
     $219 = $124 >= 0.0; //@line 293 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $220 = $169 >= 0.0; //@line 293 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $221 = $219 & $220; //@line 293 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $222 = ((($this)) + 248|0); //@line 293 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $223 = $221&1; //@line 293 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP8[$222>>0] = $223; //@line 293 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $224 = ((($this)) + 212|0); //@line 296 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     if ($221) {
      $232 = $224;
     } else {
      $259 = -$118; //@line 68 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$224>>2] = $259; //@line 302 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $260 = ((($this)) + 216|0); //@line 302 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$260>>2] = $edge1$sroa$0$0; //@line 302 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $261 = -$159; //@line 68 "./libraries/Box2D/Common/b2Math.h"
      $262 = ((($this)) + 228|0); //@line 303 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$262>>2] = $261; //@line 303 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $263 = ((($this)) + 232|0); //@line 303 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$263>>2] = $edge2$sroa$0$0; //@line 303 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $264 = ((($this)) + 236|0); //@line 304 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$264>>2] = $259; //@line 304 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $265 = ((($this)) + 240|0); //@line 304 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$265>>2] = $edge1$sroa$0$0; //@line 304 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      break;
     }
    } else {
     $217 = ((($this)) + 248|0); //@line 293 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP8[$217>>0] = 1; //@line 293 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $218 = ((($this)) + 212|0); //@line 296 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $232 = $218;
    }
    $225 = $116; //@line 296 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $226 = $225; //@line 296 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $227 = HEAP32[$226>>2]|0; //@line 296 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $228 = (($225) + 4)|0; //@line 296 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $229 = $228; //@line 296 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $230 = HEAP32[$229>>2]|0; //@line 296 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $231 = $232; //@line 296 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $233 = $231; //@line 296 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$233>>2] = $227; //@line 296 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $234 = (($231) + 4)|0; //@line 296 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $235 = $234; //@line 296 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$235>>2] = $230; //@line 296 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $236 = ((($this)) + 228|0); //@line 297 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $237 = ((($this)) + 188|0); //@line 297 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $238 = $237; //@line 297 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $239 = $238; //@line 297 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $240 = HEAP32[$239>>2]|0; //@line 297 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $241 = (($238) + 4)|0; //@line 297 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $242 = $241; //@line 297 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $243 = HEAP32[$242>>2]|0; //@line 297 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $244 = $236; //@line 297 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $245 = $244; //@line 297 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$245>>2] = $240; //@line 297 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $246 = (($244) + 4)|0; //@line 297 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $247 = $246; //@line 297 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$247>>2] = $243; //@line 297 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $248 = ((($this)) + 236|0); //@line 298 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $249 = $116; //@line 298 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $250 = $249; //@line 298 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $251 = HEAP32[$250>>2]|0; //@line 298 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $252 = (($249) + 4)|0; //@line 298 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $253 = $252; //@line 298 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $254 = HEAP32[$253>>2]|0; //@line 298 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $255 = $248; //@line 298 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $256 = $255; //@line 298 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$256>>2] = $251; //@line 298 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $257 = (($255) + 4)|0; //@line 298 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $258 = $257; //@line 298 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$258>>2] = $254; //@line 298 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    break;
   }
   if (!($164)) {
    $315 = $offset0$0 >= 0.0; //@line 325 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $316 = $124 >= 0.0; //@line 325 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $or$cond3 = $316 & $315; //@line 325 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $317 = $169 >= 0.0; //@line 325 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $$38 = $or$cond3 & $317; //@line 325 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $318 = ((($this)) + 248|0); //@line 325 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $319 = $$38&1; //@line 325 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP8[$318>>0] = $319; //@line 325 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $320 = ((($this)) + 212|0); //@line 328 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    if ($$38) {
     $321 = $116; //@line 328 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $322 = $321; //@line 328 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $323 = HEAP32[$322>>2]|0; //@line 328 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $324 = (($321) + 4)|0; //@line 328 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $325 = $324; //@line 328 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $326 = HEAP32[$325>>2]|0; //@line 328 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $327 = $320; //@line 328 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $328 = $327; //@line 328 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$328>>2] = $323; //@line 328 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $329 = (($327) + 4)|0; //@line 328 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $330 = $329; //@line 328 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$330>>2] = $326; //@line 328 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $331 = ((($this)) + 228|0); //@line 329 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $332 = $331; //@line 329 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $333 = $332; //@line 329 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$333>>2] = $323; //@line 329 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $334 = (($332) + 4)|0; //@line 329 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $335 = $334; //@line 329 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$335>>2] = $326; //@line 329 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $336 = ((($this)) + 236|0); //@line 330 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $337 = $336; //@line 330 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $338 = $337; //@line 330 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$338>>2] = $323; //@line 330 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $339 = (($337) + 4)|0; //@line 330 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $340 = $339; //@line 330 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$340>>2] = $326; //@line 330 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     break;
    } else {
     $341 = -$118; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     HEAPF32[$320>>2] = $341; //@line 334 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $342 = ((($this)) + 216|0); //@line 334 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$342>>2] = $edge1$sroa$0$0; //@line 334 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $343 = -$159; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $344 = ((($this)) + 228|0); //@line 335 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$344>>2] = $343; //@line 335 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $345 = ((($this)) + 232|0); //@line 335 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$345>>2] = $edge2$sroa$0$0; //@line 335 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $346 = ((($this)) + 188|0); //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $347 = +HEAPF32[$346>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $348 = -$347; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $349 = ((($this)) + 192|0); //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $350 = +HEAPF32[$349>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $351 = -$350; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $352 = ((($this)) + 236|0); //@line 336 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$352>>2] = $348; //@line 336 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $353 = ((($this)) + 240|0); //@line 336 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$353>>2] = $351; //@line 336 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     break;
    }
   }
   $266 = !($169 >= 0.0); //@line 309 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   if ($266) {
    $269 = $offset0$0 >= 0.0; //@line 309 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $270 = $124 >= 0.0; //@line 309 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $271 = $270 & $269; //@line 309 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $272 = ((($this)) + 248|0); //@line 309 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $273 = $271&1; //@line 309 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP8[$272>>0] = $273; //@line 309 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $274 = ((($this)) + 212|0); //@line 312 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    if ($271) {
     $282 = $274;
    } else {
     $303 = -$118; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     HEAPF32[$274>>2] = $303; //@line 318 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $304 = ((($this)) + 216|0); //@line 318 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$304>>2] = $edge1$sroa$0$0; //@line 318 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $305 = ((($this)) + 228|0); //@line 319 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$305>>2] = $303; //@line 319 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $306 = ((($this)) + 232|0); //@line 319 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$306>>2] = $edge1$sroa$0$0; //@line 319 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $307 = ((($this)) + 188|0); //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $308 = +HEAPF32[$307>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $309 = -$308; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $310 = ((($this)) + 192|0); //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $311 = +HEAPF32[$310>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $312 = -$311; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $313 = ((($this)) + 236|0); //@line 320 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$313>>2] = $309; //@line 320 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $314 = ((($this)) + 240|0); //@line 320 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$314>>2] = $312; //@line 320 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     break;
    }
   } else {
    $267 = ((($this)) + 248|0); //@line 309 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP8[$267>>0] = 1; //@line 309 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $268 = ((($this)) + 212|0); //@line 312 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $282 = $268;
   }
   $275 = $116; //@line 312 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $276 = $275; //@line 312 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $277 = HEAP32[$276>>2]|0; //@line 312 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $278 = (($275) + 4)|0; //@line 312 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $279 = $278; //@line 312 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $280 = HEAP32[$279>>2]|0; //@line 312 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $281 = $282; //@line 312 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $283 = $281; //@line 312 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP32[$283>>2] = $277; //@line 312 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $284 = (($281) + 4)|0; //@line 312 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $285 = $284; //@line 312 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP32[$285>>2] = $280; //@line 312 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $286 = ((($this)) + 228|0); //@line 313 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $287 = $286; //@line 313 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $288 = $287; //@line 313 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP32[$288>>2] = $277; //@line 313 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $289 = (($287) + 4)|0; //@line 313 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $290 = $289; //@line 313 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP32[$290>>2] = $280; //@line 313 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $291 = ((($this)) + 236|0); //@line 314 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $292 = ((($this)) + 204|0); //@line 314 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $293 = $292; //@line 314 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $294 = $293; //@line 314 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $295 = HEAP32[$294>>2]|0; //@line 314 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $296 = (($293) + 4)|0; //@line 314 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $297 = $296; //@line 314 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $298 = HEAP32[$297>>2]|0; //@line 314 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $299 = $291; //@line 314 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $300 = $299; //@line 314 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP32[$300>>2] = $295; //@line 314 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $301 = (($299) + 4)|0; //@line 314 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $302 = $301; //@line 314 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP32[$302>>2] = $298; //@line 314 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  } else {
   if (!($97)) {
    $533 = $124 >= 0.0; //@line 379 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $534 = ((($this)) + 248|0); //@line 412 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $535 = $533&1; //@line 412 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP8[$534>>0] = $535; //@line 412 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $536 = ((($this)) + 212|0); //@line 415 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    if ($533) {
     $537 = $116; //@line 415 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $538 = $537; //@line 415 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $539 = HEAP32[$538>>2]|0; //@line 415 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $540 = (($537) + 4)|0; //@line 415 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $541 = $540; //@line 415 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $542 = HEAP32[$541>>2]|0; //@line 415 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $543 = $536; //@line 415 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $544 = $543; //@line 415 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$544>>2] = $539; //@line 415 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $545 = (($543) + 4)|0; //@line 415 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $546 = $545; //@line 415 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$546>>2] = $542; //@line 415 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $bitcast145 = (HEAP32[tempDoublePtr>>2]=$539,+HEAPF32[tempDoublePtr>>2]); //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $547 = -$bitcast145; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $548 = ((($this)) + 228|0); //@line 416 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$548>>2] = $547; //@line 416 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $549 = ((($this)) + 232|0); //@line 416 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$549>>2] = $edge1$sroa$0$0; //@line 416 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $550 = ((($this)) + 236|0); //@line 417 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$550>>2] = $547; //@line 417 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $551 = ((($this)) + 240|0); //@line 417 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$551>>2] = $edge1$sroa$0$0; //@line 417 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     break;
    } else {
     $552 = -$118; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     HEAPF32[$536>>2] = $552; //@line 421 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $553 = ((($this)) + 216|0); //@line 421 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$553>>2] = $edge1$sroa$0$0; //@line 421 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $554 = ((($this)) + 228|0); //@line 422 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $555 = $116; //@line 422 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $556 = $555; //@line 422 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $557 = HEAP32[$556>>2]|0; //@line 422 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $558 = (($555) + 4)|0; //@line 422 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $559 = $558; //@line 422 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $560 = HEAP32[$559>>2]|0; //@line 422 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $561 = $554; //@line 422 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $562 = $561; //@line 422 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$562>>2] = $557; //@line 422 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $563 = (($561) + 4)|0; //@line 422 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $564 = $563; //@line 422 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$564>>2] = $560; //@line 422 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $565 = ((($this)) + 236|0); //@line 423 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $566 = $565; //@line 423 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $567 = $566; //@line 423 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$567>>2] = $557; //@line 423 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $568 = (($566) + 4)|0; //@line 423 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $569 = $568; //@line 423 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$569>>2] = $560; //@line 423 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     break;
    }
   }
   $354 = $offset0$0 >= 0.0; //@line 344 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $355 = $124 >= 0.0; //@line 344 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   if ($convex1$0$off0) {
    $356 = $355 | $354; //@line 344 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $357 = ((($this)) + 248|0); //@line 344 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $358 = $356&1; //@line 344 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP8[$357>>0] = $358; //@line 344 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $359 = ((($this)) + 212|0); //@line 347 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    if ($356) {
     $360 = $116; //@line 347 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $361 = $360; //@line 347 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $362 = HEAP32[$361>>2]|0; //@line 347 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $363 = (($360) + 4)|0; //@line 347 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $364 = $363; //@line 347 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $365 = HEAP32[$364>>2]|0; //@line 347 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $366 = $359; //@line 347 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $367 = $366; //@line 347 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$367>>2] = $362; //@line 347 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $368 = (($366) + 4)|0; //@line 347 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $369 = $368; //@line 347 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$369>>2] = $365; //@line 347 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $370 = ((($this)) + 228|0); //@line 348 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $371 = ((($this)) + 188|0); //@line 348 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $372 = $371; //@line 348 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $373 = $372; //@line 348 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $374 = HEAP32[$373>>2]|0; //@line 348 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $375 = (($372) + 4)|0; //@line 348 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $376 = $375; //@line 348 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $377 = HEAP32[$376>>2]|0; //@line 348 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $378 = $370; //@line 348 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $379 = $378; //@line 348 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$379>>2] = $374; //@line 348 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $380 = (($378) + 4)|0; //@line 348 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $381 = $380; //@line 348 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$381>>2] = $377; //@line 348 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $bitcast157 = (HEAP32[tempDoublePtr>>2]=$362,+HEAPF32[tempDoublePtr>>2]); //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $382 = -$bitcast157; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $383 = ((($this)) + 236|0); //@line 349 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$383>>2] = $382; //@line 349 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $384 = ((($this)) + 240|0); //@line 349 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$384>>2] = $edge1$sroa$0$0; //@line 349 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     break;
    } else {
     $385 = -$118; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     HEAPF32[$359>>2] = $385; //@line 353 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $386 = ((($this)) + 216|0); //@line 353 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$386>>2] = $edge1$sroa$0$0; //@line 353 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $387 = ((($this)) + 228|0); //@line 354 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $388 = $116; //@line 354 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $389 = $388; //@line 354 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $390 = HEAP32[$389>>2]|0; //@line 354 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $391 = (($388) + 4)|0; //@line 354 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $392 = $391; //@line 354 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $393 = HEAP32[$392>>2]|0; //@line 354 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $394 = $387; //@line 354 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $395 = $394; //@line 354 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$395>>2] = $390; //@line 354 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $396 = (($394) + 4)|0; //@line 354 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $397 = $396; //@line 354 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$397>>2] = $393; //@line 354 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $bitcast155 = (HEAP32[tempDoublePtr>>2]=$390,+HEAPF32[tempDoublePtr>>2]); //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $398 = -$bitcast155; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $399 = ((($this)) + 236|0); //@line 355 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$399>>2] = $398; //@line 355 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $400 = ((($this)) + 240|0); //@line 355 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$400>>2] = $edge1$sroa$0$0; //@line 355 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     break;
    }
   } else {
    $401 = $355 & $354; //@line 360 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $402 = ((($this)) + 248|0); //@line 360 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $403 = $401&1; //@line 360 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP8[$402>>0] = $403; //@line 360 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $404 = ((($this)) + 212|0); //@line 363 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    if ($401) {
     $405 = $116; //@line 363 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $406 = $405; //@line 363 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $407 = HEAP32[$406>>2]|0; //@line 363 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $408 = (($405) + 4)|0; //@line 363 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $409 = $408; //@line 363 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $410 = HEAP32[$409>>2]|0; //@line 363 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $411 = $404; //@line 363 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $412 = $411; //@line 363 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$412>>2] = $407; //@line 363 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $413 = (($411) + 4)|0; //@line 363 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $414 = $413; //@line 363 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$414>>2] = $410; //@line 363 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $415 = ((($this)) + 228|0); //@line 364 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $416 = $415; //@line 364 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $417 = $416; //@line 364 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$417>>2] = $407; //@line 364 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $418 = (($416) + 4)|0; //@line 364 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $419 = $418; //@line 364 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$419>>2] = $410; //@line 364 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $bitcast153 = (HEAP32[tempDoublePtr>>2]=$407,+HEAPF32[tempDoublePtr>>2]); //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $420 = -$bitcast153; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $421 = ((($this)) + 236|0); //@line 365 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$421>>2] = $420; //@line 365 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $422 = ((($this)) + 240|0); //@line 365 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$422>>2] = $edge1$sroa$0$0; //@line 365 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     break;
    } else {
     $423 = -$118; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     HEAPF32[$404>>2] = $423; //@line 369 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $424 = ((($this)) + 216|0); //@line 369 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$424>>2] = $edge1$sroa$0$0; //@line 369 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $425 = ((($this)) + 228|0); //@line 370 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $426 = $116; //@line 370 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $427 = $426; //@line 370 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $428 = HEAP32[$427>>2]|0; //@line 370 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $429 = (($426) + 4)|0; //@line 370 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $430 = $429; //@line 370 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $431 = HEAP32[$430>>2]|0; //@line 370 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $432 = $425; //@line 370 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $433 = $432; //@line 370 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$433>>2] = $428; //@line 370 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $434 = (($432) + 4)|0; //@line 370 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $435 = $434; //@line 370 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$435>>2] = $431; //@line 370 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $436 = ((($this)) + 188|0); //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $437 = +HEAPF32[$436>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $438 = -$437; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $439 = ((($this)) + 192|0); //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $440 = +HEAPF32[$439>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $441 = -$440; //@line 68 "./libraries/Box2D/Common/b2Math.h"
     $442 = ((($this)) + 236|0); //@line 371 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$442>>2] = $438; //@line 371 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $443 = ((($this)) + 240|0); //@line 371 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAPF32[$443>>2] = $441; //@line 371 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     break;
    }
   }
  }
 } while(0);
 $570 = ((($polygonB)) + 148|0); //@line 428 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $571 = HEAP32[$570>>2]|0; //@line 428 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $572 = ((($this)) + 128|0); //@line 428 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$572>>2] = $571; //@line 428 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $573 = ($571|0)>(0); //@line 429 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 if ($573) {
  $i$0131 = 0;
  while(1) {
   $599 = +HEAPF32[$31>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $600 = (((($polygonB)) + 20|0) + ($i$0131<<3)|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $601 = +HEAPF32[$600>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $602 = $599 * $601; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $603 = +HEAPF32[$30>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $604 = (((((($polygonB)) + 20|0) + ($i$0131<<3)|0)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $605 = +HEAPF32[$604>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $606 = $603 * $605; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $607 = $602 - $606; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $608 = +HEAPF32[$28>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $609 = $608 + $607; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $610 = $601 * $603; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $611 = $599 * $605; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $612 = $610 + $611; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $613 = +HEAPF32[$29>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $614 = $612 + $613; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $615 = (($this) + ($i$0131<<3)|0); //@line 431 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAPF32[$615>>2] = $609; //@line 431 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $616 = (((($this) + ($i$0131<<3)|0)) + 4|0); //@line 431 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAPF32[$616>>2] = $614; //@line 431 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $617 = +HEAPF32[$31>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $618 = (((($polygonB)) + 84|0) + ($i$0131<<3)|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $619 = +HEAPF32[$618>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $620 = $617 * $619; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $621 = +HEAPF32[$30>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $622 = (((((($polygonB)) + 84|0) + ($i$0131<<3)|0)) + 4|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $623 = +HEAPF32[$622>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $624 = $621 * $623; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $625 = $620 - $624; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $626 = $619 * $621; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $627 = $617 * $623; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $628 = $626 + $627; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $629 = (((($this)) + 64|0) + ($i$0131<<3)|0); //@line 432 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAPF32[$629>>2] = $625; //@line 432 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $630 = (((((($this)) + 64|0) + ($i$0131<<3)|0)) + 4|0); //@line 432 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAPF32[$630>>2] = $628; //@line 432 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $631 = (($i$0131) + 1)|0; //@line 429 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $632 = HEAP32[$570>>2]|0; //@line 429 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $633 = ($631|0)<($632|0); //@line 429 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   if ($633) {
    $i$0131 = $631;
   } else {
    break;
   }
  }
  $$pre = HEAP32[$572>>2]|0; //@line 627 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $578 = $$pre;
 } else {
  $578 = $571;
 }
 $574 = ((($this)) + 244|0); //@line 435 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAPF32[$574>>2] = 0.019999999552965164; //@line 435 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $575 = ((($manifold)) + 60|0); //@line 437 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$575>>2] = 0; //@line 437 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $576 = ((($this)) + 248|0); //@line 624 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $577 = ($578|0)>(0); //@line 627 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 if (!($577)) {
  STACKTOP = sp;return; //@line 618 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 }
 $579 = ((($this)) + 164|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $580 = +HEAPF32[$579>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $581 = +HEAPF32[$101>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $582 = ((($this)) + 212|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $583 = +HEAPF32[$582>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $584 = ((($this)) + 216|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $585 = +HEAPF32[$584>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $596 = 3.4028234663852886E+38;$edgeAxis$sroa$9$0 = 3.4028234663852886E+38;$i$01$i = 0;
 while(1) {
  $586 = (($this) + ($i$01$i<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $587 = +HEAPF32[$586>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $588 = $587 - $580; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $589 = (((($this) + ($i$01$i<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $590 = +HEAPF32[$589>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $591 = $590 - $581; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $592 = $583 * $588; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $593 = $585 * $591; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $594 = $592 + $593; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $595 = $594 < $596; //@line 630 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $edgeAxis$sroa$9$1 = $595 ? $594 : $edgeAxis$sroa$9$0; //@line 630 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $597 = $595 ? $594 : $596; //@line 630 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $598 = (($i$01$i) + 1)|0; //@line 627 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $exitcond137 = ($598|0)==($578|0); //@line 627 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  if ($exitcond137) {
   $edgeAxis$sroa$9$1$lcssa = $edgeAxis$sroa$9$1;
   break;
  } else {
   $596 = $597;$edgeAxis$sroa$9$0 = $edgeAxis$sroa$9$1;$i$01$i = $598;
  }
 }
 $634 = $edgeAxis$sroa$9$1$lcssa > 0.019999999552965164; //@line 447 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 if ($634) {
  STACKTOP = sp;return; //@line 618 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 }
 __ZN12b2EPCollider24ComputePolygonSeparationEv($polygonAxis,$this); //@line 452 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $635 = HEAP32[$polygonAxis>>2]|0; //@line 463 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $636 = ($635|0)==(0); //@line 453 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 if ($636) {
  label = 58;
 } else {
  $637 = ((($polygonAxis)) + 8|0); //@line 453 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $638 = +HEAPF32[$637>>2]; //@line 467 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $639 = +HEAPF32[$574>>2]; //@line 453 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $640 = $638 > $639; //@line 453 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  if (!($640)) {
   $641 = $edgeAxis$sroa$9$1$lcssa * 0.98000001907348633; //@line 467 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $642 = $641 + 0.0010000000474974513; //@line 467 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $643 = $638 > $642; //@line 467 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   if ($643) {
    $645 = ((($polygonAxis)) + 4|0); //@line 469 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $646 = HEAP32[$645>>2]|0; //@line 469 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $phitmp = ($635|0)==(1); //@line 470 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $647 = ((($manifold)) + 56|0); //@line 480 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    if ($phitmp) {
     $1054 = $rf;$1055 = $ie;$1056 = $ie;$648 = $647;
     label = 60;
    } else {
     HEAP32[$647>>2] = 2; //@line 529 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $775 = $58; //@line 531 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $776 = $775; //@line 531 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $777 = HEAP32[$776>>2]|0; //@line 531 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $778 = (($775) + 4)|0; //@line 531 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $779 = $778; //@line 531 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $780 = HEAP32[$779>>2]|0; //@line 531 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $781 = $ie; //@line 531 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $782 = $781; //@line 531 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$782>>2] = $777; //@line 531 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $783 = (($781) + 4)|0; //@line 531 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $784 = $783; //@line 531 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$784>>2] = $780; //@line 531 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $785 = ((($ie)) + 8|0); //@line 532 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP8[$785>>0] = 0; //@line 532 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $786 = $646&255; //@line 533 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $787 = ((($785)) + 1|0); //@line 533 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP8[$787>>0] = $786; //@line 533 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $788 = ((($785)) + 2|0); //@line 534 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP8[$788>>0] = 0; //@line 534 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $789 = ((($785)) + 3|0); //@line 535 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP8[$789>>0] = 1; //@line 535 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $790 = ((($ie)) + 12|0); //@line 537 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $791 = $70; //@line 537 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $792 = $791; //@line 537 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $793 = HEAP32[$792>>2]|0; //@line 537 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $794 = (($791) + 4)|0; //@line 537 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $795 = $794; //@line 537 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $796 = HEAP32[$795>>2]|0; //@line 537 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $797 = $790; //@line 537 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $798 = $797; //@line 537 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$798>>2] = $793; //@line 537 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $799 = (($797) + 4)|0; //@line 537 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $800 = $799; //@line 537 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$800>>2] = $796; //@line 537 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $801 = ((($ie)) + 20|0); //@line 538 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP8[$801>>0] = 0; //@line 538 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $802 = ((($801)) + 1|0); //@line 539 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP8[$802>>0] = $786; //@line 539 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $803 = ((($801)) + 2|0); //@line 540 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP8[$803>>0] = 0; //@line 540 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $804 = ((($801)) + 3|0); //@line 541 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP8[$804>>0] = 1; //@line 541 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$rf>>2] = $646; //@line 543 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $805 = (($646) + 1)|0; //@line 544 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $806 = HEAP32[$572>>2]|0; //@line 544 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $807 = ($805|0)<($806|0); //@line 544 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $808 = $807 ? $805 : 0; //@line 544 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $809 = ((($rf)) + 4|0); //@line 544 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$809>>2] = $808; //@line 544 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $810 = ((($rf)) + 8|0); //@line 545 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $811 = (($this) + ($646<<3)|0); //@line 545 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $812 = $811; //@line 545 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $813 = $812; //@line 545 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $814 = HEAP32[$813>>2]|0; //@line 545 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $815 = (($812) + 4)|0; //@line 545 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $816 = $815; //@line 545 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $817 = HEAP32[$816>>2]|0; //@line 545 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $818 = $810; //@line 545 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $819 = $818; //@line 545 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$819>>2] = $814; //@line 545 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $820 = (($818) + 4)|0; //@line 545 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $821 = $820; //@line 545 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$821>>2] = $817; //@line 545 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $822 = ((($rf)) + 16|0); //@line 546 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $823 = (($this) + ($808<<3)|0); //@line 546 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $824 = $823; //@line 546 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $825 = $824; //@line 546 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $826 = HEAP32[$825>>2]|0; //@line 546 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $827 = (($824) + 4)|0; //@line 546 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $828 = $827; //@line 546 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $829 = HEAP32[$828>>2]|0; //@line 546 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $830 = $822; //@line 546 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $831 = $830; //@line 546 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$831>>2] = $826; //@line 546 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $832 = (($830) + 4)|0; //@line 546 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $833 = $832; //@line 546 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$833>>2] = $829; //@line 546 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $834 = ((($rf)) + 24|0); //@line 547 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $835 = (((($this)) + 64|0) + ($646<<3)|0); //@line 547 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $836 = $835; //@line 547 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $837 = $836; //@line 547 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $838 = HEAP32[$837>>2]|0; //@line 547 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $839 = (($836) + 4)|0; //@line 547 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $840 = $839; //@line 547 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $841 = HEAP32[$840>>2]|0; //@line 547 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $842 = $834; //@line 547 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $843 = $842; //@line 547 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$843>>2] = $838; //@line 547 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $844 = (($842) + 4)|0; //@line 547 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $845 = $844; //@line 547 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$845>>2] = $841; //@line 547 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $846 = (HEAP32[tempDoublePtr>>2]=$838,+HEAPF32[tempDoublePtr>>2]);
     $$in = $829;$$in165 = $826;$$in166 = $817;$$in167 = $814;$$pre$phi163Z2D = $810;$$pre$phi164Z2D = $rf;$$pre$phiZ2D = $834;$1057 = $rf;$1058 = $ie;$855 = $846;$856 = $841;$871 = $ie;$872 = $646;$primaryAxis$sroa$0$0173 = 0;
     label = 67;
    }
   } else {
    label = 58;
   }
  }
 }
 if ((label|0) == 58) {
  $644 = ((($manifold)) + 56|0); //@line 480 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $1054 = $rf;$1055 = $ie;$1056 = $ie;$648 = $644;
  label = 60;
 }
 do {
  if ((label|0) == 60) {
   HEAP32[$648>>2] = 1; //@line 480 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $649 = HEAP32[$572>>2]|0; //@line 485 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $650 = ($649|0)>(1); //@line 485 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   if ($650) {
    $651 = ((($this)) + 212|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $652 = ((($this)) + 216|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $653 = +HEAPF32[$651>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $654 = ((($this)) + 64|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $655 = +HEAPF32[$654>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $656 = $653 * $655; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $657 = +HEAPF32[$652>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $658 = ((($this)) + 68|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $659 = +HEAPF32[$658>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $660 = $657 * $659; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $661 = $656 + $660; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $bestIndex$0127 = 0;$bestValue$0128 = $661;$i1$0129 = 1;
    while(1) {
     $700 = (((($this)) + 64|0) + ($i1$0129<<3)|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $701 = +HEAPF32[$700>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $702 = $653 * $701; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $703 = (((((($this)) + 64|0) + ($i1$0129<<3)|0)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $704 = +HEAPF32[$703>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $705 = $657 * $704; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $706 = $702 + $705; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $707 = $706 < $bestValue$0128; //@line 488 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $bestIndex$1 = $707 ? $i1$0129 : $bestIndex$0127; //@line 488 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $bestValue$1 = $707 ? $706 : $bestValue$0128; //@line 488 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $708 = (($i1$0129) + 1)|0; //@line 485 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $709 = ($708|0)<($649|0); //@line 485 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     if ($709) {
      $bestIndex$0127 = $bestIndex$1;$bestValue$0128 = $bestValue$1;$i1$0129 = $708;
     } else {
      $bestIndex$0$lcssa = $bestIndex$1;
      break;
     }
    }
   } else {
    $bestIndex$0$lcssa = 0;
   }
   $662 = (($bestIndex$0$lcssa) + 1)|0; //@line 496 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $663 = ($662|0)<($649|0); //@line 496 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $664 = $663 ? $662 : 0; //@line 496 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $665 = (($this) + ($bestIndex$0$lcssa<<3)|0); //@line 498 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $666 = $665; //@line 498 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $667 = $666; //@line 498 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $668 = HEAP32[$667>>2]|0; //@line 498 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $669 = (($666) + 4)|0; //@line 498 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $670 = $669; //@line 498 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $671 = HEAP32[$670>>2]|0; //@line 498 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $672 = $ie; //@line 498 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $673 = $672; //@line 498 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP32[$673>>2] = $668; //@line 498 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $674 = (($672) + 4)|0; //@line 498 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $675 = $674; //@line 498 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP32[$675>>2] = $671; //@line 498 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $676 = ((($ie)) + 8|0); //@line 499 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP8[$676>>0] = 0; //@line 499 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $677 = $bestIndex$0$lcssa&255; //@line 500 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $678 = ((($676)) + 1|0); //@line 500 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP8[$678>>0] = $677; //@line 500 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $679 = ((($676)) + 2|0); //@line 501 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP8[$679>>0] = 1; //@line 501 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $680 = ((($676)) + 3|0); //@line 502 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP8[$680>>0] = 0; //@line 502 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $681 = ((($ie)) + 12|0); //@line 504 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $682 = (($this) + ($664<<3)|0); //@line 504 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $683 = $682; //@line 504 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $684 = $683; //@line 504 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $685 = HEAP32[$684>>2]|0; //@line 504 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $686 = (($683) + 4)|0; //@line 504 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $687 = $686; //@line 504 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $688 = HEAP32[$687>>2]|0; //@line 504 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $689 = $681; //@line 504 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $690 = $689; //@line 504 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP32[$690>>2] = $685; //@line 504 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $691 = (($689) + 4)|0; //@line 504 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $692 = $691; //@line 504 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP32[$692>>2] = $688; //@line 504 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $693 = ((($ie)) + 20|0); //@line 505 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP8[$693>>0] = 0; //@line 505 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $694 = $664&255; //@line 506 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $695 = ((($693)) + 1|0); //@line 506 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP8[$695>>0] = $694; //@line 506 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $696 = ((($693)) + 2|0); //@line 507 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP8[$696>>0] = 1; //@line 507 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $697 = ((($693)) + 3|0); //@line 508 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP8[$697>>0] = 0; //@line 508 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $698 = HEAP8[$576>>0]|0; //@line 510 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $699 = ($698<<24>>24)==(0); //@line 510 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   if ($699) {
    HEAP32[$rf>>2] = 1; //@line 520 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $745 = ((($rf)) + 4|0); //@line 521 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$745>>2] = 0; //@line 521 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $746 = ((($rf)) + 8|0); //@line 522 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $747 = $70; //@line 522 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $748 = $747; //@line 522 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $749 = HEAP32[$748>>2]|0; //@line 522 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $750 = (($747) + 4)|0; //@line 522 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $751 = $750; //@line 522 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $752 = HEAP32[$751>>2]|0; //@line 522 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $753 = $746; //@line 522 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $754 = $753; //@line 522 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$754>>2] = $749; //@line 522 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $755 = (($753) + 4)|0; //@line 522 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $756 = $755; //@line 522 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$756>>2] = $752; //@line 522 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $757 = ((($rf)) + 16|0); //@line 523 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $758 = $58; //@line 523 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $759 = $758; //@line 523 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $760 = HEAP32[$759>>2]|0; //@line 523 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $761 = (($758) + 4)|0; //@line 523 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $762 = $761; //@line 523 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $763 = HEAP32[$762>>2]|0; //@line 523 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $764 = $757; //@line 523 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $765 = $764; //@line 523 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$765>>2] = $760; //@line 523 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $766 = (($764) + 4)|0; //@line 523 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $767 = $766; //@line 523 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$767>>2] = $763; //@line 523 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $768 = +HEAPF32[$116>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
    $769 = -$768; //@line 68 "./libraries/Box2D/Common/b2Math.h"
    $770 = +HEAPF32[$119>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
    $771 = -$770; //@line 68 "./libraries/Box2D/Common/b2Math.h"
    $772 = ((($rf)) + 24|0); //@line 524 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAPF32[$772>>2] = $769; //@line 524 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $773 = ((($rf)) + 28|0); //@line 524 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAPF32[$773>>2] = $771; //@line 524 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $774 = (HEAPF32[tempDoublePtr>>2]=$771,HEAP32[tempDoublePtr>>2]|0);
    $$pre162 = ((($rf)) + 24|0); //@line 550 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $$in = $763;$$in165 = $760;$$in166 = $752;$$in167 = $749;$$pre$phi163Z2D = $746;$$pre$phi164Z2D = $rf;$$pre$phiZ2D = $$pre162;$1057 = $1054;$1058 = $1056;$855 = $769;$856 = $774;$871 = $1055;$872 = 1;$primaryAxis$sroa$0$0173 = 1;
    label = 67;
    break;
   } else {
    HEAP32[$rf>>2] = 0; //@line 512 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $710 = ((($rf)) + 4|0); //@line 513 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$710>>2] = 1; //@line 513 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $711 = ((($rf)) + 8|0); //@line 514 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $712 = $58; //@line 514 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $713 = $712; //@line 514 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $714 = HEAP32[$713>>2]|0; //@line 514 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $715 = (($712) + 4)|0; //@line 514 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $716 = $715; //@line 514 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $717 = HEAP32[$716>>2]|0; //@line 514 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $718 = $711; //@line 514 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $719 = $718; //@line 514 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$719>>2] = $714; //@line 514 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $720 = (($718) + 4)|0; //@line 514 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $721 = $720; //@line 514 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$721>>2] = $717; //@line 514 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $722 = ((($rf)) + 16|0); //@line 515 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $723 = $70; //@line 515 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $724 = $723; //@line 515 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $725 = HEAP32[$724>>2]|0; //@line 515 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $726 = (($723) + 4)|0; //@line 515 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $727 = $726; //@line 515 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $728 = HEAP32[$727>>2]|0; //@line 515 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $729 = $722; //@line 515 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $730 = $729; //@line 515 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$730>>2] = $725; //@line 515 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $731 = (($729) + 4)|0; //@line 515 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $732 = $731; //@line 515 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$732>>2] = $728; //@line 515 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $733 = ((($rf)) + 24|0); //@line 516 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $734 = $116; //@line 516 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $735 = $734; //@line 516 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $736 = HEAP32[$735>>2]|0; //@line 516 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $737 = (($734) + 4)|0; //@line 516 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $738 = $737; //@line 516 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $739 = HEAP32[$738>>2]|0; //@line 516 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $740 = $733; //@line 516 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $741 = $740; //@line 516 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$741>>2] = $736; //@line 516 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $742 = (($740) + 4)|0; //@line 516 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $743 = $742; //@line 516 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    HEAP32[$743>>2] = $739; //@line 516 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $744 = (HEAP32[tempDoublePtr>>2]=$736,+HEAPF32[tempDoublePtr>>2]); //@line 517 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    $$in = $728;$$in165 = $725;$$in166 = $717;$$in167 = $714;$$pre$phi163Z2D = $711;$$pre$phi164Z2D = $rf;$$pre$phiZ2D = $733;$1057 = $1054;$1058 = $1056;$855 = $744;$856 = $739;$871 = $1055;$872 = 0;$primaryAxis$sroa$0$0173 = 1;
    label = 67;
    break;
   }
  }
 } while(0);
 if ((label|0) == 67) {
  $847 = (HEAP32[tempDoublePtr>>2]=$$in167,+HEAPF32[tempDoublePtr>>2]); //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $848 = (HEAP32[tempDoublePtr>>2]=$$in166,+HEAPF32[tempDoublePtr>>2]); //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $849 = (HEAP32[tempDoublePtr>>2]=$$in165,+HEAPF32[tempDoublePtr>>2]); //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $850 = (HEAP32[tempDoublePtr>>2]=$$in,+HEAPF32[tempDoublePtr>>2]); //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $851 = ((($rf)) + 32|0); //@line 550 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $852 = ((($rf)) + 28|0); //@line 550 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $853 = ((($rf)) + 24|0); //@line 550 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $854 = -$855; //@line 550 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAP32[$851>>2] = $856; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $857 = ((($rf)) + 36|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$857>>2] = $854; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $858 = ((($rf)) + 44|0); //@line 551 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $$cast = (HEAP32[tempDoublePtr>>2]=$856,+HEAPF32[tempDoublePtr>>2]); //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $859 = -$$cast; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$858>>2] = $859; //@line 551 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $860 = ((($rf)) + 48|0); //@line 551 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAPF32[$860>>2] = $855; //@line 551 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $861 = ((($rf)) + 8|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $862 = $$cast * $847; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $863 = ((($rf)) + 12|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $864 = $848 * $854; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $865 = $862 + $864; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $866 = ((($rf)) + 40|0); //@line 552 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAPF32[$866>>2] = $865; //@line 552 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $867 = $849 * $859; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $868 = $855 * $850; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $869 = $867 + $868; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $870 = ((($rf)) + 52|0); //@line 553 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  HEAPF32[$870>>2] = $869; //@line 553 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $873 = (__Z19b2ClipSegmentToLineP12b2ClipVertexPKS_RK6b2Vec2fi($clipPoints1,$871,$851,$865,$872)|0); //@line 561 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $874 = ($873|0)<(2); //@line 563 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  if (!($874)) {
   $875 = +HEAPF32[$870>>2]; //@line 569 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $876 = ((($rf)) + 4|0); //@line 569 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $877 = HEAP32[$876>>2]|0; //@line 569 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $878 = (__Z19b2ClipSegmentToLineP12b2ClipVertexPKS_RK6b2Vec2fi($clipPoints2,$clipPoints1,$858,$875,$877)|0); //@line 569 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $879 = ($878|0)<(2); //@line 571 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   if (!($879)) {
    $880 = ((($manifold)) + 40|0); //@line 579 "libraries/Box2D/Collision/b2CollideEdge.cpp"
    if ($primaryAxis$sroa$0$0173) {
     $881 = $$pre$phiZ2D; //@line 579 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $882 = $881; //@line 579 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $883 = HEAP32[$882>>2]|0; //@line 579 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $884 = (($881) + 4)|0; //@line 579 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $885 = $884; //@line 579 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $886 = HEAP32[$885>>2]|0; //@line 579 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $887 = $880; //@line 579 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $888 = $887; //@line 579 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$888>>2] = $883; //@line 579 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $889 = (($887) + 4)|0; //@line 579 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $890 = $889; //@line 579 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$890>>2] = $886; //@line 579 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $891 = ((($manifold)) + 48|0); //@line 580 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $892 = $$pre$phi163Z2D; //@line 580 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $893 = $892; //@line 580 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $894 = HEAP32[$893>>2]|0; //@line 580 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $895 = (($892) + 4)|0; //@line 580 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $896 = $895; //@line 580 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $897 = HEAP32[$896>>2]|0; //@line 580 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $898 = $891; //@line 580 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $899 = $898; //@line 580 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$899>>2] = $894; //@line 580 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $900 = (($898) + 4)|0; //@line 580 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $901 = $900; //@line 580 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$901>>2] = $897; //@line 580 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $902 = (HEAP32[tempDoublePtr>>2]=$894,+HEAPF32[tempDoublePtr>>2]); //@line 581 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $903 = (HEAP32[tempDoublePtr>>2]=$883,+HEAPF32[tempDoublePtr>>2]); //@line 581 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $904 = +HEAPF32[$863>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $905 = +HEAPF32[$852>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $906 = +HEAPF32[$clipPoints2>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $907 = $906 - $902; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $908 = ((($clipPoints2)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $909 = +HEAPF32[$908>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $910 = $909 - $904; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $911 = $907 * $903; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $912 = $910 * $905; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $913 = $911 + $912; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $914 = +HEAPF32[$574>>2]; //@line 595 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $915 = !($913 <= $914); //@line 595 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     if ($915) {
      $978 = $914;$pointCount$1$us = 0;
     } else {
      $952 = +HEAPF32[$28>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
      $953 = $906 - $952; //@line 587 "./libraries/Box2D/Common/b2Math.h"
      $954 = +HEAPF32[$29>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
      $955 = $909 - $954; //@line 588 "./libraries/Box2D/Common/b2Math.h"
      $956 = +HEAPF32[$31>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $957 = $953 * $956; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $958 = +HEAPF32[$30>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $959 = $955 * $958; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $960 = $957 + $959; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $961 = $953 * $958; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      $962 = $956 * $955; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      $963 = $962 - $961; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$manifold>>2] = $960; //@line 601 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $964 = ((($manifold)) + 4|0); //@line 601 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$964>>2] = $963; //@line 601 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $965 = ((($clipPoints2)) + 8|0); //@line 602 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $966 = ((($manifold)) + 16|0); //@line 602 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $967 = HEAP32[$965>>2]|0; //@line 602 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$966>>2] = $967; //@line 602 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $$pre161 = +HEAPF32[$574>>2]; //@line 595 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $978 = $$pre161;$pointCount$1$us = 1;
     }
     $968 = ((($clipPoints2)) + 12|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $969 = +HEAPF32[$968>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $970 = $969 - $902; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $971 = ((($clipPoints2)) + 16|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $972 = +HEAPF32[$971>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $973 = $972 - $904; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $974 = $970 * $903; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $975 = $973 * $905; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $976 = $974 + $975; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $977 = !($976 <= $978); //@line 595 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     if ($977) {
      $pointCount$0$lcssa = $pointCount$1$us;
     } else {
      $1012 = +HEAPF32[$28>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
      $1013 = $969 - $1012; //@line 587 "./libraries/Box2D/Common/b2Math.h"
      $1014 = +HEAPF32[$29>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
      $1015 = $972 - $1014; //@line 588 "./libraries/Box2D/Common/b2Math.h"
      $1016 = +HEAPF32[$31>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $1017 = $1013 * $1016; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $1018 = +HEAPF32[$30>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $1019 = $1015 * $1018; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $1020 = $1017 + $1019; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $1021 = $1013 * $1018; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      $1022 = $1016 * $1015; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      $1023 = $1022 - $1021; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      $1024 = (($manifold) + (($pointCount$1$us*20)|0)|0); //@line 601 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$1024>>2] = $1020; //@line 601 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1025 = (((($manifold) + (($pointCount$1$us*20)|0)|0)) + 4|0); //@line 601 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAPF32[$1025>>2] = $1023; //@line 601 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1026 = ((($clipPoints2)) + 20|0); //@line 602 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1027 = (((($manifold) + (($pointCount$1$us*20)|0)|0)) + 16|0); //@line 602 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1028 = HEAP32[$1026>>2]|0; //@line 602 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$1027>>2] = $1028; //@line 602 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1029 = (($pointCount$1$us) + 1)|0; //@line 613 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $pointCount$0$lcssa = $1029;
     }
    } else {
     $916 = HEAP32[$$pre$phi164Z2D>>2]|0; //@line 584 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $917 = (((($polygonB)) + 84|0) + ($916<<3)|0); //@line 584 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $918 = $917; //@line 584 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $919 = $918; //@line 584 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $920 = HEAP32[$919>>2]|0; //@line 584 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $921 = (($918) + 4)|0; //@line 584 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $922 = $921; //@line 584 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $923 = HEAP32[$922>>2]|0; //@line 584 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $924 = $880; //@line 584 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $925 = $924; //@line 584 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$925>>2] = $920; //@line 584 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $926 = (($924) + 4)|0; //@line 584 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $927 = $926; //@line 584 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$927>>2] = $923; //@line 584 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $928 = ((($manifold)) + 48|0); //@line 585 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $929 = (((($polygonB)) + 20|0) + ($916<<3)|0); //@line 585 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $930 = $929; //@line 585 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $931 = $930; //@line 585 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $932 = HEAP32[$931>>2]|0; //@line 585 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $933 = (($930) + 4)|0; //@line 585 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $934 = $933; //@line 585 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $935 = HEAP32[$934>>2]|0; //@line 585 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $936 = $928; //@line 585 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $937 = $936; //@line 585 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$937>>2] = $932; //@line 585 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $938 = (($936) + 4)|0; //@line 585 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $939 = $938; //@line 585 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     HEAP32[$939>>2] = $935; //@line 585 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $$pre158 = +HEAPF32[$861>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $$pre159 = +HEAPF32[$853>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $940 = +HEAPF32[$863>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $941 = +HEAPF32[$852>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $942 = +HEAPF32[$clipPoints2>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $943 = $942 - $$pre158; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $944 = ((($clipPoints2)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $945 = +HEAPF32[$944>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $946 = $945 - $940; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $947 = $943 * $$pre159; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $948 = $946 * $941; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $949 = $947 + $948; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $950 = +HEAPF32[$574>>2]; //@line 595 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $951 = !($949 <= $950); //@line 595 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     if ($951) {
      $1011 = $950;$pointCount$1 = 0;
     } else {
      $979 = $clipPoints2; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $980 = $979; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $981 = HEAP32[$980>>2]|0; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $982 = (($979) + 4)|0; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $983 = $982; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $984 = HEAP32[$983>>2]|0; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $985 = $manifold; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $986 = $985; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$986>>2] = $981; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $987 = (($985) + 4)|0; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $988 = $987; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$988>>2] = $984; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $989 = ((($clipPoints2)) + 8|0); //@line 607 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $990 = ((($989)) + 3|0); //@line 607 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $991 = HEAP8[$990>>0]|0; //@line 607 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $992 = ((($manifold)) + 16|0); //@line 607 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $993 = ((($992)) + 2|0); //@line 607 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP8[$993>>0] = $991; //@line 607 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $994 = ((($989)) + 2|0); //@line 608 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $995 = HEAP8[$994>>0]|0; //@line 608 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $996 = ((($992)) + 3|0); //@line 608 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP8[$996>>0] = $995; //@line 608 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $997 = ((($989)) + 1|0); //@line 609 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $998 = HEAP8[$997>>0]|0; //@line 609 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP8[$992>>0] = $998; //@line 609 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $999 = HEAP8[$989>>0]|0; //@line 610 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1000 = ((($992)) + 1|0); //@line 610 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP8[$1000>>0] = $999; //@line 610 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $$pre160 = +HEAPF32[$574>>2]; //@line 595 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1011 = $$pre160;$pointCount$1 = 1;
     }
     $1001 = ((($clipPoints2)) + 12|0); //@line 593 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     $1002 = +HEAPF32[$1001>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $1003 = $1002 - $$pre158; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $1004 = ((($clipPoints2)) + 16|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $1005 = +HEAPF32[$1004>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $1006 = $1005 - $940; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $1007 = $1003 * $$pre159; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $1008 = $1006 * $941; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $1009 = $1007 + $1008; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $1010 = !($1009 <= $1011); //@line 595 "libraries/Box2D/Collision/b2CollideEdge.cpp"
     if ($1010) {
      $pointCount$0$lcssa = $pointCount$1;
     } else {
      $1030 = (($manifold) + (($pointCount$1*20)|0)|0); //@line 597 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1031 = $1001; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1032 = $1031; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1033 = HEAP32[$1032>>2]|0; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1034 = (($1031) + 4)|0; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1035 = $1034; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1036 = HEAP32[$1035>>2]|0; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1037 = $1030; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1038 = $1037; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$1038>>2] = $1033; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1039 = (($1037) + 4)|0; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1040 = $1039; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP32[$1040>>2] = $1036; //@line 606 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1041 = ((($clipPoints2)) + 20|0); //@line 607 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1042 = ((($1041)) + 3|0); //@line 607 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1043 = HEAP8[$1042>>0]|0; //@line 607 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1044 = (((($manifold) + (($pointCount$1*20)|0)|0)) + 16|0); //@line 607 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1045 = ((($1044)) + 2|0); //@line 607 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP8[$1045>>0] = $1043; //@line 607 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1046 = ((($1041)) + 2|0); //@line 608 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1047 = HEAP8[$1046>>0]|0; //@line 608 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1048 = ((($1044)) + 3|0); //@line 608 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP8[$1048>>0] = $1047; //@line 608 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1049 = ((($1041)) + 1|0); //@line 609 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1050 = HEAP8[$1049>>0]|0; //@line 609 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP8[$1044>>0] = $1050; //@line 609 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1051 = HEAP8[$1041>>0]|0; //@line 610 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1052 = ((($1044)) + 1|0); //@line 610 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      HEAP8[$1052>>0] = $1051; //@line 610 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $1053 = (($pointCount$1) + 1)|0; //@line 613 "libraries/Box2D/Collision/b2CollideEdge.cpp"
      $pointCount$0$lcssa = $1053;
     }
    }
    HEAP32[$575>>2] = $pointCount$0$lcssa; //@line 617 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   }
  }
 }
 STACKTOP = sp;return; //@line 618 "libraries/Box2D/Collision/b2CollideEdge.cpp"
}
function __ZN12b2EPCollider24ComputePolygonSeparationEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$lcssa = 0.0, $$old = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0.0;
 var $25 = 0.0, $26 = 0, $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0.0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0;
 var $43 = 0, $44 = 0.0, $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0;
 var $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0.0, $8 = 0, $9 = 0, $i$08 = 0, $i$08$lcssa = 0, $or$cond = 0, $or$cond14 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$agg$result>>2] = 0; //@line 642 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $0 = ((($agg$result)) + 4|0); //@line 643 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$0>>2] = -1; //@line 643 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $1 = ((($agg$result)) + 8|0); //@line 644 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAPF32[$1>>2] = -3.4028234663852886E+38; //@line 644 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $2 = ((($this)) + 216|0); //@line 646 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 646 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $4 = ((($this)) + 212|0); //@line 646 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 646 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $6 = ((($this)) + 128|0); //@line 648 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 648 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $8 = ($7|0)>(0); //@line 648 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 if (!($8)) {
  return; //@line 690 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 }
 $9 = ((($this)) + 164|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($this)) + 168|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$11>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($this)) + 172|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $15 = ((($this)) + 176|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $16 = +HEAPF32[$15>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $17 = ((($this)) + 244|0); //@line 656 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $18 = +HEAPF32[$17>>2]; //@line 656 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 $19 = ((($this)) + 228|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $20 = ((($this)) + 232|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($this)) + 236|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($this)) + 240|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $58 = -3.4028234663852886E+38;$i$08 = 0;
 while(1) {
  $23 = (((($this)) + 64|0) + ($i$08<<3)|0); //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $24 = +HEAPF32[$23>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $25 = -$24; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $26 = (((((($this)) + 64|0) + ($i$08<<3)|0)) + 4|0); //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $27 = +HEAPF32[$26>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $28 = -$27; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $29 = (($this) + ($i$08<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $30 = +HEAPF32[$29>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $31 = $30 - $10; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $32 = (((($this) + ($i$08<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $33 = +HEAPF32[$32>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $34 = $33 - $12; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $35 = $31 * $25; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $36 = $34 * $28; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $37 = $35 + $36; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $38 = $30 - $14; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $39 = $33 - $16; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $40 = $38 * $25; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $41 = $39 * $28; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $42 = $40 + $41; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $43 = $37 < $42; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $44 = $43 ? $37 : $42; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $45 = $44 > $18; //@line 656 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  if ($45) {
   $$lcssa = $44;$i$08$lcssa = $i$08;
   break;
  }
  $46 = $3 * $24; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $47 = $5 * $28; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $48 = $46 + $47; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $49 = !($48 >= 0.0); //@line 666 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  if ($49) {
   $59 = +HEAPF32[$19>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $60 = $25 - $59; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $61 = +HEAPF32[$20>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $62 = $28 - $61; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $63 = $60 * $5; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $64 = $62 * $3; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $65 = $63 + $64; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $66 = !($65 < -0.034906588494777679); //@line 675 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $67 = $44 > $58; //@line 681 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $or$cond = $66 & $67; //@line 675 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   if ($or$cond) {
    label = 7;
   } else {
    $70 = $58;
   }
  } else {
   $50 = +HEAPF32[$21>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $51 = $25 - $50; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $52 = +HEAPF32[$22>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $53 = $28 - $52; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $54 = $51 * $5; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $55 = $53 * $3; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $56 = $54 + $55; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $57 = !($56 < -0.034906588494777679); //@line 668 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $$old = $44 > $58; //@line 681 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $or$cond14 = $57 & $$old; //@line 668 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   if ($or$cond14) {
    label = 7;
   } else {
    $70 = $58;
   }
  }
  if ((label|0) == 7) {
   label = 0;
   HEAP32[$agg$result>>2] = 2; //@line 683 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAP32[$0>>2] = $i$08; //@line 684 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   HEAPF32[$1>>2] = $44; //@line 685 "libraries/Box2D/Collision/b2CollideEdge.cpp"
   $70 = $44;
  }
  $68 = (($i$08) + 1)|0; //@line 648 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  $69 = ($68|0)<($7|0); //@line 648 "libraries/Box2D/Collision/b2CollideEdge.cpp"
  if ($69) {
   $58 = $70;$i$08 = $68;
  } else {
   label = 10;
   break;
  }
 }
 if ((label|0) == 10) {
  return; //@line 690 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 }
 HEAP32[$agg$result>>2] = 2; //@line 659 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAP32[$0>>2] = $i$08$lcssa; //@line 660 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 HEAPF32[$1>>2] = $$lcssa; //@line 661 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 return; //@line 690 "libraries/Box2D/Collision/b2CollideEdge.cpp"
}
function __Z23b2CollideEdgeAndPolygonP10b2ManifoldPK11b2EdgeShapeRK11b2TransformPK14b2PolygonShapeS6_($manifold,$edgeA,$xfA,$polygonB,$xfB) {
 $manifold = $manifold|0;
 $edgeA = $edgeA|0;
 $xfA = $xfA|0;
 $polygonB = $polygonB|0;
 $xfB = $xfB|0;
 var $collider = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $collider = sp;
 __ZN12b2EPCollider7CollideEP10b2ManifoldPK11b2EdgeShapeRK11b2TransformPK14b2PolygonShapeS7_($collider,$manifold,$edgeA,$xfA,$polygonB,$xfB); //@line 697 "libraries/Box2D/Collision/b2CollideEdge.cpp"
 STACKTOP = sp;return; //@line 698 "libraries/Box2D/Collision/b2CollideEdge.cpp"
}
function __Z17b2CollidePolygonsP10b2ManifoldPK14b2PolygonShapeRK11b2TransformS3_S6_($manifold,$polyA,$xfA,$polyB,$xfB) {
 $manifold = $manifold|0;
 $polyA = $polyA|0;
 $xfA = $xfA|0;
 $polyB = $polyB|0;
 $xfB = $xfB|0;
 var $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0.0, $109 = 0, $11 = 0.0, $110 = 0.0, $111 = 0, $112 = 0.0, $113 = 0, $114 = 0.0;
 var $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0.0, $130 = 0.0, $131 = 0.0, $132 = 0.0;
 var $133 = 0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0.0, $148 = 0.0, $149 = 0.0, $15 = 0.0, $150 = 0.0;
 var $151 = 0.0, $152 = 0.0, $153 = 0.0, $154 = 0, $155 = 0, $156 = 0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0;
 var $17 = 0.0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0.0, $177 = 0.0, $178 = 0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0.0, $182 = 0.0, $183 = 0, $184 = 0.0, $185 = 0.0, $186 = 0.0, $187 = 0.0;
 var $188 = 0.0, $189 = 0.0, $19 = 0.0, $190 = 0.0, $191 = 0.0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0.0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0.0, $200 = 0.0, $201 = 0.0, $202 = 0.0, $203 = 0.0, $204 = 0;
 var $205 = 0.0, $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0.0, $212 = 0.0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0.0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0.0, $228 = 0.0, $229 = 0, $23 = 0, $230 = 0.0, $231 = 0.0, $232 = 0.0, $233 = 0.0, $234 = 0, $235 = 0.0, $236 = 0.0, $237 = 0.0, $238 = 0.0, $239 = 0.0, $24 = 0.0, $240 = 0.0;
 var $241 = 0.0, $242 = 0.0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0.0, $25 = 0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0.0, $254 = 0.0, $255 = 0.0, $256 = 0.0, $257 = 0, $258 = 0, $259 = 0;
 var $26 = 0.0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0.0, $50 = 0.0;
 var $51 = 0, $52 = 0, $53 = 0.0, $54 = 0.0, $55 = 0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0, $6 = 0.0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0.0, $66 = 0.0, $67 = 0, $68 = 0.0, $69 = 0.0;
 var $7 = 0.0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0.0, $85 = 0.0, $86 = 0, $87 = 0.0;
 var $88 = 0.0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $clipPoints1 = 0, $clipPoints2 = 0, $edge1$0 = 0, $edge1$0$in = 0, $edgeA = 0, $edgeB = 0, $exitcond$i = 0;
 var $flip$0 = 0, $i$06$i = 0, $incidentEdge = 0, $index$0$lcssa$i = 0, $index$08$i = 0, $index$1$i = 0, $localTangent$sroa$0$0 = 0.0, $localTangent$sroa$7$0 = 0.0, $minDot$07$i = 0.0, $minDot$1$i = 0.0, $pointCount$0$lcssa = 0, $pointCount$1 = 0, $pointCount$1$us = 0, $poly1$0 = 0, $poly2$0 = 0, $tangent = 0, $trunc = 0, $trunc21 = 0, $xf1$sroa$0$0 = 0.0, $xf1$sroa$12$0 = 0.0;
 var $xf1$sroa$6$0 = 0.0, $xf1$sroa$8$0 = 0.0, $xf2$sroa$0$0 = 0.0, $xf2$sroa$10$0 = 0.0, $xf2$sroa$13$0 = 0.0, $xf2$sroa$7$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $edgeA = sp + 92|0;
 $edgeB = sp + 88|0;
 $incidentEdge = sp + 64|0;
 $tangent = sp + 56|0;
 $clipPoints1 = sp + 32|0;
 $clipPoints2 = sp + 8|0;
 $0 = sp;
 $1 = ((($manifold)) + 60|0); //@line 120 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 HEAP32[$1>>2] = 0; //@line 120 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 $2 = ((($polyA)) + 8|0); //@line 121 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 121 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 $4 = ((($polyB)) + 8|0); //@line 121 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 121 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 $6 = $3 + $5; //@line 121 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 HEAP32[$edgeA>>2] = 0; //@line 123 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 $7 = (+__ZL19b2FindMaxSeparationPiPK14b2PolygonShapeRK11b2TransformS2_S5_($edgeA,$polyA,$xfA,$polyB,$xfB)); //@line 124 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 $8 = $7 > $6; //@line 125 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 if ($8) {
  STACKTOP = sp;return; //@line 239 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 }
 HEAP32[$edgeB>>2] = 0; //@line 128 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 $9 = (+__ZL19b2FindMaxSeparationPiPK14b2PolygonShapeRK11b2TransformS2_S5_($edgeB,$polyB,$xfB,$polyA,$xfA)); //@line 129 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 $10 = $9 > $6; //@line 130 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 if (!($10)) {
  $11 = $7 + 5.0000002374872565E-4; //@line 140 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $12 = $9 > $11; //@line 140 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $13 = +HEAPF32[$xfB>>2]; //@line 144 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $14 = ((($xfB)) + 4|0); //@line 144 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $15 = +HEAPF32[$14>>2]; //@line 144 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $16 = ((($xfB)) + 8|0); //@line 144 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $17 = +HEAPF32[$16>>2]; //@line 144 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $18 = ((($xfB)) + 12|0); //@line 144 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $19 = +HEAPF32[$18>>2]; //@line 144 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $20 = +HEAPF32[$xfA>>2]; //@line 145 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $21 = ((($xfA)) + 4|0); //@line 145 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $22 = +HEAPF32[$21>>2]; //@line 145 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $23 = ((($xfA)) + 8|0); //@line 145 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $24 = +HEAPF32[$23>>2]; //@line 145 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $25 = ((($xfA)) + 12|0); //@line 145 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $26 = +HEAPF32[$25>>2]; //@line 145 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  if ($12) {
   $$sink = 2;$edge1$0$in = $edgeB;$flip$0 = 1;$poly1$0 = $polyB;$poly2$0 = $polyA;$xf1$sroa$0$0 = $13;$xf1$sroa$12$0 = $19;$xf1$sroa$6$0 = $15;$xf1$sroa$8$0 = $17;$xf2$sroa$0$0 = $20;$xf2$sroa$10$0 = $24;$xf2$sroa$13$0 = $26;$xf2$sroa$7$0 = $22;
  } else {
   $$sink = 1;$edge1$0$in = $edgeA;$flip$0 = 0;$poly1$0 = $polyA;$poly2$0 = $polyB;$xf1$sroa$0$0 = $20;$xf1$sroa$12$0 = $26;$xf1$sroa$6$0 = $22;$xf1$sroa$8$0 = $24;$xf2$sroa$0$0 = $13;$xf2$sroa$10$0 = $17;$xf2$sroa$13$0 = $19;$xf2$sroa$7$0 = $15;
  }
  $edge1$0 = HEAP32[$edge1$0$in>>2]|0; //@line 146 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $27 = ((($manifold)) + 56|0); //@line 157 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  HEAP32[$27>>2] = $$sink; //@line 157 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $28 = ((($poly2$0)) + 148|0); //@line 70 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $29 = HEAP32[$28>>2]|0; //@line 70 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $30 = ($edge1$0|0)>(-1); //@line 74 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  if ($30) {
   $31 = ((($poly1$0)) + 148|0); //@line 74 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
   $32 = HEAP32[$31>>2]|0; //@line 74 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
   $33 = ($32|0)>($edge1$0|0); //@line 74 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
   $34 = $33;
  } else {
   $34 = 0;
  }
  __Z10loveAssertbPKc($34,445484); //@line 74 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $35 = (((($poly1$0)) + 84|0) + ($edge1$0<<3)|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $36 = +HEAPF32[$35>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $37 = $xf1$sroa$12$0 * $36; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $38 = (((((($poly1$0)) + 84|0) + ($edge1$0<<3)|0)) + 4|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $39 = +HEAPF32[$38>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $40 = $xf1$sroa$8$0 * $39; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $41 = $37 - $40; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $42 = $xf1$sroa$8$0 * $36; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $43 = $xf1$sroa$12$0 * $39; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $44 = $42 + $43; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $45 = $xf2$sroa$13$0 * $41; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $46 = $xf2$sroa$10$0 * $44; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $47 = $45 + $46; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $48 = $xf2$sroa$10$0 * $41; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $49 = $xf2$sroa$13$0 * $44; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $50 = $49 - $48; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $51 = ($29|0)>(0); //@line 82 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  if ($51) {
   $i$06$i = 0;$index$08$i = 0;$minDot$07$i = 3.4028234663852886E+38;
   while(1) {
    $52 = (((($poly2$0)) + 84|0) + ($i$06$i<<3)|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $53 = +HEAPF32[$52>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $54 = $47 * $53; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $55 = (((((($poly2$0)) + 84|0) + ($i$06$i<<3)|0)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $56 = +HEAPF32[$55>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $57 = $50 * $56; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $58 = $54 + $57; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $59 = $58 < $minDot$07$i; //@line 85 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
    $minDot$1$i = $59 ? $58 : $minDot$07$i; //@line 85 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
    $index$1$i = $59 ? $i$06$i : $index$08$i; //@line 85 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
    $60 = (($i$06$i) + 1)|0; //@line 82 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
    $exitcond$i = ($60|0)==($29|0); //@line 82 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
    if ($exitcond$i) {
     $index$0$lcssa$i = $index$1$i;
     break;
    } else {
     $i$06$i = $60;$index$08$i = $index$1$i;$minDot$07$i = $minDot$1$i;
    }
   }
  } else {
   $index$0$lcssa$i = 0;
  }
  $61 = (($index$0$lcssa$i) + 1)|0; //@line 94 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $62 = ($61|0)<($29|0); //@line 94 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $63 = $62 ? $61 : 0; //@line 94 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $64 = (((($poly2$0)) + 20|0) + ($index$0$lcssa$i<<3)|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $65 = +HEAPF32[$64>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $66 = $xf2$sroa$13$0 * $65; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $67 = (((((($poly2$0)) + 20|0) + ($index$0$lcssa$i<<3)|0)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $68 = +HEAPF32[$67>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $69 = $xf2$sroa$10$0 * $68; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $70 = $66 - $69; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $71 = $xf2$sroa$0$0 + $70; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $72 = $xf2$sroa$10$0 * $65; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $73 = $xf2$sroa$13$0 * $68; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $74 = $72 + $73; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $75 = $xf2$sroa$7$0 + $74; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$incidentEdge>>2] = $71; //@line 96 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $76 = ((($incidentEdge)) + 4|0); //@line 96 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  HEAPF32[$76>>2] = $75; //@line 96 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $77 = $edge1$0&255; //@line 97 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $78 = ((($incidentEdge)) + 8|0); //@line 97 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  HEAP8[$78>>0] = $77; //@line 97 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $79 = $index$0$lcssa$i&255; //@line 98 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $80 = ((($78)) + 1|0); //@line 98 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  HEAP8[$80>>0] = $79; //@line 98 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $81 = ((($78)) + 2|0); //@line 99 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  HEAP8[$81>>0] = 1; //@line 99 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $82 = ((($78)) + 3|0); //@line 100 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  HEAP8[$82>>0] = 0; //@line 100 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $83 = (((($poly2$0)) + 20|0) + ($63<<3)|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $84 = +HEAPF32[$83>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $85 = $xf2$sroa$13$0 * $84; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $86 = (((((($poly2$0)) + 20|0) + ($63<<3)|0)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $87 = +HEAPF32[$86>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $88 = $xf2$sroa$10$0 * $87; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $89 = $85 - $88; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $90 = $xf2$sroa$0$0 + $89; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $91 = $xf2$sroa$10$0 * $84; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $92 = $xf2$sroa$13$0 * $87; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $93 = $91 + $92; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $94 = $xf2$sroa$7$0 + $93; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $95 = ((($incidentEdge)) + 12|0); //@line 102 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  HEAPF32[$95>>2] = $90; //@line 102 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $96 = ((($incidentEdge)) + 16|0); //@line 102 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  HEAPF32[$96>>2] = $94; //@line 102 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $97 = ((($incidentEdge)) + 20|0); //@line 103 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  HEAP8[$97>>0] = $77; //@line 103 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $98 = $63&255; //@line 104 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $99 = ((($97)) + 1|0); //@line 104 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  HEAP8[$99>>0] = $98; //@line 104 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $100 = ((($97)) + 2|0); //@line 105 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  HEAP8[$100>>0] = 1; //@line 105 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $101 = ((($97)) + 3|0); //@line 106 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  HEAP8[$101>>0] = 0; //@line 106 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $102 = ((($poly1$0)) + 148|0); //@line 164 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $103 = HEAP32[$102>>2]|0; //@line 164 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $104 = (($edge1$0) + 1)|0; //@line 168 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $105 = ($104|0)<($103|0); //@line 168 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $106 = $105 ? $104 : 0; //@line 168 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $107 = (((($poly1$0)) + 20|0) + ($edge1$0<<3)|0); //@line 170 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $108 = +HEAPF32[$107>>2]; //@line 170 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $109 = (((((($poly1$0)) + 20|0) + ($edge1$0<<3)|0)) + 4|0); //@line 170 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $110 = +HEAPF32[$109>>2]; //@line 170 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $111 = (((($poly1$0)) + 20|0) + ($106<<3)|0); //@line 171 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $112 = +HEAPF32[$111>>2]; //@line 171 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $113 = (((((($poly1$0)) + 20|0) + ($106<<3)|0)) + 4|0); //@line 171 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $114 = +HEAPF32[$113>>2]; //@line 171 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $115 = $112 - $108; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $116 = $114 - $110; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $117 = $115 * $115; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $118 = $116 * $116; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $119 = $117 + $118; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $120 = (+Math_sqrt((+$119))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $121 = $120 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
  if ($121) {
   $localTangent$sroa$0$0 = $115;$localTangent$sroa$7$0 = $116;
  } else {
   $122 = 1.0 / $120; //@line 121 "./libraries/Box2D/Common/b2Math.h"
   $123 = $115 * $122; //@line 122 "./libraries/Box2D/Common/b2Math.h"
   $124 = $116 * $122; //@line 123 "./libraries/Box2D/Common/b2Math.h"
   $localTangent$sroa$0$0 = $123;$localTangent$sroa$7$0 = $124;
  }
  $125 = $108 + $112; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $126 = $125 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $127 = $xf1$sroa$12$0 * $localTangent$sroa$0$0; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $128 = $xf1$sroa$8$0 * $localTangent$sroa$7$0; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $129 = $127 - $128; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $130 = $xf1$sroa$8$0 * $localTangent$sroa$0$0; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $131 = $xf1$sroa$12$0 * $localTangent$sroa$7$0; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $132 = $130 + $131; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$tangent>>2] = $129; //@line 59 "./libraries/Box2D/Common/b2Math.h"
  $133 = ((($tangent)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$133>>2] = $132; //@line 59 "./libraries/Box2D/Common/b2Math.h"
  $134 = -$129; //@line 421 "./libraries/Box2D/Common/b2Math.h"
  $135 = $xf1$sroa$12$0 * $108; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $136 = $xf1$sroa$8$0 * $110; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $137 = $135 - $136; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $138 = $xf1$sroa$0$0 + $137; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $139 = $xf1$sroa$8$0 * $108; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $140 = $xf1$sroa$12$0 * $110; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $141 = $139 + $140; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $142 = $xf1$sroa$6$0 + $141; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $143 = $142 * $134; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $144 = $138 * $129; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $145 = $142 * $132; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $146 = $144 + $145; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $147 = -$localTangent$sroa$0$0; //@line 421 "./libraries/Box2D/Common/b2Math.h"
  $148 = $110 + $114; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $149 = $148 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $150 = $138 * $132; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $151 = $150 + $143; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $152 = $6 - $146; //@line 189 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $153 = -$132; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$0>>2] = $134; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $154 = ((($0)) + 4|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$154>>2] = $153; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $155 = (__Z19b2ClipSegmentToLineP12b2ClipVertexPKS_RK6b2Vec2fi($clipPoints1,$incidentEdge,$0,$152,$edge1$0)|0); //@line 198 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $156 = ($155|0)<(2); //@line 200 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  if (!($156)) {
   $157 = $xf1$sroa$12$0 * $112; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $158 = $xf1$sroa$8$0 * $114; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $159 = $157 - $158; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $160 = $xf1$sroa$0$0 + $159; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $161 = $160 * $129; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $162 = $xf1$sroa$8$0 * $112; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $163 = $xf1$sroa$12$0 * $114; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $164 = $162 + $163; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $165 = $xf1$sroa$6$0 + $164; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $166 = $165 * $132; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $167 = $161 + $166; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $168 = $6 + $167; //@line 190 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
   $169 = (__Z19b2ClipSegmentToLineP12b2ClipVertexPKS_RK6b2Vec2fi($clipPoints2,$clipPoints1,$tangent,$168,$106)|0); //@line 204 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
   $170 = ($169|0)<(2); //@line 206 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
   if (!($170)) {
    $171 = ((($manifold)) + 40|0); //@line 212 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
    HEAPF32[$171>>2] = $localTangent$sroa$7$0; //@line 212 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
    $172 = ((($manifold)) + 44|0); //@line 212 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
    HEAPF32[$172>>2] = $147; //@line 212 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
    $173 = ((($manifold)) + 48|0); //@line 213 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
    HEAPF32[$173>>2] = $126; //@line 213 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
    $174 = ((($manifold)) + 52|0); //@line 213 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
    HEAPF32[$174>>2] = $149; //@line 213 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
    $175 = ($flip$0<<24>>24)==(0); //@line 225 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
    $176 = +HEAPF32[$clipPoints2>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $177 = $132 * $176; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $178 = ((($clipPoints2)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $179 = +HEAPF32[$178>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $180 = $179 * $134; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $181 = $177 + $180; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $182 = $181 - $151; //@line 218 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
    $183 = !($182 <= $6); //@line 220 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
    if ($175) {
     if ($183) {
      $pointCount$1$us = 0;
     } else {
      $184 = $176 - $xf2$sroa$0$0; //@line 587 "./libraries/Box2D/Common/b2Math.h"
      $185 = $179 - $xf2$sroa$7$0; //@line 588 "./libraries/Box2D/Common/b2Math.h"
      $186 = $xf2$sroa$13$0 * $184; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $187 = $xf2$sroa$10$0 * $185; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $188 = $186 + $187; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $189 = $xf2$sroa$10$0 * $184; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      $190 = $xf2$sroa$13$0 * $185; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      $191 = $190 - $189; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$manifold>>2] = $188; //@line 223 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $192 = ((($manifold)) + 4|0); //@line 223 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAPF32[$192>>2] = $191; //@line 223 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $193 = ((($clipPoints2)) + 8|0); //@line 224 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $194 = ((($manifold)) + 16|0); //@line 224 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $195 = HEAP32[$193>>2]|0; //@line 224 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAP32[$194>>2] = $195; //@line 224 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $pointCount$1$us = 1;
     }
     $196 = ((($clipPoints2)) + 12|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $197 = +HEAPF32[$196>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $198 = $132 * $197; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $199 = ((($clipPoints2)) + 16|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $200 = +HEAPF32[$199>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $201 = $200 * $134; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $202 = $198 + $201; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $203 = $202 - $151; //@line 218 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
     $204 = !($203 <= $6); //@line 220 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
     if ($204) {
      $pointCount$0$lcssa = $pointCount$1$us;
     } else {
      $235 = $197 - $xf2$sroa$0$0; //@line 587 "./libraries/Box2D/Common/b2Math.h"
      $236 = $200 - $xf2$sroa$7$0; //@line 588 "./libraries/Box2D/Common/b2Math.h"
      $237 = $xf2$sroa$13$0 * $235; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $238 = $xf2$sroa$10$0 * $236; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $239 = $237 + $238; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $240 = $xf2$sroa$10$0 * $235; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      $241 = $xf2$sroa$13$0 * $236; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      $242 = $241 - $240; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      $243 = (($manifold) + (($pointCount$1$us*20)|0)|0); //@line 223 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAPF32[$243>>2] = $239; //@line 223 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $244 = (((($manifold) + (($pointCount$1$us*20)|0)|0)) + 4|0); //@line 223 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAPF32[$244>>2] = $242; //@line 223 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $245 = ((($clipPoints2)) + 20|0); //@line 224 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $246 = (((($manifold) + (($pointCount$1$us*20)|0)|0)) + 16|0); //@line 224 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $247 = HEAP32[$245>>2]|0; //@line 224 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAP32[$246>>2] = $247; //@line 224 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $248 = (($pointCount$1$us) + 1)|0; //@line 234 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $pointCount$0$lcssa = $248;
     }
    } else {
     if ($183) {
      $pointCount$1 = 0;
     } else {
      $205 = $176 - $xf2$sroa$0$0; //@line 587 "./libraries/Box2D/Common/b2Math.h"
      $206 = $179 - $xf2$sroa$7$0; //@line 588 "./libraries/Box2D/Common/b2Math.h"
      $207 = $xf2$sroa$13$0 * $205; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $208 = $xf2$sroa$10$0 * $206; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $209 = $207 + $208; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $210 = $xf2$sroa$10$0 * $205; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      $211 = $xf2$sroa$13$0 * $206; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      $212 = $211 - $210; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$manifold>>2] = $209; //@line 223 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $213 = ((($manifold)) + 4|0); //@line 223 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAPF32[$213>>2] = $212; //@line 223 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $214 = ((($manifold)) + 16|0); //@line 224 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $215 = ((($clipPoints2)) + 8|0); //@line 224 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $216 = HEAP32[$215>>2]|0; //@line 224 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAP32[$214>>2] = $216; //@line 224 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $trunc = $216&255; //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $217 = ((($214)) + 1|0); //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $218 = $216 >>> 8; //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $219 = $218&255; //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $220 = ((($214)) + 2|0); //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $221 = $216 >>> 16; //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $222 = $221&255; //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $223 = ((($214)) + 3|0); //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $224 = $216 >>> 24; //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $225 = $224&255; //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAP8[$214>>0] = $219; //@line 229 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAP8[$217>>0] = $trunc; //@line 230 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAP8[$220>>0] = $225; //@line 231 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAP8[$223>>0] = $222; //@line 232 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $pointCount$1 = 1;
     }
     $226 = ((($clipPoints2)) + 12|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $227 = +HEAPF32[$226>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $228 = $132 * $227; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $229 = ((($clipPoints2)) + 16|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $230 = +HEAPF32[$229>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $231 = $230 * $134; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $232 = $228 + $231; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $233 = $232 - $151; //@line 218 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
     $234 = !($233 <= $6); //@line 220 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
     if ($234) {
      $pointCount$0$lcssa = $pointCount$1;
     } else {
      $249 = $227 - $xf2$sroa$0$0; //@line 587 "./libraries/Box2D/Common/b2Math.h"
      $250 = $230 - $xf2$sroa$7$0; //@line 588 "./libraries/Box2D/Common/b2Math.h"
      $251 = $xf2$sroa$13$0 * $249; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $252 = $xf2$sroa$10$0 * $250; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $253 = $251 + $252; //@line 589 "./libraries/Box2D/Common/b2Math.h"
      $254 = $xf2$sroa$10$0 * $249; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      $255 = $xf2$sroa$13$0 * $250; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      $256 = $255 - $254; //@line 590 "./libraries/Box2D/Common/b2Math.h"
      $257 = (($manifold) + (($pointCount$1*20)|0)|0); //@line 223 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAPF32[$257>>2] = $253; //@line 223 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $258 = (((($manifold) + (($pointCount$1*20)|0)|0)) + 4|0); //@line 223 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAPF32[$258>>2] = $256; //@line 223 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $259 = (((($manifold) + (($pointCount$1*20)|0)|0)) + 16|0); //@line 224 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $260 = ((($clipPoints2)) + 20|0); //@line 224 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $261 = HEAP32[$260>>2]|0; //@line 224 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAP32[$259>>2] = $261; //@line 224 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $trunc21 = $261&255; //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $262 = ((($259)) + 1|0); //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $263 = $261 >>> 8; //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $264 = $263&255; //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $265 = ((($259)) + 2|0); //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $266 = $261 >>> 16; //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $267 = $266&255; //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $268 = ((($259)) + 3|0); //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $269 = $261 >>> 24; //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $270 = $269&255; //@line 228 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAP8[$259>>0] = $264; //@line 229 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAP8[$262>>0] = $trunc21; //@line 230 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAP8[$265>>0] = $270; //@line 231 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      HEAP8[$268>>0] = $267; //@line 232 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $271 = (($pointCount$1) + 1)|0; //@line 234 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
      $pointCount$0$lcssa = $271;
     }
    }
    HEAP32[$1>>2] = $pointCount$0$lcssa; //@line 238 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
   }
  }
 }
 STACKTOP = sp;return; //@line 239 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
}
function __ZL19b2FindMaxSeparationPiPK14b2PolygonShapeRK11b2TransformS2_S5_($edgeIndex,$poly1,$xf1,$poly2,$xf2) {
 $edgeIndex = $edgeIndex|0;
 $poly1 = $poly1|0;
 $xf1 = $xf1|0;
 $poly2 = $poly2|0;
 $xf2 = $xf2|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0.0;
 var $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0;
 var $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0;
 var $63 = 0.0, $64 = 0.0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0.0, $70 = 0, $8 = 0.0, $9 = 0, $bestIndex$0$lcssa = 0, $bestIndex$05 = 0, $bestIndex$05$us = 0, $bestIndex$1 = 0, $bestIndex$1$us = 0, $exitcond = 0, $exitcond12 = 0, $exitcond13 = 0, $i$04 = 0;
 var $i$04$us = 0, $j$02$us = 0, $maxSeparation$0$lcssa = 0.0, $maxSeparation$03 = 0.0, $maxSeparation$03$us = 0.0, $maxSeparation$1 = 0.0, $maxSeparation$1$us = 0.0, $si$01$us = 0.0, $si$1$us = 0.0, $si$1$us$lcssa = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($poly1)) + 148|0); //@line 27 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 27 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 $2 = ((($poly2)) + 148|0); //@line 28 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 28 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 $4 = ((($xf2)) + 12|0); //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($xf1)) + 8|0); //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $8 = $5 * $7; //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($xf2)) + 8|0); //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($xf1)) + 12|0); //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$11>>2]; //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $13 = $10 * $12; //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $14 = $8 - $13; //@line 560 "./libraries/Box2D/Common/b2Math.h"
 $15 = $5 * $12; //@line 561 "./libraries/Box2D/Common/b2Math.h"
 $16 = $7 * $10; //@line 561 "./libraries/Box2D/Common/b2Math.h"
 $17 = $16 + $15; //@line 561 "./libraries/Box2D/Common/b2Math.h"
 $18 = +HEAPF32[$xf1>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $19 = +HEAPF32[$xf2>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $20 = $18 - $19; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($xf1)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $22 = +HEAPF32[$21>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $23 = ((($xf2)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $24 = +HEAPF32[$23>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $25 = $22 - $24; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $26 = $5 * $20; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $27 = $10 * $25; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $28 = $26 + $27; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $29 = $10 * $20; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $30 = $5 * $25; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $31 = $30 - $29; //@line 574 "./libraries/Box2D/Common/b2Math.h"
 $32 = ($1|0)>(0); //@line 36 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 if (!($32)) {
  $bestIndex$0$lcssa = 0;$maxSeparation$0$lcssa = -3.4028234663852886E+38;
  HEAP32[$edgeIndex>>2] = $bestIndex$0$lcssa; //@line 60 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  return (+$maxSeparation$0$lcssa); //@line 61 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 }
 $33 = ($3|0)>(0); //@line 44 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 if ($33) {
  $bestIndex$05$us = 0;$i$04$us = 0;$maxSeparation$03$us = -3.4028234663852886E+38;
 } else {
  $bestIndex$05 = 0;$i$04 = 0;$maxSeparation$03 = -3.4028234663852886E+38;
  while(1) {
   $69 = $maxSeparation$03 < 3.4028234663852886E+38; //@line 53 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
   $maxSeparation$1 = $69 ? 3.4028234663852886E+38 : $maxSeparation$03; //@line 53 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
   $bestIndex$1 = $69 ? $i$04 : $bestIndex$05; //@line 53 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
   $70 = (($i$04) + 1)|0; //@line 36 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
   $exitcond13 = ($70|0)==($1|0); //@line 36 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
   if ($exitcond13) {
    $bestIndex$0$lcssa = $bestIndex$1;$maxSeparation$0$lcssa = $maxSeparation$1;
    break;
   } else {
    $bestIndex$05 = $bestIndex$1;$i$04 = $70;$maxSeparation$03 = $maxSeparation$1;
   }
  }
  HEAP32[$edgeIndex>>2] = $bestIndex$0$lcssa; //@line 60 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  return (+$maxSeparation$0$lcssa); //@line 61 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 }
 while(1) {
  $34 = (((($poly1)) + 84|0) + ($i$04$us<<3)|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $35 = +HEAPF32[$34>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $36 = $17 * $35; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $37 = (((((($poly1)) + 84|0) + ($i$04$us<<3)|0)) + 4|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $38 = +HEAPF32[$37>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $39 = $14 * $38; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $40 = $36 - $39; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $41 = $14 * $35; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $42 = $17 * $38; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $43 = $41 + $42; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $44 = (((($poly1)) + 20|0) + ($i$04$us<<3)|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $45 = +HEAPF32[$44>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $46 = $17 * $45; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $47 = (((((($poly1)) + 20|0) + ($i$04$us<<3)|0)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $48 = +HEAPF32[$47>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $49 = $14 * $48; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $50 = $46 - $49; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $51 = $28 + $50; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $52 = $14 * $45; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $53 = $17 * $48; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $54 = $52 + $53; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $55 = $31 + $54; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $j$02$us = 0;$si$01$us = 3.4028234663852886E+38;
  while(1) {
   $56 = (((($poly2)) + 20|0) + ($j$02$us<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $57 = +HEAPF32[$56>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $58 = $57 - $51; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $59 = (((((($poly2)) + 20|0) + ($j$02$us<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $60 = +HEAPF32[$59>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $61 = $60 - $55; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $62 = $40 * $58; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $63 = $43 * $61; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $64 = $62 + $63; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $65 = $64 < $si$01$us; //@line 47 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
   $si$1$us = $65 ? $64 : $si$01$us; //@line 47 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
   $66 = (($j$02$us) + 1)|0; //@line 44 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
   $exitcond = ($66|0)==($3|0); //@line 44 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
   if ($exitcond) {
    $si$1$us$lcssa = $si$1$us;
    break;
   } else {
    $j$02$us = $66;$si$01$us = $si$1$us;
   }
  }
  $67 = $si$1$us$lcssa > $maxSeparation$03$us; //@line 53 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $maxSeparation$1$us = $67 ? $si$1$us$lcssa : $maxSeparation$03$us; //@line 53 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $bestIndex$1$us = $67 ? $i$04$us : $bestIndex$05$us; //@line 53 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $68 = (($i$04$us) + 1)|0; //@line 36 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  $exitcond12 = ($68|0)==($1|0); //@line 36 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
  if ($exitcond12) {
   $bestIndex$0$lcssa = $bestIndex$1$us;$maxSeparation$0$lcssa = $maxSeparation$1$us;
   break;
  } else {
   $bestIndex$05$us = $bestIndex$1$us;$i$04$us = $68;$maxSeparation$03$us = $maxSeparation$1$us;
  }
 }
 HEAP32[$edgeIndex>>2] = $bestIndex$0$lcssa; //@line 60 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
 return (+$maxSeparation$0$lcssa); //@line 61 "libraries/Box2D/Collision/b2CollidePolygon.cpp"
}
function __ZN15b2WorldManifold10InitializeEPK10b2ManifoldRK11b2TransformfS5_f($this,$manifold,$xfA,$radiusA,$xfB,$radiusB) {
 $this = $this|0;
 $manifold = $manifold|0;
 $xfA = $xfA|0;
 $radiusA = +$radiusA;
 $xfB = $xfB|0;
 $radiusB = +$radiusB;
 var $$pre = 0.0, $$pre51 = 0.0, $$pre55 = 0.0, $$pre56 = 0.0, $0 = 0, $1 = 0, $10 = 0.0, $100 = 0.0, $101 = 0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0, $111 = 0.0;
 var $112 = 0.0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0.0, $119 = 0, $12 = 0.0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0;
 var $130 = 0.0, $131 = 0.0, $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0.0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0.0, $148 = 0.0;
 var $149 = 0.0, $15 = 0.0, $150 = 0.0, $151 = 0.0, $152 = 0.0, $153 = 0.0, $154 = 0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0.0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0.0, $169 = 0, $17 = 0.0, $170 = 0.0, $171 = 0.0, $172 = 0, $173 = 0.0, $174 = 0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0.0, $180 = 0.0, $181 = 0, $182 = 0.0, $183 = 0, $184 = 0.0;
 var $185 = 0.0, $186 = 0.0, $187 = 0, $188 = 0.0, $189 = 0.0, $19 = 0.0, $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0.0, $194 = 0.0, $195 = 0.0, $196 = 0, $197 = 0.0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0.0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0.0, $205 = 0.0, $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0, $21 = 0.0, $210 = 0.0, $211 = 0.0, $212 = 0.0, $213 = 0, $214 = 0.0, $215 = 0.0, $216 = 0.0, $217 = 0.0, $218 = 0.0, $219 = 0.0, $22 = 0;
 var $220 = 0.0, $221 = 0.0, $222 = 0.0, $223 = 0.0, $224 = 0.0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0.0, $229 = 0.0, $23 = 0.0, $230 = 0.0, $231 = 0.0, $232 = 0.0, $233 = 0.0, $234 = 0.0, $235 = 0.0, $236 = 0.0, $237 = 0.0, $238 = 0.0;
 var $239 = 0.0, $24 = 0.0, $240 = 0.0, $241 = 0.0, $242 = 0.0, $243 = 0.0, $244 = 0, $245 = 0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0.0;
 var $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0;
 var $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0.0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0, $74 = 0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0.0, $86 = 0, $87 = 0.0, $88 = 0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0, $98 = 0.0;
 var $99 = 0.0, $i$047 = 0, $i4$048 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($manifold)) + 60|0); //@line 26 "libraries/Box2D/Collision/b2Collision.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 26 "libraries/Box2D/Collision/b2Collision.cpp"
 $2 = ($1|0)==(0); //@line 26 "libraries/Box2D/Collision/b2Collision.cpp"
 if ($2) {
  return; //@line 86 "libraries/Box2D/Collision/b2Collision.cpp"
 }
 $3 = ((($manifold)) + 56|0); //@line 31 "libraries/Box2D/Collision/b2Collision.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 31 "libraries/Box2D/Collision/b2Collision.cpp"
 switch ($4|0) {
 case 0:  {
  HEAPF32[$this>>2] = 1.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $5 = ((($this)) + 4|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$5>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $6 = ((($xfA)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $7 = +HEAPF32[$6>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $8 = ((($manifold)) + 48|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $9 = +HEAPF32[$8>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $10 = $7 * $9; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $11 = ((($xfA)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $12 = +HEAPF32[$11>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $13 = ((($manifold)) + 52|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $15 = $12 * $14; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $16 = $10 - $15; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $17 = +HEAPF32[$xfA>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $18 = $17 + $16; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $19 = $9 * $12; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $20 = $7 * $14; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $21 = $19 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $22 = ((($xfA)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $23 = +HEAPF32[$22>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $24 = $21 + $23; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $25 = ((($xfB)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $26 = +HEAPF32[$25>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $27 = +HEAPF32[$manifold>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $28 = $26 * $27; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $29 = ((($xfB)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $30 = +HEAPF32[$29>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $31 = ((($manifold)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $32 = +HEAPF32[$31>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $33 = $30 * $32; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $34 = $28 - $33; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $35 = +HEAPF32[$xfB>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $36 = $35 + $34; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $37 = $27 * $30; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $38 = $26 * $32; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $39 = $37 + $38; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $40 = ((($xfB)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $41 = +HEAPF32[$40>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $42 = $39 + $41; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $43 = $18 - $36; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $44 = $24 - $42; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $45 = $43 * $43; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $46 = $44 * $44; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $47 = $45 + $46; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $48 = $47 > 1.4210854715202004E-14; //@line 38 "libraries/Box2D/Collision/b2Collision.cpp"
  if ($48) {
   $49 = $36 - $18; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $50 = $42 - $24; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$this>>2] = $49; //@line 40 "libraries/Box2D/Collision/b2Collision.cpp"
   HEAPF32[$5>>2] = $50; //@line 40 "libraries/Box2D/Collision/b2Collision.cpp"
   $51 = $49 * $49; //@line 103 "./libraries/Box2D/Common/b2Math.h"
   $52 = $50 * $50; //@line 103 "./libraries/Box2D/Common/b2Math.h"
   $53 = $51 + $52; //@line 103 "./libraries/Box2D/Common/b2Math.h"
   $54 = (+Math_sqrt((+$53))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
   $55 = $54 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
   if ($55) {
    $60 = $49;$62 = $50;
   } else {
    $56 = 1.0 / $54; //@line 121 "./libraries/Box2D/Common/b2Math.h"
    $57 = $49 * $56; //@line 122 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$this>>2] = $57; //@line 122 "./libraries/Box2D/Common/b2Math.h"
    $58 = $50 * $56; //@line 123 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$5>>2] = $58; //@line 123 "./libraries/Box2D/Common/b2Math.h"
    $60 = $57;$62 = $58;
   }
  } else {
   $60 = 1.0;$62 = 0.0;
  }
  $59 = $60 * $radiusA; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $61 = $62 * $radiusA; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $63 = $18 + $59; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $64 = $24 + $61; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $65 = $60 * $radiusB; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $66 = $62 * $radiusB; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $67 = $36 - $65; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $68 = $42 - $66; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $69 = $63 + $67; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $70 = $64 + $68; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $71 = $69 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $72 = $70 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $73 = ((($this)) + 8|0); //@line 46 "libraries/Box2D/Collision/b2Collision.cpp"
  HEAPF32[$73>>2] = $71; //@line 46 "libraries/Box2D/Collision/b2Collision.cpp"
  $74 = ((($this)) + 12|0); //@line 46 "libraries/Box2D/Collision/b2Collision.cpp"
  HEAPF32[$74>>2] = $72; //@line 46 "libraries/Box2D/Collision/b2Collision.cpp"
  $75 = $67 - $63; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $76 = $68 - $64; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $77 = $75 * $60; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $78 = $76 * $62; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $79 = $77 + $78; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $80 = ((($this)) + 24|0); //@line 47 "libraries/Box2D/Collision/b2Collision.cpp"
  HEAPF32[$80>>2] = $79; //@line 47 "libraries/Box2D/Collision/b2Collision.cpp"
  return; //@line 86 "libraries/Box2D/Collision/b2Collision.cpp"
  break;
 }
 case 1:  {
  $81 = ((($xfA)) + 12|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $82 = +HEAPF32[$81>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $83 = ((($manifold)) + 40|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $84 = +HEAPF32[$83>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $85 = $82 * $84; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $86 = ((($xfA)) + 8|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $87 = +HEAPF32[$86>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $88 = ((($manifold)) + 44|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $89 = +HEAPF32[$88>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $90 = $87 * $89; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $91 = $85 - $90; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $92 = $84 * $87; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $93 = $82 * $89; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $94 = $92 + $93; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$this>>2] = $91; //@line 53 "libraries/Box2D/Collision/b2Collision.cpp"
  $95 = ((($this)) + 4|0); //@line 53 "libraries/Box2D/Collision/b2Collision.cpp"
  HEAPF32[$95>>2] = $94; //@line 53 "libraries/Box2D/Collision/b2Collision.cpp"
  $96 = +HEAPF32[$81>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $97 = ((($manifold)) + 48|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $98 = +HEAPF32[$97>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $99 = $96 * $98; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $100 = +HEAPF32[$86>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $101 = ((($manifold)) + 52|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $102 = +HEAPF32[$101>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $103 = $100 * $102; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $104 = $99 - $103; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $105 = +HEAPF32[$xfA>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $106 = $105 + $104; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $107 = $98 * $100; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $108 = $96 * $102; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $109 = $107 + $108; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $110 = ((($xfA)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $111 = +HEAPF32[$110>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $112 = $109 + $111; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $113 = HEAP32[$0>>2]|0; //@line 56 "libraries/Box2D/Collision/b2Collision.cpp"
  $114 = ($113|0)>(0); //@line 56 "libraries/Box2D/Collision/b2Collision.cpp"
  if (!($114)) {
   return; //@line 86 "libraries/Box2D/Collision/b2Collision.cpp"
  }
  $115 = ((($xfB)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $116 = ((($xfB)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $117 = ((($xfB)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $137 = $91;$139 = $94;$i$047 = 0;
  while(1) {
   $118 = +HEAPF32[$115>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $119 = (($manifold) + (($i$047*20)|0)|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $120 = +HEAPF32[$119>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $121 = $118 * $120; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $122 = +HEAPF32[$116>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $123 = (((($manifold) + (($i$047*20)|0)|0)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $124 = +HEAPF32[$123>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $125 = $122 * $124; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $126 = $121 - $125; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $127 = +HEAPF32[$xfB>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $128 = $127 + $126; //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $129 = $120 * $122; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $130 = $118 * $124; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $131 = $129 + $130; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $132 = +HEAPF32[$117>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $133 = $131 + $132; //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $134 = $128 - $106; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $135 = $133 - $112; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $136 = $137 * $134; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $138 = $135 * $139; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $140 = $136 + $138; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $141 = $radiusA - $140; //@line 59 "libraries/Box2D/Collision/b2Collision.cpp"
   $142 = $137 * $141; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $143 = $139 * $141; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $144 = $128 + $142; //@line 448 "./libraries/Box2D/Common/b2Math.h"
   $145 = $133 + $143; //@line 448 "./libraries/Box2D/Common/b2Math.h"
   $146 = $137 * $radiusB; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $147 = $139 * $radiusB; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $148 = $128 - $146; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $149 = $133 - $147; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $150 = $148 + $144; //@line 448 "./libraries/Box2D/Common/b2Math.h"
   $151 = $149 + $145; //@line 448 "./libraries/Box2D/Common/b2Math.h"
   $152 = $150 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $153 = $151 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $154 = (((($this)) + 8|0) + ($i$047<<3)|0); //@line 61 "libraries/Box2D/Collision/b2Collision.cpp"
   HEAPF32[$154>>2] = $152; //@line 61 "libraries/Box2D/Collision/b2Collision.cpp"
   $155 = (((((($this)) + 8|0) + ($i$047<<3)|0)) + 4|0); //@line 61 "libraries/Box2D/Collision/b2Collision.cpp"
   HEAPF32[$155>>2] = $153; //@line 61 "libraries/Box2D/Collision/b2Collision.cpp"
   $156 = $148 - $144; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $157 = $149 - $145; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $158 = +HEAPF32[$this>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $159 = $158 * $156; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $160 = +HEAPF32[$95>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $161 = $160 * $157; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $162 = $159 + $161; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $163 = (((($this)) + 24|0) + ($i$047<<2)|0); //@line 62 "libraries/Box2D/Collision/b2Collision.cpp"
   HEAPF32[$163>>2] = $162; //@line 62 "libraries/Box2D/Collision/b2Collision.cpp"
   $164 = (($i$047) + 1)|0; //@line 56 "libraries/Box2D/Collision/b2Collision.cpp"
   $165 = HEAP32[$0>>2]|0; //@line 56 "libraries/Box2D/Collision/b2Collision.cpp"
   $166 = ($164|0)<($165|0); //@line 56 "libraries/Box2D/Collision/b2Collision.cpp"
   if (!($166)) {
    break;
   }
   $$pre55 = +HEAPF32[$this>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $$pre56 = +HEAPF32[$95>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $137 = $$pre55;$139 = $$pre56;$i$047 = $164;
  }
  return; //@line 86 "libraries/Box2D/Collision/b2Collision.cpp"
  break;
 }
 case 2:  {
  $167 = ((($xfB)) + 12|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $168 = +HEAPF32[$167>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $169 = ((($manifold)) + 40|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $170 = +HEAPF32[$169>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $171 = $168 * $170; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $172 = ((($xfB)) + 8|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $173 = +HEAPF32[$172>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $174 = ((($manifold)) + 44|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $175 = +HEAPF32[$174>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $176 = $173 * $175; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $177 = $171 - $176; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $178 = $170 * $173; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $179 = $168 * $175; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $180 = $178 + $179; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$this>>2] = $177; //@line 69 "libraries/Box2D/Collision/b2Collision.cpp"
  $181 = ((($this)) + 4|0); //@line 69 "libraries/Box2D/Collision/b2Collision.cpp"
  HEAPF32[$181>>2] = $180; //@line 69 "libraries/Box2D/Collision/b2Collision.cpp"
  $182 = +HEAPF32[$167>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $183 = ((($manifold)) + 48|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $184 = +HEAPF32[$183>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $185 = $182 * $184; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $186 = +HEAPF32[$172>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $187 = ((($manifold)) + 52|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $188 = +HEAPF32[$187>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $189 = $186 * $188; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $190 = $185 - $189; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $191 = +HEAPF32[$xfB>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $192 = $191 + $190; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $193 = $184 * $186; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $194 = $182 * $188; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $195 = $193 + $194; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $196 = ((($xfB)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $197 = +HEAPF32[$196>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $198 = $195 + $197; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $199 = HEAP32[$0>>2]|0; //@line 72 "libraries/Box2D/Collision/b2Collision.cpp"
  $200 = ($199|0)>(0); //@line 72 "libraries/Box2D/Collision/b2Collision.cpp"
  if ($200) {
   $201 = ((($xfA)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $202 = ((($xfA)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
   $203 = ((($xfA)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
   $227 = $177;$229 = $180;$i4$048 = 0;
   while(1) {
    $208 = +HEAPF32[$201>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
    $209 = (($manifold) + (($i4$048*20)|0)|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
    $210 = +HEAPF32[$209>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
    $211 = $208 * $210; //@line 579 "./libraries/Box2D/Common/b2Math.h"
    $212 = +HEAPF32[$202>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
    $213 = (((($manifold) + (($i4$048*20)|0)|0)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
    $214 = +HEAPF32[$213>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
    $215 = $212 * $214; //@line 579 "./libraries/Box2D/Common/b2Math.h"
    $216 = $211 - $215; //@line 579 "./libraries/Box2D/Common/b2Math.h"
    $217 = +HEAPF32[$xfA>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
    $218 = $217 + $216; //@line 579 "./libraries/Box2D/Common/b2Math.h"
    $219 = $210 * $212; //@line 580 "./libraries/Box2D/Common/b2Math.h"
    $220 = $208 * $214; //@line 580 "./libraries/Box2D/Common/b2Math.h"
    $221 = $219 + $220; //@line 580 "./libraries/Box2D/Common/b2Math.h"
    $222 = +HEAPF32[$203>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
    $223 = $221 + $222; //@line 580 "./libraries/Box2D/Common/b2Math.h"
    $224 = $218 - $192; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $225 = $223 - $198; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $226 = $227 * $224; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $228 = $225 * $229; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $230 = $226 + $228; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $231 = $radiusB - $230; //@line 75 "libraries/Box2D/Collision/b2Collision.cpp"
    $232 = $227 * $231; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $233 = $229 * $231; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $234 = $218 + $232; //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $235 = $223 + $233; //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $236 = $227 * $radiusA; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $237 = $229 * $radiusA; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $238 = $218 - $236; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $239 = $223 - $237; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $240 = $238 + $234; //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $241 = $239 + $235; //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $242 = $240 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $243 = $241 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $244 = (((($this)) + 8|0) + ($i4$048<<3)|0); //@line 77 "libraries/Box2D/Collision/b2Collision.cpp"
    HEAPF32[$244>>2] = $242; //@line 77 "libraries/Box2D/Collision/b2Collision.cpp"
    $245 = (((((($this)) + 8|0) + ($i4$048<<3)|0)) + 4|0); //@line 77 "libraries/Box2D/Collision/b2Collision.cpp"
    HEAPF32[$245>>2] = $243; //@line 77 "libraries/Box2D/Collision/b2Collision.cpp"
    $246 = $238 - $234; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $247 = $239 - $235; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $248 = +HEAPF32[$this>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $249 = $248 * $246; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $250 = +HEAPF32[$181>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $251 = $250 * $247; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $252 = $249 + $251; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $253 = (((($this)) + 24|0) + ($i4$048<<2)|0); //@line 78 "libraries/Box2D/Collision/b2Collision.cpp"
    HEAPF32[$253>>2] = $252; //@line 78 "libraries/Box2D/Collision/b2Collision.cpp"
    $254 = (($i4$048) + 1)|0; //@line 72 "libraries/Box2D/Collision/b2Collision.cpp"
    $255 = HEAP32[$0>>2]|0; //@line 72 "libraries/Box2D/Collision/b2Collision.cpp"
    $256 = ($254|0)<($255|0); //@line 72 "libraries/Box2D/Collision/b2Collision.cpp"
    $$pre = +HEAPF32[$this>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $$pre51 = +HEAPF32[$181>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    if ($256) {
     $227 = $$pre;$229 = $$pre51;$i4$048 = $254;
    } else {
     $205 = $$pre;$207 = $$pre51;
     break;
    }
   }
  } else {
   $205 = $177;$207 = $180;
  }
  $204 = -$205; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $206 = -$207; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$this>>2] = $204; //@line 82 "libraries/Box2D/Collision/b2Collision.cpp"
  HEAPF32[$181>>2] = $206; //@line 82 "libraries/Box2D/Collision/b2Collision.cpp"
  return; //@line 86 "libraries/Box2D/Collision/b2Collision.cpp"
  break;
 }
 default: {
  return; //@line 86 "libraries/Box2D/Collision/b2Collision.cpp"
 }
 }
}
function __Z19b2ClipSegmentToLineP12b2ClipVertexPKS_RK6b2Vec2fi($vOut,$vIn,$normal,$offset,$vertexIndexA) {
 $vOut = $vOut|0;
 $vIn = $vIn|0;
 $normal = $normal|0;
 $offset = +$offset;
 $vertexIndexA = $vertexIndexA|0;
 var $0 = 0.0, $1 = 0.0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0.0;
 var $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $numOut$0 = 0, $numOut$1 = 0, $numOut$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = +HEAPF32[$normal>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $1 = +HEAPF32[$vIn>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $2 = $0 * $1; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($normal)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$3>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($vIn)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $7 = $4 * $6; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $8 = $2 + $7; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $9 = $8 - $offset; //@line 208 "libraries/Box2D/Collision/b2Collision.cpp"
 $10 = ((($vIn)) + 12|0); //@line 209 "libraries/Box2D/Collision/b2Collision.cpp"
 $11 = +HEAPF32[$10>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $12 = $0 * $11; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($vIn)) + 16|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $15 = $4 * $14; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $16 = $12 + $15; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $17 = $16 - $offset; //@line 209 "libraries/Box2D/Collision/b2Collision.cpp"
 $18 = !($9 <= 0.0); //@line 212 "libraries/Box2D/Collision/b2Collision.cpp"
 if ($18) {
  $numOut$0 = 0;
 } else {
  ;HEAP32[$vOut>>2]=HEAP32[$vIn>>2]|0;HEAP32[$vOut+4>>2]=HEAP32[$vIn+4>>2]|0;HEAP32[$vOut+8>>2]=HEAP32[$vIn+8>>2]|0; //@line 212 "libraries/Box2D/Collision/b2Collision.cpp"
  $numOut$0 = 1;
 }
 $19 = !($17 <= 0.0); //@line 213 "libraries/Box2D/Collision/b2Collision.cpp"
 if ($19) {
  $numOut$1 = $numOut$0;
 } else {
  $20 = (($numOut$0) + 1)|0; //@line 213 "libraries/Box2D/Collision/b2Collision.cpp"
  $21 = (($vOut) + (($numOut$0*12)|0)|0); //@line 213 "libraries/Box2D/Collision/b2Collision.cpp"
  ;HEAP32[$21>>2]=HEAP32[$10>>2]|0;HEAP32[$21+4>>2]=HEAP32[$10+4>>2]|0;HEAP32[$21+8>>2]=HEAP32[$10+8>>2]|0; //@line 213 "libraries/Box2D/Collision/b2Collision.cpp"
  $numOut$1 = $20;
 }
 $22 = $9 * $17; //@line 216 "libraries/Box2D/Collision/b2Collision.cpp"
 $23 = $22 < 0.0; //@line 216 "libraries/Box2D/Collision/b2Collision.cpp"
 if (!($23)) {
  $numOut$2 = $numOut$1;
  return ($numOut$2|0); //@line 230 "libraries/Box2D/Collision/b2Collision.cpp"
 }
 $24 = $9 - $17; //@line 219 "libraries/Box2D/Collision/b2Collision.cpp"
 $25 = $9 / $24; //@line 219 "libraries/Box2D/Collision/b2Collision.cpp"
 $26 = +HEAPF32[$10>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $27 = +HEAPF32[$vIn>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $28 = $26 - $27; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $29 = +HEAPF32[$13>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $30 = +HEAPF32[$5>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $31 = $29 - $30; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $32 = $25 * $28; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $33 = $25 * $31; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $34 = $27 + $32; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $35 = $30 + $33; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $36 = (($vOut) + (($numOut$1*12)|0)|0); //@line 220 "libraries/Box2D/Collision/b2Collision.cpp"
 HEAPF32[$36>>2] = $34; //@line 220 "libraries/Box2D/Collision/b2Collision.cpp"
 $37 = (((($vOut) + (($numOut$1*12)|0)|0)) + 4|0); //@line 220 "libraries/Box2D/Collision/b2Collision.cpp"
 HEAPF32[$37>>2] = $35; //@line 220 "libraries/Box2D/Collision/b2Collision.cpp"
 $38 = $vertexIndexA&255; //@line 223 "libraries/Box2D/Collision/b2Collision.cpp"
 $39 = (((($vOut) + (($numOut$1*12)|0)|0)) + 8|0); //@line 223 "libraries/Box2D/Collision/b2Collision.cpp"
 HEAP8[$39>>0] = $38; //@line 223 "libraries/Box2D/Collision/b2Collision.cpp"
 $40 = ((($vIn)) + 8|0); //@line 224 "libraries/Box2D/Collision/b2Collision.cpp"
 $41 = ((($40)) + 1|0); //@line 224 "libraries/Box2D/Collision/b2Collision.cpp"
 $42 = HEAP8[$41>>0]|0; //@line 224 "libraries/Box2D/Collision/b2Collision.cpp"
 $43 = ((($39)) + 1|0); //@line 224 "libraries/Box2D/Collision/b2Collision.cpp"
 HEAP8[$43>>0] = $42; //@line 224 "libraries/Box2D/Collision/b2Collision.cpp"
 $44 = ((($39)) + 2|0); //@line 225 "libraries/Box2D/Collision/b2Collision.cpp"
 HEAP8[$44>>0] = 0; //@line 225 "libraries/Box2D/Collision/b2Collision.cpp"
 $45 = ((($39)) + 3|0); //@line 226 "libraries/Box2D/Collision/b2Collision.cpp"
 HEAP8[$45>>0] = 1; //@line 226 "libraries/Box2D/Collision/b2Collision.cpp"
 $46 = (($numOut$1) + 1)|0; //@line 227 "libraries/Box2D/Collision/b2Collision.cpp"
 $numOut$2 = $46;
 return ($numOut$2|0); //@line 230 "libraries/Box2D/Collision/b2Collision.cpp"
}
function __Z13b2TestOverlapPK7b2ShapeiS1_iRK11b2TransformS4_($shapeA,$indexA,$shapeB,$indexB,$xfA,$xfB) {
 $shapeA = $shapeA|0;
 $indexA = $indexA|0;
 $shapeB = $shapeB|0;
 $indexB = $indexB|0;
 $xfA = $xfA|0;
 $xfB = $xfB|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cache = 0, $input = 0, $output = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $input = sp + 36|0;
 $cache = sp + 24|0;
 $output = sp;
 $0 = ((($input)) + 16|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAP32[$0>>2] = 0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $1 = ((($input)) + 20|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAP32[$1>>2] = 0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $2 = ((($input)) + 24|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAPF32[$2>>2] = 0.0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $3 = ((($input)) + 44|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAP32[$3>>2] = 0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $4 = ((($input)) + 48|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAP32[$4>>2] = 0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 $5 = ((($input)) + 52|0); //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 HEAPF32[$5>>2] = 0.0; //@line 31 "./libraries/Box2D/Collision/b2Distance.h"
 __ZN15b2DistanceProxy3SetEPK7b2Shapei($input,$shapeA,$indexA); //@line 238 "libraries/Box2D/Collision/b2Collision.cpp"
 $6 = ((($input)) + 28|0); //@line 239 "libraries/Box2D/Collision/b2Collision.cpp"
 __ZN15b2DistanceProxy3SetEPK7b2Shapei($6,$shapeB,$indexB); //@line 239 "libraries/Box2D/Collision/b2Collision.cpp"
 $7 = ((($input)) + 56|0); //@line 240 "libraries/Box2D/Collision/b2Collision.cpp"
 ;HEAP32[$7>>2]=HEAP32[$xfA>>2]|0;HEAP32[$7+4>>2]=HEAP32[$xfA+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$xfA+8>>2]|0;HEAP32[$7+12>>2]=HEAP32[$xfA+12>>2]|0; //@line 240 "libraries/Box2D/Collision/b2Collision.cpp"
 $8 = ((($input)) + 72|0); //@line 241 "libraries/Box2D/Collision/b2Collision.cpp"
 ;HEAP32[$8>>2]=HEAP32[$xfB>>2]|0;HEAP32[$8+4>>2]=HEAP32[$xfB+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$xfB+8>>2]|0;HEAP32[$8+12>>2]=HEAP32[$xfB+12>>2]|0; //@line 241 "libraries/Box2D/Collision/b2Collision.cpp"
 $9 = ((($input)) + 88|0); //@line 242 "libraries/Box2D/Collision/b2Collision.cpp"
 HEAP8[$9>>0] = 1; //@line 242 "libraries/Box2D/Collision/b2Collision.cpp"
 $10 = ((($cache)) + 4|0); //@line 245 "libraries/Box2D/Collision/b2Collision.cpp"
 HEAP16[$10>>1] = 0; //@line 245 "libraries/Box2D/Collision/b2Collision.cpp"
 __Z10b2DistanceP16b2DistanceOutputP14b2SimplexCachePK15b2DistanceInput($output,$cache,$input); //@line 249 "libraries/Box2D/Collision/b2Collision.cpp"
 $11 = ((($output)) + 16|0); //@line 251 "libraries/Box2D/Collision/b2Collision.cpp"
 $12 = +HEAPF32[$11>>2]; //@line 251 "libraries/Box2D/Collision/b2Collision.cpp"
 $13 = $12 < 1.1920928955078125E-6; //@line 251 "libraries/Box2D/Collision/b2Collision.cpp"
 STACKTOP = sp;return ($13|0); //@line 251 "libraries/Box2D/Collision/b2Collision.cpp"
}
function __ZN13b2DynamicTreeC2Ev($this) {
 $this = $this|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = -1; //@line 24 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $0 = ((($this)) + 12|0); //@line 26 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$0>>2] = 16; //@line 26 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $1 = ((($this)) + 8|0); //@line 27 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$1>>2] = 0; //@line 27 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $2 = (__Z7b2Alloci(576)|0); //@line 28 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $3 = ((($this)) + 4|0); //@line 28 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$3>>2] = $2; //@line 28 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $4 = HEAP32[$0>>2]|0; //@line 29 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $5 = ($4*36)|0; //@line 29 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 _memset(($2|0),0,($5|0))|0; //@line 29 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 32 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $7 = (($6) + -1)|0; //@line 32 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $8 = ($6|0)>(1); //@line 32 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($8) {
  $9 = HEAP32[$3>>2]|0; //@line 34 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $i$01 = 0;
  while(1) {
   $18 = (($i$01) + 1)|0; //@line 34 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $19 = (((($9) + (($i$01*36)|0)|0)) + 20|0); //@line 34 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   HEAP32[$19>>2] = $18; //@line 34 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $20 = (((($9) + (($i$01*36)|0)|0)) + 32|0); //@line 35 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   HEAP32[$20>>2] = -1; //@line 35 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $21 = HEAP32[$0>>2]|0; //@line 32 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $22 = (($21) + -1)|0; //@line 32 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $23 = ($18|0)<($22|0); //@line 32 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   if ($23) {
    $i$01 = $18;
   } else {
    $$lcssa = $22;
    break;
   }
  }
 } else {
  $$lcssa = $7;
 }
 $10 = HEAP32[$3>>2]|0; //@line 37 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $11 = (((($10) + (($$lcssa*36)|0)|0)) + 20|0); //@line 37 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$11>>2] = -1; //@line 37 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $12 = HEAP32[$0>>2]|0; //@line 38 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $13 = (($12) + -1)|0; //@line 38 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $14 = (((($10) + (($13*36)|0)|0)) + 32|0); //@line 38 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$14>>2] = -1; //@line 38 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $15 = ((($this)) + 16|0); //@line 39 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$15>>2] = 0; //@line 39 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $16 = ((($this)) + 20|0); //@line 41 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$16>>2] = 0; //@line 41 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $17 = ((($this)) + 24|0); //@line 43 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$17>>2] = 0; //@line 43 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 return; //@line 44 "libraries/Box2D/Collision/b2DynamicTree.cpp"
}
function __ZN13b2DynamicTreeD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 49 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 49 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 __THREW__ = 0;
 invoke_vi(3816,($1|0)); //@line 49 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  ___clang_call_terminate($4); //@line 49 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  // unreachable; //@line 49 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 } else {
  return; //@line 50 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 }
}
function __ZN13b2DynamicTree12AllocateNodeEv($this) {
 $this = $this|0;
 var $$lcssa = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phi5Z2D = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0); //@line 56 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 56 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $2 = ($1|0)==(-1); //@line 56 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($2) {
  $3 = ((($this)) + 8|0); //@line 58 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 58 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $5 = ((($this)) + 12|0); //@line 58 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 58 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $7 = ($4|0)==($6|0); //@line 58 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  __Z10loveAssertbPKc($7,445521); //@line 58 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $8 = ((($this)) + 4|0); //@line 61 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 61 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $10 = HEAP32[$5>>2]|0; //@line 62 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $11 = $10 << 1; //@line 62 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$5>>2] = $11; //@line 62 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $12 = ($10*72)|0; //@line 63 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $13 = (__Z7b2Alloci($12)|0); //@line 63 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$8>>2] = $13; //@line 63 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $14 = HEAP32[$3>>2]|0; //@line 64 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $15 = ($14*36)|0; //@line 64 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  _memcpy(($13|0),($9|0),($15|0))|0; //@line 64 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  __Z6b2FreePv($9); //@line 65 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $16 = HEAP32[$3>>2]|0; //@line 69 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $17 = HEAP32[$5>>2]|0; //@line 69 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $18 = (($17) + -1)|0; //@line 69 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $19 = ($16|0)<($18|0); //@line 69 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  if ($19) {
   $20 = HEAP32[$8>>2]|0; //@line 71 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $i$01 = $16;
   while(1) {
    $27 = (($i$01) + 1)|0; //@line 71 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $28 = (((($20) + (($i$01*36)|0)|0)) + 20|0); //@line 71 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    HEAP32[$28>>2] = $27; //@line 71 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $29 = (((($20) + (($i$01*36)|0)|0)) + 32|0); //@line 72 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    HEAP32[$29>>2] = -1; //@line 72 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $30 = HEAP32[$5>>2]|0; //@line 69 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $31 = (($30) + -1)|0; //@line 69 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $32 = ($27|0)<($31|0); //@line 69 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    if ($32) {
     $i$01 = $27;
    } else {
     $$lcssa = $31;
     break;
    }
   }
  } else {
   $$lcssa = $18;
  }
  $21 = HEAP32[$8>>2]|0; //@line 74 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $22 = (((($21) + (($$lcssa*36)|0)|0)) + 20|0); //@line 74 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$22>>2] = -1; //@line 74 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $23 = HEAP32[$5>>2]|0; //@line 75 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $24 = (($23) + -1)|0; //@line 75 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $25 = (((($21) + (($24*36)|0)|0)) + 32|0); //@line 75 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$25>>2] = -1; //@line 75 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $26 = HEAP32[$3>>2]|0; //@line 76 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$0>>2] = $26; //@line 76 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $$pre$phi5Z2D = $3;$34 = $21;$35 = $26;
 } else {
  $$phi$trans$insert = ((($this)) + 4|0); //@line 81 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 81 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $$pre4 = ((($this)) + 8|0); //@line 87 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $$pre$phi5Z2D = $$pre4;$34 = $$pre;$35 = $1;
 }
 $33 = (((($34) + (($35*36)|0)|0)) + 20|0); //@line 81 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $36 = HEAP32[$33>>2]|0; //@line 81 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$0>>2] = $36; //@line 81 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$33>>2] = -1; //@line 82 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $37 = (((($34) + (($35*36)|0)|0)) + 24|0); //@line 83 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$37>>2] = -1; //@line 83 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $38 = (((($34) + (($35*36)|0)|0)) + 28|0); //@line 84 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$38>>2] = -1; //@line 84 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $39 = (((($34) + (($35*36)|0)|0)) + 32|0); //@line 85 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$39>>2] = 0; //@line 85 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $40 = (((($34) + (($35*36)|0)|0)) + 16|0); //@line 86 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$40>>2] = 0; //@line 86 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $41 = HEAP32[$$pre$phi5Z2D>>2]|0; //@line 87 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $42 = (($41) + 1)|0; //@line 87 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$$pre$phi5Z2D>>2] = $42; //@line 87 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 return ($35|0); //@line 88 "libraries/Box2D/Collision/b2DynamicTree.cpp"
}
function __ZN13b2DynamicTree11CreateProxyERK6b2AABBPv($this,$aabb,$userData) {
 $this = $this|0;
 $aabb = $aabb|0;
 $userData = $userData|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0.0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN13b2DynamicTree12AllocateNodeEv($this)|0); //@line 107 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $1 = ((($this)) + 4|0); //@line 111 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 111 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $3 = +HEAPF32[$aabb>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $4 = $3 + -0.10000000149011612; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($aabb)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $7 = $6 + -0.10000000149011612; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $8 = (($2) + (($0*36)|0)|0); //@line 111 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAPF32[$8>>2] = $4; //@line 111 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $9 = (((($2) + (($0*36)|0)|0)) + 4|0); //@line 111 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAPF32[$9>>2] = $7; //@line 111 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $10 = HEAP32[$1>>2]|0; //@line 112 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $11 = ((($aabb)) + 8|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$11>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $13 = $12 + 0.10000000149011612; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $14 = ((($aabb)) + 12|0); //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $15 = +HEAPF32[$14>>2]; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $16 = $15 + 0.10000000149011612; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $17 = (((($10) + (($0*36)|0)|0)) + 8|0); //@line 112 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAPF32[$17>>2] = $13; //@line 112 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $18 = (((($10) + (($0*36)|0)|0)) + 12|0); //@line 112 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAPF32[$18>>2] = $16; //@line 112 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $19 = HEAP32[$1>>2]|0; //@line 113 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $20 = (((($19) + (($0*36)|0)|0)) + 16|0); //@line 113 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$20>>2] = $userData; //@line 113 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $21 = (((($19) + (($0*36)|0)|0)) + 32|0); //@line 114 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$21>>2] = 0; //@line 114 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 __ZN13b2DynamicTree10InsertLeafEi($this,$0); //@line 116 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 return ($0|0); //@line 118 "libraries/Box2D/Collision/b2DynamicTree.cpp"
}
function __ZN13b2DynamicTree10InsertLeafEi($this,$leaf) {
 $this = $this|0;
 $leaf = $leaf|0;
 var $$ = 0, $$pn = 0.0, $$pn1 = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0.0, $110 = 0.0, $111 = 0.0, $112 = 0.0;
 var $113 = 0.0, $114 = 0.0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0.0, $129 = 0, $13 = 0.0, $130 = 0.0;
 var $131 = 0, $132 = 0.0, $133 = 0, $134 = 0.0, $135 = 0, $136 = 0, $137 = 0, $138 = 0.0, $139 = 0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0, $144 = 0.0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0.0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0.0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0.0, $189 = 0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0.0;
 var $203 = 0, $204 = 0.0, $205 = 0, $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0.0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $22 = 0, $23 = 0, $24 = 0.0, $25 = 0;
 var $26 = 0.0, $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0;
 var $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0;
 var $62 = 0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0;
 var $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0.0, $88 = 0, $89 = 0.0, $9 = 0, $90 = 0, $91 = 0.0, $92 = 0, $93 = 0.0, $94 = 0, $95 = 0.0, $96 = 0, $97 = 0.0, $98 = 0;
 var $99 = 0.0, $cost1$0 = 0.0, $cost2$0 = 0.0, $index$0$lcssa = 0, $index$04 = 0, $index$4 = 0, $index$42 = 0, $index$43 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 178 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 178 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $2 = (($1) + 1)|0; //@line 178 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$0>>2] = $2; //@line 178 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $3 = HEAP32[$this>>2]|0; //@line 180 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $4 = ($3|0)==(-1); //@line 180 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($4) {
  HEAP32[$this>>2] = $leaf; //@line 182 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $5 = ((($this)) + 4|0); //@line 183 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 183 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $7 = (((($6) + (($leaf*36)|0)|0)) + 20|0); //@line 183 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$7>>2] = -1; //@line 183 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  return; //@line 314 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 }
 $8 = ((($this)) + 4|0); //@line 188 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 261 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $10 = (($9) + (($leaf*36)|0)|0); //@line 188 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $11 = +HEAPF32[$10>>2]; //@line 188 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $12 = (((($9) + (($leaf*36)|0)|0)) + 4|0); //@line 188 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $13 = +HEAPF32[$12>>2]; //@line 188 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $14 = (((($9) + (($leaf*36)|0)|0)) + 8|0); //@line 188 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 188 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $16 = (((($9) + (($leaf*36)|0)|0)) + 12|0); //@line 188 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 188 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $18 = (((($9) + (($3*36)|0)|0)) + 24|0); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $19 = HEAP32[$18>>2]|0; //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $20 = ($19|0)==(-1); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
 L5: do {
  if ($20) {
   $index$0$lcssa = $3;
  } else {
   $51 = $19;$index$04 = $3;
   while(1) {
    $21 = (((($9) + (($index$04*36)|0)|0)) + 28|0); //@line 193 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $22 = HEAP32[$21>>2]|0; //@line 193 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $23 = (((($9) + (($index$04*36)|0)|0)) + 8|0); //@line 182 "./libraries/Box2D/Collision/b2Collision.h"
    $24 = +HEAPF32[$23>>2]; //@line 182 "./libraries/Box2D/Collision/b2Collision.h"
    $25 = (($9) + (($index$04*36)|0)|0); //@line 182 "./libraries/Box2D/Collision/b2Collision.h"
    $26 = +HEAPF32[$25>>2]; //@line 182 "./libraries/Box2D/Collision/b2Collision.h"
    $27 = $24 - $26; //@line 182 "./libraries/Box2D/Collision/b2Collision.h"
    $28 = (((($9) + (($index$04*36)|0)|0)) + 12|0); //@line 183 "./libraries/Box2D/Collision/b2Collision.h"
    $29 = +HEAPF32[$28>>2]; //@line 183 "./libraries/Box2D/Collision/b2Collision.h"
    $30 = (((($9) + (($index$04*36)|0)|0)) + 4|0); //@line 183 "./libraries/Box2D/Collision/b2Collision.h"
    $31 = +HEAPF32[$30>>2]; //@line 183 "./libraries/Box2D/Collision/b2Collision.h"
    $32 = $29 - $31; //@line 183 "./libraries/Box2D/Collision/b2Collision.h"
    $33 = $27 + $32; //@line 184 "./libraries/Box2D/Collision/b2Collision.h"
    $34 = $33 * 2.0; //@line 184 "./libraries/Box2D/Collision/b2Collision.h"
    $35 = $26 < $11; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $36 = $35 ? $26 : $11; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $37 = $31 < $13; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $38 = $37 ? $31 : $13; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $39 = $24 > $15; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $40 = $39 ? $24 : $15; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $41 = $29 > $17; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $42 = $41 ? $29 : $17; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $43 = $40 - $36; //@line 182 "./libraries/Box2D/Collision/b2Collision.h"
    $44 = $42 - $38; //@line 183 "./libraries/Box2D/Collision/b2Collision.h"
    $45 = $43 + $44; //@line 184 "./libraries/Box2D/Collision/b2Collision.h"
    $46 = $45 * 2.0; //@line 184 "./libraries/Box2D/Collision/b2Collision.h"
    $47 = $46 * 2.0; //@line 202 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $48 = $46 - $34; //@line 205 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $49 = $48 * 2.0; //@line 205 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $50 = (((($9) + (($51*36)|0)|0)) + 24|0); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $52 = HEAP32[$50>>2]|0; //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $53 = ($52|0)==(-1); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $54 = (($9) + (($51*36)|0)|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
    $55 = +HEAPF32[$54>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
    $56 = $11 < $55; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $57 = $56 ? $11 : $55; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $58 = (((($9) + (($51*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
    $59 = +HEAPF32[$58>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
    $60 = $13 < $59; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $61 = $60 ? $13 : $59; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $62 = (((($9) + (($51*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
    $63 = +HEAPF32[$62>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
    $64 = $15 > $63; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $65 = $64 ? $15 : $63; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $66 = (((($9) + (($51*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
    $67 = +HEAPF32[$66>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
    $68 = $17 > $67; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $69 = $68 ? $17 : $67; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    if ($53) {
     $70 = $65 - $57; //@line 182 "./libraries/Box2D/Collision/b2Collision.h"
     $71 = $69 - $61; //@line 183 "./libraries/Box2D/Collision/b2Collision.h"
     $72 = $70 + $71; //@line 184 "./libraries/Box2D/Collision/b2Collision.h"
     $73 = $72 * 2.0; //@line 184 "./libraries/Box2D/Collision/b2Collision.h"
     $$pn = $73;
    } else {
     $74 = $63 - $55; //@line 182 "./libraries/Box2D/Collision/b2Collision.h"
     $75 = $67 - $59; //@line 183 "./libraries/Box2D/Collision/b2Collision.h"
     $76 = $74 + $75; //@line 184 "./libraries/Box2D/Collision/b2Collision.h"
     $77 = $76 * 2.0; //@line 184 "./libraries/Box2D/Collision/b2Collision.h"
     $78 = $65 - $57; //@line 182 "./libraries/Box2D/Collision/b2Collision.h"
     $79 = $69 - $61; //@line 183 "./libraries/Box2D/Collision/b2Collision.h"
     $80 = $78 + $79; //@line 184 "./libraries/Box2D/Collision/b2Collision.h"
     $81 = $80 * 2.0; //@line 184 "./libraries/Box2D/Collision/b2Collision.h"
     $82 = $81 - $77; //@line 221 "libraries/Box2D/Collision/b2DynamicTree.cpp"
     $$pn = $82;
    }
    $cost1$0 = $49 + $$pn; //@line 213 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $83 = (((($9) + (($22*36)|0)|0)) + 24|0); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $84 = HEAP32[$83>>2]|0; //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $85 = ($84|0)==(-1); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $86 = (($9) + (($22*36)|0)|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
    $87 = +HEAPF32[$86>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
    $88 = $11 < $87; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $89 = $88 ? $11 : $87; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $90 = (((($9) + (($22*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
    $91 = +HEAPF32[$90>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
    $92 = $13 < $91; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $93 = $92 ? $13 : $91; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $94 = (((($9) + (($22*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
    $95 = +HEAPF32[$94>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
    $96 = $15 > $95; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $97 = $96 ? $15 : $95; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $98 = (((($9) + (($22*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
    $99 = +HEAPF32[$98>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
    $100 = $17 > $99; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $101 = $100 ? $17 : $99; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    if ($85) {
     $102 = $97 - $89; //@line 182 "./libraries/Box2D/Collision/b2Collision.h"
     $103 = $101 - $93; //@line 183 "./libraries/Box2D/Collision/b2Collision.h"
     $104 = $102 + $103; //@line 184 "./libraries/Box2D/Collision/b2Collision.h"
     $105 = $104 * 2.0; //@line 184 "./libraries/Box2D/Collision/b2Collision.h"
     $$pn1 = $105;
    } else {
     $106 = $95 - $87; //@line 182 "./libraries/Box2D/Collision/b2Collision.h"
     $107 = $99 - $91; //@line 183 "./libraries/Box2D/Collision/b2Collision.h"
     $108 = $106 + $107; //@line 184 "./libraries/Box2D/Collision/b2Collision.h"
     $109 = $108 * 2.0; //@line 184 "./libraries/Box2D/Collision/b2Collision.h"
     $110 = $97 - $89; //@line 182 "./libraries/Box2D/Collision/b2Collision.h"
     $111 = $101 - $93; //@line 183 "./libraries/Box2D/Collision/b2Collision.h"
     $112 = $110 + $111; //@line 184 "./libraries/Box2D/Collision/b2Collision.h"
     $113 = $112 * 2.0; //@line 184 "./libraries/Box2D/Collision/b2Collision.h"
     $114 = $113 - $109; //@line 238 "libraries/Box2D/Collision/b2DynamicTree.cpp"
     $$pn1 = $114;
    }
    $cost2$0 = $49 + $$pn1; //@line 230 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $115 = $47 < $cost1$0; //@line 242 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $116 = $47 < $cost2$0; //@line 242 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $or$cond = $115 & $116; //@line 242 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    if ($or$cond) {
     $index$0$lcssa = $index$04;
     break L5;
    }
    $117 = $cost1$0 < $cost2$0; //@line 248 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $$ = $117 ? $51 : $22; //@line 251 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $118 = (((($9) + (($$*36)|0)|0)) + 24|0); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $119 = HEAP32[$118>>2]|0; //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $120 = ($119|0)==(-1); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
    if ($120) {
     $index$0$lcssa = $$;
     break;
    } else {
     $51 = $119;$index$04 = $$;
    }
   }
  }
 } while(0);
 $121 = (((($9) + (($index$0$lcssa*36)|0)|0)) + 20|0); //@line 261 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $122 = HEAP32[$121>>2]|0; //@line 261 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $123 = (__ZN13b2DynamicTree12AllocateNodeEv($this)|0); //@line 262 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $124 = HEAP32[$8>>2]|0; //@line 263 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $125 = (((($124) + (($123*36)|0)|0)) + 20|0); //@line 263 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$125>>2] = $122; //@line 263 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $126 = (((($124) + (($123*36)|0)|0)) + 16|0); //@line 264 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$126>>2] = 0; //@line 264 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $127 = (($124) + (($index$0$lcssa*36)|0)|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
 $128 = +HEAPF32[$127>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
 $129 = $11 < $128; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $130 = $129 ? $11 : $128; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $131 = (((($124) + (($index$0$lcssa*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
 $132 = +HEAPF32[$131>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
 $133 = $13 < $132; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $134 = $133 ? $13 : $132; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $135 = (($124) + (($123*36)|0)|0); //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
 HEAPF32[$135>>2] = $130; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
 $136 = (((($124) + (($123*36)|0)|0)) + 4|0); //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
 HEAPF32[$136>>2] = $134; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
 $137 = (((($124) + (($index$0$lcssa*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
 $138 = +HEAPF32[$137>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
 $139 = $15 > $138; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $140 = $139 ? $15 : $138; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $141 = (((($124) + (($index$0$lcssa*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
 $142 = +HEAPF32[$141>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
 $143 = $17 > $142; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $144 = $143 ? $17 : $142; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $145 = (((($124) + (($123*36)|0)|0)) + 8|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
 HEAPF32[$145>>2] = $140; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
 $146 = (((($124) + (($123*36)|0)|0)) + 12|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
 HEAPF32[$146>>2] = $144; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
 $147 = HEAP32[$8>>2]|0; //@line 266 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $148 = (((($147) + (($index$0$lcssa*36)|0)|0)) + 32|0); //@line 266 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $149 = HEAP32[$148>>2]|0; //@line 266 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $150 = (($149) + 1)|0; //@line 266 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $151 = (((($147) + (($123*36)|0)|0)) + 32|0); //@line 266 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$151>>2] = $150; //@line 266 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $152 = ($122|0)==(-1); //@line 268 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($152) {
  $162 = (((($147) + (($123*36)|0)|0)) + 24|0); //@line 288 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$162>>2] = $index$0$lcssa; //@line 288 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $163 = HEAP32[$8>>2]|0; //@line 289 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $164 = (((($163) + (($123*36)|0)|0)) + 28|0); //@line 289 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$164>>2] = $leaf; //@line 289 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $165 = (((($163) + (($index$0$lcssa*36)|0)|0)) + 20|0); //@line 290 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$165>>2] = $123; //@line 290 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $166 = (((($163) + (($leaf*36)|0)|0)) + 20|0); //@line 291 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$166>>2] = $123; //@line 291 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$this>>2] = $123; //@line 292 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 } else {
  $153 = (((($147) + (($122*36)|0)|0)) + 24|0); //@line 271 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $154 = HEAP32[$153>>2]|0; //@line 271 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $155 = ($154|0)==($index$0$lcssa|0); //@line 271 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  if ($155) {
   HEAP32[$153>>2] = $123; //@line 273 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  } else {
   $156 = (((($147) + (($122*36)|0)|0)) + 28|0); //@line 277 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   HEAP32[$156>>2] = $123; //@line 277 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  }
  $157 = HEAP32[$8>>2]|0; //@line 280 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $158 = (((($157) + (($123*36)|0)|0)) + 24|0); //@line 280 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$158>>2] = $index$0$lcssa; //@line 280 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $159 = (((($157) + (($123*36)|0)|0)) + 28|0); //@line 281 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$159>>2] = $leaf; //@line 281 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $160 = (((($157) + (($index$0$lcssa*36)|0)|0)) + 20|0); //@line 282 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$160>>2] = $123; //@line 282 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $161 = (((($157) + (($leaf*36)|0)|0)) + 20|0); //@line 283 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$161>>2] = $123; //@line 283 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 }
 $167 = HEAP32[$8>>2]|0; //@line 296 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $168 = (((($167) + (($leaf*36)|0)|0)) + 20|0); //@line 296 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $index$42 = HEAP32[$168>>2]|0; //@line 296 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $169 = ($index$42|0)==(-1); //@line 297 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($169) {
  return; //@line 314 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 } else {
  $index$43 = $index$42;
 }
 while(1) {
  $170 = (__ZN13b2DynamicTree7BalanceEi($this,$index$43)|0); //@line 299 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $171 = HEAP32[$8>>2]|0; //@line 301 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $172 = (((($171) + (($170*36)|0)|0)) + 24|0); //@line 301 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $173 = HEAP32[$172>>2]|0; //@line 301 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $174 = (((($171) + (($170*36)|0)|0)) + 28|0); //@line 302 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $175 = HEAP32[$174>>2]|0; //@line 302 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $176 = ($173|0)!=(-1); //@line 304 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  __Z10loveAssertbPKc($176,445874); //@line 304 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $177 = ($175|0)!=(-1); //@line 305 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  __Z10loveAssertbPKc($177,445896); //@line 305 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $178 = HEAP32[$8>>2]|0; //@line 307 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $179 = (((($178) + (($173*36)|0)|0)) + 32|0); //@line 307 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $180 = HEAP32[$179>>2]|0; //@line 307 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $181 = (((($178) + (($175*36)|0)|0)) + 32|0); //@line 307 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $182 = HEAP32[$181>>2]|0; //@line 307 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $183 = ($180|0)>($182|0); //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $184 = $183 ? $180 : $182; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $185 = (($184) + 1)|0; //@line 307 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $186 = (((($178) + (($170*36)|0)|0)) + 32|0); //@line 307 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$186>>2] = $185; //@line 307 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $187 = (($178) + (($173*36)|0)|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $188 = +HEAPF32[$187>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $189 = (($178) + (($175*36)|0)|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $190 = +HEAPF32[$189>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $191 = $188 < $190; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $192 = $191 ? $188 : $190; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $193 = (((($178) + (($173*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $194 = +HEAPF32[$193>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $195 = (((($178) + (($175*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $196 = +HEAPF32[$195>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $197 = $194 < $196; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $198 = $197 ? $194 : $196; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $199 = (($178) + (($170*36)|0)|0); //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$199>>2] = $192; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  $200 = (((($178) + (($170*36)|0)|0)) + 4|0); //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$200>>2] = $198; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  $201 = (((($178) + (($173*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $202 = +HEAPF32[$201>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $203 = (((($178) + (($175*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $204 = +HEAPF32[$203>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $205 = $202 > $204; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $206 = $205 ? $202 : $204; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $207 = (((($178) + (($173*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $208 = +HEAPF32[$207>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $209 = (((($178) + (($175*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $210 = +HEAPF32[$209>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $211 = $208 > $210; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $212 = $211 ? $208 : $210; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $213 = (((($178) + (($170*36)|0)|0)) + 8|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$213>>2] = $206; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  $214 = (((($178) + (($170*36)|0)|0)) + 12|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$214>>2] = $212; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  $215 = HEAP32[$8>>2]|0; //@line 310 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $216 = (((($215) + (($170*36)|0)|0)) + 20|0); //@line 310 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $index$4 = HEAP32[$216>>2]|0; //@line 296 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $217 = ($index$4|0)==(-1); //@line 297 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  if ($217) {
   break;
  } else {
   $index$43 = $index$4;
  }
 }
 return; //@line 314 "libraries/Box2D/Collision/b2DynamicTree.cpp"
}
function __ZN13b2DynamicTree7BalanceEi($this,$iA) {
 $this = $this|0;
 $iA = $iA|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0, $108 = 0.0, $109 = 0, $11 = 0, $110 = 0.0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0.0, $123 = 0.0, $124 = 0, $125 = 0.0, $126 = 0, $127 = 0.0, $128 = 0, $129 = 0.0, $13 = 0, $130 = 0, $131 = 0.0, $132 = 0;
 var $133 = 0, $134 = 0.0, $135 = 0, $136 = 0.0, $137 = 0, $138 = 0.0, $139 = 0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0, $144 = 0.0, $145 = 0, $146 = 0, $147 = 0.0, $148 = 0, $149 = 0.0, $15 = 0, $150 = 0;
 var $151 = 0.0, $152 = 0, $153 = 0.0, $154 = 0, $155 = 0.0, $156 = 0, $157 = 0.0, $158 = 0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0, $162 = 0.0, $163 = 0, $164 = 0.0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0.0, $214 = 0.0, $215 = 0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0, $22 = 0, $220 = 0.0, $221 = 0, $222 = 0.0;
 var $223 = 0, $224 = 0, $225 = 0.0, $226 = 0, $227 = 0.0, $228 = 0, $229 = 0.0, $23 = 0, $230 = 0, $231 = 0.0, $232 = 0, $233 = 0.0, $234 = 0, $235 = 0.0, $236 = 0, $237 = 0, $238 = 0.0, $239 = 0, $24 = 0, $240 = 0.0;
 var $241 = 0, $242 = 0.0, $243 = 0, $244 = 0.0, $245 = 0, $246 = 0.0, $247 = 0, $248 = 0.0, $249 = 0, $25 = 0, $250 = 0.0, $251 = 0.0, $252 = 0, $253 = 0.0, $254 = 0, $255 = 0.0, $256 = 0, $257 = 0, $258 = 0, $259 = 0;
 var $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0.0, $268 = 0.0, $269 = 0, $27 = 0, $270 = 0.0, $271 = 0, $272 = 0.0, $273 = 0, $274 = 0.0, $275 = 0, $276 = 0.0, $277 = 0;
 var $278 = 0, $279 = 0.0, $28 = 0, $280 = 0, $281 = 0.0, $282 = 0, $283 = 0.0, $284 = 0, $285 = 0.0, $286 = 0, $287 = 0.0, $288 = 0, $289 = 0.0, $29 = 0, $290 = 0, $291 = 0, $292 = 0.0, $293 = 0, $294 = 0.0, $295 = 0;
 var $296 = 0.0, $297 = 0, $298 = 0.0, $299 = 0, $3 = 0, $30 = 0, $300 = 0.0, $301 = 0, $302 = 0.0, $303 = 0, $304 = 0.0, $305 = 0.0, $306 = 0, $307 = 0.0, $308 = 0, $309 = 0.0, $31 = 0, $310 = 0, $311 = 0, $312 = 0;
 var $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0, $73 = 0.0, $74 = 0, $75 = 0.0, $76 = 0, $77 = 0.0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0.0, $81 = 0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0, $86 = 0.0, $87 = 0, $88 = 0.0, $89 = 0, $9 = 0, $90 = 0.0, $91 = 0, $92 = 0, $93 = 0.0, $94 = 0, $95 = 0.0, $96 = 0, $97 = 0.0, $98 = 0;
 var $99 = 0.0, $storemerge = 0, $storemerge$in = 0, $storemerge1 = 0, $storemerge1$in = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($iA|0)!=(-1); //@line 379 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 __Z10loveAssertbPKc($0,445606); //@line 379 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $1 = ((($this)) + 4|0); //@line 381 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 381 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $3 = (($2) + (($iA*36)|0)|0); //@line 381 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $4 = (((($2) + (($iA*36)|0)|0)) + 24|0); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $5 = HEAP32[$4>>2]|0; //@line 387 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $6 = ($5|0)==(-1); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
 if ($6) {
  $$1 = $iA;
  return ($$1|0); //@line 518 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 }
 $7 = (((($2) + (($iA*36)|0)|0)) + 32|0); //@line 382 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 382 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $9 = ($8|0)<(2); //@line 382 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($9) {
  $$1 = $iA;
  return ($$1|0); //@line 518 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 }
 $10 = (((($2) + (($iA*36)|0)|0)) + 28|0); //@line 388 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 388 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $12 = ($5|0)>(-1); //@line 389 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($12) {
  $13 = ((($this)) + 12|0); //@line 389 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 389 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $15 = ($5|0)<($14|0); //@line 389 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $16 = $15;
 } else {
  $16 = 0;
 }
 __Z10loveAssertbPKc($16,445624); //@line 389 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $17 = ($11|0)>(-1); //@line 390 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($17) {
  $18 = ((($this)) + 12|0); //@line 390 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 390 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $20 = ($11|0)<($19|0); //@line 390 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $21 = $20;
 } else {
  $21 = 0;
 }
 __Z10loveAssertbPKc($21,445655); //@line 390 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $22 = HEAP32[$1>>2]|0; //@line 392 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $23 = (($22) + (($5*36)|0)|0); //@line 392 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $24 = (($22) + (($11*36)|0)|0); //@line 393 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $25 = (((($22) + (($11*36)|0)|0)) + 32|0); //@line 395 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 395 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $27 = (((($22) + (($5*36)|0)|0)) + 32|0); //@line 395 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $28 = HEAP32[$27>>2]|0; //@line 395 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $29 = (($26) - ($28))|0; //@line 395 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $30 = ($29|0)>(1); //@line 398 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($30) {
  $31 = (((($22) + (($11*36)|0)|0)) + 24|0); //@line 400 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $32 = HEAP32[$31>>2]|0; //@line 400 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $33 = (((($22) + (($11*36)|0)|0)) + 28|0); //@line 401 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $34 = HEAP32[$33>>2]|0; //@line 401 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $35 = (($22) + (($32*36)|0)|0); //@line 402 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $36 = (($22) + (($34*36)|0)|0); //@line 403 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $37 = ($32|0)>(-1); //@line 404 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  if ($37) {
   $38 = ((($this)) + 12|0); //@line 404 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $39 = HEAP32[$38>>2]|0; //@line 404 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $40 = ($32|0)<($39|0); //@line 404 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $41 = $40;
  } else {
   $41 = 0;
  }
  __Z10loveAssertbPKc($41,445686); //@line 404 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $42 = ($34|0)>(-1); //@line 405 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  if ($42) {
   $43 = ((($this)) + 12|0); //@line 405 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $44 = HEAP32[$43>>2]|0; //@line 405 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $45 = ($34|0)<($44|0); //@line 405 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $46 = $45;
  } else {
   $46 = 0;
  }
  __Z10loveAssertbPKc($46,445717); //@line 405 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$31>>2] = $iA; //@line 408 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $47 = (((($2) + (($iA*36)|0)|0)) + 20|0); //@line 409 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $48 = HEAP32[$47>>2]|0; //@line 409 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $49 = (((($22) + (($11*36)|0)|0)) + 20|0); //@line 409 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$49>>2] = $48; //@line 409 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$47>>2] = $11; //@line 410 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $50 = HEAP32[$49>>2]|0; //@line 413 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $51 = ($50|0)==(-1); //@line 413 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  do {
   if ($51) {
    HEAP32[$this>>2] = $11; //@line 427 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   } else {
    $52 = HEAP32[$1>>2]|0; //@line 415 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $53 = (((($52) + (($50*36)|0)|0)) + 24|0); //@line 415 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $54 = HEAP32[$53>>2]|0; //@line 415 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $55 = ($54|0)==($iA|0); //@line 415 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    if ($55) {
     HEAP32[$53>>2] = $11; //@line 417 "libraries/Box2D/Collision/b2DynamicTree.cpp"
     break;
    } else {
     $56 = (((($52) + (($50*36)|0)|0)) + 28|0); //@line 421 "libraries/Box2D/Collision/b2DynamicTree.cpp"
     $57 = HEAP32[$56>>2]|0; //@line 421 "libraries/Box2D/Collision/b2DynamicTree.cpp"
     $58 = ($57|0)==($iA|0); //@line 421 "libraries/Box2D/Collision/b2DynamicTree.cpp"
     __Z10loveAssertbPKc($58,445748); //@line 421 "libraries/Box2D/Collision/b2DynamicTree.cpp"
     $59 = HEAP32[$49>>2]|0; //@line 422 "libraries/Box2D/Collision/b2DynamicTree.cpp"
     $60 = HEAP32[$1>>2]|0; //@line 422 "libraries/Box2D/Collision/b2DynamicTree.cpp"
     $61 = (((($60) + (($59*36)|0)|0)) + 28|0); //@line 422 "libraries/Box2D/Collision/b2DynamicTree.cpp"
     HEAP32[$61>>2] = $11; //@line 422 "libraries/Box2D/Collision/b2DynamicTree.cpp"
     break;
    }
   }
  } while(0);
  $62 = (((($22) + (($32*36)|0)|0)) + 32|0); //@line 431 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $63 = HEAP32[$62>>2]|0; //@line 431 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $64 = (((($22) + (($34*36)|0)|0)) + 32|0); //@line 431 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $65 = HEAP32[$64>>2]|0; //@line 431 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $66 = ($63|0)>($65|0); //@line 431 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  if ($66) {
   HEAP32[$33>>2] = $32; //@line 433 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   HEAP32[$10>>2] = $34; //@line 434 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $67 = (((($22) + (($34*36)|0)|0)) + 20|0); //@line 435 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   HEAP32[$67>>2] = $iA; //@line 435 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $68 = +HEAPF32[$23>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $69 = +HEAPF32[$36>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $70 = $68 < $69; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $71 = $70 ? $68 : $69; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $72 = (((($22) + (($5*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $73 = +HEAPF32[$72>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $74 = (((($22) + (($34*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $75 = +HEAPF32[$74>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $76 = $73 < $75; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $77 = $76 ? $73 : $75; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$3>>2] = $71; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
   $78 = (((($2) + (($iA*36)|0)|0)) + 4|0); //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
   HEAPF32[$78>>2] = $77; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
   $79 = (((($22) + (($5*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $80 = +HEAPF32[$79>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $81 = (((($22) + (($34*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $82 = +HEAPF32[$81>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $83 = $80 > $82; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $84 = $83 ? $80 : $82; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $85 = (((($22) + (($5*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $86 = +HEAPF32[$85>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $87 = (((($22) + (($34*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $88 = +HEAPF32[$87>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $89 = $86 > $88; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $90 = $89 ? $86 : $88; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $91 = (((($2) + (($iA*36)|0)|0)) + 8|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
   HEAPF32[$91>>2] = $84; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
   $92 = (((($2) + (($iA*36)|0)|0)) + 12|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
   HEAPF32[$92>>2] = $90; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
   $93 = +HEAPF32[$35>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $94 = $71 < $93; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $95 = $94 ? $71 : $93; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $96 = (((($22) + (($32*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $97 = +HEAPF32[$96>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $98 = $77 < $97; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $99 = $98 ? $77 : $97; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$24>>2] = $95; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
   $100 = (((($22) + (($11*36)|0)|0)) + 4|0); //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
   HEAPF32[$100>>2] = $99; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
   $101 = +HEAPF32[$91>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $102 = (((($22) + (($32*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $103 = +HEAPF32[$102>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $104 = $101 > $103; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $105 = $104 ? $101 : $103; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $106 = +HEAPF32[$92>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $107 = (((($22) + (($32*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $108 = +HEAPF32[$107>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $109 = $106 > $108; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $110 = $109 ? $106 : $108; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $111 = (((($22) + (($11*36)|0)|0)) + 8|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
   HEAPF32[$111>>2] = $105; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
   $112 = (((($22) + (($11*36)|0)|0)) + 12|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
   HEAPF32[$112>>2] = $110; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
   $113 = HEAP32[$27>>2]|0; //@line 439 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $114 = HEAP32[$64>>2]|0; //@line 439 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $115 = ($113|0)>($114|0); //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $116 = $115 ? $113 : $114; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $117 = (($116) + 1)|0; //@line 439 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   HEAP32[$7>>2] = $117; //@line 439 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $118 = HEAP32[$62>>2]|0; //@line 440 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $119 = ($117|0)>($118|0); //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $120 = $119 ? $117 : $118; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $storemerge1$in = $120;
  } else {
   HEAP32[$33>>2] = $34; //@line 444 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   HEAP32[$10>>2] = $32; //@line 445 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $121 = (((($22) + (($32*36)|0)|0)) + 20|0); //@line 446 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   HEAP32[$121>>2] = $iA; //@line 446 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $122 = +HEAPF32[$23>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $123 = +HEAPF32[$35>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $124 = $122 < $123; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $125 = $124 ? $122 : $123; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $126 = (((($22) + (($5*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $127 = +HEAPF32[$126>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $128 = (((($22) + (($32*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $129 = +HEAPF32[$128>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $130 = $127 < $129; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $131 = $130 ? $127 : $129; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$3>>2] = $125; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
   $132 = (((($2) + (($iA*36)|0)|0)) + 4|0); //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
   HEAPF32[$132>>2] = $131; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
   $133 = (((($22) + (($5*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $134 = +HEAPF32[$133>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $135 = (((($22) + (($32*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $136 = +HEAPF32[$135>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $137 = $134 > $136; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $138 = $137 ? $134 : $136; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $139 = (((($22) + (($5*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $140 = +HEAPF32[$139>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $141 = (((($22) + (($32*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $142 = +HEAPF32[$141>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $143 = $140 > $142; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $144 = $143 ? $140 : $142; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $145 = (((($2) + (($iA*36)|0)|0)) + 8|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
   HEAPF32[$145>>2] = $138; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
   $146 = (((($2) + (($iA*36)|0)|0)) + 12|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
   HEAPF32[$146>>2] = $144; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
   $147 = +HEAPF32[$36>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $148 = $125 < $147; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $149 = $148 ? $125 : $147; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $150 = (((($22) + (($34*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $151 = +HEAPF32[$150>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
   $152 = $131 < $151; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   $153 = $152 ? $131 : $151; //@line 634 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$24>>2] = $149; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
   $154 = (((($22) + (($11*36)|0)|0)) + 4|0); //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
   HEAPF32[$154>>2] = $153; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
   $155 = +HEAPF32[$145>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $156 = (((($22) + (($34*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $157 = +HEAPF32[$156>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $158 = $155 > $157; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $159 = $158 ? $155 : $157; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $160 = +HEAPF32[$146>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $161 = (((($22) + (($34*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $162 = +HEAPF32[$161>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
   $163 = $160 > $162; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $164 = $163 ? $160 : $162; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $165 = (((($22) + (($11*36)|0)|0)) + 8|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
   HEAPF32[$165>>2] = $159; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
   $166 = (((($22) + (($11*36)|0)|0)) + 12|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
   HEAPF32[$166>>2] = $164; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
   $167 = HEAP32[$27>>2]|0; //@line 450 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $168 = HEAP32[$62>>2]|0; //@line 450 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $169 = ($167|0)>($168|0); //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $170 = $169 ? $167 : $168; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $171 = (($170) + 1)|0; //@line 450 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   HEAP32[$7>>2] = $171; //@line 450 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $172 = HEAP32[$64>>2]|0; //@line 451 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $173 = ($171|0)>($172|0); //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $174 = $173 ? $171 : $172; //@line 645 "./libraries/Box2D/Common/b2Math.h"
   $storemerge1$in = $174;
  }
  $storemerge1 = (($storemerge1$in) + 1)|0; //@line 451 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$25>>2] = $storemerge1; //@line 440 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $$1 = $11;
  return ($$1|0); //@line 518 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 }
 $175 = ($29|0)<(-1); //@line 458 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if (!($175)) {
  $$1 = $iA;
  return ($$1|0); //@line 518 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 }
 $176 = (((($22) + (($5*36)|0)|0)) + 24|0); //@line 460 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $177 = HEAP32[$176>>2]|0; //@line 460 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $178 = (((($22) + (($5*36)|0)|0)) + 28|0); //@line 461 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $179 = HEAP32[$178>>2]|0; //@line 461 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $180 = (($22) + (($177*36)|0)|0); //@line 462 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $181 = (($22) + (($179*36)|0)|0); //@line 463 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $182 = ($177|0)>(-1); //@line 464 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($182) {
  $183 = ((($this)) + 12|0); //@line 464 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $184 = HEAP32[$183>>2]|0; //@line 464 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $185 = ($177|0)<($184|0); //@line 464 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $186 = $185;
 } else {
  $186 = 0;
 }
 __Z10loveAssertbPKc($186,445780); //@line 464 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $187 = ($179|0)>(-1); //@line 465 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($187) {
  $188 = ((($this)) + 12|0); //@line 465 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $189 = HEAP32[$188>>2]|0; //@line 465 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $190 = ($179|0)<($189|0); //@line 465 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $191 = $190;
 } else {
  $191 = 0;
 }
 __Z10loveAssertbPKc($191,445811); //@line 465 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$176>>2] = $iA; //@line 468 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $192 = (((($2) + (($iA*36)|0)|0)) + 20|0); //@line 469 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $193 = HEAP32[$192>>2]|0; //@line 469 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $194 = (((($22) + (($5*36)|0)|0)) + 20|0); //@line 469 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$194>>2] = $193; //@line 469 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$192>>2] = $5; //@line 470 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $195 = HEAP32[$194>>2]|0; //@line 473 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $196 = ($195|0)==(-1); //@line 473 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 do {
  if ($196) {
   HEAP32[$this>>2] = $5; //@line 487 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  } else {
   $197 = HEAP32[$1>>2]|0; //@line 475 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $198 = (((($197) + (($195*36)|0)|0)) + 24|0); //@line 475 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $199 = HEAP32[$198>>2]|0; //@line 475 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $200 = ($199|0)==($iA|0); //@line 475 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   if ($200) {
    HEAP32[$198>>2] = $5; //@line 477 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    break;
   } else {
    $201 = (((($197) + (($195*36)|0)|0)) + 28|0); //@line 481 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $202 = HEAP32[$201>>2]|0; //@line 481 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $203 = ($202|0)==($iA|0); //@line 481 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    __Z10loveAssertbPKc($203,445842); //@line 481 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $204 = HEAP32[$194>>2]|0; //@line 482 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $205 = HEAP32[$1>>2]|0; //@line 482 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    $206 = (((($205) + (($204*36)|0)|0)) + 28|0); //@line 482 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    HEAP32[$206>>2] = $5; //@line 482 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    break;
   }
  }
 } while(0);
 $207 = (((($22) + (($177*36)|0)|0)) + 32|0); //@line 491 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $208 = HEAP32[$207>>2]|0; //@line 491 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $209 = (((($22) + (($179*36)|0)|0)) + 32|0); //@line 491 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $210 = HEAP32[$209>>2]|0; //@line 491 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $211 = ($208|0)>($210|0); //@line 491 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($211) {
  HEAP32[$178>>2] = $177; //@line 493 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$4>>2] = $179; //@line 494 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $212 = (((($22) + (($179*36)|0)|0)) + 20|0); //@line 495 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$212>>2] = $iA; //@line 495 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $213 = +HEAPF32[$24>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $214 = +HEAPF32[$181>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $215 = $213 < $214; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $216 = $215 ? $213 : $214; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $217 = (((($22) + (($11*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $218 = +HEAPF32[$217>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $219 = (((($22) + (($179*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $220 = +HEAPF32[$219>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $221 = $218 < $220; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $222 = $221 ? $218 : $220; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$3>>2] = $216; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  $223 = (((($2) + (($iA*36)|0)|0)) + 4|0); //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$223>>2] = $222; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  $224 = (((($22) + (($11*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $225 = +HEAPF32[$224>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $226 = (((($22) + (($179*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $227 = +HEAPF32[$226>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $228 = $225 > $227; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $229 = $228 ? $225 : $227; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $230 = (((($22) + (($11*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $231 = +HEAPF32[$230>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $232 = (((($22) + (($179*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $233 = +HEAPF32[$232>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $234 = $231 > $233; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $235 = $234 ? $231 : $233; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $236 = (((($2) + (($iA*36)|0)|0)) + 8|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$236>>2] = $229; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  $237 = (((($2) + (($iA*36)|0)|0)) + 12|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$237>>2] = $235; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  $238 = +HEAPF32[$180>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $239 = $216 < $238; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $240 = $239 ? $216 : $238; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $241 = (((($22) + (($177*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $242 = +HEAPF32[$241>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $243 = $222 < $242; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $244 = $243 ? $222 : $242; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$23>>2] = $240; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  $245 = (((($22) + (($5*36)|0)|0)) + 4|0); //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$245>>2] = $244; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  $246 = +HEAPF32[$236>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $247 = (((($22) + (($177*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $248 = +HEAPF32[$247>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $249 = $246 > $248; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $250 = $249 ? $246 : $248; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $251 = +HEAPF32[$237>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $252 = (((($22) + (($177*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $253 = +HEAPF32[$252>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $254 = $251 > $253; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $255 = $254 ? $251 : $253; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $256 = (((($22) + (($5*36)|0)|0)) + 8|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$256>>2] = $250; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  $257 = (((($22) + (($5*36)|0)|0)) + 12|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$257>>2] = $255; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  $258 = HEAP32[$25>>2]|0; //@line 499 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $259 = HEAP32[$209>>2]|0; //@line 499 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $260 = ($258|0)>($259|0); //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $261 = $260 ? $258 : $259; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $262 = (($261) + 1)|0; //@line 499 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$7>>2] = $262; //@line 499 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $263 = HEAP32[$207>>2]|0; //@line 500 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $264 = ($262|0)>($263|0); //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $265 = $264 ? $262 : $263; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $storemerge$in = $265;
 } else {
  HEAP32[$178>>2] = $179; //@line 504 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$4>>2] = $177; //@line 505 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $266 = (((($22) + (($177*36)|0)|0)) + 20|0); //@line 506 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$266>>2] = $iA; //@line 506 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $267 = +HEAPF32[$24>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $268 = +HEAPF32[$180>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $269 = $267 < $268; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $270 = $269 ? $267 : $268; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $271 = (((($22) + (($11*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $272 = +HEAPF32[$271>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $273 = (((($22) + (($177*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $274 = +HEAPF32[$273>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $275 = $272 < $274; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $276 = $275 ? $272 : $274; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$3>>2] = $270; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  $277 = (((($2) + (($iA*36)|0)|0)) + 4|0); //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$277>>2] = $276; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  $278 = (((($22) + (($11*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $279 = +HEAPF32[$278>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $280 = (((($22) + (($177*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $281 = +HEAPF32[$280>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $282 = $279 > $281; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $283 = $282 ? $279 : $281; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $284 = (((($22) + (($11*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $285 = +HEAPF32[$284>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $286 = (((($22) + (($177*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $287 = +HEAPF32[$286>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $288 = $285 > $287; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $289 = $288 ? $285 : $287; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $290 = (((($2) + (($iA*36)|0)|0)) + 8|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$290>>2] = $283; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  $291 = (((($2) + (($iA*36)|0)|0)) + 12|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$291>>2] = $289; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  $292 = +HEAPF32[$181>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $293 = $270 < $292; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $294 = $293 ? $270 : $292; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $295 = (((($22) + (($179*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $296 = +HEAPF32[$295>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $297 = $276 < $296; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $298 = $297 ? $276 : $296; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$23>>2] = $294; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  $299 = (((($22) + (($5*36)|0)|0)) + 4|0); //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$299>>2] = $298; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  $300 = +HEAPF32[$290>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $301 = (((($22) + (($179*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $302 = +HEAPF32[$301>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $303 = $300 > $302; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $304 = $303 ? $300 : $302; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $305 = +HEAPF32[$291>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $306 = (((($22) + (($179*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $307 = +HEAPF32[$306>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $308 = $305 > $307; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $309 = $308 ? $305 : $307; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $310 = (((($22) + (($5*36)|0)|0)) + 8|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$310>>2] = $304; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  $311 = (((($22) + (($5*36)|0)|0)) + 12|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$311>>2] = $309; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  $312 = HEAP32[$25>>2]|0; //@line 510 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $313 = HEAP32[$207>>2]|0; //@line 510 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $314 = ($312|0)>($313|0); //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $315 = $314 ? $312 : $313; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $316 = (($315) + 1)|0; //@line 510 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$7>>2] = $316; //@line 510 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $317 = HEAP32[$209>>2]|0; //@line 511 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $318 = ($316|0)>($317|0); //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $319 = $318 ? $316 : $317; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $storemerge$in = $319;
 }
 $storemerge = (($storemerge$in) + 1)|0; //@line 511 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$27>>2] = $storemerge; //@line 500 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $$1 = $5;
 return ($$1|0); //@line 518 "libraries/Box2D/Collision/b2DynamicTree.cpp"
}
function __ZN13b2DynamicTree12DestroyProxyEi($this,$proxyId) {
 $this = $this|0;
 $proxyId = $proxyId|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($proxyId|0)>(-1); //@line 123 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($0) {
  $1 = ((($this)) + 12|0); //@line 123 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $2 = HEAP32[$1>>2]|0; //@line 123 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $3 = ($2|0)>($proxyId|0); //@line 123 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $4 = $3;
 } else {
  $4 = 0;
 }
 __Z10loveAssertbPKc($4,446451); //@line 123 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $5 = ((($this)) + 4|0); //@line 124 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 124 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $7 = (((($6) + (($proxyId*36)|0)|0)) + 24|0); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $8 = HEAP32[$7>>2]|0; //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $9 = ($8|0)==(-1); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
 __Z10loveAssertbPKc($9,445918); //@line 124 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 __ZN13b2DynamicTree10RemoveLeafEi($this,$proxyId); //@line 126 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($0) {
  $10 = ((($this)) + 12|0); //@line 94 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 94 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $12 = ($11|0)>($proxyId|0); //@line 94 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $13 = $12;
 } else {
  $13 = 0;
 }
 __Z10loveAssertbPKc($13,445551); //@line 94 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $14 = ((($this)) + 8|0); //@line 95 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 95 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $16 = ($15|0)>(0); //@line 95 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 __Z10loveAssertbPKc($16,445590); //@line 95 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $17 = ((($this)) + 16|0); //@line 96 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 96 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $19 = HEAP32[$5>>2]|0; //@line 96 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $20 = (((($19) + (($proxyId*36)|0)|0)) + 20|0); //@line 96 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$20>>2] = $18; //@line 96 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $21 = (((($19) + (($proxyId*36)|0)|0)) + 32|0); //@line 97 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$21>>2] = -1; //@line 97 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$17>>2] = $proxyId; //@line 98 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $22 = HEAP32[$14>>2]|0; //@line 99 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $23 = (($22) + -1)|0; //@line 99 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$14>>2] = $23; //@line 99 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 return; //@line 128 "libraries/Box2D/Collision/b2DynamicTree.cpp"
}
function __ZN13b2DynamicTree10RemoveLeafEi($this,$leaf) {
 $this = $this|0;
 $leaf = $leaf|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0, $43 = 0.0, $44 = 0;
 var $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0, $53 = 0, $54 = 0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0;
 var $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $index$02 = 0, $sibling$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 318 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $1 = ($0|0)==($leaf|0); //@line 318 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($1) {
  HEAP32[$this>>2] = -1; //@line 320 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  return; //@line 373 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 }
 $2 = ((($this)) + 4|0); //@line 324 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 368 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $4 = (((($3) + (($leaf*36)|0)|0)) + 20|0); //@line 324 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 324 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $6 = (((($3) + (($5*36)|0)|0)) + 20|0); //@line 325 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 325 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $8 = (((($3) + (($5*36)|0)|0)) + 24|0); //@line 327 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 327 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $10 = ($9|0)==($leaf|0); //@line 327 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($10) {
  $11 = (((($3) + (($5*36)|0)|0)) + 28|0); //@line 329 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 329 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $sibling$0 = $12;
 } else {
  $sibling$0 = $9;
 }
 $13 = ($7|0)==(-1); //@line 336 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($13) {
  HEAP32[$this>>2] = $sibling$0; //@line 367 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $80 = (((($3) + (($sibling$0*36)|0)|0)) + 20|0); //@line 368 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$80>>2] = -1; //@line 368 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $81 = ($5|0)>(-1); //@line 94 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  if ($81) {
   $82 = ((($this)) + 12|0); //@line 94 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $83 = HEAP32[$82>>2]|0; //@line 94 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $84 = ($83|0)>($5|0); //@line 94 "libraries/Box2D/Collision/b2DynamicTree.cpp"
   $85 = $84;
  } else {
   $85 = 0;
  }
  __Z10loveAssertbPKc($85,445551); //@line 94 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $86 = ((($this)) + 8|0); //@line 95 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $87 = HEAP32[$86>>2]|0; //@line 95 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $88 = ($87|0)>(0); //@line 95 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  __Z10loveAssertbPKc($88,445590); //@line 95 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $89 = ((($this)) + 16|0); //@line 96 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $90 = HEAP32[$89>>2]|0; //@line 96 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $91 = HEAP32[$2>>2]|0; //@line 96 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $92 = (((($91) + (($5*36)|0)|0)) + 20|0); //@line 96 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$92>>2] = $90; //@line 96 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $93 = (((($91) + (($5*36)|0)|0)) + 32|0); //@line 97 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$93>>2] = -1; //@line 97 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$89>>2] = $5; //@line 98 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $94 = HEAP32[$86>>2]|0; //@line 99 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $95 = (($94) + -1)|0; //@line 99 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$86>>2] = $95; //@line 99 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  return; //@line 373 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 }
 $14 = (((($3) + (($7*36)|0)|0)) + 24|0); //@line 339 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 339 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $16 = ($15|0)==($5|0); //@line 339 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($16) {
  HEAP32[$14>>2] = $sibling$0; //@line 341 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 } else {
  $17 = (((($3) + (($7*36)|0)|0)) + 28|0); //@line 345 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$17>>2] = $sibling$0; //@line 345 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 }
 $18 = (((($3) + (($sibling$0*36)|0)|0)) + 20|0); //@line 347 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$18>>2] = $7; //@line 347 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $19 = ($5|0)>(-1); //@line 94 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($19) {
  $20 = ((($this)) + 12|0); //@line 94 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $21 = HEAP32[$20>>2]|0; //@line 94 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $22 = ($21|0)>($5|0); //@line 94 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $23 = $22;
 } else {
  $23 = 0;
 }
 __Z10loveAssertbPKc($23,445551); //@line 94 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $24 = ((($this)) + 8|0); //@line 95 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $25 = HEAP32[$24>>2]|0; //@line 95 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $26 = ($25|0)>(0); //@line 95 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 __Z10loveAssertbPKc($26,445590); //@line 95 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $27 = ((($this)) + 16|0); //@line 96 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $28 = HEAP32[$27>>2]|0; //@line 96 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $29 = HEAP32[$2>>2]|0; //@line 96 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $30 = (((($29) + (($5*36)|0)|0)) + 20|0); //@line 96 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$30>>2] = $28; //@line 96 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $31 = (((($29) + (($5*36)|0)|0)) + 32|0); //@line 97 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$31>>2] = -1; //@line 97 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$27>>2] = $5; //@line 98 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $32 = HEAP32[$24>>2]|0; //@line 99 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $33 = (($32) + -1)|0; //@line 99 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAP32[$24>>2] = $33; //@line 99 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $index$02 = $7;
 while(1) {
  $34 = (__ZN13b2DynamicTree7BalanceEi($this,$index$02)|0); //@line 354 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $35 = HEAP32[$2>>2]|0; //@line 356 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $36 = (((($35) + (($34*36)|0)|0)) + 24|0); //@line 356 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $37 = HEAP32[$36>>2]|0; //@line 356 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $38 = (((($35) + (($34*36)|0)|0)) + 28|0); //@line 357 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $39 = HEAP32[$38>>2]|0; //@line 357 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $40 = (($35) + (($37*36)|0)|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $41 = +HEAPF32[$40>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $42 = (($35) + (($39*36)|0)|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $43 = +HEAPF32[$42>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $44 = $41 < $43; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $45 = $44 ? $41 : $43; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $46 = (((($35) + (($37*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $47 = +HEAPF32[$46>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $48 = (((($35) + (($39*36)|0)|0)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $49 = +HEAPF32[$48>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $50 = $47 < $49; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $51 = $50 ? $47 : $49; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $52 = (($35) + (($34*36)|0)|0); //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$52>>2] = $45; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  $53 = (((($35) + (($34*36)|0)|0)) + 4|0); //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$53>>2] = $51; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  $54 = (((($35) + (($37*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $55 = +HEAPF32[$54>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $56 = (((($35) + (($39*36)|0)|0)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $57 = +HEAPF32[$56>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $58 = $55 > $57; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $59 = $58 ? $55 : $57; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $60 = (((($35) + (($37*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $61 = +HEAPF32[$60>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $62 = (((($35) + (($39*36)|0)|0)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $63 = +HEAPF32[$62>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $64 = $61 > $63; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $65 = $64 ? $61 : $63; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $66 = (((($35) + (($34*36)|0)|0)) + 8|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$66>>2] = $59; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  $67 = (((($35) + (($34*36)|0)|0)) + 12|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$67>>2] = $65; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  $68 = HEAP32[$2>>2]|0; //@line 360 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $69 = (((($68) + (($37*36)|0)|0)) + 32|0); //@line 360 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $70 = HEAP32[$69>>2]|0; //@line 360 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $71 = (((($68) + (($39*36)|0)|0)) + 32|0); //@line 360 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $72 = HEAP32[$71>>2]|0; //@line 360 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $73 = ($70|0)>($72|0); //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $74 = $73 ? $70 : $72; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $75 = (($74) + 1)|0; //@line 360 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $76 = (((($68) + (($34*36)|0)|0)) + 32|0); //@line 360 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  HEAP32[$76>>2] = $75; //@line 360 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $77 = (((($68) + (($34*36)|0)|0)) + 20|0); //@line 362 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $78 = HEAP32[$77>>2]|0; //@line 362 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $79 = ($78|0)==(-1); //@line 352 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  if ($79) {
   break;
  } else {
   $index$02 = $78;
  }
 }
 return; //@line 373 "libraries/Box2D/Collision/b2DynamicTree.cpp"
}
function __ZN13b2DynamicTree9MoveProxyEiRK6b2AABBRK6b2Vec2($this,$proxyId,$aabb,$displacement) {
 $this = $this|0;
 $proxyId = $proxyId|0;
 $aabb = $aabb|0;
 $displacement = $displacement|0;
 var $$0 = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $23 = 0.0;
 var $24 = 0, $25 = 0, $26 = 0.0, $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0;
 var $42 = 0, $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b$sroa$0$0$in = 0.0;
 var $b$sroa$12$0$in = 0.0, $b$sroa$16$0$in = 0.0, $b$sroa$8$0$in = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($proxyId|0)>(-1); //@line 132 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if ($0) {
  $1 = ((($this)) + 12|0); //@line 132 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $2 = HEAP32[$1>>2]|0; //@line 132 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $3 = ($2|0)>($proxyId|0); //@line 132 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $4 = $3;
 } else {
  $4 = 0;
 }
 __Z10loveAssertbPKc($4,446451); //@line 132 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $5 = ((($this)) + 4|0); //@line 134 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 134 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $7 = (((($6) + (($proxyId*36)|0)|0)) + 24|0); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $8 = HEAP32[$7>>2]|0; //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $9 = ($8|0)==(-1); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
 __Z10loveAssertbPKc($9,445918); //@line 134 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $10 = HEAP32[$5>>2]|0; //@line 136 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $11 = (($10) + (($proxyId*36)|0)|0); //@line 205 "./libraries/Box2D/Collision/b2Collision.h"
 $12 = +HEAPF32[$11>>2]; //@line 205 "./libraries/Box2D/Collision/b2Collision.h"
 $13 = +HEAPF32[$aabb>>2]; //@line 205 "./libraries/Box2D/Collision/b2Collision.h"
 $14 = !($12 <= $13); //@line 205 "./libraries/Box2D/Collision/b2Collision.h"
 if ($14) {
  $$pre = ((($aabb)) + 4|0); //@line 144 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $$pre$phiZ2D = $$pre;
 } else {
  $15 = (((($10) + (($proxyId*36)|0)|0)) + 4|0); //@line 206 "./libraries/Box2D/Collision/b2Collision.h"
  $16 = +HEAPF32[$15>>2]; //@line 206 "./libraries/Box2D/Collision/b2Collision.h"
  $17 = ((($aabb)) + 4|0); //@line 206 "./libraries/Box2D/Collision/b2Collision.h"
  $18 = +HEAPF32[$17>>2]; //@line 206 "./libraries/Box2D/Collision/b2Collision.h"
  $19 = !($16 <= $18); //@line 206 "./libraries/Box2D/Collision/b2Collision.h"
  if ($19) {
   $$pre$phiZ2D = $17;
  } else {
   $20 = ((($aabb)) + 8|0); //@line 207 "./libraries/Box2D/Collision/b2Collision.h"
   $21 = +HEAPF32[$20>>2]; //@line 207 "./libraries/Box2D/Collision/b2Collision.h"
   $22 = (((($10) + (($proxyId*36)|0)|0)) + 8|0); //@line 207 "./libraries/Box2D/Collision/b2Collision.h"
   $23 = +HEAPF32[$22>>2]; //@line 207 "./libraries/Box2D/Collision/b2Collision.h"
   $24 = !($21 <= $23); //@line 207 "./libraries/Box2D/Collision/b2Collision.h"
   if ($24) {
    $$pre$phiZ2D = $17;
   } else {
    $25 = ((($aabb)) + 12|0); //@line 208 "./libraries/Box2D/Collision/b2Collision.h"
    $26 = +HEAPF32[$25>>2]; //@line 208 "./libraries/Box2D/Collision/b2Collision.h"
    $27 = (((($10) + (($proxyId*36)|0)|0)) + 12|0); //@line 208 "./libraries/Box2D/Collision/b2Collision.h"
    $28 = +HEAPF32[$27>>2]; //@line 208 "./libraries/Box2D/Collision/b2Collision.h"
    $29 = !($26 <= $28); //@line 208 "./libraries/Box2D/Collision/b2Collision.h"
    if ($29) {
     $$pre$phiZ2D = $17;
    } else {
     $$0 = 0;
     return ($$0|0); //@line 174 "libraries/Box2D/Collision/b2DynamicTree.cpp"
    }
   }
  }
 }
 __ZN13b2DynamicTree10RemoveLeafEi($this,$proxyId); //@line 141 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $30 = +HEAPF32[$aabb>>2]; //@line 144 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $31 = +HEAPF32[$$pre$phiZ2D>>2]; //@line 144 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $32 = ((($aabb)) + 8|0); //@line 144 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $33 = +HEAPF32[$32>>2]; //@line 144 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $34 = ((($aabb)) + 12|0); //@line 144 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $35 = +HEAPF32[$34>>2]; //@line 144 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $36 = $30 + -0.10000000149011612; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $37 = $31 + -0.10000000149011612; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $38 = $33 + 0.10000000149011612; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $39 = $35 + 0.10000000149011612; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $40 = +HEAPF32[$displacement>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $41 = $40 * 2.0; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $42 = ((($displacement)) + 4|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $43 = +HEAPF32[$42>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $44 = $43 * 2.0; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $45 = $41 < 0.0; //@line 152 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $46 = $36 + $41; //@line 154 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $47 = $38 + $41; //@line 158 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $b$sroa$0$0$in = $45 ? $46 : $36; //@line 152 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $b$sroa$12$0$in = $45 ? $38 : $47; //@line 152 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $48 = $44 < 0.0; //@line 161 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $49 = $37 + $44; //@line 163 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $50 = $39 + $44; //@line 167 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $b$sroa$8$0$in = $48 ? $49 : $37; //@line 161 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $b$sroa$16$0$in = $48 ? $39 : $50; //@line 161 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $51 = HEAP32[$5>>2]|0; //@line 170 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $52 = (($51) + (($proxyId*36)|0)|0); //@line 170 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAPF32[$52>>2] = $b$sroa$0$0$in; //@line 170 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $53 = (((($51) + (($proxyId*36)|0)|0)) + 4|0); //@line 170 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAPF32[$53>>2] = $b$sroa$8$0$in; //@line 170 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $54 = (((($51) + (($proxyId*36)|0)|0)) + 8|0); //@line 170 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAPF32[$54>>2] = $b$sroa$12$0$in; //@line 170 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $55 = (((($51) + (($proxyId*36)|0)|0)) + 12|0); //@line 170 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 HEAPF32[$55>>2] = $b$sroa$16$0$in; //@line 170 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 __ZN13b2DynamicTree10InsertLeafEi($this,$proxyId); //@line 172 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $$0 = 1;
 return ($$0|0); //@line 174 "libraries/Box2D/Collision/b2DynamicTree.cpp"
}
function __ZN13b2DynamicTree11ShiftOriginERK6b2Vec2($this,$newOrigin) {
 $this = $this|0;
 $newOrigin = $newOrigin|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0.0, $7 = 0, $8 = 0.0, $9 = 0.0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 773 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 773 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $2 = ($1|0)>(0); //@line 773 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 if (!($2)) {
  return; //@line 778 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 }
 $3 = ((($this)) + 4|0); //@line 775 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 775 "libraries/Box2D/Collision/b2DynamicTree.cpp"
 $5 = ((($newOrigin)) + 4|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $i$01 = 0;
 while(1) {
  $6 = +HEAPF32[$newOrigin>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $7 = (($4) + (($i$01*36)|0)|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $8 = +HEAPF32[$7>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $9 = $8 - $6; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$7>>2] = $9; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $10 = +HEAPF32[$5>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $11 = (((($4) + (($i$01*36)|0)|0)) + 4|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $12 = +HEAPF32[$11>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $13 = $12 - $10; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$11>>2] = $13; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $14 = +HEAPF32[$newOrigin>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $15 = (((($4) + (($i$01*36)|0)|0)) + 8|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $16 = +HEAPF32[$15>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $17 = $16 - $14; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$15>>2] = $17; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $18 = +HEAPF32[$5>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $19 = (((($4) + (($i$01*36)|0)|0)) + 12|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $20 = +HEAPF32[$19>>2]; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $21 = $20 - $18; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$19>>2] = $21; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $22 = (($i$01) + 1)|0; //@line 773 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  $23 = ($22|0)<($1|0); //@line 773 "libraries/Box2D/Collision/b2DynamicTree.cpp"
  if ($23) {
   $i$01 = $22;
  } else {
   break;
  }
 }
 return; //@line 778 "libraries/Box2D/Collision/b2DynamicTree.cpp"
}
function __Z14b2TimeOfImpactP11b2TOIOutputPK10b2TOIInput($output,$input) {
 $output = $output|0;
 $input = $input|0;
 var $$lcssa = 0, $$pre = 0.0, $$pre45 = 0.0, $$pre46 = 0.0, $$pre47 = 0.0, $$pre48 = 0.0, $$pre49 = 0.0, $$pre50 = 0.0, $$pre51 = 0.0, $$pre52 = 0.0, $$pre53 = 0.0, $$pre54 = 0.0, $$pre55 = 0.0, $$s1$0 = 0.0, $0 = 0, $1 = 0, $10 = 0.0, $100 = 0, $101 = 0, $102 = 0.0;
 var $103 = 0, $104 = 0, $105 = 0.0, $106 = 0, $107 = 0, $108 = 0.0, $109 = 0, $11 = 0.0, $110 = 0, $111 = 0, $112 = 0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0.0, $120 = 0.0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0.0, $125 = 0.0, $126 = 0, $127 = 0.0, $128 = 0.0, $129 = 0, $13 = 0.0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0.0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0.0, $146 = 0.0, $147 = 0, $148 = 0.0, $149 = 0.0, $15 = 0.0, $150 = 0.0, $151 = 0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0;
 var $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0.0, $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0, $34 = 0.0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0;
 var $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0;
 var $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $a1$0 = 0.0, $a2$0 = 0.0, $a2$0$t$0 = 0.0, $cache = 0, $distanceInput = 0, $distanceOutput = 0, $done$2$off0$ph = 0, $fcn = 0, $indexA = 0, $indexB = 0, $iter$0 = 0, $iter$0$lcssa = 0, $iter$114 = 0, $not$ = 0;
 var $pushBackIter$038 = 0, $rootIterCount$0 = 0, $s1$0 = 0.0, $s2$0 = 0.0, $s2$0$ = 0.0, $storemerge = 0.0, $storemerge18 = 0, $sweepA = 0, $sweepB = 0, $t$0 = 0.0, $t$0$a1$0 = 0.0, $t1$0 = 0.0, $t1$2$ph = 0.0, $t1$2$ph$lcssa73 = 0.0, $t2$037 = 0.0, $t2$27 = 0.0, $timer = 0, dest = 0, label = 0, sp = 0;
 var src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 320|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $timer = sp + 312|0;
 $sweepA = sp + 276|0;
 $sweepB = sp + 240|0;
 $cache = sp + 228|0;
 $distanceInput = sp + 136|0;
 $distanceOutput = sp + 112|0;
 $fcn = sp + 8|0;
 $indexA = sp + 4|0;
 $indexB = sp;
 __ZN7b2TimerC2Ev($timer); //@line 256 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $0 = HEAP32[164613]|0; //@line 258 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $1 = (($0) + 1)|0; //@line 258 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 HEAP32[164613] = $1; //@line 258 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 HEAP32[$output>>2] = 0; //@line 260 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $2 = ((($input)) + 128|0); //@line 261 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 261 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $4 = ((($output)) + 4|0); //@line 261 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 HEAP32[$4>>2] = $3; //@line 261 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $5 = ((($input)) + 28|0); //@line 264 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $6 = ((($input)) + 56|0); //@line 266 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 dest=$sweepA; src=$6; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 266 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $7 = ((($input)) + 92|0); //@line 267 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 dest=$sweepB; src=$7; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 267 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $8 = ((($sweepA)) + 24|0); //@line 715 "./libraries/Box2D/Common/b2Math.h"
 $9 = +HEAPF32[$8>>2]; //@line 715 "./libraries/Box2D/Common/b2Math.h"
 $10 = $9 / 6.2831854820251465; //@line 715 "./libraries/Box2D/Common/b2Math.h"
 $11 = (+Math_floor((+$10))); //@line 715 "./libraries/Box2D/Common/b2Math.h"
 $12 = $11 * 6.2831854820251465; //@line 715 "./libraries/Box2D/Common/b2Math.h"
 $13 = $9 - $12; //@line 716 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$8>>2] = $13; //@line 716 "./libraries/Box2D/Common/b2Math.h"
 $14 = ((($sweepA)) + 28|0); //@line 717 "./libraries/Box2D/Common/b2Math.h"
 $15 = +HEAPF32[$14>>2]; //@line 717 "./libraries/Box2D/Common/b2Math.h"
 $16 = $15 - $12; //@line 717 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$14>>2] = $16; //@line 717 "./libraries/Box2D/Common/b2Math.h"
 $17 = ((($sweepB)) + 24|0); //@line 715 "./libraries/Box2D/Common/b2Math.h"
 $18 = +HEAPF32[$17>>2]; //@line 715 "./libraries/Box2D/Common/b2Math.h"
 $19 = $18 / 6.2831854820251465; //@line 715 "./libraries/Box2D/Common/b2Math.h"
 $20 = (+Math_floor((+$19))); //@line 715 "./libraries/Box2D/Common/b2Math.h"
 $21 = $20 * 6.2831854820251465; //@line 715 "./libraries/Box2D/Common/b2Math.h"
 $22 = $18 - $21; //@line 716 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$17>>2] = $22; //@line 716 "./libraries/Box2D/Common/b2Math.h"
 $23 = ((($sweepB)) + 28|0); //@line 717 "./libraries/Box2D/Common/b2Math.h"
 $24 = +HEAPF32[$23>>2]; //@line 717 "./libraries/Box2D/Common/b2Math.h"
 $25 = $24 - $21; //@line 717 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$23>>2] = $25; //@line 717 "./libraries/Box2D/Common/b2Math.h"
 $26 = (HEAP32[tempDoublePtr>>2]=$3,+HEAPF32[tempDoublePtr>>2]); //@line 274 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $27 = ((($input)) + 24|0); //@line 276 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $28 = +HEAPF32[$27>>2]; //@line 276 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $29 = ((($input)) + 52|0); //@line 276 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $30 = +HEAPF32[$29>>2]; //@line 276 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $31 = $28 + $30; //@line 276 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $32 = $31 + -0.014999999664723873; //@line 277 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $33 = $32 < 0.004999999888241291; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $34 = $33 ? 0.004999999888241291 : $32; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $35 = $34 > 0.0012499999720603228; //@line 279 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 __Z10loveAssertbPKc($35,445944); //@line 279 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $36 = ((($cache)) + 4|0); //@line 287 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 HEAP16[$36>>1] = 0; //@line 287 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 ;HEAP32[$distanceInput>>2]=HEAP32[$input>>2]|0;HEAP32[$distanceInput+4>>2]=HEAP32[$input+4>>2]|0;HEAP32[$distanceInput+8>>2]=HEAP32[$input+8>>2]|0;HEAP32[$distanceInput+12>>2]=HEAP32[$input+12>>2]|0;HEAP32[$distanceInput+16>>2]=HEAP32[$input+16>>2]|0;HEAP32[$distanceInput+20>>2]=HEAP32[$input+20>>2]|0;HEAP32[$distanceInput+24>>2]=HEAP32[$input+24>>2]|0; //@line 289 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $37 = ((($distanceInput)) + 28|0); //@line 290 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 ;HEAP32[$37>>2]=HEAP32[$5>>2]|0;HEAP32[$37+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$37+8>>2]=HEAP32[$5+8>>2]|0;HEAP32[$37+12>>2]=HEAP32[$5+12>>2]|0;HEAP32[$37+16>>2]=HEAP32[$5+16>>2]|0;HEAP32[$37+20>>2]=HEAP32[$5+20>>2]|0;HEAP32[$37+24>>2]=HEAP32[$5+24>>2]|0; //@line 290 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $38 = ((($distanceInput)) + 88|0); //@line 291 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 HEAP8[$38>>0] = 0; //@line 291 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $39 = ((($sweepA)) + 8|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $40 = ((($sweepA)) + 12|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $41 = ((($sweepA)) + 16|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $42 = ((($sweepA)) + 20|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $43 = ((($sweepA)) + 4|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $44 = ((($sweepB)) + 8|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $45 = ((($sweepB)) + 12|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $46 = ((($sweepB)) + 16|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $47 = ((($sweepB)) + 20|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $48 = ((($sweepB)) + 4|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $49 = ((($distanceInput)) + 56|0); //@line 303 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $50 = ((($distanceInput)) + 60|0); //@line 303 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $51 = ((($distanceInput)) + 64|0); //@line 303 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $52 = ((($distanceInput)) + 68|0); //@line 303 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $53 = ((($distanceInput)) + 72|0); //@line 304 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $54 = ((($distanceInput)) + 76|0); //@line 304 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $55 = ((($distanceInput)) + 80|0); //@line 304 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $56 = ((($distanceInput)) + 84|0); //@line 304 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $57 = ((($distanceOutput)) + 16|0); //@line 309 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $58 = $34 + 0.0012499999720603228; //@line 317 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $59 = $34 + -0.0012499999720603228; //@line 376 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $$pre = +HEAPF32[$39>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $$pre45 = +HEAPF32[$40>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $$pre46 = +HEAPF32[$41>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $$pre47 = +HEAPF32[$42>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $$pre48 = +HEAPF32[$sweepA>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $$pre49 = +HEAPF32[$43>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $$pre50 = +HEAPF32[$44>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $$pre51 = +HEAPF32[$45>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $$pre52 = +HEAPF32[$46>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $$pre53 = +HEAPF32[$47>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $$pre54 = +HEAPF32[$sweepB>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $$pre55 = +HEAPF32[$48>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $iter$0 = 0;$t1$0 = 0.0;
 while(1) {
  $60 = 1.0 - $t1$0; //@line 694 "./libraries/Box2D/Common/b2Math.h"
  $61 = $60 * $$pre; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $62 = $60 * $$pre45; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $63 = $t1$0 * $$pre46; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $64 = $t1$0 * $$pre47; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $65 = $61 + $63; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $66 = $62 + $64; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $67 = $60 * $13; //@line 695 "./libraries/Box2D/Common/b2Math.h"
  $68 = $t1$0 * $16; //@line 695 "./libraries/Box2D/Common/b2Math.h"
  $69 = $67 + $68; //@line 695 "./libraries/Box2D/Common/b2Math.h"
  $70 = (+Math_sin((+$69))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
  $71 = (+Math_cos((+$69))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
  $72 = $71 * $$pre48; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $73 = $70 * $$pre49; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $74 = $72 - $73; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $75 = $70 * $$pre48; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $76 = $71 * $$pre49; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $77 = $75 + $76; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $78 = $65 - $74; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $79 = $66 - $77; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $80 = $60 * $$pre50; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $81 = $60 * $$pre51; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $82 = $t1$0 * $$pre52; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $83 = $t1$0 * $$pre53; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $84 = $80 + $82; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $85 = $81 + $83; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $86 = $60 * $22; //@line 695 "./libraries/Box2D/Common/b2Math.h"
  $87 = $t1$0 * $25; //@line 695 "./libraries/Box2D/Common/b2Math.h"
  $88 = $86 + $87; //@line 695 "./libraries/Box2D/Common/b2Math.h"
  $89 = (+Math_sin((+$88))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
  $90 = (+Math_cos((+$88))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
  $91 = $90 * $$pre54; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $92 = $89 * $$pre55; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $93 = $91 - $92; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $94 = $89 * $$pre54; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $95 = $90 * $$pre55; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $96 = $94 + $95; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $97 = $84 - $93; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $98 = $85 - $96; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$49>>2] = $78; //@line 303 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$50>>2] = $79; //@line 303 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$51>>2] = $70; //@line 303 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$52>>2] = $71; //@line 303 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$53>>2] = $97; //@line 304 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$54>>2] = $98; //@line 304 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$55>>2] = $89; //@line 304 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$56>>2] = $90; //@line 304 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  __Z10b2DistanceP16b2DistanceOutputP14b2SimplexCachePK15b2DistanceInput($distanceOutput,$cache,$distanceInput); //@line 306 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $99 = +HEAPF32[$57>>2]; //@line 309 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $100 = !($99 <= 0.0); //@line 309 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  if (!($100)) {
   $iter$0$lcssa = $iter$0;$storemerge = 0.0;$storemerge18 = 2;
   label = 21;
   break;
  }
  $101 = $99 < $58; //@line 317 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  if ($101) {
   $iter$0$lcssa = $iter$0;$storemerge = $t1$0;$storemerge18 = 3;
   label = 21;
   break;
  }
  (+__ZN20b2SeparationFunction10InitializeEPK14b2SimplexCachePK15b2DistanceProxyRK7b2SweepS5_S8_f($fcn,$cache,$input,$sweepA,$5,$sweepB,$t1$0)); //@line 327 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $102 = (+__ZNK20b2SeparationFunction17FindMinSeparationEPiS0_f($fcn,$indexA,$indexB,$26)); //@line 363 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $103 = $102 > $58; //@line 366 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  L5: do {
   if ($103) {
    label = 5;
   } else {
    $105 = $102;$pushBackIter$038 = 0;$t2$037 = $26;
    while(1) {
     $104 = $105 > $59; //@line 376 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
     if ($104) {
      $done$2$off0$ph = 0;$t1$2$ph = $t2$037;
      break L5;
     }
     $106 = HEAP32[$indexA>>2]|0; //@line 427 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
     $107 = HEAP32[$indexB>>2]|0; //@line 427 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
     $108 = (+__ZNK20b2SeparationFunction8EvaluateEiif($fcn,$106,$107,$t1$0)); //@line 384 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
     $109 = $108 < $59; //@line 388 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
     if ($109) {
      label = 8;
      break;
     }
     $110 = !($108 <= $58); //@line 397 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
     if ($110) {
      $a1$0 = $t1$0;$a2$0 = $t2$037;$rootIterCount$0 = 0;$s1$0 = $108;$s2$0 = $105;
     } else {
      label = 10;
      break;
     }
     while(1) {
      $111 = $rootIterCount$0 & 1; //@line 413 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
      $112 = ($111|0)==(0); //@line 413 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
      if ($112) {
       $119 = $a1$0 + $a2$0; //@line 421 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
       $120 = $119 * 0.5; //@line 421 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
       $t$0 = $120;
      } else {
       $113 = $34 - $s1$0; //@line 416 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
       $114 = $a2$0 - $a1$0; //@line 416 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
       $115 = $113 * $114; //@line 416 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
       $116 = $s2$0 - $s1$0; //@line 416 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
       $117 = $115 / $116; //@line 416 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
       $118 = $a1$0 + $117; //@line 416 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
       $t$0 = $118;
      }
      $121 = (($rootIterCount$0) + 1)|0; //@line 424 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
      $122 = HEAP32[164616]|0; //@line 425 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
      $123 = (($122) + 1)|0; //@line 425 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
      HEAP32[164616] = $123; //@line 425 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
      $124 = (+__ZNK20b2SeparationFunction8EvaluateEiif($fcn,$106,$107,$t$0)); //@line 427 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
      $125 = $124 - $34; //@line 429 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
      $126 = $125 > 0.0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
      $127 = -$125; //@line 618 "./libraries/Box2D/Common/b2Math.h"
      $128 = $126 ? $125 : $127; //@line 618 "./libraries/Box2D/Common/b2Math.h"
      $129 = $128 < 0.0012499999720603228; //@line 429 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
      if ($129) {
       $$lcssa = $121;$t2$27 = $t$0;
       break;
      }
      $130 = $124 > $34; //@line 437 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
      $s2$0$ = $130 ? $s2$0 : $124; //@line 441 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
      $$s1$0 = $130 ? $124 : $s1$0; //@line 441 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
      $t$0$a1$0 = $130 ? $t$0 : $a1$0; //@line 441 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
      $a2$0$t$0 = $130 ? $a2$0 : $t$0; //@line 441 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
      $not$ = ($121|0)==(50); //@line 448 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
      if ($not$) {
       $$lcssa = 50;$t2$27 = $t2$037;
       break;
      } else {
       $a1$0 = $t$0$a1$0;$a2$0 = $a2$0$t$0;$rootIterCount$0 = $121;$s1$0 = $$s1$0;$s2$0 = $s2$0$;
      }
     }
     $131 = HEAP32[164617]|0; //@line 454 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
     $132 = ($131|0)>($$lcssa|0); //@line 645 "./libraries/Box2D/Common/b2Math.h"
     $133 = $132 ? $131 : $$lcssa; //@line 645 "./libraries/Box2D/Common/b2Math.h"
     HEAP32[164617] = $133; //@line 454 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
     $134 = (($pushBackIter$038) + 1)|0; //@line 456 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
     $135 = ($134|0)==(8); //@line 458 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
     if ($135) {
      $done$2$off0$ph = 0;$t1$2$ph = $t1$0;
      break L5;
     }
     $136 = (+__ZNK20b2SeparationFunction17FindMinSeparationEPiS0_f($fcn,$indexA,$indexB,$t2$27)); //@line 363 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
     $137 = $136 > $58; //@line 366 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
     if ($137) {
      label = 5;
      break L5;
     } else {
      $105 = $136;$pushBackIter$038 = $134;$t2$037 = $t2$27;
     }
    }
    if ((label|0) == 8) {
     label = 0;
     HEAP32[$output>>2] = 1; //@line 390 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
     HEAPF32[$4>>2] = $t1$0; //@line 391 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
     $done$2$off0$ph = 1;$t1$2$ph = $t1$0;
     break;
    }
    else if ((label|0) == 10) {
     label = 0;
     HEAP32[$output>>2] = 3; //@line 400 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
     HEAPF32[$4>>2] = $t1$0; //@line 401 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
     $done$2$off0$ph = 1;$t1$2$ph = $t1$0;
     break;
    }
   }
  } while(0);
  if ((label|0) == 5) {
   label = 0;
   HEAP32[$output>>2] = 4; //@line 369 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
   HEAP32[$4>>2] = $3; //@line 370 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
   $done$2$off0$ph = 1;$t1$2$ph = $t1$0;
  }
  $138 = (($iter$0) + 1)|0; //@line 464 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $139 = HEAP32[164614]|0; //@line 465 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $140 = (($139) + 1)|0; //@line 465 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAP32[164614] = $140; //@line 465 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  if ($done$2$off0$ph) {
   $151 = $138;
   break;
  }
  $141 = ($138|0)==(20); //@line 472 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  if ($141) {
   $t1$2$ph$lcssa73 = $t1$2$ph;
   label = 20;
   break;
  }
  $iter$0 = $138;$t1$0 = $t1$2$ph;
 }
 if ((label|0) == 20) {
  HEAP32[$output>>2] = 1; //@line 475 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$4>>2] = $t1$2$ph$lcssa73; //@line 476 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $151 = 20;
 }
 else if ((label|0) == 21) {
  HEAP32[$output>>2] = $storemerge18; //@line 312 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$4>>2] = $storemerge; //@line 321 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $iter$114 = $iter$0$lcssa;
  $142 = HEAP32[164615]|0; //@line 481 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $143 = ($142|0)>($iter$114|0); //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $144 = $143 ? $142 : $iter$114; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  HEAP32[164615] = $144; //@line 481 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $145 = (+__ZNK7b2Timer15GetMillisecondsEv($timer)); //@line 483 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $146 = +HEAPF32[164612]; //@line 484 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $147 = $146 > $145; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $148 = $147 ? $146 : $145; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[164612] = $148; //@line 484 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $149 = +HEAPF32[164611]; //@line 485 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $150 = $145 + $149; //@line 485 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[164611] = $150; //@line 485 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  STACKTOP = sp;return; //@line 486 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 }
 $iter$114 = $151;
 $142 = HEAP32[164615]|0; //@line 481 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $143 = ($142|0)>($iter$114|0); //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $144 = $143 ? $142 : $iter$114; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 HEAP32[164615] = $144; //@line 481 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $145 = (+__ZNK7b2Timer15GetMillisecondsEv($timer)); //@line 483 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $146 = +HEAPF32[164612]; //@line 484 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $147 = $146 > $145; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $148 = $147 ? $146 : $145; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[164612] = $148; //@line 484 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $149 = +HEAPF32[164611]; //@line 485 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $150 = $145 + $149; //@line 485 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 HEAPF32[164611] = $150; //@line 485 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 STACKTOP = sp;return; //@line 486 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
}
function __ZN20b2SeparationFunction10InitializeEPK14b2SimplexCachePK15b2DistanceProxyRK7b2SweepS5_S8_f($this,$cache,$proxyA,$sweepA,$proxyB,$sweepB,$t1) {
 $this = $this|0;
 $cache = $cache|0;
 $proxyA = $proxyA|0;
 $sweepA = $sweepA|0;
 $proxyB = $proxyB|0;
 $sweepB = $sweepB|0;
 $t1 = +$t1;
 var $$0 = 0.0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0.0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0;
 var $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0.0, $120 = 0.0, $121 = 0, $122 = 0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0.0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0.0, $148 = 0, $149 = 0.0, $15 = 0.0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0.0, $158 = 0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0, $164 = 0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0;
 var $169 = 0, $17 = 0.0, $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0.0, $180 = 0.0, $181 = 0.0, $182 = 0.0, $183 = 0.0, $184 = 0.0, $185 = 0, $186 = 0;
 var $187 = 0.0, $188 = 0.0, $189 = 0.0, $19 = 0.0, $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0.0, $194 = 0.0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0.0, $200 = 0, $201 = 0, $202 = 0, $203 = 0.0;
 var $204 = 0, $205 = 0.0, $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0.0, $212 = 0.0, $213 = 0.0, $214 = 0.0, $215 = 0.0, $216 = 0.0, $217 = 0.0, $218 = 0.0, $219 = 0, $22 = 0.0, $220 = 0.0, $221 = 0.0;
 var $222 = 0.0, $223 = 0.0, $224 = 0.0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0.0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0.0, $236 = 0, $237 = 0.0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0.0, $248 = 0, $249 = 0.0, $25 = 0.0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0, $254 = 0, $255 = 0.0, $256 = 0.0, $257 = 0.0, $258 = 0.0;
 var $259 = 0, $26 = 0.0, $260 = 0.0, $261 = 0.0, $262 = 0.0, $263 = 0.0, $264 = 0.0, $265 = 0.0, $266 = 0.0, $267 = 0.0, $268 = 0.0, $269 = 0.0, $27 = 0.0, $270 = 0.0, $271 = 0.0, $272 = 0.0, $273 = 0.0, $274 = 0.0, $275 = 0, $276 = 0;
 var $277 = 0.0, $278 = 0.0, $279 = 0.0, $28 = 0.0, $280 = 0.0, $281 = 0.0, $282 = 0.0, $283 = 0.0, $284 = 0.0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0.0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0;
 var $295 = 0.0, $296 = 0, $297 = 0.0, $298 = 0.0, $299 = 0.0, $3 = 0, $30 = 0.0, $300 = 0.0, $301 = 0.0, $302 = 0.0, $303 = 0.0, $304 = 0.0, $305 = 0.0, $306 = 0.0, $307 = 0.0, $308 = 0.0, $309 = 0.0, $31 = 0.0, $310 = 0.0, $311 = 0;
 var $312 = 0.0, $313 = 0.0, $314 = 0.0, $315 = 0.0, $316 = 0.0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0, $45 = 0.0;
 var $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0, $56 = 0.0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0.0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0;
 var $64 = 0.0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0.0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0.0, $88 = 0, $89 = 0.0, $9 = 0.0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, dest = 0;
 var label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = $proxyA; //@line 49 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $0 = ((($this)) + 4|0); //@line 50 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 HEAP32[$0>>2] = $proxyB; //@line 50 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $1 = ((($cache)) + 4|0); //@line 51 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $2 = HEAP16[$1>>1]|0; //@line 51 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $$off = (($2) + -1)<<16>>16; //@line 52 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $3 = ($$off&65535)<(2); //@line 52 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 __Z10loveAssertbPKc($3,445963); //@line 52 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $4 = ((($this)) + 8|0); //@line 54 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 dest=$4; src=$sweepA; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 54 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $5 = ((($this)) + 44|0); //@line 55 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 dest=$5; src=$sweepB; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 55 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $6 = 1.0 - $t1; //@line 694 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($this)) + 16|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $9 = $6 * $8; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($this)) + 20|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $11 = +HEAPF32[$10>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 * $11; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($this)) + 24|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 * $t1; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $16 = ((($this)) + 28|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $17 = +HEAPF32[$16>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $18 = $17 * $t1; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $19 = $9 + $15; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $20 = $12 + $18; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($this)) + 32|0); //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $22 = +HEAPF32[$21>>2]; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $23 = $6 * $22; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($this)) + 36|0); //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $26 = $25 * $t1; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $27 = $23 + $26; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $28 = (+Math_sin((+$27))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $29 = (+Math_cos((+$27))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $30 = +HEAPF32[$4>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $31 = $29 * $30; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $32 = ((($this)) + 12|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $33 = +HEAPF32[$32>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $34 = $28 * $33; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $35 = $31 - $34; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $36 = $28 * $30; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $37 = $29 * $33; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $38 = $36 + $37; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $39 = $19 - $35; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $40 = $20 - $38; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $41 = ((($this)) + 52|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $42 = +HEAPF32[$41>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $43 = $6 * $42; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $44 = ((($this)) + 56|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $45 = +HEAPF32[$44>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $46 = $6 * $45; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $47 = ((($this)) + 60|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $48 = +HEAPF32[$47>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $49 = $48 * $t1; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $50 = ((($this)) + 64|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $51 = +HEAPF32[$50>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $52 = $51 * $t1; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $53 = $43 + $49; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $54 = $46 + $52; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $55 = ((($this)) + 68|0); //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $56 = +HEAPF32[$55>>2]; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $57 = $6 * $56; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $58 = ((($this)) + 72|0); //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $59 = +HEAPF32[$58>>2]; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $60 = $59 * $t1; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $61 = $57 + $60; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $62 = (+Math_sin((+$61))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $63 = (+Math_cos((+$61))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $64 = +HEAPF32[$5>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $65 = $63 * $64; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $66 = ((($this)) + 48|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $67 = +HEAPF32[$66>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $68 = $62 * $67; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $69 = $65 - $68; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $70 = $62 * $64; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $71 = $63 * $67; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $72 = $70 + $71; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $73 = $53 - $69; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $74 = $54 - $72; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $75 = ($2<<16>>16)==(1); //@line 61 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 if ($75) {
  $76 = ((($this)) + 80|0); //@line 63 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAP32[$76>>2] = 0; //@line 63 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $77 = HEAP32[$this>>2]|0; //@line 64 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $78 = ((($cache)) + 6|0); //@line 64 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $79 = HEAP8[$78>>0]|0; //@line 64 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $80 = $79&255; //@line 64 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $81 = ((($77)) + 20|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $82 = HEAP32[$81>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $83 = ($82|0)>($80|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  __Z10loveAssertbPKc($83,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $84 = ((($77)) + 16|0); //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $85 = HEAP32[$84>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $86 = (($85) + ($80<<3)|0); //@line 64 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $87 = +HEAPF32[$86>>2]; //@line 64 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $88 = (((($85) + ($80<<3)|0)) + 4|0); //@line 64 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $89 = +HEAPF32[$88>>2]; //@line 64 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $90 = HEAP32[$0>>2]|0; //@line 65 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $91 = ((($cache)) + 9|0); //@line 65 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $92 = HEAP8[$91>>0]|0; //@line 65 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $93 = $92&255; //@line 65 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $94 = ((($90)) + 20|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $95 = HEAP32[$94>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $96 = ($95|0)>($93|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  __Z10loveAssertbPKc($96,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $97 = ((($90)) + 16|0); //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $98 = HEAP32[$97>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $99 = (($98) + ($93<<3)|0); //@line 65 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $100 = +HEAPF32[$99>>2]; //@line 65 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $101 = (((($98) + ($93<<3)|0)) + 4|0); //@line 65 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $102 = +HEAPF32[$101>>2]; //@line 65 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $103 = $29 * $87; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $104 = $28 * $89; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $105 = $103 - $104; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $106 = $39 + $105; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $107 = $28 * $87; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $108 = $29 * $89; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $109 = $107 + $108; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $110 = $40 + $109; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $111 = $63 * $100; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $112 = $62 * $102; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $113 = $111 - $112; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $114 = $73 + $113; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $115 = $62 * $100; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $116 = $63 * $102; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $117 = $115 + $116; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $118 = $74 + $117; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $119 = $114 - $106; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $120 = $118 - $110; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $121 = ((($this)) + 92|0); //@line 68 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$121>>2] = $119; //@line 68 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $122 = ((($this)) + 96|0); //@line 68 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$122>>2] = $120; //@line 68 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $123 = $119 * $119; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $124 = $120 * $120; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $125 = $123 + $124; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $126 = (+Math_sqrt((+$125))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $127 = $126 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
  if ($127) {
   $$0 = 0.0;
   return (+$$0); //@line 122 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  }
  $128 = 1.0 / $126; //@line 121 "./libraries/Box2D/Common/b2Math.h"
  $129 = $119 * $128; //@line 122 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$121>>2] = $129; //@line 122 "./libraries/Box2D/Common/b2Math.h"
  $130 = $128 * $120; //@line 123 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$122>>2] = $130; //@line 123 "./libraries/Box2D/Common/b2Math.h"
  $$0 = $126;
  return (+$$0); //@line 122 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 }
 $131 = ((($cache)) + 6|0); //@line 72 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $132 = HEAP16[$131>>1]|0; //@line 72 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $133 = $132&255; //@line 72 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $134 = ($132&65535) >>> 8; //@line 72 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $135 = $134&255; //@line 72 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $136 = ($133<<24>>24)==($135<<24>>24); //@line 72 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $137 = ((($this)) + 80|0); //@line 75 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 if ($136) {
  HEAP32[$137>>2] = 2; //@line 75 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $138 = ((($cache)) + 9|0); //@line 76 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $139 = HEAP8[$138>>0]|0; //@line 76 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $140 = $139&255; //@line 76 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $141 = ((($proxyB)) + 20|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $142 = HEAP32[$141>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $143 = ($142|0)>($140|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  __Z10loveAssertbPKc($143,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $144 = ((($proxyB)) + 16|0); //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $145 = HEAP32[$144>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $146 = (($145) + ($140<<3)|0); //@line 76 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $147 = +HEAPF32[$146>>2]; //@line 76 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $148 = (((($145) + ($140<<3)|0)) + 4|0); //@line 76 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $149 = +HEAPF32[$148>>2]; //@line 76 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $150 = ((($cache)) + 10|0); //@line 77 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $151 = HEAP8[$150>>0]|0; //@line 77 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $152 = $151&255; //@line 77 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $153 = HEAP32[$141>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $154 = ($153|0)>($152|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  __Z10loveAssertbPKc($154,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $155 = HEAP32[$144>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $156 = (($155) + ($152<<3)|0); //@line 77 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $157 = +HEAPF32[$156>>2]; //@line 77 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $158 = (((($155) + ($152<<3)|0)) + 4|0); //@line 77 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $159 = +HEAPF32[$158>>2]; //@line 77 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $160 = $157 - $147; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $161 = $159 - $149; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $162 = -$160; //@line 421 "./libraries/Box2D/Common/b2Math.h"
  $163 = ((($this)) + 92|0); //@line 79 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$163>>2] = $161; //@line 79 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $164 = ((($this)) + 96|0); //@line 79 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$164>>2] = $162; //@line 79 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $165 = $161 * $161; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $166 = $160 * $160; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $167 = $166 + $165; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $168 = (+Math_sqrt((+$167))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $169 = $168 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
  if ($169) {
   $174 = $161;$176 = $162;
  } else {
   $170 = 1.0 / $168; //@line 121 "./libraries/Box2D/Common/b2Math.h"
   $171 = $161 * $170; //@line 122 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$163>>2] = $171; //@line 122 "./libraries/Box2D/Common/b2Math.h"
   $172 = $170 * $162; //@line 123 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$164>>2] = $172; //@line 123 "./libraries/Box2D/Common/b2Math.h"
   $174 = $171;$176 = $172;
  }
  $173 = $63 * $174; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $175 = $62 * $176; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $177 = $173 - $175; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $178 = $62 * $174; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $179 = $63 * $176; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $180 = $178 + $179; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $181 = $147 + $157; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $182 = $149 + $159; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $183 = $181 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $184 = $182 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $185 = ((($this)) + 84|0); //@line 83 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$185>>2] = $183; //@line 83 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $186 = ((($this)) + 88|0); //@line 83 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$186>>2] = $184; //@line 83 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $187 = $63 * $183; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $188 = $62 * $184; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $189 = $187 - $188; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $190 = $73 + $189; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $191 = $62 * $183; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $192 = $63 * $184; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $193 = $191 + $192; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $194 = $74 + $193; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $195 = HEAP8[$131>>0]|0; //@line 86 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $196 = $195&255; //@line 86 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $197 = ((($proxyA)) + 20|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $198 = HEAP32[$197>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $199 = ($198|0)>($196|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  __Z10loveAssertbPKc($199,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $200 = ((($proxyA)) + 16|0); //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $201 = HEAP32[$200>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $202 = (($201) + ($196<<3)|0); //@line 86 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $203 = +HEAPF32[$202>>2]; //@line 86 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $204 = (((($201) + ($196<<3)|0)) + 4|0); //@line 86 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $205 = +HEAPF32[$204>>2]; //@line 86 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $206 = $29 * $203; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $207 = $28 * $205; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $208 = $206 - $207; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $209 = $39 + $208; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $210 = $28 * $203; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $211 = $29 * $205; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $212 = $210 + $211; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $213 = $40 + $212; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $214 = $209 - $190; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $215 = $213 - $194; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $216 = $177 * $214; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $217 = $180 * $215; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $218 = $216 + $217; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $219 = $218 < 0.0; //@line 90 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  if (!($219)) {
   $$0 = $218;
   return (+$$0); //@line 122 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  }
  $220 = +HEAPF32[$163>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $221 = -$220; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $222 = +HEAPF32[$164>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $223 = -$222; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$163>>2] = $221; //@line 92 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$164>>2] = $223; //@line 92 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $224 = -$218; //@line 93 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $$0 = $224;
  return (+$$0); //@line 122 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 } else {
  $225 = ((($cache)) + 7|0); //@line 72 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAP32[$137>>2] = 1; //@line 100 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $226 = HEAP32[$this>>2]|0; //@line 101 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $227 = HEAP8[$131>>0]|0; //@line 101 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $228 = $227&255; //@line 101 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $229 = ((($226)) + 20|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $230 = HEAP32[$229>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $231 = ($230|0)>($228|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  __Z10loveAssertbPKc($231,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $232 = ((($226)) + 16|0); //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $233 = HEAP32[$232>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $234 = (($233) + ($228<<3)|0); //@line 101 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $235 = +HEAPF32[$234>>2]; //@line 101 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $236 = (((($233) + ($228<<3)|0)) + 4|0); //@line 101 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $237 = +HEAPF32[$236>>2]; //@line 101 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $238 = HEAP32[$this>>2]|0; //@line 102 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $239 = HEAP8[$225>>0]|0; //@line 102 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $240 = $239&255; //@line 102 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $241 = ((($238)) + 20|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $242 = HEAP32[$241>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $243 = ($242|0)>($240|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  __Z10loveAssertbPKc($243,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $244 = ((($238)) + 16|0); //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $245 = HEAP32[$244>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $246 = (($245) + ($240<<3)|0); //@line 102 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $247 = +HEAPF32[$246>>2]; //@line 102 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $248 = (((($245) + ($240<<3)|0)) + 4|0); //@line 102 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $249 = +HEAPF32[$248>>2]; //@line 102 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $250 = $247 - $235; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $251 = $249 - $237; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $252 = -$250; //@line 421 "./libraries/Box2D/Common/b2Math.h"
  $253 = ((($this)) + 92|0); //@line 104 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$253>>2] = $251; //@line 104 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $254 = ((($this)) + 96|0); //@line 104 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$254>>2] = $252; //@line 104 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $255 = $251 * $251; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $256 = $250 * $250; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $257 = $256 + $255; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $258 = (+Math_sqrt((+$257))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $259 = $258 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
  if ($259) {
   $264 = $251;$266 = $252;
  } else {
   $260 = 1.0 / $258; //@line 121 "./libraries/Box2D/Common/b2Math.h"
   $261 = $251 * $260; //@line 122 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$253>>2] = $261; //@line 122 "./libraries/Box2D/Common/b2Math.h"
   $262 = $260 * $252; //@line 123 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$254>>2] = $262; //@line 123 "./libraries/Box2D/Common/b2Math.h"
   $264 = $261;$266 = $262;
  }
  $263 = $29 * $264; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $265 = $28 * $266; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $267 = $263 - $265; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $268 = $28 * $264; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $269 = $29 * $266; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $270 = $268 + $269; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $271 = $235 + $247; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $272 = $237 + $249; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $273 = $271 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $274 = $272 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $275 = ((($this)) + 84|0); //@line 108 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$275>>2] = $273; //@line 108 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $276 = ((($this)) + 88|0); //@line 108 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$276>>2] = $274; //@line 108 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $277 = $29 * $273; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $278 = $28 * $274; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $279 = $277 - $278; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $280 = $39 + $279; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $281 = $28 * $273; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $282 = $29 * $274; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $283 = $281 + $282; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $284 = $40 + $283; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $285 = HEAP32[$0>>2]|0; //@line 111 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $286 = ((($cache)) + 9|0); //@line 111 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $287 = HEAP8[$286>>0]|0; //@line 111 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $288 = $287&255; //@line 111 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $289 = ((($285)) + 20|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $290 = HEAP32[$289>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $291 = ($290|0)>($288|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  __Z10loveAssertbPKc($291,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $292 = ((($285)) + 16|0); //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $293 = HEAP32[$292>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $294 = (($293) + ($288<<3)|0); //@line 111 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $295 = +HEAPF32[$294>>2]; //@line 111 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $296 = (((($293) + ($288<<3)|0)) + 4|0); //@line 111 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $297 = +HEAPF32[$296>>2]; //@line 111 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $298 = $63 * $295; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $299 = $62 * $297; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $300 = $298 - $299; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $301 = $73 + $300; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $302 = $62 * $295; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $303 = $63 * $297; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $304 = $302 + $303; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $305 = $74 + $304; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $306 = $301 - $280; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $307 = $305 - $284; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $308 = $267 * $306; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $309 = $270 * $307; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $310 = $308 + $309; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $311 = $310 < 0.0; //@line 115 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  if (!($311)) {
   $$0 = $310;
   return (+$$0); //@line 122 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  }
  $312 = +HEAPF32[$253>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $313 = -$312; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $314 = +HEAPF32[$254>>2]; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $315 = -$314; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$253>>2] = $313; //@line 117 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAPF32[$254>>2] = $315; //@line 117 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $316 = -$310; //@line 118 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $$0 = $316;
  return (+$$0); //@line 122 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 }
 return +(0.0);
}
function __ZNK20b2SeparationFunction17FindMinSeparationEPiS0_f($this,$indexA,$indexB,$t) {
 $this = $this|0;
 $indexA = $indexA|0;
 $indexB = $indexB|0;
 $t = +$t;
 var $$0 = 0.0, $0 = 0.0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0, $104 = 0.0, $105 = 0.0, $106 = 0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0.0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0.0, $12 = 0.0, $120 = 0.0, $121 = 0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0, $126 = 0.0, $127 = 0.0, $128 = 0, $129 = 0.0, $13 = 0.0, $130 = 0.0, $131 = 0.0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0.0, $140 = 0, $141 = 0.0, $142 = 0, $143 = 0.0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0.0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0.0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0.0;
 var $17 = 0.0, $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0, $181 = 0.0, $182 = 0.0, $183 = 0, $184 = 0.0, $185 = 0.0, $186 = 0.0, $187 = 0.0;
 var $188 = 0.0, $189 = 0.0, $19 = 0.0, $190 = 0, $191 = 0.0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0.0, $196 = 0.0, $197 = 0.0, $198 = 0.0, $199 = 0.0, $2 = 0.0, $20 = 0.0, $200 = 0.0, $201 = 0.0, $202 = 0.0, $203 = 0.0, $204 = 0.0;
 var $205 = 0.0, $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0.0, $21 = 0.0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0.0, $218 = 0.0, $219 = 0, $22 = 0.0, $220 = 0.0, $221 = 0.0, $222 = 0.0;
 var $223 = 0, $224 = 0.0, $225 = 0.0, $226 = 0, $227 = 0.0, $228 = 0.0, $229 = 0.0, $23 = 0.0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0.0, $239 = 0, $24 = 0, $240 = 0.0;
 var $241 = 0.0, $242 = 0.0, $243 = 0.0, $244 = 0.0, $245 = 0.0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0.0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0.0, $254 = 0, $255 = 0.0, $256 = 0.0, $257 = 0, $258 = 0.0, $259 = 0.0;
 var $26 = 0.0, $260 = 0.0, $261 = 0.0, $262 = 0.0, $263 = 0.0, $264 = 0, $265 = 0.0, $266 = 0.0, $267 = 0, $268 = 0.0, $269 = 0.0, $27 = 0, $270 = 0.0, $271 = 0.0, $272 = 0.0, $273 = 0.0, $274 = 0.0, $275 = 0.0, $276 = 0.0, $277 = 0.0;
 var $278 = 0.0, $279 = 0.0, $28 = 0.0, $280 = 0.0, $281 = 0.0, $282 = 0.0, $283 = 0.0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0.0, $290 = 0.0, $291 = 0.0, $292 = 0, $293 = 0.0, $294 = 0.0, $295 = 0.0;
 var $296 = 0, $297 = 0.0, $298 = 0.0, $299 = 0, $3 = 0.0, $30 = 0.0, $300 = 0.0, $301 = 0.0, $302 = 0.0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0.0, $310 = 0, $311 = 0.0, $312 = 0;
 var $313 = 0.0, $314 = 0.0, $315 = 0.0, $316 = 0.0, $317 = 0.0, $318 = 0.0, $319 = 0.0, $32 = 0.0, $320 = 0.0, $321 = 0.0, $322 = 0.0, $323 = 0.0, $324 = 0.0, $325 = 0.0, $326 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0.0;
 var $38 = 0.0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0.0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0.0;
 var $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0, $6 = 0.0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0.0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0.0, $98 = 0.0, $99 = 0, $bestIndex$0$lcssa$i15 = 0, $bestIndex$0$lcssa$i24 = 0, $bestIndex$0$lcssa$i43 = 0, $bestIndex$0$lcssa$i44 = 0, $bestIndex$0$lcssa$i641 = 0, $bestIndex$0$lcssa$i642 = 0, $bestIndex$03$i = 0, $bestIndex$03$i16 = 0, $bestIndex$03$i25 = 0, $bestIndex$03$i7 = 0, $bestIndex$1$i = 0, $bestIndex$1$i$lcssa = 0;
 var $bestIndex$1$i11 = 0, $bestIndex$1$i11$lcssa = 0, $bestIndex$1$i20 = 0, $bestIndex$1$i29 = 0, $bestValue$01$i = 0.0, $bestValue$01$i18 = 0.0, $bestValue$01$i27 = 0.0, $bestValue$01$i9 = 0.0, $bestValue$1$i = 0.0, $bestValue$1$i10 = 0.0, $bestValue$1$i19 = 0.0, $bestValue$1$i28 = 0.0, $exitcond = 0, $exitcond45 = 0, $exitcond46 = 0, $exitcond47 = 0, $i$02$i = 0, $i$02$i17 = 0, $i$02$i26 = 0, $i$02$i8 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = 1.0 - $t; //@line 694 "./libraries/Box2D/Common/b2Math.h"
 $1 = ((($this)) + 16|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $2 = +HEAPF32[$1>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $3 = $0 * $2; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 20|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $6 = $0 * $5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($this)) + 24|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $9 = $8 * $t; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($this)) + 28|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $11 = +HEAPF32[$10>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $12 = $11 * $t; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $13 = $3 + $9; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $14 = $6 + $12; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $15 = ((($this)) + 32|0); //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $16 = +HEAPF32[$15>>2]; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $17 = $0 * $16; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $18 = ((($this)) + 36|0); //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $19 = +HEAPF32[$18>>2]; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $20 = $19 * $t; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $21 = $17 + $20; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $22 = (+Math_sin((+$21))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $23 = (+Math_cos((+$21))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($this)) + 8|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $26 = $23 * $25; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($this)) + 12|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $28 = +HEAPF32[$27>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $29 = $22 * $28; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $30 = $26 - $29; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $31 = $22 * $25; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $32 = $23 * $28; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $33 = $31 + $32; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $34 = $13 - $30; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $35 = $14 - $33; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $36 = ((($this)) + 52|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $37 = +HEAPF32[$36>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $38 = $0 * $37; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $39 = ((($this)) + 56|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $40 = +HEAPF32[$39>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $41 = $0 * $40; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $42 = ((($this)) + 60|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $43 = +HEAPF32[$42>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $44 = $43 * $t; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $45 = ((($this)) + 64|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $46 = +HEAPF32[$45>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $47 = $46 * $t; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $48 = $38 + $44; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $49 = $41 + $47; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $50 = ((($this)) + 68|0); //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $51 = +HEAPF32[$50>>2]; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $52 = $0 * $51; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $53 = ((($this)) + 72|0); //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $54 = +HEAPF32[$53>>2]; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $55 = $54 * $t; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $56 = $52 + $55; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $57 = (+Math_sin((+$56))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $58 = (+Math_cos((+$56))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $59 = ((($this)) + 44|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $60 = +HEAPF32[$59>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $61 = $58 * $60; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $62 = ((($this)) + 48|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $63 = +HEAPF32[$62>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $64 = $57 * $63; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $65 = $61 - $64; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $66 = $57 * $60; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $67 = $58 * $63; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $68 = $66 + $67; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $69 = $48 - $65; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $70 = $49 - $68; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $71 = ((($this)) + 80|0); //@line 131 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $72 = HEAP32[$71>>2]|0; //@line 131 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 switch ($72|0) {
 case 0:  {
  $73 = ((($this)) + 92|0); //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $74 = +HEAPF32[$73>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $75 = $23 * $74; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $76 = ((($this)) + 96|0); //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $77 = +HEAPF32[$76>>2]; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $78 = $22 * $77; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $79 = $75 + $78; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $80 = $22 * $74; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $81 = $23 * $77; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $82 = $81 - $80; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $83 = -$74; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $84 = -$77; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $85 = $58 * $83; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $86 = $57 * $84; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $87 = $85 + $86; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $88 = $57 * $83; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $89 = $58 * $84; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $90 = $89 - $88; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $91 = HEAP32[$this>>2]|0; //@line 141 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $92 = ((($91)) + 16|0); //@line 110 "./libraries/Box2D/Collision/b2Distance.h"
  $93 = HEAP32[$92>>2]|0; //@line 113 "./libraries/Box2D/Collision/b2Distance.h"
  $94 = ((($91)) + 20|0); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
  $95 = HEAP32[$94>>2]|0; //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
  $96 = ($95|0)>(1); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
  if ($96) {
   $97 = +HEAPF32[$93>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $98 = $79 * $97; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $99 = ((($93)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $100 = +HEAPF32[$99>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $101 = $82 * $100; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $102 = $98 + $101; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $bestIndex$03$i16 = 0;$bestValue$01$i18 = $102;$i$02$i17 = 1;
   while(1) {
    $103 = (($93) + ($i$02$i17<<3)|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $104 = +HEAPF32[$103>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $105 = $79 * $104; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $106 = (((($93) + ($i$02$i17<<3)|0)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $107 = +HEAPF32[$106>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $108 = $82 * $107; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $109 = $105 + $108; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $110 = $109 > $bestValue$01$i18; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $bestValue$1$i19 = $110 ? $109 : $bestValue$01$i18; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $bestIndex$1$i20 = $110 ? $i$02$i17 : $bestIndex$03$i16; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $111 = (($i$02$i17) + 1)|0; //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
    $exitcond45 = ($111|0)==($95|0); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
    if ($exitcond45) {
     $bestIndex$0$lcssa$i15 = $bestIndex$1$i20;
     break;
    } else {
     $bestIndex$03$i16 = $bestIndex$1$i20;$bestValue$01$i18 = $bestValue$1$i19;$i$02$i17 = $111;
    }
   }
  } else {
   $bestIndex$0$lcssa$i15 = 0;
  }
  HEAP32[$indexA>>2] = $bestIndex$0$lcssa$i15; //@line 138 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $112 = ((($this)) + 4|0); //@line 139 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $113 = HEAP32[$112>>2]|0; //@line 139 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $114 = ((($113)) + 16|0); //@line 110 "./libraries/Box2D/Collision/b2Distance.h"
  $115 = HEAP32[$114>>2]|0; //@line 113 "./libraries/Box2D/Collision/b2Distance.h"
  $116 = ((($113)) + 20|0); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
  $117 = HEAP32[$116>>2]|0; //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
  $118 = ($117|0)>(1); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
  if ($118) {
   $119 = +HEAPF32[$115>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $120 = $87 * $119; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $121 = ((($115)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $122 = +HEAPF32[$121>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $123 = $90 * $122; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $124 = $120 + $123; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $bestIndex$03$i25 = 0;$bestValue$01$i27 = $124;$i$02$i26 = 1;
   while(1) {
    $125 = (($115) + ($i$02$i26<<3)|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $126 = +HEAPF32[$125>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $127 = $87 * $126; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $128 = (((($115) + ($i$02$i26<<3)|0)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $129 = +HEAPF32[$128>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $130 = $90 * $129; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $131 = $127 + $130; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $132 = $131 > $bestValue$01$i27; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $bestValue$1$i28 = $132 ? $131 : $bestValue$01$i27; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $bestIndex$1$i29 = $132 ? $i$02$i26 : $bestIndex$03$i25; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $133 = (($i$02$i26) + 1)|0; //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
    $exitcond = ($133|0)==($117|0); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
    if ($exitcond) {
     $bestIndex$0$lcssa$i24 = $bestIndex$1$i29;
     break;
    } else {
     $bestIndex$03$i25 = $bestIndex$1$i29;$bestValue$01$i27 = $bestValue$1$i28;$i$02$i26 = $133;
    }
   }
  } else {
   $bestIndex$0$lcssa$i24 = 0;
  }
  HEAP32[$indexB>>2] = $bestIndex$0$lcssa$i24; //@line 139 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $134 = HEAP32[$indexA>>2]|0; //@line 141 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $135 = ($134|0)>(-1); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  if ($135) {
   $136 = HEAP32[$94>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $137 = ($136|0)>($134|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $138 = $137;
  } else {
   $138 = 0;
  }
  __Z10loveAssertbPKc($138,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $139 = HEAP32[$92>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $140 = (($139) + ($134<<3)|0); //@line 141 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $141 = +HEAPF32[$140>>2]; //@line 141 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $142 = (((($139) + ($134<<3)|0)) + 4|0); //@line 141 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $143 = +HEAPF32[$142>>2]; //@line 141 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $144 = HEAP32[$112>>2]|0; //@line 142 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $145 = HEAP32[$indexB>>2]|0; //@line 142 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $146 = ($145|0)>(-1); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  if ($146) {
   $147 = ((($144)) + 20|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $148 = HEAP32[$147>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $149 = ($148|0)>($145|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $150 = $149;
  } else {
   $150 = 0;
  }
  __Z10loveAssertbPKc($150,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $151 = ((($144)) + 16|0); //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $152 = HEAP32[$151>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $153 = (($152) + ($145<<3)|0); //@line 142 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $154 = +HEAPF32[$153>>2]; //@line 142 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $155 = (((($152) + ($145<<3)|0)) + 4|0); //@line 142 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $156 = +HEAPF32[$155>>2]; //@line 142 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $157 = $23 * $141; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $158 = $22 * $143; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $159 = $157 - $158; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $160 = $34 + $159; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $161 = $22 * $141; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $162 = $23 * $143; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $163 = $161 + $162; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $164 = $35 + $163; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $165 = $58 * $154; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $166 = $57 * $156; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $167 = $165 - $166; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $168 = $69 + $167; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $169 = $57 * $154; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $170 = $58 * $156; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $171 = $169 + $170; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $172 = $70 + $171; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $173 = $168 - $160; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $174 = $172 - $164; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $175 = +HEAPF32[$73>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $176 = $175 * $173; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $177 = +HEAPF32[$76>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $178 = $177 * $174; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $179 = $176 + $178; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $$0 = $179;
  return (+$$0); //@line 191 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  break;
 }
 case 1:  {
  $180 = ((($this)) + 92|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $181 = +HEAPF32[$180>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $182 = $23 * $181; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $183 = ((($this)) + 96|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $184 = +HEAPF32[$183>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $185 = $22 * $184; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $186 = $182 - $185; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $187 = $22 * $181; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $188 = $23 * $184; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $189 = $187 + $188; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $190 = ((($this)) + 84|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $191 = +HEAPF32[$190>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $192 = $23 * $191; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $193 = ((($this)) + 88|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $194 = +HEAPF32[$193>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $195 = $22 * $194; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $196 = $192 - $195; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $197 = $34 + $196; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $198 = $22 * $191; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $199 = $23 * $194; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $200 = $198 + $199; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $201 = $35 + $200; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $202 = -$186; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $203 = -$189; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $204 = $58 * $202; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $205 = $57 * $203; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $206 = $204 + $205; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $207 = $57 * $202; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $208 = $58 * $203; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $209 = $208 - $207; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  HEAP32[$indexA>>2] = -1; //@line 158 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $210 = ((($this)) + 4|0); //@line 159 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $211 = HEAP32[$210>>2]|0; //@line 161 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $212 = ((($211)) + 16|0); //@line 110 "./libraries/Box2D/Collision/b2Distance.h"
  $213 = HEAP32[$212>>2]|0; //@line 113 "./libraries/Box2D/Collision/b2Distance.h"
  $214 = ((($211)) + 20|0); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
  $215 = HEAP32[$214>>2]|0; //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
  $216 = ($215|0)>(1); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
  if ($216) {
   $217 = +HEAPF32[$213>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $218 = $206 * $217; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $219 = ((($213)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $220 = +HEAPF32[$219>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $221 = $209 * $220; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $222 = $218 + $221; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $bestIndex$03$i7 = 0;$bestValue$01$i9 = $222;$i$02$i8 = 1;
   while(1) {
    $223 = (($213) + ($i$02$i8<<3)|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $224 = +HEAPF32[$223>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $225 = $206 * $224; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $226 = (((($213) + ($i$02$i8<<3)|0)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $227 = +HEAPF32[$226>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $228 = $209 * $227; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $229 = $225 + $228; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $230 = $229 > $bestValue$01$i9; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $bestValue$1$i10 = $230 ? $229 : $bestValue$01$i9; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $bestIndex$1$i11 = $230 ? $i$02$i8 : $bestIndex$03$i7; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $231 = (($i$02$i8) + 1)|0; //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
    $exitcond46 = ($231|0)==($215|0); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
    if ($exitcond46) {
     $bestIndex$1$i11$lcssa = $bestIndex$1$i11;
     break;
    } else {
     $bestIndex$03$i7 = $bestIndex$1$i11;$bestValue$01$i9 = $bestValue$1$i10;$i$02$i8 = $231;
    }
   }
   HEAP32[$indexB>>2] = $bestIndex$1$i11$lcssa; //@line 159 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
   $232 = ($bestIndex$1$i11$lcssa|0)>(-1); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   if ($232) {
    $bestIndex$0$lcssa$i641 = $bestIndex$1$i11$lcssa;
    label = 18;
   } else {
    $235 = 0;$bestIndex$0$lcssa$i642 = $bestIndex$1$i11$lcssa;
   }
  } else {
   HEAP32[$indexB>>2] = 0; //@line 159 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
   $bestIndex$0$lcssa$i641 = 0;
   label = 18;
  }
  if ((label|0) == 18) {
   $233 = HEAP32[$214>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $234 = ($233|0)>($bestIndex$0$lcssa$i641|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $235 = $234;$bestIndex$0$lcssa$i642 = $bestIndex$0$lcssa$i641;
  }
  __Z10loveAssertbPKc($235,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $236 = HEAP32[$212>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $237 = (($236) + ($bestIndex$0$lcssa$i642<<3)|0); //@line 161 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $238 = +HEAPF32[$237>>2]; //@line 161 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $239 = (((($236) + ($bestIndex$0$lcssa$i642<<3)|0)) + 4|0); //@line 161 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $240 = +HEAPF32[$239>>2]; //@line 161 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $241 = $58 * $238; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $242 = $57 * $240; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $243 = $241 - $242; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $244 = $69 + $243; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $245 = $57 * $238; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $246 = $58 * $240; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $247 = $245 + $246; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $248 = $70 + $247; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $249 = $244 - $197; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $250 = $248 - $201; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $251 = $186 * $249; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $252 = $189 * $250; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $253 = $251 + $252; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $$0 = $253;
  return (+$$0); //@line 191 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  break;
 }
 case 2:  {
  $254 = ((($this)) + 92|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $255 = +HEAPF32[$254>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $256 = $58 * $255; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $257 = ((($this)) + 96|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $258 = +HEAPF32[$257>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $259 = $57 * $258; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $260 = $256 - $259; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $261 = $57 * $255; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $262 = $58 * $258; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $263 = $261 + $262; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $264 = ((($this)) + 84|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $265 = +HEAPF32[$264>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $266 = $58 * $265; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $267 = ((($this)) + 88|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $268 = +HEAPF32[$267>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $269 = $57 * $268; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $270 = $266 - $269; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $271 = $69 + $270; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $272 = $57 * $265; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $273 = $58 * $268; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $274 = $272 + $273; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $275 = $70 + $274; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $276 = -$260; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $277 = -$263; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $278 = $23 * $276; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $279 = $22 * $277; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $280 = $278 + $279; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $281 = $22 * $276; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $282 = $23 * $277; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $283 = $282 - $281; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  HEAP32[$indexB>>2] = -1; //@line 175 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $284 = HEAP32[$this>>2]|0; //@line 178 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $285 = ((($284)) + 16|0); //@line 110 "./libraries/Box2D/Collision/b2Distance.h"
  $286 = HEAP32[$285>>2]|0; //@line 113 "./libraries/Box2D/Collision/b2Distance.h"
  $287 = ((($284)) + 20|0); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
  $288 = HEAP32[$287>>2]|0; //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
  $289 = ($288|0)>(1); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
  if ($289) {
   $290 = +HEAPF32[$286>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $291 = $280 * $290; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $292 = ((($286)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $293 = +HEAPF32[$292>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $294 = $283 * $293; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $295 = $291 + $294; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $bestIndex$03$i = 0;$bestValue$01$i = $295;$i$02$i = 1;
   while(1) {
    $296 = (($286) + ($i$02$i<<3)|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $297 = +HEAPF32[$296>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $298 = $280 * $297; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $299 = (((($286) + ($i$02$i<<3)|0)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $300 = +HEAPF32[$299>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $301 = $283 * $300; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $302 = $298 + $301; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $303 = $302 > $bestValue$01$i; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $bestValue$1$i = $303 ? $302 : $bestValue$01$i; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $bestIndex$1$i = $303 ? $i$02$i : $bestIndex$03$i; //@line 114 "./libraries/Box2D/Collision/b2Distance.h"
    $304 = (($i$02$i) + 1)|0; //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
    $exitcond47 = ($304|0)==($288|0); //@line 111 "./libraries/Box2D/Collision/b2Distance.h"
    if ($exitcond47) {
     $bestIndex$1$i$lcssa = $bestIndex$1$i;
     break;
    } else {
     $bestIndex$03$i = $bestIndex$1$i;$bestValue$01$i = $bestValue$1$i;$i$02$i = $304;
    }
   }
   HEAP32[$indexA>>2] = $bestIndex$1$i$lcssa; //@line 176 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
   $305 = ($bestIndex$1$i$lcssa|0)>(-1); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   if ($305) {
    $bestIndex$0$lcssa$i43 = $bestIndex$1$i$lcssa;
    label = 25;
   } else {
    $308 = 0;$bestIndex$0$lcssa$i44 = $bestIndex$1$i$lcssa;
   }
  } else {
   HEAP32[$indexA>>2] = 0; //@line 176 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
   $bestIndex$0$lcssa$i43 = 0;
   label = 25;
  }
  if ((label|0) == 25) {
   $306 = HEAP32[$287>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $307 = ($306|0)>($bestIndex$0$lcssa$i43|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $308 = $307;$bestIndex$0$lcssa$i44 = $bestIndex$0$lcssa$i43;
  }
  __Z10loveAssertbPKc($308,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $309 = HEAP32[$285>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $310 = (($309) + ($bestIndex$0$lcssa$i44<<3)|0); //@line 178 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $311 = +HEAPF32[$310>>2]; //@line 178 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $312 = (((($309) + ($bestIndex$0$lcssa$i44<<3)|0)) + 4|0); //@line 178 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $313 = +HEAPF32[$312>>2]; //@line 178 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $314 = $23 * $311; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $315 = $22 * $313; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $316 = $314 - $315; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $317 = $34 + $316; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $318 = $22 * $311; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $319 = $23 * $313; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $320 = $318 + $319; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $321 = $35 + $320; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $322 = $317 - $271; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $323 = $321 - $275; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $324 = $260 * $322; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $325 = $263 * $323; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $326 = $324 + $325; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $$0 = $326;
  return (+$$0); //@line 191 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  break;
 }
 default: {
  __Z10loveAssertbPKc(0,622525); //@line 186 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAP32[$indexA>>2] = -1; //@line 187 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  HEAP32[$indexB>>2] = -1; //@line 188 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $$0 = 0.0;
  return (+$$0); //@line 191 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 }
 }
 return +(0.0);
}
function __ZNK20b2SeparationFunction8EvaluateEiif($this,$indexA,$indexB,$t) {
 $this = $this|0;
 $indexA = $indexA|0;
 $indexB = $indexB|0;
 $t = +$t;
 var $$0 = 0.0, $0 = 0.0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0.0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0;
 var $115 = 0.0, $116 = 0, $117 = 0.0, $118 = 0.0, $119 = 0, $12 = 0.0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0, $124 = 0.0, $125 = 0.0, $126 = 0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0.0, $130 = 0.0, $131 = 0.0, $132 = 0.0;
 var $133 = 0, $134 = 0.0, $135 = 0.0, $136 = 0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0.0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0.0, $156 = 0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0.0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0.0;
 var $17 = 0.0, $170 = 0.0, $171 = 0, $172 = 0.0, $173 = 0.0, $174 = 0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0, $182 = 0.0, $183 = 0.0, $184 = 0, $185 = 0.0, $186 = 0.0, $187 = 0.0;
 var $188 = 0.0, $189 = 0.0, $19 = 0.0, $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0.0, $20 = 0.0, $200 = 0, $201 = 0, $202 = 0.0, $203 = 0, $204 = 0.0;
 var $205 = 0.0, $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0.0, $21 = 0.0, $210 = 0.0, $211 = 0.0, $212 = 0.0, $213 = 0.0, $214 = 0.0, $215 = 0.0, $216 = 0.0, $217 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0, $27 = 0;
 var $28 = 0.0, $29 = 0.0, $3 = 0.0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0.0, $44 = 0.0, $45 = 0;
 var $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0.0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0, $6 = 0.0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0.0;
 var $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0.0, $80 = 0, $81 = 0;
 var $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0.0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0.0, $96 = 0, $97 = 0.0, $98 = 0.0, $99 = 0.0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = 1.0 - $t; //@line 694 "./libraries/Box2D/Common/b2Math.h"
 $1 = ((($this)) + 16|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $2 = +HEAPF32[$1>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $3 = $0 * $2; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 20|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $6 = $0 * $5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($this)) + 24|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $9 = $8 * $t; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($this)) + 28|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $11 = +HEAPF32[$10>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $12 = $11 * $t; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $13 = $3 + $9; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $14 = $6 + $12; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $15 = ((($this)) + 32|0); //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $16 = +HEAPF32[$15>>2]; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $17 = $0 * $16; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $18 = ((($this)) + 36|0); //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $19 = +HEAPF32[$18>>2]; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $20 = $19 * $t; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $21 = $17 + $20; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $22 = (+Math_sin((+$21))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $23 = (+Math_cos((+$21))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($this)) + 8|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $26 = $23 * $25; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($this)) + 12|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $28 = +HEAPF32[$27>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $29 = $22 * $28; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $30 = $26 - $29; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $31 = $22 * $25; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $32 = $23 * $28; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $33 = $31 + $32; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $34 = $13 - $30; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $35 = $14 - $33; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $36 = ((($this)) + 52|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $37 = +HEAPF32[$36>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $38 = $0 * $37; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $39 = ((($this)) + 56|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $40 = +HEAPF32[$39>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $41 = $0 * $40; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $42 = ((($this)) + 60|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $43 = +HEAPF32[$42>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $44 = $43 * $t; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $45 = ((($this)) + 64|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $46 = +HEAPF32[$45>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $47 = $46 * $t; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $48 = $38 + $44; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $49 = $41 + $47; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $50 = ((($this)) + 68|0); //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $51 = +HEAPF32[$50>>2]; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $52 = $0 * $51; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $53 = ((($this)) + 72|0); //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $54 = +HEAPF32[$53>>2]; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $55 = $54 * $t; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $56 = $52 + $55; //@line 695 "./libraries/Box2D/Common/b2Math.h"
 $57 = (+Math_sin((+$56))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $58 = (+Math_cos((+$56))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $59 = ((($this)) + 44|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $60 = +HEAPF32[$59>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $61 = $58 * $60; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $62 = ((($this)) + 48|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $63 = +HEAPF32[$62>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $64 = $57 * $63; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $65 = $61 - $64; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $66 = $57 * $60; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $67 = $58 * $63; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $68 = $66 + $67; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $69 = $48 - $65; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $70 = $49 - $68; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $71 = ((($this)) + 80|0); //@line 200 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 $72 = HEAP32[$71>>2]|0; //@line 200 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 switch ($72|0) {
 case 0:  {
  $73 = HEAP32[$this>>2]|0; //@line 204 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $74 = ($indexA|0)>(-1); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  if ($74) {
   $75 = ((($73)) + 20|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $76 = HEAP32[$75>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $77 = ($76|0)>($indexA|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $78 = $77;
  } else {
   $78 = 0;
  }
  __Z10loveAssertbPKc($78,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $79 = ((($73)) + 16|0); //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $80 = HEAP32[$79>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $81 = (($80) + ($indexA<<3)|0); //@line 204 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $82 = +HEAPF32[$81>>2]; //@line 204 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $83 = (((($80) + ($indexA<<3)|0)) + 4|0); //@line 204 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $84 = +HEAPF32[$83>>2]; //@line 204 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $85 = ((($this)) + 4|0); //@line 205 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $86 = HEAP32[$85>>2]|0; //@line 205 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $87 = ($indexB|0)>(-1); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  if ($87) {
   $88 = ((($86)) + 20|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $89 = HEAP32[$88>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $90 = ($89|0)>($indexB|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $91 = $90;
  } else {
   $91 = 0;
  }
  __Z10loveAssertbPKc($91,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $92 = ((($86)) + 16|0); //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $93 = HEAP32[$92>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $94 = (($93) + ($indexB<<3)|0); //@line 205 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $95 = +HEAPF32[$94>>2]; //@line 205 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $96 = (((($93) + ($indexB<<3)|0)) + 4|0); //@line 205 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $97 = +HEAPF32[$96>>2]; //@line 205 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $98 = $23 * $82; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $99 = $22 * $84; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $100 = $98 - $99; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $101 = $34 + $100; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $102 = $22 * $82; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $103 = $23 * $84; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $104 = $102 + $103; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $105 = $35 + $104; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $106 = $58 * $95; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $107 = $57 * $97; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $108 = $106 - $107; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $109 = $69 + $108; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $110 = $57 * $95; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $111 = $58 * $97; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $112 = $110 + $111; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $113 = $70 + $112; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $114 = $109 - $101; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $115 = $113 - $105; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $116 = ((($this)) + 92|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $117 = +HEAPF32[$116>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $118 = $117 * $114; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $119 = ((($this)) + 96|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $120 = +HEAPF32[$119>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $121 = $120 * $115; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $122 = $118 + $121; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $$0 = $122;
  return (+$$0); //@line 242 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  break;
 }
 case 1:  {
  $123 = ((($this)) + 92|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $124 = +HEAPF32[$123>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $125 = $23 * $124; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $126 = ((($this)) + 96|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $127 = +HEAPF32[$126>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $128 = $22 * $127; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $129 = $125 - $128; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $130 = $22 * $124; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $131 = $23 * $127; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $132 = $130 + $131; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $133 = ((($this)) + 84|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $134 = +HEAPF32[$133>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $135 = $23 * $134; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $136 = ((($this)) + 88|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $137 = +HEAPF32[$136>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $138 = $22 * $137; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $139 = $135 - $138; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $140 = $34 + $139; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $141 = $22 * $134; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $142 = $23 * $137; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $143 = $141 + $142; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $144 = $35 + $143; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $145 = ((($this)) + 4|0); //@line 219 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $146 = HEAP32[$145>>2]|0; //@line 219 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $147 = ($indexB|0)>(-1); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  if ($147) {
   $148 = ((($146)) + 20|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $149 = HEAP32[$148>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $150 = ($149|0)>($indexB|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $151 = $150;
  } else {
   $151 = 0;
  }
  __Z10loveAssertbPKc($151,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $152 = ((($146)) + 16|0); //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $153 = HEAP32[$152>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $154 = (($153) + ($indexB<<3)|0); //@line 219 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $155 = +HEAPF32[$154>>2]; //@line 219 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $156 = (((($153) + ($indexB<<3)|0)) + 4|0); //@line 219 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $157 = +HEAPF32[$156>>2]; //@line 219 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $158 = $58 * $155; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $159 = $57 * $157; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $160 = $158 - $159; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $161 = $69 + $160; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $162 = $57 * $155; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $163 = $58 * $157; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $164 = $162 + $163; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $165 = $70 + $164; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $166 = $161 - $140; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $167 = $165 - $144; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $168 = $129 * $166; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $169 = $132 * $167; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $170 = $168 + $169; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $$0 = $170;
  return (+$$0); //@line 242 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  break;
 }
 case 2:  {
  $171 = ((($this)) + 92|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $172 = +HEAPF32[$171>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $173 = $58 * $172; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $174 = ((($this)) + 96|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $175 = +HEAPF32[$174>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $176 = $57 * $175; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $177 = $173 - $176; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $178 = $57 * $172; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $179 = $58 * $175; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $180 = $178 + $179; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $181 = ((($this)) + 84|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $182 = +HEAPF32[$181>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $183 = $58 * $182; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $184 = ((($this)) + 88|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $185 = +HEAPF32[$184>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $186 = $57 * $185; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $187 = $183 - $186; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $188 = $69 + $187; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $189 = $57 * $182; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $190 = $58 * $185; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $191 = $189 + $190; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $192 = $70 + $191; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $193 = HEAP32[$this>>2]|0; //@line 231 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $194 = ($indexA|0)>(-1); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  if ($194) {
   $195 = ((($193)) + 20|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $196 = HEAP32[$195>>2]|0; //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $197 = ($196|0)>($indexA|0); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
   $198 = $197;
  } else {
   $198 = 0;
  }
  __Z10loveAssertbPKc($198,445986); //@line 103 "./libraries/Box2D/Collision/b2Distance.h"
  $199 = ((($193)) + 16|0); //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $200 = HEAP32[$199>>2]|0; //@line 104 "./libraries/Box2D/Collision/b2Distance.h"
  $201 = (($200) + ($indexA<<3)|0); //@line 231 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $202 = +HEAPF32[$201>>2]; //@line 231 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $203 = (((($200) + ($indexA<<3)|0)) + 4|0); //@line 231 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $204 = +HEAPF32[$203>>2]; //@line 231 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $205 = $23 * $202; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $206 = $22 * $204; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $207 = $205 - $206; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $208 = $34 + $207; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $209 = $22 * $202; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $210 = $23 * $204; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $211 = $209 + $210; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $212 = $35 + $211; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $213 = $208 - $188; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $214 = $212 - $192; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $215 = $177 * $213; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $216 = $180 * $214; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $217 = $215 + $216; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $$0 = $217;
  return (+$$0); //@line 242 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  break;
 }
 default: {
  __Z10loveAssertbPKc(0,622525); //@line 239 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
  $$0 = 0.0;
  return (+$$0); //@line 242 "libraries/Box2D/Collision/b2TimeOfImpact.cpp"
 }
 }
 return +(0.0);
}
function __ZN16b2StackAllocatorC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 102400|0); //@line 24 "libraries/Box2D/Common/b2StackAllocator.cpp"
 HEAP32[$0>>2] = 0; //@line 24 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $1 = ((($this)) + 102404|0); //@line 25 "libraries/Box2D/Common/b2StackAllocator.cpp"
 HEAP32[$1>>2] = 0; //@line 25 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $2 = ((($this)) + 102408|0); //@line 26 "libraries/Box2D/Common/b2StackAllocator.cpp"
 HEAP32[$2>>2] = 0; //@line 26 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $3 = ((($this)) + 102796|0); //@line 27 "libraries/Box2D/Common/b2StackAllocator.cpp"
 HEAP32[$3>>2] = 0; //@line 27 "libraries/Box2D/Common/b2StackAllocator.cpp"
 return; //@line 28 "libraries/Box2D/Common/b2StackAllocator.cpp"
}
function __ZN16b2StackAllocatorD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 102400|0); //@line 32 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 32 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $2 = ($1|0)==(0); //@line 32 "libraries/Box2D/Common/b2StackAllocator.cpp"
 __THREW__ = 0;
 invoke_vii(3822,($2|0),(446016|0)); //@line 32 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  $5 = ((($this)) + 102796|0); //@line 33 "libraries/Box2D/Common/b2StackAllocator.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 33 "libraries/Box2D/Common/b2StackAllocator.cpp"
  $7 = ($6|0)==(0); //@line 33 "libraries/Box2D/Common/b2StackAllocator.cpp"
  __THREW__ = 0;
  invoke_vii(3822,($7|0),(446029|0)); //@line 33 "libraries/Box2D/Common/b2StackAllocator.cpp"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if (!($9)) {
   return; //@line 34 "libraries/Box2D/Common/b2StackAllocator.cpp"
  }
 }
 $10 = ___cxa_find_matching_catch_3(0|0)|0;
 $11 = tempRet0;
 ___clang_call_terminate($10); //@line 32 "libraries/Box2D/Common/b2StackAllocator.cpp"
 // unreachable; //@line 32 "libraries/Box2D/Common/b2StackAllocator.cpp"
}
function __ZN16b2StackAllocator8AllocateEi($this,$size) {
 $this = $this|0;
 $size = $size|0;
 var $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 102796|0); //@line 38 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 38 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $2 = ($1|0)<(32); //@line 38 "libraries/Box2D/Common/b2StackAllocator.cpp"
 __Z10loveAssertbPKc($2,446047); //@line 38 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $3 = HEAP32[$0>>2]|0; //@line 40 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $4 = (((($this)) + 102412|0) + (($3*12)|0)|0); //@line 40 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $5 = (((((($this)) + 102412|0) + (($3*12)|0)|0)) + 4|0); //@line 41 "libraries/Box2D/Common/b2StackAllocator.cpp"
 HEAP32[$5>>2] = $size; //@line 41 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $6 = ((($this)) + 102400|0); //@line 42 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 42 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $8 = (($7) + ($size))|0; //@line 42 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $9 = ($8|0)>(102400); //@line 42 "libraries/Box2D/Common/b2StackAllocator.cpp"
 if ($9) {
  $10 = (__Z7b2Alloci($size)|0); //@line 44 "libraries/Box2D/Common/b2StackAllocator.cpp"
  HEAP32[$4>>2] = $10; //@line 44 "libraries/Box2D/Common/b2StackAllocator.cpp"
  $11 = (((((($this)) + 102412|0) + (($3*12)|0)|0)) + 8|0); //@line 45 "libraries/Box2D/Common/b2StackAllocator.cpp"
  HEAP8[$11>>0] = 1; //@line 45 "libraries/Box2D/Common/b2StackAllocator.cpp"
  $$pre$phiZ2D = $4;
 } else {
  $12 = (($this) + ($7)|0); //@line 49 "libraries/Box2D/Common/b2StackAllocator.cpp"
  HEAP32[$4>>2] = $12; //@line 49 "libraries/Box2D/Common/b2StackAllocator.cpp"
  $13 = (((((($this)) + 102412|0) + (($3*12)|0)|0)) + 8|0); //@line 50 "libraries/Box2D/Common/b2StackAllocator.cpp"
  HEAP8[$13>>0] = 0; //@line 50 "libraries/Box2D/Common/b2StackAllocator.cpp"
  $14 = HEAP32[$6>>2]|0; //@line 51 "libraries/Box2D/Common/b2StackAllocator.cpp"
  $15 = (($14) + ($size))|0; //@line 51 "libraries/Box2D/Common/b2StackAllocator.cpp"
  HEAP32[$6>>2] = $15; //@line 51 "libraries/Box2D/Common/b2StackAllocator.cpp"
  $$pre$phiZ2D = $4;
 }
 $16 = ((($this)) + 102404|0); //@line 54 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 54 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $18 = (($17) + ($size))|0; //@line 54 "libraries/Box2D/Common/b2StackAllocator.cpp"
 HEAP32[$16>>2] = $18; //@line 54 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $19 = ((($this)) + 102408|0); //@line 55 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 55 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $21 = ($20|0)>($18|0); //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $22 = $21 ? $20 : $18; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 HEAP32[$19>>2] = $22; //@line 55 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $23 = HEAP32[$0>>2]|0; //@line 56 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $24 = (($23) + 1)|0; //@line 56 "libraries/Box2D/Common/b2StackAllocator.cpp"
 HEAP32[$0>>2] = $24; //@line 56 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $25 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 58 "libraries/Box2D/Common/b2StackAllocator.cpp"
 return ($25|0); //@line 58 "libraries/Box2D/Common/b2StackAllocator.cpp"
}
function __ZN16b2StackAllocator4FreeEPv($this,$p) {
 $this = $this|0;
 $p = $p|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 102796|0); //@line 63 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 63 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $2 = ($1|0)>(0); //@line 63 "libraries/Box2D/Common/b2StackAllocator.cpp"
 __Z10loveAssertbPKc($2,446081); //@line 63 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $3 = HEAP32[$0>>2]|0; //@line 64 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $4 = (((($this)) + 102412|0) + (($3*12)|0)|0); //@line 64 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $5 = ((($4)) + -12|0); //@line 65 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 65 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $7 = ($6|0)==($p|0); //@line 65 "libraries/Box2D/Common/b2StackAllocator.cpp"
 __Z10loveAssertbPKc($7,446098); //@line 65 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $8 = ((($4)) + -4|0); //@line 66 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 66 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $10 = ($9<<24>>24)==(0); //@line 66 "libraries/Box2D/Common/b2StackAllocator.cpp"
 if ($10) {
  $11 = ((($4)) + -8|0); //@line 72 "libraries/Box2D/Common/b2StackAllocator.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 72 "libraries/Box2D/Common/b2StackAllocator.cpp"
  $13 = ((($this)) + 102400|0); //@line 72 "libraries/Box2D/Common/b2StackAllocator.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 72 "libraries/Box2D/Common/b2StackAllocator.cpp"
  $15 = (($14) - ($12))|0; //@line 72 "libraries/Box2D/Common/b2StackAllocator.cpp"
  HEAP32[$13>>2] = $15; //@line 72 "libraries/Box2D/Common/b2StackAllocator.cpp"
  $$pre$phiZ2D = $11;
 } else {
  __Z6b2FreePv($p); //@line 68 "libraries/Box2D/Common/b2StackAllocator.cpp"
  $$pre = ((($4)) + -8|0); //@line 74 "libraries/Box2D/Common/b2StackAllocator.cpp"
  $$pre$phiZ2D = $$pre;
 }
 $16 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 74 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $17 = ((($this)) + 102404|0); //@line 74 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 74 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $19 = (($18) - ($16))|0; //@line 74 "libraries/Box2D/Common/b2StackAllocator.cpp"
 HEAP32[$17>>2] = $19; //@line 74 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $20 = HEAP32[$0>>2]|0; //@line 75 "libraries/Box2D/Common/b2StackAllocator.cpp"
 $21 = (($20) + -1)|0; //@line 75 "libraries/Box2D/Common/b2StackAllocator.cpp"
 HEAP32[$0>>2] = $21; //@line 75 "libraries/Box2D/Common/b2StackAllocator.cpp"
 return; //@line 78 "libraries/Box2D/Common/b2StackAllocator.cpp"
}
function __ZN7b2TimerC2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 90 "libraries/Box2D/Common/b2Timer.cpp"
}
function __ZN7b2Timer5ResetEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 94 "libraries/Box2D/Common/b2Timer.cpp"
}
function __ZNK7b2Timer15GetMillisecondsEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return +0; //@line 98 "libraries/Box2D/Common/b2Timer.cpp"
}
function __ZN6b2BodyC2EPK9b2BodyDefP7b2World($this,$bd,$world) {
 $this = $this|0;
 $bd = $bd|0;
 $world = $world|0;
 var $$ = 0, $$2 = 0, $$3 = 0.0, $$4 = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0.0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 25 "libraries/Box2D/Dynamics/b2Body.cpp"
 $1 = ((($this)) + 64|0); //@line 25 "libraries/Box2D/Dynamics/b2Body.cpp"
 $2 = ((($bd)) + 4|0); //@line 27 "libraries/Box2D/Dynamics/b2Body.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 131 "./libraries/Box2D/Common/b2Math.h"
 $4 = $3 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $5 = ($4|0)==(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
 if ($5) {
  $10 = 0;
 } else {
  $6 = ((($bd)) + 8|0); //@line 131 "./libraries/Box2D/Common/b2Math.h"
  $7 = HEAP32[$6>>2]|0; //@line 131 "./libraries/Box2D/Common/b2Math.h"
  $8 = $7 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
  $9 = ($8|0)!=(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
  $10 = $9;
 }
 __Z10loveAssertbPKc($10,446115); //@line 27 "libraries/Box2D/Dynamics/b2Body.cpp"
 $11 = ((($bd)) + 16|0); //@line 28 "libraries/Box2D/Dynamics/b2Body.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 131 "./libraries/Box2D/Common/b2Math.h"
 $13 = $12 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $14 = ($13|0)==(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
 if ($14) {
  $19 = 0;
 } else {
  $15 = ((($bd)) + 20|0); //@line 131 "./libraries/Box2D/Common/b2Math.h"
  $16 = HEAP32[$15>>2]|0; //@line 131 "./libraries/Box2D/Common/b2Math.h"
  $17 = $16 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
  $18 = ($17|0)!=(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
  $19 = $18;
 }
 __Z10loveAssertbPKc($19,446138); //@line 28 "libraries/Box2D/Dynamics/b2Body.cpp"
 $20 = ((($bd)) + 12|0); //@line 29 "libraries/Box2D/Dynamics/b2Body.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 29 "libraries/Box2D/Dynamics/b2Body.cpp"
 $22 = $21 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $23 = ($22|0)!=(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
 __Z10loveAssertbPKc($23,446167); //@line 29 "libraries/Box2D/Dynamics/b2Body.cpp"
 $24 = ((($bd)) + 24|0); //@line 30 "libraries/Box2D/Dynamics/b2Body.cpp"
 $25 = HEAP32[$24>>2]|0; //@line 30 "libraries/Box2D/Dynamics/b2Body.cpp"
 $26 = $25 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $27 = ($26|0)!=(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
 __Z10loveAssertbPKc($27,446188); //@line 30 "libraries/Box2D/Dynamics/b2Body.cpp"
 $28 = ((($bd)) + 32|0); //@line 31 "libraries/Box2D/Dynamics/b2Body.cpp"
 $29 = +HEAPF32[$28>>2]; //@line 31 "libraries/Box2D/Dynamics/b2Body.cpp"
 $30 = (HEAPF32[tempDoublePtr>>2]=$29,HEAP32[tempDoublePtr>>2]|0); //@line 31 "libraries/Box2D/Dynamics/b2Body.cpp"
 $31 = $30 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $32 = ($31|0)!=(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $33 = $29 >= 0.0; //@line 31 "libraries/Box2D/Dynamics/b2Body.cpp"
 $$2 = $33 & $32; //@line 31 "libraries/Box2D/Dynamics/b2Body.cpp"
 __Z10loveAssertbPKc($$2,446219); //@line 31 "libraries/Box2D/Dynamics/b2Body.cpp"
 $34 = ((($bd)) + 28|0); //@line 32 "libraries/Box2D/Dynamics/b2Body.cpp"
 $35 = +HEAPF32[$34>>2]; //@line 32 "libraries/Box2D/Dynamics/b2Body.cpp"
 $36 = (HEAPF32[tempDoublePtr>>2]=$35,HEAP32[tempDoublePtr>>2]|0); //@line 32 "libraries/Box2D/Dynamics/b2Body.cpp"
 $37 = $36 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $38 = ($37|0)!=(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $39 = $35 >= 0.0; //@line 32 "libraries/Box2D/Dynamics/b2Body.cpp"
 $40 = $39 & $38; //@line 32 "libraries/Box2D/Dynamics/b2Body.cpp"
 __Z10loveAssertbPKc($40,446279); //@line 32 "libraries/Box2D/Dynamics/b2Body.cpp"
 $41 = ((($this)) + 4|0); //@line 34 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP16[$41>>1] = 0; //@line 34 "libraries/Box2D/Dynamics/b2Body.cpp"
 $42 = ((($bd)) + 39|0); //@line 36 "libraries/Box2D/Dynamics/b2Body.cpp"
 $43 = HEAP8[$42>>0]|0; //@line 36 "libraries/Box2D/Dynamics/b2Body.cpp"
 $44 = ($43<<24>>24)==(0); //@line 36 "libraries/Box2D/Dynamics/b2Body.cpp"
 $$ = $44 ? 0 : 8; //@line 36 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP16[$41>>1] = $$; //@line 38 "libraries/Box2D/Dynamics/b2Body.cpp"
 $45 = ((($bd)) + 38|0); //@line 40 "libraries/Box2D/Dynamics/b2Body.cpp"
 $46 = HEAP8[$45>>0]|0; //@line 40 "libraries/Box2D/Dynamics/b2Body.cpp"
 $47 = ($46<<24>>24)==(0); //@line 40 "libraries/Box2D/Dynamics/b2Body.cpp"
 if ($47) {
  $56 = $$;
 } else {
  $48 = $$&65535; //@line 42 "libraries/Box2D/Dynamics/b2Body.cpp"
  $49 = $48 | 16; //@line 42 "libraries/Box2D/Dynamics/b2Body.cpp"
  $50 = $49&65535; //@line 42 "libraries/Box2D/Dynamics/b2Body.cpp"
  HEAP16[$41>>1] = $50; //@line 42 "libraries/Box2D/Dynamics/b2Body.cpp"
  $56 = $50;
 }
 $51 = ((($bd)) + 36|0); //@line 44 "libraries/Box2D/Dynamics/b2Body.cpp"
 $52 = HEAP16[$51>>1]|0; //@line 44 "libraries/Box2D/Dynamics/b2Body.cpp"
 $53 = $52&255; //@line 44 "libraries/Box2D/Dynamics/b2Body.cpp"
 $54 = ($53<<24>>24)==(0); //@line 44 "libraries/Box2D/Dynamics/b2Body.cpp"
 if ($54) {
  $61 = $56;
 } else {
  $55 = $56&65535; //@line 46 "libraries/Box2D/Dynamics/b2Body.cpp"
  $57 = $55 | 4; //@line 46 "libraries/Box2D/Dynamics/b2Body.cpp"
  $58 = $57&65535; //@line 46 "libraries/Box2D/Dynamics/b2Body.cpp"
  HEAP16[$41>>1] = $58; //@line 46 "libraries/Box2D/Dynamics/b2Body.cpp"
  $61 = $58;
 }
 $59 = ($52&65535)<(256); //@line 48 "libraries/Box2D/Dynamics/b2Body.cpp"
 if ($59) {
  $68 = $61;
 } else {
  $60 = $61&65535; //@line 50 "libraries/Box2D/Dynamics/b2Body.cpp"
  $62 = $60 | 2; //@line 50 "libraries/Box2D/Dynamics/b2Body.cpp"
  $63 = $62&65535; //@line 50 "libraries/Box2D/Dynamics/b2Body.cpp"
  HEAP16[$41>>1] = $63; //@line 50 "libraries/Box2D/Dynamics/b2Body.cpp"
  $68 = $63;
 }
 $64 = ((($bd)) + 40|0); //@line 52 "libraries/Box2D/Dynamics/b2Body.cpp"
 $65 = HEAP8[$64>>0]|0; //@line 52 "libraries/Box2D/Dynamics/b2Body.cpp"
 $66 = ($65<<24>>24)==(0); //@line 52 "libraries/Box2D/Dynamics/b2Body.cpp"
 if (!($66)) {
  $67 = $68&65535; //@line 54 "libraries/Box2D/Dynamics/b2Body.cpp"
  $69 = $67 | 32; //@line 54 "libraries/Box2D/Dynamics/b2Body.cpp"
  $70 = $69&65535; //@line 54 "libraries/Box2D/Dynamics/b2Body.cpp"
  HEAP16[$41>>1] = $70; //@line 54 "libraries/Box2D/Dynamics/b2Body.cpp"
 }
 $71 = ((($this)) + 88|0); //@line 57 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$71>>2] = $world; //@line 57 "libraries/Box2D/Dynamics/b2Body.cpp"
 $72 = $2; //@line 59 "libraries/Box2D/Dynamics/b2Body.cpp"
 $73 = $72; //@line 59 "libraries/Box2D/Dynamics/b2Body.cpp"
 $74 = HEAP32[$73>>2]|0; //@line 59 "libraries/Box2D/Dynamics/b2Body.cpp"
 $75 = (($72) + 4)|0; //@line 59 "libraries/Box2D/Dynamics/b2Body.cpp"
 $76 = $75; //@line 59 "libraries/Box2D/Dynamics/b2Body.cpp"
 $77 = HEAP32[$76>>2]|0; //@line 59 "libraries/Box2D/Dynamics/b2Body.cpp"
 $78 = $0; //@line 59 "libraries/Box2D/Dynamics/b2Body.cpp"
 $79 = $78; //@line 59 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$79>>2] = $74; //@line 59 "libraries/Box2D/Dynamics/b2Body.cpp"
 $80 = (($78) + 4)|0; //@line 59 "libraries/Box2D/Dynamics/b2Body.cpp"
 $81 = $80; //@line 59 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$81>>2] = $77; //@line 59 "libraries/Box2D/Dynamics/b2Body.cpp"
 $82 = +HEAPF32[$20>>2]; //@line 60 "libraries/Box2D/Dynamics/b2Body.cpp"
 $83 = (+Math_sin((+$82))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $84 = ((($this)) + 20|0); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$84>>2] = $83; //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $85 = (+Math_cos((+$82))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $86 = ((($this)) + 24|0); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$86>>2] = $85; //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $87 = ((($this)) + 28|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$87>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $88 = ((($this)) + 32|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$88>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $89 = ((($this)) + 36|0); //@line 63 "libraries/Box2D/Dynamics/b2Body.cpp"
 $90 = $89; //@line 63 "libraries/Box2D/Dynamics/b2Body.cpp"
 $91 = $90; //@line 63 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$91>>2] = $74; //@line 63 "libraries/Box2D/Dynamics/b2Body.cpp"
 $92 = (($90) + 4)|0; //@line 63 "libraries/Box2D/Dynamics/b2Body.cpp"
 $93 = $92; //@line 63 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$93>>2] = $77; //@line 63 "libraries/Box2D/Dynamics/b2Body.cpp"
 $94 = ((($this)) + 44|0); //@line 64 "libraries/Box2D/Dynamics/b2Body.cpp"
 $95 = $94; //@line 64 "libraries/Box2D/Dynamics/b2Body.cpp"
 $96 = $95; //@line 64 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$96>>2] = $74; //@line 64 "libraries/Box2D/Dynamics/b2Body.cpp"
 $97 = (($95) + 4)|0; //@line 64 "libraries/Box2D/Dynamics/b2Body.cpp"
 $98 = $97; //@line 64 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$98>>2] = $77; //@line 64 "libraries/Box2D/Dynamics/b2Body.cpp"
 $99 = HEAP32[$20>>2]|0; //@line 65 "libraries/Box2D/Dynamics/b2Body.cpp"
 $100 = ((($this)) + 52|0); //@line 65 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$100>>2] = $99; //@line 65 "libraries/Box2D/Dynamics/b2Body.cpp"
 $101 = ((($this)) + 56|0); //@line 66 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$101>>2] = $99; //@line 66 "libraries/Box2D/Dynamics/b2Body.cpp"
 $102 = ((($this)) + 60|0); //@line 67 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$102>>2] = 0.0; //@line 67 "libraries/Box2D/Dynamics/b2Body.cpp"
 $103 = ((($this)) + 108|0); //@line 69 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$103>>2] = 0; //@line 69 "libraries/Box2D/Dynamics/b2Body.cpp"
 $104 = ((($this)) + 112|0); //@line 70 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$104>>2] = 0; //@line 70 "libraries/Box2D/Dynamics/b2Body.cpp"
 $105 = ((($this)) + 92|0); //@line 71 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$105>>2] = 0; //@line 71 "libraries/Box2D/Dynamics/b2Body.cpp"
 $106 = ((($this)) + 96|0); //@line 72 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$106>>2] = 0; //@line 72 "libraries/Box2D/Dynamics/b2Body.cpp"
 $107 = $11; //@line 74 "libraries/Box2D/Dynamics/b2Body.cpp"
 $108 = $107; //@line 74 "libraries/Box2D/Dynamics/b2Body.cpp"
 $109 = HEAP32[$108>>2]|0; //@line 74 "libraries/Box2D/Dynamics/b2Body.cpp"
 $110 = (($107) + 4)|0; //@line 74 "libraries/Box2D/Dynamics/b2Body.cpp"
 $111 = $110; //@line 74 "libraries/Box2D/Dynamics/b2Body.cpp"
 $112 = HEAP32[$111>>2]|0; //@line 74 "libraries/Box2D/Dynamics/b2Body.cpp"
 $113 = $1; //@line 74 "libraries/Box2D/Dynamics/b2Body.cpp"
 $114 = $113; //@line 74 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$114>>2] = $109; //@line 74 "libraries/Box2D/Dynamics/b2Body.cpp"
 $115 = (($113) + 4)|0; //@line 74 "libraries/Box2D/Dynamics/b2Body.cpp"
 $116 = $115; //@line 74 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$116>>2] = $112; //@line 74 "libraries/Box2D/Dynamics/b2Body.cpp"
 $117 = HEAP32[$24>>2]|0; //@line 75 "libraries/Box2D/Dynamics/b2Body.cpp"
 $118 = ((($this)) + 72|0); //@line 75 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$118>>2] = $117; //@line 75 "libraries/Box2D/Dynamics/b2Body.cpp"
 $119 = HEAP32[$34>>2]|0; //@line 77 "libraries/Box2D/Dynamics/b2Body.cpp"
 $120 = ((($this)) + 132|0); //@line 77 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$120>>2] = $119; //@line 77 "libraries/Box2D/Dynamics/b2Body.cpp"
 $121 = HEAP32[$28>>2]|0; //@line 78 "libraries/Box2D/Dynamics/b2Body.cpp"
 $122 = ((($this)) + 136|0); //@line 78 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$122>>2] = $121; //@line 78 "libraries/Box2D/Dynamics/b2Body.cpp"
 $123 = ((($bd)) + 48|0); //@line 79 "libraries/Box2D/Dynamics/b2Body.cpp"
 $124 = HEAP32[$123>>2]|0; //@line 79 "libraries/Box2D/Dynamics/b2Body.cpp"
 $125 = ((($this)) + 140|0); //@line 79 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$125>>2] = $124; //@line 79 "libraries/Box2D/Dynamics/b2Body.cpp"
 $126 = ((($this)) + 76|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$126>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $127 = ((($this)) + 80|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$127>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $128 = ((($this)) + 84|0); //@line 82 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$128>>2] = 0.0; //@line 82 "libraries/Box2D/Dynamics/b2Body.cpp"
 $129 = ((($this)) + 144|0); //@line 84 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$129>>2] = 0.0; //@line 84 "libraries/Box2D/Dynamics/b2Body.cpp"
 $130 = HEAP32[$bd>>2]|0; //@line 86 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$this>>2] = $130; //@line 86 "libraries/Box2D/Dynamics/b2Body.cpp"
 $131 = ($130|0)==(2); //@line 88 "libraries/Box2D/Dynamics/b2Body.cpp"
 $132 = ((($this)) + 116|0); //@line 90 "libraries/Box2D/Dynamics/b2Body.cpp"
 $$3 = $131 ? 1.0 : 0.0; //@line 88 "libraries/Box2D/Dynamics/b2Body.cpp"
 $$4 = $131 ? 1.0 : 0.0; //@line 88 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$132>>2] = $$3; //@line 90 "libraries/Box2D/Dynamics/b2Body.cpp"
 $133 = ((($this)) + 120|0); //@line 96 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$133>>2] = $$4; //@line 96 "libraries/Box2D/Dynamics/b2Body.cpp"
 $134 = ((($this)) + 124|0); //@line 99 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$134>>2] = 0.0; //@line 99 "libraries/Box2D/Dynamics/b2Body.cpp"
 $135 = ((($this)) + 128|0); //@line 100 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$135>>2] = 0.0; //@line 100 "libraries/Box2D/Dynamics/b2Body.cpp"
 $136 = ((($bd)) + 44|0); //@line 102 "libraries/Box2D/Dynamics/b2Body.cpp"
 $137 = HEAP32[$136>>2]|0; //@line 102 "libraries/Box2D/Dynamics/b2Body.cpp"
 $138 = ((($this)) + 148|0); //@line 102 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$138>>2] = $137; //@line 102 "libraries/Box2D/Dynamics/b2Body.cpp"
 $139 = ((($this)) + 100|0); //@line 104 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$139>>2] = 0; //@line 104 "libraries/Box2D/Dynamics/b2Body.cpp"
 $140 = ((($this)) + 104|0); //@line 105 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$140>>2] = 0; //@line 105 "libraries/Box2D/Dynamics/b2Body.cpp"
 return; //@line 106 "libraries/Box2D/Dynamics/b2Body.cpp"
}
function __ZN6b2BodyD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 111 "libraries/Box2D/Dynamics/b2Body.cpp"
}
function __ZN6b2Body7SetTypeE10b2BodyType($this,$type) {
 $this = $this|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0.0;
 var $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $bitcast = 0.0, $ce$07 = 0, $exitcond = 0, $f$0 = 0, $f$0$i = 0, $f$02 = 0;
 var $f$03 = 0, $f$03$i = 0, $f$04$i = 0, $i$01 = 0, $xf1$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $xf1$i = sp;
 $0 = ((($this)) + 88|0); //@line 115 "libraries/Box2D/Dynamics/b2Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 115 "libraries/Box2D/Dynamics/b2Body.cpp"
 $2 = ((($1)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $3 = HEAP32[$2>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $4 = $3 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $5 = ($4|0)==(0); //@line 115 "libraries/Box2D/Dynamics/b2Body.cpp"
 __Z10loveAssertbPKc($5,446337); //@line 115 "libraries/Box2D/Dynamics/b2Body.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 116 "libraries/Box2D/Dynamics/b2Body.cpp"
 $7 = ((($6)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $8 = HEAP32[$7>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $9 = $8 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $10 = ($9|0)==(0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 if (!($10)) {
  STACKTOP = sp;return; //@line 164 "libraries/Box2D/Dynamics/b2Body.cpp"
 }
 $11 = HEAP32[$this>>2]|0; //@line 121 "libraries/Box2D/Dynamics/b2Body.cpp"
 $12 = ($11|0)==($type|0); //@line 121 "libraries/Box2D/Dynamics/b2Body.cpp"
 if ($12) {
  STACKTOP = sp;return; //@line 164 "libraries/Box2D/Dynamics/b2Body.cpp"
 }
 HEAP32[$this>>2] = $type; //@line 126 "libraries/Box2D/Dynamics/b2Body.cpp"
 __ZN6b2Body13ResetMassDataEv($this); //@line 128 "libraries/Box2D/Dynamics/b2Body.cpp"
 $13 = HEAP32[$this>>2]|0; //@line 130 "libraries/Box2D/Dynamics/b2Body.cpp"
 $14 = ($13|0)==(0); //@line 130 "libraries/Box2D/Dynamics/b2Body.cpp"
 if ($14) {
  $15 = ((($this)) + 64|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$15>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $16 = ((($this)) + 68|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$16>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $17 = ((($this)) + 72|0); //@line 133 "libraries/Box2D/Dynamics/b2Body.cpp"
  HEAPF32[$17>>2] = 0.0; //@line 133 "libraries/Box2D/Dynamics/b2Body.cpp"
  $18 = ((($this)) + 56|0); //@line 134 "libraries/Box2D/Dynamics/b2Body.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 134 "libraries/Box2D/Dynamics/b2Body.cpp"
  $20 = ((($this)) + 52|0); //@line 134 "libraries/Box2D/Dynamics/b2Body.cpp"
  HEAP32[$20>>2] = $19; //@line 134 "libraries/Box2D/Dynamics/b2Body.cpp"
  $21 = ((($this)) + 36|0); //@line 135 "libraries/Box2D/Dynamics/b2Body.cpp"
  $22 = ((($this)) + 44|0); //@line 135 "libraries/Box2D/Dynamics/b2Body.cpp"
  $23 = $22; //@line 135 "libraries/Box2D/Dynamics/b2Body.cpp"
  $24 = $23; //@line 135 "libraries/Box2D/Dynamics/b2Body.cpp"
  $25 = HEAP32[$24>>2]|0; //@line 135 "libraries/Box2D/Dynamics/b2Body.cpp"
  $26 = (($23) + 4)|0; //@line 135 "libraries/Box2D/Dynamics/b2Body.cpp"
  $27 = $26; //@line 135 "libraries/Box2D/Dynamics/b2Body.cpp"
  $28 = HEAP32[$27>>2]|0; //@line 135 "libraries/Box2D/Dynamics/b2Body.cpp"
  $29 = $21; //@line 135 "libraries/Box2D/Dynamics/b2Body.cpp"
  $30 = $29; //@line 135 "libraries/Box2D/Dynamics/b2Body.cpp"
  HEAP32[$30>>2] = $25; //@line 135 "libraries/Box2D/Dynamics/b2Body.cpp"
  $31 = (($29) + 4)|0; //@line 135 "libraries/Box2D/Dynamics/b2Body.cpp"
  $32 = $31; //@line 135 "libraries/Box2D/Dynamics/b2Body.cpp"
  HEAP32[$32>>2] = $28; //@line 135 "libraries/Box2D/Dynamics/b2Body.cpp"
  $33 = (HEAP32[tempDoublePtr>>2]=$19,+HEAPF32[tempDoublePtr>>2]); //@line 444 "libraries/Box2D/Dynamics/b2Body.cpp"
  $34 = (+Math_sin((+$33))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
  $35 = ((($xf1$i)) + 8|0); //@line 315 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$35>>2] = $34; //@line 315 "./libraries/Box2D/Common/b2Math.h"
  $36 = (+Math_cos((+$33))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
  $37 = ((($xf1$i)) + 12|0); //@line 316 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$37>>2] = $36; //@line 316 "./libraries/Box2D/Common/b2Math.h"
  $38 = ((($this)) + 28|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $39 = +HEAPF32[$38>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $40 = $36 * $39; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $41 = ((($this)) + 32|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $42 = +HEAPF32[$41>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $43 = $34 * $42; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $44 = $40 - $43; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $45 = $34 * $39; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $46 = $36 * $42; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $47 = $45 + $46; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $bitcast = (HEAP32[tempDoublePtr>>2]=$25,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $48 = $bitcast - $44; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $49 = (HEAP32[tempDoublePtr>>2]=$28,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $50 = $49 - $47; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$xf1$i>>2] = $48; //@line 445 "libraries/Box2D/Dynamics/b2Body.cpp"
  $51 = ((($xf1$i)) + 4|0); //@line 445 "libraries/Box2D/Dynamics/b2Body.cpp"
  HEAPF32[$51>>2] = $50; //@line 445 "libraries/Box2D/Dynamics/b2Body.cpp"
  $52 = HEAP32[$0>>2]|0; //@line 447 "libraries/Box2D/Dynamics/b2Body.cpp"
  $53 = ((($52)) + 102872|0); //@line 447 "libraries/Box2D/Dynamics/b2Body.cpp"
  $54 = ((($this)) + 100|0); //@line 448 "libraries/Box2D/Dynamics/b2Body.cpp"
  $f$03$i = HEAP32[$54>>2]|0; //@line 448 "libraries/Box2D/Dynamics/b2Body.cpp"
  $55 = ($f$03$i|0)==(0|0); //@line 448 "libraries/Box2D/Dynamics/b2Body.cpp"
  if (!($55)) {
   $56 = ((($this)) + 12|0); //@line 450 "libraries/Box2D/Dynamics/b2Body.cpp"
   $f$04$i = $f$03$i;
   while(1) {
    __ZN9b2Fixture11SynchronizeEP12b2BroadPhaseRK11b2TransformS4_($f$04$i,$53,$xf1$i,$56); //@line 450 "libraries/Box2D/Dynamics/b2Body.cpp"
    $57 = ((($f$04$i)) + 4|0); //@line 448 "libraries/Box2D/Dynamics/b2Body.cpp"
    $f$0$i = HEAP32[$57>>2]|0; //@line 448 "libraries/Box2D/Dynamics/b2Body.cpp"
    $58 = ($f$0$i|0)==(0|0); //@line 448 "libraries/Box2D/Dynamics/b2Body.cpp"
    if ($58) {
     break;
    } else {
     $f$04$i = $f$0$i;
    }
   }
  }
 }
 $59 = ((($this)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $60 = HEAP16[$59>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $61 = $60&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $62 = $61 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $63 = ($62|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($63) {
  $64 = $61 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $65 = $64&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$59>>1] = $65; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $66 = ((($this)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$66>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $67 = ((($this)) + 76|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$67>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $68 = ((($this)) + 80|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$68>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $69 = ((($this)) + 84|0); //@line 142 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$69>>2] = 0.0; //@line 142 "libraries/Box2D/Dynamics/b2Body.cpp"
 $70 = ((($this)) + 112|0); //@line 145 "libraries/Box2D/Dynamics/b2Body.cpp"
 $71 = HEAP32[$70>>2]|0; //@line 145 "libraries/Box2D/Dynamics/b2Body.cpp"
 $72 = ($71|0)==(0|0); //@line 146 "libraries/Box2D/Dynamics/b2Body.cpp"
 if (!($72)) {
  $ce$07 = $71;
  while(1) {
   $73 = ((($ce$07)) + 12|0); //@line 149 "libraries/Box2D/Dynamics/b2Body.cpp"
   $74 = HEAP32[$73>>2]|0; //@line 149 "libraries/Box2D/Dynamics/b2Body.cpp"
   $75 = HEAP32[$0>>2]|0; //@line 150 "libraries/Box2D/Dynamics/b2Body.cpp"
   $76 = ((($75)) + 102872|0); //@line 150 "libraries/Box2D/Dynamics/b2Body.cpp"
   $77 = ((($ce$07)) + 4|0); //@line 150 "libraries/Box2D/Dynamics/b2Body.cpp"
   $78 = HEAP32[$77>>2]|0; //@line 150 "libraries/Box2D/Dynamics/b2Body.cpp"
   __ZN16b2ContactManager7DestroyEP9b2Contact($76,$78); //@line 150 "libraries/Box2D/Dynamics/b2Body.cpp"
   $79 = ($74|0)==(0|0); //@line 146 "libraries/Box2D/Dynamics/b2Body.cpp"
   if ($79) {
    break;
   } else {
    $ce$07 = $74;
   }
  }
 }
 HEAP32[$70>>2] = 0; //@line 152 "libraries/Box2D/Dynamics/b2Body.cpp"
 $80 = HEAP32[$0>>2]|0; //@line 155 "libraries/Box2D/Dynamics/b2Body.cpp"
 $81 = ((($80)) + 102872|0); //@line 155 "libraries/Box2D/Dynamics/b2Body.cpp"
 $82 = ((($this)) + 100|0); //@line 156 "libraries/Box2D/Dynamics/b2Body.cpp"
 $f$02 = HEAP32[$82>>2]|0; //@line 156 "libraries/Box2D/Dynamics/b2Body.cpp"
 $83 = ($f$02|0)==(0|0); //@line 156 "libraries/Box2D/Dynamics/b2Body.cpp"
 if ($83) {
  STACKTOP = sp;return; //@line 164 "libraries/Box2D/Dynamics/b2Body.cpp"
 } else {
  $f$03 = $f$02;
 }
 while(1) {
  $84 = ((($f$03)) + 28|0); //@line 158 "libraries/Box2D/Dynamics/b2Body.cpp"
  $85 = HEAP32[$84>>2]|0; //@line 158 "libraries/Box2D/Dynamics/b2Body.cpp"
  $86 = ($85|0)>(0); //@line 159 "libraries/Box2D/Dynamics/b2Body.cpp"
  if ($86) {
   $87 = ((($f$03)) + 24|0); //@line 161 "libraries/Box2D/Dynamics/b2Body.cpp"
   $i$01 = 0;
   while(1) {
    $90 = HEAP32[$87>>2]|0; //@line 161 "libraries/Box2D/Dynamics/b2Body.cpp"
    $91 = (((($90) + (($i$01*28)|0)|0)) + 24|0); //@line 161 "libraries/Box2D/Dynamics/b2Body.cpp"
    $92 = HEAP32[$91>>2]|0; //@line 161 "libraries/Box2D/Dynamics/b2Body.cpp"
    __ZN12b2BroadPhase10TouchProxyEi($81,$92); //@line 161 "libraries/Box2D/Dynamics/b2Body.cpp"
    $93 = (($i$01) + 1)|0; //@line 159 "libraries/Box2D/Dynamics/b2Body.cpp"
    $exitcond = ($93|0)==($85|0); //@line 159 "libraries/Box2D/Dynamics/b2Body.cpp"
    if ($exitcond) {
     break;
    } else {
     $i$01 = $93;
    }
   }
  }
  $88 = ((($f$03)) + 4|0); //@line 156 "libraries/Box2D/Dynamics/b2Body.cpp"
  $f$0 = HEAP32[$88>>2]|0; //@line 156 "libraries/Box2D/Dynamics/b2Body.cpp"
  $89 = ($f$0|0)==(0|0); //@line 156 "libraries/Box2D/Dynamics/b2Body.cpp"
  if ($89) {
   break;
  } else {
   $f$03 = $f$0;
  }
 }
 STACKTOP = sp;return; //@line 164 "libraries/Box2D/Dynamics/b2Body.cpp"
}
function __ZN6b2Body13ResetMassDataEv($this) {
 $this = $this|0;
 var $$cast$pre$phiZZ2D = 0.0, $$pre = 0.0, $$pre$phiZ2D = 0.0, $$pre14 = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0, $104 = 0.0, $105 = 0.0, $106 = 0, $107 = 0, $108 = 0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0;
 var $112 = 0.0, $113 = 0.0, $114 = 0, $115 = 0.0, $116 = 0.0, $117 = 0, $118 = 0.0, $119 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0.0, $59 = 0.0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0, $66 = 0.0, $67 = 0.0, $68 = 0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0.0, $76 = 0.0, $77 = 0.0;
 var $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0.0, $86 = 0, $87 = 0.0, $88 = 0, $89 = 0.0, $9 = 0, $90 = 0, $91 = 0.0, $92 = 0.0, $93 = 0, $94 = 0.0, $95 = 0.0;
 var $96 = 0.0, $97 = 0, $98 = 0.0, $99 = 0.0, $f$0 = 0, $f$012 = 0, $f$09 = 0, $localCenter$sroa$0$0$lcssa = 0, $localCenter$sroa$0$010 = 0, $localCenter$sroa$0$1 = 0, $localCenter$sroa$0$2 = 0, $localCenter$sroa$10$0$lcssa = 0, $localCenter$sroa$10$011 = 0, $localCenter$sroa$10$1 = 0, $localCenter$sroa$10$2 = 0, $massData = 0, $storemerge = 0.0, $switch = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $massData = sp;
 $0 = ((($this)) + 116|0); //@line 284 "libraries/Box2D/Dynamics/b2Body.cpp"
 $1 = ((($this)) + 120|0); //@line 285 "libraries/Box2D/Dynamics/b2Body.cpp"
 $2 = ((($this)) + 124|0); //@line 286 "libraries/Box2D/Dynamics/b2Body.cpp"
 $3 = ((($this)) + 128|0); //@line 287 "libraries/Box2D/Dynamics/b2Body.cpp"
 $4 = ((($this)) + 28|0); //@line 288 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$4>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($this)) + 32|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$5>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;HEAP32[$0+12>>2]=0|0; //@line 285 "libraries/Box2D/Dynamics/b2Body.cpp"
 $6 = HEAP32[$this>>2]|0; //@line 291 "libraries/Box2D/Dynamics/b2Body.cpp"
 $switch = ($6>>>0)<(2); //@line 291 "libraries/Box2D/Dynamics/b2Body.cpp"
 if ($switch) {
  $7 = ((($this)) + 36|0); //@line 293 "libraries/Box2D/Dynamics/b2Body.cpp"
  $8 = ((($this)) + 12|0); //@line 293 "libraries/Box2D/Dynamics/b2Body.cpp"
  $9 = $8; //@line 293 "libraries/Box2D/Dynamics/b2Body.cpp"
  $10 = $9; //@line 293 "libraries/Box2D/Dynamics/b2Body.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 293 "libraries/Box2D/Dynamics/b2Body.cpp"
  $12 = (($9) + 4)|0; //@line 293 "libraries/Box2D/Dynamics/b2Body.cpp"
  $13 = $12; //@line 293 "libraries/Box2D/Dynamics/b2Body.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 293 "libraries/Box2D/Dynamics/b2Body.cpp"
  $15 = $7; //@line 293 "libraries/Box2D/Dynamics/b2Body.cpp"
  $16 = $15; //@line 293 "libraries/Box2D/Dynamics/b2Body.cpp"
  HEAP32[$16>>2] = $11; //@line 293 "libraries/Box2D/Dynamics/b2Body.cpp"
  $17 = (($15) + 4)|0; //@line 293 "libraries/Box2D/Dynamics/b2Body.cpp"
  $18 = $17; //@line 293 "libraries/Box2D/Dynamics/b2Body.cpp"
  HEAP32[$18>>2] = $14; //@line 293 "libraries/Box2D/Dynamics/b2Body.cpp"
  $19 = ((($this)) + 44|0); //@line 294 "libraries/Box2D/Dynamics/b2Body.cpp"
  $20 = $19; //@line 294 "libraries/Box2D/Dynamics/b2Body.cpp"
  $21 = $20; //@line 294 "libraries/Box2D/Dynamics/b2Body.cpp"
  HEAP32[$21>>2] = $11; //@line 294 "libraries/Box2D/Dynamics/b2Body.cpp"
  $22 = (($20) + 4)|0; //@line 294 "libraries/Box2D/Dynamics/b2Body.cpp"
  $23 = $22; //@line 294 "libraries/Box2D/Dynamics/b2Body.cpp"
  HEAP32[$23>>2] = $14; //@line 294 "libraries/Box2D/Dynamics/b2Body.cpp"
  $24 = ((($this)) + 56|0); //@line 295 "libraries/Box2D/Dynamics/b2Body.cpp"
  $25 = HEAP32[$24>>2]|0; //@line 295 "libraries/Box2D/Dynamics/b2Body.cpp"
  $26 = ((($this)) + 52|0); //@line 295 "libraries/Box2D/Dynamics/b2Body.cpp"
  HEAP32[$26>>2] = $25; //@line 295 "libraries/Box2D/Dynamics/b2Body.cpp"
  STACKTOP = sp;return; //@line 351 "libraries/Box2D/Dynamics/b2Body.cpp"
 }
 $27 = ($6|0)==(2); //@line 299 "libraries/Box2D/Dynamics/b2Body.cpp"
 __Z10loveAssertbPKc($27,446366); //@line 299 "libraries/Box2D/Dynamics/b2Body.cpp"
 $28 = HEAP32[158039]|0; //@line 302 "libraries/Box2D/Dynamics/b2Body.cpp"
 $29 = HEAP32[(632160)>>2]|0; //@line 302 "libraries/Box2D/Dynamics/b2Body.cpp"
 $30 = ((($this)) + 100|0); //@line 303 "libraries/Box2D/Dynamics/b2Body.cpp"
 $f$09 = HEAP32[$30>>2]|0; //@line 303 "libraries/Box2D/Dynamics/b2Body.cpp"
 $31 = ($f$09|0)==(0|0); //@line 303 "libraries/Box2D/Dynamics/b2Body.cpp"
 if ($31) {
  $localCenter$sroa$0$0$lcssa = $28;$localCenter$sroa$10$0$lcssa = $29;
 } else {
  $32 = ((($massData)) + 4|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $33 = ((($massData)) + 8|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $34 = ((($massData)) + 12|0); //@line 314 "libraries/Box2D/Dynamics/b2Body.cpp"
  $f$012 = $f$09;$localCenter$sroa$0$010 = $28;$localCenter$sroa$10$011 = $29;
  while(1) {
   $37 = +HEAPF32[$f$012>>2]; //@line 336 "./libraries/Box2D/Dynamics/b2Fixture.h"
   $38 = $37 == 0.0; //@line 305 "libraries/Box2D/Dynamics/b2Body.cpp"
   if ($38) {
    $localCenter$sroa$0$1 = $localCenter$sroa$0$010;$localCenter$sroa$10$1 = $localCenter$sroa$10$011;
   } else {
    $39 = ((($f$012)) + 12|0); //@line 336 "./libraries/Box2D/Dynamics/b2Fixture.h"
    $40 = HEAP32[$39>>2]|0; //@line 336 "./libraries/Box2D/Dynamics/b2Fixture.h"
    $41 = HEAP32[$40>>2]|0; //@line 336 "./libraries/Box2D/Dynamics/b2Fixture.h"
    $42 = ((($41)) + 28|0); //@line 336 "./libraries/Box2D/Dynamics/b2Fixture.h"
    $43 = HEAP32[$42>>2]|0; //@line 336 "./libraries/Box2D/Dynamics/b2Fixture.h"
    FUNCTION_TABLE_viid[$43 & 8191]($40,$massData,$37); //@line 336 "./libraries/Box2D/Dynamics/b2Fixture.h"
    $44 = +HEAPF32[$massData>>2]; //@line 312 "libraries/Box2D/Dynamics/b2Body.cpp"
    $45 = +HEAPF32[$0>>2]; //@line 312 "libraries/Box2D/Dynamics/b2Body.cpp"
    $46 = $44 + $45; //@line 312 "libraries/Box2D/Dynamics/b2Body.cpp"
    HEAPF32[$0>>2] = $46; //@line 312 "libraries/Box2D/Dynamics/b2Body.cpp"
    $47 = +HEAPF32[$32>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $48 = $44 * $47; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $49 = +HEAPF32[$33>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $50 = $44 * $49; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $51 = (HEAP32[tempDoublePtr>>2]=$localCenter$sroa$0$010,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $52 = $51 + $48; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $53 = (HEAPF32[tempDoublePtr>>2]=$52,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $54 = (HEAP32[tempDoublePtr>>2]=$localCenter$sroa$10$011,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $55 = $54 + $50; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $56 = (HEAPF32[tempDoublePtr>>2]=$55,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $57 = +HEAPF32[$34>>2]; //@line 314 "libraries/Box2D/Dynamics/b2Body.cpp"
    $58 = +HEAPF32[$2>>2]; //@line 314 "libraries/Box2D/Dynamics/b2Body.cpp"
    $59 = $57 + $58; //@line 314 "libraries/Box2D/Dynamics/b2Body.cpp"
    HEAPF32[$2>>2] = $59; //@line 314 "libraries/Box2D/Dynamics/b2Body.cpp"
    $localCenter$sroa$0$1 = $53;$localCenter$sroa$10$1 = $56;
   }
   $60 = ((($f$012)) + 4|0); //@line 303 "libraries/Box2D/Dynamics/b2Body.cpp"
   $f$0 = HEAP32[$60>>2]|0; //@line 303 "libraries/Box2D/Dynamics/b2Body.cpp"
   $61 = ($f$0|0)==(0|0); //@line 303 "libraries/Box2D/Dynamics/b2Body.cpp"
   if ($61) {
    $localCenter$sroa$0$0$lcssa = $localCenter$sroa$0$1;$localCenter$sroa$10$0$lcssa = $localCenter$sroa$10$1;
    break;
   } else {
    $f$012 = $f$0;$localCenter$sroa$0$010 = $localCenter$sroa$0$1;$localCenter$sroa$10$011 = $localCenter$sroa$10$1;
   }
  }
 }
 $35 = +HEAPF32[$0>>2]; //@line 318 "libraries/Box2D/Dynamics/b2Body.cpp"
 $36 = $35 > 0.0; //@line 318 "libraries/Box2D/Dynamics/b2Body.cpp"
 if ($36) {
  $62 = 1.0 / $35; //@line 320 "libraries/Box2D/Dynamics/b2Body.cpp"
  HEAPF32[$1>>2] = $62; //@line 320 "libraries/Box2D/Dynamics/b2Body.cpp"
  $63 = (HEAP32[tempDoublePtr>>2]=$localCenter$sroa$0$0$lcssa,+HEAPF32[tempDoublePtr>>2]); //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $64 = $63 * $62; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $65 = (HEAPF32[tempDoublePtr>>2]=$64,HEAP32[tempDoublePtr>>2]|0); //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $66 = (HEAP32[tempDoublePtr>>2]=$localCenter$sroa$10$0$lcssa,+HEAPF32[tempDoublePtr>>2]); //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $67 = $66 * $62; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $68 = (HEAPF32[tempDoublePtr>>2]=$67,HEAP32[tempDoublePtr>>2]|0); //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $81 = $35;$localCenter$sroa$0$2 = $65;$localCenter$sroa$10$2 = $68;
 } else {
  HEAPF32[$0>>2] = 1.0; //@line 326 "libraries/Box2D/Dynamics/b2Body.cpp"
  HEAPF32[$1>>2] = 1.0; //@line 327 "libraries/Box2D/Dynamics/b2Body.cpp"
  $81 = 1.0;$localCenter$sroa$0$2 = $localCenter$sroa$0$0$lcssa;$localCenter$sroa$10$2 = $localCenter$sroa$10$0$lcssa;
 }
 $69 = +HEAPF32[$2>>2]; //@line 330 "libraries/Box2D/Dynamics/b2Body.cpp"
 $70 = $69 > 0.0; //@line 330 "libraries/Box2D/Dynamics/b2Body.cpp"
 if ($70) {
  $71 = ((($this)) + 4|0); //@line 330 "libraries/Box2D/Dynamics/b2Body.cpp"
  $72 = HEAP16[$71>>1]|0; //@line 330 "libraries/Box2D/Dynamics/b2Body.cpp"
  $73 = $72 & 16; //@line 330 "libraries/Box2D/Dynamics/b2Body.cpp"
  $74 = ($73<<16>>16)==(0); //@line 330 "libraries/Box2D/Dynamics/b2Body.cpp"
  if ($74) {
   $75 = (HEAP32[tempDoublePtr>>2]=$localCenter$sroa$0$2,+HEAPF32[tempDoublePtr>>2]); //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $76 = $75 * $75; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $77 = (HEAP32[tempDoublePtr>>2]=$localCenter$sroa$10$2,+HEAPF32[tempDoublePtr>>2]); //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $78 = $77 * $77; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $79 = $76 + $78; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $80 = $79 * $81; //@line 333 "libraries/Box2D/Dynamics/b2Body.cpp"
   $82 = $69 - $80; //@line 333 "libraries/Box2D/Dynamics/b2Body.cpp"
   HEAPF32[$2>>2] = $82; //@line 333 "libraries/Box2D/Dynamics/b2Body.cpp"
   $83 = $82 > 0.0; //@line 334 "libraries/Box2D/Dynamics/b2Body.cpp"
   __Z10loveAssertbPKc($83,446391); //@line 334 "libraries/Box2D/Dynamics/b2Body.cpp"
   $84 = +HEAPF32[$2>>2]; //@line 335 "libraries/Box2D/Dynamics/b2Body.cpp"
   $85 = 1.0 / $84; //@line 335 "libraries/Box2D/Dynamics/b2Body.cpp"
   $$cast$pre$phiZZ2D = $75;$$pre$phiZ2D = $77;$storemerge = $85;
  } else {
   label = 14;
  }
 } else {
  label = 14;
 }
 if ((label|0) == 14) {
  HEAPF32[$2>>2] = 0.0; //@line 340 "libraries/Box2D/Dynamics/b2Body.cpp"
  $$pre = (HEAP32[tempDoublePtr>>2]=$localCenter$sroa$0$2,+HEAPF32[tempDoublePtr>>2]); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $$pre14 = (HEAP32[tempDoublePtr>>2]=$localCenter$sroa$10$2,+HEAPF32[tempDoublePtr>>2]); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $$cast$pre$phiZZ2D = $$pre;$$pre$phiZ2D = $$pre14;$storemerge = 0.0;
 }
 HEAPF32[$3>>2] = $storemerge; //@line 335 "libraries/Box2D/Dynamics/b2Body.cpp"
 $86 = ((($this)) + 44|0); //@line 345 "libraries/Box2D/Dynamics/b2Body.cpp"
 $87 = +HEAPF32[$86>>2]; //@line 345 "libraries/Box2D/Dynamics/b2Body.cpp"
 $88 = ((($this)) + 48|0); //@line 345 "libraries/Box2D/Dynamics/b2Body.cpp"
 $89 = +HEAPF32[$88>>2]; //@line 345 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$4>>2] = $localCenter$sroa$0$2; //@line 346 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$5>>2] = $localCenter$sroa$10$2; //@line 346 "libraries/Box2D/Dynamics/b2Body.cpp"
 $90 = ((($this)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $91 = +HEAPF32[$90>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $92 = $$cast$pre$phiZZ2D * $91; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $93 = ((($this)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $94 = +HEAPF32[$93>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $95 = $94 * $$pre$phiZ2D; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $96 = $92 - $95; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $97 = ((($this)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $98 = +HEAPF32[$97>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $99 = $98 + $96; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $100 = $$cast$pre$phiZZ2D * $94; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $101 = $91 * $$pre$phiZ2D; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $102 = $100 + $101; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $103 = ((($this)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $104 = +HEAPF32[$103>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $105 = $102 + $104; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$86>>2] = $99; //@line 347 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$88>>2] = $105; //@line 347 "libraries/Box2D/Dynamics/b2Body.cpp"
 $106 = ((($this)) + 36|0); //@line 347 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$106>>2] = $99; //@line 347 "libraries/Box2D/Dynamics/b2Body.cpp"
 $107 = ((($this)) + 40|0); //@line 347 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$107>>2] = $105; //@line 347 "libraries/Box2D/Dynamics/b2Body.cpp"
 $108 = ((($this)) + 72|0); //@line 350 "libraries/Box2D/Dynamics/b2Body.cpp"
 $109 = +HEAPF32[$108>>2]; //@line 350 "libraries/Box2D/Dynamics/b2Body.cpp"
 $110 = $99 - $87; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $111 = $105 - $89; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $112 = $109 * $111; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $113 = $109 * $110; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $114 = ((($this)) + 64|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $115 = +HEAPF32[$114>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $116 = $115 - $112; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$114>>2] = $116; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $117 = ((($this)) + 68|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $118 = +HEAPF32[$117>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $119 = $113 + $118; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$117>>2] = $119; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 STACKTOP = sp;return; //@line 351 "libraries/Box2D/Dynamics/b2Body.cpp"
}
function __ZN6b2Body19SynchronizeFixturesEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0.0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0, $f$0 = 0, $f$03 = 0, $f$04 = 0, $xf1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $xf1 = sp;
 $0 = ((($this)) + 52|0); //@line 444 "libraries/Box2D/Dynamics/b2Body.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 444 "libraries/Box2D/Dynamics/b2Body.cpp"
 $2 = (+Math_sin((+$1))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($xf1)) + 8|0); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$3>>2] = $2; //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $4 = (+Math_cos((+$1))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($xf1)) + 12|0); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$5>>2] = $4; //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($this)) + 28|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $8 = $4 * $7; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 32|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $11 = $2 * $10; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $12 = $8 - $11; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $13 = $7 * $2; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $14 = $4 * $10; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $15 = $13 + $14; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $16 = ((($this)) + 36|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $17 = +HEAPF32[$16>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $18 = $17 - $12; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($this)) + 40|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $21 = $20 - $15; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$xf1>>2] = $18; //@line 445 "libraries/Box2D/Dynamics/b2Body.cpp"
 $22 = ((($xf1)) + 4|0); //@line 445 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$22>>2] = $21; //@line 445 "libraries/Box2D/Dynamics/b2Body.cpp"
 $23 = ((($this)) + 88|0); //@line 447 "libraries/Box2D/Dynamics/b2Body.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 447 "libraries/Box2D/Dynamics/b2Body.cpp"
 $25 = ((($24)) + 102872|0); //@line 447 "libraries/Box2D/Dynamics/b2Body.cpp"
 $26 = ((($this)) + 100|0); //@line 448 "libraries/Box2D/Dynamics/b2Body.cpp"
 $f$03 = HEAP32[$26>>2]|0; //@line 448 "libraries/Box2D/Dynamics/b2Body.cpp"
 $27 = ($f$03|0)==(0|0); //@line 448 "libraries/Box2D/Dynamics/b2Body.cpp"
 if ($27) {
  STACKTOP = sp;return; //@line 452 "libraries/Box2D/Dynamics/b2Body.cpp"
 }
 $28 = ((($this)) + 12|0); //@line 450 "libraries/Box2D/Dynamics/b2Body.cpp"
 $f$04 = $f$03;
 while(1) {
  __ZN9b2Fixture11SynchronizeEP12b2BroadPhaseRK11b2TransformS4_($f$04,$25,$xf1,$28); //@line 450 "libraries/Box2D/Dynamics/b2Body.cpp"
  $29 = ((($f$04)) + 4|0); //@line 448 "libraries/Box2D/Dynamics/b2Body.cpp"
  $f$0 = HEAP32[$29>>2]|0; //@line 448 "libraries/Box2D/Dynamics/b2Body.cpp"
  $30 = ($f$0|0)==(0|0); //@line 448 "libraries/Box2D/Dynamics/b2Body.cpp"
  if ($30) {
   break;
  } else {
   $f$04 = $f$0;
  }
 }
 STACKTOP = sp;return; //@line 452 "libraries/Box2D/Dynamics/b2Body.cpp"
}
function __ZN6b2Body13CreateFixtureEPK12b2FixtureDef($this,$def) {
 $this = $this|0;
 $def = $def|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0.0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 88|0); //@line 168 "libraries/Box2D/Dynamics/b2Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 168 "libraries/Box2D/Dynamics/b2Body.cpp"
 $2 = ((($1)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $3 = HEAP32[$2>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $4 = $3 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $5 = ($4|0)==(0); //@line 168 "libraries/Box2D/Dynamics/b2Body.cpp"
 __Z10loveAssertbPKc($5,446337); //@line 168 "libraries/Box2D/Dynamics/b2Body.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 169 "libraries/Box2D/Dynamics/b2Body.cpp"
 $7 = ((($6)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $8 = HEAP32[$7>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $9 = $8 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $10 = ($9|0)==(0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 if (!($10)) {
  $$0 = 0;
  return ($$0|0); //@line 203 "libraries/Box2D/Dynamics/b2Body.cpp"
 }
 $11 = (__ZN16b2BlockAllocator8AllocateEi($6,44)|0); //@line 176 "libraries/Box2D/Dynamics/b2Body.cpp"
 __ZN9b2FixtureC2Ev($11); //@line 177 "libraries/Box2D/Dynamics/b2Body.cpp"
 __ZN9b2Fixture6CreateEP16b2BlockAllocatorP6b2BodyPK12b2FixtureDef($11,$6,$this,$def); //@line 178 "libraries/Box2D/Dynamics/b2Body.cpp"
 $12 = ((($this)) + 4|0); //@line 180 "libraries/Box2D/Dynamics/b2Body.cpp"
 $13 = HEAP16[$12>>1]|0; //@line 180 "libraries/Box2D/Dynamics/b2Body.cpp"
 $14 = $13 & 32; //@line 180 "libraries/Box2D/Dynamics/b2Body.cpp"
 $15 = ($14<<16>>16)==(0); //@line 180 "libraries/Box2D/Dynamics/b2Body.cpp"
 if (!($15)) {
  $16 = HEAP32[$0>>2]|0; //@line 182 "libraries/Box2D/Dynamics/b2Body.cpp"
  $17 = ((($16)) + 102872|0); //@line 182 "libraries/Box2D/Dynamics/b2Body.cpp"
  $18 = ((($this)) + 12|0); //@line 183 "libraries/Box2D/Dynamics/b2Body.cpp"
  __ZN9b2Fixture13CreateProxiesEP12b2BroadPhaseRK11b2Transform($11,$17,$18); //@line 183 "libraries/Box2D/Dynamics/b2Body.cpp"
 }
 $19 = ((($this)) + 100|0); //@line 186 "libraries/Box2D/Dynamics/b2Body.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 186 "libraries/Box2D/Dynamics/b2Body.cpp"
 $21 = ((($11)) + 4|0); //@line 186 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$21>>2] = $20; //@line 186 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$19>>2] = $11; //@line 187 "libraries/Box2D/Dynamics/b2Body.cpp"
 $22 = ((($this)) + 104|0); //@line 188 "libraries/Box2D/Dynamics/b2Body.cpp"
 $23 = HEAP32[$22>>2]|0; //@line 188 "libraries/Box2D/Dynamics/b2Body.cpp"
 $24 = (($23) + 1)|0; //@line 188 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$22>>2] = $24; //@line 188 "libraries/Box2D/Dynamics/b2Body.cpp"
 $25 = ((($11)) + 8|0); //@line 190 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$25>>2] = $this; //@line 190 "libraries/Box2D/Dynamics/b2Body.cpp"
 $26 = +HEAPF32[$11>>2]; //@line 193 "libraries/Box2D/Dynamics/b2Body.cpp"
 $27 = $26 > 0.0; //@line 193 "libraries/Box2D/Dynamics/b2Body.cpp"
 if ($27) {
  __ZN6b2Body13ResetMassDataEv($this); //@line 195 "libraries/Box2D/Dynamics/b2Body.cpp"
 }
 $28 = HEAP32[$0>>2]|0; //@line 200 "libraries/Box2D/Dynamics/b2Body.cpp"
 $29 = ((($28)) + 102868|0); //@line 200 "libraries/Box2D/Dynamics/b2Body.cpp"
 $30 = HEAP32[$29>>2]|0; //@line 200 "libraries/Box2D/Dynamics/b2Body.cpp"
 $31 = $30 | 1; //@line 200 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$29>>2] = $31; //@line 200 "libraries/Box2D/Dynamics/b2Body.cpp"
 $$0 = $11;
 return ($$0|0); //@line 203 "libraries/Box2D/Dynamics/b2Body.cpp"
}
function __ZN6b2Body14DestroyFixtureEP9b2Fixture($this,$fixture) {
 $this = $this|0;
 $fixture = $fixture|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $edge$02 = 0, $found$0$off0 = 0, $node$03 = 0, $node$03$lcssa = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 88|0); //@line 216 "libraries/Box2D/Dynamics/b2Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 216 "libraries/Box2D/Dynamics/b2Body.cpp"
 $2 = ((($1)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $3 = HEAP32[$2>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $4 = $3 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $5 = ($4|0)==(0); //@line 216 "libraries/Box2D/Dynamics/b2Body.cpp"
 __Z10loveAssertbPKc($5,446337); //@line 216 "libraries/Box2D/Dynamics/b2Body.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 217 "libraries/Box2D/Dynamics/b2Body.cpp"
 $7 = ((($6)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $8 = HEAP32[$7>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $9 = $8 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $10 = ($9|0)==(0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 if (!($10)) {
  return; //@line 279 "libraries/Box2D/Dynamics/b2Body.cpp"
 }
 $11 = ((($fixture)) + 8|0); //@line 222 "libraries/Box2D/Dynamics/b2Body.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 222 "libraries/Box2D/Dynamics/b2Body.cpp"
 $13 = ($12|0)==($this|0); //@line 222 "libraries/Box2D/Dynamics/b2Body.cpp"
 __Z10loveAssertbPKc($13,446402); //@line 222 "libraries/Box2D/Dynamics/b2Body.cpp"
 $14 = ((($this)) + 104|0); //@line 225 "libraries/Box2D/Dynamics/b2Body.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 225 "libraries/Box2D/Dynamics/b2Body.cpp"
 $16 = ($15|0)>(0); //@line 225 "libraries/Box2D/Dynamics/b2Body.cpp"
 __Z10loveAssertbPKc($16,446426); //@line 225 "libraries/Box2D/Dynamics/b2Body.cpp"
 $17 = ((($this)) + 100|0); //@line 226 "libraries/Box2D/Dynamics/b2Body.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 228 "libraries/Box2D/Dynamics/b2Body.cpp"
 $19 = ($18|0)==(0|0); //@line 228 "libraries/Box2D/Dynamics/b2Body.cpp"
 L4: do {
  if ($19) {
   $found$0$off0 = 0;
  } else {
   $24 = $18;$node$03 = $17;
   while(1) {
    $23 = ($24|0)==($fixture|0); //@line 230 "libraries/Box2D/Dynamics/b2Body.cpp"
    $21 = ((($24)) + 4|0); //@line 237 "libraries/Box2D/Dynamics/b2Body.cpp"
    if ($23) {
     $node$03$lcssa = $node$03;
     break;
    }
    $20 = HEAP32[$21>>2]|0; //@line 228 "libraries/Box2D/Dynamics/b2Body.cpp"
    $22 = ($20|0)==(0|0); //@line 228 "libraries/Box2D/Dynamics/b2Body.cpp"
    if ($22) {
     $found$0$off0 = 0;
     break L4;
    } else {
     $24 = $20;$node$03 = $21;
    }
   }
   $25 = ((($fixture)) + 4|0); //@line 232 "libraries/Box2D/Dynamics/b2Body.cpp"
   $26 = HEAP32[$25>>2]|0; //@line 232 "libraries/Box2D/Dynamics/b2Body.cpp"
   HEAP32[$node$03$lcssa>>2] = $26; //@line 232 "libraries/Box2D/Dynamics/b2Body.cpp"
   $found$0$off0 = 1;
  }
 } while(0);
 __Z10loveAssertbPKc($found$0$off0,446445); //@line 241 "libraries/Box2D/Dynamics/b2Body.cpp"
 $27 = ((($this)) + 112|0); //@line 244 "libraries/Box2D/Dynamics/b2Body.cpp"
 $28 = HEAP32[$27>>2]|0; //@line 244 "libraries/Box2D/Dynamics/b2Body.cpp"
 $29 = ($28|0)==(0|0); //@line 245 "libraries/Box2D/Dynamics/b2Body.cpp"
 if (!($29)) {
  $edge$02 = $28;
  while(1) {
   $30 = ((($edge$02)) + 4|0); //@line 247 "libraries/Box2D/Dynamics/b2Body.cpp"
   $31 = HEAP32[$30>>2]|0; //@line 247 "libraries/Box2D/Dynamics/b2Body.cpp"
   $32 = ((($edge$02)) + 12|0); //@line 248 "libraries/Box2D/Dynamics/b2Body.cpp"
   $33 = HEAP32[$32>>2]|0; //@line 248 "libraries/Box2D/Dynamics/b2Body.cpp"
   $34 = ((($31)) + 48|0); //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
   $35 = HEAP32[$34>>2]|0; //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
   $36 = ((($31)) + 52|0); //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
   $37 = HEAP32[$36>>2]|0; //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
   $38 = ($35|0)==($fixture|0); //@line 253 "libraries/Box2D/Dynamics/b2Body.cpp"
   $39 = ($37|0)==($fixture|0); //@line 253 "libraries/Box2D/Dynamics/b2Body.cpp"
   $or$cond = $38 | $39; //@line 253 "libraries/Box2D/Dynamics/b2Body.cpp"
   if ($or$cond) {
    $41 = HEAP32[$0>>2]|0; //@line 257 "libraries/Box2D/Dynamics/b2Body.cpp"
    $42 = ((($41)) + 102872|0); //@line 257 "libraries/Box2D/Dynamics/b2Body.cpp"
    __ZN16b2ContactManager7DestroyEP9b2Contact($42,$31); //@line 257 "libraries/Box2D/Dynamics/b2Body.cpp"
   }
   $40 = ($33|0)==(0|0); //@line 245 "libraries/Box2D/Dynamics/b2Body.cpp"
   if ($40) {
    break;
   } else {
    $edge$02 = $33;
   }
  }
 }
 $43 = HEAP32[$0>>2]|0; //@line 261 "libraries/Box2D/Dynamics/b2Body.cpp"
 $44 = ((($this)) + 4|0); //@line 263 "libraries/Box2D/Dynamics/b2Body.cpp"
 $45 = HEAP16[$44>>1]|0; //@line 263 "libraries/Box2D/Dynamics/b2Body.cpp"
 $46 = $45 & 32; //@line 263 "libraries/Box2D/Dynamics/b2Body.cpp"
 $47 = ($46<<16>>16)==(0); //@line 263 "libraries/Box2D/Dynamics/b2Body.cpp"
 if (!($47)) {
  $48 = ((($43)) + 102872|0); //@line 265 "libraries/Box2D/Dynamics/b2Body.cpp"
  __ZN9b2Fixture14DestroyProxiesEP12b2BroadPhase($fixture,$48); //@line 266 "libraries/Box2D/Dynamics/b2Body.cpp"
 }
 __ZN9b2Fixture7DestroyEP16b2BlockAllocator($fixture,$43); //@line 269 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$11>>2] = 0; //@line 270 "libraries/Box2D/Dynamics/b2Body.cpp"
 $49 = ((($fixture)) + 4|0); //@line 271 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$49>>2] = 0; //@line 271 "libraries/Box2D/Dynamics/b2Body.cpp"
 __ZN16b2BlockAllocator4FreeEPvi($43,$fixture,44); //@line 273 "libraries/Box2D/Dynamics/b2Body.cpp"
 $50 = HEAP32[$14>>2]|0; //@line 275 "libraries/Box2D/Dynamics/b2Body.cpp"
 $51 = (($50) + -1)|0; //@line 275 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$14>>2] = $51; //@line 275 "libraries/Box2D/Dynamics/b2Body.cpp"
 __ZN6b2Body13ResetMassDataEv($this); //@line 278 "libraries/Box2D/Dynamics/b2Body.cpp"
 return; //@line 279 "libraries/Box2D/Dynamics/b2Body.cpp"
}
function __ZN6b2Body11SetMassDataEPK10b2MassData($this,$massData) {
 $this = $this|0;
 $massData = $massData|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0, $43 = 0.0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0.0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0;
 var $63 = 0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0, $73 = 0, $74 = 0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0;
 var $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0.0, $9 = 0, $bitcast = 0.0, $storemerge = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 88|0); //@line 355 "libraries/Box2D/Dynamics/b2Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 355 "libraries/Box2D/Dynamics/b2Body.cpp"
 $2 = ((($1)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $3 = HEAP32[$2>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $4 = $3 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $5 = ($4|0)==(0); //@line 355 "libraries/Box2D/Dynamics/b2Body.cpp"
 __Z10loveAssertbPKc($5,446337); //@line 355 "libraries/Box2D/Dynamics/b2Body.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 356 "libraries/Box2D/Dynamics/b2Body.cpp"
 $7 = ((($6)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $8 = HEAP32[$7>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $9 = $8 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $10 = ($9|0)==(0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 if (!($10)) {
  return; //@line 392 "libraries/Box2D/Dynamics/b2Body.cpp"
 }
 $11 = HEAP32[$this>>2]|0; //@line 361 "libraries/Box2D/Dynamics/b2Body.cpp"
 $12 = ($11|0)==(2); //@line 361 "libraries/Box2D/Dynamics/b2Body.cpp"
 if (!($12)) {
  return; //@line 392 "libraries/Box2D/Dynamics/b2Body.cpp"
 }
 $13 = ((($this)) + 120|0); //@line 366 "libraries/Box2D/Dynamics/b2Body.cpp"
 $14 = ((($this)) + 124|0); //@line 367 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$14>>2] = 0.0; //@line 367 "libraries/Box2D/Dynamics/b2Body.cpp"
 $15 = ((($this)) + 128|0); //@line 368 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$15>>2] = 0.0; //@line 368 "libraries/Box2D/Dynamics/b2Body.cpp"
 $16 = +HEAPF32[$massData>>2]; //@line 370 "libraries/Box2D/Dynamics/b2Body.cpp"
 $17 = ((($this)) + 116|0); //@line 370 "libraries/Box2D/Dynamics/b2Body.cpp"
 $18 = !($16 <= 0.0); //@line 371 "libraries/Box2D/Dynamics/b2Body.cpp"
 $storemerge = $18 ? $16 : 1.0; //@line 380 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$17>>2] = $storemerge; //@line 370 "libraries/Box2D/Dynamics/b2Body.cpp"
 $19 = 1.0 / $storemerge; //@line 376 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$13>>2] = $19; //@line 376 "libraries/Box2D/Dynamics/b2Body.cpp"
 $20 = ((($massData)) + 12|0); //@line 378 "libraries/Box2D/Dynamics/b2Body.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 378 "libraries/Box2D/Dynamics/b2Body.cpp"
 $22 = $21 > 0.0; //@line 378 "libraries/Box2D/Dynamics/b2Body.cpp"
 if ($22) {
  $23 = ((($this)) + 4|0); //@line 378 "libraries/Box2D/Dynamics/b2Body.cpp"
  $24 = HEAP16[$23>>1]|0; //@line 378 "libraries/Box2D/Dynamics/b2Body.cpp"
  $25 = $24 & 16; //@line 378 "libraries/Box2D/Dynamics/b2Body.cpp"
  $26 = ($25<<16>>16)==(0); //@line 378 "libraries/Box2D/Dynamics/b2Body.cpp"
  if ($26) {
   $27 = ((($massData)) + 4|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $28 = +HEAPF32[$27>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $29 = $28 * $28; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $30 = ((($massData)) + 8|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $31 = +HEAPF32[$30>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $32 = $31 * $31; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $33 = $29 + $32; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $34 = $storemerge * $33; //@line 380 "libraries/Box2D/Dynamics/b2Body.cpp"
   $35 = $21 - $34; //@line 380 "libraries/Box2D/Dynamics/b2Body.cpp"
   HEAPF32[$14>>2] = $35; //@line 380 "libraries/Box2D/Dynamics/b2Body.cpp"
   $36 = $35 > 0.0; //@line 381 "libraries/Box2D/Dynamics/b2Body.cpp"
   __Z10loveAssertbPKc($36,446391); //@line 381 "libraries/Box2D/Dynamics/b2Body.cpp"
   $37 = +HEAPF32[$14>>2]; //@line 382 "libraries/Box2D/Dynamics/b2Body.cpp"
   $38 = 1.0 / $37; //@line 382 "libraries/Box2D/Dynamics/b2Body.cpp"
   HEAPF32[$15>>2] = $38; //@line 382 "libraries/Box2D/Dynamics/b2Body.cpp"
  }
 }
 $39 = ((($this)) + 28|0); //@line 386 "libraries/Box2D/Dynamics/b2Body.cpp"
 $40 = ((($this)) + 44|0); //@line 386 "libraries/Box2D/Dynamics/b2Body.cpp"
 $41 = +HEAPF32[$40>>2]; //@line 386 "libraries/Box2D/Dynamics/b2Body.cpp"
 $42 = ((($this)) + 48|0); //@line 386 "libraries/Box2D/Dynamics/b2Body.cpp"
 $43 = +HEAPF32[$42>>2]; //@line 386 "libraries/Box2D/Dynamics/b2Body.cpp"
 $44 = ((($massData)) + 4|0); //@line 387 "libraries/Box2D/Dynamics/b2Body.cpp"
 $45 = $44; //@line 387 "libraries/Box2D/Dynamics/b2Body.cpp"
 $46 = $45; //@line 387 "libraries/Box2D/Dynamics/b2Body.cpp"
 $47 = HEAP32[$46>>2]|0; //@line 387 "libraries/Box2D/Dynamics/b2Body.cpp"
 $48 = (($45) + 4)|0; //@line 387 "libraries/Box2D/Dynamics/b2Body.cpp"
 $49 = $48; //@line 387 "libraries/Box2D/Dynamics/b2Body.cpp"
 $50 = HEAP32[$49>>2]|0; //@line 387 "libraries/Box2D/Dynamics/b2Body.cpp"
 $51 = $39; //@line 387 "libraries/Box2D/Dynamics/b2Body.cpp"
 $52 = $51; //@line 387 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$52>>2] = $47; //@line 387 "libraries/Box2D/Dynamics/b2Body.cpp"
 $53 = (($51) + 4)|0; //@line 387 "libraries/Box2D/Dynamics/b2Body.cpp"
 $54 = $53; //@line 387 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$54>>2] = $50; //@line 387 "libraries/Box2D/Dynamics/b2Body.cpp"
 $55 = ((($this)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $56 = +HEAPF32[$55>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $bitcast = (HEAP32[tempDoublePtr>>2]=$47,+HEAPF32[tempDoublePtr>>2]); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $57 = $56 * $bitcast; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $58 = ((($this)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $59 = +HEAPF32[$58>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $60 = (HEAP32[tempDoublePtr>>2]=$50,+HEAPF32[tempDoublePtr>>2]); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $61 = $59 * $60; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $62 = $57 - $61; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $63 = ((($this)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $64 = +HEAPF32[$63>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $65 = $64 + $62; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $66 = $bitcast * $59; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $67 = $56 * $60; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $68 = $66 + $67; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $69 = ((($this)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $70 = +HEAPF32[$69>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $71 = $68 + $70; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$40>>2] = $65; //@line 388 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$42>>2] = $71; //@line 388 "libraries/Box2D/Dynamics/b2Body.cpp"
 $72 = ((($this)) + 36|0); //@line 388 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$72>>2] = $65; //@line 388 "libraries/Box2D/Dynamics/b2Body.cpp"
 $73 = ((($this)) + 40|0); //@line 388 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$73>>2] = $71; //@line 388 "libraries/Box2D/Dynamics/b2Body.cpp"
 $74 = ((($this)) + 72|0); //@line 391 "libraries/Box2D/Dynamics/b2Body.cpp"
 $75 = +HEAPF32[$74>>2]; //@line 391 "libraries/Box2D/Dynamics/b2Body.cpp"
 $76 = $65 - $41; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $77 = $71 - $43; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $78 = $75 * $77; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $79 = $75 * $76; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $80 = ((($this)) + 64|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $81 = +HEAPF32[$80>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $82 = $81 - $78; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$80>>2] = $82; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $83 = ((($this)) + 68|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $84 = +HEAPF32[$83>>2]; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $85 = $79 + $84; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$83>>2] = $85; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 392 "libraries/Box2D/Dynamics/b2Body.cpp"
}
function __ZNK6b2Body13ShouldCollideEPKS_($this,$other) {
 $this = $this|0;
 $other = $other|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $jn$0 = 0, $jn$02 = 0, $jn$03 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 397 "libraries/Box2D/Dynamics/b2Body.cpp"
 $1 = ($0|0)==(2); //@line 397 "libraries/Box2D/Dynamics/b2Body.cpp"
 if (!($1)) {
  $2 = HEAP32[$other>>2]|0; //@line 397 "libraries/Box2D/Dynamics/b2Body.cpp"
  $3 = ($2|0)==(2); //@line 397 "libraries/Box2D/Dynamics/b2Body.cpp"
  if (!($3)) {
   $15 = 0;
   return ($15|0);
  }
 }
 $4 = ((($this)) + 108|0); //@line 403 "libraries/Box2D/Dynamics/b2Body.cpp"
 $jn$02 = HEAP32[$4>>2]|0; //@line 403 "libraries/Box2D/Dynamics/b2Body.cpp"
 $5 = ($jn$02|0)==(0|0); //@line 403 "libraries/Box2D/Dynamics/b2Body.cpp"
 if ($5) {
  $15 = 1;
  return ($15|0);
 } else {
  $jn$03 = $jn$02;
 }
 while(1) {
  $6 = HEAP32[$jn$03>>2]|0; //@line 405 "libraries/Box2D/Dynamics/b2Body.cpp"
  $7 = ($6|0)==($other|0); //@line 405 "libraries/Box2D/Dynamics/b2Body.cpp"
  if ($7) {
   $8 = ((($jn$03)) + 4|0); //@line 407 "libraries/Box2D/Dynamics/b2Body.cpp"
   $9 = HEAP32[$8>>2]|0; //@line 407 "libraries/Box2D/Dynamics/b2Body.cpp"
   $10 = ((($9)) + 61|0); //@line 407 "libraries/Box2D/Dynamics/b2Body.cpp"
   $11 = HEAP8[$10>>0]|0; //@line 407 "libraries/Box2D/Dynamics/b2Body.cpp"
   $12 = ($11<<24>>24)==(0); //@line 407 "libraries/Box2D/Dynamics/b2Body.cpp"
   if ($12) {
    $15 = 0;
    label = 7;
    break;
   }
  }
  $13 = ((($jn$03)) + 12|0); //@line 403 "libraries/Box2D/Dynamics/b2Body.cpp"
  $jn$0 = HEAP32[$13>>2]|0; //@line 403 "libraries/Box2D/Dynamics/b2Body.cpp"
  $14 = ($jn$0|0)==(0|0); //@line 403 "libraries/Box2D/Dynamics/b2Body.cpp"
  if ($14) {
   $15 = 1;
   label = 7;
   break;
  } else {
   $jn$03 = $jn$0;
  }
 }
 if ((label|0) == 7) {
  return ($15|0);
 }
 return (0)|0;
}
function __ZN6b2Body12SetTransformERK6b2Vec2f($this,$position,$angle) {
 $this = $this|0;
 $position = $position|0;
 $angle = +$angle;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $bitcast = 0.0, $f$0 = 0, $f$01 = 0, $f$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 88|0); //@line 419 "libraries/Box2D/Dynamics/b2Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 419 "libraries/Box2D/Dynamics/b2Body.cpp"
 $2 = ((($1)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $3 = HEAP32[$2>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $4 = $3 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $5 = ($4|0)==(0); //@line 419 "libraries/Box2D/Dynamics/b2Body.cpp"
 __Z10loveAssertbPKc($5,446337); //@line 419 "libraries/Box2D/Dynamics/b2Body.cpp"
 $6 = HEAP32[$0>>2]|0; //@line 434 "libraries/Box2D/Dynamics/b2Body.cpp"
 $7 = ((($6)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $8 = HEAP32[$7>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $9 = $8 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $10 = ($9|0)==(0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 if (!($10)) {
  return; //@line 439 "libraries/Box2D/Dynamics/b2Body.cpp"
 }
 $11 = ((($this)) + 12|0); //@line 425 "libraries/Box2D/Dynamics/b2Body.cpp"
 $12 = (+Math_sin((+$angle))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($this)) + 20|0); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$13>>2] = $12; //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $14 = (+Math_cos((+$angle))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $15 = ((($this)) + 24|0); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$15>>2] = $14; //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $16 = $position; //@line 426 "libraries/Box2D/Dynamics/b2Body.cpp"
 $17 = $16; //@line 426 "libraries/Box2D/Dynamics/b2Body.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 426 "libraries/Box2D/Dynamics/b2Body.cpp"
 $19 = (($16) + 4)|0; //@line 426 "libraries/Box2D/Dynamics/b2Body.cpp"
 $20 = $19; //@line 426 "libraries/Box2D/Dynamics/b2Body.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 426 "libraries/Box2D/Dynamics/b2Body.cpp"
 $22 = $11; //@line 426 "libraries/Box2D/Dynamics/b2Body.cpp"
 $23 = $22; //@line 426 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$23>>2] = $18; //@line 426 "libraries/Box2D/Dynamics/b2Body.cpp"
 $24 = (($22) + 4)|0; //@line 426 "libraries/Box2D/Dynamics/b2Body.cpp"
 $25 = $24; //@line 426 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP32[$25>>2] = $21; //@line 426 "libraries/Box2D/Dynamics/b2Body.cpp"
 $26 = ((($this)) + 28|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $27 = +HEAPF32[$26>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $28 = $14 * $27; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $29 = ((($this)) + 32|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $30 = +HEAPF32[$29>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $31 = $12 * $30; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $32 = $28 - $31; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $bitcast = (HEAP32[tempDoublePtr>>2]=$18,+HEAPF32[tempDoublePtr>>2]); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $33 = $bitcast + $32; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $34 = $27 * $12; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $35 = $14 * $30; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $36 = $34 + $35; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $37 = (HEAP32[tempDoublePtr>>2]=$21,+HEAPF32[tempDoublePtr>>2]); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $38 = $36 + $37; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $39 = ((($this)) + 44|0); //@line 428 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$39>>2] = $33; //@line 428 "libraries/Box2D/Dynamics/b2Body.cpp"
 $40 = ((($this)) + 48|0); //@line 428 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$40>>2] = $38; //@line 428 "libraries/Box2D/Dynamics/b2Body.cpp"
 $41 = ((($this)) + 56|0); //@line 429 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$41>>2] = $angle; //@line 429 "libraries/Box2D/Dynamics/b2Body.cpp"
 $42 = ((($this)) + 36|0); //@line 431 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$42>>2] = $33; //@line 431 "libraries/Box2D/Dynamics/b2Body.cpp"
 $43 = ((($this)) + 40|0); //@line 431 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$43>>2] = $38; //@line 431 "libraries/Box2D/Dynamics/b2Body.cpp"
 $44 = ((($this)) + 52|0); //@line 432 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$44>>2] = $angle; //@line 432 "libraries/Box2D/Dynamics/b2Body.cpp"
 $45 = ((($6)) + 102872|0); //@line 434 "libraries/Box2D/Dynamics/b2Body.cpp"
 $46 = ((($this)) + 100|0); //@line 435 "libraries/Box2D/Dynamics/b2Body.cpp"
 $f$01 = HEAP32[$46>>2]|0; //@line 435 "libraries/Box2D/Dynamics/b2Body.cpp"
 $47 = ($f$01|0)==(0|0); //@line 435 "libraries/Box2D/Dynamics/b2Body.cpp"
 if ($47) {
  return; //@line 439 "libraries/Box2D/Dynamics/b2Body.cpp"
 } else {
  $f$02 = $f$01;
 }
 while(1) {
  __ZN9b2Fixture11SynchronizeEP12b2BroadPhaseRK11b2TransformS4_($f$02,$45,$11,$11); //@line 437 "libraries/Box2D/Dynamics/b2Body.cpp"
  $48 = ((($f$02)) + 4|0); //@line 435 "libraries/Box2D/Dynamics/b2Body.cpp"
  $f$0 = HEAP32[$48>>2]|0; //@line 435 "libraries/Box2D/Dynamics/b2Body.cpp"
  $49 = ($f$0|0)==(0|0); //@line 435 "libraries/Box2D/Dynamics/b2Body.cpp"
  if ($49) {
   break;
  } else {
   $f$02 = $f$0;
  }
 }
 return; //@line 439 "libraries/Box2D/Dynamics/b2Body.cpp"
}
function __ZN6b2Body9SetActiveEb($this,$flag) {
 $this = $this|0;
 $flag = $flag|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ce$03 = 0, $f$0 = 0;
 var $f$01 = 0, $f$02 = 0, $f2$0 = 0, $f2$05 = 0, $f2$06 = 0, $tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 88|0); //@line 456 "libraries/Box2D/Dynamics/b2Body.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 456 "libraries/Box2D/Dynamics/b2Body.cpp"
 $2 = ((($1)) + 102868|0); //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $3 = HEAP32[$2>>2]|0; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $4 = $3 & 2; //@line 323 "./libraries/Box2D/Dynamics/b2World.h"
 $5 = ($4|0)==(0); //@line 456 "libraries/Box2D/Dynamics/b2Body.cpp"
 __Z10loveAssertbPKc($5,446337); //@line 456 "libraries/Box2D/Dynamics/b2Body.cpp"
 $6 = ((($this)) + 4|0); //@line 661 "./libraries/Box2D/Dynamics/b2Body.h"
 $7 = HEAP16[$6>>1]|0; //@line 661 "./libraries/Box2D/Dynamics/b2Body.h"
 $8 = $7 & 32; //@line 661 "./libraries/Box2D/Dynamics/b2Body.h"
 $9 = ($8<<16>>16)!=(0); //@line 661 "./libraries/Box2D/Dynamics/b2Body.h"
 $tmp = $9 ^ $flag; //@line 458 "libraries/Box2D/Dynamics/b2Body.cpp"
 if (!($tmp)) {
  return; //@line 497 "libraries/Box2D/Dynamics/b2Body.cpp"
 }
 $10 = $7&65535; //@line 465 "libraries/Box2D/Dynamics/b2Body.cpp"
 if ($flag) {
  $11 = $10 | 32; //@line 465 "libraries/Box2D/Dynamics/b2Body.cpp"
  $12 = $11&65535; //@line 465 "libraries/Box2D/Dynamics/b2Body.cpp"
  HEAP16[$6>>1] = $12; //@line 465 "libraries/Box2D/Dynamics/b2Body.cpp"
  $13 = HEAP32[$0>>2]|0; //@line 468 "libraries/Box2D/Dynamics/b2Body.cpp"
  $14 = ((($13)) + 102872|0); //@line 468 "libraries/Box2D/Dynamics/b2Body.cpp"
  $15 = ((($this)) + 100|0); //@line 469 "libraries/Box2D/Dynamics/b2Body.cpp"
  $f$01 = HEAP32[$15>>2]|0; //@line 469 "libraries/Box2D/Dynamics/b2Body.cpp"
  $16 = ($f$01|0)==(0|0); //@line 469 "libraries/Box2D/Dynamics/b2Body.cpp"
  if ($16) {
   return; //@line 497 "libraries/Box2D/Dynamics/b2Body.cpp"
  }
  $17 = ((($this)) + 12|0); //@line 471 "libraries/Box2D/Dynamics/b2Body.cpp"
  $f$02 = $f$01;
  while(1) {
   __ZN9b2Fixture13CreateProxiesEP12b2BroadPhaseRK11b2Transform($f$02,$14,$17); //@line 471 "libraries/Box2D/Dynamics/b2Body.cpp"
   $18 = ((($f$02)) + 4|0); //@line 469 "libraries/Box2D/Dynamics/b2Body.cpp"
   $f$0 = HEAP32[$18>>2]|0; //@line 469 "libraries/Box2D/Dynamics/b2Body.cpp"
   $19 = ($f$0|0)==(0|0); //@line 469 "libraries/Box2D/Dynamics/b2Body.cpp"
   if ($19) {
    break;
   } else {
    $f$02 = $f$0;
   }
  }
  return; //@line 497 "libraries/Box2D/Dynamics/b2Body.cpp"
 }
 $20 = $10 & 65503; //@line 478 "libraries/Box2D/Dynamics/b2Body.cpp"
 $21 = $20&65535; //@line 478 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP16[$6>>1] = $21; //@line 478 "libraries/Box2D/Dynamics/b2Body.cpp"
 $22 = HEAP32[$0>>2]|0; //@line 481 "libraries/Box2D/Dynamics/b2Body.cpp"
 $23 = ((($22)) + 102872|0); //@line 481 "libraries/Box2D/Dynamics/b2Body.cpp"
 $24 = ((($this)) + 100|0); //@line 482 "libraries/Box2D/Dynamics/b2Body.cpp"
 $f2$05 = HEAP32[$24>>2]|0; //@line 482 "libraries/Box2D/Dynamics/b2Body.cpp"
 $25 = ($f2$05|0)==(0|0); //@line 482 "libraries/Box2D/Dynamics/b2Body.cpp"
 if (!($25)) {
  $f2$06 = $f2$05;
  while(1) {
   __ZN9b2Fixture14DestroyProxiesEP12b2BroadPhase($f2$06,$23); //@line 484 "libraries/Box2D/Dynamics/b2Body.cpp"
   $29 = ((($f2$06)) + 4|0); //@line 482 "libraries/Box2D/Dynamics/b2Body.cpp"
   $f2$0 = HEAP32[$29>>2]|0; //@line 482 "libraries/Box2D/Dynamics/b2Body.cpp"
   $30 = ($f2$0|0)==(0|0); //@line 482 "libraries/Box2D/Dynamics/b2Body.cpp"
   if ($30) {
    break;
   } else {
    $f2$06 = $f2$0;
   }
  }
 }
 $26 = ((($this)) + 112|0); //@line 488 "libraries/Box2D/Dynamics/b2Body.cpp"
 $27 = HEAP32[$26>>2]|0; //@line 488 "libraries/Box2D/Dynamics/b2Body.cpp"
 $28 = ($27|0)==(0|0); //@line 489 "libraries/Box2D/Dynamics/b2Body.cpp"
 if (!($28)) {
  $ce$03 = $27;
  while(1) {
   $31 = ((($ce$03)) + 12|0); //@line 492 "libraries/Box2D/Dynamics/b2Body.cpp"
   $32 = HEAP32[$31>>2]|0; //@line 492 "libraries/Box2D/Dynamics/b2Body.cpp"
   $33 = HEAP32[$0>>2]|0; //@line 493 "libraries/Box2D/Dynamics/b2Body.cpp"
   $34 = ((($33)) + 102872|0); //@line 493 "libraries/Box2D/Dynamics/b2Body.cpp"
   $35 = ((($ce$03)) + 4|0); //@line 493 "libraries/Box2D/Dynamics/b2Body.cpp"
   $36 = HEAP32[$35>>2]|0; //@line 493 "libraries/Box2D/Dynamics/b2Body.cpp"
   __ZN16b2ContactManager7DestroyEP9b2Contact($34,$36); //@line 493 "libraries/Box2D/Dynamics/b2Body.cpp"
   $37 = ($32|0)==(0|0); //@line 489 "libraries/Box2D/Dynamics/b2Body.cpp"
   if ($37) {
    break;
   } else {
    $ce$03 = $32;
   }
  }
 }
 HEAP32[$26>>2] = 0; //@line 495 "libraries/Box2D/Dynamics/b2Body.cpp"
 return; //@line 497 "libraries/Box2D/Dynamics/b2Body.cpp"
}
function __ZN6b2Body16SetFixedRotationEb($this,$flag) {
 $this = $this|0;
 $flag = $flag|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $storemerge = 0, $storemerge$in = 0, $tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 501 "libraries/Box2D/Dynamics/b2Body.cpp"
 $1 = HEAP16[$0>>1]|0; //@line 501 "libraries/Box2D/Dynamics/b2Body.cpp"
 $2 = $1&65535; //@line 501 "libraries/Box2D/Dynamics/b2Body.cpp"
 $3 = $2 & 16; //@line 502 "libraries/Box2D/Dynamics/b2Body.cpp"
 $4 = ($3|0)!=(0); //@line 502 "libraries/Box2D/Dynamics/b2Body.cpp"
 $tmp = $4 ^ $flag; //@line 502 "libraries/Box2D/Dynamics/b2Body.cpp"
 if (!($tmp)) {
  return; //@line 519 "libraries/Box2D/Dynamics/b2Body.cpp"
 }
 $5 = $2 & 65519; //@line 513 "libraries/Box2D/Dynamics/b2Body.cpp"
 $6 = $2 | 16; //@line 509 "libraries/Box2D/Dynamics/b2Body.cpp"
 $storemerge$in = $flag ? $6 : $5; //@line 507 "libraries/Box2D/Dynamics/b2Body.cpp"
 $storemerge = $storemerge$in&65535; //@line 513 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAP16[$0>>1] = $storemerge; //@line 509 "libraries/Box2D/Dynamics/b2Body.cpp"
 $7 = ((($this)) + 72|0); //@line 516 "libraries/Box2D/Dynamics/b2Body.cpp"
 HEAPF32[$7>>2] = 0.0; //@line 516 "libraries/Box2D/Dynamics/b2Body.cpp"
 __ZN6b2Body13ResetMassDataEv($this); //@line 518 "libraries/Box2D/Dynamics/b2Body.cpp"
 return; //@line 519 "libraries/Box2D/Dynamics/b2Body.cpp"
}
function __ZN17b2ContactListenerD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 84 "./libraries/Box2D/Dynamics/b2WorldCallbacks.h"
}
function __ZN17b2ContactListenerD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 84 "./libraries/Box2D/Dynamics/b2WorldCallbacks.h"
 return; //@line 84 "./libraries/Box2D/Dynamics/b2WorldCallbacks.h"
}
function __ZN17b2ContactListener12BeginContactEP9b2Contact($this,$contact) {
 $this = $this|0;
 $contact = $contact|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 87 "./libraries/Box2D/Dynamics/b2WorldCallbacks.h"
}
function __ZN17b2ContactListener10EndContactEP9b2Contact($this,$contact) {
 $this = $this|0;
 $contact = $contact|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 90 "./libraries/Box2D/Dynamics/b2WorldCallbacks.h"
}
function __ZN17b2ContactListener8PreSolveEP9b2ContactPK10b2Manifold($this,$contact,$oldManifold) {
 $this = $this|0;
 $contact = $contact|0;
 $oldManifold = $oldManifold|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 106 "./libraries/Box2D/Dynamics/b2WorldCallbacks.h"
}
function __ZN17b2ContactListener9PostSolveEP9b2ContactPK16b2ContactImpulse($this,$contact,$impulse) {
 $this = $this|0;
 $contact = $contact|0;
 $impulse = $impulse|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 118 "./libraries/Box2D/Dynamics/b2WorldCallbacks.h"
}
function __ZN16b2ContactManagerC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN12b2BroadPhaseC2Ev($this); //@line 28 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $0 = ((($this)) + 60|0); //@line 30 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 HEAP32[$0>>2] = 0; //@line 30 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $1 = ((($this)) + 64|0); //@line 31 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 HEAP32[$1>>2] = 0; //@line 31 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $2 = ((($this)) + 68|0); //@line 32 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 HEAP32[$2>>2] = 33116; //@line 32 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $3 = ((($this)) + 72|0); //@line 33 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 HEAP32[$3>>2] = 33120; //@line 33 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $4 = ((($this)) + 76|0); //@line 34 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 HEAP32[$4>>2] = 0; //@line 34 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 return; //@line 35 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
}
function __ZN16b2ContactManager7DestroyEP9b2Contact($this,$c) {
 $this = $this|0;
 $c = $c|0;
 var $$pre = 0, $$pre3 = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($c)) + 48|0); //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $1 = HEAP32[$0>>2]|0; //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $2 = ((($c)) + 52|0); //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $3 = HEAP32[$2>>2]|0; //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $4 = ((($1)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $5 = HEAP32[$4>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $6 = ((($3)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $7 = HEAP32[$6>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $8 = ((($this)) + 72|0); //@line 44 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 44 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $10 = ($9|0)==(0|0); //@line 44 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if (!($10)) {
  $11 = ((($c)) + 4|0); //@line 261 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
  $12 = HEAP32[$11>>2]|0; //@line 261 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
  $13 = $12 & 2; //@line 261 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
  $14 = ($13|0)==(0); //@line 261 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
  if (!($14)) {
   $15 = HEAP32[$9>>2]|0; //@line 46 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
   $16 = ((($15)) + 12|0); //@line 46 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
   $17 = HEAP32[$16>>2]|0; //@line 46 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
   FUNCTION_TABLE_vii[$17 & 8191]($9,$c); //@line 46 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  }
 }
 $18 = ((($c)) + 8|0); //@line 50 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 50 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $20 = ($19|0)==(0|0); //@line 50 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $21 = $19; //@line 57 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $$pre = ((($c)) + 12|0); //@line 55 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if (!($20)) {
  $22 = HEAP32[$$pre>>2]|0; //@line 52 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  $23 = ((($19)) + 12|0); //@line 52 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  HEAP32[$23>>2] = $22; //@line 52 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 }
 $24 = HEAP32[$$pre>>2]|0; //@line 55 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $25 = ($24|0)==(0|0); //@line 55 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $26 = $24; //@line 62 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if (!($25)) {
  $27 = ((($24)) + 8|0); //@line 57 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  HEAP32[$27>>2] = $21; //@line 57 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 }
 $28 = ((($this)) + 60|0); //@line 60 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 60 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $30 = ($29|0)==($c|0); //@line 60 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if ($30) {
  HEAP32[$28>>2] = $26; //@line 62 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 }
 $31 = ((($c)) + 16|0); //@line 66 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $32 = ((($c)) + 24|0); //@line 66 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $33 = HEAP32[$32>>2]|0; //@line 66 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $34 = ($33|0)==(0|0); //@line 66 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $35 = $33; //@line 73 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $$pre3 = ((($c)) + 28|0); //@line 71 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if (!($34)) {
  $36 = HEAP32[$$pre3>>2]|0; //@line 68 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  $37 = ((($33)) + 12|0); //@line 68 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  HEAP32[$37>>2] = $36; //@line 68 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 }
 $38 = HEAP32[$$pre3>>2]|0; //@line 71 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $39 = ($38|0)==(0|0); //@line 71 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $40 = $38; //@line 78 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if (!($39)) {
  $41 = ((($38)) + 8|0); //@line 73 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  HEAP32[$41>>2] = $35; //@line 73 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 }
 $42 = ((($5)) + 112|0); //@line 76 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $43 = HEAP32[$42>>2]|0; //@line 76 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $44 = ($31|0)==($43|0); //@line 76 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if ($44) {
  HEAP32[$42>>2] = $40; //@line 78 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 }
 $45 = ((($c)) + 32|0); //@line 82 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $46 = ((($c)) + 40|0); //@line 82 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $47 = HEAP32[$46>>2]|0; //@line 82 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $48 = ($47|0)==(0|0); //@line 82 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $49 = $47; //@line 89 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $$pre5 = ((($c)) + 44|0); //@line 87 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if (!($48)) {
  $50 = HEAP32[$$pre5>>2]|0; //@line 84 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  $51 = ((($47)) + 12|0); //@line 84 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  HEAP32[$51>>2] = $50; //@line 84 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 }
 $52 = HEAP32[$$pre5>>2]|0; //@line 87 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $53 = ($52|0)==(0|0); //@line 87 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $54 = $52; //@line 94 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if (!($53)) {
  $55 = ((($52)) + 8|0); //@line 89 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  HEAP32[$55>>2] = $49; //@line 89 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 }
 $56 = ((($7)) + 112|0); //@line 92 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $57 = HEAP32[$56>>2]|0; //@line 92 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $58 = ($45|0)==($57|0); //@line 92 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if (!($58)) {
  $59 = ((($this)) + 76|0); //@line 98 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  $60 = HEAP32[$59>>2]|0; //@line 98 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  __ZN9b2Contact7DestroyEPS_P16b2BlockAllocator($c,$60); //@line 98 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  $61 = ((($this)) + 64|0); //@line 99 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  $62 = HEAP32[$61>>2]|0; //@line 99 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  $63 = (($62) + -1)|0; //@line 99 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  HEAP32[$61>>2] = $63; //@line 99 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  return; //@line 100 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 }
 HEAP32[$56>>2] = $54; //@line 94 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $59 = ((($this)) + 76|0); //@line 98 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $60 = HEAP32[$59>>2]|0; //@line 98 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 __ZN9b2Contact7DestroyEPS_P16b2BlockAllocator($c,$60); //@line 98 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $61 = ((($this)) + 64|0); //@line 99 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $62 = HEAP32[$61>>2]|0; //@line 99 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $63 = (($62) + -1)|0; //@line 99 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 HEAP32[$61>>2] = $63; //@line 99 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 return; //@line 100 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
}
function __ZN16b2ContactManager7CollideEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0, $80 = 0.0;
 var $81 = 0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0.0, $86 = 0, $87 = 0.0, $88 = 0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $brmerge = 0, $c$0$be = 0, $c$01 = 0, $or$cond$i$i = 0, $or$cond10$i$i = 0, $or$cond8$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 60|0); //@line 108 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 108 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $2 = ($1|0)==(0|0); //@line 109 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if ($2) {
  return; //@line 170 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 }
 $3 = ((($this)) + 12|0); //@line 164 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $4 = ((($this)) + 4|0); //@line 165 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $5 = ((($this)) + 72|0); //@line 167 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $6 = ((($this)) + 68|0); //@line 131 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $c$01 = $1;
 while(1) {
  $7 = ((($c$01)) + 48|0); //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
  $8 = HEAP32[$7>>2]|0; //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
  $9 = ((($c$01)) + 52|0); //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
  $10 = HEAP32[$9>>2]|0; //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
  $11 = ((($c$01)) + 56|0); //@line 291 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
  $12 = HEAP32[$11>>2]|0; //@line 291 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
  $13 = ((($c$01)) + 60|0); //@line 301 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
  $14 = HEAP32[$13>>2]|0; //@line 301 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
  $15 = ((($8)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $16 = HEAP32[$15>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $17 = ((($10)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $18 = HEAP32[$17>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $19 = ((($c$01)) + 4|0); //@line 119 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  $20 = HEAP32[$19>>2]|0; //@line 119 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  $21 = $20 & 8; //@line 119 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  $22 = ($21|0)==(0); //@line 119 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  do {
   if ($22) {
    label = 10;
   } else {
    $23 = (__ZNK6b2Body13ShouldCollideEPKS_($18,$16)|0); //@line 122 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    if (!($23)) {
     $24 = ((($c$01)) + 12|0); //@line 266 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     $25 = HEAP32[$24>>2]|0; //@line 266 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     __ZN16b2ContactManager7DestroyEP9b2Contact($this,$c$01); //@line 126 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $c$0$be = $25;
     break;
    }
    $26 = HEAP32[$6>>2]|0; //@line 131 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    $27 = ($26|0)==(0|0); //@line 131 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    if (!($27)) {
     $28 = HEAP32[$26>>2]|0; //@line 131 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $29 = ((($28)) + 8|0); //@line 131 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $30 = HEAP32[$29>>2]|0; //@line 131 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $31 = (FUNCTION_TABLE_iiii[$30 & 8191]($26,$8,$10)|0); //@line 131 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     if (!($31)) {
      $32 = ((($c$01)) + 12|0); //@line 266 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
      $33 = HEAP32[$32>>2]|0; //@line 266 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
      __ZN16b2ContactManager7DestroyEP9b2Contact($this,$c$01); //@line 135 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
      $c$0$be = $33;
      break;
     }
    }
    $34 = HEAP32[$19>>2]|0; //@line 140 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    $35 = $34 & -9; //@line 140 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    HEAP32[$19>>2] = $35; //@line 140 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    label = 10;
   }
  } while(0);
  do {
   if ((label|0) == 10) {
    label = 0;
    $36 = ((($16)) + 4|0); //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
    $37 = HEAP16[$36>>1]|0; //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
    $38 = $37 & 2; //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
    $39 = ($38<<16>>16)==(0); //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
    if ($39) {
     $48 = 0;
    } else {
     $40 = HEAP32[$16>>2]|0; //@line 143 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $41 = ($40|0)!=(0); //@line 143 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $48 = $41;
    }
    $42 = ((($18)) + 4|0); //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
    $43 = HEAP16[$42>>1]|0; //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
    $44 = $43 & 2; //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
    $45 = ($44<<16>>16)==(0); //@line 656 "./libraries/Box2D/Dynamics/b2Body.h"
    if ($45) {
     $49 = 0;
    } else {
     $46 = HEAP32[$18>>2]|0; //@line 144 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $47 = ($46|0)!=(0); //@line 144 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $49 = $47;
    }
    $brmerge = $48 | $49; //@line 147 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    if (!($brmerge)) {
     $50 = ((($c$01)) + 12|0); //@line 266 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     $51 = HEAP32[$50>>2]|0; //@line 266 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     $c$0$be = $51;
     break;
    }
    $53 = ((($8)) + 24|0); //@line 153 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    $54 = HEAP32[$53>>2]|0; //@line 153 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    $55 = (((($54) + (($12*28)|0)|0)) + 24|0); //@line 153 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    $56 = HEAP32[$55>>2]|0; //@line 153 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    $57 = ((($10)) + 24|0); //@line 154 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    $58 = HEAP32[$57>>2]|0; //@line 154 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    $59 = (((($58) + (($14*28)|0)|0)) + 24|0); //@line 154 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    $60 = HEAP32[$59>>2]|0; //@line 154 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    $61 = ($56|0)>(-1); //@line 164 "./libraries/Box2D/Collision/b2DynamicTree.h"
    if ($61) {
     $62 = HEAP32[$3>>2]|0; //@line 164 "./libraries/Box2D/Collision/b2DynamicTree.h"
     $63 = ($62|0)>($56|0); //@line 164 "./libraries/Box2D/Collision/b2DynamicTree.h"
     $64 = $63;
    } else {
     $64 = 0;
    }
    __Z10loveAssertbPKc($64,446451); //@line 164 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $65 = HEAP32[$4>>2]|0; //@line 165 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $66 = ($60|0)>(-1); //@line 164 "./libraries/Box2D/Collision/b2DynamicTree.h"
    if ($66) {
     $67 = HEAP32[$3>>2]|0; //@line 164 "./libraries/Box2D/Collision/b2DynamicTree.h"
     $68 = ($67|0)>($60|0); //@line 164 "./libraries/Box2D/Collision/b2DynamicTree.h"
     $69 = $68;
    } else {
     $69 = 0;
    }
    __Z10loveAssertbPKc($69,446451); //@line 164 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $70 = HEAP32[$4>>2]|0; //@line 165 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $71 = (($70) + (($60*36)|0)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $72 = +HEAPF32[$71>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $73 = (((($65) + (($56*36)|0)|0)) + 8|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $74 = +HEAPF32[$73>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $75 = $72 - $74; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $76 = (((($70) + (($60*36)|0)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $77 = +HEAPF32[$76>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $78 = (((($65) + (($56*36)|0)|0)) + 12|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $79 = +HEAPF32[$78>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $80 = $77 - $79; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $81 = (($65) + (($56*36)|0)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $82 = +HEAPF32[$81>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $83 = (((($70) + (($60*36)|0)|0)) + 8|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $84 = +HEAPF32[$83>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $85 = $82 - $84; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $86 = (((($65) + (($56*36)|0)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $87 = +HEAPF32[$86>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $88 = (((($70) + (($60*36)|0)|0)) + 12|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $89 = +HEAPF32[$88>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $90 = $87 - $89; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $91 = $75 > 0.0; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
    $92 = $80 > 0.0; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
    $or$cond$i$i = $91 | $92; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
    $93 = $85 > 0.0; //@line 271 "./libraries/Box2D/Collision/b2Collision.h"
    $or$cond8$i$i = $or$cond$i$i | $93; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
    $94 = $90 > 0.0; //@line 271 "./libraries/Box2D/Collision/b2Collision.h"
    $or$cond10$i$i = $or$cond8$i$i | $94; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
    if ($or$cond10$i$i) {
     $95 = ((($c$01)) + 12|0); //@line 266 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     $96 = HEAP32[$95>>2]|0; //@line 266 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     __ZN16b2ContactManager7DestroyEP9b2Contact($this,$c$01); //@line 162 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $c$0$be = $96;
     break;
    } else {
     $97 = HEAP32[$5>>2]|0; //@line 167 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     __ZN9b2Contact6UpdateEP17b2ContactListener($c$01,$97); //@line 167 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $98 = ((($c$01)) + 12|0); //@line 266 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     $99 = HEAP32[$98>>2]|0; //@line 266 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     $c$0$be = $99;
     break;
    }
   }
  } while(0);
  $52 = ($c$0$be|0)==(0|0); //@line 109 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  if ($52) {
   break;
  } else {
   $c$01 = $c$0$be;
  }
 }
 return; //@line 170 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
}
function __ZN16b2ContactManager15FindNewContactsEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN12b2BroadPhase11UpdatePairsI16b2ContactManagerEEvPT_($this,$this); //@line 174 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 return; //@line 175 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
}
function __ZN12b2BroadPhase11UpdatePairsI16b2ContactManagerEEvPT_($this,$callback) {
 $this = $this|0;
 $callback = $callback|0;
 var $$i1$1 = 0, $$pre = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, $i$03 = 0, $i1$0$ph = 0, $i1$02 = 0, $i1$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = ((($this)) + 52|0); //@line 188 "./libraries/Box2D/Collision/b2BroadPhase.h"
 HEAP32[$1>>2] = 0; //@line 188 "./libraries/Box2D/Collision/b2BroadPhase.h"
 $2 = ((($this)) + 40|0); //@line 191 "./libraries/Box2D/Collision/b2BroadPhase.h"
 $3 = HEAP32[$2>>2]|0; //@line 191 "./libraries/Box2D/Collision/b2BroadPhase.h"
 $4 = ($3|0)>(0); //@line 191 "./libraries/Box2D/Collision/b2BroadPhase.h"
 if ($4) {
  $5 = ((($this)) + 32|0); //@line 193 "./libraries/Box2D/Collision/b2BroadPhase.h"
  $6 = ((($this)) + 56|0); //@line 193 "./libraries/Box2D/Collision/b2BroadPhase.h"
  $7 = ((($this)) + 12|0); //@line 164 "./libraries/Box2D/Collision/b2DynamicTree.h"
  $8 = ((($this)) + 4|0); //@line 165 "./libraries/Box2D/Collision/b2DynamicTree.h"
  $63 = $3;$i$03 = 0;
  while(1) {
   $17 = HEAP32[$5>>2]|0; //@line 193 "./libraries/Box2D/Collision/b2BroadPhase.h"
   $18 = (($17) + ($i$03<<2)|0); //@line 193 "./libraries/Box2D/Collision/b2BroadPhase.h"
   $19 = HEAP32[$18>>2]|0; //@line 193 "./libraries/Box2D/Collision/b2BroadPhase.h"
   HEAP32[$6>>2] = $19; //@line 193 "./libraries/Box2D/Collision/b2BroadPhase.h"
   $20 = ($19|0)==(-1); //@line 194 "./libraries/Box2D/Collision/b2BroadPhase.h"
   if ($20) {
    $29 = $63;
   } else {
    $21 = ($19|0)>(-1); //@line 164 "./libraries/Box2D/Collision/b2DynamicTree.h"
    if ($21) {
     $22 = HEAP32[$7>>2]|0; //@line 164 "./libraries/Box2D/Collision/b2DynamicTree.h"
     $23 = ($22|0)>($19|0); //@line 164 "./libraries/Box2D/Collision/b2DynamicTree.h"
     $24 = $23;
    } else {
     $24 = 0;
    }
    __Z10loveAssertbPKc($24,446451); //@line 164 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $25 = HEAP32[$8>>2]|0; //@line 165 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $26 = (($25) + (($19*36)|0)|0); //@line 165 "./libraries/Box2D/Collision/b2DynamicTree.h"
    __ZNK13b2DynamicTree5QueryI12b2BroadPhaseEEvPT_RK6b2AABB($this,$this,$26); //@line 204 "./libraries/Box2D/Collision/b2BroadPhase.h"
    $$pre = HEAP32[$2>>2]|0; //@line 191 "./libraries/Box2D/Collision/b2BroadPhase.h"
    $29 = $$pre;
   }
   $27 = (($i$03) + 1)|0; //@line 191 "./libraries/Box2D/Collision/b2BroadPhase.h"
   $28 = ($27|0)<($29|0); //@line 191 "./libraries/Box2D/Collision/b2BroadPhase.h"
   if ($28) {
    $63 = $29;$i$03 = $27;
   } else {
    break;
   }
  }
  $$pre5 = HEAP32[$1>>2]|0; //@line 211 "./libraries/Box2D/Collision/b2BroadPhase.h"
  $12 = $$pre5;
 } else {
  $12 = 0;
 }
 HEAP32[$2>>2] = 0; //@line 208 "./libraries/Box2D/Collision/b2BroadPhase.h"
 $9 = ((($this)) + 44|0); //@line 211 "./libraries/Box2D/Collision/b2BroadPhase.h"
 $10 = HEAP32[$9>>2]|0; //@line 211 "./libraries/Box2D/Collision/b2BroadPhase.h"
 $11 = (($10) + ($12<<3)|0); //@line 211 "./libraries/Box2D/Collision/b2BroadPhase.h"
 HEAP32[$0>>2] = 4238; //@line 211 "./libraries/Box2D/Collision/b2BroadPhase.h"
 __ZNSt3__16__sortIRPFbRK6b2PairS3_EPS1_EEvT0_S8_T_($10,$11,$0); //@line 4037 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $13 = HEAP32[$1>>2]|0; //@line 215 "./libraries/Box2D/Collision/b2BroadPhase.h"
 $14 = ($13|0)>(0); //@line 215 "./libraries/Box2D/Collision/b2BroadPhase.h"
 if (!($14)) {
  STACKTOP = sp;return; //@line 238 "./libraries/Box2D/Collision/b2BroadPhase.h"
 }
 $15 = ((($this)) + 12|0); //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $16 = ((($this)) + 4|0); //@line 159 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $i1$02 = 0;
 while(1) {
  $32 = HEAP32[$9>>2]|0; //@line 217 "./libraries/Box2D/Collision/b2BroadPhase.h"
  $33 = (($32) + ($i1$02<<3)|0); //@line 218 "./libraries/Box2D/Collision/b2BroadPhase.h"
  $34 = HEAP32[$33>>2]|0; //@line 218 "./libraries/Box2D/Collision/b2BroadPhase.h"
  $35 = ($34|0)>(-1); //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
  if ($35) {
   $36 = HEAP32[$15>>2]|0; //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
   $37 = ($36|0)>($34|0); //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
   $38 = $37;
  } else {
   $38 = 0;
  }
  __Z10loveAssertbPKc($38,446451); //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
  $39 = HEAP32[$16>>2]|0; //@line 159 "./libraries/Box2D/Collision/b2DynamicTree.h"
  $40 = (((($39) + (($34*36)|0)|0)) + 16|0); //@line 159 "./libraries/Box2D/Collision/b2DynamicTree.h"
  $41 = HEAP32[$40>>2]|0; //@line 159 "./libraries/Box2D/Collision/b2DynamicTree.h"
  $42 = (((($32) + ($i1$02<<3)|0)) + 4|0); //@line 219 "./libraries/Box2D/Collision/b2BroadPhase.h"
  $43 = HEAP32[$42>>2]|0; //@line 219 "./libraries/Box2D/Collision/b2BroadPhase.h"
  $44 = ($43|0)>(-1); //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
  if ($44) {
   $45 = HEAP32[$15>>2]|0; //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
   $46 = ($45|0)>($43|0); //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
   $47 = $46;
  } else {
   $47 = 0;
  }
  __Z10loveAssertbPKc($47,446451); //@line 158 "./libraries/Box2D/Collision/b2DynamicTree.h"
  $48 = HEAP32[$16>>2]|0; //@line 159 "./libraries/Box2D/Collision/b2DynamicTree.h"
  $49 = (((($48) + (($43*36)|0)|0)) + 16|0); //@line 159 "./libraries/Box2D/Collision/b2DynamicTree.h"
  $50 = HEAP32[$49>>2]|0; //@line 159 "./libraries/Box2D/Collision/b2DynamicTree.h"
  __ZN16b2ContactManager7AddPairEPvS0_($callback,$41,$50); //@line 221 "./libraries/Box2D/Collision/b2BroadPhase.h"
  $51 = (($i1$02) + 1)|0; //@line 222 "./libraries/Box2D/Collision/b2BroadPhase.h"
  $31 = HEAP32[$1>>2]|0; //@line 215 "./libraries/Box2D/Collision/b2BroadPhase.h"
  $i1$1 = $51;
  while(1) {
   $52 = ($i1$1|0)<($31|0); //@line 225 "./libraries/Box2D/Collision/b2BroadPhase.h"
   if (!($52)) {
    $i1$0$ph = $i1$1;
    break;
   }
   $53 = HEAP32[$9>>2]|0; //@line 227 "./libraries/Box2D/Collision/b2BroadPhase.h"
   $54 = (($53) + ($i1$1<<3)|0); //@line 228 "./libraries/Box2D/Collision/b2BroadPhase.h"
   $55 = HEAP32[$54>>2]|0; //@line 228 "./libraries/Box2D/Collision/b2BroadPhase.h"
   $56 = HEAP32[$33>>2]|0; //@line 228 "./libraries/Box2D/Collision/b2BroadPhase.h"
   $57 = ($55|0)==($56|0); //@line 228 "./libraries/Box2D/Collision/b2BroadPhase.h"
   if (!($57)) {
    $i1$0$ph = $i1$1;
    break;
   }
   $58 = (((($53) + ($i1$1<<3)|0)) + 4|0); //@line 228 "./libraries/Box2D/Collision/b2BroadPhase.h"
   $59 = HEAP32[$58>>2]|0; //@line 228 "./libraries/Box2D/Collision/b2BroadPhase.h"
   $60 = HEAP32[$42>>2]|0; //@line 228 "./libraries/Box2D/Collision/b2BroadPhase.h"
   $61 = ($59|0)==($60|0); //@line 228 "./libraries/Box2D/Collision/b2BroadPhase.h"
   $62 = $61&1; //@line 228 "./libraries/Box2D/Collision/b2BroadPhase.h"
   $$i1$1 = (($62) + ($i1$1))|0; //@line 228 "./libraries/Box2D/Collision/b2BroadPhase.h"
   if ($61) {
    $i1$1 = $$i1$1;
   } else {
    $i1$0$ph = $$i1$1;
    break;
   }
  }
  $30 = ($i1$0$ph|0)<($31|0); //@line 215 "./libraries/Box2D/Collision/b2BroadPhase.h"
  if ($30) {
   $i1$02 = $i1$0$ph;
  } else {
   break;
  }
 }
 STACKTOP = sp;return; //@line 238 "./libraries/Box2D/Collision/b2BroadPhase.h"
}
function __Z14b2PairLessThanRK6b2PairS1_($pair1,$pair2) {
 $pair1 = $pair1|0;
 $pair2 = $pair2|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$pair1>>2]|0; //@line 134 "./libraries/Box2D/Collision/b2BroadPhase.h"
 $1 = HEAP32[$pair2>>2]|0; //@line 134 "./libraries/Box2D/Collision/b2BroadPhase.h"
 $2 = ($0|0)<($1|0); //@line 134 "./libraries/Box2D/Collision/b2BroadPhase.h"
 if ($2) {
  $$0 = 1;
  return ($$0|0); //@line 145 "./libraries/Box2D/Collision/b2BroadPhase.h"
 }
 $3 = ($0|0)==($1|0); //@line 139 "./libraries/Box2D/Collision/b2BroadPhase.h"
 if (!($3)) {
  $$0 = 0;
  return ($$0|0); //@line 145 "./libraries/Box2D/Collision/b2BroadPhase.h"
 }
 $4 = ((($pair1)) + 4|0); //@line 141 "./libraries/Box2D/Collision/b2BroadPhase.h"
 $5 = HEAP32[$4>>2]|0; //@line 141 "./libraries/Box2D/Collision/b2BroadPhase.h"
 $6 = ((($pair2)) + 4|0); //@line 141 "./libraries/Box2D/Collision/b2BroadPhase.h"
 $7 = HEAP32[$6>>2]|0; //@line 141 "./libraries/Box2D/Collision/b2BroadPhase.h"
 $8 = ($5|0)<($7|0); //@line 141 "./libraries/Box2D/Collision/b2BroadPhase.h"
 $$0 = $8;
 return ($$0|0); //@line 145 "./libraries/Box2D/Collision/b2BroadPhase.h"
}
function __ZNSt3__16__sortIRPFbRK6b2PairS3_EPS1_EEvT0_S8_T_($__first,$__last,$__comp) {
 $__first = $__first|0;
 $__last = $__last|0;
 $__comp = $__comp|0;
 var $$0 = 0, $$0$lcssa204 = 0, $$0$lcssa205 = 0, $$0$lcssa206 = 0, $$0$lcssa207 = 0, $$0$lcssa208 = 0, $$0$lcssa211 = 0, $$0$ph = 0, $$04$ph = 0, $$04$ph$lcssa220 = 0, $$__m$0 = 0, $$lcssa = 0, $$lcssa199 = 0, $$lcssa201 = 0, $$lcssa202 = 0, $$lcssa203 = 0, $$lcssa224 = 0, $$lcssa225 = 0, $$lcssa226 = 0, $$lcssa227 = 0;
 var $$lcssa232 = 0, $$lcssa233 = 0, $$lcssa235 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0;
 var $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0;
 var $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0;
 var $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0;
 var $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0;
 var $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0;
 var $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0;
 var $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0;
 var $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0;
 var $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0;
 var $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0;
 var $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0;
 var $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0;
 var $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0;
 var $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0;
 var $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0;
 var $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0;
 var $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0;
 var $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__i$049 = 0, $__i$049$lcssa = 0, $__i$1 = 0;
 var $__i$2 = 0, $__i$2$lcssa = 0, $__i$2$ph = 0, $__i$3 = 0, $__i$3$lcssa = 0, $__i$3$ph = 0, $__i$4 = 0, $__i$4$lcssa213 = 0, $__j$0 = 0, $__j$1$ph = 0, $__j$2 = 0, $__j$3 = 0, $__j$4$ph = 0, $__j$5 = 0, $__m$0$ph = 0, $__m$2 = 0, $__n_swaps$0 = 0, $__n_swaps$3 = 0, $__n_swaps$4$ph = 0, $__n_swaps$5 = 0;
 var $__n_swaps$6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0$ph = $__first;$$04$ph = $__last;
 L1: while(1) {
  $0 = $$04$ph; //@line 3844 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $1 = ((($$04$ph)) + -8|0); //@line 3872 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $$0 = $$0$ph;
  L3: while(1) {
   $2 = $$0; //@line 3844 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $3 = (($0) - ($2))|0; //@line 3844 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $4 = $3 >> 3; //@line 3844 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   switch ($4|0) {
   case 2:  {
    $$0$lcssa204 = $$0;$$lcssa224 = $1;$$lcssa232 = $1;
    label = 4;
    break L1;
    break;
   }
   case 3:  {
    $$0$lcssa205 = $$0;$$lcssa225 = $1;$$lcssa233 = $1;
    label = 6;
    break L1;
    break;
   }
   case 4:  {
    $$0$lcssa206 = $$0;$$lcssa226 = $1;
    label = 14;
    break L1;
    break;
   }
   case 5:  {
    $$0$lcssa207 = $$0;$$lcssa227 = $1;$$lcssa235 = $1;
    label = 15;
    break L1;
    break;
   }
   case 1: case 0:  {
    label = 69;
    break L1;
    break;
   }
   default: {
   }
   }
   $223 = ($3|0)<(248); //@line 3864 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   if ($223) {
    $$0$lcssa208 = $$0;$$04$ph$lcssa220 = $$04$ph;
    label = 21;
    break L1;
   }
   $224 = ($3|0)>(7992); //@line 3876 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $225 = (($4|0) / 2)&-1; //@line 3878 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $226 = (($$0) + ($225<<3)|0); //@line 3879 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   do {
    if ($224) {
     $227 = (($4|0) / 4)&-1; //@line 3880 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $228 = (($$0) + ($227<<3)|0); //@line 3881 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $229 = (($226) + ($227<<3)|0); //@line 3881 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $230 = (__ZNSt3__17__sort4IRPFbRK6b2PairS3_EPS1_EEjT0_S8_S8_S8_T_($$0,$228,$226,$229,$__comp)|0); //@line 3664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $231 = HEAP32[$__comp>>2]|0; //@line 3665 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $232 = (FUNCTION_TABLE_iii[$231 & 8191]($1,$229)|0); //@line 3665 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     if ($232) {
      $233 = $229; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $234 = $233; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $235 = HEAP32[$234>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $236 = (($233) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $237 = $236; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $238 = HEAP32[$237>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $239 = $1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $240 = $239; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $241 = HEAP32[$240>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $242 = (($239) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $243 = $242; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $244 = HEAP32[$243>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $245 = $229; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $246 = $245; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$246>>2] = $241; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $247 = (($245) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $248 = $247; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$248>>2] = $244; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $249 = $1; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $250 = $249; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$250>>2] = $235; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $251 = (($249) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $252 = $251; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$252>>2] = $238; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $253 = (($230) + 1)|0; //@line 3668 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $254 = HEAP32[$__comp>>2]|0; //@line 3669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $255 = (FUNCTION_TABLE_iii[$254 & 8191]($229,$226)|0); //@line 3669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if ($255) {
       $256 = $226; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $257 = $256; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $258 = HEAP32[$257>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $259 = (($256) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $260 = $259; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $261 = HEAP32[$260>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $262 = $229; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $263 = $262; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $264 = HEAP32[$263>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $265 = (($262) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $266 = $265; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $267 = HEAP32[$266>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $268 = $226; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $269 = $268; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       HEAP32[$269>>2] = $264; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $270 = (($268) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $271 = $270; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       HEAP32[$271>>2] = $267; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $272 = $229; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $273 = $272; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       HEAP32[$273>>2] = $258; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $274 = (($272) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $275 = $274; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       HEAP32[$275>>2] = $261; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
       $276 = (($230) + 2)|0; //@line 3672 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $277 = HEAP32[$__comp>>2]|0; //@line 3673 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $278 = (FUNCTION_TABLE_iii[$277 & 8191]($226,$228)|0); //@line 3673 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       if ($278) {
        $279 = $228; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $280 = $279; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $281 = HEAP32[$280>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $282 = (($279) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $283 = $282; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $284 = HEAP32[$283>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $285 = $226; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $286 = $285; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $287 = HEAP32[$286>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $288 = (($285) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $289 = $288; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $290 = HEAP32[$289>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $291 = $228; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $292 = $291; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        HEAP32[$292>>2] = $287; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $293 = (($291) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $294 = $293; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        HEAP32[$294>>2] = $290; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $295 = $226; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $296 = $295; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        HEAP32[$296>>2] = $281; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $297 = (($295) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $298 = $297; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        HEAP32[$298>>2] = $284; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
        $299 = (($230) + 3)|0; //@line 3676 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $300 = HEAP32[$__comp>>2]|0; //@line 3677 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        $301 = (FUNCTION_TABLE_iii[$300 & 8191]($228,$$0)|0); //@line 3677 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
        if ($301) {
         $302 = $$0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $303 = $302; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $304 = HEAP32[$303>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $305 = (($302) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $306 = $305; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $307 = HEAP32[$306>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $308 = $228; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $309 = $308; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $310 = HEAP32[$309>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $311 = (($308) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $312 = $311; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $313 = HEAP32[$312>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $314 = $$0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $315 = $314; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         HEAP32[$315>>2] = $310; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $316 = (($314) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $317 = $316; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         HEAP32[$317>>2] = $313; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $318 = $228; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $319 = $318; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         HEAP32[$319>>2] = $304; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $320 = (($318) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $321 = $320; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         HEAP32[$321>>2] = $307; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
         $322 = (($230) + 4)|0; //@line 3680 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
         $__n_swaps$0 = $322;
        } else {
         $__n_swaps$0 = $299;
        }
       } else {
        $__n_swaps$0 = $276;
       }
      } else {
       $__n_swaps$0 = $253;
      }
     } else {
      $__n_swaps$0 = $230;
     }
    } else {
     $323 = HEAP32[$__comp>>2]|0; //@line 3601 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $324 = (FUNCTION_TABLE_iii[$323 & 8191]($226,$$0)|0); //@line 3601 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $325 = HEAP32[$__comp>>2]|0; //@line 3615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $326 = (FUNCTION_TABLE_iii[$325 & 8191]($1,$226)|0); //@line 3615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     if (!($324)) {
      if (!($326)) {
       $__n_swaps$0 = 0;
       break;
      }
      $327 = $226; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $328 = $327; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $329 = HEAP32[$328>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $330 = (($327) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $331 = $330; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $332 = HEAP32[$331>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $333 = $1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $334 = $333; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $335 = HEAP32[$334>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $336 = (($333) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $337 = $336; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $338 = HEAP32[$337>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $339 = $226; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $340 = $339; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$340>>2] = $335; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $341 = (($339) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $342 = $341; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$342>>2] = $338; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $343 = $1; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $344 = $343; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$344>>2] = $329; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $345 = (($343) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $346 = $345; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$346>>2] = $332; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $347 = HEAP32[$__comp>>2]|0; //@line 3608 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $348 = (FUNCTION_TABLE_iii[$347 & 8191]($226,$$0)|0); //@line 3608 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if (!($348)) {
       $__n_swaps$0 = 1;
       break;
      }
      $349 = $$0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $350 = $349; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $351 = HEAP32[$350>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $352 = (($349) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $353 = $352; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $354 = HEAP32[$353>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $355 = $226; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $356 = $355; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $357 = HEAP32[$356>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $358 = (($355) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $359 = $358; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $360 = HEAP32[$359>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $361 = $$0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $362 = $361; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$362>>2] = $357; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $363 = (($361) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $364 = $363; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$364>>2] = $360; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $365 = $226; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $366 = $365; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$366>>2] = $351; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $367 = (($365) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $368 = $367; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$368>>2] = $354; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $__n_swaps$0 = 2;
      break;
     }
     $369 = $$0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $370 = $369; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $371 = HEAP32[$370>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $372 = (($369) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $373 = $372; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $374 = HEAP32[$373>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     if ($326) {
      $375 = $1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $376 = $375; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $377 = HEAP32[$376>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $378 = (($375) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $379 = $378; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $380 = HEAP32[$379>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $381 = $$0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $382 = $381; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$382>>2] = $377; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $383 = (($381) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $384 = $383; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$384>>2] = $380; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $385 = $1; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $386 = $385; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$386>>2] = $371; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $387 = (($385) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $388 = $387; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$388>>2] = $374; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $__n_swaps$0 = 1;
      break;
     }
     $389 = $226; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $390 = $389; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $391 = HEAP32[$390>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $392 = (($389) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $393 = $392; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $394 = HEAP32[$393>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $395 = $$0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $396 = $395; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$396>>2] = $391; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $397 = (($395) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $398 = $397; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$398>>2] = $394; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $399 = $226; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $400 = $399; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$400>>2] = $371; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $401 = (($399) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $402 = $401; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$402>>2] = $374; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $403 = HEAP32[$__comp>>2]|0; //@line 3623 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $404 = (FUNCTION_TABLE_iii[$403 & 8191]($1,$226)|0); //@line 3623 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     if ($404) {
      $405 = $226; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $406 = $405; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $407 = HEAP32[$406>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $408 = (($405) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $409 = $408; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $410 = HEAP32[$409>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $411 = $1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $412 = $411; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $413 = HEAP32[$412>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $414 = (($411) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $415 = $414; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $416 = HEAP32[$415>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $417 = $226; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $418 = $417; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$418>>2] = $413; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $419 = (($417) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $420 = $419; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$420>>2] = $416; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $421 = $1; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $422 = $421; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$422>>2] = $407; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $423 = (($421) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $424 = $423; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$424>>2] = $410; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $__n_swaps$0 = 2;
     } else {
      $__n_swaps$0 = 1;
     }
    }
   } while(0);
   $425 = HEAP32[$__comp>>2]|0; //@line 3898 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $426 = (FUNCTION_TABLE_iii[$425 & 8191]($$0,$226)|0); //@line 3898 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   do {
    if ($426) {
     $__j$3 = $1;$__n_swaps$3 = $__n_swaps$0;
    } else {
     $__j$0 = $1;
     while(1) {
      $427 = ((($__j$0)) + -8|0); //@line 3904 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $428 = ($$0|0)==($427|0); //@line 3904 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if ($428) {
       break;
      }
      $486 = HEAP32[$__comp>>2]|0; //@line 3947 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $487 = (FUNCTION_TABLE_iii[$486 & 8191]($427,$226)|0); //@line 3947 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if ($487) {
       $$lcssa199 = $427;
       label = 51;
       break;
      } else {
       $__j$0 = $427;
      }
     }
     if ((label|0) == 51) {
      label = 0;
      $488 = $$0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $489 = $488; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $490 = HEAP32[$489>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $491 = (($488) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $492 = $491; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $493 = HEAP32[$492>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $494 = $$lcssa199; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $495 = $494; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $496 = HEAP32[$495>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $497 = (($494) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $498 = $497; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $499 = HEAP32[$498>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $500 = $$0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $501 = $500; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$501>>2] = $496; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $502 = (($500) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $503 = $502; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$503>>2] = $499; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $504 = $$lcssa199; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $505 = $504; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$505>>2] = $490; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $506 = (($504) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $507 = $506; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$507>>2] = $493; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $508 = (($__n_swaps$0) + 1)|0; //@line 3950 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $__j$3 = $$lcssa199;$__n_swaps$3 = $508;
      break;
     }
     $429 = ((($$0)) + 8|0); //@line 3908 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $430 = HEAP32[$__comp>>2]|0; //@line 3910 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $431 = (FUNCTION_TABLE_iii[$430 & 8191]($$0,$1)|0); //@line 3910 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     if ($431) {
      $__i$1 = $429;
     } else {
      $432 = ($429|0)==($1|0); //@line 3914 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if ($432) {
       label = 69;
       break L1;
      } else {
       $__i$049 = $429;
      }
      while(1) {
       $433 = HEAP32[$__comp>>2]|0; //@line 3916 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $434 = (FUNCTION_TABLE_iii[$433 & 8191]($$0,$__i$049)|0); //@line 3916 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       if ($434) {
        $__i$049$lcssa = $__i$049;
        break;
       }
       $456 = ((($__i$049)) + 8|0); //@line 3923 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $457 = ($456|0)==($1|0); //@line 3914 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       if ($457) {
        label = 69;
        break L1;
       } else {
        $__i$049 = $456;
       }
      }
      $435 = $__i$049$lcssa; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $436 = $435; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $437 = HEAP32[$436>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $438 = (($435) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $439 = $438; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $440 = HEAP32[$439>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $441 = $1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $442 = $441; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $443 = HEAP32[$442>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $444 = (($441) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $445 = $444; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $446 = HEAP32[$445>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $447 = $__i$049$lcssa; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $448 = $447; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$448>>2] = $443; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $449 = (($447) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $450 = $449; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$450>>2] = $446; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $451 = $1; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $452 = $451; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$452>>2] = $437; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $453 = (($451) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $454 = $453; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$454>>2] = $440; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $455 = ((($__i$049$lcssa)) + 8|0); //@line 3920 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $__i$1 = $455;
     }
     $458 = ($__i$1|0)==($1|0); //@line 3927 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     if ($458) {
      label = 69;
      break L1;
     } else {
      $__i$2$ph = $__i$1;$__j$1$ph = $1;
     }
     while(1) {
      $__i$2 = $__i$2$ph;
      while(1) {
       $459 = HEAP32[$__comp>>2]|0; //@line 3931 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $460 = (FUNCTION_TABLE_iii[$459 & 8191]($$0,$__i$2)|0); //@line 3931 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $461 = ((($__i$2)) + 8|0); //@line 3932 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       if ($460) {
        $$lcssa = $461;$__i$2$lcssa = $__i$2;
        break;
       } else {
        $__i$2 = $461;
       }
      }
      $__j$2 = $__j$1$ph;
      while(1) {
       $462 = HEAP32[$__comp>>2]|0; //@line 3933 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $463 = ((($__j$2)) + -8|0); //@line 3933 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $464 = (FUNCTION_TABLE_iii[$462 & 8191]($$0,$463)|0); //@line 3933 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       if ($464) {
        $__j$2 = $463;
       } else {
        $$lcssa201 = $463;
        break;
       }
      }
      $465 = ($__i$2$lcssa>>>0)<($$lcssa201>>>0); //@line 3935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if (!($465)) {
       $$0 = $__i$2$lcssa;
       continue L3;
      }
      $466 = $__i$2$lcssa; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $467 = $466; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $468 = HEAP32[$467>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $469 = (($466) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $470 = $469; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $471 = HEAP32[$470>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $472 = $$lcssa201; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $473 = $472; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $474 = HEAP32[$473>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $475 = (($472) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $476 = $475; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $477 = HEAP32[$476>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $478 = $__i$2$lcssa; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $479 = $478; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$479>>2] = $474; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $480 = (($478) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $481 = $480; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$481>>2] = $477; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $482 = $$lcssa201; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $483 = $482; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$483>>2] = $468; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $484 = (($482) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $485 = $484; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$485>>2] = $471; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $__i$2$ph = $$lcssa;$__j$1$ph = $$lcssa201;
     }
    }
   } while(0);
   $509 = ((($$0)) + 8|0); //@line 3956 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $510 = ($509>>>0)<($__j$3>>>0); //@line 3959 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   L48: do {
    if ($510) {
     $__i$3$ph = $509;$__j$4$ph = $__j$3;$__m$0$ph = $226;$__n_swaps$4$ph = $__n_swaps$3;
     while(1) {
      $__i$3 = $__i$3$ph;
      while(1) {
       $511 = HEAP32[$__comp>>2]|0; //@line 3966 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $512 = (FUNCTION_TABLE_iii[$511 & 8191]($__i$3,$__m$0$ph)|0); //@line 3966 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $513 = ((($__i$3)) + 8|0); //@line 3967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       if ($512) {
        $__i$3 = $513;
       } else {
        $$lcssa202 = $513;$__i$3$lcssa = $__i$3;
        break;
       }
      }
      $__j$5 = $__j$4$ph;
      while(1) {
       $514 = HEAP32[$__comp>>2]|0; //@line 3969 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $515 = ((($__j$5)) + -8|0); //@line 3969 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $516 = (FUNCTION_TABLE_iii[$514 & 8191]($515,$__m$0$ph)|0); //@line 3969 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       if ($516) {
        $$lcssa203 = $515;
        break;
       } else {
        $__j$5 = $515;
       }
      }
      $517 = ($__i$3$lcssa>>>0)>($$lcssa203>>>0); //@line 3971 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if ($517) {
       $__i$4 = $__i$3$lcssa;$__m$2 = $__m$0$ph;$__n_swaps$5 = $__n_swaps$4$ph;
       break L48;
      }
      $518 = $__i$3$lcssa; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $519 = $518; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $520 = HEAP32[$519>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $521 = (($518) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $522 = $521; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $523 = HEAP32[$522>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $524 = $$lcssa203; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $525 = $524; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $526 = HEAP32[$525>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $527 = (($524) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $528 = $527; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $529 = HEAP32[$528>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $530 = $__i$3$lcssa; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $531 = $530; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$531>>2] = $526; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $532 = (($530) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $533 = $532; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$533>>2] = $529; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $534 = $$lcssa203; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $535 = $534; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$535>>2] = $520; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $536 = (($534) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $537 = $536; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$537>>2] = $523; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $538 = (($__n_swaps$4$ph) + 1)|0; //@line 3974 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $539 = ($__m$0$ph|0)==($__i$3$lcssa|0); //@line 3977 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $$__m$0 = $539 ? $$lcssa203 : $__m$0$ph; //@line 3977 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $__i$3$ph = $$lcssa202;$__j$4$ph = $$lcssa203;$__m$0$ph = $$__m$0;$__n_swaps$4$ph = $538;
     }
    } else {
     $__i$4 = $509;$__m$2 = $226;$__n_swaps$5 = $__n_swaps$3;
    }
   } while(0);
   $540 = ($__i$4|0)==($__m$2|0); //@line 3983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   if ($540) {
    $__n_swaps$6 = $__n_swaps$5;
   } else {
    $541 = HEAP32[$__comp>>2]|0; //@line 3983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $542 = (FUNCTION_TABLE_iii[$541 & 8191]($__m$2,$__i$4)|0); //@line 3983 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    if ($542) {
     $543 = $__i$4; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $544 = $543; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $545 = HEAP32[$544>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $546 = (($543) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $547 = $546; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $548 = HEAP32[$547>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $549 = $__m$2; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $550 = $549; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $551 = HEAP32[$550>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $552 = (($549) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $553 = $552; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $554 = HEAP32[$553>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $555 = $__i$4; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $556 = $555; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$556>>2] = $551; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $557 = (($555) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $558 = $557; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$558>>2] = $554; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $559 = $__m$2; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $560 = $559; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$560>>2] = $545; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $561 = (($559) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $562 = $561; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$562>>2] = $548; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $563 = (($__n_swaps$5) + 1)|0; //@line 3986 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $__n_swaps$6 = $563;
    } else {
     $__n_swaps$6 = $__n_swaps$5;
    }
   }
   $564 = ($__n_swaps$6|0)==(0); //@line 3990 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   if ($564) {
    $565 = (__ZNSt3__127__insertion_sort_incompleteIRPFbRK6b2PairS3_EPS1_EEbT0_S8_T_($$0,$__i$4,$__comp)|0); //@line 3992 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $566 = ((($__i$4)) + 8|0); //@line 3993 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $567 = (__ZNSt3__127__insertion_sort_incompleteIRPFbRK6b2PairS3_EPS1_EEbT0_S8_T_($566,$$04$ph,$__comp)|0); //@line 3993 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    if ($567) {
     label = 68;
     break;
    }
    if ($565) {
     $$0 = $566;
     continue;
    }
   }
   $568 = $__i$4; //@line 4010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $569 = (($568) - ($2))|0; //@line 4010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $570 = (($0) - ($568))|0; //@line 4010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $571 = ($569|0)<($570|0); //@line 4010 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   if (!($571)) {
    $$0$lcssa211 = $$0;$__i$4$lcssa213 = $__i$4;
    label = 67;
    break;
   }
   __ZNSt3__16__sortIRPFbRK6b2PairS3_EPS1_EEvT0_S8_T_($$0,$__i$4,$__comp); //@line 4012 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $572 = ((($__i$4)) + 8|0); //@line 4014 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $$0 = $572;
  }
  if ((label|0) == 67) {
   label = 0;
   $573 = ((($__i$4$lcssa213)) + 8|0); //@line 4018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   __ZNSt3__16__sortIRPFbRK6b2PairS3_EPS1_EEvT0_S8_T_($573,$$04$ph,$__comp); //@line 4018 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $$0$ph = $$0$lcssa211;$$04$ph = $__i$4$lcssa213;
   continue;
  }
  else if ((label|0) == 68) {
   label = 0;
   if ($565) {
    label = 69;
    break;
   } else {
    $$0$ph = $$0;$$04$ph = $__i$4;
    continue;
   }
  }
 }
 if ((label|0) == 4) {
  $5 = HEAP32[$__comp>>2]|0; //@line 3851 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $6 = (FUNCTION_TABLE_iii[$5 & 8191]($$lcssa224,$$0$lcssa204)|0); //@line 3851 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($6)) {
   return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  $7 = $$0$lcssa204; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $8 = $7; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $9 = HEAP32[$8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $10 = (($7) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $11 = $10; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $12 = HEAP32[$11>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $13 = $$lcssa232; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $14 = $13; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $15 = HEAP32[$14>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $16 = (($13) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $17 = $16; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $18 = HEAP32[$17>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $19 = $$0$lcssa204; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $20 = $19; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$20>>2] = $15; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $21 = (($19) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $22 = $21; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$22>>2] = $18; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $23 = $$lcssa232; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $24 = $23; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$24>>2] = $9; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $25 = (($23) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $26 = $25; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$26>>2] = $12; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
 else if ((label|0) == 6) {
  $27 = ((($$0$lcssa205)) + 8|0); //@line 3855 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $28 = HEAP32[$__comp>>2]|0; //@line 3601 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $29 = (FUNCTION_TABLE_iii[$28 & 8191]($27,$$0$lcssa205)|0); //@line 3601 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $30 = HEAP32[$__comp>>2]|0; //@line 3615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $31 = (FUNCTION_TABLE_iii[$30 & 8191]($$lcssa225,$27)|0); //@line 3615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($29)) {
   if (!($31)) {
    return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   }
   $32 = $27; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $33 = $32; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $34 = HEAP32[$33>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $35 = (($32) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $36 = $35; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $37 = HEAP32[$36>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $38 = $$lcssa233; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $39 = $38; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $40 = HEAP32[$39>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $41 = (($38) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $42 = $41; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $43 = HEAP32[$42>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $44 = $27; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $45 = $44; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$45>>2] = $40; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $46 = (($44) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $47 = $46; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$47>>2] = $43; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $48 = $$lcssa233; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $49 = $48; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$49>>2] = $34; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $50 = (($48) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $51 = $50; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$51>>2] = $37; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $52 = HEAP32[$__comp>>2]|0; //@line 3608 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $53 = (FUNCTION_TABLE_iii[$52 & 8191]($27,$$0$lcssa205)|0); //@line 3608 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   if (!($53)) {
    return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   }
   $54 = $$0$lcssa205; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $55 = $54; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $56 = HEAP32[$55>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $57 = (($54) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $58 = $57; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $59 = HEAP32[$58>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $60 = $27; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $61 = $60; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $62 = HEAP32[$61>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $63 = (($60) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $64 = $63; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $65 = HEAP32[$64>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $66 = $$0$lcssa205; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $67 = $66; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$67>>2] = $62; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $68 = (($66) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $69 = $68; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$69>>2] = $65; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $70 = $27; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $71 = $70; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$71>>2] = $56; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $72 = (($70) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $73 = $72; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$73>>2] = $59; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  $74 = $$0$lcssa205; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $75 = $74; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $76 = HEAP32[$75>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $77 = (($74) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $78 = $77; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $79 = HEAP32[$78>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  if ($31) {
   $80 = $$lcssa233; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $81 = $80; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $82 = HEAP32[$81>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $83 = (($80) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $84 = $83; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $85 = HEAP32[$84>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $86 = $$0$lcssa205; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $87 = $86; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$87>>2] = $82; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $88 = (($86) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $89 = $88; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$89>>2] = $85; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $90 = $$lcssa233; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $91 = $90; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$91>>2] = $76; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $92 = (($90) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $93 = $92; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$93>>2] = $79; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  $94 = $27; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $95 = $94; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $96 = HEAP32[$95>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $97 = (($94) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $98 = $97; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $99 = HEAP32[$98>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $100 = $$0$lcssa205; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $101 = $100; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$101>>2] = $96; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $102 = (($100) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $103 = $102; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$103>>2] = $99; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $104 = $27; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $105 = $104; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$105>>2] = $76; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $106 = (($104) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $107 = $106; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$107>>2] = $79; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $108 = HEAP32[$__comp>>2]|0; //@line 3623 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $109 = (FUNCTION_TABLE_iii[$108 & 8191]($$lcssa225,$27)|0); //@line 3623 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($109)) {
   return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  $110 = $27; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $111 = $110; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $112 = HEAP32[$111>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $113 = (($110) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $114 = $113; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $115 = HEAP32[$114>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $116 = $$lcssa233; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $117 = $116; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $118 = HEAP32[$117>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $119 = (($116) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $120 = $119; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $121 = HEAP32[$120>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $122 = $27; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $123 = $122; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$123>>2] = $118; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $124 = (($122) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $125 = $124; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$125>>2] = $121; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $126 = $$lcssa233; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $127 = $126; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$127>>2] = $112; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $128 = (($126) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $129 = $128; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$129>>2] = $115; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
 else if ((label|0) == 14) {
  $130 = ((($$0$lcssa206)) + 8|0); //@line 3858 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $131 = ((($$0$lcssa206)) + 16|0); //@line 3858 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  (__ZNSt3__17__sort4IRPFbRK6b2PairS3_EPS1_EEjT0_S8_S8_S8_T_($$0$lcssa206,$130,$131,$$lcssa226,$__comp)|0); //@line 3858 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
 else if ((label|0) == 15) {
  $132 = ((($$0$lcssa207)) + 8|0); //@line 3861 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $133 = ((($$0$lcssa207)) + 16|0); //@line 3861 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $134 = ((($$0$lcssa207)) + 24|0); //@line 3861 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  (__ZNSt3__17__sort4IRPFbRK6b2PairS3_EPS1_EEjT0_S8_S8_S8_T_($$0$lcssa207,$132,$133,$134,$__comp)|0); //@line 3664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $135 = HEAP32[$__comp>>2]|0; //@line 3665 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $136 = (FUNCTION_TABLE_iii[$135 & 8191]($$lcssa227,$134)|0); //@line 3665 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($136)) {
   return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  $137 = $134; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $138 = $137; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $139 = HEAP32[$138>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $140 = (($137) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $141 = $140; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $142 = HEAP32[$141>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $143 = $$lcssa235; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $144 = $143; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $145 = HEAP32[$144>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $146 = (($143) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $147 = $146; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $148 = HEAP32[$147>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $149 = $134; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $150 = $149; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$150>>2] = $145; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $151 = (($149) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $152 = $151; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$152>>2] = $148; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $153 = $$lcssa235; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $154 = $153; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$154>>2] = $139; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $155 = (($153) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $156 = $155; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$156>>2] = $142; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $157 = HEAP32[$__comp>>2]|0; //@line 3669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $158 = (FUNCTION_TABLE_iii[$157 & 8191]($134,$133)|0); //@line 3669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($158)) {
   return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  $159 = $133; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $160 = $159; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $161 = HEAP32[$160>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $162 = (($159) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $163 = $162; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $164 = HEAP32[$163>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $165 = $134; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $166 = $165; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $167 = HEAP32[$166>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $168 = (($165) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $169 = $168; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $170 = HEAP32[$169>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $171 = $133; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $172 = $171; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$172>>2] = $167; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $173 = (($171) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $174 = $173; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$174>>2] = $170; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $175 = $134; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $176 = $175; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$176>>2] = $161; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $177 = (($175) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $178 = $177; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$178>>2] = $164; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $179 = HEAP32[$__comp>>2]|0; //@line 3673 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $180 = (FUNCTION_TABLE_iii[$179 & 8191]($133,$132)|0); //@line 3673 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($180)) {
   return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  $181 = $132; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $182 = $181; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $183 = HEAP32[$182>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $184 = (($181) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $185 = $184; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $186 = HEAP32[$185>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $187 = $133; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $188 = $187; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $189 = HEAP32[$188>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $190 = (($187) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $191 = $190; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $192 = HEAP32[$191>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $193 = $132; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $194 = $193; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$194>>2] = $189; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $195 = (($193) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $196 = $195; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$196>>2] = $192; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $197 = $133; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $198 = $197; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$198>>2] = $183; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $199 = (($197) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $200 = $199; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$200>>2] = $186; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $201 = HEAP32[$__comp>>2]|0; //@line 3677 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $202 = (FUNCTION_TABLE_iii[$201 & 8191]($132,$$0$lcssa207)|0); //@line 3677 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($202)) {
   return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  $203 = $$0$lcssa207; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $204 = $203; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $205 = HEAP32[$204>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $206 = (($203) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $207 = $206; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $208 = HEAP32[$207>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $209 = $132; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $210 = $209; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $211 = HEAP32[$210>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $212 = (($209) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $213 = $212; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $214 = HEAP32[$213>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $215 = $$0$lcssa207; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $216 = $215; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$216>>2] = $211; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $217 = (($215) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $218 = $217; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$218>>2] = $214; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $219 = $132; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $220 = $219; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$220>>2] = $205; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $221 = (($219) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $222 = $221; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$222>>2] = $208; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
 else if ((label|0) == 21) {
  __ZNSt3__118__insertion_sort_3IRPFbRK6b2PairS3_EPS1_EEvT0_S8_T_($$0$lcssa208,$$04$ph$lcssa220,$__comp); //@line 3866 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
 else if ((label|0) == 69) {
  return; //@line 4023 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
}
function __ZNSt3__17__sort4IRPFbRK6b2PairS3_EPS1_EEjT0_S8_S8_S8_T_($__x1,$__x2,$__x3,$__x4,$__c) {
 $__x1 = $__x1|0;
 $__x2 = $__x2|0;
 $__x3 = $__x3|0;
 $__x4 = $__x4|0;
 $__c = $__c|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__r$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$__c>>2]|0; //@line 3601 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $1 = (FUNCTION_TABLE_iii[$0 & 8191]($__x2,$__x1)|0); //@line 3601 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $2 = HEAP32[$__c>>2]|0; //@line 3615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $3 = (FUNCTION_TABLE_iii[$2 & 8191]($__x3,$__x2)|0); //@line 3615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 do {
  if ($1) {
   $46 = $__x1; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $47 = $46; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $48 = HEAP32[$47>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $49 = (($46) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $50 = $49; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $51 = HEAP32[$50>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   if ($3) {
    $52 = $__x3; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $53 = $52; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $54 = HEAP32[$53>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $55 = (($52) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $56 = $55; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $57 = HEAP32[$56>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $58 = $__x1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $59 = $58; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$59>>2] = $54; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $60 = (($58) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $61 = $60; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$61>>2] = $57; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $62 = $__x3; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $63 = $62; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$63>>2] = $48; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $64 = (($62) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $65 = $64; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$65>>2] = $51; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $$0$i = 1;
    break;
   }
   $66 = $__x2; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $67 = $66; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $68 = HEAP32[$67>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $69 = (($66) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $70 = $69; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $71 = HEAP32[$70>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $72 = $__x1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $73 = $72; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$73>>2] = $68; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $74 = (($72) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $75 = $74; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$75>>2] = $71; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $76 = $__x2; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $77 = $76; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$77>>2] = $48; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $78 = (($76) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $79 = $78; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$79>>2] = $51; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $80 = HEAP32[$__c>>2]|0; //@line 3623 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $81 = (FUNCTION_TABLE_iii[$80 & 8191]($__x3,$__x2)|0); //@line 3623 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   if ($81) {
    $82 = $__x2; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $83 = $82; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $84 = HEAP32[$83>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $85 = (($82) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $86 = $85; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $87 = HEAP32[$86>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $88 = $__x3; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $89 = $88; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $90 = HEAP32[$89>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $91 = (($88) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $92 = $91; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $93 = HEAP32[$92>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $94 = $__x2; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $95 = $94; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$95>>2] = $90; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $96 = (($94) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $97 = $96; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$97>>2] = $93; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $98 = $__x3; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $99 = $98; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$99>>2] = $84; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $100 = (($98) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $101 = $100; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$101>>2] = $87; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $$0$i = 2;
   } else {
    $$0$i = 1;
   }
  } else {
   if ($3) {
    $4 = $__x2; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $5 = $4; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $6 = HEAP32[$5>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $7 = (($4) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $8 = $7; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $9 = HEAP32[$8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $10 = $__x3; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $11 = $10; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $12 = HEAP32[$11>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $13 = (($10) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $14 = $13; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $15 = HEAP32[$14>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $16 = $__x2; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $17 = $16; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$17>>2] = $12; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $18 = (($16) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $19 = $18; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$19>>2] = $15; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $20 = $__x3; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $21 = $20; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$21>>2] = $6; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $22 = (($20) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $23 = $22; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$23>>2] = $9; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $24 = HEAP32[$__c>>2]|0; //@line 3608 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $25 = (FUNCTION_TABLE_iii[$24 & 8191]($__x2,$__x1)|0); //@line 3608 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    if ($25) {
     $26 = $__x1; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $27 = $26; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $28 = HEAP32[$27>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $29 = (($26) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $30 = $29; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $31 = HEAP32[$30>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $32 = $__x2; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $33 = $32; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $34 = HEAP32[$33>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $35 = (($32) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $36 = $35; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $37 = HEAP32[$36>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $38 = $__x1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $39 = $38; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$39>>2] = $34; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $40 = (($38) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $41 = $40; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$41>>2] = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $42 = $__x2; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $43 = $42; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$43>>2] = $28; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $44 = (($42) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $45 = $44; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$45>>2] = $31; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $$0$i = 2;
    } else {
     $$0$i = 1;
    }
   } else {
    $$0$i = 0;
   }
  }
 } while(0);
 $102 = HEAP32[$__c>>2]|0; //@line 3639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $103 = (FUNCTION_TABLE_iii[$102 & 8191]($__x4,$__x3)|0); //@line 3639 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 if (!($103)) {
  $__r$0 = $$0$i;
  return ($__r$0|0); //@line 3654 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
 $104 = $__x3; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $105 = $104; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $106 = HEAP32[$105>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $107 = (($104) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $108 = $107; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $109 = HEAP32[$108>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $110 = $__x4; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $111 = $110; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $112 = HEAP32[$111>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $113 = (($110) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $114 = $113; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $115 = HEAP32[$114>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $116 = $__x3; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $117 = $116; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$117>>2] = $112; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $118 = (($116) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $119 = $118; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$119>>2] = $115; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $120 = $__x4; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $121 = $120; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$121>>2] = $106; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $122 = (($120) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $123 = $122; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$123>>2] = $109; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $124 = (($$0$i) + 1)|0; //@line 3642 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $125 = HEAP32[$__c>>2]|0; //@line 3643 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $126 = (FUNCTION_TABLE_iii[$125 & 8191]($__x3,$__x2)|0); //@line 3643 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 if (!($126)) {
  $__r$0 = $124;
  return ($__r$0|0); //@line 3654 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
 $127 = $__x2; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $128 = $127; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $129 = HEAP32[$128>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $130 = (($127) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $131 = $130; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $132 = HEAP32[$131>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $133 = $__x3; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $134 = $133; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $135 = HEAP32[$134>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $136 = (($133) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $137 = $136; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $138 = HEAP32[$137>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $139 = $__x2; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $140 = $139; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$140>>2] = $135; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $141 = (($139) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $142 = $141; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$142>>2] = $138; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $143 = $__x3; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $144 = $143; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$144>>2] = $129; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $145 = (($143) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $146 = $145; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$146>>2] = $132; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $147 = (($$0$i) + 2)|0; //@line 3646 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $148 = HEAP32[$__c>>2]|0; //@line 3647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $149 = (FUNCTION_TABLE_iii[$148 & 8191]($__x2,$__x1)|0); //@line 3647 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 if (!($149)) {
  $__r$0 = $147;
  return ($__r$0|0); //@line 3654 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
 $150 = $__x1; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $151 = $150; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $152 = HEAP32[$151>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $153 = (($150) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $154 = $153; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $155 = HEAP32[$154>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $156 = $__x2; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $157 = $156; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $158 = HEAP32[$157>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $159 = (($156) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $160 = $159; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $161 = HEAP32[$160>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $162 = $__x1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $163 = $162; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$163>>2] = $158; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $164 = (($162) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $165 = $164; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$165>>2] = $161; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $166 = $__x2; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $167 = $166; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$167>>2] = $152; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $168 = (($166) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $169 = $168; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$169>>2] = $155; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $170 = (($$0$i) + 3)|0; //@line 3650 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $__r$0 = $170;
 return ($__r$0|0); //@line 3654 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
}
function __ZNSt3__118__insertion_sort_3IRPFbRK6b2PairS3_EPS1_EEvT0_S8_T_($__first,$__last,$__comp) {
 $__first = $__first|0;
 $__last = $__last|0;
 $__comp = $__comp|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__i$02 = 0, $__j$01 = 0, $__j$01$phi = 0, $__j$1 = 0, $__j$1$phi = 0, $__k$0 = 0, $__t = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__t = sp;
 $0 = ((($__first)) + 16|0); //@line 3728 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $1 = ((($__first)) + 8|0); //@line 3729 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $2 = HEAP32[$__comp>>2]|0; //@line 3601 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $3 = (FUNCTION_TABLE_iii[$2 & 8191]($1,$__first)|0); //@line 3601 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $4 = HEAP32[$__comp>>2]|0; //@line 3615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $5 = (FUNCTION_TABLE_iii[$4 & 8191]($0,$1)|0); //@line 3615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 do {
  if ($3) {
   $48 = $__first; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $49 = $48; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $50 = HEAP32[$49>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $51 = (($48) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $52 = $51; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $53 = HEAP32[$52>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   if ($5) {
    $54 = $0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $55 = $54; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $56 = HEAP32[$55>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $57 = (($54) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $58 = $57; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $59 = HEAP32[$58>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $60 = $__first; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $61 = $60; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$61>>2] = $56; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $62 = (($60) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $63 = $62; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$63>>2] = $59; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $64 = $0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $65 = $64; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$65>>2] = $50; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $66 = (($64) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $67 = $66; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$67>>2] = $53; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    break;
   }
   $68 = $1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $69 = $68; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $70 = HEAP32[$69>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $71 = (($68) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $72 = $71; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $73 = HEAP32[$72>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $74 = $__first; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $75 = $74; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$75>>2] = $70; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $76 = (($74) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $77 = $76; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$77>>2] = $73; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $78 = $1; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $79 = $78; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$79>>2] = $50; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $80 = (($78) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $81 = $80; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$81>>2] = $53; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $82 = HEAP32[$__comp>>2]|0; //@line 3623 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $83 = (FUNCTION_TABLE_iii[$82 & 8191]($0,$1)|0); //@line 3623 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   if ($83) {
    $84 = $1; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $85 = $84; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $86 = HEAP32[$85>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $87 = (($84) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $88 = $87; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $89 = HEAP32[$88>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $90 = $0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $91 = $90; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $92 = HEAP32[$91>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $93 = (($90) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $94 = $93; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $95 = HEAP32[$94>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $96 = $1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $97 = $96; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$97>>2] = $92; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $98 = (($96) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $99 = $98; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$99>>2] = $95; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $100 = $0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $101 = $100; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$101>>2] = $86; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $102 = (($100) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $103 = $102; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$103>>2] = $89; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   }
  } else {
   if ($5) {
    $6 = $1; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $7 = $6; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $8 = HEAP32[$7>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $9 = (($6) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $10 = $9; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $11 = HEAP32[$10>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $12 = $0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $13 = $12; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $14 = HEAP32[$13>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $15 = (($12) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $16 = $15; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $17 = HEAP32[$16>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $18 = $1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $19 = $18; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$19>>2] = $14; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $20 = (($18) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $21 = $20; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$21>>2] = $17; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $22 = $0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $23 = $22; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$23>>2] = $8; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $24 = (($22) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $25 = $24; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$25>>2] = $11; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $26 = HEAP32[$__comp>>2]|0; //@line 3608 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $27 = (FUNCTION_TABLE_iii[$26 & 8191]($1,$__first)|0); //@line 3608 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    if ($27) {
     $28 = $__first; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $29 = $28; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $30 = HEAP32[$29>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $31 = (($28) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $32 = $31; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $33 = HEAP32[$32>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $34 = $1; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $35 = $34; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $36 = HEAP32[$35>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $37 = (($34) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $38 = $37; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $39 = HEAP32[$38>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $40 = $__first; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $41 = $40; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$41>>2] = $36; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $42 = (($40) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $43 = $42; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$43>>2] = $39; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $44 = $1; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $45 = $44; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$45>>2] = $30; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $46 = (($44) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $47 = $46; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$47>>2] = $33; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    }
   }
  }
 } while(0);
 $104 = ((($__first)) + 24|0); //@line 3730 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $105 = ($104|0)==($__last|0); //@line 3730 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 if ($105) {
  STACKTOP = sp;return; //@line 3746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
 $__i$02 = $104;$__j$01 = $0;
 while(1) {
  $106 = HEAP32[$__comp>>2]|0; //@line 3732 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $107 = (FUNCTION_TABLE_iii[$106 & 8191]($__i$02,$__j$01)|0); //@line 3732 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if ($107) {
   $108 = $__i$02; //@line 3734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $109 = $108; //@line 3734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $110 = HEAP32[$109>>2]|0; //@line 3734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $111 = (($108) + 4)|0; //@line 3734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $112 = $111; //@line 3734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $113 = HEAP32[$112>>2]|0; //@line 3734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $114 = $__t; //@line 3734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $115 = $114; //@line 3734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   HEAP32[$115>>2] = $110; //@line 3734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $116 = (($114) + 4)|0; //@line 3734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $117 = $116; //@line 3734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   HEAP32[$117>>2] = $113; //@line 3734 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $__j$1 = $__i$02;$__k$0 = $__j$01;
   while(1) {
    $118 = $__k$0; //@line 3739 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $119 = $118; //@line 3739 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $120 = HEAP32[$119>>2]|0; //@line 3739 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $121 = (($118) + 4)|0; //@line 3739 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $122 = $121; //@line 3739 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $123 = HEAP32[$122>>2]|0; //@line 3739 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $124 = $__j$1; //@line 3739 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $125 = $124; //@line 3739 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    HEAP32[$125>>2] = $120; //@line 3739 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $126 = (($124) + 4)|0; //@line 3739 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $127 = $126; //@line 3739 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    HEAP32[$127>>2] = $123; //@line 3739 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $128 = ($__k$0|0)==($__first|0); //@line 3741 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    if ($128) {
     $$lcssa = $__k$0;
     break;
    }
    $129 = HEAP32[$__comp>>2]|0; //@line 3741 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $130 = ((($__k$0)) + -8|0); //@line 3741 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $131 = (FUNCTION_TABLE_iii[$129 & 8191]($__t,$130)|0); //@line 3741 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    if ($131) {
     $__j$1$phi = $__k$0;$__k$0 = $130;$__j$1 = $__j$1$phi;
    } else {
     $$lcssa = $__k$0;
     break;
    }
   }
   $132 = $__t; //@line 3742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $133 = $132; //@line 3742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $134 = HEAP32[$133>>2]|0; //@line 3742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $135 = (($132) + 4)|0; //@line 3742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $136 = $135; //@line 3742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $137 = HEAP32[$136>>2]|0; //@line 3742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $138 = $$lcssa; //@line 3742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $139 = $138; //@line 3742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   HEAP32[$139>>2] = $134; //@line 3742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $140 = (($138) + 4)|0; //@line 3742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $141 = $140; //@line 3742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   HEAP32[$141>>2] = $137; //@line 3742 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  $142 = ((($__i$02)) + 8|0); //@line 3730 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $143 = ($142|0)==($__last|0); //@line 3730 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if ($143) {
   break;
  } else {
   $__j$01$phi = $__i$02;$__i$02 = $142;$__j$01 = $__j$01$phi;
  }
 }
 STACKTOP = sp;return; //@line 3746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
}
function __ZNSt3__127__insertion_sort_incompleteIRPFbRK6b2PairS3_EPS1_EEbT0_S8_T_($__first,$__last,$__comp) {
 $__first = $__first|0;
 $__last = $__last|0;
 $__comp = $__comp|0;
 var $$$3 = 0, $$1 = 0, $$3 = 0, $$5 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0;
 var $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0;
 var $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0;
 var $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0;
 var $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0;
 var $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0;
 var $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0;
 var $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0;
 var $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0;
 var $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0;
 var $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__count$08 = 0;
 var $__count$1 = 0, $__i$09 = 0, $__j$06 = 0, $__j$06$phi = 0, $__j$1 = 0, $__j$1$phi = 0, $__k$0 = 0, $__t = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__t = sp;
 $0 = $__last; //@line 3752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $1 = $__first; //@line 3752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $2 = (($0) - ($1))|0; //@line 3752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $3 = $2 >> 3; //@line 3752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 switch ($3|0) {
 case 2:  {
  $4 = HEAP32[$__comp>>2]|0; //@line 3758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $5 = ((($__last)) + -8|0); //@line 3758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $6 = (FUNCTION_TABLE_iii[$4 & 8191]($5,$__first)|0); //@line 3758 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($6)) {
   $$5 = 1;
   STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  $7 = $__first; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $8 = $7; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $9 = HEAP32[$8>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $10 = (($7) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $11 = $10; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $12 = HEAP32[$11>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $13 = $5; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $14 = $13; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $15 = HEAP32[$14>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $16 = (($13) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $17 = $16; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $18 = HEAP32[$17>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $19 = $__first; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $20 = $19; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$20>>2] = $15; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $21 = (($19) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $22 = $21; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$22>>2] = $18; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $23 = $5; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $24 = $23; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$24>>2] = $9; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $25 = (($23) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $26 = $25; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$26>>2] = $12; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $$5 = 1;
  STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  break;
 }
 case 3:  {
  $27 = ((($__first)) + 8|0); //@line 3762 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $28 = ((($__last)) + -8|0); //@line 3762 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $29 = HEAP32[$__comp>>2]|0; //@line 3601 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $30 = (FUNCTION_TABLE_iii[$29 & 8191]($27,$__first)|0); //@line 3601 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $31 = HEAP32[$__comp>>2]|0; //@line 3615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $32 = (FUNCTION_TABLE_iii[$31 & 8191]($28,$27)|0); //@line 3615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($30)) {
   if (!($32)) {
    $$5 = 1;
    STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   }
   $33 = $27; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $34 = $33; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $35 = HEAP32[$34>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $36 = (($33) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $37 = $36; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $38 = HEAP32[$37>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $39 = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $40 = $39; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $41 = HEAP32[$40>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $42 = (($39) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $43 = $42; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $44 = HEAP32[$43>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $45 = $27; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $46 = $45; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$46>>2] = $41; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $47 = (($45) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $48 = $47; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$48>>2] = $44; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $49 = $28; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $50 = $49; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$50>>2] = $35; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $51 = (($49) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $52 = $51; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$52>>2] = $38; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $53 = HEAP32[$__comp>>2]|0; //@line 3608 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   $54 = (FUNCTION_TABLE_iii[$53 & 8191]($27,$__first)|0); //@line 3608 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   if (!($54)) {
    $$5 = 1;
    STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
   }
   $55 = $__first; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $56 = $55; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $57 = HEAP32[$56>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $58 = (($55) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $59 = $58; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $60 = HEAP32[$59>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $61 = $27; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $62 = $61; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $63 = HEAP32[$62>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $64 = (($61) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $65 = $64; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $66 = HEAP32[$65>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $67 = $__first; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $68 = $67; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$68>>2] = $63; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $69 = (($67) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $70 = $69; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$70>>2] = $66; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $71 = $27; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $72 = $71; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$72>>2] = $57; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $73 = (($71) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $74 = $73; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$74>>2] = $60; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $$5 = 1;
   STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  $75 = $__first; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $76 = $75; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $77 = HEAP32[$76>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $78 = (($75) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $79 = $78; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $80 = HEAP32[$79>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  if ($32) {
   $81 = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $82 = $81; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $83 = HEAP32[$82>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $84 = (($81) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $85 = $84; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $86 = HEAP32[$85>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $87 = $__first; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $88 = $87; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$88>>2] = $83; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $89 = (($87) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $90 = $89; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$90>>2] = $86; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $91 = $28; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $92 = $91; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$92>>2] = $77; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $93 = (($91) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $94 = $93; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   HEAP32[$94>>2] = $80; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
   $$5 = 1;
   STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  $95 = $27; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $96 = $95; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $97 = HEAP32[$96>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $98 = (($95) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $99 = $98; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $100 = HEAP32[$99>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $101 = $__first; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $102 = $101; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$102>>2] = $97; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $103 = (($101) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $104 = $103; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$104>>2] = $100; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $105 = $27; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $106 = $105; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$106>>2] = $77; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $107 = (($105) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $108 = $107; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$108>>2] = $80; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $109 = HEAP32[$__comp>>2]|0; //@line 3623 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $110 = (FUNCTION_TABLE_iii[$109 & 8191]($28,$27)|0); //@line 3623 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($110)) {
   $$5 = 1;
   STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  $111 = $27; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $112 = $111; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $113 = HEAP32[$112>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $114 = (($111) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $115 = $114; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $116 = HEAP32[$115>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $117 = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $118 = $117; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $119 = HEAP32[$118>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $120 = (($117) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $121 = $120; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $122 = HEAP32[$121>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $123 = $27; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $124 = $123; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$124>>2] = $119; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $125 = (($123) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $126 = $125; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$126>>2] = $122; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $127 = $28; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $128 = $127; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$128>>2] = $113; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $129 = (($127) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $130 = $129; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$130>>2] = $116; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $$5 = 1;
  STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  break;
 }
 case 4:  {
  $131 = ((($__first)) + 8|0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $132 = ((($__first)) + 16|0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $133 = ((($__last)) + -8|0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  (__ZNSt3__17__sort4IRPFbRK6b2PairS3_EPS1_EEjT0_S8_S8_S8_T_($__first,$131,$132,$133,$__comp)|0); //@line 3765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $$5 = 1;
  STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  break;
 }
 case 5:  {
  $134 = ((($__first)) + 8|0); //@line 3768 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $135 = ((($__first)) + 16|0); //@line 3768 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $136 = ((($__first)) + 24|0); //@line 3768 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $137 = ((($__last)) + -8|0); //@line 3768 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  (__ZNSt3__17__sort4IRPFbRK6b2PairS3_EPS1_EEjT0_S8_S8_S8_T_($__first,$134,$135,$136,$__comp)|0); //@line 3664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $138 = HEAP32[$__comp>>2]|0; //@line 3665 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $139 = (FUNCTION_TABLE_iii[$138 & 8191]($137,$136)|0); //@line 3665 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($139)) {
   $$5 = 1;
   STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  $140 = $136; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $141 = $140; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $142 = HEAP32[$141>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $143 = (($140) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $144 = $143; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $145 = HEAP32[$144>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $146 = $137; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $147 = $146; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $148 = HEAP32[$147>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $149 = (($146) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $150 = $149; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $151 = HEAP32[$150>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $152 = $136; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $153 = $152; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$153>>2] = $148; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $154 = (($152) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $155 = $154; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$155>>2] = $151; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $156 = $137; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $157 = $156; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$157>>2] = $142; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $158 = (($156) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $159 = $158; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$159>>2] = $145; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $160 = HEAP32[$__comp>>2]|0; //@line 3669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $161 = (FUNCTION_TABLE_iii[$160 & 8191]($136,$135)|0); //@line 3669 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($161)) {
   $$5 = 1;
   STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  $162 = $135; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $163 = $162; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $164 = HEAP32[$163>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $165 = (($162) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $166 = $165; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $167 = HEAP32[$166>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $168 = $136; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $169 = $168; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $170 = HEAP32[$169>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $171 = (($168) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $172 = $171; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $173 = HEAP32[$172>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $174 = $135; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $175 = $174; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$175>>2] = $170; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $176 = (($174) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $177 = $176; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$177>>2] = $173; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $178 = $136; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $179 = $178; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$179>>2] = $164; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $180 = (($178) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $181 = $180; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$181>>2] = $167; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $182 = HEAP32[$__comp>>2]|0; //@line 3673 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $183 = (FUNCTION_TABLE_iii[$182 & 8191]($135,$134)|0); //@line 3673 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($183)) {
   $$5 = 1;
   STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  $184 = $134; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $185 = $184; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $186 = HEAP32[$185>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $187 = (($184) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $188 = $187; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $189 = HEAP32[$188>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $190 = $135; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $191 = $190; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $192 = HEAP32[$191>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $193 = (($190) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $194 = $193; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $195 = HEAP32[$194>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $196 = $134; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $197 = $196; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$197>>2] = $192; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $198 = (($196) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $199 = $198; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$199>>2] = $195; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $200 = $135; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $201 = $200; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$201>>2] = $186; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $202 = (($200) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $203 = $202; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$203>>2] = $189; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $204 = HEAP32[$__comp>>2]|0; //@line 3677 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $205 = (FUNCTION_TABLE_iii[$204 & 8191]($134,$__first)|0); //@line 3677 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($205)) {
   $$5 = 1;
   STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  }
  $206 = $__first; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $207 = $206; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $208 = HEAP32[$207>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $209 = (($206) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $210 = $209; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $211 = HEAP32[$210>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $212 = $134; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $213 = $212; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $214 = HEAP32[$213>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $215 = (($212) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $216 = $215; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $217 = HEAP32[$216>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $218 = $__first; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $219 = $218; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$219>>2] = $214; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $220 = (($218) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $221 = $220; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$221>>2] = $217; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $222 = $134; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $223 = $222; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$223>>2] = $208; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $224 = (($222) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $225 = $224; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  HEAP32[$225>>2] = $211; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
  $$5 = 1;
  STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  break;
 }
 case 1: case 0:  {
  $$5 = 1;
  STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  break;
 }
 default: {
  $226 = ((($__first)) + 16|0); //@line 3772 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $227 = ((($__first)) + 8|0); //@line 3773 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $228 = HEAP32[$__comp>>2]|0; //@line 3601 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $229 = (FUNCTION_TABLE_iii[$228 & 8191]($227,$__first)|0); //@line 3601 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $230 = HEAP32[$__comp>>2]|0; //@line 3615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $231 = (FUNCTION_TABLE_iii[$230 & 8191]($226,$227)|0); //@line 3615 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  do {
   if ($229) {
    $274 = $__first; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $275 = $274; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $276 = HEAP32[$275>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $277 = (($274) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $278 = $277; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $279 = HEAP32[$278>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    if ($231) {
     $280 = $226; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $281 = $280; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $282 = HEAP32[$281>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $283 = (($280) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $284 = $283; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $285 = HEAP32[$284>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $286 = $__first; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $287 = $286; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$287>>2] = $282; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $288 = (($286) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $289 = $288; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$289>>2] = $285; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $290 = $226; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $291 = $290; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$291>>2] = $276; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $292 = (($290) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $293 = $292; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$293>>2] = $279; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     break;
    }
    $294 = $227; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $295 = $294; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $296 = HEAP32[$295>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $297 = (($294) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $298 = $297; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $299 = HEAP32[$298>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $300 = $__first; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $301 = $300; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$301>>2] = $296; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $302 = (($300) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $303 = $302; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$303>>2] = $299; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $304 = $227; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $305 = $304; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$305>>2] = $276; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $306 = (($304) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $307 = $306; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    HEAP32[$307>>2] = $279; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    $308 = HEAP32[$__comp>>2]|0; //@line 3623 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    $309 = (FUNCTION_TABLE_iii[$308 & 8191]($226,$227)|0); //@line 3623 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
    if ($309) {
     $310 = $227; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $311 = $310; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $312 = HEAP32[$311>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $313 = (($310) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $314 = $313; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $315 = HEAP32[$314>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $316 = $226; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $317 = $316; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $318 = HEAP32[$317>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $319 = (($316) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $320 = $319; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $321 = HEAP32[$320>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $322 = $227; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $323 = $322; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$323>>2] = $318; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $324 = (($322) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $325 = $324; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$325>>2] = $321; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $326 = $226; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $327 = $326; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$327>>2] = $312; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $328 = (($326) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $329 = $328; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$329>>2] = $315; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
    }
   } else {
    if ($231) {
     $232 = $227; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $233 = $232; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $234 = HEAP32[$233>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $235 = (($232) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $236 = $235; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $237 = HEAP32[$236>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $238 = $226; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $239 = $238; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $240 = HEAP32[$239>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $241 = (($238) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $242 = $241; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $243 = HEAP32[$242>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $244 = $227; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $245 = $244; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$245>>2] = $240; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $246 = (($244) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $247 = $246; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$247>>2] = $243; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $248 = $226; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $249 = $248; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$249>>2] = $234; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $250 = (($248) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $251 = $250; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     HEAP32[$251>>2] = $237; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     $252 = HEAP32[$__comp>>2]|0; //@line 3608 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $253 = (FUNCTION_TABLE_iii[$252 & 8191]($227,$__first)|0); //@line 3608 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     if ($253) {
      $254 = $__first; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $255 = $254; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $256 = HEAP32[$255>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $257 = (($254) + 4)|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $258 = $257; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $259 = HEAP32[$258>>2]|0; //@line 3536 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $260 = $227; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $261 = $260; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $262 = HEAP32[$261>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $263 = (($260) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $264 = $263; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $265 = HEAP32[$264>>2]|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $266 = $__first; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $267 = $266; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$267>>2] = $262; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $268 = (($266) + 4)|0; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $269 = $268; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$269>>2] = $265; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $270 = $227; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $271 = $270; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$271>>2] = $256; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $272 = (($270) + 4)|0; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      $273 = $272; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
      HEAP32[$273>>2] = $259; //@line 3538 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
     }
    }
   }
  } while(0);
  $330 = ((($__first)) + 24|0); //@line 3776 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $331 = ($330|0)==($__last|0); //@line 3776 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  L13: do {
   if ($331) {
    $$1 = 1;$$3 = 0;
   } else {
    $__count$08 = 0;$__i$09 = $330;$__j$06 = $226;
    while(1) {
     $332 = HEAP32[$__comp>>2]|0; //@line 3778 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $333 = (FUNCTION_TABLE_iii[$332 & 8191]($__i$09,$__j$06)|0); //@line 3778 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     if ($333) {
      $334 = $__i$09; //@line 3780 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $335 = $334; //@line 3780 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $336 = HEAP32[$335>>2]|0; //@line 3780 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $337 = (($334) + 4)|0; //@line 3780 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $338 = $337; //@line 3780 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $339 = HEAP32[$338>>2]|0; //@line 3780 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $340 = $__t; //@line 3780 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $341 = $340; //@line 3780 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      HEAP32[$341>>2] = $336; //@line 3780 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $342 = (($340) + 4)|0; //@line 3780 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $343 = $342; //@line 3780 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      HEAP32[$343>>2] = $339; //@line 3780 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $__j$1 = $__i$09;$__k$0 = $__j$06;
      while(1) {
       $344 = $__k$0; //@line 3785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $345 = $344; //@line 3785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $346 = HEAP32[$345>>2]|0; //@line 3785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $347 = (($344) + 4)|0; //@line 3785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $348 = $347; //@line 3785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $349 = HEAP32[$348>>2]|0; //@line 3785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $350 = $__j$1; //@line 3785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $351 = $350; //@line 3785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       HEAP32[$351>>2] = $346; //@line 3785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $352 = (($350) + 4)|0; //@line 3785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $353 = $352; //@line 3785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       HEAP32[$353>>2] = $349; //@line 3785 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $354 = ($__k$0|0)==($__first|0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       if ($354) {
        $$lcssa = $__k$0;
        break;
       }
       $355 = HEAP32[$__comp>>2]|0; //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $356 = ((($__k$0)) + -8|0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $357 = (FUNCTION_TABLE_iii[$355 & 8191]($__t,$356)|0); //@line 3787 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       if ($357) {
        $__j$1$phi = $__k$0;$__k$0 = $356;$__j$1 = $__j$1$phi;
       } else {
        $$lcssa = $__k$0;
        break;
       }
      }
      $358 = $__t; //@line 3788 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $359 = $358; //@line 3788 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $360 = HEAP32[$359>>2]|0; //@line 3788 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $361 = (($358) + 4)|0; //@line 3788 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $362 = $361; //@line 3788 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $363 = HEAP32[$362>>2]|0; //@line 3788 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $364 = $$lcssa; //@line 3788 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $365 = $364; //@line 3788 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      HEAP32[$365>>2] = $360; //@line 3788 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $366 = (($364) + 4)|0; //@line 3788 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $367 = $366; //@line 3788 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      HEAP32[$367>>2] = $363; //@line 3788 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $368 = (($__count$08) + 1)|0; //@line 3789 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $369 = ($368|0)==(8); //@line 3789 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $370 = ((($__i$09)) + 8|0); //@line 3790 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $371 = ($370|0)==($__last|0); //@line 3790 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      if ($369) {
       $$1 = 0;$$3 = $371;
       break L13;
      } else {
       $__count$1 = $368;
      }
     } else {
      $__count$1 = $__count$08;
     }
     $372 = ((($__i$09)) + 8|0); //@line 3776 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     $373 = ($372|0)==($__last|0); //@line 3776 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
     if ($373) {
      $$1 = 1;$$3 = 0;
      break;
     } else {
      $__j$06$phi = $__i$09;$__count$08 = $__count$1;$__i$09 = $372;$__j$06 = $__j$06$phi;
     }
    }
   }
  } while(0);
  $$$3 = $$3 | $$1;
  $$5 = $$$3;
  STACKTOP = sp;return ($$5|0); //@line 3795 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 }
 }
 return (0)|0;
}
function __ZNK13b2DynamicTree5QueryI12b2BroadPhaseEEvPT_RK6b2AABB($this,$callback,$aabb) {
 $this = $this|0;
 $callback = $callback|0;
 $aabb = $aabb|0;
 var $$old = 0, $$old11 = 0, $$pr = 0, $$pre = 0, $$pre$pre = 0, $$pre14 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0.0, $26 = 0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0;
 var $4 = 0, $40 = 0.0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $or$cond = 0, $or$cond$i = 0, $or$cond10$i = 0, $or$cond8$i = 0, $stack = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1040|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $stack = sp;
 $0 = ((($stack)) + 4|0); //@line 33 "./libraries/Box2D/Common/b2GrowableStack.h"
 HEAP32[$stack>>2] = $0; //@line 33 "./libraries/Box2D/Common/b2GrowableStack.h"
 $1 = ((($stack)) + 1028|0); //@line 34 "./libraries/Box2D/Common/b2GrowableStack.h"
 $2 = ((($stack)) + 1032|0); //@line 35 "./libraries/Box2D/Common/b2GrowableStack.h"
 HEAP32[$2>>2] = 256; //@line 35 "./libraries/Box2D/Common/b2GrowableStack.h"
 $3 = HEAP32[$this>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
 HEAP32[$0>>2] = $3; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
 HEAP32[$1>>2] = 1; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
 $4 = ((($this)) + 4|0); //@line 182 "./libraries/Box2D/Collision/b2DynamicTree.h"
 $5 = ((($aabb)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($aabb)) + 8|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($aabb)) + 12|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 while(1) {
  __THREW__ = 0;
  invoke_vii(3822,1,(446492|0)); //@line 67 "./libraries/Box2D/Common/b2GrowableStack.h"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if ($9) {
   label = 5;
   break;
  }
  $10 = HEAP32[$1>>2]|0; //@line 68 "./libraries/Box2D/Common/b2GrowableStack.h"
  $11 = (($10) + -1)|0; //@line 68 "./libraries/Box2D/Common/b2GrowableStack.h"
  HEAP32[$1>>2] = $11; //@line 68 "./libraries/Box2D/Common/b2GrowableStack.h"
  $12 = HEAP32[$stack>>2]|0; //@line 69 "./libraries/Box2D/Common/b2GrowableStack.h"
  $13 = (($12) + ($11<<2)|0); //@line 69 "./libraries/Box2D/Common/b2GrowableStack.h"
  $14 = HEAP32[$13>>2]|0; //@line 69 "./libraries/Box2D/Common/b2GrowableStack.h"
  $15 = ($14|0)==(-1); //@line 177 "./libraries/Box2D/Collision/b2DynamicTree.h"
  if ($15) {
   $$old = $11;$$pre14 = $12;
  } else {
   $24 = HEAP32[$4>>2]|0; //@line 182 "./libraries/Box2D/Collision/b2DynamicTree.h"
   $25 = +HEAPF32[$aabb>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $26 = (((($24) + (($14*36)|0)|0)) + 8|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $27 = +HEAPF32[$26>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $28 = $25 - $27; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $29 = +HEAPF32[$5>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $30 = (((($24) + (($14*36)|0)|0)) + 12|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $31 = +HEAPF32[$30>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $32 = $29 - $31; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $33 = (($24) + (($14*36)|0)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $34 = +HEAPF32[$33>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $35 = +HEAPF32[$6>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $36 = $34 - $35; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $37 = (((($24) + (($14*36)|0)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $38 = +HEAPF32[$37>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $39 = +HEAPF32[$7>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $40 = $38 - $39; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $41 = $28 > 0.0; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
   $42 = $32 > 0.0; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
   $or$cond$i = $41 | $42; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
   $43 = $36 > 0.0; //@line 271 "./libraries/Box2D/Collision/b2Collision.h"
   $or$cond8$i = $or$cond$i | $43; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
   $44 = $40 > 0.0; //@line 271 "./libraries/Box2D/Collision/b2Collision.h"
   $or$cond10$i = $or$cond8$i | $44; //@line 268 "./libraries/Box2D/Collision/b2Collision.h"
   if ($or$cond10$i) {
    $$old = $11;$$pre14 = $12;
   } else {
    $45 = (((($24) + (($14*36)|0)|0)) + 24|0); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $46 = HEAP32[$45>>2]|0; //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $47 = ($46|0)==(-1); //@line 32 "./libraries/Box2D/Collision/b2DynamicTree.h"
    if ($47) {
     __THREW__ = 0;
     $48 = (invoke_iii(4239,($callback|0),($14|0))|0); //@line 188 "./libraries/Box2D/Collision/b2DynamicTree.h"
     $49 = __THREW__; __THREW__ = 0;
     $50 = $49&1;
     if ($50) {
      label = 5;
      break;
     }
     $$pr = HEAP32[$1>>2]|0; //@line 74 "./libraries/Box2D/Common/b2GrowableStack.h"
     $51 = ($$pr|0)>(0); //@line 174 "./libraries/Box2D/Collision/b2DynamicTree.h"
     $or$cond = $48 & $51;
     if ($or$cond) {
      continue;
     } else {
      label = 14;
      break;
     }
    }
    $52 = HEAP32[$2>>2]|0; //@line 49 "./libraries/Box2D/Common/b2GrowableStack.h"
    $53 = ($11|0)==($52|0); //@line 49 "./libraries/Box2D/Common/b2GrowableStack.h"
    if ($53) {
     $54 = $11 << 1; //@line 52 "./libraries/Box2D/Common/b2GrowableStack.h"
     HEAP32[$2>>2] = $54; //@line 52 "./libraries/Box2D/Common/b2GrowableStack.h"
     $55 = $11 << 3; //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
     __THREW__ = 0;
     $56 = (invoke_ii(3831,($55|0))|0); //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
     $57 = __THREW__; __THREW__ = 0;
     $58 = $57&1;
     if ($58) {
      label = 5;
      break;
     }
     HEAP32[$stack>>2] = $56; //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
     $59 = HEAP32[$1>>2]|0; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
     $60 = $59 << 2; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
     _memcpy(($56|0),($12|0),($60|0))|0; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
     $61 = ($12|0)==($0|0); //@line 55 "./libraries/Box2D/Common/b2GrowableStack.h"
     if (!($61)) {
      __THREW__ = 0;
      invoke_vi(3816,($12|0)); //@line 57 "./libraries/Box2D/Common/b2GrowableStack.h"
      $62 = __THREW__; __THREW__ = 0;
      $63 = $62&1;
      if ($63) {
       label = 5;
       break;
      }
     }
    }
    $64 = HEAP32[$45>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    $65 = HEAP32[$1>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    $66 = HEAP32[$stack>>2]|0; //@line 51 "./libraries/Box2D/Common/b2GrowableStack.h"
    $67 = (($66) + ($65<<2)|0); //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    HEAP32[$67>>2] = $64; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    $68 = HEAP32[$1>>2]|0; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
    $69 = (($68) + 1)|0; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
    HEAP32[$1>>2] = $69; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
    $70 = (((($24) + (($14*36)|0)|0)) + 28|0); //@line 197 "./libraries/Box2D/Collision/b2DynamicTree.h"
    $71 = HEAP32[$2>>2]|0; //@line 49 "./libraries/Box2D/Common/b2GrowableStack.h"
    $72 = ($69|0)==($71|0); //@line 49 "./libraries/Box2D/Common/b2GrowableStack.h"
    if ($72) {
     $73 = $69 << 1; //@line 52 "./libraries/Box2D/Common/b2GrowableStack.h"
     HEAP32[$2>>2] = $73; //@line 52 "./libraries/Box2D/Common/b2GrowableStack.h"
     $74 = $69 << 3; //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
     __THREW__ = 0;
     $75 = (invoke_ii(3831,($74|0))|0); //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
     $76 = __THREW__; __THREW__ = 0;
     $77 = $76&1;
     if ($77) {
      label = 5;
      break;
     }
     HEAP32[$stack>>2] = $75; //@line 53 "./libraries/Box2D/Common/b2GrowableStack.h"
     $78 = HEAP32[$1>>2]|0; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
     $79 = $78 << 2; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
     _memcpy(($75|0),($66|0),($79|0))|0; //@line 54 "./libraries/Box2D/Common/b2GrowableStack.h"
     $80 = ($66|0)==($0|0); //@line 55 "./libraries/Box2D/Common/b2GrowableStack.h"
     if (!($80)) {
      __THREW__ = 0;
      invoke_vi(3816,($66|0)); //@line 57 "./libraries/Box2D/Common/b2GrowableStack.h"
      $81 = __THREW__; __THREW__ = 0;
      $82 = $81&1;
      if ($82) {
       label = 5;
       break;
      }
     }
    }
    $83 = HEAP32[$70>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    $84 = HEAP32[$1>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    $85 = HEAP32[$stack>>2]|0; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    $86 = (($85) + ($84<<2)|0); //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    HEAP32[$86>>2] = $83; //@line 61 "./libraries/Box2D/Common/b2GrowableStack.h"
    $87 = HEAP32[$1>>2]|0; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
    $88 = (($87) + 1)|0; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
    HEAP32[$1>>2] = $88; //@line 62 "./libraries/Box2D/Common/b2GrowableStack.h"
    $$old = $88;$$pre14 = $85;
   }
  }
  $$old11 = ($$old|0)>(0); //@line 174 "./libraries/Box2D/Collision/b2DynamicTree.h"
  if (!($$old11)) {
   $$pre = $$pre14;
   break;
  }
 }
 if ((label|0) == 5) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  $18 = HEAP32[$stack>>2]|0; //@line 40 "./libraries/Box2D/Common/b2GrowableStack.h"
  $19 = ($18|0)==($0|0); //@line 40 "./libraries/Box2D/Common/b2GrowableStack.h"
  if ($19) {
   ___resumeException($16|0);
   // unreachable;
  }
  __THREW__ = 0;
  invoke_vi(3816,($18|0)); //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
  $20 = __THREW__; __THREW__ = 0;
  $21 = $20&1;
  if ($21) {
   $22 = ___cxa_find_matching_catch_3(0|0)|0;
   $23 = tempRet0;
   ___clang_call_terminate($22); //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
   // unreachable; //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
  }
  HEAP32[$stack>>2] = 0; //@line 43 "./libraries/Box2D/Common/b2GrowableStack.h"
  ___resumeException($16|0);
  // unreachable;
 }
 else if ((label|0) == 14) {
  $$pre$pre = HEAP32[$stack>>2]|0; //@line 40 "./libraries/Box2D/Common/b2GrowableStack.h"
  $$pre = $$pre$pre;
 }
 $89 = ($$pre|0)==($0|0); //@line 40 "./libraries/Box2D/Common/b2GrowableStack.h"
 if ($89) {
  STACKTOP = sp;return; //@line 201 "./libraries/Box2D/Collision/b2DynamicTree.h"
 }
 __THREW__ = 0;
 invoke_vi(3816,($$pre|0)); //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
 $90 = __THREW__; __THREW__ = 0;
 $91 = $90&1;
 if ($91) {
  $92 = ___cxa_find_matching_catch_3(0|0)|0;
  $93 = tempRet0;
  ___clang_call_terminate($92); //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
  // unreachable; //@line 42 "./libraries/Box2D/Common/b2GrowableStack.h"
 }
 HEAP32[$stack>>2] = 0; //@line 43 "./libraries/Box2D/Common/b2GrowableStack.h"
 STACKTOP = sp;return; //@line 201 "./libraries/Box2D/Collision/b2DynamicTree.h"
}
function __ZN16b2ContactManager7AddPairEPvS0_($this,$proxyUserDataA,$proxyUserDataB) {
 $this = $this|0;
 $proxyUserDataA = $proxyUserDataA|0;
 $proxyUserDataB = $proxyUserDataB|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $edge$0 = 0, $edge$07 = 0, $edge$08 = 0, $or$cond = 0, $or$cond1 = 0, $or$cond2 = 0, $or$cond3 = 0, $or$cond4 = 0, $or$cond5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($proxyUserDataA)) + 16|0); //@line 182 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 182 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $2 = ((($proxyUserDataB)) + 16|0); //@line 183 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 183 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $4 = ((($proxyUserDataA)) + 20|0); //@line 185 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 185 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $6 = ((($proxyUserDataB)) + 20|0); //@line 186 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 186 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $8 = ((($1)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $9 = HEAP32[$8>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $10 = ((($3)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $11 = HEAP32[$10>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $12 = ($9|0)==($11|0); //@line 192 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if ($12) {
  return; //@line 296 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 }
 $13 = ((($11)) + 112|0); //@line 709 "./libraries/Box2D/Dynamics/b2Body.h"
 $edge$07 = HEAP32[$13>>2]|0; //@line 709 "./libraries/Box2D/Dynamics/b2Body.h"
 $14 = ($edge$07|0)==(0|0); //@line 201 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 L4: do {
  if (!($14)) {
   $edge$08 = $edge$07;
   while(1) {
    $15 = HEAP32[$edge$08>>2]|0; //@line 203 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    $16 = ($15|0)==($9|0); //@line 203 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    if ($16) {
     $17 = ((($edge$08)) + 4|0); //@line 205 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $18 = HEAP32[$17>>2]|0; //@line 205 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $19 = ((($18)) + 48|0); //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     $20 = HEAP32[$19>>2]|0; //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     $21 = ((($18)) + 52|0); //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     $22 = HEAP32[$21>>2]|0; //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     $23 = ((($18)) + 56|0); //@line 291 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     $24 = HEAP32[$23>>2]|0; //@line 291 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     $25 = ((($18)) + 60|0); //@line 301 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     $26 = HEAP32[$25>>2]|0; //@line 301 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
     $27 = ($20|0)==($1|0); //@line 210 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $28 = ($22|0)==($3|0); //@line 210 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $or$cond = $27 & $28; //@line 210 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $29 = ($24|0)==($5|0); //@line 210 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $or$cond1 = $or$cond & $29; //@line 210 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $30 = ($26|0)==($7|0); //@line 210 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $or$cond2 = $or$cond1 & $30; //@line 210 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     if ($or$cond2) {
      label = 24;
      break;
     }
     $31 = ($20|0)==($3|0); //@line 216 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $32 = ($22|0)==($1|0); //@line 216 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $or$cond3 = $31 & $32; //@line 216 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $33 = ($24|0)==($7|0); //@line 216 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $or$cond4 = $or$cond3 & $33; //@line 216 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $34 = ($26|0)==($5|0); //@line 216 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     $or$cond5 = $or$cond4 & $34; //@line 216 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
     if ($or$cond5) {
      label = 24;
      break;
     }
    }
    $35 = ((($edge$08)) + 12|0); //@line 223 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    $edge$0 = HEAP32[$35>>2]|0; //@line 709 "./libraries/Box2D/Dynamics/b2Body.h"
    $36 = ($edge$0|0)==(0|0); //@line 201 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
    if ($36) {
     break L4;
    } else {
     $edge$08 = $edge$0;
    }
   }
   if ((label|0) == 24) {
    return; //@line 296 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
   }
  }
 } while(0);
 $37 = (__ZNK6b2Body13ShouldCollideEPKS_($11,$9)|0); //@line 227 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if (!($37)) {
  return; //@line 296 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 }
 $38 = ((($this)) + 68|0); //@line 233 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $39 = HEAP32[$38>>2]|0; //@line 233 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $40 = ($39|0)==(0|0); //@line 233 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if (!($40)) {
  $41 = HEAP32[$39>>2]|0; //@line 233 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  $42 = ((($41)) + 8|0); //@line 233 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  $43 = HEAP32[$42>>2]|0; //@line 233 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  $44 = (FUNCTION_TABLE_iiii[$43 & 8191]($39,$1,$3)|0); //@line 233 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  if (!($44)) {
   return; //@line 296 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  }
 }
 $45 = ((($this)) + 76|0); //@line 239 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $46 = HEAP32[$45>>2]|0; //@line 239 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $47 = (__ZN9b2Contact6CreateEP9b2FixtureiS1_iP16b2BlockAllocator($1,$5,$3,$7,$46)|0); //@line 239 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $48 = ($47|0)==(0|0); //@line 240 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if ($48) {
  return; //@line 296 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 }
 $49 = ((($47)) + 48|0); //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $50 = HEAP32[$49>>2]|0; //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $51 = ((($47)) + 52|0); //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $52 = HEAP32[$51>>2]|0; //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
 $53 = ((($50)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $54 = HEAP32[$53>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $55 = ((($52)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $56 = HEAP32[$55>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $57 = ((($47)) + 8|0); //@line 254 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 HEAP32[$57>>2] = 0; //@line 254 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $58 = ((($this)) + 60|0); //@line 255 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $59 = HEAP32[$58>>2]|0; //@line 255 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $60 = ((($47)) + 12|0); //@line 255 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 HEAP32[$60>>2] = $59; //@line 255 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $61 = ($59|0)==(0); //@line 256 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if (!($61)) {
  $62 = $59; //@line 256 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  $63 = ((($62)) + 8|0); //@line 258 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  HEAP32[$63>>2] = $47; //@line 258 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 }
 HEAP32[$58>>2] = $47; //@line 260 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $64 = ((($47)) + 16|0); //@line 265 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $65 = ((($47)) + 20|0); //@line 265 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 HEAP32[$65>>2] = $47; //@line 265 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 HEAP32[$64>>2] = $56; //@line 266 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $66 = ((($47)) + 24|0); //@line 268 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 HEAP32[$66>>2] = 0; //@line 268 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $67 = ((($54)) + 112|0); //@line 269 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $68 = HEAP32[$67>>2]|0; //@line 269 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $69 = ((($47)) + 28|0); //@line 269 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 HEAP32[$69>>2] = $68; //@line 269 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $70 = ($68|0)==(0); //@line 270 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if (!($70)) {
  $71 = $68; //@line 270 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  $72 = ((($71)) + 8|0); //@line 272 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  HEAP32[$72>>2] = $64; //@line 272 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 }
 HEAP32[$67>>2] = $64; //@line 274 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $73 = ((($47)) + 32|0); //@line 277 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $74 = ((($47)) + 36|0); //@line 277 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 HEAP32[$74>>2] = $47; //@line 277 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 HEAP32[$73>>2] = $54; //@line 278 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $75 = ((($47)) + 40|0); //@line 280 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 HEAP32[$75>>2] = 0; //@line 280 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $76 = ((($56)) + 112|0); //@line 281 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $77 = HEAP32[$76>>2]|0; //@line 281 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $78 = ((($47)) + 44|0); //@line 281 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 HEAP32[$78>>2] = $77; //@line 281 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $79 = ($77|0)==(0); //@line 282 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 if (!($79)) {
  $80 = $77; //@line 282 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  $81 = ((($80)) + 8|0); //@line 284 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
  HEAP32[$81>>2] = $73; //@line 284 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 }
 HEAP32[$76>>2] = $73; //@line 286 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $82 = ((($50)) + 38|0); //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $83 = HEAP8[$82>>0]|0; //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $84 = ($83<<24>>24)==(0); //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
 if ($84) {
  $85 = ((($52)) + 38|0); //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $86 = HEAP8[$85>>0]|0; //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $87 = ($86<<24>>24)==(0); //@line 255 "./libraries/Box2D/Dynamics/b2Fixture.h"
  if ($87) {
   $88 = ((($54)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   $89 = HEAP16[$88>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   $90 = $89&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   $91 = $90 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   $92 = ($91|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   if ($92) {
    $93 = $90 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
    $94 = $93&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAP16[$88>>1] = $94; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
    $95 = ((($54)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAPF32[$95>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   }
   $96 = ((($56)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   $97 = HEAP16[$96>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   $98 = $97&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   $99 = $98 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   $100 = ($99|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
   if ($100) {
    $101 = $98 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
    $102 = $101&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAP16[$96>>1] = $102; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
    $103 = ((($56)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
    HEAPF32[$103>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
   }
  }
 }
 $104 = ((($this)) + 64|0); //@line 295 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $105 = HEAP32[$104>>2]|0; //@line 295 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 $106 = (($105) + 1)|0; //@line 295 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 HEAP32[$104>>2] = $106; //@line 295 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
 return; //@line 296 "libraries/Box2D/Dynamics/b2ContactManager.cpp"
}
function __ZN9b2FixtureC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 36 "./libraries/Box2D/Dynamics/b2Fixture.h"
 HEAP16[$0>>1] = 1; //@line 36 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $1 = ((($this)) + 34|0); //@line 37 "./libraries/Box2D/Dynamics/b2Fixture.h"
 HEAP16[$1>>1] = -1; //@line 37 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $2 = ((($this)) + 36|0); //@line 38 "./libraries/Box2D/Dynamics/b2Fixture.h"
 HEAP16[$2>>1] = 0; //@line 38 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $3 = ((($this)) + 40|0); //@line 32 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 HEAP32[$3>>2] = 0; //@line 32 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $4 = ((($this)) + 24|0); //@line 35 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 HEAP32[$4>>2] = 0; //@line 35 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $5 = ((($this)) + 28|0); //@line 36 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 HEAP32[$5>>2] = 0; //@line 36 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 ;HEAP32[$this>>2]=0|0;HEAP32[$this+4>>2]=0|0;HEAP32[$this+8>>2]=0|0;HEAP32[$this+12>>2]=0|0; //@line 33 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 return; //@line 39 "libraries/Box2D/Dynamics/b2Fixture.cpp"
}
function __ZN9b2Fixture6CreateEP16b2BlockAllocatorP6b2BodyPK12b2FixtureDef($this,$allocator,$body,$def) {
 $this = $this|0;
 $allocator = $allocator|0;
 $body = $body|0;
 $def = $def|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($def)) + 4|0); //@line 43 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 43 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $2 = ((($this)) + 40|0); //@line 43 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 HEAP32[$2>>2] = $1; //@line 43 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $3 = ((($def)) + 8|0); //@line 44 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 44 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $5 = ((($this)) + 16|0); //@line 44 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 HEAP32[$5>>2] = $4; //@line 44 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $6 = ((($def)) + 12|0); //@line 45 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 45 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $8 = ((($this)) + 20|0); //@line 45 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 HEAP32[$8>>2] = $7; //@line 45 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $9 = ((($this)) + 8|0); //@line 47 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 HEAP32[$9>>2] = $body; //@line 47 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $10 = ((($this)) + 4|0); //@line 48 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 HEAP32[$10>>2] = 0; //@line 48 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $11 = ((($this)) + 32|0); //@line 50 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $12 = ((($def)) + 22|0); //@line 50 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 ;HEAP16[$11>>1]=HEAP16[$12>>1]|0;HEAP16[$11+2>>1]=HEAP16[$12+2>>1]|0;HEAP16[$11+4>>1]=HEAP16[$12+4>>1]|0; //@line 50 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $13 = ((($def)) + 20|0); //@line 52 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $14 = HEAP8[$13>>0]|0; //@line 52 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $15 = ((($this)) + 38|0); //@line 52 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 HEAP8[$15>>0] = $14; //@line 52 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $16 = HEAP32[$def>>2]|0; //@line 54 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 54 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $18 = ((($17)) + 8|0); //@line 54 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 54 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $20 = (FUNCTION_TABLE_iii[$19 & 8191]($16,$allocator)|0); //@line 54 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $21 = ((($this)) + 12|0); //@line 54 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 HEAP32[$21>>2] = $20; //@line 54 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $22 = HEAP32[$20>>2]|0; //@line 57 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $23 = ((($22)) + 12|0); //@line 57 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 57 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $25 = (FUNCTION_TABLE_ii[$24 & 8191]($20)|0); //@line 57 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $26 = ($25*28)|0; //@line 58 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $27 = (__ZN16b2BlockAllocator8AllocateEi($allocator,$26)|0); //@line 58 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $28 = ((($this)) + 24|0); //@line 58 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 HEAP32[$28>>2] = $27; //@line 58 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $29 = ($25|0)>(0); //@line 59 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 if ($29) {
  $i$01 = 0;
 } else {
  $30 = ((($this)) + 28|0); //@line 64 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  HEAP32[$30>>2] = 0; //@line 64 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $31 = ((($def)) + 16|0); //@line 66 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $32 = HEAP32[$31>>2]|0; //@line 66 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  HEAP32[$this>>2] = $32; //@line 66 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  return; //@line 67 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 }
 while(1) {
  $33 = (((($27) + (($i$01*28)|0)|0)) + 16|0); //@line 61 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  HEAP32[$33>>2] = 0; //@line 61 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $34 = (((($27) + (($i$01*28)|0)|0)) + 24|0); //@line 62 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  HEAP32[$34>>2] = -1; //@line 62 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $35 = (($i$01) + 1)|0; //@line 59 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $exitcond = ($35|0)==($25|0); //@line 59 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  if ($exitcond) {
   break;
  } else {
   $i$01 = $35;
  }
 }
 $30 = ((($this)) + 28|0); //@line 64 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 HEAP32[$30>>2] = 0; //@line 64 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $31 = ((($def)) + 16|0); //@line 66 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $32 = HEAP32[$31>>2]|0; //@line 66 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 HEAP32[$this>>2] = $32; //@line 66 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 return; //@line 67 "libraries/Box2D/Dynamics/b2Fixture.cpp"
}
function __ZN9b2Fixture7DestroyEP16b2BlockAllocator($this,$allocator) {
 $this = $this|0;
 $allocator = $allocator|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 72 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 72 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $2 = ($1|0)==(0); //@line 72 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 __Z10loveAssertbPKc($2,446504); //@line 72 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $3 = ((($this)) + 12|0); //@line 75 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 75 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 75 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $6 = ((($5)) + 12|0); //@line 75 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 75 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $8 = (FUNCTION_TABLE_ii[$7 & 8191]($4)|0); //@line 75 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $9 = ((($this)) + 24|0); //@line 76 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 76 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $11 = ($8*28)|0; //@line 76 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 __ZN16b2BlockAllocator4FreeEPvi($allocator,$10,$11); //@line 76 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 HEAP32[$9>>2] = 0; //@line 77 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $12 = HEAP32[$3>>2]|0; //@line 80 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $13 = ((($12)) + 4|0); //@line 80 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 80 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 switch ($14|0) {
 case 0:  {
  $15 = HEAP32[$12>>2]|0; //@line 85 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 85 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  FUNCTION_TABLE_vi[$16 & 8191]($12); //@line 85 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  __ZN16b2BlockAllocator4FreeEPvi($allocator,$12,20); //@line 86 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  HEAP32[$3>>2] = 0; //@line 119 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  return; //@line 120 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  break;
 }
 case 1:  {
  $17 = HEAP32[$12>>2]|0; //@line 93 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $18 = HEAP32[$17>>2]|0; //@line 93 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  FUNCTION_TABLE_vi[$18 & 8191]($12); //@line 93 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  __ZN16b2BlockAllocator4FreeEPvi($allocator,$12,48); //@line 94 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  HEAP32[$3>>2] = 0; //@line 119 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  return; //@line 120 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  break;
 }
 case 2:  {
  $19 = HEAP32[$12>>2]|0; //@line 101 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $20 = HEAP32[$19>>2]|0; //@line 101 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  FUNCTION_TABLE_vi[$20 & 8191]($12); //@line 101 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  __ZN16b2BlockAllocator4FreeEPvi($allocator,$12,152); //@line 102 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  HEAP32[$3>>2] = 0; //@line 119 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  return; //@line 120 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  break;
 }
 case 3:  {
  $21 = HEAP32[$12>>2]|0; //@line 109 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $22 = HEAP32[$21>>2]|0; //@line 109 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  FUNCTION_TABLE_vi[$22 & 8191]($12); //@line 109 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  __ZN16b2BlockAllocator4FreeEPvi($allocator,$12,40); //@line 110 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  HEAP32[$3>>2] = 0; //@line 119 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  return; //@line 120 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  break;
 }
 default: {
  __Z10loveAssertbPKc(0,622525); //@line 115 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  HEAP32[$3>>2] = 0; //@line 119 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  return; //@line 120 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 }
 }
}
function __ZN9b2Fixture13CreateProxiesEP12b2BroadPhaseRK11b2Transform($this,$broadPhase,$xf) {
 $this = $this|0;
 $broadPhase = $broadPhase|0;
 $xf = $xf|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 124 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 124 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $2 = ($1|0)==(0); //@line 124 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 __Z10loveAssertbPKc($2,446504); //@line 124 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $3 = ((($this)) + 12|0); //@line 127 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 127 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 127 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $6 = ((($5)) + 12|0); //@line 127 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 127 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $8 = (FUNCTION_TABLE_ii[$7 & 8191]($4)|0); //@line 127 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 HEAP32[$0>>2] = $8; //@line 127 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $9 = ($8|0)>(0); //@line 129 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 if (!($9)) {
  return; //@line 137 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 }
 $10 = ((($this)) + 24|0); //@line 131 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $i$01 = 0;
 while(1) {
  $11 = HEAP32[$10>>2]|0; //@line 131 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $12 = (($11) + (($i$01*28)|0)|0); //@line 131 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $13 = HEAP32[$3>>2]|0; //@line 132 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 132 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $15 = ((($14)) + 24|0); //@line 132 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 132 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  FUNCTION_TABLE_viiii[$16 & 8191]($13,$12,$xf,$i$01); //@line 132 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $17 = (__ZN12b2BroadPhase11CreateProxyERK6b2AABBPv($broadPhase,$12,$12)|0); //@line 133 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $18 = (((($11) + (($i$01*28)|0)|0)) + 24|0); //@line 133 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  HEAP32[$18>>2] = $17; //@line 133 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $19 = (((($11) + (($i$01*28)|0)|0)) + 16|0); //@line 134 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  HEAP32[$19>>2] = $this; //@line 134 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $20 = (((($11) + (($i$01*28)|0)|0)) + 20|0); //@line 135 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  HEAP32[$20>>2] = $i$01; //@line 135 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $21 = (($i$01) + 1)|0; //@line 129 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $22 = HEAP32[$0>>2]|0; //@line 129 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $23 = ($21|0)<($22|0); //@line 129 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  if ($23) {
   $i$01 = $21;
  } else {
   break;
  }
 }
 return; //@line 137 "libraries/Box2D/Dynamics/b2Fixture.cpp"
}
function __ZN9b2Fixture14DestroyProxiesEP12b2BroadPhase($this,$broadPhase) {
 $this = $this|0;
 $broadPhase = $broadPhase|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0); //@line 142 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 142 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $2 = ($1|0)>(0); //@line 142 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 if (!($2)) {
  HEAP32[$0>>2] = 0; //@line 149 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  return; //@line 150 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 }
 $3 = ((($this)) + 24|0); //@line 144 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $i$01 = 0;
 while(1) {
  $4 = HEAP32[$3>>2]|0; //@line 144 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $5 = (((($4) + (($i$01*28)|0)|0)) + 24|0); //@line 145 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 145 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  __ZN12b2BroadPhase12DestroyProxyEi($broadPhase,$6); //@line 145 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  HEAP32[$5>>2] = -1; //@line 146 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $7 = (($i$01) + 1)|0; //@line 142 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $8 = HEAP32[$0>>2]|0; //@line 142 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $9 = ($7|0)<($8|0); //@line 142 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  if ($9) {
   $i$01 = $7;
  } else {
   break;
  }
 }
 HEAP32[$0>>2] = 0; //@line 149 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 return; //@line 150 "libraries/Box2D/Dynamics/b2Fixture.cpp"
}
function __ZN9b2Fixture11SynchronizeEP12b2BroadPhaseRK11b2TransformS4_($this,$broadPhase,$transform1,$transform2) {
 $this = $this|0;
 $broadPhase = $broadPhase|0;
 $transform1 = $transform1|0;
 $transform2 = $transform2|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0.0, $44 = 0;
 var $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $aabb1 = 0, $aabb2 = 0, $displacement = 0;
 var $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $aabb1 = sp + 24|0;
 $aabb2 = sp + 8|0;
 $displacement = sp;
 $0 = ((($this)) + 28|0); //@line 154 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 159 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $2 = ($1|0)>(0); //@line 154 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 if (!($2)) {
  STACKTOP = sp;return; //@line 174 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 }
 $3 = ((($this)) + 24|0); //@line 161 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $4 = ((($this)) + 12|0); //@line 165 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $5 = ((($aabb1)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($aabb2)) + 4|0); //@line 639 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($aabb1)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($aabb2)) + 8|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($aabb1)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($aabb2)) + 12|0); //@line 650 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($transform2)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $12 = ((($transform1)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($displacement)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $i$01 = 0;
 while(1) {
  $14 = HEAP32[$3>>2]|0; //@line 161 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $15 = HEAP32[$4>>2]|0; //@line 165 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $16 = HEAP32[$15>>2]|0; //@line 165 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $17 = ((($16)) + 24|0); //@line 165 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $18 = HEAP32[$17>>2]|0; //@line 165 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $19 = (((($14) + (($i$01*28)|0)|0)) + 20|0); //@line 165 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $20 = HEAP32[$19>>2]|0; //@line 165 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  FUNCTION_TABLE_viiii[$18 & 8191]($15,$aabb1,$transform1,$20); //@line 165 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $21 = HEAP32[$4>>2]|0; //@line 166 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $22 = HEAP32[$21>>2]|0; //@line 166 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $23 = ((($22)) + 24|0); //@line 166 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $24 = HEAP32[$23>>2]|0; //@line 166 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $25 = HEAP32[$19>>2]|0; //@line 166 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  FUNCTION_TABLE_viiii[$24 & 8191]($21,$aabb2,$transform2,$25); //@line 166 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $26 = (($14) + (($i$01*28)|0)|0); //@line 168 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $27 = +HEAPF32[$aabb1>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $28 = +HEAPF32[$aabb2>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $29 = $27 < $28; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $30 = $29 ? $27 : $28; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $31 = +HEAPF32[$5>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $32 = +HEAPF32[$6>>2]; //@line 639 "./libraries/Box2D/Common/b2Math.h"
  $33 = $31 < $32; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  $34 = $33 ? $31 : $32; //@line 634 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$26>>2] = $30; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  $35 = (((($14) + (($i$01*28)|0)|0)) + 4|0); //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$35>>2] = $34; //@line 197 "./libraries/Box2D/Collision/b2Collision.h"
  $36 = +HEAPF32[$7>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $37 = +HEAPF32[$8>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $38 = $36 > $37; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $39 = $38 ? $36 : $37; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $40 = +HEAPF32[$9>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $41 = +HEAPF32[$10>>2]; //@line 650 "./libraries/Box2D/Common/b2Math.h"
  $42 = $40 > $41; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $43 = $42 ? $40 : $41; //@line 645 "./libraries/Box2D/Common/b2Math.h"
  $44 = (((($14) + (($i$01*28)|0)|0)) + 8|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$44>>2] = $39; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  $45 = (((($14) + (($i$01*28)|0)|0)) + 12|0); //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  HEAPF32[$45>>2] = $43; //@line 198 "./libraries/Box2D/Collision/b2Collision.h"
  $46 = +HEAPF32[$transform2>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $47 = +HEAPF32[$transform1>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $48 = $46 - $47; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $49 = +HEAPF32[$11>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $50 = +HEAPF32[$12>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $51 = $49 - $50; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$displacement>>2] = $48; //@line 59 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$13>>2] = $51; //@line 59 "./libraries/Box2D/Common/b2Math.h"
  $52 = (((($14) + (($i$01*28)|0)|0)) + 24|0); //@line 172 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $53 = HEAP32[$52>>2]|0; //@line 172 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  __ZN12b2BroadPhase9MoveProxyEiRK6b2AABBRK6b2Vec2($broadPhase,$53,$26,$displacement); //@line 172 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $54 = (($i$01) + 1)|0; //@line 159 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $55 = HEAP32[$0>>2]|0; //@line 159 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $56 = ($54|0)<($55|0); //@line 159 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  if ($56) {
   $i$01 = $54;
  } else {
   break;
  }
 }
 STACKTOP = sp;return; //@line 174 "libraries/Box2D/Dynamics/b2Fixture.cpp"
}
function __ZN9b2Fixture13SetFilterDataERK8b2Filter($this,$filter) {
 $this = $this|0;
 $filter = $filter|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $edge$0$i = 0, $edge$03$i = 0, $edge$04$i = 0, $i$02$i = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 178 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 ;HEAP16[$0>>1]=HEAP16[$filter>>1]|0;HEAP16[$0+2>>1]=HEAP16[$filter+2>>1]|0;HEAP16[$0+4>>1]=HEAP16[$filter+4>>1]|0; //@line 178 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $1 = ((($this)) + 8|0); //@line 185 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 205 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $3 = ($2|0)==(0|0); //@line 185 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 if ($3) {
  return; //@line 181 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 }
 $4 = ((($2)) + 112|0); //@line 709 "./libraries/Box2D/Dynamics/b2Body.h"
 $edge$03$i = HEAP32[$4>>2]|0; //@line 709 "./libraries/Box2D/Dynamics/b2Body.h"
 $5 = ($edge$03$i|0)==(0|0); //@line 192 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 if (!($5)) {
  $edge$04$i = $edge$03$i;
  while(1) {
   $6 = ((($edge$04$i)) + 4|0); //@line 194 "libraries/Box2D/Dynamics/b2Fixture.cpp"
   $7 = HEAP32[$6>>2]|0; //@line 194 "libraries/Box2D/Dynamics/b2Fixture.cpp"
   $8 = ((($7)) + 48|0); //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
   $9 = HEAP32[$8>>2]|0; //@line 276 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
   $10 = ((($7)) + 52|0); //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
   $11 = HEAP32[$10>>2]|0; //@line 286 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
   $12 = ($9|0)==($this|0); //@line 197 "libraries/Box2D/Dynamics/b2Fixture.cpp"
   $13 = ($11|0)==($this|0); //@line 197 "libraries/Box2D/Dynamics/b2Fixture.cpp"
   $or$cond$i = $12 | $13; //@line 197 "libraries/Box2D/Dynamics/b2Fixture.cpp"
   if ($or$cond$i) {
    $14 = ((($7)) + 4|0); //@line 306 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
    $15 = HEAP32[$14>>2]|0; //@line 306 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
    $16 = $15 | 8; //@line 306 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
    HEAP32[$14>>2] = $16; //@line 306 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
   }
   $17 = ((($edge$04$i)) + 12|0); //@line 202 "libraries/Box2D/Dynamics/b2Fixture.cpp"
   $edge$0$i = HEAP32[$17>>2]|0; //@line 709 "./libraries/Box2D/Dynamics/b2Body.h"
   $18 = ($edge$0$i|0)==(0|0); //@line 192 "libraries/Box2D/Dynamics/b2Fixture.cpp"
   if ($18) {
    break;
   } else {
    $edge$04$i = $edge$0$i;
   }
  }
 }
 $19 = ((($2)) + 88|0); //@line 852 "./libraries/Box2D/Dynamics/b2Body.h"
 $20 = HEAP32[$19>>2]|0; //@line 852 "./libraries/Box2D/Dynamics/b2Body.h"
 $21 = ($20|0)==(0|0); //@line 207 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 if ($21) {
  return; //@line 181 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 }
 $22 = ((($20)) + 102872|0); //@line 213 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $23 = ((($this)) + 28|0); //@line 214 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 214 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $25 = ($24|0)>(0); //@line 214 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 if (!($25)) {
  return; //@line 181 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 }
 $26 = ((($this)) + 24|0); //@line 216 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $i$02$i = 0;
 while(1) {
  $27 = HEAP32[$26>>2]|0; //@line 216 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $28 = (((($27) + (($i$02$i*28)|0)|0)) + 24|0); //@line 216 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $29 = HEAP32[$28>>2]|0; //@line 216 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  __ZN12b2BroadPhase10TouchProxyEi($22,$29); //@line 216 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $30 = (($i$02$i) + 1)|0; //@line 214 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $31 = HEAP32[$23>>2]|0; //@line 214 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  $32 = ($30|0)<($31|0); //@line 214 "libraries/Box2D/Dynamics/b2Fixture.cpp"
  if ($32) {
   $i$02$i = $30;
  } else {
   break;
  }
 }
 return; //@line 181 "libraries/Box2D/Dynamics/b2Fixture.cpp"
}
function __ZN9b2Fixture9SetSensorEb($this,$sensor) {
 $this = $this|0;
 $sensor = $sensor|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $sensor&1; //@line 222 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $1 = ((($this)) + 38|0); //@line 222 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $2 = HEAP8[$1>>0]|0; //@line 222 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $3 = $2&255; //@line 222 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $4 = ($0|0)==($3|0); //@line 222 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 if ($4) {
  return; //@line 227 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 }
 $5 = ((($this)) + 8|0); //@line 224 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 224 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 $7 = ((($6)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $8 = HEAP16[$7>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $9 = $8&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $10 = $9 & 2; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 $11 = ($10|0)==(0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
 if ($11) {
  $12 = $9 | 2; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $13 = $12&65535; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAP16[$7>>1] = $13; //@line 639 "./libraries/Box2D/Dynamics/b2Body.h"
  $14 = ((($6)) + 144|0); //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
  HEAPF32[$14>>2] = 0.0; //@line 640 "./libraries/Box2D/Dynamics/b2Body.h"
 }
 $15 = $sensor&1; //@line 225 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 HEAP8[$1>>0] = $15; //@line 225 "libraries/Box2D/Dynamics/b2Fixture.cpp"
 return; //@line 227 "libraries/Box2D/Dynamics/b2Fixture.cpp"
}
function __ZN8b2IslandC2EiiiP16b2StackAllocatorP17b2ContactListener($this,$bodyCapacity,$contactCapacity,$jointCapacity,$allocator,$listener) {
 $this = $this|0;
 $bodyCapacity = $bodyCapacity|0;
 $contactCapacity = $contactCapacity|0;
 $jointCapacity = $jointCapacity|0;
 $allocator = $allocator|0;
 $listener = $listener|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0); //@line 155 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$0>>2] = $bodyCapacity; //@line 155 "libraries/Box2D/Dynamics/b2Island.cpp"
 $1 = ((($this)) + 44|0); //@line 156 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$1>>2] = $contactCapacity; //@line 156 "libraries/Box2D/Dynamics/b2Island.cpp"
 $2 = ((($this)) + 48|0); //@line 157 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$2>>2] = $jointCapacity; //@line 157 "libraries/Box2D/Dynamics/b2Island.cpp"
 $3 = ((($this)) + 28|0); //@line 158 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$3>>2] = 0; //@line 158 "libraries/Box2D/Dynamics/b2Island.cpp"
 $4 = ((($this)) + 36|0); //@line 159 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$4>>2] = 0; //@line 159 "libraries/Box2D/Dynamics/b2Island.cpp"
 $5 = ((($this)) + 32|0); //@line 160 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$5>>2] = 0; //@line 160 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$this>>2] = $allocator; //@line 162 "libraries/Box2D/Dynamics/b2Island.cpp"
 $6 = ((($this)) + 4|0); //@line 163 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$6>>2] = $listener; //@line 163 "libraries/Box2D/Dynamics/b2Island.cpp"
 $7 = $bodyCapacity << 2; //@line 165 "libraries/Box2D/Dynamics/b2Island.cpp"
 $8 = (__ZN16b2StackAllocator8AllocateEi($allocator,$7)|0); //@line 165 "libraries/Box2D/Dynamics/b2Island.cpp"
 $9 = ((($this)) + 8|0); //@line 165 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$9>>2] = $8; //@line 165 "libraries/Box2D/Dynamics/b2Island.cpp"
 $10 = HEAP32[$this>>2]|0; //@line 166 "libraries/Box2D/Dynamics/b2Island.cpp"
 $11 = $contactCapacity << 2; //@line 166 "libraries/Box2D/Dynamics/b2Island.cpp"
 $12 = (__ZN16b2StackAllocator8AllocateEi($10,$11)|0); //@line 166 "libraries/Box2D/Dynamics/b2Island.cpp"
 $13 = ((($this)) + 12|0); //@line 166 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$13>>2] = $12; //@line 166 "libraries/Box2D/Dynamics/b2Island.cpp"
 $14 = HEAP32[$this>>2]|0; //@line 167 "libraries/Box2D/Dynamics/b2Island.cpp"
 $15 = $jointCapacity << 2; //@line 167 "libraries/Box2D/Dynamics/b2Island.cpp"
 $16 = (__ZN16b2StackAllocator8AllocateEi($14,$15)|0); //@line 167 "libraries/Box2D/Dynamics/b2Island.cpp"
 $17 = ((($this)) + 16|0); //@line 167 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$17>>2] = $16; //@line 167 "libraries/Box2D/Dynamics/b2Island.cpp"
 $18 = HEAP32[$this>>2]|0; //@line 169 "libraries/Box2D/Dynamics/b2Island.cpp"
 $19 = HEAP32[$0>>2]|0; //@line 169 "libraries/Box2D/Dynamics/b2Island.cpp"
 $20 = ($19*12)|0; //@line 169 "libraries/Box2D/Dynamics/b2Island.cpp"
 $21 = (__ZN16b2StackAllocator8AllocateEi($18,$20)|0); //@line 169 "libraries/Box2D/Dynamics/b2Island.cpp"
 $22 = ((($this)) + 24|0); //@line 169 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$22>>2] = $21; //@line 169 "libraries/Box2D/Dynamics/b2Island.cpp"
 $23 = HEAP32[$this>>2]|0; //@line 170 "libraries/Box2D/Dynamics/b2Island.cpp"
 $24 = HEAP32[$0>>2]|0; //@line 170 "libraries/Box2D/Dynamics/b2Island.cpp"
 $25 = ($24*12)|0; //@line 170 "libraries/Box2D/Dynamics/b2Island.cpp"
 $26 = (__ZN16b2StackAllocator8AllocateEi($23,$25)|0); //@line 170 "libraries/Box2D/Dynamics/b2Island.cpp"
 $27 = ((($this)) + 20|0); //@line 170 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$27>>2] = $26; //@line 170 "libraries/Box2D/Dynamics/b2Island.cpp"
 return; //@line 171 "libraries/Box2D/Dynamics/b2Island.cpp"
}
function __ZN8b2IslandD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 176 "libraries/Box2D/Dynamics/b2Island.cpp"
 $1 = ((($this)) + 20|0); //@line 176 "libraries/Box2D/Dynamics/b2Island.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 176 "libraries/Box2D/Dynamics/b2Island.cpp"
 __THREW__ = 0;
 invoke_vii(3821,($0|0),($2|0)); //@line 176 "libraries/Box2D/Dynamics/b2Island.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  $5 = HEAP32[$this>>2]|0; //@line 177 "libraries/Box2D/Dynamics/b2Island.cpp"
  $6 = ((($this)) + 24|0); //@line 177 "libraries/Box2D/Dynamics/b2Island.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 177 "libraries/Box2D/Dynamics/b2Island.cpp"
  __THREW__ = 0;
  invoke_vii(3821,($5|0),($7|0)); //@line 177 "libraries/Box2D/Dynamics/b2Island.cpp"
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if (!($9)) {
   $10 = HEAP32[$this>>2]|0; //@line 178 "libraries/Box2D/Dynamics/b2Island.cpp"
   $11 = ((($this)) + 16|0); //@line 178 "libraries/Box2D/Dynamics/b2Island.cpp"
   $12 = HEAP32[$11>>2]|0; //@line 178 "libraries/Box2D/Dynamics/b2Island.cpp"
   __THREW__ = 0;
   invoke_vii(3821,($10|0),($12|0)); //@line 178 "libraries/Box2D/Dynamics/b2Island.cpp"
   $13 = __THREW__; __THREW__ = 0;
   $14 = $13&1;
   if (!($14)) {
    $15 = HEAP32[$this>>2]|0; //@line 179 "libraries/Box2D/Dynamics/b2Island.cpp"
    $16 = ((($this)) + 12|0); //@line 179 "libraries/Box2D/Dynamics/b2Island.cpp"
    $17 = HEAP32[$16>>2]|0; //@line 179 "libraries/Box2D/Dynamics/b2Island.cpp"
    __THREW__ = 0;
    invoke_vii(3821,($15|0),($17|0)); //@line 179 "libraries/Box2D/Dynamics/b2Island.cpp"
    $18 = __THREW__; __THREW__ = 0;
    $19 = $18&1;
    if (!($19)) {
     $20 = HEAP32[$this>>2]|0; //@line 180 "libraries/Box2D/Dynamics/b2Island.cpp"
     $21 = ((($this)) + 8|0); //@line 180 "libraries/Box2D/Dynamics/b2Island.cpp"
     $22 = HEAP32[$21>>2]|0; //@line 180 "libraries/Box2D/Dynamics/b2Island.cpp"
     __THREW__ = 0;
     invoke_vii(3821,($20|0),($22|0)); //@line 180 "libraries/Box2D/Dynamics/b2Island.cpp"
     $23 = __THREW__; __THREW__ = 0;
     $24 = $23&1;
     if (!($24)) {
      return; //@line 181 "libraries/Box2D/Dynamics/b2Island.cpp"
     }
    }
   }
  }
 }
 $25 = ___cxa_find_matching_catch_3(0|0)|0;
 $26 = tempRet0;
 ___clang_call_terminate($25); //@line 176 "libraries/Box2D/Dynamics/b2Island.cpp"
 // unreachable; //@line 176 "libraries/Box2D/Dynamics/b2Island.cpp"
}
function __ZN8b2Island5SolveEP9b2ProfileRK10b2TimeStepRK6b2Vec2b($this,$profile,$step,$gravity,$allowSleep) {
 $this = $this|0;
 $profile = $profile|0;
 $step = $step|0;
 $gravity = $gravity|0;
 $allowSleep = $allowSleep|0;
 var $$pre = 0, $$pre$phi79Z2D = 0, $$pre$phiZ2D = 0, $$pre77 = 0, $$pre78 = 0, $0 = 0.0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0.0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0.0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0.0, $189 = 0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0.0;
 var $201 = 0.0, $202 = 0.0, $203 = 0.0, $204 = 0.0, $205 = 0.0, $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0.0, $213 = 0, $214 = 0.0, $215 = 0.0, $216 = 0, $217 = 0, $218 = 0.0, $219 = 0.0;
 var $22 = 0, $220 = 0.0, $221 = 0.0, $222 = 0.0, $223 = 0.0, $224 = 0.0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0.0, $229 = 0.0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0;
 var $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0;
 var $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0.0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0;
 var $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0;
 var $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0.0, $305 = 0.0, $306 = 0, $307 = 0.0, $308 = 0, $309 = 0;
 var $31 = 0, $310 = 0.0, $311 = 0.0, $312 = 0, $313 = 0.0, $314 = 0.0, $315 = 0.0, $316 = 0.0, $317 = 0.0, $318 = 0.0, $319 = 0.0, $32 = 0, $320 = 0.0, $321 = 0.0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0;
 var $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0;
 var $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0;
 var $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0.0, $376 = 0.0, $377 = 0, $378 = 0, $379 = 0.0, $38 = 0, $380 = 0.0, $381 = 0;
 var $382 = 0.0, $383 = 0.0, $384 = 0.0, $385 = 0, $386 = 0, $387 = 0, $388 = 0.0, $389 = 0.0, $39 = 0, $390 = 0, $391 = 0.0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0;
 var $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $41 = 0.0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0.0;
 var $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0;
 var $70 = 0.0, $71 = 0.0, $72 = 0, $73 = 0.0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0, $86 = 0.0, $87 = 0, $88 = 0;
 var $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bitcast = 0.0, $brmerge$demorgan = 0, $brmerge$demorgan$not = 0, $brmerge7 = 0, $contactSolver = 0, $contactSolverDef = 0, $exitcond = 0, $i$02$i = 0;
 var $i$072 = 0, $i1$068 = 0, $i10$048 = 0, $i11$043 = 0, $i12$041 = 0, $i15$039 = 0, $i2$066 = 0, $i3$059 = 0, $i9$053 = 0, $impulse$i = 0, $j$01$i = 0, $j$063 = 0, $jointsOkay$0$off0$lcssa = 0, $jointsOkay$0$off047 = 0, $lpad$phi$index = 0, $lpad$phi$index7 = 0, $minSleepTime$0$lcssa = 0.0, $minSleepTime$040 = 0.0, $minSleepTime$2 = 0.0, $or$cond = 0;
 var $phitmp = 0, $positionSolved$0$off0$ = 0, $positionSolved$0$off0$$lcssa = 0, $positionSolved$0$off052 = 0, $positionSolved$2$off0 = 0, $solverData = 0, $timer = 0, $v$sroa$0$0 = 0, $v$sroa$8$0 = 0, $v6$sroa$0$0 = 0, $v6$sroa$8$0 = 0, $w$0 = 0.0, $w7$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 160|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $impulse$i = sp + 128|0;
 $timer = sp + 148|0;
 $solverData = sp + 96|0;
 $contactSolverDef = sp + 52|0;
 $contactSolver = sp;
 __ZN7b2TimerC2Ev($timer); //@line 185 "libraries/Box2D/Dynamics/b2Island.cpp"
 $0 = +HEAPF32[$step>>2]; //@line 187 "libraries/Box2D/Dynamics/b2Island.cpp"
 $1 = ((($this)) + 28|0); //@line 190 "libraries/Box2D/Dynamics/b2Island.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 190 "libraries/Box2D/Dynamics/b2Island.cpp"
 $3 = ($2|0)>(0); //@line 190 "libraries/Box2D/Dynamics/b2Island.cpp"
 if ($3) {
  $4 = ((($this)) + 8|0); //@line 192 "libraries/Box2D/Dynamics/b2Island.cpp"
  $5 = ((($gravity)) + 4|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $6 = ((($this)) + 20|0); //@line 220 "libraries/Box2D/Dynamics/b2Island.cpp"
  $7 = ((($this)) + 24|0); //@line 222 "libraries/Box2D/Dynamics/b2Island.cpp"
  $i$072 = 0;
  while(1) {
   $24 = HEAP32[$4>>2]|0; //@line 192 "libraries/Box2D/Dynamics/b2Island.cpp"
   $25 = (($24) + ($i$072<<2)|0); //@line 192 "libraries/Box2D/Dynamics/b2Island.cpp"
   $26 = HEAP32[$25>>2]|0; //@line 192 "libraries/Box2D/Dynamics/b2Island.cpp"
   $27 = ((($26)) + 44|0); //@line 194 "libraries/Box2D/Dynamics/b2Island.cpp"
   $28 = $27; //@line 194 "libraries/Box2D/Dynamics/b2Island.cpp"
   $29 = $28; //@line 194 "libraries/Box2D/Dynamics/b2Island.cpp"
   $30 = HEAP32[$29>>2]|0; //@line 194 "libraries/Box2D/Dynamics/b2Island.cpp"
   $31 = (($28) + 4)|0; //@line 194 "libraries/Box2D/Dynamics/b2Island.cpp"
   $32 = $31; //@line 194 "libraries/Box2D/Dynamics/b2Island.cpp"
   $33 = HEAP32[$32>>2]|0; //@line 194 "libraries/Box2D/Dynamics/b2Island.cpp"
   $34 = ((($26)) + 56|0); //@line 195 "libraries/Box2D/Dynamics/b2Island.cpp"
   $35 = HEAP32[$34>>2]|0; //@line 195 "libraries/Box2D/Dynamics/b2Island.cpp"
   $36 = ((($26)) + 64|0); //@line 196 "libraries/Box2D/Dynamics/b2Island.cpp"
   $37 = HEAP32[$36>>2]|0; //@line 196 "libraries/Box2D/Dynamics/b2Island.cpp"
   $38 = ((($26)) + 68|0); //@line 196 "libraries/Box2D/Dynamics/b2Island.cpp"
   $39 = HEAP32[$38>>2]|0; //@line 196 "libraries/Box2D/Dynamics/b2Island.cpp"
   $40 = ((($26)) + 72|0); //@line 197 "libraries/Box2D/Dynamics/b2Island.cpp"
   $41 = +HEAPF32[$40>>2]; //@line 197 "libraries/Box2D/Dynamics/b2Island.cpp"
   $42 = ((($26)) + 36|0); //@line 200 "libraries/Box2D/Dynamics/b2Island.cpp"
   $43 = $42; //@line 200 "libraries/Box2D/Dynamics/b2Island.cpp"
   $44 = $43; //@line 200 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$44>>2] = $30; //@line 200 "libraries/Box2D/Dynamics/b2Island.cpp"
   $45 = (($43) + 4)|0; //@line 200 "libraries/Box2D/Dynamics/b2Island.cpp"
   $46 = $45; //@line 200 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$46>>2] = $33; //@line 200 "libraries/Box2D/Dynamics/b2Island.cpp"
   $47 = ((($26)) + 52|0); //@line 201 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$47>>2] = $35; //@line 201 "libraries/Box2D/Dynamics/b2Island.cpp"
   $48 = HEAP32[$26>>2]|0; //@line 203 "libraries/Box2D/Dynamics/b2Island.cpp"
   $49 = ($48|0)==(2); //@line 203 "libraries/Box2D/Dynamics/b2Island.cpp"
   if ($49) {
    $50 = ((($26)) + 140|0); //@line 206 "libraries/Box2D/Dynamics/b2Island.cpp"
    $51 = +HEAPF32[$50>>2]; //@line 206 "libraries/Box2D/Dynamics/b2Island.cpp"
    $52 = +HEAPF32[$gravity>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $53 = $51 * $52; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $54 = +HEAPF32[$5>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $55 = $51 * $54; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $56 = ((($26)) + 120|0); //@line 206 "libraries/Box2D/Dynamics/b2Island.cpp"
    $57 = +HEAPF32[$56>>2]; //@line 206 "libraries/Box2D/Dynamics/b2Island.cpp"
    $58 = ((($26)) + 76|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $59 = +HEAPF32[$58>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $60 = $57 * $59; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $61 = ((($26)) + 80|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $62 = +HEAPF32[$61>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $63 = $57 * $62; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $64 = $53 + $60; //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $65 = $55 + $63; //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $66 = $0 * $64; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $67 = $0 * $65; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $68 = (HEAP32[tempDoublePtr>>2]=$37,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $69 = $68 + $66; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $70 = (HEAP32[tempDoublePtr>>2]=$39,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $71 = $70 + $67; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $72 = ((($26)) + 128|0); //@line 207 "libraries/Box2D/Dynamics/b2Island.cpp"
    $73 = +HEAPF32[$72>>2]; //@line 207 "libraries/Box2D/Dynamics/b2Island.cpp"
    $74 = $0 * $73; //@line 207 "libraries/Box2D/Dynamics/b2Island.cpp"
    $75 = ((($26)) + 84|0); //@line 207 "libraries/Box2D/Dynamics/b2Island.cpp"
    $76 = +HEAPF32[$75>>2]; //@line 207 "libraries/Box2D/Dynamics/b2Island.cpp"
    $77 = $74 * $76; //@line 207 "libraries/Box2D/Dynamics/b2Island.cpp"
    $78 = $41 + $77; //@line 207 "libraries/Box2D/Dynamics/b2Island.cpp"
    $79 = ((($26)) + 132|0); //@line 216 "libraries/Box2D/Dynamics/b2Island.cpp"
    $80 = +HEAPF32[$79>>2]; //@line 216 "libraries/Box2D/Dynamics/b2Island.cpp"
    $81 = $0 * $80; //@line 216 "libraries/Box2D/Dynamics/b2Island.cpp"
    $82 = $81 + 1.0; //@line 216 "libraries/Box2D/Dynamics/b2Island.cpp"
    $83 = 1.0 / $82; //@line 216 "libraries/Box2D/Dynamics/b2Island.cpp"
    $84 = $69 * $83; //@line 97 "./libraries/Box2D/Common/b2Math.h"
    $85 = (HEAPF32[tempDoublePtr>>2]=$84,HEAP32[tempDoublePtr>>2]|0); //@line 97 "./libraries/Box2D/Common/b2Math.h"
    $86 = $71 * $83; //@line 97 "./libraries/Box2D/Common/b2Math.h"
    $87 = (HEAPF32[tempDoublePtr>>2]=$86,HEAP32[tempDoublePtr>>2]|0); //@line 97 "./libraries/Box2D/Common/b2Math.h"
    $88 = ((($26)) + 136|0); //@line 217 "libraries/Box2D/Dynamics/b2Island.cpp"
    $89 = +HEAPF32[$88>>2]; //@line 217 "libraries/Box2D/Dynamics/b2Island.cpp"
    $90 = $0 * $89; //@line 217 "libraries/Box2D/Dynamics/b2Island.cpp"
    $91 = $90 + 1.0; //@line 217 "libraries/Box2D/Dynamics/b2Island.cpp"
    $92 = 1.0 / $91; //@line 217 "libraries/Box2D/Dynamics/b2Island.cpp"
    $93 = $78 * $92; //@line 217 "libraries/Box2D/Dynamics/b2Island.cpp"
    $v$sroa$0$0 = $85;$v$sroa$8$0 = $87;$w$0 = $93;
   } else {
    $v$sroa$0$0 = $37;$v$sroa$8$0 = $39;$w$0 = $41;
   }
   $94 = HEAP32[$6>>2]|0; //@line 220 "libraries/Box2D/Dynamics/b2Island.cpp"
   $95 = (($94) + (($i$072*12)|0)|0); //@line 220 "libraries/Box2D/Dynamics/b2Island.cpp"
   $96 = $95; //@line 220 "libraries/Box2D/Dynamics/b2Island.cpp"
   $97 = $96; //@line 220 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$97>>2] = $30; //@line 220 "libraries/Box2D/Dynamics/b2Island.cpp"
   $98 = (($96) + 4)|0; //@line 220 "libraries/Box2D/Dynamics/b2Island.cpp"
   $99 = $98; //@line 220 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$99>>2] = $33; //@line 220 "libraries/Box2D/Dynamics/b2Island.cpp"
   $100 = HEAP32[$6>>2]|0; //@line 221 "libraries/Box2D/Dynamics/b2Island.cpp"
   $101 = (((($100) + (($i$072*12)|0)|0)) + 8|0); //@line 221 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$101>>2] = $35; //@line 221 "libraries/Box2D/Dynamics/b2Island.cpp"
   $102 = HEAP32[$7>>2]|0; //@line 222 "libraries/Box2D/Dynamics/b2Island.cpp"
   $103 = (($102) + (($i$072*12)|0)|0); //@line 222 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$103>>2] = $v$sroa$0$0; //@line 222 "libraries/Box2D/Dynamics/b2Island.cpp"
   $104 = (((($102) + (($i$072*12)|0)|0)) + 4|0); //@line 222 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$104>>2] = $v$sroa$8$0; //@line 222 "libraries/Box2D/Dynamics/b2Island.cpp"
   $105 = HEAP32[$7>>2]|0; //@line 223 "libraries/Box2D/Dynamics/b2Island.cpp"
   $106 = (((($105) + (($i$072*12)|0)|0)) + 8|0); //@line 223 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAPF32[$106>>2] = $w$0; //@line 223 "libraries/Box2D/Dynamics/b2Island.cpp"
   $107 = (($i$072) + 1)|0; //@line 190 "libraries/Box2D/Dynamics/b2Island.cpp"
   $108 = HEAP32[$1>>2]|0; //@line 190 "libraries/Box2D/Dynamics/b2Island.cpp"
   $109 = ($107|0)<($108|0); //@line 190 "libraries/Box2D/Dynamics/b2Island.cpp"
   if ($109) {
    $i$072 = $107;
   } else {
    $$pre$phi79Z2D = $7;$$pre$phiZ2D = $6;
    break;
   }
  }
 } else {
  $$pre77 = ((($this)) + 20|0); //@line 231 "libraries/Box2D/Dynamics/b2Island.cpp"
  $$pre78 = ((($this)) + 24|0); //@line 232 "libraries/Box2D/Dynamics/b2Island.cpp"
  $$pre$phi79Z2D = $$pre78;$$pre$phiZ2D = $$pre77;
 }
 ;HEAP32[$solverData>>2]=HEAP32[$step>>2]|0;HEAP32[$solverData+4>>2]=HEAP32[$step+4>>2]|0;HEAP32[$solverData+8>>2]=HEAP32[$step+8>>2]|0;HEAP32[$solverData+12>>2]=HEAP32[$step+12>>2]|0;HEAP32[$solverData+16>>2]=HEAP32[$step+16>>2]|0;HEAP32[$solverData+20>>2]=HEAP32[$step+20>>2]|0; //@line 230 "libraries/Box2D/Dynamics/b2Island.cpp"
 $8 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 231 "libraries/Box2D/Dynamics/b2Island.cpp"
 $9 = ((($solverData)) + 24|0); //@line 231 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$9>>2] = $8; //@line 231 "libraries/Box2D/Dynamics/b2Island.cpp"
 $10 = HEAP32[$$pre$phi79Z2D>>2]|0; //@line 232 "libraries/Box2D/Dynamics/b2Island.cpp"
 $11 = ((($solverData)) + 28|0); //@line 232 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$11>>2] = $10; //@line 232 "libraries/Box2D/Dynamics/b2Island.cpp"
 ;HEAP32[$contactSolverDef>>2]=HEAP32[$step>>2]|0;HEAP32[$contactSolverDef+4>>2]=HEAP32[$step+4>>2]|0;HEAP32[$contactSolverDef+8>>2]=HEAP32[$step+8>>2]|0;HEAP32[$contactSolverDef+12>>2]=HEAP32[$step+12>>2]|0;HEAP32[$contactSolverDef+16>>2]=HEAP32[$step+16>>2]|0;HEAP32[$contactSolverDef+20>>2]=HEAP32[$step+20>>2]|0; //@line 236 "libraries/Box2D/Dynamics/b2Island.cpp"
 $12 = ((($this)) + 12|0); //@line 237 "libraries/Box2D/Dynamics/b2Island.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 237 "libraries/Box2D/Dynamics/b2Island.cpp"
 $14 = ((($contactSolverDef)) + 24|0); //@line 237 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$14>>2] = $13; //@line 237 "libraries/Box2D/Dynamics/b2Island.cpp"
 $15 = ((($this)) + 36|0); //@line 238 "libraries/Box2D/Dynamics/b2Island.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 238 "libraries/Box2D/Dynamics/b2Island.cpp"
 $17 = ((($contactSolverDef)) + 28|0); //@line 238 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$17>>2] = $16; //@line 238 "libraries/Box2D/Dynamics/b2Island.cpp"
 $18 = ((($contactSolverDef)) + 32|0); //@line 239 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$18>>2] = $8; //@line 239 "libraries/Box2D/Dynamics/b2Island.cpp"
 $19 = ((($contactSolverDef)) + 36|0); //@line 240 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$19>>2] = $10; //@line 240 "libraries/Box2D/Dynamics/b2Island.cpp"
 $20 = HEAP32[$this>>2]|0; //@line 241 "libraries/Box2D/Dynamics/b2Island.cpp"
 $21 = ((($contactSolverDef)) + 40|0); //@line 241 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$21>>2] = $20; //@line 241 "libraries/Box2D/Dynamics/b2Island.cpp"
 __ZN15b2ContactSolverC2EP18b2ContactSolverDef($contactSolver,$contactSolverDef); //@line 243 "libraries/Box2D/Dynamics/b2Island.cpp"
 __THREW__ = 0;
 invoke_vi(4240,($contactSolver|0)); //@line 244 "libraries/Box2D/Dynamics/b2Island.cpp"
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 do {
  if (!($23)) {
   $110 = ((($step)) + 20|0); //@line 246 "libraries/Box2D/Dynamics/b2Island.cpp"
   $111 = HEAP8[$110>>0]|0; //@line 246 "libraries/Box2D/Dynamics/b2Island.cpp"
   $112 = ($111<<24>>24)==(0); //@line 246 "libraries/Box2D/Dynamics/b2Island.cpp"
   if (!($112)) {
    __THREW__ = 0;
    invoke_vi(4241,($contactSolver|0)); //@line 248 "libraries/Box2D/Dynamics/b2Island.cpp"
    $113 = __THREW__; __THREW__ = 0;
    $114 = $113&1;
    if ($114) {
     break;
    }
   }
   $115 = ((($this)) + 32|0); //@line 251 "libraries/Box2D/Dynamics/b2Island.cpp"
   $116 = HEAP32[$115>>2]|0; //@line 251 "libraries/Box2D/Dynamics/b2Island.cpp"
   $117 = ($116|0)>(0); //@line 251 "libraries/Box2D/Dynamics/b2Island.cpp"
   L15: do {
    if ($117) {
     $118 = ((($this)) + 16|0); //@line 253 "libraries/Box2D/Dynamics/b2Island.cpp"
     $i1$068 = 0;
     while(1) {
      $136 = HEAP32[$118>>2]|0; //@line 253 "libraries/Box2D/Dynamics/b2Island.cpp"
      $137 = (($136) + ($i1$068<<2)|0); //@line 253 "libraries/Box2D/Dynamics/b2Island.cpp"
      $138 = HEAP32[$137>>2]|0; //@line 253 "libraries/Box2D/Dynamics/b2Island.cpp"
      $139 = HEAP32[$138>>2]|0; //@line 253 "libraries/Box2D/Dynamics/b2Island.cpp"
      $140 = ((($139)) + 32|0); //@line 253 "libraries/Box2D/Dynamics/b2Island.cpp"
      $141 = HEAP32[$140>>2]|0; //@line 253 "libraries/Box2D/Dynamics/b2Island.cpp"
      __THREW__ = 0;
      invoke_vii($141|0,($138|0),($solverData|0)); //@line 253 "libraries/Box2D/Dynamics/b2Island.cpp"
      $142 = __THREW__; __THREW__ = 0;
      $143 = $142&1;
      if ($143) {
       break;
      }
      $144 = (($i1$068) + 1)|0; //@line 251 "libraries/Box2D/Dynamics/b2Island.cpp"
      $145 = HEAP32[$115>>2]|0; //@line 251 "libraries/Box2D/Dynamics/b2Island.cpp"
      $146 = ($144|0)<($145|0); //@line 251 "libraries/Box2D/Dynamics/b2Island.cpp"
      if ($146) {
       $i1$068 = $144;
      } else {
       break L15;
      }
     }
     $129 = ___cxa_find_matching_catch_2()|0;
     $130 = tempRet0;
     $lpad$phi$index = $129;$lpad$phi$index7 = $130;
     __ZN15b2ContactSolverD2Ev($contactSolver); //@line 382 "libraries/Box2D/Dynamics/b2Island.cpp"
     ___resumeException($lpad$phi$index|0);
     // unreachable;
    }
   } while(0);
   __THREW__ = 0;
   $133 = (+invoke_di(3827,($timer|0))); //@line 256 "libraries/Box2D/Dynamics/b2Island.cpp"
   $134 = __THREW__; __THREW__ = 0;
   $135 = $134&1;
   if (!($135)) {
    $147 = ((($profile)) + 12|0); //@line 256 "libraries/Box2D/Dynamics/b2Island.cpp"
    HEAPF32[$147>>2] = $133; //@line 256 "libraries/Box2D/Dynamics/b2Island.cpp"
    __THREW__ = 0;
    $148 = __THREW__; __THREW__ = 0;
    $149 = $148&1;
    if (!($149)) {
     $150 = ((($step)) + 12|0); //@line 260 "libraries/Box2D/Dynamics/b2Island.cpp"
     $151 = HEAP32[$150>>2]|0; //@line 260 "libraries/Box2D/Dynamics/b2Island.cpp"
     $152 = ($151|0)>(0); //@line 260 "libraries/Box2D/Dynamics/b2Island.cpp"
     L25: do {
      if ($152) {
       $153 = ((($this)) + 16|0); //@line 264 "libraries/Box2D/Dynamics/b2Island.cpp"
       $i2$066 = 0;
       L27: while(1) {
        $154 = HEAP32[$115>>2]|0; //@line 262 "libraries/Box2D/Dynamics/b2Island.cpp"
        $155 = ($154|0)>(0); //@line 262 "libraries/Box2D/Dynamics/b2Island.cpp"
        if ($155) {
         $j$063 = 0;
         while(1) {
          $160 = HEAP32[$153>>2]|0; //@line 264 "libraries/Box2D/Dynamics/b2Island.cpp"
          $161 = (($160) + ($j$063<<2)|0); //@line 264 "libraries/Box2D/Dynamics/b2Island.cpp"
          $162 = HEAP32[$161>>2]|0; //@line 264 "libraries/Box2D/Dynamics/b2Island.cpp"
          $163 = HEAP32[$162>>2]|0; //@line 264 "libraries/Box2D/Dynamics/b2Island.cpp"
          $164 = ((($163)) + 36|0); //@line 264 "libraries/Box2D/Dynamics/b2Island.cpp"
          $165 = HEAP32[$164>>2]|0; //@line 264 "libraries/Box2D/Dynamics/b2Island.cpp"
          __THREW__ = 0;
          invoke_vii($165|0,($162|0),($solverData|0)); //@line 264 "libraries/Box2D/Dynamics/b2Island.cpp"
          $166 = __THREW__; __THREW__ = 0;
          $167 = $166&1;
          if ($167) {
           label = 15;
           break L27;
          }
          $168 = (($j$063) + 1)|0; //@line 262 "libraries/Box2D/Dynamics/b2Island.cpp"
          $169 = HEAP32[$115>>2]|0; //@line 262 "libraries/Box2D/Dynamics/b2Island.cpp"
          $170 = ($168|0)<($169|0); //@line 262 "libraries/Box2D/Dynamics/b2Island.cpp"
          if ($170) {
           $j$063 = $168;
          } else {
           break;
          }
         }
        }
        __THREW__ = 0;
        invoke_vi(4243,($contactSolver|0)); //@line 267 "libraries/Box2D/Dynamics/b2Island.cpp"
        $158 = __THREW__; __THREW__ = 0;
        $159 = $158&1;
        if ($159) {
         label = 16;
         break;
        }
        $171 = (($i2$066) + 1)|0; //@line 260 "libraries/Box2D/Dynamics/b2Island.cpp"
        $172 = HEAP32[$150>>2]|0; //@line 260 "libraries/Box2D/Dynamics/b2Island.cpp"
        $173 = ($171|0)<($172|0); //@line 260 "libraries/Box2D/Dynamics/b2Island.cpp"
        if ($173) {
         $i2$066 = $171;
        } else {
         break L25;
        }
       }
       if ((label|0) == 15) {
        $125 = ___cxa_find_matching_catch_2()|0;
        $126 = tempRet0;
        $lpad$phi$index = $125;$lpad$phi$index7 = $126;
        __ZN15b2ContactSolverD2Ev($contactSolver); //@line 382 "libraries/Box2D/Dynamics/b2Island.cpp"
        ___resumeException($lpad$phi$index|0);
        // unreachable;
       }
       else if ((label|0) == 16) {
        $127 = ___cxa_find_matching_catch_2()|0;
        $128 = tempRet0;
        $lpad$phi$index = $127;$lpad$phi$index7 = $128;
        __ZN15b2ContactSolverD2Ev($contactSolver); //@line 382 "libraries/Box2D/Dynamics/b2Island.cpp"
        ___resumeException($lpad$phi$index|0);
        // unreachable;
       }
      }
     } while(0);
     __THREW__ = 0;
     invoke_vi(4242,($contactSolver|0)); //@line 271 "libraries/Box2D/Dynamics/b2Island.cpp"
     $156 = __THREW__; __THREW__ = 0;
     $157 = $156&1;
     if (!($157)) {
      __THREW__ = 0;
      $174 = (+invoke_di(3827,($timer|0))); //@line 272 "libraries/Box2D/Dynamics/b2Island.cpp"
      $175 = __THREW__; __THREW__ = 0;
      $176 = $175&1;
      if (!($176)) {
       $177 = ((($profile)) + 16|0); //@line 272 "libraries/Box2D/Dynamics/b2Island.cpp"
       HEAPF32[$177>>2] = $174; //@line 272 "libraries/Box2D/Dynamics/b2Island.cpp"
       $178 = HEAP32[$1>>2]|0; //@line 275 "libraries/Box2D/Dynamics/b2Island.cpp"
       $179 = ($178|0)>(0); //@line 275 "libraries/Box2D/Dynamics/b2Island.cpp"
       if ($179) {
        $$pre = HEAP32[$$pre$phi79Z2D>>2]|0; //@line 279 "libraries/Box2D/Dynamics/b2Island.cpp"
        $194 = $$pre;$i3$059 = 0;
        while(1) {
         $186 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 301 "libraries/Box2D/Dynamics/b2Island.cpp"
         $187 = (($186) + (($i3$059*12)|0)|0); //@line 277 "libraries/Box2D/Dynamics/b2Island.cpp"
         $188 = +HEAPF32[$187>>2]; //@line 277 "libraries/Box2D/Dynamics/b2Island.cpp"
         $189 = (((($186) + (($i3$059*12)|0)|0)) + 4|0); //@line 277 "libraries/Box2D/Dynamics/b2Island.cpp"
         $190 = +HEAPF32[$189>>2]; //@line 277 "libraries/Box2D/Dynamics/b2Island.cpp"
         $191 = (((($186) + (($i3$059*12)|0)|0)) + 8|0); //@line 278 "libraries/Box2D/Dynamics/b2Island.cpp"
         $192 = +HEAPF32[$191>>2]; //@line 278 "libraries/Box2D/Dynamics/b2Island.cpp"
         $193 = (($194) + (($i3$059*12)|0)|0); //@line 279 "libraries/Box2D/Dynamics/b2Island.cpp"
         $195 = HEAP32[$193>>2]|0; //@line 279 "libraries/Box2D/Dynamics/b2Island.cpp"
         $196 = (((($194) + (($i3$059*12)|0)|0)) + 4|0); //@line 279 "libraries/Box2D/Dynamics/b2Island.cpp"
         $197 = HEAP32[$196>>2]|0; //@line 279 "libraries/Box2D/Dynamics/b2Island.cpp"
         $198 = (((($194) + (($i3$059*12)|0)|0)) + 8|0); //@line 280 "libraries/Box2D/Dynamics/b2Island.cpp"
         $199 = +HEAPF32[$198>>2]; //@line 280 "libraries/Box2D/Dynamics/b2Island.cpp"
         $200 = (HEAP32[tempDoublePtr>>2]=$195,+HEAPF32[tempDoublePtr>>2]); //@line 459 "./libraries/Box2D/Common/b2Math.h"
         $201 = $0 * $200; //@line 459 "./libraries/Box2D/Common/b2Math.h"
         $202 = (HEAP32[tempDoublePtr>>2]=$197,+HEAPF32[tempDoublePtr>>2]); //@line 459 "./libraries/Box2D/Common/b2Math.h"
         $203 = $0 * $202; //@line 459 "./libraries/Box2D/Common/b2Math.h"
         $204 = $201 * $201; //@line 408 "./libraries/Box2D/Common/b2Math.h"
         $205 = $203 * $203; //@line 408 "./libraries/Box2D/Common/b2Math.h"
         $206 = $204 + $205; //@line 408 "./libraries/Box2D/Common/b2Math.h"
         $207 = $206 > 4.0; //@line 284 "libraries/Box2D/Dynamics/b2Island.cpp"
         if ($207) {
          $208 = (+Math_sqrt((+$206))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
          $209 = 2.0 / $208; //@line 286 "libraries/Box2D/Dynamics/b2Island.cpp"
          $210 = $200 * $209; //@line 97 "./libraries/Box2D/Common/b2Math.h"
          $211 = (HEAPF32[tempDoublePtr>>2]=$210,HEAP32[tempDoublePtr>>2]|0); //@line 97 "./libraries/Box2D/Common/b2Math.h"
          $212 = $202 * $209; //@line 97 "./libraries/Box2D/Common/b2Math.h"
          $213 = (HEAPF32[tempDoublePtr>>2]=$212,HEAP32[tempDoublePtr>>2]|0); //@line 97 "./libraries/Box2D/Common/b2Math.h"
          $v6$sroa$0$0 = $211;$v6$sroa$8$0 = $213;
         } else {
          $v6$sroa$0$0 = $195;$v6$sroa$8$0 = $197;
         }
         $214 = $0 * $199; //@line 290 "libraries/Box2D/Dynamics/b2Island.cpp"
         $215 = $214 * $214; //@line 291 "libraries/Box2D/Dynamics/b2Island.cpp"
         $216 = $215 > 2.4674012660980225; //@line 291 "libraries/Box2D/Dynamics/b2Island.cpp"
         if ($216) {
          $217 = $214 > 0.0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
          $218 = -$214; //@line 618 "./libraries/Box2D/Common/b2Math.h"
          $219 = $217 ? $214 : $218; //@line 618 "./libraries/Box2D/Common/b2Math.h"
          $220 = 1.5707963705062866 / $219; //@line 293 "libraries/Box2D/Dynamics/b2Island.cpp"
          $221 = $199 * $220; //@line 294 "libraries/Box2D/Dynamics/b2Island.cpp"
          $w7$0 = $221;
         } else {
          $w7$0 = $199;
         }
         $222 = (HEAP32[tempDoublePtr>>2]=$v6$sroa$0$0,+HEAPF32[tempDoublePtr>>2]); //@line 459 "./libraries/Box2D/Common/b2Math.h"
         $223 = $0 * $222; //@line 459 "./libraries/Box2D/Common/b2Math.h"
         $224 = (HEAP32[tempDoublePtr>>2]=$v6$sroa$8$0,+HEAPF32[tempDoublePtr>>2]); //@line 459 "./libraries/Box2D/Common/b2Math.h"
         $225 = $0 * $224; //@line 459 "./libraries/Box2D/Common/b2Math.h"
         $226 = $188 + $223; //@line 85 "./libraries/Box2D/Common/b2Math.h"
         $227 = $190 + $225; //@line 85 "./libraries/Box2D/Common/b2Math.h"
         $228 = $0 * $w7$0; //@line 299 "libraries/Box2D/Dynamics/b2Island.cpp"
         $229 = $192 + $228; //@line 299 "libraries/Box2D/Dynamics/b2Island.cpp"
         HEAPF32[$187>>2] = $226; //@line 301 "libraries/Box2D/Dynamics/b2Island.cpp"
         HEAPF32[$189>>2] = $227; //@line 301 "libraries/Box2D/Dynamics/b2Island.cpp"
         $230 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 302 "libraries/Box2D/Dynamics/b2Island.cpp"
         $231 = (((($230) + (($i3$059*12)|0)|0)) + 8|0); //@line 302 "libraries/Box2D/Dynamics/b2Island.cpp"
         HEAPF32[$231>>2] = $229; //@line 302 "libraries/Box2D/Dynamics/b2Island.cpp"
         $232 = HEAP32[$$pre$phi79Z2D>>2]|0; //@line 303 "libraries/Box2D/Dynamics/b2Island.cpp"
         $233 = (($232) + (($i3$059*12)|0)|0); //@line 303 "libraries/Box2D/Dynamics/b2Island.cpp"
         HEAP32[$233>>2] = $v6$sroa$0$0; //@line 303 "libraries/Box2D/Dynamics/b2Island.cpp"
         $234 = (((($232) + (($i3$059*12)|0)|0)) + 4|0); //@line 303 "libraries/Box2D/Dynamics/b2Island.cpp"
         HEAP32[$234>>2] = $v6$sroa$8$0; //@line 303 "libraries/Box2D/Dynamics/b2Island.cpp"
         $235 = HEAP32[$$pre$phi79Z2D>>2]|0; //@line 304 "libraries/Box2D/Dynamics/b2Island.cpp"
         $236 = (((($235) + (($i3$059*12)|0)|0)) + 8|0); //@line 304 "libraries/Box2D/Dynamics/b2Island.cpp"
         HEAPF32[$236>>2] = $w7$0; //@line 304 "libraries/Box2D/Dynamics/b2Island.cpp"
         $237 = (($i3$059) + 1)|0; //@line 275 "libraries/Box2D/Dynamics/b2Island.cpp"
         $238 = HEAP32[$1>>2]|0; //@line 275 "libraries/Box2D/Dynamics/b2Island.cpp"
         $239 = ($237|0)<($238|0); //@line 275 "libraries/Box2D/Dynamics/b2Island.cpp"
         if ($239) {
          $194 = $235;$i3$059 = $237;
         } else {
          break;
         }
        }
       }
       __THREW__ = 0;
       $180 = __THREW__; __THREW__ = 0;
       $181 = $180&1;
       if (!($181)) {
        $182 = ((($step)) + 16|0); //@line 310 "libraries/Box2D/Dynamics/b2Island.cpp"
        $183 = HEAP32[$182>>2]|0; //@line 310 "libraries/Box2D/Dynamics/b2Island.cpp"
        $184 = ($183|0)>(0); //@line 310 "libraries/Box2D/Dynamics/b2Island.cpp"
        do {
         if ($184) {
          $185 = ((($this)) + 16|0); //@line 317 "libraries/Box2D/Dynamics/b2Island.cpp"
          $i9$053 = 0;$positionSolved$0$off052 = 0;
          L57: while(1) {
           __THREW__ = 0;
           $240 = (invoke_ii(4244,($contactSolver|0))|0); //@line 312 "libraries/Box2D/Dynamics/b2Island.cpp"
           $241 = __THREW__; __THREW__ = 0;
           $242 = $241&1;
           if ($242) {
            label = 14;
            break;
           }
           $243 = HEAP32[$115>>2]|0; //@line 315 "libraries/Box2D/Dynamics/b2Island.cpp"
           $244 = ($243|0)>(0); //@line 315 "libraries/Box2D/Dynamics/b2Island.cpp"
           if ($244) {
            $i10$048 = 0;$jointsOkay$0$off047 = 1;
            while(1) {
             $248 = HEAP32[$185>>2]|0; //@line 317 "libraries/Box2D/Dynamics/b2Island.cpp"
             $249 = (($248) + ($i10$048<<2)|0); //@line 317 "libraries/Box2D/Dynamics/b2Island.cpp"
             $250 = HEAP32[$249>>2]|0; //@line 317 "libraries/Box2D/Dynamics/b2Island.cpp"
             $251 = HEAP32[$250>>2]|0; //@line 317 "libraries/Box2D/Dynamics/b2Island.cpp"
             $252 = ((($251)) + 40|0); //@line 317 "libraries/Box2D/Dynamics/b2Island.cpp"
             $253 = HEAP32[$252>>2]|0; //@line 317 "libraries/Box2D/Dynamics/b2Island.cpp"
             __THREW__ = 0;
             $254 = (invoke_iii($253|0,($250|0),($solverData|0))|0); //@line 317 "libraries/Box2D/Dynamics/b2Island.cpp"
             $255 = __THREW__; __THREW__ = 0;
             $256 = $255&1;
             if ($256) {
              label = 13;
              break L57;
             }
             $257 = $jointsOkay$0$off047 & $254; //@line 318 "libraries/Box2D/Dynamics/b2Island.cpp"
             $258 = (($i10$048) + 1)|0; //@line 315 "libraries/Box2D/Dynamics/b2Island.cpp"
             $259 = HEAP32[$115>>2]|0; //@line 315 "libraries/Box2D/Dynamics/b2Island.cpp"
             $260 = ($258|0)<($259|0); //@line 315 "libraries/Box2D/Dynamics/b2Island.cpp"
             if ($260) {
              $i10$048 = $258;$jointsOkay$0$off047 = $257;
             } else {
              $jointsOkay$0$off0$lcssa = $257;
              break;
             }
            }
           } else {
            $jointsOkay$0$off0$lcssa = 1;
           }
           $brmerge$demorgan = $240 & $jointsOkay$0$off0$lcssa; //@line 321 "libraries/Box2D/Dynamics/b2Island.cpp"
           $positionSolved$0$off0$ = $positionSolved$0$off052 | $brmerge$demorgan; //@line 321 "libraries/Box2D/Dynamics/b2Island.cpp"
           $245 = (($i9$053) + 1)|0; //@line 310 "libraries/Box2D/Dynamics/b2Island.cpp"
           $brmerge$demorgan$not = $brmerge$demorgan ^ 1;
           $246 = HEAP32[$182>>2]|0; //@line 310 "libraries/Box2D/Dynamics/b2Island.cpp"
           $247 = ($245|0)<($246|0); //@line 310 "libraries/Box2D/Dynamics/b2Island.cpp"
           $or$cond = $247 & $brmerge$demorgan$not;
           if ($or$cond) {
            $i9$053 = $245;$positionSolved$0$off052 = $positionSolved$0$off0$;
           } else {
            $positionSolved$0$off0$$lcssa = $positionSolved$0$off0$;
            label = 48;
            break;
           }
          }
          if ((label|0) == 13) {
           $121 = ___cxa_find_matching_catch_2()|0;
           $122 = tempRet0;
           $lpad$phi$index = $121;$lpad$phi$index7 = $122;
           __ZN15b2ContactSolverD2Ev($contactSolver); //@line 382 "libraries/Box2D/Dynamics/b2Island.cpp"
           ___resumeException($lpad$phi$index|0);
           // unreachable;
          }
          else if ((label|0) == 14) {
           $123 = ___cxa_find_matching_catch_2()|0;
           $124 = tempRet0;
           $lpad$phi$index = $123;$lpad$phi$index7 = $124;
           __ZN15b2ContactSolverD2Ev($contactSolver); //@line 382 "libraries/Box2D/Dynamics/b2Island.cpp"
           ___resumeException($lpad$phi$index|0);
           // unreachable;
          }
          else if ((label|0) == 48) {
           $phitmp = $positionSolved$0$off0$$lcssa ^ 1; //@line 330 "libraries/Box2D/Dynamics/b2Island.cpp"
           $positionSolved$2$off0 = $phitmp;
           break;
          }
         } else {
          $positionSolved$2$off0 = 1;
         }
        } while(0);
        $261 = HEAP32[$1>>2]|0; //@line 330 "libraries/Box2D/Dynamics/b2Island.cpp"
        $262 = ($261|0)>(0); //@line 330 "libraries/Box2D/Dynamics/b2Island.cpp"
        if ($262) {
         $263 = ((($this)) + 8|0); //@line 332 "libraries/Box2D/Dynamics/b2Island.cpp"
         $i11$043 = 0;
         while(1) {
          $267 = HEAP32[$263>>2]|0; //@line 332 "libraries/Box2D/Dynamics/b2Island.cpp"
          $268 = (($267) + ($i11$043<<2)|0); //@line 332 "libraries/Box2D/Dynamics/b2Island.cpp"
          $269 = HEAP32[$268>>2]|0; //@line 332 "libraries/Box2D/Dynamics/b2Island.cpp"
          $270 = ((($269)) + 44|0); //@line 333 "libraries/Box2D/Dynamics/b2Island.cpp"
          $271 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 333 "libraries/Box2D/Dynamics/b2Island.cpp"
          $272 = (($271) + (($i11$043*12)|0)|0); //@line 333 "libraries/Box2D/Dynamics/b2Island.cpp"
          $273 = $272; //@line 333 "libraries/Box2D/Dynamics/b2Island.cpp"
          $274 = $273; //@line 333 "libraries/Box2D/Dynamics/b2Island.cpp"
          $275 = HEAP32[$274>>2]|0; //@line 333 "libraries/Box2D/Dynamics/b2Island.cpp"
          $276 = (($273) + 4)|0; //@line 333 "libraries/Box2D/Dynamics/b2Island.cpp"
          $277 = $276; //@line 333 "libraries/Box2D/Dynamics/b2Island.cpp"
          $278 = HEAP32[$277>>2]|0; //@line 333 "libraries/Box2D/Dynamics/b2Island.cpp"
          $279 = $270; //@line 333 "libraries/Box2D/Dynamics/b2Island.cpp"
          $280 = $279; //@line 333 "libraries/Box2D/Dynamics/b2Island.cpp"
          HEAP32[$280>>2] = $275; //@line 333 "libraries/Box2D/Dynamics/b2Island.cpp"
          $281 = (($279) + 4)|0; //@line 333 "libraries/Box2D/Dynamics/b2Island.cpp"
          $282 = $281; //@line 333 "libraries/Box2D/Dynamics/b2Island.cpp"
          HEAP32[$282>>2] = $278; //@line 333 "libraries/Box2D/Dynamics/b2Island.cpp"
          $283 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 334 "libraries/Box2D/Dynamics/b2Island.cpp"
          $284 = (((($283) + (($i11$043*12)|0)|0)) + 8|0); //@line 334 "libraries/Box2D/Dynamics/b2Island.cpp"
          $285 = HEAP32[$284>>2]|0; //@line 334 "libraries/Box2D/Dynamics/b2Island.cpp"
          $286 = ((($269)) + 56|0); //@line 334 "libraries/Box2D/Dynamics/b2Island.cpp"
          HEAP32[$286>>2] = $285; //@line 334 "libraries/Box2D/Dynamics/b2Island.cpp"
          $287 = ((($269)) + 64|0); //@line 335 "libraries/Box2D/Dynamics/b2Island.cpp"
          $288 = HEAP32[$$pre$phi79Z2D>>2]|0; //@line 335 "libraries/Box2D/Dynamics/b2Island.cpp"
          $289 = (($288) + (($i11$043*12)|0)|0); //@line 335 "libraries/Box2D/Dynamics/b2Island.cpp"
          $290 = $289; //@line 335 "libraries/Box2D/Dynamics/b2Island.cpp"
          $291 = $290; //@line 335 "libraries/Box2D/Dynamics/b2Island.cpp"
          $292 = HEAP32[$291>>2]|0; //@line 335 "libraries/Box2D/Dynamics/b2Island.cpp"
          $293 = (($290) + 4)|0; //@line 335 "libraries/Box2D/Dynamics/b2Island.cpp"
          $294 = $293; //@line 335 "libraries/Box2D/Dynamics/b2Island.cpp"
          $295 = HEAP32[$294>>2]|0; //@line 335 "libraries/Box2D/Dynamics/b2Island.cpp"
          $296 = $287; //@line 335 "libraries/Box2D/Dynamics/b2Island.cpp"
          $297 = $296; //@line 335 "libraries/Box2D/Dynamics/b2Island.cpp"
          HEAP32[$297>>2] = $292; //@line 335 "libraries/Box2D/Dynamics/b2Island.cpp"
          $298 = (($296) + 4)|0; //@line 335 "libraries/Box2D/Dynamics/b2Island.cpp"
          $299 = $298; //@line 335 "libraries/Box2D/Dynamics/b2Island.cpp"
          HEAP32[$299>>2] = $295; //@line 335 "libraries/Box2D/Dynamics/b2Island.cpp"
          $300 = HEAP32[$$pre$phi79Z2D>>2]|0; //@line 336 "libraries/Box2D/Dynamics/b2Island.cpp"
          $301 = (((($300) + (($i11$043*12)|0)|0)) + 8|0); //@line 336 "libraries/Box2D/Dynamics/b2Island.cpp"
          $302 = HEAP32[$301>>2]|0; //@line 336 "libraries/Box2D/Dynamics/b2Island.cpp"
          $303 = ((($269)) + 72|0); //@line 336 "libraries/Box2D/Dynamics/b2Island.cpp"
          HEAP32[$303>>2] = $302; //@line 336 "libraries/Box2D/Dynamics/b2Island.cpp"
          $304 = (HEAP32[tempDoublePtr>>2]=$285,+HEAPF32[tempDoublePtr>>2]); //@line 836 "./libraries/Box2D/Dynamics/b2Body.h"
          $305 = (+Math_sin((+$304))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
          $306 = ((($269)) + 20|0); //@line 315 "./libraries/Box2D/Common/b2Math.h"
          HEAPF32[$306>>2] = $305; //@line 315 "./libraries/Box2D/Common/b2Math.h"
          $307 = (+Math_cos((+$304))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
          $308 = ((($269)) + 24|0); //@line 316 "./libraries/Box2D/Common/b2Math.h"
          HEAPF32[$308>>2] = $307; //@line 316 "./libraries/Box2D/Common/b2Math.h"
          $309 = ((($269)) + 28|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
          $310 = +HEAPF32[$309>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
          $311 = $307 * $310; //@line 568 "./libraries/Box2D/Common/b2Math.h"
          $312 = ((($269)) + 32|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
          $313 = +HEAPF32[$312>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
          $314 = $305 * $313; //@line 568 "./libraries/Box2D/Common/b2Math.h"
          $315 = $311 - $314; //@line 568 "./libraries/Box2D/Common/b2Math.h"
          $316 = $305 * $310; //@line 568 "./libraries/Box2D/Common/b2Math.h"
          $317 = $307 * $313; //@line 568 "./libraries/Box2D/Common/b2Math.h"
          $318 = $316 + $317; //@line 568 "./libraries/Box2D/Common/b2Math.h"
          $bitcast = (HEAP32[tempDoublePtr>>2]=$275,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
          $319 = $bitcast - $315; //@line 454 "./libraries/Box2D/Common/b2Math.h"
          $320 = (HEAP32[tempDoublePtr>>2]=$278,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
          $321 = $320 - $318; //@line 454 "./libraries/Box2D/Common/b2Math.h"
          $322 = ((($269)) + 12|0); //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
          HEAPF32[$322>>2] = $319; //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
          $323 = ((($269)) + 16|0); //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
          HEAPF32[$323>>2] = $321; //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
          $324 = (($i11$043) + 1)|0; //@line 330 "libraries/Box2D/Dynamics/b2Island.cpp"
          $325 = HEAP32[$1>>2]|0; //@line 330 "libraries/Box2D/Dynamics/b2Island.cpp"
          $326 = ($324|0)<($325|0); //@line 330 "libraries/Box2D/Dynamics/b2Island.cpp"
          if ($326) {
           $i11$043 = $324;
          } else {
           break;
          }
         }
        }
        __THREW__ = 0;
        $264 = (+invoke_di(3827,($timer|0))); //@line 340 "libraries/Box2D/Dynamics/b2Island.cpp"
        $265 = __THREW__; __THREW__ = 0;
        $266 = $265&1;
        if (!($266)) {
         $327 = ((($profile)) + 20|0); //@line 340 "libraries/Box2D/Dynamics/b2Island.cpp"
         HEAPF32[$327>>2] = $264; //@line 340 "libraries/Box2D/Dynamics/b2Island.cpp"
         $328 = ((($contactSolver)) + 40|0); //@line 342 "libraries/Box2D/Dynamics/b2Island.cpp"
         $329 = HEAP32[$328>>2]|0; //@line 342 "libraries/Box2D/Dynamics/b2Island.cpp"
         $330 = ((($this)) + 4|0); //@line 518 "libraries/Box2D/Dynamics/b2Island.cpp"
         $331 = HEAP32[$330>>2]|0; //@line 518 "libraries/Box2D/Dynamics/b2Island.cpp"
         $332 = ($331|0)==(0|0); //@line 518 "libraries/Box2D/Dynamics/b2Island.cpp"
         L78: do {
          if (!($332)) {
           $333 = HEAP32[$15>>2]|0; //@line 523 "libraries/Box2D/Dynamics/b2Island.cpp"
           $334 = ($333|0)>(0); //@line 523 "libraries/Box2D/Dynamics/b2Island.cpp"
           if ($334) {
            $335 = ((($impulse$i)) + 16|0); //@line 530 "libraries/Box2D/Dynamics/b2Island.cpp"
            $i$02$i = 0;
            while(1) {
             $336 = HEAP32[$12>>2]|0; //@line 525 "libraries/Box2D/Dynamics/b2Island.cpp"
             $337 = (($336) + ($i$02$i<<2)|0); //@line 525 "libraries/Box2D/Dynamics/b2Island.cpp"
             $338 = HEAP32[$337>>2]|0; //@line 525 "libraries/Box2D/Dynamics/b2Island.cpp"
             $339 = (((($329) + (($i$02$i*156)|0)|0)) + 148|0); //@line 530 "libraries/Box2D/Dynamics/b2Island.cpp"
             $340 = HEAP32[$339>>2]|0; //@line 531 "libraries/Box2D/Dynamics/b2Island.cpp"
             HEAP32[$335>>2] = $340; //@line 530 "libraries/Box2D/Dynamics/b2Island.cpp"
             $341 = ($340|0)>(0); //@line 531 "libraries/Box2D/Dynamics/b2Island.cpp"
             if ($341) {
              $j$01$i = 0;
              while(1) {
               $351 = ((((($329) + (($i$02$i*156)|0)|0) + (($j$01$i*36)|0)|0)) + 16|0); //@line 533 "libraries/Box2D/Dynamics/b2Island.cpp"
               $352 = HEAP32[$351>>2]|0; //@line 533 "libraries/Box2D/Dynamics/b2Island.cpp"
               $353 = (($impulse$i) + ($j$01$i<<2)|0); //@line 533 "libraries/Box2D/Dynamics/b2Island.cpp"
               HEAP32[$353>>2] = $352; //@line 533 "libraries/Box2D/Dynamics/b2Island.cpp"
               $354 = ((((($329) + (($i$02$i*156)|0)|0) + (($j$01$i*36)|0)|0)) + 20|0); //@line 534 "libraries/Box2D/Dynamics/b2Island.cpp"
               $355 = HEAP32[$354>>2]|0; //@line 534 "libraries/Box2D/Dynamics/b2Island.cpp"
               $356 = (((($impulse$i)) + 8|0) + ($j$01$i<<2)|0); //@line 534 "libraries/Box2D/Dynamics/b2Island.cpp"
               HEAP32[$356>>2] = $355; //@line 534 "libraries/Box2D/Dynamics/b2Island.cpp"
               $357 = (($j$01$i) + 1)|0; //@line 531 "libraries/Box2D/Dynamics/b2Island.cpp"
               $exitcond = ($357|0)==($340|0); //@line 531 "libraries/Box2D/Dynamics/b2Island.cpp"
               if ($exitcond) {
                break;
               } else {
                $j$01$i = $357;
               }
              }
             }
             $342 = HEAP32[$330>>2]|0; //@line 537 "libraries/Box2D/Dynamics/b2Island.cpp"
             $343 = HEAP32[$342>>2]|0; //@line 537 "libraries/Box2D/Dynamics/b2Island.cpp"
             $344 = ((($343)) + 20|0); //@line 537 "libraries/Box2D/Dynamics/b2Island.cpp"
             $345 = HEAP32[$344>>2]|0; //@line 537 "libraries/Box2D/Dynamics/b2Island.cpp"
             __THREW__ = 0;
             invoke_viii($345|0,($342|0),($338|0),($impulse$i|0)); //@line 537 "libraries/Box2D/Dynamics/b2Island.cpp"
             $346 = __THREW__; __THREW__ = 0;
             $347 = $346&1;
             if ($347) {
              break;
             }
             $348 = (($i$02$i) + 1)|0; //@line 523 "libraries/Box2D/Dynamics/b2Island.cpp"
             $349 = HEAP32[$15>>2]|0; //@line 523 "libraries/Box2D/Dynamics/b2Island.cpp"
             $350 = ($348|0)<($349|0); //@line 523 "libraries/Box2D/Dynamics/b2Island.cpp"
             if ($350) {
              $i$02$i = $348;
             } else {
              break L78;
             }
            }
            $119 = ___cxa_find_matching_catch_2()|0;
            $120 = tempRet0;
            $lpad$phi$index = $119;$lpad$phi$index7 = $120;
            __ZN15b2ContactSolverD2Ev($contactSolver); //@line 382 "libraries/Box2D/Dynamics/b2Island.cpp"
            ___resumeException($lpad$phi$index|0);
            // unreachable;
           }
          }
         } while(0);
         if (!($allowSleep)) {
          __ZN15b2ContactSolverD2Ev($contactSolver); //@line 382 "libraries/Box2D/Dynamics/b2Island.cpp"
          STACKTOP = sp;return; //@line 382 "libraries/Box2D/Dynamics/b2Island.cpp"
         }
         $358 = HEAP32[$1>>2]|0; //@line 351 "libraries/Box2D/Dynamics/b2Island.cpp"
         $359 = ($358|0)>(0); //@line 351 "libraries/Box2D/Dynamics/b2Island.cpp"
         if ($359) {
          $360 = ((($this)) + 8|0); //@line 353 "libraries/Box2D/Dynamics/b2Island.cpp"
          $361 = HEAP32[$360>>2]|0; //@line 353 "libraries/Box2D/Dynamics/b2Island.cpp"
          $i12$041 = 0;$minSleepTime$040 = 3.4028234663852886E+38;
          while(1) {
           $366 = (($361) + ($i12$041<<2)|0); //@line 353 "libraries/Box2D/Dynamics/b2Island.cpp"
           $367 = HEAP32[$366>>2]|0; //@line 353 "libraries/Box2D/Dynamics/b2Island.cpp"
           $368 = HEAP32[$367>>2]|0; //@line 471 "./libraries/Box2D/Dynamics/b2Body.h"
           $369 = ($368|0)==(0); //@line 354 "libraries/Box2D/Dynamics/b2Island.cpp"
           L98: do {
            if ($369) {
             $minSleepTime$2 = $minSleepTime$040;
            } else {
             $370 = ((($367)) + 4|0); //@line 359 "libraries/Box2D/Dynamics/b2Island.cpp"
             $371 = HEAP16[$370>>1]|0; //@line 359 "libraries/Box2D/Dynamics/b2Island.cpp"
             $372 = $371 & 4; //@line 359 "libraries/Box2D/Dynamics/b2Island.cpp"
             $373 = ($372<<16>>16)==(0); //@line 359 "libraries/Box2D/Dynamics/b2Island.cpp"
             do {
              if (!($373)) {
               $374 = ((($367)) + 72|0); //@line 360 "libraries/Box2D/Dynamics/b2Island.cpp"
               $375 = +HEAPF32[$374>>2]; //@line 360 "libraries/Box2D/Dynamics/b2Island.cpp"
               $376 = $375 * $375; //@line 360 "libraries/Box2D/Dynamics/b2Island.cpp"
               $377 = $376 > 0.0012184699298813939; //@line 360 "libraries/Box2D/Dynamics/b2Island.cpp"
               if ($377) {
                break;
               }
               $378 = ((($367)) + 64|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
               $379 = +HEAPF32[$378>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
               $380 = $379 * $379; //@line 408 "./libraries/Box2D/Common/b2Math.h"
               $381 = ((($367)) + 68|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
               $382 = +HEAPF32[$381>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
               $383 = $382 * $382; //@line 408 "./libraries/Box2D/Common/b2Math.h"
               $384 = $380 + $383; //@line 408 "./libraries/Box2D/Common/b2Math.h"
               $385 = $384 > 9.9999997473787516E-5; //@line 361 "libraries/Box2D/Dynamics/b2Island.cpp"
               if ($385) {
                break;
               }
               $387 = ((($367)) + 144|0); //@line 368 "libraries/Box2D/Dynamics/b2Island.cpp"
               $388 = +HEAPF32[$387>>2]; //@line 368 "libraries/Box2D/Dynamics/b2Island.cpp"
               $389 = $0 + $388; //@line 368 "libraries/Box2D/Dynamics/b2Island.cpp"
               HEAPF32[$387>>2] = $389; //@line 368 "libraries/Box2D/Dynamics/b2Island.cpp"
               $390 = $minSleepTime$040 < $389; //@line 634 "./libraries/Box2D/Common/b2Math.h"
               $391 = $390 ? $minSleepTime$040 : $389; //@line 634 "./libraries/Box2D/Common/b2Math.h"
               $minSleepTime$2 = $391;
               break L98;
              }
             } while(0);
             $386 = ((($367)) + 144|0); //@line 363 "libraries/Box2D/Dynamics/b2Island.cpp"
             HEAPF32[$386>>2] = 0.0; //@line 363 "libraries/Box2D/Dynamics/b2Island.cpp"
             $minSleepTime$2 = 0.0;
            }
           } while(0);
           $392 = (($i12$041) + 1)|0; //@line 351 "libraries/Box2D/Dynamics/b2Island.cpp"
           $393 = ($392|0)<($358|0); //@line 351 "libraries/Box2D/Dynamics/b2Island.cpp"
           if ($393) {
            $i12$041 = $392;$minSleepTime$040 = $minSleepTime$2;
           } else {
            $minSleepTime$0$lcssa = $minSleepTime$2;
            break;
           }
          }
         } else {
          $minSleepTime$0$lcssa = 3.4028234663852886E+38;
         }
         $362 = !($minSleepTime$0$lcssa >= 0.5); //@line 373 "libraries/Box2D/Dynamics/b2Island.cpp"
         $brmerge7 = $362 | $positionSolved$2$off0; //@line 373 "libraries/Box2D/Dynamics/b2Island.cpp"
         if ($brmerge7) {
          __ZN15b2ContactSolverD2Ev($contactSolver); //@line 382 "libraries/Box2D/Dynamics/b2Island.cpp"
          STACKTOP = sp;return; //@line 382 "libraries/Box2D/Dynamics/b2Island.cpp"
         }
         $363 = HEAP32[$1>>2]|0; //@line 375 "libraries/Box2D/Dynamics/b2Island.cpp"
         $364 = ($363|0)>(0); //@line 375 "libraries/Box2D/Dynamics/b2Island.cpp"
         if (!($364)) {
          __ZN15b2ContactSolverD2Ev($contactSolver); //@line 382 "libraries/Box2D/Dynamics/b2Island.cpp"
          STACKTOP = sp;return; //@line 382 "libraries/Box2D/Dynamics/b2Island.cpp"
         }
         $365 = ((($this)) + 8|0); //@line 377 "libraries/Box2D/Dynamics/b2Island.cpp"
         $i15$039 = 0;
         while(1) {
          $394 = HEAP32[$365>>2]|0; //@line 377 "libraries/Box2D/Dynamics/b2Island.cpp"
          $395 = (($394) + ($i15$039<<2)|0); //@line 377 "libraries/Box2D/Dynamics/b2Island.cpp"
          $396 = HEAP32[$395>>2]|0; //@line 377 "libraries/Box2D/Dynamics/b2Island.cpp"
          $397 = ((($396)) + 4|0); //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
          $398 = HEAP16[$397>>1]|0; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
          $399 = $398&65535; //@line 637 "./libraries/Box2D/Dynamics/b2Body.h"
          $400 = $399 & 65533; //@line 645 "./libraries/Box2D/Dynamics/b2Body.h"
          $401 = $400&65535; //@line 645 "./libraries/Box2D/Dynamics/b2Body.h"
          HEAP16[$397>>1] = $401; //@line 645 "./libraries/Box2D/Dynamics/b2Body.h"
          $402 = ((($396)) + 144|0); //@line 646 "./libraries/Box2D/Dynamics/b2Body.h"
          HEAPF32[$402>>2] = 0.0; //@line 646 "./libraries/Box2D/Dynamics/b2Body.h"
          $403 = ((($396)) + 64|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
          ;HEAP32[$403>>2]=0|0;HEAP32[$403+4>>2]=0|0;HEAP32[$403+8>>2]=0|0;HEAP32[$403+12>>2]=0|0;HEAP32[$403+16>>2]=0|0;HEAP32[$403+20>>2]=0|0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
          $404 = (($i15$039) + 1)|0; //@line 375 "libraries/Box2D/Dynamics/b2Island.cpp"
          $405 = HEAP32[$1>>2]|0; //@line 375 "libraries/Box2D/Dynamics/b2Island.cpp"
          $406 = ($404|0)<($405|0); //@line 375 "libraries/Box2D/Dynamics/b2Island.cpp"
          if ($406) {
           $i15$039 = $404;
          } else {
           break;
          }
         }
         __ZN15b2ContactSolverD2Ev($contactSolver); //@line 382 "libraries/Box2D/Dynamics/b2Island.cpp"
         STACKTOP = sp;return; //@line 382 "libraries/Box2D/Dynamics/b2Island.cpp"
        }
       }
      }
     }
    }
   }
  }
 } while(0);
 $131 = ___cxa_find_matching_catch_2()|0;
 $132 = tempRet0;
 $lpad$phi$index = $131;$lpad$phi$index7 = $132;
 __ZN15b2ContactSolverD2Ev($contactSolver); //@line 382 "libraries/Box2D/Dynamics/b2Island.cpp"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZN8b2Island8SolveTOIERK10b2TimeStepii($this,$subStep,$toiIndexA,$toiIndexB) {
 $this = $this|0;
 $subStep = $subStep|0;
 $toiIndexA = $toiIndexA|0;
 $toiIndexB = $toiIndexB|0;
 var $$lcssa = 0, $$not = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phi22Z2D = 0, $$pre$phiZ2D = 0, $$pre21 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0.0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0.0, $171 = 0, $172 = 0.0, $173 = 0, $174 = 0.0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0.0;
 var $182 = 0.0, $183 = 0.0, $184 = 0.0, $185 = 0.0, $186 = 0.0, $187 = 0.0, $188 = 0.0, $189 = 0, $19 = 0, $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0.0, $198 = 0, $199 = 0, $2 = 0;
 var $20 = 0, $200 = 0.0, $201 = 0.0, $202 = 0.0, $203 = 0.0, $204 = 0.0, $205 = 0.0, $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0.0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0;
 var $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0.0, $229 = 0, $23 = 0, $230 = 0.0, $231 = 0, $232 = 0, $233 = 0.0, $234 = 0.0, $235 = 0;
 var $236 = 0.0, $237 = 0.0, $238 = 0.0, $239 = 0.0, $24 = 0, $240 = 0.0, $241 = 0.0, $242 = 0.0, $243 = 0.0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $contactSolver = 0, $contactSolverDef = 0, $exitcond = 0, $i$016 = 0;
 var $i$02$i = 0, $i1$011 = 0, $i2$08 = 0, $i3$07 = 0, $impulse$i = 0, $j$01$i = 0, $lpad$phi$index = 0, $lpad$phi$index4 = 0, $or$cond = 0, $v$sroa$0$0 = 0, $v$sroa$9$0 = 0, $w$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $impulse$i = sp + 96|0;
 $contactSolverDef = sp + 52|0;
 $contactSolver = sp;
 $0 = ((($this)) + 28|0); //@line 386 "libraries/Box2D/Dynamics/b2Island.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 386 "libraries/Box2D/Dynamics/b2Island.cpp"
 $2 = ($1|0)>($toiIndexA|0); //@line 386 "libraries/Box2D/Dynamics/b2Island.cpp"
 __Z10loveAssertbPKc($2,446522); //@line 386 "libraries/Box2D/Dynamics/b2Island.cpp"
 $3 = HEAP32[$0>>2]|0; //@line 387 "libraries/Box2D/Dynamics/b2Island.cpp"
 $4 = ($3|0)>($toiIndexB|0); //@line 387 "libraries/Box2D/Dynamics/b2Island.cpp"
 __Z10loveAssertbPKc($4,446546); //@line 387 "libraries/Box2D/Dynamics/b2Island.cpp"
 $5 = HEAP32[$0>>2]|0; //@line 390 "libraries/Box2D/Dynamics/b2Island.cpp"
 $6 = ($5|0)>(0); //@line 390 "libraries/Box2D/Dynamics/b2Island.cpp"
 if ($6) {
  $7 = ((($this)) + 8|0); //@line 392 "libraries/Box2D/Dynamics/b2Island.cpp"
  $8 = ((($this)) + 20|0); //@line 393 "libraries/Box2D/Dynamics/b2Island.cpp"
  $9 = ((($this)) + 24|0); //@line 395 "libraries/Box2D/Dynamics/b2Island.cpp"
  $i$016 = 0;
  while(1) {
   $26 = HEAP32[$7>>2]|0; //@line 392 "libraries/Box2D/Dynamics/b2Island.cpp"
   $27 = (($26) + ($i$016<<2)|0); //@line 392 "libraries/Box2D/Dynamics/b2Island.cpp"
   $28 = HEAP32[$27>>2]|0; //@line 392 "libraries/Box2D/Dynamics/b2Island.cpp"
   $29 = HEAP32[$8>>2]|0; //@line 393 "libraries/Box2D/Dynamics/b2Island.cpp"
   $30 = (($29) + (($i$016*12)|0)|0); //@line 393 "libraries/Box2D/Dynamics/b2Island.cpp"
   $31 = ((($28)) + 44|0); //@line 393 "libraries/Box2D/Dynamics/b2Island.cpp"
   $32 = $31; //@line 393 "libraries/Box2D/Dynamics/b2Island.cpp"
   $33 = $32; //@line 393 "libraries/Box2D/Dynamics/b2Island.cpp"
   $34 = HEAP32[$33>>2]|0; //@line 393 "libraries/Box2D/Dynamics/b2Island.cpp"
   $35 = (($32) + 4)|0; //@line 393 "libraries/Box2D/Dynamics/b2Island.cpp"
   $36 = $35; //@line 393 "libraries/Box2D/Dynamics/b2Island.cpp"
   $37 = HEAP32[$36>>2]|0; //@line 393 "libraries/Box2D/Dynamics/b2Island.cpp"
   $38 = $30; //@line 393 "libraries/Box2D/Dynamics/b2Island.cpp"
   $39 = $38; //@line 393 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$39>>2] = $34; //@line 393 "libraries/Box2D/Dynamics/b2Island.cpp"
   $40 = (($38) + 4)|0; //@line 393 "libraries/Box2D/Dynamics/b2Island.cpp"
   $41 = $40; //@line 393 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$41>>2] = $37; //@line 393 "libraries/Box2D/Dynamics/b2Island.cpp"
   $42 = ((($28)) + 56|0); //@line 394 "libraries/Box2D/Dynamics/b2Island.cpp"
   $43 = HEAP32[$42>>2]|0; //@line 394 "libraries/Box2D/Dynamics/b2Island.cpp"
   $44 = HEAP32[$8>>2]|0; //@line 394 "libraries/Box2D/Dynamics/b2Island.cpp"
   $45 = (((($44) + (($i$016*12)|0)|0)) + 8|0); //@line 394 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$45>>2] = $43; //@line 394 "libraries/Box2D/Dynamics/b2Island.cpp"
   $46 = HEAP32[$9>>2]|0; //@line 395 "libraries/Box2D/Dynamics/b2Island.cpp"
   $47 = (($46) + (($i$016*12)|0)|0); //@line 395 "libraries/Box2D/Dynamics/b2Island.cpp"
   $48 = ((($28)) + 64|0); //@line 395 "libraries/Box2D/Dynamics/b2Island.cpp"
   $49 = $48; //@line 395 "libraries/Box2D/Dynamics/b2Island.cpp"
   $50 = $49; //@line 395 "libraries/Box2D/Dynamics/b2Island.cpp"
   $51 = HEAP32[$50>>2]|0; //@line 395 "libraries/Box2D/Dynamics/b2Island.cpp"
   $52 = (($49) + 4)|0; //@line 395 "libraries/Box2D/Dynamics/b2Island.cpp"
   $53 = $52; //@line 395 "libraries/Box2D/Dynamics/b2Island.cpp"
   $54 = HEAP32[$53>>2]|0; //@line 395 "libraries/Box2D/Dynamics/b2Island.cpp"
   $55 = $47; //@line 395 "libraries/Box2D/Dynamics/b2Island.cpp"
   $56 = $55; //@line 395 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$56>>2] = $51; //@line 395 "libraries/Box2D/Dynamics/b2Island.cpp"
   $57 = (($55) + 4)|0; //@line 395 "libraries/Box2D/Dynamics/b2Island.cpp"
   $58 = $57; //@line 395 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$58>>2] = $54; //@line 395 "libraries/Box2D/Dynamics/b2Island.cpp"
   $59 = ((($28)) + 72|0); //@line 396 "libraries/Box2D/Dynamics/b2Island.cpp"
   $60 = HEAP32[$59>>2]|0; //@line 396 "libraries/Box2D/Dynamics/b2Island.cpp"
   $61 = HEAP32[$9>>2]|0; //@line 396 "libraries/Box2D/Dynamics/b2Island.cpp"
   $62 = (((($61) + (($i$016*12)|0)|0)) + 8|0); //@line 396 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$62>>2] = $60; //@line 396 "libraries/Box2D/Dynamics/b2Island.cpp"
   $63 = (($i$016) + 1)|0; //@line 390 "libraries/Box2D/Dynamics/b2Island.cpp"
   $64 = HEAP32[$0>>2]|0; //@line 390 "libraries/Box2D/Dynamics/b2Island.cpp"
   $65 = ($63|0)<($64|0); //@line 390 "libraries/Box2D/Dynamics/b2Island.cpp"
   if ($65) {
    $i$016 = $63;
   } else {
    $$lcssa = $61;
    break;
   }
  }
  $10 = $$lcssa; //@line 390 "libraries/Box2D/Dynamics/b2Island.cpp"
  $$pre$phi22Z2D = $9;$$pre$phiZ2D = $8;$22 = $10;
 } else {
  $$phi$trans$insert = ((($this)) + 24|0); //@line 405 "libraries/Box2D/Dynamics/b2Island.cpp"
  $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 405 "libraries/Box2D/Dynamics/b2Island.cpp"
  $$pre21 = ((($this)) + 20|0); //@line 404 "libraries/Box2D/Dynamics/b2Island.cpp"
  $$pre$phi22Z2D = $$phi$trans$insert;$$pre$phiZ2D = $$pre21;$22 = $$pre;
 }
 $11 = ((($this)) + 12|0); //@line 400 "libraries/Box2D/Dynamics/b2Island.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 400 "libraries/Box2D/Dynamics/b2Island.cpp"
 $13 = ((($contactSolverDef)) + 24|0); //@line 400 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$13>>2] = $12; //@line 400 "libraries/Box2D/Dynamics/b2Island.cpp"
 $14 = ((($this)) + 36|0); //@line 401 "libraries/Box2D/Dynamics/b2Island.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 401 "libraries/Box2D/Dynamics/b2Island.cpp"
 $16 = ((($contactSolverDef)) + 28|0); //@line 401 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$16>>2] = $15; //@line 401 "libraries/Box2D/Dynamics/b2Island.cpp"
 $17 = HEAP32[$this>>2]|0; //@line 402 "libraries/Box2D/Dynamics/b2Island.cpp"
 $18 = ((($contactSolverDef)) + 40|0); //@line 402 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$18>>2] = $17; //@line 402 "libraries/Box2D/Dynamics/b2Island.cpp"
 ;HEAP32[$contactSolverDef>>2]=HEAP32[$subStep>>2]|0;HEAP32[$contactSolverDef+4>>2]=HEAP32[$subStep+4>>2]|0;HEAP32[$contactSolverDef+8>>2]=HEAP32[$subStep+8>>2]|0;HEAP32[$contactSolverDef+12>>2]=HEAP32[$subStep+12>>2]|0;HEAP32[$contactSolverDef+16>>2]=HEAP32[$subStep+16>>2]|0;HEAP32[$contactSolverDef+20>>2]=HEAP32[$subStep+20>>2]|0; //@line 403 "libraries/Box2D/Dynamics/b2Island.cpp"
 $19 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 404 "libraries/Box2D/Dynamics/b2Island.cpp"
 $20 = ((($contactSolverDef)) + 32|0); //@line 404 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$20>>2] = $19; //@line 404 "libraries/Box2D/Dynamics/b2Island.cpp"
 $21 = ((($contactSolverDef)) + 36|0); //@line 405 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$21>>2] = $22; //@line 405 "libraries/Box2D/Dynamics/b2Island.cpp"
 __ZN15b2ContactSolverC2EP18b2ContactSolverDef($contactSolver,$contactSolverDef); //@line 406 "libraries/Box2D/Dynamics/b2Island.cpp"
 $23 = ((($subStep)) + 16|0); //@line 409 "libraries/Box2D/Dynamics/b2Island.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 409 "libraries/Box2D/Dynamics/b2Island.cpp"
 $25 = ($24|0)>(0); //@line 409 "libraries/Box2D/Dynamics/b2Island.cpp"
 L8: do {
  if ($25) {
   $i1$011 = 0;
   while(1) {
    __THREW__ = 0;
    $66 = (invoke_iiii(4245,($contactSolver|0),($toiIndexA|0),($toiIndexB|0))|0); //@line 411 "libraries/Box2D/Dynamics/b2Island.cpp"
    $67 = __THREW__; __THREW__ = 0;
    $68 = $67&1;
    if ($68) {
     break;
    }
    $69 = (($i1$011) + 1)|0; //@line 409 "libraries/Box2D/Dynamics/b2Island.cpp"
    $$not = $66 ^ 1;
    $70 = HEAP32[$23>>2]|0; //@line 409 "libraries/Box2D/Dynamics/b2Island.cpp"
    $71 = ($69|0)<($70|0); //@line 409 "libraries/Box2D/Dynamics/b2Island.cpp"
    $or$cond = $71 & $$not;
    if ($or$cond) {
     $i1$011 = $69;
    } else {
     break L8;
    }
   }
   $76 = ___cxa_find_matching_catch_2()|0;
   $77 = tempRet0;
   $lpad$phi$index = $76;$lpad$phi$index4 = $77;
   __ZN15b2ContactSolverD2Ev($contactSolver); //@line 514 "libraries/Box2D/Dynamics/b2Island.cpp"
   ___resumeException($lpad$phi$index|0);
   // unreachable;
  }
 } while(0);
 $80 = ((($this)) + 8|0); //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 $81 = HEAP32[$80>>2]|0; //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 $82 = (($81) + ($toiIndexA<<2)|0); //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 $83 = HEAP32[$82>>2]|0; //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 $84 = ((($83)) + 36|0); //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 $85 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 $86 = (($85) + (($toiIndexA*12)|0)|0); //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 $87 = $86; //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 $88 = $87; //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 $89 = HEAP32[$88>>2]|0; //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 $90 = (($87) + 4)|0; //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 $91 = $90; //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 $92 = HEAP32[$91>>2]|0; //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 $93 = $84; //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 $94 = $93; //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$94>>2] = $89; //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 $95 = (($93) + 4)|0; //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 $96 = $95; //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$96>>2] = $92; //@line 452 "libraries/Box2D/Dynamics/b2Island.cpp"
 $97 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 453 "libraries/Box2D/Dynamics/b2Island.cpp"
 $98 = (((($97) + (($toiIndexA*12)|0)|0)) + 8|0); //@line 453 "libraries/Box2D/Dynamics/b2Island.cpp"
 $99 = HEAP32[$98>>2]|0; //@line 453 "libraries/Box2D/Dynamics/b2Island.cpp"
 $100 = HEAP32[$80>>2]|0; //@line 453 "libraries/Box2D/Dynamics/b2Island.cpp"
 $101 = (($100) + ($toiIndexA<<2)|0); //@line 453 "libraries/Box2D/Dynamics/b2Island.cpp"
 $102 = HEAP32[$101>>2]|0; //@line 453 "libraries/Box2D/Dynamics/b2Island.cpp"
 $103 = ((($102)) + 52|0); //@line 453 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$103>>2] = $99; //@line 453 "libraries/Box2D/Dynamics/b2Island.cpp"
 $104 = (($100) + ($toiIndexB<<2)|0); //@line 454 "libraries/Box2D/Dynamics/b2Island.cpp"
 $105 = HEAP32[$104>>2]|0; //@line 454 "libraries/Box2D/Dynamics/b2Island.cpp"
 $106 = ((($105)) + 36|0); //@line 454 "libraries/Box2D/Dynamics/b2Island.cpp"
 $107 = (($97) + (($toiIndexB*12)|0)|0); //@line 454 "libraries/Box2D/Dynamics/b2Island.cpp"
 $108 = $107; //@line 454 "libraries/Box2D/Dynamics/b2Island.cpp"
 $109 = $108; //@line 454 "libraries/Box2D/Dynamics/b2Island.cpp"
 $110 = HEAP32[$109>>2]|0; //@line 454 "libraries/Box2D/Dynamics/b2Island.cpp"
 $111 = (($108) + 4)|0; //@line 454 "libraries/Box2D/Dynamics/b2Island.cpp"
 $112 = $111; //@line 454 "libraries/Box2D/Dynamics/b2Island.cpp"
 $113 = HEAP32[$112>>2]|0; //@line 454 "libraries/Box2D/Dynamics/b2Island.cpp"
 $114 = $106; //@line 454 "libraries/Box2D/Dynamics/b2Island.cpp"
 $115 = $114; //@line 454 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$115>>2] = $110; //@line 454 "libraries/Box2D/Dynamics/b2Island.cpp"
 $116 = (($114) + 4)|0; //@line 454 "libraries/Box2D/Dynamics/b2Island.cpp"
 $117 = $116; //@line 454 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$117>>2] = $113; //@line 454 "libraries/Box2D/Dynamics/b2Island.cpp"
 $118 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 455 "libraries/Box2D/Dynamics/b2Island.cpp"
 $119 = (((($118) + (($toiIndexB*12)|0)|0)) + 8|0); //@line 455 "libraries/Box2D/Dynamics/b2Island.cpp"
 $120 = HEAP32[$119>>2]|0; //@line 455 "libraries/Box2D/Dynamics/b2Island.cpp"
 $121 = HEAP32[$80>>2]|0; //@line 455 "libraries/Box2D/Dynamics/b2Island.cpp"
 $122 = (($121) + ($toiIndexB<<2)|0); //@line 455 "libraries/Box2D/Dynamics/b2Island.cpp"
 $123 = HEAP32[$122>>2]|0; //@line 455 "libraries/Box2D/Dynamics/b2Island.cpp"
 $124 = ((($123)) + 52|0); //@line 455 "libraries/Box2D/Dynamics/b2Island.cpp"
 HEAP32[$124>>2] = $120; //@line 455 "libraries/Box2D/Dynamics/b2Island.cpp"
 __THREW__ = 0;
 invoke_vi(4240,($contactSolver|0)); //@line 459 "libraries/Box2D/Dynamics/b2Island.cpp"
 $125 = __THREW__; __THREW__ = 0;
 $126 = $125&1;
 if ($126) {
  $78 = ___cxa_find_matching_catch_2()|0;
  $79 = tempRet0;
  $lpad$phi$index = $78;$lpad$phi$index4 = $79;
  __ZN15b2ContactSolverD2Ev($contactSolver); //@line 514 "libraries/Box2D/Dynamics/b2Island.cpp"
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 $127 = ((($subStep)) + 12|0); //@line 462 "libraries/Box2D/Dynamics/b2Island.cpp"
 $128 = HEAP32[$127>>2]|0; //@line 462 "libraries/Box2D/Dynamics/b2Island.cpp"
 $129 = ($128|0)>(0); //@line 462 "libraries/Box2D/Dynamics/b2Island.cpp"
 L19: do {
  if ($129) {
   $i2$08 = 0;
   while(1) {
    __THREW__ = 0;
    invoke_vi(4243,($contactSolver|0)); //@line 464 "libraries/Box2D/Dynamics/b2Island.cpp"
    $133 = __THREW__; __THREW__ = 0;
    $134 = $133&1;
    if ($134) {
     break;
    }
    $135 = (($i2$08) + 1)|0; //@line 462 "libraries/Box2D/Dynamics/b2Island.cpp"
    $136 = HEAP32[$127>>2]|0; //@line 462 "libraries/Box2D/Dynamics/b2Island.cpp"
    $137 = ($135|0)<($136|0); //@line 462 "libraries/Box2D/Dynamics/b2Island.cpp"
    if ($137) {
     $i2$08 = $135;
    } else {
     break L19;
    }
   }
   $74 = ___cxa_find_matching_catch_2()|0;
   $75 = tempRet0;
   $lpad$phi$index = $74;$lpad$phi$index4 = $75;
   __ZN15b2ContactSolverD2Ev($contactSolver); //@line 514 "libraries/Box2D/Dynamics/b2Island.cpp"
   ___resumeException($lpad$phi$index|0);
   // unreachable;
  }
 } while(0);
 $130 = +HEAPF32[$subStep>>2]; //@line 470 "libraries/Box2D/Dynamics/b2Island.cpp"
 $131 = HEAP32[$0>>2]|0; //@line 473 "libraries/Box2D/Dynamics/b2Island.cpp"
 $132 = ($131|0)>(0); //@line 473 "libraries/Box2D/Dynamics/b2Island.cpp"
 if ($132) {
  $i3$07 = 0;
  while(1) {
   $168 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 499 "libraries/Box2D/Dynamics/b2Island.cpp"
   $169 = (($168) + (($i3$07*12)|0)|0); //@line 475 "libraries/Box2D/Dynamics/b2Island.cpp"
   $170 = +HEAPF32[$169>>2]; //@line 475 "libraries/Box2D/Dynamics/b2Island.cpp"
   $171 = (((($168) + (($i3$07*12)|0)|0)) + 4|0); //@line 475 "libraries/Box2D/Dynamics/b2Island.cpp"
   $172 = +HEAPF32[$171>>2]; //@line 475 "libraries/Box2D/Dynamics/b2Island.cpp"
   $173 = (((($168) + (($i3$07*12)|0)|0)) + 8|0); //@line 476 "libraries/Box2D/Dynamics/b2Island.cpp"
   $174 = +HEAPF32[$173>>2]; //@line 476 "libraries/Box2D/Dynamics/b2Island.cpp"
   $175 = HEAP32[$$pre$phi22Z2D>>2]|0; //@line 477 "libraries/Box2D/Dynamics/b2Island.cpp"
   $176 = (($175) + (($i3$07*12)|0)|0); //@line 477 "libraries/Box2D/Dynamics/b2Island.cpp"
   $177 = HEAP32[$176>>2]|0; //@line 477 "libraries/Box2D/Dynamics/b2Island.cpp"
   $178 = (((($175) + (($i3$07*12)|0)|0)) + 4|0); //@line 477 "libraries/Box2D/Dynamics/b2Island.cpp"
   $179 = HEAP32[$178>>2]|0; //@line 477 "libraries/Box2D/Dynamics/b2Island.cpp"
   $180 = (((($175) + (($i3$07*12)|0)|0)) + 8|0); //@line 478 "libraries/Box2D/Dynamics/b2Island.cpp"
   $181 = +HEAPF32[$180>>2]; //@line 478 "libraries/Box2D/Dynamics/b2Island.cpp"
   $182 = (HEAP32[tempDoublePtr>>2]=$177,+HEAPF32[tempDoublePtr>>2]); //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $183 = $130 * $182; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $184 = (HEAP32[tempDoublePtr>>2]=$179,+HEAPF32[tempDoublePtr>>2]); //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $185 = $130 * $184; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $186 = $183 * $183; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $187 = $185 * $185; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $188 = $186 + $187; //@line 408 "./libraries/Box2D/Common/b2Math.h"
   $189 = $188 > 4.0; //@line 482 "libraries/Box2D/Dynamics/b2Island.cpp"
   if ($189) {
    $190 = (+Math_sqrt((+$188))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
    $191 = 2.0 / $190; //@line 484 "libraries/Box2D/Dynamics/b2Island.cpp"
    $192 = $182 * $191; //@line 97 "./libraries/Box2D/Common/b2Math.h"
    $193 = (HEAPF32[tempDoublePtr>>2]=$192,HEAP32[tempDoublePtr>>2]|0); //@line 97 "./libraries/Box2D/Common/b2Math.h"
    $194 = $184 * $191; //@line 97 "./libraries/Box2D/Common/b2Math.h"
    $195 = (HEAPF32[tempDoublePtr>>2]=$194,HEAP32[tempDoublePtr>>2]|0); //@line 97 "./libraries/Box2D/Common/b2Math.h"
    $v$sroa$0$0 = $193;$v$sroa$9$0 = $195;
   } else {
    $v$sroa$0$0 = $177;$v$sroa$9$0 = $179;
   }
   $196 = $130 * $181; //@line 488 "libraries/Box2D/Dynamics/b2Island.cpp"
   $197 = $196 * $196; //@line 489 "libraries/Box2D/Dynamics/b2Island.cpp"
   $198 = $197 > 2.4674012660980225; //@line 489 "libraries/Box2D/Dynamics/b2Island.cpp"
   if ($198) {
    $199 = $196 > 0.0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
    $200 = -$196; //@line 618 "./libraries/Box2D/Common/b2Math.h"
    $201 = $199 ? $196 : $200; //@line 618 "./libraries/Box2D/Common/b2Math.h"
    $202 = 1.5707963705062866 / $201; //@line 491 "libraries/Box2D/Dynamics/b2Island.cpp"
    $203 = $181 * $202; //@line 492 "libraries/Box2D/Dynamics/b2Island.cpp"
    $w$0 = $203;
   } else {
    $w$0 = $181;
   }
   $204 = (HEAP32[tempDoublePtr>>2]=$v$sroa$0$0,+HEAPF32[tempDoublePtr>>2]); //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $205 = $130 * $204; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $206 = (HEAP32[tempDoublePtr>>2]=$v$sroa$9$0,+HEAPF32[tempDoublePtr>>2]); //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $207 = $130 * $206; //@line 459 "./libraries/Box2D/Common/b2Math.h"
   $208 = $170 + $205; //@line 85 "./libraries/Box2D/Common/b2Math.h"
   $209 = $172 + $207; //@line 85 "./libraries/Box2D/Common/b2Math.h"
   $210 = $130 * $w$0; //@line 497 "libraries/Box2D/Dynamics/b2Island.cpp"
   $211 = $174 + $210; //@line 497 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAPF32[$169>>2] = $208; //@line 499 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAPF32[$171>>2] = $209; //@line 499 "libraries/Box2D/Dynamics/b2Island.cpp"
   $212 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 500 "libraries/Box2D/Dynamics/b2Island.cpp"
   $213 = (((($212) + (($i3$07*12)|0)|0)) + 8|0); //@line 500 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAPF32[$213>>2] = $211; //@line 500 "libraries/Box2D/Dynamics/b2Island.cpp"
   $214 = HEAP32[$$pre$phi22Z2D>>2]|0; //@line 501 "libraries/Box2D/Dynamics/b2Island.cpp"
   $215 = (($214) + (($i3$07*12)|0)|0); //@line 501 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$215>>2] = $v$sroa$0$0; //@line 501 "libraries/Box2D/Dynamics/b2Island.cpp"
   $216 = (((($214) + (($i3$07*12)|0)|0)) + 4|0); //@line 501 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$216>>2] = $v$sroa$9$0; //@line 501 "libraries/Box2D/Dynamics/b2Island.cpp"
   $217 = HEAP32[$$pre$phi22Z2D>>2]|0; //@line 502 "libraries/Box2D/Dynamics/b2Island.cpp"
   $218 = (((($217) + (($i3$07*12)|0)|0)) + 8|0); //@line 502 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAPF32[$218>>2] = $w$0; //@line 502 "libraries/Box2D/Dynamics/b2Island.cpp"
   $219 = HEAP32[$80>>2]|0; //@line 505 "libraries/Box2D/Dynamics/b2Island.cpp"
   $220 = (($219) + ($i3$07<<2)|0); //@line 505 "libraries/Box2D/Dynamics/b2Island.cpp"
   $221 = HEAP32[$220>>2]|0; //@line 505 "libraries/Box2D/Dynamics/b2Island.cpp"
   $222 = ((($221)) + 44|0); //@line 506 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAPF32[$222>>2] = $208; //@line 506 "libraries/Box2D/Dynamics/b2Island.cpp"
   $223 = ((($221)) + 48|0); //@line 506 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAPF32[$223>>2] = $209; //@line 506 "libraries/Box2D/Dynamics/b2Island.cpp"
   $224 = ((($221)) + 56|0); //@line 507 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAPF32[$224>>2] = $211; //@line 507 "libraries/Box2D/Dynamics/b2Island.cpp"
   $225 = ((($221)) + 64|0); //@line 508 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$225>>2] = $v$sroa$0$0; //@line 508 "libraries/Box2D/Dynamics/b2Island.cpp"
   $226 = ((($221)) + 68|0); //@line 508 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAP32[$226>>2] = $v$sroa$9$0; //@line 508 "libraries/Box2D/Dynamics/b2Island.cpp"
   $227 = ((($221)) + 72|0); //@line 509 "libraries/Box2D/Dynamics/b2Island.cpp"
   HEAPF32[$227>>2] = $w$0; //@line 509 "libraries/Box2D/Dynamics/b2Island.cpp"
   $228 = (+Math_sin((+$211))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
   $229 = ((($221)) + 20|0); //@line 315 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$229>>2] = $228; //@line 315 "./libraries/Box2D/Common/b2Math.h"
   $230 = (+Math_cos((+$211))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
   $231 = ((($221)) + 24|0); //@line 316 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$231>>2] = $230; //@line 316 "./libraries/Box2D/Common/b2Math.h"
   $232 = ((($221)) + 28|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $233 = +HEAPF32[$232>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $234 = $230 * $233; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $235 = ((($221)) + 32|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $236 = +HEAPF32[$235>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $237 = $228 * $236; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $238 = $234 - $237; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $239 = $228 * $233; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $240 = $230 * $236; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $241 = $239 + $240; //@line 568 "./libraries/Box2D/Common/b2Math.h"
   $242 = $208 - $238; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $243 = $209 - $241; //@line 454 "./libraries/Box2D/Common/b2Math.h"
   $244 = ((($221)) + 12|0); //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAPF32[$244>>2] = $242; //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
   $245 = ((($221)) + 16|0); //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
   HEAPF32[$245>>2] = $243; //@line 837 "./libraries/Box2D/Dynamics/b2Body.h"
   $246 = (($i3$07) + 1)|0; //@line 473 "libraries/Box2D/Dynamics/b2Island.cpp"
   $247 = HEAP32[$0>>2]|0; //@line 473 "libraries/Box2D/Dynamics/b2Island.cpp"
   $248 = ($246|0)<($247|0); //@line 473 "libraries/Box2D/Dynamics/b2Island.cpp"
   if ($248) {
    $i3$07 = $246;
   } else {
    break;
   }
  }
 }
 $138 = ((($contactSolver)) + 40|0); //@line 513 "libraries/Box2D/Dynamics/b2Island.cpp"
 $139 = HEAP32[$138>>2]|0; //@line 513 "libraries/Box2D/Dynamics/b2Island.cpp"
 $140 = ((($this)) + 4|0); //@line 518 "libraries/Box2D/Dynamics/b2Island.cpp"
 $141 = HEAP32[$140>>2]|0; //@line 518 "libraries/Box2D/Dynamics/b2Island.cpp"
 $142 = ($141|0)==(0|0); //@line 518 "libraries/Box2D/Dynamics/b2Island.cpp"
 if ($142) {
  __ZN15b2ContactSolverD2Ev($contactSolver); //@line 514 "libraries/Box2D/Dynamics/b2Island.cpp"
  STACKTOP = sp;return; //@line 514 "libraries/Box2D/Dynamics/b2Island.cpp"
 }
 $143 = HEAP32[$14>>2]|0; //@line 523 "libraries/Box2D/Dynamics/b2Island.cpp"
 $144 = ($143|0)>(0); //@line 523 "libraries/Box2D/Dynamics/b2Island.cpp"
 if (!($144)) {
  __ZN15b2ContactSolverD2Ev($contactSolver); //@line 514 "libraries/Box2D/Dynamics/b2Island.cpp"
  STACKTOP = sp;return; //@line 514 "libraries/Box2D/Dynamics/b2Island.cpp"
 }
 $145 = ((($impulse$i)) + 16|0); //@line 530 "libraries/Box2D/Dynamics/b2Island.cpp"
 $i$02$i = 0;
 while(1) {
  $146 = HEAP32[$11>>2]|0; //@line 525 "libraries/Box2D/Dynamics/b2Island.cpp"
  $147 = (($146) + ($i$02$i<<2)|0); //@line 525 "libraries/Box2D/Dynamics/b2Island.cpp"
  $148 = HEAP32[$147>>2]|0; //@line 525 "libraries/Box2D/Dynamics/b2Island.cpp"
  $149 = (((($139) + (($i$02$i*156)|0)|0)) + 148|0); //@line 530 "libraries/Box2D/Dynamics/b2Island.cpp"
  $150 = HEAP32[$149>>2]|0; //@line 531 "libraries/Box2D/Dynamics/b2Island.cpp"
  HEAP32[$145>>2] = $150; //@line 530 "libraries/Box2D/Dynamics/b2Island.cpp"
  $151 = ($150|0)>(0); //@line 531 "libraries/Box2D/Dynamics/b2Island.cpp"
  if ($151) {
   $j$01$i = 0;
   while(1) {
    $161 = ((((($139) + (($i$02$i*156)|0)|0) + (($j$01$i*36)|0)|0)) + 16|0); //@line 533 "libraries/Box2D/Dynamics/b2Island.cpp"
    $162 = HEAP32[$161>>2]|0; //@line 533 "libraries/Box2D/Dynamics/b2Island.cpp"
    $163 = (($impulse$i) + ($j$01$i<<2)|0); //@line 533 "libraries/Box2D/Dynamics/b2Island.cpp"
    HEAP32[$163>>2] = $162; //@line 533 "libraries/Box2D/Dynamics/b2Island.cpp"
    $164 = ((((($139) + (($i$02$i*156)|0)|0) + (($j$01$i*36)|0)|0)) + 20|0); //@line 534 "libraries/Box2D/Dynamics/b2Island.cpp"
    $165 = HEAP32[$164>>2]|0; //@line 534 "libraries/Box2D/Dynamics/b2Island.cpp"
    $166 = (((($impulse$i)) + 8|0) + ($j$01$i<<2)|0); //@line 534 "libraries/Box2D/Dynamics/b2Island.cpp"
    HEAP32[$166>>2] = $165; //@line 534 "libraries/Box2D/Dynamics/b2Island.cpp"
    $167 = (($j$01$i) + 1)|0; //@line 531 "libraries/Box2D/Dynamics/b2Island.cpp"
    $exitcond = ($167|0)==($150|0); //@line 531 "libraries/Box2D/Dynamics/b2Island.cpp"
    if ($exitcond) {
     break;
    } else {
     $j$01$i = $167;
    }
   }
  }
  $152 = HEAP32[$140>>2]|0; //@line 537 "libraries/Box2D/Dynamics/b2Island.cpp"
  $153 = HEAP32[$152>>2]|0; //@line 537 "libraries/Box2D/Dynamics/b2Island.cpp"
  $154 = ((($153)) + 20|0); //@line 537 "libraries/Box2D/Dynamics/b2Island.cpp"
  $155 = HEAP32[$154>>2]|0; //@line 537 "libraries/Box2D/Dynamics/b2Island.cpp"
  __THREW__ = 0;
  invoke_viii($155|0,($152|0),($148|0),($impulse$i|0)); //@line 537 "libraries/Box2D/Dynamics/b2Island.cpp"
  $156 = __THREW__; __THREW__ = 0;
  $157 = $156&1;
  if ($157) {
   break;
  }
  $158 = (($i$02$i) + 1)|0; //@line 523 "libraries/Box2D/Dynamics/b2Island.cpp"
  $159 = HEAP32[$14>>2]|0; //@line 523 "libraries/Box2D/Dynamics/b2Island.cpp"
  $160 = ($158|0)<($159|0); //@line 523 "libraries/Box2D/Dynamics/b2Island.cpp"
  if ($160) {
   $i$02$i = $158;
  } else {
   label = 31;
   break;
  }
 }
 if ((label|0) == 31) {
  __ZN15b2ContactSolverD2Ev($contactSolver); //@line 514 "libraries/Box2D/Dynamics/b2Island.cpp"
  STACKTOP = sp;return; //@line 514 "libraries/Box2D/Dynamics/b2Island.cpp"
 }
 $72 = ___cxa_find_matching_catch_2()|0;
 $73 = tempRet0;
 $lpad$phi$index = $72;$lpad$phi$index4 = $73;
 __ZN15b2ContactSolverD2Ev($contactSolver); //@line 514 "libraries/Box2D/Dynamics/b2Island.cpp"
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZN23b2ChainAndCircleContact8EvaluateEP10b2ManifoldRK11b2TransformS4_($this,$manifold,$xfA,$xfB) {
 $this = $this|0;
 $manifold = $manifold|0;
 $xfA = $xfA|0;
 $xfB = $xfB|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $edge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $edge = sp;
 $0 = ((($this)) + 48|0); //@line 48 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 48 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 $2 = ((($1)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $3 = HEAP32[$2>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 HEAP32[$edge>>2] = (20456); //@line 63 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $4 = ((($edge)) + 4|0); //@line 64 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 HEAP32[$4>>2] = 1; //@line 64 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $5 = ((($edge)) + 8|0); //@line 65 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 HEAPF32[$5>>2] = 0.0099999997764825821; //@line 65 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $6 = ((($edge)) + 28|0); //@line 66 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;HEAP32[$6+12>>2]=0|0;HEAP16[$6+16>>1]=0|0; //@line 67 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $7 = ((($this)) + 56|0); //@line 50 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 50 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 __ZNK12b2ChainShape12GetChildEdgeEP11b2EdgeShapei($3,$edge,$8); //@line 50 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 $9 = ((($this)) + 52|0); //@line 52 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 52 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 $11 = ((($10)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $12 = HEAP32[$11>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 __Z22b2CollideEdgeAndCircleP10b2ManifoldPK11b2EdgeShapeRK11b2TransformPK13b2CircleShapeS6_($manifold,$edge,$xfA,$12,$xfB); //@line 51 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 STACKTOP = sp;return; //@line 53 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
}
function __ZN23b2ChainAndCircleContactD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 34 "./libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.h"
 return; //@line 34 "./libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.h"
}
function __ZN23b2ChainAndCircleContact6CreateEP9b2FixtureiS1_iP16b2BlockAllocator($fixtureA,$indexA,$fixtureB,$indexB,$allocator) {
 $fixtureA = $fixtureA|0;
 $indexA = $indexA|0;
 $fixtureB = $fixtureB|0;
 $indexB = $indexB|0;
 $allocator = $allocator|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN16b2BlockAllocator8AllocateEi($allocator,148)|0); //@line 29 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 __ZN9b2ContactC2EP9b2FixtureiS1_i($0,$fixtureA,$indexA,$fixtureB,$indexB); //@line 40 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 HEAP32[$0>>2] = (33164); //@line 41 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 $1 = ((($0)) + 48|0); //@line 42 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 42 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 $3 = ((($2)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $4 = HEAP32[$3>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $5 = ((($4)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $6 = HEAP32[$5>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $7 = ($6|0)==(3); //@line 42 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 __Z10loveAssertbPKc($7,446623); //@line 42 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 $8 = ((($0)) + 52|0); //@line 43 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 43 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 $10 = ((($9)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $11 = HEAP32[$10>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $12 = ((($11)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $13 = HEAP32[$12>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $14 = ($13|0)==(0); //@line 43 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 __Z10loveAssertbPKc($14,446770); //@line 43 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 return ($0|0); //@line 30 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
}
function __ZN23b2ChainAndCircleContact7DestroyEP9b2ContactP16b2BlockAllocator($contact,$allocator) {
 $contact = $contact|0;
 $allocator = $allocator|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$contact>>2]|0; //@line 35 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 $1 = ((($0)) + 4|0); //@line 35 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 35 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 FUNCTION_TABLE_vi[$2 & 8191]($contact); //@line 35 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 __ZN16b2BlockAllocator4FreeEPvi($allocator,$contact,148); //@line 36 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
 return; //@line 37 "libraries/Box2D/Dynamics/Contacts/b2ChainAndCircleContact.cpp"
}
function __ZN24b2ChainAndPolygonContact8EvaluateEP10b2ManifoldRK11b2TransformS4_($this,$manifold,$xfA,$xfB) {
 $this = $this|0;
 $manifold = $manifold|0;
 $xfA = $xfA|0;
 $xfB = $xfB|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $edge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $edge = sp;
 $0 = ((($this)) + 48|0); //@line 48 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 48 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 $2 = ((($1)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $3 = HEAP32[$2>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 HEAP32[$edge>>2] = (20456); //@line 63 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $4 = ((($edge)) + 4|0); //@line 64 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 HEAP32[$4>>2] = 1; //@line 64 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $5 = ((($edge)) + 8|0); //@line 65 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 HEAPF32[$5>>2] = 0.0099999997764825821; //@line 65 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $6 = ((($edge)) + 28|0); //@line 66 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;HEAP32[$6+12>>2]=0|0;HEAP16[$6+16>>1]=0|0; //@line 67 "./libraries/Box2D/Collision/Shapes/b2EdgeShape.h"
 $7 = ((($this)) + 56|0); //@line 50 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 50 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 __ZNK12b2ChainShape12GetChildEdgeEP11b2EdgeShapei($3,$edge,$8); //@line 50 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 $9 = ((($this)) + 52|0); //@line 52 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 52 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 $11 = ((($10)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $12 = HEAP32[$11>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 __Z23b2CollideEdgeAndPolygonP10b2ManifoldPK11b2EdgeShapeRK11b2TransformPK14b2PolygonShapeS6_($manifold,$edge,$xfA,$12,$xfB); //@line 51 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 STACKTOP = sp;return; //@line 53 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
}
function __ZN24b2ChainAndPolygonContactD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 34 "./libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.h"
 return; //@line 34 "./libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.h"
}
function __ZN24b2ChainAndPolygonContact6CreateEP9b2FixtureiS1_iP16b2BlockAllocator($fixtureA,$indexA,$fixtureB,$indexB,$allocator) {
 $fixtureA = $fixtureA|0;
 $indexA = $indexA|0;
 $fixtureB = $fixtureB|0;
 $indexB = $indexB|0;
 $allocator = $allocator|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN16b2BlockAllocator8AllocateEi($allocator,148)|0); //@line 29 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 __ZN9b2ContactC2EP9b2FixtureiS1_i($0,$fixtureA,$indexA,$fixtureB,$indexB); //@line 40 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 HEAP32[$0>>2] = (33184); //@line 41 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 $1 = ((($0)) + 48|0); //@line 42 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 42 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 $3 = ((($2)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $4 = HEAP32[$3>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $5 = ((($4)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $6 = HEAP32[$5>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $7 = ($6|0)==(3); //@line 42 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 __Z10loveAssertbPKc($7,446623); //@line 42 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 $8 = ((($0)) + 52|0); //@line 43 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 43 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 $10 = ((($9)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $11 = HEAP32[$10>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $12 = ((($11)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $13 = HEAP32[$12>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $14 = ($13|0)==(2); //@line 43 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 __Z10loveAssertbPKc($14,446665); //@line 43 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 return ($0|0); //@line 30 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
}
function __ZN24b2ChainAndPolygonContact7DestroyEP9b2ContactP16b2BlockAllocator($contact,$allocator) {
 $contact = $contact|0;
 $allocator = $allocator|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$contact>>2]|0; //@line 35 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 $1 = ((($0)) + 4|0); //@line 35 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 35 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 FUNCTION_TABLE_vi[$2 & 8191]($contact); //@line 35 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 __ZN16b2BlockAllocator4FreeEPvi($allocator,$contact,148); //@line 36 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
 return; //@line 37 "libraries/Box2D/Dynamics/Contacts/b2ChainAndPolygonContact.cpp"
}
function __ZN15b2CircleContact8EvaluateEP10b2ManifoldRK11b2TransformS4_($this,$manifold,$xfA,$xfB) {
 $this = $this|0;
 $manifold = $manifold|0;
 $xfA = $xfA|0;
 $xfB = $xfB|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 50 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 50 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 $2 = ((($1)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $3 = HEAP32[$2>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $4 = ((($this)) + 52|0); //@line 51 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 51 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 $6 = ((($5)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $7 = HEAP32[$6>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
 __Z16b2CollideCirclesP10b2ManifoldPK13b2CircleShapeRK11b2TransformS3_S6_($manifold,$3,$xfA,$7,$xfB); //@line 49 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 return; //@line 52 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
}
function __ZN15b2CircleContactD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 34 "./libraries/Box2D/Dynamics/Contacts/b2CircleContact.h"
 return; //@line 34 "./libraries/Box2D/Dynamics/Contacts/b2CircleContact.h"
}
function __ZN15b2CircleContact6CreateEP9b2FixtureiS1_iP16b2BlockAllocator($fixtureA,$0,$fixtureB,$1,$allocator) {
 $fixtureA = $fixtureA|0;
 $0 = $0|0;
 $fixtureB = $fixtureB|0;
 $1 = $1|0;
 $allocator = $allocator|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZN16b2BlockAllocator8AllocateEi($allocator,148)|0); //@line 30 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 __ZN9b2ContactC2EP9b2FixtureiS1_i($2,$fixtureA,0,$fixtureB,0); //@line 41 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 HEAP32[$2>>2] = (33204); //@line 42 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 $3 = ((($2)) + 48|0); //@line 43 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 43 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 $5 = ((($4)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $6 = HEAP32[$5>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $7 = ((($6)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $8 = HEAP32[$7>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $9 = ($8|0)==(0); //@line 43 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 __Z10loveAssertbPKc($9,446727); //@line 43 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 $10 = ((($2)) + 52|0); //@line 44 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 44 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 $12 = ((($11)) + 12|0); //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $13 = HEAP32[$12>>2]|0; //@line 240 "./libraries/Box2D/Dynamics/b2Fixture.h"
 $14 = ((($13)) + 4|0); //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $15 = HEAP32[$14>>2]|0; //@line 98 "./libraries/Box2D/Collision/Shapes/b2Shape.h"
 $16 = ($15|0)==(0); //@line 44 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 __Z10loveAssertbPKc($16,446770); //@line 44 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 return ($2|0); //@line 31 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
}
function __ZN15b2CircleContact7DestroyEP9b2ContactP16b2BlockAllocator($contact,$allocator) {
 $contact = $contact|0;
 $allocator = $allocator|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$contact>>2]|0; //@line 36 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 $1 = ((($0)) + 4|0); //@line 36 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 36 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 FUNCTION_TABLE_vi[$2 & 8191]($contact); //@line 36 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 __ZN16b2BlockAllocator4FreeEPvi($allocator,$contact,148); //@line 37 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
 return; //@line 38 "libraries/Box2D/Dynamics/Contacts/b2CircleContact.cpp"
}
function __ZN15b2ContactSolverC2EP18b2ContactSolverDef($this,$def) {
 $this = $this|0;
 $def = $def|0;
 var $$cast = 0, $$pre = 0, $$sink = 0.0, $$sink7 = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0.0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $i$02 = 0, $j$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$this>>2]=HEAP32[$def>>2]|0;HEAP32[$this+4>>2]=HEAP32[$def+4>>2]|0;HEAP32[$this+8>>2]=HEAP32[$def+8>>2]|0;HEAP32[$this+12>>2]=HEAP32[$def+12>>2]|0;HEAP32[$this+16>>2]=HEAP32[$def+16>>2]|0;HEAP32[$this+20>>2]=HEAP32[$def+20>>2]|0; //@line 48 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $0 = ((($def)) + 40|0); //@line 49 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 49 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $2 = ((($this)) + 32|0); //@line 49 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 HEAP32[$2>>2] = $1; //@line 49 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $3 = ((($def)) + 28|0); //@line 50 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 50 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $5 = ((($this)) + 48|0); //@line 50 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 HEAP32[$5>>2] = $4; //@line 50 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $$cast = $1; //@line 51 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $6 = ($4*88)|0; //@line 51 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $7 = (__ZN16b2StackAllocator8AllocateEi($$cast,$6)|0); //@line 51 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $8 = ((($this)) + 36|0); //@line 51 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 HEAP32[$8>>2] = $7; //@line 51 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $9 = HEAP32[$2>>2]|0; //@line 52 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $10 = HEAP32[$5>>2]|0; //@line 52 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $11 = ($10*156)|0; //@line 52 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $12 = (__ZN16b2StackAllocator8AllocateEi($9,$11)|0); //@line 52 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $13 = ((($this)) + 40|0); //@line 52 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 HEAP32[$13>>2] = $12; //@line 52 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $14 = ((($def)) + 32|0); //@line 53 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 53 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $16 = ((($this)) + 24|0); //@line 53 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 HEAP32[$16>>2] = $15; //@line 53 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $17 = ((($def)) + 36|0); //@line 54 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 54 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $19 = ((($this)) + 28|0); //@line 54 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 HEAP32[$19>>2] = $18; //@line 54 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $20 = ((($def)) + 24|0); //@line 55 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 55 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $22 = ((($this)) + 44|0); //@line 55 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 HEAP32[$22>>2] = $21; //@line 55 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $23 = HEAP32[$5>>2]|0; //@line 58 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $24 = ($23|0)>(0); //@line 58 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 if (!($24)) {
  return; //@line 131 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 }
 $25 = $21; //@line 58 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $26 = ((($this)) + 20|0); //@line 111 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $27 = ((($this)) + 8|0); //@line 113 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $29 = $25;$i$02 = 0;
 while(1) {
  $28 = (($29) + ($i$02<<2)|0); //@line 60 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $30 = HEAP32[$28>>2]|0; //@line 60 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $31 = ((($30)) + 48|0); //@line 62 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $32 = HEAP32[$31>>2]|0; //@line 62 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $33 = ((($30)) + 52|0); //@line 63 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $34 = HEAP32[$33>>2]|0; //@line 63 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $35 = ((($32)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $36 = HEAP32[$35>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $37 = ((($34)) + 12|0); //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $38 = HEAP32[$37>>2]|0; //@line 245 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $39 = ((($36)) + 8|0); //@line 66 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $40 = HEAP32[$39>>2]|0; //@line 66 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $41 = ((($38)) + 8|0); //@line 67 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $42 = HEAP32[$41>>2]|0; //@line 67 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $43 = ((($32)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $44 = HEAP32[$43>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $45 = ((($34)) + 8|0); //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $46 = HEAP32[$45>>2]|0; //@line 275 "./libraries/Box2D/Dynamics/b2Fixture.h"
  $47 = ((($30)) + 124|0); //@line 72 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $48 = HEAP32[$47>>2]|0; //@line 72 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $49 = ($48|0)>(0); //@line 73 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  __Z10loveAssertbPKc($49,446814); //@line 73 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $50 = HEAP32[$13>>2]|0; //@line 75 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $51 = ((($30)) + 136|0); //@line 76 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $52 = HEAP32[$51>>2]|0; //@line 76 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $53 = (((($50) + (($i$02*156)|0)|0)) + 136|0); //@line 76 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$53>>2] = $52; //@line 76 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $54 = ((($30)) + 140|0); //@line 77 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $55 = HEAP32[$54>>2]|0; //@line 77 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $56 = (((($50) + (($i$02*156)|0)|0)) + 140|0); //@line 77 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$56>>2] = $55; //@line 77 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $57 = ((($30)) + 144|0); //@line 78 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $58 = HEAP32[$57>>2]|0; //@line 78 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $59 = (((($50) + (($i$02*156)|0)|0)) + 144|0); //@line 78 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$59>>2] = $58; //@line 78 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $60 = ((($44)) + 8|0); //@line 79 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $61 = HEAP32[$60>>2]|0; //@line 79 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $62 = (((($50) + (($i$02*156)|0)|0)) + 112|0); //@line 79 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$62>>2] = $61; //@line 79 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $63 = ((($46)) + 8|0); //@line 80 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $64 = HEAP32[$63>>2]|0; //@line 80 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $65 = (((($50) + (($i$02*156)|0)|0)) + 116|0); //@line 80 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$65>>2] = $64; //@line 80 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $66 = ((($44)) + 120|0); //@line 81 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $67 = HEAP32[$66>>2]|0; //@line 81 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $68 = (((($50) + (($i$02*156)|0)|0)) + 120|0); //@line 81 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$68>>2] = $67; //@line 81 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $69 = ((($46)) + 120|0); //@line 82 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $70 = HEAP32[$69>>2]|0; //@line 82 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $71 = (((($50) + (($i$02*156)|0)|0)) + 124|0); //@line 82 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$71>>2] = $70; //@line 82 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $72 = ((($44)) + 128|0); //@line 83 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $73 = HEAP32[$72>>2]|0; //@line 83 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $74 = (((($50) + (($i$02*156)|0)|0)) + 128|0); //@line 83 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$74>>2] = $73; //@line 83 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $75 = ((($46)) + 128|0); //@line 84 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $76 = HEAP32[$75>>2]|0; //@line 84 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $77 = (((($50) + (($i$02*156)|0)|0)) + 132|0); //@line 84 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$77>>2] = $76; //@line 84 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $78 = (((($50) + (($i$02*156)|0)|0)) + 152|0); //@line 85 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$78>>2] = $i$02; //@line 85 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $79 = (((($50) + (($i$02*156)|0)|0)) + 148|0); //@line 86 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$79>>2] = $48; //@line 86 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $80 = (((($50) + (($i$02*156)|0)|0)) + 80|0); //@line 88 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  ;HEAP32[$80>>2]=0|0;HEAP32[$80+4>>2]=0|0;HEAP32[$80+8>>2]=0|0;HEAP32[$80+12>>2]=0|0;HEAP32[$80+16>>2]=0|0;HEAP32[$80+20>>2]=0|0;HEAP32[$80+24>>2]=0|0;HEAP32[$80+28>>2]=0|0; //@line 219 "./libraries/Box2D/Common/b2Math.h"
  $81 = HEAP32[$8>>2]|0; //@line 90 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $82 = HEAP32[$60>>2]|0; //@line 91 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $83 = (((($81) + (($i$02*88)|0)|0)) + 32|0); //@line 91 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$83>>2] = $82; //@line 91 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $84 = HEAP32[$63>>2]|0; //@line 92 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $85 = (((($81) + (($i$02*88)|0)|0)) + 36|0); //@line 92 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$85>>2] = $84; //@line 92 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $86 = HEAP32[$66>>2]|0; //@line 93 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $87 = (((($81) + (($i$02*88)|0)|0)) + 40|0); //@line 93 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$87>>2] = $86; //@line 93 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $88 = HEAP32[$69>>2]|0; //@line 94 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $89 = (((($81) + (($i$02*88)|0)|0)) + 44|0); //@line 94 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$89>>2] = $88; //@line 94 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $90 = (((($81) + (($i$02*88)|0)|0)) + 48|0); //@line 95 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $91 = ((($44)) + 28|0); //@line 95 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $92 = $91; //@line 95 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $93 = $92; //@line 95 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $94 = HEAP32[$93>>2]|0; //@line 95 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $95 = (($92) + 4)|0; //@line 95 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $96 = $95; //@line 95 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $97 = HEAP32[$96>>2]|0; //@line 95 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $98 = $90; //@line 95 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $99 = $98; //@line 95 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$99>>2] = $94; //@line 95 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $100 = (($98) + 4)|0; //@line 95 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $101 = $100; //@line 95 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$101>>2] = $97; //@line 95 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $102 = (((($81) + (($i$02*88)|0)|0)) + 56|0); //@line 96 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $103 = ((($46)) + 28|0); //@line 96 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $104 = $103; //@line 96 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $105 = $104; //@line 96 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $106 = HEAP32[$105>>2]|0; //@line 96 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $107 = (($104) + 4)|0; //@line 96 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $108 = $107; //@line 96 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $109 = HEAP32[$108>>2]|0; //@line 96 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $110 = $102; //@line 96 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $111 = $110; //@line 96 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$111>>2] = $106; //@line 96 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $112 = (($110) + 4)|0; //@line 96 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $113 = $112; //@line 96 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$113>>2] = $109; //@line 96 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $114 = HEAP32[$72>>2]|0; //@line 97 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $115 = (((($81) + (($i$02*88)|0)|0)) + 64|0); //@line 97 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$115>>2] = $114; //@line 97 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $116 = HEAP32[$75>>2]|0; //@line 98 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $117 = (((($81) + (($i$02*88)|0)|0)) + 68|0); //@line 98 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$117>>2] = $116; //@line 98 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $118 = (((($81) + (($i$02*88)|0)|0)) + 16|0); //@line 99 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $119 = ((($30)) + 104|0); //@line 99 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $120 = $119; //@line 99 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $121 = $120; //@line 99 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $122 = HEAP32[$121>>2]|0; //@line 99 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $123 = (($120) + 4)|0; //@line 99 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $124 = $123; //@line 99 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $125 = HEAP32[$124>>2]|0; //@line 99 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $126 = $118; //@line 99 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $127 = $126; //@line 99 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$127>>2] = $122; //@line 99 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $128 = (($126) + 4)|0; //@line 99 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $129 = $128; //@line 99 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$129>>2] = $125; //@line 99 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $130 = (((($81) + (($i$02*88)|0)|0)) + 24|0); //@line 100 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $131 = ((($30)) + 112|0); //@line 100 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $132 = $131; //@line 100 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $133 = $132; //@line 100 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $134 = HEAP32[$133>>2]|0; //@line 100 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $135 = (($132) + 4)|0; //@line 100 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $136 = $135; //@line 100 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $137 = HEAP32[$136>>2]|0; //@line 100 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $138 = $130; //@line 100 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $139 = $138; //@line 100 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$139>>2] = $134; //@line 100 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $140 = (($138) + 4)|0; //@line 100 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $141 = $140; //@line 100 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$141>>2] = $137; //@line 100 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $142 = (((($81) + (($i$02*88)|0)|0)) + 84|0); //@line 101 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$142>>2] = $48; //@line 101 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $143 = (((($81) + (($i$02*88)|0)|0)) + 76|0); //@line 102 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$143>>2] = $40; //@line 102 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $144 = (((($81) + (($i$02*88)|0)|0)) + 80|0); //@line 103 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$144>>2] = $42; //@line 103 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $145 = ((($30)) + 120|0); //@line 104 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $146 = HEAP32[$145>>2]|0; //@line 104 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $147 = (((($81) + (($i$02*88)|0)|0)) + 72|0); //@line 104 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$147>>2] = $146; //@line 104 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  if ($49) {
   $j$01 = 0;
   while(1) {
    $151 = (((($30)) + 64|0) + (($j$01*20)|0)|0); //@line 108 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $152 = HEAP8[$26>>0]|0; //@line 111 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $153 = ($152<<24>>24)==(0); //@line 111 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    if ($153) {
     $$sink = 0.0;$$sink7 = 0.0;
    } else {
     $154 = +HEAPF32[$27>>2]; //@line 113 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $155 = (((((($30)) + 64|0) + (($j$01*20)|0)|0)) + 8|0); //@line 113 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $156 = +HEAPF32[$155>>2]; //@line 113 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $157 = $154 * $156; //@line 113 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $158 = (((((($30)) + 64|0) + (($j$01*20)|0)|0)) + 12|0); //@line 114 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $159 = +HEAPF32[$158>>2]; //@line 114 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $160 = $154 * $159; //@line 114 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $$sink = $160;$$sink7 = $157;
    }
    $161 = ((((($50) + (($i$02*156)|0)|0) + (($j$01*36)|0)|0)) + 16|0); //@line 118 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    HEAPF32[$161>>2] = $$sink7; //@line 118 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $162 = ((((($50) + (($i$02*156)|0)|0) + (($j$01*36)|0)|0)) + 20|0); //@line 119 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    HEAPF32[$162>>2] = $$sink; //@line 119 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $163 = ((($50) + (($i$02*156)|0)|0) + (($j$01*36)|0)|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
    $164 = ((((($50) + (($i$02*156)|0)|0) + (($j$01*36)|0)|0)) + 24|0); //@line 124 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    HEAPF32[$164>>2] = 0.0; //@line 124 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $165 = ((((($50) + (($i$02*156)|0)|0) + (($j$01*36)|0)|0)) + 28|0); //@line 125 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    HEAPF32[$165>>2] = 0.0; //@line 125 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $166 = ((((($50) + (($i$02*156)|0)|0) + (($j$01*36)|0)|0)) + 32|0); //@line 126 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    HEAPF32[$166>>2] = 0.0; //@line 126 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $167 = ((($81) + (($i$02*88)|0)|0) + ($j$01<<3)|0); //@line 128 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    ;HEAP32[$163>>2]=0|0;HEAP32[$163+4>>2]=0|0;HEAP32[$163+8>>2]=0|0;HEAP32[$163+12>>2]=0|0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
    $168 = $151; //@line 128 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $169 = $168; //@line 128 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $170 = HEAP32[$169>>2]|0; //@line 128 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $171 = (($168) + 4)|0; //@line 128 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $172 = $171; //@line 128 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $173 = HEAP32[$172>>2]|0; //@line 128 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $174 = $167; //@line 128 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $175 = $174; //@line 128 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    HEAP32[$175>>2] = $170; //@line 128 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $176 = (($174) + 4)|0; //@line 128 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $177 = $176; //@line 128 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    HEAP32[$177>>2] = $173; //@line 128 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $178 = (($j$01) + 1)|0; //@line 106 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $exitcond = ($178|0)==($48|0); //@line 106 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    if ($exitcond) {
     break;
    } else {
     $j$01 = $178;
    }
   }
  }
  $148 = (($i$02) + 1)|0; //@line 58 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $149 = HEAP32[$5>>2]|0; //@line 58 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $150 = ($148|0)<($149|0); //@line 58 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  if (!($150)) {
   break;
  }
  $$pre = HEAP32[$22>>2]|0; //@line 60 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $29 = $$pre;$i$02 = $148;
 }
 return; //@line 131 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
}
function __ZN15b2ContactSolverD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 135 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 135 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $2 = ((($this)) + 40|0); //@line 135 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 135 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 __THREW__ = 0;
 invoke_vii(3821,($1|0),($3|0)); //@line 135 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $11 = ___cxa_find_matching_catch_3(0|0)|0;
  $12 = tempRet0;
  ___clang_call_terminate($11); //@line 135 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  // unreachable; //@line 135 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 }
 $6 = HEAP32[$0>>2]|0; //@line 136 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $7 = ((($this)) + 36|0); //@line 136 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 136 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 __THREW__ = 0;
 invoke_vii(3821,($6|0),($8|0)); //@line 136 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $11 = ___cxa_find_matching_catch_3(0|0)|0;
  $12 = tempRet0;
  ___clang_call_terminate($11); //@line 135 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  // unreachable; //@line 135 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 } else {
  return; //@line 137 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 }
}
function __ZN15b2ContactSolver29InitializeVelocityConstraintsEv($this) {
 $this = $this|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0.0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0.0, $117 = 0.0, $118 = 0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0, $128 = 0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0.0, $132 = 0.0;
 var $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0, $144 = 0.0, $145 = 0.0, $146 = 0, $147 = 0.0, $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0.0;
 var $151 = 0.0, $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0, $163 = 0.0, $164 = 0.0, $165 = 0, $166 = 0, $167 = 0.0, $168 = 0.0, $169 = 0.0;
 var $17 = 0.0, $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0.0, $182 = 0, $183 = 0.0, $184 = 0.0, $185 = 0.0, $186 = 0, $187 = 0;
 var $188 = 0.0, $189 = 0, $19 = 0.0, $190 = 0.0, $191 = 0.0, $192 = 0, $193 = 0.0, $194 = 0.0, $195 = 0.0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0, $201 = 0.0, $202 = 0.0, $203 = 0.0, $204 = 0;
 var $205 = 0.0, $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0.0, $213 = 0.0, $214 = 0, $215 = 0.0, $216 = 0.0, $217 = 0.0, $218 = 0.0, $219 = 0.0, $22 = 0, $220 = 0.0, $221 = 0.0, $222 = 0.0;
 var $223 = 0.0, $224 = 0.0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0.0, $229 = 0.0, $23 = 0, $230 = 0.0, $231 = 0.0, $232 = 0.0, $233 = 0.0, $234 = 0.0, $235 = 0.0, $236 = 0.0, $237 = 0.0, $238 = 0.0, $239 = 0.0, $24 = 0, $240 = 0;
 var $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0.0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0, $43 = 0.0, $44 = 0, $45 = 0.0;
 var $46 = 0, $47 = 0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0.0, $53 = 0, $54 = 0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0, $63 = 0.0;
 var $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0, $73 = 0, $74 = 0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0;
 var $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $det$0$i = 0.0;
 var $exitcond = 0, $i$034 = 0, $j$033 = 0, $or$cond = 0, $worldManifold = 0, $xfA = 0, $xfB = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $xfA = sp + 48|0;
 $xfB = sp + 32|0;
 $worldManifold = sp;
 $0 = ((($this)) + 48|0); //@line 142 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 142 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $2 = ($1|0)>(0); //@line 142 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 if (!($2)) {
  STACKTOP = sp;return; //@line 249 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 }
 $3 = ((($this)) + 40|0); //@line 144 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $4 = ((($this)) + 36|0); //@line 145 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $5 = ((($this)) + 44|0); //@line 149 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $6 = ((($this)) + 24|0); //@line 161 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $7 = ((($this)) + 28|0); //@line 163 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $8 = ((($xfA)) + 8|0); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($xfA)) + 12|0); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($xfB)) + 8|0); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($xfB)) + 12|0); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $12 = ((($xfA)) + 4|0); //@line 176 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $13 = ((($xfB)) + 4|0); //@line 177 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $i$034 = 0;
 while(1) {
  $14 = HEAP32[$3>>2]|0; //@line 144 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $15 = HEAP32[$4>>2]|0; //@line 145 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $16 = (((($15) + (($i$034*88)|0)|0)) + 76|0); //@line 147 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $17 = +HEAPF32[$16>>2]; //@line 147 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $18 = (((($15) + (($i$034*88)|0)|0)) + 80|0); //@line 148 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $19 = +HEAPF32[$18>>2]; //@line 148 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $20 = (((($14) + (($i$034*156)|0)|0)) + 152|0); //@line 149 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $21 = HEAP32[$20>>2]|0; //@line 149 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $22 = HEAP32[$5>>2]|0; //@line 149 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $23 = (($22) + ($21<<2)|0); //@line 149 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $24 = HEAP32[$23>>2]|0; //@line 149 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $25 = ((($24)) + 64|0); //@line 224 "./libraries/Box2D/Dynamics/Contacts/b2Contact.h"
  $26 = (((($14) + (($i$034*156)|0)|0)) + 112|0); //@line 151 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $27 = HEAP32[$26>>2]|0; //@line 151 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $28 = (((($14) + (($i$034*156)|0)|0)) + 116|0); //@line 152 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $29 = HEAP32[$28>>2]|0; //@line 152 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $30 = (((($14) + (($i$034*156)|0)|0)) + 120|0); //@line 154 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $31 = +HEAPF32[$30>>2]; //@line 154 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $32 = (((($14) + (($i$034*156)|0)|0)) + 124|0); //@line 155 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $33 = +HEAPF32[$32>>2]; //@line 155 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $34 = (((($14) + (($i$034*156)|0)|0)) + 128|0); //@line 156 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $35 = +HEAPF32[$34>>2]; //@line 156 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $36 = (((($14) + (($i$034*156)|0)|0)) + 132|0); //@line 157 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $37 = +HEAPF32[$36>>2]; //@line 157 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $38 = (((($15) + (($i$034*88)|0)|0)) + 48|0); //@line 158 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $39 = +HEAPF32[$38>>2]; //@line 158 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $40 = (((($15) + (($i$034*88)|0)|0)) + 52|0); //@line 158 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $41 = +HEAPF32[$40>>2]; //@line 158 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $42 = (((($15) + (($i$034*88)|0)|0)) + 56|0); //@line 159 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $43 = +HEAPF32[$42>>2]; //@line 159 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $44 = (((($15) + (($i$034*88)|0)|0)) + 60|0); //@line 159 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $45 = +HEAPF32[$44>>2]; //@line 159 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $46 = HEAP32[$6>>2]|0; //@line 161 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $47 = (($46) + (($27*12)|0)|0); //@line 161 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $48 = +HEAPF32[$47>>2]; //@line 161 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $49 = (((($46) + (($27*12)|0)|0)) + 4|0); //@line 161 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $50 = +HEAPF32[$49>>2]; //@line 161 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $51 = (((($46) + (($27*12)|0)|0)) + 8|0); //@line 162 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $52 = +HEAPF32[$51>>2]; //@line 162 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $53 = HEAP32[$7>>2]|0; //@line 163 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $54 = (($53) + (($27*12)|0)|0); //@line 163 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $55 = +HEAPF32[$54>>2]; //@line 163 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $56 = (((($53) + (($27*12)|0)|0)) + 4|0); //@line 163 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $57 = +HEAPF32[$56>>2]; //@line 163 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $58 = (((($53) + (($27*12)|0)|0)) + 8|0); //@line 164 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $59 = +HEAPF32[$58>>2]; //@line 164 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $60 = (($46) + (($29*12)|0)|0); //@line 166 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $61 = +HEAPF32[$60>>2]; //@line 166 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $62 = (((($46) + (($29*12)|0)|0)) + 4|0); //@line 166 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $63 = +HEAPF32[$62>>2]; //@line 166 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $64 = (((($46) + (($29*12)|0)|0)) + 8|0); //@line 167 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $65 = +HEAPF32[$64>>2]; //@line 167 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $66 = (($53) + (($29*12)|0)|0); //@line 168 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $67 = +HEAPF32[$66>>2]; //@line 168 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $68 = (((($53) + (($29*12)|0)|0)) + 4|0); //@line 168 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $69 = +HEAPF32[$68>>2]; //@line 168 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $70 = (((($53) + (($29*12)|0)|0)) + 8|0); //@line 169 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $71 = +HEAPF32[$70>>2]; //@line 169 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $72 = ((($24)) + 124|0); //@line 171 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $73 = HEAP32[$72>>2]|0; //@line 171 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $74 = ($73|0)>(0); //@line 171 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  __Z10loveAssertbPKc($74,446829); //@line 171 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $75 = (+Math_sin((+$52))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$8>>2] = $75; //@line 315 "./libraries/Box2D/Common/b2Math.h"
  $76 = (+Math_cos((+$52))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$9>>2] = $76; //@line 316 "./libraries/Box2D/Common/b2Math.h"
  $77 = (+Math_sin((+$65))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$10>>2] = $77; //@line 315 "./libraries/Box2D/Common/b2Math.h"
  $78 = (+Math_cos((+$65))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$11>>2] = $78; //@line 316 "./libraries/Box2D/Common/b2Math.h"
  $79 = $39 * $76; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $80 = $41 * $75; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $81 = $79 - $80; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $82 = $39 * $75; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $83 = $41 * $76; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $84 = $83 + $82; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $85 = $48 - $81; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $86 = $50 - $84; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$xfA>>2] = $85; //@line 176 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$12>>2] = $86; //@line 176 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $87 = $43 * $78; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $88 = $45 * $77; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $89 = $87 - $88; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $90 = $43 * $77; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $91 = $45 * $78; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $92 = $91 + $90; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $93 = $61 - $89; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $94 = $63 - $92; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$xfB>>2] = $93; //@line 177 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$13>>2] = $94; //@line 177 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  __ZN15b2WorldManifold10InitializeEPK10b2ManifoldRK11b2TransformfS5_f($worldManifold,$25,$xfA,$17,$xfB,$19); //@line 180 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $95 = (((($14) + (($i$034*156)|0)|0)) + 72|0); //@line 182 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $96 = $worldManifold; //@line 182 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $97 = $96; //@line 182 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $98 = HEAP32[$97>>2]|0; //@line 182 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $99 = (($96) + 4)|0; //@line 182 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $100 = $99; //@line 182 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $101 = HEAP32[$100>>2]|0; //@line 182 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $102 = $95; //@line 182 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $103 = $102; //@line 182 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$103>>2] = $98; //@line 182 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $104 = (($102) + 4)|0; //@line 182 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $105 = $104; //@line 182 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$105>>2] = $101; //@line 182 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $106 = (((($14) + (($i$034*156)|0)|0)) + 148|0); //@line 184 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $107 = HEAP32[$106>>2]|0; //@line 184 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $108 = ($107|0)>(0); //@line 185 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  do {
   if ($108) {
    $109 = (((($14) + (($i$034*156)|0)|0)) + 76|0); //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $110 = $31 + $33; //@line 195 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $111 = (((($14) + (($i$034*156)|0)|0)) + 140|0); //@line 213 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $j$033 = 0;
    while(1) {
     $115 = (((($worldManifold)) + 8|0) + ($j$033<<3)|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $116 = +HEAPF32[$115>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $117 = $116 - $48; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $118 = (((((($worldManifold)) + 8|0) + ($j$033<<3)|0)) + 4|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $119 = +HEAPF32[$118>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $120 = $119 - $50; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $121 = ((($14) + (($i$034*156)|0)|0) + (($j$033*36)|0)|0); //@line 189 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     HEAPF32[$121>>2] = $117; //@line 189 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $122 = ((((($14) + (($i$034*156)|0)|0) + (($j$033*36)|0)|0)) + 4|0); //@line 189 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     HEAPF32[$122>>2] = $120; //@line 189 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $123 = +HEAPF32[$115>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $124 = $123 - $61; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $125 = +HEAPF32[$118>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $126 = $125 - $63; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $127 = ((((($14) + (($i$034*156)|0)|0) + (($j$033*36)|0)|0)) + 8|0); //@line 190 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     HEAPF32[$127>>2] = $124; //@line 190 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $128 = ((((($14) + (($i$034*156)|0)|0) + (($j$033*36)|0)|0)) + 12|0); //@line 190 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     HEAPF32[$128>>2] = $126; //@line 190 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $129 = +HEAPF32[$109>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $130 = $117 * $129; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $131 = +HEAPF32[$95>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $132 = $120 * $131; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $133 = $130 - $132; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $134 = $129 * $124; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $135 = $131 * $126; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $136 = $134 - $135; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $137 = $35 * $133; //@line 195 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $138 = $133 * $137; //@line 195 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $139 = $110 + $138; //@line 195 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $140 = $37 * $136; //@line 195 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $141 = $136 * $140; //@line 195 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $142 = $139 + $141; //@line 195 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $143 = $142 > 0.0; //@line 197 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $144 = 1.0 / $142; //@line 197 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $145 = $143 ? $144 : 0.0; //@line 197 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $146 = ((((($14) + (($i$034*156)|0)|0) + (($j$033*36)|0)|0)) + 24|0); //@line 197 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     HEAPF32[$146>>2] = $145; //@line 197 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $147 = +HEAPF32[$109>>2]; //@line 421 "./libraries/Box2D/Common/b2Math.h"
     $148 = +HEAPF32[$95>>2]; //@line 421 "./libraries/Box2D/Common/b2Math.h"
     $149 = -$148; //@line 421 "./libraries/Box2D/Common/b2Math.h"
     $150 = $117 * $149; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $151 = $147 * $120; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $152 = $150 - $151; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $153 = $124 * $149; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $154 = $147 * $126; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $155 = $153 - $154; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $156 = $35 * $152; //@line 204 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $157 = $152 * $156; //@line 204 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $158 = $110 + $157; //@line 204 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $159 = $37 * $155; //@line 204 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $160 = $155 * $159; //@line 204 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $161 = $158 + $160; //@line 204 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $162 = $161 > 0.0; //@line 206 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $163 = 1.0 / $161; //@line 206 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $164 = $162 ? $163 : 0.0; //@line 206 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $165 = ((((($14) + (($i$034*156)|0)|0) + (($j$033*36)|0)|0)) + 28|0); //@line 206 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     HEAPF32[$165>>2] = $164; //@line 206 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $166 = ((((($14) + (($i$034*156)|0)|0) + (($j$033*36)|0)|0)) + 32|0); //@line 209 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     HEAPF32[$166>>2] = 0.0; //@line 209 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $167 = $71 * $126; //@line 428 "./libraries/Box2D/Common/b2Math.h"
     $168 = $71 * $124; //@line 428 "./libraries/Box2D/Common/b2Math.h"
     $169 = $67 - $167; //@line 448 "./libraries/Box2D/Common/b2Math.h"
     $170 = $69 + $168; //@line 448 "./libraries/Box2D/Common/b2Math.h"
     $171 = $169 - $55; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $172 = $170 - $57; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $173 = $59 * $120; //@line 428 "./libraries/Box2D/Common/b2Math.h"
     $174 = $59 * $117; //@line 428 "./libraries/Box2D/Common/b2Math.h"
     $175 = $171 + $173; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $176 = $172 - $174; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $177 = +HEAPF32[$95>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $178 = $177 * $175; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $179 = +HEAPF32[$109>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $180 = $179 * $176; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $181 = $178 + $180; //@line 408 "./libraries/Box2D/Common/b2Math.h"
     $182 = $181 < -1.0; //@line 211 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     if ($182) {
      $183 = +HEAPF32[$111>>2]; //@line 213 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $184 = $181 * $183; //@line 213 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $185 = -$184; //@line 213 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      HEAPF32[$166>>2] = $185; //@line 213 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     }
     $186 = (($j$033) + 1)|0; //@line 185 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $exitcond = ($186|0)==($107|0); //@line 185 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     if ($exitcond) {
      break;
     } else {
      $j$033 = $186;
     }
    }
    $$pre = HEAP32[$106>>2]|0; //@line 218 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $112 = ($$pre|0)!=(2); //@line 218 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $113 = HEAP8[446813]|0; //@line 218 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $114 = ($113<<24>>24)==(0); //@line 218 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $or$cond = $112 | $114; //@line 218 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    if (!($or$cond)) {
     $187 = (($14) + (($i$034*156)|0)|0); //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $188 = +HEAPF32[$187>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $189 = (((($14) + (($i$034*156)|0)|0)) + 76|0); //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $190 = +HEAPF32[$189>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $191 = $188 * $190; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $192 = (((($14) + (($i$034*156)|0)|0)) + 4|0); //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $193 = +HEAPF32[$192>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $194 = +HEAPF32[$95>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $195 = $193 * $194; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $196 = $191 - $195; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $197 = (((($14) + (($i$034*156)|0)|0)) + 8|0); //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $198 = +HEAPF32[$197>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $199 = $190 * $198; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $200 = (((($14) + (($i$034*156)|0)|0)) + 12|0); //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $201 = +HEAPF32[$200>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $202 = $194 * $201; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $203 = $199 - $202; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $204 = (((($14) + (($i$034*156)|0)|0)) + 36|0); //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $205 = +HEAPF32[$204>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $206 = $190 * $205; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $207 = (((($14) + (($i$034*156)|0)|0)) + 40|0); //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $208 = +HEAPF32[$207>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $209 = $194 * $208; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $210 = $206 - $209; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $211 = (((($14) + (($i$034*156)|0)|0)) + 44|0); //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $212 = +HEAPF32[$211>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $213 = $190 * $212; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $214 = (((($14) + (($i$034*156)|0)|0)) + 48|0); //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $215 = +HEAPF32[$214>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $216 = $194 * $215; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $217 = $213 - $216; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $218 = $31 + $33; //@line 228 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $219 = $35 * $196; //@line 228 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $220 = $196 * $219; //@line 228 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $221 = $218 + $220; //@line 228 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $222 = $37 * $203; //@line 228 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $223 = $203 * $222; //@line 228 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $224 = $221 + $223; //@line 228 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $225 = $35 * $210; //@line 229 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $226 = $210 * $225; //@line 229 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $227 = $218 + $226; //@line 229 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $228 = $37 * $217; //@line 229 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $229 = $217 * $228; //@line 229 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $230 = $227 + $229; //@line 229 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $231 = $219 * $210; //@line 230 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $232 = $218 + $231; //@line 230 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $233 = $222 * $217; //@line 230 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $234 = $232 + $233; //@line 230 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $235 = $224 * $224; //@line 234 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $236 = $224 * $230; //@line 234 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $237 = $234 * $234; //@line 234 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $238 = $236 - $237; //@line 234 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $239 = $238 * 1000.0; //@line 234 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $240 = $235 < $239; //@line 234 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     if ($240) {
      $241 = (((($14) + (($i$034*156)|0)|0)) + 96|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$241>>2] = $224; //@line 65 "./libraries/Box2D/Common/b2Math.h"
      $242 = (((($14) + (($i$034*156)|0)|0)) + 100|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$242>>2] = $234; //@line 65 "./libraries/Box2D/Common/b2Math.h"
      $243 = (((($14) + (($i$034*156)|0)|0)) + 104|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$243>>2] = $234; //@line 65 "./libraries/Box2D/Common/b2Math.h"
      $244 = (((($14) + (($i$034*156)|0)|0)) + 108|0); //@line 65 "./libraries/Box2D/Common/b2Math.h"
      HEAPF32[$244>>2] = $230; //@line 65 "./libraries/Box2D/Common/b2Math.h"
      $245 = $238 != 0.0; //@line 228 "./libraries/Box2D/Common/b2Math.h"
      $246 = 1.0 / $238; //@line 230 "./libraries/Box2D/Common/b2Math.h"
      $det$0$i = $245 ? $246 : $238; //@line 228 "./libraries/Box2D/Common/b2Math.h"
      $247 = $230 * $det$0$i; //@line 232 "./libraries/Box2D/Common/b2Math.h"
      $248 = $det$0$i * $234; //@line 232 "./libraries/Box2D/Common/b2Math.h"
      $249 = -$248; //@line 232 "./libraries/Box2D/Common/b2Math.h"
      $250 = $224 * $det$0$i; //@line 233 "./libraries/Box2D/Common/b2Math.h"
      $251 = (((($14) + (($i$034*156)|0)|0)) + 80|0); //@line 239 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      HEAPF32[$251>>2] = $247; //@line 239 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $252 = (((($14) + (($i$034*156)|0)|0)) + 84|0); //@line 239 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      HEAPF32[$252>>2] = $249; //@line 239 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $253 = (((($14) + (($i$034*156)|0)|0)) + 88|0); //@line 239 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      HEAPF32[$253>>2] = $249; //@line 239 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $254 = (((($14) + (($i$034*156)|0)|0)) + 92|0); //@line 239 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      HEAPF32[$254>>2] = $250; //@line 239 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      break;
     } else {
      HEAP32[$106>>2] = 1; //@line 245 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      break;
     }
    }
   }
  } while(0);
  $255 = (($i$034) + 1)|0; //@line 142 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $256 = HEAP32[$0>>2]|0; //@line 142 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $257 = ($255|0)<($256|0); //@line 142 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  if ($257) {
   $i$034 = $255;
  } else {
   break;
  }
 }
 STACKTOP = sp;return; //@line 249 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
}
function __ZN15b2ContactSolver9WarmStartEv($this) {
 $this = $this|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0.0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0;
 var $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0, $74 = 0, $75 = 0.0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0;
 var $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0, $91 = 0, $exitcond = 0, $i$021 = 0, $j$015 = 0, $vA$sroa$0$0$lcssa = 0, $vA$sroa$0$011 = 0, $vA$sroa$6$0$lcssa = 0, $vA$sroa$6$012 = 0;
 var $vB$sroa$0$0$lcssa = 0, $vB$sroa$0$09 = 0, $vB$sroa$6$0$lcssa = 0, $vB$sroa$6$010 = 0, $wA$0$lcssa = 0.0, $wA$013 = 0.0, $wB$0$lcssa = 0.0, $wB$014 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 254 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 254 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $2 = ($1|0)>(0); //@line 254 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 if (!($2)) {
  return; //@line 289 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 }
 $3 = ((($this)) + 40|0); //@line 256 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $4 = ((($this)) + 28|0); //@line 266 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $$pre = HEAP32[$4>>2]|0; //@line 266 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $21 = $$pre;$i$021 = 0;
 while(1) {
  $5 = HEAP32[$3>>2]|0; //@line 256 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $6 = (((($5) + (($i$021*156)|0)|0)) + 112|0); //@line 258 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 258 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $8 = (((($5) + (($i$021*156)|0)|0)) + 116|0); //@line 259 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 259 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $10 = (((($5) + (($i$021*156)|0)|0)) + 120|0); //@line 260 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $11 = +HEAPF32[$10>>2]; //@line 260 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $12 = (((($5) + (($i$021*156)|0)|0)) + 128|0); //@line 261 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $13 = +HEAPF32[$12>>2]; //@line 261 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $14 = (((($5) + (($i$021*156)|0)|0)) + 124|0); //@line 262 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $15 = +HEAPF32[$14>>2]; //@line 262 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $16 = (((($5) + (($i$021*156)|0)|0)) + 132|0); //@line 263 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $17 = +HEAPF32[$16>>2]; //@line 263 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $18 = (((($5) + (($i$021*156)|0)|0)) + 148|0); //@line 264 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 264 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $20 = (($21) + (($7*12)|0)|0); //@line 266 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $22 = HEAP32[$20>>2]|0; //@line 266 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $23 = (((($21) + (($7*12)|0)|0)) + 4|0); //@line 266 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $24 = HEAP32[$23>>2]|0; //@line 266 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $25 = (((($21) + (($7*12)|0)|0)) + 8|0); //@line 267 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $26 = +HEAPF32[$25>>2]; //@line 267 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $27 = (($21) + (($9*12)|0)|0); //@line 268 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $28 = HEAP32[$27>>2]|0; //@line 268 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $29 = (((($21) + (($9*12)|0)|0)) + 4|0); //@line 268 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $30 = HEAP32[$29>>2]|0; //@line 268 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $31 = (((($21) + (($9*12)|0)|0)) + 8|0); //@line 269 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $32 = +HEAPF32[$31>>2]; //@line 269 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $33 = (((($5) + (($i$021*156)|0)|0)) + 72|0); //@line 271 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $34 = +HEAPF32[$33>>2]; //@line 271 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $35 = (((($5) + (($i$021*156)|0)|0)) + 76|0); //@line 271 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $36 = +HEAPF32[$35>>2]; //@line 271 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $37 = ($19|0)>(0); //@line 274 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  if ($37) {
   $j$015 = 0;$vA$sroa$0$011 = $22;$vA$sroa$6$012 = $24;$vB$sroa$0$09 = $28;$vB$sroa$6$010 = $30;$wA$013 = $26;$wB$014 = $32;
   while(1) {
    $47 = ((((($5) + (($i$021*156)|0)|0) + (($j$015*36)|0)|0)) + 16|0); //@line 277 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $48 = +HEAPF32[$47>>2]; //@line 277 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $49 = $34 * $48; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $50 = $36 * $48; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $51 = ((((($5) + (($i$021*156)|0)|0) + (($j$015*36)|0)|0)) + 20|0); //@line 277 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $52 = +HEAPF32[$51>>2]; //@line 277 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $53 = $36 * $52; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $54 = $34 * $52; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $55 = $49 + $53; //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $56 = $50 - $54; //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $57 = ((($5) + (($i$021*156)|0)|0) + (($j$015*36)|0)|0); //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $58 = +HEAPF32[$57>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $59 = $58 * $56; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $60 = ((((($5) + (($i$021*156)|0)|0) + (($j$015*36)|0)|0)) + 4|0); //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $61 = +HEAPF32[$60>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $62 = $61 * $55; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $63 = $59 - $62; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $64 = $13 * $63; //@line 278 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $65 = $wA$013 - $64; //@line 278 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $66 = $11 * $55; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $67 = $11 * $56; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $68 = (HEAP32[tempDoublePtr>>2]=$vA$sroa$0$011,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $69 = $68 - $66; //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $70 = (HEAPF32[tempDoublePtr>>2]=$69,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $71 = (HEAP32[tempDoublePtr>>2]=$vA$sroa$6$012,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $72 = $71 - $67; //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $73 = (HEAPF32[tempDoublePtr>>2]=$72,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $74 = ((((($5) + (($i$021*156)|0)|0) + (($j$015*36)|0)|0)) + 8|0); //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $75 = +HEAPF32[$74>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $76 = $56 * $75; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $77 = ((((($5) + (($i$021*156)|0)|0) + (($j$015*36)|0)|0)) + 12|0); //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $78 = +HEAPF32[$77>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $79 = $55 * $78; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $80 = $76 - $79; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $81 = $17 * $80; //@line 280 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $82 = $wB$014 + $81; //@line 280 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $83 = $15 * $55; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $84 = $15 * $56; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $85 = (HEAP32[tempDoublePtr>>2]=$vB$sroa$0$09,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $86 = $85 + $83; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $87 = (HEAPF32[tempDoublePtr>>2]=$86,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $88 = (HEAP32[tempDoublePtr>>2]=$vB$sroa$6$010,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $89 = $88 + $84; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $90 = (HEAPF32[tempDoublePtr>>2]=$89,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $91 = (($j$015) + 1)|0; //@line 274 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $exitcond = ($91|0)==($19|0); //@line 274 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    if ($exitcond) {
     $vA$sroa$0$0$lcssa = $70;$vA$sroa$6$0$lcssa = $73;$vB$sroa$0$0$lcssa = $87;$vB$sroa$6$0$lcssa = $90;$wA$0$lcssa = $65;$wB$0$lcssa = $82;
     break;
    } else {
     $j$015 = $91;$vA$sroa$0$011 = $70;$vA$sroa$6$012 = $73;$vB$sroa$0$09 = $87;$vB$sroa$6$010 = $90;$wA$013 = $65;$wB$014 = $82;
    }
   }
  } else {
   $vA$sroa$0$0$lcssa = $22;$vA$sroa$6$0$lcssa = $24;$vB$sroa$0$0$lcssa = $28;$vB$sroa$6$0$lcssa = $30;$wA$0$lcssa = $26;$wB$0$lcssa = $32;
  }
  HEAP32[$20>>2] = $vA$sroa$0$0$lcssa; //@line 284 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$23>>2] = $vA$sroa$6$0$lcssa; //@line 284 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $38 = HEAP32[$4>>2]|0; //@line 285 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $39 = (((($38) + (($7*12)|0)|0)) + 8|0); //@line 285 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$39>>2] = $wA$0$lcssa; //@line 285 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $40 = (($38) + (($9*12)|0)|0); //@line 286 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$40>>2] = $vB$sroa$0$0$lcssa; //@line 286 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $41 = (((($38) + (($9*12)|0)|0)) + 4|0); //@line 286 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$41>>2] = $vB$sroa$6$0$lcssa; //@line 286 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $42 = HEAP32[$4>>2]|0; //@line 287 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $43 = (((($42) + (($9*12)|0)|0)) + 8|0); //@line 287 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$43>>2] = $wB$0$lcssa; //@line 287 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $44 = (($i$021) + 1)|0; //@line 254 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $45 = HEAP32[$0>>2]|0; //@line 254 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $46 = ($44|0)<($45|0); //@line 254 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  if ($46) {
   $21 = $42;$i$021 = $44;
  } else {
   break;
  }
 }
 return; //@line 289 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
}
function __ZN15b2ContactSolver24SolveVelocityConstraintsEv($this) {
 $this = $this|0;
 var $$off = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0.0, $110 = 0.0, $111 = 0.0, $112 = 0, $113 = 0;
 var $114 = 0.0, $115 = 0.0, $116 = 0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0, $128 = 0.0, $129 = 0.0, $13 = 0.0, $130 = 0, $131 = 0.0;
 var $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0, $139 = 0.0, $14 = 0, $140 = 0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0, $145 = 0.0, $146 = 0.0, $147 = 0, $148 = 0.0, $149 = 0.0, $15 = 0.0;
 var $150 = 0.0, $151 = 0.0, $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0, $156 = 0.0, $157 = 0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0, $167 = 0.0, $168 = 0;
 var $169 = 0.0, $17 = 0.0, $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0, $175 = 0, $176 = 0.0, $177 = 0, $178 = 0.0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0.0, $184 = 0.0, $185 = 0, $186 = 0.0;
 var $187 = 0.0, $188 = 0.0, $189 = 0.0, $19 = 0, $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0.0, $194 = 0.0, $195 = 0.0, $196 = 0, $197 = 0.0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0, $200 = 0.0, $201 = 0.0, $202 = 0.0, $203 = 0.0;
 var $204 = 0, $205 = 0.0, $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0.0, $212 = 0.0, $213 = 0.0, $214 = 0, $215 = 0.0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0.0, $22 = 0, $220 = 0.0, $221 = 0.0;
 var $222 = 0.0, $223 = 0.0, $224 = 0.0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0, $229 = 0.0, $23 = 0, $230 = 0.0, $231 = 0, $232 = 0.0, $233 = 0.0, $234 = 0, $235 = 0.0, $236 = 0.0, $237 = 0, $238 = 0.0, $239 = 0.0, $24 = 0;
 var $240 = 0.0, $241 = 0, $242 = 0.0, $243 = 0.0, $244 = 0, $245 = 0.0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0, $251 = 0.0, $252 = 0.0, $253 = 0, $254 = 0.0, $255 = 0.0, $256 = 0.0, $257 = 0, $258 = 0.0;
 var $259 = 0.0, $26 = 0.0, $260 = 0, $261 = 0.0, $262 = 0.0, $263 = 0.0, $264 = 0.0, $265 = 0.0, $266 = 0, $267 = 0, $268 = 0.0, $269 = 0.0, $27 = 0, $270 = 0.0, $271 = 0.0, $272 = 0.0, $273 = 0.0, $274 = 0.0, $275 = 0.0, $276 = 0.0;
 var $277 = 0.0, $278 = 0.0, $279 = 0, $28 = 0, $280 = 0.0, $281 = 0, $282 = 0.0, $283 = 0.0, $284 = 0.0, $285 = 0.0, $286 = 0.0, $287 = 0.0, $288 = 0.0, $289 = 0.0, $29 = 0, $290 = 0.0, $291 = 0.0, $292 = 0.0, $293 = 0.0, $294 = 0;
 var $295 = 0.0, $296 = 0, $297 = 0.0, $298 = 0.0, $299 = 0.0, $3 = 0, $30 = 0, $300 = 0.0, $301 = 0.0, $302 = 0.0, $303 = 0.0, $304 = 0.0, $305 = 0.0, $306 = 0, $307 = 0.0, $308 = 0.0, $309 = 0.0, $31 = 0, $310 = 0.0, $311 = 0.0;
 var $312 = 0, $313 = 0, $314 = 0.0, $315 = 0.0, $316 = 0.0, $317 = 0.0, $318 = 0.0, $319 = 0.0, $32 = 0.0, $320 = 0.0, $321 = 0.0, $322 = 0.0, $323 = 0.0, $324 = 0.0, $325 = 0, $326 = 0.0, $327 = 0, $328 = 0.0, $329 = 0.0, $33 = 0;
 var $330 = 0.0, $331 = 0.0, $332 = 0.0, $333 = 0.0, $334 = 0.0, $335 = 0.0, $336 = 0.0, $337 = 0.0, $338 = 0.0, $339 = 0.0, $34 = 0.0, $340 = 0, $341 = 0.0, $342 = 0, $343 = 0.0, $344 = 0.0, $345 = 0.0, $346 = 0.0, $347 = 0.0, $348 = 0.0;
 var $349 = 0.0, $35 = 0, $350 = 0.0, $351 = 0.0, $352 = 0, $353 = 0.0, $354 = 0.0, $355 = 0.0, $356 = 0.0, $357 = 0.0, $358 = 0, $359 = 0, $36 = 0.0, $360 = 0.0, $361 = 0.0, $362 = 0.0, $363 = 0.0, $364 = 0.0, $365 = 0.0, $366 = 0.0;
 var $367 = 0.0, $368 = 0.0, $369 = 0.0, $37 = 0.0, $370 = 0.0, $371 = 0, $372 = 0.0, $373 = 0, $374 = 0.0, $375 = 0.0, $376 = 0.0, $377 = 0.0, $378 = 0.0, $379 = 0.0, $38 = 0, $380 = 0.0, $381 = 0.0, $382 = 0.0, $383 = 0.0, $384 = 0.0;
 var $385 = 0.0, $386 = 0, $387 = 0.0, $388 = 0, $389 = 0.0, $39 = 0.0, $390 = 0.0, $391 = 0.0, $392 = 0.0, $393 = 0.0, $394 = 0.0, $395 = 0.0, $396 = 0.0, $397 = 0.0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0.0, $401 = 0.0;
 var $402 = 0.0, $403 = 0.0, $404 = 0.0, $405 = 0.0, $406 = 0.0, $407 = 0.0, $408 = 0.0, $409 = 0.0, $41 = 0, $410 = 0.0, $411 = 0, $412 = 0.0, $413 = 0, $414 = 0.0, $415 = 0.0, $416 = 0.0, $417 = 0.0, $418 = 0.0, $419 = 0.0, $42 = 0;
 var $420 = 0.0, $421 = 0.0, $422 = 0.0, $423 = 0.0, $424 = 0.0, $425 = 0.0, $426 = 0, $427 = 0.0, $428 = 0, $429 = 0.0, $43 = 0, $430 = 0.0, $431 = 0.0, $432 = 0.0, $433 = 0.0, $434 = 0.0, $435 = 0.0, $436 = 0.0, $437 = 0.0, $438 = 0;
 var $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0.0;
 var $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0.0, $63 = 0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0;
 var $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0.0, $78 = 0.0, $79 = 0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0, $86 = 0.0, $87 = 0.0, $88 = 0.0;
 var $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $exitcond = 0, $exitcond128 = 0, $i$0111 = 0, $i1$0101 = 0, $j$089 = 0, $or$cond = 0, $or$cond3 = 0, $or$cond5 = 0;
 var $or$cond84 = 0, $or$cond9 = 0, $vA$sroa$0$0$lcssa = 0, $vA$sroa$0$087 = 0, $vA$sroa$0$199 = 0, $vA$sroa$0$3 = 0, $vA$sroa$20$0$lcssa = 0, $vA$sroa$20$088 = 0, $vA$sroa$20$1100 = 0, $vA$sroa$20$3 = 0, $vB$sroa$0$0$lcssa = 0, $vB$sroa$0$085 = 0, $vB$sroa$0$197 = 0, $vB$sroa$0$3 = 0, $vB$sroa$20$0$lcssa = 0, $vB$sroa$20$086 = 0, $vB$sroa$20$198 = 0, $vB$sroa$20$3 = 0, $wA$0$lcssa = 0.0, $wA$091 = 0.0;
 var $wA$1103 = 0.0, $wA$3 = 0.0, $wB$0$lcssa = 0.0, $wB$090 = 0.0, $wB$1102 = 0.0, $wB$3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 293 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 293 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $2 = ($1|0)>(0); //@line 293 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 if (!($2)) {
  return; //@line 602 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 }
 $3 = ((($this)) + 40|0); //@line 295 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $4 = ((($this)) + 28|0); //@line 305 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $$pre = HEAP32[$4>>2]|0; //@line 305 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $21 = $$pre;$i$0111 = 0;
 while(1) {
  $5 = HEAP32[$3>>2]|0; //@line 295 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $6 = (((($5) + (($i$0111*156)|0)|0)) + 112|0); //@line 297 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 297 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $8 = (((($5) + (($i$0111*156)|0)|0)) + 116|0); //@line 298 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 298 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $10 = (((($5) + (($i$0111*156)|0)|0)) + 120|0); //@line 299 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $11 = +HEAPF32[$10>>2]; //@line 299 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $12 = (((($5) + (($i$0111*156)|0)|0)) + 128|0); //@line 300 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $13 = +HEAPF32[$12>>2]; //@line 300 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $14 = (((($5) + (($i$0111*156)|0)|0)) + 124|0); //@line 301 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $15 = +HEAPF32[$14>>2]; //@line 301 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $16 = (((($5) + (($i$0111*156)|0)|0)) + 132|0); //@line 302 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $17 = +HEAPF32[$16>>2]; //@line 302 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $18 = (((($5) + (($i$0111*156)|0)|0)) + 148|0); //@line 303 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 303 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $20 = (($21) + (($7*12)|0)|0); //@line 305 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $22 = HEAP32[$20>>2]|0; //@line 305 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $23 = (((($21) + (($7*12)|0)|0)) + 4|0); //@line 305 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $24 = HEAP32[$23>>2]|0; //@line 305 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $25 = (((($21) + (($7*12)|0)|0)) + 8|0); //@line 306 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $26 = +HEAPF32[$25>>2]; //@line 306 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $27 = (($21) + (($9*12)|0)|0); //@line 307 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $28 = HEAP32[$27>>2]|0; //@line 307 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $29 = (((($21) + (($9*12)|0)|0)) + 4|0); //@line 307 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $30 = HEAP32[$29>>2]|0; //@line 307 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $31 = (((($21) + (($9*12)|0)|0)) + 8|0); //@line 308 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $32 = +HEAPF32[$31>>2]; //@line 308 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $33 = (((($5) + (($i$0111*156)|0)|0)) + 72|0); //@line 310 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $34 = +HEAPF32[$33>>2]; //@line 310 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $35 = (((($5) + (($i$0111*156)|0)|0)) + 76|0); //@line 310 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $36 = +HEAPF32[$35>>2]; //@line 310 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $37 = -$34; //@line 421 "./libraries/Box2D/Common/b2Math.h"
  $38 = (((($5) + (($i$0111*156)|0)|0)) + 136|0); //@line 312 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $39 = +HEAPF32[$38>>2]; //@line 312 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $40 = ($19|0)==(1); //@line 314 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $$off = (($19) + -1)|0; //@line 314 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $41 = ($$off>>>0)<(2); //@line 314 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  __Z10loveAssertbPKc($41,446854); //@line 314 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $42 = ($19|0)>(0); //@line 318 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  if ($42) {
   $43 = (((($5) + (($i$0111*156)|0)|0)) + 144|0); //@line 326 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
   $j$089 = 0;$vA$sroa$0$087 = $22;$vA$sroa$20$088 = $24;$vB$sroa$0$085 = $28;$vB$sroa$20$086 = $30;$wA$091 = $26;$wB$090 = $32;
   while(1) {
    $46 = ((((($5) + (($i$0111*156)|0)|0) + (($j$089*36)|0)|0)) + 12|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $47 = +HEAPF32[$46>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $48 = $wB$090 * $47; //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $49 = ((((($5) + (($i$0111*156)|0)|0) + (($j$089*36)|0)|0)) + 8|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $50 = +HEAPF32[$49>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $51 = $wB$090 * $50; //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $52 = (HEAP32[tempDoublePtr>>2]=$vB$sroa$0$085,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $53 = $52 - $48; //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $54 = (HEAP32[tempDoublePtr>>2]=$vB$sroa$20$086,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $55 = $54 + $51; //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $56 = (HEAP32[tempDoublePtr>>2]=$vA$sroa$0$087,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $57 = $53 - $56; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $58 = (HEAP32[tempDoublePtr>>2]=$vA$sroa$20$088,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $59 = $55 - $58; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $60 = ((((($5) + (($i$0111*156)|0)|0) + (($j$089*36)|0)|0)) + 4|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $61 = +HEAPF32[$60>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $62 = $wA$091 * $61; //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $63 = ((($5) + (($i$0111*156)|0)|0) + (($j$089*36)|0)|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $64 = +HEAPF32[$63>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $65 = $wA$091 * $64; //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $66 = $57 + $62; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $67 = $59 - $65; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $68 = $36 * $66; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $69 = $67 * $37; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $70 = $68 + $69; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $71 = +HEAPF32[$43>>2]; //@line 326 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $72 = $70 - $71; //@line 326 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $73 = ((((($5) + (($i$0111*156)|0)|0) + (($j$089*36)|0)|0)) + 28|0); //@line 327 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $74 = +HEAPF32[$73>>2]; //@line 327 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $75 = $74 * $72; //@line 327 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $76 = ((((($5) + (($i$0111*156)|0)|0) + (($j$089*36)|0)|0)) + 16|0); //@line 330 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $77 = +HEAPF32[$76>>2]; //@line 330 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $78 = $39 * $77; //@line 330 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $79 = ((((($5) + (($i$0111*156)|0)|0) + (($j$089*36)|0)|0)) + 20|0); //@line 331 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $80 = +HEAPF32[$79>>2]; //@line 331 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $81 = $80 - $75; //@line 331 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $82 = -$78; //@line 331 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $83 = $81 < $78; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $84 = $83 ? $81 : $78; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $85 = $84 < $82; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $86 = $85 ? $82 : $84; //@line 645 "./libraries/Box2D/Common/b2Math.h"
    $87 = $86 - $80; //@line 332 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    HEAPF32[$79>>2] = $86; //@line 333 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $88 = $36 * $87; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $89 = $87 * $37; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $90 = $11 * $88; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $91 = $11 * $89; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $92 = $56 - $90; //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $93 = (HEAPF32[tempDoublePtr>>2]=$92,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $94 = $58 - $91; //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $95 = (HEAPF32[tempDoublePtr>>2]=$94,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $96 = $64 * $89; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $97 = $61 * $88; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $98 = $96 - $97; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $99 = $13 * $98; //@line 339 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $100 = $wA$091 - $99; //@line 339 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $101 = $15 * $88; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $102 = $15 * $89; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $103 = $52 + $101; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $104 = (HEAPF32[tempDoublePtr>>2]=$103,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $105 = $54 + $102; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $106 = (HEAPF32[tempDoublePtr>>2]=$105,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $107 = $50 * $89; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $108 = $47 * $88; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $109 = $107 - $108; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $110 = $17 * $109; //@line 342 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $111 = $wB$090 + $110; //@line 342 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $112 = (($j$089) + 1)|0; //@line 318 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $exitcond = ($112|0)==($19|0); //@line 318 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    if ($exitcond) {
     $vA$sroa$0$0$lcssa = $93;$vA$sroa$20$0$lcssa = $95;$vB$sroa$0$0$lcssa = $104;$vB$sroa$20$0$lcssa = $106;$wA$0$lcssa = $100;$wB$0$lcssa = $111;
     break;
    } else {
     $j$089 = $112;$vA$sroa$0$087 = $93;$vA$sroa$20$088 = $95;$vB$sroa$0$085 = $104;$vB$sroa$20$086 = $106;$wA$091 = $100;$wB$090 = $111;
    }
   }
  } else {
   $vA$sroa$0$0$lcssa = $22;$vA$sroa$20$0$lcssa = $24;$vB$sroa$0$0$lcssa = $28;$vB$sroa$20$0$lcssa = $30;$wA$0$lcssa = $26;$wB$0$lcssa = $32;
  }
  $44 = HEAP8[446813]|0; //@line 346 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $45 = ($44<<24>>24)==(0); //@line 346 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $or$cond9 = $40 | $45; //@line 346 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  do {
   if ($or$cond9) {
    if ($42) {
     $i1$0101 = 0;$vA$sroa$0$199 = $vA$sroa$0$0$lcssa;$vA$sroa$20$1100 = $vA$sroa$20$0$lcssa;$vB$sroa$0$197 = $vB$sroa$0$0$lcssa;$vB$sroa$20$198 = $vB$sroa$20$0$lcssa;$wA$1103 = $wA$0$lcssa;$wB$1102 = $wB$0$lcssa;
     while(1) {
      $113 = ((((($5) + (($i$0111*156)|0)|0) + (($i1$0101*36)|0)|0)) + 12|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
      $114 = +HEAPF32[$113>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
      $115 = $wB$1102 * $114; //@line 428 "./libraries/Box2D/Common/b2Math.h"
      $116 = ((((($5) + (($i$0111*156)|0)|0) + (($i1$0101*36)|0)|0)) + 8|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
      $117 = +HEAPF32[$116>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
      $118 = $wB$1102 * $117; //@line 428 "./libraries/Box2D/Common/b2Math.h"
      $119 = (HEAP32[tempDoublePtr>>2]=$vB$sroa$0$197,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
      $120 = $119 - $115; //@line 448 "./libraries/Box2D/Common/b2Math.h"
      $121 = (HEAP32[tempDoublePtr>>2]=$vB$sroa$20$198,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
      $122 = $121 + $118; //@line 448 "./libraries/Box2D/Common/b2Math.h"
      $123 = (HEAP32[tempDoublePtr>>2]=$vA$sroa$0$199,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $124 = $120 - $123; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $125 = (HEAP32[tempDoublePtr>>2]=$vA$sroa$20$1100,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $126 = $122 - $125; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $127 = ((((($5) + (($i$0111*156)|0)|0) + (($i1$0101*36)|0)|0)) + 4|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
      $128 = +HEAPF32[$127>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
      $129 = $wA$1103 * $128; //@line 428 "./libraries/Box2D/Common/b2Math.h"
      $130 = ((($5) + (($i$0111*156)|0)|0) + (($i1$0101*36)|0)|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
      $131 = +HEAPF32[$130>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
      $132 = $wA$1103 * $131; //@line 428 "./libraries/Box2D/Common/b2Math.h"
      $133 = $124 + $129; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $134 = $126 - $132; //@line 454 "./libraries/Box2D/Common/b2Math.h"
      $135 = $34 * $133; //@line 408 "./libraries/Box2D/Common/b2Math.h"
      $136 = $36 * $134; //@line 408 "./libraries/Box2D/Common/b2Math.h"
      $137 = $135 + $136; //@line 408 "./libraries/Box2D/Common/b2Math.h"
      $138 = ((((($5) + (($i$0111*156)|0)|0) + (($i1$0101*36)|0)|0)) + 24|0); //@line 357 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $139 = +HEAPF32[$138>>2]; //@line 357 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $140 = ((((($5) + (($i$0111*156)|0)|0) + (($i1$0101*36)|0)|0)) + 32|0); //@line 357 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $141 = +HEAPF32[$140>>2]; //@line 357 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $142 = $137 - $141; //@line 357 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $143 = $139 * $142; //@line 357 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $144 = ((((($5) + (($i$0111*156)|0)|0) + (($i1$0101*36)|0)|0)) + 16|0); //@line 360 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $145 = +HEAPF32[$144>>2]; //@line 360 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $146 = $145 - $143; //@line 360 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $147 = $146 > 0.0; //@line 645 "./libraries/Box2D/Common/b2Math.h"
      $148 = $147 ? $146 : 0.0; //@line 645 "./libraries/Box2D/Common/b2Math.h"
      $149 = $148 - $145; //@line 361 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      HEAPF32[$144>>2] = $148; //@line 362 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $150 = $34 * $149; //@line 459 "./libraries/Box2D/Common/b2Math.h"
      $151 = $36 * $149; //@line 459 "./libraries/Box2D/Common/b2Math.h"
      $152 = $11 * $150; //@line 459 "./libraries/Box2D/Common/b2Math.h"
      $153 = $11 * $151; //@line 459 "./libraries/Box2D/Common/b2Math.h"
      $154 = $123 - $152; //@line 91 "./libraries/Box2D/Common/b2Math.h"
      $155 = (HEAPF32[tempDoublePtr>>2]=$154,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
      $156 = $125 - $153; //@line 91 "./libraries/Box2D/Common/b2Math.h"
      $157 = (HEAPF32[tempDoublePtr>>2]=$156,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
      $158 = $131 * $151; //@line 414 "./libraries/Box2D/Common/b2Math.h"
      $159 = $128 * $150; //@line 414 "./libraries/Box2D/Common/b2Math.h"
      $160 = $158 - $159; //@line 414 "./libraries/Box2D/Common/b2Math.h"
      $161 = $13 * $160; //@line 367 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $162 = $wA$1103 - $161; //@line 367 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $163 = $15 * $150; //@line 459 "./libraries/Box2D/Common/b2Math.h"
      $164 = $15 * $151; //@line 459 "./libraries/Box2D/Common/b2Math.h"
      $165 = $119 + $163; //@line 85 "./libraries/Box2D/Common/b2Math.h"
      $166 = (HEAPF32[tempDoublePtr>>2]=$165,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
      $167 = $121 + $164; //@line 85 "./libraries/Box2D/Common/b2Math.h"
      $168 = (HEAPF32[tempDoublePtr>>2]=$167,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
      $169 = $117 * $151; //@line 414 "./libraries/Box2D/Common/b2Math.h"
      $170 = $114 * $150; //@line 414 "./libraries/Box2D/Common/b2Math.h"
      $171 = $169 - $170; //@line 414 "./libraries/Box2D/Common/b2Math.h"
      $172 = $17 * $171; //@line 370 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $173 = $wB$1102 + $172; //@line 370 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $174 = (($i1$0101) + 1)|0; //@line 348 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      $exitcond128 = ($174|0)==($19|0); //@line 348 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
      if ($exitcond128) {
       $vA$sroa$0$3 = $155;$vA$sroa$20$3 = $157;$vB$sroa$0$3 = $166;$vB$sroa$20$3 = $168;$wA$3 = $162;$wB$3 = $173;
       break;
      } else {
       $i1$0101 = $174;$vA$sroa$0$199 = $155;$vA$sroa$20$1100 = $157;$vB$sroa$0$197 = $166;$vB$sroa$20$198 = $168;$wA$1103 = $162;$wB$1102 = $173;
      }
     }
    } else {
     $vA$sroa$0$3 = $vA$sroa$0$0$lcssa;$vA$sroa$20$3 = $vA$sroa$20$0$lcssa;$vB$sroa$0$3 = $vB$sroa$0$0$lcssa;$vB$sroa$20$3 = $vB$sroa$20$0$lcssa;$wA$3 = $wA$0$lcssa;$wB$3 = $wB$0$lcssa;
    }
   } else {
    $175 = (((($5) + (($i$0111*156)|0)|0)) + 16|0); //@line 411 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $176 = +HEAPF32[$175>>2]; //@line 411 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $177 = (((($5) + (($i$0111*156)|0)|0)) + 52|0); //@line 411 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $178 = +HEAPF32[$177>>2]; //@line 411 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $179 = $176 >= 0.0; //@line 412 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $180 = $178 >= 0.0; //@line 412 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $181 = $179 & $180; //@line 412 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    __Z10loveAssertbPKc($181,446889); //@line 412 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $182 = (((($5) + (($i$0111*156)|0)|0)) + 12|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $183 = +HEAPF32[$182>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $184 = $wB$0$lcssa * $183; //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $185 = (((($5) + (($i$0111*156)|0)|0)) + 8|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $186 = +HEAPF32[$185>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $187 = $wB$0$lcssa * $186; //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $188 = (HEAP32[tempDoublePtr>>2]=$vB$sroa$0$0$lcssa,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $189 = $188 - $184; //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $190 = (HEAP32[tempDoublePtr>>2]=$vB$sroa$20$0$lcssa,+HEAPF32[tempDoublePtr>>2]); //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $191 = $190 + $187; //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $192 = (HEAP32[tempDoublePtr>>2]=$vA$sroa$0$0$lcssa,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $193 = $189 - $192; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $194 = (HEAP32[tempDoublePtr>>2]=$vA$sroa$20$0$lcssa,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $195 = $191 - $194; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $196 = (((($5) + (($i$0111*156)|0)|0)) + 4|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $197 = +HEAPF32[$196>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $198 = $wA$0$lcssa * $197; //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $199 = (($5) + (($i$0111*156)|0)|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $200 = +HEAPF32[$199>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $201 = $wA$0$lcssa * $200; //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $202 = $193 + $198; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $203 = $195 - $201; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $204 = (((($5) + (($i$0111*156)|0)|0)) + 48|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $205 = +HEAPF32[$204>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $206 = $wB$0$lcssa * $205; //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $207 = (((($5) + (($i$0111*156)|0)|0)) + 44|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $208 = +HEAPF32[$207>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $209 = $wB$0$lcssa * $208; //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $210 = $188 - $206; //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $211 = $190 + $209; //@line 448 "./libraries/Box2D/Common/b2Math.h"
    $212 = $210 - $192; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $213 = $211 - $194; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $214 = (((($5) + (($i$0111*156)|0)|0)) + 40|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $215 = +HEAPF32[$214>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $216 = $wA$0$lcssa * $215; //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $217 = (((($5) + (($i$0111*156)|0)|0)) + 36|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $218 = +HEAPF32[$217>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $219 = $wA$0$lcssa * $218; //@line 428 "./libraries/Box2D/Common/b2Math.h"
    $220 = $212 + $216; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $221 = $213 - $219; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $222 = $34 * $202; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $223 = $36 * $203; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $224 = $222 + $223; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $225 = $34 * $220; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $226 = $36 * $221; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $227 = $225 + $226; //@line 408 "./libraries/Box2D/Common/b2Math.h"
    $228 = (((($5) + (($i$0111*156)|0)|0)) + 32|0); //@line 423 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $229 = +HEAPF32[$228>>2]; //@line 423 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $230 = $224 - $229; //@line 423 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $231 = (((($5) + (($i$0111*156)|0)|0)) + 68|0); //@line 424 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $232 = +HEAPF32[$231>>2]; //@line 424 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $233 = $227 - $232; //@line 424 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $234 = (((($5) + (($i$0111*156)|0)|0)) + 96|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $235 = +HEAPF32[$234>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $236 = $176 * $235; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $237 = (((($5) + (($i$0111*156)|0)|0)) + 104|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $238 = +HEAPF32[$237>>2]; //@line 528 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $239 = $178 * $238; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $240 = $236 + $239; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $241 = (((($5) + (($i$0111*156)|0)|0)) + 100|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $242 = +HEAPF32[$241>>2]; //@line 487 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $243 = $176 * $242; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $244 = (((($5) + (($i$0111*156)|0)|0)) + 108|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $245 = +HEAPF32[$244>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $246 = $178 * $245; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $247 = $243 + $246; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $248 = $230 - $240; //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $249 = $233 - $247; //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $250 = (((($5) + (($i$0111*156)|0)|0)) + 80|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $251 = +HEAPF32[$250>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $252 = $251 * $248; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $253 = (((($5) + (($i$0111*156)|0)|0)) + 88|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $254 = +HEAPF32[$253>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $255 = $254 * $249; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $256 = $252 + $255; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $257 = (((($5) + (($i$0111*156)|0)|0)) + 84|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $258 = +HEAPF32[$257>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $259 = $248 * $258; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $260 = (((($5) + (($i$0111*156)|0)|0)) + 92|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $261 = +HEAPF32[$260>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $262 = $249 * $261; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $263 = $259 + $262; //@line 435 "./libraries/Box2D/Common/b2Math.h"
    $264 = -$256; //@line 68 "./libraries/Box2D/Common/b2Math.h"
    $265 = -$263; //@line 68 "./libraries/Box2D/Common/b2Math.h"
    $266 = !($256 <= -0.0); //@line 445 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $267 = !($263 <= -0.0); //@line 445 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $or$cond84 = $266 | $267; //@line 445 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    if (!($or$cond84)) {
     $268 = $264 - $176; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $269 = $265 - $178; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $270 = $34 * $268; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $271 = $36 * $268; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $272 = $34 * $269; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $273 = $36 * $269; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $274 = $270 + $272; //@line 448 "./libraries/Box2D/Common/b2Math.h"
     $275 = $271 + $273; //@line 448 "./libraries/Box2D/Common/b2Math.h"
     $276 = $11 * $274; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $277 = $11 * $275; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $278 = $192 - $276; //@line 91 "./libraries/Box2D/Common/b2Math.h"
     $279 = (HEAPF32[tempDoublePtr>>2]=$278,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
     $280 = $194 - $277; //@line 91 "./libraries/Box2D/Common/b2Math.h"
     $281 = (HEAPF32[tempDoublePtr>>2]=$280,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
     $282 = $200 * $271; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $283 = $197 * $270; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $284 = $282 - $283; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $285 = $218 * $273; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $286 = $215 * $272; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $287 = $285 - $286; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $288 = $284 + $287; //@line 454 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $289 = $13 * $288; //@line 454 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $290 = $wA$0$lcssa - $289; //@line 454 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $291 = $15 * $274; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $292 = $15 * $275; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $293 = $188 + $291; //@line 85 "./libraries/Box2D/Common/b2Math.h"
     $294 = (HEAPF32[tempDoublePtr>>2]=$293,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
     $295 = $190 + $292; //@line 85 "./libraries/Box2D/Common/b2Math.h"
     $296 = (HEAPF32[tempDoublePtr>>2]=$295,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
     $297 = $186 * $271; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $298 = $183 * $270; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $299 = $297 - $298; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $300 = $208 * $273; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $301 = $205 * $272; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $302 = $300 - $301; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $303 = $299 + $302; //@line 457 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $304 = $17 * $303; //@line 457 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $305 = $wB$0$lcssa + $304; //@line 457 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     HEAPF32[$175>>2] = $264; //@line 460 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     HEAPF32[$177>>2] = $265; //@line 461 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $vA$sroa$0$3 = $279;$vA$sroa$20$3 = $281;$vB$sroa$0$3 = $294;$vB$sroa$20$3 = $296;$wA$3 = $290;$wB$3 = $305;
     break;
    }
    $306 = (((($5) + (($i$0111*156)|0)|0)) + 24|0); //@line 484 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $307 = +HEAPF32[$306>>2]; //@line 484 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $308 = $248 * $307; //@line 484 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $309 = -$308; //@line 484 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $310 = $242 * $309; //@line 487 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $311 = $249 + $310; //@line 487 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $312 = $308 <= -0.0; //@line 489 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $313 = $311 >= 0.0; //@line 489 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $or$cond = $312 & $313; //@line 489 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    if ($or$cond) {
     $314 = $309 - $176; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $315 = 0.0 - $178; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $316 = $34 * $314; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $317 = $36 * $314; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $318 = $34 * $315; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $319 = $36 * $315; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $320 = $318 + $316; //@line 448 "./libraries/Box2D/Common/b2Math.h"
     $321 = $319 + $317; //@line 448 "./libraries/Box2D/Common/b2Math.h"
     $322 = $11 * $320; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $323 = $11 * $321; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $324 = $192 - $322; //@line 91 "./libraries/Box2D/Common/b2Math.h"
     $325 = (HEAPF32[tempDoublePtr>>2]=$324,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
     $326 = $194 - $323; //@line 91 "./libraries/Box2D/Common/b2Math.h"
     $327 = (HEAPF32[tempDoublePtr>>2]=$326,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
     $328 = $317 * $200; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $329 = $316 * $197; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $330 = $328 - $329; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $331 = $319 * $218; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $332 = $318 * $215; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $333 = $331 - $332; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $334 = $330 + $333; //@line 498 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $335 = $13 * $334; //@line 498 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $336 = $wA$0$lcssa - $335; //@line 498 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $337 = $15 * $320; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $338 = $15 * $321; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $339 = $188 + $337; //@line 85 "./libraries/Box2D/Common/b2Math.h"
     $340 = (HEAPF32[tempDoublePtr>>2]=$339,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
     $341 = $190 + $338; //@line 85 "./libraries/Box2D/Common/b2Math.h"
     $342 = (HEAPF32[tempDoublePtr>>2]=$341,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
     $343 = $317 * $186; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $344 = $316 * $183; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $345 = $343 - $344; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $346 = $319 * $208; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $347 = $318 * $205; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $348 = $346 - $347; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $349 = $345 + $348; //@line 501 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $350 = $17 * $349; //@line 501 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $351 = $wB$0$lcssa + $350; //@line 501 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     HEAPF32[$175>>2] = $309; //@line 504 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     HEAP32[$177>>2] = 0; //@line 505 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $vA$sroa$0$3 = $325;$vA$sroa$20$3 = $327;$vB$sroa$0$3 = $340;$vB$sroa$20$3 = $342;$wA$3 = $336;$wB$3 = $351;
     break;
    }
    $352 = (((($5) + (($i$0111*156)|0)|0)) + 60|0); //@line 527 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $353 = +HEAPF32[$352>>2]; //@line 527 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $354 = $249 * $353; //@line 527 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $355 = -$354; //@line 527 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $356 = $238 * $355; //@line 528 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $357 = $248 + $356; //@line 528 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $358 = $354 <= -0.0; //@line 531 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $359 = $357 >= 0.0; //@line 531 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $or$cond3 = $358 & $359; //@line 531 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    if ($or$cond3) {
     $360 = 0.0 - $176; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $361 = $355 - $178; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $362 = $34 * $360; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $363 = $36 * $360; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $364 = $34 * $361; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $365 = $36 * $361; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $366 = $362 + $364; //@line 448 "./libraries/Box2D/Common/b2Math.h"
     $367 = $363 + $365; //@line 448 "./libraries/Box2D/Common/b2Math.h"
     $368 = $11 * $366; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $369 = $11 * $367; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $370 = $192 - $368; //@line 91 "./libraries/Box2D/Common/b2Math.h"
     $371 = (HEAPF32[tempDoublePtr>>2]=$370,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
     $372 = $194 - $369; //@line 91 "./libraries/Box2D/Common/b2Math.h"
     $373 = (HEAPF32[tempDoublePtr>>2]=$372,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
     $374 = $363 * $200; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $375 = $362 * $197; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $376 = $374 - $375; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $377 = $365 * $218; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $378 = $364 * $215; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $379 = $377 - $378; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $380 = $376 + $379; //@line 540 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $381 = $13 * $380; //@line 540 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $382 = $wA$0$lcssa - $381; //@line 540 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $383 = $15 * $366; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $384 = $15 * $367; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $385 = $188 + $383; //@line 85 "./libraries/Box2D/Common/b2Math.h"
     $386 = (HEAPF32[tempDoublePtr>>2]=$385,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
     $387 = $190 + $384; //@line 85 "./libraries/Box2D/Common/b2Math.h"
     $388 = (HEAPF32[tempDoublePtr>>2]=$387,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
     $389 = $363 * $186; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $390 = $362 * $183; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $391 = $389 - $390; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $392 = $365 * $208; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $393 = $364 * $205; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $394 = $392 - $393; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $395 = $391 + $394; //@line 543 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $396 = $17 * $395; //@line 543 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $397 = $wB$0$lcssa + $396; //@line 543 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     HEAP32[$175>>2] = 0; //@line 546 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     HEAPF32[$177>>2] = $355; //@line 547 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $vA$sroa$0$3 = $371;$vA$sroa$20$3 = $373;$vB$sroa$0$3 = $386;$vB$sroa$20$3 = $388;$wA$3 = $382;$wB$3 = $397;
     break;
    }
    $398 = $248 >= 0.0; //@line 571 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $399 = $249 >= 0.0; //@line 571 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $or$cond5 = $398 & $399; //@line 571 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    if ($or$cond5) {
     $400 = 0.0 - $176; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $401 = 0.0 - $178; //@line 454 "./libraries/Box2D/Common/b2Math.h"
     $402 = $34 * $400; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $403 = $36 * $400; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $404 = $34 * $401; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $405 = $36 * $401; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $406 = $402 + $404; //@line 448 "./libraries/Box2D/Common/b2Math.h"
     $407 = $403 + $405; //@line 448 "./libraries/Box2D/Common/b2Math.h"
     $408 = $11 * $406; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $409 = $11 * $407; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $410 = $192 - $408; //@line 91 "./libraries/Box2D/Common/b2Math.h"
     $411 = (HEAPF32[tempDoublePtr>>2]=$410,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
     $412 = $194 - $409; //@line 91 "./libraries/Box2D/Common/b2Math.h"
     $413 = (HEAPF32[tempDoublePtr>>2]=$412,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
     $414 = $403 * $200; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $415 = $402 * $197; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $416 = $414 - $415; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $417 = $405 * $218; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $418 = $404 * $215; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $419 = $417 - $418; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $420 = $416 + $419; //@line 580 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $421 = $13 * $420; //@line 580 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $422 = $wA$0$lcssa - $421; //@line 580 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $423 = $15 * $406; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $424 = $15 * $407; //@line 459 "./libraries/Box2D/Common/b2Math.h"
     $425 = $188 + $423; //@line 85 "./libraries/Box2D/Common/b2Math.h"
     $426 = (HEAPF32[tempDoublePtr>>2]=$425,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
     $427 = $190 + $424; //@line 85 "./libraries/Box2D/Common/b2Math.h"
     $428 = (HEAPF32[tempDoublePtr>>2]=$427,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
     $429 = $403 * $186; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $430 = $402 * $183; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $431 = $429 - $430; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $432 = $405 * $208; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $433 = $404 * $205; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $434 = $432 - $433; //@line 414 "./libraries/Box2D/Common/b2Math.h"
     $435 = $431 + $434; //@line 583 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $436 = $17 * $435; //@line 583 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $437 = $wB$0$lcssa + $436; //@line 583 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     HEAP32[$175>>2] = 0; //@line 586 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     HEAP32[$177>>2] = 0; //@line 587 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $vA$sroa$0$3 = $411;$vA$sroa$20$3 = $413;$vB$sroa$0$3 = $426;$vB$sroa$20$3 = $428;$wA$3 = $422;$wB$3 = $437;
    } else {
     $vA$sroa$0$3 = $vA$sroa$0$0$lcssa;$vA$sroa$20$3 = $vA$sroa$20$0$lcssa;$vB$sroa$0$3 = $vB$sroa$0$0$lcssa;$vB$sroa$20$3 = $vB$sroa$20$0$lcssa;$wA$3 = $wA$0$lcssa;$wB$3 = $wB$0$lcssa;
    }
   }
  } while(0);
  $438 = HEAP32[$4>>2]|0; //@line 597 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $439 = (($438) + (($7*12)|0)|0); //@line 597 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$439>>2] = $vA$sroa$0$3; //@line 597 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $440 = (((($438) + (($7*12)|0)|0)) + 4|0); //@line 597 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$440>>2] = $vA$sroa$20$3; //@line 597 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $441 = HEAP32[$4>>2]|0; //@line 598 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $442 = (((($441) + (($7*12)|0)|0)) + 8|0); //@line 598 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$442>>2] = $wA$3; //@line 598 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $443 = (($441) + (($9*12)|0)|0); //@line 599 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$443>>2] = $vB$sroa$0$3; //@line 599 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $444 = (((($441) + (($9*12)|0)|0)) + 4|0); //@line 599 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$444>>2] = $vB$sroa$20$3; //@line 599 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $445 = HEAP32[$4>>2]|0; //@line 600 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $446 = (((($445) + (($9*12)|0)|0)) + 8|0); //@line 600 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$446>>2] = $wB$3; //@line 600 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $447 = (($i$0111) + 1)|0; //@line 293 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $448 = HEAP32[$0>>2]|0; //@line 293 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $449 = ($447|0)<($448|0); //@line 293 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  if ($449) {
   $21 = $445;$i$0111 = $447;
  } else {
   break;
  }
 }
 return; //@line 602 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
}
function __ZN15b2ContactSolver13StoreImpulsesEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$02 = 0, $j$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 606 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 606 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $2 = ($1|0)>(0); //@line 606 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 if (!($2)) {
  return; //@line 617 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 }
 $3 = ((($this)) + 40|0); //@line 608 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 608 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $5 = ((($this)) + 44|0); //@line 609 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 609 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $i$02 = 0;
 while(1) {
  $7 = (((($4) + (($i$02*156)|0)|0)) + 152|0); //@line 609 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $8 = HEAP32[$7>>2]|0; //@line 609 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $9 = (($6) + ($8<<2)|0); //@line 609 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $10 = HEAP32[$9>>2]|0; //@line 609 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $11 = (((($4) + (($i$02*156)|0)|0)) + 148|0); //@line 611 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 611 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $13 = ($12|0)>(0); //@line 611 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  if ($13) {
   $j$01 = 0;
   while(1) {
    $16 = ((((($4) + (($i$02*156)|0)|0) + (($j$01*36)|0)|0)) + 16|0); //@line 613 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $17 = HEAP32[$16>>2]|0; //@line 613 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $18 = (((((($10)) + 64|0) + (($j$01*20)|0)|0)) + 8|0); //@line 613 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    HEAP32[$18>>2] = $17; //@line 613 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $19 = ((((($4) + (($i$02*156)|0)|0) + (($j$01*36)|0)|0)) + 20|0); //@line 614 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $20 = HEAP32[$19>>2]|0; //@line 614 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $21 = (((((($10)) + 64|0) + (($j$01*20)|0)|0)) + 12|0); //@line 614 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    HEAP32[$21>>2] = $20; //@line 614 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $22 = (($j$01) + 1)|0; //@line 611 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $23 = ($22|0)<($12|0); //@line 611 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    if ($23) {
     $j$01 = $22;
    } else {
     break;
    }
   }
  }
  $14 = (($i$02) + 1)|0; //@line 606 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $15 = ($14|0)<($1|0); //@line 606 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  if ($15) {
   $i$02 = $14;
  } else {
   break;
  }
 }
 return; //@line 617 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
}
function __ZN15b2ContactSolver24SolvePositionConstraintsEv($this) {
 $this = $this|0;
 var $$lcssa = 0.0, $$lcssa63 = 0, $$lcssa64 = 0, $$lcssa65 = 0.0, $$lcssa66 = 0, $$lcssa67 = 0, $$lcssa68 = 0.0, $$pre = 0, $$pre44 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0.0;
 var $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0, $119 = 0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0;
 var $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0, $13 = 0, $130 = 0.0, $131 = 0, $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0, $141 = 0.0, $142 = 0, $143 = 0.0;
 var $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0.0, $148 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0, $27 = 0.0, $28 = 0;
 var $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0.0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0;
 var $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $aA$0$lcssa = 0.0, $aA$024 = 0.0;
 var $aB$0$lcssa = 0.0, $aB$023 = 0.0, $cA$sroa$0$0$lcssa = 0, $cA$sroa$0$020 = 0, $cA$sroa$8$0$lcssa = 0, $cA$sroa$8$021 = 0, $cB$sroa$0$0$lcssa = 0, $cB$sroa$0$018 = 0, $cB$sroa$8$0$lcssa = 0, $cB$sroa$8$019 = 0, $exitcond = 0, $i$032 = 0, $j$022 = 0, $minSeparation$0$lcssa = 0.0, $minSeparation$033 = 0.0, $minSeparation$1$lcssa = 0.0, $minSeparation$125 = 0.0, $psm = 0, $xfA = 0, $xfB = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $xfA = sp + 40|0;
 $xfB = sp + 24|0;
 $psm = sp;
 $0 = ((($this)) + 48|0); //@line 675 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 675 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $2 = ($1|0)>(0); //@line 675 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 if (!($2)) {
  $minSeparation$0$lcssa = 0.0;
  $15 = $minSeparation$0$lcssa >= -0.014999999664723873; //@line 746 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  STACKTOP = sp;return ($15|0); //@line 746 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 }
 $3 = ((($this)) + 36|0); //@line 677 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $4 = ((($this)) + 24|0); //@line 689 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $5 = ((($xfA)) + 8|0); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($xfA)) + 12|0); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($xfB)) + 8|0); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($xfB)) + 12|0); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($xfA)) + 4|0); //@line 701 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $10 = ((($xfB)) + 4|0); //@line 702 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $11 = ((($psm)) + 4|0); //@line 706 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $12 = ((($psm)) + 8|0); //@line 708 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $13 = ((($psm)) + 12|0); //@line 708 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $14 = ((($psm)) + 16|0); //@line 709 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $$pre = HEAP32[$4>>2]|0; //@line 689 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $41 = $$pre;$i$032 = 0;$minSeparation$033 = 0.0;
 while(1) {
  $16 = HEAP32[$3>>2]|0; //@line 677 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $17 = (($16) + (($i$032*88)|0)|0); //@line 677 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $18 = (((($16) + (($i$032*88)|0)|0)) + 32|0); //@line 679 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 679 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $20 = (((($16) + (($i$032*88)|0)|0)) + 36|0); //@line 680 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $21 = HEAP32[$20>>2]|0; //@line 680 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $22 = (((($16) + (($i$032*88)|0)|0)) + 48|0); //@line 681 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $23 = +HEAPF32[$22>>2]; //@line 681 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $24 = (((($16) + (($i$032*88)|0)|0)) + 52|0); //@line 681 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $25 = +HEAPF32[$24>>2]; //@line 681 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $26 = (((($16) + (($i$032*88)|0)|0)) + 40|0); //@line 682 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $27 = +HEAPF32[$26>>2]; //@line 682 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $28 = (((($16) + (($i$032*88)|0)|0)) + 64|0); //@line 683 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $29 = +HEAPF32[$28>>2]; //@line 683 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $30 = (((($16) + (($i$032*88)|0)|0)) + 56|0); //@line 684 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $31 = +HEAPF32[$30>>2]; //@line 684 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $32 = (((($16) + (($i$032*88)|0)|0)) + 60|0); //@line 684 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $33 = +HEAPF32[$32>>2]; //@line 684 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $34 = (((($16) + (($i$032*88)|0)|0)) + 44|0); //@line 685 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $35 = +HEAPF32[$34>>2]; //@line 685 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $36 = (((($16) + (($i$032*88)|0)|0)) + 68|0); //@line 686 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $37 = +HEAPF32[$36>>2]; //@line 686 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $38 = (((($16) + (($i$032*88)|0)|0)) + 84|0); //@line 687 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $39 = HEAP32[$38>>2]|0; //@line 687 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $40 = (($41) + (($19*12)|0)|0); //@line 689 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $42 = HEAP32[$40>>2]|0; //@line 689 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $43 = (((($41) + (($19*12)|0)|0)) + 4|0); //@line 689 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $44 = HEAP32[$43>>2]|0; //@line 689 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $45 = (((($41) + (($19*12)|0)|0)) + 8|0); //@line 690 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $46 = +HEAPF32[$45>>2]; //@line 690 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $47 = (($41) + (($21*12)|0)|0); //@line 692 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $48 = HEAP32[$47>>2]|0; //@line 692 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $49 = (((($41) + (($21*12)|0)|0)) + 4|0); //@line 692 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $50 = HEAP32[$49>>2]|0; //@line 692 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $51 = (((($41) + (($21*12)|0)|0)) + 8|0); //@line 693 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $52 = +HEAPF32[$51>>2]; //@line 693 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $53 = ($39|0)>(0); //@line 696 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  if ($53) {
   $54 = $27 + $35; //@line 723 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
   $aA$024 = $46;$aB$023 = $52;$cA$sroa$0$020 = $42;$cA$sroa$8$021 = $44;$cB$sroa$0$018 = $48;$cB$sroa$8$019 = $50;$j$022 = 0;$minSeparation$125 = $minSeparation$033;
   while(1) {
    $67 = (+Math_sin((+$aA$024))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$5>>2] = $67; //@line 315 "./libraries/Box2D/Common/b2Math.h"
    $68 = (+Math_cos((+$aA$024))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$6>>2] = $68; //@line 316 "./libraries/Box2D/Common/b2Math.h"
    $69 = (+Math_sin((+$aB$023))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$7>>2] = $69; //@line 315 "./libraries/Box2D/Common/b2Math.h"
    $70 = (+Math_cos((+$aB$023))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$8>>2] = $70; //@line 316 "./libraries/Box2D/Common/b2Math.h"
    $71 = $23 * $68; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $72 = $25 * $67; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $73 = $71 - $72; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $74 = $23 * $67; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $75 = $25 * $68; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $76 = $75 + $74; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $77 = (HEAP32[tempDoublePtr>>2]=$cA$sroa$0$020,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $78 = $77 - $73; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $79 = (HEAP32[tempDoublePtr>>2]=$cA$sroa$8$021,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $80 = $79 - $76; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$xfA>>2] = $78; //@line 701 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    HEAPF32[$9>>2] = $80; //@line 701 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $81 = $31 * $70; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $82 = $33 * $69; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $83 = $81 - $82; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $84 = $31 * $69; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $85 = $33 * $70; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $86 = $85 + $84; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $87 = (HEAP32[tempDoublePtr>>2]=$cB$sroa$0$018,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $88 = $87 - $83; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $89 = (HEAP32[tempDoublePtr>>2]=$cB$sroa$8$019,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $90 = $89 - $86; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$xfB>>2] = $88; //@line 702 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    HEAPF32[$10>>2] = $90; //@line 702 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    __ZN24b2PositionSolverManifold10InitializeEP27b2ContactPositionConstraintRK11b2TransformS4_i($psm,$17,$xfA,$xfB,$j$022); //@line 705 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $91 = +HEAPF32[$psm>>2]; //@line 706 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $92 = +HEAPF32[$11>>2]; //@line 706 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $93 = +HEAPF32[$12>>2]; //@line 708 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $94 = +HEAPF32[$13>>2]; //@line 708 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $95 = +HEAPF32[$14>>2]; //@line 709 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $96 = $93 - $77; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $97 = $94 - $79; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $98 = $93 - $87; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $99 = $94 - $89; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $100 = $minSeparation$125 < $95; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $101 = $100 ? $minSeparation$125 : $95; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $102 = $95 + 0.004999999888241291; //@line 718 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $103 = $102 * 0.20000000298023224; //@line 718 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $104 = $103 < 0.0; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $105 = $104 ? $103 : 0.0; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $106 = $92 * $96; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $107 = $91 * $97; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $108 = $106 - $107; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $109 = $92 * $98; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $110 = $91 * $99; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $111 = $109 - $110; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $112 = $29 * $108; //@line 723 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $113 = $108 * $112; //@line 723 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $114 = $54 + $113; //@line 723 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $115 = $37 * $111; //@line 723 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $116 = $111 * $115; //@line 723 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $117 = $116 + $114; //@line 723 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $118 = $117 > 0.0; //@line 726 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    if ($118) {
     $119 = $105 < -0.20000000298023224; //@line 645 "./libraries/Box2D/Common/b2Math.h"
     $120 = $119 ? -0.20000000298023224 : $105; //@line 645 "./libraries/Box2D/Common/b2Math.h"
     $121 = -$120; //@line 726 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $122 = $121 / $117; //@line 726 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $124 = $122;
    } else {
     $124 = 0.0;
    }
    $123 = $91 * $124; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $125 = $92 * $124; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $126 = $27 * $123; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $127 = $27 * $125; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $128 = $77 - $126; //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $129 = (HEAPF32[tempDoublePtr>>2]=$128,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $130 = $79 - $127; //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $131 = (HEAPF32[tempDoublePtr>>2]=$130,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $132 = $96 * $125; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $133 = $97 * $123; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $134 = $132 - $133; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $135 = $29 * $134; //@line 731 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $136 = $aA$024 - $135; //@line 731 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $137 = $35 * $123; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $138 = $35 * $125; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $139 = $87 + $137; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $140 = (HEAPF32[tempDoublePtr>>2]=$139,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $141 = $89 + $138; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $142 = (HEAPF32[tempDoublePtr>>2]=$141,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $143 = $98 * $125; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $144 = $99 * $123; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $145 = $143 - $144; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $146 = $37 * $145; //@line 734 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $147 = $aB$023 + $146; //@line 734 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $148 = (($j$022) + 1)|0; //@line 696 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $exitcond = ($148|0)==($39|0); //@line 696 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    if ($exitcond) {
     $$lcssa = $101;$$lcssa63 = $129;$$lcssa64 = $131;$$lcssa65 = $136;$$lcssa66 = $140;$$lcssa67 = $142;$$lcssa68 = $147;
     break;
    } else {
     $aA$024 = $136;$aB$023 = $147;$cA$sroa$0$020 = $129;$cA$sroa$8$021 = $131;$cB$sroa$0$018 = $140;$cB$sroa$8$019 = $142;$j$022 = $148;$minSeparation$125 = $101;
    }
   }
   $$pre44 = HEAP32[$4>>2]|0; //@line 737 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
   $56 = $$pre44;$aA$0$lcssa = $$lcssa65;$aB$0$lcssa = $$lcssa68;$cA$sroa$0$0$lcssa = $$lcssa63;$cA$sroa$8$0$lcssa = $$lcssa64;$cB$sroa$0$0$lcssa = $$lcssa66;$cB$sroa$8$0$lcssa = $$lcssa67;$minSeparation$1$lcssa = $$lcssa;
  } else {
   $56 = $41;$aA$0$lcssa = $46;$aB$0$lcssa = $52;$cA$sroa$0$0$lcssa = $42;$cA$sroa$8$0$lcssa = $44;$cB$sroa$0$0$lcssa = $48;$cB$sroa$8$0$lcssa = $50;$minSeparation$1$lcssa = $minSeparation$033;
  }
  $55 = (($56) + (($19*12)|0)|0); //@line 737 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$55>>2] = $cA$sroa$0$0$lcssa; //@line 737 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $57 = (((($56) + (($19*12)|0)|0)) + 4|0); //@line 737 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$57>>2] = $cA$sroa$8$0$lcssa; //@line 737 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $58 = HEAP32[$4>>2]|0; //@line 738 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $59 = (((($58) + (($19*12)|0)|0)) + 8|0); //@line 738 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$59>>2] = $aA$0$lcssa; //@line 738 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $60 = (($58) + (($21*12)|0)|0); //@line 740 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$60>>2] = $cB$sroa$0$0$lcssa; //@line 740 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $61 = (((($58) + (($21*12)|0)|0)) + 4|0); //@line 740 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$61>>2] = $cB$sroa$8$0$lcssa; //@line 740 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $62 = HEAP32[$4>>2]|0; //@line 741 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $63 = (((($62) + (($21*12)|0)|0)) + 8|0); //@line 741 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$63>>2] = $aB$0$lcssa; //@line 741 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $64 = (($i$032) + 1)|0; //@line 675 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $65 = HEAP32[$0>>2]|0; //@line 675 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $66 = ($64|0)<($65|0); //@line 675 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  if ($66) {
   $41 = $62;$i$032 = $64;$minSeparation$033 = $minSeparation$1$lcssa;
  } else {
   $minSeparation$0$lcssa = $minSeparation$1$lcssa;
   break;
  }
 }
 $15 = $minSeparation$0$lcssa >= -0.014999999664723873; //@line 746 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 STACKTOP = sp;return ($15|0); //@line 746 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
}
function __ZN24b2PositionSolverManifold10InitializeEP27b2ContactPositionConstraintRK11b2TransformS4_i($this,$pc,$xfA,$xfB,$index) {
 $this = $this|0;
 $pc = $pc|0;
 $xfA = $xfA|0;
 $xfB = $xfB|0;
 $index = $index|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0.0, $103 = 0, $104 = 0.0, $105 = 0, $106 = 0.0, $107 = 0.0, $108 = 0, $109 = 0.0, $11 = 0.0, $110 = 0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0;
 var $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0, $128 = 0.0, $129 = 0.0, $13 = 0.0, $130 = 0, $131 = 0.0, $132 = 0.0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0.0, $138 = 0, $139 = 0.0, $14 = 0.0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0.0, $148 = 0.0, $149 = 0.0, $15 = 0.0, $150 = 0, $151 = 0.0;
 var $152 = 0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0.0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0, $166 = 0.0, $167 = 0.0, $168 = 0, $169 = 0.0, $17 = 0.0;
 var $170 = 0, $171 = 0.0, $172 = 0.0, $173 = 0, $174 = 0.0, $175 = 0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0.0, $180 = 0.0, $181 = 0.0, $182 = 0.0, $183 = 0.0, $184 = 0, $185 = 0.0, $186 = 0.0, $187 = 0.0, $188 = 0.0;
 var $189 = 0.0, $19 = 0.0, $190 = 0.0, $191 = 0.0, $192 = 0, $193 = 0.0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0.0, $198 = 0, $199 = 0, $2 = 0, $20 = 0.0, $200 = 0, $201 = 0.0, $202 = 0.0, $21 = 0, $22 = 0.0, $23 = 0.0;
 var $24 = 0, $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0.0;
 var $42 = 0.0, $43 = 0.0, $44 = 0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0, $58 = 0, $59 = 0.0, $6 = 0.0;
 var $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0.0, $67 = 0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0.0, $78 = 0;
 var $79 = 0.0, $8 = 0.0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0, $86 = 0.0, $87 = 0, $88 = 0.0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0;
 var $97 = 0.0, $98 = 0.0, $99 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pc)) + 84|0); //@line 623 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 623 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $2 = ($1|0)>(0); //@line 623 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 __Z10loveAssertbPKc($2,446916); //@line 623 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $3 = ((($pc)) + 72|0); //@line 625 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 625 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 switch ($4|0) {
 case 0:  {
  $5 = ((($xfA)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $6 = +HEAPF32[$5>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $7 = ((($pc)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $9 = $6 * $8; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $10 = ((($xfA)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $11 = +HEAPF32[$10>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $12 = ((($pc)) + 28|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $13 = +HEAPF32[$12>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $14 = $11 * $13; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $15 = $9 - $14; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $16 = +HEAPF32[$xfA>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $17 = $16 + $15; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $18 = $8 * $11; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $19 = $6 * $13; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $20 = $18 + $19; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $21 = ((($xfA)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $22 = +HEAPF32[$21>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $23 = $20 + $22; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $24 = ((($xfB)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $25 = +HEAPF32[$24>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $26 = +HEAPF32[$pc>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $27 = $25 * $26; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $28 = ((($xfB)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $29 = +HEAPF32[$28>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $30 = ((($pc)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $31 = +HEAPF32[$30>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $32 = $29 * $31; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $33 = $27 - $32; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $34 = +HEAPF32[$xfB>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $35 = $34 + $33; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $36 = $26 * $29; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $37 = $25 * $31; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $38 = $36 + $37; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $39 = ((($xfB)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $40 = +HEAPF32[$39>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $41 = $38 + $40; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $42 = $35 - $17; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $43 = $41 - $23; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$this>>2] = $42; //@line 631 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $44 = ((($this)) + 4|0); //@line 631 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$44>>2] = $43; //@line 631 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $45 = $42 * $42; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $46 = $43 * $43; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $47 = $45 + $46; //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $48 = (+Math_sqrt((+$47))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $49 = $48 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
  if ($49) {
   $60 = $42;$62 = $43;
  } else {
   $50 = 1.0 / $48; //@line 121 "./libraries/Box2D/Common/b2Math.h"
   $51 = $42 * $50; //@line 122 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$this>>2] = $51; //@line 122 "./libraries/Box2D/Common/b2Math.h"
   $52 = $43 * $50; //@line 123 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$44>>2] = $52; //@line 123 "./libraries/Box2D/Common/b2Math.h"
   $60 = $51;$62 = $52;
  }
  $53 = $17 + $35; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $54 = $23 + $41; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $55 = $53 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $56 = $54 * 0.5; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $57 = ((($this)) + 8|0); //@line 633 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$57>>2] = $55; //@line 633 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $58 = ((($this)) + 12|0); //@line 633 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$58>>2] = $56; //@line 633 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $59 = $42 * $60; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $61 = $43 * $62; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $63 = $59 + $61; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $64 = ((($pc)) + 76|0); //@line 634 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $65 = +HEAPF32[$64>>2]; //@line 634 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $66 = $63 - $65; //@line 634 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $67 = ((($pc)) + 80|0); //@line 634 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $68 = +HEAPF32[$67>>2]; //@line 634 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $69 = $66 - $68; //@line 634 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $70 = ((($this)) + 16|0); //@line 634 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$70>>2] = $69; //@line 634 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  return; //@line 663 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  break;
 }
 case 1:  {
  $71 = ((($xfA)) + 12|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $72 = +HEAPF32[$71>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $73 = ((($pc)) + 16|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $74 = +HEAPF32[$73>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $75 = $72 * $74; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $76 = ((($xfA)) + 8|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $77 = +HEAPF32[$76>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $78 = ((($pc)) + 20|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $79 = +HEAPF32[$78>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $80 = $77 * $79; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $81 = $75 - $80; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $82 = $74 * $77; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $83 = $72 * $79; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $84 = $82 + $83; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$this>>2] = $81; //@line 640 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $85 = ((($this)) + 4|0); //@line 640 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$85>>2] = $84; //@line 640 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $86 = +HEAPF32[$71>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $87 = ((($pc)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $88 = +HEAPF32[$87>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $89 = $86 * $88; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $90 = +HEAPF32[$76>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $91 = ((($pc)) + 28|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $92 = +HEAPF32[$91>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $93 = $90 * $92; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $94 = $89 - $93; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $95 = +HEAPF32[$xfA>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $96 = $95 + $94; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $97 = $88 * $90; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $98 = $86 * $92; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $99 = $97 + $98; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $100 = ((($xfA)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $101 = +HEAPF32[$100>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $102 = $99 + $101; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $103 = ((($xfB)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $104 = +HEAPF32[$103>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $105 = (($pc) + ($index<<3)|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $106 = +HEAPF32[$105>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $107 = $104 * $106; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $108 = ((($xfB)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $109 = +HEAPF32[$108>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $110 = (((($pc) + ($index<<3)|0)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $111 = +HEAPF32[$110>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $112 = $109 * $111; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $113 = $107 - $112; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $114 = +HEAPF32[$xfB>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $115 = $114 + $113; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $116 = $106 * $109; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $117 = $104 * $111; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $118 = $116 + $117; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $119 = ((($xfB)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $120 = +HEAPF32[$119>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $121 = $118 + $120; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $122 = $115 - $96; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $123 = $121 - $102; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $124 = $81 * $122; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $125 = $123 * $84; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $126 = $124 + $125; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $127 = ((($pc)) + 76|0); //@line 644 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $128 = +HEAPF32[$127>>2]; //@line 644 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $129 = $126 - $128; //@line 644 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $130 = ((($pc)) + 80|0); //@line 644 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $131 = +HEAPF32[$130>>2]; //@line 644 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $132 = $129 - $131; //@line 644 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $133 = ((($this)) + 16|0); //@line 644 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$133>>2] = $132; //@line 644 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $134 = ((($this)) + 8|0); //@line 645 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$134>>2] = $115; //@line 645 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $135 = ((($this)) + 12|0); //@line 645 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$135>>2] = $121; //@line 645 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  return; //@line 663 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  break;
 }
 case 2:  {
  $136 = ((($xfB)) + 12|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $137 = +HEAPF32[$136>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $138 = ((($pc)) + 16|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $139 = +HEAPF32[$138>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $140 = $137 * $139; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $141 = ((($xfB)) + 8|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $142 = +HEAPF32[$141>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $143 = ((($pc)) + 20|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $144 = +HEAPF32[$143>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $145 = $142 * $144; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $146 = $140 - $145; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $147 = $139 * $142; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $148 = $137 * $144; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $149 = $147 + $148; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$this>>2] = $146; //@line 651 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $150 = ((($this)) + 4|0); //@line 651 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$150>>2] = $149; //@line 651 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $151 = +HEAPF32[$136>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $152 = ((($pc)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $153 = +HEAPF32[$152>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $154 = $151 * $153; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $155 = +HEAPF32[$141>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $156 = ((($pc)) + 28|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $157 = +HEAPF32[$156>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $158 = $155 * $157; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $159 = $154 - $158; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $160 = +HEAPF32[$xfB>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $161 = $160 + $159; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $162 = $153 * $155; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $163 = $151 * $157; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $164 = $162 + $163; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $165 = ((($xfB)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $166 = +HEAPF32[$165>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $167 = $164 + $166; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $168 = ((($xfA)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $169 = +HEAPF32[$168>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $170 = (($pc) + ($index<<3)|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $171 = +HEAPF32[$170>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $172 = $169 * $171; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $173 = ((($xfA)) + 8|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $174 = +HEAPF32[$173>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $175 = (((($pc) + ($index<<3)|0)) + 4|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $176 = +HEAPF32[$175>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $177 = $174 * $176; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $178 = $172 - $177; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $179 = +HEAPF32[$xfA>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $180 = $179 + $178; //@line 579 "./libraries/Box2D/Common/b2Math.h"
  $181 = $171 * $174; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $182 = $169 * $176; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $183 = $181 + $182; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $184 = ((($xfA)) + 4|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $185 = +HEAPF32[$184>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $186 = $183 + $185; //@line 580 "./libraries/Box2D/Common/b2Math.h"
  $187 = $180 - $161; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $188 = $186 - $167; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $189 = $146 * $187; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $190 = $188 * $149; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $191 = $189 + $190; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $192 = ((($pc)) + 76|0); //@line 655 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $193 = +HEAPF32[$192>>2]; //@line 655 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $194 = $191 - $193; //@line 655 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $195 = ((($pc)) + 80|0); //@line 655 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $196 = +HEAPF32[$195>>2]; //@line 655 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $197 = $194 - $196; //@line 655 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $198 = ((($this)) + 16|0); //@line 655 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$198>>2] = $197; //@line 655 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $199 = ((($this)) + 8|0); //@line 656 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$199>>2] = $180; //@line 656 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $200 = ((($this)) + 12|0); //@line 656 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$200>>2] = $186; //@line 656 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $201 = -$146; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  $202 = -$149; //@line 68 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$this>>2] = $201; //@line 659 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$150>>2] = $202; //@line 659 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  return; //@line 663 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  break;
 }
 default: {
  return; //@line 663 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 }
 }
}
function __ZN15b2ContactSolver27SolveTOIPositionConstraintsEii($this,$toiIndexA,$toiIndexB) {
 $this = $this|0;
 $toiIndexA = $toiIndexA|0;
 $toiIndexB = $toiIndexB|0;
 var $$lcssa = 0.0, $$lcssa63 = 0, $$lcssa64 = 0, $$lcssa65 = 0.0, $$lcssa66 = 0, $$lcssa67 = 0, $$lcssa68 = 0.0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0;
 var $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0, $123 = 0, $124 = 0.0, $125 = 0.0, $126 = 0.0;
 var $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0.0, $132 = 0.0, $133 = 0, $134 = 0.0, $135 = 0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0;
 var $145 = 0.0, $146 = 0, $147 = 0.0, $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0.0, $152 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0;
 var $26 = 0, $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0, $43 = 0.0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0;
 var $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0;
 var $99 = 0.0, $aA$0$lcssa = 0.0, $aA$023 = 0.0, $aB$0$lcssa = 0.0, $aB$024 = 0.0, $cA$sroa$0$0$lcssa = 0, $cA$sroa$0$021 = 0, $cA$sroa$8$0$lcssa = 0, $cA$sroa$8$022 = 0, $cB$sroa$0$0$lcssa = 0, $cB$sroa$0$019 = 0, $cB$sroa$8$0$lcssa = 0, $cB$sroa$8$020 = 0, $exitcond = 0, $i$033 = 0, $iA$0 = 0.0, $iB$0 = 0.0, $j$025 = 0, $mA$0 = 0.0, $mB$0 = 0.0;
 var $minSeparation$0$lcssa = 0.0, $minSeparation$034 = 0.0, $minSeparation$1$lcssa = 0.0, $minSeparation$126 = 0.0, $or$cond = 0, $or$cond8 = 0, $psm = 0, $xfA = 0, $xfB = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $xfA = sp + 40|0;
 $xfB = sp + 24|0;
 $psm = sp;
 $0 = ((($this)) + 48|0); //@line 754 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 754 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $2 = ($1|0)>(0); //@line 754 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 if (!($2)) {
  $minSeparation$0$lcssa = 0.0;
  $15 = $minSeparation$0$lcssa >= -0.0074999998323619366; //@line 837 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  STACKTOP = sp;return ($15|0); //@line 837 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 }
 $3 = ((($this)) + 36|0); //@line 756 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $4 = ((($this)) + 24|0); //@line 780 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $5 = ((($xfA)) + 8|0); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($xfA)) + 12|0); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($xfB)) + 8|0); //@line 315 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($xfB)) + 12|0); //@line 316 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($xfA)) + 4|0); //@line 792 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $10 = ((($xfB)) + 4|0); //@line 793 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $11 = ((($psm)) + 4|0); //@line 797 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $12 = ((($psm)) + 8|0); //@line 799 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $13 = ((($psm)) + 12|0); //@line 799 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $14 = ((($psm)) + 16|0); //@line 800 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 $i$033 = 0;$minSeparation$034 = 0.0;
 while(1) {
  $16 = HEAP32[$3>>2]|0; //@line 756 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $17 = (($16) + (($i$033*88)|0)|0); //@line 756 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $18 = (((($16) + (($i$033*88)|0)|0)) + 32|0); //@line 758 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 758 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $20 = (((($16) + (($i$033*88)|0)|0)) + 36|0); //@line 759 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $21 = HEAP32[$20>>2]|0; //@line 759 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $22 = (((($16) + (($i$033*88)|0)|0)) + 48|0); //@line 760 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $23 = +HEAPF32[$22>>2]; //@line 760 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $24 = (((($16) + (($i$033*88)|0)|0)) + 52|0); //@line 760 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $25 = +HEAPF32[$24>>2]; //@line 760 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $26 = (((($16) + (($i$033*88)|0)|0)) + 56|0); //@line 761 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $27 = +HEAPF32[$26>>2]; //@line 761 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $28 = (((($16) + (($i$033*88)|0)|0)) + 60|0); //@line 761 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $29 = +HEAPF32[$28>>2]; //@line 761 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $30 = (((($16) + (($i$033*88)|0)|0)) + 84|0); //@line 762 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $31 = HEAP32[$30>>2]|0; //@line 762 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $32 = ($19|0)==($toiIndexA|0); //@line 766 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $33 = ($19|0)==($toiIndexB|0); //@line 766 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $or$cond = $32 | $33; //@line 766 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  if ($or$cond) {
   $34 = (((($16) + (($i$033*88)|0)|0)) + 40|0); //@line 768 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
   $35 = +HEAPF32[$34>>2]; //@line 768 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
   $36 = (((($16) + (($i$033*88)|0)|0)) + 64|0); //@line 769 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
   $37 = +HEAPF32[$36>>2]; //@line 769 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
   $iA$0 = $37;$mA$0 = $35;
  } else {
   $iA$0 = 0.0;$mA$0 = 0.0;
  }
  $38 = ($21|0)==($toiIndexA|0); //@line 774 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $39 = ($21|0)==($toiIndexB|0); //@line 774 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $or$cond8 = $38 | $39; //@line 774 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  if ($or$cond8) {
   $40 = (((($16) + (($i$033*88)|0)|0)) + 44|0); //@line 776 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
   $41 = +HEAPF32[$40>>2]; //@line 776 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
   $42 = (((($16) + (($i$033*88)|0)|0)) + 68|0); //@line 777 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
   $43 = +HEAPF32[$42>>2]; //@line 777 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
   $iB$0 = $43;$mB$0 = $41;
  } else {
   $iB$0 = 0.0;$mB$0 = 0.0;
  }
  $44 = HEAP32[$4>>2]|0; //@line 780 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $45 = (($44) + (($19*12)|0)|0); //@line 780 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $46 = HEAP32[$45>>2]|0; //@line 780 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $47 = (((($44) + (($19*12)|0)|0)) + 4|0); //@line 780 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $48 = HEAP32[$47>>2]|0; //@line 780 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $49 = (((($44) + (($19*12)|0)|0)) + 8|0); //@line 781 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $50 = +HEAPF32[$49>>2]; //@line 781 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $51 = (($44) + (($21*12)|0)|0); //@line 783 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $52 = HEAP32[$51>>2]|0; //@line 783 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $53 = (((($44) + (($21*12)|0)|0)) + 4|0); //@line 783 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $54 = HEAP32[$53>>2]|0; //@line 783 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $55 = (((($44) + (($21*12)|0)|0)) + 8|0); //@line 784 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $56 = +HEAPF32[$55>>2]; //@line 784 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $57 = ($31|0)>(0); //@line 787 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  if ($57) {
   $58 = $mA$0 + $mB$0; //@line 814 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
   $aA$023 = $50;$aB$024 = $56;$cA$sroa$0$021 = $46;$cA$sroa$8$022 = $48;$cB$sroa$0$019 = $52;$cB$sroa$8$020 = $54;$j$025 = 0;$minSeparation$126 = $minSeparation$034;
   while(1) {
    $71 = (+Math_sin((+$aA$023))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$5>>2] = $71; //@line 315 "./libraries/Box2D/Common/b2Math.h"
    $72 = (+Math_cos((+$aA$023))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$6>>2] = $72; //@line 316 "./libraries/Box2D/Common/b2Math.h"
    $73 = (+Math_sin((+$aB$024))); //@line 315 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$7>>2] = $73; //@line 315 "./libraries/Box2D/Common/b2Math.h"
    $74 = (+Math_cos((+$aB$024))); //@line 316 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$8>>2] = $74; //@line 316 "./libraries/Box2D/Common/b2Math.h"
    $75 = $23 * $72; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $76 = $25 * $71; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $77 = $75 - $76; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $78 = $23 * $71; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $79 = $25 * $72; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $80 = $79 + $78; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $81 = (HEAP32[tempDoublePtr>>2]=$cA$sroa$0$021,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $82 = $81 - $77; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $83 = (HEAP32[tempDoublePtr>>2]=$cA$sroa$8$022,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $84 = $83 - $80; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$xfA>>2] = $82; //@line 792 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    HEAPF32[$9>>2] = $84; //@line 792 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $85 = $27 * $74; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $86 = $29 * $73; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $87 = $85 - $86; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $88 = $27 * $73; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $89 = $29 * $74; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $90 = $89 + $88; //@line 568 "./libraries/Box2D/Common/b2Math.h"
    $91 = (HEAP32[tempDoublePtr>>2]=$cB$sroa$0$019,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $92 = $91 - $87; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $93 = (HEAP32[tempDoublePtr>>2]=$cB$sroa$8$020,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $94 = $93 - $90; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    HEAPF32[$xfB>>2] = $92; //@line 793 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    HEAPF32[$10>>2] = $94; //@line 793 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    __ZN24b2PositionSolverManifold10InitializeEP27b2ContactPositionConstraintRK11b2TransformS4_i($psm,$17,$xfA,$xfB,$j$025); //@line 796 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $95 = +HEAPF32[$psm>>2]; //@line 797 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $96 = +HEAPF32[$11>>2]; //@line 797 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $97 = +HEAPF32[$12>>2]; //@line 799 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $98 = +HEAPF32[$13>>2]; //@line 799 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $99 = +HEAPF32[$14>>2]; //@line 800 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $100 = $97 - $81; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $101 = $98 - $83; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $102 = $97 - $91; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $103 = $98 - $93; //@line 454 "./libraries/Box2D/Common/b2Math.h"
    $104 = $minSeparation$126 < $99; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $105 = $104 ? $minSeparation$126 : $99; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $106 = $99 + 0.004999999888241291; //@line 809 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $107 = $106 * 0.75; //@line 809 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $108 = $107 < 0.0; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $109 = $108 ? $107 : 0.0; //@line 634 "./libraries/Box2D/Common/b2Math.h"
    $110 = $96 * $100; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $111 = $95 * $101; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $112 = $110 - $111; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $113 = $96 * $102; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $114 = $95 * $103; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $115 = $113 - $114; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $116 = $iA$0 * $112; //@line 814 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $117 = $112 * $116; //@line 814 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $118 = $58 + $117; //@line 814 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $119 = $iB$0 * $115; //@line 814 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $120 = $115 * $119; //@line 814 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $121 = $120 + $118; //@line 814 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $122 = $121 > 0.0; //@line 817 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    if ($122) {
     $123 = $109 < -0.20000000298023224; //@line 645 "./libraries/Box2D/Common/b2Math.h"
     $124 = $123 ? -0.20000000298023224 : $109; //@line 645 "./libraries/Box2D/Common/b2Math.h"
     $125 = -$124; //@line 817 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $126 = $125 / $121; //@line 817 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
     $128 = $126;
    } else {
     $128 = 0.0;
    }
    $127 = $95 * $128; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $129 = $96 * $128; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $130 = $mA$0 * $127; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $131 = $mA$0 * $129; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $132 = $81 - $130; //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $133 = (HEAPF32[tempDoublePtr>>2]=$132,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $134 = $83 - $131; //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $135 = (HEAPF32[tempDoublePtr>>2]=$134,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
    $136 = $100 * $129; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $137 = $101 * $127; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $138 = $136 - $137; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $139 = $iA$0 * $138; //@line 822 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $140 = $aA$023 - $139; //@line 822 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $141 = $mB$0 * $127; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $142 = $mB$0 * $129; //@line 459 "./libraries/Box2D/Common/b2Math.h"
    $143 = $91 + $141; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $144 = (HEAPF32[tempDoublePtr>>2]=$143,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $145 = $93 + $142; //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $146 = (HEAPF32[tempDoublePtr>>2]=$145,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
    $147 = $102 * $129; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $148 = $103 * $127; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $149 = $147 - $148; //@line 414 "./libraries/Box2D/Common/b2Math.h"
    $150 = $iB$0 * $149; //@line 825 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $151 = $aB$024 + $150; //@line 825 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $152 = (($j$025) + 1)|0; //@line 787 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    $exitcond = ($152|0)==($31|0); //@line 787 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
    if ($exitcond) {
     $$lcssa = $105;$$lcssa63 = $133;$$lcssa64 = $135;$$lcssa65 = $140;$$lcssa66 = $144;$$lcssa67 = $146;$$lcssa68 = $151;
     break;
    } else {
     $aA$023 = $140;$aB$024 = $151;$cA$sroa$0$021 = $133;$cA$sroa$8$022 = $135;$cB$sroa$0$019 = $144;$cB$sroa$8$020 = $146;$j$025 = $152;$minSeparation$126 = $105;
    }
   }
   $$pre = HEAP32[$4>>2]|0; //@line 828 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
   $60 = $$pre;$aA$0$lcssa = $$lcssa65;$aB$0$lcssa = $$lcssa68;$cA$sroa$0$0$lcssa = $$lcssa63;$cA$sroa$8$0$lcssa = $$lcssa64;$cB$sroa$0$0$lcssa = $$lcssa66;$cB$sroa$8$0$lcssa = $$lcssa67;$minSeparation$1$lcssa = $$lcssa;
  } else {
   $60 = $44;$aA$0$lcssa = $50;$aB$0$lcssa = $56;$cA$sroa$0$0$lcssa = $46;$cA$sroa$8$0$lcssa = $48;$cB$sroa$0$0$lcssa = $52;$cB$sroa$8$0$lcssa = $54;$minSeparation$1$lcssa = $minSeparation$034;
  }
  $59 = (($60) + (($19*12)|0)|0); //@line 828 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$59>>2] = $cA$sroa$0$0$lcssa; //@line 828 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $61 = (((($60) + (($19*12)|0)|0)) + 4|0); //@line 828 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$61>>2] = $cA$sroa$8$0$lcssa; //@line 828 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $62 = HEAP32[$4>>2]|0; //@line 829 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $63 = (((($62) + (($19*12)|0)|0)) + 8|0); //@line 829 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$63>>2] = $aA$0$lcssa; //@line 829 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $64 = (($62) + (($21*12)|0)|0); //@line 831 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$64>>2] = $cB$sroa$0$0$lcssa; //@line 831 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $65 = (((($62) + (($21*12)|0)|0)) + 4|0); //@line 831 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAP32[$65>>2] = $cB$sroa$8$0$lcssa; //@line 831 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $66 = HEAP32[$4>>2]|0; //@line 832 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $67 = (((($66) + (($21*12)|0)|0)) + 8|0); //@line 832 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  HEAPF32[$67>>2] = $aB$0$lcssa; //@line 832 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $68 = (($i$033) + 1)|0; //@line 754 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $69 = HEAP32[$0>>2]|0; //@line 754 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  $70 = ($68|0)<($69|0); //@line 754 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
  if ($70) {
   $i$033 = $68;$minSeparation$034 = $minSeparation$1$lcssa;
  } else {
   $minSeparation$0$lcssa = $minSeparation$1$lcssa;
   break;
  }
 }
 $15 = $minSeparation$0$lcssa >= -0.0074999998323619366; //@line 837 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
 STACKTOP = sp;return ($15|0); //@line 837 "libraries/Box2D/Dynamics/Contacts/b2ContactSolver.cpp"
}
function __ZNK15b2DistanceJoint10GetAnchorAEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 225 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 225 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 80|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 84|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 225 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
}
function __ZNK15b2DistanceJoint10GetAnchorBEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 230 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 230 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 88|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 92|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 230 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
}
function __ZNK15b2DistanceJoint16GetReactionForceEf($agg$result,$this,$inv_dt) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 100|0); //@line 235 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 235 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $2 = $1 * $inv_dt; //@line 235 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $3 = ((($this)) + 116|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$3>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $5 = $4 * $2; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($this)) + 120|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $8 = $2 * $7; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $5; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$9>>2] = $8; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 236 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
}
function __ZNK15b2DistanceJoint17GetReactionTorqueEf($this,$inv_dt) {
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return +0; //@line 242 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
}
function __ZN15b2DistanceJoint4DumpEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer13 = 0, $vararg_buffer17 = 0, $vararg_buffer20 = 0, $vararg_buffer23 = 0;
 var $vararg_buffer26 = 0, $vararg_buffer3 = 0, $vararg_buffer6 = 0, $vararg_buffer9 = 0, $vararg_ptr12 = 0, $vararg_ptr16 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer26 = sp + 88|0;
 $vararg_buffer23 = sp + 80|0;
 $vararg_buffer20 = sp + 72|0;
 $vararg_buffer17 = sp + 64|0;
 $vararg_buffer13 = sp + 48|0;
 $vararg_buffer9 = sp + 32|0;
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($this)) + 48|0); //@line 247 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 247 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 247 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 247 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $4 = ((($this)) + 52|0); //@line 248 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 248 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $6 = ((($5)) + 8|0); //@line 248 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 248 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 __Z5b2LogPKcz(446953,$vararg_buffer); //@line 250 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$vararg_buffer1>>2] = $3; //@line 251 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 __Z5b2LogPKcz(447349,$vararg_buffer1); //@line 251 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$vararg_buffer3>>2] = $7; //@line 252 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 __Z5b2LogPKcz(447375,$vararg_buffer3); //@line 252 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $8 = ((($this)) + 61|0); //@line 253 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 253 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $10 = $9&255; //@line 253 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$vararg_buffer6>>2] = $10; //@line 253 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 __Z5b2LogPKcz(447401,$vararg_buffer6); //@line 253 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $11 = ((($this)) + 80|0); //@line 254 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $12 = +HEAPF32[$11>>2]; //@line 254 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $13 = $12; //@line 254 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $14 = ((($this)) + 84|0); //@line 254 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 254 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $16 = $15; //@line 254 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF64[$vararg_buffer9>>3] = $13; //@line 254 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $vararg_ptr12 = ((($vararg_buffer9)) + 8|0); //@line 254 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF64[$vararg_ptr12>>3] = $16; //@line 254 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 __Z5b2LogPKcz(447106,$vararg_buffer9); //@line 254 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $17 = ((($this)) + 88|0); //@line 255 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $18 = +HEAPF32[$17>>2]; //@line 255 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $19 = $18; //@line 255 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $20 = ((($this)) + 92|0); //@line 255 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 255 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $22 = $21; //@line 255 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF64[$vararg_buffer13>>3] = $19; //@line 255 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $vararg_ptr16 = ((($vararg_buffer13)) + 8|0); //@line 255 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF64[$vararg_ptr16>>3] = $22; //@line 255 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 __Z5b2LogPKcz(447148,$vararg_buffer13); //@line 255 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $23 = ((($this)) + 104|0); //@line 256 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $24 = +HEAPF32[$23>>2]; //@line 256 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $25 = $24; //@line 256 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF64[$vararg_buffer17>>3] = $25; //@line 256 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 __Z5b2LogPKcz(446979,$vararg_buffer17); //@line 256 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $26 = ((($this)) + 68|0); //@line 257 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $27 = +HEAPF32[$26>>2]; //@line 257 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $28 = $27; //@line 257 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF64[$vararg_buffer20>>3] = $28; //@line 257 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 __Z5b2LogPKcz(447003,$vararg_buffer20); //@line 257 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $29 = ((($this)) + 72|0); //@line 258 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $30 = +HEAPF32[$29>>2]; //@line 258 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $31 = $30; //@line 258 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF64[$vararg_buffer23>>3] = $31; //@line 258 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 __Z5b2LogPKcz(447032,$vararg_buffer23); //@line 258 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $32 = ((($this)) + 56|0); //@line 259 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $33 = HEAP32[$32>>2]|0; //@line 259 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$vararg_buffer26>>2] = $33; //@line 259 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 __Z5b2LogPKcz(447513,$vararg_buffer26); //@line 259 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 STACKTOP = sp;return; //@line 260 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
}
function __ZN15b2DistanceJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 67 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
 return; //@line 67 "./libraries/Box2D/Dynamics/Joints/b2DistanceJoint.h"
}
function __ZN15b2DistanceJoint23InitVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$sink = 0.0, $$sink23 = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0, $107 = 0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0;
 var $114 = 0, $115 = 0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0.0;
 var $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0, $144 = 0.0, $145 = 0.0, $146 = 0, $147 = 0, $148 = 0.0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0.0, $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0, $166 = 0.0, $167 = 0.0, $168 = 0.0;
 var $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0.0, $172 = 0, $173 = 0.0, $174 = 0.0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0.0, $182 = 0, $183 = 0.0, $184 = 0.0, $185 = 0.0, $186 = 0.0;
 var $187 = 0.0, $188 = 0.0, $189 = 0.0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0.0, $194 = 0, $195 = 0.0, $196 = 0.0, $197 = 0.0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0.0, $201 = 0.0, $202 = 0.0, $203 = 0.0;
 var $204 = 0, $205 = 0.0, $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0.0, $212 = 0.0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0, $53 = 0.0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0.0, $95 = 0.0, $96 = 0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $bitcast = 0.0, $bitcast25 = 0.0, $vA$sroa$0$0 = 0, $vA$sroa$6$0 = 0, $vB$sroa$0$0 = 0, $vB$sroa$6$0 = 0, $wA$0 = 0.0, $wB$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 64 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 64 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $4 = ((($this)) + 108|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$4>>2] = $3; //@line 64 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $5 = ((($this)) + 52|0); //@line 65 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 65 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $7 = ((($6)) + 8|0); //@line 65 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 65 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $9 = ((($this)) + 112|0); //@line 65 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$9>>2] = $8; //@line 65 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $10 = ((($this)) + 140|0); //@line 66 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $11 = ((($1)) + 28|0); //@line 66 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $12 = $11; //@line 66 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $13 = $12; //@line 66 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 66 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $15 = (($12) + 4)|0; //@line 66 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $16 = $15; //@line 66 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 66 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $18 = $10; //@line 66 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $19 = $18; //@line 66 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$19>>2] = $14; //@line 66 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $20 = (($18) + 4)|0; //@line 66 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $21 = $20; //@line 66 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$21>>2] = $17; //@line 66 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $22 = ((($this)) + 148|0); //@line 67 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $23 = ((($6)) + 28|0); //@line 67 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $24 = $23; //@line 67 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $25 = $24; //@line 67 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 67 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $27 = (($24) + 4)|0; //@line 67 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $28 = $27; //@line 67 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 67 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $30 = $22; //@line 67 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $31 = $30; //@line 67 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$31>>2] = $26; //@line 67 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $32 = (($30) + 4)|0; //@line 67 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $33 = $32; //@line 67 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$33>>2] = $29; //@line 67 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $34 = ((($1)) + 120|0); //@line 68 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 68 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $36 = ((($this)) + 156|0); //@line 68 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$36>>2] = $35; //@line 68 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $37 = ((($6)) + 120|0); //@line 69 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $38 = HEAP32[$37>>2]|0; //@line 69 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $39 = ((($this)) + 160|0); //@line 69 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$39>>2] = $38; //@line 69 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $40 = ((($1)) + 128|0); //@line 70 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $41 = HEAP32[$40>>2]|0; //@line 70 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $42 = ((($this)) + 164|0); //@line 70 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$42>>2] = $41; //@line 70 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $43 = ((($6)) + 128|0); //@line 71 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $44 = HEAP32[$43>>2]|0; //@line 71 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $45 = ((($this)) + 168|0); //@line 71 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$45>>2] = $44; //@line 71 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $46 = ((($data)) + 24|0); //@line 73 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $47 = HEAP32[$46>>2]|0; //@line 73 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $48 = (($47) + (($3*12)|0)|0); //@line 73 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $49 = +HEAPF32[$48>>2]; //@line 73 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $50 = (((($47) + (($3*12)|0)|0)) + 4|0); //@line 73 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $51 = +HEAPF32[$50>>2]; //@line 73 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $52 = (((($47) + (($3*12)|0)|0)) + 8|0); //@line 74 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $53 = +HEAPF32[$52>>2]; //@line 74 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $54 = ((($data)) + 28|0); //@line 75 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $55 = HEAP32[$54>>2]|0; //@line 75 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $56 = (($55) + (($3*12)|0)|0); //@line 75 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $57 = HEAP32[$56>>2]|0; //@line 75 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $58 = (((($55) + (($3*12)|0)|0)) + 4|0); //@line 75 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $59 = HEAP32[$58>>2]|0; //@line 75 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $60 = (((($55) + (($3*12)|0)|0)) + 8|0); //@line 76 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $61 = +HEAPF32[$60>>2]; //@line 76 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $62 = (($47) + (($8*12)|0)|0); //@line 78 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $63 = +HEAPF32[$62>>2]; //@line 78 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $64 = (((($47) + (($8*12)|0)|0)) + 4|0); //@line 78 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $65 = +HEAPF32[$64>>2]; //@line 78 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $66 = (((($47) + (($8*12)|0)|0)) + 8|0); //@line 79 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $67 = +HEAPF32[$66>>2]; //@line 79 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $68 = (($55) + (($8*12)|0)|0); //@line 80 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $69 = HEAP32[$68>>2]|0; //@line 80 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $70 = (((($55) + (($8*12)|0)|0)) + 4|0); //@line 80 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $71 = HEAP32[$70>>2]|0; //@line 80 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $72 = (((($55) + (($8*12)|0)|0)) + 8|0); //@line 81 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $73 = +HEAPF32[$72>>2]; //@line 81 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $74 = (+Math_sin((+$53))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $75 = (+Math_cos((+$53))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $76 = (+Math_sin((+$67))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $77 = (+Math_cos((+$67))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $78 = ((($this)) + 80|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $79 = +HEAPF32[$78>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $bitcast = (HEAP32[tempDoublePtr>>2]=$14,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $80 = $79 - $bitcast; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $81 = ((($this)) + 84|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $82 = +HEAPF32[$81>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $83 = (HEAP32[tempDoublePtr>>2]=$17,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $84 = $82 - $83; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $85 = $75 * $80; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $86 = $74 * $84; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $87 = $85 - $86; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $88 = $74 * $80; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $89 = $75 * $84; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $90 = $88 + $89; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $91 = ((($this)) + 124|0); //@line 85 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$91>>2] = $87; //@line 85 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $92 = ((($this)) + 128|0); //@line 85 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$92>>2] = $90; //@line 85 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $93 = ((($this)) + 88|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $94 = +HEAPF32[$93>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $bitcast25 = (HEAP32[tempDoublePtr>>2]=$26,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $95 = $94 - $bitcast25; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $96 = ((($this)) + 92|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $97 = +HEAPF32[$96>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $98 = (HEAP32[tempDoublePtr>>2]=$29,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $99 = $97 - $98; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $100 = $77 * $95; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $101 = $76 * $99; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $102 = $100 - $101; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $103 = $76 * $95; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $104 = $77 * $99; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $105 = $103 + $104; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $106 = ((($this)) + 132|0); //@line 86 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$106>>2] = $102; //@line 86 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $107 = ((($this)) + 136|0); //@line 86 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$107>>2] = $105; //@line 86 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $108 = $63 + $102; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $109 = $65 + $105; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $110 = $108 - $49; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $111 = $109 - $51; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $112 = $110 - $87; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $113 = $111 - $90; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 $114 = ((($this)) + 116|0); //@line 87 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$114>>2] = $112; //@line 87 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $115 = ((($this)) + 120|0); //@line 87 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$115>>2] = $113; //@line 87 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $116 = $112 * $112; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $117 = $113 * $113; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $118 = $116 + $117; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $119 = (+Math_sqrt((+$118))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $120 = $119 > 0.004999999888241291; //@line 91 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 if ($120) {
  $121 = 1.0 / $119; //@line 93 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $122 = $112 * $121; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$114>>2] = $122; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $123 = $121 * $113; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $124 = $123;$127 = $122;
 } else {
  HEAPF32[$114>>2] = 0.0; //@line 65 "./libraries/Box2D/Common/b2Math.h"
  $124 = 0.0;$127 = 0.0;
 }
 HEAPF32[$115>>2] = $124; //@line 97 "./libraries/Box2D/Common/b2Math.h"
 $125 = $124 * $87; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $126 = $90 * $127; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $128 = $125 - $126; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $129 = $124 * $102; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $130 = $127 * $105; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $131 = $129 - $130; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $132 = +HEAPF32[$36>>2]; //@line 141 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $133 = +HEAPF32[$42>>2]; //@line 142 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $134 = $128 * $133; //@line 102 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $135 = $128 * $134; //@line 102 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $136 = $132 + $135; //@line 102 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $137 = +HEAPF32[$39>>2]; //@line 143 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $138 = $137 + $136; //@line 102 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $139 = +HEAPF32[$45>>2]; //@line 144 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $140 = $131 * $139; //@line 102 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $141 = $131 * $140; //@line 102 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $142 = $138 + $141; //@line 102 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $143 = $142 != 0.0; //@line 105 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $144 = 1.0 / $142; //@line 105 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $145 = $143 ? $144 : 0.0; //@line 105 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $146 = ((($this)) + 172|0); //@line 105 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$146>>2] = $145; //@line 105 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $147 = ((($this)) + 68|0); //@line 107 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $148 = +HEAPF32[$147>>2]; //@line 107 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $149 = $148 > 0.0; //@line 107 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 if ($149) {
  $150 = ((($this)) + 104|0); //@line 109 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $151 = +HEAPF32[$150>>2]; //@line 109 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $152 = $119 - $151; //@line 109 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $153 = $148 * 6.2831854820251465; //@line 112 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $154 = $145 * 2.0; //@line 115 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $155 = ((($this)) + 72|0); //@line 115 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $156 = +HEAPF32[$155>>2]; //@line 115 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $157 = $154 * $156; //@line 115 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $158 = $153 * $157; //@line 115 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $159 = $153 * $145; //@line 118 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $160 = $153 * $159; //@line 118 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $161 = +HEAPF32[$data>>2]; //@line 121 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $162 = $160 * $161; //@line 122 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $163 = $158 + $162; //@line 122 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $164 = $161 * $163; //@line 122 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $165 = $164 != 0.0; //@line 123 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $166 = 1.0 / $164; //@line 123 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $167 = $165 ? $166 : 0.0; //@line 123 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $168 = $152 * $161; //@line 124 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $169 = $160 * $168; //@line 124 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $170 = $169 * $167; //@line 124 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $171 = $142 + $167; //@line 126 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $172 = $171 != 0.0; //@line 127 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $173 = 1.0 / $171; //@line 127 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $174 = $172 ? $173 : 0.0; //@line 127 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  HEAPF32[$146>>2] = $174; //@line 127 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $$sink = $170;$$sink23 = $167;
 } else {
  $$sink = 0.0;$$sink23 = 0.0;
 }
 $175 = ((($this)) + 96|0); //@line 131 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$175>>2] = $$sink23; //@line 131 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $176 = ((($this)) + 76|0); //@line 132 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$176>>2] = $$sink; //@line 132 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $177 = ((($data)) + 20|0); //@line 135 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $178 = HEAP8[$177>>0]|0; //@line 135 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $179 = ($178<<24>>24)==(0); //@line 135 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 if ($179) {
  $213 = ((($this)) + 100|0); //@line 148 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  HEAPF32[$213>>2] = 0.0; //@line 148 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $vA$sroa$0$0 = $57;$vA$sroa$6$0 = $59;$vB$sroa$0$0 = $69;$vB$sroa$6$0 = $71;$wA$0 = $61;$wB$0 = $73;
  $214 = HEAP32[$4>>2]|0; //@line 151 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $215 = HEAP32[$54>>2]|0; //@line 151 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $216 = (($215) + (($214*12)|0)|0); //@line 151 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  HEAP32[$216>>2] = $vA$sroa$0$0; //@line 151 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $217 = (((($215) + (($214*12)|0)|0)) + 4|0); //@line 151 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  HEAP32[$217>>2] = $vA$sroa$6$0; //@line 151 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $218 = HEAP32[$4>>2]|0; //@line 152 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $219 = HEAP32[$54>>2]|0; //@line 152 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $220 = (((($219) + (($218*12)|0)|0)) + 8|0); //@line 152 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  HEAPF32[$220>>2] = $wA$0; //@line 152 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $221 = HEAP32[$9>>2]|0; //@line 153 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $222 = (($219) + (($221*12)|0)|0); //@line 153 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  HEAP32[$222>>2] = $vB$sroa$0$0; //@line 153 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $223 = (((($219) + (($221*12)|0)|0)) + 4|0); //@line 153 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  HEAP32[$223>>2] = $vB$sroa$6$0; //@line 153 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $224 = HEAP32[$9>>2]|0; //@line 154 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $225 = HEAP32[$54>>2]|0; //@line 154 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $226 = (((($225) + (($224*12)|0)|0)) + 8|0); //@line 154 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  HEAPF32[$226>>2] = $wB$0; //@line 154 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  return; //@line 155 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 } else {
  $180 = ((($data)) + 8|0); //@line 138 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $181 = +HEAPF32[$180>>2]; //@line 138 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $182 = ((($this)) + 100|0); //@line 138 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $183 = +HEAPF32[$182>>2]; //@line 138 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $184 = $181 * $183; //@line 138 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  HEAPF32[$182>>2] = $184; //@line 138 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $185 = $184 * $127; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $186 = $184 * $124; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $187 = $185 * $132; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $188 = $186 * $132; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $189 = (HEAP32[tempDoublePtr>>2]=$57,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $190 = $189 - $187; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $191 = (HEAPF32[tempDoublePtr>>2]=$190,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $192 = (HEAP32[tempDoublePtr>>2]=$59,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $193 = $192 - $188; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $194 = (HEAPF32[tempDoublePtr>>2]=$193,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $195 = $186 * $87; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $196 = $185 * $90; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $197 = $195 - $196; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $198 = $133 * $197; //@line 142 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $199 = $61 - $198; //@line 142 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $200 = $185 * $137; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $201 = $186 * $137; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $202 = (HEAP32[tempDoublePtr>>2]=$69,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $203 = $202 + $200; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $204 = (HEAPF32[tempDoublePtr>>2]=$203,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $205 = (HEAP32[tempDoublePtr>>2]=$71,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $206 = $205 + $201; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $207 = (HEAPF32[tempDoublePtr>>2]=$206,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $208 = $186 * $102; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $209 = $185 * $105; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $210 = $208 - $209; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $211 = $139 * $210; //@line 144 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $212 = $73 + $211; //@line 144 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $vA$sroa$0$0 = $191;$vA$sroa$6$0 = $194;$vB$sroa$0$0 = $204;$vB$sroa$6$0 = $207;$wA$0 = $199;$wB$0 = $212;
  $214 = HEAP32[$4>>2]|0; //@line 151 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $215 = HEAP32[$54>>2]|0; //@line 151 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $216 = (($215) + (($214*12)|0)|0); //@line 151 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  HEAP32[$216>>2] = $vA$sroa$0$0; //@line 151 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $217 = (((($215) + (($214*12)|0)|0)) + 4|0); //@line 151 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  HEAP32[$217>>2] = $vA$sroa$6$0; //@line 151 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $218 = HEAP32[$4>>2]|0; //@line 152 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $219 = HEAP32[$54>>2]|0; //@line 152 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $220 = (((($219) + (($218*12)|0)|0)) + 8|0); //@line 152 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  HEAPF32[$220>>2] = $wA$0; //@line 152 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $221 = HEAP32[$9>>2]|0; //@line 153 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $222 = (($219) + (($221*12)|0)|0); //@line 153 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  HEAP32[$222>>2] = $vB$sroa$0$0; //@line 153 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $223 = (((($219) + (($221*12)|0)|0)) + 4|0); //@line 153 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  HEAP32[$223>>2] = $vB$sroa$6$0; //@line 153 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $224 = HEAP32[$9>>2]|0; //@line 154 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $225 = HEAP32[$54>>2]|0; //@line 154 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  $226 = (((($225) + (($224*12)|0)|0)) + 8|0); //@line 154 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  HEAPF32[$226>>2] = $wB$0; //@line 154 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
  return; //@line 155 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 }
}
function __ZN15b2DistanceJoint24SolveVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0.0;
 var $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0, $49 = 0.0, $5 = 0.0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0;
 var $63 = 0.0, $64 = 0.0, $65 = 0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0.0, $70 = 0.0, $71 = 0.0, $72 = 0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0, $80 = 0.0;
 var $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0.0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 108|0); //@line 159 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 159 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $2 = ((($data)) + 28|0); //@line 159 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 159 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $4 = (($3) + (($1*12)|0)|0); //@line 159 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 159 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $6 = (((($3) + (($1*12)|0)|0)) + 4|0); //@line 159 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 159 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $8 = (((($3) + (($1*12)|0)|0)) + 8|0); //@line 160 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 160 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $10 = ((($this)) + 112|0); //@line 161 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 161 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $12 = (($3) + (($11*12)|0)|0); //@line 161 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $13 = +HEAPF32[$12>>2]; //@line 161 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $14 = (((($3) + (($11*12)|0)|0)) + 4|0); //@line 161 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 161 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $16 = (((($3) + (($11*12)|0)|0)) + 8|0); //@line 162 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 162 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $18 = ((($this)) + 128|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $19 = +HEAPF32[$18>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $20 = $9 * $19; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($this)) + 124|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $22 = +HEAPF32[$21>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $23 = $9 * $22; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $24 = $5 - $20; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $25 = $7 + $23; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $26 = ((($this)) + 136|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $27 = +HEAPF32[$26>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $28 = $17 * $27; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $29 = ((($this)) + 132|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $30 = +HEAPF32[$29>>2]; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $31 = $17 * $30; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $32 = $13 - $28; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $33 = $15 + $31; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $34 = $32 - $24; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $35 = $33 - $25; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $36 = ((($this)) + 116|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $37 = +HEAPF32[$36>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $38 = $37 * $34; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $39 = ((($this)) + 120|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $40 = +HEAPF32[$39>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $41 = $40 * $35; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $42 = $38 + $41; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $43 = ((($this)) + 172|0); //@line 169 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $44 = +HEAPF32[$43>>2]; //@line 169 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $45 = ((($this)) + 76|0); //@line 169 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $46 = +HEAPF32[$45>>2]; //@line 169 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $47 = $46 + $42; //@line 169 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $48 = ((($this)) + 96|0); //@line 169 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $49 = +HEAPF32[$48>>2]; //@line 169 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $50 = ((($this)) + 100|0); //@line 169 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $51 = +HEAPF32[$50>>2]; //@line 169 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $52 = $49 * $51; //@line 169 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $53 = $47 + $52; //@line 169 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $54 = $44 * $53; //@line 169 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $55 = -$54; //@line 169 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $56 = $51 - $54; //@line 170 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$50>>2] = $56; //@line 170 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $57 = $37 * $55; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $58 = $40 * $55; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $59 = ((($this)) + 156|0); //@line 173 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $60 = +HEAPF32[$59>>2]; //@line 173 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $61 = $57 * $60; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $62 = $58 * $60; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $63 = $5 - $61; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $64 = $7 - $62; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $65 = ((($this)) + 164|0); //@line 174 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $66 = +HEAPF32[$65>>2]; //@line 174 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $67 = $58 * $22; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $68 = $57 * $19; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $69 = $67 - $68; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $70 = $66 * $69; //@line 174 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $71 = $9 - $70; //@line 174 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $72 = ((($this)) + 160|0); //@line 175 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $73 = +HEAPF32[$72>>2]; //@line 175 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $74 = $57 * $73; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $75 = $58 * $73; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $76 = $13 + $74; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $77 = $15 + $75; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $78 = ((($this)) + 168|0); //@line 176 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $79 = +HEAPF32[$78>>2]; //@line 176 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $80 = $58 * $30; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $81 = $57 * $27; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $82 = $80 - $81; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $83 = $79 * $82; //@line 176 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $84 = $17 + $83; //@line 176 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$4>>2] = $63; //@line 178 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$6>>2] = $64; //@line 178 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $85 = HEAP32[$0>>2]|0; //@line 179 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $86 = HEAP32[$2>>2]|0; //@line 179 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $87 = (((($86) + (($85*12)|0)|0)) + 8|0); //@line 179 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$87>>2] = $71; //@line 179 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $88 = HEAP32[$10>>2]|0; //@line 180 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $89 = (($86) + (($88*12)|0)|0); //@line 180 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$89>>2] = $76; //@line 180 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $90 = (((($86) + (($88*12)|0)|0)) + 4|0); //@line 180 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$90>>2] = $77; //@line 180 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $91 = HEAP32[$10>>2]|0; //@line 181 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $92 = HEAP32[$2>>2]|0; //@line 181 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $93 = (((($92) + (($91*12)|0)|0)) + 8|0); //@line 181 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$93>>2] = $84; //@line 181 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 return; //@line 182 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
}
function __ZN15b2DistanceJoint24SolvePositionConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$0 = 0, $$0$i = 0.0, $0 = 0, $1 = 0.0, $10 = 0.0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0.0, $120 = 0.0, $121 = 0.0, $122 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0;
 var $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0.0, $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0;
 var $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0, $44 = 0.0, $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0;
 var $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0.0, $73 = 0.0, $74 = 0, $75 = 0.0, $76 = 0;
 var $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0.0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0;
 var $95 = 0.0, $96 = 0.0, $97 = 0, $98 = 0.0, $99 = 0.0, $u$sroa$0$0 = 0.0, $u$sroa$6$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 68|0); //@line 186 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 186 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $2 = $1 > 0.0; //@line 186 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 if ($2) {
  $$0 = 1;
  return ($$0|0); //@line 221 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 }
 $3 = ((($this)) + 108|0); //@line 192 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 215 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $5 = ((($data)) + 24|0); //@line 192 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 215 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $7 = (($6) + (($4*12)|0)|0); //@line 192 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $8 = +HEAPF32[$7>>2]; //@line 192 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $9 = (((($6) + (($4*12)|0)|0)) + 4|0); //@line 192 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $10 = +HEAPF32[$9>>2]; //@line 192 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $11 = (((($6) + (($4*12)|0)|0)) + 8|0); //@line 193 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $12 = +HEAPF32[$11>>2]; //@line 193 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $13 = ((($this)) + 112|0); //@line 194 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 194 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $15 = (($6) + (($14*12)|0)|0); //@line 194 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $16 = +HEAPF32[$15>>2]; //@line 194 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $17 = (((($6) + (($14*12)|0)|0)) + 4|0); //@line 194 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $18 = +HEAPF32[$17>>2]; //@line 194 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $19 = (((($6) + (($14*12)|0)|0)) + 8|0); //@line 195 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $20 = +HEAPF32[$19>>2]; //@line 195 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $21 = (+Math_sin((+$12))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $22 = (+Math_cos((+$12))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $23 = (+Math_sin((+$20))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $24 = (+Math_cos((+$20))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $25 = ((($this)) + 80|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $26 = +HEAPF32[$25>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($this)) + 140|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $28 = +HEAPF32[$27>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $29 = $26 - $28; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $30 = ((($this)) + 84|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $31 = +HEAPF32[$30>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $32 = ((($this)) + 144|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $33 = +HEAPF32[$32>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $34 = $31 - $33; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $35 = $22 * $29; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $36 = $21 * $34; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $37 = $35 - $36; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $38 = $21 * $29; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $39 = $22 * $34; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $40 = $38 + $39; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $41 = ((($this)) + 88|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $42 = +HEAPF32[$41>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $43 = ((($this)) + 148|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $44 = +HEAPF32[$43>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $45 = $42 - $44; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $46 = ((($this)) + 92|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $47 = +HEAPF32[$46>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $48 = ((($this)) + 152|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $49 = +HEAPF32[$48>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $50 = $47 - $49; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $51 = $24 * $45; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $52 = $23 * $50; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $53 = $51 - $52; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $54 = $23 * $45; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $55 = $24 * $50; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $56 = $54 + $55; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $57 = $16 + $53; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $58 = $18 + $56; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $59 = $57 - $8; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $60 = $58 - $10; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $61 = $59 - $37; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $62 = $60 - $40; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $63 = $61 * $61; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $64 = $62 * $62; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $65 = $63 + $64; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $66 = (+Math_sqrt((+$65))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $67 = $66 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
 if ($67) {
  $$0$i = 0.0;$u$sroa$0$0 = $61;$u$sroa$6$0 = $62;
 } else {
  $68 = 1.0 / $66; //@line 121 "./libraries/Box2D/Common/b2Math.h"
  $69 = $68 * $61; //@line 122 "./libraries/Box2D/Common/b2Math.h"
  $70 = $68 * $62; //@line 123 "./libraries/Box2D/Common/b2Math.h"
  $$0$i = $66;$u$sroa$0$0 = $69;$u$sroa$6$0 = $70;
 }
 $71 = ((($this)) + 104|0); //@line 204 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $72 = +HEAPF32[$71>>2]; //@line 204 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $73 = $$0$i - $72; //@line 204 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $74 = $73 < 0.20000000298023224; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $75 = $74 ? $73 : 0.20000000298023224; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $76 = $75 < -0.20000000298023224; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $77 = $76 ? -0.20000000298023224 : $75; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $78 = ((($this)) + 172|0); //@line 207 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $79 = +HEAPF32[$78>>2]; //@line 207 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $80 = $79 * $77; //@line 207 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $81 = -$80; //@line 207 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $82 = $u$sroa$0$0 * $81; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $83 = $u$sroa$6$0 * $81; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $84 = ((($this)) + 156|0); //@line 210 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $85 = +HEAPF32[$84>>2]; //@line 210 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $86 = $85 * $82; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $87 = $85 * $83; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $88 = $8 - $86; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $89 = $10 - $87; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $90 = ((($this)) + 164|0); //@line 211 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $91 = +HEAPF32[$90>>2]; //@line 211 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $92 = $37 * $83; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $93 = $40 * $82; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $94 = $92 - $93; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $95 = $91 * $94; //@line 211 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $96 = $12 - $95; //@line 211 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $97 = ((($this)) + 160|0); //@line 212 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $98 = +HEAPF32[$97>>2]; //@line 212 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $99 = $98 * $82; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $100 = $98 * $83; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $101 = $16 + $99; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $102 = $18 + $100; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $103 = ((($this)) + 168|0); //@line 213 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $104 = +HEAPF32[$103>>2]; //@line 213 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $105 = $53 * $83; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $106 = $56 * $82; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $107 = $105 - $106; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $108 = $104 * $107; //@line 213 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $109 = $20 + $108; //@line 213 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$7>>2] = $88; //@line 215 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$9>>2] = $89; //@line 215 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $110 = HEAP32[$3>>2]|0; //@line 216 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $111 = HEAP32[$5>>2]|0; //@line 216 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $112 = (((($111) + (($110*12)|0)|0)) + 8|0); //@line 216 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$112>>2] = $96; //@line 216 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $113 = HEAP32[$13>>2]|0; //@line 217 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $114 = (($111) + (($113*12)|0)|0); //@line 217 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$114>>2] = $101; //@line 217 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $115 = (((($111) + (($113*12)|0)|0)) + 4|0); //@line 217 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$115>>2] = $102; //@line 217 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $116 = HEAP32[$13>>2]|0; //@line 218 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $117 = HEAP32[$5>>2]|0; //@line 218 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $118 = (((($117) + (($116*12)|0)|0)) + 8|0); //@line 218 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$118>>2] = $109; //@line 218 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $119 = $77 > 0.0; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $120 = -$77; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $121 = $119 ? $77 : $120; //@line 618 "./libraries/Box2D/Common/b2Math.h"
 $122 = $121 < 0.004999999888241291; //@line 220 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $$0 = $122;
 return ($$0|0); //@line 221 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
}
function __ZN18b2DistanceJointDef10InitializeEP6b2BodyS1_RK6b2Vec2S4_($this,$b1,$b2,$anchor1,$anchor2) {
 $this = $this|0;
 $b1 = $b1|0;
 $b2 = $b2|0;
 $anchor1 = $anchor1|0;
 $anchor2 = $anchor2|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0;
 var $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0.0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0, $44 = 0.0;
 var $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 41 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$0>>2] = $b1; //@line 41 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $1 = ((($this)) + 12|0); //@line 42 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$1>>2] = $b2; //@line 42 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $2 = +HEAPF32[$anchor1>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($b1)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$3>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $5 = $2 - $4; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($anchor1)) + 4|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($b1)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $9 = +HEAPF32[$8>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $10 = $7 - $9; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($b1)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$11>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $13 = $5 * $12; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $14 = ((($b1)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $15 = +HEAPF32[$14>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $16 = $10 * $15; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $17 = $13 + $16; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $18 = $5 * $15; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $19 = $12 * $10; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $20 = $19 - $18; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($this)) + 20|0); //@line 43 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$21>>2] = $17; //@line 43 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $22 = ((($this)) + 24|0); //@line 43 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$22>>2] = $20; //@line 43 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $23 = +HEAPF32[$anchor2>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($b2)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $26 = $23 - $25; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $27 = ((($anchor2)) + 4|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $28 = +HEAPF32[$27>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $29 = ((($b2)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $30 = +HEAPF32[$29>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $31 = $28 - $30; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $32 = ((($b2)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $33 = +HEAPF32[$32>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $34 = $26 * $33; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $35 = ((($b2)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $36 = +HEAPF32[$35>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $37 = $31 * $36; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $38 = $34 + $37; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $39 = $26 * $36; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $40 = $33 * $31; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $41 = $40 - $39; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $42 = ((($this)) + 28|0); //@line 44 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$42>>2] = $38; //@line 44 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $43 = ((($this)) + 32|0); //@line 44 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$43>>2] = $41; //@line 44 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $44 = +HEAPF32[$anchor2>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $45 = +HEAPF32[$anchor1>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $46 = $44 - $45; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $47 = +HEAPF32[$27>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $48 = +HEAPF32[$6>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $49 = $47 - $48; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $50 = $46 * $46; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $51 = $49 * $49; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $52 = $50 + $51; //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $53 = (+Math_sqrt((+$52))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
 $54 = ((($this)) + 36|0); //@line 46 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$54>>2] = $53; //@line 46 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 return; //@line 47 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
}
function __ZN15b2DistanceJointC2EPK18b2DistanceJointDef($this,$def) {
 $this = $this|0;
 $def = $def|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN7b2JointC2EPK10b2JointDef($this,$def); //@line 50 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$this>>2] = (33224); //@line 51 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $0 = ((($this)) + 80|0); //@line 49 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $1 = ((($this)) + 88|0); //@line 49 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $2 = ((($def)) + 20|0); //@line 52 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $3 = $2; //@line 52 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $4 = $3; //@line 52 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 52 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $6 = (($3) + 4)|0; //@line 52 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $7 = $6; //@line 52 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 52 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $9 = $0; //@line 52 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $10 = $9; //@line 52 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$10>>2] = $5; //@line 52 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $11 = (($9) + 4)|0; //@line 52 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $12 = $11; //@line 52 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$12>>2] = $8; //@line 52 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $13 = ((($def)) + 28|0); //@line 53 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $14 = $13; //@line 53 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $15 = $14; //@line 53 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 53 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $17 = (($14) + 4)|0; //@line 53 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $18 = $17; //@line 53 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 53 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $20 = $1; //@line 53 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $21 = $20; //@line 53 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$21>>2] = $16; //@line 53 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $22 = (($20) + 4)|0; //@line 53 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $23 = $22; //@line 53 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$23>>2] = $19; //@line 53 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $24 = ((($def)) + 36|0); //@line 54 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $25 = HEAP32[$24>>2]|0; //@line 54 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $26 = ((($this)) + 104|0); //@line 54 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$26>>2] = $25; //@line 54 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $27 = ((($def)) + 40|0); //@line 55 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $28 = HEAP32[$27>>2]|0; //@line 55 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $29 = ((($this)) + 68|0); //@line 55 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$29>>2] = $28; //@line 55 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $30 = ((($def)) + 44|0); //@line 56 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 56 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $32 = ((($this)) + 72|0); //@line 56 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAP32[$32>>2] = $31; //@line 56 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $33 = ((($this)) + 100|0); //@line 57 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$33>>2] = 0.0; //@line 57 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $34 = ((($this)) + 96|0); //@line 58 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$34>>2] = 0.0; //@line 58 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 $35 = ((($this)) + 76|0); //@line 59 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 HEAPF32[$35>>2] = 0.0; //@line 59 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
 return; //@line 60 "libraries/Box2D/Dynamics/Joints/b2DistanceJoint.cpp"
}
function __ZNK15b2FrictionJoint10GetAnchorAEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 197 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 197 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 68|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 72|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 197 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
}
function __ZNK15b2FrictionJoint10GetAnchorBEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 202 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 202 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 76|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 80|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 202 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
}
function __ZNK15b2FrictionJoint16GetReactionForceEf($agg$result,$this,$inv_dt) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 84|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $1 = +HEAPF32[$0>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $2 = $1 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($this)) + 88|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$3>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $5 = $4 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $2; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$6>>2] = $5; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 207 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
}
function __ZNK15b2FrictionJoint17GetReactionTorqueEf($this,$inv_dt) {
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 92|0); //@line 212 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 212 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $2 = $1 * $inv_dt; //@line 212 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 return (+$2); //@line 212 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
}
function __ZN15b2FrictionJoint4DumpEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer13 = 0, $vararg_buffer17 = 0, $vararg_buffer20 = 0, $vararg_buffer23 = 0, $vararg_buffer3 = 0, $vararg_buffer6 = 0, $vararg_buffer9 = 0;
 var $vararg_ptr12 = 0, $vararg_ptr16 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer23 = sp + 80|0;
 $vararg_buffer20 = sp + 72|0;
 $vararg_buffer17 = sp + 64|0;
 $vararg_buffer13 = sp + 48|0;
 $vararg_buffer9 = sp + 32|0;
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($this)) + 48|0); //@line 239 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 239 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 239 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 239 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $4 = ((($this)) + 52|0); //@line 240 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 240 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $6 = ((($5)) + 8|0); //@line 240 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 240 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 __Z5b2LogPKcz(447080,$vararg_buffer); //@line 242 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$vararg_buffer1>>2] = $3; //@line 243 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 __Z5b2LogPKcz(447349,$vararg_buffer1); //@line 243 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$vararg_buffer3>>2] = $7; //@line 244 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 __Z5b2LogPKcz(447375,$vararg_buffer3); //@line 244 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $8 = ((($this)) + 61|0); //@line 245 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 245 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $10 = $9&255; //@line 245 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$vararg_buffer6>>2] = $10; //@line 245 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 __Z5b2LogPKcz(447401,$vararg_buffer6); //@line 245 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $11 = ((($this)) + 68|0); //@line 246 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $12 = +HEAPF32[$11>>2]; //@line 246 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $13 = $12; //@line 246 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $14 = ((($this)) + 72|0); //@line 246 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 246 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $16 = $15; //@line 246 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF64[$vararg_buffer9>>3] = $13; //@line 246 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $vararg_ptr12 = ((($vararg_buffer9)) + 8|0); //@line 246 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF64[$vararg_ptr12>>3] = $16; //@line 246 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 __Z5b2LogPKcz(447106,$vararg_buffer9); //@line 246 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $17 = ((($this)) + 76|0); //@line 247 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $18 = +HEAPF32[$17>>2]; //@line 247 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $19 = $18; //@line 247 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $20 = ((($this)) + 80|0); //@line 247 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 247 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $22 = $21; //@line 247 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF64[$vararg_buffer13>>3] = $19; //@line 247 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $vararg_ptr16 = ((($vararg_buffer13)) + 8|0); //@line 247 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF64[$vararg_ptr16>>3] = $22; //@line 247 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 __Z5b2LogPKcz(447148,$vararg_buffer13); //@line 247 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $23 = ((($this)) + 96|0); //@line 248 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $24 = +HEAPF32[$23>>2]; //@line 248 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $25 = $24; //@line 248 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF64[$vararg_buffer17>>3] = $25; //@line 248 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 __Z5b2LogPKcz(447190,$vararg_buffer17); //@line 248 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $26 = ((($this)) + 100|0); //@line 249 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $27 = +HEAPF32[$26>>2]; //@line 249 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $28 = $27; //@line 249 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF64[$vararg_buffer20>>3] = $28; //@line 249 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 __Z5b2LogPKcz(447216,$vararg_buffer20); //@line 249 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $29 = ((($this)) + 56|0); //@line 250 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $30 = HEAP32[$29>>2]|0; //@line 250 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$vararg_buffer23>>2] = $30; //@line 250 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 __Z5b2LogPKcz(447513,$vararg_buffer23); //@line 250 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 STACKTOP = sp;return; //@line 251 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
}
function __ZN15b2FrictionJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 55 "./libraries/Box2D/Dynamics/Joints/b2FrictionJoint.h"
 return; //@line 55 "./libraries/Box2D/Dynamics/Joints/b2FrictionJoint.h"
}
function __ZN15b2FrictionJoint23InitVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0;
 var $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0.0, $135 = 0, $136 = 0, $137 = 0.0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0, $147 = 0.0, $148 = 0.0, $149 = 0, $15 = 0, $150 = 0.0, $151 = 0.0;
 var $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0, $157 = 0.0, $158 = 0.0, $159 = 0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0.0, $17 = 0;
 var $170 = 0, $171 = 0.0, $172 = 0.0, $173 = 0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0.0, $66 = 0.0, $67 = 0.0;
 var $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0.0, $87 = 0.0, $88 = 0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0, $99 = 0, $bitcast = 0.0, $bitcast17 = 0.0, $det$0$i = 0.0, $storemerge = 0.0, $vA$sroa$0$0 = 0;
 var $vA$sroa$6$0 = 0, $vB$sroa$0$0 = 0, $vB$sroa$6$0 = 0, $wA$0 = 0.0, $wB$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 58 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 58 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 58 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 58 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $4 = ((($this)) + 104|0); //@line 58 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$4>>2] = $3; //@line 58 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $5 = ((($this)) + 52|0); //@line 59 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 59 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $7 = ((($6)) + 8|0); //@line 59 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 59 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $9 = ((($this)) + 108|0); //@line 59 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$9>>2] = $8; //@line 59 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $10 = ((($this)) + 128|0); //@line 60 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $11 = ((($1)) + 28|0); //@line 60 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $12 = $11; //@line 60 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $13 = $12; //@line 60 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 60 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $15 = (($12) + 4)|0; //@line 60 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $16 = $15; //@line 60 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 60 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $18 = $10; //@line 60 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $19 = $18; //@line 60 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$19>>2] = $14; //@line 60 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $20 = (($18) + 4)|0; //@line 60 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $21 = $20; //@line 60 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$21>>2] = $17; //@line 60 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $22 = ((($this)) + 136|0); //@line 61 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $23 = ((($6)) + 28|0); //@line 61 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $24 = $23; //@line 61 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $25 = $24; //@line 61 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 61 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $27 = (($24) + 4)|0; //@line 61 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $28 = $27; //@line 61 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 61 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $30 = $22; //@line 61 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $31 = $30; //@line 61 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$31>>2] = $26; //@line 61 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $32 = (($30) + 4)|0; //@line 61 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $33 = $32; //@line 61 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$33>>2] = $29; //@line 61 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $34 = ((($1)) + 120|0); //@line 62 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 62 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $36 = ((($this)) + 144|0); //@line 62 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$36>>2] = $35; //@line 62 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $37 = ((($6)) + 120|0); //@line 63 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $38 = HEAP32[$37>>2]|0; //@line 63 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $39 = ((($this)) + 148|0); //@line 63 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$39>>2] = $38; //@line 63 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $40 = ((($1)) + 128|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $41 = HEAP32[$40>>2]|0; //@line 64 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $42 = ((($this)) + 152|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$42>>2] = $41; //@line 64 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $43 = ((($6)) + 128|0); //@line 65 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $44 = HEAP32[$43>>2]|0; //@line 65 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $45 = ((($this)) + 156|0); //@line 65 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$45>>2] = $44; //@line 65 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $46 = ((($data)) + 24|0); //@line 67 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $47 = HEAP32[$46>>2]|0; //@line 67 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $48 = (((($47) + (($3*12)|0)|0)) + 8|0); //@line 67 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $49 = +HEAPF32[$48>>2]; //@line 67 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $50 = ((($data)) + 28|0); //@line 68 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $51 = HEAP32[$50>>2]|0; //@line 68 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $52 = (($51) + (($3*12)|0)|0); //@line 68 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $53 = HEAP32[$52>>2]|0; //@line 68 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $54 = (((($51) + (($3*12)|0)|0)) + 4|0); //@line 68 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $55 = HEAP32[$54>>2]|0; //@line 68 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $56 = (((($51) + (($3*12)|0)|0)) + 8|0); //@line 69 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $57 = +HEAPF32[$56>>2]; //@line 69 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $58 = (((($47) + (($8*12)|0)|0)) + 8|0); //@line 71 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $59 = +HEAPF32[$58>>2]; //@line 71 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $60 = (($51) + (($8*12)|0)|0); //@line 72 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $61 = HEAP32[$60>>2]|0; //@line 72 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $62 = (((($51) + (($8*12)|0)|0)) + 4|0); //@line 72 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $63 = HEAP32[$62>>2]|0; //@line 72 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $64 = (((($51) + (($8*12)|0)|0)) + 8|0); //@line 73 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $65 = +HEAPF32[$64>>2]; //@line 73 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $66 = (+Math_sin((+$49))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $67 = (+Math_cos((+$49))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $68 = (+Math_sin((+$59))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $69 = (+Math_cos((+$59))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $70 = ((($this)) + 68|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $71 = +HEAPF32[$70>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $bitcast = (HEAP32[tempDoublePtr>>2]=$14,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $72 = $71 - $bitcast; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $73 = ((($this)) + 72|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $74 = +HEAPF32[$73>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $75 = (HEAP32[tempDoublePtr>>2]=$17,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $76 = $74 - $75; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $77 = $67 * $72; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $78 = $66 * $76; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $79 = $77 - $78; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $80 = $66 * $72; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $81 = $67 * $76; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $82 = $80 + $81; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $83 = ((($this)) + 112|0); //@line 78 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$83>>2] = $79; //@line 78 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $84 = ((($this)) + 116|0); //@line 78 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$84>>2] = $82; //@line 78 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $85 = ((($this)) + 76|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $86 = +HEAPF32[$85>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $bitcast17 = (HEAP32[tempDoublePtr>>2]=$26,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $87 = $86 - $bitcast17; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $88 = ((($this)) + 80|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $89 = +HEAPF32[$88>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $90 = (HEAP32[tempDoublePtr>>2]=$29,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $91 = $89 - $90; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $92 = $69 * $87; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $93 = $68 * $91; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $94 = $92 - $93; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $95 = $68 * $87; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $96 = $69 * $91; //@line 568 "./libraries/Box2D/Common/b2Math.h"
 $97 = $95 + $96; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $98 = ((($this)) + 120|0); //@line 79 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$98>>2] = $94; //@line 79 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $99 = ((($this)) + 124|0); //@line 79 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$99>>2] = $97; //@line 79 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $100 = (HEAP32[tempDoublePtr>>2]=$35,+HEAPF32[tempDoublePtr>>2]); //@line 90 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $101 = (HEAP32[tempDoublePtr>>2]=$38,+HEAPF32[tempDoublePtr>>2]); //@line 90 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $102 = (HEAP32[tempDoublePtr>>2]=$41,+HEAPF32[tempDoublePtr>>2]); //@line 91 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $103 = (HEAP32[tempDoublePtr>>2]=$44,+HEAPF32[tempDoublePtr>>2]); //@line 91 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $104 = $100 + $101; //@line 94 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $105 = $102 * $82; //@line 94 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $106 = $82 * $105; //@line 94 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $107 = $104 + $106; //@line 94 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $108 = $103 * $97; //@line 94 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $109 = $97 * $108; //@line 94 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $110 = $107 + $109; //@line 94 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $111 = $102 * $79; //@line 95 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $112 = $82 * $111; //@line 95 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $113 = -$112; //@line 95 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $114 = $103 * $94; //@line 95 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $115 = $97 * $114; //@line 95 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $116 = $113 - $115; //@line 95 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $117 = $79 * $111; //@line 97 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $118 = $104 + $117; //@line 97 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $119 = $94 * $114; //@line 97 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $120 = $118 + $119; //@line 97 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $121 = $110 * $120; //@line 227 "./libraries/Box2D/Common/b2Math.h"
 $122 = $116 * $116; //@line 227 "./libraries/Box2D/Common/b2Math.h"
 $123 = $121 - $122; //@line 227 "./libraries/Box2D/Common/b2Math.h"
 $124 = $123 != 0.0; //@line 228 "./libraries/Box2D/Common/b2Math.h"
 $125 = 1.0 / $123; //@line 230 "./libraries/Box2D/Common/b2Math.h"
 $det$0$i = $124 ? $125 : $123; //@line 228 "./libraries/Box2D/Common/b2Math.h"
 $126 = $120 * $det$0$i; //@line 232 "./libraries/Box2D/Common/b2Math.h"
 $127 = $116 * $det$0$i; //@line 232 "./libraries/Box2D/Common/b2Math.h"
 $128 = -$127; //@line 232 "./libraries/Box2D/Common/b2Math.h"
 $129 = $110 * $det$0$i; //@line 233 "./libraries/Box2D/Common/b2Math.h"
 $130 = ((($this)) + 160|0); //@line 99 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$130>>2] = $126; //@line 99 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $131 = ((($this)) + 164|0); //@line 99 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$131>>2] = $128; //@line 99 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $132 = ((($this)) + 168|0); //@line 99 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$132>>2] = $128; //@line 99 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $133 = ((($this)) + 172|0); //@line 99 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$133>>2] = $129; //@line 99 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $134 = $102 + $103; //@line 101 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $135 = ((($this)) + 176|0); //@line 101 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $136 = $134 > 0.0; //@line 102 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $137 = 1.0 / $134; //@line 104 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $storemerge = $136 ? $137 : $134; //@line 102 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$135>>2] = $storemerge; //@line 101 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $138 = ((($data)) + 20|0); //@line 107 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $139 = HEAP8[$138>>0]|0; //@line 107 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $140 = ($139<<24>>24)==(0); //@line 107 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $141 = ((($this)) + 84|0); //@line 110 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 if ($140) {
  HEAPF32[$141>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $180 = ((($this)) + 88|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$180>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $181 = ((($this)) + 92|0); //@line 122 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  HEAPF32[$181>>2] = 0.0; //@line 122 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $vA$sroa$0$0 = $53;$vA$sroa$6$0 = $55;$vB$sroa$0$0 = $61;$vB$sroa$6$0 = $63;$wA$0 = $57;$wB$0 = $65;
  $182 = HEAP32[$4>>2]|0; //@line 125 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $183 = HEAP32[$50>>2]|0; //@line 125 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $184 = (($183) + (($182*12)|0)|0); //@line 125 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  HEAP32[$184>>2] = $vA$sroa$0$0; //@line 125 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $185 = (((($183) + (($182*12)|0)|0)) + 4|0); //@line 125 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  HEAP32[$185>>2] = $vA$sroa$6$0; //@line 125 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $186 = HEAP32[$4>>2]|0; //@line 126 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $187 = HEAP32[$50>>2]|0; //@line 126 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $188 = (((($187) + (($186*12)|0)|0)) + 8|0); //@line 126 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  HEAPF32[$188>>2] = $wA$0; //@line 126 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $189 = HEAP32[$9>>2]|0; //@line 127 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $190 = (($187) + (($189*12)|0)|0); //@line 127 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  HEAP32[$190>>2] = $vB$sroa$0$0; //@line 127 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $191 = (((($187) + (($189*12)|0)|0)) + 4|0); //@line 127 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  HEAP32[$191>>2] = $vB$sroa$6$0; //@line 127 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $192 = HEAP32[$9>>2]|0; //@line 128 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $193 = HEAP32[$50>>2]|0; //@line 128 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $194 = (((($193) + (($192*12)|0)|0)) + 8|0); //@line 128 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  HEAPF32[$194>>2] = $wB$0; //@line 128 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  return; //@line 129 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 } else {
  $142 = ((($data)) + 8|0); //@line 110 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $143 = +HEAPF32[$142>>2]; //@line 110 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $144 = +HEAPF32[$141>>2]; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $145 = $143 * $144; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$141>>2] = $145; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $146 = ((($this)) + 88|0); //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $147 = +HEAPF32[$146>>2]; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $148 = $143 * $147; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$146>>2] = $148; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $149 = ((($this)) + 92|0); //@line 111 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $150 = +HEAPF32[$149>>2]; //@line 111 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $151 = $143 * $150; //@line 111 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  HEAPF32[$149>>2] = $151; //@line 111 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $152 = $100 * $145; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $153 = $100 * $148; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $154 = (HEAP32[tempDoublePtr>>2]=$53,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $155 = $154 - $152; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $156 = (HEAPF32[tempDoublePtr>>2]=$155,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $157 = (HEAP32[tempDoublePtr>>2]=$55,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $158 = $157 - $153; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $159 = (HEAPF32[tempDoublePtr>>2]=$158,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $160 = $148 * $79; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $161 = $145 * $82; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $162 = $160 - $161; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $163 = $151 + $162; //@line 115 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $164 = $102 * $163; //@line 115 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $165 = $57 - $164; //@line 115 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $166 = $101 * $145; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $167 = $101 * $148; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $168 = (HEAP32[tempDoublePtr>>2]=$61,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $169 = $168 + $166; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $170 = (HEAPF32[tempDoublePtr>>2]=$169,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $171 = (HEAP32[tempDoublePtr>>2]=$63,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $172 = $171 + $167; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $173 = (HEAPF32[tempDoublePtr>>2]=$172,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $174 = $148 * $94; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $175 = $145 * $97; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $176 = $174 - $175; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $177 = $151 + $176; //@line 117 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $178 = $103 * $177; //@line 117 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $179 = $65 + $178; //@line 117 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $vA$sroa$0$0 = $156;$vA$sroa$6$0 = $159;$vB$sroa$0$0 = $170;$vB$sroa$6$0 = $173;$wA$0 = $165;$wB$0 = $179;
  $182 = HEAP32[$4>>2]|0; //@line 125 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $183 = HEAP32[$50>>2]|0; //@line 125 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $184 = (($183) + (($182*12)|0)|0); //@line 125 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  HEAP32[$184>>2] = $vA$sroa$0$0; //@line 125 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $185 = (((($183) + (($182*12)|0)|0)) + 4|0); //@line 125 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  HEAP32[$185>>2] = $vA$sroa$6$0; //@line 125 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $186 = HEAP32[$4>>2]|0; //@line 126 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $187 = HEAP32[$50>>2]|0; //@line 126 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $188 = (((($187) + (($186*12)|0)|0)) + 8|0); //@line 126 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  HEAPF32[$188>>2] = $wA$0; //@line 126 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $189 = HEAP32[$9>>2]|0; //@line 127 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $190 = (($187) + (($189*12)|0)|0); //@line 127 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  HEAP32[$190>>2] = $vB$sroa$0$0; //@line 127 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $191 = (((($187) + (($189*12)|0)|0)) + 4|0); //@line 127 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  HEAP32[$191>>2] = $vB$sroa$6$0; //@line 127 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $192 = HEAP32[$9>>2]|0; //@line 128 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $193 = HEAP32[$50>>2]|0; //@line 128 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  $194 = (((($193) + (($192*12)|0)|0)) + 8|0); //@line 128 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  HEAPF32[$194>>2] = $wB$0; //@line 128 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
  return; //@line 129 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 }
}
function __ZN15b2FrictionJoint24SolveVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0;
 var $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0.0, $130 = 0, $131 = 0, $132 = 0, $14 = 0;
 var $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0;
 var $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0.0, $50 = 0;
 var $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0, $66 = 0.0, $67 = 0.0, $68 = 0, $69 = 0.0;
 var $7 = 0.0, $70 = 0.0, $71 = 0.0, $72 = 0, $73 = 0.0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0, $86 = 0.0, $87 = 0.0;
 var $88 = 0.0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0.0, $92 = 0, $93 = 0.0, $94 = 0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 104|0); //@line 133 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 182 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $2 = ((($data)) + 28|0); //@line 133 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 182 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $4 = (($3) + (($1*12)|0)|0); //@line 133 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 133 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $6 = (((($3) + (($1*12)|0)|0)) + 4|0); //@line 133 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 133 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $8 = (((($3) + (($1*12)|0)|0)) + 8|0); //@line 134 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 134 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $10 = ((($this)) + 108|0); //@line 135 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 135 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $12 = (($3) + (($11*12)|0)|0); //@line 135 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $13 = +HEAPF32[$12>>2]; //@line 135 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $14 = (((($3) + (($11*12)|0)|0)) + 4|0); //@line 135 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 135 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $16 = (((($3) + (($11*12)|0)|0)) + 8|0); //@line 136 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 136 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $18 = ((($this)) + 144|0); //@line 138 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $19 = +HEAPF32[$18>>2]; //@line 138 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $20 = ((($this)) + 148|0); //@line 138 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 138 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $22 = ((($this)) + 152|0); //@line 139 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $23 = +HEAPF32[$22>>2]; //@line 139 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $24 = ((($this)) + 156|0); //@line 139 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $25 = +HEAPF32[$24>>2]; //@line 139 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $26 = +HEAPF32[$data>>2]; //@line 141 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $27 = $17 - $9; //@line 145 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $28 = ((($this)) + 176|0); //@line 146 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $29 = +HEAPF32[$28>>2]; //@line 146 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $30 = $27 * $29; //@line 146 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $31 = ((($this)) + 92|0); //@line 148 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $32 = +HEAPF32[$31>>2]; //@line 148 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $33 = ((($this)) + 100|0); //@line 149 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $34 = +HEAPF32[$33>>2]; //@line 149 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $35 = $26 * $34; //@line 149 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $36 = $32 - $30; //@line 150 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $37 = -$35; //@line 150 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $38 = $36 < $35; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $39 = $38 ? $36 : $35; //@line 634 "./libraries/Box2D/Common/b2Math.h"
 $40 = $39 < $37; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 $41 = $40 ? $37 : $39; //@line 645 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$31>>2] = $41; //@line 150 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $42 = $41 - $32; //@line 151 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $43 = $23 * $42; //@line 153 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $44 = $9 - $43; //@line 153 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $45 = $25 * $42; //@line 154 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $46 = $17 + $45; //@line 154 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $47 = ((($this)) + 124|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $48 = +HEAPF32[$47>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $49 = $48 * $46; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $50 = ((($this)) + 120|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $51 = +HEAPF32[$50>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $52 = $51 * $46; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $53 = $13 - $49; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $54 = $15 + $52; //@line 448 "./libraries/Box2D/Common/b2Math.h"
 $55 = $53 - $5; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $56 = $54 - $7; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $57 = ((($this)) + 116|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $58 = +HEAPF32[$57>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $59 = $58 * $44; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $60 = ((($this)) + 112|0); //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $61 = +HEAPF32[$60>>2]; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $62 = $61 * $44; //@line 428 "./libraries/Box2D/Common/b2Math.h"
 $63 = $59 + $55; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $64 = $56 - $62; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $65 = ((($this)) + 160|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $66 = +HEAPF32[$65>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $67 = $66 * $63; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $68 = ((($this)) + 168|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $69 = +HEAPF32[$68>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $70 = $69 * $64; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $71 = $67 + $70; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $72 = ((($this)) + 164|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $73 = +HEAPF32[$72>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $74 = $73 * $63; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $75 = ((($this)) + 172|0); //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $76 = +HEAPF32[$75>>2]; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $77 = $76 * $64; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $78 = $74 + $77; //@line 435 "./libraries/Box2D/Common/b2Math.h"
 $79 = ((($this)) + 84|0); //@line 162 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $80 = +HEAPF32[$79>>2]; //@line 162 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $81 = ((($this)) + 88|0); //@line 162 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $82 = +HEAPF32[$81>>2]; //@line 162 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $83 = $80 - $71; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$79>>2] = $83; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $84 = $82 - $78; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$81>>2] = $84; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $85 = ((($this)) + 96|0); //@line 165 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $86 = +HEAPF32[$85>>2]; //@line 165 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $87 = $26 * $86; //@line 165 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $88 = $83 * $83; //@line 110 "./libraries/Box2D/Common/b2Math.h"
 $89 = $84 * $84; //@line 110 "./libraries/Box2D/Common/b2Math.h"
 $90 = $88 + $89; //@line 110 "./libraries/Box2D/Common/b2Math.h"
 $91 = $87 * $87; //@line 167 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $92 = $90 > $91; //@line 167 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 if ($92) {
  $93 = (+Math_sqrt((+$90))); //@line 103 "./libraries/Box2D/Common/b2Math.h"
  $94 = $93 < 1.1920928955078125E-7; //@line 117 "./libraries/Box2D/Common/b2Math.h"
  if ($94) {
   $101 = $84;$99 = $83;
  } else {
   $95 = 1.0 / $93; //@line 121 "./libraries/Box2D/Common/b2Math.h"
   $96 = $83 * $95; //@line 122 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$79>>2] = $96; //@line 122 "./libraries/Box2D/Common/b2Math.h"
   $97 = $84 * $95; //@line 123 "./libraries/Box2D/Common/b2Math.h"
   HEAPF32[$81>>2] = $97; //@line 123 "./libraries/Box2D/Common/b2Math.h"
   $101 = $97;$99 = $96;
  }
  $98 = $87 * $99; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$79>>2] = $98; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $100 = $87 * $101; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$81>>2] = $100; //@line 97 "./libraries/Box2D/Common/b2Math.h"
  $103 = $98;$105 = $100;
 } else {
  $103 = $83;$105 = $84;
 }
 $102 = $103 - $80; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $104 = $105 - $82; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $106 = $19 * $102; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $107 = $19 * $104; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $108 = $5 - $106; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $109 = $7 - $107; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $110 = $61 * $104; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $111 = $102 * $58; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $112 = $110 - $111; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $113 = $23 * $112; //@line 176 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $114 = $44 - $113; //@line 176 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $115 = $21 * $102; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $116 = $21 * $104; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $117 = $13 + $115; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $118 = $15 + $116; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $119 = $104 * $51; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $120 = $102 * $48; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $121 = $119 - $120; //@line 414 "./libraries/Box2D/Common/b2Math.h"
 $122 = $25 * $121; //@line 179 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $123 = $46 + $122; //@line 179 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$4>>2] = $108; //@line 182 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$6>>2] = $109; //@line 182 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $124 = HEAP32[$0>>2]|0; //@line 183 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $125 = HEAP32[$2>>2]|0; //@line 183 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $126 = (((($125) + (($124*12)|0)|0)) + 8|0); //@line 183 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$126>>2] = $114; //@line 183 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $127 = HEAP32[$10>>2]|0; //@line 184 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $128 = (($125) + (($127*12)|0)|0); //@line 184 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$128>>2] = $117; //@line 184 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $129 = (((($125) + (($127*12)|0)|0)) + 4|0); //@line 184 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$129>>2] = $118; //@line 184 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $130 = HEAP32[$10>>2]|0; //@line 185 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $131 = HEAP32[$2>>2]|0; //@line 185 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $132 = (((($131) + (($130*12)|0)|0)) + 8|0); //@line 185 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$132>>2] = $123; //@line 185 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 return; //@line 186 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
}
function __ZN15b2FrictionJoint24SolvePositionConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1; //@line 192 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
}
function __ZN18b2FrictionJointDef10InitializeEP6b2BodyS1_RK6b2Vec2($this,$bA,$bB,$anchor) {
 $this = $this|0;
 $bA = $bA|0;
 $bB = $bB|0;
 $anchor = $anchor|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0;
 var $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0.0, $40 = 0.0, $41 = 0, $42 = 0, $5 = 0.0, $6 = 0;
 var $7 = 0.0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 37 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$0>>2] = $bA; //@line 37 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $1 = ((($this)) + 12|0); //@line 38 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$1>>2] = $bB; //@line 38 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $2 = +HEAPF32[$anchor>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $3 = ((($bA)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $4 = +HEAPF32[$3>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $5 = $2 - $4; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $6 = ((($anchor)) + 4|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $7 = +HEAPF32[$6>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $8 = ((($bA)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $9 = +HEAPF32[$8>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $10 = $7 - $9; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $11 = ((($bA)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $12 = +HEAPF32[$11>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $13 = $5 * $12; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $14 = ((($bA)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $15 = +HEAPF32[$14>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $16 = $10 * $15; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $17 = $13 + $16; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $18 = $5 * $15; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $19 = $12 * $10; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $20 = $19 - $18; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $21 = ((($this)) + 20|0); //@line 39 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$21>>2] = $17; //@line 39 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $22 = ((($this)) + 24|0); //@line 39 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$22>>2] = $20; //@line 39 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $23 = +HEAPF32[$anchor>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $24 = ((($bB)) + 12|0); //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $25 = +HEAPF32[$24>>2]; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $26 = $23 - $25; //@line 587 "./libraries/Box2D/Common/b2Math.h"
 $27 = +HEAPF32[$6>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $28 = ((($bB)) + 16|0); //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $29 = +HEAPF32[$28>>2]; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $30 = $27 - $29; //@line 588 "./libraries/Box2D/Common/b2Math.h"
 $31 = ((($bB)) + 24|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $32 = +HEAPF32[$31>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $33 = $26 * $32; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $34 = ((($bB)) + 20|0); //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $35 = +HEAPF32[$34>>2]; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $36 = $30 * $35; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $37 = $33 + $36; //@line 589 "./libraries/Box2D/Common/b2Math.h"
 $38 = $26 * $35; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $39 = $32 * $30; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $40 = $39 - $38; //@line 590 "./libraries/Box2D/Common/b2Math.h"
 $41 = ((($this)) + 28|0); //@line 40 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$41>>2] = $37; //@line 40 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $42 = ((($this)) + 32|0); //@line 40 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$42>>2] = $40; //@line 40 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 return; //@line 41 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
}
function __ZN15b2FrictionJointC2EPK18b2FrictionJointDef($this,$def) {
 $this = $this|0;
 $def = $def|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN7b2JointC2EPK10b2JointDef($this,$def); //@line 44 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$this>>2] = (33276); //@line 45 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $0 = ((($this)) + 68|0); //@line 43 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $1 = ((($this)) + 76|0); //@line 43 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $2 = ((($def)) + 20|0); //@line 46 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $3 = $2; //@line 46 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $4 = $3; //@line 46 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 46 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $6 = (($3) + 4)|0; //@line 46 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $7 = $6; //@line 46 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 46 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $9 = $0; //@line 46 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $10 = $9; //@line 46 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$10>>2] = $5; //@line 46 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $11 = (($9) + 4)|0; //@line 46 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $12 = $11; //@line 46 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$12>>2] = $8; //@line 46 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $13 = ((($def)) + 28|0); //@line 47 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $14 = $13; //@line 47 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $15 = $14; //@line 47 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 47 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $17 = (($14) + 4)|0; //@line 47 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $18 = $17; //@line 47 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 47 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $20 = $1; //@line 47 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $21 = $20; //@line 47 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$21>>2] = $16; //@line 47 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $22 = (($20) + 4)|0; //@line 47 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $23 = $22; //@line 47 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$23>>2] = $19; //@line 47 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $24 = ((($this)) + 84|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$24>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $25 = ((($this)) + 88|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$25>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
 $26 = ((($this)) + 92|0); //@line 50 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$26>>2] = 0.0; //@line 50 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $27 = ((($def)) + 36|0); //@line 52 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $28 = HEAP32[$27>>2]|0; //@line 52 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $29 = ((($this)) + 96|0); //@line 52 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$29>>2] = $28; //@line 52 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $30 = ((($def)) + 40|0); //@line 53 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 53 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $32 = ((($this)) + 100|0); //@line 53 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAP32[$32>>2] = $31; //@line 53 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 return; //@line 54 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
}
function __ZN15b2FrictionJoint11SetMaxForceEf($this,$force) {
 $this = $this|0;
 $force = +$force;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (HEAPF32[tempDoublePtr>>2]=$force,HEAP32[tempDoublePtr>>2]|0); //@line 217 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $1 = $0 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $2 = ($1|0)!=(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $3 = $force >= 0.0; //@line 217 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $4 = $3 & $2; //@line 217 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 __Z10loveAssertbPKc($4,447243); //@line 217 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $5 = ((($this)) + 96|0); //@line 218 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$5>>2] = $force; //@line 218 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 return; //@line 219 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
}
function __ZNK15b2FrictionJoint11GetMaxForceEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 96|0); //@line 223 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 223 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 return (+$1); //@line 223 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
}
function __ZN15b2FrictionJoint12SetMaxTorqueEf($this,$torque) {
 $this = $this|0;
 $torque = +$torque;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (HEAPF32[tempDoublePtr>>2]=$torque,HEAP32[tempDoublePtr>>2]|0); //@line 228 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $1 = $0 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $2 = ($1|0)!=(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $3 = $torque >= 0.0; //@line 228 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $4 = $3 & $2; //@line 228 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 __Z10loveAssertbPKc($4,447277); //@line 228 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $5 = ((($this)) + 100|0); //@line 229 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 HEAPF32[$5>>2] = $torque; //@line 229 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 return; //@line 230 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
}
function __ZNK15b2FrictionJoint12GetMaxTorqueEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 100|0); //@line 234 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 234 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
 return (+$1); //@line 234 "libraries/Box2D/Dynamics/Joints/b2FrictionJoint.cpp"
}
function __ZNK11b2GearJoint10GetAnchorAEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 372 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 372 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 92|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 96|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 372 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
}
function __ZNK11b2GearJoint10GetAnchorBEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $3 = 0.0, $4 = 0, $5 = 0.0, $6 = 0.0;
 var $7 = 0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0); //@line 377 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 377 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $2 = ((($1)) + 24|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $4 = ((($this)) + 100|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $5 = +HEAPF32[$4>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $6 = $3 * $5; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $7 = ((($1)) + 20|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $8 = +HEAPF32[$7>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $9 = ((($this)) + 104|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $10 = +HEAPF32[$9>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $11 = $8 * $10; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $12 = $6 - $11; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $13 = ((($1)) + 12|0); //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $14 = +HEAPF32[$13>>2]; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $15 = $14 + $12; //@line 579 "./libraries/Box2D/Common/b2Math.h"
 $16 = $5 * $8; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $17 = $3 * $10; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $18 = $16 + $17; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $19 = ((($1)) + 16|0); //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $20 = +HEAPF32[$19>>2]; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 $21 = $18 + $20; //@line 580 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $15; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $22 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$22>>2] = $21; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 377 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
}
function __ZNK11b2GearJoint16GetReactionForceEf($agg$result,$this,$inv_dt) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $10 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 156|0); //@line 382 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 382 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $2 = ((($this)) + 240|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $3 = +HEAPF32[$2>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $4 = $1 * $3; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $5 = ((($this)) + 244|0); //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $6 = +HEAPF32[$5>>2]; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $7 = $1 * $6; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $8 = $4 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $9 = $7 * $inv_dt; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$agg$result>>2] = $8; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 $10 = ((($agg$result)) + 4|0); //@line 59 "./libraries/Box2D/Common/b2Math.h"
 HEAPF32[$10>>2] = $9; //@line 59 "./libraries/Box2D/Common/b2Math.h"
 return; //@line 384 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
}
function __ZNK11b2GearJoint17GetReactionTorqueEf($this,$inv_dt) {
 $this = $this|0;
 $inv_dt = +$inv_dt;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 156|0); //@line 388 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 388 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $2 = ((($this)) + 256|0); //@line 388 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 388 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $4 = $1 * $3; //@line 388 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $5 = $4 * $inv_dt; //@line 389 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 return (+$5); //@line 389 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
}
function __ZN11b2GearJoint4DumpEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer12 = 0, $vararg_buffer15 = 0, $vararg_buffer18 = 0, $vararg_buffer3 = 0, $vararg_buffer6 = 0, $vararg_buffer9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer18 = sp + 56|0;
 $vararg_buffer15 = sp + 48|0;
 $vararg_buffer12 = sp + 40|0;
 $vararg_buffer9 = sp + 32|0;
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($this)) + 48|0); //@line 405 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 405 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 405 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 405 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $4 = ((($this)) + 52|0); //@line 406 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 406 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $6 = ((($5)) + 8|0); //@line 406 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 406 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $8 = ((($this)) + 68|0); //@line 408 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 408 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $10 = ((($9)) + 56|0); //@line 408 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 408 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $12 = ((($this)) + 72|0); //@line 409 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 409 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $14 = ((($13)) + 56|0); //@line 409 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 409 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 __Z5b2LogPKcz(447327,$vararg_buffer); //@line 411 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$vararg_buffer1>>2] = $3; //@line 412 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 __Z5b2LogPKcz(447349,$vararg_buffer1); //@line 412 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$vararg_buffer3>>2] = $7; //@line 413 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 __Z5b2LogPKcz(447375,$vararg_buffer3); //@line 413 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $16 = ((($this)) + 61|0); //@line 414 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $17 = HEAP8[$16>>0]|0; //@line 414 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $18 = $17&255; //@line 414 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$vararg_buffer6>>2] = $18; //@line 414 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 __Z5b2LogPKcz(447401,$vararg_buffer6); //@line 414 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$vararg_buffer9>>2] = $11; //@line 415 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 __Z5b2LogPKcz(447436,$vararg_buffer9); //@line 415 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$vararg_buffer12>>2] = $15; //@line 416 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 __Z5b2LogPKcz(447463,$vararg_buffer12); //@line 416 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $19 = ((($this)) + 152|0); //@line 417 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $20 = +HEAPF32[$19>>2]; //@line 417 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $21 = $20; //@line 417 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF64[$vararg_buffer15>>3] = $21; //@line 417 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 __Z5b2LogPKcz(447490,$vararg_buffer15); //@line 417 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $22 = ((($this)) + 56|0); //@line 418 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $23 = HEAP32[$22>>2]|0; //@line 418 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$vararg_buffer18>>2] = $23; //@line 418 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 __Z5b2LogPKcz(447513,$vararg_buffer18); //@line 418 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 STACKTOP = sp;return; //@line 419 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
}
function __ZN11b2GearJointD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 56 "./libraries/Box2D/Dynamics/Joints/b2GearJoint.h"
 return; //@line 56 "./libraries/Box2D/Dynamics/Joints/b2GearJoint.h"
}
function __ZN11b2GearJoint23InitVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$in = 0.0, $$pn = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0.0, $112 = 0, $113 = 0;
 var $114 = 0.0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0, $123 = 0.0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0.0;
 var $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0.0, $148 = 0, $149 = 0.0, $15 = 0;
 var $150 = 0, $151 = 0.0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0, $165 = 0.0, $166 = 0, $167 = 0.0, $168 = 0.0;
 var $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0, $18 = 0, $180 = 0.0, $181 = 0, $182 = 0.0, $183 = 0.0, $184 = 0.0, $185 = 0.0, $186 = 0.0;
 var $187 = 0.0, $188 = 0.0, $189 = 0.0, $19 = 0, $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0.0, $196 = 0.0, $197 = 0.0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0.0, $201 = 0.0, $202 = 0.0, $203 = 0.0;
 var $204 = 0.0, $205 = 0.0, $206 = 0, $207 = 0.0, $208 = 0, $209 = 0.0, $21 = 0, $210 = 0, $211 = 0.0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0.0, $22 = 0, $220 = 0.0, $221 = 0.0;
 var $222 = 0.0, $223 = 0, $224 = 0.0, $225 = 0.0, $226 = 0, $227 = 0.0, $228 = 0.0, $229 = 0.0, $23 = 0, $230 = 0.0, $231 = 0.0, $232 = 0.0, $233 = 0, $234 = 0.0, $235 = 0.0, $236 = 0.0, $237 = 0, $238 = 0.0, $239 = 0, $24 = 0;
 var $240 = 0.0, $241 = 0.0, $242 = 0.0, $243 = 0.0, $244 = 0.0, $245 = 0.0, $246 = 0.0, $247 = 0.0, $248 = 0, $249 = 0.0, $25 = 0, $250 = 0.0, $251 = 0.0, $252 = 0, $253 = 0.0, $254 = 0, $255 = 0.0, $256 = 0.0, $257 = 0.0, $258 = 0.0;
 var $259 = 0.0, $26 = 0, $260 = 0.0, $261 = 0.0, $262 = 0.0, $263 = 0, $264 = 0.0, $265 = 0.0, $266 = 0.0, $267 = 0.0, $268 = 0.0, $269 = 0.0, $27 = 0, $270 = 0.0, $271 = 0, $272 = 0.0, $273 = 0.0, $274 = 0.0, $275 = 0.0, $276 = 0;
 var $277 = 0.0, $278 = 0.0, $279 = 0.0, $28 = 0, $280 = 0.0, $281 = 0.0, $282 = 0.0, $283 = 0.0, $284 = 0.0, $285 = 0.0, $286 = 0.0, $287 = 0, $288 = 0.0, $289 = 0, $29 = 0, $290 = 0.0, $291 = 0, $292 = 0.0, $293 = 0.0, $294 = 0;
 var $295 = 0, $296 = 0, $297 = 0.0, $298 = 0, $299 = 0.0, $3 = 0, $30 = 0, $300 = 0.0, $301 = 0.0, $302 = 0.0, $303 = 0.0, $304 = 0.0, $305 = 0, $306 = 0.0, $307 = 0.0, $308 = 0, $309 = 0.0, $31 = 0, $310 = 0.0, $311 = 0.0;
 var $312 = 0.0, $313 = 0.0, $314 = 0.0, $315 = 0.0, $316 = 0.0, $317 = 0.0, $318 = 0, $319 = 0.0, $32 = 0, $320 = 0.0, $321 = 0, $322 = 0.0, $323 = 0.0, $324 = 0.0, $325 = 0.0, $326 = 0.0, $327 = 0.0, $328 = 0.0, $329 = 0.0, $33 = 0;
 var $330 = 0.0, $331 = 0, $332 = 0.0, $333 = 0.0, $334 = 0, $335 = 0.0, $336 = 0.0, $337 = 0.0, $338 = 0.0, $339 = 0.0, $34 = 0, $340 = 0.0, $341 = 0.0, $342 = 0.0, $343 = 0.0, $344 = 0, $345 = 0.0, $346 = 0.0, $347 = 0, $348 = 0.0;
 var $349 = 0.0, $35 = 0, $350 = 0.0, $351 = 0.0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0;
 var $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0.0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $storemerge22 = 0.0, $vA$sroa$0$0 = 0, $vA$sroa$6$0 = 0, $vB$sroa$0$0 = 0, $vB$sroa$6$0 = 0, $vC$sroa$0$0 = 0, $vC$sroa$6$0 = 0, $vD$sroa$0$0 = 0, $vD$sroa$6$0 = 0, $wA$0 = 0.0, $wB$0 = 0.0, $wC$0 = 0.0, $wD$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0); //@line 133 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 133 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $2 = ((($1)) + 8|0); //@line 133 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 133 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $4 = ((($this)) + 160|0); //@line 133 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$4>>2] = $3; //@line 133 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $5 = ((($this)) + 52|0); //@line 134 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 134 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $7 = ((($6)) + 8|0); //@line 134 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 134 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $9 = ((($this)) + 164|0); //@line 134 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$9>>2] = $8; //@line 134 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $10 = ((($this)) + 84|0); //@line 135 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 135 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $12 = ((($11)) + 8|0); //@line 135 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 135 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $14 = ((($this)) + 168|0); //@line 135 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$14>>2] = $13; //@line 135 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $15 = ((($this)) + 88|0); //@line 136 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 136 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $17 = ((($16)) + 8|0); //@line 136 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 136 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $19 = ((($this)) + 172|0); //@line 136 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$19>>2] = $18; //@line 136 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $20 = ((($this)) + 176|0); //@line 137 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $21 = ((($1)) + 28|0); //@line 137 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $22 = $21; //@line 137 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $23 = $22; //@line 137 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 137 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $25 = (($22) + 4)|0; //@line 137 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $26 = $25; //@line 137 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $27 = HEAP32[$26>>2]|0; //@line 137 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $28 = $20; //@line 137 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $29 = $28; //@line 137 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$29>>2] = $24; //@line 137 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $30 = (($28) + 4)|0; //@line 137 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $31 = $30; //@line 137 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$31>>2] = $27; //@line 137 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $32 = ((($this)) + 184|0); //@line 138 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $33 = ((($6)) + 28|0); //@line 138 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $34 = $33; //@line 138 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $35 = $34; //@line 138 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $36 = HEAP32[$35>>2]|0; //@line 138 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $37 = (($34) + 4)|0; //@line 138 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $38 = $37; //@line 138 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $39 = HEAP32[$38>>2]|0; //@line 138 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $40 = $32; //@line 138 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $41 = $40; //@line 138 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$41>>2] = $36; //@line 138 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $42 = (($40) + 4)|0; //@line 138 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $43 = $42; //@line 138 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$43>>2] = $39; //@line 138 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $44 = ((($this)) + 192|0); //@line 139 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $45 = ((($11)) + 28|0); //@line 139 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $46 = $45; //@line 139 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $47 = $46; //@line 139 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $48 = HEAP32[$47>>2]|0; //@line 139 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $49 = (($46) + 4)|0; //@line 139 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $50 = $49; //@line 139 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $51 = HEAP32[$50>>2]|0; //@line 139 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $52 = $44; //@line 139 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $53 = $52; //@line 139 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$53>>2] = $48; //@line 139 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $54 = (($52) + 4)|0; //@line 139 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $55 = $54; //@line 139 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$55>>2] = $51; //@line 139 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $56 = ((($this)) + 200|0); //@line 140 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $57 = ((($16)) + 28|0); //@line 140 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $58 = $57; //@line 140 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $59 = $58; //@line 140 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $60 = HEAP32[$59>>2]|0; //@line 140 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $61 = (($58) + 4)|0; //@line 140 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $62 = $61; //@line 140 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $63 = HEAP32[$62>>2]|0; //@line 140 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $64 = $56; //@line 140 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $65 = $64; //@line 140 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$65>>2] = $60; //@line 140 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $66 = (($64) + 4)|0; //@line 140 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $67 = $66; //@line 140 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$67>>2] = $63; //@line 140 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $68 = ((($1)) + 120|0); //@line 141 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $69 = HEAP32[$68>>2]|0; //@line 141 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $70 = ((($this)) + 208|0); //@line 141 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$70>>2] = $69; //@line 141 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $71 = ((($6)) + 120|0); //@line 142 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $72 = HEAP32[$71>>2]|0; //@line 142 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $73 = ((($this)) + 212|0); //@line 142 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$73>>2] = $72; //@line 142 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $74 = ((($11)) + 120|0); //@line 143 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $75 = HEAP32[$74>>2]|0; //@line 143 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $76 = ((($this)) + 216|0); //@line 143 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$76>>2] = $75; //@line 143 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $77 = ((($16)) + 120|0); //@line 144 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $78 = HEAP32[$77>>2]|0; //@line 144 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $79 = ((($this)) + 220|0); //@line 144 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$79>>2] = $78; //@line 144 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $80 = ((($1)) + 128|0); //@line 145 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $81 = HEAP32[$80>>2]|0; //@line 145 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $82 = ((($this)) + 224|0); //@line 145 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$82>>2] = $81; //@line 145 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $83 = ((($6)) + 128|0); //@line 146 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $84 = HEAP32[$83>>2]|0; //@line 146 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $85 = ((($this)) + 228|0); //@line 146 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$85>>2] = $84; //@line 146 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $86 = ((($11)) + 128|0); //@line 147 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $87 = HEAP32[$86>>2]|0; //@line 147 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $88 = ((($this)) + 232|0); //@line 147 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$88>>2] = $87; //@line 147 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $89 = ((($16)) + 128|0); //@line 148 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $90 = HEAP32[$89>>2]|0; //@line 148 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $91 = ((($this)) + 236|0); //@line 148 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$91>>2] = $90; //@line 148 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $92 = ((($data)) + 24|0); //@line 150 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $93 = HEAP32[$92>>2]|0; //@line 150 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $94 = (((($93) + (($3*12)|0)|0)) + 8|0); //@line 150 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $95 = +HEAPF32[$94>>2]; //@line 150 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $96 = ((($data)) + 28|0); //@line 151 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $97 = HEAP32[$96>>2]|0; //@line 151 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $98 = (($97) + (($3*12)|0)|0); //@line 151 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $99 = HEAP32[$98>>2]|0; //@line 151 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $100 = (((($97) + (($3*12)|0)|0)) + 4|0); //@line 151 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $101 = HEAP32[$100>>2]|0; //@line 151 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $102 = (((($97) + (($3*12)|0)|0)) + 8|0); //@line 152 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $103 = +HEAPF32[$102>>2]; //@line 152 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $104 = (((($93) + (($8*12)|0)|0)) + 8|0); //@line 154 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $105 = +HEAPF32[$104>>2]; //@line 154 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $106 = (($97) + (($8*12)|0)|0); //@line 155 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $107 = HEAP32[$106>>2]|0; //@line 155 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $108 = (((($97) + (($8*12)|0)|0)) + 4|0); //@line 155 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $109 = HEAP32[$108>>2]|0; //@line 155 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $110 = (((($97) + (($8*12)|0)|0)) + 8|0); //@line 156 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $111 = +HEAPF32[$110>>2]; //@line 156 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $112 = HEAP32[$14>>2]|0; //@line 158 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $113 = (((($93) + (($112*12)|0)|0)) + 8|0); //@line 158 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $114 = +HEAPF32[$113>>2]; //@line 158 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $115 = (($97) + (($112*12)|0)|0); //@line 159 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $116 = HEAP32[$115>>2]|0; //@line 159 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $117 = (((($97) + (($112*12)|0)|0)) + 4|0); //@line 159 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $118 = HEAP32[$117>>2]|0; //@line 159 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $119 = (((($97) + (($112*12)|0)|0)) + 8|0); //@line 160 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $120 = +HEAPF32[$119>>2]; //@line 160 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $121 = HEAP32[$19>>2]|0; //@line 162 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $122 = (((($93) + (($121*12)|0)|0)) + 8|0); //@line 162 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $123 = +HEAPF32[$122>>2]; //@line 162 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $124 = (($97) + (($121*12)|0)|0); //@line 163 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $125 = HEAP32[$124>>2]|0; //@line 163 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $126 = (((($97) + (($121*12)|0)|0)) + 4|0); //@line 163 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $127 = HEAP32[$126>>2]|0; //@line 163 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $128 = (((($97) + (($121*12)|0)|0)) + 8|0); //@line 164 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $129 = +HEAPF32[$128>>2]; //@line 164 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $130 = (+Math_sin((+$95))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $131 = (+Math_cos((+$95))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $132 = (+Math_sin((+$105))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $133 = (+Math_cos((+$105))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $134 = (+Math_sin((+$114))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $135 = (+Math_cos((+$114))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $136 = (+Math_sin((+$123))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $137 = (+Math_cos((+$123))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $138 = ((($this)) + 272|0); //@line 168 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$138>>2] = 0.0; //@line 168 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $139 = ((($this)) + 76|0); //@line 170 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $140 = HEAP32[$139>>2]|0; //@line 170 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $141 = ($140|0)==(1); //@line 170 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $142 = (HEAP32[tempDoublePtr>>2]=$75,+HEAPF32[tempDoublePtr>>2]); //@line 185 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $143 = (HEAP32[tempDoublePtr>>2]=$87,+HEAPF32[tempDoublePtr>>2]); //@line 185 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $144 = (HEAP32[tempDoublePtr>>2]=$81,+HEAPF32[tempDoublePtr>>2]); //@line 185 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $145 = (HEAP32[tempDoublePtr>>2]=$78,+HEAPF32[tempDoublePtr>>2]); //@line 203 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $146 = (HEAP32[tempDoublePtr>>2]=$90,+HEAPF32[tempDoublePtr>>2]); //@line 203 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $147 = (HEAP32[tempDoublePtr>>2]=$84,+HEAPF32[tempDoublePtr>>2]); //@line 203 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 if ($141) {
  $148 = ((($this)) + 264|0); //@line 174 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$148>>2] = 1.0; //@line 174 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $149 = $144 + $143; //@line 175 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $$in = $149;$207 = 0.0;$209 = 0.0;$211 = 1.0;$337 = 1.0;
 } else {
  $150 = ((($this)) + 124|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $151 = +HEAPF32[$150>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $152 = $135 * $151; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $153 = ((($this)) + 128|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $154 = +HEAPF32[$153>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $155 = $134 * $154; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $156 = $152 - $155; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $157 = $134 * $151; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $158 = $135 * $154; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $159 = $157 + $158; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $160 = ((($this)) + 108|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $161 = +HEAPF32[$160>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $162 = +HEAPF32[$44>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $163 = $161 - $162; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $164 = ((($this)) + 112|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $165 = +HEAPF32[$164>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $166 = ((($this)) + 196|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $167 = +HEAPF32[$166>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $168 = $165 - $167; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $169 = $135 * $163; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $170 = $134 * $168; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $171 = $169 - $170; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $172 = $134 * $163; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $173 = $135 * $168; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $174 = $172 + $173; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $175 = ((($this)) + 92|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $176 = +HEAPF32[$175>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $177 = +HEAPF32[$20>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $178 = $176 - $177; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $179 = ((($this)) + 96|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $180 = +HEAPF32[$179>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $181 = ((($this)) + 180|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $182 = +HEAPF32[$181>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $183 = $180 - $182; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $184 = $131 * $178; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $185 = $130 * $183; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $186 = $184 - $185; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $187 = $130 * $178; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $188 = $131 * $183; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $189 = $187 + $188; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $190 = $159 * $171; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $191 = $156 * $174; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $192 = $190 - $191; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $193 = ((($this)) + 264|0); //@line 183 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$193>>2] = $192; //@line 183 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $194 = $159 * $186; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $195 = $156 * $189; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $196 = $194 - $195; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $197 = +HEAPF32[$70>>2]; //@line 185 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $198 = $142 + $197; //@line 185 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $199 = $143 * $192; //@line 185 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $200 = $192 * $199; //@line 185 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $201 = $198 + $200; //@line 185 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $202 = $196 * $144; //@line 185 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $203 = $196 * $202; //@line 185 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $204 = $201 + $203; //@line 185 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $$in = $204;$207 = $156;$209 = $159;$211 = $196;$337 = $192;
 }
 $205 = $$in + 0.0; //@line 193 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $206 = ((($this)) + 240|0); //@line 182 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$206>>2] = $207; //@line 182 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $208 = ((($this)) + 244|0); //@line 182 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$208>>2] = $209; //@line 182 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $210 = ((($this)) + 256|0); //@line 184 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$210>>2] = $211; //@line 184 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$138>>2] = $205; //@line 175 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $212 = ((($this)) + 80|0); //@line 188 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $213 = HEAP32[$212>>2]|0; //@line 188 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $214 = ($213|0)==(1); //@line 188 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 if ($214) {
  $215 = ((($this)) + 152|0); //@line 191 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $216 = HEAP32[$215>>2]|0; //@line 191 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $217 = ((($this)) + 260|0); //@line 191 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$217>>2] = $216; //@line 191 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $218 = ((($this)) + 268|0); //@line 192 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$218>>2] = $216; //@line 192 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $219 = (HEAP32[tempDoublePtr>>2]=$216,+HEAPF32[tempDoublePtr>>2]); //@line 193 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $220 = $219 * $219; //@line 193 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $221 = $147 + $146; //@line 193 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $222 = $220 * $221; //@line 193 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $$pn = $222;$288 = 0.0;$290 = 0.0;$324 = $219;$350 = $219;
 } else {
  $223 = ((($this)) + 132|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $224 = +HEAPF32[$223>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $225 = $137 * $224; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $226 = ((($this)) + 136|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $227 = +HEAPF32[$226>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $228 = $136 * $227; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $229 = $225 - $228; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $230 = $136 * $224; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $231 = $137 * $227; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $232 = $230 + $231; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $233 = ((($this)) + 116|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $234 = +HEAPF32[$233>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $235 = +HEAPF32[$56>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $236 = $234 - $235; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $237 = ((($this)) + 120|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $238 = +HEAPF32[$237>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $239 = ((($this)) + 204|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $240 = +HEAPF32[$239>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $241 = $238 - $240; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $242 = $137 * $236; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $243 = $136 * $241; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $244 = $242 - $243; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $245 = $136 * $236; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $246 = $137 * $241; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $247 = $245 + $246; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $248 = ((($this)) + 100|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $249 = +HEAPF32[$248>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $250 = +HEAPF32[$32>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $251 = $249 - $250; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $252 = ((($this)) + 104|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $253 = +HEAPF32[$252>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $254 = ((($this)) + 188|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $255 = +HEAPF32[$254>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $256 = $253 - $255; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $257 = $133 * $251; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $258 = $132 * $256; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $259 = $257 - $258; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $260 = $132 * $251; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $261 = $133 * $256; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $262 = $260 + $261; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $263 = ((($this)) + 152|0); //@line 200 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $264 = +HEAPF32[$263>>2]; //@line 200 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $265 = $229 * $264; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $266 = $232 * $264; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $267 = $232 * $244; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $268 = $229 * $247; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $269 = $267 - $268; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $270 = $269 * $264; //@line 201 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $271 = ((($this)) + 268|0); //@line 201 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$271>>2] = $270; //@line 201 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $272 = $232 * $259; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $273 = $229 * $262; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $274 = $272 - $273; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $275 = $274 * $264; //@line 202 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $276 = ((($this)) + 260|0); //@line 202 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$276>>2] = $275; //@line 202 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $277 = $264 * $264; //@line 203 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $278 = +HEAPF32[$73>>2]; //@line 203 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $279 = $145 + $278; //@line 203 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $280 = $277 * $279; //@line 203 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $281 = $146 * $270; //@line 203 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $282 = $270 * $281; //@line 203 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $283 = $280 + $282; //@line 203 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $284 = $275 * $147; //@line 203 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $285 = $275 * $284; //@line 203 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $286 = $283 + $285; //@line 203 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $$pn = $286;$288 = $265;$290 = $266;$324 = $275;$350 = $270;
 }
 $storemerge22 = $205 + $$pn; //@line 203 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $287 = ((($this)) + 248|0); //@line 200 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$287>>2] = $288; //@line 200 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $289 = ((($this)) + 252|0); //@line 200 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$289>>2] = $290; //@line 200 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $291 = $storemerge22 > 0.0; //@line 207 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $292 = 1.0 / $storemerge22; //@line 207 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $293 = $291 ? $292 : 0.0; //@line 207 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$138>>2] = $293; //@line 207 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $294 = ((($data)) + 20|0); //@line 209 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $295 = HEAP8[$294>>0]|0; //@line 209 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $296 = ($295<<24>>24)==(0); //@line 209 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 if ($296) {
  $352 = ((($this)) + 156|0); //@line 222 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$352>>2] = 0.0; //@line 222 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $vA$sroa$0$0 = $99;$vA$sroa$6$0 = $101;$vB$sroa$0$0 = $107;$vB$sroa$6$0 = $109;$vC$sroa$0$0 = $116;$vC$sroa$6$0 = $118;$vD$sroa$0$0 = $125;$vD$sroa$6$0 = $127;$wA$0 = $103;$wB$0 = $111;$wC$0 = $120;$wD$0 = $129;
  $353 = HEAP32[$4>>2]|0; //@line 225 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $354 = HEAP32[$96>>2]|0; //@line 225 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $355 = (($354) + (($353*12)|0)|0); //@line 225 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$355>>2] = $vA$sroa$0$0; //@line 225 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $356 = (((($354) + (($353*12)|0)|0)) + 4|0); //@line 225 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$356>>2] = $vA$sroa$6$0; //@line 225 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $357 = HEAP32[$4>>2]|0; //@line 226 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $358 = HEAP32[$96>>2]|0; //@line 226 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $359 = (((($358) + (($357*12)|0)|0)) + 8|0); //@line 226 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$359>>2] = $wA$0; //@line 226 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $360 = HEAP32[$9>>2]|0; //@line 227 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $361 = (($358) + (($360*12)|0)|0); //@line 227 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$361>>2] = $vB$sroa$0$0; //@line 227 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $362 = (((($358) + (($360*12)|0)|0)) + 4|0); //@line 227 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$362>>2] = $vB$sroa$6$0; //@line 227 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $363 = HEAP32[$9>>2]|0; //@line 228 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $364 = HEAP32[$96>>2]|0; //@line 228 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $365 = (((($364) + (($363*12)|0)|0)) + 8|0); //@line 228 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$365>>2] = $wB$0; //@line 228 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $366 = HEAP32[$14>>2]|0; //@line 229 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $367 = (($364) + (($366*12)|0)|0); //@line 229 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$367>>2] = $vC$sroa$0$0; //@line 229 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $368 = (((($364) + (($366*12)|0)|0)) + 4|0); //@line 229 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$368>>2] = $vC$sroa$6$0; //@line 229 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $369 = HEAP32[$14>>2]|0; //@line 230 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $370 = HEAP32[$96>>2]|0; //@line 230 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $371 = (((($370) + (($369*12)|0)|0)) + 8|0); //@line 230 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$371>>2] = $wC$0; //@line 230 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $372 = HEAP32[$19>>2]|0; //@line 231 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $373 = (($370) + (($372*12)|0)|0); //@line 231 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$373>>2] = $vD$sroa$0$0; //@line 231 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $374 = (((($370) + (($372*12)|0)|0)) + 4|0); //@line 231 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$374>>2] = $vD$sroa$6$0; //@line 231 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $375 = HEAP32[$19>>2]|0; //@line 232 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $376 = HEAP32[$96>>2]|0; //@line 232 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $377 = (((($376) + (($375*12)|0)|0)) + 8|0); //@line 232 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$377>>2] = $wD$0; //@line 232 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  return; //@line 233 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 } else {
  $297 = +HEAPF32[$70>>2]; //@line 211 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $298 = ((($this)) + 156|0); //@line 211 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $299 = +HEAPF32[$298>>2]; //@line 211 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $300 = $297 * $299; //@line 211 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $301 = $207 * $300; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $302 = $300 * $209; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $303 = (HEAP32[tempDoublePtr>>2]=$99,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $304 = $303 + $301; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $305 = (HEAPF32[tempDoublePtr>>2]=$304,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $306 = (HEAP32[tempDoublePtr>>2]=$101,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $307 = $306 + $302; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $308 = (HEAPF32[tempDoublePtr>>2]=$307,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $309 = $299 * $144; //@line 212 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $310 = $309 * $211; //@line 212 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $311 = $103 + $310; //@line 212 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $312 = +HEAPF32[$73>>2]; //@line 213 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $313 = $299 * $312; //@line 213 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $314 = $288 * $313; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $315 = $313 * $290; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $316 = (HEAP32[tempDoublePtr>>2]=$107,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $317 = $316 + $314; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $318 = (HEAPF32[tempDoublePtr>>2]=$317,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $319 = (HEAP32[tempDoublePtr>>2]=$109,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $320 = $319 + $315; //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $321 = (HEAPF32[tempDoublePtr>>2]=$320,HEAP32[tempDoublePtr>>2]|0); //@line 85 "./libraries/Box2D/Common/b2Math.h"
  $322 = $299 * $147; //@line 214 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $323 = $322 * $324; //@line 214 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $325 = $111 + $323; //@line 214 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $326 = $299 * $142; //@line 215 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $327 = $207 * $326; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $328 = $209 * $326; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $329 = (HEAP32[tempDoublePtr>>2]=$116,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $330 = $329 - $327; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $331 = (HEAPF32[tempDoublePtr>>2]=$330,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $332 = (HEAP32[tempDoublePtr>>2]=$118,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $333 = $332 - $328; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $334 = (HEAPF32[tempDoublePtr>>2]=$333,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $335 = $299 * $143; //@line 216 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $336 = $335 * $337; //@line 216 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $338 = $120 - $336; //@line 216 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $339 = $299 * $145; //@line 217 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $340 = $288 * $339; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $341 = $290 * $339; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $342 = (HEAP32[tempDoublePtr>>2]=$125,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $343 = $342 - $340; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $344 = (HEAPF32[tempDoublePtr>>2]=$343,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $345 = (HEAP32[tempDoublePtr>>2]=$127,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $346 = $345 - $341; //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $347 = (HEAPF32[tempDoublePtr>>2]=$346,HEAP32[tempDoublePtr>>2]|0); //@line 91 "./libraries/Box2D/Common/b2Math.h"
  $348 = $299 * $146; //@line 218 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $349 = $348 * $350; //@line 218 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $351 = $129 - $349; //@line 218 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $vA$sroa$0$0 = $305;$vA$sroa$6$0 = $308;$vB$sroa$0$0 = $318;$vB$sroa$6$0 = $321;$vC$sroa$0$0 = $331;$vC$sroa$6$0 = $334;$vD$sroa$0$0 = $344;$vD$sroa$6$0 = $347;$wA$0 = $311;$wB$0 = $325;$wC$0 = $338;$wD$0 = $351;
  $353 = HEAP32[$4>>2]|0; //@line 225 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $354 = HEAP32[$96>>2]|0; //@line 225 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $355 = (($354) + (($353*12)|0)|0); //@line 225 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$355>>2] = $vA$sroa$0$0; //@line 225 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $356 = (((($354) + (($353*12)|0)|0)) + 4|0); //@line 225 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$356>>2] = $vA$sroa$6$0; //@line 225 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $357 = HEAP32[$4>>2]|0; //@line 226 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $358 = HEAP32[$96>>2]|0; //@line 226 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $359 = (((($358) + (($357*12)|0)|0)) + 8|0); //@line 226 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$359>>2] = $wA$0; //@line 226 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $360 = HEAP32[$9>>2]|0; //@line 227 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $361 = (($358) + (($360*12)|0)|0); //@line 227 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$361>>2] = $vB$sroa$0$0; //@line 227 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $362 = (((($358) + (($360*12)|0)|0)) + 4|0); //@line 227 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$362>>2] = $vB$sroa$6$0; //@line 227 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $363 = HEAP32[$9>>2]|0; //@line 228 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $364 = HEAP32[$96>>2]|0; //@line 228 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $365 = (((($364) + (($363*12)|0)|0)) + 8|0); //@line 228 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$365>>2] = $wB$0; //@line 228 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $366 = HEAP32[$14>>2]|0; //@line 229 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $367 = (($364) + (($366*12)|0)|0); //@line 229 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$367>>2] = $vC$sroa$0$0; //@line 229 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $368 = (((($364) + (($366*12)|0)|0)) + 4|0); //@line 229 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$368>>2] = $vC$sroa$6$0; //@line 229 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $369 = HEAP32[$14>>2]|0; //@line 230 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $370 = HEAP32[$96>>2]|0; //@line 230 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $371 = (((($370) + (($369*12)|0)|0)) + 8|0); //@line 230 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$371>>2] = $wC$0; //@line 230 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $372 = HEAP32[$19>>2]|0; //@line 231 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $373 = (($370) + (($372*12)|0)|0); //@line 231 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$373>>2] = $vD$sroa$0$0; //@line 231 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $374 = (((($370) + (($372*12)|0)|0)) + 4|0); //@line 231 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$374>>2] = $vD$sroa$6$0; //@line 231 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $375 = HEAP32[$19>>2]|0; //@line 232 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $376 = HEAP32[$96>>2]|0; //@line 232 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $377 = (((($376) + (($375*12)|0)|0)) + 8|0); //@line 232 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$377>>2] = $wD$0; //@line 232 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  return; //@line 233 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 }
}
function __ZN11b2GearJoint24SolveVelocityConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0, $113 = 0.0, $114 = 0.0, $115 = 0.0;
 var $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0.0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0;
 var $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0, $49 = 0.0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0;
 var $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0.0, $63 = 0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0, $7 = 0.0, $70 = 0.0, $71 = 0.0, $72 = 0.0;
 var $73 = 0, $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0, $89 = 0.0, $9 = 0.0, $90 = 0.0;
 var $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 160|0); //@line 237 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 237 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $2 = ((($data)) + 28|0); //@line 237 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 237 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $4 = (($3) + (($1*12)|0)|0); //@line 237 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $5 = +HEAPF32[$4>>2]; //@line 237 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $6 = (((($3) + (($1*12)|0)|0)) + 4|0); //@line 237 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 237 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $8 = (((($3) + (($1*12)|0)|0)) + 8|0); //@line 238 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 238 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $10 = ((($this)) + 164|0); //@line 239 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 239 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $12 = (($3) + (($11*12)|0)|0); //@line 239 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $13 = +HEAPF32[$12>>2]; //@line 239 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $14 = (((($3) + (($11*12)|0)|0)) + 4|0); //@line 239 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 239 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $16 = (((($3) + (($11*12)|0)|0)) + 8|0); //@line 240 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 240 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $18 = ((($this)) + 168|0); //@line 241 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 241 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $20 = (($3) + (($19*12)|0)|0); //@line 241 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 241 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $22 = (((($3) + (($19*12)|0)|0)) + 4|0); //@line 241 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $23 = +HEAPF32[$22>>2]; //@line 241 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $24 = (((($3) + (($19*12)|0)|0)) + 8|0); //@line 242 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $25 = +HEAPF32[$24>>2]; //@line 242 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $26 = ((($this)) + 172|0); //@line 243 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $27 = HEAP32[$26>>2]|0; //@line 243 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $28 = (($3) + (($27*12)|0)|0); //@line 243 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $29 = +HEAPF32[$28>>2]; //@line 243 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $30 = (((($3) + (($27*12)|0)|0)) + 4|0); //@line 243 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $31 = +HEAPF32[$30>>2]; //@line 243 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $32 = (((($3) + (($27*12)|0)|0)) + 8|0); //@line 244 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $33 = +HEAPF32[$32>>2]; //@line 244 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $34 = $5 - $21; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $35 = $7 - $23; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $36 = ((($this)) + 240|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $37 = +HEAPF32[$36>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $38 = $34 * $37; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $39 = ((($this)) + 244|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $40 = +HEAPF32[$39>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $41 = $35 * $40; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $42 = $38 + $41; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $43 = $13 - $29; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $44 = $15 - $31; //@line 454 "./libraries/Box2D/Common/b2Math.h"
 $45 = ((($this)) + 248|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $46 = +HEAPF32[$45>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $47 = $43 * $46; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $48 = ((($this)) + 252|0); //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $49 = +HEAPF32[$48>>2]; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $50 = $44 * $49; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $51 = $47 + $50; //@line 408 "./libraries/Box2D/Common/b2Math.h"
 $52 = $42 + $51; //@line 246 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $53 = ((($this)) + 256|0); //@line 247 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $54 = +HEAPF32[$53>>2]; //@line 247 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $55 = $9 * $54; //@line 247 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $56 = ((($this)) + 264|0); //@line 247 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $57 = +HEAPF32[$56>>2]; //@line 247 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $58 = $25 * $57; //@line 247 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $59 = $55 - $58; //@line 247 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $60 = ((($this)) + 260|0); //@line 247 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $61 = +HEAPF32[$60>>2]; //@line 247 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $62 = $17 * $61; //@line 247 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $63 = ((($this)) + 268|0); //@line 247 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $64 = +HEAPF32[$63>>2]; //@line 247 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $65 = $33 * $64; //@line 247 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $66 = $62 - $65; //@line 247 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $67 = $59 + $66; //@line 247 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $68 = $52 + $67; //@line 247 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $69 = ((($this)) + 272|0); //@line 249 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $70 = +HEAPF32[$69>>2]; //@line 249 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $71 = $70 * $68; //@line 249 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $72 = -$71; //@line 249 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $73 = ((($this)) + 156|0); //@line 250 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $74 = +HEAPF32[$73>>2]; //@line 250 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $75 = $74 - $71; //@line 250 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$73>>2] = $75; //@line 250 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $76 = ((($this)) + 208|0); //@line 252 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $77 = +HEAPF32[$76>>2]; //@line 252 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $78 = $77 * $72; //@line 252 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $79 = $37 * $78; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $80 = $40 * $78; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $81 = $5 + $79; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $82 = $7 + $80; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $83 = ((($this)) + 224|0); //@line 253 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $84 = +HEAPF32[$83>>2]; //@line 253 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $85 = $84 * $72; //@line 253 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $86 = $85 * $54; //@line 253 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $87 = $9 + $86; //@line 253 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $88 = ((($this)) + 212|0); //@line 254 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $89 = +HEAPF32[$88>>2]; //@line 254 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $90 = $89 * $72; //@line 254 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $91 = $46 * $90; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $92 = $90 * $49; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $93 = $13 + $91; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $94 = $15 + $92; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $95 = ((($this)) + 228|0); //@line 255 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $96 = +HEAPF32[$95>>2]; //@line 255 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $97 = $96 * $72; //@line 255 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $98 = $97 * $61; //@line 255 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $99 = $17 + $98; //@line 255 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $100 = ((($this)) + 216|0); //@line 256 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $101 = +HEAPF32[$100>>2]; //@line 256 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $102 = $101 * $72; //@line 256 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $103 = $37 * $102; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $104 = $40 * $102; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $105 = $21 - $103; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $106 = $23 - $104; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $107 = ((($this)) + 232|0); //@line 257 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $108 = +HEAPF32[$107>>2]; //@line 257 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $109 = $108 * $72; //@line 257 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $110 = $109 * $57; //@line 257 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $111 = $25 - $110; //@line 257 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $112 = ((($this)) + 220|0); //@line 258 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $113 = +HEAPF32[$112>>2]; //@line 258 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $114 = $113 * $72; //@line 258 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $115 = $46 * $114; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $116 = $49 * $114; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $117 = $29 - $115; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $118 = $31 - $116; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $119 = ((($this)) + 236|0); //@line 259 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $120 = +HEAPF32[$119>>2]; //@line 259 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $121 = $120 * $72; //@line 259 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $122 = $121 * $64; //@line 259 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $123 = $33 - $122; //@line 259 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $124 = HEAP32[$0>>2]|0; //@line 261 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $125 = HEAP32[$2>>2]|0; //@line 261 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $126 = (($125) + (($124*12)|0)|0); //@line 261 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$126>>2] = $81; //@line 261 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $127 = (((($125) + (($124*12)|0)|0)) + 4|0); //@line 261 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$127>>2] = $82; //@line 261 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $128 = HEAP32[$0>>2]|0; //@line 262 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $129 = HEAP32[$2>>2]|0; //@line 262 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $130 = (((($129) + (($128*12)|0)|0)) + 8|0); //@line 262 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$130>>2] = $87; //@line 262 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $131 = HEAP32[$10>>2]|0; //@line 263 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $132 = (($129) + (($131*12)|0)|0); //@line 263 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$132>>2] = $93; //@line 263 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $133 = (((($129) + (($131*12)|0)|0)) + 4|0); //@line 263 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$133>>2] = $94; //@line 263 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $134 = HEAP32[$10>>2]|0; //@line 264 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $135 = HEAP32[$2>>2]|0; //@line 264 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $136 = (((($135) + (($134*12)|0)|0)) + 8|0); //@line 264 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$136>>2] = $99; //@line 264 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $137 = HEAP32[$18>>2]|0; //@line 265 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $138 = (($135) + (($137*12)|0)|0); //@line 265 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$138>>2] = $105; //@line 265 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $139 = (((($135) + (($137*12)|0)|0)) + 4|0); //@line 265 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$139>>2] = $106; //@line 265 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $140 = HEAP32[$18>>2]|0; //@line 266 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $141 = HEAP32[$2>>2]|0; //@line 266 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $142 = (((($141) + (($140*12)|0)|0)) + 8|0); //@line 266 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$142>>2] = $111; //@line 266 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $143 = HEAP32[$26>>2]|0; //@line 267 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $144 = (($141) + (($143*12)|0)|0); //@line 267 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$144>>2] = $117; //@line 267 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $145 = (((($141) + (($143*12)|0)|0)) + 4|0); //@line 267 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$145>>2] = $118; //@line 267 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $146 = HEAP32[$26>>2]|0; //@line 268 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $147 = HEAP32[$2>>2]|0; //@line 268 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $148 = (((($147) + (($146*12)|0)|0)) + 8|0); //@line 268 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$148>>2] = $123; //@line 268 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 return; //@line 269 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
}
function __ZN11b2GearJoint24SolvePositionConstraintsERK12b2SolverData($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$pn = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0;
 var $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0.0, $132 = 0.0;
 var $133 = 0.0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0, $141 = 0.0, $142 = 0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0, $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0;
 var $151 = 0.0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0, $161 = 0.0, $162 = 0, $163 = 0.0, $164 = 0.0, $165 = 0, $166 = 0.0, $167 = 0, $168 = 0.0, $169 = 0.0;
 var $17 = 0.0, $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0, $177 = 0.0, $178 = 0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0, $182 = 0.0, $183 = 0, $184 = 0.0, $185 = 0.0, $186 = 0.0, $187 = 0.0;
 var $188 = 0.0, $189 = 0.0, $19 = 0, $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0.0, $194 = 0.0, $195 = 0.0, $196 = 0.0, $197 = 0.0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0.0, $201 = 0.0, $202 = 0.0, $203 = 0, $204 = 0.0;
 var $205 = 0, $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0, $21 = 0, $210 = 0.0, $211 = 0.0, $212 = 0.0, $213 = 0.0, $214 = 0, $215 = 0.0, $216 = 0.0, $217 = 0.0, $218 = 0.0, $219 = 0.0, $22 = 0, $220 = 0.0, $221 = 0.0, $222 = 0.0;
 var $223 = 0.0, $224 = 0.0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0.0, $229 = 0.0, $23 = 0, $230 = 0.0, $231 = 0.0, $232 = 0.0, $233 = 0.0, $234 = 0.0, $235 = 0.0, $236 = 0.0, $237 = 0.0, $238 = 0, $239 = 0.0, $24 = 0, $240 = 0.0;
 var $241 = 0, $242 = 0.0, $243 = 0.0, $244 = 0.0, $245 = 0.0, $246 = 0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0.0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0.0, $254 = 0.0, $255 = 0.0, $256 = 0.0, $257 = 0.0, $258 = 0, $259 = 0.0;
 var $26 = 0, $260 = 0.0, $261 = 0.0, $262 = 0.0, $263 = 0.0, $264 = 0.0, $265 = 0.0, $266 = 0.0, $267 = 0.0, $268 = 0.0, $269 = 0.0, $27 = 0, $270 = 0.0, $271 = 0, $272 = 0.0, $273 = 0.0, $274 = 0.0, $275 = 0.0, $276 = 0.0, $277 = 0.0;
 var $278 = 0.0, $279 = 0.0, $28 = 0, $280 = 0.0, $281 = 0.0, $282 = 0.0, $283 = 0.0, $284 = 0, $285 = 0.0, $286 = 0.0, $287 = 0.0, $288 = 0.0, $289 = 0.0, $29 = 0, $290 = 0.0, $291 = 0.0, $292 = 0.0, $293 = 0.0, $294 = 0.0, $295 = 0.0;
 var $296 = 0.0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0;
 var $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0, $55 = 0.0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0;
 var $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0;
 var $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0.0, $85 = 0, $86 = 0.0, $87 = 0, $88 = 0.0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0;
 var $JvAC$sroa$0$0 = 0.0, $JvAC$sroa$6$0 = 0.0, $JvBD$sroa$0$0 = 0.0, $JvBD$sroa$6$0 = 0.0, $JwA$0 = 0.0, $JwB$0 = 0.0, $JwC$0 = 0.0, $JwD$0 = 0.0, $coordinateA$0 = 0.0, $coordinateB$0 = 0.0, $impulse$0 = 0.0, $mass$0 = 0.0, $mass$0$in = 0.0, $mass$1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 160|0); //@line 273 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 357 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $2 = ((($data)) + 24|0); //@line 273 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 357 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $4 = (($3) + (($1*12)|0)|0); //@line 273 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 273 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $6 = (((($3) + (($1*12)|0)|0)) + 4|0); //@line 273 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 273 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $8 = (((($3) + (($1*12)|0)|0)) + 8|0); //@line 274 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 274 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $10 = ((($this)) + 164|0); //@line 275 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 275 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $12 = (($3) + (($11*12)|0)|0); //@line 275 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 275 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $14 = (((($3) + (($11*12)|0)|0)) + 4|0); //@line 275 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 275 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $16 = (((($3) + (($11*12)|0)|0)) + 8|0); //@line 276 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $17 = +HEAPF32[$16>>2]; //@line 276 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $18 = ((($this)) + 168|0); //@line 277 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 277 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $20 = (($3) + (($19*12)|0)|0); //@line 277 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $21 = HEAP32[$20>>2]|0; //@line 277 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $22 = (((($3) + (($19*12)|0)|0)) + 4|0); //@line 277 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $23 = HEAP32[$22>>2]|0; //@line 277 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $24 = (((($3) + (($19*12)|0)|0)) + 8|0); //@line 278 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $25 = +HEAPF32[$24>>2]; //@line 278 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $26 = ((($this)) + 172|0); //@line 279 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $27 = HEAP32[$26>>2]|0; //@line 279 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $28 = (($3) + (($27*12)|0)|0); //@line 279 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 279 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $30 = (((($3) + (($27*12)|0)|0)) + 4|0); //@line 279 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 279 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $32 = (((($3) + (($27*12)|0)|0)) + 8|0); //@line 280 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $33 = +HEAPF32[$32>>2]; //@line 280 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $34 = (+Math_sin((+$9))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $35 = (+Math_cos((+$9))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $36 = (+Math_sin((+$17))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $37 = (+Math_cos((+$17))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $38 = (+Math_sin((+$25))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $39 = (+Math_cos((+$25))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $40 = (+Math_sin((+$33))); //@line 307 "./libraries/Box2D/Common/b2Math.h"
 $41 = (+Math_cos((+$33))); //@line 308 "./libraries/Box2D/Common/b2Math.h"
 $42 = ((($this)) + 76|0); //@line 292 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $43 = HEAP32[$42>>2]|0; //@line 292 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $44 = ($43|0)==(1); //@line 292 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $45 = ((($this)) + 224|0); //@line 297 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $46 = +HEAPF32[$45>>2]; //@line 349 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 if ($44) {
  $47 = ((($this)) + 232|0); //@line 297 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $48 = +HEAPF32[$47>>2]; //@line 297 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $49 = $46 + $48; //@line 297 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $50 = $9 - $25; //@line 299 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $51 = ((($this)) + 140|0); //@line 299 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $52 = +HEAPF32[$51>>2]; //@line 299 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $53 = $50 - $52; //@line 299 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $281 = $48;$JvAC$sroa$0$0 = 0.0;$JvAC$sroa$6$0 = 0.0;$JwA$0 = 1.0;$JwC$0 = 1.0;$coordinateA$0 = $53;$mass$0$in = $49;
 } else {
  $54 = ((($this)) + 124|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $55 = +HEAPF32[$54>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $56 = $39 * $55; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $57 = ((($this)) + 128|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $58 = +HEAPF32[$57>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $59 = $38 * $58; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $60 = $56 - $59; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $61 = $38 * $55; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $62 = $39 * $58; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $63 = $61 + $62; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $64 = ((($this)) + 108|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $65 = +HEAPF32[$64>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $66 = ((($this)) + 192|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $67 = +HEAPF32[$66>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $68 = $65 - $67; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $69 = ((($this)) + 112|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $70 = +HEAPF32[$69>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $71 = ((($this)) + 196|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $72 = +HEAPF32[$71>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $73 = $70 - $72; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $74 = $39 * $68; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $75 = $38 * $73; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $76 = $74 - $75; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $77 = $38 * $68; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $78 = $39 * $73; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $79 = $77 + $78; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $80 = ((($this)) + 92|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $81 = +HEAPF32[$80>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $82 = ((($this)) + 176|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $83 = +HEAPF32[$82>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $84 = $81 - $83; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $85 = ((($this)) + 96|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $86 = +HEAPF32[$85>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $87 = ((($this)) + 180|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $88 = +HEAPF32[$87>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $89 = $86 - $88; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $90 = $35 * $84; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $91 = $34 * $89; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $92 = $90 - $91; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $93 = $34 * $84; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $94 = $35 * $89; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $95 = $93 + $94; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $96 = $63 * $76; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $97 = $60 * $79; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $98 = $96 - $97; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $99 = $63 * $92; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $100 = $60 * $95; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $101 = $99 - $100; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $102 = ((($this)) + 216|0); //@line 309 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $103 = +HEAPF32[$102>>2]; //@line 309 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $104 = ((($this)) + 208|0); //@line 309 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $105 = +HEAPF32[$104>>2]; //@line 309 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $106 = $103 + $105; //@line 309 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $107 = ((($this)) + 232|0); //@line 309 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $108 = +HEAPF32[$107>>2]; //@line 309 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $109 = $98 * $108; //@line 309 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $110 = $98 * $109; //@line 309 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $111 = $106 + $110; //@line 309 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $112 = $46 * $101; //@line 309 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $113 = $101 * $112; //@line 309 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $114 = $111 + $113; //@line 309 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $115 = (HEAP32[tempDoublePtr>>2]=$5,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $116 = (HEAP32[tempDoublePtr>>2]=$21,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $117 = $115 - $116; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $118 = (HEAP32[tempDoublePtr>>2]=$7,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $119 = (HEAP32[tempDoublePtr>>2]=$23,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $120 = $118 - $119; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $121 = $117 + $92; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $122 = $120 + $95; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $123 = $39 * $121; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $124 = $38 * $122; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $125 = $123 + $124; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $126 = $38 * $121; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $127 = $39 * $122; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $128 = $127 - $126; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $129 = $125 - $68; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $130 = $128 - $73; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $131 = $55 * $129; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $132 = $58 * $130; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $133 = $131 + $132; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $281 = $108;$JvAC$sroa$0$0 = $60;$JvAC$sroa$6$0 = $63;$JwA$0 = $101;$JwC$0 = $98;$coordinateA$0 = $133;$mass$0$in = $114;
 }
 $mass$0 = $mass$0$in + 0.0; //@line 297 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $134 = ((($this)) + 80|0); //@line 316 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $135 = HEAP32[$134>>2]|0; //@line 316 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $136 = ($135|0)==(1); //@line 316 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $137 = ((($this)) + 152|0); //@line 319 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $138 = +HEAPF32[$137>>2]; //@line 340 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 if ($136) {
  $139 = $138 * $138; //@line 321 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $140 = ((($this)) + 228|0); //@line 321 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $141 = +HEAPF32[$140>>2]; //@line 321 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $142 = ((($this)) + 236|0); //@line 321 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $143 = +HEAPF32[$142>>2]; //@line 321 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $144 = $141 + $143; //@line 321 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $145 = $139 * $144; //@line 321 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $146 = $17 - $33; //@line 323 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $147 = ((($this)) + 144|0); //@line 323 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $148 = +HEAPF32[$147>>2]; //@line 323 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $149 = $146 - $148; //@line 323 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $$pn = $145;$268 = $141;$294 = $143;$JvBD$sroa$0$0 = 0.0;$JvBD$sroa$6$0 = 0.0;$JwB$0 = $138;$JwD$0 = $138;$coordinateB$0 = $149;
 } else {
  $150 = ((($this)) + 132|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $151 = +HEAPF32[$150>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $152 = $41 * $151; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $153 = ((($this)) + 136|0); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $154 = +HEAPF32[$153>>2]; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $155 = $40 * $154; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $156 = $152 - $155; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $157 = $40 * $151; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $158 = $41 * $154; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $159 = $157 + $158; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $160 = ((($this)) + 116|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $161 = +HEAPF32[$160>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $162 = ((($this)) + 200|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $163 = +HEAPF32[$162>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $164 = $161 - $163; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $165 = ((($this)) + 120|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $166 = +HEAPF32[$165>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $167 = ((($this)) + 204|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $168 = +HEAPF32[$167>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $169 = $166 - $168; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $170 = $41 * $164; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $171 = $40 * $169; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $172 = $170 - $171; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $173 = $40 * $164; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $174 = $41 * $169; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $175 = $173 + $174; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $176 = ((($this)) + 100|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $177 = +HEAPF32[$176>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $178 = ((($this)) + 184|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $179 = +HEAPF32[$178>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $180 = $177 - $179; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $181 = ((($this)) + 104|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $182 = +HEAPF32[$181>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $183 = ((($this)) + 188|0); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $184 = +HEAPF32[$183>>2]; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $185 = $182 - $184; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $186 = $37 * $180; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $187 = $36 * $185; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $188 = $186 - $187; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $189 = $36 * $180; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $190 = $37 * $185; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $191 = $189 + $190; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $192 = $156 * $138; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $193 = $159 * $138; //@line 459 "./libraries/Box2D/Common/b2Math.h"
  $194 = $159 * $172; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $195 = $156 * $175; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $196 = $194 - $195; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $197 = $138 * $196; //@line 331 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $198 = $159 * $188; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $199 = $156 * $191; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $200 = $198 - $199; //@line 414 "./libraries/Box2D/Common/b2Math.h"
  $201 = $138 * $200; //@line 332 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $202 = $138 * $138; //@line 333 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $203 = ((($this)) + 220|0); //@line 333 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $204 = +HEAPF32[$203>>2]; //@line 333 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $205 = ((($this)) + 212|0); //@line 333 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $206 = +HEAPF32[$205>>2]; //@line 333 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $207 = $204 + $206; //@line 333 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $208 = $202 * $207; //@line 333 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $209 = ((($this)) + 236|0); //@line 333 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $210 = +HEAPF32[$209>>2]; //@line 333 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $211 = $197 * $210; //@line 333 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $212 = $197 * $211; //@line 333 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $213 = $208 + $212; //@line 333 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $214 = ((($this)) + 228|0); //@line 333 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $215 = +HEAPF32[$214>>2]; //@line 333 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $216 = $215 * $201; //@line 333 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $217 = $201 * $216; //@line 333 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $218 = $213 + $217; //@line 333 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $219 = (HEAP32[tempDoublePtr>>2]=$13,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $220 = (HEAP32[tempDoublePtr>>2]=$29,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $221 = $219 - $220; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $222 = (HEAP32[tempDoublePtr>>2]=$15,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $223 = (HEAP32[tempDoublePtr>>2]=$31,+HEAPF32[tempDoublePtr>>2]); //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $224 = $222 - $223; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $225 = $221 + $188; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $226 = $224 + $191; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $227 = $41 * $225; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $228 = $40 * $226; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $229 = $227 + $228; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $230 = $40 * $225; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $231 = $41 * $226; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $232 = $231 - $230; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $233 = $229 - $164; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $234 = $232 - $169; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $235 = $151 * $233; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $236 = $154 * $234; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $237 = $235 + $236; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $$pn = $218;$268 = $215;$294 = $210;$JvBD$sroa$0$0 = $192;$JvBD$sroa$6$0 = $193;$JwB$0 = $201;$JwD$0 = $197;$coordinateB$0 = $237;
 }
 $mass$1 = $mass$0 + $$pn; //@line 321 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $238 = $mass$1 > 0.0; //@line 343 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 if ($238) {
  $239 = $coordinateB$0 * $138; //@line 340 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $240 = $coordinateA$0 + $239; //@line 340 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $241 = ((($this)) + 148|0); //@line 340 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $242 = +HEAPF32[$241>>2]; //@line 340 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $243 = $240 - $242; //@line 340 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $244 = -$243; //@line 345 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $245 = $244 / $mass$1; //@line 345 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $impulse$0 = $245;
 } else {
  $impulse$0 = 0.0;
 }
 $246 = ((($this)) + 208|0); //@line 348 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $247 = +HEAPF32[$246>>2]; //@line 348 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $248 = $impulse$0 * $247; //@line 348 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $249 = $JvAC$sroa$0$0 * $248; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $250 = $JvAC$sroa$6$0 * $248; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $251 = (HEAP32[tempDoublePtr>>2]=$5,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $252 = $251 + $249; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $253 = (HEAP32[tempDoublePtr>>2]=$7,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $254 = $253 + $250; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $255 = $impulse$0 * $46; //@line 349 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $256 = $JwA$0 * $255; //@line 349 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $257 = $9 + $256; //@line 349 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $258 = ((($this)) + 212|0); //@line 350 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $259 = +HEAPF32[$258>>2]; //@line 350 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $260 = $impulse$0 * $259; //@line 350 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $261 = $JvBD$sroa$0$0 * $260; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $262 = $JvBD$sroa$6$0 * $260; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $263 = (HEAP32[tempDoublePtr>>2]=$13,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $264 = $263 + $261; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $265 = (HEAP32[tempDoublePtr>>2]=$15,+HEAPF32[tempDoublePtr>>2]); //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $266 = $265 + $262; //@line 85 "./libraries/Box2D/Common/b2Math.h"
 $267 = $impulse$0 * $268; //@line 351 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $269 = $JwB$0 * $267; //@line 351 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $270 = $17 + $269; //@line 351 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $271 = ((($this)) + 216|0); //@line 352 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $272 = +HEAPF32[$271>>2]; //@line 352 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $273 = $impulse$0 * $272; //@line 352 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $274 = $JvAC$sroa$0$0 * $273; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $275 = $JvAC$sroa$6$0 * $273; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $276 = (HEAP32[tempDoublePtr>>2]=$21,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $277 = $276 - $274; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $278 = (HEAP32[tempDoublePtr>>2]=$23,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $279 = $278 - $275; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $280 = $impulse$0 * $281; //@line 353 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $282 = $JwC$0 * $280; //@line 353 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $283 = $25 - $282; //@line 353 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $284 = ((($this)) + 220|0); //@line 354 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $285 = +HEAPF32[$284>>2]; //@line 354 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $286 = $impulse$0 * $285; //@line 354 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $287 = $JvBD$sroa$0$0 * $286; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $288 = $JvBD$sroa$6$0 * $286; //@line 459 "./libraries/Box2D/Common/b2Math.h"
 $289 = (HEAP32[tempDoublePtr>>2]=$29,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $290 = $289 - $287; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $291 = (HEAP32[tempDoublePtr>>2]=$31,+HEAPF32[tempDoublePtr>>2]); //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $292 = $291 - $288; //@line 91 "./libraries/Box2D/Common/b2Math.h"
 $293 = $impulse$0 * $294; //@line 355 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $295 = $JwD$0 * $293; //@line 355 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $296 = $33 - $295; //@line 355 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $297 = (($3) + (($1*12)|0)|0); //@line 357 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$297>>2] = $252; //@line 357 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$6>>2] = $254; //@line 357 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $298 = HEAP32[$0>>2]|0; //@line 358 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $299 = HEAP32[$2>>2]|0; //@line 358 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $300 = (((($299) + (($298*12)|0)|0)) + 8|0); //@line 358 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$300>>2] = $257; //@line 358 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $301 = HEAP32[$10>>2]|0; //@line 359 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $302 = (($299) + (($301*12)|0)|0); //@line 359 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$302>>2] = $264; //@line 359 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $303 = (((($299) + (($301*12)|0)|0)) + 4|0); //@line 359 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$303>>2] = $266; //@line 359 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $304 = HEAP32[$10>>2]|0; //@line 360 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $305 = HEAP32[$2>>2]|0; //@line 360 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $306 = (((($305) + (($304*12)|0)|0)) + 8|0); //@line 360 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$306>>2] = $270; //@line 360 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $307 = HEAP32[$18>>2]|0; //@line 361 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $308 = (($305) + (($307*12)|0)|0); //@line 361 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$308>>2] = $277; //@line 361 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $309 = (((($305) + (($307*12)|0)|0)) + 4|0); //@line 361 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$309>>2] = $279; //@line 361 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $310 = HEAP32[$18>>2]|0; //@line 362 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $311 = HEAP32[$2>>2]|0; //@line 362 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $312 = (((($311) + (($310*12)|0)|0)) + 8|0); //@line 362 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$312>>2] = $283; //@line 362 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $313 = HEAP32[$26>>2]|0; //@line 363 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $314 = (($311) + (($313*12)|0)|0); //@line 363 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$314>>2] = $290; //@line 363 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $315 = (((($311) + (($313*12)|0)|0)) + 4|0); //@line 363 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$315>>2] = $292; //@line 363 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $316 = HEAP32[$26>>2]|0; //@line 364 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $317 = HEAP32[$2>>2]|0; //@line 364 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $318 = (((($317) + (($316*12)|0)|0)) + 8|0); //@line 364 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$318>>2] = $296; //@line 364 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 return 1; //@line 367 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
}
function __ZN11b2GearJointC2EPK14b2GearJointDef($this,$def) {
 $this = $this|0;
 $def = $def|0;
 var $$cast = 0, $$cast18 = 0.0, $$cast19 = 0.0, $$off = 0, $$off17 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0;
 var $13 = 0, $130 = 0.0, $131 = 0.0, $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0.0, $151 = 0, $152 = 0.0, $153 = 0, $154 = 0, $155 = 0.0, $156 = 0, $157 = 0.0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0.0, $162 = 0, $163 = 0.0, $164 = 0, $165 = 0;
 var $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0.0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0, $200 = 0.0;
 var $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0;
 var $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0;
 var $238 = 0.0, $239 = 0.0, $24 = 0, $240 = 0.0, $241 = 0.0, $242 = 0.0, $243 = 0.0, $244 = 0.0, $245 = 0.0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0.0, $254 = 0.0, $255 = 0.0;
 var $256 = 0.0, $257 = 0.0, $258 = 0.0, $259 = 0.0, $26 = 0, $260 = 0.0, $261 = 0.0, $262 = 0, $263 = 0.0, $264 = 0, $265 = 0.0, $266 = 0.0, $267 = 0, $268 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0;
 var $32 = 0.0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0, $43 = 0.0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bitcast = 0.0, $bitcast21 = 0.0, $bitcast23 = 0.0, $bitcast25 = 0.0, $bitcast27 = 0.0, $bitcast29 = 0.0;
 var $coordinateA$0 = 0.0, $coordinateB$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN7b2JointC2EPK10b2JointDef($this,$def); //@line 45 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$this>>2] = (33328); //@line 46 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $0 = ((($this)) + 92|0); //@line 44 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $1 = ((($this)) + 100|0); //@line 44 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $2 = ((($this)) + 108|0); //@line 44 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $3 = ((($this)) + 116|0); //@line 44 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $4 = ((($this)) + 124|0); //@line 44 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $5 = ((($this)) + 132|0); //@line 44 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $6 = ((($def)) + 20|0); //@line 47 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 47 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $8 = ((($this)) + 68|0); //@line 47 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$8>>2] = $7; //@line 47 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $9 = ((($def)) + 24|0); //@line 48 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 48 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $11 = ((($this)) + 72|0); //@line 48 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$11>>2] = $10; //@line 48 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $12 = $7; //@line 50 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $13 = ((($12)) + 4|0); //@line 188 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $14 = HEAP32[$13>>2]|0; //@line 188 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $15 = ((($this)) + 76|0); //@line 50 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$15>>2] = $14; //@line 50 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $$cast = $10; //@line 51 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $16 = ((($$cast)) + 4|0); //@line 188 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $17 = HEAP32[$16>>2]|0; //@line 188 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $18 = ((($this)) + 80|0); //@line 51 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$18>>2] = $17; //@line 51 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $$off = (($14) + -1)|0; //@line 53 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $19 = ($$off>>>0)<(2); //@line 53 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 __Z10loveAssertbPKc($19,447556); //@line 53 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $20 = HEAP32[$18>>2]|0; //@line 54 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $$off17 = (($20) + -1)|0; //@line 54 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $21 = ($$off17>>>0)<(2); //@line 54 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 __Z10loveAssertbPKc($21,447614); //@line 54 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $22 = HEAP32[$8>>2]|0; //@line 60 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $23 = ((($22)) + 48|0); //@line 193 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $24 = HEAP32[$23>>2]|0; //@line 193 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $25 = ((($this)) + 84|0); //@line 60 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$25>>2] = $24; //@line 60 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $26 = ((($22)) + 52|0); //@line 198 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $27 = HEAP32[$26>>2]|0; //@line 198 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $28 = ((($this)) + 48|0); //@line 61 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$28>>2] = $27; //@line 61 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $29 = ((($27)) + 20|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $30 = +HEAPF32[$29>>2]; //@line 64 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $31 = ((($27)) + 24|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $32 = +HEAPF32[$31>>2]; //@line 64 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $33 = $24; //@line 66 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $34 = ((($33)) + 20|0); //@line 66 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $35 = +HEAPF32[$34>>2]; //@line 66 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $36 = ((($33)) + 24|0); //@line 66 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $37 = +HEAPF32[$36>>2]; //@line 66 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $38 = HEAP32[$15>>2]|0; //@line 69 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $39 = ($38|0)==(1); //@line 69 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 if ($39) {
  $40 = ((($27)) + 56|0); //@line 65 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $41 = +HEAPF32[$40>>2]; //@line 65 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $42 = ((($33)) + 56|0); //@line 67 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $43 = +HEAPF32[$42>>2]; //@line 67 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $44 = HEAP32[$6>>2]|0; //@line 71 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $45 = ((($44)) + 68|0); //@line 72 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $46 = $45; //@line 72 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $47 = $46; //@line 72 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $48 = HEAP32[$47>>2]|0; //@line 72 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $49 = (($46) + 4)|0; //@line 72 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $50 = $49; //@line 72 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $51 = HEAP32[$50>>2]|0; //@line 72 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $52 = $2; //@line 72 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $53 = $52; //@line 72 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$53>>2] = $48; //@line 72 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $54 = (($52) + 4)|0; //@line 72 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $55 = $54; //@line 72 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$55>>2] = $51; //@line 72 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $56 = ((($44)) + 76|0); //@line 73 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $57 = $56; //@line 73 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $58 = $57; //@line 73 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $59 = HEAP32[$58>>2]|0; //@line 73 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $60 = (($57) + 4)|0; //@line 73 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $61 = $60; //@line 73 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $62 = HEAP32[$61>>2]|0; //@line 73 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $63 = $0; //@line 73 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $64 = $63; //@line 73 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$64>>2] = $59; //@line 73 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $65 = (($63) + 4)|0; //@line 73 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $66 = $65; //@line 73 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$66>>2] = $62; //@line 73 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $67 = ((($44)) + 116|0); //@line 74 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $68 = HEAP32[$67>>2]|0; //@line 74 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $69 = ((($this)) + 140|0); //@line 74 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$69>>2] = $68; //@line 74 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$4>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $70 = ((($this)) + 128|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$70>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $71 = $41 - $43; //@line 77 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $$cast19 = (HEAP32[tempDoublePtr>>2]=$68,+HEAPF32[tempDoublePtr>>2]); //@line 77 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $72 = $71 - $$cast19; //@line 77 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $coordinateA$0 = $72;
 } else {
  $73 = ((($33)) + 16|0); //@line 66 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $74 = +HEAPF32[$73>>2]; //@line 66 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $75 = ((($33)) + 12|0); //@line 66 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $76 = +HEAPF32[$75>>2]; //@line 66 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $77 = ((($27)) + 16|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $78 = +HEAPF32[$77>>2]; //@line 64 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $79 = ((($27)) + 12|0); //@line 64 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $80 = +HEAPF32[$79>>2]; //@line 64 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $81 = HEAP32[$6>>2]|0; //@line 81 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $82 = ((($81)) + 68|0); //@line 82 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $83 = $82; //@line 82 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $84 = $83; //@line 82 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $85 = HEAP32[$84>>2]|0; //@line 82 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $86 = (($83) + 4)|0; //@line 82 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $87 = $86; //@line 82 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $88 = HEAP32[$87>>2]|0; //@line 82 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $89 = $2; //@line 82 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $90 = $89; //@line 82 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$90>>2] = $85; //@line 82 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $91 = (($89) + 4)|0; //@line 82 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $92 = $91; //@line 82 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$92>>2] = $88; //@line 82 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $93 = ((($81)) + 76|0); //@line 83 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $94 = $93; //@line 83 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $95 = $94; //@line 83 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $96 = HEAP32[$95>>2]|0; //@line 83 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $97 = (($94) + 4)|0; //@line 83 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $98 = $97; //@line 83 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $99 = HEAP32[$98>>2]|0; //@line 83 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $100 = $0; //@line 83 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $101 = $100; //@line 83 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$101>>2] = $96; //@line 83 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $102 = (($100) + 4)|0; //@line 83 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $103 = $102; //@line 83 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$103>>2] = $99; //@line 83 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $104 = ((($81)) + 100|0); //@line 84 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $105 = HEAP32[$104>>2]|0; //@line 84 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $106 = ((($this)) + 140|0); //@line 84 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$106>>2] = $105; //@line 84 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $107 = ((($81)) + 84|0); //@line 85 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $108 = $107; //@line 85 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $109 = $108; //@line 85 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $110 = HEAP32[$109>>2]|0; //@line 85 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $111 = (($108) + 4)|0; //@line 85 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $112 = $111; //@line 85 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $113 = HEAP32[$112>>2]|0; //@line 85 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $114 = $4; //@line 85 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $115 = $114; //@line 85 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$115>>2] = $110; //@line 85 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $116 = (($114) + 4)|0; //@line 85 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $117 = $116; //@line 85 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$117>>2] = $113; //@line 85 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $bitcast = (HEAP32[tempDoublePtr>>2]=$85,+HEAPF32[tempDoublePtr>>2]); //@line 87 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $118 = (HEAP32[tempDoublePtr>>2]=$88,+HEAPF32[tempDoublePtr>>2]); //@line 87 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $bitcast21 = (HEAP32[tempDoublePtr>>2]=$96,+HEAPF32[tempDoublePtr>>2]); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $119 = $32 * $bitcast21; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $120 = (HEAP32[tempDoublePtr>>2]=$99,+HEAPF32[tempDoublePtr>>2]); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $121 = $30 * $120; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $122 = $119 - $121; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $123 = $30 * $bitcast21; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $124 = $32 * $120; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $125 = $123 + $124; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $126 = $80 - $76; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $127 = $78 - $74; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $128 = $126 + $122; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $129 = $127 + $125; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $130 = $37 * $128; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $131 = $35 * $129; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $132 = $130 + $131; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $133 = $35 * $128; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $134 = $37 * $129; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $135 = $134 - $133; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $136 = $132 - $bitcast; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $137 = $135 - $118; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $bitcast23 = (HEAP32[tempDoublePtr>>2]=$110,+HEAPF32[tempDoublePtr>>2]); //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $138 = $bitcast23 * $136; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $139 = (HEAP32[tempDoublePtr>>2]=$113,+HEAPF32[tempDoublePtr>>2]); //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $140 = $139 * $137; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $141 = $138 + $140; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $coordinateA$0 = $141;
 }
 $142 = HEAP32[$11>>2]|0; //@line 92 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $143 = ((($142)) + 48|0); //@line 193 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $144 = HEAP32[$143>>2]|0; //@line 193 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $145 = ((($this)) + 88|0); //@line 92 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$145>>2] = $144; //@line 92 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $146 = ((($142)) + 52|0); //@line 198 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $147 = HEAP32[$146>>2]|0; //@line 198 "./libraries/Box2D/Dynamics/Joints/b2Joint.h"
 $148 = ((($this)) + 52|0); //@line 93 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAP32[$148>>2] = $147; //@line 93 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $149 = ((($147)) + 20|0); //@line 96 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $150 = +HEAPF32[$149>>2]; //@line 96 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $151 = ((($147)) + 24|0); //@line 96 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $152 = +HEAPF32[$151>>2]; //@line 96 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $153 = $144; //@line 98 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $154 = ((($153)) + 20|0); //@line 98 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $155 = +HEAPF32[$154>>2]; //@line 98 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $156 = ((($153)) + 24|0); //@line 98 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $157 = +HEAPF32[$156>>2]; //@line 98 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $158 = HEAP32[$18>>2]|0; //@line 101 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $159 = ($158|0)==(1); //@line 101 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 if ($159) {
  $160 = ((($147)) + 56|0); //@line 97 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $161 = +HEAPF32[$160>>2]; //@line 97 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $162 = ((($153)) + 56|0); //@line 99 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $163 = +HEAPF32[$162>>2]; //@line 99 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $164 = HEAP32[$9>>2]|0; //@line 103 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $165 = ((($164)) + 68|0); //@line 104 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $166 = $165; //@line 104 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $167 = $166; //@line 104 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $168 = HEAP32[$167>>2]|0; //@line 104 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $169 = (($166) + 4)|0; //@line 104 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $170 = $169; //@line 104 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $171 = HEAP32[$170>>2]|0; //@line 104 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $172 = $3; //@line 104 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $173 = $172; //@line 104 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$173>>2] = $168; //@line 104 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $174 = (($172) + 4)|0; //@line 104 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $175 = $174; //@line 104 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$175>>2] = $171; //@line 104 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $176 = ((($164)) + 76|0); //@line 105 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $177 = $176; //@line 105 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $178 = $177; //@line 105 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $179 = HEAP32[$178>>2]|0; //@line 105 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $180 = (($177) + 4)|0; //@line 105 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $181 = $180; //@line 105 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $182 = HEAP32[$181>>2]|0; //@line 105 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $183 = $1; //@line 105 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $184 = $183; //@line 105 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$184>>2] = $179; //@line 105 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $185 = (($183) + 4)|0; //@line 105 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $186 = $185; //@line 105 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$186>>2] = $182; //@line 105 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $187 = ((($164)) + 116|0); //@line 106 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $188 = HEAP32[$187>>2]|0; //@line 106 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $189 = ((($this)) + 144|0); //@line 106 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$189>>2] = $188; //@line 106 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$5>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $190 = ((($this)) + 136|0); //@line 62 "./libraries/Box2D/Common/b2Math.h"
  HEAPF32[$190>>2] = 0.0; //@line 62 "./libraries/Box2D/Common/b2Math.h"
  $191 = $161 - $163; //@line 109 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $$cast18 = (HEAP32[tempDoublePtr>>2]=$188,+HEAPF32[tempDoublePtr>>2]); //@line 109 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $192 = $191 - $$cast18; //@line 109 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $coordinateB$0 = $192;
  $262 = ((($def)) + 28|0); //@line 124 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $263 = +HEAPF32[$262>>2]; //@line 124 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $264 = ((($this)) + 152|0); //@line 124 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$264>>2] = $263; //@line 124 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $265 = $coordinateB$0 * $263; //@line 126 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $266 = $coordinateA$0 + $265; //@line 126 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $267 = ((($this)) + 148|0); //@line 126 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$267>>2] = $266; //@line 126 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $268 = ((($this)) + 156|0); //@line 128 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$268>>2] = 0.0; //@line 128 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  return; //@line 129 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 } else {
  $193 = ((($153)) + 16|0); //@line 98 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $194 = +HEAPF32[$193>>2]; //@line 98 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $195 = ((($153)) + 12|0); //@line 98 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $196 = +HEAPF32[$195>>2]; //@line 98 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $197 = ((($147)) + 16|0); //@line 96 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $198 = +HEAPF32[$197>>2]; //@line 96 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $199 = ((($147)) + 12|0); //@line 96 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $200 = +HEAPF32[$199>>2]; //@line 96 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $201 = HEAP32[$9>>2]|0; //@line 113 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $202 = ((($201)) + 68|0); //@line 114 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $203 = $202; //@line 114 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $204 = $203; //@line 114 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $205 = HEAP32[$204>>2]|0; //@line 114 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $206 = (($203) + 4)|0; //@line 114 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $207 = $206; //@line 114 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $208 = HEAP32[$207>>2]|0; //@line 114 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $209 = $3; //@line 114 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $210 = $209; //@line 114 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$210>>2] = $205; //@line 114 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $211 = (($209) + 4)|0; //@line 114 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $212 = $211; //@line 114 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$212>>2] = $208; //@line 114 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $213 = ((($201)) + 76|0); //@line 115 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $214 = $213; //@line 115 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $215 = $214; //@line 115 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $216 = HEAP32[$215>>2]|0; //@line 115 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $217 = (($214) + 4)|0; //@line 115 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $218 = $217; //@line 115 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $219 = HEAP32[$218>>2]|0; //@line 115 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $220 = $1; //@line 115 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $221 = $220; //@line 115 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$221>>2] = $216; //@line 115 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $222 = (($220) + 4)|0; //@line 115 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $223 = $222; //@line 115 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$223>>2] = $219; //@line 115 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $224 = ((($201)) + 100|0); //@line 116 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $225 = HEAP32[$224>>2]|0; //@line 116 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $226 = ((($this)) + 144|0); //@line 116 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$226>>2] = $225; //@line 116 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $227 = ((($201)) + 84|0); //@line 117 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $228 = $227; //@line 117 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $229 = $228; //@line 117 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $230 = HEAP32[$229>>2]|0; //@line 117 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $231 = (($228) + 4)|0; //@line 117 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $232 = $231; //@line 117 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $233 = HEAP32[$232>>2]|0; //@line 117 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $234 = $5; //@line 117 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $235 = $234; //@line 117 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$235>>2] = $230; //@line 117 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $236 = (($234) + 4)|0; //@line 117 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $237 = $236; //@line 117 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAP32[$237>>2] = $233; //@line 117 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $bitcast25 = (HEAP32[tempDoublePtr>>2]=$205,+HEAPF32[tempDoublePtr>>2]); //@line 119 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $238 = (HEAP32[tempDoublePtr>>2]=$208,+HEAPF32[tempDoublePtr>>2]); //@line 119 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $bitcast27 = (HEAP32[tempDoublePtr>>2]=$216,+HEAPF32[tempDoublePtr>>2]); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $239 = $152 * $bitcast27; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $240 = (HEAP32[tempDoublePtr>>2]=$219,+HEAPF32[tempDoublePtr>>2]); //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $241 = $150 * $240; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $242 = $239 - $241; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $243 = $150 * $bitcast27; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $244 = $152 * $240; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $245 = $243 + $244; //@line 568 "./libraries/Box2D/Common/b2Math.h"
  $246 = $200 - $196; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $247 = $198 - $194; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $248 = $246 + $242; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $249 = $247 + $245; //@line 448 "./libraries/Box2D/Common/b2Math.h"
  $250 = $157 * $248; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $251 = $155 * $249; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $252 = $250 + $251; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $253 = $155 * $248; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $254 = $157 * $249; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $255 = $254 - $253; //@line 574 "./libraries/Box2D/Common/b2Math.h"
  $256 = $252 - $bitcast25; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $257 = $255 - $238; //@line 454 "./libraries/Box2D/Common/b2Math.h"
  $bitcast29 = (HEAP32[tempDoublePtr>>2]=$230,+HEAPF32[tempDoublePtr>>2]); //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $258 = $bitcast29 * $256; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $259 = (HEAP32[tempDoublePtr>>2]=$233,+HEAPF32[tempDoublePtr>>2]); //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $260 = $259 * $257; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $261 = $258 + $260; //@line 408 "./libraries/Box2D/Common/b2Math.h"
  $coordinateB$0 = $261;
  $262 = ((($def)) + 28|0); //@line 124 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $263 = +HEAPF32[$262>>2]; //@line 124 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $264 = ((($this)) + 152|0); //@line 124 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$264>>2] = $263; //@line 124 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $265 = $coordinateB$0 * $263; //@line 126 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $266 = $coordinateA$0 + $265; //@line 126 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $267 = ((($this)) + 148|0); //@line 126 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$267>>2] = $266; //@line 126 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  $268 = ((($this)) + 156|0); //@line 128 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  HEAPF32[$268>>2] = 0.0; //@line 128 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
  return; //@line 129 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 }
}
function __ZN11b2GearJoint8SetRatioEf($this,$ratio) {
 $this = $this|0;
 $ratio = +$ratio;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (HEAPF32[tempDoublePtr>>2]=$ratio,HEAP32[tempDoublePtr>>2]|0); //@line 394 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $1 = $0 & 2139095040; //@line 29 "./libraries/Box2D/Common/b2Math.h"
 $2 = ($1|0)!=(2139095040); //@line 29 "./libraries/Box2D/Common/b2Math.h"
 __Z10loveAssertbPKc($2,447672); //@line 394 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $3 = ((($this)) + 152|0); //@line 395 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 HEAPF32[$3>>2] = $ratio; //@line 395 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 return; //@line 396 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
}
function __ZNK11b2GearJoint8GetRatioEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 152|0); //@line 400 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 400 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
 return (+$1); //@line 400 "libraries/Box2D/Dynamics/Joints/b2GearJoint.cpp"
}
function _wuff_int8_to_int8($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $0;
 $7 = $1;
 $8 = $2;
 $9 = $4;
 $10 = $9&255;
 $11 = (($8) + ($10))|0;
 $12 = $5;
 $13 = $12&255;
 $14 = (($11) + ($13))|0;
 _memcpy(($6|0),($7|0),($14|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int8_to_int16($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $i = 0, $i16 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $i16 = sp + 16|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  $10 = HEAP8[$9>>0]|0;
  $11 = $10&255;
  $12 = (($11) - 128)|0;
  $13 = $12 << 8;
  $14 = $13&65535;
  HEAP16[$i16>>1] = $14;
  $15 = $0;
  $16 = $3;
  $17 = $16&255;
  $18 = (($i16) + ($17)|0);
  $19 = $4;
  $20 = $19&255;
  _memcpy(($15|0),($18|0),($20|0))|0;
  $21 = $1;
  $22 = ((($21)) + 1|0);
  $1 = $22;
  $23 = $4;
  $24 = $23&255;
  $25 = $0;
  $26 = (($25) + ($24)|0);
  $0 = $26;
 }
 $i = 0;
 while(1) {
  $27 = $i;
  $28 = $2;
  $29 = ($27>>>0)<($28>>>0);
  if (!($29)) {
   break;
  }
  $30 = $i;
  $31 = $1;
  $32 = (($31) + ($30)|0);
  $33 = HEAP8[$32>>0]|0;
  $34 = $33&255;
  $35 = (($34) - 128)|0;
  $36 = $35 << 8;
  $37 = $36&65535;
  HEAP16[$i16>>1] = $37;
  $38 = $0;
  $39 = $i;
  $40 = $39<<1;
  $41 = (($38) + ($40)|0);
  ;HEAP8[$41>>0]=HEAP8[$i16>>0]|0;HEAP8[$41+1>>0]=HEAP8[$i16+1>>0]|0;
  $42 = $i;
  $43 = (($42) + 1)|0;
  $i = $43;
 }
 $44 = $5;
 $45 = $44&255;
 $46 = ($45|0)!=(0);
 if (!($46)) {
  STACKTOP = sp;return;
 }
 $47 = $2;
 $48 = $1;
 $49 = (($48) + ($47)|0);
 $50 = HEAP8[$49>>0]|0;
 $51 = $50&255;
 $52 = (($51) - 128)|0;
 $53 = $52 << 8;
 $54 = $53&65535;
 HEAP16[$i16>>1] = $54;
 $55 = $0;
 $56 = $2;
 $57 = $56<<1;
 $58 = (($55) + ($57)|0);
 $59 = $5;
 $60 = $59&255;
 _memcpy(($58|0),($i16|0),($60|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int8_to_int24($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $i = 0, $i24 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $i24 = sp + 4|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  $10 = HEAP8[$9>>0]|0;
  $11 = $10&255;
  $12 = (($11) - 128)|0;
  $13 = $12 << 24;
  HEAP32[$i24>>2] = $13;
  $14 = $0;
  $15 = ((($i24)) + 1|0);
  $16 = $3;
  $17 = $16&255;
  $18 = (($15) + ($17)|0);
  $19 = $4;
  $20 = $19&255;
  _memcpy(($14|0),($18|0),($20|0))|0;
  $21 = $1;
  $22 = ((($21)) + 1|0);
  $1 = $22;
  $23 = $4;
  $24 = $23&255;
  $25 = $0;
  $26 = (($25) + ($24)|0);
  $0 = $26;
 }
 $i = 0;
 while(1) {
  $27 = $i;
  $28 = $2;
  $29 = ($27>>>0)<($28>>>0);
  if (!($29)) {
   break;
  }
  $30 = $i;
  $31 = $1;
  $32 = (($31) + ($30)|0);
  $33 = HEAP8[$32>>0]|0;
  $34 = $33&255;
  $35 = (($34) - 128)|0;
  $36 = $35 << 24;
  HEAP32[$i24>>2] = $36;
  $37 = $0;
  $38 = $i;
  $39 = ($38*3)|0;
  $40 = (($37) + ($39)|0);
  $41 = ((($i24)) + 1|0);
  ;HEAP8[$40>>0]=HEAP8[$41>>0]|0;HEAP8[$40+1>>0]=HEAP8[$41+1>>0]|0;HEAP8[$40+2>>0]=HEAP8[$41+2>>0]|0;
  $42 = $i;
  $43 = (($42) + 1)|0;
  $i = $43;
 }
 $44 = $5;
 $45 = $44&255;
 $46 = ($45|0)!=(0);
 if (!($46)) {
  STACKTOP = sp;return;
 }
 $47 = $2;
 $48 = $1;
 $49 = (($48) + ($47)|0);
 $50 = HEAP8[$49>>0]|0;
 $51 = $50&255;
 $52 = (($51) - 128)|0;
 $53 = $52 << 24;
 HEAP32[$i24>>2] = $53;
 $54 = $0;
 $55 = $2;
 $56 = ($55*3)|0;
 $57 = (($54) + ($56)|0);
 $58 = ((($i24)) + 1|0);
 $59 = $5;
 $60 = $59&255;
 _memcpy(($57|0),($58|0),($60|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int8_to_int32($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0, $i32 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $i32 = sp + 4|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  $10 = HEAP8[$9>>0]|0;
  $11 = $10&255;
  $12 = (($11) - 128)|0;
  $13 = $12 << 24;
  HEAP32[$i32>>2] = $13;
  $14 = $0;
  $15 = $3;
  $16 = $15&255;
  $17 = (($i32) + ($16)|0);
  $18 = $4;
  $19 = $18&255;
  _memcpy(($14|0),($17|0),($19|0))|0;
  $20 = $1;
  $21 = ((($20)) + 1|0);
  $1 = $21;
  $22 = $4;
  $23 = $22&255;
  $24 = $0;
  $25 = (($24) + ($23)|0);
  $0 = $25;
 }
 $i = 0;
 while(1) {
  $26 = $i;
  $27 = $2;
  $28 = ($26>>>0)<($27>>>0);
  if (!($28)) {
   break;
  }
  $29 = $i;
  $30 = $1;
  $31 = (($30) + ($29)|0);
  $32 = HEAP8[$31>>0]|0;
  $33 = $32&255;
  $34 = (($33) - 128)|0;
  $35 = $34 << 24;
  HEAP32[$i32>>2] = $35;
  $36 = $0;
  $37 = $i;
  $38 = $37<<2;
  $39 = (($36) + ($38)|0);
  ;HEAP8[$39>>0]=HEAP8[$i32>>0]|0;HEAP8[$39+1>>0]=HEAP8[$i32+1>>0]|0;HEAP8[$39+2>>0]=HEAP8[$i32+2>>0]|0;HEAP8[$39+3>>0]=HEAP8[$i32+3>>0]|0;
  $40 = $i;
  $41 = (($40) + 1)|0;
  $i = $41;
 }
 $42 = $5;
 $43 = $42&255;
 $44 = ($43|0)!=(0);
 if (!($44)) {
  STACKTOP = sp;return;
 }
 $45 = $2;
 $46 = $1;
 $47 = (($46) + ($45)|0);
 $48 = HEAP8[$47>>0]|0;
 $49 = $48&255;
 $50 = (($49) - 128)|0;
 $51 = $50 << 24;
 HEAP32[$i32>>2] = $51;
 $52 = $0;
 $53 = $2;
 $54 = $53<<2;
 $55 = (($52) + ($54)|0);
 $56 = $5;
 $57 = $56&255;
 _memcpy(($55|0),($i32|0),($57|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int8_to_float32($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0.0, $54 = 0.0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $f32 = 0, $i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f32 = sp + 4|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  $10 = HEAP8[$9>>0]|0;
  $11 = $10&255;
  $12 = (($11) - 128)|0;
  $13 = (+($12|0));
  $14 = $13 / 128.0;
  HEAPF32[$f32>>2] = $14;
  $15 = $0;
  $16 = $3;
  $17 = $16&255;
  $18 = (($f32) + ($17)|0);
  $19 = $4;
  $20 = $19&255;
  _memcpy(($15|0),($18|0),($20|0))|0;
  $21 = $1;
  $22 = ((($21)) + 1|0);
  $1 = $22;
  $23 = $4;
  $24 = $23&255;
  $25 = $0;
  $26 = (($25) + ($24)|0);
  $0 = $26;
 }
 $i = 0;
 while(1) {
  $27 = $i;
  $28 = $2;
  $29 = ($27>>>0)<($28>>>0);
  if (!($29)) {
   break;
  }
  $30 = $i;
  $31 = $1;
  $32 = (($31) + ($30)|0);
  $33 = HEAP8[$32>>0]|0;
  $34 = $33&255;
  $35 = (($34) - 128)|0;
  $36 = (+($35|0));
  $37 = $36 / 128.0;
  HEAPF32[$f32>>2] = $37;
  $38 = $0;
  $39 = $i;
  $40 = $39<<2;
  $41 = (($38) + ($40)|0);
  ;HEAP8[$41>>0]=HEAP8[$f32>>0]|0;HEAP8[$41+1>>0]=HEAP8[$f32+1>>0]|0;HEAP8[$41+2>>0]=HEAP8[$f32+2>>0]|0;HEAP8[$41+3>>0]=HEAP8[$f32+3>>0]|0;
  $42 = $i;
  $43 = (($42) + 1)|0;
  $i = $43;
 }
 $44 = $5;
 $45 = $44&255;
 $46 = ($45|0)!=(0);
 if (!($46)) {
  STACKTOP = sp;return;
 }
 $47 = $2;
 $48 = $1;
 $49 = (($48) + ($47)|0);
 $50 = HEAP8[$49>>0]|0;
 $51 = $50&255;
 $52 = (($51) - 128)|0;
 $53 = (+($52|0));
 $54 = $53 / 128.0;
 HEAPF32[$f32>>2] = $54;
 $55 = $0;
 $56 = $2;
 $57 = $56<<2;
 $58 = (($55) + ($57)|0);
 $59 = $5;
 $60 = $59&255;
 _memcpy(($58|0),($f32|0),($60|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int8_to_float64($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0.0, $54 = 0.0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $f64 = 0, $i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f64 = sp;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  $10 = HEAP8[$9>>0]|0;
  $11 = $10&255;
  $12 = (($11) - 128)|0;
  $13 = (+($12|0));
  $14 = $13 / 128.0;
  HEAPF64[$f64>>3] = $14;
  $15 = $0;
  $16 = $3;
  $17 = $16&255;
  $18 = (($f64) + ($17)|0);
  $19 = $4;
  $20 = $19&255;
  _memcpy(($15|0),($18|0),($20|0))|0;
  $21 = $1;
  $22 = ((($21)) + 1|0);
  $1 = $22;
  $23 = $4;
  $24 = $23&255;
  $25 = $0;
  $26 = (($25) + ($24)|0);
  $0 = $26;
 }
 $i = 0;
 while(1) {
  $27 = $i;
  $28 = $2;
  $29 = ($27>>>0)<($28>>>0);
  if (!($29)) {
   break;
  }
  $30 = $i;
  $31 = $1;
  $32 = (($31) + ($30)|0);
  $33 = HEAP8[$32>>0]|0;
  $34 = $33&255;
  $35 = (($34) - 128)|0;
  $36 = (+($35|0));
  $37 = $36 / 128.0;
  HEAPF64[$f64>>3] = $37;
  $38 = $0;
  $39 = $i;
  $40 = $39<<3;
  $41 = (($38) + ($40)|0);
  ;HEAP8[$41>>0]=HEAP8[$f64>>0]|0;HEAP8[$41+1>>0]=HEAP8[$f64+1>>0]|0;HEAP8[$41+2>>0]=HEAP8[$f64+2>>0]|0;HEAP8[$41+3>>0]=HEAP8[$f64+3>>0]|0;HEAP8[$41+4>>0]=HEAP8[$f64+4>>0]|0;HEAP8[$41+5>>0]=HEAP8[$f64+5>>0]|0;HEAP8[$41+6>>0]=HEAP8[$f64+6>>0]|0;HEAP8[$41+7>>0]=HEAP8[$f64+7>>0]|0;
  $42 = $i;
  $43 = (($42) + 1)|0;
  $i = $43;
 }
 $44 = $5;
 $45 = $44&255;
 $46 = ($45|0)!=(0);
 if (!($46)) {
  STACKTOP = sp;return;
 }
 $47 = $2;
 $48 = $1;
 $49 = (($48) + ($47)|0);
 $50 = HEAP8[$49>>0]|0;
 $51 = $50&255;
 $52 = (($51) - 128)|0;
 $53 = (+($52|0));
 $54 = $53 / 128.0;
 HEAPF64[$f64>>3] = $54;
 $55 = $0;
 $56 = $2;
 $57 = $56<<3;
 $58 = (($55) + ($57)|0);
 $59 = $5;
 $60 = $59&255;
 _memcpy(($58|0),($f64|0),($60|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int16_to_int8($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $i = 0, $i16 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $i16 = sp + 16|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $i = 0;
 while(1) {
  $6 = $i;
  $7 = $2;
  $8 = ($6>>>0)<($7>>>0);
  if (!($8)) {
   break;
  }
  $9 = $1;
  $10 = $i;
  $11 = $10<<1;
  $12 = (($9) + ($11)|0);
  ;HEAP8[$i16>>0]=HEAP8[$12>>0]|0;HEAP8[$i16+1>>0]=HEAP8[$12+1>>0]|0;
  $13 = HEAP16[$i16>>1]|0;
  $14 = $13 << 16 >> 16;
  $15 = $14 >> 8;
  $16 = (($15) + 128)|0;
  $17 = $16&255;
  $18 = $i;
  $19 = $0;
  $20 = (($19) + ($18)|0);
  HEAP8[$20>>0] = $17;
  $21 = $i;
  $22 = (($21) + 1)|0;
  $i = $22;
 }
 STACKTOP = sp;return;
}
function _wuff_int16_to_int16($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $0;
 $7 = $1;
 $8 = $3;
 $9 = $8&255;
 $10 = (($7) + ($9)|0);
 $11 = $2;
 $12 = $11<<1;
 $13 = $4;
 $14 = $13&255;
 $15 = (($12) + ($14))|0;
 $16 = $5;
 $17 = $16&255;
 $18 = (($15) + ($17))|0;
 _memcpy(($6|0),($10|0),($18|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int16_to_int24($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $i = 0, $i16 = 0, $i24 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $i16 = sp + 20|0;
 $i24 = sp + 4|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  ;HEAP8[$i16>>0]=HEAP8[$9>>0]|0;HEAP8[$i16+1>>0]=HEAP8[$9+1>>0]|0;
  $10 = HEAP16[$i16>>1]|0;
  $11 = $10 << 16 >> 16;
  $12 = $11 << 16;
  HEAP32[$i24>>2] = $12;
  $13 = $0;
  $14 = ((($i24)) + 1|0);
  $15 = $3;
  $16 = $15&255;
  $17 = (($14) + ($16)|0);
  $18 = $4;
  $19 = $18&255;
  _memcpy(($13|0),($17|0),($19|0))|0;
  $20 = $1;
  $21 = ((($20)) + 2|0);
  $1 = $21;
  $22 = $4;
  $23 = $22&255;
  $24 = $0;
  $25 = (($24) + ($23)|0);
  $0 = $25;
 }
 $i = 0;
 while(1) {
  $26 = $i;
  $27 = $2;
  $28 = ($26>>>0)<($27>>>0);
  if (!($28)) {
   break;
  }
  $29 = $1;
  $30 = $i;
  $31 = $30<<1;
  $32 = (($29) + ($31)|0);
  ;HEAP8[$i16>>0]=HEAP8[$32>>0]|0;HEAP8[$i16+1>>0]=HEAP8[$32+1>>0]|0;
  $33 = HEAP16[$i16>>1]|0;
  $34 = $33 << 16 >> 16;
  $35 = $34 << 16;
  HEAP32[$i24>>2] = $35;
  $36 = $0;
  $37 = $i;
  $38 = ($37*3)|0;
  $39 = (($36) + ($38)|0);
  $40 = ((($i24)) + 1|0);
  ;HEAP8[$39>>0]=HEAP8[$40>>0]|0;HEAP8[$39+1>>0]=HEAP8[$40+1>>0]|0;HEAP8[$39+2>>0]=HEAP8[$40+2>>0]|0;
  $41 = $i;
  $42 = (($41) + 1)|0;
  $i = $42;
 }
 $43 = $5;
 $44 = $43&255;
 $45 = ($44|0)!=(0);
 if (!($45)) {
  STACKTOP = sp;return;
 }
 $46 = $1;
 $47 = $2;
 $48 = $47<<1;
 $49 = (($46) + ($48)|0);
 ;HEAP8[$i16>>0]=HEAP8[$49>>0]|0;HEAP8[$i16+1>>0]=HEAP8[$49+1>>0]|0;
 $50 = HEAP16[$i16>>1]|0;
 $51 = $50 << 16 >> 16;
 $52 = $51 << 16;
 HEAP32[$i24>>2] = $52;
 $53 = $0;
 $54 = $2;
 $55 = ($54*3)|0;
 $56 = (($53) + ($55)|0);
 $57 = ((($i24)) + 1|0);
 $58 = $5;
 $59 = $58&255;
 _memcpy(($56|0),($57|0),($59|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int16_to_int32($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0, $i16 = 0, $i32 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $i16 = sp + 20|0;
 $i32 = sp + 4|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  ;HEAP8[$i16>>0]=HEAP8[$9>>0]|0;HEAP8[$i16+1>>0]=HEAP8[$9+1>>0]|0;
  $10 = HEAP16[$i16>>1]|0;
  $11 = $10 << 16 >> 16;
  $12 = $11 << 16;
  HEAP32[$i32>>2] = $12;
  $13 = $0;
  $14 = $3;
  $15 = $14&255;
  $16 = (($i32) + ($15)|0);
  $17 = $4;
  $18 = $17&255;
  _memcpy(($13|0),($16|0),($18|0))|0;
  $19 = $1;
  $20 = ((($19)) + 2|0);
  $1 = $20;
  $21 = $4;
  $22 = $21&255;
  $23 = $0;
  $24 = (($23) + ($22)|0);
  $0 = $24;
 }
 $i = 0;
 while(1) {
  $25 = $i;
  $26 = $2;
  $27 = ($25>>>0)<($26>>>0);
  if (!($27)) {
   break;
  }
  $28 = $1;
  $29 = $i;
  $30 = $29<<1;
  $31 = (($28) + ($30)|0);
  ;HEAP8[$i16>>0]=HEAP8[$31>>0]|0;HEAP8[$i16+1>>0]=HEAP8[$31+1>>0]|0;
  $32 = HEAP16[$i16>>1]|0;
  $33 = $32 << 16 >> 16;
  $34 = $33 << 16;
  HEAP32[$i32>>2] = $34;
  $35 = $0;
  $36 = $i;
  $37 = $36<<2;
  $38 = (($35) + ($37)|0);
  ;HEAP8[$38>>0]=HEAP8[$i32>>0]|0;HEAP8[$38+1>>0]=HEAP8[$i32+1>>0]|0;HEAP8[$38+2>>0]=HEAP8[$i32+2>>0]|0;HEAP8[$38+3>>0]=HEAP8[$i32+3>>0]|0;
  $39 = $i;
  $40 = (($39) + 1)|0;
  $i = $40;
 }
 $41 = $5;
 $42 = $41&255;
 $43 = ($42|0)!=(0);
 if (!($43)) {
  STACKTOP = sp;return;
 }
 $44 = $1;
 $45 = $2;
 $46 = $45<<1;
 $47 = (($44) + ($46)|0);
 ;HEAP8[$i16>>0]=HEAP8[$47>>0]|0;HEAP8[$i16+1>>0]=HEAP8[$47+1>>0]|0;
 $48 = HEAP16[$i16>>1]|0;
 $49 = $48 << 16 >> 16;
 $50 = $49 << 16;
 HEAP32[$i32>>2] = $50;
 $51 = $0;
 $52 = $2;
 $53 = $52<<2;
 $54 = (($51) + ($53)|0);
 $55 = $5;
 $56 = $55&255;
 _memcpy(($54|0),($i32|0),($56|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int16_to_float32($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f32 = 0, $i = 0, $i16 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $i16 = sp + 20|0;
 $f32 = sp + 4|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  ;HEAP8[$i16>>0]=HEAP8[$9>>0]|0;HEAP8[$i16+1>>0]=HEAP8[$9+1>>0]|0;
  $10 = HEAP16[$i16>>1]|0;
  $11 = (+($10<<16>>16));
  $12 = $11 / 32768.0;
  HEAPF32[$f32>>2] = $12;
  $13 = $0;
  $14 = $3;
  $15 = $14&255;
  $16 = (($f32) + ($15)|0);
  $17 = $4;
  $18 = $17&255;
  _memcpy(($13|0),($16|0),($18|0))|0;
  $19 = $1;
  $20 = ((($19)) + 2|0);
  $1 = $20;
  $21 = $4;
  $22 = $21&255;
  $23 = $0;
  $24 = (($23) + ($22)|0);
  $0 = $24;
 }
 $i = 0;
 while(1) {
  $25 = $i;
  $26 = $2;
  $27 = ($25>>>0)<($26>>>0);
  if (!($27)) {
   break;
  }
  $28 = $1;
  $29 = $i;
  $30 = $29<<1;
  $31 = (($28) + ($30)|0);
  ;HEAP8[$i16>>0]=HEAP8[$31>>0]|0;HEAP8[$i16+1>>0]=HEAP8[$31+1>>0]|0;
  $32 = HEAP16[$i16>>1]|0;
  $33 = (+($32<<16>>16));
  $34 = $33 / 32768.0;
  HEAPF32[$f32>>2] = $34;
  $35 = $0;
  $36 = $i;
  $37 = $36<<2;
  $38 = (($35) + ($37)|0);
  ;HEAP8[$38>>0]=HEAP8[$f32>>0]|0;HEAP8[$38+1>>0]=HEAP8[$f32+1>>0]|0;HEAP8[$38+2>>0]=HEAP8[$f32+2>>0]|0;HEAP8[$38+3>>0]=HEAP8[$f32+3>>0]|0;
  $39 = $i;
  $40 = (($39) + 1)|0;
  $i = $40;
 }
 $41 = $5;
 $42 = $41&255;
 $43 = ($42|0)!=(0);
 if (!($43)) {
  STACKTOP = sp;return;
 }
 $44 = $1;
 $45 = $2;
 $46 = $45<<1;
 $47 = (($44) + ($46)|0);
 ;HEAP8[$i16>>0]=HEAP8[$47>>0]|0;HEAP8[$i16+1>>0]=HEAP8[$47+1>>0]|0;
 $48 = HEAP16[$i16>>1]|0;
 $49 = (+($48<<16>>16));
 $50 = $49 / 32768.0;
 HEAPF32[$f32>>2] = $50;
 $51 = $0;
 $52 = $2;
 $53 = $52<<2;
 $54 = (($51) + ($53)|0);
 $55 = $5;
 $56 = $55&255;
 _memcpy(($54|0),($f32|0),($56|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int16_to_float64($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f64 = 0, $i = 0, $i16 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $i16 = sp + 24|0;
 $f64 = sp;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  ;HEAP8[$i16>>0]=HEAP8[$9>>0]|0;HEAP8[$i16+1>>0]=HEAP8[$9+1>>0]|0;
  $10 = HEAP16[$i16>>1]|0;
  $11 = (+($10<<16>>16));
  $12 = $11 / 32768.0;
  HEAPF64[$f64>>3] = $12;
  $13 = $0;
  $14 = $3;
  $15 = $14&255;
  $16 = (($f64) + ($15)|0);
  $17 = $4;
  $18 = $17&255;
  _memcpy(($13|0),($16|0),($18|0))|0;
  $19 = $1;
  $20 = ((($19)) + 2|0);
  $1 = $20;
  $21 = $4;
  $22 = $21&255;
  $23 = $0;
  $24 = (($23) + ($22)|0);
  $0 = $24;
 }
 $i = 0;
 while(1) {
  $25 = $i;
  $26 = $2;
  $27 = ($25>>>0)<($26>>>0);
  if (!($27)) {
   break;
  }
  $28 = $1;
  $29 = $i;
  $30 = $29<<1;
  $31 = (($28) + ($30)|0);
  ;HEAP8[$i16>>0]=HEAP8[$31>>0]|0;HEAP8[$i16+1>>0]=HEAP8[$31+1>>0]|0;
  $32 = HEAP16[$i16>>1]|0;
  $33 = (+($32<<16>>16));
  $34 = $33 / 32768.0;
  HEAPF64[$f64>>3] = $34;
  $35 = $0;
  $36 = $i;
  $37 = $36<<3;
  $38 = (($35) + ($37)|0);
  ;HEAP8[$38>>0]=HEAP8[$f64>>0]|0;HEAP8[$38+1>>0]=HEAP8[$f64+1>>0]|0;HEAP8[$38+2>>0]=HEAP8[$f64+2>>0]|0;HEAP8[$38+3>>0]=HEAP8[$f64+3>>0]|0;HEAP8[$38+4>>0]=HEAP8[$f64+4>>0]|0;HEAP8[$38+5>>0]=HEAP8[$f64+5>>0]|0;HEAP8[$38+6>>0]=HEAP8[$f64+6>>0]|0;HEAP8[$38+7>>0]=HEAP8[$f64+7>>0]|0;
  $39 = $i;
  $40 = (($39) + 1)|0;
  $i = $40;
 }
 $41 = $5;
 $42 = $41&255;
 $43 = ($42|0)!=(0);
 if (!($43)) {
  STACKTOP = sp;return;
 }
 $44 = $1;
 $45 = $2;
 $46 = $45<<1;
 $47 = (($44) + ($46)|0);
 ;HEAP8[$i16>>0]=HEAP8[$47>>0]|0;HEAP8[$i16+1>>0]=HEAP8[$47+1>>0]|0;
 $48 = HEAP16[$i16>>1]|0;
 $49 = (+($48<<16>>16));
 $50 = $49 / 32768.0;
 HEAPF64[$f64>>3] = $50;
 $51 = $0;
 $52 = $2;
 $53 = $52<<3;
 $54 = (($51) + ($53)|0);
 $55 = $5;
 $56 = $55&255;
 _memcpy(($54|0),($f64|0),($56|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int24_to_int8($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $i = 0, $i24 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $i24 = sp + 4|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 HEAP32[$i24>>2] = 0;
 $i = 0;
 while(1) {
  $6 = $i;
  $7 = $2;
  $8 = ($6>>>0)<($7>>>0);
  if (!($8)) {
   break;
  }
  $9 = ((($i24)) + 1|0);
  $10 = $1;
  $11 = $i;
  $12 = ($11*3)|0;
  $13 = (($10) + ($12)|0);
  ;HEAP8[$9>>0]=HEAP8[$13>>0]|0;HEAP8[$9+1>>0]=HEAP8[$13+1>>0]|0;HEAP8[$9+2>>0]=HEAP8[$13+2>>0]|0;
  $14 = HEAP32[$i24>>2]|0;
  $15 = $14 >> 16;
  $16 = (($15) + 128)|0;
  $17 = $16&255;
  $18 = $i;
  $19 = $0;
  $20 = (($19) + ($18)|0);
  HEAP8[$20>>0] = $17;
  $21 = $i;
  $22 = (($21) + 1)|0;
  $i = $22;
 }
 STACKTOP = sp;return;
}
function _wuff_int24_to_int16($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $0;
  $10 = $1;
  $11 = ((($10)) + 1|0);
  $12 = $3;
  $13 = $12&255;
  $14 = (($11) + ($13)|0);
  $15 = $4;
  $16 = $15&255;
  _memcpy(($9|0),($14|0),($16|0))|0;
  $17 = $1;
  $18 = ((($17)) + 3|0);
  $1 = $18;
  $19 = $4;
  $20 = $19&255;
  $21 = $0;
  $22 = (($21) + ($20)|0);
  $0 = $22;
 }
 $i = 0;
 while(1) {
  $23 = $i;
  $24 = $2;
  $25 = ($23>>>0)<($24>>>0);
  if (!($25)) {
   break;
  }
  $26 = $0;
  $27 = $i;
  $28 = $27<<1;
  $29 = (($26) + ($28)|0);
  $30 = $1;
  $31 = ((($30)) + 1|0);
  $32 = $i;
  $33 = ($32*3)|0;
  $34 = (($31) + ($33)|0);
  ;HEAP8[$29>>0]=HEAP8[$34>>0]|0;HEAP8[$29+1>>0]=HEAP8[$34+1>>0]|0;
  $35 = $i;
  $36 = (($35) + 1)|0;
  $i = $36;
 }
 $37 = $5;
 $38 = $37&255;
 $39 = ($38|0)!=(0);
 if (!($39)) {
  STACKTOP = sp;return;
 }
 $40 = $0;
 $41 = $2;
 $42 = $41<<1;
 $43 = (($40) + ($42)|0);
 $44 = $1;
 $45 = ((($44)) + 1|0);
 $46 = $2;
 $47 = ($46*3)|0;
 $48 = (($45) + ($47)|0);
 $49 = $5;
 $50 = $49&255;
 _memcpy(($43|0),($48|0),($50|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int24_to_int24($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $0;
 $7 = $1;
 $8 = $3;
 $9 = $8&255;
 $10 = (($7) + ($9)|0);
 $11 = $2;
 $12 = ($11*3)|0;
 $13 = $4;
 $14 = $13&255;
 $15 = (($12) + ($14))|0;
 $16 = $5;
 $17 = $16&255;
 $18 = (($15) + ($17))|0;
 _memcpy(($6|0),($10|0),($18|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int24_to_int32($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0, $i32 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $i32 = sp + 4|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 HEAP32[$i32>>2] = 0;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = ((($i32)) + 1|0);
  $10 = $1;
  ;HEAP8[$9>>0]=HEAP8[$10>>0]|0;HEAP8[$9+1>>0]=HEAP8[$10+1>>0]|0;HEAP8[$9+2>>0]=HEAP8[$10+2>>0]|0;
  $11 = $0;
  $12 = $3;
  $13 = $12&255;
  $14 = (($i32) + ($13)|0);
  $15 = $4;
  $16 = $15&255;
  _memcpy(($11|0),($14|0),($16|0))|0;
  $17 = $1;
  $18 = ((($17)) + 3|0);
  $1 = $18;
  $19 = $4;
  $20 = $19&255;
  $21 = $0;
  $22 = (($21) + ($20)|0);
  $0 = $22;
 }
 $i = 0;
 while(1) {
  $23 = $i;
  $24 = $2;
  $25 = ($23>>>0)<($24>>>0);
  if (!($25)) {
   break;
  }
  $26 = ((($i32)) + 1|0);
  $27 = $1;
  $28 = $i;
  $29 = ($28*3)|0;
  $30 = (($27) + ($29)|0);
  ;HEAP8[$26>>0]=HEAP8[$30>>0]|0;HEAP8[$26+1>>0]=HEAP8[$30+1>>0]|0;HEAP8[$26+2>>0]=HEAP8[$30+2>>0]|0;
  $31 = $0;
  $32 = $i;
  $33 = $32<<2;
  $34 = (($31) + ($33)|0);
  ;HEAP8[$34>>0]=HEAP8[$i32>>0]|0;HEAP8[$34+1>>0]=HEAP8[$i32+1>>0]|0;HEAP8[$34+2>>0]=HEAP8[$i32+2>>0]|0;HEAP8[$34+3>>0]=HEAP8[$i32+3>>0]|0;
  $35 = $i;
  $36 = (($35) + 1)|0;
  $i = $36;
 }
 $37 = $5;
 $38 = $37&255;
 $39 = ($38|0)!=(0);
 if (!($39)) {
  STACKTOP = sp;return;
 }
 $40 = ((($i32)) + 1|0);
 $41 = $1;
 $42 = $2;
 $43 = ($42*3)|0;
 $44 = (($41) + ($43)|0);
 ;HEAP8[$40>>0]=HEAP8[$44>>0]|0;HEAP8[$40+1>>0]=HEAP8[$44+1>>0]|0;HEAP8[$40+2>>0]=HEAP8[$44+2>>0]|0;
 $45 = $0;
 $46 = $2;
 $47 = $46<<2;
 $48 = (($45) + ($47)|0);
 $49 = $5;
 $50 = $49&255;
 _memcpy(($48|0),($i32|0),($50|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int24_to_float32($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $f32 = 0, $i = 0, $i24 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $i24 = sp + 8|0;
 $f32 = sp + 4|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 HEAP32[$i24>>2] = 0;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = ((($i24)) + 1|0);
  $10 = $1;
  ;HEAP8[$9>>0]=HEAP8[$10>>0]|0;HEAP8[$9+1>>0]=HEAP8[$10+1>>0]|0;HEAP8[$9+2>>0]=HEAP8[$10+2>>0]|0;
  $11 = HEAP32[$i24>>2]|0;
  $12 = (+($11|0));
  $13 = $12 / 2147483648.0;
  $14 = $13;
  HEAPF32[$f32>>2] = $14;
  $15 = $0;
  $16 = $3;
  $17 = $16&255;
  $18 = (($f32) + ($17)|0);
  $19 = $4;
  $20 = $19&255;
  _memcpy(($15|0),($18|0),($20|0))|0;
  $21 = $1;
  $22 = ((($21)) + 3|0);
  $1 = $22;
  $23 = $4;
  $24 = $23&255;
  $25 = $0;
  $26 = (($25) + ($24)|0);
  $0 = $26;
 }
 $i = 0;
 while(1) {
  $27 = $i;
  $28 = $2;
  $29 = ($27>>>0)<($28>>>0);
  if (!($29)) {
   break;
  }
  $30 = ((($i24)) + 1|0);
  $31 = $1;
  $32 = $i;
  $33 = ($32*3)|0;
  $34 = (($31) + ($33)|0);
  ;HEAP8[$30>>0]=HEAP8[$34>>0]|0;HEAP8[$30+1>>0]=HEAP8[$34+1>>0]|0;HEAP8[$30+2>>0]=HEAP8[$34+2>>0]|0;
  $35 = HEAP32[$i24>>2]|0;
  $36 = (+($35|0));
  $37 = $36 / 2147483648.0;
  $38 = $37;
  HEAPF32[$f32>>2] = $38;
  $39 = $0;
  $40 = $i;
  $41 = $40<<2;
  $42 = (($39) + ($41)|0);
  ;HEAP8[$42>>0]=HEAP8[$f32>>0]|0;HEAP8[$42+1>>0]=HEAP8[$f32+1>>0]|0;HEAP8[$42+2>>0]=HEAP8[$f32+2>>0]|0;HEAP8[$42+3>>0]=HEAP8[$f32+3>>0]|0;
  $43 = $i;
  $44 = (($43) + 1)|0;
  $i = $44;
 }
 $45 = $5;
 $46 = $45&255;
 $47 = ($46|0)!=(0);
 if (!($47)) {
  STACKTOP = sp;return;
 }
 $48 = ((($i24)) + 1|0);
 $49 = $1;
 $50 = $2;
 $51 = ($50*3)|0;
 $52 = (($49) + ($51)|0);
 ;HEAP8[$48>>0]=HEAP8[$52>>0]|0;HEAP8[$48+1>>0]=HEAP8[$52+1>>0]|0;HEAP8[$48+2>>0]=HEAP8[$52+2>>0]|0;
 $53 = HEAP32[$i24>>2]|0;
 $54 = (+($53|0));
 $55 = $54 / 2147483648.0;
 $56 = $55;
 HEAPF32[$f32>>2] = $56;
 $57 = $0;
 $58 = $2;
 $59 = $58<<2;
 $60 = (($57) + ($59)|0);
 $61 = $5;
 $62 = $61&255;
 _memcpy(($60|0),($f32|0),($62|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int24_to_float64($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $f64 = 0, $i = 0, $i24 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $i24 = sp + 12|0;
 $f64 = sp;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 HEAP32[$i24>>2] = 0;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = ((($i24)) + 1|0);
  $10 = $1;
  ;HEAP8[$9>>0]=HEAP8[$10>>0]|0;HEAP8[$9+1>>0]=HEAP8[$10+1>>0]|0;HEAP8[$9+2>>0]=HEAP8[$10+2>>0]|0;
  $11 = HEAP32[$i24>>2]|0;
  $12 = (+($11|0));
  $13 = $12 / 2147483648.0;
  HEAPF64[$f64>>3] = $13;
  $14 = $0;
  $15 = $3;
  $16 = $15&255;
  $17 = (($f64) + ($16)|0);
  $18 = $4;
  $19 = $18&255;
  _memcpy(($14|0),($17|0),($19|0))|0;
  $20 = $1;
  $21 = ((($20)) + 3|0);
  $1 = $21;
  $22 = $4;
  $23 = $22&255;
  $24 = $0;
  $25 = (($24) + ($23)|0);
  $0 = $25;
 }
 $i = 0;
 while(1) {
  $26 = $i;
  $27 = $2;
  $28 = ($26>>>0)<($27>>>0);
  if (!($28)) {
   break;
  }
  $29 = ((($i24)) + 1|0);
  $30 = $1;
  $31 = $i;
  $32 = ($31*3)|0;
  $33 = (($30) + ($32)|0);
  ;HEAP8[$29>>0]=HEAP8[$33>>0]|0;HEAP8[$29+1>>0]=HEAP8[$33+1>>0]|0;HEAP8[$29+2>>0]=HEAP8[$33+2>>0]|0;
  $34 = HEAP32[$i24>>2]|0;
  $35 = (+($34|0));
  $36 = $35 / 2147483648.0;
  HEAPF64[$f64>>3] = $36;
  $37 = $0;
  $38 = $i;
  $39 = $38<<3;
  $40 = (($37) + ($39)|0);
  ;HEAP8[$40>>0]=HEAP8[$f64>>0]|0;HEAP8[$40+1>>0]=HEAP8[$f64+1>>0]|0;HEAP8[$40+2>>0]=HEAP8[$f64+2>>0]|0;HEAP8[$40+3>>0]=HEAP8[$f64+3>>0]|0;HEAP8[$40+4>>0]=HEAP8[$f64+4>>0]|0;HEAP8[$40+5>>0]=HEAP8[$f64+5>>0]|0;HEAP8[$40+6>>0]=HEAP8[$f64+6>>0]|0;HEAP8[$40+7>>0]=HEAP8[$f64+7>>0]|0;
  $41 = $i;
  $42 = (($41) + 1)|0;
  $i = $42;
 }
 $43 = $5;
 $44 = $43&255;
 $45 = ($44|0)!=(0);
 if (!($45)) {
  STACKTOP = sp;return;
 }
 $46 = ((($i24)) + 1|0);
 $47 = $1;
 $48 = $2;
 $49 = ($48*3)|0;
 $50 = (($47) + ($49)|0);
 ;HEAP8[$46>>0]=HEAP8[$50>>0]|0;HEAP8[$46+1>>0]=HEAP8[$50+1>>0]|0;HEAP8[$46+2>>0]=HEAP8[$50+2>>0]|0;
 $51 = HEAP32[$i24>>2]|0;
 $52 = (+($51|0));
 $53 = $52 / 2147483648.0;
 HEAPF64[$f64>>3] = $53;
 $54 = $0;
 $55 = $2;
 $56 = $55<<3;
 $57 = (($54) + ($56)|0);
 $58 = $5;
 $59 = $58&255;
 _memcpy(($57|0),($f64|0),($59|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int32_to_int8($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $i = 0, $i32 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $i32 = sp + 4|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 HEAP32[$i32>>2] = 0;
 $i = 0;
 while(1) {
  $6 = $i;
  $7 = $2;
  $8 = ($6>>>0)<($7>>>0);
  if (!($8)) {
   break;
  }
  $9 = $1;
  $10 = $i;
  $11 = $10<<2;
  $12 = (($9) + ($11)|0);
  ;HEAP8[$i32>>0]=HEAP8[$12>>0]|0;HEAP8[$i32+1>>0]=HEAP8[$12+1>>0]|0;HEAP8[$i32+2>>0]=HEAP8[$12+2>>0]|0;HEAP8[$i32+3>>0]=HEAP8[$12+3>>0]|0;
  $13 = HEAP32[$i32>>2]|0;
  $14 = $13 >> 24;
  $15 = (($14) + 128)|0;
  $16 = $15&255;
  $17 = $i;
  $18 = $0;
  $19 = (($18) + ($17)|0);
  HEAP8[$19>>0] = $16;
  $20 = $i;
  $21 = (($20) + 1)|0;
  $i = $21;
 }
 STACKTOP = sp;return;
}
function _wuff_int32_to_int16($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $0;
  $10 = $1;
  $11 = ((($10)) + 2|0);
  $12 = $3;
  $13 = $12&255;
  $14 = (($11) + ($13)|0);
  $15 = $4;
  $16 = $15&255;
  _memcpy(($9|0),($14|0),($16|0))|0;
  $17 = $1;
  $18 = ((($17)) + 4|0);
  $1 = $18;
  $19 = $4;
  $20 = $19&255;
  $21 = $0;
  $22 = (($21) + ($20)|0);
  $0 = $22;
 }
 $i = 0;
 while(1) {
  $23 = $i;
  $24 = $2;
  $25 = ($23>>>0)<($24>>>0);
  if (!($25)) {
   break;
  }
  $26 = $0;
  $27 = $i;
  $28 = $27<<1;
  $29 = (($26) + ($28)|0);
  $30 = $1;
  $31 = ((($30)) + 2|0);
  $32 = $i;
  $33 = $32<<2;
  $34 = (($31) + ($33)|0);
  ;HEAP8[$29>>0]=HEAP8[$34>>0]|0;HEAP8[$29+1>>0]=HEAP8[$34+1>>0]|0;
  $35 = $i;
  $36 = (($35) + 1)|0;
  $i = $36;
 }
 $37 = $5;
 $38 = $37&255;
 $39 = ($38|0)!=(0);
 if (!($39)) {
  STACKTOP = sp;return;
 }
 $40 = $0;
 $41 = $2;
 $42 = $41<<1;
 $43 = (($40) + ($42)|0);
 $44 = $1;
 $45 = ((($44)) + 2|0);
 $46 = $2;
 $47 = $46<<2;
 $48 = (($45) + ($47)|0);
 $49 = $5;
 $50 = $49&255;
 _memcpy(($43|0),($48|0),($50|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int32_to_int24($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $0;
  $10 = $1;
  $11 = ((($10)) + 1|0);
  $12 = $3;
  $13 = $12&255;
  $14 = (($11) + ($13)|0);
  $15 = $4;
  $16 = $15&255;
  _memcpy(($9|0),($14|0),($16|0))|0;
  $17 = $1;
  $18 = ((($17)) + 4|0);
  $1 = $18;
  $19 = $4;
  $20 = $19&255;
  $21 = $0;
  $22 = (($21) + ($20)|0);
  $0 = $22;
 }
 $i = 0;
 while(1) {
  $23 = $i;
  $24 = $2;
  $25 = ($23>>>0)<($24>>>0);
  if (!($25)) {
   break;
  }
  $26 = $0;
  $27 = $i;
  $28 = ($27*3)|0;
  $29 = (($26) + ($28)|0);
  $30 = $1;
  $31 = ((($30)) + 1|0);
  $32 = $i;
  $33 = $32<<2;
  $34 = (($31) + ($33)|0);
  ;HEAP8[$29>>0]=HEAP8[$34>>0]|0;HEAP8[$29+1>>0]=HEAP8[$34+1>>0]|0;HEAP8[$29+2>>0]=HEAP8[$34+2>>0]|0;
  $35 = $i;
  $36 = (($35) + 1)|0;
  $i = $36;
 }
 $37 = $5;
 $38 = $37&255;
 $39 = ($38|0)!=(0);
 if (!($39)) {
  STACKTOP = sp;return;
 }
 $40 = $0;
 $41 = $2;
 $42 = ($41*3)|0;
 $43 = (($40) + ($42)|0);
 $44 = $1;
 $45 = ((($44)) + 1|0);
 $46 = $2;
 $47 = $46<<2;
 $48 = (($45) + ($47)|0);
 $49 = $5;
 $50 = $49&255;
 _memcpy(($43|0),($48|0),($50|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int32_to_int32($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $0;
 $7 = $1;
 $8 = $3;
 $9 = $8&255;
 $10 = (($7) + ($9)|0);
 $11 = $2;
 $12 = $11<<2;
 $13 = $4;
 $14 = $13&255;
 $15 = (($12) + ($14))|0;
 $16 = $5;
 $17 = $16&255;
 $18 = (($15) + ($17))|0;
 _memcpy(($6|0),($10|0),($18|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int32_to_float32($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $f32 = 0, $i = 0, $i32 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $i32 = sp + 8|0;
 $f32 = sp + 4|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  ;HEAP8[$i32>>0]=HEAP8[$9>>0]|0;HEAP8[$i32+1>>0]=HEAP8[$9+1>>0]|0;HEAP8[$i32+2>>0]=HEAP8[$9+2>>0]|0;HEAP8[$i32+3>>0]=HEAP8[$9+3>>0]|0;
  $10 = HEAP32[$i32>>2]|0;
  $11 = (+($10|0));
  $12 = $11 / 2147483648.0;
  $13 = $12;
  HEAPF32[$f32>>2] = $13;
  $14 = $0;
  $15 = $3;
  $16 = $15&255;
  $17 = (($f32) + ($16)|0);
  $18 = $4;
  $19 = $18&255;
  _memcpy(($14|0),($17|0),($19|0))|0;
  $20 = $1;
  $21 = ((($20)) + 4|0);
  $1 = $21;
  $22 = $4;
  $23 = $22&255;
  $24 = $0;
  $25 = (($24) + ($23)|0);
  $0 = $25;
 }
 $i = 0;
 while(1) {
  $26 = $i;
  $27 = $2;
  $28 = ($26>>>0)<($27>>>0);
  if (!($28)) {
   break;
  }
  $29 = $1;
  $30 = $i;
  $31 = $30<<2;
  $32 = (($29) + ($31)|0);
  ;HEAP8[$i32>>0]=HEAP8[$32>>0]|0;HEAP8[$i32+1>>0]=HEAP8[$32+1>>0]|0;HEAP8[$i32+2>>0]=HEAP8[$32+2>>0]|0;HEAP8[$i32+3>>0]=HEAP8[$32+3>>0]|0;
  $33 = HEAP32[$i32>>2]|0;
  $34 = (+($33|0));
  $35 = $34 / 2147483648.0;
  $36 = $35;
  HEAPF32[$f32>>2] = $36;
  $37 = $0;
  $38 = $i;
  $39 = $38<<2;
  $40 = (($37) + ($39)|0);
  ;HEAP8[$40>>0]=HEAP8[$f32>>0]|0;HEAP8[$40+1>>0]=HEAP8[$f32+1>>0]|0;HEAP8[$40+2>>0]=HEAP8[$f32+2>>0]|0;HEAP8[$40+3>>0]=HEAP8[$f32+3>>0]|0;
  $41 = $i;
  $42 = (($41) + 1)|0;
  $i = $42;
 }
 $43 = $5;
 $44 = $43&255;
 $45 = ($44|0)!=(0);
 if (!($45)) {
  STACKTOP = sp;return;
 }
 $46 = $1;
 $47 = $2;
 $48 = $47<<2;
 $49 = (($46) + ($48)|0);
 ;HEAP8[$i32>>0]=HEAP8[$49>>0]|0;HEAP8[$i32+1>>0]=HEAP8[$49+1>>0]|0;HEAP8[$i32+2>>0]=HEAP8[$49+2>>0]|0;HEAP8[$i32+3>>0]=HEAP8[$49+3>>0]|0;
 $50 = HEAP32[$i32>>2]|0;
 $51 = (+($50|0));
 $52 = $51 / 2147483648.0;
 $53 = $52;
 HEAPF32[$f32>>2] = $53;
 $54 = $0;
 $55 = $2;
 $56 = $55<<2;
 $57 = (($54) + ($56)|0);
 $58 = $5;
 $59 = $58&255;
 _memcpy(($57|0),($f32|0),($59|0))|0;
 STACKTOP = sp;return;
}
function _wuff_int32_to_float64($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f64 = 0, $i = 0, $i32 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $i32 = sp + 12|0;
 $f64 = sp;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  ;HEAP8[$i32>>0]=HEAP8[$9>>0]|0;HEAP8[$i32+1>>0]=HEAP8[$9+1>>0]|0;HEAP8[$i32+2>>0]=HEAP8[$9+2>>0]|0;HEAP8[$i32+3>>0]=HEAP8[$9+3>>0]|0;
  $10 = HEAP32[$i32>>2]|0;
  $11 = (+($10|0));
  $12 = $11 / 2147483648.0;
  HEAPF64[$f64>>3] = $12;
  $13 = $0;
  $14 = $3;
  $15 = $14&255;
  $16 = (($f64) + ($15)|0);
  $17 = $4;
  $18 = $17&255;
  _memcpy(($13|0),($16|0),($18|0))|0;
  $19 = $1;
  $20 = ((($19)) + 4|0);
  $1 = $20;
  $21 = $4;
  $22 = $21&255;
  $23 = $0;
  $24 = (($23) + ($22)|0);
  $0 = $24;
 }
 $i = 0;
 while(1) {
  $25 = $i;
  $26 = $2;
  $27 = ($25>>>0)<($26>>>0);
  if (!($27)) {
   break;
  }
  $28 = $1;
  $29 = $i;
  $30 = $29<<2;
  $31 = (($28) + ($30)|0);
  ;HEAP8[$i32>>0]=HEAP8[$31>>0]|0;HEAP8[$i32+1>>0]=HEAP8[$31+1>>0]|0;HEAP8[$i32+2>>0]=HEAP8[$31+2>>0]|0;HEAP8[$i32+3>>0]=HEAP8[$31+3>>0]|0;
  $32 = HEAP32[$i32>>2]|0;
  $33 = (+($32|0));
  $34 = $33 / 2147483648.0;
  HEAPF64[$f64>>3] = $34;
  $35 = $0;
  $36 = $i;
  $37 = $36<<3;
  $38 = (($35) + ($37)|0);
  ;HEAP8[$38>>0]=HEAP8[$f64>>0]|0;HEAP8[$38+1>>0]=HEAP8[$f64+1>>0]|0;HEAP8[$38+2>>0]=HEAP8[$f64+2>>0]|0;HEAP8[$38+3>>0]=HEAP8[$f64+3>>0]|0;HEAP8[$38+4>>0]=HEAP8[$f64+4>>0]|0;HEAP8[$38+5>>0]=HEAP8[$f64+5>>0]|0;HEAP8[$38+6>>0]=HEAP8[$f64+6>>0]|0;HEAP8[$38+7>>0]=HEAP8[$f64+7>>0]|0;
  $39 = $i;
  $40 = (($39) + 1)|0;
  $i = $40;
 }
 $41 = $5;
 $42 = $41&255;
 $43 = ($42|0)!=(0);
 if (!($43)) {
  STACKTOP = sp;return;
 }
 $44 = $1;
 $45 = $2;
 $46 = $45<<2;
 $47 = (($44) + ($46)|0);
 ;HEAP8[$i32>>0]=HEAP8[$47>>0]|0;HEAP8[$i32+1>>0]=HEAP8[$47+1>>0]|0;HEAP8[$i32+2>>0]=HEAP8[$47+2>>0]|0;HEAP8[$i32+3>>0]=HEAP8[$47+3>>0]|0;
 $48 = HEAP32[$i32>>2]|0;
 $49 = (+($48|0));
 $50 = $49 / 2147483648.0;
 HEAPF64[$f64>>3] = $50;
 $51 = $0;
 $52 = $2;
 $53 = $52<<3;
 $54 = (($51) + ($53)|0);
 $55 = $5;
 $56 = $55&255;
 _memcpy(($54|0),($f64|0),($56|0))|0;
 STACKTOP = sp;return;
}
function _wuff_float32_to_int8($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $f32 = 0, $i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f32 = sp + 4|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $i = 0;
 while(1) {
  $6 = $i;
  $7 = $2;
  $8 = ($6>>>0)<($7>>>0);
  if (!($8)) {
   break;
  }
  $9 = $1;
  $10 = $i;
  $11 = $10<<2;
  $12 = (($9) + ($11)|0);
  ;HEAP8[$f32>>0]=HEAP8[$12>>0]|0;HEAP8[$f32+1>>0]=HEAP8[$12+1>>0]|0;HEAP8[$f32+2>>0]=HEAP8[$12+2>>0]|0;HEAP8[$f32+3>>0]=HEAP8[$12+3>>0]|0;
  $13 = +HEAPF32[$f32>>2];
  $14 = $13 * 127.5;
  $15 = $14 + 128.0;
  $16 = (~~(($15))&255);
  $17 = $i;
  $18 = $0;
  $19 = (($18) + ($17)|0);
  HEAP8[$19>>0] = $16;
  $20 = $i;
  $21 = (($20) + 1)|0;
  $i = $21;
 }
 STACKTOP = sp;return;
}
function _wuff_float32_to_int16($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f32 = 0, $i = 0, $i16 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f32 = sp + 4|0;
 $i16 = sp + 20|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  ;HEAP8[$f32>>0]=HEAP8[$9>>0]|0;HEAP8[$f32+1>>0]=HEAP8[$9+1>>0]|0;HEAP8[$f32+2>>0]=HEAP8[$9+2>>0]|0;HEAP8[$f32+3>>0]=HEAP8[$9+3>>0]|0;
  $10 = +HEAPF32[$f32>>2];
  $11 = $10 * 32767.5;
  $12 = (~~(($11)));
  HEAP16[$i16>>1] = $12;
  $13 = $0;
  $14 = $3;
  $15 = $14&255;
  $16 = (($i16) + ($15)|0);
  $17 = $4;
  $18 = $17&255;
  _memcpy(($13|0),($16|0),($18|0))|0;
  $19 = $1;
  $20 = ((($19)) + 4|0);
  $1 = $20;
  $21 = $4;
  $22 = $21&255;
  $23 = $0;
  $24 = (($23) + ($22)|0);
  $0 = $24;
 }
 $i = 0;
 while(1) {
  $25 = $i;
  $26 = $2;
  $27 = ($25>>>0)<($26>>>0);
  if (!($27)) {
   break;
  }
  $28 = $1;
  $29 = $i;
  $30 = $29<<2;
  $31 = (($28) + ($30)|0);
  ;HEAP8[$f32>>0]=HEAP8[$31>>0]|0;HEAP8[$f32+1>>0]=HEAP8[$31+1>>0]|0;HEAP8[$f32+2>>0]=HEAP8[$31+2>>0]|0;HEAP8[$f32+3>>0]=HEAP8[$31+3>>0]|0;
  $32 = +HEAPF32[$f32>>2];
  $33 = $32 * 32767.5;
  $34 = (~~(($33)));
  HEAP16[$i16>>1] = $34;
  $35 = $0;
  $36 = $i;
  $37 = $36<<1;
  $38 = (($35) + ($37)|0);
  ;HEAP8[$38>>0]=HEAP8[$i16>>0]|0;HEAP8[$38+1>>0]=HEAP8[$i16+1>>0]|0;
  $39 = $i;
  $40 = (($39) + 1)|0;
  $i = $40;
 }
 $41 = $5;
 $42 = $41&255;
 $43 = ($42|0)!=(0);
 if (!($43)) {
  STACKTOP = sp;return;
 }
 $44 = $1;
 $45 = $i;
 $46 = $45<<2;
 $47 = (($44) + ($46)|0);
 ;HEAP8[$f32>>0]=HEAP8[$47>>0]|0;HEAP8[$f32+1>>0]=HEAP8[$47+1>>0]|0;HEAP8[$f32+2>>0]=HEAP8[$47+2>>0]|0;HEAP8[$f32+3>>0]=HEAP8[$47+3>>0]|0;
 $48 = +HEAPF32[$f32>>2];
 $49 = $48 * 32767.5;
 $50 = (~~(($49)));
 HEAP16[$i16>>1] = $50;
 $51 = $0;
 $52 = $i;
 $53 = $52<<1;
 $54 = (($51) + ($53)|0);
 $55 = $5;
 $56 = $55&255;
 _memcpy(($54|0),($i16|0),($56|0))|0;
 STACKTOP = sp;return;
}
function _wuff_float32_to_int24($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $f32 = 0, $i = 0, $i24 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f32 = sp + 8|0;
 $i24 = sp + 4|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  ;HEAP8[$f32>>0]=HEAP8[$9>>0]|0;HEAP8[$f32+1>>0]=HEAP8[$9+1>>0]|0;HEAP8[$f32+2>>0]=HEAP8[$9+2>>0]|0;HEAP8[$f32+3>>0]=HEAP8[$9+3>>0]|0;
  $10 = +HEAPF32[$f32>>2];
  $11 = $10;
  $12 = $11 * 2147483647.5;
  $13 = (~~(($12)));
  HEAP32[$i24>>2] = $13;
  $14 = $0;
  $15 = ((($i24)) + 1|0);
  $16 = $3;
  $17 = $16&255;
  $18 = (($15) + ($17)|0);
  $19 = $4;
  $20 = $19&255;
  _memcpy(($14|0),($18|0),($20|0))|0;
  $21 = $1;
  $22 = ((($21)) + 4|0);
  $1 = $22;
  $23 = $4;
  $24 = $23&255;
  $25 = $0;
  $26 = (($25) + ($24)|0);
  $0 = $26;
 }
 $i = 0;
 while(1) {
  $27 = $i;
  $28 = $2;
  $29 = ($27>>>0)<($28>>>0);
  if (!($29)) {
   break;
  }
  $30 = $1;
  $31 = $i;
  $32 = $31<<2;
  $33 = (($30) + ($32)|0);
  ;HEAP8[$f32>>0]=HEAP8[$33>>0]|0;HEAP8[$f32+1>>0]=HEAP8[$33+1>>0]|0;HEAP8[$f32+2>>0]=HEAP8[$33+2>>0]|0;HEAP8[$f32+3>>0]=HEAP8[$33+3>>0]|0;
  $34 = +HEAPF32[$f32>>2];
  $35 = $34;
  $36 = $35 * 2147483647.5;
  $37 = (~~(($36)));
  HEAP32[$i24>>2] = $37;
  $38 = $0;
  $39 = $i;
  $40 = ($39*3)|0;
  $41 = (($38) + ($40)|0);
  $42 = ((($i24)) + 1|0);
  ;HEAP8[$41>>0]=HEAP8[$42>>0]|0;HEAP8[$41+1>>0]=HEAP8[$42+1>>0]|0;HEAP8[$41+2>>0]=HEAP8[$42+2>>0]|0;
  $43 = $i;
  $44 = (($43) + 1)|0;
  $i = $44;
 }
 $45 = $5;
 $46 = $45&255;
 $47 = ($46|0)!=(0);
 if (!($47)) {
  STACKTOP = sp;return;
 }
 $48 = $1;
 $49 = $2;
 $50 = $49<<2;
 $51 = (($48) + ($50)|0);
 ;HEAP8[$f32>>0]=HEAP8[$51>>0]|0;HEAP8[$f32+1>>0]=HEAP8[$51+1>>0]|0;HEAP8[$f32+2>>0]=HEAP8[$51+2>>0]|0;HEAP8[$f32+3>>0]=HEAP8[$51+3>>0]|0;
 $52 = +HEAPF32[$f32>>2];
 $53 = $52;
 $54 = $53 * 2147483647.5;
 $55 = (~~(($54)));
 HEAP32[$i24>>2] = $55;
 $56 = $0;
 $57 = $2;
 $58 = ($57*3)|0;
 $59 = (($56) + ($58)|0);
 $60 = ((($i24)) + 1|0);
 $61 = $5;
 $62 = $61&255;
 _memcpy(($59|0),($60|0),($62|0))|0;
 STACKTOP = sp;return;
}
function _wuff_float32_to_int32($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $f32 = 0, $i = 0, $i32 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f32 = sp + 8|0;
 $i32 = sp + 4|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  ;HEAP8[$f32>>0]=HEAP8[$9>>0]|0;HEAP8[$f32+1>>0]=HEAP8[$9+1>>0]|0;HEAP8[$f32+2>>0]=HEAP8[$9+2>>0]|0;HEAP8[$f32+3>>0]=HEAP8[$9+3>>0]|0;
  $10 = +HEAPF32[$f32>>2];
  $11 = $10;
  $12 = $11 * 2147483647.5;
  $13 = (~~(($12)));
  HEAP32[$i32>>2] = $13;
  $14 = $0;
  $15 = $3;
  $16 = $15&255;
  $17 = (($i32) + ($16)|0);
  $18 = $4;
  $19 = $18&255;
  _memcpy(($14|0),($17|0),($19|0))|0;
  $20 = $1;
  $21 = ((($20)) + 4|0);
  $1 = $21;
  $22 = $4;
  $23 = $22&255;
  $24 = $0;
  $25 = (($24) + ($23)|0);
  $0 = $25;
 }
 $i = 0;
 while(1) {
  $26 = $i;
  $27 = $2;
  $28 = ($26>>>0)<($27>>>0);
  if (!($28)) {
   break;
  }
  $29 = $1;
  $30 = $i;
  $31 = $30<<2;
  $32 = (($29) + ($31)|0);
  ;HEAP8[$f32>>0]=HEAP8[$32>>0]|0;HEAP8[$f32+1>>0]=HEAP8[$32+1>>0]|0;HEAP8[$f32+2>>0]=HEAP8[$32+2>>0]|0;HEAP8[$f32+3>>0]=HEAP8[$32+3>>0]|0;
  $33 = +HEAPF32[$f32>>2];
  $34 = $33;
  $35 = $34 * 2147483647.5;
  $36 = (~~(($35)));
  HEAP32[$i32>>2] = $36;
  $37 = $0;
  $38 = $i;
  $39 = $38<<2;
  $40 = (($37) + ($39)|0);
  ;HEAP8[$40>>0]=HEAP8[$i32>>0]|0;HEAP8[$40+1>>0]=HEAP8[$i32+1>>0]|0;HEAP8[$40+2>>0]=HEAP8[$i32+2>>0]|0;HEAP8[$40+3>>0]=HEAP8[$i32+3>>0]|0;
  $41 = $i;
  $42 = (($41) + 1)|0;
  $i = $42;
 }
 $43 = $5;
 $44 = $43&255;
 $45 = ($44|0)!=(0);
 if (!($45)) {
  STACKTOP = sp;return;
 }
 $46 = $1;
 $47 = $2;
 $48 = $47<<2;
 $49 = (($46) + ($48)|0);
 ;HEAP8[$f32>>0]=HEAP8[$49>>0]|0;HEAP8[$f32+1>>0]=HEAP8[$49+1>>0]|0;HEAP8[$f32+2>>0]=HEAP8[$49+2>>0]|0;HEAP8[$f32+3>>0]=HEAP8[$49+3>>0]|0;
 $50 = +HEAPF32[$f32>>2];
 $51 = $50;
 $52 = $51 * 2147483647.5;
 $53 = (~~(($52)));
 HEAP32[$i32>>2] = $53;
 $54 = $0;
 $55 = $2;
 $56 = $55<<2;
 $57 = (($54) + ($56)|0);
 $58 = $5;
 $59 = $58&255;
 _memcpy(($57|0),($i32|0),($59|0))|0;
 STACKTOP = sp;return;
}
function _wuff_float32_to_float32($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $0;
 $7 = $1;
 $8 = $3;
 $9 = $8&255;
 $10 = (($7) + ($9)|0);
 $11 = $2;
 $12 = $11<<2;
 $13 = $4;
 $14 = $13&255;
 $15 = (($12) + ($14))|0;
 $16 = $5;
 $17 = $16&255;
 $18 = (($15) + ($17))|0;
 _memcpy(($6|0),($10|0),($18|0))|0;
 STACKTOP = sp;return;
}
function _wuff_float32_to_float64($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f32 = 0, $f64 = 0, $i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f32 = sp + 12|0;
 $f64 = sp;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  ;HEAP8[$f32>>0]=HEAP8[$9>>0]|0;HEAP8[$f32+1>>0]=HEAP8[$9+1>>0]|0;HEAP8[$f32+2>>0]=HEAP8[$9+2>>0]|0;HEAP8[$f32+3>>0]=HEAP8[$9+3>>0]|0;
  $10 = +HEAPF32[$f32>>2];
  $11 = $10;
  HEAPF64[$f64>>3] = $11;
  $12 = $0;
  $13 = $3;
  $14 = $13&255;
  $15 = (($f64) + ($14)|0);
  $16 = $4;
  $17 = $16&255;
  _memcpy(($12|0),($15|0),($17|0))|0;
  $18 = $1;
  $19 = ((($18)) + 4|0);
  $1 = $19;
  $20 = $4;
  $21 = $20&255;
  $22 = $0;
  $23 = (($22) + ($21)|0);
  $0 = $23;
 }
 $i = 0;
 while(1) {
  $24 = $i;
  $25 = $2;
  $26 = ($24>>>0)<($25>>>0);
  if (!($26)) {
   break;
  }
  $27 = $1;
  $28 = $i;
  $29 = $28<<2;
  $30 = (($27) + ($29)|0);
  ;HEAP8[$f32>>0]=HEAP8[$30>>0]|0;HEAP8[$f32+1>>0]=HEAP8[$30+1>>0]|0;HEAP8[$f32+2>>0]=HEAP8[$30+2>>0]|0;HEAP8[$f32+3>>0]=HEAP8[$30+3>>0]|0;
  $31 = +HEAPF32[$f32>>2];
  $32 = $31;
  HEAPF64[$f64>>3] = $32;
  $33 = $0;
  $34 = $i;
  $35 = $34<<3;
  $36 = (($33) + ($35)|0);
  ;HEAP8[$36>>0]=HEAP8[$f64>>0]|0;HEAP8[$36+1>>0]=HEAP8[$f64+1>>0]|0;HEAP8[$36+2>>0]=HEAP8[$f64+2>>0]|0;HEAP8[$36+3>>0]=HEAP8[$f64+3>>0]|0;HEAP8[$36+4>>0]=HEAP8[$f64+4>>0]|0;HEAP8[$36+5>>0]=HEAP8[$f64+5>>0]|0;HEAP8[$36+6>>0]=HEAP8[$f64+6>>0]|0;HEAP8[$36+7>>0]=HEAP8[$f64+7>>0]|0;
  $37 = $i;
  $38 = (($37) + 1)|0;
  $i = $38;
 }
 $39 = $5;
 $40 = $39&255;
 $41 = ($40|0)!=(0);
 if (!($41)) {
  STACKTOP = sp;return;
 }
 $42 = $1;
 $43 = $2;
 $44 = $43<<2;
 $45 = (($42) + ($44)|0);
 ;HEAP8[$f32>>0]=HEAP8[$45>>0]|0;HEAP8[$f32+1>>0]=HEAP8[$45+1>>0]|0;HEAP8[$f32+2>>0]=HEAP8[$45+2>>0]|0;HEAP8[$f32+3>>0]=HEAP8[$45+3>>0]|0;
 $46 = +HEAPF32[$f32>>2];
 $47 = $46;
 HEAPF64[$f64>>3] = $47;
 $48 = $0;
 $49 = $2;
 $50 = $49<<3;
 $51 = (($48) + ($50)|0);
 $52 = $5;
 $53 = $52&255;
 _memcpy(($51|0),($f64|0),($53|0))|0;
 STACKTOP = sp;return;
}
function _wuff_float64_to_int8($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $f64 = 0, $i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f64 = sp;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $i = 0;
 while(1) {
  $6 = $i;
  $7 = $2;
  $8 = ($6>>>0)<($7>>>0);
  if (!($8)) {
   break;
  }
  $9 = $1;
  $10 = $i;
  $11 = $10<<3;
  $12 = (($9) + ($11)|0);
  ;HEAP8[$f64>>0]=HEAP8[$12>>0]|0;HEAP8[$f64+1>>0]=HEAP8[$12+1>>0]|0;HEAP8[$f64+2>>0]=HEAP8[$12+2>>0]|0;HEAP8[$f64+3>>0]=HEAP8[$12+3>>0]|0;HEAP8[$f64+4>>0]=HEAP8[$12+4>>0]|0;HEAP8[$f64+5>>0]=HEAP8[$12+5>>0]|0;HEAP8[$f64+6>>0]=HEAP8[$12+6>>0]|0;HEAP8[$f64+7>>0]=HEAP8[$12+7>>0]|0;
  $13 = +HEAPF64[$f64>>3];
  $14 = $13 * 127.5;
  $15 = $14 + 128.0;
  $16 = (~~(($15))&255);
  $17 = $i;
  $18 = $0;
  $19 = (($18) + ($17)|0);
  HEAP8[$19>>0] = $16;
  $20 = $i;
  $21 = (($20) + 1)|0;
  $i = $21;
 }
 STACKTOP = sp;return;
}
function _wuff_float64_to_int16($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f64 = 0, $i = 0, $i16 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f64 = sp;
 $i16 = sp + 24|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  ;HEAP8[$f64>>0]=HEAP8[$9>>0]|0;HEAP8[$f64+1>>0]=HEAP8[$9+1>>0]|0;HEAP8[$f64+2>>0]=HEAP8[$9+2>>0]|0;HEAP8[$f64+3>>0]=HEAP8[$9+3>>0]|0;HEAP8[$f64+4>>0]=HEAP8[$9+4>>0]|0;HEAP8[$f64+5>>0]=HEAP8[$9+5>>0]|0;HEAP8[$f64+6>>0]=HEAP8[$9+6>>0]|0;HEAP8[$f64+7>>0]=HEAP8[$9+7>>0]|0;
  $10 = +HEAPF64[$f64>>3];
  $11 = $10 * 32767.5;
  $12 = (~~(($11)));
  HEAP16[$i16>>1] = $12;
  $13 = $0;
  $14 = $3;
  $15 = $14&255;
  $16 = (($i16) + ($15)|0);
  $17 = $4;
  $18 = $17&255;
  _memcpy(($13|0),($16|0),($18|0))|0;
  $19 = $1;
  $20 = ((($19)) + 8|0);
  $1 = $20;
  $21 = $4;
  $22 = $21&255;
  $23 = $0;
  $24 = (($23) + ($22)|0);
  $0 = $24;
 }
 $i = 0;
 while(1) {
  $25 = $i;
  $26 = $2;
  $27 = ($25>>>0)<($26>>>0);
  if (!($27)) {
   break;
  }
  $28 = $1;
  $29 = $i;
  $30 = $29<<3;
  $31 = (($28) + ($30)|0);
  ;HEAP8[$f64>>0]=HEAP8[$31>>0]|0;HEAP8[$f64+1>>0]=HEAP8[$31+1>>0]|0;HEAP8[$f64+2>>0]=HEAP8[$31+2>>0]|0;HEAP8[$f64+3>>0]=HEAP8[$31+3>>0]|0;HEAP8[$f64+4>>0]=HEAP8[$31+4>>0]|0;HEAP8[$f64+5>>0]=HEAP8[$31+5>>0]|0;HEAP8[$f64+6>>0]=HEAP8[$31+6>>0]|0;HEAP8[$f64+7>>0]=HEAP8[$31+7>>0]|0;
  $32 = +HEAPF64[$f64>>3];
  $33 = $32 * 32767.5;
  $34 = (~~(($33)));
  HEAP16[$i16>>1] = $34;
  $35 = $0;
  $36 = $i;
  $37 = $36<<1;
  $38 = (($35) + ($37)|0);
  ;HEAP8[$38>>0]=HEAP8[$i16>>0]|0;HEAP8[$38+1>>0]=HEAP8[$i16+1>>0]|0;
  $39 = $i;
  $40 = (($39) + 1)|0;
  $i = $40;
 }
 $41 = $5;
 $42 = $41&255;
 $43 = ($42|0)!=(0);
 if (!($43)) {
  STACKTOP = sp;return;
 }
 $44 = $1;
 $45 = $i;
 $46 = $45<<3;
 $47 = (($44) + ($46)|0);
 ;HEAP8[$f64>>0]=HEAP8[$47>>0]|0;HEAP8[$f64+1>>0]=HEAP8[$47+1>>0]|0;HEAP8[$f64+2>>0]=HEAP8[$47+2>>0]|0;HEAP8[$f64+3>>0]=HEAP8[$47+3>>0]|0;HEAP8[$f64+4>>0]=HEAP8[$47+4>>0]|0;HEAP8[$f64+5>>0]=HEAP8[$47+5>>0]|0;HEAP8[$f64+6>>0]=HEAP8[$47+6>>0]|0;HEAP8[$f64+7>>0]=HEAP8[$47+7>>0]|0;
 $48 = +HEAPF64[$f64>>3];
 $49 = $48 * 32767.5;
 $50 = (~~(($49)));
 HEAP16[$i16>>1] = $50;
 $51 = $0;
 $52 = $i;
 $53 = $52<<1;
 $54 = (($51) + ($53)|0);
 $55 = $5;
 $56 = $55&255;
 _memcpy(($54|0),($i16|0),($56|0))|0;
 STACKTOP = sp;return;
}
function _wuff_float64_to_int24($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $f64 = 0, $i = 0, $i24 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f64 = sp;
 $i24 = sp + 12|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  ;HEAP8[$f64>>0]=HEAP8[$9>>0]|0;HEAP8[$f64+1>>0]=HEAP8[$9+1>>0]|0;HEAP8[$f64+2>>0]=HEAP8[$9+2>>0]|0;HEAP8[$f64+3>>0]=HEAP8[$9+3>>0]|0;HEAP8[$f64+4>>0]=HEAP8[$9+4>>0]|0;HEAP8[$f64+5>>0]=HEAP8[$9+5>>0]|0;HEAP8[$f64+6>>0]=HEAP8[$9+6>>0]|0;HEAP8[$f64+7>>0]=HEAP8[$9+7>>0]|0;
  $10 = +HEAPF64[$f64>>3];
  $11 = $10 * 2147483647.5;
  $12 = (~~(($11)));
  HEAP32[$i24>>2] = $12;
  $13 = $0;
  $14 = ((($i24)) + 1|0);
  $15 = $3;
  $16 = $15&255;
  $17 = (($14) + ($16)|0);
  $18 = $4;
  $19 = $18&255;
  _memcpy(($13|0),($17|0),($19|0))|0;
  $20 = $1;
  $21 = ((($20)) + 8|0);
  $1 = $21;
  $22 = $4;
  $23 = $22&255;
  $24 = $0;
  $25 = (($24) + ($23)|0);
  $0 = $25;
 }
 $i = 0;
 while(1) {
  $26 = $i;
  $27 = $2;
  $28 = ($26>>>0)<($27>>>0);
  if (!($28)) {
   break;
  }
  $29 = $1;
  $30 = $i;
  $31 = $30<<3;
  $32 = (($29) + ($31)|0);
  ;HEAP8[$f64>>0]=HEAP8[$32>>0]|0;HEAP8[$f64+1>>0]=HEAP8[$32+1>>0]|0;HEAP8[$f64+2>>0]=HEAP8[$32+2>>0]|0;HEAP8[$f64+3>>0]=HEAP8[$32+3>>0]|0;HEAP8[$f64+4>>0]=HEAP8[$32+4>>0]|0;HEAP8[$f64+5>>0]=HEAP8[$32+5>>0]|0;HEAP8[$f64+6>>0]=HEAP8[$32+6>>0]|0;HEAP8[$f64+7>>0]=HEAP8[$32+7>>0]|0;
  $33 = +HEAPF64[$f64>>3];
  $34 = $33 * 2147483647.5;
  $35 = (~~(($34)));
  HEAP32[$i24>>2] = $35;
  $36 = $0;
  $37 = $i;
  $38 = ($37*3)|0;
  $39 = (($36) + ($38)|0);
  $40 = ((($i24)) + 1|0);
  ;HEAP8[$39>>0]=HEAP8[$40>>0]|0;HEAP8[$39+1>>0]=HEAP8[$40+1>>0]|0;HEAP8[$39+2>>0]=HEAP8[$40+2>>0]|0;
  $41 = $i;
  $42 = (($41) + 1)|0;
  $i = $42;
 }
 $43 = $5;
 $44 = $43&255;
 $45 = ($44|0)!=(0);
 if (!($45)) {
  STACKTOP = sp;return;
 }
 $46 = $1;
 $47 = $2;
 $48 = $47<<3;
 $49 = (($46) + ($48)|0);
 ;HEAP8[$f64>>0]=HEAP8[$49>>0]|0;HEAP8[$f64+1>>0]=HEAP8[$49+1>>0]|0;HEAP8[$f64+2>>0]=HEAP8[$49+2>>0]|0;HEAP8[$f64+3>>0]=HEAP8[$49+3>>0]|0;HEAP8[$f64+4>>0]=HEAP8[$49+4>>0]|0;HEAP8[$f64+5>>0]=HEAP8[$49+5>>0]|0;HEAP8[$f64+6>>0]=HEAP8[$49+6>>0]|0;HEAP8[$f64+7>>0]=HEAP8[$49+7>>0]|0;
 $50 = +HEAPF64[$f64>>3];
 $51 = $50 * 2147483647.5;
 $52 = (~~(($51)));
 HEAP32[$i24>>2] = $52;
 $53 = $0;
 $54 = $2;
 $55 = ($54*3)|0;
 $56 = (($53) + ($55)|0);
 $57 = ((($i24)) + 1|0);
 $58 = $5;
 $59 = $58&255;
 _memcpy(($56|0),($57|0),($59|0))|0;
 STACKTOP = sp;return;
}
function _wuff_float64_to_int32($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f64 = 0, $i = 0, $i32 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f64 = sp;
 $i32 = sp + 12|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  ;HEAP8[$f64>>0]=HEAP8[$9>>0]|0;HEAP8[$f64+1>>0]=HEAP8[$9+1>>0]|0;HEAP8[$f64+2>>0]=HEAP8[$9+2>>0]|0;HEAP8[$f64+3>>0]=HEAP8[$9+3>>0]|0;HEAP8[$f64+4>>0]=HEAP8[$9+4>>0]|0;HEAP8[$f64+5>>0]=HEAP8[$9+5>>0]|0;HEAP8[$f64+6>>0]=HEAP8[$9+6>>0]|0;HEAP8[$f64+7>>0]=HEAP8[$9+7>>0]|0;
  $10 = +HEAPF64[$f64>>3];
  $11 = $10 * 2147483647.5;
  $12 = (~~(($11)));
  HEAP32[$i32>>2] = $12;
  $13 = $0;
  $14 = $3;
  $15 = $14&255;
  $16 = (($i32) + ($15)|0);
  $17 = $4;
  $18 = $17&255;
  _memcpy(($13|0),($16|0),($18|0))|0;
  $19 = $1;
  $20 = ((($19)) + 8|0);
  $1 = $20;
  $21 = $4;
  $22 = $21&255;
  $23 = $0;
  $24 = (($23) + ($22)|0);
  $0 = $24;
 }
 $i = 0;
 while(1) {
  $25 = $i;
  $26 = $2;
  $27 = ($25>>>0)<($26>>>0);
  if (!($27)) {
   break;
  }
  $28 = $1;
  $29 = $i;
  $30 = $29<<3;
  $31 = (($28) + ($30)|0);
  ;HEAP8[$f64>>0]=HEAP8[$31>>0]|0;HEAP8[$f64+1>>0]=HEAP8[$31+1>>0]|0;HEAP8[$f64+2>>0]=HEAP8[$31+2>>0]|0;HEAP8[$f64+3>>0]=HEAP8[$31+3>>0]|0;HEAP8[$f64+4>>0]=HEAP8[$31+4>>0]|0;HEAP8[$f64+5>>0]=HEAP8[$31+5>>0]|0;HEAP8[$f64+6>>0]=HEAP8[$31+6>>0]|0;HEAP8[$f64+7>>0]=HEAP8[$31+7>>0]|0;
  $32 = +HEAPF64[$f64>>3];
  $33 = $32 * 2147483647.5;
  $34 = (~~(($33)));
  HEAP32[$i32>>2] = $34;
  $35 = $0;
  $36 = $i;
  $37 = $36<<2;
  $38 = (($35) + ($37)|0);
  ;HEAP8[$38>>0]=HEAP8[$i32>>0]|0;HEAP8[$38+1>>0]=HEAP8[$i32+1>>0]|0;HEAP8[$38+2>>0]=HEAP8[$i32+2>>0]|0;HEAP8[$38+3>>0]=HEAP8[$i32+3>>0]|0;
  $39 = $i;
  $40 = (($39) + 1)|0;
  $i = $40;
 }
 $41 = $5;
 $42 = $41&255;
 $43 = ($42|0)!=(0);
 if (!($43)) {
  STACKTOP = sp;return;
 }
 $44 = $1;
 $45 = $2;
 $46 = $45<<3;
 $47 = (($44) + ($46)|0);
 ;HEAP8[$f64>>0]=HEAP8[$47>>0]|0;HEAP8[$f64+1>>0]=HEAP8[$47+1>>0]|0;HEAP8[$f64+2>>0]=HEAP8[$47+2>>0]|0;HEAP8[$f64+3>>0]=HEAP8[$47+3>>0]|0;HEAP8[$f64+4>>0]=HEAP8[$47+4>>0]|0;HEAP8[$f64+5>>0]=HEAP8[$47+5>>0]|0;HEAP8[$f64+6>>0]=HEAP8[$47+6>>0]|0;HEAP8[$f64+7>>0]=HEAP8[$47+7>>0]|0;
 $48 = +HEAPF64[$f64>>3];
 $49 = $48 * 2147483647.5;
 $50 = (~~(($49)));
 HEAP32[$i32>>2] = $50;
 $51 = $0;
 $52 = $2;
 $53 = $52<<2;
 $54 = (($51) + ($53)|0);
 $55 = $5;
 $56 = $55&255;
 _memcpy(($54|0),($i32|0),($56|0))|0;
 STACKTOP = sp;return;
}
function _wuff_float64_to_float32($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f32 = 0, $f64 = 0, $i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f64 = sp;
 $f32 = sp + 12|0;
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $4;
 $7 = $6&255;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = $1;
  ;HEAP8[$f64>>0]=HEAP8[$9>>0]|0;HEAP8[$f64+1>>0]=HEAP8[$9+1>>0]|0;HEAP8[$f64+2>>0]=HEAP8[$9+2>>0]|0;HEAP8[$f64+3>>0]=HEAP8[$9+3>>0]|0;HEAP8[$f64+4>>0]=HEAP8[$9+4>>0]|0;HEAP8[$f64+5>>0]=HEAP8[$9+5>>0]|0;HEAP8[$f64+6>>0]=HEAP8[$9+6>>0]|0;HEAP8[$f64+7>>0]=HEAP8[$9+7>>0]|0;
  $10 = +HEAPF64[$f64>>3];
  $11 = $10;
  HEAPF32[$f32>>2] = $11;
  $12 = $0;
  $13 = $3;
  $14 = $13&255;
  $15 = (($f32) + ($14)|0);
  $16 = $4;
  $17 = $16&255;
  _memcpy(($12|0),($15|0),($17|0))|0;
  $18 = $1;
  $19 = ((($18)) + 8|0);
  $1 = $19;
  $20 = $4;
  $21 = $20&255;
  $22 = $0;
  $23 = (($22) + ($21)|0);
  $0 = $23;
 }
 $i = 0;
 while(1) {
  $24 = $i;
  $25 = $2;
  $26 = ($24>>>0)<($25>>>0);
  if (!($26)) {
   break;
  }
  $27 = $1;
  $28 = $i;
  $29 = $28<<3;
  $30 = (($27) + ($29)|0);
  ;HEAP8[$f64>>0]=HEAP8[$30>>0]|0;HEAP8[$f64+1>>0]=HEAP8[$30+1>>0]|0;HEAP8[$f64+2>>0]=HEAP8[$30+2>>0]|0;HEAP8[$f64+3>>0]=HEAP8[$30+3>>0]|0;HEAP8[$f64+4>>0]=HEAP8[$30+4>>0]|0;HEAP8[$f64+5>>0]=HEAP8[$30+5>>0]|0;HEAP8[$f64+6>>0]=HEAP8[$30+6>>0]|0;HEAP8[$f64+7>>0]=HEAP8[$30+7>>0]|0;
  $31 = +HEAPF64[$f64>>3];
  $32 = $31;
  HEAPF32[$f32>>2] = $32;
  $33 = $0;
  $34 = $i;
  $35 = $34<<2;
  $36 = (($33) + ($35)|0);
  ;HEAP8[$36>>0]=HEAP8[$f32>>0]|0;HEAP8[$36+1>>0]=HEAP8[$f32+1>>0]|0;HEAP8[$36+2>>0]=HEAP8[$f32+2>>0]|0;HEAP8[$36+3>>0]=HEAP8[$f32+3>>0]|0;
  $37 = $i;
  $38 = (($37) + 1)|0;
  $i = $38;
 }
 $39 = $5;
 $40 = $39&255;
 $41 = ($40|0)!=(0);
 if (!($41)) {
  STACKTOP = sp;return;
 }
 $42 = $1;
 $43 = $2;
 $44 = $43<<3;
 $45 = (($42) + ($44)|0);
 ;HEAP8[$f64>>0]=HEAP8[$45>>0]|0;HEAP8[$f64+1>>0]=HEAP8[$45+1>>0]|0;HEAP8[$f64+2>>0]=HEAP8[$45+2>>0]|0;HEAP8[$f64+3>>0]=HEAP8[$45+3>>0]|0;HEAP8[$f64+4>>0]=HEAP8[$45+4>>0]|0;HEAP8[$f64+5>>0]=HEAP8[$45+5>>0]|0;HEAP8[$f64+6>>0]=HEAP8[$45+6>>0]|0;HEAP8[$f64+7>>0]=HEAP8[$45+7>>0]|0;
 $46 = +HEAPF64[$f64>>3];
 $47 = $46;
 HEAPF32[$f32>>2] = $47;
 $48 = $0;
 $49 = $2;
 $50 = $49<<2;
 $51 = (($48) + ($50)|0);
 $52 = $5;
 $53 = $52&255;
 _memcpy(($51|0),($f32|0),($53|0))|0;
 STACKTOP = sp;return;
}
function _wuff_float64_to_float64($dst,$src,$samples,$offset,$head,$tail) {
 $dst = $dst|0;
 $src = $src|0;
 $samples = $samples|0;
 $offset = $offset|0;
 $head = $head|0;
 $tail = $tail|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $dst;
 $1 = $src;
 $2 = $samples;
 $3 = $offset;
 $4 = $head;
 $5 = $tail;
 $6 = $0;
 $7 = $1;
 $8 = $3;
 $9 = $8&255;
 $10 = (($7) + ($9)|0);
 $11 = $2;
 $12 = $11<<3;
 $13 = $4;
 $14 = $13&255;
 $15 = (($12) + ($14))|0;
 $16 = $5;
 $17 = $16&255;
 $18 = (($15) + ($17))|0;
 _memcpy(($6|0),($10|0),($18|0))|0;
 STACKTOP = sp;return;
}
function __ZN4glad16gladLoadGLLoaderEPFPvPKcE($load) {
 $load = $load|0;
 var $$$i = 0, $$0 = 0, $$57$i = 0, $$59$i = 0, $$61$i = 0, $$63$i = 0, $$65$i = 0, $$67$i = 0, $$69$i = 0, $$71$i = 0, $$73$i = 0, $$75$i = 0, $$ph$i = 0, $$ph174$i = 0, $$phitmp47$i = 0, $$phitmp51$i = 0, $$phitmp55$i = 0, $$pre = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0;
 var $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0;
 var $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0;
 var $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0;
 var $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0, $1088 = 0, $1089 = 0;
 var $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0, $1106 = 0;
 var $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0, $1124 = 0;
 var $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0, $1142 = 0;
 var $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0, $116 = 0, $1160 = 0;
 var $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0, $1178 = 0, $1179 = 0;
 var $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0, $1196 = 0, $1197 = 0;
 var $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0, $1213 = 0, $1214 = 0;
 var $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0, $1231 = 0, $1232 = 0;
 var $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0, $125 = 0, $1250 = 0;
 var $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0, $1267 = 0, $1268 = 0, $1269 = 0;
 var $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0, $1285 = 0, $1286 = 0, $1287 = 0;
 var $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0, $1302 = 0, $1303 = 0, $1304 = 0;
 var $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0, $1320 = 0, $1321 = 0, $1322 = 0;
 var $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0, $1339 = 0, $134 = 0, $1340 = 0;
 var $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0, $1357 = 0, $1358 = 0, $1359 = 0;
 var $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0, $1375 = 0, $1376 = 0, $1377 = 0;
 var $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0, $1393 = 0, $1394 = 0, $1395 = 0;
 var $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0, $1410 = 0, $1411 = 0, $1412 = 0;
 var $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0, $1429 = 0, $143 = 0, $1430 = 0;
 var $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0, $1447 = 0, $1448 = 0, $1449 = 0;
 var $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0, $1465 = 0, $1466 = 0, $1467 = 0;
 var $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0, $1483 = 0, $1484 = 0, $1485 = 0;
 var $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0, $1500 = 0, $1501 = 0, $1502 = 0;
 var $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0, $1511 = 0, $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0, $1518 = 0, $1519 = 0, $152 = 0, $1520 = 0;
 var $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0, $153 = 0, $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0, $1536 = 0, $1537 = 0, $1538 = 0, $1539 = 0;
 var $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0, $1548 = 0, $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0, $1554 = 0, $1555 = 0, $1556 = 0, $1557 = 0;
 var $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $1565 = 0, $1566 = 0, $1567 = 0, $1568 = 0, $1569 = 0, $157 = 0, $1570 = 0, $1571 = 0, $1572 = 0, $1573 = 0, $1574 = 0, $1575 = 0;
 var $1576 = 0, $1577 = 0, $1578 = 0, $1579 = 0, $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0, $1583 = 0, $1584 = 0, $1585 = 0, $1586 = 0, $1587 = 0, $1588 = 0, $1589 = 0, $159 = 0, $1590 = 0, $1591 = 0, $1592 = 0, $1593 = 0;
 var $1594 = 0, $1595 = 0, $1596 = 0, $1597 = 0, $1598 = 0, $1599 = 0, $16 = 0, $160 = 0, $1600 = 0, $1601 = 0, $1602 = 0, $1603 = 0, $1604 = 0, $1605 = 0, $1606 = 0, $1607 = 0, $1608 = 0, $1609 = 0, $161 = 0, $1610 = 0;
 var $1611 = 0, $1612 = 0, $1613 = 0, $1614 = 0, $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0, $1619 = 0, $162 = 0, $1620 = 0, $1621 = 0, $1622 = 0, $1623 = 0, $1624 = 0, $1625 = 0, $1626 = 0, $1627 = 0, $1628 = 0, $1629 = 0;
 var $163 = 0, $1630 = 0, $1631 = 0, $1632 = 0, $1633 = 0, $1634 = 0, $1635 = 0, $1636 = 0, $1637 = 0, $1638 = 0, $1639 = 0, $164 = 0, $1640 = 0, $1641 = 0, $1642 = 0, $1643 = 0, $1644 = 0, $1645 = 0, $1646 = 0, $1647 = 0;
 var $1648 = 0, $1649 = 0, $165 = 0, $1650 = 0, $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0, $1655 = 0, $1656 = 0, $1657 = 0, $1658 = 0, $1659 = 0, $166 = 0, $1660 = 0, $1661 = 0, $1662 = 0, $1663 = 0, $1664 = 0, $1665 = 0;
 var $1666 = 0, $1667 = 0, $1668 = 0, $1669 = 0, $167 = 0, $1670 = 0, $1671 = 0, $1672 = 0, $1673 = 0, $1674 = 0, $1675 = 0, $1676 = 0, $1677 = 0, $1678 = 0, $1679 = 0, $168 = 0, $1680 = 0, $1681 = 0, $1682 = 0, $1683 = 0;
 var $1684 = 0, $1685 = 0, $1686 = 0, $1687 = 0, $1688 = 0, $1689 = 0, $169 = 0, $1690 = 0, $1691 = 0, $1692 = 0, $1693 = 0, $1694 = 0, $1695 = 0, $1696 = 0, $1697 = 0, $1698 = 0, $1699 = 0, $17 = 0, $170 = 0, $1700 = 0;
 var $1701 = 0, $1702 = 0, $1703 = 0, $1704 = 0, $1705 = 0, $1706 = 0, $1707 = 0, $1708 = 0, $1709 = 0, $171 = 0, $1710 = 0, $1711 = 0, $1712 = 0, $1713 = 0, $1714 = 0, $1715 = 0, $1716 = 0, $1717 = 0, $1718 = 0, $1719 = 0;
 var $172 = 0, $1720 = 0, $1721 = 0, $1722 = 0, $1723 = 0, $1724 = 0, $1725 = 0, $1726 = 0, $1727 = 0, $1728 = 0, $1729 = 0, $173 = 0, $1730 = 0, $1731 = 0, $1732 = 0, $1733 = 0, $1734 = 0, $1735 = 0, $1736 = 0, $1737 = 0;
 var $1738 = 0, $1739 = 0, $174 = 0, $1740 = 0, $1741 = 0, $1742 = 0, $1743 = 0, $1744 = 0, $1745 = 0, $1746 = 0, $1747 = 0, $1748 = 0, $1749 = 0, $175 = 0, $1750 = 0, $1751 = 0, $1752 = 0, $1753 = 0, $1754 = 0, $1755 = 0;
 var $1756 = 0, $1757 = 0, $1758 = 0, $1759 = 0, $176 = 0, $1760 = 0, $1761 = 0, $1762 = 0, $1763 = 0, $1764 = 0, $1765 = 0, $1766 = 0, $1767 = 0, $1768 = 0, $1769 = 0, $177 = 0, $1770 = 0, $1771 = 0, $1772 = 0, $1773 = 0;
 var $1774 = 0, $1775 = 0, $1776 = 0, $1777 = 0, $1778 = 0, $1779 = 0, $178 = 0, $1780 = 0, $1781 = 0, $1782 = 0, $1783 = 0, $1784 = 0, $1785 = 0, $1786 = 0, $1787 = 0, $1788 = 0, $1789 = 0, $179 = 0, $1790 = 0, $1791 = 0;
 var $1792 = 0, $1793 = 0, $1794 = 0, $1795 = 0, $1796 = 0, $1797 = 0, $1798 = 0, $1799 = 0, $18 = 0, $180 = 0, $1800 = 0, $1801 = 0, $1802 = 0, $1803 = 0, $1804 = 0, $1805 = 0, $1806 = 0, $1807 = 0, $1808 = 0, $1809 = 0;
 var $181 = 0, $1810 = 0, $1811 = 0, $1812 = 0, $1813 = 0, $1814 = 0, $1815 = 0, $1816 = 0, $1817 = 0, $1818 = 0, $1819 = 0, $182 = 0, $1820 = 0, $1821 = 0, $1822 = 0, $1823 = 0, $1824 = 0, $1825 = 0, $1826 = 0, $1827 = 0;
 var $1828 = 0, $1829 = 0, $183 = 0, $1830 = 0, $1831 = 0, $1832 = 0, $1833 = 0, $1834 = 0, $1835 = 0, $1836 = 0, $1837 = 0, $1838 = 0, $1839 = 0, $184 = 0, $1840 = 0, $1841 = 0, $1842 = 0, $1843 = 0, $1844 = 0, $1845 = 0;
 var $1846 = 0, $1847 = 0, $1848 = 0, $1849 = 0, $185 = 0, $1850 = 0, $1851 = 0, $1852 = 0, $1853 = 0, $1854 = 0, $1855 = 0, $1856 = 0, $1857 = 0, $1858 = 0, $1859 = 0, $186 = 0, $1860 = 0, $1861 = 0, $1862 = 0, $1863 = 0;
 var $1864 = 0, $1865 = 0, $1866 = 0, $1867 = 0, $1868 = 0, $1869 = 0, $187 = 0, $1870 = 0, $1871 = 0, $1872 = 0, $1873 = 0, $1874 = 0, $1875 = 0, $1876 = 0, $1877 = 0, $1878 = 0, $1879 = 0, $188 = 0, $1880 = 0, $1881 = 0;
 var $1882 = 0, $1883 = 0, $1884 = 0, $1885 = 0, $1886 = 0, $1887 = 0, $1888 = 0, $1889 = 0, $189 = 0, $1890 = 0, $1891 = 0, $1892 = 0, $1893 = 0, $1894 = 0, $1895 = 0, $1896 = 0, $1897 = 0, $1898 = 0, $1899 = 0, $19 = 0;
 var $190 = 0, $1900 = 0, $1901 = 0, $1902 = 0, $1903 = 0, $1904 = 0, $1905 = 0, $1906 = 0, $1907 = 0, $1908 = 0, $1909 = 0, $191 = 0, $1910 = 0, $1911 = 0, $1912 = 0, $1913 = 0, $1914 = 0, $1915 = 0, $1916 = 0, $1917 = 0;
 var $1918 = 0, $1919 = 0, $192 = 0, $1920 = 0, $1921 = 0, $1922 = 0, $1923 = 0, $1924 = 0, $1925 = 0, $1926 = 0, $1927 = 0, $1928 = 0, $1929 = 0, $193 = 0, $1930 = 0, $1931 = 0, $1932 = 0, $1933 = 0, $1934 = 0, $1935 = 0;
 var $1936 = 0, $1937 = 0, $1938 = 0, $1939 = 0, $194 = 0, $1940 = 0, $1941 = 0, $1942 = 0, $1943 = 0, $1944 = 0, $1945 = 0, $1946 = 0, $1947 = 0, $1948 = 0, $1949 = 0, $195 = 0, $1950 = 0, $1951 = 0, $1952 = 0, $1953 = 0;
 var $1954 = 0, $1955 = 0, $1956 = 0, $1957 = 0, $1958 = 0, $1959 = 0, $196 = 0, $1960 = 0, $1961 = 0, $1962 = 0, $1963 = 0, $1964 = 0, $1965 = 0, $1966 = 0, $1967 = 0, $1968 = 0, $1969 = 0, $197 = 0, $1970 = 0, $1971 = 0;
 var $1972 = 0, $1973 = 0, $1974 = 0, $1975 = 0, $1976 = 0, $1977 = 0, $1978 = 0, $1979 = 0, $198 = 0, $1980 = 0, $1981 = 0, $1982 = 0, $1983 = 0, $1984 = 0, $1985 = 0, $1986 = 0, $1987 = 0, $1988 = 0, $1989 = 0, $199 = 0;
 var $1990 = 0, $1991 = 0, $1992 = 0, $1993 = 0, $1994 = 0, $1995 = 0, $1996 = 0, $1997 = 0, $1998 = 0, $1999 = 0, $2 = 0, $20 = 0, $200 = 0, $2000 = 0, $2001 = 0, $2002 = 0, $2003 = 0, $2004 = 0, $2005 = 0, $2006 = 0;
 var $2007 = 0, $2008 = 0, $2009 = 0, $201 = 0, $2010 = 0, $2011 = 0, $2012 = 0, $2013 = 0, $2014 = 0, $2015 = 0, $2016 = 0, $2017 = 0, $2018 = 0, $2019 = 0, $202 = 0, $2020 = 0, $2021 = 0, $2022 = 0, $2023 = 0, $2024 = 0;
 var $2025 = 0, $2026 = 0, $2027 = 0, $2028 = 0, $2029 = 0, $203 = 0, $2030 = 0, $2031 = 0, $2032 = 0, $2033 = 0, $2034 = 0, $2035 = 0, $2036 = 0, $2037 = 0, $2038 = 0, $2039 = 0, $204 = 0, $2040 = 0, $2041 = 0, $2042 = 0;
 var $2043 = 0, $2044 = 0, $2045 = 0, $2046 = 0, $2047 = 0, $2048 = 0, $2049 = 0, $205 = 0, $2050 = 0, $2051 = 0, $2052 = 0, $2053 = 0, $2054 = 0, $2055 = 0, $2056 = 0, $2057 = 0, $2058 = 0, $2059 = 0, $206 = 0, $2060 = 0;
 var $2061 = 0, $2062 = 0, $2063 = 0, $2064 = 0, $2065 = 0, $2066 = 0, $2067 = 0, $2068 = 0, $2069 = 0, $207 = 0, $2070 = 0, $2071 = 0, $2072 = 0, $2073 = 0, $2074 = 0, $2075 = 0, $2076 = 0, $2077 = 0, $2078 = 0, $2079 = 0;
 var $208 = 0, $2080 = 0, $2081 = 0, $2082 = 0, $2083 = 0, $2084 = 0, $2085 = 0, $2086 = 0, $2087 = 0, $2088 = 0, $2089 = 0, $209 = 0, $2090 = 0, $2091 = 0, $2092 = 0, $2093 = 0, $2094 = 0, $2095 = 0, $2096 = 0, $2097 = 0;
 var $2098 = 0, $2099 = 0, $21 = 0, $210 = 0, $2100 = 0, $2101 = 0, $2102 = 0, $2103 = 0, $2104 = 0, $2105 = 0, $2106 = 0, $2107 = 0, $2108 = 0, $2109 = 0, $211 = 0, $2110 = 0, $2111 = 0, $2112 = 0, $2113 = 0, $2114 = 0;
 var $2115 = 0, $2116 = 0, $2117 = 0, $2118 = 0, $2119 = 0, $212 = 0, $2120 = 0, $2121 = 0, $2122 = 0, $2123 = 0, $2124 = 0, $2125 = 0, $2126 = 0, $2127 = 0, $2128 = 0, $2129 = 0, $213 = 0, $2130 = 0, $2131 = 0, $2132 = 0;
 var $2133 = 0, $2134 = 0, $2135 = 0, $2136 = 0, $2137 = 0, $2138 = 0, $2139 = 0, $214 = 0, $2140 = 0, $2141 = 0, $2142 = 0, $2143 = 0, $2144 = 0, $2145 = 0, $2146 = 0, $2147 = 0, $2148 = 0, $2149 = 0, $215 = 0, $2150 = 0;
 var $2151 = 0, $2152 = 0, $2153 = 0, $2154 = 0, $2155 = 0, $2156 = 0, $2157 = 0, $2158 = 0, $2159 = 0, $216 = 0, $2160 = 0, $2161 = 0, $2162 = 0, $2163 = 0, $2164 = 0, $2165 = 0, $2166 = 0, $2167 = 0, $2168 = 0, $2169 = 0;
 var $217 = 0, $2170 = 0, $2171 = 0, $2172 = 0, $2173 = 0, $2174 = 0, $2175 = 0, $2176 = 0, $2177 = 0, $2178 = 0, $2179 = 0, $218 = 0, $2180 = 0, $2181 = 0, $2182 = 0, $2183 = 0, $2184 = 0, $2185 = 0, $2186 = 0, $2187 = 0;
 var $2188 = 0, $2189 = 0, $219 = 0, $2190 = 0, $2191 = 0, $2192 = 0, $2193 = 0, $2194 = 0, $2195 = 0, $2196 = 0, $2197 = 0, $2198 = 0, $2199 = 0, $22 = 0, $220 = 0, $2200 = 0, $2201 = 0, $2202 = 0, $2203 = 0, $2204 = 0;
 var $2205 = 0, $2206 = 0, $2207 = 0, $2208 = 0, $2209 = 0, $221 = 0, $2210 = 0, $2211 = 0, $2212 = 0, $2213 = 0, $2214 = 0, $2215 = 0, $2216 = 0, $2217 = 0, $2218 = 0, $2219 = 0, $222 = 0, $2220 = 0, $2221 = 0, $2222 = 0;
 var $2223 = 0, $2224 = 0, $2225 = 0, $2226 = 0, $2227 = 0, $2228 = 0, $2229 = 0, $223 = 0, $2230 = 0, $2231 = 0, $2232 = 0, $2233 = 0, $2234 = 0, $2235 = 0, $2236 = 0, $2237 = 0, $2238 = 0, $2239 = 0, $224 = 0, $2240 = 0;
 var $2241 = 0, $2242 = 0, $2243 = 0, $2244 = 0, $2245 = 0, $2246 = 0, $2247 = 0, $2248 = 0, $2249 = 0, $225 = 0, $2250 = 0, $2251 = 0, $2252 = 0, $2253 = 0, $2254 = 0, $2255 = 0, $2256 = 0, $2257 = 0, $2258 = 0, $2259 = 0;
 var $226 = 0, $2260 = 0, $2261 = 0, $2262 = 0, $2263 = 0, $2264 = 0, $2265 = 0, $2266 = 0, $2267 = 0, $2268 = 0, $2269 = 0, $227 = 0, $2270 = 0, $2271 = 0, $2272 = 0, $2273 = 0, $2274 = 0, $2275 = 0, $2276 = 0, $2277 = 0;
 var $2278 = 0, $2279 = 0, $228 = 0, $2280 = 0, $2281 = 0, $2282 = 0, $2283 = 0, $2284 = 0, $2285 = 0, $2286 = 0, $2287 = 0, $2288 = 0, $2289 = 0, $229 = 0, $2290 = 0, $2291 = 0, $2292 = 0, $2293 = 0, $2294 = 0, $2295 = 0;
 var $2296 = 0, $2297 = 0, $2298 = 0, $2299 = 0, $23 = 0, $230 = 0, $2300 = 0, $2301 = 0, $2302 = 0, $2303 = 0, $2304 = 0, $2305 = 0, $2306 = 0, $2307 = 0, $2308 = 0, $2309 = 0, $231 = 0, $2310 = 0, $2311 = 0, $2312 = 0;
 var $2313 = 0, $2314 = 0, $2315 = 0, $2316 = 0, $2317 = 0, $2318 = 0, $2319 = 0, $232 = 0, $2320 = 0, $2321 = 0, $2322 = 0, $2323 = 0, $2324 = 0, $2325 = 0, $2326 = 0, $2327 = 0, $2328 = 0, $2329 = 0, $233 = 0, $2330 = 0;
 var $2331 = 0, $2332 = 0, $2333 = 0, $2334 = 0, $2335 = 0, $2336 = 0, $2337 = 0, $2338 = 0, $2339 = 0, $234 = 0, $2340 = 0, $2341 = 0, $2342 = 0, $2343 = 0, $2344 = 0, $2345 = 0, $2346 = 0, $2347 = 0, $2348 = 0, $2349 = 0;
 var $235 = 0, $2350 = 0, $2351 = 0, $2352 = 0, $2353 = 0, $2354 = 0, $2355 = 0, $2356 = 0, $2357 = 0, $2358 = 0, $2359 = 0, $236 = 0, $2360 = 0, $2361 = 0, $2362 = 0, $2363 = 0, $2364 = 0, $2365 = 0, $2366 = 0, $2367 = 0;
 var $2368 = 0, $2369 = 0, $237 = 0, $2370 = 0, $2371 = 0, $2372 = 0, $2373 = 0, $2374 = 0, $2375 = 0, $2376 = 0, $2377 = 0, $2378 = 0, $2379 = 0, $238 = 0, $2380 = 0, $2381 = 0, $2382 = 0, $2383 = 0, $2384 = 0, $2385 = 0;
 var $2386 = 0, $2387 = 0, $2388 = 0, $2389 = 0, $239 = 0, $2390 = 0, $2391 = 0, $2392 = 0, $2393 = 0, $2394 = 0, $2395 = 0, $2396 = 0, $2397 = 0, $2398 = 0, $2399 = 0, $24 = 0, $240 = 0, $2400 = 0, $2401 = 0, $2402 = 0;
 var $2403 = 0, $2404 = 0, $2405 = 0, $2406 = 0, $2407 = 0, $2408 = 0, $2409 = 0, $241 = 0, $2410 = 0, $2411 = 0, $2412 = 0, $2413 = 0, $2414 = 0, $2415 = 0, $2416 = 0, $2417 = 0, $2418 = 0, $2419 = 0, $242 = 0, $2420 = 0;
 var $2421 = 0, $2422 = 0, $2423 = 0, $2424 = 0, $2425 = 0, $2426 = 0, $2427 = 0, $2428 = 0, $2429 = 0, $243 = 0, $2430 = 0, $2431 = 0, $2432 = 0, $2433 = 0, $2434 = 0, $2435 = 0, $2436 = 0, $2437 = 0, $2438 = 0, $2439 = 0;
 var $244 = 0, $2440 = 0, $2441 = 0, $2442 = 0, $2443 = 0, $2444 = 0, $2445 = 0, $2446 = 0, $2447 = 0, $2448 = 0, $2449 = 0, $245 = 0, $2450 = 0, $2451 = 0, $2452 = 0, $2453 = 0, $2454 = 0, $2455 = 0, $2456 = 0, $2457 = 0;
 var $2458 = 0, $2459 = 0, $246 = 0, $2460 = 0, $2461 = 0, $2462 = 0, $2463 = 0, $2464 = 0, $2465 = 0, $2466 = 0, $2467 = 0, $2468 = 0, $2469 = 0, $247 = 0, $2470 = 0, $2471 = 0, $2472 = 0, $2473 = 0, $2474 = 0, $2475 = 0;
 var $2476 = 0, $2477 = 0, $2478 = 0, $2479 = 0, $248 = 0, $2480 = 0, $2481 = 0, $2482 = 0, $2483 = 0, $2484 = 0, $2485 = 0, $2486 = 0, $2487 = 0, $2488 = 0, $2489 = 0, $249 = 0, $2490 = 0, $2491 = 0, $2492 = 0, $2493 = 0;
 var $2494 = 0, $2495 = 0, $2496 = 0, $2497 = 0, $2498 = 0, $2499 = 0, $25 = 0, $250 = 0, $2500 = 0, $2501 = 0, $2502 = 0, $2503 = 0, $2504 = 0, $2505 = 0, $2506 = 0, $2507 = 0, $2508 = 0, $2509 = 0, $251 = 0, $2510 = 0;
 var $2511 = 0, $2512 = 0, $2513 = 0, $2514 = 0, $2515 = 0, $2516 = 0, $2517 = 0, $2518 = 0, $2519 = 0, $252 = 0, $2520 = 0, $2521 = 0, $2522 = 0, $2523 = 0, $2524 = 0, $2525 = 0, $2526 = 0, $2527 = 0, $2528 = 0, $2529 = 0;
 var $253 = 0, $2530 = 0, $2531 = 0, $2532 = 0, $2533 = 0, $2534 = 0, $2535 = 0, $2536 = 0, $2537 = 0, $2538 = 0, $2539 = 0, $254 = 0, $2540 = 0, $2541 = 0, $2542 = 0, $2543 = 0, $2544 = 0, $2545 = 0, $2546 = 0, $2547 = 0;
 var $2548 = 0, $2549 = 0, $255 = 0, $2550 = 0, $2551 = 0, $2552 = 0, $2553 = 0, $2554 = 0, $2555 = 0, $2556 = 0, $2557 = 0, $2558 = 0, $2559 = 0, $256 = 0, $2560 = 0, $2561 = 0, $2562 = 0, $2563 = 0, $2564 = 0, $2565 = 0;
 var $2566 = 0, $2567 = 0, $2568 = 0, $2569 = 0, $257 = 0, $2570 = 0, $2571 = 0, $2572 = 0, $2573 = 0, $2574 = 0, $2575 = 0, $2576 = 0, $2577 = 0, $2578 = 0, $2579 = 0, $258 = 0, $2580 = 0, $2581 = 0, $2582 = 0, $2583 = 0;
 var $2584 = 0, $2585 = 0, $2586 = 0, $2587 = 0, $2588 = 0, $2589 = 0, $259 = 0, $2590 = 0, $2591 = 0, $2592 = 0, $2593 = 0, $2594 = 0, $2595 = 0, $2596 = 0, $2597 = 0, $2598 = 0, $2599 = 0, $26 = 0, $260 = 0, $2600 = 0;
 var $2601 = 0, $2602 = 0, $2603 = 0, $2604 = 0, $2605 = 0, $2606 = 0, $2607 = 0, $2608 = 0, $2609 = 0, $261 = 0, $2610 = 0, $2611 = 0, $2612 = 0, $2613 = 0, $2614 = 0, $2615 = 0, $2616 = 0, $2617 = 0, $2618 = 0, $2619 = 0;
 var $262 = 0, $2620 = 0, $2621 = 0, $2622 = 0, $2623 = 0, $2624 = 0, $2625 = 0, $2626 = 0, $2627 = 0, $2628 = 0, $2629 = 0, $263 = 0, $2630 = 0, $2631 = 0, $2632 = 0, $2633 = 0, $2634 = 0, $2635 = 0, $2636 = 0, $2637 = 0;
 var $2638 = 0, $2639 = 0, $264 = 0, $2640 = 0, $2641 = 0, $2642 = 0, $2643 = 0, $2644 = 0, $2645 = 0, $2646 = 0, $2647 = 0, $2648 = 0, $2649 = 0, $265 = 0, $2650 = 0, $2651 = 0, $2652 = 0, $2653 = 0, $2654 = 0, $2655 = 0;
 var $2656 = 0, $2657 = 0, $2658 = 0, $2659 = 0, $266 = 0, $2660 = 0, $2661 = 0, $2662 = 0, $2663 = 0, $2664 = 0, $2665 = 0, $2666 = 0, $2667 = 0, $2668 = 0, $2669 = 0, $267 = 0, $2670 = 0, $2671 = 0, $2672 = 0, $2673 = 0;
 var $2674 = 0, $2675 = 0, $2676 = 0, $2677 = 0, $2678 = 0, $2679 = 0, $268 = 0, $2680 = 0, $2681 = 0, $2682 = 0, $2683 = 0, $2684 = 0, $2685 = 0, $2686 = 0, $2687 = 0, $2688 = 0, $2689 = 0, $269 = 0, $2690 = 0, $2691 = 0;
 var $2692 = 0, $2693 = 0, $2694 = 0, $2695 = 0, $2696 = 0, $2697 = 0, $2698 = 0, $2699 = 0, $27 = 0, $270 = 0, $2700 = 0, $2701 = 0, $2702 = 0, $2703 = 0, $2704 = 0, $2705 = 0, $2706 = 0, $2707 = 0, $2708 = 0, $2709 = 0;
 var $271 = 0, $2710 = 0, $2711 = 0, $2712 = 0, $2713 = 0, $2714 = 0, $2715 = 0, $2716 = 0, $2717 = 0, $2718 = 0, $2719 = 0, $272 = 0, $2720 = 0, $2721 = 0, $2722 = 0, $2723 = 0, $2724 = 0, $2725 = 0, $2726 = 0, $2727 = 0;
 var $2728 = 0, $2729 = 0, $273 = 0, $2730 = 0, $2731 = 0, $2732 = 0, $2733 = 0, $2734 = 0, $2735 = 0, $2736 = 0, $2737 = 0, $2738 = 0, $2739 = 0, $274 = 0, $2740 = 0, $2741 = 0, $2742 = 0, $2743 = 0, $2744 = 0, $2745 = 0;
 var $2746 = 0, $2747 = 0, $2748 = 0, $2749 = 0, $275 = 0, $2750 = 0, $2751 = 0, $2752 = 0, $2753 = 0, $2754 = 0, $2755 = 0, $2756 = 0, $2757 = 0, $2758 = 0, $2759 = 0, $276 = 0, $2760 = 0, $2761 = 0, $2762 = 0, $2763 = 0;
 var $2764 = 0, $2765 = 0, $2766 = 0, $2767 = 0, $2768 = 0, $2769 = 0, $277 = 0, $2770 = 0, $2771 = 0, $2772 = 0, $2773 = 0, $2774 = 0, $2775 = 0, $2776 = 0, $2777 = 0, $2778 = 0, $2779 = 0, $278 = 0, $2780 = 0, $2781 = 0;
 var $2782 = 0, $2783 = 0, $2784 = 0, $2785 = 0, $2786 = 0, $2787 = 0, $2788 = 0, $2789 = 0, $279 = 0, $2790 = 0, $2791 = 0, $2792 = 0, $2793 = 0, $2794 = 0, $2795 = 0, $2796 = 0, $2797 = 0, $2798 = 0, $2799 = 0, $28 = 0;
 var $280 = 0, $2800 = 0, $2801 = 0, $2802 = 0, $2803 = 0, $2804 = 0, $2805 = 0, $2806 = 0, $2807 = 0, $2808 = 0, $2809 = 0, $281 = 0, $2810 = 0, $2811 = 0, $2812 = 0, $2813 = 0, $2814 = 0, $2815 = 0, $2816 = 0, $2817 = 0;
 var $2818 = 0, $2819 = 0, $282 = 0, $2820 = 0, $2821 = 0, $2822 = 0, $2823 = 0, $2824 = 0, $2825 = 0, $2826 = 0, $2827 = 0, $2828 = 0, $2829 = 0, $283 = 0, $2830 = 0, $2831 = 0, $2832 = 0, $2833 = 0, $2834 = 0, $2835 = 0;
 var $2836 = 0, $2837 = 0, $2838 = 0, $2839 = 0, $284 = 0, $2840 = 0, $2841 = 0, $2842 = 0, $2843 = 0, $2844 = 0, $2845 = 0, $2846 = 0, $2847 = 0, $2848 = 0, $2849 = 0, $285 = 0, $2850 = 0, $2851 = 0, $2852 = 0, $2853 = 0;
 var $2854 = 0, $2855 = 0, $2856 = 0, $2857 = 0, $2858 = 0, $2859 = 0, $286 = 0, $2860 = 0, $2861 = 0, $2862 = 0, $2863 = 0, $2864 = 0, $2865 = 0, $2866 = 0, $2867 = 0, $2868 = 0, $2869 = 0, $287 = 0, $2870 = 0, $2871 = 0;
 var $2872 = 0, $2873 = 0, $2874 = 0, $2875 = 0, $2876 = 0, $2877 = 0, $2878 = 0, $2879 = 0, $288 = 0, $2880 = 0, $2881 = 0, $2882 = 0, $2883 = 0, $2884 = 0, $2885 = 0, $2886 = 0, $2887 = 0, $2888 = 0, $2889 = 0, $289 = 0;
 var $2890 = 0, $2891 = 0, $2892 = 0, $2893 = 0, $2894 = 0, $2895 = 0, $2896 = 0, $2897 = 0, $2898 = 0, $2899 = 0, $29 = 0, $290 = 0, $2900 = 0, $2901 = 0, $2902 = 0, $2903 = 0, $2904 = 0, $2905 = 0, $2906 = 0, $2907 = 0;
 var $2908 = 0, $2909 = 0, $291 = 0, $2910 = 0, $2911 = 0, $2912 = 0, $2913 = 0, $2914 = 0, $2915 = 0, $2916 = 0, $2917 = 0, $2918 = 0, $2919 = 0, $292 = 0, $2920 = 0, $2921 = 0, $2922 = 0, $2923 = 0, $2924 = 0, $2925 = 0;
 var $2926 = 0, $2927 = 0, $2928 = 0, $2929 = 0, $293 = 0, $2930 = 0, $2931 = 0, $2932 = 0, $2933 = 0, $2934 = 0, $2935 = 0, $2936 = 0, $2937 = 0, $2938 = 0, $2939 = 0, $294 = 0, $2940 = 0, $2941 = 0, $2942 = 0, $2943 = 0;
 var $2944 = 0, $2945 = 0, $2946 = 0, $2947 = 0, $2948 = 0, $2949 = 0, $295 = 0, $2950 = 0, $2951 = 0, $2952 = 0, $2953 = 0, $2954 = 0, $2955 = 0, $2956 = 0, $2957 = 0, $2958 = 0, $2959 = 0, $296 = 0, $2960 = 0, $2961 = 0;
 var $2962 = 0, $2963 = 0, $2964 = 0, $2965 = 0, $2966 = 0, $2967 = 0, $2968 = 0, $2969 = 0, $297 = 0, $2970 = 0, $2971 = 0, $2972 = 0, $2973 = 0, $2974 = 0, $2975 = 0, $2976 = 0, $2977 = 0, $2978 = 0, $2979 = 0, $298 = 0;
 var $2980 = 0, $2981 = 0, $2982 = 0, $2983 = 0, $2984 = 0, $2985 = 0, $2986 = 0, $2987 = 0, $2988 = 0, $2989 = 0, $299 = 0, $2990 = 0, $2991 = 0, $2992 = 0, $2993 = 0, $2994 = 0, $2995 = 0, $2996 = 0, $2997 = 0, $2998 = 0;
 var $2999 = 0, $3 = 0, $30 = 0, $300 = 0, $3000 = 0, $3001 = 0, $3002 = 0, $3003 = 0, $3004 = 0, $3005 = 0, $3006 = 0, $3007 = 0, $3008 = 0, $3009 = 0, $301 = 0, $3010 = 0, $3011 = 0, $3012 = 0, $3013 = 0, $3014 = 0;
 var $3015 = 0, $3016 = 0, $3017 = 0, $3018 = 0, $3019 = 0, $302 = 0, $3020 = 0, $3021 = 0, $3022 = 0, $3023 = 0, $3024 = 0, $3025 = 0, $3026 = 0, $3027 = 0, $3028 = 0, $3029 = 0, $303 = 0, $3030 = 0, $3031 = 0, $3032 = 0;
 var $3033 = 0, $3034 = 0, $3035 = 0, $3036 = 0, $3037 = 0, $3038 = 0, $3039 = 0, $304 = 0, $3040 = 0, $3041 = 0, $3042 = 0, $3043 = 0, $3044 = 0, $3045 = 0, $3046 = 0, $3047 = 0, $3048 = 0, $3049 = 0, $305 = 0, $3050 = 0;
 var $3051 = 0, $3052 = 0, $3053 = 0, $3054 = 0, $3055 = 0, $3056 = 0, $3057 = 0, $3058 = 0, $3059 = 0, $306 = 0, $3060 = 0, $3061 = 0, $3062 = 0, $3063 = 0, $3064 = 0, $3065 = 0, $3066 = 0, $3067 = 0, $3068 = 0, $3069 = 0;
 var $307 = 0, $3070 = 0, $3071 = 0, $3072 = 0, $3073 = 0, $3074 = 0, $3075 = 0, $3076 = 0, $3077 = 0, $3078 = 0, $3079 = 0, $308 = 0, $3080 = 0, $3081 = 0, $3082 = 0, $3083 = 0, $3084 = 0, $3085 = 0, $3086 = 0, $3087 = 0;
 var $3088 = 0, $3089 = 0, $309 = 0, $3090 = 0, $3091 = 0, $3092 = 0, $3093 = 0, $3094 = 0, $3095 = 0, $3096 = 0, $3097 = 0, $3098 = 0, $3099 = 0, $31 = 0, $310 = 0, $3100 = 0, $3101 = 0, $3102 = 0, $3103 = 0, $3104 = 0;
 var $3105 = 0, $3106 = 0, $3107 = 0, $3108 = 0, $3109 = 0, $311 = 0, $3110 = 0, $3111 = 0, $3112 = 0, $3113 = 0, $3114 = 0, $3115 = 0, $3116 = 0, $3117 = 0, $3118 = 0, $3119 = 0, $312 = 0, $3120 = 0, $3121 = 0, $3122 = 0;
 var $3123 = 0, $3124 = 0, $3125 = 0, $3126 = 0, $3127 = 0, $3128 = 0, $3129 = 0, $313 = 0, $3130 = 0, $3131 = 0, $3132 = 0, $3133 = 0, $3134 = 0, $3135 = 0, $3136 = 0, $3137 = 0, $3138 = 0, $3139 = 0, $314 = 0, $3140 = 0;
 var $3141 = 0, $3142 = 0, $3143 = 0, $3144 = 0, $3145 = 0, $3146 = 0, $3147 = 0, $3148 = 0, $3149 = 0, $315 = 0, $3150 = 0, $3151 = 0, $3152 = 0, $3153 = 0, $3154 = 0, $3155 = 0, $3156 = 0, $3157 = 0, $3158 = 0, $3159 = 0;
 var $316 = 0, $3160 = 0, $3161 = 0, $3162 = 0, $3163 = 0, $3164 = 0, $3165 = 0, $3166 = 0, $3167 = 0, $3168 = 0, $3169 = 0, $317 = 0, $3170 = 0, $3171 = 0, $3172 = 0, $3173 = 0, $3174 = 0, $3175 = 0, $3176 = 0, $3177 = 0;
 var $3178 = 0, $3179 = 0, $318 = 0, $3180 = 0, $3181 = 0, $3182 = 0, $3183 = 0, $3184 = 0, $3185 = 0, $3186 = 0, $3187 = 0, $3188 = 0, $3189 = 0, $319 = 0, $3190 = 0, $3191 = 0, $3192 = 0, $3193 = 0, $3194 = 0, $3195 = 0;
 var $3196 = 0, $3197 = 0, $3198 = 0, $3199 = 0, $32 = 0, $320 = 0, $3200 = 0, $3201 = 0, $3202 = 0, $3203 = 0, $3204 = 0, $3205 = 0, $3206 = 0, $3207 = 0, $3208 = 0, $3209 = 0, $321 = 0, $3210 = 0, $3211 = 0, $3212 = 0;
 var $3213 = 0, $3214 = 0, $3215 = 0, $3216 = 0, $3217 = 0, $3218 = 0, $3219 = 0, $322 = 0, $3220 = 0, $3221 = 0, $3222 = 0, $3223 = 0, $3224 = 0, $3225 = 0, $3226 = 0, $3227 = 0, $3228 = 0, $3229 = 0, $323 = 0, $3230 = 0;
 var $3231 = 0, $3232 = 0, $3233 = 0, $3234 = 0, $3235 = 0, $3236 = 0, $3237 = 0, $3238 = 0, $3239 = 0, $324 = 0, $3240 = 0, $3241 = 0, $3242 = 0, $3243 = 0, $3244 = 0, $3245 = 0, $3246 = 0, $3247 = 0, $3248 = 0, $3249 = 0;
 var $325 = 0, $3250 = 0, $3251 = 0, $3252 = 0, $3253 = 0, $3254 = 0, $3255 = 0, $3256 = 0, $3257 = 0, $3258 = 0, $3259 = 0, $326 = 0, $3260 = 0, $3261 = 0, $3262 = 0, $3263 = 0, $3264 = 0, $3265 = 0, $3266 = 0, $3267 = 0;
 var $3268 = 0, $3269 = 0, $327 = 0, $3270 = 0, $3271 = 0, $3272 = 0, $3273 = 0, $3274 = 0, $3275 = 0, $3276 = 0, $3277 = 0, $3278 = 0, $3279 = 0, $328 = 0, $3280 = 0, $3281 = 0, $3282 = 0, $3283 = 0, $3284 = 0, $3285 = 0;
 var $3286 = 0, $3287 = 0, $3288 = 0, $3289 = 0, $329 = 0, $3290 = 0, $3291 = 0, $3292 = 0, $3293 = 0, $3294 = 0, $3295 = 0, $3296 = 0, $3297 = 0, $3298 = 0, $3299 = 0, $33 = 0, $330 = 0, $3300 = 0, $3301 = 0, $3302 = 0;
 var $3303 = 0, $3304 = 0, $3305 = 0, $3306 = 0, $3307 = 0, $3308 = 0, $3309 = 0, $331 = 0, $3310 = 0, $3311 = 0, $3312 = 0, $3313 = 0, $3314 = 0, $3315 = 0, $3316 = 0, $3317 = 0, $3318 = 0, $3319 = 0, $332 = 0, $3320 = 0;
 var $3321 = 0, $3322 = 0, $3323 = 0, $3324 = 0, $3325 = 0, $3326 = 0, $3327 = 0, $3328 = 0, $3329 = 0, $333 = 0, $3330 = 0, $3331 = 0, $3332 = 0, $3333 = 0, $3334 = 0, $3335 = 0, $3336 = 0, $3337 = 0, $3338 = 0, $3339 = 0;
 var $334 = 0, $3340 = 0, $3341 = 0, $3342 = 0, $3343 = 0, $3344 = 0, $3345 = 0, $3346 = 0, $3347 = 0, $3348 = 0, $3349 = 0, $335 = 0, $3350 = 0, $3351 = 0, $3352 = 0, $3353 = 0, $3354 = 0, $3355 = 0, $3356 = 0, $3357 = 0;
 var $3358 = 0, $3359 = 0, $336 = 0, $3360 = 0, $3361 = 0, $3362 = 0, $3363 = 0, $3364 = 0, $3365 = 0, $3366 = 0, $3367 = 0, $3368 = 0, $3369 = 0, $337 = 0, $3370 = 0, $3371 = 0, $3372 = 0, $3373 = 0, $3374 = 0, $3375 = 0;
 var $3376 = 0, $3377 = 0, $3378 = 0, $3379 = 0, $338 = 0, $3380 = 0, $3381 = 0, $3382 = 0, $3383 = 0, $3384 = 0, $3385 = 0, $3386 = 0, $3387 = 0, $3388 = 0, $3389 = 0, $339 = 0, $3390 = 0, $3391 = 0, $3392 = 0, $3393 = 0;
 var $3394 = 0, $3395 = 0, $3396 = 0, $3397 = 0, $3398 = 0, $3399 = 0, $34 = 0, $340 = 0, $3400 = 0, $3401 = 0, $3402 = 0, $3403 = 0, $3404 = 0, $3405 = 0, $3406 = 0, $3407 = 0, $3408 = 0, $3409 = 0, $341 = 0, $3410 = 0;
 var $3411 = 0, $3412 = 0, $3413 = 0, $3414 = 0, $3415 = 0, $3416 = 0, $3417 = 0, $3418 = 0, $3419 = 0, $342 = 0, $3420 = 0, $3421 = 0, $3422 = 0, $3423 = 0, $3424 = 0, $3425 = 0, $3426 = 0, $3427 = 0, $3428 = 0, $3429 = 0;
 var $343 = 0, $3430 = 0, $3431 = 0, $3432 = 0, $3433 = 0, $3434 = 0, $3435 = 0, $3436 = 0, $3437 = 0, $3438 = 0, $3439 = 0, $344 = 0, $3440 = 0, $3441 = 0, $3442 = 0, $3443 = 0, $3444 = 0, $3445 = 0, $3446 = 0, $3447 = 0;
 var $3448 = 0, $3449 = 0, $345 = 0, $3450 = 0, $3451 = 0, $3452 = 0, $3453 = 0, $3454 = 0, $3455 = 0, $3456 = 0, $3457 = 0, $3458 = 0, $3459 = 0, $346 = 0, $3460 = 0, $3461 = 0, $3462 = 0, $3463 = 0, $3464 = 0, $3465 = 0;
 var $3466 = 0, $3467 = 0, $3468 = 0, $3469 = 0, $347 = 0, $3470 = 0, $3471 = 0, $3472 = 0, $3473 = 0, $3474 = 0, $3475 = 0, $3476 = 0, $3477 = 0, $3478 = 0, $3479 = 0, $348 = 0, $3480 = 0, $3481 = 0, $3482 = 0, $3483 = 0;
 var $3484 = 0, $3485 = 0, $3486 = 0, $3487 = 0, $3488 = 0, $3489 = 0, $349 = 0, $3490 = 0, $3491 = 0, $3492 = 0, $3493 = 0, $3494 = 0, $3495 = 0, $3496 = 0, $3497 = 0, $3498 = 0, $3499 = 0, $35 = 0, $350 = 0, $3500 = 0;
 var $3501 = 0, $3502 = 0, $3503 = 0, $3504 = 0, $3505 = 0, $3506 = 0, $3507 = 0, $3508 = 0, $3509 = 0, $351 = 0, $3510 = 0, $3511 = 0, $3512 = 0, $3513 = 0, $3514 = 0, $3515 = 0, $3516 = 0, $3517 = 0, $3518 = 0, $3519 = 0;
 var $352 = 0, $3520 = 0, $3521 = 0, $3522 = 0, $3523 = 0, $3524 = 0, $3525 = 0, $3526 = 0, $3527 = 0, $3528 = 0, $3529 = 0, $353 = 0, $3530 = 0, $3531 = 0, $3532 = 0, $3533 = 0, $3534 = 0, $3535 = 0, $3536 = 0, $3537 = 0;
 var $3538 = 0, $3539 = 0, $354 = 0, $3540 = 0, $3541 = 0, $3542 = 0, $3543 = 0, $3544 = 0, $3545 = 0, $3546 = 0, $3547 = 0, $3548 = 0, $3549 = 0, $355 = 0, $3550 = 0, $3551 = 0, $3552 = 0, $3553 = 0, $3554 = 0, $3555 = 0;
 var $3556 = 0, $3557 = 0, $3558 = 0, $3559 = 0, $356 = 0, $3560 = 0, $3561 = 0, $3562 = 0, $3563 = 0, $3564 = 0, $3565 = 0, $3566 = 0, $3567 = 0, $3568 = 0, $3569 = 0, $357 = 0, $3570 = 0, $3571 = 0, $358 = 0, $359 = 0;
 var $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0;
 var $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0;
 var $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0;
 var $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0;
 var $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0;
 var $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0;
 var $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0;
 var $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0;
 var $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0;
 var $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0;
 var $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0;
 var $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0;
 var $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0;
 var $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0;
 var $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0;
 var $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0;
 var $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0;
 var $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0;
 var $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0;
 var $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0;
 var $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0;
 var $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0;
 var $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0;
 var $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0;
 var $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0;
 var $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0;
 var $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0;
 var $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0;
 var $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0;
 var $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0;
 var $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0;
 var $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0;
 var $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0;
 var $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0;
 var $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0;
 var $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $cmp$i = 0, $cmp46$i = 0, $gles$0$i = 0, $major$0$i = 0, $major$0$in$i = 0, $major$188$ph$ph$i = 0, $minor$0$in$i = 0, $minor$1$i = 0, $minor$1$in90$ph$ph$i = 0, $minor$186$i = 0;
 var $minor$194$ph$ph$i = 0, $or$cond$i = 0, $or$cond11$i = 0, $or$cond13$i = 0, $or$cond15$i = 0, $or$cond17$i = 0, $or$cond19$i = 0, $or$cond21$i = 0, $or$cond23$i = 0, $or$cond25$i = 0, $or$cond27$i = 0, $or$cond29$i = 0, $or$cond3$i = 0, $or$cond31$i = 0, $or$cond33$i = 0, $or$cond35$i = 0, $or$cond37$i = 0, $or$cond39$i = 0, $or$cond41$i = 0, $or$cond43$i = 0;
 var $or$cond5$i = 0, $or$cond7$i = 0, $or$cond9$i = 0, $phitmp$i = 0, $phitmp47$i = 0, $phitmp53$i = 0, $phitmp54$i = 0, $storemerge$i = 0, $strncmp$i = 0, $strncmp45$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[166754] = 0; //@line 7276 "libraries/glad/glad.cpp"
 $0 = (FUNCTION_TABLE_ii[$load & 8191](613977)|0); //@line 7277 "libraries/glad/glad.cpp"
 HEAP32[165140] = $0; //@line 7277 "libraries/glad/glad.cpp"
 $1 = ($0|0)==(0|0); //@line 7278 "libraries/glad/glad.cpp"
 if ($1) {
  $$0 = 0;
  return ($$0|0); //@line 7537 "libraries/glad/glad.cpp"
 }
 $2 = (FUNCTION_TABLE_ii[$0 & 8191](7938)|0); //@line 7240 "libraries/glad/glad.cpp"
 $3 = HEAP8[$2>>0]|0; //@line 7241 "libraries/glad/glad.cpp"
 $4 = $3 << 24 >> 24; //@line 7241 "libraries/glad/glad.cpp"
 $5 = ((($2)) + 2|0); //@line 7241 "libraries/glad/glad.cpp"
 $6 = HEAP8[$5>>0]|0; //@line 7241 "libraries/glad/glad.cpp"
 $7 = $6 << 24 >> 24; //@line 7241 "libraries/glad/glad.cpp"
 $strncmp$i = (_strncmp($2,447689,10)|0); //@line 7243 "libraries/glad/glad.cpp"
 $cmp$i = ($strncmp$i|0)==(0); //@line 7243 "libraries/glad/glad.cpp"
 if ($cmp$i) {
  $8 = ((($2)) + 10|0); //@line 7244 "libraries/glad/glad.cpp"
  $9 = HEAP8[$8>>0]|0; //@line 7244 "libraries/glad/glad.cpp"
  $10 = $9 << 24 >> 24; //@line 7244 "libraries/glad/glad.cpp"
  $11 = ((($2)) + 12|0); //@line 7244 "libraries/glad/glad.cpp"
  $12 = HEAP8[$11>>0]|0; //@line 7244 "libraries/glad/glad.cpp"
  $13 = $12 << 24 >> 24; //@line 7244 "libraries/glad/glad.cpp"
  $gles$0$i = 1;$major$0$in$i = $10;$minor$0$in$i = $13;
 } else {
  $gles$0$i = 0;$major$0$in$i = $4;$minor$0$in$i = $7;
 }
 $major$0$i = (($major$0$in$i) + -48)|0; //@line 7244 "libraries/glad/glad.cpp"
 $strncmp45$i = (_strncmp($2,447700,6)|0); //@line 7246 "libraries/glad/glad.cpp"
 $cmp46$i = ($strncmp45$i|0)==(0); //@line 7246 "libraries/glad/glad.cpp"
 if ($cmp46$i) {
  $14 = ((($2)) + 6|0); //@line 7247 "libraries/glad/glad.cpp"
  $15 = HEAP8[$14>>0]|0; //@line 7247 "libraries/glad/glad.cpp"
  $16 = $15 << 24 >> 24; //@line 7247 "libraries/glad/glad.cpp"
  $17 = ((($2)) + 8|0); //@line 7247 "libraries/glad/glad.cpp"
  $18 = HEAP8[$17>>0]|0; //@line 7247 "libraries/glad/glad.cpp"
  $19 = $18 << 24 >> 24; //@line 7247 "libraries/glad/glad.cpp"
  $20 = (($16) + -47)|0; //@line 7248 "libraries/glad/glad.cpp"
  $minor$186$i = (($19) + -48)|0; //@line 7247 "libraries/glad/glad.cpp"
  HEAP32[166754] = $20; //@line 7250 "libraries/glad/glad.cpp"
  $major$188$ph$ph$i = $20;$minor$1$in90$ph$ph$i = $19;$minor$194$ph$ph$i = $minor$186$i;
  label = 7;
 } else {
  $minor$1$i = (($minor$0$in$i) + -48)|0; //@line 7247 "libraries/glad/glad.cpp"
  HEAP32[166754] = $major$0$i; //@line 7250 "libraries/glad/glad.cpp"
  $21 = ($gles$0$i|0)==(0); //@line 7251 "libraries/glad/glad.cpp"
  if ($21) {
   $31 = ($major$0$i|0)==(1); //@line 7251 "libraries/glad/glad.cpp"
   $32 = ($minor$1$i|0)>(-1); //@line 7251 "libraries/glad/glad.cpp"
   $or$cond$i = $31 & $32; //@line 7251 "libraries/glad/glad.cpp"
   $33 = ($major$0$i|0)>(1); //@line 7251 "libraries/glad/glad.cpp"
   $$$i = $33 | $or$cond$i; //@line 7251 "libraries/glad/glad.cpp"
   $34 = $$$i&1; //@line 7251 "libraries/glad/glad.cpp"
   HEAP8[1251994] = $34; //@line 7251 "libraries/glad/glad.cpp"
   $35 = ($minor$0$in$i|0)>(48); //@line 7252 "libraries/glad/glad.cpp"
   $or$cond3$i = $31 & $35; //@line 7252 "libraries/glad/glad.cpp"
   $$57$i = $33 | $or$cond3$i; //@line 7252 "libraries/glad/glad.cpp"
   $36 = $$57$i&1; //@line 7252 "libraries/glad/glad.cpp"
   HEAP8[1251995] = $36; //@line 7252 "libraries/glad/glad.cpp"
   $37 = ($minor$1$i|0)>(1); //@line 7253 "libraries/glad/glad.cpp"
   $or$cond5$i = $31 & $37; //@line 7253 "libraries/glad/glad.cpp"
   $$59$i = $33 | $or$cond5$i; //@line 7253 "libraries/glad/glad.cpp"
   $38 = $$59$i&1; //@line 7253 "libraries/glad/glad.cpp"
   HEAP8[1251996] = $38; //@line 7253 "libraries/glad/glad.cpp"
   $39 = ($minor$1$i|0)>(2); //@line 7254 "libraries/glad/glad.cpp"
   $or$cond7$i = $31 & $39; //@line 7254 "libraries/glad/glad.cpp"
   $$61$i = $33 | $or$cond7$i; //@line 7254 "libraries/glad/glad.cpp"
   $40 = $$61$i&1; //@line 7254 "libraries/glad/glad.cpp"
   HEAP8[1251997] = $40; //@line 7254 "libraries/glad/glad.cpp"
   $41 = ($minor$1$i|0)>(3); //@line 7255 "libraries/glad/glad.cpp"
   $or$cond9$i = $31 & $41; //@line 7255 "libraries/glad/glad.cpp"
   $$63$i = $33 | $or$cond9$i; //@line 7255 "libraries/glad/glad.cpp"
   $42 = $$63$i&1; //@line 7255 "libraries/glad/glad.cpp"
   HEAP8[1251998] = $42; //@line 7255 "libraries/glad/glad.cpp"
   $43 = ($minor$1$i|0)>(4); //@line 7256 "libraries/glad/glad.cpp"
   $or$cond11$i = $31 & $43; //@line 7256 "libraries/glad/glad.cpp"
   $$65$i = $33 | $or$cond11$i; //@line 7256 "libraries/glad/glad.cpp"
   $44 = $$65$i&1; //@line 7256 "libraries/glad/glad.cpp"
   HEAP8[1251999] = $44; //@line 7256 "libraries/glad/glad.cpp"
   $45 = ($major$0$i|0)==(2); //@line 7257 "libraries/glad/glad.cpp"
   $or$cond13$i = $45 & $32; //@line 7257 "libraries/glad/glad.cpp"
   $46 = ($major$0$i|0)>(2); //@line 7257 "libraries/glad/glad.cpp"
   $$67$i = $46 | $or$cond13$i; //@line 7257 "libraries/glad/glad.cpp"
   $47 = $$67$i&1; //@line 7257 "libraries/glad/glad.cpp"
   HEAP8[1252000] = $47; //@line 7257 "libraries/glad/glad.cpp"
   $or$cond15$i = $45 & $35; //@line 7258 "libraries/glad/glad.cpp"
   $$69$i = $46 | $or$cond15$i; //@line 7258 "libraries/glad/glad.cpp"
   $48 = $$69$i&1; //@line 7258 "libraries/glad/glad.cpp"
   HEAP8[1252001] = $48; //@line 7258 "libraries/glad/glad.cpp"
   $49 = ($major$0$i|0)==(3); //@line 7259 "libraries/glad/glad.cpp"
   $or$cond17$i = $49 & $32; //@line 7259 "libraries/glad/glad.cpp"
   $50 = ($major$0$i|0)>(3); //@line 7259 "libraries/glad/glad.cpp"
   $$71$i = $50 | $or$cond17$i; //@line 7259 "libraries/glad/glad.cpp"
   $51 = $$71$i&1; //@line 7259 "libraries/glad/glad.cpp"
   HEAP8[1252002] = $51; //@line 7259 "libraries/glad/glad.cpp"
   $or$cond19$i = $49 & $35; //@line 7260 "libraries/glad/glad.cpp"
   $$73$i = $50 | $or$cond19$i; //@line 7260 "libraries/glad/glad.cpp"
   $52 = $$73$i&1; //@line 7260 "libraries/glad/glad.cpp"
   HEAP8[1252003] = $52; //@line 7260 "libraries/glad/glad.cpp"
   $or$cond21$i = $49 & $37; //@line 7261 "libraries/glad/glad.cpp"
   $$75$i = $50 | $or$cond21$i; //@line 7261 "libraries/glad/glad.cpp"
   $53 = $$75$i&1; //@line 7261 "libraries/glad/glad.cpp"
   HEAP8[1252004] = $53; //@line 7261 "libraries/glad/glad.cpp"
   $or$cond23$i = $49 & $39; //@line 7262 "libraries/glad/glad.cpp"
   $phitmp$i = $50&1; //@line 7262 "libraries/glad/glad.cpp"
   $$ph$i = $or$cond23$i ? 1 : $phitmp$i; //@line 7262 "libraries/glad/glad.cpp"
   HEAP8[1252005] = $$ph$i; //@line 7262 "libraries/glad/glad.cpp"
   $54 = ($major$0$i|0)==(4); //@line 7263 "libraries/glad/glad.cpp"
   $or$cond25$i = $54 & $32; //@line 7263 "libraries/glad/glad.cpp"
   $55 = ($major$0$i|0)>(4); //@line 7263 "libraries/glad/glad.cpp"
   $phitmp47$i = $55&1; //@line 7263 "libraries/glad/glad.cpp"
   $56 = $or$cond25$i ? 1 : $phitmp47$i; //@line 7263 "libraries/glad/glad.cpp"
   HEAP8[1252006] = $56; //@line 7263 "libraries/glad/glad.cpp"
   $or$cond27$i = $54 & $35; //@line 7264 "libraries/glad/glad.cpp"
   $$phitmp47$i = $or$cond27$i ? 1 : $phitmp47$i; //@line 7264 "libraries/glad/glad.cpp"
   HEAP8[1252007] = $$phitmp47$i; //@line 7264 "libraries/glad/glad.cpp"
   $or$cond29$i = $54 & $37; //@line 7265 "libraries/glad/glad.cpp"
   $57 = $or$cond29$i ? 1 : $phitmp47$i; //@line 7265 "libraries/glad/glad.cpp"
   HEAP8[1252008] = $57; //@line 7265 "libraries/glad/glad.cpp"
   $or$cond31$i = $54 & $39; //@line 7266 "libraries/glad/glad.cpp"
   $$ph174$i = $or$cond31$i ? 1 : $phitmp47$i; //@line 7266 "libraries/glad/glad.cpp"
   HEAP8[1252009] = $$ph174$i; //@line 7266 "libraries/glad/glad.cpp"
   $or$cond33$i = $54 & $41; //@line 7267 "libraries/glad/glad.cpp"
   $58 = $or$cond33$i ? 1 : $phitmp47$i; //@line 7267 "libraries/glad/glad.cpp"
   HEAP8[1252010] = $58; //@line 7267 "libraries/glad/glad.cpp"
   $or$cond35$i = $54 & $43; //@line 7268 "libraries/glad/glad.cpp"
   $$phitmp51$i = $or$cond35$i ? 1 : $phitmp47$i; //@line 7268 "libraries/glad/glad.cpp"
   HEAP8[1252011] = $$phitmp51$i; //@line 7268 "libraries/glad/glad.cpp"
   HEAP8[1252012] = 0; //@line 7269 "libraries/glad/glad.cpp"
   HEAP8[1252013] = 0; //@line 7270 "libraries/glad/glad.cpp"
   HEAP8[1252014] = 0; //@line 7271 "libraries/glad/glad.cpp"
   HEAP8[1252015] = 0; //@line 7272 "libraries/glad/glad.cpp"
   if ($$$i) {
    $59 = (FUNCTION_TABLE_ii[$load & 8191](616030)|0); //@line 2212 "libraries/glad/glad.cpp"
    HEAP32[165205] = $59; //@line 2212 "libraries/glad/glad.cpp"
    $60 = (FUNCTION_TABLE_ii[$load & 8191](616018)|0); //@line 2213 "libraries/glad/glad.cpp"
    HEAP32[165653] = $60; //@line 2213 "libraries/glad/glad.cpp"
    $61 = (FUNCTION_TABLE_ii[$load & 8191](616273)|0); //@line 2214 "libraries/glad/glad.cpp"
    HEAP32[165134] = $61; //@line 2214 "libraries/glad/glad.cpp"
    $62 = (FUNCTION_TABLE_ii[$load & 8191](616049)|0); //@line 2215 "libraries/glad/glad.cpp"
    HEAP32[164641] = $62; //@line 2215 "libraries/glad/glad.cpp"
    $63 = (FUNCTION_TABLE_ii[$load & 8191](447707)|0); //@line 2216 "libraries/glad/glad.cpp"
    HEAP32[164694] = $63; //@line 2216 "libraries/glad/glad.cpp"
    $64 = (FUNCTION_TABLE_ii[$load & 8191](447719)|0); //@line 2217 "libraries/glad/glad.cpp"
    HEAP32[165045] = $64; //@line 2217 "libraries/glad/glad.cpp"
    $65 = (FUNCTION_TABLE_ii[$load & 8191](616429)|0); //@line 2218 "libraries/glad/glad.cpp"
    HEAP32[165383] = $65; //@line 2218 "libraries/glad/glad.cpp"
    $66 = (FUNCTION_TABLE_ii[$load & 8191](616346)|0); //@line 2219 "libraries/glad/glad.cpp"
    HEAP32[165106] = $66; //@line 2219 "libraries/glad/glad.cpp"
    $67 = (FUNCTION_TABLE_ii[$load & 8191](614205)|0); //@line 2220 "libraries/glad/glad.cpp"
    HEAP32[165617] = $67; //@line 2220 "libraries/glad/glad.cpp"
    $68 = (FUNCTION_TABLE_ii[$load & 8191](616330)|0); //@line 2221 "libraries/glad/glad.cpp"
    HEAP32[165108] = $68; //@line 2221 "libraries/glad/glad.cpp"
    $69 = (FUNCTION_TABLE_ii[$load & 8191](614222)|0); //@line 2222 "libraries/glad/glad.cpp"
    HEAP32[164788] = $69; //@line 2222 "libraries/glad/glad.cpp"
    $70 = (FUNCTION_TABLE_ii[$load & 8191](447733)|0); //@line 2223 "libraries/glad/glad.cpp"
    HEAP32[164785] = $70; //@line 2223 "libraries/glad/glad.cpp"
    $71 = (FUNCTION_TABLE_ii[$load & 8191](614109)|0); //@line 2224 "libraries/glad/glad.cpp"
    HEAP32[165602] = $71; //@line 2224 "libraries/glad/glad.cpp"
    $72 = (FUNCTION_TABLE_ii[$load & 8191](447746)|0); //@line 2225 "libraries/glad/glad.cpp"
    HEAP32[164868] = $72; //@line 2225 "libraries/glad/glad.cpp"
    $73 = (FUNCTION_TABLE_ii[$load & 8191](616041)|0); //@line 2226 "libraries/glad/glad.cpp"
    HEAP32[164768] = $73; //@line 2226 "libraries/glad/glad.cpp"
    $74 = (FUNCTION_TABLE_ii[$load & 8191](616416)|0); //@line 2227 "libraries/glad/glad.cpp"
    HEAP32[165395] = $74; //@line 2227 "libraries/glad/glad.cpp"
    $75 = (FUNCTION_TABLE_ii[$load & 8191](616061)|0); //@line 2228 "libraries/glad/glad.cpp"
    HEAP32[164916] = $75; //@line 2228 "libraries/glad/glad.cpp"
    $76 = (FUNCTION_TABLE_ii[$load & 8191](615960)|0); //@line 2229 "libraries/glad/glad.cpp"
    HEAP32[165104] = $76; //@line 2229 "libraries/glad/glad.cpp"
    $77 = (FUNCTION_TABLE_ii[$load & 8191](616088)|0); //@line 2230 "libraries/glad/glad.cpp"
    HEAP32[165603] = $77; //@line 2230 "libraries/glad/glad.cpp"
    $78 = (FUNCTION_TABLE_ii[$load & 8191](616456)|0); //@line 2231 "libraries/glad/glad.cpp"
    HEAP32[165466] = $78; //@line 2231 "libraries/glad/glad.cpp"
    $79 = (FUNCTION_TABLE_ii[$load & 8191](616076)|0); //@line 2232 "libraries/glad/glad.cpp"
    HEAP32[165644] = $79; //@line 2232 "libraries/glad/glad.cpp"
    $80 = (FUNCTION_TABLE_ii[$load & 8191](616008)|0); //@line 2233 "libraries/glad/glad.cpp"
    HEAP32[165193] = $80; //@line 2233 "libraries/glad/glad.cpp"
    $81 = (FUNCTION_TABLE_ii[$load & 8191](615999)|0); //@line 2234 "libraries/glad/glad.cpp"
    HEAP32[164832] = $81; //@line 2234 "libraries/glad/glad.cpp"
    $82 = (FUNCTION_TABLE_ii[$load & 8191](615943)|0); //@line 2235 "libraries/glad/glad.cpp"
    HEAP32[165082] = $82; //@line 2235 "libraries/glad/glad.cpp"
    $83 = (FUNCTION_TABLE_ii[$load & 8191](615952)|0); //@line 2236 "libraries/glad/glad.cpp"
    HEAP32[164883] = $83; //@line 2236 "libraries/glad/glad.cpp"
    $84 = (FUNCTION_TABLE_ii[$load & 8191](616188)|0); //@line 2237 "libraries/glad/glad.cpp"
    HEAP32[165364] = $84; //@line 2237 "libraries/glad/glad.cpp"
    $85 = (FUNCTION_TABLE_ii[$load & 8191](447759)|0); //@line 2238 "libraries/glad/glad.cpp"
    HEAP32[165194] = $85; //@line 2238 "libraries/glad/glad.cpp"
    $86 = (FUNCTION_TABLE_ii[$load & 8191](616379)|0); //@line 2239 "libraries/glad/glad.cpp"
    HEAP32[165322] = $86; //@line 2239 "libraries/glad/glad.cpp"
    $87 = (FUNCTION_TABLE_ii[$load & 8191](616393)|0); //@line 2240 "libraries/glad/glad.cpp"
    HEAP32[164775] = $87; //@line 2240 "libraries/glad/glad.cpp"
    $88 = (FUNCTION_TABLE_ii[$load & 8191](615987)|0); //@line 2241 "libraries/glad/glad.cpp"
    HEAP32[165448] = $88; //@line 2241 "libraries/glad/glad.cpp"
    $89 = (FUNCTION_TABLE_ii[$load & 8191](447769)|0); //@line 2242 "libraries/glad/glad.cpp"
    HEAP32[165116] = $89; //@line 2242 "libraries/glad/glad.cpp"
    $90 = (FUNCTION_TABLE_ii[$load & 8191](613963)|0); //@line 2243 "libraries/glad/glad.cpp"
    HEAP32[165018] = $90; //@line 2243 "libraries/glad/glad.cpp"
    $91 = (FUNCTION_TABLE_ii[$load & 8191](447783)|0); //@line 2244 "libraries/glad/glad.cpp"
    HEAP32[165245] = $91; //@line 2244 "libraries/glad/glad.cpp"
    $92 = (FUNCTION_TABLE_ii[$load & 8191](614138)|0); //@line 2245 "libraries/glad/glad.cpp"
    HEAP32[165294] = $92; //@line 2245 "libraries/glad/glad.cpp"
    $93 = (FUNCTION_TABLE_ii[$load & 8191](614015)|0); //@line 2246 "libraries/glad/glad.cpp"
    HEAP32[165084] = $93; //@line 2246 "libraries/glad/glad.cpp"
    $94 = (FUNCTION_TABLE_ii[$load & 8191](447796)|0); //@line 2247 "libraries/glad/glad.cpp"
    HEAP32[165004] = $94; //@line 2247 "libraries/glad/glad.cpp"
    $95 = (FUNCTION_TABLE_ii[$load & 8191](615864)|0); //@line 2248 "libraries/glad/glad.cpp"
    HEAP32[165571] = $95; //@line 2248 "libraries/glad/glad.cpp"
    $96 = (FUNCTION_TABLE_ii[$load & 8191](614003)|0); //@line 2249 "libraries/glad/glad.cpp"
    HEAP32[165575] = $96; //@line 2249 "libraries/glad/glad.cpp"
    $97 = (FUNCTION_TABLE_ii[$load & 8191](613989)|0); //@line 2250 "libraries/glad/glad.cpp"
    HEAP32[165587] = $97; //@line 2250 "libraries/glad/glad.cpp"
    $98 = (FUNCTION_TABLE_ii[$load & 8191](613977)|0); //@line 2251 "libraries/glad/glad.cpp"
    HEAP32[165140] = $98; //@line 2251 "libraries/glad/glad.cpp"
    $99 = (FUNCTION_TABLE_ii[$load & 8191](447809)|0); //@line 2252 "libraries/glad/glad.cpp"
    HEAP32[164792] = $99; //@line 2252 "libraries/glad/glad.cpp"
    $100 = (FUNCTION_TABLE_ii[$load & 8191](614165)|0); //@line 2253 "libraries/glad/glad.cpp"
    HEAP32[165456] = $100; //@line 2253 "libraries/glad/glad.cpp"
    $101 = (FUNCTION_TABLE_ii[$load & 8191](614185)|0); //@line 2254 "libraries/glad/glad.cpp"
    HEAP32[165170] = $101; //@line 2254 "libraries/glad/glad.cpp"
    $102 = (FUNCTION_TABLE_ii[$load & 8191](447823)|0); //@line 2255 "libraries/glad/glad.cpp"
    HEAP32[165581] = $102; //@line 2255 "libraries/glad/glad.cpp"
    $103 = (FUNCTION_TABLE_ii[$load & 8191](447848)|0); //@line 2256 "libraries/glad/glad.cpp"
    HEAP32[164903] = $103; //@line 2256 "libraries/glad/glad.cpp"
    $104 = (FUNCTION_TABLE_ii[$load & 8191](616176)|0); //@line 2257 "libraries/glad/glad.cpp"
    HEAP32[164774] = $104; //@line 2257 "libraries/glad/glad.cpp"
    $105 = (FUNCTION_TABLE_ii[$load & 8191](616224)|0); //@line 2258 "libraries/glad/glad.cpp"
    HEAP32[164864] = $105; //@line 2258 "libraries/glad/glad.cpp"
    $106 = (FUNCTION_TABLE_ii[$load & 8191](616405)|0); //@line 2259 "libraries/glad/glad.cpp"
    HEAP32[165094] = $106; //@line 2259 "libraries/glad/glad.cpp"
    $107 = (FUNCTION_TABLE_ii[$load & 8191](447873)|0); //@line 2260 "libraries/glad/glad.cpp"
    HEAP32[165129] = $107; //@line 2260 "libraries/glad/glad.cpp"
    $108 = (FUNCTION_TABLE_ii[$load & 8191](447883)|0); //@line 2261 "libraries/glad/glad.cpp"
    HEAP32[165525] = $108; //@line 2261 "libraries/glad/glad.cpp"
    $109 = (FUNCTION_TABLE_ii[$load & 8191](447893)|0); //@line 2262 "libraries/glad/glad.cpp"
    HEAP32[165001] = $109; //@line 2262 "libraries/glad/glad.cpp"
    $110 = (FUNCTION_TABLE_ii[$load & 8191](447904)|0); //@line 2263 "libraries/glad/glad.cpp"
    HEAP32[165020] = $110; //@line 2263 "libraries/glad/glad.cpp"
    $111 = (FUNCTION_TABLE_ii[$load & 8191](447916)|0); //@line 2264 "libraries/glad/glad.cpp"
    HEAP32[165256] = $111; //@line 2264 "libraries/glad/glad.cpp"
    $112 = (FUNCTION_TABLE_ii[$load & 8191](447930)|0); //@line 2265 "libraries/glad/glad.cpp"
    HEAP32[165515] = $112; //@line 2265 "libraries/glad/glad.cpp"
    $113 = (FUNCTION_TABLE_ii[$load & 8191](447941)|0); //@line 2266 "libraries/glad/glad.cpp"
    HEAP32[165450] = $113; //@line 2266 "libraries/glad/glad.cpp"
    $114 = (FUNCTION_TABLE_ii[$load & 8191](447952)|0); //@line 2267 "libraries/glad/glad.cpp"
    HEAP32[164753] = $114; //@line 2267 "libraries/glad/glad.cpp"
    $115 = (FUNCTION_TABLE_ii[$load & 8191](447960)|0); //@line 2268 "libraries/glad/glad.cpp"
    HEAP32[164806] = $115; //@line 2268 "libraries/glad/glad.cpp"
    $116 = (FUNCTION_TABLE_ii[$load & 8191](447969)|0); //@line 2269 "libraries/glad/glad.cpp"
    HEAP32[164706] = $116; //@line 2269 "libraries/glad/glad.cpp"
    $117 = (FUNCTION_TABLE_ii[$load & 8191](447979)|0); //@line 2270 "libraries/glad/glad.cpp"
    HEAP32[165516] = $117; //@line 2270 "libraries/glad/glad.cpp"
    $118 = (FUNCTION_TABLE_ii[$load & 8191](447990)|0); //@line 2271 "libraries/glad/glad.cpp"
    HEAP32[165208] = $118; //@line 2271 "libraries/glad/glad.cpp"
    $119 = (FUNCTION_TABLE_ii[$load & 8191](448000)|0); //@line 2272 "libraries/glad/glad.cpp"
    HEAP32[165168] = $119; //@line 2272 "libraries/glad/glad.cpp"
    $120 = (FUNCTION_TABLE_ii[$load & 8191](448011)|0); //@line 2273 "libraries/glad/glad.cpp"
    HEAP32[165200] = $120; //@line 2273 "libraries/glad/glad.cpp"
    $121 = (FUNCTION_TABLE_ii[$load & 8191](448021)|0); //@line 2274 "libraries/glad/glad.cpp"
    HEAP32[165646] = $121; //@line 2274 "libraries/glad/glad.cpp"
    $122 = (FUNCTION_TABLE_ii[$load & 8191](448032)|0); //@line 2275 "libraries/glad/glad.cpp"
    HEAP32[164714] = $122; //@line 2275 "libraries/glad/glad.cpp"
    $123 = (FUNCTION_TABLE_ii[$load & 8191](448042)|0); //@line 2276 "libraries/glad/glad.cpp"
    HEAP32[165493] = $123; //@line 2276 "libraries/glad/glad.cpp"
    $124 = (FUNCTION_TABLE_ii[$load & 8191](448053)|0); //@line 2277 "libraries/glad/glad.cpp"
    HEAP32[164718] = $124; //@line 2277 "libraries/glad/glad.cpp"
    $125 = (FUNCTION_TABLE_ii[$load & 8191](448063)|0); //@line 2278 "libraries/glad/glad.cpp"
    HEAP32[164740] = $125; //@line 2278 "libraries/glad/glad.cpp"
    $126 = (FUNCTION_TABLE_ii[$load & 8191](448074)|0); //@line 2279 "libraries/glad/glad.cpp"
    HEAP32[165452] = $126; //@line 2279 "libraries/glad/glad.cpp"
    $127 = (FUNCTION_TABLE_ii[$load & 8191](448085)|0); //@line 2280 "libraries/glad/glad.cpp"
    HEAP32[164755] = $127; //@line 2280 "libraries/glad/glad.cpp"
    $128 = (FUNCTION_TABLE_ii[$load & 8191](448097)|0); //@line 2281 "libraries/glad/glad.cpp"
    HEAP32[165457] = $128; //@line 2281 "libraries/glad/glad.cpp"
    $129 = (FUNCTION_TABLE_ii[$load & 8191](448108)|0); //@line 2282 "libraries/glad/glad.cpp"
    HEAP32[165445] = $129; //@line 2282 "libraries/glad/glad.cpp"
    $130 = (FUNCTION_TABLE_ii[$load & 8191](448120)|0); //@line 2283 "libraries/glad/glad.cpp"
    HEAP32[165444] = $130; //@line 2283 "libraries/glad/glad.cpp"
    $131 = (FUNCTION_TABLE_ii[$load & 8191](448131)|0); //@line 2284 "libraries/glad/glad.cpp"
    HEAP32[165531] = $131; //@line 2284 "libraries/glad/glad.cpp"
    $132 = (FUNCTION_TABLE_ii[$load & 8191](448143)|0); //@line 2285 "libraries/glad/glad.cpp"
    HEAP32[165067] = $132; //@line 2285 "libraries/glad/glad.cpp"
    $133 = (FUNCTION_TABLE_ii[$load & 8191](448153)|0); //@line 2286 "libraries/glad/glad.cpp"
    HEAP32[164698] = $133; //@line 2286 "libraries/glad/glad.cpp"
    $134 = (FUNCTION_TABLE_ii[$load & 8191](448164)|0); //@line 2287 "libraries/glad/glad.cpp"
    HEAP32[165070] = $134; //@line 2287 "libraries/glad/glad.cpp"
    $135 = (FUNCTION_TABLE_ii[$load & 8191](448174)|0); //@line 2288 "libraries/glad/glad.cpp"
    HEAP32[165414] = $135; //@line 2288 "libraries/glad/glad.cpp"
    $136 = (FUNCTION_TABLE_ii[$load & 8191](448185)|0); //@line 2289 "libraries/glad/glad.cpp"
    HEAP32[165069] = $136; //@line 2289 "libraries/glad/glad.cpp"
    $137 = (FUNCTION_TABLE_ii[$load & 8191](448195)|0); //@line 2290 "libraries/glad/glad.cpp"
    HEAP32[164834] = $137; //@line 2290 "libraries/glad/glad.cpp"
    $138 = (FUNCTION_TABLE_ii[$load & 8191](448206)|0); //@line 2291 "libraries/glad/glad.cpp"
    HEAP32[165071] = $138; //@line 2291 "libraries/glad/glad.cpp"
    $139 = (FUNCTION_TABLE_ii[$load & 8191](448216)|0); //@line 2292 "libraries/glad/glad.cpp"
    HEAP32[165262] = $139; //@line 2292 "libraries/glad/glad.cpp"
    $140 = (FUNCTION_TABLE_ii[$load & 8191](448227)|0); //@line 2293 "libraries/glad/glad.cpp"
    HEAP32[165056] = $140; //@line 2293 "libraries/glad/glad.cpp"
    $141 = (FUNCTION_TABLE_ii[$load & 8191](448237)|0); //@line 2294 "libraries/glad/glad.cpp"
    HEAP32[165480] = $141; //@line 2294 "libraries/glad/glad.cpp"
    $142 = (FUNCTION_TABLE_ii[$load & 8191](448248)|0); //@line 2295 "libraries/glad/glad.cpp"
    HEAP32[165124] = $142; //@line 2295 "libraries/glad/glad.cpp"
    $143 = (FUNCTION_TABLE_ii[$load & 8191](448259)|0); //@line 2296 "libraries/glad/glad.cpp"
    HEAP32[165343] = $143; //@line 2296 "libraries/glad/glad.cpp"
    $144 = (FUNCTION_TABLE_ii[$load & 8191](448271)|0); //@line 2297 "libraries/glad/glad.cpp"
    HEAP32[164658] = $144; //@line 2297 "libraries/glad/glad.cpp"
    $145 = (FUNCTION_TABLE_ii[$load & 8191](448282)|0); //@line 2298 "libraries/glad/glad.cpp"
    HEAP32[165326] = $145; //@line 2298 "libraries/glad/glad.cpp"
    $146 = (FUNCTION_TABLE_ii[$load & 8191](448294)|0); //@line 2299 "libraries/glad/glad.cpp"
    HEAP32[165131] = $146; //@line 2299 "libraries/glad/glad.cpp"
    $147 = (FUNCTION_TABLE_ii[$load & 8191](448305)|0); //@line 2300 "libraries/glad/glad.cpp"
    HEAP32[165621] = $147; //@line 2300 "libraries/glad/glad.cpp"
    $148 = (FUNCTION_TABLE_ii[$load & 8191](448317)|0); //@line 2301 "libraries/glad/glad.cpp"
    HEAP32[164708] = $148; //@line 2301 "libraries/glad/glad.cpp"
    $149 = (FUNCTION_TABLE_ii[$load & 8191](448328)|0); //@line 2302 "libraries/glad/glad.cpp"
    HEAP32[165142] = $149; //@line 2302 "libraries/glad/glad.cpp"
    $150 = (FUNCTION_TABLE_ii[$load & 8191](448340)|0); //@line 2303 "libraries/glad/glad.cpp"
    HEAP32[165380] = $150; //@line 2303 "libraries/glad/glad.cpp"
    $151 = (FUNCTION_TABLE_ii[$load & 8191](448346)|0); //@line 2304 "libraries/glad/glad.cpp"
    HEAP32[164637] = $151; //@line 2304 "libraries/glad/glad.cpp"
    $152 = (FUNCTION_TABLE_ii[$load & 8191](448355)|0); //@line 2305 "libraries/glad/glad.cpp"
    HEAP32[165097] = $152; //@line 2305 "libraries/glad/glad.cpp"
    $153 = (FUNCTION_TABLE_ii[$load & 8191](448365)|0); //@line 2306 "libraries/glad/glad.cpp"
    HEAP32[164639] = $153; //@line 2306 "libraries/glad/glad.cpp"
    $154 = (FUNCTION_TABLE_ii[$load & 8191](448374)|0); //@line 2307 "libraries/glad/glad.cpp"
    HEAP32[164650] = $154; //@line 2307 "libraries/glad/glad.cpp"
    $155 = (FUNCTION_TABLE_ii[$load & 8191](448384)|0); //@line 2308 "libraries/glad/glad.cpp"
    HEAP32[164629] = $155; //@line 2308 "libraries/glad/glad.cpp"
    $156 = (FUNCTION_TABLE_ii[$load & 8191](448393)|0); //@line 2309 "libraries/glad/glad.cpp"
    HEAP32[164887] = $156; //@line 2309 "libraries/glad/glad.cpp"
    $157 = (FUNCTION_TABLE_ii[$load & 8191](448403)|0); //@line 2310 "libraries/glad/glad.cpp"
    HEAP32[164645] = $157; //@line 2310 "libraries/glad/glad.cpp"
    $158 = (FUNCTION_TABLE_ii[$load & 8191](448412)|0); //@line 2311 "libraries/glad/glad.cpp"
    HEAP32[164800] = $158; //@line 2311 "libraries/glad/glad.cpp"
    $159 = (FUNCTION_TABLE_ii[$load & 8191](448422)|0); //@line 2312 "libraries/glad/glad.cpp"
    HEAP32[165631] = $159; //@line 2312 "libraries/glad/glad.cpp"
    $160 = (FUNCTION_TABLE_ii[$load & 8191](448433)|0); //@line 2313 "libraries/glad/glad.cpp"
    HEAP32[164948] = $160; //@line 2313 "libraries/glad/glad.cpp"
    $161 = (FUNCTION_TABLE_ii[$load & 8191](448445)|0); //@line 2314 "libraries/glad/glad.cpp"
    HEAP32[165630] = $161; //@line 2314 "libraries/glad/glad.cpp"
    $162 = (FUNCTION_TABLE_ii[$load & 8191](448456)|0); //@line 2315 "libraries/glad/glad.cpp"
    HEAP32[165285] = $162; //@line 2315 "libraries/glad/glad.cpp"
    $163 = (FUNCTION_TABLE_ii[$load & 8191](448468)|0); //@line 2316 "libraries/glad/glad.cpp"
    HEAP32[165629] = $163; //@line 2316 "libraries/glad/glad.cpp"
    $164 = (FUNCTION_TABLE_ii[$load & 8191](448479)|0); //@line 2317 "libraries/glad/glad.cpp"
    HEAP32[164861] = $164; //@line 2317 "libraries/glad/glad.cpp"
    $165 = (FUNCTION_TABLE_ii[$load & 8191](448491)|0); //@line 2318 "libraries/glad/glad.cpp"
    HEAP32[165628] = $165; //@line 2318 "libraries/glad/glad.cpp"
    $166 = (FUNCTION_TABLE_ii[$load & 8191](448502)|0); //@line 2319 "libraries/glad/glad.cpp"
    HEAP32[165431] = $166; //@line 2319 "libraries/glad/glad.cpp"
    $167 = (FUNCTION_TABLE_ii[$load & 8191](448514)|0); //@line 2320 "libraries/glad/glad.cpp"
    HEAP32[165397] = $167; //@line 2320 "libraries/glad/glad.cpp"
    $168 = (FUNCTION_TABLE_ii[$load & 8191](448525)|0); //@line 2321 "libraries/glad/glad.cpp"
    HEAP32[165128] = $168; //@line 2321 "libraries/glad/glad.cpp"
    $169 = (FUNCTION_TABLE_ii[$load & 8191](448537)|0); //@line 2322 "libraries/glad/glad.cpp"
    HEAP32[164700] = $169; //@line 2322 "libraries/glad/glad.cpp"
    $170 = (FUNCTION_TABLE_ii[$load & 8191](448551)|0); //@line 2323 "libraries/glad/glad.cpp"
    HEAP32[165236] = $170; //@line 2323 "libraries/glad/glad.cpp"
    $171 = (FUNCTION_TABLE_ii[$load & 8191](448566)|0); //@line 2324 "libraries/glad/glad.cpp"
    HEAP32[164699] = $171; //@line 2324 "libraries/glad/glad.cpp"
    $172 = (FUNCTION_TABLE_ii[$load & 8191](448580)|0); //@line 2325 "libraries/glad/glad.cpp"
    HEAP32[164655] = $172; //@line 2325 "libraries/glad/glad.cpp"
    $173 = (FUNCTION_TABLE_ii[$load & 8191](448595)|0); //@line 2326 "libraries/glad/glad.cpp"
    HEAP32[164702] = $173; //@line 2326 "libraries/glad/glad.cpp"
    $174 = (FUNCTION_TABLE_ii[$load & 8191](448609)|0); //@line 2327 "libraries/glad/glad.cpp"
    HEAP32[165349] = $174; //@line 2327 "libraries/glad/glad.cpp"
    $175 = (FUNCTION_TABLE_ii[$load & 8191](448624)|0); //@line 2328 "libraries/glad/glad.cpp"
    HEAP32[165088] = $175; //@line 2328 "libraries/glad/glad.cpp"
    $176 = (FUNCTION_TABLE_ii[$load & 8191](448638)|0); //@line 2329 "libraries/glad/glad.cpp"
    HEAP32[165054] = $176; //@line 2329 "libraries/glad/glad.cpp"
    $177 = (FUNCTION_TABLE_ii[$load & 8191](448653)|0); //@line 2330 "libraries/glad/glad.cpp"
    HEAP32[165029] = $177; //@line 2330 "libraries/glad/glad.cpp"
    $178 = (FUNCTION_TABLE_ii[$load & 8191](448667)|0); //@line 2331 "libraries/glad/glad.cpp"
    HEAP32[165330] = $178; //@line 2331 "libraries/glad/glad.cpp"
    $179 = (FUNCTION_TABLE_ii[$load & 8191](448682)|0); //@line 2332 "libraries/glad/glad.cpp"
    HEAP32[165030] = $179; //@line 2332 "libraries/glad/glad.cpp"
    $180 = (FUNCTION_TABLE_ii[$load & 8191](448696)|0); //@line 2333 "libraries/glad/glad.cpp"
    HEAP32[164871] = $180; //@line 2333 "libraries/glad/glad.cpp"
    $181 = (FUNCTION_TABLE_ii[$load & 8191](448711)|0); //@line 2334 "libraries/glad/glad.cpp"
    HEAP32[165027] = $181; //@line 2334 "libraries/glad/glad.cpp"
    $182 = (FUNCTION_TABLE_ii[$load & 8191](448725)|0); //@line 2335 "libraries/glad/glad.cpp"
    HEAP32[165075] = $182; //@line 2335 "libraries/glad/glad.cpp"
    $183 = (FUNCTION_TABLE_ii[$load & 8191](448740)|0); //@line 2336 "libraries/glad/glad.cpp"
    HEAP32[165044] = $183; //@line 2336 "libraries/glad/glad.cpp"
    $184 = (FUNCTION_TABLE_ii[$load & 8191](448754)|0); //@line 2337 "libraries/glad/glad.cpp"
    HEAP32[164889] = $184; //@line 2337 "libraries/glad/glad.cpp"
    $185 = (FUNCTION_TABLE_ii[$load & 8191](448769)|0); //@line 2338 "libraries/glad/glad.cpp"
    HEAP32[165267] = $185; //@line 2338 "libraries/glad/glad.cpp"
    $186 = (FUNCTION_TABLE_ii[$load & 8191](448783)|0); //@line 2339 "libraries/glad/glad.cpp"
    HEAP32[165592] = $186; //@line 2339 "libraries/glad/glad.cpp"
    $187 = (FUNCTION_TABLE_ii[$load & 8191](448798)|0); //@line 2340 "libraries/glad/glad.cpp"
    HEAP32[165268] = $187; //@line 2340 "libraries/glad/glad.cpp"
    $188 = (FUNCTION_TABLE_ii[$load & 8191](448812)|0); //@line 2341 "libraries/glad/glad.cpp"
    HEAP32[165119] = $188; //@line 2341 "libraries/glad/glad.cpp"
    $189 = (FUNCTION_TABLE_ii[$load & 8191](448827)|0); //@line 2342 "libraries/glad/glad.cpp"
    HEAP32[165393] = $189; //@line 2342 "libraries/glad/glad.cpp"
    $190 = (FUNCTION_TABLE_ii[$load & 8191](448841)|0); //@line 2343 "libraries/glad/glad.cpp"
    HEAP32[165137] = $190; //@line 2343 "libraries/glad/glad.cpp"
    $191 = (FUNCTION_TABLE_ii[$load & 8191](448856)|0); //@line 2344 "libraries/glad/glad.cpp"
    HEAP32[165474] = $191; //@line 2344 "libraries/glad/glad.cpp"
    $192 = (FUNCTION_TABLE_ii[$load & 8191](448870)|0); //@line 2345 "libraries/glad/glad.cpp"
    HEAP32[165055] = $192; //@line 2345 "libraries/glad/glad.cpp"
    $193 = (FUNCTION_TABLE_ii[$load & 8191](448885)|0); //@line 2346 "libraries/glad/glad.cpp"
    HEAP32[165127] = $193; //@line 2346 "libraries/glad/glad.cpp"
    $194 = (FUNCTION_TABLE_ii[$load & 8191](448893)|0); //@line 2347 "libraries/glad/glad.cpp"
    HEAP32[164632] = $194; //@line 2347 "libraries/glad/glad.cpp"
    $195 = (FUNCTION_TABLE_ii[$load & 8191](448902)|0); //@line 2348 "libraries/glad/glad.cpp"
    HEAP32[165126] = $195; //@line 2348 "libraries/glad/glad.cpp"
    $196 = (FUNCTION_TABLE_ii[$load & 8191](448910)|0); //@line 2349 "libraries/glad/glad.cpp"
    HEAP32[165204] = $196; //@line 2349 "libraries/glad/glad.cpp"
    $197 = (FUNCTION_TABLE_ii[$load & 8191](448919)|0); //@line 2350 "libraries/glad/glad.cpp"
    HEAP32[165123] = $197; //@line 2350 "libraries/glad/glad.cpp"
    $198 = (FUNCTION_TABLE_ii[$load & 8191](448927)|0); //@line 2351 "libraries/glad/glad.cpp"
    HEAP32[165327] = $198; //@line 2351 "libraries/glad/glad.cpp"
    $199 = (FUNCTION_TABLE_ii[$load & 8191](448936)|0); //@line 2352 "libraries/glad/glad.cpp"
    HEAP32[165135] = $199; //@line 2352 "libraries/glad/glad.cpp"
    $200 = (FUNCTION_TABLE_ii[$load & 8191](448944)|0); //@line 2353 "libraries/glad/glad.cpp"
    HEAP32[165620] = $200; //@line 2353 "libraries/glad/glad.cpp"
    $201 = (FUNCTION_TABLE_ii[$load & 8191](448953)|0); //@line 2354 "libraries/glad/glad.cpp"
    HEAP32[165499] = $201; //@line 2354 "libraries/glad/glad.cpp"
    $202 = (FUNCTION_TABLE_ii[$load & 8191](448966)|0); //@line 2355 "libraries/glad/glad.cpp"
    HEAP32[165569] = $202; //@line 2355 "libraries/glad/glad.cpp"
    $203 = (FUNCTION_TABLE_ii[$load & 8191](448980)|0); //@line 2356 "libraries/glad/glad.cpp"
    HEAP32[165500] = $203; //@line 2356 "libraries/glad/glad.cpp"
    $204 = (FUNCTION_TABLE_ii[$load & 8191](448993)|0); //@line 2357 "libraries/glad/glad.cpp"
    HEAP32[164835] = $204; //@line 2357 "libraries/glad/glad.cpp"
    $205 = (FUNCTION_TABLE_ii[$load & 8191](449007)|0); //@line 2358 "libraries/glad/glad.cpp"
    HEAP32[165497] = $205; //@line 2358 "libraries/glad/glad.cpp"
    $206 = (FUNCTION_TABLE_ii[$load & 8191](449020)|0); //@line 2359 "libraries/glad/glad.cpp"
    HEAP32[164724] = $206; //@line 2359 "libraries/glad/glad.cpp"
    $207 = (FUNCTION_TABLE_ii[$load & 8191](449034)|0); //@line 2360 "libraries/glad/glad.cpp"
    HEAP32[165511] = $207; //@line 2360 "libraries/glad/glad.cpp"
    $208 = (FUNCTION_TABLE_ii[$load & 8191](449047)|0); //@line 2361 "libraries/glad/glad.cpp"
    HEAP32[165564] = $208; //@line 2361 "libraries/glad/glad.cpp"
    $209 = (FUNCTION_TABLE_ii[$load & 8191](449061)|0); //@line 2362 "libraries/glad/glad.cpp"
    HEAP32[165259] = $209; //@line 2362 "libraries/glad/glad.cpp"
    $210 = (FUNCTION_TABLE_ii[$load & 8191](449074)|0); //@line 2363 "libraries/glad/glad.cpp"
    HEAP32[165136] = $210; //@line 2363 "libraries/glad/glad.cpp"
    $211 = (FUNCTION_TABLE_ii[$load & 8191](449088)|0); //@line 2364 "libraries/glad/glad.cpp"
    HEAP32[165261] = $211; //@line 2364 "libraries/glad/glad.cpp"
    $212 = (FUNCTION_TABLE_ii[$load & 8191](449101)|0); //@line 2365 "libraries/glad/glad.cpp"
    HEAP32[164687] = $212; //@line 2365 "libraries/glad/glad.cpp"
    $213 = (FUNCTION_TABLE_ii[$load & 8191](449115)|0); //@line 2366 "libraries/glad/glad.cpp"
    HEAP32[165266] = $213; //@line 2366 "libraries/glad/glad.cpp"
    $214 = (FUNCTION_TABLE_ii[$load & 8191](449128)|0); //@line 2367 "libraries/glad/glad.cpp"
    HEAP32[165594] = $214; //@line 2367 "libraries/glad/glad.cpp"
    $215 = (FUNCTION_TABLE_ii[$load & 8191](449142)|0); //@line 2368 "libraries/glad/glad.cpp"
    HEAP32[165273] = $215; //@line 2368 "libraries/glad/glad.cpp"
    $216 = (FUNCTION_TABLE_ii[$load & 8191](449155)|0); //@line 2369 "libraries/glad/glad.cpp"
    HEAP32[164840] = $216; //@line 2369 "libraries/glad/glad.cpp"
    $217 = (FUNCTION_TABLE_ii[$load & 8191](449169)|0); //@line 2370 "libraries/glad/glad.cpp"
    HEAP32[165034] = $217; //@line 2370 "libraries/glad/glad.cpp"
    $218 = (FUNCTION_TABLE_ii[$load & 8191](449182)|0); //@line 2371 "libraries/glad/glad.cpp"
    HEAP32[165425] = $218; //@line 2371 "libraries/glad/glad.cpp"
    $219 = (FUNCTION_TABLE_ii[$load & 8191](449196)|0); //@line 2372 "libraries/glad/glad.cpp"
    HEAP32[165032] = $219; //@line 2372 "libraries/glad/glad.cpp"
    $220 = (FUNCTION_TABLE_ii[$load & 8191](449209)|0); //@line 2373 "libraries/glad/glad.cpp"
    HEAP32[164845] = $220; //@line 2373 "libraries/glad/glad.cpp"
    $221 = (FUNCTION_TABLE_ii[$load & 8191](449223)|0); //@line 2374 "libraries/glad/glad.cpp"
    HEAP32[165025] = $221; //@line 2374 "libraries/glad/glad.cpp"
    $222 = (FUNCTION_TABLE_ii[$load & 8191](449236)|0); //@line 2375 "libraries/glad/glad.cpp"
    HEAP32[165099] = $222; //@line 2375 "libraries/glad/glad.cpp"
    $223 = (FUNCTION_TABLE_ii[$load & 8191](449250)|0); //@line 2376 "libraries/glad/glad.cpp"
    HEAP32[165042] = $223; //@line 2376 "libraries/glad/glad.cpp"
    $224 = (FUNCTION_TABLE_ii[$load & 8191](449263)|0); //@line 2377 "libraries/glad/glad.cpp"
    HEAP32[164782] = $224; //@line 2377 "libraries/glad/glad.cpp"
    $225 = (FUNCTION_TABLE_ii[$load & 8191](449277)|0); //@line 2378 "libraries/glad/glad.cpp"
    HEAP32[164816] = $225; //@line 2378 "libraries/glad/glad.cpp"
    $226 = (FUNCTION_TABLE_ii[$load & 8191](449290)|0); //@line 2379 "libraries/glad/glad.cpp"
    HEAP32[165276] = $226; //@line 2379 "libraries/glad/glad.cpp"
    $227 = (FUNCTION_TABLE_ii[$load & 8191](449304)|0); //@line 2380 "libraries/glad/glad.cpp"
    HEAP32[164814] = $227; //@line 2380 "libraries/glad/glad.cpp"
    $228 = (FUNCTION_TABLE_ii[$load & 8191](449317)|0); //@line 2381 "libraries/glad/glad.cpp"
    HEAP32[164691] = $228; //@line 2381 "libraries/glad/glad.cpp"
    $229 = (FUNCTION_TABLE_ii[$load & 8191](449331)|0); //@line 2382 "libraries/glad/glad.cpp"
    HEAP32[164819] = $229; //@line 2382 "libraries/glad/glad.cpp"
    $230 = (FUNCTION_TABLE_ii[$load & 8191](449344)|0); //@line 2383 "libraries/glad/glad.cpp"
    HEAP32[164923] = $230; //@line 2383 "libraries/glad/glad.cpp"
    $231 = (FUNCTION_TABLE_ii[$load & 8191](449358)|0); //@line 2384 "libraries/glad/glad.cpp"
    HEAP32[164822] = $231; //@line 2384 "libraries/glad/glad.cpp"
    $232 = (FUNCTION_TABLE_ii[$load & 8191](449371)|0); //@line 2385 "libraries/glad/glad.cpp"
    HEAP32[165079] = $232; //@line 2385 "libraries/glad/glad.cpp"
    $233 = (FUNCTION_TABLE_ii[$load & 8191](449385)|0); //@line 2386 "libraries/glad/glad.cpp"
    HEAP32[164935] = $233; //@line 2386 "libraries/glad/glad.cpp"
    $234 = (FUNCTION_TABLE_ii[$load & 8191](449396)|0); //@line 2387 "libraries/glad/glad.cpp"
    HEAP32[165078] = $234; //@line 2387 "libraries/glad/glad.cpp"
    $235 = (FUNCTION_TABLE_ii[$load & 8191](449408)|0); //@line 2388 "libraries/glad/glad.cpp"
    HEAP32[164934] = $235; //@line 2388 "libraries/glad/glad.cpp"
    $236 = (FUNCTION_TABLE_ii[$load & 8191](449419)|0); //@line 2389 "libraries/glad/glad.cpp"
    HEAP32[164628] = $236; //@line 2389 "libraries/glad/glad.cpp"
    $237 = (FUNCTION_TABLE_ii[$load & 8191](449431)|0); //@line 2390 "libraries/glad/glad.cpp"
    HEAP32[164941] = $237; //@line 2390 "libraries/glad/glad.cpp"
    $238 = (FUNCTION_TABLE_ii[$load & 8191](449442)|0); //@line 2391 "libraries/glad/glad.cpp"
    HEAP32[164738] = $238; //@line 2391 "libraries/glad/glad.cpp"
    $239 = (FUNCTION_TABLE_ii[$load & 8191](449454)|0); //@line 2392 "libraries/glad/glad.cpp"
    HEAP32[164945] = $239; //@line 2392 "libraries/glad/glad.cpp"
    $240 = (FUNCTION_TABLE_ii[$load & 8191](449465)|0); //@line 2393 "libraries/glad/glad.cpp"
    HEAP32[165011] = $240; //@line 2393 "libraries/glad/glad.cpp"
    $241 = (FUNCTION_TABLE_ii[$load & 8191](449477)|0); //@line 2394 "libraries/glad/glad.cpp"
    HEAP32[164710] = $241; //@line 2394 "libraries/glad/glad.cpp"
    $242 = (FUNCTION_TABLE_ii[$load & 8191](449488)|0); //@line 2395 "libraries/glad/glad.cpp"
    HEAP32[164828] = $242; //@line 2395 "libraries/glad/glad.cpp"
    $243 = (FUNCTION_TABLE_ii[$load & 8191](449500)|0); //@line 2396 "libraries/glad/glad.cpp"
    HEAP32[164711] = $243; //@line 2396 "libraries/glad/glad.cpp"
    $244 = (FUNCTION_TABLE_ii[$load & 8191](449511)|0); //@line 2397 "libraries/glad/glad.cpp"
    HEAP32[165275] = $244; //@line 2397 "libraries/glad/glad.cpp"
    $245 = (FUNCTION_TABLE_ii[$load & 8191](449523)|0); //@line 2398 "libraries/glad/glad.cpp"
    HEAP32[164713] = $245; //@line 2398 "libraries/glad/glad.cpp"
    $246 = (FUNCTION_TABLE_ii[$load & 8191](449534)|0); //@line 2399 "libraries/glad/glad.cpp"
    HEAP32[164805] = $246; //@line 2399 "libraries/glad/glad.cpp"
    $247 = (FUNCTION_TABLE_ii[$load & 8191](449546)|0); //@line 2400 "libraries/glad/glad.cpp"
    HEAP32[164719] = $247; //@line 2400 "libraries/glad/glad.cpp"
    $248 = (FUNCTION_TABLE_ii[$load & 8191](449557)|0); //@line 2401 "libraries/glad/glad.cpp"
    HEAP32[164958] = $248; //@line 2401 "libraries/glad/glad.cpp"
    $249 = (FUNCTION_TABLE_ii[$load & 8191](449569)|0); //@line 2402 "libraries/glad/glad.cpp"
    HEAP32[165562] = $249; //@line 2402 "libraries/glad/glad.cpp"
    $250 = (FUNCTION_TABLE_ii[$load & 8191](449580)|0); //@line 2403 "libraries/glad/glad.cpp"
    HEAP32[165258] = $250; //@line 2403 "libraries/glad/glad.cpp"
    $251 = (FUNCTION_TABLE_ii[$load & 8191](449592)|0); //@line 2404 "libraries/glad/glad.cpp"
    HEAP32[165619] = $251; //@line 2404 "libraries/glad/glad.cpp"
    $252 = (FUNCTION_TABLE_ii[$load & 8191](449603)|0); //@line 2405 "libraries/glad/glad.cpp"
    HEAP32[164681] = $252; //@line 2405 "libraries/glad/glad.cpp"
    $253 = (FUNCTION_TABLE_ii[$load & 8191](449615)|0); //@line 2406 "libraries/glad/glad.cpp"
    HEAP32[165293] = $253; //@line 2406 "libraries/glad/glad.cpp"
    $254 = (FUNCTION_TABLE_ii[$load & 8191](449626)|0); //@line 2407 "libraries/glad/glad.cpp"
    HEAP32[164668] = $254; //@line 2407 "libraries/glad/glad.cpp"
    $255 = (FUNCTION_TABLE_ii[$load & 8191](449638)|0); //@line 2408 "libraries/glad/glad.cpp"
    HEAP32[165282] = $255; //@line 2408 "libraries/glad/glad.cpp"
    $256 = (FUNCTION_TABLE_ii[$load & 8191](449649)|0); //@line 2409 "libraries/glad/glad.cpp"
    HEAP32[164747] = $256; //@line 2409 "libraries/glad/glad.cpp"
    $257 = (FUNCTION_TABLE_ii[$load & 8191](449661)|0); //@line 2410 "libraries/glad/glad.cpp"
    HEAP32[165385] = $257; //@line 2410 "libraries/glad/glad.cpp"
    $258 = (FUNCTION_TABLE_ii[$load & 8191](449673)|0); //@line 2411 "libraries/glad/glad.cpp"
    HEAP32[165175] = $258; //@line 2411 "libraries/glad/glad.cpp"
    $259 = (FUNCTION_TABLE_ii[$load & 8191](449689)|0); //@line 2412 "libraries/glad/glad.cpp"
    HEAP32[165486] = $259; //@line 2412 "libraries/glad/glad.cpp"
    $260 = (FUNCTION_TABLE_ii[$load & 8191](449696)|0); //@line 2413 "libraries/glad/glad.cpp"
    HEAP32[164663] = $260; //@line 2413 "libraries/glad/glad.cpp"
    $261 = (FUNCTION_TABLE_ii[$load & 8191](449704)|0); //@line 2414 "libraries/glad/glad.cpp"
    HEAP32[164752] = $261; //@line 2414 "libraries/glad/glad.cpp"
    $262 = (FUNCTION_TABLE_ii[$load & 8191](449711)|0); //@line 2415 "libraries/glad/glad.cpp"
    HEAP32[164653] = $262; //@line 2415 "libraries/glad/glad.cpp"
    $263 = (FUNCTION_TABLE_ii[$load & 8191](449719)|0); //@line 2416 "libraries/glad/glad.cpp"
    HEAP32[164907] = $263; //@line 2416 "libraries/glad/glad.cpp"
    $264 = (FUNCTION_TABLE_ii[$load & 8191](449728)|0); //@line 2417 "libraries/glad/glad.cpp"
    HEAP32[164765] = $264; //@line 2417 "libraries/glad/glad.cpp"
    $265 = (FUNCTION_TABLE_ii[$load & 8191](449738)|0); //@line 2418 "libraries/glad/glad.cpp"
    HEAP32[164904] = $265; //@line 2418 "libraries/glad/glad.cpp"
    $266 = (FUNCTION_TABLE_ii[$load & 8191](449747)|0); //@line 2419 "libraries/glad/glad.cpp"
    HEAP32[165370] = $266; //@line 2419 "libraries/glad/glad.cpp"
    $267 = (FUNCTION_TABLE_ii[$load & 8191](449757)|0); //@line 2420 "libraries/glad/glad.cpp"
    HEAP32[165008] = $267; //@line 2420 "libraries/glad/glad.cpp"
    $268 = (FUNCTION_TABLE_ii[$load & 8191](449771)|0); //@line 2421 "libraries/glad/glad.cpp"
    HEAP32[165254] = $268; //@line 2421 "libraries/glad/glad.cpp"
    $269 = (FUNCTION_TABLE_ii[$load & 8191](449786)|0); //@line 2422 "libraries/glad/glad.cpp"
    HEAP32[165014] = $269; //@line 2422 "libraries/glad/glad.cpp"
    $270 = (FUNCTION_TABLE_ii[$load & 8191](449800)|0); //@line 2423 "libraries/glad/glad.cpp"
    HEAP32[164656] = $270; //@line 2423 "libraries/glad/glad.cpp"
    $271 = (FUNCTION_TABLE_ii[$load & 8191](449815)|0); //@line 2424 "libraries/glad/glad.cpp"
    HEAP32[165059] = $271; //@line 2424 "libraries/glad/glad.cpp"
    $272 = (FUNCTION_TABLE_ii[$load & 8191](449829)|0); //@line 2425 "libraries/glad/glad.cpp"
    HEAP32[164821] = $272; //@line 2425 "libraries/glad/glad.cpp"
    $273 = (FUNCTION_TABLE_ii[$load & 8191](449841)|0); //@line 2426 "libraries/glad/glad.cpp"
    HEAP32[165092] = $273; //@line 2426 "libraries/glad/glad.cpp"
    $274 = (FUNCTION_TABLE_ii[$load & 8191](449854)|0); //@line 2427 "libraries/glad/glad.cpp"
    HEAP32[164809] = $274; //@line 2427 "libraries/glad/glad.cpp"
    $275 = (FUNCTION_TABLE_ii[$load & 8191](449866)|0); //@line 2428 "libraries/glad/glad.cpp"
    HEAP32[165277] = $275; //@line 2428 "libraries/glad/glad.cpp"
    $276 = (FUNCTION_TABLE_ii[$load & 8191](449879)|0); //@line 2429 "libraries/glad/glad.cpp"
    HEAP32[165624] = $276; //@line 2429 "libraries/glad/glad.cpp"
    $277 = (FUNCTION_TABLE_ii[$load & 8191](449896)|0); //@line 2430 "libraries/glad/glad.cpp"
    HEAP32[165301] = $277; //@line 2430 "libraries/glad/glad.cpp"
    $278 = (FUNCTION_TABLE_ii[$load & 8191](449909)|0); //@line 2431 "libraries/glad/glad.cpp"
    HEAP32[164961] = $278; //@line 2431 "libraries/glad/glad.cpp"
    $279 = (FUNCTION_TABLE_ii[$load & 8191](449919)|0); //@line 2432 "libraries/glad/glad.cpp"
    HEAP32[165244] = $279; //@line 2432 "libraries/glad/glad.cpp"
    $280 = (FUNCTION_TABLE_ii[$load & 8191](449930)|0); //@line 2433 "libraries/glad/glad.cpp"
    HEAP32[164968] = $280; //@line 2433 "libraries/glad/glad.cpp"
    $281 = (FUNCTION_TABLE_ii[$load & 8191](449940)|0); //@line 2434 "libraries/glad/glad.cpp"
    HEAP32[165354] = $281; //@line 2434 "libraries/glad/glad.cpp"
    $282 = (FUNCTION_TABLE_ii[$load & 8191](449951)|0); //@line 2435 "libraries/glad/glad.cpp"
    HEAP32[165210] = $282; //@line 2435 "libraries/glad/glad.cpp"
    $283 = (FUNCTION_TABLE_ii[$load & 8191](449961)|0); //@line 2436 "libraries/glad/glad.cpp"
    HEAP32[165387] = $283; //@line 2436 "libraries/glad/glad.cpp"
    $284 = (FUNCTION_TABLE_ii[$load & 8191](449972)|0); //@line 2437 "libraries/glad/glad.cpp"
    HEAP32[164847] = $284; //@line 2437 "libraries/glad/glad.cpp"
    $285 = (FUNCTION_TABLE_ii[$load & 8191](449982)|0); //@line 2438 "libraries/glad/glad.cpp"
    HEAP32[165376] = $285; //@line 2438 "libraries/glad/glad.cpp"
    $286 = (FUNCTION_TABLE_ii[$load & 8191](449993)|0); //@line 2439 "libraries/glad/glad.cpp"
    HEAP32[165212] = $286; //@line 2439 "libraries/glad/glad.cpp"
    $287 = (FUNCTION_TABLE_ii[$load & 8191](450003)|0); //@line 2440 "libraries/glad/glad.cpp"
    HEAP32[165235] = $287; //@line 2440 "libraries/glad/glad.cpp"
    $288 = (FUNCTION_TABLE_ii[$load & 8191](450014)|0); //@line 2441 "libraries/glad/glad.cpp"
    HEAP32[164983] = $288; //@line 2441 "libraries/glad/glad.cpp"
    $289 = (FUNCTION_TABLE_ii[$load & 8191](450031)|0); //@line 2442 "libraries/glad/glad.cpp"
    HEAP32[165358] = $289; //@line 2442 "libraries/glad/glad.cpp"
    $290 = (FUNCTION_TABLE_ii[$load & 8191](450046)|0); //@line 2443 "libraries/glad/glad.cpp"
    HEAP32[165184] = $290; //@line 2443 "libraries/glad/glad.cpp"
    $291 = (FUNCTION_TABLE_ii[$load & 8191](450059)|0); //@line 2444 "libraries/glad/glad.cpp"
    HEAP32[165166] = $291; //@line 2444 "libraries/glad/glad.cpp"
    $292 = (FUNCTION_TABLE_ii[$load & 8191](450071)|0); //@line 2445 "libraries/glad/glad.cpp"
    HEAP32[165296] = $292; //@line 2445 "libraries/glad/glad.cpp"
    $293 = (FUNCTION_TABLE_ii[$load & 8191](450082)|0); //@line 2446 "libraries/glad/glad.cpp"
    HEAP32[164750] = $293; //@line 2446 "libraries/glad/glad.cpp"
    $294 = (FUNCTION_TABLE_ii[$load & 8191](450096)|0); //@line 2447 "libraries/glad/glad.cpp"
    HEAP32[165114] = $294; //@line 2447 "libraries/glad/glad.cpp"
    $295 = (FUNCTION_TABLE_ii[$load & 8191](450106)|0); //@line 2448 "libraries/glad/glad.cpp"
    HEAP32[165386] = $295; //@line 2448 "libraries/glad/glad.cpp"
    $296 = (FUNCTION_TABLE_ii[$load & 8191](450117)|0); //@line 2449 "libraries/glad/glad.cpp"
    HEAP32[165546] = $296; //@line 2449 "libraries/glad/glad.cpp"
    $297 = (FUNCTION_TABLE_ii[$load & 8191](450130)|0); //@line 2450 "libraries/glad/glad.cpp"
    HEAP32[164880] = $297; //@line 2450 "libraries/glad/glad.cpp"
    $298 = (FUNCTION_TABLE_ii[$load & 8191](450143)|0); //@line 2451 "libraries/glad/glad.cpp"
    HEAP32[165537] = $298; //@line 2451 "libraries/glad/glad.cpp"
    $299 = (FUNCTION_TABLE_ii[$load & 8191](450155)|0); //@line 2452 "libraries/glad/glad.cpp"
    HEAP32[165588] = $299; //@line 2452 "libraries/glad/glad.cpp"
    $300 = (FUNCTION_TABLE_ii[$load & 8191](450163)|0); //@line 2453 "libraries/glad/glad.cpp"
    HEAP32[165156] = $300; //@line 2453 "libraries/glad/glad.cpp"
    $301 = (FUNCTION_TABLE_ii[$load & 8191](450175)|0); //@line 2454 "libraries/glad/glad.cpp"
    HEAP32[165362] = $301; //@line 2454 "libraries/glad/glad.cpp"
    $302 = (FUNCTION_TABLE_ii[$load & 8191](450188)|0); //@line 2455 "libraries/glad/glad.cpp"
    HEAP32[164881] = $302; //@line 2455 "libraries/glad/glad.cpp"
    $303 = (FUNCTION_TABLE_ii[$load & 8191](450196)|0); //@line 2456 "libraries/glad/glad.cpp"
    HEAP32[164882] = $303; //@line 2456 "libraries/glad/glad.cpp"
    $304 = (FUNCTION_TABLE_ii[$load & 8191](450204)|0); //@line 2457 "libraries/glad/glad.cpp"
    HEAP32[165558] = $304; //@line 2457 "libraries/glad/glad.cpp"
    $305 = (FUNCTION_TABLE_ii[$load & 8191](450212)|0); //@line 2458 "libraries/glad/glad.cpp"
    HEAP32[165559] = $305; //@line 2458 "libraries/glad/glad.cpp"
    $306 = (FUNCTION_TABLE_ii[$load & 8191](450220)|0); //@line 2459 "libraries/glad/glad.cpp"
    HEAP32[165302] = $306; //@line 2459 "libraries/glad/glad.cpp"
    $307 = (FUNCTION_TABLE_ii[$load & 8191](450232)|0); //@line 2460 "libraries/glad/glad.cpp"
    HEAP32[165304] = $307; //@line 2460 "libraries/glad/glad.cpp"
    $308 = (FUNCTION_TABLE_ii[$load & 8191](450244)|0); //@line 2461 "libraries/glad/glad.cpp"
    HEAP32[164938] = $308; //@line 2461 "libraries/glad/glad.cpp"
    $309 = (FUNCTION_TABLE_ii[$load & 8191](450256)|0); //@line 2462 "libraries/glad/glad.cpp"
    HEAP32[164939] = $309; //@line 2462 "libraries/glad/glad.cpp"
    $310 = (FUNCTION_TABLE_ii[$load & 8191](450268)|0); //@line 2463 "libraries/glad/glad.cpp"
    HEAP32[165580] = $310; //@line 2463 "libraries/glad/glad.cpp"
    $311 = (FUNCTION_TABLE_ii[$load & 8191](450282)|0); //@line 2464 "libraries/glad/glad.cpp"
    HEAP32[165120] = $311; //@line 2464 "libraries/glad/glad.cpp"
    $312 = (FUNCTION_TABLE_ii[$load & 8191](450297)|0); //@line 2465 "libraries/glad/glad.cpp"
    HEAP32[165582] = $312; //@line 2465 "libraries/glad/glad.cpp"
    $313 = (FUNCTION_TABLE_ii[$load & 8191](450311)|0); //@line 2466 "libraries/glad/glad.cpp"
    HEAP32[164669] = $313; //@line 2466 "libraries/glad/glad.cpp"
    $314 = (FUNCTION_TABLE_ii[$load & 8191](450326)|0); //@line 2467 "libraries/glad/glad.cpp"
    HEAP32[164634] = $314; //@line 2467 "libraries/glad/glad.cpp"
    $315 = (FUNCTION_TABLE_ii[$load & 8191](450340)|0); //@line 2468 "libraries/glad/glad.cpp"
    HEAP32[164754] = $315; //@line 2468 "libraries/glad/glad.cpp"
    $316 = (FUNCTION_TABLE_ii[$load & 8191](450355)|0); //@line 2469 "libraries/glad/glad.cpp"
    HEAP32[164635] = $316; //@line 2469 "libraries/glad/glad.cpp"
    $317 = (FUNCTION_TABLE_ii[$load & 8191](450369)|0); //@line 2470 "libraries/glad/glad.cpp"
    HEAP32[165340] = $317; //@line 2470 "libraries/glad/glad.cpp"
    $318 = (FUNCTION_TABLE_ii[$load & 8191](450384)|0); //@line 2471 "libraries/glad/glad.cpp"
    HEAP32[165332] = $318; //@line 2471 "libraries/glad/glad.cpp"
    $319 = (FUNCTION_TABLE_ii[$load & 8191](450396)|0); //@line 2472 "libraries/glad/glad.cpp"
    HEAP32[165566] = $319; //@line 2472 "libraries/glad/glad.cpp"
    $320 = (FUNCTION_TABLE_ii[$load & 8191](450409)|0); //@line 2473 "libraries/glad/glad.cpp"
    HEAP32[165331] = $320; //@line 2473 "libraries/glad/glad.cpp"
    $321 = (FUNCTION_TABLE_ii[$load & 8191](450421)|0); //@line 2474 "libraries/glad/glad.cpp"
    HEAP32[165195] = $321; //@line 2474 "libraries/glad/glad.cpp"
    $322 = (FUNCTION_TABLE_ii[$load & 8191](450434)|0); //@line 2475 "libraries/glad/glad.cpp"
    HEAP32[165315] = $322; //@line 2475 "libraries/glad/glad.cpp"
    $323 = (FUNCTION_TABLE_ii[$load & 8191](450446)|0); //@line 2476 "libraries/glad/glad.cpp"
    HEAP32[164891] = $323; //@line 2476 "libraries/glad/glad.cpp"
    $324 = (FUNCTION_TABLE_ii[$load & 8191](450458)|0); //@line 2477 "libraries/glad/glad.cpp"
    HEAP32[165196] = $324; //@line 2477 "libraries/glad/glad.cpp"
    $325 = (FUNCTION_TABLE_ii[$load & 8191](450475)|0); //@line 2478 "libraries/glad/glad.cpp"
    HEAP32[165191] = $325; //@line 2478 "libraries/glad/glad.cpp"
    $326 = (FUNCTION_TABLE_ii[$load & 8191](450492)|0); //@line 2479 "libraries/glad/glad.cpp"
    HEAP32[165583] = $326; //@line 2479 "libraries/glad/glad.cpp"
    $327 = (FUNCTION_TABLE_ii[$load & 8191](450505)|0); //@line 2480 "libraries/glad/glad.cpp"
    HEAP32[164794] = $327; //@line 2480 "libraries/glad/glad.cpp"
    $328 = (FUNCTION_TABLE_ii[$load & 8191](450519)|0); //@line 2481 "libraries/glad/glad.cpp"
    HEAP32[165633] = $328; //@line 2481 "libraries/glad/glad.cpp"
    $329 = (FUNCTION_TABLE_ii[$load & 8191](450533)|0); //@line 2482 "libraries/glad/glad.cpp"
    HEAP32[165153] = $329; //@line 2482 "libraries/glad/glad.cpp"
    $330 = (FUNCTION_TABLE_ii[$load & 8191](450546)|0); //@line 2483 "libraries/glad/glad.cpp"
    HEAP32[165604] = $330; //@line 2483 "libraries/glad/glad.cpp"
    $331 = (FUNCTION_TABLE_ii[$load & 8191](450559)|0); //@line 2484 "libraries/glad/glad.cpp"
    HEAP32[165257] = $331; //@line 2484 "libraries/glad/glad.cpp"
    $332 = (FUNCTION_TABLE_ii[$load & 8191](450574)|0); //@line 2485 "libraries/glad/glad.cpp"
    HEAP32[165206] = $332; //@line 2485 "libraries/glad/glad.cpp"
    $333 = (FUNCTION_TABLE_ii[$load & 8191](450587)|0); //@line 2486 "libraries/glad/glad.cpp"
    HEAP32[165447] = $333; //@line 2486 "libraries/glad/glad.cpp"
    $334 = (FUNCTION_TABLE_ii[$load & 8191](450600)|0); //@line 2487 "libraries/glad/glad.cpp"
    HEAP32[165090] = $334; //@line 2487 "libraries/glad/glad.cpp"
    $335 = (FUNCTION_TABLE_ii[$load & 8191](450611)|0); //@line 2488 "libraries/glad/glad.cpp"
    HEAP32[164644] = $335; //@line 2488 "libraries/glad/glad.cpp"
    $336 = (FUNCTION_TABLE_ii[$load & 8191](450622)|0); //@line 2489 "libraries/glad/glad.cpp"
    HEAP32[164986] = $336; //@line 2489 "libraries/glad/glad.cpp"
    $337 = (FUNCTION_TABLE_ii[$load & 8191](450633)|0); //@line 2490 "libraries/glad/glad.cpp"
    HEAP32[164791] = $337; //@line 2490 "libraries/glad/glad.cpp"
    $338 = (FUNCTION_TABLE_ii[$load & 8191](450649)|0); //@line 2491 "libraries/glad/glad.cpp"
    HEAP32[165608] = $338; //@line 2491 "libraries/glad/glad.cpp"
    $339 = (FUNCTION_TABLE_ii[$load & 8191](450665)|0); //@line 2492 "libraries/glad/glad.cpp"
    HEAP32[164960] = $339; //@line 2492 "libraries/glad/glad.cpp"
    $340 = (FUNCTION_TABLE_ii[$load & 8191](450681)|0); //@line 2493 "libraries/glad/glad.cpp"
    HEAP32[164921] = $340; //@line 2493 "libraries/glad/glad.cpp"
    $341 = (FUNCTION_TABLE_ii[$load & 8191](450698)|0); //@line 2494 "libraries/glad/glad.cpp"
    HEAP32[165585] = $341; //@line 2494 "libraries/glad/glad.cpp"
    $342 = (FUNCTION_TABLE_ii[$load & 8191](450715)|0); //@line 2495 "libraries/glad/glad.cpp"
    HEAP32[165461] = $342; //@line 2495 "libraries/glad/glad.cpp"
    $343 = (FUNCTION_TABLE_ii[$load & 8191](450735)|0); //@line 2496 "libraries/glad/glad.cpp"
    HEAP32[165203] = $343; //@line 2496 "libraries/glad/glad.cpp"
    $344 = (FUNCTION_TABLE_ii[$load & 8191](450749)|0); //@line 2497 "libraries/glad/glad.cpp"
    HEAP32[165572] = $344; //@line 2497 "libraries/glad/glad.cpp"
    $345 = (FUNCTION_TABLE_ii[$load & 8191](450763)|0); //@line 2498 "libraries/glad/glad.cpp"
    HEAP32[165522] = $345; //@line 2498 "libraries/glad/glad.cpp"
    $346 = (FUNCTION_TABLE_ii[$load & 8191](450777)|0); //@line 2499 "libraries/glad/glad.cpp"
    HEAP32[165226] = $346; //@line 2499 "libraries/glad/glad.cpp"
    $347 = (FUNCTION_TABLE_ii[$load & 8191](450791)|0); //@line 2500 "libraries/glad/glad.cpp"
    HEAP32[165634] = $347; //@line 2500 "libraries/glad/glad.cpp"
    $348 = (FUNCTION_TABLE_ii[$load & 8191](450805)|0); //@line 2501 "libraries/glad/glad.cpp"
    HEAP32[165052] = $348; //@line 2501 "libraries/glad/glad.cpp"
    $349 = (FUNCTION_TABLE_ii[$load & 8191](615702)|0); //@line 2502 "libraries/glad/glad.cpp"
    HEAP32[164865] = $349; //@line 2502 "libraries/glad/glad.cpp"
    $350 = (FUNCTION_TABLE_ii[$load & 8191](615673)|0); //@line 2503 "libraries/glad/glad.cpp"
    HEAP32[164701] = $350; //@line 2503 "libraries/glad/glad.cpp"
    $351 = (FUNCTION_TABLE_ii[$load & 8191](615688)|0); //@line 2504 "libraries/glad/glad.cpp"
    HEAP32[165611] = $351; //@line 2504 "libraries/glad/glad.cpp"
    $352 = (FUNCTION_TABLE_ii[$load & 8191](450814)|0); //@line 2505 "libraries/glad/glad.cpp"
    HEAP32[165613] = $352; //@line 2505 "libraries/glad/glad.cpp"
    $353 = (FUNCTION_TABLE_ii[$load & 8191](615660)|0); //@line 2506 "libraries/glad/glad.cpp"
    HEAP32[164981] = $353; //@line 2506 "libraries/glad/glad.cpp"
    $354 = (FUNCTION_TABLE_ii[$load & 8191](450828)|0); //@line 2507 "libraries/glad/glad.cpp"
    HEAP32[165606] = $354; //@line 2507 "libraries/glad/glad.cpp"
    $355 = (FUNCTION_TABLE_ii[$load & 8191](450842)|0); //@line 2508 "libraries/glad/glad.cpp"
    HEAP32[165605] = $355; //@line 2508 "libraries/glad/glad.cpp"
    $356 = (FUNCTION_TABLE_ii[$load & 8191](450856)|0); //@line 2509 "libraries/glad/glad.cpp"
    HEAP32[164872] = $356; //@line 2509 "libraries/glad/glad.cpp"
    $357 = (FUNCTION_TABLE_ii[$load & 8191](450864)|0); //@line 2510 "libraries/glad/glad.cpp"
    HEAP32[165260] = $357; //@line 2510 "libraries/glad/glad.cpp"
    $358 = (FUNCTION_TABLE_ii[$load & 8191](450876)|0); //@line 2511 "libraries/glad/glad.cpp"
    HEAP32[164869] = $358; //@line 2511 "libraries/glad/glad.cpp"
    $359 = (FUNCTION_TABLE_ii[$load & 8191](450889)|0); //@line 2512 "libraries/glad/glad.cpp"
    HEAP32[165290] = $359; //@line 2512 "libraries/glad/glad.cpp"
    $360 = (FUNCTION_TABLE_ii[$load & 8191](615712)|0); //@line 2513 "libraries/glad/glad.cpp"
    HEAP32[165292] = $360; //@line 2513 "libraries/glad/glad.cpp"
    $361 = (FUNCTION_TABLE_ii[$load & 8191](450899)|0); //@line 2514 "libraries/glad/glad.cpp"
    HEAP32[165147] = $361; //@line 2514 "libraries/glad/glad.cpp"
    $362 = (FUNCTION_TABLE_ii[$load & 8191](450908)|0); //@line 2515 "libraries/glad/glad.cpp"
    HEAP32[165145] = $362; //@line 2515 "libraries/glad/glad.cpp"
    $363 = (FUNCTION_TABLE_ii[$load & 8191](450917)|0); //@line 2516 "libraries/glad/glad.cpp"
    HEAP32[164781] = $363; //@line 2516 "libraries/glad/glad.cpp"
    $364 = (FUNCTION_TABLE_ii[$load & 8191](450930)|0); //@line 2517 "libraries/glad/glad.cpp"
    HEAP32[164779] = $364; //@line 2517 "libraries/glad/glad.cpp"
    $$pre = HEAP8[1251995]|0; //@line 2522 "libraries/glad/glad.cpp"
    $366 = $$pre;
   } else {
    $366 = $36;
   }
   $365 = ($366<<24>>24)==(0); //@line 2522 "libraries/glad/glad.cpp"
   if (!($365)) {
    $367 = (FUNCTION_TABLE_ii[$load & 8191](615797)|0); //@line 2523 "libraries/glad/glad.cpp"
    HEAP32[164759] = $367; //@line 2523 "libraries/glad/glad.cpp"
    $368 = (FUNCTION_TABLE_ii[$load & 8191](615810)|0); //@line 2524 "libraries/glad/glad.cpp"
    HEAP32[165086] = $368; //@line 2524 "libraries/glad/glad.cpp"
    $369 = (FUNCTION_TABLE_ii[$load & 8191](615462)|0); //@line 2525 "libraries/glad/glad.cpp"
    HEAP32[164857] = $369; //@line 2525 "libraries/glad/glad.cpp"
    $370 = (FUNCTION_TABLE_ii[$load & 8191](616280)|0); //@line 2526 "libraries/glad/glad.cpp"
    HEAP32[164859] = $370; //@line 2526 "libraries/glad/glad.cpp"
    $371 = (FUNCTION_TABLE_ii[$load & 8191](450943)|0); //@line 2527 "libraries/glad/glad.cpp"
    HEAP32[164618] = $371; //@line 2527 "libraries/glad/glad.cpp"
    $372 = (FUNCTION_TABLE_ii[$load & 8191](616582)|0); //@line 2528 "libraries/glad/glad.cpp"
    HEAP32[164967] = $372; //@line 2528 "libraries/glad/glad.cpp"
    $373 = (FUNCTION_TABLE_ii[$load & 8191](450960)|0); //@line 2529 "libraries/glad/glad.cpp"
    HEAP32[165496] = $373; //@line 2529 "libraries/glad/glad.cpp"
    $374 = (FUNCTION_TABLE_ii[$load & 8191](616599)|0); //@line 2530 "libraries/glad/glad.cpp"
    HEAP32[164677] = $374; //@line 2530 "libraries/glad/glad.cpp"
    $375 = (FUNCTION_TABLE_ii[$load & 8191](450980)|0); //@line 2531 "libraries/glad/glad.cpp"
    HEAP32[165576] = $375; //@line 2531 "libraries/glad/glad.cpp"
    $376 = (FUNCTION_TABLE_ii[$load & 8191](614122)|0); //@line 2532 "libraries/glad/glad.cpp"
    HEAP32[165192] = $376; //@line 2532 "libraries/glad/glad.cpp"
    $377 = (FUNCTION_TABLE_ii[$load & 8191](614151)|0); //@line 2533 "libraries/glad/glad.cpp"
    HEAP32[164685] = $377; //@line 2533 "libraries/glad/glad.cpp"
    $378 = (FUNCTION_TABLE_ii[$load & 8191](614043)|0); //@line 2534 "libraries/glad/glad.cpp"
    HEAP32[165157] = $378; //@line 2534 "libraries/glad/glad.cpp"
    $379 = (FUNCTION_TABLE_ii[$load & 8191](614029)|0); //@line 2535 "libraries/glad/glad.cpp"
    HEAP32[164922] = $379; //@line 2535 "libraries/glad/glad.cpp"
    $380 = (FUNCTION_TABLE_ii[$load & 8191](614239)|0); //@line 2536 "libraries/glad/glad.cpp"
    HEAP32[165607] = $380; //@line 2536 "libraries/glad/glad.cpp"
    $381 = (FUNCTION_TABLE_ii[$load & 8191](450996)|0); //@line 2537 "libraries/glad/glad.cpp"
    HEAP32[165635] = $381; //@line 2537 "libraries/glad/glad.cpp"
    $382 = (FUNCTION_TABLE_ii[$load & 8191](615566)|0); //@line 2538 "libraries/glad/glad.cpp"
    HEAP32[165652] = $382; //@line 2538 "libraries/glad/glad.cpp"
    $383 = (FUNCTION_TABLE_ii[$load & 8191](451011)|0); //@line 2539 "libraries/glad/glad.cpp"
    HEAP32[165307] = $383; //@line 2539 "libraries/glad/glad.cpp"
    $384 = (FUNCTION_TABLE_ii[$load & 8191](451032)|0); //@line 2540 "libraries/glad/glad.cpp"
    HEAP32[165148] = $384; //@line 2540 "libraries/glad/glad.cpp"
    $385 = (FUNCTION_TABLE_ii[$load & 8191](615496)|0); //@line 2541 "libraries/glad/glad.cpp"
    HEAP32[165503] = $385; //@line 2541 "libraries/glad/glad.cpp"
    $386 = (FUNCTION_TABLE_ii[$load & 8191](451050)|0); //@line 2542 "libraries/glad/glad.cpp"
    HEAP32[164928] = $386; //@line 2542 "libraries/glad/glad.cpp"
    $387 = (FUNCTION_TABLE_ii[$load & 8191](451065)|0); //@line 2543 "libraries/glad/glad.cpp"
    HEAP32[165626] = $387; //@line 2543 "libraries/glad/glad.cpp"
    $388 = (FUNCTION_TABLE_ii[$load & 8191](615550)|0); //@line 2544 "libraries/glad/glad.cpp"
    HEAP32[164745] = $388; //@line 2544 "libraries/glad/glad.cpp"
    $389 = (FUNCTION_TABLE_ii[$load & 8191](615532)|0); //@line 2545 "libraries/glad/glad.cpp"
    HEAP32[165333] = $389; //@line 2545 "libraries/glad/glad.cpp"
    $390 = (FUNCTION_TABLE_ii[$load & 8191](615516)|0); //@line 2546 "libraries/glad/glad.cpp"
    HEAP32[164756] = $390; //@line 2546 "libraries/glad/glad.cpp"
    $391 = (FUNCTION_TABLE_ii[$load & 8191](451085)|0); //@line 2547 "libraries/glad/glad.cpp"
    HEAP32[165049] = $391; //@line 2547 "libraries/glad/glad.cpp"
    $392 = (FUNCTION_TABLE_ii[$load & 8191](451107)|0); //@line 2548 "libraries/glad/glad.cpp"
    HEAP32[165000] = $392; //@line 2548 "libraries/glad/glad.cpp"
    $393 = (FUNCTION_TABLE_ii[$load & 8191](451128)|0); //@line 2549 "libraries/glad/glad.cpp"
    HEAP32[165388] = $393; //@line 2549 "libraries/glad/glad.cpp"
    $394 = (FUNCTION_TABLE_ii[$load & 8191](451138)|0); //@line 2550 "libraries/glad/glad.cpp"
    HEAP32[165263] = $394; //@line 2550 "libraries/glad/glad.cpp"
    $395 = (FUNCTION_TABLE_ii[$load & 8191](451149)|0); //@line 2551 "libraries/glad/glad.cpp"
    HEAP32[165483] = $395; //@line 2551 "libraries/glad/glad.cpp"
    $396 = (FUNCTION_TABLE_ii[$load & 8191](451167)|0); //@line 2552 "libraries/glad/glad.cpp"
    HEAP32[165538] = $396; //@line 2552 "libraries/glad/glad.cpp"
   }
  } else {
   $major$188$ph$ph$i = $major$0$i;$minor$1$in90$ph$ph$i = $minor$0$in$i;$minor$194$ph$ph$i = $minor$1$i;
   label = 7;
  }
 }
 if ((label|0) == 7) {
  HEAP8[1251994] = 0; //@line 7251 "libraries/glad/glad.cpp"
  HEAP8[1251995] = 0; //@line 7252 "libraries/glad/glad.cpp"
  HEAP8[1251996] = 0; //@line 7253 "libraries/glad/glad.cpp"
  HEAP8[1251997] = 0; //@line 7254 "libraries/glad/glad.cpp"
  HEAP8[1251998] = 0; //@line 7255 "libraries/glad/glad.cpp"
  HEAP8[1251999] = 0; //@line 7256 "libraries/glad/glad.cpp"
  HEAP8[1252000] = 0; //@line 7257 "libraries/glad/glad.cpp"
  HEAP8[1252001] = 0; //@line 7258 "libraries/glad/glad.cpp"
  HEAP8[1252002] = 0; //@line 7259 "libraries/glad/glad.cpp"
  HEAP8[1252003] = 0; //@line 7260 "libraries/glad/glad.cpp"
  HEAP8[1252004] = 0; //@line 7261 "libraries/glad/glad.cpp"
  HEAP8[1252005] = 0; //@line 7262 "libraries/glad/glad.cpp"
  HEAP8[1252006] = 0; //@line 7263 "libraries/glad/glad.cpp"
  HEAP8[1252007] = 0; //@line 7264 "libraries/glad/glad.cpp"
  HEAP8[1252008] = 0; //@line 7265 "libraries/glad/glad.cpp"
  HEAP8[1252009] = 0; //@line 7266 "libraries/glad/glad.cpp"
  HEAP8[1252010] = 0; //@line 7267 "libraries/glad/glad.cpp"
  HEAP8[1252011] = 0; //@line 7268 "libraries/glad/glad.cpp"
  $22 = ($major$188$ph$ph$i|0)==(2); //@line 7269 "libraries/glad/glad.cpp"
  $23 = ($minor$194$ph$ph$i|0)>(-1); //@line 7269 "libraries/glad/glad.cpp"
  $or$cond37$i = $23 & $22; //@line 7269 "libraries/glad/glad.cpp"
  $24 = ($major$188$ph$ph$i|0)>(2); //@line 7269 "libraries/glad/glad.cpp"
  $phitmp53$i = $24&1; //@line 7269 "libraries/glad/glad.cpp"
  $25 = $or$cond37$i ? 1 : $phitmp53$i; //@line 7269 "libraries/glad/glad.cpp"
  HEAP8[1252012] = $25; //@line 7269 "libraries/glad/glad.cpp"
  $26 = ($major$188$ph$ph$i|0)==(3); //@line 7270 "libraries/glad/glad.cpp"
  $or$cond39$i = $23 & $26; //@line 7270 "libraries/glad/glad.cpp"
  $27 = ($major$188$ph$ph$i|0)>(3); //@line 7270 "libraries/glad/glad.cpp"
  $phitmp54$i = $27&1; //@line 7270 "libraries/glad/glad.cpp"
  $storemerge$i = $or$cond39$i ? 1 : $phitmp54$i; //@line 7270 "libraries/glad/glad.cpp"
  HEAP8[1252013] = $storemerge$i; //@line 7270 "libraries/glad/glad.cpp"
  $28 = ($minor$1$in90$ph$ph$i|0)>(48); //@line 7271 "libraries/glad/glad.cpp"
  $or$cond41$i = $28 & $26; //@line 7271 "libraries/glad/glad.cpp"
  $29 = $or$cond41$i ? 1 : $phitmp54$i; //@line 7271 "libraries/glad/glad.cpp"
  HEAP8[1252014] = $29; //@line 7271 "libraries/glad/glad.cpp"
  $30 = ($minor$194$ph$ph$i|0)>(1); //@line 7272 "libraries/glad/glad.cpp"
  $or$cond43$i = $30 & $26; //@line 7272 "libraries/glad/glad.cpp"
  $$phitmp55$i = $or$cond43$i ? 1 : $phitmp54$i; //@line 7272 "libraries/glad/glad.cpp"
  HEAP8[1252015] = $$phitmp55$i; //@line 7272 "libraries/glad/glad.cpp"
 }
 $397 = HEAP8[1251996]|0; //@line 2557 "libraries/glad/glad.cpp"
 $398 = ($397<<24>>24)==(0); //@line 2557 "libraries/glad/glad.cpp"
 if (!($398)) {
  $399 = (FUNCTION_TABLE_ii[$load & 8191](615476)|0); //@line 2558 "libraries/glad/glad.cpp"
  HEAP32[165510] = $399; //@line 2558 "libraries/glad/glad.cpp"
  $400 = (FUNCTION_TABLE_ii[$load & 8191](451186)|0); //@line 2559 "libraries/glad/glad.cpp"
  HEAP32[165367] = $400; //@line 2559 "libraries/glad/glad.cpp"
  $401 = (FUNCTION_TABLE_ii[$load & 8191](451199)|0); //@line 2560 "libraries/glad/glad.cpp"
  HEAP32[164963] = $401; //@line 2560 "libraries/glad/glad.cpp"
  $402 = (FUNCTION_TABLE_ii[$load & 8191](451215)|0); //@line 2561 "libraries/glad/glad.cpp"
  HEAP32[165098] = $402; //@line 2561 "libraries/glad/glad.cpp"
 }
 $403 = HEAP8[1251997]|0; //@line 2566 "libraries/glad/glad.cpp"
 $404 = ($403<<24>>24)==(0); //@line 2566 "libraries/glad/glad.cpp"
 if (!($404)) {
  $405 = (FUNCTION_TABLE_ii[$load & 8191](616144)|0); //@line 2567 "libraries/glad/glad.cpp"
  HEAP32[165288] = $405; //@line 2567 "libraries/glad/glad.cpp"
  $406 = (FUNCTION_TABLE_ii[$load & 8191](616313)|0); //@line 2568 "libraries/glad/glad.cpp"
  HEAP32[165176] = $406; //@line 2568 "libraries/glad/glad.cpp"
  $407 = (FUNCTION_TABLE_ii[$load & 8191](451235)|0); //@line 2569 "libraries/glad/glad.cpp"
  HEAP32[165031] = $407; //@line 2569 "libraries/glad/glad.cpp"
  $408 = (FUNCTION_TABLE_ii[$load & 8191](614060)|0); //@line 2570 "libraries/glad/glad.cpp"
  HEAP32[164678] = $408; //@line 2570 "libraries/glad/glad.cpp"
  $409 = (FUNCTION_TABLE_ii[$load & 8191](451258)|0); //@line 2571 "libraries/glad/glad.cpp"
  HEAP32[165494] = $409; //@line 2571 "libraries/glad/glad.cpp"
  $410 = (FUNCTION_TABLE_ii[$load & 8191](451281)|0); //@line 2572 "libraries/glad/glad.cpp"
  HEAP32[164633] = $410; //@line 2572 "libraries/glad/glad.cpp"
  $411 = (FUNCTION_TABLE_ii[$load & 8191](614083)|0); //@line 2573 "libraries/glad/glad.cpp"
  HEAP32[165449] = $411; //@line 2573 "libraries/glad/glad.cpp"
  $412 = (FUNCTION_TABLE_ii[$load & 8191](451307)|0); //@line 2574 "libraries/glad/glad.cpp"
  HEAP32[165081] = $412; //@line 2574 "libraries/glad/glad.cpp"
  $413 = (FUNCTION_TABLE_ii[$load & 8191](451333)|0); //@line 2575 "libraries/glad/glad.cpp"
  HEAP32[165185] = $413; //@line 2575 "libraries/glad/glad.cpp"
  $414 = (FUNCTION_TABLE_ii[$load & 8191](615581)|0); //@line 2576 "libraries/glad/glad.cpp"
  HEAP32[165238] = $414; //@line 2576 "libraries/glad/glad.cpp"
  $415 = (FUNCTION_TABLE_ii[$load & 8191](451357)|0); //@line 2577 "libraries/glad/glad.cpp"
  HEAP32[164762] = $415; //@line 2577 "libraries/glad/glad.cpp"
  $416 = (FUNCTION_TABLE_ii[$load & 8191](451375)|0); //@line 2578 "libraries/glad/glad.cpp"
  HEAP32[164842] = $416; //@line 2578 "libraries/glad/glad.cpp"
  $417 = (FUNCTION_TABLE_ii[$load & 8191](451394)|0); //@line 2579 "libraries/glad/glad.cpp"
  HEAP32[164763] = $417; //@line 2579 "libraries/glad/glad.cpp"
  $418 = (FUNCTION_TABLE_ii[$load & 8191](451412)|0); //@line 2580 "libraries/glad/glad.cpp"
  HEAP32[165324] = $418; //@line 2580 "libraries/glad/glad.cpp"
  $419 = (FUNCTION_TABLE_ii[$load & 8191](451431)|0); //@line 2581 "libraries/glad/glad.cpp"
  HEAP32[164769] = $419; //@line 2581 "libraries/glad/glad.cpp"
  $420 = (FUNCTION_TABLE_ii[$load & 8191](451449)|0); //@line 2582 "libraries/glad/glad.cpp"
  HEAP32[164969] = $420; //@line 2582 "libraries/glad/glad.cpp"
  $421 = (FUNCTION_TABLE_ii[$load & 8191](451468)|0); //@line 2583 "libraries/glad/glad.cpp"
  HEAP32[164773] = $421; //@line 2583 "libraries/glad/glad.cpp"
  $422 = (FUNCTION_TABLE_ii[$load & 8191](451486)|0); //@line 2584 "libraries/glad/glad.cpp"
  HEAP32[165060] = $422; //@line 2584 "libraries/glad/glad.cpp"
  $423 = (FUNCTION_TABLE_ii[$load & 8191](451505)|0); //@line 2585 "libraries/glad/glad.cpp"
  HEAP32[164991] = $423; //@line 2585 "libraries/glad/glad.cpp"
  $424 = (FUNCTION_TABLE_ii[$load & 8191](451523)|0); //@line 2586 "libraries/glad/glad.cpp"
  HEAP32[164976] = $424; //@line 2586 "libraries/glad/glad.cpp"
  $425 = (FUNCTION_TABLE_ii[$load & 8191](451542)|0); //@line 2587 "libraries/glad/glad.cpp"
  HEAP32[164992] = $425; //@line 2587 "libraries/glad/glad.cpp"
  $426 = (FUNCTION_TABLE_ii[$load & 8191](451560)|0); //@line 2588 "libraries/glad/glad.cpp"
  HEAP32[165577] = $426; //@line 2588 "libraries/glad/glad.cpp"
  $427 = (FUNCTION_TABLE_ii[$load & 8191](451579)|0); //@line 2589 "libraries/glad/glad.cpp"
  HEAP32[164844] = $427; //@line 2589 "libraries/glad/glad.cpp"
  $428 = (FUNCTION_TABLE_ii[$load & 8191](451597)|0); //@line 2590 "libraries/glad/glad.cpp"
  HEAP32[165064] = $428; //@line 2590 "libraries/glad/glad.cpp"
  $429 = (FUNCTION_TABLE_ii[$load & 8191](451616)|0); //@line 2591 "libraries/glad/glad.cpp"
  HEAP32[164826] = $429; //@line 2591 "libraries/glad/glad.cpp"
  $430 = (FUNCTION_TABLE_ii[$load & 8191](451634)|0); //@line 2592 "libraries/glad/glad.cpp"
  HEAP32[165508] = $430; //@line 2592 "libraries/glad/glad.cpp"
  $431 = (FUNCTION_TABLE_ii[$load & 8191](451653)|0); //@line 2593 "libraries/glad/glad.cpp"
  HEAP32[165222] = $431; //@line 2593 "libraries/glad/glad.cpp"
  $432 = (FUNCTION_TABLE_ii[$load & 8191](451671)|0); //@line 2594 "libraries/glad/glad.cpp"
  HEAP32[165478] = $432; //@line 2594 "libraries/glad/glad.cpp"
  $433 = (FUNCTION_TABLE_ii[$load & 8191](451690)|0); //@line 2595 "libraries/glad/glad.cpp"
  HEAP32[165220] = $433; //@line 2595 "libraries/glad/glad.cpp"
  $434 = (FUNCTION_TABLE_ii[$load & 8191](451708)|0); //@line 2596 "libraries/glad/glad.cpp"
  HEAP32[164920] = $434; //@line 2596 "libraries/glad/glad.cpp"
  $435 = (FUNCTION_TABLE_ii[$load & 8191](451727)|0); //@line 2597 "libraries/glad/glad.cpp"
  HEAP32[165219] = $435; //@line 2597 "libraries/glad/glad.cpp"
  $436 = (FUNCTION_TABLE_ii[$load & 8191](451745)|0); //@line 2598 "libraries/glad/glad.cpp"
  HEAP32[165005] = $436; //@line 2598 "libraries/glad/glad.cpp"
  $437 = (FUNCTION_TABLE_ii[$load & 8191](451764)|0); //@line 2599 "libraries/glad/glad.cpp"
  HEAP32[165215] = $437; //@line 2599 "libraries/glad/glad.cpp"
  $438 = (FUNCTION_TABLE_ii[$load & 8191](451782)|0); //@line 2600 "libraries/glad/glad.cpp"
  HEAP32[165308] = $438; //@line 2600 "libraries/glad/glad.cpp"
  $439 = (FUNCTION_TABLE_ii[$load & 8191](451801)|0); //@line 2601 "libraries/glad/glad.cpp"
  HEAP32[165453] = $439; //@line 2601 "libraries/glad/glad.cpp"
  $440 = (FUNCTION_TABLE_ii[$load & 8191](451819)|0); //@line 2602 "libraries/glad/glad.cpp"
  HEAP32[165465] = $440; //@line 2602 "libraries/glad/glad.cpp"
  $441 = (FUNCTION_TABLE_ii[$load & 8191](451838)|0); //@line 2603 "libraries/glad/glad.cpp"
  HEAP32[165451] = $441; //@line 2603 "libraries/glad/glad.cpp"
  $442 = (FUNCTION_TABLE_ii[$load & 8191](451856)|0); //@line 2604 "libraries/glad/glad.cpp"
  HEAP32[165026] = $442; //@line 2604 "libraries/glad/glad.cpp"
  $443 = (FUNCTION_TABLE_ii[$load & 8191](451875)|0); //@line 2605 "libraries/glad/glad.cpp"
  HEAP32[165459] = $443; //@line 2605 "libraries/glad/glad.cpp"
  $444 = (FUNCTION_TABLE_ii[$load & 8191](451893)|0); //@line 2606 "libraries/glad/glad.cpp"
  HEAP32[165319] = $444; //@line 2606 "libraries/glad/glad.cpp"
  $445 = (FUNCTION_TABLE_ii[$load & 8191](451912)|0); //@line 2607 "libraries/glad/glad.cpp"
  HEAP32[165440] = $445; //@line 2607 "libraries/glad/glad.cpp"
  $446 = (FUNCTION_TABLE_ii[$load & 8191](451930)|0); //@line 2608 "libraries/glad/glad.cpp"
  HEAP32[164866] = $446; //@line 2608 "libraries/glad/glad.cpp"
  $447 = (FUNCTION_TABLE_ii[$load & 8191](451949)|0); //@line 2609 "libraries/glad/glad.cpp"
  HEAP32[165345] = $447; //@line 2609 "libraries/glad/glad.cpp"
  $448 = (FUNCTION_TABLE_ii[$load & 8191](451972)|0); //@line 2610 "libraries/glad/glad.cpp"
  HEAP32[165344] = $448; //@line 2610 "libraries/glad/glad.cpp"
  $449 = (FUNCTION_TABLE_ii[$load & 8191](451995)|0); //@line 2611 "libraries/glad/glad.cpp"
  HEAP32[165125] = $449; //@line 2611 "libraries/glad/glad.cpp"
  $450 = (FUNCTION_TABLE_ii[$load & 8191](452018)|0); //@line 2612 "libraries/glad/glad.cpp"
  HEAP32[165394] = $450; //@line 2612 "libraries/glad/glad.cpp"
 }
 $451 = HEAP8[1251998]|0; //@line 2617 "libraries/glad/glad.cpp"
 $452 = ($451<<24>>24)==(0); //@line 2617 "libraries/glad/glad.cpp"
 if (!($452)) {
  $453 = (FUNCTION_TABLE_ii[$load & 8191](616490)|0); //@line 2618 "libraries/glad/glad.cpp"
  HEAP32[164680] = $453; //@line 2618 "libraries/glad/glad.cpp"
  $454 = (FUNCTION_TABLE_ii[$load & 8191](452041)|0); //@line 2619 "libraries/glad/glad.cpp"
  HEAP32[165524] = $454; //@line 2619 "libraries/glad/glad.cpp"
  $455 = (FUNCTION_TABLE_ii[$load & 8191](452059)|0); //@line 2620 "libraries/glad/glad.cpp"
  HEAP32[165040] = $455; //@line 2620 "libraries/glad/glad.cpp"
  $456 = (FUNCTION_TABLE_ii[$load & 8191](452079)|0); //@line 2621 "libraries/glad/glad.cpp"
  HEAP32[165436] = $456; //@line 2621 "libraries/glad/glad.cpp"
  $457 = (FUNCTION_TABLE_ii[$load & 8191](452097)|0); //@line 2622 "libraries/glad/glad.cpp"
  HEAP32[165416] = $457; //@line 2622 "libraries/glad/glad.cpp"
  $458 = (FUNCTION_TABLE_ii[$load & 8191](452116)|0); //@line 2623 "libraries/glad/glad.cpp"
  HEAP32[165401] = $458; //@line 2623 "libraries/glad/glad.cpp"
  $459 = (FUNCTION_TABLE_ii[$load & 8191](452134)|0); //@line 2624 "libraries/glad/glad.cpp"
  HEAP32[164830] = $459; //@line 2624 "libraries/glad/glad.cpp"
  $460 = (FUNCTION_TABLE_ii[$load & 8191](452153)|0); //@line 2625 "libraries/glad/glad.cpp"
  HEAP32[164966] = $460; //@line 2625 "libraries/glad/glad.cpp"
  $461 = (FUNCTION_TABLE_ii[$load & 8191](452165)|0); //@line 2626 "libraries/glad/glad.cpp"
  HEAP32[164793] = $461; //@line 2626 "libraries/glad/glad.cpp"
  $462 = (FUNCTION_TABLE_ii[$load & 8191](452178)|0); //@line 2627 "libraries/glad/glad.cpp"
  HEAP32[164965] = $462; //@line 2627 "libraries/glad/glad.cpp"
  $463 = (FUNCTION_TABLE_ii[$load & 8191](452190)|0); //@line 2628 "libraries/glad/glad.cpp"
  HEAP32[165224] = $463; //@line 2628 "libraries/glad/glad.cpp"
  $464 = (FUNCTION_TABLE_ii[$load & 8191](452203)|0); //@line 2629 "libraries/glad/glad.cpp"
  HEAP32[165228] = $464; //@line 2629 "libraries/glad/glad.cpp"
  $465 = (FUNCTION_TABLE_ii[$load & 8191](452221)|0); //@line 2630 "libraries/glad/glad.cpp"
  HEAP32[165028] = $465; //@line 2630 "libraries/glad/glad.cpp"
  $466 = (FUNCTION_TABLE_ii[$load & 8191](452240)|0); //@line 2631 "libraries/glad/glad.cpp"
  HEAP32[164790] = $466; //@line 2631 "libraries/glad/glad.cpp"
  $467 = (FUNCTION_TABLE_ii[$load & 8191](452260)|0); //@line 2632 "libraries/glad/glad.cpp"
  HEAP32[165024] = $467; //@line 2632 "libraries/glad/glad.cpp"
  $468 = (FUNCTION_TABLE_ii[$load & 8191](452279)|0); //@line 2633 "libraries/glad/glad.cpp"
  HEAP32[165237] = $468; //@line 2633 "libraries/glad/glad.cpp"
  $469 = (FUNCTION_TABLE_ii[$load & 8191](452299)|0); //@line 2634 "libraries/glad/glad.cpp"
  HEAP32[165068] = $469; //@line 2634 "libraries/glad/glad.cpp"
  $470 = (FUNCTION_TABLE_ii[$load & 8191](452318)|0); //@line 2635 "libraries/glad/glad.cpp"
  HEAP32[164659] = $470; //@line 2635 "libraries/glad/glad.cpp"
  $471 = (FUNCTION_TABLE_ii[$load & 8191](452338)|0); //@line 2636 "libraries/glad/glad.cpp"
  HEAP32[165197] = $471; //@line 2636 "libraries/glad/glad.cpp"
  $472 = (FUNCTION_TABLE_ii[$load & 8191](452357)|0); //@line 2637 "libraries/glad/glad.cpp"
  HEAP32[165392] = $472; //@line 2637 "libraries/glad/glad.cpp"
  $473 = (FUNCTION_TABLE_ii[$load & 8191](452377)|0); //@line 2638 "libraries/glad/glad.cpp"
  HEAP32[164862] = $473; //@line 2638 "libraries/glad/glad.cpp"
  $474 = (FUNCTION_TABLE_ii[$load & 8191](452396)|0); //@line 2639 "libraries/glad/glad.cpp"
  HEAP32[164954] = $474; //@line 2639 "libraries/glad/glad.cpp"
  $475 = (FUNCTION_TABLE_ii[$load & 8191](452416)|0); //@line 2640 "libraries/glad/glad.cpp"
  HEAP32[165421] = $475; //@line 2640 "libraries/glad/glad.cpp"
  $476 = (FUNCTION_TABLE_ii[$load & 8191](452436)|0); //@line 2641 "libraries/glad/glad.cpp"
  HEAP32[165003] = $476; //@line 2641 "libraries/glad/glad.cpp"
  $477 = (FUNCTION_TABLE_ii[$load & 8191](452457)|0); //@line 2642 "libraries/glad/glad.cpp"
  HEAP32[165424] = $477; //@line 2642 "libraries/glad/glad.cpp"
  $478 = (FUNCTION_TABLE_ii[$load & 8191](452477)|0); //@line 2643 "libraries/glad/glad.cpp"
  HEAP32[164757] = $478; //@line 2643 "libraries/glad/glad.cpp"
  $479 = (FUNCTION_TABLE_ii[$load & 8191](452498)|0); //@line 2644 "libraries/glad/glad.cpp"
  HEAP32[165242] = $479; //@line 2644 "libraries/glad/glad.cpp"
  $480 = (FUNCTION_TABLE_ii[$load & 8191](452518)|0); //@line 2645 "libraries/glad/glad.cpp"
  HEAP32[164846] = $480; //@line 2645 "libraries/glad/glad.cpp"
  $481 = (FUNCTION_TABLE_ii[$load & 8191](452539)|0); //@line 2646 "libraries/glad/glad.cpp"
  HEAP32[165314] = $481; //@line 2646 "libraries/glad/glad.cpp"
  $482 = (FUNCTION_TABLE_ii[$load & 8191](452563)|0); //@line 2647 "libraries/glad/glad.cpp"
  HEAP32[164627] = $482; //@line 2647 "libraries/glad/glad.cpp"
  $483 = (FUNCTION_TABLE_ii[$load & 8191](452577)|0); //@line 2648 "libraries/glad/glad.cpp"
  HEAP32[165187] = $483; //@line 2648 "libraries/glad/glad.cpp"
  $484 = (FUNCTION_TABLE_ii[$load & 8191](452592)|0); //@line 2649 "libraries/glad/glad.cpp"
  HEAP32[164626] = $484; //@line 2649 "libraries/glad/glad.cpp"
  $485 = (FUNCTION_TABLE_ii[$load & 8191](452606)|0); //@line 2650 "libraries/glad/glad.cpp"
  HEAP32[165532] = $485; //@line 2650 "libraries/glad/glad.cpp"
  $486 = (FUNCTION_TABLE_ii[$load & 8191](452621)|0); //@line 2651 "libraries/glad/glad.cpp"
  HEAP32[164625] = $486; //@line 2651 "libraries/glad/glad.cpp"
  $487 = (FUNCTION_TABLE_ii[$load & 8191](452635)|0); //@line 2652 "libraries/glad/glad.cpp"
  HEAP32[164648] = $487; //@line 2652 "libraries/glad/glad.cpp"
  $488 = (FUNCTION_TABLE_ii[$load & 8191](452650)|0); //@line 2653 "libraries/glad/glad.cpp"
  HEAP32[164622] = $488; //@line 2653 "libraries/glad/glad.cpp"
  $489 = (FUNCTION_TABLE_ii[$load & 8191](452664)|0); //@line 2654 "libraries/glad/glad.cpp"
  HEAP32[164670] = $489; //@line 2654 "libraries/glad/glad.cpp"
  $490 = (FUNCTION_TABLE_ii[$load & 8191](452679)|0); //@line 2655 "libraries/glad/glad.cpp"
  HEAP32[165407] = $490; //@line 2655 "libraries/glad/glad.cpp"
  $491 = (FUNCTION_TABLE_ii[$load & 8191](452693)|0); //@line 2656 "libraries/glad/glad.cpp"
  HEAP32[165601] = $491; //@line 2656 "libraries/glad/glad.cpp"
  $492 = (FUNCTION_TABLE_ii[$load & 8191](452708)|0); //@line 2657 "libraries/glad/glad.cpp"
  HEAP32[165442] = $492; //@line 2657 "libraries/glad/glad.cpp"
  $493 = (FUNCTION_TABLE_ii[$load & 8191](452722)|0); //@line 2658 "libraries/glad/glad.cpp"
  HEAP32[165252] = $493; //@line 2658 "libraries/glad/glad.cpp"
  $494 = (FUNCTION_TABLE_ii[$load & 8191](452737)|0); //@line 2659 "libraries/glad/glad.cpp"
  HEAP32[165439] = $494; //@line 2659 "libraries/glad/glad.cpp"
  $495 = (FUNCTION_TABLE_ii[$load & 8191](452751)|0); //@line 2660 "libraries/glad/glad.cpp"
  HEAP32[165015] = $495; //@line 2660 "libraries/glad/glad.cpp"
  $496 = (FUNCTION_TABLE_ii[$load & 8191](452766)|0); //@line 2661 "libraries/glad/glad.cpp"
  HEAP32[165435] = $496; //@line 2661 "libraries/glad/glad.cpp"
  $497 = (FUNCTION_TABLE_ii[$load & 8191](452780)|0); //@line 2662 "libraries/glad/glad.cpp"
  HEAP32[164874] = $497; //@line 2662 "libraries/glad/glad.cpp"
  $498 = (FUNCTION_TABLE_ii[$load & 8191](616510)|0); //@line 2663 "libraries/glad/glad.cpp"
  HEAP32[165404] = $498; //@line 2663 "libraries/glad/glad.cpp"
  $499 = (FUNCTION_TABLE_ii[$load & 8191](616160)|0); //@line 2664 "libraries/glad/glad.cpp"
  HEAP32[165469] = $499; //@line 2664 "libraries/glad/glad.cpp"
 }
 $500 = HEAP8[1251999]|0; //@line 2669 "libraries/glad/glad.cpp"
 $501 = ($500<<24>>24)==(0); //@line 2669 "libraries/glad/glad.cpp"
 if (!($501)) {
  $502 = (FUNCTION_TABLE_ii[$load & 8191](452795)|0); //@line 2670 "libraries/glad/glad.cpp"
  HEAP32[164959] = $502; //@line 2670 "libraries/glad/glad.cpp"
  $503 = (FUNCTION_TABLE_ii[$load & 8191](452808)|0); //@line 2671 "libraries/glad/glad.cpp"
  HEAP32[165161] = $503; //@line 2671 "libraries/glad/glad.cpp"
  $504 = (FUNCTION_TABLE_ii[$load & 8191](452824)|0); //@line 2672 "libraries/glad/glad.cpp"
  HEAP32[165596] = $504; //@line 2672 "libraries/glad/glad.cpp"
  $505 = (FUNCTION_TABLE_ii[$load & 8191](452834)|0); //@line 2673 "libraries/glad/glad.cpp"
  HEAP32[165555] = $505; //@line 2673 "libraries/glad/glad.cpp"
  $506 = (FUNCTION_TABLE_ii[$load & 8191](452847)|0); //@line 2674 "libraries/glad/glad.cpp"
  HEAP32[165146] = $506; //@line 2674 "libraries/glad/glad.cpp"
  $507 = (FUNCTION_TABLE_ii[$load & 8191](452858)|0); //@line 2675 "libraries/glad/glad.cpp"
  HEAP32[164813] = $507; //@line 2675 "libraries/glad/glad.cpp"
  $508 = (FUNCTION_TABLE_ii[$load & 8191](452871)|0); //@line 2676 "libraries/glad/glad.cpp"
  HEAP32[165321] = $508; //@line 2676 "libraries/glad/glad.cpp"
  $509 = (FUNCTION_TABLE_ii[$load & 8191](452890)|0); //@line 2677 "libraries/glad/glad.cpp"
  HEAP32[164776] = $509; //@line 2677 "libraries/glad/glad.cpp"
  $510 = (FUNCTION_TABLE_ii[$load & 8191](614842)|0); //@line 2678 "libraries/glad/glad.cpp"
  HEAP32[165557] = $510; //@line 2678 "libraries/glad/glad.cpp"
  $511 = (FUNCTION_TABLE_ii[$load & 8191](614264)|0); //@line 2679 "libraries/glad/glad.cpp"
  HEAP32[165381] = $511; //@line 2679 "libraries/glad/glad.cpp"
  $512 = (FUNCTION_TABLE_ii[$load & 8191](614251)|0); //@line 2680 "libraries/glad/glad.cpp"
  HEAP32[165357] = $512; //@line 2680 "libraries/glad/glad.cpp"
  $513 = (FUNCTION_TABLE_ii[$load & 8191](614332)|0); //@line 2681 "libraries/glad/glad.cpp"
  HEAP32[164673] = $513; //@line 2681 "libraries/glad/glad.cpp"
  $514 = (FUNCTION_TABLE_ii[$load & 8191](614303)|0); //@line 2682 "libraries/glad/glad.cpp"
  HEAP32[165565] = $514; //@line 2682 "libraries/glad/glad.cpp"
  $515 = (FUNCTION_TABLE_ii[$load & 8191](614316)|0); //@line 2683 "libraries/glad/glad.cpp"
  HEAP32[165513] = $515; //@line 2683 "libraries/glad/glad.cpp"
  $516 = (FUNCTION_TABLE_ii[$load & 8191](452910)|0); //@line 2684 "libraries/glad/glad.cpp"
  HEAP32[165350] = $516; //@line 2684 "libraries/glad/glad.cpp"
  $517 = (FUNCTION_TABLE_ii[$load & 8191](452929)|0); //@line 2685 "libraries/glad/glad.cpp"
  HEAP32[165023] = $517; //@line 2685 "libraries/glad/glad.cpp"
  $518 = (FUNCTION_TABLE_ii[$load & 8191](452941)|0); //@line 2686 "libraries/glad/glad.cpp"
  HEAP32[165265] = $518; //@line 2686 "libraries/glad/glad.cpp"
  $519 = (FUNCTION_TABLE_ii[$load & 8191](614280)|0); //@line 2687 "libraries/glad/glad.cpp"
  HEAP32[165063] = $519; //@line 2687 "libraries/glad/glad.cpp"
  $520 = (FUNCTION_TABLE_ii[$load & 8191](452955)|0); //@line 2688 "libraries/glad/glad.cpp"
  HEAP32[165590] = $520; //@line 2688 "libraries/glad/glad.cpp"
 }
 $521 = HEAP8[1252000]|0; //@line 2693 "libraries/glad/glad.cpp"
 $522 = ($521<<24>>24)==(0); //@line 2693 "libraries/glad/glad.cpp"
 if (!($522)) {
  $523 = (FUNCTION_TABLE_ii[$load & 8191](616200)|0); //@line 2694 "libraries/glad/glad.cpp"
  HEAP32[165355] = $523; //@line 2694 "libraries/glad/glad.cpp"
  $524 = (FUNCTION_TABLE_ii[$load & 8191](616619)|0); //@line 2695 "libraries/glad/glad.cpp"
  HEAP32[165046] = $524; //@line 2695 "libraries/glad/glad.cpp"
  $525 = (FUNCTION_TABLE_ii[$load & 8191](616545)|0); //@line 2696 "libraries/glad/glad.cpp"
  HEAP32[165159] = $525; //@line 2696 "libraries/glad/glad.cpp"
  $526 = (FUNCTION_TABLE_ii[$load & 8191](616523)|0); //@line 2697 "libraries/glad/glad.cpp"
  HEAP32[164911] = $526; //@line 2697 "libraries/glad/glad.cpp"
  $527 = (FUNCTION_TABLE_ii[$load & 8191](616251)|0); //@line 2698 "libraries/glad/glad.cpp"
  HEAP32[164654] = $527; //@line 2698 "libraries/glad/glad.cpp"
  $528 = (FUNCTION_TABLE_ii[$load & 8191](613918)|0); //@line 2699 "libraries/glad/glad.cpp"
  HEAP32[165223] = $528; //@line 2699 "libraries/glad/glad.cpp"
  $529 = (FUNCTION_TABLE_ii[$load & 8191](615216)|0); //@line 2700 "libraries/glad/glad.cpp"
  HEAP32[165504] = $529; //@line 2700 "libraries/glad/glad.cpp"
  $530 = (FUNCTION_TABLE_ii[$load & 8191](615034)|0); //@line 2701 "libraries/glad/glad.cpp"
  HEAP32[164646] = $530; //@line 2701 "libraries/glad/glad.cpp"
  $531 = (FUNCTION_TABLE_ii[$load & 8191](613819)|0); //@line 2702 "libraries/glad/glad.cpp"
  HEAP32[165365] = $531; //@line 2702 "libraries/glad/glad.cpp"
  $532 = (FUNCTION_TABLE_ii[$load & 8191](613888)|0); //@line 2703 "libraries/glad/glad.cpp"
  HEAP32[164672] = $532; //@line 2703 "libraries/glad/glad.cpp"
  $533 = (FUNCTION_TABLE_ii[$load & 8191](615109)|0); //@line 2704 "libraries/glad/glad.cpp"
  HEAP32[164697] = $533; //@line 2704 "libraries/glad/glad.cpp"
  $534 = (FUNCTION_TABLE_ii[$load & 8191](614965)|0); //@line 2705 "libraries/glad/glad.cpp"
  HEAP32[165085] = $534; //@line 2705 "libraries/glad/glad.cpp"
  $535 = (FUNCTION_TABLE_ii[$load & 8191](613948)|0); //@line 2706 "libraries/glad/glad.cpp"
  HEAP32[165143] = $535; //@line 2706 "libraries/glad/glad.cpp"
  $536 = (FUNCTION_TABLE_ii[$load & 8191](615770)|0); //@line 2707 "libraries/glad/glad.cpp"
  HEAP32[164811] = $536; //@line 2707 "libraries/glad/glad.cpp"
  $537 = (FUNCTION_TABLE_ii[$load & 8191](615744)|0); //@line 2708 "libraries/glad/glad.cpp"
  HEAP32[165289] = $537; //@line 2708 "libraries/glad/glad.cpp"
  $538 = (FUNCTION_TABLE_ii[$load & 8191](614947)|0); //@line 2709 "libraries/glad/glad.cpp"
  HEAP32[165189] = $538; //@line 2709 "libraries/glad/glad.cpp"
  $539 = (FUNCTION_TABLE_ii[$load & 8191](614566)|0); //@line 2710 "libraries/glad/glad.cpp"
  HEAP32[165651] = $539; //@line 2710 "libraries/glad/glad.cpp"
  $540 = (FUNCTION_TABLE_ii[$load & 8191](614980)|0); //@line 2711 "libraries/glad/glad.cpp"
  HEAP32[164803] = $540; //@line 2711 "libraries/glad/glad.cpp"
  $541 = (FUNCTION_TABLE_ii[$load & 8191](614927)|0); //@line 2712 "libraries/glad/glad.cpp"
  HEAP32[164909] = $541; //@line 2712 "libraries/glad/glad.cpp"
  $542 = (FUNCTION_TABLE_ii[$load & 8191](615083)|0); //@line 2713 "libraries/glad/glad.cpp"
  HEAP32[165573] = $542; //@line 2713 "libraries/glad/glad.cpp"
  $543 = (FUNCTION_TABLE_ii[$load & 8191](615166)|0); //@line 2714 "libraries/glad/glad.cpp"
  HEAP32[165061] = $543; //@line 2714 "libraries/glad/glad.cpp"
  $544 = (FUNCTION_TABLE_ii[$load & 8191](615069)|0); //@line 2715 "libraries/glad/glad.cpp"
  HEAP32[164990] = $544; //@line 2715 "libraries/glad/glad.cpp"
  $545 = (FUNCTION_TABLE_ii[$load & 8191](615050)|0); //@line 2716 "libraries/glad/glad.cpp"
  HEAP32[164795] = $545; //@line 2716 "libraries/glad/glad.cpp"
  $546 = (FUNCTION_TABLE_ii[$load & 8191](615016)|0); //@line 2717 "libraries/glad/glad.cpp"
  HEAP32[165110] = $546; //@line 2717 "libraries/glad/glad.cpp"
  $547 = (FUNCTION_TABLE_ii[$load & 8191](614479)|0); //@line 2718 "libraries/glad/glad.cpp"
  HEAP32[165470] = $547; //@line 2718 "libraries/glad/glad.cpp"
  $548 = (FUNCTION_TABLE_ii[$load & 8191](614449)|0); //@line 2719 "libraries/glad/glad.cpp"
  HEAP32[165303] = $548; //@line 2719 "libraries/glad/glad.cpp"
  $549 = (FUNCTION_TABLE_ii[$load & 8191](614464)|0); //@line 2720 "libraries/glad/glad.cpp"
  HEAP32[165009] = $549; //@line 2720 "libraries/glad/glad.cpp"
  $550 = (FUNCTION_TABLE_ii[$load & 8191](452975)|0); //@line 2721 "libraries/glad/glad.cpp"
  HEAP32[164741] = $550; //@line 2721 "libraries/glad/glad.cpp"
  $551 = (FUNCTION_TABLE_ii[$load & 8191](614500)|0); //@line 2722 "libraries/glad/glad.cpp"
  HEAP32[165182] = $551; //@line 2722 "libraries/glad/glad.cpp"
  $552 = (FUNCTION_TABLE_ii[$load & 8191](614520)|0); //@line 2723 "libraries/glad/glad.cpp"
  HEAP32[165038] = $552; //@line 2723 "libraries/glad/glad.cpp"
  $553 = (FUNCTION_TABLE_ii[$load & 8191](614540)|0); //@line 2724 "libraries/glad/glad.cpp"
  HEAP32[164890] = $553; //@line 2724 "libraries/glad/glad.cpp"
  $554 = (FUNCTION_TABLE_ii[$load & 8191](615204)|0); //@line 2725 "libraries/glad/glad.cpp"
  HEAP32[165280] = $554; //@line 2725 "libraries/glad/glad.cpp"
  $555 = (FUNCTION_TABLE_ii[$load & 8191](615098)|0); //@line 2726 "libraries/glad/glad.cpp"
  HEAP32[164825] = $555; //@line 2726 "libraries/glad/glad.cpp"
  $556 = (FUNCTION_TABLE_ii[$load & 8191](615152)|0); //@line 2727 "libraries/glad/glad.cpp"
  HEAP32[165133] = $556; //@line 2727 "libraries/glad/glad.cpp"
  $557 = (FUNCTION_TABLE_ii[$load & 8191](615001)|0); //@line 2728 "libraries/glad/glad.cpp"
  HEAP32[165539] = $557; //@line 2728 "libraries/glad/glad.cpp"
  $558 = (FUNCTION_TABLE_ii[$load & 8191](613854)|0); //@line 2729 "libraries/glad/glad.cpp"
  HEAP32[165058] = $558; //@line 2729 "libraries/glad/glad.cpp"
  $559 = (FUNCTION_TABLE_ii[$load & 8191](614585)|0); //@line 2730 "libraries/glad/glad.cpp"
  HEAP32[165181] = $559; //@line 2730 "libraries/glad/glad.cpp"
  $560 = (FUNCTION_TABLE_ii[$load & 8191](614597)|0); //@line 2731 "libraries/glad/glad.cpp"
  HEAP32[165553] = $560; //@line 2731 "libraries/glad/glad.cpp"
  $561 = (FUNCTION_TABLE_ii[$load & 8191](614609)|0); //@line 2732 "libraries/glad/glad.cpp"
  HEAP32[164716] = $561; //@line 2732 "libraries/glad/glad.cpp"
  $562 = (FUNCTION_TABLE_ii[$load & 8191](614621)|0); //@line 2733 "libraries/glad/glad.cpp"
  HEAP32[165297] = $562; //@line 2733 "libraries/glad/glad.cpp"
  $563 = (FUNCTION_TABLE_ii[$load & 8191](614633)|0); //@line 2734 "libraries/glad/glad.cpp"
  HEAP32[165188] = $563; //@line 2734 "libraries/glad/glad.cpp"
  $564 = (FUNCTION_TABLE_ii[$load & 8191](614645)|0); //@line 2735 "libraries/glad/glad.cpp"
  HEAP32[164937] = $564; //@line 2735 "libraries/glad/glad.cpp"
  $565 = (FUNCTION_TABLE_ii[$load & 8191](614657)|0); //@line 2736 "libraries/glad/glad.cpp"
  HEAP32[165190] = $565; //@line 2736 "libraries/glad/glad.cpp"
  $566 = (FUNCTION_TABLE_ii[$load & 8191](614669)|0); //@line 2737 "libraries/glad/glad.cpp"
  HEAP32[165287] = $566; //@line 2737 "libraries/glad/glad.cpp"
  $567 = (FUNCTION_TABLE_ii[$load & 8191](614733)|0); //@line 2738 "libraries/glad/glad.cpp"
  HEAP32[165610] = $567; //@line 2738 "libraries/glad/glad.cpp"
  $568 = (FUNCTION_TABLE_ii[$load & 8191](614746)|0); //@line 2739 "libraries/glad/glad.cpp"
  HEAP32[165418] = $568; //@line 2739 "libraries/glad/glad.cpp"
  $569 = (FUNCTION_TABLE_ii[$load & 8191](614759)|0); //@line 2740 "libraries/glad/glad.cpp"
  HEAP32[165017] = $569; //@line 2740 "libraries/glad/glad.cpp"
  $570 = (FUNCTION_TABLE_ii[$load & 8191](614772)|0); //@line 2741 "libraries/glad/glad.cpp"
  HEAP32[165650] = $570; //@line 2741 "libraries/glad/glad.cpp"
  $571 = (FUNCTION_TABLE_ii[$load & 8191](614681)|0); //@line 2742 "libraries/glad/glad.cpp"
  HEAP32[165316] = $571; //@line 2742 "libraries/glad/glad.cpp"
  $572 = (FUNCTION_TABLE_ii[$load & 8191](614694)|0); //@line 2743 "libraries/glad/glad.cpp"
  HEAP32[164977] = $572; //@line 2743 "libraries/glad/glad.cpp"
  $573 = (FUNCTION_TABLE_ii[$load & 8191](614707)|0); //@line 2744 "libraries/glad/glad.cpp"
  HEAP32[165043] = $573; //@line 2744 "libraries/glad/glad.cpp"
  $574 = (FUNCTION_TABLE_ii[$load & 8191](614720)|0); //@line 2745 "libraries/glad/glad.cpp"
  HEAP32[164915] = $574; //@line 2745 "libraries/glad/glad.cpp"
  $575 = (FUNCTION_TABLE_ii[$load & 8191](614785)|0); //@line 2746 "libraries/glad/glad.cpp"
  HEAP32[165563] = $575; //@line 2746 "libraries/glad/glad.cpp"
  $576 = (FUNCTION_TABLE_ii[$load & 8191](614804)|0); //@line 2747 "libraries/glad/glad.cpp"
  HEAP32[165618] = $576; //@line 2747 "libraries/glad/glad.cpp"
  $577 = (FUNCTION_TABLE_ii[$load & 8191](614823)|0); //@line 2748 "libraries/glad/glad.cpp"
  HEAP32[165649] = $577; //@line 2748 "libraries/glad/glad.cpp"
  $578 = (FUNCTION_TABLE_ii[$load & 8191](615186)|0); //@line 2749 "libraries/glad/glad.cpp"
  HEAP32[165115] = $578; //@line 2749 "libraries/glad/glad.cpp"
  $579 = (FUNCTION_TABLE_ii[$load & 8191](452995)|0); //@line 2750 "libraries/glad/glad.cpp"
  HEAP32[165006] = $579; //@line 2750 "libraries/glad/glad.cpp"
  $580 = (FUNCTION_TABLE_ii[$load & 8191](453012)|0); //@line 2751 "libraries/glad/glad.cpp"
  HEAP32[165509] = $580; //@line 2751 "libraries/glad/glad.cpp"
  $581 = (FUNCTION_TABLE_ii[$load & 8191](616296)|0); //@line 2752 "libraries/glad/glad.cpp"
  HEAP32[164679] = $581; //@line 2752 "libraries/glad/glad.cpp"
  $582 = (FUNCTION_TABLE_ii[$load & 8191](614855)|0); //@line 2753 "libraries/glad/glad.cpp"
  HEAP32[164978] = $582; //@line 2753 "libraries/glad/glad.cpp"
  $583 = (FUNCTION_TABLE_ii[$load & 8191](453030)|0); //@line 2754 "libraries/glad/glad.cpp"
  HEAP32[164686] = $583; //@line 2754 "libraries/glad/glad.cpp"
  $584 = (FUNCTION_TABLE_ii[$load & 8191](453047)|0); //@line 2755 "libraries/glad/glad.cpp"
  HEAP32[164638] = $584; //@line 2755 "libraries/glad/glad.cpp"
  $585 = (FUNCTION_TABLE_ii[$load & 8191](453065)|0); //@line 2756 "libraries/glad/glad.cpp"
  HEAP32[165639] = $585; //@line 2756 "libraries/glad/glad.cpp"
  $586 = (FUNCTION_TABLE_ii[$load & 8191](453082)|0); //@line 2757 "libraries/glad/glad.cpp"
  HEAP32[164696] = $586; //@line 2757 "libraries/glad/glad.cpp"
  $587 = (FUNCTION_TABLE_ii[$load & 8191](616362)|0); //@line 2758 "libraries/glad/glad.cpp"
  HEAP32[165641] = $587; //@line 2758 "libraries/glad/glad.cpp"
  $588 = (FUNCTION_TABLE_ii[$load & 8191](614873)|0); //@line 2759 "libraries/glad/glad.cpp"
  HEAP32[165579] = $588; //@line 2759 "libraries/glad/glad.cpp"
  $589 = (FUNCTION_TABLE_ii[$load & 8191](453100)|0); //@line 2760 "libraries/glad/glad.cpp"
  HEAP32[165645] = $589; //@line 2760 "libraries/glad/glad.cpp"
  $590 = (FUNCTION_TABLE_ii[$load & 8191](453117)|0); //@line 2761 "libraries/glad/glad.cpp"
  HEAP32[165360] = $590; //@line 2761 "libraries/glad/glad.cpp"
  $591 = (FUNCTION_TABLE_ii[$load & 8191](453135)|0); //@line 2762 "libraries/glad/glad.cpp"
  HEAP32[164767] = $591; //@line 2762 "libraries/glad/glad.cpp"
  $592 = (FUNCTION_TABLE_ii[$load & 8191](453152)|0); //@line 2763 "libraries/glad/glad.cpp"
  HEAP32[165642] = $592; //@line 2763 "libraries/glad/glad.cpp"
  $593 = (FUNCTION_TABLE_ii[$load & 8191](616439)|0); //@line 2764 "libraries/glad/glad.cpp"
  HEAP32[164736] = $593; //@line 2764 "libraries/glad/glad.cpp"
  $594 = (FUNCTION_TABLE_ii[$load & 8191](614891)|0); //@line 2765 "libraries/glad/glad.cpp"
  HEAP32[165041] = $594; //@line 2765 "libraries/glad/glad.cpp"
  $595 = (FUNCTION_TABLE_ii[$load & 8191](453170)|0); //@line 2766 "libraries/glad/glad.cpp"
  HEAP32[165272] = $595; //@line 2766 "libraries/glad/glad.cpp"
  $596 = (FUNCTION_TABLE_ii[$load & 8191](453187)|0); //@line 2767 "libraries/glad/glad.cpp"
  HEAP32[164897] = $596; //@line 2767 "libraries/glad/glad.cpp"
  $597 = (FUNCTION_TABLE_ii[$load & 8191](453205)|0); //@line 2768 "libraries/glad/glad.cpp"
  HEAP32[164930] = $597; //@line 2768 "libraries/glad/glad.cpp"
  $598 = (FUNCTION_TABLE_ii[$load & 8191](453224)|0); //@line 2769 "libraries/glad/glad.cpp"
  HEAP32[165398] = $598; //@line 2769 "libraries/glad/glad.cpp"
  $599 = (FUNCTION_TABLE_ii[$load & 8191](453243)|0); //@line 2770 "libraries/glad/glad.cpp"
  HEAP32[165091] = $599; //@line 2770 "libraries/glad/glad.cpp"
  $600 = (FUNCTION_TABLE_ii[$load & 8191](453262)|0); //@line 2771 "libraries/glad/glad.cpp"
  HEAP32[164780] = $600; //@line 2771 "libraries/glad/glad.cpp"
  $601 = (FUNCTION_TABLE_ii[$load & 8191](453281)|0); //@line 2772 "libraries/glad/glad.cpp"
  HEAP32[165336] = $601; //@line 2772 "libraries/glad/glad.cpp"
  $602 = (FUNCTION_TABLE_ii[$load & 8191](453301)|0); //@line 2773 "libraries/glad/glad.cpp"
  HEAP32[164949] = $602; //@line 2773 "libraries/glad/glad.cpp"
  $603 = (FUNCTION_TABLE_ii[$load & 8191](453321)|0); //@line 2774 "libraries/glad/glad.cpp"
  HEAP32[165446] = $603; //@line 2774 "libraries/glad/glad.cpp"
  $604 = (FUNCTION_TABLE_ii[$load & 8191](453341)|0); //@line 2775 "libraries/glad/glad.cpp"
  HEAP32[165281] = $604; //@line 2775 "libraries/glad/glad.cpp"
  $605 = (FUNCTION_TABLE_ii[$load & 8191](453359)|0); //@line 2776 "libraries/glad/glad.cpp"
  HEAP32[165164] = $605; //@line 2776 "libraries/glad/glad.cpp"
  $606 = (FUNCTION_TABLE_ii[$load & 8191](453376)|0); //@line 2777 "libraries/glad/glad.cpp"
  HEAP32[164841] = $606; //@line 2777 "libraries/glad/glad.cpp"
  $607 = (FUNCTION_TABLE_ii[$load & 8191](616565)|0); //@line 2778 "libraries/glad/glad.cpp"
  HEAP32[165163] = $607; //@line 2778 "libraries/glad/glad.cpp"
  $608 = (FUNCTION_TABLE_ii[$load & 8191](614909)|0); //@line 2779 "libraries/glad/glad.cpp"
  HEAP32[165284] = $608; //@line 2779 "libraries/glad/glad.cpp"
  $609 = (FUNCTION_TABLE_ii[$load & 8191](453394)|0); //@line 2780 "libraries/glad/glad.cpp"
  HEAP32[165514] = $609; //@line 2780 "libraries/glad/glad.cpp"
  $610 = (FUNCTION_TABLE_ii[$load & 8191](453412)|0); //@line 2781 "libraries/glad/glad.cpp"
  HEAP32[165174] = $610; //@line 2781 "libraries/glad/glad.cpp"
  $611 = (FUNCTION_TABLE_ii[$load & 8191](453429)|0); //@line 2782 "libraries/glad/glad.cpp"
  HEAP32[165600] = $611; //@line 2782 "libraries/glad/glad.cpp"
  $612 = (FUNCTION_TABLE_ii[$load & 8191](453447)|0); //@line 2783 "libraries/glad/glad.cpp"
  HEAP32[164717] = $612; //@line 2783 "libraries/glad/glad.cpp"
  $613 = (FUNCTION_TABLE_ii[$load & 8191](453466)|0); //@line 2784 "libraries/glad/glad.cpp"
  HEAP32[165396] = $613; //@line 2784 "libraries/glad/glad.cpp"
  $614 = (FUNCTION_TABLE_ii[$load & 8191](453485)|0); //@line 2785 "libraries/glad/glad.cpp"
  HEAP32[165476] = $614; //@line 2785 "libraries/glad/glad.cpp"
  $615 = (FUNCTION_TABLE_ii[$load & 8191](615722)|0); //@line 2786 "libraries/glad/glad.cpp"
  HEAP32[164942] = $615; //@line 2786 "libraries/glad/glad.cpp"
 }
 $616 = HEAP8[1252001]|0; //@line 2791 "libraries/glad/glad.cpp"
 $617 = ($616<<24>>24)==(0); //@line 2791 "libraries/glad/glad.cpp"
 if (!($617)) {
  $618 = (FUNCTION_TABLE_ii[$load & 8191](453504)|0); //@line 2792 "libraries/glad/glad.cpp"
  HEAP32[165225] = $618; //@line 2792 "libraries/glad/glad.cpp"
  $619 = (FUNCTION_TABLE_ii[$load & 8191](453525)|0); //@line 2793 "libraries/glad/glad.cpp"
  HEAP32[164692] = $619; //@line 2793 "libraries/glad/glad.cpp"
  $620 = (FUNCTION_TABLE_ii[$load & 8191](453546)|0); //@line 2794 "libraries/glad/glad.cpp"
  HEAP32[165065] = $620; //@line 2794 "libraries/glad/glad.cpp"
  $621 = (FUNCTION_TABLE_ii[$load & 8191](453567)|0); //@line 2795 "libraries/glad/glad.cpp"
  HEAP32[165505] = $621; //@line 2795 "libraries/glad/glad.cpp"
  $622 = (FUNCTION_TABLE_ii[$load & 8191](453588)|0); //@line 2796 "libraries/glad/glad.cpp"
  HEAP32[164743] = $622; //@line 2796 "libraries/glad/glad.cpp"
  $623 = (FUNCTION_TABLE_ii[$load & 8191](453609)|0); //@line 2797 "libraries/glad/glad.cpp"
  HEAP32[164705] = $623; //@line 2797 "libraries/glad/glad.cpp"
 }
 $624 = HEAP8[1252002]|0; //@line 2802 "libraries/glad/glad.cpp"
 $625 = ($624<<24>>24)==(0); //@line 2802 "libraries/glad/glad.cpp"
 if (!($625)) {
  $626 = (FUNCTION_TABLE_ii[$load & 8191](453630)|0); //@line 2803 "libraries/glad/glad.cpp"
  HEAP32[164721] = $626; //@line 2803 "libraries/glad/glad.cpp"
  $627 = (FUNCTION_TABLE_ii[$load & 8191](453643)|0); //@line 2804 "libraries/glad/glad.cpp"
  HEAP32[165658] = $627; //@line 2804 "libraries/glad/glad.cpp"
  $628 = (FUNCTION_TABLE_ii[$load & 8191](453659)|0); //@line 2805 "libraries/glad/glad.cpp"
  HEAP32[164643] = $628; //@line 2805 "libraries/glad/glad.cpp"
  $629 = (FUNCTION_TABLE_ii[$load & 8191](453675)|0); //@line 2806 "libraries/glad/glad.cpp"
  HEAP32[164666] = $629; //@line 2806 "libraries/glad/glad.cpp"
  $630 = (FUNCTION_TABLE_ii[$load & 8191](453685)|0); //@line 2807 "libraries/glad/glad.cpp"
  HEAP32[165534] = $630; //@line 2807 "libraries/glad/glad.cpp"
  $631 = (FUNCTION_TABLE_ii[$load & 8191](453696)|0); //@line 2808 "libraries/glad/glad.cpp"
  HEAP32[164970] = $631; //@line 2808 "libraries/glad/glad.cpp"
  $632 = (FUNCTION_TABLE_ii[$load & 8191](453709)|0); //@line 2809 "libraries/glad/glad.cpp"
  HEAP32[165485] = $632; //@line 2809 "libraries/glad/glad.cpp"
  $633 = (FUNCTION_TABLE_ii[$load & 8191](453734)|0); //@line 2810 "libraries/glad/glad.cpp"
  HEAP32[165475] = $633; //@line 2810 "libraries/glad/glad.cpp"
  $634 = (FUNCTION_TABLE_ii[$load & 8191](453757)|0); //@line 2811 "libraries/glad/glad.cpp"
  HEAP32[165430] = $634; //@line 2811 "libraries/glad/glad.cpp"
  $635 = (FUNCTION_TABLE_ii[$load & 8191](453775)|0); //@line 2812 "libraries/glad/glad.cpp"
  HEAP32[165512] = $635; //@line 2812 "libraries/glad/glad.cpp"
  $636 = (FUNCTION_TABLE_ii[$load & 8191](453792)|0); //@line 2813 "libraries/glad/glad.cpp"
  HEAP32[165096] = $636; //@line 2813 "libraries/glad/glad.cpp"
  $637 = (FUNCTION_TABLE_ii[$load & 8191](453820)|0); //@line 2814 "libraries/glad/glad.cpp"
  HEAP32[164647] = $637; //@line 2814 "libraries/glad/glad.cpp"
  $638 = (FUNCTION_TABLE_ii[$load & 8191](453850)|0); //@line 2815 "libraries/glad/glad.cpp"
  HEAP32[164914] = $638; //@line 2815 "libraries/glad/glad.cpp"
  $639 = (FUNCTION_TABLE_ii[$load & 8191](453863)|0); //@line 2816 "libraries/glad/glad.cpp"
  HEAP32[164898] = $639; //@line 2816 "libraries/glad/glad.cpp"
  $640 = (FUNCTION_TABLE_ii[$load & 8191](453888)|0); //@line 2817 "libraries/glad/glad.cpp"
  HEAP32[165502] = $640; //@line 2817 "libraries/glad/glad.cpp"
  $641 = (FUNCTION_TABLE_ii[$load & 8191](453911)|0); //@line 2818 "libraries/glad/glad.cpp"
  HEAP32[165363] = $641; //@line 2818 "libraries/glad/glad.cpp"
  $642 = (FUNCTION_TABLE_ii[$load & 8191](453934)|0); //@line 2819 "libraries/glad/glad.cpp"
  HEAP32[165591] = $642; //@line 2819 "libraries/glad/glad.cpp"
  $643 = (FUNCTION_TABLE_ii[$load & 8191](453955)|0); //@line 2820 "libraries/glad/glad.cpp"
  HEAP32[165438] = $643; //@line 2820 "libraries/glad/glad.cpp"
  $644 = (FUNCTION_TABLE_ii[$load & 8191](453977)|0); //@line 2821 "libraries/glad/glad.cpp"
  HEAP32[165169] = $644; //@line 2821 "libraries/glad/glad.cpp"
  $645 = (FUNCTION_TABLE_ii[$load & 8191](453995)|0); //@line 2822 "libraries/glad/glad.cpp"
  HEAP32[164973] = $645; //@line 2822 "libraries/glad/glad.cpp"
  $646 = (FUNCTION_TABLE_ii[$load & 8191](454013)|0); //@line 2823 "libraries/glad/glad.cpp"
  HEAP32[165101] = $646; //@line 2823 "libraries/glad/glad.cpp"
  $647 = (FUNCTION_TABLE_ii[$load & 8191](454031)|0); //@line 2824 "libraries/glad/glad.cpp"
  HEAP32[165348] = $647; //@line 2824 "libraries/glad/glad.cpp"
  $648 = (FUNCTION_TABLE_ii[$load & 8191](454049)|0); //@line 2825 "libraries/glad/glad.cpp"
  HEAP32[165356] = $648; //@line 2825 "libraries/glad/glad.cpp"
  $649 = (FUNCTION_TABLE_ii[$load & 8191](454068)|0); //@line 2826 "libraries/glad/glad.cpp"
  HEAP32[165154] = $649; //@line 2826 "libraries/glad/glad.cpp"
  $650 = (FUNCTION_TABLE_ii[$load & 8191](454087)|0); //@line 2827 "libraries/glad/glad.cpp"
  HEAP32[164620] = $650; //@line 2827 "libraries/glad/glad.cpp"
  $651 = (FUNCTION_TABLE_ii[$load & 8191](454106)|0); //@line 2828 "libraries/glad/glad.cpp"
  HEAP32[165463] = $651; //@line 2828 "libraries/glad/glad.cpp"
  $652 = (FUNCTION_TABLE_ii[$load & 8191](454125)|0); //@line 2829 "libraries/glad/glad.cpp"
  HEAP32[165487] = $652; //@line 2829 "libraries/glad/glad.cpp"
  $653 = (FUNCTION_TABLE_ii[$load & 8191](454144)|0); //@line 2830 "libraries/glad/glad.cpp"
  HEAP32[165529] = $653; //@line 2830 "libraries/glad/glad.cpp"
  $654 = (FUNCTION_TABLE_ii[$load & 8191](454163)|0); //@line 2831 "libraries/glad/glad.cpp"
  HEAP32[165295] = $654; //@line 2831 "libraries/glad/glad.cpp"
  $655 = (FUNCTION_TABLE_ii[$load & 8191](454182)|0); //@line 2832 "libraries/glad/glad.cpp"
  HEAP32[165337] = $655; //@line 2832 "libraries/glad/glad.cpp"
  $656 = (FUNCTION_TABLE_ii[$load & 8191](454201)|0); //@line 2833 "libraries/glad/glad.cpp"
  HEAP32[165638] = $656; //@line 2833 "libraries/glad/glad.cpp"
  $657 = (FUNCTION_TABLE_ii[$load & 8191](454221)|0); //@line 2834 "libraries/glad/glad.cpp"
  HEAP32[165409] = $657; //@line 2834 "libraries/glad/glad.cpp"
  $658 = (FUNCTION_TABLE_ii[$load & 8191](454241)|0); //@line 2835 "libraries/glad/glad.cpp"
  HEAP32[165542] = $658; //@line 2835 "libraries/glad/glad.cpp"
  $659 = (FUNCTION_TABLE_ii[$load & 8191](454261)|0); //@line 2836 "libraries/glad/glad.cpp"
  HEAP32[165415] = $659; //@line 2836 "libraries/glad/glad.cpp"
  $660 = (FUNCTION_TABLE_ii[$load & 8191](454281)|0); //@line 2837 "libraries/glad/glad.cpp"
  HEAP32[165454] = $660; //@line 2837 "libraries/glad/glad.cpp"
  $661 = (FUNCTION_TABLE_ii[$load & 8191](454300)|0); //@line 2838 "libraries/glad/glad.cpp"
  HEAP32[165240] = $661; //@line 2838 "libraries/glad/glad.cpp"
  $662 = (FUNCTION_TABLE_ii[$load & 8191](454319)|0); //@line 2839 "libraries/glad/glad.cpp"
  HEAP32[164827] = $662; //@line 2839 "libraries/glad/glad.cpp"
  $663 = (FUNCTION_TABLE_ii[$load & 8191](454339)|0); //@line 2840 "libraries/glad/glad.cpp"
  HEAP32[165105] = $663; //@line 2840 "libraries/glad/glad.cpp"
  $664 = (FUNCTION_TABLE_ii[$load & 8191](454359)|0); //@line 2841 "libraries/glad/glad.cpp"
  HEAP32[164860] = $664; //@line 2841 "libraries/glad/glad.cpp"
  $665 = (FUNCTION_TABLE_ii[$load & 8191](454375)|0); //@line 2842 "libraries/glad/glad.cpp"
  HEAP32[164997] = $665; //@line 2842 "libraries/glad/glad.cpp"
  $666 = (FUNCTION_TABLE_ii[$load & 8191](454398)|0); //@line 2843 "libraries/glad/glad.cpp"
  HEAP32[164783] = $666; //@line 2843 "libraries/glad/glad.cpp"
  $667 = (FUNCTION_TABLE_ii[$load & 8191](454420)|0); //@line 2844 "libraries/glad/glad.cpp"
  HEAP32[164760] = $667; //@line 2844 "libraries/glad/glad.cpp"
  $668 = (FUNCTION_TABLE_ii[$load & 8191](454433)|0); //@line 2845 "libraries/glad/glad.cpp"
  HEAP32[165528] = $668; //@line 2845 "libraries/glad/glad.cpp"
  $669 = (FUNCTION_TABLE_ii[$load & 8191](454446)|0); //@line 2846 "libraries/glad/glad.cpp"
  HEAP32[165412] = $669; //@line 2846 "libraries/glad/glad.cpp"
  $670 = (FUNCTION_TABLE_ii[$load & 8191](454459)|0); //@line 2847 "libraries/glad/glad.cpp"
  HEAP32[165199] = $670; //@line 2847 "libraries/glad/glad.cpp"
  $671 = (FUNCTION_TABLE_ii[$load & 8191](454472)|0); //@line 2848 "libraries/glad/glad.cpp"
  HEAP32[165095] = $671; //@line 2848 "libraries/glad/glad.cpp"
  $672 = (FUNCTION_TABLE_ii[$load & 8191](454486)|0); //@line 2849 "libraries/glad/glad.cpp"
  HEAP32[165080] = $672; //@line 2849 "libraries/glad/glad.cpp"
  $673 = (FUNCTION_TABLE_ii[$load & 8191](454500)|0); //@line 2850 "libraries/glad/glad.cpp"
  HEAP32[165117] = $673; //@line 2850 "libraries/glad/glad.cpp"
  $674 = (FUNCTION_TABLE_ii[$load & 8191](454514)|0); //@line 2851 "libraries/glad/glad.cpp"
  HEAP32[164980] = $674; //@line 2851 "libraries/glad/glad.cpp"
  $675 = (FUNCTION_TABLE_ii[$load & 8191](454528)|0); //@line 2852 "libraries/glad/glad.cpp"
  HEAP32[165468] = $675; //@line 2852 "libraries/glad/glad.cpp"
  $676 = (FUNCTION_TABLE_ii[$load & 8191](454546)|0); //@line 2853 "libraries/glad/glad.cpp"
  HEAP32[165247] = $676; //@line 2853 "libraries/glad/glad.cpp"
  $677 = (FUNCTION_TABLE_ii[$load & 8191](454565)|0); //@line 2854 "libraries/glad/glad.cpp"
  HEAP32[165567] = $677; //@line 2854 "libraries/glad/glad.cpp"
  $678 = (FUNCTION_TABLE_ii[$load & 8191](454586)|0); //@line 2855 "libraries/glad/glad.cpp"
  HEAP32[164926] = $678; //@line 2855 "libraries/glad/glad.cpp"
  $679 = (FUNCTION_TABLE_ii[$load & 8191](454608)|0); //@line 2856 "libraries/glad/glad.cpp"
  HEAP32[165400] = $679; //@line 2856 "libraries/glad/glad.cpp"
  $680 = (FUNCTION_TABLE_ii[$load & 8191](454624)|0); //@line 2857 "libraries/glad/glad.cpp"
  HEAP32[165661] = $680; //@line 2857 "libraries/glad/glad.cpp"
  $681 = (FUNCTION_TABLE_ii[$load & 8191](454641)|0); //@line 2858 "libraries/glad/glad.cpp"
  HEAP32[164707] = $681; //@line 2858 "libraries/glad/glad.cpp"
  $682 = (FUNCTION_TABLE_ii[$load & 8191](454657)|0); //@line 2859 "libraries/glad/glad.cpp"
  HEAP32[164722] = $682; //@line 2859 "libraries/glad/glad.cpp"
  $683 = (FUNCTION_TABLE_ii[$load & 8191](454673)|0); //@line 2860 "libraries/glad/glad.cpp"
  HEAP32[165216] = $683; //@line 2860 "libraries/glad/glad.cpp"
  $684 = (FUNCTION_TABLE_ii[$load & 8191](614432)|0); //@line 2861 "libraries/glad/glad.cpp"
  HEAP32[164804] = $684; //@line 2861 "libraries/glad/glad.cpp"
  $685 = (FUNCTION_TABLE_ii[$load & 8191](614384)|0); //@line 2862 "libraries/glad/glad.cpp"
  HEAP32[165274] = $685; //@line 2862 "libraries/glad/glad.cpp"
  $686 = (FUNCTION_TABLE_ii[$load & 8191](614362)|0); //@line 2863 "libraries/glad/glad.cpp"
  HEAP32[164936] = $686; //@line 2863 "libraries/glad/glad.cpp"
  $687 = (FUNCTION_TABLE_ii[$load & 8191](614343)|0); //@line 2864 "libraries/glad/glad.cpp"
  HEAP32[164676] = $687; //@line 2864 "libraries/glad/glad.cpp"
  $688 = (FUNCTION_TABLE_ii[$load & 8191](616468)|0); //@line 2865 "libraries/glad/glad.cpp"
  HEAP32[164704] = $688; //@line 2865 "libraries/glad/glad.cpp"
  $689 = (FUNCTION_TABLE_ii[$load & 8191](614403)|0); //@line 2866 "libraries/glad/glad.cpp"
  HEAP32[164884] = $689; //@line 2866 "libraries/glad/glad.cpp"
  $690 = (FUNCTION_TABLE_ii[$load & 8191](615381)|0); //@line 2867 "libraries/glad/glad.cpp"
  HEAP32[165368] = $690; //@line 2867 "libraries/glad/glad.cpp"
  $691 = (FUNCTION_TABLE_ii[$load & 8191](615237)|0); //@line 2868 "libraries/glad/glad.cpp"
  HEAP32[165202] = $691; //@line 2868 "libraries/glad/glad.cpp"
  $692 = (FUNCTION_TABLE_ii[$load & 8191](615273)|0); //@line 2869 "libraries/glad/glad.cpp"
  HEAP32[164758] = $692; //@line 2869 "libraries/glad/glad.cpp"
  $693 = (FUNCTION_TABLE_ii[$load & 8191](615255)|0); //@line 2870 "libraries/glad/glad.cpp"
  HEAP32[164797] = $693; //@line 2870 "libraries/glad/glad.cpp"
  $694 = (FUNCTION_TABLE_ii[$load & 8191](616102)|0); //@line 2871 "libraries/glad/glad.cpp"
  HEAP32[165498] = $694; //@line 2871 "libraries/glad/glad.cpp"
  $695 = (FUNCTION_TABLE_ii[$load & 8191](454686)|0); //@line 2872 "libraries/glad/glad.cpp"
  HEAP32[164988] = $695; //@line 2872 "libraries/glad/glad.cpp"
  $696 = (FUNCTION_TABLE_ii[$load & 8191](615320)|0); //@line 2873 "libraries/glad/glad.cpp"
  HEAP32[164777] = $696; //@line 2873 "libraries/glad/glad.cpp"
  $697 = (FUNCTION_TABLE_ii[$load & 8191](454709)|0); //@line 2874 "libraries/glad/glad.cpp"
  HEAP32[165234] = $697; //@line 2874 "libraries/glad/glad.cpp"
  $698 = (FUNCTION_TABLE_ii[$load & 8191](615294)|0); //@line 2875 "libraries/glad/glad.cpp"
  HEAP32[164630] = $698; //@line 2875 "libraries/glad/glad.cpp"
  $699 = (FUNCTION_TABLE_ii[$load & 8191](615343)|0); //@line 2876 "libraries/glad/glad.cpp"
  HEAP32[164778] = $699; //@line 2876 "libraries/glad/glad.cpp"
  $700 = (FUNCTION_TABLE_ii[$load & 8191](616127)|0); //@line 2877 "libraries/glad/glad.cpp"
  HEAP32[165249] = $700; //@line 2877 "libraries/glad/glad.cpp"
  $701 = (FUNCTION_TABLE_ii[$load & 8191](454732)|0); //@line 2878 "libraries/glad/glad.cpp"
  HEAP32[164725] = $701; //@line 2878 "libraries/glad/glad.cpp"
  $702 = (FUNCTION_TABLE_ii[$load & 8191](454750)|0); //@line 2879 "libraries/glad/glad.cpp"
  HEAP32[165298] = $702; //@line 2879 "libraries/glad/glad.cpp"
  $703 = (FUNCTION_TABLE_ii[$load & 8191](454783)|0); //@line 2880 "libraries/glad/glad.cpp"
  HEAP32[164944] = $703; //@line 2880 "libraries/glad/glad.cpp"
  $704 = (FUNCTION_TABLE_ii[$load & 8191](454809)|0); //@line 2881 "libraries/glad/glad.cpp"
  HEAP32[165517] = $704; //@line 2881 "libraries/glad/glad.cpp"
  $705 = (FUNCTION_TABLE_ii[$load & 8191](454826)|0); //@line 2882 "libraries/glad/glad.cpp"
  HEAP32[164950] = $705; //@line 2882 "libraries/glad/glad.cpp"
  $706 = (FUNCTION_TABLE_ii[$load & 8191](615642)|0); //@line 2883 "libraries/glad/glad.cpp"
  HEAP32[165066] = $706; //@line 2883 "libraries/glad/glad.cpp"
  $707 = (FUNCTION_TABLE_ii[$load & 8191](615621)|0); //@line 2884 "libraries/glad/glad.cpp"
  HEAP32[164893] = $707; //@line 2884 "libraries/glad/glad.cpp"
  $708 = (FUNCTION_TABLE_ii[$load & 8191](615603)|0); //@line 2885 "libraries/glad/glad.cpp"
  HEAP32[165300] = $708; //@line 2885 "libraries/glad/glad.cpp"
  $709 = (FUNCTION_TABLE_ii[$load & 8191](454851)|0); //@line 2886 "libraries/glad/glad.cpp"
  HEAP32[164810] = $709; //@line 2886 "libraries/glad/glad.cpp"
 }
 $710 = HEAP8[1252003]|0; //@line 2891 "libraries/glad/glad.cpp"
 $711 = ($710<<24>>24)==(0); //@line 2891 "libraries/glad/glad.cpp"
 if (!($711)) {
  $712 = (FUNCTION_TABLE_ii[$load & 8191](615897)|0); //@line 2892 "libraries/glad/glad.cpp"
  HEAP32[165248] = $712; //@line 2892 "libraries/glad/glad.cpp"
  $713 = (FUNCTION_TABLE_ii[$load & 8191](615919)|0); //@line 2893 "libraries/glad/glad.cpp"
  HEAP32[164873] = $713; //@line 2893 "libraries/glad/glad.cpp"
  $714 = (FUNCTION_TABLE_ii[$load & 8191](454867)|0); //@line 2894 "libraries/glad/glad.cpp"
  HEAP32[165113] = $714; //@line 2894 "libraries/glad/glad.cpp"
  $715 = (FUNCTION_TABLE_ii[$load & 8191](454879)|0); //@line 2895 "libraries/glad/glad.cpp"
  HEAP32[165373] = $715; //@line 2895 "libraries/glad/glad.cpp"
  $716 = (FUNCTION_TABLE_ii[$load & 8191](454903)|0); //@line 2896 "libraries/glad/glad.cpp"
  HEAP32[165637] = $716; //@line 2896 "libraries/glad/glad.cpp"
  $717 = (FUNCTION_TABLE_ii[$load & 8191](454923)|0); //@line 2897 "libraries/glad/glad.cpp"
  HEAP32[164824] = $717; //@line 2897 "libraries/glad/glad.cpp"
  $718 = (FUNCTION_TABLE_ii[$load & 8191](454943)|0); //@line 2898 "libraries/glad/glad.cpp"
  HEAP32[164833] = $718; //@line 2898 "libraries/glad/glad.cpp"
  $719 = (FUNCTION_TABLE_ii[$load & 8191](454965)|0); //@line 2899 "libraries/glad/glad.cpp"
  HEAP32[164770] = $719; //@line 2899 "libraries/glad/glad.cpp"
  $720 = (FUNCTION_TABLE_ii[$load & 8191](454988)|0); //@line 2900 "libraries/glad/glad.cpp"
  HEAP32[165556] = $720; //@line 2900 "libraries/glad/glad.cpp"
  $721 = (FUNCTION_TABLE_ii[$load & 8191](455011)|0); //@line 2901 "libraries/glad/glad.cpp"
  HEAP32[165048] = $721; //@line 2901 "libraries/glad/glad.cpp"
  $722 = (FUNCTION_TABLE_ii[$load & 8191](455037)|0); //@line 2902 "libraries/glad/glad.cpp"
  HEAP32[165541] = $722; //@line 2902 "libraries/glad/glad.cpp"
  $723 = (FUNCTION_TABLE_ii[$load & 8191](455065)|0); //@line 2903 "libraries/glad/glad.cpp"
  HEAP32[165325] = $723; //@line 2903 "libraries/glad/glad.cpp"
  $724 = (FUNCTION_TABLE_ii[$load & 8191](453757)|0); //@line 2904 "libraries/glad/glad.cpp"
  HEAP32[165430] = $724; //@line 2904 "libraries/glad/glad.cpp"
  $725 = (FUNCTION_TABLE_ii[$load & 8191](453775)|0); //@line 2905 "libraries/glad/glad.cpp"
  HEAP32[165512] = $725; //@line 2905 "libraries/glad/glad.cpp"
  $726 = (FUNCTION_TABLE_ii[$load & 8191](453659)|0); //@line 2906 "libraries/glad/glad.cpp"
  HEAP32[164643] = $726; //@line 2906 "libraries/glad/glad.cpp"
 }
 $727 = HEAP8[1252004]|0; //@line 2911 "libraries/glad/glad.cpp"
 $728 = ($727<<24>>24)==(0); //@line 2911 "libraries/glad/glad.cpp"
 if (!($728)) {
  $729 = (FUNCTION_TABLE_ii[$load & 8191](455087)|0); //@line 2912 "libraries/glad/glad.cpp"
  HEAP32[164902] = $729; //@line 2912 "libraries/glad/glad.cpp"
  $730 = (FUNCTION_TABLE_ii[$load & 8191](455112)|0); //@line 2913 "libraries/glad/glad.cpp"
  HEAP32[164690] = $730; //@line 2913 "libraries/glad/glad.cpp"
  $731 = (FUNCTION_TABLE_ii[$load & 8191](455142)|0); //@line 2914 "libraries/glad/glad.cpp"
  HEAP32[165310] = $731; //@line 2914 "libraries/glad/glad.cpp"
  $732 = (FUNCTION_TABLE_ii[$load & 8191](455176)|0); //@line 2915 "libraries/glad/glad.cpp"
  HEAP32[165073] = $732; //@line 2915 "libraries/glad/glad.cpp"
  $733 = (FUNCTION_TABLE_ii[$load & 8191](455206)|0); //@line 2916 "libraries/glad/glad.cpp"
  HEAP32[165230] = $733; //@line 2916 "libraries/glad/glad.cpp"
  $734 = (FUNCTION_TABLE_ii[$load & 8191](455224)|0); //@line 2917 "libraries/glad/glad.cpp"
  HEAP32[164892] = $734; //@line 2917 "libraries/glad/glad.cpp"
  $735 = (FUNCTION_TABLE_ii[$load & 8191](455236)|0); //@line 2918 "libraries/glad/glad.cpp"
  HEAP32[164933] = $735; //@line 2918 "libraries/glad/glad.cpp"
  $736 = (FUNCTION_TABLE_ii[$load & 8191](455245)|0); //@line 2919 "libraries/glad/glad.cpp"
  HEAP32[165033] = $736; //@line 2919 "libraries/glad/glad.cpp"
  $737 = (FUNCTION_TABLE_ii[$load & 8191](455258)|0); //@line 2920 "libraries/glad/glad.cpp"
  HEAP32[165462] = $737; //@line 2920 "libraries/glad/glad.cpp"
  $738 = (FUNCTION_TABLE_ii[$load & 8191](455275)|0); //@line 2921 "libraries/glad/glad.cpp"
  HEAP32[165171] = $738; //@line 2921 "libraries/glad/glad.cpp"
  $739 = (FUNCTION_TABLE_ii[$load & 8191](455286)|0); //@line 2922 "libraries/glad/glad.cpp"
  HEAP32[164829] = $739; //@line 2922 "libraries/glad/glad.cpp"
  $740 = (FUNCTION_TABLE_ii[$load & 8191](455302)|0); //@line 2923 "libraries/glad/glad.cpp"
  HEAP32[165547] = $740; //@line 2923 "libraries/glad/glad.cpp"
  $741 = (FUNCTION_TABLE_ii[$load & 8191](455314)|0); //@line 2924 "libraries/glad/glad.cpp"
  HEAP32[164964] = $741; //@line 2924 "libraries/glad/glad.cpp"
  $742 = (FUNCTION_TABLE_ii[$load & 8191](455332)|0); //@line 2925 "libraries/glad/glad.cpp"
  HEAP32[165167] = $742; //@line 2925 "libraries/glad/glad.cpp"
  $743 = (FUNCTION_TABLE_ii[$load & 8191](455357)|0); //@line 2926 "libraries/glad/glad.cpp"
  HEAP32[165518] = $743; //@line 2926 "libraries/glad/glad.cpp"
  $744 = (FUNCTION_TABLE_ii[$load & 8191](455378)|0); //@line 2927 "libraries/glad/glad.cpp"
  HEAP32[165037] = $744; //@line 2927 "libraries/glad/glad.cpp"
  $745 = (FUNCTION_TABLE_ii[$load & 8191](455402)|0); //@line 2928 "libraries/glad/glad.cpp"
  HEAP32[165647] = $745; //@line 2928 "libraries/glad/glad.cpp"
  $746 = (FUNCTION_TABLE_ii[$load & 8191](455426)|0); //@line 2929 "libraries/glad/glad.cpp"
  HEAP32[164674] = $746; //@line 2929 "libraries/glad/glad.cpp"
  $747 = (FUNCTION_TABLE_ii[$load & 8191](455445)|0); //@line 2930 "libraries/glad/glad.cpp"
  HEAP32[164688] = $747; //@line 2930 "libraries/glad/glad.cpp"
 }
 $748 = HEAP8[1252005]|0; //@line 2935 "libraries/glad/glad.cpp"
 $749 = ($748<<24>>24)==(0); //@line 2935 "libraries/glad/glad.cpp"
 if (!($749)) {
  $750 = (FUNCTION_TABLE_ii[$load & 8191](455459)|0); //@line 2936 "libraries/glad/glad.cpp"
  HEAP32[164974] = $750; //@line 2936 "libraries/glad/glad.cpp"
  $751 = (FUNCTION_TABLE_ii[$load & 8191](455489)|0); //@line 2937 "libraries/glad/glad.cpp"
  HEAP32[164733] = $751; //@line 2937 "libraries/glad/glad.cpp"
  $752 = (FUNCTION_TABLE_ii[$load & 8191](455508)|0); //@line 2938 "libraries/glad/glad.cpp"
  HEAP32[164913] = $752; //@line 2938 "libraries/glad/glad.cpp"
  $753 = (FUNCTION_TABLE_ii[$load & 8191](455522)|0); //@line 2939 "libraries/glad/glad.cpp"
  HEAP32[164709] = $753; //@line 2939 "libraries/glad/glad.cpp"
  $754 = (FUNCTION_TABLE_ii[$load & 8191](455539)|0); //@line 2940 "libraries/glad/glad.cpp"
  HEAP32[165489] = $754; //@line 2940 "libraries/glad/glad.cpp"
  $755 = (FUNCTION_TABLE_ii[$load & 8191](455551)|0); //@line 2941 "libraries/glad/glad.cpp"
  HEAP32[164640] = $755; //@line 2941 "libraries/glad/glad.cpp"
  $756 = (FUNCTION_TABLE_ii[$load & 8191](455565)|0); //@line 2942 "libraries/glad/glad.cpp"
  HEAP32[165177] = $756; //@line 2942 "libraries/glad/glad.cpp"
  $757 = (FUNCTION_TABLE_ii[$load & 8191](455585)|0); //@line 2943 "libraries/glad/glad.cpp"
  HEAP32[165255] = $757; //@line 2943 "libraries/glad/glad.cpp"
  $758 = (FUNCTION_TABLE_ii[$load & 8191](455606)|0); //@line 2944 "libraries/glad/glad.cpp"
  HEAP32[165179] = $758; //@line 2944 "libraries/glad/glad.cpp"
  $759 = (FUNCTION_TABLE_ii[$load & 8191](455626)|0); //@line 2945 "libraries/glad/glad.cpp"
  HEAP32[165305] = $759; //@line 2945 "libraries/glad/glad.cpp"
  $760 = (FUNCTION_TABLE_ii[$load & 8191](455647)|0); //@line 2946 "libraries/glad/glad.cpp"
  HEAP32[165072] = $760; //@line 2946 "libraries/glad/glad.cpp"
  $761 = (FUNCTION_TABLE_ii[$load & 8191](455669)|0); //@line 2947 "libraries/glad/glad.cpp"
  HEAP32[165411] = $761; //@line 2947 "libraries/glad/glad.cpp"
  $762 = (FUNCTION_TABLE_ii[$load & 8191](455692)|0); //@line 2948 "libraries/glad/glad.cpp"
  HEAP32[165471] = $762; //@line 2948 "libraries/glad/glad.cpp"
  $763 = (FUNCTION_TABLE_ii[$load & 8191](455716)|0); //@line 2949 "libraries/glad/glad.cpp"
  HEAP32[164732] = $763; //@line 2949 "libraries/glad/glad.cpp"
  $764 = (FUNCTION_TABLE_ii[$load & 8191](455741)|0); //@line 2950 "libraries/glad/glad.cpp"
  HEAP32[164818] = $764; //@line 2950 "libraries/glad/glad.cpp"
  $765 = (FUNCTION_TABLE_ii[$load & 8191](455765)|0); //@line 2951 "libraries/glad/glad.cpp"
  HEAP32[165429] = $765; //@line 2951 "libraries/glad/glad.cpp"
  $766 = (FUNCTION_TABLE_ii[$load & 8191](455791)|0); //@line 2952 "libraries/glad/glad.cpp"
  HEAP32[165227] = $766; //@line 2952 "libraries/glad/glad.cpp"
  $767 = (FUNCTION_TABLE_ii[$load & 8191](455806)|0); //@line 2953 "libraries/glad/glad.cpp"
  HEAP32[164796] = $767; //@line 2953 "libraries/glad/glad.cpp"
  $768 = (FUNCTION_TABLE_ii[$load & 8191](455827)|0); //@line 2954 "libraries/glad/glad.cpp"
  HEAP32[165643] = $768; //@line 2954 "libraries/glad/glad.cpp"
  $769 = (FUNCTION_TABLE_ii[$load & 8191](615875)|0); //@line 2955 "libraries/glad/glad.cpp"
  HEAP32[165491] = $769; //@line 2955 "libraries/glad/glad.cpp"
  $770 = (FUNCTION_TABLE_ii[$load & 8191](455849)|0); //@line 2956 "libraries/glad/glad.cpp"
  HEAP32[164962] = $770; //@line 2956 "libraries/glad/glad.cpp"
  $771 = (FUNCTION_TABLE_ii[$load & 8191](455868)|0); //@line 2957 "libraries/glad/glad.cpp"
  HEAP32[165132] = $771; //@line 2957 "libraries/glad/glad.cpp"
  $772 = (FUNCTION_TABLE_ii[$load & 8191](455888)|0); //@line 2958 "libraries/glad/glad.cpp"
  HEAP32[165574] = $772; //@line 2958 "libraries/glad/glad.cpp"
  $773 = (FUNCTION_TABLE_ii[$load & 8191](455907)|0); //@line 2959 "libraries/glad/glad.cpp"
  HEAP32[165141] = $773; //@line 2959 "libraries/glad/glad.cpp"
  $774 = (FUNCTION_TABLE_ii[$load & 8191](455927)|0); //@line 2960 "libraries/glad/glad.cpp"
  HEAP32[164850] = $774; //@line 2960 "libraries/glad/glad.cpp"
  $775 = (FUNCTION_TABLE_ii[$load & 8191](455946)|0); //@line 2961 "libraries/glad/glad.cpp"
  HEAP32[165584] = $775; //@line 2961 "libraries/glad/glad.cpp"
  $776 = (FUNCTION_TABLE_ii[$load & 8191](455966)|0); //@line 2962 "libraries/glad/glad.cpp"
  HEAP32[164671] = $776; //@line 2962 "libraries/glad/glad.cpp"
  $777 = (FUNCTION_TABLE_ii[$load & 8191](455985)|0); //@line 2963 "libraries/glad/glad.cpp"
  HEAP32[165278] = $777; //@line 2963 "libraries/glad/glad.cpp"
  $778 = (FUNCTION_TABLE_ii[$load & 8191](456005)|0); //@line 2964 "libraries/glad/glad.cpp"
  HEAP32[164689] = $778; //@line 2964 "libraries/glad/glad.cpp"
  $779 = (FUNCTION_TABLE_ii[$load & 8191](456018)|0); //@line 2965 "libraries/glad/glad.cpp"
  HEAP32[165391] = $779; //@line 2965 "libraries/glad/glad.cpp"
  $780 = (FUNCTION_TABLE_ii[$load & 8191](456032)|0); //@line 2966 "libraries/glad/glad.cpp"
  HEAP32[165490] = $780; //@line 2966 "libraries/glad/glad.cpp"
  $781 = (FUNCTION_TABLE_ii[$load & 8191](456045)|0); //@line 2967 "libraries/glad/glad.cpp"
  HEAP32[164787] = $781; //@line 2967 "libraries/glad/glad.cpp"
  $782 = (FUNCTION_TABLE_ii[$load & 8191](456059)|0); //@line 2968 "libraries/glad/glad.cpp"
  HEAP32[164664] = $782; //@line 2968 "libraries/glad/glad.cpp"
  $783 = (FUNCTION_TABLE_ii[$load & 8191](456072)|0); //@line 2969 "libraries/glad/glad.cpp"
  HEAP32[165527] = $783; //@line 2969 "libraries/glad/glad.cpp"
  $784 = (FUNCTION_TABLE_ii[$load & 8191](456086)|0); //@line 2970 "libraries/glad/glad.cpp"
  HEAP32[165570] = $784; //@line 2970 "libraries/glad/glad.cpp"
  $785 = (FUNCTION_TABLE_ii[$load & 8191](456101)|0); //@line 2971 "libraries/glad/glad.cpp"
  HEAP32[164917] = $785; //@line 2971 "libraries/glad/glad.cpp"
  $786 = (FUNCTION_TABLE_ii[$load & 8191](456117)|0); //@line 2972 "libraries/glad/glad.cpp"
  HEAP32[164720] = $786; //@line 2972 "libraries/glad/glad.cpp"
  $787 = (FUNCTION_TABLE_ii[$load & 8191](456132)|0); //@line 2973 "libraries/glad/glad.cpp"
  HEAP32[165550] = $787; //@line 2973 "libraries/glad/glad.cpp"
  $788 = (FUNCTION_TABLE_ii[$load & 8191](456148)|0); //@line 2974 "libraries/glad/glad.cpp"
  HEAP32[164766] = $788; //@line 2974 "libraries/glad/glad.cpp"
  $789 = (FUNCTION_TABLE_ii[$load & 8191](456163)|0); //@line 2975 "libraries/glad/glad.cpp"
  HEAP32[164746] = $789; //@line 2975 "libraries/glad/glad.cpp"
  $790 = (FUNCTION_TABLE_ii[$load & 8191](456179)|0); //@line 2976 "libraries/glad/glad.cpp"
  HEAP32[164940] = $790; //@line 2976 "libraries/glad/glad.cpp"
  $791 = (FUNCTION_TABLE_ii[$load & 8191](456194)|0); //@line 2977 "libraries/glad/glad.cpp"
  HEAP32[165384] = $791; //@line 2977 "libraries/glad/glad.cpp"
  $792 = (FUNCTION_TABLE_ii[$load & 8191](456210)|0); //@line 2978 "libraries/glad/glad.cpp"
  HEAP32[164854] = $792; //@line 2978 "libraries/glad/glad.cpp"
  $793 = (FUNCTION_TABLE_ii[$load & 8191](456230)|0); //@line 2979 "libraries/glad/glad.cpp"
  HEAP32[165016] = $793; //@line 2979 "libraries/glad/glad.cpp"
  $794 = (FUNCTION_TABLE_ii[$load & 8191](456251)|0); //@line 2980 "libraries/glad/glad.cpp"
  HEAP32[164727] = $794; //@line 2980 "libraries/glad/glad.cpp"
  $795 = (FUNCTION_TABLE_ii[$load & 8191](456271)|0); //@line 2981 "libraries/glad/glad.cpp"
  HEAP32[165121] = $795; //@line 2981 "libraries/glad/glad.cpp"
  $796 = (FUNCTION_TABLE_ii[$load & 8191](456292)|0); //@line 2982 "libraries/glad/glad.cpp"
  HEAP32[164660] = $796; //@line 2982 "libraries/glad/glad.cpp"
  $797 = (FUNCTION_TABLE_ii[$load & 8191](456312)|0); //@line 2983 "libraries/glad/glad.cpp"
  HEAP32[164849] = $797; //@line 2983 "libraries/glad/glad.cpp"
  $798 = (FUNCTION_TABLE_ii[$load & 8191](456333)|0); //@line 2984 "libraries/glad/glad.cpp"
  HEAP32[164751] = $798; //@line 2984 "libraries/glad/glad.cpp"
  $799 = (FUNCTION_TABLE_ii[$load & 8191](456353)|0); //@line 2985 "libraries/glad/glad.cpp"
  HEAP32[164905] = $799; //@line 2985 "libraries/glad/glad.cpp"
  $800 = (FUNCTION_TABLE_ii[$load & 8191](456374)|0); //@line 2986 "libraries/glad/glad.cpp"
  HEAP32[165243] = $800; //@line 2986 "libraries/glad/glad.cpp"
  $801 = (FUNCTION_TABLE_ii[$load & 8191](456387)|0); //@line 2987 "libraries/glad/glad.cpp"
  HEAP32[165162] = $801; //@line 2987 "libraries/glad/glad.cpp"
  $802 = (FUNCTION_TABLE_ii[$load & 8191](456401)|0); //@line 2988 "libraries/glad/glad.cpp"
  HEAP32[164894] = $802; //@line 2988 "libraries/glad/glad.cpp"
  $803 = (FUNCTION_TABLE_ii[$load & 8191](456413)|0); //@line 2989 "libraries/glad/glad.cpp"
  HEAP32[164642] = $803; //@line 2989 "libraries/glad/glad.cpp"
  $804 = (FUNCTION_TABLE_ii[$load & 8191](456426)|0); //@line 2990 "libraries/glad/glad.cpp"
  HEAP32[165328] = $804; //@line 2990 "libraries/glad/glad.cpp"
  $805 = (FUNCTION_TABLE_ii[$load & 8191](456438)|0); //@line 2991 "libraries/glad/glad.cpp"
  HEAP32[165403] = $805; //@line 2991 "libraries/glad/glad.cpp"
  $806 = (FUNCTION_TABLE_ii[$load & 8191](456451)|0); //@line 2992 "libraries/glad/glad.cpp"
  HEAP32[164972] = $806; //@line 2992 "libraries/glad/glad.cpp"
  $807 = (FUNCTION_TABLE_ii[$load & 8191](456472)|0); //@line 2993 "libraries/glad/glad.cpp"
  HEAP32[165586] = $807; //@line 2993 "libraries/glad/glad.cpp"
 }
 $808 = HEAP8[1252006]|0; //@line 2998 "libraries/glad/glad.cpp"
 $809 = ($808<<24>>24)==(0); //@line 2998 "libraries/glad/glad.cpp"
 if (!($809)) {
  $810 = (FUNCTION_TABLE_ii[$load & 8191](456494)|0); //@line 2999 "libraries/glad/glad.cpp"
  HEAP32[164682] = $810; //@line 2999 "libraries/glad/glad.cpp"
  $811 = (FUNCTION_TABLE_ii[$load & 8191](456513)|0); //@line 3000 "libraries/glad/glad.cpp"
  HEAP32[164839] = $811; //@line 3000 "libraries/glad/glad.cpp"
  $812 = (FUNCTION_TABLE_ii[$load & 8191](456530)|0); //@line 3001 "libraries/glad/glad.cpp"
  HEAP32[165455] = $812; //@line 3001 "libraries/glad/glad.cpp"
  $813 = (FUNCTION_TABLE_ii[$load & 8191](456555)|0); //@line 3002 "libraries/glad/glad.cpp"
  HEAP32[165507] = $813; //@line 3002 "libraries/glad/glad.cpp"
  $814 = (FUNCTION_TABLE_ii[$load & 8191](456568)|0); //@line 3003 "libraries/glad/glad.cpp"
  HEAP32[165519] = $814; //@line 3003 "libraries/glad/glad.cpp"
  $815 = (FUNCTION_TABLE_ii[$load & 8191](456589)|0); //@line 3004 "libraries/glad/glad.cpp"
  HEAP32[164723] = $815; //@line 3004 "libraries/glad/glad.cpp"
  $816 = (FUNCTION_TABLE_ii[$load & 8191](456610)|0); //@line 3005 "libraries/glad/glad.cpp"
  HEAP32[165422] = $816; //@line 3005 "libraries/glad/glad.cpp"
  $817 = (FUNCTION_TABLE_ii[$load & 8191](456633)|0); //@line 3006 "libraries/glad/glad.cpp"
  HEAP32[165183] = $817; //@line 3006 "libraries/glad/glad.cpp"
  $818 = (FUNCTION_TABLE_ii[$load & 8191](456645)|0); //@line 3007 "libraries/glad/glad.cpp"
  HEAP32[165561] = $818; //@line 3007 "libraries/glad/glad.cpp"
  $819 = (FUNCTION_TABLE_ii[$load & 8191](456657)|0); //@line 3008 "libraries/glad/glad.cpp"
  HEAP32[164715] = $819; //@line 3008 "libraries/glad/glad.cpp"
  $820 = (FUNCTION_TABLE_ii[$load & 8191](456669)|0); //@line 3009 "libraries/glad/glad.cpp"
  HEAP32[164989] = $820; //@line 3009 "libraries/glad/glad.cpp"
  $821 = (FUNCTION_TABLE_ii[$load & 8191](456681)|0); //@line 3010 "libraries/glad/glad.cpp"
  HEAP32[164925] = $821; //@line 3010 "libraries/glad/glad.cpp"
  $822 = (FUNCTION_TABLE_ii[$load & 8191](456694)|0); //@line 3011 "libraries/glad/glad.cpp"
  HEAP32[164877] = $822; //@line 3011 "libraries/glad/glad.cpp"
  $823 = (FUNCTION_TABLE_ii[$load & 8191](456707)|0); //@line 3012 "libraries/glad/glad.cpp"
  HEAP32[165622] = $823; //@line 3012 "libraries/glad/glad.cpp"
  $824 = (FUNCTION_TABLE_ii[$load & 8191](456720)|0); //@line 3013 "libraries/glad/glad.cpp"
  HEAP32[165007] = $824; //@line 3013 "libraries/glad/glad.cpp"
  $825 = (FUNCTION_TABLE_ii[$load & 8191](456733)|0); //@line 3014 "libraries/glad/glad.cpp"
  HEAP32[165506] = $825; //@line 3014 "libraries/glad/glad.cpp"
  $826 = (FUNCTION_TABLE_ii[$load & 8191](456752)|0); //@line 3015 "libraries/glad/glad.cpp"
  HEAP32[164744] = $826; //@line 3015 "libraries/glad/glad.cpp"
  $827 = (FUNCTION_TABLE_ii[$load & 8191](456771)|0); //@line 3016 "libraries/glad/glad.cpp"
  HEAP32[165036] = $827; //@line 3016 "libraries/glad/glad.cpp"
  $828 = (FUNCTION_TABLE_ii[$load & 8191](456790)|0); //@line 3017 "libraries/glad/glad.cpp"
  HEAP32[164952] = $828; //@line 3017 "libraries/glad/glad.cpp"
  $829 = (FUNCTION_TABLE_ii[$load & 8191](456811)|0); //@line 3018 "libraries/glad/glad.cpp"
  HEAP32[165578] = $829; //@line 3018 "libraries/glad/glad.cpp"
  $830 = (FUNCTION_TABLE_ii[$load & 8191](456832)|0); //@line 3019 "libraries/glad/glad.cpp"
  HEAP32[165283] = $830; //@line 3019 "libraries/glad/glad.cpp"
  $831 = (FUNCTION_TABLE_ii[$load & 8191](456853)|0); //@line 3020 "libraries/glad/glad.cpp"
  HEAP32[165253] = $831; //@line 3020 "libraries/glad/glad.cpp"
  $832 = (FUNCTION_TABLE_ii[$load & 8191](456874)|0); //@line 3021 "libraries/glad/glad.cpp"
  HEAP32[165039] = $832; //@line 3021 "libraries/glad/glad.cpp"
  $833 = (FUNCTION_TABLE_ii[$load & 8191](456895)|0); //@line 3022 "libraries/glad/glad.cpp"
  HEAP32[165151] = $833; //@line 3022 "libraries/glad/glad.cpp"
  $834 = (FUNCTION_TABLE_ii[$load & 8191](456916)|0); //@line 3023 "libraries/glad/glad.cpp"
  HEAP32[165434] = $834; //@line 3023 "libraries/glad/glad.cpp"
  $835 = (FUNCTION_TABLE_ii[$load & 8191](456931)|0); //@line 3024 "libraries/glad/glad.cpp"
  HEAP32[165627] = $835; //@line 3024 "libraries/glad/glad.cpp"
  $836 = (FUNCTION_TABLE_ii[$load & 8191](456962)|0); //@line 3025 "libraries/glad/glad.cpp"
  HEAP32[165359] = $836; //@line 3025 "libraries/glad/glad.cpp"
  $837 = (FUNCTION_TABLE_ii[$load & 8191](456983)|0); //@line 3026 "libraries/glad/glad.cpp"
  HEAP32[165109] = $837; //@line 3026 "libraries/glad/glad.cpp"
  $838 = (FUNCTION_TABLE_ii[$load & 8191](457014)|0); //@line 3027 "libraries/glad/glad.cpp"
  HEAP32[165419] = $838; //@line 3027 "libraries/glad/glad.cpp"
  $839 = (FUNCTION_TABLE_ii[$load & 8191](457047)|0); //@line 3028 "libraries/glad/glad.cpp"
  HEAP32[165318] = $839; //@line 3028 "libraries/glad/glad.cpp"
  $840 = (FUNCTION_TABLE_ii[$load & 8191](457073)|0); //@line 3029 "libraries/glad/glad.cpp"
  HEAP32[164631] = $840; //@line 3029 "libraries/glad/glad.cpp"
  $841 = (FUNCTION_TABLE_ii[$load & 8191](457097)|0); //@line 3030 "libraries/glad/glad.cpp"
  HEAP32[165477] = $841; //@line 3030 "libraries/glad/glad.cpp"
  $842 = (FUNCTION_TABLE_ii[$load & 8191](457123)|0); //@line 3031 "libraries/glad/glad.cpp"
  HEAP32[165279] = $842; //@line 3031 "libraries/glad/glad.cpp"
  $843 = (FUNCTION_TABLE_ii[$load & 8191](457143)|0); //@line 3032 "libraries/glad/glad.cpp"
  HEAP32[165551] = $843; //@line 3032 "libraries/glad/glad.cpp"
  $844 = (FUNCTION_TABLE_ii[$load & 8191](457161)|0); //@line 3033 "libraries/glad/glad.cpp"
  HEAP32[165002] = $844; //@line 3033 "libraries/glad/glad.cpp"
  $845 = (FUNCTION_TABLE_ii[$load & 8191](457180)|0); //@line 3034 "libraries/glad/glad.cpp"
  HEAP32[165640] = $845; //@line 3034 "libraries/glad/glad.cpp"
  $846 = (FUNCTION_TABLE_ii[$load & 8191](457204)|0); //@line 3035 "libraries/glad/glad.cpp"
  HEAP32[165495] = $846; //@line 3035 "libraries/glad/glad.cpp"
  $847 = (FUNCTION_TABLE_ii[$load & 8191](457231)|0); //@line 3036 "libraries/glad/glad.cpp"
  HEAP32[165443] = $847; //@line 3036 "libraries/glad/glad.cpp"
  $848 = (FUNCTION_TABLE_ii[$load & 8191](457255)|0); //@line 3037 "libraries/glad/glad.cpp"
  HEAP32[165286] = $848; //@line 3037 "libraries/glad/glad.cpp"
  $849 = (FUNCTION_TABLE_ii[$load & 8191](457277)|0); //@line 3038 "libraries/glad/glad.cpp"
  HEAP32[164726] = $849; //@line 3038 "libraries/glad/glad.cpp"
  $850 = (FUNCTION_TABLE_ii[$load & 8191](457302)|0); //@line 3039 "libraries/glad/glad.cpp"
  HEAP32[165417] = $850; //@line 3039 "libraries/glad/glad.cpp"
  $851 = (FUNCTION_TABLE_ii[$load & 8191](457328)|0); //@line 3040 "libraries/glad/glad.cpp"
  HEAP32[164924] = $851; //@line 3040 "libraries/glad/glad.cpp"
  $852 = (FUNCTION_TABLE_ii[$load & 8191](457352)|0); //@line 3041 "libraries/glad/glad.cpp"
  HEAP32[165221] = $852; //@line 3041 "libraries/glad/glad.cpp"
  $853 = (FUNCTION_TABLE_ii[$load & 8191](457382)|0); //@line 3042 "libraries/glad/glad.cpp"
  HEAP32[164885] = $853; //@line 3042 "libraries/glad/glad.cpp"
  $854 = (FUNCTION_TABLE_ii[$load & 8191](457402)|0); //@line 3043 "libraries/glad/glad.cpp"
  HEAP32[164985] = $854; //@line 3043 "libraries/glad/glad.cpp"
  $855 = (FUNCTION_TABLE_ii[$load & 8191](457420)|0); //@line 3044 "libraries/glad/glad.cpp"
  HEAP32[165342] = $855; //@line 3044 "libraries/glad/glad.cpp"
 }
 $856 = HEAP8[1252007]|0; //@line 3049 "libraries/glad/glad.cpp"
 $857 = ($856<<24>>24)==(0); //@line 3049 "libraries/glad/glad.cpp"
 if (!($857)) {
  $858 = (FUNCTION_TABLE_ii[$load & 8191](615840)|0); //@line 3050 "libraries/glad/glad.cpp"
  HEAP32[165543] = $858; //@line 3050 "libraries/glad/glad.cpp"
  $859 = (FUNCTION_TABLE_ii[$load & 8191](615825)|0); //@line 3051 "libraries/glad/glad.cpp"
  HEAP32[165231] = $859; //@line 3051 "libraries/glad/glad.cpp"
  $860 = (FUNCTION_TABLE_ii[$load & 8191](615125)|0); //@line 3052 "libraries/glad/glad.cpp"
  HEAP32[164900] = $860; //@line 3052 "libraries/glad/glad.cpp"
  $861 = (FUNCTION_TABLE_ii[$load & 8191](616237)|0); //@line 3053 "libraries/glad/glad.cpp"
  HEAP32[164657] = $861; //@line 3053 "libraries/glad/glad.cpp"
  $862 = (FUNCTION_TABLE_ii[$load & 8191](615973)|0); //@line 3054 "libraries/glad/glad.cpp"
  HEAP32[165437] = $862; //@line 3054 "libraries/glad/glad.cpp"
  $863 = (FUNCTION_TABLE_ii[$load & 8191](457440)|0); //@line 3055 "libraries/glad/glad.cpp"
  HEAP32[165057] = $863; //@line 3055 "libraries/glad/glad.cpp"
  $864 = (FUNCTION_TABLE_ii[$load & 8191](457459)|0); //@line 3056 "libraries/glad/glad.cpp"
  HEAP32[165488] = $864; //@line 3056 "libraries/glad/glad.cpp"
  $865 = (FUNCTION_TABLE_ii[$load & 8191](457475)|0); //@line 3057 "libraries/glad/glad.cpp"
  HEAP32[164764] = $865; //@line 3057 "libraries/glad/glad.cpp"
  $866 = (FUNCTION_TABLE_ii[$load & 8191](457495)|0); //@line 3058 "libraries/glad/glad.cpp"
  HEAP32[165329] = $866; //@line 3058 "libraries/glad/glad.cpp"
  $867 = (FUNCTION_TABLE_ii[$load & 8191](457514)|0); //@line 3059 "libraries/glad/glad.cpp"
  HEAP32[164742] = $867; //@line 3059 "libraries/glad/glad.cpp"
  $868 = (FUNCTION_TABLE_ii[$load & 8191](457536)|0); //@line 3060 "libraries/glad/glad.cpp"
  HEAP32[165317] = $868; //@line 3060 "libraries/glad/glad.cpp"
  $869 = (FUNCTION_TABLE_ii[$load & 8191](457559)|0); //@line 3061 "libraries/glad/glad.cpp"
  HEAP32[165382] = $869; //@line 3061 "libraries/glad/glad.cpp"
  $870 = (FUNCTION_TABLE_ii[$load & 8191](457581)|0); //@line 3062 "libraries/glad/glad.cpp"
  HEAP32[165660] = $870; //@line 3062 "libraries/glad/glad.cpp"
  $871 = (FUNCTION_TABLE_ii[$load & 8191](457606)|0); //@line 3063 "libraries/glad/glad.cpp"
  HEAP32[165122] = $871; //@line 3063 "libraries/glad/glad.cpp"
  $872 = (FUNCTION_TABLE_ii[$load & 8191](457628)|0); //@line 3064 "libraries/glad/glad.cpp"
  HEAP32[165291] = $872; //@line 3064 "libraries/glad/glad.cpp"
  $873 = (FUNCTION_TABLE_ii[$load & 8191](457648)|0); //@line 3065 "libraries/glad/glad.cpp"
  HEAP32[165338] = $873; //@line 3065 "libraries/glad/glad.cpp"
  $874 = (FUNCTION_TABLE_ii[$load & 8191](457671)|0); //@line 3066 "libraries/glad/glad.cpp"
  HEAP32[164876] = $874; //@line 3066 "libraries/glad/glad.cpp"
  $875 = (FUNCTION_TABLE_ii[$load & 8191](457690)|0); //@line 3067 "libraries/glad/glad.cpp"
  HEAP32[165269] = $875; //@line 3067 "libraries/glad/glad.cpp"
  $876 = (FUNCTION_TABLE_ii[$load & 8191](457710)|0); //@line 3068 "libraries/glad/glad.cpp"
  HEAP32[164879] = $876; //@line 3068 "libraries/glad/glad.cpp"
  $877 = (FUNCTION_TABLE_ii[$load & 8191](457729)|0); //@line 3069 "libraries/glad/glad.cpp"
  HEAP32[165648] = $877; //@line 3069 "libraries/glad/glad.cpp"
  $878 = (FUNCTION_TABLE_ii[$load & 8191](457749)|0); //@line 3070 "libraries/glad/glad.cpp"
  HEAP32[164878] = $878; //@line 3070 "libraries/glad/glad.cpp"
  $879 = (FUNCTION_TABLE_ii[$load & 8191](457768)|0); //@line 3071 "libraries/glad/glad.cpp"
  HEAP32[165173] = $879; //@line 3071 "libraries/glad/glad.cpp"
  $880 = (FUNCTION_TABLE_ii[$load & 8191](457788)|0); //@line 3072 "libraries/glad/glad.cpp"
  HEAP32[164943] = $880; //@line 3072 "libraries/glad/glad.cpp"
  $881 = (FUNCTION_TABLE_ii[$load & 8191](457808)|0); //@line 3073 "libraries/glad/glad.cpp"
  HEAP32[164808] = $881; //@line 3073 "libraries/glad/glad.cpp"
  $882 = (FUNCTION_TABLE_ii[$load & 8191](457829)|0); //@line 3074 "libraries/glad/glad.cpp"
  HEAP32[165552] = $882; //@line 3074 "libraries/glad/glad.cpp"
  $883 = (FUNCTION_TABLE_ii[$load & 8191](457848)|0); //@line 3075 "libraries/glad/glad.cpp"
  HEAP32[164812] = $883; //@line 3075 "libraries/glad/glad.cpp"
  $884 = (FUNCTION_TABLE_ii[$load & 8191](457868)|0); //@line 3076 "libraries/glad/glad.cpp"
  HEAP32[165139] = $884; //@line 3076 "libraries/glad/glad.cpp"
  $885 = (FUNCTION_TABLE_ii[$load & 8191](457887)|0); //@line 3077 "libraries/glad/glad.cpp"
  HEAP32[164951] = $885; //@line 3077 "libraries/glad/glad.cpp"
  $886 = (FUNCTION_TABLE_ii[$load & 8191](457907)|0); //@line 3078 "libraries/glad/glad.cpp"
  HEAP32[165050] = $886; //@line 3078 "libraries/glad/glad.cpp"
  $887 = (FUNCTION_TABLE_ii[$load & 8191](457926)|0); //@line 3079 "libraries/glad/glad.cpp"
  HEAP32[164984] = $887; //@line 3079 "libraries/glad/glad.cpp"
  $888 = (FUNCTION_TABLE_ii[$load & 8191](457946)|0); //@line 3080 "libraries/glad/glad.cpp"
  HEAP32[165598] = $888; //@line 3080 "libraries/glad/glad.cpp"
  $889 = (FUNCTION_TABLE_ii[$load & 8191](457966)|0); //@line 3081 "libraries/glad/glad.cpp"
  HEAP32[165548] = $889; //@line 3081 "libraries/glad/glad.cpp"
  $890 = (FUNCTION_TABLE_ii[$load & 8191](457987)|0); //@line 3082 "libraries/glad/glad.cpp"
  HEAP32[165144] = $890; //@line 3082 "libraries/glad/glad.cpp"
  $891 = (FUNCTION_TABLE_ii[$load & 8191](458006)|0); //@line 3083 "libraries/glad/glad.cpp"
  HEAP32[164886] = $891; //@line 3083 "libraries/glad/glad.cpp"
  $892 = (FUNCTION_TABLE_ii[$load & 8191](458026)|0); //@line 3084 "libraries/glad/glad.cpp"
  HEAP32[165334] = $892; //@line 3084 "libraries/glad/glad.cpp"
  $893 = (FUNCTION_TABLE_ii[$load & 8191](458045)|0); //@line 3085 "libraries/glad/glad.cpp"
  HEAP32[165408] = $893; //@line 3085 "libraries/glad/glad.cpp"
  $894 = (FUNCTION_TABLE_ii[$load & 8191](458065)|0); //@line 3086 "libraries/glad/glad.cpp"
  HEAP32[165335] = $894; //@line 3086 "libraries/glad/glad.cpp"
  $895 = (FUNCTION_TABLE_ii[$load & 8191](458084)|0); //@line 3087 "libraries/glad/glad.cpp"
  HEAP32[164667] = $895; //@line 3087 "libraries/glad/glad.cpp"
  $896 = (FUNCTION_TABLE_ii[$load & 8191](458104)|0); //@line 3088 "libraries/glad/glad.cpp"
  HEAP32[164855] = $896; //@line 3088 "libraries/glad/glad.cpp"
  $897 = (FUNCTION_TABLE_ii[$load & 8191](458124)|0); //@line 3089 "libraries/glad/glad.cpp"
  HEAP32[165413] = $897; //@line 3089 "libraries/glad/glad.cpp"
  $898 = (FUNCTION_TABLE_ii[$load & 8191](458145)|0); //@line 3090 "libraries/glad/glad.cpp"
  HEAP32[165209] = $898; //@line 3090 "libraries/glad/glad.cpp"
  $899 = (FUNCTION_TABLE_ii[$load & 8191](458164)|0); //@line 3091 "libraries/glad/glad.cpp"
  HEAP32[165053] = $899; //@line 3091 "libraries/glad/glad.cpp"
  $900 = (FUNCTION_TABLE_ii[$load & 8191](458184)|0); //@line 3092 "libraries/glad/glad.cpp"
  HEAP32[165211] = $900; //@line 3092 "libraries/glad/glad.cpp"
  $901 = (FUNCTION_TABLE_ii[$load & 8191](458203)|0); //@line 3093 "libraries/glad/glad.cpp"
  HEAP32[165593] = $901; //@line 3093 "libraries/glad/glad.cpp"
  $902 = (FUNCTION_TABLE_ii[$load & 8191](458223)|0); //@line 3094 "libraries/glad/glad.cpp"
  HEAP32[165213] = $902; //@line 3094 "libraries/glad/glad.cpp"
  $903 = (FUNCTION_TABLE_ii[$load & 8191](458242)|0); //@line 3095 "libraries/glad/glad.cpp"
  HEAP32[165402] = $903; //@line 3095 "libraries/glad/glad.cpp"
  $904 = (FUNCTION_TABLE_ii[$load & 8191](458262)|0); //@line 3096 "libraries/glad/glad.cpp"
  HEAP32[165599] = $904; //@line 3096 "libraries/glad/glad.cpp"
  $905 = (FUNCTION_TABLE_ii[$load & 8191](458282)|0); //@line 3097 "libraries/glad/glad.cpp"
  HEAP32[165198] = $905; //@line 3097 "libraries/glad/glad.cpp"
  $906 = (FUNCTION_TABLE_ii[$load & 8191](458303)|0); //@line 3098 "libraries/glad/glad.cpp"
  HEAP32[165251] = $906; //@line 3098 "libraries/glad/glad.cpp"
  $907 = (FUNCTION_TABLE_ii[$load & 8191](458329)|0); //@line 3099 "libraries/glad/glad.cpp"
  HEAP32[164993] = $907; //@line 3099 "libraries/glad/glad.cpp"
  $908 = (FUNCTION_TABLE_ii[$load & 8191](458355)|0); //@line 3100 "libraries/glad/glad.cpp"
  HEAP32[164737] = $908; //@line 3100 "libraries/glad/glad.cpp"
  $909 = (FUNCTION_TABLE_ii[$load & 8191](458381)|0); //@line 3101 "libraries/glad/glad.cpp"
  HEAP32[164675] = $909; //@line 3101 "libraries/glad/glad.cpp"
  $910 = (FUNCTION_TABLE_ii[$load & 8191](458407)|0); //@line 3102 "libraries/glad/glad.cpp"
  HEAP32[165229] = $910; //@line 3102 "libraries/glad/glad.cpp"
  $911 = (FUNCTION_TABLE_ii[$load & 8191](458433)|0); //@line 3103 "libraries/glad/glad.cpp"
  HEAP32[164953] = $911; //@line 3103 "libraries/glad/glad.cpp"
  $912 = (FUNCTION_TABLE_ii[$load & 8191](458459)|0); //@line 3104 "libraries/glad/glad.cpp"
  HEAP32[165399] = $912; //@line 3104 "libraries/glad/glad.cpp"
  $913 = (FUNCTION_TABLE_ii[$load & 8191](458487)|0); //@line 3105 "libraries/glad/glad.cpp"
  HEAP32[164731] = $913; //@line 3105 "libraries/glad/glad.cpp"
  $914 = (FUNCTION_TABLE_ii[$load & 8191](458515)|0); //@line 3106 "libraries/glad/glad.cpp"
  HEAP32[165535] = $914; //@line 3106 "libraries/glad/glad.cpp"
  $915 = (FUNCTION_TABLE_ii[$load & 8191](458543)|0); //@line 3107 "libraries/glad/glad.cpp"
  HEAP32[165520] = $915; //@line 3107 "libraries/glad/glad.cpp"
  $916 = (FUNCTION_TABLE_ii[$load & 8191](458571)|0); //@line 3108 "libraries/glad/glad.cpp"
  HEAP32[165021] = $916; //@line 3108 "libraries/glad/glad.cpp"
  $917 = (FUNCTION_TABLE_ii[$load & 8191](458599)|0); //@line 3109 "libraries/glad/glad.cpp"
  HEAP32[165623] = $917; //@line 3109 "libraries/glad/glad.cpp"
  $918 = (FUNCTION_TABLE_ii[$load & 8191](458627)|0); //@line 3110 "libraries/glad/glad.cpp"
  HEAP32[165130] = $918; //@line 3110 "libraries/glad/glad.cpp"
  $919 = (FUNCTION_TABLE_ii[$load & 8191](458655)|0); //@line 3111 "libraries/glad/glad.cpp"
  HEAP32[164728] = $919; //@line 3111 "libraries/glad/glad.cpp"
  $920 = (FUNCTION_TABLE_ii[$load & 8191](458683)|0); //@line 3112 "libraries/glad/glad.cpp"
  HEAP32[164956] = $920; //@line 3112 "libraries/glad/glad.cpp"
  $921 = (FUNCTION_TABLE_ii[$load & 8191](458711)|0); //@line 3113 "libraries/glad/glad.cpp"
  HEAP32[164823] = $921; //@line 3113 "libraries/glad/glad.cpp"
  $922 = (FUNCTION_TABLE_ii[$load & 8191](458739)|0); //@line 3114 "libraries/glad/glad.cpp"
  HEAP32[164802] = $922; //@line 3114 "libraries/glad/glad.cpp"
  $923 = (FUNCTION_TABLE_ii[$load & 8191](458767)|0); //@line 3115 "libraries/glad/glad.cpp"
  HEAP32[165051] = $923; //@line 3115 "libraries/glad/glad.cpp"
  $924 = (FUNCTION_TABLE_ii[$load & 8191](458795)|0); //@line 3116 "libraries/glad/glad.cpp"
  HEAP32[164901] = $924; //@line 3116 "libraries/glad/glad.cpp"
  $925 = (FUNCTION_TABLE_ii[$load & 8191](458821)|0); //@line 3117 "libraries/glad/glad.cpp"
  HEAP32[165019] = $925; //@line 3117 "libraries/glad/glad.cpp"
  $926 = (FUNCTION_TABLE_ii[$load & 8191](458849)|0); //@line 3118 "libraries/glad/glad.cpp"
  HEAP32[165521] = $926; //@line 3118 "libraries/glad/glad.cpp"
  $927 = (FUNCTION_TABLE_ii[$load & 8191](458867)|0); //@line 3119 "libraries/glad/glad.cpp"
  HEAP32[165320] = $927; //@line 3119 "libraries/glad/glad.cpp"
  $928 = (FUNCTION_TABLE_ii[$load & 8191](458885)|0); //@line 3120 "libraries/glad/glad.cpp"
  HEAP32[165659] = $928; //@line 3120 "libraries/glad/glad.cpp"
  $929 = (FUNCTION_TABLE_ii[$load & 8191](458903)|0); //@line 3121 "libraries/glad/glad.cpp"
  HEAP32[164734] = $929; //@line 3121 "libraries/glad/glad.cpp"
  $930 = (FUNCTION_TABLE_ii[$load & 8191](458921)|0); //@line 3122 "libraries/glad/glad.cpp"
  HEAP32[164851] = $930; //@line 3122 "libraries/glad/glad.cpp"
  $931 = (FUNCTION_TABLE_ii[$load & 8191](458940)|0); //@line 3123 "libraries/glad/glad.cpp"
  HEAP32[165311] = $931; //@line 3123 "libraries/glad/glad.cpp"
  $932 = (FUNCTION_TABLE_ii[$load & 8191](458959)|0); //@line 3124 "libraries/glad/glad.cpp"
  HEAP32[165433] = $932; //@line 3124 "libraries/glad/glad.cpp"
  $933 = (FUNCTION_TABLE_ii[$load & 8191](458978)|0); //@line 3125 "libraries/glad/glad.cpp"
  HEAP32[164875] = $933; //@line 3125 "libraries/glad/glad.cpp"
  $934 = (FUNCTION_TABLE_ii[$load & 8191](458997)|0); //@line 3126 "libraries/glad/glad.cpp"
  HEAP32[165306] = $934; //@line 3126 "libraries/glad/glad.cpp"
  $935 = (FUNCTION_TABLE_ii[$load & 8191](459020)|0); //@line 3127 "libraries/glad/glad.cpp"
  HEAP32[165351] = $935; //@line 3127 "libraries/glad/glad.cpp"
  $936 = (FUNCTION_TABLE_ii[$load & 8191](459041)|0); //@line 3128 "libraries/glad/glad.cpp"
  HEAP32[165246] = $936; //@line 3128 "libraries/glad/glad.cpp"
  $937 = (FUNCTION_TABLE_ii[$load & 8191](459058)|0); //@line 3129 "libraries/glad/glad.cpp"
  HEAP32[165118] = $937; //@line 3129 "libraries/glad/glad.cpp"
  $938 = (FUNCTION_TABLE_ii[$load & 8191](459077)|0); //@line 3130 "libraries/glad/glad.cpp"
  HEAP32[165165] = $938; //@line 3130 "libraries/glad/glad.cpp"
  $939 = (FUNCTION_TABLE_ii[$load & 8191](459097)|0); //@line 3131 "libraries/glad/glad.cpp"
  HEAP32[164999] = $939; //@line 3131 "libraries/glad/glad.cpp"
  $940 = (FUNCTION_TABLE_ii[$load & 8191](459113)|0); //@line 3132 "libraries/glad/glad.cpp"
  HEAP32[165526] = $940; //@line 3132 "libraries/glad/glad.cpp"
  $941 = (FUNCTION_TABLE_ii[$load & 8191](459130)|0); //@line 3133 "libraries/glad/glad.cpp"
  HEAP32[165361] = $941; //@line 3133 "libraries/glad/glad.cpp"
  $942 = (FUNCTION_TABLE_ii[$load & 8191](459148)|0); //@line 3134 "libraries/glad/glad.cpp"
  HEAP32[165152] = $942; //@line 3134 "libraries/glad/glad.cpp"
  $943 = (FUNCTION_TABLE_ii[$load & 8191](459167)|0); //@line 3135 "libraries/glad/glad.cpp"
  HEAP32[165083] = $943; //@line 3135 "libraries/glad/glad.cpp"
  $944 = (FUNCTION_TABLE_ii[$load & 8191](459187)|0); //@line 3136 "libraries/glad/glad.cpp"
  HEAP32[165464] = $944; //@line 3136 "libraries/glad/glad.cpp"
  $945 = (FUNCTION_TABLE_ii[$load & 8191](459201)|0); //@line 3137 "libraries/glad/glad.cpp"
  HEAP32[164636] = $945; //@line 3137 "libraries/glad/glad.cpp"
 }
 $946 = HEAP8[1252008]|0; //@line 3142 "libraries/glad/glad.cpp"
 $947 = ($946<<24>>24)==(0); //@line 3142 "libraries/glad/glad.cpp"
 if (!($947)) {
  $948 = (FUNCTION_TABLE_ii[$load & 8191](459216)|0); //@line 3143 "libraries/glad/glad.cpp"
  HEAP32[165657] = $948; //@line 3143 "libraries/glad/glad.cpp"
  $949 = (FUNCTION_TABLE_ii[$load & 8191](459250)|0); //@line 3144 "libraries/glad/glad.cpp"
  HEAP32[164665] = $949; //@line 3144 "libraries/glad/glad.cpp"
  $950 = (FUNCTION_TABLE_ii[$load & 8191](459286)|0); //@line 3145 "libraries/glad/glad.cpp"
  HEAP32[164895] = $950; //@line 3145 "libraries/glad/glad.cpp"
  $951 = (FUNCTION_TABLE_ii[$load & 8191](459332)|0); //@line 3146 "libraries/glad/glad.cpp"
  HEAP32[164979] = $951; //@line 3146 "libraries/glad/glad.cpp"
  $952 = (FUNCTION_TABLE_ii[$load & 8191](459354)|0); //@line 3147 "libraries/glad/glad.cpp"
  HEAP32[165158] = $952; //@line 3147 "libraries/glad/glad.cpp"
  $953 = (FUNCTION_TABLE_ii[$load & 8191](459387)|0); //@line 3148 "libraries/glad/glad.cpp"
  HEAP32[164735] = $953; //@line 3148 "libraries/glad/glad.cpp"
  $954 = (FUNCTION_TABLE_ii[$load & 8191](459406)|0); //@line 3149 "libraries/glad/glad.cpp"
  HEAP32[164771] = $954; //@line 3149 "libraries/glad/glad.cpp"
  $955 = (FUNCTION_TABLE_ii[$load & 8191](459422)|0); //@line 3150 "libraries/glad/glad.cpp"
  HEAP32[165180] = $955; //@line 3150 "libraries/glad/glad.cpp"
  $956 = (FUNCTION_TABLE_ii[$load & 8191](459437)|0); //@line 3151 "libraries/glad/glad.cpp"
  HEAP32[165612] = $956; //@line 3151 "libraries/glad/glad.cpp"
  $957 = (FUNCTION_TABLE_ii[$load & 8191](459452)|0); //@line 3152 "libraries/glad/glad.cpp"
  HEAP32[165339] = $957; //@line 3152 "libraries/glad/glad.cpp"
  $958 = (FUNCTION_TABLE_ii[$load & 8191](459467)|0); //@line 3153 "libraries/glad/glad.cpp"
  HEAP32[164919] = $958; //@line 3153 "libraries/glad/glad.cpp"
  $959 = (FUNCTION_TABLE_ii[$load & 8191](459500)|0); //@line 3154 "libraries/glad/glad.cpp"
  HEAP32[164649] = $959; //@line 3154 "libraries/glad/glad.cpp"
 }
 $960 = HEAP8[1252009]|0; //@line 3159 "libraries/glad/glad.cpp"
 $961 = ($960<<24>>24)==(0); //@line 3159 "libraries/glad/glad.cpp"
 if (!($961)) {
  $962 = (FUNCTION_TABLE_ii[$load & 8191](459539)|0); //@line 3160 "libraries/glad/glad.cpp"
  HEAP32[165484] = $962; //@line 3160 "libraries/glad/glad.cpp"
  $963 = (FUNCTION_TABLE_ii[$load & 8191](459557)|0); //@line 3161 "libraries/glad/glad.cpp"
  HEAP32[165178] = $963; //@line 3161 "libraries/glad/glad.cpp"
  $964 = (FUNCTION_TABLE_ii[$load & 8191](459578)|0); //@line 3162 "libraries/glad/glad.cpp"
  HEAP32[164957] = $964; //@line 3162 "libraries/glad/glad.cpp"
  $965 = (FUNCTION_TABLE_ii[$load & 8191](459596)|0); //@line 3163 "libraries/glad/glad.cpp"
  HEAP32[165427] = $965; //@line 3163 "libraries/glad/glad.cpp"
  $966 = (FUNCTION_TABLE_ii[$load & 8191](459622)|0); //@line 3164 "libraries/glad/glad.cpp"
  HEAP32[164975] = $966; //@line 3164 "libraries/glad/glad.cpp"
  $967 = (FUNCTION_TABLE_ii[$load & 8191](459641)|0); //@line 3165 "libraries/glad/glad.cpp"
  HEAP32[165149] = $967; //@line 3165 "libraries/glad/glad.cpp"
  $968 = (FUNCTION_TABLE_ii[$load & 8191](459665)|0); //@line 3166 "libraries/glad/glad.cpp"
  HEAP32[165632] = $968; //@line 3166 "libraries/glad/glad.cpp"
  $969 = (FUNCTION_TABLE_ii[$load & 8191](459693)|0); //@line 3167 "libraries/glad/glad.cpp"
  HEAP32[165087] = $969; //@line 3167 "libraries/glad/glad.cpp"
  $970 = (FUNCTION_TABLE_ii[$load & 8191](459717)|0); //@line 3168 "libraries/glad/glad.cpp"
  HEAP32[165375] = $970; //@line 3168 "libraries/glad/glad.cpp"
  $971 = (FUNCTION_TABLE_ii[$load & 8191](459741)|0); //@line 3169 "libraries/glad/glad.cpp"
  HEAP32[165102] = $971; //@line 3169 "libraries/glad/glad.cpp"
  $972 = (FUNCTION_TABLE_ii[$load & 8191](459762)|0); //@line 3170 "libraries/glad/glad.cpp"
  HEAP32[164749] = $972; //@line 3170 "libraries/glad/glad.cpp"
  $973 = (FUNCTION_TABLE_ii[$load & 8191](459788)|0); //@line 3171 "libraries/glad/glad.cpp"
  HEAP32[165010] = $973; //@line 3171 "libraries/glad/glad.cpp"
  $974 = (FUNCTION_TABLE_ii[$load & 8191](459811)|0); //@line 3172 "libraries/glad/glad.cpp"
  HEAP32[164995] = $974; //@line 3172 "libraries/glad/glad.cpp"
  $975 = (FUNCTION_TABLE_ii[$load & 8191](459835)|0); //@line 3173 "libraries/glad/glad.cpp"
  HEAP32[165022] = $975; //@line 3173 "libraries/glad/glad.cpp"
  $976 = (FUNCTION_TABLE_ii[$load & 8191](459862)|0); //@line 3174 "libraries/glad/glad.cpp"
  HEAP32[164703] = $976; //@line 3174 "libraries/glad/glad.cpp"
  $977 = (FUNCTION_TABLE_ii[$load & 8191](459888)|0); //@line 3175 "libraries/glad/glad.cpp"
  HEAP32[165312] = $977; //@line 3175 "libraries/glad/glad.cpp"
  $978 = (FUNCTION_TABLE_ii[$load & 8191](459916)|0); //@line 3176 "libraries/glad/glad.cpp"
  HEAP32[164843] = $978; //@line 3176 "libraries/glad/glad.cpp"
  $979 = (FUNCTION_TABLE_ii[$load & 8191](459940)|0); //@line 3177 "libraries/glad/glad.cpp"
  HEAP32[165420] = $979; //@line 3177 "libraries/glad/glad.cpp"
  $980 = (FUNCTION_TABLE_ii[$load & 8191](459966)|0); //@line 3178 "libraries/glad/glad.cpp"
  HEAP32[165150] = $980; //@line 3178 "libraries/glad/glad.cpp"
  $981 = (FUNCTION_TABLE_ii[$load & 8191](459991)|0); //@line 3179 "libraries/glad/glad.cpp"
  HEAP32[165663] = $981; //@line 3179 "libraries/glad/glad.cpp"
  $982 = (FUNCTION_TABLE_ii[$load & 8191](460014)|0); //@line 3180 "libraries/glad/glad.cpp"
  HEAP32[165155] = $982; //@line 3180 "libraries/glad/glad.cpp"
  $983 = (FUNCTION_TABLE_ii[$load & 8191](460043)|0); //@line 3181 "libraries/glad/glad.cpp"
  HEAP32[165458] = $983; //@line 3181 "libraries/glad/glad.cpp"
  $984 = (FUNCTION_TABLE_ii[$load & 8191](460077)|0); //@line 3182 "libraries/glad/glad.cpp"
  HEAP32[164820] = $984; //@line 3182 "libraries/glad/glad.cpp"
  $985 = (FUNCTION_TABLE_ii[$load & 8191](460105)|0); //@line 3183 "libraries/glad/glad.cpp"
  HEAP32[165654] = $985; //@line 3183 "libraries/glad/glad.cpp"
  $986 = (FUNCTION_TABLE_ii[$load & 8191](460122)|0); //@line 3184 "libraries/glad/glad.cpp"
  HEAP32[165560] = $986; //@line 3184 "libraries/glad/glad.cpp"
  $987 = (FUNCTION_TABLE_ii[$load & 8191](460148)|0); //@line 3185 "libraries/glad/glad.cpp"
  HEAP32[164910] = $987; //@line 3185 "libraries/glad/glad.cpp"
  $988 = (FUNCTION_TABLE_ii[$load & 8191](460174)|0); //@line 3186 "libraries/glad/glad.cpp"
  HEAP32[165533] = $988; //@line 3186 "libraries/glad/glad.cpp"
  $989 = (FUNCTION_TABLE_ii[$load & 8191](460188)|0); //@line 3187 "libraries/glad/glad.cpp"
  HEAP32[165479] = $989; //@line 3187 "libraries/glad/glad.cpp"
  $990 = (FUNCTION_TABLE_ii[$load & 8191](460207)|0); //@line 3188 "libraries/glad/glad.cpp"
  HEAP32[165103] = $990; //@line 3188 "libraries/glad/glad.cpp"
  $991 = (FUNCTION_TABLE_ii[$load & 8191](460228)|0); //@line 3189 "libraries/glad/glad.cpp"
  HEAP32[165544] = $991; //@line 3189 "libraries/glad/glad.cpp"
  $992 = (FUNCTION_TABLE_ii[$load & 8191](460250)|0); //@line 3190 "libraries/glad/glad.cpp"
  HEAP32[165074] = $992; //@line 3190 "libraries/glad/glad.cpp"
  $993 = (FUNCTION_TABLE_ii[$load & 8191](460272)|0); //@line 3191 "libraries/glad/glad.cpp"
  HEAP32[165107] = $993; //@line 3191 "libraries/glad/glad.cpp"
  $994 = (FUNCTION_TABLE_ii[$load & 8191](460294)|0); //@line 3192 "libraries/glad/glad.cpp"
  HEAP32[164858] = $994; //@line 3192 "libraries/glad/glad.cpp"
  $995 = (FUNCTION_TABLE_ii[$load & 8191](460317)|0); //@line 3193 "libraries/glad/glad.cpp"
  HEAP32[164693] = $995; //@line 3193 "libraries/glad/glad.cpp"
  $996 = (FUNCTION_TABLE_ii[$load & 8191](460339)|0); //@line 3194 "libraries/glad/glad.cpp"
  HEAP32[165481] = $996; //@line 3194 "libraries/glad/glad.cpp"
  $997 = (FUNCTION_TABLE_ii[$load & 8191](460360)|0); //@line 3195 "libraries/glad/glad.cpp"
  HEAP32[165100] = $997; //@line 3195 "libraries/glad/glad.cpp"
  $998 = (FUNCTION_TABLE_ii[$load & 8191](460383)|0); //@line 3196 "libraries/glad/glad.cpp"
  HEAP32[164807] = $998; //@line 3196 "libraries/glad/glad.cpp"
  $999 = (FUNCTION_TABLE_ii[$load & 8191](460404)|0); //@line 3197 "libraries/glad/glad.cpp"
  HEAP32[164899] = $999; //@line 3197 "libraries/glad/glad.cpp"
  $1000 = (FUNCTION_TABLE_ii[$load & 8191](460421)|0); //@line 3198 "libraries/glad/glad.cpp"
  HEAP32[164838] = $1000; //@line 3198 "libraries/glad/glad.cpp"
  $1001 = (FUNCTION_TABLE_ii[$load & 8191](460437)|0); //@line 3199 "libraries/glad/glad.cpp"
  HEAP32[165347] = $1001; //@line 3199 "libraries/glad/glad.cpp"
  $1002 = (FUNCTION_TABLE_ii[$load & 8191](460451)|0); //@line 3200 "libraries/glad/glad.cpp"
  HEAP32[165138] = $1002; //@line 3200 "libraries/glad/glad.cpp"
  $1003 = (FUNCTION_TABLE_ii[$load & 8191](460468)|0); //@line 3201 "libraries/glad/glad.cpp"
  HEAP32[164994] = $1003; //@line 3201 "libraries/glad/glad.cpp"
  $1004 = (FUNCTION_TABLE_ii[$load & 8191](460485)|0); //@line 3202 "libraries/glad/glad.cpp"
  HEAP32[164946] = $1004; //@line 3202 "libraries/glad/glad.cpp"
  $1005 = (FUNCTION_TABLE_ii[$load & 8191](615462)|0); //@line 3203 "libraries/glad/glad.cpp"
  HEAP32[164857] = $1005; //@line 3203 "libraries/glad/glad.cpp"
  $1006 = (FUNCTION_TABLE_ii[$load & 8191](615462)|0); //@line 3204 "libraries/glad/glad.cpp"
  HEAP32[164857] = $1006; //@line 3204 "libraries/glad/glad.cpp"
 }
 $1007 = HEAP8[1252010]|0; //@line 3209 "libraries/glad/glad.cpp"
 $1008 = ($1007<<24>>24)==(0); //@line 3209 "libraries/glad/glad.cpp"
 if (!($1008)) {
  $1009 = (FUNCTION_TABLE_ii[$load & 8191](460505)|0); //@line 3210 "libraries/glad/glad.cpp"
  HEAP32[165460] = $1009; //@line 3210 "libraries/glad/glad.cpp"
  $1010 = (FUNCTION_TABLE_ii[$load & 8191](460521)|0); //@line 3211 "libraries/glad/glad.cpp"
  HEAP32[165062] = $1010; //@line 3211 "libraries/glad/glad.cpp"
  $1011 = (FUNCTION_TABLE_ii[$load & 8191](460537)|0); //@line 3212 "libraries/glad/glad.cpp"
  HEAP32[164801] = $1011; //@line 3212 "libraries/glad/glad.cpp"
  $1012 = (FUNCTION_TABLE_ii[$load & 8191](460556)|0); //@line 3213 "libraries/glad/glad.cpp"
  HEAP32[165625] = $1012; //@line 3213 "libraries/glad/glad.cpp"
  $1013 = (FUNCTION_TABLE_ii[$load & 8191](460574)|0); //@line 3214 "libraries/glad/glad.cpp"
  HEAP32[164971] = $1013; //@line 3214 "libraries/glad/glad.cpp"
  $1014 = (FUNCTION_TABLE_ii[$load & 8191](460593)|0); //@line 3215 "libraries/glad/glad.cpp"
  HEAP32[164996] = $1014; //@line 3215 "libraries/glad/glad.cpp"
  $1015 = (FUNCTION_TABLE_ii[$load & 8191](460608)|0); //@line 3216 "libraries/glad/glad.cpp"
  HEAP32[165492] = $1015; //@line 3216 "libraries/glad/glad.cpp"
  $1016 = (FUNCTION_TABLE_ii[$load & 8191](460623)|0); //@line 3217 "libraries/glad/glad.cpp"
  HEAP32[165379] = $1016; //@line 3217 "libraries/glad/glad.cpp"
  $1017 = (FUNCTION_TABLE_ii[$load & 8191](460643)|0); //@line 3218 "libraries/glad/glad.cpp"
  HEAP32[164908] = $1017; //@line 3218 "libraries/glad/glad.cpp"
 }
 $1018 = HEAP8[1252011]|0; //@line 3223 "libraries/glad/glad.cpp"
 $1019 = ($1018<<24>>24)==(0); //@line 3223 "libraries/glad/glad.cpp"
 if (!($1019)) {
  $1020 = (FUNCTION_TABLE_ii[$load & 8191](460663)|0); //@line 3224 "libraries/glad/glad.cpp"
  HEAP32[165662] = $1020; //@line 3224 "libraries/glad/glad.cpp"
  $1021 = (FUNCTION_TABLE_ii[$load & 8191](460677)|0); //@line 3225 "libraries/glad/glad.cpp"
  HEAP32[165111] = $1021; //@line 3225 "libraries/glad/glad.cpp"
  $1022 = (FUNCTION_TABLE_ii[$load & 8191](460704)|0); //@line 3226 "libraries/glad/glad.cpp"
  HEAP32[164799] = $1022; //@line 3226 "libraries/glad/glad.cpp"
  $1023 = (FUNCTION_TABLE_ii[$load & 8191](460734)|0); //@line 3227 "libraries/glad/glad.cpp"
  HEAP32[164856] = $1023; //@line 3227 "libraries/glad/glad.cpp"
  $1024 = (FUNCTION_TABLE_ii[$load & 8191](460765)|0); //@line 3228 "libraries/glad/glad.cpp"
  HEAP32[164761] = $1024; //@line 3228 "libraries/glad/glad.cpp"
  $1025 = (FUNCTION_TABLE_ii[$load & 8191](460790)|0); //@line 3229 "libraries/glad/glad.cpp"
  HEAP32[164927] = $1025; //@line 3229 "libraries/glad/glad.cpp"
  $1026 = (FUNCTION_TABLE_ii[$load & 8191](460816)|0); //@line 3230 "libraries/glad/glad.cpp"
  HEAP32[165035] = $1026; //@line 3230 "libraries/glad/glad.cpp"
  $1027 = (FUNCTION_TABLE_ii[$load & 8191](460844)|0); //@line 3231 "libraries/glad/glad.cpp"
  HEAP32[165655] = $1027; //@line 3231 "libraries/glad/glad.cpp"
  $1028 = (FUNCTION_TABLE_ii[$load & 8191](460860)|0); //@line 3232 "libraries/glad/glad.cpp"
  HEAP32[164998] = $1028; //@line 3232 "libraries/glad/glad.cpp"
  $1029 = (FUNCTION_TABLE_ii[$load & 8191](460881)|0); //@line 3233 "libraries/glad/glad.cpp"
  HEAP32[165609] = $1029; //@line 3233 "libraries/glad/glad.cpp"
  $1030 = (FUNCTION_TABLE_ii[$load & 8191](460899)|0); //@line 3234 "libraries/glad/glad.cpp"
  HEAP32[165366] = $1030; //@line 3234 "libraries/glad/glad.cpp"
  $1031 = (FUNCTION_TABLE_ii[$load & 8191](460920)|0); //@line 3235 "libraries/glad/glad.cpp"
  HEAP32[164729] = $1031; //@line 3235 "libraries/glad/glad.cpp"
  $1032 = (FUNCTION_TABLE_ii[$load & 8191](460945)|0); //@line 3236 "libraries/glad/glad.cpp"
  HEAP32[165241] = $1032; //@line 3236 "libraries/glad/glad.cpp"
  $1033 = (FUNCTION_TABLE_ii[$load & 8191](460968)|0); //@line 3237 "libraries/glad/glad.cpp"
  HEAP32[165614] = $1033; //@line 3237 "libraries/glad/glad.cpp"
  $1034 = (FUNCTION_TABLE_ii[$load & 8191](460994)|0); //@line 3238 "libraries/glad/glad.cpp"
  HEAP32[164848] = $1034; //@line 3238 "libraries/glad/glad.cpp"
  $1035 = (FUNCTION_TABLE_ii[$load & 8191](461011)|0); //@line 3239 "libraries/glad/glad.cpp"
  HEAP32[165615] = $1035; //@line 3239 "libraries/glad/glad.cpp"
  $1036 = (FUNCTION_TABLE_ii[$load & 8191](461033)|0); //@line 3240 "libraries/glad/glad.cpp"
  HEAP32[165232] = $1036; //@line 3240 "libraries/glad/glad.cpp"
  $1037 = (FUNCTION_TABLE_ii[$load & 8191](461052)|0); //@line 3241 "libraries/glad/glad.cpp"
  HEAP32[165374] = $1037; //@line 3241 "libraries/glad/glad.cpp"
  $1038 = (FUNCTION_TABLE_ii[$load & 8191](461082)|0); //@line 3242 "libraries/glad/glad.cpp"
  HEAP32[164662] = $1038; //@line 3242 "libraries/glad/glad.cpp"
  $1039 = (FUNCTION_TABLE_ii[$load & 8191](461110)|0); //@line 3243 "libraries/glad/glad.cpp"
  HEAP32[165554] = $1039; //@line 3243 "libraries/glad/glad.cpp"
  $1040 = (FUNCTION_TABLE_ii[$load & 8191](461140)|0); //@line 3244 "libraries/glad/glad.cpp"
  HEAP32[165426] = $1040; //@line 3244 "libraries/glad/glad.cpp"
  $1041 = (FUNCTION_TABLE_ii[$load & 8191](461165)|0); //@line 3245 "libraries/glad/glad.cpp"
  HEAP32[165371] = $1041; //@line 3245 "libraries/glad/glad.cpp"
  $1042 = (FUNCTION_TABLE_ii[$load & 8191](461189)|0); //@line 3246 "libraries/glad/glad.cpp"
  HEAP32[165545] = $1042; //@line 3246 "libraries/glad/glad.cpp"
  $1043 = (FUNCTION_TABLE_ii[$load & 8191](461210)|0); //@line 3247 "libraries/glad/glad.cpp"
  HEAP32[165390] = $1043; //@line 3247 "libraries/glad/glad.cpp"
  $1044 = (FUNCTION_TABLE_ii[$load & 8191](461241)|0); //@line 3248 "libraries/glad/glad.cpp"
  HEAP32[165656] = $1044; //@line 3248 "libraries/glad/glad.cpp"
  $1045 = (FUNCTION_TABLE_ii[$load & 8191](461270)|0); //@line 3249 "libraries/glad/glad.cpp"
  HEAP32[164730] = $1045; //@line 3249 "libraries/glad/glad.cpp"
  $1046 = (FUNCTION_TABLE_ii[$load & 8191](461296)|0); //@line 3250 "libraries/glad/glad.cpp"
  HEAP32[165616] = $1046; //@line 3250 "libraries/glad/glad.cpp"
  $1047 = (FUNCTION_TABLE_ii[$load & 8191](461327)|0); //@line 3251 "libraries/glad/glad.cpp"
  HEAP32[165341] = $1047; //@line 3251 "libraries/glad/glad.cpp"
  $1048 = (FUNCTION_TABLE_ii[$load & 8191](461356)|0); //@line 3252 "libraries/glad/glad.cpp"
  HEAP32[164863] = $1048; //@line 3252 "libraries/glad/glad.cpp"
  $1049 = (FUNCTION_TABLE_ii[$load & 8191](461386)|0); //@line 3253 "libraries/glad/glad.cpp"
  HEAP32[165201] = $1049; //@line 3253 "libraries/glad/glad.cpp"
  $1050 = (FUNCTION_TABLE_ii[$load & 8191](461415)|0); //@line 3254 "libraries/glad/glad.cpp"
  HEAP32[165323] = $1050; //@line 3254 "libraries/glad/glad.cpp"
  $1051 = (FUNCTION_TABLE_ii[$load & 8191](461448)|0); //@line 3255 "libraries/glad/glad.cpp"
  HEAP32[165428] = $1051; //@line 3255 "libraries/glad/glad.cpp"
  $1052 = (FUNCTION_TABLE_ii[$load & 8191](461484)|0); //@line 3256 "libraries/glad/glad.cpp"
  HEAP32[165523] = $1052; //@line 3256 "libraries/glad/glad.cpp"
  $1053 = (FUNCTION_TABLE_ii[$load & 8191](461510)|0); //@line 3257 "libraries/glad/glad.cpp"
  HEAP32[164931] = $1053; //@line 3257 "libraries/glad/glad.cpp"
  $1054 = (FUNCTION_TABLE_ii[$load & 8191](461537)|0); //@line 3258 "libraries/glad/glad.cpp"
  HEAP32[165369] = $1054; //@line 3258 "libraries/glad/glad.cpp"
  $1055 = (FUNCTION_TABLE_ii[$load & 8191](461563)|0); //@line 3259 "libraries/glad/glad.cpp"
  HEAP32[164683] = $1055; //@line 3259 "libraries/glad/glad.cpp"
  $1056 = (FUNCTION_TABLE_ii[$load & 8191](461589)|0); //@line 3260 "libraries/glad/glad.cpp"
  HEAP32[164815] = $1056; //@line 3260 "libraries/glad/glad.cpp"
  $1057 = (FUNCTION_TABLE_ii[$load & 8191](461612)|0); //@line 3261 "libraries/glad/glad.cpp"
  HEAP32[165264] = $1057; //@line 3261 "libraries/glad/glad.cpp"
  $1058 = (FUNCTION_TABLE_ii[$load & 8191](461642)|0); //@line 3262 "libraries/glad/glad.cpp"
  HEAP32[165410] = $1058; //@line 3262 "libraries/glad/glad.cpp"
  $1059 = (FUNCTION_TABLE_ii[$load & 8191](461675)|0); //@line 3263 "libraries/glad/glad.cpp"
  HEAP32[165389] = $1059; //@line 3263 "libraries/glad/glad.cpp"
  $1060 = (FUNCTION_TABLE_ii[$load & 8191](461718)|0); //@line 3264 "libraries/glad/glad.cpp"
  HEAP32[165536] = $1060; //@line 3264 "libraries/glad/glad.cpp"
  $1061 = (FUNCTION_TABLE_ii[$load & 8191](461740)|0); //@line 3265 "libraries/glad/glad.cpp"
  HEAP32[164888] = $1061; //@line 3265 "libraries/glad/glad.cpp"
  $1062 = (FUNCTION_TABLE_ii[$load & 8191](461767)|0); //@line 3266 "libraries/glad/glad.cpp"
  HEAP32[165239] = $1062; //@line 3266 "libraries/glad/glad.cpp"
  $1063 = (FUNCTION_TABLE_ii[$load & 8191](461805)|0); //@line 3267 "libraries/glad/glad.cpp"
  HEAP32[164918] = $1063; //@line 3267 "libraries/glad/glad.cpp"
  $1064 = (FUNCTION_TABLE_ii[$load & 8191](461839)|0); //@line 3268 "libraries/glad/glad.cpp"
  HEAP32[164798] = $1064; //@line 3268 "libraries/glad/glad.cpp"
  $1065 = (FUNCTION_TABLE_ii[$load & 8191](461856)|0); //@line 3269 "libraries/glad/glad.cpp"
  HEAP32[165467] = $1065; //@line 3269 "libraries/glad/glad.cpp"
  $1066 = (FUNCTION_TABLE_ii[$load & 8191](461872)|0); //@line 3270 "libraries/glad/glad.cpp"
  HEAP32[164852] = $1066; //@line 3270 "libraries/glad/glad.cpp"
  $1067 = (FUNCTION_TABLE_ii[$load & 8191](461893)|0); //@line 3271 "libraries/glad/glad.cpp"
  HEAP32[164837] = $1067; //@line 3271 "libraries/glad/glad.cpp"
  $1068 = (FUNCTION_TABLE_ii[$load & 8191](461912)|0); //@line 3272 "libraries/glad/glad.cpp"
  HEAP32[165076] = $1068; //@line 3272 "libraries/glad/glad.cpp"
  $1069 = (FUNCTION_TABLE_ii[$load & 8191](461931)|0); //@line 3273 "libraries/glad/glad.cpp"
  HEAP32[165501] = $1069; //@line 3273 "libraries/glad/glad.cpp"
  $1070 = (FUNCTION_TABLE_ii[$load & 8191](461950)|0); //@line 3274 "libraries/glad/glad.cpp"
  HEAP32[164896] = $1070; //@line 3274 "libraries/glad/glad.cpp"
  $1071 = (FUNCTION_TABLE_ii[$load & 8191](461980)|0); //@line 3275 "libraries/glad/glad.cpp"
  HEAP32[164623] = $1071; //@line 3275 "libraries/glad/glad.cpp"
  $1072 = (FUNCTION_TABLE_ii[$load & 8191](462010)|0); //@line 3276 "libraries/glad/glad.cpp"
  HEAP32[165432] = $1072; //@line 3276 "libraries/glad/glad.cpp"
  $1073 = (FUNCTION_TABLE_ii[$load & 8191](462030)|0); //@line 3277 "libraries/glad/glad.cpp"
  HEAP32[165218] = $1073; //@line 3277 "libraries/glad/glad.cpp"
  $1074 = (FUNCTION_TABLE_ii[$load & 8191](462050)|0); //@line 3278 "libraries/glad/glad.cpp"
  HEAP32[164987] = $1074; //@line 3278 "libraries/glad/glad.cpp"
  $1075 = (FUNCTION_TABLE_ii[$load & 8191](462070)|0); //@line 3279 "libraries/glad/glad.cpp"
  HEAP32[165013] = $1075; //@line 3279 "libraries/glad/glad.cpp"
  $1076 = (FUNCTION_TABLE_ii[$load & 8191](462100)|0); //@line 3280 "libraries/glad/glad.cpp"
  HEAP32[165250] = $1076; //@line 3280 "libraries/glad/glad.cpp"
  $1077 = (FUNCTION_TABLE_ii[$load & 8191](462130)|0); //@line 3281 "libraries/glad/glad.cpp"
  HEAP32[165372] = $1077; //@line 3281 "libraries/glad/glad.cpp"
  $1078 = (FUNCTION_TABLE_ii[$load & 8191](462160)|0); //@line 3282 "libraries/glad/glad.cpp"
  HEAP32[165089] = $1078; //@line 3282 "libraries/glad/glad.cpp"
  $1079 = (FUNCTION_TABLE_ii[$load & 8191](462184)|0); //@line 3283 "libraries/glad/glad.cpp"
  HEAP32[165271] = $1079; //@line 3283 "libraries/glad/glad.cpp"
  $1080 = (FUNCTION_TABLE_ii[$load & 8191](462208)|0); //@line 3284 "libraries/glad/glad.cpp"
  HEAP32[164786] = $1080; //@line 3284 "libraries/glad/glad.cpp"
  $1081 = (FUNCTION_TABLE_ii[$load & 8191](462232)|0); //@line 3285 "libraries/glad/glad.cpp"
  HEAP32[164619] = $1081; //@line 3285 "libraries/glad/glad.cpp"
  $1082 = (FUNCTION_TABLE_ii[$load & 8191](462252)|0); //@line 3286 "libraries/glad/glad.cpp"
  HEAP32[164624] = $1082; //@line 3286 "libraries/glad/glad.cpp"
  $1083 = (FUNCTION_TABLE_ii[$load & 8191](462273)|0); //@line 3287 "libraries/glad/glad.cpp"
  HEAP32[164947] = $1083; //@line 3287 "libraries/glad/glad.cpp"
  $1084 = (FUNCTION_TABLE_ii[$load & 8191](462293)|0); //@line 3288 "libraries/glad/glad.cpp"
  HEAP32[165214] = $1084; //@line 3288 "libraries/glad/glad.cpp"
  $1085 = (FUNCTION_TABLE_ii[$load & 8191](462315)|0); //@line 3289 "libraries/glad/glad.cpp"
  HEAP32[165093] = $1085; //@line 3289 "libraries/glad/glad.cpp"
  $1086 = (FUNCTION_TABLE_ii[$load & 8191](462338)|0); //@line 3290 "libraries/glad/glad.cpp"
  HEAP32[165346] = $1086; //@line 3290 "libraries/glad/glad.cpp"
  $1087 = (FUNCTION_TABLE_ii[$load & 8191](462359)|0); //@line 3291 "libraries/glad/glad.cpp"
  HEAP32[165077] = $1087; //@line 3291 "libraries/glad/glad.cpp"
  $1088 = (FUNCTION_TABLE_ii[$load & 8191](462383)|0); //@line 3292 "libraries/glad/glad.cpp"
  HEAP32[164695] = $1088; //@line 3292 "libraries/glad/glad.cpp"
  $1089 = (FUNCTION_TABLE_ii[$load & 8191](462401)|0); //@line 3293 "libraries/glad/glad.cpp"
  HEAP32[164982] = $1089; //@line 3293 "libraries/glad/glad.cpp"
  $1090 = (FUNCTION_TABLE_ii[$load & 8191](462419)|0); //@line 3294 "libraries/glad/glad.cpp"
  HEAP32[164651] = $1090; //@line 3294 "libraries/glad/glad.cpp"
  $1091 = (FUNCTION_TABLE_ii[$load & 8191](462447)|0); //@line 3295 "libraries/glad/glad.cpp"
  HEAP32[165423] = $1091; //@line 3295 "libraries/glad/glad.cpp"
  $1092 = (FUNCTION_TABLE_ii[$load & 8191](462476)|0); //@line 3296 "libraries/glad/glad.cpp"
  HEAP32[165406] = $1092; //@line 3296 "libraries/glad/glad.cpp"
  $1093 = (FUNCTION_TABLE_ii[$load & 8191](462505)|0); //@line 3297 "libraries/glad/glad.cpp"
  HEAP32[165217] = $1093; //@line 3297 "libraries/glad/glad.cpp"
  $1094 = (FUNCTION_TABLE_ii[$load & 8191](462529)|0); //@line 3298 "libraries/glad/glad.cpp"
  HEAP32[164784] = $1094; //@line 3298 "libraries/glad/glad.cpp"
  $1095 = (FUNCTION_TABLE_ii[$load & 8191](462554)|0); //@line 3299 "libraries/glad/glad.cpp"
  HEAP32[165377] = $1095; //@line 3299 "libraries/glad/glad.cpp"
  $1096 = (FUNCTION_TABLE_ii[$load & 8191](462580)|0); //@line 3300 "libraries/glad/glad.cpp"
  HEAP32[165160] = $1096; //@line 3300 "libraries/glad/glad.cpp"
  $1097 = (FUNCTION_TABLE_ii[$load & 8191](462604)|0); //@line 3301 "libraries/glad/glad.cpp"
  HEAP32[165172] = $1097; //@line 3301 "libraries/glad/glad.cpp"
  $1098 = (FUNCTION_TABLE_ii[$load & 8191](462625)|0); //@line 3302 "libraries/glad/glad.cpp"
  HEAP32[164912] = $1098; //@line 3302 "libraries/glad/glad.cpp"
  $1099 = (FUNCTION_TABLE_ii[$load & 8191](462652)|0); //@line 3303 "libraries/glad/glad.cpp"
  HEAP32[165313] = $1099; //@line 3303 "libraries/glad/glad.cpp"
  $1100 = (FUNCTION_TABLE_ii[$load & 8191](462678)|0); //@line 3304 "libraries/glad/glad.cpp"
  HEAP32[164621] = $1100; //@line 3304 "libraries/glad/glad.cpp"
  $1101 = (FUNCTION_TABLE_ii[$load & 8191](462705)|0); //@line 3305 "libraries/glad/glad.cpp"
  HEAP32[164906] = $1101; //@line 3305 "libraries/glad/glad.cpp"
  $1102 = (FUNCTION_TABLE_ii[$load & 8191](462731)|0); //@line 3306 "libraries/glad/glad.cpp"
  HEAP32[165012] = $1102; //@line 3306 "libraries/glad/glad.cpp"
  $1103 = (FUNCTION_TABLE_ii[$load & 8191](462758)|0); //@line 3307 "libraries/glad/glad.cpp"
  HEAP32[164836] = $1103; //@line 3307 "libraries/glad/glad.cpp"
  $1104 = (FUNCTION_TABLE_ii[$load & 8191](462785)|0); //@line 3308 "libraries/glad/glad.cpp"
  HEAP32[165472] = $1104; //@line 3308 "libraries/glad/glad.cpp"
  $1105 = (FUNCTION_TABLE_ii[$load & 8191](462811)|0); //@line 3309 "libraries/glad/glad.cpp"
  HEAP32[164789] = $1105; //@line 3309 "libraries/glad/glad.cpp"
  $1106 = (FUNCTION_TABLE_ii[$load & 8191](462838)|0); //@line 3310 "libraries/glad/glad.cpp"
  HEAP32[164748] = $1106; //@line 3310 "libraries/glad/glad.cpp"
  $1107 = (FUNCTION_TABLE_ii[$load & 8191](462865)|0); //@line 3311 "libraries/glad/glad.cpp"
  HEAP32[164867] = $1107; //@line 3311 "libraries/glad/glad.cpp"
  $1108 = (FUNCTION_TABLE_ii[$load & 8191](462893)|0); //@line 3312 "libraries/glad/glad.cpp"
  HEAP32[165270] = $1108; //@line 3312 "libraries/glad/glad.cpp"
  $1109 = (FUNCTION_TABLE_ii[$load & 8191](462912)|0); //@line 3313 "libraries/glad/glad.cpp"
  HEAP32[164932] = $1109; //@line 3313 "libraries/glad/glad.cpp"
  $1110 = (FUNCTION_TABLE_ii[$load & 8191](462938)|0); //@line 3314 "libraries/glad/glad.cpp"
  HEAP32[165589] = $1110; //@line 3314 "libraries/glad/glad.cpp"
  $1111 = (FUNCTION_TABLE_ii[$load & 8191](462966)|0); //@line 3315 "libraries/glad/glad.cpp"
  HEAP32[165482] = $1111; //@line 3315 "libraries/glad/glad.cpp"
  $1112 = (FUNCTION_TABLE_ii[$load & 8191](462983)|0); //@line 3316 "libraries/glad/glad.cpp"
  HEAP32[165299] = $1112; //@line 3316 "libraries/glad/glad.cpp"
  $1113 = (FUNCTION_TABLE_ii[$load & 8191](463008)|0); //@line 3317 "libraries/glad/glad.cpp"
  HEAP32[164817] = $1113; //@line 3317 "libraries/glad/glad.cpp"
  $1114 = (FUNCTION_TABLE_ii[$load & 8191](463024)|0); //@line 3318 "libraries/glad/glad.cpp"
  HEAP32[164739] = $1114; //@line 3318 "libraries/glad/glad.cpp"
  $1115 = (FUNCTION_TABLE_ii[$load & 8191](463051)|0); //@line 3319 "libraries/glad/glad.cpp"
  HEAP32[165353] = $1115; //@line 3319 "libraries/glad/glad.cpp"
  $1116 = (FUNCTION_TABLE_ii[$load & 8191](463076)|0); //@line 3320 "libraries/glad/glad.cpp"
  HEAP32[165568] = $1116; //@line 3320 "libraries/glad/glad.cpp"
  $1117 = (FUNCTION_TABLE_ii[$load & 8191](463104)|0); //@line 3321 "libraries/glad/glad.cpp"
  HEAP32[164684] = $1117; //@line 3321 "libraries/glad/glad.cpp"
  $1118 = (FUNCTION_TABLE_ii[$load & 8191](463130)|0); //@line 3322 "libraries/glad/glad.cpp"
  HEAP32[164661] = $1118; //@line 3322 "libraries/glad/glad.cpp"
  $1119 = (FUNCTION_TABLE_ii[$load & 8191](463154)|0); //@line 3323 "libraries/glad/glad.cpp"
  HEAP32[165441] = $1119; //@line 3323 "libraries/glad/glad.cpp"
  $1120 = (FUNCTION_TABLE_ii[$load & 8191](463175)|0); //@line 3324 "libraries/glad/glad.cpp"
  HEAP32[165636] = $1120; //@line 3324 "libraries/glad/glad.cpp"
  $1121 = (FUNCTION_TABLE_ii[$load & 8191](463206)|0); //@line 3325 "libraries/glad/glad.cpp"
  HEAP32[164772] = $1121; //@line 3325 "libraries/glad/glad.cpp"
  $1122 = (FUNCTION_TABLE_ii[$load & 8191](463231)|0); //@line 3326 "libraries/glad/glad.cpp"
  HEAP32[165186] = $1122; //@line 3326 "libraries/glad/glad.cpp"
  $1123 = (FUNCTION_TABLE_ii[$load & 8191](463256)|0); //@line 3327 "libraries/glad/glad.cpp"
  HEAP32[165112] = $1123; //@line 3327 "libraries/glad/glad.cpp"
  $1124 = (FUNCTION_TABLE_ii[$load & 8191](463271)|0); //@line 3328 "libraries/glad/glad.cpp"
  HEAP32[164853] = $1124; //@line 3328 "libraries/glad/glad.cpp"
  $1125 = (FUNCTION_TABLE_ii[$load & 8191](463287)|0); //@line 3329 "libraries/glad/glad.cpp"
  HEAP32[165309] = $1125; //@line 3329 "libraries/glad/glad.cpp"
  $1126 = (FUNCTION_TABLE_ii[$load & 8191](463303)|0); //@line 3330 "libraries/glad/glad.cpp"
  HEAP32[165207] = $1126; //@line 3330 "libraries/glad/glad.cpp"
  $1127 = (FUNCTION_TABLE_ii[$load & 8191](463319)|0); //@line 3331 "libraries/glad/glad.cpp"
  HEAP32[165352] = $1127; //@line 3331 "libraries/glad/glad.cpp"
  $1128 = (FUNCTION_TABLE_ii[$load & 8191](463336)|0); //@line 3332 "libraries/glad/glad.cpp"
  HEAP32[165473] = $1128; //@line 3332 "libraries/glad/glad.cpp"
  $1129 = (FUNCTION_TABLE_ii[$load & 8191](463350)|0); //@line 3333 "libraries/glad/glad.cpp"
  HEAP32[165530] = $1129; //@line 3333 "libraries/glad/glad.cpp"
  $1130 = (FUNCTION_TABLE_ii[$load & 8191](463362)|0); //@line 3334 "libraries/glad/glad.cpp"
  HEAP32[164652] = $1130; //@line 3334 "libraries/glad/glad.cpp"
  $1131 = (FUNCTION_TABLE_ii[$load & 8191](463374)|0); //@line 3335 "libraries/glad/glad.cpp"
  HEAP32[164712] = $1131; //@line 3335 "libraries/glad/glad.cpp"
  $1132 = (FUNCTION_TABLE_ii[$load & 8191](463386)|0); //@line 3336 "libraries/glad/glad.cpp"
  HEAP32[165549] = $1132; //@line 3336 "libraries/glad/glad.cpp"
  $1133 = (FUNCTION_TABLE_ii[$load & 8191](463403)|0); //@line 3337 "libraries/glad/glad.cpp"
  HEAP32[165405] = $1133; //@line 3337 "libraries/glad/glad.cpp"
  $1134 = (FUNCTION_TABLE_ii[$load & 8191](463421)|0); //@line 3338 "libraries/glad/glad.cpp"
  HEAP32[164870] = $1134; //@line 3338 "libraries/glad/glad.cpp"
  $1135 = (FUNCTION_TABLE_ii[$load & 8191](463439)|0); //@line 3339 "libraries/glad/glad.cpp"
  HEAP32[164929] = $1135; //@line 3339 "libraries/glad/glad.cpp"
  $1136 = (FUNCTION_TABLE_ii[$load & 8191](463460)|0); //@line 3340 "libraries/glad/glad.cpp"
  HEAP32[165233] = $1136; //@line 3340 "libraries/glad/glad.cpp"
  $1137 = (FUNCTION_TABLE_ii[$load & 8191](463477)|0); //@line 3341 "libraries/glad/glad.cpp"
  HEAP32[165378] = $1137; //@line 3341 "libraries/glad/glad.cpp"
  $1138 = (FUNCTION_TABLE_ii[$load & 8191](463501)|0); //@line 3342 "libraries/glad/glad.cpp"
  HEAP32[165540] = $1138; //@line 3342 "libraries/glad/glad.cpp"
  $1139 = (FUNCTION_TABLE_ii[$load & 8191](463523)|0); //@line 3343 "libraries/glad/glad.cpp"
  HEAP32[165047] = $1139; //@line 3343 "libraries/glad/glad.cpp"
  $1140 = (FUNCTION_TABLE_ii[$load & 8191](463539)|0); //@line 3344 "libraries/glad/glad.cpp"
  HEAP32[164831] = $1140; //@line 3344 "libraries/glad/glad.cpp"
  $1141 = (FUNCTION_TABLE_ii[$load & 8191](463552)|0); //@line 3345 "libraries/glad/glad.cpp"
  HEAP32[165595] = $1141; //@line 3345 "libraries/glad/glad.cpp"
 }
 $1142 = HEAP8[1252012]|0; //@line 3350 "libraries/glad/glad.cpp"
 $1143 = ($1142<<24>>24)==(0); //@line 3350 "libraries/glad/glad.cpp"
 if (!($1143)) {
  $1144 = (FUNCTION_TABLE_ii[$load & 8191](616144)|0); //@line 3351 "libraries/glad/glad.cpp"
  HEAP32[165288] = $1144; //@line 3351 "libraries/glad/glad.cpp"
  $1145 = (FUNCTION_TABLE_ii[$load & 8191](613918)|0); //@line 3352 "libraries/glad/glad.cpp"
  HEAP32[165223] = $1145; //@line 3352 "libraries/glad/glad.cpp"
  $1146 = (FUNCTION_TABLE_ii[$load & 8191](615216)|0); //@line 3353 "libraries/glad/glad.cpp"
  HEAP32[165504] = $1146; //@line 3353 "libraries/glad/glad.cpp"
  $1147 = (FUNCTION_TABLE_ii[$load & 8191](614842)|0); //@line 3354 "libraries/glad/glad.cpp"
  HEAP32[165557] = $1147; //@line 3354 "libraries/glad/glad.cpp"
  $1148 = (FUNCTION_TABLE_ii[$load & 8191](615237)|0); //@line 3355 "libraries/glad/glad.cpp"
  HEAP32[165202] = $1148; //@line 3355 "libraries/glad/glad.cpp"
  $1149 = (FUNCTION_TABLE_ii[$load & 8191](614384)|0); //@line 3356 "libraries/glad/glad.cpp"
  HEAP32[165274] = $1149; //@line 3356 "libraries/glad/glad.cpp"
  $1150 = (FUNCTION_TABLE_ii[$load & 8191](614151)|0); //@line 3357 "libraries/glad/glad.cpp"
  HEAP32[164685] = $1150; //@line 3357 "libraries/glad/glad.cpp"
  $1151 = (FUNCTION_TABLE_ii[$load & 8191](616510)|0); //@line 3358 "libraries/glad/glad.cpp"
  HEAP32[165404] = $1151; //@line 3358 "libraries/glad/glad.cpp"
  $1152 = (FUNCTION_TABLE_ii[$load & 8191](616160)|0); //@line 3359 "libraries/glad/glad.cpp"
  HEAP32[165469] = $1152; //@line 3359 "libraries/glad/glad.cpp"
  $1153 = (FUNCTION_TABLE_ii[$load & 8191](616200)|0); //@line 3360 "libraries/glad/glad.cpp"
  HEAP32[165355] = $1153; //@line 3360 "libraries/glad/glad.cpp"
  $1154 = (FUNCTION_TABLE_ii[$load & 8191](616188)|0); //@line 3361 "libraries/glad/glad.cpp"
  HEAP32[165364] = $1154; //@line 3361 "libraries/glad/glad.cpp"
  $1155 = (FUNCTION_TABLE_ii[$load & 8191](616490)|0); //@line 3362 "libraries/glad/glad.cpp"
  HEAP32[164680] = $1155; //@line 3362 "libraries/glad/glad.cpp"
  $1156 = (FUNCTION_TABLE_ii[$load & 8191](614303)|0); //@line 3363 "libraries/glad/glad.cpp"
  HEAP32[165565] = $1156; //@line 3363 "libraries/glad/glad.cpp"
  $1157 = (FUNCTION_TABLE_ii[$load & 8191](614316)|0); //@line 3364 "libraries/glad/glad.cpp"
  HEAP32[165513] = $1157; //@line 3364 "libraries/glad/glad.cpp"
  $1158 = (FUNCTION_TABLE_ii[$load & 8191](616102)|0); //@line 3365 "libraries/glad/glad.cpp"
  HEAP32[165498] = $1158; //@line 3365 "libraries/glad/glad.cpp"
  $1159 = (FUNCTION_TABLE_ii[$load & 8191](616041)|0); //@line 3366 "libraries/glad/glad.cpp"
  HEAP32[164768] = $1159; //@line 3366 "libraries/glad/glad.cpp"
  $1160 = (FUNCTION_TABLE_ii[$load & 8191](616416)|0); //@line 3367 "libraries/glad/glad.cpp"
  HEAP32[165395] = $1160; //@line 3367 "libraries/glad/glad.cpp"
  $1161 = (FUNCTION_TABLE_ii[$load & 8191](615973)|0); //@line 3368 "libraries/glad/glad.cpp"
  HEAP32[165437] = $1161; //@line 3368 "libraries/glad/glad.cpp"
  $1162 = (FUNCTION_TABLE_ii[$load & 8191](616061)|0); //@line 3369 "libraries/glad/glad.cpp"
  HEAP32[164916] = $1162; //@line 3369 "libraries/glad/glad.cpp"
  $1163 = (FUNCTION_TABLE_ii[$load & 8191](616456)|0); //@line 3370 "libraries/glad/glad.cpp"
  HEAP32[165466] = $1163; //@line 3370 "libraries/glad/glad.cpp"
  $1164 = (FUNCTION_TABLE_ii[$load & 8191](615034)|0); //@line 3371 "libraries/glad/glad.cpp"
  HEAP32[164646] = $1164; //@line 3371 "libraries/glad/glad.cpp"
  $1165 = (FUNCTION_TABLE_ii[$load & 8191](614060)|0); //@line 3372 "libraries/glad/glad.cpp"
  HEAP32[164678] = $1165; //@line 3372 "libraries/glad/glad.cpp"
  $1166 = (FUNCTION_TABLE_ii[$load & 8191](614083)|0); //@line 3373 "libraries/glad/glad.cpp"
  HEAP32[165449] = $1166; //@line 3373 "libraries/glad/glad.cpp"
  $1167 = (FUNCTION_TABLE_ii[$load & 8191](616582)|0); //@line 3374 "libraries/glad/glad.cpp"
  HEAP32[164967] = $1167; //@line 3374 "libraries/glad/glad.cpp"
  $1168 = (FUNCTION_TABLE_ii[$load & 8191](616599)|0); //@line 3375 "libraries/glad/glad.cpp"
  HEAP32[164677] = $1168; //@line 3375 "libraries/glad/glad.cpp"
  $1169 = (FUNCTION_TABLE_ii[$load & 8191](613819)|0); //@line 3376 "libraries/glad/glad.cpp"
  HEAP32[165365] = $1169; //@line 3376 "libraries/glad/glad.cpp"
  $1170 = (FUNCTION_TABLE_ii[$load & 8191](613888)|0); //@line 3377 "libraries/glad/glad.cpp"
  HEAP32[164672] = $1170; //@line 3377 "libraries/glad/glad.cpp"
  $1171 = (FUNCTION_TABLE_ii[$load & 8191](616030)|0); //@line 3378 "libraries/glad/glad.cpp"
  HEAP32[165205] = $1171; //@line 3378 "libraries/glad/glad.cpp"
  $1172 = (FUNCTION_TABLE_ii[$load & 8191](614264)|0); //@line 3379 "libraries/glad/glad.cpp"
  HEAP32[165381] = $1172; //@line 3379 "libraries/glad/glad.cpp"
  $1173 = (FUNCTION_TABLE_ii[$load & 8191](615273)|0); //@line 3380 "libraries/glad/glad.cpp"
  HEAP32[164758] = $1173; //@line 3380 "libraries/glad/glad.cpp"
  $1174 = (FUNCTION_TABLE_ii[$load & 8191](615109)|0); //@line 3381 "libraries/glad/glad.cpp"
  HEAP32[164697] = $1174; //@line 3381 "libraries/glad/glad.cpp"
  $1175 = (FUNCTION_TABLE_ii[$load & 8191](614362)|0); //@line 3382 "libraries/glad/glad.cpp"
  HEAP32[164936] = $1175; //@line 3382 "libraries/glad/glad.cpp"
  $1176 = (FUNCTION_TABLE_ii[$load & 8191](614965)|0); //@line 3383 "libraries/glad/glad.cpp"
  HEAP32[165085] = $1176; //@line 3383 "libraries/glad/glad.cpp"
  $1177 = (FUNCTION_TABLE_ii[$load & 8191](614043)|0); //@line 3384 "libraries/glad/glad.cpp"
  HEAP32[165157] = $1177; //@line 3384 "libraries/glad/glad.cpp"
  $1178 = (FUNCTION_TABLE_ii[$load & 8191](615987)|0); //@line 3385 "libraries/glad/glad.cpp"
  HEAP32[165448] = $1178; //@line 3385 "libraries/glad/glad.cpp"
  $1179 = (FUNCTION_TABLE_ii[$load & 8191](616076)|0); //@line 3386 "libraries/glad/glad.cpp"
  HEAP32[165644] = $1179; //@line 3386 "libraries/glad/glad.cpp"
  $1180 = (FUNCTION_TABLE_ii[$load & 8191](616237)|0); //@line 3387 "libraries/glad/glad.cpp"
  HEAP32[164657] = $1180; //@line 3387 "libraries/glad/glad.cpp"
  $1181 = (FUNCTION_TABLE_ii[$load & 8191](613948)|0); //@line 3388 "libraries/glad/glad.cpp"
  HEAP32[165143] = $1181; //@line 3388 "libraries/glad/glad.cpp"
  $1182 = (FUNCTION_TABLE_ii[$load & 8191](616008)|0); //@line 3389 "libraries/glad/glad.cpp"
  HEAP32[165193] = $1182; //@line 3389 "libraries/glad/glad.cpp"
  $1183 = (FUNCTION_TABLE_ii[$load & 8191](615770)|0); //@line 3390 "libraries/glad/glad.cpp"
  HEAP32[164811] = $1183; //@line 3390 "libraries/glad/glad.cpp"
  $1184 = (FUNCTION_TABLE_ii[$load & 8191](615797)|0); //@line 3391 "libraries/glad/glad.cpp"
  HEAP32[164759] = $1184; //@line 3391 "libraries/glad/glad.cpp"
  $1185 = (FUNCTION_TABLE_ii[$load & 8191](615810)|0); //@line 3392 "libraries/glad/glad.cpp"
  HEAP32[165086] = $1185; //@line 3392 "libraries/glad/glad.cpp"
  $1186 = (FUNCTION_TABLE_ii[$load & 8191](615999)|0); //@line 3393 "libraries/glad/glad.cpp"
  HEAP32[164832] = $1186; //@line 3393 "libraries/glad/glad.cpp"
  $1187 = (FUNCTION_TABLE_ii[$load & 8191](615744)|0); //@line 3394 "libraries/glad/glad.cpp"
  HEAP32[165289] = $1187; //@line 3394 "libraries/glad/glad.cpp"
  $1188 = (FUNCTION_TABLE_ii[$load & 8191](615943)|0); //@line 3395 "libraries/glad/glad.cpp"
  HEAP32[165082] = $1188; //@line 3395 "libraries/glad/glad.cpp"
  $1189 = (FUNCTION_TABLE_ii[$load & 8191](615952)|0); //@line 3396 "libraries/glad/glad.cpp"
  HEAP32[164883] = $1189; //@line 3396 "libraries/glad/glad.cpp"
  $1190 = (FUNCTION_TABLE_ii[$load & 8191](615294)|0); //@line 3397 "libraries/glad/glad.cpp"
  HEAP32[164630] = $1190; //@line 3397 "libraries/glad/glad.cpp"
  $1191 = (FUNCTION_TABLE_ii[$load & 8191](615320)|0); //@line 3398 "libraries/glad/glad.cpp"
  HEAP32[164777] = $1191; //@line 3398 "libraries/glad/glad.cpp"
  $1192 = (FUNCTION_TABLE_ii[$load & 8191](616018)|0); //@line 3399 "libraries/glad/glad.cpp"
  HEAP32[165653] = $1192; //@line 3399 "libraries/glad/glad.cpp"
  $1193 = (FUNCTION_TABLE_ii[$load & 8191](614251)|0); //@line 3400 "libraries/glad/glad.cpp"
  HEAP32[165357] = $1193; //@line 3400 "libraries/glad/glad.cpp"
  $1194 = (FUNCTION_TABLE_ii[$load & 8191](616127)|0); //@line 3401 "libraries/glad/glad.cpp"
  HEAP32[165249] = $1194; //@line 3401 "libraries/glad/glad.cpp"
  $1195 = (FUNCTION_TABLE_ii[$load & 8191](615255)|0); //@line 3402 "libraries/glad/glad.cpp"
  HEAP32[164797] = $1195; //@line 3402 "libraries/glad/glad.cpp"
  $1196 = (FUNCTION_TABLE_ii[$load & 8191](614343)|0); //@line 3403 "libraries/glad/glad.cpp"
  HEAP32[164676] = $1196; //@line 3403 "libraries/glad/glad.cpp"
  $1197 = (FUNCTION_TABLE_ii[$load & 8191](614029)|0); //@line 3404 "libraries/glad/glad.cpp"
  HEAP32[164922] = $1197; //@line 3404 "libraries/glad/glad.cpp"
  $1198 = (FUNCTION_TABLE_ii[$load & 8191](614947)|0); //@line 3405 "libraries/glad/glad.cpp"
  HEAP32[165189] = $1198; //@line 3405 "libraries/glad/glad.cpp"
  $1199 = (FUNCTION_TABLE_ii[$load & 8191](614566)|0); //@line 3406 "libraries/glad/glad.cpp"
  HEAP32[165651] = $1199; //@line 3406 "libraries/glad/glad.cpp"
  $1200 = (FUNCTION_TABLE_ii[$load & 8191](614980)|0); //@line 3407 "libraries/glad/glad.cpp"
  HEAP32[164803] = $1200; //@line 3407 "libraries/glad/glad.cpp"
  $1201 = (FUNCTION_TABLE_ii[$load & 8191](614927)|0); //@line 3408 "libraries/glad/glad.cpp"
  HEAP32[164909] = $1201; //@line 3408 "libraries/glad/glad.cpp"
  $1202 = (FUNCTION_TABLE_ii[$load & 8191](614015)|0); //@line 3409 "libraries/glad/glad.cpp"
  HEAP32[165084] = $1202; //@line 3409 "libraries/glad/glad.cpp"
  $1203 = (FUNCTION_TABLE_ii[$load & 8191](614280)|0); //@line 3410 "libraries/glad/glad.cpp"
  HEAP32[165063] = $1203; //@line 3410 "libraries/glad/glad.cpp"
  $1204 = (FUNCTION_TABLE_ii[$load & 8191](615864)|0); //@line 3411 "libraries/glad/glad.cpp"
  HEAP32[165571] = $1204; //@line 3411 "libraries/glad/glad.cpp"
  $1205 = (FUNCTION_TABLE_ii[$load & 8191](614003)|0); //@line 3412 "libraries/glad/glad.cpp"
  HEAP32[165575] = $1205; //@line 3412 "libraries/glad/glad.cpp"
  $1206 = (FUNCTION_TABLE_ii[$load & 8191](615343)|0); //@line 3413 "libraries/glad/glad.cpp"
  HEAP32[164778] = $1206; //@line 3413 "libraries/glad/glad.cpp"
  $1207 = (FUNCTION_TABLE_ii[$load & 8191](613989)|0); //@line 3414 "libraries/glad/glad.cpp"
  HEAP32[165587] = $1207; //@line 3414 "libraries/glad/glad.cpp"
  $1208 = (FUNCTION_TABLE_ii[$load & 8191](615083)|0); //@line 3415 "libraries/glad/glad.cpp"
  HEAP32[165573] = $1208; //@line 3415 "libraries/glad/glad.cpp"
  $1209 = (FUNCTION_TABLE_ii[$load & 8191](615166)|0); //@line 3416 "libraries/glad/glad.cpp"
  HEAP32[165061] = $1209; //@line 3416 "libraries/glad/glad.cpp"
  $1210 = (FUNCTION_TABLE_ii[$load & 8191](614403)|0); //@line 3417 "libraries/glad/glad.cpp"
  HEAP32[164884] = $1210; //@line 3417 "libraries/glad/glad.cpp"
  $1211 = (FUNCTION_TABLE_ii[$load & 8191](615069)|0); //@line 3418 "libraries/glad/glad.cpp"
  HEAP32[164990] = $1211; //@line 3418 "libraries/glad/glad.cpp"
  $1212 = (FUNCTION_TABLE_ii[$load & 8191](615050)|0); //@line 3419 "libraries/glad/glad.cpp"
  HEAP32[164795] = $1212; //@line 3419 "libraries/glad/glad.cpp"
  $1213 = (FUNCTION_TABLE_ii[$load & 8191](615125)|0); //@line 3420 "libraries/glad/glad.cpp"
  HEAP32[164900] = $1213; //@line 3420 "libraries/glad/glad.cpp"
  $1214 = (FUNCTION_TABLE_ii[$load & 8191](615016)|0); //@line 3421 "libraries/glad/glad.cpp"
  HEAP32[165110] = $1214; //@line 3421 "libraries/glad/glad.cpp"
  $1215 = (FUNCTION_TABLE_ii[$load & 8191](613977)|0); //@line 3422 "libraries/glad/glad.cpp"
  HEAP32[165140] = $1215; //@line 3422 "libraries/glad/glad.cpp"
  $1216 = (FUNCTION_TABLE_ii[$load & 8191](614165)|0); //@line 3423 "libraries/glad/glad.cpp"
  HEAP32[165456] = $1216; //@line 3423 "libraries/glad/glad.cpp"
  $1217 = (FUNCTION_TABLE_ii[$load & 8191](614185)|0); //@line 3424 "libraries/glad/glad.cpp"
  HEAP32[165170] = $1217; //@line 3424 "libraries/glad/glad.cpp"
  $1218 = (FUNCTION_TABLE_ii[$load & 8191](614449)|0); //@line 3425 "libraries/glad/glad.cpp"
  HEAP32[165303] = $1218; //@line 3425 "libraries/glad/glad.cpp"
  $1219 = (FUNCTION_TABLE_ii[$load & 8191](614464)|0); //@line 3426 "libraries/glad/glad.cpp"
  HEAP32[165009] = $1219; //@line 3426 "libraries/glad/glad.cpp"
  $1220 = (FUNCTION_TABLE_ii[$load & 8191](614479)|0); //@line 3427 "libraries/glad/glad.cpp"
  HEAP32[165470] = $1220; //@line 3427 "libraries/glad/glad.cpp"
  $1221 = (FUNCTION_TABLE_ii[$load & 8191](614500)|0); //@line 3428 "libraries/glad/glad.cpp"
  HEAP32[165182] = $1221; //@line 3428 "libraries/glad/glad.cpp"
  $1222 = (FUNCTION_TABLE_ii[$load & 8191](614520)|0); //@line 3429 "libraries/glad/glad.cpp"
  HEAP32[165038] = $1222; //@line 3429 "libraries/glad/glad.cpp"
  $1223 = (FUNCTION_TABLE_ii[$load & 8191](614540)|0); //@line 3430 "libraries/glad/glad.cpp"
  HEAP32[164890] = $1223; //@line 3430 "libraries/glad/glad.cpp"
  $1224 = (FUNCTION_TABLE_ii[$load & 8191](616273)|0); //@line 3431 "libraries/glad/glad.cpp"
  HEAP32[165134] = $1224; //@line 3431 "libraries/glad/glad.cpp"
  $1225 = (FUNCTION_TABLE_ii[$load & 8191](614332)|0); //@line 3432 "libraries/glad/glad.cpp"
  HEAP32[164673] = $1225; //@line 3432 "libraries/glad/glad.cpp"
  $1226 = (FUNCTION_TABLE_ii[$load & 8191](616176)|0); //@line 3433 "libraries/glad/glad.cpp"
  HEAP32[164774] = $1226; //@line 3433 "libraries/glad/glad.cpp"
  $1227 = (FUNCTION_TABLE_ii[$load & 8191](615381)|0); //@line 3434 "libraries/glad/glad.cpp"
  HEAP32[165368] = $1227; //@line 3434 "libraries/glad/glad.cpp"
  $1228 = (FUNCTION_TABLE_ii[$load & 8191](615204)|0); //@line 3435 "libraries/glad/glad.cpp"
  HEAP32[165280] = $1228; //@line 3435 "libraries/glad/glad.cpp"
  $1229 = (FUNCTION_TABLE_ii[$load & 8191](614432)|0); //@line 3436 "libraries/glad/glad.cpp"
  HEAP32[164804] = $1229; //@line 3436 "libraries/glad/glad.cpp"
  $1230 = (FUNCTION_TABLE_ii[$load & 8191](615098)|0); //@line 3437 "libraries/glad/glad.cpp"
  HEAP32[164825] = $1230; //@line 3437 "libraries/glad/glad.cpp"
  $1231 = (FUNCTION_TABLE_ii[$load & 8191](614239)|0); //@line 3438 "libraries/glad/glad.cpp"
  HEAP32[165607] = $1231; //@line 3438 "libraries/glad/glad.cpp"
  $1232 = (FUNCTION_TABLE_ii[$load & 8191](616049)|0); //@line 3439 "libraries/glad/glad.cpp"
  HEAP32[164641] = $1232; //@line 3439 "libraries/glad/glad.cpp"
  $1233 = (FUNCTION_TABLE_ii[$load & 8191](615152)|0); //@line 3440 "libraries/glad/glad.cpp"
  HEAP32[165133] = $1233; //@line 3440 "libraries/glad/glad.cpp"
  $1234 = (FUNCTION_TABLE_ii[$load & 8191](613963)|0); //@line 3441 "libraries/glad/glad.cpp"
  HEAP32[165018] = $1234; //@line 3441 "libraries/glad/glad.cpp"
  $1235 = (FUNCTION_TABLE_ii[$load & 8191](616280)|0); //@line 3442 "libraries/glad/glad.cpp"
  HEAP32[164859] = $1235; //@line 3442 "libraries/glad/glad.cpp"
  $1236 = (FUNCTION_TABLE_ii[$load & 8191](614138)|0); //@line 3443 "libraries/glad/glad.cpp"
  HEAP32[165294] = $1236; //@line 3443 "libraries/glad/glad.cpp"
  $1237 = (FUNCTION_TABLE_ii[$load & 8191](615840)|0); //@line 3444 "libraries/glad/glad.cpp"
  HEAP32[165543] = $1237; //@line 3444 "libraries/glad/glad.cpp"
  $1238 = (FUNCTION_TABLE_ii[$load & 8191](616468)|0); //@line 3445 "libraries/glad/glad.cpp"
  HEAP32[164704] = $1238; //@line 3445 "libraries/glad/glad.cpp"
  $1239 = (FUNCTION_TABLE_ii[$load & 8191](616313)|0); //@line 3446 "libraries/glad/glad.cpp"
  HEAP32[165176] = $1239; //@line 3446 "libraries/glad/glad.cpp"
  $1240 = (FUNCTION_TABLE_ii[$load & 8191](616429)|0); //@line 3447 "libraries/glad/glad.cpp"
  HEAP32[165383] = $1240; //@line 3447 "libraries/glad/glad.cpp"
  $1241 = (FUNCTION_TABLE_ii[$load & 8191](615825)|0); //@line 3448 "libraries/glad/glad.cpp"
  HEAP32[165231] = $1241; //@line 3448 "libraries/glad/glad.cpp"
  $1242 = (FUNCTION_TABLE_ii[$load & 8191](615001)|0); //@line 3449 "libraries/glad/glad.cpp"
  HEAP32[165539] = $1242; //@line 3449 "libraries/glad/glad.cpp"
  $1243 = (FUNCTION_TABLE_ii[$load & 8191](616379)|0); //@line 3450 "libraries/glad/glad.cpp"
  HEAP32[165322] = $1243; //@line 3450 "libraries/glad/glad.cpp"
  $1244 = (FUNCTION_TABLE_ii[$load & 8191](616523)|0); //@line 3451 "libraries/glad/glad.cpp"
  HEAP32[164911] = $1244; //@line 3451 "libraries/glad/glad.cpp"
  $1245 = (FUNCTION_TABLE_ii[$load & 8191](616088)|0); //@line 3452 "libraries/glad/glad.cpp"
  HEAP32[165603] = $1245; //@line 3452 "libraries/glad/glad.cpp"
  $1246 = (FUNCTION_TABLE_ii[$load & 8191](616251)|0); //@line 3453 "libraries/glad/glad.cpp"
  HEAP32[164654] = $1246; //@line 3453 "libraries/glad/glad.cpp"
  $1247 = (FUNCTION_TABLE_ii[$load & 8191](616393)|0); //@line 3454 "libraries/glad/glad.cpp"
  HEAP32[164775] = $1247; //@line 3454 "libraries/glad/glad.cpp"
  $1248 = (FUNCTION_TABLE_ii[$load & 8191](616545)|0); //@line 3455 "libraries/glad/glad.cpp"
  HEAP32[165159] = $1248; //@line 3455 "libraries/glad/glad.cpp"
  $1249 = (FUNCTION_TABLE_ii[$load & 8191](614109)|0); //@line 3456 "libraries/glad/glad.cpp"
  HEAP32[165602] = $1249; //@line 3456 "libraries/glad/glad.cpp"
  $1250 = (FUNCTION_TABLE_ii[$load & 8191](616346)|0); //@line 3457 "libraries/glad/glad.cpp"
  HEAP32[165106] = $1250; //@line 3457 "libraries/glad/glad.cpp"
  $1251 = (FUNCTION_TABLE_ii[$load & 8191](614205)|0); //@line 3458 "libraries/glad/glad.cpp"
  HEAP32[165617] = $1251; //@line 3458 "libraries/glad/glad.cpp"
  $1252 = (FUNCTION_TABLE_ii[$load & 8191](616330)|0); //@line 3459 "libraries/glad/glad.cpp"
  HEAP32[165108] = $1252; //@line 3459 "libraries/glad/glad.cpp"
  $1253 = (FUNCTION_TABLE_ii[$load & 8191](614222)|0); //@line 3460 "libraries/glad/glad.cpp"
  HEAP32[164788] = $1253; //@line 3460 "libraries/glad/glad.cpp"
  $1254 = (FUNCTION_TABLE_ii[$load & 8191](614122)|0); //@line 3461 "libraries/glad/glad.cpp"
  HEAP32[165192] = $1254; //@line 3461 "libraries/glad/glad.cpp"
  $1255 = (FUNCTION_TABLE_ii[$load & 8191](614585)|0); //@line 3462 "libraries/glad/glad.cpp"
  HEAP32[165181] = $1255; //@line 3462 "libraries/glad/glad.cpp"
  $1256 = (FUNCTION_TABLE_ii[$load & 8191](614733)|0); //@line 3463 "libraries/glad/glad.cpp"
  HEAP32[165610] = $1256; //@line 3463 "libraries/glad/glad.cpp"
  $1257 = (FUNCTION_TABLE_ii[$load & 8191](614633)|0); //@line 3464 "libraries/glad/glad.cpp"
  HEAP32[165188] = $1257; //@line 3464 "libraries/glad/glad.cpp"
  $1258 = (FUNCTION_TABLE_ii[$load & 8191](614681)|0); //@line 3465 "libraries/glad/glad.cpp"
  HEAP32[165316] = $1258; //@line 3465 "libraries/glad/glad.cpp"
  $1259 = (FUNCTION_TABLE_ii[$load & 8191](614597)|0); //@line 3466 "libraries/glad/glad.cpp"
  HEAP32[165553] = $1259; //@line 3466 "libraries/glad/glad.cpp"
  $1260 = (FUNCTION_TABLE_ii[$load & 8191](614746)|0); //@line 3467 "libraries/glad/glad.cpp"
  HEAP32[165418] = $1260; //@line 3467 "libraries/glad/glad.cpp"
  $1261 = (FUNCTION_TABLE_ii[$load & 8191](614645)|0); //@line 3468 "libraries/glad/glad.cpp"
  HEAP32[164937] = $1261; //@line 3468 "libraries/glad/glad.cpp"
  $1262 = (FUNCTION_TABLE_ii[$load & 8191](614694)|0); //@line 3469 "libraries/glad/glad.cpp"
  HEAP32[164977] = $1262; //@line 3469 "libraries/glad/glad.cpp"
  $1263 = (FUNCTION_TABLE_ii[$load & 8191](614609)|0); //@line 3470 "libraries/glad/glad.cpp"
  HEAP32[164716] = $1263; //@line 3470 "libraries/glad/glad.cpp"
  $1264 = (FUNCTION_TABLE_ii[$load & 8191](614759)|0); //@line 3471 "libraries/glad/glad.cpp"
  HEAP32[165017] = $1264; //@line 3471 "libraries/glad/glad.cpp"
  $1265 = (FUNCTION_TABLE_ii[$load & 8191](614657)|0); //@line 3472 "libraries/glad/glad.cpp"
  HEAP32[165190] = $1265; //@line 3472 "libraries/glad/glad.cpp"
  $1266 = (FUNCTION_TABLE_ii[$load & 8191](614707)|0); //@line 3473 "libraries/glad/glad.cpp"
  HEAP32[165043] = $1266; //@line 3473 "libraries/glad/glad.cpp"
  $1267 = (FUNCTION_TABLE_ii[$load & 8191](614621)|0); //@line 3474 "libraries/glad/glad.cpp"
  HEAP32[165297] = $1267; //@line 3474 "libraries/glad/glad.cpp"
  $1268 = (FUNCTION_TABLE_ii[$load & 8191](614772)|0); //@line 3475 "libraries/glad/glad.cpp"
  HEAP32[165650] = $1268; //@line 3475 "libraries/glad/glad.cpp"
  $1269 = (FUNCTION_TABLE_ii[$load & 8191](614669)|0); //@line 3476 "libraries/glad/glad.cpp"
  HEAP32[165287] = $1269; //@line 3476 "libraries/glad/glad.cpp"
  $1270 = (FUNCTION_TABLE_ii[$load & 8191](614720)|0); //@line 3477 "libraries/glad/glad.cpp"
  HEAP32[164915] = $1270; //@line 3477 "libraries/glad/glad.cpp"
  $1271 = (FUNCTION_TABLE_ii[$load & 8191](614785)|0); //@line 3478 "libraries/glad/glad.cpp"
  HEAP32[165563] = $1271; //@line 3478 "libraries/glad/glad.cpp"
  $1272 = (FUNCTION_TABLE_ii[$load & 8191](614804)|0); //@line 3479 "libraries/glad/glad.cpp"
  HEAP32[165618] = $1272; //@line 3479 "libraries/glad/glad.cpp"
  $1273 = (FUNCTION_TABLE_ii[$load & 8191](614823)|0); //@line 3480 "libraries/glad/glad.cpp"
  HEAP32[165649] = $1273; //@line 3480 "libraries/glad/glad.cpp"
  $1274 = (FUNCTION_TABLE_ii[$load & 8191](613854)|0); //@line 3481 "libraries/glad/glad.cpp"
  HEAP32[165058] = $1274; //@line 3481 "libraries/glad/glad.cpp"
  $1275 = (FUNCTION_TABLE_ii[$load & 8191](615186)|0); //@line 3482 "libraries/glad/glad.cpp"
  HEAP32[165115] = $1275; //@line 3482 "libraries/glad/glad.cpp"
  $1276 = (FUNCTION_TABLE_ii[$load & 8191](616296)|0); //@line 3483 "libraries/glad/glad.cpp"
  HEAP32[164679] = $1276; //@line 3483 "libraries/glad/glad.cpp"
  $1277 = (FUNCTION_TABLE_ii[$load & 8191](614855)|0); //@line 3484 "libraries/glad/glad.cpp"
  HEAP32[164978] = $1277; //@line 3484 "libraries/glad/glad.cpp"
  $1278 = (FUNCTION_TABLE_ii[$load & 8191](616362)|0); //@line 3485 "libraries/glad/glad.cpp"
  HEAP32[165641] = $1278; //@line 3485 "libraries/glad/glad.cpp"
  $1279 = (FUNCTION_TABLE_ii[$load & 8191](614873)|0); //@line 3486 "libraries/glad/glad.cpp"
  HEAP32[165579] = $1279; //@line 3486 "libraries/glad/glad.cpp"
  $1280 = (FUNCTION_TABLE_ii[$load & 8191](616439)|0); //@line 3487 "libraries/glad/glad.cpp"
  HEAP32[164736] = $1280; //@line 3487 "libraries/glad/glad.cpp"
  $1281 = (FUNCTION_TABLE_ii[$load & 8191](614891)|0); //@line 3488 "libraries/glad/glad.cpp"
  HEAP32[165041] = $1281; //@line 3488 "libraries/glad/glad.cpp"
  $1282 = (FUNCTION_TABLE_ii[$load & 8191](616565)|0); //@line 3489 "libraries/glad/glad.cpp"
  HEAP32[165163] = $1282; //@line 3489 "libraries/glad/glad.cpp"
  $1283 = (FUNCTION_TABLE_ii[$load & 8191](614909)|0); //@line 3490 "libraries/glad/glad.cpp"
  HEAP32[165284] = $1283; //@line 3490 "libraries/glad/glad.cpp"
  $1284 = (FUNCTION_TABLE_ii[$load & 8191](615722)|0); //@line 3491 "libraries/glad/glad.cpp"
  HEAP32[164942] = $1284; //@line 3491 "libraries/glad/glad.cpp"
  $1285 = (FUNCTION_TABLE_ii[$load & 8191](616405)|0); //@line 3492 "libraries/glad/glad.cpp"
  HEAP32[165094] = $1285; //@line 3492 "libraries/glad/glad.cpp"
 }
 $1286 = HEAP8[1252013]|0; //@line 3497 "libraries/glad/glad.cpp"
 $1287 = ($1286<<24>>24)==(0); //@line 3497 "libraries/glad/glad.cpp"
 if (!($1287)) {
  $1288 = (FUNCTION_TABLE_ii[$load & 8191](447783)|0); //@line 3498 "libraries/glad/glad.cpp"
  HEAP32[165245] = $1288; //@line 3498 "libraries/glad/glad.cpp"
  $1289 = (FUNCTION_TABLE_ii[$load & 8191](615476)|0); //@line 3499 "libraries/glad/glad.cpp"
  HEAP32[165510] = $1289; //@line 3499 "libraries/glad/glad.cpp"
  $1290 = (FUNCTION_TABLE_ii[$load & 8191](451186)|0); //@line 3500 "libraries/glad/glad.cpp"
  HEAP32[165367] = $1290; //@line 3500 "libraries/glad/glad.cpp"
  $1291 = (FUNCTION_TABLE_ii[$load & 8191](451199)|0); //@line 3501 "libraries/glad/glad.cpp"
  HEAP32[164963] = $1291; //@line 3501 "libraries/glad/glad.cpp"
  $1292 = (FUNCTION_TABLE_ii[$load & 8191](451215)|0); //@line 3502 "libraries/glad/glad.cpp"
  HEAP32[165098] = $1292; //@line 3502 "libraries/glad/glad.cpp"
  $1293 = (FUNCTION_TABLE_ii[$load & 8191](451235)|0); //@line 3503 "libraries/glad/glad.cpp"
  HEAP32[165031] = $1293; //@line 3503 "libraries/glad/glad.cpp"
  $1294 = (FUNCTION_TABLE_ii[$load & 8191](451281)|0); //@line 3504 "libraries/glad/glad.cpp"
  HEAP32[164633] = $1294; //@line 3504 "libraries/glad/glad.cpp"
  $1295 = (FUNCTION_TABLE_ii[$load & 8191](452795)|0); //@line 3505 "libraries/glad/glad.cpp"
  HEAP32[164959] = $1295; //@line 3505 "libraries/glad/glad.cpp"
  $1296 = (FUNCTION_TABLE_ii[$load & 8191](452808)|0); //@line 3506 "libraries/glad/glad.cpp"
  HEAP32[165161] = $1296; //@line 3506 "libraries/glad/glad.cpp"
  $1297 = (FUNCTION_TABLE_ii[$load & 8191](452824)|0); //@line 3507 "libraries/glad/glad.cpp"
  HEAP32[165596] = $1297; //@line 3507 "libraries/glad/glad.cpp"
  $1298 = (FUNCTION_TABLE_ii[$load & 8191](452834)|0); //@line 3508 "libraries/glad/glad.cpp"
  HEAP32[165555] = $1298; //@line 3508 "libraries/glad/glad.cpp"
  $1299 = (FUNCTION_TABLE_ii[$load & 8191](452847)|0); //@line 3509 "libraries/glad/glad.cpp"
  HEAP32[165146] = $1299; //@line 3509 "libraries/glad/glad.cpp"
  $1300 = (FUNCTION_TABLE_ii[$load & 8191](452858)|0); //@line 3510 "libraries/glad/glad.cpp"
  HEAP32[164813] = $1300; //@line 3510 "libraries/glad/glad.cpp"
  $1301 = (FUNCTION_TABLE_ii[$load & 8191](452890)|0); //@line 3511 "libraries/glad/glad.cpp"
  HEAP32[164776] = $1301; //@line 3511 "libraries/glad/glad.cpp"
  $1302 = (FUNCTION_TABLE_ii[$load & 8191](452941)|0); //@line 3512 "libraries/glad/glad.cpp"
  HEAP32[165265] = $1302; //@line 3512 "libraries/glad/glad.cpp"
  $1303 = (FUNCTION_TABLE_ii[$load & 8191](452955)|0); //@line 3513 "libraries/glad/glad.cpp"
  HEAP32[165590] = $1303; //@line 3513 "libraries/glad/glad.cpp"
  $1304 = (FUNCTION_TABLE_ii[$load & 8191](616619)|0); //@line 3514 "libraries/glad/glad.cpp"
  HEAP32[165046] = $1304; //@line 3514 "libraries/glad/glad.cpp"
  $1305 = (FUNCTION_TABLE_ii[$load & 8191](453504)|0); //@line 3515 "libraries/glad/glad.cpp"
  HEAP32[165225] = $1305; //@line 3515 "libraries/glad/glad.cpp"
  $1306 = (FUNCTION_TABLE_ii[$load & 8191](453525)|0); //@line 3516 "libraries/glad/glad.cpp"
  HEAP32[164692] = $1306; //@line 3516 "libraries/glad/glad.cpp"
  $1307 = (FUNCTION_TABLE_ii[$load & 8191](453546)|0); //@line 3517 "libraries/glad/glad.cpp"
  HEAP32[165065] = $1307; //@line 3517 "libraries/glad/glad.cpp"
  $1308 = (FUNCTION_TABLE_ii[$load & 8191](453567)|0); //@line 3518 "libraries/glad/glad.cpp"
  HEAP32[165505] = $1308; //@line 3518 "libraries/glad/glad.cpp"
  $1309 = (FUNCTION_TABLE_ii[$load & 8191](453588)|0); //@line 3519 "libraries/glad/glad.cpp"
  HEAP32[164743] = $1309; //@line 3519 "libraries/glad/glad.cpp"
  $1310 = (FUNCTION_TABLE_ii[$load & 8191](453609)|0); //@line 3520 "libraries/glad/glad.cpp"
  HEAP32[164705] = $1310; //@line 3520 "libraries/glad/glad.cpp"
  $1311 = (FUNCTION_TABLE_ii[$load & 8191](454732)|0); //@line 3521 "libraries/glad/glad.cpp"
  HEAP32[164725] = $1311; //@line 3521 "libraries/glad/glad.cpp"
  $1312 = (FUNCTION_TABLE_ii[$load & 8191](454750)|0); //@line 3522 "libraries/glad/glad.cpp"
  HEAP32[165298] = $1312; //@line 3522 "libraries/glad/glad.cpp"
  $1313 = (FUNCTION_TABLE_ii[$load & 8191](454783)|0); //@line 3523 "libraries/glad/glad.cpp"
  HEAP32[164944] = $1313; //@line 3523 "libraries/glad/glad.cpp"
  $1314 = (FUNCTION_TABLE_ii[$load & 8191](454809)|0); //@line 3524 "libraries/glad/glad.cpp"
  HEAP32[165517] = $1314; //@line 3524 "libraries/glad/glad.cpp"
  $1315 = (FUNCTION_TABLE_ii[$load & 8191](454826)|0); //@line 3525 "libraries/glad/glad.cpp"
  HEAP32[164950] = $1315; //@line 3525 "libraries/glad/glad.cpp"
  $1316 = (FUNCTION_TABLE_ii[$load & 8191](615642)|0); //@line 3526 "libraries/glad/glad.cpp"
  HEAP32[165066] = $1316; //@line 3526 "libraries/glad/glad.cpp"
  $1317 = (FUNCTION_TABLE_ii[$load & 8191](615621)|0); //@line 3527 "libraries/glad/glad.cpp"
  HEAP32[164893] = $1317; //@line 3527 "libraries/glad/glad.cpp"
  $1318 = (FUNCTION_TABLE_ii[$load & 8191](615603)|0); //@line 3528 "libraries/glad/glad.cpp"
  HEAP32[165300] = $1318; //@line 3528 "libraries/glad/glad.cpp"
  $1319 = (FUNCTION_TABLE_ii[$load & 8191](454851)|0); //@line 3529 "libraries/glad/glad.cpp"
  HEAP32[164810] = $1319; //@line 3529 "libraries/glad/glad.cpp"
  $1320 = (FUNCTION_TABLE_ii[$load & 8191](453659)|0); //@line 3530 "libraries/glad/glad.cpp"
  HEAP32[164643] = $1320; //@line 3530 "libraries/glad/glad.cpp"
  $1321 = (FUNCTION_TABLE_ii[$load & 8191](453709)|0); //@line 3531 "libraries/glad/glad.cpp"
  HEAP32[165485] = $1321; //@line 3531 "libraries/glad/glad.cpp"
  $1322 = (FUNCTION_TABLE_ii[$load & 8191](453734)|0); //@line 3532 "libraries/glad/glad.cpp"
  HEAP32[165475] = $1322; //@line 3532 "libraries/glad/glad.cpp"
  $1323 = (FUNCTION_TABLE_ii[$load & 8191](453757)|0); //@line 3533 "libraries/glad/glad.cpp"
  HEAP32[165430] = $1323; //@line 3533 "libraries/glad/glad.cpp"
  $1324 = (FUNCTION_TABLE_ii[$load & 8191](453775)|0); //@line 3534 "libraries/glad/glad.cpp"
  HEAP32[165512] = $1324; //@line 3534 "libraries/glad/glad.cpp"
  $1325 = (FUNCTION_TABLE_ii[$load & 8191](453792)|0); //@line 3535 "libraries/glad/glad.cpp"
  HEAP32[165096] = $1325; //@line 3535 "libraries/glad/glad.cpp"
  $1326 = (FUNCTION_TABLE_ii[$load & 8191](453820)|0); //@line 3536 "libraries/glad/glad.cpp"
  HEAP32[164647] = $1326; //@line 3536 "libraries/glad/glad.cpp"
  $1327 = (FUNCTION_TABLE_ii[$load & 8191](453911)|0); //@line 3537 "libraries/glad/glad.cpp"
  HEAP32[165363] = $1327; //@line 3537 "libraries/glad/glad.cpp"
  $1328 = (FUNCTION_TABLE_ii[$load & 8191](453934)|0); //@line 3538 "libraries/glad/glad.cpp"
  HEAP32[165591] = $1328; //@line 3538 "libraries/glad/glad.cpp"
  $1329 = (FUNCTION_TABLE_ii[$load & 8191](453955)|0); //@line 3539 "libraries/glad/glad.cpp"
  HEAP32[165438] = $1329; //@line 3539 "libraries/glad/glad.cpp"
  $1330 = (FUNCTION_TABLE_ii[$load & 8191](454031)|0); //@line 3540 "libraries/glad/glad.cpp"
  HEAP32[165348] = $1330; //@line 3540 "libraries/glad/glad.cpp"
  $1331 = (FUNCTION_TABLE_ii[$load & 8191](454106)|0); //@line 3541 "libraries/glad/glad.cpp"
  HEAP32[165463] = $1331; //@line 3541 "libraries/glad/glad.cpp"
  $1332 = (FUNCTION_TABLE_ii[$load & 8191](454182)|0); //@line 3542 "libraries/glad/glad.cpp"
  HEAP32[165337] = $1332; //@line 3542 "libraries/glad/glad.cpp"
  $1333 = (FUNCTION_TABLE_ii[$load & 8191](454261)|0); //@line 3543 "libraries/glad/glad.cpp"
  HEAP32[165415] = $1333; //@line 3543 "libraries/glad/glad.cpp"
  $1334 = (FUNCTION_TABLE_ii[$load & 8191](454359)|0); //@line 3544 "libraries/glad/glad.cpp"
  HEAP32[164860] = $1334; //@line 3544 "libraries/glad/glad.cpp"
  $1335 = (FUNCTION_TABLE_ii[$load & 8191](454398)|0); //@line 3545 "libraries/glad/glad.cpp"
  HEAP32[164783] = $1335; //@line 3545 "libraries/glad/glad.cpp"
  $1336 = (FUNCTION_TABLE_ii[$load & 8191](454420)|0); //@line 3546 "libraries/glad/glad.cpp"
  HEAP32[164760] = $1336; //@line 3546 "libraries/glad/glad.cpp"
  $1337 = (FUNCTION_TABLE_ii[$load & 8191](454433)|0); //@line 3547 "libraries/glad/glad.cpp"
  HEAP32[165528] = $1337; //@line 3547 "libraries/glad/glad.cpp"
  $1338 = (FUNCTION_TABLE_ii[$load & 8191](454446)|0); //@line 3548 "libraries/glad/glad.cpp"
  HEAP32[165412] = $1338; //@line 3548 "libraries/glad/glad.cpp"
  $1339 = (FUNCTION_TABLE_ii[$load & 8191](454459)|0); //@line 3549 "libraries/glad/glad.cpp"
  HEAP32[165199] = $1339; //@line 3549 "libraries/glad/glad.cpp"
  $1340 = (FUNCTION_TABLE_ii[$load & 8191](454472)|0); //@line 3550 "libraries/glad/glad.cpp"
  HEAP32[165095] = $1340; //@line 3550 "libraries/glad/glad.cpp"
  $1341 = (FUNCTION_TABLE_ii[$load & 8191](454486)|0); //@line 3551 "libraries/glad/glad.cpp"
  HEAP32[165080] = $1341; //@line 3551 "libraries/glad/glad.cpp"
  $1342 = (FUNCTION_TABLE_ii[$load & 8191](454500)|0); //@line 3552 "libraries/glad/glad.cpp"
  HEAP32[165117] = $1342; //@line 3552 "libraries/glad/glad.cpp"
  $1343 = (FUNCTION_TABLE_ii[$load & 8191](454514)|0); //@line 3553 "libraries/glad/glad.cpp"
  HEAP32[164980] = $1343; //@line 3553 "libraries/glad/glad.cpp"
  $1344 = (FUNCTION_TABLE_ii[$load & 8191](454608)|0); //@line 3554 "libraries/glad/glad.cpp"
  HEAP32[165400] = $1344; //@line 3554 "libraries/glad/glad.cpp"
  $1345 = (FUNCTION_TABLE_ii[$load & 8191](454624)|0); //@line 3555 "libraries/glad/glad.cpp"
  HEAP32[165661] = $1345; //@line 3555 "libraries/glad/glad.cpp"
  $1346 = (FUNCTION_TABLE_ii[$load & 8191](454641)|0); //@line 3556 "libraries/glad/glad.cpp"
  HEAP32[164707] = $1346; //@line 3556 "libraries/glad/glad.cpp"
  $1347 = (FUNCTION_TABLE_ii[$load & 8191](454657)|0); //@line 3557 "libraries/glad/glad.cpp"
  HEAP32[164722] = $1347; //@line 3557 "libraries/glad/glad.cpp"
  $1348 = (FUNCTION_TABLE_ii[$load & 8191](454673)|0); //@line 3558 "libraries/glad/glad.cpp"
  HEAP32[165216] = $1348; //@line 3558 "libraries/glad/glad.cpp"
  $1349 = (FUNCTION_TABLE_ii[$load & 8191](454903)|0); //@line 3559 "libraries/glad/glad.cpp"
  HEAP32[165637] = $1349; //@line 3559 "libraries/glad/glad.cpp"
  $1350 = (FUNCTION_TABLE_ii[$load & 8191](454923)|0); //@line 3560 "libraries/glad/glad.cpp"
  HEAP32[164824] = $1350; //@line 3560 "libraries/glad/glad.cpp"
  $1351 = (FUNCTION_TABLE_ii[$load & 8191](454943)|0); //@line 3561 "libraries/glad/glad.cpp"
  HEAP32[164833] = $1351; //@line 3561 "libraries/glad/glad.cpp"
  $1352 = (FUNCTION_TABLE_ii[$load & 8191](454988)|0); //@line 3562 "libraries/glad/glad.cpp"
  HEAP32[165556] = $1352; //@line 3562 "libraries/glad/glad.cpp"
  $1353 = (FUNCTION_TABLE_ii[$load & 8191](455011)|0); //@line 3563 "libraries/glad/glad.cpp"
  HEAP32[165048] = $1353; //@line 3563 "libraries/glad/glad.cpp"
  $1354 = (FUNCTION_TABLE_ii[$load & 8191](455037)|0); //@line 3564 "libraries/glad/glad.cpp"
  HEAP32[165541] = $1354; //@line 3564 "libraries/glad/glad.cpp"
  $1355 = (FUNCTION_TABLE_ii[$load & 8191](455065)|0); //@line 3565 "libraries/glad/glad.cpp"
  HEAP32[165325] = $1355; //@line 3565 "libraries/glad/glad.cpp"
  $1356 = (FUNCTION_TABLE_ii[$load & 8191](615897)|0); //@line 3566 "libraries/glad/glad.cpp"
  HEAP32[165248] = $1356; //@line 3566 "libraries/glad/glad.cpp"
  $1357 = (FUNCTION_TABLE_ii[$load & 8191](615919)|0); //@line 3567 "libraries/glad/glad.cpp"
  HEAP32[164873] = $1357; //@line 3567 "libraries/glad/glad.cpp"
  $1358 = (FUNCTION_TABLE_ii[$load & 8191](455224)|0); //@line 3568 "libraries/glad/glad.cpp"
  HEAP32[164892] = $1358; //@line 3568 "libraries/glad/glad.cpp"
  $1359 = (FUNCTION_TABLE_ii[$load & 8191](455236)|0); //@line 3569 "libraries/glad/glad.cpp"
  HEAP32[164933] = $1359; //@line 3569 "libraries/glad/glad.cpp"
  $1360 = (FUNCTION_TABLE_ii[$load & 8191](455245)|0); //@line 3570 "libraries/glad/glad.cpp"
  HEAP32[165033] = $1360; //@line 3570 "libraries/glad/glad.cpp"
  $1361 = (FUNCTION_TABLE_ii[$load & 8191](455258)|0); //@line 3571 "libraries/glad/glad.cpp"
  HEAP32[165462] = $1361; //@line 3571 "libraries/glad/glad.cpp"
  $1362 = (FUNCTION_TABLE_ii[$load & 8191](455275)|0); //@line 3572 "libraries/glad/glad.cpp"
  HEAP32[165171] = $1362; //@line 3572 "libraries/glad/glad.cpp"
  $1363 = (FUNCTION_TABLE_ii[$load & 8191](455286)|0); //@line 3573 "libraries/glad/glad.cpp"
  HEAP32[164829] = $1363; //@line 3573 "libraries/glad/glad.cpp"
  $1364 = (FUNCTION_TABLE_ii[$load & 8191](455302)|0); //@line 3574 "libraries/glad/glad.cpp"
  HEAP32[165547] = $1364; //@line 3574 "libraries/glad/glad.cpp"
  $1365 = (FUNCTION_TABLE_ii[$load & 8191](455314)|0); //@line 3575 "libraries/glad/glad.cpp"
  HEAP32[164964] = $1365; //@line 3575 "libraries/glad/glad.cpp"
  $1366 = (FUNCTION_TABLE_ii[$load & 8191](455332)|0); //@line 3576 "libraries/glad/glad.cpp"
  HEAP32[165167] = $1366; //@line 3576 "libraries/glad/glad.cpp"
  $1367 = (FUNCTION_TABLE_ii[$load & 8191](455508)|0); //@line 3577 "libraries/glad/glad.cpp"
  HEAP32[164913] = $1367; //@line 3577 "libraries/glad/glad.cpp"
  $1368 = (FUNCTION_TABLE_ii[$load & 8191](455522)|0); //@line 3578 "libraries/glad/glad.cpp"
  HEAP32[164709] = $1368; //@line 3578 "libraries/glad/glad.cpp"
  $1369 = (FUNCTION_TABLE_ii[$load & 8191](455539)|0); //@line 3579 "libraries/glad/glad.cpp"
  HEAP32[165489] = $1369; //@line 3579 "libraries/glad/glad.cpp"
  $1370 = (FUNCTION_TABLE_ii[$load & 8191](455551)|0); //@line 3580 "libraries/glad/glad.cpp"
  HEAP32[164640] = $1370; //@line 3580 "libraries/glad/glad.cpp"
  $1371 = (FUNCTION_TABLE_ii[$load & 8191](455565)|0); //@line 3581 "libraries/glad/glad.cpp"
  HEAP32[165177] = $1371; //@line 3581 "libraries/glad/glad.cpp"
  $1372 = (FUNCTION_TABLE_ii[$load & 8191](455585)|0); //@line 3582 "libraries/glad/glad.cpp"
  HEAP32[165255] = $1372; //@line 3582 "libraries/glad/glad.cpp"
  $1373 = (FUNCTION_TABLE_ii[$load & 8191](455606)|0); //@line 3583 "libraries/glad/glad.cpp"
  HEAP32[165179] = $1373; //@line 3583 "libraries/glad/glad.cpp"
  $1374 = (FUNCTION_TABLE_ii[$load & 8191](455626)|0); //@line 3584 "libraries/glad/glad.cpp"
  HEAP32[165305] = $1374; //@line 3584 "libraries/glad/glad.cpp"
  $1375 = (FUNCTION_TABLE_ii[$load & 8191](455692)|0); //@line 3585 "libraries/glad/glad.cpp"
  HEAP32[165471] = $1375; //@line 3585 "libraries/glad/glad.cpp"
  $1376 = (FUNCTION_TABLE_ii[$load & 8191](455741)|0); //@line 3586 "libraries/glad/glad.cpp"
  HEAP32[164818] = $1376; //@line 3586 "libraries/glad/glad.cpp"
  $1377 = (FUNCTION_TABLE_ii[$load & 8191](615875)|0); //@line 3587 "libraries/glad/glad.cpp"
  HEAP32[165491] = $1377; //@line 3587 "libraries/glad/glad.cpp"
  $1378 = (FUNCTION_TABLE_ii[$load & 8191](457180)|0); //@line 3588 "libraries/glad/glad.cpp"
  HEAP32[165640] = $1378; //@line 3588 "libraries/glad/glad.cpp"
  $1379 = (FUNCTION_TABLE_ii[$load & 8191](457204)|0); //@line 3589 "libraries/glad/glad.cpp"
  HEAP32[165495] = $1379; //@line 3589 "libraries/glad/glad.cpp"
  $1380 = (FUNCTION_TABLE_ii[$load & 8191](457231)|0); //@line 3590 "libraries/glad/glad.cpp"
  HEAP32[165443] = $1380; //@line 3590 "libraries/glad/glad.cpp"
  $1381 = (FUNCTION_TABLE_ii[$load & 8191](457255)|0); //@line 3591 "libraries/glad/glad.cpp"
  HEAP32[165286] = $1381; //@line 3591 "libraries/glad/glad.cpp"
  $1382 = (FUNCTION_TABLE_ii[$load & 8191](457277)|0); //@line 3592 "libraries/glad/glad.cpp"
  HEAP32[164726] = $1382; //@line 3592 "libraries/glad/glad.cpp"
  $1383 = (FUNCTION_TABLE_ii[$load & 8191](457302)|0); //@line 3593 "libraries/glad/glad.cpp"
  HEAP32[165417] = $1383; //@line 3593 "libraries/glad/glad.cpp"
  $1384 = (FUNCTION_TABLE_ii[$load & 8191](457440)|0); //@line 3594 "libraries/glad/glad.cpp"
  HEAP32[165057] = $1384; //@line 3594 "libraries/glad/glad.cpp"
  $1385 = (FUNCTION_TABLE_ii[$load & 8191](457459)|0); //@line 3595 "libraries/glad/glad.cpp"
  HEAP32[165488] = $1385; //@line 3595 "libraries/glad/glad.cpp"
  $1386 = (FUNCTION_TABLE_ii[$load & 8191](457475)|0); //@line 3596 "libraries/glad/glad.cpp"
  HEAP32[164764] = $1386; //@line 3596 "libraries/glad/glad.cpp"
  $1387 = (FUNCTION_TABLE_ii[$load & 8191](459811)|0); //@line 3597 "libraries/glad/glad.cpp"
  HEAP32[164995] = $1387; //@line 3597 "libraries/glad/glad.cpp"
  $1388 = (FUNCTION_TABLE_ii[$load & 8191](459835)|0); //@line 3598 "libraries/glad/glad.cpp"
  HEAP32[165022] = $1388; //@line 3598 "libraries/glad/glad.cpp"
  $1389 = (FUNCTION_TABLE_ii[$load & 8191](459437)|0); //@line 3599 "libraries/glad/glad.cpp"
  HEAP32[165612] = $1389; //@line 3599 "libraries/glad/glad.cpp"
  $1390 = (FUNCTION_TABLE_ii[$load & 8191](459452)|0); //@line 3600 "libraries/glad/glad.cpp"
  HEAP32[165339] = $1390; //@line 3600 "libraries/glad/glad.cpp"
  $1391 = (FUNCTION_TABLE_ii[$load & 8191](459332)|0); //@line 3601 "libraries/glad/glad.cpp"
  HEAP32[164979] = $1391; //@line 3601 "libraries/glad/glad.cpp"
 }
 $1392 = HEAP8[1252014]|0; //@line 3606 "libraries/glad/glad.cpp"
 $1393 = ($1392<<24>>24)==(0); //@line 3606 "libraries/glad/glad.cpp"
 if (!($1393)) {
  $1394 = (FUNCTION_TABLE_ii[$load & 8191](459578)|0); //@line 3607 "libraries/glad/glad.cpp"
  HEAP32[164957] = $1394; //@line 3607 "libraries/glad/glad.cpp"
  $1395 = (FUNCTION_TABLE_ii[$load & 8191](459596)|0); //@line 3608 "libraries/glad/glad.cpp"
  HEAP32[165427] = $1395; //@line 3608 "libraries/glad/glad.cpp"
  $1396 = (FUNCTION_TABLE_ii[$load & 8191](456589)|0); //@line 3609 "libraries/glad/glad.cpp"
  HEAP32[164723] = $1396; //@line 3609 "libraries/glad/glad.cpp"
  $1397 = (FUNCTION_TABLE_ii[$load & 8191](456610)|0); //@line 3610 "libraries/glad/glad.cpp"
  HEAP32[165422] = $1397; //@line 3610 "libraries/glad/glad.cpp"
  $1398 = (FUNCTION_TABLE_ii[$load & 8191](459641)|0); //@line 3611 "libraries/glad/glad.cpp"
  HEAP32[165149] = $1398; //@line 3611 "libraries/glad/glad.cpp"
  $1399 = (FUNCTION_TABLE_ii[$load & 8191](459665)|0); //@line 3612 "libraries/glad/glad.cpp"
  HEAP32[165632] = $1399; //@line 3612 "libraries/glad/glad.cpp"
  $1400 = (FUNCTION_TABLE_ii[$load & 8191](459916)|0); //@line 3613 "libraries/glad/glad.cpp"
  HEAP32[164843] = $1400; //@line 3613 "libraries/glad/glad.cpp"
  $1401 = (FUNCTION_TABLE_ii[$load & 8191](459940)|0); //@line 3614 "libraries/glad/glad.cpp"
  HEAP32[165420] = $1401; //@line 3614 "libraries/glad/glad.cpp"
  $1402 = (FUNCTION_TABLE_ii[$load & 8191](459966)|0); //@line 3615 "libraries/glad/glad.cpp"
  HEAP32[165150] = $1402; //@line 3615 "libraries/glad/glad.cpp"
  $1403 = (FUNCTION_TABLE_ii[$load & 8191](459991)|0); //@line 3616 "libraries/glad/glad.cpp"
  HEAP32[165663] = $1403; //@line 3616 "libraries/glad/glad.cpp"
  $1404 = (FUNCTION_TABLE_ii[$load & 8191](460014)|0); //@line 3617 "libraries/glad/glad.cpp"
  HEAP32[165155] = $1404; //@line 3617 "libraries/glad/glad.cpp"
  $1405 = (FUNCTION_TABLE_ii[$load & 8191](457495)|0); //@line 3618 "libraries/glad/glad.cpp"
  HEAP32[165329] = $1405; //@line 3618 "libraries/glad/glad.cpp"
  $1406 = (FUNCTION_TABLE_ii[$load & 8191](457514)|0); //@line 3619 "libraries/glad/glad.cpp"
  HEAP32[164742] = $1406; //@line 3619 "libraries/glad/glad.cpp"
  $1407 = (FUNCTION_TABLE_ii[$load & 8191](457536)|0); //@line 3620 "libraries/glad/glad.cpp"
  HEAP32[165317] = $1407; //@line 3620 "libraries/glad/glad.cpp"
  $1408 = (FUNCTION_TABLE_ii[$load & 8191](457559)|0); //@line 3621 "libraries/glad/glad.cpp"
  HEAP32[165382] = $1408; //@line 3621 "libraries/glad/glad.cpp"
  $1409 = (FUNCTION_TABLE_ii[$load & 8191](457581)|0); //@line 3622 "libraries/glad/glad.cpp"
  HEAP32[165660] = $1409; //@line 3622 "libraries/glad/glad.cpp"
  $1410 = (FUNCTION_TABLE_ii[$load & 8191](457606)|0); //@line 3623 "libraries/glad/glad.cpp"
  HEAP32[165122] = $1410; //@line 3623 "libraries/glad/glad.cpp"
  $1411 = (FUNCTION_TABLE_ii[$load & 8191](457628)|0); //@line 3624 "libraries/glad/glad.cpp"
  HEAP32[165291] = $1411; //@line 3624 "libraries/glad/glad.cpp"
  $1412 = (FUNCTION_TABLE_ii[$load & 8191](457648)|0); //@line 3625 "libraries/glad/glad.cpp"
  HEAP32[165338] = $1412; //@line 3625 "libraries/glad/glad.cpp"
  $1413 = (FUNCTION_TABLE_ii[$load & 8191](457671)|0); //@line 3626 "libraries/glad/glad.cpp"
  HEAP32[164876] = $1413; //@line 3626 "libraries/glad/glad.cpp"
  $1414 = (FUNCTION_TABLE_ii[$load & 8191](457829)|0); //@line 3627 "libraries/glad/glad.cpp"
  HEAP32[165552] = $1414; //@line 3627 "libraries/glad/glad.cpp"
  $1415 = (FUNCTION_TABLE_ii[$load & 8191](457987)|0); //@line 3628 "libraries/glad/glad.cpp"
  HEAP32[165144] = $1415; //@line 3628 "libraries/glad/glad.cpp"
  $1416 = (FUNCTION_TABLE_ii[$load & 8191](458145)|0); //@line 3629 "libraries/glad/glad.cpp"
  HEAP32[165209] = $1416; //@line 3629 "libraries/glad/glad.cpp"
  $1417 = (FUNCTION_TABLE_ii[$load & 8191](457788)|0); //@line 3630 "libraries/glad/glad.cpp"
  HEAP32[164943] = $1417; //@line 3630 "libraries/glad/glad.cpp"
  $1418 = (FUNCTION_TABLE_ii[$load & 8191](457946)|0); //@line 3631 "libraries/glad/glad.cpp"
  HEAP32[165598] = $1418; //@line 3631 "libraries/glad/glad.cpp"
  $1419 = (FUNCTION_TABLE_ii[$load & 8191](458104)|0); //@line 3632 "libraries/glad/glad.cpp"
  HEAP32[164855] = $1419; //@line 3632 "libraries/glad/glad.cpp"
  $1420 = (FUNCTION_TABLE_ii[$load & 8191](458262)|0); //@line 3633 "libraries/glad/glad.cpp"
  HEAP32[165599] = $1420; //@line 3633 "libraries/glad/glad.cpp"
  $1421 = (FUNCTION_TABLE_ii[$load & 8191](457710)|0); //@line 3634 "libraries/glad/glad.cpp"
  HEAP32[164879] = $1421; //@line 3634 "libraries/glad/glad.cpp"
  $1422 = (FUNCTION_TABLE_ii[$load & 8191](457868)|0); //@line 3635 "libraries/glad/glad.cpp"
  HEAP32[165139] = $1422; //@line 3635 "libraries/glad/glad.cpp"
  $1423 = (FUNCTION_TABLE_ii[$load & 8191](458026)|0); //@line 3636 "libraries/glad/glad.cpp"
  HEAP32[165334] = $1423; //@line 3636 "libraries/glad/glad.cpp"
  $1424 = (FUNCTION_TABLE_ii[$load & 8191](458184)|0); //@line 3637 "libraries/glad/glad.cpp"
  HEAP32[165211] = $1424; //@line 3637 "libraries/glad/glad.cpp"
  $1425 = (FUNCTION_TABLE_ii[$load & 8191](457690)|0); //@line 3638 "libraries/glad/glad.cpp"
  HEAP32[165269] = $1425; //@line 3638 "libraries/glad/glad.cpp"
  $1426 = (FUNCTION_TABLE_ii[$load & 8191](457848)|0); //@line 3639 "libraries/glad/glad.cpp"
  HEAP32[164812] = $1426; //@line 3639 "libraries/glad/glad.cpp"
  $1427 = (FUNCTION_TABLE_ii[$load & 8191](458006)|0); //@line 3640 "libraries/glad/glad.cpp"
  HEAP32[164886] = $1427; //@line 3640 "libraries/glad/glad.cpp"
  $1428 = (FUNCTION_TABLE_ii[$load & 8191](458164)|0); //@line 3641 "libraries/glad/glad.cpp"
  HEAP32[165053] = $1428; //@line 3641 "libraries/glad/glad.cpp"
  $1429 = (FUNCTION_TABLE_ii[$load & 8191](457808)|0); //@line 3642 "libraries/glad/glad.cpp"
  HEAP32[164808] = $1429; //@line 3642 "libraries/glad/glad.cpp"
  $1430 = (FUNCTION_TABLE_ii[$load & 8191](457966)|0); //@line 3643 "libraries/glad/glad.cpp"
  HEAP32[165548] = $1430; //@line 3643 "libraries/glad/glad.cpp"
  $1431 = (FUNCTION_TABLE_ii[$load & 8191](458124)|0); //@line 3644 "libraries/glad/glad.cpp"
  HEAP32[165413] = $1431; //@line 3644 "libraries/glad/glad.cpp"
  $1432 = (FUNCTION_TABLE_ii[$load & 8191](458282)|0); //@line 3645 "libraries/glad/glad.cpp"
  HEAP32[165198] = $1432; //@line 3645 "libraries/glad/glad.cpp"
  $1433 = (FUNCTION_TABLE_ii[$load & 8191](457729)|0); //@line 3646 "libraries/glad/glad.cpp"
  HEAP32[165648] = $1433; //@line 3646 "libraries/glad/glad.cpp"
  $1434 = (FUNCTION_TABLE_ii[$load & 8191](457887)|0); //@line 3647 "libraries/glad/glad.cpp"
  HEAP32[164951] = $1434; //@line 3647 "libraries/glad/glad.cpp"
  $1435 = (FUNCTION_TABLE_ii[$load & 8191](458045)|0); //@line 3648 "libraries/glad/glad.cpp"
  HEAP32[165408] = $1435; //@line 3648 "libraries/glad/glad.cpp"
  $1436 = (FUNCTION_TABLE_ii[$load & 8191](458203)|0); //@line 3649 "libraries/glad/glad.cpp"
  HEAP32[165593] = $1436; //@line 3649 "libraries/glad/glad.cpp"
  $1437 = (FUNCTION_TABLE_ii[$load & 8191](458303)|0); //@line 3650 "libraries/glad/glad.cpp"
  HEAP32[165251] = $1437; //@line 3650 "libraries/glad/glad.cpp"
  $1438 = (FUNCTION_TABLE_ii[$load & 8191](458329)|0); //@line 3651 "libraries/glad/glad.cpp"
  HEAP32[164993] = $1438; //@line 3651 "libraries/glad/glad.cpp"
  $1439 = (FUNCTION_TABLE_ii[$load & 8191](458355)|0); //@line 3652 "libraries/glad/glad.cpp"
  HEAP32[164737] = $1439; //@line 3652 "libraries/glad/glad.cpp"
  $1440 = (FUNCTION_TABLE_ii[$load & 8191](458459)|0); //@line 3653 "libraries/glad/glad.cpp"
  HEAP32[165399] = $1440; //@line 3653 "libraries/glad/glad.cpp"
  $1441 = (FUNCTION_TABLE_ii[$load & 8191](458487)|0); //@line 3654 "libraries/glad/glad.cpp"
  HEAP32[164731] = $1441; //@line 3654 "libraries/glad/glad.cpp"
  $1442 = (FUNCTION_TABLE_ii[$load & 8191](458515)|0); //@line 3655 "libraries/glad/glad.cpp"
  HEAP32[165535] = $1442; //@line 3655 "libraries/glad/glad.cpp"
  $1443 = (FUNCTION_TABLE_ii[$load & 8191](458543)|0); //@line 3656 "libraries/glad/glad.cpp"
  HEAP32[165520] = $1443; //@line 3656 "libraries/glad/glad.cpp"
  $1444 = (FUNCTION_TABLE_ii[$load & 8191](458571)|0); //@line 3657 "libraries/glad/glad.cpp"
  HEAP32[165021] = $1444; //@line 3657 "libraries/glad/glad.cpp"
  $1445 = (FUNCTION_TABLE_ii[$load & 8191](458599)|0); //@line 3658 "libraries/glad/glad.cpp"
  HEAP32[165623] = $1445; //@line 3658 "libraries/glad/glad.cpp"
  $1446 = (FUNCTION_TABLE_ii[$load & 8191](458795)|0); //@line 3659 "libraries/glad/glad.cpp"
  HEAP32[164901] = $1446; //@line 3659 "libraries/glad/glad.cpp"
  $1447 = (FUNCTION_TABLE_ii[$load & 8191](458821)|0); //@line 3660 "libraries/glad/glad.cpp"
  HEAP32[165019] = $1447; //@line 3660 "libraries/glad/glad.cpp"
  $1448 = (FUNCTION_TABLE_ii[$load & 8191](459387)|0); //@line 3661 "libraries/glad/glad.cpp"
  HEAP32[164735] = $1448; //@line 3661 "libraries/glad/glad.cpp"
  $1449 = (FUNCTION_TABLE_ii[$load & 8191](453643)|0); //@line 3662 "libraries/glad/glad.cpp"
  HEAP32[165658] = $1449; //@line 3662 "libraries/glad/glad.cpp"
  $1450 = (FUNCTION_TABLE_ii[$load & 8191](459406)|0); //@line 3663 "libraries/glad/glad.cpp"
  HEAP32[164771] = $1450; //@line 3663 "libraries/glad/glad.cpp"
  $1451 = (FUNCTION_TABLE_ii[$load & 8191](463130)|0); //@line 3664 "libraries/glad/glad.cpp"
  HEAP32[164661] = $1451; //@line 3664 "libraries/glad/glad.cpp"
  $1452 = (FUNCTION_TABLE_ii[$load & 8191](460122)|0); //@line 3665 "libraries/glad/glad.cpp"
  HEAP32[165560] = $1452; //@line 3665 "libraries/glad/glad.cpp"
  $1453 = (FUNCTION_TABLE_ii[$load & 8191](455426)|0); //@line 3666 "libraries/glad/glad.cpp"
  HEAP32[164674] = $1453; //@line 3666 "libraries/glad/glad.cpp"
  $1454 = (FUNCTION_TABLE_ii[$load & 8191](455445)|0); //@line 3667 "libraries/glad/glad.cpp"
  HEAP32[164688] = $1454; //@line 3667 "libraries/glad/glad.cpp"
  $1455 = (FUNCTION_TABLE_ii[$load & 8191](447848)|0); //@line 3668 "libraries/glad/glad.cpp"
  HEAP32[164903] = $1455; //@line 3668 "libraries/glad/glad.cpp"
  $1456 = (FUNCTION_TABLE_ii[$load & 8191](447823)|0); //@line 3669 "libraries/glad/glad.cpp"
  HEAP32[165581] = $1456; //@line 3669 "libraries/glad/glad.cpp"
  $1457 = (FUNCTION_TABLE_ii[$load & 8191](460188)|0); //@line 3670 "libraries/glad/glad.cpp"
  HEAP32[165479] = $1457; //@line 3670 "libraries/glad/glad.cpp"
  $1458 = (FUNCTION_TABLE_ii[$load & 8191](460207)|0); //@line 3671 "libraries/glad/glad.cpp"
  HEAP32[165103] = $1458; //@line 3671 "libraries/glad/glad.cpp"
  $1459 = (FUNCTION_TABLE_ii[$load & 8191](460228)|0); //@line 3672 "libraries/glad/glad.cpp"
  HEAP32[165544] = $1459; //@line 3672 "libraries/glad/glad.cpp"
  $1460 = (FUNCTION_TABLE_ii[$load & 8191](460272)|0); //@line 3673 "libraries/glad/glad.cpp"
  HEAP32[165107] = $1460; //@line 3673 "libraries/glad/glad.cpp"
  $1461 = (FUNCTION_TABLE_ii[$load & 8191](460294)|0); //@line 3674 "libraries/glad/glad.cpp"
  HEAP32[164858] = $1461; //@line 3674 "libraries/glad/glad.cpp"
 }
 $1462 = HEAP8[1252015]|0; //@line 3679 "libraries/glad/glad.cpp"
 $1463 = ($1462<<24>>24)==(0); //@line 3679 "libraries/glad/glad.cpp"
 if (!($1463)) {
  $1464 = (FUNCTION_TABLE_ii[$load & 8191](463569)|0); //@line 3680 "libraries/glad/glad.cpp"
  HEAP32[165597] = $1464; //@line 3680 "libraries/glad/glad.cpp"
  $1465 = (FUNCTION_TABLE_ii[$load & 8191](459622)|0); //@line 3681 "libraries/glad/glad.cpp"
  HEAP32[164975] = $1465; //@line 3681 "libraries/glad/glad.cpp"
  $1466 = (FUNCTION_TABLE_ii[$load & 8191](460317)|0); //@line 3682 "libraries/glad/glad.cpp"
  HEAP32[164693] = $1466; //@line 3682 "libraries/glad/glad.cpp"
  $1467 = (FUNCTION_TABLE_ii[$load & 8191](460339)|0); //@line 3683 "libraries/glad/glad.cpp"
  HEAP32[165481] = $1467; //@line 3683 "libraries/glad/glad.cpp"
  $1468 = (FUNCTION_TABLE_ii[$load & 8191](460360)|0); //@line 3684 "libraries/glad/glad.cpp"
  HEAP32[165100] = $1468; //@line 3684 "libraries/glad/glad.cpp"
  $1469 = (FUNCTION_TABLE_ii[$load & 8191](460383)|0); //@line 3685 "libraries/glad/glad.cpp"
  HEAP32[164807] = $1469; //@line 3685 "libraries/glad/glad.cpp"
  $1470 = (FUNCTION_TABLE_ii[$load & 8191](460404)|0); //@line 3686 "libraries/glad/glad.cpp"
  HEAP32[164899] = $1470; //@line 3686 "libraries/glad/glad.cpp"
  $1471 = (FUNCTION_TABLE_ii[$load & 8191](460421)|0); //@line 3687 "libraries/glad/glad.cpp"
  HEAP32[164838] = $1471; //@line 3687 "libraries/glad/glad.cpp"
  $1472 = (FUNCTION_TABLE_ii[$load & 8191](460437)|0); //@line 3688 "libraries/glad/glad.cpp"
  HEAP32[165347] = $1472; //@line 3688 "libraries/glad/glad.cpp"
  $1473 = (FUNCTION_TABLE_ii[$load & 8191](460451)|0); //@line 3689 "libraries/glad/glad.cpp"
  HEAP32[165138] = $1473; //@line 3689 "libraries/glad/glad.cpp"
  $1474 = (FUNCTION_TABLE_ii[$load & 8191](460468)|0); //@line 3690 "libraries/glad/glad.cpp"
  HEAP32[164994] = $1474; //@line 3690 "libraries/glad/glad.cpp"
  $1475 = (FUNCTION_TABLE_ii[$load & 8191](460485)|0); //@line 3691 "libraries/glad/glad.cpp"
  HEAP32[164946] = $1475; //@line 3691 "libraries/glad/glad.cpp"
  $1476 = (FUNCTION_TABLE_ii[$load & 8191](615462)|0); //@line 3692 "libraries/glad/glad.cpp"
  HEAP32[164857] = $1476; //@line 3692 "libraries/glad/glad.cpp"
  $1477 = (FUNCTION_TABLE_ii[$load & 8191](453675)|0); //@line 3693 "libraries/glad/glad.cpp"
  HEAP32[164666] = $1477; //@line 3693 "libraries/glad/glad.cpp"
  $1478 = (FUNCTION_TABLE_ii[$load & 8191](453685)|0); //@line 3694 "libraries/glad/glad.cpp"
  HEAP32[165534] = $1478; //@line 3694 "libraries/glad/glad.cpp"
  $1479 = (FUNCTION_TABLE_ii[$load & 8191](456513)|0); //@line 3695 "libraries/glad/glad.cpp"
  HEAP32[164839] = $1479; //@line 3695 "libraries/glad/glad.cpp"
  $1480 = (FUNCTION_TABLE_ii[$load & 8191](456530)|0); //@line 3696 "libraries/glad/glad.cpp"
  HEAP32[165455] = $1480; //@line 3696 "libraries/glad/glad.cpp"
  $1481 = (FUNCTION_TABLE_ii[$load & 8191](456555)|0); //@line 3697 "libraries/glad/glad.cpp"
  HEAP32[165507] = $1481; //@line 3697 "libraries/glad/glad.cpp"
  $1482 = (FUNCTION_TABLE_ii[$load & 8191](456568)|0); //@line 3698 "libraries/glad/glad.cpp"
  HEAP32[165519] = $1482; //@line 3698 "libraries/glad/glad.cpp"
  $1483 = (FUNCTION_TABLE_ii[$load & 8191](453630)|0); //@line 3699 "libraries/glad/glad.cpp"
  HEAP32[164721] = $1483; //@line 3699 "libraries/glad/glad.cpp"
  $1484 = (FUNCTION_TABLE_ii[$load & 8191](453696)|0); //@line 3700 "libraries/glad/glad.cpp"
  HEAP32[164970] = $1484; //@line 3700 "libraries/glad/glad.cpp"
  $1485 = (FUNCTION_TABLE_ii[$load & 8191](455087)|0); //@line 3701 "libraries/glad/glad.cpp"
  HEAP32[164902] = $1485; //@line 3701 "libraries/glad/glad.cpp"
  $1486 = (FUNCTION_TABLE_ii[$load & 8191](455112)|0); //@line 3702 "libraries/glad/glad.cpp"
  HEAP32[164690] = $1486; //@line 3702 "libraries/glad/glad.cpp"
  $1487 = (FUNCTION_TABLE_ii[$load & 8191](455142)|0); //@line 3703 "libraries/glad/glad.cpp"
  HEAP32[165310] = $1487; //@line 3703 "libraries/glad/glad.cpp"
  $1488 = (FUNCTION_TABLE_ii[$load & 8191](455357)|0); //@line 3704 "libraries/glad/glad.cpp"
  HEAP32[165518] = $1488; //@line 3704 "libraries/glad/glad.cpp"
  $1489 = (FUNCTION_TABLE_ii[$load & 8191](463584)|0); //@line 3705 "libraries/glad/glad.cpp"
  HEAP32[164955] = $1489; //@line 3705 "libraries/glad/glad.cpp"
  $1490 = (FUNCTION_TABLE_ii[$load & 8191](463206)|0); //@line 3706 "libraries/glad/glad.cpp"
  HEAP32[164772] = $1490; //@line 3706 "libraries/glad/glad.cpp"
  $1491 = (FUNCTION_TABLE_ii[$load & 8191](463336)|0); //@line 3707 "libraries/glad/glad.cpp"
  HEAP32[165473] = $1491; //@line 3707 "libraries/glad/glad.cpp"
  $1492 = (FUNCTION_TABLE_ii[$load & 8191](463287)|0); //@line 3708 "libraries/glad/glad.cpp"
  HEAP32[165309] = $1492; //@line 3708 "libraries/glad/glad.cpp"
  $1493 = (FUNCTION_TABLE_ii[$load & 8191](463303)|0); //@line 3709 "libraries/glad/glad.cpp"
  HEAP32[165207] = $1493; //@line 3709 "libraries/glad/glad.cpp"
  $1494 = (FUNCTION_TABLE_ii[$load & 8191](463319)|0); //@line 3710 "libraries/glad/glad.cpp"
  HEAP32[165352] = $1494; //@line 3710 "libraries/glad/glad.cpp"
  $1495 = (FUNCTION_TABLE_ii[$load & 8191](456494)|0); //@line 3711 "libraries/glad/glad.cpp"
  HEAP32[164682] = $1495; //@line 3711 "libraries/glad/glad.cpp"
  $1496 = (FUNCTION_TABLE_ii[$load & 8191](457143)|0); //@line 3712 "libraries/glad/glad.cpp"
  HEAP32[165551] = $1496; //@line 3712 "libraries/glad/glad.cpp"
  $1497 = (FUNCTION_TABLE_ii[$load & 8191](454528)|0); //@line 3713 "libraries/glad/glad.cpp"
  HEAP32[165468] = $1497; //@line 3713 "libraries/glad/glad.cpp"
  $1498 = (FUNCTION_TABLE_ii[$load & 8191](454546)|0); //@line 3714 "libraries/glad/glad.cpp"
  HEAP32[165247] = $1498; //@line 3714 "libraries/glad/glad.cpp"
  $1499 = (FUNCTION_TABLE_ii[$load & 8191](454565)|0); //@line 3715 "libraries/glad/glad.cpp"
  HEAP32[165567] = $1499; //@line 3715 "libraries/glad/glad.cpp"
  $1500 = (FUNCTION_TABLE_ii[$load & 8191](454586)|0); //@line 3716 "libraries/glad/glad.cpp"
  HEAP32[164926] = $1500; //@line 3716 "libraries/glad/glad.cpp"
  $1501 = (FUNCTION_TABLE_ii[$load & 8191](455647)|0); //@line 3717 "libraries/glad/glad.cpp"
  HEAP32[165072] = $1501; //@line 3717 "libraries/glad/glad.cpp"
  $1502 = (FUNCTION_TABLE_ii[$load & 8191](455669)|0); //@line 3718 "libraries/glad/glad.cpp"
  HEAP32[165411] = $1502; //@line 3718 "libraries/glad/glad.cpp"
  $1503 = (FUNCTION_TABLE_ii[$load & 8191](455716)|0); //@line 3719 "libraries/glad/glad.cpp"
  HEAP32[164732] = $1503; //@line 3719 "libraries/glad/glad.cpp"
  $1504 = (FUNCTION_TABLE_ii[$load & 8191](455765)|0); //@line 3720 "libraries/glad/glad.cpp"
  HEAP32[165429] = $1504; //@line 3720 "libraries/glad/glad.cpp"
  $1505 = (FUNCTION_TABLE_ii[$load & 8191](454867)|0); //@line 3721 "libraries/glad/glad.cpp"
  HEAP32[165113] = $1505; //@line 3721 "libraries/glad/glad.cpp"
  $1506 = (FUNCTION_TABLE_ii[$load & 8191](460105)|0); //@line 3722 "libraries/glad/glad.cpp"
  HEAP32[165654] = $1506; //@line 3722 "libraries/glad/glad.cpp"
  $1507 = (FUNCTION_TABLE_ii[$load & 8191](460148)|0); //@line 3723 "libraries/glad/glad.cpp"
  HEAP32[164910] = $1507; //@line 3723 "libraries/glad/glad.cpp"
 }
 __ZN4gladL15find_extensionsEv(); //@line 7303 "libraries/glad/glad.cpp"
 $1508 = HEAP8[1252016]|0; //@line 3728 "libraries/glad/glad.cpp"
 $1509 = ($1508<<24>>24)==(0); //@line 3728 "libraries/glad/glad.cpp"
 if (!($1509)) {
  $1510 = (FUNCTION_TABLE_ii[$load & 8191](476500)|0); //@line 3729 "libraries/glad/glad.cpp"
  HEAP32[165664] = $1510; //@line 3729 "libraries/glad/glad.cpp"
 }
 $1511 = HEAP8[1252019]|0; //@line 3736 "libraries/glad/glad.cpp"
 $1512 = ($1511<<24>>24)==(0); //@line 3736 "libraries/glad/glad.cpp"
 if (!($1512)) {
  $1513 = (FUNCTION_TABLE_ii[$load & 8191](460317)|0); //@line 3737 "libraries/glad/glad.cpp"
  HEAP32[164693] = $1513; //@line 3737 "libraries/glad/glad.cpp"
  $1514 = (FUNCTION_TABLE_ii[$load & 8191](460339)|0); //@line 3738 "libraries/glad/glad.cpp"
  HEAP32[165481] = $1514; //@line 3738 "libraries/glad/glad.cpp"
  $1515 = (FUNCTION_TABLE_ii[$load & 8191](460360)|0); //@line 3739 "libraries/glad/glad.cpp"
  HEAP32[165100] = $1515; //@line 3739 "libraries/glad/glad.cpp"
  $1516 = (FUNCTION_TABLE_ii[$load & 8191](460383)|0); //@line 3740 "libraries/glad/glad.cpp"
  HEAP32[164807] = $1516; //@line 3740 "libraries/glad/glad.cpp"
  $1517 = (FUNCTION_TABLE_ii[$load & 8191](460404)|0); //@line 3741 "libraries/glad/glad.cpp"
  HEAP32[164899] = $1517; //@line 3741 "libraries/glad/glad.cpp"
  $1518 = (FUNCTION_TABLE_ii[$load & 8191](460421)|0); //@line 3742 "libraries/glad/glad.cpp"
  HEAP32[164838] = $1518; //@line 3742 "libraries/glad/glad.cpp"
  $1519 = (FUNCTION_TABLE_ii[$load & 8191](460437)|0); //@line 3743 "libraries/glad/glad.cpp"
  HEAP32[165347] = $1519; //@line 3743 "libraries/glad/glad.cpp"
  $1520 = (FUNCTION_TABLE_ii[$load & 8191](460451)|0); //@line 3744 "libraries/glad/glad.cpp"
  HEAP32[165138] = $1520; //@line 3744 "libraries/glad/glad.cpp"
  $1521 = (FUNCTION_TABLE_ii[$load & 8191](460468)|0); //@line 3745 "libraries/glad/glad.cpp"
  HEAP32[164994] = $1521; //@line 3745 "libraries/glad/glad.cpp"
  $1522 = (FUNCTION_TABLE_ii[$load & 8191](460485)|0); //@line 3746 "libraries/glad/glad.cpp"
  HEAP32[164946] = $1522; //@line 3746 "libraries/glad/glad.cpp"
  $1523 = (FUNCTION_TABLE_ii[$load & 8191](615462)|0); //@line 3747 "libraries/glad/glad.cpp"
  HEAP32[164857] = $1523; //@line 3747 "libraries/glad/glad.cpp"
  $1524 = (FUNCTION_TABLE_ii[$load & 8191](476518)|0); //@line 3748 "libraries/glad/glad.cpp"
  HEAP32[165665] = $1524; //@line 3748 "libraries/glad/glad.cpp"
  $1525 = (FUNCTION_TABLE_ii[$load & 8191](476543)|0); //@line 3749 "libraries/glad/glad.cpp"
  HEAP32[165666] = $1525; //@line 3749 "libraries/glad/glad.cpp"
  $1526 = (FUNCTION_TABLE_ii[$load & 8191](476567)|0); //@line 3750 "libraries/glad/glad.cpp"
  HEAP32[165667] = $1526; //@line 3750 "libraries/glad/glad.cpp"
  $1527 = (FUNCTION_TABLE_ii[$load & 8191](476593)|0); //@line 3751 "libraries/glad/glad.cpp"
  HEAP32[165668] = $1527; //@line 3751 "libraries/glad/glad.cpp"
  $1528 = (FUNCTION_TABLE_ii[$load & 8191](476617)|0); //@line 3752 "libraries/glad/glad.cpp"
  HEAP32[165669] = $1528; //@line 3752 "libraries/glad/glad.cpp"
  $1529 = (FUNCTION_TABLE_ii[$load & 8191](476637)|0); //@line 3753 "libraries/glad/glad.cpp"
  HEAP32[165670] = $1529; //@line 3753 "libraries/glad/glad.cpp"
  $1530 = (FUNCTION_TABLE_ii[$load & 8191](476656)|0); //@line 3754 "libraries/glad/glad.cpp"
  HEAP32[165671] = $1530; //@line 3754 "libraries/glad/glad.cpp"
  $1531 = (FUNCTION_TABLE_ii[$load & 8191](476673)|0); //@line 3755 "libraries/glad/glad.cpp"
  HEAP32[165672] = $1531; //@line 3755 "libraries/glad/glad.cpp"
  $1532 = (FUNCTION_TABLE_ii[$load & 8191](476693)|0); //@line 3756 "libraries/glad/glad.cpp"
  HEAP32[165673] = $1532; //@line 3756 "libraries/glad/glad.cpp"
  $1533 = (FUNCTION_TABLE_ii[$load & 8191](476713)|0); //@line 3757 "libraries/glad/glad.cpp"
  HEAP32[165674] = $1533; //@line 3757 "libraries/glad/glad.cpp"
  $1534 = (FUNCTION_TABLE_ii[$load & 8191](476736)|0); //@line 3758 "libraries/glad/glad.cpp"
  HEAP32[165675] = $1534; //@line 3758 "libraries/glad/glad.cpp"
 }
 $1535 = HEAP8[1252022]|0; //@line 3765 "libraries/glad/glad.cpp"
 $1536 = ($1535<<24>>24)==(0); //@line 3765 "libraries/glad/glad.cpp"
 if (!($1536)) {
  $1537 = (FUNCTION_TABLE_ii[$load & 8191](463206)|0); //@line 3766 "libraries/glad/glad.cpp"
  HEAP32[164772] = $1537; //@line 3766 "libraries/glad/glad.cpp"
  $1538 = (FUNCTION_TABLE_ii[$load & 8191](463336)|0); //@line 3767 "libraries/glad/glad.cpp"
  HEAP32[165473] = $1538; //@line 3767 "libraries/glad/glad.cpp"
  $1539 = (FUNCTION_TABLE_ii[$load & 8191](463287)|0); //@line 3768 "libraries/glad/glad.cpp"
  HEAP32[165309] = $1539; //@line 3768 "libraries/glad/glad.cpp"
  $1540 = (FUNCTION_TABLE_ii[$load & 8191](463303)|0); //@line 3769 "libraries/glad/glad.cpp"
  HEAP32[165207] = $1540; //@line 3769 "libraries/glad/glad.cpp"
  $1541 = (FUNCTION_TABLE_ii[$load & 8191](463319)|0); //@line 3770 "libraries/glad/glad.cpp"
  HEAP32[165352] = $1541; //@line 3770 "libraries/glad/glad.cpp"
  $1542 = (FUNCTION_TABLE_ii[$load & 8191](476753)|0); //@line 3771 "libraries/glad/glad.cpp"
  HEAP32[165676] = $1542; //@line 3771 "libraries/glad/glad.cpp"
  $1543 = (FUNCTION_TABLE_ii[$load & 8191](476781)|0); //@line 3772 "libraries/glad/glad.cpp"
  HEAP32[165677] = $1543; //@line 3772 "libraries/glad/glad.cpp"
  $1544 = (FUNCTION_TABLE_ii[$load & 8191](476798)|0); //@line 3773 "libraries/glad/glad.cpp"
  HEAP32[165678] = $1544; //@line 3773 "libraries/glad/glad.cpp"
  $1545 = (FUNCTION_TABLE_ii[$load & 8191](476817)|0); //@line 3774 "libraries/glad/glad.cpp"
  HEAP32[165679] = $1545; //@line 3774 "libraries/glad/glad.cpp"
  $1546 = (FUNCTION_TABLE_ii[$load & 8191](476836)|0); //@line 3775 "libraries/glad/glad.cpp"
  HEAP32[165680] = $1546; //@line 3775 "libraries/glad/glad.cpp"
 }
 $1547 = HEAP8[1252027]|0; //@line 3784 "libraries/glad/glad.cpp"
 $1548 = ($1547<<24>>24)==(0); //@line 3784 "libraries/glad/glad.cpp"
 if (!($1548)) {
  $1549 = (FUNCTION_TABLE_ii[$load & 8191](459216)|0); //@line 3785 "libraries/glad/glad.cpp"
  HEAP32[165657] = $1549; //@line 3785 "libraries/glad/glad.cpp"
  $1550 = (FUNCTION_TABLE_ii[$load & 8191](459250)|0); //@line 3786 "libraries/glad/glad.cpp"
  HEAP32[164665] = $1550; //@line 3786 "libraries/glad/glad.cpp"
  $1551 = (FUNCTION_TABLE_ii[$load & 8191](459286)|0); //@line 3787 "libraries/glad/glad.cpp"
  HEAP32[164895] = $1551; //@line 3787 "libraries/glad/glad.cpp"
 }
 $1552 = HEAP8[1252028]|0; //@line 3792 "libraries/glad/glad.cpp"
 $1553 = ($1552<<24>>24)==(0); //@line 3792 "libraries/glad/glad.cpp"
 if (!($1553)) {
  $1554 = (FUNCTION_TABLE_ii[$load & 8191](476856)|0); //@line 3793 "libraries/glad/glad.cpp"
  HEAP32[165681] = $1554; //@line 3793 "libraries/glad/glad.cpp"
  $1555 = (FUNCTION_TABLE_ii[$load & 8191](476878)|0); //@line 3794 "libraries/glad/glad.cpp"
  HEAP32[165682] = $1555; //@line 3794 "libraries/glad/glad.cpp"
  $1556 = (FUNCTION_TABLE_ii[$load & 8191](476907)|0); //@line 3795 "libraries/glad/glad.cpp"
  HEAP32[165683] = $1556; //@line 3795 "libraries/glad/glad.cpp"
  $1557 = (FUNCTION_TABLE_ii[$load & 8191](476938)|0); //@line 3796 "libraries/glad/glad.cpp"
  HEAP32[165684] = $1557; //@line 3796 "libraries/glad/glad.cpp"
  $1558 = (FUNCTION_TABLE_ii[$load & 8191](476972)|0); //@line 3797 "libraries/glad/glad.cpp"
  HEAP32[165685] = $1558; //@line 3797 "libraries/glad/glad.cpp"
  $1559 = (FUNCTION_TABLE_ii[$load & 8191](476992)|0); //@line 3798 "libraries/glad/glad.cpp"
  HEAP32[165686] = $1559; //@line 3798 "libraries/glad/glad.cpp"
  $1560 = (FUNCTION_TABLE_ii[$load & 8191](477021)|0); //@line 3799 "libraries/glad/glad.cpp"
  HEAP32[165687] = $1560; //@line 3799 "libraries/glad/glad.cpp"
  $1561 = (FUNCTION_TABLE_ii[$load & 8191](477053)|0); //@line 3800 "libraries/glad/glad.cpp"
  HEAP32[165688] = $1561; //@line 3800 "libraries/glad/glad.cpp"
  $1562 = (FUNCTION_TABLE_ii[$load & 8191](477076)|0); //@line 3801 "libraries/glad/glad.cpp"
  HEAP32[165689] = $1562; //@line 3801 "libraries/glad/glad.cpp"
  $1563 = (FUNCTION_TABLE_ii[$load & 8191](477100)|0); //@line 3802 "libraries/glad/glad.cpp"
  HEAP32[165690] = $1563; //@line 3802 "libraries/glad/glad.cpp"
  $1564 = (FUNCTION_TABLE_ii[$load & 8191](477130)|0); //@line 3803 "libraries/glad/glad.cpp"
  HEAP32[165691] = $1564; //@line 3803 "libraries/glad/glad.cpp"
  $1565 = (FUNCTION_TABLE_ii[$load & 8191](477161)|0); //@line 3804 "libraries/glad/glad.cpp"
  HEAP32[165692] = $1565; //@line 3804 "libraries/glad/glad.cpp"
  $1566 = (FUNCTION_TABLE_ii[$load & 8191](477190)|0); //@line 3805 "libraries/glad/glad.cpp"
  HEAP32[165693] = $1566; //@line 3805 "libraries/glad/glad.cpp"
  $1567 = (FUNCTION_TABLE_ii[$load & 8191](477217)|0); //@line 3806 "libraries/glad/glad.cpp"
  HEAP32[165694] = $1567; //@line 3806 "libraries/glad/glad.cpp"
  $1568 = (FUNCTION_TABLE_ii[$load & 8191](477241)|0); //@line 3807 "libraries/glad/glad.cpp"
  HEAP32[165695] = $1568; //@line 3807 "libraries/glad/glad.cpp"
  $1569 = (FUNCTION_TABLE_ii[$load & 8191](477266)|0); //@line 3808 "libraries/glad/glad.cpp"
  HEAP32[165696] = $1569; //@line 3808 "libraries/glad/glad.cpp"
 }
 $1570 = HEAP8[1252029]|0; //@line 3813 "libraries/glad/glad.cpp"
 $1571 = ($1570<<24>>24)==(0); //@line 3813 "libraries/glad/glad.cpp"
 if (!($1571)) {
  $1572 = (FUNCTION_TABLE_ii[$load & 8191](455459)|0); //@line 3814 "libraries/glad/glad.cpp"
  HEAP32[164974] = $1572; //@line 3814 "libraries/glad/glad.cpp"
  $1573 = (FUNCTION_TABLE_ii[$load & 8191](455489)|0); //@line 3815 "libraries/glad/glad.cpp"
  HEAP32[164733] = $1573; //@line 3815 "libraries/glad/glad.cpp"
 }
 $1574 = HEAP8[1252030]|0; //@line 3820 "libraries/glad/glad.cpp"
 $1575 = ($1574<<24>>24)==(0); //@line 3820 "libraries/glad/glad.cpp"
 if (!($1575)) {
  $1576 = (FUNCTION_TABLE_ii[$load & 8191](460505)|0); //@line 3821 "libraries/glad/glad.cpp"
  HEAP32[165460] = $1576; //@line 3821 "libraries/glad/glad.cpp"
 }
 $1577 = HEAP8[1252031]|0; //@line 3826 "libraries/glad/glad.cpp"
 $1578 = ($1577<<24>>24)==(0); //@line 3826 "libraries/glad/glad.cpp"
 if (!($1578)) {
  $1579 = (FUNCTION_TABLE_ii[$load & 8191](477293)|0); //@line 3827 "libraries/glad/glad.cpp"
  HEAP32[165697] = $1579; //@line 3827 "libraries/glad/glad.cpp"
 }
 $1580 = HEAP8[1252032]|0; //@line 3832 "libraries/glad/glad.cpp"
 $1581 = ($1580<<24>>24)==(0); //@line 3832 "libraries/glad/glad.cpp"
 if (!($1581)) {
  $1582 = (FUNCTION_TABLE_ii[$load & 8191](459539)|0); //@line 3833 "libraries/glad/glad.cpp"
  HEAP32[165484] = $1582; //@line 3833 "libraries/glad/glad.cpp"
  $1583 = (FUNCTION_TABLE_ii[$load & 8191](459557)|0); //@line 3834 "libraries/glad/glad.cpp"
  HEAP32[165178] = $1583; //@line 3834 "libraries/glad/glad.cpp"
 }
 $1584 = HEAP8[1252033]|0; //@line 3839 "libraries/glad/glad.cpp"
 $1585 = ($1584<<24>>24)==(0); //@line 3839 "libraries/glad/glad.cpp"
 if (!($1585)) {
  $1586 = (FUNCTION_TABLE_ii[$load & 8191](460521)|0); //@line 3840 "libraries/glad/glad.cpp"
  HEAP32[165062] = $1586; //@line 3840 "libraries/glad/glad.cpp"
  $1587 = (FUNCTION_TABLE_ii[$load & 8191](460537)|0); //@line 3841 "libraries/glad/glad.cpp"
  HEAP32[164801] = $1587; //@line 3841 "libraries/glad/glad.cpp"
 }
 $1588 = HEAP8[1252034]|0; //@line 3846 "libraries/glad/glad.cpp"
 $1589 = ($1588<<24>>24)==(0); //@line 3846 "libraries/glad/glad.cpp"
 if (!($1589)) {
  $1590 = (FUNCTION_TABLE_ii[$load & 8191](460663)|0); //@line 3847 "libraries/glad/glad.cpp"
  HEAP32[165662] = $1590; //@line 3847 "libraries/glad/glad.cpp"
 }
 $1591 = HEAP8[1252035]|0; //@line 3852 "libraries/glad/glad.cpp"
 $1592 = ($1591<<24>>24)==(0); //@line 3852 "libraries/glad/glad.cpp"
 if (!($1592)) {
  $1593 = (FUNCTION_TABLE_ii[$load & 8191](477320)|0); //@line 3853 "libraries/glad/glad.cpp"
  HEAP32[165698] = $1593; //@line 3853 "libraries/glad/glad.cpp"
 }
 $1594 = HEAP8[1252038]|0; //@line 3860 "libraries/glad/glad.cpp"
 $1595 = ($1594<<24>>24)==(0); //@line 3860 "libraries/glad/glad.cpp"
 if (!($1595)) {
  $1596 = (FUNCTION_TABLE_ii[$load & 8191](459578)|0); //@line 3861 "libraries/glad/glad.cpp"
  HEAP32[164957] = $1596; //@line 3861 "libraries/glad/glad.cpp"
  $1597 = (FUNCTION_TABLE_ii[$load & 8191](459596)|0); //@line 3862 "libraries/glad/glad.cpp"
  HEAP32[165427] = $1597; //@line 3862 "libraries/glad/glad.cpp"
 }
 $1598 = HEAP8[1252039]|0; //@line 3867 "libraries/glad/glad.cpp"
 $1599 = ($1598<<24>>24)==(0); //@line 3867 "libraries/glad/glad.cpp"
 if (!($1599)) {
  $1600 = (FUNCTION_TABLE_ii[$load & 8191](477336)|0); //@line 3868 "libraries/glad/glad.cpp"
  HEAP32[165699] = $1600; //@line 3868 "libraries/glad/glad.cpp"
 }
 $1601 = HEAP8[1252042]|0; //@line 3875 "libraries/glad/glad.cpp"
 $1602 = ($1601<<24>>24)==(0); //@line 3875 "libraries/glad/glad.cpp"
 if (!($1602)) {
  $1603 = (FUNCTION_TABLE_ii[$load & 8191](454903)|0); //@line 3876 "libraries/glad/glad.cpp"
  HEAP32[165637] = $1603; //@line 3876 "libraries/glad/glad.cpp"
 }
 $1604 = HEAP8[1252043]|0; //@line 3881 "libraries/glad/glad.cpp"
 $1605 = ($1604<<24>>24)==(0); //@line 3881 "libraries/glad/glad.cpp"
 if (!($1605)) {
  $1606 = (FUNCTION_TABLE_ii[$load & 8191](459622)|0); //@line 3882 "libraries/glad/glad.cpp"
  HEAP32[164975] = $1606; //@line 3882 "libraries/glad/glad.cpp"
 }
 $1607 = HEAP8[1252045]|0; //@line 3888 "libraries/glad/glad.cpp"
 $1608 = ($1607<<24>>24)==(0); //@line 3888 "libraries/glad/glad.cpp"
 if (!($1608)) {
  $1609 = (FUNCTION_TABLE_ii[$load & 8191](477366)|0); //@line 3889 "libraries/glad/glad.cpp"
  HEAP32[165700] = $1609; //@line 3889 "libraries/glad/glad.cpp"
  $1610 = (FUNCTION_TABLE_ii[$load & 8191](477391)|0); //@line 3890 "libraries/glad/glad.cpp"
  HEAP32[165701] = $1610; //@line 3890 "libraries/glad/glad.cpp"
  $1611 = (FUNCTION_TABLE_ii[$load & 8191](477415)|0); //@line 3891 "libraries/glad/glad.cpp"
  HEAP32[165702] = $1611; //@line 3891 "libraries/glad/glad.cpp"
  $1612 = (FUNCTION_TABLE_ii[$load & 8191](477441)|0); //@line 3892 "libraries/glad/glad.cpp"
  HEAP32[165703] = $1612; //@line 3892 "libraries/glad/glad.cpp"
 }
 $1613 = HEAP8[1252049]|0; //@line 3900 "libraries/glad/glad.cpp"
 $1614 = ($1613<<24>>24)==(0); //@line 3900 "libraries/glad/glad.cpp"
 if (!($1614)) {
  $1615 = (FUNCTION_TABLE_ii[$load & 8191](460677)|0); //@line 3901 "libraries/glad/glad.cpp"
  HEAP32[165111] = $1615; //@line 3901 "libraries/glad/glad.cpp"
  $1616 = (FUNCTION_TABLE_ii[$load & 8191](460704)|0); //@line 3902 "libraries/glad/glad.cpp"
  HEAP32[164799] = $1616; //@line 3902 "libraries/glad/glad.cpp"
  $1617 = (FUNCTION_TABLE_ii[$load & 8191](460734)|0); //@line 3903 "libraries/glad/glad.cpp"
  HEAP32[164856] = $1617; //@line 3903 "libraries/glad/glad.cpp"
  $1618 = (FUNCTION_TABLE_ii[$load & 8191](460765)|0); //@line 3904 "libraries/glad/glad.cpp"
  HEAP32[164761] = $1618; //@line 3904 "libraries/glad/glad.cpp"
  $1619 = (FUNCTION_TABLE_ii[$load & 8191](460790)|0); //@line 3905 "libraries/glad/glad.cpp"
  HEAP32[164927] = $1619; //@line 3905 "libraries/glad/glad.cpp"
  $1620 = (FUNCTION_TABLE_ii[$load & 8191](460816)|0); //@line 3906 "libraries/glad/glad.cpp"
  HEAP32[165035] = $1620; //@line 3906 "libraries/glad/glad.cpp"
  $1621 = (FUNCTION_TABLE_ii[$load & 8191](460844)|0); //@line 3907 "libraries/glad/glad.cpp"
  HEAP32[165655] = $1621; //@line 3907 "libraries/glad/glad.cpp"
  $1622 = (FUNCTION_TABLE_ii[$load & 8191](460860)|0); //@line 3908 "libraries/glad/glad.cpp"
  HEAP32[164998] = $1622; //@line 3908 "libraries/glad/glad.cpp"
  $1623 = (FUNCTION_TABLE_ii[$load & 8191](460881)|0); //@line 3909 "libraries/glad/glad.cpp"
  HEAP32[165609] = $1623; //@line 3909 "libraries/glad/glad.cpp"
  $1624 = (FUNCTION_TABLE_ii[$load & 8191](460899)|0); //@line 3910 "libraries/glad/glad.cpp"
  HEAP32[165366] = $1624; //@line 3910 "libraries/glad/glad.cpp"
  $1625 = (FUNCTION_TABLE_ii[$load & 8191](460920)|0); //@line 3911 "libraries/glad/glad.cpp"
  HEAP32[164729] = $1625; //@line 3911 "libraries/glad/glad.cpp"
  $1626 = (FUNCTION_TABLE_ii[$load & 8191](460945)|0); //@line 3912 "libraries/glad/glad.cpp"
  HEAP32[165241] = $1626; //@line 3912 "libraries/glad/glad.cpp"
  $1627 = (FUNCTION_TABLE_ii[$load & 8191](460968)|0); //@line 3913 "libraries/glad/glad.cpp"
  HEAP32[165614] = $1627; //@line 3913 "libraries/glad/glad.cpp"
  $1628 = (FUNCTION_TABLE_ii[$load & 8191](460994)|0); //@line 3914 "libraries/glad/glad.cpp"
  HEAP32[164848] = $1628; //@line 3914 "libraries/glad/glad.cpp"
  $1629 = (FUNCTION_TABLE_ii[$load & 8191](461011)|0); //@line 3915 "libraries/glad/glad.cpp"
  HEAP32[165615] = $1629; //@line 3915 "libraries/glad/glad.cpp"
  $1630 = (FUNCTION_TABLE_ii[$load & 8191](461033)|0); //@line 3916 "libraries/glad/glad.cpp"
  HEAP32[165232] = $1630; //@line 3916 "libraries/glad/glad.cpp"
  $1631 = (FUNCTION_TABLE_ii[$load & 8191](461052)|0); //@line 3917 "libraries/glad/glad.cpp"
  HEAP32[165374] = $1631; //@line 3917 "libraries/glad/glad.cpp"
  $1632 = (FUNCTION_TABLE_ii[$load & 8191](461082)|0); //@line 3918 "libraries/glad/glad.cpp"
  HEAP32[164662] = $1632; //@line 3918 "libraries/glad/glad.cpp"
  $1633 = (FUNCTION_TABLE_ii[$load & 8191](461110)|0); //@line 3919 "libraries/glad/glad.cpp"
  HEAP32[165554] = $1633; //@line 3919 "libraries/glad/glad.cpp"
  $1634 = (FUNCTION_TABLE_ii[$load & 8191](461140)|0); //@line 3920 "libraries/glad/glad.cpp"
  HEAP32[165426] = $1634; //@line 3920 "libraries/glad/glad.cpp"
  $1635 = (FUNCTION_TABLE_ii[$load & 8191](461165)|0); //@line 3921 "libraries/glad/glad.cpp"
  HEAP32[165371] = $1635; //@line 3921 "libraries/glad/glad.cpp"
  $1636 = (FUNCTION_TABLE_ii[$load & 8191](461189)|0); //@line 3922 "libraries/glad/glad.cpp"
  HEAP32[165545] = $1636; //@line 3922 "libraries/glad/glad.cpp"
  $1637 = (FUNCTION_TABLE_ii[$load & 8191](461210)|0); //@line 3923 "libraries/glad/glad.cpp"
  HEAP32[165390] = $1637; //@line 3923 "libraries/glad/glad.cpp"
  $1638 = (FUNCTION_TABLE_ii[$load & 8191](461241)|0); //@line 3924 "libraries/glad/glad.cpp"
  HEAP32[165656] = $1638; //@line 3924 "libraries/glad/glad.cpp"
  $1639 = (FUNCTION_TABLE_ii[$load & 8191](461270)|0); //@line 3925 "libraries/glad/glad.cpp"
  HEAP32[164730] = $1639; //@line 3925 "libraries/glad/glad.cpp"
  $1640 = (FUNCTION_TABLE_ii[$load & 8191](461296)|0); //@line 3926 "libraries/glad/glad.cpp"
  HEAP32[165616] = $1640; //@line 3926 "libraries/glad/glad.cpp"
  $1641 = (FUNCTION_TABLE_ii[$load & 8191](461327)|0); //@line 3927 "libraries/glad/glad.cpp"
  HEAP32[165341] = $1641; //@line 3927 "libraries/glad/glad.cpp"
  $1642 = (FUNCTION_TABLE_ii[$load & 8191](461356)|0); //@line 3928 "libraries/glad/glad.cpp"
  HEAP32[164863] = $1642; //@line 3928 "libraries/glad/glad.cpp"
  $1643 = (FUNCTION_TABLE_ii[$load & 8191](461386)|0); //@line 3929 "libraries/glad/glad.cpp"
  HEAP32[165201] = $1643; //@line 3929 "libraries/glad/glad.cpp"
  $1644 = (FUNCTION_TABLE_ii[$load & 8191](461415)|0); //@line 3930 "libraries/glad/glad.cpp"
  HEAP32[165323] = $1644; //@line 3930 "libraries/glad/glad.cpp"
  $1645 = (FUNCTION_TABLE_ii[$load & 8191](461448)|0); //@line 3931 "libraries/glad/glad.cpp"
  HEAP32[165428] = $1645; //@line 3931 "libraries/glad/glad.cpp"
  $1646 = (FUNCTION_TABLE_ii[$load & 8191](461484)|0); //@line 3932 "libraries/glad/glad.cpp"
  HEAP32[165523] = $1646; //@line 3932 "libraries/glad/glad.cpp"
  $1647 = (FUNCTION_TABLE_ii[$load & 8191](461510)|0); //@line 3933 "libraries/glad/glad.cpp"
  HEAP32[164931] = $1647; //@line 3933 "libraries/glad/glad.cpp"
  $1648 = (FUNCTION_TABLE_ii[$load & 8191](461537)|0); //@line 3934 "libraries/glad/glad.cpp"
  HEAP32[165369] = $1648; //@line 3934 "libraries/glad/glad.cpp"
  $1649 = (FUNCTION_TABLE_ii[$load & 8191](461563)|0); //@line 3935 "libraries/glad/glad.cpp"
  HEAP32[164683] = $1649; //@line 3935 "libraries/glad/glad.cpp"
  $1650 = (FUNCTION_TABLE_ii[$load & 8191](461589)|0); //@line 3936 "libraries/glad/glad.cpp"
  HEAP32[164815] = $1650; //@line 3936 "libraries/glad/glad.cpp"
  $1651 = (FUNCTION_TABLE_ii[$load & 8191](461612)|0); //@line 3937 "libraries/glad/glad.cpp"
  HEAP32[165264] = $1651; //@line 3937 "libraries/glad/glad.cpp"
  $1652 = (FUNCTION_TABLE_ii[$load & 8191](461642)|0); //@line 3938 "libraries/glad/glad.cpp"
  HEAP32[165410] = $1652; //@line 3938 "libraries/glad/glad.cpp"
  $1653 = (FUNCTION_TABLE_ii[$load & 8191](461675)|0); //@line 3939 "libraries/glad/glad.cpp"
  HEAP32[165389] = $1653; //@line 3939 "libraries/glad/glad.cpp"
  $1654 = (FUNCTION_TABLE_ii[$load & 8191](461718)|0); //@line 3940 "libraries/glad/glad.cpp"
  HEAP32[165536] = $1654; //@line 3940 "libraries/glad/glad.cpp"
  $1655 = (FUNCTION_TABLE_ii[$load & 8191](461740)|0); //@line 3941 "libraries/glad/glad.cpp"
  HEAP32[164888] = $1655; //@line 3941 "libraries/glad/glad.cpp"
  $1656 = (FUNCTION_TABLE_ii[$load & 8191](461767)|0); //@line 3942 "libraries/glad/glad.cpp"
  HEAP32[165239] = $1656; //@line 3942 "libraries/glad/glad.cpp"
  $1657 = (FUNCTION_TABLE_ii[$load & 8191](461805)|0); //@line 3943 "libraries/glad/glad.cpp"
  HEAP32[164918] = $1657; //@line 3943 "libraries/glad/glad.cpp"
  $1658 = (FUNCTION_TABLE_ii[$load & 8191](461839)|0); //@line 3944 "libraries/glad/glad.cpp"
  HEAP32[164798] = $1658; //@line 3944 "libraries/glad/glad.cpp"
  $1659 = (FUNCTION_TABLE_ii[$load & 8191](461856)|0); //@line 3945 "libraries/glad/glad.cpp"
  HEAP32[165467] = $1659; //@line 3945 "libraries/glad/glad.cpp"
  $1660 = (FUNCTION_TABLE_ii[$load & 8191](461872)|0); //@line 3946 "libraries/glad/glad.cpp"
  HEAP32[164852] = $1660; //@line 3946 "libraries/glad/glad.cpp"
  $1661 = (FUNCTION_TABLE_ii[$load & 8191](461893)|0); //@line 3947 "libraries/glad/glad.cpp"
  HEAP32[164837] = $1661; //@line 3947 "libraries/glad/glad.cpp"
  $1662 = (FUNCTION_TABLE_ii[$load & 8191](461912)|0); //@line 3948 "libraries/glad/glad.cpp"
  HEAP32[165076] = $1662; //@line 3948 "libraries/glad/glad.cpp"
  $1663 = (FUNCTION_TABLE_ii[$load & 8191](461931)|0); //@line 3949 "libraries/glad/glad.cpp"
  HEAP32[165501] = $1663; //@line 3949 "libraries/glad/glad.cpp"
  $1664 = (FUNCTION_TABLE_ii[$load & 8191](461950)|0); //@line 3950 "libraries/glad/glad.cpp"
  HEAP32[164896] = $1664; //@line 3950 "libraries/glad/glad.cpp"
  $1665 = (FUNCTION_TABLE_ii[$load & 8191](461980)|0); //@line 3951 "libraries/glad/glad.cpp"
  HEAP32[164623] = $1665; //@line 3951 "libraries/glad/glad.cpp"
  $1666 = (FUNCTION_TABLE_ii[$load & 8191](462010)|0); //@line 3952 "libraries/glad/glad.cpp"
  HEAP32[165432] = $1666; //@line 3952 "libraries/glad/glad.cpp"
  $1667 = (FUNCTION_TABLE_ii[$load & 8191](462030)|0); //@line 3953 "libraries/glad/glad.cpp"
  HEAP32[165218] = $1667; //@line 3953 "libraries/glad/glad.cpp"
  $1668 = (FUNCTION_TABLE_ii[$load & 8191](462050)|0); //@line 3954 "libraries/glad/glad.cpp"
  HEAP32[164987] = $1668; //@line 3954 "libraries/glad/glad.cpp"
  $1669 = (FUNCTION_TABLE_ii[$load & 8191](462070)|0); //@line 3955 "libraries/glad/glad.cpp"
  HEAP32[165013] = $1669; //@line 3955 "libraries/glad/glad.cpp"
  $1670 = (FUNCTION_TABLE_ii[$load & 8191](462100)|0); //@line 3956 "libraries/glad/glad.cpp"
  HEAP32[165250] = $1670; //@line 3956 "libraries/glad/glad.cpp"
  $1671 = (FUNCTION_TABLE_ii[$load & 8191](462130)|0); //@line 3957 "libraries/glad/glad.cpp"
  HEAP32[165372] = $1671; //@line 3957 "libraries/glad/glad.cpp"
  $1672 = (FUNCTION_TABLE_ii[$load & 8191](462160)|0); //@line 3958 "libraries/glad/glad.cpp"
  HEAP32[165089] = $1672; //@line 3958 "libraries/glad/glad.cpp"
  $1673 = (FUNCTION_TABLE_ii[$load & 8191](462184)|0); //@line 3959 "libraries/glad/glad.cpp"
  HEAP32[165271] = $1673; //@line 3959 "libraries/glad/glad.cpp"
  $1674 = (FUNCTION_TABLE_ii[$load & 8191](462208)|0); //@line 3960 "libraries/glad/glad.cpp"
  HEAP32[164786] = $1674; //@line 3960 "libraries/glad/glad.cpp"
  $1675 = (FUNCTION_TABLE_ii[$load & 8191](462232)|0); //@line 3961 "libraries/glad/glad.cpp"
  HEAP32[164619] = $1675; //@line 3961 "libraries/glad/glad.cpp"
  $1676 = (FUNCTION_TABLE_ii[$load & 8191](462252)|0); //@line 3962 "libraries/glad/glad.cpp"
  HEAP32[164624] = $1676; //@line 3962 "libraries/glad/glad.cpp"
  $1677 = (FUNCTION_TABLE_ii[$load & 8191](462273)|0); //@line 3963 "libraries/glad/glad.cpp"
  HEAP32[164947] = $1677; //@line 3963 "libraries/glad/glad.cpp"
  $1678 = (FUNCTION_TABLE_ii[$load & 8191](462293)|0); //@line 3964 "libraries/glad/glad.cpp"
  HEAP32[165214] = $1678; //@line 3964 "libraries/glad/glad.cpp"
  $1679 = (FUNCTION_TABLE_ii[$load & 8191](462315)|0); //@line 3965 "libraries/glad/glad.cpp"
  HEAP32[165093] = $1679; //@line 3965 "libraries/glad/glad.cpp"
  $1680 = (FUNCTION_TABLE_ii[$load & 8191](462338)|0); //@line 3966 "libraries/glad/glad.cpp"
  HEAP32[165346] = $1680; //@line 3966 "libraries/glad/glad.cpp"
  $1681 = (FUNCTION_TABLE_ii[$load & 8191](462359)|0); //@line 3967 "libraries/glad/glad.cpp"
  HEAP32[165077] = $1681; //@line 3967 "libraries/glad/glad.cpp"
  $1682 = (FUNCTION_TABLE_ii[$load & 8191](462383)|0); //@line 3968 "libraries/glad/glad.cpp"
  HEAP32[164695] = $1682; //@line 3968 "libraries/glad/glad.cpp"
  $1683 = (FUNCTION_TABLE_ii[$load & 8191](462401)|0); //@line 3969 "libraries/glad/glad.cpp"
  HEAP32[164982] = $1683; //@line 3969 "libraries/glad/glad.cpp"
  $1684 = (FUNCTION_TABLE_ii[$load & 8191](462419)|0); //@line 3970 "libraries/glad/glad.cpp"
  HEAP32[164651] = $1684; //@line 3970 "libraries/glad/glad.cpp"
  $1685 = (FUNCTION_TABLE_ii[$load & 8191](462447)|0); //@line 3971 "libraries/glad/glad.cpp"
  HEAP32[165423] = $1685; //@line 3971 "libraries/glad/glad.cpp"
  $1686 = (FUNCTION_TABLE_ii[$load & 8191](462476)|0); //@line 3972 "libraries/glad/glad.cpp"
  HEAP32[165406] = $1686; //@line 3972 "libraries/glad/glad.cpp"
  $1687 = (FUNCTION_TABLE_ii[$load & 8191](462505)|0); //@line 3973 "libraries/glad/glad.cpp"
  HEAP32[165217] = $1687; //@line 3973 "libraries/glad/glad.cpp"
  $1688 = (FUNCTION_TABLE_ii[$load & 8191](462529)|0); //@line 3974 "libraries/glad/glad.cpp"
  HEAP32[164784] = $1688; //@line 3974 "libraries/glad/glad.cpp"
  $1689 = (FUNCTION_TABLE_ii[$load & 8191](462554)|0); //@line 3975 "libraries/glad/glad.cpp"
  HEAP32[165377] = $1689; //@line 3975 "libraries/glad/glad.cpp"
  $1690 = (FUNCTION_TABLE_ii[$load & 8191](462580)|0); //@line 3976 "libraries/glad/glad.cpp"
  HEAP32[165160] = $1690; //@line 3976 "libraries/glad/glad.cpp"
  $1691 = (FUNCTION_TABLE_ii[$load & 8191](462604)|0); //@line 3977 "libraries/glad/glad.cpp"
  HEAP32[165172] = $1691; //@line 3977 "libraries/glad/glad.cpp"
  $1692 = (FUNCTION_TABLE_ii[$load & 8191](462625)|0); //@line 3978 "libraries/glad/glad.cpp"
  HEAP32[164912] = $1692; //@line 3978 "libraries/glad/glad.cpp"
  $1693 = (FUNCTION_TABLE_ii[$load & 8191](462652)|0); //@line 3979 "libraries/glad/glad.cpp"
  HEAP32[165313] = $1693; //@line 3979 "libraries/glad/glad.cpp"
  $1694 = (FUNCTION_TABLE_ii[$load & 8191](462678)|0); //@line 3980 "libraries/glad/glad.cpp"
  HEAP32[164621] = $1694; //@line 3980 "libraries/glad/glad.cpp"
  $1695 = (FUNCTION_TABLE_ii[$load & 8191](462705)|0); //@line 3981 "libraries/glad/glad.cpp"
  HEAP32[164906] = $1695; //@line 3981 "libraries/glad/glad.cpp"
  $1696 = (FUNCTION_TABLE_ii[$load & 8191](462731)|0); //@line 3982 "libraries/glad/glad.cpp"
  HEAP32[165012] = $1696; //@line 3982 "libraries/glad/glad.cpp"
  $1697 = (FUNCTION_TABLE_ii[$load & 8191](462758)|0); //@line 3983 "libraries/glad/glad.cpp"
  HEAP32[164836] = $1697; //@line 3983 "libraries/glad/glad.cpp"
  $1698 = (FUNCTION_TABLE_ii[$load & 8191](462785)|0); //@line 3984 "libraries/glad/glad.cpp"
  HEAP32[165472] = $1698; //@line 3984 "libraries/glad/glad.cpp"
  $1699 = (FUNCTION_TABLE_ii[$load & 8191](462811)|0); //@line 3985 "libraries/glad/glad.cpp"
  HEAP32[164789] = $1699; //@line 3985 "libraries/glad/glad.cpp"
  $1700 = (FUNCTION_TABLE_ii[$load & 8191](462838)|0); //@line 3986 "libraries/glad/glad.cpp"
  HEAP32[164748] = $1700; //@line 3986 "libraries/glad/glad.cpp"
  $1701 = (FUNCTION_TABLE_ii[$load & 8191](462865)|0); //@line 3987 "libraries/glad/glad.cpp"
  HEAP32[164867] = $1701; //@line 3987 "libraries/glad/glad.cpp"
  $1702 = (FUNCTION_TABLE_ii[$load & 8191](462893)|0); //@line 3988 "libraries/glad/glad.cpp"
  HEAP32[165270] = $1702; //@line 3988 "libraries/glad/glad.cpp"
  $1703 = (FUNCTION_TABLE_ii[$load & 8191](462912)|0); //@line 3989 "libraries/glad/glad.cpp"
  HEAP32[164932] = $1703; //@line 3989 "libraries/glad/glad.cpp"
  $1704 = (FUNCTION_TABLE_ii[$load & 8191](462938)|0); //@line 3990 "libraries/glad/glad.cpp"
  HEAP32[165589] = $1704; //@line 3990 "libraries/glad/glad.cpp"
  $1705 = (FUNCTION_TABLE_ii[$load & 8191](462966)|0); //@line 3991 "libraries/glad/glad.cpp"
  HEAP32[165482] = $1705; //@line 3991 "libraries/glad/glad.cpp"
  $1706 = (FUNCTION_TABLE_ii[$load & 8191](462983)|0); //@line 3992 "libraries/glad/glad.cpp"
  HEAP32[165299] = $1706; //@line 3992 "libraries/glad/glad.cpp"
  $1707 = (FUNCTION_TABLE_ii[$load & 8191](463008)|0); //@line 3993 "libraries/glad/glad.cpp"
  HEAP32[164817] = $1707; //@line 3993 "libraries/glad/glad.cpp"
  $1708 = (FUNCTION_TABLE_ii[$load & 8191](463024)|0); //@line 3994 "libraries/glad/glad.cpp"
  HEAP32[164739] = $1708; //@line 3994 "libraries/glad/glad.cpp"
  $1709 = (FUNCTION_TABLE_ii[$load & 8191](463051)|0); //@line 3995 "libraries/glad/glad.cpp"
  HEAP32[165353] = $1709; //@line 3995 "libraries/glad/glad.cpp"
  $1710 = (FUNCTION_TABLE_ii[$load & 8191](463076)|0); //@line 3996 "libraries/glad/glad.cpp"
  HEAP32[165568] = $1710; //@line 3996 "libraries/glad/glad.cpp"
  $1711 = (FUNCTION_TABLE_ii[$load & 8191](463104)|0); //@line 3997 "libraries/glad/glad.cpp"
  HEAP32[164684] = $1711; //@line 3997 "libraries/glad/glad.cpp"
 }
 $1712 = HEAP8[1252050]|0; //@line 4002 "libraries/glad/glad.cpp"
 $1713 = ($1712<<24>>24)==(0); //@line 4002 "libraries/glad/glad.cpp"
 if (!($1713)) {
  $1714 = (FUNCTION_TABLE_ii[$load & 8191](477465)|0); //@line 4003 "libraries/glad/glad.cpp"
  HEAP32[165704] = $1714; //@line 4003 "libraries/glad/glad.cpp"
  $1715 = (FUNCTION_TABLE_ii[$load & 8191](477485)|0); //@line 4004 "libraries/glad/glad.cpp"
  HEAP32[165705] = $1715; //@line 4004 "libraries/glad/glad.cpp"
  $1716 = (FUNCTION_TABLE_ii[$load & 8191](477513)|0); //@line 4005 "libraries/glad/glad.cpp"
  HEAP32[165706] = $1716; //@line 4005 "libraries/glad/glad.cpp"
  $1717 = (FUNCTION_TABLE_ii[$load & 8191](477529)|0); //@line 4006 "libraries/glad/glad.cpp"
  HEAP32[165707] = $1717; //@line 4006 "libraries/glad/glad.cpp"
 }
 $1718 = HEAP8[1252051]|0; //@line 4011 "libraries/glad/glad.cpp"
 $1719 = ($1718<<24>>24)==(0); //@line 4011 "libraries/glad/glad.cpp"
 if (!($1719)) {
  $1720 = (FUNCTION_TABLE_ii[$load & 8191](455087)|0); //@line 4012 "libraries/glad/glad.cpp"
  HEAP32[164902] = $1720; //@line 4012 "libraries/glad/glad.cpp"
  $1721 = (FUNCTION_TABLE_ii[$load & 8191](455112)|0); //@line 4013 "libraries/glad/glad.cpp"
  HEAP32[164690] = $1721; //@line 4013 "libraries/glad/glad.cpp"
  $1722 = (FUNCTION_TABLE_ii[$load & 8191](455142)|0); //@line 4014 "libraries/glad/glad.cpp"
  HEAP32[165310] = $1722; //@line 4014 "libraries/glad/glad.cpp"
  $1723 = (FUNCTION_TABLE_ii[$load & 8191](455176)|0); //@line 4015 "libraries/glad/glad.cpp"
  HEAP32[165073] = $1723; //@line 4015 "libraries/glad/glad.cpp"
 }
 $1724 = HEAP8[1252052]|0; //@line 4020 "libraries/glad/glad.cpp"
 $1725 = ($1724<<24>>24)==(0); //@line 4020 "libraries/glad/glad.cpp"
 if (!($1725)) {
  $1726 = (FUNCTION_TABLE_ii[$load & 8191](456589)|0); //@line 4021 "libraries/glad/glad.cpp"
  HEAP32[164723] = $1726; //@line 4021 "libraries/glad/glad.cpp"
  $1727 = (FUNCTION_TABLE_ii[$load & 8191](456610)|0); //@line 4022 "libraries/glad/glad.cpp"
  HEAP32[165422] = $1727; //@line 4022 "libraries/glad/glad.cpp"
 }
 $1728 = HEAP8[1252053]|0; //@line 4027 "libraries/glad/glad.cpp"
 $1729 = ($1728<<24>>24)==(0); //@line 4027 "libraries/glad/glad.cpp"
 if (!($1729)) {
  $1730 = (FUNCTION_TABLE_ii[$load & 8191](477553)|0); //@line 4028 "libraries/glad/glad.cpp"
  HEAP32[165708] = $1730; //@line 4028 "libraries/glad/glad.cpp"
  $1731 = (FUNCTION_TABLE_ii[$load & 8191](477578)|0); //@line 4029 "libraries/glad/glad.cpp"
  HEAP32[165709] = $1731; //@line 4029 "libraries/glad/glad.cpp"
 }
 $1732 = HEAP8[1252055]|0; //@line 4035 "libraries/glad/glad.cpp"
 $1733 = ($1732<<24>>24)==(0); //@line 4035 "libraries/glad/glad.cpp"
 if (!($1733)) {
  $1734 = (FUNCTION_TABLE_ii[$load & 8191](615840)|0); //@line 4036 "libraries/glad/glad.cpp"
  HEAP32[165543] = $1734; //@line 4036 "libraries/glad/glad.cpp"
  $1735 = (FUNCTION_TABLE_ii[$load & 8191](615825)|0); //@line 4037 "libraries/glad/glad.cpp"
  HEAP32[165231] = $1735; //@line 4037 "libraries/glad/glad.cpp"
  $1736 = (FUNCTION_TABLE_ii[$load & 8191](615125)|0); //@line 4038 "libraries/glad/glad.cpp"
  HEAP32[164900] = $1736; //@line 4038 "libraries/glad/glad.cpp"
  $1737 = (FUNCTION_TABLE_ii[$load & 8191](616237)|0); //@line 4039 "libraries/glad/glad.cpp"
  HEAP32[164657] = $1737; //@line 4039 "libraries/glad/glad.cpp"
  $1738 = (FUNCTION_TABLE_ii[$load & 8191](615973)|0); //@line 4040 "libraries/glad/glad.cpp"
  HEAP32[165437] = $1738; //@line 4040 "libraries/glad/glad.cpp"
 }
 $1739 = HEAP8[1252056]|0; //@line 4045 "libraries/glad/glad.cpp"
 $1740 = ($1739<<24>>24)==(0); //@line 4045 "libraries/glad/glad.cpp"
 if (!($1740)) {
  $1741 = (FUNCTION_TABLE_ii[$load & 8191](463130)|0); //@line 4046 "libraries/glad/glad.cpp"
  HEAP32[164661] = $1741; //@line 4046 "libraries/glad/glad.cpp"
 }
 $1742 = HEAP8[1252057]|0; //@line 4051 "libraries/glad/glad.cpp"
 $1743 = ($1742<<24>>24)==(0); //@line 4051 "libraries/glad/glad.cpp"
 if (!($1743)) {
  $1744 = (FUNCTION_TABLE_ii[$load & 8191](477605)|0); //@line 4052 "libraries/glad/glad.cpp"
  HEAP32[165710] = $1744; //@line 4052 "libraries/glad/glad.cpp"
 }
 $1745 = HEAP8[1252064]|0; //@line 4063 "libraries/glad/glad.cpp"
 $1746 = ($1745<<24>>24)==(0); //@line 4063 "libraries/glad/glad.cpp"
 if (!($1746)) {
  $1747 = (FUNCTION_TABLE_ii[$load & 8191](459641)|0); //@line 4064 "libraries/glad/glad.cpp"
  HEAP32[165149] = $1747; //@line 4064 "libraries/glad/glad.cpp"
  $1748 = (FUNCTION_TABLE_ii[$load & 8191](459665)|0); //@line 4065 "libraries/glad/glad.cpp"
  HEAP32[165632] = $1748; //@line 4065 "libraries/glad/glad.cpp"
 }
 $1749 = HEAP8[1252065]|0; //@line 4070 "libraries/glad/glad.cpp"
 $1750 = ($1749<<24>>24)==(0); //@line 4070 "libraries/glad/glad.cpp"
 if (!($1750)) {
  $1751 = (FUNCTION_TABLE_ii[$load & 8191](614432)|0); //@line 4071 "libraries/glad/glad.cpp"
  HEAP32[164804] = $1751; //@line 4071 "libraries/glad/glad.cpp"
  $1752 = (FUNCTION_TABLE_ii[$load & 8191](614384)|0); //@line 4072 "libraries/glad/glad.cpp"
  HEAP32[165274] = $1752; //@line 4072 "libraries/glad/glad.cpp"
  $1753 = (FUNCTION_TABLE_ii[$load & 8191](614362)|0); //@line 4073 "libraries/glad/glad.cpp"
  HEAP32[164936] = $1753; //@line 4073 "libraries/glad/glad.cpp"
  $1754 = (FUNCTION_TABLE_ii[$load & 8191](614343)|0); //@line 4074 "libraries/glad/glad.cpp"
  HEAP32[164676] = $1754; //@line 4074 "libraries/glad/glad.cpp"
  $1755 = (FUNCTION_TABLE_ii[$load & 8191](616468)|0); //@line 4075 "libraries/glad/glad.cpp"
  HEAP32[164704] = $1755; //@line 4075 "libraries/glad/glad.cpp"
  $1756 = (FUNCTION_TABLE_ii[$load & 8191](614403)|0); //@line 4076 "libraries/glad/glad.cpp"
  HEAP32[164884] = $1756; //@line 4076 "libraries/glad/glad.cpp"
  $1757 = (FUNCTION_TABLE_ii[$load & 8191](615381)|0); //@line 4077 "libraries/glad/glad.cpp"
  HEAP32[165368] = $1757; //@line 4077 "libraries/glad/glad.cpp"
  $1758 = (FUNCTION_TABLE_ii[$load & 8191](615237)|0); //@line 4078 "libraries/glad/glad.cpp"
  HEAP32[165202] = $1758; //@line 4078 "libraries/glad/glad.cpp"
  $1759 = (FUNCTION_TABLE_ii[$load & 8191](615273)|0); //@line 4079 "libraries/glad/glad.cpp"
  HEAP32[164758] = $1759; //@line 4079 "libraries/glad/glad.cpp"
  $1760 = (FUNCTION_TABLE_ii[$load & 8191](615255)|0); //@line 4080 "libraries/glad/glad.cpp"
  HEAP32[164797] = $1760; //@line 4080 "libraries/glad/glad.cpp"
  $1761 = (FUNCTION_TABLE_ii[$load & 8191](616102)|0); //@line 4081 "libraries/glad/glad.cpp"
  HEAP32[165498] = $1761; //@line 4081 "libraries/glad/glad.cpp"
  $1762 = (FUNCTION_TABLE_ii[$load & 8191](454686)|0); //@line 4082 "libraries/glad/glad.cpp"
  HEAP32[164988] = $1762; //@line 4082 "libraries/glad/glad.cpp"
  $1763 = (FUNCTION_TABLE_ii[$load & 8191](615320)|0); //@line 4083 "libraries/glad/glad.cpp"
  HEAP32[164777] = $1763; //@line 4083 "libraries/glad/glad.cpp"
  $1764 = (FUNCTION_TABLE_ii[$load & 8191](454709)|0); //@line 4084 "libraries/glad/glad.cpp"
  HEAP32[165234] = $1764; //@line 4084 "libraries/glad/glad.cpp"
  $1765 = (FUNCTION_TABLE_ii[$load & 8191](615294)|0); //@line 4085 "libraries/glad/glad.cpp"
  HEAP32[164630] = $1765; //@line 4085 "libraries/glad/glad.cpp"
  $1766 = (FUNCTION_TABLE_ii[$load & 8191](615343)|0); //@line 4086 "libraries/glad/glad.cpp"
  HEAP32[164778] = $1766; //@line 4086 "libraries/glad/glad.cpp"
  $1767 = (FUNCTION_TABLE_ii[$load & 8191](616127)|0); //@line 4087 "libraries/glad/glad.cpp"
  HEAP32[165249] = $1767; //@line 4087 "libraries/glad/glad.cpp"
  $1768 = (FUNCTION_TABLE_ii[$load & 8191](454732)|0); //@line 4088 "libraries/glad/glad.cpp"
  HEAP32[164725] = $1768; //@line 4088 "libraries/glad/glad.cpp"
  $1769 = (FUNCTION_TABLE_ii[$load & 8191](454750)|0); //@line 4089 "libraries/glad/glad.cpp"
  HEAP32[165298] = $1769; //@line 4089 "libraries/glad/glad.cpp"
  $1770 = (FUNCTION_TABLE_ii[$load & 8191](454783)|0); //@line 4090 "libraries/glad/glad.cpp"
  HEAP32[164944] = $1770; //@line 4090 "libraries/glad/glad.cpp"
 }
 $1771 = HEAP8[1252067]|0; //@line 4096 "libraries/glad/glad.cpp"
 $1772 = ($1771<<24>>24)==(0); //@line 4096 "libraries/glad/glad.cpp"
 if (!($1772)) {
  $1773 = (FUNCTION_TABLE_ii[$load & 8191](477631)|0); //@line 4097 "libraries/glad/glad.cpp"
  HEAP32[165711] = $1773; //@line 4097 "libraries/glad/glad.cpp"
  $1774 = (FUNCTION_TABLE_ii[$load & 8191](477654)|0); //@line 4098 "libraries/glad/glad.cpp"
  HEAP32[165712] = $1774; //@line 4098 "libraries/glad/glad.cpp"
  $1775 = (FUNCTION_TABLE_ii[$load & 8191](477678)|0); //@line 4099 "libraries/glad/glad.cpp"
  HEAP32[165713] = $1775; //@line 4099 "libraries/glad/glad.cpp"
  $1776 = (FUNCTION_TABLE_ii[$load & 8191](477707)|0); //@line 4100 "libraries/glad/glad.cpp"
  HEAP32[165714] = $1776; //@line 4100 "libraries/glad/glad.cpp"
 }
 $1777 = HEAP8[1252068]|0; //@line 4105 "libraries/glad/glad.cpp"
 $1778 = ($1777<<24>>24)==(0); //@line 4105 "libraries/glad/glad.cpp"
 if (!($1778)) {
  $1779 = (FUNCTION_TABLE_ii[$load & 8191](457440)|0); //@line 4106 "libraries/glad/glad.cpp"
  HEAP32[165057] = $1779; //@line 4106 "libraries/glad/glad.cpp"
  $1780 = (FUNCTION_TABLE_ii[$load & 8191](457459)|0); //@line 4107 "libraries/glad/glad.cpp"
  HEAP32[165488] = $1780; //@line 4107 "libraries/glad/glad.cpp"
  $1781 = (FUNCTION_TABLE_ii[$load & 8191](457475)|0); //@line 4108 "libraries/glad/glad.cpp"
  HEAP32[164764] = $1781; //@line 4108 "libraries/glad/glad.cpp"
 }
 $1782 = HEAP8[1252069]|0; //@line 4113 "libraries/glad/glad.cpp"
 $1783 = ($1782<<24>>24)==(0); //@line 4113 "libraries/glad/glad.cpp"
 if (!($1783)) {
  $1784 = (FUNCTION_TABLE_ii[$load & 8191](463154)|0); //@line 4114 "libraries/glad/glad.cpp"
  HEAP32[165441] = $1784; //@line 4114 "libraries/glad/glad.cpp"
  $1785 = (FUNCTION_TABLE_ii[$load & 8191](463175)|0); //@line 4115 "libraries/glad/glad.cpp"
  HEAP32[165636] = $1785; //@line 4115 "libraries/glad/glad.cpp"
 }
 $1786 = HEAP8[1252071]|0; //@line 4121 "libraries/glad/glad.cpp"
 $1787 = ($1786<<24>>24)==(0); //@line 4121 "libraries/glad/glad.cpp"
 if (!($1787)) {
  $1788 = (FUNCTION_TABLE_ii[$load & 8191](456633)|0); //@line 4122 "libraries/glad/glad.cpp"
  HEAP32[165183] = $1788; //@line 4122 "libraries/glad/glad.cpp"
  $1789 = (FUNCTION_TABLE_ii[$load & 8191](456645)|0); //@line 4123 "libraries/glad/glad.cpp"
  HEAP32[165561] = $1789; //@line 4123 "libraries/glad/glad.cpp"
  $1790 = (FUNCTION_TABLE_ii[$load & 8191](456657)|0); //@line 4124 "libraries/glad/glad.cpp"
  HEAP32[164715] = $1790; //@line 4124 "libraries/glad/glad.cpp"
  $1791 = (FUNCTION_TABLE_ii[$load & 8191](456669)|0); //@line 4125 "libraries/glad/glad.cpp"
  HEAP32[164989] = $1791; //@line 4125 "libraries/glad/glad.cpp"
  $1792 = (FUNCTION_TABLE_ii[$load & 8191](456681)|0); //@line 4126 "libraries/glad/glad.cpp"
  HEAP32[164925] = $1792; //@line 4126 "libraries/glad/glad.cpp"
  $1793 = (FUNCTION_TABLE_ii[$load & 8191](456694)|0); //@line 4127 "libraries/glad/glad.cpp"
  HEAP32[164877] = $1793; //@line 4127 "libraries/glad/glad.cpp"
  $1794 = (FUNCTION_TABLE_ii[$load & 8191](456707)|0); //@line 4128 "libraries/glad/glad.cpp"
  HEAP32[165622] = $1794; //@line 4128 "libraries/glad/glad.cpp"
  $1795 = (FUNCTION_TABLE_ii[$load & 8191](456720)|0); //@line 4129 "libraries/glad/glad.cpp"
  HEAP32[165007] = $1795; //@line 4129 "libraries/glad/glad.cpp"
  $1796 = (FUNCTION_TABLE_ii[$load & 8191](456733)|0); //@line 4130 "libraries/glad/glad.cpp"
  HEAP32[165506] = $1796; //@line 4130 "libraries/glad/glad.cpp"
  $1797 = (FUNCTION_TABLE_ii[$load & 8191](456752)|0); //@line 4131 "libraries/glad/glad.cpp"
  HEAP32[164744] = $1797; //@line 4131 "libraries/glad/glad.cpp"
  $1798 = (FUNCTION_TABLE_ii[$load & 8191](456771)|0); //@line 4132 "libraries/glad/glad.cpp"
  HEAP32[165036] = $1798; //@line 4132 "libraries/glad/glad.cpp"
  $1799 = (FUNCTION_TABLE_ii[$load & 8191](456790)|0); //@line 4133 "libraries/glad/glad.cpp"
  HEAP32[164952] = $1799; //@line 4133 "libraries/glad/glad.cpp"
  $1800 = (FUNCTION_TABLE_ii[$load & 8191](456811)|0); //@line 4134 "libraries/glad/glad.cpp"
  HEAP32[165578] = $1800; //@line 4134 "libraries/glad/glad.cpp"
  $1801 = (FUNCTION_TABLE_ii[$load & 8191](456832)|0); //@line 4135 "libraries/glad/glad.cpp"
  HEAP32[165283] = $1801; //@line 4135 "libraries/glad/glad.cpp"
  $1802 = (FUNCTION_TABLE_ii[$load & 8191](456853)|0); //@line 4136 "libraries/glad/glad.cpp"
  HEAP32[165253] = $1802; //@line 4136 "libraries/glad/glad.cpp"
  $1803 = (FUNCTION_TABLE_ii[$load & 8191](456874)|0); //@line 4137 "libraries/glad/glad.cpp"
  HEAP32[165039] = $1803; //@line 4137 "libraries/glad/glad.cpp"
  $1804 = (FUNCTION_TABLE_ii[$load & 8191](456895)|0); //@line 4138 "libraries/glad/glad.cpp"
  HEAP32[165151] = $1804; //@line 4138 "libraries/glad/glad.cpp"
  $1805 = (FUNCTION_TABLE_ii[$load & 8191](456916)|0); //@line 4139 "libraries/glad/glad.cpp"
  HEAP32[165434] = $1805; //@line 4139 "libraries/glad/glad.cpp"
 }
 $1806 = HEAP8[1252072]|0; //@line 4144 "libraries/glad/glad.cpp"
 $1807 = ($1806<<24>>24)==(0); //@line 4144 "libraries/glad/glad.cpp"
 if (!($1807)) {
  $1808 = (FUNCTION_TABLE_ii[$load & 8191](477735)|0); //@line 4145 "libraries/glad/glad.cpp"
  HEAP32[165715] = $1808; //@line 4145 "libraries/glad/glad.cpp"
  $1809 = (FUNCTION_TABLE_ii[$load & 8191](477752)|0); //@line 4146 "libraries/glad/glad.cpp"
  HEAP32[165716] = $1809; //@line 4146 "libraries/glad/glad.cpp"
  $1810 = (FUNCTION_TABLE_ii[$load & 8191](477769)|0); //@line 4147 "libraries/glad/glad.cpp"
  HEAP32[165717] = $1810; //@line 4147 "libraries/glad/glad.cpp"
  $1811 = (FUNCTION_TABLE_ii[$load & 8191](477786)|0); //@line 4148 "libraries/glad/glad.cpp"
  HEAP32[165718] = $1811; //@line 4148 "libraries/glad/glad.cpp"
  $1812 = (FUNCTION_TABLE_ii[$load & 8191](477803)|0); //@line 4149 "libraries/glad/glad.cpp"
  HEAP32[165719] = $1812; //@line 4149 "libraries/glad/glad.cpp"
  $1813 = (FUNCTION_TABLE_ii[$load & 8191](477821)|0); //@line 4150 "libraries/glad/glad.cpp"
  HEAP32[165720] = $1813; //@line 4150 "libraries/glad/glad.cpp"
  $1814 = (FUNCTION_TABLE_ii[$load & 8191](477839)|0); //@line 4151 "libraries/glad/glad.cpp"
  HEAP32[165721] = $1814; //@line 4151 "libraries/glad/glad.cpp"
  $1815 = (FUNCTION_TABLE_ii[$load & 8191](477857)|0); //@line 4152 "libraries/glad/glad.cpp"
  HEAP32[165722] = $1815; //@line 4152 "libraries/glad/glad.cpp"
  $1816 = (FUNCTION_TABLE_ii[$load & 8191](477875)|0); //@line 4153 "libraries/glad/glad.cpp"
  HEAP32[165723] = $1816; //@line 4153 "libraries/glad/glad.cpp"
  $1817 = (FUNCTION_TABLE_ii[$load & 8191](477893)|0); //@line 4154 "libraries/glad/glad.cpp"
  HEAP32[165724] = $1817; //@line 4154 "libraries/glad/glad.cpp"
  $1818 = (FUNCTION_TABLE_ii[$load & 8191](477911)|0); //@line 4155 "libraries/glad/glad.cpp"
  HEAP32[165725] = $1818; //@line 4155 "libraries/glad/glad.cpp"
  $1819 = (FUNCTION_TABLE_ii[$load & 8191](477929)|0); //@line 4156 "libraries/glad/glad.cpp"
  HEAP32[165726] = $1819; //@line 4156 "libraries/glad/glad.cpp"
  $1820 = (FUNCTION_TABLE_ii[$load & 8191](477947)|0); //@line 4157 "libraries/glad/glad.cpp"
  HEAP32[165727] = $1820; //@line 4157 "libraries/glad/glad.cpp"
  $1821 = (FUNCTION_TABLE_ii[$load & 8191](477966)|0); //@line 4158 "libraries/glad/glad.cpp"
  HEAP32[165728] = $1821; //@line 4158 "libraries/glad/glad.cpp"
  $1822 = (FUNCTION_TABLE_ii[$load & 8191](477985)|0); //@line 4159 "libraries/glad/glad.cpp"
  HEAP32[165729] = $1822; //@line 4159 "libraries/glad/glad.cpp"
  $1823 = (FUNCTION_TABLE_ii[$load & 8191](478004)|0); //@line 4160 "libraries/glad/glad.cpp"
  HEAP32[165730] = $1823; //@line 4160 "libraries/glad/glad.cpp"
  $1824 = (FUNCTION_TABLE_ii[$load & 8191](478023)|0); //@line 4161 "libraries/glad/glad.cpp"
  HEAP32[165731] = $1824; //@line 4161 "libraries/glad/glad.cpp"
  $1825 = (FUNCTION_TABLE_ii[$load & 8191](478043)|0); //@line 4162 "libraries/glad/glad.cpp"
  HEAP32[165732] = $1825; //@line 4162 "libraries/glad/glad.cpp"
  $1826 = (FUNCTION_TABLE_ii[$load & 8191](478064)|0); //@line 4163 "libraries/glad/glad.cpp"
  HEAP32[165733] = $1826; //@line 4163 "libraries/glad/glad.cpp"
  $1827 = (FUNCTION_TABLE_ii[$load & 8191](478085)|0); //@line 4164 "libraries/glad/glad.cpp"
  HEAP32[165734] = $1827; //@line 4164 "libraries/glad/glad.cpp"
  $1828 = (FUNCTION_TABLE_ii[$load & 8191](478107)|0); //@line 4165 "libraries/glad/glad.cpp"
  HEAP32[165735] = $1828; //@line 4165 "libraries/glad/glad.cpp"
  $1829 = (FUNCTION_TABLE_ii[$load & 8191](478131)|0); //@line 4166 "libraries/glad/glad.cpp"
  HEAP32[165736] = $1829; //@line 4166 "libraries/glad/glad.cpp"
  $1830 = (FUNCTION_TABLE_ii[$load & 8191](478155)|0); //@line 4167 "libraries/glad/glad.cpp"
  HEAP32[165737] = $1830; //@line 4167 "libraries/glad/glad.cpp"
  $1831 = (FUNCTION_TABLE_ii[$load & 8191](478179)|0); //@line 4168 "libraries/glad/glad.cpp"
  HEAP32[165738] = $1831; //@line 4168 "libraries/glad/glad.cpp"
  $1832 = (FUNCTION_TABLE_ii[$load & 8191](478203)|0); //@line 4169 "libraries/glad/glad.cpp"
  HEAP32[165739] = $1832; //@line 4169 "libraries/glad/glad.cpp"
  $1833 = (FUNCTION_TABLE_ii[$load & 8191](478228)|0); //@line 4170 "libraries/glad/glad.cpp"
  HEAP32[165740] = $1833; //@line 4170 "libraries/glad/glad.cpp"
  $1834 = (FUNCTION_TABLE_ii[$load & 8191](478253)|0); //@line 4171 "libraries/glad/glad.cpp"
  HEAP32[165741] = $1834; //@line 4171 "libraries/glad/glad.cpp"
  $1835 = (FUNCTION_TABLE_ii[$load & 8191](478278)|0); //@line 4172 "libraries/glad/glad.cpp"
  HEAP32[165742] = $1835; //@line 4172 "libraries/glad/glad.cpp"
  $1836 = (FUNCTION_TABLE_ii[$load & 8191](478303)|0); //@line 4173 "libraries/glad/glad.cpp"
  HEAP32[165743] = $1836; //@line 4173 "libraries/glad/glad.cpp"
  $1837 = (FUNCTION_TABLE_ii[$load & 8191](478328)|0); //@line 4174 "libraries/glad/glad.cpp"
  HEAP32[165744] = $1837; //@line 4174 "libraries/glad/glad.cpp"
  $1838 = (FUNCTION_TABLE_ii[$load & 8191](478353)|0); //@line 4175 "libraries/glad/glad.cpp"
  HEAP32[165745] = $1838; //@line 4175 "libraries/glad/glad.cpp"
  $1839 = (FUNCTION_TABLE_ii[$load & 8191](478378)|0); //@line 4176 "libraries/glad/glad.cpp"
  HEAP32[165746] = $1839; //@line 4176 "libraries/glad/glad.cpp"
  $1840 = (FUNCTION_TABLE_ii[$load & 8191](478403)|0); //@line 4177 "libraries/glad/glad.cpp"
  HEAP32[165747] = $1840; //@line 4177 "libraries/glad/glad.cpp"
  $1841 = (FUNCTION_TABLE_ii[$load & 8191](478429)|0); //@line 4178 "libraries/glad/glad.cpp"
  HEAP32[165748] = $1841; //@line 4178 "libraries/glad/glad.cpp"
  $1842 = (FUNCTION_TABLE_ii[$load & 8191](478455)|0); //@line 4179 "libraries/glad/glad.cpp"
  HEAP32[165749] = $1842; //@line 4179 "libraries/glad/glad.cpp"
  $1843 = (FUNCTION_TABLE_ii[$load & 8191](478481)|0); //@line 4180 "libraries/glad/glad.cpp"
  HEAP32[165750] = $1843; //@line 4180 "libraries/glad/glad.cpp"
 }
 $1844 = HEAP8[1252075]|0; //@line 4187 "libraries/glad/glad.cpp"
 $1845 = ($1844<<24>>24)==(0); //@line 4187 "libraries/glad/glad.cpp"
 if (!($1845)) {
  $1846 = (FUNCTION_TABLE_ii[$load & 8191](478507)|0); //@line 4188 "libraries/glad/glad.cpp"
  HEAP32[165751] = $1846; //@line 4188 "libraries/glad/glad.cpp"
  $1847 = (FUNCTION_TABLE_ii[$load & 8191](478541)|0); //@line 4189 "libraries/glad/glad.cpp"
  HEAP32[165752] = $1847; //@line 4189 "libraries/glad/glad.cpp"
 }
 $1848 = HEAP8[1252076]|0; //@line 4194 "libraries/glad/glad.cpp"
 $1849 = ($1848<<24>>24)==(0); //@line 4194 "libraries/glad/glad.cpp"
 if (!($1849)) {
  $1850 = (FUNCTION_TABLE_ii[$load & 8191](478577)|0); //@line 4195 "libraries/glad/glad.cpp"
  HEAP32[165753] = $1850; //@line 4195 "libraries/glad/glad.cpp"
 }
 $1851 = HEAP8[1252077]|0; //@line 4200 "libraries/glad/glad.cpp"
 $1852 = ($1851<<24>>24)==(0); //@line 4200 "libraries/glad/glad.cpp"
 if (!($1852)) {
  $1853 = (FUNCTION_TABLE_ii[$load & 8191](459332)|0); //@line 4201 "libraries/glad/glad.cpp"
  HEAP32[164979] = $1853; //@line 4201 "libraries/glad/glad.cpp"
 }
 $1854 = HEAP8[1252078]|0; //@line 4206 "libraries/glad/glad.cpp"
 $1855 = ($1854<<24>>24)==(0); //@line 4206 "libraries/glad/glad.cpp"
 if (!($1855)) {
  $1856 = (FUNCTION_TABLE_ii[$load & 8191](459693)|0); //@line 4207 "libraries/glad/glad.cpp"
  HEAP32[165087] = $1856; //@line 4207 "libraries/glad/glad.cpp"
 }
 $1857 = HEAP8[1252079]|0; //@line 4212 "libraries/glad/glad.cpp"
 $1858 = ($1857<<24>>24)==(0); //@line 4212 "libraries/glad/glad.cpp"
 if (!($1858)) {
  $1859 = (FUNCTION_TABLE_ii[$load & 8191](459717)|0); //@line 4213 "libraries/glad/glad.cpp"
  HEAP32[165375] = $1859; //@line 4213 "libraries/glad/glad.cpp"
  $1860 = (FUNCTION_TABLE_ii[$load & 8191](459741)|0); //@line 4214 "libraries/glad/glad.cpp"
  HEAP32[165102] = $1860; //@line 4214 "libraries/glad/glad.cpp"
  $1861 = (FUNCTION_TABLE_ii[$load & 8191](459762)|0); //@line 4215 "libraries/glad/glad.cpp"
  HEAP32[164749] = $1861; //@line 4215 "libraries/glad/glad.cpp"
  $1862 = (FUNCTION_TABLE_ii[$load & 8191](459788)|0); //@line 4216 "libraries/glad/glad.cpp"
  HEAP32[165010] = $1862; //@line 4216 "libraries/glad/glad.cpp"
  $1863 = (FUNCTION_TABLE_ii[$load & 8191](459811)|0); //@line 4217 "libraries/glad/glad.cpp"
  HEAP32[164995] = $1863; //@line 4217 "libraries/glad/glad.cpp"
  $1864 = (FUNCTION_TABLE_ii[$load & 8191](459835)|0); //@line 4218 "libraries/glad/glad.cpp"
  HEAP32[165022] = $1864; //@line 4218 "libraries/glad/glad.cpp"
 }
 $1865 = HEAP8[1252081]|0; //@line 4224 "libraries/glad/glad.cpp"
 $1866 = ($1865<<24>>24)==(0); //@line 4224 "libraries/glad/glad.cpp"
 if (!($1866)) {
  $1867 = (FUNCTION_TABLE_ii[$load & 8191](454809)|0); //@line 4225 "libraries/glad/glad.cpp"
  HEAP32[165517] = $1867; //@line 4225 "libraries/glad/glad.cpp"
  $1868 = (FUNCTION_TABLE_ii[$load & 8191](454826)|0); //@line 4226 "libraries/glad/glad.cpp"
  HEAP32[164950] = $1868; //@line 4226 "libraries/glad/glad.cpp"
 }
 $1869 = HEAP8[1252082]|0; //@line 4231 "libraries/glad/glad.cpp"
 $1870 = ($1869<<24>>24)==(0); //@line 4231 "libraries/glad/glad.cpp"
 if (!($1870)) {
  $1871 = (FUNCTION_TABLE_ii[$load & 8191](460556)|0); //@line 4232 "libraries/glad/glad.cpp"
  HEAP32[165625] = $1871; //@line 4232 "libraries/glad/glad.cpp"
  $1872 = (FUNCTION_TABLE_ii[$load & 8191](460574)|0); //@line 4233 "libraries/glad/glad.cpp"
  HEAP32[164971] = $1872; //@line 4233 "libraries/glad/glad.cpp"
  $1873 = (FUNCTION_TABLE_ii[$load & 8191](460593)|0); //@line 4234 "libraries/glad/glad.cpp"
  HEAP32[164996] = $1873; //@line 4234 "libraries/glad/glad.cpp"
  $1874 = (FUNCTION_TABLE_ii[$load & 8191](460608)|0); //@line 4235 "libraries/glad/glad.cpp"
  HEAP32[165492] = $1874; //@line 4235 "libraries/glad/glad.cpp"
  $1875 = (FUNCTION_TABLE_ii[$load & 8191](460623)|0); //@line 4236 "libraries/glad/glad.cpp"
  HEAP32[165379] = $1875; //@line 4236 "libraries/glad/glad.cpp"
  $1876 = (FUNCTION_TABLE_ii[$load & 8191](460643)|0); //@line 4237 "libraries/glad/glad.cpp"
  HEAP32[164908] = $1876; //@line 4237 "libraries/glad/glad.cpp"
 }
 $1877 = HEAP8[1252083]|0; //@line 4242 "libraries/glad/glad.cpp"
 $1878 = ($1877<<24>>24)==(0); //@line 4242 "libraries/glad/glad.cpp"
 if (!($1878)) {
  $1879 = (FUNCTION_TABLE_ii[$load & 8191](459862)|0); //@line 4243 "libraries/glad/glad.cpp"
  HEAP32[164703] = $1879; //@line 4243 "libraries/glad/glad.cpp"
  $1880 = (FUNCTION_TABLE_ii[$load & 8191](459888)|0); //@line 4244 "libraries/glad/glad.cpp"
  HEAP32[165312] = $1880; //@line 4244 "libraries/glad/glad.cpp"
 }
 $1881 = HEAP8[1252085]|0; //@line 4250 "libraries/glad/glad.cpp"
 $1882 = ($1881<<24>>24)==(0); //@line 4250 "libraries/glad/glad.cpp"
 if (!($1882)) {
  $1883 = (FUNCTION_TABLE_ii[$load & 8191](478602)|0); //@line 4251 "libraries/glad/glad.cpp"
  HEAP32[165754] = $1883; //@line 4251 "libraries/glad/glad.cpp"
 }
 $1884 = HEAP8[1252089]|0; //@line 4259 "libraries/glad/glad.cpp"
 $1885 = ($1884<<24>>24)==(0); //@line 4259 "libraries/glad/glad.cpp"
 if (!($1885)) {
  $1886 = (FUNCTION_TABLE_ii[$load & 8191](459916)|0); //@line 4260 "libraries/glad/glad.cpp"
  HEAP32[164843] = $1886; //@line 4260 "libraries/glad/glad.cpp"
  $1887 = (FUNCTION_TABLE_ii[$load & 8191](459940)|0); //@line 4261 "libraries/glad/glad.cpp"
  HEAP32[165420] = $1887; //@line 4261 "libraries/glad/glad.cpp"
  $1888 = (FUNCTION_TABLE_ii[$load & 8191](459966)|0); //@line 4262 "libraries/glad/glad.cpp"
  HEAP32[165150] = $1888; //@line 4262 "libraries/glad/glad.cpp"
  $1889 = (FUNCTION_TABLE_ii[$load & 8191](459991)|0); //@line 4263 "libraries/glad/glad.cpp"
  HEAP32[165663] = $1889; //@line 4263 "libraries/glad/glad.cpp"
  $1890 = (FUNCTION_TABLE_ii[$load & 8191](460014)|0); //@line 4264 "libraries/glad/glad.cpp"
  HEAP32[165155] = $1890; //@line 4264 "libraries/glad/glad.cpp"
  $1891 = (FUNCTION_TABLE_ii[$load & 8191](460043)|0); //@line 4265 "libraries/glad/glad.cpp"
  HEAP32[165458] = $1891; //@line 4265 "libraries/glad/glad.cpp"
 }
 $1892 = HEAP8[1252090]|0; //@line 4270 "libraries/glad/glad.cpp"
 $1893 = ($1892<<24>>24)==(0); //@line 4270 "libraries/glad/glad.cpp"
 if (!($1893)) {
  $1894 = (FUNCTION_TABLE_ii[$load & 8191](455206)|0); //@line 4271 "libraries/glad/glad.cpp"
  HEAP32[165230] = $1894; //@line 4271 "libraries/glad/glad.cpp"
 }
 $1895 = HEAP8[1252093]|0; //@line 4278 "libraries/glad/glad.cpp"
 $1896 = ($1895<<24>>24)==(0); //@line 4278 "libraries/glad/glad.cpp"
 if (!($1896)) {
  $1897 = (FUNCTION_TABLE_ii[$load & 8191](478632)|0); //@line 4279 "libraries/glad/glad.cpp"
  HEAP32[165755] = $1897; //@line 4279 "libraries/glad/glad.cpp"
  $1898 = (FUNCTION_TABLE_ii[$load & 8191](478660)|0); //@line 4280 "libraries/glad/glad.cpp"
  HEAP32[165756] = $1898; //@line 4280 "libraries/glad/glad.cpp"
  $1899 = (FUNCTION_TABLE_ii[$load & 8191](478678)|0); //@line 4281 "libraries/glad/glad.cpp"
  HEAP32[165757] = $1899; //@line 4281 "libraries/glad/glad.cpp"
  $1900 = (FUNCTION_TABLE_ii[$load & 8191](478695)|0); //@line 4282 "libraries/glad/glad.cpp"
  HEAP32[165758] = $1900; //@line 4282 "libraries/glad/glad.cpp"
  $1901 = (FUNCTION_TABLE_ii[$load & 8191](478723)|0); //@line 4283 "libraries/glad/glad.cpp"
  HEAP32[165759] = $1901; //@line 4283 "libraries/glad/glad.cpp"
  $1902 = (FUNCTION_TABLE_ii[$load & 8191](478742)|0); //@line 4284 "libraries/glad/glad.cpp"
  HEAP32[165760] = $1902; //@line 4284 "libraries/glad/glad.cpp"
  $1903 = (FUNCTION_TABLE_ii[$load & 8191](478761)|0); //@line 4285 "libraries/glad/glad.cpp"
  HEAP32[165761] = $1903; //@line 4285 "libraries/glad/glad.cpp"
  $1904 = (FUNCTION_TABLE_ii[$load & 8191](478781)|0); //@line 4286 "libraries/glad/glad.cpp"
  HEAP32[165762] = $1904; //@line 4286 "libraries/glad/glad.cpp"
  $1905 = (FUNCTION_TABLE_ii[$load & 8191](478800)|0); //@line 4287 "libraries/glad/glad.cpp"
  HEAP32[165763] = $1905; //@line 4287 "libraries/glad/glad.cpp"
  $1906 = (FUNCTION_TABLE_ii[$load & 8191](478815)|0); //@line 4288 "libraries/glad/glad.cpp"
  HEAP32[165764] = $1906; //@line 4288 "libraries/glad/glad.cpp"
  $1907 = (FUNCTION_TABLE_ii[$load & 8191](478830)|0); //@line 4289 "libraries/glad/glad.cpp"
  HEAP32[165765] = $1907; //@line 4289 "libraries/glad/glad.cpp"
  $1908 = (FUNCTION_TABLE_ii[$load & 8191](478845)|0); //@line 4290 "libraries/glad/glad.cpp"
  HEAP32[165766] = $1908; //@line 4290 "libraries/glad/glad.cpp"
  $1909 = (FUNCTION_TABLE_ii[$load & 8191](478865)|0); //@line 4291 "libraries/glad/glad.cpp"
  HEAP32[165767] = $1909; //@line 4291 "libraries/glad/glad.cpp"
  $1910 = (FUNCTION_TABLE_ii[$load & 8191](478886)|0); //@line 4292 "libraries/glad/glad.cpp"
  HEAP32[165768] = $1910; //@line 4292 "libraries/glad/glad.cpp"
  $1911 = (FUNCTION_TABLE_ii[$load & 8191](478907)|0); //@line 4293 "libraries/glad/glad.cpp"
  HEAP32[165769] = $1911; //@line 4293 "libraries/glad/glad.cpp"
  $1912 = (FUNCTION_TABLE_ii[$load & 8191](478931)|0); //@line 4294 "libraries/glad/glad.cpp"
  HEAP32[165770] = $1912; //@line 4294 "libraries/glad/glad.cpp"
  $1913 = (FUNCTION_TABLE_ii[$load & 8191](478951)|0); //@line 4295 "libraries/glad/glad.cpp"
  HEAP32[165771] = $1913; //@line 4295 "libraries/glad/glad.cpp"
  $1914 = (FUNCTION_TABLE_ii[$load & 8191](478978)|0); //@line 4296 "libraries/glad/glad.cpp"
  HEAP32[165772] = $1914; //@line 4296 "libraries/glad/glad.cpp"
  $1915 = (FUNCTION_TABLE_ii[$load & 8191](479003)|0); //@line 4297 "libraries/glad/glad.cpp"
  HEAP32[165773] = $1915; //@line 4297 "libraries/glad/glad.cpp"
  $1916 = (FUNCTION_TABLE_ii[$load & 8191](479022)|0); //@line 4298 "libraries/glad/glad.cpp"
  HEAP32[165774] = $1916; //@line 4298 "libraries/glad/glad.cpp"
 }
 $1917 = HEAP8[1252095]|0; //@line 4304 "libraries/glad/glad.cpp"
 $1918 = ($1917<<24>>24)==(0); //@line 4304 "libraries/glad/glad.cpp"
 if (!($1918)) {
  $1919 = (FUNCTION_TABLE_ii[$load & 8191](479038)|0); //@line 4305 "libraries/glad/glad.cpp"
  HEAP32[165775] = $1919; //@line 4305 "libraries/glad/glad.cpp"
  $1920 = (FUNCTION_TABLE_ii[$load & 8191](479072)|0); //@line 4306 "libraries/glad/glad.cpp"
  HEAP32[165776] = $1920; //@line 4306 "libraries/glad/glad.cpp"
  $1921 = (FUNCTION_TABLE_ii[$load & 8191](479111)|0); //@line 4307 "libraries/glad/glad.cpp"
  HEAP32[165777] = $1921; //@line 4307 "libraries/glad/glad.cpp"
 }
 $1922 = HEAP8[1252096]|0; //@line 4312 "libraries/glad/glad.cpp"
 $1923 = ($1922<<24>>24)==(0); //@line 4312 "libraries/glad/glad.cpp"
 if (!($1923)) {
  $1924 = (FUNCTION_TABLE_ii[$load & 8191](479136)|0); //@line 4313 "libraries/glad/glad.cpp"
  HEAP32[165778] = $1924; //@line 4313 "libraries/glad/glad.cpp"
 }
 $1925 = HEAP8[1252097]|0; //@line 4318 "libraries/glad/glad.cpp"
 $1926 = ($1925<<24>>24)==(0); //@line 4318 "libraries/glad/glad.cpp"
 if (!($1926)) {
  $1927 = (FUNCTION_TABLE_ii[$load & 8191](455508)|0); //@line 4319 "libraries/glad/glad.cpp"
  HEAP32[164913] = $1927; //@line 4319 "libraries/glad/glad.cpp"
  $1928 = (FUNCTION_TABLE_ii[$load & 8191](455522)|0); //@line 4320 "libraries/glad/glad.cpp"
  HEAP32[164709] = $1928; //@line 4320 "libraries/glad/glad.cpp"
  $1929 = (FUNCTION_TABLE_ii[$load & 8191](455539)|0); //@line 4321 "libraries/glad/glad.cpp"
  HEAP32[165489] = $1929; //@line 4321 "libraries/glad/glad.cpp"
  $1930 = (FUNCTION_TABLE_ii[$load & 8191](455551)|0); //@line 4322 "libraries/glad/glad.cpp"
  HEAP32[164640] = $1930; //@line 4322 "libraries/glad/glad.cpp"
  $1931 = (FUNCTION_TABLE_ii[$load & 8191](455565)|0); //@line 4323 "libraries/glad/glad.cpp"
  HEAP32[165177] = $1931; //@line 4323 "libraries/glad/glad.cpp"
  $1932 = (FUNCTION_TABLE_ii[$load & 8191](455585)|0); //@line 4324 "libraries/glad/glad.cpp"
  HEAP32[165255] = $1932; //@line 4324 "libraries/glad/glad.cpp"
  $1933 = (FUNCTION_TABLE_ii[$load & 8191](455606)|0); //@line 4325 "libraries/glad/glad.cpp"
  HEAP32[165179] = $1933; //@line 4325 "libraries/glad/glad.cpp"
  $1934 = (FUNCTION_TABLE_ii[$load & 8191](455626)|0); //@line 4326 "libraries/glad/glad.cpp"
  HEAP32[165305] = $1934; //@line 4326 "libraries/glad/glad.cpp"
  $1935 = (FUNCTION_TABLE_ii[$load & 8191](455647)|0); //@line 4327 "libraries/glad/glad.cpp"
  HEAP32[165072] = $1935; //@line 4327 "libraries/glad/glad.cpp"
  $1936 = (FUNCTION_TABLE_ii[$load & 8191](455669)|0); //@line 4328 "libraries/glad/glad.cpp"
  HEAP32[165411] = $1936; //@line 4328 "libraries/glad/glad.cpp"
  $1937 = (FUNCTION_TABLE_ii[$load & 8191](455692)|0); //@line 4329 "libraries/glad/glad.cpp"
  HEAP32[165471] = $1937; //@line 4329 "libraries/glad/glad.cpp"
  $1938 = (FUNCTION_TABLE_ii[$load & 8191](455716)|0); //@line 4330 "libraries/glad/glad.cpp"
  HEAP32[164732] = $1938; //@line 4330 "libraries/glad/glad.cpp"
  $1939 = (FUNCTION_TABLE_ii[$load & 8191](455741)|0); //@line 4331 "libraries/glad/glad.cpp"
  HEAP32[164818] = $1939; //@line 4331 "libraries/glad/glad.cpp"
  $1940 = (FUNCTION_TABLE_ii[$load & 8191](455765)|0); //@line 4332 "libraries/glad/glad.cpp"
  HEAP32[165429] = $1940; //@line 4332 "libraries/glad/glad.cpp"
 }
 $1941 = HEAP8[1252100]|0; //@line 4339 "libraries/glad/glad.cpp"
 $1942 = ($1941<<24>>24)==(0); //@line 4339 "libraries/glad/glad.cpp"
 if (!($1942)) {
  $1943 = (FUNCTION_TABLE_ii[$load & 8191](457495)|0); //@line 4340 "libraries/glad/glad.cpp"
  HEAP32[165329] = $1943; //@line 4340 "libraries/glad/glad.cpp"
  $1944 = (FUNCTION_TABLE_ii[$load & 8191](457514)|0); //@line 4341 "libraries/glad/glad.cpp"
  HEAP32[164742] = $1944; //@line 4341 "libraries/glad/glad.cpp"
  $1945 = (FUNCTION_TABLE_ii[$load & 8191](457536)|0); //@line 4342 "libraries/glad/glad.cpp"
  HEAP32[165317] = $1945; //@line 4342 "libraries/glad/glad.cpp"
  $1946 = (FUNCTION_TABLE_ii[$load & 8191](457559)|0); //@line 4343 "libraries/glad/glad.cpp"
  HEAP32[165382] = $1946; //@line 4343 "libraries/glad/glad.cpp"
  $1947 = (FUNCTION_TABLE_ii[$load & 8191](457581)|0); //@line 4344 "libraries/glad/glad.cpp"
  HEAP32[165660] = $1947; //@line 4344 "libraries/glad/glad.cpp"
  $1948 = (FUNCTION_TABLE_ii[$load & 8191](457606)|0); //@line 4345 "libraries/glad/glad.cpp"
  HEAP32[165122] = $1948; //@line 4345 "libraries/glad/glad.cpp"
  $1949 = (FUNCTION_TABLE_ii[$load & 8191](457628)|0); //@line 4346 "libraries/glad/glad.cpp"
  HEAP32[165291] = $1949; //@line 4346 "libraries/glad/glad.cpp"
  $1950 = (FUNCTION_TABLE_ii[$load & 8191](457648)|0); //@line 4347 "libraries/glad/glad.cpp"
  HEAP32[165338] = $1950; //@line 4347 "libraries/glad/glad.cpp"
  $1951 = (FUNCTION_TABLE_ii[$load & 8191](457671)|0); //@line 4348 "libraries/glad/glad.cpp"
  HEAP32[164876] = $1951; //@line 4348 "libraries/glad/glad.cpp"
  $1952 = (FUNCTION_TABLE_ii[$load & 8191](457690)|0); //@line 4349 "libraries/glad/glad.cpp"
  HEAP32[165269] = $1952; //@line 4349 "libraries/glad/glad.cpp"
  $1953 = (FUNCTION_TABLE_ii[$load & 8191](457710)|0); //@line 4350 "libraries/glad/glad.cpp"
  HEAP32[164879] = $1953; //@line 4350 "libraries/glad/glad.cpp"
  $1954 = (FUNCTION_TABLE_ii[$load & 8191](457729)|0); //@line 4351 "libraries/glad/glad.cpp"
  HEAP32[165648] = $1954; //@line 4351 "libraries/glad/glad.cpp"
  $1955 = (FUNCTION_TABLE_ii[$load & 8191](457749)|0); //@line 4352 "libraries/glad/glad.cpp"
  HEAP32[164878] = $1955; //@line 4352 "libraries/glad/glad.cpp"
  $1956 = (FUNCTION_TABLE_ii[$load & 8191](457768)|0); //@line 4353 "libraries/glad/glad.cpp"
  HEAP32[165173] = $1956; //@line 4353 "libraries/glad/glad.cpp"
  $1957 = (FUNCTION_TABLE_ii[$load & 8191](457788)|0); //@line 4354 "libraries/glad/glad.cpp"
  HEAP32[164943] = $1957; //@line 4354 "libraries/glad/glad.cpp"
  $1958 = (FUNCTION_TABLE_ii[$load & 8191](457808)|0); //@line 4355 "libraries/glad/glad.cpp"
  HEAP32[164808] = $1958; //@line 4355 "libraries/glad/glad.cpp"
  $1959 = (FUNCTION_TABLE_ii[$load & 8191](457829)|0); //@line 4356 "libraries/glad/glad.cpp"
  HEAP32[165552] = $1959; //@line 4356 "libraries/glad/glad.cpp"
  $1960 = (FUNCTION_TABLE_ii[$load & 8191](457848)|0); //@line 4357 "libraries/glad/glad.cpp"
  HEAP32[164812] = $1960; //@line 4357 "libraries/glad/glad.cpp"
  $1961 = (FUNCTION_TABLE_ii[$load & 8191](457868)|0); //@line 4358 "libraries/glad/glad.cpp"
  HEAP32[165139] = $1961; //@line 4358 "libraries/glad/glad.cpp"
  $1962 = (FUNCTION_TABLE_ii[$load & 8191](457887)|0); //@line 4359 "libraries/glad/glad.cpp"
  HEAP32[164951] = $1962; //@line 4359 "libraries/glad/glad.cpp"
  $1963 = (FUNCTION_TABLE_ii[$load & 8191](457907)|0); //@line 4360 "libraries/glad/glad.cpp"
  HEAP32[165050] = $1963; //@line 4360 "libraries/glad/glad.cpp"
  $1964 = (FUNCTION_TABLE_ii[$load & 8191](457926)|0); //@line 4361 "libraries/glad/glad.cpp"
  HEAP32[164984] = $1964; //@line 4361 "libraries/glad/glad.cpp"
  $1965 = (FUNCTION_TABLE_ii[$load & 8191](457946)|0); //@line 4362 "libraries/glad/glad.cpp"
  HEAP32[165598] = $1965; //@line 4362 "libraries/glad/glad.cpp"
  $1966 = (FUNCTION_TABLE_ii[$load & 8191](457966)|0); //@line 4363 "libraries/glad/glad.cpp"
  HEAP32[165548] = $1966; //@line 4363 "libraries/glad/glad.cpp"
  $1967 = (FUNCTION_TABLE_ii[$load & 8191](457987)|0); //@line 4364 "libraries/glad/glad.cpp"
  HEAP32[165144] = $1967; //@line 4364 "libraries/glad/glad.cpp"
  $1968 = (FUNCTION_TABLE_ii[$load & 8191](458006)|0); //@line 4365 "libraries/glad/glad.cpp"
  HEAP32[164886] = $1968; //@line 4365 "libraries/glad/glad.cpp"
  $1969 = (FUNCTION_TABLE_ii[$load & 8191](458026)|0); //@line 4366 "libraries/glad/glad.cpp"
  HEAP32[165334] = $1969; //@line 4366 "libraries/glad/glad.cpp"
  $1970 = (FUNCTION_TABLE_ii[$load & 8191](458045)|0); //@line 4367 "libraries/glad/glad.cpp"
  HEAP32[165408] = $1970; //@line 4367 "libraries/glad/glad.cpp"
  $1971 = (FUNCTION_TABLE_ii[$load & 8191](458065)|0); //@line 4368 "libraries/glad/glad.cpp"
  HEAP32[165335] = $1971; //@line 4368 "libraries/glad/glad.cpp"
  $1972 = (FUNCTION_TABLE_ii[$load & 8191](458084)|0); //@line 4369 "libraries/glad/glad.cpp"
  HEAP32[164667] = $1972; //@line 4369 "libraries/glad/glad.cpp"
  $1973 = (FUNCTION_TABLE_ii[$load & 8191](458104)|0); //@line 4370 "libraries/glad/glad.cpp"
  HEAP32[164855] = $1973; //@line 4370 "libraries/glad/glad.cpp"
  $1974 = (FUNCTION_TABLE_ii[$load & 8191](458124)|0); //@line 4371 "libraries/glad/glad.cpp"
  HEAP32[165413] = $1974; //@line 4371 "libraries/glad/glad.cpp"
  $1975 = (FUNCTION_TABLE_ii[$load & 8191](458145)|0); //@line 4372 "libraries/glad/glad.cpp"
  HEAP32[165209] = $1975; //@line 4372 "libraries/glad/glad.cpp"
  $1976 = (FUNCTION_TABLE_ii[$load & 8191](458164)|0); //@line 4373 "libraries/glad/glad.cpp"
  HEAP32[165053] = $1976; //@line 4373 "libraries/glad/glad.cpp"
  $1977 = (FUNCTION_TABLE_ii[$load & 8191](458184)|0); //@line 4374 "libraries/glad/glad.cpp"
  HEAP32[165211] = $1977; //@line 4374 "libraries/glad/glad.cpp"
  $1978 = (FUNCTION_TABLE_ii[$load & 8191](458203)|0); //@line 4375 "libraries/glad/glad.cpp"
  HEAP32[165593] = $1978; //@line 4375 "libraries/glad/glad.cpp"
  $1979 = (FUNCTION_TABLE_ii[$load & 8191](458223)|0); //@line 4376 "libraries/glad/glad.cpp"
  HEAP32[165213] = $1979; //@line 4376 "libraries/glad/glad.cpp"
  $1980 = (FUNCTION_TABLE_ii[$load & 8191](458242)|0); //@line 4377 "libraries/glad/glad.cpp"
  HEAP32[165402] = $1980; //@line 4377 "libraries/glad/glad.cpp"
  $1981 = (FUNCTION_TABLE_ii[$load & 8191](458262)|0); //@line 4378 "libraries/glad/glad.cpp"
  HEAP32[165599] = $1981; //@line 4378 "libraries/glad/glad.cpp"
  $1982 = (FUNCTION_TABLE_ii[$load & 8191](458282)|0); //@line 4379 "libraries/glad/glad.cpp"
  HEAP32[165198] = $1982; //@line 4379 "libraries/glad/glad.cpp"
  $1983 = (FUNCTION_TABLE_ii[$load & 8191](458303)|0); //@line 4380 "libraries/glad/glad.cpp"
  HEAP32[165251] = $1983; //@line 4380 "libraries/glad/glad.cpp"
  $1984 = (FUNCTION_TABLE_ii[$load & 8191](458329)|0); //@line 4381 "libraries/glad/glad.cpp"
  HEAP32[164993] = $1984; //@line 4381 "libraries/glad/glad.cpp"
  $1985 = (FUNCTION_TABLE_ii[$load & 8191](458355)|0); //@line 4382 "libraries/glad/glad.cpp"
  HEAP32[164737] = $1985; //@line 4382 "libraries/glad/glad.cpp"
  $1986 = (FUNCTION_TABLE_ii[$load & 8191](458381)|0); //@line 4383 "libraries/glad/glad.cpp"
  HEAP32[164675] = $1986; //@line 4383 "libraries/glad/glad.cpp"
  $1987 = (FUNCTION_TABLE_ii[$load & 8191](458407)|0); //@line 4384 "libraries/glad/glad.cpp"
  HEAP32[165229] = $1987; //@line 4384 "libraries/glad/glad.cpp"
  $1988 = (FUNCTION_TABLE_ii[$load & 8191](458433)|0); //@line 4385 "libraries/glad/glad.cpp"
  HEAP32[164953] = $1988; //@line 4385 "libraries/glad/glad.cpp"
  $1989 = (FUNCTION_TABLE_ii[$load & 8191](458459)|0); //@line 4386 "libraries/glad/glad.cpp"
  HEAP32[165399] = $1989; //@line 4386 "libraries/glad/glad.cpp"
  $1990 = (FUNCTION_TABLE_ii[$load & 8191](458487)|0); //@line 4387 "libraries/glad/glad.cpp"
  HEAP32[164731] = $1990; //@line 4387 "libraries/glad/glad.cpp"
  $1991 = (FUNCTION_TABLE_ii[$load & 8191](458515)|0); //@line 4388 "libraries/glad/glad.cpp"
  HEAP32[165535] = $1991; //@line 4388 "libraries/glad/glad.cpp"
  $1992 = (FUNCTION_TABLE_ii[$load & 8191](458543)|0); //@line 4389 "libraries/glad/glad.cpp"
  HEAP32[165520] = $1992; //@line 4389 "libraries/glad/glad.cpp"
  $1993 = (FUNCTION_TABLE_ii[$load & 8191](458571)|0); //@line 4390 "libraries/glad/glad.cpp"
  HEAP32[165021] = $1993; //@line 4390 "libraries/glad/glad.cpp"
  $1994 = (FUNCTION_TABLE_ii[$load & 8191](458599)|0); //@line 4391 "libraries/glad/glad.cpp"
  HEAP32[165623] = $1994; //@line 4391 "libraries/glad/glad.cpp"
  $1995 = (FUNCTION_TABLE_ii[$load & 8191](458627)|0); //@line 4392 "libraries/glad/glad.cpp"
  HEAP32[165130] = $1995; //@line 4392 "libraries/glad/glad.cpp"
  $1996 = (FUNCTION_TABLE_ii[$load & 8191](458655)|0); //@line 4393 "libraries/glad/glad.cpp"
  HEAP32[164728] = $1996; //@line 4393 "libraries/glad/glad.cpp"
  $1997 = (FUNCTION_TABLE_ii[$load & 8191](458683)|0); //@line 4394 "libraries/glad/glad.cpp"
  HEAP32[164956] = $1997; //@line 4394 "libraries/glad/glad.cpp"
  $1998 = (FUNCTION_TABLE_ii[$load & 8191](458711)|0); //@line 4395 "libraries/glad/glad.cpp"
  HEAP32[164823] = $1998; //@line 4395 "libraries/glad/glad.cpp"
  $1999 = (FUNCTION_TABLE_ii[$load & 8191](458739)|0); //@line 4396 "libraries/glad/glad.cpp"
  HEAP32[164802] = $1999; //@line 4396 "libraries/glad/glad.cpp"
  $2000 = (FUNCTION_TABLE_ii[$load & 8191](458767)|0); //@line 4397 "libraries/glad/glad.cpp"
  HEAP32[165051] = $2000; //@line 4397 "libraries/glad/glad.cpp"
  $2001 = (FUNCTION_TABLE_ii[$load & 8191](458795)|0); //@line 4398 "libraries/glad/glad.cpp"
  HEAP32[164901] = $2001; //@line 4398 "libraries/glad/glad.cpp"
  $2002 = (FUNCTION_TABLE_ii[$load & 8191](458821)|0); //@line 4399 "libraries/glad/glad.cpp"
  HEAP32[165019] = $2002; //@line 4399 "libraries/glad/glad.cpp"
 }
 $2003 = HEAP8[1252102]|0; //@line 4405 "libraries/glad/glad.cpp"
 $2004 = ($2003<<24>>24)==(0); //@line 4405 "libraries/glad/glad.cpp"
 if (!($2004)) {
  $2005 = (FUNCTION_TABLE_ii[$load & 8191](459354)|0); //@line 4406 "libraries/glad/glad.cpp"
  HEAP32[165158] = $2005; //@line 4406 "libraries/glad/glad.cpp"
 }
 $2006 = HEAP8[1252108]|0; //@line 4416 "libraries/glad/glad.cpp"
 $2007 = ($2006<<24>>24)==(0); //@line 4416 "libraries/glad/glad.cpp"
 if (!($2007)) {
  $2008 = (FUNCTION_TABLE_ii[$load & 8191](459387)|0); //@line 4417 "libraries/glad/glad.cpp"
  HEAP32[164735] = $2008; //@line 4417 "libraries/glad/glad.cpp"
  $2009 = (FUNCTION_TABLE_ii[$load & 8191](459406)|0); //@line 4418 "libraries/glad/glad.cpp"
  HEAP32[164771] = $2009; //@line 4418 "libraries/glad/glad.cpp"
 }
 $2010 = HEAP8[1252112]|0; //@line 4426 "libraries/glad/glad.cpp"
 $2011 = ($2010<<24>>24)==(0); //@line 4426 "libraries/glad/glad.cpp"
 if (!($2011)) {
  $2012 = (FUNCTION_TABLE_ii[$load & 8191](460077)|0); //@line 4427 "libraries/glad/glad.cpp"
  HEAP32[164820] = $2012; //@line 4427 "libraries/glad/glad.cpp"
 }
 $2013 = HEAP8[1252113]|0; //@line 4432 "libraries/glad/glad.cpp"
 $2014 = ($2013<<24>>24)==(0); //@line 4432 "libraries/glad/glad.cpp"
 if (!($2014)) {
  $2015 = (FUNCTION_TABLE_ii[$load & 8191](456931)|0); //@line 4433 "libraries/glad/glad.cpp"
  HEAP32[165627] = $2015; //@line 4433 "libraries/glad/glad.cpp"
  $2016 = (FUNCTION_TABLE_ii[$load & 8191](456962)|0); //@line 4434 "libraries/glad/glad.cpp"
  HEAP32[165359] = $2016; //@line 4434 "libraries/glad/glad.cpp"
  $2017 = (FUNCTION_TABLE_ii[$load & 8191](456983)|0); //@line 4435 "libraries/glad/glad.cpp"
  HEAP32[165109] = $2017; //@line 4435 "libraries/glad/glad.cpp"
  $2018 = (FUNCTION_TABLE_ii[$load & 8191](457014)|0); //@line 4436 "libraries/glad/glad.cpp"
  HEAP32[165419] = $2018; //@line 4436 "libraries/glad/glad.cpp"
  $2019 = (FUNCTION_TABLE_ii[$load & 8191](457047)|0); //@line 4437 "libraries/glad/glad.cpp"
  HEAP32[165318] = $2019; //@line 4437 "libraries/glad/glad.cpp"
  $2020 = (FUNCTION_TABLE_ii[$load & 8191](457073)|0); //@line 4438 "libraries/glad/glad.cpp"
  HEAP32[164631] = $2020; //@line 4438 "libraries/glad/glad.cpp"
  $2021 = (FUNCTION_TABLE_ii[$load & 8191](457097)|0); //@line 4439 "libraries/glad/glad.cpp"
  HEAP32[165477] = $2021; //@line 4439 "libraries/glad/glad.cpp"
  $2022 = (FUNCTION_TABLE_ii[$load & 8191](457123)|0); //@line 4440 "libraries/glad/glad.cpp"
  HEAP32[165279] = $2022; //@line 4440 "libraries/glad/glad.cpp"
 }
 $2023 = HEAP8[1252118]|0; //@line 4449 "libraries/glad/glad.cpp"
 $2024 = ($2023<<24>>24)==(0); //@line 4449 "libraries/glad/glad.cpp"
 if (!($2024)) {
  $2025 = (FUNCTION_TABLE_ii[$load & 8191](479158)|0); //@line 4450 "libraries/glad/glad.cpp"
  HEAP32[165779] = $2025; //@line 4450 "libraries/glad/glad.cpp"
  $2026 = (FUNCTION_TABLE_ii[$load & 8191](479175)|0); //@line 4451 "libraries/glad/glad.cpp"
  HEAP32[165780] = $2026; //@line 4451 "libraries/glad/glad.cpp"
  $2027 = (FUNCTION_TABLE_ii[$load & 8191](479198)|0); //@line 4452 "libraries/glad/glad.cpp"
  HEAP32[165781] = $2027; //@line 4452 "libraries/glad/glad.cpp"
  $2028 = (FUNCTION_TABLE_ii[$load & 8191](479224)|0); //@line 4453 "libraries/glad/glad.cpp"
  HEAP32[165782] = $2028; //@line 4453 "libraries/glad/glad.cpp"
  $2029 = (FUNCTION_TABLE_ii[$load & 8191](479243)|0); //@line 4454 "libraries/glad/glad.cpp"
  HEAP32[165783] = $2029; //@line 4454 "libraries/glad/glad.cpp"
  $2030 = (FUNCTION_TABLE_ii[$load & 8191](479263)|0); //@line 4455 "libraries/glad/glad.cpp"
  HEAP32[165784] = $2030; //@line 4455 "libraries/glad/glad.cpp"
 }
 $2031 = HEAP8[1252120]|0; //@line 4461 "libraries/glad/glad.cpp"
 $2032 = ($2031<<24>>24)==(0); //@line 4461 "libraries/glad/glad.cpp"
 if (!($2032)) {
  $2033 = (FUNCTION_TABLE_ii[$load & 8191](479285)|0); //@line 4462 "libraries/glad/glad.cpp"
  HEAP32[165785] = $2033; //@line 4462 "libraries/glad/glad.cpp"
  $2034 = (FUNCTION_TABLE_ii[$load & 8191](479311)|0); //@line 4463 "libraries/glad/glad.cpp"
  HEAP32[165786] = $2034; //@line 4463 "libraries/glad/glad.cpp"
  $2035 = (FUNCTION_TABLE_ii[$load & 8191](479342)|0); //@line 4464 "libraries/glad/glad.cpp"
  HEAP32[165787] = $2035; //@line 4464 "libraries/glad/glad.cpp"
 }
 $2036 = HEAP8[1252121]|0; //@line 4469 "libraries/glad/glad.cpp"
 $2037 = ($2036<<24>>24)==(0); //@line 4469 "libraries/glad/glad.cpp"
 if (!($2037)) {
  $2038 = (FUNCTION_TABLE_ii[$load & 8191](479373)|0); //@line 4470 "libraries/glad/glad.cpp"
  HEAP32[165788] = $2038; //@line 4470 "libraries/glad/glad.cpp"
 }
 $2039 = HEAP8[1252125]|0; //@line 4478 "libraries/glad/glad.cpp"
 $2040 = ($2039<<24>>24)==(0); //@line 4478 "libraries/glad/glad.cpp"
 if (!($2040)) {
  $2041 = (FUNCTION_TABLE_ii[$load & 8191](455224)|0); //@line 4479 "libraries/glad/glad.cpp"
  HEAP32[164892] = $2041; //@line 4479 "libraries/glad/glad.cpp"
  $2042 = (FUNCTION_TABLE_ii[$load & 8191](455236)|0); //@line 4480 "libraries/glad/glad.cpp"
  HEAP32[164933] = $2042; //@line 4480 "libraries/glad/glad.cpp"
  $2043 = (FUNCTION_TABLE_ii[$load & 8191](455245)|0); //@line 4481 "libraries/glad/glad.cpp"
  HEAP32[165033] = $2043; //@line 4481 "libraries/glad/glad.cpp"
  $2044 = (FUNCTION_TABLE_ii[$load & 8191](455258)|0); //@line 4482 "libraries/glad/glad.cpp"
  HEAP32[165462] = $2044; //@line 4482 "libraries/glad/glad.cpp"
  $2045 = (FUNCTION_TABLE_ii[$load & 8191](455275)|0); //@line 4483 "libraries/glad/glad.cpp"
  HEAP32[165171] = $2045; //@line 4483 "libraries/glad/glad.cpp"
  $2046 = (FUNCTION_TABLE_ii[$load & 8191](455286)|0); //@line 4484 "libraries/glad/glad.cpp"
  HEAP32[164829] = $2046; //@line 4484 "libraries/glad/glad.cpp"
  $2047 = (FUNCTION_TABLE_ii[$load & 8191](455302)|0); //@line 4485 "libraries/glad/glad.cpp"
  HEAP32[165547] = $2047; //@line 4485 "libraries/glad/glad.cpp"
 }
 $2048 = HEAP8[1252126]|0; //@line 4490 "libraries/glad/glad.cpp"
 $2049 = ($2048<<24>>24)==(0); //@line 4490 "libraries/glad/glad.cpp"
 if (!($2049)) {
  $2050 = (FUNCTION_TABLE_ii[$load & 8191](457143)|0); //@line 4491 "libraries/glad/glad.cpp"
  HEAP32[165551] = $2050; //@line 4491 "libraries/glad/glad.cpp"
  $2051 = (FUNCTION_TABLE_ii[$load & 8191](457161)|0); //@line 4492 "libraries/glad/glad.cpp"
  HEAP32[165002] = $2051; //@line 4492 "libraries/glad/glad.cpp"
 }
 $2052 = HEAP8[1252127]|0; //@line 4497 "libraries/glad/glad.cpp"
 $2053 = ($2052<<24>>24)==(0); //@line 4497 "libraries/glad/glad.cpp"
 if (!($2053)) {
  $2054 = (FUNCTION_TABLE_ii[$load & 8191](463552)|0); //@line 4498 "libraries/glad/glad.cpp"
  HEAP32[165595] = $2054; //@line 4498 "libraries/glad/glad.cpp"
 }
 $2055 = HEAP8[1252128]|0; //@line 4503 "libraries/glad/glad.cpp"
 $2056 = ($2055<<24>>24)==(0); //@line 4503 "libraries/glad/glad.cpp"
 if (!($2056)) {
  $2057 = (FUNCTION_TABLE_ii[$load & 8191](479396)|0); //@line 4504 "libraries/glad/glad.cpp"
  HEAP32[165789] = $2057; //@line 4504 "libraries/glad/glad.cpp"
 }
 $2058 = HEAP8[1252130]|0; //@line 4510 "libraries/glad/glad.cpp"
 $2059 = ($2058<<24>>24)==(0); //@line 4510 "libraries/glad/glad.cpp"
 if (!($2059)) {
  $2060 = (FUNCTION_TABLE_ii[$load & 8191](460105)|0); //@line 4511 "libraries/glad/glad.cpp"
  HEAP32[165654] = $2060; //@line 4511 "libraries/glad/glad.cpp"
 }
 $2061 = HEAP8[1252138]|0; //@line 4523 "libraries/glad/glad.cpp"
 $2062 = ($2061<<24>>24)==(0); //@line 4523 "libraries/glad/glad.cpp"
 if (!($2062)) {
  $2063 = (FUNCTION_TABLE_ii[$load & 8191](455378)|0); //@line 4524 "libraries/glad/glad.cpp"
  HEAP32[165037] = $2063; //@line 4524 "libraries/glad/glad.cpp"
  $2064 = (FUNCTION_TABLE_ii[$load & 8191](455402)|0); //@line 4525 "libraries/glad/glad.cpp"
  HEAP32[165647] = $2064; //@line 4525 "libraries/glad/glad.cpp"
  $2065 = (FUNCTION_TABLE_ii[$load & 8191](455426)|0); //@line 4526 "libraries/glad/glad.cpp"
  HEAP32[164674] = $2065; //@line 4526 "libraries/glad/glad.cpp"
  $2066 = (FUNCTION_TABLE_ii[$load & 8191](455445)|0); //@line 4527 "libraries/glad/glad.cpp"
  HEAP32[164688] = $2066; //@line 4527 "libraries/glad/glad.cpp"
 }
 $2067 = HEAP8[1252145]|0; //@line 4538 "libraries/glad/glad.cpp"
 $2068 = ($2067<<24>>24)==(0); //@line 4538 "libraries/glad/glad.cpp"
 if (!($2068)) {
  $2069 = (FUNCTION_TABLE_ii[$load & 8191](459422)|0); //@line 4539 "libraries/glad/glad.cpp"
  HEAP32[165180] = $2069; //@line 4539 "libraries/glad/glad.cpp"
  $2070 = (FUNCTION_TABLE_ii[$load & 8191](459437)|0); //@line 4540 "libraries/glad/glad.cpp"
  HEAP32[165612] = $2070; //@line 4540 "libraries/glad/glad.cpp"
  $2071 = (FUNCTION_TABLE_ii[$load & 8191](459452)|0); //@line 4541 "libraries/glad/glad.cpp"
  HEAP32[165339] = $2071; //@line 4541 "libraries/glad/glad.cpp"
 }
 $2072 = HEAP8[1252146]|0; //@line 4546 "libraries/glad/glad.cpp"
 $2073 = ($2072<<24>>24)==(0); //@line 4546 "libraries/glad/glad.cpp"
 if (!($2073)) {
  $2074 = (FUNCTION_TABLE_ii[$load & 8191](460122)|0); //@line 4547 "libraries/glad/glad.cpp"
  HEAP32[165560] = $2074; //@line 4547 "libraries/glad/glad.cpp"
  $2075 = (FUNCTION_TABLE_ii[$load & 8191](460148)|0); //@line 4548 "libraries/glad/glad.cpp"
  HEAP32[164910] = $2075; //@line 4548 "libraries/glad/glad.cpp"
 }
 $2076 = HEAP8[1252148]|0; //@line 4554 "libraries/glad/glad.cpp"
 $2077 = ($2076<<24>>24)==(0); //@line 4554 "libraries/glad/glad.cpp"
 if (!($2077)) {
  $2078 = (FUNCTION_TABLE_ii[$load & 8191](460174)|0); //@line 4555 "libraries/glad/glad.cpp"
  HEAP32[165533] = $2078; //@line 4555 "libraries/glad/glad.cpp"
 }
 $2079 = HEAP8[1252149]|0; //@line 4560 "libraries/glad/glad.cpp"
 $2080 = ($2079<<24>>24)==(0); //@line 4560 "libraries/glad/glad.cpp"
 if (!($2080)) {
  $2081 = (FUNCTION_TABLE_ii[$load & 8191](455791)|0); //@line 4561 "libraries/glad/glad.cpp"
  HEAP32[165227] = $2081; //@line 4561 "libraries/glad/glad.cpp"
  $2082 = (FUNCTION_TABLE_ii[$load & 8191](455806)|0); //@line 4562 "libraries/glad/glad.cpp"
  HEAP32[164796] = $2082; //@line 4562 "libraries/glad/glad.cpp"
  $2083 = (FUNCTION_TABLE_ii[$load & 8191](455827)|0); //@line 4563 "libraries/glad/glad.cpp"
  HEAP32[165643] = $2083; //@line 4563 "libraries/glad/glad.cpp"
 }
 $2084 = HEAP8[1252150]|0; //@line 4568 "libraries/glad/glad.cpp"
 $2085 = ($2084<<24>>24)==(0); //@line 4568 "libraries/glad/glad.cpp"
 if (!($2085)) {
  $2086 = (FUNCTION_TABLE_ii[$load & 8191](457180)|0); //@line 4569 "libraries/glad/glad.cpp"
  HEAP32[165640] = $2086; //@line 4569 "libraries/glad/glad.cpp"
  $2087 = (FUNCTION_TABLE_ii[$load & 8191](457204)|0); //@line 4570 "libraries/glad/glad.cpp"
  HEAP32[165495] = $2087; //@line 4570 "libraries/glad/glad.cpp"
  $2088 = (FUNCTION_TABLE_ii[$load & 8191](457231)|0); //@line 4571 "libraries/glad/glad.cpp"
  HEAP32[165443] = $2088; //@line 4571 "libraries/glad/glad.cpp"
  $2089 = (FUNCTION_TABLE_ii[$load & 8191](457255)|0); //@line 4572 "libraries/glad/glad.cpp"
  HEAP32[165286] = $2089; //@line 4572 "libraries/glad/glad.cpp"
  $2090 = (FUNCTION_TABLE_ii[$load & 8191](457277)|0); //@line 4573 "libraries/glad/glad.cpp"
  HEAP32[164726] = $2090; //@line 4573 "libraries/glad/glad.cpp"
  $2091 = (FUNCTION_TABLE_ii[$load & 8191](457302)|0); //@line 4574 "libraries/glad/glad.cpp"
  HEAP32[165417] = $2091; //@line 4574 "libraries/glad/glad.cpp"
  $2092 = (FUNCTION_TABLE_ii[$load & 8191](457328)|0); //@line 4575 "libraries/glad/glad.cpp"
  HEAP32[164924] = $2092; //@line 4575 "libraries/glad/glad.cpp"
 }
 $2093 = HEAP8[1252151]|0; //@line 4580 "libraries/glad/glad.cpp"
 $2094 = ($2093<<24>>24)==(0); //@line 4580 "libraries/glad/glad.cpp"
 if (!($2094)) {
  $2095 = (FUNCTION_TABLE_ii[$load & 8191](457352)|0); //@line 4581 "libraries/glad/glad.cpp"
  HEAP32[165221] = $2095; //@line 4581 "libraries/glad/glad.cpp"
  $2096 = (FUNCTION_TABLE_ii[$load & 8191](457382)|0); //@line 4582 "libraries/glad/glad.cpp"
  HEAP32[164885] = $2096; //@line 4582 "libraries/glad/glad.cpp"
  $2097 = (FUNCTION_TABLE_ii[$load & 8191](457402)|0); //@line 4583 "libraries/glad/glad.cpp"
  HEAP32[164985] = $2097; //@line 4583 "libraries/glad/glad.cpp"
  $2098 = (FUNCTION_TABLE_ii[$load & 8191](457420)|0); //@line 4584 "libraries/glad/glad.cpp"
  HEAP32[165342] = $2098; //@line 4584 "libraries/glad/glad.cpp"
 }
 $2099 = HEAP8[1252152]|0; //@line 4589 "libraries/glad/glad.cpp"
 $2100 = ($2099<<24>>24)==(0); //@line 4589 "libraries/glad/glad.cpp"
 if (!($2100)) {
  $2101 = (FUNCTION_TABLE_ii[$load & 8191](459467)|0); //@line 4590 "libraries/glad/glad.cpp"
  HEAP32[164919] = $2101; //@line 4590 "libraries/glad/glad.cpp"
  $2102 = (FUNCTION_TABLE_ii[$load & 8191](459500)|0); //@line 4591 "libraries/glad/glad.cpp"
  HEAP32[164649] = $2102; //@line 4591 "libraries/glad/glad.cpp"
 }
 $2103 = HEAP8[1252154]|0; //@line 4597 "libraries/glad/glad.cpp"
 $2104 = ($2103<<24>>24)==(0); //@line 4597 "libraries/glad/glad.cpp"
 if (!($2104)) {
  $2105 = (FUNCTION_TABLE_ii[$load & 8191](454923)|0); //@line 4598 "libraries/glad/glad.cpp"
  HEAP32[164824] = $2105; //@line 4598 "libraries/glad/glad.cpp"
  $2106 = (FUNCTION_TABLE_ii[$load & 8191](454943)|0); //@line 4599 "libraries/glad/glad.cpp"
  HEAP32[164833] = $2106; //@line 4599 "libraries/glad/glad.cpp"
  $2107 = (FUNCTION_TABLE_ii[$load & 8191](454965)|0); //@line 4600 "libraries/glad/glad.cpp"
  HEAP32[164770] = $2107; //@line 4600 "libraries/glad/glad.cpp"
  $2108 = (FUNCTION_TABLE_ii[$load & 8191](454988)|0); //@line 4601 "libraries/glad/glad.cpp"
  HEAP32[165556] = $2108; //@line 4601 "libraries/glad/glad.cpp"
  $2109 = (FUNCTION_TABLE_ii[$load & 8191](455011)|0); //@line 4602 "libraries/glad/glad.cpp"
  HEAP32[165048] = $2109; //@line 4602 "libraries/glad/glad.cpp"
  $2110 = (FUNCTION_TABLE_ii[$load & 8191](455037)|0); //@line 4603 "libraries/glad/glad.cpp"
  HEAP32[165541] = $2110; //@line 4603 "libraries/glad/glad.cpp"
  $2111 = (FUNCTION_TABLE_ii[$load & 8191](455065)|0); //@line 4604 "libraries/glad/glad.cpp"
  HEAP32[165325] = $2111; //@line 4604 "libraries/glad/glad.cpp"
  $2112 = (FUNCTION_TABLE_ii[$load & 8191](453757)|0); //@line 4605 "libraries/glad/glad.cpp"
  HEAP32[165430] = $2112; //@line 4605 "libraries/glad/glad.cpp"
  $2113 = (FUNCTION_TABLE_ii[$load & 8191](453775)|0); //@line 4606 "libraries/glad/glad.cpp"
  HEAP32[165512] = $2113; //@line 4606 "libraries/glad/glad.cpp"
  $2114 = (FUNCTION_TABLE_ii[$load & 8191](453659)|0); //@line 4607 "libraries/glad/glad.cpp"
  HEAP32[164643] = $2114; //@line 4607 "libraries/glad/glad.cpp"
 }
 $2115 = HEAP8[1252156]|0; //@line 4613 "libraries/glad/glad.cpp"
 $2116 = ($2115<<24>>24)==(0); //@line 4613 "libraries/glad/glad.cpp"
 if (!($2116)) {
  $2117 = (FUNCTION_TABLE_ii[$load & 8191](615642)|0); //@line 4614 "libraries/glad/glad.cpp"
  HEAP32[165066] = $2117; //@line 4614 "libraries/glad/glad.cpp"
  $2118 = (FUNCTION_TABLE_ii[$load & 8191](615621)|0); //@line 4615 "libraries/glad/glad.cpp"
  HEAP32[164893] = $2118; //@line 4615 "libraries/glad/glad.cpp"
  $2119 = (FUNCTION_TABLE_ii[$load & 8191](615603)|0); //@line 4616 "libraries/glad/glad.cpp"
  HEAP32[165300] = $2119; //@line 4616 "libraries/glad/glad.cpp"
  $2120 = (FUNCTION_TABLE_ii[$load & 8191](454851)|0); //@line 4617 "libraries/glad/glad.cpp"
  HEAP32[164810] = $2120; //@line 4617 "libraries/glad/glad.cpp"
 }
 $2121 = HEAP8[1252157]|0; //@line 4622 "libraries/glad/glad.cpp"
 $2122 = ($2121<<24>>24)==(0); //@line 4622 "libraries/glad/glad.cpp"
 if (!($2122)) {
  $2123 = (FUNCTION_TABLE_ii[$load & 8191](458849)|0); //@line 4623 "libraries/glad/glad.cpp"
  HEAP32[165521] = $2123; //@line 4623 "libraries/glad/glad.cpp"
  $2124 = (FUNCTION_TABLE_ii[$load & 8191](458867)|0); //@line 4624 "libraries/glad/glad.cpp"
  HEAP32[165320] = $2124; //@line 4624 "libraries/glad/glad.cpp"
  $2125 = (FUNCTION_TABLE_ii[$load & 8191](458885)|0); //@line 4625 "libraries/glad/glad.cpp"
  HEAP32[165659] = $2125; //@line 4625 "libraries/glad/glad.cpp"
  $2126 = (FUNCTION_TABLE_ii[$load & 8191](458903)|0); //@line 4626 "libraries/glad/glad.cpp"
  HEAP32[164734] = $2126; //@line 4626 "libraries/glad/glad.cpp"
  $2127 = (FUNCTION_TABLE_ii[$load & 8191](458921)|0); //@line 4627 "libraries/glad/glad.cpp"
  HEAP32[164851] = $2127; //@line 4627 "libraries/glad/glad.cpp"
  $2128 = (FUNCTION_TABLE_ii[$load & 8191](458940)|0); //@line 4628 "libraries/glad/glad.cpp"
  HEAP32[165311] = $2128; //@line 4628 "libraries/glad/glad.cpp"
  $2129 = (FUNCTION_TABLE_ii[$load & 8191](458959)|0); //@line 4629 "libraries/glad/glad.cpp"
  HEAP32[165433] = $2129; //@line 4629 "libraries/glad/glad.cpp"
  $2130 = (FUNCTION_TABLE_ii[$load & 8191](458978)|0); //@line 4630 "libraries/glad/glad.cpp"
  HEAP32[164875] = $2130; //@line 4630 "libraries/glad/glad.cpp"
  $2131 = (FUNCTION_TABLE_ii[$load & 8191](458997)|0); //@line 4631 "libraries/glad/glad.cpp"
  HEAP32[165306] = $2131; //@line 4631 "libraries/glad/glad.cpp"
  $2132 = (FUNCTION_TABLE_ii[$load & 8191](459020)|0); //@line 4632 "libraries/glad/glad.cpp"
  HEAP32[165351] = $2132; //@line 4632 "libraries/glad/glad.cpp"
 }
 $2133 = HEAP8[1252158]|0; //@line 4637 "libraries/glad/glad.cpp"
 $2134 = ($2133<<24>>24)==(0); //@line 4637 "libraries/glad/glad.cpp"
 if (!($2134)) {
  $2135 = (FUNCTION_TABLE_ii[$load & 8191](460188)|0); //@line 4638 "libraries/glad/glad.cpp"
  HEAP32[165479] = $2135; //@line 4638 "libraries/glad/glad.cpp"
  $2136 = (FUNCTION_TABLE_ii[$load & 8191](460207)|0); //@line 4639 "libraries/glad/glad.cpp"
  HEAP32[165103] = $2136; //@line 4639 "libraries/glad/glad.cpp"
  $2137 = (FUNCTION_TABLE_ii[$load & 8191](460228)|0); //@line 4640 "libraries/glad/glad.cpp"
  HEAP32[165544] = $2137; //@line 4640 "libraries/glad/glad.cpp"
  $2138 = (FUNCTION_TABLE_ii[$load & 8191](460250)|0); //@line 4641 "libraries/glad/glad.cpp"
  HEAP32[165074] = $2138; //@line 4641 "libraries/glad/glad.cpp"
  $2139 = (FUNCTION_TABLE_ii[$load & 8191](460272)|0); //@line 4642 "libraries/glad/glad.cpp"
  HEAP32[165107] = $2139; //@line 4642 "libraries/glad/glad.cpp"
  $2140 = (FUNCTION_TABLE_ii[$load & 8191](460294)|0); //@line 4643 "libraries/glad/glad.cpp"
  HEAP32[164858] = $2140; //@line 4643 "libraries/glad/glad.cpp"
 }
 $2141 = HEAP8[1252160]|0; //@line 4649 "libraries/glad/glad.cpp"
 $2142 = ($2141<<24>>24)==(0); //@line 4649 "libraries/glad/glad.cpp"
 if (!($2142)) {
  $2143 = (FUNCTION_TABLE_ii[$load & 8191](455849)|0); //@line 4650 "libraries/glad/glad.cpp"
  HEAP32[164962] = $2143; //@line 4650 "libraries/glad/glad.cpp"
  $2144 = (FUNCTION_TABLE_ii[$load & 8191](455868)|0); //@line 4651 "libraries/glad/glad.cpp"
  HEAP32[165132] = $2144; //@line 4651 "libraries/glad/glad.cpp"
  $2145 = (FUNCTION_TABLE_ii[$load & 8191](455888)|0); //@line 4652 "libraries/glad/glad.cpp"
  HEAP32[165574] = $2145; //@line 4652 "libraries/glad/glad.cpp"
  $2146 = (FUNCTION_TABLE_ii[$load & 8191](455907)|0); //@line 4653 "libraries/glad/glad.cpp"
  HEAP32[165141] = $2146; //@line 4653 "libraries/glad/glad.cpp"
  $2147 = (FUNCTION_TABLE_ii[$load & 8191](455927)|0); //@line 4654 "libraries/glad/glad.cpp"
  HEAP32[164850] = $2147; //@line 4654 "libraries/glad/glad.cpp"
  $2148 = (FUNCTION_TABLE_ii[$load & 8191](455946)|0); //@line 4655 "libraries/glad/glad.cpp"
  HEAP32[165584] = $2148; //@line 4655 "libraries/glad/glad.cpp"
  $2149 = (FUNCTION_TABLE_ii[$load & 8191](455966)|0); //@line 4656 "libraries/glad/glad.cpp"
  HEAP32[164671] = $2149; //@line 4656 "libraries/glad/glad.cpp"
  $2150 = (FUNCTION_TABLE_ii[$load & 8191](455985)|0); //@line 4657 "libraries/glad/glad.cpp"
  HEAP32[165278] = $2150; //@line 4657 "libraries/glad/glad.cpp"
  $2151 = (FUNCTION_TABLE_ii[$load & 8191](456005)|0); //@line 4658 "libraries/glad/glad.cpp"
  HEAP32[164689] = $2151; //@line 4658 "libraries/glad/glad.cpp"
  $2152 = (FUNCTION_TABLE_ii[$load & 8191](456018)|0); //@line 4659 "libraries/glad/glad.cpp"
  HEAP32[165391] = $2152; //@line 4659 "libraries/glad/glad.cpp"
  $2153 = (FUNCTION_TABLE_ii[$load & 8191](456032)|0); //@line 4660 "libraries/glad/glad.cpp"
  HEAP32[165490] = $2153; //@line 4660 "libraries/glad/glad.cpp"
  $2154 = (FUNCTION_TABLE_ii[$load & 8191](456045)|0); //@line 4661 "libraries/glad/glad.cpp"
  HEAP32[164787] = $2154; //@line 4661 "libraries/glad/glad.cpp"
  $2155 = (FUNCTION_TABLE_ii[$load & 8191](456059)|0); //@line 4662 "libraries/glad/glad.cpp"
  HEAP32[164664] = $2155; //@line 4662 "libraries/glad/glad.cpp"
  $2156 = (FUNCTION_TABLE_ii[$load & 8191](456072)|0); //@line 4663 "libraries/glad/glad.cpp"
  HEAP32[165527] = $2156; //@line 4663 "libraries/glad/glad.cpp"
  $2157 = (FUNCTION_TABLE_ii[$load & 8191](456086)|0); //@line 4664 "libraries/glad/glad.cpp"
  HEAP32[165570] = $2157; //@line 4664 "libraries/glad/glad.cpp"
  $2158 = (FUNCTION_TABLE_ii[$load & 8191](456101)|0); //@line 4665 "libraries/glad/glad.cpp"
  HEAP32[164917] = $2158; //@line 4665 "libraries/glad/glad.cpp"
  $2159 = (FUNCTION_TABLE_ii[$load & 8191](456117)|0); //@line 4666 "libraries/glad/glad.cpp"
  HEAP32[164720] = $2159; //@line 4666 "libraries/glad/glad.cpp"
  $2160 = (FUNCTION_TABLE_ii[$load & 8191](456132)|0); //@line 4667 "libraries/glad/glad.cpp"
  HEAP32[165550] = $2160; //@line 4667 "libraries/glad/glad.cpp"
  $2161 = (FUNCTION_TABLE_ii[$load & 8191](456148)|0); //@line 4668 "libraries/glad/glad.cpp"
  HEAP32[164766] = $2161; //@line 4668 "libraries/glad/glad.cpp"
  $2162 = (FUNCTION_TABLE_ii[$load & 8191](456163)|0); //@line 4669 "libraries/glad/glad.cpp"
  HEAP32[164746] = $2162; //@line 4669 "libraries/glad/glad.cpp"
  $2163 = (FUNCTION_TABLE_ii[$load & 8191](456179)|0); //@line 4670 "libraries/glad/glad.cpp"
  HEAP32[164940] = $2163; //@line 4670 "libraries/glad/glad.cpp"
  $2164 = (FUNCTION_TABLE_ii[$load & 8191](456194)|0); //@line 4671 "libraries/glad/glad.cpp"
  HEAP32[165384] = $2164; //@line 4671 "libraries/glad/glad.cpp"
  $2165 = (FUNCTION_TABLE_ii[$load & 8191](456210)|0); //@line 4672 "libraries/glad/glad.cpp"
  HEAP32[164854] = $2165; //@line 4672 "libraries/glad/glad.cpp"
  $2166 = (FUNCTION_TABLE_ii[$load & 8191](456230)|0); //@line 4673 "libraries/glad/glad.cpp"
  HEAP32[165016] = $2166; //@line 4673 "libraries/glad/glad.cpp"
  $2167 = (FUNCTION_TABLE_ii[$load & 8191](456251)|0); //@line 4674 "libraries/glad/glad.cpp"
  HEAP32[164727] = $2167; //@line 4674 "libraries/glad/glad.cpp"
  $2168 = (FUNCTION_TABLE_ii[$load & 8191](456271)|0); //@line 4675 "libraries/glad/glad.cpp"
  HEAP32[165121] = $2168; //@line 4675 "libraries/glad/glad.cpp"
  $2169 = (FUNCTION_TABLE_ii[$load & 8191](456292)|0); //@line 4676 "libraries/glad/glad.cpp"
  HEAP32[164660] = $2169; //@line 4676 "libraries/glad/glad.cpp"
  $2170 = (FUNCTION_TABLE_ii[$load & 8191](456312)|0); //@line 4677 "libraries/glad/glad.cpp"
  HEAP32[164849] = $2170; //@line 4677 "libraries/glad/glad.cpp"
  $2171 = (FUNCTION_TABLE_ii[$load & 8191](456333)|0); //@line 4678 "libraries/glad/glad.cpp"
  HEAP32[164751] = $2171; //@line 4678 "libraries/glad/glad.cpp"
  $2172 = (FUNCTION_TABLE_ii[$load & 8191](456353)|0); //@line 4679 "libraries/glad/glad.cpp"
  HEAP32[164905] = $2172; //@line 4679 "libraries/glad/glad.cpp"
  $2173 = (FUNCTION_TABLE_ii[$load & 8191](456374)|0); //@line 4680 "libraries/glad/glad.cpp"
  HEAP32[165243] = $2173; //@line 4680 "libraries/glad/glad.cpp"
  $2174 = (FUNCTION_TABLE_ii[$load & 8191](456387)|0); //@line 4681 "libraries/glad/glad.cpp"
  HEAP32[165162] = $2174; //@line 4681 "libraries/glad/glad.cpp"
  $2175 = (FUNCTION_TABLE_ii[$load & 8191](456401)|0); //@line 4682 "libraries/glad/glad.cpp"
  HEAP32[164894] = $2175; //@line 4682 "libraries/glad/glad.cpp"
  $2176 = (FUNCTION_TABLE_ii[$load & 8191](456413)|0); //@line 4683 "libraries/glad/glad.cpp"
  HEAP32[164642] = $2176; //@line 4683 "libraries/glad/glad.cpp"
  $2177 = (FUNCTION_TABLE_ii[$load & 8191](456426)|0); //@line 4684 "libraries/glad/glad.cpp"
  HEAP32[165328] = $2177; //@line 4684 "libraries/glad/glad.cpp"
  $2178 = (FUNCTION_TABLE_ii[$load & 8191](456438)|0); //@line 4685 "libraries/glad/glad.cpp"
  HEAP32[165403] = $2178; //@line 4685 "libraries/glad/glad.cpp"
  $2179 = (FUNCTION_TABLE_ii[$load & 8191](456451)|0); //@line 4686 "libraries/glad/glad.cpp"
  HEAP32[164972] = $2179; //@line 4686 "libraries/glad/glad.cpp"
  $2180 = (FUNCTION_TABLE_ii[$load & 8191](456472)|0); //@line 4687 "libraries/glad/glad.cpp"
  HEAP32[165586] = $2180; //@line 4687 "libraries/glad/glad.cpp"
 }
 $2181 = HEAP8[1252161]|0; //@line 4692 "libraries/glad/glad.cpp"
 $2182 = ($2181<<24>>24)==(0); //@line 4692 "libraries/glad/glad.cpp"
 if (!($2182)) {
  $2183 = (FUNCTION_TABLE_ii[$load & 8191](459041)|0); //@line 4693 "libraries/glad/glad.cpp"
  HEAP32[165246] = $2183; //@line 4693 "libraries/glad/glad.cpp"
  $2184 = (FUNCTION_TABLE_ii[$load & 8191](459058)|0); //@line 4694 "libraries/glad/glad.cpp"
  HEAP32[165118] = $2184; //@line 4694 "libraries/glad/glad.cpp"
  $2185 = (FUNCTION_TABLE_ii[$load & 8191](459077)|0); //@line 4695 "libraries/glad/glad.cpp"
  HEAP32[165165] = $2185; //@line 4695 "libraries/glad/glad.cpp"
  $2186 = (FUNCTION_TABLE_ii[$load & 8191](459097)|0); //@line 4696 "libraries/glad/glad.cpp"
  HEAP32[164999] = $2186; //@line 4696 "libraries/glad/glad.cpp"
  $2187 = (FUNCTION_TABLE_ii[$load & 8191](459113)|0); //@line 4697 "libraries/glad/glad.cpp"
  HEAP32[165526] = $2187; //@line 4697 "libraries/glad/glad.cpp"
  $2188 = (FUNCTION_TABLE_ii[$load & 8191](459130)|0); //@line 4698 "libraries/glad/glad.cpp"
  HEAP32[165361] = $2188; //@line 4698 "libraries/glad/glad.cpp"
  $2189 = (FUNCTION_TABLE_ii[$load & 8191](459148)|0); //@line 4699 "libraries/glad/glad.cpp"
  HEAP32[165152] = $2189; //@line 4699 "libraries/glad/glad.cpp"
  $2190 = (FUNCTION_TABLE_ii[$load & 8191](459167)|0); //@line 4700 "libraries/glad/glad.cpp"
  HEAP32[165083] = $2190; //@line 4700 "libraries/glad/glad.cpp"
  $2191 = (FUNCTION_TABLE_ii[$load & 8191](459187)|0); //@line 4701 "libraries/glad/glad.cpp"
  HEAP32[165464] = $2191; //@line 4701 "libraries/glad/glad.cpp"
  $2192 = (FUNCTION_TABLE_ii[$load & 8191](459201)|0); //@line 4702 "libraries/glad/glad.cpp"
  HEAP32[164636] = $2192; //@line 4702 "libraries/glad/glad.cpp"
 }
 $2193 = HEAP8[1252162]|0; //@line 4707 "libraries/glad/glad.cpp"
 $2194 = ($2193<<24>>24)==(0); //@line 4707 "libraries/glad/glad.cpp"
 if (!($2194)) {
  $2195 = (FUNCTION_TABLE_ii[$load & 8191](479411)|0); //@line 4708 "libraries/glad/glad.cpp"
  HEAP32[165790] = $2195; //@line 4708 "libraries/glad/glad.cpp"
  $2196 = (FUNCTION_TABLE_ii[$load & 8191](479448)|0); //@line 4709 "libraries/glad/glad.cpp"
  HEAP32[165791] = $2196; //@line 4709 "libraries/glad/glad.cpp"
  $2197 = (FUNCTION_TABLE_ii[$load & 8191](479487)|0); //@line 4710 "libraries/glad/glad.cpp"
  HEAP32[165792] = $2197; //@line 4710 "libraries/glad/glad.cpp"
 }
 $2198 = HEAP8[1252164]|0; //@line 4716 "libraries/glad/glad.cpp"
 $2199 = ($2198<<24>>24)==(0); //@line 4716 "libraries/glad/glad.cpp"
 if (!($2199)) {
  $2200 = (FUNCTION_TABLE_ii[$load & 8191](479536)|0); //@line 4717 "libraries/glad/glad.cpp"
  HEAP32[165793] = $2200; //@line 4717 "libraries/glad/glad.cpp"
  $2201 = (FUNCTION_TABLE_ii[$load & 8191](479555)|0); //@line 4718 "libraries/glad/glad.cpp"
  HEAP32[165794] = $2201; //@line 4718 "libraries/glad/glad.cpp"
  $2202 = (FUNCTION_TABLE_ii[$load & 8191](479581)|0); //@line 4719 "libraries/glad/glad.cpp"
  HEAP32[165795] = $2202; //@line 4719 "libraries/glad/glad.cpp"
 }
 $2203 = HEAP8[1252165]|0; //@line 4724 "libraries/glad/glad.cpp"
 $2204 = ($2203<<24>>24)==(0); //@line 4724 "libraries/glad/glad.cpp"
 if (!($2204)) {
  $2205 = (FUNCTION_TABLE_ii[$load & 8191](479603)|0); //@line 4725 "libraries/glad/glad.cpp"
  HEAP32[165796] = $2205; //@line 4725 "libraries/glad/glad.cpp"
  $2206 = (FUNCTION_TABLE_ii[$load & 8191](479636)|0); //@line 4726 "libraries/glad/glad.cpp"
  HEAP32[165797] = $2206; //@line 4726 "libraries/glad/glad.cpp"
  $2207 = (FUNCTION_TABLE_ii[$load & 8191](479662)|0); //@line 4727 "libraries/glad/glad.cpp"
  HEAP32[165798] = $2207; //@line 4727 "libraries/glad/glad.cpp"
  $2208 = (FUNCTION_TABLE_ii[$load & 8191](479699)|0); //@line 4728 "libraries/glad/glad.cpp"
  HEAP32[165799] = $2208; //@line 4728 "libraries/glad/glad.cpp"
 }
 $2209 = HEAP8[1252166]|0; //@line 4733 "libraries/glad/glad.cpp"
 $2210 = ($2209<<24>>24)==(0); //@line 4733 "libraries/glad/glad.cpp"
 if (!($2210)) {
  $2211 = (FUNCTION_TABLE_ii[$load & 8191](479721)|0); //@line 4734 "libraries/glad/glad.cpp"
  HEAP32[165800] = $2211; //@line 4734 "libraries/glad/glad.cpp"
 }
 $2212 = HEAP8[1252167]|0; //@line 4739 "libraries/glad/glad.cpp"
 $2213 = ($2212<<24>>24)==(0); //@line 4739 "libraries/glad/glad.cpp"
 if (!($2213)) {
  $2214 = (FUNCTION_TABLE_ii[$load & 8191](479740)|0); //@line 4740 "libraries/glad/glad.cpp"
  HEAP32[165801] = $2214; //@line 4740 "libraries/glad/glad.cpp"
 }
 $2215 = HEAP8[1252170]|0; //@line 4747 "libraries/glad/glad.cpp"
 $2216 = ($2215<<24>>24)==(0); //@line 4747 "libraries/glad/glad.cpp"
 if (!($2216)) {
  $2217 = (FUNCTION_TABLE_ii[$load & 8191](479759)|0); //@line 4748 "libraries/glad/glad.cpp"
  HEAP32[165802] = $2217; //@line 4748 "libraries/glad/glad.cpp"
 }
 $2218 = HEAP8[1252171]|0; //@line 4753 "libraries/glad/glad.cpp"
 $2219 = ($2218<<24>>24)==(0); //@line 4753 "libraries/glad/glad.cpp"
 if (!($2219)) {
  $2220 = (FUNCTION_TABLE_ii[$load & 8191](479781)|0); //@line 4754 "libraries/glad/glad.cpp"
  HEAP32[165803] = $2220; //@line 4754 "libraries/glad/glad.cpp"
  $2221 = (FUNCTION_TABLE_ii[$load & 8191](479798)|0); //@line 4755 "libraries/glad/glad.cpp"
  HEAP32[165804] = $2221; //@line 4755 "libraries/glad/glad.cpp"
 }
 $2222 = HEAP8[1252172]|0; //@line 4760 "libraries/glad/glad.cpp"
 $2223 = ($2222<<24>>24)==(0); //@line 4760 "libraries/glad/glad.cpp"
 if (!($2223)) {
  $2224 = (FUNCTION_TABLE_ii[$load & 8191](479818)|0); //@line 4761 "libraries/glad/glad.cpp"
  HEAP32[165805] = $2224; //@line 4761 "libraries/glad/glad.cpp"
  $2225 = (FUNCTION_TABLE_ii[$load & 8191](479841)|0); //@line 4762 "libraries/glad/glad.cpp"
  HEAP32[165806] = $2225; //@line 4762 "libraries/glad/glad.cpp"
  $2226 = (FUNCTION_TABLE_ii[$load & 8191](479862)|0); //@line 4763 "libraries/glad/glad.cpp"
  HEAP32[165807] = $2226; //@line 4763 "libraries/glad/glad.cpp"
 }
 $2227 = HEAP8[1252173]|0; //@line 4768 "libraries/glad/glad.cpp"
 $2228 = ($2227<<24>>24)==(0); //@line 4768 "libraries/glad/glad.cpp"
 if (!($2228)) {
  $2229 = (FUNCTION_TABLE_ii[$load & 8191](479882)|0); //@line 4769 "libraries/glad/glad.cpp"
  HEAP32[165808] = $2229; //@line 4769 "libraries/glad/glad.cpp"
  $2230 = (FUNCTION_TABLE_ii[$load & 8191](479899)|0); //@line 4770 "libraries/glad/glad.cpp"
  HEAP32[165809] = $2230; //@line 4770 "libraries/glad/glad.cpp"
  $2231 = (FUNCTION_TABLE_ii[$load & 8191](479916)|0); //@line 4771 "libraries/glad/glad.cpp"
  HEAP32[165810] = $2231; //@line 4771 "libraries/glad/glad.cpp"
  $2232 = (FUNCTION_TABLE_ii[$load & 8191](479933)|0); //@line 4772 "libraries/glad/glad.cpp"
  HEAP32[165811] = $2232; //@line 4772 "libraries/glad/glad.cpp"
  $2233 = (FUNCTION_TABLE_ii[$load & 8191](479950)|0); //@line 4773 "libraries/glad/glad.cpp"
  HEAP32[165812] = $2233; //@line 4773 "libraries/glad/glad.cpp"
  $2234 = (FUNCTION_TABLE_ii[$load & 8191](479974)|0); //@line 4774 "libraries/glad/glad.cpp"
  HEAP32[165813] = $2234; //@line 4774 "libraries/glad/glad.cpp"
  $2235 = (FUNCTION_TABLE_ii[$load & 8191](479993)|0); //@line 4775 "libraries/glad/glad.cpp"
  HEAP32[165814] = $2235; //@line 4775 "libraries/glad/glad.cpp"
  $2236 = (FUNCTION_TABLE_ii[$load & 8191](480012)|0); //@line 4776 "libraries/glad/glad.cpp"
  HEAP32[165815] = $2236; //@line 4776 "libraries/glad/glad.cpp"
  $2237 = (FUNCTION_TABLE_ii[$load & 8191](480030)|0); //@line 4777 "libraries/glad/glad.cpp"
  HEAP32[165816] = $2237; //@line 4777 "libraries/glad/glad.cpp"
  $2238 = (FUNCTION_TABLE_ii[$load & 8191](480048)|0); //@line 4778 "libraries/glad/glad.cpp"
  HEAP32[165817] = $2238; //@line 4778 "libraries/glad/glad.cpp"
  $2239 = (FUNCTION_TABLE_ii[$load & 8191](480070)|0); //@line 4779 "libraries/glad/glad.cpp"
  HEAP32[165818] = $2239; //@line 4779 "libraries/glad/glad.cpp"
  $2240 = (FUNCTION_TABLE_ii[$load & 8191](480092)|0); //@line 4780 "libraries/glad/glad.cpp"
  HEAP32[165819] = $2240; //@line 4780 "libraries/glad/glad.cpp"
  $2241 = (FUNCTION_TABLE_ii[$load & 8191](480111)|0); //@line 4781 "libraries/glad/glad.cpp"
  HEAP32[165820] = $2241; //@line 4781 "libraries/glad/glad.cpp"
  $2242 = (FUNCTION_TABLE_ii[$load & 8191](480128)|0); //@line 4782 "libraries/glad/glad.cpp"
  HEAP32[165821] = $2242; //@line 4782 "libraries/glad/glad.cpp"
  $2243 = (FUNCTION_TABLE_ii[$load & 8191](480143)|0); //@line 4783 "libraries/glad/glad.cpp"
  HEAP32[165822] = $2243; //@line 4783 "libraries/glad/glad.cpp"
  $2244 = (FUNCTION_TABLE_ii[$load & 8191](480159)|0); //@line 4784 "libraries/glad/glad.cpp"
  HEAP32[165823] = $2244; //@line 4784 "libraries/glad/glad.cpp"
  $2245 = (FUNCTION_TABLE_ii[$load & 8191](480184)|0); //@line 4785 "libraries/glad/glad.cpp"
  HEAP32[165824] = $2245; //@line 4785 "libraries/glad/glad.cpp"
  $2246 = (FUNCTION_TABLE_ii[$load & 8191](480213)|0); //@line 4786 "libraries/glad/glad.cpp"
  HEAP32[165825] = $2246; //@line 4786 "libraries/glad/glad.cpp"
  $2247 = (FUNCTION_TABLE_ii[$load & 8191](480236)|0); //@line 4787 "libraries/glad/glad.cpp"
  HEAP32[165826] = $2247; //@line 4787 "libraries/glad/glad.cpp"
  $2248 = (FUNCTION_TABLE_ii[$load & 8191](480260)|0); //@line 4788 "libraries/glad/glad.cpp"
  HEAP32[165827] = $2248; //@line 4788 "libraries/glad/glad.cpp"
  $2249 = (FUNCTION_TABLE_ii[$load & 8191](480283)|0); //@line 4789 "libraries/glad/glad.cpp"
  HEAP32[165828] = $2249; //@line 4789 "libraries/glad/glad.cpp"
  $2250 = (FUNCTION_TABLE_ii[$load & 8191](480307)|0); //@line 4790 "libraries/glad/glad.cpp"
  HEAP32[165829] = $2250; //@line 4790 "libraries/glad/glad.cpp"
  $2251 = (FUNCTION_TABLE_ii[$load & 8191](480327)|0); //@line 4791 "libraries/glad/glad.cpp"
  HEAP32[165830] = $2251; //@line 4791 "libraries/glad/glad.cpp"
  $2252 = (FUNCTION_TABLE_ii[$load & 8191](480347)|0); //@line 4792 "libraries/glad/glad.cpp"
  HEAP32[165831] = $2252; //@line 4792 "libraries/glad/glad.cpp"
  $2253 = (FUNCTION_TABLE_ii[$load & 8191](480370)|0); //@line 4793 "libraries/glad/glad.cpp"
  HEAP32[165832] = $2253; //@line 4793 "libraries/glad/glad.cpp"
  $2254 = (FUNCTION_TABLE_ii[$load & 8191](480393)|0); //@line 4794 "libraries/glad/glad.cpp"
  HEAP32[165833] = $2254; //@line 4794 "libraries/glad/glad.cpp"
  $2255 = (FUNCTION_TABLE_ii[$load & 8191](480417)|0); //@line 4795 "libraries/glad/glad.cpp"
  HEAP32[165834] = $2255; //@line 4795 "libraries/glad/glad.cpp"
  $2256 = (FUNCTION_TABLE_ii[$load & 8191](480441)|0); //@line 4796 "libraries/glad/glad.cpp"
  HEAP32[165835] = $2256; //@line 4796 "libraries/glad/glad.cpp"
  $2257 = (FUNCTION_TABLE_ii[$load & 8191](480468)|0); //@line 4797 "libraries/glad/glad.cpp"
  HEAP32[165836] = $2257; //@line 4797 "libraries/glad/glad.cpp"
  $2258 = (FUNCTION_TABLE_ii[$load & 8191](480495)|0); //@line 4798 "libraries/glad/glad.cpp"
  HEAP32[165837] = $2258; //@line 4798 "libraries/glad/glad.cpp"
  $2259 = (FUNCTION_TABLE_ii[$load & 8191](480516)|0); //@line 4799 "libraries/glad/glad.cpp"
  HEAP32[165838] = $2259; //@line 4799 "libraries/glad/glad.cpp"
  $2260 = (FUNCTION_TABLE_ii[$load & 8191](480543)|0); //@line 4800 "libraries/glad/glad.cpp"
  HEAP32[165839] = $2260; //@line 4800 "libraries/glad/glad.cpp"
  $2261 = (FUNCTION_TABLE_ii[$load & 8191](480570)|0); //@line 4801 "libraries/glad/glad.cpp"
  HEAP32[165840] = $2261; //@line 4801 "libraries/glad/glad.cpp"
  $2262 = (FUNCTION_TABLE_ii[$load & 8191](480602)|0); //@line 4802 "libraries/glad/glad.cpp"
  HEAP32[165841] = $2262; //@line 4802 "libraries/glad/glad.cpp"
  $2263 = (FUNCTION_TABLE_ii[$load & 8191](480634)|0); //@line 4803 "libraries/glad/glad.cpp"
  HEAP32[165842] = $2263; //@line 4803 "libraries/glad/glad.cpp"
  $2264 = (FUNCTION_TABLE_ii[$load & 8191](480654)|0); //@line 4804 "libraries/glad/glad.cpp"
  HEAP32[165843] = $2264; //@line 4804 "libraries/glad/glad.cpp"
  $2265 = (FUNCTION_TABLE_ii[$load & 8191](480677)|0); //@line 4805 "libraries/glad/glad.cpp"
  HEAP32[165844] = $2265; //@line 4805 "libraries/glad/glad.cpp"
  $2266 = (FUNCTION_TABLE_ii[$load & 8191](480704)|0); //@line 4806 "libraries/glad/glad.cpp"
  HEAP32[165845] = $2266; //@line 4806 "libraries/glad/glad.cpp"
  $2267 = (FUNCTION_TABLE_ii[$load & 8191](480726)|0); //@line 4807 "libraries/glad/glad.cpp"
  HEAP32[165846] = $2267; //@line 4807 "libraries/glad/glad.cpp"
  $2268 = (FUNCTION_TABLE_ii[$load & 8191](480752)|0); //@line 4808 "libraries/glad/glad.cpp"
  HEAP32[165847] = $2268; //@line 4808 "libraries/glad/glad.cpp"
  $2269 = (FUNCTION_TABLE_ii[$load & 8191](480770)|0); //@line 4809 "libraries/glad/glad.cpp"
  HEAP32[165848] = $2269; //@line 4809 "libraries/glad/glad.cpp"
  $2270 = (FUNCTION_TABLE_ii[$load & 8191](480789)|0); //@line 4810 "libraries/glad/glad.cpp"
  HEAP32[165849] = $2270; //@line 4810 "libraries/glad/glad.cpp"
  $2271 = (FUNCTION_TABLE_ii[$load & 8191](480807)|0); //@line 4811 "libraries/glad/glad.cpp"
  HEAP32[165850] = $2271; //@line 4811 "libraries/glad/glad.cpp"
  $2272 = (FUNCTION_TABLE_ii[$load & 8191](480826)|0); //@line 4812 "libraries/glad/glad.cpp"
  HEAP32[165851] = $2272; //@line 4812 "libraries/glad/glad.cpp"
  $2273 = (FUNCTION_TABLE_ii[$load & 8191](480844)|0); //@line 4813 "libraries/glad/glad.cpp"
  HEAP32[165852] = $2273; //@line 4813 "libraries/glad/glad.cpp"
  $2274 = (FUNCTION_TABLE_ii[$load & 8191](480863)|0); //@line 4814 "libraries/glad/glad.cpp"
  HEAP32[165853] = $2274; //@line 4814 "libraries/glad/glad.cpp"
  $2275 = (FUNCTION_TABLE_ii[$load & 8191](480881)|0); //@line 4815 "libraries/glad/glad.cpp"
  HEAP32[165854] = $2275; //@line 4815 "libraries/glad/glad.cpp"
  $2276 = (FUNCTION_TABLE_ii[$load & 8191](480900)|0); //@line 4816 "libraries/glad/glad.cpp"
  HEAP32[165855] = $2276; //@line 4816 "libraries/glad/glad.cpp"
  $2277 = (FUNCTION_TABLE_ii[$load & 8191](480918)|0); //@line 4817 "libraries/glad/glad.cpp"
  HEAP32[165856] = $2277; //@line 4817 "libraries/glad/glad.cpp"
  $2278 = (FUNCTION_TABLE_ii[$load & 8191](480937)|0); //@line 4818 "libraries/glad/glad.cpp"
  HEAP32[165857] = $2278; //@line 4818 "libraries/glad/glad.cpp"
  $2279 = (FUNCTION_TABLE_ii[$load & 8191](480959)|0); //@line 4819 "libraries/glad/glad.cpp"
  HEAP32[165858] = $2279; //@line 4819 "libraries/glad/glad.cpp"
  $2280 = (FUNCTION_TABLE_ii[$load & 8191](480981)|0); //@line 4820 "libraries/glad/glad.cpp"
  HEAP32[165859] = $2280; //@line 4820 "libraries/glad/glad.cpp"
  $2281 = (FUNCTION_TABLE_ii[$load & 8191](481003)|0); //@line 4821 "libraries/glad/glad.cpp"
  HEAP32[165860] = $2281; //@line 4821 "libraries/glad/glad.cpp"
  $2282 = (FUNCTION_TABLE_ii[$load & 8191](481025)|0); //@line 4822 "libraries/glad/glad.cpp"
  HEAP32[165861] = $2282; //@line 4822 "libraries/glad/glad.cpp"
  $2283 = (FUNCTION_TABLE_ii[$load & 8191](481047)|0); //@line 4823 "libraries/glad/glad.cpp"
  HEAP32[165862] = $2283; //@line 4823 "libraries/glad/glad.cpp"
  $2284 = (FUNCTION_TABLE_ii[$load & 8191](481071)|0); //@line 4824 "libraries/glad/glad.cpp"
  HEAP32[165863] = $2284; //@line 4824 "libraries/glad/glad.cpp"
  $2285 = (FUNCTION_TABLE_ii[$load & 8191](481096)|0); //@line 4825 "libraries/glad/glad.cpp"
  HEAP32[165864] = $2285; //@line 4825 "libraries/glad/glad.cpp"
  $2286 = (FUNCTION_TABLE_ii[$load & 8191](481120)|0); //@line 4826 "libraries/glad/glad.cpp"
  HEAP32[165865] = $2286; //@line 4826 "libraries/glad/glad.cpp"
  $2287 = (FUNCTION_TABLE_ii[$load & 8191](481145)|0); //@line 4827 "libraries/glad/glad.cpp"
  HEAP32[165866] = $2287; //@line 4827 "libraries/glad/glad.cpp"
  $2288 = (FUNCTION_TABLE_ii[$load & 8191](481166)|0); //@line 4828 "libraries/glad/glad.cpp"
  HEAP32[165867] = $2288; //@line 4828 "libraries/glad/glad.cpp"
  $2289 = (FUNCTION_TABLE_ii[$load & 8191](481187)|0); //@line 4829 "libraries/glad/glad.cpp"
  HEAP32[165868] = $2289; //@line 4829 "libraries/glad/glad.cpp"
  $2290 = (FUNCTION_TABLE_ii[$load & 8191](481211)|0); //@line 4830 "libraries/glad/glad.cpp"
  HEAP32[165869] = $2290; //@line 4830 "libraries/glad/glad.cpp"
  $2291 = (FUNCTION_TABLE_ii[$load & 8191](481235)|0); //@line 4831 "libraries/glad/glad.cpp"
  HEAP32[165870] = $2291; //@line 4831 "libraries/glad/glad.cpp"
  $2292 = (FUNCTION_TABLE_ii[$load & 8191](481260)|0); //@line 4832 "libraries/glad/glad.cpp"
  HEAP32[165871] = $2292; //@line 4832 "libraries/glad/glad.cpp"
  $2293 = (FUNCTION_TABLE_ii[$load & 8191](481285)|0); //@line 4833 "libraries/glad/glad.cpp"
  HEAP32[165872] = $2293; //@line 4833 "libraries/glad/glad.cpp"
  $2294 = (FUNCTION_TABLE_ii[$load & 8191](481313)|0); //@line 4834 "libraries/glad/glad.cpp"
  HEAP32[165873] = $2294; //@line 4834 "libraries/glad/glad.cpp"
  $2295 = (FUNCTION_TABLE_ii[$load & 8191](481341)|0); //@line 4835 "libraries/glad/glad.cpp"
  HEAP32[165874] = $2295; //@line 4835 "libraries/glad/glad.cpp"
  $2296 = (FUNCTION_TABLE_ii[$load & 8191](481363)|0); //@line 4836 "libraries/glad/glad.cpp"
  HEAP32[165875] = $2296; //@line 4836 "libraries/glad/glad.cpp"
  $2297 = (FUNCTION_TABLE_ii[$load & 8191](481391)|0); //@line 4837 "libraries/glad/glad.cpp"
  HEAP32[165876] = $2297; //@line 4837 "libraries/glad/glad.cpp"
  $2298 = (FUNCTION_TABLE_ii[$load & 8191](481419)|0); //@line 4838 "libraries/glad/glad.cpp"
  HEAP32[165877] = $2298; //@line 4838 "libraries/glad/glad.cpp"
  $2299 = (FUNCTION_TABLE_ii[$load & 8191](481452)|0); //@line 4839 "libraries/glad/glad.cpp"
  HEAP32[165878] = $2299; //@line 4839 "libraries/glad/glad.cpp"
  $2300 = (FUNCTION_TABLE_ii[$load & 8191](481485)|0); //@line 4840 "libraries/glad/glad.cpp"
  HEAP32[165879] = $2300; //@line 4840 "libraries/glad/glad.cpp"
  $2301 = (FUNCTION_TABLE_ii[$load & 8191](481506)|0); //@line 4841 "libraries/glad/glad.cpp"
  HEAP32[165880] = $2301; //@line 4841 "libraries/glad/glad.cpp"
  $2302 = (FUNCTION_TABLE_ii[$load & 8191](481530)|0); //@line 4842 "libraries/glad/glad.cpp"
  HEAP32[165881] = $2302; //@line 4842 "libraries/glad/glad.cpp"
  $2303 = (FUNCTION_TABLE_ii[$load & 8191](481558)|0); //@line 4843 "libraries/glad/glad.cpp"
  HEAP32[165882] = $2303; //@line 4843 "libraries/glad/glad.cpp"
  $2304 = (FUNCTION_TABLE_ii[$load & 8191](481588)|0); //@line 4844 "libraries/glad/glad.cpp"
  HEAP32[165883] = $2304; //@line 4844 "libraries/glad/glad.cpp"
  $2305 = (FUNCTION_TABLE_ii[$load & 8191](481619)|0); //@line 4845 "libraries/glad/glad.cpp"
  HEAP32[165884] = $2305; //@line 4845 "libraries/glad/glad.cpp"
  $2306 = (FUNCTION_TABLE_ii[$load & 8191](481641)|0); //@line 4846 "libraries/glad/glad.cpp"
  HEAP32[165885] = $2306; //@line 4846 "libraries/glad/glad.cpp"
  $2307 = (FUNCTION_TABLE_ii[$load & 8191](481664)|0); //@line 4847 "libraries/glad/glad.cpp"
  HEAP32[165886] = $2307; //@line 4847 "libraries/glad/glad.cpp"
  $2308 = (FUNCTION_TABLE_ii[$load & 8191](481688)|0); //@line 4848 "libraries/glad/glad.cpp"
  HEAP32[165887] = $2308; //@line 4848 "libraries/glad/glad.cpp"
  $2309 = (FUNCTION_TABLE_ii[$load & 8191](481707)|0); //@line 4849 "libraries/glad/glad.cpp"
  HEAP32[165888] = $2309; //@line 4849 "libraries/glad/glad.cpp"
  $2310 = (FUNCTION_TABLE_ii[$load & 8191](481727)|0); //@line 4850 "libraries/glad/glad.cpp"
  HEAP32[165889] = $2310; //@line 4850 "libraries/glad/glad.cpp"
  $2311 = (FUNCTION_TABLE_ii[$load & 8191](481749)|0); //@line 4851 "libraries/glad/glad.cpp"
  HEAP32[165890] = $2311; //@line 4851 "libraries/glad/glad.cpp"
  $2312 = (FUNCTION_TABLE_ii[$load & 8191](481773)|0); //@line 4852 "libraries/glad/glad.cpp"
  HEAP32[165891] = $2312; //@line 4852 "libraries/glad/glad.cpp"
  $2313 = (FUNCTION_TABLE_ii[$load & 8191](481797)|0); //@line 4853 "libraries/glad/glad.cpp"
  HEAP32[165892] = $2313; //@line 4853 "libraries/glad/glad.cpp"
  $2314 = (FUNCTION_TABLE_ii[$load & 8191](481827)|0); //@line 4854 "libraries/glad/glad.cpp"
  HEAP32[165893] = $2314; //@line 4854 "libraries/glad/glad.cpp"
  $2315 = (FUNCTION_TABLE_ii[$load & 8191](481857)|0); //@line 4855 "libraries/glad/glad.cpp"
  HEAP32[165894] = $2315; //@line 4855 "libraries/glad/glad.cpp"
  $2316 = (FUNCTION_TABLE_ii[$load & 8191](481887)|0); //@line 4856 "libraries/glad/glad.cpp"
  HEAP32[165895] = $2316; //@line 4856 "libraries/glad/glad.cpp"
  $2317 = (FUNCTION_TABLE_ii[$load & 8191](481920)|0); //@line 4857 "libraries/glad/glad.cpp"
  HEAP32[165896] = $2317; //@line 4857 "libraries/glad/glad.cpp"
  $2318 = (FUNCTION_TABLE_ii[$load & 8191](481953)|0); //@line 4858 "libraries/glad/glad.cpp"
  HEAP32[165897] = $2318; //@line 4858 "libraries/glad/glad.cpp"
  $2319 = (FUNCTION_TABLE_ii[$load & 8191](481986)|0); //@line 4859 "libraries/glad/glad.cpp"
  HEAP32[165898] = $2319; //@line 4859 "libraries/glad/glad.cpp"
  $2320 = (FUNCTION_TABLE_ii[$load & 8191](482017)|0); //@line 4860 "libraries/glad/glad.cpp"
  HEAP32[165899] = $2320; //@line 4860 "libraries/glad/glad.cpp"
  $2321 = (FUNCTION_TABLE_ii[$load & 8191](482048)|0); //@line 4861 "libraries/glad/glad.cpp"
  HEAP32[165900] = $2321; //@line 4861 "libraries/glad/glad.cpp"
  $2322 = (FUNCTION_TABLE_ii[$load & 8191](482079)|0); //@line 4862 "libraries/glad/glad.cpp"
  HEAP32[165901] = $2322; //@line 4862 "libraries/glad/glad.cpp"
  $2323 = (FUNCTION_TABLE_ii[$load & 8191](482110)|0); //@line 4863 "libraries/glad/glad.cpp"
  HEAP32[165902] = $2323; //@line 4863 "libraries/glad/glad.cpp"
  $2324 = (FUNCTION_TABLE_ii[$load & 8191](482144)|0); //@line 4864 "libraries/glad/glad.cpp"
  HEAP32[165903] = $2324; //@line 4864 "libraries/glad/glad.cpp"
  $2325 = (FUNCTION_TABLE_ii[$load & 8191](482178)|0); //@line 4865 "libraries/glad/glad.cpp"
  HEAP32[165904] = $2325; //@line 4865 "libraries/glad/glad.cpp"
  $2326 = (FUNCTION_TABLE_ii[$load & 8191](482212)|0); //@line 4866 "libraries/glad/glad.cpp"
  HEAP32[165905] = $2326; //@line 4866 "libraries/glad/glad.cpp"
  $2327 = (FUNCTION_TABLE_ii[$load & 8191](482244)|0); //@line 4867 "libraries/glad/glad.cpp"
  HEAP32[165906] = $2327; //@line 4867 "libraries/glad/glad.cpp"
  $2328 = (FUNCTION_TABLE_ii[$load & 8191](482270)|0); //@line 4868 "libraries/glad/glad.cpp"
  HEAP32[165907] = $2328; //@line 4868 "libraries/glad/glad.cpp"
  $2329 = (FUNCTION_TABLE_ii[$load & 8191](482296)|0); //@line 4869 "libraries/glad/glad.cpp"
  HEAP32[165908] = $2329; //@line 4869 "libraries/glad/glad.cpp"
  $2330 = (FUNCTION_TABLE_ii[$load & 8191](482322)|0); //@line 4870 "libraries/glad/glad.cpp"
  HEAP32[165909] = $2330; //@line 4870 "libraries/glad/glad.cpp"
  $2331 = (FUNCTION_TABLE_ii[$load & 8191](482348)|0); //@line 4871 "libraries/glad/glad.cpp"
  HEAP32[165910] = $2331; //@line 4871 "libraries/glad/glad.cpp"
  $2332 = (FUNCTION_TABLE_ii[$load & 8191](482369)|0); //@line 4872 "libraries/glad/glad.cpp"
  HEAP32[165911] = $2332; //@line 4872 "libraries/glad/glad.cpp"
  $2333 = (FUNCTION_TABLE_ii[$load & 8191](482393)|0); //@line 4873 "libraries/glad/glad.cpp"
  HEAP32[165912] = $2333; //@line 4873 "libraries/glad/glad.cpp"
  $2334 = (FUNCTION_TABLE_ii[$load & 8191](482413)|0); //@line 4874 "libraries/glad/glad.cpp"
  HEAP32[165913] = $2334; //@line 4874 "libraries/glad/glad.cpp"
  $2335 = (FUNCTION_TABLE_ii[$load & 8191](482435)|0); //@line 4875 "libraries/glad/glad.cpp"
  HEAP32[165914] = $2335; //@line 4875 "libraries/glad/glad.cpp"
  $2336 = (FUNCTION_TABLE_ii[$load & 8191](482466)|0); //@line 4876 "libraries/glad/glad.cpp"
  HEAP32[165915] = $2336; //@line 4876 "libraries/glad/glad.cpp"
  $2337 = (FUNCTION_TABLE_ii[$load & 8191](482494)|0); //@line 4877 "libraries/glad/glad.cpp"
  HEAP32[165916] = $2337; //@line 4877 "libraries/glad/glad.cpp"
  $2338 = (FUNCTION_TABLE_ii[$load & 8191](482521)|0); //@line 4878 "libraries/glad/glad.cpp"
  HEAP32[165917] = $2338; //@line 4878 "libraries/glad/glad.cpp"
  $2339 = (FUNCTION_TABLE_ii[$load & 8191](482543)|0); //@line 4879 "libraries/glad/glad.cpp"
  HEAP32[165918] = $2339; //@line 4879 "libraries/glad/glad.cpp"
  $2340 = (FUNCTION_TABLE_ii[$load & 8191](482565)|0); //@line 4880 "libraries/glad/glad.cpp"
  HEAP32[165919] = $2340; //@line 4880 "libraries/glad/glad.cpp"
  $2341 = (FUNCTION_TABLE_ii[$load & 8191](482587)|0); //@line 4881 "libraries/glad/glad.cpp"
  HEAP32[165920] = $2341; //@line 4881 "libraries/glad/glad.cpp"
  $2342 = (FUNCTION_TABLE_ii[$load & 8191](482609)|0); //@line 4882 "libraries/glad/glad.cpp"
  HEAP32[165921] = $2342; //@line 4882 "libraries/glad/glad.cpp"
  $2343 = (FUNCTION_TABLE_ii[$load & 8191](482631)|0); //@line 4883 "libraries/glad/glad.cpp"
  HEAP32[165922] = $2343; //@line 4883 "libraries/glad/glad.cpp"
  $2344 = (FUNCTION_TABLE_ii[$load & 8191](482653)|0); //@line 4884 "libraries/glad/glad.cpp"
  HEAP32[165923] = $2344; //@line 4884 "libraries/glad/glad.cpp"
  $2345 = (FUNCTION_TABLE_ii[$load & 8191](482675)|0); //@line 4885 "libraries/glad/glad.cpp"
  HEAP32[165924] = $2345; //@line 4885 "libraries/glad/glad.cpp"
  $2346 = (FUNCTION_TABLE_ii[$load & 8191](482697)|0); //@line 4886 "libraries/glad/glad.cpp"
  HEAP32[165925] = $2346; //@line 4886 "libraries/glad/glad.cpp"
  $2347 = (FUNCTION_TABLE_ii[$load & 8191](482720)|0); //@line 4887 "libraries/glad/glad.cpp"
  HEAP32[165926] = $2347; //@line 4887 "libraries/glad/glad.cpp"
  $2348 = (FUNCTION_TABLE_ii[$load & 8191](482743)|0); //@line 4888 "libraries/glad/glad.cpp"
  HEAP32[165927] = $2348; //@line 4888 "libraries/glad/glad.cpp"
  $2349 = (FUNCTION_TABLE_ii[$load & 8191](482766)|0); //@line 4889 "libraries/glad/glad.cpp"
  HEAP32[165928] = $2349; //@line 4889 "libraries/glad/glad.cpp"
  $2350 = (FUNCTION_TABLE_ii[$load & 8191](482789)|0); //@line 4890 "libraries/glad/glad.cpp"
  HEAP32[165929] = $2350; //@line 4890 "libraries/glad/glad.cpp"
  $2351 = (FUNCTION_TABLE_ii[$load & 8191](482812)|0); //@line 4891 "libraries/glad/glad.cpp"
  HEAP32[165930] = $2351; //@line 4891 "libraries/glad/glad.cpp"
  $2352 = (FUNCTION_TABLE_ii[$load & 8191](482835)|0); //@line 4892 "libraries/glad/glad.cpp"
  HEAP32[165931] = $2352; //@line 4892 "libraries/glad/glad.cpp"
  $2353 = (FUNCTION_TABLE_ii[$load & 8191](482858)|0); //@line 4893 "libraries/glad/glad.cpp"
  HEAP32[165932] = $2353; //@line 4893 "libraries/glad/glad.cpp"
  $2354 = (FUNCTION_TABLE_ii[$load & 8191](482881)|0); //@line 4894 "libraries/glad/glad.cpp"
  HEAP32[165933] = $2354; //@line 4894 "libraries/glad/glad.cpp"
  $2355 = (FUNCTION_TABLE_ii[$load & 8191](482910)|0); //@line 4895 "libraries/glad/glad.cpp"
  HEAP32[165934] = $2355; //@line 4895 "libraries/glad/glad.cpp"
  $2356 = (FUNCTION_TABLE_ii[$load & 8191](482939)|0); //@line 4896 "libraries/glad/glad.cpp"
  HEAP32[165935] = $2356; //@line 4896 "libraries/glad/glad.cpp"
  $2357 = (FUNCTION_TABLE_ii[$load & 8191](482968)|0); //@line 4897 "libraries/glad/glad.cpp"
  HEAP32[165936] = $2357; //@line 4897 "libraries/glad/glad.cpp"
  $2358 = (FUNCTION_TABLE_ii[$load & 8191](482999)|0); //@line 4898 "libraries/glad/glad.cpp"
  HEAP32[165937] = $2358; //@line 4898 "libraries/glad/glad.cpp"
  $2359 = (FUNCTION_TABLE_ii[$load & 8191](483030)|0); //@line 4899 "libraries/glad/glad.cpp"
  HEAP32[165938] = $2359; //@line 4899 "libraries/glad/glad.cpp"
  $2360 = (FUNCTION_TABLE_ii[$load & 8191](483061)|0); //@line 4900 "libraries/glad/glad.cpp"
  HEAP32[165939] = $2360; //@line 4900 "libraries/glad/glad.cpp"
  $2361 = (FUNCTION_TABLE_ii[$load & 8191](483092)|0); //@line 4901 "libraries/glad/glad.cpp"
  HEAP32[165940] = $2361; //@line 4901 "libraries/glad/glad.cpp"
  $2362 = (FUNCTION_TABLE_ii[$load & 8191](483123)|0); //@line 4902 "libraries/glad/glad.cpp"
  HEAP32[165941] = $2362; //@line 4902 "libraries/glad/glad.cpp"
  $2363 = (FUNCTION_TABLE_ii[$load & 8191](483154)|0); //@line 4903 "libraries/glad/glad.cpp"
  HEAP32[165942] = $2363; //@line 4903 "libraries/glad/glad.cpp"
  $2364 = (FUNCTION_TABLE_ii[$load & 8191](483173)|0); //@line 4904 "libraries/glad/glad.cpp"
  HEAP32[165943] = $2364; //@line 4904 "libraries/glad/glad.cpp"
  $2365 = (FUNCTION_TABLE_ii[$load & 8191](483193)|0); //@line 4905 "libraries/glad/glad.cpp"
  HEAP32[165944] = $2365; //@line 4905 "libraries/glad/glad.cpp"
  $2366 = (FUNCTION_TABLE_ii[$load & 8191](483218)|0); //@line 4906 "libraries/glad/glad.cpp"
  HEAP32[165945] = $2366; //@line 4906 "libraries/glad/glad.cpp"
  $2367 = (FUNCTION_TABLE_ii[$load & 8191](483244)|0); //@line 4907 "libraries/glad/glad.cpp"
  HEAP32[165946] = $2367; //@line 4907 "libraries/glad/glad.cpp"
  $2368 = (FUNCTION_TABLE_ii[$load & 8191](483272)|0); //@line 4908 "libraries/glad/glad.cpp"
  HEAP32[165947] = $2368; //@line 4908 "libraries/glad/glad.cpp"
  $2369 = (FUNCTION_TABLE_ii[$load & 8191](483301)|0); //@line 4909 "libraries/glad/glad.cpp"
  HEAP32[165948] = $2369; //@line 4909 "libraries/glad/glad.cpp"
  $2370 = (FUNCTION_TABLE_ii[$load & 8191](483327)|0); //@line 4910 "libraries/glad/glad.cpp"
  HEAP32[165949] = $2370; //@line 4910 "libraries/glad/glad.cpp"
  $2371 = (FUNCTION_TABLE_ii[$load & 8191](483354)|0); //@line 4911 "libraries/glad/glad.cpp"
  HEAP32[165950] = $2371; //@line 4911 "libraries/glad/glad.cpp"
  $2372 = (FUNCTION_TABLE_ii[$load & 8191](483383)|0); //@line 4912 "libraries/glad/glad.cpp"
  HEAP32[165951] = $2372; //@line 4912 "libraries/glad/glad.cpp"
  $2373 = (FUNCTION_TABLE_ii[$load & 8191](483413)|0); //@line 4913 "libraries/glad/glad.cpp"
  HEAP32[165952] = $2373; //@line 4913 "libraries/glad/glad.cpp"
  $2374 = (FUNCTION_TABLE_ii[$load & 8191](483436)|0); //@line 4914 "libraries/glad/glad.cpp"
  HEAP32[165953] = $2374; //@line 4914 "libraries/glad/glad.cpp"
  $2375 = (FUNCTION_TABLE_ii[$load & 8191](483459)|0); //@line 4915 "libraries/glad/glad.cpp"
  HEAP32[165954] = $2375; //@line 4915 "libraries/glad/glad.cpp"
  $2376 = (FUNCTION_TABLE_ii[$load & 8191](483482)|0); //@line 4916 "libraries/glad/glad.cpp"
  HEAP32[165955] = $2376; //@line 4916 "libraries/glad/glad.cpp"
  $2377 = (FUNCTION_TABLE_ii[$load & 8191](483505)|0); //@line 4917 "libraries/glad/glad.cpp"
  HEAP32[165956] = $2377; //@line 4917 "libraries/glad/glad.cpp"
  $2378 = (FUNCTION_TABLE_ii[$load & 8191](483529)|0); //@line 4918 "libraries/glad/glad.cpp"
  HEAP32[165957] = $2378; //@line 4918 "libraries/glad/glad.cpp"
  $2379 = (FUNCTION_TABLE_ii[$load & 8191](483553)|0); //@line 4919 "libraries/glad/glad.cpp"
  HEAP32[165958] = $2379; //@line 4919 "libraries/glad/glad.cpp"
  $2380 = (FUNCTION_TABLE_ii[$load & 8191](483577)|0); //@line 4920 "libraries/glad/glad.cpp"
  HEAP32[165959] = $2380; //@line 4920 "libraries/glad/glad.cpp"
  $2381 = (FUNCTION_TABLE_ii[$load & 8191](483601)|0); //@line 4921 "libraries/glad/glad.cpp"
  HEAP32[165960] = $2381; //@line 4921 "libraries/glad/glad.cpp"
  $2382 = (FUNCTION_TABLE_ii[$load & 8191](483637)|0); //@line 4922 "libraries/glad/glad.cpp"
  HEAP32[165961] = $2382; //@line 4922 "libraries/glad/glad.cpp"
  $2383 = (FUNCTION_TABLE_ii[$load & 8191](483672)|0); //@line 4923 "libraries/glad/glad.cpp"
  HEAP32[165962] = $2383; //@line 4923 "libraries/glad/glad.cpp"
  $2384 = (FUNCTION_TABLE_ii[$load & 8191](483708)|0); //@line 4924 "libraries/glad/glad.cpp"
  HEAP32[165963] = $2384; //@line 4924 "libraries/glad/glad.cpp"
  $2385 = (FUNCTION_TABLE_ii[$load & 8191](483745)|0); //@line 4925 "libraries/glad/glad.cpp"
  HEAP32[165964] = $2385; //@line 4925 "libraries/glad/glad.cpp"
  $2386 = (FUNCTION_TABLE_ii[$load & 8191](483781)|0); //@line 4926 "libraries/glad/glad.cpp"
  HEAP32[165965] = $2386; //@line 4926 "libraries/glad/glad.cpp"
  $2387 = (FUNCTION_TABLE_ii[$load & 8191](483818)|0); //@line 4927 "libraries/glad/glad.cpp"
  HEAP32[165966] = $2387; //@line 4927 "libraries/glad/glad.cpp"
  $2388 = (FUNCTION_TABLE_ii[$load & 8191](483856)|0); //@line 4928 "libraries/glad/glad.cpp"
  HEAP32[165967] = $2388; //@line 4928 "libraries/glad/glad.cpp"
  $2389 = (FUNCTION_TABLE_ii[$load & 8191](483894)|0); //@line 4929 "libraries/glad/glad.cpp"
  HEAP32[165968] = $2389; //@line 4929 "libraries/glad/glad.cpp"
  $2390 = (FUNCTION_TABLE_ii[$load & 8191](483933)|0); //@line 4930 "libraries/glad/glad.cpp"
  HEAP32[165969] = $2390; //@line 4930 "libraries/glad/glad.cpp"
  $2391 = (FUNCTION_TABLE_ii[$load & 8191](483957)|0); //@line 4931 "libraries/glad/glad.cpp"
  HEAP32[165970] = $2391; //@line 4931 "libraries/glad/glad.cpp"
  $2392 = (FUNCTION_TABLE_ii[$load & 8191](483982)|0); //@line 4932 "libraries/glad/glad.cpp"
  HEAP32[165971] = $2392; //@line 4932 "libraries/glad/glad.cpp"
  $2393 = (FUNCTION_TABLE_ii[$load & 8191](483999)|0); //@line 4933 "libraries/glad/glad.cpp"
  HEAP32[165972] = $2393; //@line 4933 "libraries/glad/glad.cpp"
  $2394 = (FUNCTION_TABLE_ii[$load & 8191](484017)|0); //@line 4934 "libraries/glad/glad.cpp"
  HEAP32[165973] = $2394; //@line 4934 "libraries/glad/glad.cpp"
  $2395 = (FUNCTION_TABLE_ii[$load & 8191](484036)|0); //@line 4935 "libraries/glad/glad.cpp"
  HEAP32[165974] = $2395; //@line 4935 "libraries/glad/glad.cpp"
  $2396 = (FUNCTION_TABLE_ii[$load & 8191](484060)|0); //@line 4936 "libraries/glad/glad.cpp"
  HEAP32[165975] = $2396; //@line 4936 "libraries/glad/glad.cpp"
  $2397 = (FUNCTION_TABLE_ii[$load & 8191](484094)|0); //@line 4937 "libraries/glad/glad.cpp"
  HEAP32[165976] = $2397; //@line 4937 "libraries/glad/glad.cpp"
  $2398 = (FUNCTION_TABLE_ii[$load & 8191](484129)|0); //@line 4938 "libraries/glad/glad.cpp"
  HEAP32[165977] = $2398; //@line 4938 "libraries/glad/glad.cpp"
  $2399 = (FUNCTION_TABLE_ii[$load & 8191](484163)|0); //@line 4939 "libraries/glad/glad.cpp"
  HEAP32[165978] = $2399; //@line 4939 "libraries/glad/glad.cpp"
  $2400 = (FUNCTION_TABLE_ii[$load & 8191](484198)|0); //@line 4940 "libraries/glad/glad.cpp"
  HEAP32[165979] = $2400; //@line 4940 "libraries/glad/glad.cpp"
  $2401 = (FUNCTION_TABLE_ii[$load & 8191](484235)|0); //@line 4941 "libraries/glad/glad.cpp"
  HEAP32[165980] = $2401; //@line 4941 "libraries/glad/glad.cpp"
  $2402 = (FUNCTION_TABLE_ii[$load & 8191](484272)|0); //@line 4942 "libraries/glad/glad.cpp"
  HEAP32[165981] = $2402; //@line 4942 "libraries/glad/glad.cpp"
  $2403 = (FUNCTION_TABLE_ii[$load & 8191](484295)|0); //@line 4943 "libraries/glad/glad.cpp"
  HEAP32[165982] = $2403; //@line 4943 "libraries/glad/glad.cpp"
  $2404 = (FUNCTION_TABLE_ii[$load & 8191](484322)|0); //@line 4944 "libraries/glad/glad.cpp"
  HEAP32[165983] = $2404; //@line 4944 "libraries/glad/glad.cpp"
  $2405 = (FUNCTION_TABLE_ii[$load & 8191](484352)|0); //@line 4945 "libraries/glad/glad.cpp"
  HEAP32[165984] = $2405; //@line 4945 "libraries/glad/glad.cpp"
  $2406 = (FUNCTION_TABLE_ii[$load & 8191](484389)|0); //@line 4946 "libraries/glad/glad.cpp"
  HEAP32[165985] = $2406; //@line 4946 "libraries/glad/glad.cpp"
  $2407 = (FUNCTION_TABLE_ii[$load & 8191](484430)|0); //@line 4947 "libraries/glad/glad.cpp"
  HEAP32[165986] = $2407; //@line 4947 "libraries/glad/glad.cpp"
  $2408 = (FUNCTION_TABLE_ii[$load & 8191](484479)|0); //@line 4948 "libraries/glad/glad.cpp"
  HEAP32[165987] = $2408; //@line 4948 "libraries/glad/glad.cpp"
  $2409 = (FUNCTION_TABLE_ii[$load & 8191](484512)|0); //@line 4949 "libraries/glad/glad.cpp"
  HEAP32[165988] = $2409; //@line 4949 "libraries/glad/glad.cpp"
  $2410 = (FUNCTION_TABLE_ii[$load & 8191](484543)|0); //@line 4950 "libraries/glad/glad.cpp"
  HEAP32[165989] = $2410; //@line 4950 "libraries/glad/glad.cpp"
  $2411 = (FUNCTION_TABLE_ii[$load & 8191](484574)|0); //@line 4951 "libraries/glad/glad.cpp"
  HEAP32[165990] = $2411; //@line 4951 "libraries/glad/glad.cpp"
  $2412 = (FUNCTION_TABLE_ii[$load & 8191](484605)|0); //@line 4952 "libraries/glad/glad.cpp"
  HEAP32[165991] = $2412; //@line 4952 "libraries/glad/glad.cpp"
  $2413 = (FUNCTION_TABLE_ii[$load & 8191](484639)|0); //@line 4953 "libraries/glad/glad.cpp"
  HEAP32[165992] = $2413; //@line 4953 "libraries/glad/glad.cpp"
  $2414 = (FUNCTION_TABLE_ii[$load & 8191](484685)|0); //@line 4954 "libraries/glad/glad.cpp"
  HEAP32[165993] = $2414; //@line 4954 "libraries/glad/glad.cpp"
  $2415 = (FUNCTION_TABLE_ii[$load & 8191](484712)|0); //@line 4955 "libraries/glad/glad.cpp"
  HEAP32[165994] = $2415; //@line 4955 "libraries/glad/glad.cpp"
  $2416 = (FUNCTION_TABLE_ii[$load & 8191](484740)|0); //@line 4956 "libraries/glad/glad.cpp"
  HEAP32[165995] = $2416; //@line 4956 "libraries/glad/glad.cpp"
  $2417 = (FUNCTION_TABLE_ii[$load & 8191](484767)|0); //@line 4957 "libraries/glad/glad.cpp"
  HEAP32[165996] = $2417; //@line 4957 "libraries/glad/glad.cpp"
  $2418 = (FUNCTION_TABLE_ii[$load & 8191](484795)|0); //@line 4958 "libraries/glad/glad.cpp"
  HEAP32[165997] = $2418; //@line 4958 "libraries/glad/glad.cpp"
  $2419 = (FUNCTION_TABLE_ii[$load & 8191](484822)|0); //@line 4959 "libraries/glad/glad.cpp"
  HEAP32[165998] = $2419; //@line 4959 "libraries/glad/glad.cpp"
  $2420 = (FUNCTION_TABLE_ii[$load & 8191](484853)|0); //@line 4960 "libraries/glad/glad.cpp"
  HEAP32[165999] = $2420; //@line 4960 "libraries/glad/glad.cpp"
  $2421 = (FUNCTION_TABLE_ii[$load & 8191](484881)|0); //@line 4961 "libraries/glad/glad.cpp"
  HEAP32[166000] = $2421; //@line 4961 "libraries/glad/glad.cpp"
  $2422 = (FUNCTION_TABLE_ii[$load & 8191](484910)|0); //@line 4962 "libraries/glad/glad.cpp"
  HEAP32[166001] = $2422; //@line 4962 "libraries/glad/glad.cpp"
  $2423 = (FUNCTION_TABLE_ii[$load & 8191](484944)|0); //@line 4963 "libraries/glad/glad.cpp"
  HEAP32[166002] = $2423; //@line 4963 "libraries/glad/glad.cpp"
  $2424 = (FUNCTION_TABLE_ii[$load & 8191](484977)|0); //@line 4964 "libraries/glad/glad.cpp"
  HEAP32[166003] = $2424; //@line 4964 "libraries/glad/glad.cpp"
  $2425 = (FUNCTION_TABLE_ii[$load & 8191](485002)|0); //@line 4965 "libraries/glad/glad.cpp"
  HEAP32[166004] = $2425; //@line 4965 "libraries/glad/glad.cpp"
  $2426 = (FUNCTION_TABLE_ii[$load & 8191](485028)|0); //@line 4966 "libraries/glad/glad.cpp"
  HEAP32[166005] = $2426; //@line 4966 "libraries/glad/glad.cpp"
  $2427 = (FUNCTION_TABLE_ii[$load & 8191](485057)|0); //@line 4967 "libraries/glad/glad.cpp"
  HEAP32[166006] = $2427; //@line 4967 "libraries/glad/glad.cpp"
  $2428 = (FUNCTION_TABLE_ii[$load & 8191](485085)|0); //@line 4968 "libraries/glad/glad.cpp"
  HEAP32[166007] = $2428; //@line 4968 "libraries/glad/glad.cpp"
  $2429 = (FUNCTION_TABLE_ii[$load & 8191](485116)|0); //@line 4969 "libraries/glad/glad.cpp"
  HEAP32[166008] = $2429; //@line 4969 "libraries/glad/glad.cpp"
  $2430 = (FUNCTION_TABLE_ii[$load & 8191](485144)|0); //@line 4970 "libraries/glad/glad.cpp"
  HEAP32[166009] = $2430; //@line 4970 "libraries/glad/glad.cpp"
  $2431 = (FUNCTION_TABLE_ii[$load & 8191](485173)|0); //@line 4971 "libraries/glad/glad.cpp"
  HEAP32[166010] = $2431; //@line 4971 "libraries/glad/glad.cpp"
  $2432 = (FUNCTION_TABLE_ii[$load & 8191](485204)|0); //@line 4972 "libraries/glad/glad.cpp"
  HEAP32[166011] = $2432; //@line 4972 "libraries/glad/glad.cpp"
  $2433 = (FUNCTION_TABLE_ii[$load & 8191](485240)|0); //@line 4973 "libraries/glad/glad.cpp"
  HEAP32[166012] = $2433; //@line 4973 "libraries/glad/glad.cpp"
  $2434 = (FUNCTION_TABLE_ii[$load & 8191](485271)|0); //@line 4974 "libraries/glad/glad.cpp"
  HEAP32[166013] = $2434; //@line 4974 "libraries/glad/glad.cpp"
  $2435 = (FUNCTION_TABLE_ii[$load & 8191](485308)|0); //@line 4975 "libraries/glad/glad.cpp"
  HEAP32[166014] = $2435; //@line 4975 "libraries/glad/glad.cpp"
  $2436 = (FUNCTION_TABLE_ii[$load & 8191](485343)|0); //@line 4976 "libraries/glad/glad.cpp"
  HEAP32[166015] = $2436; //@line 4976 "libraries/glad/glad.cpp"
  $2437 = (FUNCTION_TABLE_ii[$load & 8191](485379)|0); //@line 4977 "libraries/glad/glad.cpp"
  HEAP32[166016] = $2437; //@line 4977 "libraries/glad/glad.cpp"
  $2438 = (FUNCTION_TABLE_ii[$load & 8191](485402)|0); //@line 4978 "libraries/glad/glad.cpp"
  HEAP32[166017] = $2438; //@line 4978 "libraries/glad/glad.cpp"
  $2439 = (FUNCTION_TABLE_ii[$load & 8191](485426)|0); //@line 4979 "libraries/glad/glad.cpp"
  HEAP32[166018] = $2439; //@line 4979 "libraries/glad/glad.cpp"
  $2440 = (FUNCTION_TABLE_ii[$load & 8191](485455)|0); //@line 4980 "libraries/glad/glad.cpp"
  HEAP32[166019] = $2440; //@line 4980 "libraries/glad/glad.cpp"
  $2441 = (FUNCTION_TABLE_ii[$load & 8191](485485)|0); //@line 4981 "libraries/glad/glad.cpp"
  HEAP32[166020] = $2441; //@line 4981 "libraries/glad/glad.cpp"
  $2442 = (FUNCTION_TABLE_ii[$load & 8191](485513)|0); //@line 4982 "libraries/glad/glad.cpp"
  HEAP32[166021] = $2442; //@line 4982 "libraries/glad/glad.cpp"
  $2443 = (FUNCTION_TABLE_ii[$load & 8191](485541)|0); //@line 4983 "libraries/glad/glad.cpp"
  HEAP32[166022] = $2443; //@line 4983 "libraries/glad/glad.cpp"
  $2444 = (FUNCTION_TABLE_ii[$load & 8191](485571)|0); //@line 4984 "libraries/glad/glad.cpp"
  HEAP32[166023] = $2444; //@line 4984 "libraries/glad/glad.cpp"
  $2445 = (FUNCTION_TABLE_ii[$load & 8191](485601)|0); //@line 4985 "libraries/glad/glad.cpp"
  HEAP32[166024] = $2445; //@line 4985 "libraries/glad/glad.cpp"
  $2446 = (FUNCTION_TABLE_ii[$load & 8191](485626)|0); //@line 4986 "libraries/glad/glad.cpp"
  HEAP32[166025] = $2446; //@line 4986 "libraries/glad/glad.cpp"
  $2447 = (FUNCTION_TABLE_ii[$load & 8191](485659)|0); //@line 4987 "libraries/glad/glad.cpp"
  HEAP32[166026] = $2447; //@line 4987 "libraries/glad/glad.cpp"
  $2448 = (FUNCTION_TABLE_ii[$load & 8191](485683)|0); //@line 4988 "libraries/glad/glad.cpp"
  HEAP32[166027] = $2448; //@line 4988 "libraries/glad/glad.cpp"
  $2449 = (FUNCTION_TABLE_ii[$load & 8191](485709)|0); //@line 4989 "libraries/glad/glad.cpp"
  HEAP32[166028] = $2449; //@line 4989 "libraries/glad/glad.cpp"
  $2450 = (FUNCTION_TABLE_ii[$load & 8191](485738)|0); //@line 4990 "libraries/glad/glad.cpp"
  HEAP32[166029] = $2450; //@line 4990 "libraries/glad/glad.cpp"
  $2451 = (FUNCTION_TABLE_ii[$load & 8191](485770)|0); //@line 4991 "libraries/glad/glad.cpp"
  HEAP32[166030] = $2451; //@line 4991 "libraries/glad/glad.cpp"
  $2452 = (FUNCTION_TABLE_ii[$load & 8191](485806)|0); //@line 4992 "libraries/glad/glad.cpp"
  HEAP32[166031] = $2452; //@line 4992 "libraries/glad/glad.cpp"
  $2453 = (FUNCTION_TABLE_ii[$load & 8191](485828)|0); //@line 4993 "libraries/glad/glad.cpp"
  HEAP32[166032] = $2453; //@line 4993 "libraries/glad/glad.cpp"
  $2454 = (FUNCTION_TABLE_ii[$load & 8191](485850)|0); //@line 4994 "libraries/glad/glad.cpp"
  HEAP32[166033] = $2454; //@line 4994 "libraries/glad/glad.cpp"
  $2455 = (FUNCTION_TABLE_ii[$load & 8191](485872)|0); //@line 4995 "libraries/glad/glad.cpp"
  HEAP32[166034] = $2455; //@line 4995 "libraries/glad/glad.cpp"
  $2456 = (FUNCTION_TABLE_ii[$load & 8191](485894)|0); //@line 4996 "libraries/glad/glad.cpp"
  HEAP32[166035] = $2456; //@line 4996 "libraries/glad/glad.cpp"
  $2457 = (FUNCTION_TABLE_ii[$load & 8191](485917)|0); //@line 4997 "libraries/glad/glad.cpp"
  HEAP32[166036] = $2457; //@line 4997 "libraries/glad/glad.cpp"
  $2458 = (FUNCTION_TABLE_ii[$load & 8191](485940)|0); //@line 4998 "libraries/glad/glad.cpp"
  HEAP32[166037] = $2458; //@line 4998 "libraries/glad/glad.cpp"
  $2459 = (FUNCTION_TABLE_ii[$load & 8191](485963)|0); //@line 4999 "libraries/glad/glad.cpp"
  HEAP32[166038] = $2459; //@line 4999 "libraries/glad/glad.cpp"
  $2460 = (FUNCTION_TABLE_ii[$load & 8191](485986)|0); //@line 5000 "libraries/glad/glad.cpp"
  HEAP32[166039] = $2460; //@line 5000 "libraries/glad/glad.cpp"
  $2461 = (FUNCTION_TABLE_ii[$load & 8191](486015)|0); //@line 5001 "libraries/glad/glad.cpp"
  HEAP32[166040] = $2461; //@line 5001 "libraries/glad/glad.cpp"
  $2462 = (FUNCTION_TABLE_ii[$load & 8191](486044)|0); //@line 5002 "libraries/glad/glad.cpp"
  HEAP32[166041] = $2462; //@line 5002 "libraries/glad/glad.cpp"
  $2463 = (FUNCTION_TABLE_ii[$load & 8191](486073)|0); //@line 5003 "libraries/glad/glad.cpp"
  HEAP32[166042] = $2463; //@line 5003 "libraries/glad/glad.cpp"
  $2464 = (FUNCTION_TABLE_ii[$load & 8191](486104)|0); //@line 5004 "libraries/glad/glad.cpp"
  HEAP32[166043] = $2464; //@line 5004 "libraries/glad/glad.cpp"
  $2465 = (FUNCTION_TABLE_ii[$load & 8191](486135)|0); //@line 5005 "libraries/glad/glad.cpp"
  HEAP32[166044] = $2465; //@line 5005 "libraries/glad/glad.cpp"
  $2466 = (FUNCTION_TABLE_ii[$load & 8191](486166)|0); //@line 5006 "libraries/glad/glad.cpp"
  HEAP32[166045] = $2466; //@line 5006 "libraries/glad/glad.cpp"
  $2467 = (FUNCTION_TABLE_ii[$load & 8191](486197)|0); //@line 5007 "libraries/glad/glad.cpp"
  HEAP32[166046] = $2467; //@line 5007 "libraries/glad/glad.cpp"
  $2468 = (FUNCTION_TABLE_ii[$load & 8191](486228)|0); //@line 5008 "libraries/glad/glad.cpp"
  HEAP32[166047] = $2468; //@line 5008 "libraries/glad/glad.cpp"
  $2469 = (FUNCTION_TABLE_ii[$load & 8191](486259)|0); //@line 5009 "libraries/glad/glad.cpp"
  HEAP32[166048] = $2469; //@line 5009 "libraries/glad/glad.cpp"
  $2470 = (FUNCTION_TABLE_ii[$load & 8191](486283)|0); //@line 5010 "libraries/glad/glad.cpp"
  HEAP32[166049] = $2470; //@line 5010 "libraries/glad/glad.cpp"
  $2471 = (FUNCTION_TABLE_ii[$load & 8191](486305)|0); //@line 5011 "libraries/glad/glad.cpp"
  HEAP32[166050] = $2471; //@line 5011 "libraries/glad/glad.cpp"
  $2472 = (FUNCTION_TABLE_ii[$load & 8191](486327)|0); //@line 5012 "libraries/glad/glad.cpp"
  HEAP32[166051] = $2472; //@line 5012 "libraries/glad/glad.cpp"
  $2473 = (FUNCTION_TABLE_ii[$load & 8191](486349)|0); //@line 5013 "libraries/glad/glad.cpp"
  HEAP32[166052] = $2473; //@line 5013 "libraries/glad/glad.cpp"
  $2474 = (FUNCTION_TABLE_ii[$load & 8191](486382)|0); //@line 5014 "libraries/glad/glad.cpp"
  HEAP32[166053] = $2474; //@line 5014 "libraries/glad/glad.cpp"
  $2475 = (FUNCTION_TABLE_ii[$load & 8191](486415)|0); //@line 5015 "libraries/glad/glad.cpp"
  HEAP32[166054] = $2475; //@line 5015 "libraries/glad/glad.cpp"
  $2476 = (FUNCTION_TABLE_ii[$load & 8191](486448)|0); //@line 5016 "libraries/glad/glad.cpp"
  HEAP32[166055] = $2476; //@line 5016 "libraries/glad/glad.cpp"
  $2477 = (FUNCTION_TABLE_ii[$load & 8191](486483)|0); //@line 5017 "libraries/glad/glad.cpp"
  HEAP32[166056] = $2477; //@line 5017 "libraries/glad/glad.cpp"
  $2478 = (FUNCTION_TABLE_ii[$load & 8191](486519)|0); //@line 5018 "libraries/glad/glad.cpp"
  HEAP32[166057] = $2478; //@line 5018 "libraries/glad/glad.cpp"
  $2479 = (FUNCTION_TABLE_ii[$load & 8191](486555)|0); //@line 5019 "libraries/glad/glad.cpp"
  HEAP32[166058] = $2479; //@line 5019 "libraries/glad/glad.cpp"
  $2480 = (FUNCTION_TABLE_ii[$load & 8191](486591)|0); //@line 5020 "libraries/glad/glad.cpp"
  HEAP32[166059] = $2480; //@line 5020 "libraries/glad/glad.cpp"
  $2481 = (FUNCTION_TABLE_ii[$load & 8191](486628)|0); //@line 5021 "libraries/glad/glad.cpp"
  HEAP32[166060] = $2481; //@line 5021 "libraries/glad/glad.cpp"
  $2482 = (FUNCTION_TABLE_ii[$load & 8191](486664)|0); //@line 5022 "libraries/glad/glad.cpp"
  HEAP32[166061] = $2482; //@line 5022 "libraries/glad/glad.cpp"
  $2483 = (FUNCTION_TABLE_ii[$load & 8191](486691)|0); //@line 5023 "libraries/glad/glad.cpp"
  HEAP32[166062] = $2483; //@line 5023 "libraries/glad/glad.cpp"
 }
 $2484 = HEAP8[1252174]|0; //@line 5028 "libraries/glad/glad.cpp"
 $2485 = ($2484<<24>>24)==(0); //@line 5028 "libraries/glad/glad.cpp"
 if (!($2485)) {
  $2486 = (FUNCTION_TABLE_ii[$load & 8191](486727)|0); //@line 5029 "libraries/glad/glad.cpp"
  HEAP32[166063] = $2486; //@line 5029 "libraries/glad/glad.cpp"
 }
 $2487 = HEAP8[1252175]|0; //@line 5034 "libraries/glad/glad.cpp"
 $2488 = ($2487<<24>>24)==(0); //@line 5034 "libraries/glad/glad.cpp"
 if (!($2488)) {
  $2489 = (FUNCTION_TABLE_ii[$load & 8191](486751)|0); //@line 5035 "libraries/glad/glad.cpp"
  HEAP32[166064] = $2489; //@line 5035 "libraries/glad/glad.cpp"
  $2490 = (FUNCTION_TABLE_ii[$load & 8191](486767)|0); //@line 5036 "libraries/glad/glad.cpp"
  HEAP32[166065] = $2490; //@line 5036 "libraries/glad/glad.cpp"
  $2491 = (FUNCTION_TABLE_ii[$load & 8191](486786)|0); //@line 5037 "libraries/glad/glad.cpp"
  HEAP32[166066] = $2491; //@line 5037 "libraries/glad/glad.cpp"
  $2492 = (FUNCTION_TABLE_ii[$load & 8191](486799)|0); //@line 5038 "libraries/glad/glad.cpp"
  HEAP32[166067] = $2492; //@line 5038 "libraries/glad/glad.cpp"
  $2493 = (FUNCTION_TABLE_ii[$load & 8191](486815)|0); //@line 5039 "libraries/glad/glad.cpp"
  HEAP32[166068] = $2493; //@line 5039 "libraries/glad/glad.cpp"
  $2494 = (FUNCTION_TABLE_ii[$load & 8191](486829)|0); //@line 5040 "libraries/glad/glad.cpp"
  HEAP32[166069] = $2494; //@line 5040 "libraries/glad/glad.cpp"
  $2495 = (FUNCTION_TABLE_ii[$load & 8191](486847)|0); //@line 5041 "libraries/glad/glad.cpp"
  HEAP32[166070] = $2495; //@line 5041 "libraries/glad/glad.cpp"
  $2496 = (FUNCTION_TABLE_ii[$load & 8191](486863)|0); //@line 5042 "libraries/glad/glad.cpp"
  HEAP32[166071] = $2496; //@line 5042 "libraries/glad/glad.cpp"
  $2497 = (FUNCTION_TABLE_ii[$load & 8191](486885)|0); //@line 5043 "libraries/glad/glad.cpp"
  HEAP32[166072] = $2497; //@line 5043 "libraries/glad/glad.cpp"
  $2498 = (FUNCTION_TABLE_ii[$load & 8191](486908)|0); //@line 5044 "libraries/glad/glad.cpp"
  HEAP32[166073] = $2498; //@line 5044 "libraries/glad/glad.cpp"
  $2499 = (FUNCTION_TABLE_ii[$load & 8191](486932)|0); //@line 5045 "libraries/glad/glad.cpp"
  HEAP32[166074] = $2499; //@line 5045 "libraries/glad/glad.cpp"
 }
 $2500 = HEAP8[1252176]|0; //@line 5050 "libraries/glad/glad.cpp"
 $2501 = ($2500<<24>>24)==(0); //@line 5050 "libraries/glad/glad.cpp"
 if (!($2501)) {
  $2502 = (FUNCTION_TABLE_ii[$load & 8191](486957)|0); //@line 5051 "libraries/glad/glad.cpp"
  HEAP32[166075] = $2502; //@line 5051 "libraries/glad/glad.cpp"
 }
 $2503 = HEAP8[1252177]|0; //@line 5056 "libraries/glad/glad.cpp"
 $2504 = ($2503<<24>>24)==(0); //@line 5056 "libraries/glad/glad.cpp"
 if (!($2504)) {
  $2505 = (FUNCTION_TABLE_ii[$load & 8191](486974)|0); //@line 5057 "libraries/glad/glad.cpp"
  HEAP32[166076] = $2505; //@line 5057 "libraries/glad/glad.cpp"
  $2506 = (FUNCTION_TABLE_ii[$load & 8191](481773)|0); //@line 5058 "libraries/glad/glad.cpp"
  HEAP32[165891] = $2506; //@line 5058 "libraries/glad/glad.cpp"
  $2507 = (FUNCTION_TABLE_ii[$load & 8191](481749)|0); //@line 5059 "libraries/glad/glad.cpp"
  HEAP32[165890] = $2507; //@line 5059 "libraries/glad/glad.cpp"
  $2508 = (FUNCTION_TABLE_ii[$load & 8191](481688)|0); //@line 5060 "libraries/glad/glad.cpp"
  HEAP32[165887] = $2508; //@line 5060 "libraries/glad/glad.cpp"
  $2509 = (FUNCTION_TABLE_ii[$load & 8191](481707)|0); //@line 5061 "libraries/glad/glad.cpp"
  HEAP32[165888] = $2509; //@line 5061 "libraries/glad/glad.cpp"
  $2510 = (FUNCTION_TABLE_ii[$load & 8191](481727)|0); //@line 5062 "libraries/glad/glad.cpp"
  HEAP32[165889] = $2510; //@line 5062 "libraries/glad/glad.cpp"
 }
 $2511 = HEAP8[1252178]|0; //@line 5067 "libraries/glad/glad.cpp"
 $2512 = ($2511<<24>>24)==(0); //@line 5067 "libraries/glad/glad.cpp"
 if (!($2512)) {
  $2513 = (FUNCTION_TABLE_ii[$load & 8191](486996)|0); //@line 5068 "libraries/glad/glad.cpp"
  HEAP32[166077] = $2513; //@line 5068 "libraries/glad/glad.cpp"
  $2514 = (FUNCTION_TABLE_ii[$load & 8191](487009)|0); //@line 5069 "libraries/glad/glad.cpp"
  HEAP32[166078] = $2514; //@line 5069 "libraries/glad/glad.cpp"
  $2515 = (FUNCTION_TABLE_ii[$load & 8191](487023)|0); //@line 5070 "libraries/glad/glad.cpp"
  HEAP32[166079] = $2515; //@line 5070 "libraries/glad/glad.cpp"
  $2516 = (FUNCTION_TABLE_ii[$load & 8191](487043)|0); //@line 5071 "libraries/glad/glad.cpp"
  HEAP32[166080] = $2516; //@line 5071 "libraries/glad/glad.cpp"
  $2517 = (FUNCTION_TABLE_ii[$load & 8191](487071)|0); //@line 5072 "libraries/glad/glad.cpp"
  HEAP32[166081] = $2517; //@line 5072 "libraries/glad/glad.cpp"
  $2518 = (FUNCTION_TABLE_ii[$load & 8191](487087)|0); //@line 5073 "libraries/glad/glad.cpp"
  HEAP32[166082] = $2518; //@line 5073 "libraries/glad/glad.cpp"
  $2519 = (FUNCTION_TABLE_ii[$load & 8191](487111)|0); //@line 5074 "libraries/glad/glad.cpp"
  HEAP32[166083] = $2519; //@line 5074 "libraries/glad/glad.cpp"
  $2520 = (FUNCTION_TABLE_ii[$load & 8191](487127)|0); //@line 5075 "libraries/glad/glad.cpp"
  HEAP32[166084] = $2520; //@line 5075 "libraries/glad/glad.cpp"
 }
 $2521 = HEAP8[1252179]|0; //@line 5080 "libraries/glad/glad.cpp"
 $2522 = ($2521<<24>>24)==(0); //@line 5080 "libraries/glad/glad.cpp"
 if (!($2522)) {
  $2523 = (FUNCTION_TABLE_ii[$load & 8191](487143)|0); //@line 5081 "libraries/glad/glad.cpp"
  HEAP32[166085] = $2523; //@line 5081 "libraries/glad/glad.cpp"
  $2524 = (FUNCTION_TABLE_ii[$load & 8191](487171)|0); //@line 5082 "libraries/glad/glad.cpp"
  HEAP32[166086] = $2524; //@line 5082 "libraries/glad/glad.cpp"
  $2525 = (FUNCTION_TABLE_ii[$load & 8191](487204)|0); //@line 5083 "libraries/glad/glad.cpp"
  HEAP32[166087] = $2525; //@line 5083 "libraries/glad/glad.cpp"
  $2526 = (FUNCTION_TABLE_ii[$load & 8191](487241)|0); //@line 5084 "libraries/glad/glad.cpp"
  HEAP32[166088] = $2526; //@line 5084 "libraries/glad/glad.cpp"
 }
 $2527 = HEAP8[1252180]|0; //@line 5089 "libraries/glad/glad.cpp"
 $2528 = ($2527<<24>>24)==(0); //@line 5089 "libraries/glad/glad.cpp"
 if (!($2528)) {
  $2529 = (FUNCTION_TABLE_ii[$load & 8191](487274)|0); //@line 5090 "libraries/glad/glad.cpp"
  HEAP32[166089] = $2529; //@line 5090 "libraries/glad/glad.cpp"
  $2530 = (FUNCTION_TABLE_ii[$load & 8191](487299)|0); //@line 5091 "libraries/glad/glad.cpp"
  HEAP32[166090] = $2530; //@line 5091 "libraries/glad/glad.cpp"
 }
 $2531 = HEAP8[1252182]|0; //@line 5097 "libraries/glad/glad.cpp"
 $2532 = ($2531<<24>>24)==(0); //@line 5097 "libraries/glad/glad.cpp"
 if (!($2532)) {
  $2533 = (FUNCTION_TABLE_ii[$load & 8191](487326)|0); //@line 5098 "libraries/glad/glad.cpp"
  HEAP32[166091] = $2533; //@line 5098 "libraries/glad/glad.cpp"
 }
 $2534 = HEAP8[1252183]|0; //@line 5103 "libraries/glad/glad.cpp"
 $2535 = ($2534<<24>>24)==(0); //@line 5103 "libraries/glad/glad.cpp"
 if (!($2535)) {
  $2536 = (FUNCTION_TABLE_ii[$load & 8191](487347)|0); //@line 5104 "libraries/glad/glad.cpp"
  HEAP32[166092] = $2536; //@line 5104 "libraries/glad/glad.cpp"
 }
 $2537 = HEAP8[1252185]|0; //@line 5110 "libraries/glad/glad.cpp"
 $2538 = ($2537<<24>>24)==(0); //@line 5110 "libraries/glad/glad.cpp"
 if (!($2538)) {
  $2539 = (FUNCTION_TABLE_ii[$load & 8191](487383)|0); //@line 5111 "libraries/glad/glad.cpp"
  HEAP32[166093] = $2539; //@line 5111 "libraries/glad/glad.cpp"
  $2540 = (FUNCTION_TABLE_ii[$load & 8191](487403)|0); //@line 5112 "libraries/glad/glad.cpp"
  HEAP32[166094] = $2540; //@line 5112 "libraries/glad/glad.cpp"
  $2541 = (FUNCTION_TABLE_ii[$load & 8191](487425)|0); //@line 5113 "libraries/glad/glad.cpp"
  HEAP32[166095] = $2541; //@line 5113 "libraries/glad/glad.cpp"
  $2542 = (FUNCTION_TABLE_ii[$load & 8191](487450)|0); //@line 5114 "libraries/glad/glad.cpp"
  HEAP32[166096] = $2542; //@line 5114 "libraries/glad/glad.cpp"
  $2543 = (FUNCTION_TABLE_ii[$load & 8191](487472)|0); //@line 5115 "libraries/glad/glad.cpp"
  HEAP32[166097] = $2543; //@line 5115 "libraries/glad/glad.cpp"
  $2544 = (FUNCTION_TABLE_ii[$load & 8191](487497)|0); //@line 5116 "libraries/glad/glad.cpp"
  HEAP32[166098] = $2544; //@line 5116 "libraries/glad/glad.cpp"
  $2545 = (FUNCTION_TABLE_ii[$load & 8191](487529)|0); //@line 5117 "libraries/glad/glad.cpp"
  HEAP32[166099] = $2545; //@line 5117 "libraries/glad/glad.cpp"
  $2546 = (FUNCTION_TABLE_ii[$load & 8191](487548)|0); //@line 5118 "libraries/glad/glad.cpp"
  HEAP32[166100] = $2546; //@line 5118 "libraries/glad/glad.cpp"
  $2547 = (FUNCTION_TABLE_ii[$load & 8191](487569)|0); //@line 5119 "libraries/glad/glad.cpp"
  HEAP32[166101] = $2547; //@line 5119 "libraries/glad/glad.cpp"
  $2548 = (FUNCTION_TABLE_ii[$load & 8191](487593)|0); //@line 5120 "libraries/glad/glad.cpp"
  HEAP32[166102] = $2548; //@line 5120 "libraries/glad/glad.cpp"
  $2549 = (FUNCTION_TABLE_ii[$load & 8191](487614)|0); //@line 5121 "libraries/glad/glad.cpp"
  HEAP32[166103] = $2549; //@line 5121 "libraries/glad/glad.cpp"
  $2550 = (FUNCTION_TABLE_ii[$load & 8191](487642)|0); //@line 5122 "libraries/glad/glad.cpp"
  HEAP32[166104] = $2550; //@line 5122 "libraries/glad/glad.cpp"
  $2551 = (FUNCTION_TABLE_ii[$load & 8191](487668)|0); //@line 5123 "libraries/glad/glad.cpp"
  HEAP32[166105] = $2551; //@line 5123 "libraries/glad/glad.cpp"
  $2552 = (FUNCTION_TABLE_ii[$load & 8191](487694)|0); //@line 5124 "libraries/glad/glad.cpp"
  HEAP32[166106] = $2552; //@line 5124 "libraries/glad/glad.cpp"
  $2553 = (FUNCTION_TABLE_ii[$load & 8191](487720)|0); //@line 5125 "libraries/glad/glad.cpp"
  HEAP32[166107] = $2553; //@line 5125 "libraries/glad/glad.cpp"
  $2554 = (FUNCTION_TABLE_ii[$load & 8191](487749)|0); //@line 5126 "libraries/glad/glad.cpp"
  HEAP32[166108] = $2554; //@line 5126 "libraries/glad/glad.cpp"
  $2555 = (FUNCTION_TABLE_ii[$load & 8191](487790)|0); //@line 5127 "libraries/glad/glad.cpp"
  HEAP32[166109] = $2555; //@line 5127 "libraries/glad/glad.cpp"
 }
 $2556 = HEAP8[1252188]|0; //@line 5134 "libraries/glad/glad.cpp"
 $2557 = ($2556<<24>>24)==(0); //@line 5134 "libraries/glad/glad.cpp"
 if (!($2557)) {
  $2558 = (FUNCTION_TABLE_ii[$load & 8191](487810)|0); //@line 5135 "libraries/glad/glad.cpp"
  HEAP32[166110] = $2558; //@line 5135 "libraries/glad/glad.cpp"
 }
 $2559 = HEAP8[1252189]|0; //@line 5140 "libraries/glad/glad.cpp"
 $2560 = ($2559<<24>>24)==(0); //@line 5140 "libraries/glad/glad.cpp"
 if (!($2560)) {
  $2561 = (FUNCTION_TABLE_ii[$load & 8191](487834)|0); //@line 5141 "libraries/glad/glad.cpp"
  HEAP32[166111] = $2561; //@line 5141 "libraries/glad/glad.cpp"
 }
 $2562 = HEAP8[1252190]|0; //@line 5146 "libraries/glad/glad.cpp"
 $2563 = ($2562<<24>>24)==(0); //@line 5146 "libraries/glad/glad.cpp"
 if (!($2563)) {
  $2564 = (FUNCTION_TABLE_ii[$load & 8191](487857)|0); //@line 5147 "libraries/glad/glad.cpp"
  HEAP32[166112] = $2564; //@line 5147 "libraries/glad/glad.cpp"
  $2565 = (FUNCTION_TABLE_ii[$load & 8191](487886)|0); //@line 5148 "libraries/glad/glad.cpp"
  HEAP32[166113] = $2565; //@line 5148 "libraries/glad/glad.cpp"
 }
 $2566 = HEAP8[1252191]|0; //@line 5153 "libraries/glad/glad.cpp"
 $2567 = ($2566<<24>>24)==(0); //@line 5153 "libraries/glad/glad.cpp"
 if (!($2567)) {
  $2568 = (FUNCTION_TABLE_ii[$load & 8191](487917)|0); //@line 5154 "libraries/glad/glad.cpp"
  HEAP32[166114] = $2568; //@line 5154 "libraries/glad/glad.cpp"
  $2569 = (FUNCTION_TABLE_ii[$load & 8191](479636)|0); //@line 5155 "libraries/glad/glad.cpp"
  HEAP32[165797] = $2569; //@line 5155 "libraries/glad/glad.cpp"
  $2570 = (FUNCTION_TABLE_ii[$load & 8191](487936)|0); //@line 5156 "libraries/glad/glad.cpp"
  HEAP32[166115] = $2570; //@line 5156 "libraries/glad/glad.cpp"
  $2571 = (FUNCTION_TABLE_ii[$load & 8191](487961)|0); //@line 5157 "libraries/glad/glad.cpp"
  HEAP32[166116] = $2571; //@line 5157 "libraries/glad/glad.cpp"
  $2572 = (FUNCTION_TABLE_ii[$load & 8191](487977)|0); //@line 5158 "libraries/glad/glad.cpp"
  HEAP32[166117] = $2572; //@line 5158 "libraries/glad/glad.cpp"
  $2573 = (FUNCTION_TABLE_ii[$load & 8191](487993)|0); //@line 5159 "libraries/glad/glad.cpp"
  HEAP32[166118] = $2573; //@line 5159 "libraries/glad/glad.cpp"
  $2574 = (FUNCTION_TABLE_ii[$load & 8191](488009)|0); //@line 5160 "libraries/glad/glad.cpp"
  HEAP32[166119] = $2574; //@line 5160 "libraries/glad/glad.cpp"
  $2575 = (FUNCTION_TABLE_ii[$load & 8191](488025)|0); //@line 5161 "libraries/glad/glad.cpp"
  HEAP32[166120] = $2575; //@line 5161 "libraries/glad/glad.cpp"
  $2576 = (FUNCTION_TABLE_ii[$load & 8191](488042)|0); //@line 5162 "libraries/glad/glad.cpp"
  HEAP32[166121] = $2576; //@line 5162 "libraries/glad/glad.cpp"
  $2577 = (FUNCTION_TABLE_ii[$load & 8191](488059)|0); //@line 5163 "libraries/glad/glad.cpp"
  HEAP32[166122] = $2577; //@line 5163 "libraries/glad/glad.cpp"
  $2578 = (FUNCTION_TABLE_ii[$load & 8191](488076)|0); //@line 5164 "libraries/glad/glad.cpp"
  HEAP32[166123] = $2578; //@line 5164 "libraries/glad/glad.cpp"
 }
 $2579 = HEAP8[1252193]|0; //@line 5170 "libraries/glad/glad.cpp"
 $2580 = ($2579<<24>>24)==(0); //@line 5170 "libraries/glad/glad.cpp"
 if (!($2580)) {
  $2581 = (FUNCTION_TABLE_ii[$load & 8191](487274)|0); //@line 5171 "libraries/glad/glad.cpp"
  HEAP32[166089] = $2581; //@line 5171 "libraries/glad/glad.cpp"
  $2582 = (FUNCTION_TABLE_ii[$load & 8191](487299)|0); //@line 5172 "libraries/glad/glad.cpp"
  HEAP32[166090] = $2582; //@line 5172 "libraries/glad/glad.cpp"
  $2583 = (FUNCTION_TABLE_ii[$load & 8191](488093)|0); //@line 5173 "libraries/glad/glad.cpp"
  HEAP32[166124] = $2583; //@line 5173 "libraries/glad/glad.cpp"
 }
 $2584 = HEAP8[1252194]|0; //@line 5178 "libraries/glad/glad.cpp"
 $2585 = ($2584<<24>>24)==(0); //@line 5178 "libraries/glad/glad.cpp"
 if (!($2585)) {
  $2586 = (FUNCTION_TABLE_ii[$load & 8191](488118)|0); //@line 5179 "libraries/glad/glad.cpp"
  HEAP32[166125] = $2586; //@line 5179 "libraries/glad/glad.cpp"
  $2587 = (FUNCTION_TABLE_ii[$load & 8191](488138)|0); //@line 5180 "libraries/glad/glad.cpp"
  HEAP32[166126] = $2587; //@line 5180 "libraries/glad/glad.cpp"
 }
 $2588 = HEAP8[1252195]|0; //@line 5185 "libraries/glad/glad.cpp"
 $2589 = ($2588<<24>>24)==(0); //@line 5185 "libraries/glad/glad.cpp"
 if (!($2589)) {
  $2590 = (FUNCTION_TABLE_ii[$load & 8191](488166)|0); //@line 5186 "libraries/glad/glad.cpp"
  HEAP32[166127] = $2590; //@line 5186 "libraries/glad/glad.cpp"
  $2591 = (FUNCTION_TABLE_ii[$load & 8191](488195)|0); //@line 5187 "libraries/glad/glad.cpp"
  HEAP32[166128] = $2591; //@line 5187 "libraries/glad/glad.cpp"
 }
 $2592 = HEAP8[1252197]|0; //@line 5193 "libraries/glad/glad.cpp"
 $2593 = ($2592<<24>>24)==(0); //@line 5193 "libraries/glad/glad.cpp"
 if (!($2593)) {
  $2594 = (FUNCTION_TABLE_ii[$load & 8191](487347)|0); //@line 5194 "libraries/glad/glad.cpp"
  HEAP32[166092] = $2594; //@line 5194 "libraries/glad/glad.cpp"
  $2595 = (FUNCTION_TABLE_ii[$load & 8191](488226)|0); //@line 5195 "libraries/glad/glad.cpp"
  HEAP32[166129] = $2595; //@line 5195 "libraries/glad/glad.cpp"
 }
 $2596 = HEAP8[1252198]|0; //@line 5200 "libraries/glad/glad.cpp"
 $2597 = ($2596<<24>>24)==(0); //@line 5200 "libraries/glad/glad.cpp"
 if (!($2597)) {
  $2598 = (FUNCTION_TABLE_ii[$load & 8191](488263)|0); //@line 5201 "libraries/glad/glad.cpp"
  HEAP32[166130] = $2598; //@line 5201 "libraries/glad/glad.cpp"
  $2599 = (FUNCTION_TABLE_ii[$load & 8191](488286)|0); //@line 5202 "libraries/glad/glad.cpp"
  HEAP32[166131] = $2599; //@line 5202 "libraries/glad/glad.cpp"
  $2600 = (FUNCTION_TABLE_ii[$load & 8191](488310)|0); //@line 5203 "libraries/glad/glad.cpp"
  HEAP32[166132] = $2600; //@line 5203 "libraries/glad/glad.cpp"
 }
 $2601 = HEAP8[1252199]|0; //@line 5208 "libraries/glad/glad.cpp"
 $2602 = ($2601<<24>>24)==(0); //@line 5208 "libraries/glad/glad.cpp"
 if (!($2602)) {
  $2603 = (FUNCTION_TABLE_ii[$load & 8191](486751)|0); //@line 5209 "libraries/glad/glad.cpp"
  HEAP32[166064] = $2603; //@line 5209 "libraries/glad/glad.cpp"
  $2604 = (FUNCTION_TABLE_ii[$load & 8191](486767)|0); //@line 5210 "libraries/glad/glad.cpp"
  HEAP32[166065] = $2604; //@line 5210 "libraries/glad/glad.cpp"
  $2605 = (FUNCTION_TABLE_ii[$load & 8191](486786)|0); //@line 5211 "libraries/glad/glad.cpp"
  HEAP32[166066] = $2605; //@line 5211 "libraries/glad/glad.cpp"
  $2606 = (FUNCTION_TABLE_ii[$load & 8191](486799)|0); //@line 5212 "libraries/glad/glad.cpp"
  HEAP32[166067] = $2606; //@line 5212 "libraries/glad/glad.cpp"
  $2607 = (FUNCTION_TABLE_ii[$load & 8191](486815)|0); //@line 5213 "libraries/glad/glad.cpp"
  HEAP32[166068] = $2607; //@line 5213 "libraries/glad/glad.cpp"
  $2608 = (FUNCTION_TABLE_ii[$load & 8191](486847)|0); //@line 5214 "libraries/glad/glad.cpp"
  HEAP32[166070] = $2608; //@line 5214 "libraries/glad/glad.cpp"
  $2609 = (FUNCTION_TABLE_ii[$load & 8191](486885)|0); //@line 5215 "libraries/glad/glad.cpp"
  HEAP32[166072] = $2609; //@line 5215 "libraries/glad/glad.cpp"
 }
 $2610 = HEAP8[1252203]|0; //@line 5223 "libraries/glad/glad.cpp"
 $2611 = ($2610<<24>>24)==(0); //@line 5223 "libraries/glad/glad.cpp"
 if (!($2611)) {
  $2612 = (FUNCTION_TABLE_ii[$load & 8191](488329)|0); //@line 5224 "libraries/glad/glad.cpp"
  HEAP32[166133] = $2612; //@line 5224 "libraries/glad/glad.cpp"
 }
 $2613 = HEAP8[1252205]|0; //@line 5230 "libraries/glad/glad.cpp"
 $2614 = ($2613<<24>>24)==(0); //@line 5230 "libraries/glad/glad.cpp"
 if (!($2614)) {
  $2615 = (FUNCTION_TABLE_ii[$load & 8191](488353)|0); //@line 5231 "libraries/glad/glad.cpp"
  HEAP32[166134] = $2615; //@line 5231 "libraries/glad/glad.cpp"
 }
 $2616 = HEAP8[1252206]|0; //@line 5236 "libraries/glad/glad.cpp"
 $2617 = ($2616<<24>>24)==(0); //@line 5236 "libraries/glad/glad.cpp"
 if (!($2617)) {
  $2618 = (FUNCTION_TABLE_ii[$load & 8191](488379)|0); //@line 5237 "libraries/glad/glad.cpp"
  HEAP32[166135] = $2618; //@line 5237 "libraries/glad/glad.cpp"
 }
 $2619 = HEAP8[1252208]|0; //@line 5243 "libraries/glad/glad.cpp"
 $2620 = ($2619<<24>>24)==(0); //@line 5243 "libraries/glad/glad.cpp"
 if (!($2620)) {
  $2621 = (FUNCTION_TABLE_ii[$load & 8191](488400)|0); //@line 5244 "libraries/glad/glad.cpp"
  HEAP32[166136] = $2621; //@line 5244 "libraries/glad/glad.cpp"
 }
 $2622 = HEAP8[1252211]|0; //@line 5251 "libraries/glad/glad.cpp"
 $2623 = ($2622<<24>>24)==(0); //@line 5251 "libraries/glad/glad.cpp"
 if (!($2623)) {
  $2624 = (FUNCTION_TABLE_ii[$load & 8191](488419)|0); //@line 5252 "libraries/glad/glad.cpp"
  HEAP32[166137] = $2624; //@line 5252 "libraries/glad/glad.cpp"
  $2625 = (FUNCTION_TABLE_ii[$load & 8191](488447)|0); //@line 5253 "libraries/glad/glad.cpp"
  HEAP32[166138] = $2625; //@line 5253 "libraries/glad/glad.cpp"
  $2626 = (FUNCTION_TABLE_ii[$load & 8191](488464)|0); //@line 5254 "libraries/glad/glad.cpp"
  HEAP32[166139] = $2626; //@line 5254 "libraries/glad/glad.cpp"
  $2627 = (FUNCTION_TABLE_ii[$load & 8191](488483)|0); //@line 5255 "libraries/glad/glad.cpp"
  HEAP32[166140] = $2627; //@line 5255 "libraries/glad/glad.cpp"
 }
 $2628 = HEAP8[1252212]|0; //@line 5260 "libraries/glad/glad.cpp"
 $2629 = ($2628<<24>>24)==(0); //@line 5260 "libraries/glad/glad.cpp"
 if (!($2629)) {
  $2630 = (FUNCTION_TABLE_ii[$load & 8191](488502)|0); //@line 5261 "libraries/glad/glad.cpp"
  HEAP32[166141] = $2630; //@line 5261 "libraries/glad/glad.cpp"
  $2631 = (FUNCTION_TABLE_ii[$load & 8191](488524)|0); //@line 5262 "libraries/glad/glad.cpp"
  HEAP32[166142] = $2631; //@line 5262 "libraries/glad/glad.cpp"
  $2632 = (FUNCTION_TABLE_ii[$load & 8191](488543)|0); //@line 5263 "libraries/glad/glad.cpp"
  HEAP32[166143] = $2632; //@line 5263 "libraries/glad/glad.cpp"
  $2633 = (FUNCTION_TABLE_ii[$load & 8191](488568)|0); //@line 5264 "libraries/glad/glad.cpp"
  HEAP32[166144] = $2633; //@line 5264 "libraries/glad/glad.cpp"
  $2634 = (FUNCTION_TABLE_ii[$load & 8191](488593)|0); //@line 5265 "libraries/glad/glad.cpp"
  HEAP32[166145] = $2634; //@line 5265 "libraries/glad/glad.cpp"
  $2635 = (FUNCTION_TABLE_ii[$load & 8191](488618)|0); //@line 5266 "libraries/glad/glad.cpp"
  HEAP32[166146] = $2635; //@line 5266 "libraries/glad/glad.cpp"
  $2636 = (FUNCTION_TABLE_ii[$load & 8191](488644)|0); //@line 5267 "libraries/glad/glad.cpp"
  HEAP32[166147] = $2636; //@line 5267 "libraries/glad/glad.cpp"
  $2637 = (FUNCTION_TABLE_ii[$load & 8191](488672)|0); //@line 5268 "libraries/glad/glad.cpp"
  HEAP32[166148] = $2637; //@line 5268 "libraries/glad/glad.cpp"
  $2638 = (FUNCTION_TABLE_ii[$load & 8191](488697)|0); //@line 5269 "libraries/glad/glad.cpp"
  HEAP32[166149] = $2638; //@line 5269 "libraries/glad/glad.cpp"
  $2639 = (FUNCTION_TABLE_ii[$load & 8191](488728)|0); //@line 5270 "libraries/glad/glad.cpp"
  HEAP32[166150] = $2639; //@line 5270 "libraries/glad/glad.cpp"
  $2640 = (FUNCTION_TABLE_ii[$load & 8191](488754)|0); //@line 5271 "libraries/glad/glad.cpp"
  HEAP32[166151] = $2640; //@line 5271 "libraries/glad/glad.cpp"
  $2641 = (FUNCTION_TABLE_ii[$load & 8191](487834)|0); //@line 5272 "libraries/glad/glad.cpp"
  HEAP32[166111] = $2641; //@line 5272 "libraries/glad/glad.cpp"
  $2642 = (FUNCTION_TABLE_ii[$load & 8191](482521)|0); //@line 5273 "libraries/glad/glad.cpp"
  HEAP32[165917] = $2642; //@line 5273 "libraries/glad/glad.cpp"
  $2643 = (FUNCTION_TABLE_ii[$load & 8191](482697)|0); //@line 5274 "libraries/glad/glad.cpp"
  HEAP32[165925] = $2643; //@line 5274 "libraries/glad/glad.cpp"
  $2644 = (FUNCTION_TABLE_ii[$load & 8191](482609)|0); //@line 5275 "libraries/glad/glad.cpp"
  HEAP32[165921] = $2644; //@line 5275 "libraries/glad/glad.cpp"
  $2645 = (FUNCTION_TABLE_ii[$load & 8191](482789)|0); //@line 5276 "libraries/glad/glad.cpp"
  HEAP32[165929] = $2645; //@line 5276 "libraries/glad/glad.cpp"
  $2646 = (FUNCTION_TABLE_ii[$load & 8191](482543)|0); //@line 5277 "libraries/glad/glad.cpp"
  HEAP32[165918] = $2646; //@line 5277 "libraries/glad/glad.cpp"
  $2647 = (FUNCTION_TABLE_ii[$load & 8191](482720)|0); //@line 5278 "libraries/glad/glad.cpp"
  HEAP32[165926] = $2647; //@line 5278 "libraries/glad/glad.cpp"
  $2648 = (FUNCTION_TABLE_ii[$load & 8191](482631)|0); //@line 5279 "libraries/glad/glad.cpp"
  HEAP32[165922] = $2648; //@line 5279 "libraries/glad/glad.cpp"
  $2649 = (FUNCTION_TABLE_ii[$load & 8191](482812)|0); //@line 5280 "libraries/glad/glad.cpp"
  HEAP32[165930] = $2649; //@line 5280 "libraries/glad/glad.cpp"
  $2650 = (FUNCTION_TABLE_ii[$load & 8191](482565)|0); //@line 5281 "libraries/glad/glad.cpp"
  HEAP32[165919] = $2650; //@line 5281 "libraries/glad/glad.cpp"
  $2651 = (FUNCTION_TABLE_ii[$load & 8191](482743)|0); //@line 5282 "libraries/glad/glad.cpp"
  HEAP32[165927] = $2651; //@line 5282 "libraries/glad/glad.cpp"
  $2652 = (FUNCTION_TABLE_ii[$load & 8191](482653)|0); //@line 5283 "libraries/glad/glad.cpp"
  HEAP32[165923] = $2652; //@line 5283 "libraries/glad/glad.cpp"
  $2653 = (FUNCTION_TABLE_ii[$load & 8191](482835)|0); //@line 5284 "libraries/glad/glad.cpp"
  HEAP32[165931] = $2653; //@line 5284 "libraries/glad/glad.cpp"
  $2654 = (FUNCTION_TABLE_ii[$load & 8191](482587)|0); //@line 5285 "libraries/glad/glad.cpp"
  HEAP32[165920] = $2654; //@line 5285 "libraries/glad/glad.cpp"
  $2655 = (FUNCTION_TABLE_ii[$load & 8191](482766)|0); //@line 5286 "libraries/glad/glad.cpp"
  HEAP32[165928] = $2655; //@line 5286 "libraries/glad/glad.cpp"
  $2656 = (FUNCTION_TABLE_ii[$load & 8191](482675)|0); //@line 5287 "libraries/glad/glad.cpp"
  HEAP32[165924] = $2656; //@line 5287 "libraries/glad/glad.cpp"
  $2657 = (FUNCTION_TABLE_ii[$load & 8191](482858)|0); //@line 5288 "libraries/glad/glad.cpp"
  HEAP32[165932] = $2657; //@line 5288 "libraries/glad/glad.cpp"
  $2658 = (FUNCTION_TABLE_ii[$load & 8191](482881)|0); //@line 5289 "libraries/glad/glad.cpp"
  HEAP32[165933] = $2658; //@line 5289 "libraries/glad/glad.cpp"
  $2659 = (FUNCTION_TABLE_ii[$load & 8191](482910)|0); //@line 5290 "libraries/glad/glad.cpp"
  HEAP32[165934] = $2659; //@line 5290 "libraries/glad/glad.cpp"
  $2660 = (FUNCTION_TABLE_ii[$load & 8191](482939)|0); //@line 5291 "libraries/glad/glad.cpp"
  HEAP32[165935] = $2660; //@line 5291 "libraries/glad/glad.cpp"
  $2661 = (FUNCTION_TABLE_ii[$load & 8191](488777)|0); //@line 5292 "libraries/glad/glad.cpp"
  HEAP32[166152] = $2661; //@line 5292 "libraries/glad/glad.cpp"
  $2662 = (FUNCTION_TABLE_ii[$load & 8191](488799)|0); //@line 5293 "libraries/glad/glad.cpp"
  HEAP32[166153] = $2662; //@line 5293 "libraries/glad/glad.cpp"
  $2663 = (FUNCTION_TABLE_ii[$load & 8191](483413)|0); //@line 5294 "libraries/glad/glad.cpp"
  HEAP32[165952] = $2663; //@line 5294 "libraries/glad/glad.cpp"
  $2664 = (FUNCTION_TABLE_ii[$load & 8191](483436)|0); //@line 5295 "libraries/glad/glad.cpp"
  HEAP32[165953] = $2664; //@line 5295 "libraries/glad/glad.cpp"
  $2665 = (FUNCTION_TABLE_ii[$load & 8191](483459)|0); //@line 5296 "libraries/glad/glad.cpp"
  HEAP32[165954] = $2665; //@line 5296 "libraries/glad/glad.cpp"
  $2666 = (FUNCTION_TABLE_ii[$load & 8191](483482)|0); //@line 5297 "libraries/glad/glad.cpp"
  HEAP32[165955] = $2666; //@line 5297 "libraries/glad/glad.cpp"
  $2667 = (FUNCTION_TABLE_ii[$load & 8191](483505)|0); //@line 5298 "libraries/glad/glad.cpp"
  HEAP32[165956] = $2667; //@line 5298 "libraries/glad/glad.cpp"
  $2668 = (FUNCTION_TABLE_ii[$load & 8191](483529)|0); //@line 5299 "libraries/glad/glad.cpp"
  HEAP32[165957] = $2668; //@line 5299 "libraries/glad/glad.cpp"
  $2669 = (FUNCTION_TABLE_ii[$load & 8191](483553)|0); //@line 5300 "libraries/glad/glad.cpp"
  HEAP32[165958] = $2669; //@line 5300 "libraries/glad/glad.cpp"
  $2670 = (FUNCTION_TABLE_ii[$load & 8191](483577)|0); //@line 5301 "libraries/glad/glad.cpp"
  HEAP32[165959] = $2670; //@line 5301 "libraries/glad/glad.cpp"
  $2671 = (FUNCTION_TABLE_ii[$load & 8191](482939)|0); //@line 5302 "libraries/glad/glad.cpp"
  HEAP32[165935] = $2671; //@line 5302 "libraries/glad/glad.cpp"
  $2672 = (FUNCTION_TABLE_ii[$load & 8191](482968)|0); //@line 5303 "libraries/glad/glad.cpp"
  HEAP32[165936] = $2672; //@line 5303 "libraries/glad/glad.cpp"
  $2673 = (FUNCTION_TABLE_ii[$load & 8191](482999)|0); //@line 5304 "libraries/glad/glad.cpp"
  HEAP32[165937] = $2673; //@line 5304 "libraries/glad/glad.cpp"
  $2674 = (FUNCTION_TABLE_ii[$load & 8191](483030)|0); //@line 5305 "libraries/glad/glad.cpp"
  HEAP32[165938] = $2674; //@line 5305 "libraries/glad/glad.cpp"
  $2675 = (FUNCTION_TABLE_ii[$load & 8191](483061)|0); //@line 5306 "libraries/glad/glad.cpp"
  HEAP32[165939] = $2675; //@line 5306 "libraries/glad/glad.cpp"
  $2676 = (FUNCTION_TABLE_ii[$load & 8191](483092)|0); //@line 5307 "libraries/glad/glad.cpp"
  HEAP32[165940] = $2676; //@line 5307 "libraries/glad/glad.cpp"
  $2677 = (FUNCTION_TABLE_ii[$load & 8191](483123)|0); //@line 5308 "libraries/glad/glad.cpp"
  HEAP32[165941] = $2677; //@line 5308 "libraries/glad/glad.cpp"
 }
 $2678 = HEAP8[1252215]|0; //@line 5315 "libraries/glad/glad.cpp"
 $2679 = ($2678<<24>>24)==(0); //@line 5315 "libraries/glad/glad.cpp"
 if (!($2679)) {
  $2680 = (FUNCTION_TABLE_ii[$load & 8191](488828)|0); //@line 5316 "libraries/glad/glad.cpp"
  HEAP32[166154] = $2680; //@line 5316 "libraries/glad/glad.cpp"
  $2681 = (FUNCTION_TABLE_ii[$load & 8191](488850)|0); //@line 5317 "libraries/glad/glad.cpp"
  HEAP32[166155] = $2681; //@line 5317 "libraries/glad/glad.cpp"
 }
 $2682 = HEAP8[1252222]|0; //@line 5328 "libraries/glad/glad.cpp"
 $2683 = ($2682<<24>>24)==(0); //@line 5328 "libraries/glad/glad.cpp"
 if (!($2683)) {
  $2684 = (FUNCTION_TABLE_ii[$load & 8191](488869)|0); //@line 5329 "libraries/glad/glad.cpp"
  HEAP32[166156] = $2684; //@line 5329 "libraries/glad/glad.cpp"
 }
 $2685 = HEAP8[1252226]|0; //@line 5337 "libraries/glad/glad.cpp"
 $2686 = ($2685<<24>>24)==(0); //@line 5337 "libraries/glad/glad.cpp"
 if (!($2686)) {
  $2687 = (FUNCTION_TABLE_ii[$load & 8191](488892)|0); //@line 5338 "libraries/glad/glad.cpp"
  HEAP32[166157] = $2687; //@line 5338 "libraries/glad/glad.cpp"
 }
 $2688 = HEAP8[1252228]|0; //@line 5344 "libraries/glad/glad.cpp"
 $2689 = ($2688<<24>>24)==(0); //@line 5344 "libraries/glad/glad.cpp"
 if (!($2689)) {
  $2690 = (FUNCTION_TABLE_ii[$load & 8191](488913)|0); //@line 5345 "libraries/glad/glad.cpp"
  HEAP32[166158] = $2690; //@line 5345 "libraries/glad/glad.cpp"
 }
 $2691 = HEAP8[1252229]|0; //@line 5350 "libraries/glad/glad.cpp"
 $2692 = ($2691<<24>>24)==(0); //@line 5350 "libraries/glad/glad.cpp"
 if (!($2692)) {
  $2693 = (FUNCTION_TABLE_ii[$load & 8191](488934)|0); //@line 5351 "libraries/glad/glad.cpp"
  HEAP32[166159] = $2693; //@line 5351 "libraries/glad/glad.cpp"
 }
 $2694 = HEAP8[1252230]|0; //@line 5356 "libraries/glad/glad.cpp"
 $2695 = ($2694<<24>>24)==(0); //@line 5356 "libraries/glad/glad.cpp"
 if (!($2695)) {
  $2696 = (FUNCTION_TABLE_ii[$load & 8191](488963)|0); //@line 5357 "libraries/glad/glad.cpp"
  HEAP32[166160] = $2696; //@line 5357 "libraries/glad/glad.cpp"
  $2697 = (FUNCTION_TABLE_ii[$load & 8191](488984)|0); //@line 5358 "libraries/glad/glad.cpp"
  HEAP32[166161] = $2697; //@line 5358 "libraries/glad/glad.cpp"
  $2698 = (FUNCTION_TABLE_ii[$load & 8191](489006)|0); //@line 5359 "libraries/glad/glad.cpp"
  HEAP32[166162] = $2698; //@line 5359 "libraries/glad/glad.cpp"
  $2699 = (FUNCTION_TABLE_ii[$load & 8191](489030)|0); //@line 5360 "libraries/glad/glad.cpp"
  HEAP32[166163] = $2699; //@line 5360 "libraries/glad/glad.cpp"
  $2700 = (FUNCTION_TABLE_ii[$load & 8191](489055)|0); //@line 5361 "libraries/glad/glad.cpp"
  HEAP32[166164] = $2700; //@line 5361 "libraries/glad/glad.cpp"
  $2701 = (FUNCTION_TABLE_ii[$load & 8191](489080)|0); //@line 5362 "libraries/glad/glad.cpp"
  HEAP32[166165] = $2701; //@line 5362 "libraries/glad/glad.cpp"
  $2702 = (FUNCTION_TABLE_ii[$load & 8191](489106)|0); //@line 5363 "libraries/glad/glad.cpp"
  HEAP32[166166] = $2702; //@line 5363 "libraries/glad/glad.cpp"
  $2703 = (FUNCTION_TABLE_ii[$load & 8191](489134)|0); //@line 5364 "libraries/glad/glad.cpp"
  HEAP32[166167] = $2703; //@line 5364 "libraries/glad/glad.cpp"
 }
 $2704 = HEAP8[1252231]|0; //@line 5369 "libraries/glad/glad.cpp"
 $2705 = ($2704<<24>>24)==(0); //@line 5369 "libraries/glad/glad.cpp"
 if (!($2705)) {
  $2706 = (FUNCTION_TABLE_ii[$load & 8191](489163)|0); //@line 5370 "libraries/glad/glad.cpp"
  HEAP32[166168] = $2706; //@line 5370 "libraries/glad/glad.cpp"
  $2707 = (FUNCTION_TABLE_ii[$load & 8191](489178)|0); //@line 5371 "libraries/glad/glad.cpp"
  HEAP32[166169] = $2707; //@line 5371 "libraries/glad/glad.cpp"
 }
 $2708 = HEAP8[1252232]|0; //@line 5376 "libraries/glad/glad.cpp"
 $2709 = ($2708<<24>>24)==(0); //@line 5376 "libraries/glad/glad.cpp"
 if (!($2709)) {
  $2710 = (FUNCTION_TABLE_ii[$load & 8191](489163)|0); //@line 5377 "libraries/glad/glad.cpp"
  HEAP32[166168] = $2710; //@line 5377 "libraries/glad/glad.cpp"
 }
 $2711 = HEAP8[1252240]|0; //@line 5389 "libraries/glad/glad.cpp"
 $2712 = ($2711<<24>>24)==(0); //@line 5389 "libraries/glad/glad.cpp"
 if (!($2712)) {
  $2713 = (FUNCTION_TABLE_ii[$load & 8191](488400)|0); //@line 5390 "libraries/glad/glad.cpp"
  HEAP32[166136] = $2713; //@line 5390 "libraries/glad/glad.cpp"
 }
 $2714 = HEAP8[1252242]|0; //@line 5396 "libraries/glad/glad.cpp"
 $2715 = ($2714<<24>>24)==(0); //@line 5396 "libraries/glad/glad.cpp"
 if (!($2715)) {
  $2716 = (FUNCTION_TABLE_ii[$load & 8191](488963)|0); //@line 5397 "libraries/glad/glad.cpp"
  HEAP32[166160] = $2716; //@line 5397 "libraries/glad/glad.cpp"
  $2717 = (FUNCTION_TABLE_ii[$load & 8191](488984)|0); //@line 5398 "libraries/glad/glad.cpp"
  HEAP32[166161] = $2717; //@line 5398 "libraries/glad/glad.cpp"
  $2718 = (FUNCTION_TABLE_ii[$load & 8191](489006)|0); //@line 5399 "libraries/glad/glad.cpp"
  HEAP32[166162] = $2718; //@line 5399 "libraries/glad/glad.cpp"
  $2719 = (FUNCTION_TABLE_ii[$load & 8191](489030)|0); //@line 5400 "libraries/glad/glad.cpp"
  HEAP32[166163] = $2719; //@line 5400 "libraries/glad/glad.cpp"
  $2720 = (FUNCTION_TABLE_ii[$load & 8191](489198)|0); //@line 5401 "libraries/glad/glad.cpp"
  HEAP32[166170] = $2720; //@line 5401 "libraries/glad/glad.cpp"
  $2721 = (FUNCTION_TABLE_ii[$load & 8191](489216)|0); //@line 5402 "libraries/glad/glad.cpp"
  HEAP32[166171] = $2721; //@line 5402 "libraries/glad/glad.cpp"
 }
 $2722 = HEAP8[1252251]|0; //@line 5415 "libraries/glad/glad.cpp"
 $2723 = ($2722<<24>>24)==(0); //@line 5415 "libraries/glad/glad.cpp"
 if (!($2723)) {
  $2724 = (FUNCTION_TABLE_ii[$load & 8191](489235)|0); //@line 5416 "libraries/glad/glad.cpp"
  HEAP32[166172] = $2724; //@line 5416 "libraries/glad/glad.cpp"
  $2725 = (FUNCTION_TABLE_ii[$load & 8191](489253)|0); //@line 5417 "libraries/glad/glad.cpp"
  HEAP32[166173] = $2725; //@line 5417 "libraries/glad/glad.cpp"
  $2726 = (FUNCTION_TABLE_ii[$load & 8191](489271)|0); //@line 5418 "libraries/glad/glad.cpp"
  HEAP32[166174] = $2726; //@line 5418 "libraries/glad/glad.cpp"
  $2727 = (FUNCTION_TABLE_ii[$load & 8191](486283)|0); //@line 5419 "libraries/glad/glad.cpp"
  HEAP32[166049] = $2727; //@line 5419 "libraries/glad/glad.cpp"
  $2728 = (FUNCTION_TABLE_ii[$load & 8191](486305)|0); //@line 5420 "libraries/glad/glad.cpp"
  HEAP32[166050] = $2728; //@line 5420 "libraries/glad/glad.cpp"
  $2729 = (FUNCTION_TABLE_ii[$load & 8191](486327)|0); //@line 5421 "libraries/glad/glad.cpp"
  HEAP32[166051] = $2729; //@line 5421 "libraries/glad/glad.cpp"
 }
 $2730 = HEAP8[1252254]|0; //@line 5428 "libraries/glad/glad.cpp"
 $2731 = ($2730<<24>>24)==(0); //@line 5428 "libraries/glad/glad.cpp"
 if (!($2731)) {
  $2732 = (FUNCTION_TABLE_ii[$load & 8191](489289)|0); //@line 5429 "libraries/glad/glad.cpp"
  HEAP32[166175] = $2732; //@line 5429 "libraries/glad/glad.cpp"
 }
 $2733 = HEAP8[1252255]|0; //@line 5434 "libraries/glad/glad.cpp"
 $2734 = ($2733<<24>>24)==(0); //@line 5434 "libraries/glad/glad.cpp"
 if (!($2734)) {
  $2735 = (FUNCTION_TABLE_ii[$load & 8191](486908)|0); //@line 5435 "libraries/glad/glad.cpp"
  HEAP32[166073] = $2735; //@line 5435 "libraries/glad/glad.cpp"
  $2736 = (FUNCTION_TABLE_ii[$load & 8191](486932)|0); //@line 5436 "libraries/glad/glad.cpp"
  HEAP32[166074] = $2736; //@line 5436 "libraries/glad/glad.cpp"
 }
 $2737 = HEAP8[1252256]|0; //@line 5441 "libraries/glad/glad.cpp"
 $2738 = ($2737<<24>>24)==(0); //@line 5441 "libraries/glad/glad.cpp"
 if (!($2738)) {
  $2739 = (FUNCTION_TABLE_ii[$load & 8191](489306)|0); //@line 5442 "libraries/glad/glad.cpp"
  HEAP32[166176] = $2739; //@line 5442 "libraries/glad/glad.cpp"
  $2740 = (FUNCTION_TABLE_ii[$load & 8191](489334)|0); //@line 5443 "libraries/glad/glad.cpp"
  HEAP32[166177] = $2740; //@line 5443 "libraries/glad/glad.cpp"
  $2741 = (FUNCTION_TABLE_ii[$load & 8191](489360)|0); //@line 5444 "libraries/glad/glad.cpp"
  HEAP32[166178] = $2741; //@line 5444 "libraries/glad/glad.cpp"
  $2742 = (FUNCTION_TABLE_ii[$load & 8191](489381)|0); //@line 5445 "libraries/glad/glad.cpp"
  HEAP32[166179] = $2742; //@line 5445 "libraries/glad/glad.cpp"
  $2743 = (FUNCTION_TABLE_ii[$load & 8191](489403)|0); //@line 5446 "libraries/glad/glad.cpp"
  HEAP32[166180] = $2743; //@line 5446 "libraries/glad/glad.cpp"
  $2744 = (FUNCTION_TABLE_ii[$load & 8191](489423)|0); //@line 5447 "libraries/glad/glad.cpp"
  HEAP32[166181] = $2744; //@line 5447 "libraries/glad/glad.cpp"
  $2745 = (FUNCTION_TABLE_ii[$load & 8191](489454)|0); //@line 5448 "libraries/glad/glad.cpp"
  HEAP32[166182] = $2745; //@line 5448 "libraries/glad/glad.cpp"
 }
 $2746 = HEAP8[1252259]|0; //@line 5455 "libraries/glad/glad.cpp"
 $2747 = ($2746<<24>>24)==(0); //@line 5455 "libraries/glad/glad.cpp"
 if (!($2747)) {
  $2748 = (FUNCTION_TABLE_ii[$load & 8191](489487)|0); //@line 5456 "libraries/glad/glad.cpp"
  HEAP32[166183] = $2748; //@line 5456 "libraries/glad/glad.cpp"
  $2749 = (FUNCTION_TABLE_ii[$load & 8191](489508)|0); //@line 5457 "libraries/glad/glad.cpp"
  HEAP32[166184] = $2749; //@line 5457 "libraries/glad/glad.cpp"
  $2750 = (FUNCTION_TABLE_ii[$load & 8191](489529)|0); //@line 5458 "libraries/glad/glad.cpp"
  HEAP32[166185] = $2750; //@line 5458 "libraries/glad/glad.cpp"
  $2751 = (FUNCTION_TABLE_ii[$load & 8191](489550)|0); //@line 5459 "libraries/glad/glad.cpp"
  HEAP32[166186] = $2751; //@line 5459 "libraries/glad/glad.cpp"
  $2752 = (FUNCTION_TABLE_ii[$load & 8191](489571)|0); //@line 5460 "libraries/glad/glad.cpp"
  HEAP32[166187] = $2752; //@line 5460 "libraries/glad/glad.cpp"
  $2753 = (FUNCTION_TABLE_ii[$load & 8191](489593)|0); //@line 5461 "libraries/glad/glad.cpp"
  HEAP32[166188] = $2753; //@line 5461 "libraries/glad/glad.cpp"
  $2754 = (FUNCTION_TABLE_ii[$load & 8191](489615)|0); //@line 5462 "libraries/glad/glad.cpp"
  HEAP32[166189] = $2754; //@line 5462 "libraries/glad/glad.cpp"
  $2755 = (FUNCTION_TABLE_ii[$load & 8191](489637)|0); //@line 5463 "libraries/glad/glad.cpp"
  HEAP32[166190] = $2755; //@line 5463 "libraries/glad/glad.cpp"
  $2756 = (FUNCTION_TABLE_ii[$load & 8191](489659)|0); //@line 5464 "libraries/glad/glad.cpp"
  HEAP32[166191] = $2756; //@line 5464 "libraries/glad/glad.cpp"
  $2757 = (FUNCTION_TABLE_ii[$load & 8191](489685)|0); //@line 5465 "libraries/glad/glad.cpp"
  HEAP32[166192] = $2757; //@line 5465 "libraries/glad/glad.cpp"
 }
 $2758 = HEAP8[1252260]|0; //@line 5470 "libraries/glad/glad.cpp"
 $2759 = ($2758<<24>>24)==(0); //@line 5470 "libraries/glad/glad.cpp"
 if (!($2759)) {
  $2760 = (FUNCTION_TABLE_ii[$load & 8191](489709)|0); //@line 5471 "libraries/glad/glad.cpp"
  HEAP32[166193] = $2760; //@line 5471 "libraries/glad/glad.cpp"
 }
 $2761 = HEAP8[1252264]|0; //@line 5479 "libraries/glad/glad.cpp"
 $2762 = ($2761<<24>>24)==(0); //@line 5479 "libraries/glad/glad.cpp"
 if (!($2762)) {
  $2763 = (FUNCTION_TABLE_ii[$load & 8191](489725)|0); //@line 5480 "libraries/glad/glad.cpp"
  HEAP32[166194] = $2763; //@line 5480 "libraries/glad/glad.cpp"
 }
 $2764 = HEAP8[1252268]|0; //@line 5488 "libraries/glad/glad.cpp"
 $2765 = ($2764<<24>>24)==(0); //@line 5488 "libraries/glad/glad.cpp"
 if (!($2765)) {
  $2766 = (FUNCTION_TABLE_ii[$load & 8191](489747)|0); //@line 5489 "libraries/glad/glad.cpp"
  HEAP32[166195] = $2766; //@line 5489 "libraries/glad/glad.cpp"
  $2767 = (FUNCTION_TABLE_ii[$load & 8191](489760)|0); //@line 5490 "libraries/glad/glad.cpp"
  HEAP32[166196] = $2767; //@line 5490 "libraries/glad/glad.cpp"
  $2768 = (FUNCTION_TABLE_ii[$load & 8191](489774)|0); //@line 5491 "libraries/glad/glad.cpp"
  HEAP32[166197] = $2768; //@line 5491 "libraries/glad/glad.cpp"
  $2769 = (FUNCTION_TABLE_ii[$load & 8191](489794)|0); //@line 5492 "libraries/glad/glad.cpp"
  HEAP32[166198] = $2769; //@line 5492 "libraries/glad/glad.cpp"
  $2770 = (FUNCTION_TABLE_ii[$load & 8191](489822)|0); //@line 5493 "libraries/glad/glad.cpp"
  HEAP32[166199] = $2770; //@line 5493 "libraries/glad/glad.cpp"
  $2771 = (FUNCTION_TABLE_ii[$load & 8191](489838)|0); //@line 5494 "libraries/glad/glad.cpp"
  HEAP32[166200] = $2771; //@line 5494 "libraries/glad/glad.cpp"
  $2772 = (FUNCTION_TABLE_ii[$load & 8191](489862)|0); //@line 5495 "libraries/glad/glad.cpp"
  HEAP32[166201] = $2772; //@line 5495 "libraries/glad/glad.cpp"
  $2773 = (FUNCTION_TABLE_ii[$load & 8191](489878)|0); //@line 5496 "libraries/glad/glad.cpp"
  HEAP32[166202] = $2773; //@line 5496 "libraries/glad/glad.cpp"
 }
 $2774 = HEAP8[1252269]|0; //@line 5501 "libraries/glad/glad.cpp"
 $2775 = ($2774<<24>>24)==(0); //@line 5501 "libraries/glad/glad.cpp"
 if (!($2775)) {
  $2776 = (FUNCTION_TABLE_ii[$load & 8191](489894)|0); //@line 5502 "libraries/glad/glad.cpp"
  HEAP32[166203] = $2776; //@line 5502 "libraries/glad/glad.cpp"
  $2777 = (FUNCTION_TABLE_ii[$load & 8191](489922)|0); //@line 5503 "libraries/glad/glad.cpp"
  HEAP32[166204] = $2777; //@line 5503 "libraries/glad/glad.cpp"
  $2778 = (FUNCTION_TABLE_ii[$load & 8191](489955)|0); //@line 5504 "libraries/glad/glad.cpp"
  HEAP32[166205] = $2778; //@line 5504 "libraries/glad/glad.cpp"
  $2779 = (FUNCTION_TABLE_ii[$load & 8191](489992)|0); //@line 5505 "libraries/glad/glad.cpp"
  HEAP32[166206] = $2779; //@line 5505 "libraries/glad/glad.cpp"
 }
 $2780 = HEAP8[1252270]|0; //@line 5510 "libraries/glad/glad.cpp"
 $2781 = ($2780<<24>>24)==(0); //@line 5510 "libraries/glad/glad.cpp"
 if (!($2781)) {
  $2782 = (FUNCTION_TABLE_ii[$load & 8191](490025)|0); //@line 5511 "libraries/glad/glad.cpp"
  HEAP32[166207] = $2782; //@line 5511 "libraries/glad/glad.cpp"
  $2783 = (FUNCTION_TABLE_ii[$load & 8191](490054)|0); //@line 5512 "libraries/glad/glad.cpp"
  HEAP32[166208] = $2783; //@line 5512 "libraries/glad/glad.cpp"
 }
 $2784 = HEAP8[1252277]|0; //@line 5523 "libraries/glad/glad.cpp"
 $2785 = ($2784<<24>>24)==(0); //@line 5523 "libraries/glad/glad.cpp"
 if (!($2785)) {
  $2786 = (FUNCTION_TABLE_ii[$load & 8191](490093)|0); //@line 5524 "libraries/glad/glad.cpp"
  HEAP32[166209] = $2786; //@line 5524 "libraries/glad/glad.cpp"
 }
 $2787 = HEAP8[1252278]|0; //@line 5529 "libraries/glad/glad.cpp"
 $2788 = ($2787<<24>>24)==(0); //@line 5529 "libraries/glad/glad.cpp"
 if (!($2788)) {
  $2789 = (FUNCTION_TABLE_ii[$load & 8191](490117)|0); //@line 5530 "libraries/glad/glad.cpp"
  HEAP32[166210] = $2789; //@line 5530 "libraries/glad/glad.cpp"
  $2790 = (FUNCTION_TABLE_ii[$load & 8191](490139)|0); //@line 5531 "libraries/glad/glad.cpp"
  HEAP32[166211] = $2790; //@line 5531 "libraries/glad/glad.cpp"
 }
 $2791 = HEAP8[1252280]|0; //@line 5537 "libraries/glad/glad.cpp"
 $2792 = ($2791<<24>>24)==(0); //@line 5537 "libraries/glad/glad.cpp"
 if (!($2792)) {
  $2793 = (FUNCTION_TABLE_ii[$load & 8191](490158)|0); //@line 5538 "libraries/glad/glad.cpp"
  HEAP32[166212] = $2793; //@line 5538 "libraries/glad/glad.cpp"
  $2794 = (FUNCTION_TABLE_ii[$load & 8191](490173)|0); //@line 5539 "libraries/glad/glad.cpp"
  HEAP32[166213] = $2794; //@line 5539 "libraries/glad/glad.cpp"
  $2795 = (FUNCTION_TABLE_ii[$load & 8191](490190)|0); //@line 5540 "libraries/glad/glad.cpp"
  HEAP32[166214] = $2795; //@line 5540 "libraries/glad/glad.cpp"
 }
 $2796 = HEAP8[1252282]|0; //@line 5546 "libraries/glad/glad.cpp"
 $2797 = ($2796<<24>>24)==(0); //@line 5546 "libraries/glad/glad.cpp"
 if (!($2797)) {
  $2798 = (FUNCTION_TABLE_ii[$load & 8191](490213)|0); //@line 5547 "libraries/glad/glad.cpp"
  HEAP32[166215] = $2798; //@line 5547 "libraries/glad/glad.cpp"
 }
 $2799 = HEAP8[1252285]|0; //@line 5554 "libraries/glad/glad.cpp"
 $2800 = ($2799<<24>>24)==(0); //@line 5554 "libraries/glad/glad.cpp"
 if (!($2800)) {
  $2801 = (FUNCTION_TABLE_ii[$load & 8191](490239)|0); //@line 5555 "libraries/glad/glad.cpp"
  HEAP32[166216] = $2801; //@line 5555 "libraries/glad/glad.cpp"
 }
 $2802 = HEAP8[1252295]|0; //@line 5569 "libraries/glad/glad.cpp"
 $2803 = ($2802<<24>>24)==(0); //@line 5569 "libraries/glad/glad.cpp"
 if (!($2803)) {
  $2804 = (FUNCTION_TABLE_ii[$load & 8191](490261)|0); //@line 5570 "libraries/glad/glad.cpp"
  HEAP32[166217] = $2804; //@line 5570 "libraries/glad/glad.cpp"
 }
 $2805 = HEAP8[1252296]|0; //@line 5575 "libraries/glad/glad.cpp"
 $2806 = ($2805<<24>>24)==(0); //@line 5575 "libraries/glad/glad.cpp"
 if (!($2806)) {
  $2807 = (FUNCTION_TABLE_ii[$load & 8191](490282)|0); //@line 5576 "libraries/glad/glad.cpp"
  HEAP32[166218] = $2807; //@line 5576 "libraries/glad/glad.cpp"
  $2808 = (FUNCTION_TABLE_ii[$load & 8191](490298)|0); //@line 5577 "libraries/glad/glad.cpp"
  HEAP32[166219] = $2808; //@line 5577 "libraries/glad/glad.cpp"
  $2809 = (FUNCTION_TABLE_ii[$load & 8191](490317)|0); //@line 5578 "libraries/glad/glad.cpp"
  HEAP32[166220] = $2809; //@line 5578 "libraries/glad/glad.cpp"
  $2810 = (FUNCTION_TABLE_ii[$load & 8191](490340)|0); //@line 5579 "libraries/glad/glad.cpp"
  HEAP32[166221] = $2810; //@line 5579 "libraries/glad/glad.cpp"
  $2811 = (FUNCTION_TABLE_ii[$load & 8191](490366)|0); //@line 5580 "libraries/glad/glad.cpp"
  HEAP32[166222] = $2811; //@line 5580 "libraries/glad/glad.cpp"
  $2812 = (FUNCTION_TABLE_ii[$load & 8191](490395)|0); //@line 5581 "libraries/glad/glad.cpp"
  HEAP32[166223] = $2812; //@line 5581 "libraries/glad/glad.cpp"
 }
 $2813 = HEAP8[1252297]|0; //@line 5586 "libraries/glad/glad.cpp"
 $2814 = ($2813<<24>>24)==(0); //@line 5586 "libraries/glad/glad.cpp"
 if (!($2814)) {
  $2815 = (FUNCTION_TABLE_ii[$load & 8191](490421)|0); //@line 5587 "libraries/glad/glad.cpp"
  HEAP32[166224] = $2815; //@line 5587 "libraries/glad/glad.cpp"
  $2816 = (FUNCTION_TABLE_ii[$load & 8191](490442)|0); //@line 5588 "libraries/glad/glad.cpp"
  HEAP32[166225] = $2816; //@line 5588 "libraries/glad/glad.cpp"
  $2817 = (FUNCTION_TABLE_ii[$load & 8191](490464)|0); //@line 5589 "libraries/glad/glad.cpp"
  HEAP32[166226] = $2817; //@line 5589 "libraries/glad/glad.cpp"
  $2818 = (FUNCTION_TABLE_ii[$load & 8191](490488)|0); //@line 5590 "libraries/glad/glad.cpp"
  HEAP32[166227] = $2818; //@line 5590 "libraries/glad/glad.cpp"
  $2819 = (FUNCTION_TABLE_ii[$load & 8191](490513)|0); //@line 5591 "libraries/glad/glad.cpp"
  HEAP32[166228] = $2819; //@line 5591 "libraries/glad/glad.cpp"
  $2820 = (FUNCTION_TABLE_ii[$load & 8191](490538)|0); //@line 5592 "libraries/glad/glad.cpp"
  HEAP32[166229] = $2820; //@line 5592 "libraries/glad/glad.cpp"
  $2821 = (FUNCTION_TABLE_ii[$load & 8191](490564)|0); //@line 5593 "libraries/glad/glad.cpp"
  HEAP32[166230] = $2821; //@line 5593 "libraries/glad/glad.cpp"
  $2822 = (FUNCTION_TABLE_ii[$load & 8191](490592)|0); //@line 5594 "libraries/glad/glad.cpp"
  HEAP32[166231] = $2822; //@line 5594 "libraries/glad/glad.cpp"
 }
 $2823 = HEAP8[1252298]|0; //@line 5599 "libraries/glad/glad.cpp"
 $2824 = ($2823<<24>>24)==(0); //@line 5599 "libraries/glad/glad.cpp"
 if (!($2824)) {
  $2825 = (FUNCTION_TABLE_ii[$load & 8191](490621)|0); //@line 5600 "libraries/glad/glad.cpp"
  HEAP32[166232] = $2825; //@line 5600 "libraries/glad/glad.cpp"
  $2826 = (FUNCTION_TABLE_ii[$load & 8191](490636)|0); //@line 5601 "libraries/glad/glad.cpp"
  HEAP32[166233] = $2826; //@line 5601 "libraries/glad/glad.cpp"
 }
 $2827 = HEAP8[1252307]|0; //@line 5614 "libraries/glad/glad.cpp"
 $2828 = ($2827<<24>>24)==(0); //@line 5614 "libraries/glad/glad.cpp"
 if (!($2828)) {
  $2829 = (FUNCTION_TABLE_ii[$load & 8191](490656)|0); //@line 5615 "libraries/glad/glad.cpp"
  HEAP32[166234] = $2829; //@line 5615 "libraries/glad/glad.cpp"
 }
 $2830 = HEAP8[1252308]|0; //@line 5620 "libraries/glad/glad.cpp"
 $2831 = ($2830<<24>>24)==(0); //@line 5620 "libraries/glad/glad.cpp"
 if (!($2831)) {
  $2832 = (FUNCTION_TABLE_ii[$load & 8191](490685)|0); //@line 5621 "libraries/glad/glad.cpp"
  HEAP32[166235] = $2832; //@line 5621 "libraries/glad/glad.cpp"
 }
 $2833 = HEAP8[1252309]|0; //@line 5626 "libraries/glad/glad.cpp"
 $2834 = ($2833<<24>>24)==(0); //@line 5626 "libraries/glad/glad.cpp"
 if (!($2834)) {
  $2835 = (FUNCTION_TABLE_ii[$load & 8191](490702)|0); //@line 5627 "libraries/glad/glad.cpp"
  HEAP32[166236] = $2835; //@line 5627 "libraries/glad/glad.cpp"
  $2836 = (FUNCTION_TABLE_ii[$load & 8191](490723)|0); //@line 5628 "libraries/glad/glad.cpp"
  HEAP32[166237] = $2836; //@line 5628 "libraries/glad/glad.cpp"
  $2837 = (FUNCTION_TABLE_ii[$load & 8191](490747)|0); //@line 5629 "libraries/glad/glad.cpp"
  HEAP32[166238] = $2837; //@line 5629 "libraries/glad/glad.cpp"
  $2838 = (FUNCTION_TABLE_ii[$load & 8191](490768)|0); //@line 5630 "libraries/glad/glad.cpp"
  HEAP32[166239] = $2838; //@line 5630 "libraries/glad/glad.cpp"
 }
 $2839 = HEAP8[1252316]|0; //@line 5641 "libraries/glad/glad.cpp"
 $2840 = ($2839<<24>>24)==(0); //@line 5641 "libraries/glad/glad.cpp"
 if (!($2840)) {
  $2841 = (FUNCTION_TABLE_ii[$load & 8191](490787)|0); //@line 5642 "libraries/glad/glad.cpp"
  HEAP32[166240] = $2841; //@line 5642 "libraries/glad/glad.cpp"
  $2842 = (FUNCTION_TABLE_ii[$load & 8191](490811)|0); //@line 5643 "libraries/glad/glad.cpp"
  HEAP32[166241] = $2842; //@line 5643 "libraries/glad/glad.cpp"
  $2843 = (FUNCTION_TABLE_ii[$load & 8191](490835)|0); //@line 5644 "libraries/glad/glad.cpp"
  HEAP32[166242] = $2843; //@line 5644 "libraries/glad/glad.cpp"
  $2844 = (FUNCTION_TABLE_ii[$load & 8191](490861)|0); //@line 5645 "libraries/glad/glad.cpp"
  HEAP32[166243] = $2844; //@line 5645 "libraries/glad/glad.cpp"
 }
 $2845 = HEAP8[1252318]|0; //@line 5651 "libraries/glad/glad.cpp"
 $2846 = ($2845<<24>>24)==(0); //@line 5651 "libraries/glad/glad.cpp"
 if (!($2846)) {
  $2847 = (FUNCTION_TABLE_ii[$load & 8191](490885)|0); //@line 5652 "libraries/glad/glad.cpp"
  HEAP32[166244] = $2847; //@line 5652 "libraries/glad/glad.cpp"
  $2848 = (FUNCTION_TABLE_ii[$load & 8191](490907)|0); //@line 5653 "libraries/glad/glad.cpp"
  HEAP32[166245] = $2848; //@line 5653 "libraries/glad/glad.cpp"
  $2849 = (FUNCTION_TABLE_ii[$load & 8191](490937)|0); //@line 5654 "libraries/glad/glad.cpp"
  HEAP32[166246] = $2849; //@line 5654 "libraries/glad/glad.cpp"
  $2850 = (FUNCTION_TABLE_ii[$load & 8191](490963)|0); //@line 5655 "libraries/glad/glad.cpp"
  HEAP32[166247] = $2850; //@line 5655 "libraries/glad/glad.cpp"
 }
 $2851 = HEAP8[1252320]|0; //@line 5661 "libraries/glad/glad.cpp"
 $2852 = ($2851<<24>>24)==(0); //@line 5661 "libraries/glad/glad.cpp"
 if (!($2852)) {
  $2853 = (FUNCTION_TABLE_ii[$load & 8191](490997)|0); //@line 5662 "libraries/glad/glad.cpp"
  HEAP32[166248] = $2853; //@line 5662 "libraries/glad/glad.cpp"
  $2854 = (FUNCTION_TABLE_ii[$load & 8191](491013)|0); //@line 5663 "libraries/glad/glad.cpp"
  HEAP32[166249] = $2854; //@line 5663 "libraries/glad/glad.cpp"
  $2855 = (FUNCTION_TABLE_ii[$load & 8191](491029)|0); //@line 5664 "libraries/glad/glad.cpp"
  HEAP32[166250] = $2855; //@line 5664 "libraries/glad/glad.cpp"
  $2856 = (FUNCTION_TABLE_ii[$load & 8191](491045)|0); //@line 5665 "libraries/glad/glad.cpp"
  HEAP32[166251] = $2856; //@line 5665 "libraries/glad/glad.cpp"
  $2857 = (FUNCTION_TABLE_ii[$load & 8191](491061)|0); //@line 5666 "libraries/glad/glad.cpp"
  HEAP32[166252] = $2857; //@line 5666 "libraries/glad/glad.cpp"
  $2858 = (FUNCTION_TABLE_ii[$load & 8191](491078)|0); //@line 5667 "libraries/glad/glad.cpp"
  HEAP32[166253] = $2858; //@line 5667 "libraries/glad/glad.cpp"
  $2859 = (FUNCTION_TABLE_ii[$load & 8191](491095)|0); //@line 5668 "libraries/glad/glad.cpp"
  HEAP32[166254] = $2859; //@line 5668 "libraries/glad/glad.cpp"
  $2860 = (FUNCTION_TABLE_ii[$load & 8191](491112)|0); //@line 5669 "libraries/glad/glad.cpp"
  HEAP32[166255] = $2860; //@line 5669 "libraries/glad/glad.cpp"
  $2861 = (FUNCTION_TABLE_ii[$load & 8191](491129)|0); //@line 5670 "libraries/glad/glad.cpp"
  HEAP32[166256] = $2861; //@line 5670 "libraries/glad/glad.cpp"
  $2862 = (FUNCTION_TABLE_ii[$load & 8191](491146)|0); //@line 5671 "libraries/glad/glad.cpp"
  HEAP32[166257] = $2862; //@line 5671 "libraries/glad/glad.cpp"
  $2863 = (FUNCTION_TABLE_ii[$load & 8191](491163)|0); //@line 5672 "libraries/glad/glad.cpp"
  HEAP32[166258] = $2863; //@line 5672 "libraries/glad/glad.cpp"
  $2864 = (FUNCTION_TABLE_ii[$load & 8191](491180)|0); //@line 5673 "libraries/glad/glad.cpp"
  HEAP32[166259] = $2864; //@line 5673 "libraries/glad/glad.cpp"
  $2865 = (FUNCTION_TABLE_ii[$load & 8191](491197)|0); //@line 5674 "libraries/glad/glad.cpp"
  HEAP32[166260] = $2865; //@line 5674 "libraries/glad/glad.cpp"
  $2866 = (FUNCTION_TABLE_ii[$load & 8191](491215)|0); //@line 5675 "libraries/glad/glad.cpp"
  HEAP32[166261] = $2866; //@line 5675 "libraries/glad/glad.cpp"
  $2867 = (FUNCTION_TABLE_ii[$load & 8191](491233)|0); //@line 5676 "libraries/glad/glad.cpp"
  HEAP32[166262] = $2867; //@line 5676 "libraries/glad/glad.cpp"
  $2868 = (FUNCTION_TABLE_ii[$load & 8191](491251)|0); //@line 5677 "libraries/glad/glad.cpp"
  HEAP32[166263] = $2868; //@line 5677 "libraries/glad/glad.cpp"
  $2869 = (FUNCTION_TABLE_ii[$load & 8191](491269)|0); //@line 5678 "libraries/glad/glad.cpp"
  HEAP32[166264] = $2869; //@line 5678 "libraries/glad/glad.cpp"
  $2870 = (FUNCTION_TABLE_ii[$load & 8191](491288)|0); //@line 5679 "libraries/glad/glad.cpp"
  HEAP32[166265] = $2870; //@line 5679 "libraries/glad/glad.cpp"
  $2871 = (FUNCTION_TABLE_ii[$load & 8191](491308)|0); //@line 5680 "libraries/glad/glad.cpp"
  HEAP32[166266] = $2871; //@line 5680 "libraries/glad/glad.cpp"
  $2872 = (FUNCTION_TABLE_ii[$load & 8191](491331)|0); //@line 5681 "libraries/glad/glad.cpp"
  HEAP32[166267] = $2872; //@line 5681 "libraries/glad/glad.cpp"
  $2873 = (FUNCTION_TABLE_ii[$load & 8191](491354)|0); //@line 5682 "libraries/glad/glad.cpp"
  HEAP32[166268] = $2873; //@line 5682 "libraries/glad/glad.cpp"
  $2874 = (FUNCTION_TABLE_ii[$load & 8191](491377)|0); //@line 5683 "libraries/glad/glad.cpp"
  HEAP32[166269] = $2874; //@line 5683 "libraries/glad/glad.cpp"
  $2875 = (FUNCTION_TABLE_ii[$load & 8191](491400)|0); //@line 5684 "libraries/glad/glad.cpp"
  HEAP32[166270] = $2875; //@line 5684 "libraries/glad/glad.cpp"
  $2876 = (FUNCTION_TABLE_ii[$load & 8191](491424)|0); //@line 5685 "libraries/glad/glad.cpp"
  HEAP32[166271] = $2876; //@line 5685 "libraries/glad/glad.cpp"
  $2877 = (FUNCTION_TABLE_ii[$load & 8191](491448)|0); //@line 5686 "libraries/glad/glad.cpp"
  HEAP32[166272] = $2877; //@line 5686 "libraries/glad/glad.cpp"
  $2878 = (FUNCTION_TABLE_ii[$load & 8191](491472)|0); //@line 5687 "libraries/glad/glad.cpp"
  HEAP32[166273] = $2878; //@line 5687 "libraries/glad/glad.cpp"
  $2879 = (FUNCTION_TABLE_ii[$load & 8191](491496)|0); //@line 5688 "libraries/glad/glad.cpp"
  HEAP32[166274] = $2879; //@line 5688 "libraries/glad/glad.cpp"
  $2880 = (FUNCTION_TABLE_ii[$load & 8191](491520)|0); //@line 5689 "libraries/glad/glad.cpp"
  HEAP32[166275] = $2880; //@line 5689 "libraries/glad/glad.cpp"
  $2881 = (FUNCTION_TABLE_ii[$load & 8191](491544)|0); //@line 5690 "libraries/glad/glad.cpp"
  HEAP32[166276] = $2881; //@line 5690 "libraries/glad/glad.cpp"
  $2882 = (FUNCTION_TABLE_ii[$load & 8191](491568)|0); //@line 5691 "libraries/glad/glad.cpp"
  HEAP32[166277] = $2882; //@line 5691 "libraries/glad/glad.cpp"
  $2883 = (FUNCTION_TABLE_ii[$load & 8191](491592)|0); //@line 5692 "libraries/glad/glad.cpp"
  HEAP32[166278] = $2883; //@line 5692 "libraries/glad/glad.cpp"
  $2884 = (FUNCTION_TABLE_ii[$load & 8191](491617)|0); //@line 5693 "libraries/glad/glad.cpp"
  HEAP32[166279] = $2884; //@line 5693 "libraries/glad/glad.cpp"
  $2885 = (FUNCTION_TABLE_ii[$load & 8191](491642)|0); //@line 5694 "libraries/glad/glad.cpp"
  HEAP32[166280] = $2885; //@line 5694 "libraries/glad/glad.cpp"
  $2886 = (FUNCTION_TABLE_ii[$load & 8191](491667)|0); //@line 5695 "libraries/glad/glad.cpp"
  HEAP32[166281] = $2886; //@line 5695 "libraries/glad/glad.cpp"
 }
 $2887 = HEAP8[1252321]|0; //@line 5700 "libraries/glad/glad.cpp"
 $2888 = ($2887<<24>>24)==(0); //@line 5700 "libraries/glad/glad.cpp"
 if (!($2888)) {
  $2889 = (FUNCTION_TABLE_ii[$load & 8191](491692)|0); //@line 5701 "libraries/glad/glad.cpp"
  HEAP32[166282] = $2889; //@line 5701 "libraries/glad/glad.cpp"
 }
 $2890 = HEAP8[1252322]|0; //@line 5706 "libraries/glad/glad.cpp"
 $2891 = ($2890<<24>>24)==(0); //@line 5706 "libraries/glad/glad.cpp"
 if (!($2891)) {
  $2892 = (FUNCTION_TABLE_ii[$load & 8191](491720)|0); //@line 5707 "libraries/glad/glad.cpp"
  HEAP32[166283] = $2892; //@line 5707 "libraries/glad/glad.cpp"
  $2893 = (FUNCTION_TABLE_ii[$load & 8191](491749)|0); //@line 5708 "libraries/glad/glad.cpp"
  HEAP32[166284] = $2893; //@line 5708 "libraries/glad/glad.cpp"
 }
 $2894 = HEAP8[1252323]|0; //@line 5713 "libraries/glad/glad.cpp"
 $2895 = ($2894<<24>>24)==(0); //@line 5713 "libraries/glad/glad.cpp"
 if (!($2895)) {
  $2896 = (FUNCTION_TABLE_ii[$load & 8191](491780)|0); //@line 5714 "libraries/glad/glad.cpp"
  HEAP32[166285] = $2896; //@line 5714 "libraries/glad/glad.cpp"
  $2897 = (FUNCTION_TABLE_ii[$load & 8191](491794)|0); //@line 5715 "libraries/glad/glad.cpp"
  HEAP32[166286] = $2897; //@line 5715 "libraries/glad/glad.cpp"
  $2898 = (FUNCTION_TABLE_ii[$load & 8191](491811)|0); //@line 5716 "libraries/glad/glad.cpp"
  HEAP32[166287] = $2898; //@line 5716 "libraries/glad/glad.cpp"
 }
 $2899 = HEAP8[1252324]|0; //@line 5721 "libraries/glad/glad.cpp"
 $2900 = ($2899<<24>>24)==(0); //@line 5721 "libraries/glad/glad.cpp"
 if (!($2900)) {
  $2901 = (FUNCTION_TABLE_ii[$load & 8191](491823)|0); //@line 5722 "libraries/glad/glad.cpp"
  HEAP32[166288] = $2901; //@line 5722 "libraries/glad/glad.cpp"
 }
 $2902 = HEAP8[1252325]|0; //@line 5727 "libraries/glad/glad.cpp"
 $2903 = ($2902<<24>>24)==(0); //@line 5727 "libraries/glad/glad.cpp"
 if (!($2903)) {
  $2904 = (FUNCTION_TABLE_ii[$load & 8191](491851)|0); //@line 5728 "libraries/glad/glad.cpp"
  HEAP32[166289] = $2904; //@line 5728 "libraries/glad/glad.cpp"
  $2905 = (FUNCTION_TABLE_ii[$load & 8191](491877)|0); //@line 5729 "libraries/glad/glad.cpp"
  HEAP32[166290] = $2905; //@line 5729 "libraries/glad/glad.cpp"
  $2906 = (FUNCTION_TABLE_ii[$load & 8191](491905)|0); //@line 5730 "libraries/glad/glad.cpp"
  HEAP32[166291] = $2906; //@line 5730 "libraries/glad/glad.cpp"
  $2907 = (FUNCTION_TABLE_ii[$load & 8191](491936)|0); //@line 5731 "libraries/glad/glad.cpp"
  HEAP32[166292] = $2907; //@line 5731 "libraries/glad/glad.cpp"
  $2908 = (FUNCTION_TABLE_ii[$load & 8191](491969)|0); //@line 5732 "libraries/glad/glad.cpp"
  HEAP32[166293] = $2908; //@line 5732 "libraries/glad/glad.cpp"
  $2909 = (FUNCTION_TABLE_ii[$load & 8191](492000)|0); //@line 5733 "libraries/glad/glad.cpp"
  HEAP32[166294] = $2909; //@line 5733 "libraries/glad/glad.cpp"
  $2910 = (FUNCTION_TABLE_ii[$load & 8191](492021)|0); //@line 5734 "libraries/glad/glad.cpp"
  HEAP32[166295] = $2910; //@line 5734 "libraries/glad/glad.cpp"
  $2911 = (FUNCTION_TABLE_ii[$load & 8191](492045)|0); //@line 5735 "libraries/glad/glad.cpp"
  HEAP32[166296] = $2911; //@line 5735 "libraries/glad/glad.cpp"
  $2912 = (FUNCTION_TABLE_ii[$load & 8191](492076)|0); //@line 5736 "libraries/glad/glad.cpp"
  HEAP32[166297] = $2912; //@line 5736 "libraries/glad/glad.cpp"
  $2913 = (FUNCTION_TABLE_ii[$load & 8191](492098)|0); //@line 5737 "libraries/glad/glad.cpp"
  HEAP32[166298] = $2913; //@line 5737 "libraries/glad/glad.cpp"
  $2914 = (FUNCTION_TABLE_ii[$load & 8191](492118)|0); //@line 5738 "libraries/glad/glad.cpp"
  HEAP32[166299] = $2914; //@line 5738 "libraries/glad/glad.cpp"
 }
 $2915 = HEAP8[1252329]|0; //@line 5746 "libraries/glad/glad.cpp"
 $2916 = ($2915<<24>>24)==(0); //@line 5746 "libraries/glad/glad.cpp"
 if (!($2916)) {
  $2917 = (FUNCTION_TABLE_ii[$load & 8191](492149)|0); //@line 5747 "libraries/glad/glad.cpp"
  HEAP32[166300] = $2917; //@line 5747 "libraries/glad/glad.cpp"
 }
 $2918 = HEAP8[1252334]|0; //@line 5756 "libraries/glad/glad.cpp"
 $2919 = ($2918<<24>>24)==(0); //@line 5756 "libraries/glad/glad.cpp"
 if (!($2919)) {
  $2920 = (FUNCTION_TABLE_ii[$load & 8191](492171)|0); //@line 5757 "libraries/glad/glad.cpp"
  HEAP32[166301] = $2920; //@line 5757 "libraries/glad/glad.cpp"
  $2921 = (FUNCTION_TABLE_ii[$load & 8191](492193)|0); //@line 5758 "libraries/glad/glad.cpp"
  HEAP32[166302] = $2921; //@line 5758 "libraries/glad/glad.cpp"
 }
 $2922 = HEAP8[1252335]|0; //@line 5763 "libraries/glad/glad.cpp"
 $2923 = ($2922<<24>>24)==(0); //@line 5763 "libraries/glad/glad.cpp"
 if (!($2923)) {
  $2924 = (FUNCTION_TABLE_ii[$load & 8191](492219)|0); //@line 5764 "libraries/glad/glad.cpp"
  HEAP32[166303] = $2924; //@line 5764 "libraries/glad/glad.cpp"
 }
 $2925 = HEAP8[1252340]|0; //@line 5773 "libraries/glad/glad.cpp"
 $2926 = ($2925<<24>>24)==(0); //@line 5773 "libraries/glad/glad.cpp"
 if (!($2926)) {
  $2927 = (FUNCTION_TABLE_ii[$load & 8191](492239)|0); //@line 5774 "libraries/glad/glad.cpp"
  HEAP32[166304] = $2927; //@line 5774 "libraries/glad/glad.cpp"
  $2928 = (FUNCTION_TABLE_ii[$load & 8191](492263)|0); //@line 5775 "libraries/glad/glad.cpp"
  HEAP32[166305] = $2928; //@line 5775 "libraries/glad/glad.cpp"
 }
 $2929 = HEAP8[1252344]|0; //@line 5783 "libraries/glad/glad.cpp"
 $2930 = ($2929<<24>>24)==(0); //@line 5783 "libraries/glad/glad.cpp"
 if (!($2930)) {
  $2931 = (FUNCTION_TABLE_ii[$load & 8191](492285)|0); //@line 5784 "libraries/glad/glad.cpp"
  HEAP32[166306] = $2931; //@line 5784 "libraries/glad/glad.cpp"
 }
 $2932 = HEAP8[1252345]|0; //@line 5789 "libraries/glad/glad.cpp"
 $2933 = ($2932<<24>>24)==(0); //@line 5789 "libraries/glad/glad.cpp"
 if (!($2933)) {
  $2934 = (FUNCTION_TABLE_ii[$load & 8191](492308)|0); //@line 5790 "libraries/glad/glad.cpp"
  HEAP32[166307] = $2934; //@line 5790 "libraries/glad/glad.cpp"
 }
 $2935 = HEAP8[1252346]|0; //@line 5795 "libraries/glad/glad.cpp"
 $2936 = ($2935<<24>>24)==(0); //@line 5795 "libraries/glad/glad.cpp"
 if (!($2936)) {
  $2937 = (FUNCTION_TABLE_ii[$load & 8191](492346)|0); //@line 5796 "libraries/glad/glad.cpp"
  HEAP32[166308] = $2937; //@line 5796 "libraries/glad/glad.cpp"
  $2938 = (FUNCTION_TABLE_ii[$load & 8191](492373)|0); //@line 5797 "libraries/glad/glad.cpp"
  HEAP32[166309] = $2938; //@line 5797 "libraries/glad/glad.cpp"
  $2939 = (FUNCTION_TABLE_ii[$load & 8191](492402)|0); //@line 5798 "libraries/glad/glad.cpp"
  HEAP32[166310] = $2939; //@line 5798 "libraries/glad/glad.cpp"
 }
 $2940 = HEAP8[1252352]|0; //@line 5808 "libraries/glad/glad.cpp"
 $2941 = ($2940<<24>>24)==(0); //@line 5808 "libraries/glad/glad.cpp"
 if (!($2941)) {
  $2942 = (FUNCTION_TABLE_ii[$load & 8191](492429)|0); //@line 5809 "libraries/glad/glad.cpp"
  HEAP32[166311] = $2942; //@line 5809 "libraries/glad/glad.cpp"
 }
 $2943 = HEAP8[1252357]|0; //@line 5818 "libraries/glad/glad.cpp"
 $2944 = ($2943<<24>>24)==(0); //@line 5818 "libraries/glad/glad.cpp"
 if (!($2944)) {
  $2945 = (FUNCTION_TABLE_ii[$load & 8191](492462)|0); //@line 5819 "libraries/glad/glad.cpp"
  HEAP32[166312] = $2945; //@line 5819 "libraries/glad/glad.cpp"
 }
 $2946 = HEAP8[1252358]|0; //@line 5824 "libraries/glad/glad.cpp"
 $2947 = ($2946<<24>>24)==(0); //@line 5824 "libraries/glad/glad.cpp"
 if (!($2947)) {
  $2948 = (FUNCTION_TABLE_ii[$load & 8191](492487)|0); //@line 5825 "libraries/glad/glad.cpp"
  HEAP32[166313] = $2948; //@line 5825 "libraries/glad/glad.cpp"
  $2949 = (FUNCTION_TABLE_ii[$load & 8191](492509)|0); //@line 5826 "libraries/glad/glad.cpp"
  HEAP32[166314] = $2949; //@line 5826 "libraries/glad/glad.cpp"
  $2950 = (FUNCTION_TABLE_ii[$load & 8191](492533)|0); //@line 5827 "libraries/glad/glad.cpp"
  HEAP32[166315] = $2950; //@line 5827 "libraries/glad/glad.cpp"
  $2951 = (FUNCTION_TABLE_ii[$load & 8191](492562)|0); //@line 5828 "libraries/glad/glad.cpp"
  HEAP32[166316] = $2951; //@line 5828 "libraries/glad/glad.cpp"
  $2952 = (FUNCTION_TABLE_ii[$load & 8191](492591)|0); //@line 5829 "libraries/glad/glad.cpp"
  HEAP32[166317] = $2952; //@line 5829 "libraries/glad/glad.cpp"
 }
 $2953 = HEAP8[1252359]|0; //@line 5834 "libraries/glad/glad.cpp"
 $2954 = ($2953<<24>>24)==(0); //@line 5834 "libraries/glad/glad.cpp"
 if (!($2954)) {
  $2955 = (FUNCTION_TABLE_ii[$load & 8191](492625)|0); //@line 5835 "libraries/glad/glad.cpp"
  HEAP32[166318] = $2955; //@line 5835 "libraries/glad/glad.cpp"
  $2956 = (FUNCTION_TABLE_ii[$load & 8191](492642)|0); //@line 5836 "libraries/glad/glad.cpp"
  HEAP32[166319] = $2956; //@line 5836 "libraries/glad/glad.cpp"
  $2957 = (FUNCTION_TABLE_ii[$load & 8191](492662)|0); //@line 5837 "libraries/glad/glad.cpp"
  HEAP32[166320] = $2957; //@line 5837 "libraries/glad/glad.cpp"
  $2958 = (FUNCTION_TABLE_ii[$load & 8191](492678)|0); //@line 5838 "libraries/glad/glad.cpp"
  HEAP32[166321] = $2958; //@line 5838 "libraries/glad/glad.cpp"
  $2959 = (FUNCTION_TABLE_ii[$load & 8191](492693)|0); //@line 5839 "libraries/glad/glad.cpp"
  HEAP32[166322] = $2959; //@line 5839 "libraries/glad/glad.cpp"
  $2960 = (FUNCTION_TABLE_ii[$load & 8191](492710)|0); //@line 5840 "libraries/glad/glad.cpp"
  HEAP32[166323] = $2960; //@line 5840 "libraries/glad/glad.cpp"
  $2961 = (FUNCTION_TABLE_ii[$load & 8191](492729)|0); //@line 5841 "libraries/glad/glad.cpp"
  HEAP32[166324] = $2961; //@line 5841 "libraries/glad/glad.cpp"
  $2962 = (FUNCTION_TABLE_ii[$load & 8191](492747)|0); //@line 5842 "libraries/glad/glad.cpp"
  HEAP32[166325] = $2962; //@line 5842 "libraries/glad/glad.cpp"
 }
 $2963 = HEAP8[1252361]|0; //@line 5848 "libraries/glad/glad.cpp"
 $2964 = ($2963<<24>>24)==(0); //@line 5848 "libraries/glad/glad.cpp"
 if (!($2964)) {
  $2965 = (FUNCTION_TABLE_ii[$load & 8191](492767)|0); //@line 5849 "libraries/glad/glad.cpp"
  HEAP32[166326] = $2965; //@line 5849 "libraries/glad/glad.cpp"
  $2966 = (FUNCTION_TABLE_ii[$load & 8191](492791)|0); //@line 5850 "libraries/glad/glad.cpp"
  HEAP32[166327] = $2966; //@line 5850 "libraries/glad/glad.cpp"
 }
 $2967 = HEAP8[1252362]|0; //@line 5855 "libraries/glad/glad.cpp"
 $2968 = ($2967<<24>>24)==(0); //@line 5855 "libraries/glad/glad.cpp"
 if (!($2968)) {
  $2969 = (FUNCTION_TABLE_ii[$load & 8191](492821)|0); //@line 5856 "libraries/glad/glad.cpp"
  HEAP32[166328] = $2969; //@line 5856 "libraries/glad/glad.cpp"
  $2970 = (FUNCTION_TABLE_ii[$load & 8191](492859)|0); //@line 5857 "libraries/glad/glad.cpp"
  HEAP32[166329] = $2970; //@line 5857 "libraries/glad/glad.cpp"
 }
 $2971 = HEAP8[1252363]|0; //@line 5862 "libraries/glad/glad.cpp"
 $2972 = ($2971<<24>>24)==(0); //@line 5862 "libraries/glad/glad.cpp"
 if (!($2972)) {
  $2973 = (FUNCTION_TABLE_ii[$load & 8191](492896)|0); //@line 5863 "libraries/glad/glad.cpp"
  HEAP32[166330] = $2973; //@line 5863 "libraries/glad/glad.cpp"
  $2974 = (FUNCTION_TABLE_ii[$load & 8191](492919)|0); //@line 5864 "libraries/glad/glad.cpp"
  HEAP32[166331] = $2974; //@line 5864 "libraries/glad/glad.cpp"
  $2975 = (FUNCTION_TABLE_ii[$load & 8191](492944)|0); //@line 5865 "libraries/glad/glad.cpp"
  HEAP32[166332] = $2975; //@line 5865 "libraries/glad/glad.cpp"
 }
 $2976 = HEAP8[1252366]|0; //@line 5872 "libraries/glad/glad.cpp"
 $2977 = ($2976<<24>>24)==(0); //@line 5872 "libraries/glad/glad.cpp"
 if (!($2977)) {
  $2978 = (FUNCTION_TABLE_ii[$load & 8191](492972)|0); //@line 5873 "libraries/glad/glad.cpp"
  HEAP32[166333] = $2978; //@line 5873 "libraries/glad/glad.cpp"
  $2979 = (FUNCTION_TABLE_ii[$load & 8191](492989)|0); //@line 5874 "libraries/glad/glad.cpp"
  HEAP32[166334] = $2979; //@line 5874 "libraries/glad/glad.cpp"
  $2980 = (FUNCTION_TABLE_ii[$load & 8191](493003)|0); //@line 5875 "libraries/glad/glad.cpp"
  HEAP32[166335] = $2980; //@line 5875 "libraries/glad/glad.cpp"
  $2981 = (FUNCTION_TABLE_ii[$load & 8191](493021)|0); //@line 5876 "libraries/glad/glad.cpp"
  HEAP32[166336] = $2981; //@line 5876 "libraries/glad/glad.cpp"
  $2982 = (FUNCTION_TABLE_ii[$load & 8191](493043)|0); //@line 5877 "libraries/glad/glad.cpp"
  HEAP32[166337] = $2982; //@line 5877 "libraries/glad/glad.cpp"
  $2983 = (FUNCTION_TABLE_ii[$load & 8191](493059)|0); //@line 5878 "libraries/glad/glad.cpp"
  HEAP32[166338] = $2983; //@line 5878 "libraries/glad/glad.cpp"
  $2984 = (FUNCTION_TABLE_ii[$load & 8191](493080)|0); //@line 5879 "libraries/glad/glad.cpp"
  HEAP32[166339] = $2984; //@line 5879 "libraries/glad/glad.cpp"
 }
 $2985 = HEAP8[1252370]|0; //@line 5887 "libraries/glad/glad.cpp"
 $2986 = ($2985<<24>>24)==(0); //@line 5887 "libraries/glad/glad.cpp"
 if (!($2986)) {
  $2987 = (FUNCTION_TABLE_ii[$load & 8191](493097)|0); //@line 5888 "libraries/glad/glad.cpp"
  HEAP32[166340] = $2987; //@line 5888 "libraries/glad/glad.cpp"
  $2988 = (FUNCTION_TABLE_ii[$load & 8191](493117)|0); //@line 5889 "libraries/glad/glad.cpp"
  HEAP32[166341] = $2988; //@line 5889 "libraries/glad/glad.cpp"
 }
 $2989 = HEAP8[1252371]|0; //@line 5894 "libraries/glad/glad.cpp"
 $2990 = ($2989<<24>>24)==(0); //@line 5894 "libraries/glad/glad.cpp"
 if (!($2990)) {
  $2991 = (FUNCTION_TABLE_ii[$load & 8191](493148)|0); //@line 5895 "libraries/glad/glad.cpp"
  HEAP32[166342] = $2991; //@line 5895 "libraries/glad/glad.cpp"
  $2992 = (FUNCTION_TABLE_ii[$load & 8191](493171)|0); //@line 5896 "libraries/glad/glad.cpp"
  HEAP32[166343] = $2992; //@line 5896 "libraries/glad/glad.cpp"
  $2993 = (FUNCTION_TABLE_ii[$load & 8191](493197)|0); //@line 5897 "libraries/glad/glad.cpp"
  HEAP32[166344] = $2993; //@line 5897 "libraries/glad/glad.cpp"
  $2994 = (FUNCTION_TABLE_ii[$load & 8191](493220)|0); //@line 5898 "libraries/glad/glad.cpp"
  HEAP32[166345] = $2994; //@line 5898 "libraries/glad/glad.cpp"
 }
 $2995 = HEAP8[1252372]|0; //@line 5903 "libraries/glad/glad.cpp"
 $2996 = ($2995<<24>>24)==(0); //@line 5903 "libraries/glad/glad.cpp"
 if (!($2996)) {
  $2997 = (FUNCTION_TABLE_ii[$load & 8191](493241)|0); //@line 5904 "libraries/glad/glad.cpp"
  HEAP32[166346] = $2997; //@line 5904 "libraries/glad/glad.cpp"
  $2998 = (FUNCTION_TABLE_ii[$load & 8191](493265)|0); //@line 5905 "libraries/glad/glad.cpp"
  HEAP32[166347] = $2998; //@line 5905 "libraries/glad/glad.cpp"
  $2999 = (FUNCTION_TABLE_ii[$load & 8191](493294)|0); //@line 5906 "libraries/glad/glad.cpp"
  HEAP32[166348] = $2999; //@line 5906 "libraries/glad/glad.cpp"
 }
 $3000 = HEAP8[1252373]|0; //@line 5911 "libraries/glad/glad.cpp"
 $3001 = ($3000<<24>>24)==(0); //@line 5911 "libraries/glad/glad.cpp"
 if (!($3001)) {
  $3002 = (FUNCTION_TABLE_ii[$load & 8191](493323)|0); //@line 5912 "libraries/glad/glad.cpp"
  HEAP32[166349] = $3002; //@line 5912 "libraries/glad/glad.cpp"
  $3003 = (FUNCTION_TABLE_ii[$load & 8191](493349)|0); //@line 5913 "libraries/glad/glad.cpp"
  HEAP32[166350] = $3003; //@line 5913 "libraries/glad/glad.cpp"
  $3004 = (FUNCTION_TABLE_ii[$load & 8191](493376)|0); //@line 5914 "libraries/glad/glad.cpp"
  HEAP32[166351] = $3004; //@line 5914 "libraries/glad/glad.cpp"
  $3005 = (FUNCTION_TABLE_ii[$load & 8191](493405)|0); //@line 5915 "libraries/glad/glad.cpp"
  HEAP32[166352] = $3005; //@line 5915 "libraries/glad/glad.cpp"
  $3006 = (FUNCTION_TABLE_ii[$load & 8191](493430)|0); //@line 5916 "libraries/glad/glad.cpp"
  HEAP32[166353] = $3006; //@line 5916 "libraries/glad/glad.cpp"
  $3007 = (FUNCTION_TABLE_ii[$load & 8191](493455)|0); //@line 5917 "libraries/glad/glad.cpp"
  HEAP32[166354] = $3007; //@line 5917 "libraries/glad/glad.cpp"
  $3008 = (FUNCTION_TABLE_ii[$load & 8191](493480)|0); //@line 5918 "libraries/glad/glad.cpp"
  HEAP32[166355] = $3008; //@line 5918 "libraries/glad/glad.cpp"
 }
 $3009 = HEAP8[1252385]|0; //@line 5934 "libraries/glad/glad.cpp"
 $3010 = ($3009<<24>>24)==(0); //@line 5934 "libraries/glad/glad.cpp"
 if (!($3010)) {
  $3011 = (FUNCTION_TABLE_ii[$load & 8191](493505)|0); //@line 5935 "libraries/glad/glad.cpp"
  HEAP32[166356] = $3011; //@line 5935 "libraries/glad/glad.cpp"
 }
 $3012 = HEAP8[1252386]|0; //@line 5940 "libraries/glad/glad.cpp"
 $3013 = ($3012<<24>>24)==(0); //@line 5940 "libraries/glad/glad.cpp"
 if (!($3013)) {
  $3014 = (FUNCTION_TABLE_ii[$load & 8191](493530)|0); //@line 5941 "libraries/glad/glad.cpp"
  HEAP32[166357] = $3014; //@line 5941 "libraries/glad/glad.cpp"
 }
 $3015 = HEAP8[1252387]|0; //@line 5946 "libraries/glad/glad.cpp"
 $3016 = ($3015<<24>>24)==(0); //@line 5946 "libraries/glad/glad.cpp"
 if (!($3016)) {
  $3017 = (FUNCTION_TABLE_ii[$load & 8191](493552)|0); //@line 5947 "libraries/glad/glad.cpp"
  HEAP32[166358] = $3017; //@line 5947 "libraries/glad/glad.cpp"
  $3018 = (FUNCTION_TABLE_ii[$load & 8191](493588)|0); //@line 5948 "libraries/glad/glad.cpp"
  HEAP32[166359] = $3018; //@line 5948 "libraries/glad/glad.cpp"
 }
 $3019 = HEAP8[1252394]|0; //@line 5959 "libraries/glad/glad.cpp"
 $3020 = ($3019<<24>>24)==(0); //@line 5959 "libraries/glad/glad.cpp"
 if (!($3020)) {
  $3021 = (FUNCTION_TABLE_ii[$load & 8191](493625)|0); //@line 5960 "libraries/glad/glad.cpp"
  HEAP32[166360] = $3021; //@line 5960 "libraries/glad/glad.cpp"
 }
 $3022 = HEAP8[1252396]|0; //@line 5966 "libraries/glad/glad.cpp"
 $3023 = ($3022<<24>>24)==(0); //@line 5966 "libraries/glad/glad.cpp"
 if (!($3023)) {
  $3024 = (FUNCTION_TABLE_ii[$load & 8191](493649)|0); //@line 5967 "libraries/glad/glad.cpp"
  HEAP32[166361] = $3024; //@line 5967 "libraries/glad/glad.cpp"
 }
 $3025 = HEAP8[1252397]|0; //@line 5972 "libraries/glad/glad.cpp"
 $3026 = ($3025<<24>>24)==(0); //@line 5972 "libraries/glad/glad.cpp"
 if (!($3026)) {
  $3027 = (FUNCTION_TABLE_ii[$load & 8191](493687)|0); //@line 5973 "libraries/glad/glad.cpp"
  HEAP32[166362] = $3027; //@line 5973 "libraries/glad/glad.cpp"
  $3028 = (FUNCTION_TABLE_ii[$load & 8191](493706)|0); //@line 5974 "libraries/glad/glad.cpp"
  HEAP32[166363] = $3028; //@line 5974 "libraries/glad/glad.cpp"
  $3029 = (FUNCTION_TABLE_ii[$load & 8191](493728)|0); //@line 5975 "libraries/glad/glad.cpp"
  HEAP32[166364] = $3029; //@line 5975 "libraries/glad/glad.cpp"
 }
 $3030 = HEAP8[1252398]|0; //@line 5980 "libraries/glad/glad.cpp"
 $3031 = ($3030<<24>>24)==(0); //@line 5980 "libraries/glad/glad.cpp"
 if (!($3031)) {
  $3032 = (FUNCTION_TABLE_ii[$load & 8191](493748)|0); //@line 5981 "libraries/glad/glad.cpp"
  HEAP32[166365] = $3032; //@line 5981 "libraries/glad/glad.cpp"
  $3033 = (FUNCTION_TABLE_ii[$load & 8191](493770)|0); //@line 5982 "libraries/glad/glad.cpp"
  HEAP32[166366] = $3033; //@line 5982 "libraries/glad/glad.cpp"
  $3034 = (FUNCTION_TABLE_ii[$load & 8191](493793)|0); //@line 5983 "libraries/glad/glad.cpp"
  HEAP32[166367] = $3034; //@line 5983 "libraries/glad/glad.cpp"
  $3035 = (FUNCTION_TABLE_ii[$load & 8191](493816)|0); //@line 5984 "libraries/glad/glad.cpp"
  HEAP32[166368] = $3035; //@line 5984 "libraries/glad/glad.cpp"
  $3036 = (FUNCTION_TABLE_ii[$load & 8191](493836)|0); //@line 5985 "libraries/glad/glad.cpp"
  HEAP32[166369] = $3036; //@line 5985 "libraries/glad/glad.cpp"
  $3037 = (FUNCTION_TABLE_ii[$load & 8191](493863)|0); //@line 5986 "libraries/glad/glad.cpp"
  HEAP32[166370] = $3037; //@line 5986 "libraries/glad/glad.cpp"
  $3038 = (FUNCTION_TABLE_ii[$load & 8191](493889)|0); //@line 5987 "libraries/glad/glad.cpp"
  HEAP32[166371] = $3038; //@line 5987 "libraries/glad/glad.cpp"
  $3039 = (FUNCTION_TABLE_ii[$load & 8191](493915)|0); //@line 5988 "libraries/glad/glad.cpp"
  HEAP32[166372] = $3039; //@line 5988 "libraries/glad/glad.cpp"
  $3040 = (FUNCTION_TABLE_ii[$load & 8191](493939)|0); //@line 5989 "libraries/glad/glad.cpp"
  HEAP32[166373] = $3040; //@line 5989 "libraries/glad/glad.cpp"
  $3041 = (FUNCTION_TABLE_ii[$load & 8191](493967)|0); //@line 5990 "libraries/glad/glad.cpp"
  HEAP32[166374] = $3041; //@line 5990 "libraries/glad/glad.cpp"
 }
 $3042 = HEAP8[1252402]|0; //@line 5998 "libraries/glad/glad.cpp"
 $3043 = ($3042<<24>>24)==(0); //@line 5998 "libraries/glad/glad.cpp"
 if (!($3043)) {
  $3044 = (FUNCTION_TABLE_ii[$load & 8191](493991)|0); //@line 5999 "libraries/glad/glad.cpp"
  HEAP32[166375] = $3044; //@line 5999 "libraries/glad/glad.cpp"
  $3045 = (FUNCTION_TABLE_ii[$load & 8191](494027)|0); //@line 6000 "libraries/glad/glad.cpp"
  HEAP32[166376] = $3045; //@line 6000 "libraries/glad/glad.cpp"
 }
 $3046 = HEAP8[1252403]|0; //@line 6005 "libraries/glad/glad.cpp"
 $3047 = ($3046<<24>>24)==(0); //@line 6005 "libraries/glad/glad.cpp"
 if (!($3047)) {
  $3048 = (FUNCTION_TABLE_ii[$load & 8191](494065)|0); //@line 6006 "libraries/glad/glad.cpp"
  HEAP32[166377] = $3048; //@line 6006 "libraries/glad/glad.cpp"
  $3049 = (FUNCTION_TABLE_ii[$load & 8191](494106)|0); //@line 6007 "libraries/glad/glad.cpp"
  HEAP32[166378] = $3049; //@line 6007 "libraries/glad/glad.cpp"
 }
 $3050 = HEAP8[1252404]|0; //@line 6012 "libraries/glad/glad.cpp"
 $3051 = ($3050<<24>>24)==(0); //@line 6012 "libraries/glad/glad.cpp"
 if (!($3051)) {
  $3052 = (FUNCTION_TABLE_ii[$load & 8191](494149)|0); //@line 6013 "libraries/glad/glad.cpp"
  HEAP32[166379] = $3052; //@line 6013 "libraries/glad/glad.cpp"
  $3053 = (FUNCTION_TABLE_ii[$load & 8191](494170)|0); //@line 6014 "libraries/glad/glad.cpp"
  HEAP32[166380] = $3053; //@line 6014 "libraries/glad/glad.cpp"
  $3054 = (FUNCTION_TABLE_ii[$load & 8191](494198)|0); //@line 6015 "libraries/glad/glad.cpp"
  HEAP32[166381] = $3054; //@line 6015 "libraries/glad/glad.cpp"
  $3055 = (FUNCTION_TABLE_ii[$load & 8191](494228)|0); //@line 6016 "libraries/glad/glad.cpp"
  HEAP32[166382] = $3055; //@line 6016 "libraries/glad/glad.cpp"
  $3056 = (FUNCTION_TABLE_ii[$load & 8191](494261)|0); //@line 6017 "libraries/glad/glad.cpp"
  HEAP32[166383] = $3056; //@line 6017 "libraries/glad/glad.cpp"
  $3057 = (FUNCTION_TABLE_ii[$load & 8191](494280)|0); //@line 6018 "libraries/glad/glad.cpp"
  HEAP32[166384] = $3057; //@line 6018 "libraries/glad/glad.cpp"
  $3058 = (FUNCTION_TABLE_ii[$load & 8191](494308)|0); //@line 6019 "libraries/glad/glad.cpp"
  HEAP32[166385] = $3058; //@line 6019 "libraries/glad/glad.cpp"
  $3059 = (FUNCTION_TABLE_ii[$load & 8191](494339)|0); //@line 6020 "libraries/glad/glad.cpp"
  HEAP32[166386] = $3059; //@line 6020 "libraries/glad/glad.cpp"
  $3060 = (FUNCTION_TABLE_ii[$load & 8191](494361)|0); //@line 6021 "libraries/glad/glad.cpp"
  HEAP32[166387] = $3060; //@line 6021 "libraries/glad/glad.cpp"
  $3061 = (FUNCTION_TABLE_ii[$load & 8191](494384)|0); //@line 6022 "libraries/glad/glad.cpp"
  HEAP32[166388] = $3061; //@line 6022 "libraries/glad/glad.cpp"
  $3062 = (FUNCTION_TABLE_ii[$load & 8191](494413)|0); //@line 6023 "libraries/glad/glad.cpp"
  HEAP32[166389] = $3062; //@line 6023 "libraries/glad/glad.cpp"
  $3063 = (FUNCTION_TABLE_ii[$load & 8191](494443)|0); //@line 6024 "libraries/glad/glad.cpp"
  HEAP32[166390] = $3063; //@line 6024 "libraries/glad/glad.cpp"
  $3064 = (FUNCTION_TABLE_ii[$load & 8191](494471)|0); //@line 6025 "libraries/glad/glad.cpp"
  HEAP32[166391] = $3064; //@line 6025 "libraries/glad/glad.cpp"
 }
 $3065 = HEAP8[1252405]|0; //@line 6030 "libraries/glad/glad.cpp"
 $3066 = ($3065<<24>>24)==(0); //@line 6030 "libraries/glad/glad.cpp"
 if (!($3066)) {
  $3067 = (FUNCTION_TABLE_ii[$load & 8191](494497)|0); //@line 6031 "libraries/glad/glad.cpp"
  HEAP32[166392] = $3067; //@line 6031 "libraries/glad/glad.cpp"
  $3068 = (FUNCTION_TABLE_ii[$load & 8191](494517)|0); //@line 6032 "libraries/glad/glad.cpp"
  HEAP32[166393] = $3068; //@line 6032 "libraries/glad/glad.cpp"
 }
 $3069 = HEAP8[1252407]|0; //@line 6038 "libraries/glad/glad.cpp"
 $3070 = ($3069<<24>>24)==(0); //@line 6038 "libraries/glad/glad.cpp"
 if (!($3070)) {
  $3071 = (FUNCTION_TABLE_ii[$load & 8191](494534)|0); //@line 6039 "libraries/glad/glad.cpp"
  HEAP32[166394] = $3071; //@line 6039 "libraries/glad/glad.cpp"
  $3072 = (FUNCTION_TABLE_ii[$load & 8191](494551)|0); //@line 6040 "libraries/glad/glad.cpp"
  HEAP32[166395] = $3072; //@line 6040 "libraries/glad/glad.cpp"
  $3073 = (FUNCTION_TABLE_ii[$load & 8191](494568)|0); //@line 6041 "libraries/glad/glad.cpp"
  HEAP32[166396] = $3073; //@line 6041 "libraries/glad/glad.cpp"
  $3074 = (FUNCTION_TABLE_ii[$load & 8191](494580)|0); //@line 6042 "libraries/glad/glad.cpp"
  HEAP32[166397] = $3074; //@line 6042 "libraries/glad/glad.cpp"
  $3075 = (FUNCTION_TABLE_ii[$load & 8191](494597)|0); //@line 6043 "libraries/glad/glad.cpp"
  HEAP32[166398] = $3075; //@line 6043 "libraries/glad/glad.cpp"
  $3076 = (FUNCTION_TABLE_ii[$load & 8191](494618)|0); //@line 6044 "libraries/glad/glad.cpp"
  HEAP32[166399] = $3076; //@line 6044 "libraries/glad/glad.cpp"
  $3077 = (FUNCTION_TABLE_ii[$load & 8191](494636)|0); //@line 6045 "libraries/glad/glad.cpp"
  HEAP32[166400] = $3077; //@line 6045 "libraries/glad/glad.cpp"
  $3078 = (FUNCTION_TABLE_ii[$load & 8191](494653)|0); //@line 6046 "libraries/glad/glad.cpp"
  HEAP32[166401] = $3078; //@line 6046 "libraries/glad/glad.cpp"
  $3079 = (FUNCTION_TABLE_ii[$load & 8191](494677)|0); //@line 6047 "libraries/glad/glad.cpp"
  HEAP32[166402] = $3079; //@line 6047 "libraries/glad/glad.cpp"
  $3080 = (FUNCTION_TABLE_ii[$load & 8191](494700)|0); //@line 6048 "libraries/glad/glad.cpp"
  HEAP32[166403] = $3080; //@line 6048 "libraries/glad/glad.cpp"
  $3081 = (FUNCTION_TABLE_ii[$load & 8191](494730)|0); //@line 6049 "libraries/glad/glad.cpp"
  HEAP32[166404] = $3081; //@line 6049 "libraries/glad/glad.cpp"
  $3082 = (FUNCTION_TABLE_ii[$load & 8191](494753)|0); //@line 6050 "libraries/glad/glad.cpp"
  HEAP32[166405] = $3082; //@line 6050 "libraries/glad/glad.cpp"
  $3083 = (FUNCTION_TABLE_ii[$load & 8191](494776)|0); //@line 6051 "libraries/glad/glad.cpp"
  HEAP32[166406] = $3083; //@line 6051 "libraries/glad/glad.cpp"
  $3084 = (FUNCTION_TABLE_ii[$load & 8191](494794)|0); //@line 6052 "libraries/glad/glad.cpp"
  HEAP32[166407] = $3084; //@line 6052 "libraries/glad/glad.cpp"
  $3085 = (FUNCTION_TABLE_ii[$load & 8191](494827)|0); //@line 6053 "libraries/glad/glad.cpp"
  HEAP32[166408] = $3085; //@line 6053 "libraries/glad/glad.cpp"
  $3086 = (FUNCTION_TABLE_ii[$load & 8191](494851)|0); //@line 6054 "libraries/glad/glad.cpp"
  HEAP32[166409] = $3086; //@line 6054 "libraries/glad/glad.cpp"
  $3087 = (FUNCTION_TABLE_ii[$load & 8191](494874)|0); //@line 6055 "libraries/glad/glad.cpp"
  HEAP32[166410] = $3087; //@line 6055 "libraries/glad/glad.cpp"
 }
 $3088 = HEAP8[1252409]|0; //@line 6061 "libraries/glad/glad.cpp"
 $3089 = ($3088<<24>>24)==(0); //@line 6061 "libraries/glad/glad.cpp"
 if (!($3089)) {
  $3090 = (FUNCTION_TABLE_ii[$load & 8191](494894)|0); //@line 6062 "libraries/glad/glad.cpp"
  HEAP32[166411] = $3090; //@line 6062 "libraries/glad/glad.cpp"
  $3091 = (FUNCTION_TABLE_ii[$load & 8191](494921)|0); //@line 6063 "libraries/glad/glad.cpp"
  HEAP32[166412] = $3091; //@line 6063 "libraries/glad/glad.cpp"
 }
 $3092 = HEAP8[1252410]|0; //@line 6068 "libraries/glad/glad.cpp"
 $3093 = ($3092<<24>>24)==(0); //@line 6068 "libraries/glad/glad.cpp"
 if (!($3093)) {
  $3094 = (FUNCTION_TABLE_ii[$load & 8191](494946)|0); //@line 6069 "libraries/glad/glad.cpp"
  HEAP32[166413] = $3094; //@line 6069 "libraries/glad/glad.cpp"
 }
 $3095 = HEAP8[1252411]|0; //@line 6074 "libraries/glad/glad.cpp"
 $3096 = ($3095<<24>>24)==(0); //@line 6074 "libraries/glad/glad.cpp"
 if (!($3096)) {
  $3097 = (FUNCTION_TABLE_ii[$load & 8191](494972)|0); //@line 6075 "libraries/glad/glad.cpp"
  HEAP32[166414] = $3097; //@line 6075 "libraries/glad/glad.cpp"
 }
 $3098 = HEAP8[1252412]|0; //@line 6080 "libraries/glad/glad.cpp"
 $3099 = ($3098<<24>>24)==(0); //@line 6080 "libraries/glad/glad.cpp"
 if (!($3099)) {
  $3100 = (FUNCTION_TABLE_ii[$load & 8191](495005)|0); //@line 6081 "libraries/glad/glad.cpp"
  HEAP32[166415] = $3100; //@line 6081 "libraries/glad/glad.cpp"
 }
 $3101 = HEAP8[1252413]|0; //@line 6086 "libraries/glad/glad.cpp"
 $3102 = ($3101<<24>>24)==(0); //@line 6086 "libraries/glad/glad.cpp"
 if (!($3102)) {
  $3103 = (FUNCTION_TABLE_ii[$load & 8191](495027)|0); //@line 6087 "libraries/glad/glad.cpp"
  HEAP32[166416] = $3103; //@line 6087 "libraries/glad/glad.cpp"
 }
 $3104 = HEAP8[1252414]|0; //@line 6092 "libraries/glad/glad.cpp"
 $3105 = ($3104<<24>>24)==(0); //@line 6092 "libraries/glad/glad.cpp"
 if (!($3105)) {
  $3106 = (FUNCTION_TABLE_ii[$load & 8191](495048)|0); //@line 6093 "libraries/glad/glad.cpp"
  HEAP32[166417] = $3106; //@line 6093 "libraries/glad/glad.cpp"
  $3107 = (FUNCTION_TABLE_ii[$load & 8191](495065)|0); //@line 6094 "libraries/glad/glad.cpp"
  HEAP32[166418] = $3107; //@line 6094 "libraries/glad/glad.cpp"
 }
 $3108 = HEAP8[1252416]|0; //@line 6100 "libraries/glad/glad.cpp"
 $3109 = ($3108<<24>>24)==(0); //@line 6100 "libraries/glad/glad.cpp"
 if (!($3109)) {
  $3110 = (FUNCTION_TABLE_ii[$load & 8191](495087)|0); //@line 6101 "libraries/glad/glad.cpp"
  HEAP32[166419] = $3110; //@line 6101 "libraries/glad/glad.cpp"
  $3111 = (FUNCTION_TABLE_ii[$load & 8191](495103)|0); //@line 6102 "libraries/glad/glad.cpp"
  HEAP32[166420] = $3111; //@line 6102 "libraries/glad/glad.cpp"
  $3112 = (FUNCTION_TABLE_ii[$load & 8191](495119)|0); //@line 6103 "libraries/glad/glad.cpp"
  HEAP32[166421] = $3112; //@line 6103 "libraries/glad/glad.cpp"
 }
 $3113 = HEAP8[1252419]|0; //@line 6110 "libraries/glad/glad.cpp"
 $3114 = ($3113<<24>>24)==(0); //@line 6110 "libraries/glad/glad.cpp"
 if (!($3114)) {
  $3115 = (FUNCTION_TABLE_ii[$load & 8191](495136)|0); //@line 6111 "libraries/glad/glad.cpp"
  HEAP32[166422] = $3115; //@line 6111 "libraries/glad/glad.cpp"
 }
 $3116 = HEAP8[1252420]|0; //@line 6116 "libraries/glad/glad.cpp"
 $3117 = ($3116<<24>>24)==(0); //@line 6116 "libraries/glad/glad.cpp"
 if (!($3117)) {
  $3118 = (FUNCTION_TABLE_ii[$load & 8191](495152)|0); //@line 6117 "libraries/glad/glad.cpp"
  HEAP32[166423] = $3118; //@line 6117 "libraries/glad/glad.cpp"
  $3119 = (FUNCTION_TABLE_ii[$load & 8191](495176)|0); //@line 6118 "libraries/glad/glad.cpp"
  HEAP32[166424] = $3119; //@line 6118 "libraries/glad/glad.cpp"
 }
 $3120 = HEAP8[1252421]|0; //@line 6123 "libraries/glad/glad.cpp"
 $3121 = ($3120<<24>>24)==(0); //@line 6123 "libraries/glad/glad.cpp"
 if (!($3121)) {
  $3122 = (FUNCTION_TABLE_ii[$load & 8191](495202)|0); //@line 6124 "libraries/glad/glad.cpp"
  HEAP32[166425] = $3122; //@line 6124 "libraries/glad/glad.cpp"
 }
 $3123 = HEAP8[1252423]|0; //@line 6130 "libraries/glad/glad.cpp"
 $3124 = ($3123<<24>>24)==(0); //@line 6130 "libraries/glad/glad.cpp"
 if (!($3124)) {
  $3125 = (FUNCTION_TABLE_ii[$load & 8191](495218)|0); //@line 6131 "libraries/glad/glad.cpp"
  HEAP32[166426] = $3125; //@line 6131 "libraries/glad/glad.cpp"
  $3126 = (FUNCTION_TABLE_ii[$load & 8191](495239)|0); //@line 6132 "libraries/glad/glad.cpp"
  HEAP32[166427] = $3126; //@line 6132 "libraries/glad/glad.cpp"
  $3127 = (FUNCTION_TABLE_ii[$load & 8191](495261)|0); //@line 6133 "libraries/glad/glad.cpp"
  HEAP32[166428] = $3127; //@line 6133 "libraries/glad/glad.cpp"
 }
 $3128 = HEAP8[1252425]|0; //@line 6139 "libraries/glad/glad.cpp"
 $3129 = ($3128<<24>>24)==(0); //@line 6139 "libraries/glad/glad.cpp"
 if (!($3129)) {
  $3130 = (FUNCTION_TABLE_ii[$load & 8191](495281)|0); //@line 6140 "libraries/glad/glad.cpp"
  HEAP32[166429] = $3130; //@line 6140 "libraries/glad/glad.cpp"
  $3131 = (FUNCTION_TABLE_ii[$load & 8191](495298)|0); //@line 6141 "libraries/glad/glad.cpp"
  HEAP32[166430] = $3131; //@line 6141 "libraries/glad/glad.cpp"
  $3132 = (FUNCTION_TABLE_ii[$load & 8191](495312)|0); //@line 6142 "libraries/glad/glad.cpp"
  HEAP32[166431] = $3132; //@line 6142 "libraries/glad/glad.cpp"
  $3133 = (FUNCTION_TABLE_ii[$load & 8191](495324)|0); //@line 6143 "libraries/glad/glad.cpp"
  HEAP32[166432] = $3133; //@line 6143 "libraries/glad/glad.cpp"
  $3134 = (FUNCTION_TABLE_ii[$load & 8191](495338)|0); //@line 6144 "libraries/glad/glad.cpp"
  HEAP32[166433] = $3134; //@line 6144 "libraries/glad/glad.cpp"
  $3135 = (FUNCTION_TABLE_ii[$load & 8191](495353)|0); //@line 6145 "libraries/glad/glad.cpp"
  HEAP32[166434] = $3135; //@line 6145 "libraries/glad/glad.cpp"
  $3136 = (FUNCTION_TABLE_ii[$load & 8191](495369)|0); //@line 6146 "libraries/glad/glad.cpp"
  HEAP32[166435] = $3136; //@line 6146 "libraries/glad/glad.cpp"
 }
 $3137 = HEAP8[1252428]|0; //@line 6153 "libraries/glad/glad.cpp"
 $3138 = ($3137<<24>>24)==(0); //@line 6153 "libraries/glad/glad.cpp"
 if (!($3138)) {
  $3139 = (FUNCTION_TABLE_ii[$load & 8191](495382)|0); //@line 6154 "libraries/glad/glad.cpp"
  HEAP32[166436] = $3139; //@line 6154 "libraries/glad/glad.cpp"
 }
 $3140 = HEAP8[1252431]|0; //@line 6161 "libraries/glad/glad.cpp"
 $3141 = ($3140<<24>>24)==(0); //@line 6161 "libraries/glad/glad.cpp"
 if (!($3141)) {
  $3142 = (FUNCTION_TABLE_ii[$load & 8191](495408)|0); //@line 6162 "libraries/glad/glad.cpp"
  HEAP32[166437] = $3142; //@line 6162 "libraries/glad/glad.cpp"
 }
 $3143 = HEAP8[1252432]|0; //@line 6167 "libraries/glad/glad.cpp"
 $3144 = ($3143<<24>>24)==(0); //@line 6167 "libraries/glad/glad.cpp"
 if (!($3144)) {
  $3145 = (FUNCTION_TABLE_ii[$load & 8191](488400)|0); //@line 6168 "libraries/glad/glad.cpp"
  HEAP32[166136] = $3145; //@line 6168 "libraries/glad/glad.cpp"
  $3146 = (FUNCTION_TABLE_ii[$load & 8191](495428)|0); //@line 6169 "libraries/glad/glad.cpp"
  HEAP32[166438] = $3146; //@line 6169 "libraries/glad/glad.cpp"
  $3147 = (FUNCTION_TABLE_ii[$load & 8191](495456)|0); //@line 6170 "libraries/glad/glad.cpp"
  HEAP32[166439] = $3147; //@line 6170 "libraries/glad/glad.cpp"
  $3148 = (FUNCTION_TABLE_ii[$load & 8191](495487)|0); //@line 6171 "libraries/glad/glad.cpp"
  HEAP32[166440] = $3148; //@line 6171 "libraries/glad/glad.cpp"
 }
 $3149 = HEAP8[1252433]|0; //@line 6176 "libraries/glad/glad.cpp"
 $3150 = ($3149<<24>>24)==(0); //@line 6176 "libraries/glad/glad.cpp"
 if (!($3150)) {
  $3151 = (FUNCTION_TABLE_ii[$load & 8191](495510)|0); //@line 6177 "libraries/glad/glad.cpp"
  HEAP32[166441] = $3151; //@line 6177 "libraries/glad/glad.cpp"
 }
 $3152 = HEAP8[1252434]|0; //@line 6182 "libraries/glad/glad.cpp"
 $3153 = ($3152<<24>>24)==(0); //@line 6182 "libraries/glad/glad.cpp"
 if (!($3153)) {
  $3154 = (FUNCTION_TABLE_ii[$load & 8191](495545)|0); //@line 6183 "libraries/glad/glad.cpp"
  HEAP32[166442] = $3154; //@line 6183 "libraries/glad/glad.cpp"
 }
 $3155 = HEAP8[1252436]|0; //@line 6189 "libraries/glad/glad.cpp"
 $3156 = ($3155<<24>>24)==(0); //@line 6189 "libraries/glad/glad.cpp"
 if (!($3156)) {
  $3157 = (FUNCTION_TABLE_ii[$load & 8191](495588)|0); //@line 6190 "libraries/glad/glad.cpp"
  HEAP32[166443] = $3157; //@line 6190 "libraries/glad/glad.cpp"
  $3158 = (FUNCTION_TABLE_ii[$load & 8191](487810)|0); //@line 6191 "libraries/glad/glad.cpp"
  HEAP32[166110] = $3158; //@line 6191 "libraries/glad/glad.cpp"
  $3159 = (FUNCTION_TABLE_ii[$load & 8191](488934)|0); //@line 6192 "libraries/glad/glad.cpp"
  HEAP32[166159] = $3159; //@line 6192 "libraries/glad/glad.cpp"
  $3160 = (FUNCTION_TABLE_ii[$load & 8191](495611)|0); //@line 6193 "libraries/glad/glad.cpp"
  HEAP32[166444] = $3160; //@line 6193 "libraries/glad/glad.cpp"
 }
 $3161 = HEAP8[1252439]|0; //@line 6200 "libraries/glad/glad.cpp"
 $3162 = ($3161<<24>>24)==(0); //@line 6200 "libraries/glad/glad.cpp"
 if (!($3162)) {
  $3163 = (FUNCTION_TABLE_ii[$load & 8191](495639)|0); //@line 6201 "libraries/glad/glad.cpp"
  HEAP32[166445] = $3163; //@line 6201 "libraries/glad/glad.cpp"
  $3164 = (FUNCTION_TABLE_ii[$load & 8191](495668)|0); //@line 6202 "libraries/glad/glad.cpp"
  HEAP32[166446] = $3164; //@line 6202 "libraries/glad/glad.cpp"
  $3165 = (FUNCTION_TABLE_ii[$load & 8191](495698)|0); //@line 6203 "libraries/glad/glad.cpp"
  HEAP32[166447] = $3165; //@line 6203 "libraries/glad/glad.cpp"
  $3166 = (FUNCTION_TABLE_ii[$load & 8191](495729)|0); //@line 6204 "libraries/glad/glad.cpp"
  HEAP32[166448] = $3166; //@line 6204 "libraries/glad/glad.cpp"
  $3167 = (FUNCTION_TABLE_ii[$load & 8191](495759)|0); //@line 6205 "libraries/glad/glad.cpp"
  HEAP32[166449] = $3167; //@line 6205 "libraries/glad/glad.cpp"
  $3168 = (FUNCTION_TABLE_ii[$load & 8191](495790)|0); //@line 6206 "libraries/glad/glad.cpp"
  HEAP32[166450] = $3168; //@line 6206 "libraries/glad/glad.cpp"
  $3169 = (FUNCTION_TABLE_ii[$load & 8191](495822)|0); //@line 6207 "libraries/glad/glad.cpp"
  HEAP32[166451] = $3169; //@line 6207 "libraries/glad/glad.cpp"
  $3170 = (FUNCTION_TABLE_ii[$load & 8191](495849)|0); //@line 6208 "libraries/glad/glad.cpp"
  HEAP32[166452] = $3170; //@line 6208 "libraries/glad/glad.cpp"
  $3171 = (FUNCTION_TABLE_ii[$load & 8191](495877)|0); //@line 6209 "libraries/glad/glad.cpp"
  HEAP32[166453] = $3171; //@line 6209 "libraries/glad/glad.cpp"
  $3172 = (FUNCTION_TABLE_ii[$load & 8191](495906)|0); //@line 6210 "libraries/glad/glad.cpp"
  HEAP32[166454] = $3172; //@line 6210 "libraries/glad/glad.cpp"
  $3173 = (FUNCTION_TABLE_ii[$load & 8191](495934)|0); //@line 6211 "libraries/glad/glad.cpp"
  HEAP32[166455] = $3173; //@line 6211 "libraries/glad/glad.cpp"
  $3174 = (FUNCTION_TABLE_ii[$load & 8191](495963)|0); //@line 6212 "libraries/glad/glad.cpp"
  HEAP32[166456] = $3174; //@line 6212 "libraries/glad/glad.cpp"
  $3175 = (FUNCTION_TABLE_ii[$load & 8191](495993)|0); //@line 6213 "libraries/glad/glad.cpp"
  HEAP32[166457] = $3175; //@line 6213 "libraries/glad/glad.cpp"
  $3176 = (FUNCTION_TABLE_ii[$load & 8191](496025)|0); //@line 6214 "libraries/glad/glad.cpp"
  HEAP32[166458] = $3176; //@line 6214 "libraries/glad/glad.cpp"
  $3177 = (FUNCTION_TABLE_ii[$load & 8191](496058)|0); //@line 6215 "libraries/glad/glad.cpp"
  HEAP32[166459] = $3177; //@line 6215 "libraries/glad/glad.cpp"
  $3178 = (FUNCTION_TABLE_ii[$load & 8191](496088)|0); //@line 6216 "libraries/glad/glad.cpp"
  HEAP32[166460] = $3178; //@line 6216 "libraries/glad/glad.cpp"
 }
 $3179 = HEAP8[1252440]|0; //@line 6221 "libraries/glad/glad.cpp"
 $3180 = ($3179<<24>>24)==(0); //@line 6221 "libraries/glad/glad.cpp"
 if (!($3180)) {
  $3181 = (FUNCTION_TABLE_ii[$load & 8191](496119)|0); //@line 6222 "libraries/glad/glad.cpp"
  HEAP32[166461] = $3181; //@line 6222 "libraries/glad/glad.cpp"
  $3182 = (FUNCTION_TABLE_ii[$load & 8191](496154)|0); //@line 6223 "libraries/glad/glad.cpp"
  HEAP32[166462] = $3182; //@line 6223 "libraries/glad/glad.cpp"
 }
 $3183 = HEAP8[1252442]|0; //@line 6229 "libraries/glad/glad.cpp"
 $3184 = ($3183<<24>>24)==(0); //@line 6229 "libraries/glad/glad.cpp"
 if (!($3184)) {
  $3185 = (FUNCTION_TABLE_ii[$load & 8191](490997)|0); //@line 6230 "libraries/glad/glad.cpp"
  HEAP32[166248] = $3185; //@line 6230 "libraries/glad/glad.cpp"
  $3186 = (FUNCTION_TABLE_ii[$load & 8191](491013)|0); //@line 6231 "libraries/glad/glad.cpp"
  HEAP32[166249] = $3186; //@line 6231 "libraries/glad/glad.cpp"
  $3187 = (FUNCTION_TABLE_ii[$load & 8191](491029)|0); //@line 6232 "libraries/glad/glad.cpp"
  HEAP32[166250] = $3187; //@line 6232 "libraries/glad/glad.cpp"
  $3188 = (FUNCTION_TABLE_ii[$load & 8191](491045)|0); //@line 6233 "libraries/glad/glad.cpp"
  HEAP32[166251] = $3188; //@line 6233 "libraries/glad/glad.cpp"
  $3189 = (FUNCTION_TABLE_ii[$load & 8191](491061)|0); //@line 6234 "libraries/glad/glad.cpp"
  HEAP32[166252] = $3189; //@line 6234 "libraries/glad/glad.cpp"
  $3190 = (FUNCTION_TABLE_ii[$load & 8191](491078)|0); //@line 6235 "libraries/glad/glad.cpp"
  HEAP32[166253] = $3190; //@line 6235 "libraries/glad/glad.cpp"
  $3191 = (FUNCTION_TABLE_ii[$load & 8191](491095)|0); //@line 6236 "libraries/glad/glad.cpp"
  HEAP32[166254] = $3191; //@line 6236 "libraries/glad/glad.cpp"
  $3192 = (FUNCTION_TABLE_ii[$load & 8191](491112)|0); //@line 6237 "libraries/glad/glad.cpp"
  HEAP32[166255] = $3192; //@line 6237 "libraries/glad/glad.cpp"
  $3193 = (FUNCTION_TABLE_ii[$load & 8191](491129)|0); //@line 6238 "libraries/glad/glad.cpp"
  HEAP32[166256] = $3193; //@line 6238 "libraries/glad/glad.cpp"
  $3194 = (FUNCTION_TABLE_ii[$load & 8191](491146)|0); //@line 6239 "libraries/glad/glad.cpp"
  HEAP32[166257] = $3194; //@line 6239 "libraries/glad/glad.cpp"
  $3195 = (FUNCTION_TABLE_ii[$load & 8191](491163)|0); //@line 6240 "libraries/glad/glad.cpp"
  HEAP32[166258] = $3195; //@line 6240 "libraries/glad/glad.cpp"
  $3196 = (FUNCTION_TABLE_ii[$load & 8191](491180)|0); //@line 6241 "libraries/glad/glad.cpp"
  HEAP32[166259] = $3196; //@line 6241 "libraries/glad/glad.cpp"
  $3197 = (FUNCTION_TABLE_ii[$load & 8191](491197)|0); //@line 6242 "libraries/glad/glad.cpp"
  HEAP32[166260] = $3197; //@line 6242 "libraries/glad/glad.cpp"
  $3198 = (FUNCTION_TABLE_ii[$load & 8191](491215)|0); //@line 6243 "libraries/glad/glad.cpp"
  HEAP32[166261] = $3198; //@line 6243 "libraries/glad/glad.cpp"
  $3199 = (FUNCTION_TABLE_ii[$load & 8191](491233)|0); //@line 6244 "libraries/glad/glad.cpp"
  HEAP32[166262] = $3199; //@line 6244 "libraries/glad/glad.cpp"
  $3200 = (FUNCTION_TABLE_ii[$load & 8191](491251)|0); //@line 6245 "libraries/glad/glad.cpp"
  HEAP32[166263] = $3200; //@line 6245 "libraries/glad/glad.cpp"
  $3201 = (FUNCTION_TABLE_ii[$load & 8191](491269)|0); //@line 6246 "libraries/glad/glad.cpp"
  HEAP32[166264] = $3201; //@line 6246 "libraries/glad/glad.cpp"
  $3202 = (FUNCTION_TABLE_ii[$load & 8191](491308)|0); //@line 6247 "libraries/glad/glad.cpp"
  HEAP32[166266] = $3202; //@line 6247 "libraries/glad/glad.cpp"
  $3203 = (FUNCTION_TABLE_ii[$load & 8191](491331)|0); //@line 6248 "libraries/glad/glad.cpp"
  HEAP32[166267] = $3203; //@line 6248 "libraries/glad/glad.cpp"
  $3204 = (FUNCTION_TABLE_ii[$load & 8191](491354)|0); //@line 6249 "libraries/glad/glad.cpp"
  HEAP32[166268] = $3204; //@line 6249 "libraries/glad/glad.cpp"
  $3205 = (FUNCTION_TABLE_ii[$load & 8191](491377)|0); //@line 6250 "libraries/glad/glad.cpp"
  HEAP32[166269] = $3205; //@line 6250 "libraries/glad/glad.cpp"
  $3206 = (FUNCTION_TABLE_ii[$load & 8191](491400)|0); //@line 6251 "libraries/glad/glad.cpp"
  HEAP32[166270] = $3206; //@line 6251 "libraries/glad/glad.cpp"
  $3207 = (FUNCTION_TABLE_ii[$load & 8191](491424)|0); //@line 6252 "libraries/glad/glad.cpp"
  HEAP32[166271] = $3207; //@line 6252 "libraries/glad/glad.cpp"
  $3208 = (FUNCTION_TABLE_ii[$load & 8191](491448)|0); //@line 6253 "libraries/glad/glad.cpp"
  HEAP32[166272] = $3208; //@line 6253 "libraries/glad/glad.cpp"
  $3209 = (FUNCTION_TABLE_ii[$load & 8191](491472)|0); //@line 6254 "libraries/glad/glad.cpp"
  HEAP32[166273] = $3209; //@line 6254 "libraries/glad/glad.cpp"
  $3210 = (FUNCTION_TABLE_ii[$load & 8191](491496)|0); //@line 6255 "libraries/glad/glad.cpp"
  HEAP32[166274] = $3210; //@line 6255 "libraries/glad/glad.cpp"
  $3211 = (FUNCTION_TABLE_ii[$load & 8191](491520)|0); //@line 6256 "libraries/glad/glad.cpp"
  HEAP32[166275] = $3211; //@line 6256 "libraries/glad/glad.cpp"
  $3212 = (FUNCTION_TABLE_ii[$load & 8191](491544)|0); //@line 6257 "libraries/glad/glad.cpp"
  HEAP32[166276] = $3212; //@line 6257 "libraries/glad/glad.cpp"
  $3213 = (FUNCTION_TABLE_ii[$load & 8191](491568)|0); //@line 6258 "libraries/glad/glad.cpp"
  HEAP32[166277] = $3213; //@line 6258 "libraries/glad/glad.cpp"
  $3214 = (FUNCTION_TABLE_ii[$load & 8191](491592)|0); //@line 6259 "libraries/glad/glad.cpp"
  HEAP32[166278] = $3214; //@line 6259 "libraries/glad/glad.cpp"
  $3215 = (FUNCTION_TABLE_ii[$load & 8191](491617)|0); //@line 6260 "libraries/glad/glad.cpp"
  HEAP32[166279] = $3215; //@line 6260 "libraries/glad/glad.cpp"
  $3216 = (FUNCTION_TABLE_ii[$load & 8191](491642)|0); //@line 6261 "libraries/glad/glad.cpp"
  HEAP32[166280] = $3216; //@line 6261 "libraries/glad/glad.cpp"
  $3217 = (FUNCTION_TABLE_ii[$load & 8191](491667)|0); //@line 6262 "libraries/glad/glad.cpp"
  HEAP32[166281] = $3217; //@line 6262 "libraries/glad/glad.cpp"
 }
 $3218 = HEAP8[1252443]|0; //@line 6267 "libraries/glad/glad.cpp"
 $3219 = ($3218<<24>>24)==(0); //@line 6267 "libraries/glad/glad.cpp"
 if (!($3219)) {
  $3220 = (FUNCTION_TABLE_ii[$load & 8191](496191)|0); //@line 6268 "libraries/glad/glad.cpp"
  HEAP32[166463] = $3220; //@line 6268 "libraries/glad/glad.cpp"
  $3221 = (FUNCTION_TABLE_ii[$load & 8191](496204)|0); //@line 6269 "libraries/glad/glad.cpp"
  HEAP32[166464] = $3221; //@line 6269 "libraries/glad/glad.cpp"
  $3222 = (FUNCTION_TABLE_ii[$load & 8191](496218)|0); //@line 6270 "libraries/glad/glad.cpp"
  HEAP32[166465] = $3222; //@line 6270 "libraries/glad/glad.cpp"
  $3223 = (FUNCTION_TABLE_ii[$load & 8191](496231)|0); //@line 6271 "libraries/glad/glad.cpp"
  HEAP32[166466] = $3223; //@line 6271 "libraries/glad/glad.cpp"
  $3224 = (FUNCTION_TABLE_ii[$load & 8191](496245)|0); //@line 6272 "libraries/glad/glad.cpp"
  HEAP32[166467] = $3224; //@line 6272 "libraries/glad/glad.cpp"
  $3225 = (FUNCTION_TABLE_ii[$load & 8191](496258)|0); //@line 6273 "libraries/glad/glad.cpp"
  HEAP32[166468] = $3225; //@line 6273 "libraries/glad/glad.cpp"
  $3226 = (FUNCTION_TABLE_ii[$load & 8191](496272)|0); //@line 6274 "libraries/glad/glad.cpp"
  HEAP32[166469] = $3226; //@line 6274 "libraries/glad/glad.cpp"
  $3227 = (FUNCTION_TABLE_ii[$load & 8191](496285)|0); //@line 6275 "libraries/glad/glad.cpp"
  HEAP32[166470] = $3227; //@line 6275 "libraries/glad/glad.cpp"
  $3228 = (FUNCTION_TABLE_ii[$load & 8191](496299)|0); //@line 6276 "libraries/glad/glad.cpp"
  HEAP32[166471] = $3228; //@line 6276 "libraries/glad/glad.cpp"
  $3229 = (FUNCTION_TABLE_ii[$load & 8191](496311)|0); //@line 6277 "libraries/glad/glad.cpp"
  HEAP32[166472] = $3229; //@line 6277 "libraries/glad/glad.cpp"
  $3230 = (FUNCTION_TABLE_ii[$load & 8191](496324)|0); //@line 6278 "libraries/glad/glad.cpp"
  HEAP32[166473] = $3230; //@line 6278 "libraries/glad/glad.cpp"
  $3231 = (FUNCTION_TABLE_ii[$load & 8191](496336)|0); //@line 6279 "libraries/glad/glad.cpp"
  HEAP32[166474] = $3231; //@line 6279 "libraries/glad/glad.cpp"
  $3232 = (FUNCTION_TABLE_ii[$load & 8191](496349)|0); //@line 6280 "libraries/glad/glad.cpp"
  HEAP32[166475] = $3232; //@line 6280 "libraries/glad/glad.cpp"
  $3233 = (FUNCTION_TABLE_ii[$load & 8191](496364)|0); //@line 6281 "libraries/glad/glad.cpp"
  HEAP32[166476] = $3233; //@line 6281 "libraries/glad/glad.cpp"
  $3234 = (FUNCTION_TABLE_ii[$load & 8191](496380)|0); //@line 6282 "libraries/glad/glad.cpp"
  HEAP32[166477] = $3234; //@line 6282 "libraries/glad/glad.cpp"
  $3235 = (FUNCTION_TABLE_ii[$load & 8191](496395)|0); //@line 6283 "libraries/glad/glad.cpp"
  HEAP32[166478] = $3235; //@line 6283 "libraries/glad/glad.cpp"
  $3236 = (FUNCTION_TABLE_ii[$load & 8191](496411)|0); //@line 6284 "libraries/glad/glad.cpp"
  HEAP32[166479] = $3236; //@line 6284 "libraries/glad/glad.cpp"
  $3237 = (FUNCTION_TABLE_ii[$load & 8191](496426)|0); //@line 6285 "libraries/glad/glad.cpp"
  HEAP32[166480] = $3237; //@line 6285 "libraries/glad/glad.cpp"
  $3238 = (FUNCTION_TABLE_ii[$load & 8191](496442)|0); //@line 6286 "libraries/glad/glad.cpp"
  HEAP32[166481] = $3238; //@line 6286 "libraries/glad/glad.cpp"
  $3239 = (FUNCTION_TABLE_ii[$load & 8191](496457)|0); //@line 6287 "libraries/glad/glad.cpp"
  HEAP32[166482] = $3239; //@line 6287 "libraries/glad/glad.cpp"
  $3240 = (FUNCTION_TABLE_ii[$load & 8191](496473)|0); //@line 6288 "libraries/glad/glad.cpp"
  HEAP32[166483] = $3240; //@line 6288 "libraries/glad/glad.cpp"
  $3241 = (FUNCTION_TABLE_ii[$load & 8191](496493)|0); //@line 6289 "libraries/glad/glad.cpp"
  HEAP32[166484] = $3241; //@line 6289 "libraries/glad/glad.cpp"
  $3242 = (FUNCTION_TABLE_ii[$load & 8191](496514)|0); //@line 6290 "libraries/glad/glad.cpp"
  HEAP32[166485] = $3242; //@line 6290 "libraries/glad/glad.cpp"
  $3243 = (FUNCTION_TABLE_ii[$load & 8191](496534)|0); //@line 6291 "libraries/glad/glad.cpp"
  HEAP32[166486] = $3243; //@line 6291 "libraries/glad/glad.cpp"
  $3244 = (FUNCTION_TABLE_ii[$load & 8191](496555)|0); //@line 6292 "libraries/glad/glad.cpp"
  HEAP32[166487] = $3244; //@line 6292 "libraries/glad/glad.cpp"
  $3245 = (FUNCTION_TABLE_ii[$load & 8191](496575)|0); //@line 6293 "libraries/glad/glad.cpp"
  HEAP32[166488] = $3245; //@line 6293 "libraries/glad/glad.cpp"
  $3246 = (FUNCTION_TABLE_ii[$load & 8191](496596)|0); //@line 6294 "libraries/glad/glad.cpp"
  HEAP32[166489] = $3246; //@line 6294 "libraries/glad/glad.cpp"
  $3247 = (FUNCTION_TABLE_ii[$load & 8191](496616)|0); //@line 6295 "libraries/glad/glad.cpp"
  HEAP32[166490] = $3247; //@line 6295 "libraries/glad/glad.cpp"
  $3248 = (FUNCTION_TABLE_ii[$load & 8191](496637)|0); //@line 6296 "libraries/glad/glad.cpp"
  HEAP32[166491] = $3248; //@line 6296 "libraries/glad/glad.cpp"
  $3249 = (FUNCTION_TABLE_ii[$load & 8191](496651)|0); //@line 6297 "libraries/glad/glad.cpp"
  HEAP32[166492] = $3249; //@line 6297 "libraries/glad/glad.cpp"
  $3250 = (FUNCTION_TABLE_ii[$load & 8191](496666)|0); //@line 6298 "libraries/glad/glad.cpp"
  HEAP32[166493] = $3250; //@line 6298 "libraries/glad/glad.cpp"
  $3251 = (FUNCTION_TABLE_ii[$load & 8191](496687)|0); //@line 6299 "libraries/glad/glad.cpp"
  HEAP32[166494] = $3251; //@line 6299 "libraries/glad/glad.cpp"
  $3252 = (FUNCTION_TABLE_ii[$load & 8191](496709)|0); //@line 6300 "libraries/glad/glad.cpp"
  HEAP32[166495] = $3252; //@line 6300 "libraries/glad/glad.cpp"
  $3253 = (FUNCTION_TABLE_ii[$load & 8191](496727)|0); //@line 6301 "libraries/glad/glad.cpp"
  HEAP32[166496] = $3253; //@line 6301 "libraries/glad/glad.cpp"
  $3254 = (FUNCTION_TABLE_ii[$load & 8191](496746)|0); //@line 6302 "libraries/glad/glad.cpp"
  HEAP32[166497] = $3254; //@line 6302 "libraries/glad/glad.cpp"
  $3255 = (FUNCTION_TABLE_ii[$load & 8191](496765)|0); //@line 6303 "libraries/glad/glad.cpp"
  HEAP32[166498] = $3255; //@line 6303 "libraries/glad/glad.cpp"
  $3256 = (FUNCTION_TABLE_ii[$load & 8191](496785)|0); //@line 6304 "libraries/glad/glad.cpp"
  HEAP32[166499] = $3256; //@line 6304 "libraries/glad/glad.cpp"
  $3257 = (FUNCTION_TABLE_ii[$load & 8191](496804)|0); //@line 6305 "libraries/glad/glad.cpp"
  HEAP32[166500] = $3257; //@line 6305 "libraries/glad/glad.cpp"
  $3258 = (FUNCTION_TABLE_ii[$load & 8191](496824)|0); //@line 6306 "libraries/glad/glad.cpp"
  HEAP32[166501] = $3258; //@line 6306 "libraries/glad/glad.cpp"
  $3259 = (FUNCTION_TABLE_ii[$load & 8191](496843)|0); //@line 6307 "libraries/glad/glad.cpp"
  HEAP32[166502] = $3259; //@line 6307 "libraries/glad/glad.cpp"
  $3260 = (FUNCTION_TABLE_ii[$load & 8191](496863)|0); //@line 6308 "libraries/glad/glad.cpp"
  HEAP32[166503] = $3260; //@line 6308 "libraries/glad/glad.cpp"
  $3261 = (FUNCTION_TABLE_ii[$load & 8191](496882)|0); //@line 6309 "libraries/glad/glad.cpp"
  HEAP32[166504] = $3261; //@line 6309 "libraries/glad/glad.cpp"
  $3262 = (FUNCTION_TABLE_ii[$load & 8191](496902)|0); //@line 6310 "libraries/glad/glad.cpp"
  HEAP32[166505] = $3262; //@line 6310 "libraries/glad/glad.cpp"
  $3263 = (FUNCTION_TABLE_ii[$load & 8191](496923)|0); //@line 6311 "libraries/glad/glad.cpp"
  HEAP32[166506] = $3263; //@line 6311 "libraries/glad/glad.cpp"
  $3264 = (FUNCTION_TABLE_ii[$load & 8191](496944)|0); //@line 6312 "libraries/glad/glad.cpp"
  HEAP32[166507] = $3264; //@line 6312 "libraries/glad/glad.cpp"
  $3265 = (FUNCTION_TABLE_ii[$load & 8191](496965)|0); //@line 6313 "libraries/glad/glad.cpp"
  HEAP32[166508] = $3265; //@line 6313 "libraries/glad/glad.cpp"
 }
 $3266 = HEAP8[1252445]|0; //@line 6319 "libraries/glad/glad.cpp"
 $3267 = ($3266<<24>>24)==(0); //@line 6319 "libraries/glad/glad.cpp"
 if (!($3267)) {
  $3268 = (FUNCTION_TABLE_ii[$load & 8191](496986)|0); //@line 6320 "libraries/glad/glad.cpp"
  HEAP32[166509] = $3268; //@line 6320 "libraries/glad/glad.cpp"
 }
 $3269 = HEAP8[1252446]|0; //@line 6325 "libraries/glad/glad.cpp"
 $3270 = ($3269<<24>>24)==(0); //@line 6325 "libraries/glad/glad.cpp"
 if (!($3270)) {
  $3271 = (FUNCTION_TABLE_ii[$load & 8191](497010)|0); //@line 6326 "libraries/glad/glad.cpp"
  HEAP32[166510] = $3271; //@line 6326 "libraries/glad/glad.cpp"
 }
 $3272 = HEAP8[1252448]|0; //@line 6332 "libraries/glad/glad.cpp"
 $3273 = ($3272<<24>>24)==(0); //@line 6332 "libraries/glad/glad.cpp"
 if (!($3273)) {
  $3274 = (FUNCTION_TABLE_ii[$load & 8191](497040)|0); //@line 6333 "libraries/glad/glad.cpp"
  HEAP32[166511] = $3274; //@line 6333 "libraries/glad/glad.cpp"
  $3275 = (FUNCTION_TABLE_ii[$load & 8191](497063)|0); //@line 6334 "libraries/glad/glad.cpp"
  HEAP32[166512] = $3275; //@line 6334 "libraries/glad/glad.cpp"
  $3276 = (FUNCTION_TABLE_ii[$load & 8191](497086)|0); //@line 6335 "libraries/glad/glad.cpp"
  HEAP32[166513] = $3276; //@line 6335 "libraries/glad/glad.cpp"
  $3277 = (FUNCTION_TABLE_ii[$load & 8191](497109)|0); //@line 6336 "libraries/glad/glad.cpp"
  HEAP32[166514] = $3277; //@line 6336 "libraries/glad/glad.cpp"
  $3278 = (FUNCTION_TABLE_ii[$load & 8191](497132)|0); //@line 6337 "libraries/glad/glad.cpp"
  HEAP32[166515] = $3278; //@line 6337 "libraries/glad/glad.cpp"
  $3279 = (FUNCTION_TABLE_ii[$load & 8191](497155)|0); //@line 6338 "libraries/glad/glad.cpp"
  HEAP32[166516] = $3279; //@line 6338 "libraries/glad/glad.cpp"
 }
 $3280 = HEAP8[1252449]|0; //@line 6343 "libraries/glad/glad.cpp"
 $3281 = ($3280<<24>>24)==(0); //@line 6343 "libraries/glad/glad.cpp"
 if (!($3281)) {
  $3282 = (FUNCTION_TABLE_ii[$load & 8191](497178)|0); //@line 6344 "libraries/glad/glad.cpp"
  HEAP32[166517] = $3282; //@line 6344 "libraries/glad/glad.cpp"
  $3283 = (FUNCTION_TABLE_ii[$load & 8191](497202)|0); //@line 6345 "libraries/glad/glad.cpp"
  HEAP32[166518] = $3283; //@line 6345 "libraries/glad/glad.cpp"
  $3284 = (FUNCTION_TABLE_ii[$load & 8191](497229)|0); //@line 6346 "libraries/glad/glad.cpp"
  HEAP32[166519] = $3284; //@line 6346 "libraries/glad/glad.cpp"
  $3285 = (FUNCTION_TABLE_ii[$load & 8191](497250)|0); //@line 6347 "libraries/glad/glad.cpp"
  HEAP32[166520] = $3285; //@line 6347 "libraries/glad/glad.cpp"
  $3286 = (FUNCTION_TABLE_ii[$load & 8191](497274)|0); //@line 6348 "libraries/glad/glad.cpp"
  HEAP32[166521] = $3286; //@line 6348 "libraries/glad/glad.cpp"
  $3287 = (FUNCTION_TABLE_ii[$load & 8191](497296)|0); //@line 6349 "libraries/glad/glad.cpp"
  HEAP32[166522] = $3287; //@line 6349 "libraries/glad/glad.cpp"
  $3288 = (FUNCTION_TABLE_ii[$load & 8191](497320)|0); //@line 6350 "libraries/glad/glad.cpp"
  HEAP32[166523] = $3288; //@line 6350 "libraries/glad/glad.cpp"
 }
 $3289 = HEAP8[1252450]|0; //@line 6355 "libraries/glad/glad.cpp"
 $3290 = ($3289<<24>>24)==(0); //@line 6355 "libraries/glad/glad.cpp"
 if (!($3290)) {
  $3291 = (FUNCTION_TABLE_ii[$load & 8191](497345)|0); //@line 6356 "libraries/glad/glad.cpp"
  HEAP32[166524] = $3291; //@line 6356 "libraries/glad/glad.cpp"
  $3292 = (FUNCTION_TABLE_ii[$load & 8191](497375)|0); //@line 6357 "libraries/glad/glad.cpp"
  HEAP32[166525] = $3292; //@line 6357 "libraries/glad/glad.cpp"
  $3293 = (FUNCTION_TABLE_ii[$load & 8191](497406)|0); //@line 6358 "libraries/glad/glad.cpp"
  HEAP32[166526] = $3293; //@line 6358 "libraries/glad/glad.cpp"
 }
 $3294 = HEAP8[1252452]|0; //@line 6364 "libraries/glad/glad.cpp"
 $3295 = ($3294<<24>>24)==(0); //@line 6364 "libraries/glad/glad.cpp"
 if (!($3295)) {
  $3296 = (FUNCTION_TABLE_ii[$load & 8191](497438)|0); //@line 6365 "libraries/glad/glad.cpp"
  HEAP32[166527] = $3296; //@line 6365 "libraries/glad/glad.cpp"
  $3297 = (FUNCTION_TABLE_ii[$load & 8191](497451)|0); //@line 6366 "libraries/glad/glad.cpp"
  HEAP32[166528] = $3297; //@line 6366 "libraries/glad/glad.cpp"
  $3298 = (FUNCTION_TABLE_ii[$load & 8191](497467)|0); //@line 6367 "libraries/glad/glad.cpp"
  HEAP32[166529] = $3298; //@line 6367 "libraries/glad/glad.cpp"
  $3299 = (FUNCTION_TABLE_ii[$load & 8191](497478)|0); //@line 6368 "libraries/glad/glad.cpp"
  HEAP32[166530] = $3299; //@line 6368 "libraries/glad/glad.cpp"
  $3300 = (FUNCTION_TABLE_ii[$load & 8191](497495)|0); //@line 6369 "libraries/glad/glad.cpp"
  HEAP32[166531] = $3300; //@line 6369 "libraries/glad/glad.cpp"
  $3301 = (FUNCTION_TABLE_ii[$load & 8191](497510)|0); //@line 6370 "libraries/glad/glad.cpp"
  HEAP32[166532] = $3301; //@line 6370 "libraries/glad/glad.cpp"
  $3302 = (FUNCTION_TABLE_ii[$load & 8191](497530)|0); //@line 6371 "libraries/glad/glad.cpp"
  HEAP32[166533] = $3302; //@line 6371 "libraries/glad/glad.cpp"
  $3303 = (FUNCTION_TABLE_ii[$load & 8191](497548)|0); //@line 6372 "libraries/glad/glad.cpp"
  HEAP32[166534] = $3303; //@line 6372 "libraries/glad/glad.cpp"
  $3304 = (FUNCTION_TABLE_ii[$load & 8191](497563)|0); //@line 6373 "libraries/glad/glad.cpp"
  HEAP32[166535] = $3304; //@line 6373 "libraries/glad/glad.cpp"
  $3305 = (FUNCTION_TABLE_ii[$load & 8191](497578)|0); //@line 6374 "libraries/glad/glad.cpp"
  HEAP32[166536] = $3305; //@line 6374 "libraries/glad/glad.cpp"
  $3306 = (FUNCTION_TABLE_ii[$load & 8191](497597)|0); //@line 6375 "libraries/glad/glad.cpp"
  HEAP32[166537] = $3306; //@line 6375 "libraries/glad/glad.cpp"
  $3307 = (FUNCTION_TABLE_ii[$load & 8191](497613)|0); //@line 6376 "libraries/glad/glad.cpp"
  HEAP32[166538] = $3307; //@line 6376 "libraries/glad/glad.cpp"
  $3308 = (FUNCTION_TABLE_ii[$load & 8191](497626)|0); //@line 6377 "libraries/glad/glad.cpp"
  HEAP32[166539] = $3308; //@line 6377 "libraries/glad/glad.cpp"
  $3309 = (FUNCTION_TABLE_ii[$load & 8191](497647)|0); //@line 6378 "libraries/glad/glad.cpp"
  HEAP32[166540] = $3309; //@line 6378 "libraries/glad/glad.cpp"
  $3310 = (FUNCTION_TABLE_ii[$load & 8191](497665)|0); //@line 6379 "libraries/glad/glad.cpp"
  HEAP32[166541] = $3310; //@line 6379 "libraries/glad/glad.cpp"
  $3311 = (FUNCTION_TABLE_ii[$load & 8191](497685)|0); //@line 6380 "libraries/glad/glad.cpp"
  HEAP32[166542] = $3311; //@line 6380 "libraries/glad/glad.cpp"
  $3312 = (FUNCTION_TABLE_ii[$load & 8191](497704)|0); //@line 6381 "libraries/glad/glad.cpp"
  HEAP32[166543] = $3312; //@line 6381 "libraries/glad/glad.cpp"
  $3313 = (FUNCTION_TABLE_ii[$load & 8191](497724)|0); //@line 6382 "libraries/glad/glad.cpp"
  HEAP32[166544] = $3313; //@line 6382 "libraries/glad/glad.cpp"
  $3314 = (FUNCTION_TABLE_ii[$load & 8191](497743)|0); //@line 6383 "libraries/glad/glad.cpp"
  HEAP32[166545] = $3314; //@line 6383 "libraries/glad/glad.cpp"
  $3315 = (FUNCTION_TABLE_ii[$load & 8191](497761)|0); //@line 6384 "libraries/glad/glad.cpp"
  HEAP32[166546] = $3315; //@line 6384 "libraries/glad/glad.cpp"
  $3316 = (FUNCTION_TABLE_ii[$load & 8191](497781)|0); //@line 6385 "libraries/glad/glad.cpp"
  HEAP32[166547] = $3316; //@line 6385 "libraries/glad/glad.cpp"
  $3317 = (FUNCTION_TABLE_ii[$load & 8191](497808)|0); //@line 6386 "libraries/glad/glad.cpp"
  HEAP32[166548] = $3317; //@line 6386 "libraries/glad/glad.cpp"
  $3318 = (FUNCTION_TABLE_ii[$load & 8191](497828)|0); //@line 6387 "libraries/glad/glad.cpp"
  HEAP32[166549] = $3318; //@line 6387 "libraries/glad/glad.cpp"
  $3319 = (FUNCTION_TABLE_ii[$load & 8191](497850)|0); //@line 6388 "libraries/glad/glad.cpp"
  HEAP32[166550] = $3319; //@line 6388 "libraries/glad/glad.cpp"
  $3320 = (FUNCTION_TABLE_ii[$load & 8191](497879)|0); //@line 6389 "libraries/glad/glad.cpp"
  HEAP32[166551] = $3320; //@line 6389 "libraries/glad/glad.cpp"
  $3321 = (FUNCTION_TABLE_ii[$load & 8191](497910)|0); //@line 6390 "libraries/glad/glad.cpp"
  HEAP32[166552] = $3321; //@line 6390 "libraries/glad/glad.cpp"
  $3322 = (FUNCTION_TABLE_ii[$load & 8191](497933)|0); //@line 6391 "libraries/glad/glad.cpp"
  HEAP32[166553] = $3322; //@line 6391 "libraries/glad/glad.cpp"
  $3323 = (FUNCTION_TABLE_ii[$load & 8191](497951)|0); //@line 6392 "libraries/glad/glad.cpp"
  HEAP32[166554] = $3323; //@line 6392 "libraries/glad/glad.cpp"
  $3324 = (FUNCTION_TABLE_ii[$load & 8191](497971)|0); //@line 6393 "libraries/glad/glad.cpp"
  HEAP32[166555] = $3324; //@line 6393 "libraries/glad/glad.cpp"
  $3325 = (FUNCTION_TABLE_ii[$load & 8191](497998)|0); //@line 6394 "libraries/glad/glad.cpp"
  HEAP32[166556] = $3325; //@line 6394 "libraries/glad/glad.cpp"
  $3326 = (FUNCTION_TABLE_ii[$load & 8191](498027)|0); //@line 6395 "libraries/glad/glad.cpp"
  HEAP32[166557] = $3326; //@line 6395 "libraries/glad/glad.cpp"
  $3327 = (FUNCTION_TABLE_ii[$load & 8191](498050)|0); //@line 6396 "libraries/glad/glad.cpp"
  HEAP32[166558] = $3327; //@line 6396 "libraries/glad/glad.cpp"
  $3328 = (FUNCTION_TABLE_ii[$load & 8191](498073)|0); //@line 6397 "libraries/glad/glad.cpp"
  HEAP32[166559] = $3328; //@line 6397 "libraries/glad/glad.cpp"
  $3329 = (FUNCTION_TABLE_ii[$load & 8191](498093)|0); //@line 6398 "libraries/glad/glad.cpp"
  HEAP32[166560] = $3329; //@line 6398 "libraries/glad/glad.cpp"
  $3330 = (FUNCTION_TABLE_ii[$load & 8191](498111)|0); //@line 6399 "libraries/glad/glad.cpp"
  HEAP32[166561] = $3330; //@line 6399 "libraries/glad/glad.cpp"
  $3331 = (FUNCTION_TABLE_ii[$load & 8191](498132)|0); //@line 6400 "libraries/glad/glad.cpp"
  HEAP32[166562] = $3331; //@line 6400 "libraries/glad/glad.cpp"
  $3332 = (FUNCTION_TABLE_ii[$load & 8191](498151)|0); //@line 6401 "libraries/glad/glad.cpp"
  HEAP32[166563] = $3332; //@line 6401 "libraries/glad/glad.cpp"
  $3333 = (FUNCTION_TABLE_ii[$load & 8191](498174)|0); //@line 6402 "libraries/glad/glad.cpp"
  HEAP32[166564] = $3333; //@line 6402 "libraries/glad/glad.cpp"
  $3334 = (FUNCTION_TABLE_ii[$load & 8191](498193)|0); //@line 6403 "libraries/glad/glad.cpp"
  HEAP32[166565] = $3334; //@line 6403 "libraries/glad/glad.cpp"
  $3335 = (FUNCTION_TABLE_ii[$load & 8191](498215)|0); //@line 6404 "libraries/glad/glad.cpp"
  HEAP32[166566] = $3335; //@line 6404 "libraries/glad/glad.cpp"
  $3336 = (FUNCTION_TABLE_ii[$load & 8191](498239)|0); //@line 6405 "libraries/glad/glad.cpp"
  HEAP32[166567] = $3336; //@line 6405 "libraries/glad/glad.cpp"
  $3337 = (FUNCTION_TABLE_ii[$load & 8191](498257)|0); //@line 6406 "libraries/glad/glad.cpp"
  HEAP32[166568] = $3337; //@line 6406 "libraries/glad/glad.cpp"
  $3338 = (FUNCTION_TABLE_ii[$load & 8191](498276)|0); //@line 6407 "libraries/glad/glad.cpp"
  HEAP32[166569] = $3338; //@line 6407 "libraries/glad/glad.cpp"
  $3339 = (FUNCTION_TABLE_ii[$load & 8191](498295)|0); //@line 6408 "libraries/glad/glad.cpp"
  HEAP32[166570] = $3339; //@line 6408 "libraries/glad/glad.cpp"
  $3340 = (FUNCTION_TABLE_ii[$load & 8191](498314)|0); //@line 6409 "libraries/glad/glad.cpp"
  HEAP32[166571] = $3340; //@line 6409 "libraries/glad/glad.cpp"
  $3341 = (FUNCTION_TABLE_ii[$load & 8191](498342)|0); //@line 6410 "libraries/glad/glad.cpp"
  HEAP32[166572] = $3341; //@line 6410 "libraries/glad/glad.cpp"
  $3342 = (FUNCTION_TABLE_ii[$load & 8191](498361)|0); //@line 6411 "libraries/glad/glad.cpp"
  HEAP32[166573] = $3342; //@line 6411 "libraries/glad/glad.cpp"
  $3343 = (FUNCTION_TABLE_ii[$load & 8191](498380)|0); //@line 6412 "libraries/glad/glad.cpp"
  HEAP32[166574] = $3343; //@line 6412 "libraries/glad/glad.cpp"
  $3344 = (FUNCTION_TABLE_ii[$load & 8191](498408)|0); //@line 6413 "libraries/glad/glad.cpp"
  HEAP32[166575] = $3344; //@line 6413 "libraries/glad/glad.cpp"
  $3345 = (FUNCTION_TABLE_ii[$load & 8191](498437)|0); //@line 6414 "libraries/glad/glad.cpp"
  HEAP32[166576] = $3345; //@line 6414 "libraries/glad/glad.cpp"
  $3346 = (FUNCTION_TABLE_ii[$load & 8191](498468)|0); //@line 6415 "libraries/glad/glad.cpp"
  HEAP32[166577] = $3346; //@line 6415 "libraries/glad/glad.cpp"
  $3347 = (FUNCTION_TABLE_ii[$load & 8191](498506)|0); //@line 6416 "libraries/glad/glad.cpp"
  HEAP32[166578] = $3347; //@line 6416 "libraries/glad/glad.cpp"
  $3348 = (FUNCTION_TABLE_ii[$load & 8191](498546)|0); //@line 6417 "libraries/glad/glad.cpp"
  HEAP32[166579] = $3348; //@line 6417 "libraries/glad/glad.cpp"
  $3349 = (FUNCTION_TABLE_ii[$load & 8191](498570)|0); //@line 6418 "libraries/glad/glad.cpp"
  HEAP32[166580] = $3349; //@line 6418 "libraries/glad/glad.cpp"
  $3350 = (FUNCTION_TABLE_ii[$load & 8191](498594)|0); //@line 6419 "libraries/glad/glad.cpp"
  HEAP32[166581] = $3350; //@line 6419 "libraries/glad/glad.cpp"
  $3351 = (FUNCTION_TABLE_ii[$load & 8191](498624)|0); //@line 6420 "libraries/glad/glad.cpp"
  HEAP32[166582] = $3351; //@line 6420 "libraries/glad/glad.cpp"
  $3352 = (FUNCTION_TABLE_ii[$load & 8191](498656)|0); //@line 6421 "libraries/glad/glad.cpp"
  HEAP32[166583] = $3352; //@line 6421 "libraries/glad/glad.cpp"
  $3353 = (FUNCTION_TABLE_ii[$load & 8191](498681)|0); //@line 6422 "libraries/glad/glad.cpp"
  HEAP32[166584] = $3353; //@line 6422 "libraries/glad/glad.cpp"
  $3354 = (FUNCTION_TABLE_ii[$load & 8191](498698)|0); //@line 6423 "libraries/glad/glad.cpp"
  HEAP32[166585] = $3354; //@line 6423 "libraries/glad/glad.cpp"
  $3355 = (FUNCTION_TABLE_ii[$load & 8191](498713)|0); //@line 6424 "libraries/glad/glad.cpp"
  HEAP32[166586] = $3355; //@line 6424 "libraries/glad/glad.cpp"
  $3356 = (FUNCTION_TABLE_ii[$load & 8191](498728)|0); //@line 6425 "libraries/glad/glad.cpp"
  HEAP32[166587] = $3356; //@line 6425 "libraries/glad/glad.cpp"
  $3357 = (FUNCTION_TABLE_ii[$load & 8191](498750)|0); //@line 6426 "libraries/glad/glad.cpp"
  HEAP32[166588] = $3357; //@line 6426 "libraries/glad/glad.cpp"
  $3358 = (FUNCTION_TABLE_ii[$load & 8191](498772)|0); //@line 6427 "libraries/glad/glad.cpp"
  HEAP32[166589] = $3358; //@line 6427 "libraries/glad/glad.cpp"
  $3359 = (FUNCTION_TABLE_ii[$load & 8191](498792)|0); //@line 6428 "libraries/glad/glad.cpp"
  HEAP32[166590] = $3359; //@line 6428 "libraries/glad/glad.cpp"
 }
 $3360 = HEAP8[1252454]|0; //@line 6434 "libraries/glad/glad.cpp"
 $3361 = ($3360<<24>>24)==(0); //@line 6434 "libraries/glad/glad.cpp"
 if (!($3361)) {
  $3362 = (FUNCTION_TABLE_ii[$load & 8191](498812)|0); //@line 6435 "libraries/glad/glad.cpp"
  HEAP32[166591] = $3362; //@line 6435 "libraries/glad/glad.cpp"
 }
 $3363 = HEAP8[1252455]|0; //@line 6440 "libraries/glad/glad.cpp"
 $3364 = ($3363<<24>>24)==(0); //@line 6440 "libraries/glad/glad.cpp"
 if (!($3364)) {
  $3365 = (FUNCTION_TABLE_ii[$load & 8191](498828)|0); //@line 6441 "libraries/glad/glad.cpp"
  HEAP32[166592] = $3365; //@line 6441 "libraries/glad/glad.cpp"
  $3366 = (FUNCTION_TABLE_ii[$load & 8191](498850)|0); //@line 6442 "libraries/glad/glad.cpp"
  HEAP32[166593] = $3366; //@line 6442 "libraries/glad/glad.cpp"
  $3367 = (FUNCTION_TABLE_ii[$load & 8191](498875)|0); //@line 6443 "libraries/glad/glad.cpp"
  HEAP32[166594] = $3367; //@line 6443 "libraries/glad/glad.cpp"
  $3368 = (FUNCTION_TABLE_ii[$load & 8191](498890)|0); //@line 6444 "libraries/glad/glad.cpp"
  HEAP32[166595] = $3368; //@line 6444 "libraries/glad/glad.cpp"
  $3369 = (FUNCTION_TABLE_ii[$load & 8191](498906)|0); //@line 6445 "libraries/glad/glad.cpp"
  HEAP32[166596] = $3369; //@line 6445 "libraries/glad/glad.cpp"
  $3370 = (FUNCTION_TABLE_ii[$load & 8191](498923)|0); //@line 6446 "libraries/glad/glad.cpp"
  HEAP32[166597] = $3370; //@line 6446 "libraries/glad/glad.cpp"
 }
 $3371 = HEAP8[1252456]|0; //@line 6451 "libraries/glad/glad.cpp"
 $3372 = ($3371<<24>>24)==(0); //@line 6451 "libraries/glad/glad.cpp"
 if (!($3372)) {
  $3373 = (FUNCTION_TABLE_ii[$load & 8191](498941)|0); //@line 6452 "libraries/glad/glad.cpp"
  HEAP32[166598] = $3373; //@line 6452 "libraries/glad/glad.cpp"
  $3374 = (FUNCTION_TABLE_ii[$load & 8191](498962)|0); //@line 6453 "libraries/glad/glad.cpp"
  HEAP32[166599] = $3374; //@line 6453 "libraries/glad/glad.cpp"
 }
 $3375 = HEAP8[1252457]|0; //@line 6458 "libraries/glad/glad.cpp"
 $3376 = ($3375<<24>>24)==(0); //@line 6458 "libraries/glad/glad.cpp"
 if (!($3376)) {
  $3377 = (FUNCTION_TABLE_ii[$load & 8191](498988)|0); //@line 6459 "libraries/glad/glad.cpp"
  HEAP32[166600] = $3377; //@line 6459 "libraries/glad/glad.cpp"
 }
 $3378 = HEAP8[1252462]|0; //@line 6468 "libraries/glad/glad.cpp"
 $3379 = ($3378<<24>>24)==(0); //@line 6468 "libraries/glad/glad.cpp"
 if (!($3379)) {
  $3380 = (FUNCTION_TABLE_ii[$load & 8191](499003)|0); //@line 6469 "libraries/glad/glad.cpp"
  HEAP32[166601] = $3380; //@line 6469 "libraries/glad/glad.cpp"
  $3381 = (FUNCTION_TABLE_ii[$load & 8191](499036)|0); //@line 6470 "libraries/glad/glad.cpp"
  HEAP32[166602] = $3381; //@line 6470 "libraries/glad/glad.cpp"
  $3382 = (FUNCTION_TABLE_ii[$load & 8191](499074)|0); //@line 6471 "libraries/glad/glad.cpp"
  HEAP32[166603] = $3382; //@line 6471 "libraries/glad/glad.cpp"
 }
 $3383 = HEAP8[1252468]|0; //@line 6481 "libraries/glad/glad.cpp"
 $3384 = ($3383<<24>>24)==(0); //@line 6481 "libraries/glad/glad.cpp"
 if (!($3384)) {
  $3385 = (FUNCTION_TABLE_ii[$load & 8191](499097)|0); //@line 6482 "libraries/glad/glad.cpp"
  HEAP32[166604] = $3385; //@line 6482 "libraries/glad/glad.cpp"
  $3386 = (FUNCTION_TABLE_ii[$load & 8191](499120)|0); //@line 6483 "libraries/glad/glad.cpp"
  HEAP32[166605] = $3386; //@line 6483 "libraries/glad/glad.cpp"
  $3387 = (FUNCTION_TABLE_ii[$load & 8191](499146)|0); //@line 6484 "libraries/glad/glad.cpp"
  HEAP32[166606] = $3387; //@line 6484 "libraries/glad/glad.cpp"
  $3388 = (FUNCTION_TABLE_ii[$load & 8191](499167)|0); //@line 6485 "libraries/glad/glad.cpp"
  HEAP32[166607] = $3388; //@line 6485 "libraries/glad/glad.cpp"
  $3389 = (FUNCTION_TABLE_ii[$load & 8191](499195)|0); //@line 6486 "libraries/glad/glad.cpp"
  HEAP32[166608] = $3389; //@line 6486 "libraries/glad/glad.cpp"
  $3390 = (FUNCTION_TABLE_ii[$load & 8191](499226)|0); //@line 6487 "libraries/glad/glad.cpp"
  HEAP32[166609] = $3390; //@line 6487 "libraries/glad/glad.cpp"
  $3391 = (FUNCTION_TABLE_ii[$load & 8191](499252)|0); //@line 6488 "libraries/glad/glad.cpp"
  HEAP32[166610] = $3391; //@line 6488 "libraries/glad/glad.cpp"
  $3392 = (FUNCTION_TABLE_ii[$load & 8191](499280)|0); //@line 6489 "libraries/glad/glad.cpp"
  HEAP32[166611] = $3392; //@line 6489 "libraries/glad/glad.cpp"
  $3393 = (FUNCTION_TABLE_ii[$load & 8191](499313)|0); //@line 6490 "libraries/glad/glad.cpp"
  HEAP32[166612] = $3393; //@line 6490 "libraries/glad/glad.cpp"
  $3394 = (FUNCTION_TABLE_ii[$load & 8191](499333)|0); //@line 6491 "libraries/glad/glad.cpp"
  HEAP32[166613] = $3394; //@line 6491 "libraries/glad/glad.cpp"
  $3395 = (FUNCTION_TABLE_ii[$load & 8191](499349)|0); //@line 6492 "libraries/glad/glad.cpp"
  HEAP32[166614] = $3395; //@line 6492 "libraries/glad/glad.cpp"
  $3396 = (FUNCTION_TABLE_ii[$load & 8191](491288)|0); //@line 6493 "libraries/glad/glad.cpp"
  HEAP32[166265] = $3396; //@line 6493 "libraries/glad/glad.cpp"
  $3397 = (FUNCTION_TABLE_ii[$load & 8191](499366)|0); //@line 6494 "libraries/glad/glad.cpp"
  HEAP32[166615] = $3397; //@line 6494 "libraries/glad/glad.cpp"
  $3398 = (FUNCTION_TABLE_ii[$load & 8191](499389)|0); //@line 6495 "libraries/glad/glad.cpp"
  HEAP32[166616] = $3398; //@line 6495 "libraries/glad/glad.cpp"
 }
 $3399 = HEAP8[1252478]|0; //@line 6509 "libraries/glad/glad.cpp"
 $3400 = ($3399<<24>>24)==(0); //@line 6509 "libraries/glad/glad.cpp"
 if (!($3400)) {
  $3401 = (FUNCTION_TABLE_ii[$load & 8191](499413)|0); //@line 6510 "libraries/glad/glad.cpp"
  HEAP32[166617] = $3401; //@line 6510 "libraries/glad/glad.cpp"
 }
 $3402 = HEAP8[1252481]|0; //@line 6517 "libraries/glad/glad.cpp"
 $3403 = ($3402<<24>>24)==(0); //@line 6517 "libraries/glad/glad.cpp"
 if (!($3403)) {
  $3404 = (FUNCTION_TABLE_ii[$load & 8191](499432)|0); //@line 6518 "libraries/glad/glad.cpp"
  HEAP32[166618] = $3404; //@line 6518 "libraries/glad/glad.cpp"
  $3405 = (FUNCTION_TABLE_ii[$load & 8191](499466)|0); //@line 6519 "libraries/glad/glad.cpp"
  HEAP32[166619] = $3405; //@line 6519 "libraries/glad/glad.cpp"
  $3406 = (FUNCTION_TABLE_ii[$load & 8191](499500)|0); //@line 6520 "libraries/glad/glad.cpp"
  HEAP32[166620] = $3406; //@line 6520 "libraries/glad/glad.cpp"
  $3407 = (FUNCTION_TABLE_ii[$load & 8191](499530)|0); //@line 6521 "libraries/glad/glad.cpp"
  HEAP32[166621] = $3407; //@line 6521 "libraries/glad/glad.cpp"
  $3408 = (FUNCTION_TABLE_ii[$load & 8191](499560)|0); //@line 6522 "libraries/glad/glad.cpp"
  HEAP32[166622] = $3408; //@line 6522 "libraries/glad/glad.cpp"
  $3409 = (FUNCTION_TABLE_ii[$load & 8191](499598)|0); //@line 6523 "libraries/glad/glad.cpp"
  HEAP32[166623] = $3409; //@line 6523 "libraries/glad/glad.cpp"
 }
 $3410 = HEAP8[1252483]|0; //@line 6529 "libraries/glad/glad.cpp"
 $3411 = ($3410<<24>>24)==(0); //@line 6529 "libraries/glad/glad.cpp"
 if (!($3411)) {
  $3412 = (FUNCTION_TABLE_ii[$load & 8191](499636)|0); //@line 6530 "libraries/glad/glad.cpp"
  HEAP32[166624] = $3412; //@line 6530 "libraries/glad/glad.cpp"
  $3413 = (FUNCTION_TABLE_ii[$load & 8191](499663)|0); //@line 6531 "libraries/glad/glad.cpp"
  HEAP32[166625] = $3413; //@line 6531 "libraries/glad/glad.cpp"
  $3414 = (FUNCTION_TABLE_ii[$load & 8191](499688)|0); //@line 6532 "libraries/glad/glad.cpp"
  HEAP32[166626] = $3414; //@line 6532 "libraries/glad/glad.cpp"
  $3415 = (FUNCTION_TABLE_ii[$load & 8191](499717)|0); //@line 6533 "libraries/glad/glad.cpp"
  HEAP32[166627] = $3415; //@line 6533 "libraries/glad/glad.cpp"
  $3416 = (FUNCTION_TABLE_ii[$load & 8191](499737)|0); //@line 6534 "libraries/glad/glad.cpp"
  HEAP32[166628] = $3416; //@line 6534 "libraries/glad/glad.cpp"
  $3417 = (FUNCTION_TABLE_ii[$load & 8191](499758)|0); //@line 6535 "libraries/glad/glad.cpp"
  HEAP32[166629] = $3417; //@line 6535 "libraries/glad/glad.cpp"
  $3418 = (FUNCTION_TABLE_ii[$load & 8191](499777)|0); //@line 6536 "libraries/glad/glad.cpp"
  HEAP32[166630] = $3418; //@line 6536 "libraries/glad/glad.cpp"
  $3419 = (FUNCTION_TABLE_ii[$load & 8191](499807)|0); //@line 6537 "libraries/glad/glad.cpp"
  HEAP32[166631] = $3419; //@line 6537 "libraries/glad/glad.cpp"
  $3420 = (FUNCTION_TABLE_ii[$load & 8191](499825)|0); //@line 6538 "libraries/glad/glad.cpp"
  HEAP32[166632] = $3420; //@line 6538 "libraries/glad/glad.cpp"
  $3421 = (FUNCTION_TABLE_ii[$load & 8191](499848)|0); //@line 6539 "libraries/glad/glad.cpp"
  HEAP32[166633] = $3421; //@line 6539 "libraries/glad/glad.cpp"
  $3422 = (FUNCTION_TABLE_ii[$load & 8191](499869)|0); //@line 6540 "libraries/glad/glad.cpp"
  HEAP32[166634] = $3422; //@line 6540 "libraries/glad/glad.cpp"
  $3423 = (FUNCTION_TABLE_ii[$load & 8191](499901)|0); //@line 6541 "libraries/glad/glad.cpp"
  HEAP32[166635] = $3423; //@line 6541 "libraries/glad/glad.cpp"
 }
 $3424 = HEAP8[1252484]|0; //@line 6546 "libraries/glad/glad.cpp"
 $3425 = ($3424<<24>>24)==(0); //@line 6546 "libraries/glad/glad.cpp"
 if (!($3425)) {
  $3426 = (FUNCTION_TABLE_ii[$load & 8191](499936)|0); //@line 6547 "libraries/glad/glad.cpp"
  HEAP32[166636] = $3426; //@line 6547 "libraries/glad/glad.cpp"
  $3427 = (FUNCTION_TABLE_ii[$load & 8191](499962)|0); //@line 6548 "libraries/glad/glad.cpp"
  HEAP32[166637] = $3427; //@line 6548 "libraries/glad/glad.cpp"
  $3428 = (FUNCTION_TABLE_ii[$load & 8191](499991)|0); //@line 6549 "libraries/glad/glad.cpp"
  HEAP32[166638] = $3428; //@line 6549 "libraries/glad/glad.cpp"
  $3429 = (FUNCTION_TABLE_ii[$load & 8191](500017)|0); //@line 6550 "libraries/glad/glad.cpp"
  HEAP32[166639] = $3429; //@line 6550 "libraries/glad/glad.cpp"
  $3430 = (FUNCTION_TABLE_ii[$load & 8191](500041)|0); //@line 6551 "libraries/glad/glad.cpp"
  HEAP32[166640] = $3430; //@line 6551 "libraries/glad/glad.cpp"
  $3431 = (FUNCTION_TABLE_ii[$load & 8191](500068)|0); //@line 6552 "libraries/glad/glad.cpp"
  HEAP32[166641] = $3431; //@line 6552 "libraries/glad/glad.cpp"
  $3432 = (FUNCTION_TABLE_ii[$load & 8191](500096)|0); //@line 6553 "libraries/glad/glad.cpp"
  HEAP32[166642] = $3432; //@line 6553 "libraries/glad/glad.cpp"
 }
 $3433 = HEAP8[1252486]|0; //@line 6559 "libraries/glad/glad.cpp"
 $3434 = ($3433<<24>>24)==(0); //@line 6559 "libraries/glad/glad.cpp"
 if (!($3434)) {
  $3435 = (FUNCTION_TABLE_ii[$load & 8191](500122)|0); //@line 6560 "libraries/glad/glad.cpp"
  HEAP32[166643] = $3435; //@line 6560 "libraries/glad/glad.cpp"
  $3436 = (FUNCTION_TABLE_ii[$load & 8191](500136)|0); //@line 6561 "libraries/glad/glad.cpp"
  HEAP32[166644] = $3436; //@line 6561 "libraries/glad/glad.cpp"
  $3437 = (FUNCTION_TABLE_ii[$load & 8191](500150)|0); //@line 6562 "libraries/glad/glad.cpp"
  HEAP32[166645] = $3437; //@line 6562 "libraries/glad/glad.cpp"
  $3438 = (FUNCTION_TABLE_ii[$load & 8191](500180)|0); //@line 6563 "libraries/glad/glad.cpp"
  HEAP32[166646] = $3438; //@line 6563 "libraries/glad/glad.cpp"
  $3439 = (FUNCTION_TABLE_ii[$load & 8191](500211)|0); //@line 6564 "libraries/glad/glad.cpp"
  HEAP32[166647] = $3439; //@line 6564 "libraries/glad/glad.cpp"
  $3440 = (FUNCTION_TABLE_ii[$load & 8191](500230)|0); //@line 6565 "libraries/glad/glad.cpp"
  HEAP32[166648] = $3440; //@line 6565 "libraries/glad/glad.cpp"
  $3441 = (FUNCTION_TABLE_ii[$load & 8191](500257)|0); //@line 6566 "libraries/glad/glad.cpp"
  HEAP32[166649] = $3441; //@line 6566 "libraries/glad/glad.cpp"
  $3442 = (FUNCTION_TABLE_ii[$load & 8191](500279)|0); //@line 6567 "libraries/glad/glad.cpp"
  HEAP32[166650] = $3442; //@line 6567 "libraries/glad/glad.cpp"
  $3443 = (FUNCTION_TABLE_ii[$load & 8191](500302)|0); //@line 6568 "libraries/glad/glad.cpp"
  HEAP32[166651] = $3443; //@line 6568 "libraries/glad/glad.cpp"
  $3444 = (FUNCTION_TABLE_ii[$load & 8191](500323)|0); //@line 6569 "libraries/glad/glad.cpp"
  HEAP32[166652] = $3444; //@line 6569 "libraries/glad/glad.cpp"
 }
 $3445 = HEAP8[1252487]|0; //@line 6574 "libraries/glad/glad.cpp"
 $3446 = ($3445<<24>>24)==(0); //@line 6574 "libraries/glad/glad.cpp"
 if (!($3446)) {
  $3447 = (FUNCTION_TABLE_ii[$load & 8191](500346)|0); //@line 6575 "libraries/glad/glad.cpp"
  HEAP32[166653] = $3447; //@line 6575 "libraries/glad/glad.cpp"
  $3448 = (FUNCTION_TABLE_ii[$load & 8191](500368)|0); //@line 6576 "libraries/glad/glad.cpp"
  HEAP32[166654] = $3448; //@line 6576 "libraries/glad/glad.cpp"
  $3449 = (FUNCTION_TABLE_ii[$load & 8191](500390)|0); //@line 6577 "libraries/glad/glad.cpp"
  HEAP32[166655] = $3449; //@line 6577 "libraries/glad/glad.cpp"
  $3450 = (FUNCTION_TABLE_ii[$load & 8191](500412)|0); //@line 6578 "libraries/glad/glad.cpp"
  HEAP32[166656] = $3450; //@line 6578 "libraries/glad/glad.cpp"
  $3451 = (FUNCTION_TABLE_ii[$load & 8191](500434)|0); //@line 6579 "libraries/glad/glad.cpp"
  HEAP32[166657] = $3451; //@line 6579 "libraries/glad/glad.cpp"
  $3452 = (FUNCTION_TABLE_ii[$load & 8191](500457)|0); //@line 6580 "libraries/glad/glad.cpp"
  HEAP32[166658] = $3452; //@line 6580 "libraries/glad/glad.cpp"
  $3453 = (FUNCTION_TABLE_ii[$load & 8191](500480)|0); //@line 6581 "libraries/glad/glad.cpp"
  HEAP32[166659] = $3453; //@line 6581 "libraries/glad/glad.cpp"
  $3454 = (FUNCTION_TABLE_ii[$load & 8191](500503)|0); //@line 6582 "libraries/glad/glad.cpp"
  HEAP32[166660] = $3454; //@line 6582 "libraries/glad/glad.cpp"
  $3455 = (FUNCTION_TABLE_ii[$load & 8191](500526)|0); //@line 6583 "libraries/glad/glad.cpp"
  HEAP32[166661] = $3455; //@line 6583 "libraries/glad/glad.cpp"
  $3456 = (FUNCTION_TABLE_ii[$load & 8191](500549)|0); //@line 6584 "libraries/glad/glad.cpp"
  HEAP32[166662] = $3456; //@line 6584 "libraries/glad/glad.cpp"
  $3457 = (FUNCTION_TABLE_ii[$load & 8191](500572)|0); //@line 6585 "libraries/glad/glad.cpp"
  HEAP32[166663] = $3457; //@line 6585 "libraries/glad/glad.cpp"
  $3458 = (FUNCTION_TABLE_ii[$load & 8191](500595)|0); //@line 6586 "libraries/glad/glad.cpp"
  HEAP32[166664] = $3458; //@line 6586 "libraries/glad/glad.cpp"
  $3459 = (FUNCTION_TABLE_ii[$load & 8191](500618)|0); //@line 6587 "libraries/glad/glad.cpp"
  HEAP32[166665] = $3459; //@line 6587 "libraries/glad/glad.cpp"
  $3460 = (FUNCTION_TABLE_ii[$load & 8191](500642)|0); //@line 6588 "libraries/glad/glad.cpp"
  HEAP32[166666] = $3460; //@line 6588 "libraries/glad/glad.cpp"
  $3461 = (FUNCTION_TABLE_ii[$load & 8191](500666)|0); //@line 6589 "libraries/glad/glad.cpp"
  HEAP32[166667] = $3461; //@line 6589 "libraries/glad/glad.cpp"
  $3462 = (FUNCTION_TABLE_ii[$load & 8191](500690)|0); //@line 6590 "libraries/glad/glad.cpp"
  HEAP32[166668] = $3462; //@line 6590 "libraries/glad/glad.cpp"
  $3463 = (FUNCTION_TABLE_ii[$load & 8191](500714)|0); //@line 6591 "libraries/glad/glad.cpp"
  HEAP32[166669] = $3463; //@line 6591 "libraries/glad/glad.cpp"
  $3464 = (FUNCTION_TABLE_ii[$load & 8191](500739)|0); //@line 6592 "libraries/glad/glad.cpp"
  HEAP32[166670] = $3464; //@line 6592 "libraries/glad/glad.cpp"
  $3465 = (FUNCTION_TABLE_ii[$load & 8191](500765)|0); //@line 6593 "libraries/glad/glad.cpp"
  HEAP32[166671] = $3465; //@line 6593 "libraries/glad/glad.cpp"
 }
 $3466 = HEAP8[1252488]|0; //@line 6598 "libraries/glad/glad.cpp"
 $3467 = ($3466<<24>>24)==(0); //@line 6598 "libraries/glad/glad.cpp"
 if (!($3467)) {
  $3468 = (FUNCTION_TABLE_ii[$load & 8191](500789)|0); //@line 6599 "libraries/glad/glad.cpp"
  HEAP32[166672] = $3468; //@line 6599 "libraries/glad/glad.cpp"
  $3469 = (FUNCTION_TABLE_ii[$load & 8191](500812)|0); //@line 6600 "libraries/glad/glad.cpp"
  HEAP32[166673] = $3469; //@line 6600 "libraries/glad/glad.cpp"
  $3470 = (FUNCTION_TABLE_ii[$load & 8191](500829)|0); //@line 6601 "libraries/glad/glad.cpp"
  HEAP32[166674] = $3470; //@line 6601 "libraries/glad/glad.cpp"
  $3471 = (FUNCTION_TABLE_ii[$load & 8191](500846)|0); //@line 6602 "libraries/glad/glad.cpp"
  HEAP32[166675] = $3471; //@line 6602 "libraries/glad/glad.cpp"
  $3472 = (FUNCTION_TABLE_ii[$load & 8191](500862)|0); //@line 6603 "libraries/glad/glad.cpp"
  HEAP32[166676] = $3472; //@line 6603 "libraries/glad/glad.cpp"
  $3473 = (FUNCTION_TABLE_ii[$load & 8191](500878)|0); //@line 6604 "libraries/glad/glad.cpp"
  HEAP32[166677] = $3473; //@line 6604 "libraries/glad/glad.cpp"
  $3474 = (FUNCTION_TABLE_ii[$load & 8191](500897)|0); //@line 6605 "libraries/glad/glad.cpp"
  HEAP32[166678] = $3474; //@line 6605 "libraries/glad/glad.cpp"
  $3475 = (FUNCTION_TABLE_ii[$load & 8191](500916)|0); //@line 6606 "libraries/glad/glad.cpp"
  HEAP32[166679] = $3475; //@line 6606 "libraries/glad/glad.cpp"
  $3476 = (FUNCTION_TABLE_ii[$load & 8191](500941)|0); //@line 6607 "libraries/glad/glad.cpp"
  HEAP32[166680] = $3476; //@line 6607 "libraries/glad/glad.cpp"
  $3477 = (FUNCTION_TABLE_ii[$load & 8191](500960)|0); //@line 6608 "libraries/glad/glad.cpp"
  HEAP32[166681] = $3477; //@line 6608 "libraries/glad/glad.cpp"
  $3478 = (FUNCTION_TABLE_ii[$load & 8191](500983)|0); //@line 6609 "libraries/glad/glad.cpp"
  HEAP32[166682] = $3478; //@line 6609 "libraries/glad/glad.cpp"
  $3479 = (FUNCTION_TABLE_ii[$load & 8191](501007)|0); //@line 6610 "libraries/glad/glad.cpp"
  HEAP32[166683] = $3479; //@line 6610 "libraries/glad/glad.cpp"
 }
 $3480 = HEAP8[1252489]|0; //@line 6615 "libraries/glad/glad.cpp"
 $3481 = ($3480<<24>>24)==(0); //@line 6615 "libraries/glad/glad.cpp"
 if (!($3481)) {
  $3482 = (FUNCTION_TABLE_ii[$load & 8191](501029)|0); //@line 6616 "libraries/glad/glad.cpp"
  HEAP32[166684] = $3482; //@line 6616 "libraries/glad/glad.cpp"
  $3483 = (FUNCTION_TABLE_ii[$load & 8191](501050)|0); //@line 6617 "libraries/glad/glad.cpp"
  HEAP32[166685] = $3483; //@line 6617 "libraries/glad/glad.cpp"
  $3484 = (FUNCTION_TABLE_ii[$load & 8191](501071)|0); //@line 6618 "libraries/glad/glad.cpp"
  HEAP32[166686] = $3484; //@line 6618 "libraries/glad/glad.cpp"
  $3485 = (FUNCTION_TABLE_ii[$load & 8191](501092)|0); //@line 6619 "libraries/glad/glad.cpp"
  HEAP32[166687] = $3485; //@line 6619 "libraries/glad/glad.cpp"
  $3486 = (FUNCTION_TABLE_ii[$load & 8191](501113)|0); //@line 6620 "libraries/glad/glad.cpp"
  HEAP32[166688] = $3486; //@line 6620 "libraries/glad/glad.cpp"
  $3487 = (FUNCTION_TABLE_ii[$load & 8191](501135)|0); //@line 6621 "libraries/glad/glad.cpp"
  HEAP32[166689] = $3487; //@line 6621 "libraries/glad/glad.cpp"
  $3488 = (FUNCTION_TABLE_ii[$load & 8191](501157)|0); //@line 6622 "libraries/glad/glad.cpp"
  HEAP32[166690] = $3488; //@line 6622 "libraries/glad/glad.cpp"
  $3489 = (FUNCTION_TABLE_ii[$load & 8191](501179)|0); //@line 6623 "libraries/glad/glad.cpp"
  HEAP32[166691] = $3489; //@line 6623 "libraries/glad/glad.cpp"
  $3490 = (FUNCTION_TABLE_ii[$load & 8191](501201)|0); //@line 6624 "libraries/glad/glad.cpp"
  HEAP32[166692] = $3490; //@line 6624 "libraries/glad/glad.cpp"
  $3491 = (FUNCTION_TABLE_ii[$load & 8191](501223)|0); //@line 6625 "libraries/glad/glad.cpp"
  HEAP32[166693] = $3491; //@line 6625 "libraries/glad/glad.cpp"
  $3492 = (FUNCTION_TABLE_ii[$load & 8191](501245)|0); //@line 6626 "libraries/glad/glad.cpp"
  HEAP32[166694] = $3492; //@line 6626 "libraries/glad/glad.cpp"
  $3493 = (FUNCTION_TABLE_ii[$load & 8191](501267)|0); //@line 6627 "libraries/glad/glad.cpp"
  HEAP32[166695] = $3493; //@line 6627 "libraries/glad/glad.cpp"
  $3494 = (FUNCTION_TABLE_ii[$load & 8191](501289)|0); //@line 6628 "libraries/glad/glad.cpp"
  HEAP32[166696] = $3494; //@line 6628 "libraries/glad/glad.cpp"
  $3495 = (FUNCTION_TABLE_ii[$load & 8191](501312)|0); //@line 6629 "libraries/glad/glad.cpp"
  HEAP32[166697] = $3495; //@line 6629 "libraries/glad/glad.cpp"
  $3496 = (FUNCTION_TABLE_ii[$load & 8191](501335)|0); //@line 6630 "libraries/glad/glad.cpp"
  HEAP32[166698] = $3496; //@line 6630 "libraries/glad/glad.cpp"
  $3497 = (FUNCTION_TABLE_ii[$load & 8191](501358)|0); //@line 6631 "libraries/glad/glad.cpp"
  HEAP32[166699] = $3497; //@line 6631 "libraries/glad/glad.cpp"
  $3498 = (FUNCTION_TABLE_ii[$load & 8191](501381)|0); //@line 6632 "libraries/glad/glad.cpp"
  HEAP32[166700] = $3498; //@line 6632 "libraries/glad/glad.cpp"
  $3499 = (FUNCTION_TABLE_ii[$load & 8191](501403)|0); //@line 6633 "libraries/glad/glad.cpp"
  HEAP32[166701] = $3499; //@line 6633 "libraries/glad/glad.cpp"
  $3500 = (FUNCTION_TABLE_ii[$load & 8191](501425)|0); //@line 6634 "libraries/glad/glad.cpp"
  HEAP32[166702] = $3500; //@line 6634 "libraries/glad/glad.cpp"
  $3501 = (FUNCTION_TABLE_ii[$load & 8191](501448)|0); //@line 6635 "libraries/glad/glad.cpp"
  HEAP32[166703] = $3501; //@line 6635 "libraries/glad/glad.cpp"
  $3502 = (FUNCTION_TABLE_ii[$load & 8191](501471)|0); //@line 6636 "libraries/glad/glad.cpp"
  HEAP32[166704] = $3502; //@line 6636 "libraries/glad/glad.cpp"
  $3503 = (FUNCTION_TABLE_ii[$load & 8191](501497)|0); //@line 6637 "libraries/glad/glad.cpp"
  HEAP32[166705] = $3503; //@line 6637 "libraries/glad/glad.cpp"
  $3504 = (FUNCTION_TABLE_ii[$load & 8191](501521)|0); //@line 6638 "libraries/glad/glad.cpp"
  HEAP32[166706] = $3504; //@line 6638 "libraries/glad/glad.cpp"
 }
 $3505 = HEAP8[1252490]|0; //@line 6643 "libraries/glad/glad.cpp"
 $3506 = ($3505<<24>>24)==(0); //@line 6643 "libraries/glad/glad.cpp"
 if (!($3506)) {
  $3507 = (FUNCTION_TABLE_ii[$load & 8191](501546)|0); //@line 6644 "libraries/glad/glad.cpp"
  HEAP32[166707] = $3507; //@line 6644 "libraries/glad/glad.cpp"
  $3508 = (FUNCTION_TABLE_ii[$load & 8191](501568)|0); //@line 6645 "libraries/glad/glad.cpp"
  HEAP32[166708] = $3508; //@line 6645 "libraries/glad/glad.cpp"
  $3509 = (FUNCTION_TABLE_ii[$load & 8191](501601)|0); //@line 6646 "libraries/glad/glad.cpp"
  HEAP32[166709] = $3509; //@line 6646 "libraries/glad/glad.cpp"
  $3510 = (FUNCTION_TABLE_ii[$load & 8191](501635)|0); //@line 6647 "libraries/glad/glad.cpp"
  HEAP32[166710] = $3510; //@line 6647 "libraries/glad/glad.cpp"
  $3511 = (FUNCTION_TABLE_ii[$load & 8191](501655)|0); //@line 6648 "libraries/glad/glad.cpp"
  HEAP32[166711] = $3511; //@line 6648 "libraries/glad/glad.cpp"
  $3512 = (FUNCTION_TABLE_ii[$load & 8191](501677)|0); //@line 6649 "libraries/glad/glad.cpp"
  HEAP32[166712] = $3512; //@line 6649 "libraries/glad/glad.cpp"
  $3513 = (FUNCTION_TABLE_ii[$load & 8191](501705)|0); //@line 6650 "libraries/glad/glad.cpp"
  HEAP32[166713] = $3513; //@line 6650 "libraries/glad/glad.cpp"
  $3514 = (FUNCTION_TABLE_ii[$load & 8191](501733)|0); //@line 6651 "libraries/glad/glad.cpp"
  HEAP32[166714] = $3514; //@line 6651 "libraries/glad/glad.cpp"
  $3515 = (FUNCTION_TABLE_ii[$load & 8191](501761)|0); //@line 6652 "libraries/glad/glad.cpp"
  HEAP32[166715] = $3515; //@line 6652 "libraries/glad/glad.cpp"
  $3516 = (FUNCTION_TABLE_ii[$load & 8191](501778)|0); //@line 6653 "libraries/glad/glad.cpp"
  HEAP32[166716] = $3516; //@line 6653 "libraries/glad/glad.cpp"
  $3517 = (FUNCTION_TABLE_ii[$load & 8191](501812)|0); //@line 6654 "libraries/glad/glad.cpp"
  HEAP32[166717] = $3517; //@line 6654 "libraries/glad/glad.cpp"
  $3518 = (FUNCTION_TABLE_ii[$load & 8191](501846)|0); //@line 6655 "libraries/glad/glad.cpp"
  HEAP32[166718] = $3518; //@line 6655 "libraries/glad/glad.cpp"
 }
 $3519 = HEAP8[1252491]|0; //@line 6660 "libraries/glad/glad.cpp"
 $3520 = ($3519<<24>>24)==(0); //@line 6660 "libraries/glad/glad.cpp"
 if (!($3520)) {
  $3521 = (FUNCTION_TABLE_ii[$load & 8191](501880)|0); //@line 6661 "libraries/glad/glad.cpp"
  HEAP32[166719] = $3521; //@line 6661 "libraries/glad/glad.cpp"
  $3522 = (FUNCTION_TABLE_ii[$load & 8191](501899)|0); //@line 6662 "libraries/glad/glad.cpp"
  HEAP32[166720] = $3522; //@line 6662 "libraries/glad/glad.cpp"
  $3523 = (FUNCTION_TABLE_ii[$load & 8191](501920)|0); //@line 6663 "libraries/glad/glad.cpp"
  HEAP32[166721] = $3523; //@line 6663 "libraries/glad/glad.cpp"
  $3524 = (FUNCTION_TABLE_ii[$load & 8191](501942)|0); //@line 6664 "libraries/glad/glad.cpp"
  HEAP32[166722] = $3524; //@line 6664 "libraries/glad/glad.cpp"
  $3525 = (FUNCTION_TABLE_ii[$load & 8191](501960)|0); //@line 6665 "libraries/glad/glad.cpp"
  HEAP32[166723] = $3525; //@line 6665 "libraries/glad/glad.cpp"
  $3526 = (FUNCTION_TABLE_ii[$load & 8191](501979)|0); //@line 6666 "libraries/glad/glad.cpp"
  HEAP32[166724] = $3526; //@line 6666 "libraries/glad/glad.cpp"
  $3527 = (FUNCTION_TABLE_ii[$load & 8191](501999)|0); //@line 6667 "libraries/glad/glad.cpp"
  HEAP32[166725] = $3527; //@line 6667 "libraries/glad/glad.cpp"
  $3528 = (FUNCTION_TABLE_ii[$load & 8191](502021)|0); //@line 6668 "libraries/glad/glad.cpp"
  HEAP32[166726] = $3528; //@line 6668 "libraries/glad/glad.cpp"
  $3529 = (FUNCTION_TABLE_ii[$load & 8191](502044)|0); //@line 6669 "libraries/glad/glad.cpp"
  HEAP32[166727] = $3529; //@line 6669 "libraries/glad/glad.cpp"
  $3530 = (FUNCTION_TABLE_ii[$load & 8191](502060)|0); //@line 6670 "libraries/glad/glad.cpp"
  HEAP32[166728] = $3530; //@line 6670 "libraries/glad/glad.cpp"
  $3531 = (FUNCTION_TABLE_ii[$load & 8191](502072)|0); //@line 6671 "libraries/glad/glad.cpp"
  HEAP32[166729] = $3531; //@line 6671 "libraries/glad/glad.cpp"
  $3532 = (FUNCTION_TABLE_ii[$load & 8191](502085)|0); //@line 6672 "libraries/glad/glad.cpp"
  HEAP32[166730] = $3532; //@line 6672 "libraries/glad/glad.cpp"
 }
 $3533 = HEAP8[1252493]|0; //@line 6678 "libraries/glad/glad.cpp"
 $3534 = ($3533<<24>>24)==(0); //@line 6678 "libraries/glad/glad.cpp"
 if (!($3534)) {
  $3535 = (FUNCTION_TABLE_ii[$load & 8191](502100)|0); //@line 6679 "libraries/glad/glad.cpp"
  HEAP32[166731] = $3535; //@line 6679 "libraries/glad/glad.cpp"
  $3536 = (FUNCTION_TABLE_ii[$load & 8191](502128)|0); //@line 6680 "libraries/glad/glad.cpp"
  HEAP32[166732] = $3536; //@line 6680 "libraries/glad/glad.cpp"
 }
 $3537 = HEAP8[1252495]|0; //@line 6686 "libraries/glad/glad.cpp"
 $3538 = ($3537<<24>>24)==(0); //@line 6686 "libraries/glad/glad.cpp"
 if (!($3538)) {
  $3539 = (FUNCTION_TABLE_ii[$load & 8191](502154)|0); //@line 6687 "libraries/glad/glad.cpp"
  HEAP32[166733] = $3539; //@line 6687 "libraries/glad/glad.cpp"
 }
 $3540 = HEAP8[1252497]|0; //@line 6693 "libraries/glad/glad.cpp"
 $3541 = ($3540<<24>>24)==(0); //@line 6693 "libraries/glad/glad.cpp"
 if (!($3541)) {
  $3542 = (FUNCTION_TABLE_ii[$load & 8191](502187)|0); //@line 6694 "libraries/glad/glad.cpp"
  HEAP32[166734] = $3542; //@line 6694 "libraries/glad/glad.cpp"
 }
 $3543 = HEAP8[1252498]|0; //@line 6699 "libraries/glad/glad.cpp"
 $3544 = ($3543<<24>>24)==(0); //@line 6699 "libraries/glad/glad.cpp"
 if (!($3544)) {
  $3545 = (FUNCTION_TABLE_ii[$load & 8191](502231)|0); //@line 6700 "libraries/glad/glad.cpp"
  HEAP32[166735] = $3545; //@line 6700 "libraries/glad/glad.cpp"
 }
 $3546 = HEAP8[1252500]|0; //@line 6706 "libraries/glad/glad.cpp"
 $3547 = ($3546<<24>>24)==(0); //@line 6706 "libraries/glad/glad.cpp"
 if (!($3547)) {
  $3548 = (FUNCTION_TABLE_ii[$load & 8191](502247)|0); //@line 6707 "libraries/glad/glad.cpp"
  HEAP32[166736] = $3548; //@line 6707 "libraries/glad/glad.cpp"
  $3549 = (FUNCTION_TABLE_ii[$load & 8191](502271)|0); //@line 6708 "libraries/glad/glad.cpp"
  HEAP32[166737] = $3549; //@line 6708 "libraries/glad/glad.cpp"
  $3550 = (FUNCTION_TABLE_ii[$load & 8191](502300)|0); //@line 6709 "libraries/glad/glad.cpp"
  HEAP32[166738] = $3550; //@line 6709 "libraries/glad/glad.cpp"
  $3551 = (FUNCTION_TABLE_ii[$load & 8191](502326)|0); //@line 6710 "libraries/glad/glad.cpp"
  HEAP32[166739] = $3551; //@line 6710 "libraries/glad/glad.cpp"
 }
 $3552 = HEAP8[1252501]|0; //@line 6715 "libraries/glad/glad.cpp"
 $3553 = ($3552<<24>>24)==(0); //@line 6715 "libraries/glad/glad.cpp"
 if (!($3553)) {
  $3554 = (FUNCTION_TABLE_ii[$load & 8191](502353)|0); //@line 6716 "libraries/glad/glad.cpp"
  HEAP32[166740] = $3554; //@line 6716 "libraries/glad/glad.cpp"
  $3555 = (FUNCTION_TABLE_ii[$load & 8191](502374)|0); //@line 6717 "libraries/glad/glad.cpp"
  HEAP32[166741] = $3555; //@line 6717 "libraries/glad/glad.cpp"
  $3556 = (FUNCTION_TABLE_ii[$load & 8191](502394)|0); //@line 6718 "libraries/glad/glad.cpp"
  HEAP32[166742] = $3556; //@line 6718 "libraries/glad/glad.cpp"
  $3557 = (FUNCTION_TABLE_ii[$load & 8191](502420)|0); //@line 6719 "libraries/glad/glad.cpp"
  HEAP32[166743] = $3557; //@line 6719 "libraries/glad/glad.cpp"
  $3558 = (FUNCTION_TABLE_ii[$load & 8191](502445)|0); //@line 6720 "libraries/glad/glad.cpp"
  HEAP32[166744] = $3558; //@line 6720 "libraries/glad/glad.cpp"
  $3559 = (FUNCTION_TABLE_ii[$load & 8191](502477)|0); //@line 6721 "libraries/glad/glad.cpp"
  HEAP32[166745] = $3559; //@line 6721 "libraries/glad/glad.cpp"
  $3560 = (FUNCTION_TABLE_ii[$load & 8191](502510)|0); //@line 6722 "libraries/glad/glad.cpp"
  HEAP32[166746] = $3560; //@line 6722 "libraries/glad/glad.cpp"
  $3561 = (FUNCTION_TABLE_ii[$load & 8191](502534)|0); //@line 6723 "libraries/glad/glad.cpp"
  HEAP32[166747] = $3561; //@line 6723 "libraries/glad/glad.cpp"
 }
 $3562 = HEAP8[1252502]|0; //@line 6728 "libraries/glad/glad.cpp"
 $3563 = ($3562<<24>>24)==(0); //@line 6728 "libraries/glad/glad.cpp"
 if (!($3563)) {
  $3564 = (FUNCTION_TABLE_ii[$load & 8191](502561)|0); //@line 6729 "libraries/glad/glad.cpp"
  HEAP32[166748] = $3564; //@line 6729 "libraries/glad/glad.cpp"
  $3565 = (FUNCTION_TABLE_ii[$load & 8191](502581)|0); //@line 6730 "libraries/glad/glad.cpp"
  HEAP32[166749] = $3565; //@line 6730 "libraries/glad/glad.cpp"
  $3566 = (FUNCTION_TABLE_ii[$load & 8191](502602)|0); //@line 6731 "libraries/glad/glad.cpp"
  HEAP32[166750] = $3566; //@line 6731 "libraries/glad/glad.cpp"
  $3567 = (FUNCTION_TABLE_ii[$load & 8191](502627)|0); //@line 6732 "libraries/glad/glad.cpp"
  HEAP32[166751] = $3567; //@line 6732 "libraries/glad/glad.cpp"
 }
 $3568 = HEAP8[1252504]|0; //@line 6738 "libraries/glad/glad.cpp"
 $3569 = ($3568<<24>>24)==(0); //@line 6738 "libraries/glad/glad.cpp"
 if ($3569) {
  $$0 = 1;
  return ($$0|0); //@line 7537 "libraries/glad/glad.cpp"
 }
 $3570 = (FUNCTION_TABLE_ii[$load & 8191](502659)|0); //@line 6739 "libraries/glad/glad.cpp"
 HEAP32[166752] = $3570; //@line 6739 "libraries/glad/glad.cpp"
 $3571 = (FUNCTION_TABLE_ii[$load & 8191](502677)|0); //@line 6740 "libraries/glad/glad.cpp"
 HEAP32[166753] = $3571; //@line 6740 "libraries/glad/glad.cpp"
 $$0 = 1;
 return ($$0|0); //@line 7537 "libraries/glad/glad.cpp"
}
function __ZN4gladL15find_extensionsEv() {
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0;
 var $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0;
 var $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0;
 var $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0;
 var $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0;
 var $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0;
 var $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0;
 var $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0;
 var $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0;
 var $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0;
 var $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0;
 var $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0;
 var $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0;
 var $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0;
 var $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0;
 var $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0;
 var $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0;
 var $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0;
 var $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0;
 var $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0;
 var $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0;
 var $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0;
 var $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0;
 var $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0;
 var $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0;
 var $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0;
 var $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0;
 var $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0;
 var $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0;
 var $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0;
 var $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0;
 var $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0;
 var $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0;
 var $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0;
 var $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0;
 var $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0;
 var $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0;
 var $981 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN4gladL7has_extEPKc(463607)|0); //@line 6746 "libraries/glad/glad.cpp"
 $1 = $0&1; //@line 6746 "libraries/glad/glad.cpp"
 HEAP8[1252016] = $1; //@line 6746 "libraries/glad/glad.cpp"
 $2 = (__ZN4gladL7has_extEPKc(463638)|0); //@line 6747 "libraries/glad/glad.cpp"
 $3 = $2&1; //@line 6747 "libraries/glad/glad.cpp"
 HEAP8[1252017] = $3; //@line 6747 "libraries/glad/glad.cpp"
 $4 = (__ZN4gladL7has_extEPKc(463678)|0); //@line 6748 "libraries/glad/glad.cpp"
 $5 = $4&1; //@line 6748 "libraries/glad/glad.cpp"
 HEAP8[1252018] = $5; //@line 6748 "libraries/glad/glad.cpp"
 $6 = (__ZN4gladL7has_extEPKc(463707)|0); //@line 6749 "libraries/glad/glad.cpp"
 $7 = $6&1; //@line 6749 "libraries/glad/glad.cpp"
 HEAP8[1252019] = $7; //@line 6749 "libraries/glad/glad.cpp"
 $8 = (__ZN4gladL7has_extEPKc(463720)|0); //@line 6750 "libraries/glad/glad.cpp"
 $9 = $8&1; //@line 6750 "libraries/glad/glad.cpp"
 HEAP8[1252020] = $9; //@line 6750 "libraries/glad/glad.cpp"
 $10 = (__ZN4gladL7has_extEPKc(463736)|0); //@line 6751 "libraries/glad/glad.cpp"
 $11 = $10&1; //@line 6751 "libraries/glad/glad.cpp"
 HEAP8[1252021] = $11; //@line 6751 "libraries/glad/glad.cpp"
 $12 = (__ZN4gladL7has_extEPKc(463773)|0); //@line 6752 "libraries/glad/glad.cpp"
 $13 = $12&1; //@line 6752 "libraries/glad/glad.cpp"
 HEAP8[1252022] = $13; //@line 6752 "libraries/glad/glad.cpp"
 $14 = (__ZN4gladL7has_extEPKc(463791)|0); //@line 6753 "libraries/glad/glad.cpp"
 $15 = $14&1; //@line 6753 "libraries/glad/glad.cpp"
 HEAP8[1252023] = $15; //@line 6753 "libraries/glad/glad.cpp"
 $16 = (__ZN4gladL7has_extEPKc(463827)|0); //@line 6754 "libraries/glad/glad.cpp"
 $17 = $16&1; //@line 6754 "libraries/glad/glad.cpp"
 HEAP8[1252024] = $17; //@line 6754 "libraries/glad/glad.cpp"
 $18 = (__ZN4gladL7has_extEPKc(463863)|0); //@line 6755 "libraries/glad/glad.cpp"
 $19 = $18&1; //@line 6755 "libraries/glad/glad.cpp"
 HEAP8[1252025] = $19; //@line 6755 "libraries/glad/glad.cpp"
 $20 = (__ZN4gladL7has_extEPKc(463905)|0); //@line 6756 "libraries/glad/glad.cpp"
 $21 = $20&1; //@line 6756 "libraries/glad/glad.cpp"
 HEAP8[1252026] = $21; //@line 6756 "libraries/glad/glad.cpp"
 $22 = (__ZN4gladL7has_extEPKc(463929)|0); //@line 6757 "libraries/glad/glad.cpp"
 $23 = $22&1; //@line 6757 "libraries/glad/glad.cpp"
 HEAP8[1252027] = $23; //@line 6757 "libraries/glad/glad.cpp"
 $24 = (__ZN4gladL7has_extEPKc(463950)|0); //@line 6758 "libraries/glad/glad.cpp"
 $25 = $24&1; //@line 6758 "libraries/glad/glad.cpp"
 HEAP8[1252028] = $25; //@line 6758 "libraries/glad/glad.cpp"
 $26 = (__ZN4gladL7has_extEPKc(463974)|0); //@line 6759 "libraries/glad/glad.cpp"
 $27 = $26&1; //@line 6759 "libraries/glad/glad.cpp"
 HEAP8[1252029] = $27; //@line 6759 "libraries/glad/glad.cpp"
 $28 = (__ZN4gladL7has_extEPKc(464001)|0); //@line 6760 "libraries/glad/glad.cpp"
 $29 = $28&1; //@line 6760 "libraries/glad/glad.cpp"
 HEAP8[1252030] = $29; //@line 6760 "libraries/glad/glad.cpp"
 $30 = (__ZN4gladL7has_extEPKc(464023)|0); //@line 6761 "libraries/glad/glad.cpp"
 $31 = $30&1; //@line 6761 "libraries/glad/glad.cpp"
 HEAP8[1252031] = $31; //@line 6761 "libraries/glad/glad.cpp"
 $32 = (__ZN4gladL7has_extEPKc(464039)|0); //@line 6762 "libraries/glad/glad.cpp"
 $33 = $32&1; //@line 6762 "libraries/glad/glad.cpp"
 HEAP8[1252032] = $33; //@line 6762 "libraries/glad/glad.cpp"
 $34 = (__ZN4gladL7has_extEPKc(464066)|0); //@line 6763 "libraries/glad/glad.cpp"
 $35 = $34&1; //@line 6763 "libraries/glad/glad.cpp"
 HEAP8[1252033] = $35; //@line 6763 "libraries/glad/glad.cpp"
 $36 = (__ZN4gladL7has_extEPKc(464087)|0); //@line 6764 "libraries/glad/glad.cpp"
 $37 = $36&1; //@line 6764 "libraries/glad/glad.cpp"
 HEAP8[1252034] = $37; //@line 6764 "libraries/glad/glad.cpp"
 $38 = (__ZN4gladL7has_extEPKc(464107)|0); //@line 6765 "libraries/glad/glad.cpp"
 $39 = $38&1; //@line 6765 "libraries/glad/glad.cpp"
 HEAP8[1252035] = $39; //@line 6765 "libraries/glad/glad.cpp"
 $40 = (__ZN4gladL7has_extEPKc(464133)|0); //@line 6766 "libraries/glad/glad.cpp"
 $41 = $40&1; //@line 6766 "libraries/glad/glad.cpp"
 HEAP8[1252036] = $41; //@line 6766 "libraries/glad/glad.cpp"
 $42 = (__ZN4gladL7has_extEPKc(464154)|0); //@line 6767 "libraries/glad/glad.cpp"
 $43 = $42&1; //@line 6767 "libraries/glad/glad.cpp"
 HEAP8[1252037] = $43; //@line 6767 "libraries/glad/glad.cpp"
 $44 = (__ZN4gladL7has_extEPKc(464194)|0); //@line 6768 "libraries/glad/glad.cpp"
 $45 = $44&1; //@line 6768 "libraries/glad/glad.cpp"
 HEAP8[1252038] = $45; //@line 6768 "libraries/glad/glad.cpp"
 $46 = (__ZN4gladL7has_extEPKc(464216)|0); //@line 6769 "libraries/glad/glad.cpp"
 $47 = $46&1; //@line 6769 "libraries/glad/glad.cpp"
 HEAP8[1252039] = $47; //@line 6769 "libraries/glad/glad.cpp"
 $48 = (__ZN4gladL7has_extEPKc(464251)|0); //@line 6770 "libraries/glad/glad.cpp"
 $49 = $48&1; //@line 6770 "libraries/glad/glad.cpp"
 HEAP8[1252040] = $49; //@line 6770 "libraries/glad/glad.cpp"
 $50 = (__ZN4gladL7has_extEPKc(464286)|0); //@line 6771 "libraries/glad/glad.cpp"
 $51 = $50&1; //@line 6771 "libraries/glad/glad.cpp"
 HEAP8[1252041] = $51; //@line 6771 "libraries/glad/glad.cpp"
 $52 = (__ZN4gladL7has_extEPKc(464312)|0); //@line 6772 "libraries/glad/glad.cpp"
 $53 = $52&1; //@line 6772 "libraries/glad/glad.cpp"
 HEAP8[1252042] = $53; //@line 6772 "libraries/glad/glad.cpp"
 $54 = (__ZN4gladL7has_extEPKc(464331)|0); //@line 6773 "libraries/glad/glad.cpp"
 $55 = $54&1; //@line 6773 "libraries/glad/glad.cpp"
 HEAP8[1252043] = $55; //@line 6773 "libraries/glad/glad.cpp"
 $56 = (__ZN4gladL7has_extEPKc(464349)|0); //@line 6774 "libraries/glad/glad.cpp"
 $57 = $56&1; //@line 6774 "libraries/glad/glad.cpp"
 HEAP8[1252044] = $57; //@line 6774 "libraries/glad/glad.cpp"
 $58 = (__ZN4gladL7has_extEPKc(464370)|0); //@line 6775 "libraries/glad/glad.cpp"
 $59 = $58&1; //@line 6775 "libraries/glad/glad.cpp"
 HEAP8[1252045] = $59; //@line 6775 "libraries/glad/glad.cpp"
 $60 = (__ZN4gladL7has_extEPKc(464390)|0); //@line 6776 "libraries/glad/glad.cpp"
 $61 = $60&1; //@line 6776 "libraries/glad/glad.cpp"
 HEAP8[1252046] = $61; //@line 6776 "libraries/glad/glad.cpp"
 $62 = (__ZN4gladL7has_extEPKc(464416)|0); //@line 6777 "libraries/glad/glad.cpp"
 $63 = $62&1; //@line 6777 "libraries/glad/glad.cpp"
 HEAP8[1252047] = $63; //@line 6777 "libraries/glad/glad.cpp"
 $64 = (__ZN4gladL7has_extEPKc(464435)|0); //@line 6778 "libraries/glad/glad.cpp"
 $65 = $64&1; //@line 6778 "libraries/glad/glad.cpp"
 HEAP8[1252048] = $65; //@line 6778 "libraries/glad/glad.cpp"
 $66 = (__ZN4gladL7has_extEPKc(464461)|0); //@line 6779 "libraries/glad/glad.cpp"
 $67 = $66&1; //@line 6779 "libraries/glad/glad.cpp"
 HEAP8[1252049] = $67; //@line 6779 "libraries/glad/glad.cpp"
 $68 = (__ZN4gladL7has_extEPKc(464488)|0); //@line 6780 "libraries/glad/glad.cpp"
 $69 = $68&1; //@line 6780 "libraries/glad/glad.cpp"
 HEAP8[1252050] = $69; //@line 6780 "libraries/glad/glad.cpp"
 $70 = (__ZN4gladL7has_extEPKc(464514)|0); //@line 6781 "libraries/glad/glad.cpp"
 $71 = $70&1; //@line 6781 "libraries/glad/glad.cpp"
 HEAP8[1252051] = $71; //@line 6781 "libraries/glad/glad.cpp"
 $72 = (__ZN4gladL7has_extEPKc(464547)|0); //@line 6782 "libraries/glad/glad.cpp"
 $73 = $72&1; //@line 6782 "libraries/glad/glad.cpp"
 HEAP8[1252052] = $73; //@line 6782 "libraries/glad/glad.cpp"
 $74 = (__ZN4gladL7has_extEPKc(464568)|0); //@line 6783 "libraries/glad/glad.cpp"
 $75 = $74&1; //@line 6783 "libraries/glad/glad.cpp"
 HEAP8[1252053] = $75; //@line 6783 "libraries/glad/glad.cpp"
 $76 = (__ZN4gladL7has_extEPKc(464590)|0); //@line 6784 "libraries/glad/glad.cpp"
 $77 = $76&1; //@line 6784 "libraries/glad/glad.cpp"
 HEAP8[1252054] = $77; //@line 6784 "libraries/glad/glad.cpp"
 $78 = (__ZN4gladL7has_extEPKc(464614)|0); //@line 6785 "libraries/glad/glad.cpp"
 $79 = $78&1; //@line 6785 "libraries/glad/glad.cpp"
 HEAP8[1252055] = $79; //@line 6785 "libraries/glad/glad.cpp"
 $80 = (__ZN4gladL7has_extEPKc(464639)|0); //@line 6786 "libraries/glad/glad.cpp"
 $81 = $80&1; //@line 6786 "libraries/glad/glad.cpp"
 HEAP8[1252056] = $81; //@line 6786 "libraries/glad/glad.cpp"
 $82 = (__ZN4gladL7has_extEPKc(464666)|0); //@line 6787 "libraries/glad/glad.cpp"
 $83 = $82&1; //@line 6787 "libraries/glad/glad.cpp"
 HEAP8[1252057] = $83; //@line 6787 "libraries/glad/glad.cpp"
 $84 = (__ZN4gladL7has_extEPKc(464693)|0); //@line 6788 "libraries/glad/glad.cpp"
 $85 = $84&1; //@line 6788 "libraries/glad/glad.cpp"
 HEAP8[1252058] = $85; //@line 6788 "libraries/glad/glad.cpp"
 $86 = (__ZN4gladL7has_extEPKc(464718)|0); //@line 6789 "libraries/glad/glad.cpp"
 $87 = $86&1; //@line 6789 "libraries/glad/glad.cpp"
 HEAP8[1252059] = $87; //@line 6789 "libraries/glad/glad.cpp"
 $88 = (__ZN4gladL7has_extEPKc(464750)|0); //@line 6790 "libraries/glad/glad.cpp"
 $89 = $88&1; //@line 6790 "libraries/glad/glad.cpp"
 HEAP8[1252060] = $89; //@line 6790 "libraries/glad/glad.cpp"
 $90 = (__ZN4gladL7has_extEPKc(464783)|0); //@line 6791 "libraries/glad/glad.cpp"
 $91 = $90&1; //@line 6791 "libraries/glad/glad.cpp"
 HEAP8[1252061] = $91; //@line 6791 "libraries/glad/glad.cpp"
 $92 = (__ZN4gladL7has_extEPKc(464817)|0); //@line 6792 "libraries/glad/glad.cpp"
 $93 = $92&1; //@line 6792 "libraries/glad/glad.cpp"
 HEAP8[1252062] = $93; //@line 6792 "libraries/glad/glad.cpp"
 $94 = (__ZN4gladL7has_extEPKc(464848)|0); //@line 6793 "libraries/glad/glad.cpp"
 $95 = $94&1; //@line 6793 "libraries/glad/glad.cpp"
 HEAP8[1252063] = $95; //@line 6793 "libraries/glad/glad.cpp"
 $96 = (__ZN4gladL7has_extEPKc(464881)|0); //@line 6794 "libraries/glad/glad.cpp"
 $97 = $96&1; //@line 6794 "libraries/glad/glad.cpp"
 HEAP8[1252064] = $97; //@line 6794 "libraries/glad/glad.cpp"
 $98 = (__ZN4gladL7has_extEPKc(464915)|0); //@line 6795 "libraries/glad/glad.cpp"
 $99 = $98&1; //@line 6795 "libraries/glad/glad.cpp"
 HEAP8[1252065] = $99; //@line 6795 "libraries/glad/glad.cpp"
 $100 = (__ZN4gladL7has_extEPKc(464941)|0); //@line 6796 "libraries/glad/glad.cpp"
 $101 = $100&1; //@line 6796 "libraries/glad/glad.cpp"
 HEAP8[1252066] = $101; //@line 6796 "libraries/glad/glad.cpp"
 $102 = (__ZN4gladL7has_extEPKc(464965)|0); //@line 6797 "libraries/glad/glad.cpp"
 $103 = $102&1; //@line 6797 "libraries/glad/glad.cpp"
 HEAP8[1252067] = $103; //@line 6797 "libraries/glad/glad.cpp"
 $104 = (__ZN4gladL7has_extEPKc(464989)|0); //@line 6798 "libraries/glad/glad.cpp"
 $105 = $104&1; //@line 6798 "libraries/glad/glad.cpp"
 HEAP8[1252068] = $105; //@line 6798 "libraries/glad/glad.cpp"
 $106 = (__ZN4gladL7has_extEPKc(465015)|0); //@line 6799 "libraries/glad/glad.cpp"
 $107 = $106&1; //@line 6799 "libraries/glad/glad.cpp"
 HEAP8[1252069] = $107; //@line 6799 "libraries/glad/glad.cpp"
 $108 = (__ZN4gladL7has_extEPKc(465044)|0); //@line 6800 "libraries/glad/glad.cpp"
 $109 = $108&1; //@line 6800 "libraries/glad/glad.cpp"
 HEAP8[1252070] = $109; //@line 6800 "libraries/glad/glad.cpp"
 $110 = (__ZN4gladL7has_extEPKc(465063)|0); //@line 6801 "libraries/glad/glad.cpp"
 $111 = $110&1; //@line 6801 "libraries/glad/glad.cpp"
 HEAP8[1252071] = $111; //@line 6801 "libraries/glad/glad.cpp"
 $112 = (__ZN4gladL7has_extEPKc(465086)|0); //@line 6802 "libraries/glad/glad.cpp"
 $113 = $112&1; //@line 6802 "libraries/glad/glad.cpp"
 HEAP8[1252072] = $113; //@line 6802 "libraries/glad/glad.cpp"
 $114 = (__ZN4gladL7has_extEPKc(465110)|0); //@line 6803 "libraries/glad/glad.cpp"
 $115 = $114&1; //@line 6803 "libraries/glad/glad.cpp"
 HEAP8[1252073] = $115; //@line 6803 "libraries/glad/glad.cpp"
 $116 = (__ZN4gladL7has_extEPKc(465134)|0); //@line 6804 "libraries/glad/glad.cpp"
 $117 = $116&1; //@line 6804 "libraries/glad/glad.cpp"
 HEAP8[1252074] = $117; //@line 6804 "libraries/glad/glad.cpp"
 $118 = (__ZN4gladL7has_extEPKc(465159)|0); //@line 6805 "libraries/glad/glad.cpp"
 $119 = $118&1; //@line 6805 "libraries/glad/glad.cpp"
 HEAP8[1252075] = $119; //@line 6805 "libraries/glad/glad.cpp"
 $120 = (__ZN4gladL7has_extEPKc(465186)|0); //@line 6806 "libraries/glad/glad.cpp"
 $121 = $120&1; //@line 6806 "libraries/glad/glad.cpp"
 HEAP8[1252076] = $121; //@line 6806 "libraries/glad/glad.cpp"
 $122 = (__ZN4gladL7has_extEPKc(465210)|0); //@line 6807 "libraries/glad/glad.cpp"
 $123 = $122&1; //@line 6807 "libraries/glad/glad.cpp"
 HEAP8[1252077] = $123; //@line 6807 "libraries/glad/glad.cpp"
 $124 = (__ZN4gladL7has_extEPKc(465238)|0); //@line 6808 "libraries/glad/glad.cpp"
 $125 = $124&1; //@line 6808 "libraries/glad/glad.cpp"
 HEAP8[1252078] = $125; //@line 6808 "libraries/glad/glad.cpp"
 $126 = (__ZN4gladL7has_extEPKc(465267)|0); //@line 6809 "libraries/glad/glad.cpp"
 $127 = $126&1; //@line 6809 "libraries/glad/glad.cpp"
 HEAP8[1252079] = $127; //@line 6809 "libraries/glad/glad.cpp"
 $128 = (__ZN4gladL7has_extEPKc(465293)|0); //@line 6810 "libraries/glad/glad.cpp"
 $129 = $128&1; //@line 6810 "libraries/glad/glad.cpp"
 HEAP8[1252080] = $129; //@line 6810 "libraries/glad/glad.cpp"
 $130 = (__ZN4gladL7has_extEPKc(465321)|0); //@line 6811 "libraries/glad/glad.cpp"
 $131 = $130&1; //@line 6811 "libraries/glad/glad.cpp"
 HEAP8[1252081] = $131; //@line 6811 "libraries/glad/glad.cpp"
 $132 = (__ZN4gladL7has_extEPKc(465345)|0); //@line 6812 "libraries/glad/glad.cpp"
 $133 = $132&1; //@line 6812 "libraries/glad/glad.cpp"
 HEAP8[1252082] = $133; //@line 6812 "libraries/glad/glad.cpp"
 $134 = (__ZN4gladL7has_extEPKc(465363)|0); //@line 6813 "libraries/glad/glad.cpp"
 $135 = $134&1; //@line 6813 "libraries/glad/glad.cpp"
 HEAP8[1252083] = $135; //@line 6813 "libraries/glad/glad.cpp"
 $136 = (__ZN4gladL7has_extEPKc(465390)|0); //@line 6814 "libraries/glad/glad.cpp"
 $137 = $136&1; //@line 6814 "libraries/glad/glad.cpp"
 HEAP8[1252084] = $137; //@line 6814 "libraries/glad/glad.cpp"
 $138 = (__ZN4gladL7has_extEPKc(465414)|0); //@line 6815 "libraries/glad/glad.cpp"
 $139 = $138&1; //@line 6815 "libraries/glad/glad.cpp"
 HEAP8[1252085] = $139; //@line 6815 "libraries/glad/glad.cpp"
 $140 = (__ZN4gladL7has_extEPKc(465445)|0); //@line 6816 "libraries/glad/glad.cpp"
 $141 = $140&1; //@line 6816 "libraries/glad/glad.cpp"
 HEAP8[1252086] = $141; //@line 6816 "libraries/glad/glad.cpp"
 $142 = (__ZN4gladL7has_extEPKc(465478)|0); //@line 6817 "libraries/glad/glad.cpp"
 $143 = $142&1; //@line 6817 "libraries/glad/glad.cpp"
 HEAP8[1252087] = $143; //@line 6817 "libraries/glad/glad.cpp"
 $144 = (__ZN4gladL7has_extEPKc(465505)|0); //@line 6818 "libraries/glad/glad.cpp"
 $145 = $144&1; //@line 6818 "libraries/glad/glad.cpp"
 HEAP8[1252088] = $145; //@line 6818 "libraries/glad/glad.cpp"
 $146 = (__ZN4gladL7has_extEPKc(465532)|0); //@line 6819 "libraries/glad/glad.cpp"
 $147 = $146&1; //@line 6819 "libraries/glad/glad.cpp"
 HEAP8[1252089] = $147; //@line 6819 "libraries/glad/glad.cpp"
 $148 = (__ZN4gladL7has_extEPKc(465563)|0); //@line 6820 "libraries/glad/glad.cpp"
 $149 = $148&1; //@line 6820 "libraries/glad/glad.cpp"
 HEAP8[1252090] = $149; //@line 6820 "libraries/glad/glad.cpp"
 $150 = (__ZN4gladL7has_extEPKc(465587)|0); //@line 6821 "libraries/glad/glad.cpp"
 $151 = $150&1; //@line 6821 "libraries/glad/glad.cpp"
 HEAP8[1252091] = $151; //@line 6821 "libraries/glad/glad.cpp"
 $152 = (__ZN4gladL7has_extEPKc(465614)|0); //@line 6822 "libraries/glad/glad.cpp"
 $153 = $152&1; //@line 6822 "libraries/glad/glad.cpp"
 HEAP8[1252092] = $153; //@line 6822 "libraries/glad/glad.cpp"
 $154 = (__ZN4gladL7has_extEPKc(465651)|0); //@line 6823 "libraries/glad/glad.cpp"
 $155 = $154&1; //@line 6823 "libraries/glad/glad.cpp"
 HEAP8[1252093] = $155; //@line 6823 "libraries/glad/glad.cpp"
 $156 = (__ZN4gladL7has_extEPKc(465669)|0); //@line 6824 "libraries/glad/glad.cpp"
 $157 = $156&1; //@line 6824 "libraries/glad/glad.cpp"
 HEAP8[1252094] = $157; //@line 6824 "libraries/glad/glad.cpp"
 $158 = (__ZN4gladL7has_extEPKc(465697)|0); //@line 6825 "libraries/glad/glad.cpp"
 $159 = $158&1; //@line 6825 "libraries/glad/glad.cpp"
 HEAP8[1252095] = $159; //@line 6825 "libraries/glad/glad.cpp"
 $160 = (__ZN4gladL7has_extEPKc(465721)|0); //@line 6826 "libraries/glad/glad.cpp"
 $161 = $160&1; //@line 6826 "libraries/glad/glad.cpp"
 HEAP8[1252096] = $161; //@line 6826 "libraries/glad/glad.cpp"
 $162 = (__ZN4gladL7has_extEPKc(465743)|0); //@line 6827 "libraries/glad/glad.cpp"
 $163 = $162&1; //@line 6827 "libraries/glad/glad.cpp"
 HEAP8[1252097] = $163; //@line 6827 "libraries/glad/glad.cpp"
 $164 = (__ZN4gladL7has_extEPKc(465766)|0); //@line 6828 "libraries/glad/glad.cpp"
 $165 = $164&1; //@line 6828 "libraries/glad/glad.cpp"
 HEAP8[1252098] = $165; //@line 6828 "libraries/glad/glad.cpp"
 $166 = (__ZN4gladL7has_extEPKc(465791)|0); //@line 6829 "libraries/glad/glad.cpp"
 $167 = $166&1; //@line 6829 "libraries/glad/glad.cpp"
 HEAP8[1252099] = $167; //@line 6829 "libraries/glad/glad.cpp"
 $168 = (__ZN4gladL7has_extEPKc(465827)|0); //@line 6830 "libraries/glad/glad.cpp"
 $169 = $168&1; //@line 6830 "libraries/glad/glad.cpp"
 HEAP8[1252100] = $169; //@line 6830 "libraries/glad/glad.cpp"
 $170 = (__ZN4gladL7has_extEPKc(465858)|0); //@line 6831 "libraries/glad/glad.cpp"
 $171 = $170&1; //@line 6831 "libraries/glad/glad.cpp"
 HEAP8[1252101] = $171; //@line 6831 "libraries/glad/glad.cpp"
 $172 = (__ZN4gladL7has_extEPKc(465891)|0); //@line 6832 "libraries/glad/glad.cpp"
 $173 = $172&1; //@line 6832 "libraries/glad/glad.cpp"
 HEAP8[1252102] = $173; //@line 6832 "libraries/glad/glad.cpp"
 $174 = (__ZN4gladL7has_extEPKc(465921)|0); //@line 6833 "libraries/glad/glad.cpp"
 $175 = $174&1; //@line 6833 "libraries/glad/glad.cpp"
 HEAP8[1252103] = $175; //@line 6833 "libraries/glad/glad.cpp"
 $176 = (__ZN4gladL7has_extEPKc(465942)|0); //@line 6834 "libraries/glad/glad.cpp"
 $177 = $176&1; //@line 6834 "libraries/glad/glad.cpp"
 HEAP8[1252104] = $177; //@line 6834 "libraries/glad/glad.cpp"
 $178 = (__ZN4gladL7has_extEPKc(465969)|0); //@line 6835 "libraries/glad/glad.cpp"
 $179 = $178&1; //@line 6835 "libraries/glad/glad.cpp"
 HEAP8[1252105] = $179; //@line 6835 "libraries/glad/glad.cpp"
 $180 = (__ZN4gladL7has_extEPKc(465989)|0); //@line 6836 "libraries/glad/glad.cpp"
 $181 = $180&1; //@line 6836 "libraries/glad/glad.cpp"
 HEAP8[1252106] = $181; //@line 6836 "libraries/glad/glad.cpp"
 $182 = (__ZN4gladL7has_extEPKc(466019)|0); //@line 6837 "libraries/glad/glad.cpp"
 $183 = $182&1; //@line 6837 "libraries/glad/glad.cpp"
 HEAP8[1252107] = $183; //@line 6837 "libraries/glad/glad.cpp"
 $184 = (__ZN4gladL7has_extEPKc(466044)|0); //@line 6838 "libraries/glad/glad.cpp"
 $185 = $184&1; //@line 6838 "libraries/glad/glad.cpp"
 HEAP8[1252108] = $185; //@line 6838 "libraries/glad/glad.cpp"
 $186 = (__ZN4gladL7has_extEPKc(466075)|0); //@line 6839 "libraries/glad/glad.cpp"
 $187 = $186&1; //@line 6839 "libraries/glad/glad.cpp"
 HEAP8[1252109] = $187; //@line 6839 "libraries/glad/glad.cpp"
 $188 = (__ZN4gladL7has_extEPKc(466100)|0); //@line 6840 "libraries/glad/glad.cpp"
 $189 = $188&1; //@line 6840 "libraries/glad/glad.cpp"
 HEAP8[1252110] = $189; //@line 6840 "libraries/glad/glad.cpp"
 $190 = (__ZN4gladL7has_extEPKc(466124)|0); //@line 6841 "libraries/glad/glad.cpp"
 $191 = $190&1; //@line 6841 "libraries/glad/glad.cpp"
 HEAP8[1252111] = $191; //@line 6841 "libraries/glad/glad.cpp"
 $192 = (__ZN4gladL7has_extEPKc(466153)|0); //@line 6842 "libraries/glad/glad.cpp"
 $193 = $192&1; //@line 6842 "libraries/glad/glad.cpp"
 HEAP8[1252112] = $193; //@line 6842 "libraries/glad/glad.cpp"
 $194 = (__ZN4gladL7has_extEPKc(466189)|0); //@line 6843 "libraries/glad/glad.cpp"
 $195 = $194&1; //@line 6843 "libraries/glad/glad.cpp"
 HEAP8[1252113] = $195; //@line 6843 "libraries/glad/glad.cpp"
 $196 = (__ZN4gladL7has_extEPKc(466214)|0); //@line 6844 "libraries/glad/glad.cpp"
 $197 = $196&1; //@line 6844 "libraries/glad/glad.cpp"
 HEAP8[1252114] = $197; //@line 6844 "libraries/glad/glad.cpp"
 $198 = (__ZN4gladL7has_extEPKc(466250)|0); //@line 6845 "libraries/glad/glad.cpp"
 $199 = $198&1; //@line 6845 "libraries/glad/glad.cpp"
 HEAP8[1252115] = $199; //@line 6845 "libraries/glad/glad.cpp"
 $200 = (__ZN4gladL7has_extEPKc(466276)|0); //@line 6846 "libraries/glad/glad.cpp"
 $201 = $200&1; //@line 6846 "libraries/glad/glad.cpp"
 HEAP8[1252116] = $201; //@line 6846 "libraries/glad/glad.cpp"
 $202 = (__ZN4gladL7has_extEPKc(466311)|0); //@line 6847 "libraries/glad/glad.cpp"
 $203 = $202&1; //@line 6847 "libraries/glad/glad.cpp"
 HEAP8[1252117] = $203; //@line 6847 "libraries/glad/glad.cpp"
 $204 = (__ZN4gladL7has_extEPKc(466343)|0); //@line 6848 "libraries/glad/glad.cpp"
 $205 = $204&1; //@line 6848 "libraries/glad/glad.cpp"
 HEAP8[1252118] = $205; //@line 6848 "libraries/glad/glad.cpp"
 $206 = (__ZN4gladL7has_extEPKc(466375)|0); //@line 6849 "libraries/glad/glad.cpp"
 $207 = $206&1; //@line 6849 "libraries/glad/glad.cpp"
 HEAP8[1252119] = $207; //@line 6849 "libraries/glad/glad.cpp"
 $208 = (__ZN4gladL7has_extEPKc(466407)|0); //@line 6850 "libraries/glad/glad.cpp"
 $209 = $208&1; //@line 6850 "libraries/glad/glad.cpp"
 HEAP8[1252120] = $209; //@line 6850 "libraries/glad/glad.cpp"
 $210 = (__ZN4gladL7has_extEPKc(466428)|0); //@line 6851 "libraries/glad/glad.cpp"
 $211 = $210&1; //@line 6851 "libraries/glad/glad.cpp"
 HEAP8[1252121] = $211; //@line 6851 "libraries/glad/glad.cpp"
 $212 = (__ZN4gladL7has_extEPKc(466450)|0); //@line 6852 "libraries/glad/glad.cpp"
 $213 = $212&1; //@line 6852 "libraries/glad/glad.cpp"
 HEAP8[1252122] = $213; //@line 6852 "libraries/glad/glad.cpp"
 $214 = (__ZN4gladL7has_extEPKc(466473)|0); //@line 6853 "libraries/glad/glad.cpp"
 $215 = $214&1; //@line 6853 "libraries/glad/glad.cpp"
 HEAP8[1252123] = $215; //@line 6853 "libraries/glad/glad.cpp"
 $216 = (__ZN4gladL7has_extEPKc(466501)|0); //@line 6854 "libraries/glad/glad.cpp"
 $217 = $216&1; //@line 6854 "libraries/glad/glad.cpp"
 HEAP8[1252124] = $217; //@line 6854 "libraries/glad/glad.cpp"
 $218 = (__ZN4gladL7has_extEPKc(466526)|0); //@line 6855 "libraries/glad/glad.cpp"
 $219 = $218&1; //@line 6855 "libraries/glad/glad.cpp"
 HEAP8[1252125] = $219; //@line 6855 "libraries/glad/glad.cpp"
 $220 = (__ZN4gladL7has_extEPKc(466538)|0); //@line 6856 "libraries/glad/glad.cpp"
 $221 = $220&1; //@line 6856 "libraries/glad/glad.cpp"
 HEAP8[1252126] = $221; //@line 6856 "libraries/glad/glad.cpp"
 $222 = (__ZN4gladL7has_extEPKc(466565)|0); //@line 6857 "libraries/glad/glad.cpp"
 $223 = $222&1; //@line 6857 "libraries/glad/glad.cpp"
 HEAP8[1252127] = $223; //@line 6857 "libraries/glad/glad.cpp"
 $224 = (__ZN4gladL7has_extEPKc(466588)|0); //@line 6858 "libraries/glad/glad.cpp"
 $225 = $224&1; //@line 6858 "libraries/glad/glad.cpp"
 HEAP8[1252128] = $225; //@line 6858 "libraries/glad/glad.cpp"
 $226 = (__ZN4gladL7has_extEPKc(466617)|0); //@line 6859 "libraries/glad/glad.cpp"
 $227 = $226&1; //@line 6859 "libraries/glad/glad.cpp"
 HEAP8[1252129] = $227; //@line 6859 "libraries/glad/glad.cpp"
 $228 = (__ZN4gladL7has_extEPKc(466652)|0); //@line 6860 "libraries/glad/glad.cpp"
 $229 = $228&1; //@line 6860 "libraries/glad/glad.cpp"
 HEAP8[1252130] = $229; //@line 6860 "libraries/glad/glad.cpp"
 $230 = (__ZN4gladL7has_extEPKc(466680)|0); //@line 6861 "libraries/glad/glad.cpp"
 $231 = $230&1; //@line 6861 "libraries/glad/glad.cpp"
 HEAP8[1252131] = $231; //@line 6861 "libraries/glad/glad.cpp"
 $232 = (__ZN4gladL7has_extEPKc(466712)|0); //@line 6862 "libraries/glad/glad.cpp"
 $233 = $232&1; //@line 6862 "libraries/glad/glad.cpp"
 HEAP8[1252132] = $233; //@line 6862 "libraries/glad/glad.cpp"
 $234 = (__ZN4gladL7has_extEPKc(466744)|0); //@line 6863 "libraries/glad/glad.cpp"
 $235 = $234&1; //@line 6863 "libraries/glad/glad.cpp"
 HEAP8[1252133] = $235; //@line 6863 "libraries/glad/glad.cpp"
 $236 = (__ZN4gladL7has_extEPKc(466774)|0); //@line 6864 "libraries/glad/glad.cpp"
 $237 = $236&1; //@line 6864 "libraries/glad/glad.cpp"
 HEAP8[1252134] = $237; //@line 6864 "libraries/glad/glad.cpp"
 $238 = (__ZN4gladL7has_extEPKc(466803)|0); //@line 6865 "libraries/glad/glad.cpp"
 $239 = $238&1; //@line 6865 "libraries/glad/glad.cpp"
 HEAP8[1252135] = $239; //@line 6865 "libraries/glad/glad.cpp"
 $240 = (__ZN4gladL7has_extEPKc(466824)|0); //@line 6866 "libraries/glad/glad.cpp"
 $241 = $240&1; //@line 6866 "libraries/glad/glad.cpp"
 HEAP8[1252136] = $241; //@line 6866 "libraries/glad/glad.cpp"
 $242 = (__ZN4gladL7has_extEPKc(466846)|0); //@line 6867 "libraries/glad/glad.cpp"
 $243 = $242&1; //@line 6867 "libraries/glad/glad.cpp"
 HEAP8[1252137] = $243; //@line 6867 "libraries/glad/glad.cpp"
 $244 = (__ZN4gladL7has_extEPKc(466882)|0); //@line 6868 "libraries/glad/glad.cpp"
 $245 = $244&1; //@line 6868 "libraries/glad/glad.cpp"
 HEAP8[1252138] = $245; //@line 6868 "libraries/glad/glad.cpp"
 $246 = (__ZN4gladL7has_extEPKc(466909)|0); //@line 6869 "libraries/glad/glad.cpp"
 $247 = $246&1; //@line 6869 "libraries/glad/glad.cpp"
 HEAP8[1252139] = $247; //@line 6869 "libraries/glad/glad.cpp"
 $248 = (__ZN4gladL7has_extEPKc(466937)|0); //@line 6870 "libraries/glad/glad.cpp"
 $249 = $248&1; //@line 6870 "libraries/glad/glad.cpp"
 HEAP8[1252140] = $249; //@line 6870 "libraries/glad/glad.cpp"
 $250 = (__ZN4gladL7has_extEPKc(466962)|0); //@line 6871 "libraries/glad/glad.cpp"
 $251 = $250&1; //@line 6871 "libraries/glad/glad.cpp"
 HEAP8[1252141] = $251; //@line 6871 "libraries/glad/glad.cpp"
 $252 = (__ZN4gladL7has_extEPKc(466987)|0); //@line 6872 "libraries/glad/glad.cpp"
 $253 = $252&1; //@line 6872 "libraries/glad/glad.cpp"
 HEAP8[1252142] = $253; //@line 6872 "libraries/glad/glad.cpp"
 $254 = (__ZN4gladL7has_extEPKc(467005)|0); //@line 6873 "libraries/glad/glad.cpp"
 $255 = $254&1; //@line 6873 "libraries/glad/glad.cpp"
 HEAP8[1252143] = $255; //@line 6873 "libraries/glad/glad.cpp"
 $256 = (__ZN4gladL7has_extEPKc(467031)|0); //@line 6874 "libraries/glad/glad.cpp"
 $257 = $256&1; //@line 6874 "libraries/glad/glad.cpp"
 HEAP8[1252144] = $257; //@line 6874 "libraries/glad/glad.cpp"
 $258 = (__ZN4gladL7has_extEPKc(467055)|0); //@line 6875 "libraries/glad/glad.cpp"
 $259 = $258&1; //@line 6875 "libraries/glad/glad.cpp"
 HEAP8[1252145] = $259; //@line 6875 "libraries/glad/glad.cpp"
 $260 = (__ZN4gladL7has_extEPKc(467078)|0); //@line 6876 "libraries/glad/glad.cpp"
 $261 = $260&1; //@line 6876 "libraries/glad/glad.cpp"
 HEAP8[1252146] = $261; //@line 6876 "libraries/glad/glad.cpp"
 $262 = (__ZN4gladL7has_extEPKc(467113)|0); //@line 6877 "libraries/glad/glad.cpp"
 $263 = $262&1; //@line 6877 "libraries/glad/glad.cpp"
 HEAP8[1252147] = $263; //@line 6877 "libraries/glad/glad.cpp"
 $264 = (__ZN4gladL7has_extEPKc(467136)|0); //@line 6878 "libraries/glad/glad.cpp"
 $265 = $264&1; //@line 6878 "libraries/glad/glad.cpp"
 HEAP8[1252148] = $265; //@line 6878 "libraries/glad/glad.cpp"
 $266 = (__ZN4gladL7has_extEPKc(467156)|0); //@line 6879 "libraries/glad/glad.cpp"
 $267 = $266&1; //@line 6879 "libraries/glad/glad.cpp"
 HEAP8[1252149] = $267; //@line 6879 "libraries/glad/glad.cpp"
 $268 = (__ZN4gladL7has_extEPKc(467175)|0); //@line 6880 "libraries/glad/glad.cpp"
 $269 = $268&1; //@line 6880 "libraries/glad/glad.cpp"
 HEAP8[1252150] = $269; //@line 6880 "libraries/glad/glad.cpp"
 $270 = (__ZN4gladL7has_extEPKc(467202)|0); //@line 6881 "libraries/glad/glad.cpp"
 $271 = $270&1; //@line 6881 "libraries/glad/glad.cpp"
 HEAP8[1252151] = $271; //@line 6881 "libraries/glad/glad.cpp"
 $272 = (__ZN4gladL7has_extEPKc(467229)|0); //@line 6882 "libraries/glad/glad.cpp"
 $273 = $272&1; //@line 6882 "libraries/glad/glad.cpp"
 HEAP8[1252152] = $273; //@line 6882 "libraries/glad/glad.cpp"
 $274 = (__ZN4gladL7has_extEPKc(467265)|0); //@line 6883 "libraries/glad/glad.cpp"
 $275 = $274&1; //@line 6883 "libraries/glad/glad.cpp"
 HEAP8[1252153] = $275; //@line 6883 "libraries/glad/glad.cpp"
 $276 = (__ZN4gladL7has_extEPKc(467306)|0); //@line 6884 "libraries/glad/glad.cpp"
 $277 = $276&1; //@line 6884 "libraries/glad/glad.cpp"
 HEAP8[1252154] = $277; //@line 6884 "libraries/glad/glad.cpp"
 $278 = (__ZN4gladL7has_extEPKc(467335)|0); //@line 6885 "libraries/glad/glad.cpp"
 $279 = $278&1; //@line 6885 "libraries/glad/glad.cpp"
 HEAP8[1252155] = $279; //@line 6885 "libraries/glad/glad.cpp"
 $280 = (__ZN4gladL7has_extEPKc(467360)|0); //@line 6886 "libraries/glad/glad.cpp"
 $281 = $280&1; //@line 6886 "libraries/glad/glad.cpp"
 HEAP8[1252156] = $281; //@line 6886 "libraries/glad/glad.cpp"
 $282 = (__ZN4gladL7has_extEPKc(467387)|0); //@line 6887 "libraries/glad/glad.cpp"
 $283 = $282&1; //@line 6887 "libraries/glad/glad.cpp"
 HEAP8[1252157] = $283; //@line 6887 "libraries/glad/glad.cpp"
 $284 = (__ZN4gladL7has_extEPKc(467414)|0); //@line 6888 "libraries/glad/glad.cpp"
 $285 = $284&1; //@line 6888 "libraries/glad/glad.cpp"
 HEAP8[1252158] = $285; //@line 6888 "libraries/glad/glad.cpp"
 $286 = (__ZN4gladL7has_extEPKc(467443)|0); //@line 6889 "libraries/glad/glad.cpp"
 $287 = $286&1; //@line 6889 "libraries/glad/glad.cpp"
 HEAP8[1252159] = $287; //@line 6889 "libraries/glad/glad.cpp"
 $288 = (__ZN4gladL7has_extEPKc(467478)|0); //@line 6890 "libraries/glad/glad.cpp"
 $289 = $288&1; //@line 6890 "libraries/glad/glad.cpp"
 HEAP8[1252160] = $289; //@line 6890 "libraries/glad/glad.cpp"
 $290 = (__ZN4gladL7has_extEPKc(467512)|0); //@line 6891 "libraries/glad/glad.cpp"
 $291 = $290&1; //@line 6891 "libraries/glad/glad.cpp"
 HEAP8[1252161] = $291; //@line 6891 "libraries/glad/glad.cpp"
 $292 = (__ZN4gladL7has_extEPKc(467534)|0); //@line 6892 "libraries/glad/glad.cpp"
 $293 = $292&1; //@line 6892 "libraries/glad/glad.cpp"
 HEAP8[1252162] = $293; //@line 6892 "libraries/glad/glad.cpp"
 $294 = (__ZN4gladL7has_extEPKc(467555)|0); //@line 6893 "libraries/glad/glad.cpp"
 $295 = $294&1; //@line 6893 "libraries/glad/glad.cpp"
 HEAP8[1252163] = $295; //@line 6893 "libraries/glad/glad.cpp"
 $296 = (__ZN4gladL7has_extEPKc(467567)|0); //@line 6894 "libraries/glad/glad.cpp"
 $297 = $296&1; //@line 6894 "libraries/glad/glad.cpp"
 HEAP8[1252164] = $297; //@line 6894 "libraries/glad/glad.cpp"
 $298 = (__ZN4gladL7has_extEPKc(467591)|0); //@line 6895 "libraries/glad/glad.cpp"
 $299 = $298&1; //@line 6895 "libraries/glad/glad.cpp"
 HEAP8[1252165] = $299; //@line 6895 "libraries/glad/glad.cpp"
 $300 = (__ZN4gladL7has_extEPKc(467618)|0); //@line 6896 "libraries/glad/glad.cpp"
 $301 = $300&1; //@line 6896 "libraries/glad/glad.cpp"
 HEAP8[1252166] = $301; //@line 6896 "libraries/glad/glad.cpp"
 $302 = (__ZN4gladL7has_extEPKc(467638)|0); //@line 6897 "libraries/glad/glad.cpp"
 $303 = $302&1; //@line 6897 "libraries/glad/glad.cpp"
 HEAP8[1252167] = $303; //@line 6897 "libraries/glad/glad.cpp"
 $304 = (__ZN4gladL7has_extEPKc(467660)|0); //@line 6898 "libraries/glad/glad.cpp"
 $305 = $304&1; //@line 6898 "libraries/glad/glad.cpp"
 HEAP8[1252168] = $305; //@line 6898 "libraries/glad/glad.cpp"
 $306 = (__ZN4gladL7has_extEPKc(467686)|0); //@line 6899 "libraries/glad/glad.cpp"
 $307 = $306&1; //@line 6899 "libraries/glad/glad.cpp"
 HEAP8[1252169] = $307; //@line 6899 "libraries/glad/glad.cpp"
 $308 = (__ZN4gladL7has_extEPKc(467717)|0); //@line 6900 "libraries/glad/glad.cpp"
 $309 = $308&1; //@line 6900 "libraries/glad/glad.cpp"
 HEAP8[1252170] = $309; //@line 6900 "libraries/glad/glad.cpp"
 $310 = (__ZN4gladL7has_extEPKc(467735)|0); //@line 6901 "libraries/glad/glad.cpp"
 $311 = $310&1; //@line 6901 "libraries/glad/glad.cpp"
 HEAP8[1252171] = $311; //@line 6901 "libraries/glad/glad.cpp"
 $312 = (__ZN4gladL7has_extEPKc(467754)|0); //@line 6902 "libraries/glad/glad.cpp"
 $313 = $312&1; //@line 6902 "libraries/glad/glad.cpp"
 HEAP8[1252172] = $313; //@line 6902 "libraries/glad/glad.cpp"
 $314 = (__ZN4gladL7has_extEPKc(467774)|0); //@line 6903 "libraries/glad/glad.cpp"
 $315 = $314&1; //@line 6903 "libraries/glad/glad.cpp"
 HEAP8[1252173] = $315; //@line 6903 "libraries/glad/glad.cpp"
 $316 = (__ZN4gladL7has_extEPKc(467801)|0); //@line 6904 "libraries/glad/glad.cpp"
 $317 = $316&1; //@line 6904 "libraries/glad/glad.cpp"
 HEAP8[1252174] = $317; //@line 6904 "libraries/glad/glad.cpp"
 $318 = (__ZN4gladL7has_extEPKc(467828)|0); //@line 6905 "libraries/glad/glad.cpp"
 $319 = $318&1; //@line 6905 "libraries/glad/glad.cpp"
 HEAP8[1252175] = $319; //@line 6905 "libraries/glad/glad.cpp"
 $320 = (__ZN4gladL7has_extEPKc(467856)|0); //@line 6906 "libraries/glad/glad.cpp"
 $321 = $320&1; //@line 6906 "libraries/glad/glad.cpp"
 HEAP8[1252176] = $321; //@line 6906 "libraries/glad/glad.cpp"
 $322 = (__ZN4gladL7has_extEPKc(467876)|0); //@line 6907 "libraries/glad/glad.cpp"
 $323 = $322&1; //@line 6907 "libraries/glad/glad.cpp"
 HEAP8[1252177] = $323; //@line 6907 "libraries/glad/glad.cpp"
 $324 = (__ZN4gladL7has_extEPKc(467897)|0); //@line 6908 "libraries/glad/glad.cpp"
 $325 = $324&1; //@line 6908 "libraries/glad/glad.cpp"
 HEAP8[1252178] = $325; //@line 6908 "libraries/glad/glad.cpp"
 $326 = (__ZN4gladL7has_extEPKc(467925)|0); //@line 6909 "libraries/glad/glad.cpp"
 $327 = $326&1; //@line 6909 "libraries/glad/glad.cpp"
 HEAP8[1252179] = $327; //@line 6909 "libraries/glad/glad.cpp"
 $328 = (__ZN4gladL7has_extEPKc(467958)|0); //@line 6910 "libraries/glad/glad.cpp"
 $329 = $328&1; //@line 6910 "libraries/glad/glad.cpp"
 HEAP8[1252180] = $329; //@line 6910 "libraries/glad/glad.cpp"
 $330 = (__ZN4gladL7has_extEPKc(467980)|0); //@line 6911 "libraries/glad/glad.cpp"
 $331 = $330&1; //@line 6911 "libraries/glad/glad.cpp"
 HEAP8[1252181] = $331; //@line 6911 "libraries/glad/glad.cpp"
 $332 = (__ZN4gladL7has_extEPKc(467999)|0); //@line 6912 "libraries/glad/glad.cpp"
 $333 = $332&1; //@line 6912 "libraries/glad/glad.cpp"
 HEAP8[1252182] = $333; //@line 6912 "libraries/glad/glad.cpp"
 $334 = (__ZN4gladL7has_extEPKc(468023)|0); //@line 6913 "libraries/glad/glad.cpp"
 $335 = $334&1; //@line 6913 "libraries/glad/glad.cpp"
 HEAP8[1252183] = $335; //@line 6913 "libraries/glad/glad.cpp"
 $336 = (__ZN4gladL7has_extEPKc(468054)|0); //@line 6914 "libraries/glad/glad.cpp"
 $337 = $336&1; //@line 6914 "libraries/glad/glad.cpp"
 HEAP8[1252184] = $337; //@line 6914 "libraries/glad/glad.cpp"
 $338 = (__ZN4gladL7has_extEPKc(468097)|0); //@line 6915 "libraries/glad/glad.cpp"
 $339 = $338&1; //@line 6915 "libraries/glad/glad.cpp"
 HEAP8[1252185] = $339; //@line 6915 "libraries/glad/glad.cpp"
 $340 = (__ZN4gladL7has_extEPKc(468123)|0); //@line 6916 "libraries/glad/glad.cpp"
 $341 = $340&1; //@line 6916 "libraries/glad/glad.cpp"
 HEAP8[1252186] = $341; //@line 6916 "libraries/glad/glad.cpp"
 $342 = (__ZN4gladL7has_extEPKc(468147)|0); //@line 6917 "libraries/glad/glad.cpp"
 $343 = $342&1; //@line 6917 "libraries/glad/glad.cpp"
 HEAP8[1252187] = $343; //@line 6917 "libraries/glad/glad.cpp"
 $344 = (__ZN4gladL7has_extEPKc(468174)|0); //@line 6918 "libraries/glad/glad.cpp"
 $345 = $344&1; //@line 6918 "libraries/glad/glad.cpp"
 HEAP8[1252188] = $345; //@line 6918 "libraries/glad/glad.cpp"
 $346 = (__ZN4gladL7has_extEPKc(468197)|0); //@line 6919 "libraries/glad/glad.cpp"
 $347 = $346&1; //@line 6919 "libraries/glad/glad.cpp"
 HEAP8[1252189] = $347; //@line 6919 "libraries/glad/glad.cpp"
 $348 = (__ZN4gladL7has_extEPKc(468221)|0); //@line 6920 "libraries/glad/glad.cpp"
 $349 = $348&1; //@line 6920 "libraries/glad/glad.cpp"
 HEAP8[1252190] = $349; //@line 6920 "libraries/glad/glad.cpp"
 $350 = (__ZN4gladL7has_extEPKc(468251)|0); //@line 6921 "libraries/glad/glad.cpp"
 $351 = $350&1; //@line 6921 "libraries/glad/glad.cpp"
 HEAP8[1252191] = $351; //@line 6921 "libraries/glad/glad.cpp"
 $352 = (__ZN4gladL7has_extEPKc(468270)|0); //@line 6922 "libraries/glad/glad.cpp"
 $353 = $352&1; //@line 6922 "libraries/glad/glad.cpp"
 HEAP8[1252192] = $353; //@line 6922 "libraries/glad/glad.cpp"
 $354 = (__ZN4gladL7has_extEPKc(468289)|0); //@line 6923 "libraries/glad/glad.cpp"
 $355 = $354&1; //@line 6923 "libraries/glad/glad.cpp"
 HEAP8[1252193] = $355; //@line 6923 "libraries/glad/glad.cpp"
 $356 = (__ZN4gladL7has_extEPKc(468313)|0); //@line 6924 "libraries/glad/glad.cpp"
 $357 = $356&1; //@line 6924 "libraries/glad/glad.cpp"
 HEAP8[1252194] = $357; //@line 6924 "libraries/glad/glad.cpp"
 $358 = (__ZN4gladL7has_extEPKc(468337)|0); //@line 6925 "libraries/glad/glad.cpp"
 $359 = $358&1; //@line 6925 "libraries/glad/glad.cpp"
 HEAP8[1252195] = $359; //@line 6925 "libraries/glad/glad.cpp"
 $360 = (__ZN4gladL7has_extEPKc(468364)|0); //@line 6926 "libraries/glad/glad.cpp"
 $361 = $360&1; //@line 6926 "libraries/glad/glad.cpp"
 HEAP8[1252196] = $361; //@line 6926 "libraries/glad/glad.cpp"
 $362 = (__ZN4gladL7has_extEPKc(468398)|0); //@line 6927 "libraries/glad/glad.cpp"
 $363 = $362&1; //@line 6927 "libraries/glad/glad.cpp"
 HEAP8[1252197] = $363; //@line 6927 "libraries/glad/glad.cpp"
 $364 = (__ZN4gladL7has_extEPKc(468436)|0); //@line 6928 "libraries/glad/glad.cpp"
 $365 = $364&1; //@line 6928 "libraries/glad/glad.cpp"
 HEAP8[1252198] = $365; //@line 6928 "libraries/glad/glad.cpp"
 $366 = (__ZN4gladL7has_extEPKc(468466)|0); //@line 6929 "libraries/glad/glad.cpp"
 $367 = $366&1; //@line 6929 "libraries/glad/glad.cpp"
 HEAP8[1252199] = $367; //@line 6929 "libraries/glad/glad.cpp"
 $368 = (__ZN4gladL7has_extEPKc(468497)|0); //@line 6930 "libraries/glad/glad.cpp"
 $369 = $368&1; //@line 6930 "libraries/glad/glad.cpp"
 HEAP8[1252200] = $369; //@line 6930 "libraries/glad/glad.cpp"
 $370 = (__ZN4gladL7has_extEPKc(468525)|0); //@line 6931 "libraries/glad/glad.cpp"
 $371 = $370&1; //@line 6931 "libraries/glad/glad.cpp"
 HEAP8[1252201] = $371; //@line 6931 "libraries/glad/glad.cpp"
 $372 = (__ZN4gladL7has_extEPKc(468545)|0); //@line 6932 "libraries/glad/glad.cpp"
 $373 = $372&1; //@line 6932 "libraries/glad/glad.cpp"
 HEAP8[1252202] = $373; //@line 6932 "libraries/glad/glad.cpp"
 $374 = (__ZN4gladL7has_extEPKc(468572)|0); //@line 6933 "libraries/glad/glad.cpp"
 $375 = $374&1; //@line 6933 "libraries/glad/glad.cpp"
 HEAP8[1252203] = $375; //@line 6933 "libraries/glad/glad.cpp"
 $376 = (__ZN4gladL7has_extEPKc(468600)|0); //@line 6934 "libraries/glad/glad.cpp"
 $377 = $376&1; //@line 6934 "libraries/glad/glad.cpp"
 HEAP8[1252204] = $377; //@line 6934 "libraries/glad/glad.cpp"
 $378 = (__ZN4gladL7has_extEPKc(468627)|0); //@line 6935 "libraries/glad/glad.cpp"
 $379 = $378&1; //@line 6935 "libraries/glad/glad.cpp"
 HEAP8[1252205] = $379; //@line 6935 "libraries/glad/glad.cpp"
 $380 = (__ZN4gladL7has_extEPKc(468657)|0); //@line 6936 "libraries/glad/glad.cpp"
 $381 = $380&1; //@line 6936 "libraries/glad/glad.cpp"
 HEAP8[1252206] = $381; //@line 6936 "libraries/glad/glad.cpp"
 $382 = (__ZN4gladL7has_extEPKc(468681)|0); //@line 6937 "libraries/glad/glad.cpp"
 $383 = $382&1; //@line 6937 "libraries/glad/glad.cpp"
 HEAP8[1252207] = $383; //@line 6937 "libraries/glad/glad.cpp"
 $384 = (__ZN4gladL7has_extEPKc(468699)|0); //@line 6938 "libraries/glad/glad.cpp"
 $385 = $384&1; //@line 6938 "libraries/glad/glad.cpp"
 HEAP8[1252208] = $385; //@line 6938 "libraries/glad/glad.cpp"
 $386 = (__ZN4gladL7has_extEPKc(468725)|0); //@line 6939 "libraries/glad/glad.cpp"
 $387 = $386&1; //@line 6939 "libraries/glad/glad.cpp"
 HEAP8[1252209] = $387; //@line 6939 "libraries/glad/glad.cpp"
 $388 = (__ZN4gladL7has_extEPKc(468749)|0); //@line 6940 "libraries/glad/glad.cpp"
 $389 = $388&1; //@line 6940 "libraries/glad/glad.cpp"
 HEAP8[1252210] = $389; //@line 6940 "libraries/glad/glad.cpp"
 $390 = (__ZN4gladL7has_extEPKc(468769)|0); //@line 6941 "libraries/glad/glad.cpp"
 $391 = $390&1; //@line 6941 "libraries/glad/glad.cpp"
 HEAP8[1252211] = $391; //@line 6941 "libraries/glad/glad.cpp"
 $392 = (__ZN4gladL7has_extEPKc(468787)|0); //@line 6942 "libraries/glad/glad.cpp"
 $393 = $392&1; //@line 6942 "libraries/glad/glad.cpp"
 HEAP8[1252212] = $393; //@line 6942 "libraries/glad/glad.cpp"
 $394 = (__ZN4gladL7has_extEPKc(468818)|0); //@line 6943 "libraries/glad/glad.cpp"
 $395 = $394&1; //@line 6943 "libraries/glad/glad.cpp"
 HEAP8[1252213] = $395; //@line 6943 "libraries/glad/glad.cpp"
 $396 = (__ZN4gladL7has_extEPKc(468850)|0); //@line 6944 "libraries/glad/glad.cpp"
 $397 = $396&1; //@line 6944 "libraries/glad/glad.cpp"
 HEAP8[1252214] = $397; //@line 6944 "libraries/glad/glad.cpp"
 $398 = (__ZN4gladL7has_extEPKc(468885)|0); //@line 6945 "libraries/glad/glad.cpp"
 $399 = $398&1; //@line 6945 "libraries/glad/glad.cpp"
 HEAP8[1252215] = $399; //@line 6945 "libraries/glad/glad.cpp"
 $400 = (__ZN4gladL7has_extEPKc(468916)|0); //@line 6946 "libraries/glad/glad.cpp"
 $401 = $400&1; //@line 6946 "libraries/glad/glad.cpp"
 HEAP8[1252216] = $401; //@line 6946 "libraries/glad/glad.cpp"
 $402 = (__ZN4gladL7has_extEPKc(468951)|0); //@line 6947 "libraries/glad/glad.cpp"
 $403 = $402&1; //@line 6947 "libraries/glad/glad.cpp"
 HEAP8[1252217] = $403; //@line 6947 "libraries/glad/glad.cpp"
 $404 = (__ZN4gladL7has_extEPKc(468977)|0); //@line 6948 "libraries/glad/glad.cpp"
 $405 = $404&1; //@line 6948 "libraries/glad/glad.cpp"
 HEAP8[1252218] = $405; //@line 6948 "libraries/glad/glad.cpp"
 $406 = (__ZN4gladL7has_extEPKc(469001)|0); //@line 6949 "libraries/glad/glad.cpp"
 $407 = $406&1; //@line 6949 "libraries/glad/glad.cpp"
 HEAP8[1252219] = $407; //@line 6949 "libraries/glad/glad.cpp"
 $408 = (__ZN4gladL7has_extEPKc(469035)|0); //@line 6950 "libraries/glad/glad.cpp"
 $409 = $408&1; //@line 6950 "libraries/glad/glad.cpp"
 HEAP8[1252220] = $409; //@line 6950 "libraries/glad/glad.cpp"
 $410 = (__ZN4gladL7has_extEPKc(469061)|0); //@line 6951 "libraries/glad/glad.cpp"
 $411 = $410&1; //@line 6951 "libraries/glad/glad.cpp"
 HEAP8[1252221] = $411; //@line 6951 "libraries/glad/glad.cpp"
 $412 = (__ZN4gladL7has_extEPKc(469084)|0); //@line 6952 "libraries/glad/glad.cpp"
 $413 = $412&1; //@line 6952 "libraries/glad/glad.cpp"
 HEAP8[1252222] = $413; //@line 6952 "libraries/glad/glad.cpp"
 $414 = (__ZN4gladL7has_extEPKc(469106)|0); //@line 6953 "libraries/glad/glad.cpp"
 $415 = $414&1; //@line 6953 "libraries/glad/glad.cpp"
 HEAP8[1252223] = $415; //@line 6953 "libraries/glad/glad.cpp"
 $416 = (__ZN4gladL7has_extEPKc(469129)|0); //@line 6954 "libraries/glad/glad.cpp"
 $417 = $416&1; //@line 6954 "libraries/glad/glad.cpp"
 HEAP8[1252224] = $417; //@line 6954 "libraries/glad/glad.cpp"
 $418 = (__ZN4gladL7has_extEPKc(469141)|0); //@line 6955 "libraries/glad/glad.cpp"
 $419 = $418&1; //@line 6955 "libraries/glad/glad.cpp"
 HEAP8[1252225] = $419; //@line 6955 "libraries/glad/glad.cpp"
 $420 = (__ZN4gladL7has_extEPKc(469167)|0); //@line 6956 "libraries/glad/glad.cpp"
 $421 = $420&1; //@line 6956 "libraries/glad/glad.cpp"
 HEAP8[1252226] = $421; //@line 6956 "libraries/glad/glad.cpp"
 $422 = (__ZN4gladL7has_extEPKc(469192)|0); //@line 6957 "libraries/glad/glad.cpp"
 $423 = $422&1; //@line 6957 "libraries/glad/glad.cpp"
 HEAP8[1252227] = $423; //@line 6957 "libraries/glad/glad.cpp"
 $424 = (__ZN4gladL7has_extEPKc(469223)|0); //@line 6958 "libraries/glad/glad.cpp"
 $425 = $424&1; //@line 6958 "libraries/glad/glad.cpp"
 HEAP8[1252228] = $425; //@line 6958 "libraries/glad/glad.cpp"
 $426 = (__ZN4gladL7has_extEPKc(469250)|0); //@line 6959 "libraries/glad/glad.cpp"
 $427 = $426&1; //@line 6959 "libraries/glad/glad.cpp"
 HEAP8[1252229] = $427; //@line 6959 "libraries/glad/glad.cpp"
 $428 = (__ZN4gladL7has_extEPKc(469271)|0); //@line 6960 "libraries/glad/glad.cpp"
 $429 = $428&1; //@line 6960 "libraries/glad/glad.cpp"
 HEAP8[1252230] = $429; //@line 6960 "libraries/glad/glad.cpp"
 $430 = (__ZN4gladL7has_extEPKc(469299)|0); //@line 6961 "libraries/glad/glad.cpp"
 $431 = $430&1; //@line 6961 "libraries/glad/glad.cpp"
 HEAP8[1252231] = $431; //@line 6961 "libraries/glad/glad.cpp"
 $432 = (__ZN4gladL7has_extEPKc(469321)|0); //@line 6962 "libraries/glad/glad.cpp"
 $433 = $432&1; //@line 6962 "libraries/glad/glad.cpp"
 HEAP8[1252232] = $433; //@line 6962 "libraries/glad/glad.cpp"
 $434 = (__ZN4gladL7has_extEPKc(469350)|0); //@line 6963 "libraries/glad/glad.cpp"
 $435 = $434&1; //@line 6963 "libraries/glad/glad.cpp"
 HEAP8[1252233] = $435; //@line 6963 "libraries/glad/glad.cpp"
 $436 = (__ZN4gladL7has_extEPKc(469382)|0); //@line 6964 "libraries/glad/glad.cpp"
 $437 = $436&1; //@line 6964 "libraries/glad/glad.cpp"
 HEAP8[1252234] = $437; //@line 6964 "libraries/glad/glad.cpp"
 $438 = (__ZN4gladL7has_extEPKc(469414)|0); //@line 6965 "libraries/glad/glad.cpp"
 $439 = $438&1; //@line 6965 "libraries/glad/glad.cpp"
 HEAP8[1252235] = $439; //@line 6965 "libraries/glad/glad.cpp"
 $440 = (__ZN4gladL7has_extEPKc(469446)|0); //@line 6966 "libraries/glad/glad.cpp"
 $441 = $440&1; //@line 6966 "libraries/glad/glad.cpp"
 HEAP8[1252236] = $441; //@line 6966 "libraries/glad/glad.cpp"
 $442 = (__ZN4gladL7has_extEPKc(469478)|0); //@line 6967 "libraries/glad/glad.cpp"
 $443 = $442&1; //@line 6967 "libraries/glad/glad.cpp"
 HEAP8[1252237] = $443; //@line 6967 "libraries/glad/glad.cpp"
 $444 = (__ZN4gladL7has_extEPKc(469502)|0); //@line 6968 "libraries/glad/glad.cpp"
 $445 = $444&1; //@line 6968 "libraries/glad/glad.cpp"
 HEAP8[1252238] = $445; //@line 6968 "libraries/glad/glad.cpp"
 $446 = (__ZN4gladL7has_extEPKc(469532)|0); //@line 6969 "libraries/glad/glad.cpp"
 $447 = $446&1; //@line 6969 "libraries/glad/glad.cpp"
 HEAP8[1252239] = $447; //@line 6969 "libraries/glad/glad.cpp"
 $448 = (__ZN4gladL7has_extEPKc(469566)|0); //@line 6970 "libraries/glad/glad.cpp"
 $449 = $448&1; //@line 6970 "libraries/glad/glad.cpp"
 HEAP8[1252240] = $449; //@line 6970 "libraries/glad/glad.cpp"
 $450 = (__ZN4gladL7has_extEPKc(469595)|0); //@line 6971 "libraries/glad/glad.cpp"
 $451 = $450&1; //@line 6971 "libraries/glad/glad.cpp"
 HEAP8[1252241] = $451; //@line 6971 "libraries/glad/glad.cpp"
 $452 = (__ZN4gladL7has_extEPKc(469626)|0); //@line 6972 "libraries/glad/glad.cpp"
 $453 = $452&1; //@line 6972 "libraries/glad/glad.cpp"
 HEAP8[1252242] = $453; //@line 6972 "libraries/glad/glad.cpp"
 $454 = (__ZN4gladL7has_extEPKc(469649)|0); //@line 6973 "libraries/glad/glad.cpp"
 $455 = $454&1; //@line 6973 "libraries/glad/glad.cpp"
 HEAP8[1252243] = $455; //@line 6973 "libraries/glad/glad.cpp"
 $456 = (__ZN4gladL7has_extEPKc(469671)|0); //@line 6974 "libraries/glad/glad.cpp"
 $457 = $456&1; //@line 6974 "libraries/glad/glad.cpp"
 HEAP8[1252244] = $457; //@line 6974 "libraries/glad/glad.cpp"
 $458 = (__ZN4gladL7has_extEPKc(469689)|0); //@line 6975 "libraries/glad/glad.cpp"
 $459 = $458&1; //@line 6975 "libraries/glad/glad.cpp"
 HEAP8[1252245] = $459; //@line 6975 "libraries/glad/glad.cpp"
 $460 = (__ZN4gladL7has_extEPKc(469720)|0); //@line 6976 "libraries/glad/glad.cpp"
 $461 = $460&1; //@line 6976 "libraries/glad/glad.cpp"
 HEAP8[1252246] = $461; //@line 6976 "libraries/glad/glad.cpp"
 $462 = (__ZN4gladL7has_extEPKc(469741)|0); //@line 6977 "libraries/glad/glad.cpp"
 $463 = $462&1; //@line 6977 "libraries/glad/glad.cpp"
 HEAP8[1252247] = $463; //@line 6977 "libraries/glad/glad.cpp"
 $464 = (__ZN4gladL7has_extEPKc(469761)|0); //@line 6978 "libraries/glad/glad.cpp"
 $465 = $464&1; //@line 6978 "libraries/glad/glad.cpp"
 HEAP8[1252248] = $465; //@line 6978 "libraries/glad/glad.cpp"
 $466 = (__ZN4gladL7has_extEPKc(469788)|0); //@line 6979 "libraries/glad/glad.cpp"
 $467 = $466&1; //@line 6979 "libraries/glad/glad.cpp"
 HEAP8[1252249] = $467; //@line 6979 "libraries/glad/glad.cpp"
 $468 = (__ZN4gladL7has_extEPKc(469811)|0); //@line 6980 "libraries/glad/glad.cpp"
 $469 = $468&1; //@line 6980 "libraries/glad/glad.cpp"
 HEAP8[1252250] = $469; //@line 6980 "libraries/glad/glad.cpp"
 $470 = (__ZN4gladL7has_extEPKc(469835)|0); //@line 6981 "libraries/glad/glad.cpp"
 $471 = $470&1; //@line 6981 "libraries/glad/glad.cpp"
 HEAP8[1252251] = $471; //@line 6981 "libraries/glad/glad.cpp"
 $472 = (__ZN4gladL7has_extEPKc(469858)|0); //@line 6982 "libraries/glad/glad.cpp"
 $473 = $472&1; //@line 6982 "libraries/glad/glad.cpp"
 HEAP8[1252252] = $473; //@line 6982 "libraries/glad/glad.cpp"
 $474 = (__ZN4gladL7has_extEPKc(469881)|0); //@line 6983 "libraries/glad/glad.cpp"
 $475 = $474&1; //@line 6983 "libraries/glad/glad.cpp"
 HEAP8[1252253] = $475; //@line 6983 "libraries/glad/glad.cpp"
 $476 = (__ZN4gladL7has_extEPKc(469916)|0); //@line 6984 "libraries/glad/glad.cpp"
 $477 = $476&1; //@line 6984 "libraries/glad/glad.cpp"
 HEAP8[1252254] = $477; //@line 6984 "libraries/glad/glad.cpp"
 $478 = (__ZN4gladL7has_extEPKc(469936)|0); //@line 6985 "libraries/glad/glad.cpp"
 $479 = $478&1; //@line 6985 "libraries/glad/glad.cpp"
 HEAP8[1252255] = $479; //@line 6985 "libraries/glad/glad.cpp"
 $480 = (__ZN4gladL7has_extEPKc(469955)|0); //@line 6986 "libraries/glad/glad.cpp"
 $481 = $480&1; //@line 6986 "libraries/glad/glad.cpp"
 HEAP8[1252256] = $481; //@line 6986 "libraries/glad/glad.cpp"
 $482 = (__ZN4gladL7has_extEPKc(469981)|0); //@line 6987 "libraries/glad/glad.cpp"
 $483 = $482&1; //@line 6987 "libraries/glad/glad.cpp"
 HEAP8[1252257] = $483; //@line 6987 "libraries/glad/glad.cpp"
 $484 = (__ZN4gladL7has_extEPKc(470004)|0); //@line 6988 "libraries/glad/glad.cpp"
 $485 = $484&1; //@line 6988 "libraries/glad/glad.cpp"
 HEAP8[1252258] = $485; //@line 6988 "libraries/glad/glad.cpp"
 $486 = (__ZN4gladL7has_extEPKc(470029)|0); //@line 6989 "libraries/glad/glad.cpp"
 $487 = $486&1; //@line 6989 "libraries/glad/glad.cpp"
 HEAP8[1252259] = $487; //@line 6989 "libraries/glad/glad.cpp"
 $488 = (__ZN4gladL7has_extEPKc(470056)|0); //@line 6990 "libraries/glad/glad.cpp"
 $489 = $488&1; //@line 6990 "libraries/glad/glad.cpp"
 HEAP8[1252260] = $489; //@line 6990 "libraries/glad/glad.cpp"
 $490 = (__ZN4gladL7has_extEPKc(470079)|0); //@line 6991 "libraries/glad/glad.cpp"
 $491 = $490&1; //@line 6991 "libraries/glad/glad.cpp"
 HEAP8[1252261] = $491; //@line 6991 "libraries/glad/glad.cpp"
 $492 = (__ZN4gladL7has_extEPKc(470097)|0); //@line 6992 "libraries/glad/glad.cpp"
 $493 = $492&1; //@line 6992 "libraries/glad/glad.cpp"
 HEAP8[1252262] = $493; //@line 6992 "libraries/glad/glad.cpp"
 $494 = (__ZN4gladL7has_extEPKc(470137)|0); //@line 6993 "libraries/glad/glad.cpp"
 $495 = $494&1; //@line 6993 "libraries/glad/glad.cpp"
 HEAP8[1252263] = $495; //@line 6993 "libraries/glad/glad.cpp"
 $496 = (__ZN4gladL7has_extEPKc(470173)|0); //@line 6994 "libraries/glad/glad.cpp"
 $497 = $496&1; //@line 6994 "libraries/glad/glad.cpp"
 HEAP8[1252264] = $497; //@line 6994 "libraries/glad/glad.cpp"
 $498 = (__ZN4gladL7has_extEPKc(470191)|0); //@line 6995 "libraries/glad/glad.cpp"
 $499 = $498&1; //@line 6995 "libraries/glad/glad.cpp"
 HEAP8[1252265] = $499; //@line 6995 "libraries/glad/glad.cpp"
 $500 = (__ZN4gladL7has_extEPKc(470206)|0); //@line 6996 "libraries/glad/glad.cpp"
 $501 = $500&1; //@line 6996 "libraries/glad/glad.cpp"
 HEAP8[1252266] = $501; //@line 6996 "libraries/glad/glad.cpp"
 $502 = (__ZN4gladL7has_extEPKc(470221)|0); //@line 6997 "libraries/glad/glad.cpp"
 $503 = $502&1; //@line 6997 "libraries/glad/glad.cpp"
 HEAP8[1252267] = $503; //@line 6997 "libraries/glad/glad.cpp"
 $504 = (__ZN4gladL7has_extEPKc(470242)|0); //@line 6998 "libraries/glad/glad.cpp"
 $505 = $504&1; //@line 6998 "libraries/glad/glad.cpp"
 HEAP8[1252268] = $505; //@line 6998 "libraries/glad/glad.cpp"
 $506 = (__ZN4gladL7has_extEPKc(470270)|0); //@line 6999 "libraries/glad/glad.cpp"
 $507 = $506&1; //@line 6999 "libraries/glad/glad.cpp"
 HEAP8[1252269] = $507; //@line 6999 "libraries/glad/glad.cpp"
 $508 = (__ZN4gladL7has_extEPKc(470303)|0); //@line 7000 "libraries/glad/glad.cpp"
 $509 = $508&1; //@line 7000 "libraries/glad/glad.cpp"
 HEAP8[1252270] = $509; //@line 7000 "libraries/glad/glad.cpp"
 $510 = (__ZN4gladL7has_extEPKc(470320)|0); //@line 7001 "libraries/glad/glad.cpp"
 $511 = $510&1; //@line 7001 "libraries/glad/glad.cpp"
 HEAP8[1252271] = $511; //@line 7001 "libraries/glad/glad.cpp"
 $512 = (__ZN4gladL7has_extEPKc(470346)|0); //@line 7002 "libraries/glad/glad.cpp"
 $513 = $512&1; //@line 7002 "libraries/glad/glad.cpp"
 HEAP8[1252272] = $513; //@line 7002 "libraries/glad/glad.cpp"
 $514 = (__ZN4gladL7has_extEPKc(470378)|0); //@line 7003 "libraries/glad/glad.cpp"
 $515 = $514&1; //@line 7003 "libraries/glad/glad.cpp"
 HEAP8[1252273] = $515; //@line 7003 "libraries/glad/glad.cpp"
 $516 = (__ZN4gladL7has_extEPKc(470404)|0); //@line 7004 "libraries/glad/glad.cpp"
 $517 = $516&1; //@line 7004 "libraries/glad/glad.cpp"
 HEAP8[1252274] = $517; //@line 7004 "libraries/glad/glad.cpp"
 $518 = (__ZN4gladL7has_extEPKc(470429)|0); //@line 7005 "libraries/glad/glad.cpp"
 $519 = $518&1; //@line 7005 "libraries/glad/glad.cpp"
 HEAP8[1252275] = $519; //@line 7005 "libraries/glad/glad.cpp"
 $520 = (__ZN4gladL7has_extEPKc(470460)|0); //@line 7006 "libraries/glad/glad.cpp"
 $521 = $520&1; //@line 7006 "libraries/glad/glad.cpp"
 HEAP8[1252276] = $521; //@line 7006 "libraries/glad/glad.cpp"
 $522 = (__ZN4gladL7has_extEPKc(470487)|0); //@line 7007 "libraries/glad/glad.cpp"
 $523 = $522&1; //@line 7007 "libraries/glad/glad.cpp"
 HEAP8[1252277] = $523; //@line 7007 "libraries/glad/glad.cpp"
 $524 = (__ZN4gladL7has_extEPKc(470510)|0); //@line 7008 "libraries/glad/glad.cpp"
 $525 = $524&1; //@line 7008 "libraries/glad/glad.cpp"
 HEAP8[1252278] = $525; //@line 7008 "libraries/glad/glad.cpp"
 $526 = (__ZN4gladL7has_extEPKc(470536)|0); //@line 7009 "libraries/glad/glad.cpp"
 $527 = $526&1; //@line 7009 "libraries/glad/glad.cpp"
 HEAP8[1252279] = $527; //@line 7009 "libraries/glad/glad.cpp"
 $528 = (__ZN4gladL7has_extEPKc(470555)|0); //@line 7010 "libraries/glad/glad.cpp"
 $529 = $528&1; //@line 7010 "libraries/glad/glad.cpp"
 HEAP8[1252280] = $529; //@line 7010 "libraries/glad/glad.cpp"
 $530 = (__ZN4gladL7has_extEPKc(470572)|0); //@line 7011 "libraries/glad/glad.cpp"
 $531 = $530&1; //@line 7011 "libraries/glad/glad.cpp"
 HEAP8[1252281] = $531; //@line 7011 "libraries/glad/glad.cpp"
 $532 = (__ZN4gladL7has_extEPKc(470600)|0); //@line 7012 "libraries/glad/glad.cpp"
 $533 = $532&1; //@line 7012 "libraries/glad/glad.cpp"
 HEAP8[1252282] = $533; //@line 7012 "libraries/glad/glad.cpp"
 $534 = (__ZN4gladL7has_extEPKc(470630)|0); //@line 7013 "libraries/glad/glad.cpp"
 $535 = $534&1; //@line 7013 "libraries/glad/glad.cpp"
 HEAP8[1252283] = $535; //@line 7013 "libraries/glad/glad.cpp"
 $536 = (__ZN4gladL7has_extEPKc(470661)|0); //@line 7014 "libraries/glad/glad.cpp"
 $537 = $536&1; //@line 7014 "libraries/glad/glad.cpp"
 HEAP8[1252284] = $537; //@line 7014 "libraries/glad/glad.cpp"
 $538 = (__ZN4gladL7has_extEPKc(470679)|0); //@line 7015 "libraries/glad/glad.cpp"
 $539 = $538&1; //@line 7015 "libraries/glad/glad.cpp"
 HEAP8[1252285] = $539; //@line 7015 "libraries/glad/glad.cpp"
 $540 = (__ZN4gladL7has_extEPKc(470701)|0); //@line 7016 "libraries/glad/glad.cpp"
 $541 = $540&1; //@line 7016 "libraries/glad/glad.cpp"
 HEAP8[1252286] = $541; //@line 7016 "libraries/glad/glad.cpp"
 $542 = (__ZN4gladL7has_extEPKc(470725)|0); //@line 7017 "libraries/glad/glad.cpp"
 $543 = $542&1; //@line 7017 "libraries/glad/glad.cpp"
 HEAP8[1252287] = $543; //@line 7017 "libraries/glad/glad.cpp"
 $544 = (__ZN4gladL7has_extEPKc(470752)|0); //@line 7018 "libraries/glad/glad.cpp"
 $545 = $544&1; //@line 7018 "libraries/glad/glad.cpp"
 HEAP8[1252288] = $545; //@line 7018 "libraries/glad/glad.cpp"
 $546 = (__ZN4gladL7has_extEPKc(470776)|0); //@line 7019 "libraries/glad/glad.cpp"
 $547 = $546&1; //@line 7019 "libraries/glad/glad.cpp"
 HEAP8[1252289] = $547; //@line 7019 "libraries/glad/glad.cpp"
 $548 = (__ZN4gladL7has_extEPKc(470816)|0); //@line 7020 "libraries/glad/glad.cpp"
 $549 = $548&1; //@line 7020 "libraries/glad/glad.cpp"
 HEAP8[1252290] = $549; //@line 7020 "libraries/glad/glad.cpp"
 $550 = (__ZN4gladL7has_extEPKc(470844)|0); //@line 7021 "libraries/glad/glad.cpp"
 $551 = $550&1; //@line 7021 "libraries/glad/glad.cpp"
 HEAP8[1252291] = $551; //@line 7021 "libraries/glad/glad.cpp"
 $552 = (__ZN4gladL7has_extEPKc(470860)|0); //@line 7022 "libraries/glad/glad.cpp"
 $553 = $552&1; //@line 7022 "libraries/glad/glad.cpp"
 HEAP8[1252292] = $553; //@line 7022 "libraries/glad/glad.cpp"
 $554 = (__ZN4gladL7has_extEPKc(470876)|0); //@line 7023 "libraries/glad/glad.cpp"
 $555 = $554&1; //@line 7023 "libraries/glad/glad.cpp"
 HEAP8[1252293] = $555; //@line 7023 "libraries/glad/glad.cpp"
 $556 = (__ZN4gladL7has_extEPKc(470903)|0); //@line 7024 "libraries/glad/glad.cpp"
 $557 = $556&1; //@line 7024 "libraries/glad/glad.cpp"
 HEAP8[1252294] = $557; //@line 7024 "libraries/glad/glad.cpp"
 $558 = (__ZN4gladL7has_extEPKc(470934)|0); //@line 7025 "libraries/glad/glad.cpp"
 $559 = $558&1; //@line 7025 "libraries/glad/glad.cpp"
 HEAP8[1252295] = $559; //@line 7025 "libraries/glad/glad.cpp"
 $560 = (__ZN4gladL7has_extEPKc(470961)|0); //@line 7026 "libraries/glad/glad.cpp"
 $561 = $560&1; //@line 7026 "libraries/glad/glad.cpp"
 HEAP8[1252296] = $561; //@line 7026 "libraries/glad/glad.cpp"
 $562 = (__ZN4gladL7has_extEPKc(470979)|0); //@line 7027 "libraries/glad/glad.cpp"
 $563 = $562&1; //@line 7027 "libraries/glad/glad.cpp"
 HEAP8[1252297] = $563; //@line 7027 "libraries/glad/glad.cpp"
 $564 = (__ZN4gladL7has_extEPKc(471007)|0); //@line 7028 "libraries/glad/glad.cpp"
 $565 = $564&1; //@line 7028 "libraries/glad/glad.cpp"
 HEAP8[1252298] = $565; //@line 7028 "libraries/glad/glad.cpp"
 $566 = (__ZN4gladL7has_extEPKc(471029)|0); //@line 7029 "libraries/glad/glad.cpp"
 $567 = $566&1; //@line 7029 "libraries/glad/glad.cpp"
 HEAP8[1252299] = $567; //@line 7029 "libraries/glad/glad.cpp"
 $568 = (__ZN4gladL7has_extEPKc(471061)|0); //@line 7030 "libraries/glad/glad.cpp"
 $569 = $568&1; //@line 7030 "libraries/glad/glad.cpp"
 HEAP8[1252300] = $569; //@line 7030 "libraries/glad/glad.cpp"
 $570 = (__ZN4gladL7has_extEPKc(471091)|0); //@line 7031 "libraries/glad/glad.cpp"
 $571 = $570&1; //@line 7031 "libraries/glad/glad.cpp"
 HEAP8[1252301] = $571; //@line 7031 "libraries/glad/glad.cpp"
 $572 = (__ZN4gladL7has_extEPKc(471112)|0); //@line 7032 "libraries/glad/glad.cpp"
 $573 = $572&1; //@line 7032 "libraries/glad/glad.cpp"
 HEAP8[1252302] = $573; //@line 7032 "libraries/glad/glad.cpp"
 $574 = (__ZN4gladL7has_extEPKc(471140)|0); //@line 7033 "libraries/glad/glad.cpp"
 $575 = $574&1; //@line 7033 "libraries/glad/glad.cpp"
 HEAP8[1252303] = $575; //@line 7033 "libraries/glad/glad.cpp"
 $576 = (__ZN4gladL7has_extEPKc(471166)|0); //@line 7034 "libraries/glad/glad.cpp"
 $577 = $576&1; //@line 7034 "libraries/glad/glad.cpp"
 HEAP8[1252304] = $577; //@line 7034 "libraries/glad/glad.cpp"
 $578 = (__ZN4gladL7has_extEPKc(471199)|0); //@line 7035 "libraries/glad/glad.cpp"
 $579 = $578&1; //@line 7035 "libraries/glad/glad.cpp"
 HEAP8[1252305] = $579; //@line 7035 "libraries/glad/glad.cpp"
 $580 = (__ZN4gladL7has_extEPKc(471219)|0); //@line 7036 "libraries/glad/glad.cpp"
 $581 = $580&1; //@line 7036 "libraries/glad/glad.cpp"
 HEAP8[1252306] = $581; //@line 7036 "libraries/glad/glad.cpp"
 $582 = (__ZN4gladL7has_extEPKc(471243)|0); //@line 7037 "libraries/glad/glad.cpp"
 $583 = $582&1; //@line 7037 "libraries/glad/glad.cpp"
 HEAP8[1252307] = $583; //@line 7037 "libraries/glad/glad.cpp"
 $584 = (__ZN4gladL7has_extEPKc(471287)|0); //@line 7038 "libraries/glad/glad.cpp"
 $585 = $584&1; //@line 7038 "libraries/glad/glad.cpp"
 HEAP8[1252308] = $585; //@line 7038 "libraries/glad/glad.cpp"
 $586 = (__ZN4gladL7has_extEPKc(471307)|0); //@line 7039 "libraries/glad/glad.cpp"
 $587 = $586&1; //@line 7039 "libraries/glad/glad.cpp"
 HEAP8[1252309] = $587; //@line 7039 "libraries/glad/glad.cpp"
 $588 = (__ZN4gladL7has_extEPKc(471334)|0); //@line 7040 "libraries/glad/glad.cpp"
 $589 = $588&1; //@line 7040 "libraries/glad/glad.cpp"
 HEAP8[1252310] = $589; //@line 7040 "libraries/glad/glad.cpp"
 $590 = (__ZN4gladL7has_extEPKc(471359)|0); //@line 7041 "libraries/glad/glad.cpp"
 $591 = $590&1; //@line 7041 "libraries/glad/glad.cpp"
 HEAP8[1252311] = $591; //@line 7041 "libraries/glad/glad.cpp"
 $592 = (__ZN4gladL7has_extEPKc(471389)|0); //@line 7042 "libraries/glad/glad.cpp"
 $593 = $592&1; //@line 7042 "libraries/glad/glad.cpp"
 HEAP8[1252312] = $593; //@line 7042 "libraries/glad/glad.cpp"
 $594 = (__ZN4gladL7has_extEPKc(471416)|0); //@line 7043 "libraries/glad/glad.cpp"
 $595 = $594&1; //@line 7043 "libraries/glad/glad.cpp"
 HEAP8[1252313] = $595; //@line 7043 "libraries/glad/glad.cpp"
 $596 = (__ZN4gladL7has_extEPKc(471446)|0); //@line 7044 "libraries/glad/glad.cpp"
 $597 = $596&1; //@line 7044 "libraries/glad/glad.cpp"
 HEAP8[1252314] = $597; //@line 7044 "libraries/glad/glad.cpp"
 $598 = (__ZN4gladL7has_extEPKc(471476)|0); //@line 7045 "libraries/glad/glad.cpp"
 $599 = $598&1; //@line 7045 "libraries/glad/glad.cpp"
 HEAP8[1252315] = $599; //@line 7045 "libraries/glad/glad.cpp"
 $600 = (__ZN4gladL7has_extEPKc(471502)|0); //@line 7046 "libraries/glad/glad.cpp"
 $601 = $600&1; //@line 7046 "libraries/glad/glad.cpp"
 HEAP8[1252316] = $601; //@line 7046 "libraries/glad/glad.cpp"
 $602 = (__ZN4gladL7has_extEPKc(471522)|0); //@line 7047 "libraries/glad/glad.cpp"
 $603 = $602&1; //@line 7047 "libraries/glad/glad.cpp"
 HEAP8[1252317] = $603; //@line 7047 "libraries/glad/glad.cpp"
 $604 = (__ZN4gladL7has_extEPKc(471550)|0); //@line 7048 "libraries/glad/glad.cpp"
 $605 = $604&1; //@line 7048 "libraries/glad/glad.cpp"
 HEAP8[1252318] = $605; //@line 7048 "libraries/glad/glad.cpp"
 $606 = (__ZN4gladL7has_extEPKc(471576)|0); //@line 7049 "libraries/glad/glad.cpp"
 $607 = $606&1; //@line 7049 "libraries/glad/glad.cpp"
 HEAP8[1252319] = $607; //@line 7049 "libraries/glad/glad.cpp"
 $608 = (__ZN4gladL7has_extEPKc(471594)|0); //@line 7050 "libraries/glad/glad.cpp"
 $609 = $608&1; //@line 7050 "libraries/glad/glad.cpp"
 HEAP8[1252320] = $609; //@line 7050 "libraries/glad/glad.cpp"
 $610 = (__ZN4gladL7has_extEPKc(471618)|0); //@line 7051 "libraries/glad/glad.cpp"
 $611 = $610&1; //@line 7051 "libraries/glad/glad.cpp"
 HEAP8[1252321] = $611; //@line 7051 "libraries/glad/glad.cpp"
 $612 = (__ZN4gladL7has_extEPKc(471646)|0); //@line 7052 "libraries/glad/glad.cpp"
 $613 = $612&1; //@line 7052 "libraries/glad/glad.cpp"
 HEAP8[1252322] = $613; //@line 7052 "libraries/glad/glad.cpp"
 $614 = (__ZN4gladL7has_extEPKc(471673)|0); //@line 7053 "libraries/glad/glad.cpp"
 $615 = $614&1; //@line 7053 "libraries/glad/glad.cpp"
 HEAP8[1252323] = $615; //@line 7053 "libraries/glad/glad.cpp"
 $616 = (__ZN4gladL7has_extEPKc(471696)|0); //@line 7054 "libraries/glad/glad.cpp"
 $617 = $616&1; //@line 7054 "libraries/glad/glad.cpp"
 HEAP8[1252324] = $617; //@line 7054 "libraries/glad/glad.cpp"
 $618 = (__ZN4gladL7has_extEPKc(471725)|0); //@line 7055 "libraries/glad/glad.cpp"
 $619 = $618&1; //@line 7055 "libraries/glad/glad.cpp"
 HEAP8[1252325] = $619; //@line 7055 "libraries/glad/glad.cpp"
 $620 = (__ZN4gladL7has_extEPKc(471752)|0); //@line 7056 "libraries/glad/glad.cpp"
 $621 = $620&1; //@line 7056 "libraries/glad/glad.cpp"
 HEAP8[1252326] = $621; //@line 7056 "libraries/glad/glad.cpp"
 $622 = (__ZN4gladL7has_extEPKc(471773)|0); //@line 7057 "libraries/glad/glad.cpp"
 $623 = $622&1; //@line 7057 "libraries/glad/glad.cpp"
 HEAP8[1252327] = $623; //@line 7057 "libraries/glad/glad.cpp"
 $624 = (__ZN4gladL7has_extEPKc(471800)|0); //@line 7058 "libraries/glad/glad.cpp"
 $625 = $624&1; //@line 7058 "libraries/glad/glad.cpp"
 HEAP8[1252328] = $625; //@line 7058 "libraries/glad/glad.cpp"
 $626 = (__ZN4gladL7has_extEPKc(471827)|0); //@line 7059 "libraries/glad/glad.cpp"
 $627 = $626&1; //@line 7059 "libraries/glad/glad.cpp"
 HEAP8[1252329] = $627; //@line 7059 "libraries/glad/glad.cpp"
 $628 = (__ZN4gladL7has_extEPKc(471851)|0); //@line 7060 "libraries/glad/glad.cpp"
 $629 = $628&1; //@line 7060 "libraries/glad/glad.cpp"
 HEAP8[1252330] = $629; //@line 7060 "libraries/glad/glad.cpp"
 $630 = (__ZN4gladL7has_extEPKc(471887)|0); //@line 7061 "libraries/glad/glad.cpp"
 $631 = $630&1; //@line 7061 "libraries/glad/glad.cpp"
 HEAP8[1252331] = $631; //@line 7061 "libraries/glad/glad.cpp"
 $632 = (__ZN4gladL7has_extEPKc(471920)|0); //@line 7062 "libraries/glad/glad.cpp"
 $633 = $632&1; //@line 7062 "libraries/glad/glad.cpp"
 HEAP8[1252332] = $633; //@line 7062 "libraries/glad/glad.cpp"
 $634 = (__ZN4gladL7has_extEPKc(471949)|0); //@line 7063 "libraries/glad/glad.cpp"
 $635 = $634&1; //@line 7063 "libraries/glad/glad.cpp"
 HEAP8[1252333] = $635; //@line 7063 "libraries/glad/glad.cpp"
 $636 = (__ZN4gladL7has_extEPKc(471978)|0); //@line 7064 "libraries/glad/glad.cpp"
 $637 = $636&1; //@line 7064 "libraries/glad/glad.cpp"
 HEAP8[1252334] = $637; //@line 7064 "libraries/glad/glad.cpp"
 $638 = (__ZN4gladL7has_extEPKc(472000)|0); //@line 7065 "libraries/glad/glad.cpp"
 $639 = $638&1; //@line 7065 "libraries/glad/glad.cpp"
 HEAP8[1252335] = $639; //@line 7065 "libraries/glad/glad.cpp"
 $640 = (__ZN4gladL7has_extEPKc(472034)|0); //@line 7066 "libraries/glad/glad.cpp"
 $641 = $640&1; //@line 7066 "libraries/glad/glad.cpp"
 HEAP8[1252336] = $641; //@line 7066 "libraries/glad/glad.cpp"
 $642 = (__ZN4gladL7has_extEPKc(472058)|0); //@line 7067 "libraries/glad/glad.cpp"
 $643 = $642&1; //@line 7067 "libraries/glad/glad.cpp"
 HEAP8[1252337] = $643; //@line 7067 "libraries/glad/glad.cpp"
 $644 = (__ZN4gladL7has_extEPKc(472101)|0); //@line 7068 "libraries/glad/glad.cpp"
 $645 = $644&1; //@line 7068 "libraries/glad/glad.cpp"
 HEAP8[1252338] = $645; //@line 7068 "libraries/glad/glad.cpp"
 $646 = (__ZN4gladL7has_extEPKc(472128)|0); //@line 7069 "libraries/glad/glad.cpp"
 $647 = $646&1; //@line 7069 "libraries/glad/glad.cpp"
 HEAP8[1252339] = $647; //@line 7069 "libraries/glad/glad.cpp"
 $648 = (__ZN4gladL7has_extEPKc(472155)|0); //@line 7070 "libraries/glad/glad.cpp"
 $649 = $648&1; //@line 7070 "libraries/glad/glad.cpp"
 HEAP8[1252340] = $649; //@line 7070 "libraries/glad/glad.cpp"
 $650 = (__ZN4gladL7has_extEPKc(472188)|0); //@line 7071 "libraries/glad/glad.cpp"
 $651 = $650&1; //@line 7071 "libraries/glad/glad.cpp"
 HEAP8[1252341] = $651; //@line 7071 "libraries/glad/glad.cpp"
 $652 = (__ZN4gladL7has_extEPKc(472224)|0); //@line 7072 "libraries/glad/glad.cpp"
 $653 = $652&1; //@line 7072 "libraries/glad/glad.cpp"
 HEAP8[1252342] = $653; //@line 7072 "libraries/glad/glad.cpp"
 $654 = (__ZN4gladL7has_extEPKc(472256)|0); //@line 7073 "libraries/glad/glad.cpp"
 $655 = $654&1; //@line 7073 "libraries/glad/glad.cpp"
 HEAP8[1252343] = $655; //@line 7073 "libraries/glad/glad.cpp"
 $656 = (__ZN4gladL7has_extEPKc(472279)|0); //@line 7074 "libraries/glad/glad.cpp"
 $657 = $656&1; //@line 7074 "libraries/glad/glad.cpp"
 HEAP8[1252344] = $657; //@line 7074 "libraries/glad/glad.cpp"
 $658 = (__ZN4gladL7has_extEPKc(472305)|0); //@line 7075 "libraries/glad/glad.cpp"
 $659 = $658&1; //@line 7075 "libraries/glad/glad.cpp"
 HEAP8[1252345] = $659; //@line 7075 "libraries/glad/glad.cpp"
 $660 = (__ZN4gladL7has_extEPKc(472338)|0); //@line 7076 "libraries/glad/glad.cpp"
 $661 = $660&1; //@line 7076 "libraries/glad/glad.cpp"
 HEAP8[1252346] = $661; //@line 7076 "libraries/glad/glad.cpp"
 $662 = (__ZN4gladL7has_extEPKc(472364)|0); //@line 7077 "libraries/glad/glad.cpp"
 $663 = $662&1; //@line 7077 "libraries/glad/glad.cpp"
 HEAP8[1252347] = $663; //@line 7077 "libraries/glad/glad.cpp"
 $664 = (__ZN4gladL7has_extEPKc(472396)|0); //@line 7078 "libraries/glad/glad.cpp"
 $665 = $664&1; //@line 7078 "libraries/glad/glad.cpp"
 HEAP8[1252348] = $665; //@line 7078 "libraries/glad/glad.cpp"
 $666 = (__ZN4gladL7has_extEPKc(472420)|0); //@line 7079 "libraries/glad/glad.cpp"
 $667 = $666&1; //@line 7079 "libraries/glad/glad.cpp"
 HEAP8[1252349] = $667; //@line 7079 "libraries/glad/glad.cpp"
 $668 = (__ZN4gladL7has_extEPKc(472454)|0); //@line 7080 "libraries/glad/glad.cpp"
 $669 = $668&1; //@line 7080 "libraries/glad/glad.cpp"
 HEAP8[1252350] = $669; //@line 7080 "libraries/glad/glad.cpp"
 $670 = (__ZN4gladL7has_extEPKc(472488)|0); //@line 7081 "libraries/glad/glad.cpp"
 $671 = $670&1; //@line 7081 "libraries/glad/glad.cpp"
 HEAP8[1252351] = $671; //@line 7081 "libraries/glad/glad.cpp"
 $672 = (__ZN4gladL7has_extEPKc(472511)|0); //@line 7082 "libraries/glad/glad.cpp"
 $673 = $672&1; //@line 7082 "libraries/glad/glad.cpp"
 HEAP8[1252352] = $673; //@line 7082 "libraries/glad/glad.cpp"
 $674 = (__ZN4gladL7has_extEPKc(472545)|0); //@line 7083 "libraries/glad/glad.cpp"
 $675 = $674&1; //@line 7083 "libraries/glad/glad.cpp"
 HEAP8[1252353] = $675; //@line 7083 "libraries/glad/glad.cpp"
 $676 = (__ZN4gladL7has_extEPKc(472572)|0); //@line 7084 "libraries/glad/glad.cpp"
 $677 = $676&1; //@line 7084 "libraries/glad/glad.cpp"
 HEAP8[1252354] = $677; //@line 7084 "libraries/glad/glad.cpp"
 $678 = (__ZN4gladL7has_extEPKc(472596)|0); //@line 7085 "libraries/glad/glad.cpp"
 $679 = $678&1; //@line 7085 "libraries/glad/glad.cpp"
 HEAP8[1252355] = $679; //@line 7085 "libraries/glad/glad.cpp"
 $680 = (__ZN4gladL7has_extEPKc(472619)|0); //@line 7086 "libraries/glad/glad.cpp"
 $681 = $680&1; //@line 7086 "libraries/glad/glad.cpp"
 HEAP8[1252356] = $681; //@line 7086 "libraries/glad/glad.cpp"
 $682 = (__ZN4gladL7has_extEPKc(472654)|0); //@line 7087 "libraries/glad/glad.cpp"
 $683 = $682&1; //@line 7087 "libraries/glad/glad.cpp"
 HEAP8[1252357] = $683; //@line 7087 "libraries/glad/glad.cpp"
 $684 = (__ZN4gladL7has_extEPKc(472683)|0); //@line 7088 "libraries/glad/glad.cpp"
 $685 = $684&1; //@line 7088 "libraries/glad/glad.cpp"
 HEAP8[1252358] = $685; //@line 7088 "libraries/glad/glad.cpp"
 $686 = (__ZN4gladL7has_extEPKc(472706)|0); //@line 7089 "libraries/glad/glad.cpp"
 $687 = $686&1; //@line 7089 "libraries/glad/glad.cpp"
 HEAP8[1252359] = $687; //@line 7089 "libraries/glad/glad.cpp"
 $688 = (__ZN4gladL7has_extEPKc(472721)|0); //@line 7090 "libraries/glad/glad.cpp"
 $689 = $688&1; //@line 7090 "libraries/glad/glad.cpp"
 HEAP8[1252360] = $689; //@line 7090 "libraries/glad/glad.cpp"
 $690 = (__ZN4gladL7has_extEPKc(472743)|0); //@line 7091 "libraries/glad/glad.cpp"
 $691 = $690&1; //@line 7091 "libraries/glad/glad.cpp"
 HEAP8[1252361] = $691; //@line 7091 "libraries/glad/glad.cpp"
 $692 = (__ZN4gladL7has_extEPKc(472771)|0); //@line 7092 "libraries/glad/glad.cpp"
 $693 = $692&1; //@line 7092 "libraries/glad/glad.cpp"
 HEAP8[1252362] = $693; //@line 7092 "libraries/glad/glad.cpp"
 $694 = (__ZN4gladL7has_extEPKc(472804)|0); //@line 7093 "libraries/glad/glad.cpp"
 $695 = $694&1; //@line 7093 "libraries/glad/glad.cpp"
 HEAP8[1252363] = $695; //@line 7093 "libraries/glad/glad.cpp"
 $696 = (__ZN4gladL7has_extEPKc(472830)|0); //@line 7094 "libraries/glad/glad.cpp"
 $697 = $696&1; //@line 7094 "libraries/glad/glad.cpp"
 HEAP8[1252364] = $697; //@line 7094 "libraries/glad/glad.cpp"
 $698 = (__ZN4gladL7has_extEPKc(472847)|0); //@line 7095 "libraries/glad/glad.cpp"
 $699 = $698&1; //@line 7095 "libraries/glad/glad.cpp"
 HEAP8[1252365] = $699; //@line 7095 "libraries/glad/glad.cpp"
 $700 = (__ZN4gladL7has_extEPKc(472866)|0); //@line 7096 "libraries/glad/glad.cpp"
 $701 = $700&1; //@line 7096 "libraries/glad/glad.cpp"
 HEAP8[1252366] = $701; //@line 7096 "libraries/glad/glad.cpp"
 $702 = (__ZN4gladL7has_extEPKc(472880)|0); //@line 7097 "libraries/glad/glad.cpp"
 $703 = $702&1; //@line 7097 "libraries/glad/glad.cpp"
 HEAP8[1252367] = $703; //@line 7097 "libraries/glad/glad.cpp"
 $704 = (__ZN4gladL7has_extEPKc(472913)|0); //@line 7098 "libraries/glad/glad.cpp"
 $705 = $704&1; //@line 7098 "libraries/glad/glad.cpp"
 HEAP8[1252368] = $705; //@line 7098 "libraries/glad/glad.cpp"
 $706 = (__ZN4gladL7has_extEPKc(472940)|0); //@line 7099 "libraries/glad/glad.cpp"
 $707 = $706&1; //@line 7099 "libraries/glad/glad.cpp"
 HEAP8[1252369] = $707; //@line 7099 "libraries/glad/glad.cpp"
 $708 = (__ZN4gladL7has_extEPKc(472970)|0); //@line 7100 "libraries/glad/glad.cpp"
 $709 = $708&1; //@line 7100 "libraries/glad/glad.cpp"
 HEAP8[1252370] = $709; //@line 7100 "libraries/glad/glad.cpp"
 $710 = (__ZN4gladL7has_extEPKc(472993)|0); //@line 7101 "libraries/glad/glad.cpp"
 $711 = $710&1; //@line 7101 "libraries/glad/glad.cpp"
 HEAP8[1252371] = $711; //@line 7101 "libraries/glad/glad.cpp"
 $712 = (__ZN4gladL7has_extEPKc(473022)|0); //@line 7102 "libraries/glad/glad.cpp"
 $713 = $712&1; //@line 7102 "libraries/glad/glad.cpp"
 HEAP8[1252372] = $713; //@line 7102 "libraries/glad/glad.cpp"
 $714 = (__ZN4gladL7has_extEPKc(473050)|0); //@line 7103 "libraries/glad/glad.cpp"
 $715 = $714&1; //@line 7103 "libraries/glad/glad.cpp"
 HEAP8[1252373] = $715; //@line 7103 "libraries/glad/glad.cpp"
 $716 = (__ZN4gladL7has_extEPKc(473085)|0); //@line 7104 "libraries/glad/glad.cpp"
 $717 = $716&1; //@line 7104 "libraries/glad/glad.cpp"
 HEAP8[1252374] = $717; //@line 7104 "libraries/glad/glad.cpp"
 $718 = (__ZN4gladL7has_extEPKc(473104)|0); //@line 7105 "libraries/glad/glad.cpp"
 $719 = $718&1; //@line 7105 "libraries/glad/glad.cpp"
 HEAP8[1252375] = $719; //@line 7105 "libraries/glad/glad.cpp"
 $720 = (__ZN4gladL7has_extEPKc(473131)|0); //@line 7106 "libraries/glad/glad.cpp"
 $721 = $720&1; //@line 7106 "libraries/glad/glad.cpp"
 HEAP8[1252376] = $721; //@line 7106 "libraries/glad/glad.cpp"
 $722 = (__ZN4gladL7has_extEPKc(473157)|0); //@line 7107 "libraries/glad/glad.cpp"
 $723 = $722&1; //@line 7107 "libraries/glad/glad.cpp"
 HEAP8[1252377] = $723; //@line 7107 "libraries/glad/glad.cpp"
 $724 = (__ZN4gladL7has_extEPKc(473170)|0); //@line 7108 "libraries/glad/glad.cpp"
 $725 = $724&1; //@line 7108 "libraries/glad/glad.cpp"
 HEAP8[1252378] = $725; //@line 7108 "libraries/glad/glad.cpp"
 $726 = (__ZN4gladL7has_extEPKc(473202)|0); //@line 7109 "libraries/glad/glad.cpp"
 $727 = $726&1; //@line 7109 "libraries/glad/glad.cpp"
 HEAP8[1252379] = $727; //@line 7109 "libraries/glad/glad.cpp"
 $728 = (__ZN4gladL7has_extEPKc(473248)|0); //@line 7110 "libraries/glad/glad.cpp"
 $729 = $728&1; //@line 7110 "libraries/glad/glad.cpp"
 HEAP8[1252380] = $729; //@line 7110 "libraries/glad/glad.cpp"
 $730 = (__ZN4gladL7has_extEPKc(473263)|0); //@line 7111 "libraries/glad/glad.cpp"
 $731 = $730&1; //@line 7111 "libraries/glad/glad.cpp"
 HEAP8[1252381] = $731; //@line 7111 "libraries/glad/glad.cpp"
 $732 = (__ZN4gladL7has_extEPKc(473289)|0); //@line 7112 "libraries/glad/glad.cpp"
 $733 = $732&1; //@line 7112 "libraries/glad/glad.cpp"
 HEAP8[1252382] = $733; //@line 7112 "libraries/glad/glad.cpp"
 $734 = (__ZN4gladL7has_extEPKc(473311)|0); //@line 7113 "libraries/glad/glad.cpp"
 $735 = $734&1; //@line 7113 "libraries/glad/glad.cpp"
 HEAP8[1252383] = $735; //@line 7113 "libraries/glad/glad.cpp"
 $736 = (__ZN4gladL7has_extEPKc(473332)|0); //@line 7114 "libraries/glad/glad.cpp"
 $737 = $736&1; //@line 7114 "libraries/glad/glad.cpp"
 HEAP8[1252384] = $737; //@line 7114 "libraries/glad/glad.cpp"
 $738 = (__ZN4gladL7has_extEPKc(473358)|0); //@line 7115 "libraries/glad/glad.cpp"
 $739 = $738&1; //@line 7115 "libraries/glad/glad.cpp"
 HEAP8[1252385] = $739; //@line 7115 "libraries/glad/glad.cpp"
 $740 = (__ZN4gladL7has_extEPKc(473386)|0); //@line 7116 "libraries/glad/glad.cpp"
 $741 = $740&1; //@line 7116 "libraries/glad/glad.cpp"
 HEAP8[1252386] = $741; //@line 7116 "libraries/glad/glad.cpp"
 $742 = (__ZN4gladL7has_extEPKc(473411)|0); //@line 7117 "libraries/glad/glad.cpp"
 $743 = $742&1; //@line 7117 "libraries/glad/glad.cpp"
 HEAP8[1252387] = $743; //@line 7117 "libraries/glad/glad.cpp"
 $744 = (__ZN4gladL7has_extEPKc(473449)|0); //@line 7118 "libraries/glad/glad.cpp"
 $745 = $744&1; //@line 7118 "libraries/glad/glad.cpp"
 HEAP8[1252388] = $745; //@line 7118 "libraries/glad/glad.cpp"
 $746 = (__ZN4gladL7has_extEPKc(473471)|0); //@line 7119 "libraries/glad/glad.cpp"
 $747 = $746&1; //@line 7119 "libraries/glad/glad.cpp"
 HEAP8[1252389] = $747; //@line 7119 "libraries/glad/glad.cpp"
 $748 = (__ZN4gladL7has_extEPKc(473490)|0); //@line 7120 "libraries/glad/glad.cpp"
 $749 = $748&1; //@line 7120 "libraries/glad/glad.cpp"
 HEAP8[1252390] = $749; //@line 7120 "libraries/glad/glad.cpp"
 $750 = (__ZN4gladL7has_extEPKc(473511)|0); //@line 7121 "libraries/glad/glad.cpp"
 $751 = $750&1; //@line 7121 "libraries/glad/glad.cpp"
 HEAP8[1252391] = $751; //@line 7121 "libraries/glad/glad.cpp"
 $752 = (__ZN4gladL7has_extEPKc(473544)|0); //@line 7122 "libraries/glad/glad.cpp"
 $753 = $752&1; //@line 7122 "libraries/glad/glad.cpp"
 HEAP8[1252392] = $753; //@line 7122 "libraries/glad/glad.cpp"
 $754 = (__ZN4gladL7has_extEPKc(473578)|0); //@line 7123 "libraries/glad/glad.cpp"
 $755 = $754&1; //@line 7123 "libraries/glad/glad.cpp"
 HEAP8[1252393] = $755; //@line 7123 "libraries/glad/glad.cpp"
 $756 = (__ZN4gladL7has_extEPKc(473606)|0); //@line 7124 "libraries/glad/glad.cpp"
 $757 = $756&1; //@line 7124 "libraries/glad/glad.cpp"
 HEAP8[1252394] = $757; //@line 7124 "libraries/glad/glad.cpp"
 $758 = (__ZN4gladL7has_extEPKc(473634)|0); //@line 7125 "libraries/glad/glad.cpp"
 $759 = $758&1; //@line 7125 "libraries/glad/glad.cpp"
 HEAP8[1252395] = $759; //@line 7125 "libraries/glad/glad.cpp"
 $760 = (__ZN4gladL7has_extEPKc(473668)|0); //@line 7126 "libraries/glad/glad.cpp"
 $761 = $760&1; //@line 7126 "libraries/glad/glad.cpp"
 HEAP8[1252396] = $761; //@line 7126 "libraries/glad/glad.cpp"
 $762 = (__ZN4gladL7has_extEPKc(473694)|0); //@line 7127 "libraries/glad/glad.cpp"
 $763 = $762&1; //@line 7127 "libraries/glad/glad.cpp"
 HEAP8[1252397] = $763; //@line 7127 "libraries/glad/glad.cpp"
 $764 = (__ZN4gladL7has_extEPKc(473715)|0); //@line 7128 "libraries/glad/glad.cpp"
 $765 = $764&1; //@line 7128 "libraries/glad/glad.cpp"
 HEAP8[1252398] = $765; //@line 7128 "libraries/glad/glad.cpp"
 $766 = (__ZN4gladL7has_extEPKc(473742)|0); //@line 7129 "libraries/glad/glad.cpp"
 $767 = $766&1; //@line 7129 "libraries/glad/glad.cpp"
 HEAP8[1252399] = $767; //@line 7129 "libraries/glad/glad.cpp"
 $768 = (__ZN4gladL7has_extEPKc(473762)|0); //@line 7130 "libraries/glad/glad.cpp"
 $769 = $768&1; //@line 7130 "libraries/glad/glad.cpp"
 HEAP8[1252400] = $769; //@line 7130 "libraries/glad/glad.cpp"
 $770 = (__ZN4gladL7has_extEPKc(473784)|0); //@line 7131 "libraries/glad/glad.cpp"
 $771 = $770&1; //@line 7131 "libraries/glad/glad.cpp"
 HEAP8[1252401] = $771; //@line 7131 "libraries/glad/glad.cpp"
 $772 = (__ZN4gladL7has_extEPKc(473807)|0); //@line 7132 "libraries/glad/glad.cpp"
 $773 = $772&1; //@line 7132 "libraries/glad/glad.cpp"
 HEAP8[1252402] = $773; //@line 7132 "libraries/glad/glad.cpp"
 $774 = (__ZN4gladL7has_extEPKc(473842)|0); //@line 7133 "libraries/glad/glad.cpp"
 $775 = $774&1; //@line 7133 "libraries/glad/glad.cpp"
 HEAP8[1252403] = $775; //@line 7133 "libraries/glad/glad.cpp"
 $776 = (__ZN4gladL7has_extEPKc(473883)|0); //@line 7134 "libraries/glad/glad.cpp"
 $777 = $776&1; //@line 7134 "libraries/glad/glad.cpp"
 HEAP8[1252404] = $777; //@line 7134 "libraries/glad/glad.cpp"
 $778 = (__ZN4gladL7has_extEPKc(473906)|0); //@line 7135 "libraries/glad/glad.cpp"
 $779 = $778&1; //@line 7135 "libraries/glad/glad.cpp"
 HEAP8[1252405] = $779; //@line 7135 "libraries/glad/glad.cpp"
 $780 = (__ZN4gladL7has_extEPKc(473936)|0); //@line 7136 "libraries/glad/glad.cpp"
 $781 = $780&1; //@line 7136 "libraries/glad/glad.cpp"
 HEAP8[1252406] = $781; //@line 7136 "libraries/glad/glad.cpp"
 $782 = (__ZN4gladL7has_extEPKc(473975)|0); //@line 7137 "libraries/glad/glad.cpp"
 $783 = $782&1; //@line 7137 "libraries/glad/glad.cpp"
 HEAP8[1252407] = $783; //@line 7137 "libraries/glad/glad.cpp"
 $784 = (__ZN4gladL7has_extEPKc(473994)|0); //@line 7138 "libraries/glad/glad.cpp"
 $785 = $784&1; //@line 7138 "libraries/glad/glad.cpp"
 HEAP8[1252408] = $785; //@line 7138 "libraries/glad/glad.cpp"
 $786 = (__ZN4gladL7has_extEPKc(474017)|0); //@line 7139 "libraries/glad/glad.cpp"
 $787 = $786&1; //@line 7139 "libraries/glad/glad.cpp"
 HEAP8[1252409] = $787; //@line 7139 "libraries/glad/glad.cpp"
 $788 = (__ZN4gladL7has_extEPKc(474042)|0); //@line 7140 "libraries/glad/glad.cpp"
 $789 = $788&1; //@line 7140 "libraries/glad/glad.cpp"
 HEAP8[1252410] = $789; //@line 7140 "libraries/glad/glad.cpp"
 $790 = (__ZN4gladL7has_extEPKc(474068)|0); //@line 7141 "libraries/glad/glad.cpp"
 $791 = $790&1; //@line 7141 "libraries/glad/glad.cpp"
 HEAP8[1252411] = $791; //@line 7141 "libraries/glad/glad.cpp"
 $792 = (__ZN4gladL7has_extEPKc(474101)|0); //@line 7142 "libraries/glad/glad.cpp"
 $793 = $792&1; //@line 7142 "libraries/glad/glad.cpp"
 HEAP8[1252412] = $793; //@line 7142 "libraries/glad/glad.cpp"
 $794 = (__ZN4gladL7has_extEPKc(474119)|0); //@line 7143 "libraries/glad/glad.cpp"
 $795 = $794&1; //@line 7143 "libraries/glad/glad.cpp"
 HEAP8[1252413] = $795; //@line 7143 "libraries/glad/glad.cpp"
 $796 = (__ZN4gladL7has_extEPKc(474136)|0); //@line 7144 "libraries/glad/glad.cpp"
 $797 = $796&1; //@line 7144 "libraries/glad/glad.cpp"
 HEAP8[1252414] = $797; //@line 7144 "libraries/glad/glad.cpp"
 $798 = (__ZN4gladL7has_extEPKc(474158)|0); //@line 7145 "libraries/glad/glad.cpp"
 $799 = $798&1; //@line 7145 "libraries/glad/glad.cpp"
 HEAP8[1252415] = $799; //@line 7145 "libraries/glad/glad.cpp"
 $800 = (__ZN4gladL7has_extEPKc(474179)|0); //@line 7146 "libraries/glad/glad.cpp"
 $801 = $800&1; //@line 7146 "libraries/glad/glad.cpp"
 HEAP8[1252416] = $801; //@line 7146 "libraries/glad/glad.cpp"
 $802 = (__ZN4gladL7has_extEPKc(474204)|0); //@line 7147 "libraries/glad/glad.cpp"
 $803 = $802&1; //@line 7147 "libraries/glad/glad.cpp"
 HEAP8[1252417] = $803; //@line 7147 "libraries/glad/glad.cpp"
 $804 = (__ZN4gladL7has_extEPKc(474222)|0); //@line 7148 "libraries/glad/glad.cpp"
 $805 = $804&1; //@line 7148 "libraries/glad/glad.cpp"
 HEAP8[1252418] = $805; //@line 7148 "libraries/glad/glad.cpp"
 $806 = (__ZN4gladL7has_extEPKc(474244)|0); //@line 7149 "libraries/glad/glad.cpp"
 $807 = $806&1; //@line 7149 "libraries/glad/glad.cpp"
 HEAP8[1252419] = $807; //@line 7149 "libraries/glad/glad.cpp"
 $808 = (__ZN4gladL7has_extEPKc(474263)|0); //@line 7150 "libraries/glad/glad.cpp"
 $809 = $808&1; //@line 7150 "libraries/glad/glad.cpp"
 HEAP8[1252420] = $809; //@line 7150 "libraries/glad/glad.cpp"
 $810 = (__ZN4gladL7has_extEPKc(474284)|0); //@line 7151 "libraries/glad/glad.cpp"
 $811 = $810&1; //@line 7151 "libraries/glad/glad.cpp"
 HEAP8[1252421] = $811; //@line 7151 "libraries/glad/glad.cpp"
 $812 = (__ZN4gladL7has_extEPKc(474303)|0); //@line 7152 "libraries/glad/glad.cpp"
 $813 = $812&1; //@line 7152 "libraries/glad/glad.cpp"
 HEAP8[1252422] = $813; //@line 7152 "libraries/glad/glad.cpp"
 $814 = (__ZN4gladL7has_extEPKc(474334)|0); //@line 7153 "libraries/glad/glad.cpp"
 $815 = $814&1; //@line 7153 "libraries/glad/glad.cpp"
 HEAP8[1252423] = $815; //@line 7153 "libraries/glad/glad.cpp"
 $816 = (__ZN4gladL7has_extEPKc(474361)|0); //@line 7154 "libraries/glad/glad.cpp"
 $817 = $816&1; //@line 7154 "libraries/glad/glad.cpp"
 HEAP8[1252424] = $817; //@line 7154 "libraries/glad/glad.cpp"
 $818 = (__ZN4gladL7has_extEPKc(474389)|0); //@line 7155 "libraries/glad/glad.cpp"
 $819 = $818&1; //@line 7155 "libraries/glad/glad.cpp"
 HEAP8[1252425] = $819; //@line 7155 "libraries/glad/glad.cpp"
 $820 = (__ZN4gladL7has_extEPKc(474401)|0); //@line 7156 "libraries/glad/glad.cpp"
 $821 = $820&1; //@line 7156 "libraries/glad/glad.cpp"
 HEAP8[1252426] = $821; //@line 7156 "libraries/glad/glad.cpp"
 $822 = (__ZN4gladL7has_extEPKc(474422)|0); //@line 7157 "libraries/glad/glad.cpp"
 $823 = $822&1; //@line 7157 "libraries/glad/glad.cpp"
 HEAP8[1252427] = $823; //@line 7157 "libraries/glad/glad.cpp"
 $824 = (__ZN4gladL7has_extEPKc(474441)|0); //@line 7158 "libraries/glad/glad.cpp"
 $825 = $824&1; //@line 7158 "libraries/glad/glad.cpp"
 HEAP8[1252428] = $825; //@line 7158 "libraries/glad/glad.cpp"
 $826 = (__ZN4gladL7has_extEPKc(474474)|0); //@line 7159 "libraries/glad/glad.cpp"
 $827 = $826&1; //@line 7159 "libraries/glad/glad.cpp"
 HEAP8[1252429] = $827; //@line 7159 "libraries/glad/glad.cpp"
 $828 = (__ZN4gladL7has_extEPKc(474498)|0); //@line 7160 "libraries/glad/glad.cpp"
 $829 = $828&1; //@line 7160 "libraries/glad/glad.cpp"
 HEAP8[1252430] = $829; //@line 7160 "libraries/glad/glad.cpp"
 $830 = (__ZN4gladL7has_extEPKc(474530)|0); //@line 7161 "libraries/glad/glad.cpp"
 $831 = $830&1; //@line 7161 "libraries/glad/glad.cpp"
 HEAP8[1252431] = $831; //@line 7161 "libraries/glad/glad.cpp"
 $832 = (__ZN4gladL7has_extEPKc(474553)|0); //@line 7162 "libraries/glad/glad.cpp"
 $833 = $832&1; //@line 7162 "libraries/glad/glad.cpp"
 HEAP8[1252432] = $833; //@line 7162 "libraries/glad/glad.cpp"
 $834 = (__ZN4gladL7has_extEPKc(474585)|0); //@line 7163 "libraries/glad/glad.cpp"
 $835 = $834&1; //@line 7163 "libraries/glad/glad.cpp"
 HEAP8[1252433] = $835; //@line 7163 "libraries/glad/glad.cpp"
 $836 = (__ZN4gladL7has_extEPKc(474615)|0); //@line 7164 "libraries/glad/glad.cpp"
 $837 = $836&1; //@line 7164 "libraries/glad/glad.cpp"
 HEAP8[1252434] = $837; //@line 7164 "libraries/glad/glad.cpp"
 $838 = (__ZN4gladL7has_extEPKc(474654)|0); //@line 7165 "libraries/glad/glad.cpp"
 $839 = $838&1; //@line 7165 "libraries/glad/glad.cpp"
 HEAP8[1252435] = $839; //@line 7165 "libraries/glad/glad.cpp"
 $840 = (__ZN4gladL7has_extEPKc(474681)|0); //@line 7166 "libraries/glad/glad.cpp"
 $841 = $840&1; //@line 7166 "libraries/glad/glad.cpp"
 HEAP8[1252436] = $841; //@line 7166 "libraries/glad/glad.cpp"
 $842 = (__ZN4gladL7has_extEPKc(474705)|0); //@line 7167 "libraries/glad/glad.cpp"
 $843 = $842&1; //@line 7167 "libraries/glad/glad.cpp"
 HEAP8[1252437] = $843; //@line 7167 "libraries/glad/glad.cpp"
 $844 = (__ZN4gladL7has_extEPKc(474728)|0); //@line 7168 "libraries/glad/glad.cpp"
 $845 = $844&1; //@line 7168 "libraries/glad/glad.cpp"
 HEAP8[1252438] = $845; //@line 7168 "libraries/glad/glad.cpp"
 $846 = (__ZN4gladL7has_extEPKc(474762)|0); //@line 7169 "libraries/glad/glad.cpp"
 $847 = $846&1; //@line 7169 "libraries/glad/glad.cpp"
 HEAP8[1252439] = $847; //@line 7169 "libraries/glad/glad.cpp"
 $848 = (__ZN4gladL7has_extEPKc(474781)|0); //@line 7170 "libraries/glad/glad.cpp"
 $849 = $848&1; //@line 7170 "libraries/glad/glad.cpp"
 HEAP8[1252440] = $849; //@line 7170 "libraries/glad/glad.cpp"
 $850 = (__ZN4gladL7has_extEPKc(474800)|0); //@line 7171 "libraries/glad/glad.cpp"
 $851 = $850&1; //@line 7171 "libraries/glad/glad.cpp"
 HEAP8[1252441] = $851; //@line 7171 "libraries/glad/glad.cpp"
 $852 = (__ZN4gladL7has_extEPKc(474832)|0); //@line 7172 "libraries/glad/glad.cpp"
 $853 = $852&1; //@line 7172 "libraries/glad/glad.cpp"
 HEAP8[1252442] = $853; //@line 7172 "libraries/glad/glad.cpp"
 $854 = (__ZN4gladL7has_extEPKc(474850)|0); //@line 7173 "libraries/glad/glad.cpp"
 $855 = $854&1; //@line 7173 "libraries/glad/glad.cpp"
 HEAP8[1252443] = $855; //@line 7173 "libraries/glad/glad.cpp"
 $856 = (__ZN4gladL7has_extEPKc(474867)|0); //@line 7174 "libraries/glad/glad.cpp"
 $857 = $856&1; //@line 7174 "libraries/glad/glad.cpp"
 HEAP8[1252444] = $857; //@line 7174 "libraries/glad/glad.cpp"
 $858 = (__ZN4gladL7has_extEPKc(474887)|0); //@line 7175 "libraries/glad/glad.cpp"
 $859 = $858&1; //@line 7175 "libraries/glad/glad.cpp"
 HEAP8[1252445] = $859; //@line 7175 "libraries/glad/glad.cpp"
 $860 = (__ZN4gladL7has_extEPKc(474910)|0); //@line 7176 "libraries/glad/glad.cpp"
 $861 = $860&1; //@line 7176 "libraries/glad/glad.cpp"
 HEAP8[1252446] = $861; //@line 7176 "libraries/glad/glad.cpp"
 $862 = (__ZN4gladL7has_extEPKc(474944)|0); //@line 7177 "libraries/glad/glad.cpp"
 $863 = $862&1; //@line 7177 "libraries/glad/glad.cpp"
 HEAP8[1252447] = $863; //@line 7177 "libraries/glad/glad.cpp"
 $864 = (__ZN4gladL7has_extEPKc(474971)|0); //@line 7178 "libraries/glad/glad.cpp"
 $865 = $864&1; //@line 7178 "libraries/glad/glad.cpp"
 HEAP8[1252448] = $865; //@line 7178 "libraries/glad/glad.cpp"
 $866 = (__ZN4gladL7has_extEPKc(474997)|0); //@line 7179 "libraries/glad/glad.cpp"
 $867 = $866&1; //@line 7179 "libraries/glad/glad.cpp"
 HEAP8[1252449] = $867; //@line 7179 "libraries/glad/glad.cpp"
 $868 = (__ZN4gladL7has_extEPKc(475019)|0); //@line 7180 "libraries/glad/glad.cpp"
 $869 = $868&1; //@line 7180 "libraries/glad/glad.cpp"
 HEAP8[1252450] = $869; //@line 7180 "libraries/glad/glad.cpp"
 $870 = (__ZN4gladL7has_extEPKc(475049)|0); //@line 7181 "libraries/glad/glad.cpp"
 $871 = $870&1; //@line 7181 "libraries/glad/glad.cpp"
 HEAP8[1252451] = $871; //@line 7181 "libraries/glad/glad.cpp"
 $872 = (__ZN4gladL7has_extEPKc(475080)|0); //@line 7182 "libraries/glad/glad.cpp"
 $873 = $872&1; //@line 7182 "libraries/glad/glad.cpp"
 HEAP8[1252452] = $873; //@line 7182 "libraries/glad/glad.cpp"
 $874 = (__ZN4gladL7has_extEPKc(475101)|0); //@line 7183 "libraries/glad/glad.cpp"
 $875 = $874&1; //@line 7183 "libraries/glad/glad.cpp"
 HEAP8[1252453] = $875; //@line 7183 "libraries/glad/glad.cpp"
 $876 = (__ZN4gladL7has_extEPKc(475134)|0); //@line 7184 "libraries/glad/glad.cpp"
 $877 = $876&1; //@line 7184 "libraries/glad/glad.cpp"
 HEAP8[1252454] = $877; //@line 7184 "libraries/glad/glad.cpp"
 $878 = (__ZN4gladL7has_extEPKc(475153)|0); //@line 7185 "libraries/glad/glad.cpp"
 $879 = $878&1; //@line 7185 "libraries/glad/glad.cpp"
 HEAP8[1252455] = $879; //@line 7185 "libraries/glad/glad.cpp"
 $880 = (__ZN4gladL7has_extEPKc(475173)|0); //@line 7186 "libraries/glad/glad.cpp"
 $881 = $880&1; //@line 7186 "libraries/glad/glad.cpp"
 HEAP8[1252456] = $881; //@line 7186 "libraries/glad/glad.cpp"
 $882 = (__ZN4gladL7has_extEPKc(475197)|0); //@line 7187 "libraries/glad/glad.cpp"
 $883 = $882&1; //@line 7187 "libraries/glad/glad.cpp"
 HEAP8[1252457] = $883; //@line 7187 "libraries/glad/glad.cpp"
 $884 = (__ZN4gladL7has_extEPKc(475215)|0); //@line 7188 "libraries/glad/glad.cpp"
 $885 = $884&1; //@line 7188 "libraries/glad/glad.cpp"
 HEAP8[1252458] = $885; //@line 7188 "libraries/glad/glad.cpp"
 $886 = (__ZN4gladL7has_extEPKc(475239)|0); //@line 7189 "libraries/glad/glad.cpp"
 $887 = $886&1; //@line 7189 "libraries/glad/glad.cpp"
 HEAP8[1252459] = $887; //@line 7189 "libraries/glad/glad.cpp"
 $888 = (__ZN4gladL7has_extEPKc(475256)|0); //@line 7190 "libraries/glad/glad.cpp"
 $889 = $888&1; //@line 7190 "libraries/glad/glad.cpp"
 HEAP8[1252460] = $889; //@line 7190 "libraries/glad/glad.cpp"
 $890 = (__ZN4gladL7has_extEPKc(475281)|0); //@line 7191 "libraries/glad/glad.cpp"
 $891 = $890&1; //@line 7191 "libraries/glad/glad.cpp"
 HEAP8[1252461] = $891; //@line 7191 "libraries/glad/glad.cpp"
 $892 = (__ZN4gladL7has_extEPKc(475300)|0); //@line 7192 "libraries/glad/glad.cpp"
 $893 = $892&1; //@line 7192 "libraries/glad/glad.cpp"
 HEAP8[1252462] = $893; //@line 7192 "libraries/glad/glad.cpp"
 $894 = (__ZN4gladL7has_extEPKc(475323)|0); //@line 7193 "libraries/glad/glad.cpp"
 $895 = $894&1; //@line 7193 "libraries/glad/glad.cpp"
 HEAP8[1252463] = $895; //@line 7193 "libraries/glad/glad.cpp"
 $896 = (__ZN4gladL7has_extEPKc(475359)|0); //@line 7194 "libraries/glad/glad.cpp"
 $897 = $896&1; //@line 7194 "libraries/glad/glad.cpp"
 HEAP8[1252464] = $897; //@line 7194 "libraries/glad/glad.cpp"
 $898 = (__ZN4gladL7has_extEPKc(475388)|0); //@line 7195 "libraries/glad/glad.cpp"
 $899 = $898&1; //@line 7195 "libraries/glad/glad.cpp"
 HEAP8[1252465] = $899; //@line 7195 "libraries/glad/glad.cpp"
 $900 = (__ZN4gladL7has_extEPKc(475414)|0); //@line 7196 "libraries/glad/glad.cpp"
 $901 = $900&1; //@line 7196 "libraries/glad/glad.cpp"
 HEAP8[1252466] = $901; //@line 7196 "libraries/glad/glad.cpp"
 $902 = (__ZN4gladL7has_extEPKc(475446)|0); //@line 7197 "libraries/glad/glad.cpp"
 $903 = $902&1; //@line 7197 "libraries/glad/glad.cpp"
 HEAP8[1252467] = $903; //@line 7197 "libraries/glad/glad.cpp"
 $904 = (__ZN4gladL7has_extEPKc(475472)|0); //@line 7198 "libraries/glad/glad.cpp"
 $905 = $904&1; //@line 7198 "libraries/glad/glad.cpp"
 HEAP8[1252468] = $905; //@line 7198 "libraries/glad/glad.cpp"
 $906 = (__ZN4gladL7has_extEPKc(475497)|0); //@line 7199 "libraries/glad/glad.cpp"
 $907 = $906&1; //@line 7199 "libraries/glad/glad.cpp"
 HEAP8[1252469] = $907; //@line 7199 "libraries/glad/glad.cpp"
 $908 = (__ZN4gladL7has_extEPKc(475523)|0); //@line 7200 "libraries/glad/glad.cpp"
 $909 = $908&1; //@line 7200 "libraries/glad/glad.cpp"
 HEAP8[1252470] = $909; //@line 7200 "libraries/glad/glad.cpp"
 $910 = (__ZN4gladL7has_extEPKc(475564)|0); //@line 7201 "libraries/glad/glad.cpp"
 $911 = $910&1; //@line 7201 "libraries/glad/glad.cpp"
 HEAP8[1252471] = $911; //@line 7201 "libraries/glad/glad.cpp"
 $912 = (__ZN4gladL7has_extEPKc(475599)|0); //@line 7202 "libraries/glad/glad.cpp"
 $913 = $912&1; //@line 7202 "libraries/glad/glad.cpp"
 HEAP8[1252472] = $913; //@line 7202 "libraries/glad/glad.cpp"
 $914 = (__ZN4gladL7has_extEPKc(475625)|0); //@line 7203 "libraries/glad/glad.cpp"
 $915 = $914&1; //@line 7203 "libraries/glad/glad.cpp"
 HEAP8[1252473] = $915; //@line 7203 "libraries/glad/glad.cpp"
 $916 = (__ZN4gladL7has_extEPKc(475653)|0); //@line 7204 "libraries/glad/glad.cpp"
 $917 = $916&1; //@line 7204 "libraries/glad/glad.cpp"
 HEAP8[1252474] = $917; //@line 7204 "libraries/glad/glad.cpp"
 $918 = (__ZN4gladL7has_extEPKc(475681)|0); //@line 7205 "libraries/glad/glad.cpp"
 $919 = $918&1; //@line 7205 "libraries/glad/glad.cpp"
 HEAP8[1252475] = $919; //@line 7205 "libraries/glad/glad.cpp"
 $920 = (__ZN4gladL7has_extEPKc(475708)|0); //@line 7206 "libraries/glad/glad.cpp"
 $921 = $920&1; //@line 7206 "libraries/glad/glad.cpp"
 HEAP8[1252476] = $921; //@line 7206 "libraries/glad/glad.cpp"
 $922 = (__ZN4gladL7has_extEPKc(475727)|0); //@line 7207 "libraries/glad/glad.cpp"
 $923 = $922&1; //@line 7207 "libraries/glad/glad.cpp"
 HEAP8[1252477] = $923; //@line 7207 "libraries/glad/glad.cpp"
 $924 = (__ZN4gladL7has_extEPKc(475755)|0); //@line 7208 "libraries/glad/glad.cpp"
 $925 = $924&1; //@line 7208 "libraries/glad/glad.cpp"
 HEAP8[1252478] = $925; //@line 7208 "libraries/glad/glad.cpp"
 $926 = (__ZN4gladL7has_extEPKc(475777)|0); //@line 7209 "libraries/glad/glad.cpp"
 $927 = $926&1; //@line 7209 "libraries/glad/glad.cpp"
 HEAP8[1252479] = $927; //@line 7209 "libraries/glad/glad.cpp"
 $928 = (__ZN4gladL7has_extEPKc(475804)|0); //@line 7210 "libraries/glad/glad.cpp"
 $929 = $928&1; //@line 7210 "libraries/glad/glad.cpp"
 HEAP8[1252480] = $929; //@line 7210 "libraries/glad/glad.cpp"
 $930 = (__ZN4gladL7has_extEPKc(475842)|0); //@line 7211 "libraries/glad/glad.cpp"
 $931 = $930&1; //@line 7211 "libraries/glad/glad.cpp"
 HEAP8[1252481] = $931; //@line 7211 "libraries/glad/glad.cpp"
 $932 = (__ZN4gladL7has_extEPKc(475868)|0); //@line 7212 "libraries/glad/glad.cpp"
 $933 = $932&1; //@line 7212 "libraries/glad/glad.cpp"
 HEAP8[1252482] = $933; //@line 7212 "libraries/glad/glad.cpp"
 $934 = (__ZN4gladL7has_extEPKc(475897)|0); //@line 7213 "libraries/glad/glad.cpp"
 $935 = $934&1; //@line 7213 "libraries/glad/glad.cpp"
 HEAP8[1252483] = $935; //@line 7213 "libraries/glad/glad.cpp"
 $936 = (__ZN4gladL7has_extEPKc(475922)|0); //@line 7214 "libraries/glad/glad.cpp"
 $937 = $936&1; //@line 7214 "libraries/glad/glad.cpp"
 HEAP8[1252484] = $937; //@line 7214 "libraries/glad/glad.cpp"
 $938 = (__ZN4gladL7has_extEPKc(475948)|0); //@line 7215 "libraries/glad/glad.cpp"
 $939 = $938&1; //@line 7215 "libraries/glad/glad.cpp"
 HEAP8[1252485] = $939; //@line 7215 "libraries/glad/glad.cpp"
 $940 = (__ZN4gladL7has_extEPKc(475984)|0); //@line 7216 "libraries/glad/glad.cpp"
 $941 = $940&1; //@line 7216 "libraries/glad/glad.cpp"
 HEAP8[1252486] = $941; //@line 7216 "libraries/glad/glad.cpp"
 $942 = (__ZN4gladL7has_extEPKc(476004)|0); //@line 7217 "libraries/glad/glad.cpp"
 $943 = $942&1; //@line 7217 "libraries/glad/glad.cpp"
 HEAP8[1252487] = $943; //@line 7217 "libraries/glad/glad.cpp"
 $944 = (__ZN4gladL7has_extEPKc(476038)|0); //@line 7218 "libraries/glad/glad.cpp"
 $945 = $944&1; //@line 7218 "libraries/glad/glad.cpp"
 HEAP8[1252488] = $945; //@line 7218 "libraries/glad/glad.cpp"
 $946 = (__ZN4gladL7has_extEPKc(476073)|0); //@line 7219 "libraries/glad/glad.cpp"
 $947 = $946&1; //@line 7219 "libraries/glad/glad.cpp"
 HEAP8[1252489] = $947; //@line 7219 "libraries/glad/glad.cpp"
 $948 = (__ZN4gladL7has_extEPKc(476095)|0); //@line 7220 "libraries/glad/glad.cpp"
 $949 = $948&1; //@line 7220 "libraries/glad/glad.cpp"
 HEAP8[1252490] = $949; //@line 7220 "libraries/glad/glad.cpp"
 $950 = (__ZN4gladL7has_extEPKc(476115)|0); //@line 7221 "libraries/glad/glad.cpp"
 $951 = $950&1; //@line 7221 "libraries/glad/glad.cpp"
 HEAP8[1252491] = $951; //@line 7221 "libraries/glad/glad.cpp"
 $952 = (__ZN4gladL7has_extEPKc(476136)|0); //@line 7222 "libraries/glad/glad.cpp"
 $953 = $952&1; //@line 7222 "libraries/glad/glad.cpp"
 HEAP8[1252492] = $953; //@line 7222 "libraries/glad/glad.cpp"
 $954 = (__ZN4gladL7has_extEPKc(476158)|0); //@line 7223 "libraries/glad/glad.cpp"
 $955 = $954&1; //@line 7223 "libraries/glad/glad.cpp"
 HEAP8[1252493] = $955; //@line 7223 "libraries/glad/glad.cpp"
 $956 = (__ZN4gladL7has_extEPKc(476184)|0); //@line 7224 "libraries/glad/glad.cpp"
 $957 = $956&1; //@line 7224 "libraries/glad/glad.cpp"
 HEAP8[1252494] = $957; //@line 7224 "libraries/glad/glad.cpp"
 $958 = (__ZN4gladL7has_extEPKc(476207)|0); //@line 7225 "libraries/glad/glad.cpp"
 $959 = $958&1; //@line 7225 "libraries/glad/glad.cpp"
 HEAP8[1252495] = $959; //@line 7225 "libraries/glad/glad.cpp"
 $960 = (__ZN4gladL7has_extEPKc(476224)|0); //@line 7226 "libraries/glad/glad.cpp"
 $961 = $960&1; //@line 7226 "libraries/glad/glad.cpp"
 HEAP8[1252496] = $961; //@line 7226 "libraries/glad/glad.cpp"
 $962 = (__ZN4gladL7has_extEPKc(476242)|0); //@line 7227 "libraries/glad/glad.cpp"
 $963 = $962&1; //@line 7227 "libraries/glad/glad.cpp"
 HEAP8[1252497] = $963; //@line 7227 "libraries/glad/glad.cpp"
 $964 = (__ZN4gladL7has_extEPKc(476290)|0); //@line 7228 "libraries/glad/glad.cpp"
 $965 = $964&1; //@line 7228 "libraries/glad/glad.cpp"
 HEAP8[1252498] = $965; //@line 7228 "libraries/glad/glad.cpp"
 $966 = (__ZN4gladL7has_extEPKc(476309)|0); //@line 7229 "libraries/glad/glad.cpp"
 $967 = $966&1; //@line 7229 "libraries/glad/glad.cpp"
 HEAP8[1252499] = $967; //@line 7229 "libraries/glad/glad.cpp"
 $968 = (__ZN4gladL7has_extEPKc(476333)|0); //@line 7230 "libraries/glad/glad.cpp"
 $969 = $968&1; //@line 7230 "libraries/glad/glad.cpp"
 HEAP8[1252500] = $969; //@line 7230 "libraries/glad/glad.cpp"
 $970 = (__ZN4gladL7has_extEPKc(476356)|0); //@line 7231 "libraries/glad/glad.cpp"
 $971 = $970&1; //@line 7231 "libraries/glad/glad.cpp"
 HEAP8[1252501] = $971; //@line 7231 "libraries/glad/glad.cpp"
 $972 = (__ZN4gladL7has_extEPKc(476377)|0); //@line 7232 "libraries/glad/glad.cpp"
 $973 = $972&1; //@line 7232 "libraries/glad/glad.cpp"
 HEAP8[1252502] = $973; //@line 7232 "libraries/glad/glad.cpp"
 $974 = (__ZN4gladL7has_extEPKc(476399)|0); //@line 7233 "libraries/glad/glad.cpp"
 $975 = $974&1; //@line 7233 "libraries/glad/glad.cpp"
 HEAP8[1252503] = $975; //@line 7233 "libraries/glad/glad.cpp"
 $976 = (__ZN4gladL7has_extEPKc(476427)|0); //@line 7234 "libraries/glad/glad.cpp"
 $977 = $976&1; //@line 7234 "libraries/glad/glad.cpp"
 HEAP8[1252504] = $977; //@line 7234 "libraries/glad/glad.cpp"
 $978 = (__ZN4gladL7has_extEPKc(476451)|0); //@line 7235 "libraries/glad/glad.cpp"
 $979 = $978&1; //@line 7235 "libraries/glad/glad.cpp"
 HEAP8[1252505] = $979; //@line 7235 "libraries/glad/glad.cpp"
 $980 = (__ZN4gladL7has_extEPKc(476479)|0); //@line 7236 "libraries/glad/glad.cpp"
 $981 = $980&1; //@line 7236 "libraries/glad/glad.cpp"
 HEAP8[1252506] = $981; //@line 7236 "libraries/glad/glad.cpp"
 return; //@line 7237 "libraries/glad/glad.cpp"
}
function __ZN4gladL7has_extEPKc($ext) {
 $ext = $ext|0;
 var $$13 = 0, $$4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $extensions$06 = 0, $index$011 = 0, $num = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $num = sp;
 $0 = HEAP32[166754]|0; //@line 34 "libraries/glad/glad.cpp"
 $1 = ($0|0)<(3); //@line 34 "libraries/glad/glad.cpp"
 if (!($1)) {
  $18 = HEAP32[165587]|0; //@line 58 "libraries/glad/glad.cpp"
  FUNCTION_TABLE_vii[$18 & 8191](33309,$num); //@line 58 "libraries/glad/glad.cpp"
  $19 = HEAP32[$num>>2]|0; //@line 61 "libraries/glad/glad.cpp"
  $20 = ($19|0)>(0); //@line 61 "libraries/glad/glad.cpp"
  if ($20) {
   $index$011 = 0;
  } else {
   $$13 = 0;
   STACKTOP = sp;return ($$13|0);
  }
  while(1) {
   $24 = HEAP32[165216]|0; //@line 62 "libraries/glad/glad.cpp"
   $25 = (FUNCTION_TABLE_iii[$24 & 8191](7939,$index$011)|0); //@line 62 "libraries/glad/glad.cpp"
   $26 = (_strcmp($25,$ext)|0); //@line 63 "libraries/glad/glad.cpp"
   $27 = ($26|0)==(0); //@line 63 "libraries/glad/glad.cpp"
   $23 = (($index$011) + 1)|0; //@line 61 "libraries/glad/glad.cpp"
   if ($27) {
    $$13 = 1;
    label = 12;
    break;
   }
   $21 = HEAP32[$num>>2]|0; //@line 61 "libraries/glad/glad.cpp"
   $22 = ($23|0)<($21|0); //@line 61 "libraries/glad/glad.cpp"
   if ($22) {
    $index$011 = $23;
   } else {
    $$13 = 0;
    label = 12;
    break;
   }
  }
  if ((label|0) == 12) {
   STACKTOP = sp;return ($$13|0);
  }
 }
 $2 = HEAP32[165140]|0; //@line 38 "libraries/glad/glad.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191](7939)|0); //@line 38 "libraries/glad/glad.cpp"
 $4 = ($3|0)==(0|0); //@line 39 "libraries/glad/glad.cpp"
 $5 = ($ext|0)==(0|0); //@line 39 "libraries/glad/glad.cpp"
 $or$cond = $5 | $4; //@line 39 "libraries/glad/glad.cpp"
 if ($or$cond) {
  $$4 = 0;
  STACKTOP = sp;return ($$4|0); //@line 70 "libraries/glad/glad.cpp"
 }
 $6 = (_strstr($3,$ext)|0); //@line 44 "libraries/glad/glad.cpp"
 $7 = ($6|0)==(0|0); //@line 45 "libraries/glad/glad.cpp"
 if ($7) {
  $$4 = 0;
  STACKTOP = sp;return ($$4|0); //@line 70 "libraries/glad/glad.cpp"
 }
 $8 = (_strlen($ext)|0); //@line 49 "libraries/glad/glad.cpp"
 $10 = $6;$extensions$06 = $3;
 L18: while(1) {
  $9 = (($10) + ($8)|0); //@line 49 "libraries/glad/glad.cpp"
  $11 = ($10|0)==($extensions$06|0); //@line 50 "libraries/glad/glad.cpp"
  if ($11) {
   label = 7;
  } else {
   $12 = ((($10)) + -1|0); //@line 50 "libraries/glad/glad.cpp"
   $13 = HEAP8[$12>>0]|0; //@line 50 "libraries/glad/glad.cpp"
   $14 = ($13<<24>>24)==(32); //@line 50 "libraries/glad/glad.cpp"
   if ($14) {
    label = 7;
   }
  }
  if ((label|0) == 7) {
   label = 0;
   $15 = HEAP8[$9>>0]|0; //@line 51 "libraries/glad/glad.cpp"
   switch ($15<<24>>24) {
   case 0: case 32:  {
    $$4 = 1;
    label = 13;
    break L18;
    break;
   }
   default: {
   }
   }
  }
  $16 = (_strstr($9,$ext)|0); //@line 44 "libraries/glad/glad.cpp"
  $17 = ($16|0)==(0|0); //@line 45 "libraries/glad/glad.cpp"
  if ($17) {
   $$4 = 0;
   label = 13;
   break;
  } else {
   $10 = $16;$extensions$06 = $9;
  }
 }
 if ((label|0) == 13) {
  STACKTOP = sp;return ($$4|0); //@line 70 "libraries/glad/glad.cpp"
 }
 return (0)|0;
}
function __ZN7lodepng5StateD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (33380); //@line 6005 "libraries/lodepng/lodepng.cpp"
 $0 = ((($this)) + 108|0); //@line 2605 "libraries/lodepng/lodepng.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 2605 "libraries/lodepng/lodepng.cpp"
 $2 = ($1|0)==(0|0); //@line 2605 "libraries/lodepng/lodepng.cpp"
 if (!($2)) {
  _free($1); //@line 76 "libraries/lodepng/lodepng.cpp"
 }
 HEAP32[$0>>2] = 0; //@line 2606 "libraries/lodepng/lodepng.cpp"
 $3 = ((($this)) + 112|0); //@line 2607 "libraries/lodepng/lodepng.cpp"
 HEAP32[$3>>2] = 0; //@line 2607 "libraries/lodepng/lodepng.cpp"
 $4 = ((($this)) + 132|0); //@line 4751 "libraries/lodepng/lodepng.cpp"
 __Z20lodepng_info_cleanupP11LodePNGInfo($4); //@line 4751 "libraries/lodepng/lodepng.cpp"
 return; //@line 6007 "libraries/lodepng/lodepng.cpp"
}
function __Z20lodepng_info_cleanupP11LodePNGInfo($info) {
 $info = $info|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 20|0); //@line 2605 "libraries/lodepng/lodepng.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 2605 "libraries/lodepng/lodepng.cpp"
 $2 = ($1|0)==(0|0); //@line 2605 "libraries/lodepng/lodepng.cpp"
 if (!($2)) {
  _free($1); //@line 76 "libraries/lodepng/lodepng.cpp"
 }
 HEAP32[$0>>2] = 0; //@line 2606 "libraries/lodepng/lodepng.cpp"
 $3 = ((($info)) + 24|0); //@line 2607 "libraries/lodepng/lodepng.cpp"
 HEAP32[$3>>2] = 0; //@line 2607 "libraries/lodepng/lodepng.cpp"
 $4 = ((($info)) + 60|0); //@line 2743 "libraries/lodepng/lodepng.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 2743 "libraries/lodepng/lodepng.cpp"
 $6 = ($5|0)==(0); //@line 2743 "libraries/lodepng/lodepng.cpp"
 $7 = ((($info)) + 64|0); //@line 2748 "libraries/lodepng/lodepng.cpp"
 $$pre$i = ((($info)) + 68|0); //@line 2749 "libraries/lodepng/lodepng.cpp"
 if (!($6)) {
  $i$01$i = 0;
  while(1) {
   $8 = HEAP32[$7>>2]|0; //@line 2745 "libraries/lodepng/lodepng.cpp"
   $9 = (($8) + ($i$01$i<<2)|0); //@line 2745 "libraries/lodepng/lodepng.cpp"
   $10 = HEAP32[$9>>2]|0; //@line 321 "libraries/lodepng/lodepng.cpp"
   _free($10); //@line 76 "libraries/lodepng/lodepng.cpp"
   HEAP32[$9>>2] = 0; //@line 322 "libraries/lodepng/lodepng.cpp"
   $11 = HEAP32[$$pre$i>>2]|0; //@line 2746 "libraries/lodepng/lodepng.cpp"
   $12 = (($11) + ($i$01$i<<2)|0); //@line 2746 "libraries/lodepng/lodepng.cpp"
   $13 = HEAP32[$12>>2]|0; //@line 321 "libraries/lodepng/lodepng.cpp"
   _free($13); //@line 76 "libraries/lodepng/lodepng.cpp"
   HEAP32[$12>>2] = 0; //@line 322 "libraries/lodepng/lodepng.cpp"
   $14 = (($i$01$i) + 1)|0; //@line 2743 "libraries/lodepng/lodepng.cpp"
   $15 = HEAP32[$4>>2]|0; //@line 2743 "libraries/lodepng/lodepng.cpp"
   $16 = ($14|0)==($15|0); //@line 2743 "libraries/lodepng/lodepng.cpp"
   if ($16) {
    break;
   } else {
    $i$01$i = $14;
   }
  }
 }
 $17 = HEAP32[$7>>2]|0; //@line 2748 "libraries/lodepng/lodepng.cpp"
 _free($17); //@line 76 "libraries/lodepng/lodepng.cpp"
 $18 = HEAP32[$$pre$i>>2]|0; //@line 2749 "libraries/lodepng/lodepng.cpp"
 _free($18); //@line 76 "libraries/lodepng/lodepng.cpp"
 __ZL20LodePNGIText_cleanupP11LodePNGInfo($info); //@line 2905 "libraries/lodepng/lodepng.cpp"
 $19 = ((($info)) + 136|0); //@line 2707 "libraries/lodepng/lodepng.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 2707 "libraries/lodepng/lodepng.cpp"
 _free($20); //@line 76 "libraries/lodepng/lodepng.cpp"
 $21 = ((($info)) + 140|0); //@line 2707 "libraries/lodepng/lodepng.cpp"
 $22 = HEAP32[$21>>2]|0; //@line 2707 "libraries/lodepng/lodepng.cpp"
 _free($22); //@line 76 "libraries/lodepng/lodepng.cpp"
 $23 = ((($info)) + 144|0); //@line 2707 "libraries/lodepng/lodepng.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 2707 "libraries/lodepng/lodepng.cpp"
 _free($24); //@line 76 "libraries/lodepng/lodepng.cpp"
 return; //@line 2909 "libraries/lodepng/lodepng.cpp"
}
function __ZL20LodePNGIText_cleanupP11LodePNGInfo($info) {
 $info = $info|0;
 var $$pre = 0, $$pre2 = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 72|0); //@line 2808 "libraries/lodepng/lodepng.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 2808 "libraries/lodepng/lodepng.cpp"
 $2 = ($1|0)==(0); //@line 2808 "libraries/lodepng/lodepng.cpp"
 $3 = ((($info)) + 76|0); //@line 2815 "libraries/lodepng/lodepng.cpp"
 $$pre = ((($info)) + 80|0); //@line 2816 "libraries/lodepng/lodepng.cpp"
 $$pre2 = ((($info)) + 84|0); //@line 2817 "libraries/lodepng/lodepng.cpp"
 $$pre4 = ((($info)) + 88|0); //@line 2818 "libraries/lodepng/lodepng.cpp"
 if (!($2)) {
  $i$01 = 0;
  while(1) {
   $4 = HEAP32[$3>>2]|0; //@line 2810 "libraries/lodepng/lodepng.cpp"
   $5 = (($4) + ($i$01<<2)|0); //@line 2810 "libraries/lodepng/lodepng.cpp"
   $6 = HEAP32[$5>>2]|0; //@line 321 "libraries/lodepng/lodepng.cpp"
   _free($6); //@line 76 "libraries/lodepng/lodepng.cpp"
   HEAP32[$5>>2] = 0; //@line 322 "libraries/lodepng/lodepng.cpp"
   $7 = HEAP32[$$pre>>2]|0; //@line 2811 "libraries/lodepng/lodepng.cpp"
   $8 = (($7) + ($i$01<<2)|0); //@line 2811 "libraries/lodepng/lodepng.cpp"
   $9 = HEAP32[$8>>2]|0; //@line 321 "libraries/lodepng/lodepng.cpp"
   _free($9); //@line 76 "libraries/lodepng/lodepng.cpp"
   HEAP32[$8>>2] = 0; //@line 322 "libraries/lodepng/lodepng.cpp"
   $10 = HEAP32[$$pre2>>2]|0; //@line 2812 "libraries/lodepng/lodepng.cpp"
   $11 = (($10) + ($i$01<<2)|0); //@line 2812 "libraries/lodepng/lodepng.cpp"
   $12 = HEAP32[$11>>2]|0; //@line 321 "libraries/lodepng/lodepng.cpp"
   _free($12); //@line 76 "libraries/lodepng/lodepng.cpp"
   HEAP32[$11>>2] = 0; //@line 322 "libraries/lodepng/lodepng.cpp"
   $13 = HEAP32[$$pre4>>2]|0; //@line 2813 "libraries/lodepng/lodepng.cpp"
   $14 = (($13) + ($i$01<<2)|0); //@line 2813 "libraries/lodepng/lodepng.cpp"
   $15 = HEAP32[$14>>2]|0; //@line 321 "libraries/lodepng/lodepng.cpp"
   _free($15); //@line 76 "libraries/lodepng/lodepng.cpp"
   HEAP32[$14>>2] = 0; //@line 322 "libraries/lodepng/lodepng.cpp"
   $16 = (($i$01) + 1)|0; //@line 2808 "libraries/lodepng/lodepng.cpp"
   $17 = HEAP32[$0>>2]|0; //@line 2808 "libraries/lodepng/lodepng.cpp"
   $18 = ($16|0)==($17|0); //@line 2808 "libraries/lodepng/lodepng.cpp"
   if ($18) {
    break;
   } else {
    $i$01 = $16;
   }
  }
 }
 $19 = HEAP32[$3>>2]|0; //@line 2815 "libraries/lodepng/lodepng.cpp"
 _free($19); //@line 76 "libraries/lodepng/lodepng.cpp"
 $20 = HEAP32[$$pre>>2]|0; //@line 2816 "libraries/lodepng/lodepng.cpp"
 _free($20); //@line 76 "libraries/lodepng/lodepng.cpp"
 $21 = HEAP32[$$pre2>>2]|0; //@line 2817 "libraries/lodepng/lodepng.cpp"
 _free($21); //@line 76 "libraries/lodepng/lodepng.cpp"
 $22 = HEAP32[$$pre4>>2]|0; //@line 2818 "libraries/lodepng/lodepng.cpp"
 _free($22); //@line 76 "libraries/lodepng/lodepng.cpp"
 return; //@line 2819 "libraries/lodepng/lodepng.cpp"
}
function __ZN7lodepng5StateD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (33380); //@line 6005 "libraries/lodepng/lodepng.cpp"
 $0 = ((($this)) + 108|0); //@line 2605 "libraries/lodepng/lodepng.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 2605 "libraries/lodepng/lodepng.cpp"
 $2 = ($1|0)==(0|0); //@line 2605 "libraries/lodepng/lodepng.cpp"
 if (!($2)) {
  _free($1); //@line 76 "libraries/lodepng/lodepng.cpp"
 }
 HEAP32[$0>>2] = 0; //@line 2606 "libraries/lodepng/lodepng.cpp"
 $3 = ((($this)) + 112|0); //@line 2607 "libraries/lodepng/lodepng.cpp"
 HEAP32[$3>>2] = 0; //@line 2607 "libraries/lodepng/lodepng.cpp"
 $4 = ((($this)) + 132|0); //@line 4751 "libraries/lodepng/lodepng.cpp"
 __Z20lodepng_info_cleanupP11LodePNGInfo($4); //@line 4751 "libraries/lodepng/lodepng.cpp"
 __ZdlPv($this); //@line 6005 "libraries/lodepng/lodepng.cpp"
 return; //@line 6007 "libraries/lodepng/lodepng.cpp"
}
function __Z28lodepng_huffman_code_lengthsPjPKjjj($lengths,$frequencies,$numcodes,$maxbitlen) {
 $lengths = $lengths|0;
 $frequencies = $frequencies|0;
 $numcodes = $numcodes|0;
 $maxbitlen = $maxbitlen|0;
 var $$0 = 0, $$lcssa76 = 0, $$pre = 0, $$pre$i$i = 0, $$pre$i$i19 = 0, $$pre$i$i27 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0.0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $coins$0$lcssa = 0, $coins$047 = 0, $coins$047$phi = 0, $exitcond = 0, $i$01$i = 0, $i$01$i$i = 0, $i$01$i12 = 0, $i$01$i18 = 0, $i$01$i2 = 0, $i$01$i25 = 0, $i$01$i33 = 0;
 var $i$01$i5 = 0, $i$01$i8 = 0, $i$01$i9 = 0, $i$061 = 0, $i$255 = 0, $i$255$lcssa = 0, $i$344 = 0, $i$442 = 0, $i$442$phi = 0, $j$04$i = 0, $j$04$i24 = 0, $j$051 = 0, $j$1$i = 0, $j$1$i31 = 0, $j$141 = 0, $not$ = 0, $numcoins$049 = 0, $numcoins$1$lcssa = 0, $numcoins$143 = 0, $numcoins$2 = 0;
 var $numpresent$059 = 0, $numpresent$1 = 0, $numpresent$1$lcssa = 0, $numprev$050 = 0, $numprev$050$phi = 0, $or$cond = 0, $prev_row$0$lcssa = 0, $prev_row$048 = 0, $prev_row$048$phi = 0, $sum$060 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($numcodes|0)==(0); //@line 749 "libraries/lodepng/lodepng.cpp"
 if ($0) {
  $$0 = 80;
  return ($$0|0); //@line 855 "libraries/lodepng/lodepng.cpp"
 } else {
  $i$061 = 0;$numpresent$059 = 0;$sum$060 = 0;
 }
 while(1) {
  $1 = (($frequencies) + ($i$061<<2)|0); //@line 753 "libraries/lodepng/lodepng.cpp"
  $2 = HEAP32[$1>>2]|0; //@line 753 "libraries/lodepng/lodepng.cpp"
  $3 = (($2) + ($sum$060))|0; //@line 756 "libraries/lodepng/lodepng.cpp"
  $not$ = ($2|0)!=(0); //@line 753 "libraries/lodepng/lodepng.cpp"
  $4 = $not$&1; //@line 753 "libraries/lodepng/lodepng.cpp"
  $numpresent$1 = (($4) + ($numpresent$059))|0; //@line 753 "libraries/lodepng/lodepng.cpp"
  $5 = (($i$061) + 1)|0; //@line 751 "libraries/lodepng/lodepng.cpp"
  $6 = ($5|0)==($numcodes|0); //@line 751 "libraries/lodepng/lodepng.cpp"
  if ($6) {
   $$lcssa76 = $3;$numpresent$1$lcssa = $numpresent$1;
   break;
  } else {
   $i$061 = $5;$numpresent$059 = $numpresent$1;$sum$060 = $3;
  }
 }
 $7 = $numcodes << 2; //@line 760 "libraries/lodepng/lodepng.cpp"
 _memset(($lengths|0),0,($7|0))|0; //@line 760 "libraries/lodepng/lodepng.cpp"
 switch ($numpresent$1$lcssa|0) {
 case 0:  {
  $8 = ((($lengths)) + 4|0); //@line 769 "libraries/lodepng/lodepng.cpp"
  HEAP32[$8>>2] = 1; //@line 769 "libraries/lodepng/lodepng.cpp"
  HEAP32[$lengths>>2] = 1; //@line 769 "libraries/lodepng/lodepng.cpp"
  $$0 = 0;
  return ($$0|0); //@line 855 "libraries/lodepng/lodepng.cpp"
  break;
 }
 case 1:  {
  $i$255 = 0;
  while(1) {
   $9 = (($frequencies) + ($i$255<<2)|0); //@line 775 "libraries/lodepng/lodepng.cpp"
   $10 = HEAP32[$9>>2]|0; //@line 775 "libraries/lodepng/lodepng.cpp"
   $11 = ($10|0)==(0); //@line 775 "libraries/lodepng/lodepng.cpp"
   if (!($11)) {
    $i$255$lcssa = $i$255;
    break;
   }
   $16 = (($i$255) + 1)|0; //@line 773 "libraries/lodepng/lodepng.cpp"
   $17 = ($16|0)==($numcodes|0); //@line 773 "libraries/lodepng/lodepng.cpp"
   if ($17) {
    $$0 = 0;
    label = 66;
    break;
   } else {
    $i$255 = $16;
   }
  }
  if ((label|0) == 66) {
   return ($$0|0); //@line 855 "libraries/lodepng/lodepng.cpp"
  }
  $12 = (($lengths) + ($i$255$lcssa<<2)|0); //@line 777 "libraries/lodepng/lodepng.cpp"
  HEAP32[$12>>2] = 1; //@line 777 "libraries/lodepng/lodepng.cpp"
  $13 = ($i$255$lcssa|0)==(0); //@line 778 "libraries/lodepng/lodepng.cpp"
  $14 = $13&1; //@line 778 "libraries/lodepng/lodepng.cpp"
  $15 = (($lengths) + ($14<<2)|0); //@line 778 "libraries/lodepng/lodepng.cpp"
  HEAP32[$15>>2] = 1; //@line 778 "libraries/lodepng/lodepng.cpp"
  $$0 = 0;
  return ($$0|0); //@line 855 "libraries/lodepng/lodepng.cpp"
  break;
 }
 default: {
  $18 = $numpresent$1$lcssa << 1; //@line 788 "libraries/lodepng/lodepng.cpp"
  $19 = $numpresent$1$lcssa << 5; //@line 789 "libraries/lodepng/lodepng.cpp"
  $20 = (_malloc($19)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
  $21 = (_malloc($19)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
  $22 = ($20|0)!=(0|0); //@line 791 "libraries/lodepng/lodepng.cpp"
  $23 = ($21|0)!=(0|0); //@line 791 "libraries/lodepng/lodepng.cpp"
  $or$cond = $22 & $23; //@line 791 "libraries/lodepng/lodepng.cpp"
  if (!($or$cond)) {
   _free($20); //@line 76 "libraries/lodepng/lodepng.cpp"
   _free($21); //@line 76 "libraries/lodepng/lodepng.cpp"
   $$0 = 83;
   return ($$0|0); //@line 855 "libraries/lodepng/lodepng.cpp"
  }
  $24 = ($18|0)==(0); //@line 707 "libraries/lodepng/lodepng.cpp"
  if (!($24)) {
   $i$01$i = 0;
   while(1) {
    $25 = (($20) + ($i$01$i<<4)|0); //@line 188 "libraries/lodepng/lodepng.cpp"
    HEAP32[$25>>2] = 0; //@line 188 "libraries/lodepng/lodepng.cpp"
    $26 = (((($20) + ($i$01$i<<4)|0)) + 8|0); //@line 189 "libraries/lodepng/lodepng.cpp"
    HEAP32[$26>>2] = 0; //@line 189 "libraries/lodepng/lodepng.cpp"
    $27 = (((($20) + ($i$01$i<<4)|0)) + 4|0); //@line 189 "libraries/lodepng/lodepng.cpp"
    HEAP32[$27>>2] = 0; //@line 189 "libraries/lodepng/lodepng.cpp"
    $28 = (($i$01$i) + 1)|0; //@line 707 "libraries/lodepng/lodepng.cpp"
    $29 = ($28|0)==($18|0); //@line 707 "libraries/lodepng/lodepng.cpp"
    if ($29) {
     $i$01$i2 = 0;
     break;
    } else {
     $i$01$i = $28;
    }
   }
   while(1) {
    $30 = (($21) + ($i$01$i2<<4)|0); //@line 188 "libraries/lodepng/lodepng.cpp"
    HEAP32[$30>>2] = 0; //@line 188 "libraries/lodepng/lodepng.cpp"
    $31 = (((($21) + ($i$01$i2<<4)|0)) + 8|0); //@line 189 "libraries/lodepng/lodepng.cpp"
    HEAP32[$31>>2] = 0; //@line 189 "libraries/lodepng/lodepng.cpp"
    $32 = (((($21) + ($i$01$i2<<4)|0)) + 4|0); //@line 189 "libraries/lodepng/lodepng.cpp"
    HEAP32[$32>>2] = 0; //@line 189 "libraries/lodepng/lodepng.cpp"
    $33 = (($i$01$i2) + 1)|0; //@line 707 "libraries/lodepng/lodepng.cpp"
    $34 = ($33|0)==($18|0); //@line 707 "libraries/lodepng/lodepng.cpp"
    if ($34) {
     break;
    } else {
     $i$01$i2 = $33;
    }
   }
  }
  $35 = (+($$lcssa76>>>0)); //@line 730 "libraries/lodepng/lodepng.cpp"
  $i$01$i8 = 0;$j$04$i = 0;
  while(1) {
   $36 = (($frequencies) + ($i$01$i8<<2)|0); //@line 728 "libraries/lodepng/lodepng.cpp"
   $37 = HEAP32[$36>>2]|0; //@line 728 "libraries/lodepng/lodepng.cpp"
   $38 = ($37|0)==(0); //@line 728 "libraries/lodepng/lodepng.cpp"
   if ($38) {
    $j$1$i = $j$04$i;
   } else {
    $39 = (+($37>>>0)); //@line 730 "libraries/lodepng/lodepng.cpp"
    $40 = $39 / $35; //@line 730 "libraries/lodepng/lodepng.cpp"
    $41 = (((($20) + ($j$04$i<<4)|0)) + 12|0); //@line 730 "libraries/lodepng/lodepng.cpp"
    HEAPF32[$41>>2] = $40; //@line 730 "libraries/lodepng/lodepng.cpp"
    $42 = (($20) + ($j$04$i<<4)|0); //@line 731 "libraries/lodepng/lodepng.cpp"
    $43 = (((($20) + ($j$04$i<<4)|0)) + 4|0); //@line 196 "libraries/lodepng/lodepng.cpp"
    $44 = HEAP32[$43>>2]|0; //@line 196 "libraries/lodepng/lodepng.cpp"
    $45 = (($44) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
    $46 = $45 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
    $47 = (((($20) + ($j$04$i<<4)|0)) + 8|0); //@line 155 "libraries/lodepng/lodepng.cpp"
    $48 = HEAP32[$47>>2]|0; //@line 155 "libraries/lodepng/lodepng.cpp"
    $49 = ($48>>>0)<($46>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
    if ($49) {
     $50 = $48 << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
     $51 = ($50>>>0)<($46>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
     $52 = ($45*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
     $53 = $52 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
     $54 = $51 ? $46 : $53; //@line 157 "libraries/lodepng/lodepng.cpp"
     $55 = HEAP32[$42>>2]|0; //@line 158 "libraries/lodepng/lodepng.cpp"
     $56 = (_realloc($55,$54)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
     $57 = ($56|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
     if (!($57)) {
      HEAP32[$47>>2] = $54; //@line 161 "libraries/lodepng/lodepng.cpp"
      HEAP32[$42>>2] = $56; //@line 162 "libraries/lodepng/lodepng.cpp"
      $59 = $56;
      label = 19;
     }
    } else {
     $$pre$i$i = HEAP32[$42>>2]|0; //@line 197 "libraries/lodepng/lodepng.cpp"
     $59 = $$pre$i$i;
     label = 19;
    }
    if ((label|0) == 19) {
     label = 0;
     HEAP32[$43>>2] = $45; //@line 173 "libraries/lodepng/lodepng.cpp"
     $58 = (($59) + ($44<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
     HEAP32[$58>>2] = $i$01$i8; //@line 197 "libraries/lodepng/lodepng.cpp"
    }
    $60 = (($j$04$i) + 1)|0; //@line 732 "libraries/lodepng/lodepng.cpp"
    $j$1$i = $60;
   }
   $61 = (($i$01$i8) + 1)|0; //@line 726 "libraries/lodepng/lodepng.cpp"
   $62 = ($61|0)==($numcodes|0); //@line 726 "libraries/lodepng/lodepng.cpp"
   if ($62) {
    break;
   } else {
    $i$01$i8 = $61;$j$04$i = $j$1$i;
   }
  }
  _qsort($20,$numpresent$1$lcssa,16,4246); //@line 803 "libraries/lodepng/lodepng.cpp"
  $63 = ($maxbitlen|0)==(0); //@line 807 "libraries/lodepng/lodepng.cpp"
  if ($63) {
   $coins$0$lcssa = $20;$prev_row$0$lcssa = $21;
  } else {
   $64 = (+($$lcssa76>>>0)); //@line 730 "libraries/lodepng/lodepng.cpp"
   $coins$047 = $20;$j$051 = 1;$numcoins$049 = $numpresent$1$lcssa;$numprev$050 = 0;$prev_row$048 = $21;
   while(1) {
    $66 = ($numprev$050|0)==(0); //@line 713 "libraries/lodepng/lodepng.cpp"
    if (!($66)) {
     $i$01$i9 = 0;
     while(1) {
      $67 = (($prev_row$048) + ($i$01$i9<<4)|0); //@line 713 "libraries/lodepng/lodepng.cpp"
      $68 = (((($prev_row$048) + ($i$01$i9<<4)|0)) + 8|0); //@line 147 "libraries/lodepng/lodepng.cpp"
      HEAP32[$68>>2] = 0; //@line 147 "libraries/lodepng/lodepng.cpp"
      $69 = (((($prev_row$048) + ($i$01$i9<<4)|0)) + 4|0); //@line 147 "libraries/lodepng/lodepng.cpp"
      HEAP32[$69>>2] = 0; //@line 147 "libraries/lodepng/lodepng.cpp"
      $70 = HEAP32[$67>>2]|0; //@line 148 "libraries/lodepng/lodepng.cpp"
      _free($70); //@line 76 "libraries/lodepng/lodepng.cpp"
      HEAP32[$67>>2] = 0; //@line 149 "libraries/lodepng/lodepng.cpp"
      $71 = (($i$01$i9) + 1)|0; //@line 713 "libraries/lodepng/lodepng.cpp"
      $72 = ($71|0)==($numprev$050|0); //@line 713 "libraries/lodepng/lodepng.cpp"
      if ($72) {
       $i$01$i12 = 0;
       break;
      } else {
       $i$01$i9 = $71;
      }
     }
     while(1) {
      $73 = (($prev_row$048) + ($i$01$i12<<4)|0); //@line 188 "libraries/lodepng/lodepng.cpp"
      HEAP32[$73>>2] = 0; //@line 188 "libraries/lodepng/lodepng.cpp"
      $74 = (((($prev_row$048) + ($i$01$i12<<4)|0)) + 8|0); //@line 189 "libraries/lodepng/lodepng.cpp"
      HEAP32[$74>>2] = 0; //@line 189 "libraries/lodepng/lodepng.cpp"
      $75 = (((($prev_row$048) + ($i$01$i12<<4)|0)) + 4|0); //@line 189 "libraries/lodepng/lodepng.cpp"
      HEAP32[$75>>2] = 0; //@line 189 "libraries/lodepng/lodepng.cpp"
      $76 = (($i$01$i12) + 1)|0; //@line 707 "libraries/lodepng/lodepng.cpp"
      $77 = ($76|0)==($numprev$050|0); //@line 707 "libraries/lodepng/lodepng.cpp"
      if ($77) {
       break;
      } else {
       $i$01$i12 = $76;
      }
     }
    }
    $78 = ($numcoins$049>>>0)>(1); //@line 821 "libraries/lodepng/lodepng.cpp"
    if ($78) {
     $111 = 1;$i$344 = 0;$numcoins$143 = 0;
     while(1) {
      $79 = (($numcoins$143) + 1)|0; //@line 824 "libraries/lodepng/lodepng.cpp"
      $80 = (($prev_row$048) + ($numcoins$143<<4)|0); //@line 824 "libraries/lodepng/lodepng.cpp"
      $81 = (((($coins$047) + ($i$344<<4)|0)) + 12|0); //@line 693 "libraries/lodepng/lodepng.cpp"
      $82 = HEAP32[$81>>2]|0; //@line 693 "libraries/lodepng/lodepng.cpp"
      $83 = (((($prev_row$048) + ($numcoins$143<<4)|0)) + 12|0); //@line 693 "libraries/lodepng/lodepng.cpp"
      HEAP32[$83>>2] = $82; //@line 693 "libraries/lodepng/lodepng.cpp"
      $84 = (((($coins$047) + ($i$344<<4)|0)) + 4|0); //@line 205 "libraries/lodepng/lodepng.cpp"
      $85 = HEAP32[$84>>2]|0; //@line 205 "libraries/lodepng/lodepng.cpp"
      $86 = $85 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
      $87 = (((($prev_row$048) + ($numcoins$143<<4)|0)) + 8|0); //@line 155 "libraries/lodepng/lodepng.cpp"
      $88 = HEAP32[$87>>2]|0; //@line 155 "libraries/lodepng/lodepng.cpp"
      $89 = ($88>>>0)<($86>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
      if ($89) {
       $90 = $88 << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
       $91 = ($90>>>0)<($86>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
       $92 = ($85*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
       $93 = $92 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
       $94 = $91 ? $86 : $93; //@line 157 "libraries/lodepng/lodepng.cpp"
       $95 = HEAP32[$80>>2]|0; //@line 158 "libraries/lodepng/lodepng.cpp"
       $96 = (_realloc($95,$94)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       $97 = ($96|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
       if (!($97)) {
        HEAP32[$87>>2] = $94; //@line 161 "libraries/lodepng/lodepng.cpp"
        HEAP32[$80>>2] = $96; //@line 162 "libraries/lodepng/lodepng.cpp"
        $$pre = HEAP32[$84>>2]|0; //@line 206 "libraries/lodepng/lodepng.cpp"
        $100 = $$pre;
        label = 32;
       }
      } else {
       $100 = $85;
       label = 32;
      }
      if ((label|0) == 32) {
       label = 0;
       $98 = (((($prev_row$048) + ($numcoins$143<<4)|0)) + 4|0); //@line 173 "libraries/lodepng/lodepng.cpp"
       HEAP32[$98>>2] = $85; //@line 173 "libraries/lodepng/lodepng.cpp"
       $99 = ($100|0)==(0); //@line 206 "libraries/lodepng/lodepng.cpp"
       if (!($99)) {
        $101 = (($coins$047) + ($i$344<<4)|0); //@line 206 "libraries/lodepng/lodepng.cpp"
        $102 = HEAP32[$101>>2]|0; //@line 206 "libraries/lodepng/lodepng.cpp"
        $103 = HEAP32[$80>>2]|0; //@line 206 "libraries/lodepng/lodepng.cpp"
        $i$01$i$i = 0;
        while(1) {
         $104 = (($102) + ($i$01$i$i<<2)|0); //@line 206 "libraries/lodepng/lodepng.cpp"
         $105 = HEAP32[$104>>2]|0; //@line 206 "libraries/lodepng/lodepng.cpp"
         $106 = (($103) + ($i$01$i$i<<2)|0); //@line 206 "libraries/lodepng/lodepng.cpp"
         HEAP32[$106>>2] = $105; //@line 206 "libraries/lodepng/lodepng.cpp"
         $107 = (($i$01$i$i) + 1)|0; //@line 206 "libraries/lodepng/lodepng.cpp"
         $108 = HEAP32[$84>>2]|0; //@line 206 "libraries/lodepng/lodepng.cpp"
         $109 = ($107|0)==($108|0); //@line 206 "libraries/lodepng/lodepng.cpp"
         if ($109) {
          break;
         } else {
          $i$01$i$i = $107;
         }
        }
       }
      }
      $110 = (((($coins$047) + ($111<<4)|0)) + 4|0); //@line 700 "libraries/lodepng/lodepng.cpp"
      $112 = HEAP32[$110>>2]|0; //@line 700 "libraries/lodepng/lodepng.cpp"
      $113 = ($112|0)==(0); //@line 700 "libraries/lodepng/lodepng.cpp"
      if (!($113)) {
       $114 = (($coins$047) + ($111<<4)|0); //@line 700 "libraries/lodepng/lodepng.cpp"
       $115 = (((($prev_row$048) + ($numcoins$143<<4)|0)) + 4|0); //@line 196 "libraries/lodepng/lodepng.cpp"
       $i$01$i18 = 0;
       while(1) {
        $116 = HEAP32[$114>>2]|0; //@line 700 "libraries/lodepng/lodepng.cpp"
        $117 = (($116) + ($i$01$i18<<2)|0); //@line 700 "libraries/lodepng/lodepng.cpp"
        $118 = HEAP32[$117>>2]|0; //@line 700 "libraries/lodepng/lodepng.cpp"
        $119 = HEAP32[$115>>2]|0; //@line 196 "libraries/lodepng/lodepng.cpp"
        $120 = (($119) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
        $121 = $120 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
        $122 = HEAP32[$87>>2]|0; //@line 155 "libraries/lodepng/lodepng.cpp"
        $123 = ($122>>>0)<($121>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
        if ($123) {
         $124 = $122 << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
         $125 = ($124>>>0)<($121>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
         $126 = ($120*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
         $127 = $126 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
         $128 = $125 ? $121 : $127; //@line 157 "libraries/lodepng/lodepng.cpp"
         $129 = HEAP32[$80>>2]|0; //@line 158 "libraries/lodepng/lodepng.cpp"
         $130 = (_realloc($129,$128)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
         $131 = ($130|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
         if (!($131)) {
          HEAP32[$87>>2] = $128; //@line 161 "libraries/lodepng/lodepng.cpp"
          HEAP32[$80>>2] = $130; //@line 162 "libraries/lodepng/lodepng.cpp"
          $133 = $130;
          label = 41;
         }
        } else {
         $$pre$i$i19 = HEAP32[$80>>2]|0; //@line 197 "libraries/lodepng/lodepng.cpp"
         $133 = $$pre$i$i19;
         label = 41;
        }
        if ((label|0) == 41) {
         label = 0;
         HEAP32[$115>>2] = $120; //@line 173 "libraries/lodepng/lodepng.cpp"
         $132 = (($133) + ($119<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
         HEAP32[$132>>2] = $118; //@line 197 "libraries/lodepng/lodepng.cpp"
        }
        $134 = (($i$01$i18) + 1)|0; //@line 700 "libraries/lodepng/lodepng.cpp"
        $135 = HEAP32[$110>>2]|0; //@line 700 "libraries/lodepng/lodepng.cpp"
        $136 = ($134|0)==($135|0); //@line 700 "libraries/lodepng/lodepng.cpp"
        if ($136) {
         break;
        } else {
         $i$01$i18 = $134;
        }
       }
      }
      $137 = (((($coins$047) + ($111<<4)|0)) + 12|0); //@line 701 "libraries/lodepng/lodepng.cpp"
      $138 = +HEAPF32[$137>>2]; //@line 701 "libraries/lodepng/lodepng.cpp"
      $139 = +HEAPF32[$83>>2]; //@line 701 "libraries/lodepng/lodepng.cpp"
      $140 = $138 + $139; //@line 701 "libraries/lodepng/lodepng.cpp"
      HEAPF32[$83>>2] = $140; //@line 701 "libraries/lodepng/lodepng.cpp"
      $141 = (($i$344) + 2)|0; //@line 821 "libraries/lodepng/lodepng.cpp"
      $142 = $141 | 1; //@line 821 "libraries/lodepng/lodepng.cpp"
      $143 = ($142>>>0)<($numcoins$049>>>0); //@line 821 "libraries/lodepng/lodepng.cpp"
      if ($143) {
       $111 = $142;$i$344 = $141;$numcoins$143 = $79;
      } else {
       $numcoins$1$lcssa = $79;
       break;
      }
     }
    } else {
     $numcoins$1$lcssa = 0;
    }
    $144 = ($j$051>>>0)<($maxbitlen>>>0); //@line 829 "libraries/lodepng/lodepng.cpp"
    if ($144) {
     $145 = (($prev_row$048) + ($numcoins$1$lcssa<<4)|0); //@line 831 "libraries/lodepng/lodepng.cpp"
     $i$01$i25 = 0;$j$04$i24 = 0;
     while(1) {
      $146 = (($frequencies) + ($i$01$i25<<2)|0); //@line 728 "libraries/lodepng/lodepng.cpp"
      $147 = HEAP32[$146>>2]|0; //@line 728 "libraries/lodepng/lodepng.cpp"
      $148 = ($147|0)==(0); //@line 728 "libraries/lodepng/lodepng.cpp"
      if ($148) {
       $j$1$i31 = $j$04$i24;
      } else {
       $149 = (+($147>>>0)); //@line 730 "libraries/lodepng/lodepng.cpp"
       $150 = $149 / $64; //@line 730 "libraries/lodepng/lodepng.cpp"
       $151 = (((($145) + ($j$04$i24<<4)|0)) + 12|0); //@line 730 "libraries/lodepng/lodepng.cpp"
       HEAPF32[$151>>2] = $150; //@line 730 "libraries/lodepng/lodepng.cpp"
       $152 = (($145) + ($j$04$i24<<4)|0); //@line 731 "libraries/lodepng/lodepng.cpp"
       $153 = (((($145) + ($j$04$i24<<4)|0)) + 4|0); //@line 196 "libraries/lodepng/lodepng.cpp"
       $154 = HEAP32[$153>>2]|0; //@line 196 "libraries/lodepng/lodepng.cpp"
       $155 = (($154) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
       $156 = $155 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
       $157 = (((($145) + ($j$04$i24<<4)|0)) + 8|0); //@line 155 "libraries/lodepng/lodepng.cpp"
       $158 = HEAP32[$157>>2]|0; //@line 155 "libraries/lodepng/lodepng.cpp"
       $159 = ($158>>>0)<($156>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
       if ($159) {
        $160 = $158 << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
        $161 = ($160>>>0)<($156>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
        $162 = ($155*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
        $163 = $162 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
        $164 = $161 ? $156 : $163; //@line 157 "libraries/lodepng/lodepng.cpp"
        $165 = HEAP32[$152>>2]|0; //@line 158 "libraries/lodepng/lodepng.cpp"
        $166 = (_realloc($165,$164)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
        $167 = ($166|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
        if (!($167)) {
         HEAP32[$157>>2] = $164; //@line 161 "libraries/lodepng/lodepng.cpp"
         HEAP32[$152>>2] = $166; //@line 162 "libraries/lodepng/lodepng.cpp"
         $169 = $166;
         label = 51;
        }
       } else {
        $$pre$i$i27 = HEAP32[$152>>2]|0; //@line 197 "libraries/lodepng/lodepng.cpp"
        $169 = $$pre$i$i27;
        label = 51;
       }
       if ((label|0) == 51) {
        label = 0;
        HEAP32[$153>>2] = $155; //@line 173 "libraries/lodepng/lodepng.cpp"
        $168 = (($169) + ($154<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
        HEAP32[$168>>2] = $i$01$i25; //@line 197 "libraries/lodepng/lodepng.cpp"
       }
       $170 = (($j$04$i24) + 1)|0; //@line 732 "libraries/lodepng/lodepng.cpp"
       $j$1$i31 = $170;
      }
      $171 = (($i$01$i25) + 1)|0; //@line 726 "libraries/lodepng/lodepng.cpp"
      $172 = ($171|0)==($numcodes|0); //@line 726 "libraries/lodepng/lodepng.cpp"
      if ($172) {
       break;
      } else {
       $i$01$i25 = $171;$j$04$i24 = $j$1$i31;
      }
     }
     $173 = (($numcoins$1$lcssa) + ($numpresent$1$lcssa))|0; //@line 832 "libraries/lodepng/lodepng.cpp"
     $numcoins$2 = $173;
    } else {
     $numcoins$2 = $numcoins$1$lcssa;
    }
    _qsort($prev_row$048,$numcoins$2,16,4246); //@line 834 "libraries/lodepng/lodepng.cpp"
    $174 = (($j$051) + 1)|0; //@line 807 "libraries/lodepng/lodepng.cpp"
    $175 = ($174>>>0)>($maxbitlen>>>0); //@line 807 "libraries/lodepng/lodepng.cpp"
    if ($175) {
     $coins$0$lcssa = $prev_row$048;$prev_row$0$lcssa = $coins$047;
     break;
    } else {
     $prev_row$048$phi = $coins$047;$numprev$050$phi = $numcoins$049;$coins$047$phi = $prev_row$048;$j$051 = $174;$numcoins$049 = $numcoins$2;$prev_row$048 = $prev_row$048$phi;$numprev$050 = $numprev$050$phi;$coins$047 = $coins$047$phi;
    }
   }
  }
  $65 = ($numpresent$1$lcssa>>>0)>(1); //@line 841 "libraries/lodepng/lodepng.cpp"
  if ($65) {
   $177 = 1;$i$442 = 0;
   while(1) {
    $178 = (((($coins$0$lcssa) + ($i$442<<4)|0)) + 4|0); //@line 844 "libraries/lodepng/lodepng.cpp"
    $179 = HEAP32[$178>>2]|0; //@line 844 "libraries/lodepng/lodepng.cpp"
    $180 = ($179|0)==(0); //@line 844 "libraries/lodepng/lodepng.cpp"
    if (!($180)) {
     $181 = (($coins$0$lcssa) + ($i$442<<4)|0); //@line 844 "libraries/lodepng/lodepng.cpp"
     $182 = HEAP32[$181>>2]|0; //@line 844 "libraries/lodepng/lodepng.cpp"
     $j$141 = 0;
     while(1) {
      $183 = (($182) + ($j$141<<2)|0); //@line 844 "libraries/lodepng/lodepng.cpp"
      $184 = HEAP32[$183>>2]|0; //@line 844 "libraries/lodepng/lodepng.cpp"
      $185 = (($lengths) + ($184<<2)|0); //@line 844 "libraries/lodepng/lodepng.cpp"
      $186 = HEAP32[$185>>2]|0; //@line 844 "libraries/lodepng/lodepng.cpp"
      $187 = (($186) + 1)|0; //@line 844 "libraries/lodepng/lodepng.cpp"
      HEAP32[$185>>2] = $187; //@line 844 "libraries/lodepng/lodepng.cpp"
      $188 = (($j$141) + 1)|0; //@line 844 "libraries/lodepng/lodepng.cpp"
      $189 = HEAP32[$178>>2]|0; //@line 844 "libraries/lodepng/lodepng.cpp"
      $190 = ($188>>>0)<($189>>>0); //@line 844 "libraries/lodepng/lodepng.cpp"
      if ($190) {
       $j$141 = $188;
      } else {
       break;
      }
     }
    }
    $176 = (($177) + 1)|0; //@line 841 "libraries/lodepng/lodepng.cpp"
    $exitcond = ($176|0)==($numpresent$1$lcssa|0); //@line 841 "libraries/lodepng/lodepng.cpp"
    if ($exitcond) {
     break;
    } else {
     $i$442$phi = $177;$177 = $176;$i$442 = $i$442$phi;
    }
   }
  }
  if ($24) {
   _free($coins$0$lcssa); //@line 76 "libraries/lodepng/lodepng.cpp"
  } else {
   $i$01$i33 = 0;
   while(1) {
    $191 = (($coins$0$lcssa) + ($i$01$i33<<4)|0); //@line 713 "libraries/lodepng/lodepng.cpp"
    $192 = (((($coins$0$lcssa) + ($i$01$i33<<4)|0)) + 8|0); //@line 147 "libraries/lodepng/lodepng.cpp"
    HEAP32[$192>>2] = 0; //@line 147 "libraries/lodepng/lodepng.cpp"
    $193 = (((($coins$0$lcssa) + ($i$01$i33<<4)|0)) + 4|0); //@line 147 "libraries/lodepng/lodepng.cpp"
    HEAP32[$193>>2] = 0; //@line 147 "libraries/lodepng/lodepng.cpp"
    $194 = HEAP32[$191>>2]|0; //@line 148 "libraries/lodepng/lodepng.cpp"
    _free($194); //@line 76 "libraries/lodepng/lodepng.cpp"
    HEAP32[$191>>2] = 0; //@line 149 "libraries/lodepng/lodepng.cpp"
    $195 = (($i$01$i33) + 1)|0; //@line 713 "libraries/lodepng/lodepng.cpp"
    $196 = ($195|0)==($18|0); //@line 713 "libraries/lodepng/lodepng.cpp"
    if ($196) {
     break;
    } else {
     $i$01$i33 = $195;
    }
   }
   _free($coins$0$lcssa); //@line 76 "libraries/lodepng/lodepng.cpp"
   $i$01$i5 = 0;
   while(1) {
    $197 = (($prev_row$0$lcssa) + ($i$01$i5<<4)|0); //@line 713 "libraries/lodepng/lodepng.cpp"
    $198 = (((($prev_row$0$lcssa) + ($i$01$i5<<4)|0)) + 8|0); //@line 147 "libraries/lodepng/lodepng.cpp"
    HEAP32[$198>>2] = 0; //@line 147 "libraries/lodepng/lodepng.cpp"
    $199 = (((($prev_row$0$lcssa) + ($i$01$i5<<4)|0)) + 4|0); //@line 147 "libraries/lodepng/lodepng.cpp"
    HEAP32[$199>>2] = 0; //@line 147 "libraries/lodepng/lodepng.cpp"
    $200 = HEAP32[$197>>2]|0; //@line 148 "libraries/lodepng/lodepng.cpp"
    _free($200); //@line 76 "libraries/lodepng/lodepng.cpp"
    HEAP32[$197>>2] = 0; //@line 149 "libraries/lodepng/lodepng.cpp"
    $201 = (($i$01$i5) + 1)|0; //@line 713 "libraries/lodepng/lodepng.cpp"
    $202 = ($201|0)==($18|0); //@line 713 "libraries/lodepng/lodepng.cpp"
    if ($202) {
     break;
    } else {
     $i$01$i5 = $201;
    }
   }
  }
  _free($prev_row$0$lcssa); //@line 76 "libraries/lodepng/lodepng.cpp"
  $$0 = 0;
  return ($$0|0); //@line 855 "libraries/lodepng/lodepng.cpp"
 }
 }
 return (0)|0;
}
function __ZL12coin_comparePKvS0_($a,$b) {
 $a = $a|0;
 $b = $b|0;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($a)) + 12|0); //@line 717 "libraries/lodepng/lodepng.cpp"
 $1 = +HEAPF32[$0>>2]; //@line 717 "libraries/lodepng/lodepng.cpp"
 $2 = ((($b)) + 12|0); //@line 718 "libraries/lodepng/lodepng.cpp"
 $3 = +HEAPF32[$2>>2]; //@line 718 "libraries/lodepng/lodepng.cpp"
 $4 = $1 > $3; //@line 719 "libraries/lodepng/lodepng.cpp"
 $5 = $1 < $3; //@line 719 "libraries/lodepng/lodepng.cpp"
 $6 = $5 << 31 >> 31; //@line 719 "libraries/lodepng/lodepng.cpp"
 $7 = $4 ? 1 : $6; //@line 719 "libraries/lodepng/lodepng.cpp"
 return ($7|0); //@line 719 "libraries/lodepng/lodepng.cpp"
}
function __Z15lodepng_inflatePPhPjPKhjPK25LodePNGDecompressSettings($out,$outsize,$in,$insize,$settings) {
 $out = $out|0;
 $outsize = $outsize|0;
 $in = $in|0;
 $insize = $insize|0;
 $settings = $settings|0;
 var $$$i$i = 0, $$0$i$ph$i$i = 0, $$0$i116$i$i = 0, $$2$i = 0, $$lcssa480 = 0, $$lcssa481 = 0, $$lcssa482 = 0, $$lcssa485 = 0, $$lcssa485$lcssa = 0, $$lcssa487 = 0, $$lcssa487$lcssa = 0, $$lcssa497 = 0, $$lcssa497$lcssa = 0, $$lcssa497$lcssa498 = 0, $$lcssa497503 = 0, $$lcssa504 = 0, $$lcssa504$lcssa = 0, $$lcssa504$lcssa505 = 0, $$lcssa511 = 0, $$lcssa512 = 0;
 var $$lcssa541 = 0, $$lcssa542 = 0, $$lcssa543 = 0, $$off$i$i = 0, $$phi$trans$insert$i$i = 0, $$pre$i$i = 0, $$pre$i15$i$i$i = 0, $$pre$i15149$i$i$i = 0, $$pre108$i$i = 0, $$pre110$i$i = 0, $$sink$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0;
 var $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0;
 var $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0;
 var $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0;
 var $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0;
 var $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0;
 var $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0;
 var $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0;
 var $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0;
 var $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0;
 var $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0;
 var $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0;
 var $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0;
 var $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0;
 var $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0;
 var $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0;
 var $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0;
 var $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0;
 var $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0;
 var $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0;
 var $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0;
 var $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0;
 var $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $BFINAL$0$ph$i = 0;
 var $backward$061$i$i = 0, $bitlen_d$0$i$i$i = 0, $bitlen_ll$0$i$i$i = 0, $bp$0$ph$i = 0, $bp$079$i = 0, $bp$079$i$lcssa479 = 0, $bp$1$i = 0, $bp$10$i = 0, $bp$11$i = 0, $bp$12$i = 0, $bp$13$i = 0, $bp$14$i = 0, $bp$15$ph$i = 0, $bp$16$i = 0, $bp$16$i$ph = 0, $bp$17$i = 0, $bp$18$i = 0, $bp$19$i = 0, $bp$20$i = 0, $bp$20$i$lcssa = 0;
 var $bp$21$i = 0, $bp$22$i = 0, $bp$3$i = 0, $bp$4$i = 0, $bp$4$i$lcssa = 0, $bp$5$i = 0, $bp$6$i = 0, $bp$7$i = 0, $bp$8$i = 0, $bp$8$i$lcssa483 = 0, $bp$8$i$lcssa483$lcssa = 0, $bp$9$i = 0, $error$0$ph71$be$i$i$i = 0, $error$0$ph71$lcssa84$i$i$i = 0, $error$0$ph71119$i$i$i = 0, $error$11$i$i$i = 0, $error$5$i$i = 0, $exitcond$i$i = 0, $exitcond$i4$i = 0, $forward$060$i$i = 0;
 var $i$0123$i$i$i = 0, $i$02$i$i$i = 0, $i$02$i13$i$i = 0, $i$3$ph$be$i$i$i = 0, $i$3$ph118$i$i$i = 0, $i$392$i$i$i = 0, $i$392$i$i$i$lcssa491 = 0, $i$4111$i$i$i = 0, $i$6103$i$i$i = 0, $i$895$i$i$i = 0, $n$0110$i$i$i = 0, $n$02$i$i = 0, $n$1102$i$i$i = 0, $n$294$i$i$i = 0, $or$cond$i$i$i = 0, $p$0$lcssa$i$i = 0, $p$01$i$i = 0, $pos$0$ph$i = 0, $pos$080$i = 0, $pos$080$i$lcssa477 = 0;
 var $pos$080$i478 = 0, $pos$1$i = 0, $pos$2$i = 0, $pos$4$ph$i = 0, $pos$4$ph$i$lcssa544 = 0, $pos$4$ph$i549 = 0, $pos$4$ph$i550 = 0, $pos$5$i = 0, $pos$6$i = 0, $result$0$lcssa$i$i$i = 0, $result$0$lcssa$i16$i$i = 0, $result$01$i$i$i = 0, $result$01$i14$i$i = 0, $tree_cl$i$i$i = 0, $tree_d$i$i = 0, $tree_ll$i$i = 0, $treepos$0$i$i$i = 0, $treepos$0$i$i$i$i = 0, $treepos$0$i8$i$i = 0, $v$sroa$0$0 = 0;
 var $v$sroa$0$1 = 0, $v$sroa$0$1$lcssa475 = 0, $v$sroa$0$1476 = 0, $v$sroa$0$2 = 0, $v$sroa$0$3 = 0, $v$sroa$0$4$ph = 0, $v$sroa$0$4$ph$lcssa531 = 0, $v$sroa$0$4$ph$lcssa532 = 0, $v$sroa$0$4$ph537 = 0, $v$sroa$0$4$ph538 = 0, $v$sroa$0$5 = 0, $v$sroa$0$6 = 0, $v$sroa$0$6$lcssa = 0, $v$sroa$0$7 = 0, $v$sroa$0$8 = 0, $v$sroa$13$0 = 0, $v$sroa$13$1 = 0, $v$sroa$13$1$lcssa473 = 0, $v$sroa$13$1474 = 0, $v$sroa$13$2 = 0;
 var $v$sroa$13$3$ph = 0, $v$sroa$13$3$ph$lcssa523 = 0, $v$sroa$13$3$ph$lcssa524 = 0, $v$sroa$13$3$ph529 = 0, $v$sroa$13$3$ph530 = 0, $v$sroa$13$4 = 0, $v$sroa$13$5 = 0, $v$sroa$18$0 = 0, $v$sroa$18$1 = 0, $v$sroa$18$1$lcssa471 = 0, $v$sroa$18$1472 = 0, $v$sroa$18$2 = 0, $v$sroa$18$3 = 0, $v$sroa$18$4$ph = 0, $v$sroa$18$4$ph$lcssa515 = 0, $v$sroa$18$4$ph$lcssa516 = 0, $v$sroa$18$4$ph521 = 0, $v$sroa$18$4$ph522 = 0, $v$sroa$18$5 = 0, $v$sroa$18$6 = 0;
 var $v$sroa$18$6$lcssa = 0, $v$sroa$18$7 = 0, $value$0$i$i$i = 0, $value$0$in$i$i$i = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tree_cl$i$i$i = sp + 40|0;
 $tree_ll$i$i = sp + 20|0;
 $tree_d$i$i = sp;
 $0 = HEAP32[$out>>2]|0; //@line 1266 "libraries/lodepng/lodepng.cpp"
 $1 = HEAP32[$outsize>>2]|0; //@line 1266 "libraries/lodepng/lodepng.cpp"
 $2 = $insize << 3; //@line 1245 "libraries/lodepng/lodepng.cpp"
 $3 = ((($tree_ll$i$i)) + 4|0); //@line 533 "libraries/lodepng/lodepng.cpp"
 $4 = ((($tree_ll$i$i)) + 8|0); //@line 534 "libraries/lodepng/lodepng.cpp"
 $5 = ((($tree_d$i$i)) + 4|0); //@line 533 "libraries/lodepng/lodepng.cpp"
 $6 = ((($tree_d$i$i)) + 8|0); //@line 534 "libraries/lodepng/lodepng.cpp"
 $7 = ((($tree_ll$i$i)) + 16|0); //@line 939 "libraries/lodepng/lodepng.cpp"
 $8 = ((($tree_d$i$i)) + 16|0); //@line 939 "libraries/lodepng/lodepng.cpp"
 $9 = ((($tree_cl$i$i$i)) + 4|0); //@line 533 "libraries/lodepng/lodepng.cpp"
 $10 = ((($tree_cl$i$i$i)) + 8|0); //@line 534 "libraries/lodepng/lodepng.cpp"
 $11 = ((($tree_cl$i$i$i)) + 16|0); //@line 662 "libraries/lodepng/lodepng.cpp"
 $12 = ((($tree_cl$i$i$i)) + 12|0); //@line 663 "libraries/lodepng/lodepng.cpp"
 $13 = ((($tree_ll$i$i)) + 12|0); //@line 663 "libraries/lodepng/lodepng.cpp"
 $14 = ((($tree_d$i$i)) + 12|0); //@line 663 "libraries/lodepng/lodepng.cpp"
 $BFINAL$0$ph$i = 0;$bp$0$ph$i = 0;$pos$0$ph$i = 0;$v$sroa$0$0 = $0;$v$sroa$13$0 = $1;$v$sroa$18$0 = $1;
 L1: while(1) {
  $15 = ($BFINAL$0$ph$i|0)==(0); //@line 1242 "libraries/lodepng/lodepng.cpp"
  if ($15) {
   $bp$079$i = $bp$0$ph$i;$pos$080$i = $pos$0$ph$i;$v$sroa$0$1 = $v$sroa$0$0;$v$sroa$13$1 = $v$sroa$13$0;$v$sroa$18$1 = $v$sroa$18$0;
  } else {
   $$2$i = 0;$v$sroa$0$8 = $v$sroa$0$0;$v$sroa$13$5 = $v$sroa$13$0;
   label = 110;
   break;
  }
  L3: while(1) {
   $16 = (($bp$079$i) + 2)|0; //@line 1245 "libraries/lodepng/lodepng.cpp"
   $17 = ($16>>>0)<($2>>>0); //@line 1245 "libraries/lodepng/lodepng.cpp"
   if (!($17)) {
    $$2$i = 52;$v$sroa$0$8 = $v$sroa$0$1;$v$sroa$13$5 = $v$sroa$13$1;
    label = 110;
    break L1;
   }
   $18 = $bp$079$i >>> 3; //@line 448 "libraries/lodepng/lodepng.cpp"
   $19 = (($in) + ($18)|0); //@line 448 "libraries/lodepng/lodepng.cpp"
   $20 = HEAP8[$19>>0]|0; //@line 448 "libraries/lodepng/lodepng.cpp"
   $21 = $20&255; //@line 448 "libraries/lodepng/lodepng.cpp"
   $22 = $bp$079$i & 7; //@line 448 "libraries/lodepng/lodepng.cpp"
   $23 = $21 >>> $22; //@line 448 "libraries/lodepng/lodepng.cpp"
   $24 = $23 & 1; //@line 448 "libraries/lodepng/lodepng.cpp"
   $25 = (($bp$079$i) + 1)|0; //@line 449 "libraries/lodepng/lodepng.cpp"
   $26 = $25 >>> 3; //@line 448 "libraries/lodepng/lodepng.cpp"
   $27 = (($in) + ($26)|0); //@line 448 "libraries/lodepng/lodepng.cpp"
   $28 = HEAP8[$27>>0]|0; //@line 448 "libraries/lodepng/lodepng.cpp"
   $29 = $28&255; //@line 448 "libraries/lodepng/lodepng.cpp"
   $30 = $25 & 7; //@line 448 "libraries/lodepng/lodepng.cpp"
   $31 = $29 >>> $30; //@line 448 "libraries/lodepng/lodepng.cpp"
   $32 = $31 & 1; //@line 448 "libraries/lodepng/lodepng.cpp"
   $33 = $16 >>> 3; //@line 448 "libraries/lodepng/lodepng.cpp"
   $34 = (($in) + ($33)|0); //@line 448 "libraries/lodepng/lodepng.cpp"
   $35 = HEAP8[$34>>0]|0; //@line 448 "libraries/lodepng/lodepng.cpp"
   $36 = $35&255; //@line 448 "libraries/lodepng/lodepng.cpp"
   $37 = $16 & 7; //@line 448 "libraries/lodepng/lodepng.cpp"
   $38 = $36 >>> $37; //@line 448 "libraries/lodepng/lodepng.cpp"
   $39 = (($bp$079$i) + 3)|0; //@line 449 "libraries/lodepng/lodepng.cpp"
   $40 = $38 << 1; //@line 1248 "libraries/lodepng/lodepng.cpp"
   $41 = $40 & 2; //@line 1248 "libraries/lodepng/lodepng.cpp"
   $42 = $41 | $32; //@line 1248 "libraries/lodepng/lodepng.cpp"
   switch ($42|0) {
   case 3:  {
    $$2$i = 20;$v$sroa$0$8 = $v$sroa$0$1;$v$sroa$13$5 = $v$sroa$13$1;
    label = 110;
    break L1;
    break;
   }
   case 0:  {
    break;
   }
   default: {
    $$lcssa480 = $24;$$lcssa481 = $39;$$lcssa482 = $42;$bp$079$i$lcssa479 = $bp$079$i;$pos$080$i$lcssa477 = $pos$080$i;$v$sroa$0$1$lcssa475 = $v$sroa$0$1;$v$sroa$13$1$lcssa473 = $v$sroa$13$1;$v$sroa$18$1$lcssa471 = $v$sroa$18$1;
    break L3;
   }
   }
   $43 = $39 & 7; //@line 1208 "libraries/lodepng/lodepng.cpp"
   $44 = ($43|0)==(0); //@line 1208 "libraries/lodepng/lodepng.cpp"
   if ($44) {
    $bp$1$i = $39;
   } else {
    $46 = $39;
    while(1) {
     $45 = (($46) + 1)|0; //@line 1208 "libraries/lodepng/lodepng.cpp"
     $47 = $45 & 7; //@line 1208 "libraries/lodepng/lodepng.cpp"
     $48 = ($47|0)==(0); //@line 1208 "libraries/lodepng/lodepng.cpp"
     if ($48) {
      $bp$1$i = $45;
      break;
     } else {
      $46 = $45;
     }
    }
   }
   $49 = $bp$1$i >>> 3; //@line 1209 "libraries/lodepng/lodepng.cpp"
   $50 = (($49) + 4)|0; //@line 1212 "libraries/lodepng/lodepng.cpp"
   $51 = ($50>>>0)<($insize>>>0); //@line 1212 "libraries/lodepng/lodepng.cpp"
   if (!($51)) {
    $$2$i = 52;$v$sroa$0$8 = $v$sroa$0$1;$v$sroa$13$5 = $v$sroa$13$1;
    label = 110;
    break L1;
   }
   $52 = (($in) + ($49)|0); //@line 1213 "libraries/lodepng/lodepng.cpp"
   $53 = HEAP8[$52>>0]|0; //@line 1213 "libraries/lodepng/lodepng.cpp"
   $54 = $53&255; //@line 1213 "libraries/lodepng/lodepng.cpp"
   $55 = (($49) + 1)|0; //@line 1213 "libraries/lodepng/lodepng.cpp"
   $56 = (($in) + ($55)|0); //@line 1213 "libraries/lodepng/lodepng.cpp"
   $57 = HEAP8[$56>>0]|0; //@line 1213 "libraries/lodepng/lodepng.cpp"
   $58 = $57&255; //@line 1213 "libraries/lodepng/lodepng.cpp"
   $59 = $58 << 8; //@line 1213 "libraries/lodepng/lodepng.cpp"
   $60 = $59 | $54; //@line 1213 "libraries/lodepng/lodepng.cpp"
   $61 = (($49) + 2)|0; //@line 1213 "libraries/lodepng/lodepng.cpp"
   $62 = (($in) + ($61)|0); //@line 1214 "libraries/lodepng/lodepng.cpp"
   $63 = HEAP8[$62>>0]|0; //@line 1214 "libraries/lodepng/lodepng.cpp"
   $64 = $63&255; //@line 1214 "libraries/lodepng/lodepng.cpp"
   $65 = (($49) + 3)|0; //@line 1214 "libraries/lodepng/lodepng.cpp"
   $66 = (($in) + ($65)|0); //@line 1214 "libraries/lodepng/lodepng.cpp"
   $67 = HEAP8[$66>>0]|0; //@line 1214 "libraries/lodepng/lodepng.cpp"
   $68 = $67&255; //@line 1214 "libraries/lodepng/lodepng.cpp"
   $69 = $68 << 8; //@line 1214 "libraries/lodepng/lodepng.cpp"
   $70 = $69 | $64; //@line 1214 "libraries/lodepng/lodepng.cpp"
   $71 = (($70) + ($60))|0; //@line 1217 "libraries/lodepng/lodepng.cpp"
   $72 = ($71|0)==(65535); //@line 1217 "libraries/lodepng/lodepng.cpp"
   if (!($72)) {
    $$2$i = 21;$v$sroa$0$8 = $v$sroa$0$1;$v$sroa$13$5 = $v$sroa$13$1;
    label = 110;
    break L1;
   }
   $73 = (($60) + ($pos$080$i))|0; //@line 1219 "libraries/lodepng/lodepng.cpp"
   $74 = ($v$sroa$18$1>>>0)<($73>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($74) {
    $75 = $v$sroa$18$1 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $76 = ($75>>>0)<($73>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $77 = ($73*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $78 = $77 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $79 = $76 ? $73 : $78; //@line 227 "libraries/lodepng/lodepng.cpp"
    $80 = $v$sroa$0$1; //@line 228 "libraries/lodepng/lodepng.cpp"
    $81 = (_realloc($80,$79)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $82 = ($81|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if ($82) {
     $$2$i = 83;$v$sroa$0$8 = $v$sroa$0$1;$v$sroa$13$5 = $v$sroa$13$1;
     label = 110;
     break L1;
    }
    $83 = $81; //@line 232 "libraries/lodepng/lodepng.cpp"
    $v$sroa$0$2 = $83;$v$sroa$18$2 = $79;
   } else {
    $v$sroa$0$2 = $v$sroa$0$1;$v$sroa$18$2 = $v$sroa$18$1;
   }
   $84 = (($60) + ($50))|0; //@line 1222 "libraries/lodepng/lodepng.cpp"
   $85 = ($84>>>0)>($insize>>>0); //@line 1222 "libraries/lodepng/lodepng.cpp"
   if ($85) {
    $$2$i = 23;$v$sroa$0$8 = $v$sroa$0$2;$v$sroa$13$5 = $73;
    label = 110;
    break L1;
   }
   $86 = ($60|0)==(0); //@line 1223 "libraries/lodepng/lodepng.cpp"
   if ($86) {
    $p$0$lcssa$i$i = $50;$pos$2$i = $pos$080$i;
   } else {
    $87 = $v$sroa$0$2; //@line 1223 "libraries/lodepng/lodepng.cpp"
    $n$02$i$i = 0;$p$01$i$i = $50;$pos$1$i = $pos$080$i;
    while(1) {
     $88 = (($p$01$i$i) + 1)|0; //@line 1223 "libraries/lodepng/lodepng.cpp"
     $89 = (($in) + ($p$01$i$i)|0); //@line 1223 "libraries/lodepng/lodepng.cpp"
     $90 = HEAP8[$89>>0]|0; //@line 1223 "libraries/lodepng/lodepng.cpp"
     $91 = (($pos$1$i) + 1)|0; //@line 1223 "libraries/lodepng/lodepng.cpp"
     $92 = (($87) + ($pos$1$i)|0); //@line 1223 "libraries/lodepng/lodepng.cpp"
     HEAP8[$92>>0] = $90; //@line 1223 "libraries/lodepng/lodepng.cpp"
     $93 = (($n$02$i$i) + 1)|0; //@line 1223 "libraries/lodepng/lodepng.cpp"
     $exitcond$i4$i = ($93|0)==($60|0); //@line 1223 "libraries/lodepng/lodepng.cpp"
     if ($exitcond$i4$i) {
      $p$0$lcssa$i$i = $84;$pos$2$i = $73;
      break;
     } else {
      $n$02$i$i = $93;$p$01$i$i = $88;$pos$1$i = $91;
     }
    }
   }
   $94 = $p$0$lcssa$i$i << 3; //@line 1225 "libraries/lodepng/lodepng.cpp"
   $95 = ($24|0)==(0); //@line 1242 "libraries/lodepng/lodepng.cpp"
   if ($95) {
    $bp$079$i = $94;$pos$080$i = $pos$2$i;$v$sroa$0$1 = $v$sroa$0$2;$v$sroa$13$1 = $73;$v$sroa$18$1 = $v$sroa$18$2;
   } else {
    $$2$i = 0;$v$sroa$0$8 = $v$sroa$0$2;$v$sroa$13$5 = $73;
    label = 110;
    break L1;
   }
  }
  HEAP32[$tree_ll$i$i>>2] = 0; //@line 532 "libraries/lodepng/lodepng.cpp"
  HEAP32[$3>>2] = 0; //@line 533 "libraries/lodepng/lodepng.cpp"
  HEAP32[$4>>2] = 0; //@line 534 "libraries/lodepng/lodepng.cpp"
  HEAP32[$tree_d$i$i>>2] = 0; //@line 532 "libraries/lodepng/lodepng.cpp"
  HEAP32[$5>>2] = 0; //@line 533 "libraries/lodepng/lodepng.cpp"
  HEAP32[$6>>2] = 0; //@line 534 "libraries/lodepng/lodepng.cpp"
  L24: do {
   switch ($$lcssa482|0) {
   case 1:  {
    __ZL23generateFixedLitLenTreeP11HuffmanTree($tree_ll$i$i); //@line 957 "libraries/lodepng/lodepng.cpp"
    __ZL25generateFixedDistanceTreeP11HuffmanTree($tree_d$i$i); //@line 958 "libraries/lodepng/lodepng.cpp"
    $bp$14$i = $$lcssa481;
    label = 74;
    break;
   }
   case 2:  {
    $96 = (($bp$079$i$lcssa479) + 17)|0; //@line 977 "libraries/lodepng/lodepng.cpp"
    $97 = ($96>>>0)>($2>>>0); //@line 977 "libraries/lodepng/lodepng.cpp"
    if ($97) {
     $$0$i$ph$i$i = 49;$bp$11$i = $$lcssa481;
     label = 71;
    } else {
     $98 = $$lcssa481 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
     $99 = (($in) + ($98)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
     $100 = HEAP8[$99>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
     $101 = $100&255; //@line 458 "libraries/lodepng/lodepng.cpp"
     $102 = $$lcssa481 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
     $103 = $101 >>> $102; //@line 458 "libraries/lodepng/lodepng.cpp"
     $104 = $103 & 1; //@line 458 "libraries/lodepng/lodepng.cpp"
     $105 = (($bp$079$i$lcssa479) + 4)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
     $106 = $105 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
     $107 = (($in) + ($106)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
     $108 = HEAP8[$107>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
     $109 = $108&255; //@line 458 "libraries/lodepng/lodepng.cpp"
     $110 = $105 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
     $111 = $109 >>> $110; //@line 458 "libraries/lodepng/lodepng.cpp"
     $112 = $111 << 1; //@line 458 "libraries/lodepng/lodepng.cpp"
     $113 = $112 & 2; //@line 458 "libraries/lodepng/lodepng.cpp"
     $114 = $113 | $104; //@line 458 "libraries/lodepng/lodepng.cpp"
     $115 = (($bp$079$i$lcssa479) + 5)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
     $116 = $115 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
     $117 = (($in) + ($116)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
     $118 = HEAP8[$117>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
     $119 = $118&255; //@line 458 "libraries/lodepng/lodepng.cpp"
     $120 = $115 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
     $121 = $119 >>> $120; //@line 458 "libraries/lodepng/lodepng.cpp"
     $122 = $121 << 2; //@line 458 "libraries/lodepng/lodepng.cpp"
     $123 = $122 & 4; //@line 458 "libraries/lodepng/lodepng.cpp"
     $124 = $114 | $123; //@line 458 "libraries/lodepng/lodepng.cpp"
     $125 = (($bp$079$i$lcssa479) + 6)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
     $126 = $125 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
     $127 = (($in) + ($126)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
     $128 = HEAP8[$127>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
     $129 = $128&255; //@line 458 "libraries/lodepng/lodepng.cpp"
     $130 = $125 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
     $131 = $129 >>> $130; //@line 458 "libraries/lodepng/lodepng.cpp"
     $132 = $131 << 3; //@line 458 "libraries/lodepng/lodepng.cpp"
     $133 = $132 & 8; //@line 458 "libraries/lodepng/lodepng.cpp"
     $134 = $124 | $133; //@line 458 "libraries/lodepng/lodepng.cpp"
     $135 = (($bp$079$i$lcssa479) + 7)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
     $136 = $135 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
     $137 = (($in) + ($136)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
     $138 = HEAP8[$137>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
     $139 = $138&255; //@line 458 "libraries/lodepng/lodepng.cpp"
     $140 = $135 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
     $141 = $139 >>> $140; //@line 458 "libraries/lodepng/lodepng.cpp"
     $142 = $141 << 4; //@line 458 "libraries/lodepng/lodepng.cpp"
     $143 = $142 & 16; //@line 458 "libraries/lodepng/lodepng.cpp"
     $144 = $134 | $143; //@line 458 "libraries/lodepng/lodepng.cpp"
     $145 = (($bp$079$i$lcssa479) + 8)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
     $146 = (($144) + 257)|0; //@line 980 "libraries/lodepng/lodepng.cpp"
     $147 = $145 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
     $148 = (($in) + ($147)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
     $149 = HEAP8[$148>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
     $150 = $149&255; //@line 458 "libraries/lodepng/lodepng.cpp"
     $151 = $145 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
     $152 = $150 >>> $151; //@line 458 "libraries/lodepng/lodepng.cpp"
     $153 = $152 & 1; //@line 458 "libraries/lodepng/lodepng.cpp"
     $154 = (($bp$079$i$lcssa479) + 9)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
     $155 = $154 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
     $156 = (($in) + ($155)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
     $157 = HEAP8[$156>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
     $158 = $157&255; //@line 458 "libraries/lodepng/lodepng.cpp"
     $159 = $154 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
     $160 = $158 >>> $159; //@line 458 "libraries/lodepng/lodepng.cpp"
     $161 = $160 << 1; //@line 458 "libraries/lodepng/lodepng.cpp"
     $162 = $161 & 2; //@line 458 "libraries/lodepng/lodepng.cpp"
     $163 = $162 | $153; //@line 458 "libraries/lodepng/lodepng.cpp"
     $164 = (($bp$079$i$lcssa479) + 10)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
     $165 = $164 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
     $166 = (($in) + ($165)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
     $167 = HEAP8[$166>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
     $168 = $167&255; //@line 458 "libraries/lodepng/lodepng.cpp"
     $169 = $164 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
     $170 = $168 >>> $169; //@line 458 "libraries/lodepng/lodepng.cpp"
     $171 = $170 << 2; //@line 458 "libraries/lodepng/lodepng.cpp"
     $172 = $171 & 4; //@line 458 "libraries/lodepng/lodepng.cpp"
     $173 = $163 | $172; //@line 458 "libraries/lodepng/lodepng.cpp"
     $174 = (($bp$079$i$lcssa479) + 11)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
     $175 = $174 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
     $176 = (($in) + ($175)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
     $177 = HEAP8[$176>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
     $178 = $177&255; //@line 458 "libraries/lodepng/lodepng.cpp"
     $179 = $174 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
     $180 = $178 >>> $179; //@line 458 "libraries/lodepng/lodepng.cpp"
     $181 = $180 << 3; //@line 458 "libraries/lodepng/lodepng.cpp"
     $182 = $181 & 8; //@line 458 "libraries/lodepng/lodepng.cpp"
     $183 = $173 | $182; //@line 458 "libraries/lodepng/lodepng.cpp"
     $184 = (($bp$079$i$lcssa479) + 12)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
     $185 = $184 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
     $186 = (($in) + ($185)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
     $187 = HEAP8[$186>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
     $188 = $187&255; //@line 458 "libraries/lodepng/lodepng.cpp"
     $189 = $184 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
     $190 = $188 >>> $189; //@line 458 "libraries/lodepng/lodepng.cpp"
     $191 = $190 << 4; //@line 458 "libraries/lodepng/lodepng.cpp"
     $192 = $191 & 16; //@line 458 "libraries/lodepng/lodepng.cpp"
     $193 = $183 | $192; //@line 458 "libraries/lodepng/lodepng.cpp"
     $194 = (($bp$079$i$lcssa479) + 13)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
     $195 = $194 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
     $196 = (($in) + ($195)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
     $197 = HEAP8[$196>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
     $198 = $197&255; //@line 458 "libraries/lodepng/lodepng.cpp"
     $199 = $194 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
     $200 = $198 >>> $199; //@line 458 "libraries/lodepng/lodepng.cpp"
     $201 = $200 & 1; //@line 458 "libraries/lodepng/lodepng.cpp"
     $202 = (($bp$079$i$lcssa479) + 14)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
     $203 = $202 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
     $204 = (($in) + ($203)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
     $205 = HEAP8[$204>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
     $206 = $205&255; //@line 458 "libraries/lodepng/lodepng.cpp"
     $207 = $202 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
     $208 = $206 >>> $207; //@line 458 "libraries/lodepng/lodepng.cpp"
     $209 = $208 << 1; //@line 458 "libraries/lodepng/lodepng.cpp"
     $210 = $209 & 2; //@line 458 "libraries/lodepng/lodepng.cpp"
     $211 = $210 | $201; //@line 458 "libraries/lodepng/lodepng.cpp"
     $212 = (($bp$079$i$lcssa479) + 15)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
     $213 = $212 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
     $214 = (($in) + ($213)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
     $215 = HEAP8[$214>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
     $216 = $215&255; //@line 458 "libraries/lodepng/lodepng.cpp"
     $217 = $212 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
     $218 = $216 >>> $217; //@line 458 "libraries/lodepng/lodepng.cpp"
     $219 = $218 << 2; //@line 458 "libraries/lodepng/lodepng.cpp"
     $220 = $219 & 4; //@line 458 "libraries/lodepng/lodepng.cpp"
     $221 = $211 | $220; //@line 458 "libraries/lodepng/lodepng.cpp"
     $222 = (($bp$079$i$lcssa479) + 16)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
     $223 = $222 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
     $224 = (($in) + ($223)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
     $225 = HEAP8[$224>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
     $226 = $225&255; //@line 458 "libraries/lodepng/lodepng.cpp"
     $227 = $222 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
     $228 = $226 >>> $227; //@line 458 "libraries/lodepng/lodepng.cpp"
     $229 = $228 << 3; //@line 458 "libraries/lodepng/lodepng.cpp"
     $230 = $229 & 8; //@line 458 "libraries/lodepng/lodepng.cpp"
     $231 = $221 | $230; //@line 458 "libraries/lodepng/lodepng.cpp"
     $232 = (($231) + 4)|0; //@line 984 "libraries/lodepng/lodepng.cpp"
     $233 = ($232*3)|0; //@line 986 "libraries/lodepng/lodepng.cpp"
     $234 = (($233) + ($96))|0; //@line 986 "libraries/lodepng/lodepng.cpp"
     $235 = ($234>>>0)>($2>>>0); //@line 986 "libraries/lodepng/lodepng.cpp"
     if ($235) {
      $$0$i$ph$i$i = 50;$bp$11$i = $96;
      label = 71;
     } else {
      HEAP32[$tree_cl$i$i$i>>2] = 0; //@line 532 "libraries/lodepng/lodepng.cpp"
      HEAP32[$9>>2] = 0; //@line 533 "libraries/lodepng/lodepng.cpp"
      HEAP32[$10>>2] = 0; //@line 534 "libraries/lodepng/lodepng.cpp"
      $236 = (_malloc(76)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
      $237 = ($236|0)==(0|0); //@line 995 "libraries/lodepng/lodepng.cpp"
      L30: do {
       if ($237) {
        $bitlen_d$0$i$i$i = 0;$bitlen_ll$0$i$i$i = 0;$bp$12$i = $96;$error$11$i$i$i = 83;
       } else {
        $$pre$i15$i$i$i = $96;$bp$3$i = $96;$i$0123$i$i$i = 0;
        while(1) {
         $238 = ($i$0123$i$i$i>>>0)<($232>>>0); //@line 999 "libraries/lodepng/lodepng.cpp"
         $239 = (66980 + ($i$0123$i$i$i<<2)|0); //@line 999 "libraries/lodepng/lodepng.cpp"
         $240 = HEAP32[$239>>2]|0; //@line 999 "libraries/lodepng/lodepng.cpp"
         if ($238) {
          $241 = $$pre$i15$i$i$i >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
          $242 = (($in) + ($241)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
          $243 = HEAP8[$242>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
          $244 = $243&255; //@line 458 "libraries/lodepng/lodepng.cpp"
          $245 = $$pre$i15$i$i$i & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
          $246 = $244 >>> $245; //@line 458 "libraries/lodepng/lodepng.cpp"
          $247 = $246 & 1; //@line 458 "libraries/lodepng/lodepng.cpp"
          $248 = (($$pre$i15$i$i$i) + 1)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
          $249 = $248 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
          $250 = (($in) + ($249)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
          $251 = HEAP8[$250>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
          $252 = $251&255; //@line 458 "libraries/lodepng/lodepng.cpp"
          $253 = $248 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
          $254 = $252 >>> $253; //@line 458 "libraries/lodepng/lodepng.cpp"
          $255 = $254 << 1; //@line 458 "libraries/lodepng/lodepng.cpp"
          $256 = $255 & 2; //@line 458 "libraries/lodepng/lodepng.cpp"
          $257 = $256 | $247; //@line 458 "libraries/lodepng/lodepng.cpp"
          $258 = (($$pre$i15$i$i$i) + 2)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
          $259 = $258 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
          $260 = (($in) + ($259)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
          $261 = HEAP8[$260>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
          $262 = $261&255; //@line 458 "libraries/lodepng/lodepng.cpp"
          $263 = $258 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
          $264 = $262 >>> $263; //@line 458 "libraries/lodepng/lodepng.cpp"
          $265 = $264 << 2; //@line 458 "libraries/lodepng/lodepng.cpp"
          $266 = $265 & 4; //@line 458 "libraries/lodepng/lodepng.cpp"
          $267 = $257 | $266; //@line 458 "libraries/lodepng/lodepng.cpp"
          $268 = (($$pre$i15$i$i$i) + 3)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
          $$pre$i15149$i$i$i = $268;$$sink$i$i = $267;$bp$4$i = $268;
         } else {
          $$pre$i15149$i$i$i = $$pre$i15$i$i$i;$$sink$i$i = 0;$bp$4$i = $bp$3$i;
         }
         $269 = (($236) + ($240<<2)|0); //@line 1000 "libraries/lodepng/lodepng.cpp"
         HEAP32[$269>>2] = $$sink$i$i; //@line 1000 "libraries/lodepng/lodepng.cpp"
         $270 = (($i$0123$i$i$i) + 1)|0; //@line 997 "libraries/lodepng/lodepng.cpp"
         $271 = ($270|0)==(19); //@line 997 "libraries/lodepng/lodepng.cpp"
         if ($271) {
          $bp$4$i$lcssa = $bp$4$i;
          break;
         } else {
          $$pre$i15$i$i$i = $$pre$i15149$i$i$i;$bp$3$i = $bp$4$i;$i$0123$i$i$i = $270;
         }
        }
        $272 = (_malloc(76)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
        HEAP32[$10>>2] = $272; //@line 659 "libraries/lodepng/lodepng.cpp"
        $273 = ($272|0)==(0|0); //@line 660 "libraries/lodepng/lodepng.cpp"
        if ($273) {
         $bitlen_d$0$i$i$i = 0;$bitlen_ll$0$i$i$i = 0;$bp$12$i = $bp$4$i$lcssa;$error$11$i$i$i = 83;
        } else {
         dest=$272; src=$236; stop=dest+76|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 661 "libraries/lodepng/lodepng.cpp"
         HEAP32[$11>>2] = 19; //@line 662 "libraries/lodepng/lodepng.cpp"
         HEAP32[$12>>2] = 7; //@line 663 "libraries/lodepng/lodepng.cpp"
         $274 = (__ZL28HuffmanTree_makeFromLengths2P11HuffmanTree($tree_cl$i$i$i)|0); //@line 664 "libraries/lodepng/lodepng.cpp"
         $275 = ($274|0)==(0); //@line 1004 "libraries/lodepng/lodepng.cpp"
         if ($275) {
          $276 = (_malloc(1152)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
          $277 = (_malloc(128)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
          $278 = ($276|0)!=(0|0); //@line 1009 "libraries/lodepng/lodepng.cpp"
          $279 = ($277|0)!=(0|0); //@line 1009 "libraries/lodepng/lodepng.cpp"
          $or$cond$i$i$i = $278 & $279; //@line 1009 "libraries/lodepng/lodepng.cpp"
          if ($or$cond$i$i$i) {
           _memset(($276|0),0,1152)|0; //@line 1010 "libraries/lodepng/lodepng.cpp"
           dest=$277; stop=dest+128|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 1011 "libraries/lodepng/lodepng.cpp"
           $280 = (($144) + 258)|0; //@line 982 "libraries/lodepng/lodepng.cpp"
           $281 = (($193) + ($280))|0; //@line 1015 "libraries/lodepng/lodepng.cpp"
           $282 = ($281|0)==(0); //@line 1015 "libraries/lodepng/lodepng.cpp"
           if ($282) {
            $bitlen_d$0$i$i$i = $277;$bitlen_ll$0$i$i$i = $276;$bp$12$i = $bp$4$i$lcssa;$error$11$i$i$i = 64;
           } else {
            $283 = HEAP32[$tree_cl$i$i$i>>2]|0; //@line 937 "libraries/lodepng/lodepng.cpp"
            $284 = HEAP32[$11>>2]|0; //@line 939 "libraries/lodepng/lodepng.cpp"
            $285 = (-258 - ($144))|0; //@line 1035 "libraries/lodepng/lodepng.cpp"
            $bp$6$i = $bp$4$i$lcssa;$error$0$ph71119$i$i$i = 0;$i$3$ph118$i$i$i = 0;
            L41: while(1) {
             $bp$7$i = $bp$6$i;$i$392$i$i$i = $i$3$ph118$i$i$i;
             while(1) {
              $bp$8$i = $bp$7$i;$treepos$0$i$i$i$i = 0;
              while(1) {
               $287 = ($bp$8$i>>>0)<($2>>>0); //@line 932 "libraries/lodepng/lodepng.cpp"
               if (!($287)) {
                $bp$9$i = $bp$8$i;
                label = 64;
                break L41;
               }
               $288 = $treepos$0$i$i$i$i << 1; //@line 937 "libraries/lodepng/lodepng.cpp"
               $289 = $bp$8$i >>> 3; //@line 937 "libraries/lodepng/lodepng.cpp"
               $290 = (($in) + ($289)|0); //@line 937 "libraries/lodepng/lodepng.cpp"
               $291 = HEAP8[$290>>0]|0; //@line 937 "libraries/lodepng/lodepng.cpp"
               $292 = $291&255; //@line 937 "libraries/lodepng/lodepng.cpp"
               $293 = $bp$8$i & 7; //@line 937 "libraries/lodepng/lodepng.cpp"
               $294 = $292 >>> $293; //@line 937 "libraries/lodepng/lodepng.cpp"
               $295 = $294 & 1; //@line 937 "libraries/lodepng/lodepng.cpp"
               $296 = $295 | $288; //@line 937 "libraries/lodepng/lodepng.cpp"
               $297 = (($283) + ($296<<2)|0); //@line 937 "libraries/lodepng/lodepng.cpp"
               $298 = HEAP32[$297>>2]|0; //@line 937 "libraries/lodepng/lodepng.cpp"
               $299 = (($bp$8$i) + 1)|0; //@line 1031 "libraries/lodepng/lodepng.cpp"
               $300 = ($298>>>0)<($284>>>0); //@line 939 "libraries/lodepng/lodepng.cpp"
               if ($300) {
                $$lcssa485 = $298;$$lcssa487 = $299;$bp$8$i$lcssa483 = $bp$8$i;
                break;
               }
               $301 = (($298) - ($284))|0; //@line 940 "libraries/lodepng/lodepng.cpp"
               $302 = ($301>>>0)<($284>>>0); //@line 942 "libraries/lodepng/lodepng.cpp"
               if ($302) {
                $bp$8$i = $299;$treepos$0$i$i$i$i = $301;
               } else {
                $bp$9$i = $299;
                label = 64;
                break L41;
               }
              }
              $303 = ($$lcssa485>>>0)<(16); //@line 1018 "libraries/lodepng/lodepng.cpp"
              if (!($303)) {
               $$lcssa485$lcssa = $$lcssa485;$$lcssa487$lcssa = $$lcssa487;$bp$8$i$lcssa483$lcssa = $bp$8$i$lcssa483;$i$392$i$i$i$lcssa491 = $i$392$i$i$i;
               break;
              }
              $304 = ($i$392$i$i$i>>>0)<($146>>>0); //@line 1020 "libraries/lodepng/lodepng.cpp"
              if ($304) {
               $305 = (($276) + ($i$392$i$i$i<<2)|0); //@line 1020 "libraries/lodepng/lodepng.cpp"
               HEAP32[$305>>2] = $$lcssa485; //@line 1020 "libraries/lodepng/lodepng.cpp"
              } else {
               $306 = (($i$392$i$i$i) - ($146))|0; //@line 1021 "libraries/lodepng/lodepng.cpp"
               $307 = (($277) + ($306<<2)|0); //@line 1021 "libraries/lodepng/lodepng.cpp"
               HEAP32[$307>>2] = $$lcssa485; //@line 1021 "libraries/lodepng/lodepng.cpp"
              }
              $308 = (($i$392$i$i$i) + 1)|0; //@line 1022 "libraries/lodepng/lodepng.cpp"
              $309 = ($308>>>0)<($281>>>0); //@line 1015 "libraries/lodepng/lodepng.cpp"
              if ($309) {
               $bp$7$i = $$lcssa487;$i$392$i$i$i = $308;
              } else {
               $bp$10$i = $$lcssa487;$error$0$ph71$lcssa84$i$i$i = $error$0$ph71119$i$i$i;
               break L41;
              }
             }
             L56: do {
              switch ($$lcssa485$lcssa|0) {
              case -1:  {
               $bp$9$i = $$lcssa487$lcssa;
               label = 64;
               break L41;
               break;
              }
              case 16:  {
               $310 = ($i$392$i$i$i$lcssa491|0)==(0); //@line 1029 "libraries/lodepng/lodepng.cpp"
               if ($310) {
                $bitlen_d$0$i$i$i = $277;$bitlen_ll$0$i$i$i = $276;$bp$12$i = $$lcssa487$lcssa;$error$11$i$i$i = 54;
                break L30;
               }
               $311 = (($bp$8$i$lcssa483$lcssa) + 3)|0; //@line 1031 "libraries/lodepng/lodepng.cpp"
               $312 = ($311>>>0)>($2>>>0); //@line 1031 "libraries/lodepng/lodepng.cpp"
               if ($312) {
                $bitlen_d$0$i$i$i = $277;$bitlen_ll$0$i$i$i = $276;$bp$12$i = $$lcssa487$lcssa;$error$11$i$i$i = 50;
                break L30;
               }
               $313 = $$lcssa487$lcssa >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
               $314 = (($in) + ($313)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
               $315 = HEAP8[$314>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
               $316 = $315&255; //@line 458 "libraries/lodepng/lodepng.cpp"
               $317 = $$lcssa487$lcssa & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
               $318 = $316 >>> $317; //@line 458 "libraries/lodepng/lodepng.cpp"
               $319 = $318 & 1; //@line 458 "libraries/lodepng/lodepng.cpp"
               $320 = (($bp$8$i$lcssa483$lcssa) + 2)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
               $321 = $320 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
               $322 = (($in) + ($321)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
               $323 = HEAP8[$322>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
               $324 = $323&255; //@line 458 "libraries/lodepng/lodepng.cpp"
               $325 = $320 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
               $326 = $324 >>> $325; //@line 458 "libraries/lodepng/lodepng.cpp"
               $327 = $326 << 1; //@line 458 "libraries/lodepng/lodepng.cpp"
               $328 = $327 & 2; //@line 458 "libraries/lodepng/lodepng.cpp"
               $329 = $328 | $319; //@line 458 "libraries/lodepng/lodepng.cpp"
               $330 = (($329) + 3)|0; //@line 1032 "libraries/lodepng/lodepng.cpp"
               $331 = ($i$392$i$i$i$lcssa491>>>0)<($280>>>0); //@line 1034 "libraries/lodepng/lodepng.cpp"
               $332 = (($i$392$i$i$i$lcssa491) + -1)|0; //@line 1034 "libraries/lodepng/lodepng.cpp"
               $333 = (($276) + ($332<<2)|0); //@line 1034 "libraries/lodepng/lodepng.cpp"
               $334 = (($285) + ($i$392$i$i$i$lcssa491))|0; //@line 1035 "libraries/lodepng/lodepng.cpp"
               $335 = (($277) + ($334<<2)|0); //@line 1035 "libraries/lodepng/lodepng.cpp"
               $value$0$in$i$i$i = $331 ? $333 : $335; //@line 1034 "libraries/lodepng/lodepng.cpp"
               $value$0$i$i$i = HEAP32[$value$0$in$i$i$i>>2]|0; //@line 1034 "libraries/lodepng/lodepng.cpp"
               $i$4111$i$i$i = $i$392$i$i$i$lcssa491;$n$0110$i$i$i = 0;
               while(1) {
                $336 = ($i$4111$i$i$i>>>0)<($281>>>0); //@line 1039 "libraries/lodepng/lodepng.cpp"
                if (!($336)) {
                 $bp$5$i = $311;$error$0$ph71$be$i$i$i = 13;$i$3$ph$be$i$i$i = $i$4111$i$i$i;
                 break L56;
                }
                $337 = ($i$4111$i$i$i>>>0)<($146>>>0); //@line 1040 "libraries/lodepng/lodepng.cpp"
                if ($337) {
                 $338 = (($276) + ($i$4111$i$i$i<<2)|0); //@line 1040 "libraries/lodepng/lodepng.cpp"
                 HEAP32[$338>>2] = $value$0$i$i$i; //@line 1040 "libraries/lodepng/lodepng.cpp"
                } else {
                 $339 = (($i$4111$i$i$i) - ($146))|0; //@line 1041 "libraries/lodepng/lodepng.cpp"
                 $340 = (($277) + ($339<<2)|0); //@line 1041 "libraries/lodepng/lodepng.cpp"
                 HEAP32[$340>>2] = $value$0$i$i$i; //@line 1041 "libraries/lodepng/lodepng.cpp"
                }
                $341 = (($i$4111$i$i$i) + 1)|0; //@line 1042 "libraries/lodepng/lodepng.cpp"
                $342 = (($n$0110$i$i$i) + 1)|0; //@line 1037 "libraries/lodepng/lodepng.cpp"
                $343 = ($342>>>0)<($330>>>0); //@line 1037 "libraries/lodepng/lodepng.cpp"
                if ($343) {
                 $i$4111$i$i$i = $341;$n$0110$i$i$i = $342;
                } else {
                 $bp$5$i = $311;$error$0$ph71$be$i$i$i = $error$0$ph71119$i$i$i;$i$3$ph$be$i$i$i = $341;
                 break;
                }
               }
               break;
              }
              case 17:  {
               $344 = (($bp$8$i$lcssa483$lcssa) + 4)|0; //@line 1048 "libraries/lodepng/lodepng.cpp"
               $345 = ($344>>>0)>($2>>>0); //@line 1048 "libraries/lodepng/lodepng.cpp"
               if ($345) {
                $bitlen_d$0$i$i$i = $277;$bitlen_ll$0$i$i$i = $276;$bp$12$i = $$lcssa487$lcssa;$error$11$i$i$i = 50;
                break L30;
               }
               $346 = $$lcssa487$lcssa >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
               $347 = (($in) + ($346)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
               $348 = HEAP8[$347>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
               $349 = $348&255; //@line 458 "libraries/lodepng/lodepng.cpp"
               $350 = $$lcssa487$lcssa & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
               $351 = $349 >>> $350; //@line 458 "libraries/lodepng/lodepng.cpp"
               $352 = $351 & 1; //@line 458 "libraries/lodepng/lodepng.cpp"
               $353 = (($bp$8$i$lcssa483$lcssa) + 2)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
               $354 = $353 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
               $355 = (($in) + ($354)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
               $356 = HEAP8[$355>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
               $357 = $356&255; //@line 458 "libraries/lodepng/lodepng.cpp"
               $358 = $353 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
               $359 = $357 >>> $358; //@line 458 "libraries/lodepng/lodepng.cpp"
               $360 = $359 << 1; //@line 458 "libraries/lodepng/lodepng.cpp"
               $361 = $360 & 2; //@line 458 "libraries/lodepng/lodepng.cpp"
               $362 = $361 | $352; //@line 458 "libraries/lodepng/lodepng.cpp"
               $363 = (($bp$8$i$lcssa483$lcssa) + 3)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
               $364 = $363 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
               $365 = (($in) + ($364)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
               $366 = HEAP8[$365>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
               $367 = $366&255; //@line 458 "libraries/lodepng/lodepng.cpp"
               $368 = $363 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
               $369 = $367 >>> $368; //@line 458 "libraries/lodepng/lodepng.cpp"
               $370 = $369 << 2; //@line 458 "libraries/lodepng/lodepng.cpp"
               $371 = $370 & 4; //@line 458 "libraries/lodepng/lodepng.cpp"
               $372 = $362 | $371; //@line 458 "libraries/lodepng/lodepng.cpp"
               $373 = (($372) + 3)|0; //@line 1049 "libraries/lodepng/lodepng.cpp"
               $i$6103$i$i$i = $i$392$i$i$i$lcssa491;$n$1102$i$i$i = 0;
               while(1) {
                $374 = ($i$6103$i$i$i>>>0)<($281>>>0); //@line 1054 "libraries/lodepng/lodepng.cpp"
                if (!($374)) {
                 $bp$5$i = $344;$error$0$ph71$be$i$i$i = 14;$i$3$ph$be$i$i$i = $i$6103$i$i$i;
                 break L56;
                }
                $375 = ($i$6103$i$i$i>>>0)<($146>>>0); //@line 1056 "libraries/lodepng/lodepng.cpp"
                if ($375) {
                 $376 = (($276) + ($i$6103$i$i$i<<2)|0); //@line 1056 "libraries/lodepng/lodepng.cpp"
                 HEAP32[$376>>2] = 0; //@line 1056 "libraries/lodepng/lodepng.cpp"
                } else {
                 $377 = (($i$6103$i$i$i) - ($146))|0; //@line 1057 "libraries/lodepng/lodepng.cpp"
                 $378 = (($277) + ($377<<2)|0); //@line 1057 "libraries/lodepng/lodepng.cpp"
                 HEAP32[$378>>2] = 0; //@line 1057 "libraries/lodepng/lodepng.cpp"
                }
                $379 = (($i$6103$i$i$i) + 1)|0; //@line 1058 "libraries/lodepng/lodepng.cpp"
                $380 = (($n$1102$i$i$i) + 1)|0; //@line 1052 "libraries/lodepng/lodepng.cpp"
                $381 = ($380>>>0)<($373>>>0); //@line 1052 "libraries/lodepng/lodepng.cpp"
                if ($381) {
                 $i$6103$i$i$i = $379;$n$1102$i$i$i = $380;
                } else {
                 $bp$5$i = $344;$error$0$ph71$be$i$i$i = $error$0$ph71119$i$i$i;$i$3$ph$be$i$i$i = $379;
                 break;
                }
               }
               break;
              }
              case 18:  {
               $382 = (($bp$8$i$lcssa483$lcssa) + 8)|0; //@line 1064 "libraries/lodepng/lodepng.cpp"
               $383 = ($382>>>0)>($2>>>0); //@line 1064 "libraries/lodepng/lodepng.cpp"
               if ($383) {
                $bitlen_d$0$i$i$i = $277;$bitlen_ll$0$i$i$i = $276;$bp$12$i = $$lcssa487$lcssa;$error$11$i$i$i = 50;
                break L30;
               }
               $384 = $$lcssa487$lcssa >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
               $385 = (($in) + ($384)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
               $386 = HEAP8[$385>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
               $387 = $386&255; //@line 458 "libraries/lodepng/lodepng.cpp"
               $388 = $$lcssa487$lcssa & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
               $389 = $387 >>> $388; //@line 458 "libraries/lodepng/lodepng.cpp"
               $390 = $389 & 1; //@line 458 "libraries/lodepng/lodepng.cpp"
               $391 = (($bp$8$i$lcssa483$lcssa) + 2)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
               $392 = $391 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
               $393 = (($in) + ($392)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
               $394 = HEAP8[$393>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
               $395 = $394&255; //@line 458 "libraries/lodepng/lodepng.cpp"
               $396 = $391 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
               $397 = $395 >>> $396; //@line 458 "libraries/lodepng/lodepng.cpp"
               $398 = $397 << 1; //@line 458 "libraries/lodepng/lodepng.cpp"
               $399 = $398 & 2; //@line 458 "libraries/lodepng/lodepng.cpp"
               $400 = $399 | $390; //@line 458 "libraries/lodepng/lodepng.cpp"
               $401 = (($bp$8$i$lcssa483$lcssa) + 3)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
               $402 = $401 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
               $403 = (($in) + ($402)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
               $404 = HEAP8[$403>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
               $405 = $404&255; //@line 458 "libraries/lodepng/lodepng.cpp"
               $406 = $401 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
               $407 = $405 >>> $406; //@line 458 "libraries/lodepng/lodepng.cpp"
               $408 = $407 << 2; //@line 458 "libraries/lodepng/lodepng.cpp"
               $409 = $408 & 4; //@line 458 "libraries/lodepng/lodepng.cpp"
               $410 = $400 | $409; //@line 458 "libraries/lodepng/lodepng.cpp"
               $411 = (($bp$8$i$lcssa483$lcssa) + 4)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
               $412 = $411 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
               $413 = (($in) + ($412)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
               $414 = HEAP8[$413>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
               $415 = $414&255; //@line 458 "libraries/lodepng/lodepng.cpp"
               $416 = $411 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
               $417 = $415 >>> $416; //@line 458 "libraries/lodepng/lodepng.cpp"
               $418 = $417 << 3; //@line 458 "libraries/lodepng/lodepng.cpp"
               $419 = $418 & 8; //@line 458 "libraries/lodepng/lodepng.cpp"
               $420 = $410 | $419; //@line 458 "libraries/lodepng/lodepng.cpp"
               $421 = (($bp$8$i$lcssa483$lcssa) + 5)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
               $422 = $421 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
               $423 = (($in) + ($422)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
               $424 = HEAP8[$423>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
               $425 = $424&255; //@line 458 "libraries/lodepng/lodepng.cpp"
               $426 = $421 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
               $427 = $425 >>> $426; //@line 458 "libraries/lodepng/lodepng.cpp"
               $428 = $427 << 4; //@line 458 "libraries/lodepng/lodepng.cpp"
               $429 = $428 & 16; //@line 458 "libraries/lodepng/lodepng.cpp"
               $430 = $420 | $429; //@line 458 "libraries/lodepng/lodepng.cpp"
               $431 = (($bp$8$i$lcssa483$lcssa) + 6)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
               $432 = $431 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
               $433 = (($in) + ($432)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
               $434 = HEAP8[$433>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
               $435 = $434&255; //@line 458 "libraries/lodepng/lodepng.cpp"
               $436 = $431 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
               $437 = $435 >>> $436; //@line 458 "libraries/lodepng/lodepng.cpp"
               $438 = $437 << 5; //@line 458 "libraries/lodepng/lodepng.cpp"
               $439 = $438 & 32; //@line 458 "libraries/lodepng/lodepng.cpp"
               $440 = $430 | $439; //@line 458 "libraries/lodepng/lodepng.cpp"
               $441 = (($bp$8$i$lcssa483$lcssa) + 7)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
               $442 = $441 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
               $443 = (($in) + ($442)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
               $444 = HEAP8[$443>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
               $445 = $444&255; //@line 458 "libraries/lodepng/lodepng.cpp"
               $446 = $441 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
               $447 = $445 >>> $446; //@line 458 "libraries/lodepng/lodepng.cpp"
               $448 = $447 << 6; //@line 458 "libraries/lodepng/lodepng.cpp"
               $449 = $448 & 64; //@line 458 "libraries/lodepng/lodepng.cpp"
               $450 = $440 | $449; //@line 458 "libraries/lodepng/lodepng.cpp"
               $451 = (($450) + 11)|0; //@line 1065 "libraries/lodepng/lodepng.cpp"
               $452 = ($451|0)==(0); //@line 1068 "libraries/lodepng/lodepng.cpp"
               if ($452) {
                $bp$5$i = $382;$error$0$ph71$be$i$i$i = $error$0$ph71119$i$i$i;$i$3$ph$be$i$i$i = $i$392$i$i$i$lcssa491;
               } else {
                $i$895$i$i$i = $i$392$i$i$i$lcssa491;$n$294$i$i$i = 0;
                while(1) {
                 $453 = ($i$895$i$i$i>>>0)<($281>>>0); //@line 1070 "libraries/lodepng/lodepng.cpp"
                 if (!($453)) {
                  $bp$5$i = $382;$error$0$ph71$be$i$i$i = 15;$i$3$ph$be$i$i$i = $i$895$i$i$i;
                  break L56;
                 }
                 $454 = ($i$895$i$i$i>>>0)<($146>>>0); //@line 1072 "libraries/lodepng/lodepng.cpp"
                 if ($454) {
                  $455 = (($276) + ($i$895$i$i$i<<2)|0); //@line 1072 "libraries/lodepng/lodepng.cpp"
                  HEAP32[$455>>2] = 0; //@line 1072 "libraries/lodepng/lodepng.cpp"
                 } else {
                  $456 = (($i$895$i$i$i) - ($146))|0; //@line 1073 "libraries/lodepng/lodepng.cpp"
                  $457 = (($277) + ($456<<2)|0); //@line 1073 "libraries/lodepng/lodepng.cpp"
                  HEAP32[$457>>2] = 0; //@line 1073 "libraries/lodepng/lodepng.cpp"
                 }
                 $458 = (($i$895$i$i$i) + 1)|0; //@line 1074 "libraries/lodepng/lodepng.cpp"
                 $459 = (($n$294$i$i$i) + 1)|0; //@line 1068 "libraries/lodepng/lodepng.cpp"
                 $460 = ($459>>>0)<($451>>>0); //@line 1068 "libraries/lodepng/lodepng.cpp"
                 if ($460) {
                  $i$895$i$i$i = $458;$n$294$i$i$i = $459;
                 } else {
                  $bp$5$i = $382;$error$0$ph71$be$i$i$i = $error$0$ph71119$i$i$i;$i$3$ph$be$i$i$i = $458;
                  break;
                 }
                }
               }
               break;
              }
              default: {
               $bitlen_d$0$i$i$i = $277;$bitlen_ll$0$i$i$i = $276;$bp$12$i = $$lcssa487$lcssa;$error$11$i$i$i = 16;
               break L30;
              }
              }
             } while(0);
             $286 = ($i$3$ph$be$i$i$i>>>0)<($281>>>0); //@line 1015 "libraries/lodepng/lodepng.cpp"
             if ($286) {
              $bp$6$i = $bp$5$i;$error$0$ph71119$i$i$i = $error$0$ph71$be$i$i$i;$i$3$ph118$i$i$i = $i$3$ph$be$i$i$i;
             } else {
              $bp$10$i = $bp$5$i;$error$0$ph71$lcssa84$i$i$i = $error$0$ph71$be$i$i$i;
              break;
             }
            }
            if ((label|0) == 64) {
             label = 0;
             $461 = ($bp$9$i>>>0)>($2>>>0); //@line 1083 "libraries/lodepng/lodepng.cpp"
             $462 = $461 ? 10 : 11; //@line 1083 "libraries/lodepng/lodepng.cpp"
             $bitlen_d$0$i$i$i = $277;$bitlen_ll$0$i$i$i = $276;$bp$12$i = $bp$9$i;$error$11$i$i$i = $462;
             break;
            }
            $463 = ($error$0$ph71$lcssa84$i$i$i|0)==(0); //@line 1089 "libraries/lodepng/lodepng.cpp"
            if ($463) {
             $$phi$trans$insert$i$i = ((($276)) + 1024|0); //@line 1091 "libraries/lodepng/lodepng.cpp"
             $$pre$i$i = HEAP32[$$phi$trans$insert$i$i>>2]|0; //@line 1091 "libraries/lodepng/lodepng.cpp"
             $464 = ($$pre$i$i|0)==(0); //@line 1091 "libraries/lodepng/lodepng.cpp"
             if ($464) {
              $bitlen_d$0$i$i$i = $277;$bitlen_ll$0$i$i$i = $276;$bp$12$i = $bp$10$i;$error$11$i$i$i = 64;
             } else {
              $465 = (_malloc(1152)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
              HEAP32[$4>>2] = $465; //@line 659 "libraries/lodepng/lodepng.cpp"
              $466 = ($465|0)==(0|0); //@line 660 "libraries/lodepng/lodepng.cpp"
              if ($466) {
               $bitlen_d$0$i$i$i = $277;$bitlen_ll$0$i$i$i = $276;$bp$12$i = $bp$10$i;$error$11$i$i$i = 83;
              } else {
               _memcpy(($465|0),($276|0),1152)|0; //@line 661 "libraries/lodepng/lodepng.cpp"
               HEAP32[$7>>2] = 288; //@line 662 "libraries/lodepng/lodepng.cpp"
               HEAP32[$13>>2] = 15; //@line 663 "libraries/lodepng/lodepng.cpp"
               $467 = (__ZL28HuffmanTree_makeFromLengths2P11HuffmanTree($tree_ll$i$i)|0); //@line 664 "libraries/lodepng/lodepng.cpp"
               $468 = ($467|0)==(0); //@line 1095 "libraries/lodepng/lodepng.cpp"
               if ($468) {
                $469 = (_malloc(128)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
                HEAP32[$6>>2] = $469; //@line 659 "libraries/lodepng/lodepng.cpp"
                $470 = ($469|0)==(0|0); //@line 660 "libraries/lodepng/lodepng.cpp"
                if ($470) {
                 $bitlen_d$0$i$i$i = $277;$bitlen_ll$0$i$i$i = $276;$bp$12$i = $bp$10$i;$error$11$i$i$i = 83;
                } else {
                 dest=$469; src=$277; stop=dest+128|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 661 "libraries/lodepng/lodepng.cpp"
                 HEAP32[$8>>2] = 32; //@line 662 "libraries/lodepng/lodepng.cpp"
                 HEAP32[$14>>2] = 15; //@line 663 "libraries/lodepng/lodepng.cpp"
                 $471 = (__ZL28HuffmanTree_makeFromLengths2P11HuffmanTree($tree_d$i$i)|0); //@line 664 "libraries/lodepng/lodepng.cpp"
                 $bitlen_d$0$i$i$i = $277;$bitlen_ll$0$i$i$i = $276;$bp$12$i = $bp$10$i;$error$11$i$i$i = $471;
                }
               } else {
                $bitlen_d$0$i$i$i = $277;$bitlen_ll$0$i$i$i = $276;$bp$12$i = $bp$10$i;$error$11$i$i$i = $467;
               }
              }
             }
            } else {
             $bitlen_d$0$i$i$i = $277;$bitlen_ll$0$i$i$i = $276;$bp$12$i = $bp$10$i;$error$11$i$i$i = $error$0$ph71$lcssa84$i$i$i;
            }
           }
          } else {
           $bitlen_d$0$i$i$i = $277;$bitlen_ll$0$i$i$i = $276;$bp$12$i = $bp$4$i$lcssa;$error$11$i$i$i = 83;
          }
         } else {
          $bitlen_d$0$i$i$i = 0;$bitlen_ll$0$i$i$i = 0;$bp$12$i = $bp$4$i$lcssa;$error$11$i$i$i = $274;
         }
        }
       }
      } while(0);
      _free($236); //@line 76 "libraries/lodepng/lodepng.cpp"
      _free($bitlen_ll$0$i$i$i); //@line 76 "libraries/lodepng/lodepng.cpp"
      _free($bitlen_d$0$i$i$i); //@line 76 "libraries/lodepng/lodepng.cpp"
      $472 = HEAP32[$tree_cl$i$i$i>>2]|0; //@line 539 "libraries/lodepng/lodepng.cpp"
      _free($472); //@line 76 "libraries/lodepng/lodepng.cpp"
      $473 = HEAP32[$9>>2]|0; //@line 540 "libraries/lodepng/lodepng.cpp"
      _free($473); //@line 76 "libraries/lodepng/lodepng.cpp"
      $474 = HEAP32[$10>>2]|0; //@line 541 "libraries/lodepng/lodepng.cpp"
      _free($474); //@line 76 "libraries/lodepng/lodepng.cpp"
      $475 = ($error$11$i$i$i|0)==(0); //@line 1124 "libraries/lodepng/lodepng.cpp"
      if ($475) {
       $bp$14$i = $bp$12$i;
       label = 74;
       break L24;
      } else {
       $$0$i116$i$i = $error$11$i$i$i;$bp$13$i = $bp$12$i;$pos$080$i478 = $pos$080$i$lcssa477;$v$sroa$0$1476 = $v$sroa$0$1$lcssa475;$v$sroa$13$1474 = $v$sroa$13$1$lcssa473;$v$sroa$18$1472 = $v$sroa$18$1$lcssa471;
      }
     }
    }
    if ((label|0) == 71) {
     label = 0;
     $$0$i116$i$i = $$0$i$ph$i$i;$bp$13$i = $bp$11$i;$pos$080$i478 = $pos$080$i$lcssa477;$v$sroa$0$1476 = $v$sroa$0$1$lcssa475;$v$sroa$13$1474 = $v$sroa$13$1$lcssa473;$v$sroa$18$1472 = $v$sroa$18$1$lcssa471;
    }
    $$pre108$i$i = HEAP32[$tree_ll$i$i>>2]|0; //@line 539 "libraries/lodepng/lodepng.cpp"
    $$pre110$i$i = HEAP32[$tree_d$i$i>>2]|0; //@line 539 "libraries/lodepng/lodepng.cpp"
    $601 = $$pre108$i$i;$604 = $$pre110$i$i;$bp$22$i = $bp$13$i;$error$5$i$i = $$0$i116$i$i;$pos$6$i = $pos$080$i478;$v$sroa$0$7 = $v$sroa$0$1476;$v$sroa$13$4 = $v$sroa$13$1474;$v$sroa$18$7 = $v$sroa$18$1472;
    break;
   }
   default: {
    $bp$14$i = $$lcssa481;
    label = 74;
   }
   }
  } while(0);
  L98: do {
   if ((label|0) == 74) {
    label = 0;
    $476 = HEAP32[$tree_ll$i$i>>2]|0; //@line 937 "libraries/lodepng/lodepng.cpp"
    $477 = HEAP32[$7>>2]|0; //@line 939 "libraries/lodepng/lodepng.cpp"
    $478 = HEAP32[$tree_d$i$i>>2]|0; //@line 937 "libraries/lodepng/lodepng.cpp"
    $479 = HEAP32[$8>>2]|0; //@line 939 "libraries/lodepng/lodepng.cpp"
    $bp$15$ph$i = $bp$14$i;$pos$4$ph$i = $pos$080$i$lcssa477;$v$sroa$0$3 = $v$sroa$0$1$lcssa475;$v$sroa$13$2 = $v$sroa$13$1$lcssa473;$v$sroa$18$3 = $v$sroa$18$1$lcssa471;
    L100: while(1) {
     $bp$16$i$ph = $bp$15$ph$i;$v$sroa$0$4$ph = $v$sroa$0$3;$v$sroa$13$3$ph = $v$sroa$13$2;$v$sroa$18$4$ph = $v$sroa$18$3;
     while(1) {
      $bp$16$i = $bp$16$i$ph;$treepos$0$i$i$i = 0;
      while(1) {
       $480 = ($bp$16$i>>>0)<($2>>>0); //@line 932 "libraries/lodepng/lodepng.cpp"
       if (!($480)) {
        $bp$21$i = $bp$16$i;$pos$4$ph$i550 = $pos$4$ph$i;$v$sroa$0$4$ph538 = $v$sroa$0$4$ph;$v$sroa$13$3$ph530 = $v$sroa$13$3$ph;$v$sroa$18$4$ph522 = $v$sroa$18$4$ph;
        break L100;
       }
       $481 = $treepos$0$i$i$i << 1; //@line 937 "libraries/lodepng/lodepng.cpp"
       $482 = $bp$16$i >>> 3; //@line 937 "libraries/lodepng/lodepng.cpp"
       $483 = (($in) + ($482)|0); //@line 937 "libraries/lodepng/lodepng.cpp"
       $484 = HEAP8[$483>>0]|0; //@line 937 "libraries/lodepng/lodepng.cpp"
       $485 = $484&255; //@line 937 "libraries/lodepng/lodepng.cpp"
       $486 = $bp$16$i & 7; //@line 937 "libraries/lodepng/lodepng.cpp"
       $487 = $485 >>> $486; //@line 937 "libraries/lodepng/lodepng.cpp"
       $488 = $487 & 1; //@line 937 "libraries/lodepng/lodepng.cpp"
       $489 = $488 | $481; //@line 937 "libraries/lodepng/lodepng.cpp"
       $490 = (($476) + ($489<<2)|0); //@line 937 "libraries/lodepng/lodepng.cpp"
       $491 = HEAP32[$490>>2]|0; //@line 937 "libraries/lodepng/lodepng.cpp"
       $492 = (($bp$16$i) + 1)|0; //@line 1146 "libraries/lodepng/lodepng.cpp"
       $493 = ($491>>>0)<($477>>>0); //@line 939 "libraries/lodepng/lodepng.cpp"
       if ($493) {
        $$lcssa497 = $491;$$lcssa504 = $492;
        break;
       }
       $494 = (($491) - ($477))|0; //@line 940 "libraries/lodepng/lodepng.cpp"
       $495 = ($494>>>0)<($477>>>0); //@line 942 "libraries/lodepng/lodepng.cpp"
       if ($495) {
        $bp$16$i = $492;$treepos$0$i$i$i = $494;
       } else {
        $bp$21$i = $492;$pos$4$ph$i550 = $pos$4$ph$i;$v$sroa$0$4$ph538 = $v$sroa$0$4$ph;$v$sroa$13$3$ph530 = $v$sroa$13$3$ph;$v$sroa$18$4$ph522 = $v$sroa$18$4$ph;
        break L100;
       }
      }
      $496 = ($$lcssa497>>>0)<(256); //@line 1128 "libraries/lodepng/lodepng.cpp"
      if ($496) {
       $$lcssa497$lcssa = $$lcssa497;$$lcssa504$lcssa = $$lcssa504;$v$sroa$0$4$ph$lcssa531 = $v$sroa$0$4$ph;$v$sroa$13$3$ph$lcssa523 = $v$sroa$13$3$ph;$v$sroa$18$4$ph$lcssa515 = $v$sroa$18$4$ph;
       break;
      }
      $$off$i$i = (($$lcssa497) + -257)|0; //@line 1135 "libraries/lodepng/lodepng.cpp"
      $512 = ($$off$i$i>>>0)<(29); //@line 1135 "libraries/lodepng/lodepng.cpp"
      if (!($512)) {
       $$lcssa497$lcssa498 = $$lcssa497;$$lcssa504$lcssa505 = $$lcssa504;$pos$4$ph$i$lcssa544 = $pos$4$ph$i;$v$sroa$0$4$ph$lcssa532 = $v$sroa$0$4$ph;$v$sroa$13$3$ph$lcssa524 = $v$sroa$13$3$ph;$v$sroa$18$4$ph$lcssa516 = $v$sroa$18$4$ph;
       label = 107;
       break L100;
      }
      $513 = (33388 + ($$off$i$i<<2)|0); //@line 1145 "libraries/lodepng/lodepng.cpp"
      $514 = HEAP32[$513>>2]|0; //@line 1145 "libraries/lodepng/lodepng.cpp"
      $515 = (($514) + ($$lcssa504))|0; //@line 1146 "libraries/lodepng/lodepng.cpp"
      $516 = ($515>>>0)>($2>>>0); //@line 1146 "libraries/lodepng/lodepng.cpp"
      if ($516) {
       $601 = $476;$604 = $478;$bp$22$i = $$lcssa504;$error$5$i$i = 51;$pos$6$i = $pos$4$ph$i;$v$sroa$0$7 = $v$sroa$0$4$ph;$v$sroa$13$4 = $v$sroa$13$3$ph;$v$sroa$18$7 = $v$sroa$18$4$ph;
       break L98;
      }
      $517 = (33504 + ($$off$i$i<<2)|0); //@line 1142 "libraries/lodepng/lodepng.cpp"
      $518 = HEAP32[$517>>2]|0; //@line 1142 "libraries/lodepng/lodepng.cpp"
      $519 = (($$lcssa497) + -265)|0; //@line 456 "libraries/lodepng/lodepng.cpp"
      $520 = ($519>>>0)>(19); //@line 456 "libraries/lodepng/lodepng.cpp"
      if ($520) {
       $bp$17$i = $$lcssa504;$result$0$lcssa$i$i$i = 0;
      } else {
       $522 = $$lcssa504;$i$02$i$i$i = 0;$result$01$i$i$i = 0;
       while(1) {
        $521 = $522 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
        $523 = (($in) + ($521)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
        $524 = HEAP8[$523>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
        $525 = $524&255; //@line 458 "libraries/lodepng/lodepng.cpp"
        $526 = $522 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
        $527 = $525 >>> $526; //@line 458 "libraries/lodepng/lodepng.cpp"
        $528 = $527 & 1; //@line 458 "libraries/lodepng/lodepng.cpp"
        $529 = $528 << $i$02$i$i$i; //@line 458 "libraries/lodepng/lodepng.cpp"
        $530 = (($529) + ($result$01$i$i$i))|0; //@line 458 "libraries/lodepng/lodepng.cpp"
        $531 = (($522) + 1)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
        $532 = (($i$02$i$i$i) + 1)|0; //@line 456 "libraries/lodepng/lodepng.cpp"
        $533 = ($532|0)==($514|0); //@line 456 "libraries/lodepng/lodepng.cpp"
        if ($533) {
         $bp$17$i = $515;$result$0$lcssa$i$i$i = $530;
         break;
        } else {
         $522 = $531;$i$02$i$i$i = $532;$result$01$i$i$i = $530;
        }
       }
      }
      $534 = (($result$0$lcssa$i$i$i) + ($518))|0; //@line 1147 "libraries/lodepng/lodepng.cpp"
      $bp$18$i = $bp$17$i;$treepos$0$i8$i$i = 0;
      while(1) {
       $535 = ($bp$18$i>>>0)<($2>>>0); //@line 932 "libraries/lodepng/lodepng.cpp"
       if (!($535)) {
        $$lcssa497503 = $$lcssa497;$bp$19$i = $bp$18$i;$pos$4$ph$i549 = $pos$4$ph$i;$v$sroa$0$4$ph537 = $v$sroa$0$4$ph;$v$sroa$13$3$ph529 = $v$sroa$13$3$ph;$v$sroa$18$4$ph521 = $v$sroa$18$4$ph;
        label = 95;
        break L100;
       }
       $536 = $treepos$0$i8$i$i << 1; //@line 937 "libraries/lodepng/lodepng.cpp"
       $537 = $bp$18$i >>> 3; //@line 937 "libraries/lodepng/lodepng.cpp"
       $538 = (($in) + ($537)|0); //@line 937 "libraries/lodepng/lodepng.cpp"
       $539 = HEAP8[$538>>0]|0; //@line 937 "libraries/lodepng/lodepng.cpp"
       $540 = $539&255; //@line 937 "libraries/lodepng/lodepng.cpp"
       $541 = $bp$18$i & 7; //@line 937 "libraries/lodepng/lodepng.cpp"
       $542 = $540 >>> $541; //@line 937 "libraries/lodepng/lodepng.cpp"
       $543 = $542 & 1; //@line 937 "libraries/lodepng/lodepng.cpp"
       $544 = $543 | $536; //@line 937 "libraries/lodepng/lodepng.cpp"
       $545 = (($478) + ($544<<2)|0); //@line 937 "libraries/lodepng/lodepng.cpp"
       $546 = HEAP32[$545>>2]|0; //@line 937 "libraries/lodepng/lodepng.cpp"
       $547 = (($bp$18$i) + 1)|0; //@line 1166 "libraries/lodepng/lodepng.cpp"
       $548 = ($546>>>0)<($479>>>0); //@line 939 "libraries/lodepng/lodepng.cpp"
       if ($548) {
        $$lcssa511 = $546;$$lcssa512 = $547;
        break;
       }
       $549 = (($546) - ($479))|0; //@line 940 "libraries/lodepng/lodepng.cpp"
       $550 = ($549>>>0)<($479>>>0); //@line 942 "libraries/lodepng/lodepng.cpp"
       if ($550) {
        $bp$18$i = $547;$treepos$0$i8$i$i = $549;
       } else {
        $$lcssa497503 = $$lcssa497;$bp$19$i = $547;$pos$4$ph$i549 = $pos$4$ph$i;$v$sroa$0$4$ph537 = $v$sroa$0$4$ph;$v$sroa$13$3$ph529 = $v$sroa$13$3$ph;$v$sroa$18$4$ph521 = $v$sroa$18$4$ph;
        label = 95;
        break L100;
       }
      }
      $551 = ($$lcssa511>>>0)>(29); //@line 1151 "libraries/lodepng/lodepng.cpp"
      if ($551) {
       $$lcssa497503 = $$lcssa497;$bp$19$i = $$lcssa512;$pos$4$ph$i549 = $pos$4$ph$i;$v$sroa$0$4$ph537 = $v$sroa$0$4$ph;$v$sroa$13$3$ph529 = $v$sroa$13$3$ph;$v$sroa$18$4$ph521 = $v$sroa$18$4$ph;
       label = 95;
       break L100;
      }
      $555 = (33620 + ($$lcssa511<<2)|0); //@line 1165 "libraries/lodepng/lodepng.cpp"
      $556 = HEAP32[$555>>2]|0; //@line 1165 "libraries/lodepng/lodepng.cpp"
      $557 = (($556) + ($$lcssa512))|0; //@line 1166 "libraries/lodepng/lodepng.cpp"
      $558 = ($557>>>0)>($2>>>0); //@line 1166 "libraries/lodepng/lodepng.cpp"
      if ($558) {
       $601 = $476;$604 = $478;$bp$22$i = $$lcssa512;$error$5$i$i = 51;$pos$6$i = $pos$4$ph$i;$v$sroa$0$7 = $v$sroa$0$4$ph;$v$sroa$13$4 = $v$sroa$13$3$ph;$v$sroa$18$7 = $v$sroa$18$4$ph;
       break L98;
      }
      $559 = (33740 + ($$lcssa511<<2)|0); //@line 1162 "libraries/lodepng/lodepng.cpp"
      $560 = HEAP32[$559>>2]|0; //@line 1162 "libraries/lodepng/lodepng.cpp"
      $561 = ($$lcssa511>>>0)<(4); //@line 456 "libraries/lodepng/lodepng.cpp"
      if ($561) {
       $bp$20$i = $$lcssa512;$result$0$lcssa$i16$i$i = 0;
      } else {
       $563 = $$lcssa512;$i$02$i13$i$i = 0;$result$01$i14$i$i = 0;
       while(1) {
        $562 = $563 >>> 3; //@line 458 "libraries/lodepng/lodepng.cpp"
        $564 = (($in) + ($562)|0); //@line 458 "libraries/lodepng/lodepng.cpp"
        $565 = HEAP8[$564>>0]|0; //@line 458 "libraries/lodepng/lodepng.cpp"
        $566 = $565&255; //@line 458 "libraries/lodepng/lodepng.cpp"
        $567 = $563 & 7; //@line 458 "libraries/lodepng/lodepng.cpp"
        $568 = $566 >>> $567; //@line 458 "libraries/lodepng/lodepng.cpp"
        $569 = $568 & 1; //@line 458 "libraries/lodepng/lodepng.cpp"
        $570 = $569 << $i$02$i13$i$i; //@line 458 "libraries/lodepng/lodepng.cpp"
        $571 = (($570) + ($result$01$i14$i$i))|0; //@line 458 "libraries/lodepng/lodepng.cpp"
        $572 = (($563) + 1)|0; //@line 459 "libraries/lodepng/lodepng.cpp"
        $573 = (($i$02$i13$i$i) + 1)|0; //@line 456 "libraries/lodepng/lodepng.cpp"
        $574 = ($573|0)==($556|0); //@line 456 "libraries/lodepng/lodepng.cpp"
        if ($574) {
         $bp$20$i = $557;$result$0$lcssa$i16$i$i = $571;
         break;
        } else {
         $563 = $572;$i$02$i13$i$i = $573;$result$01$i14$i$i = $571;
        }
       }
      }
      $575 = (($result$0$lcssa$i16$i$i) + ($560))|0; //@line 1167 "libraries/lodepng/lodepng.cpp"
      $576 = ($pos$4$ph$i>>>0)<($575>>>0); //@line 1171 "libraries/lodepng/lodepng.cpp"
      if ($576) {
       $601 = $476;$604 = $478;$bp$22$i = $bp$20$i;$error$5$i$i = 52;$pos$6$i = $pos$4$ph$i;$v$sroa$0$7 = $v$sroa$0$4$ph;$v$sroa$13$4 = $v$sroa$13$3$ph;$v$sroa$18$7 = $v$sroa$18$4$ph;
       break L98;
      }
      $577 = (($534) + ($pos$4$ph$i))|0; //@line 1174 "libraries/lodepng/lodepng.cpp"
      $578 = ($v$sroa$18$4$ph>>>0)<($577>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
      if ($578) {
       $579 = $v$sroa$18$4$ph << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
       $580 = ($579>>>0)<($577>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
       $581 = ($577*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
       $582 = $581 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
       $583 = $580 ? $577 : $582; //@line 227 "libraries/lodepng/lodepng.cpp"
       $584 = $v$sroa$0$4$ph; //@line 228 "libraries/lodepng/lodepng.cpp"
       $585 = (_realloc($584,$583)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       $586 = ($585|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
       if ($586) {
        $601 = $476;$604 = $478;$bp$22$i = $bp$20$i;$error$5$i$i = 83;$pos$6$i = $pos$4$ph$i;$v$sroa$0$7 = $v$sroa$0$4$ph;$v$sroa$13$4 = $v$sroa$13$3$ph;$v$sroa$18$7 = $v$sroa$18$4$ph;
        break L98;
       }
       $587 = $585; //@line 232 "libraries/lodepng/lodepng.cpp"
       $v$sroa$0$6 = $587;$v$sroa$18$6 = $583;
      } else {
       $v$sroa$0$6 = $v$sroa$0$4$ph;$v$sroa$18$6 = $v$sroa$18$4$ph;
      }
      $588 = ($534|0)==(0); //@line 1175 "libraries/lodepng/lodepng.cpp"
      if ($588) {
       $bp$16$i$ph = $bp$20$i;$v$sroa$0$4$ph = $v$sroa$0$6;$v$sroa$13$3$ph = $577;$v$sroa$18$4$ph = $v$sroa$18$6;
      } else {
       $$lcssa541 = $534;$$lcssa542 = $575;$$lcssa543 = $577;$bp$20$i$lcssa = $bp$20$i;$v$sroa$0$6$lcssa = $v$sroa$0$6;$v$sroa$18$6$lcssa = $v$sroa$18$6;
       label = 105;
       break;
      }
     }
     if ((label|0) == 105) {
      label = 0;
      $589 = (($pos$4$ph$i) - ($$lcssa542))|0; //@line 1172 "libraries/lodepng/lodepng.cpp"
      $590 = $v$sroa$0$6$lcssa; //@line 1177 "libraries/lodepng/lodepng.cpp"
      $backward$061$i$i = $589;$forward$060$i$i = 0;$pos$5$i = $pos$4$ph$i;
      while(1) {
       $591 = (($590) + ($backward$061$i$i)|0); //@line 1177 "libraries/lodepng/lodepng.cpp"
       $592 = HEAP8[$591>>0]|0; //@line 1177 "libraries/lodepng/lodepng.cpp"
       $593 = (($590) + ($pos$5$i)|0); //@line 1177 "libraries/lodepng/lodepng.cpp"
       HEAP8[$593>>0] = $592; //@line 1177 "libraries/lodepng/lodepng.cpp"
       $594 = (($pos$5$i) + 1)|0; //@line 1178 "libraries/lodepng/lodepng.cpp"
       $595 = (($backward$061$i$i) + 1)|0; //@line 1179 "libraries/lodepng/lodepng.cpp"
       $596 = ($595>>>0)<($pos$4$ph$i>>>0); //@line 1180 "libraries/lodepng/lodepng.cpp"
       $$$i$i = $596 ? $595 : $589; //@line 1180 "libraries/lodepng/lodepng.cpp"
       $597 = (($forward$060$i$i) + 1)|0; //@line 1175 "libraries/lodepng/lodepng.cpp"
       $exitcond$i$i = ($597|0)==($$lcssa541|0); //@line 1175 "libraries/lodepng/lodepng.cpp"
       if ($exitcond$i$i) {
        $bp$15$ph$i = $bp$20$i$lcssa;$pos$4$ph$i = $$lcssa543;$v$sroa$0$3 = $v$sroa$0$6$lcssa;$v$sroa$13$2 = $$lcssa543;$v$sroa$18$3 = $v$sroa$18$6$lcssa;
        continue L100;
       } else {
        $backward$061$i$i = $$$i$i;$forward$060$i$i = $597;$pos$5$i = $594;
       }
      }
     }
     $497 = (($pos$4$ph$i) + 1)|0; //@line 1131 "libraries/lodepng/lodepng.cpp"
     $498 = ($v$sroa$18$4$ph$lcssa515>>>0)<($497>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
     if ($498) {
      $500 = $v$sroa$18$4$ph$lcssa515 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
      $501 = ($500>>>0)<($497>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
      $502 = ($497*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
      $503 = $502 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
      $504 = $501 ? $497 : $503; //@line 227 "libraries/lodepng/lodepng.cpp"
      $505 = $v$sroa$0$4$ph$lcssa531; //@line 228 "libraries/lodepng/lodepng.cpp"
      $506 = (_realloc($505,$504)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $507 = ($506|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
      if ($507) {
       $601 = $476;$604 = $478;$bp$22$i = $$lcssa504$lcssa;$error$5$i$i = 83;$pos$6$i = $pos$4$ph$i;$v$sroa$0$7 = $v$sroa$0$4$ph$lcssa531;$v$sroa$13$4 = $v$sroa$13$3$ph$lcssa523;$v$sroa$18$7 = $v$sroa$18$4$ph$lcssa515;
       break L98;
      }
      $508 = $506; //@line 232 "libraries/lodepng/lodepng.cpp"
      $511 = $506;$v$sroa$0$5 = $508;$v$sroa$18$5 = $504;
     } else {
      $499 = $v$sroa$0$4$ph$lcssa531; //@line 1132 "libraries/lodepng/lodepng.cpp"
      $511 = $499;$v$sroa$0$5 = $v$sroa$0$4$ph$lcssa531;$v$sroa$18$5 = $v$sroa$18$4$ph$lcssa515;
     }
     $509 = $$lcssa497$lcssa&255; //@line 1132 "libraries/lodepng/lodepng.cpp"
     $510 = (($511) + ($pos$4$ph$i)|0); //@line 1132 "libraries/lodepng/lodepng.cpp"
     HEAP8[$510>>0] = $509; //@line 1132 "libraries/lodepng/lodepng.cpp"
     $bp$15$ph$i = $$lcssa504$lcssa;$pos$4$ph$i = $497;$v$sroa$0$3 = $v$sroa$0$5;$v$sroa$13$2 = $497;$v$sroa$18$3 = $v$sroa$18$5;
    }
    if ((label|0) == 95) {
     label = 0;
     $552 = ($$lcssa497503|0)==(-1); //@line 1153 "libraries/lodepng/lodepng.cpp"
     if (!($552)) {
      $601 = $476;$604 = $478;$bp$22$i = $bp$19$i;$error$5$i$i = 18;$pos$6$i = $pos$4$ph$i549;$v$sroa$0$7 = $v$sroa$0$4$ph537;$v$sroa$13$4 = $v$sroa$13$3$ph529;$v$sroa$18$7 = $v$sroa$18$4$ph521;
      break;
     }
     $553 = ($bp$19$i>>>0)>($2>>>0); //@line 1157 "libraries/lodepng/lodepng.cpp"
     $554 = $553 ? 10 : 11; //@line 1157 "libraries/lodepng/lodepng.cpp"
     $601 = $476;$604 = $478;$bp$22$i = $bp$19$i;$error$5$i$i = $554;$pos$6$i = $pos$4$ph$i549;$v$sroa$0$7 = $v$sroa$0$4$ph537;$v$sroa$13$4 = $v$sroa$13$3$ph529;$v$sroa$18$7 = $v$sroa$18$4$ph521;
     break;
    }
    else if ((label|0) == 107) {
     label = 0;
     $598 = ($$lcssa497$lcssa498|0)==(256); //@line 1183 "libraries/lodepng/lodepng.cpp"
     if ($598) {
      $601 = $476;$604 = $478;$bp$22$i = $$lcssa504$lcssa505;$error$5$i$i = 0;$pos$6$i = $pos$4$ph$i$lcssa544;$v$sroa$0$7 = $v$sroa$0$4$ph$lcssa532;$v$sroa$13$4 = $v$sroa$13$3$ph$lcssa524;$v$sroa$18$7 = $v$sroa$18$4$ph$lcssa516;
      break;
     } else {
      $bp$21$i = $$lcssa504$lcssa505;$pos$4$ph$i550 = $pos$4$ph$i$lcssa544;$v$sroa$0$4$ph538 = $v$sroa$0$4$ph$lcssa532;$v$sroa$13$3$ph530 = $v$sroa$13$3$ph$lcssa524;$v$sroa$18$4$ph522 = $v$sroa$18$4$ph$lcssa516;
     }
    }
    $599 = ($bp$21$i>>>0)>($2>>>0); //@line 1191 "libraries/lodepng/lodepng.cpp"
    $600 = $599 ? 10 : 11; //@line 1191 "libraries/lodepng/lodepng.cpp"
    $601 = $476;$604 = $478;$bp$22$i = $bp$21$i;$error$5$i$i = $600;$pos$6$i = $pos$4$ph$i550;$v$sroa$0$7 = $v$sroa$0$4$ph538;$v$sroa$13$4 = $v$sroa$13$3$ph530;$v$sroa$18$7 = $v$sroa$18$4$ph522;
   }
  } while(0);
  _free($601); //@line 76 "libraries/lodepng/lodepng.cpp"
  $602 = HEAP32[$3>>2]|0; //@line 540 "libraries/lodepng/lodepng.cpp"
  _free($602); //@line 76 "libraries/lodepng/lodepng.cpp"
  $603 = HEAP32[$4>>2]|0; //@line 541 "libraries/lodepng/lodepng.cpp"
  _free($603); //@line 76 "libraries/lodepng/lodepng.cpp"
  _free($604); //@line 76 "libraries/lodepng/lodepng.cpp"
  $605 = HEAP32[$5>>2]|0; //@line 540 "libraries/lodepng/lodepng.cpp"
  _free($605); //@line 76 "libraries/lodepng/lodepng.cpp"
  $606 = HEAP32[$6>>2]|0; //@line 541 "libraries/lodepng/lodepng.cpp"
  _free($606); //@line 76 "libraries/lodepng/lodepng.cpp"
  $607 = ($error$5$i$i|0)==(0); //@line 1254 "libraries/lodepng/lodepng.cpp"
  if ($607) {
   $BFINAL$0$ph$i = $$lcssa480;$bp$0$ph$i = $bp$22$i;$pos$0$ph$i = $pos$6$i;$v$sroa$0$0 = $v$sroa$0$7;$v$sroa$13$0 = $v$sroa$13$4;$v$sroa$18$0 = $v$sroa$18$7;
  } else {
   $$2$i = $error$5$i$i;$v$sroa$0$8 = $v$sroa$0$7;$v$sroa$13$5 = $v$sroa$13$4;
   label = 110;
   break;
  }
 }
 if ((label|0) == 110) {
  HEAP32[$out>>2] = $v$sroa$0$8; //@line 1268 "libraries/lodepng/lodepng.cpp"
  HEAP32[$outsize>>2] = $v$sroa$13$5; //@line 1269 "libraries/lodepng/lodepng.cpp"
  STACKTOP = sp;return ($$2$i|0); //@line 1270 "libraries/lodepng/lodepng.cpp"
 }
 return (0)|0;
}
function __ZL23generateFixedLitLenTreeP11HuffmanTree($tree) {
 $tree = $tree|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond10 = 0, $exitcond9 = 0;
 var $i$07 = 0, $i$16 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_malloc(1152)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
 $1 = ($0|0)==(0|0); //@line 891 "libraries/lodepng/lodepng.cpp"
 if ($1) {
  return;
 } else {
  $i$07 = 0;
 }
 while(1) {
  $2 = (($0) + ($i$07<<2)|0); //@line 894 "libraries/lodepng/lodepng.cpp"
  HEAP32[$2>>2] = 8; //@line 894 "libraries/lodepng/lodepng.cpp"
  $3 = (($i$07) + 1)|0; //@line 894 "libraries/lodepng/lodepng.cpp"
  $exitcond10 = ($3|0)==(144); //@line 894 "libraries/lodepng/lodepng.cpp"
  if ($exitcond10) {
   $i$16 = 144;
   break;
  } else {
   $i$07 = $3;
  }
 }
 while(1) {
  $4 = (($0) + ($i$16<<2)|0); //@line 895 "libraries/lodepng/lodepng.cpp"
  HEAP32[$4>>2] = 9; //@line 895 "libraries/lodepng/lodepng.cpp"
  $5 = (($i$16) + 1)|0; //@line 895 "libraries/lodepng/lodepng.cpp"
  $exitcond9 = ($5|0)==(256); //@line 895 "libraries/lodepng/lodepng.cpp"
  if ($exitcond9) {
   break;
  } else {
   $i$16 = $5;
  }
 }
 $6 = ((($0)) + 1024|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$6>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $7 = ((($0)) + 1028|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$7>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $8 = ((($0)) + 1032|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$8>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $9 = ((($0)) + 1036|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$9>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $10 = ((($0)) + 1040|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$10>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $11 = ((($0)) + 1044|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$11>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $12 = ((($0)) + 1048|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$12>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $13 = ((($0)) + 1052|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$13>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $14 = ((($0)) + 1056|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$14>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $15 = ((($0)) + 1060|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$15>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $16 = ((($0)) + 1064|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$16>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $17 = ((($0)) + 1068|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$17>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $18 = ((($0)) + 1072|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$18>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $19 = ((($0)) + 1076|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$19>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $20 = ((($0)) + 1080|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$20>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $21 = ((($0)) + 1084|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$21>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $22 = ((($0)) + 1088|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$22>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $23 = ((($0)) + 1092|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$23>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $24 = ((($0)) + 1096|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$24>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $25 = ((($0)) + 1100|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$25>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $26 = ((($0)) + 1104|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$26>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $27 = ((($0)) + 1108|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$27>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $28 = ((($0)) + 1112|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$28>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $29 = ((($0)) + 1116|0); //@line 896 "libraries/lodepng/lodepng.cpp"
 HEAP32[$29>>2] = 7; //@line 896 "libraries/lodepng/lodepng.cpp"
 $30 = ((($0)) + 1120|0); //@line 897 "libraries/lodepng/lodepng.cpp"
 HEAP32[$30>>2] = 8; //@line 897 "libraries/lodepng/lodepng.cpp"
 $31 = ((($0)) + 1124|0); //@line 897 "libraries/lodepng/lodepng.cpp"
 HEAP32[$31>>2] = 8; //@line 897 "libraries/lodepng/lodepng.cpp"
 $32 = ((($0)) + 1128|0); //@line 897 "libraries/lodepng/lodepng.cpp"
 HEAP32[$32>>2] = 8; //@line 897 "libraries/lodepng/lodepng.cpp"
 $33 = ((($0)) + 1132|0); //@line 897 "libraries/lodepng/lodepng.cpp"
 HEAP32[$33>>2] = 8; //@line 897 "libraries/lodepng/lodepng.cpp"
 $34 = ((($0)) + 1136|0); //@line 897 "libraries/lodepng/lodepng.cpp"
 HEAP32[$34>>2] = 8; //@line 897 "libraries/lodepng/lodepng.cpp"
 $35 = ((($0)) + 1140|0); //@line 897 "libraries/lodepng/lodepng.cpp"
 HEAP32[$35>>2] = 8; //@line 897 "libraries/lodepng/lodepng.cpp"
 $36 = ((($0)) + 1144|0); //@line 897 "libraries/lodepng/lodepng.cpp"
 HEAP32[$36>>2] = 8; //@line 897 "libraries/lodepng/lodepng.cpp"
 $37 = ((($0)) + 1148|0); //@line 897 "libraries/lodepng/lodepng.cpp"
 HEAP32[$37>>2] = 8; //@line 897 "libraries/lodepng/lodepng.cpp"
 (__ZL27HuffmanTree_makeFromLengthsP11HuffmanTreePKjjj($tree,$0,288,15)|0); //@line 899 "libraries/lodepng/lodepng.cpp"
 _free($0); //@line 76 "libraries/lodepng/lodepng.cpp"
 return;
}
function __ZL27HuffmanTree_makeFromLengthsP11HuffmanTreePKjjj($tree,$bitlen,$numcodes,$maxbitlen) {
 $tree = $tree|0;
 $bitlen = $bitlen|0;
 $numcodes = $numcodes|0;
 $maxbitlen = $maxbitlen|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $numcodes << 2; //@line 659 "libraries/lodepng/lodepng.cpp"
 $1 = (_malloc($0)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
 $2 = ((($tree)) + 8|0); //@line 659 "libraries/lodepng/lodepng.cpp"
 HEAP32[$2>>2] = $1; //@line 659 "libraries/lodepng/lodepng.cpp"
 $3 = ($1|0)==(0|0); //@line 660 "libraries/lodepng/lodepng.cpp"
 if ($3) {
  $$0 = 83;
  return ($$0|0); //@line 665 "libraries/lodepng/lodepng.cpp"
 }
 $4 = ($numcodes|0)==(0); //@line 661 "libraries/lodepng/lodepng.cpp"
 if (!($4)) {
  $i$01 = 0;
  while(1) {
   $5 = (($bitlen) + ($i$01<<2)|0); //@line 661 "libraries/lodepng/lodepng.cpp"
   $6 = HEAP32[$5>>2]|0; //@line 661 "libraries/lodepng/lodepng.cpp"
   $7 = (($1) + ($i$01<<2)|0); //@line 661 "libraries/lodepng/lodepng.cpp"
   HEAP32[$7>>2] = $6; //@line 661 "libraries/lodepng/lodepng.cpp"
   $8 = (($i$01) + 1)|0; //@line 661 "libraries/lodepng/lodepng.cpp"
   $9 = ($8|0)==($numcodes|0); //@line 661 "libraries/lodepng/lodepng.cpp"
   if ($9) {
    break;
   } else {
    $i$01 = $8;
   }
  }
 }
 $10 = ((($tree)) + 16|0); //@line 662 "libraries/lodepng/lodepng.cpp"
 HEAP32[$10>>2] = $numcodes; //@line 662 "libraries/lodepng/lodepng.cpp"
 $11 = ((($tree)) + 12|0); //@line 663 "libraries/lodepng/lodepng.cpp"
 HEAP32[$11>>2] = $maxbitlen; //@line 663 "libraries/lodepng/lodepng.cpp"
 $12 = (__ZL28HuffmanTree_makeFromLengths2P11HuffmanTree($tree)|0); //@line 664 "libraries/lodepng/lodepng.cpp"
 $$0 = $12;
 return ($$0|0); //@line 665 "libraries/lodepng/lodepng.cpp"
}
function __ZL28HuffmanTree_makeFromLengths2P11HuffmanTree($tree) {
 $tree = $tree|0;
 var $$0 = 0, $$lcssa = 0, $$mask$i = 0, $$mask21$i = 0, $$pre = 0, $$pre$i = 0, $$pre$phi$iZ2D = 0, $$pre23$i = 0, $$pre24 = 0, $$pre24$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $bits$021 = 0, $bits$119 = 0, $blcount$sroa$0$0 = 0, $blcount$sroa$0$1912 = 0, $i$013$i = 0, $n$018 = 0, $n$019$i = 0, $n$118$i = 0, $n$210$i = 0, $nextcode$sroa$0$0 = 0, $nextcode$sroa$0$313 = 0, $nodefilled$016$i = 0, $nodefilled$1$lcssa$i = 0, $nodefilled$111$i = 0, $nodefilled$3$ph$i = 0, $nodefilled$3$ph$i$lcssa = 0, $treepos$017$i = 0, $treepos$1$lcssa$i = 0;
 var $treepos$112$i = 0, $treepos$3$ph$i = 0, $treepos$3$ph$i$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($tree)) + 16|0); //@line 620 "libraries/lodepng/lodepng.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 620 "libraries/lodepng/lodepng.cpp"
 $2 = $1 << 2; //@line 620 "libraries/lodepng/lodepng.cpp"
 $3 = (_malloc($2)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
 $4 = ((($tree)) + 4|0); //@line 620 "libraries/lodepng/lodepng.cpp"
 HEAP32[$4>>2] = $3; //@line 620 "libraries/lodepng/lodepng.cpp"
 $5 = ($3|0)==(0|0); //@line 621 "libraries/lodepng/lodepng.cpp"
 $6 = ((($tree)) + 12|0); //@line 623 "libraries/lodepng/lodepng.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 623 "libraries/lodepng/lodepng.cpp"
 $8 = (($7) + 1)|0; //@line 623 "libraries/lodepng/lodepng.cpp"
 $9 = $8 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
 $10 = ($9|0)==(0); //@line 155 "libraries/lodepng/lodepng.cpp"
 if ($10) {
  $blcount$sroa$0$0 = 0;
  label = 4;
 } else {
  $11 = (_realloc(0,$9)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
  $12 = ($11|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
  if ($12) {
   $blcount$sroa$0$1912 = 0;$nextcode$sroa$0$313 = 0;
  } else {
   $blcount$sroa$0$0 = $11;
   label = 4;
  }
 }
 do {
  if ((label|0) == 4) {
   $13 = ($8|0)==(0); //@line 182 "libraries/lodepng/lodepng.cpp"
   if (!($13)) {
    _memset(($blcount$sroa$0$0|0),0,($9|0))|0; //@line 182 "libraries/lodepng/lodepng.cpp"
   }
   $14 = HEAP32[$6>>2]|0; //@line 624 "libraries/lodepng/lodepng.cpp"
   $15 = (($14) + 1)|0; //@line 624 "libraries/lodepng/lodepng.cpp"
   $16 = $15 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
   $17 = ($16|0)==(0); //@line 155 "libraries/lodepng/lodepng.cpp"
   if ($17) {
    $nextcode$sroa$0$0 = 0;
   } else {
    $18 = (_realloc(0,$16)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $19 = ($18|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
    if ($19) {
     $blcount$sroa$0$1912 = $blcount$sroa$0$0;$nextcode$sroa$0$313 = 0;
     break;
    }
    $nextcode$sroa$0$0 = $18;
   }
   $20 = ($15|0)==(0); //@line 182 "libraries/lodepng/lodepng.cpp"
   if (!($20)) {
    _memset(($nextcode$sroa$0$0|0),0,($16|0))|0; //@line 182 "libraries/lodepng/lodepng.cpp"
   }
   if ($5) {
    $blcount$sroa$0$1912 = $blcount$sroa$0$0;$nextcode$sroa$0$313 = $nextcode$sroa$0$0;
   } else {
    $21 = HEAP32[$0>>2]|0; //@line 637 "libraries/lodepng/lodepng.cpp"
    $22 = ($21|0)==(0); //@line 630 "libraries/lodepng/lodepng.cpp"
    if (!($22)) {
     $23 = ((($tree)) + 8|0); //@line 630 "libraries/lodepng/lodepng.cpp"
     $24 = HEAP32[$23>>2]|0; //@line 630 "libraries/lodepng/lodepng.cpp"
     $bits$021 = 0;
     while(1) {
      $27 = (($24) + ($bits$021<<2)|0); //@line 630 "libraries/lodepng/lodepng.cpp"
      $28 = HEAP32[$27>>2]|0; //@line 630 "libraries/lodepng/lodepng.cpp"
      $29 = (($blcount$sroa$0$0) + ($28<<2)|0); //@line 630 "libraries/lodepng/lodepng.cpp"
      $30 = HEAP32[$29>>2]|0; //@line 630 "libraries/lodepng/lodepng.cpp"
      $31 = (($30) + 1)|0; //@line 630 "libraries/lodepng/lodepng.cpp"
      HEAP32[$29>>2] = $31; //@line 630 "libraries/lodepng/lodepng.cpp"
      $32 = (($bits$021) + 1)|0; //@line 630 "libraries/lodepng/lodepng.cpp"
      $33 = ($32|0)==($21|0); //@line 630 "libraries/lodepng/lodepng.cpp"
      if ($33) {
       break;
      } else {
       $bits$021 = $32;
      }
     }
    }
    $25 = HEAP32[$6>>2]|0; //@line 632 "libraries/lodepng/lodepng.cpp"
    $26 = ($25|0)==(0); //@line 632 "libraries/lodepng/lodepng.cpp"
    if (!($26)) {
     $$pre = HEAP32[$nextcode$sroa$0$0>>2]|0; //@line 634 "libraries/lodepng/lodepng.cpp"
     $40 = $$pre;$bits$119 = 1;
     while(1) {
      $36 = (($bits$119) + -1)|0; //@line 634 "libraries/lodepng/lodepng.cpp"
      $37 = (($blcount$sroa$0$0) + ($36<<2)|0); //@line 634 "libraries/lodepng/lodepng.cpp"
      $38 = HEAP32[$37>>2]|0; //@line 634 "libraries/lodepng/lodepng.cpp"
      $39 = (($38) + ($40))|0; //@line 634 "libraries/lodepng/lodepng.cpp"
      $41 = $39 << 1; //@line 634 "libraries/lodepng/lodepng.cpp"
      $42 = (($nextcode$sroa$0$0) + ($bits$119<<2)|0); //@line 634 "libraries/lodepng/lodepng.cpp"
      HEAP32[$42>>2] = $41; //@line 634 "libraries/lodepng/lodepng.cpp"
      $43 = (($bits$119) + 1)|0; //@line 632 "libraries/lodepng/lodepng.cpp"
      $44 = ($43>>>0)>($25>>>0); //@line 632 "libraries/lodepng/lodepng.cpp"
      if ($44) {
       break;
      } else {
       $40 = $41;$bits$119 = $43;
      }
     }
    }
    if (!($22)) {
     $34 = ((($tree)) + 8|0); //@line 639 "libraries/lodepng/lodepng.cpp"
     $35 = HEAP32[$34>>2]|0; //@line 639 "libraries/lodepng/lodepng.cpp"
     $109 = $21;$n$018 = 0;
     while(1) {
      $45 = (($35) + ($n$018<<2)|0); //@line 639 "libraries/lodepng/lodepng.cpp"
      $46 = HEAP32[$45>>2]|0; //@line 639 "libraries/lodepng/lodepng.cpp"
      $47 = ($46|0)==(0); //@line 639 "libraries/lodepng/lodepng.cpp"
      if ($47) {
       $55 = $109;
      } else {
       $48 = (($nextcode$sroa$0$0) + ($46<<2)|0); //@line 639 "libraries/lodepng/lodepng.cpp"
       $49 = HEAP32[$48>>2]|0; //@line 639 "libraries/lodepng/lodepng.cpp"
       $50 = (($49) + 1)|0; //@line 639 "libraries/lodepng/lodepng.cpp"
       HEAP32[$48>>2] = $50; //@line 639 "libraries/lodepng/lodepng.cpp"
       $51 = HEAP32[$4>>2]|0; //@line 639 "libraries/lodepng/lodepng.cpp"
       $52 = (($51) + ($n$018<<2)|0); //@line 639 "libraries/lodepng/lodepng.cpp"
       HEAP32[$52>>2] = $49; //@line 639 "libraries/lodepng/lodepng.cpp"
       $$pre24 = HEAP32[$0>>2]|0; //@line 637 "libraries/lodepng/lodepng.cpp"
       $55 = $$pre24;
      }
      $53 = (($n$018) + 1)|0; //@line 637 "libraries/lodepng/lodepng.cpp"
      $54 = ($53|0)==($55|0); //@line 637 "libraries/lodepng/lodepng.cpp"
      if ($54) {
       break;
      } else {
       $109 = $55;$n$018 = $53;
      }
     }
    }
    _free($blcount$sroa$0$0); //@line 76 "libraries/lodepng/lodepng.cpp"
    _free($nextcode$sroa$0$0); //@line 76 "libraries/lodepng/lodepng.cpp"
    $56 = HEAP32[$0>>2]|0; //@line 564 "libraries/lodepng/lodepng.cpp"
    $57 = $56 << 3; //@line 551 "libraries/lodepng/lodepng.cpp"
    $58 = (_malloc($57)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
    HEAP32[$tree>>2] = $58; //@line 551 "libraries/lodepng/lodepng.cpp"
    $59 = ($58|0)==(0|0); //@line 552 "libraries/lodepng/lodepng.cpp"
    if ($59) {
     $$0 = 83;
     return ($$0|0); //@line 648 "libraries/lodepng/lodepng.cpp"
    }
    $$mask$i = $56 & 2147483647; //@line 564 "libraries/lodepng/lodepng.cpp"
    $60 = ($$mask$i|0)==(0); //@line 564 "libraries/lodepng/lodepng.cpp"
    if (!($60)) {
     $61 = $56 << 1; //@line 564 "libraries/lodepng/lodepng.cpp"
     $n$019$i = 0;
     while(1) {
      $65 = (($58) + ($n$019$i<<2)|0); //@line 566 "libraries/lodepng/lodepng.cpp"
      HEAP32[$65>>2] = 32767; //@line 566 "libraries/lodepng/lodepng.cpp"
      $66 = (($n$019$i) + 1)|0; //@line 564 "libraries/lodepng/lodepng.cpp"
      $67 = ($66>>>0)<($61>>>0); //@line 564 "libraries/lodepng/lodepng.cpp"
      if ($67) {
       $n$019$i = $66;
      } else {
       break;
      }
     }
    }
    $62 = ($56|0)==(0); //@line 569 "libraries/lodepng/lodepng.cpp"
    if ($62) {
     $$0 = 0;
     return ($$0|0); //@line 648 "libraries/lodepng/lodepng.cpp"
    }
    $63 = ((($tree)) + 8|0); //@line 571 "libraries/lodepng/lodepng.cpp"
    $64 = HEAP32[$63>>2]|0; //@line 571 "libraries/lodepng/lodepng.cpp"
    $110 = $56;$n$118$i = 0;$nodefilled$016$i = 0;$treepos$017$i = 0;
    L46: while(1) {
     $68 = (($64) + ($n$118$i<<2)|0); //@line 571 "libraries/lodepng/lodepng.cpp"
     $69 = HEAP32[$68>>2]|0; //@line 571 "libraries/lodepng/lodepng.cpp"
     $70 = ($69|0)==(0); //@line 571 "libraries/lodepng/lodepng.cpp"
     if ($70) {
      $101 = $110;$nodefilled$1$lcssa$i = $nodefilled$016$i;$treepos$1$lcssa$i = $treepos$017$i;
     } else {
      $71 = HEAP32[$4>>2]|0; //@line 573 "libraries/lodepng/lodepng.cpp"
      $72 = (($71) + ($n$118$i<<2)|0); //@line 573 "libraries/lodepng/lodepng.cpp"
      $78 = $69;$i$013$i = 0;$nodefilled$111$i = $nodefilled$016$i;$treepos$112$i = $treepos$017$i;
      while(1) {
       $75 = HEAP32[$72>>2]|0; //@line 573 "libraries/lodepng/lodepng.cpp"
       $76 = $i$013$i ^ -1; //@line 573 "libraries/lodepng/lodepng.cpp"
       $77 = (($78) + ($76))|0; //@line 573 "libraries/lodepng/lodepng.cpp"
       $79 = $75 >>> $77; //@line 573 "libraries/lodepng/lodepng.cpp"
       $80 = $79 & 1; //@line 573 "libraries/lodepng/lodepng.cpp"
       $81 = ($treepos$112$i|0)<(0); //@line 575 "libraries/lodepng/lodepng.cpp"
       if ($81) {
        $$0 = 55;
        label = 47;
        break L46;
       }
       $82 = (($treepos$112$i) + 2)|0; //@line 575 "libraries/lodepng/lodepng.cpp"
       $83 = HEAP32[$0>>2]|0; //@line 575 "libraries/lodepng/lodepng.cpp"
       $84 = ($82>>>0)>($83>>>0); //@line 575 "libraries/lodepng/lodepng.cpp"
       if ($84) {
        $$0 = 55;
        label = 47;
        break L46;
       }
       $85 = $treepos$112$i << 1; //@line 576 "libraries/lodepng/lodepng.cpp"
       $86 = $80 | $85; //@line 576 "libraries/lodepng/lodepng.cpp"
       $87 = (($58) + ($86<<2)|0); //@line 576 "libraries/lodepng/lodepng.cpp"
       $88 = HEAP32[$87>>2]|0; //@line 576 "libraries/lodepng/lodepng.cpp"
       $89 = ($88|0)==(32767); //@line 576 "libraries/lodepng/lodepng.cpp"
       do {
        if ($89) {
         $90 = (($i$013$i) + 1)|0; //@line 578 "libraries/lodepng/lodepng.cpp"
         $91 = ($90|0)==($78|0); //@line 578 "libraries/lodepng/lodepng.cpp"
         if ($91) {
          HEAP32[$87>>2] = $n$118$i; //@line 580 "libraries/lodepng/lodepng.cpp"
          $$pre$phi$iZ2D = $78;$nodefilled$3$ph$i = $nodefilled$111$i;$treepos$3$ph$i = 0;
          break;
         } else {
          $92 = (($nodefilled$111$i) + 1)|0; //@line 587 "libraries/lodepng/lodepng.cpp"
          $93 = (($83) + ($92))|0; //@line 589 "libraries/lodepng/lodepng.cpp"
          HEAP32[$87>>2] = $93; //@line 589 "libraries/lodepng/lodepng.cpp"
          $$pre$phi$iZ2D = $90;$nodefilled$3$ph$i = $92;$treepos$3$ph$i = $92;
          break;
         }
        } else {
         $94 = (($88) - ($83))|0; //@line 593 "libraries/lodepng/lodepng.cpp"
         $$pre24$i = (($i$013$i) + 1)|0; //@line 571 "libraries/lodepng/lodepng.cpp"
         $$pre$phi$iZ2D = $$pre24$i;$nodefilled$3$ph$i = $nodefilled$111$i;$treepos$3$ph$i = $94;
        }
       } while(0);
       $95 = HEAP32[$63>>2]|0; //@line 571 "libraries/lodepng/lodepng.cpp"
       $96 = (($95) + ($n$118$i<<2)|0); //@line 571 "libraries/lodepng/lodepng.cpp"
       $97 = HEAP32[$96>>2]|0; //@line 571 "libraries/lodepng/lodepng.cpp"
       $98 = ($97|0)==($$pre$phi$iZ2D|0); //@line 571 "libraries/lodepng/lodepng.cpp"
       if ($98) {
        $nodefilled$3$ph$i$lcssa = $nodefilled$3$ph$i;$treepos$3$ph$i$lcssa = $treepos$3$ph$i;
        break;
       } else {
        $78 = $97;$i$013$i = $$pre$phi$iZ2D;$nodefilled$111$i = $nodefilled$3$ph$i;$treepos$112$i = $treepos$3$ph$i;
       }
      }
      $$pre$i = HEAP32[$0>>2]|0; //@line 569 "libraries/lodepng/lodepng.cpp"
      $101 = $$pre$i;$nodefilled$1$lcssa$i = $nodefilled$3$ph$i$lcssa;$treepos$1$lcssa$i = $treepos$3$ph$i$lcssa;
     }
     $99 = (($n$118$i) + 1)|0; //@line 569 "libraries/lodepng/lodepng.cpp"
     $100 = ($99>>>0)<($101>>>0); //@line 569 "libraries/lodepng/lodepng.cpp"
     if ($100) {
      $110 = $101;$n$118$i = $99;$nodefilled$016$i = $nodefilled$1$lcssa$i;$treepos$017$i = $treepos$1$lcssa$i;
     } else {
      $$lcssa = $101;
      label = 31;
      break;
     }
    }
    if ((label|0) == 31) {
     $$mask21$i = $$lcssa & 2147483647; //@line 597 "libraries/lodepng/lodepng.cpp"
     $73 = ($$mask21$i|0)==(0); //@line 597 "libraries/lodepng/lodepng.cpp"
     if ($73) {
      $$0 = 0;
      return ($$0|0); //@line 648 "libraries/lodepng/lodepng.cpp"
     }
     $74 = HEAP32[$tree>>2]|0; //@line 599 "libraries/lodepng/lodepng.cpp"
     $111 = $$lcssa;$n$210$i = 0;
     while(1) {
      $102 = (($74) + ($n$210$i<<2)|0); //@line 599 "libraries/lodepng/lodepng.cpp"
      $103 = HEAP32[$102>>2]|0; //@line 599 "libraries/lodepng/lodepng.cpp"
      $104 = ($103|0)==(32767); //@line 599 "libraries/lodepng/lodepng.cpp"
      if ($104) {
       HEAP32[$102>>2] = 0; //@line 599 "libraries/lodepng/lodepng.cpp"
       $$pre23$i = HEAP32[$0>>2]|0; //@line 597 "libraries/lodepng/lodepng.cpp"
       $107 = $$pre23$i;
      } else {
       $107 = $111;
      }
      $105 = (($n$210$i) + 1)|0; //@line 597 "libraries/lodepng/lodepng.cpp"
      $106 = $107 << 1; //@line 597 "libraries/lodepng/lodepng.cpp"
      $108 = ($105>>>0)<($106>>>0); //@line 597 "libraries/lodepng/lodepng.cpp"
      if ($108) {
       $111 = $107;$n$210$i = $105;
      } else {
       $$0 = 0;
       break;
      }
     }
     return ($$0|0); //@line 648 "libraries/lodepng/lodepng.cpp"
    }
    else if ((label|0) == 47) {
     return ($$0|0); //@line 648 "libraries/lodepng/lodepng.cpp"
    }
   }
  }
 } while(0);
 _free($blcount$sroa$0$1912); //@line 76 "libraries/lodepng/lodepng.cpp"
 _free($nextcode$sroa$0$313); //@line 76 "libraries/lodepng/lodepng.cpp"
 $$0 = 83;
 return ($$0|0); //@line 648 "libraries/lodepng/lodepng.cpp"
}
function __ZL25generateFixedDistanceTreeP11HuffmanTree($tree) {
 $tree = $tree|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, dest = 0, label = 0;
 var sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = (_malloc(128)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
 $1 = ($0|0)==(0|0); //@line 910 "libraries/lodepng/lodepng.cpp"
 if ($1) {
  return;
 }
 HEAP32[$0>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $2 = ((($0)) + 4|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$2>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $3 = ((($0)) + 8|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$3>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $4 = ((($0)) + 12|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$4>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $5 = ((($0)) + 16|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$5>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $6 = ((($0)) + 20|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$6>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $7 = ((($0)) + 24|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$7>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $8 = ((($0)) + 28|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$8>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $9 = ((($0)) + 32|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$9>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $10 = ((($0)) + 36|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$10>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $11 = ((($0)) + 40|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$11>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $12 = ((($0)) + 44|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$12>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $13 = ((($0)) + 48|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$13>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $14 = ((($0)) + 52|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$14>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $15 = ((($0)) + 56|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$15>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $16 = ((($0)) + 60|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$16>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $17 = ((($0)) + 64|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$17>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $18 = ((($0)) + 68|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$18>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $19 = ((($0)) + 72|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$19>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $20 = ((($0)) + 76|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$20>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $21 = ((($0)) + 80|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$21>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $22 = ((($0)) + 84|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$22>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $23 = ((($0)) + 88|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$23>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $24 = ((($0)) + 92|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$24>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $25 = ((($0)) + 96|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$25>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $26 = ((($0)) + 100|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$26>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $27 = ((($0)) + 104|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$27>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $28 = ((($0)) + 108|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$28>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $29 = ((($0)) + 112|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$29>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $30 = ((($0)) + 116|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$30>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $31 = ((($0)) + 120|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$31>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $32 = ((($0)) + 124|0); //@line 913 "libraries/lodepng/lodepng.cpp"
 HEAP32[$32>>2] = 5; //@line 913 "libraries/lodepng/lodepng.cpp"
 $33 = (_malloc(128)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
 $34 = ((($tree)) + 8|0); //@line 659 "libraries/lodepng/lodepng.cpp"
 HEAP32[$34>>2] = $33; //@line 659 "libraries/lodepng/lodepng.cpp"
 $35 = ($33|0)==(0|0); //@line 660 "libraries/lodepng/lodepng.cpp"
 if (!($35)) {
  dest=$33; src=$0; stop=dest+128|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 661 "libraries/lodepng/lodepng.cpp"
  $36 = ((($tree)) + 16|0); //@line 662 "libraries/lodepng/lodepng.cpp"
  HEAP32[$36>>2] = 32; //@line 662 "libraries/lodepng/lodepng.cpp"
  $37 = ((($tree)) + 12|0); //@line 663 "libraries/lodepng/lodepng.cpp"
  HEAP32[$37>>2] = 15; //@line 663 "libraries/lodepng/lodepng.cpp"
  (__ZL28HuffmanTree_makeFromLengths2P11HuffmanTree($tree)|0); //@line 664 "libraries/lodepng/lodepng.cpp"
 }
 _free($0); //@line 76 "libraries/lodepng/lodepng.cpp"
 return;
}
function __Z15lodepng_deflatePPhPjPKhjPK23LodePNGCompressSettings($out,$outsize,$in,$insize,$settings) {
 $out = $out|0;
 $outsize = $outsize|0;
 $in = $in|0;
 $insize = $insize|0;
 $settings = $settings|0;
 var $$$i = 0, $$$i$i = 0, $$$i13$i = 0, $$0$i = 0, $$01$i$i$i = 0, $$01$i$i$i$lcssa = 0, $$01$i18$i$i = 0, $$01$i18$i$i$lcssa = 0, $$1$i = 0, $$2$i$i = 0, $$3$i$i = 0, $$i$7$i$i = 0, $$lcssa = 0, $$lcssa72 = 0, $$lcssa74 = 0, $$lcssa76 = 0, $$lcssa78 = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0;
 var $$pre$i$i$i = 0, $$pre$i$i$i$i = 0, $$pre$i$i$i$i$i = 0, $$pre$i$i$i$i24$i = 0, $$pre$i$i$i17$i$i = 0, $$pre$i$i$i184$i$i = 0, $$pre$i$i$i19$i = 0, $$pre$i$i129$i$i = 0, $$pre$i$i138$i$i = 0, $$pre$i$i14$i$i = 0, $$pre$i$i147$1$i$i = 0, $$pre$i$i147$2$i$i = 0, $$pre$i$i147$i$i = 0, $$pre$i$i153$i$i = 0, $$pre$i$i153268$i$i = 0, $$pre$i$i153269$i$i = 0, $$pre$i$i157$1$i$i = 0, $$pre$i$i157$i$i = 0, $$pre$i$i166$1$i$i = 0, $$pre$i$i166$2$i$i = 0;
 var $$pre$i$i166$i$i = 0, $$pre$i$i175$i$i = 0, $$pre$i$i181$i$i = 0, $$pre$i$i26$i$i = 0, $$pre$i$i4$i = 0, $$pre$i10$i$i = 0, $$pre$i110$i$i = 0, $$pre$i114$i$i = 0, $$pre$i119$i$i = 0, $$pre$i12$i$i = 0, $$pre$i144$i$i = 0, $$pre$i154$i$i = 0, $$pre$i17$i$i = 0, $$pre$i2$i$i = 0, $$pre$i22$i$i = 0, $$pre$i5$i$i = 0, $$pre$i7$i$i = 0, $$pre25 = 0, $$pre26 = 0, $$pre27 = 0;
 var $$pre270$i$i = 0, $$pre28 = 0, $$pre29 = 0, $$pre30 = 0, $$pre31 = 0, $$pre32 = 0, $$pre33 = 0, $$pre34 = 0, $$pre56$i = 0, $$pre57$i = 0, $$pre58$i = 0, $$pre59$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0;
 var $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0;
 var $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0;
 var $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0;
 var $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0;
 var $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0;
 var $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0;
 var $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0;
 var $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0;
 var $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0;
 var $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0;
 var $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0;
 var $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0;
 var $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0;
 var $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0;
 var $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0;
 var $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0;
 var $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0;
 var $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0;
 var $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0;
 var $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0;
 var $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0;
 var $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0;
 var $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0;
 var $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0;
 var $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0;
 var $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0;
 var $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0;
 var $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0;
 var $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0;
 var $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0;
 var $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0;
 var $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0;
 var $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0;
 var $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0;
 var $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0;
 var $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0;
 var $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0;
 var $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0;
 var $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0;
 var $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0;
 var $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0;
 var $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0;
 var $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0;
 var $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0;
 var $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0;
 var $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0;
 var $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0;
 var $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0;
 var $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0;
 var $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0;
 var $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0;
 var $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0;
 var $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0;
 var $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0;
 var $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0;
 var $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0;
 var $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0;
 var $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0;
 var $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $HCLEN$0$i$i = 0, $HCLEN$0$i$i$lcssa = 0, $bitlen_cl$sroa$0$1191$ph$i$i = 0, $bitlen_cl$sroa$0$2$ph$i$i = 0, $bitlen_cl$sroa$0$2$ph$i$i$lcssa = 0, $bitlen_cl$sroa$0$6$i$i = 0;
 var $bitlen_cl$sroa$13$1$i$i = 0, $bitlen_cl$sroa$13$1$i$i$lcssa = 0, $bitlen_cl$sroa$13$1$ph$i$i = 0, $bitlen_cl$sroa$19$1189$ph$i$i = 0, $bitlen_cl$sroa$19$2$ph$i$i = 0, $bitlen_lld$sroa$0$0$lcssa$i$i = 0, $bitlen_lld$sroa$0$0243$i$i = 0, $bitlen_lld$sroa$0$1$i$i = 0, $bitlen_lld$sroa$0$2$i$i = 0, $bitlen_lld$sroa$0$3$lcssa$i$i = 0, $bitlen_lld$sroa$0$3236$i$i = 0, $bitlen_lld$sroa$0$4$i$i = 0, $bitlen_lld$sroa$0$5$i$i = 0, $bitlen_lld$sroa$0$6$i$i = 0, $bitlen_lld$sroa$13$0$lcssa$i$i = 0, $bitlen_lld$sroa$13$0242$i$i = 0, $bitlen_lld$sroa$13$1$i$i = 0, $bitlen_lld$sroa$13$2$lcssa$i$i = 0, $bitlen_lld$sroa$13$2235$i$i = 0, $bitlen_lld$sroa$13$3$i$i = 0;
 var $bitlen_lld$sroa$21$0$lcssa$i$i = 0, $bitlen_lld$sroa$21$0241$i$i = 0, $bitlen_lld$sroa$21$1$i$i = 0, $bitlen_lld$sroa$21$2$i$i = 0, $bitlen_lld$sroa$21$3234$i$i = 0, $bitlen_lld$sroa$21$4$i$i = 0, $bitlen_lld$sroa$21$5$i$i = 0, $bitlen_lld_e$sroa$0$0$lcssa$i$i = 0, $bitlen_lld_e$sroa$0$0229$i$i = 0, $bitlen_lld_e$sroa$0$1$i$i = 0, $bitlen_lld_e$sroa$0$10$i$i = 0, $bitlen_lld_e$sroa$0$12$lcssa$i$i = 0, $bitlen_lld_e$sroa$0$12$ph$i$i = 0, $bitlen_lld_e$sroa$0$12221$i$i = 0, $bitlen_lld_e$sroa$0$13$i$i = 0, $bitlen_lld_e$sroa$0$14$i$i = 0, $bitlen_lld_e$sroa$0$15$i$i = 0, $bitlen_lld_e$sroa$0$16$i$i = 0, $bitlen_lld_e$sroa$0$17$i$i = 0, $bitlen_lld_e$sroa$0$18$i$i = 0;
 var $bitlen_lld_e$sroa$0$19$i$i = 0, $bitlen_lld_e$sroa$0$2$i$i = 0, $bitlen_lld_e$sroa$0$21$i$i = 0, $bitlen_lld_e$sroa$0$22$i$i = 0, $bitlen_lld_e$sroa$0$24$i$i = 0, $bitlen_lld_e$sroa$0$25$i$i = 0, $bitlen_lld_e$sroa$0$3$i$i = 0, $bitlen_lld_e$sroa$0$5$i$i = 0, $bitlen_lld_e$sroa$0$6$i$i = 0, $bitlen_lld_e$sroa$0$7$i$i = 0, $bitlen_lld_e$sroa$0$9$i$i = 0, $bitlen_lld_e$sroa$40$0$lcssa$i$i = 0, $bitlen_lld_e$sroa$40$0228$i$i = 0, $bitlen_lld_e$sroa$40$1$i$i = 0, $bitlen_lld_e$sroa$40$10$i$i = 0, $bitlen_lld_e$sroa$40$12$i$i = 0, $bitlen_lld_e$sroa$40$14$i$i = 0, $bitlen_lld_e$sroa$40$3$i$i = 0, $bitlen_lld_e$sroa$40$5$i$i = 0, $bitlen_lld_e$sroa$40$7$lcssa$i$i = 0;
 var $bitlen_lld_e$sroa$40$7$ph$i$i = 0, $bitlen_lld_e$sroa$40$7220$i$i = 0, $bitlen_lld_e$sroa$40$8$i$i = 0, $bitlen_lld_e$sroa$40$9$i$i = 0, $bitlen_lld_e$sroa$64$0227$i$i = 0, $bitlen_lld_e$sroa$64$1$i$i = 0, $bitlen_lld_e$sroa$64$10$i$i = 0, $bitlen_lld_e$sroa$64$12$lcssa$i$i = 0, $bitlen_lld_e$sroa$64$12$ph$i$i = 0, $bitlen_lld_e$sroa$64$12219$i$i = 0, $bitlen_lld_e$sroa$64$13$i$i = 0, $bitlen_lld_e$sroa$64$14$i$i = 0, $bitlen_lld_e$sroa$64$15$i$i = 0, $bitlen_lld_e$sroa$64$16$i$i = 0, $bitlen_lld_e$sroa$64$17$i$i = 0, $bitlen_lld_e$sroa$64$18$i$i = 0, $bitlen_lld_e$sroa$64$19$i$i = 0, $bitlen_lld_e$sroa$64$2$i$i = 0, $bitlen_lld_e$sroa$64$21$i$i = 0, $bitlen_lld_e$sroa$64$22$i$i = 0;
 var $bitlen_lld_e$sroa$64$24$i$i = 0, $bitlen_lld_e$sroa$64$3$i$i = 0, $bitlen_lld_e$sroa$64$5$i$i = 0, $bitlen_lld_e$sroa$64$6$i$i = 0, $bitlen_lld_e$sroa$64$7$i$i = 0, $bitlen_lld_e$sroa$64$9$i$i = 0, $blocksize$0$i = 0, $bp$i = 0, $datapos$029$i$i = 0, $datapos$1$lcssa$i$i = 0, $datapos$127$i$i = 0, $error$022$i$i = 0, $error$1$i = 0, $error$1$i$i = 0, $error$1$i$lcssa = 0, $exitcond$i$i = 0, $exitcond$i12$i = 0, $frequencies_cl$sroa$0$2$i$i = 0, $frequencies_d$sroa$0$2$i$i = 0, $frequencies_ll$sroa$0$2$i$i = 0;
 var $hash$i = 0, $i$01$i$i$i = 0, $i$01$i$i$i$i = 0, $i$01$i$i$i23$i = 0, $i$01$i$i16$i$i = 0, $i$01$i$i183$i$i = 0, $i$01$i128$i$i = 0, $i$01$i137$i$i = 0, $i$01$i174$i$i = 0, $i$024$i$i = 0, $i$0253$i$i = 0, $i$030$i$i = 0, $i$042$i = 0, $i$10209$i$i = 0, $i$11208$i$i = 0, $i$12$i$i = 0, $i$1250$i$i = 0, $i$2$i$i = 0, $i$27$i$i = 0, $i$3244$i$i = 0;
 var $i$4237$i$i = 0, $i$45$i$i = 0, $i$5230$i$i = 0, $i$6$i$i = 0, $i$7213$i$i = 0, $i$9211$i$i = 0, $insize$$i = 0, $j$0$lcssa$i$i = 0, $j$0215$i$i = 0, $j$028$i$i = 0, $j$2$i$i = 0, $j$3$i$i = 0, $k$0222$i$i = 0, $lz77_encoded$i$i = 0, $lz77_encoded$i9$i = 0, $not$$i$i = 0, $or$cond$i = 0, $or$cond$i$i = 0, $or$cond$i$i$i = 0, $or$cond$i15$i = 0;
 var $or$cond$i19$i$i = 0, $or$cond$i2$i = 0, $or$cond12$i$i = 0, $or$cond13$i$i = 0, $or$cond14$i$i = 0, $tree_cl$i$i = 0, $tree_cl$idx$val$i$i = 0, $tree_cl$idx6$val$i$i = 0, $tree_cl$idx7$val$i$i = 0, $tree_d$i$i = 0, $tree_d$i11$i = 0, $tree_d$idx$val$i$i = 0, $tree_ll$i$i = 0, $tree_ll$i10$i = 0, $tree_ll$idx$val$i$i = 0, $tree_ll$idx$val$i29$i = 0, $tree_ll$idx1$val$i$i = 0, $tree_ll$idx2$val$i$i = 0, $tree_ll$idx3$val$i$i = 0, $tree_ll$idx5$val$i$i = 0;
 var $tree_ll$idx8$val$i$i = 0, $v = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lz77_encoded$i9$i = sp + 156|0;
 $tree_ll$i10$i = sp + 136|0;
 $tree_d$i11$i = sp + 116|0;
 $tree_cl$i$i = sp + 96|0;
 $tree_ll$i$i = sp + 76|0;
 $tree_d$i$i = sp + 56|0;
 $lz77_encoded$i$i = sp + 44|0;
 $bp$i = sp + 40|0;
 $hash$i = sp + 16|0;
 $v = sp;
 $0 = HEAP32[$out>>2]|0; //@line 2040 "libraries/lodepng/lodepng.cpp"
 $1 = HEAP32[$outsize>>2]|0; //@line 2040 "libraries/lodepng/lodepng.cpp"
 HEAP32[$v>>2] = $0; //@line 279 "libraries/lodepng/lodepng.cpp"
 $2 = ((($v)) + 4|0); //@line 280 "libraries/lodepng/lodepng.cpp"
 HEAP32[$2>>2] = $1; //@line 280 "libraries/lodepng/lodepng.cpp"
 $3 = ((($v)) + 8|0); //@line 280 "libraries/lodepng/lodepng.cpp"
 HEAP32[$3>>2] = $1; //@line 280 "libraries/lodepng/lodepng.cpp"
 HEAP32[$bp$i>>2] = 0; //@line 2000 "libraries/lodepng/lodepng.cpp"
 $4 = HEAP32[$settings>>2]|0; //@line 2003 "libraries/lodepng/lodepng.cpp"
 $5 = ($4>>>0)>(2); //@line 2003 "libraries/lodepng/lodepng.cpp"
 if ($5) {
  $$0$i = 61;
  $1240 = HEAP32[$v>>2]|0; //@line 2042 "libraries/lodepng/lodepng.cpp"
  HEAP32[$out>>2] = $1240; //@line 2042 "libraries/lodepng/lodepng.cpp"
  $1241 = HEAP32[$2>>2]|0; //@line 2043 "libraries/lodepng/lodepng.cpp"
  HEAP32[$outsize>>2] = $1241; //@line 2043 "libraries/lodepng/lodepng.cpp"
  STACKTOP = sp;return ($$0$i|0); //@line 2044 "libraries/lodepng/lodepng.cpp"
 }
 switch ($4|0) {
 case 0:  {
  $6 = (($insize) + 65534)|0; //@line 1636 "libraries/lodepng/lodepng.cpp"
  $7 = (($6>>>0) / 65535)&-1; //@line 1636 "libraries/lodepng/lodepng.cpp"
  $8 = ($6>>>0)<(65535); //@line 1638 "libraries/lodepng/lodepng.cpp"
  if ($8) {
   $$0$i = 0;
   $1240 = HEAP32[$v>>2]|0; //@line 2042 "libraries/lodepng/lodepng.cpp"
   HEAP32[$out>>2] = $1240; //@line 2042 "libraries/lodepng/lodepng.cpp"
   $1241 = HEAP32[$2>>2]|0; //@line 2043 "libraries/lodepng/lodepng.cpp"
   HEAP32[$outsize>>2] = $1241; //@line 2043 "libraries/lodepng/lodepng.cpp"
   STACKTOP = sp;return ($$0$i|0); //@line 2044 "libraries/lodepng/lodepng.cpp"
  }
  $9 = (($7) + -1)|0; //@line 1643 "libraries/lodepng/lodepng.cpp"
  $13 = $1;$15 = $1;$datapos$029$i$i = 0;$i$030$i$i = 0;
  while(1) {
   $10 = ($i$030$i$i|0)==($9|0); //@line 1643 "libraries/lodepng/lodepng.cpp"
   $11 = $10&1; //@line 1646 "libraries/lodepng/lodepng.cpp"
   $12 = (($13) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $14 = ($15>>>0)<($12>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($14) {
    $16 = $15 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $17 = ($16>>>0)<($12>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $18 = ($12*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $19 = $18 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $20 = $17 ? $12 : $19; //@line 227 "libraries/lodepng/lodepng.cpp"
    $21 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
    $22 = (_realloc($21,$20)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $23 = ($22|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if ($23) {
     $31 = $13;$33 = $15;
    } else {
     HEAP32[$3>>2] = $20; //@line 231 "libraries/lodepng/lodepng.cpp"
     HEAP32[$v>>2] = $22; //@line 232 "libraries/lodepng/lodepng.cpp"
     $25 = $22;
     label = 9;
    }
   } else {
    $$pre$i$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
    $25 = $$pre$i$i$i;
    label = 9;
   }
   if ((label|0) == 9) {
    label = 0;
    HEAP32[$2>>2] = $12; //@line 243 "libraries/lodepng/lodepng.cpp"
    $24 = (($25) + ($13)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$24>>0] = $11; //@line 289 "libraries/lodepng/lodepng.cpp"
    $$pre26 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
    $$pre27 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
    $31 = $$pre26;$33 = $$pre27;
   }
   $26 = (($insize) - ($datapos$029$i$i))|0; //@line 1650 "libraries/lodepng/lodepng.cpp"
   $27 = ($26>>>0)<(65535); //@line 1650 "libraries/lodepng/lodepng.cpp"
   $$$i$i = $27 ? $26 : 65535; //@line 1650 "libraries/lodepng/lodepng.cpp"
   $28 = (65535 - ($$$i$i))|0; //@line 1651 "libraries/lodepng/lodepng.cpp"
   $29 = $$$i$i&255; //@line 1653 "libraries/lodepng/lodepng.cpp"
   $30 = (($31) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $32 = ($33>>>0)<($30>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($32) {
    $34 = $33 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $35 = ($34>>>0)<($30>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $36 = ($30*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $37 = $36 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $38 = $35 ? $30 : $37; //@line 227 "libraries/lodepng/lodepng.cpp"
    $39 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
    $40 = (_realloc($39,$38)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $41 = ($40|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if ($41) {
     $47 = $31;$49 = $33;
    } else {
     HEAP32[$3>>2] = $38; //@line 231 "libraries/lodepng/lodepng.cpp"
     HEAP32[$v>>2] = $40; //@line 232 "libraries/lodepng/lodepng.cpp"
     $43 = $40;
     label = 14;
    }
   } else {
    $$pre$i2$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
    $43 = $$pre$i2$i$i;
    label = 14;
   }
   if ((label|0) == 14) {
    label = 0;
    HEAP32[$2>>2] = $30; //@line 243 "libraries/lodepng/lodepng.cpp"
    $42 = (($43) + ($31)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$42>>0] = $29; //@line 289 "libraries/lodepng/lodepng.cpp"
    $$pre28 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
    $$pre29 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
    $47 = $$pre28;$49 = $$pre29;
   }
   $44 = $$$i$i >>> 8; //@line 1654 "libraries/lodepng/lodepng.cpp"
   $45 = $44&255; //@line 1654 "libraries/lodepng/lodepng.cpp"
   $46 = (($47) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $48 = ($49>>>0)<($46>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($48) {
    $50 = $49 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $51 = ($50>>>0)<($46>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $52 = ($46*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $53 = $52 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $54 = $51 ? $46 : $53; //@line 227 "libraries/lodepng/lodepng.cpp"
    $55 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
    $56 = (_realloc($55,$54)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $57 = ($56|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if ($57) {
     $62 = $47;$64 = $49;
    } else {
     HEAP32[$3>>2] = $54; //@line 231 "libraries/lodepng/lodepng.cpp"
     HEAP32[$v>>2] = $56; //@line 232 "libraries/lodepng/lodepng.cpp"
     $59 = $56;
     label = 19;
    }
   } else {
    $$pre$i7$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
    $59 = $$pre$i7$i$i;
    label = 19;
   }
   if ((label|0) == 19) {
    label = 0;
    HEAP32[$2>>2] = $46; //@line 243 "libraries/lodepng/lodepng.cpp"
    $58 = (($59) + ($47)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$58>>0] = $45; //@line 289 "libraries/lodepng/lodepng.cpp"
    $$pre30 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
    $$pre31 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
    $62 = $$pre30;$64 = $$pre31;
   }
   $60 = $28&255; //@line 1655 "libraries/lodepng/lodepng.cpp"
   $61 = (($62) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $63 = ($64>>>0)<($61>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($63) {
    $65 = $64 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $66 = ($65>>>0)<($61>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $67 = ($61*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $68 = $67 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $69 = $66 ? $61 : $68; //@line 227 "libraries/lodepng/lodepng.cpp"
    $70 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
    $71 = (_realloc($70,$69)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $72 = ($71|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if ($72) {
     $78 = $62;$80 = $64;
    } else {
     HEAP32[$3>>2] = $69; //@line 231 "libraries/lodepng/lodepng.cpp"
     HEAP32[$v>>2] = $71; //@line 232 "libraries/lodepng/lodepng.cpp"
     $74 = $71;
     label = 24;
    }
   } else {
    $$pre$i12$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
    $74 = $$pre$i12$i$i;
    label = 24;
   }
   if ((label|0) == 24) {
    label = 0;
    HEAP32[$2>>2] = $61; //@line 243 "libraries/lodepng/lodepng.cpp"
    $73 = (($74) + ($62)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$73>>0] = $60; //@line 289 "libraries/lodepng/lodepng.cpp"
    $$pre32 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
    $$pre33 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
    $78 = $$pre32;$80 = $$pre33;
   }
   $75 = $28 >>> 8; //@line 1656 "libraries/lodepng/lodepng.cpp"
   $76 = $75&255; //@line 1656 "libraries/lodepng/lodepng.cpp"
   $77 = (($78) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $79 = ($80>>>0)<($77>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($79) {
    $81 = $80 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $82 = ($81>>>0)<($77>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $83 = ($77*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $84 = $83 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $85 = $82 ? $77 : $84; //@line 227 "libraries/lodepng/lodepng.cpp"
    $86 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
    $87 = (_realloc($86,$85)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $88 = ($87|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if (!($88)) {
     HEAP32[$3>>2] = $85; //@line 231 "libraries/lodepng/lodepng.cpp"
     HEAP32[$v>>2] = $87; //@line 232 "libraries/lodepng/lodepng.cpp"
     $90 = $87;
     label = 29;
    }
   } else {
    $$pre$i17$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
    $90 = $$pre$i17$i$i;
    label = 29;
   }
   if ((label|0) == 29) {
    label = 0;
    HEAP32[$2>>2] = $77; //@line 243 "libraries/lodepng/lodepng.cpp"
    $89 = (($90) + ($78)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$89>>0] = $76; //@line 289 "libraries/lodepng/lodepng.cpp"
   }
   $91 = ($datapos$029$i$i>>>0)<($insize>>>0); //@line 1659 "libraries/lodepng/lodepng.cpp"
   if ($91) {
    $datapos$127$i$i = $datapos$029$i$i;$j$028$i$i = 0;
    while(1) {
     $92 = (($datapos$127$i$i) + 1)|0; //@line 1661 "libraries/lodepng/lodepng.cpp"
     $93 = (($in) + ($datapos$127$i$i)|0); //@line 1661 "libraries/lodepng/lodepng.cpp"
     $94 = HEAP8[$93>>0]|0; //@line 1661 "libraries/lodepng/lodepng.cpp"
     $95 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
     $96 = (($95) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
     $97 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
     $98 = ($97>>>0)<($96>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
     if ($98) {
      $99 = $97 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
      $100 = ($99>>>0)<($96>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
      $101 = ($96*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
      $102 = $101 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
      $103 = $100 ? $96 : $102; //@line 227 "libraries/lodepng/lodepng.cpp"
      $104 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
      $105 = (_realloc($104,$103)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $106 = ($105|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
      if (!($106)) {
       HEAP32[$3>>2] = $103; //@line 231 "libraries/lodepng/lodepng.cpp"
       HEAP32[$v>>2] = $105; //@line 232 "libraries/lodepng/lodepng.cpp"
       $108 = $105;
       label = 35;
      }
     } else {
      $$pre$i22$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
      $108 = $$pre$i22$i$i;
      label = 35;
     }
     if ((label|0) == 35) {
      label = 0;
      HEAP32[$2>>2] = $96; //@line 243 "libraries/lodepng/lodepng.cpp"
      $107 = (($108) + ($95)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
      HEAP8[$107>>0] = $94; //@line 289 "libraries/lodepng/lodepng.cpp"
     }
     $109 = (($j$028$i$i) + 1)|0; //@line 1659 "libraries/lodepng/lodepng.cpp"
     $110 = ($109>>>0)<(65535); //@line 1659 "libraries/lodepng/lodepng.cpp"
     $111 = ($92>>>0)<($insize>>>0); //@line 1659 "libraries/lodepng/lodepng.cpp"
     $or$cond$i$i = $110 & $111; //@line 1659 "libraries/lodepng/lodepng.cpp"
     if ($or$cond$i$i) {
      $datapos$127$i$i = $92;$j$028$i$i = $109;
     } else {
      $datapos$1$lcssa$i$i = $92;
      break;
     }
    }
   } else {
    $datapos$1$lcssa$i$i = $datapos$029$i$i;
   }
   $112 = (($i$030$i$i) + 1)|0; //@line 1638 "libraries/lodepng/lodepng.cpp"
   $113 = ($112|0)==($7|0); //@line 1638 "libraries/lodepng/lodepng.cpp"
   if ($113) {
    $$0$i = 0;
    break;
   }
   $$pre = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $$pre25 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
   $13 = $$pre;$15 = $$pre25;$datapos$029$i$i = $datapos$1$lcssa$i$i;$i$030$i$i = $112;
  }
  $1240 = HEAP32[$v>>2]|0; //@line 2042 "libraries/lodepng/lodepng.cpp"
  HEAP32[$out>>2] = $1240; //@line 2042 "libraries/lodepng/lodepng.cpp"
  $1241 = HEAP32[$2>>2]|0; //@line 2043 "libraries/lodepng/lodepng.cpp"
  HEAP32[$outsize>>2] = $1241; //@line 2043 "libraries/lodepng/lodepng.cpp"
  STACKTOP = sp;return ($$0$i|0); //@line 2044 "libraries/lodepng/lodepng.cpp"
  break;
 }
 case 1:  {
  $blocksize$0$i = $insize;
  break;
 }
 default: {
  $114 = $insize >>> 3; //@line 2008 "libraries/lodepng/lodepng.cpp"
  $115 = (($114) + 8)|0; //@line 2008 "libraries/lodepng/lodepng.cpp"
  $116 = ($115>>>0)<(65535); //@line 2009 "libraries/lodepng/lodepng.cpp"
  $$$i = $116 ? 65535 : $115; //@line 2009 "libraries/lodepng/lodepng.cpp"
  $blocksize$0$i = $$$i;
 }
 }
 $117 = (($insize) + -1)|0; //@line 2012 "libraries/lodepng/lodepng.cpp"
 $118 = (($117) + ($blocksize$0$i))|0; //@line 2012 "libraries/lodepng/lodepng.cpp"
 $119 = (($118>>>0) / ($blocksize$0$i>>>0))&-1; //@line 2012 "libraries/lodepng/lodepng.cpp"
 $120 = ($119|0)==(0); //@line 2013 "libraries/lodepng/lodepng.cpp"
 $$1$i = $120 ? 1 : $119; //@line 2013 "libraries/lodepng/lodepng.cpp"
 $121 = ((($settings)) + 8|0); //@line 2015 "libraries/lodepng/lodepng.cpp"
 $122 = HEAP32[$121>>2]|0; //@line 2015 "libraries/lodepng/lodepng.cpp"
 $123 = (_malloc(262144)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
 HEAP32[$hash$i>>2] = $123; //@line 1365 "libraries/lodepng/lodepng.cpp"
 $124 = $122 << 2; //@line 1366 "libraries/lodepng/lodepng.cpp"
 $125 = (_malloc($124)|0); //@line 1380 "libraries/lodepng/lodepng.cpp"
 $126 = ((($hash$i)) + 8|0); //@line 1366 "libraries/lodepng/lodepng.cpp"
 HEAP32[$126>>2] = $125; //@line 1366 "libraries/lodepng/lodepng.cpp"
 $127 = $122 << 1; //@line 1367 "libraries/lodepng/lodepng.cpp"
 $128 = (_malloc($127)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
 $129 = ((($hash$i)) + 4|0); //@line 1367 "libraries/lodepng/lodepng.cpp"
 HEAP32[$129>>2] = $128; //@line 1367 "libraries/lodepng/lodepng.cpp"
 $130 = (_malloc($127)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
 $131 = ((($hash$i)) + 20|0); //@line 1369 "libraries/lodepng/lodepng.cpp"
 HEAP32[$131>>2] = $130; //@line 1369 "libraries/lodepng/lodepng.cpp"
 $132 = (_malloc(1036)|0); //@line 1383 "libraries/lodepng/lodepng.cpp"
 $133 = ((($hash$i)) + 12|0); //@line 1370 "libraries/lodepng/lodepng.cpp"
 HEAP32[$133>>2] = $132; //@line 1370 "libraries/lodepng/lodepng.cpp"
 $134 = (_malloc($127)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
 $135 = ((($hash$i)) + 16|0); //@line 1371 "libraries/lodepng/lodepng.cpp"
 HEAP32[$135>>2] = $134; //@line 1371 "libraries/lodepng/lodepng.cpp"
 $136 = ($123|0)==(0|0); //@line 1373 "libraries/lodepng/lodepng.cpp"
 $137 = ($128|0)==(0|0); //@line 1373 "libraries/lodepng/lodepng.cpp"
 $or$cond12$i$i = $136 | $137; //@line 1373 "libraries/lodepng/lodepng.cpp"
 $138 = ($125|0)==(0|0); //@line 1373 "libraries/lodepng/lodepng.cpp"
 $or$cond13$i$i = $138 | $or$cond12$i$i; //@line 1373 "libraries/lodepng/lodepng.cpp"
 if ($or$cond13$i$i) {
  $$0$i = 83;
  $1240 = HEAP32[$v>>2]|0; //@line 2042 "libraries/lodepng/lodepng.cpp"
  HEAP32[$out>>2] = $1240; //@line 2042 "libraries/lodepng/lodepng.cpp"
  $1241 = HEAP32[$2>>2]|0; //@line 2043 "libraries/lodepng/lodepng.cpp"
  HEAP32[$outsize>>2] = $1241; //@line 2043 "libraries/lodepng/lodepng.cpp"
  STACKTOP = sp;return ($$0$i|0); //@line 2044 "libraries/lodepng/lodepng.cpp"
 }
 $139 = ($132|0)==(0|0); //@line 1373 "libraries/lodepng/lodepng.cpp"
 $140 = ($134|0)==(0|0); //@line 1373 "libraries/lodepng/lodepng.cpp"
 $or$cond$i2$i = $139 | $140; //@line 1373 "libraries/lodepng/lodepng.cpp"
 $141 = ($130|0)==(0|0); //@line 1373 "libraries/lodepng/lodepng.cpp"
 $or$cond14$i$i = $141 | $or$cond$i2$i; //@line 1373 "libraries/lodepng/lodepng.cpp"
 if ($or$cond14$i$i) {
  $$0$i = 83;
  $1240 = HEAP32[$v>>2]|0; //@line 2042 "libraries/lodepng/lodepng.cpp"
  HEAP32[$out>>2] = $1240; //@line 2042 "libraries/lodepng/lodepng.cpp"
  $1241 = HEAP32[$2>>2]|0; //@line 2043 "libraries/lodepng/lodepng.cpp"
  HEAP32[$outsize>>2] = $1241; //@line 2043 "libraries/lodepng/lodepng.cpp"
  STACKTOP = sp;return ($$0$i|0); //@line 2044 "libraries/lodepng/lodepng.cpp"
 }
 _memset(($123|0),-1,262144)|0; //@line 1379 "libraries/lodepng/lodepng.cpp"
 $142 = ($122|0)==(0); //@line 1380 "libraries/lodepng/lodepng.cpp"
 if ($142) {
  _memset(($132|0),-1,1036)|0; //@line 1383 "libraries/lodepng/lodepng.cpp"
 } else {
  _memset(($125|0),-1,($124|0))|0; //@line 1380 "libraries/lodepng/lodepng.cpp"
  $i$27$i$i = 0;
  while(1) {
   $143 = $i$27$i$i&65535; //@line 1381 "libraries/lodepng/lodepng.cpp"
   $144 = (($128) + ($i$27$i$i<<1)|0); //@line 1381 "libraries/lodepng/lodepng.cpp"
   HEAP16[$144>>1] = $143; //@line 1381 "libraries/lodepng/lodepng.cpp"
   $145 = (($i$27$i$i) + 1)|0; //@line 1381 "libraries/lodepng/lodepng.cpp"
   $146 = ($145|0)==($122|0); //@line 1381 "libraries/lodepng/lodepng.cpp"
   if ($146) {
    break;
   } else {
    $i$27$i$i = $145;
   }
  }
  _memset(($132|0),-1,1036)|0; //@line 1383 "libraries/lodepng/lodepng.cpp"
  $i$45$i$i = 0;
  while(1) {
   $147 = $i$45$i$i&65535; //@line 1384 "libraries/lodepng/lodepng.cpp"
   $148 = (($134) + ($i$45$i$i<<1)|0); //@line 1384 "libraries/lodepng/lodepng.cpp"
   HEAP16[$148>>1] = $147; //@line 1384 "libraries/lodepng/lodepng.cpp"
   $149 = (($i$45$i$i) + 1)|0; //@line 1384 "libraries/lodepng/lodepng.cpp"
   $150 = ($149|0)==($122|0); //@line 1384 "libraries/lodepng/lodepng.cpp"
   if ($150) {
    break;
   } else {
    $i$45$i$i = $149;
   }
  }
 }
 $151 = (($$1$i) + -1)|0; //@line 2020 "libraries/lodepng/lodepng.cpp"
 $152 = ((($tree_ll$i$i)) + 4|0); //@line 533 "libraries/lodepng/lodepng.cpp"
 $153 = ((($tree_ll$i$i)) + 8|0); //@line 534 "libraries/lodepng/lodepng.cpp"
 $154 = ((($tree_d$i$i)) + 4|0); //@line 533 "libraries/lodepng/lodepng.cpp"
 $155 = ((($tree_d$i$i)) + 8|0); //@line 534 "libraries/lodepng/lodepng.cpp"
 $156 = ((($settings)) + 4|0); //@line 1969 "libraries/lodepng/lodepng.cpp"
 $157 = ((($lz77_encoded$i$i)) + 8|0); //@line 189 "libraries/lodepng/lodepng.cpp"
 $158 = ((($lz77_encoded$i$i)) + 4|0); //@line 189 "libraries/lodepng/lodepng.cpp"
 $159 = ((($settings)) + 12|0); //@line 1974 "libraries/lodepng/lodepng.cpp"
 $160 = ((($settings)) + 16|0); //@line 1974 "libraries/lodepng/lodepng.cpp"
 $161 = ((($settings)) + 20|0); //@line 1974 "libraries/lodepng/lodepng.cpp"
 $162 = ((($lz77_encoded$i9$i)) + 8|0); //@line 189 "libraries/lodepng/lodepng.cpp"
 $163 = ((($lz77_encoded$i9$i)) + 4|0); //@line 189 "libraries/lodepng/lodepng.cpp"
 $164 = ((($tree_ll$i10$i)) + 4|0); //@line 533 "libraries/lodepng/lodepng.cpp"
 $165 = ((($tree_ll$i10$i)) + 8|0); //@line 534 "libraries/lodepng/lodepng.cpp"
 $166 = ((($tree_d$i11$i)) + 4|0); //@line 533 "libraries/lodepng/lodepng.cpp"
 $167 = ((($tree_d$i11$i)) + 8|0); //@line 534 "libraries/lodepng/lodepng.cpp"
 $168 = ((($tree_cl$i$i)) + 4|0); //@line 533 "libraries/lodepng/lodepng.cpp"
 $169 = ((($tree_cl$i$i)) + 8|0); //@line 534 "libraries/lodepng/lodepng.cpp"
 $170 = ((($tree_ll$i10$i)) + 12|0); //@line 863 "libraries/lodepng/lodepng.cpp"
 $171 = ((($tree_ll$i10$i)) + 16|0); //@line 864 "libraries/lodepng/lodepng.cpp"
 $172 = ((($tree_d$i11$i)) + 12|0); //@line 863 "libraries/lodepng/lodepng.cpp"
 $173 = ((($tree_d$i11$i)) + 16|0); //@line 864 "libraries/lodepng/lodepng.cpp"
 $174 = ((($tree_cl$i$i)) + 12|0); //@line 863 "libraries/lodepng/lodepng.cpp"
 $175 = ((($tree_cl$i$i)) + 16|0); //@line 864 "libraries/lodepng/lodepng.cpp"
 $181 = $4;$i$042$i = 0;
 while(1) {
  $176 = ($i$042$i|0)==($151|0); //@line 2020 "libraries/lodepng/lodepng.cpp"
  $177 = $176&1; //@line 2020 "libraries/lodepng/lodepng.cpp"
  $178 = Math_imul($i$042$i, $blocksize$0$i)|0; //@line 2021 "libraries/lodepng/lodepng.cpp"
  $179 = (($178) + ($blocksize$0$i))|0; //@line 2022 "libraries/lodepng/lodepng.cpp"
  $180 = ($179>>>0)>($insize>>>0); //@line 2023 "libraries/lodepng/lodepng.cpp"
  $insize$$i = $180 ? $insize : $179; //@line 2023 "libraries/lodepng/lodepng.cpp"
  switch ($181|0) {
  case 1:  {
   HEAP32[$tree_ll$i$i>>2] = 0; //@line 532 "libraries/lodepng/lodepng.cpp"
   HEAP32[$152>>2] = 0; //@line 533 "libraries/lodepng/lodepng.cpp"
   HEAP32[$153>>2] = 0; //@line 534 "libraries/lodepng/lodepng.cpp"
   HEAP32[$tree_d$i$i>>2] = 0; //@line 532 "libraries/lodepng/lodepng.cpp"
   HEAP32[$154>>2] = 0; //@line 533 "libraries/lodepng/lodepng.cpp"
   HEAP32[$155>>2] = 0; //@line 534 "libraries/lodepng/lodepng.cpp"
   __ZL23generateFixedLitLenTreeP11HuffmanTree($tree_ll$i$i); //@line 1962 "libraries/lodepng/lodepng.cpp"
   __ZL25generateFixedDistanceTreeP11HuffmanTree($tree_d$i$i); //@line 1963 "libraries/lodepng/lodepng.cpp"
   $182 = HEAP32[$bp$i>>2]|0; //@line 1965 "libraries/lodepng/lodepng.cpp"
   $183 = $182 & 7; //@line 1965 "libraries/lodepng/lodepng.cpp"
   $184 = ($183|0)==(0); //@line 1965 "libraries/lodepng/lodepng.cpp"
   do {
    if ($184) {
     $185 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
     $186 = (($185) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
     $187 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
     $188 = ($187>>>0)<($186>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
     if ($188) {
      $189 = $187 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
      $190 = ($189>>>0)<($186>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
      $191 = ($186*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
      $192 = $191 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
      $193 = $190 ? $186 : $192; //@line 227 "libraries/lodepng/lodepng.cpp"
      $194 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
      $195 = (_realloc($194,$193)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $196 = ($195|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
      if ($196) {
       break;
      }
      HEAP32[$3>>2] = $193; //@line 231 "libraries/lodepng/lodepng.cpp"
      HEAP32[$v>>2] = $195; //@line 232 "libraries/lodepng/lodepng.cpp"
      $198 = $195;
     } else {
      $$pre$i$i4$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
      $198 = $$pre$i$i4$i;
     }
     HEAP32[$2>>2] = $186; //@line 243 "libraries/lodepng/lodepng.cpp"
     $197 = (($198) + ($185)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
     HEAP8[$197>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
    }
   } while(0);
   $199 = HEAP32[$bp$i>>2]|0; //@line 1965 "libraries/lodepng/lodepng.cpp"
   $200 = $199 & 7; //@line 1965 "libraries/lodepng/lodepng.cpp"
   $201 = $177 << $200; //@line 1965 "libraries/lodepng/lodepng.cpp"
   $202 = HEAP32[$2>>2]|0; //@line 1965 "libraries/lodepng/lodepng.cpp"
   $203 = (($202) + -1)|0; //@line 1965 "libraries/lodepng/lodepng.cpp"
   $204 = HEAP32[$v>>2]|0; //@line 1965 "libraries/lodepng/lodepng.cpp"
   $205 = (($204) + ($203)|0); //@line 1965 "libraries/lodepng/lodepng.cpp"
   $206 = HEAP8[$205>>0]|0; //@line 1965 "libraries/lodepng/lodepng.cpp"
   $207 = $206&255; //@line 1965 "libraries/lodepng/lodepng.cpp"
   $208 = $207 | $201; //@line 1965 "libraries/lodepng/lodepng.cpp"
   $209 = $208&255; //@line 1965 "libraries/lodepng/lodepng.cpp"
   HEAP8[$205>>0] = $209; //@line 1965 "libraries/lodepng/lodepng.cpp"
   $210 = HEAP32[$bp$i>>2]|0; //@line 1965 "libraries/lodepng/lodepng.cpp"
   $211 = (($210) + 1)|0; //@line 1965 "libraries/lodepng/lodepng.cpp"
   HEAP32[$bp$i>>2] = $211; //@line 1965 "libraries/lodepng/lodepng.cpp"
   $212 = $211 & 7; //@line 1966 "libraries/lodepng/lodepng.cpp"
   $213 = ($212|0)==(0); //@line 1966 "libraries/lodepng/lodepng.cpp"
   do {
    if ($213) {
     $214 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
     $215 = (($214) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
     $216 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
     $217 = ($216>>>0)<($215>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
     if ($217) {
      $218 = $216 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
      $219 = ($218>>>0)<($215>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
      $220 = ($215*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
      $221 = $220 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
      $222 = $219 ? $215 : $221; //@line 227 "libraries/lodepng/lodepng.cpp"
      $223 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
      $224 = (_realloc($223,$222)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $225 = ($224|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
      if ($225) {
       break;
      }
      HEAP32[$3>>2] = $222; //@line 231 "libraries/lodepng/lodepng.cpp"
      HEAP32[$v>>2] = $224; //@line 232 "libraries/lodepng/lodepng.cpp"
      $227 = $224;
     } else {
      $$pre$i5$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
      $227 = $$pre$i5$i$i;
     }
     HEAP32[$2>>2] = $215; //@line 243 "libraries/lodepng/lodepng.cpp"
     $226 = (($227) + ($214)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
     HEAP8[$226>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
    }
   } while(0);
   $228 = HEAP32[$bp$i>>2]|0; //@line 1966 "libraries/lodepng/lodepng.cpp"
   $229 = $228 & 7; //@line 1966 "libraries/lodepng/lodepng.cpp"
   $230 = 1 << $229; //@line 1966 "libraries/lodepng/lodepng.cpp"
   $231 = HEAP32[$2>>2]|0; //@line 1966 "libraries/lodepng/lodepng.cpp"
   $232 = (($231) + -1)|0; //@line 1966 "libraries/lodepng/lodepng.cpp"
   $233 = HEAP32[$v>>2]|0; //@line 1966 "libraries/lodepng/lodepng.cpp"
   $234 = (($233) + ($232)|0); //@line 1966 "libraries/lodepng/lodepng.cpp"
   $235 = HEAP8[$234>>0]|0; //@line 1966 "libraries/lodepng/lodepng.cpp"
   $236 = $235&255; //@line 1966 "libraries/lodepng/lodepng.cpp"
   $237 = $236 | $230; //@line 1966 "libraries/lodepng/lodepng.cpp"
   $238 = $237&255; //@line 1966 "libraries/lodepng/lodepng.cpp"
   HEAP8[$234>>0] = $238; //@line 1966 "libraries/lodepng/lodepng.cpp"
   $239 = HEAP32[$bp$i>>2]|0; //@line 1966 "libraries/lodepng/lodepng.cpp"
   $240 = (($239) + 1)|0; //@line 1966 "libraries/lodepng/lodepng.cpp"
   HEAP32[$bp$i>>2] = $240; //@line 1966 "libraries/lodepng/lodepng.cpp"
   $241 = $240 & 7; //@line 1967 "libraries/lodepng/lodepng.cpp"
   $242 = ($241|0)==(0); //@line 1967 "libraries/lodepng/lodepng.cpp"
   do {
    if ($242) {
     $243 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
     $244 = (($243) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
     $245 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
     $246 = ($245>>>0)<($244>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
     if ($246) {
      $247 = $245 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
      $248 = ($247>>>0)<($244>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
      $249 = ($244*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
      $250 = $249 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
      $251 = $248 ? $244 : $250; //@line 227 "libraries/lodepng/lodepng.cpp"
      $252 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
      $253 = (_realloc($252,$251)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $254 = ($253|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
      if ($254) {
       break;
      }
      HEAP32[$3>>2] = $251; //@line 231 "libraries/lodepng/lodepng.cpp"
      HEAP32[$v>>2] = $253; //@line 232 "libraries/lodepng/lodepng.cpp"
      $256 = $253;
     } else {
      $$pre$i10$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
      $256 = $$pre$i10$i$i;
     }
     HEAP32[$2>>2] = $244; //@line 243 "libraries/lodepng/lodepng.cpp"
     $255 = (($256) + ($243)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
     HEAP8[$255>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
    }
   } while(0);
   $257 = HEAP32[$bp$i>>2]|0; //@line 1967 "libraries/lodepng/lodepng.cpp"
   $258 = (($257) + 1)|0; //@line 1967 "libraries/lodepng/lodepng.cpp"
   HEAP32[$bp$i>>2] = $258; //@line 1967 "libraries/lodepng/lodepng.cpp"
   $259 = HEAP32[$156>>2]|0; //@line 1969 "libraries/lodepng/lodepng.cpp"
   $260 = ($259|0)==(0); //@line 1969 "libraries/lodepng/lodepng.cpp"
   do {
    if ($260) {
     $261 = ($insize$$i>>>0)>($178>>>0); //@line 1980 "libraries/lodepng/lodepng.cpp"
     if ($261) {
      $tree_ll$idx1$val$i$i = HEAP32[$152>>2]|0; //@line 2025 "libraries/lodepng/lodepng.cpp"
      $tree_ll$idx3$val$i$i = HEAP32[$153>>2]|0; //@line 2025 "libraries/lodepng/lodepng.cpp"
      $$pre$i$i$i$i = $258;$i$024$i$i = $178;
      while(1) {
       $269 = (($in) + ($i$024$i$i)|0); //@line 1982 "libraries/lodepng/lodepng.cpp"
       $270 = HEAP8[$269>>0]|0; //@line 1982 "libraries/lodepng/lodepng.cpp"
       $271 = $270&255; //@line 1982 "libraries/lodepng/lodepng.cpp"
       $272 = (($tree_ll$idx1$val$i$i) + ($271<<2)|0); //@line 877 "libraries/lodepng/lodepng.cpp"
       $273 = HEAP32[$272>>2]|0; //@line 877 "libraries/lodepng/lodepng.cpp"
       $274 = (($tree_ll$idx3$val$i$i) + ($271<<2)|0); //@line 882 "libraries/lodepng/lodepng.cpp"
       $275 = HEAP32[$274>>2]|0; //@line 882 "libraries/lodepng/lodepng.cpp"
       $276 = ($275|0)==(0); //@line 438 "libraries/lodepng/lodepng.cpp"
       if ($276) {
        $$pre$i$i26$i$i = $$pre$i$i$i$i;
       } else {
        $277 = (($275) + -1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
        $279 = $$pre$i$i$i$i;$i$01$i$i$i$i = 0;
        while(1) {
         $278 = $279 & 7; //@line 438 "libraries/lodepng/lodepng.cpp"
         $280 = ($278|0)==(0); //@line 438 "libraries/lodepng/lodepng.cpp"
         do {
          if ($280) {
           $281 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
           $282 = (($281) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
           $283 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
           $284 = ($283>>>0)<($282>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
           if ($284) {
            $285 = $283 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
            $286 = ($285>>>0)<($282>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
            $287 = ($282*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
            $288 = $287 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
            $289 = $286 ? $282 : $288; //@line 227 "libraries/lodepng/lodepng.cpp"
            $290 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
            $291 = (_realloc($290,$289)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
            $292 = ($291|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
            if ($292) {
             break;
            }
            HEAP32[$3>>2] = $289; //@line 231 "libraries/lodepng/lodepng.cpp"
            HEAP32[$v>>2] = $291; //@line 232 "libraries/lodepng/lodepng.cpp"
            $294 = $291;
           } else {
            $$pre$i$i$i$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
            $294 = $$pre$i$i$i$i$i;
           }
           HEAP32[$2>>2] = $282; //@line 243 "libraries/lodepng/lodepng.cpp"
           $293 = (($294) + ($281)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
           HEAP8[$293>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
          }
         } while(0);
         $295 = (($277) - ($i$01$i$i$i$i))|0; //@line 438 "libraries/lodepng/lodepng.cpp"
         $296 = $273 >>> $295; //@line 438 "libraries/lodepng/lodepng.cpp"
         $297 = $296 & 1; //@line 438 "libraries/lodepng/lodepng.cpp"
         $298 = HEAP32[$bp$i>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
         $299 = $298 & 7; //@line 438 "libraries/lodepng/lodepng.cpp"
         $300 = $297 << $299; //@line 438 "libraries/lodepng/lodepng.cpp"
         $301 = HEAP32[$2>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
         $302 = (($301) + -1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
         $303 = HEAP32[$v>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
         $304 = (($303) + ($302)|0); //@line 438 "libraries/lodepng/lodepng.cpp"
         $305 = HEAP8[$304>>0]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
         $306 = $305&255; //@line 438 "libraries/lodepng/lodepng.cpp"
         $307 = $306 | $300; //@line 438 "libraries/lodepng/lodepng.cpp"
         $308 = $307&255; //@line 438 "libraries/lodepng/lodepng.cpp"
         HEAP8[$304>>0] = $308; //@line 438 "libraries/lodepng/lodepng.cpp"
         $309 = HEAP32[$bp$i>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
         $310 = (($309) + 1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
         HEAP32[$bp$i>>2] = $310; //@line 438 "libraries/lodepng/lodepng.cpp"
         $311 = (($i$01$i$i$i$i) + 1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
         $312 = ($311|0)==($275|0); //@line 438 "libraries/lodepng/lodepng.cpp"
         if ($312) {
          $$pre$i$i26$i$i = $310;
          break;
         } else {
          $279 = $310;$i$01$i$i$i$i = $311;
         }
        }
       }
       $313 = (($i$024$i$i) + 1)|0; //@line 1980 "libraries/lodepng/lodepng.cpp"
       $exitcond$i$i = ($313|0)==($insize$$i|0); //@line 1980 "libraries/lodepng/lodepng.cpp"
       if ($exitcond$i$i) {
        label = 84;
        break;
       } else {
        $$pre$i$i$i$i = $$pre$i$i26$i$i;$i$024$i$i = $313;
       }
      }
     } else {
      label = 84;
     }
    } else {
     HEAP32[$lz77_encoded$i$i>>2] = 0; //@line 188 "libraries/lodepng/lodepng.cpp"
     HEAP32[$157>>2] = 0; //@line 189 "libraries/lodepng/lodepng.cpp"
     HEAP32[$158>>2] = 0; //@line 189 "libraries/lodepng/lodepng.cpp"
     $262 = HEAP32[$121>>2]|0; //@line 1973 "libraries/lodepng/lodepng.cpp"
     $263 = HEAP32[$159>>2]|0; //@line 1974 "libraries/lodepng/lodepng.cpp"
     $264 = HEAP32[$160>>2]|0; //@line 1974 "libraries/lodepng/lodepng.cpp"
     $265 = HEAP32[$161>>2]|0; //@line 1974 "libraries/lodepng/lodepng.cpp"
     $266 = (__ZL10encodeLZ77P8uivectorP4HashPKhjjjjjj($lz77_encoded$i$i,$hash$i,$in,$178,$insize$$i,$262,$263,$264,$265)|0); //@line 1973 "libraries/lodepng/lodepng.cpp"
     $267 = ($266|0)==(0); //@line 1975 "libraries/lodepng/lodepng.cpp"
     if ($267) {
      __ZL13writeLZ77dataPjP8ucvectorPK8uivectorPK11HuffmanTreeS7_($bp$i,$v,$lz77_encoded$i$i,$tree_ll$i$i,$tree_d$i$i); //@line 1975 "libraries/lodepng/lodepng.cpp"
      HEAP32[$157>>2] = 0; //@line 147 "libraries/lodepng/lodepng.cpp"
      HEAP32[$158>>2] = 0; //@line 147 "libraries/lodepng/lodepng.cpp"
      $268 = HEAP32[$lz77_encoded$i$i>>2]|0; //@line 148 "libraries/lodepng/lodepng.cpp"
      _free($268); //@line 76 "libraries/lodepng/lodepng.cpp"
      label = 84;
      break;
     } else {
      HEAP32[$157>>2] = 0; //@line 147 "libraries/lodepng/lodepng.cpp"
      HEAP32[$158>>2] = 0; //@line 147 "libraries/lodepng/lodepng.cpp"
      $314 = HEAP32[$lz77_encoded$i$i>>2]|0; //@line 148 "libraries/lodepng/lodepng.cpp"
      _free($314); //@line 76 "libraries/lodepng/lodepng.cpp"
      $$pre$i$i = HEAP32[$152>>2]|0; //@line 540 "libraries/lodepng/lodepng.cpp"
      $$pre$i = HEAP32[$153>>2]|0; //@line 541 "libraries/lodepng/lodepng.cpp"
      $357 = $$pre$i$i;$358 = $$pre$i;$error$022$i$i = $266;
      break;
     }
    }
   } while(0);
   if ((label|0) == 84) {
    label = 0;
    $tree_ll$idx$val$i$i = HEAP32[$152>>2]|0; //@line 2025 "libraries/lodepng/lodepng.cpp"
    $315 = ((($tree_ll$idx$val$i$i)) + 1024|0); //@line 877 "libraries/lodepng/lodepng.cpp"
    $316 = HEAP32[$315>>2]|0; //@line 877 "libraries/lodepng/lodepng.cpp"
    $tree_ll$idx2$val$i$i = HEAP32[$153>>2]|0; //@line 2025 "libraries/lodepng/lodepng.cpp"
    $317 = ((($tree_ll$idx2$val$i$i)) + 1024|0); //@line 882 "libraries/lodepng/lodepng.cpp"
    $318 = HEAP32[$317>>2]|0; //@line 882 "libraries/lodepng/lodepng.cpp"
    $319 = ($318|0)==(0); //@line 438 "libraries/lodepng/lodepng.cpp"
    if ($319) {
     $357 = $tree_ll$idx$val$i$i;$358 = $tree_ll$idx2$val$i$i;$error$022$i$i = 0;
    } else {
     $320 = (($318) + -1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
     $$pre$i$i14$i$i = HEAP32[$bp$i>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
     $322 = $$pre$i$i14$i$i;$i$01$i$i16$i$i = 0;
     while(1) {
      $321 = $322 & 7; //@line 438 "libraries/lodepng/lodepng.cpp"
      $323 = ($321|0)==(0); //@line 438 "libraries/lodepng/lodepng.cpp"
      do {
       if ($323) {
        $324 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
        $325 = (($324) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
        $326 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
        $327 = ($326>>>0)<($325>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
        if ($327) {
         $328 = $326 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
         $329 = ($328>>>0)<($325>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
         $330 = ($325*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
         $331 = $330 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
         $332 = $329 ? $325 : $331; //@line 227 "libraries/lodepng/lodepng.cpp"
         $333 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
         $334 = (_realloc($333,$332)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
         $335 = ($334|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
         if ($335) {
          break;
         }
         HEAP32[$3>>2] = $332; //@line 231 "libraries/lodepng/lodepng.cpp"
         HEAP32[$v>>2] = $334; //@line 232 "libraries/lodepng/lodepng.cpp"
         $337 = $334;
        } else {
         $$pre$i$i$i17$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
         $337 = $$pre$i$i$i17$i$i;
        }
        HEAP32[$2>>2] = $325; //@line 243 "libraries/lodepng/lodepng.cpp"
        $336 = (($337) + ($324)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
        HEAP8[$336>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
       }
      } while(0);
      $338 = (($320) - ($i$01$i$i16$i$i))|0; //@line 438 "libraries/lodepng/lodepng.cpp"
      $339 = $316 >>> $338; //@line 438 "libraries/lodepng/lodepng.cpp"
      $340 = $339 & 1; //@line 438 "libraries/lodepng/lodepng.cpp"
      $341 = HEAP32[$bp$i>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
      $342 = $341 & 7; //@line 438 "libraries/lodepng/lodepng.cpp"
      $343 = $340 << $342; //@line 438 "libraries/lodepng/lodepng.cpp"
      $344 = HEAP32[$2>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
      $345 = (($344) + -1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
      $346 = HEAP32[$v>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
      $347 = (($346) + ($345)|0); //@line 438 "libraries/lodepng/lodepng.cpp"
      $348 = HEAP8[$347>>0]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
      $349 = $348&255; //@line 438 "libraries/lodepng/lodepng.cpp"
      $350 = $349 | $343; //@line 438 "libraries/lodepng/lodepng.cpp"
      $351 = $350&255; //@line 438 "libraries/lodepng/lodepng.cpp"
      HEAP8[$347>>0] = $351; //@line 438 "libraries/lodepng/lodepng.cpp"
      $352 = HEAP32[$bp$i>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
      $353 = (($352) + 1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
      HEAP32[$bp$i>>2] = $353; //@line 438 "libraries/lodepng/lodepng.cpp"
      $354 = (($i$01$i$i16$i$i) + 1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
      $355 = ($354|0)==($318|0); //@line 438 "libraries/lodepng/lodepng.cpp"
      if ($355) {
       $357 = $tree_ll$idx$val$i$i;$358 = $tree_ll$idx2$val$i$i;$error$022$i$i = 0;
       break;
      } else {
       $322 = $353;$i$01$i$i16$i$i = $354;
      }
     }
    }
   }
   $356 = HEAP32[$tree_ll$i$i>>2]|0; //@line 539 "libraries/lodepng/lodepng.cpp"
   _free($356); //@line 76 "libraries/lodepng/lodepng.cpp"
   _free($357); //@line 76 "libraries/lodepng/lodepng.cpp"
   _free($358); //@line 76 "libraries/lodepng/lodepng.cpp"
   $359 = HEAP32[$tree_d$i$i>>2]|0; //@line 539 "libraries/lodepng/lodepng.cpp"
   _free($359); //@line 76 "libraries/lodepng/lodepng.cpp"
   $360 = HEAP32[$154>>2]|0; //@line 540 "libraries/lodepng/lodepng.cpp"
   _free($360); //@line 76 "libraries/lodepng/lodepng.cpp"
   $361 = HEAP32[$155>>2]|0; //@line 541 "libraries/lodepng/lodepng.cpp"
   _free($361); //@line 76 "libraries/lodepng/lodepng.cpp"
   $error$1$i = $error$022$i$i;
   break;
  }
  case 2:  {
   $362 = (($insize$$i) - ($178))|0; //@line 1733 "libraries/lodepng/lodepng.cpp"
   HEAP32[$lz77_encoded$i9$i>>2] = 0; //@line 188 "libraries/lodepng/lodepng.cpp"
   HEAP32[$162>>2] = 0; //@line 189 "libraries/lodepng/lodepng.cpp"
   HEAP32[$163>>2] = 0; //@line 189 "libraries/lodepng/lodepng.cpp"
   HEAP32[$tree_ll$i10$i>>2] = 0; //@line 532 "libraries/lodepng/lodepng.cpp"
   HEAP32[$164>>2] = 0; //@line 533 "libraries/lodepng/lodepng.cpp"
   HEAP32[$165>>2] = 0; //@line 534 "libraries/lodepng/lodepng.cpp"
   HEAP32[$tree_d$i11$i>>2] = 0; //@line 532 "libraries/lodepng/lodepng.cpp"
   HEAP32[$166>>2] = 0; //@line 533 "libraries/lodepng/lodepng.cpp"
   HEAP32[$167>>2] = 0; //@line 534 "libraries/lodepng/lodepng.cpp"
   HEAP32[$tree_cl$i$i>>2] = 0; //@line 532 "libraries/lodepng/lodepng.cpp"
   HEAP32[$168>>2] = 0; //@line 533 "libraries/lodepng/lodepng.cpp"
   HEAP32[$169>>2] = 0; //@line 534 "libraries/lodepng/lodepng.cpp"
   $363 = HEAP32[$156>>2]|0; //@line 1761 "libraries/lodepng/lodepng.cpp"
   $364 = ($363|0)==(0); //@line 1761 "libraries/lodepng/lodepng.cpp"
   do {
    if ($364) {
     $371 = $362 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
     $372 = ($371|0)==(0); //@line 155 "libraries/lodepng/lodepng.cpp"
     if ($372) {
      $380 = 0;
     } else {
      $373 = (_realloc(0,$371)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $374 = ($373|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
      if ($374) {
       $bitlen_cl$sroa$0$6$i$i = 0;$bitlen_lld$sroa$0$6$i$i = 0;$bitlen_lld_e$sroa$0$25$i$i = 0;$error$1$i$i = 83;$frequencies_cl$sroa$0$2$i$i = 0;$frequencies_d$sroa$0$2$i$i = 0;$frequencies_ll$sroa$0$2$i$i = 0;
       break;
      }
      HEAP32[$162>>2] = $371; //@line 161 "libraries/lodepng/lodepng.cpp"
      HEAP32[$lz77_encoded$i9$i>>2] = $373; //@line 162 "libraries/lodepng/lodepng.cpp"
      $380 = $373;
     }
     HEAP32[$163>>2] = $362; //@line 173 "libraries/lodepng/lodepng.cpp"
     $375 = ($insize$$i>>>0)>($178>>>0); //@line 1770 "libraries/lodepng/lodepng.cpp"
     if ($375) {
      $i$0253$i$i = $178;
      while(1) {
       $376 = (($in) + ($i$0253$i$i)|0); //@line 1770 "libraries/lodepng/lodepng.cpp"
       $377 = HEAP8[$376>>0]|0; //@line 1770 "libraries/lodepng/lodepng.cpp"
       $378 = $377&255; //@line 1770 "libraries/lodepng/lodepng.cpp"
       $379 = (($380) + ($i$0253$i$i<<2)|0); //@line 1770 "libraries/lodepng/lodepng.cpp"
       HEAP32[$379>>2] = $378; //@line 1770 "libraries/lodepng/lodepng.cpp"
       $381 = (($i$0253$i$i) + 1)|0; //@line 1770 "libraries/lodepng/lodepng.cpp"
       $exitcond$i12$i = ($381|0)==($insize$$i|0); //@line 1770 "libraries/lodepng/lodepng.cpp"
       if ($exitcond$i12$i) {
        label = 101;
        break;
       } else {
        $i$0253$i$i = $381;
       }
      }
     } else {
      label = 101;
     }
    } else {
     $365 = HEAP32[$121>>2]|0; //@line 1763 "libraries/lodepng/lodepng.cpp"
     $366 = HEAP32[$159>>2]|0; //@line 1764 "libraries/lodepng/lodepng.cpp"
     $367 = HEAP32[$160>>2]|0; //@line 1764 "libraries/lodepng/lodepng.cpp"
     $368 = HEAP32[$161>>2]|0; //@line 1764 "libraries/lodepng/lodepng.cpp"
     $369 = (__ZL10encodeLZ77P8uivectorP4HashPKhjjjjjj($lz77_encoded$i9$i,$hash$i,$in,$178,$insize$$i,$365,$366,$367,$368)|0); //@line 1763 "libraries/lodepng/lodepng.cpp"
     $370 = ($369|0)==(0); //@line 1765 "libraries/lodepng/lodepng.cpp"
     if ($370) {
      label = 101;
     } else {
      $bitlen_cl$sroa$0$6$i$i = 0;$bitlen_lld$sroa$0$6$i$i = 0;$bitlen_lld_e$sroa$0$25$i$i = 0;$error$1$i$i = $369;$frequencies_cl$sroa$0$2$i$i = 0;$frequencies_d$sroa$0$2$i$i = 0;$frequencies_ll$sroa$0$2$i$i = 0;
     }
    }
   } while(0);
   L150: do {
    if ((label|0) == 101) {
     label = 0;
     $382 = (_realloc(0,1144)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
     $383 = ($382|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
     if ($383) {
      $bitlen_cl$sroa$0$6$i$i = 0;$bitlen_lld$sroa$0$6$i$i = 0;$bitlen_lld_e$sroa$0$25$i$i = 0;$error$1$i$i = 83;$frequencies_cl$sroa$0$2$i$i = 0;$frequencies_d$sroa$0$2$i$i = 0;$frequencies_ll$sroa$0$2$i$i = 0;
     } else {
      _memset(($382|0),0,1144)|0; //@line 182 "libraries/lodepng/lodepng.cpp"
      $384 = (_realloc(0,120)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $385 = ($384|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
      if ($385) {
       $bitlen_cl$sroa$0$6$i$i = 0;$bitlen_lld$sroa$0$6$i$i = 0;$bitlen_lld_e$sroa$0$25$i$i = 0;$error$1$i$i = 83;$frequencies_cl$sroa$0$2$i$i = 0;$frequencies_d$sroa$0$2$i$i = 0;$frequencies_ll$sroa$0$2$i$i = $382;
      } else {
       dest=$384; stop=dest+120|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 182 "libraries/lodepng/lodepng.cpp"
       $386 = HEAP32[$163>>2]|0; //@line 1777 "libraries/lodepng/lodepng.cpp"
       $387 = ($386|0)==(0); //@line 1777 "libraries/lodepng/lodepng.cpp"
       if (!($387)) {
        $388 = HEAP32[$lz77_encoded$i9$i>>2]|0; //@line 1779 "libraries/lodepng/lodepng.cpp"
        $i$1250$i$i = 0;
        while(1) {
         $389 = (($388) + ($i$1250$i$i<<2)|0); //@line 1779 "libraries/lodepng/lodepng.cpp"
         $390 = HEAP32[$389>>2]|0; //@line 1779 "libraries/lodepng/lodepng.cpp"
         $391 = (($382) + ($390<<2)|0); //@line 1780 "libraries/lodepng/lodepng.cpp"
         $392 = HEAP32[$391>>2]|0; //@line 1780 "libraries/lodepng/lodepng.cpp"
         $393 = (($392) + 1)|0; //@line 1780 "libraries/lodepng/lodepng.cpp"
         HEAP32[$391>>2] = $393; //@line 1780 "libraries/lodepng/lodepng.cpp"
         $394 = ($390>>>0)>(256); //@line 1781 "libraries/lodepng/lodepng.cpp"
         if ($394) {
          $395 = (($i$1250$i$i) + 2)|0; //@line 1783 "libraries/lodepng/lodepng.cpp"
          $396 = (($388) + ($395<<2)|0); //@line 1783 "libraries/lodepng/lodepng.cpp"
          $397 = HEAP32[$396>>2]|0; //@line 1783 "libraries/lodepng/lodepng.cpp"
          $398 = (($384) + ($397<<2)|0); //@line 1784 "libraries/lodepng/lodepng.cpp"
          $399 = HEAP32[$398>>2]|0; //@line 1784 "libraries/lodepng/lodepng.cpp"
          $400 = (($399) + 1)|0; //@line 1784 "libraries/lodepng/lodepng.cpp"
          HEAP32[$398>>2] = $400; //@line 1784 "libraries/lodepng/lodepng.cpp"
          $401 = (($i$1250$i$i) + 3)|0; //@line 1785 "libraries/lodepng/lodepng.cpp"
          $i$2$i$i = $401;
         } else {
          $i$2$i$i = $i$1250$i$i;
         }
         $402 = (($i$2$i$i) + 1)|0; //@line 1777 "libraries/lodepng/lodepng.cpp"
         $403 = ($402|0)==($386|0); //@line 1777 "libraries/lodepng/lodepng.cpp"
         if ($403) {
          break;
         } else {
          $i$1250$i$i = $402;
         }
        }
       }
       $404 = ((($382)) + 1024|0); //@line 1788 "libraries/lodepng/lodepng.cpp"
       HEAP32[$404>>2] = 1; //@line 1788 "libraries/lodepng/lodepng.cpp"
       $$01$i$i$i = 286;
       while(1) {
        $405 = (($$01$i$i$i) + -1)|0; //@line 862 "libraries/lodepng/lodepng.cpp"
        $406 = (($382) + ($405<<2)|0); //@line 862 "libraries/lodepng/lodepng.cpp"
        $407 = HEAP32[$406>>2]|0; //@line 862 "libraries/lodepng/lodepng.cpp"
        $408 = ($407|0)==(0); //@line 862 "libraries/lodepng/lodepng.cpp"
        $409 = ($$01$i$i$i>>>0)>(257); //@line 862 "libraries/lodepng/lodepng.cpp"
        $or$cond$i$i$i = $409 & $408; //@line 862 "libraries/lodepng/lodepng.cpp"
        if ($or$cond$i$i$i) {
         $$01$i$i$i = $405;
        } else {
         $$01$i$i$i$lcssa = $$01$i$i$i;
         break;
        }
       }
       HEAP32[$170>>2] = 15; //@line 863 "libraries/lodepng/lodepng.cpp"
       HEAP32[$171>>2] = $$01$i$i$i$lcssa; //@line 864 "libraries/lodepng/lodepng.cpp"
       $410 = $$01$i$i$i$lcssa << 2; //@line 865 "libraries/lodepng/lodepng.cpp"
       $411 = (_realloc(0,$410)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       HEAP32[$165>>2] = $411; //@line 865 "libraries/lodepng/lodepng.cpp"
       $412 = ($411|0)==(0|0); //@line 866 "libraries/lodepng/lodepng.cpp"
       if ($412) {
        $bitlen_cl$sroa$0$6$i$i = 0;$bitlen_lld$sroa$0$6$i$i = 0;$bitlen_lld_e$sroa$0$25$i$i = 0;$error$1$i$i = 83;$frequencies_cl$sroa$0$2$i$i = 0;$frequencies_d$sroa$0$2$i$i = $384;$frequencies_ll$sroa$0$2$i$i = $382;
       } else {
        _memset(($411|0),0,($410|0))|0; //@line 868 "libraries/lodepng/lodepng.cpp"
        $413 = (__Z28lodepng_huffman_code_lengthsPjPKjjj($411,$382,$$01$i$i$i$lcssa,15)|0); //@line 870 "libraries/lodepng/lodepng.cpp"
        $414 = ($413|0)==(0); //@line 871 "libraries/lodepng/lodepng.cpp"
        if ($414) {
         $415 = (__ZL28HuffmanTree_makeFromLengths2P11HuffmanTree($tree_ll$i10$i)|0); //@line 871 "libraries/lodepng/lodepng.cpp"
         $416 = ($415|0)==(0); //@line 1792 "libraries/lodepng/lodepng.cpp"
         if ($416) {
          $$01$i18$i$i = 30;
          while(1) {
           $417 = (($$01$i18$i$i) + -1)|0; //@line 862 "libraries/lodepng/lodepng.cpp"
           $418 = (($384) + ($417<<2)|0); //@line 862 "libraries/lodepng/lodepng.cpp"
           $419 = HEAP32[$418>>2]|0; //@line 862 "libraries/lodepng/lodepng.cpp"
           $420 = ($419|0)==(0); //@line 862 "libraries/lodepng/lodepng.cpp"
           $421 = ($$01$i18$i$i>>>0)>(2); //@line 862 "libraries/lodepng/lodepng.cpp"
           $or$cond$i19$i$i = $421 & $420; //@line 862 "libraries/lodepng/lodepng.cpp"
           if ($or$cond$i19$i$i) {
            $$01$i18$i$i = $417;
           } else {
            $$01$i18$i$i$lcssa = $$01$i18$i$i;
            break;
           }
          }
          HEAP32[$172>>2] = 15; //@line 863 "libraries/lodepng/lodepng.cpp"
          HEAP32[$173>>2] = $$01$i18$i$i$lcssa; //@line 864 "libraries/lodepng/lodepng.cpp"
          $422 = $$01$i18$i$i$lcssa << 2; //@line 865 "libraries/lodepng/lodepng.cpp"
          $423 = (_realloc(0,$422)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
          HEAP32[$167>>2] = $423; //@line 865 "libraries/lodepng/lodepng.cpp"
          $424 = ($423|0)==(0|0); //@line 866 "libraries/lodepng/lodepng.cpp"
          if ($424) {
           $bitlen_cl$sroa$0$6$i$i = 0;$bitlen_lld$sroa$0$6$i$i = 0;$bitlen_lld_e$sroa$0$25$i$i = 0;$error$1$i$i = 83;$frequencies_cl$sroa$0$2$i$i = 0;$frequencies_d$sroa$0$2$i$i = $384;$frequencies_ll$sroa$0$2$i$i = $382;
          } else {
           _memset(($423|0),0,($422|0))|0; //@line 868 "libraries/lodepng/lodepng.cpp"
           $425 = (__Z28lodepng_huffman_code_lengthsPjPKjjj($423,$384,$$01$i18$i$i$lcssa,15)|0); //@line 870 "libraries/lodepng/lodepng.cpp"
           $426 = ($425|0)==(0); //@line 871 "libraries/lodepng/lodepng.cpp"
           if ($426) {
            $427 = (__ZL28HuffmanTree_makeFromLengths2P11HuffmanTree($tree_d$i11$i)|0); //@line 871 "libraries/lodepng/lodepng.cpp"
            $428 = ($427|0)==(0); //@line 1795 "libraries/lodepng/lodepng.cpp"
            if ($428) {
             $429 = HEAP32[$171>>2]|0; //@line 1797 "libraries/lodepng/lodepng.cpp"
             $430 = ($429>>>0)>(286); //@line 1797 "libraries/lodepng/lodepng.cpp"
             $$$i13$i = $430 ? 286 : $429; //@line 1797 "libraries/lodepng/lodepng.cpp"
             $431 = HEAP32[$173>>2]|0; //@line 1798 "libraries/lodepng/lodepng.cpp"
             $432 = ($431>>>0)>(30); //@line 1798 "libraries/lodepng/lodepng.cpp"
             $$2$i$i = $432 ? 30 : $431; //@line 1798 "libraries/lodepng/lodepng.cpp"
             $433 = ($$$i13$i|0)==(0); //@line 1800 "libraries/lodepng/lodepng.cpp"
             if ($433) {
              $bitlen_lld$sroa$0$0$lcssa$i$i = 0;$bitlen_lld$sroa$13$0$lcssa$i$i = 0;$bitlen_lld$sroa$21$0$lcssa$i$i = 0;
             } else {
              $bitlen_lld$sroa$0$0243$i$i = 0;$bitlen_lld$sroa$13$0242$i$i = 0;$bitlen_lld$sroa$21$0241$i$i = 0;$i$3244$i$i = 0;
              while(1) {
               $tree_ll$idx8$val$i$i = HEAP32[$165>>2]|0; //@line 2026 "libraries/lodepng/lodepng.cpp"
               $435 = (($tree_ll$idx8$val$i$i) + ($i$3244$i$i<<2)|0); //@line 882 "libraries/lodepng/lodepng.cpp"
               $436 = HEAP32[$435>>2]|0; //@line 882 "libraries/lodepng/lodepng.cpp"
               $437 = (($bitlen_lld$sroa$13$0242$i$i) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
               $438 = $437 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
               $439 = ($bitlen_lld$sroa$21$0241$i$i>>>0)<($438>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
               do {
                if ($439) {
                 $440 = $bitlen_lld$sroa$21$0241$i$i << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                 $441 = ($440>>>0)<($438>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
                 $442 = ($437*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
                 $443 = $442 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                 $444 = $441 ? $438 : $443; //@line 157 "libraries/lodepng/lodepng.cpp"
                 $445 = (_realloc($bitlen_lld$sroa$0$0243$i$i,$444)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                 $446 = ($445|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
                 if ($446) {
                  $bitlen_lld$sroa$0$2$i$i = $bitlen_lld$sroa$0$0243$i$i;$bitlen_lld$sroa$13$1$i$i = $bitlen_lld$sroa$13$0242$i$i;$bitlen_lld$sroa$21$2$i$i = $bitlen_lld$sroa$21$0241$i$i;
                  break;
                 }
                 $bitlen_lld$sroa$0$1$i$i = $445;$bitlen_lld$sroa$21$1$i$i = $444;
                 label = 122;
                } else {
                 $bitlen_lld$sroa$0$1$i$i = $bitlen_lld$sroa$0$0243$i$i;$bitlen_lld$sroa$21$1$i$i = $bitlen_lld$sroa$21$0241$i$i;
                 label = 122;
                }
               } while(0);
               if ((label|0) == 122) {
                label = 0;
                $447 = (($bitlen_lld$sroa$0$1$i$i) + ($bitlen_lld$sroa$13$0242$i$i<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
                HEAP32[$447>>2] = $436; //@line 197 "libraries/lodepng/lodepng.cpp"
                $bitlen_lld$sroa$0$2$i$i = $bitlen_lld$sroa$0$1$i$i;$bitlen_lld$sroa$13$1$i$i = $437;$bitlen_lld$sroa$21$2$i$i = $bitlen_lld$sroa$21$1$i$i;
               }
               $448 = (($i$3244$i$i) + 1)|0; //@line 1800 "libraries/lodepng/lodepng.cpp"
               $449 = ($448|0)==($$$i13$i|0); //@line 1800 "libraries/lodepng/lodepng.cpp"
               if ($449) {
                $bitlen_lld$sroa$0$0$lcssa$i$i = $bitlen_lld$sroa$0$2$i$i;$bitlen_lld$sroa$13$0$lcssa$i$i = $bitlen_lld$sroa$13$1$i$i;$bitlen_lld$sroa$21$0$lcssa$i$i = $bitlen_lld$sroa$21$2$i$i;
                break;
               } else {
                $bitlen_lld$sroa$0$0243$i$i = $bitlen_lld$sroa$0$2$i$i;$bitlen_lld$sroa$13$0242$i$i = $bitlen_lld$sroa$13$1$i$i;$bitlen_lld$sroa$21$0241$i$i = $bitlen_lld$sroa$21$2$i$i;$i$3244$i$i = $448;
               }
              }
             }
             $434 = ($$2$i$i|0)==(0); //@line 1801 "libraries/lodepng/lodepng.cpp"
             if ($434) {
              $bitlen_lld$sroa$0$3$lcssa$i$i = $bitlen_lld$sroa$0$0$lcssa$i$i;$bitlen_lld$sroa$13$2$lcssa$i$i = $bitlen_lld$sroa$13$0$lcssa$i$i;
             } else {
              $bitlen_lld$sroa$0$3236$i$i = $bitlen_lld$sroa$0$0$lcssa$i$i;$bitlen_lld$sroa$13$2235$i$i = $bitlen_lld$sroa$13$0$lcssa$i$i;$bitlen_lld$sroa$21$3234$i$i = $bitlen_lld$sroa$21$0$lcssa$i$i;$i$4237$i$i = 0;
              while(1) {
               $tree_d$idx$val$i$i = HEAP32[$167>>2]|0; //@line 2026 "libraries/lodepng/lodepng.cpp"
               $451 = (($tree_d$idx$val$i$i) + ($i$4237$i$i<<2)|0); //@line 882 "libraries/lodepng/lodepng.cpp"
               $452 = HEAP32[$451>>2]|0; //@line 882 "libraries/lodepng/lodepng.cpp"
               $453 = (($bitlen_lld$sroa$13$2235$i$i) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
               $454 = $453 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
               $455 = ($bitlen_lld$sroa$21$3234$i$i>>>0)<($454>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
               do {
                if ($455) {
                 $456 = $bitlen_lld$sroa$21$3234$i$i << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                 $457 = ($456>>>0)<($454>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
                 $458 = ($453*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
                 $459 = $458 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                 $460 = $457 ? $454 : $459; //@line 157 "libraries/lodepng/lodepng.cpp"
                 $461 = (_realloc($bitlen_lld$sroa$0$3236$i$i,$460)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                 $462 = ($461|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
                 if ($462) {
                  $bitlen_lld$sroa$0$5$i$i = $bitlen_lld$sroa$0$3236$i$i;$bitlen_lld$sroa$13$3$i$i = $bitlen_lld$sroa$13$2235$i$i;$bitlen_lld$sroa$21$5$i$i = $bitlen_lld$sroa$21$3234$i$i;
                  break;
                 }
                 $bitlen_lld$sroa$0$4$i$i = $461;$bitlen_lld$sroa$21$4$i$i = $460;
                 label = 128;
                } else {
                 $bitlen_lld$sroa$0$4$i$i = $bitlen_lld$sroa$0$3236$i$i;$bitlen_lld$sroa$21$4$i$i = $bitlen_lld$sroa$21$3234$i$i;
                 label = 128;
                }
               } while(0);
               if ((label|0) == 128) {
                label = 0;
                $463 = (($bitlen_lld$sroa$0$4$i$i) + ($bitlen_lld$sroa$13$2235$i$i<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
                HEAP32[$463>>2] = $452; //@line 197 "libraries/lodepng/lodepng.cpp"
                $bitlen_lld$sroa$0$5$i$i = $bitlen_lld$sroa$0$4$i$i;$bitlen_lld$sroa$13$3$i$i = $453;$bitlen_lld$sroa$21$5$i$i = $bitlen_lld$sroa$21$4$i$i;
               }
               $464 = (($i$4237$i$i) + 1)|0; //@line 1801 "libraries/lodepng/lodepng.cpp"
               $465 = ($464|0)==($$2$i$i|0); //@line 1801 "libraries/lodepng/lodepng.cpp"
               if ($465) {
                $bitlen_lld$sroa$0$3$lcssa$i$i = $bitlen_lld$sroa$0$5$i$i;$bitlen_lld$sroa$13$2$lcssa$i$i = $bitlen_lld$sroa$13$3$i$i;
                break;
               } else {
                $bitlen_lld$sroa$0$3236$i$i = $bitlen_lld$sroa$0$5$i$i;$bitlen_lld$sroa$13$2235$i$i = $bitlen_lld$sroa$13$3$i$i;$bitlen_lld$sroa$21$3234$i$i = $bitlen_lld$sroa$21$5$i$i;$i$4237$i$i = $464;
               }
              }
             }
             $450 = ($bitlen_lld$sroa$13$2$lcssa$i$i|0)==(0); //@line 1805 "libraries/lodepng/lodepng.cpp"
             if ($450) {
              $bitlen_lld_e$sroa$0$0$lcssa$i$i = 0;$bitlen_lld_e$sroa$40$0$lcssa$i$i = 0;
             } else {
              $bitlen_lld_e$sroa$0$0229$i$i = 0;$bitlen_lld_e$sroa$40$0228$i$i = 0;$bitlen_lld_e$sroa$64$0227$i$i = 0;$i$5230$i$i = 0;
              while(1) {
               $466 = (($i$5230$i$i) + 1)|0; //@line 1808 "libraries/lodepng/lodepng.cpp"
               $467 = ($466>>>0)<($bitlen_lld$sroa$13$2$lcssa$i$i>>>0); //@line 1808 "libraries/lodepng/lodepng.cpp"
               $468 = (($bitlen_lld$sroa$0$3$lcssa$i$i) + ($i$5230$i$i<<2)|0); //@line 1808 "libraries/lodepng/lodepng.cpp"
               $469 = HEAP32[$468>>2]|0; //@line 1808 "libraries/lodepng/lodepng.cpp"
               do {
                if ($467) {
                 $471 = $466;$j$0215$i$i = 0;
                 while(1) {
                  $470 = (($bitlen_lld$sroa$0$3$lcssa$i$i) + ($471<<2)|0); //@line 1808 "libraries/lodepng/lodepng.cpp"
                  $472 = HEAP32[$470>>2]|0; //@line 1808 "libraries/lodepng/lodepng.cpp"
                  $473 = ($472|0)==($469|0); //@line 1808 "libraries/lodepng/lodepng.cpp"
                  if (!($473)) {
                   $j$0$lcssa$i$i = $j$0215$i$i;
                   break;
                  }
                  $474 = (($j$0215$i$i) + 1)|0; //@line 1808 "libraries/lodepng/lodepng.cpp"
                  $475 = (($474) + ($466))|0; //@line 1808 "libraries/lodepng/lodepng.cpp"
                  $476 = ($475>>>0)<($bitlen_lld$sroa$13$2$lcssa$i$i>>>0); //@line 1808 "libraries/lodepng/lodepng.cpp"
                  if ($476) {
                   $471 = $475;$j$0215$i$i = $474;
                  } else {
                   $j$0$lcssa$i$i = $474;
                   break;
                  }
                 }
                 $477 = ($469|0)==(0); //@line 1810 "libraries/lodepng/lodepng.cpp"
                 $478 = ($j$0$lcssa$i$i>>>0)>(1); //@line 1810 "libraries/lodepng/lodepng.cpp"
                 $or$cond$i15$i = $477 & $478; //@line 1810 "libraries/lodepng/lodepng.cpp"
                 if ($or$cond$i15$i) {
                  $479 = (($j$0$lcssa$i$i) + 1)|0; //@line 1812 "libraries/lodepng/lodepng.cpp"
                  $480 = ($479>>>0)<(11); //@line 1813 "libraries/lodepng/lodepng.cpp"
                  do {
                   if ($480) {
                    $481 = (($bitlen_lld_e$sroa$40$0228$i$i) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
                    $482 = $481 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
                    $483 = ($bitlen_lld_e$sroa$64$0227$i$i>>>0)<($482>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
                    do {
                     if ($483) {
                      $484 = $bitlen_lld_e$sroa$64$0227$i$i << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                      $485 = ($484>>>0)<($482>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
                      $486 = ($481*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
                      $487 = $486 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                      $488 = $485 ? $482 : $487; //@line 157 "libraries/lodepng/lodepng.cpp"
                      $489 = (_realloc($bitlen_lld_e$sroa$0$0229$i$i,$488)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                      $490 = ($489|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
                      if ($490) {
                       $bitlen_lld_e$sroa$0$2$i$i = $bitlen_lld_e$sroa$0$0229$i$i;$bitlen_lld_e$sroa$40$1$i$i = $bitlen_lld_e$sroa$40$0228$i$i;$bitlen_lld_e$sroa$64$2$i$i = $bitlen_lld_e$sroa$64$0227$i$i;
                       break;
                      }
                      $bitlen_lld_e$sroa$0$1$i$i = $489;$bitlen_lld_e$sroa$64$1$i$i = $488;
                      label = 138;
                     } else {
                      $bitlen_lld_e$sroa$0$1$i$i = $bitlen_lld_e$sroa$0$0229$i$i;$bitlen_lld_e$sroa$64$1$i$i = $bitlen_lld_e$sroa$64$0227$i$i;
                      label = 138;
                     }
                    } while(0);
                    if ((label|0) == 138) {
                     label = 0;
                     $491 = (($bitlen_lld_e$sroa$0$1$i$i) + ($bitlen_lld_e$sroa$40$0228$i$i<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
                     HEAP32[$491>>2] = 17; //@line 197 "libraries/lodepng/lodepng.cpp"
                     $bitlen_lld_e$sroa$0$2$i$i = $bitlen_lld_e$sroa$0$1$i$i;$bitlen_lld_e$sroa$40$1$i$i = $481;$bitlen_lld_e$sroa$64$2$i$i = $bitlen_lld_e$sroa$64$1$i$i;
                    }
                    $492 = (($j$0$lcssa$i$i) + -2)|0; //@line 1816 "libraries/lodepng/lodepng.cpp"
                    $493 = (($bitlen_lld_e$sroa$40$1$i$i) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
                    $494 = $493 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
                    $495 = ($bitlen_lld_e$sroa$64$2$i$i>>>0)<($494>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
                    if ($495) {
                     $496 = $bitlen_lld_e$sroa$64$2$i$i << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $497 = ($496>>>0)<($494>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
                     $498 = ($493*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $499 = $498 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $500 = $497 ? $494 : $499; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $501 = (_realloc($bitlen_lld_e$sroa$0$2$i$i,$500)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                     $502 = ($501|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
                     if ($502) {
                      $bitlen_lld_e$sroa$0$9$i$i = $bitlen_lld_e$sroa$0$2$i$i;$bitlen_lld_e$sroa$40$5$i$i = $bitlen_lld_e$sroa$40$1$i$i;$bitlen_lld_e$sroa$64$9$i$i = $bitlen_lld_e$sroa$64$2$i$i;$j$2$i$i = $479;
                      break;
                     }
                     $bitlen_lld_e$sroa$0$3$i$i = $501;$bitlen_lld_e$sroa$64$3$i$i = $500;
                    } else {
                     $bitlen_lld_e$sroa$0$3$i$i = $bitlen_lld_e$sroa$0$2$i$i;$bitlen_lld_e$sroa$64$3$i$i = $bitlen_lld_e$sroa$64$2$i$i;
                    }
                    $503 = (($bitlen_lld_e$sroa$0$3$i$i) + ($bitlen_lld_e$sroa$40$1$i$i<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
                    HEAP32[$503>>2] = $492; //@line 197 "libraries/lodepng/lodepng.cpp"
                    $bitlen_lld_e$sroa$0$9$i$i = $bitlen_lld_e$sroa$0$3$i$i;$bitlen_lld_e$sroa$40$5$i$i = $493;$bitlen_lld_e$sroa$64$9$i$i = $bitlen_lld_e$sroa$64$3$i$i;$j$2$i$i = $479;
                   } else {
                    $504 = ($479>>>0)>(138); //@line 1820 "libraries/lodepng/lodepng.cpp"
                    $$3$i$i = $504 ? 138 : $479; //@line 1820 "libraries/lodepng/lodepng.cpp"
                    $505 = (($bitlen_lld_e$sroa$40$0228$i$i) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
                    $506 = $505 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
                    $507 = ($bitlen_lld_e$sroa$64$0227$i$i>>>0)<($506>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
                    do {
                     if ($507) {
                      $508 = $bitlen_lld_e$sroa$64$0227$i$i << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                      $509 = ($508>>>0)<($506>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
                      $510 = ($505*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
                      $511 = $510 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                      $512 = $509 ? $506 : $511; //@line 157 "libraries/lodepng/lodepng.cpp"
                      $513 = (_realloc($bitlen_lld_e$sroa$0$0229$i$i,$512)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                      $514 = ($513|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
                      if ($514) {
                       $bitlen_lld_e$sroa$0$6$i$i = $bitlen_lld_e$sroa$0$0229$i$i;$bitlen_lld_e$sroa$40$3$i$i = $bitlen_lld_e$sroa$40$0228$i$i;$bitlen_lld_e$sroa$64$6$i$i = $bitlen_lld_e$sroa$64$0227$i$i;
                       break;
                      }
                      $bitlen_lld_e$sroa$0$5$i$i = $513;$bitlen_lld_e$sroa$64$5$i$i = $512;
                      label = 146;
                     } else {
                      $bitlen_lld_e$sroa$0$5$i$i = $bitlen_lld_e$sroa$0$0229$i$i;$bitlen_lld_e$sroa$64$5$i$i = $bitlen_lld_e$sroa$64$0227$i$i;
                      label = 146;
                     }
                    } while(0);
                    if ((label|0) == 146) {
                     label = 0;
                     $515 = (($bitlen_lld_e$sroa$0$5$i$i) + ($bitlen_lld_e$sroa$40$0228$i$i<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
                     HEAP32[$515>>2] = 18; //@line 197 "libraries/lodepng/lodepng.cpp"
                     $bitlen_lld_e$sroa$0$6$i$i = $bitlen_lld_e$sroa$0$5$i$i;$bitlen_lld_e$sroa$40$3$i$i = $505;$bitlen_lld_e$sroa$64$6$i$i = $bitlen_lld_e$sroa$64$5$i$i;
                    }
                    $516 = (($$3$i$i) + -11)|0; //@line 1822 "libraries/lodepng/lodepng.cpp"
                    $517 = (($bitlen_lld_e$sroa$40$3$i$i) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
                    $518 = $517 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
                    $519 = ($bitlen_lld_e$sroa$64$6$i$i>>>0)<($518>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
                    if ($519) {
                     $520 = $bitlen_lld_e$sroa$64$6$i$i << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $521 = ($520>>>0)<($518>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
                     $522 = ($517*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $523 = $522 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $524 = $521 ? $518 : $523; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $525 = (_realloc($bitlen_lld_e$sroa$0$6$i$i,$524)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                     $526 = ($525|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
                     if ($526) {
                      $bitlen_lld_e$sroa$0$9$i$i = $bitlen_lld_e$sroa$0$6$i$i;$bitlen_lld_e$sroa$40$5$i$i = $bitlen_lld_e$sroa$40$3$i$i;$bitlen_lld_e$sroa$64$9$i$i = $bitlen_lld_e$sroa$64$6$i$i;$j$2$i$i = $$3$i$i;
                      break;
                     }
                     $bitlen_lld_e$sroa$0$7$i$i = $525;$bitlen_lld_e$sroa$64$7$i$i = $524;
                    } else {
                     $bitlen_lld_e$sroa$0$7$i$i = $bitlen_lld_e$sroa$0$6$i$i;$bitlen_lld_e$sroa$64$7$i$i = $bitlen_lld_e$sroa$64$6$i$i;
                    }
                    $527 = (($bitlen_lld_e$sroa$0$7$i$i) + ($bitlen_lld_e$sroa$40$3$i$i<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
                    HEAP32[$527>>2] = $516; //@line 197 "libraries/lodepng/lodepng.cpp"
                    $bitlen_lld_e$sroa$0$9$i$i = $bitlen_lld_e$sroa$0$7$i$i;$bitlen_lld_e$sroa$40$5$i$i = $517;$bitlen_lld_e$sroa$64$9$i$i = $bitlen_lld_e$sroa$64$7$i$i;$j$2$i$i = $$3$i$i;
                   }
                  } while(0);
                  $528 = (($i$5230$i$i) + -1)|0; //@line 1824 "libraries/lodepng/lodepng.cpp"
                  $529 = (($528) + ($j$2$i$i))|0; //@line 1824 "libraries/lodepng/lodepng.cpp"
                  $bitlen_lld_e$sroa$0$24$i$i = $bitlen_lld_e$sroa$0$9$i$i;$bitlen_lld_e$sroa$40$14$i$i = $bitlen_lld_e$sroa$40$5$i$i;$bitlen_lld_e$sroa$64$24$i$i = $bitlen_lld_e$sroa$64$9$i$i;$i$6$i$i = $529;
                  break;
                 }
                 $530 = ($j$0$lcssa$i$i>>>0)>(2); //@line 1826 "libraries/lodepng/lodepng.cpp"
                 if (!($530)) {
                  label = 178;
                  break;
                 }
                 $531 = (($j$0$lcssa$i$i>>>0) / 6)&-1; //@line 1829 "libraries/lodepng/lodepng.cpp"
                 $532 = (($j$0$lcssa$i$i>>>0) % 6)&-1; //@line 1829 "libraries/lodepng/lodepng.cpp"
                 $533 = (($bitlen_lld_e$sroa$40$0228$i$i) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
                 $534 = $533 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
                 $535 = ($bitlen_lld_e$sroa$64$0227$i$i>>>0)<($534>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
                 do {
                  if ($535) {
                   $536 = $bitlen_lld_e$sroa$64$0227$i$i << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                   $537 = ($536>>>0)<($534>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
                   $538 = ($533*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
                   $539 = $538 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                   $540 = $537 ? $534 : $539; //@line 157 "libraries/lodepng/lodepng.cpp"
                   $541 = (_realloc($bitlen_lld_e$sroa$0$0229$i$i,$540)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                   $542 = ($541|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
                   if ($542) {
                    $bitlen_lld_e$sroa$0$12$ph$i$i = $bitlen_lld_e$sroa$0$0229$i$i;$bitlen_lld_e$sroa$40$7$ph$i$i = $bitlen_lld_e$sroa$40$0228$i$i;$bitlen_lld_e$sroa$64$12$ph$i$i = $bitlen_lld_e$sroa$64$0227$i$i;
                    break;
                   }
                   $bitlen_lld_e$sroa$0$10$i$i = $541;$bitlen_lld_e$sroa$64$10$i$i = $540;
                   label = 156;
                  } else {
                   $bitlen_lld_e$sroa$0$10$i$i = $bitlen_lld_e$sroa$0$0229$i$i;$bitlen_lld_e$sroa$64$10$i$i = $bitlen_lld_e$sroa$64$0227$i$i;
                   label = 156;
                  }
                 } while(0);
                 if ((label|0) == 156) {
                  label = 0;
                  $543 = (($bitlen_lld_e$sroa$0$10$i$i) + ($bitlen_lld_e$sroa$40$0228$i$i<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
                  HEAP32[$543>>2] = $469; //@line 197 "libraries/lodepng/lodepng.cpp"
                  $bitlen_lld_e$sroa$0$12$ph$i$i = $bitlen_lld_e$sroa$0$10$i$i;$bitlen_lld_e$sroa$40$7$ph$i$i = $533;$bitlen_lld_e$sroa$64$12$ph$i$i = $bitlen_lld_e$sroa$64$10$i$i;
                 }
                 $544 = ($j$0$lcssa$i$i>>>0)>(5); //@line 1831 "libraries/lodepng/lodepng.cpp"
                 if ($544) {
                  $bitlen_lld_e$sroa$0$12221$i$i = $bitlen_lld_e$sroa$0$12$ph$i$i;$bitlen_lld_e$sroa$40$7220$i$i = $bitlen_lld_e$sroa$40$7$ph$i$i;$bitlen_lld_e$sroa$64$12219$i$i = $bitlen_lld_e$sroa$64$12$ph$i$i;$k$0222$i$i = 0;
                  while(1) {
                   $545 = (($bitlen_lld_e$sroa$40$7220$i$i) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
                   $546 = $545 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
                   $547 = ($bitlen_lld_e$sroa$64$12219$i$i>>>0)<($546>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
                   do {
                    if ($547) {
                     $548 = $bitlen_lld_e$sroa$64$12219$i$i << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $549 = ($548>>>0)<($546>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
                     $550 = ($545*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $551 = $550 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $552 = $549 ? $546 : $551; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $553 = (_realloc($bitlen_lld_e$sroa$0$12221$i$i,$552)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                     $554 = ($553|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
                     if ($554) {
                      $bitlen_lld_e$sroa$0$14$i$i = $bitlen_lld_e$sroa$0$12221$i$i;$bitlen_lld_e$sroa$40$8$i$i = $bitlen_lld_e$sroa$40$7220$i$i;$bitlen_lld_e$sroa$64$14$i$i = $bitlen_lld_e$sroa$64$12219$i$i;
                      break;
                     }
                     $bitlen_lld_e$sroa$0$13$i$i = $553;$bitlen_lld_e$sroa$64$13$i$i = $552;
                     label = 161;
                    } else {
                     $bitlen_lld_e$sroa$0$13$i$i = $bitlen_lld_e$sroa$0$12221$i$i;$bitlen_lld_e$sroa$64$13$i$i = $bitlen_lld_e$sroa$64$12219$i$i;
                     label = 161;
                    }
                   } while(0);
                   if ((label|0) == 161) {
                    label = 0;
                    $555 = (($bitlen_lld_e$sroa$0$13$i$i) + ($bitlen_lld_e$sroa$40$7220$i$i<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
                    HEAP32[$555>>2] = 16; //@line 197 "libraries/lodepng/lodepng.cpp"
                    $bitlen_lld_e$sroa$0$14$i$i = $bitlen_lld_e$sroa$0$13$i$i;$bitlen_lld_e$sroa$40$8$i$i = $545;$bitlen_lld_e$sroa$64$14$i$i = $bitlen_lld_e$sroa$64$13$i$i;
                   }
                   $556 = (($bitlen_lld_e$sroa$40$8$i$i) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
                   $557 = $556 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
                   $558 = ($bitlen_lld_e$sroa$64$14$i$i>>>0)<($557>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
                   do {
                    if ($558) {
                     $559 = $bitlen_lld_e$sroa$64$14$i$i << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $560 = ($559>>>0)<($557>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
                     $561 = ($556*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $562 = $561 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $563 = $560 ? $557 : $562; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $564 = (_realloc($bitlen_lld_e$sroa$0$14$i$i,$563)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                     $565 = ($564|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
                     if ($565) {
                      $bitlen_lld_e$sroa$0$16$i$i = $bitlen_lld_e$sroa$0$14$i$i;$bitlen_lld_e$sroa$40$9$i$i = $bitlen_lld_e$sroa$40$8$i$i;$bitlen_lld_e$sroa$64$16$i$i = $bitlen_lld_e$sroa$64$14$i$i;
                      break;
                     }
                     $bitlen_lld_e$sroa$0$15$i$i = $564;$bitlen_lld_e$sroa$64$15$i$i = $563;
                     label = 165;
                    } else {
                     $bitlen_lld_e$sroa$0$15$i$i = $bitlen_lld_e$sroa$0$14$i$i;$bitlen_lld_e$sroa$64$15$i$i = $bitlen_lld_e$sroa$64$14$i$i;
                     label = 165;
                    }
                   } while(0);
                   if ((label|0) == 165) {
                    label = 0;
                    $566 = (($bitlen_lld_e$sroa$0$15$i$i) + ($bitlen_lld_e$sroa$40$8$i$i<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
                    HEAP32[$566>>2] = 3; //@line 197 "libraries/lodepng/lodepng.cpp"
                    $bitlen_lld_e$sroa$0$16$i$i = $bitlen_lld_e$sroa$0$15$i$i;$bitlen_lld_e$sroa$40$9$i$i = $556;$bitlen_lld_e$sroa$64$16$i$i = $bitlen_lld_e$sroa$64$15$i$i;
                   }
                   $567 = (($k$0222$i$i) + 1)|0; //@line 1831 "libraries/lodepng/lodepng.cpp"
                   $568 = ($567>>>0)<($531>>>0); //@line 1831 "libraries/lodepng/lodepng.cpp"
                   if ($568) {
                    $bitlen_lld_e$sroa$0$12221$i$i = $bitlen_lld_e$sroa$0$16$i$i;$bitlen_lld_e$sroa$40$7220$i$i = $bitlen_lld_e$sroa$40$9$i$i;$bitlen_lld_e$sroa$64$12219$i$i = $bitlen_lld_e$sroa$64$16$i$i;$k$0222$i$i = $567;
                   } else {
                    $bitlen_lld_e$sroa$0$12$lcssa$i$i = $bitlen_lld_e$sroa$0$16$i$i;$bitlen_lld_e$sroa$40$7$lcssa$i$i = $bitlen_lld_e$sroa$40$9$i$i;$bitlen_lld_e$sroa$64$12$lcssa$i$i = $bitlen_lld_e$sroa$64$16$i$i;
                    break;
                   }
                  }
                 } else {
                  $bitlen_lld_e$sroa$0$12$lcssa$i$i = $bitlen_lld_e$sroa$0$12$ph$i$i;$bitlen_lld_e$sroa$40$7$lcssa$i$i = $bitlen_lld_e$sroa$40$7$ph$i$i;$bitlen_lld_e$sroa$64$12$lcssa$i$i = $bitlen_lld_e$sroa$64$12$ph$i$i;
                 }
                 $569 = ($532>>>0)>(2); //@line 1836 "libraries/lodepng/lodepng.cpp"
                 do {
                  if ($569) {
                   $570 = (($bitlen_lld_e$sroa$40$7$lcssa$i$i) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
                   $571 = $570 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
                   $572 = ($bitlen_lld_e$sroa$64$12$lcssa$i$i>>>0)<($571>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
                   do {
                    if ($572) {
                     $573 = $bitlen_lld_e$sroa$64$12$lcssa$i$i << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $574 = ($573>>>0)<($571>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
                     $575 = ($570*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $576 = $575 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $577 = $574 ? $571 : $576; //@line 157 "libraries/lodepng/lodepng.cpp"
                     $578 = (_realloc($bitlen_lld_e$sroa$0$12$lcssa$i$i,$577)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                     $579 = ($578|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
                     if ($579) {
                      $bitlen_lld_e$sroa$0$18$i$i = $bitlen_lld_e$sroa$0$12$lcssa$i$i;$bitlen_lld_e$sroa$40$10$i$i = $bitlen_lld_e$sroa$40$7$lcssa$i$i;$bitlen_lld_e$sroa$64$18$i$i = $bitlen_lld_e$sroa$64$12$lcssa$i$i;
                      break;
                     }
                     $bitlen_lld_e$sroa$0$17$i$i = $578;$bitlen_lld_e$sroa$64$17$i$i = $577;
                     label = 171;
                    } else {
                     $bitlen_lld_e$sroa$0$17$i$i = $bitlen_lld_e$sroa$0$12$lcssa$i$i;$bitlen_lld_e$sroa$64$17$i$i = $bitlen_lld_e$sroa$64$12$lcssa$i$i;
                     label = 171;
                    }
                   } while(0);
                   if ((label|0) == 171) {
                    label = 0;
                    $580 = (($bitlen_lld_e$sroa$0$17$i$i) + ($bitlen_lld_e$sroa$40$7$lcssa$i$i<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
                    HEAP32[$580>>2] = 16; //@line 197 "libraries/lodepng/lodepng.cpp"
                    $bitlen_lld_e$sroa$0$18$i$i = $bitlen_lld_e$sroa$0$17$i$i;$bitlen_lld_e$sroa$40$10$i$i = $570;$bitlen_lld_e$sroa$64$18$i$i = $bitlen_lld_e$sroa$64$17$i$i;
                   }
                   $581 = (($532) + -3)|0; //@line 1839 "libraries/lodepng/lodepng.cpp"
                   $582 = (($bitlen_lld_e$sroa$40$10$i$i) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
                   $583 = $582 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
                   $584 = ($bitlen_lld_e$sroa$64$18$i$i>>>0)<($583>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
                   if ($584) {
                    $585 = $bitlen_lld_e$sroa$64$18$i$i << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                    $586 = ($585>>>0)<($583>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
                    $587 = ($582*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
                    $588 = $587 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                    $589 = $586 ? $583 : $588; //@line 157 "libraries/lodepng/lodepng.cpp"
                    $590 = (_realloc($bitlen_lld_e$sroa$0$18$i$i,$589)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                    $591 = ($590|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
                    if ($591) {
                     $bitlen_lld_e$sroa$0$21$i$i = $bitlen_lld_e$sroa$0$18$i$i;$bitlen_lld_e$sroa$40$12$i$i = $bitlen_lld_e$sroa$40$10$i$i;$bitlen_lld_e$sroa$64$21$i$i = $bitlen_lld_e$sroa$64$18$i$i;$j$3$i$i = $j$0$lcssa$i$i;
                     break;
                    }
                    $bitlen_lld_e$sroa$0$19$i$i = $590;$bitlen_lld_e$sroa$64$19$i$i = $589;
                   } else {
                    $bitlen_lld_e$sroa$0$19$i$i = $bitlen_lld_e$sroa$0$18$i$i;$bitlen_lld_e$sroa$64$19$i$i = $bitlen_lld_e$sroa$64$18$i$i;
                   }
                   $592 = (($bitlen_lld_e$sroa$0$19$i$i) + ($bitlen_lld_e$sroa$40$10$i$i<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
                   HEAP32[$592>>2] = $581; //@line 197 "libraries/lodepng/lodepng.cpp"
                   $bitlen_lld_e$sroa$0$21$i$i = $bitlen_lld_e$sroa$0$19$i$i;$bitlen_lld_e$sroa$40$12$i$i = $582;$bitlen_lld_e$sroa$64$21$i$i = $bitlen_lld_e$sroa$64$19$i$i;$j$3$i$i = $j$0$lcssa$i$i;
                  } else {
                   $593 = (($j$0$lcssa$i$i) - ($532))|0; //@line 1841 "libraries/lodepng/lodepng.cpp"
                   $bitlen_lld_e$sroa$0$21$i$i = $bitlen_lld_e$sroa$0$12$lcssa$i$i;$bitlen_lld_e$sroa$40$12$i$i = $bitlen_lld_e$sroa$40$7$lcssa$i$i;$bitlen_lld_e$sroa$64$21$i$i = $bitlen_lld_e$sroa$64$12$lcssa$i$i;$j$3$i$i = $593;
                  }
                 } while(0);
                 $594 = (($j$3$i$i) + ($i$5230$i$i))|0; //@line 1842 "libraries/lodepng/lodepng.cpp"
                 $bitlen_lld_e$sroa$0$24$i$i = $bitlen_lld_e$sroa$0$21$i$i;$bitlen_lld_e$sroa$40$14$i$i = $bitlen_lld_e$sroa$40$12$i$i;$bitlen_lld_e$sroa$64$24$i$i = $bitlen_lld_e$sroa$64$21$i$i;$i$6$i$i = $594;
                } else {
                 label = 178;
                }
               } while(0);
               do {
                if ((label|0) == 178) {
                 label = 0;
                 $595 = (($bitlen_lld_e$sroa$40$0228$i$i) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
                 $596 = $595 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
                 $597 = ($bitlen_lld_e$sroa$64$0227$i$i>>>0)<($596>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
                 if ($597) {
                  $598 = $bitlen_lld_e$sroa$64$0227$i$i << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                  $599 = ($598>>>0)<($596>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
                  $600 = ($595*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
                  $601 = $600 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
                  $602 = $599 ? $596 : $601; //@line 157 "libraries/lodepng/lodepng.cpp"
                  $603 = (_realloc($bitlen_lld_e$sroa$0$0229$i$i,$602)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                  $604 = ($603|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
                  if ($604) {
                   $bitlen_lld_e$sroa$0$24$i$i = $bitlen_lld_e$sroa$0$0229$i$i;$bitlen_lld_e$sroa$40$14$i$i = $bitlen_lld_e$sroa$40$0228$i$i;$bitlen_lld_e$sroa$64$24$i$i = $bitlen_lld_e$sroa$64$0227$i$i;$i$6$i$i = $i$5230$i$i;
                   break;
                  }
                  $bitlen_lld_e$sroa$0$22$i$i = $603;$bitlen_lld_e$sroa$64$22$i$i = $602;
                 } else {
                  $bitlen_lld_e$sroa$0$22$i$i = $bitlen_lld_e$sroa$0$0229$i$i;$bitlen_lld_e$sroa$64$22$i$i = $bitlen_lld_e$sroa$64$0227$i$i;
                 }
                 $605 = (($bitlen_lld_e$sroa$0$22$i$i) + ($bitlen_lld_e$sroa$40$0228$i$i<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
                 HEAP32[$605>>2] = $469; //@line 197 "libraries/lodepng/lodepng.cpp"
                 $bitlen_lld_e$sroa$0$24$i$i = $bitlen_lld_e$sroa$0$22$i$i;$bitlen_lld_e$sroa$40$14$i$i = $595;$bitlen_lld_e$sroa$64$24$i$i = $bitlen_lld_e$sroa$64$22$i$i;$i$6$i$i = $i$5230$i$i;
                }
               } while(0);
               $606 = (($i$6$i$i) + 1)|0; //@line 1805 "libraries/lodepng/lodepng.cpp"
               $607 = ($606|0)==($bitlen_lld$sroa$13$2$lcssa$i$i|0); //@line 1805 "libraries/lodepng/lodepng.cpp"
               if ($607) {
                $bitlen_lld_e$sroa$0$0$lcssa$i$i = $bitlen_lld_e$sroa$0$24$i$i;$bitlen_lld_e$sroa$40$0$lcssa$i$i = $bitlen_lld_e$sroa$40$14$i$i;
                break;
               } else {
                $bitlen_lld_e$sroa$0$0229$i$i = $bitlen_lld_e$sroa$0$24$i$i;$bitlen_lld_e$sroa$40$0228$i$i = $bitlen_lld_e$sroa$40$14$i$i;$bitlen_lld_e$sroa$64$0227$i$i = $bitlen_lld_e$sroa$64$24$i$i;$i$5230$i$i = $606;
               }
              }
             }
             $608 = (_realloc(0,76)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
             $609 = ($608|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
             if ($609) {
              $bitlen_cl$sroa$0$6$i$i = 0;$bitlen_lld$sroa$0$6$i$i = $bitlen_lld$sroa$0$3$lcssa$i$i;$bitlen_lld_e$sroa$0$25$i$i = $bitlen_lld_e$sroa$0$0$lcssa$i$i;$error$1$i$i = 83;$frequencies_cl$sroa$0$2$i$i = 0;$frequencies_d$sroa$0$2$i$i = $384;$frequencies_ll$sroa$0$2$i$i = $382;
              break;
             }
             dest=$608; stop=dest+76|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 182 "libraries/lodepng/lodepng.cpp"
             $610 = ($bitlen_lld_e$sroa$40$0$lcssa$i$i|0)==(0); //@line 1853 "libraries/lodepng/lodepng.cpp"
             if (!($610)) {
              $i$7213$i$i = 0;
              while(1) {
               $611 = (($bitlen_lld_e$sroa$0$0$lcssa$i$i) + ($i$7213$i$i<<2)|0); //@line 1855 "libraries/lodepng/lodepng.cpp"
               $612 = HEAP32[$611>>2]|0; //@line 1855 "libraries/lodepng/lodepng.cpp"
               $613 = (($608) + ($612<<2)|0); //@line 1855 "libraries/lodepng/lodepng.cpp"
               $614 = HEAP32[$613>>2]|0; //@line 1855 "libraries/lodepng/lodepng.cpp"
               $615 = (($614) + 1)|0; //@line 1855 "libraries/lodepng/lodepng.cpp"
               HEAP32[$613>>2] = $615; //@line 1855 "libraries/lodepng/lodepng.cpp"
               $616 = HEAP32[$611>>2]|0; //@line 1858 "libraries/lodepng/lodepng.cpp"
               $617 = ($616>>>0)>(15); //@line 1858 "libraries/lodepng/lodepng.cpp"
               $618 = $617&1; //@line 1858 "libraries/lodepng/lodepng.cpp"
               $$i$7$i$i = (($i$7213$i$i) + 1)|0; //@line 1858 "libraries/lodepng/lodepng.cpp"
               $619 = (($$i$7$i$i) + ($618))|0; //@line 1853 "libraries/lodepng/lodepng.cpp"
               $620 = ($619|0)==($bitlen_lld_e$sroa$40$0$lcssa$i$i|0); //@line 1853 "libraries/lodepng/lodepng.cpp"
               if ($620) {
                break;
               } else {
                $i$7213$i$i = $619;
               }
              }
             }
             HEAP32[$174>>2] = 7; //@line 863 "libraries/lodepng/lodepng.cpp"
             HEAP32[$175>>2] = 19; //@line 864 "libraries/lodepng/lodepng.cpp"
             $621 = HEAP32[$169>>2]|0; //@line 865 "libraries/lodepng/lodepng.cpp"
             $622 = (_realloc($621,76)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
             HEAP32[$169>>2] = $622; //@line 865 "libraries/lodepng/lodepng.cpp"
             $623 = ($622|0)==(0|0); //@line 866 "libraries/lodepng/lodepng.cpp"
             if ($623) {
              $bitlen_cl$sroa$0$6$i$i = 0;$bitlen_lld$sroa$0$6$i$i = $bitlen_lld$sroa$0$3$lcssa$i$i;$bitlen_lld_e$sroa$0$25$i$i = $bitlen_lld_e$sroa$0$0$lcssa$i$i;$error$1$i$i = 83;$frequencies_cl$sroa$0$2$i$i = $608;$frequencies_d$sroa$0$2$i$i = $384;$frequencies_ll$sroa$0$2$i$i = $382;
              break;
             }
             dest=$622; stop=dest+76|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 868 "libraries/lodepng/lodepng.cpp"
             $624 = (__Z28lodepng_huffman_code_lengthsPjPKjjj($622,$608,19,7)|0); //@line 870 "libraries/lodepng/lodepng.cpp"
             $625 = ($624|0)==(0); //@line 871 "libraries/lodepng/lodepng.cpp"
             if (!($625)) {
              $bitlen_cl$sroa$0$6$i$i = 0;$bitlen_lld$sroa$0$6$i$i = $bitlen_lld$sroa$0$3$lcssa$i$i;$bitlen_lld_e$sroa$0$25$i$i = $bitlen_lld_e$sroa$0$0$lcssa$i$i;$error$1$i$i = $624;$frequencies_cl$sroa$0$2$i$i = $608;$frequencies_d$sroa$0$2$i$i = $384;$frequencies_ll$sroa$0$2$i$i = $382;
              break;
             }
             $626 = (__ZL28HuffmanTree_makeFromLengths2P11HuffmanTree($tree_cl$i$i)|0); //@line 871 "libraries/lodepng/lodepng.cpp"
             $627 = ($626|0)==(0); //@line 1863 "libraries/lodepng/lodepng.cpp"
             if (!($627)) {
              $bitlen_cl$sroa$0$6$i$i = 0;$bitlen_lld$sroa$0$6$i$i = $bitlen_lld$sroa$0$3$lcssa$i$i;$bitlen_lld_e$sroa$0$25$i$i = $bitlen_lld_e$sroa$0$0$lcssa$i$i;$error$1$i$i = $626;$frequencies_cl$sroa$0$2$i$i = $608;$frequencies_d$sroa$0$2$i$i = $384;$frequencies_ll$sroa$0$2$i$i = $382;
              break;
             }
             $628 = HEAP32[$175>>2]|0; //@line 1866 "libraries/lodepng/lodepng.cpp"
             $629 = $628 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
             $630 = ($629|0)==(0); //@line 155 "libraries/lodepng/lodepng.cpp"
             if ($630) {
              $bitlen_cl$sroa$0$1191$ph$i$i = 0;$bitlen_cl$sroa$19$1189$ph$i$i = 0;
             } else {
              $631 = (_realloc(0,$629)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
              $632 = ($631|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
              if ($632) {
               $bitlen_cl$sroa$0$6$i$i = 0;$bitlen_lld$sroa$0$6$i$i = $bitlen_lld$sroa$0$3$lcssa$i$i;$bitlen_lld_e$sroa$0$25$i$i = $bitlen_lld_e$sroa$0$0$lcssa$i$i;$error$1$i$i = 83;$frequencies_cl$sroa$0$2$i$i = $608;$frequencies_d$sroa$0$2$i$i = $384;$frequencies_ll$sroa$0$2$i$i = $382;
               break;
              }
              $bitlen_cl$sroa$0$1191$ph$i$i = $631;$bitlen_cl$sroa$19$1189$ph$i$i = $629;
             }
             $633 = ($628|0)==(0); //@line 1866 "libraries/lodepng/lodepng.cpp"
             if ($633) {
              $bitlen_cl$sroa$0$2$ph$i$i = $bitlen_cl$sroa$0$1191$ph$i$i;$bitlen_cl$sroa$13$1$ph$i$i = 0;$bitlen_cl$sroa$19$2$ph$i$i = $bitlen_cl$sroa$19$1189$ph$i$i;
             } else {
              $tree_cl$idx7$val$i$i = HEAP32[$169>>2]|0; //@line 2026 "libraries/lodepng/lodepng.cpp"
              $i$9211$i$i = 0;
              while(1) {
               $634 = (66980 + ($i$9211$i$i<<2)|0); //@line 1869 "libraries/lodepng/lodepng.cpp"
               $635 = HEAP32[$634>>2]|0; //@line 1869 "libraries/lodepng/lodepng.cpp"
               $636 = (($tree_cl$idx7$val$i$i) + ($635<<2)|0); //@line 882 "libraries/lodepng/lodepng.cpp"
               $637 = HEAP32[$636>>2]|0; //@line 882 "libraries/lodepng/lodepng.cpp"
               $638 = (($bitlen_cl$sroa$0$1191$ph$i$i) + ($i$9211$i$i<<2)|0); //@line 1869 "libraries/lodepng/lodepng.cpp"
               HEAP32[$638>>2] = $637; //@line 1869 "libraries/lodepng/lodepng.cpp"
               $639 = (($i$9211$i$i) + 1)|0; //@line 1866 "libraries/lodepng/lodepng.cpp"
               $640 = ($639|0)==($628|0); //@line 1866 "libraries/lodepng/lodepng.cpp"
               if ($640) {
                $bitlen_cl$sroa$0$2$ph$i$i = $bitlen_cl$sroa$0$1191$ph$i$i;$bitlen_cl$sroa$13$1$ph$i$i = $628;$bitlen_cl$sroa$19$2$ph$i$i = $bitlen_cl$sroa$19$1189$ph$i$i;
                break;
               } else {
                $i$9211$i$i = $639;
               }
              }
             }
             L289: while(1) {
              $bitlen_cl$sroa$13$1$i$i = $bitlen_cl$sroa$13$1$ph$i$i;
              while(1) {
               $641 = (($bitlen_cl$sroa$13$1$i$i) + -1)|0; //@line 1871 "libraries/lodepng/lodepng.cpp"
               $642 = (($bitlen_cl$sroa$0$2$ph$i$i) + ($641<<2)|0); //@line 1871 "libraries/lodepng/lodepng.cpp"
               $643 = HEAP32[$642>>2]|0; //@line 1871 "libraries/lodepng/lodepng.cpp"
               $644 = ($643|0)==(0); //@line 1871 "libraries/lodepng/lodepng.cpp"
               $645 = ($bitlen_cl$sroa$13$1$i$i>>>0)>(4); //@line 1871 "libraries/lodepng/lodepng.cpp"
               $646 = $645 & $644; //@line 1871 "libraries/lodepng/lodepng.cpp"
               if (!($646)) {
                $bitlen_cl$sroa$0$2$ph$i$i$lcssa = $bitlen_cl$sroa$0$2$ph$i$i;$bitlen_cl$sroa$13$1$i$i$lcssa = $bitlen_cl$sroa$13$1$i$i;
                break L289;
               }
               $647 = $641 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
               $648 = ($bitlen_cl$sroa$19$2$ph$i$i>>>0)<($647>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
               if ($648) {
                $$lcssa = $647;$$lcssa72 = $641;
                break;
               } else {
                $bitlen_cl$sroa$13$1$i$i = $641;
               }
              }
              $649 = $bitlen_cl$sroa$19$2$ph$i$i << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
              $650 = ($649>>>0)<($$lcssa>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
              $651 = ($$lcssa72*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
              $652 = $651 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
              $653 = $650 ? $$lcssa : $652; //@line 157 "libraries/lodepng/lodepng.cpp"
              $654 = (_realloc($bitlen_cl$sroa$0$2$ph$i$i,$653)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
              $655 = ($654|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
              if ($655) {
               $bitlen_cl$sroa$0$6$i$i = $bitlen_cl$sroa$0$2$ph$i$i;$bitlen_lld$sroa$0$6$i$i = $bitlen_lld$sroa$0$3$lcssa$i$i;$bitlen_lld_e$sroa$0$25$i$i = $bitlen_lld_e$sroa$0$0$lcssa$i$i;$error$1$i$i = 83;$frequencies_cl$sroa$0$2$i$i = $608;$frequencies_d$sroa$0$2$i$i = $384;$frequencies_ll$sroa$0$2$i$i = $382;
               break L150;
              }
              $bitlen_cl$sroa$0$2$ph$i$i = $654;$bitlen_cl$sroa$13$1$ph$i$i = $$lcssa72;$bitlen_cl$sroa$19$2$ph$i$i = $653;
             }
             $656 = HEAP32[$bp$i>>2]|0; //@line 1893 "libraries/lodepng/lodepng.cpp"
             $657 = $656 & 7; //@line 1893 "libraries/lodepng/lodepng.cpp"
             $658 = ($657|0)==(0); //@line 1893 "libraries/lodepng/lodepng.cpp"
             do {
              if ($658) {
               $659 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
               $660 = (($659) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
               $661 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
               $662 = ($661>>>0)<($660>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
               if ($662) {
                $663 = $661 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                $664 = ($663>>>0)<($660>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                $665 = ($660*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                $666 = $665 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                $667 = $664 ? $660 : $666; //@line 227 "libraries/lodepng/lodepng.cpp"
                $668 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                $669 = (_realloc($668,$667)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                $670 = ($669|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                if ($670) {
                 break;
                }
                HEAP32[$3>>2] = $667; //@line 231 "libraries/lodepng/lodepng.cpp"
                HEAP32[$v>>2] = $669; //@line 232 "libraries/lodepng/lodepng.cpp"
                $672 = $669;
               } else {
                $$pre$i110$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                $672 = $$pre$i110$i$i;
               }
               HEAP32[$2>>2] = $660; //@line 243 "libraries/lodepng/lodepng.cpp"
               $671 = (($672) + ($659)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
               HEAP8[$671>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
              }
             } while(0);
             $673 = HEAP32[$bp$i>>2]|0; //@line 1893 "libraries/lodepng/lodepng.cpp"
             $674 = $673 & 7; //@line 1893 "libraries/lodepng/lodepng.cpp"
             $675 = $177 << $674; //@line 1893 "libraries/lodepng/lodepng.cpp"
             $676 = HEAP32[$2>>2]|0; //@line 1893 "libraries/lodepng/lodepng.cpp"
             $677 = (($676) + -1)|0; //@line 1893 "libraries/lodepng/lodepng.cpp"
             $678 = HEAP32[$v>>2]|0; //@line 1893 "libraries/lodepng/lodepng.cpp"
             $679 = (($678) + ($677)|0); //@line 1893 "libraries/lodepng/lodepng.cpp"
             $680 = HEAP8[$679>>0]|0; //@line 1893 "libraries/lodepng/lodepng.cpp"
             $681 = $680&255; //@line 1893 "libraries/lodepng/lodepng.cpp"
             $682 = $681 | $675; //@line 1893 "libraries/lodepng/lodepng.cpp"
             $683 = $682&255; //@line 1893 "libraries/lodepng/lodepng.cpp"
             HEAP8[$679>>0] = $683; //@line 1893 "libraries/lodepng/lodepng.cpp"
             $684 = HEAP32[$bp$i>>2]|0; //@line 1893 "libraries/lodepng/lodepng.cpp"
             $685 = (($684) + 1)|0; //@line 1893 "libraries/lodepng/lodepng.cpp"
             HEAP32[$bp$i>>2] = $685; //@line 1893 "libraries/lodepng/lodepng.cpp"
             $686 = $685 & 7; //@line 1894 "libraries/lodepng/lodepng.cpp"
             $687 = ($686|0)==(0); //@line 1894 "libraries/lodepng/lodepng.cpp"
             do {
              if ($687) {
               $688 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
               $689 = (($688) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
               $690 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
               $691 = ($690>>>0)<($689>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
               if ($691) {
                $692 = $690 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                $693 = ($692>>>0)<($689>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                $694 = ($689*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                $695 = $694 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                $696 = $693 ? $689 : $695; //@line 227 "libraries/lodepng/lodepng.cpp"
                $697 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                $698 = (_realloc($697,$696)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                $699 = ($698|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                if ($699) {
                 break;
                }
                HEAP32[$3>>2] = $696; //@line 231 "libraries/lodepng/lodepng.cpp"
                HEAP32[$v>>2] = $698; //@line 232 "libraries/lodepng/lodepng.cpp"
                $701 = $698;
               } else {
                $$pre$i114$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                $701 = $$pre$i114$i$i;
               }
               HEAP32[$2>>2] = $689; //@line 243 "libraries/lodepng/lodepng.cpp"
               $700 = (($701) + ($688)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
               HEAP8[$700>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
              }
             } while(0);
             $702 = HEAP32[$bp$i>>2]|0; //@line 1894 "libraries/lodepng/lodepng.cpp"
             $703 = (($702) + 1)|0; //@line 1894 "libraries/lodepng/lodepng.cpp"
             HEAP32[$bp$i>>2] = $703; //@line 1894 "libraries/lodepng/lodepng.cpp"
             $704 = $703 & 7; //@line 1895 "libraries/lodepng/lodepng.cpp"
             $705 = ($704|0)==(0); //@line 1895 "libraries/lodepng/lodepng.cpp"
             do {
              if ($705) {
               $706 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
               $707 = (($706) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
               $708 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
               $709 = ($708>>>0)<($707>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
               if ($709) {
                $710 = $708 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                $711 = ($710>>>0)<($707>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                $712 = ($707*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                $713 = $712 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                $714 = $711 ? $707 : $713; //@line 227 "libraries/lodepng/lodepng.cpp"
                $715 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                $716 = (_realloc($715,$714)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                $717 = ($716|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                if ($717) {
                 break;
                }
                HEAP32[$3>>2] = $714; //@line 231 "libraries/lodepng/lodepng.cpp"
                HEAP32[$v>>2] = $716; //@line 232 "libraries/lodepng/lodepng.cpp"
                $719 = $716;
               } else {
                $$pre$i119$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                $719 = $$pre$i119$i$i;
               }
               HEAP32[$2>>2] = $707; //@line 243 "libraries/lodepng/lodepng.cpp"
               $718 = (($719) + ($706)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
               HEAP8[$718>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
              }
             } while(0);
             $720 = HEAP32[$bp$i>>2]|0; //@line 1895 "libraries/lodepng/lodepng.cpp"
             $721 = $720 & 7; //@line 1895 "libraries/lodepng/lodepng.cpp"
             $722 = 1 << $721; //@line 1895 "libraries/lodepng/lodepng.cpp"
             $723 = HEAP32[$2>>2]|0; //@line 1895 "libraries/lodepng/lodepng.cpp"
             $724 = (($723) + -1)|0; //@line 1895 "libraries/lodepng/lodepng.cpp"
             $725 = HEAP32[$v>>2]|0; //@line 1895 "libraries/lodepng/lodepng.cpp"
             $726 = (($725) + ($724)|0); //@line 1895 "libraries/lodepng/lodepng.cpp"
             $727 = HEAP8[$726>>0]|0; //@line 1895 "libraries/lodepng/lodepng.cpp"
             $728 = $727&255; //@line 1895 "libraries/lodepng/lodepng.cpp"
             $729 = $728 | $722; //@line 1895 "libraries/lodepng/lodepng.cpp"
             $730 = $729&255; //@line 1895 "libraries/lodepng/lodepng.cpp"
             HEAP8[$726>>0] = $730; //@line 1895 "libraries/lodepng/lodepng.cpp"
             $731 = HEAP32[$bp$i>>2]|0; //@line 1895 "libraries/lodepng/lodepng.cpp"
             $732 = (($731) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
             HEAP32[$bp$i>>2] = $732; //@line 1895 "libraries/lodepng/lodepng.cpp"
             $733 = (($bitlen_cl$sroa$13$1$i$i$lcssa) + -4)|0; //@line 1900 "libraries/lodepng/lodepng.cpp"
             $HCLEN$0$i$i = $733;
             while(1) {
              $734 = (($HCLEN$0$i$i) + 3)|0; //@line 1902 "libraries/lodepng/lodepng.cpp"
              $735 = (($bitlen_cl$sroa$0$2$ph$i$i$lcssa) + ($734<<2)|0); //@line 1902 "libraries/lodepng/lodepng.cpp"
              $736 = HEAP32[$735>>2]|0; //@line 1902 "libraries/lodepng/lodepng.cpp"
              $737 = ($HCLEN$0$i$i|0)!=(0); //@line 1902 "libraries/lodepng/lodepng.cpp"
              $not$$i$i = ($736|0)==(0); //@line 1902 "libraries/lodepng/lodepng.cpp"
              $738 = $737 & $not$$i$i; //@line 1902 "libraries/lodepng/lodepng.cpp"
              $739 = (($HCLEN$0$i$i) + -1)|0; //@line 1902 "libraries/lodepng/lodepng.cpp"
              if ($738) {
               $HCLEN$0$i$i = $739;
              } else {
               $HCLEN$0$i$i$lcssa = $HCLEN$0$i$i;
               break;
              }
             }
             $740 = (($$$i13$i) + -257)|0; //@line 1898 "libraries/lodepng/lodepng.cpp"
             $742 = $732;$i$01$i$i$i = 0;
             while(1) {
              $741 = $742 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
              $743 = ($741|0)==(0); //@line 432 "libraries/lodepng/lodepng.cpp"
              do {
               if ($743) {
                $744 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                $745 = (($744) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                $746 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
                $747 = ($746>>>0)<($745>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
                if ($747) {
                 $748 = $746 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                 $749 = ($748>>>0)<($745>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                 $750 = ($745*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                 $751 = $750 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                 $752 = $749 ? $745 : $751; //@line 227 "libraries/lodepng/lodepng.cpp"
                 $753 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                 $754 = (_realloc($753,$752)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                 $755 = ($754|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                 if ($755) {
                  break;
                 }
                 HEAP32[$3>>2] = $752; //@line 231 "libraries/lodepng/lodepng.cpp"
                 HEAP32[$v>>2] = $754; //@line 232 "libraries/lodepng/lodepng.cpp"
                 $757 = $754;
                } else {
                 $$pre$i$i$i19$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                 $757 = $$pre$i$i$i19$i;
                }
                HEAP32[$2>>2] = $745; //@line 243 "libraries/lodepng/lodepng.cpp"
                $756 = (($757) + ($744)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
                HEAP8[$756>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
               }
              } while(0);
              $758 = $740 >>> $i$01$i$i$i; //@line 432 "libraries/lodepng/lodepng.cpp"
              $759 = $758 & 1; //@line 432 "libraries/lodepng/lodepng.cpp"
              $760 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $761 = $760 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
              $762 = $759 << $761; //@line 432 "libraries/lodepng/lodepng.cpp"
              $763 = HEAP32[$2>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $764 = (($763) + -1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $765 = HEAP32[$v>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $766 = (($765) + ($764)|0); //@line 432 "libraries/lodepng/lodepng.cpp"
              $767 = HEAP8[$766>>0]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $768 = $767&255; //@line 432 "libraries/lodepng/lodepng.cpp"
              $769 = $768 | $762; //@line 432 "libraries/lodepng/lodepng.cpp"
              $770 = $769&255; //@line 432 "libraries/lodepng/lodepng.cpp"
              HEAP8[$766>>0] = $770; //@line 432 "libraries/lodepng/lodepng.cpp"
              $771 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $772 = (($771) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              HEAP32[$bp$i>>2] = $772; //@line 432 "libraries/lodepng/lodepng.cpp"
              $773 = (($i$01$i$i$i) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $774 = ($773|0)==(5); //@line 432 "libraries/lodepng/lodepng.cpp"
              if ($774) {
               $$lcssa74 = $772;
               break;
              } else {
               $742 = $772;$i$01$i$i$i = $773;
              }
             }
             $775 = (($$2$i$i) + -1)|0; //@line 1899 "libraries/lodepng/lodepng.cpp"
             $777 = $$lcssa74;$i$01$i128$i$i = 0;
             while(1) {
              $776 = $777 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
              $778 = ($776|0)==(0); //@line 432 "libraries/lodepng/lodepng.cpp"
              do {
               if ($778) {
                $779 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                $780 = (($779) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                $781 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
                $782 = ($781>>>0)<($780>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
                if ($782) {
                 $783 = $781 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                 $784 = ($783>>>0)<($780>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                 $785 = ($780*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                 $786 = $785 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                 $787 = $784 ? $780 : $786; //@line 227 "libraries/lodepng/lodepng.cpp"
                 $788 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                 $789 = (_realloc($788,$787)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                 $790 = ($789|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                 if ($790) {
                  break;
                 }
                 HEAP32[$3>>2] = $787; //@line 231 "libraries/lodepng/lodepng.cpp"
                 HEAP32[$v>>2] = $789; //@line 232 "libraries/lodepng/lodepng.cpp"
                 $792 = $789;
                } else {
                 $$pre$i$i129$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                 $792 = $$pre$i$i129$i$i;
                }
                HEAP32[$2>>2] = $780; //@line 243 "libraries/lodepng/lodepng.cpp"
                $791 = (($792) + ($779)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
                HEAP8[$791>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
               }
              } while(0);
              $793 = $775 >>> $i$01$i128$i$i; //@line 432 "libraries/lodepng/lodepng.cpp"
              $794 = $793 & 1; //@line 432 "libraries/lodepng/lodepng.cpp"
              $795 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $796 = $795 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
              $797 = $794 << $796; //@line 432 "libraries/lodepng/lodepng.cpp"
              $798 = HEAP32[$2>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $799 = (($798) + -1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $800 = HEAP32[$v>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $801 = (($800) + ($799)|0); //@line 432 "libraries/lodepng/lodepng.cpp"
              $802 = HEAP8[$801>>0]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $803 = $802&255; //@line 432 "libraries/lodepng/lodepng.cpp"
              $804 = $803 | $797; //@line 432 "libraries/lodepng/lodepng.cpp"
              $805 = $804&255; //@line 432 "libraries/lodepng/lodepng.cpp"
              HEAP8[$801>>0] = $805; //@line 432 "libraries/lodepng/lodepng.cpp"
              $806 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $807 = (($806) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              HEAP32[$bp$i>>2] = $807; //@line 432 "libraries/lodepng/lodepng.cpp"
              $808 = (($i$01$i128$i$i) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $809 = ($808|0)==(5); //@line 432 "libraries/lodepng/lodepng.cpp"
              if ($809) {
               $811 = $807;$i$01$i137$i$i = 0;
               break;
              } else {
               $777 = $807;$i$01$i128$i$i = $808;
              }
             }
             while(1) {
              $810 = $811 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
              $812 = ($810|0)==(0); //@line 432 "libraries/lodepng/lodepng.cpp"
              do {
               if ($812) {
                $813 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                $814 = (($813) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                $815 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
                $816 = ($815>>>0)<($814>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
                if ($816) {
                 $817 = $815 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                 $818 = ($817>>>0)<($814>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                 $819 = ($814*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                 $820 = $819 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                 $821 = $818 ? $814 : $820; //@line 227 "libraries/lodepng/lodepng.cpp"
                 $822 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                 $823 = (_realloc($822,$821)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                 $824 = ($823|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                 if ($824) {
                  break;
                 }
                 HEAP32[$3>>2] = $821; //@line 231 "libraries/lodepng/lodepng.cpp"
                 HEAP32[$v>>2] = $823; //@line 232 "libraries/lodepng/lodepng.cpp"
                 $826 = $823;
                } else {
                 $$pre$i$i138$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                 $826 = $$pre$i$i138$i$i;
                }
                HEAP32[$2>>2] = $814; //@line 243 "libraries/lodepng/lodepng.cpp"
                $825 = (($826) + ($813)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
                HEAP8[$825>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
               }
              } while(0);
              $827 = $HCLEN$0$i$i$lcssa >>> $i$01$i137$i$i; //@line 432 "libraries/lodepng/lodepng.cpp"
              $828 = $827 & 1; //@line 432 "libraries/lodepng/lodepng.cpp"
              $829 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $830 = $829 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
              $831 = $828 << $830; //@line 432 "libraries/lodepng/lodepng.cpp"
              $832 = HEAP32[$2>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $833 = (($832) + -1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $834 = HEAP32[$v>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $835 = (($834) + ($833)|0); //@line 432 "libraries/lodepng/lodepng.cpp"
              $836 = HEAP8[$835>>0]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $837 = $836&255; //@line 432 "libraries/lodepng/lodepng.cpp"
              $838 = $837 | $831; //@line 432 "libraries/lodepng/lodepng.cpp"
              $839 = $838&255; //@line 432 "libraries/lodepng/lodepng.cpp"
              HEAP8[$835>>0] = $839; //@line 432 "libraries/lodepng/lodepng.cpp"
              $840 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $841 = (($840) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              HEAP32[$bp$i>>2] = $841; //@line 432 "libraries/lodepng/lodepng.cpp"
              $842 = (($i$01$i137$i$i) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
              $843 = ($842|0)==(4); //@line 432 "libraries/lodepng/lodepng.cpp"
              if ($843) {
               $$lcssa76 = $841;
               break;
              } else {
               $811 = $841;$i$01$i137$i$i = $842;
              }
             }
             $844 = (($HCLEN$0$i$i$lcssa) + 4)|0; //@line 1902 "libraries/lodepng/lodepng.cpp"
             $845 = ($844|0)==(0); //@line 1908 "libraries/lodepng/lodepng.cpp"
             if ($845) {
              $$pre$i$i153269$i$i = $$lcssa76;
             } else {
              $$pre$i144$i$i = $$lcssa76;$i$10209$i$i = 0;
              while(1) {
               $846 = (($bitlen_cl$sroa$0$2$ph$i$i$lcssa) + ($i$10209$i$i<<2)|0); //@line 1908 "libraries/lodepng/lodepng.cpp"
               $847 = HEAP32[$846>>2]|0; //@line 1908 "libraries/lodepng/lodepng.cpp"
               $848 = $$pre$i144$i$i & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
               $849 = ($848|0)==(0); //@line 432 "libraries/lodepng/lodepng.cpp"
               do {
                if ($849) {
                 $850 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                 $851 = (($850) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                 $852 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
                 $853 = ($852>>>0)<($851>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
                 if ($853) {
                  $854 = $852 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                  $855 = ($854>>>0)<($851>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                  $856 = ($851*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                  $857 = $856 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                  $858 = $855 ? $851 : $857; //@line 227 "libraries/lodepng/lodepng.cpp"
                  $859 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                  $860 = (_realloc($859,$858)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                  $861 = ($860|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                  if ($861) {
                   break;
                  }
                  HEAP32[$3>>2] = $858; //@line 231 "libraries/lodepng/lodepng.cpp"
                  HEAP32[$v>>2] = $860; //@line 232 "libraries/lodepng/lodepng.cpp"
                  $863 = $860;
                 } else {
                  $$pre$i$i147$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                  $863 = $$pre$i$i147$i$i;
                 }
                 HEAP32[$2>>2] = $851; //@line 243 "libraries/lodepng/lodepng.cpp"
                 $862 = (($863) + ($850)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
                 HEAP8[$862>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
                }
               } while(0);
               $864 = $847 & 1; //@line 432 "libraries/lodepng/lodepng.cpp"
               $865 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $866 = $865 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
               $867 = $864 << $866; //@line 432 "libraries/lodepng/lodepng.cpp"
               $868 = HEAP32[$2>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $869 = (($868) + -1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $870 = HEAP32[$v>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $871 = (($870) + ($869)|0); //@line 432 "libraries/lodepng/lodepng.cpp"
               $872 = HEAP8[$871>>0]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $873 = $872&255; //@line 432 "libraries/lodepng/lodepng.cpp"
               $874 = $873 | $867; //@line 432 "libraries/lodepng/lodepng.cpp"
               $875 = $874&255; //@line 432 "libraries/lodepng/lodepng.cpp"
               HEAP8[$871>>0] = $875; //@line 432 "libraries/lodepng/lodepng.cpp"
               $876 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $877 = (($876) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               HEAP32[$bp$i>>2] = $877; //@line 432 "libraries/lodepng/lodepng.cpp"
               $878 = $877 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
               $879 = ($878|0)==(0); //@line 432 "libraries/lodepng/lodepng.cpp"
               do {
                if ($879) {
                 $1163 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                 $1164 = (($1163) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                 $1165 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
                 $1166 = ($1165>>>0)<($1164>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
                 if ($1166) {
                  $1167 = $1165 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                  $1168 = ($1167>>>0)<($1164>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                  $1169 = ($1164*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                  $1170 = $1169 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                  $1171 = $1168 ? $1164 : $1170; //@line 227 "libraries/lodepng/lodepng.cpp"
                  $1172 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                  $1173 = (_realloc($1172,$1171)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                  $1174 = ($1173|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                  if ($1174) {
                   break;
                  }
                  HEAP32[$3>>2] = $1171; //@line 231 "libraries/lodepng/lodepng.cpp"
                  HEAP32[$v>>2] = $1173; //@line 232 "libraries/lodepng/lodepng.cpp"
                  $1176 = $1173;
                 } else {
                  $$pre$i$i147$1$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                  $1176 = $$pre$i$i147$1$i$i;
                 }
                 HEAP32[$2>>2] = $1164; //@line 243 "libraries/lodepng/lodepng.cpp"
                 $1175 = (($1176) + ($1163)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
                 HEAP8[$1175>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
                }
               } while(0);
               $1177 = $847 >>> 1; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1178 = $1177 & 1; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1179 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1180 = $1179 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1181 = $1178 << $1180; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1182 = HEAP32[$2>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1183 = (($1182) + -1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1184 = HEAP32[$v>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1185 = (($1184) + ($1183)|0); //@line 432 "libraries/lodepng/lodepng.cpp"
               $1186 = HEAP8[$1185>>0]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1187 = $1186&255; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1188 = $1187 | $1181; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1189 = $1188&255; //@line 432 "libraries/lodepng/lodepng.cpp"
               HEAP8[$1185>>0] = $1189; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1190 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1191 = (($1190) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               HEAP32[$bp$i>>2] = $1191; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1192 = $1191 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1193 = ($1192|0)==(0); //@line 432 "libraries/lodepng/lodepng.cpp"
               do {
                if ($1193) {
                 $1194 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                 $1195 = (($1194) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                 $1196 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
                 $1197 = ($1196>>>0)<($1195>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
                 if ($1197) {
                  $1198 = $1196 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                  $1199 = ($1198>>>0)<($1195>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                  $1200 = ($1195*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                  $1201 = $1200 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                  $1202 = $1199 ? $1195 : $1201; //@line 227 "libraries/lodepng/lodepng.cpp"
                  $1203 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                  $1204 = (_realloc($1203,$1202)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                  $1205 = ($1204|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                  if ($1205) {
                   break;
                  }
                  HEAP32[$3>>2] = $1202; //@line 231 "libraries/lodepng/lodepng.cpp"
                  HEAP32[$v>>2] = $1204; //@line 232 "libraries/lodepng/lodepng.cpp"
                  $1207 = $1204;
                 } else {
                  $$pre$i$i147$2$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                  $1207 = $$pre$i$i147$2$i$i;
                 }
                 HEAP32[$2>>2] = $1195; //@line 243 "libraries/lodepng/lodepng.cpp"
                 $1206 = (($1207) + ($1194)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
                 HEAP8[$1206>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
                }
               } while(0);
               $1208 = $847 >>> 2; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1209 = $1208 & 1; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1210 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1211 = $1210 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1212 = $1209 << $1211; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1213 = HEAP32[$2>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1214 = (($1213) + -1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1215 = HEAP32[$v>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1216 = (($1215) + ($1214)|0); //@line 432 "libraries/lodepng/lodepng.cpp"
               $1217 = HEAP8[$1216>>0]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1218 = $1217&255; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1219 = $1218 | $1212; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1220 = $1219&255; //@line 432 "libraries/lodepng/lodepng.cpp"
               HEAP8[$1216>>0] = $1220; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1221 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1222 = (($1221) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
               HEAP32[$bp$i>>2] = $1222; //@line 432 "libraries/lodepng/lodepng.cpp"
               $1223 = (($i$10209$i$i) + 1)|0; //@line 1908 "libraries/lodepng/lodepng.cpp"
               $1224 = ($1223|0)==($844|0); //@line 1908 "libraries/lodepng/lodepng.cpp"
               if ($1224) {
                $$pre$i$i153269$i$i = $1222;
                break;
               } else {
                $$pre$i144$i$i = $1222;$i$10209$i$i = $1223;
               }
              }
             }
             L384: do {
              if (!($610)) {
               $tree_cl$idx$val$i$i = HEAP32[$168>>2]|0; //@line 2026 "libraries/lodepng/lodepng.cpp"
               $tree_cl$idx6$val$i$i = HEAP32[$169>>2]|0; //@line 2026 "libraries/lodepng/lodepng.cpp"
               $$pre$i$i153$i$i = $$pre$i$i153269$i$i;$i$11208$i$i = 0;
               while(1) {
                $880 = (($bitlen_lld_e$sroa$0$0$lcssa$i$i) + ($i$11208$i$i<<2)|0); //@line 1913 "libraries/lodepng/lodepng.cpp"
                $881 = HEAP32[$880>>2]|0; //@line 1913 "libraries/lodepng/lodepng.cpp"
                $882 = (($tree_cl$idx$val$i$i) + ($881<<2)|0); //@line 877 "libraries/lodepng/lodepng.cpp"
                $883 = HEAP32[$882>>2]|0; //@line 877 "libraries/lodepng/lodepng.cpp"
                $884 = (($tree_cl$idx6$val$i$i) + ($881<<2)|0); //@line 882 "libraries/lodepng/lodepng.cpp"
                $885 = HEAP32[$884>>2]|0; //@line 882 "libraries/lodepng/lodepng.cpp"
                $886 = ($885|0)==(0); //@line 438 "libraries/lodepng/lodepng.cpp"
                if ($886) {
                 $$pre$i154$i$i = $$pre$i$i153$i$i;$923 = $881;
                } else {
                 $887 = (($885) + -1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
                 $889 = $$pre$i$i153$i$i;$i$01$i$i$i23$i = 0;
                 while(1) {
                  $888 = $889 & 7; //@line 438 "libraries/lodepng/lodepng.cpp"
                  $890 = ($888|0)==(0); //@line 438 "libraries/lodepng/lodepng.cpp"
                  do {
                   if ($890) {
                    $891 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                    $892 = (($891) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                    $893 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
                    $894 = ($893>>>0)<($892>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
                    if ($894) {
                     $895 = $893 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                     $896 = ($895>>>0)<($892>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                     $897 = ($892*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                     $898 = $897 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                     $899 = $896 ? $892 : $898; //@line 227 "libraries/lodepng/lodepng.cpp"
                     $900 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                     $901 = (_realloc($900,$899)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                     $902 = ($901|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                     if ($902) {
                      break;
                     }
                     HEAP32[$3>>2] = $899; //@line 231 "libraries/lodepng/lodepng.cpp"
                     HEAP32[$v>>2] = $901; //@line 232 "libraries/lodepng/lodepng.cpp"
                     $904 = $901;
                    } else {
                     $$pre$i$i$i$i24$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                     $904 = $$pre$i$i$i$i24$i;
                    }
                    HEAP32[$2>>2] = $892; //@line 243 "libraries/lodepng/lodepng.cpp"
                    $903 = (($904) + ($891)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
                    HEAP8[$903>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
                   }
                  } while(0);
                  $905 = (($887) - ($i$01$i$i$i23$i))|0; //@line 438 "libraries/lodepng/lodepng.cpp"
                  $906 = $883 >>> $905; //@line 438 "libraries/lodepng/lodepng.cpp"
                  $907 = $906 & 1; //@line 438 "libraries/lodepng/lodepng.cpp"
                  $908 = HEAP32[$bp$i>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
                  $909 = $908 & 7; //@line 438 "libraries/lodepng/lodepng.cpp"
                  $910 = $907 << $909; //@line 438 "libraries/lodepng/lodepng.cpp"
                  $911 = HEAP32[$2>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
                  $912 = (($911) + -1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
                  $913 = HEAP32[$v>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
                  $914 = (($913) + ($912)|0); //@line 438 "libraries/lodepng/lodepng.cpp"
                  $915 = HEAP8[$914>>0]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
                  $916 = $915&255; //@line 438 "libraries/lodepng/lodepng.cpp"
                  $917 = $916 | $910; //@line 438 "libraries/lodepng/lodepng.cpp"
                  $918 = $917&255; //@line 438 "libraries/lodepng/lodepng.cpp"
                  HEAP8[$914>>0] = $918; //@line 438 "libraries/lodepng/lodepng.cpp"
                  $919 = HEAP32[$bp$i>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
                  $920 = (($919) + 1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
                  HEAP32[$bp$i>>2] = $920; //@line 438 "libraries/lodepng/lodepng.cpp"
                  $921 = (($i$01$i$i$i23$i) + 1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
                  $922 = ($921|0)==($885|0); //@line 438 "libraries/lodepng/lodepng.cpp"
                  if ($922) {
                   $$lcssa78 = $920;
                   break;
                  } else {
                   $889 = $920;$i$01$i$i$i23$i = $921;
                  }
                 }
                 $$pre270$i$i = HEAP32[$880>>2]|0; //@line 1916 "libraries/lodepng/lodepng.cpp"
                 $$pre$i154$i$i = $$lcssa78;$923 = $$pre270$i$i;
                }
                switch ($923|0) {
                case 16:  {
                 $924 = (($i$11208$i$i) + 1)|0; //@line 1916 "libraries/lodepng/lodepng.cpp"
                 $925 = (($bitlen_lld_e$sroa$0$0$lcssa$i$i) + ($924<<2)|0); //@line 1916 "libraries/lodepng/lodepng.cpp"
                 $926 = HEAP32[$925>>2]|0; //@line 1916 "libraries/lodepng/lodepng.cpp"
                 $927 = $$pre$i154$i$i & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $928 = ($927|0)==(0); //@line 432 "libraries/lodepng/lodepng.cpp"
                 do {
                  if ($928) {
                   $929 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                   $930 = (($929) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                   $931 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
                   $932 = ($931>>>0)<($930>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
                   if ($932) {
                    $933 = $931 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $934 = ($933>>>0)<($930>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                    $935 = ($930*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $936 = $935 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $937 = $934 ? $930 : $936; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $938 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                    $939 = (_realloc($938,$937)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                    $940 = ($939|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                    if ($940) {
                     break;
                    }
                    HEAP32[$3>>2] = $937; //@line 231 "libraries/lodepng/lodepng.cpp"
                    HEAP32[$v>>2] = $939; //@line 232 "libraries/lodepng/lodepng.cpp"
                    $942 = $939;
                   } else {
                    $$pre$i$i157$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                    $942 = $$pre$i$i157$i$i;
                   }
                   HEAP32[$2>>2] = $930; //@line 243 "libraries/lodepng/lodepng.cpp"
                   $941 = (($942) + ($929)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
                   HEAP8[$941>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
                  }
                 } while(0);
                 $943 = $926 & 1; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $944 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $945 = $944 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $946 = $943 << $945; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $947 = HEAP32[$2>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $948 = (($947) + -1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $949 = HEAP32[$v>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $950 = (($949) + ($948)|0); //@line 432 "libraries/lodepng/lodepng.cpp"
                 $951 = HEAP8[$950>>0]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $952 = $951&255; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $953 = $952 | $946; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $954 = $953&255; //@line 432 "libraries/lodepng/lodepng.cpp"
                 HEAP8[$950>>0] = $954; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $955 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $956 = (($955) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 HEAP32[$bp$i>>2] = $956; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $957 = $956 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $958 = ($957|0)==(0); //@line 432 "libraries/lodepng/lodepng.cpp"
                 do {
                  if ($958) {
                   $1134 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                   $1135 = (($1134) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                   $1136 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
                   $1137 = ($1136>>>0)<($1135>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
                   if ($1137) {
                    $1138 = $1136 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $1139 = ($1138>>>0)<($1135>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                    $1140 = ($1135*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $1141 = $1140 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $1142 = $1139 ? $1135 : $1141; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $1143 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                    $1144 = (_realloc($1143,$1142)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                    $1145 = ($1144|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                    if ($1145) {
                     break;
                    }
                    HEAP32[$3>>2] = $1142; //@line 231 "libraries/lodepng/lodepng.cpp"
                    HEAP32[$v>>2] = $1144; //@line 232 "libraries/lodepng/lodepng.cpp"
                    $1147 = $1144;
                   } else {
                    $$pre$i$i157$1$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                    $1147 = $$pre$i$i157$1$i$i;
                   }
                   HEAP32[$2>>2] = $1135; //@line 243 "libraries/lodepng/lodepng.cpp"
                   $1146 = (($1147) + ($1134)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
                   HEAP8[$1146>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
                  }
                 } while(0);
                 $1148 = $926 >>> 1; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1149 = $1148 & 1; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1150 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1151 = $1150 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1152 = $1149 << $1151; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1153 = HEAP32[$2>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1154 = (($1153) + -1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1155 = HEAP32[$v>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1156 = (($1155) + ($1154)|0); //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1157 = HEAP8[$1156>>0]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1158 = $1157&255; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1159 = $1158 | $1152; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1160 = $1159&255; //@line 432 "libraries/lodepng/lodepng.cpp"
                 HEAP8[$1156>>0] = $1160; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1161 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1162 = (($1161) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 HEAP32[$bp$i>>2] = $1162; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $$pre$i$i153268$i$i = $1162;$i$12$i$i = $924;
                 break;
                }
                case 17:  {
                 $959 = (($i$11208$i$i) + 1)|0; //@line 1917 "libraries/lodepng/lodepng.cpp"
                 $960 = (($bitlen_lld_e$sroa$0$0$lcssa$i$i) + ($959<<2)|0); //@line 1917 "libraries/lodepng/lodepng.cpp"
                 $961 = HEAP32[$960>>2]|0; //@line 1917 "libraries/lodepng/lodepng.cpp"
                 $962 = $$pre$i154$i$i & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $963 = ($962|0)==(0); //@line 432 "libraries/lodepng/lodepng.cpp"
                 do {
                  if ($963) {
                   $964 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                   $965 = (($964) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                   $966 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
                   $967 = ($966>>>0)<($965>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
                   if ($967) {
                    $968 = $966 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $969 = ($968>>>0)<($965>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                    $970 = ($965*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $971 = $970 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $972 = $969 ? $965 : $971; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $973 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                    $974 = (_realloc($973,$972)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                    $975 = ($974|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                    if ($975) {
                     break;
                    }
                    HEAP32[$3>>2] = $972; //@line 231 "libraries/lodepng/lodepng.cpp"
                    HEAP32[$v>>2] = $974; //@line 232 "libraries/lodepng/lodepng.cpp"
                    $977 = $974;
                   } else {
                    $$pre$i$i166$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                    $977 = $$pre$i$i166$i$i;
                   }
                   HEAP32[$2>>2] = $965; //@line 243 "libraries/lodepng/lodepng.cpp"
                   $976 = (($977) + ($964)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
                   HEAP8[$976>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
                  }
                 } while(0);
                 $978 = $961 & 1; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $979 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $980 = $979 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $981 = $978 << $980; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $982 = HEAP32[$2>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $983 = (($982) + -1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $984 = HEAP32[$v>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $985 = (($984) + ($983)|0); //@line 432 "libraries/lodepng/lodepng.cpp"
                 $986 = HEAP8[$985>>0]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $987 = $986&255; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $988 = $987 | $981; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $989 = $988&255; //@line 432 "libraries/lodepng/lodepng.cpp"
                 HEAP8[$985>>0] = $989; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $990 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $991 = (($990) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 HEAP32[$bp$i>>2] = $991; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $992 = $991 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $993 = ($992|0)==(0); //@line 432 "libraries/lodepng/lodepng.cpp"
                 do {
                  if ($993) {
                   $1074 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                   $1075 = (($1074) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                   $1076 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
                   $1077 = ($1076>>>0)<($1075>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
                   if ($1077) {
                    $1078 = $1076 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $1079 = ($1078>>>0)<($1075>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                    $1080 = ($1075*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $1081 = $1080 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $1082 = $1079 ? $1075 : $1081; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $1083 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                    $1084 = (_realloc($1083,$1082)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                    $1085 = ($1084|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                    if ($1085) {
                     break;
                    }
                    HEAP32[$3>>2] = $1082; //@line 231 "libraries/lodepng/lodepng.cpp"
                    HEAP32[$v>>2] = $1084; //@line 232 "libraries/lodepng/lodepng.cpp"
                    $1087 = $1084;
                   } else {
                    $$pre$i$i166$1$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                    $1087 = $$pre$i$i166$1$i$i;
                   }
                   HEAP32[$2>>2] = $1075; //@line 243 "libraries/lodepng/lodepng.cpp"
                   $1086 = (($1087) + ($1074)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
                   HEAP8[$1086>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
                  }
                 } while(0);
                 $1088 = $961 >>> 1; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1089 = $1088 & 1; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1090 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1091 = $1090 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1092 = $1089 << $1091; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1093 = HEAP32[$2>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1094 = (($1093) + -1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1095 = HEAP32[$v>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1096 = (($1095) + ($1094)|0); //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1097 = HEAP8[$1096>>0]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1098 = $1097&255; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1099 = $1098 | $1092; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1100 = $1099&255; //@line 432 "libraries/lodepng/lodepng.cpp"
                 HEAP8[$1096>>0] = $1100; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1101 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1102 = (($1101) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 HEAP32[$bp$i>>2] = $1102; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1103 = $1102 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1104 = ($1103|0)==(0); //@line 432 "libraries/lodepng/lodepng.cpp"
                 do {
                  if ($1104) {
                   $1105 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                   $1106 = (($1105) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                   $1107 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
                   $1108 = ($1107>>>0)<($1106>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
                   if ($1108) {
                    $1109 = $1107 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $1110 = ($1109>>>0)<($1106>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                    $1111 = ($1106*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $1112 = $1111 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $1113 = $1110 ? $1106 : $1112; //@line 227 "libraries/lodepng/lodepng.cpp"
                    $1114 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                    $1115 = (_realloc($1114,$1113)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                    $1116 = ($1115|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                    if ($1116) {
                     break;
                    }
                    HEAP32[$3>>2] = $1113; //@line 231 "libraries/lodepng/lodepng.cpp"
                    HEAP32[$v>>2] = $1115; //@line 232 "libraries/lodepng/lodepng.cpp"
                    $1118 = $1115;
                   } else {
                    $$pre$i$i166$2$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                    $1118 = $$pre$i$i166$2$i$i;
                   }
                   HEAP32[$2>>2] = $1106; //@line 243 "libraries/lodepng/lodepng.cpp"
                   $1117 = (($1118) + ($1105)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
                   HEAP8[$1117>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
                  }
                 } while(0);
                 $1119 = $961 >>> 2; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1120 = $1119 & 1; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1121 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1122 = $1121 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1123 = $1120 << $1122; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1124 = HEAP32[$2>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1125 = (($1124) + -1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1126 = HEAP32[$v>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1127 = (($1126) + ($1125)|0); //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1128 = HEAP8[$1127>>0]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1129 = $1128&255; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1130 = $1129 | $1123; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1131 = $1130&255; //@line 432 "libraries/lodepng/lodepng.cpp"
                 HEAP8[$1127>>0] = $1131; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1132 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $1133 = (($1132) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                 HEAP32[$bp$i>>2] = $1133; //@line 432 "libraries/lodepng/lodepng.cpp"
                 $$pre$i$i153268$i$i = $1133;$i$12$i$i = $959;
                 break;
                }
                case 18:  {
                 $994 = (($i$11208$i$i) + 1)|0; //@line 1918 "libraries/lodepng/lodepng.cpp"
                 $995 = (($bitlen_lld_e$sroa$0$0$lcssa$i$i) + ($994<<2)|0); //@line 1918 "libraries/lodepng/lodepng.cpp"
                 $996 = HEAP32[$995>>2]|0; //@line 1918 "libraries/lodepng/lodepng.cpp"
                 $998 = $$pre$i154$i$i;$i$01$i174$i$i = 0;
                 while(1) {
                  $997 = $998 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
                  $999 = ($997|0)==(0); //@line 432 "libraries/lodepng/lodepng.cpp"
                  do {
                   if ($999) {
                    $1000 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                    $1001 = (($1000) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                    $1002 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
                    $1003 = ($1002>>>0)<($1001>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
                    if ($1003) {
                     $1004 = $1002 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                     $1005 = ($1004>>>0)<($1001>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                     $1006 = ($1001*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                     $1007 = $1006 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                     $1008 = $1005 ? $1001 : $1007; //@line 227 "libraries/lodepng/lodepng.cpp"
                     $1009 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                     $1010 = (_realloc($1009,$1008)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                     $1011 = ($1010|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                     if ($1011) {
                      break;
                     }
                     HEAP32[$3>>2] = $1008; //@line 231 "libraries/lodepng/lodepng.cpp"
                     HEAP32[$v>>2] = $1010; //@line 232 "libraries/lodepng/lodepng.cpp"
                     $1013 = $1010;
                    } else {
                     $$pre$i$i175$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                     $1013 = $$pre$i$i175$i$i;
                    }
                    HEAP32[$2>>2] = $1001; //@line 243 "libraries/lodepng/lodepng.cpp"
                    $1012 = (($1013) + ($1000)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
                    HEAP8[$1012>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
                   }
                  } while(0);
                  $1014 = $996 >>> $i$01$i174$i$i; //@line 432 "libraries/lodepng/lodepng.cpp"
                  $1015 = $1014 & 1; //@line 432 "libraries/lodepng/lodepng.cpp"
                  $1016 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                  $1017 = $1016 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
                  $1018 = $1015 << $1017; //@line 432 "libraries/lodepng/lodepng.cpp"
                  $1019 = HEAP32[$2>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                  $1020 = (($1019) + -1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                  $1021 = HEAP32[$v>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                  $1022 = (($1021) + ($1020)|0); //@line 432 "libraries/lodepng/lodepng.cpp"
                  $1023 = HEAP8[$1022>>0]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                  $1024 = $1023&255; //@line 432 "libraries/lodepng/lodepng.cpp"
                  $1025 = $1024 | $1018; //@line 432 "libraries/lodepng/lodepng.cpp"
                  $1026 = $1025&255; //@line 432 "libraries/lodepng/lodepng.cpp"
                  HEAP8[$1022>>0] = $1026; //@line 432 "libraries/lodepng/lodepng.cpp"
                  $1027 = HEAP32[$bp$i>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                  $1028 = (($1027) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                  HEAP32[$bp$i>>2] = $1028; //@line 432 "libraries/lodepng/lodepng.cpp"
                  $1029 = (($i$01$i174$i$i) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
                  $1030 = ($1029|0)==(7); //@line 432 "libraries/lodepng/lodepng.cpp"
                  if ($1030) {
                   $$pre$i$i153268$i$i = $1028;$i$12$i$i = $994;
                   break;
                  } else {
                   $998 = $1028;$i$01$i174$i$i = $1029;
                  }
                 }
                 break;
                }
                default: {
                 $$pre$i$i153268$i$i = $$pre$i154$i$i;$i$12$i$i = $i$11208$i$i;
                }
                }
                $1031 = (($i$12$i$i) + 1)|0; //@line 1911 "libraries/lodepng/lodepng.cpp"
                $1032 = ($1031|0)==($bitlen_lld_e$sroa$40$0$lcssa$i$i|0); //@line 1911 "libraries/lodepng/lodepng.cpp"
                if ($1032) {
                 break L384;
                } else {
                 $$pre$i$i153$i$i = $$pre$i$i153268$i$i;$i$11208$i$i = $1031;
                }
               }
              }
             } while(0);
             __ZL13writeLZ77dataPjP8ucvectorPK8uivectorPK11HuffmanTreeS7_($bp$i,$v,$lz77_encoded$i9$i,$tree_ll$i10$i,$tree_d$i11$i); //@line 1922 "libraries/lodepng/lodepng.cpp"
             $tree_ll$idx5$val$i$i = HEAP32[$165>>2]|0; //@line 2026 "libraries/lodepng/lodepng.cpp"
             $1033 = ((($tree_ll$idx5$val$i$i)) + 1024|0); //@line 882 "libraries/lodepng/lodepng.cpp"
             $1034 = HEAP32[$1033>>2]|0; //@line 882 "libraries/lodepng/lodepng.cpp"
             $1035 = ($1034|0)==(0); //@line 1924 "libraries/lodepng/lodepng.cpp"
             if ($1035) {
              $bitlen_cl$sroa$0$6$i$i = $bitlen_cl$sroa$0$2$ph$i$i$lcssa;$bitlen_lld$sroa$0$6$i$i = $bitlen_lld$sroa$0$3$lcssa$i$i;$bitlen_lld_e$sroa$0$25$i$i = $bitlen_lld_e$sroa$0$0$lcssa$i$i;$error$1$i$i = 64;$frequencies_cl$sroa$0$2$i$i = $608;$frequencies_d$sroa$0$2$i$i = $384;$frequencies_ll$sroa$0$2$i$i = $382;
              break;
             }
             $tree_ll$idx$val$i29$i = HEAP32[$164>>2]|0; //@line 2026 "libraries/lodepng/lodepng.cpp"
             $1036 = ((($tree_ll$idx$val$i29$i)) + 1024|0); //@line 877 "libraries/lodepng/lodepng.cpp"
             $1037 = HEAP32[$1036>>2]|0; //@line 877 "libraries/lodepng/lodepng.cpp"
             $1038 = (($1034) + -1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
             $$pre$i$i181$i$i = HEAP32[$bp$i>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
             $1040 = $$pre$i$i181$i$i;$i$01$i$i183$i$i = 0;
             while(1) {
              $1039 = $1040 & 7; //@line 438 "libraries/lodepng/lodepng.cpp"
              $1041 = ($1039|0)==(0); //@line 438 "libraries/lodepng/lodepng.cpp"
              do {
               if ($1041) {
                $1042 = HEAP32[$2>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                $1043 = (($1042) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                $1044 = HEAP32[$3>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
                $1045 = ($1044>>>0)<($1043>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
                if ($1045) {
                 $1046 = $1044 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                 $1047 = ($1046>>>0)<($1043>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                 $1048 = ($1043*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                 $1049 = $1048 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                 $1050 = $1047 ? $1043 : $1049; //@line 227 "libraries/lodepng/lodepng.cpp"
                 $1051 = HEAP32[$v>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                 $1052 = (_realloc($1051,$1050)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                 $1053 = ($1052|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                 if ($1053) {
                  break;
                 }
                 HEAP32[$3>>2] = $1050; //@line 231 "libraries/lodepng/lodepng.cpp"
                 HEAP32[$v>>2] = $1052; //@line 232 "libraries/lodepng/lodepng.cpp"
                 $1055 = $1052;
                } else {
                 $$pre$i$i$i184$i$i = HEAP32[$v>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                 $1055 = $$pre$i$i$i184$i$i;
                }
                HEAP32[$2>>2] = $1043; //@line 243 "libraries/lodepng/lodepng.cpp"
                $1054 = (($1055) + ($1042)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
                HEAP8[$1054>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
               }
              } while(0);
              $1056 = (($1038) - ($i$01$i$i183$i$i))|0; //@line 438 "libraries/lodepng/lodepng.cpp"
              $1057 = $1037 >>> $1056; //@line 438 "libraries/lodepng/lodepng.cpp"
              $1058 = $1057 & 1; //@line 438 "libraries/lodepng/lodepng.cpp"
              $1059 = HEAP32[$bp$i>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
              $1060 = $1059 & 7; //@line 438 "libraries/lodepng/lodepng.cpp"
              $1061 = $1058 << $1060; //@line 438 "libraries/lodepng/lodepng.cpp"
              $1062 = HEAP32[$2>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
              $1063 = (($1062) + -1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
              $1064 = HEAP32[$v>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
              $1065 = (($1064) + ($1063)|0); //@line 438 "libraries/lodepng/lodepng.cpp"
              $1066 = HEAP8[$1065>>0]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
              $1067 = $1066&255; //@line 438 "libraries/lodepng/lodepng.cpp"
              $1068 = $1067 | $1061; //@line 438 "libraries/lodepng/lodepng.cpp"
              $1069 = $1068&255; //@line 438 "libraries/lodepng/lodepng.cpp"
              HEAP8[$1065>>0] = $1069; //@line 438 "libraries/lodepng/lodepng.cpp"
              $1070 = HEAP32[$bp$i>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
              $1071 = (($1070) + 1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
              HEAP32[$bp$i>>2] = $1071; //@line 438 "libraries/lodepng/lodepng.cpp"
              $1072 = (($i$01$i$i183$i$i) + 1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
              $1073 = ($1072|0)==($1034|0); //@line 438 "libraries/lodepng/lodepng.cpp"
              if ($1073) {
               $bitlen_cl$sroa$0$6$i$i = $bitlen_cl$sroa$0$2$ph$i$i$lcssa;$bitlen_lld$sroa$0$6$i$i = $bitlen_lld$sroa$0$3$lcssa$i$i;$bitlen_lld_e$sroa$0$25$i$i = $bitlen_lld_e$sroa$0$0$lcssa$i$i;$error$1$i$i = 0;$frequencies_cl$sroa$0$2$i$i = $608;$frequencies_d$sroa$0$2$i$i = $384;$frequencies_ll$sroa$0$2$i$i = $382;
               break;
              } else {
               $1040 = $1071;$i$01$i$i183$i$i = $1072;
              }
             }
            } else {
             $bitlen_cl$sroa$0$6$i$i = 0;$bitlen_lld$sroa$0$6$i$i = 0;$bitlen_lld_e$sroa$0$25$i$i = 0;$error$1$i$i = $427;$frequencies_cl$sroa$0$2$i$i = 0;$frequencies_d$sroa$0$2$i$i = $384;$frequencies_ll$sroa$0$2$i$i = $382;
            }
           } else {
            $bitlen_cl$sroa$0$6$i$i = 0;$bitlen_lld$sroa$0$6$i$i = 0;$bitlen_lld_e$sroa$0$25$i$i = 0;$error$1$i$i = $425;$frequencies_cl$sroa$0$2$i$i = 0;$frequencies_d$sroa$0$2$i$i = $384;$frequencies_ll$sroa$0$2$i$i = $382;
           }
          }
         } else {
          $bitlen_cl$sroa$0$6$i$i = 0;$bitlen_lld$sroa$0$6$i$i = 0;$bitlen_lld_e$sroa$0$25$i$i = 0;$error$1$i$i = $415;$frequencies_cl$sroa$0$2$i$i = 0;$frequencies_d$sroa$0$2$i$i = $384;$frequencies_ll$sroa$0$2$i$i = $382;
         }
        } else {
         $bitlen_cl$sroa$0$6$i$i = 0;$bitlen_lld$sroa$0$6$i$i = 0;$bitlen_lld_e$sroa$0$25$i$i = 0;$error$1$i$i = $413;$frequencies_cl$sroa$0$2$i$i = 0;$frequencies_d$sroa$0$2$i$i = $384;$frequencies_ll$sroa$0$2$i$i = $382;
        }
       }
      }
     }
    }
   } while(0);
   HEAP32[$162>>2] = 0; //@line 147 "libraries/lodepng/lodepng.cpp"
   HEAP32[$163>>2] = 0; //@line 147 "libraries/lodepng/lodepng.cpp"
   $1225 = HEAP32[$lz77_encoded$i9$i>>2]|0; //@line 148 "libraries/lodepng/lodepng.cpp"
   _free($1225); //@line 76 "libraries/lodepng/lodepng.cpp"
   HEAP32[$lz77_encoded$i9$i>>2] = 0; //@line 149 "libraries/lodepng/lodepng.cpp"
   $1226 = HEAP32[$tree_ll$i10$i>>2]|0; //@line 539 "libraries/lodepng/lodepng.cpp"
   _free($1226); //@line 76 "libraries/lodepng/lodepng.cpp"
   $1227 = HEAP32[$164>>2]|0; //@line 540 "libraries/lodepng/lodepng.cpp"
   _free($1227); //@line 76 "libraries/lodepng/lodepng.cpp"
   $1228 = HEAP32[$165>>2]|0; //@line 541 "libraries/lodepng/lodepng.cpp"
   _free($1228); //@line 76 "libraries/lodepng/lodepng.cpp"
   $1229 = HEAP32[$tree_d$i11$i>>2]|0; //@line 539 "libraries/lodepng/lodepng.cpp"
   _free($1229); //@line 76 "libraries/lodepng/lodepng.cpp"
   $1230 = HEAP32[$166>>2]|0; //@line 540 "libraries/lodepng/lodepng.cpp"
   _free($1230); //@line 76 "libraries/lodepng/lodepng.cpp"
   $1231 = HEAP32[$167>>2]|0; //@line 541 "libraries/lodepng/lodepng.cpp"
   _free($1231); //@line 76 "libraries/lodepng/lodepng.cpp"
   $1232 = HEAP32[$tree_cl$i$i>>2]|0; //@line 539 "libraries/lodepng/lodepng.cpp"
   _free($1232); //@line 76 "libraries/lodepng/lodepng.cpp"
   $1233 = HEAP32[$168>>2]|0; //@line 540 "libraries/lodepng/lodepng.cpp"
   _free($1233); //@line 76 "libraries/lodepng/lodepng.cpp"
   $1234 = HEAP32[$169>>2]|0; //@line 541 "libraries/lodepng/lodepng.cpp"
   _free($1234); //@line 76 "libraries/lodepng/lodepng.cpp"
   _free($frequencies_ll$sroa$0$2$i$i); //@line 76 "libraries/lodepng/lodepng.cpp"
   _free($frequencies_d$sroa$0$2$i$i); //@line 76 "libraries/lodepng/lodepng.cpp"
   _free($frequencies_cl$sroa$0$2$i$i); //@line 76 "libraries/lodepng/lodepng.cpp"
   _free($bitlen_lld_e$sroa$0$25$i$i); //@line 76 "libraries/lodepng/lodepng.cpp"
   _free($bitlen_lld$sroa$0$6$i$i); //@line 76 "libraries/lodepng/lodepng.cpp"
   _free($bitlen_cl$sroa$0$6$i$i); //@line 76 "libraries/lodepng/lodepng.cpp"
   $error$1$i = $error$1$i$i;
   break;
  }
  default: {
   $error$1$i = 0;
  }
  }
  $1235 = (($i$042$i) + 1)|0; //@line 2018 "libraries/lodepng/lodepng.cpp"
  $1236 = ($1235|0)!=($$1$i|0); //@line 2018 "libraries/lodepng/lodepng.cpp"
  $1237 = ($error$1$i|0)==(0); //@line 2018 "libraries/lodepng/lodepng.cpp"
  $or$cond$i = $1236 & $1237; //@line 2018 "libraries/lodepng/lodepng.cpp"
  if (!($or$cond$i)) {
   $error$1$i$lcssa = $error$1$i;
   break;
  }
  $$pre34 = HEAP32[$settings>>2]|0; //@line 2025 "libraries/lodepng/lodepng.cpp"
  $181 = $$pre34;$i$042$i = $1235;
 }
 $$pre56$i = HEAP32[$hash$i>>2]|0; //@line 1391 "libraries/lodepng/lodepng.cpp"
 $$pre57$i = HEAP32[$129>>2]|0; //@line 1393 "libraries/lodepng/lodepng.cpp"
 $$pre58$i = HEAP32[$131>>2]|0; //@line 1395 "libraries/lodepng/lodepng.cpp"
 $$pre59$i = HEAP32[$133>>2]|0; //@line 1396 "libraries/lodepng/lodepng.cpp"
 _free($$pre56$i); //@line 76 "libraries/lodepng/lodepng.cpp"
 $1238 = HEAP32[$126>>2]|0; //@line 1392 "libraries/lodepng/lodepng.cpp"
 _free($1238); //@line 76 "libraries/lodepng/lodepng.cpp"
 _free($$pre57$i); //@line 76 "libraries/lodepng/lodepng.cpp"
 _free($$pre58$i); //@line 76 "libraries/lodepng/lodepng.cpp"
 _free($$pre59$i); //@line 76 "libraries/lodepng/lodepng.cpp"
 $1239 = HEAP32[$135>>2]|0; //@line 1397 "libraries/lodepng/lodepng.cpp"
 _free($1239); //@line 76 "libraries/lodepng/lodepng.cpp"
 $$0$i = $error$1$i$lcssa;
 $1240 = HEAP32[$v>>2]|0; //@line 2042 "libraries/lodepng/lodepng.cpp"
 HEAP32[$out>>2] = $1240; //@line 2042 "libraries/lodepng/lodepng.cpp"
 $1241 = HEAP32[$2>>2]|0; //@line 2043 "libraries/lodepng/lodepng.cpp"
 HEAP32[$outsize>>2] = $1241; //@line 2043 "libraries/lodepng/lodepng.cpp"
 STACKTOP = sp;return ($$0$i|0); //@line 2044 "libraries/lodepng/lodepng.cpp"
}
function __ZL10encodeLZ77P8uivectorP4HashPKhjjjjjj($out,$hash,$in,$inpos,$insize,$windowsize,$minmatch,$nicematch,$lazymatching) {
 $out = $out|0;
 $hash = $hash|0;
 $in = $in|0;
 $inpos = $inpos|0;
 $insize = $insize|0;
 $windowsize = $windowsize|0;
 $minmatch = $minmatch|0;
 $nicematch = $nicematch|0;
 $lazymatching = $lazymatching|0;
 var $$$i = 0, $$$i23 = 0, $$0$lcssa$i = 0, $$0$lcssa$i26 = 0, $$01$i = 0, $$01$i24 = 0, $$012 = 0, $$1$i2141 = 0, $$2$i$i = 0, $$2$i6$i = 0, $$lcssa134 = 0, $$lcssa136 = 0, $$nicematch = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i15$i = 0, $$pre$i21$i = 0, $$pre$i30 = 0, $$pre$i36 = 0, $$pre$i9$i = 0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0;
 var $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0;
 var $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0;
 var $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0;
 var $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0;
 var $404 = 0, $405 = 0, $406 = 0, $407 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $backptr$050 = 0, $backptr$050$us = 0, $exitcond = 0, $foreptr$0$lcssa = 0, $foreptr$0$lcssa$us = 0, $foreptr$049 = 0, $foreptr$049$us = 0, $hashpos$0$be$us = 0, $hashpos$056 = 0, $hashpos$056$us = 0, $i$03$i = 0, $i$03$i16 = 0;
 var $i$075 = 0, $lazy$082 = 0, $lazy$1 = 0, $lazy$2 = 0, $lazylength$083 = 0, $lazylength$1 = 0, $lazyoffset$084 = 0, $lazyoffset$1 = 0, $left$0$ph$i$i = 0, $left$0$ph$i1$i = 0, $length$054 = 0, $length$054$us = 0, $length$1 = 0, $length$1$us = 0, $length$2 = 0, $length$3 = 0, $numzeros$081 = 0, $numzeros$1 = 0, $numzeros$1$$us = 0, $numzeros$276 = 0;
 var $numzeros$3 = 0, $numzeros$3$lcssa = 0, $numzeros$5 = 0, $offset$053 = 0, $offset$053$us = 0, $offset$1 = 0, $offset$1$us = 0, $offset$2 = 0, $offset$3 = 0, $or$cond14 = 0, $or$cond5 = 0, $or$cond7 = 0, $or$cond9 = 0, $pos$085 = 0, $pos$1 = 0, $pos$277 = 0, $pos$4 = 0, $prev_offset$055 = 0, $prev_offset$055$us = 0, $result$02$i = 0;
 var $result$02$i17 = 0, $result$2$i = 0, $result$2$i19 = 0, $right$0$i$i = 0, $right$0$i$i$lcssa133 = 0, $right$0$i4$i = 0, $right$0$i4$i$lcssa135 = 0, $right$0$ph$i$i = 0, $right$0$ph$i2$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($windowsize>>>0)>(8191); //@line 1462 "libraries/lodepng/lodepng.cpp"
 $1 = $windowsize >>> 3; //@line 1462 "libraries/lodepng/lodepng.cpp"
 $2 = $0 ? $windowsize : $1; //@line 1462 "libraries/lodepng/lodepng.cpp"
 $3 = $0 ? 258 : 64; //@line 1463 "libraries/lodepng/lodepng.cpp"
 $4 = (($windowsize) + -1)|0; //@line 1478 "libraries/lodepng/lodepng.cpp"
 $5 = ($4>>>0)>(32767); //@line 1478 "libraries/lodepng/lodepng.cpp"
 if ($5) {
  $$012 = 60;
  return ($$012|0); //@line 1627 "libraries/lodepng/lodepng.cpp"
 }
 $6 = $4 & $windowsize; //@line 1479 "libraries/lodepng/lodepng.cpp"
 $7 = ($6|0)==(0); //@line 1479 "libraries/lodepng/lodepng.cpp"
 if (!($7)) {
  $$012 = 90;
  return ($$012|0); //@line 1627 "libraries/lodepng/lodepng.cpp"
 }
 $8 = ($nicematch>>>0)>(258); //@line 1481 "libraries/lodepng/lodepng.cpp"
 $$nicematch = $8 ? 258 : $nicematch; //@line 1481 "libraries/lodepng/lodepng.cpp"
 $9 = ($inpos>>>0)<($insize>>>0); //@line 1483 "libraries/lodepng/lodepng.cpp"
 if (!($9)) {
  $$012 = 0;
  return ($$012|0); //@line 1627 "libraries/lodepng/lodepng.cpp"
 }
 $10 = (($in) + ($insize)|0); //@line 1427 "libraries/lodepng/lodepng.cpp"
 $11 = ((($hash)) + 8|0); //@line 1437 "libraries/lodepng/lodepng.cpp"
 $12 = ((($hash)) + 20|0); //@line 1441 "libraries/lodepng/lodepng.cpp"
 $13 = ((($hash)) + 12|0); //@line 1442 "libraries/lodepng/lodepng.cpp"
 $14 = ((($hash)) + 4|0); //@line 1506 "libraries/lodepng/lodepng.cpp"
 $15 = ((($hash)) + 16|0); //@line 1554 "libraries/lodepng/lodepng.cpp"
 $16 = ($2|0)==(0); //@line 1514 "libraries/lodepng/lodepng.cpp"
 $17 = ($lazymatching|0)==(0); //@line 1563 "libraries/lodepng/lodepng.cpp"
 $18 = ((($out)) + 4|0); //@line 196 "libraries/lodepng/lodepng.cpp"
 $19 = ((($out)) + 8|0); //@line 155 "libraries/lodepng/lodepng.cpp"
 $lazy$082 = 0;$lazylength$083 = 0;$lazyoffset$084 = 0;$numzeros$081 = 0;$pos$085 = $inpos;
 L10: while(1) {
  $20 = $pos$085 & $4; //@line 1485 "libraries/lodepng/lodepng.cpp"
  $21 = (($pos$085) + 2)|0; //@line 1405 "libraries/lodepng/lodepng.cpp"
  $22 = ($21>>>0)<($insize>>>0); //@line 1405 "libraries/lodepng/lodepng.cpp"
  if ($22) {
   $23 = (($in) + ($pos$085)|0); //@line 1411 "libraries/lodepng/lodepng.cpp"
   $24 = HEAP8[$23>>0]|0; //@line 1411 "libraries/lodepng/lodepng.cpp"
   $25 = $24&255; //@line 1411 "libraries/lodepng/lodepng.cpp"
   $26 = (($pos$085) + 1)|0; //@line 1412 "libraries/lodepng/lodepng.cpp"
   $27 = (($in) + ($26)|0); //@line 1412 "libraries/lodepng/lodepng.cpp"
   $28 = HEAP8[$27>>0]|0; //@line 1412 "libraries/lodepng/lodepng.cpp"
   $29 = $28&255; //@line 1412 "libraries/lodepng/lodepng.cpp"
   $30 = $29 << 4; //@line 1412 "libraries/lodepng/lodepng.cpp"
   $31 = $30 ^ $25; //@line 1412 "libraries/lodepng/lodepng.cpp"
   $32 = (($in) + ($21)|0); //@line 1413 "libraries/lodepng/lodepng.cpp"
   $33 = HEAP8[$32>>0]|0; //@line 1413 "libraries/lodepng/lodepng.cpp"
   $34 = $33&255; //@line 1413 "libraries/lodepng/lodepng.cpp"
   $35 = $34 << 8; //@line 1413 "libraries/lodepng/lodepng.cpp"
   $36 = $31 ^ $35; //@line 1413 "libraries/lodepng/lodepng.cpp"
   $result$2$i = $36;
   label = 9;
  } else {
   $37 = (($insize) - ($pos$085))|0; //@line 1417 "libraries/lodepng/lodepng.cpp"
   $38 = ($pos$085|0)==($insize|0); //@line 1418 "libraries/lodepng/lodepng.cpp"
   if ($38) {
    label = 10;
   } else {
    $i$03$i = 0;$result$02$i = 0;
    while(1) {
     $39 = (($i$03$i) + ($pos$085))|0; //@line 1418 "libraries/lodepng/lodepng.cpp"
     $40 = (($in) + ($39)|0); //@line 1418 "libraries/lodepng/lodepng.cpp"
     $41 = HEAP8[$40>>0]|0; //@line 1418 "libraries/lodepng/lodepng.cpp"
     $42 = $41&255; //@line 1418 "libraries/lodepng/lodepng.cpp"
     $43 = $i$03$i << 3; //@line 1418 "libraries/lodepng/lodepng.cpp"
     $44 = $42 << $43; //@line 1418 "libraries/lodepng/lodepng.cpp"
     $45 = $44 ^ $result$02$i; //@line 1418 "libraries/lodepng/lodepng.cpp"
     $46 = (($i$03$i) + 1)|0; //@line 1418 "libraries/lodepng/lodepng.cpp"
     $47 = ($46|0)==($37|0); //@line 1418 "libraries/lodepng/lodepng.cpp"
     if ($47) {
      $result$2$i = $45;
      label = 9;
      break;
     } else {
      $i$03$i = $46;$result$02$i = $45;
     }
    }
   }
  }
  if ((label|0) == 9) {
   label = 0;
   $48 = $result$2$i & 65535; //@line 1420 "libraries/lodepng/lodepng.cpp"
   $49 = ($48|0)==(0); //@line 1490 "libraries/lodepng/lodepng.cpp"
   if ($49) {
    label = 10;
   } else {
    $72 = $48;$numzeros$1 = 0;
   }
  }
  do {
   if ((label|0) == 10) {
    label = 0;
    $50 = ($numzeros$081|0)==(0); //@line 1492 "libraries/lodepng/lodepng.cpp"
    if (!($50)) {
     $62 = (($pos$085) + ($numzeros$081))|0; //@line 1493 "libraries/lodepng/lodepng.cpp"
     $63 = ($62>>>0)>($insize>>>0); //@line 1493 "libraries/lodepng/lodepng.cpp"
     if (!($63)) {
      $64 = (($62) + -1)|0; //@line 1493 "libraries/lodepng/lodepng.cpp"
      $65 = (($in) + ($64)|0); //@line 1493 "libraries/lodepng/lodepng.cpp"
      $66 = HEAP8[$65>>0]|0; //@line 1493 "libraries/lodepng/lodepng.cpp"
      $67 = ($66<<24>>24)==(0); //@line 1493 "libraries/lodepng/lodepng.cpp"
      if ($67) {
       $72 = 0;$numzeros$1 = $numzeros$081;
       break;
      }
     }
     $68 = (($numzeros$081) + -1)|0; //@line 1493 "libraries/lodepng/lodepng.cpp"
     $72 = 0;$numzeros$1 = $68;
     break;
    }
    $51 = (($in) + ($pos$085)|0); //@line 1425 "libraries/lodepng/lodepng.cpp"
    $52 = ((($51)) + 258|0); //@line 1426 "libraries/lodepng/lodepng.cpp"
    $53 = ($52>>>0)>($10>>>0); //@line 1427 "libraries/lodepng/lodepng.cpp"
    $$$i23 = $53 ? $10 : $52; //@line 1427 "libraries/lodepng/lodepng.cpp"
    $54 = ($51|0)==($$$i23|0); //@line 1429 "libraries/lodepng/lodepng.cpp"
    L27: do {
     if ($54) {
      $$0$lcssa$i26 = $51;
     } else {
      $$01$i24 = $51;
      while(1) {
       $55 = HEAP8[$$01$i24>>0]|0; //@line 1429 "libraries/lodepng/lodepng.cpp"
       $56 = ($55<<24>>24)==(0); //@line 1429 "libraries/lodepng/lodepng.cpp"
       if (!($56)) {
        $$0$lcssa$i26 = $$01$i24;
        break L27;
       }
       $57 = ((($$01$i24)) + 1|0); //@line 1429 "libraries/lodepng/lodepng.cpp"
       $58 = ($57|0)==($$$i23|0); //@line 1429 "libraries/lodepng/lodepng.cpp"
       if ($58) {
        $$0$lcssa$i26 = $$$i23;
        break;
       } else {
        $$01$i24 = $57;
       }
      }
     }
    } while(0);
    $59 = $$0$lcssa$i26; //@line 1431 "libraries/lodepng/lodepng.cpp"
    $60 = $51; //@line 1431 "libraries/lodepng/lodepng.cpp"
    $61 = (($59) - ($60))|0; //@line 1431 "libraries/lodepng/lodepng.cpp"
    $72 = 0;$numzeros$1 = $61;
   }
  } while(0);
  $69 = $numzeros$1&65535; //@line 1500 "libraries/lodepng/lodepng.cpp"
  $70 = HEAP32[$11>>2]|0; //@line 1437 "libraries/lodepng/lodepng.cpp"
  $71 = (($70) + ($20<<2)|0); //@line 1437 "libraries/lodepng/lodepng.cpp"
  HEAP32[$71>>2] = $72; //@line 1437 "libraries/lodepng/lodepng.cpp"
  $73 = HEAP32[$hash>>2]|0; //@line 1585 "libraries/lodepng/lodepng.cpp"
  $74 = (($73) + ($72<<2)|0); //@line 1438 "libraries/lodepng/lodepng.cpp"
  $75 = HEAP32[$74>>2]|0; //@line 1438 "libraries/lodepng/lodepng.cpp"
  $76 = ($75|0)==(-1); //@line 1438 "libraries/lodepng/lodepng.cpp"
  if (!($76)) {
   $77 = $75&65535; //@line 1438 "libraries/lodepng/lodepng.cpp"
   $78 = HEAP32[$14>>2]|0; //@line 1438 "libraries/lodepng/lodepng.cpp"
   $79 = (($78) + ($20<<1)|0); //@line 1438 "libraries/lodepng/lodepng.cpp"
   HEAP16[$79>>1] = $77; //@line 1438 "libraries/lodepng/lodepng.cpp"
  }
  HEAP32[$74>>2] = $20; //@line 1439 "libraries/lodepng/lodepng.cpp"
  $80 = HEAP32[$12>>2]|0; //@line 1441 "libraries/lodepng/lodepng.cpp"
  $81 = (($80) + ($20<<1)|0); //@line 1441 "libraries/lodepng/lodepng.cpp"
  HEAP16[$81>>1] = $69; //@line 1441 "libraries/lodepng/lodepng.cpp"
  $82 = $numzeros$1 & 65535; //@line 1442 "libraries/lodepng/lodepng.cpp"
  $83 = HEAP32[$13>>2]|0; //@line 1586 "libraries/lodepng/lodepng.cpp"
  $84 = (($83) + ($82<<2)|0); //@line 1442 "libraries/lodepng/lodepng.cpp"
  $85 = HEAP32[$84>>2]|0; //@line 1442 "libraries/lodepng/lodepng.cpp"
  $86 = ($85|0)==(-1); //@line 1442 "libraries/lodepng/lodepng.cpp"
  if (!($86)) {
   $87 = $85&65535; //@line 1442 "libraries/lodepng/lodepng.cpp"
   $88 = HEAP32[$15>>2]|0; //@line 1442 "libraries/lodepng/lodepng.cpp"
   $89 = (($88) + ($20<<1)|0); //@line 1442 "libraries/lodepng/lodepng.cpp"
   HEAP16[$89>>1] = $87; //@line 1442 "libraries/lodepng/lodepng.cpp"
  }
  HEAP32[$84>>2] = $20; //@line 1443 "libraries/lodepng/lodepng.cpp"
  $90 = HEAP32[$14>>2]|0; //@line 1506 "libraries/lodepng/lodepng.cpp"
  $91 = (($90) + ($20<<1)|0); //@line 1506 "libraries/lodepng/lodepng.cpp"
  $92 = HEAP16[$91>>1]|0; //@line 1506 "libraries/lodepng/lodepng.cpp"
  $93 = $92&65535; //@line 1506 "libraries/lodepng/lodepng.cpp"
  $94 = (($pos$085) + 258)|0; //@line 1508 "libraries/lodepng/lodepng.cpp"
  $95 = ($94>>>0)>($insize>>>0); //@line 1508 "libraries/lodepng/lodepng.cpp"
  $96 = $95 ? $insize : $94; //@line 1508 "libraries/lodepng/lodepng.cpp"
  $97 = (($in) + ($96)|0); //@line 1508 "libraries/lodepng/lodepng.cpp"
  L39: do {
   if ($16) {
    $length$2 = 0;$offset$2 = 0;
   } else {
    $98 = ($numzeros$1>>>0)>(2); //@line 1553 "libraries/lodepng/lodepng.cpp"
    $99 = (($in) + ($pos$085)|0); //@line 1522 "libraries/lodepng/lodepng.cpp"
    $100 = $99; //@line 1539 "libraries/lodepng/lodepng.cpp"
    if ($98) {
     $144 = 1;$hashpos$056$us = $93;$length$054$us = 0;$offset$053$us = 0;$prev_offset$055$us = 0;
    } else {
     $101 = ($pos$085|0)==($96|0); //@line 1534 "libraries/lodepng/lodepng.cpp"
     $172 = 1;$hashpos$056 = $93;$length$054 = 0;$offset$053 = 0;$prev_offset$055 = 0;
     while(1) {
      $146 = ($20>>>0)>=($hashpos$056>>>0); //@line 1515 "libraries/lodepng/lodepng.cpp"
      $147 = (($20) - ($hashpos$056))|0; //@line 1515 "libraries/lodepng/lodepng.cpp"
      $148 = $146 ? 0 : $windowsize; //@line 1515 "libraries/lodepng/lodepng.cpp"
      $149 = (($148) + ($147))|0; //@line 1515 "libraries/lodepng/lodepng.cpp"
      $150 = ($149>>>0)<($prev_offset$055>>>0); //@line 1517 "libraries/lodepng/lodepng.cpp"
      if ($150) {
       $length$2 = $length$054;$offset$2 = $offset$053;
       break L39;
      }
      $151 = ($149|0)==(0); //@line 1519 "libraries/lodepng/lodepng.cpp"
      if ($151) {
       $length$1 = $length$054;$offset$1 = $offset$053;
      } else {
       L48: do {
        if ($101) {
         $foreptr$0$lcssa = $99;
        } else {
         $152 = (($pos$085) - ($149))|0; //@line 1523 "libraries/lodepng/lodepng.cpp"
         $153 = (($in) + ($152)|0); //@line 1523 "libraries/lodepng/lodepng.cpp"
         $backptr$050 = $153;$foreptr$049 = $99;
         while(1) {
          $154 = HEAP8[$backptr$050>>0]|0; //@line 1534 "libraries/lodepng/lodepng.cpp"
          $155 = HEAP8[$foreptr$049>>0]|0; //@line 1534 "libraries/lodepng/lodepng.cpp"
          $156 = ($154<<24>>24)==($155<<24>>24); //@line 1534 "libraries/lodepng/lodepng.cpp"
          if (!($156)) {
           $foreptr$0$lcssa = $foreptr$049;
           break L48;
          }
          $157 = ((($backptr$050)) + 1|0); //@line 1536 "libraries/lodepng/lodepng.cpp"
          $158 = ((($foreptr$049)) + 1|0); //@line 1537 "libraries/lodepng/lodepng.cpp"
          $159 = ($158|0)==($97|0); //@line 1534 "libraries/lodepng/lodepng.cpp"
          if ($159) {
           $foreptr$0$lcssa = $97;
           break;
          } else {
           $backptr$050 = $157;$foreptr$049 = $158;
          }
         }
        }
       } while(0);
       $160 = $foreptr$0$lcssa; //@line 1539 "libraries/lodepng/lodepng.cpp"
       $161 = (($160) - ($100))|0; //@line 1539 "libraries/lodepng/lodepng.cpp"
       $162 = ($161>>>0)>($length$054>>>0); //@line 1541 "libraries/lodepng/lodepng.cpp"
       if ($162) {
        $163 = ($161>>>0)<($$nicematch>>>0); //@line 1547 "libraries/lodepng/lodepng.cpp"
        if ($163) {
         $length$1 = $161;$offset$1 = $149;
        } else {
         $length$2 = $161;$offset$2 = $149;
         break L39;
        }
       } else {
        $length$1 = $length$054;$offset$1 = $offset$053;
       }
      }
      $164 = (($90) + ($hashpos$056<<1)|0); //@line 1551 "libraries/lodepng/lodepng.cpp"
      $165 = HEAP16[$164>>1]|0; //@line 1551 "libraries/lodepng/lodepng.cpp"
      $166 = $165&65535; //@line 1551 "libraries/lodepng/lodepng.cpp"
      $167 = ($hashpos$056|0)==($166|0); //@line 1551 "libraries/lodepng/lodepng.cpp"
      if ($167) {
       $length$2 = $length$1;$offset$2 = $offset$1;
       break L39;
      }
      $168 = (($70) + ($166<<2)|0); //@line 1559 "libraries/lodepng/lodepng.cpp"
      $169 = HEAP32[$168>>2]|0; //@line 1559 "libraries/lodepng/lodepng.cpp"
      $170 = ($169|0)==($72|0); //@line 1559 "libraries/lodepng/lodepng.cpp"
      if (!($170)) {
       $length$2 = $length$1;$offset$2 = $offset$1;
       break L39;
      }
      $171 = (($172) + 1)|0; //@line 1514 "libraries/lodepng/lodepng.cpp"
      $173 = ($172>>>0)<($2>>>0); //@line 1514 "libraries/lodepng/lodepng.cpp"
      if ($173) {
       $172 = $171;$hashpos$056 = $166;$length$054 = $length$1;$offset$053 = $offset$1;$prev_offset$055 = $149;
      } else {
       $length$2 = $length$1;$offset$2 = $offset$1;
       break L39;
      }
     }
    }
    while(1) {
     $102 = ($20>>>0)>=($hashpos$056$us>>>0); //@line 1515 "libraries/lodepng/lodepng.cpp"
     $103 = (($20) - ($hashpos$056$us))|0; //@line 1515 "libraries/lodepng/lodepng.cpp"
     $104 = $102 ? 0 : $windowsize; //@line 1515 "libraries/lodepng/lodepng.cpp"
     $105 = (($104) + ($103))|0; //@line 1515 "libraries/lodepng/lodepng.cpp"
     $106 = ($105>>>0)<($prev_offset$055$us>>>0); //@line 1517 "libraries/lodepng/lodepng.cpp"
     if ($106) {
      $length$2 = $length$054$us;$offset$2 = $offset$053$us;
      break L39;
     }
     $107 = ($105|0)==(0); //@line 1519 "libraries/lodepng/lodepng.cpp"
     if ($107) {
      $length$1$us = $length$054$us;$offset$1$us = $offset$053$us;
     } else {
      $108 = (($80) + ($hashpos$056$us<<1)|0); //@line 1528 "libraries/lodepng/lodepng.cpp"
      $109 = HEAP16[$108>>1]|0; //@line 1528 "libraries/lodepng/lodepng.cpp"
      $110 = $109&65535; //@line 1528 "libraries/lodepng/lodepng.cpp"
      $111 = ($110>>>0)>($numzeros$1>>>0); //@line 1529 "libraries/lodepng/lodepng.cpp"
      $numzeros$1$$us = $111 ? $numzeros$1 : $110; //@line 1529 "libraries/lodepng/lodepng.cpp"
      $112 = (($99) + ($numzeros$1$$us)|0); //@line 1531 "libraries/lodepng/lodepng.cpp"
      $113 = ($112|0)==($97|0); //@line 1534 "libraries/lodepng/lodepng.cpp"
      L63: do {
       if ($113) {
        $foreptr$0$lcssa$us = $97;
       } else {
        $114 = (($pos$085) - ($105))|0; //@line 1523 "libraries/lodepng/lodepng.cpp"
        $115 = (($in) + ($114)|0); //@line 1523 "libraries/lodepng/lodepng.cpp"
        $116 = (($115) + ($numzeros$1$$us)|0); //@line 1530 "libraries/lodepng/lodepng.cpp"
        $backptr$050$us = $116;$foreptr$049$us = $112;
        while(1) {
         $117 = HEAP8[$backptr$050$us>>0]|0; //@line 1534 "libraries/lodepng/lodepng.cpp"
         $118 = HEAP8[$foreptr$049$us>>0]|0; //@line 1534 "libraries/lodepng/lodepng.cpp"
         $119 = ($117<<24>>24)==($118<<24>>24); //@line 1534 "libraries/lodepng/lodepng.cpp"
         if (!($119)) {
          $foreptr$0$lcssa$us = $foreptr$049$us;
          break L63;
         }
         $120 = ((($backptr$050$us)) + 1|0); //@line 1536 "libraries/lodepng/lodepng.cpp"
         $121 = ((($foreptr$049$us)) + 1|0); //@line 1537 "libraries/lodepng/lodepng.cpp"
         $122 = ($121|0)==($97|0); //@line 1534 "libraries/lodepng/lodepng.cpp"
         if ($122) {
          $foreptr$0$lcssa$us = $97;
          break;
         } else {
          $backptr$050$us = $120;$foreptr$049$us = $121;
         }
        }
       }
      } while(0);
      $123 = $foreptr$0$lcssa$us; //@line 1539 "libraries/lodepng/lodepng.cpp"
      $124 = (($123) - ($100))|0; //@line 1539 "libraries/lodepng/lodepng.cpp"
      $125 = ($124>>>0)>($length$054$us>>>0); //@line 1541 "libraries/lodepng/lodepng.cpp"
      if ($125) {
       $126 = ($124>>>0)<($$nicematch>>>0); //@line 1547 "libraries/lodepng/lodepng.cpp"
       if ($126) {
        $length$1$us = $124;$offset$1$us = $105;
       } else {
        $length$2 = $124;$offset$2 = $105;
        break L39;
       }
      } else {
       $length$1$us = $length$054$us;$offset$1$us = $offset$053$us;
      }
     }
     $127 = (($90) + ($hashpos$056$us<<1)|0); //@line 1551 "libraries/lodepng/lodepng.cpp"
     $128 = HEAP16[$127>>1]|0; //@line 1551 "libraries/lodepng/lodepng.cpp"
     $129 = $128&65535; //@line 1551 "libraries/lodepng/lodepng.cpp"
     $130 = ($hashpos$056$us|0)==($129|0); //@line 1551 "libraries/lodepng/lodepng.cpp"
     if ($130) {
      $length$2 = $length$1$us;$offset$2 = $offset$1$us;
      break L39;
     }
     $131 = ($length$1$us>>>0)>($numzeros$1>>>0); //@line 1553 "libraries/lodepng/lodepng.cpp"
     if ($131) {
      $135 = HEAP32[$15>>2]|0; //@line 1554 "libraries/lodepng/lodepng.cpp"
      $136 = (($135) + ($hashpos$056$us<<1)|0); //@line 1554 "libraries/lodepng/lodepng.cpp"
      $137 = HEAP16[$136>>1]|0; //@line 1554 "libraries/lodepng/lodepng.cpp"
      $138 = $137&65535; //@line 1554 "libraries/lodepng/lodepng.cpp"
      $139 = (($80) + ($138<<1)|0); //@line 1555 "libraries/lodepng/lodepng.cpp"
      $140 = HEAP16[$139>>1]|0; //@line 1555 "libraries/lodepng/lodepng.cpp"
      $141 = $140&65535; //@line 1555 "libraries/lodepng/lodepng.cpp"
      $142 = ($141|0)==($numzeros$1|0); //@line 1555 "libraries/lodepng/lodepng.cpp"
      if ($142) {
       $hashpos$0$be$us = $138;
      } else {
       $length$2 = $length$1$us;$offset$2 = $offset$1$us;
       break L39;
      }
     } else {
      $132 = (($70) + ($129<<2)|0); //@line 1559 "libraries/lodepng/lodepng.cpp"
      $133 = HEAP32[$132>>2]|0; //@line 1559 "libraries/lodepng/lodepng.cpp"
      $134 = ($133|0)==($72|0); //@line 1559 "libraries/lodepng/lodepng.cpp"
      if ($134) {
       $hashpos$0$be$us = $129;
      } else {
       $length$2 = $length$1$us;$offset$2 = $offset$1$us;
       break L39;
      }
     }
     $143 = (($144) + 1)|0; //@line 1514 "libraries/lodepng/lodepng.cpp"
     $145 = ($144>>>0)<($2>>>0); //@line 1514 "libraries/lodepng/lodepng.cpp"
     if ($145) {
      $144 = $143;$hashpos$056$us = $hashpos$0$be$us;$length$054$us = $length$1$us;$offset$053$us = $offset$1$us;$prev_offset$055$us = $105;
     } else {
      $length$2 = $length$1$us;$offset$2 = $offset$1$us;
      break;
     }
    }
   }
  } while(0);
  do {
   if ($17) {
    $lazy$1 = $lazy$082;$length$3 = $length$2;$offset$3 = $offset$2;$pos$1 = $pos$085;
    label = 61;
   } else {
    $174 = ($lazy$082|0)==(0); //@line 1565 "libraries/lodepng/lodepng.cpp"
    $175 = ($length$2>>>0)>(2); //@line 1565 "libraries/lodepng/lodepng.cpp"
    $or$cond5 = $174 & $175; //@line 1565 "libraries/lodepng/lodepng.cpp"
    if ($or$cond5) {
     $176 = ($length$2>>>0)<=($3>>>0); //@line 1565 "libraries/lodepng/lodepng.cpp"
     $177 = ($length$2>>>0)<(258); //@line 1565 "libraries/lodepng/lodepng.cpp"
     $or$cond7 = $176 & $177; //@line 1565 "libraries/lodepng/lodepng.cpp"
     if ($or$cond7) {
      $lazy$2 = 1;$lazylength$1 = $length$2;$lazyoffset$1 = $offset$2;$numzeros$5 = $numzeros$1;$pos$4 = $pos$085;
      break;
     }
    }
    if ($174) {
     $lazy$1 = 0;$length$3 = $length$2;$offset$3 = $offset$2;$pos$1 = $pos$085;
     label = 61;
    } else {
     $178 = ($pos$085|0)==(0); //@line 1575 "libraries/lodepng/lodepng.cpp"
     if ($178) {
      $$012 = 81;
      label = 129;
      break L10;
     }
     $179 = (($lazylength$083) + 1)|0; //@line 1576 "libraries/lodepng/lodepng.cpp"
     $180 = ($length$2>>>0)>($179>>>0); //@line 1576 "libraries/lodepng/lodepng.cpp"
     if (!($180)) {
      HEAP32[$74>>2] = -1; //@line 1585 "libraries/lodepng/lodepng.cpp"
      $200 = (($83) + ($numzeros$1<<2)|0); //@line 1586 "libraries/lodepng/lodepng.cpp"
      HEAP32[$200>>2] = -1; //@line 1586 "libraries/lodepng/lodepng.cpp"
      $201 = (($pos$085) + -1)|0; //@line 1587 "libraries/lodepng/lodepng.cpp"
      $lazy$1 = 0;$length$3 = $lazylength$083;$offset$3 = $lazyoffset$084;$pos$1 = $201;
      label = 61;
      break;
     }
     $181 = (($pos$085) + -1)|0; //@line 1579 "libraries/lodepng/lodepng.cpp"
     $182 = (($in) + ($181)|0); //@line 1579 "libraries/lodepng/lodepng.cpp"
     $183 = HEAP8[$182>>0]|0; //@line 1579 "libraries/lodepng/lodepng.cpp"
     $184 = $183&255; //@line 1579 "libraries/lodepng/lodepng.cpp"
     $185 = HEAP32[$18>>2]|0; //@line 196 "libraries/lodepng/lodepng.cpp"
     $186 = (($185) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
     $187 = $186 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
     $188 = HEAP32[$19>>2]|0; //@line 155 "libraries/lodepng/lodepng.cpp"
     $189 = ($188>>>0)<($187>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
     if ($189) {
      $190 = $188 << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
      $191 = ($190>>>0)<($187>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
      $192 = ($186*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
      $193 = $192 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
      $194 = $191 ? $187 : $193; //@line 157 "libraries/lodepng/lodepng.cpp"
      $195 = HEAP32[$out>>2]|0; //@line 158 "libraries/lodepng/lodepng.cpp"
      $196 = (_realloc($195,$194)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $197 = ($196|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
      if ($197) {
       $$012 = 83;
       label = 129;
       break L10;
      }
      HEAP32[$19>>2] = $194; //@line 161 "libraries/lodepng/lodepng.cpp"
      HEAP32[$out>>2] = $196; //@line 162 "libraries/lodepng/lodepng.cpp"
      $199 = $196;
     } else {
      $$pre$i = HEAP32[$out>>2]|0; //@line 197 "libraries/lodepng/lodepng.cpp"
      $199 = $$pre$i;
     }
     HEAP32[$18>>2] = $186; //@line 173 "libraries/lodepng/lodepng.cpp"
     $198 = (($199) + ($185<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
     HEAP32[$198>>2] = $184; //@line 197 "libraries/lodepng/lodepng.cpp"
     $lazy$1 = 0;$length$3 = $length$2;$offset$3 = $offset$2;$pos$1 = $pos$085;
     label = 61;
    }
   }
  } while(0);
  do {
   if ((label|0) == 61) {
    label = 0;
    $202 = ($length$3>>>0)>(2); //@line 1591 "libraries/lodepng/lodepng.cpp"
    $203 = ($offset$3>>>0)>($windowsize>>>0); //@line 1591 "libraries/lodepng/lodepng.cpp"
    $or$cond14 = $203 & $202; //@line 1591 "libraries/lodepng/lodepng.cpp"
    if ($or$cond14) {
     $$012 = 86;
     label = 129;
     break L10;
    }
    $204 = ($length$3>>>0)<(3); //@line 1594 "libraries/lodepng/lodepng.cpp"
    if ($204) {
     $205 = (($in) + ($pos$1)|0); //@line 1596 "libraries/lodepng/lodepng.cpp"
     $206 = HEAP8[$205>>0]|0; //@line 1596 "libraries/lodepng/lodepng.cpp"
     $207 = $206&255; //@line 1596 "libraries/lodepng/lodepng.cpp"
     $208 = HEAP32[$18>>2]|0; //@line 196 "libraries/lodepng/lodepng.cpp"
     $209 = (($208) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
     $210 = $209 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
     $211 = HEAP32[$19>>2]|0; //@line 155 "libraries/lodepng/lodepng.cpp"
     $212 = ($211>>>0)<($210>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
     if ($212) {
      $213 = $211 << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
      $214 = ($213>>>0)<($210>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
      $215 = ($209*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
      $216 = $215 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
      $217 = $214 ? $210 : $216; //@line 157 "libraries/lodepng/lodepng.cpp"
      $218 = HEAP32[$out>>2]|0; //@line 158 "libraries/lodepng/lodepng.cpp"
      $219 = (_realloc($218,$217)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $220 = ($219|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
      if ($220) {
       $$012 = 83;
       label = 129;
       break L10;
      }
      HEAP32[$19>>2] = $217; //@line 161 "libraries/lodepng/lodepng.cpp"
      HEAP32[$out>>2] = $219; //@line 162 "libraries/lodepng/lodepng.cpp"
      $222 = $219;
     } else {
      $$pre$i30 = HEAP32[$out>>2]|0; //@line 197 "libraries/lodepng/lodepng.cpp"
      $222 = $$pre$i30;
     }
     HEAP32[$18>>2] = $209; //@line 173 "libraries/lodepng/lodepng.cpp"
     $221 = (($222) + ($208<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
     HEAP32[$221>>2] = $207; //@line 197 "libraries/lodepng/lodepng.cpp"
     $lazy$2 = $lazy$1;$lazylength$1 = $lazylength$083;$lazyoffset$1 = $lazyoffset$084;$numzeros$5 = $numzeros$1;$pos$4 = $pos$1;
     break;
    }
    $223 = ($length$3>>>0)<($minmatch>>>0); //@line 1598 "libraries/lodepng/lodepng.cpp"
    if (!($223)) {
     $224 = ($length$3|0)==(3); //@line 1598 "libraries/lodepng/lodepng.cpp"
     $225 = ($offset$3>>>0)>(4096); //@line 1598 "libraries/lodepng/lodepng.cpp"
     $or$cond9 = $225 & $224; //@line 1598 "libraries/lodepng/lodepng.cpp"
     if (!($or$cond9)) {
      $left$0$ph$i$i = 1;$right$0$ph$i$i = 28;
      L105: while(1) {
       $right$0$i$i = $right$0$ph$i$i;
       while(1) {
        $244 = ($left$0$ph$i$i>>>0)>($right$0$i$i>>>0); //@line 1314 "libraries/lodepng/lodepng.cpp"
        if ($244) {
         $$2$i$i = 28;$256 = 258;
         break L105;
        }
        $245 = (($right$0$i$i) + ($left$0$ph$i$i))|0; //@line 1316 "libraries/lodepng/lodepng.cpp"
        $246 = $245 >>> 1; //@line 1316 "libraries/lodepng/lodepng.cpp"
        $247 = (33504 + ($246<<2)|0); //@line 1317 "libraries/lodepng/lodepng.cpp"
        $248 = HEAP32[$247>>2]|0; //@line 1317 "libraries/lodepng/lodepng.cpp"
        $249 = ($248>>>0)>($length$3>>>0); //@line 1317 "libraries/lodepng/lodepng.cpp"
        if (!($249)) {
         $$lcssa134 = $246;$right$0$i$i$lcssa133 = $right$0$i$i;
         break;
        }
        $251 = (($246) + -1)|0; //@line 1318 "libraries/lodepng/lodepng.cpp"
        $252 = (33504 + ($251<<2)|0); //@line 1318 "libraries/lodepng/lodepng.cpp"
        $253 = HEAP32[$252>>2]|0; //@line 1318 "libraries/lodepng/lodepng.cpp"
        $254 = ($253>>>0)>($length$3>>>0); //@line 1318 "libraries/lodepng/lodepng.cpp"
        if ($254) {
         $right$0$i$i = $251;
        } else {
         $$2$i$i = $251;$256 = $253;
         break L105;
        }
       }
       $250 = (($$lcssa134) + 1)|0; //@line 1317 "libraries/lodepng/lodepng.cpp"
       $left$0$ph$i$i = $250;$right$0$ph$i$i = $right$0$i$i$lcssa133;
      }
      $255 = (($length$3) - ($256))|0; //@line 1333 "libraries/lodepng/lodepng.cpp"
      $left$0$ph$i1$i = 1;$right$0$ph$i2$i = 29;
      L113: while(1) {
       $right$0$i4$i = $right$0$ph$i2$i;
       while(1) {
        $257 = ($left$0$ph$i1$i>>>0)>($right$0$i4$i>>>0); //@line 1314 "libraries/lodepng/lodepng.cpp"
        if ($257) {
         $$2$i6$i = 29;$269 = 24577;
         break L113;
        }
        $258 = (($right$0$i4$i) + ($left$0$ph$i1$i))|0; //@line 1316 "libraries/lodepng/lodepng.cpp"
        $259 = $258 >>> 1; //@line 1316 "libraries/lodepng/lodepng.cpp"
        $260 = (33740 + ($259<<2)|0); //@line 1317 "libraries/lodepng/lodepng.cpp"
        $261 = HEAP32[$260>>2]|0; //@line 1317 "libraries/lodepng/lodepng.cpp"
        $262 = ($261>>>0)>($offset$3>>>0); //@line 1317 "libraries/lodepng/lodepng.cpp"
        if (!($262)) {
         $$lcssa136 = $259;$right$0$i4$i$lcssa135 = $right$0$i4$i;
         break;
        }
        $264 = (($259) + -1)|0; //@line 1318 "libraries/lodepng/lodepng.cpp"
        $265 = (33740 + ($264<<2)|0); //@line 1318 "libraries/lodepng/lodepng.cpp"
        $266 = HEAP32[$265>>2]|0; //@line 1318 "libraries/lodepng/lodepng.cpp"
        $267 = ($266>>>0)>($offset$3>>>0); //@line 1318 "libraries/lodepng/lodepng.cpp"
        if ($267) {
         $right$0$i4$i = $264;
        } else {
         $$2$i6$i = $264;$269 = $266;
         break L113;
        }
       }
       $263 = (($$lcssa136) + 1)|0; //@line 1317 "libraries/lodepng/lodepng.cpp"
       $left$0$ph$i1$i = $263;$right$0$ph$i2$i = $right$0$i4$i$lcssa135;
      }
      $268 = (($offset$3) - ($269))|0; //@line 1335 "libraries/lodepng/lodepng.cpp"
      $270 = (($$2$i$i) + 257)|0; //@line 1337 "libraries/lodepng/lodepng.cpp"
      $271 = HEAP32[$18>>2]|0; //@line 196 "libraries/lodepng/lodepng.cpp"
      $272 = (($271) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
      $273 = $272 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
      $274 = HEAP32[$19>>2]|0; //@line 155 "libraries/lodepng/lodepng.cpp"
      $275 = ($274>>>0)<($273>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
      if ($275) {
       $276 = $274 << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
       $277 = ($276>>>0)<($273>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
       $278 = ($272*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
       $279 = $278 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
       $280 = $277 ? $273 : $279; //@line 157 "libraries/lodepng/lodepng.cpp"
       $281 = HEAP32[$out>>2]|0; //@line 158 "libraries/lodepng/lodepng.cpp"
       $282 = (_realloc($281,$280)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       $283 = ($282|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
       if (!($283)) {
        HEAP32[$19>>2] = $280; //@line 161 "libraries/lodepng/lodepng.cpp"
        HEAP32[$out>>2] = $282; //@line 162 "libraries/lodepng/lodepng.cpp"
        $285 = $282;
        label = 90;
       }
      } else {
       $$pre$i15$i = HEAP32[$out>>2]|0; //@line 197 "libraries/lodepng/lodepng.cpp"
       $285 = $$pre$i15$i;
       label = 90;
      }
      if ((label|0) == 90) {
       label = 0;
       HEAP32[$18>>2] = $272; //@line 173 "libraries/lodepng/lodepng.cpp"
       $284 = (($285) + ($271<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
       HEAP32[$284>>2] = $270; //@line 197 "libraries/lodepng/lodepng.cpp"
      }
      $286 = HEAP32[$18>>2]|0; //@line 196 "libraries/lodepng/lodepng.cpp"
      $287 = (($286) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
      $288 = $287 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
      $289 = HEAP32[$19>>2]|0; //@line 155 "libraries/lodepng/lodepng.cpp"
      $290 = ($289>>>0)<($288>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
      if ($290) {
       $291 = $289 << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
       $292 = ($291>>>0)<($288>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
       $293 = ($287*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
       $294 = $293 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
       $295 = $292 ? $288 : $294; //@line 157 "libraries/lodepng/lodepng.cpp"
       $296 = HEAP32[$out>>2]|0; //@line 158 "libraries/lodepng/lodepng.cpp"
       $297 = (_realloc($296,$295)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       $298 = ($297|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
       if (!($298)) {
        HEAP32[$19>>2] = $295; //@line 161 "libraries/lodepng/lodepng.cpp"
        HEAP32[$out>>2] = $297; //@line 162 "libraries/lodepng/lodepng.cpp"
        $300 = $297;
        label = 95;
       }
      } else {
       $$pre$i21$i = HEAP32[$out>>2]|0; //@line 197 "libraries/lodepng/lodepng.cpp"
       $300 = $$pre$i21$i;
       label = 95;
      }
      if ((label|0) == 95) {
       label = 0;
       HEAP32[$18>>2] = $287; //@line 173 "libraries/lodepng/lodepng.cpp"
       $299 = (($300) + ($286<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
       HEAP32[$299>>2] = $255; //@line 197 "libraries/lodepng/lodepng.cpp"
      }
      $301 = HEAP32[$18>>2]|0; //@line 196 "libraries/lodepng/lodepng.cpp"
      $302 = (($301) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
      $303 = $302 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
      $304 = HEAP32[$19>>2]|0; //@line 155 "libraries/lodepng/lodepng.cpp"
      $305 = ($304>>>0)<($303>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
      if ($305) {
       $306 = $304 << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
       $307 = ($306>>>0)<($303>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
       $308 = ($302*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
       $309 = $308 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
       $310 = $307 ? $303 : $309; //@line 157 "libraries/lodepng/lodepng.cpp"
       $311 = HEAP32[$out>>2]|0; //@line 158 "libraries/lodepng/lodepng.cpp"
       $312 = (_realloc($311,$310)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       $313 = ($312|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
       if (!($313)) {
        HEAP32[$19>>2] = $310; //@line 161 "libraries/lodepng/lodepng.cpp"
        HEAP32[$out>>2] = $312; //@line 162 "libraries/lodepng/lodepng.cpp"
        $315 = $312;
        label = 100;
       }
      } else {
       $$pre$i9$i = HEAP32[$out>>2]|0; //@line 197 "libraries/lodepng/lodepng.cpp"
       $315 = $$pre$i9$i;
       label = 100;
      }
      if ((label|0) == 100) {
       label = 0;
       HEAP32[$18>>2] = $302; //@line 173 "libraries/lodepng/lodepng.cpp"
       $314 = (($315) + ($301<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
       HEAP32[$314>>2] = $$2$i6$i; //@line 197 "libraries/lodepng/lodepng.cpp"
      }
      $316 = HEAP32[$18>>2]|0; //@line 196 "libraries/lodepng/lodepng.cpp"
      $317 = (($316) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
      $318 = $317 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
      $319 = HEAP32[$19>>2]|0; //@line 155 "libraries/lodepng/lodepng.cpp"
      $320 = ($319>>>0)<($318>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
      do {
       if ($320) {
        $321 = $319 << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
        $322 = ($321>>>0)<($318>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
        $323 = ($317*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
        $324 = $323 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
        $325 = $322 ? $318 : $324; //@line 157 "libraries/lodepng/lodepng.cpp"
        $326 = HEAP32[$out>>2]|0; //@line 158 "libraries/lodepng/lodepng.cpp"
        $327 = (_realloc($326,$325)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
        $328 = ($327|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
        if ($328) {
         break;
        }
        HEAP32[$19>>2] = $325; //@line 161 "libraries/lodepng/lodepng.cpp"
        HEAP32[$out>>2] = $327; //@line 162 "libraries/lodepng/lodepng.cpp"
        $330 = $327;
        label = 105;
       } else {
        $$pre$i$i = HEAP32[$out>>2]|0; //@line 197 "libraries/lodepng/lodepng.cpp"
        $330 = $$pre$i$i;
        label = 105;
       }
      } while(0);
      if ((label|0) == 105) {
       label = 0;
       HEAP32[$18>>2] = $317; //@line 173 "libraries/lodepng/lodepng.cpp"
       $329 = (($330) + ($316<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
       HEAP32[$329>>2] = $268; //@line 197 "libraries/lodepng/lodepng.cpp"
      }
      $331 = ($length$3>>>0)>(1); //@line 1607 "libraries/lodepng/lodepng.cpp"
      if (!($331)) {
       $lazy$2 = $lazy$1;$lazylength$1 = $lazylength$083;$lazyoffset$1 = $lazyoffset$084;$numzeros$5 = $numzeros$1;$pos$4 = $pos$1;
       break;
      }
      $332 = (($length$3) + -1)|0; //@line 1609 "libraries/lodepng/lodepng.cpp"
      $i$075 = 1;$numzeros$276 = $numzeros$1;$pos$277 = $pos$1;
      while(1) {
       $333 = (($pos$277) + 1)|0; //@line 1609 "libraries/lodepng/lodepng.cpp"
       $334 = $333 & $4; //@line 1610 "libraries/lodepng/lodepng.cpp"
       $335 = (($pos$277) + 3)|0; //@line 1405 "libraries/lodepng/lodepng.cpp"
       $336 = ($335>>>0)<($insize>>>0); //@line 1405 "libraries/lodepng/lodepng.cpp"
       do {
        if ($336) {
         $337 = (($in) + ($333)|0); //@line 1411 "libraries/lodepng/lodepng.cpp"
         $338 = HEAP8[$337>>0]|0; //@line 1411 "libraries/lodepng/lodepng.cpp"
         $339 = $338&255; //@line 1411 "libraries/lodepng/lodepng.cpp"
         $340 = (($pos$277) + 2)|0; //@line 1412 "libraries/lodepng/lodepng.cpp"
         $341 = (($in) + ($340)|0); //@line 1412 "libraries/lodepng/lodepng.cpp"
         $342 = HEAP8[$341>>0]|0; //@line 1412 "libraries/lodepng/lodepng.cpp"
         $343 = $342&255; //@line 1412 "libraries/lodepng/lodepng.cpp"
         $344 = $343 << 4; //@line 1412 "libraries/lodepng/lodepng.cpp"
         $345 = $344 ^ $339; //@line 1412 "libraries/lodepng/lodepng.cpp"
         $346 = (($in) + ($335)|0); //@line 1413 "libraries/lodepng/lodepng.cpp"
         $347 = HEAP8[$346>>0]|0; //@line 1413 "libraries/lodepng/lodepng.cpp"
         $348 = $347&255; //@line 1413 "libraries/lodepng/lodepng.cpp"
         $349 = $348 << 8; //@line 1413 "libraries/lodepng/lodepng.cpp"
         $350 = $345 ^ $349; //@line 1413 "libraries/lodepng/lodepng.cpp"
         $result$2$i19 = $350;
         label = 113;
        } else {
         $351 = ($333>>>0)<($insize>>>0); //@line 1416 "libraries/lodepng/lodepng.cpp"
         if (!($351)) {
          label = 114;
          break;
         }
         $352 = (($insize) - ($333))|0; //@line 1417 "libraries/lodepng/lodepng.cpp"
         $353 = ($333|0)==($insize|0); //@line 1418 "libraries/lodepng/lodepng.cpp"
         if ($353) {
          label = 114;
          break;
         } else {
          $i$03$i16 = 0;$result$02$i17 = 0;
         }
         while(1) {
          $354 = (($i$03$i16) + ($333))|0; //@line 1418 "libraries/lodepng/lodepng.cpp"
          $355 = (($in) + ($354)|0); //@line 1418 "libraries/lodepng/lodepng.cpp"
          $356 = HEAP8[$355>>0]|0; //@line 1418 "libraries/lodepng/lodepng.cpp"
          $357 = $356&255; //@line 1418 "libraries/lodepng/lodepng.cpp"
          $358 = $i$03$i16 << 3; //@line 1418 "libraries/lodepng/lodepng.cpp"
          $359 = $357 << $358; //@line 1418 "libraries/lodepng/lodepng.cpp"
          $360 = $359 ^ $result$02$i17; //@line 1418 "libraries/lodepng/lodepng.cpp"
          $361 = (($i$03$i16) + 1)|0; //@line 1418 "libraries/lodepng/lodepng.cpp"
          $362 = ($361|0)==($352|0); //@line 1418 "libraries/lodepng/lodepng.cpp"
          if ($362) {
           $result$2$i19 = $360;
           label = 113;
           break;
          } else {
           $i$03$i16 = $361;$result$02$i17 = $360;
          }
         }
        }
       } while(0);
       if ((label|0) == 113) {
        label = 0;
        $363 = $result$2$i19 & 65535; //@line 1420 "libraries/lodepng/lodepng.cpp"
        $364 = ($363|0)==(0); //@line 1612 "libraries/lodepng/lodepng.cpp"
        if ($364) {
         label = 114;
        } else {
         $$1$i2141 = $363;$numzeros$3 = 0;
        }
       }
       do {
        if ((label|0) == 114) {
         label = 0;
         $365 = ($numzeros$276|0)==(0); //@line 1614 "libraries/lodepng/lodepng.cpp"
         if (!($365)) {
          $377 = (($333) + ($numzeros$276))|0; //@line 1615 "libraries/lodepng/lodepng.cpp"
          $378 = ($377>>>0)>($insize>>>0); //@line 1615 "libraries/lodepng/lodepng.cpp"
          if (!($378)) {
           $379 = (($pos$277) + ($numzeros$276))|0; //@line 1615 "libraries/lodepng/lodepng.cpp"
           $380 = (($in) + ($379)|0); //@line 1615 "libraries/lodepng/lodepng.cpp"
           $381 = HEAP8[$380>>0]|0; //@line 1615 "libraries/lodepng/lodepng.cpp"
           $382 = ($381<<24>>24)==(0); //@line 1615 "libraries/lodepng/lodepng.cpp"
           if ($382) {
            $$1$i2141 = 0;$numzeros$3 = $numzeros$276;
            break;
           }
          }
          $383 = (($numzeros$276) + -1)|0; //@line 1615 "libraries/lodepng/lodepng.cpp"
          $$1$i2141 = 0;$numzeros$3 = $383;
          break;
         }
         $366 = (($in) + ($333)|0); //@line 1425 "libraries/lodepng/lodepng.cpp"
         $367 = ((($366)) + 258|0); //@line 1426 "libraries/lodepng/lodepng.cpp"
         $368 = ($367>>>0)>($10>>>0); //@line 1427 "libraries/lodepng/lodepng.cpp"
         $$$i = $368 ? $10 : $367; //@line 1427 "libraries/lodepng/lodepng.cpp"
         $369 = ($366|0)==($$$i|0); //@line 1429 "libraries/lodepng/lodepng.cpp"
         L168: do {
          if ($369) {
           $$0$lcssa$i = $366;
          } else {
           $$01$i = $366;
           while(1) {
            $370 = HEAP8[$$01$i>>0]|0; //@line 1429 "libraries/lodepng/lodepng.cpp"
            $371 = ($370<<24>>24)==(0); //@line 1429 "libraries/lodepng/lodepng.cpp"
            if (!($371)) {
             $$0$lcssa$i = $$01$i;
             break L168;
            }
            $372 = ((($$01$i)) + 1|0); //@line 1429 "libraries/lodepng/lodepng.cpp"
            $373 = ($372|0)==($$$i|0); //@line 1429 "libraries/lodepng/lodepng.cpp"
            if ($373) {
             $$0$lcssa$i = $$$i;
             break;
            } else {
             $$01$i = $372;
            }
           }
          }
         } while(0);
         $374 = $$0$lcssa$i; //@line 1431 "libraries/lodepng/lodepng.cpp"
         $375 = $366; //@line 1431 "libraries/lodepng/lodepng.cpp"
         $376 = (($374) - ($375))|0; //@line 1431 "libraries/lodepng/lodepng.cpp"
         $$1$i2141 = 0;$numzeros$3 = $376;
        }
       } while(0);
       $384 = $numzeros$3&65535; //@line 1621 "libraries/lodepng/lodepng.cpp"
       $385 = HEAP32[$11>>2]|0; //@line 1437 "libraries/lodepng/lodepng.cpp"
       $386 = (($385) + ($334<<2)|0); //@line 1437 "libraries/lodepng/lodepng.cpp"
       HEAP32[$386>>2] = $$1$i2141; //@line 1437 "libraries/lodepng/lodepng.cpp"
       $387 = HEAP32[$hash>>2]|0; //@line 1439 "libraries/lodepng/lodepng.cpp"
       $388 = (($387) + ($$1$i2141<<2)|0); //@line 1438 "libraries/lodepng/lodepng.cpp"
       $389 = HEAP32[$388>>2]|0; //@line 1438 "libraries/lodepng/lodepng.cpp"
       $390 = ($389|0)==(-1); //@line 1438 "libraries/lodepng/lodepng.cpp"
       if (!($390)) {
        $391 = $389&65535; //@line 1438 "libraries/lodepng/lodepng.cpp"
        $392 = HEAP32[$14>>2]|0; //@line 1438 "libraries/lodepng/lodepng.cpp"
        $393 = (($392) + ($334<<1)|0); //@line 1438 "libraries/lodepng/lodepng.cpp"
        HEAP16[$393>>1] = $391; //@line 1438 "libraries/lodepng/lodepng.cpp"
       }
       HEAP32[$388>>2] = $334; //@line 1439 "libraries/lodepng/lodepng.cpp"
       $394 = HEAP32[$12>>2]|0; //@line 1441 "libraries/lodepng/lodepng.cpp"
       $395 = (($394) + ($334<<1)|0); //@line 1441 "libraries/lodepng/lodepng.cpp"
       HEAP16[$395>>1] = $384; //@line 1441 "libraries/lodepng/lodepng.cpp"
       $396 = $numzeros$3 & 65535; //@line 1442 "libraries/lodepng/lodepng.cpp"
       $397 = HEAP32[$13>>2]|0; //@line 1443 "libraries/lodepng/lodepng.cpp"
       $398 = (($397) + ($396<<2)|0); //@line 1442 "libraries/lodepng/lodepng.cpp"
       $399 = HEAP32[$398>>2]|0; //@line 1442 "libraries/lodepng/lodepng.cpp"
       $400 = ($399|0)==(-1); //@line 1442 "libraries/lodepng/lodepng.cpp"
       if (!($400)) {
        $401 = $399&65535; //@line 1442 "libraries/lodepng/lodepng.cpp"
        $402 = HEAP32[$15>>2]|0; //@line 1442 "libraries/lodepng/lodepng.cpp"
        $403 = (($402) + ($334<<1)|0); //@line 1442 "libraries/lodepng/lodepng.cpp"
        HEAP16[$403>>1] = $401; //@line 1442 "libraries/lodepng/lodepng.cpp"
       }
       HEAP32[$398>>2] = $334; //@line 1443 "libraries/lodepng/lodepng.cpp"
       $404 = (($i$075) + 1)|0; //@line 1607 "libraries/lodepng/lodepng.cpp"
       $exitcond = ($404|0)==($length$3|0); //@line 1607 "libraries/lodepng/lodepng.cpp"
       if ($exitcond) {
        $numzeros$3$lcssa = $numzeros$3;
        break;
       } else {
        $i$075 = $404;$numzeros$276 = $numzeros$3;$pos$277 = $333;
       }
      }
      $405 = (($332) + ($pos$1))|0; //@line 1609 "libraries/lodepng/lodepng.cpp"
      $lazy$2 = $lazy$1;$lazylength$1 = $lazylength$083;$lazyoffset$1 = $lazyoffset$084;$numzeros$5 = $numzeros$3$lcssa;$pos$4 = $405;
      break;
     }
    }
    $226 = (($in) + ($pos$1)|0); //@line 1602 "libraries/lodepng/lodepng.cpp"
    $227 = HEAP8[$226>>0]|0; //@line 1602 "libraries/lodepng/lodepng.cpp"
    $228 = $227&255; //@line 1602 "libraries/lodepng/lodepng.cpp"
    $229 = HEAP32[$18>>2]|0; //@line 196 "libraries/lodepng/lodepng.cpp"
    $230 = (($229) + 1)|0; //@line 196 "libraries/lodepng/lodepng.cpp"
    $231 = $230 << 2; //@line 172 "libraries/lodepng/lodepng.cpp"
    $232 = HEAP32[$19>>2]|0; //@line 155 "libraries/lodepng/lodepng.cpp"
    $233 = ($232>>>0)<($231>>>0); //@line 155 "libraries/lodepng/lodepng.cpp"
    if ($233) {
     $234 = $232 << 1; //@line 157 "libraries/lodepng/lodepng.cpp"
     $235 = ($234>>>0)<($231>>>0); //@line 157 "libraries/lodepng/lodepng.cpp"
     $236 = ($230*12)|0; //@line 157 "libraries/lodepng/lodepng.cpp"
     $237 = $236 >>> 1; //@line 157 "libraries/lodepng/lodepng.cpp"
     $238 = $235 ? $231 : $237; //@line 157 "libraries/lodepng/lodepng.cpp"
     $239 = HEAP32[$out>>2]|0; //@line 158 "libraries/lodepng/lodepng.cpp"
     $240 = (_realloc($239,$238)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
     $241 = ($240|0)==(0|0); //@line 159 "libraries/lodepng/lodepng.cpp"
     if ($241) {
      $$012 = 83;
      label = 129;
      break L10;
     }
     HEAP32[$19>>2] = $238; //@line 161 "libraries/lodepng/lodepng.cpp"
     HEAP32[$out>>2] = $240; //@line 162 "libraries/lodepng/lodepng.cpp"
     $243 = $240;
    } else {
     $$pre$i36 = HEAP32[$out>>2]|0; //@line 197 "libraries/lodepng/lodepng.cpp"
     $243 = $$pre$i36;
    }
    HEAP32[$18>>2] = $230; //@line 173 "libraries/lodepng/lodepng.cpp"
    $242 = (($243) + ($229<<2)|0); //@line 197 "libraries/lodepng/lodepng.cpp"
    HEAP32[$242>>2] = $228; //@line 197 "libraries/lodepng/lodepng.cpp"
    $lazy$2 = $lazy$1;$lazylength$1 = $lazylength$083;$lazyoffset$1 = $lazyoffset$084;$numzeros$5 = $numzeros$1;$pos$4 = $pos$1;
   }
  } while(0);
  $406 = (($pos$4) + 1)|0; //@line 1483 "libraries/lodepng/lodepng.cpp"
  $407 = ($406>>>0)<($insize>>>0); //@line 1483 "libraries/lodepng/lodepng.cpp"
  if ($407) {
   $lazy$082 = $lazy$2;$lazylength$083 = $lazylength$1;$lazyoffset$084 = $lazyoffset$1;$numzeros$081 = $numzeros$5;$pos$085 = $406;
  } else {
   $$012 = 0;
   label = 129;
   break;
  }
 }
 if ((label|0) == 129) {
  return ($$012|0); //@line 1627 "libraries/lodepng/lodepng.cpp"
 }
 return (0)|0;
}
function __ZL13writeLZ77dataPjP8ucvectorPK8uivectorPK11HuffmanTreeS7_($bp,$out,$lz77_encoded,$tree_ll,$tree_d) {
 $bp = $bp|0;
 $out = $out|0;
 $lz77_encoded = $lz77_encoded|0;
 $tree_ll = $tree_ll|0;
 $tree_d = $tree_d|0;
 var $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i7 = 0, $$pre$i$i15 = 0, $$pre$i$i3 = 0, $$pre$i$i4 = 0, $$pre$i12 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$01$i = 0, $i$01$i$i = 0, $i$01$i$i6 = 0, $i$01$i14 = 0, $i$020 = 0, $i$1 = 0, $tree_d$idx = 0, $tree_d$idx$val = 0, $tree_d$idx1 = 0, $tree_d$idx1$val = 0, $tree_ll$idx = 0, $tree_ll$idx$val = 0;
 var $tree_ll$idx2 = 0, $tree_ll$idx2$val = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($lz77_encoded)) + 4|0); //@line 1677 "libraries/lodepng/lodepng.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1677 "libraries/lodepng/lodepng.cpp"
 $2 = ($1|0)==(0); //@line 1677 "libraries/lodepng/lodepng.cpp"
 if ($2) {
  return; //@line 1699 "libraries/lodepng/lodepng.cpp"
 }
 $tree_ll$idx = ((($tree_ll)) + 4|0);
 $tree_ll$idx2 = ((($tree_ll)) + 8|0);
 $tree_d$idx = ((($tree_d)) + 4|0);
 $tree_d$idx1 = ((($tree_d)) + 8|0);
 $3 = ((($out)) + 4|0); //@line 288 "libraries/lodepng/lodepng.cpp"
 $4 = ((($out)) + 8|0); //@line 225 "libraries/lodepng/lodepng.cpp"
 $i$020 = 0;
 while(1) {
  $5 = HEAP32[$lz77_encoded>>2]|0; //@line 1679 "libraries/lodepng/lodepng.cpp"
  $6 = (($5) + ($i$020<<2)|0); //@line 1679 "libraries/lodepng/lodepng.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 1679 "libraries/lodepng/lodepng.cpp"
  $tree_ll$idx$val = HEAP32[$tree_ll$idx>>2]|0;
  $8 = (($tree_ll$idx$val) + ($7<<2)|0); //@line 877 "libraries/lodepng/lodepng.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 877 "libraries/lodepng/lodepng.cpp"
  $tree_ll$idx2$val = HEAP32[$tree_ll$idx2>>2]|0;
  $10 = (($tree_ll$idx2$val) + ($7<<2)|0); //@line 882 "libraries/lodepng/lodepng.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 882 "libraries/lodepng/lodepng.cpp"
  $12 = ($11|0)==(0); //@line 438 "libraries/lodepng/lodepng.cpp"
  if (!($12)) {
   $13 = (($11) + -1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
   $$pre$i$i4 = HEAP32[$bp>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
   $15 = $$pre$i$i4;$i$01$i$i6 = 0;
   while(1) {
    $14 = $15 & 7; //@line 438 "libraries/lodepng/lodepng.cpp"
    $16 = ($14|0)==(0); //@line 438 "libraries/lodepng/lodepng.cpp"
    do {
     if ($16) {
      $17 = HEAP32[$3>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
      $18 = (($17) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
      $19 = HEAP32[$4>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
      $20 = ($19>>>0)<($18>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
      if ($20) {
       $21 = $19 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
       $22 = ($21>>>0)<($18>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
       $23 = ($18*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
       $24 = $23 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
       $25 = $22 ? $18 : $24; //@line 227 "libraries/lodepng/lodepng.cpp"
       $26 = HEAP32[$out>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
       $27 = (_realloc($26,$25)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       $28 = ($27|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
       if ($28) {
        break;
       }
       HEAP32[$4>>2] = $25; //@line 231 "libraries/lodepng/lodepng.cpp"
       HEAP32[$out>>2] = $27; //@line 232 "libraries/lodepng/lodepng.cpp"
       $30 = $27;
      } else {
       $$pre$i$i$i7 = HEAP32[$out>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
       $30 = $$pre$i$i$i7;
      }
      HEAP32[$3>>2] = $18; //@line 243 "libraries/lodepng/lodepng.cpp"
      $29 = (($30) + ($17)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
      HEAP8[$29>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
     }
    } while(0);
    $31 = (($13) - ($i$01$i$i6))|0; //@line 438 "libraries/lodepng/lodepng.cpp"
    $32 = $9 >>> $31; //@line 438 "libraries/lodepng/lodepng.cpp"
    $33 = $32 & 1; //@line 438 "libraries/lodepng/lodepng.cpp"
    $34 = HEAP32[$bp>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
    $35 = $34 & 7; //@line 438 "libraries/lodepng/lodepng.cpp"
    $36 = $33 << $35; //@line 438 "libraries/lodepng/lodepng.cpp"
    $37 = HEAP32[$3>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
    $38 = (($37) + -1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
    $39 = HEAP32[$out>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
    $40 = (($39) + ($38)|0); //@line 438 "libraries/lodepng/lodepng.cpp"
    $41 = HEAP8[$40>>0]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
    $42 = $41&255; //@line 438 "libraries/lodepng/lodepng.cpp"
    $43 = $42 | $36; //@line 438 "libraries/lodepng/lodepng.cpp"
    $44 = $43&255; //@line 438 "libraries/lodepng/lodepng.cpp"
    HEAP8[$40>>0] = $44; //@line 438 "libraries/lodepng/lodepng.cpp"
    $45 = HEAP32[$bp>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
    $46 = (($45) + 1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
    HEAP32[$bp>>2] = $46; //@line 438 "libraries/lodepng/lodepng.cpp"
    $47 = (($i$01$i$i6) + 1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
    $48 = ($47|0)==($11|0); //@line 438 "libraries/lodepng/lodepng.cpp"
    if ($48) {
     break;
    } else {
     $15 = $46;$i$01$i$i6 = $47;
    }
   }
  }
  $49 = ($7>>>0)>(256); //@line 1681 "libraries/lodepng/lodepng.cpp"
  if ($49) {
   $50 = (($7) + -257)|0; //@line 1683 "libraries/lodepng/lodepng.cpp"
   $51 = (33388 + ($50<<2)|0); //@line 1684 "libraries/lodepng/lodepng.cpp"
   $52 = HEAP32[$51>>2]|0; //@line 1684 "libraries/lodepng/lodepng.cpp"
   $53 = (($i$020) + 1)|0; //@line 1685 "libraries/lodepng/lodepng.cpp"
   $54 = HEAP32[$lz77_encoded>>2]|0; //@line 1685 "libraries/lodepng/lodepng.cpp"
   $55 = (($54) + ($53<<2)|0); //@line 1685 "libraries/lodepng/lodepng.cpp"
   $56 = HEAP32[$55>>2]|0; //@line 1685 "libraries/lodepng/lodepng.cpp"
   $57 = (($i$020) + 2)|0; //@line 1687 "libraries/lodepng/lodepng.cpp"
   $58 = (($54) + ($57<<2)|0); //@line 1687 "libraries/lodepng/lodepng.cpp"
   $59 = HEAP32[$58>>2]|0; //@line 1687 "libraries/lodepng/lodepng.cpp"
   $60 = (33620 + ($59<<2)|0); //@line 1690 "libraries/lodepng/lodepng.cpp"
   $61 = HEAP32[$60>>2]|0; //@line 1690 "libraries/lodepng/lodepng.cpp"
   $62 = (($i$020) + 3)|0; //@line 1691 "libraries/lodepng/lodepng.cpp"
   $63 = (($54) + ($62<<2)|0); //@line 1691 "libraries/lodepng/lodepng.cpp"
   $64 = HEAP32[$63>>2]|0; //@line 1691 "libraries/lodepng/lodepng.cpp"
   $65 = (($7) + -265)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
   $66 = ($65>>>0)>(19); //@line 432 "libraries/lodepng/lodepng.cpp"
   if (!($66)) {
    $$pre$i12 = HEAP32[$bp>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
    $68 = $$pre$i12;$i$01$i14 = 0;
    while(1) {
     $67 = $68 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
     $69 = ($67|0)==(0); //@line 432 "libraries/lodepng/lodepng.cpp"
     do {
      if ($69) {
       $70 = HEAP32[$3>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
       $71 = (($70) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
       $72 = HEAP32[$4>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
       $73 = ($72>>>0)<($71>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
       if ($73) {
        $74 = $72 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $75 = ($74>>>0)<($71>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
        $76 = ($71*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
        $77 = $76 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $78 = $75 ? $71 : $77; //@line 227 "libraries/lodepng/lodepng.cpp"
        $79 = HEAP32[$out>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
        $80 = (_realloc($79,$78)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
        $81 = ($80|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
        if ($81) {
         break;
        }
        HEAP32[$4>>2] = $78; //@line 231 "libraries/lodepng/lodepng.cpp"
        HEAP32[$out>>2] = $80; //@line 232 "libraries/lodepng/lodepng.cpp"
        $83 = $80;
       } else {
        $$pre$i$i15 = HEAP32[$out>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
        $83 = $$pre$i$i15;
       }
       HEAP32[$3>>2] = $71; //@line 243 "libraries/lodepng/lodepng.cpp"
       $82 = (($83) + ($70)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
       HEAP8[$82>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
      }
     } while(0);
     $84 = $56 >>> $i$01$i14; //@line 432 "libraries/lodepng/lodepng.cpp"
     $85 = $84 & 1; //@line 432 "libraries/lodepng/lodepng.cpp"
     $86 = HEAP32[$bp>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
     $87 = $86 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
     $88 = $85 << $87; //@line 432 "libraries/lodepng/lodepng.cpp"
     $89 = HEAP32[$3>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
     $90 = (($89) + -1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
     $91 = HEAP32[$out>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
     $92 = (($91) + ($90)|0); //@line 432 "libraries/lodepng/lodepng.cpp"
     $93 = HEAP8[$92>>0]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
     $94 = $93&255; //@line 432 "libraries/lodepng/lodepng.cpp"
     $95 = $94 | $88; //@line 432 "libraries/lodepng/lodepng.cpp"
     $96 = $95&255; //@line 432 "libraries/lodepng/lodepng.cpp"
     HEAP8[$92>>0] = $96; //@line 432 "libraries/lodepng/lodepng.cpp"
     $97 = HEAP32[$bp>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
     $98 = (($97) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
     HEAP32[$bp>>2] = $98; //@line 432 "libraries/lodepng/lodepng.cpp"
     $99 = (($i$01$i14) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
     $100 = ($99|0)==($52|0); //@line 432 "libraries/lodepng/lodepng.cpp"
     if ($100) {
      break;
     } else {
      $68 = $98;$i$01$i14 = $99;
     }
    }
   }
   $tree_d$idx$val = HEAP32[$tree_d$idx>>2]|0;
   $101 = (($tree_d$idx$val) + ($59<<2)|0); //@line 877 "libraries/lodepng/lodepng.cpp"
   $102 = HEAP32[$101>>2]|0; //@line 877 "libraries/lodepng/lodepng.cpp"
   $tree_d$idx1$val = HEAP32[$tree_d$idx1>>2]|0;
   $103 = (($tree_d$idx1$val) + ($59<<2)|0); //@line 882 "libraries/lodepng/lodepng.cpp"
   $104 = HEAP32[$103>>2]|0; //@line 882 "libraries/lodepng/lodepng.cpp"
   $105 = ($104|0)==(0); //@line 438 "libraries/lodepng/lodepng.cpp"
   if (!($105)) {
    $106 = (($104) + -1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
    $$pre$i$i3 = HEAP32[$bp>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
    $108 = $$pre$i$i3;$i$01$i$i = 0;
    while(1) {
     $107 = $108 & 7; //@line 438 "libraries/lodepng/lodepng.cpp"
     $109 = ($107|0)==(0); //@line 438 "libraries/lodepng/lodepng.cpp"
     do {
      if ($109) {
       $110 = HEAP32[$3>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
       $111 = (($110) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
       $112 = HEAP32[$4>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
       $113 = ($112>>>0)<($111>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
       if ($113) {
        $114 = $112 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $115 = ($114>>>0)<($111>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
        $116 = ($111*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
        $117 = $116 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $118 = $115 ? $111 : $117; //@line 227 "libraries/lodepng/lodepng.cpp"
        $119 = HEAP32[$out>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
        $120 = (_realloc($119,$118)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
        $121 = ($120|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
        if ($121) {
         break;
        }
        HEAP32[$4>>2] = $118; //@line 231 "libraries/lodepng/lodepng.cpp"
        HEAP32[$out>>2] = $120; //@line 232 "libraries/lodepng/lodepng.cpp"
        $123 = $120;
       } else {
        $$pre$i$i$i = HEAP32[$out>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
        $123 = $$pre$i$i$i;
       }
       HEAP32[$3>>2] = $111; //@line 243 "libraries/lodepng/lodepng.cpp"
       $122 = (($123) + ($110)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
       HEAP8[$122>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
      }
     } while(0);
     $124 = (($106) - ($i$01$i$i))|0; //@line 438 "libraries/lodepng/lodepng.cpp"
     $125 = $102 >>> $124; //@line 438 "libraries/lodepng/lodepng.cpp"
     $126 = $125 & 1; //@line 438 "libraries/lodepng/lodepng.cpp"
     $127 = HEAP32[$bp>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
     $128 = $127 & 7; //@line 438 "libraries/lodepng/lodepng.cpp"
     $129 = $126 << $128; //@line 438 "libraries/lodepng/lodepng.cpp"
     $130 = HEAP32[$3>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
     $131 = (($130) + -1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
     $132 = HEAP32[$out>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
     $133 = (($132) + ($131)|0); //@line 438 "libraries/lodepng/lodepng.cpp"
     $134 = HEAP8[$133>>0]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
     $135 = $134&255; //@line 438 "libraries/lodepng/lodepng.cpp"
     $136 = $135 | $129; //@line 438 "libraries/lodepng/lodepng.cpp"
     $137 = $136&255; //@line 438 "libraries/lodepng/lodepng.cpp"
     HEAP8[$133>>0] = $137; //@line 438 "libraries/lodepng/lodepng.cpp"
     $138 = HEAP32[$bp>>2]|0; //@line 438 "libraries/lodepng/lodepng.cpp"
     $139 = (($138) + 1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
     HEAP32[$bp>>2] = $139; //@line 438 "libraries/lodepng/lodepng.cpp"
     $140 = (($i$01$i$i) + 1)|0; //@line 438 "libraries/lodepng/lodepng.cpp"
     $141 = ($140|0)==($104|0); //@line 438 "libraries/lodepng/lodepng.cpp"
     if ($141) {
      break;
     } else {
      $108 = $139;$i$01$i$i = $140;
     }
    }
   }
   $142 = ($59>>>0)<(4); //@line 432 "libraries/lodepng/lodepng.cpp"
   if ($142) {
    $i$1 = $62;
   } else {
    $$pre$i = HEAP32[$bp>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
    $144 = $$pre$i;$i$01$i = 0;
    while(1) {
     $143 = $144 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
     $145 = ($143|0)==(0); //@line 432 "libraries/lodepng/lodepng.cpp"
     do {
      if ($145) {
       $146 = HEAP32[$3>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
       $147 = (($146) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
       $148 = HEAP32[$4>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
       $149 = ($148>>>0)<($147>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
       if ($149) {
        $150 = $148 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $151 = ($150>>>0)<($147>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
        $152 = ($147*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
        $153 = $152 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $154 = $151 ? $147 : $153; //@line 227 "libraries/lodepng/lodepng.cpp"
        $155 = HEAP32[$out>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
        $156 = (_realloc($155,$154)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
        $157 = ($156|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
        if ($157) {
         break;
        }
        HEAP32[$4>>2] = $154; //@line 231 "libraries/lodepng/lodepng.cpp"
        HEAP32[$out>>2] = $156; //@line 232 "libraries/lodepng/lodepng.cpp"
        $159 = $156;
       } else {
        $$pre$i$i = HEAP32[$out>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
        $159 = $$pre$i$i;
       }
       HEAP32[$3>>2] = $147; //@line 243 "libraries/lodepng/lodepng.cpp"
       $158 = (($159) + ($146)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
       HEAP8[$158>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
      }
     } while(0);
     $160 = $64 >>> $i$01$i; //@line 432 "libraries/lodepng/lodepng.cpp"
     $161 = $160 & 1; //@line 432 "libraries/lodepng/lodepng.cpp"
     $162 = HEAP32[$bp>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
     $163 = $162 & 7; //@line 432 "libraries/lodepng/lodepng.cpp"
     $164 = $161 << $163; //@line 432 "libraries/lodepng/lodepng.cpp"
     $165 = HEAP32[$3>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
     $166 = (($165) + -1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
     $167 = HEAP32[$out>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
     $168 = (($167) + ($166)|0); //@line 432 "libraries/lodepng/lodepng.cpp"
     $169 = HEAP8[$168>>0]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
     $170 = $169&255; //@line 432 "libraries/lodepng/lodepng.cpp"
     $171 = $170 | $164; //@line 432 "libraries/lodepng/lodepng.cpp"
     $172 = $171&255; //@line 432 "libraries/lodepng/lodepng.cpp"
     HEAP8[$168>>0] = $172; //@line 432 "libraries/lodepng/lodepng.cpp"
     $173 = HEAP32[$bp>>2]|0; //@line 432 "libraries/lodepng/lodepng.cpp"
     $174 = (($173) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
     HEAP32[$bp>>2] = $174; //@line 432 "libraries/lodepng/lodepng.cpp"
     $175 = (($i$01$i) + 1)|0; //@line 432 "libraries/lodepng/lodepng.cpp"
     $176 = ($175|0)==($61|0); //@line 432 "libraries/lodepng/lodepng.cpp"
     if ($176) {
      $i$1 = $62;
      break;
     } else {
      $144 = $174;$i$01$i = $175;
     }
    }
   }
  } else {
   $i$1 = $i$020;
  }
  $177 = (($i$1) + 1)|0; //@line 1677 "libraries/lodepng/lodepng.cpp"
  $178 = HEAP32[$0>>2]|0; //@line 1677 "libraries/lodepng/lodepng.cpp"
  $179 = ($177|0)==($178|0); //@line 1677 "libraries/lodepng/lodepng.cpp"
  if ($179) {
   break;
  } else {
   $i$020 = $177;
  }
 }
 return; //@line 1699 "libraries/lodepng/lodepng.cpp"
}
function __Z23lodepng_zlib_decompressPPhPjPKhjPK25LodePNGDecompressSettings($out,$outsize,$in,$insize,$settings) {
 $out = $out|0;
 $outsize = $outsize|0;
 $in = $in|0;
 $insize = $insize|0;
 $settings = $settings|0;
 var $$0$i = 0, $$017$i$i = 0, $$08$i$i = 0, $$1 = 0, $$1$lcssa$i$i = 0, $$11$i$i = 0, $$lcssa26 = 0, $$lcssa27 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $7 = 0, $8 = 0, $9 = 0, $amount$04$i$i = 0, $or$cond = 0, $phitmp$i$i = 0, $s1$0$lcssa$i$i = 0, $s1$09$i$i = 0, $s1$1$lcssa$i$i = 0, $s1$12$i$i = 0, $s2$0$lcssa$i$i = 0, $s2$010$i$i = 0;
 var $s2$1$lcssa$i$i = 0, $s2$13$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($insize>>>0)<(2); //@line 2108 "libraries/lodepng/lodepng.cpp"
 if ($0) {
  $$1 = 53;
  return ($$1|0); //@line 2145 "libraries/lodepng/lodepng.cpp"
 }
 $1 = HEAP8[$in>>0]|0; //@line 2110 "libraries/lodepng/lodepng.cpp"
 $2 = $1&255; //@line 2110 "libraries/lodepng/lodepng.cpp"
 $3 = $2 << 8; //@line 2110 "libraries/lodepng/lodepng.cpp"
 $4 = ((($in)) + 1|0); //@line 2110 "libraries/lodepng/lodepng.cpp"
 $5 = HEAP8[$4>>0]|0; //@line 2110 "libraries/lodepng/lodepng.cpp"
 $6 = $5&255; //@line 2110 "libraries/lodepng/lodepng.cpp"
 $7 = $3 | $6; //@line 2110 "libraries/lodepng/lodepng.cpp"
 $8 = (($7>>>0) % 31)&-1; //@line 2110 "libraries/lodepng/lodepng.cpp"
 $9 = ($8|0)==(0); //@line 2110 "libraries/lodepng/lodepng.cpp"
 if (!($9)) {
  $$1 = 24;
  return ($$1|0); //@line 2145 "libraries/lodepng/lodepng.cpp"
 }
 $10 = $2 & 15; //@line 2116 "libraries/lodepng/lodepng.cpp"
 $11 = ($10|0)!=(8); //@line 2122 "libraries/lodepng/lodepng.cpp"
 $12 = ($1<<24>>24)<(0); //@line 2122 "libraries/lodepng/lodepng.cpp"
 $or$cond = $12 | $11; //@line 2122 "libraries/lodepng/lodepng.cpp"
 if ($or$cond) {
  $$1 = 25;
  return ($$1|0); //@line 2145 "libraries/lodepng/lodepng.cpp"
 }
 $13 = $6 & 32; //@line 2119 "libraries/lodepng/lodepng.cpp"
 $14 = ($13|0)==(0); //@line 2127 "libraries/lodepng/lodepng.cpp"
 if (!($14)) {
  $$1 = 26;
  return ($$1|0); //@line 2145 "libraries/lodepng/lodepng.cpp"
 }
 $15 = ((($in)) + 2|0); //@line 2134 "libraries/lodepng/lodepng.cpp"
 $16 = (($insize) + -2)|0; //@line 2134 "libraries/lodepng/lodepng.cpp"
 $17 = ((($settings)) + 8|0); //@line 1277 "libraries/lodepng/lodepng.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 1277 "libraries/lodepng/lodepng.cpp"
 $19 = ($18|0)==(0|0); //@line 1277 "libraries/lodepng/lodepng.cpp"
 if ($19) {
  $21 = (__Z15lodepng_inflatePPhPjPKhjPK25LodePNGDecompressSettings($out,$outsize,$15,$16,0)|0); //@line 1283 "libraries/lodepng/lodepng.cpp"
  $$0$i = $21;
 } else {
  $20 = (FUNCTION_TABLE_iiiiii[$18 & 8191]($out,$outsize,$15,$16,$settings)|0); //@line 1279 "libraries/lodepng/lodepng.cpp"
  $$0$i = $20;
 }
 $22 = ($$0$i|0)==(0); //@line 2135 "libraries/lodepng/lodepng.cpp"
 if (!($22)) {
  $$1 = $$0$i;
  return ($$1|0); //@line 2145 "libraries/lodepng/lodepng.cpp"
 }
 $23 = HEAP32[$settings>>2]|0; //@line 2137 "libraries/lodepng/lodepng.cpp"
 $24 = ($23|0)==(0); //@line 2137 "libraries/lodepng/lodepng.cpp"
 if ($24) {
  $25 = (($insize) + -4)|0; //@line 2139 "libraries/lodepng/lodepng.cpp"
  $26 = (($in) + ($25)|0); //@line 2139 "libraries/lodepng/lodepng.cpp"
  $27 = HEAP8[$26>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
  $28 = $27&255; //@line 343 "libraries/lodepng/lodepng.cpp"
  $29 = $28 << 24; //@line 343 "libraries/lodepng/lodepng.cpp"
  $30 = ((($26)) + 1|0); //@line 343 "libraries/lodepng/lodepng.cpp"
  $31 = HEAP8[$30>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
  $32 = $31&255; //@line 343 "libraries/lodepng/lodepng.cpp"
  $33 = $32 << 16; //@line 343 "libraries/lodepng/lodepng.cpp"
  $34 = $33 | $29; //@line 343 "libraries/lodepng/lodepng.cpp"
  $35 = ((($26)) + 2|0); //@line 343 "libraries/lodepng/lodepng.cpp"
  $36 = HEAP8[$35>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
  $37 = $36&255; //@line 343 "libraries/lodepng/lodepng.cpp"
  $38 = $37 << 8; //@line 343 "libraries/lodepng/lodepng.cpp"
  $39 = $34 | $38; //@line 343 "libraries/lodepng/lodepng.cpp"
  $40 = ((($26)) + 3|0); //@line 343 "libraries/lodepng/lodepng.cpp"
  $41 = HEAP8[$40>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
  $42 = $41&255; //@line 343 "libraries/lodepng/lodepng.cpp"
  $43 = $39 | $42; //@line 343 "libraries/lodepng/lodepng.cpp"
  $44 = HEAP32[$outsize>>2]|0; //@line 2140 "libraries/lodepng/lodepng.cpp"
  $45 = ($44|0)==(0); //@line 2072 "libraries/lodepng/lodepng.cpp"
  if ($45) {
   $s1$0$lcssa$i$i = 1;$s2$0$lcssa$i$i = 0;
  } else {
   $46 = HEAP32[$out>>2]|0; //@line 2140 "libraries/lodepng/lodepng.cpp"
   $$017$i$i = $46;$$08$i$i = $44;$s1$09$i$i = 1;$s2$010$i$i = 0;
   while(1) {
    $47 = ($$08$i$i>>>0)>(5550); //@line 2075 "libraries/lodepng/lodepng.cpp"
    $48 = $47 ? 5550 : $$08$i$i; //@line 2075 "libraries/lodepng/lodepng.cpp"
    $49 = (($$08$i$i) - ($48))|0; //@line 2076 "libraries/lodepng/lodepng.cpp"
    $50 = ($48|0)==(0); //@line 2077 "libraries/lodepng/lodepng.cpp"
    if ($50) {
     $$1$lcssa$i$i = $$017$i$i;$s1$1$lcssa$i$i = $s1$09$i$i;$s2$1$lcssa$i$i = $s2$010$i$i;
    } else {
     $$11$i$i = $$017$i$i;$amount$04$i$i = $48;$s1$12$i$i = $s1$09$i$i;$s2$13$i$i = $s2$010$i$i;
     while(1) {
      $51 = ((($$11$i$i)) + 1|0); //@line 2079 "libraries/lodepng/lodepng.cpp"
      $52 = HEAP8[$$11$i$i>>0]|0; //@line 2079 "libraries/lodepng/lodepng.cpp"
      $53 = $52&255; //@line 2079 "libraries/lodepng/lodepng.cpp"
      $54 = (($53) + ($s1$12$i$i))|0; //@line 2079 "libraries/lodepng/lodepng.cpp"
      $55 = (($54) + ($s2$13$i$i))|0; //@line 2080 "libraries/lodepng/lodepng.cpp"
      $56 = (($amount$04$i$i) + -1)|0; //@line 2081 "libraries/lodepng/lodepng.cpp"
      $57 = ($56|0)==(0); //@line 2077 "libraries/lodepng/lodepng.cpp"
      if ($57) {
       $$1$lcssa$i$i = $51;$s1$1$lcssa$i$i = $54;$s2$1$lcssa$i$i = $55;
       break;
      } else {
       $$11$i$i = $51;$amount$04$i$i = $56;$s1$12$i$i = $54;$s2$13$i$i = $55;
      }
     }
    }
    $58 = (($s1$1$lcssa$i$i>>>0) % 65521)&-1; //@line 2083 "libraries/lodepng/lodepng.cpp"
    $59 = (($s2$1$lcssa$i$i>>>0) % 65521)&-1; //@line 2084 "libraries/lodepng/lodepng.cpp"
    $60 = ($$08$i$i|0)==($48|0); //@line 2072 "libraries/lodepng/lodepng.cpp"
    if ($60) {
     $$lcssa26 = $58;$$lcssa27 = $59;
     break;
    } else {
     $$017$i$i = $$1$lcssa$i$i;$$08$i$i = $49;$s1$09$i$i = $58;$s2$010$i$i = $59;
    }
   }
   $phitmp$i$i = $$lcssa27 << 16; //@line 2087 "libraries/lodepng/lodepng.cpp"
   $s1$0$lcssa$i$i = $$lcssa26;$s2$0$lcssa$i$i = $phitmp$i$i;
  }
  $61 = $s1$0$lcssa$i$i | $s2$0$lcssa$i$i; //@line 2087 "libraries/lodepng/lodepng.cpp"
  $62 = ($61|0)==($43|0); //@line 2141 "libraries/lodepng/lodepng.cpp"
  if (!($62)) {
   $$1 = 58;
   return ($$1|0); //@line 2145 "libraries/lodepng/lodepng.cpp"
  }
 }
 $$1 = 0;
 return ($$1|0); //@line 2145 "libraries/lodepng/lodepng.cpp"
}
function __Z21lodepng_zlib_compressPPhPjPKhjPK23LodePNGCompressSettings($out,$outsize,$in,$insize,$settings) {
 $out = $out|0;
 $outsize = $outsize|0;
 $in = $in|0;
 $insize = $insize|0;
 $settings = $settings|0;
 var $$0$i = 0, $$017$i$i = 0, $$08$i$i = 0, $$1$lcssa$i$i = 0, $$11$i$i = 0, $$lcssa = 0, $$lcssa54 = 0, $$lcssa55 = 0, $$pre = 0, $$pre31 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $amount$04$i$i = 0, $deflatedata = 0;
 var $deflatesize = 0, $i$021 = 0, $outv$sroa$0 = 0, $outv$sroa$18$0 = 0, $outv$sroa$18$1 = 0, $outv$sroa$18$2$lcssa = 0, $outv$sroa$18$220 = 0, $outv$sroa$18$3 = 0, $outv$sroa$18$4 = 0, $outv$sroa$18$5 = 0, $outv$sroa$29$0 = 0, $outv$sroa$29$1 = 0, $outv$sroa$29$2 = 0, $outv$sroa$29$3 = 0, $outv$sroa$29$4$lcssa = 0, $outv$sroa$29$419 = 0, $outv$sroa$29$5 = 0, $outv$sroa$29$6 = 0, $phitmp$i$i = 0, $s1$0$lcssa$i$i = 0;
 var $s1$09$i$i = 0, $s1$1$lcssa$i$i = 0, $s1$12$i$i = 0, $s2$0$lcssa$i$i = 0, $s2$010$i$i = 0, $s2$1$lcssa$i$i = 0, $s2$13$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $outv$sroa$0 = sp + 8|0;
 $deflatedata = sp + 4|0;
 $deflatesize = sp;
 HEAP32[$deflatedata>>2] = 0; //@line 2172 "libraries/lodepng/lodepng.cpp"
 HEAP32[$deflatesize>>2] = 0; //@line 2173 "libraries/lodepng/lodepng.cpp"
 $0 = HEAP32[$out>>2]|0; //@line 2184 "libraries/lodepng/lodepng.cpp"
 $1 = HEAP32[$outsize>>2]|0; //@line 2184 "libraries/lodepng/lodepng.cpp"
 HEAP32[$outv$sroa$0>>2] = $0; //@line 279 "libraries/lodepng/lodepng.cpp"
 $2 = (($1) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
 $3 = ($1|0)==(-1); //@line 225 "libraries/lodepng/lodepng.cpp"
 if ($3) {
  $4 = $0; //@line 289 "libraries/lodepng/lodepng.cpp"
  $15 = $4;$outv$sroa$29$0 = -1;
  label = 5;
 } else {
  $5 = $1 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
  $6 = ($5>>>0)<($2>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
  $7 = ($2*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
  $8 = $7 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
  $9 = $6 ? $2 : $8; //@line 227 "libraries/lodepng/lodepng.cpp"
  $10 = $0; //@line 228 "libraries/lodepng/lodepng.cpp"
  $11 = (_realloc($10,$9)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
  $12 = ($11|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
  if ($12) {
   $outv$sroa$18$0 = $1;$outv$sroa$29$1 = $1;
  } else {
   $13 = $11; //@line 232 "libraries/lodepng/lodepng.cpp"
   HEAP32[$outv$sroa$0>>2] = $13; //@line 232 "libraries/lodepng/lodepng.cpp"
   $15 = $11;$outv$sroa$29$0 = $9;
   label = 5;
  }
 }
 if ((label|0) == 5) {
  $14 = (($15) + ($1)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
  HEAP8[$14>>0] = 120; //@line 289 "libraries/lodepng/lodepng.cpp"
  $outv$sroa$18$0 = $2;$outv$sroa$29$1 = $outv$sroa$29$0;
 }
 $16 = (($outv$sroa$18$0) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
 $17 = ($outv$sroa$29$1>>>0)<($16>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
 if ($17) {
  $20 = $outv$sroa$29$1 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
  $21 = ($20>>>0)<($16>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
  $22 = ($16*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
  $23 = $22 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
  $24 = $21 ? $16 : $23; //@line 227 "libraries/lodepng/lodepng.cpp"
  $25 = HEAP32[$outv$sroa$0>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
  $26 = (_realloc($25,$24)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
  $27 = ($26|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
  $28 = $25; //@line 229 "libraries/lodepng/lodepng.cpp"
  if ($27) {
   $102 = $28;$103 = $25;$outv$sroa$18$1 = $outv$sroa$18$0;$outv$sroa$29$3 = $outv$sroa$29$1;
  } else {
   $29 = $26; //@line 232 "libraries/lodepng/lodepng.cpp"
   HEAP32[$outv$sroa$0>>2] = $29; //@line 232 "libraries/lodepng/lodepng.cpp"
   $101 = $29;$31 = $26;$outv$sroa$29$2 = $24;
   label = 10;
  }
 } else {
  $18 = HEAP32[$outv$sroa$0>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
  $19 = $18; //@line 225 "libraries/lodepng/lodepng.cpp"
  $101 = $19;$31 = $18;$outv$sroa$29$2 = $outv$sroa$29$1;
  label = 10;
 }
 if ((label|0) == 10) {
  $30 = (($31) + ($outv$sroa$18$0)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
  HEAP8[$30>>0] = 1; //@line 289 "libraries/lodepng/lodepng.cpp"
  $102 = $101;$103 = $31;$outv$sroa$18$1 = $16;$outv$sroa$29$3 = $outv$sroa$29$2;
 }
 $32 = ((($settings)) + 28|0); //@line 2051 "libraries/lodepng/lodepng.cpp"
 $33 = HEAP32[$32>>2]|0; //@line 2051 "libraries/lodepng/lodepng.cpp"
 $34 = ($33|0)==(0|0); //@line 2051 "libraries/lodepng/lodepng.cpp"
 if ($34) {
  $36 = (__Z15lodepng_deflatePPhPjPKhjPK23LodePNGCompressSettings($deflatedata,$deflatesize,$in,$insize,$settings)|0); //@line 2057 "libraries/lodepng/lodepng.cpp"
  $$0$i = $36;
 } else {
  $35 = (FUNCTION_TABLE_iiiiii[$33 & 8191]($deflatedata,$deflatesize,$in,$insize,$settings)|0); //@line 2053 "libraries/lodepng/lodepng.cpp"
  $$0$i = $35;
 }
 $37 = ($$0$i|0)==(0); //@line 2191 "libraries/lodepng/lodepng.cpp"
 if (!($37)) {
  $100 = $102;$outv$sroa$18$5 = $outv$sroa$18$1;
  HEAP32[$out>>2] = $100; //@line 2199 "libraries/lodepng/lodepng.cpp"
  HEAP32[$outsize>>2] = $outv$sroa$18$5; //@line 2200 "libraries/lodepng/lodepng.cpp"
  STACKTOP = sp;return ($$0$i|0); //@line 2202 "libraries/lodepng/lodepng.cpp"
 }
 $38 = ($insize|0)==(0); //@line 2072 "libraries/lodepng/lodepng.cpp"
 if ($38) {
  $s1$0$lcssa$i$i = 1;$s2$0$lcssa$i$i = 0;
 } else {
  $$017$i$i = $in;$$08$i$i = $insize;$s1$09$i$i = 1;$s2$010$i$i = 0;
  while(1) {
   $39 = ($$08$i$i>>>0)>(5550); //@line 2075 "libraries/lodepng/lodepng.cpp"
   $40 = $39 ? 5550 : $$08$i$i; //@line 2075 "libraries/lodepng/lodepng.cpp"
   $41 = (($$08$i$i) - ($40))|0; //@line 2076 "libraries/lodepng/lodepng.cpp"
   $42 = ($40|0)==(0); //@line 2077 "libraries/lodepng/lodepng.cpp"
   if ($42) {
    $$1$lcssa$i$i = $$017$i$i;$s1$1$lcssa$i$i = $s1$09$i$i;$s2$1$lcssa$i$i = $s2$010$i$i;
   } else {
    $$11$i$i = $$017$i$i;$amount$04$i$i = $40;$s1$12$i$i = $s1$09$i$i;$s2$13$i$i = $s2$010$i$i;
    while(1) {
     $43 = ((($$11$i$i)) + 1|0); //@line 2079 "libraries/lodepng/lodepng.cpp"
     $44 = HEAP8[$$11$i$i>>0]|0; //@line 2079 "libraries/lodepng/lodepng.cpp"
     $45 = $44&255; //@line 2079 "libraries/lodepng/lodepng.cpp"
     $46 = (($45) + ($s1$12$i$i))|0; //@line 2079 "libraries/lodepng/lodepng.cpp"
     $47 = (($46) + ($s2$13$i$i))|0; //@line 2080 "libraries/lodepng/lodepng.cpp"
     $48 = (($amount$04$i$i) + -1)|0; //@line 2081 "libraries/lodepng/lodepng.cpp"
     $49 = ($48|0)==(0); //@line 2077 "libraries/lodepng/lodepng.cpp"
     if ($49) {
      $$1$lcssa$i$i = $43;$s1$1$lcssa$i$i = $46;$s2$1$lcssa$i$i = $47;
      break;
     } else {
      $$11$i$i = $43;$amount$04$i$i = $48;$s1$12$i$i = $46;$s2$13$i$i = $47;
     }
    }
   }
   $50 = (($s1$1$lcssa$i$i>>>0) % 65521)&-1; //@line 2083 "libraries/lodepng/lodepng.cpp"
   $51 = (($s2$1$lcssa$i$i>>>0) % 65521)&-1; //@line 2084 "libraries/lodepng/lodepng.cpp"
   $52 = ($$08$i$i|0)==($40|0); //@line 2072 "libraries/lodepng/lodepng.cpp"
   if ($52) {
    $$lcssa54 = $50;$$lcssa55 = $51;
    break;
   } else {
    $$017$i$i = $$1$lcssa$i$i;$$08$i$i = $41;$s1$09$i$i = $50;$s2$010$i$i = $51;
   }
  }
  $phitmp$i$i = $$lcssa55 << 16; //@line 2087 "libraries/lodepng/lodepng.cpp"
  $s1$0$lcssa$i$i = $$lcssa54;$s2$0$lcssa$i$i = $phitmp$i$i;
 }
 $53 = HEAP32[$deflatesize>>2]|0; //@line 2194 "libraries/lodepng/lodepng.cpp"
 $54 = ($53|0)==(0); //@line 2194 "libraries/lodepng/lodepng.cpp"
 $55 = HEAP32[$deflatedata>>2]|0; //@line 2195 "libraries/lodepng/lodepng.cpp"
 if ($54) {
  $$lcssa = $55;$outv$sroa$18$2$lcssa = $outv$sroa$18$1;$outv$sroa$29$4$lcssa = $outv$sroa$29$3;
 } else {
  $104 = $103;$57 = $55;$66 = $103;$i$021 = 0;$outv$sroa$18$220 = $outv$sroa$18$1;$outv$sroa$29$419 = $outv$sroa$29$3;
  while(1) {
   $56 = (($57) + ($i$021)|0); //@line 2194 "libraries/lodepng/lodepng.cpp"
   $58 = HEAP8[$56>>0]|0; //@line 2194 "libraries/lodepng/lodepng.cpp"
   $59 = (($outv$sroa$18$220) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $60 = ($outv$sroa$29$419>>>0)<($59>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($60) {
    $61 = $outv$sroa$29$419 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $62 = ($61>>>0)<($59>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $63 = ($59*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $64 = $63 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $65 = $62 ? $59 : $64; //@line 227 "libraries/lodepng/lodepng.cpp"
    $67 = (_realloc($66,$65)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $68 = ($67|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if ($68) {
     $106 = $66;$107 = $66;$outv$sroa$18$3 = $outv$sroa$18$220;$outv$sroa$29$6 = $outv$sroa$29$419;
    } else {
     $69 = $67; //@line 232 "libraries/lodepng/lodepng.cpp"
     HEAP32[$outv$sroa$0>>2] = $69; //@line 232 "libraries/lodepng/lodepng.cpp"
     $105 = $67;$71 = $67;$outv$sroa$29$5 = $65;
     label = 24;
    }
   } else {
    $105 = $66;$71 = $104;$outv$sroa$29$5 = $outv$sroa$29$419;
    label = 24;
   }
   if ((label|0) == 24) {
    label = 0;
    $70 = (($71) + ($outv$sroa$18$220)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$70>>0] = $58; //@line 289 "libraries/lodepng/lodepng.cpp"
    $106 = $105;$107 = $71;$outv$sroa$18$3 = $59;$outv$sroa$29$6 = $outv$sroa$29$5;
   }
   $72 = (($i$021) + 1)|0; //@line 2194 "libraries/lodepng/lodepng.cpp"
   $73 = HEAP32[$deflatesize>>2]|0; //@line 2194 "libraries/lodepng/lodepng.cpp"
   $74 = ($72|0)==($73|0); //@line 2194 "libraries/lodepng/lodepng.cpp"
   $75 = HEAP32[$deflatedata>>2]|0; //@line 2195 "libraries/lodepng/lodepng.cpp"
   if ($74) {
    $$lcssa = $75;$outv$sroa$18$2$lcssa = $outv$sroa$18$3;$outv$sroa$29$4$lcssa = $outv$sroa$29$6;
    break;
   } else {
    $104 = $107;$57 = $75;$66 = $106;$i$021 = $72;$outv$sroa$18$220 = $outv$sroa$18$3;$outv$sroa$29$419 = $outv$sroa$29$6;
   }
  }
 }
 _free($$lcssa); //@line 76 "libraries/lodepng/lodepng.cpp"
 $76 = (($outv$sroa$18$2$lcssa) + 4)|0; //@line 360 "libraries/lodepng/lodepng.cpp"
 $77 = ($outv$sroa$29$4$lcssa>>>0)<($76>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
 if ($77) {
  $78 = $outv$sroa$29$4$lcssa << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
  $79 = ($78>>>0)<($76>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
  $80 = ($76*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
  $81 = $80 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
  $82 = $79 ? $76 : $81; //@line 227 "libraries/lodepng/lodepng.cpp"
  $83 = HEAP32[$outv$sroa$0>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
  $84 = (_realloc($83,$82)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
  $85 = ($84|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
  if ($85) {
   $89 = $83;$outv$sroa$18$4 = $outv$sroa$18$2$lcssa;
  } else {
   $86 = $84; //@line 232 "libraries/lodepng/lodepng.cpp"
   HEAP32[$outv$sroa$0>>2] = $86; //@line 232 "libraries/lodepng/lodepng.cpp"
   $89 = $84;$outv$sroa$18$4 = $76;
  }
 } else {
  $$pre = HEAP32[$outv$sroa$0>>2]|0; //@line 361 "libraries/lodepng/lodepng.cpp"
  $89 = $$pre;$outv$sroa$18$4 = $76;
 }
 $87 = (($outv$sroa$18$4) + -4)|0; //@line 361 "libraries/lodepng/lodepng.cpp"
 $88 = (($89) + ($87)|0); //@line 361 "libraries/lodepng/lodepng.cpp"
 $90 = $s2$0$lcssa$i$i >>> 24; //@line 350 "libraries/lodepng/lodepng.cpp"
 $91 = $90&255; //@line 350 "libraries/lodepng/lodepng.cpp"
 HEAP8[$88>>0] = $91; //@line 350 "libraries/lodepng/lodepng.cpp"
 $92 = $s2$0$lcssa$i$i >>> 16; //@line 351 "libraries/lodepng/lodepng.cpp"
 $93 = $92&255; //@line 351 "libraries/lodepng/lodepng.cpp"
 $94 = ((($88)) + 1|0); //@line 351 "libraries/lodepng/lodepng.cpp"
 HEAP8[$94>>0] = $93; //@line 351 "libraries/lodepng/lodepng.cpp"
 $95 = $s1$0$lcssa$i$i >>> 8; //@line 352 "libraries/lodepng/lodepng.cpp"
 $96 = $95&255; //@line 352 "libraries/lodepng/lodepng.cpp"
 $97 = ((($88)) + 2|0); //@line 352 "libraries/lodepng/lodepng.cpp"
 HEAP8[$97>>0] = $96; //@line 352 "libraries/lodepng/lodepng.cpp"
 $98 = $s1$0$lcssa$i$i&255; //@line 353 "libraries/lodepng/lodepng.cpp"
 $99 = ((($88)) + 3|0); //@line 353 "libraries/lodepng/lodepng.cpp"
 HEAP8[$99>>0] = $98; //@line 353 "libraries/lodepng/lodepng.cpp"
 $$pre31 = HEAP32[$outv$sroa$0>>2]|0; //@line 2199 "libraries/lodepng/lodepng.cpp"
 $100 = $$pre31;$outv$sroa$18$5 = $outv$sroa$18$4;
 HEAP32[$out>>2] = $100; //@line 2199 "libraries/lodepng/lodepng.cpp"
 HEAP32[$outsize>>2] = $outv$sroa$18$5; //@line 2200 "libraries/lodepng/lodepng.cpp"
 STACKTOP = sp;return ($$0$i|0); //@line 2202 "libraries/lodepng/lodepng.cpp"
}
function __Z20lodepng_chunk_createPPhPjjPKcPKh($out,$outlength,$length,$type,$data) {
 $out = $out|0;
 $outlength = $outlength|0;
 $length = $length|0;
 $type = $type|0;
 $data = $data|0;
 var $$0 = 0, $$lcssa = 0, $$pre = 0, $$pre2 = 0, $$pre3 = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $8 = 0, $9 = 0;
 var $c$0$lcssa$i$i = 0, $c$0$masked$i$i = 0, $c$01$i$i = 0, $exitcond$i$i = 0, $i$01 = 0, $n$02$i$i = 0, $or$cond = 0, $phitmp$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$outlength>>2]|0; //@line 2488 "libraries/lodepng/lodepng.cpp"
 $1 = (($length) + 12)|0; //@line 2488 "libraries/lodepng/lodepng.cpp"
 $2 = (($1) + ($0))|0; //@line 2488 "libraries/lodepng/lodepng.cpp"
 $3 = ($2>>>0)<($1>>>0); //@line 2489 "libraries/lodepng/lodepng.cpp"
 $4 = ($2>>>0)<($0>>>0); //@line 2489 "libraries/lodepng/lodepng.cpp"
 $or$cond = $3 | $4; //@line 2489 "libraries/lodepng/lodepng.cpp"
 if ($or$cond) {
  $$0 = 77;
  return ($$0|0); //@line 2512 "libraries/lodepng/lodepng.cpp"
 }
 $5 = HEAP32[$out>>2]|0; //@line 2490 "libraries/lodepng/lodepng.cpp"
 $6 = (_realloc($5,$2)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $7 = ($6|0)==(0|0); //@line 2491 "libraries/lodepng/lodepng.cpp"
 if ($7) {
  $$0 = 83;
  return ($$0|0); //@line 2512 "libraries/lodepng/lodepng.cpp"
 }
 HEAP32[$out>>2] = $6; //@line 2492 "libraries/lodepng/lodepng.cpp"
 HEAP32[$outlength>>2] = $2; //@line 2493 "libraries/lodepng/lodepng.cpp"
 $8 = (($6) + ($0)|0); //@line 2494 "libraries/lodepng/lodepng.cpp"
 $9 = $length >>> 24; //@line 350 "libraries/lodepng/lodepng.cpp"
 $10 = $9&255; //@line 350 "libraries/lodepng/lodepng.cpp"
 HEAP8[$8>>0] = $10; //@line 350 "libraries/lodepng/lodepng.cpp"
 $11 = $length >>> 16; //@line 351 "libraries/lodepng/lodepng.cpp"
 $12 = $11&255; //@line 351 "libraries/lodepng/lodepng.cpp"
 $13 = ((($8)) + 1|0); //@line 351 "libraries/lodepng/lodepng.cpp"
 HEAP8[$13>>0] = $12; //@line 351 "libraries/lodepng/lodepng.cpp"
 $14 = $length >>> 8; //@line 352 "libraries/lodepng/lodepng.cpp"
 $15 = $14&255; //@line 352 "libraries/lodepng/lodepng.cpp"
 $16 = ((($8)) + 2|0); //@line 352 "libraries/lodepng/lodepng.cpp"
 HEAP8[$16>>0] = $15; //@line 352 "libraries/lodepng/lodepng.cpp"
 $17 = $length&255; //@line 353 "libraries/lodepng/lodepng.cpp"
 $18 = ((($8)) + 3|0); //@line 353 "libraries/lodepng/lodepng.cpp"
 HEAP8[$18>>0] = $17; //@line 353 "libraries/lodepng/lodepng.cpp"
 $19 = HEAP8[$type>>0]|0; //@line 2500 "libraries/lodepng/lodepng.cpp"
 $20 = ((($8)) + 4|0); //@line 2500 "libraries/lodepng/lodepng.cpp"
 HEAP8[$20>>0] = $19; //@line 2500 "libraries/lodepng/lodepng.cpp"
 $21 = ((($type)) + 1|0); //@line 2501 "libraries/lodepng/lodepng.cpp"
 $22 = HEAP8[$21>>0]|0; //@line 2501 "libraries/lodepng/lodepng.cpp"
 $23 = ((($8)) + 5|0); //@line 2501 "libraries/lodepng/lodepng.cpp"
 HEAP8[$23>>0] = $22; //@line 2501 "libraries/lodepng/lodepng.cpp"
 $24 = ((($type)) + 2|0); //@line 2502 "libraries/lodepng/lodepng.cpp"
 $25 = HEAP8[$24>>0]|0; //@line 2502 "libraries/lodepng/lodepng.cpp"
 $26 = ((($8)) + 6|0); //@line 2502 "libraries/lodepng/lodepng.cpp"
 HEAP8[$26>>0] = $25; //@line 2502 "libraries/lodepng/lodepng.cpp"
 $27 = ((($type)) + 3|0); //@line 2503 "libraries/lodepng/lodepng.cpp"
 $28 = HEAP8[$27>>0]|0; //@line 2503 "libraries/lodepng/lodepng.cpp"
 $29 = ((($8)) + 7|0); //@line 2503 "libraries/lodepng/lodepng.cpp"
 HEAP8[$29>>0] = $28; //@line 2503 "libraries/lodepng/lodepng.cpp"
 $30 = ($length|0)==(0); //@line 2506 "libraries/lodepng/lodepng.cpp"
 if ($30) {
  $38 = $10;$41 = $12;$45 = $15;$49 = $17;
 } else {
  $i$01 = 0;
  while(1) {
   $31 = (($data) + ($i$01)|0); //@line 2506 "libraries/lodepng/lodepng.cpp"
   $32 = HEAP8[$31>>0]|0; //@line 2506 "libraries/lodepng/lodepng.cpp"
   $33 = (($i$01) + 8)|0; //@line 2506 "libraries/lodepng/lodepng.cpp"
   $34 = (($8) + ($33)|0); //@line 2506 "libraries/lodepng/lodepng.cpp"
   HEAP8[$34>>0] = $32; //@line 2506 "libraries/lodepng/lodepng.cpp"
   $35 = (($i$01) + 1)|0; //@line 2506 "libraries/lodepng/lodepng.cpp"
   $36 = ($35|0)==($length|0); //@line 2506 "libraries/lodepng/lodepng.cpp"
   if ($36) {
    break;
   } else {
    $i$01 = $35;
   }
  }
  $$pre = HEAP8[$8>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
  $$pre2 = HEAP8[$13>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
  $$pre3 = HEAP8[$16>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
  $$pre4 = HEAP8[$18>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
  $38 = $$pre;$41 = $$pre2;$45 = $$pre3;$49 = $$pre4;
 }
 $37 = $38&255; //@line 343 "libraries/lodepng/lodepng.cpp"
 $39 = $37 << 24; //@line 343 "libraries/lodepng/lodepng.cpp"
 $40 = $41&255; //@line 343 "libraries/lodepng/lodepng.cpp"
 $42 = $40 << 16; //@line 343 "libraries/lodepng/lodepng.cpp"
 $43 = $42 | $39; //@line 343 "libraries/lodepng/lodepng.cpp"
 $44 = $45&255; //@line 343 "libraries/lodepng/lodepng.cpp"
 $46 = $44 << 8; //@line 343 "libraries/lodepng/lodepng.cpp"
 $47 = $43 | $46; //@line 343 "libraries/lodepng/lodepng.cpp"
 $48 = $49&255; //@line 343 "libraries/lodepng/lodepng.cpp"
 $50 = $47 | $48; //@line 343 "libraries/lodepng/lodepng.cpp"
 $51 = (($50) + 4)|0; //@line 2448 "libraries/lodepng/lodepng.cpp"
 $52 = ($51|0)==(0); //@line 2338 "libraries/lodepng/lodepng.cpp"
 if ($52) {
  $c$0$lcssa$i$i = 0;
 } else {
  $c$01$i$i = -1;$n$02$i$i = 0;
  while(1) {
   $53 = (($20) + ($n$02$i$i)|0); //@line 2340 "libraries/lodepng/lodepng.cpp"
   $54 = HEAP8[$53>>0]|0; //@line 2340 "libraries/lodepng/lodepng.cpp"
   $55 = $54&255; //@line 2340 "libraries/lodepng/lodepng.cpp"
   $c$0$masked$i$i = $c$01$i$i & 255; //@line 2340 "libraries/lodepng/lodepng.cpp"
   $56 = $55 ^ $c$0$masked$i$i; //@line 2340 "libraries/lodepng/lodepng.cpp"
   $57 = (33860 + ($56<<2)|0); //@line 2340 "libraries/lodepng/lodepng.cpp"
   $58 = HEAP32[$57>>2]|0; //@line 2340 "libraries/lodepng/lodepng.cpp"
   $59 = $c$01$i$i >>> 8; //@line 2340 "libraries/lodepng/lodepng.cpp"
   $60 = $58 ^ $59; //@line 2340 "libraries/lodepng/lodepng.cpp"
   $61 = (($n$02$i$i) + 1)|0; //@line 2338 "libraries/lodepng/lodepng.cpp"
   $exitcond$i$i = ($61|0)==($51|0); //@line 2338 "libraries/lodepng/lodepng.cpp"
   if ($exitcond$i$i) {
    $$lcssa = $60;
    break;
   } else {
    $c$01$i$i = $60;$n$02$i$i = $61;
   }
  }
  $phitmp$i$i = $$lcssa ^ -1; //@line 2342 "libraries/lodepng/lodepng.cpp"
  $c$0$lcssa$i$i = $phitmp$i$i;
 }
 $62 = ((($8)) + 8|0); //@line 2449 "libraries/lodepng/lodepng.cpp"
 $63 = (($62) + ($50)|0); //@line 2449 "libraries/lodepng/lodepng.cpp"
 $64 = $c$0$lcssa$i$i >>> 24; //@line 350 "libraries/lodepng/lodepng.cpp"
 $65 = $64&255; //@line 350 "libraries/lodepng/lodepng.cpp"
 HEAP8[$63>>0] = $65; //@line 350 "libraries/lodepng/lodepng.cpp"
 $66 = $c$0$lcssa$i$i >>> 16; //@line 351 "libraries/lodepng/lodepng.cpp"
 $67 = $66&255; //@line 351 "libraries/lodepng/lodepng.cpp"
 $68 = ((($63)) + 1|0); //@line 351 "libraries/lodepng/lodepng.cpp"
 HEAP8[$68>>0] = $67; //@line 351 "libraries/lodepng/lodepng.cpp"
 $69 = $c$0$lcssa$i$i >>> 8; //@line 352 "libraries/lodepng/lodepng.cpp"
 $70 = $69&255; //@line 352 "libraries/lodepng/lodepng.cpp"
 $71 = ((($63)) + 2|0); //@line 352 "libraries/lodepng/lodepng.cpp"
 HEAP8[$71>>0] = $70; //@line 352 "libraries/lodepng/lodepng.cpp"
 $72 = $c$0$lcssa$i$i&255; //@line 353 "libraries/lodepng/lodepng.cpp"
 $73 = ((($63)) + 3|0); //@line 353 "libraries/lodepng/lodepng.cpp"
 HEAP8[$73>>0] = $72; //@line 353 "libraries/lodepng/lodepng.cpp"
 $$0 = 0;
 return ($$0|0); //@line 2512 "libraries/lodepng/lodepng.cpp"
}
function __Z16lodepng_add_textP11LodePNGInfoPKcS2_($info,$key,$str) {
 $info = $info|0;
 $key = $key|0;
 $str = $str|0;
 var $$0 = 0, $$pre$i = 0, $$pre$i7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 64|0); //@line 2772 "libraries/lodepng/lodepng.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 2772 "libraries/lodepng/lodepng.cpp"
 $2 = ((($info)) + 60|0); //@line 2772 "libraries/lodepng/lodepng.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 2772 "libraries/lodepng/lodepng.cpp"
 $4 = $3 << 2; //@line 2772 "libraries/lodepng/lodepng.cpp"
 $5 = (($4) + 4)|0; //@line 2772 "libraries/lodepng/lodepng.cpp"
 $6 = (_realloc($1,$5)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $7 = ((($info)) + 68|0); //@line 2773 "libraries/lodepng/lodepng.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 2773 "libraries/lodepng/lodepng.cpp"
 $9 = HEAP32[$2>>2]|0; //@line 2773 "libraries/lodepng/lodepng.cpp"
 $10 = $9 << 2; //@line 2773 "libraries/lodepng/lodepng.cpp"
 $11 = (($10) + 4)|0; //@line 2773 "libraries/lodepng/lodepng.cpp"
 $12 = (_realloc($8,$11)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $13 = ($6|0)!=(0|0); //@line 2774 "libraries/lodepng/lodepng.cpp"
 $14 = ($12|0)!=(0|0); //@line 2774 "libraries/lodepng/lodepng.cpp"
 $or$cond = $13 & $14; //@line 2774 "libraries/lodepng/lodepng.cpp"
 if (!($or$cond)) {
  _free($6); //@line 76 "libraries/lodepng/lodepng.cpp"
  _free($12); //@line 76 "libraries/lodepng/lodepng.cpp"
  $$0 = 83;
  return ($$0|0); //@line 2792 "libraries/lodepng/lodepng.cpp"
 }
 $15 = HEAP32[$2>>2]|0; //@line 2781 "libraries/lodepng/lodepng.cpp"
 $16 = (($15) + 1)|0; //@line 2781 "libraries/lodepng/lodepng.cpp"
 HEAP32[$2>>2] = $16; //@line 2781 "libraries/lodepng/lodepng.cpp"
 HEAP32[$0>>2] = $6; //@line 2782 "libraries/lodepng/lodepng.cpp"
 HEAP32[$7>>2] = $12; //@line 2783 "libraries/lodepng/lodepng.cpp"
 $17 = (($6) + ($15<<2)|0); //@line 2785 "libraries/lodepng/lodepng.cpp"
 HEAP32[$17>>2] = 0; //@line 314 "libraries/lodepng/lodepng.cpp"
 $18 = (_realloc(0,1)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $19 = ($18|0)==(0|0); //@line 303 "libraries/lodepng/lodepng.cpp"
 if (!($19)) {
  HEAP8[$18>>0] = 0; //@line 305 "libraries/lodepng/lodepng.cpp"
  HEAP32[$17>>2] = $18; //@line 306 "libraries/lodepng/lodepng.cpp"
 }
 $20 = HEAP32[$2>>2]|0; //@line 2786 "libraries/lodepng/lodepng.cpp"
 $21 = (($20) + -1)|0; //@line 2786 "libraries/lodepng/lodepng.cpp"
 $22 = HEAP32[$0>>2]|0; //@line 2786 "libraries/lodepng/lodepng.cpp"
 $23 = (($22) + ($21<<2)|0); //@line 2786 "libraries/lodepng/lodepng.cpp"
 $24 = (_strlen($key)|0); //@line 327 "libraries/lodepng/lodepng.cpp"
 $25 = HEAP32[$23>>2]|0; //@line 302 "libraries/lodepng/lodepng.cpp"
 $26 = (($24) + 1)|0; //@line 302 "libraries/lodepng/lodepng.cpp"
 $27 = (_realloc($25,$26)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $28 = ($27|0)==(0|0); //@line 303 "libraries/lodepng/lodepng.cpp"
 if (!($28)) {
  $29 = (($27) + ($24)|0); //@line 305 "libraries/lodepng/lodepng.cpp"
  HEAP8[$29>>0] = 0; //@line 305 "libraries/lodepng/lodepng.cpp"
  HEAP32[$23>>2] = $27; //@line 306 "libraries/lodepng/lodepng.cpp"
  $30 = ($24|0)==(0); //@line 330 "libraries/lodepng/lodepng.cpp"
  if (!($30)) {
   $31 = HEAP8[$key>>0]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
   HEAP8[$27>>0] = $31; //@line 332 "libraries/lodepng/lodepng.cpp"
   $32 = ($24|0)==(1); //@line 330 "libraries/lodepng/lodepng.cpp"
   if (!($32)) {
    $34 = 1;
    while(1) {
     $$pre$i = HEAP32[$23>>2]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
     $33 = (($key) + ($34)|0); //@line 332 "libraries/lodepng/lodepng.cpp"
     $35 = HEAP8[$33>>0]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
     $36 = (($$pre$i) + ($34)|0); //@line 332 "libraries/lodepng/lodepng.cpp"
     HEAP8[$36>>0] = $35; //@line 332 "libraries/lodepng/lodepng.cpp"
     $37 = (($34) + 1)|0; //@line 330 "libraries/lodepng/lodepng.cpp"
     $38 = ($37|0)==($24|0); //@line 330 "libraries/lodepng/lodepng.cpp"
     if ($38) {
      break;
     } else {
      $34 = $37;
     }
    }
   }
  }
 }
 $39 = HEAP32[$2>>2]|0; //@line 2788 "libraries/lodepng/lodepng.cpp"
 $40 = (($39) + -1)|0; //@line 2788 "libraries/lodepng/lodepng.cpp"
 $41 = HEAP32[$7>>2]|0; //@line 2788 "libraries/lodepng/lodepng.cpp"
 $42 = (($41) + ($40<<2)|0); //@line 2788 "libraries/lodepng/lodepng.cpp"
 HEAP32[$42>>2] = 0; //@line 314 "libraries/lodepng/lodepng.cpp"
 $43 = (_realloc(0,1)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $44 = ($43|0)==(0|0); //@line 303 "libraries/lodepng/lodepng.cpp"
 if (!($44)) {
  HEAP8[$43>>0] = 0; //@line 305 "libraries/lodepng/lodepng.cpp"
  HEAP32[$42>>2] = $43; //@line 306 "libraries/lodepng/lodepng.cpp"
 }
 $45 = HEAP32[$2>>2]|0; //@line 2789 "libraries/lodepng/lodepng.cpp"
 $46 = (($45) + -1)|0; //@line 2789 "libraries/lodepng/lodepng.cpp"
 $47 = HEAP32[$7>>2]|0; //@line 2789 "libraries/lodepng/lodepng.cpp"
 $48 = (($47) + ($46<<2)|0); //@line 2789 "libraries/lodepng/lodepng.cpp"
 $49 = (_strlen($str)|0); //@line 327 "libraries/lodepng/lodepng.cpp"
 $50 = HEAP32[$48>>2]|0; //@line 302 "libraries/lodepng/lodepng.cpp"
 $51 = (($49) + 1)|0; //@line 302 "libraries/lodepng/lodepng.cpp"
 $52 = (_realloc($50,$51)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $53 = ($52|0)==(0|0); //@line 303 "libraries/lodepng/lodepng.cpp"
 if ($53) {
  $$0 = 0;
  return ($$0|0); //@line 2792 "libraries/lodepng/lodepng.cpp"
 }
 $54 = (($52) + ($49)|0); //@line 305 "libraries/lodepng/lodepng.cpp"
 HEAP8[$54>>0] = 0; //@line 305 "libraries/lodepng/lodepng.cpp"
 HEAP32[$48>>2] = $52; //@line 306 "libraries/lodepng/lodepng.cpp"
 $55 = ($49|0)==(0); //@line 330 "libraries/lodepng/lodepng.cpp"
 if ($55) {
  $$0 = 0;
  return ($$0|0); //@line 2792 "libraries/lodepng/lodepng.cpp"
 }
 $56 = HEAP8[$str>>0]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
 HEAP8[$52>>0] = $56; //@line 332 "libraries/lodepng/lodepng.cpp"
 $57 = ($49|0)==(1); //@line 330 "libraries/lodepng/lodepng.cpp"
 if ($57) {
  $$0 = 0;
  return ($$0|0); //@line 2792 "libraries/lodepng/lodepng.cpp"
 } else {
  $59 = 1;
 }
 while(1) {
  $$pre$i7 = HEAP32[$48>>2]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
  $58 = (($str) + ($59)|0); //@line 332 "libraries/lodepng/lodepng.cpp"
  $60 = HEAP8[$58>>0]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
  $61 = (($$pre$i7) + ($59)|0); //@line 332 "libraries/lodepng/lodepng.cpp"
  HEAP8[$61>>0] = $60; //@line 332 "libraries/lodepng/lodepng.cpp"
  $62 = (($59) + 1)|0; //@line 330 "libraries/lodepng/lodepng.cpp"
  $63 = ($62|0)==($49|0); //@line 330 "libraries/lodepng/lodepng.cpp"
  if ($63) {
   $$0 = 0;
   break;
  } else {
   $59 = $62;
  }
 }
 return ($$0|0); //@line 2792 "libraries/lodepng/lodepng.cpp"
}
function __Z17lodepng_add_itextP11LodePNGInfoPKcS2_S2_S2_($info,$key,$langtag,$transkey,$str) {
 $info = $info|0;
 $key = $key|0;
 $langtag = $langtag|0;
 $transkey = $transkey|0;
 $str = $str|0;
 var $$0 = 0, $$pre$i = 0, $$pre$i11 = 0, $$pre$i19 = 0, $$pre$i27 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, $or$cond3 = 0, $or$cond5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 76|0); //@line 2845 "libraries/lodepng/lodepng.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 2845 "libraries/lodepng/lodepng.cpp"
 $2 = ((($info)) + 72|0); //@line 2845 "libraries/lodepng/lodepng.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 2845 "libraries/lodepng/lodepng.cpp"
 $4 = $3 << 2; //@line 2845 "libraries/lodepng/lodepng.cpp"
 $5 = (($4) + 4)|0; //@line 2845 "libraries/lodepng/lodepng.cpp"
 $6 = (_realloc($1,$5)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $7 = ((($info)) + 80|0); //@line 2846 "libraries/lodepng/lodepng.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 2846 "libraries/lodepng/lodepng.cpp"
 $9 = HEAP32[$2>>2]|0; //@line 2846 "libraries/lodepng/lodepng.cpp"
 $10 = $9 << 2; //@line 2846 "libraries/lodepng/lodepng.cpp"
 $11 = (($10) + 4)|0; //@line 2846 "libraries/lodepng/lodepng.cpp"
 $12 = (_realloc($8,$11)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $13 = ((($info)) + 84|0); //@line 2847 "libraries/lodepng/lodepng.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 2847 "libraries/lodepng/lodepng.cpp"
 $15 = HEAP32[$2>>2]|0; //@line 2847 "libraries/lodepng/lodepng.cpp"
 $16 = $15 << 2; //@line 2847 "libraries/lodepng/lodepng.cpp"
 $17 = (($16) + 4)|0; //@line 2847 "libraries/lodepng/lodepng.cpp"
 $18 = (_realloc($14,$17)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $19 = ((($info)) + 88|0); //@line 2848 "libraries/lodepng/lodepng.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 2848 "libraries/lodepng/lodepng.cpp"
 $21 = HEAP32[$2>>2]|0; //@line 2848 "libraries/lodepng/lodepng.cpp"
 $22 = $21 << 2; //@line 2848 "libraries/lodepng/lodepng.cpp"
 $23 = (($22) + 4)|0; //@line 2848 "libraries/lodepng/lodepng.cpp"
 $24 = (_realloc($20,$23)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $25 = ($6|0)!=(0|0); //@line 2849 "libraries/lodepng/lodepng.cpp"
 $26 = ($12|0)!=(0|0); //@line 2849 "libraries/lodepng/lodepng.cpp"
 $or$cond = $25 & $26; //@line 2849 "libraries/lodepng/lodepng.cpp"
 $27 = ($18|0)!=(0|0); //@line 2849 "libraries/lodepng/lodepng.cpp"
 $or$cond3 = $or$cond & $27; //@line 2849 "libraries/lodepng/lodepng.cpp"
 $28 = ($24|0)!=(0|0); //@line 2849 "libraries/lodepng/lodepng.cpp"
 $or$cond5 = $or$cond3 & $28; //@line 2849 "libraries/lodepng/lodepng.cpp"
 if (!($or$cond5)) {
  _free($6); //@line 76 "libraries/lodepng/lodepng.cpp"
  _free($12); //@line 76 "libraries/lodepng/lodepng.cpp"
  _free($18); //@line 76 "libraries/lodepng/lodepng.cpp"
  _free($24); //@line 76 "libraries/lodepng/lodepng.cpp"
  $$0 = 83;
  return ($$0|0); //@line 2877 "libraries/lodepng/lodepng.cpp"
 }
 $29 = HEAP32[$2>>2]|0; //@line 2858 "libraries/lodepng/lodepng.cpp"
 $30 = (($29) + 1)|0; //@line 2858 "libraries/lodepng/lodepng.cpp"
 HEAP32[$2>>2] = $30; //@line 2858 "libraries/lodepng/lodepng.cpp"
 HEAP32[$0>>2] = $6; //@line 2859 "libraries/lodepng/lodepng.cpp"
 HEAP32[$7>>2] = $12; //@line 2860 "libraries/lodepng/lodepng.cpp"
 HEAP32[$13>>2] = $18; //@line 2861 "libraries/lodepng/lodepng.cpp"
 HEAP32[$19>>2] = $24; //@line 2862 "libraries/lodepng/lodepng.cpp"
 $31 = (($6) + ($29<<2)|0); //@line 2864 "libraries/lodepng/lodepng.cpp"
 HEAP32[$31>>2] = 0; //@line 314 "libraries/lodepng/lodepng.cpp"
 $32 = (_realloc(0,1)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $33 = ($32|0)==(0|0); //@line 303 "libraries/lodepng/lodepng.cpp"
 if (!($33)) {
  HEAP8[$32>>0] = 0; //@line 305 "libraries/lodepng/lodepng.cpp"
  HEAP32[$31>>2] = $32; //@line 306 "libraries/lodepng/lodepng.cpp"
 }
 $34 = HEAP32[$2>>2]|0; //@line 2865 "libraries/lodepng/lodepng.cpp"
 $35 = (($34) + -1)|0; //@line 2865 "libraries/lodepng/lodepng.cpp"
 $36 = HEAP32[$0>>2]|0; //@line 2865 "libraries/lodepng/lodepng.cpp"
 $37 = (($36) + ($35<<2)|0); //@line 2865 "libraries/lodepng/lodepng.cpp"
 $38 = (_strlen($key)|0); //@line 327 "libraries/lodepng/lodepng.cpp"
 $39 = HEAP32[$37>>2]|0; //@line 302 "libraries/lodepng/lodepng.cpp"
 $40 = (($38) + 1)|0; //@line 302 "libraries/lodepng/lodepng.cpp"
 $41 = (_realloc($39,$40)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $42 = ($41|0)==(0|0); //@line 303 "libraries/lodepng/lodepng.cpp"
 if (!($42)) {
  $43 = (($41) + ($38)|0); //@line 305 "libraries/lodepng/lodepng.cpp"
  HEAP8[$43>>0] = 0; //@line 305 "libraries/lodepng/lodepng.cpp"
  HEAP32[$37>>2] = $41; //@line 306 "libraries/lodepng/lodepng.cpp"
  $44 = ($38|0)==(0); //@line 330 "libraries/lodepng/lodepng.cpp"
  if (!($44)) {
   $45 = HEAP8[$key>>0]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
   HEAP8[$41>>0] = $45; //@line 332 "libraries/lodepng/lodepng.cpp"
   $46 = ($38|0)==(1); //@line 330 "libraries/lodepng/lodepng.cpp"
   if (!($46)) {
    $48 = 1;
    while(1) {
     $$pre$i = HEAP32[$37>>2]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
     $47 = (($key) + ($48)|0); //@line 332 "libraries/lodepng/lodepng.cpp"
     $49 = HEAP8[$47>>0]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
     $50 = (($$pre$i) + ($48)|0); //@line 332 "libraries/lodepng/lodepng.cpp"
     HEAP8[$50>>0] = $49; //@line 332 "libraries/lodepng/lodepng.cpp"
     $51 = (($48) + 1)|0; //@line 330 "libraries/lodepng/lodepng.cpp"
     $52 = ($51|0)==($38|0); //@line 330 "libraries/lodepng/lodepng.cpp"
     if ($52) {
      break;
     } else {
      $48 = $51;
     }
    }
   }
  }
 }
 $53 = HEAP32[$2>>2]|0; //@line 2867 "libraries/lodepng/lodepng.cpp"
 $54 = (($53) + -1)|0; //@line 2867 "libraries/lodepng/lodepng.cpp"
 $55 = HEAP32[$7>>2]|0; //@line 2867 "libraries/lodepng/lodepng.cpp"
 $56 = (($55) + ($54<<2)|0); //@line 2867 "libraries/lodepng/lodepng.cpp"
 HEAP32[$56>>2] = 0; //@line 314 "libraries/lodepng/lodepng.cpp"
 $57 = (_realloc(0,1)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $58 = ($57|0)==(0|0); //@line 303 "libraries/lodepng/lodepng.cpp"
 if (!($58)) {
  HEAP8[$57>>0] = 0; //@line 305 "libraries/lodepng/lodepng.cpp"
  HEAP32[$56>>2] = $57; //@line 306 "libraries/lodepng/lodepng.cpp"
 }
 $59 = HEAP32[$2>>2]|0; //@line 2868 "libraries/lodepng/lodepng.cpp"
 $60 = (($59) + -1)|0; //@line 2868 "libraries/lodepng/lodepng.cpp"
 $61 = HEAP32[$7>>2]|0; //@line 2868 "libraries/lodepng/lodepng.cpp"
 $62 = (($61) + ($60<<2)|0); //@line 2868 "libraries/lodepng/lodepng.cpp"
 $63 = (_strlen($langtag)|0); //@line 327 "libraries/lodepng/lodepng.cpp"
 $64 = HEAP32[$62>>2]|0; //@line 302 "libraries/lodepng/lodepng.cpp"
 $65 = (($63) + 1)|0; //@line 302 "libraries/lodepng/lodepng.cpp"
 $66 = (_realloc($64,$65)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $67 = ($66|0)==(0|0); //@line 303 "libraries/lodepng/lodepng.cpp"
 if (!($67)) {
  $68 = (($66) + ($63)|0); //@line 305 "libraries/lodepng/lodepng.cpp"
  HEAP8[$68>>0] = 0; //@line 305 "libraries/lodepng/lodepng.cpp"
  HEAP32[$62>>2] = $66; //@line 306 "libraries/lodepng/lodepng.cpp"
  $69 = ($63|0)==(0); //@line 330 "libraries/lodepng/lodepng.cpp"
  if (!($69)) {
   $70 = HEAP8[$langtag>>0]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
   HEAP8[$66>>0] = $70; //@line 332 "libraries/lodepng/lodepng.cpp"
   $71 = ($63|0)==(1); //@line 330 "libraries/lodepng/lodepng.cpp"
   if (!($71)) {
    $73 = 1;
    while(1) {
     $$pre$i11 = HEAP32[$62>>2]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
     $72 = (($langtag) + ($73)|0); //@line 332 "libraries/lodepng/lodepng.cpp"
     $74 = HEAP8[$72>>0]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
     $75 = (($$pre$i11) + ($73)|0); //@line 332 "libraries/lodepng/lodepng.cpp"
     HEAP8[$75>>0] = $74; //@line 332 "libraries/lodepng/lodepng.cpp"
     $76 = (($73) + 1)|0; //@line 330 "libraries/lodepng/lodepng.cpp"
     $77 = ($76|0)==($63|0); //@line 330 "libraries/lodepng/lodepng.cpp"
     if ($77) {
      break;
     } else {
      $73 = $76;
     }
    }
   }
  }
 }
 $78 = HEAP32[$2>>2]|0; //@line 2870 "libraries/lodepng/lodepng.cpp"
 $79 = (($78) + -1)|0; //@line 2870 "libraries/lodepng/lodepng.cpp"
 $80 = HEAP32[$13>>2]|0; //@line 2870 "libraries/lodepng/lodepng.cpp"
 $81 = (($80) + ($79<<2)|0); //@line 2870 "libraries/lodepng/lodepng.cpp"
 HEAP32[$81>>2] = 0; //@line 314 "libraries/lodepng/lodepng.cpp"
 $82 = (_realloc(0,1)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $83 = ($82|0)==(0|0); //@line 303 "libraries/lodepng/lodepng.cpp"
 if (!($83)) {
  HEAP8[$82>>0] = 0; //@line 305 "libraries/lodepng/lodepng.cpp"
  HEAP32[$81>>2] = $82; //@line 306 "libraries/lodepng/lodepng.cpp"
 }
 $84 = HEAP32[$2>>2]|0; //@line 2871 "libraries/lodepng/lodepng.cpp"
 $85 = (($84) + -1)|0; //@line 2871 "libraries/lodepng/lodepng.cpp"
 $86 = HEAP32[$13>>2]|0; //@line 2871 "libraries/lodepng/lodepng.cpp"
 $87 = (($86) + ($85<<2)|0); //@line 2871 "libraries/lodepng/lodepng.cpp"
 $88 = (_strlen($transkey)|0); //@line 327 "libraries/lodepng/lodepng.cpp"
 $89 = HEAP32[$87>>2]|0; //@line 302 "libraries/lodepng/lodepng.cpp"
 $90 = (($88) + 1)|0; //@line 302 "libraries/lodepng/lodepng.cpp"
 $91 = (_realloc($89,$90)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $92 = ($91|0)==(0|0); //@line 303 "libraries/lodepng/lodepng.cpp"
 if (!($92)) {
  $93 = (($91) + ($88)|0); //@line 305 "libraries/lodepng/lodepng.cpp"
  HEAP8[$93>>0] = 0; //@line 305 "libraries/lodepng/lodepng.cpp"
  HEAP32[$87>>2] = $91; //@line 306 "libraries/lodepng/lodepng.cpp"
  $94 = ($88|0)==(0); //@line 330 "libraries/lodepng/lodepng.cpp"
  if (!($94)) {
   $95 = HEAP8[$transkey>>0]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
   HEAP8[$91>>0] = $95; //@line 332 "libraries/lodepng/lodepng.cpp"
   $96 = ($88|0)==(1); //@line 330 "libraries/lodepng/lodepng.cpp"
   if (!($96)) {
    $98 = 1;
    while(1) {
     $$pre$i19 = HEAP32[$87>>2]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
     $97 = (($transkey) + ($98)|0); //@line 332 "libraries/lodepng/lodepng.cpp"
     $99 = HEAP8[$97>>0]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
     $100 = (($$pre$i19) + ($98)|0); //@line 332 "libraries/lodepng/lodepng.cpp"
     HEAP8[$100>>0] = $99; //@line 332 "libraries/lodepng/lodepng.cpp"
     $101 = (($98) + 1)|0; //@line 330 "libraries/lodepng/lodepng.cpp"
     $102 = ($101|0)==($88|0); //@line 330 "libraries/lodepng/lodepng.cpp"
     if ($102) {
      break;
     } else {
      $98 = $101;
     }
    }
   }
  }
 }
 $103 = HEAP32[$2>>2]|0; //@line 2873 "libraries/lodepng/lodepng.cpp"
 $104 = (($103) + -1)|0; //@line 2873 "libraries/lodepng/lodepng.cpp"
 $105 = HEAP32[$19>>2]|0; //@line 2873 "libraries/lodepng/lodepng.cpp"
 $106 = (($105) + ($104<<2)|0); //@line 2873 "libraries/lodepng/lodepng.cpp"
 HEAP32[$106>>2] = 0; //@line 314 "libraries/lodepng/lodepng.cpp"
 $107 = (_realloc(0,1)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $108 = ($107|0)==(0|0); //@line 303 "libraries/lodepng/lodepng.cpp"
 if (!($108)) {
  HEAP8[$107>>0] = 0; //@line 305 "libraries/lodepng/lodepng.cpp"
  HEAP32[$106>>2] = $107; //@line 306 "libraries/lodepng/lodepng.cpp"
 }
 $109 = HEAP32[$2>>2]|0; //@line 2874 "libraries/lodepng/lodepng.cpp"
 $110 = (($109) + -1)|0; //@line 2874 "libraries/lodepng/lodepng.cpp"
 $111 = HEAP32[$19>>2]|0; //@line 2874 "libraries/lodepng/lodepng.cpp"
 $112 = (($111) + ($110<<2)|0); //@line 2874 "libraries/lodepng/lodepng.cpp"
 $113 = (_strlen($str)|0); //@line 327 "libraries/lodepng/lodepng.cpp"
 $114 = HEAP32[$112>>2]|0; //@line 302 "libraries/lodepng/lodepng.cpp"
 $115 = (($113) + 1)|0; //@line 302 "libraries/lodepng/lodepng.cpp"
 $116 = (_realloc($114,$115)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
 $117 = ($116|0)==(0|0); //@line 303 "libraries/lodepng/lodepng.cpp"
 if ($117) {
  $$0 = 0;
  return ($$0|0); //@line 2877 "libraries/lodepng/lodepng.cpp"
 }
 $118 = (($116) + ($113)|0); //@line 305 "libraries/lodepng/lodepng.cpp"
 HEAP8[$118>>0] = 0; //@line 305 "libraries/lodepng/lodepng.cpp"
 HEAP32[$112>>2] = $116; //@line 306 "libraries/lodepng/lodepng.cpp"
 $119 = ($113|0)==(0); //@line 330 "libraries/lodepng/lodepng.cpp"
 if ($119) {
  $$0 = 0;
  return ($$0|0); //@line 2877 "libraries/lodepng/lodepng.cpp"
 }
 $120 = HEAP8[$str>>0]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
 HEAP8[$116>>0] = $120; //@line 332 "libraries/lodepng/lodepng.cpp"
 $121 = ($113|0)==(1); //@line 330 "libraries/lodepng/lodepng.cpp"
 if ($121) {
  $$0 = 0;
  return ($$0|0); //@line 2877 "libraries/lodepng/lodepng.cpp"
 } else {
  $123 = 1;
 }
 while(1) {
  $$pre$i27 = HEAP32[$112>>2]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
  $122 = (($str) + ($123)|0); //@line 332 "libraries/lodepng/lodepng.cpp"
  $124 = HEAP8[$122>>0]|0; //@line 332 "libraries/lodepng/lodepng.cpp"
  $125 = (($$pre$i27) + ($123)|0); //@line 332 "libraries/lodepng/lodepng.cpp"
  HEAP8[$125>>0] = $124; //@line 332 "libraries/lodepng/lodepng.cpp"
  $126 = (($123) + 1)|0; //@line 330 "libraries/lodepng/lodepng.cpp"
  $127 = ($126|0)==($113|0); //@line 330 "libraries/lodepng/lodepng.cpp"
  if ($127) {
   $$0 = 0;
   break;
  } else {
   $123 = $126;
  }
 }
 return ($$0|0); //@line 2877 "libraries/lodepng/lodepng.cpp"
}
function __Z17lodepng_info_copyP11LodePNGInfoPKS_($dest,$source) {
 $dest = $dest|0;
 $source = $source|0;
 var $$2$i = 0, $$pre = 0, $$pre$i$pre = 0, $$pre2$i$pre = 0, $$pre47 = 0, $$pre49 = 0, $$pre51 = 0, $$pre7$i$pre = 0, $$pre9$i$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$02$i = 0;
 var $i$02$i18 = 0, $or$cond$i = 0, $or$cond10$i = 0, $or$cond11$i = 0, $phitmp$i = 0, $phitmp3$i = 0, $scevgep4$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __Z20lodepng_info_cleanupP11LodePNGInfo($dest); //@line 2913 "libraries/lodepng/lodepng.cpp"
 _memcpy(($dest|0),($source|0),160)|0; //@line 2914 "libraries/lodepng/lodepng.cpp"
 $0 = ((($dest)) + 12|0); //@line 2915 "libraries/lodepng/lodepng.cpp"
 $1 = ((($dest)) + 28|0); //@line 2556 "libraries/lodepng/lodepng.cpp"
 ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0;HEAP32[$1+12>>2]=0|0; //@line 2557 "libraries/lodepng/lodepng.cpp"
 HEAP32[$0>>2] = 6; //@line 2558 "libraries/lodepng/lodepng.cpp"
 $2 = ((($dest)) + 16|0); //@line 2559 "libraries/lodepng/lodepng.cpp"
 HEAP32[$2>>2] = 8; //@line 2559 "libraries/lodepng/lodepng.cpp"
 $3 = ((($dest)) + 20|0); //@line 2560 "libraries/lodepng/lodepng.cpp"
 $4 = ((($source)) + 12|0); //@line 2916 "libraries/lodepng/lodepng.cpp"
 $5 = $3; //@line 2561 "libraries/lodepng/lodepng.cpp"
 $6 = $5; //@line 2561 "libraries/lodepng/lodepng.cpp"
 HEAP32[$6>>2] = 0; //@line 2561 "libraries/lodepng/lodepng.cpp"
 $7 = (($5) + 4)|0; //@line 2561 "libraries/lodepng/lodepng.cpp"
 $8 = $7; //@line 2561 "libraries/lodepng/lodepng.cpp"
 HEAP32[$8>>2] = 0; //@line 2561 "libraries/lodepng/lodepng.cpp"
 ;HEAP32[$0>>2]=HEAP32[$4>>2]|0;HEAP32[$0+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$4+8>>2]|0;HEAP32[$0+12>>2]=HEAP32[$4+12>>2]|0;HEAP32[$0+16>>2]=HEAP32[$4+16>>2]|0;HEAP32[$0+20>>2]=HEAP32[$4+20>>2]|0;HEAP32[$0+24>>2]=HEAP32[$4+24>>2]|0;HEAP32[$0+28>>2]=HEAP32[$4+28>>2]|0; //@line 2573 "libraries/lodepng/lodepng.cpp"
 $9 = ((($source)) + 20|0); //@line 2574 "libraries/lodepng/lodepng.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 2574 "libraries/lodepng/lodepng.cpp"
 $11 = ($10|0)==(0|0); //@line 2574 "libraries/lodepng/lodepng.cpp"
 do {
  if (!($11)) {
   $12 = (_malloc(1024)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
   HEAP32[$3>>2] = $12; //@line 2576 "libraries/lodepng/lodepng.cpp"
   $13 = ($12|0)==(0|0); //@line 2577 "libraries/lodepng/lodepng.cpp"
   $14 = ((($source)) + 24|0); //@line 2577 "libraries/lodepng/lodepng.cpp"
   $15 = HEAP32[$14>>2]|0; //@line 2577 "libraries/lodepng/lodepng.cpp"
   if ($13) {
    $29 = ($15|0)==(0); //@line 2577 "libraries/lodepng/lodepng.cpp"
    if ($29) {
     break;
    } else {
     $$2$i = 83;
    }
    return ($$2$i|0);
   }
   $phitmp$i = $15 & 1073741823; //@line 2577 "libraries/lodepng/lodepng.cpp"
   $phitmp3$i = ($phitmp$i|0)==(0); //@line 2577 "libraries/lodepng/lodepng.cpp"
   if (!($phitmp3$i)) {
    $16 = HEAP32[$9>>2]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
    $17 = HEAP8[$16>>0]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
    HEAP8[$12>>0] = $17; //@line 2578 "libraries/lodepng/lodepng.cpp"
    $18 = ((($16)) + 1|0); //@line 2578 "libraries/lodepng/lodepng.cpp"
    $19 = HEAP8[$18>>0]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
    $20 = ((($12)) + 1|0); //@line 2578 "libraries/lodepng/lodepng.cpp"
    HEAP8[$20>>0] = $19; //@line 2578 "libraries/lodepng/lodepng.cpp"
    $22 = 2;
    while(1) {
     $$pre2$i$pre = HEAP32[$3>>2]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
     $$pre = HEAP32[$9>>2]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
     $21 = (($$pre) + ($22)|0); //@line 2578 "libraries/lodepng/lodepng.cpp"
     $23 = HEAP8[$21>>0]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
     $24 = (($$pre2$i$pre) + ($22)|0); //@line 2578 "libraries/lodepng/lodepng.cpp"
     HEAP8[$24>>0] = $23; //@line 2578 "libraries/lodepng/lodepng.cpp"
     $25 = (($22) + 1)|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
     $26 = HEAP32[$14>>2]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
     $27 = $26 << 2; //@line 2578 "libraries/lodepng/lodepng.cpp"
     $28 = ($25|0)==($27|0); //@line 2578 "libraries/lodepng/lodepng.cpp"
     if ($28) {
      break;
     } else {
      $22 = $25;
     }
    }
   }
  }
 } while(0);
 $30 = ((($dest)) + 64|0); //@line 2755 "libraries/lodepng/lodepng.cpp"
 HEAP32[$30>>2] = 0; //@line 2755 "libraries/lodepng/lodepng.cpp"
 $31 = ((($dest)) + 68|0); //@line 2756 "libraries/lodepng/lodepng.cpp"
 HEAP32[$31>>2] = 0; //@line 2756 "libraries/lodepng/lodepng.cpp"
 $32 = ((($dest)) + 60|0); //@line 2757 "libraries/lodepng/lodepng.cpp"
 HEAP32[$32>>2] = 0; //@line 2757 "libraries/lodepng/lodepng.cpp"
 $33 = ((($source)) + 64|0); //@line 2760 "libraries/lodepng/lodepng.cpp"
 $34 = ((($source)) + 60|0); //@line 2758 "libraries/lodepng/lodepng.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 2758 "libraries/lodepng/lodepng.cpp"
 $36 = ($35|0)==(0); //@line 2758 "libraries/lodepng/lodepng.cpp"
 L11: do {
  if (!($36)) {
   $37 = ((($source)) + 68|0); //@line 2760 "libraries/lodepng/lodepng.cpp"
   $i$02$i = 0;
   while(1) {
    $41 = HEAP32[$33>>2]|0; //@line 2760 "libraries/lodepng/lodepng.cpp"
    $42 = (($41) + ($i$02$i<<2)|0); //@line 2760 "libraries/lodepng/lodepng.cpp"
    $43 = HEAP32[$42>>2]|0; //@line 2760 "libraries/lodepng/lodepng.cpp"
    $44 = HEAP32[$37>>2]|0; //@line 2760 "libraries/lodepng/lodepng.cpp"
    $45 = (($44) + ($i$02$i<<2)|0); //@line 2760 "libraries/lodepng/lodepng.cpp"
    $46 = HEAP32[$45>>2]|0; //@line 2760 "libraries/lodepng/lodepng.cpp"
    $47 = (__Z16lodepng_add_textP11LodePNGInfoPKcS2_($dest,$43,$46)|0); //@line 2760 "libraries/lodepng/lodepng.cpp"
    $48 = ($47|0)==(0); //@line 2760 "libraries/lodepng/lodepng.cpp"
    $40 = (($i$02$i) + 1)|0; //@line 2758 "libraries/lodepng/lodepng.cpp"
    if (!($48)) {
     $$2$i = $47;
     break;
    }
    $38 = HEAP32[$34>>2]|0; //@line 2758 "libraries/lodepng/lodepng.cpp"
    $39 = ($40|0)==($38|0); //@line 2758 "libraries/lodepng/lodepng.cpp"
    if ($39) {
     break L11;
    } else {
     $i$02$i = $40;
    }
   }
   return ($$2$i|0);
  }
 } while(0);
 $49 = ((($dest)) + 72|0); //@line 2828 "libraries/lodepng/lodepng.cpp"
 $50 = ((($source)) + 76|0); //@line 2831 "libraries/lodepng/lodepng.cpp"
 $51 = ((($source)) + 72|0); //@line 2829 "libraries/lodepng/lodepng.cpp"
 ;HEAP32[$49>>2]=0|0;HEAP32[$49+4>>2]=0|0;HEAP32[$49+8>>2]=0|0;HEAP32[$49+12>>2]=0|0;HEAP32[$49+16>>2]=0|0; //@line 2828 "libraries/lodepng/lodepng.cpp"
 $52 = HEAP32[$51>>2]|0; //@line 2829 "libraries/lodepng/lodepng.cpp"
 $53 = ($52|0)==(0); //@line 2829 "libraries/lodepng/lodepng.cpp"
 L18: do {
  if (!($53)) {
   $54 = ((($source)) + 80|0); //@line 2831 "libraries/lodepng/lodepng.cpp"
   $55 = ((($source)) + 84|0); //@line 2831 "libraries/lodepng/lodepng.cpp"
   $56 = ((($source)) + 88|0); //@line 2831 "libraries/lodepng/lodepng.cpp"
   $i$02$i18 = 0;
   while(1) {
    $60 = HEAP32[$50>>2]|0; //@line 2831 "libraries/lodepng/lodepng.cpp"
    $61 = (($60) + ($i$02$i18<<2)|0); //@line 2831 "libraries/lodepng/lodepng.cpp"
    $62 = HEAP32[$61>>2]|0; //@line 2831 "libraries/lodepng/lodepng.cpp"
    $63 = HEAP32[$54>>2]|0; //@line 2831 "libraries/lodepng/lodepng.cpp"
    $64 = (($63) + ($i$02$i18<<2)|0); //@line 2831 "libraries/lodepng/lodepng.cpp"
    $65 = HEAP32[$64>>2]|0; //@line 2831 "libraries/lodepng/lodepng.cpp"
    $66 = HEAP32[$55>>2]|0; //@line 2831 "libraries/lodepng/lodepng.cpp"
    $67 = (($66) + ($i$02$i18<<2)|0); //@line 2831 "libraries/lodepng/lodepng.cpp"
    $68 = HEAP32[$67>>2]|0; //@line 2831 "libraries/lodepng/lodepng.cpp"
    $69 = HEAP32[$56>>2]|0; //@line 2831 "libraries/lodepng/lodepng.cpp"
    $70 = (($69) + ($i$02$i18<<2)|0); //@line 2831 "libraries/lodepng/lodepng.cpp"
    $71 = HEAP32[$70>>2]|0; //@line 2831 "libraries/lodepng/lodepng.cpp"
    $72 = (__Z17lodepng_add_itextP11LodePNGInfoPKcS2_S2_S2_($dest,$62,$65,$68,$71)|0); //@line 2831 "libraries/lodepng/lodepng.cpp"
    $73 = ($72|0)==(0); //@line 2831 "libraries/lodepng/lodepng.cpp"
    $59 = (($i$02$i18) + 1)|0; //@line 2829 "libraries/lodepng/lodepng.cpp"
    if (!($73)) {
     $$2$i = $72;
     break;
    }
    $57 = HEAP32[$51>>2]|0; //@line 2829 "libraries/lodepng/lodepng.cpp"
    $58 = ($59|0)==($57|0); //@line 2829 "libraries/lodepng/lodepng.cpp"
    if ($58) {
     break L18;
    } else {
     $i$02$i18 = $59;
    }
   }
   return ($$2$i|0);
  }
 } while(0);
 $scevgep4$i = ((($dest)) + 136|0); //@line 2700 "libraries/lodepng/lodepng.cpp"
 ;HEAP32[$scevgep4$i>>2]=0|0;HEAP32[$scevgep4$i+4>>2]=0|0;HEAP32[$scevgep4$i+8>>2]=0|0;HEAP32[$scevgep4$i+12>>2]=0|0;HEAP32[$scevgep4$i+16>>2]=0|0;HEAP32[$scevgep4$i+20>>2]=0|0; //@line 2701 "libraries/lodepng/lodepng.cpp"
 $74 = ((($dest)) + 140|0); //@line 2707 "libraries/lodepng/lodepng.cpp"
 $75 = HEAP32[$74>>2]|0; //@line 2707 "libraries/lodepng/lodepng.cpp"
 _free($75); //@line 76 "libraries/lodepng/lodepng.cpp"
 $76 = ((($dest)) + 144|0); //@line 2707 "libraries/lodepng/lodepng.cpp"
 $77 = HEAP32[$76>>2]|0; //@line 2707 "libraries/lodepng/lodepng.cpp"
 _free($77); //@line 76 "libraries/lodepng/lodepng.cpp"
 $78 = ((($source)) + 148|0); //@line 2719 "libraries/lodepng/lodepng.cpp"
 $79 = HEAP32[$78>>2]|0; //@line 2721 "libraries/lodepng/lodepng.cpp"
 $80 = ((($dest)) + 148|0); //@line 2719 "libraries/lodepng/lodepng.cpp"
 HEAP32[$80>>2] = $79; //@line 2719 "libraries/lodepng/lodepng.cpp"
 $81 = HEAP32[$78>>2]|0; //@line 2722 "libraries/lodepng/lodepng.cpp"
 $82 = (_malloc($81)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
 HEAP32[$scevgep4$i>>2] = $82; //@line 2720 "libraries/lodepng/lodepng.cpp"
 $83 = ($82|0)!=(0|0); //@line 2721 "libraries/lodepng/lodepng.cpp"
 $84 = ($79|0)==(0); //@line 2721 "libraries/lodepng/lodepng.cpp"
 $or$cond$i = $84 | $83; //@line 2721 "libraries/lodepng/lodepng.cpp"
 if (!($or$cond$i)) {
  $$2$i = 83;
  return ($$2$i|0);
 }
 $85 = ($81|0)==(0); //@line 2722 "libraries/lodepng/lodepng.cpp"
 if (!($85)) {
  $86 = ((($source)) + 136|0); //@line 2724 "libraries/lodepng/lodepng.cpp"
  $87 = HEAP32[$86>>2]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
  $88 = HEAP8[$87>>0]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
  HEAP8[$82>>0] = $88; //@line 2724 "libraries/lodepng/lodepng.cpp"
  $89 = ($81>>>0)>(1); //@line 2722 "libraries/lodepng/lodepng.cpp"
  if ($89) {
   $90 = ((($87)) + 1|0); //@line 2724 "libraries/lodepng/lodepng.cpp"
   $91 = HEAP8[$90>>0]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
   $92 = ((($82)) + 1|0); //@line 2724 "libraries/lodepng/lodepng.cpp"
   HEAP8[$92>>0] = $91; //@line 2724 "libraries/lodepng/lodepng.cpp"
   $93 = HEAP32[$78>>2]|0; //@line 2722 "libraries/lodepng/lodepng.cpp"
   $94 = ($93>>>0)>(2); //@line 2722 "libraries/lodepng/lodepng.cpp"
   if ($94) {
    $96 = 2;
    while(1) {
     $$pre$i$pre = HEAP32[$scevgep4$i>>2]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
     $$pre47 = HEAP32[$86>>2]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
     $95 = (($$pre47) + ($96)|0); //@line 2724 "libraries/lodepng/lodepng.cpp"
     $97 = HEAP8[$95>>0]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
     $98 = (($$pre$i$pre) + ($96)|0); //@line 2724 "libraries/lodepng/lodepng.cpp"
     HEAP8[$98>>0] = $97; //@line 2724 "libraries/lodepng/lodepng.cpp"
     $99 = (($96) + 1)|0; //@line 2722 "libraries/lodepng/lodepng.cpp"
     $100 = HEAP32[$78>>2]|0; //@line 2722 "libraries/lodepng/lodepng.cpp"
     $101 = ($99>>>0)<($100>>>0); //@line 2722 "libraries/lodepng/lodepng.cpp"
     if ($101) {
      $96 = $99;
     } else {
      break;
     }
    }
   }
  }
 }
 $102 = ((($source)) + 152|0); //@line 2719 "libraries/lodepng/lodepng.cpp"
 $103 = HEAP32[$102>>2]|0; //@line 2721 "libraries/lodepng/lodepng.cpp"
 $104 = ((($dest)) + 152|0); //@line 2719 "libraries/lodepng/lodepng.cpp"
 HEAP32[$104>>2] = $103; //@line 2719 "libraries/lodepng/lodepng.cpp"
 $105 = HEAP32[$102>>2]|0; //@line 2722 "libraries/lodepng/lodepng.cpp"
 $106 = (_malloc($105)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
 HEAP32[$74>>2] = $106; //@line 2720 "libraries/lodepng/lodepng.cpp"
 $107 = ($106|0)!=(0|0); //@line 2721 "libraries/lodepng/lodepng.cpp"
 $108 = ($103|0)==(0); //@line 2721 "libraries/lodepng/lodepng.cpp"
 $or$cond10$i = $108 | $107; //@line 2721 "libraries/lodepng/lodepng.cpp"
 if (!($or$cond10$i)) {
  $$2$i = 83;
  return ($$2$i|0);
 }
 $109 = ($105|0)==(0); //@line 2722 "libraries/lodepng/lodepng.cpp"
 if (!($109)) {
  $110 = ((($source)) + 140|0); //@line 2724 "libraries/lodepng/lodepng.cpp"
  $111 = HEAP32[$110>>2]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
  $112 = HEAP8[$111>>0]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
  HEAP8[$106>>0] = $112; //@line 2724 "libraries/lodepng/lodepng.cpp"
  $113 = ($105>>>0)>(1); //@line 2722 "libraries/lodepng/lodepng.cpp"
  if ($113) {
   $114 = ((($111)) + 1|0); //@line 2724 "libraries/lodepng/lodepng.cpp"
   $115 = HEAP8[$114>>0]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
   $116 = ((($106)) + 1|0); //@line 2724 "libraries/lodepng/lodepng.cpp"
   HEAP8[$116>>0] = $115; //@line 2724 "libraries/lodepng/lodepng.cpp"
   $117 = HEAP32[$102>>2]|0; //@line 2722 "libraries/lodepng/lodepng.cpp"
   $118 = ($117>>>0)>(2); //@line 2722 "libraries/lodepng/lodepng.cpp"
   if ($118) {
    $120 = 2;
    while(1) {
     $$pre7$i$pre = HEAP32[$74>>2]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
     $$pre49 = HEAP32[$110>>2]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
     $119 = (($$pre49) + ($120)|0); //@line 2724 "libraries/lodepng/lodepng.cpp"
     $121 = HEAP8[$119>>0]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
     $122 = (($$pre7$i$pre) + ($120)|0); //@line 2724 "libraries/lodepng/lodepng.cpp"
     HEAP8[$122>>0] = $121; //@line 2724 "libraries/lodepng/lodepng.cpp"
     $123 = (($120) + 1)|0; //@line 2722 "libraries/lodepng/lodepng.cpp"
     $124 = HEAP32[$102>>2]|0; //@line 2722 "libraries/lodepng/lodepng.cpp"
     $125 = ($123>>>0)<($124>>>0); //@line 2722 "libraries/lodepng/lodepng.cpp"
     if ($125) {
      $120 = $123;
     } else {
      break;
     }
    }
   }
  }
 }
 $126 = ((($source)) + 156|0); //@line 2719 "libraries/lodepng/lodepng.cpp"
 $127 = HEAP32[$126>>2]|0; //@line 2721 "libraries/lodepng/lodepng.cpp"
 $128 = ((($dest)) + 156|0); //@line 2719 "libraries/lodepng/lodepng.cpp"
 HEAP32[$128>>2] = $127; //@line 2719 "libraries/lodepng/lodepng.cpp"
 $129 = HEAP32[$126>>2]|0; //@line 2722 "libraries/lodepng/lodepng.cpp"
 $130 = (_malloc($129)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
 HEAP32[$76>>2] = $130; //@line 2720 "libraries/lodepng/lodepng.cpp"
 $131 = ($130|0)!=(0|0); //@line 2721 "libraries/lodepng/lodepng.cpp"
 $132 = ($127|0)==(0); //@line 2721 "libraries/lodepng/lodepng.cpp"
 $or$cond11$i = $132 | $131; //@line 2721 "libraries/lodepng/lodepng.cpp"
 if (!($or$cond11$i)) {
  $$2$i = 83;
  return ($$2$i|0);
 }
 $133 = ($129|0)==(0); //@line 2722 "libraries/lodepng/lodepng.cpp"
 if ($133) {
  $$2$i = 0;
  return ($$2$i|0);
 }
 $134 = ((($source)) + 144|0); //@line 2724 "libraries/lodepng/lodepng.cpp"
 $135 = HEAP32[$134>>2]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
 $136 = HEAP8[$135>>0]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
 HEAP8[$130>>0] = $136; //@line 2724 "libraries/lodepng/lodepng.cpp"
 $137 = ($129>>>0)>(1); //@line 2722 "libraries/lodepng/lodepng.cpp"
 if (!($137)) {
  $$2$i = 0;
  return ($$2$i|0);
 }
 $138 = ((($135)) + 1|0); //@line 2724 "libraries/lodepng/lodepng.cpp"
 $139 = HEAP8[$138>>0]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
 $140 = ((($130)) + 1|0); //@line 2724 "libraries/lodepng/lodepng.cpp"
 HEAP8[$140>>0] = $139; //@line 2724 "libraries/lodepng/lodepng.cpp"
 $141 = HEAP32[$126>>2]|0; //@line 2722 "libraries/lodepng/lodepng.cpp"
 $142 = ($141>>>0)>(2); //@line 2722 "libraries/lodepng/lodepng.cpp"
 if ($142) {
  $144 = 2;
 } else {
  $$2$i = 0;
  return ($$2$i|0);
 }
 while(1) {
  $$pre9$i$pre = HEAP32[$76>>2]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
  $$pre51 = HEAP32[$134>>2]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
  $143 = (($$pre51) + ($144)|0); //@line 2724 "libraries/lodepng/lodepng.cpp"
  $145 = HEAP8[$143>>0]|0; //@line 2724 "libraries/lodepng/lodepng.cpp"
  $146 = (($$pre9$i$pre) + ($144)|0); //@line 2724 "libraries/lodepng/lodepng.cpp"
  HEAP8[$146>>0] = $145; //@line 2724 "libraries/lodepng/lodepng.cpp"
  $147 = (($144) + 1)|0; //@line 2722 "libraries/lodepng/lodepng.cpp"
  $148 = HEAP32[$126>>2]|0; //@line 2722 "libraries/lodepng/lodepng.cpp"
  $149 = ($147>>>0)<($148>>>0); //@line 2722 "libraries/lodepng/lodepng.cpp"
  if ($149) {
   $144 = $147;
  } else {
   $$2$i = 0;
   break;
  }
 }
 return ($$2$i|0);
}
function __Z15lodepng_convertPhPKhP16LodePNGColorModePKS2_jj($out,$in,$mode_out,$mode_in,$w,$h) {
 $out = $out|0;
 $in = $in|0;
 $mode_out = $mode_out|0;
 $mode_in = $mode_in|0;
 $w = $w|0;
 $h = $h|0;
 var $$ = 0, $$0$i$i$i$i = 0, $$02$i = 0, $$025$i$i = 0, $$lcssa = 0, $$lcssa19 = 0, $$mask$i = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0;
 var $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0;
 var $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0;
 var $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0;
 var $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0;
 var $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0;
 var $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0;
 var $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0;
 var $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0;
 var $343 = 0, $344 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $a = 0, $a4 = 0, $b = 0, $b3 = 0, $bit$01$i = 0, $bit$06$i$i = 0, $exitcond$i = 0;
 var $g = 0, $g2 = 0, $i$01$i = 0, $i$06 = 0, $i$112 = 0, $i$27 = 0, $i$39 = 0, $mode_out$idx$val = 0, $or$cond = 0, $r = 0, $r1 = 0, $tree = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tree = sp;
 $r = sp + 74|0;
 $g = sp + 72|0;
 $b = sp + 70|0;
 $a = sp + 68|0;
 $r1 = sp + 79|0;
 $g2 = sp + 78|0;
 $b3 = sp + 77|0;
 $a4 = sp + 76|0;
 $0 = Math_imul($h, $w)|0; //@line 3421 "libraries/lodepng/lodepng.cpp"
 $1 = HEAP32[$mode_out>>2]|0; //@line 3455 "libraries/lodepng/lodepng.cpp"
 $2 = HEAP32[$mode_in>>2]|0; //@line 2634 "libraries/lodepng/lodepng.cpp"
 $3 = ($1|0)==($2|0); //@line 2586 "libraries/lodepng/lodepng.cpp"
 L1: do {
  if ($3) {
   $4 = ((($mode_out)) + 4|0); //@line 2587 "libraries/lodepng/lodepng.cpp"
   $5 = HEAP32[$4>>2]|0; //@line 2587 "libraries/lodepng/lodepng.cpp"
   $6 = ((($mode_in)) + 4|0); //@line 2587 "libraries/lodepng/lodepng.cpp"
   $7 = HEAP32[$6>>2]|0; //@line 2634 "libraries/lodepng/lodepng.cpp"
   $8 = ($5|0)==($7|0); //@line 2587 "libraries/lodepng/lodepng.cpp"
   if ($8) {
    $9 = ((($mode_out)) + 16|0); //@line 2588 "libraries/lodepng/lodepng.cpp"
    $10 = HEAP32[$9>>2]|0; //@line 2588 "libraries/lodepng/lodepng.cpp"
    $11 = ((($mode_in)) + 16|0); //@line 2588 "libraries/lodepng/lodepng.cpp"
    $12 = HEAP32[$11>>2]|0; //@line 2588 "libraries/lodepng/lodepng.cpp"
    $13 = ($10|0)==($12|0); //@line 2588 "libraries/lodepng/lodepng.cpp"
    if ($13) {
     $14 = ($10|0)==(0); //@line 2589 "libraries/lodepng/lodepng.cpp"
     if (!($14)) {
      $15 = ((($mode_out)) + 20|0); //@line 2591 "libraries/lodepng/lodepng.cpp"
      $16 = HEAP32[$15>>2]|0; //@line 2591 "libraries/lodepng/lodepng.cpp"
      $17 = ((($mode_in)) + 20|0); //@line 2591 "libraries/lodepng/lodepng.cpp"
      $18 = HEAP32[$17>>2]|0; //@line 2591 "libraries/lodepng/lodepng.cpp"
      $19 = ($16|0)==($18|0); //@line 2591 "libraries/lodepng/lodepng.cpp"
      if (!($19)) {
       break;
      }
      $20 = ((($mode_out)) + 24|0); //@line 2592 "libraries/lodepng/lodepng.cpp"
      $21 = HEAP32[$20>>2]|0; //@line 2592 "libraries/lodepng/lodepng.cpp"
      $22 = ((($mode_in)) + 24|0); //@line 2592 "libraries/lodepng/lodepng.cpp"
      $23 = HEAP32[$22>>2]|0; //@line 2592 "libraries/lodepng/lodepng.cpp"
      $24 = ($21|0)==($23|0); //@line 2592 "libraries/lodepng/lodepng.cpp"
      if (!($24)) {
       break;
      }
      $25 = ((($mode_out)) + 28|0); //@line 2593 "libraries/lodepng/lodepng.cpp"
      $26 = HEAP32[$25>>2]|0; //@line 2593 "libraries/lodepng/lodepng.cpp"
      $27 = ((($mode_in)) + 28|0); //@line 2593 "libraries/lodepng/lodepng.cpp"
      $28 = HEAP32[$27>>2]|0; //@line 2593 "libraries/lodepng/lodepng.cpp"
      $29 = ($26|0)==($28|0); //@line 2593 "libraries/lodepng/lodepng.cpp"
      if (!($29)) {
       break;
      }
     }
     $30 = ((($mode_out)) + 12|0); //@line 2595 "libraries/lodepng/lodepng.cpp"
     $31 = HEAP32[$30>>2]|0; //@line 2596 "libraries/lodepng/lodepng.cpp"
     $32 = ((($mode_in)) + 12|0); //@line 2595 "libraries/lodepng/lodepng.cpp"
     $33 = HEAP32[$32>>2]|0; //@line 2595 "libraries/lodepng/lodepng.cpp"
     $34 = ($31|0)==($33|0); //@line 2595 "libraries/lodepng/lodepng.cpp"
     if ($34) {
      $$mask$i = $31 & 1073741823; //@line 2596 "libraries/lodepng/lodepng.cpp"
      $35 = ($$mask$i|0)==(0); //@line 2596 "libraries/lodepng/lodepng.cpp"
      if (!($35)) {
       $36 = ((($mode_out)) + 8|0); //@line 2598 "libraries/lodepng/lodepng.cpp"
       $37 = HEAP32[$36>>2]|0; //@line 2598 "libraries/lodepng/lodepng.cpp"
       $38 = ((($mode_in)) + 8|0); //@line 2598 "libraries/lodepng/lodepng.cpp"
       $39 = HEAP32[$38>>2]|0; //@line 2598 "libraries/lodepng/lodepng.cpp"
       $40 = $31 << 2; //@line 2596 "libraries/lodepng/lodepng.cpp"
       $i$01$i = 0;
       while(1) {
        $43 = (($37) + ($i$01$i)|0); //@line 2598 "libraries/lodepng/lodepng.cpp"
        $44 = HEAP8[$43>>0]|0; //@line 2598 "libraries/lodepng/lodepng.cpp"
        $45 = (($39) + ($i$01$i)|0); //@line 2598 "libraries/lodepng/lodepng.cpp"
        $46 = HEAP8[$45>>0]|0; //@line 2598 "libraries/lodepng/lodepng.cpp"
        $47 = ($44<<24>>24)==($46<<24>>24); //@line 2598 "libraries/lodepng/lodepng.cpp"
        $42 = (($i$01$i) + 1)|0; //@line 2596 "libraries/lodepng/lodepng.cpp"
        if (!($47)) {
         break L1;
        }
        $41 = ($42|0)==($40|0); //@line 2596 "libraries/lodepng/lodepng.cpp"
        if ($41) {
         break;
        } else {
         $i$01$i = $42;
        }
       }
      }
      switch ($1|0) {
      case 3: case 0:  {
       $$0$i$i$i$i = 1;
       break;
      }
      case 2:  {
       $$0$i$i$i$i = 3;
       break;
      }
      case 4:  {
       $$0$i$i$i$i = 2;
       break;
      }
      case 6:  {
       $$0$i$i$i$i = 4;
       break;
      }
      default: {
       $$0$i$i$i$i = 0;
      }
      }
      $48 = Math_imul($5, $0)|0; //@line 2676 "libraries/lodepng/lodepng.cpp"
      $49 = Math_imul($48, $$0$i$i$i$i)|0; //@line 2676 "libraries/lodepng/lodepng.cpp"
      $50 = (($49) + 7)|0; //@line 2676 "libraries/lodepng/lodepng.cpp"
      $51 = $50 >>> 3; //@line 2676 "libraries/lodepng/lodepng.cpp"
      $52 = ($51|0)==(0); //@line 3426 "libraries/lodepng/lodepng.cpp"
      if ($52) {
       STACKTOP = sp;return 0; //@line 3475 "libraries/lodepng/lodepng.cpp"
      } else {
       $i$06 = 0;
      }
      while(1) {
       $53 = (($in) + ($i$06)|0); //@line 3426 "libraries/lodepng/lodepng.cpp"
       $54 = HEAP8[$53>>0]|0; //@line 3426 "libraries/lodepng/lodepng.cpp"
       $55 = (($out) + ($i$06)|0); //@line 3426 "libraries/lodepng/lodepng.cpp"
       HEAP8[$55>>0] = $54; //@line 3426 "libraries/lodepng/lodepng.cpp"
       $56 = (($i$06) + 1)|0; //@line 3426 "libraries/lodepng/lodepng.cpp"
       $57 = ($56|0)==($51|0); //@line 3426 "libraries/lodepng/lodepng.cpp"
       if ($57) {
        break;
       } else {
        $i$06 = $56;
       }
      }
      STACKTOP = sp;return 0; //@line 3475 "libraries/lodepng/lodepng.cpp"
     }
    }
   }
  }
 } while(0);
 $58 = ($1|0)==(3); //@line 3430 "libraries/lodepng/lodepng.cpp"
 if ($58) {
  $59 = ((($mode_out)) + 4|0); //@line 3432 "libraries/lodepng/lodepng.cpp"
  $60 = HEAP32[$59>>2]|0; //@line 3432 "libraries/lodepng/lodepng.cpp"
  $61 = 1 << $60; //@line 3432 "libraries/lodepng/lodepng.cpp"
  $62 = ((($mode_out)) + 12|0); //@line 3433 "libraries/lodepng/lodepng.cpp"
  $63 = HEAP32[$62>>2]|0; //@line 3433 "libraries/lodepng/lodepng.cpp"
  $64 = ($63>>>0)<($61>>>0); //@line 3433 "libraries/lodepng/lodepng.cpp"
  $$ = $64 ? $63 : $61; //@line 3433 "libraries/lodepng/lodepng.cpp"
  dest=$tree; stop=dest+64|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 2966 "libraries/lodepng/lodepng.cpp"
  $65 = ((($tree)) + 64|0); //@line 2967 "libraries/lodepng/lodepng.cpp"
  HEAP32[$65>>2] = -1; //@line 2967 "libraries/lodepng/lodepng.cpp"
  $66 = ($$|0)==(0); //@line 3435 "libraries/lodepng/lodepng.cpp"
  if (!($66)) {
   $67 = ((($mode_out)) + 8|0); //@line 3437 "libraries/lodepng/lodepng.cpp"
   $i$112 = 0;
   while(1) {
    $68 = $i$112 << 2; //@line 3437 "libraries/lodepng/lodepng.cpp"
    $69 = HEAP32[$67>>2]|0; //@line 3437 "libraries/lodepng/lodepng.cpp"
    $70 = (($69) + ($68)|0); //@line 3437 "libraries/lodepng/lodepng.cpp"
    $71 = HEAP8[$70>>0]|0; //@line 3438 "libraries/lodepng/lodepng.cpp"
    $72 = ((($70)) + 1|0); //@line 3438 "libraries/lodepng/lodepng.cpp"
    $73 = HEAP8[$72>>0]|0; //@line 3438 "libraries/lodepng/lodepng.cpp"
    $74 = ((($70)) + 2|0); //@line 3438 "libraries/lodepng/lodepng.cpp"
    $75 = HEAP8[$74>>0]|0; //@line 3438 "libraries/lodepng/lodepng.cpp"
    $76 = ((($70)) + 3|0); //@line 3438 "libraries/lodepng/lodepng.cpp"
    $77 = HEAP8[$76>>0]|0; //@line 3438 "libraries/lodepng/lodepng.cpp"
    $78 = $71&255; //@line 3011 "libraries/lodepng/lodepng.cpp"
    $79 = $73&255; //@line 3011 "libraries/lodepng/lodepng.cpp"
    $80 = $75&255; //@line 3011 "libraries/lodepng/lodepng.cpp"
    $81 = $77&255; //@line 3011 "libraries/lodepng/lodepng.cpp"
    $$02$i = $tree;$bit$01$i = 0;
    while(1) {
     $82 = $78 >>> $bit$01$i; //@line 3011 "libraries/lodepng/lodepng.cpp"
     $83 = $82 << 3; //@line 3011 "libraries/lodepng/lodepng.cpp"
     $84 = $83 & 8; //@line 3011 "libraries/lodepng/lodepng.cpp"
     $85 = $79 >>> $bit$01$i; //@line 3011 "libraries/lodepng/lodepng.cpp"
     $86 = $85 << 2; //@line 3011 "libraries/lodepng/lodepng.cpp"
     $87 = $86 & 4; //@line 3011 "libraries/lodepng/lodepng.cpp"
     $88 = $80 >>> $bit$01$i; //@line 3011 "libraries/lodepng/lodepng.cpp"
     $89 = $88 << 1; //@line 3011 "libraries/lodepng/lodepng.cpp"
     $90 = $89 & 2; //@line 3011 "libraries/lodepng/lodepng.cpp"
     $91 = $81 >>> $bit$01$i; //@line 3011 "libraries/lodepng/lodepng.cpp"
     $92 = $91 & 1; //@line 3011 "libraries/lodepng/lodepng.cpp"
     $93 = $87 | $92; //@line 3011 "libraries/lodepng/lodepng.cpp"
     $94 = $93 | $84; //@line 3011 "libraries/lodepng/lodepng.cpp"
     $95 = $94 | $90; //@line 3011 "libraries/lodepng/lodepng.cpp"
     $96 = (($$02$i) + ($95<<2)|0); //@line 3012 "libraries/lodepng/lodepng.cpp"
     $97 = HEAP32[$96>>2]|0; //@line 3012 "libraries/lodepng/lodepng.cpp"
     $98 = ($97|0)==(0|0); //@line 3012 "libraries/lodepng/lodepng.cpp"
     if ($98) {
      $99 = (_malloc(68)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
      HEAP32[$96>>2] = $99; //@line 3014 "libraries/lodepng/lodepng.cpp"
      dest=$99; stop=dest+64|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 2966 "libraries/lodepng/lodepng.cpp"
      $100 = ((($99)) + 64|0); //@line 2967 "libraries/lodepng/lodepng.cpp"
      HEAP32[$100>>2] = -1; //@line 2967 "libraries/lodepng/lodepng.cpp"
      $$pre$i = HEAP32[$96>>2]|0; //@line 3017 "libraries/lodepng/lodepng.cpp"
      $344 = $$pre$i;
     } else {
      $344 = $97;
     }
     $101 = (($bit$01$i) + 1)|0; //@line 3009 "libraries/lodepng/lodepng.cpp"
     $exitcond$i = ($101|0)==(8); //@line 3009 "libraries/lodepng/lodepng.cpp"
     if ($exitcond$i) {
      $$lcssa = $344;
      break;
     } else {
      $$02$i = $344;$bit$01$i = $101;
     }
    }
    $102 = ((($$lcssa)) + 64|0); //@line 3019 "libraries/lodepng/lodepng.cpp"
    HEAP32[$102>>2] = $i$112; //@line 3019 "libraries/lodepng/lodepng.cpp"
    $103 = (($i$112) + 1)|0; //@line 3435 "libraries/lodepng/lodepng.cpp"
    $104 = ($103|0)==($$|0); //@line 3435 "libraries/lodepng/lodepng.cpp"
    if ($104) {
     break;
    } else {
     $i$112 = $103;
    }
   }
  }
 }
 $105 = ((($mode_in)) + 4|0); //@line 3442 "libraries/lodepng/lodepng.cpp"
 $106 = HEAP32[$105>>2]|0; //@line 3442 "libraries/lodepng/lodepng.cpp"
 $107 = ($106|0)==(16); //@line 3442 "libraries/lodepng/lodepng.cpp"
 $108 = ((($mode_out)) + 4|0); //@line 3442 "libraries/lodepng/lodepng.cpp"
 $109 = HEAP32[$108>>2]|0; //@line 3442 "libraries/lodepng/lodepng.cpp"
 $110 = ($109|0)==(16); //@line 3442 "libraries/lodepng/lodepng.cpp"
 $or$cond = $107 & $110; //@line 3442 "libraries/lodepng/lodepng.cpp"
 L41: do {
  if ($or$cond) {
   $111 = ($0|0)==(0); //@line 3444 "libraries/lodepng/lodepng.cpp"
   if (!($111)) {
    $i$27 = 0;
    while(1) {
     HEAP16[$r>>1] = 0; //@line 3446 "libraries/lodepng/lodepng.cpp"
     HEAP16[$g>>1] = 0; //@line 3446 "libraries/lodepng/lodepng.cpp"
     HEAP16[$b>>1] = 0; //@line 3446 "libraries/lodepng/lodepng.cpp"
     HEAP16[$a>>1] = 0; //@line 3446 "libraries/lodepng/lodepng.cpp"
     __ZL19getPixelColorRGBA16PtS_S_S_PKhjPK16LodePNGColorMode($r,$g,$b,$a,$in,$i$27,$mode_in); //@line 3447 "libraries/lodepng/lodepng.cpp"
     $112 = HEAP16[$r>>1]|0; //@line 3448 "libraries/lodepng/lodepng.cpp"
     $113 = HEAP16[$g>>1]|0; //@line 3448 "libraries/lodepng/lodepng.cpp"
     $114 = HEAP16[$b>>1]|0; //@line 3448 "libraries/lodepng/lodepng.cpp"
     $115 = HEAP16[$a>>1]|0; //@line 3448 "libraries/lodepng/lodepng.cpp"
     $mode_out$idx$val = HEAP32[$mode_out>>2]|0;
     switch ($mode_out$idx$val|0) {
     case 0:  {
      $116 = ($112&65535) >>> 8; //@line 3104 "libraries/lodepng/lodepng.cpp"
      $117 = $116&255; //@line 3104 "libraries/lodepng/lodepng.cpp"
      $118 = $i$27 << 1; //@line 3104 "libraries/lodepng/lodepng.cpp"
      $119 = (($out) + ($118)|0); //@line 3104 "libraries/lodepng/lodepng.cpp"
      HEAP8[$119>>0] = $117; //@line 3104 "libraries/lodepng/lodepng.cpp"
      $120 = $112&255; //@line 3105 "libraries/lodepng/lodepng.cpp"
      $121 = $118 | 1; //@line 3105 "libraries/lodepng/lodepng.cpp"
      $122 = (($out) + ($121)|0); //@line 3105 "libraries/lodepng/lodepng.cpp"
      HEAP8[$122>>0] = $120; //@line 3105 "libraries/lodepng/lodepng.cpp"
      break;
     }
     case 2:  {
      $123 = ($112&65535) >>> 8; //@line 3109 "libraries/lodepng/lodepng.cpp"
      $124 = $123&255; //@line 3109 "libraries/lodepng/lodepng.cpp"
      $125 = ($i$27*6)|0; //@line 3109 "libraries/lodepng/lodepng.cpp"
      $126 = (($out) + ($125)|0); //@line 3109 "libraries/lodepng/lodepng.cpp"
      HEAP8[$126>>0] = $124; //@line 3109 "libraries/lodepng/lodepng.cpp"
      $127 = $112&255; //@line 3110 "libraries/lodepng/lodepng.cpp"
      $128 = $125 | 1; //@line 3110 "libraries/lodepng/lodepng.cpp"
      $129 = (($out) + ($128)|0); //@line 3110 "libraries/lodepng/lodepng.cpp"
      HEAP8[$129>>0] = $127; //@line 3110 "libraries/lodepng/lodepng.cpp"
      $130 = ($113&65535) >>> 8; //@line 3111 "libraries/lodepng/lodepng.cpp"
      $131 = $130&255; //@line 3111 "libraries/lodepng/lodepng.cpp"
      $132 = (($125) + 2)|0; //@line 3111 "libraries/lodepng/lodepng.cpp"
      $133 = (($out) + ($132)|0); //@line 3111 "libraries/lodepng/lodepng.cpp"
      HEAP8[$133>>0] = $131; //@line 3111 "libraries/lodepng/lodepng.cpp"
      $134 = $113&255; //@line 3112 "libraries/lodepng/lodepng.cpp"
      $135 = (($125) + 3)|0; //@line 3112 "libraries/lodepng/lodepng.cpp"
      $136 = (($out) + ($135)|0); //@line 3112 "libraries/lodepng/lodepng.cpp"
      HEAP8[$136>>0] = $134; //@line 3112 "libraries/lodepng/lodepng.cpp"
      $137 = ($114&65535) >>> 8; //@line 3113 "libraries/lodepng/lodepng.cpp"
      $138 = $137&255; //@line 3113 "libraries/lodepng/lodepng.cpp"
      $139 = (($125) + 4)|0; //@line 3113 "libraries/lodepng/lodepng.cpp"
      $140 = (($out) + ($139)|0); //@line 3113 "libraries/lodepng/lodepng.cpp"
      HEAP8[$140>>0] = $138; //@line 3113 "libraries/lodepng/lodepng.cpp"
      $141 = $114&255; //@line 3114 "libraries/lodepng/lodepng.cpp"
      $142 = (($125) + 5)|0; //@line 3114 "libraries/lodepng/lodepng.cpp"
      $143 = (($out) + ($142)|0); //@line 3114 "libraries/lodepng/lodepng.cpp"
      HEAP8[$143>>0] = $141; //@line 3114 "libraries/lodepng/lodepng.cpp"
      break;
     }
     case 4:  {
      $144 = ($112&65535) >>> 8; //@line 3119 "libraries/lodepng/lodepng.cpp"
      $145 = $144&255; //@line 3119 "libraries/lodepng/lodepng.cpp"
      $146 = $i$27 << 2; //@line 3119 "libraries/lodepng/lodepng.cpp"
      $147 = (($out) + ($146)|0); //@line 3119 "libraries/lodepng/lodepng.cpp"
      HEAP8[$147>>0] = $145; //@line 3119 "libraries/lodepng/lodepng.cpp"
      $148 = $112&255; //@line 3120 "libraries/lodepng/lodepng.cpp"
      $149 = $146 | 1; //@line 3120 "libraries/lodepng/lodepng.cpp"
      $150 = (($out) + ($149)|0); //@line 3120 "libraries/lodepng/lodepng.cpp"
      HEAP8[$150>>0] = $148; //@line 3120 "libraries/lodepng/lodepng.cpp"
      $151 = ($115&65535) >>> 8; //@line 3121 "libraries/lodepng/lodepng.cpp"
      $152 = $151&255; //@line 3121 "libraries/lodepng/lodepng.cpp"
      $153 = $146 | 2; //@line 3121 "libraries/lodepng/lodepng.cpp"
      $154 = (($out) + ($153)|0); //@line 3121 "libraries/lodepng/lodepng.cpp"
      HEAP8[$154>>0] = $152; //@line 3121 "libraries/lodepng/lodepng.cpp"
      $155 = $115&255; //@line 3122 "libraries/lodepng/lodepng.cpp"
      $156 = $146 | 3; //@line 3122 "libraries/lodepng/lodepng.cpp"
      $157 = (($out) + ($156)|0); //@line 3122 "libraries/lodepng/lodepng.cpp"
      HEAP8[$157>>0] = $155; //@line 3122 "libraries/lodepng/lodepng.cpp"
      break;
     }
     case 6:  {
      $158 = ($112&65535) >>> 8; //@line 3126 "libraries/lodepng/lodepng.cpp"
      $159 = $158&255; //@line 3126 "libraries/lodepng/lodepng.cpp"
      $160 = $i$27 << 3; //@line 3126 "libraries/lodepng/lodepng.cpp"
      $161 = (($out) + ($160)|0); //@line 3126 "libraries/lodepng/lodepng.cpp"
      HEAP8[$161>>0] = $159; //@line 3126 "libraries/lodepng/lodepng.cpp"
      $162 = $112&255; //@line 3127 "libraries/lodepng/lodepng.cpp"
      $163 = $160 | 1; //@line 3127 "libraries/lodepng/lodepng.cpp"
      $164 = (($out) + ($163)|0); //@line 3127 "libraries/lodepng/lodepng.cpp"
      HEAP8[$164>>0] = $162; //@line 3127 "libraries/lodepng/lodepng.cpp"
      $165 = ($113&65535) >>> 8; //@line 3128 "libraries/lodepng/lodepng.cpp"
      $166 = $165&255; //@line 3128 "libraries/lodepng/lodepng.cpp"
      $167 = $160 | 2; //@line 3128 "libraries/lodepng/lodepng.cpp"
      $168 = (($out) + ($167)|0); //@line 3128 "libraries/lodepng/lodepng.cpp"
      HEAP8[$168>>0] = $166; //@line 3128 "libraries/lodepng/lodepng.cpp"
      $169 = $113&255; //@line 3129 "libraries/lodepng/lodepng.cpp"
      $170 = $160 | 3; //@line 3129 "libraries/lodepng/lodepng.cpp"
      $171 = (($out) + ($170)|0); //@line 3129 "libraries/lodepng/lodepng.cpp"
      HEAP8[$171>>0] = $169; //@line 3129 "libraries/lodepng/lodepng.cpp"
      $172 = ($114&65535) >>> 8; //@line 3130 "libraries/lodepng/lodepng.cpp"
      $173 = $172&255; //@line 3130 "libraries/lodepng/lodepng.cpp"
      $174 = $160 | 4; //@line 3130 "libraries/lodepng/lodepng.cpp"
      $175 = (($out) + ($174)|0); //@line 3130 "libraries/lodepng/lodepng.cpp"
      HEAP8[$175>>0] = $173; //@line 3130 "libraries/lodepng/lodepng.cpp"
      $176 = $114&255; //@line 3131 "libraries/lodepng/lodepng.cpp"
      $177 = $160 | 5; //@line 3131 "libraries/lodepng/lodepng.cpp"
      $178 = (($out) + ($177)|0); //@line 3131 "libraries/lodepng/lodepng.cpp"
      HEAP8[$178>>0] = $176; //@line 3131 "libraries/lodepng/lodepng.cpp"
      $179 = ($115&65535) >>> 8; //@line 3132 "libraries/lodepng/lodepng.cpp"
      $180 = $179&255; //@line 3132 "libraries/lodepng/lodepng.cpp"
      $181 = $160 | 6; //@line 3132 "libraries/lodepng/lodepng.cpp"
      $182 = (($out) + ($181)|0); //@line 3132 "libraries/lodepng/lodepng.cpp"
      HEAP8[$182>>0] = $180; //@line 3132 "libraries/lodepng/lodepng.cpp"
      $183 = $115&255; //@line 3133 "libraries/lodepng/lodepng.cpp"
      $184 = $160 | 7; //@line 3133 "libraries/lodepng/lodepng.cpp"
      $185 = (($out) + ($184)|0); //@line 3133 "libraries/lodepng/lodepng.cpp"
      HEAP8[$185>>0] = $183; //@line 3133 "libraries/lodepng/lodepng.cpp"
      break;
     }
     default: {
     }
     }
     $186 = (($i$27) + 1)|0; //@line 3444 "libraries/lodepng/lodepng.cpp"
     $187 = ($186|0)==($0|0); //@line 3444 "libraries/lodepng/lodepng.cpp"
     if ($187) {
      break L41;
     } else {
      $i$27 = $186;
     }
    }
   }
  } else {
   $188 = ($109|0)==(8); //@line 3451 "libraries/lodepng/lodepng.cpp"
   L53: do {
    if ($188) {
     switch ($1|0) {
     case 6:  {
      __ZL19getPixelColorsRGBA8PhjjPKhPK16LodePNGColorMode($out,$0,1,$in,$mode_in); //@line 3453 "libraries/lodepng/lodepng.cpp"
      break L41;
      break;
     }
     case 2:  {
      __ZL19getPixelColorsRGBA8PhjjPKhPK16LodePNGColorMode($out,$0,0,$in,$mode_in); //@line 3457 "libraries/lodepng/lodepng.cpp"
      break L41;
      break;
     }
     default: {
      break L53;
     }
     }
    }
   } while(0);
   HEAP8[$r1>>0] = 0; //@line 3461 "libraries/lodepng/lodepng.cpp"
   HEAP8[$g2>>0] = 0; //@line 3461 "libraries/lodepng/lodepng.cpp"
   HEAP8[$b3>>0] = 0; //@line 3461 "libraries/lodepng/lodepng.cpp"
   HEAP8[$a4>>0] = 0; //@line 3461 "libraries/lodepng/lodepng.cpp"
   $189 = ($0|0)==(0); //@line 3462 "libraries/lodepng/lodepng.cpp"
   L59: do {
    if (!($189)) {
     $i$39 = 0;
     while(1) {
      __ZL18getPixelColorRGBA8PhS_S_S_PKhjPK16LodePNGColorMode($r1,$g2,$b3,$a4,$in,$i$39,$mode_in); //@line 3464 "libraries/lodepng/lodepng.cpp"
      $190 = HEAP8[$r1>>0]|0; //@line 3465 "libraries/lodepng/lodepng.cpp"
      $191 = HEAP8[$g2>>0]|0; //@line 3465 "libraries/lodepng/lodepng.cpp"
      $192 = HEAP8[$b3>>0]|0; //@line 3465 "libraries/lodepng/lodepng.cpp"
      $193 = HEAP8[$a4>>0]|0; //@line 3465 "libraries/lodepng/lodepng.cpp"
      $194 = HEAP32[$mode_out>>2]|0; //@line 3027 "libraries/lodepng/lodepng.cpp"
      L62: do {
       switch ($194|0) {
       case 0:  {
        $195 = HEAP32[$108>>2]|0; //@line 3030 "libraries/lodepng/lodepng.cpp"
        switch ($195|0) {
        case 8:  {
         $196 = (($out) + ($i$39)|0); //@line 3030 "libraries/lodepng/lodepng.cpp"
         HEAP8[$196>>0] = $190; //@line 3030 "libraries/lodepng/lodepng.cpp"
         break L62;
         break;
        }
        case 16:  {
         $197 = $i$39 << 1; //@line 3031 "libraries/lodepng/lodepng.cpp"
         $198 = $197 | 1; //@line 3031 "libraries/lodepng/lodepng.cpp"
         $199 = (($out) + ($198)|0); //@line 3031 "libraries/lodepng/lodepng.cpp"
         HEAP8[$199>>0] = $190; //@line 3031 "libraries/lodepng/lodepng.cpp"
         $200 = (($out) + ($197)|0); //@line 3031 "libraries/lodepng/lodepng.cpp"
         HEAP8[$200>>0] = $190; //@line 3031 "libraries/lodepng/lodepng.cpp"
         break L62;
         break;
        }
        default: {
         $201 = $190&255; //@line 3035 "libraries/lodepng/lodepng.cpp"
         $202 = (8 - ($195))|0; //@line 3035 "libraries/lodepng/lodepng.cpp"
         $203 = $201 >>> $202; //@line 3035 "libraries/lodepng/lodepng.cpp"
         $204 = 1 << $195; //@line 3035 "libraries/lodepng/lodepng.cpp"
         $205 = (($204) + 255)|0; //@line 3035 "libraries/lodepng/lodepng.cpp"
         $206 = ($195|0)==(1); //@line 2940 "libraries/lodepng/lodepng.cpp"
         $207 = ($195|0)==(2); //@line 2940 "libraries/lodepng/lodepng.cpp"
         $208 = $207 ? 3 : 1; //@line 2940 "libraries/lodepng/lodepng.cpp"
         $209 = $206 ? 7 : $208; //@line 2940 "libraries/lodepng/lodepng.cpp"
         $210 = $209 & $i$39; //@line 2942 "libraries/lodepng/lodepng.cpp"
         $211 = $205 & 255; //@line 3035 "libraries/lodepng/lodepng.cpp"
         $212 = $211 & $203; //@line 3036 "libraries/lodepng/lodepng.cpp"
         $213 = (($209) - ($210))|0; //@line 2944 "libraries/lodepng/lodepng.cpp"
         $214 = Math_imul($213, $195)|0; //@line 2944 "libraries/lodepng/lodepng.cpp"
         $215 = $212 << $214; //@line 2944 "libraries/lodepng/lodepng.cpp"
         $216 = ($210|0)==(0); //@line 2945 "libraries/lodepng/lodepng.cpp"
         if ($216) {
          $217 = $215&255; //@line 2945 "libraries/lodepng/lodepng.cpp"
          $218 = Math_imul($195, $i$39)|0; //@line 2945 "libraries/lodepng/lodepng.cpp"
          $219 = $218 >>> 3; //@line 2945 "libraries/lodepng/lodepng.cpp"
          $220 = (($out) + ($219)|0); //@line 2945 "libraries/lodepng/lodepng.cpp"
          HEAP8[$220>>0] = $217; //@line 2945 "libraries/lodepng/lodepng.cpp"
          break L62;
         } else {
          $221 = Math_imul($195, $i$39)|0; //@line 2946 "libraries/lodepng/lodepng.cpp"
          $222 = $221 >>> 3; //@line 2946 "libraries/lodepng/lodepng.cpp"
          $223 = (($out) + ($222)|0); //@line 2946 "libraries/lodepng/lodepng.cpp"
          $224 = HEAP8[$223>>0]|0; //@line 2946 "libraries/lodepng/lodepng.cpp"
          $225 = $224&255; //@line 2946 "libraries/lodepng/lodepng.cpp"
          $226 = $225 | $215; //@line 2946 "libraries/lodepng/lodepng.cpp"
          $227 = $226&255; //@line 2946 "libraries/lodepng/lodepng.cpp"
          HEAP8[$223>>0] = $227; //@line 2946 "libraries/lodepng/lodepng.cpp"
          break L62;
         }
        }
        }
        break;
       }
       case 2:  {
        $228 = HEAP32[$108>>2]|0; //@line 3041 "libraries/lodepng/lodepng.cpp"
        $229 = ($228|0)==(8); //@line 3041 "libraries/lodepng/lodepng.cpp"
        if ($229) {
         $230 = ($i$39*3)|0; //@line 3043 "libraries/lodepng/lodepng.cpp"
         $231 = (($out) + ($230)|0); //@line 3043 "libraries/lodepng/lodepng.cpp"
         HEAP8[$231>>0] = $190; //@line 3043 "libraries/lodepng/lodepng.cpp"
         $232 = (($230) + 1)|0; //@line 3044 "libraries/lodepng/lodepng.cpp"
         $233 = (($out) + ($232)|0); //@line 3044 "libraries/lodepng/lodepng.cpp"
         HEAP8[$233>>0] = $191; //@line 3044 "libraries/lodepng/lodepng.cpp"
         $234 = (($230) + 2)|0; //@line 3045 "libraries/lodepng/lodepng.cpp"
         $235 = (($out) + ($234)|0); //@line 3045 "libraries/lodepng/lodepng.cpp"
         HEAP8[$235>>0] = $192; //@line 3045 "libraries/lodepng/lodepng.cpp"
         break L62;
        } else {
         $236 = ($i$39*6)|0; //@line 3049 "libraries/lodepng/lodepng.cpp"
         $237 = $236 | 1; //@line 3049 "libraries/lodepng/lodepng.cpp"
         $238 = (($out) + ($237)|0); //@line 3049 "libraries/lodepng/lodepng.cpp"
         HEAP8[$238>>0] = $190; //@line 3049 "libraries/lodepng/lodepng.cpp"
         $239 = (($out) + ($236)|0); //@line 3049 "libraries/lodepng/lodepng.cpp"
         HEAP8[$239>>0] = $190; //@line 3049 "libraries/lodepng/lodepng.cpp"
         $240 = (($236) + 3)|0; //@line 3050 "libraries/lodepng/lodepng.cpp"
         $241 = (($out) + ($240)|0); //@line 3050 "libraries/lodepng/lodepng.cpp"
         HEAP8[$241>>0] = $191; //@line 3050 "libraries/lodepng/lodepng.cpp"
         $242 = (($236) + 2)|0; //@line 3050 "libraries/lodepng/lodepng.cpp"
         $243 = (($out) + ($242)|0); //@line 3050 "libraries/lodepng/lodepng.cpp"
         HEAP8[$243>>0] = $191; //@line 3050 "libraries/lodepng/lodepng.cpp"
         $244 = (($236) + 5)|0; //@line 3051 "libraries/lodepng/lodepng.cpp"
         $245 = (($out) + ($244)|0); //@line 3051 "libraries/lodepng/lodepng.cpp"
         HEAP8[$245>>0] = $192; //@line 3051 "libraries/lodepng/lodepng.cpp"
         $246 = (($236) + 4)|0; //@line 3051 "libraries/lodepng/lodepng.cpp"
         $247 = (($out) + ($246)|0); //@line 3051 "libraries/lodepng/lodepng.cpp"
         HEAP8[$247>>0] = $192; //@line 3051 "libraries/lodepng/lodepng.cpp"
         break L62;
        }
        break;
       }
       case 3:  {
        $248 = $190&255; //@line 2989 "libraries/lodepng/lodepng.cpp"
        $249 = $191&255; //@line 2989 "libraries/lodepng/lodepng.cpp"
        $250 = $192&255; //@line 2989 "libraries/lodepng/lodepng.cpp"
        $251 = $193&255; //@line 2989 "libraries/lodepng/lodepng.cpp"
        $$025$i$i = $tree;$bit$06$i$i = 0;
        while(1) {
         $252 = $248 >>> $bit$06$i$i; //@line 2989 "libraries/lodepng/lodepng.cpp"
         $253 = $252 << 3; //@line 2989 "libraries/lodepng/lodepng.cpp"
         $254 = $253 & 8; //@line 2989 "libraries/lodepng/lodepng.cpp"
         $255 = $249 >>> $bit$06$i$i; //@line 2989 "libraries/lodepng/lodepng.cpp"
         $256 = $255 << 2; //@line 2989 "libraries/lodepng/lodepng.cpp"
         $257 = $256 & 4; //@line 2989 "libraries/lodepng/lodepng.cpp"
         $258 = $250 >>> $bit$06$i$i; //@line 2989 "libraries/lodepng/lodepng.cpp"
         $259 = $258 << 1; //@line 2989 "libraries/lodepng/lodepng.cpp"
         $260 = $259 & 2; //@line 2989 "libraries/lodepng/lodepng.cpp"
         $261 = $251 >>> $bit$06$i$i; //@line 2989 "libraries/lodepng/lodepng.cpp"
         $262 = $261 & 1; //@line 2989 "libraries/lodepng/lodepng.cpp"
         $263 = $257 | $262; //@line 2989 "libraries/lodepng/lodepng.cpp"
         $264 = $263 | $254; //@line 2989 "libraries/lodepng/lodepng.cpp"
         $265 = $264 | $260; //@line 2989 "libraries/lodepng/lodepng.cpp"
         $266 = (($$025$i$i) + ($265<<2)|0); //@line 2990 "libraries/lodepng/lodepng.cpp"
         $267 = HEAP32[$266>>2]|0; //@line 2990 "libraries/lodepng/lodepng.cpp"
         $268 = ($267|0)==(0|0); //@line 2990 "libraries/lodepng/lodepng.cpp"
         if ($268) {
          break L62;
         }
         $269 = (($bit$06$i$i) + 1)|0; //@line 2987 "libraries/lodepng/lodepng.cpp"
         $270 = ($269|0)<(8); //@line 2987 "libraries/lodepng/lodepng.cpp"
         if ($270) {
          $$025$i$i = $267;$bit$06$i$i = $269;
         } else {
          $$lcssa19 = $267;
          break;
         }
        }
        $271 = ((($$lcssa19)) + 64|0); //@line 2993 "libraries/lodepng/lodepng.cpp"
        $272 = HEAP32[$271>>2]|0; //@line 2993 "libraries/lodepng/lodepng.cpp"
        $273 = ($272|0)<(0); //@line 3057 "libraries/lodepng/lodepng.cpp"
        if (!($273)) {
         $274 = HEAP32[$108>>2]|0; //@line 3058 "libraries/lodepng/lodepng.cpp"
         $275 = ($274|0)==(8); //@line 3058 "libraries/lodepng/lodepng.cpp"
         if ($275) {
          $276 = $272&255; //@line 3058 "libraries/lodepng/lodepng.cpp"
          $277 = (($out) + ($i$39)|0); //@line 3058 "libraries/lodepng/lodepng.cpp"
          HEAP8[$277>>0] = $276; //@line 3058 "libraries/lodepng/lodepng.cpp"
          break L62;
         }
         $278 = ($274|0)==(1); //@line 2940 "libraries/lodepng/lodepng.cpp"
         $279 = ($274|0)==(2); //@line 2940 "libraries/lodepng/lodepng.cpp"
         $280 = $279 ? 3 : 1; //@line 2940 "libraries/lodepng/lodepng.cpp"
         $281 = $278 ? 7 : $280; //@line 2940 "libraries/lodepng/lodepng.cpp"
         $282 = $281 & $i$39; //@line 2942 "libraries/lodepng/lodepng.cpp"
         $283 = 1 << $274; //@line 2943 "libraries/lodepng/lodepng.cpp"
         $284 = (($283) + -1)|0; //@line 2943 "libraries/lodepng/lodepng.cpp"
         $285 = $284 & $272; //@line 2943 "libraries/lodepng/lodepng.cpp"
         $286 = (($281) - ($282))|0; //@line 2944 "libraries/lodepng/lodepng.cpp"
         $287 = Math_imul($286, $274)|0; //@line 2944 "libraries/lodepng/lodepng.cpp"
         $288 = $285 << $287; //@line 2944 "libraries/lodepng/lodepng.cpp"
         $289 = ($282|0)==(0); //@line 2945 "libraries/lodepng/lodepng.cpp"
         if ($289) {
          $290 = $288&255; //@line 2945 "libraries/lodepng/lodepng.cpp"
          $291 = Math_imul($274, $i$39)|0; //@line 2945 "libraries/lodepng/lodepng.cpp"
          $292 = $291 >>> 3; //@line 2945 "libraries/lodepng/lodepng.cpp"
          $293 = (($out) + ($292)|0); //@line 2945 "libraries/lodepng/lodepng.cpp"
          HEAP8[$293>>0] = $290; //@line 2945 "libraries/lodepng/lodepng.cpp"
          break L62;
         } else {
          $294 = Math_imul($274, $i$39)|0; //@line 2946 "libraries/lodepng/lodepng.cpp"
          $295 = $294 >>> 3; //@line 2946 "libraries/lodepng/lodepng.cpp"
          $296 = (($out) + ($295)|0); //@line 2946 "libraries/lodepng/lodepng.cpp"
          $297 = HEAP8[$296>>0]|0; //@line 2946 "libraries/lodepng/lodepng.cpp"
          $298 = $297&255; //@line 2946 "libraries/lodepng/lodepng.cpp"
          $299 = $298 | $288; //@line 2946 "libraries/lodepng/lodepng.cpp"
          $300 = $299&255; //@line 2946 "libraries/lodepng/lodepng.cpp"
          HEAP8[$296>>0] = $300; //@line 2946 "libraries/lodepng/lodepng.cpp"
          break L62;
         }
        }
        break;
       }
       case 4:  {
        $301 = HEAP32[$108>>2]|0; //@line 3064 "libraries/lodepng/lodepng.cpp"
        switch ($301|0) {
        case 8:  {
         $302 = $i$39 << 1; //@line 3066 "libraries/lodepng/lodepng.cpp"
         $303 = (($out) + ($302)|0); //@line 3066 "libraries/lodepng/lodepng.cpp"
         HEAP8[$303>>0] = $190; //@line 3066 "libraries/lodepng/lodepng.cpp"
         $304 = $302 | 1; //@line 3067 "libraries/lodepng/lodepng.cpp"
         $305 = (($out) + ($304)|0); //@line 3067 "libraries/lodepng/lodepng.cpp"
         HEAP8[$305>>0] = $193; //@line 3067 "libraries/lodepng/lodepng.cpp"
         break L62;
         break;
        }
        case 16:  {
         $306 = $i$39 << 2; //@line 3071 "libraries/lodepng/lodepng.cpp"
         $307 = $306 | 1; //@line 3071 "libraries/lodepng/lodepng.cpp"
         $308 = (($out) + ($307)|0); //@line 3071 "libraries/lodepng/lodepng.cpp"
         HEAP8[$308>>0] = $190; //@line 3071 "libraries/lodepng/lodepng.cpp"
         $309 = (($out) + ($306)|0); //@line 3071 "libraries/lodepng/lodepng.cpp"
         HEAP8[$309>>0] = $190; //@line 3071 "libraries/lodepng/lodepng.cpp"
         $310 = $306 | 3; //@line 3072 "libraries/lodepng/lodepng.cpp"
         $311 = (($out) + ($310)|0); //@line 3072 "libraries/lodepng/lodepng.cpp"
         HEAP8[$311>>0] = $193; //@line 3072 "libraries/lodepng/lodepng.cpp"
         $312 = $306 | 2; //@line 3072 "libraries/lodepng/lodepng.cpp"
         $313 = (($out) + ($312)|0); //@line 3072 "libraries/lodepng/lodepng.cpp"
         HEAP8[$313>>0] = $193; //@line 3072 "libraries/lodepng/lodepng.cpp"
         break L62;
         break;
        }
        default: {
         break L62;
        }
        }
        break;
       }
       case 6:  {
        $314 = HEAP32[$108>>2]|0; //@line 3077 "libraries/lodepng/lodepng.cpp"
        $315 = ($314|0)==(8); //@line 3077 "libraries/lodepng/lodepng.cpp"
        if ($315) {
         $316 = $i$39 << 2; //@line 3079 "libraries/lodepng/lodepng.cpp"
         $317 = (($out) + ($316)|0); //@line 3079 "libraries/lodepng/lodepng.cpp"
         HEAP8[$317>>0] = $190; //@line 3079 "libraries/lodepng/lodepng.cpp"
         $318 = $316 | 1; //@line 3080 "libraries/lodepng/lodepng.cpp"
         $319 = (($out) + ($318)|0); //@line 3080 "libraries/lodepng/lodepng.cpp"
         HEAP8[$319>>0] = $191; //@line 3080 "libraries/lodepng/lodepng.cpp"
         $320 = $316 | 2; //@line 3081 "libraries/lodepng/lodepng.cpp"
         $321 = (($out) + ($320)|0); //@line 3081 "libraries/lodepng/lodepng.cpp"
         HEAP8[$321>>0] = $192; //@line 3081 "libraries/lodepng/lodepng.cpp"
         $322 = $316 | 3; //@line 3082 "libraries/lodepng/lodepng.cpp"
         $323 = (($out) + ($322)|0); //@line 3082 "libraries/lodepng/lodepng.cpp"
         HEAP8[$323>>0] = $193; //@line 3082 "libraries/lodepng/lodepng.cpp"
         break L62;
        } else {
         $324 = $i$39 << 3; //@line 3086 "libraries/lodepng/lodepng.cpp"
         $325 = $324 | 1; //@line 3086 "libraries/lodepng/lodepng.cpp"
         $326 = (($out) + ($325)|0); //@line 3086 "libraries/lodepng/lodepng.cpp"
         HEAP8[$326>>0] = $190; //@line 3086 "libraries/lodepng/lodepng.cpp"
         $327 = (($out) + ($324)|0); //@line 3086 "libraries/lodepng/lodepng.cpp"
         HEAP8[$327>>0] = $190; //@line 3086 "libraries/lodepng/lodepng.cpp"
         $328 = $324 | 3; //@line 3087 "libraries/lodepng/lodepng.cpp"
         $329 = (($out) + ($328)|0); //@line 3087 "libraries/lodepng/lodepng.cpp"
         HEAP8[$329>>0] = $191; //@line 3087 "libraries/lodepng/lodepng.cpp"
         $330 = $324 | 2; //@line 3087 "libraries/lodepng/lodepng.cpp"
         $331 = (($out) + ($330)|0); //@line 3087 "libraries/lodepng/lodepng.cpp"
         HEAP8[$331>>0] = $191; //@line 3087 "libraries/lodepng/lodepng.cpp"
         $332 = $324 | 5; //@line 3088 "libraries/lodepng/lodepng.cpp"
         $333 = (($out) + ($332)|0); //@line 3088 "libraries/lodepng/lodepng.cpp"
         HEAP8[$333>>0] = $192; //@line 3088 "libraries/lodepng/lodepng.cpp"
         $334 = $324 | 4; //@line 3088 "libraries/lodepng/lodepng.cpp"
         $335 = (($out) + ($334)|0); //@line 3088 "libraries/lodepng/lodepng.cpp"
         HEAP8[$335>>0] = $192; //@line 3088 "libraries/lodepng/lodepng.cpp"
         $336 = $324 | 7; //@line 3089 "libraries/lodepng/lodepng.cpp"
         $337 = (($out) + ($336)|0); //@line 3089 "libraries/lodepng/lodepng.cpp"
         HEAP8[$337>>0] = $193; //@line 3089 "libraries/lodepng/lodepng.cpp"
         $338 = $324 | 6; //@line 3089 "libraries/lodepng/lodepng.cpp"
         $339 = (($out) + ($338)|0); //@line 3089 "libraries/lodepng/lodepng.cpp"
         HEAP8[$339>>0] = $193; //@line 3089 "libraries/lodepng/lodepng.cpp"
         break L62;
        }
        break;
       }
       default: {
       }
       }
      } while(0);
      $340 = (($i$39) + 1)|0; //@line 3462 "libraries/lodepng/lodepng.cpp"
      $341 = ($340|0)==($0|0); //@line 3462 "libraries/lodepng/lodepng.cpp"
      if ($341) {
       break L59;
      } else {
       $i$39 = $340;
      }
     }
    }
   } while(0);
  }
 } while(0);
 $342 = HEAP32[$mode_out>>2]|0; //@line 3469 "libraries/lodepng/lodepng.cpp"
 $343 = ($342|0)==(3); //@line 3469 "libraries/lodepng/lodepng.cpp"
 if (!($343)) {
  STACKTOP = sp;return 0; //@line 3475 "libraries/lodepng/lodepng.cpp"
 }
 __ZL18color_tree_cleanupP9ColorTree($tree); //@line 3471 "libraries/lodepng/lodepng.cpp"
 STACKTOP = sp;return 0; //@line 3475 "libraries/lodepng/lodepng.cpp"
}
function __ZL19getPixelColorRGBA16PtS_S_S_PKhjPK16LodePNGColorMode($r,$g,$b,$a,$in,$i,$mode) {
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 $in = $in|0;
 $i = $i|0;
 $mode = $mode|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$mode>>2]|0; //@line 3385 "libraries/lodepng/lodepng.cpp"
 switch ($0|0) {
 case 0:  {
  $1 = $i << 1; //@line 3387 "libraries/lodepng/lodepng.cpp"
  $2 = (($in) + ($1)|0); //@line 3387 "libraries/lodepng/lodepng.cpp"
  $3 = HEAP8[$2>>0]|0; //@line 3387 "libraries/lodepng/lodepng.cpp"
  $4 = $3&255; //@line 3387 "libraries/lodepng/lodepng.cpp"
  $5 = $4 << 8; //@line 3387 "libraries/lodepng/lodepng.cpp"
  $6 = $1 | 1; //@line 3387 "libraries/lodepng/lodepng.cpp"
  $7 = (($in) + ($6)|0); //@line 3387 "libraries/lodepng/lodepng.cpp"
  $8 = HEAP8[$7>>0]|0; //@line 3387 "libraries/lodepng/lodepng.cpp"
  $9 = $8&255; //@line 3387 "libraries/lodepng/lodepng.cpp"
  $10 = $5 | $9; //@line 3387 "libraries/lodepng/lodepng.cpp"
  $11 = $10&65535; //@line 3387 "libraries/lodepng/lodepng.cpp"
  HEAP16[$b>>1] = $11; //@line 3387 "libraries/lodepng/lodepng.cpp"
  HEAP16[$g>>1] = $11; //@line 3387 "libraries/lodepng/lodepng.cpp"
  HEAP16[$r>>1] = $11; //@line 3387 "libraries/lodepng/lodepng.cpp"
  $12 = ((($mode)) + 16|0); //@line 3388 "libraries/lodepng/lodepng.cpp"
  $13 = HEAP32[$12>>2]|0; //@line 3388 "libraries/lodepng/lodepng.cpp"
  $14 = ($13|0)==(0); //@line 3388 "libraries/lodepng/lodepng.cpp"
  if (!($14)) {
   $15 = HEAP8[$2>>0]|0; //@line 3388 "libraries/lodepng/lodepng.cpp"
   $16 = $15&255; //@line 3388 "libraries/lodepng/lodepng.cpp"
   $17 = $16 << 8; //@line 3388 "libraries/lodepng/lodepng.cpp"
   $18 = HEAP8[$7>>0]|0; //@line 3388 "libraries/lodepng/lodepng.cpp"
   $19 = $18&255; //@line 3388 "libraries/lodepng/lodepng.cpp"
   $20 = $17 | $19; //@line 3388 "libraries/lodepng/lodepng.cpp"
   $21 = ((($mode)) + 20|0); //@line 3388 "libraries/lodepng/lodepng.cpp"
   $22 = HEAP32[$21>>2]|0; //@line 3388 "libraries/lodepng/lodepng.cpp"
   $23 = ($20|0)==($22|0); //@line 3388 "libraries/lodepng/lodepng.cpp"
   if ($23) {
    HEAP16[$a>>1] = 0; //@line 3388 "libraries/lodepng/lodepng.cpp"
    return; //@line 3413 "libraries/lodepng/lodepng.cpp"
   }
  }
  HEAP16[$a>>1] = -1; //@line 3389 "libraries/lodepng/lodepng.cpp"
  return; //@line 3413 "libraries/lodepng/lodepng.cpp"
  break;
 }
 case 2:  {
  $24 = ($i*6)|0; //@line 3393 "libraries/lodepng/lodepng.cpp"
  $25 = (($in) + ($24)|0); //@line 3393 "libraries/lodepng/lodepng.cpp"
  $26 = HEAP8[$25>>0]|0; //@line 3393 "libraries/lodepng/lodepng.cpp"
  $27 = $26&255; //@line 3393 "libraries/lodepng/lodepng.cpp"
  $28 = $27 << 8; //@line 3393 "libraries/lodepng/lodepng.cpp"
  $29 = $24 | 1; //@line 3393 "libraries/lodepng/lodepng.cpp"
  $30 = (($in) + ($29)|0); //@line 3393 "libraries/lodepng/lodepng.cpp"
  $31 = HEAP8[$30>>0]|0; //@line 3393 "libraries/lodepng/lodepng.cpp"
  $32 = $31&255; //@line 3393 "libraries/lodepng/lodepng.cpp"
  $33 = $28 | $32; //@line 3393 "libraries/lodepng/lodepng.cpp"
  $34 = $33&65535; //@line 3393 "libraries/lodepng/lodepng.cpp"
  HEAP16[$r>>1] = $34; //@line 3393 "libraries/lodepng/lodepng.cpp"
  $35 = (($24) + 2)|0; //@line 3394 "libraries/lodepng/lodepng.cpp"
  $36 = (($in) + ($35)|0); //@line 3394 "libraries/lodepng/lodepng.cpp"
  $37 = HEAP8[$36>>0]|0; //@line 3394 "libraries/lodepng/lodepng.cpp"
  $38 = $37&255; //@line 3394 "libraries/lodepng/lodepng.cpp"
  $39 = $38 << 8; //@line 3394 "libraries/lodepng/lodepng.cpp"
  $40 = (($24) + 3)|0; //@line 3394 "libraries/lodepng/lodepng.cpp"
  $41 = (($in) + ($40)|0); //@line 3394 "libraries/lodepng/lodepng.cpp"
  $42 = HEAP8[$41>>0]|0; //@line 3394 "libraries/lodepng/lodepng.cpp"
  $43 = $42&255; //@line 3394 "libraries/lodepng/lodepng.cpp"
  $44 = $39 | $43; //@line 3394 "libraries/lodepng/lodepng.cpp"
  $45 = $44&65535; //@line 3394 "libraries/lodepng/lodepng.cpp"
  HEAP16[$g>>1] = $45; //@line 3394 "libraries/lodepng/lodepng.cpp"
  $46 = (($24) + 4)|0; //@line 3395 "libraries/lodepng/lodepng.cpp"
  $47 = (($in) + ($46)|0); //@line 3395 "libraries/lodepng/lodepng.cpp"
  $48 = HEAP8[$47>>0]|0; //@line 3395 "libraries/lodepng/lodepng.cpp"
  $49 = $48&255; //@line 3395 "libraries/lodepng/lodepng.cpp"
  $50 = $49 << 8; //@line 3395 "libraries/lodepng/lodepng.cpp"
  $51 = (($24) + 5)|0; //@line 3395 "libraries/lodepng/lodepng.cpp"
  $52 = (($in) + ($51)|0); //@line 3395 "libraries/lodepng/lodepng.cpp"
  $53 = HEAP8[$52>>0]|0; //@line 3395 "libraries/lodepng/lodepng.cpp"
  $54 = $53&255; //@line 3395 "libraries/lodepng/lodepng.cpp"
  $55 = $50 | $54; //@line 3395 "libraries/lodepng/lodepng.cpp"
  $56 = $55&65535; //@line 3395 "libraries/lodepng/lodepng.cpp"
  HEAP16[$b>>1] = $56; //@line 3395 "libraries/lodepng/lodepng.cpp"
  $57 = ((($mode)) + 16|0); //@line 3396 "libraries/lodepng/lodepng.cpp"
  $58 = HEAP32[$57>>2]|0; //@line 3396 "libraries/lodepng/lodepng.cpp"
  $59 = ($58|0)==(0); //@line 3396 "libraries/lodepng/lodepng.cpp"
  if (!($59)) {
   $60 = HEAP8[$25>>0]|0; //@line 3396 "libraries/lodepng/lodepng.cpp"
   $61 = $60&255; //@line 3396 "libraries/lodepng/lodepng.cpp"
   $62 = $61 << 8; //@line 3396 "libraries/lodepng/lodepng.cpp"
   $63 = HEAP8[$30>>0]|0; //@line 3396 "libraries/lodepng/lodepng.cpp"
   $64 = $63&255; //@line 3396 "libraries/lodepng/lodepng.cpp"
   $65 = $62 | $64; //@line 3396 "libraries/lodepng/lodepng.cpp"
   $66 = ((($mode)) + 20|0); //@line 3396 "libraries/lodepng/lodepng.cpp"
   $67 = HEAP32[$66>>2]|0; //@line 3396 "libraries/lodepng/lodepng.cpp"
   $68 = ($65|0)==($67|0); //@line 3396 "libraries/lodepng/lodepng.cpp"
   if ($68) {
    $69 = HEAP8[$36>>0]|0; //@line 3397 "libraries/lodepng/lodepng.cpp"
    $70 = $69&255; //@line 3397 "libraries/lodepng/lodepng.cpp"
    $71 = $70 << 8; //@line 3397 "libraries/lodepng/lodepng.cpp"
    $72 = HEAP8[$41>>0]|0; //@line 3397 "libraries/lodepng/lodepng.cpp"
    $73 = $72&255; //@line 3397 "libraries/lodepng/lodepng.cpp"
    $74 = $71 | $73; //@line 3397 "libraries/lodepng/lodepng.cpp"
    $75 = ((($mode)) + 24|0); //@line 3397 "libraries/lodepng/lodepng.cpp"
    $76 = HEAP32[$75>>2]|0; //@line 3397 "libraries/lodepng/lodepng.cpp"
    $77 = ($74|0)==($76|0); //@line 3397 "libraries/lodepng/lodepng.cpp"
    if ($77) {
     $78 = HEAP8[$47>>0]|0; //@line 3398 "libraries/lodepng/lodepng.cpp"
     $79 = $78&255; //@line 3398 "libraries/lodepng/lodepng.cpp"
     $80 = $79 << 8; //@line 3398 "libraries/lodepng/lodepng.cpp"
     $81 = HEAP8[$52>>0]|0; //@line 3398 "libraries/lodepng/lodepng.cpp"
     $82 = $81&255; //@line 3398 "libraries/lodepng/lodepng.cpp"
     $83 = $80 | $82; //@line 3398 "libraries/lodepng/lodepng.cpp"
     $84 = ((($mode)) + 28|0); //@line 3398 "libraries/lodepng/lodepng.cpp"
     $85 = HEAP32[$84>>2]|0; //@line 3398 "libraries/lodepng/lodepng.cpp"
     $86 = ($83|0)==($85|0); //@line 3398 "libraries/lodepng/lodepng.cpp"
     if ($86) {
      HEAP16[$a>>1] = 0; //@line 3398 "libraries/lodepng/lodepng.cpp"
      return; //@line 3413 "libraries/lodepng/lodepng.cpp"
     }
    }
   }
  }
  HEAP16[$a>>1] = -1; //@line 3399 "libraries/lodepng/lodepng.cpp"
  return; //@line 3413 "libraries/lodepng/lodepng.cpp"
  break;
 }
 case 4:  {
  $87 = $i << 2; //@line 3403 "libraries/lodepng/lodepng.cpp"
  $88 = (($in) + ($87)|0); //@line 3403 "libraries/lodepng/lodepng.cpp"
  $89 = HEAP8[$88>>0]|0; //@line 3403 "libraries/lodepng/lodepng.cpp"
  $90 = $89&255; //@line 3403 "libraries/lodepng/lodepng.cpp"
  $91 = $90 << 8; //@line 3403 "libraries/lodepng/lodepng.cpp"
  $92 = $87 | 1; //@line 3403 "libraries/lodepng/lodepng.cpp"
  $93 = (($in) + ($92)|0); //@line 3403 "libraries/lodepng/lodepng.cpp"
  $94 = HEAP8[$93>>0]|0; //@line 3403 "libraries/lodepng/lodepng.cpp"
  $95 = $94&255; //@line 3403 "libraries/lodepng/lodepng.cpp"
  $96 = $91 | $95; //@line 3403 "libraries/lodepng/lodepng.cpp"
  $97 = $96&65535; //@line 3403 "libraries/lodepng/lodepng.cpp"
  HEAP16[$b>>1] = $97; //@line 3403 "libraries/lodepng/lodepng.cpp"
  HEAP16[$g>>1] = $97; //@line 3403 "libraries/lodepng/lodepng.cpp"
  HEAP16[$r>>1] = $97; //@line 3403 "libraries/lodepng/lodepng.cpp"
  $98 = $87 | 2; //@line 3404 "libraries/lodepng/lodepng.cpp"
  $99 = (($in) + ($98)|0); //@line 3404 "libraries/lodepng/lodepng.cpp"
  $100 = HEAP8[$99>>0]|0; //@line 3404 "libraries/lodepng/lodepng.cpp"
  $101 = $100&255; //@line 3404 "libraries/lodepng/lodepng.cpp"
  $102 = $101 << 8; //@line 3404 "libraries/lodepng/lodepng.cpp"
  $103 = $87 | 3; //@line 3404 "libraries/lodepng/lodepng.cpp"
  $104 = (($in) + ($103)|0); //@line 3404 "libraries/lodepng/lodepng.cpp"
  $105 = HEAP8[$104>>0]|0; //@line 3404 "libraries/lodepng/lodepng.cpp"
  $106 = $105&255; //@line 3404 "libraries/lodepng/lodepng.cpp"
  $107 = $102 | $106; //@line 3404 "libraries/lodepng/lodepng.cpp"
  $108 = $107&65535; //@line 3404 "libraries/lodepng/lodepng.cpp"
  HEAP16[$a>>1] = $108; //@line 3404 "libraries/lodepng/lodepng.cpp"
  return; //@line 3413 "libraries/lodepng/lodepng.cpp"
  break;
 }
 case 6:  {
  $109 = $i << 3; //@line 3408 "libraries/lodepng/lodepng.cpp"
  $110 = (($in) + ($109)|0); //@line 3408 "libraries/lodepng/lodepng.cpp"
  $111 = HEAP8[$110>>0]|0; //@line 3408 "libraries/lodepng/lodepng.cpp"
  $112 = $111&255; //@line 3408 "libraries/lodepng/lodepng.cpp"
  $113 = $112 << 8; //@line 3408 "libraries/lodepng/lodepng.cpp"
  $114 = $109 | 1; //@line 3408 "libraries/lodepng/lodepng.cpp"
  $115 = (($in) + ($114)|0); //@line 3408 "libraries/lodepng/lodepng.cpp"
  $116 = HEAP8[$115>>0]|0; //@line 3408 "libraries/lodepng/lodepng.cpp"
  $117 = $116&255; //@line 3408 "libraries/lodepng/lodepng.cpp"
  $118 = $113 | $117; //@line 3408 "libraries/lodepng/lodepng.cpp"
  $119 = $118&65535; //@line 3408 "libraries/lodepng/lodepng.cpp"
  HEAP16[$r>>1] = $119; //@line 3408 "libraries/lodepng/lodepng.cpp"
  $120 = $109 | 2; //@line 3409 "libraries/lodepng/lodepng.cpp"
  $121 = (($in) + ($120)|0); //@line 3409 "libraries/lodepng/lodepng.cpp"
  $122 = HEAP8[$121>>0]|0; //@line 3409 "libraries/lodepng/lodepng.cpp"
  $123 = $122&255; //@line 3409 "libraries/lodepng/lodepng.cpp"
  $124 = $123 << 8; //@line 3409 "libraries/lodepng/lodepng.cpp"
  $125 = $109 | 3; //@line 3409 "libraries/lodepng/lodepng.cpp"
  $126 = (($in) + ($125)|0); //@line 3409 "libraries/lodepng/lodepng.cpp"
  $127 = HEAP8[$126>>0]|0; //@line 3409 "libraries/lodepng/lodepng.cpp"
  $128 = $127&255; //@line 3409 "libraries/lodepng/lodepng.cpp"
  $129 = $124 | $128; //@line 3409 "libraries/lodepng/lodepng.cpp"
  $130 = $129&65535; //@line 3409 "libraries/lodepng/lodepng.cpp"
  HEAP16[$g>>1] = $130; //@line 3409 "libraries/lodepng/lodepng.cpp"
  $131 = $109 | 4; //@line 3410 "libraries/lodepng/lodepng.cpp"
  $132 = (($in) + ($131)|0); //@line 3410 "libraries/lodepng/lodepng.cpp"
  $133 = HEAP8[$132>>0]|0; //@line 3410 "libraries/lodepng/lodepng.cpp"
  $134 = $133&255; //@line 3410 "libraries/lodepng/lodepng.cpp"
  $135 = $134 << 8; //@line 3410 "libraries/lodepng/lodepng.cpp"
  $136 = $109 | 5; //@line 3410 "libraries/lodepng/lodepng.cpp"
  $137 = (($in) + ($136)|0); //@line 3410 "libraries/lodepng/lodepng.cpp"
  $138 = HEAP8[$137>>0]|0; //@line 3410 "libraries/lodepng/lodepng.cpp"
  $139 = $138&255; //@line 3410 "libraries/lodepng/lodepng.cpp"
  $140 = $135 | $139; //@line 3410 "libraries/lodepng/lodepng.cpp"
  $141 = $140&65535; //@line 3410 "libraries/lodepng/lodepng.cpp"
  HEAP16[$b>>1] = $141; //@line 3410 "libraries/lodepng/lodepng.cpp"
  $142 = $109 | 6; //@line 3411 "libraries/lodepng/lodepng.cpp"
  $143 = (($in) + ($142)|0); //@line 3411 "libraries/lodepng/lodepng.cpp"
  $144 = HEAP8[$143>>0]|0; //@line 3411 "libraries/lodepng/lodepng.cpp"
  $145 = $144&255; //@line 3411 "libraries/lodepng/lodepng.cpp"
  $146 = $145 << 8; //@line 3411 "libraries/lodepng/lodepng.cpp"
  $147 = $109 | 7; //@line 3411 "libraries/lodepng/lodepng.cpp"
  $148 = (($in) + ($147)|0); //@line 3411 "libraries/lodepng/lodepng.cpp"
  $149 = HEAP8[$148>>0]|0; //@line 3411 "libraries/lodepng/lodepng.cpp"
  $150 = $149&255; //@line 3411 "libraries/lodepng/lodepng.cpp"
  $151 = $146 | $150; //@line 3411 "libraries/lodepng/lodepng.cpp"
  $152 = $151&65535; //@line 3411 "libraries/lodepng/lodepng.cpp"
  HEAP16[$a>>1] = $152; //@line 3411 "libraries/lodepng/lodepng.cpp"
  return; //@line 3413 "libraries/lodepng/lodepng.cpp"
  break;
 }
 default: {
  return; //@line 3413 "libraries/lodepng/lodepng.cpp"
 }
 }
}
function __ZL19getPixelColorsRGBA8PhjjPKhPK16LodePNGColorMode($buffer,$numpixels,$has_alpha,$in,$mode) {
 $buffer = $buffer|0;
 $numpixels = $numpixels|0;
 $has_alpha = $has_alpha|0;
 $in = $in|0;
 $mode = $mode|0;
 var $$038 = 0, $$038$us = 0, $$141 = 0, $$141$us = 0, $$236 = 0, $$236$us = 0, $$344 = 0, $$344$us = 0, $$447 = 0, $$447$us = 0, $$551 = 0, $$551$us = 0, $$654 = 0, $$654$us = 0, $$757 = 0, $$757$us = 0, $$860 = 0, $$860$us = 0, $$963 = 0, $$963$us = 0;
 var $$pre = 0, $$pre92 = 0, $$sink = 0, $$sink91 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0;
 var $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0;
 var $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0;
 var $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0;
 var $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0;
 var $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0;
 var $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0;
 var $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0;
 var $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0;
 var $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $i$0$i = 0, $i$0$i$us = 0, $i$0$i10 = 0, $i$0$i10$us = 0, $i$01$i = 0, $i$01$i$us = 0, $i$01$i5 = 0, $i$01$i5$us = 0, $i$03$i = 0, $i$03$i$us = 0, $i$03$i8 = 0, $i$03$i8$us = 0, $i$037 = 0, $i$037$us = 0, $i$140 = 0, $i$140$us = 0, $i$235 = 0, $i$235$us = 0, $i$343 = 0, $i$343$us = 0;
 var $i$446 = 0, $i$446$us = 0, $i$550 = 0, $i$550$us = 0, $i$653 = 0, $i$653$us = 0, $i$756 = 0, $i$756$us = 0, $i$859 = 0, $i$859$us = 0, $i$962 = 0, $i$962$us = 0, $index$0 = 0, $index$0$us = 0, $j$034 = 0, $j$034$us = 0, $j$1 = 0, $j$1$us = 0, $j1$049 = 0, $j1$049$us = 0;
 var $j1$2 = 0, $j1$2$us = 0, $not$$us = 0, $not$1$us = 0, $not$2$us = 0, $not$3$us = 0, $not$4$us = 0, $result$0$lcssa$i = 0, $result$0$lcssa$i$us = 0, $result$02$i = 0, $result$02$i$us = 0, $result$02$i9 = 0, $result$02$i9$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($has_alpha|0)!=(0); //@line 3252 "libraries/lodepng/lodepng.cpp"
 $1 = $0 ? 4 : 3; //@line 3252 "libraries/lodepng/lodepng.cpp"
 $2 = HEAP32[$mode>>2]|0; //@line 3254 "libraries/lodepng/lodepng.cpp"
 switch ($2|0) {
 case 0:  {
  $50 = ((($mode)) + 4|0); //@line 3256 "libraries/lodepng/lodepng.cpp"
  $51 = HEAP32[$50>>2]|0; //@line 3256 "libraries/lodepng/lodepng.cpp"
  switch ($51|0) {
  case 8:  {
   $78 = ($numpixels|0)==(0); //@line 3258 "libraries/lodepng/lodepng.cpp"
   if ($78) {
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
   $79 = ((($mode)) + 16|0); //@line 3261 "libraries/lodepng/lodepng.cpp"
   $80 = ((($mode)) + 20|0); //@line 3261 "libraries/lodepng/lodepng.cpp"
   if ($0) {
    $$038$us = $buffer;$i$037$us = 0;
   } else {
    $$038 = $buffer;$i$037 = 0;
    while(1) {
     $97 = (($in) + ($i$037)|0); //@line 3260 "libraries/lodepng/lodepng.cpp"
     $98 = HEAP8[$97>>0]|0; //@line 3260 "libraries/lodepng/lodepng.cpp"
     $99 = ((($$038)) + 2|0); //@line 3260 "libraries/lodepng/lodepng.cpp"
     HEAP8[$99>>0] = $98; //@line 3260 "libraries/lodepng/lodepng.cpp"
     $100 = ((($$038)) + 1|0); //@line 3260 "libraries/lodepng/lodepng.cpp"
     HEAP8[$100>>0] = $98; //@line 3260 "libraries/lodepng/lodepng.cpp"
     HEAP8[$$038>>0] = $98; //@line 3260 "libraries/lodepng/lodepng.cpp"
     $101 = (($i$037) + 1)|0; //@line 3258 "libraries/lodepng/lodepng.cpp"
     $102 = (($$038) + ($1)|0); //@line 3258 "libraries/lodepng/lodepng.cpp"
     $103 = ($101|0)==($numpixels|0); //@line 3258 "libraries/lodepng/lodepng.cpp"
     if ($103) {
      break;
     } else {
      $$038 = $102;$i$037 = $101;
     }
    }
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
   while(1) {
    $81 = (($in) + ($i$037$us)|0); //@line 3260 "libraries/lodepng/lodepng.cpp"
    $82 = HEAP8[$81>>0]|0; //@line 3260 "libraries/lodepng/lodepng.cpp"
    $83 = ((($$038$us)) + 2|0); //@line 3260 "libraries/lodepng/lodepng.cpp"
    HEAP8[$83>>0] = $82; //@line 3260 "libraries/lodepng/lodepng.cpp"
    $84 = ((($$038$us)) + 1|0); //@line 3260 "libraries/lodepng/lodepng.cpp"
    HEAP8[$84>>0] = $82; //@line 3260 "libraries/lodepng/lodepng.cpp"
    HEAP8[$$038$us>>0] = $82; //@line 3260 "libraries/lodepng/lodepng.cpp"
    $85 = HEAP32[$79>>2]|0; //@line 3261 "libraries/lodepng/lodepng.cpp"
    $86 = ($85|0)==(0); //@line 3261 "libraries/lodepng/lodepng.cpp"
    if ($86) {
     $91 = 0;
    } else {
     $87 = HEAP8[$81>>0]|0; //@line 3261 "libraries/lodepng/lodepng.cpp"
     $88 = $87&255; //@line 3261 "libraries/lodepng/lodepng.cpp"
     $89 = HEAP32[$80>>2]|0; //@line 3261 "libraries/lodepng/lodepng.cpp"
     $90 = ($88|0)==($89|0); //@line 3261 "libraries/lodepng/lodepng.cpp"
     $91 = $90;
    }
    $not$4$us = $91 ^ 1; //@line 3261 "libraries/lodepng/lodepng.cpp"
    $92 = $not$4$us << 31 >> 31; //@line 3261 "libraries/lodepng/lodepng.cpp"
    $93 = ((($$038$us)) + 3|0); //@line 3261 "libraries/lodepng/lodepng.cpp"
    HEAP8[$93>>0] = $92; //@line 3261 "libraries/lodepng/lodepng.cpp"
    $94 = (($i$037$us) + 1)|0; //@line 3258 "libraries/lodepng/lodepng.cpp"
    $95 = (($$038$us) + ($1)|0); //@line 3258 "libraries/lodepng/lodepng.cpp"
    $96 = ($94|0)==($numpixels|0); //@line 3258 "libraries/lodepng/lodepng.cpp"
    if ($96) {
     break;
    } else {
     $$038$us = $95;$i$037$us = $94;
    }
   }
   return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   break;
  }
  case 16:  {
   $52 = ($numpixels|0)==(0); //@line 3266 "libraries/lodepng/lodepng.cpp"
   if ($52) {
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
   $53 = ((($mode)) + 16|0); //@line 3269 "libraries/lodepng/lodepng.cpp"
   $54 = ((($mode)) + 20|0); //@line 3269 "libraries/lodepng/lodepng.cpp"
   if ($0) {
    $$141$us = $buffer;$i$140$us = 0;
   } else {
    $$141 = $buffer;$i$140 = 0;
    while(1) {
     $104 = $i$140 << 1; //@line 3268 "libraries/lodepng/lodepng.cpp"
     $105 = (($in) + ($104)|0); //@line 3268 "libraries/lodepng/lodepng.cpp"
     $106 = HEAP8[$105>>0]|0; //@line 3268 "libraries/lodepng/lodepng.cpp"
     $107 = ((($$141)) + 2|0); //@line 3268 "libraries/lodepng/lodepng.cpp"
     HEAP8[$107>>0] = $106; //@line 3268 "libraries/lodepng/lodepng.cpp"
     $108 = ((($$141)) + 1|0); //@line 3268 "libraries/lodepng/lodepng.cpp"
     HEAP8[$108>>0] = $106; //@line 3268 "libraries/lodepng/lodepng.cpp"
     HEAP8[$$141>>0] = $106; //@line 3268 "libraries/lodepng/lodepng.cpp"
     $109 = (($i$140) + 1)|0; //@line 3266 "libraries/lodepng/lodepng.cpp"
     $110 = (($$141) + ($1)|0); //@line 3266 "libraries/lodepng/lodepng.cpp"
     $111 = ($109|0)==($numpixels|0); //@line 3266 "libraries/lodepng/lodepng.cpp"
     if ($111) {
      break;
     } else {
      $$141 = $110;$i$140 = $109;
     }
    }
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
   while(1) {
    $55 = $i$140$us << 1; //@line 3268 "libraries/lodepng/lodepng.cpp"
    $56 = (($in) + ($55)|0); //@line 3268 "libraries/lodepng/lodepng.cpp"
    $57 = HEAP8[$56>>0]|0; //@line 3268 "libraries/lodepng/lodepng.cpp"
    $58 = ((($$141$us)) + 2|0); //@line 3268 "libraries/lodepng/lodepng.cpp"
    HEAP8[$58>>0] = $57; //@line 3268 "libraries/lodepng/lodepng.cpp"
    $59 = ((($$141$us)) + 1|0); //@line 3268 "libraries/lodepng/lodepng.cpp"
    HEAP8[$59>>0] = $57; //@line 3268 "libraries/lodepng/lodepng.cpp"
    HEAP8[$$141$us>>0] = $57; //@line 3268 "libraries/lodepng/lodepng.cpp"
    $60 = HEAP32[$53>>2]|0; //@line 3269 "libraries/lodepng/lodepng.cpp"
    $61 = ($60|0)==(0); //@line 3269 "libraries/lodepng/lodepng.cpp"
    if ($61) {
     $72 = 0;
    } else {
     $62 = HEAP8[$56>>0]|0; //@line 3269 "libraries/lodepng/lodepng.cpp"
     $63 = $62&255; //@line 3269 "libraries/lodepng/lodepng.cpp"
     $64 = $63 << 8; //@line 3269 "libraries/lodepng/lodepng.cpp"
     $65 = $55 | 1; //@line 3269 "libraries/lodepng/lodepng.cpp"
     $66 = (($in) + ($65)|0); //@line 3269 "libraries/lodepng/lodepng.cpp"
     $67 = HEAP8[$66>>0]|0; //@line 3269 "libraries/lodepng/lodepng.cpp"
     $68 = $67&255; //@line 3269 "libraries/lodepng/lodepng.cpp"
     $69 = $64 | $68; //@line 3269 "libraries/lodepng/lodepng.cpp"
     $70 = HEAP32[$54>>2]|0; //@line 3269 "libraries/lodepng/lodepng.cpp"
     $71 = ($69|0)==($70|0); //@line 3269 "libraries/lodepng/lodepng.cpp"
     $72 = $71;
    }
    $not$3$us = $72 ^ 1; //@line 3269 "libraries/lodepng/lodepng.cpp"
    $73 = $not$3$us << 31 >> 31; //@line 3269 "libraries/lodepng/lodepng.cpp"
    $74 = ((($$141$us)) + 3|0); //@line 3269 "libraries/lodepng/lodepng.cpp"
    HEAP8[$74>>0] = $73; //@line 3269 "libraries/lodepng/lodepng.cpp"
    $75 = (($i$140$us) + 1)|0; //@line 3266 "libraries/lodepng/lodepng.cpp"
    $76 = (($$141$us) + ($1)|0); //@line 3266 "libraries/lodepng/lodepng.cpp"
    $77 = ($75|0)==($numpixels|0); //@line 3266 "libraries/lodepng/lodepng.cpp"
    if ($77) {
     break;
    } else {
     $$141$us = $76;$i$140$us = $75;
    }
   }
   return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   break;
  }
  default: {
   $112 = 1 << $51; //@line 3274 "libraries/lodepng/lodepng.cpp"
   $113 = (($112) + -1)|0; //@line 3274 "libraries/lodepng/lodepng.cpp"
   $114 = ($numpixels|0)==(0); //@line 3276 "libraries/lodepng/lodepng.cpp"
   if ($114) {
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
   $115 = ((($mode)) + 16|0); //@line 3280 "libraries/lodepng/lodepng.cpp"
   $116 = ((($mode)) + 20|0); //@line 3280 "libraries/lodepng/lodepng.cpp"
   if ($0) {
    $$236$us = $buffer;$118 = $51;$i$235$us = 0;$j$034$us = 0;
   } else {
    $$236 = $buffer;$148 = $51;$i$235 = 0;$j$034 = 0;
    while(1) {
     $147 = ($148|0)==(0); //@line 2360 "libraries/lodepng/lodepng.cpp"
     if ($147) {
      $j$1 = $j$034;$result$0$lcssa$i = 0;
     } else {
      $i$01$i = (($148) + -1)|0; //@line 2360 "libraries/lodepng/lodepng.cpp"
      $150 = $j$034;$i$03$i = $i$01$i;$result$02$i = 0;
      while(1) {
       $149 = $150 >>> 3; //@line 2351 "libraries/lodepng/lodepng.cpp"
       $151 = (($in) + ($149)|0); //@line 2351 "libraries/lodepng/lodepng.cpp"
       $152 = HEAP8[$151>>0]|0; //@line 2351 "libraries/lodepng/lodepng.cpp"
       $153 = $152&255; //@line 2351 "libraries/lodepng/lodepng.cpp"
       $154 = $150 & 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
       $155 = $154 ^ 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
       $156 = $153 >>> $155; //@line 2351 "libraries/lodepng/lodepng.cpp"
       $157 = $156 & 1; //@line 2351 "libraries/lodepng/lodepng.cpp"
       $158 = (($150) + 1)|0; //@line 2352 "libraries/lodepng/lodepng.cpp"
       $159 = $157 << $i$03$i; //@line 2362 "libraries/lodepng/lodepng.cpp"
       $160 = (($159) + ($result$02$i))|0; //@line 2362 "libraries/lodepng/lodepng.cpp"
       $i$0$i = (($i$03$i) + -1)|0; //@line 2360 "libraries/lodepng/lodepng.cpp"
       $161 = ($i$0$i>>>0)<($148>>>0); //@line 2360 "libraries/lodepng/lodepng.cpp"
       if ($161) {
        $150 = $158;$i$03$i = $i$0$i;$result$02$i = $160;
       } else {
        $j$1 = $158;$result$0$lcssa$i = $160;
        break;
       }
      }
     }
     $162 = ($result$0$lcssa$i*255)|0; //@line 3279 "libraries/lodepng/lodepng.cpp"
     $163 = (($162>>>0) / ($113>>>0))&-1; //@line 3279 "libraries/lodepng/lodepng.cpp"
     $164 = $163&255; //@line 3279 "libraries/lodepng/lodepng.cpp"
     $165 = ((($$236)) + 2|0); //@line 3279 "libraries/lodepng/lodepng.cpp"
     HEAP8[$165>>0] = $164; //@line 3279 "libraries/lodepng/lodepng.cpp"
     $166 = ((($$236)) + 1|0); //@line 3279 "libraries/lodepng/lodepng.cpp"
     HEAP8[$166>>0] = $164; //@line 3279 "libraries/lodepng/lodepng.cpp"
     HEAP8[$$236>>0] = $164; //@line 3279 "libraries/lodepng/lodepng.cpp"
     $167 = (($i$235) + 1)|0; //@line 3276 "libraries/lodepng/lodepng.cpp"
     $168 = ($167|0)==($numpixels|0); //@line 3276 "libraries/lodepng/lodepng.cpp"
     if ($168) {
      break;
     }
     $169 = (($$236) + ($1)|0); //@line 3276 "libraries/lodepng/lodepng.cpp"
     $$pre = HEAP32[$50>>2]|0; //@line 3278 "libraries/lodepng/lodepng.cpp"
     $$236 = $169;$148 = $$pre;$i$235 = $167;$j$034 = $j$1;
    }
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
   while(1) {
    $117 = ($118|0)==(0); //@line 2360 "libraries/lodepng/lodepng.cpp"
    if ($117) {
     $j$1$us = $j$034$us;$result$0$lcssa$i$us = 0;
    } else {
     $i$01$i$us = (($118) + -1)|0; //@line 2360 "libraries/lodepng/lodepng.cpp"
     $120 = $j$034$us;$i$03$i$us = $i$01$i$us;$result$02$i$us = 0;
     while(1) {
      $119 = $120 >>> 3; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $121 = (($in) + ($119)|0); //@line 2351 "libraries/lodepng/lodepng.cpp"
      $122 = HEAP8[$121>>0]|0; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $123 = $122&255; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $124 = $120 & 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $125 = $124 ^ 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $126 = $123 >>> $125; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $127 = $126 & 1; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $128 = (($120) + 1)|0; //@line 2352 "libraries/lodepng/lodepng.cpp"
      $129 = $127 << $i$03$i$us; //@line 2362 "libraries/lodepng/lodepng.cpp"
      $130 = (($129) + ($result$02$i$us))|0; //@line 2362 "libraries/lodepng/lodepng.cpp"
      $i$0$i$us = (($i$03$i$us) + -1)|0; //@line 2360 "libraries/lodepng/lodepng.cpp"
      $131 = ($i$0$i$us>>>0)<($118>>>0); //@line 2360 "libraries/lodepng/lodepng.cpp"
      if ($131) {
       $120 = $128;$i$03$i$us = $i$0$i$us;$result$02$i$us = $130;
      } else {
       $j$1$us = $128;$result$0$lcssa$i$us = $130;
       break;
      }
     }
    }
    $132 = ($result$0$lcssa$i$us*255)|0; //@line 3279 "libraries/lodepng/lodepng.cpp"
    $133 = (($132>>>0) / ($113>>>0))&-1; //@line 3279 "libraries/lodepng/lodepng.cpp"
    $134 = $133&255; //@line 3279 "libraries/lodepng/lodepng.cpp"
    $135 = ((($$236$us)) + 2|0); //@line 3279 "libraries/lodepng/lodepng.cpp"
    HEAP8[$135>>0] = $134; //@line 3279 "libraries/lodepng/lodepng.cpp"
    $136 = ((($$236$us)) + 1|0); //@line 3279 "libraries/lodepng/lodepng.cpp"
    HEAP8[$136>>0] = $134; //@line 3279 "libraries/lodepng/lodepng.cpp"
    HEAP8[$$236$us>>0] = $134; //@line 3279 "libraries/lodepng/lodepng.cpp"
    $137 = HEAP32[$115>>2]|0; //@line 3280 "libraries/lodepng/lodepng.cpp"
    $138 = ($137|0)==(0); //@line 3280 "libraries/lodepng/lodepng.cpp"
    if ($138) {
     $141 = 0;
    } else {
     $139 = HEAP32[$116>>2]|0; //@line 3280 "libraries/lodepng/lodepng.cpp"
     $140 = ($result$0$lcssa$i$us|0)==($139|0); //@line 3280 "libraries/lodepng/lodepng.cpp"
     $141 = $140;
    }
    $not$2$us = $141 ^ 1; //@line 3280 "libraries/lodepng/lodepng.cpp"
    $142 = $not$2$us << 31 >> 31; //@line 3280 "libraries/lodepng/lodepng.cpp"
    $143 = ((($$236$us)) + 3|0); //@line 3280 "libraries/lodepng/lodepng.cpp"
    HEAP8[$143>>0] = $142; //@line 3280 "libraries/lodepng/lodepng.cpp"
    $144 = (($i$235$us) + 1)|0; //@line 3276 "libraries/lodepng/lodepng.cpp"
    $145 = ($144|0)==($numpixels|0); //@line 3276 "libraries/lodepng/lodepng.cpp"
    if ($145) {
     break;
    }
    $146 = (($$236$us) + ($1)|0); //@line 3276 "libraries/lodepng/lodepng.cpp"
    $$pre92 = HEAP32[$50>>2]|0; //@line 3278 "libraries/lodepng/lodepng.cpp"
    $$236$us = $146;$118 = $$pre92;$i$235$us = $144;$j$034$us = $j$1$us;
   }
   return; //@line 3378 "libraries/lodepng/lodepng.cpp"
  }
  }
  break;
 }
 case 2:  {
  $170 = ((($mode)) + 4|0); //@line 3286 "libraries/lodepng/lodepng.cpp"
  $171 = HEAP32[$170>>2]|0; //@line 3286 "libraries/lodepng/lodepng.cpp"
  $172 = ($171|0)==(8); //@line 3286 "libraries/lodepng/lodepng.cpp"
  $173 = ($numpixels|0)==(0); //@line 3288 "libraries/lodepng/lodepng.cpp"
  if ($172) {
   if ($173) {
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
   $227 = ((($mode)) + 16|0); //@line 3293 "libraries/lodepng/lodepng.cpp"
   $228 = ((($mode)) + 20|0); //@line 3293 "libraries/lodepng/lodepng.cpp"
   $229 = ((($mode)) + 24|0); //@line 3294 "libraries/lodepng/lodepng.cpp"
   $230 = ((($mode)) + 28|0); //@line 3294 "libraries/lodepng/lodepng.cpp"
   if ($0) {
    $$344$us = $buffer;$i$343$us = 0;
   } else {
    $$344 = $buffer;$i$343 = 0;
    while(1) {
     $259 = ($i$343*3)|0; //@line 3290 "libraries/lodepng/lodepng.cpp"
     $260 = (($in) + ($259)|0); //@line 3290 "libraries/lodepng/lodepng.cpp"
     $261 = HEAP8[$260>>0]|0; //@line 3290 "libraries/lodepng/lodepng.cpp"
     HEAP8[$$344>>0] = $261; //@line 3290 "libraries/lodepng/lodepng.cpp"
     $262 = (($259) + 1)|0; //@line 3291 "libraries/lodepng/lodepng.cpp"
     $263 = (($in) + ($262)|0); //@line 3291 "libraries/lodepng/lodepng.cpp"
     $264 = HEAP8[$263>>0]|0; //@line 3291 "libraries/lodepng/lodepng.cpp"
     $265 = ((($$344)) + 1|0); //@line 3291 "libraries/lodepng/lodepng.cpp"
     HEAP8[$265>>0] = $264; //@line 3291 "libraries/lodepng/lodepng.cpp"
     $266 = (($259) + 2)|0; //@line 3292 "libraries/lodepng/lodepng.cpp"
     $267 = (($in) + ($266)|0); //@line 3292 "libraries/lodepng/lodepng.cpp"
     $268 = HEAP8[$267>>0]|0; //@line 3292 "libraries/lodepng/lodepng.cpp"
     $269 = ((($$344)) + 2|0); //@line 3292 "libraries/lodepng/lodepng.cpp"
     HEAP8[$269>>0] = $268; //@line 3292 "libraries/lodepng/lodepng.cpp"
     $270 = (($i$343) + 1)|0; //@line 3288 "libraries/lodepng/lodepng.cpp"
     $271 = (($$344) + ($1)|0); //@line 3288 "libraries/lodepng/lodepng.cpp"
     $272 = ($270|0)==($numpixels|0); //@line 3288 "libraries/lodepng/lodepng.cpp"
     if ($272) {
      break;
     } else {
      $$344 = $271;$i$343 = $270;
     }
    }
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
   while(1) {
    $231 = ($i$343$us*3)|0; //@line 3290 "libraries/lodepng/lodepng.cpp"
    $232 = (($in) + ($231)|0); //@line 3290 "libraries/lodepng/lodepng.cpp"
    $233 = HEAP8[$232>>0]|0; //@line 3293 "libraries/lodepng/lodepng.cpp"
    HEAP8[$$344$us>>0] = $233; //@line 3290 "libraries/lodepng/lodepng.cpp"
    $234 = (($231) + 1)|0; //@line 3291 "libraries/lodepng/lodepng.cpp"
    $235 = (($in) + ($234)|0); //@line 3291 "libraries/lodepng/lodepng.cpp"
    $236 = HEAP8[$235>>0]|0; //@line 3294 "libraries/lodepng/lodepng.cpp"
    $237 = ((($$344$us)) + 1|0); //@line 3291 "libraries/lodepng/lodepng.cpp"
    HEAP8[$237>>0] = $236; //@line 3291 "libraries/lodepng/lodepng.cpp"
    $238 = (($231) + 2)|0; //@line 3292 "libraries/lodepng/lodepng.cpp"
    $239 = (($in) + ($238)|0); //@line 3292 "libraries/lodepng/lodepng.cpp"
    $240 = HEAP8[$239>>0]|0; //@line 3292 "libraries/lodepng/lodepng.cpp"
    $241 = ((($$344$us)) + 2|0); //@line 3292 "libraries/lodepng/lodepng.cpp"
    HEAP8[$241>>0] = $240; //@line 3292 "libraries/lodepng/lodepng.cpp"
    $242 = HEAP32[$227>>2]|0; //@line 3293 "libraries/lodepng/lodepng.cpp"
    $243 = ($242|0)==(0); //@line 3293 "libraries/lodepng/lodepng.cpp"
    if ($243) {
     $253 = 0;
    } else {
     $244 = $233&255; //@line 3293 "libraries/lodepng/lodepng.cpp"
     $245 = HEAP32[$228>>2]|0; //@line 3293 "libraries/lodepng/lodepng.cpp"
     $246 = ($244|0)==($245|0); //@line 3293 "libraries/lodepng/lodepng.cpp"
     if ($246) {
      $247 = $236&255; //@line 3294 "libraries/lodepng/lodepng.cpp"
      $248 = HEAP32[$229>>2]|0; //@line 3294 "libraries/lodepng/lodepng.cpp"
      $249 = ($247|0)==($248|0); //@line 3294 "libraries/lodepng/lodepng.cpp"
      if ($249) {
       $250 = $240&255; //@line 3294 "libraries/lodepng/lodepng.cpp"
       $251 = HEAP32[$230>>2]|0; //@line 3294 "libraries/lodepng/lodepng.cpp"
       $252 = ($250|0)==($251|0); //@line 3294 "libraries/lodepng/lodepng.cpp"
       $253 = $252;
      } else {
       $253 = 0;
      }
     } else {
      $253 = 0;
     }
    }
    $not$1$us = $253 ^ 1; //@line 3293 "libraries/lodepng/lodepng.cpp"
    $254 = $not$1$us << 31 >> 31; //@line 3293 "libraries/lodepng/lodepng.cpp"
    $255 = ((($$344$us)) + 3|0); //@line 3293 "libraries/lodepng/lodepng.cpp"
    HEAP8[$255>>0] = $254; //@line 3293 "libraries/lodepng/lodepng.cpp"
    $256 = (($i$343$us) + 1)|0; //@line 3288 "libraries/lodepng/lodepng.cpp"
    $257 = (($$344$us) + ($1)|0); //@line 3288 "libraries/lodepng/lodepng.cpp"
    $258 = ($256|0)==($numpixels|0); //@line 3288 "libraries/lodepng/lodepng.cpp"
    if ($258) {
     break;
    } else {
     $$344$us = $257;$i$343$us = $256;
    }
   }
   return; //@line 3378 "libraries/lodepng/lodepng.cpp"
  } else {
   if ($173) {
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
   $174 = ((($mode)) + 16|0); //@line 3304 "libraries/lodepng/lodepng.cpp"
   $175 = ((($mode)) + 20|0); //@line 3305 "libraries/lodepng/lodepng.cpp"
   $176 = ((($mode)) + 24|0); //@line 3306 "libraries/lodepng/lodepng.cpp"
   $177 = ((($mode)) + 28|0); //@line 3307 "libraries/lodepng/lodepng.cpp"
   if ($0) {
    $$447$us = $buffer;$i$446$us = 0;
   } else {
    $$447 = $buffer;$i$446 = 0;
    while(1) {
     $273 = ($i$446*6)|0; //@line 3301 "libraries/lodepng/lodepng.cpp"
     $274 = (($in) + ($273)|0); //@line 3301 "libraries/lodepng/lodepng.cpp"
     $275 = HEAP8[$274>>0]|0; //@line 3301 "libraries/lodepng/lodepng.cpp"
     HEAP8[$$447>>0] = $275; //@line 3301 "libraries/lodepng/lodepng.cpp"
     $276 = (($273) + 2)|0; //@line 3302 "libraries/lodepng/lodepng.cpp"
     $277 = (($in) + ($276)|0); //@line 3302 "libraries/lodepng/lodepng.cpp"
     $278 = HEAP8[$277>>0]|0; //@line 3302 "libraries/lodepng/lodepng.cpp"
     $279 = ((($$447)) + 1|0); //@line 3302 "libraries/lodepng/lodepng.cpp"
     HEAP8[$279>>0] = $278; //@line 3302 "libraries/lodepng/lodepng.cpp"
     $280 = (($273) + 4)|0; //@line 3303 "libraries/lodepng/lodepng.cpp"
     $281 = (($in) + ($280)|0); //@line 3303 "libraries/lodepng/lodepng.cpp"
     $282 = HEAP8[$281>>0]|0; //@line 3303 "libraries/lodepng/lodepng.cpp"
     $283 = ((($$447)) + 2|0); //@line 3303 "libraries/lodepng/lodepng.cpp"
     HEAP8[$283>>0] = $282; //@line 3303 "libraries/lodepng/lodepng.cpp"
     $284 = (($i$446) + 1)|0; //@line 3299 "libraries/lodepng/lodepng.cpp"
     $285 = (($$447) + ($1)|0); //@line 3299 "libraries/lodepng/lodepng.cpp"
     $286 = ($284|0)==($numpixels|0); //@line 3299 "libraries/lodepng/lodepng.cpp"
     if ($286) {
      break;
     } else {
      $$447 = $285;$i$446 = $284;
     }
    }
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
   while(1) {
    $178 = ($i$446$us*6)|0; //@line 3301 "libraries/lodepng/lodepng.cpp"
    $179 = (($in) + ($178)|0); //@line 3301 "libraries/lodepng/lodepng.cpp"
    $180 = HEAP8[$179>>0]|0; //@line 3301 "libraries/lodepng/lodepng.cpp"
    HEAP8[$$447$us>>0] = $180; //@line 3301 "libraries/lodepng/lodepng.cpp"
    $181 = (($178) + 2)|0; //@line 3302 "libraries/lodepng/lodepng.cpp"
    $182 = (($in) + ($181)|0); //@line 3302 "libraries/lodepng/lodepng.cpp"
    $183 = HEAP8[$182>>0]|0; //@line 3302 "libraries/lodepng/lodepng.cpp"
    $184 = ((($$447$us)) + 1|0); //@line 3302 "libraries/lodepng/lodepng.cpp"
    HEAP8[$184>>0] = $183; //@line 3302 "libraries/lodepng/lodepng.cpp"
    $185 = (($178) + 4)|0; //@line 3303 "libraries/lodepng/lodepng.cpp"
    $186 = (($in) + ($185)|0); //@line 3303 "libraries/lodepng/lodepng.cpp"
    $187 = HEAP8[$186>>0]|0; //@line 3303 "libraries/lodepng/lodepng.cpp"
    $188 = ((($$447$us)) + 2|0); //@line 3303 "libraries/lodepng/lodepng.cpp"
    HEAP8[$188>>0] = $187; //@line 3303 "libraries/lodepng/lodepng.cpp"
    $189 = HEAP32[$174>>2]|0; //@line 3304 "libraries/lodepng/lodepng.cpp"
    $190 = ($189|0)==(0); //@line 3304 "libraries/lodepng/lodepng.cpp"
    if ($190) {
     $221 = 0;
    } else {
     $191 = HEAP8[$179>>0]|0; //@line 3305 "libraries/lodepng/lodepng.cpp"
     $192 = $191&255; //@line 3305 "libraries/lodepng/lodepng.cpp"
     $193 = $192 << 8; //@line 3305 "libraries/lodepng/lodepng.cpp"
     $194 = $178 | 1; //@line 3305 "libraries/lodepng/lodepng.cpp"
     $195 = (($in) + ($194)|0); //@line 3305 "libraries/lodepng/lodepng.cpp"
     $196 = HEAP8[$195>>0]|0; //@line 3305 "libraries/lodepng/lodepng.cpp"
     $197 = $196&255; //@line 3305 "libraries/lodepng/lodepng.cpp"
     $198 = $193 | $197; //@line 3305 "libraries/lodepng/lodepng.cpp"
     $199 = HEAP32[$175>>2]|0; //@line 3305 "libraries/lodepng/lodepng.cpp"
     $200 = ($198|0)==($199|0); //@line 3305 "libraries/lodepng/lodepng.cpp"
     if ($200) {
      $201 = HEAP8[$182>>0]|0; //@line 3306 "libraries/lodepng/lodepng.cpp"
      $202 = $201&255; //@line 3306 "libraries/lodepng/lodepng.cpp"
      $203 = $202 << 8; //@line 3306 "libraries/lodepng/lodepng.cpp"
      $204 = (($178) + 3)|0; //@line 3306 "libraries/lodepng/lodepng.cpp"
      $205 = (($in) + ($204)|0); //@line 3306 "libraries/lodepng/lodepng.cpp"
      $206 = HEAP8[$205>>0]|0; //@line 3306 "libraries/lodepng/lodepng.cpp"
      $207 = $206&255; //@line 3306 "libraries/lodepng/lodepng.cpp"
      $208 = $203 | $207; //@line 3306 "libraries/lodepng/lodepng.cpp"
      $209 = HEAP32[$176>>2]|0; //@line 3306 "libraries/lodepng/lodepng.cpp"
      $210 = ($208|0)==($209|0); //@line 3306 "libraries/lodepng/lodepng.cpp"
      if ($210) {
       $211 = HEAP8[$186>>0]|0; //@line 3307 "libraries/lodepng/lodepng.cpp"
       $212 = $211&255; //@line 3307 "libraries/lodepng/lodepng.cpp"
       $213 = $212 << 8; //@line 3307 "libraries/lodepng/lodepng.cpp"
       $214 = (($178) + 5)|0; //@line 3307 "libraries/lodepng/lodepng.cpp"
       $215 = (($in) + ($214)|0); //@line 3307 "libraries/lodepng/lodepng.cpp"
       $216 = HEAP8[$215>>0]|0; //@line 3307 "libraries/lodepng/lodepng.cpp"
       $217 = $216&255; //@line 3307 "libraries/lodepng/lodepng.cpp"
       $218 = $213 | $217; //@line 3307 "libraries/lodepng/lodepng.cpp"
       $219 = HEAP32[$177>>2]|0; //@line 3307 "libraries/lodepng/lodepng.cpp"
       $220 = ($218|0)==($219|0); //@line 3307 "libraries/lodepng/lodepng.cpp"
       $221 = $220;
      } else {
       $221 = 0;
      }
     } else {
      $221 = 0;
     }
    }
    $not$$us = $221 ^ 1; //@line 3304 "libraries/lodepng/lodepng.cpp"
    $222 = $not$$us << 31 >> 31; //@line 3304 "libraries/lodepng/lodepng.cpp"
    $223 = ((($$447$us)) + 3|0); //@line 3304 "libraries/lodepng/lodepng.cpp"
    HEAP8[$223>>0] = $222; //@line 3304 "libraries/lodepng/lodepng.cpp"
    $224 = (($i$446$us) + 1)|0; //@line 3299 "libraries/lodepng/lodepng.cpp"
    $225 = (($$447$us) + ($1)|0); //@line 3299 "libraries/lodepng/lodepng.cpp"
    $226 = ($224|0)==($numpixels|0); //@line 3299 "libraries/lodepng/lodepng.cpp"
    if ($226) {
     break;
    } else {
     $$447$us = $225;$i$446$us = $224;
    }
   }
   return; //@line 3378 "libraries/lodepng/lodepng.cpp"
  }
  break;
 }
 case 3:  {
  $3 = ($numpixels|0)==(0); //@line 3315 "libraries/lodepng/lodepng.cpp"
  if ($3) {
   return; //@line 3378 "libraries/lodepng/lodepng.cpp"
  }
  $4 = ((($mode)) + 4|0); //@line 3317 "libraries/lodepng/lodepng.cpp"
  $5 = ((($mode)) + 12|0); //@line 3320 "libraries/lodepng/lodepng.cpp"
  $6 = ((($mode)) + 8|0); //@line 3329 "libraries/lodepng/lodepng.cpp"
  if ($0) {
   $$551$us = $buffer;$i$550$us = 0;$j1$049$us = 0;
   while(1) {
    $7 = HEAP32[$4>>2]|0; //@line 3317 "libraries/lodepng/lodepng.cpp"
    switch ($7|0) {
    case 8:  {
     $8 = (($in) + ($i$550$us)|0); //@line 3317 "libraries/lodepng/lodepng.cpp"
     $9 = HEAP8[$8>>0]|0; //@line 3317 "libraries/lodepng/lodepng.cpp"
     $10 = $9&255; //@line 3317 "libraries/lodepng/lodepng.cpp"
     $index$0$us = $10;$j1$2$us = $j1$049$us;
     break;
    }
    case 0:  {
     $index$0$us = 0;$j1$2$us = $j1$049$us;
     break;
    }
    default: {
     $i$01$i5$us = (($7) + -1)|0; //@line 2360 "libraries/lodepng/lodepng.cpp"
     $12 = $j1$049$us;$i$03$i8$us = $i$01$i5$us;$result$02$i9$us = 0;
     while(1) {
      $11 = $12 >>> 3; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $13 = (($in) + ($11)|0); //@line 2351 "libraries/lodepng/lodepng.cpp"
      $14 = HEAP8[$13>>0]|0; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $15 = $14&255; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $16 = $12 & 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $17 = $16 ^ 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $18 = $15 >>> $17; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $19 = $18 & 1; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $20 = (($12) + 1)|0; //@line 2352 "libraries/lodepng/lodepng.cpp"
      $21 = $19 << $i$03$i8$us; //@line 2362 "libraries/lodepng/lodepng.cpp"
      $22 = (($21) + ($result$02$i9$us))|0; //@line 2362 "libraries/lodepng/lodepng.cpp"
      $i$0$i10$us = (($i$03$i8$us) + -1)|0; //@line 2360 "libraries/lodepng/lodepng.cpp"
      $23 = ($i$0$i10$us>>>0)<($7>>>0); //@line 2360 "libraries/lodepng/lodepng.cpp"
      if ($23) {
       $12 = $20;$i$03$i8$us = $i$0$i10$us;$result$02$i9$us = $22;
      } else {
       $index$0$us = $22;$j1$2$us = $20;
       break;
      }
     }
    }
    }
    $24 = HEAP32[$5>>2]|0; //@line 3320 "libraries/lodepng/lodepng.cpp"
    $25 = ($index$0$us>>>0)<($24>>>0); //@line 3320 "libraries/lodepng/lodepng.cpp"
    if ($25) {
     $28 = $index$0$us << 2; //@line 3329 "libraries/lodepng/lodepng.cpp"
     $29 = HEAP32[$6>>2]|0; //@line 3329 "libraries/lodepng/lodepng.cpp"
     $30 = (($29) + ($28)|0); //@line 3329 "libraries/lodepng/lodepng.cpp"
     $31 = HEAP8[$30>>0]|0; //@line 3329 "libraries/lodepng/lodepng.cpp"
     HEAP8[$$551$us>>0] = $31; //@line 3329 "libraries/lodepng/lodepng.cpp"
     $32 = $28 | 1; //@line 3330 "libraries/lodepng/lodepng.cpp"
     $33 = HEAP32[$6>>2]|0; //@line 3330 "libraries/lodepng/lodepng.cpp"
     $34 = (($33) + ($32)|0); //@line 3330 "libraries/lodepng/lodepng.cpp"
     $35 = HEAP8[$34>>0]|0; //@line 3330 "libraries/lodepng/lodepng.cpp"
     $36 = ((($$551$us)) + 1|0); //@line 3330 "libraries/lodepng/lodepng.cpp"
     HEAP8[$36>>0] = $35; //@line 3330 "libraries/lodepng/lodepng.cpp"
     $37 = $28 | 2; //@line 3331 "libraries/lodepng/lodepng.cpp"
     $38 = HEAP32[$6>>2]|0; //@line 3331 "libraries/lodepng/lodepng.cpp"
     $39 = (($38) + ($37)|0); //@line 3331 "libraries/lodepng/lodepng.cpp"
     $40 = HEAP8[$39>>0]|0; //@line 3331 "libraries/lodepng/lodepng.cpp"
     $41 = ((($$551$us)) + 2|0); //@line 3331 "libraries/lodepng/lodepng.cpp"
     HEAP8[$41>>0] = $40; //@line 3331 "libraries/lodepng/lodepng.cpp"
     $42 = $28 | 3; //@line 3332 "libraries/lodepng/lodepng.cpp"
     $43 = HEAP32[$6>>2]|0; //@line 3332 "libraries/lodepng/lodepng.cpp"
     $44 = (($43) + ($42)|0); //@line 3332 "libraries/lodepng/lodepng.cpp"
     $45 = HEAP8[$44>>0]|0; //@line 3332 "libraries/lodepng/lodepng.cpp"
     $$sink = $45;
    } else {
     $26 = ((($$551$us)) + 2|0); //@line 3324 "libraries/lodepng/lodepng.cpp"
     HEAP8[$26>>0] = 0; //@line 3324 "libraries/lodepng/lodepng.cpp"
     $27 = ((($$551$us)) + 1|0); //@line 3324 "libraries/lodepng/lodepng.cpp"
     HEAP8[$27>>0] = 0; //@line 3324 "libraries/lodepng/lodepng.cpp"
     HEAP8[$$551$us>>0] = 0; //@line 3324 "libraries/lodepng/lodepng.cpp"
     $$sink = -1;
    }
    $46 = ((($$551$us)) + 3|0); //@line 3332 "libraries/lodepng/lodepng.cpp"
    HEAP8[$46>>0] = $$sink; //@line 3332 "libraries/lodepng/lodepng.cpp"
    $47 = (($i$550$us) + 1)|0; //@line 3315 "libraries/lodepng/lodepng.cpp"
    $48 = (($$551$us) + ($1)|0); //@line 3315 "libraries/lodepng/lodepng.cpp"
    $49 = ($47|0)==($numpixels|0); //@line 3315 "libraries/lodepng/lodepng.cpp"
    if ($49) {
     break;
    } else {
     $$551$us = $48;$i$550$us = $47;$j1$049$us = $j1$2$us;
    }
   }
   return; //@line 3378 "libraries/lodepng/lodepng.cpp"
  } else {
   $$551 = $buffer;$i$550 = 0;$j1$049 = 0;
   while(1) {
    $287 = HEAP32[$4>>2]|0; //@line 3317 "libraries/lodepng/lodepng.cpp"
    switch ($287|0) {
    case 8:  {
     $288 = (($in) + ($i$550)|0); //@line 3317 "libraries/lodepng/lodepng.cpp"
     $289 = HEAP8[$288>>0]|0; //@line 3317 "libraries/lodepng/lodepng.cpp"
     $290 = $289&255; //@line 3317 "libraries/lodepng/lodepng.cpp"
     $index$0 = $290;$j1$2 = $j1$049;
     break;
    }
    case 0:  {
     $index$0 = 0;$j1$2 = $j1$049;
     break;
    }
    default: {
     $i$01$i5 = (($287) + -1)|0; //@line 2360 "libraries/lodepng/lodepng.cpp"
     $292 = $j1$049;$i$03$i8 = $i$01$i5;$result$02$i9 = 0;
     while(1) {
      $291 = $292 >>> 3; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $293 = (($in) + ($291)|0); //@line 2351 "libraries/lodepng/lodepng.cpp"
      $294 = HEAP8[$293>>0]|0; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $295 = $294&255; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $296 = $292 & 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $297 = $296 ^ 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $298 = $295 >>> $297; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $299 = $298 & 1; //@line 2351 "libraries/lodepng/lodepng.cpp"
      $300 = (($292) + 1)|0; //@line 2352 "libraries/lodepng/lodepng.cpp"
      $301 = $299 << $i$03$i8; //@line 2362 "libraries/lodepng/lodepng.cpp"
      $302 = (($301) + ($result$02$i9))|0; //@line 2362 "libraries/lodepng/lodepng.cpp"
      $i$0$i10 = (($i$03$i8) + -1)|0; //@line 2360 "libraries/lodepng/lodepng.cpp"
      $303 = ($i$0$i10>>>0)<($287>>>0); //@line 2360 "libraries/lodepng/lodepng.cpp"
      if ($303) {
       $292 = $300;$i$03$i8 = $i$0$i10;$result$02$i9 = $302;
      } else {
       $index$0 = $302;$j1$2 = $300;
       break;
      }
     }
    }
    }
    $304 = HEAP32[$5>>2]|0; //@line 3320 "libraries/lodepng/lodepng.cpp"
    $305 = ($index$0>>>0)<($304>>>0); //@line 3320 "libraries/lodepng/lodepng.cpp"
    if ($305) {
     $307 = $index$0 << 2; //@line 3329 "libraries/lodepng/lodepng.cpp"
     $308 = HEAP32[$6>>2]|0; //@line 3329 "libraries/lodepng/lodepng.cpp"
     $309 = (($308) + ($307)|0); //@line 3329 "libraries/lodepng/lodepng.cpp"
     $310 = HEAP8[$309>>0]|0; //@line 3329 "libraries/lodepng/lodepng.cpp"
     HEAP8[$$551>>0] = $310; //@line 3329 "libraries/lodepng/lodepng.cpp"
     $311 = $307 | 1; //@line 3330 "libraries/lodepng/lodepng.cpp"
     $312 = HEAP32[$6>>2]|0; //@line 3330 "libraries/lodepng/lodepng.cpp"
     $313 = (($312) + ($311)|0); //@line 3330 "libraries/lodepng/lodepng.cpp"
     $314 = HEAP8[$313>>0]|0; //@line 3330 "libraries/lodepng/lodepng.cpp"
     $315 = ((($$551)) + 1|0); //@line 3330 "libraries/lodepng/lodepng.cpp"
     HEAP8[$315>>0] = $314; //@line 3330 "libraries/lodepng/lodepng.cpp"
     $316 = $307 | 2; //@line 3331 "libraries/lodepng/lodepng.cpp"
     $317 = HEAP32[$6>>2]|0; //@line 3331 "libraries/lodepng/lodepng.cpp"
     $318 = (($317) + ($316)|0); //@line 3331 "libraries/lodepng/lodepng.cpp"
     $319 = HEAP8[$318>>0]|0; //@line 3331 "libraries/lodepng/lodepng.cpp"
     $$sink91 = $319;
    } else {
     $306 = ((($$551)) + 1|0); //@line 3324 "libraries/lodepng/lodepng.cpp"
     HEAP8[$306>>0] = 0; //@line 3324 "libraries/lodepng/lodepng.cpp"
     HEAP8[$$551>>0] = 0; //@line 3324 "libraries/lodepng/lodepng.cpp"
     $$sink91 = 0;
    }
    $320 = ((($$551)) + 2|0); //@line 3331 "libraries/lodepng/lodepng.cpp"
    HEAP8[$320>>0] = $$sink91; //@line 3331 "libraries/lodepng/lodepng.cpp"
    $321 = (($i$550) + 1)|0; //@line 3315 "libraries/lodepng/lodepng.cpp"
    $322 = (($$551) + ($1)|0); //@line 3315 "libraries/lodepng/lodepng.cpp"
    $323 = ($321|0)==($numpixels|0); //@line 3315 "libraries/lodepng/lodepng.cpp"
    if ($323) {
     break;
    } else {
     $$551 = $322;$i$550 = $321;$j1$049 = $j1$2;
    }
   }
   return; //@line 3378 "libraries/lodepng/lodepng.cpp"
  }
  break;
 }
 case 4:  {
  $324 = ((($mode)) + 4|0); //@line 3338 "libraries/lodepng/lodepng.cpp"
  $325 = HEAP32[$324>>2]|0; //@line 3338 "libraries/lodepng/lodepng.cpp"
  $326 = ($325|0)==(8); //@line 3338 "libraries/lodepng/lodepng.cpp"
  $327 = ($numpixels|0)==(0); //@line 3340 "libraries/lodepng/lodepng.cpp"
  if ($326) {
   if ($327) {
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
   if ($0) {
    $$654$us = $buffer;$i$653$us = 0;
    while(1) {
     $340 = $i$653$us << 1; //@line 3342 "libraries/lodepng/lodepng.cpp"
     $341 = (($in) + ($340)|0); //@line 3342 "libraries/lodepng/lodepng.cpp"
     $342 = HEAP8[$341>>0]|0; //@line 3342 "libraries/lodepng/lodepng.cpp"
     $343 = ((($$654$us)) + 2|0); //@line 3342 "libraries/lodepng/lodepng.cpp"
     HEAP8[$343>>0] = $342; //@line 3342 "libraries/lodepng/lodepng.cpp"
     $344 = ((($$654$us)) + 1|0); //@line 3342 "libraries/lodepng/lodepng.cpp"
     HEAP8[$344>>0] = $342; //@line 3342 "libraries/lodepng/lodepng.cpp"
     HEAP8[$$654$us>>0] = $342; //@line 3342 "libraries/lodepng/lodepng.cpp"
     $345 = $340 | 1; //@line 3343 "libraries/lodepng/lodepng.cpp"
     $346 = (($in) + ($345)|0); //@line 3343 "libraries/lodepng/lodepng.cpp"
     $347 = HEAP8[$346>>0]|0; //@line 3343 "libraries/lodepng/lodepng.cpp"
     $348 = ((($$654$us)) + 3|0); //@line 3343 "libraries/lodepng/lodepng.cpp"
     HEAP8[$348>>0] = $347; //@line 3343 "libraries/lodepng/lodepng.cpp"
     $349 = (($i$653$us) + 1)|0; //@line 3340 "libraries/lodepng/lodepng.cpp"
     $350 = (($$654$us) + ($1)|0); //@line 3340 "libraries/lodepng/lodepng.cpp"
     $351 = ($349|0)==($numpixels|0); //@line 3340 "libraries/lodepng/lodepng.cpp"
     if ($351) {
      break;
     } else {
      $$654$us = $350;$i$653$us = $349;
     }
    }
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   } else {
    $$654 = $buffer;$i$653 = 0;
    while(1) {
     $352 = $i$653 << 1; //@line 3342 "libraries/lodepng/lodepng.cpp"
     $353 = (($in) + ($352)|0); //@line 3342 "libraries/lodepng/lodepng.cpp"
     $354 = HEAP8[$353>>0]|0; //@line 3342 "libraries/lodepng/lodepng.cpp"
     $355 = ((($$654)) + 2|0); //@line 3342 "libraries/lodepng/lodepng.cpp"
     HEAP8[$355>>0] = $354; //@line 3342 "libraries/lodepng/lodepng.cpp"
     $356 = ((($$654)) + 1|0); //@line 3342 "libraries/lodepng/lodepng.cpp"
     HEAP8[$356>>0] = $354; //@line 3342 "libraries/lodepng/lodepng.cpp"
     HEAP8[$$654>>0] = $354; //@line 3342 "libraries/lodepng/lodepng.cpp"
     $357 = (($i$653) + 1)|0; //@line 3340 "libraries/lodepng/lodepng.cpp"
     $358 = (($$654) + ($1)|0); //@line 3340 "libraries/lodepng/lodepng.cpp"
     $359 = ($357|0)==($numpixels|0); //@line 3340 "libraries/lodepng/lodepng.cpp"
     if ($359) {
      break;
     } else {
      $$654 = $358;$i$653 = $357;
     }
    }
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
  } else {
   if ($327) {
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
   if ($0) {
    $$757$us = $buffer;$i$756$us = 0;
    while(1) {
     $328 = $i$756$us << 2; //@line 3350 "libraries/lodepng/lodepng.cpp"
     $329 = (($in) + ($328)|0); //@line 3350 "libraries/lodepng/lodepng.cpp"
     $330 = HEAP8[$329>>0]|0; //@line 3350 "libraries/lodepng/lodepng.cpp"
     $331 = ((($$757$us)) + 2|0); //@line 3350 "libraries/lodepng/lodepng.cpp"
     HEAP8[$331>>0] = $330; //@line 3350 "libraries/lodepng/lodepng.cpp"
     $332 = ((($$757$us)) + 1|0); //@line 3350 "libraries/lodepng/lodepng.cpp"
     HEAP8[$332>>0] = $330; //@line 3350 "libraries/lodepng/lodepng.cpp"
     HEAP8[$$757$us>>0] = $330; //@line 3350 "libraries/lodepng/lodepng.cpp"
     $333 = $328 | 2; //@line 3351 "libraries/lodepng/lodepng.cpp"
     $334 = (($in) + ($333)|0); //@line 3351 "libraries/lodepng/lodepng.cpp"
     $335 = HEAP8[$334>>0]|0; //@line 3351 "libraries/lodepng/lodepng.cpp"
     $336 = ((($$757$us)) + 3|0); //@line 3351 "libraries/lodepng/lodepng.cpp"
     HEAP8[$336>>0] = $335; //@line 3351 "libraries/lodepng/lodepng.cpp"
     $337 = (($i$756$us) + 1)|0; //@line 3348 "libraries/lodepng/lodepng.cpp"
     $338 = (($$757$us) + ($1)|0); //@line 3348 "libraries/lodepng/lodepng.cpp"
     $339 = ($337|0)==($numpixels|0); //@line 3348 "libraries/lodepng/lodepng.cpp"
     if ($339) {
      break;
     } else {
      $$757$us = $338;$i$756$us = $337;
     }
    }
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   } else {
    $$757 = $buffer;$i$756 = 0;
    while(1) {
     $360 = $i$756 << 2; //@line 3350 "libraries/lodepng/lodepng.cpp"
     $361 = (($in) + ($360)|0); //@line 3350 "libraries/lodepng/lodepng.cpp"
     $362 = HEAP8[$361>>0]|0; //@line 3350 "libraries/lodepng/lodepng.cpp"
     $363 = ((($$757)) + 2|0); //@line 3350 "libraries/lodepng/lodepng.cpp"
     HEAP8[$363>>0] = $362; //@line 3350 "libraries/lodepng/lodepng.cpp"
     $364 = ((($$757)) + 1|0); //@line 3350 "libraries/lodepng/lodepng.cpp"
     HEAP8[$364>>0] = $362; //@line 3350 "libraries/lodepng/lodepng.cpp"
     HEAP8[$$757>>0] = $362; //@line 3350 "libraries/lodepng/lodepng.cpp"
     $365 = (($i$756) + 1)|0; //@line 3348 "libraries/lodepng/lodepng.cpp"
     $366 = (($$757) + ($1)|0); //@line 3348 "libraries/lodepng/lodepng.cpp"
     $367 = ($365|0)==($numpixels|0); //@line 3348 "libraries/lodepng/lodepng.cpp"
     if ($367) {
      break;
     } else {
      $$757 = $366;$i$756 = $365;
     }
    }
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
  }
  break;
 }
 case 6:  {
  $368 = ((($mode)) + 4|0); //@line 3357 "libraries/lodepng/lodepng.cpp"
  $369 = HEAP32[$368>>2]|0; //@line 3357 "libraries/lodepng/lodepng.cpp"
  $370 = ($369|0)==(8); //@line 3357 "libraries/lodepng/lodepng.cpp"
  $371 = ($numpixels|0)==(0); //@line 3359 "libraries/lodepng/lodepng.cpp"
  if ($370) {
   if ($371) {
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
   if ($0) {
    $$860$us = $buffer;$i$859$us = 0;
    while(1) {
     $390 = $i$859$us << 2; //@line 3361 "libraries/lodepng/lodepng.cpp"
     $391 = (($in) + ($390)|0); //@line 3361 "libraries/lodepng/lodepng.cpp"
     $392 = HEAP8[$391>>0]|0; //@line 3361 "libraries/lodepng/lodepng.cpp"
     HEAP8[$$860$us>>0] = $392; //@line 3361 "libraries/lodepng/lodepng.cpp"
     $393 = $390 | 1; //@line 3362 "libraries/lodepng/lodepng.cpp"
     $394 = (($in) + ($393)|0); //@line 3362 "libraries/lodepng/lodepng.cpp"
     $395 = HEAP8[$394>>0]|0; //@line 3362 "libraries/lodepng/lodepng.cpp"
     $396 = ((($$860$us)) + 1|0); //@line 3362 "libraries/lodepng/lodepng.cpp"
     HEAP8[$396>>0] = $395; //@line 3362 "libraries/lodepng/lodepng.cpp"
     $397 = $390 | 2; //@line 3363 "libraries/lodepng/lodepng.cpp"
     $398 = (($in) + ($397)|0); //@line 3363 "libraries/lodepng/lodepng.cpp"
     $399 = HEAP8[$398>>0]|0; //@line 3363 "libraries/lodepng/lodepng.cpp"
     $400 = ((($$860$us)) + 2|0); //@line 3363 "libraries/lodepng/lodepng.cpp"
     HEAP8[$400>>0] = $399; //@line 3363 "libraries/lodepng/lodepng.cpp"
     $401 = $390 | 3; //@line 3364 "libraries/lodepng/lodepng.cpp"
     $402 = (($in) + ($401)|0); //@line 3364 "libraries/lodepng/lodepng.cpp"
     $403 = HEAP8[$402>>0]|0; //@line 3364 "libraries/lodepng/lodepng.cpp"
     $404 = ((($$860$us)) + 3|0); //@line 3364 "libraries/lodepng/lodepng.cpp"
     HEAP8[$404>>0] = $403; //@line 3364 "libraries/lodepng/lodepng.cpp"
     $405 = (($i$859$us) + 1)|0; //@line 3359 "libraries/lodepng/lodepng.cpp"
     $406 = (($$860$us) + ($1)|0); //@line 3359 "libraries/lodepng/lodepng.cpp"
     $407 = ($405|0)==($numpixels|0); //@line 3359 "libraries/lodepng/lodepng.cpp"
     if ($407) {
      break;
     } else {
      $$860$us = $406;$i$859$us = $405;
     }
    }
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   } else {
    $$860 = $buffer;$i$859 = 0;
    while(1) {
     $408 = $i$859 << 2; //@line 3361 "libraries/lodepng/lodepng.cpp"
     $409 = (($in) + ($408)|0); //@line 3361 "libraries/lodepng/lodepng.cpp"
     $410 = HEAP8[$409>>0]|0; //@line 3361 "libraries/lodepng/lodepng.cpp"
     HEAP8[$$860>>0] = $410; //@line 3361 "libraries/lodepng/lodepng.cpp"
     $411 = $408 | 1; //@line 3362 "libraries/lodepng/lodepng.cpp"
     $412 = (($in) + ($411)|0); //@line 3362 "libraries/lodepng/lodepng.cpp"
     $413 = HEAP8[$412>>0]|0; //@line 3362 "libraries/lodepng/lodepng.cpp"
     $414 = ((($$860)) + 1|0); //@line 3362 "libraries/lodepng/lodepng.cpp"
     HEAP8[$414>>0] = $413; //@line 3362 "libraries/lodepng/lodepng.cpp"
     $415 = $408 | 2; //@line 3363 "libraries/lodepng/lodepng.cpp"
     $416 = (($in) + ($415)|0); //@line 3363 "libraries/lodepng/lodepng.cpp"
     $417 = HEAP8[$416>>0]|0; //@line 3363 "libraries/lodepng/lodepng.cpp"
     $418 = ((($$860)) + 2|0); //@line 3363 "libraries/lodepng/lodepng.cpp"
     HEAP8[$418>>0] = $417; //@line 3363 "libraries/lodepng/lodepng.cpp"
     $419 = (($i$859) + 1)|0; //@line 3359 "libraries/lodepng/lodepng.cpp"
     $420 = (($$860) + ($1)|0); //@line 3359 "libraries/lodepng/lodepng.cpp"
     $421 = ($419|0)==($numpixels|0); //@line 3359 "libraries/lodepng/lodepng.cpp"
     if ($421) {
      break;
     } else {
      $$860 = $420;$i$859 = $419;
     }
    }
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
  } else {
   if ($371) {
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
   if ($0) {
    $$963$us = $buffer;$i$962$us = 0;
    while(1) {
     $372 = $i$962$us << 3; //@line 3371 "libraries/lodepng/lodepng.cpp"
     $373 = (($in) + ($372)|0); //@line 3371 "libraries/lodepng/lodepng.cpp"
     $374 = HEAP8[$373>>0]|0; //@line 3371 "libraries/lodepng/lodepng.cpp"
     HEAP8[$$963$us>>0] = $374; //@line 3371 "libraries/lodepng/lodepng.cpp"
     $375 = $372 | 2; //@line 3372 "libraries/lodepng/lodepng.cpp"
     $376 = (($in) + ($375)|0); //@line 3372 "libraries/lodepng/lodepng.cpp"
     $377 = HEAP8[$376>>0]|0; //@line 3372 "libraries/lodepng/lodepng.cpp"
     $378 = ((($$963$us)) + 1|0); //@line 3372 "libraries/lodepng/lodepng.cpp"
     HEAP8[$378>>0] = $377; //@line 3372 "libraries/lodepng/lodepng.cpp"
     $379 = $372 | 4; //@line 3373 "libraries/lodepng/lodepng.cpp"
     $380 = (($in) + ($379)|0); //@line 3373 "libraries/lodepng/lodepng.cpp"
     $381 = HEAP8[$380>>0]|0; //@line 3373 "libraries/lodepng/lodepng.cpp"
     $382 = ((($$963$us)) + 2|0); //@line 3373 "libraries/lodepng/lodepng.cpp"
     HEAP8[$382>>0] = $381; //@line 3373 "libraries/lodepng/lodepng.cpp"
     $383 = $372 | 6; //@line 3374 "libraries/lodepng/lodepng.cpp"
     $384 = (($in) + ($383)|0); //@line 3374 "libraries/lodepng/lodepng.cpp"
     $385 = HEAP8[$384>>0]|0; //@line 3374 "libraries/lodepng/lodepng.cpp"
     $386 = ((($$963$us)) + 3|0); //@line 3374 "libraries/lodepng/lodepng.cpp"
     HEAP8[$386>>0] = $385; //@line 3374 "libraries/lodepng/lodepng.cpp"
     $387 = (($i$962$us) + 1)|0; //@line 3369 "libraries/lodepng/lodepng.cpp"
     $388 = (($$963$us) + ($1)|0); //@line 3369 "libraries/lodepng/lodepng.cpp"
     $389 = ($387|0)==($numpixels|0); //@line 3369 "libraries/lodepng/lodepng.cpp"
     if ($389) {
      break;
     } else {
      $$963$us = $388;$i$962$us = $387;
     }
    }
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   } else {
    $$963 = $buffer;$i$962 = 0;
    while(1) {
     $422 = $i$962 << 3; //@line 3371 "libraries/lodepng/lodepng.cpp"
     $423 = (($in) + ($422)|0); //@line 3371 "libraries/lodepng/lodepng.cpp"
     $424 = HEAP8[$423>>0]|0; //@line 3371 "libraries/lodepng/lodepng.cpp"
     HEAP8[$$963>>0] = $424; //@line 3371 "libraries/lodepng/lodepng.cpp"
     $425 = $422 | 2; //@line 3372 "libraries/lodepng/lodepng.cpp"
     $426 = (($in) + ($425)|0); //@line 3372 "libraries/lodepng/lodepng.cpp"
     $427 = HEAP8[$426>>0]|0; //@line 3372 "libraries/lodepng/lodepng.cpp"
     $428 = ((($$963)) + 1|0); //@line 3372 "libraries/lodepng/lodepng.cpp"
     HEAP8[$428>>0] = $427; //@line 3372 "libraries/lodepng/lodepng.cpp"
     $429 = $422 | 4; //@line 3373 "libraries/lodepng/lodepng.cpp"
     $430 = (($in) + ($429)|0); //@line 3373 "libraries/lodepng/lodepng.cpp"
     $431 = HEAP8[$430>>0]|0; //@line 3373 "libraries/lodepng/lodepng.cpp"
     $432 = ((($$963)) + 2|0); //@line 3373 "libraries/lodepng/lodepng.cpp"
     HEAP8[$432>>0] = $431; //@line 3373 "libraries/lodepng/lodepng.cpp"
     $433 = (($i$962) + 1)|0; //@line 3369 "libraries/lodepng/lodepng.cpp"
     $434 = (($$963) + ($1)|0); //@line 3369 "libraries/lodepng/lodepng.cpp"
     $435 = ($433|0)==($numpixels|0); //@line 3369 "libraries/lodepng/lodepng.cpp"
     if ($435) {
      break;
     } else {
      $$963 = $434;$i$962 = $433;
     }
    }
    return; //@line 3378 "libraries/lodepng/lodepng.cpp"
   }
  }
  break;
 }
 default: {
  return; //@line 3378 "libraries/lodepng/lodepng.cpp"
 }
 }
}
function __ZL18getPixelColorRGBA8PhS_S_S_PKhjPK16LodePNGColorMode($r,$g,$b,$a,$in,$i,$mode) {
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 $in = $in|0;
 $i = $i|0;
 $mode = $mode|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$0$i = 0, $i$0$i7 = 0, $i$01$i = 0, $i$01$i2 = 0, $i$03$i = 0, $i$03$i5 = 0, $index$0 = 0, $result$0$lcssa$i = 0, $result$02$i = 0, $result$02$i6 = 0;
 var $storemerge = 0, $storemerge1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$mode>>2]|0; //@line 3143 "libraries/lodepng/lodepng.cpp"
 switch ($0|0) {
 case 0:  {
  $1 = ((($mode)) + 4|0); //@line 3145 "libraries/lodepng/lodepng.cpp"
  $2 = HEAP32[$1>>2]|0; //@line 3145 "libraries/lodepng/lodepng.cpp"
  switch ($2|0) {
  case 8:  {
   $3 = (($in) + ($i)|0); //@line 3147 "libraries/lodepng/lodepng.cpp"
   $4 = HEAP8[$3>>0]|0; //@line 3147 "libraries/lodepng/lodepng.cpp"
   HEAP8[$b>>0] = $4; //@line 3147 "libraries/lodepng/lodepng.cpp"
   HEAP8[$g>>0] = $4; //@line 3147 "libraries/lodepng/lodepng.cpp"
   HEAP8[$r>>0] = $4; //@line 3147 "libraries/lodepng/lodepng.cpp"
   $5 = ((($mode)) + 16|0); //@line 3148 "libraries/lodepng/lodepng.cpp"
   $6 = HEAP32[$5>>2]|0; //@line 3148 "libraries/lodepng/lodepng.cpp"
   $7 = ($6|0)==(0); //@line 3148 "libraries/lodepng/lodepng.cpp"
   if (!($7)) {
    $8 = $4&255; //@line 3148 "libraries/lodepng/lodepng.cpp"
    $9 = ((($mode)) + 20|0); //@line 3148 "libraries/lodepng/lodepng.cpp"
    $10 = HEAP32[$9>>2]|0; //@line 3148 "libraries/lodepng/lodepng.cpp"
    $11 = ($8|0)==($10|0); //@line 3148 "libraries/lodepng/lodepng.cpp"
    if ($11) {
     HEAP8[$a>>0] = 0; //@line 3148 "libraries/lodepng/lodepng.cpp"
     return; //@line 3241 "libraries/lodepng/lodepng.cpp"
    }
   }
   HEAP8[$a>>0] = -1; //@line 3149 "libraries/lodepng/lodepng.cpp"
   return; //@line 3241 "libraries/lodepng/lodepng.cpp"
   break;
  }
  case 16:  {
   $12 = $i << 1; //@line 3153 "libraries/lodepng/lodepng.cpp"
   $13 = (($in) + ($12)|0); //@line 3153 "libraries/lodepng/lodepng.cpp"
   $14 = HEAP8[$13>>0]|0; //@line 3153 "libraries/lodepng/lodepng.cpp"
   HEAP8[$b>>0] = $14; //@line 3153 "libraries/lodepng/lodepng.cpp"
   HEAP8[$g>>0] = $14; //@line 3153 "libraries/lodepng/lodepng.cpp"
   HEAP8[$r>>0] = $14; //@line 3153 "libraries/lodepng/lodepng.cpp"
   $15 = ((($mode)) + 16|0); //@line 3154 "libraries/lodepng/lodepng.cpp"
   $16 = HEAP32[$15>>2]|0; //@line 3154 "libraries/lodepng/lodepng.cpp"
   $17 = ($16|0)==(0); //@line 3154 "libraries/lodepng/lodepng.cpp"
   if (!($17)) {
    $18 = HEAP8[$13>>0]|0; //@line 3154 "libraries/lodepng/lodepng.cpp"
    $19 = $18&255; //@line 3154 "libraries/lodepng/lodepng.cpp"
    $20 = $19 << 8; //@line 3154 "libraries/lodepng/lodepng.cpp"
    $21 = $12 | 1; //@line 3154 "libraries/lodepng/lodepng.cpp"
    $22 = (($in) + ($21)|0); //@line 3154 "libraries/lodepng/lodepng.cpp"
    $23 = HEAP8[$22>>0]|0; //@line 3154 "libraries/lodepng/lodepng.cpp"
    $24 = $23&255; //@line 3154 "libraries/lodepng/lodepng.cpp"
    $25 = $20 | $24; //@line 3154 "libraries/lodepng/lodepng.cpp"
    $26 = ((($mode)) + 20|0); //@line 3154 "libraries/lodepng/lodepng.cpp"
    $27 = HEAP32[$26>>2]|0; //@line 3154 "libraries/lodepng/lodepng.cpp"
    $28 = ($25|0)==($27|0); //@line 3154 "libraries/lodepng/lodepng.cpp"
    if ($28) {
     HEAP8[$a>>0] = 0; //@line 3154 "libraries/lodepng/lodepng.cpp"
     return; //@line 3241 "libraries/lodepng/lodepng.cpp"
    }
   }
   HEAP8[$a>>0] = -1; //@line 3155 "libraries/lodepng/lodepng.cpp"
   return; //@line 3241 "libraries/lodepng/lodepng.cpp"
   break;
  }
  default: {
   $29 = 1 << $2; //@line 3159 "libraries/lodepng/lodepng.cpp"
   $30 = (($29) + -1)|0; //@line 3159 "libraries/lodepng/lodepng.cpp"
   $31 = ($2|0)==(0); //@line 2360 "libraries/lodepng/lodepng.cpp"
   if ($31) {
    $result$0$lcssa$i = 0;
   } else {
    $32 = Math_imul($2, $i)|0; //@line 3160 "libraries/lodepng/lodepng.cpp"
    $i$01$i = (($2) + -1)|0; //@line 2360 "libraries/lodepng/lodepng.cpp"
    $34 = $32;$i$03$i = $i$01$i;$result$02$i = 0;
    while(1) {
     $33 = $34 >>> 3; //@line 2351 "libraries/lodepng/lodepng.cpp"
     $35 = (($in) + ($33)|0); //@line 2351 "libraries/lodepng/lodepng.cpp"
     $36 = HEAP8[$35>>0]|0; //@line 2351 "libraries/lodepng/lodepng.cpp"
     $37 = $36&255; //@line 2351 "libraries/lodepng/lodepng.cpp"
     $38 = $34 & 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
     $39 = $38 ^ 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
     $40 = $37 >>> $39; //@line 2351 "libraries/lodepng/lodepng.cpp"
     $41 = $40 & 1; //@line 2351 "libraries/lodepng/lodepng.cpp"
     $42 = (($34) + 1)|0; //@line 2352 "libraries/lodepng/lodepng.cpp"
     $43 = $41 << $i$03$i; //@line 2362 "libraries/lodepng/lodepng.cpp"
     $44 = (($43) + ($result$02$i))|0; //@line 2362 "libraries/lodepng/lodepng.cpp"
     $i$0$i = (($i$03$i) + -1)|0; //@line 2360 "libraries/lodepng/lodepng.cpp"
     $45 = ($i$0$i>>>0)<($2>>>0); //@line 2360 "libraries/lodepng/lodepng.cpp"
     if ($45) {
      $34 = $42;$i$03$i = $i$0$i;$result$02$i = $44;
     } else {
      $result$0$lcssa$i = $44;
      break;
     }
    }
   }
   $46 = ($result$0$lcssa$i*255)|0; //@line 3162 "libraries/lodepng/lodepng.cpp"
   $47 = (($46>>>0) / ($30>>>0))&-1; //@line 3162 "libraries/lodepng/lodepng.cpp"
   $48 = $47&255; //@line 3162 "libraries/lodepng/lodepng.cpp"
   HEAP8[$b>>0] = $48; //@line 3162 "libraries/lodepng/lodepng.cpp"
   HEAP8[$g>>0] = $48; //@line 3162 "libraries/lodepng/lodepng.cpp"
   HEAP8[$r>>0] = $48; //@line 3162 "libraries/lodepng/lodepng.cpp"
   $49 = ((($mode)) + 16|0); //@line 3163 "libraries/lodepng/lodepng.cpp"
   $50 = HEAP32[$49>>2]|0; //@line 3163 "libraries/lodepng/lodepng.cpp"
   $51 = ($50|0)==(0); //@line 3163 "libraries/lodepng/lodepng.cpp"
   if ($51) {
    label = 16;
   } else {
    $52 = ((($mode)) + 20|0); //@line 3163 "libraries/lodepng/lodepng.cpp"
    $53 = HEAP32[$52>>2]|0; //@line 3163 "libraries/lodepng/lodepng.cpp"
    $54 = ($result$0$lcssa$i|0)==($53|0); //@line 3163 "libraries/lodepng/lodepng.cpp"
    if ($54) {
     $storemerge1 = 0;
    } else {
     label = 16;
    }
   }
   if ((label|0) == 16) {
    $storemerge1 = -1;
   }
   HEAP8[$a>>0] = $storemerge1; //@line 3163 "libraries/lodepng/lodepng.cpp"
   return; //@line 3241 "libraries/lodepng/lodepng.cpp"
  }
  }
  break;
 }
 case 2:  {
  $55 = ((($mode)) + 4|0); //@line 3169 "libraries/lodepng/lodepng.cpp"
  $56 = HEAP32[$55>>2]|0; //@line 3169 "libraries/lodepng/lodepng.cpp"
  $57 = ($56|0)==(8); //@line 3169 "libraries/lodepng/lodepng.cpp"
  if ($57) {
   $58 = ($i*3)|0; //@line 3171 "libraries/lodepng/lodepng.cpp"
   $59 = (($in) + ($58)|0); //@line 3171 "libraries/lodepng/lodepng.cpp"
   $60 = HEAP8[$59>>0]|0; //@line 3171 "libraries/lodepng/lodepng.cpp"
   HEAP8[$r>>0] = $60; //@line 3171 "libraries/lodepng/lodepng.cpp"
   $61 = (($58) + 1)|0; //@line 3171 "libraries/lodepng/lodepng.cpp"
   $62 = (($in) + ($61)|0); //@line 3171 "libraries/lodepng/lodepng.cpp"
   $63 = HEAP8[$62>>0]|0; //@line 3171 "libraries/lodepng/lodepng.cpp"
   HEAP8[$g>>0] = $63; //@line 3171 "libraries/lodepng/lodepng.cpp"
   $64 = (($58) + 2)|0; //@line 3171 "libraries/lodepng/lodepng.cpp"
   $65 = (($in) + ($64)|0); //@line 3171 "libraries/lodepng/lodepng.cpp"
   $66 = HEAP8[$65>>0]|0; //@line 3171 "libraries/lodepng/lodepng.cpp"
   HEAP8[$b>>0] = $66; //@line 3171 "libraries/lodepng/lodepng.cpp"
   $67 = ((($mode)) + 16|0); //@line 3172 "libraries/lodepng/lodepng.cpp"
   $68 = HEAP32[$67>>2]|0; //@line 3172 "libraries/lodepng/lodepng.cpp"
   $69 = ($68|0)==(0); //@line 3172 "libraries/lodepng/lodepng.cpp"
   if (!($69)) {
    $70 = HEAP8[$r>>0]|0; //@line 3172 "libraries/lodepng/lodepng.cpp"
    $71 = $70&255; //@line 3172 "libraries/lodepng/lodepng.cpp"
    $72 = ((($mode)) + 20|0); //@line 3172 "libraries/lodepng/lodepng.cpp"
    $73 = HEAP32[$72>>2]|0; //@line 3172 "libraries/lodepng/lodepng.cpp"
    $74 = ($71|0)==($73|0); //@line 3172 "libraries/lodepng/lodepng.cpp"
    if ($74) {
     $75 = HEAP8[$g>>0]|0; //@line 3172 "libraries/lodepng/lodepng.cpp"
     $76 = $75&255; //@line 3172 "libraries/lodepng/lodepng.cpp"
     $77 = ((($mode)) + 24|0); //@line 3172 "libraries/lodepng/lodepng.cpp"
     $78 = HEAP32[$77>>2]|0; //@line 3172 "libraries/lodepng/lodepng.cpp"
     $79 = ($76|0)==($78|0); //@line 3172 "libraries/lodepng/lodepng.cpp"
     if ($79) {
      $80 = $66&255; //@line 3172 "libraries/lodepng/lodepng.cpp"
      $81 = ((($mode)) + 28|0); //@line 3172 "libraries/lodepng/lodepng.cpp"
      $82 = HEAP32[$81>>2]|0; //@line 3172 "libraries/lodepng/lodepng.cpp"
      $83 = ($80|0)==($82|0); //@line 3172 "libraries/lodepng/lodepng.cpp"
      if ($83) {
       HEAP8[$a>>0] = 0; //@line 3172 "libraries/lodepng/lodepng.cpp"
       return; //@line 3241 "libraries/lodepng/lodepng.cpp"
      }
     }
    }
   }
   HEAP8[$a>>0] = -1; //@line 3173 "libraries/lodepng/lodepng.cpp"
   return; //@line 3241 "libraries/lodepng/lodepng.cpp"
  } else {
   $84 = ($i*6)|0; //@line 3177 "libraries/lodepng/lodepng.cpp"
   $85 = (($in) + ($84)|0); //@line 3177 "libraries/lodepng/lodepng.cpp"
   $86 = HEAP8[$85>>0]|0; //@line 3177 "libraries/lodepng/lodepng.cpp"
   HEAP8[$r>>0] = $86; //@line 3177 "libraries/lodepng/lodepng.cpp"
   $87 = (($84) + 2)|0; //@line 3178 "libraries/lodepng/lodepng.cpp"
   $88 = (($in) + ($87)|0); //@line 3178 "libraries/lodepng/lodepng.cpp"
   $89 = HEAP8[$88>>0]|0; //@line 3178 "libraries/lodepng/lodepng.cpp"
   HEAP8[$g>>0] = $89; //@line 3178 "libraries/lodepng/lodepng.cpp"
   $90 = (($84) + 4)|0; //@line 3179 "libraries/lodepng/lodepng.cpp"
   $91 = (($in) + ($90)|0); //@line 3179 "libraries/lodepng/lodepng.cpp"
   $92 = HEAP8[$91>>0]|0; //@line 3179 "libraries/lodepng/lodepng.cpp"
   HEAP8[$b>>0] = $92; //@line 3179 "libraries/lodepng/lodepng.cpp"
   $93 = ((($mode)) + 16|0); //@line 3180 "libraries/lodepng/lodepng.cpp"
   $94 = HEAP32[$93>>2]|0; //@line 3180 "libraries/lodepng/lodepng.cpp"
   $95 = ($94|0)==(0); //@line 3180 "libraries/lodepng/lodepng.cpp"
   if (!($95)) {
    $96 = HEAP8[$85>>0]|0; //@line 3180 "libraries/lodepng/lodepng.cpp"
    $97 = $96&255; //@line 3180 "libraries/lodepng/lodepng.cpp"
    $98 = $97 << 8; //@line 3180 "libraries/lodepng/lodepng.cpp"
    $99 = $84 | 1; //@line 3180 "libraries/lodepng/lodepng.cpp"
    $100 = (($in) + ($99)|0); //@line 3180 "libraries/lodepng/lodepng.cpp"
    $101 = HEAP8[$100>>0]|0; //@line 3180 "libraries/lodepng/lodepng.cpp"
    $102 = $101&255; //@line 3180 "libraries/lodepng/lodepng.cpp"
    $103 = $98 | $102; //@line 3180 "libraries/lodepng/lodepng.cpp"
    $104 = ((($mode)) + 20|0); //@line 3180 "libraries/lodepng/lodepng.cpp"
    $105 = HEAP32[$104>>2]|0; //@line 3180 "libraries/lodepng/lodepng.cpp"
    $106 = ($103|0)==($105|0); //@line 3180 "libraries/lodepng/lodepng.cpp"
    if ($106) {
     $107 = HEAP8[$88>>0]|0; //@line 3181 "libraries/lodepng/lodepng.cpp"
     $108 = $107&255; //@line 3181 "libraries/lodepng/lodepng.cpp"
     $109 = $108 << 8; //@line 3181 "libraries/lodepng/lodepng.cpp"
     $110 = (($84) + 3)|0; //@line 3181 "libraries/lodepng/lodepng.cpp"
     $111 = (($in) + ($110)|0); //@line 3181 "libraries/lodepng/lodepng.cpp"
     $112 = HEAP8[$111>>0]|0; //@line 3181 "libraries/lodepng/lodepng.cpp"
     $113 = $112&255; //@line 3181 "libraries/lodepng/lodepng.cpp"
     $114 = $109 | $113; //@line 3181 "libraries/lodepng/lodepng.cpp"
     $115 = ((($mode)) + 24|0); //@line 3181 "libraries/lodepng/lodepng.cpp"
     $116 = HEAP32[$115>>2]|0; //@line 3181 "libraries/lodepng/lodepng.cpp"
     $117 = ($114|0)==($116|0); //@line 3181 "libraries/lodepng/lodepng.cpp"
     if ($117) {
      $118 = HEAP8[$91>>0]|0; //@line 3182 "libraries/lodepng/lodepng.cpp"
      $119 = $118&255; //@line 3182 "libraries/lodepng/lodepng.cpp"
      $120 = $119 << 8; //@line 3182 "libraries/lodepng/lodepng.cpp"
      $121 = (($84) + 5)|0; //@line 3182 "libraries/lodepng/lodepng.cpp"
      $122 = (($in) + ($121)|0); //@line 3182 "libraries/lodepng/lodepng.cpp"
      $123 = HEAP8[$122>>0]|0; //@line 3182 "libraries/lodepng/lodepng.cpp"
      $124 = $123&255; //@line 3182 "libraries/lodepng/lodepng.cpp"
      $125 = $120 | $124; //@line 3182 "libraries/lodepng/lodepng.cpp"
      $126 = ((($mode)) + 28|0); //@line 3182 "libraries/lodepng/lodepng.cpp"
      $127 = HEAP32[$126>>2]|0; //@line 3182 "libraries/lodepng/lodepng.cpp"
      $128 = ($125|0)==($127|0); //@line 3182 "libraries/lodepng/lodepng.cpp"
      if ($128) {
       HEAP8[$a>>0] = 0; //@line 3182 "libraries/lodepng/lodepng.cpp"
       return; //@line 3241 "libraries/lodepng/lodepng.cpp"
      }
     }
    }
   }
   HEAP8[$a>>0] = -1; //@line 3183 "libraries/lodepng/lodepng.cpp"
   return; //@line 3241 "libraries/lodepng/lodepng.cpp"
  }
  break;
 }
 case 3:  {
  $129 = ((($mode)) + 4|0); //@line 3189 "libraries/lodepng/lodepng.cpp"
  $130 = HEAP32[$129>>2]|0; //@line 3189 "libraries/lodepng/lodepng.cpp"
  switch ($130|0) {
  case 8:  {
   $131 = (($in) + ($i)|0); //@line 3189 "libraries/lodepng/lodepng.cpp"
   $132 = HEAP8[$131>>0]|0; //@line 3189 "libraries/lodepng/lodepng.cpp"
   $133 = $132&255; //@line 3189 "libraries/lodepng/lodepng.cpp"
   $index$0 = $133;
   break;
  }
  case 0:  {
   $index$0 = 0;
   break;
  }
  default: {
   $134 = Math_imul($130, $i)|0; //@line 3192 "libraries/lodepng/lodepng.cpp"
   $i$01$i2 = (($130) + -1)|0; //@line 2360 "libraries/lodepng/lodepng.cpp"
   $136 = $134;$i$03$i5 = $i$01$i2;$result$02$i6 = 0;
   while(1) {
    $135 = $136 >>> 3; //@line 2351 "libraries/lodepng/lodepng.cpp"
    $137 = (($in) + ($135)|0); //@line 2351 "libraries/lodepng/lodepng.cpp"
    $138 = HEAP8[$137>>0]|0; //@line 2351 "libraries/lodepng/lodepng.cpp"
    $139 = $138&255; //@line 2351 "libraries/lodepng/lodepng.cpp"
    $140 = $136 & 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
    $141 = $140 ^ 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
    $142 = $139 >>> $141; //@line 2351 "libraries/lodepng/lodepng.cpp"
    $143 = $142 & 1; //@line 2351 "libraries/lodepng/lodepng.cpp"
    $144 = (($136) + 1)|0; //@line 2352 "libraries/lodepng/lodepng.cpp"
    $145 = $143 << $i$03$i5; //@line 2362 "libraries/lodepng/lodepng.cpp"
    $146 = (($145) + ($result$02$i6))|0; //@line 2362 "libraries/lodepng/lodepng.cpp"
    $i$0$i7 = (($i$03$i5) + -1)|0; //@line 2360 "libraries/lodepng/lodepng.cpp"
    $147 = ($i$0$i7>>>0)<($130>>>0); //@line 2360 "libraries/lodepng/lodepng.cpp"
    if ($147) {
     $136 = $144;$i$03$i5 = $i$0$i7;$result$02$i6 = $146;
    } else {
     $index$0 = $146;
     break;
    }
   }
  }
  }
  $148 = ((($mode)) + 12|0); //@line 3196 "libraries/lodepng/lodepng.cpp"
  $149 = HEAP32[$148>>2]|0; //@line 3196 "libraries/lodepng/lodepng.cpp"
  $150 = ($index$0>>>0)<($149>>>0); //@line 3196 "libraries/lodepng/lodepng.cpp"
  if ($150) {
   $151 = $index$0 << 2; //@line 3205 "libraries/lodepng/lodepng.cpp"
   $152 = ((($mode)) + 8|0); //@line 3205 "libraries/lodepng/lodepng.cpp"
   $153 = HEAP32[$152>>2]|0; //@line 3205 "libraries/lodepng/lodepng.cpp"
   $154 = (($153) + ($151)|0); //@line 3205 "libraries/lodepng/lodepng.cpp"
   $155 = HEAP8[$154>>0]|0; //@line 3205 "libraries/lodepng/lodepng.cpp"
   HEAP8[$r>>0] = $155; //@line 3205 "libraries/lodepng/lodepng.cpp"
   $156 = $151 | 1; //@line 3206 "libraries/lodepng/lodepng.cpp"
   $157 = HEAP32[$152>>2]|0; //@line 3206 "libraries/lodepng/lodepng.cpp"
   $158 = (($157) + ($156)|0); //@line 3206 "libraries/lodepng/lodepng.cpp"
   $159 = HEAP8[$158>>0]|0; //@line 3206 "libraries/lodepng/lodepng.cpp"
   HEAP8[$g>>0] = $159; //@line 3206 "libraries/lodepng/lodepng.cpp"
   $160 = $151 | 2; //@line 3207 "libraries/lodepng/lodepng.cpp"
   $161 = HEAP32[$152>>2]|0; //@line 3207 "libraries/lodepng/lodepng.cpp"
   $162 = (($161) + ($160)|0); //@line 3207 "libraries/lodepng/lodepng.cpp"
   $163 = HEAP8[$162>>0]|0; //@line 3207 "libraries/lodepng/lodepng.cpp"
   HEAP8[$b>>0] = $163; //@line 3207 "libraries/lodepng/lodepng.cpp"
   $164 = $151 | 3; //@line 3208 "libraries/lodepng/lodepng.cpp"
   $165 = HEAP32[$152>>2]|0; //@line 3208 "libraries/lodepng/lodepng.cpp"
   $166 = (($165) + ($164)|0); //@line 3208 "libraries/lodepng/lodepng.cpp"
   $167 = HEAP8[$166>>0]|0; //@line 3208 "libraries/lodepng/lodepng.cpp"
   $storemerge = $167;
  } else {
   HEAP8[$b>>0] = 0; //@line 3200 "libraries/lodepng/lodepng.cpp"
   HEAP8[$g>>0] = 0; //@line 3200 "libraries/lodepng/lodepng.cpp"
   HEAP8[$r>>0] = 0; //@line 3200 "libraries/lodepng/lodepng.cpp"
   $storemerge = -1;
  }
  HEAP8[$a>>0] = $storemerge; //@line 3201 "libraries/lodepng/lodepng.cpp"
  return; //@line 3241 "libraries/lodepng/lodepng.cpp"
  break;
 }
 case 4:  {
  $168 = ((($mode)) + 4|0); //@line 3213 "libraries/lodepng/lodepng.cpp"
  $169 = HEAP32[$168>>2]|0; //@line 3213 "libraries/lodepng/lodepng.cpp"
  $170 = ($169|0)==(8); //@line 3213 "libraries/lodepng/lodepng.cpp"
  if ($170) {
   $171 = $i << 1; //@line 3215 "libraries/lodepng/lodepng.cpp"
   $172 = (($in) + ($171)|0); //@line 3215 "libraries/lodepng/lodepng.cpp"
   $173 = HEAP8[$172>>0]|0; //@line 3215 "libraries/lodepng/lodepng.cpp"
   HEAP8[$b>>0] = $173; //@line 3215 "libraries/lodepng/lodepng.cpp"
   HEAP8[$g>>0] = $173; //@line 3215 "libraries/lodepng/lodepng.cpp"
   HEAP8[$r>>0] = $173; //@line 3215 "libraries/lodepng/lodepng.cpp"
   $174 = $171 | 1; //@line 3216 "libraries/lodepng/lodepng.cpp"
   $175 = (($in) + ($174)|0); //@line 3216 "libraries/lodepng/lodepng.cpp"
   $176 = HEAP8[$175>>0]|0; //@line 3216 "libraries/lodepng/lodepng.cpp"
   HEAP8[$a>>0] = $176; //@line 3216 "libraries/lodepng/lodepng.cpp"
   return; //@line 3241 "libraries/lodepng/lodepng.cpp"
  } else {
   $177 = $i << 2; //@line 3220 "libraries/lodepng/lodepng.cpp"
   $178 = (($in) + ($177)|0); //@line 3220 "libraries/lodepng/lodepng.cpp"
   $179 = HEAP8[$178>>0]|0; //@line 3220 "libraries/lodepng/lodepng.cpp"
   HEAP8[$b>>0] = $179; //@line 3220 "libraries/lodepng/lodepng.cpp"
   HEAP8[$g>>0] = $179; //@line 3220 "libraries/lodepng/lodepng.cpp"
   HEAP8[$r>>0] = $179; //@line 3220 "libraries/lodepng/lodepng.cpp"
   $180 = $177 | 2; //@line 3221 "libraries/lodepng/lodepng.cpp"
   $181 = (($in) + ($180)|0); //@line 3221 "libraries/lodepng/lodepng.cpp"
   $182 = HEAP8[$181>>0]|0; //@line 3221 "libraries/lodepng/lodepng.cpp"
   HEAP8[$a>>0] = $182; //@line 3221 "libraries/lodepng/lodepng.cpp"
   return; //@line 3241 "libraries/lodepng/lodepng.cpp"
  }
  break;
 }
 case 6:  {
  $183 = ((($mode)) + 4|0); //@line 3226 "libraries/lodepng/lodepng.cpp"
  $184 = HEAP32[$183>>2]|0; //@line 3226 "libraries/lodepng/lodepng.cpp"
  $185 = ($184|0)==(8); //@line 3226 "libraries/lodepng/lodepng.cpp"
  if ($185) {
   $186 = $i << 2; //@line 3228 "libraries/lodepng/lodepng.cpp"
   $187 = (($in) + ($186)|0); //@line 3228 "libraries/lodepng/lodepng.cpp"
   $188 = HEAP8[$187>>0]|0; //@line 3228 "libraries/lodepng/lodepng.cpp"
   HEAP8[$r>>0] = $188; //@line 3228 "libraries/lodepng/lodepng.cpp"
   $189 = $186 | 1; //@line 3229 "libraries/lodepng/lodepng.cpp"
   $190 = (($in) + ($189)|0); //@line 3229 "libraries/lodepng/lodepng.cpp"
   $191 = HEAP8[$190>>0]|0; //@line 3229 "libraries/lodepng/lodepng.cpp"
   HEAP8[$g>>0] = $191; //@line 3229 "libraries/lodepng/lodepng.cpp"
   $192 = $186 | 2; //@line 3230 "libraries/lodepng/lodepng.cpp"
   $193 = (($in) + ($192)|0); //@line 3230 "libraries/lodepng/lodepng.cpp"
   $194 = HEAP8[$193>>0]|0; //@line 3230 "libraries/lodepng/lodepng.cpp"
   HEAP8[$b>>0] = $194; //@line 3230 "libraries/lodepng/lodepng.cpp"
   $195 = $186 | 3; //@line 3231 "libraries/lodepng/lodepng.cpp"
   $196 = (($in) + ($195)|0); //@line 3231 "libraries/lodepng/lodepng.cpp"
   $197 = HEAP8[$196>>0]|0; //@line 3231 "libraries/lodepng/lodepng.cpp"
   HEAP8[$a>>0] = $197; //@line 3231 "libraries/lodepng/lodepng.cpp"
   return; //@line 3241 "libraries/lodepng/lodepng.cpp"
  } else {
   $198 = $i << 3; //@line 3235 "libraries/lodepng/lodepng.cpp"
   $199 = (($in) + ($198)|0); //@line 3235 "libraries/lodepng/lodepng.cpp"
   $200 = HEAP8[$199>>0]|0; //@line 3235 "libraries/lodepng/lodepng.cpp"
   HEAP8[$r>>0] = $200; //@line 3235 "libraries/lodepng/lodepng.cpp"
   $201 = $198 | 2; //@line 3236 "libraries/lodepng/lodepng.cpp"
   $202 = (($in) + ($201)|0); //@line 3236 "libraries/lodepng/lodepng.cpp"
   $203 = HEAP8[$202>>0]|0; //@line 3236 "libraries/lodepng/lodepng.cpp"
   HEAP8[$g>>0] = $203; //@line 3236 "libraries/lodepng/lodepng.cpp"
   $204 = $198 | 4; //@line 3237 "libraries/lodepng/lodepng.cpp"
   $205 = (($in) + ($204)|0); //@line 3237 "libraries/lodepng/lodepng.cpp"
   $206 = HEAP8[$205>>0]|0; //@line 3237 "libraries/lodepng/lodepng.cpp"
   HEAP8[$b>>0] = $206; //@line 3237 "libraries/lodepng/lodepng.cpp"
   $207 = $198 | 6; //@line 3238 "libraries/lodepng/lodepng.cpp"
   $208 = (($in) + ($207)|0); //@line 3238 "libraries/lodepng/lodepng.cpp"
   $209 = HEAP8[$208>>0]|0; //@line 3238 "libraries/lodepng/lodepng.cpp"
   HEAP8[$a>>0] = $209; //@line 3238 "libraries/lodepng/lodepng.cpp"
   return; //@line 3241 "libraries/lodepng/lodepng.cpp"
  }
  break;
 }
 default: {
  return; //@line 3241 "libraries/lodepng/lodepng.cpp"
 }
 }
}
function __ZL18color_tree_cleanupP9ColorTree($tree) {
 $tree = $tree|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $i$01 = 0;
 while(1) {
  $0 = (($tree) + ($i$01<<2)|0); //@line 2975 "libraries/lodepng/lodepng.cpp"
  $1 = HEAP32[$0>>2]|0; //@line 2975 "libraries/lodepng/lodepng.cpp"
  $2 = ($1|0)==(0|0); //@line 2975 "libraries/lodepng/lodepng.cpp"
  if (!($2)) {
   __ZL18color_tree_cleanupP9ColorTree($1); //@line 2977 "libraries/lodepng/lodepng.cpp"
   $3 = HEAP32[$0>>2]|0; //@line 2978 "libraries/lodepng/lodepng.cpp"
   _free($3); //@line 76 "libraries/lodepng/lodepng.cpp"
  }
  $4 = (($i$01) + 1)|0; //@line 2973 "libraries/lodepng/lodepng.cpp"
  $5 = ($4|0)==(16); //@line 2973 "libraries/lodepng/lodepng.cpp"
  if ($5) {
   break;
  } else {
   $i$01 = $4;
  }
 }
 return; //@line 2981 "libraries/lodepng/lodepng.cpp"
}
function __Z25lodepng_get_color_profileP19LodePNGColorProfilePKhjjPK16LodePNGColorMode($profile,$in,$w,$h,$mode) {
 $profile = $profile|0;
 $in = $in|0;
 $w = $w|0;
 $h = $h|0;
 $mode = $mode|0;
 var $$0$i = 0, $$0$i$i$i = 0, $$02$i = 0, $$025$i$i = 0, $$lcssa = 0, $$lcssa74 = 0, $$lobit$i$i = 0, $$mux = 0, $$not = 0, $$not65 = 0, $$not66 = 0, $$not69 = 0, $$not70 = 0, $$pre$i = 0, $$pre$phi60Z2D = 0, $$pre$phi62Z2D = 0, $$pre$phiZ2D = 0, $$pre55 = 0, $$pre58 = 0, $$pre59 = 0;
 var $$pre61 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $a = 0, $a4 = 0, $a8 = 0, $alpha_done$042 = 0;
 var $alpha_done$2 = 0, $alpha_done$228 = 0, $alpha_done$339 = 0, $alpha_done$5 = 0, $b = 0, $b3 = 0, $b7 = 0, $bit$01$i = 0, $bit$06$i$i = 0, $bits_done$037 = 0, $brmerge = 0, $brmerge67 = 0, $brmerge71 = 0, $colored_done$043 = 0, $colored_done$1 = 0, $colored_done$240 = 0, $colored_done$3 = 0, $exitcond$i = 0, $g = 0, $g2 = 0;
 var $g6 = 0, $i$01$i$i = 0, $i$049 = 0, $i$144 = 0, $i$241 = 0, $maxnumcolors$0 = 0, $numcolors_done$038 = 0, $numcolors_done$1 = 0, $or$cond = 0, $or$cond11 = 0, $or$cond13 = 0, $or$cond15 = 0, $or$cond17 = 0, $or$cond19 = 0, $or$cond21 = 0, $or$cond23 = 0, $or$cond3 = 0, $or$cond52 = 0, $or$cond68 = 0, $or$cond7 = 0;
 var $r = 0, $r1 = 0, $r5 = 0, $tree = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tree = sp;
 $r = sp + 82|0;
 $g = sp + 80|0;
 $b = sp + 78|0;
 $a = sp + 76|0;
 $r1 = sp + 74|0;
 $g2 = sp + 72|0;
 $b3 = sp + 70|0;
 $a4 = sp + 68|0;
 $r5 = sp + 87|0;
 $g6 = sp + 86|0;
 $b7 = sp + 85|0;
 $a8 = sp + 84|0;
 $0 = Math_imul($h, $w)|0; //@line 3520 "libraries/lodepng/lodepng.cpp"
 $1 = HEAP32[$mode>>2]|0; //@line 2634 "libraries/lodepng/lodepng.cpp"
 $2 = $1 | 4; //@line 2644 "libraries/lodepng/lodepng.cpp"
 $3 = ($2|0)==(4); //@line 2644 "libraries/lodepng/lodepng.cpp"
 $4 = $3&1; //@line 2644 "libraries/lodepng/lodepng.cpp"
 $5 = ((($mode)) + 16|0); //@line 2669 "libraries/lodepng/lodepng.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 2669 "libraries/lodepng/lodepng.cpp"
 $$lobit$i$i = $1 & 4; //@line 2649 "libraries/lodepng/lodepng.cpp"
 $7 = $6 | $$lobit$i$i; //@line 2670 "libraries/lodepng/lodepng.cpp"
 $8 = ($7|0)==(0); //@line 2670 "libraries/lodepng/lodepng.cpp"
 L1: do {
  if ($8) {
   $9 = ((($mode)) + 12|0); //@line 2660 "libraries/lodepng/lodepng.cpp"
   $10 = HEAP32[$9>>2]|0; //@line 2660 "libraries/lodepng/lodepng.cpp"
   $11 = ($10|0)==(0); //@line 2660 "libraries/lodepng/lodepng.cpp"
   if ($11) {
    $233 = 1;
   } else {
    $12 = ((($mode)) + 8|0); //@line 2662 "libraries/lodepng/lodepng.cpp"
    $13 = HEAP32[$12>>2]|0; //@line 2662 "libraries/lodepng/lodepng.cpp"
    $i$01$i$i = 0;
    while(1) {
     $16 = $i$01$i$i << 2; //@line 2662 "libraries/lodepng/lodepng.cpp"
     $17 = $16 | 3; //@line 2662 "libraries/lodepng/lodepng.cpp"
     $18 = (($13) + ($17)|0); //@line 2662 "libraries/lodepng/lodepng.cpp"
     $19 = HEAP8[$18>>0]|0; //@line 2662 "libraries/lodepng/lodepng.cpp"
     $20 = ($19<<24>>24)==(-1); //@line 2662 "libraries/lodepng/lodepng.cpp"
     $15 = (($i$01$i$i) + 1)|0; //@line 2660 "libraries/lodepng/lodepng.cpp"
     if (!($20)) {
      $233 = 0;
      break L1;
     }
     $14 = ($15|0)==($10|0); //@line 2660 "libraries/lodepng/lodepng.cpp"
     if ($14) {
      $233 = 1;
      break;
     } else {
      $i$01$i$i = $15;
     }
    }
   }
  } else {
   $233 = 0;
  }
 } while(0);
 $21 = ((($mode)) + 4|0); //@line 2634 "libraries/lodepng/lodepng.cpp"
 $22 = HEAP32[$21>>2]|0; //@line 3534 "libraries/lodepng/lodepng.cpp"
 switch ($1|0) {
 case 3: case 0:  {
  $$0$i$i$i = 1;
  break;
 }
 case 2:  {
  $$0$i$i$i = 3;
  break;
 }
 case 4:  {
  $$0$i$i$i = 2;
  break;
 }
 case 6:  {
  $$0$i$i$i = 4;
  break;
 }
 default: {
  $$0$i$i$i = 0;
 }
 }
 $23 = Math_imul($$0$i$i$i, $22)|0; //@line 2549 "libraries/lodepng/lodepng.cpp"
 $24 = ($23|0)==(1); //@line 3526 "libraries/lodepng/lodepng.cpp"
 $25 = $24&1; //@line 3526 "libraries/lodepng/lodepng.cpp"
 $$not = ($23>>>0)>(8); //@line 3529 "libraries/lodepng/lodepng.cpp"
 $brmerge = $$not | $24; //@line 3529 "libraries/lodepng/lodepng.cpp"
 $$mux = $$not ? 257 : 2; //@line 3529 "libraries/lodepng/lodepng.cpp"
 if ($brmerge) {
  $maxnumcolors$0 = $$mux;
 } else {
  $26 = ($23|0)==(2); //@line 3529 "libraries/lodepng/lodepng.cpp"
  if ($26) {
   $maxnumcolors$0 = 4;
  } else {
   $27 = ($23|0)==(4); //@line 3529 "libraries/lodepng/lodepng.cpp"
   $28 = $27 ? 16 : 256; //@line 3529 "libraries/lodepng/lodepng.cpp"
   $maxnumcolors$0 = $28;
  }
 }
 dest=$tree; stop=dest+64|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 2966 "libraries/lodepng/lodepng.cpp"
 $29 = ((($tree)) + 64|0); //@line 2967 "libraries/lodepng/lodepng.cpp"
 HEAP32[$29>>2] = -1; //@line 2967 "libraries/lodepng/lodepng.cpp"
 $30 = ($22|0)==(16); //@line 3534 "libraries/lodepng/lodepng.cpp"
 if ($30) {
  $31 = ($0|0)==(0); //@line 3537 "libraries/lodepng/lodepng.cpp"
  L20: do {
   if (!($31)) {
    $i$049 = 0;
    while(1) {
     __ZL19getPixelColorRGBA16PtS_S_S_PKhjPK16LodePNGColorMode($r,$g,$b,$a,$in,$i$049,$mode); //@line 3539 "libraries/lodepng/lodepng.cpp"
     $34 = HEAP16[$r>>1]|0; //@line 3540 "libraries/lodepng/lodepng.cpp"
     $35 = $34&65535; //@line 3540 "libraries/lodepng/lodepng.cpp"
     $36 = $35 & 255; //@line 3540 "libraries/lodepng/lodepng.cpp"
     $37 = $35 >>> 8; //@line 3540 "libraries/lodepng/lodepng.cpp"
     $38 = ($36|0)==($37|0); //@line 3540 "libraries/lodepng/lodepng.cpp"
     if (!($38)) {
      break;
     }
     $39 = HEAP16[$g>>1]|0; //@line 3540 "libraries/lodepng/lodepng.cpp"
     $40 = $39&65535; //@line 3540 "libraries/lodepng/lodepng.cpp"
     $41 = $40 & 255; //@line 3540 "libraries/lodepng/lodepng.cpp"
     $42 = $40 >>> 8; //@line 3540 "libraries/lodepng/lodepng.cpp"
     $43 = ($41|0)==($42|0); //@line 3540 "libraries/lodepng/lodepng.cpp"
     if (!($43)) {
      break;
     }
     $44 = HEAP16[$b>>1]|0; //@line 3541 "libraries/lodepng/lodepng.cpp"
     $45 = $44&65535; //@line 3541 "libraries/lodepng/lodepng.cpp"
     $46 = $45 & 255; //@line 3541 "libraries/lodepng/lodepng.cpp"
     $47 = $45 >>> 8; //@line 3541 "libraries/lodepng/lodepng.cpp"
     $48 = ($46|0)==($47|0); //@line 3541 "libraries/lodepng/lodepng.cpp"
     if (!($48)) {
      break;
     }
     $49 = HEAP16[$a>>1]|0; //@line 3541 "libraries/lodepng/lodepng.cpp"
     $50 = $49&65535; //@line 3541 "libraries/lodepng/lodepng.cpp"
     $51 = $50 & 255; //@line 3541 "libraries/lodepng/lodepng.cpp"
     $52 = $50 >>> 8; //@line 3541 "libraries/lodepng/lodepng.cpp"
     $53 = ($51|0)==($52|0); //@line 3541 "libraries/lodepng/lodepng.cpp"
     $33 = (($i$049) + 1)|0; //@line 3537 "libraries/lodepng/lodepng.cpp"
     if (!($53)) {
      break;
     }
     $32 = ($33|0)==($0|0); //@line 3537 "libraries/lodepng/lodepng.cpp"
     if ($32) {
      break L20;
     } else {
      $i$049 = $33;
     }
    }
    HEAP16[$r1>>1] = 0; //@line 3551 "libraries/lodepng/lodepng.cpp"
    HEAP16[$g2>>1] = 0; //@line 3551 "libraries/lodepng/lodepng.cpp"
    HEAP16[$b3>>1] = 0; //@line 3551 "libraries/lodepng/lodepng.cpp"
    HEAP16[$a4>>1] = 0; //@line 3551 "libraries/lodepng/lodepng.cpp"
    $54 = ((($profile)) + 1048|0); //@line 3552 "libraries/lodepng/lodepng.cpp"
    HEAP32[$54>>2] = 16; //@line 3552 "libraries/lodepng/lodepng.cpp"
    $55 = ((($profile)) + 8|0); //@line 3567 "libraries/lodepng/lodepng.cpp"
    $56 = ((($profile)) + 10|0); //@line 3567 "libraries/lodepng/lodepng.cpp"
    $57 = ((($profile)) + 12|0); //@line 3567 "libraries/lodepng/lodepng.cpp"
    $58 = ((($profile)) + 16|0); //@line 3570 "libraries/lodepng/lodepng.cpp"
    $59 = ((($profile)) + 4|0); //@line 3581 "libraries/lodepng/lodepng.cpp"
    $alpha_done$042 = $233;$colored_done$043 = $4;$i$144 = 0;
    while(1) {
     __ZL19getPixelColorRGBA16PtS_S_S_PKhjPK16LodePNGColorMode($r1,$g2,$b3,$a4,$in,$i$144,$mode); //@line 3557 "libraries/lodepng/lodepng.cpp"
     $60 = ($colored_done$043|0)==(0); //@line 3559 "libraries/lodepng/lodepng.cpp"
     if ($60) {
      $61 = HEAP16[$r1>>1]|0; //@line 3559 "libraries/lodepng/lodepng.cpp"
      $62 = HEAP16[$g2>>1]|0; //@line 3559 "libraries/lodepng/lodepng.cpp"
      $63 = ($61<<16>>16)==($62<<16>>16); //@line 3559 "libraries/lodepng/lodepng.cpp"
      $64 = HEAP16[$b3>>1]|0; //@line 3559 "libraries/lodepng/lodepng.cpp"
      $65 = ($61<<16>>16)==($64<<16>>16); //@line 3559 "libraries/lodepng/lodepng.cpp"
      $or$cond21 = $63 & $65; //@line 3559 "libraries/lodepng/lodepng.cpp"
      if ($or$cond21) {
       $colored_done$1 = 0;
      } else {
       HEAP32[$profile>>2] = 1; //@line 3561 "libraries/lodepng/lodepng.cpp"
       $colored_done$1 = 1;
      }
     } else {
      $colored_done$1 = $colored_done$043;
     }
     $66 = ($alpha_done$042|0)==(0); //@line 3565 "libraries/lodepng/lodepng.cpp"
     L34: do {
      if ($66) {
       $67 = HEAP16[$r1>>1]|0; //@line 3577 "libraries/lodepng/lodepng.cpp"
       $68 = HEAP32[$55>>2]|0; //@line 3567 "libraries/lodepng/lodepng.cpp"
       $69 = $68&65535; //@line 3567 "libraries/lodepng/lodepng.cpp"
       $70 = ($67<<16>>16)==($69<<16>>16); //@line 3567 "libraries/lodepng/lodepng.cpp"
       if ($70) {
        $71 = $68 >>> 16; //@line 3567 "libraries/lodepng/lodepng.cpp"
        $72 = $71&65535; //@line 3567 "libraries/lodepng/lodepng.cpp"
        $73 = HEAP16[$g2>>1]|0; //@line 3567 "libraries/lodepng/lodepng.cpp"
        $74 = ($73<<16>>16)==($72<<16>>16); //@line 3567 "libraries/lodepng/lodepng.cpp"
        if ($74) {
         $75 = HEAP16[$b3>>1]|0; //@line 3567 "libraries/lodepng/lodepng.cpp"
         $76 = HEAP16[$57>>1]|0; //@line 3567 "libraries/lodepng/lodepng.cpp"
         $77 = ($75<<16>>16)==($76<<16>>16); //@line 3567 "libraries/lodepng/lodepng.cpp"
         $81 = $77;
        } else {
         $81 = 0;
        }
       } else {
        $81 = 0;
       }
       $78 = HEAP16[$a4>>1]|0; //@line 3581 "libraries/lodepng/lodepng.cpp"
       switch ($78<<16>>16) {
       case -1:  {
        $87 = HEAP32[$59>>2]|0; //@line 3581 "libraries/lodepng/lodepng.cpp"
        $88 = ($87|0)!=(0); //@line 3581 "libraries/lodepng/lodepng.cpp"
        $or$cond3 = $81 & $88; //@line 3581 "libraries/lodepng/lodepng.cpp"
        if (!($or$cond3)) {
         $alpha_done$228 = 0;
         break L34;
        }
        HEAP32[$58>>2] = 1; //@line 3584 "libraries/lodepng/lodepng.cpp"
        $alpha_done$2 = 1;
        label = 37;
        break L34;
        break;
       }
       case 0:  {
        label = 30;
        break;
       }
       default: {
       }
       }
       if ((label|0) == 30) {
        label = 0;
        $79 = HEAP32[$59>>2]|0; //@line 3574 "libraries/lodepng/lodepng.cpp"
        $80 = ($79|0)==(0); //@line 3568 "libraries/lodepng/lodepng.cpp"
        $or$cond = $81 | $80; //@line 3568 "libraries/lodepng/lodepng.cpp"
        if ($or$cond) {
         $84 = HEAP32[$58>>2]|0; //@line 3574 "libraries/lodepng/lodepng.cpp"
         $$not65 = ($84|0)!=(0); //@line 3574 "libraries/lodepng/lodepng.cpp"
         $$not66 = $80 ^ 1; //@line 3574 "libraries/lodepng/lodepng.cpp"
         $brmerge67 = $$not65 | $$not66; //@line 3574 "libraries/lodepng/lodepng.cpp"
         if ($brmerge67) {
          $alpha_done$228 = 0;
          break;
         }
         HEAP32[$59>>2] = 1; //@line 3576 "libraries/lodepng/lodepng.cpp"
         HEAP16[$55>>1] = $67; //@line 3577 "libraries/lodepng/lodepng.cpp"
         $85 = HEAP16[$g2>>1]|0; //@line 3578 "libraries/lodepng/lodepng.cpp"
         HEAP16[$56>>1] = $85; //@line 3578 "libraries/lodepng/lodepng.cpp"
         $86 = HEAP16[$b3>>1]|0; //@line 3579 "libraries/lodepng/lodepng.cpp"
         HEAP16[$57>>1] = $86; //@line 3579 "libraries/lodepng/lodepng.cpp"
         $alpha_done$228 = 0;
         break;
        }
       }
       HEAP32[$58>>2] = 1; //@line 3570 "libraries/lodepng/lodepng.cpp"
       $82 = HEAP32[$54>>2]|0; //@line 3572 "libraries/lodepng/lodepng.cpp"
       $83 = ($82>>>0)<(8); //@line 3572 "libraries/lodepng/lodepng.cpp"
       if ($83) {
        HEAP32[$54>>2] = 8; //@line 3572 "libraries/lodepng/lodepng.cpp"
        $alpha_done$2 = 1;
        label = 37;
       } else {
        $alpha_done$2 = 1;
        label = 37;
       }
      } else {
       $alpha_done$2 = $alpha_done$042;
       label = 37;
      }
     } while(0);
     if ((label|0) == 37) {
      label = 0;
      $89 = ($alpha_done$2|0)!=(0); //@line 3589 "libraries/lodepng/lodepng.cpp"
      $90 = ($colored_done$1|0)!=(0); //@line 3589 "libraries/lodepng/lodepng.cpp"
      $or$cond7 = $90 & $89; //@line 3589 "libraries/lodepng/lodepng.cpp"
      if ($or$cond7) {
       break;
      } else {
       $alpha_done$228 = $alpha_done$2;
      }
     }
     $91 = (($i$144) + 1)|0; //@line 3555 "libraries/lodepng/lodepng.cpp"
     $92 = ($91|0)==($0|0); //@line 3555 "libraries/lodepng/lodepng.cpp"
     if ($92) {
      break;
     } else {
      $alpha_done$042 = $alpha_done$228;$colored_done$043 = $colored_done$1;$i$144 = $91;
     }
    }
    __ZL18color_tree_cleanupP9ColorTree($tree); //@line 3667 "libraries/lodepng/lodepng.cpp"
    STACKTOP = sp;return 0; //@line 3668 "libraries/lodepng/lodepng.cpp"
   }
  } while(0);
 }
 $93 = ($0|0)==(0); //@line 3594 "libraries/lodepng/lodepng.cpp"
 if ($93) {
  $$pre58 = ((($profile)) + 8|0); //@line 3662 "libraries/lodepng/lodepng.cpp"
  $$pre59 = ((($profile)) + 10|0); //@line 3663 "libraries/lodepng/lodepng.cpp"
  $$pre61 = ((($profile)) + 12|0); //@line 3664 "libraries/lodepng/lodepng.cpp"
  $$pre$phi60Z2D = $$pre59;$$pre$phi62Z2D = $$pre61;$$pre$phiZ2D = $$pre58;
 } else {
  $94 = ((($profile)) + 1048|0); //@line 3599 "libraries/lodepng/lodepng.cpp"
  $95 = ((($profile)) + 8|0); //@line 3616 "libraries/lodepng/lodepng.cpp"
  $96 = ((($profile)) + 10|0); //@line 3616 "libraries/lodepng/lodepng.cpp"
  $97 = ((($profile)) + 12|0); //@line 3616 "libraries/lodepng/lodepng.cpp"
  $98 = ((($profile)) + 16|0); //@line 3619 "libraries/lodepng/lodepng.cpp"
  $99 = ((($profile)) + 4|0); //@line 3630 "libraries/lodepng/lodepng.cpp"
  $100 = ((($profile)) + 20|0); //@line 3643 "libraries/lodepng/lodepng.cpp"
  $alpha_done$339 = $233;$bits_done$037 = $25;$colored_done$240 = $4;$i$241 = 0;$numcolors_done$038 = 0;
  while(1) {
   HEAP8[$r5>>0] = 0; //@line 3596 "libraries/lodepng/lodepng.cpp"
   HEAP8[$g6>>0] = 0; //@line 3596 "libraries/lodepng/lodepng.cpp"
   HEAP8[$b7>>0] = 0; //@line 3596 "libraries/lodepng/lodepng.cpp"
   HEAP8[$a8>>0] = 0; //@line 3596 "libraries/lodepng/lodepng.cpp"
   __ZL18getPixelColorRGBA8PhS_S_S_PKhjPK16LodePNGColorMode($r5,$g6,$b7,$a8,$in,$i$241,$mode); //@line 3597 "libraries/lodepng/lodepng.cpp"
   $101 = ($bits_done$037|0)==(0); //@line 3599 "libraries/lodepng/lodepng.cpp"
   $102 = HEAP32[$94>>2]|0; //@line 3603 "libraries/lodepng/lodepng.cpp"
   $103 = ($102>>>0)<(8); //@line 3599 "libraries/lodepng/lodepng.cpp"
   $or$cond68 = $101 & $103; //@line 3599 "libraries/lodepng/lodepng.cpp"
   if ($or$cond68) {
    $104 = HEAP8[$r5>>0]|0; //@line 3602 "libraries/lodepng/lodepng.cpp"
    switch ($104<<24>>24) {
    case -1: case 0:  {
     $$0$i = 1;
     break;
    }
    default: {
     $105 = (($104&255) % 17)&-1; //@line 3507 "libraries/lodepng/lodepng.cpp"
     $106 = ($105<<24>>24)==(0); //@line 3507 "libraries/lodepng/lodepng.cpp"
     if ($106) {
      $107 = (($104&255) % 85)&-1; //@line 3507 "libraries/lodepng/lodepng.cpp"
      $108 = ($107<<24>>24)==(0); //@line 3507 "libraries/lodepng/lodepng.cpp"
      $109 = $108 ? 2 : 4; //@line 3507 "libraries/lodepng/lodepng.cpp"
      $$0$i = $109;
     } else {
      $$0$i = 8;
     }
    }
    }
    $110 = ($$0$i>>>0)>($102>>>0); //@line 3603 "libraries/lodepng/lodepng.cpp"
    if ($110) {
     HEAP32[$94>>2] = $$0$i; //@line 3603 "libraries/lodepng/lodepng.cpp"
     $112 = $$0$i;
    } else {
     $112 = $102;
    }
   } else {
    $112 = $102;
   }
   $111 = ($112>>>0)>=($23>>>0); //@line 3605 "libraries/lodepng/lodepng.cpp"
   $113 = $111&1; //@line 3605 "libraries/lodepng/lodepng.cpp"
   $114 = ($colored_done$240|0)==(0); //@line 3607 "libraries/lodepng/lodepng.cpp"
   if ($114) {
    $115 = HEAP8[$r5>>0]|0; //@line 3607 "libraries/lodepng/lodepng.cpp"
    $116 = HEAP8[$g6>>0]|0; //@line 3607 "libraries/lodepng/lodepng.cpp"
    $117 = ($115<<24>>24)==($116<<24>>24); //@line 3607 "libraries/lodepng/lodepng.cpp"
    $118 = HEAP8[$b7>>0]|0; //@line 3607 "libraries/lodepng/lodepng.cpp"
    $119 = ($115<<24>>24)==($118<<24>>24); //@line 3607 "libraries/lodepng/lodepng.cpp"
    $or$cond23 = $117 & $119; //@line 3607 "libraries/lodepng/lodepng.cpp"
    if ($or$cond23) {
     $141 = $112;$colored_done$3 = 0;
    } else {
     HEAP32[$profile>>2] = 1; //@line 3609 "libraries/lodepng/lodepng.cpp"
     $120 = ($112>>>0)<(8); //@line 3611 "libraries/lodepng/lodepng.cpp"
     if ($120) {
      HEAP32[$94>>2] = 8; //@line 3611 "libraries/lodepng/lodepng.cpp"
      $141 = 8;$colored_done$3 = 1;
     } else {
      $141 = $112;$colored_done$3 = 1;
     }
    }
   } else {
    $141 = $112;$colored_done$3 = $colored_done$240;
   }
   $121 = ($alpha_done$339|0)==(0); //@line 3614 "libraries/lodepng/lodepng.cpp"
   L74: do {
    if ($121) {
     $122 = HEAP8[$r5>>0]|0; //@line 3626 "libraries/lodepng/lodepng.cpp"
     $123 = $122&255; //@line 3616 "libraries/lodepng/lodepng.cpp"
     $124 = HEAP32[$95>>2]|0; //@line 3616 "libraries/lodepng/lodepng.cpp"
     $125 = $124 & 65535; //@line 3616 "libraries/lodepng/lodepng.cpp"
     $126 = ($123|0)==($125|0); //@line 3616 "libraries/lodepng/lodepng.cpp"
     if ($126) {
      $127 = $124 >>> 16; //@line 3616 "libraries/lodepng/lodepng.cpp"
      $128 = HEAP8[$g6>>0]|0; //@line 3616 "libraries/lodepng/lodepng.cpp"
      $129 = $128&255; //@line 3616 "libraries/lodepng/lodepng.cpp"
      $130 = ($129|0)==($127|0); //@line 3616 "libraries/lodepng/lodepng.cpp"
      if ($130) {
       $131 = HEAP8[$b7>>0]|0; //@line 3616 "libraries/lodepng/lodepng.cpp"
       $132 = $131&255; //@line 3616 "libraries/lodepng/lodepng.cpp"
       $133 = HEAP16[$97>>1]|0; //@line 3616 "libraries/lodepng/lodepng.cpp"
       $134 = $133&65535; //@line 3616 "libraries/lodepng/lodepng.cpp"
       $135 = ($132|0)==($134|0); //@line 3616 "libraries/lodepng/lodepng.cpp"
       $139 = $135;
      } else {
       $139 = 0;
      }
     } else {
      $139 = 0;
     }
     $136 = HEAP8[$a8>>0]|0; //@line 3630 "libraries/lodepng/lodepng.cpp"
     switch ($136<<24>>24) {
     case -1:  {
      $148 = HEAP32[$99>>2]|0; //@line 3630 "libraries/lodepng/lodepng.cpp"
      $149 = ($148|0)!=(0); //@line 3630 "libraries/lodepng/lodepng.cpp"
      $or$cond13 = $139 & $149; //@line 3630 "libraries/lodepng/lodepng.cpp"
      if (!($or$cond13)) {
       $alpha_done$5 = 0;
       break L74;
      }
      HEAP32[$98>>2] = 1; //@line 3633 "libraries/lodepng/lodepng.cpp"
      $150 = ($141>>>0)<(8); //@line 3635 "libraries/lodepng/lodepng.cpp"
      if (!($150)) {
       $alpha_done$5 = 1;
       break L74;
      }
      HEAP32[$94>>2] = 8; //@line 3635 "libraries/lodepng/lodepng.cpp"
      $alpha_done$5 = 1;
      break L74;
      break;
     }
     case 0:  {
      $137 = HEAP32[$99>>2]|0; //@line 3623 "libraries/lodepng/lodepng.cpp"
      $138 = ($137|0)==(0); //@line 3617 "libraries/lodepng/lodepng.cpp"
      $or$cond11 = $139 | $138; //@line 3617 "libraries/lodepng/lodepng.cpp"
      if ($or$cond11) {
       $142 = HEAP32[$98>>2]|0; //@line 3623 "libraries/lodepng/lodepng.cpp"
       $$not69 = ($142|0)!=(0); //@line 3623 "libraries/lodepng/lodepng.cpp"
       $$not70 = $138 ^ 1; //@line 3623 "libraries/lodepng/lodepng.cpp"
       $brmerge71 = $$not69 | $$not70; //@line 3623 "libraries/lodepng/lodepng.cpp"
       if ($brmerge71) {
        $alpha_done$5 = 0;
        break L74;
       }
       HEAP32[$99>>2] = 1; //@line 3625 "libraries/lodepng/lodepng.cpp"
       $143 = $122&255; //@line 3626 "libraries/lodepng/lodepng.cpp"
       HEAP16[$95>>1] = $143; //@line 3626 "libraries/lodepng/lodepng.cpp"
       $144 = HEAP8[$g6>>0]|0; //@line 3627 "libraries/lodepng/lodepng.cpp"
       $145 = $144&255; //@line 3627 "libraries/lodepng/lodepng.cpp"
       HEAP16[$96>>1] = $145; //@line 3627 "libraries/lodepng/lodepng.cpp"
       $146 = HEAP8[$b7>>0]|0; //@line 3628 "libraries/lodepng/lodepng.cpp"
       $147 = $146&255; //@line 3628 "libraries/lodepng/lodepng.cpp"
       HEAP16[$97>>1] = $147; //@line 3628 "libraries/lodepng/lodepng.cpp"
       $alpha_done$5 = 0;
       break L74;
      }
      break;
     }
     default: {
     }
     }
     HEAP32[$98>>2] = 1; //@line 3619 "libraries/lodepng/lodepng.cpp"
     $140 = ($141>>>0)<(8); //@line 3621 "libraries/lodepng/lodepng.cpp"
     if ($140) {
      HEAP32[$94>>2] = 8; //@line 3621 "libraries/lodepng/lodepng.cpp"
      $alpha_done$5 = 1;
     } else {
      $alpha_done$5 = 1;
     }
    } else {
     $alpha_done$5 = $alpha_done$339;
    }
   } while(0);
   $151 = ($numcolors_done$038|0)==(0); //@line 3639 "libraries/lodepng/lodepng.cpp"
   do {
    if ($151) {
     $152 = HEAP8[$r5>>0]|0; //@line 3648 "libraries/lodepng/lodepng.cpp"
     $153 = HEAP8[$g6>>0]|0; //@line 3649 "libraries/lodepng/lodepng.cpp"
     $154 = HEAP8[$b7>>0]|0; //@line 3650 "libraries/lodepng/lodepng.cpp"
     $155 = HEAP8[$a8>>0]|0; //@line 3651 "libraries/lodepng/lodepng.cpp"
     $156 = $152&255; //@line 2989 "libraries/lodepng/lodepng.cpp"
     $157 = $153&255; //@line 2989 "libraries/lodepng/lodepng.cpp"
     $158 = $154&255; //@line 2989 "libraries/lodepng/lodepng.cpp"
     $159 = $155&255; //@line 2989 "libraries/lodepng/lodepng.cpp"
     $$025$i$i = $tree;$bit$06$i$i = 0;
     while(1) {
      $160 = $156 >>> $bit$06$i$i; //@line 2989 "libraries/lodepng/lodepng.cpp"
      $161 = $160 << 3; //@line 2989 "libraries/lodepng/lodepng.cpp"
      $162 = $161 & 8; //@line 2989 "libraries/lodepng/lodepng.cpp"
      $163 = $157 >>> $bit$06$i$i; //@line 2989 "libraries/lodepng/lodepng.cpp"
      $164 = $163 << 2; //@line 2989 "libraries/lodepng/lodepng.cpp"
      $165 = $164 & 4; //@line 2989 "libraries/lodepng/lodepng.cpp"
      $166 = $158 >>> $bit$06$i$i; //@line 2989 "libraries/lodepng/lodepng.cpp"
      $167 = $166 << 1; //@line 2989 "libraries/lodepng/lodepng.cpp"
      $168 = $167 & 2; //@line 2989 "libraries/lodepng/lodepng.cpp"
      $169 = $159 >>> $bit$06$i$i; //@line 2989 "libraries/lodepng/lodepng.cpp"
      $170 = $169 & 1; //@line 2989 "libraries/lodepng/lodepng.cpp"
      $171 = $165 | $170; //@line 2989 "libraries/lodepng/lodepng.cpp"
      $172 = $171 | $162; //@line 2989 "libraries/lodepng/lodepng.cpp"
      $173 = $172 | $168; //@line 2989 "libraries/lodepng/lodepng.cpp"
      $174 = (($$025$i$i) + ($173<<2)|0); //@line 2990 "libraries/lodepng/lodepng.cpp"
      $175 = HEAP32[$174>>2]|0; //@line 2990 "libraries/lodepng/lodepng.cpp"
      $176 = ($175|0)==(0|0); //@line 2990 "libraries/lodepng/lodepng.cpp"
      if ($176) {
       break;
      }
      $177 = (($bit$06$i$i) + 1)|0; //@line 2987 "libraries/lodepng/lodepng.cpp"
      $178 = ($177|0)<(8); //@line 2987 "libraries/lodepng/lodepng.cpp"
      if ($178) {
       $$025$i$i = $175;$bit$06$i$i = $177;
      } else {
       $$lcssa74 = $175;
       label = 71;
       break;
      }
     }
     if ((label|0) == 71) {
      label = 0;
      $179 = ((($$lcssa74)) + 64|0); //@line 2993 "libraries/lodepng/lodepng.cpp"
      $180 = HEAP32[$179>>2]|0; //@line 2993 "libraries/lodepng/lodepng.cpp"
      $181 = ($180|0)<(0); //@line 3641 "libraries/lodepng/lodepng.cpp"
      if (!($181)) {
       $numcolors_done$1 = 0;
       break;
      }
     }
     $182 = HEAP32[$100>>2]|0; //@line 3644 "libraries/lodepng/lodepng.cpp"
     $$02$i = $tree;$bit$01$i = 0;
     while(1) {
      $183 = $156 >>> $bit$01$i; //@line 3011 "libraries/lodepng/lodepng.cpp"
      $184 = $183 << 3; //@line 3011 "libraries/lodepng/lodepng.cpp"
      $185 = $184 & 8; //@line 3011 "libraries/lodepng/lodepng.cpp"
      $186 = $157 >>> $bit$01$i; //@line 3011 "libraries/lodepng/lodepng.cpp"
      $187 = $186 << 2; //@line 3011 "libraries/lodepng/lodepng.cpp"
      $188 = $187 & 4; //@line 3011 "libraries/lodepng/lodepng.cpp"
      $189 = $158 >>> $bit$01$i; //@line 3011 "libraries/lodepng/lodepng.cpp"
      $190 = $189 << 1; //@line 3011 "libraries/lodepng/lodepng.cpp"
      $191 = $190 & 2; //@line 3011 "libraries/lodepng/lodepng.cpp"
      $192 = $159 >>> $bit$01$i; //@line 3011 "libraries/lodepng/lodepng.cpp"
      $193 = $192 & 1; //@line 3011 "libraries/lodepng/lodepng.cpp"
      $194 = $188 | $193; //@line 3011 "libraries/lodepng/lodepng.cpp"
      $195 = $194 | $185; //@line 3011 "libraries/lodepng/lodepng.cpp"
      $196 = $195 | $191; //@line 3011 "libraries/lodepng/lodepng.cpp"
      $197 = (($$02$i) + ($196<<2)|0); //@line 3012 "libraries/lodepng/lodepng.cpp"
      $198 = HEAP32[$197>>2]|0; //@line 3012 "libraries/lodepng/lodepng.cpp"
      $199 = ($198|0)==(0|0); //@line 3012 "libraries/lodepng/lodepng.cpp"
      if ($199) {
       $200 = (_malloc(68)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
       HEAP32[$197>>2] = $200; //@line 3014 "libraries/lodepng/lodepng.cpp"
       dest=$200; stop=dest+64|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 2966 "libraries/lodepng/lodepng.cpp"
       $201 = ((($200)) + 64|0); //@line 2967 "libraries/lodepng/lodepng.cpp"
       HEAP32[$201>>2] = -1; //@line 2967 "libraries/lodepng/lodepng.cpp"
       $$pre$i = HEAP32[$197>>2]|0; //@line 3017 "libraries/lodepng/lodepng.cpp"
       $234 = $$pre$i;
      } else {
       $234 = $198;
      }
      $202 = (($bit$01$i) + 1)|0; //@line 3009 "libraries/lodepng/lodepng.cpp"
      $exitcond$i = ($202|0)==(8); //@line 3009 "libraries/lodepng/lodepng.cpp"
      if ($exitcond$i) {
       $$lcssa = $234;
       break;
      } else {
       $$02$i = $234;$bit$01$i = $202;
      }
     }
     $203 = ((($$lcssa)) + 64|0); //@line 3019 "libraries/lodepng/lodepng.cpp"
     HEAP32[$203>>2] = $182; //@line 3019 "libraries/lodepng/lodepng.cpp"
     $204 = ($182>>>0)<(256); //@line 3644 "libraries/lodepng/lodepng.cpp"
     if ($204) {
      $205 = $182 << 2; //@line 3648 "libraries/lodepng/lodepng.cpp"
      $206 = (((($profile)) + 24|0) + ($205)|0); //@line 3648 "libraries/lodepng/lodepng.cpp"
      HEAP8[$206>>0] = $152; //@line 3648 "libraries/lodepng/lodepng.cpp"
      $207 = $205 | 1; //@line 3649 "libraries/lodepng/lodepng.cpp"
      $208 = (((($profile)) + 24|0) + ($207)|0); //@line 3649 "libraries/lodepng/lodepng.cpp"
      HEAP8[$208>>0] = $153; //@line 3649 "libraries/lodepng/lodepng.cpp"
      $209 = $205 | 2; //@line 3650 "libraries/lodepng/lodepng.cpp"
      $210 = (((($profile)) + 24|0) + ($209)|0); //@line 3650 "libraries/lodepng/lodepng.cpp"
      HEAP8[$210>>0] = $154; //@line 3650 "libraries/lodepng/lodepng.cpp"
      $211 = $205 | 3; //@line 3651 "libraries/lodepng/lodepng.cpp"
      $212 = (((($profile)) + 24|0) + ($211)|0); //@line 3651 "libraries/lodepng/lodepng.cpp"
      HEAP8[$212>>0] = $155; //@line 3651 "libraries/lodepng/lodepng.cpp"
      $$pre55 = HEAP32[$100>>2]|0; //@line 3653 "libraries/lodepng/lodepng.cpp"
      $214 = $$pre55;
     } else {
      $214 = $182;
     }
     $213 = (($214) + 1)|0; //@line 3653 "libraries/lodepng/lodepng.cpp"
     HEAP32[$100>>2] = $213; //@line 3653 "libraries/lodepng/lodepng.cpp"
     $215 = ($213>>>0)>=($maxnumcolors$0>>>0); //@line 3654 "libraries/lodepng/lodepng.cpp"
     $216 = $215&1; //@line 3654 "libraries/lodepng/lodepng.cpp"
     $numcolors_done$1 = $216;
    } else {
     $numcolors_done$1 = $numcolors_done$038;
    }
   } while(0);
   $217 = ($alpha_done$5|0)!=(0); //@line 3658 "libraries/lodepng/lodepng.cpp"
   $218 = ($numcolors_done$1|0)!=(0); //@line 3658 "libraries/lodepng/lodepng.cpp"
   $or$cond15 = $217 & $218; //@line 3658 "libraries/lodepng/lodepng.cpp"
   $219 = ($colored_done$3|0)!=(0); //@line 3658 "libraries/lodepng/lodepng.cpp"
   $or$cond17 = $219 & $or$cond15; //@line 3658 "libraries/lodepng/lodepng.cpp"
   $or$cond19 = $111 & $or$cond17; //@line 3658 "libraries/lodepng/lodepng.cpp"
   $220 = (($i$241) + 1)|0; //@line 3594 "libraries/lodepng/lodepng.cpp"
   $221 = ($220|0)==($0|0); //@line 3594 "libraries/lodepng/lodepng.cpp"
   $or$cond52 = $or$cond19 | $221;
   if ($or$cond52) {
    $$pre$phi60Z2D = $96;$$pre$phi62Z2D = $97;$$pre$phiZ2D = $95;
    break;
   } else {
    $alpha_done$339 = $alpha_done$5;$bits_done$037 = $113;$colored_done$240 = $colored_done$3;$i$241 = $220;$numcolors_done$038 = $numcolors_done$1;
   }
  }
 }
 $222 = HEAP32[$$pre$phiZ2D>>2]|0; //@line 3662 "libraries/lodepng/lodepng.cpp"
 $223 = $222 & 65535; //@line 3662 "libraries/lodepng/lodepng.cpp"
 $224 = ($223*257)|0; //@line 3662 "libraries/lodepng/lodepng.cpp"
 $225 = $224&65535; //@line 3662 "libraries/lodepng/lodepng.cpp"
 HEAP16[$$pre$phiZ2D>>1] = $225; //@line 3662 "libraries/lodepng/lodepng.cpp"
 $226 = $222 >>> 16; //@line 3663 "libraries/lodepng/lodepng.cpp"
 $227 = ($226*257)|0; //@line 3663 "libraries/lodepng/lodepng.cpp"
 $228 = $227&65535; //@line 3663 "libraries/lodepng/lodepng.cpp"
 HEAP16[$$pre$phi60Z2D>>1] = $228; //@line 3663 "libraries/lodepng/lodepng.cpp"
 $229 = HEAP16[$$pre$phi62Z2D>>1]|0; //@line 3664 "libraries/lodepng/lodepng.cpp"
 $230 = $229&65535; //@line 3664 "libraries/lodepng/lodepng.cpp"
 $231 = ($230*257)|0; //@line 3664 "libraries/lodepng/lodepng.cpp"
 $232 = $231&65535; //@line 3664 "libraries/lodepng/lodepng.cpp"
 HEAP16[$$pre$phi62Z2D>>1] = $232; //@line 3664 "libraries/lodepng/lodepng.cpp"
 __ZL18color_tree_cleanupP9ColorTree($tree); //@line 3667 "libraries/lodepng/lodepng.cpp"
 STACKTOP = sp;return 0; //@line 3668 "libraries/lodepng/lodepng.cpp"
}
function __Z25lodepng_auto_choose_colorP16LodePNGColorModePKhjjPKS_($mode_out,$image,$w,$h,$mode_in) {
 $mode_out = $mode_out|0;
 $image = $image|0;
 $w = $w|0;
 $h = $h|0;
 $mode_in = $mode_in|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $$pre12 = 0, $$pre13 = 0, $$pre14 = 0, $$pre16 = 0, $$pre18 = 0, $$pre2$i$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $error$1 = 0, $error$2 = 0, $i$09 = 0, $notlhs = 0, $notrhs = 0, $or$cond = 0, $or$cond3 = 0, $or$cond7$not = 0, $or$cond8 = 0, $phitmp = 0, $phitmp$i = 0, $phitmp3$i = 0, $prof = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1056|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $prof = sp;
 $0 = ((($prof)) + 16|0); //@line 3483 "libraries/lodepng/lodepng.cpp"
 HEAP32[$0>>2] = 0; //@line 3483 "libraries/lodepng/lodepng.cpp"
 $1 = ((($prof)) + 20|0); //@line 3485 "libraries/lodepng/lodepng.cpp"
 HEAP32[$1>>2] = 0; //@line 3485 "libraries/lodepng/lodepng.cpp"
 $2 = ((($prof)) + 1048|0); //@line 3486 "libraries/lodepng/lodepng.cpp"
 ;HEAP32[$prof>>2]=0|0;HEAP32[$prof+4>>2]=0|0;HEAP32[$prof+8>>2]=0|0;HEAP16[$prof+12>>1]=0|0; //@line 3482 "libraries/lodepng/lodepng.cpp"
 HEAP32[$2>>2] = 1; //@line 3486 "libraries/lodepng/lodepng.cpp"
 (__Z25lodepng_get_color_profileP19LodePNGColorProfilePKhjjPK16LodePNGColorMode($prof,$image,$w,$h,$mode_in)|0); //@line 3685 "libraries/lodepng/lodepng.cpp"
 $3 = ((($mode_out)) + 16|0); //@line 3687 "libraries/lodepng/lodepng.cpp"
 HEAP32[$3>>2] = 0; //@line 3687 "libraries/lodepng/lodepng.cpp"
 $4 = ((($prof)) + 4|0); //@line 3689 "libraries/lodepng/lodepng.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 3727 "libraries/lodepng/lodepng.cpp"
 $6 = ($5|0)!=(0); //@line 3689 "libraries/lodepng/lodepng.cpp"
 $7 = Math_imul($h, $w)|0; //@line 3689 "libraries/lodepng/lodepng.cpp"
 $8 = ($7>>>0)<(17); //@line 3689 "libraries/lodepng/lodepng.cpp"
 $or$cond = $8 & $6; //@line 3689 "libraries/lodepng/lodepng.cpp"
 if ($or$cond) {
  HEAP32[$0>>2] = 1; //@line 3690 "libraries/lodepng/lodepng.cpp"
  $9 = HEAP32[$2>>2]|0; //@line 3691 "libraries/lodepng/lodepng.cpp"
  $10 = ($9>>>0)<(8); //@line 3691 "libraries/lodepng/lodepng.cpp"
  if ($10) {
   HEAP32[$2>>2] = 8; //@line 3691 "libraries/lodepng/lodepng.cpp"
   $13 = 1;
  } else {
   $13 = 1;
  }
 } else {
  $$pre = HEAP32[$0>>2]|0; //@line 3693 "libraries/lodepng/lodepng.cpp"
  $13 = $$pre;
 }
 $11 = HEAP32[$prof>>2]|0; //@line 3724 "libraries/lodepng/lodepng.cpp"
 $12 = $13 | $11; //@line 3693 "libraries/lodepng/lodepng.cpp"
 $14 = HEAP32[$1>>2]|0; //@line 3694 "libraries/lodepng/lodepng.cpp"
 $15 = ($14>>>0)<(3); //@line 3695 "libraries/lodepng/lodepng.cpp"
 if ($15) {
  $131 = 1;
  label = 9;
 } else {
  $16 = ($14>>>0)<(5); //@line 3695 "libraries/lodepng/lodepng.cpp"
  if ($16) {
   $131 = 2;
   label = 9;
  } else {
   $17 = ($14>>>0)<(17); //@line 3695 "libraries/lodepng/lodepng.cpp"
   $18 = $17 ? 4 : 8; //@line 3695 "libraries/lodepng/lodepng.cpp"
   $19 = ($14>>>0)<(257); //@line 3696 "libraries/lodepng/lodepng.cpp"
   if ($19) {
    $131 = $18;
    label = 9;
   } else {
    $$pre18 = $14 << 1; //@line 3697 "libraries/lodepng/lodepng.cpp"
    $$pre$phiZ2D = $$pre18;$25 = $18;$27 = 1;
   }
  }
 }
 if ((label|0) == 9) {
  $20 = $14 << 1; //@line 3696 "libraries/lodepng/lodepng.cpp"
  $21 = ($20>>>0)<($7>>>0); //@line 3696 "libraries/lodepng/lodepng.cpp"
  if ($21) {
   $22 = HEAP32[$2>>2]|0; //@line 3696 "libraries/lodepng/lodepng.cpp"
   $phitmp = ($22>>>0)>(8);
   $$pre$phiZ2D = $20;$25 = $131;$27 = $phitmp;
  } else {
   $$pre$phiZ2D = $20;$25 = $131;$27 = 1;
  }
 }
 $23 = ($7>>>0)<($$pre$phiZ2D>>>0); //@line 3697 "libraries/lodepng/lodepng.cpp"
 $24 = HEAP32[$2>>2]|0; //@line 3729 "libraries/lodepng/lodepng.cpp"
 $notlhs = ($12|0)==(0); //@line 3698 "libraries/lodepng/lodepng.cpp"
 $notrhs = ($24>>>0)<=($25>>>0); //@line 3698 "libraries/lodepng/lodepng.cpp"
 $or$cond7$not = $notlhs & $notrhs; //@line 3698 "libraries/lodepng/lodepng.cpp"
 $26 = $23 | $27; //@line 3700 "libraries/lodepng/lodepng.cpp"
 $or$cond8 = $26 | $or$cond7$not; //@line 3698 "libraries/lodepng/lodepng.cpp"
 if ($or$cond8) {
  $109 = ((($mode_out)) + 4|0); //@line 3723 "libraries/lodepng/lodepng.cpp"
  HEAP32[$109>>2] = $24; //@line 3723 "libraries/lodepng/lodepng.cpp"
  $110 = ($13|0)!=(0); //@line 3724 "libraries/lodepng/lodepng.cpp"
  $111 = ($11|0)!=(0); //@line 3724 "libraries/lodepng/lodepng.cpp"
  $112 = $111 ? 6 : 4; //@line 3724 "libraries/lodepng/lodepng.cpp"
  $113 = $111 ? 2 : 0; //@line 3725 "libraries/lodepng/lodepng.cpp"
  $114 = $110 ? $112 : $113; //@line 3724 "libraries/lodepng/lodepng.cpp"
  HEAP32[$mode_out>>2] = $114; //@line 3724 "libraries/lodepng/lodepng.cpp"
  $115 = ($13|0)==(0); //@line 3727 "libraries/lodepng/lodepng.cpp"
  $or$cond3 = $6 & $115; //@line 3727 "libraries/lodepng/lodepng.cpp"
  if (!($or$cond3)) {
   $error$2 = 0;
   STACKTOP = sp;return ($error$2|0); //@line 3738 "libraries/lodepng/lodepng.cpp"
  }
  $116 = 1 << $24; //@line 3729 "libraries/lodepng/lodepng.cpp"
  $117 = (($116) + -1)|0; //@line 3729 "libraries/lodepng/lodepng.cpp"
  $118 = ((($prof)) + 8|0); //@line 3730 "libraries/lodepng/lodepng.cpp"
  $119 = HEAP32[$118>>2]|0; //@line 3730 "libraries/lodepng/lodepng.cpp"
  $120 = $119 & 65535; //@line 3730 "libraries/lodepng/lodepng.cpp"
  $121 = $120 & $117; //@line 3730 "libraries/lodepng/lodepng.cpp"
  $122 = ((($mode_out)) + 20|0); //@line 3730 "libraries/lodepng/lodepng.cpp"
  HEAP32[$122>>2] = $121; //@line 3730 "libraries/lodepng/lodepng.cpp"
  $123 = $119 >>> 16; //@line 3731 "libraries/lodepng/lodepng.cpp"
  $124 = $123 & $117; //@line 3731 "libraries/lodepng/lodepng.cpp"
  $125 = ((($mode_out)) + 24|0); //@line 3731 "libraries/lodepng/lodepng.cpp"
  HEAP32[$125>>2] = $124; //@line 3731 "libraries/lodepng/lodepng.cpp"
  $126 = ((($prof)) + 12|0); //@line 3732 "libraries/lodepng/lodepng.cpp"
  $127 = HEAP16[$126>>1]|0; //@line 3732 "libraries/lodepng/lodepng.cpp"
  $128 = $127&65535; //@line 3732 "libraries/lodepng/lodepng.cpp"
  $129 = $128 & $117; //@line 3732 "libraries/lodepng/lodepng.cpp"
  $130 = ((($mode_out)) + 28|0); //@line 3732 "libraries/lodepng/lodepng.cpp"
  HEAP32[$130>>2] = $129; //@line 3732 "libraries/lodepng/lodepng.cpp"
  HEAP32[$3>>2] = 1; //@line 3733 "libraries/lodepng/lodepng.cpp"
  $error$2 = 0;
  STACKTOP = sp;return ($error$2|0); //@line 3738 "libraries/lodepng/lodepng.cpp"
 }
 $28 = ((($mode_out)) + 8|0); //@line 2605 "libraries/lodepng/lodepng.cpp"
 $29 = HEAP32[$28>>2]|0; //@line 2605 "libraries/lodepng/lodepng.cpp"
 $30 = ($29|0)==(0|0); //@line 2605 "libraries/lodepng/lodepng.cpp"
 if ($30) {
  $33 = $14;
 } else {
  _free($29); //@line 76 "libraries/lodepng/lodepng.cpp"
  $$pre12 = HEAP32[$1>>2]|0; //@line 3704 "libraries/lodepng/lodepng.cpp"
  $33 = $$pre12;
 }
 HEAP32[$28>>2] = 0; //@line 2606 "libraries/lodepng/lodepng.cpp"
 $31 = ((($mode_out)) + 12|0); //@line 2607 "libraries/lodepng/lodepng.cpp"
 HEAP32[$31>>2] = 0; //@line 2607 "libraries/lodepng/lodepng.cpp"
 $32 = ($33|0)==(0); //@line 3704 "libraries/lodepng/lodepng.cpp"
 L24: do {
  if ($32) {
   $error$1 = 0;
  } else {
   $132 = 0;$47 = 0;$i$09 = 0;
   while(1) {
    $34 = $i$09 << 2; //@line 3706 "libraries/lodepng/lodepng.cpp"
    $35 = (((($prof)) + 24|0) + ($34)|0); //@line 3706 "libraries/lodepng/lodepng.cpp"
    $36 = HEAP8[$35>>0]|0; //@line 3706 "libraries/lodepng/lodepng.cpp"
    $37 = $34 | 1; //@line 3706 "libraries/lodepng/lodepng.cpp"
    $38 = (((($prof)) + 24|0) + ($37)|0); //@line 3706 "libraries/lodepng/lodepng.cpp"
    $39 = HEAP8[$38>>0]|0; //@line 3706 "libraries/lodepng/lodepng.cpp"
    $40 = $34 | 2; //@line 3706 "libraries/lodepng/lodepng.cpp"
    $41 = (((($prof)) + 24|0) + ($40)|0); //@line 3706 "libraries/lodepng/lodepng.cpp"
    $42 = HEAP8[$41>>0]|0; //@line 3706 "libraries/lodepng/lodepng.cpp"
    $43 = $34 | 3; //@line 3706 "libraries/lodepng/lodepng.cpp"
    $44 = (((($prof)) + 24|0) + ($43)|0); //@line 3706 "libraries/lodepng/lodepng.cpp"
    $45 = HEAP8[$44>>0]|0; //@line 3706 "libraries/lodepng/lodepng.cpp"
    $46 = ($47|0)==(0|0); //@line 2616 "libraries/lodepng/lodepng.cpp"
    if ($46) {
     $48 = (_realloc(0,1024)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
     $49 = ($48|0)==(0|0); //@line 2620 "libraries/lodepng/lodepng.cpp"
     if ($49) {
      $error$1 = 83;
      break L24;
     }
     HEAP32[$28>>2] = $48; //@line 2621 "libraries/lodepng/lodepng.cpp"
     $$pre14 = HEAP32[$31>>2]|0; //@line 2623 "libraries/lodepng/lodepng.cpp"
     $51 = $$pre14;$53 = $48;
    } else {
     $51 = $132;$53 = $47;
    }
    $50 = $51 << 2; //@line 2623 "libraries/lodepng/lodepng.cpp"
    $52 = (($53) + ($50)|0); //@line 2623 "libraries/lodepng/lodepng.cpp"
    HEAP8[$52>>0] = $36; //@line 2623 "libraries/lodepng/lodepng.cpp"
    $54 = HEAP32[$31>>2]|0; //@line 2624 "libraries/lodepng/lodepng.cpp"
    $55 = $54 << 2; //@line 2624 "libraries/lodepng/lodepng.cpp"
    $56 = $55 | 1; //@line 2624 "libraries/lodepng/lodepng.cpp"
    $57 = HEAP32[$28>>2]|0; //@line 2624 "libraries/lodepng/lodepng.cpp"
    $58 = (($57) + ($56)|0); //@line 2624 "libraries/lodepng/lodepng.cpp"
    HEAP8[$58>>0] = $39; //@line 2624 "libraries/lodepng/lodepng.cpp"
    $59 = HEAP32[$31>>2]|0; //@line 2625 "libraries/lodepng/lodepng.cpp"
    $60 = $59 << 2; //@line 2625 "libraries/lodepng/lodepng.cpp"
    $61 = $60 | 2; //@line 2625 "libraries/lodepng/lodepng.cpp"
    $62 = HEAP32[$28>>2]|0; //@line 2625 "libraries/lodepng/lodepng.cpp"
    $63 = (($62) + ($61)|0); //@line 2625 "libraries/lodepng/lodepng.cpp"
    HEAP8[$63>>0] = $42; //@line 2625 "libraries/lodepng/lodepng.cpp"
    $64 = HEAP32[$31>>2]|0; //@line 2626 "libraries/lodepng/lodepng.cpp"
    $65 = $64 << 2; //@line 2626 "libraries/lodepng/lodepng.cpp"
    $66 = $65 | 3; //@line 2626 "libraries/lodepng/lodepng.cpp"
    $67 = HEAP32[$28>>2]|0; //@line 2626 "libraries/lodepng/lodepng.cpp"
    $68 = (($67) + ($66)|0); //@line 2626 "libraries/lodepng/lodepng.cpp"
    HEAP8[$68>>0] = $45; //@line 2626 "libraries/lodepng/lodepng.cpp"
    $69 = HEAP32[$31>>2]|0; //@line 2627 "libraries/lodepng/lodepng.cpp"
    $70 = (($69) + 1)|0; //@line 2627 "libraries/lodepng/lodepng.cpp"
    HEAP32[$31>>2] = $70; //@line 2627 "libraries/lodepng/lodepng.cpp"
    $71 = (($i$09) + 1)|0; //@line 3704 "libraries/lodepng/lodepng.cpp"
    $72 = HEAP32[$1>>2]|0; //@line 3704 "libraries/lodepng/lodepng.cpp"
    $73 = ($71|0)==($72|0); //@line 3704 "libraries/lodepng/lodepng.cpp"
    if ($73) {
     $error$1 = 0;
     break L24;
    }
    $$pre13 = HEAP32[$28>>2]|0; //@line 2616 "libraries/lodepng/lodepng.cpp"
    $132 = $70;$47 = $$pre13;$i$09 = $71;
   }
  }
 } while(0);
 HEAP32[$mode_out>>2] = 3; //@line 3710 "libraries/lodepng/lodepng.cpp"
 $74 = ((($mode_out)) + 4|0); //@line 3711 "libraries/lodepng/lodepng.cpp"
 HEAP32[$74>>2] = $25; //@line 3711 "libraries/lodepng/lodepng.cpp"
 $75 = HEAP32[$mode_in>>2]|0; //@line 3713 "libraries/lodepng/lodepng.cpp"
 $76 = ($75|0)==(3); //@line 3713 "libraries/lodepng/lodepng.cpp"
 if (!($76)) {
  $error$2 = $error$1;
  STACKTOP = sp;return ($error$2|0); //@line 3738 "libraries/lodepng/lodepng.cpp"
 }
 $77 = ((($mode_in)) + 12|0); //@line 3713 "libraries/lodepng/lodepng.cpp"
 $78 = HEAP32[$77>>2]|0; //@line 3713 "libraries/lodepng/lodepng.cpp"
 $79 = HEAP32[$31>>2]|0; //@line 3713 "libraries/lodepng/lodepng.cpp"
 $80 = ($78>>>0)<($79>>>0); //@line 3713 "libraries/lodepng/lodepng.cpp"
 if ($80) {
  $error$2 = $error$1;
  STACKTOP = sp;return ($error$2|0); //@line 3738 "libraries/lodepng/lodepng.cpp"
 }
 $81 = ((($mode_in)) + 4|0); //@line 3714 "libraries/lodepng/lodepng.cpp"
 $82 = HEAP32[$81>>2]|0; //@line 3714 "libraries/lodepng/lodepng.cpp"
 $83 = ($82|0)==($25|0); //@line 3714 "libraries/lodepng/lodepng.cpp"
 if (!($83)) {
  $error$2 = $error$1;
  STACKTOP = sp;return ($error$2|0); //@line 3738 "libraries/lodepng/lodepng.cpp"
 }
 $84 = HEAP32[$28>>2]|0; //@line 2605 "libraries/lodepng/lodepng.cpp"
 $85 = ($84|0)==(0|0); //@line 2605 "libraries/lodepng/lodepng.cpp"
 if (!($85)) {
  _free($84); //@line 76 "libraries/lodepng/lodepng.cpp"
 }
 $86 = $28; //@line 2607 "libraries/lodepng/lodepng.cpp"
 $87 = $86; //@line 2607 "libraries/lodepng/lodepng.cpp"
 HEAP32[$87>>2] = 0; //@line 2607 "libraries/lodepng/lodepng.cpp"
 $88 = (($86) + 4)|0; //@line 2607 "libraries/lodepng/lodepng.cpp"
 $89 = $88; //@line 2607 "libraries/lodepng/lodepng.cpp"
 HEAP32[$89>>2] = 0; //@line 2607 "libraries/lodepng/lodepng.cpp"
 ;HEAP32[$mode_out>>2]=HEAP32[$mode_in>>2]|0;HEAP32[$mode_out+4>>2]=HEAP32[$mode_in+4>>2]|0;HEAP32[$mode_out+8>>2]=HEAP32[$mode_in+8>>2]|0;HEAP32[$mode_out+12>>2]=HEAP32[$mode_in+12>>2]|0;HEAP32[$mode_out+16>>2]=HEAP32[$mode_in+16>>2]|0;HEAP32[$mode_out+20>>2]=HEAP32[$mode_in+20>>2]|0;HEAP32[$mode_out+24>>2]=HEAP32[$mode_in+24>>2]|0;HEAP32[$mode_out+28>>2]=HEAP32[$mode_in+28>>2]|0; //@line 2573 "libraries/lodepng/lodepng.cpp"
 $90 = ((($mode_in)) + 8|0); //@line 2574 "libraries/lodepng/lodepng.cpp"
 $91 = HEAP32[$90>>2]|0; //@line 2574 "libraries/lodepng/lodepng.cpp"
 $92 = ($91|0)==(0|0); //@line 2574 "libraries/lodepng/lodepng.cpp"
 if ($92) {
  $error$2 = $error$1;
  STACKTOP = sp;return ($error$2|0); //@line 3738 "libraries/lodepng/lodepng.cpp"
 }
 $93 = (_malloc(1024)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
 HEAP32[$28>>2] = $93; //@line 2576 "libraries/lodepng/lodepng.cpp"
 $94 = ($93|0)==(0|0); //@line 2577 "libraries/lodepng/lodepng.cpp"
 if ($94) {
  $error$2 = $error$1;
  STACKTOP = sp;return ($error$2|0); //@line 3738 "libraries/lodepng/lodepng.cpp"
 }
 $95 = HEAP32[$77>>2]|0; //@line 2577 "libraries/lodepng/lodepng.cpp"
 $phitmp$i = $95 & 1073741823; //@line 2577 "libraries/lodepng/lodepng.cpp"
 $phitmp3$i = ($phitmp$i|0)==(0); //@line 2577 "libraries/lodepng/lodepng.cpp"
 if ($phitmp3$i) {
  $error$2 = $error$1;
  STACKTOP = sp;return ($error$2|0); //@line 3738 "libraries/lodepng/lodepng.cpp"
 }
 $96 = HEAP32[$90>>2]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
 $97 = HEAP8[$96>>0]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
 HEAP8[$93>>0] = $97; //@line 2578 "libraries/lodepng/lodepng.cpp"
 $98 = ((($96)) + 1|0); //@line 2578 "libraries/lodepng/lodepng.cpp"
 $99 = HEAP8[$98>>0]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
 $100 = ((($93)) + 1|0); //@line 2578 "libraries/lodepng/lodepng.cpp"
 HEAP8[$100>>0] = $99; //@line 2578 "libraries/lodepng/lodepng.cpp"
 $102 = 2;
 while(1) {
  $$pre2$i$pre = HEAP32[$28>>2]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
  $$pre16 = HEAP32[$90>>2]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
  $101 = (($$pre16) + ($102)|0); //@line 2578 "libraries/lodepng/lodepng.cpp"
  $103 = HEAP8[$101>>0]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
  $104 = (($$pre2$i$pre) + ($102)|0); //@line 2578 "libraries/lodepng/lodepng.cpp"
  HEAP8[$104>>0] = $103; //@line 2578 "libraries/lodepng/lodepng.cpp"
  $105 = (($102) + 1)|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
  $106 = HEAP32[$77>>2]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
  $107 = $106 << 2; //@line 2578 "libraries/lodepng/lodepng.cpp"
  $108 = ($105|0)==($107|0); //@line 2578 "libraries/lodepng/lodepng.cpp"
  if ($108) {
   $error$2 = $error$1;
   break;
  } else {
   $102 = $105;
  }
 }
 STACKTOP = sp;return ($error$2|0); //@line 3738 "libraries/lodepng/lodepng.cpp"
}
function __Z15lodepng_inspectPjS_P12LodePNGStatePKhj($w,$h,$state,$in,$insize) {
 $w = $w|0;
 $h = $h|0;
 $state = $state|0;
 $in = $in|0;
 $insize = $insize|0;
 var $$0$i = 0, $$1 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $c$0$masked$i = 0, $c$01$i = 0, $exitcond$i = 0, $n$02$i = 0, $or$cond = 0, $phitmp$i = 0, $scevgep4$i$i = 0, dest = 0, label = 0, sp = 0;
 var stop = 0;
 sp = STACKTOP;
 $0 = ((($state)) + 132|0); //@line 3818 "libraries/lodepng/lodepng.cpp"
 $1 = ($insize|0)==(0); //@line 3819 "libraries/lodepng/lodepng.cpp"
 $2 = ($in|0)==(0|0); //@line 3819 "libraries/lodepng/lodepng.cpp"
 $or$cond = $2 | $1; //@line 3819 "libraries/lodepng/lodepng.cpp"
 if ($or$cond) {
  $3 = ((($state)) + 292|0); //@line 3821 "libraries/lodepng/lodepng.cpp"
  HEAP32[$3>>2] = 48; //@line 3821 "libraries/lodepng/lodepng.cpp"
  $$1 = 48;
  return ($$1|0); //@line 3875 "libraries/lodepng/lodepng.cpp"
 }
 $4 = ($insize>>>0)<(33); //@line 3823 "libraries/lodepng/lodepng.cpp"
 if ($4) {
  $5 = ((($state)) + 292|0); //@line 3825 "libraries/lodepng/lodepng.cpp"
  HEAP32[$5>>2] = 27; //@line 3825 "libraries/lodepng/lodepng.cpp"
  $$1 = 27;
  return ($$1|0); //@line 3875 "libraries/lodepng/lodepng.cpp"
 }
 __Z20lodepng_info_cleanupP11LodePNGInfo($0); //@line 3829 "libraries/lodepng/lodepng.cpp"
 $6 = ((($state)) + 160|0); //@line 2556 "libraries/lodepng/lodepng.cpp"
 $7 = ((($state)) + 144|0); //@line 2558 "libraries/lodepng/lodepng.cpp"
 ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;HEAP32[$6+12>>2]=0|0; //@line 2557 "libraries/lodepng/lodepng.cpp"
 HEAP32[$7>>2] = 6; //@line 2558 "libraries/lodepng/lodepng.cpp"
 $8 = ((($state)) + 148|0); //@line 2559 "libraries/lodepng/lodepng.cpp"
 HEAP32[$8>>2] = 8; //@line 2559 "libraries/lodepng/lodepng.cpp"
 $9 = ((($state)) + 152|0); //@line 2560 "libraries/lodepng/lodepng.cpp"
 HEAP32[$9>>2] = 0; //@line 2560 "libraries/lodepng/lodepng.cpp"
 $10 = ((($state)) + 156|0); //@line 2561 "libraries/lodepng/lodepng.cpp"
 HEAP32[$10>>2] = 0; //@line 2561 "libraries/lodepng/lodepng.cpp"
 $11 = ((($state)) + 140|0); //@line 2883 "libraries/lodepng/lodepng.cpp"
 HEAP32[$11>>2] = 0; //@line 2883 "libraries/lodepng/lodepng.cpp"
 HEAP32[$0>>2] = 0; //@line 2884 "libraries/lodepng/lodepng.cpp"
 $12 = ((($state)) + 136|0); //@line 2885 "libraries/lodepng/lodepng.cpp"
 HEAP32[$12>>2] = 0; //@line 2885 "libraries/lodepng/lodepng.cpp"
 $13 = ((($state)) + 176|0); //@line 2887 "libraries/lodepng/lodepng.cpp"
 $14 = ((($state)) + 252|0); //@line 2894 "libraries/lodepng/lodepng.cpp"
 HEAP32[$14>>2] = 0; //@line 2894 "libraries/lodepng/lodepng.cpp"
 $scevgep4$i$i = ((($state)) + 268|0); //@line 2700 "libraries/lodepng/lodepng.cpp"
 ;HEAP32[$scevgep4$i$i>>2]=0|0;HEAP32[$scevgep4$i$i+4>>2]=0|0;HEAP32[$scevgep4$i$i+8>>2]=0|0;HEAP32[$scevgep4$i$i+12>>2]=0|0;HEAP32[$scevgep4$i$i+16>>2]=0|0;HEAP32[$scevgep4$i$i+20>>2]=0|0; //@line 2701 "libraries/lodepng/lodepng.cpp"
 dest=$13; stop=dest+52|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 2887 "libraries/lodepng/lodepng.cpp"
 $15 = HEAP8[$in>>0]|0; //@line 3832 "libraries/lodepng/lodepng.cpp"
 $16 = ($15<<24>>24)==(-119); //@line 3832 "libraries/lodepng/lodepng.cpp"
 if ($16) {
  $17 = ((($in)) + 1|0); //@line 3832 "libraries/lodepng/lodepng.cpp"
  $18 = HEAP8[$17>>0]|0; //@line 3832 "libraries/lodepng/lodepng.cpp"
  $19 = ($18<<24>>24)==(80); //@line 3832 "libraries/lodepng/lodepng.cpp"
  if ($19) {
   $20 = ((($in)) + 2|0); //@line 3832 "libraries/lodepng/lodepng.cpp"
   $21 = HEAP8[$20>>0]|0; //@line 3832 "libraries/lodepng/lodepng.cpp"
   $22 = ($21<<24>>24)==(78); //@line 3832 "libraries/lodepng/lodepng.cpp"
   if ($22) {
    $23 = ((($in)) + 3|0); //@line 3832 "libraries/lodepng/lodepng.cpp"
    $24 = HEAP8[$23>>0]|0; //@line 3832 "libraries/lodepng/lodepng.cpp"
    $25 = ($24<<24>>24)==(71); //@line 3832 "libraries/lodepng/lodepng.cpp"
    if ($25) {
     $26 = ((($in)) + 4|0); //@line 3833 "libraries/lodepng/lodepng.cpp"
     $27 = HEAP8[$26>>0]|0; //@line 3833 "libraries/lodepng/lodepng.cpp"
     $28 = ($27<<24>>24)==(13); //@line 3833 "libraries/lodepng/lodepng.cpp"
     if ($28) {
      $29 = ((($in)) + 5|0); //@line 3833 "libraries/lodepng/lodepng.cpp"
      $30 = HEAP8[$29>>0]|0; //@line 3833 "libraries/lodepng/lodepng.cpp"
      $31 = ($30<<24>>24)==(10); //@line 3833 "libraries/lodepng/lodepng.cpp"
      if ($31) {
       $32 = ((($in)) + 6|0); //@line 3833 "libraries/lodepng/lodepng.cpp"
       $33 = HEAP8[$32>>0]|0; //@line 3833 "libraries/lodepng/lodepng.cpp"
       $34 = ($33<<24>>24)==(26); //@line 3833 "libraries/lodepng/lodepng.cpp"
       if ($34) {
        $35 = ((($in)) + 7|0); //@line 3833 "libraries/lodepng/lodepng.cpp"
        $36 = HEAP8[$35>>0]|0; //@line 3833 "libraries/lodepng/lodepng.cpp"
        $37 = ($36<<24>>24)==(10); //@line 3833 "libraries/lodepng/lodepng.cpp"
        if ($37) {
         $39 = ((($in)) + 12|0); //@line 3837 "libraries/lodepng/lodepng.cpp"
         $40 = HEAP8[$39>>0]|0; //@line 3837 "libraries/lodepng/lodepng.cpp"
         $41 = ($40<<24>>24)==(73); //@line 3837 "libraries/lodepng/lodepng.cpp"
         if ($41) {
          $42 = ((($in)) + 13|0); //@line 3837 "libraries/lodepng/lodepng.cpp"
          $43 = HEAP8[$42>>0]|0; //@line 3837 "libraries/lodepng/lodepng.cpp"
          $44 = ($43<<24>>24)==(72); //@line 3837 "libraries/lodepng/lodepng.cpp"
          if ($44) {
           $45 = ((($in)) + 14|0); //@line 3837 "libraries/lodepng/lodepng.cpp"
           $46 = HEAP8[$45>>0]|0; //@line 3837 "libraries/lodepng/lodepng.cpp"
           $47 = ($46<<24>>24)==(68); //@line 3837 "libraries/lodepng/lodepng.cpp"
           if ($47) {
            $48 = ((($in)) + 15|0); //@line 3837 "libraries/lodepng/lodepng.cpp"
            $49 = HEAP8[$48>>0]|0; //@line 3837 "libraries/lodepng/lodepng.cpp"
            $50 = ($49<<24>>24)==(82); //@line 3837 "libraries/lodepng/lodepng.cpp"
            if ($50) {
             $52 = ((($in)) + 16|0); //@line 3843 "libraries/lodepng/lodepng.cpp"
             $53 = HEAP8[$52>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
             $54 = $53&255; //@line 343 "libraries/lodepng/lodepng.cpp"
             $55 = $54 << 24; //@line 343 "libraries/lodepng/lodepng.cpp"
             $56 = ((($in)) + 17|0); //@line 343 "libraries/lodepng/lodepng.cpp"
             $57 = HEAP8[$56>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
             $58 = $57&255; //@line 343 "libraries/lodepng/lodepng.cpp"
             $59 = $58 << 16; //@line 343 "libraries/lodepng/lodepng.cpp"
             $60 = $59 | $55; //@line 343 "libraries/lodepng/lodepng.cpp"
             $61 = ((($in)) + 18|0); //@line 343 "libraries/lodepng/lodepng.cpp"
             $62 = HEAP8[$61>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
             $63 = $62&255; //@line 343 "libraries/lodepng/lodepng.cpp"
             $64 = $63 << 8; //@line 343 "libraries/lodepng/lodepng.cpp"
             $65 = $60 | $64; //@line 343 "libraries/lodepng/lodepng.cpp"
             $66 = ((($in)) + 19|0); //@line 343 "libraries/lodepng/lodepng.cpp"
             $67 = HEAP8[$66>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
             $68 = $67&255; //@line 343 "libraries/lodepng/lodepng.cpp"
             $69 = $65 | $68; //@line 343 "libraries/lodepng/lodepng.cpp"
             HEAP32[$w>>2] = $69; //@line 3843 "libraries/lodepng/lodepng.cpp"
             $70 = ((($in)) + 20|0); //@line 3844 "libraries/lodepng/lodepng.cpp"
             $71 = HEAP8[$70>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
             $72 = $71&255; //@line 343 "libraries/lodepng/lodepng.cpp"
             $73 = $72 << 24; //@line 343 "libraries/lodepng/lodepng.cpp"
             $74 = ((($in)) + 21|0); //@line 343 "libraries/lodepng/lodepng.cpp"
             $75 = HEAP8[$74>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
             $76 = $75&255; //@line 343 "libraries/lodepng/lodepng.cpp"
             $77 = $76 << 16; //@line 343 "libraries/lodepng/lodepng.cpp"
             $78 = $77 | $73; //@line 343 "libraries/lodepng/lodepng.cpp"
             $79 = ((($in)) + 22|0); //@line 343 "libraries/lodepng/lodepng.cpp"
             $80 = HEAP8[$79>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
             $81 = $80&255; //@line 343 "libraries/lodepng/lodepng.cpp"
             $82 = $81 << 8; //@line 343 "libraries/lodepng/lodepng.cpp"
             $83 = $78 | $82; //@line 343 "libraries/lodepng/lodepng.cpp"
             $84 = ((($in)) + 23|0); //@line 343 "libraries/lodepng/lodepng.cpp"
             $85 = HEAP8[$84>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
             $86 = $85&255; //@line 343 "libraries/lodepng/lodepng.cpp"
             $87 = $83 | $86; //@line 343 "libraries/lodepng/lodepng.cpp"
             HEAP32[$h>>2] = $87; //@line 3844 "libraries/lodepng/lodepng.cpp"
             $88 = ((($in)) + 24|0); //@line 3845 "libraries/lodepng/lodepng.cpp"
             $89 = HEAP8[$88>>0]|0; //@line 3845 "libraries/lodepng/lodepng.cpp"
             $90 = $89&255; //@line 3873 "libraries/lodepng/lodepng.cpp"
             HEAP32[$8>>2] = $90; //@line 3845 "libraries/lodepng/lodepng.cpp"
             $91 = ((($in)) + 25|0); //@line 3846 "libraries/lodepng/lodepng.cpp"
             $92 = HEAP8[$91>>0]|0; //@line 3846 "libraries/lodepng/lodepng.cpp"
             $93 = $92&255; //@line 3873 "libraries/lodepng/lodepng.cpp"
             HEAP32[$7>>2] = $93; //@line 3846 "libraries/lodepng/lodepng.cpp"
             $94 = ((($in)) + 26|0); //@line 3847 "libraries/lodepng/lodepng.cpp"
             $95 = HEAP8[$94>>0]|0; //@line 3847 "libraries/lodepng/lodepng.cpp"
             $96 = $95&255; //@line 3867 "libraries/lodepng/lodepng.cpp"
             HEAP32[$0>>2] = $96; //@line 3847 "libraries/lodepng/lodepng.cpp"
             $97 = ((($in)) + 27|0); //@line 3848 "libraries/lodepng/lodepng.cpp"
             $98 = HEAP8[$97>>0]|0; //@line 3848 "libraries/lodepng/lodepng.cpp"
             $99 = $98&255; //@line 3869 "libraries/lodepng/lodepng.cpp"
             HEAP32[$12>>2] = $99; //@line 3848 "libraries/lodepng/lodepng.cpp"
             $100 = ((($in)) + 28|0); //@line 3849 "libraries/lodepng/lodepng.cpp"
             $101 = HEAP8[$100>>0]|0; //@line 3849 "libraries/lodepng/lodepng.cpp"
             $102 = $101&255; //@line 3871 "libraries/lodepng/lodepng.cpp"
             HEAP32[$11>>2] = $102; //@line 3849 "libraries/lodepng/lodepng.cpp"
             $103 = HEAP32[$w>>2]|0; //@line 3851 "libraries/lodepng/lodepng.cpp"
             $104 = ($103|0)==(0); //@line 3851 "libraries/lodepng/lodepng.cpp"
             if (!($104)) {
              $105 = HEAP32[$h>>2]|0; //@line 3851 "libraries/lodepng/lodepng.cpp"
              $106 = ($105|0)==(0); //@line 3851 "libraries/lodepng/lodepng.cpp"
              if (!($106)) {
               $108 = ((($state)) + 20|0); //@line 3856 "libraries/lodepng/lodepng.cpp"
               $109 = HEAP32[$108>>2]|0; //@line 3856 "libraries/lodepng/lodepng.cpp"
               $110 = ($109|0)==(0); //@line 3856 "libraries/lodepng/lodepng.cpp"
               do {
                if ($110) {
                 $111 = ((($in)) + 29|0); //@line 3858 "libraries/lodepng/lodepng.cpp"
                 $112 = HEAP8[$111>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
                 $113 = $112&255; //@line 343 "libraries/lodepng/lodepng.cpp"
                 $114 = ((($in)) + 30|0); //@line 343 "libraries/lodepng/lodepng.cpp"
                 $115 = HEAP8[$114>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
                 $116 = $115&255; //@line 343 "libraries/lodepng/lodepng.cpp"
                 $117 = $116 << 16; //@line 343 "libraries/lodepng/lodepng.cpp"
                 $118 = ((($in)) + 31|0); //@line 343 "libraries/lodepng/lodepng.cpp"
                 $119 = HEAP8[$118>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
                 $120 = $119&255; //@line 343 "libraries/lodepng/lodepng.cpp"
                 $121 = $120 << 8; //@line 343 "libraries/lodepng/lodepng.cpp"
                 $122 = ((($in)) + 32|0); //@line 343 "libraries/lodepng/lodepng.cpp"
                 $123 = HEAP8[$122>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
                 $124 = $123&255; //@line 343 "libraries/lodepng/lodepng.cpp"
                 $c$01$i = -1;$n$02$i = 0;
                 while(1) {
                  $125 = (($39) + ($n$02$i)|0); //@line 2340 "libraries/lodepng/lodepng.cpp"
                  $126 = HEAP8[$125>>0]|0; //@line 2340 "libraries/lodepng/lodepng.cpp"
                  $127 = $126&255; //@line 2340 "libraries/lodepng/lodepng.cpp"
                  $c$0$masked$i = $c$01$i & 255; //@line 2340 "libraries/lodepng/lodepng.cpp"
                  $128 = $127 ^ $c$0$masked$i; //@line 2340 "libraries/lodepng/lodepng.cpp"
                  $129 = (33860 + ($128<<2)|0); //@line 2340 "libraries/lodepng/lodepng.cpp"
                  $130 = HEAP32[$129>>2]|0; //@line 2340 "libraries/lodepng/lodepng.cpp"
                  $131 = $c$01$i >>> 8; //@line 2340 "libraries/lodepng/lodepng.cpp"
                  $132 = $130 ^ $131; //@line 2340 "libraries/lodepng/lodepng.cpp"
                  $133 = (($n$02$i) + 1)|0; //@line 2338 "libraries/lodepng/lodepng.cpp"
                  $exitcond$i = ($133|0)==(17); //@line 2338 "libraries/lodepng/lodepng.cpp"
                  if ($exitcond$i) {
                   $$lcssa = $132;
                   break;
                  } else {
                   $c$01$i = $132;$n$02$i = $133;
                  }
                 }
                 $134 = $113 << 24; //@line 343 "libraries/lodepng/lodepng.cpp"
                 $135 = $117 | $134; //@line 343 "libraries/lodepng/lodepng.cpp"
                 $136 = $135 | $121; //@line 343 "libraries/lodepng/lodepng.cpp"
                 $137 = $136 | $124; //@line 343 "libraries/lodepng/lodepng.cpp"
                 $phitmp$i = $$lcssa ^ -1; //@line 2342 "libraries/lodepng/lodepng.cpp"
                 $138 = ($137|0)==($phitmp$i|0); //@line 3860 "libraries/lodepng/lodepng.cpp"
                 if ($138) {
                  break;
                 }
                 $139 = ((($state)) + 292|0); //@line 3862 "libraries/lodepng/lodepng.cpp"
                 HEAP32[$139>>2] = 57; //@line 3862 "libraries/lodepng/lodepng.cpp"
                 $$1 = 57;
                 return ($$1|0); //@line 3875 "libraries/lodepng/lodepng.cpp"
                }
               } while(0);
               $140 = ($95<<24>>24)==(0); //@line 3867 "libraries/lodepng/lodepng.cpp"
               if (!($140)) {
                $141 = ((($state)) + 292|0); //@line 3867 "libraries/lodepng/lodepng.cpp"
                HEAP32[$141>>2] = 32; //@line 3867 "libraries/lodepng/lodepng.cpp"
                $$1 = 32;
                return ($$1|0); //@line 3875 "libraries/lodepng/lodepng.cpp"
               }
               $142 = ($98<<24>>24)==(0); //@line 3869 "libraries/lodepng/lodepng.cpp"
               if (!($142)) {
                $143 = ((($state)) + 292|0); //@line 3869 "libraries/lodepng/lodepng.cpp"
                HEAP32[$143>>2] = 33; //@line 3869 "libraries/lodepng/lodepng.cpp"
                $$1 = 33;
                return ($$1|0); //@line 3875 "libraries/lodepng/lodepng.cpp"
               }
               $144 = ($101&255)>(1); //@line 3871 "libraries/lodepng/lodepng.cpp"
               if ($144) {
                $145 = ((($state)) + 292|0); //@line 3871 "libraries/lodepng/lodepng.cpp"
                HEAP32[$145>>2] = 34; //@line 3871 "libraries/lodepng/lodepng.cpp"
                $$1 = 34;
                return ($$1|0); //@line 3875 "libraries/lodepng/lodepng.cpp"
               }
               switch ($93|0) {
               case 0:  {
                switch ($90|0) {
                case 1: case 2: case 4: case 8: case 16:  {
                 label = 39;
                 break;
                }
                default: {
                 $$0$i = 37;
                }
                }
                break;
               }
               case 2:  {
                switch ($90|0) {
                case 8: case 16:  {
                 label = 39;
                 break;
                }
                default: {
                 $$0$i = 37;
                }
                }
                break;
               }
               case 3:  {
                switch ($90|0) {
                case 1: case 2: case 4: case 8:  {
                 label = 39;
                 break;
                }
                default: {
                 $$0$i = 37;
                }
                }
                break;
               }
               case 4:  {
                switch ($90|0) {
                case 8: case 16:  {
                 label = 39;
                 break;
                }
                default: {
                 $$0$i = 37;
                }
                }
                break;
               }
               case 6:  {
                switch ($90|0) {
                case 8: case 16:  {
                 label = 39;
                 break;
                }
                default: {
                 $$0$i = 37;
                }
                }
                break;
               }
               default: {
                $$0$i = 31;
               }
               }
               if ((label|0) == 39) {
                $$0$i = 0;
               }
               $146 = ((($state)) + 292|0); //@line 3873 "libraries/lodepng/lodepng.cpp"
               HEAP32[$146>>2] = $$0$i; //@line 3873 "libraries/lodepng/lodepng.cpp"
               $$1 = $$0$i;
               return ($$1|0); //@line 3875 "libraries/lodepng/lodepng.cpp"
              }
             }
             $107 = ((($state)) + 292|0); //@line 3853 "libraries/lodepng/lodepng.cpp"
             HEAP32[$107>>2] = 93; //@line 3853 "libraries/lodepng/lodepng.cpp"
             $$1 = 93;
             return ($$1|0); //@line 3875 "libraries/lodepng/lodepng.cpp"
            }
           }
          }
         }
         $51 = ((($state)) + 292|0); //@line 3839 "libraries/lodepng/lodepng.cpp"
         HEAP32[$51>>2] = 29; //@line 3839 "libraries/lodepng/lodepng.cpp"
         $$1 = 29;
         return ($$1|0); //@line 3875 "libraries/lodepng/lodepng.cpp"
        }
       }
      }
     }
    }
   }
  }
 }
 $38 = ((($state)) + 292|0); //@line 3835 "libraries/lodepng/lodepng.cpp"
 HEAP32[$38>>2] = 28; //@line 3835 "libraries/lodepng/lodepng.cpp"
 $$1 = 28;
 return ($$1|0); //@line 3875 "libraries/lodepng/lodepng.cpp"
}
function __Z14lodepng_decodePPhPjS1_P12LodePNGStatePKhj($out,$w,$h,$state,$in,$insize) {
 $out = $out|0;
 $w = $w|0;
 $h = $h|0;
 $state = $state|0;
 $in = $in|0;
 $insize = $insize|0;
 var $$$i = 0, $$$i$i = 0, $$$i61$i = 0, $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i$i = 0, $$0$i$i$i$i101$i = 0, $$0$i$i$i$i104$i = 0, $$0$i$i$i$i107$i = 0, $$0$i$i$i$i112$i = 0, $$0$i$i$i$i87$i = 0, $$0$i$i$i$i89$i = 0, $$0$i$i$i$i92$i = 0, $$0$i$i$i$i95$i = 0, $$0$i$i$i$i98$i = 0, $$0$i$i63$i = 0, $$0$i111$i = 0, $$0$i28$i = 0, $$0$i33$i = 0, $$0$i38$i = 0;
 var $$0$i83$i = 0, $$1 = 0, $$2$$us$i$i = 0, $$2$lcssa$i$i = 0, $$227$us$i$i = 0, $$37$i$i = 0, $$4$i$i = 0, $$5$i$i = 0, $$idx17$val$i = 0, $$idx18$i = 0, $$idx18$val$i = 0, $$idx19$val$i = 0, $$idx20$i = 0, $$idx20$val$i = 0, $$lcssa218 = 0, $$lcssa222 = 0, $$lobit$i$i = 0, $$mask$i = 0, $$ph = 0, $$pr$pre$i = 0;
 var $$pr133$pre$i = 0, $$pre = 0, $$pre$i$i = 0, $$pre$i$i$i = 0, $$pre$i$i49$i = 0, $$pre$i$i65$i = 0, $$pre$i69$pre$i$pre = 0, $$pre2$i$pre = 0, $$pre54$i$i$i = 0, $$pre55$i$i$i = 0, $$pre56$i$i$i = 0, $$pre57$i$i$i = 0, $$pre58$i$i$i = 0, $$pre59$i$i$i = 0, $$pre60$i$i$i = 0, $$pre61$i$i$i = 0, $$pre62$i$i$i = 0, $$pre90 = 0, $$pre91 = 0, $$pre92 = 0;
 var $$pre93 = 0, $$pre95 = 0, $$pre96 = 0, $$pre97 = 0, $$pre98 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0;
 var $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0;
 var $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0;
 var $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0;
 var $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0;
 var $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0;
 var $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0;
 var $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0;
 var $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0;
 var $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0;
 var $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0;
 var $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0;
 var $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0;
 var $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0;
 var $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0;
 var $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0;
 var $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0;
 var $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0;
 var $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0;
 var $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0;
 var $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0;
 var $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0;
 var $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0;
 var $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0;
 var $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0;
 var $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0;
 var $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0;
 var $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0;
 var $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0;
 var $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0;
 var $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0;
 var $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0;
 var $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0;
 var $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0;
 var $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0;
 var $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0;
 var $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0;
 var $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0;
 var $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0;
 var $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0;
 var $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0;
 var $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0;
 var $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0;
 var $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0;
 var $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0;
 var $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0;
 var $997 = 0, $998 = 0, $999 = 0, $IEND$1$i = 0, $b$08$us$us$i$i$i = 0, $b3$018$us$us$i$i$i = 0, $c$0$lcssa$i$i$i = 0, $c$0$masked$i$i$i = 0, $c$01$i$i$i = 0, $chunk$0184$i = 0, $critical_pos$0182$i = 0, $critical_pos$2$i = 0, $decoded$i$i = 0, $decoded$i55$i = 0, $error$0$i$i = 0, $error$0$i50$i = 0, $error$0$i71$i = 0, $exitcond$i$i$i = 0, $exitcond$i$i85$i = 0, $exitcond11$i$i$i = 0;
 var $exitcond11$i15$us$i$i = 0, $exitcond12$i$i$i = 0, $exitcond12$i21$us$i$i = 0, $exitcond49$i$i = 0, $exitcond50$i$i = 0, $exitcond51$i$i = 0, $exitcond52$i$i = 0, $exitcond53$i$i$i = 0, $filter_passstart$i$i = 0, $filter_passstart$i$i$i = 0, $i$01$i = 0, $i$01$i$i = 0, $i$01$i81$i = 0, $i$015$i$i$i = 0, $i$0176$i = 0, $i$02$i$i = 0, $i$026$us$i$i = 0, $i$122$i$i = 0, $i$128$i$i$i = 0, $i$314$i$i = 0;
 var $ibp$017$us$us$i$i$i = 0, $ibp$05$us$i$i$i = 0, $ibp$05$us$i19$us$i$i = 0, $ibp$11$us$i$i$i = 0, $ibp$11$us$i14$us$i$i = 0, $idat$sroa$0$0180$i = 0, $idat$sroa$0$0180$i$lcssa = 0, $idat$sroa$0$0180$i$lcssa223 = 0, $idat$sroa$0$0180$i$lcssa224 = 0, $idat$sroa$0$0180$i$lcssa225 = 0, $idat$sroa$0$0180$i$lcssa227 = 0, $idat$sroa$0$0180$i$lcssa228 = 0, $idat$sroa$0$0180$i$lcssa229 = 0, $idat$sroa$0$0180$i$lcssa230 = 0, $idat$sroa$0$0180$i$lcssa232 = 0, $idat$sroa$0$0180$i$lcssa233 = 0, $idat$sroa$0$0180$i$lcssa234 = 0, $idat$sroa$0$0180$i236 = 0, $idat$sroa$0$1$i = 0, $idat$sroa$0$3$i = 0;
 var $idat$sroa$0$3$i$lcssa = 0, $idat$sroa$0$5$i = 0, $idat$sroa$14$0177$i = 0, $idat$sroa$14$1$i = 0, $idat$sroa$14$3$i = 0, $idat$sroa$9$0178$i = 0, $idat$sroa$9$0178$i$lcssa = 0, $idat$sroa$9$0178$i$lcssa237 = 0, $idat$sroa$9$0178$i$lcssa238 = 0, $idat$sroa$9$0178$i$lcssa239 = 0, $idat$sroa$9$0178$i$lcssa241 = 0, $idat$sroa$9$0178$i$lcssa242 = 0, $idat$sroa$9$0178$i$lcssa243 = 0, $idat$sroa$9$0178$i$lcssa244 = 0, $idat$sroa$9$0178$i$lcssa246 = 0, $idat$sroa$9$0178$i$lcssa247 = 0, $idat$sroa$9$0178$i$lcssa248 = 0, $idat$sroa$9$0178$i250 = 0, $idat$sroa$9$2$i = 0, $idat$sroa$9$2$i$lcssa = 0;
 var $idat$sroa$9$4$i = 0, $key$0$i$i = 0, $key$0$i51$i = 0, $key$0$i72$i = 0, $langtag$0$i$i = 0, $length$0$lcssa$i$i = 0, $length$0$lcssa$i45$i = 0, $length$0$lcssa$i56$i = 0, $length$03$i$i = 0, $length$030$i$i = 0, $length$07$i$i = 0, $length$123$i$i = 0, $length$123$i$i$lcssa216 = 0, $length$19$i$i = 0, $length$2$lcssa$i$i = 0, $length$215$i$i = 0, $length$27$i$i = 0, $n$02$i$i$i = 0, $not$$i$i = 0, $obp$016$us$us$i$i$i = 0;
 var $obp$06$us$i$i$i = 0, $obp$06$us$i18$us$i$i = 0, $obp$12$us$i$i$i = 0, $obp$12$us$i13$us$i$i = 0, $or$cond$i = 0, $or$cond$i$i = 0, $or$cond$i$i$i = 0, $or$cond$i10$i$i = 0, $or$cond$i11$us$i$i = 0, $or$cond5$i = 0, $outv$sroa$0$1$ph247$i = 0, $outv$sroa$0$1132135$i = 0, $padded_passstart$i$i = 0, $padded_passstart$i$i$i = 0, $passh$i$i = 0, $passh$i$i$i = 0, $passstart$i$i = 0, $passstart$i$i$i = 0, $passw$i$i = 0, $passw$i$i$i = 0;
 var $phitmp$i = 0, $phitmp$i$i$i = 0, $phitmp3$i = 0, $pos$01$i$i = 0, $predict$0$i = 0, $predict$1$i = 0, $predict$2$i = 0, $predict$3$i = 0, $scanlines$i = 0, $scevgep$i$i = 0, $scevgep$i60$i = 0, $scevgep40$i$i = 0, $storemerge = 0, $str$0$i$i = 0, $tmp$us$us$i$i$i = 0, $tmp1$us$us$i$i$i = 0, $tmp2$us$us$i$i$i = 0, $tmp3$us$us$i$i$i = 0, $transkey$0$i$i = 0, $umax$i$i = 0;
 var $unknown$0183$i = 0, $unknown$1$i = 0, $x$03$us$i$i$i = 0, $x$03$us$i12$us$i$i = 0, $x$09$us$us$i$i$i = 0, $x1$022$us$us$i$i$i = 0, $y$012$us$i$i$i = 0, $y$07$us$i$i$i = 0, $y$07$us$i17$us$i$i = 0, $y2$025$us$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 352|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $decoded$i55$i = sp + 332|0;
 $decoded$i$i = sp + 320|0;
 $passw$i$i$i = sp + 292|0;
 $passh$i$i$i = sp + 264|0;
 $filter_passstart$i$i$i = sp + 232|0;
 $padded_passstart$i$i$i = sp + 200|0;
 $passstart$i$i$i = sp + 168|0;
 $passw$i$i = sp + 140|0;
 $passh$i$i = sp + 112|0;
 $filter_passstart$i$i = sp + 80|0;
 $padded_passstart$i$i = sp + 48|0;
 $passstart$i$i = sp + 16|0;
 $scanlines$i = sp;
 HEAP32[$out>>2] = 0; //@line 4449 "libraries/lodepng/lodepng.cpp"
 $0 = (__Z15lodepng_inspectPjS_P12LodePNGStatePKhj($w,$h,$state,$in,$insize)|0); //@line 4467 "libraries/lodepng/lodepng.cpp"
 $1 = ((($state)) + 292|0); //@line 4451 "libraries/lodepng/lodepng.cpp"
 HEAP32[$1>>2] = $0; //@line 4451 "libraries/lodepng/lodepng.cpp"
 $2 = ($0|0)==(0); //@line 4452 "libraries/lodepng/lodepng.cpp"
 do {
  if ($2) {
   $3 = HEAP32[$w>>2]|0; //@line 4454 "libraries/lodepng/lodepng.cpp"
   $4 = HEAP32[$h>>2]|0; //@line 4454 "libraries/lodepng/lodepng.cpp"
   $5 = Math_imul($4, $3)|0; //@line 4454 "libraries/lodepng/lodepng.cpp"
   $6 = ($4|0)==(0); //@line 4457 "libraries/lodepng/lodepng.cpp"
   if (!($6)) {
    $7 = (($5>>>0) / ($4>>>0))&-1; //@line 4457 "libraries/lodepng/lodepng.cpp"
    $8 = ($7|0)==($3|0); //@line 4457 "libraries/lodepng/lodepng.cpp"
    if (!($8)) {
     HEAP32[$1>>2] = 92; //@line 4457 "libraries/lodepng/lodepng.cpp"
     $$ph = 92;
     break;
    }
   }
   $9 = ($5>>>0)>(268435455); //@line 4460 "libraries/lodepng/lodepng.cpp"
   if ($9) {
    HEAP32[$1>>2] = 92; //@line 4460 "libraries/lodepng/lodepng.cpp"
    $$ph = 92;
    break;
   }
   $10 = ((($in)) + 33|0); //@line 4463 "libraries/lodepng/lodepng.cpp"
   $11 = $in; //@line 4473 "libraries/lodepng/lodepng.cpp"
   $12 = ((($state)) + 20|0); //@line 4578 "libraries/lodepng/lodepng.cpp"
   $13 = ((($state)) + 152|0); //@line 4126 "libraries/lodepng/lodepng.cpp"
   $14 = ((($state)) + 156|0); //@line 4127 "libraries/lodepng/lodepng.cpp"
   $15 = ((($state)) + 144|0); //@line 4150 "libraries/lodepng/lodepng.cpp"
   $16 = ((($state)) + 160|0); //@line 4162 "libraries/lodepng/lodepng.cpp"
   $17 = ((($state)) + 172|0); //@line 4163 "libraries/lodepng/lodepng.cpp"
   $18 = ((($state)) + 168|0); //@line 4163 "libraries/lodepng/lodepng.cpp"
   $19 = ((($state)) + 164|0); //@line 4163 "libraries/lodepng/lodepng.cpp"
   $20 = ((($state)) + 176|0); //@line 4190 "libraries/lodepng/lodepng.cpp"
   $21 = ((($state)) + 188|0); //@line 4191 "libraries/lodepng/lodepng.cpp"
   $22 = ((($state)) + 184|0); //@line 4191 "libraries/lodepng/lodepng.cpp"
   $23 = ((($state)) + 180|0); //@line 4191 "libraries/lodepng/lodepng.cpp"
   $24 = ((($state)) + 28|0); //@line 4527 "libraries/lodepng/lodepng.cpp"
   $25 = ((($state)) + 132|0); //@line 4529 "libraries/lodepng/lodepng.cpp"
   $26 = ((($state)) + 4|0); //@line 4538 "libraries/lodepng/lodepng.cpp"
   $27 = ((($decoded$i$i)) + 8|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   $28 = ((($decoded$i$i)) + 4|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   $29 = ((($state)) + 8|0); //@line 2150 "libraries/lodepng/lodepng.cpp"
   $30 = ((($decoded$i55$i)) + 8|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   $31 = ((($decoded$i55$i)) + 4|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   $32 = ((($state)) + 224|0); //@line 4405 "libraries/lodepng/lodepng.cpp"
   $33 = ((($state)) + 228|0); //@line 4406 "libraries/lodepng/lodepng.cpp"
   $34 = ((($state)) + 232|0); //@line 4407 "libraries/lodepng/lodepng.cpp"
   $35 = ((($state)) + 236|0); //@line 4408 "libraries/lodepng/lodepng.cpp"
   $36 = ((($state)) + 240|0); //@line 4409 "libraries/lodepng/lodepng.cpp"
   $37 = ((($state)) + 244|0); //@line 4410 "libraries/lodepng/lodepng.cpp"
   $38 = ((($state)) + 248|0); //@line 4411 "libraries/lodepng/lodepng.cpp"
   $39 = ((($state)) + 252|0); //@line 4420 "libraries/lodepng/lodepng.cpp"
   $40 = ((($state)) + 256|0); //@line 4421 "libraries/lodepng/lodepng.cpp"
   $41 = ((($state)) + 260|0); //@line 4422 "libraries/lodepng/lodepng.cpp"
   $42 = ((($state)) + 264|0); //@line 4423 "libraries/lodepng/lodepng.cpp"
   $43 = ((($state)) + 32|0); //@line 4569 "libraries/lodepng/lodepng.cpp"
   $chunk$0184$i = $10;$critical_pos$0182$i = 1;$idat$sroa$0$0180$i = 0;$idat$sroa$14$0177$i = 0;$idat$sroa$9$0178$i = 0;$unknown$0183$i = 0;
   L10: while(1) {
    $44 = $chunk$0184$i; //@line 4473 "libraries/lodepng/lodepng.cpp"
    $45 = (($44) - ($11))|0; //@line 4473 "libraries/lodepng/lodepng.cpp"
    $46 = (($45) + 12)|0; //@line 4473 "libraries/lodepng/lodepng.cpp"
    $47 = ($46>>>0)>($insize>>>0); //@line 4473 "libraries/lodepng/lodepng.cpp"
    $48 = ($chunk$0184$i>>>0)<($in>>>0); //@line 4473 "libraries/lodepng/lodepng.cpp"
    $or$cond$i = $48 | $47; //@line 4473 "libraries/lodepng/lodepng.cpp"
    if ($or$cond$i) {
     $idat$sroa$0$0180$i$lcssa = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa = $idat$sroa$9$0178$i;
     label = 9;
     break;
    }
    $49 = HEAP8[$chunk$0184$i>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
    $50 = $49&255; //@line 343 "libraries/lodepng/lodepng.cpp"
    $51 = $50 << 24; //@line 343 "libraries/lodepng/lodepng.cpp"
    $52 = ((($chunk$0184$i)) + 1|0); //@line 343 "libraries/lodepng/lodepng.cpp"
    $53 = HEAP8[$52>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
    $54 = $53&255; //@line 343 "libraries/lodepng/lodepng.cpp"
    $55 = $54 << 16; //@line 343 "libraries/lodepng/lodepng.cpp"
    $56 = $55 | $51; //@line 343 "libraries/lodepng/lodepng.cpp"
    $57 = ((($chunk$0184$i)) + 2|0); //@line 343 "libraries/lodepng/lodepng.cpp"
    $58 = HEAP8[$57>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
    $59 = $58&255; //@line 343 "libraries/lodepng/lodepng.cpp"
    $60 = $59 << 8; //@line 343 "libraries/lodepng/lodepng.cpp"
    $61 = $56 | $60; //@line 343 "libraries/lodepng/lodepng.cpp"
    $62 = ((($chunk$0184$i)) + 3|0); //@line 343 "libraries/lodepng/lodepng.cpp"
    $63 = HEAP8[$62>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
    $64 = $63&255; //@line 343 "libraries/lodepng/lodepng.cpp"
    $65 = $61 | $64; //@line 343 "libraries/lodepng/lodepng.cpp"
    $66 = ($65|0)<(0); //@line 4478 "libraries/lodepng/lodepng.cpp"
    if ($66) {
     $idat$sroa$0$0180$i$lcssa223 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa237 = $idat$sroa$9$0178$i;
     label = 11;
     break;
    }
    $67 = (($65) + ($46))|0; //@line 4480 "libraries/lodepng/lodepng.cpp"
    $68 = ($67>>>0)>($insize>>>0); //@line 4480 "libraries/lodepng/lodepng.cpp"
    if ($68) {
     $idat$sroa$0$0180$i$lcssa224 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa238 = $idat$sroa$9$0178$i;
     label = 14;
     break;
    }
    $69 = (($chunk$0184$i) + ($65)|0); //@line 4480 "libraries/lodepng/lodepng.cpp"
    $70 = ((($69)) + 12|0); //@line 4480 "libraries/lodepng/lodepng.cpp"
    $71 = ($70>>>0)<($in>>>0); //@line 4480 "libraries/lodepng/lodepng.cpp"
    if ($71) {
     $idat$sroa$0$0180$i$lcssa224 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa238 = $idat$sroa$9$0178$i;
     label = 14;
     break;
    }
    $72 = ((($chunk$0184$i)) + 8|0); //@line 2432 "libraries/lodepng/lodepng.cpp"
    $73 = ((($chunk$0184$i)) + 4|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
    $74 = HEAP8[$73>>0]|0; //@line 2412 "libraries/lodepng/lodepng.cpp"
    L16: do {
     switch ($74<<24>>24) {
     case 73:  {
      $75 = ((($chunk$0184$i)) + 5|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
      $76 = HEAP8[$75>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
      switch ($76<<24>>24) {
      case 68:  {
       break;
      }
      case 69:  {
       $99 = ((($chunk$0184$i)) + 6|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
       $100 = HEAP8[$99>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
       $101 = ($100<<24>>24)==(78); //@line 2407 "libraries/lodepng/lodepng.cpp"
       if (!($101)) {
        $idat$sroa$0$0180$i$lcssa230 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa244 = $idat$sroa$9$0178$i;
        label = 165;
        break L10;
       }
       $102 = ((($chunk$0184$i)) + 7|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
       $103 = HEAP8[$102>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
       $104 = ($103<<24>>24)==(68); //@line 2407 "libraries/lodepng/lodepng.cpp"
       if ($104) {
        $IEND$1$i = 1;$critical_pos$2$i = $critical_pos$0182$i;$idat$sroa$0$3$i = $idat$sroa$0$0180$i;$idat$sroa$14$3$i = $idat$sroa$14$0177$i;$idat$sroa$9$2$i = $idat$sroa$9$0178$i;$unknown$1$i = $unknown$0183$i;
        break L16;
       } else {
        $idat$sroa$0$0180$i$lcssa230 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa244 = $idat$sroa$9$0178$i;
        label = 165;
        break L10;
       }
       break;
      }
      default: {
       $idat$sroa$0$0180$i$lcssa230 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa244 = $idat$sroa$9$0178$i;
       label = 165;
       break L10;
      }
      }
      $77 = ((($chunk$0184$i)) + 6|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
      $78 = HEAP8[$77>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
      $79 = ($78<<24>>24)==(65); //@line 2407 "libraries/lodepng/lodepng.cpp"
      if (!($79)) {
       $idat$sroa$0$0180$i$lcssa230 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa244 = $idat$sroa$9$0178$i;
       label = 165;
       break L10;
      }
      $80 = ((($chunk$0184$i)) + 7|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
      $81 = HEAP8[$80>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
      $82 = ($81<<24>>24)==(84); //@line 2407 "libraries/lodepng/lodepng.cpp"
      if (!($82)) {
       $idat$sroa$0$0180$i$lcssa230 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa244 = $idat$sroa$9$0178$i;
       label = 165;
       break L10;
      }
      $83 = (($65) + ($idat$sroa$9$0178$i))|0; //@line 4491 "libraries/lodepng/lodepng.cpp"
      $84 = ($idat$sroa$14$0177$i>>>0)<($83>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
      if ($84) {
       $85 = $idat$sroa$14$0177$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
       $86 = ($85>>>0)<($83>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
       $87 = ($83*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
       $88 = $87 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
       $89 = $86 ? $83 : $88; //@line 227 "libraries/lodepng/lodepng.cpp"
       $90 = (_realloc($idat$sroa$0$0180$i,$89)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       $91 = ($90|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
       if ($91) {
        $idat$sroa$0$0180$i$lcssa233 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa247 = $idat$sroa$9$0178$i;
        label = 23;
        break L10;
       } else {
        $idat$sroa$0$1$i = $90;$idat$sroa$14$1$i = $89;
       }
      } else {
       $idat$sroa$0$1$i = $idat$sroa$0$0180$i;$idat$sroa$14$1$i = $idat$sroa$14$0177$i;
      }
      $92 = ($65|0)==(0); //@line 4492 "libraries/lodepng/lodepng.cpp"
      if ($92) {
       $IEND$1$i = 0;$critical_pos$2$i = 3;$idat$sroa$0$3$i = $idat$sroa$0$1$i;$idat$sroa$14$3$i = $idat$sroa$14$1$i;$idat$sroa$9$2$i = $83;$unknown$1$i = $unknown$0183$i;
      } else {
       $i$0176$i = 0;
       while(1) {
        $93 = (($72) + ($i$0176$i)|0); //@line 4492 "libraries/lodepng/lodepng.cpp"
        $94 = HEAP8[$93>>0]|0; //@line 4492 "libraries/lodepng/lodepng.cpp"
        $95 = (($i$0176$i) + ($idat$sroa$9$0178$i))|0; //@line 4492 "libraries/lodepng/lodepng.cpp"
        $96 = (($idat$sroa$0$1$i) + ($95)|0); //@line 4492 "libraries/lodepng/lodepng.cpp"
        HEAP8[$96>>0] = $94; //@line 4492 "libraries/lodepng/lodepng.cpp"
        $97 = (($i$0176$i) + 1)|0; //@line 4492 "libraries/lodepng/lodepng.cpp"
        $98 = ($97|0)==($65|0); //@line 4492 "libraries/lodepng/lodepng.cpp"
        if ($98) {
         $IEND$1$i = 0;$critical_pos$2$i = 3;$idat$sroa$0$3$i = $idat$sroa$0$1$i;$idat$sroa$14$3$i = $idat$sroa$14$1$i;$idat$sroa$9$2$i = $83;$unknown$1$i = $unknown$0183$i;
         break;
        } else {
         $i$0176$i = $97;
        }
       }
      }
      break;
     }
     case 80:  {
      $105 = ((($chunk$0184$i)) + 5|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
      $106 = HEAP8[$105>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
      $107 = ($106<<24>>24)==(76); //@line 2407 "libraries/lodepng/lodepng.cpp"
      if (!($107)) {
       $idat$sroa$0$0180$i$lcssa230 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa244 = $idat$sroa$9$0178$i;
       label = 165;
       break L10;
      }
      $108 = ((($chunk$0184$i)) + 6|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
      $109 = HEAP8[$108>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
      $110 = ($109<<24>>24)==(84); //@line 2407 "libraries/lodepng/lodepng.cpp"
      if (!($110)) {
       $idat$sroa$0$0180$i$lcssa230 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa244 = $idat$sroa$9$0178$i;
       label = 165;
       break L10;
      }
      $111 = ((($chunk$0184$i)) + 7|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
      $112 = HEAP8[$111>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
      $113 = ($112<<24>>24)==(69); //@line 2407 "libraries/lodepng/lodepng.cpp"
      if (!($113)) {
       $idat$sroa$0$0180$i$lcssa230 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa244 = $idat$sroa$9$0178$i;
       label = 165;
       break L10;
      }
      $114 = HEAP32[$13>>2]|0; //@line 4126 "libraries/lodepng/lodepng.cpp"
      $115 = ($114|0)==(0|0); //@line 4126 "libraries/lodepng/lodepng.cpp"
      if (!($115)) {
       _free($114); //@line 76 "libraries/lodepng/lodepng.cpp"
      }
      $116 = (($65>>>0) / 3)&-1; //@line 4136 "libraries/lodepng/lodepng.cpp"
      HEAP32[$14>>2] = $116; //@line 4127 "libraries/lodepng/lodepng.cpp"
      $117 = $116 << 2; //@line 4128 "libraries/lodepng/lodepng.cpp"
      $118 = (_malloc($117)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
      HEAP32[$13>>2] = $118; //@line 4128 "libraries/lodepng/lodepng.cpp"
      $119 = ($118|0)==(0|0); //@line 4129 "libraries/lodepng/lodepng.cpp"
      L37: do {
       if ($119) {
        $120 = ($65>>>0)<(3); //@line 4129 "libraries/lodepng/lodepng.cpp"
        if (!($120)) {
         $idat$sroa$0$0180$i$lcssa232 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa246 = $idat$sroa$9$0178$i;
         label = 33;
         break L10;
        }
       } else {
        $121 = ($65>>>0)>(770); //@line 4134 "libraries/lodepng/lodepng.cpp"
        if ($121) {
         $$0$i28$i = 38;$idat$sroa$0$0180$i236 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i250 = $idat$sroa$9$0178$i;
         label = 39;
         break L10;
        }
        $122 = ($65>>>0)<(3); //@line 4136 "libraries/lodepng/lodepng.cpp"
        if (!($122)) {
         $128 = $118;$i$02$i$i = 0;$pos$01$i$i = 0;
         while(1) {
          $123 = (($pos$01$i$i) + 1)|0; //@line 4138 "libraries/lodepng/lodepng.cpp"
          $124 = (($72) + ($pos$01$i$i)|0); //@line 4138 "libraries/lodepng/lodepng.cpp"
          $125 = HEAP8[$124>>0]|0; //@line 4138 "libraries/lodepng/lodepng.cpp"
          $126 = $i$02$i$i << 2; //@line 4138 "libraries/lodepng/lodepng.cpp"
          $127 = (($128) + ($126)|0); //@line 4138 "libraries/lodepng/lodepng.cpp"
          HEAP8[$127>>0] = $125; //@line 4138 "libraries/lodepng/lodepng.cpp"
          $129 = (($pos$01$i$i) + 2)|0; //@line 4139 "libraries/lodepng/lodepng.cpp"
          $130 = (($72) + ($123)|0); //@line 4139 "libraries/lodepng/lodepng.cpp"
          $131 = HEAP8[$130>>0]|0; //@line 4139 "libraries/lodepng/lodepng.cpp"
          $132 = $126 | 1; //@line 4139 "libraries/lodepng/lodepng.cpp"
          $133 = HEAP32[$13>>2]|0; //@line 4139 "libraries/lodepng/lodepng.cpp"
          $134 = (($133) + ($132)|0); //@line 4139 "libraries/lodepng/lodepng.cpp"
          HEAP8[$134>>0] = $131; //@line 4139 "libraries/lodepng/lodepng.cpp"
          $135 = (($72) + ($129)|0); //@line 4140 "libraries/lodepng/lodepng.cpp"
          $136 = HEAP8[$135>>0]|0; //@line 4140 "libraries/lodepng/lodepng.cpp"
          $137 = $126 | 2; //@line 4140 "libraries/lodepng/lodepng.cpp"
          $138 = HEAP32[$13>>2]|0; //@line 4140 "libraries/lodepng/lodepng.cpp"
          $139 = (($138) + ($137)|0); //@line 4140 "libraries/lodepng/lodepng.cpp"
          HEAP8[$139>>0] = $136; //@line 4140 "libraries/lodepng/lodepng.cpp"
          $140 = $126 | 3; //@line 4141 "libraries/lodepng/lodepng.cpp"
          $141 = HEAP32[$13>>2]|0; //@line 4141 "libraries/lodepng/lodepng.cpp"
          $142 = (($141) + ($140)|0); //@line 4141 "libraries/lodepng/lodepng.cpp"
          HEAP8[$142>>0] = -1; //@line 4141 "libraries/lodepng/lodepng.cpp"
          $143 = (($i$02$i$i) + 1)|0; //@line 4136 "libraries/lodepng/lodepng.cpp"
          $144 = HEAP32[$14>>2]|0; //@line 4136 "libraries/lodepng/lodepng.cpp"
          $145 = ($143|0)==($144|0); //@line 4136 "libraries/lodepng/lodepng.cpp"
          if ($145) {
           break L37;
          }
          $146 = (($pos$01$i$i) + 3)|0; //@line 4140 "libraries/lodepng/lodepng.cpp"
          $$pre$i$i = HEAP32[$13>>2]|0; //@line 4138 "libraries/lodepng/lodepng.cpp"
          $128 = $$pre$i$i;$i$02$i$i = $143;$pos$01$i$i = $146;
         }
        }
       }
      } while(0);
      HEAP32[$1>>2] = 0; //@line 4505 "libraries/lodepng/lodepng.cpp"
      $IEND$1$i = 0;$critical_pos$2$i = 2;$idat$sroa$0$3$i = $idat$sroa$0$0180$i;$idat$sroa$14$3$i = $idat$sroa$14$0177$i;$idat$sroa$9$2$i = $idat$sroa$9$0178$i;$unknown$1$i = $unknown$0183$i;
      break;
     }
     case 116:  {
      $147 = ((($chunk$0184$i)) + 5|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
      $148 = HEAP8[$147>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
      switch ($148<<24>>24) {
      case 82:  {
       $149 = ((($chunk$0184$i)) + 6|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
       $150 = HEAP8[$149>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
       $151 = ($150<<24>>24)==(78); //@line 2407 "libraries/lodepng/lodepng.cpp"
       if (!($151)) {
        label = 166;
        break L16;
       }
       $152 = ((($chunk$0184$i)) + 7|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
       $153 = HEAP8[$152>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
       $154 = ($153<<24>>24)==(83); //@line 2407 "libraries/lodepng/lodepng.cpp"
       if (!($154)) {
        label = 166;
        break L16;
       }
       $155 = HEAP32[$15>>2]|0; //@line 4150 "libraries/lodepng/lodepng.cpp"
       switch ($155|0) {
       case 3:  {
        $156 = HEAP32[$14>>2]|0; //@line 4153 "libraries/lodepng/lodepng.cpp"
        $157 = ($156>>>0)<($65>>>0); //@line 4153 "libraries/lodepng/lodepng.cpp"
        if ($157) {
         $$0$i33$i = 38;$idat$sroa$0$0180$i$lcssa229 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa243 = $idat$sroa$9$0178$i;
         label = 52;
         break L10;
        }
        $158 = ($65|0)==(0); //@line 4155 "libraries/lodepng/lodepng.cpp"
        if (!($158)) {
         $i$01$i$i = 0;
         while(1) {
          $159 = (($72) + ($i$01$i$i)|0); //@line 4155 "libraries/lodepng/lodepng.cpp"
          $160 = HEAP8[$159>>0]|0; //@line 4155 "libraries/lodepng/lodepng.cpp"
          $161 = $i$01$i$i << 2; //@line 4155 "libraries/lodepng/lodepng.cpp"
          $162 = $161 | 3; //@line 4155 "libraries/lodepng/lodepng.cpp"
          $163 = HEAP32[$13>>2]|0; //@line 4155 "libraries/lodepng/lodepng.cpp"
          $164 = (($163) + ($162)|0); //@line 4155 "libraries/lodepng/lodepng.cpp"
          HEAP8[$164>>0] = $160; //@line 4155 "libraries/lodepng/lodepng.cpp"
          $165 = (($i$01$i$i) + 1)|0; //@line 4155 "libraries/lodepng/lodepng.cpp"
          $166 = ($165|0)==($65|0); //@line 4155 "libraries/lodepng/lodepng.cpp"
          if ($166) {
           break;
          } else {
           $i$01$i$i = $165;
          }
         }
        }
        break;
       }
       case 0:  {
        $167 = ($65|0)==(2); //@line 4160 "libraries/lodepng/lodepng.cpp"
        if (!($167)) {
         $$0$i33$i = 30;$idat$sroa$0$0180$i$lcssa229 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa243 = $idat$sroa$9$0178$i;
         label = 52;
         break L10;
        }
        HEAP32[$16>>2] = 1; //@line 4162 "libraries/lodepng/lodepng.cpp"
        $168 = HEAP8[$72>>0]|0; //@line 4163 "libraries/lodepng/lodepng.cpp"
        $169 = $168&255; //@line 4163 "libraries/lodepng/lodepng.cpp"
        $170 = $169 << 8; //@line 4163 "libraries/lodepng/lodepng.cpp"
        $171 = ((($chunk$0184$i)) + 9|0); //@line 4163 "libraries/lodepng/lodepng.cpp"
        $172 = HEAP8[$171>>0]|0; //@line 4163 "libraries/lodepng/lodepng.cpp"
        $173 = $172&255; //@line 4163 "libraries/lodepng/lodepng.cpp"
        $174 = $170 | $173; //@line 4163 "libraries/lodepng/lodepng.cpp"
        HEAP32[$17>>2] = $174; //@line 4163 "libraries/lodepng/lodepng.cpp"
        HEAP32[$18>>2] = $174; //@line 4163 "libraries/lodepng/lodepng.cpp"
        HEAP32[$19>>2] = $174; //@line 4163 "libraries/lodepng/lodepng.cpp"
        break;
       }
       case 2:  {
        $175 = ($65|0)==(6); //@line 4168 "libraries/lodepng/lodepng.cpp"
        if (!($175)) {
         $$0$i33$i = 41;$idat$sroa$0$0180$i$lcssa229 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa243 = $idat$sroa$9$0178$i;
         label = 52;
         break L10;
        }
        HEAP32[$16>>2] = 1; //@line 4170 "libraries/lodepng/lodepng.cpp"
        $176 = HEAP8[$72>>0]|0; //@line 4171 "libraries/lodepng/lodepng.cpp"
        $177 = $176&255; //@line 4171 "libraries/lodepng/lodepng.cpp"
        $178 = $177 << 8; //@line 4171 "libraries/lodepng/lodepng.cpp"
        $179 = ((($chunk$0184$i)) + 9|0); //@line 4171 "libraries/lodepng/lodepng.cpp"
        $180 = HEAP8[$179>>0]|0; //@line 4171 "libraries/lodepng/lodepng.cpp"
        $181 = $180&255; //@line 4171 "libraries/lodepng/lodepng.cpp"
        $182 = $178 | $181; //@line 4171 "libraries/lodepng/lodepng.cpp"
        HEAP32[$19>>2] = $182; //@line 4171 "libraries/lodepng/lodepng.cpp"
        $183 = ((($chunk$0184$i)) + 10|0); //@line 4172 "libraries/lodepng/lodepng.cpp"
        $184 = HEAP8[$183>>0]|0; //@line 4172 "libraries/lodepng/lodepng.cpp"
        $185 = $184&255; //@line 4172 "libraries/lodepng/lodepng.cpp"
        $186 = $185 << 8; //@line 4172 "libraries/lodepng/lodepng.cpp"
        $187 = ((($chunk$0184$i)) + 11|0); //@line 4172 "libraries/lodepng/lodepng.cpp"
        $188 = HEAP8[$187>>0]|0; //@line 4172 "libraries/lodepng/lodepng.cpp"
        $189 = $188&255; //@line 4172 "libraries/lodepng/lodepng.cpp"
        $190 = $186 | $189; //@line 4172 "libraries/lodepng/lodepng.cpp"
        HEAP32[$18>>2] = $190; //@line 4172 "libraries/lodepng/lodepng.cpp"
        $191 = ((($chunk$0184$i)) + 12|0); //@line 4173 "libraries/lodepng/lodepng.cpp"
        $192 = HEAP8[$191>>0]|0; //@line 4173 "libraries/lodepng/lodepng.cpp"
        $193 = $192&255; //@line 4173 "libraries/lodepng/lodepng.cpp"
        $194 = $193 << 8; //@line 4173 "libraries/lodepng/lodepng.cpp"
        $195 = ((($chunk$0184$i)) + 13|0); //@line 4173 "libraries/lodepng/lodepng.cpp"
        $196 = HEAP8[$195>>0]|0; //@line 4173 "libraries/lodepng/lodepng.cpp"
        $197 = $196&255; //@line 4173 "libraries/lodepng/lodepng.cpp"
        $198 = $194 | $197; //@line 4173 "libraries/lodepng/lodepng.cpp"
        HEAP32[$17>>2] = $198; //@line 4173 "libraries/lodepng/lodepng.cpp"
        break;
       }
       default: {
        $$0$i33$i = 42;$idat$sroa$0$0180$i$lcssa229 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa243 = $idat$sroa$9$0178$i;
        label = 52;
        break L10;
       }
       }
       HEAP32[$1>>2] = 0; //@line 4514 "libraries/lodepng/lodepng.cpp"
       $IEND$1$i = 0;$critical_pos$2$i = $critical_pos$0182$i;$idat$sroa$0$3$i = $idat$sroa$0$0180$i;$idat$sroa$14$3$i = $idat$sroa$14$0177$i;$idat$sroa$9$2$i = $idat$sroa$9$0178$i;$unknown$1$i = $unknown$0183$i;
       break L16;
       break;
      }
      case 69:  {
       $244 = ((($chunk$0184$i)) + 6|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
       $245 = HEAP8[$244>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
       $246 = ($245<<24>>24)==(88); //@line 2407 "libraries/lodepng/lodepng.cpp"
       if (!($246)) {
        label = 166;
        break L16;
       }
       $247 = ((($chunk$0184$i)) + 7|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
       $248 = HEAP8[$247>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
       $249 = ($248<<24>>24)==(116); //@line 2407 "libraries/lodepng/lodepng.cpp"
       if (!($249)) {
        label = 166;
        break L16;
       }
       $250 = HEAP32[$24>>2]|0; //@line 4527 "libraries/lodepng/lodepng.cpp"
       $251 = ($250|0)==(0); //@line 4527 "libraries/lodepng/lodepng.cpp"
       if ($251) {
        $IEND$1$i = 0;$critical_pos$2$i = $critical_pos$0182$i;$idat$sroa$0$3$i = $idat$sroa$0$0180$i;$idat$sroa$14$3$i = $idat$sroa$14$0177$i;$idat$sroa$9$2$i = $idat$sroa$9$0178$i;$unknown$1$i = $unknown$0183$i;
        break L16;
       }
       $252 = ($65|0)==(0); //@line 4227 "libraries/lodepng/lodepng.cpp"
       if ($252) {
        $error$0$i$i = 89;$key$0$i$i = 0;$str$0$i$i = 0;
       } else {
        $length$07$i$i = 0;
        while(1) {
         $253 = (($72) + ($length$07$i$i)|0); //@line 4227 "libraries/lodepng/lodepng.cpp"
         $254 = HEAP8[$253>>0]|0; //@line 4227 "libraries/lodepng/lodepng.cpp"
         $255 = ($254<<24>>24)==(0); //@line 4227 "libraries/lodepng/lodepng.cpp"
         if ($255) {
          $length$0$lcssa$i$i = $length$07$i$i;
          break;
         }
         $256 = (($length$07$i$i) + 1)|0; //@line 4227 "libraries/lodepng/lodepng.cpp"
         $257 = ($256>>>0)<($65>>>0); //@line 4227 "libraries/lodepng/lodepng.cpp"
         if ($257) {
          $length$07$i$i = $256;
         } else {
          $length$0$lcssa$i$i = $256;
          break;
         }
        }
        $258 = (($length$0$lcssa$i$i) + -1)|0; //@line 4230 "libraries/lodepng/lodepng.cpp"
        $259 = ($258>>>0)>(78); //@line 4230 "libraries/lodepng/lodepng.cpp"
        if ($259) {
         $error$0$i$i = 89;$key$0$i$i = 0;$str$0$i$i = 0;
        } else {
         $260 = (($length$0$lcssa$i$i) + 1)|0; //@line 4232 "libraries/lodepng/lodepng.cpp"
         $261 = (_malloc($260)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
         $262 = ($261|0)==(0|0); //@line 4233 "libraries/lodepng/lodepng.cpp"
         if ($262) {
          $error$0$i$i = 83;$key$0$i$i = 0;$str$0$i$i = 0;
         } else {
          $263 = (($261) + ($length$0$lcssa$i$i)|0); //@line 4235 "libraries/lodepng/lodepng.cpp"
          HEAP8[$263>>0] = 0; //@line 4235 "libraries/lodepng/lodepng.cpp"
          $264 = ($length$0$lcssa$i$i|0)==(0); //@line 4236 "libraries/lodepng/lodepng.cpp"
          if (!($264)) {
           _memcpy(($261|0),($72|0),($length$0$lcssa$i$i|0))|0; //@line 4236 "libraries/lodepng/lodepng.cpp"
          }
          $265 = ($65>>>0)<($260>>>0); //@line 4240 "libraries/lodepng/lodepng.cpp"
          $266 = (($65) - ($260))|0; //@line 4240 "libraries/lodepng/lodepng.cpp"
          $$$i$i = $265 ? 0 : $266; //@line 4240 "libraries/lodepng/lodepng.cpp"
          $267 = (($$$i$i) + 1)|0; //@line 4241 "libraries/lodepng/lodepng.cpp"
          $268 = (_malloc($267)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
          $269 = ($268|0)==(0|0); //@line 4242 "libraries/lodepng/lodepng.cpp"
          if ($269) {
           $error$0$i$i = 83;$key$0$i$i = $261;$str$0$i$i = 0;
          } else {
           $270 = (($268) + ($$$i$i)|0); //@line 4244 "libraries/lodepng/lodepng.cpp"
           HEAP8[$270>>0] = 0; //@line 4244 "libraries/lodepng/lodepng.cpp"
           $271 = ($$$i$i|0)==(0); //@line 4245 "libraries/lodepng/lodepng.cpp"
           if (!($271)) {
            $scevgep$i$i = (($72) + ($260)|0); //@line 4245 "libraries/lodepng/lodepng.cpp"
            $272 = ($65>>>0)>($260>>>0); //@line 4245 "libraries/lodepng/lodepng.cpp"
            $umax$i$i = $272 ? $65 : $260; //@line 4245 "libraries/lodepng/lodepng.cpp"
            $273 = $length$0$lcssa$i$i ^ -1; //@line 4245 "libraries/lodepng/lodepng.cpp"
            $274 = (($umax$i$i) + ($273))|0; //@line 4245 "libraries/lodepng/lodepng.cpp"
            _memcpy(($268|0),($scevgep$i$i|0),($274|0))|0; //@line 4245 "libraries/lodepng/lodepng.cpp"
           }
           $275 = (__Z16lodepng_add_textP11LodePNGInfoPKcS2_($25,$261,$268)|0); //@line 4247 "libraries/lodepng/lodepng.cpp"
           $error$0$i$i = $275;$key$0$i$i = $261;$str$0$i$i = $268;
          }
         }
        }
       }
       _free($key$0$i$i); //@line 76 "libraries/lodepng/lodepng.cpp"
       _free($str$0$i$i); //@line 76 "libraries/lodepng/lodepng.cpp"
       HEAP32[$1>>2] = $error$0$i$i; //@line 4529 "libraries/lodepng/lodepng.cpp"
       $276 = ($error$0$i$i|0)==(0); //@line 4530 "libraries/lodepng/lodepng.cpp"
       if ($276) {
        $IEND$1$i = 0;$critical_pos$2$i = $critical_pos$0182$i;$idat$sroa$0$3$i = $idat$sroa$0$0180$i;$idat$sroa$14$3$i = $idat$sroa$14$0177$i;$idat$sroa$9$2$i = $idat$sroa$9$0178$i;$unknown$1$i = $unknown$0183$i;
        break L16;
       } else {
        $idat$sroa$0$5$i = $idat$sroa$0$0180$i;$idat$sroa$9$4$i = $idat$sroa$9$0178$i;
        break L10;
       }
       break;
      }
      case 73:  {
       $447 = ((($chunk$0184$i)) + 6|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
       $448 = HEAP8[$447>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
       $449 = ($448<<24>>24)==(77); //@line 2407 "libraries/lodepng/lodepng.cpp"
       if (!($449)) {
        label = 166;
        break L16;
       }
       $450 = ((($chunk$0184$i)) + 7|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
       $451 = HEAP8[$450>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
       $452 = ($451<<24>>24)==(69); //@line 2407 "libraries/lodepng/lodepng.cpp"
       if (!($452)) {
        label = 166;
        break L16;
       }
       $453 = ($65|0)==(7); //@line 4403 "libraries/lodepng/lodepng.cpp"
       if (!($453)) {
        $idat$sroa$0$0180$i$lcssa228 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa242 = $idat$sroa$9$0178$i;
        label = 157;
        break L10;
       }
       HEAP32[$32>>2] = 1; //@line 4405 "libraries/lodepng/lodepng.cpp"
       $454 = HEAP8[$72>>0]|0; //@line 4406 "libraries/lodepng/lodepng.cpp"
       $455 = $454&255; //@line 4406 "libraries/lodepng/lodepng.cpp"
       $456 = $455 << 8; //@line 4406 "libraries/lodepng/lodepng.cpp"
       $457 = ((($chunk$0184$i)) + 9|0); //@line 4406 "libraries/lodepng/lodepng.cpp"
       $458 = HEAP8[$457>>0]|0; //@line 4406 "libraries/lodepng/lodepng.cpp"
       $459 = $458&255; //@line 4406 "libraries/lodepng/lodepng.cpp"
       $460 = $456 | $459; //@line 4406 "libraries/lodepng/lodepng.cpp"
       HEAP32[$33>>2] = $460; //@line 4406 "libraries/lodepng/lodepng.cpp"
       $461 = ((($chunk$0184$i)) + 10|0); //@line 4407 "libraries/lodepng/lodepng.cpp"
       $462 = HEAP8[$461>>0]|0; //@line 4407 "libraries/lodepng/lodepng.cpp"
       $463 = $462&255; //@line 4407 "libraries/lodepng/lodepng.cpp"
       HEAP32[$34>>2] = $463; //@line 4407 "libraries/lodepng/lodepng.cpp"
       $464 = ((($chunk$0184$i)) + 11|0); //@line 4408 "libraries/lodepng/lodepng.cpp"
       $465 = HEAP8[$464>>0]|0; //@line 4408 "libraries/lodepng/lodepng.cpp"
       $466 = $465&255; //@line 4408 "libraries/lodepng/lodepng.cpp"
       HEAP32[$35>>2] = $466; //@line 4408 "libraries/lodepng/lodepng.cpp"
       $467 = ((($chunk$0184$i)) + 12|0); //@line 4409 "libraries/lodepng/lodepng.cpp"
       $468 = HEAP8[$467>>0]|0; //@line 4409 "libraries/lodepng/lodepng.cpp"
       $469 = $468&255; //@line 4409 "libraries/lodepng/lodepng.cpp"
       HEAP32[$36>>2] = $469; //@line 4409 "libraries/lodepng/lodepng.cpp"
       $470 = ((($chunk$0184$i)) + 13|0); //@line 4410 "libraries/lodepng/lodepng.cpp"
       $471 = HEAP8[$470>>0]|0; //@line 4410 "libraries/lodepng/lodepng.cpp"
       $472 = $471&255; //@line 4410 "libraries/lodepng/lodepng.cpp"
       HEAP32[$37>>2] = $472; //@line 4410 "libraries/lodepng/lodepng.cpp"
       $473 = ((($chunk$0184$i)) + 14|0); //@line 4411 "libraries/lodepng/lodepng.cpp"
       $474 = HEAP8[$473>>0]|0; //@line 4411 "libraries/lodepng/lodepng.cpp"
       $475 = $474&255; //@line 4411 "libraries/lodepng/lodepng.cpp"
       HEAP32[$38>>2] = $475; //@line 4411 "libraries/lodepng/lodepng.cpp"
       HEAP32[$1>>2] = 0; //@line 4553 "libraries/lodepng/lodepng.cpp"
       $IEND$1$i = 0;$critical_pos$2$i = $critical_pos$0182$i;$idat$sroa$0$3$i = $idat$sroa$0$0180$i;$idat$sroa$14$3$i = $idat$sroa$14$0177$i;$idat$sroa$9$2$i = $idat$sroa$9$0178$i;$unknown$1$i = $unknown$0183$i;
       break L16;
       break;
      }
      default: {
       label = 166;
       break L16;
      }
      }
      break;
     }
     case 98:  {
      $199 = ((($chunk$0184$i)) + 5|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
      $200 = HEAP8[$199>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
      $201 = ($200<<24>>24)==(75); //@line 2407 "libraries/lodepng/lodepng.cpp"
      if ($201) {
       $202 = ((($chunk$0184$i)) + 6|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
       $203 = HEAP8[$202>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
       $204 = ($203<<24>>24)==(71); //@line 2407 "libraries/lodepng/lodepng.cpp"
       if ($204) {
        $205 = ((($chunk$0184$i)) + 7|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
        $206 = HEAP8[$205>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
        $207 = ($206<<24>>24)==(68); //@line 2407 "libraries/lodepng/lodepng.cpp"
        if ($207) {
         $208 = HEAP32[$15>>2]|0; //@line 4185 "libraries/lodepng/lodepng.cpp"
         switch ($208|0) {
         case 3:  {
          $209 = ($65|0)==(1); //@line 4188 "libraries/lodepng/lodepng.cpp"
          if (!($209)) {
           $$0$i38$i = 43;$idat$sroa$0$0180$i$lcssa227 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa241 = $idat$sroa$9$0178$i;
           label = 64;
           break L10;
          }
          HEAP32[$20>>2] = 1; //@line 4190 "libraries/lodepng/lodepng.cpp"
          $210 = HEAP8[$72>>0]|0; //@line 4191 "libraries/lodepng/lodepng.cpp"
          $211 = $210&255; //@line 4191 "libraries/lodepng/lodepng.cpp"
          HEAP32[$21>>2] = $211; //@line 4191 "libraries/lodepng/lodepng.cpp"
          HEAP32[$22>>2] = $211; //@line 4191 "libraries/lodepng/lodepng.cpp"
          HEAP32[$23>>2] = $211; //@line 4191 "libraries/lodepng/lodepng.cpp"
          break;
         }
         case 4: case 0:  {
          $212 = ($65|0)==(2); //@line 4196 "libraries/lodepng/lodepng.cpp"
          if (!($212)) {
           $$0$i38$i = 44;$idat$sroa$0$0180$i$lcssa227 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa241 = $idat$sroa$9$0178$i;
           label = 64;
           break L10;
          }
          HEAP32[$20>>2] = 1; //@line 4198 "libraries/lodepng/lodepng.cpp"
          $213 = HEAP8[$72>>0]|0; //@line 4199 "libraries/lodepng/lodepng.cpp"
          $214 = $213&255; //@line 4199 "libraries/lodepng/lodepng.cpp"
          $215 = $214 << 8; //@line 4199 "libraries/lodepng/lodepng.cpp"
          $216 = ((($chunk$0184$i)) + 9|0); //@line 4199 "libraries/lodepng/lodepng.cpp"
          $217 = HEAP8[$216>>0]|0; //@line 4199 "libraries/lodepng/lodepng.cpp"
          $218 = $217&255; //@line 4199 "libraries/lodepng/lodepng.cpp"
          $219 = $215 | $218; //@line 4199 "libraries/lodepng/lodepng.cpp"
          HEAP32[$21>>2] = $219; //@line 4199 "libraries/lodepng/lodepng.cpp"
          HEAP32[$22>>2] = $219; //@line 4199 "libraries/lodepng/lodepng.cpp"
          HEAP32[$23>>2] = $219; //@line 4199 "libraries/lodepng/lodepng.cpp"
          break;
         }
         case 6: case 2:  {
          $220 = ($65|0)==(6); //@line 4204 "libraries/lodepng/lodepng.cpp"
          if (!($220)) {
           $$0$i38$i = 45;$idat$sroa$0$0180$i$lcssa227 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa241 = $idat$sroa$9$0178$i;
           label = 64;
           break L10;
          }
          HEAP32[$20>>2] = 1; //@line 4206 "libraries/lodepng/lodepng.cpp"
          $221 = HEAP8[$72>>0]|0; //@line 4207 "libraries/lodepng/lodepng.cpp"
          $222 = $221&255; //@line 4207 "libraries/lodepng/lodepng.cpp"
          $223 = $222 << 8; //@line 4207 "libraries/lodepng/lodepng.cpp"
          $224 = ((($chunk$0184$i)) + 9|0); //@line 4207 "libraries/lodepng/lodepng.cpp"
          $225 = HEAP8[$224>>0]|0; //@line 4207 "libraries/lodepng/lodepng.cpp"
          $226 = $225&255; //@line 4207 "libraries/lodepng/lodepng.cpp"
          $227 = $223 | $226; //@line 4207 "libraries/lodepng/lodepng.cpp"
          HEAP32[$23>>2] = $227; //@line 4207 "libraries/lodepng/lodepng.cpp"
          $228 = ((($chunk$0184$i)) + 10|0); //@line 4208 "libraries/lodepng/lodepng.cpp"
          $229 = HEAP8[$228>>0]|0; //@line 4208 "libraries/lodepng/lodepng.cpp"
          $230 = $229&255; //@line 4208 "libraries/lodepng/lodepng.cpp"
          $231 = $230 << 8; //@line 4208 "libraries/lodepng/lodepng.cpp"
          $232 = ((($chunk$0184$i)) + 11|0); //@line 4208 "libraries/lodepng/lodepng.cpp"
          $233 = HEAP8[$232>>0]|0; //@line 4208 "libraries/lodepng/lodepng.cpp"
          $234 = $233&255; //@line 4208 "libraries/lodepng/lodepng.cpp"
          $235 = $231 | $234; //@line 4208 "libraries/lodepng/lodepng.cpp"
          HEAP32[$22>>2] = $235; //@line 4208 "libraries/lodepng/lodepng.cpp"
          $236 = ((($chunk$0184$i)) + 12|0); //@line 4209 "libraries/lodepng/lodepng.cpp"
          $237 = HEAP8[$236>>0]|0; //@line 4209 "libraries/lodepng/lodepng.cpp"
          $238 = $237&255; //@line 4209 "libraries/lodepng/lodepng.cpp"
          $239 = $238 << 8; //@line 4209 "libraries/lodepng/lodepng.cpp"
          $240 = ((($chunk$0184$i)) + 13|0); //@line 4209 "libraries/lodepng/lodepng.cpp"
          $241 = HEAP8[$240>>0]|0; //@line 4209 "libraries/lodepng/lodepng.cpp"
          $242 = $241&255; //@line 4209 "libraries/lodepng/lodepng.cpp"
          $243 = $239 | $242; //@line 4209 "libraries/lodepng/lodepng.cpp"
          HEAP32[$21>>2] = $243; //@line 4209 "libraries/lodepng/lodepng.cpp"
          break;
         }
         default: {
         }
         }
         HEAP32[$1>>2] = 0; //@line 4521 "libraries/lodepng/lodepng.cpp"
         $IEND$1$i = 0;$critical_pos$2$i = $critical_pos$0182$i;$idat$sroa$0$3$i = $idat$sroa$0$0180$i;$idat$sroa$14$3$i = $idat$sroa$14$0177$i;$idat$sroa$9$2$i = $idat$sroa$9$0178$i;$unknown$1$i = $unknown$0183$i;
        } else {
         label = 166;
        }
       } else {
        label = 166;
       }
      } else {
       label = 166;
      }
      break;
     }
     case 122:  {
      $277 = ((($chunk$0184$i)) + 5|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
      $278 = HEAP8[$277>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
      $279 = ($278<<24>>24)==(84); //@line 2407 "libraries/lodepng/lodepng.cpp"
      if ($279) {
       $280 = ((($chunk$0184$i)) + 6|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
       $281 = HEAP8[$280>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
       $282 = ($281<<24>>24)==(88); //@line 2407 "libraries/lodepng/lodepng.cpp"
       if ($282) {
        $283 = ((($chunk$0184$i)) + 7|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
        $284 = HEAP8[$283>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
        $285 = ($284<<24>>24)==(116); //@line 2407 "libraries/lodepng/lodepng.cpp"
        if ($285) {
         $286 = HEAP32[$24>>2]|0; //@line 4536 "libraries/lodepng/lodepng.cpp"
         $287 = ($286|0)==(0); //@line 4536 "libraries/lodepng/lodepng.cpp"
         if ($287) {
          $IEND$1$i = 0;$critical_pos$2$i = $critical_pos$0182$i;$idat$sroa$0$3$i = $idat$sroa$0$0180$i;$idat$sroa$14$3$i = $idat$sroa$14$0177$i;$idat$sroa$9$2$i = $idat$sroa$9$0178$i;$unknown$1$i = $unknown$0183$i;
         } else {
          HEAP32[$decoded$i$i>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
          HEAP32[$27>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
          HEAP32[$28>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
          $288 = ($65|0)==(0); //@line 4273 "libraries/lodepng/lodepng.cpp"
          L100: do {
           if ($288) {
            $length$0$lcssa$i45$i = 0;
           } else {
            $length$03$i$i = 0;
            while(1) {
             $289 = (($72) + ($length$03$i$i)|0); //@line 4273 "libraries/lodepng/lodepng.cpp"
             $290 = HEAP8[$289>>0]|0; //@line 4273 "libraries/lodepng/lodepng.cpp"
             $291 = ($290<<24>>24)==(0); //@line 4273 "libraries/lodepng/lodepng.cpp"
             if ($291) {
              $length$0$lcssa$i45$i = $length$03$i$i;
              break L100;
             }
             $292 = (($length$03$i$i) + 1)|0; //@line 4273 "libraries/lodepng/lodepng.cpp"
             $293 = ($292>>>0)<($65>>>0); //@line 4273 "libraries/lodepng/lodepng.cpp"
             if ($293) {
              $length$03$i$i = $292;
             } else {
              $length$0$lcssa$i45$i = $292;
              break;
             }
            }
           }
          } while(0);
          $294 = (($length$0$lcssa$i45$i) + 2)|0; //@line 4274 "libraries/lodepng/lodepng.cpp"
          $295 = ($65>>>0)>($294>>>0); //@line 4274 "libraries/lodepng/lodepng.cpp"
          do {
           if ($295) {
            $296 = (($length$0$lcssa$i45$i) + -1)|0; //@line 4275 "libraries/lodepng/lodepng.cpp"
            $297 = ($296>>>0)>(78); //@line 4275 "libraries/lodepng/lodepng.cpp"
            if ($297) {
             $error$0$i50$i = 89;$key$0$i51$i = 0;
            } else {
             $298 = (($length$0$lcssa$i45$i) + 1)|0; //@line 4277 "libraries/lodepng/lodepng.cpp"
             $299 = (_malloc($298)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
             $300 = ($299|0)==(0|0); //@line 4278 "libraries/lodepng/lodepng.cpp"
             if ($300) {
              $error$0$i50$i = 83;$key$0$i51$i = 0;
             } else {
              $301 = (($299) + ($length$0$lcssa$i45$i)|0); //@line 4280 "libraries/lodepng/lodepng.cpp"
              HEAP8[$301>>0] = 0; //@line 4280 "libraries/lodepng/lodepng.cpp"
              $302 = ($length$0$lcssa$i45$i|0)==(0); //@line 4281 "libraries/lodepng/lodepng.cpp"
              if (!($302)) {
               _memcpy(($299|0),($72|0),($length$0$lcssa$i45$i|0))|0; //@line 4281 "libraries/lodepng/lodepng.cpp"
              }
              $303 = (($72) + ($298)|0); //@line 4283 "libraries/lodepng/lodepng.cpp"
              $304 = HEAP8[$303>>0]|0; //@line 4283 "libraries/lodepng/lodepng.cpp"
              $305 = ($304<<24>>24)==(0); //@line 4283 "libraries/lodepng/lodepng.cpp"
              if (!($305)) {
               $error$0$i50$i = 72;$key$0$i51$i = $299;
               break;
              }
              $306 = ($65>>>0)<($294>>>0); //@line 4286 "libraries/lodepng/lodepng.cpp"
              if ($306) {
               $error$0$i50$i = 75;$key$0$i51$i = $299;
               break;
              }
              $307 = (($65) - ($294))|0; //@line 4288 "libraries/lodepng/lodepng.cpp"
              $308 = (($72) + ($294)|0); //@line 4291 "libraries/lodepng/lodepng.cpp"
              $309 = HEAP32[$29>>2]|0; //@line 2150 "libraries/lodepng/lodepng.cpp"
              $310 = ($309|0)==(0|0); //@line 2150 "libraries/lodepng/lodepng.cpp"
              if ($310) {
               $312 = (__Z23lodepng_zlib_decompressPPhPjPKhjPK25LodePNGDecompressSettings($decoded$i$i,$28,$308,$307,$26)|0); //@line 2156 "libraries/lodepng/lodepng.cpp"
               $$0$i$i$i = $312;
              } else {
               $311 = (FUNCTION_TABLE_iiiiii[$309 & 8191]($decoded$i$i,$28,$308,$307,$26)|0); //@line 2152 "libraries/lodepng/lodepng.cpp"
               $$0$i$i$i = $311;
              }
              $313 = ($$0$i$i$i|0)==(0); //@line 4293 "libraries/lodepng/lodepng.cpp"
              if (!($313)) {
               $error$0$i50$i = $$0$i$i$i;$key$0$i51$i = $299;
               break;
              }
              $314 = HEAP32[$28>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
              $315 = (($314) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
              $316 = HEAP32[$27>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
              $317 = ($316>>>0)<($315>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
              do {
               if ($317) {
                $318 = $316 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                $319 = ($318>>>0)<($315>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                $320 = ($315*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                $321 = $320 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                $322 = $319 ? $315 : $321; //@line 227 "libraries/lodepng/lodepng.cpp"
                $323 = HEAP32[$decoded$i$i>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                $324 = (_realloc($323,$322)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                $325 = ($324|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                if ($325) {
                 break;
                }
                HEAP32[$27>>2] = $322; //@line 231 "libraries/lodepng/lodepng.cpp"
                HEAP32[$decoded$i$i>>2] = $324; //@line 232 "libraries/lodepng/lodepng.cpp"
                $327 = $324;
                label = 102;
               } else {
                $$pre$i$i49$i = HEAP32[$decoded$i$i>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                $327 = $$pre$i$i49$i;
                label = 102;
               }
              } while(0);
              if ((label|0) == 102) {
               label = 0;
               HEAP32[$28>>2] = $315; //@line 243 "libraries/lodepng/lodepng.cpp"
               $326 = (($327) + ($314)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
               HEAP8[$326>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
              }
              $328 = HEAP32[$decoded$i$i>>2]|0; //@line 4296 "libraries/lodepng/lodepng.cpp"
              $329 = (__Z16lodepng_add_textP11LodePNGInfoPKcS2_($25,$299,$328)|0); //@line 4296 "libraries/lodepng/lodepng.cpp"
              $error$0$i50$i = $329;$key$0$i51$i = $299;
             }
            }
           } else {
            $error$0$i50$i = 75;$key$0$i51$i = 0;
           }
          } while(0);
          _free($key$0$i51$i); //@line 76 "libraries/lodepng/lodepng.cpp"
          HEAP32[$27>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
          HEAP32[$28>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
          $330 = HEAP32[$decoded$i$i>>2]|0; //@line 252 "libraries/lodepng/lodepng.cpp"
          _free($330); //@line 76 "libraries/lodepng/lodepng.cpp"
          HEAP32[$1>>2] = $error$0$i50$i; //@line 4538 "libraries/lodepng/lodepng.cpp"
          $331 = ($error$0$i50$i|0)==(0); //@line 4539 "libraries/lodepng/lodepng.cpp"
          if ($331) {
           $IEND$1$i = 0;$critical_pos$2$i = $critical_pos$0182$i;$idat$sroa$0$3$i = $idat$sroa$0$0180$i;$idat$sroa$14$3$i = $idat$sroa$14$0177$i;$idat$sroa$9$2$i = $idat$sroa$9$0178$i;$unknown$1$i = $unknown$0183$i;
          } else {
           $idat$sroa$0$5$i = $idat$sroa$0$0180$i;$idat$sroa$9$4$i = $idat$sroa$9$0178$i;
           break L10;
          }
         }
        } else {
         label = 166;
        }
       } else {
        label = 166;
       }
      } else {
       label = 166;
      }
      break;
     }
     case 105:  {
      $332 = ((($chunk$0184$i)) + 5|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
      $333 = HEAP8[$332>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
      $334 = ($333<<24>>24)==(84); //@line 2407 "libraries/lodepng/lodepng.cpp"
      if ($334) {
       $335 = ((($chunk$0184$i)) + 6|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
       $336 = HEAP8[$335>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
       $337 = ($336<<24>>24)==(88); //@line 2407 "libraries/lodepng/lodepng.cpp"
       if ($337) {
        $338 = ((($chunk$0184$i)) + 7|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
        $339 = HEAP8[$338>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
        $340 = ($339<<24>>24)==(116); //@line 2407 "libraries/lodepng/lodepng.cpp"
        if ($340) {
         $341 = HEAP32[$24>>2]|0; //@line 4545 "libraries/lodepng/lodepng.cpp"
         $342 = ($341|0)==(0); //@line 4545 "libraries/lodepng/lodepng.cpp"
         if ($342) {
          $IEND$1$i = 0;$critical_pos$2$i = $critical_pos$0182$i;$idat$sroa$0$3$i = $idat$sroa$0$0180$i;$idat$sroa$14$3$i = $idat$sroa$14$0177$i;$idat$sroa$9$2$i = $idat$sroa$9$0178$i;$unknown$1$i = $unknown$0183$i;
         } else {
          HEAP32[$decoded$i55$i>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
          HEAP32[$30>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
          HEAP32[$31>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
          $343 = ($65>>>0)<(5); //@line 4323 "libraries/lodepng/lodepng.cpp"
          L132: do {
           if ($343) {
            $error$0$i71$i = 30;$key$0$i72$i = 0;$langtag$0$i$i = 0;$transkey$0$i$i = 0;
           } else {
            $length$030$i$i = 0;
            while(1) {
             $344 = (($72) + ($length$030$i$i)|0); //@line 4326 "libraries/lodepng/lodepng.cpp"
             $345 = HEAP8[$344>>0]|0; //@line 4326 "libraries/lodepng/lodepng.cpp"
             $346 = ($345<<24>>24)==(0); //@line 4326 "libraries/lodepng/lodepng.cpp"
             if ($346) {
              $length$0$lcssa$i56$i = $length$030$i$i;
              break;
             }
             $347 = (($length$030$i$i) + 1)|0; //@line 4326 "libraries/lodepng/lodepng.cpp"
             $348 = ($347>>>0)<($65>>>0); //@line 4326 "libraries/lodepng/lodepng.cpp"
             if ($348) {
              $length$030$i$i = $347;
             } else {
              $length$0$lcssa$i56$i = $347;
              break;
             }
            }
            $349 = (($length$0$lcssa$i56$i) + 3)|0; //@line 4327 "libraries/lodepng/lodepng.cpp"
            $350 = ($349>>>0)<($65>>>0); //@line 4327 "libraries/lodepng/lodepng.cpp"
            if ($350) {
             $351 = (($length$0$lcssa$i56$i) + -1)|0; //@line 4328 "libraries/lodepng/lodepng.cpp"
             $352 = ($351>>>0)>(78); //@line 4328 "libraries/lodepng/lodepng.cpp"
             if ($352) {
              $error$0$i71$i = 89;$key$0$i72$i = 0;$langtag$0$i$i = 0;$transkey$0$i$i = 0;
             } else {
              $353 = (($length$0$lcssa$i56$i) + 1)|0; //@line 4330 "libraries/lodepng/lodepng.cpp"
              $354 = (_malloc($353)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
              $355 = ($354|0)==(0|0); //@line 4331 "libraries/lodepng/lodepng.cpp"
              if ($355) {
               $error$0$i71$i = 83;$key$0$i72$i = 0;$langtag$0$i$i = 0;$transkey$0$i$i = 0;
              } else {
               $356 = (($354) + ($length$0$lcssa$i56$i)|0); //@line 4333 "libraries/lodepng/lodepng.cpp"
               HEAP8[$356>>0] = 0; //@line 4333 "libraries/lodepng/lodepng.cpp"
               $357 = ($length$0$lcssa$i56$i|0)==(0); //@line 4334 "libraries/lodepng/lodepng.cpp"
               if (!($357)) {
                _memcpy(($354|0),($72|0),($length$0$lcssa$i56$i|0))|0; //@line 4334 "libraries/lodepng/lodepng.cpp"
               }
               $358 = (($72) + ($353)|0); //@line 4337 "libraries/lodepng/lodepng.cpp"
               $359 = HEAP8[$358>>0]|0; //@line 4337 "libraries/lodepng/lodepng.cpp"
               $360 = (($length$0$lcssa$i56$i) + 2)|0; //@line 4338 "libraries/lodepng/lodepng.cpp"
               $361 = (($72) + ($360)|0); //@line 4338 "libraries/lodepng/lodepng.cpp"
               $362 = HEAP8[$361>>0]|0; //@line 4338 "libraries/lodepng/lodepng.cpp"
               $363 = ($362<<24>>24)==(0); //@line 4338 "libraries/lodepng/lodepng.cpp"
               if ($363) {
                $i$122$i$i = $349;$length$123$i$i = 0;
               } else {
                $error$0$i71$i = 72;$key$0$i72$i = $354;$langtag$0$i$i = 0;$transkey$0$i$i = 0;
                break;
               }
               while(1) {
                $367 = (($72) + ($i$122$i$i)|0); //@line 4346 "libraries/lodepng/lodepng.cpp"
                $368 = HEAP8[$367>>0]|0; //@line 4346 "libraries/lodepng/lodepng.cpp"
                $369 = ($368<<24>>24)==(0); //@line 4346 "libraries/lodepng/lodepng.cpp"
                $370 = (($length$123$i$i) + 1)|0; //@line 4346 "libraries/lodepng/lodepng.cpp"
                $365 = (($i$122$i$i) + 1)|0; //@line 4346 "libraries/lodepng/lodepng.cpp"
                if ($369) {
                 $371 = $370;$length$19$i$i = $length$123$i$i;
                 break;
                }
                $364 = ($365>>>0)<($65>>>0); //@line 4346 "libraries/lodepng/lodepng.cpp"
                if ($364) {
                 $i$122$i$i = $365;$length$123$i$i = $370;
                } else {
                 $$lcssa218 = $370;$length$123$i$i$lcssa216 = $length$123$i$i;
                 label = 119;
                 break;
                }
               }
               if ((label|0) == 119) {
                label = 0;
                $366 = (($length$123$i$i$lcssa216) + 2)|0; //@line 4346 "libraries/lodepng/lodepng.cpp"
                $371 = $366;$length$19$i$i = $$lcssa218;
               }
               $372 = (_malloc($371)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
               $373 = ($372|0)==(0|0); //@line 4349 "libraries/lodepng/lodepng.cpp"
               if ($373) {
                $error$0$i71$i = 83;$key$0$i72$i = $354;$langtag$0$i$i = 0;$transkey$0$i$i = 0;
                break;
               }
               $374 = (($372) + ($length$19$i$i)|0); //@line 4351 "libraries/lodepng/lodepng.cpp"
               HEAP8[$374>>0] = 0; //@line 4351 "libraries/lodepng/lodepng.cpp"
               $375 = ($length$19$i$i|0)==(0); //@line 4352 "libraries/lodepng/lodepng.cpp"
               if (!($375)) {
                $scevgep40$i$i = (($72) + ($349)|0); //@line 4352 "libraries/lodepng/lodepng.cpp"
                _memcpy(($372|0),($scevgep40$i$i|0),($length$19$i$i|0))|0; //@line 4352 "libraries/lodepng/lodepng.cpp"
               }
               $376 = (($371) + ($349))|0; //@line 4355 "libraries/lodepng/lodepng.cpp"
               $377 = ($376>>>0)<($65>>>0); //@line 4357 "libraries/lodepng/lodepng.cpp"
               L153: do {
                if ($377) {
                 $i$314$i$i = $376;$length$215$i$i = 0;
                 while(1) {
                  $381 = (($72) + ($i$314$i$i)|0); //@line 4357 "libraries/lodepng/lodepng.cpp"
                  $382 = HEAP8[$381>>0]|0; //@line 4357 "libraries/lodepng/lodepng.cpp"
                  $383 = ($382<<24>>24)==(0); //@line 4357 "libraries/lodepng/lodepng.cpp"
                  $384 = (($length$215$i$i) + 1)|0; //@line 4357 "libraries/lodepng/lodepng.cpp"
                  $379 = (($i$314$i$i) + 1)|0; //@line 4357 "libraries/lodepng/lodepng.cpp"
                  if ($383) {
                   $385 = $384;$length$27$i$i = $length$215$i$i;
                   break L153;
                  }
                  $378 = ($379>>>0)<($65>>>0); //@line 4357 "libraries/lodepng/lodepng.cpp"
                  if ($378) {
                   $i$314$i$i = $379;$length$215$i$i = $384;
                  } else {
                   $length$2$lcssa$i$i = $384;
                   label = 126;
                   break;
                  }
                 }
                } else {
                 $length$2$lcssa$i$i = 0;
                 label = 126;
                }
               } while(0);
               if ((label|0) == 126) {
                label = 0;
                $380 = (($length$2$lcssa$i$i) + 1)|0; //@line 4357 "libraries/lodepng/lodepng.cpp"
                $385 = $380;$length$27$i$i = $length$2$lcssa$i$i;
               }
               $386 = (_malloc($385)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
               $387 = ($386|0)==(0|0); //@line 4360 "libraries/lodepng/lodepng.cpp"
               if ($387) {
                $error$0$i71$i = 83;$key$0$i72$i = $354;$langtag$0$i$i = $372;$transkey$0$i$i = 0;
                break;
               }
               $388 = (($386) + ($length$27$i$i)|0); //@line 4362 "libraries/lodepng/lodepng.cpp"
               HEAP8[$388>>0] = 0; //@line 4362 "libraries/lodepng/lodepng.cpp"
               $389 = ($length$27$i$i|0)==(0); //@line 4363 "libraries/lodepng/lodepng.cpp"
               if (!($389)) {
                $scevgep$i60$i = (($72) + ($376)|0); //@line 4363 "libraries/lodepng/lodepng.cpp"
                _memcpy(($386|0),($scevgep$i60$i|0),($length$27$i$i|0))|0; //@line 4363 "libraries/lodepng/lodepng.cpp"
               }
               $390 = (($385) + ($376))|0; //@line 4366 "libraries/lodepng/lodepng.cpp"
               $391 = ($65>>>0)<($390>>>0); //@line 4368 "libraries/lodepng/lodepng.cpp"
               $392 = (($65) - ($390))|0; //@line 4368 "libraries/lodepng/lodepng.cpp"
               $$$i61$i = $391 ? 0 : $392; //@line 4368 "libraries/lodepng/lodepng.cpp"
               $393 = ($359<<24>>24)==(0); //@line 4370 "libraries/lodepng/lodepng.cpp"
               do {
                if ($393) {
                 $416 = (($$$i61$i) + 1)|0; //@line 4382 "libraries/lodepng/lodepng.cpp"
                 $417 = ($416|0)==(0); //@line 225 "libraries/lodepng/lodepng.cpp"
                 if ($417) {
                  $421 = 0;
                 } else {
                  $418 = (_realloc(0,$416)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                  $419 = ($418|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                  if ($419) {
                   $error$0$i71$i = 83;$key$0$i72$i = $354;$langtag$0$i$i = $372;$transkey$0$i$i = $386;
                   break L132;
                  }
                  HEAP32[$30>>2] = $416; //@line 231 "libraries/lodepng/lodepng.cpp"
                  HEAP32[$decoded$i55$i>>2] = $418; //@line 232 "libraries/lodepng/lodepng.cpp"
                  $421 = $418;
                 }
                 HEAP32[$31>>2] = $416; //@line 243 "libraries/lodepng/lodepng.cpp"
                 $420 = (($421) + ($$$i61$i)|0); //@line 4384 "libraries/lodepng/lodepng.cpp"
                 HEAP8[$420>>0] = 0; //@line 4384 "libraries/lodepng/lodepng.cpp"
                 $422 = ($$$i61$i|0)==(0); //@line 4385 "libraries/lodepng/lodepng.cpp"
                 if ($422) {
                  break;
                 }
                 $423 = (($72) + ($390)|0); //@line 4385 "libraries/lodepng/lodepng.cpp"
                 $424 = HEAP8[$423>>0]|0; //@line 4385 "libraries/lodepng/lodepng.cpp"
                 HEAP8[$421>>0] = $424; //@line 4385 "libraries/lodepng/lodepng.cpp"
                 $425 = ($$$i61$i|0)==(1); //@line 4385 "libraries/lodepng/lodepng.cpp"
                 if ($425) {
                  break;
                 }
                 $426 = (($390) + 1)|0; //@line 4385 "libraries/lodepng/lodepng.cpp"
                 $427 = (($72) + ($426)|0); //@line 4385 "libraries/lodepng/lodepng.cpp"
                 $428 = HEAP8[$427>>0]|0; //@line 4385 "libraries/lodepng/lodepng.cpp"
                 $429 = ((($421)) + 1|0); //@line 4385 "libraries/lodepng/lodepng.cpp"
                 HEAP8[$429>>0] = $428; //@line 4385 "libraries/lodepng/lodepng.cpp"
                 $430 = ($$$i61$i|0)==(2); //@line 4385 "libraries/lodepng/lodepng.cpp"
                 if ($430) {
                  break;
                 }
                 $431 = (($390) + 2)|0; //@line 4385 "libraries/lodepng/lodepng.cpp"
                 $432 = (($72) + ($431)|0); //@line 4385 "libraries/lodepng/lodepng.cpp"
                 $433 = HEAP8[$432>>0]|0; //@line 4385 "libraries/lodepng/lodepng.cpp"
                 $434 = ((($421)) + 2|0); //@line 4385 "libraries/lodepng/lodepng.cpp"
                 HEAP8[$434>>0] = $433; //@line 4385 "libraries/lodepng/lodepng.cpp"
                 $435 = ($$$i61$i|0)==(3); //@line 4385 "libraries/lodepng/lodepng.cpp"
                 if ($435) {
                  break;
                 } else {
                  $437 = 3;
                 }
                 while(1) {
                  $$pre$i69$pre$i$pre = HEAP32[$decoded$i55$i>>2]|0; //@line 4385 "libraries/lodepng/lodepng.cpp"
                  $436 = (($437) + ($390))|0; //@line 4385 "libraries/lodepng/lodepng.cpp"
                  $438 = (($72) + ($436)|0); //@line 4385 "libraries/lodepng/lodepng.cpp"
                  $439 = HEAP8[$438>>0]|0; //@line 4385 "libraries/lodepng/lodepng.cpp"
                  $440 = (($$pre$i69$pre$i$pre) + ($437)|0); //@line 4385 "libraries/lodepng/lodepng.cpp"
                  HEAP8[$440>>0] = $439; //@line 4385 "libraries/lodepng/lodepng.cpp"
                  $441 = (($437) + 1)|0; //@line 4385 "libraries/lodepng/lodepng.cpp"
                  $442 = ($441|0)==($$$i61$i|0); //@line 4385 "libraries/lodepng/lodepng.cpp"
                  if ($442) {
                   break;
                  } else {
                   $437 = $441;
                  }
                 }
                } else {
                 $394 = (($72) + ($390)|0); //@line 4374 "libraries/lodepng/lodepng.cpp"
                 $395 = HEAP32[$29>>2]|0; //@line 2150 "libraries/lodepng/lodepng.cpp"
                 $396 = ($395|0)==(0|0); //@line 2150 "libraries/lodepng/lodepng.cpp"
                 if ($396) {
                  $398 = (__Z23lodepng_zlib_decompressPPhPjPKhjPK25LodePNGDecompressSettings($decoded$i55$i,$31,$394,$$$i61$i,$26)|0); //@line 2156 "libraries/lodepng/lodepng.cpp"
                  $$0$i$i63$i = $398;
                 } else {
                  $397 = (FUNCTION_TABLE_iiiiii[$395 & 8191]($decoded$i55$i,$31,$394,$$$i61$i,$26)|0); //@line 2152 "libraries/lodepng/lodepng.cpp"
                  $$0$i$i63$i = $397;
                 }
                 $399 = ($$0$i$i63$i|0)==(0); //@line 4376 "libraries/lodepng/lodepng.cpp"
                 if (!($399)) {
                  $error$0$i71$i = $$0$i$i63$i;$key$0$i72$i = $354;$langtag$0$i$i = $372;$transkey$0$i$i = $386;
                  break L132;
                 }
                 $400 = HEAP32[$30>>2]|0; //@line 4377 "libraries/lodepng/lodepng.cpp"
                 $401 = HEAP32[$31>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                 $402 = ($400>>>0)<($401>>>0); //@line 4377 "libraries/lodepng/lodepng.cpp"
                 if ($402) {
                  HEAP32[$30>>2] = $401; //@line 4377 "libraries/lodepng/lodepng.cpp"
                  $405 = $401;
                 } else {
                  $405 = $400;
                 }
                 $403 = (($401) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
                 $404 = ($405>>>0)<($403>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
                 if ($404) {
                  $406 = $405 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                  $407 = ($406>>>0)<($403>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
                  $408 = ($403*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
                  $409 = $408 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
                  $410 = $407 ? $403 : $409; //@line 227 "libraries/lodepng/lodepng.cpp"
                  $411 = HEAP32[$decoded$i55$i>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
                  $412 = (_realloc($411,$410)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
                  $413 = ($412|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
                  if ($413) {
                   break;
                  }
                  HEAP32[$30>>2] = $410; //@line 231 "libraries/lodepng/lodepng.cpp"
                  HEAP32[$decoded$i55$i>>2] = $412; //@line 232 "libraries/lodepng/lodepng.cpp"
                  $415 = $412;
                 } else {
                  $$pre$i$i65$i = HEAP32[$decoded$i55$i>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
                  $415 = $$pre$i$i65$i;
                 }
                 HEAP32[$31>>2] = $403; //@line 243 "libraries/lodepng/lodepng.cpp"
                 $414 = (($415) + ($401)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
                 HEAP8[$414>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
                }
               } while(0);
               $443 = HEAP32[$decoded$i55$i>>2]|0; //@line 4388 "libraries/lodepng/lodepng.cpp"
               $444 = (__Z17lodepng_add_itextP11LodePNGInfoPKcS2_S2_S2_($25,$354,$372,$386,$443)|0); //@line 4388 "libraries/lodepng/lodepng.cpp"
               $error$0$i71$i = $444;$key$0$i72$i = $354;$langtag$0$i$i = $372;$transkey$0$i$i = $386;
              }
             }
            } else {
             $error$0$i71$i = 75;$key$0$i72$i = 0;$langtag$0$i$i = 0;$transkey$0$i$i = 0;
            }
           }
          } while(0);
          _free($key$0$i72$i); //@line 76 "libraries/lodepng/lodepng.cpp"
          _free($langtag$0$i$i); //@line 76 "libraries/lodepng/lodepng.cpp"
          _free($transkey$0$i$i); //@line 76 "libraries/lodepng/lodepng.cpp"
          HEAP32[$30>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
          HEAP32[$31>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
          $445 = HEAP32[$decoded$i55$i>>2]|0; //@line 252 "libraries/lodepng/lodepng.cpp"
          _free($445); //@line 76 "libraries/lodepng/lodepng.cpp"
          HEAP32[$1>>2] = $error$0$i71$i; //@line 4547 "libraries/lodepng/lodepng.cpp"
          $446 = ($error$0$i71$i|0)==(0); //@line 4548 "libraries/lodepng/lodepng.cpp"
          if ($446) {
           $IEND$1$i = 0;$critical_pos$2$i = $critical_pos$0182$i;$idat$sroa$0$3$i = $idat$sroa$0$0180$i;$idat$sroa$14$3$i = $idat$sroa$14$0177$i;$idat$sroa$9$2$i = $idat$sroa$9$0178$i;$unknown$1$i = $unknown$0183$i;
          } else {
           $idat$sroa$0$5$i = $idat$sroa$0$0180$i;$idat$sroa$9$4$i = $idat$sroa$9$0178$i;
           break L10;
          }
         }
        } else {
         label = 166;
        }
       } else {
        label = 166;
       }
      } else {
       label = 166;
      }
      break;
     }
     case 112:  {
      $476 = ((($chunk$0184$i)) + 5|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
      $477 = HEAP8[$476>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
      $478 = ($477<<24>>24)==(72); //@line 2407 "libraries/lodepng/lodepng.cpp"
      if ($478) {
       $479 = ((($chunk$0184$i)) + 6|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
       $480 = HEAP8[$479>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
       $481 = ($480<<24>>24)==(89); //@line 2407 "libraries/lodepng/lodepng.cpp"
       if ($481) {
        $482 = ((($chunk$0184$i)) + 7|0); //@line 2407 "libraries/lodepng/lodepng.cpp"
        $483 = HEAP8[$482>>0]|0; //@line 2407 "libraries/lodepng/lodepng.cpp"
        $484 = ($483<<24>>24)==(115); //@line 2407 "libraries/lodepng/lodepng.cpp"
        if ($484) {
         $485 = ($65|0)==(9); //@line 4418 "libraries/lodepng/lodepng.cpp"
         if (!($485)) {
          $idat$sroa$0$0180$i$lcssa225 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa239 = $idat$sroa$9$0178$i;
          label = 163;
          break L10;
         }
         HEAP32[$39>>2] = 1; //@line 4420 "libraries/lodepng/lodepng.cpp"
         $486 = HEAP8[$72>>0]|0; //@line 4421 "libraries/lodepng/lodepng.cpp"
         $487 = $486&255; //@line 4421 "libraries/lodepng/lodepng.cpp"
         $488 = $487 << 24; //@line 4421 "libraries/lodepng/lodepng.cpp"
         $489 = ((($chunk$0184$i)) + 9|0); //@line 4421 "libraries/lodepng/lodepng.cpp"
         $490 = HEAP8[$489>>0]|0; //@line 4421 "libraries/lodepng/lodepng.cpp"
         $491 = $490&255; //@line 4421 "libraries/lodepng/lodepng.cpp"
         $492 = $491 << 16; //@line 4421 "libraries/lodepng/lodepng.cpp"
         $493 = $492 | $488; //@line 4421 "libraries/lodepng/lodepng.cpp"
         $494 = ((($chunk$0184$i)) + 10|0); //@line 4421 "libraries/lodepng/lodepng.cpp"
         $495 = HEAP8[$494>>0]|0; //@line 4421 "libraries/lodepng/lodepng.cpp"
         $496 = $495&255; //@line 4421 "libraries/lodepng/lodepng.cpp"
         $497 = $496 << 8; //@line 4421 "libraries/lodepng/lodepng.cpp"
         $498 = $493 | $497; //@line 4421 "libraries/lodepng/lodepng.cpp"
         $499 = ((($chunk$0184$i)) + 11|0); //@line 4421 "libraries/lodepng/lodepng.cpp"
         $500 = HEAP8[$499>>0]|0; //@line 4421 "libraries/lodepng/lodepng.cpp"
         $501 = $500&255; //@line 4421 "libraries/lodepng/lodepng.cpp"
         $502 = $498 | $501; //@line 4421 "libraries/lodepng/lodepng.cpp"
         HEAP32[$40>>2] = $502; //@line 4421 "libraries/lodepng/lodepng.cpp"
         $503 = ((($chunk$0184$i)) + 12|0); //@line 4422 "libraries/lodepng/lodepng.cpp"
         $504 = HEAP8[$503>>0]|0; //@line 4422 "libraries/lodepng/lodepng.cpp"
         $505 = $504&255; //@line 4422 "libraries/lodepng/lodepng.cpp"
         $506 = $505 << 24; //@line 4422 "libraries/lodepng/lodepng.cpp"
         $507 = ((($chunk$0184$i)) + 13|0); //@line 4422 "libraries/lodepng/lodepng.cpp"
         $508 = HEAP8[$507>>0]|0; //@line 4422 "libraries/lodepng/lodepng.cpp"
         $509 = $508&255; //@line 4422 "libraries/lodepng/lodepng.cpp"
         $510 = $509 << 16; //@line 4422 "libraries/lodepng/lodepng.cpp"
         $511 = $510 | $506; //@line 4422 "libraries/lodepng/lodepng.cpp"
         $512 = ((($chunk$0184$i)) + 14|0); //@line 4422 "libraries/lodepng/lodepng.cpp"
         $513 = HEAP8[$512>>0]|0; //@line 4422 "libraries/lodepng/lodepng.cpp"
         $514 = $513&255; //@line 4422 "libraries/lodepng/lodepng.cpp"
         $515 = $514 << 8; //@line 4422 "libraries/lodepng/lodepng.cpp"
         $516 = $511 | $515; //@line 4422 "libraries/lodepng/lodepng.cpp"
         $517 = ((($chunk$0184$i)) + 15|0); //@line 4422 "libraries/lodepng/lodepng.cpp"
         $518 = HEAP8[$517>>0]|0; //@line 4422 "libraries/lodepng/lodepng.cpp"
         $519 = $518&255; //@line 4422 "libraries/lodepng/lodepng.cpp"
         $520 = $516 | $519; //@line 4422 "libraries/lodepng/lodepng.cpp"
         HEAP32[$41>>2] = $520; //@line 4422 "libraries/lodepng/lodepng.cpp"
         $521 = ((($chunk$0184$i)) + 16|0); //@line 4423 "libraries/lodepng/lodepng.cpp"
         $522 = HEAP8[$521>>0]|0; //@line 4423 "libraries/lodepng/lodepng.cpp"
         $523 = $522&255; //@line 4423 "libraries/lodepng/lodepng.cpp"
         HEAP32[$42>>2] = $523; //@line 4423 "libraries/lodepng/lodepng.cpp"
         HEAP32[$1>>2] = 0; //@line 4558 "libraries/lodepng/lodepng.cpp"
         $IEND$1$i = 0;$critical_pos$2$i = $critical_pos$0182$i;$idat$sroa$0$3$i = $idat$sroa$0$0180$i;$idat$sroa$14$3$i = $idat$sroa$14$0177$i;$idat$sroa$9$2$i = $idat$sroa$9$0178$i;$unknown$1$i = $unknown$0183$i;
        } else {
         label = 166;
        }
       } else {
        label = 166;
       }
      } else {
       label = 166;
      }
      break;
     }
     default: {
      $$lobit$i$i = $74 & 32; //@line 2412 "libraries/lodepng/lodepng.cpp"
      $524 = ($$lobit$i$i<<24>>24)==(0); //@line 4565 "libraries/lodepng/lodepng.cpp"
      if ($524) {
       $idat$sroa$0$0180$i$lcssa230 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa244 = $idat$sroa$9$0178$i;
       label = 165;
       break L10;
      } else {
       label = 166;
      }
     }
     }
    } while(0);
    if ((label|0) == 166) {
     label = 0;
     $525 = HEAP32[$43>>2]|0; //@line 4569 "libraries/lodepng/lodepng.cpp"
     $526 = ($525|0)==(0); //@line 4569 "libraries/lodepng/lodepng.cpp"
     if ($526) {
      $IEND$1$i = 0;$critical_pos$2$i = $critical_pos$0182$i;$idat$sroa$0$3$i = $idat$sroa$0$0180$i;$idat$sroa$14$3$i = $idat$sroa$14$0177$i;$idat$sroa$9$2$i = $idat$sroa$9$0178$i;$unknown$1$i = 1;
     } else {
      $527 = (($critical_pos$0182$i) + -1)|0; //@line 4571 "libraries/lodepng/lodepng.cpp"
      $528 = (((($state)) + 268|0) + ($527<<2)|0); //@line 4571 "libraries/lodepng/lodepng.cpp"
      $529 = (((($state)) + 280|0) + ($527<<2)|0); //@line 4572 "libraries/lodepng/lodepng.cpp"
      $530 = (($65) + 12)|0; //@line 2467 "libraries/lodepng/lodepng.cpp"
      $531 = HEAP32[$529>>2]|0; //@line 2469 "libraries/lodepng/lodepng.cpp"
      $532 = (($531) + ($530))|0; //@line 2469 "libraries/lodepng/lodepng.cpp"
      $533 = ($532>>>0)<($530>>>0); //@line 2470 "libraries/lodepng/lodepng.cpp"
      $534 = ($532>>>0)<($531>>>0); //@line 2470 "libraries/lodepng/lodepng.cpp"
      $or$cond$i$i = $533 | $534; //@line 2470 "libraries/lodepng/lodepng.cpp"
      if ($or$cond$i$i) {
       $$0$i83$i = 77;$idat$sroa$0$0180$i$lcssa234 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa248 = $idat$sroa$9$0178$i;
       label = 172;
       break;
      }
      $535 = HEAP32[$528>>2]|0; //@line 2472 "libraries/lodepng/lodepng.cpp"
      $536 = (_realloc($535,$532)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $537 = ($536|0)==(0|0); //@line 2473 "libraries/lodepng/lodepng.cpp"
      if ($537) {
       $$0$i83$i = 83;$idat$sroa$0$0180$i$lcssa234 = $idat$sroa$0$0180$i;$idat$sroa$9$0178$i$lcssa248 = $idat$sroa$9$0178$i;
       label = 172;
       break;
      }
      HEAP32[$528>>2] = $536; //@line 2474 "libraries/lodepng/lodepng.cpp"
      HEAP32[$529>>2] = $532; //@line 2475 "libraries/lodepng/lodepng.cpp"
      $538 = (($536) + ($531)|0); //@line 2476 "libraries/lodepng/lodepng.cpp"
      $i$01$i81$i = 0;
      while(1) {
       $539 = (($chunk$0184$i) + ($i$01$i81$i)|0); //@line 2478 "libraries/lodepng/lodepng.cpp"
       $540 = HEAP8[$539>>0]|0; //@line 2478 "libraries/lodepng/lodepng.cpp"
       $541 = (($538) + ($i$01$i81$i)|0); //@line 2478 "libraries/lodepng/lodepng.cpp"
       HEAP8[$541>>0] = $540; //@line 2478 "libraries/lodepng/lodepng.cpp"
       $542 = (($i$01$i81$i) + 1)|0; //@line 2478 "libraries/lodepng/lodepng.cpp"
       $543 = ($542|0)==($530|0); //@line 2478 "libraries/lodepng/lodepng.cpp"
       if ($543) {
        break;
       } else {
        $i$01$i81$i = $542;
       }
      }
      HEAP32[$1>>2] = 0; //@line 4571 "libraries/lodepng/lodepng.cpp"
      $IEND$1$i = 0;$critical_pos$2$i = $critical_pos$0182$i;$idat$sroa$0$3$i = $idat$sroa$0$0180$i;$idat$sroa$14$3$i = $idat$sroa$14$0177$i;$idat$sroa$9$2$i = $idat$sroa$9$0178$i;$unknown$1$i = 1;
     }
    }
    $544 = HEAP32[$12>>2]|0; //@line 4578 "libraries/lodepng/lodepng.cpp"
    $545 = $544 | $unknown$1$i; //@line 4578 "libraries/lodepng/lodepng.cpp"
    $546 = ($545|0)==(0); //@line 4578 "libraries/lodepng/lodepng.cpp"
    if ($546) {
     $547 = HEAP8[$chunk$0184$i>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
     $548 = $547&255; //@line 343 "libraries/lodepng/lodepng.cpp"
     $549 = $548 << 24; //@line 343 "libraries/lodepng/lodepng.cpp"
     $550 = HEAP8[$52>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
     $551 = $550&255; //@line 343 "libraries/lodepng/lodepng.cpp"
     $552 = $551 << 16; //@line 343 "libraries/lodepng/lodepng.cpp"
     $553 = $552 | $549; //@line 343 "libraries/lodepng/lodepng.cpp"
     $554 = HEAP8[$57>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
     $555 = $554&255; //@line 343 "libraries/lodepng/lodepng.cpp"
     $556 = $555 << 8; //@line 343 "libraries/lodepng/lodepng.cpp"
     $557 = $553 | $556; //@line 343 "libraries/lodepng/lodepng.cpp"
     $558 = HEAP8[$62>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
     $559 = $558&255; //@line 343 "libraries/lodepng/lodepng.cpp"
     $560 = $557 | $559; //@line 343 "libraries/lodepng/lodepng.cpp"
     $561 = (($560) + 8)|0; //@line 2438 "libraries/lodepng/lodepng.cpp"
     $562 = (($chunk$0184$i) + ($561)|0); //@line 2438 "libraries/lodepng/lodepng.cpp"
     $563 = HEAP8[$562>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
     $564 = $563&255; //@line 343 "libraries/lodepng/lodepng.cpp"
     $565 = $564 << 24; //@line 343 "libraries/lodepng/lodepng.cpp"
     $566 = ((($562)) + 1|0); //@line 343 "libraries/lodepng/lodepng.cpp"
     $567 = HEAP8[$566>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
     $568 = $567&255; //@line 343 "libraries/lodepng/lodepng.cpp"
     $569 = $568 << 16; //@line 343 "libraries/lodepng/lodepng.cpp"
     $570 = $569 | $565; //@line 343 "libraries/lodepng/lodepng.cpp"
     $571 = ((($562)) + 2|0); //@line 343 "libraries/lodepng/lodepng.cpp"
     $572 = HEAP8[$571>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
     $573 = $572&255; //@line 343 "libraries/lodepng/lodepng.cpp"
     $574 = $573 << 8; //@line 343 "libraries/lodepng/lodepng.cpp"
     $575 = $570 | $574; //@line 343 "libraries/lodepng/lodepng.cpp"
     $576 = ((($562)) + 3|0); //@line 343 "libraries/lodepng/lodepng.cpp"
     $577 = HEAP8[$576>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
     $578 = $577&255; //@line 343 "libraries/lodepng/lodepng.cpp"
     $579 = $575 | $578; //@line 343 "libraries/lodepng/lodepng.cpp"
     $580 = (($560) + 4)|0; //@line 2440 "libraries/lodepng/lodepng.cpp"
     $581 = ($580|0)==(0); //@line 2338 "libraries/lodepng/lodepng.cpp"
     if ($581) {
      $c$0$lcssa$i$i$i = 0;
     } else {
      $c$01$i$i$i = -1;$n$02$i$i$i = 0;
      while(1) {
       $582 = (($73) + ($n$02$i$i$i)|0); //@line 2340 "libraries/lodepng/lodepng.cpp"
       $583 = HEAP8[$582>>0]|0; //@line 2340 "libraries/lodepng/lodepng.cpp"
       $584 = $583&255; //@line 2340 "libraries/lodepng/lodepng.cpp"
       $c$0$masked$i$i$i = $c$01$i$i$i & 255; //@line 2340 "libraries/lodepng/lodepng.cpp"
       $585 = $584 ^ $c$0$masked$i$i$i; //@line 2340 "libraries/lodepng/lodepng.cpp"
       $586 = (33860 + ($585<<2)|0); //@line 2340 "libraries/lodepng/lodepng.cpp"
       $587 = HEAP32[$586>>2]|0; //@line 2340 "libraries/lodepng/lodepng.cpp"
       $588 = $c$01$i$i$i >>> 8; //@line 2340 "libraries/lodepng/lodepng.cpp"
       $589 = $587 ^ $588; //@line 2340 "libraries/lodepng/lodepng.cpp"
       $590 = (($n$02$i$i$i) + 1)|0; //@line 2338 "libraries/lodepng/lodepng.cpp"
       $exitcond$i$i85$i = ($590|0)==($580|0); //@line 2338 "libraries/lodepng/lodepng.cpp"
       if ($exitcond$i$i85$i) {
        $$lcssa222 = $589;
        break;
       } else {
        $c$01$i$i$i = $589;$n$02$i$i$i = $590;
       }
      }
      $phitmp$i$i$i = $$lcssa222 ^ -1; //@line 2342 "libraries/lodepng/lodepng.cpp"
      $c$0$lcssa$i$i$i = $phitmp$i$i$i;
     }
     $not$$i$i = ($579|0)==($c$0$lcssa$i$i$i|0); //@line 2442 "libraries/lodepng/lodepng.cpp"
     if (!($not$$i$i)) {
      $idat$sroa$0$3$i$lcssa = $idat$sroa$0$3$i;$idat$sroa$9$2$i$lcssa = $idat$sroa$9$2$i;
      label = 178;
      break;
     }
    }
    $591 = ($IEND$1$i<<24>>24)==(0); //@line 4583 "libraries/lodepng/lodepng.cpp"
    if (!($591)) {
     $idat$sroa$0$5$i = $idat$sroa$0$3$i;$idat$sroa$9$4$i = $idat$sroa$9$2$i;
     break;
    }
    $592 = HEAP8[$chunk$0184$i>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
    $593 = $592&255; //@line 343 "libraries/lodepng/lodepng.cpp"
    $594 = $593 << 24; //@line 343 "libraries/lodepng/lodepng.cpp"
    $595 = HEAP8[$52>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
    $596 = $595&255; //@line 343 "libraries/lodepng/lodepng.cpp"
    $597 = $596 << 16; //@line 343 "libraries/lodepng/lodepng.cpp"
    $598 = $597 | $594; //@line 343 "libraries/lodepng/lodepng.cpp"
    $599 = HEAP8[$57>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
    $600 = $599&255; //@line 343 "libraries/lodepng/lodepng.cpp"
    $601 = $600 << 8; //@line 343 "libraries/lodepng/lodepng.cpp"
    $602 = $598 | $601; //@line 343 "libraries/lodepng/lodepng.cpp"
    $603 = HEAP8[$62>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
    $604 = $603&255; //@line 343 "libraries/lodepng/lodepng.cpp"
    $605 = $602 | $604; //@line 343 "libraries/lodepng/lodepng.cpp"
    $606 = (($605) + 12)|0; //@line 2460 "libraries/lodepng/lodepng.cpp"
    $607 = (($chunk$0184$i) + ($606)|0); //@line 2461 "libraries/lodepng/lodepng.cpp"
    $608 = HEAP32[$1>>2]|0; //@line 4467 "libraries/lodepng/lodepng.cpp"
    $609 = ($608|0)==(0); //@line 4467 "libraries/lodepng/lodepng.cpp"
    if ($609) {
     $chunk$0184$i = $607;$critical_pos$0182$i = $critical_pos$2$i;$idat$sroa$0$0180$i = $idat$sroa$0$3$i;$idat$sroa$14$0177$i = $idat$sroa$14$3$i;$idat$sroa$9$0178$i = $idat$sroa$9$2$i;$unknown$0183$i = $unknown$1$i;
    } else {
     $idat$sroa$0$5$i = $idat$sroa$0$3$i;$idat$sroa$9$4$i = $idat$sroa$9$2$i;
     break;
    }
   }
   switch (label|0) {
    case 9: {
     HEAP32[$1>>2] = 30; //@line 4473 "libraries/lodepng/lodepng.cpp"
     $idat$sroa$0$5$i = $idat$sroa$0$0180$i$lcssa;$idat$sroa$9$4$i = $idat$sroa$9$0178$i$lcssa;
     break;
    }
    case 11: {
     HEAP32[$1>>2] = 63; //@line 4478 "libraries/lodepng/lodepng.cpp"
     $idat$sroa$0$5$i = $idat$sroa$0$0180$i$lcssa223;$idat$sroa$9$4$i = $idat$sroa$9$0178$i$lcssa237;
     break;
    }
    case 14: {
     HEAP32[$1>>2] = 64; //@line 4482 "libraries/lodepng/lodepng.cpp"
     $idat$sroa$0$5$i = $idat$sroa$0$0180$i$lcssa224;$idat$sroa$9$4$i = $idat$sroa$9$0178$i$lcssa238;
     break;
    }
    case 23: {
     HEAP32[$1>>2] = 83; //@line 4491 "libraries/lodepng/lodepng.cpp"
     $idat$sroa$0$5$i = $idat$sroa$0$0180$i$lcssa233;$idat$sroa$9$4$i = $idat$sroa$9$0178$i$lcssa247;
     break;
    }
    case 33: {
     HEAP32[$14>>2] = 0; //@line 4131 "libraries/lodepng/lodepng.cpp"
     $$0$i28$i = 83;$idat$sroa$0$0180$i236 = $idat$sroa$0$0180$i$lcssa232;$idat$sroa$9$0178$i250 = $idat$sroa$9$0178$i$lcssa246;
     label = 39;
     break;
    }
    case 52: {
     HEAP32[$1>>2] = $$0$i33$i; //@line 4514 "libraries/lodepng/lodepng.cpp"
     $idat$sroa$0$5$i = $idat$sroa$0$0180$i$lcssa229;$idat$sroa$9$4$i = $idat$sroa$9$0178$i$lcssa243;
     break;
    }
    case 64: {
     HEAP32[$1>>2] = $$0$i38$i; //@line 4521 "libraries/lodepng/lodepng.cpp"
     $idat$sroa$0$5$i = $idat$sroa$0$0180$i$lcssa227;$idat$sroa$9$4$i = $idat$sroa$9$0178$i$lcssa241;
     break;
    }
    case 157: {
     HEAP32[$1>>2] = 73; //@line 4553 "libraries/lodepng/lodepng.cpp"
     $idat$sroa$0$5$i = $idat$sroa$0$0180$i$lcssa228;$idat$sroa$9$4$i = $idat$sroa$9$0178$i$lcssa242;
     break;
    }
    case 163: {
     HEAP32[$1>>2] = 74; //@line 4558 "libraries/lodepng/lodepng.cpp"
     $idat$sroa$0$5$i = $idat$sroa$0$0180$i$lcssa225;$idat$sroa$9$4$i = $idat$sroa$9$0178$i$lcssa239;
     break;
    }
    case 165: {
     HEAP32[$1>>2] = 69; //@line 4565 "libraries/lodepng/lodepng.cpp"
     $idat$sroa$0$5$i = $idat$sroa$0$0180$i$lcssa230;$idat$sroa$9$4$i = $idat$sroa$9$0178$i$lcssa244;
     break;
    }
    case 172: {
     HEAP32[$1>>2] = $$0$i83$i; //@line 4571 "libraries/lodepng/lodepng.cpp"
     $idat$sroa$0$5$i = $idat$sroa$0$0180$i$lcssa234;$idat$sroa$9$4$i = $idat$sroa$9$0178$i$lcssa248;
     break;
    }
    case 178: {
     HEAP32[$1>>2] = 57; //@line 4580 "libraries/lodepng/lodepng.cpp"
     $idat$sroa$0$5$i = $idat$sroa$0$3$i$lcssa;$idat$sroa$9$4$i = $idat$sroa$9$2$i$lcssa;
     break;
    }
   }
   if ((label|0) == 39) {
    HEAP32[$1>>2] = $$0$i28$i; //@line 4505 "libraries/lodepng/lodepng.cpp"
    $idat$sroa$0$5$i = $idat$sroa$0$0180$i236;$idat$sroa$9$4$i = $idat$sroa$9$0178$i250;
   }
   HEAP32[$scanlines$i>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
   $610 = ((($scanlines$i)) + 8|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   HEAP32[$610>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
   $611 = ((($scanlines$i)) + 4|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   HEAP32[$611>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
   $612 = ((($state)) + 140|0); //@line 4589 "libraries/lodepng/lodepng.cpp"
   $613 = HEAP32[$612>>2]|0; //@line 4589 "libraries/lodepng/lodepng.cpp"
   $614 = ($613|0)==(0); //@line 4589 "libraries/lodepng/lodepng.cpp"
   $615 = HEAP32[$w>>2]|0; //@line 4605 "libraries/lodepng/lodepng.cpp"
   if ($614) {
    $616 = HEAP32[$h>>2]|0; //@line 4592 "libraries/lodepng/lodepng.cpp"
    $$idx19$val$i = HEAP32[$15>>2]|0; //@line 4633 "libraries/lodepng/lodepng.cpp"
    $$idx20$i = ((($state)) + 148|0); //@line 4633 "libraries/lodepng/lodepng.cpp"
    $$idx20$val$i = HEAP32[$$idx20$i>>2]|0; //@line 4633 "libraries/lodepng/lodepng.cpp"
    switch ($$idx19$val$i|0) {
    case 3: case 0:  {
     $$0$i$i$i$i87$i = 1;
     break;
    }
    case 2:  {
     $$0$i$i$i$i87$i = 3;
     break;
    }
    case 4:  {
     $$0$i$i$i$i87$i = 2;
     break;
    }
    case 6:  {
     $$0$i$i$i$i87$i = 4;
     break;
    }
    default: {
     $$0$i$i$i$i87$i = 0;
    }
    }
    $617 = Math_imul($$idx20$val$i, $615)|0; //@line 2549 "libraries/lodepng/lodepng.cpp"
    $618 = Math_imul($617, $$0$i$i$i$i87$i)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $619 = (($618) + 7)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $620 = $619 >>> 3; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $621 = Math_imul($620, $616)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $622 = (($621) + ($616))|0; //@line 4592 "libraries/lodepng/lodepng.cpp"
    $predict$3$i = $622;
   } else {
    $623 = (($615) + 7)|0; //@line 4599 "libraries/lodepng/lodepng.cpp"
    $624 = $623 >>> 3; //@line 4599 "libraries/lodepng/lodepng.cpp"
    $625 = HEAP32[$h>>2]|0; //@line 4605 "libraries/lodepng/lodepng.cpp"
    $626 = (($625) + 7)|0; //@line 4599 "libraries/lodepng/lodepng.cpp"
    $627 = $626 >>> 3; //@line 4599 "libraries/lodepng/lodepng.cpp"
    $$idx17$val$i = HEAP32[$15>>2]|0; //@line 4633 "libraries/lodepng/lodepng.cpp"
    $$idx18$i = ((($state)) + 148|0); //@line 4633 "libraries/lodepng/lodepng.cpp"
    $$idx18$val$i = HEAP32[$$idx18$i>>2]|0; //@line 4633 "libraries/lodepng/lodepng.cpp"
    switch ($$idx17$val$i|0) {
    case 3: case 0:  {
     $$0$i$i$i$i89$i = 1;
     break;
    }
    case 2:  {
     $$0$i$i$i$i89$i = 3;
     break;
    }
    case 4:  {
     $$0$i$i$i$i89$i = 2;
     break;
    }
    case 6:  {
     $$0$i$i$i$i89$i = 4;
     break;
    }
    default: {
     $$0$i$i$i$i89$i = 0;
    }
    }
    $628 = Math_imul($$idx18$val$i, $624)|0; //@line 2549 "libraries/lodepng/lodepng.cpp"
    $629 = Math_imul($628, $$0$i$i$i$i89$i)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $630 = (($629) + 7)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $631 = $630 >>> 3; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $632 = Math_imul($631, $627)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $633 = (($632) + ($627))|0; //@line 4599 "libraries/lodepng/lodepng.cpp"
    $634 = ($615>>>0)>(4); //@line 4600 "libraries/lodepng/lodepng.cpp"
    $635 = (($615) + 3)|0; //@line 4600 "libraries/lodepng/lodepng.cpp"
    if ($634) {
     $636 = $635 >>> 3; //@line 4600 "libraries/lodepng/lodepng.cpp"
     switch ($$idx17$val$i|0) {
     case 3: case 0:  {
      $$0$i$i$i$i92$i = 1;
      break;
     }
     case 2:  {
      $$0$i$i$i$i92$i = 3;
      break;
     }
     case 4:  {
      $$0$i$i$i$i92$i = 2;
      break;
     }
     case 6:  {
      $$0$i$i$i$i92$i = 4;
      break;
     }
     default: {
      $$0$i$i$i$i92$i = 0;
     }
     }
     $637 = Math_imul($$idx18$val$i, $636)|0; //@line 2549 "libraries/lodepng/lodepng.cpp"
     $638 = Math_imul($637, $$0$i$i$i$i92$i)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
     $639 = (($638) + 7)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
     $640 = $639 >>> 3; //@line 2690 "libraries/lodepng/lodepng.cpp"
     $641 = Math_imul($640, $627)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
     $642 = (($633) + ($627))|0; //@line 4600 "libraries/lodepng/lodepng.cpp"
     $643 = (($642) + ($641))|0; //@line 4600 "libraries/lodepng/lodepng.cpp"
     $predict$0$i = $643;
    } else {
     $predict$0$i = $633;
    }
    $644 = $635 >>> 2; //@line 4601 "libraries/lodepng/lodepng.cpp"
    $645 = (($625) + 3)|0; //@line 4601 "libraries/lodepng/lodepng.cpp"
    $646 = $645 >>> 3; //@line 4601 "libraries/lodepng/lodepng.cpp"
    switch ($$idx17$val$i|0) {
    case 3: case 0:  {
     $$0$i$i$i$i95$i = 1;
     break;
    }
    case 2:  {
     $$0$i$i$i$i95$i = 3;
     break;
    }
    case 4:  {
     $$0$i$i$i$i95$i = 2;
     break;
    }
    case 6:  {
     $$0$i$i$i$i95$i = 4;
     break;
    }
    default: {
     $$0$i$i$i$i95$i = 0;
    }
    }
    $647 = Math_imul($$idx18$val$i, $644)|0; //@line 2549 "libraries/lodepng/lodepng.cpp"
    $648 = Math_imul($647, $$0$i$i$i$i95$i)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $649 = (($648) + 7)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $650 = $649 >>> 3; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $651 = Math_imul($650, $646)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $652 = (($predict$0$i) + ($646))|0; //@line 4601 "libraries/lodepng/lodepng.cpp"
    $653 = (($652) + ($651))|0; //@line 4601 "libraries/lodepng/lodepng.cpp"
    $654 = ($615>>>0)>(2); //@line 4602 "libraries/lodepng/lodepng.cpp"
    $655 = (($615) + 1)|0; //@line 4602 "libraries/lodepng/lodepng.cpp"
    if ($654) {
     $656 = $655 >>> 2; //@line 4602 "libraries/lodepng/lodepng.cpp"
     $657 = $645 >>> 2; //@line 4602 "libraries/lodepng/lodepng.cpp"
     switch ($$idx17$val$i|0) {
     case 3: case 0:  {
      $$0$i$i$i$i98$i = 1;
      break;
     }
     case 2:  {
      $$0$i$i$i$i98$i = 3;
      break;
     }
     case 4:  {
      $$0$i$i$i$i98$i = 2;
      break;
     }
     case 6:  {
      $$0$i$i$i$i98$i = 4;
      break;
     }
     default: {
      $$0$i$i$i$i98$i = 0;
     }
     }
     $658 = Math_imul($$idx18$val$i, $656)|0; //@line 2549 "libraries/lodepng/lodepng.cpp"
     $659 = Math_imul($658, $$0$i$i$i$i98$i)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
     $660 = (($659) + 7)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
     $661 = $660 >>> 3; //@line 2690 "libraries/lodepng/lodepng.cpp"
     $662 = Math_imul($661, $657)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
     $663 = (($653) + ($657))|0; //@line 4602 "libraries/lodepng/lodepng.cpp"
     $664 = (($663) + ($662))|0; //@line 4602 "libraries/lodepng/lodepng.cpp"
     $predict$1$i = $664;
    } else {
     $predict$1$i = $653;
    }
    $665 = $655 >>> 1; //@line 4603 "libraries/lodepng/lodepng.cpp"
    $666 = (($625) + 1)|0; //@line 4603 "libraries/lodepng/lodepng.cpp"
    $667 = $666 >>> 2; //@line 4603 "libraries/lodepng/lodepng.cpp"
    switch ($$idx17$val$i|0) {
    case 3: case 0:  {
     $$0$i$i$i$i101$i = 1;
     break;
    }
    case 2:  {
     $$0$i$i$i$i101$i = 3;
     break;
    }
    case 4:  {
     $$0$i$i$i$i101$i = 2;
     break;
    }
    case 6:  {
     $$0$i$i$i$i101$i = 4;
     break;
    }
    default: {
     $$0$i$i$i$i101$i = 0;
    }
    }
    $668 = Math_imul($$idx18$val$i, $665)|0; //@line 2549 "libraries/lodepng/lodepng.cpp"
    $669 = Math_imul($668, $$0$i$i$i$i101$i)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $670 = (($669) + 7)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $671 = $670 >>> 3; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $672 = Math_imul($671, $667)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $673 = (($predict$1$i) + ($667))|0; //@line 4603 "libraries/lodepng/lodepng.cpp"
    $674 = (($673) + ($672))|0; //@line 4603 "libraries/lodepng/lodepng.cpp"
    $675 = ($615>>>0)>(1); //@line 4604 "libraries/lodepng/lodepng.cpp"
    if ($675) {
     $676 = $615 >>> 1; //@line 4604 "libraries/lodepng/lodepng.cpp"
     $677 = $666 >>> 1; //@line 4604 "libraries/lodepng/lodepng.cpp"
     switch ($$idx17$val$i|0) {
     case 3: case 0:  {
      $$0$i$i$i$i104$i = 1;
      break;
     }
     case 2:  {
      $$0$i$i$i$i104$i = 3;
      break;
     }
     case 4:  {
      $$0$i$i$i$i104$i = 2;
      break;
     }
     case 6:  {
      $$0$i$i$i$i104$i = 4;
      break;
     }
     default: {
      $$0$i$i$i$i104$i = 0;
     }
     }
     $678 = Math_imul($$idx18$val$i, $676)|0; //@line 2549 "libraries/lodepng/lodepng.cpp"
     $679 = Math_imul($678, $$0$i$i$i$i104$i)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
     $680 = (($679) + 7)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
     $681 = $680 >>> 3; //@line 2690 "libraries/lodepng/lodepng.cpp"
     $682 = Math_imul($681, $677)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
     $683 = (($674) + ($677))|0; //@line 4604 "libraries/lodepng/lodepng.cpp"
     $684 = (($683) + ($682))|0; //@line 4604 "libraries/lodepng/lodepng.cpp"
     $predict$2$i = $684;
    } else {
     $predict$2$i = $674;
    }
    $685 = $625 >>> 1; //@line 4605 "libraries/lodepng/lodepng.cpp"
    switch ($$idx17$val$i|0) {
    case 3: case 0:  {
     $$0$i$i$i$i107$i = 1;
     break;
    }
    case 2:  {
     $$0$i$i$i$i107$i = 3;
     break;
    }
    case 4:  {
     $$0$i$i$i$i107$i = 2;
     break;
    }
    case 6:  {
     $$0$i$i$i$i107$i = 4;
     break;
    }
    default: {
     $$0$i$i$i$i107$i = 0;
    }
    }
    $686 = Math_imul($$idx18$val$i, $615)|0; //@line 2549 "libraries/lodepng/lodepng.cpp"
    $687 = Math_imul($686, $$0$i$i$i$i107$i)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $688 = (($687) + 7)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $689 = $688 >>> 3; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $690 = Math_imul($689, $685)|0; //@line 2690 "libraries/lodepng/lodepng.cpp"
    $691 = (($predict$2$i) + ($685))|0; //@line 4605 "libraries/lodepng/lodepng.cpp"
    $692 = (($691) + ($690))|0; //@line 4605 "libraries/lodepng/lodepng.cpp"
    $predict$3$i = $692;
   }
   $693 = HEAP32[$1>>2]|0; //@line 4607 "libraries/lodepng/lodepng.cpp"
   $694 = ($693|0)==(0); //@line 4607 "libraries/lodepng/lodepng.cpp"
   L291: do {
    if ($694) {
     $695 = ($predict$3$i|0)==(0); //@line 225 "libraries/lodepng/lodepng.cpp"
     do {
      if (!($695)) {
       $696 = (_realloc(0,$predict$3$i)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       $697 = ($696|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
       if ($697) {
        HEAP32[$1>>2] = 83; //@line 4607 "libraries/lodepng/lodepng.cpp"
        break L291;
       } else {
        HEAP32[$610>>2] = $predict$3$i; //@line 231 "libraries/lodepng/lodepng.cpp"
        HEAP32[$scanlines$i>>2] = $696; //@line 232 "libraries/lodepng/lodepng.cpp"
        $$pr$pre$i = HEAP32[$1>>2]|0; //@line 4608 "libraries/lodepng/lodepng.cpp"
        $698 = ($$pr$pre$i|0)==(0); //@line 4608 "libraries/lodepng/lodepng.cpp"
        if ($698) {
         break;
        } else {
         break L291;
        }
       }
      }
     } while(0);
     $699 = HEAP32[$29>>2]|0; //@line 2150 "libraries/lodepng/lodepng.cpp"
     $700 = ($699|0)==(0|0); //@line 2150 "libraries/lodepng/lodepng.cpp"
     if ($700) {
      $702 = (__Z23lodepng_zlib_decompressPPhPjPKhjPK25LodePNGDecompressSettings($scanlines$i,$611,$idat$sroa$0$5$i,$idat$sroa$9$4$i,$26)|0); //@line 2156 "libraries/lodepng/lodepng.cpp"
      $$0$i111$i = $702;
     } else {
      $701 = (FUNCTION_TABLE_iiiiii[$699 & 8191]($scanlines$i,$611,$idat$sroa$0$5$i,$idat$sroa$9$4$i,$26)|0); //@line 2152 "libraries/lodepng/lodepng.cpp"
      $$0$i111$i = $701;
     }
     $703 = ($$0$i111$i|0)!=(0); //@line 4612 "libraries/lodepng/lodepng.cpp"
     $704 = HEAP32[$611>>2]|0; //@line 4612 "libraries/lodepng/lodepng.cpp"
     $705 = ($704|0)==($predict$3$i|0); //@line 4612 "libraries/lodepng/lodepng.cpp"
     $or$cond5$i = $703 | $705; //@line 4612 "libraries/lodepng/lodepng.cpp"
     $$$i = $or$cond5$i ? $$0$i111$i : 91; //@line 4612 "libraries/lodepng/lodepng.cpp"
     HEAP32[$1>>2] = $$$i; //@line 4612 "libraries/lodepng/lodepng.cpp"
    }
   } while(0);
   _free($idat$sroa$0$5$i); //@line 76 "libraries/lodepng/lodepng.cpp"
   $706 = HEAP32[$1>>2]|0; //@line 4616 "libraries/lodepng/lodepng.cpp"
   $707 = ($706|0)==(0); //@line 4616 "libraries/lodepng/lodepng.cpp"
   if ($707) {
    $708 = HEAP32[$w>>2]|0; //@line 4618 "libraries/lodepng/lodepng.cpp"
    $709 = HEAP32[$h>>2]|0; //@line 4618 "libraries/lodepng/lodepng.cpp"
    $710 = HEAP32[$15>>2]|0; //@line 2634 "libraries/lodepng/lodepng.cpp"
    $711 = ((($state)) + 148|0); //@line 2634 "libraries/lodepng/lodepng.cpp"
    $712 = HEAP32[$711>>2]|0; //@line 2634 "libraries/lodepng/lodepng.cpp"
    switch ($710|0) {
    case 3: case 0:  {
     $$0$i$i$i$i112$i = 1;
     break;
    }
    case 2:  {
     $$0$i$i$i$i112$i = 3;
     break;
    }
    case 4:  {
     $$0$i$i$i$i112$i = 2;
     break;
    }
    case 6:  {
     $$0$i$i$i$i112$i = 4;
     break;
    }
    default: {
     $$0$i$i$i$i112$i = 0;
    }
    }
    $713 = Math_imul($709, $708)|0; //@line 2549 "libraries/lodepng/lodepng.cpp"
    $714 = Math_imul($713, $712)|0; //@line 2676 "libraries/lodepng/lodepng.cpp"
    $715 = Math_imul($714, $$0$i$i$i$i112$i)|0; //@line 2676 "libraries/lodepng/lodepng.cpp"
    $716 = (($715) + 7)|0; //@line 2676 "libraries/lodepng/lodepng.cpp"
    $717 = $716 >>> 3; //@line 2676 "libraries/lodepng/lodepng.cpp"
    $718 = ($717|0)==(0); //@line 225 "libraries/lodepng/lodepng.cpp"
    do {
     if ($718) {
      $725 = $710;$727 = $712;$733 = $708;$737 = $709;$outv$sroa$0$1$ph247$i = 0;
      label = 250;
     } else {
      $719 = (_realloc(0,$717)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $720 = ($719|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
      if ($720) {
       HEAP32[$1>>2] = 83; //@line 4621 "libraries/lodepng/lodepng.cpp"
       $outv$sroa$0$1132135$i = 0;
       break;
      }
      _memset(($719|0),0,($717|0))|0; //@line 268 "libraries/lodepng/lodepng.cpp"
      $721 = $719; //@line 232 "libraries/lodepng/lodepng.cpp"
      $$pr133$pre$i = HEAP32[$1>>2]|0; //@line 4622 "libraries/lodepng/lodepng.cpp"
      $722 = ($$pr133$pre$i|0)==(0); //@line 4622 "libraries/lodepng/lodepng.cpp"
      if ($722) {
       $$pre = HEAP32[$w>>2]|0; //@line 4622 "libraries/lodepng/lodepng.cpp"
       $$pre90 = HEAP32[$h>>2]|0; //@line 4622 "libraries/lodepng/lodepng.cpp"
       $$pre91 = HEAP32[$15>>2]|0; //@line 2634 "libraries/lodepng/lodepng.cpp"
       $$pre92 = HEAP32[$711>>2]|0; //@line 2634 "libraries/lodepng/lodepng.cpp"
       $725 = $$pre91;$727 = $$pre92;$733 = $$pre;$737 = $$pre90;$outv$sroa$0$1$ph247$i = $721;
       label = 250;
      } else {
       $outv$sroa$0$1132135$i = $721;
      }
     }
    } while(0);
    if ((label|0) == 250) {
     $723 = $outv$sroa$0$1$ph247$i; //@line 4622 "libraries/lodepng/lodepng.cpp"
     $724 = HEAP32[$scanlines$i>>2]|0; //@line 4622 "libraries/lodepng/lodepng.cpp"
     switch ($725|0) {
     case 3: case 0:  {
      $$0$i$i$i$i$i = 1;
      break;
     }
     case 2:  {
      $$0$i$i$i$i$i = 3;
      break;
     }
     case 4:  {
      $$0$i$i$i$i$i = 2;
      break;
     }
     case 6:  {
      $$0$i$i$i$i$i = 4;
      break;
     }
     default: {
      $$0$i$i$i$i$i = 0;
     }
     }
     $726 = Math_imul($$0$i$i$i$i$i, $727)|0; //@line 2549 "libraries/lodepng/lodepng.cpp"
     $728 = ($726|0)==(0); //@line 4084 "libraries/lodepng/lodepng.cpp"
     L326: do {
      if ($728) {
       $$5$i$i = 31;
      } else {
       $729 = HEAP32[$612>>2]|0; //@line 4086 "libraries/lodepng/lodepng.cpp"
       $730 = ($729|0)==(0); //@line 4086 "libraries/lodepng/lodepng.cpp"
       L328: do {
        if ($730) {
         $731 = ($726>>>0)<(8); //@line 4088 "libraries/lodepng/lodepng.cpp"
         if ($731) {
          $732 = Math_imul($726, $733)|0; //@line 4088 "libraries/lodepng/lodepng.cpp"
          $734 = (($732) + 7)|0; //@line 4088 "libraries/lodepng/lodepng.cpp"
          $735 = $734 & -8; //@line 4088 "libraries/lodepng/lodepng.cpp"
          $736 = ($732|0)==($735|0); //@line 4088 "libraries/lodepng/lodepng.cpp"
          if (!($736)) {
           $738 = (__ZL8unfilterPhPKhjjj($724,$724,$733,$737,$726)|0); //@line 4090 "libraries/lodepng/lodepng.cpp"
           $739 = ($738|0)==(0); //@line 4090 "libraries/lodepng/lodepng.cpp"
           if (!($739)) {
            $$5$i$i = $738;
            break L326;
           }
           $740 = ($737|0)==(0); //@line 4058 "libraries/lodepng/lodepng.cpp"
           $741 = ($732|0)==(0); //@line 4061 "libraries/lodepng/lodepng.cpp"
           $or$cond$i10$i$i = $740 | $741; //@line 4058 "libraries/lodepng/lodepng.cpp"
           if ($or$cond$i10$i$i) {
            break;
           } else {
            $ibp$05$us$i$i$i = 0;$obp$06$us$i$i$i = 0;$y$07$us$i$i$i = 0;
           }
           while(1) {
            $ibp$11$us$i$i$i = $ibp$05$us$i$i$i;$obp$12$us$i$i$i = $obp$06$us$i$i$i;$x$03$us$i$i$i = 0;
            while(1) {
             $742 = $ibp$11$us$i$i$i >>> 3; //@line 2351 "libraries/lodepng/lodepng.cpp"
             $743 = (($724) + ($742)|0); //@line 2351 "libraries/lodepng/lodepng.cpp"
             $744 = HEAP8[$743>>0]|0; //@line 2351 "libraries/lodepng/lodepng.cpp"
             $745 = $744&255; //@line 2351 "libraries/lodepng/lodepng.cpp"
             $746 = $ibp$11$us$i$i$i & 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
             $747 = $746 ^ 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
             $748 = (($ibp$11$us$i$i$i) + 1)|0; //@line 2352 "libraries/lodepng/lodepng.cpp"
             $749 = 1 << $747; //@line 2383 "libraries/lodepng/lodepng.cpp"
             $750 = $749 & $745; //@line 2383 "libraries/lodepng/lodepng.cpp"
             $751 = ($750|0)==(0); //@line 2383 "libraries/lodepng/lodepng.cpp"
             $752 = $obp$12$us$i$i$i & 7; //@line 2383 "libraries/lodepng/lodepng.cpp"
             $753 = $752 ^ 7; //@line 2383 "libraries/lodepng/lodepng.cpp"
             $754 = 1 << $753; //@line 2383 "libraries/lodepng/lodepng.cpp"
             if ($751) {
              $761 = $754 ^ 255; //@line 2383 "libraries/lodepng/lodepng.cpp"
              $762 = $obp$12$us$i$i$i >>> 3; //@line 2383 "libraries/lodepng/lodepng.cpp"
              $763 = (($723) + ($762)|0); //@line 2383 "libraries/lodepng/lodepng.cpp"
              $764 = HEAP8[$763>>0]|0; //@line 2383 "libraries/lodepng/lodepng.cpp"
              $765 = $764&255; //@line 2383 "libraries/lodepng/lodepng.cpp"
              $766 = $765 & $761; //@line 2383 "libraries/lodepng/lodepng.cpp"
              $767 = $766&255; //@line 2383 "libraries/lodepng/lodepng.cpp"
              HEAP8[$763>>0] = $767; //@line 2383 "libraries/lodepng/lodepng.cpp"
             } else {
              $755 = $obp$12$us$i$i$i >>> 3; //@line 2384 "libraries/lodepng/lodepng.cpp"
              $756 = (($723) + ($755)|0); //@line 2384 "libraries/lodepng/lodepng.cpp"
              $757 = HEAP8[$756>>0]|0; //@line 2384 "libraries/lodepng/lodepng.cpp"
              $758 = $757&255; //@line 2384 "libraries/lodepng/lodepng.cpp"
              $759 = $758 | $754; //@line 2384 "libraries/lodepng/lodepng.cpp"
              $760 = $759&255; //@line 2384 "libraries/lodepng/lodepng.cpp"
              HEAP8[$756>>0] = $760; //@line 2384 "libraries/lodepng/lodepng.cpp"
             }
             $768 = (($obp$12$us$i$i$i) + 1)|0; //@line 2385 "libraries/lodepng/lodepng.cpp"
             $769 = (($x$03$us$i$i$i) + 1)|0; //@line 4061 "libraries/lodepng/lodepng.cpp"
             $exitcond11$i$i$i = ($769|0)==($732|0); //@line 4061 "libraries/lodepng/lodepng.cpp"
             if ($exitcond11$i$i$i) {
              break;
             } else {
              $ibp$11$us$i$i$i = $748;$obp$12$us$i$i$i = $768;$x$03$us$i$i$i = $769;
             }
            }
            $770 = (($obp$06$us$i$i$i) + ($732))|0; //@line 4061 "libraries/lodepng/lodepng.cpp"
            $771 = (($ibp$05$us$i$i$i) + ($735))|0; //@line 4066 "libraries/lodepng/lodepng.cpp"
            $772 = (($y$07$us$i$i$i) + 1)|0; //@line 4058 "libraries/lodepng/lodepng.cpp"
            $exitcond12$i$i$i = ($772|0)==($737|0); //@line 4058 "libraries/lodepng/lodepng.cpp"
            if ($exitcond12$i$i$i) {
             break L328;
            } else {
             $ibp$05$us$i$i$i = $771;$obp$06$us$i$i$i = $770;$y$07$us$i$i$i = $772;
            }
           }
          }
         }
         $773 = (__ZL8unfilterPhPKhjjj($723,$724,$733,$737,$726)|0); //@line 4094 "libraries/lodepng/lodepng.cpp"
         $774 = ($773|0)==(0); //@line 4094 "libraries/lodepng/lodepng.cpp"
         if (!($774)) {
          $$5$i$i = $773;
          break L326;
         }
        } else {
         __ZL19Adam7_getpassvaluesPjS_S_S_S_jjj($passw$i$i,$passh$i$i,$filter_passstart$i$i,$padded_passstart$i$i,$passstart$i$i,$733,$737,$726); //@line 4101 "libraries/lodepng/lodepng.cpp"
         $775 = ($726>>>0)<(8); //@line 4108 "libraries/lodepng/lodepng.cpp"
         L345: do {
          if ($775) {
           $$227$us$i$i = 0;$i$026$us$i$i = 0;
           while(1) {
            $784 = (($padded_passstart$i$i) + ($i$026$us$i$i<<2)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
            $785 = HEAP32[$784>>2]|0; //@line 4112 "libraries/lodepng/lodepng.cpp"
            $786 = (($724) + ($785)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
            $787 = (($filter_passstart$i$i) + ($i$026$us$i$i<<2)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
            $788 = HEAP32[$787>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
            $789 = (($724) + ($788)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
            $790 = (($passw$i$i) + ($i$026$us$i$i<<2)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
            $791 = HEAP32[$790>>2]|0; //@line 4112 "libraries/lodepng/lodepng.cpp"
            $792 = (($passh$i$i) + ($i$026$us$i$i<<2)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
            $793 = HEAP32[$792>>2]|0; //@line 4113 "libraries/lodepng/lodepng.cpp"
            $794 = (__ZL8unfilterPhPKhjjj($786,$789,$791,$793,$726)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
            $795 = ($794|0)==(0); //@line 4105 "libraries/lodepng/lodepng.cpp"
            $$2$$us$i$i = $795 ? $$227$us$i$i : $794; //@line 4105 "libraries/lodepng/lodepng.cpp"
            if (!($795)) {
             $$37$i$i = 0;$$4$i$i = $794;
             break L345;
            }
            $796 = (($passstart$i$i) + ($i$026$us$i$i<<2)|0); //@line 4112 "libraries/lodepng/lodepng.cpp"
            $797 = HEAP32[$796>>2]|0; //@line 4112 "libraries/lodepng/lodepng.cpp"
            $798 = (($724) + ($797)|0); //@line 4112 "libraries/lodepng/lodepng.cpp"
            $799 = Math_imul($791, $726)|0; //@line 4112 "libraries/lodepng/lodepng.cpp"
            $800 = (($799) + 7)|0; //@line 4113 "libraries/lodepng/lodepng.cpp"
            $801 = $800 & -8; //@line 4113 "libraries/lodepng/lodepng.cpp"
            $802 = ($793|0)==(0); //@line 4058 "libraries/lodepng/lodepng.cpp"
            $803 = ($799|0)==(0); //@line 4061 "libraries/lodepng/lodepng.cpp"
            $or$cond$i11$us$i$i = $802 | $803; //@line 4058 "libraries/lodepng/lodepng.cpp"
            if (!($or$cond$i11$us$i$i)) {
             $ibp$05$us$i19$us$i$i = 0;$obp$06$us$i18$us$i$i = 0;$y$07$us$i17$us$i$i = 0;
             while(1) {
              $ibp$11$us$i14$us$i$i = $ibp$05$us$i19$us$i$i;$obp$12$us$i13$us$i$i = $obp$06$us$i18$us$i$i;$x$03$us$i12$us$i$i = 0;
              while(1) {
               $804 = $ibp$11$us$i14$us$i$i >>> 3; //@line 2351 "libraries/lodepng/lodepng.cpp"
               $805 = (($786) + ($804)|0); //@line 2351 "libraries/lodepng/lodepng.cpp"
               $806 = HEAP8[$805>>0]|0; //@line 2351 "libraries/lodepng/lodepng.cpp"
               $807 = $806&255; //@line 2351 "libraries/lodepng/lodepng.cpp"
               $808 = $ibp$11$us$i14$us$i$i & 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
               $809 = $808 ^ 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
               $810 = (($ibp$11$us$i14$us$i$i) + 1)|0; //@line 2352 "libraries/lodepng/lodepng.cpp"
               $811 = 1 << $809; //@line 2383 "libraries/lodepng/lodepng.cpp"
               $812 = $811 & $807; //@line 2383 "libraries/lodepng/lodepng.cpp"
               $813 = ($812|0)==(0); //@line 2383 "libraries/lodepng/lodepng.cpp"
               $814 = $obp$12$us$i13$us$i$i & 7; //@line 2383 "libraries/lodepng/lodepng.cpp"
               $815 = $814 ^ 7; //@line 2383 "libraries/lodepng/lodepng.cpp"
               $816 = 1 << $815; //@line 2383 "libraries/lodepng/lodepng.cpp"
               if ($813) {
                $823 = $816 ^ 255; //@line 2383 "libraries/lodepng/lodepng.cpp"
                $824 = $obp$12$us$i13$us$i$i >>> 3; //@line 2383 "libraries/lodepng/lodepng.cpp"
                $825 = (($798) + ($824)|0); //@line 2383 "libraries/lodepng/lodepng.cpp"
                $826 = HEAP8[$825>>0]|0; //@line 2383 "libraries/lodepng/lodepng.cpp"
                $827 = $826&255; //@line 2383 "libraries/lodepng/lodepng.cpp"
                $828 = $827 & $823; //@line 2383 "libraries/lodepng/lodepng.cpp"
                $829 = $828&255; //@line 2383 "libraries/lodepng/lodepng.cpp"
                HEAP8[$825>>0] = $829; //@line 2383 "libraries/lodepng/lodepng.cpp"
               } else {
                $817 = $obp$12$us$i13$us$i$i >>> 3; //@line 2384 "libraries/lodepng/lodepng.cpp"
                $818 = (($798) + ($817)|0); //@line 2384 "libraries/lodepng/lodepng.cpp"
                $819 = HEAP8[$818>>0]|0; //@line 2384 "libraries/lodepng/lodepng.cpp"
                $820 = $819&255; //@line 2384 "libraries/lodepng/lodepng.cpp"
                $821 = $820 | $816; //@line 2384 "libraries/lodepng/lodepng.cpp"
                $822 = $821&255; //@line 2384 "libraries/lodepng/lodepng.cpp"
                HEAP8[$818>>0] = $822; //@line 2384 "libraries/lodepng/lodepng.cpp"
               }
               $830 = (($obp$12$us$i13$us$i$i) + 1)|0; //@line 2385 "libraries/lodepng/lodepng.cpp"
               $831 = (($x$03$us$i12$us$i$i) + 1)|0; //@line 4061 "libraries/lodepng/lodepng.cpp"
               $exitcond11$i15$us$i$i = ($831|0)==($799|0); //@line 4061 "libraries/lodepng/lodepng.cpp"
               if ($exitcond11$i15$us$i$i) {
                break;
               } else {
                $ibp$11$us$i14$us$i$i = $810;$obp$12$us$i13$us$i$i = $830;$x$03$us$i12$us$i$i = $831;
               }
              }
              $832 = (($obp$06$us$i18$us$i$i) + ($799))|0; //@line 4061 "libraries/lodepng/lodepng.cpp"
              $833 = (($ibp$05$us$i19$us$i$i) + ($801))|0; //@line 4066 "libraries/lodepng/lodepng.cpp"
              $834 = (($y$07$us$i17$us$i$i) + 1)|0; //@line 4058 "libraries/lodepng/lodepng.cpp"
              $exitcond12$i21$us$i$i = ($834|0)==($793|0); //@line 4058 "libraries/lodepng/lodepng.cpp"
              if ($exitcond12$i21$us$i$i) {
               break;
              } else {
               $ibp$05$us$i19$us$i$i = $833;$obp$06$us$i18$us$i$i = $832;$y$07$us$i17$us$i$i = $834;
              }
             }
            }
            $835 = (($i$026$us$i$i) + 1)|0; //@line 4103 "libraries/lodepng/lodepng.cpp"
            $836 = ($835|0)==(7); //@line 4103 "libraries/lodepng/lodepng.cpp"
            if ($836) {
             $$2$lcssa$i$i = $$2$$us$i$i;
             label = 280;
             break;
            } else {
             $$227$us$i$i = $$2$$us$i$i;$i$026$us$i$i = $835;
            }
           }
          } else {
           $776 = HEAP32[$padded_passstart$i$i>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
           $777 = (($724) + ($776)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
           $778 = HEAP32[$filter_passstart$i$i>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
           $779 = (($724) + ($778)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
           $780 = HEAP32[$passw$i$i>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
           $781 = HEAP32[$passh$i$i>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
           $782 = (__ZL8unfilterPhPKhjjj($777,$779,$780,$781,$726)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
           $783 = ($782|0)==(0); //@line 4105 "libraries/lodepng/lodepng.cpp"
           if ($783) {
            $837 = ((($padded_passstart$i$i)) + 4|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
            $838 = HEAP32[$837>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
            $839 = (($724) + ($838)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
            $840 = ((($filter_passstart$i$i)) + 4|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
            $841 = HEAP32[$840>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
            $842 = (($724) + ($841)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
            $843 = ((($passw$i$i)) + 4|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
            $844 = HEAP32[$843>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
            $845 = ((($passh$i$i)) + 4|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
            $846 = HEAP32[$845>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
            $847 = (__ZL8unfilterPhPKhjjj($839,$842,$844,$846,$726)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
            $848 = ($847|0)==(0); //@line 4105 "libraries/lodepng/lodepng.cpp"
            if ($848) {
             $928 = ((($padded_passstart$i$i)) + 8|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
             $929 = HEAP32[$928>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
             $930 = (($724) + ($929)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
             $931 = ((($filter_passstart$i$i)) + 8|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
             $932 = HEAP32[$931>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
             $933 = (($724) + ($932)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
             $934 = ((($passw$i$i)) + 8|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
             $935 = HEAP32[$934>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
             $936 = ((($passh$i$i)) + 8|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
             $937 = HEAP32[$936>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
             $938 = (__ZL8unfilterPhPKhjjj($930,$933,$935,$937,$726)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
             $939 = ($938|0)==(0); //@line 4105 "libraries/lodepng/lodepng.cpp"
             if ($939) {
              $940 = ((($padded_passstart$i$i)) + 12|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
              $941 = HEAP32[$940>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
              $942 = (($724) + ($941)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
              $943 = ((($filter_passstart$i$i)) + 12|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
              $944 = HEAP32[$943>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
              $945 = (($724) + ($944)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
              $946 = ((($passw$i$i)) + 12|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
              $947 = HEAP32[$946>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
              $948 = ((($passh$i$i)) + 12|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
              $949 = HEAP32[$948>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
              $950 = (__ZL8unfilterPhPKhjjj($942,$945,$947,$949,$726)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
              $951 = ($950|0)==(0); //@line 4105 "libraries/lodepng/lodepng.cpp"
              if ($951) {
               $952 = ((($padded_passstart$i$i)) + 16|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $953 = HEAP32[$952>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
               $954 = (($724) + ($953)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $955 = ((($filter_passstart$i$i)) + 16|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $956 = HEAP32[$955>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
               $957 = (($724) + ($956)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $958 = ((($passw$i$i)) + 16|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $959 = HEAP32[$958>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
               $960 = ((($passh$i$i)) + 16|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $961 = HEAP32[$960>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
               $962 = (__ZL8unfilterPhPKhjjj($954,$957,$959,$961,$726)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $963 = ($962|0)==(0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               if (!($963)) {
                $$37$i$i = 0;$$4$i$i = $962;
                break;
               }
               $964 = ((($padded_passstart$i$i)) + 20|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $965 = HEAP32[$964>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
               $966 = (($724) + ($965)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $967 = ((($filter_passstart$i$i)) + 20|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $968 = HEAP32[$967>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
               $969 = (($724) + ($968)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $970 = ((($passw$i$i)) + 20|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $971 = HEAP32[$970>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
               $972 = ((($passh$i$i)) + 20|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $973 = HEAP32[$972>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
               $974 = (__ZL8unfilterPhPKhjjj($966,$969,$971,$973,$726)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $975 = ($974|0)==(0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               if (!($975)) {
                $$37$i$i = 0;$$4$i$i = $974;
                break;
               }
               $976 = ((($padded_passstart$i$i)) + 24|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $977 = HEAP32[$976>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
               $978 = (($724) + ($977)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $979 = ((($filter_passstart$i$i)) + 24|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $980 = HEAP32[$979>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
               $981 = (($724) + ($980)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $982 = ((($passw$i$i)) + 24|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $983 = HEAP32[$982>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
               $984 = ((($passh$i$i)) + 24|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $985 = HEAP32[$984>>2]|0; //@line 4105 "libraries/lodepng/lodepng.cpp"
               $986 = (__ZL8unfilterPhPKhjjj($978,$981,$983,$985,$726)|0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               $987 = ($986|0)==(0); //@line 4105 "libraries/lodepng/lodepng.cpp"
               if ($987) {
                $$2$lcssa$i$i = 0;
                label = 280;
               } else {
                $$37$i$i = 0;$$4$i$i = $986;
               }
              } else {
               $$37$i$i = 0;$$4$i$i = $950;
              }
             } else {
              $$37$i$i = 0;$$4$i$i = $938;
             }
            } else {
             $$37$i$i = 0;$$4$i$i = $847;
            }
           } else {
            $$37$i$i = 0;$$4$i$i = $782;
           }
          }
         } while(0);
         if ((label|0) == 280) {
          __ZL19Adam7_getpassvaluesPjS_S_S_S_jjj($passw$i$i$i,$passh$i$i$i,$filter_passstart$i$i$i,$padded_passstart$i$i$i,$passstart$i$i$i,$733,$737,$726); //@line 3999 "libraries/lodepng/lodepng.cpp"
          $849 = ($726>>>0)>(7); //@line 4001 "libraries/lodepng/lodepng.cpp"
          if ($849) {
           $850 = $726 >>> 3; //@line 4006 "libraries/lodepng/lodepng.cpp"
           $851 = ($850|0)==(0); //@line 4012 "libraries/lodepng/lodepng.cpp"
           if (!($851)) {
            $i$015$i$i$i = 0;
            while(1) {
             $852 = (($passh$i$i$i) + ($i$015$i$i$i<<2)|0); //@line 4007 "libraries/lodepng/lodepng.cpp"
             $853 = HEAP32[$852>>2]|0; //@line 4007 "libraries/lodepng/lodepng.cpp"
             $854 = ($853|0)==(0); //@line 4007 "libraries/lodepng/lodepng.cpp"
             do {
              if (!($854)) {
               $855 = (($passw$i$i$i) + ($i$015$i$i$i<<2)|0); //@line 4008 "libraries/lodepng/lodepng.cpp"
               $856 = HEAP32[$855>>2]|0; //@line 4008 "libraries/lodepng/lodepng.cpp"
               $857 = ($856|0)==(0); //@line 4008 "libraries/lodepng/lodepng.cpp"
               if ($857) {
                break;
               }
               $858 = (34884 + ($i$015$i$i$i<<2)|0); //@line 4011 "libraries/lodepng/lodepng.cpp"
               $859 = (34912 + ($i$015$i$i$i<<2)|0); //@line 4011 "libraries/lodepng/lodepng.cpp"
               $860 = (34940 + ($i$015$i$i$i<<2)|0); //@line 4011 "libraries/lodepng/lodepng.cpp"
               $861 = (34968 + ($i$015$i$i$i<<2)|0); //@line 4011 "libraries/lodepng/lodepng.cpp"
               $862 = (($passstart$i$i$i) + ($i$015$i$i$i<<2)|0); //@line 4010 "libraries/lodepng/lodepng.cpp"
               $$pre58$i$i$i = HEAP32[$862>>2]|0; //@line 4010 "libraries/lodepng/lodepng.cpp"
               $$pre59$i$i$i = HEAP32[$861>>2]|0; //@line 4011 "libraries/lodepng/lodepng.cpp"
               $$pre60$i$i$i = HEAP32[$860>>2]|0; //@line 4011 "libraries/lodepng/lodepng.cpp"
               $$pre61$i$i$i = HEAP32[$859>>2]|0; //@line 4011 "libraries/lodepng/lodepng.cpp"
               $$pre62$i$i$i = HEAP32[$858>>2]|0; //@line 4011 "libraries/lodepng/lodepng.cpp"
               $y$012$us$i$i$i = 0;
               while(1) {
                $863 = Math_imul($y$012$us$i$i$i, $$pre60$i$i$i)|0; //@line 4011 "libraries/lodepng/lodepng.cpp"
                $864 = (($863) + ($$pre59$i$i$i))|0; //@line 4011 "libraries/lodepng/lodepng.cpp"
                $865 = Math_imul($864, $733)|0; //@line 4011 "libraries/lodepng/lodepng.cpp"
                $866 = (($865) + ($$pre61$i$i$i))|0; //@line 4011 "libraries/lodepng/lodepng.cpp"
                $867 = Math_imul($y$012$us$i$i$i, $856)|0; //@line 4010 "libraries/lodepng/lodepng.cpp"
                $x$09$us$us$i$i$i = 0;
                while(1) {
                 $869 = (($x$09$us$us$i$i$i) + ($867))|0; //@line 4010 "libraries/lodepng/lodepng.cpp"
                 $870 = Math_imul($869, $850)|0; //@line 4010 "libraries/lodepng/lodepng.cpp"
                 $871 = (($870) + ($$pre58$i$i$i))|0; //@line 4010 "libraries/lodepng/lodepng.cpp"
                 $872 = Math_imul($x$09$us$us$i$i$i, $$pre62$i$i$i)|0; //@line 4011 "libraries/lodepng/lodepng.cpp"
                 $873 = (($866) + ($872))|0; //@line 4011 "libraries/lodepng/lodepng.cpp"
                 $874 = Math_imul($873, $850)|0; //@line 4011 "libraries/lodepng/lodepng.cpp"
                 $b$08$us$us$i$i$i = 0;
                 while(1) {
                  $876 = (($871) + ($b$08$us$us$i$i$i))|0; //@line 4014 "libraries/lodepng/lodepng.cpp"
                  $877 = (($724) + ($876)|0); //@line 4014 "libraries/lodepng/lodepng.cpp"
                  $878 = HEAP8[$877>>0]|0; //@line 4014 "libraries/lodepng/lodepng.cpp"
                  $879 = (($b$08$us$us$i$i$i) + ($874))|0; //@line 4014 "libraries/lodepng/lodepng.cpp"
                  $880 = (($723) + ($879)|0); //@line 4014 "libraries/lodepng/lodepng.cpp"
                  HEAP8[$880>>0] = $878; //@line 4014 "libraries/lodepng/lodepng.cpp"
                  $881 = (($b$08$us$us$i$i$i) + 1)|0; //@line 4012 "libraries/lodepng/lodepng.cpp"
                  $exitcond$i$i$i = ($881|0)==($850|0); //@line 4012 "libraries/lodepng/lodepng.cpp"
                  if ($exitcond$i$i$i) {
                   break;
                  } else {
                   $b$08$us$us$i$i$i = $881;
                  }
                 }
                 $875 = (($x$09$us$us$i$i$i) + 1)|0; //@line 4008 "libraries/lodepng/lodepng.cpp"
                 $exitcond49$i$i = ($875|0)==($856|0); //@line 4008 "libraries/lodepng/lodepng.cpp"
                 if ($exitcond49$i$i) {
                  break;
                 } else {
                  $x$09$us$us$i$i$i = $875;
                 }
                }
                $868 = (($y$012$us$i$i$i) + 1)|0; //@line 4007 "libraries/lodepng/lodepng.cpp"
                $exitcond50$i$i = ($868|0)==($853|0); //@line 4007 "libraries/lodepng/lodepng.cpp"
                if ($exitcond50$i$i) {
                 break;
                } else {
                 $y$012$us$i$i$i = $868;
                }
               }
              }
             } while(0);
             $882 = (($i$015$i$i$i) + 1)|0; //@line 4003 "libraries/lodepng/lodepng.cpp"
             $883 = ($882|0)==(7); //@line 4003 "libraries/lodepng/lodepng.cpp"
             if ($883) {
              break;
             } else {
              $i$015$i$i$i = $882;
             }
            }
           }
          } else {
           $i$128$i$i$i = 0;
           while(1) {
            $884 = (($passw$i$i$i) + ($i$128$i$i$i<<2)|0); //@line 4024 "libraries/lodepng/lodepng.cpp"
            $885 = HEAP32[$884>>2]|0; //@line 4028 "libraries/lodepng/lodepng.cpp"
            $886 = (($passh$i$i$i) + ($i$128$i$i$i<<2)|0); //@line 4027 "libraries/lodepng/lodepng.cpp"
            $887 = HEAP32[$886>>2]|0; //@line 4027 "libraries/lodepng/lodepng.cpp"
            $888 = ($887|0)==(0); //@line 4027 "libraries/lodepng/lodepng.cpp"
            $889 = ($885|0)==(0); //@line 4028 "libraries/lodepng/lodepng.cpp"
            $or$cond$i$i$i = $889 | $888; //@line 4027 "libraries/lodepng/lodepng.cpp"
            if (!($or$cond$i$i$i)) {
             $890 = (34884 + ($i$128$i$i$i<<2)|0); //@line 4031 "libraries/lodepng/lodepng.cpp"
             $891 = (34912 + ($i$128$i$i$i<<2)|0); //@line 4031 "libraries/lodepng/lodepng.cpp"
             $892 = (34940 + ($i$128$i$i$i<<2)|0); //@line 4031 "libraries/lodepng/lodepng.cpp"
             $893 = (34968 + ($i$128$i$i$i<<2)|0); //@line 4031 "libraries/lodepng/lodepng.cpp"
             $894 = (($passstart$i$i$i) + ($i$128$i$i$i<<2)|0); //@line 4030 "libraries/lodepng/lodepng.cpp"
             $$pre$i$i$i = HEAP32[$894>>2]|0; //@line 4030 "libraries/lodepng/lodepng.cpp"
             $$pre54$i$i$i = HEAP32[$893>>2]|0; //@line 4031 "libraries/lodepng/lodepng.cpp"
             $$pre55$i$i$i = HEAP32[$892>>2]|0; //@line 4031 "libraries/lodepng/lodepng.cpp"
             $$pre56$i$i$i = HEAP32[$891>>2]|0; //@line 4031 "libraries/lodepng/lodepng.cpp"
             $$pre57$i$i$i = HEAP32[$890>>2]|0; //@line 4031 "libraries/lodepng/lodepng.cpp"
             $895 = $$pre$i$i$i << 3; //@line 4030 "libraries/lodepng/lodepng.cpp"
             $y2$025$us$i$i$i = 0;
             while(1) {
              $896 = Math_imul($y2$025$us$i$i$i, $885)|0; //@line 4030 "libraries/lodepng/lodepng.cpp"
              $897 = Math_imul($y2$025$us$i$i$i, $$pre55$i$i$i)|0; //@line 4031 "libraries/lodepng/lodepng.cpp"
              $898 = (($897) + ($$pre54$i$i$i))|0; //@line 4031 "libraries/lodepng/lodepng.cpp"
              $899 = Math_imul($898, $733)|0; //@line 4031 "libraries/lodepng/lodepng.cpp"
              $900 = (($899) + ($$pre56$i$i$i))|0; //@line 4031 "libraries/lodepng/lodepng.cpp"
              $x1$022$us$us$i$i$i = 0;
              while(1) {
               $tmp$us$us$i$i$i = (($x1$022$us$us$i$i$i) + ($896))|0; //@line 4117 "libraries/lodepng/lodepng.cpp"
               $tmp1$us$us$i$i$i = Math_imul($tmp$us$us$i$i$i, $726)|0; //@line 4117 "libraries/lodepng/lodepng.cpp"
               $902 = (($tmp1$us$us$i$i$i) + ($895))|0; //@line 4030 "libraries/lodepng/lodepng.cpp"
               $903 = Math_imul($x1$022$us$us$i$i$i, $$pre57$i$i$i)|0; //@line 4031 "libraries/lodepng/lodepng.cpp"
               $tmp2$us$us$i$i$i = (($900) + ($903))|0; //@line 4117 "libraries/lodepng/lodepng.cpp"
               $tmp3$us$us$i$i$i = Math_imul($tmp2$us$us$i$i$i, $726)|0; //@line 4031 "libraries/lodepng/lodepng.cpp"
               $b3$018$us$us$i$i$i = 0;$ibp$017$us$us$i$i$i = $902;$obp$016$us$us$i$i$i = $tmp3$us$us$i$i$i;
               while(1) {
                $905 = $ibp$017$us$us$i$i$i >>> 3; //@line 2351 "libraries/lodepng/lodepng.cpp"
                $906 = (($724) + ($905)|0); //@line 2351 "libraries/lodepng/lodepng.cpp"
                $907 = HEAP8[$906>>0]|0; //@line 2351 "libraries/lodepng/lodepng.cpp"
                $908 = $907&255; //@line 2351 "libraries/lodepng/lodepng.cpp"
                $909 = $ibp$017$us$us$i$i$i & 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
                $910 = $909 ^ 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
                $911 = $908 >>> $910; //@line 2351 "libraries/lodepng/lodepng.cpp"
                $912 = $911 & 1; //@line 2351 "libraries/lodepng/lodepng.cpp"
                $913 = (($ibp$017$us$us$i$i$i) + 1)|0; //@line 2352 "libraries/lodepng/lodepng.cpp"
                $914 = ($912|0)==(0); //@line 2371 "libraries/lodepng/lodepng.cpp"
                if (!($914)) {
                 $915 = $obp$016$us$us$i$i$i & 7; //@line 2374 "libraries/lodepng/lodepng.cpp"
                 $916 = $915 ^ 7; //@line 2374 "libraries/lodepng/lodepng.cpp"
                 $917 = $912 << $916; //@line 2374 "libraries/lodepng/lodepng.cpp"
                 $918 = $obp$016$us$us$i$i$i >>> 3; //@line 2374 "libraries/lodepng/lodepng.cpp"
                 $919 = (($723) + ($918)|0); //@line 2374 "libraries/lodepng/lodepng.cpp"
                 $920 = HEAP8[$919>>0]|0; //@line 2374 "libraries/lodepng/lodepng.cpp"
                 $921 = $920&255; //@line 2374 "libraries/lodepng/lodepng.cpp"
                 $922 = $921 | $917; //@line 2374 "libraries/lodepng/lodepng.cpp"
                 $923 = $922&255; //@line 2374 "libraries/lodepng/lodepng.cpp"
                 HEAP8[$919>>0] = $923; //@line 2374 "libraries/lodepng/lodepng.cpp"
                }
                $924 = (($obp$016$us$us$i$i$i) + 1)|0; //@line 2376 "libraries/lodepng/lodepng.cpp"
                $925 = (($b3$018$us$us$i$i$i) + 1)|0; //@line 4032 "libraries/lodepng/lodepng.cpp"
                $exitcond53$i$i$i = ($925|0)==($726|0); //@line 4032 "libraries/lodepng/lodepng.cpp"
                if ($exitcond53$i$i$i) {
                 break;
                } else {
                 $b3$018$us$us$i$i$i = $925;$ibp$017$us$us$i$i$i = $913;$obp$016$us$us$i$i$i = $924;
                }
               }
               $904 = (($x1$022$us$us$i$i$i) + 1)|0; //@line 4028 "libraries/lodepng/lodepng.cpp"
               $exitcond51$i$i = ($904|0)==($885|0); //@line 4028 "libraries/lodepng/lodepng.cpp"
               if ($exitcond51$i$i) {
                break;
               } else {
                $x1$022$us$us$i$i$i = $904;
               }
              }
              $901 = (($y2$025$us$i$i$i) + 1)|0; //@line 4027 "libraries/lodepng/lodepng.cpp"
              $exitcond52$i$i = ($901|0)==($887|0); //@line 4027 "libraries/lodepng/lodepng.cpp"
              if ($exitcond52$i$i) {
               break;
              } else {
               $y2$025$us$i$i$i = $901;
              }
             }
            }
            $926 = (($i$128$i$i$i) + 1)|0; //@line 4021 "libraries/lodepng/lodepng.cpp"
            $927 = ($926|0)==(7); //@line 4021 "libraries/lodepng/lodepng.cpp"
            if ($927) {
             break;
            } else {
             $i$128$i$i$i = $926;
            }
           }
          }
          $$37$i$i = 1;$$4$i$i = $$2$lcssa$i$i;
         }
         if (!($$37$i$i)) {
          $$5$i$i = $$4$i$i;
          break L326;
         }
        }
       } while(0);
       $$5$i$i = 0;
      }
     } while(0);
     HEAP32[$1>>2] = $$5$i$i; //@line 4622 "libraries/lodepng/lodepng.cpp"
     $outv$sroa$0$1132135$i = $outv$sroa$0$1$ph247$i;
    }
    HEAP32[$out>>2] = $outv$sroa$0$1132135$i; //@line 4623 "libraries/lodepng/lodepng.cpp"
   }
   HEAP32[$610>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
   HEAP32[$611>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
   $988 = HEAP32[$scanlines$i>>2]|0; //@line 252 "libraries/lodepng/lodepng.cpp"
   _free($988); //@line 76 "libraries/lodepng/lodepng.cpp"
   $$pre93 = HEAP32[$1>>2]|0; //@line 4634 "libraries/lodepng/lodepng.cpp"
   $989 = ($$pre93|0)==(0); //@line 4634 "libraries/lodepng/lodepng.cpp"
   if (!($989)) {
    $$1 = $$pre93;
    STACKTOP = sp;return ($$1|0); //@line 4671 "libraries/lodepng/lodepng.cpp"
   }
   $990 = ((($state)) + 24|0); //@line 4635 "libraries/lodepng/lodepng.cpp"
   $991 = HEAP32[$990>>2]|0; //@line 4640 "libraries/lodepng/lodepng.cpp"
   $992 = ($991|0)==(0); //@line 4635 "libraries/lodepng/lodepng.cpp"
   $$pre97 = ((($state)) + 100|0); //@line 4642 "libraries/lodepng/lodepng.cpp"
   $$pre98 = ((($state)) + 144|0); //@line 4642 "libraries/lodepng/lodepng.cpp"
   if ($992) {
    $1040 = ((($state)) + 108|0); //@line 2605 "libraries/lodepng/lodepng.cpp"
    $1041 = HEAP32[$1040>>2]|0; //@line 2605 "libraries/lodepng/lodepng.cpp"
    $1042 = ($1041|0)==(0|0); //@line 2605 "libraries/lodepng/lodepng.cpp"
    if (!($1042)) {
     _free($1041); //@line 76 "libraries/lodepng/lodepng.cpp"
    }
    ;HEAP32[$$pre97>>2]=HEAP32[$$pre98>>2]|0;HEAP32[$$pre97+4>>2]=HEAP32[$$pre98+4>>2]|0;HEAP32[$$pre97+8>>2]=HEAP32[$$pre98+8>>2]|0;HEAP32[$$pre97+12>>2]=HEAP32[$$pre98+12>>2]|0;HEAP32[$$pre97+16>>2]=HEAP32[$$pre98+16>>2]|0;HEAP32[$$pre97+20>>2]=HEAP32[$$pre98+20>>2]|0;HEAP32[$$pre97+24>>2]=HEAP32[$$pre98+24>>2]|0;HEAP32[$$pre97+28>>2]=HEAP32[$$pre98+28>>2]|0; //@line 2573 "libraries/lodepng/lodepng.cpp"
    $1043 = ((($state)) + 152|0); //@line 2574 "libraries/lodepng/lodepng.cpp"
    $1044 = HEAP32[$1043>>2]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
    $1045 = ($1044|0)==(0|0); //@line 2574 "libraries/lodepng/lodepng.cpp"
    do {
     if (!($1045)) {
      $1046 = (_malloc(1024)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
      HEAP32[$1040>>2] = $1046; //@line 2576 "libraries/lodepng/lodepng.cpp"
      $1047 = ($1046|0)==(0|0); //@line 2577 "libraries/lodepng/lodepng.cpp"
      $1048 = ((($state)) + 156|0); //@line 2577 "libraries/lodepng/lodepng.cpp"
      $1049 = HEAP32[$1048>>2]|0; //@line 2577 "libraries/lodepng/lodepng.cpp"
      if ($1047) {
       $1050 = ($1049|0)==(0); //@line 2577 "libraries/lodepng/lodepng.cpp"
       if ($1050) {
        break;
       }
       HEAP32[$1>>2] = 83; //@line 4642 "libraries/lodepng/lodepng.cpp"
       $$1 = 83;
       STACKTOP = sp;return ($$1|0); //@line 4671 "libraries/lodepng/lodepng.cpp"
      } else {
       $phitmp$i = $1049 & 1073741823; //@line 2577 "libraries/lodepng/lodepng.cpp"
       $phitmp3$i = ($phitmp$i|0)==(0); //@line 2577 "libraries/lodepng/lodepng.cpp"
       if ($phitmp3$i) {
        break;
       }
       $1051 = HEAP8[$1044>>0]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
       HEAP8[$1046>>0] = $1051; //@line 2578 "libraries/lodepng/lodepng.cpp"
       $1052 = ((($1044)) + 1|0); //@line 2578 "libraries/lodepng/lodepng.cpp"
       $1053 = HEAP8[$1052>>0]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
       $1054 = ((($1046)) + 1|0); //@line 2578 "libraries/lodepng/lodepng.cpp"
       HEAP8[$1054>>0] = $1053; //@line 2578 "libraries/lodepng/lodepng.cpp"
       $1056 = 2;
       while(1) {
        $$pre2$i$pre = HEAP32[$1040>>2]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
        $$pre95 = HEAP32[$1043>>2]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
        $1055 = (($$pre95) + ($1056)|0); //@line 2578 "libraries/lodepng/lodepng.cpp"
        $1057 = HEAP8[$1055>>0]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
        $1058 = (($$pre2$i$pre) + ($1056)|0); //@line 2578 "libraries/lodepng/lodepng.cpp"
        HEAP8[$1058>>0] = $1057; //@line 2578 "libraries/lodepng/lodepng.cpp"
        $1059 = (($1056) + 1)|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
        $1060 = HEAP32[$1048>>2]|0; //@line 2578 "libraries/lodepng/lodepng.cpp"
        $1061 = $1060 << 2; //@line 2578 "libraries/lodepng/lodepng.cpp"
        $1062 = ($1059|0)==($1061|0); //@line 2578 "libraries/lodepng/lodepng.cpp"
        if ($1062) {
         break;
        } else {
         $1056 = $1059;
        }
       }
      }
     }
    } while(0);
    HEAP32[$1>>2] = 0; //@line 4642 "libraries/lodepng/lodepng.cpp"
    $$1 = 0;
    STACKTOP = sp;return ($$1|0); //@line 4671 "libraries/lodepng/lodepng.cpp"
   }
   $993 = HEAP32[$$pre97>>2]|0; //@line 2634 "libraries/lodepng/lodepng.cpp"
   $994 = HEAP32[$$pre98>>2]|0; //@line 2586 "libraries/lodepng/lodepng.cpp"
   $995 = ($993|0)==($994|0); //@line 2586 "libraries/lodepng/lodepng.cpp"
   L428: do {
    if ($995) {
     $996 = ((($state)) + 104|0); //@line 2587 "libraries/lodepng/lodepng.cpp"
     $997 = HEAP32[$996>>2]|0; //@line 2587 "libraries/lodepng/lodepng.cpp"
     $998 = ((($state)) + 148|0); //@line 2587 "libraries/lodepng/lodepng.cpp"
     $999 = HEAP32[$998>>2]|0; //@line 2587 "libraries/lodepng/lodepng.cpp"
     $1000 = ($997|0)==($999|0); //@line 2587 "libraries/lodepng/lodepng.cpp"
     if ($1000) {
      $1001 = ((($state)) + 116|0); //@line 2588 "libraries/lodepng/lodepng.cpp"
      $1002 = HEAP32[$1001>>2]|0; //@line 2588 "libraries/lodepng/lodepng.cpp"
      $1003 = ((($state)) + 160|0); //@line 2588 "libraries/lodepng/lodepng.cpp"
      $1004 = HEAP32[$1003>>2]|0; //@line 2588 "libraries/lodepng/lodepng.cpp"
      $1005 = ($1002|0)==($1004|0); //@line 2588 "libraries/lodepng/lodepng.cpp"
      if ($1005) {
       $1006 = ($1002|0)==(0); //@line 2589 "libraries/lodepng/lodepng.cpp"
       if (!($1006)) {
        $1007 = ((($state)) + 120|0); //@line 2591 "libraries/lodepng/lodepng.cpp"
        $1008 = HEAP32[$1007>>2]|0; //@line 2591 "libraries/lodepng/lodepng.cpp"
        $1009 = ((($state)) + 164|0); //@line 2591 "libraries/lodepng/lodepng.cpp"
        $1010 = HEAP32[$1009>>2]|0; //@line 2591 "libraries/lodepng/lodepng.cpp"
        $1011 = ($1008|0)==($1010|0); //@line 2591 "libraries/lodepng/lodepng.cpp"
        if (!($1011)) {
         break;
        }
        $1012 = ((($state)) + 124|0); //@line 2592 "libraries/lodepng/lodepng.cpp"
        $1013 = HEAP32[$1012>>2]|0; //@line 2592 "libraries/lodepng/lodepng.cpp"
        $1014 = ((($state)) + 168|0); //@line 2592 "libraries/lodepng/lodepng.cpp"
        $1015 = HEAP32[$1014>>2]|0; //@line 2592 "libraries/lodepng/lodepng.cpp"
        $1016 = ($1013|0)==($1015|0); //@line 2592 "libraries/lodepng/lodepng.cpp"
        if (!($1016)) {
         break;
        }
        $1017 = ((($state)) + 128|0); //@line 2593 "libraries/lodepng/lodepng.cpp"
        $1018 = HEAP32[$1017>>2]|0; //@line 2593 "libraries/lodepng/lodepng.cpp"
        $1019 = ((($state)) + 172|0); //@line 2593 "libraries/lodepng/lodepng.cpp"
        $1020 = HEAP32[$1019>>2]|0; //@line 2593 "libraries/lodepng/lodepng.cpp"
        $1021 = ($1018|0)==($1020|0); //@line 2593 "libraries/lodepng/lodepng.cpp"
        if (!($1021)) {
         break;
        }
       }
       $1022 = ((($state)) + 112|0); //@line 2595 "libraries/lodepng/lodepng.cpp"
       $1023 = HEAP32[$1022>>2]|0; //@line 2596 "libraries/lodepng/lodepng.cpp"
       $1024 = ((($state)) + 156|0); //@line 2595 "libraries/lodepng/lodepng.cpp"
       $1025 = HEAP32[$1024>>2]|0; //@line 2595 "libraries/lodepng/lodepng.cpp"
       $1026 = ($1023|0)==($1025|0); //@line 2595 "libraries/lodepng/lodepng.cpp"
       if ($1026) {
        $$mask$i = $1023 & 1073741823; //@line 2596 "libraries/lodepng/lodepng.cpp"
        $1027 = ($$mask$i|0)==(0); //@line 2596 "libraries/lodepng/lodepng.cpp"
        if ($1027) {
         $$1 = 0;
         STACKTOP = sp;return ($$1|0); //@line 4671 "libraries/lodepng/lodepng.cpp"
        }
        $1028 = ((($state)) + 108|0); //@line 2598 "libraries/lodepng/lodepng.cpp"
        $1029 = HEAP32[$1028>>2]|0; //@line 2598 "libraries/lodepng/lodepng.cpp"
        $1030 = ((($state)) + 152|0); //@line 2598 "libraries/lodepng/lodepng.cpp"
        $1031 = HEAP32[$1030>>2]|0; //@line 2598 "libraries/lodepng/lodepng.cpp"
        $1032 = $1023 << 2; //@line 2596 "libraries/lodepng/lodepng.cpp"
        $i$01$i = 0;
        while(1) {
         $1035 = (($1029) + ($i$01$i)|0); //@line 2598 "libraries/lodepng/lodepng.cpp"
         $1036 = HEAP8[$1035>>0]|0; //@line 2598 "libraries/lodepng/lodepng.cpp"
         $1037 = (($1031) + ($i$01$i)|0); //@line 2598 "libraries/lodepng/lodepng.cpp"
         $1038 = HEAP8[$1037>>0]|0; //@line 2598 "libraries/lodepng/lodepng.cpp"
         $1039 = ($1036<<24>>24)==($1038<<24>>24); //@line 2598 "libraries/lodepng/lodepng.cpp"
         $1034 = (($i$01$i) + 1)|0; //@line 2596 "libraries/lodepng/lodepng.cpp"
         if (!($1039)) {
          break L428;
         }
         $1033 = ($1034|0)==($1032|0); //@line 2596 "libraries/lodepng/lodepng.cpp"
         if ($1033) {
          $$1 = 0;
          break;
         } else {
          $i$01$i = $1034;
         }
        }
        STACKTOP = sp;return ($$1|0); //@line 4671 "libraries/lodepng/lodepng.cpp"
       }
      }
     }
    }
   } while(0);
   $1063 = HEAP32[$out>>2]|0; //@line 4649 "libraries/lodepng/lodepng.cpp"
   switch ($993|0) {
   case 6: case 2:  {
    break;
   }
   default: {
    $1064 = ((($state)) + 104|0); //@line 4655 "libraries/lodepng/lodepng.cpp"
    $1065 = HEAP32[$1064>>2]|0; //@line 4655 "libraries/lodepng/lodepng.cpp"
    $1066 = ($1065|0)==(8); //@line 4655 "libraries/lodepng/lodepng.cpp"
    if (!($1066)) {
     $$1 = 56;
     STACKTOP = sp;return ($$1|0); //@line 4671 "libraries/lodepng/lodepng.cpp"
    }
   }
   }
   $1067 = HEAP32[$w>>2]|0; //@line 4667 "libraries/lodepng/lodepng.cpp"
   $1068 = HEAP32[$h>>2]|0; //@line 4667 "libraries/lodepng/lodepng.cpp"
   $1069 = ((($state)) + 104|0); //@line 2634 "libraries/lodepng/lodepng.cpp"
   $1070 = HEAP32[$1069>>2]|0; //@line 2634 "libraries/lodepng/lodepng.cpp"
   switch ($993|0) {
   case 3: case 0:  {
    $$0$i$i$i$i = 1;
    break;
   }
   case 2:  {
    $$0$i$i$i$i = 3;
    break;
   }
   case 4:  {
    $$0$i$i$i$i = 2;
    break;
   }
   case 6:  {
    $$0$i$i$i$i = 4;
    break;
   }
   default: {
    $$0$i$i$i$i = 0;
   }
   }
   $1071 = Math_imul($1068, $1067)|0; //@line 2549 "libraries/lodepng/lodepng.cpp"
   $1072 = Math_imul($1071, $1070)|0; //@line 2676 "libraries/lodepng/lodepng.cpp"
   $1073 = Math_imul($1072, $$0$i$i$i$i)|0; //@line 2676 "libraries/lodepng/lodepng.cpp"
   $1074 = (($1073) + 7)|0; //@line 2676 "libraries/lodepng/lodepng.cpp"
   $1075 = $1074 >>> 3; //@line 2676 "libraries/lodepng/lodepng.cpp"
   $1076 = (_malloc($1075)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
   HEAP32[$out>>2] = $1076; //@line 4661 "libraries/lodepng/lodepng.cpp"
   $1077 = ($1076|0)==(0|0); //@line 4662 "libraries/lodepng/lodepng.cpp"
   if ($1077) {
    $storemerge = 83;
   } else {
    (__Z15lodepng_convertPhPKhP16LodePNGColorModePKS2_jj($1076,$1063,$$pre97,$$pre98,$1067,$1068)|0); //@line 4666 "libraries/lodepng/lodepng.cpp"
    $storemerge = 0;
   }
   HEAP32[$1>>2] = $storemerge; //@line 4666 "libraries/lodepng/lodepng.cpp"
   _free($1063); //@line 76 "libraries/lodepng/lodepng.cpp"
   $$pre96 = HEAP32[$1>>2]|0; //@line 4670 "libraries/lodepng/lodepng.cpp"
   $$1 = $$pre96;
   STACKTOP = sp;return ($$1|0); //@line 4671 "libraries/lodepng/lodepng.cpp"
  } else {
   $$ph = $0;
  }
 } while(0);
 $$1 = $$ph;
 STACKTOP = sp;return ($$1|0); //@line 4671 "libraries/lodepng/lodepng.cpp"
}
function __ZL8unfilterPhPKhjjj($out,$in,$w,$h,$bpp) {
 $out = $out|0;
 $in = $in|0;
 $w = $w|0;
 $h = $h|0;
 $bpp = $bpp|0;
 var $$$i = 0, $$0$i$i = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $div$i = 0, $div1$i = 0, $exitcond$i = 0, $exitcond52$i = 0, $exitcond53$i = 0, $exitcond54$i = 0, $exitcond55$i = 0, $factor$i$i = 0, $i$021$i = 0, $i$1042$i = 0, $i$1140$i = 0, $i$1238$i = 0, $i$124$i = 0, $i$222$i = 0, $i$328$i = 0, $i$426$i = 0, $i$536$i = 0, $i$634$i = 0;
 var $i$732$i = 0, $i$830$i = 0, $i$943$i = 0, $ispos$i$i = 0, $ispos1$i$i = 0, $ispos3$i$i = 0, $neg$i$i = 0, $neg2$i$i = 0, $neg4$i$i = 0, $or$cond$i$i = 0, $prevline$017 = 0, $sext$i$i = 0, $y$018 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($bpp) + 7)|0; //@line 3965 "libraries/lodepng/lodepng.cpp"
 $1 = $0 >>> 3; //@line 3965 "libraries/lodepng/lodepng.cpp"
 $2 = Math_imul($bpp, $w)|0; //@line 3966 "libraries/lodepng/lodepng.cpp"
 $3 = (($2) + 7)|0; //@line 3966 "libraries/lodepng/lodepng.cpp"
 $4 = $3 >>> 3; //@line 3966 "libraries/lodepng/lodepng.cpp"
 $5 = ($h|0)==(0); //@line 3968 "libraries/lodepng/lodepng.cpp"
 if ($5) {
  $$2 = 0;
  return ($$2|0); //@line 3980 "libraries/lodepng/lodepng.cpp"
 }
 $6 = (($4) + 1)|0; //@line 3971 "libraries/lodepng/lodepng.cpp"
 $7 = ($4|0)==(0); //@line 3893 "libraries/lodepng/lodepng.cpp"
 $8 = ($1|0)==(0); //@line 3896 "libraries/lodepng/lodepng.cpp"
 $9 = ($1>>>0)<($4>>>0); //@line 3897 "libraries/lodepng/lodepng.cpp"
 $prevline$017 = 0;$y$018 = 0;
 L4: while(1) {
  $10 = Math_imul($y$018, $4)|0; //@line 3970 "libraries/lodepng/lodepng.cpp"
  $11 = Math_imul($y$018, $6)|0; //@line 3971 "libraries/lodepng/lodepng.cpp"
  $12 = (($in) + ($11)|0); //@line 3972 "libraries/lodepng/lodepng.cpp"
  $13 = HEAP8[$12>>0]|0; //@line 3972 "libraries/lodepng/lodepng.cpp"
  $14 = (($out) + ($10)|0); //@line 3974 "libraries/lodepng/lodepng.cpp"
  $15 = (($11) + 1)|0; //@line 3974 "libraries/lodepng/lodepng.cpp"
  $16 = (($in) + ($15)|0); //@line 3974 "libraries/lodepng/lodepng.cpp"
  $17 = $13&255; //@line 3890 "libraries/lodepng/lodepng.cpp"
  L6: do {
   switch ($17|0) {
   case 0:  {
    if (!($7)) {
     $i$021$i = 0;
     while(1) {
      $18 = (($16) + ($i$021$i)|0); //@line 3893 "libraries/lodepng/lodepng.cpp"
      $19 = HEAP8[$18>>0]|0; //@line 3893 "libraries/lodepng/lodepng.cpp"
      $20 = (($14) + ($i$021$i)|0); //@line 3893 "libraries/lodepng/lodepng.cpp"
      HEAP8[$20>>0] = $19; //@line 3893 "libraries/lodepng/lodepng.cpp"
      $21 = (($i$021$i) + 1)|0; //@line 3893 "libraries/lodepng/lodepng.cpp"
      $22 = ($21|0)==($4|0); //@line 3893 "libraries/lodepng/lodepng.cpp"
      if ($22) {
       break;
      } else {
       $i$021$i = $21;
      }
     }
    }
    break;
   }
   case 1:  {
    if (!($8)) {
     $i$124$i = 0;
     while(1) {
      $23 = (($16) + ($i$124$i)|0); //@line 3896 "libraries/lodepng/lodepng.cpp"
      $24 = HEAP8[$23>>0]|0; //@line 3896 "libraries/lodepng/lodepng.cpp"
      $25 = (($14) + ($i$124$i)|0); //@line 3896 "libraries/lodepng/lodepng.cpp"
      HEAP8[$25>>0] = $24; //@line 3896 "libraries/lodepng/lodepng.cpp"
      $26 = (($i$124$i) + 1)|0; //@line 3896 "libraries/lodepng/lodepng.cpp"
      $27 = ($26|0)==($1|0); //@line 3896 "libraries/lodepng/lodepng.cpp"
      if ($27) {
       break;
      } else {
       $i$124$i = $26;
      }
     }
    }
    if ($9) {
     $i$222$i = $1;
     while(1) {
      $28 = (($16) + ($i$222$i)|0); //@line 3897 "libraries/lodepng/lodepng.cpp"
      $29 = HEAP8[$28>>0]|0; //@line 3897 "libraries/lodepng/lodepng.cpp"
      $30 = $29&255; //@line 3897 "libraries/lodepng/lodepng.cpp"
      $31 = (($i$222$i) - ($1))|0; //@line 3897 "libraries/lodepng/lodepng.cpp"
      $32 = (($14) + ($31)|0); //@line 3897 "libraries/lodepng/lodepng.cpp"
      $33 = HEAP8[$32>>0]|0; //@line 3897 "libraries/lodepng/lodepng.cpp"
      $34 = $33&255; //@line 3897 "libraries/lodepng/lodepng.cpp"
      $35 = (($34) + ($30))|0; //@line 3897 "libraries/lodepng/lodepng.cpp"
      $36 = $35&255; //@line 3897 "libraries/lodepng/lodepng.cpp"
      $37 = (($14) + ($i$222$i)|0); //@line 3897 "libraries/lodepng/lodepng.cpp"
      HEAP8[$37>>0] = $36; //@line 3897 "libraries/lodepng/lodepng.cpp"
      $38 = (($i$222$i) + 1)|0; //@line 3897 "libraries/lodepng/lodepng.cpp"
      $exitcond$i = ($38|0)==($4|0); //@line 3897 "libraries/lodepng/lodepng.cpp"
      if ($exitcond$i) {
       break;
      } else {
       $i$222$i = $38;
      }
     }
    }
    break;
   }
   case 2:  {
    $39 = ($prevline$017|0)==(0|0); //@line 3900 "libraries/lodepng/lodepng.cpp"
    if ($39) {
     if ($7) {
      break L6;
     } else {
      $i$426$i = 0;
     }
     while(1) {
      $51 = (($16) + ($i$426$i)|0); //@line 3906 "libraries/lodepng/lodepng.cpp"
      $52 = HEAP8[$51>>0]|0; //@line 3906 "libraries/lodepng/lodepng.cpp"
      $53 = (($14) + ($i$426$i)|0); //@line 3906 "libraries/lodepng/lodepng.cpp"
      HEAP8[$53>>0] = $52; //@line 3906 "libraries/lodepng/lodepng.cpp"
      $54 = (($i$426$i) + 1)|0; //@line 3906 "libraries/lodepng/lodepng.cpp"
      $55 = ($54|0)==($4|0); //@line 3906 "libraries/lodepng/lodepng.cpp"
      if ($55) {
       break;
      } else {
       $i$426$i = $54;
      }
     }
    } else {
     if ($7) {
      break L6;
     } else {
      $i$328$i = 0;
     }
     while(1) {
      $40 = (($16) + ($i$328$i)|0); //@line 3902 "libraries/lodepng/lodepng.cpp"
      $41 = HEAP8[$40>>0]|0; //@line 3902 "libraries/lodepng/lodepng.cpp"
      $42 = $41&255; //@line 3902 "libraries/lodepng/lodepng.cpp"
      $43 = (($prevline$017) + ($i$328$i)|0); //@line 3902 "libraries/lodepng/lodepng.cpp"
      $44 = HEAP8[$43>>0]|0; //@line 3902 "libraries/lodepng/lodepng.cpp"
      $45 = $44&255; //@line 3902 "libraries/lodepng/lodepng.cpp"
      $46 = (($45) + ($42))|0; //@line 3902 "libraries/lodepng/lodepng.cpp"
      $47 = $46&255; //@line 3902 "libraries/lodepng/lodepng.cpp"
      $48 = (($14) + ($i$328$i)|0); //@line 3902 "libraries/lodepng/lodepng.cpp"
      HEAP8[$48>>0] = $47; //@line 3902 "libraries/lodepng/lodepng.cpp"
      $49 = (($i$328$i) + 1)|0; //@line 3902 "libraries/lodepng/lodepng.cpp"
      $50 = ($49|0)==($4|0); //@line 3902 "libraries/lodepng/lodepng.cpp"
      if ($50) {
       break;
      } else {
       $i$328$i = $49;
      }
     }
    }
    break;
   }
   case 3:  {
    $56 = ($prevline$017|0)==(0|0); //@line 3910 "libraries/lodepng/lodepng.cpp"
    if ($56) {
     if (!($8)) {
      $i$732$i = 0;
      while(1) {
       $84 = (($16) + ($i$732$i)|0); //@line 3917 "libraries/lodepng/lodepng.cpp"
       $85 = HEAP8[$84>>0]|0; //@line 3917 "libraries/lodepng/lodepng.cpp"
       $86 = (($14) + ($i$732$i)|0); //@line 3917 "libraries/lodepng/lodepng.cpp"
       HEAP8[$86>>0] = $85; //@line 3917 "libraries/lodepng/lodepng.cpp"
       $87 = (($i$732$i) + 1)|0; //@line 3917 "libraries/lodepng/lodepng.cpp"
       $88 = ($87|0)==($1|0); //@line 3917 "libraries/lodepng/lodepng.cpp"
       if ($88) {
        break;
       } else {
        $i$732$i = $87;
       }
      }
     }
     if ($9) {
      $i$830$i = $1;
     } else {
      break L6;
     }
     while(1) {
      $89 = (($16) + ($i$830$i)|0); //@line 3918 "libraries/lodepng/lodepng.cpp"
      $90 = HEAP8[$89>>0]|0; //@line 3918 "libraries/lodepng/lodepng.cpp"
      $91 = $90&255; //@line 3918 "libraries/lodepng/lodepng.cpp"
      $92 = (($i$830$i) - ($1))|0; //@line 3918 "libraries/lodepng/lodepng.cpp"
      $93 = (($14) + ($92)|0); //@line 3918 "libraries/lodepng/lodepng.cpp"
      $94 = HEAP8[$93>>0]|0; //@line 3918 "libraries/lodepng/lodepng.cpp"
      $div$i = ($94&255) >>> 1; //@line 3918 "libraries/lodepng/lodepng.cpp"
      $95 = $div$i&255; //@line 3918 "libraries/lodepng/lodepng.cpp"
      $96 = (($95) + ($91))|0; //@line 3918 "libraries/lodepng/lodepng.cpp"
      $97 = $96&255; //@line 3918 "libraries/lodepng/lodepng.cpp"
      $98 = (($14) + ($i$830$i)|0); //@line 3918 "libraries/lodepng/lodepng.cpp"
      HEAP8[$98>>0] = $97; //@line 3918 "libraries/lodepng/lodepng.cpp"
      $99 = (($i$830$i) + 1)|0; //@line 3918 "libraries/lodepng/lodepng.cpp"
      $exitcond52$i = ($99|0)==($4|0); //@line 3918 "libraries/lodepng/lodepng.cpp"
      if ($exitcond52$i) {
       break;
      } else {
       $i$830$i = $99;
      }
     }
    } else {
     if (!($8)) {
      $i$536$i = 0;
      while(1) {
       $57 = (($16) + ($i$536$i)|0); //@line 3912 "libraries/lodepng/lodepng.cpp"
       $58 = HEAP8[$57>>0]|0; //@line 3912 "libraries/lodepng/lodepng.cpp"
       $59 = $58&255; //@line 3912 "libraries/lodepng/lodepng.cpp"
       $60 = (($prevline$017) + ($i$536$i)|0); //@line 3912 "libraries/lodepng/lodepng.cpp"
       $61 = HEAP8[$60>>0]|0; //@line 3912 "libraries/lodepng/lodepng.cpp"
       $div1$i = ($61&255) >>> 1; //@line 3912 "libraries/lodepng/lodepng.cpp"
       $62 = $div1$i&255; //@line 3912 "libraries/lodepng/lodepng.cpp"
       $63 = (($62) + ($59))|0; //@line 3912 "libraries/lodepng/lodepng.cpp"
       $64 = $63&255; //@line 3912 "libraries/lodepng/lodepng.cpp"
       $65 = (($14) + ($i$536$i)|0); //@line 3912 "libraries/lodepng/lodepng.cpp"
       HEAP8[$65>>0] = $64; //@line 3912 "libraries/lodepng/lodepng.cpp"
       $66 = (($i$536$i) + 1)|0; //@line 3912 "libraries/lodepng/lodepng.cpp"
       $67 = ($66|0)==($1|0); //@line 3912 "libraries/lodepng/lodepng.cpp"
       if ($67) {
        break;
       } else {
        $i$536$i = $66;
       }
      }
     }
     if ($9) {
      $i$634$i = $1;
     } else {
      break L6;
     }
     while(1) {
      $68 = (($16) + ($i$634$i)|0); //@line 3913 "libraries/lodepng/lodepng.cpp"
      $69 = HEAP8[$68>>0]|0; //@line 3913 "libraries/lodepng/lodepng.cpp"
      $70 = $69&255; //@line 3913 "libraries/lodepng/lodepng.cpp"
      $71 = (($i$634$i) - ($1))|0; //@line 3913 "libraries/lodepng/lodepng.cpp"
      $72 = (($14) + ($71)|0); //@line 3913 "libraries/lodepng/lodepng.cpp"
      $73 = HEAP8[$72>>0]|0; //@line 3913 "libraries/lodepng/lodepng.cpp"
      $74 = $73&255; //@line 3913 "libraries/lodepng/lodepng.cpp"
      $75 = (($prevline$017) + ($i$634$i)|0); //@line 3913 "libraries/lodepng/lodepng.cpp"
      $76 = HEAP8[$75>>0]|0; //@line 3913 "libraries/lodepng/lodepng.cpp"
      $77 = $76&255; //@line 3913 "libraries/lodepng/lodepng.cpp"
      $78 = (($77) + ($74))|0; //@line 3913 "libraries/lodepng/lodepng.cpp"
      $79 = $78 >>> 1; //@line 3913 "libraries/lodepng/lodepng.cpp"
      $80 = (($79) + ($70))|0; //@line 3913 "libraries/lodepng/lodepng.cpp"
      $81 = $80&255; //@line 3913 "libraries/lodepng/lodepng.cpp"
      $82 = (($14) + ($i$634$i)|0); //@line 3913 "libraries/lodepng/lodepng.cpp"
      HEAP8[$82>>0] = $81; //@line 3913 "libraries/lodepng/lodepng.cpp"
      $83 = (($i$634$i) + 1)|0; //@line 3913 "libraries/lodepng/lodepng.cpp"
      $exitcond53$i = ($83|0)==($4|0); //@line 3913 "libraries/lodepng/lodepng.cpp"
      if ($exitcond53$i) {
       break;
      } else {
       $i$634$i = $83;
      }
     }
    }
    break;
   }
   case 4:  {
    $100 = ($prevline$017|0)==(0|0); //@line 3922 "libraries/lodepng/lodepng.cpp"
    if ($100) {
     if (!($8)) {
      $i$1140$i = 0;
      while(1) {
       $141 = (($16) + ($i$1140$i)|0); //@line 3937 "libraries/lodepng/lodepng.cpp"
       $142 = HEAP8[$141>>0]|0; //@line 3937 "libraries/lodepng/lodepng.cpp"
       $143 = (($14) + ($i$1140$i)|0); //@line 3937 "libraries/lodepng/lodepng.cpp"
       HEAP8[$143>>0] = $142; //@line 3937 "libraries/lodepng/lodepng.cpp"
       $144 = (($i$1140$i) + 1)|0; //@line 3935 "libraries/lodepng/lodepng.cpp"
       $145 = ($144|0)==($1|0); //@line 3935 "libraries/lodepng/lodepng.cpp"
       if ($145) {
        break;
       } else {
        $i$1140$i = $144;
       }
      }
     }
     if ($9) {
      $i$1238$i = $1;
     } else {
      break L6;
     }
     while(1) {
      $146 = (($16) + ($i$1238$i)|0); //@line 3942 "libraries/lodepng/lodepng.cpp"
      $147 = HEAP8[$146>>0]|0; //@line 3942 "libraries/lodepng/lodepng.cpp"
      $148 = $147&255; //@line 3942 "libraries/lodepng/lodepng.cpp"
      $149 = (($i$1238$i) - ($1))|0; //@line 3942 "libraries/lodepng/lodepng.cpp"
      $150 = (($14) + ($149)|0); //@line 3942 "libraries/lodepng/lodepng.cpp"
      $151 = HEAP8[$150>>0]|0; //@line 3942 "libraries/lodepng/lodepng.cpp"
      $152 = $151&255; //@line 3942 "libraries/lodepng/lodepng.cpp"
      $153 = (($152) + ($148))|0; //@line 3942 "libraries/lodepng/lodepng.cpp"
      $154 = $153&255; //@line 3942 "libraries/lodepng/lodepng.cpp"
      $155 = (($14) + ($i$1238$i)|0); //@line 3942 "libraries/lodepng/lodepng.cpp"
      HEAP8[$155>>0] = $154; //@line 3942 "libraries/lodepng/lodepng.cpp"
      $156 = (($i$1238$i) + 1)|0; //@line 3939 "libraries/lodepng/lodepng.cpp"
      $exitcond54$i = ($156|0)==($4|0); //@line 3939 "libraries/lodepng/lodepng.cpp"
      if ($exitcond54$i) {
       break;
      } else {
       $i$1238$i = $156;
      }
     }
    } else {
     if (!($8)) {
      $i$943$i = 0;
      while(1) {
       $101 = (($16) + ($i$943$i)|0); //@line 3926 "libraries/lodepng/lodepng.cpp"
       $102 = HEAP8[$101>>0]|0; //@line 3926 "libraries/lodepng/lodepng.cpp"
       $103 = $102&255; //@line 3926 "libraries/lodepng/lodepng.cpp"
       $104 = (($prevline$017) + ($i$943$i)|0); //@line 3926 "libraries/lodepng/lodepng.cpp"
       $105 = HEAP8[$104>>0]|0; //@line 3926 "libraries/lodepng/lodepng.cpp"
       $106 = $105&255; //@line 3926 "libraries/lodepng/lodepng.cpp"
       $107 = (($106) + ($103))|0; //@line 3926 "libraries/lodepng/lodepng.cpp"
       $108 = $107&255; //@line 3926 "libraries/lodepng/lodepng.cpp"
       $109 = (($14) + ($i$943$i)|0); //@line 3926 "libraries/lodepng/lodepng.cpp"
       HEAP8[$109>>0] = $108; //@line 3926 "libraries/lodepng/lodepng.cpp"
       $110 = (($i$943$i) + 1)|0; //@line 3924 "libraries/lodepng/lodepng.cpp"
       $111 = ($110|0)==($1|0); //@line 3924 "libraries/lodepng/lodepng.cpp"
       if ($111) {
        break;
       } else {
        $i$943$i = $110;
       }
      }
     }
     if ($9) {
      $i$1042$i = $1;
     } else {
      break L6;
     }
     while(1) {
      $112 = (($16) + ($i$1042$i)|0); //@line 3930 "libraries/lodepng/lodepng.cpp"
      $113 = HEAP8[$112>>0]|0; //@line 3930 "libraries/lodepng/lodepng.cpp"
      $114 = $113&255; //@line 3930 "libraries/lodepng/lodepng.cpp"
      $115 = (($i$1042$i) - ($1))|0; //@line 3930 "libraries/lodepng/lodepng.cpp"
      $116 = (($14) + ($115)|0); //@line 3930 "libraries/lodepng/lodepng.cpp"
      $117 = HEAP8[$116>>0]|0; //@line 3930 "libraries/lodepng/lodepng.cpp"
      $118 = (($prevline$017) + ($i$1042$i)|0); //@line 3930 "libraries/lodepng/lodepng.cpp"
      $119 = HEAP8[$118>>0]|0; //@line 3930 "libraries/lodepng/lodepng.cpp"
      $120 = (($prevline$017) + ($115)|0); //@line 3930 "libraries/lodepng/lodepng.cpp"
      $121 = HEAP8[$120>>0]|0; //@line 3930 "libraries/lodepng/lodepng.cpp"
      $122 = $119&255; //@line 3749 "libraries/lodepng/lodepng.cpp"
      $123 = $121&255; //@line 3749 "libraries/lodepng/lodepng.cpp"
      $124 = (($122) - ($123))|0; //@line 3749 "libraries/lodepng/lodepng.cpp"
      $ispos$i$i = ($124|0)>(-1); //@line 3749 "libraries/lodepng/lodepng.cpp"
      $neg$i$i = (0 - ($124))|0; //@line 3749 "libraries/lodepng/lodepng.cpp"
      $125 = $ispos$i$i ? $124 : $neg$i$i; //@line 3749 "libraries/lodepng/lodepng.cpp"
      $126 = $117&255; //@line 3750 "libraries/lodepng/lodepng.cpp"
      $127 = (($126) - ($123))|0; //@line 3750 "libraries/lodepng/lodepng.cpp"
      $ispos1$i$i = ($127|0)>(-1); //@line 3750 "libraries/lodepng/lodepng.cpp"
      $neg2$i$i = (0 - ($127))|0; //@line 3750 "libraries/lodepng/lodepng.cpp"
      $128 = $ispos1$i$i ? $127 : $neg2$i$i; //@line 3750 "libraries/lodepng/lodepng.cpp"
      $factor$i$i = Math_imul($123, -2)|0; //@line 3930 "libraries/lodepng/lodepng.cpp"
      $129 = (($122) + ($126))|0; //@line 3751 "libraries/lodepng/lodepng.cpp"
      $130 = (($129) + ($factor$i$i))|0; //@line 3751 "libraries/lodepng/lodepng.cpp"
      $ispos3$i$i = ($130|0)>(-1); //@line 3751 "libraries/lodepng/lodepng.cpp"
      $neg4$i$i = (0 - ($130))|0; //@line 3751 "libraries/lodepng/lodepng.cpp"
      $131 = $ispos3$i$i ? $130 : $neg4$i$i; //@line 3751 "libraries/lodepng/lodepng.cpp"
      $sext$i$i = $131 << 16; //@line 3753 "libraries/lodepng/lodepng.cpp"
      $132 = $sext$i$i >> 16; //@line 3753 "libraries/lodepng/lodepng.cpp"
      $133 = ($132|0)<($125|0); //@line 3753 "libraries/lodepng/lodepng.cpp"
      $134 = ($132|0)<($128|0); //@line 3753 "libraries/lodepng/lodepng.cpp"
      $or$cond$i$i = $133 & $134; //@line 3753 "libraries/lodepng/lodepng.cpp"
      $135 = ($128|0)<($125|0); //@line 3754 "libraries/lodepng/lodepng.cpp"
      $$$i = $135 ? $119 : $117; //@line 3754 "libraries/lodepng/lodepng.cpp"
      $$0$i$i = $or$cond$i$i ? $121 : $$$i; //@line 3753 "libraries/lodepng/lodepng.cpp"
      $136 = $$0$i$i&255; //@line 3930 "libraries/lodepng/lodepng.cpp"
      $137 = (($136) + ($114))|0; //@line 3930 "libraries/lodepng/lodepng.cpp"
      $138 = $137&255; //@line 3930 "libraries/lodepng/lodepng.cpp"
      $139 = (($14) + ($i$1042$i)|0); //@line 3930 "libraries/lodepng/lodepng.cpp"
      HEAP8[$139>>0] = $138; //@line 3930 "libraries/lodepng/lodepng.cpp"
      $140 = (($i$1042$i) + 1)|0; //@line 3928 "libraries/lodepng/lodepng.cpp"
      $exitcond55$i = ($140|0)==($4|0); //@line 3928 "libraries/lodepng/lodepng.cpp"
      if ($exitcond55$i) {
       break;
      } else {
       $i$1042$i = $140;
      }
     }
    }
    break;
   }
   default: {
    $$2 = 36;
    label = 34;
    break L4;
   }
   }
  } while(0);
  $157 = (($y$018) + 1)|0; //@line 3968 "libraries/lodepng/lodepng.cpp"
  $158 = ($157>>>0)<($h>>>0); //@line 3968 "libraries/lodepng/lodepng.cpp"
  if ($158) {
   $prevline$017 = $14;$y$018 = $157;
  } else {
   $$2 = 0;
   label = 34;
   break;
  }
 }
 if ((label|0) == 34) {
  return ($$2|0); //@line 3980 "libraries/lodepng/lodepng.cpp"
 }
 return (0)|0;
}
function __ZL19Adam7_getpassvaluesPjS_S_S_S_jjj($passw,$passh,$filter_passstart,$padded_passstart,$passstart,$w,$h,$bpp) {
 $passw = $passw|0;
 $passh = $passh|0;
 $filter_passstart = $filter_passstart|0;
 $padded_passstart = $padded_passstart|0;
 $passstart = $passstart|0;
 $w = $w|0;
 $h = $h|0;
 $bpp = $bpp|0;
 var $$ = 0, $$pre = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $i$03 = 0, $i$12 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($w) + -1)|0; //@line 3789 "libraries/lodepng/lodepng.cpp"
 $1 = (($h) + -1)|0; //@line 3790 "libraries/lodepng/lodepng.cpp"
 $i$03 = 0;
 while(1) {
  $2 = (34884 + ($i$03<<2)|0); //@line 3789 "libraries/lodepng/lodepng.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 3789 "libraries/lodepng/lodepng.cpp"
  $4 = (34912 + ($i$03<<2)|0); //@line 3789 "libraries/lodepng/lodepng.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 3789 "libraries/lodepng/lodepng.cpp"
  $6 = (($0) + ($3))|0; //@line 3789 "libraries/lodepng/lodepng.cpp"
  $7 = (($6) - ($5))|0; //@line 3789 "libraries/lodepng/lodepng.cpp"
  $8 = (($7>>>0) / ($3>>>0))&-1; //@line 3789 "libraries/lodepng/lodepng.cpp"
  $9 = (($passw) + ($i$03<<2)|0); //@line 3789 "libraries/lodepng/lodepng.cpp"
  HEAP32[$9>>2] = $8; //@line 3789 "libraries/lodepng/lodepng.cpp"
  $10 = (34940 + ($i$03<<2)|0); //@line 3790 "libraries/lodepng/lodepng.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 3790 "libraries/lodepng/lodepng.cpp"
  $12 = (34968 + ($i$03<<2)|0); //@line 3790 "libraries/lodepng/lodepng.cpp"
  $13 = HEAP32[$12>>2]|0; //@line 3790 "libraries/lodepng/lodepng.cpp"
  $14 = (($1) + ($11))|0; //@line 3790 "libraries/lodepng/lodepng.cpp"
  $15 = (($14) - ($13))|0; //@line 3790 "libraries/lodepng/lodepng.cpp"
  $16 = (($15>>>0) / ($11>>>0))&-1; //@line 3790 "libraries/lodepng/lodepng.cpp"
  $17 = (($passh) + ($i$03<<2)|0); //@line 3790 "libraries/lodepng/lodepng.cpp"
  HEAP32[$17>>2] = $16; //@line 3790 "libraries/lodepng/lodepng.cpp"
  $18 = HEAP32[$9>>2]|0; //@line 3791 "libraries/lodepng/lodepng.cpp"
  $19 = ($18|0)==(0); //@line 3791 "libraries/lodepng/lodepng.cpp"
  $$ = $19 ? 0 : $16; //@line 3791 "libraries/lodepng/lodepng.cpp"
  HEAP32[$17>>2] = $$; //@line 3791 "libraries/lodepng/lodepng.cpp"
  $20 = ($$|0)==(0); //@line 3792 "libraries/lodepng/lodepng.cpp"
  if ($20) {
   HEAP32[$9>>2] = 0; //@line 3792 "libraries/lodepng/lodepng.cpp"
  }
  $21 = (($i$03) + 1)|0; //@line 3787 "libraries/lodepng/lodepng.cpp"
  $22 = ($21|0)==(7); //@line 3787 "libraries/lodepng/lodepng.cpp"
  if ($22) {
   break;
  } else {
   $i$03 = $21;
  }
 }
 HEAP32[$passstart>>2] = 0; //@line 3795 "libraries/lodepng/lodepng.cpp"
 HEAP32[$padded_passstart>>2] = 0; //@line 3795 "libraries/lodepng/lodepng.cpp"
 HEAP32[$filter_passstart>>2] = 0; //@line 3795 "libraries/lodepng/lodepng.cpp"
 $35 = 0;$i$12 = 0;
 while(1) {
  $23 = (($passw) + ($i$12<<2)|0); //@line 3800 "libraries/lodepng/lodepng.cpp"
  $24 = HEAP32[$23>>2]|0; //@line 3800 "libraries/lodepng/lodepng.cpp"
  $25 = ($24|0)==(0); //@line 3800 "libraries/lodepng/lodepng.cpp"
  $$pre5 = (($passh) + ($i$12<<2)|0); //@line 3802 "libraries/lodepng/lodepng.cpp"
  if ($25) {
   $34 = 0;
  } else {
   $26 = HEAP32[$$pre5>>2]|0; //@line 3800 "libraries/lodepng/lodepng.cpp"
   $27 = ($26|0)==(0); //@line 3800 "libraries/lodepng/lodepng.cpp"
   if ($27) {
    $34 = 0;
   } else {
    $28 = Math_imul($24, $bpp)|0; //@line 3800 "libraries/lodepng/lodepng.cpp"
    $29 = (($28) + 7)|0; //@line 3800 "libraries/lodepng/lodepng.cpp"
    $30 = $29 >>> 3; //@line 3800 "libraries/lodepng/lodepng.cpp"
    $31 = (($30) + 1)|0; //@line 3800 "libraries/lodepng/lodepng.cpp"
    $32 = Math_imul($26, $31)|0; //@line 3800 "libraries/lodepng/lodepng.cpp"
    $34 = $32;
   }
  }
  $33 = (($34) + ($35))|0; //@line 3800 "libraries/lodepng/lodepng.cpp"
  $36 = (($i$12) + 1)|0; //@line 3799 "libraries/lodepng/lodepng.cpp"
  $37 = (($filter_passstart) + ($36<<2)|0); //@line 3799 "libraries/lodepng/lodepng.cpp"
  HEAP32[$37>>2] = $33; //@line 3799 "libraries/lodepng/lodepng.cpp"
  $38 = (($padded_passstart) + ($i$12<<2)|0); //@line 3802 "libraries/lodepng/lodepng.cpp"
  $39 = HEAP32[$38>>2]|0; //@line 3802 "libraries/lodepng/lodepng.cpp"
  $40 = HEAP32[$$pre5>>2]|0; //@line 3802 "libraries/lodepng/lodepng.cpp"
  $41 = HEAP32[$23>>2]|0; //@line 3802 "libraries/lodepng/lodepng.cpp"
  $42 = Math_imul($41, $bpp)|0; //@line 3802 "libraries/lodepng/lodepng.cpp"
  $43 = (($42) + 7)|0; //@line 3802 "libraries/lodepng/lodepng.cpp"
  $44 = $43 >>> 3; //@line 3802 "libraries/lodepng/lodepng.cpp"
  $45 = Math_imul($44, $40)|0; //@line 3802 "libraries/lodepng/lodepng.cpp"
  $46 = (($45) + ($39))|0; //@line 3802 "libraries/lodepng/lodepng.cpp"
  $47 = (($padded_passstart) + ($36<<2)|0); //@line 3802 "libraries/lodepng/lodepng.cpp"
  HEAP32[$47>>2] = $46; //@line 3802 "libraries/lodepng/lodepng.cpp"
  $48 = (($passstart) + ($i$12<<2)|0); //@line 3804 "libraries/lodepng/lodepng.cpp"
  $49 = HEAP32[$48>>2]|0; //@line 3804 "libraries/lodepng/lodepng.cpp"
  $50 = HEAP32[$$pre5>>2]|0; //@line 3804 "libraries/lodepng/lodepng.cpp"
  $51 = HEAP32[$23>>2]|0; //@line 3804 "libraries/lodepng/lodepng.cpp"
  $52 = Math_imul($50, $bpp)|0; //@line 3804 "libraries/lodepng/lodepng.cpp"
  $53 = Math_imul($52, $51)|0; //@line 3804 "libraries/lodepng/lodepng.cpp"
  $54 = (($53) + 7)|0; //@line 3804 "libraries/lodepng/lodepng.cpp"
  $55 = $54 >>> 3; //@line 3804 "libraries/lodepng/lodepng.cpp"
  $56 = (($55) + ($49))|0; //@line 3804 "libraries/lodepng/lodepng.cpp"
  $57 = (($passstart) + ($36<<2)|0); //@line 3804 "libraries/lodepng/lodepng.cpp"
  HEAP32[$57>>2] = $56; //@line 3804 "libraries/lodepng/lodepng.cpp"
  $58 = ($36|0)==(7); //@line 3796 "libraries/lodepng/lodepng.cpp"
  if ($58) {
   break;
  }
  $$pre = HEAP32[$37>>2]|0; //@line 3799 "libraries/lodepng/lodepng.cpp"
  $35 = $$pre;$i$12 = $36;
 }
 return; //@line 3806 "libraries/lodepng/lodepng.cpp"
}
function __Z14lodepng_encodePPhPjPKhjjP12LodePNGState($out,$outsize,$image,$w,$h,$state) {
 $out = $out|0;
 $outsize = $outsize|0;
 $image = $image|0;
 $w = $w|0;
 $h = $h|0;
 $state = $state|0;
 var $$$i = 0, $$$i82 = 0, $$0 = 0, $$0$i$i = 0, $$0$i$i$i = 0, $$0$i$i115 = 0, $$0$i$i99 = 0, $$0$i$ph = 0, $$0$i4$ph = 0, $$12$$i = 0, $$12$16$i = 0, $$12$i = 0, $$2$i = 0, $$2$i31 = 0, $$2$i90 = 0, $$32$$i = 0, $$32$36$i = 0, $$32$i = 0, $$lcssa$i = 0, $$lcssa185 = 0;
 var $$lcssa186 = 0, $$lcssa47$i = 0, $$lcssa48$i = 0, $$lcssa49$i = 0, $$mask$i = 0, $$mask$i40 = 0, $$mask$i46 = 0, $$not = 0, $$not$i = 0, $$pr = 0, $$pr132 = 0, $$pr133 = 0, $$pre = 0, $$pre$i = 0, $$pre$i12$i = 0, $$pre$i17$i = 0, $$pre$i2$i = 0, $$pre$i22$i = 0, $$pre$i27$i = 0, $$pre$i32$i = 0;
 var $$pre$i7$i = 0, $$pre10$i = 0, $$pre12$i = 0, $$pre155 = 0, $$pre156 = 0, $$pre157 = 0, $$pre158 = 0, $$pre159 = 0, $$sink$i = 0, $$sink49$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0;
 var $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0;
 var $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0;
 var $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0;
 var $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0;
 var $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0;
 var $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0;
 var $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0;
 var $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0;
 var $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0;
 var $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0;
 var $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0;
 var $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0;
 var $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0;
 var $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0;
 var $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0;
 var $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0;
 var $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0;
 var $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0;
 var $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0;
 var $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0;
 var $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0;
 var $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0;
 var $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0;
 var $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0;
 var $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0;
 var $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0;
 var $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0;
 var $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0;
 var $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0;
 var $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0;
 var $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0;
 var $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0;
 var $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0;
 var $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0;
 var $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0;
 var $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0;
 var $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0;
 var $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0;
 var $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0;
 var $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0;
 var $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0;
 var $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0;
 var $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0;
 var $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0;
 var $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0;
 var $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0;
 var $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $PLTE$sroa$0$0$lcssa$i = 0, $PLTE$sroa$0$0$lcssa$i59 = 0, $PLTE$sroa$0$03$i = 0, $PLTE$sroa$0$03$i49 = 0, $PLTE$sroa$0$1$i = 0, $PLTE$sroa$0$1$i53 = 0, $PLTE$sroa$0$3$i = 0, $PLTE$sroa$0$3$i57 = 0, $PLTE$sroa$14$01$i = 0, $PLTE$sroa$14$01$i51 = 0, $PLTE$sroa$14$1$i = 0;
 var $PLTE$sroa$14$1$i52 = 0, $PLTE$sroa$14$3$i = 0, $PLTE$sroa$14$3$i55 = 0, $PLTE$sroa$9$0$lcssa$i = 0, $PLTE$sroa$9$0$lcssa$i60 = 0, $PLTE$sroa$9$02$i = 0, $PLTE$sroa$9$02$i50 = 0, $PLTE$sroa$9$2$i = 0, $PLTE$sroa$9$2$i56 = 0, $b$05$i = 0, $b$1$i = 0, $bKGD$sroa$0$10$i = 0, $bKGD$sroa$0$11$i = 0, $bKGD$sroa$0$12$i = 0, $bKGD$sroa$0$13$i = 0, $bKGD$sroa$0$14$i = 0, $bKGD$sroa$0$144$i = 0, $bKGD$sroa$0$18$i = 0, $bKGD$sroa$0$548$i = 0, $bKGD$sroa$0$7$i = 0;
 var $bKGD$sroa$0$8$i = 0, $bKGD$sroa$0$9$i = 0, $bKGD$sroa$33$043$i = 0, $bKGD$sroa$33$247$i = 0, $bKGD$sroa$33$3$i = 0, $bKGD$sroa$33$4$i = 0, $bKGD$sroa$33$5$i = 0, $bKGD$sroa$33$6$i = 0, $bKGD$sroa$33$9$i = 0, $bKGD$sroa$54$10$i = 0, $bKGD$sroa$54$11$i = 0, $bKGD$sroa$54$5$i = 0, $bKGD$sroa$54$6$i = 0, $bKGD$sroa$54$7$i = 0, $bKGD$sroa$54$8$i = 0, $bKGD$sroa$54$9$i = 0, $compressed$i = 0, $compressed_data$i = 0, $data = 0, $data$sroa$0$023$i = 0;
 var $data$sroa$0$084$i = 0, $data$sroa$0$1$i = 0, $data$sroa$0$1$i103 = 0, $data$sroa$0$10$i = 0, $data$sroa$0$10$i122 = 0, $data$sroa$0$11$i = 0, $data$sroa$0$12$i = 0, $data$sroa$0$14$lcssa$i = 0, $data$sroa$0$14$ph$i = 0, $data$sroa$0$1464$i = 0, $data$sroa$0$15$i = 0, $data$sroa$0$16$i = 0, $data$sroa$0$17$i = 0, $data$sroa$0$18$i = 0, $data$sroa$0$1952$i = 0, $data$sroa$0$2$i = 0, $data$sroa$0$2$i$lcssa = 0, $data$sroa$0$2$i105 = 0, $data$sroa$0$2$i105$lcssa = 0, $data$sroa$0$20$i = 0;
 var $data$sroa$0$21$i = 0, $data$sroa$0$22$i = 0, $data$sroa$0$2357$i = 0, $data$sroa$0$24$i = 0, $data$sroa$0$25$i = 0, $data$sroa$0$26$i = 0, $data$sroa$0$27$i = 0, $data$sroa$0$3$i = 0, $data$sroa$0$3$i107 = 0, $data$sroa$0$4$i = 0, $data$sroa$0$4$i109 = 0, $data$sroa$0$4$i83 = 0, $data$sroa$0$5$i = 0, $data$sroa$0$5$i111 = 0, $data$sroa$0$5$i85 = 0, $data$sroa$0$6$i = 0, $data$sroa$0$6$i113 = 0, $data$sroa$0$7$i = 0, $data$sroa$0$7$lcssa$i = 0, $data$sroa$0$718$i = 0;
 var $data$sroa$0$8$i = 0, $data$sroa$0$9$i = 0, $data$sroa$0$9$lcssa$i = 0, $data$sroa$0$9$ph$i = 0, $data$sroa$0$974$i = 0, $data$sroa$18$022$i = 0, $data$sroa$18$1$i = 0, $data$sroa$18$1$i$lcssa = 0, $data$sroa$18$2$i = 0, $data$sroa$18$3$i = 0, $data$sroa$18$4$lcssa$i = 0, $data$sroa$18$417$i = 0, $data$sroa$18$5$i = 0, $data$sroa$19$2$i = 0, $data$sroa$29$021$i = 0, $data$sroa$29$1$i = 0, $data$sroa$29$2$i = 0, $data$sroa$29$2$i$lcssa = 0, $data$sroa$29$3$i = 0, $data$sroa$29$4$i = 0;
 var $data$sroa$29$5$i = 0, $data$sroa$29$6$i = 0, $data$sroa$29$716$i = 0, $data$sroa$29$8$i = 0, $data$sroa$29$9$i = 0, $data$sroa$36$083$i = 0, $data$sroa$36$1$i = 0, $data$sroa$36$1$i$lcssa = 0, $data$sroa$36$10$i = 0, $data$sroa$36$1151$i = 0, $data$sroa$36$12$i = 0, $data$sroa$36$13$i = 0, $data$sroa$36$1456$i = 0, $data$sroa$36$15$i = 0, $data$sroa$36$16$i = 0, $data$sroa$36$2$i = 0, $data$sroa$36$3$i = 0, $data$sroa$36$5$lcssa$i = 0, $data$sroa$36$5$ph$i = 0, $data$sroa$36$573$i = 0;
 var $data$sroa$36$6$i = 0, $data$sroa$36$8$lcssa$i = 0, $data$sroa$36$8$ph$i = 0, $data$sroa$36$863$i = 0, $data$sroa$36$9$i = 0, $data$sroa$59$082$i = 0, $data$sroa$59$1$i = 0, $data$sroa$59$10$i = 0, $data$sroa$59$11$i = 0, $data$sroa$59$12$i = 0, $data$sroa$59$14$lcssa$i = 0, $data$sroa$59$14$ph$i = 0, $data$sroa$59$1462$i = 0, $data$sroa$59$15$i = 0, $data$sroa$59$16$i = 0, $data$sroa$59$17$i = 0, $data$sroa$59$18$i = 0, $data$sroa$59$1950$i = 0, $data$sroa$59$2$i = 0, $data$sroa$59$2$i$lcssa = 0;
 var $data$sroa$59$20$i = 0, $data$sroa$59$21$i = 0, $data$sroa$59$2255$i = 0, $data$sroa$59$23$i = 0, $data$sroa$59$24$i = 0, $data$sroa$59$3$i = 0, $data$sroa$59$4$i = 0, $data$sroa$59$5$i = 0, $data$sroa$59$6$i = 0, $data$sroa$59$7$i = 0, $data$sroa$59$9$lcssa$i = 0, $data$sroa$59$9$ph$i = 0, $data$sroa$59$972$i = 0, $datasize = 0, $error$0$i = 0, $g$04$i = 0, $g$1$i = 0, $header$sroa$0$10$i = 0, $header$sroa$0$11$i = 0, $header$sroa$0$12$i = 0;
 var $header$sroa$0$13$i = 0, $header$sroa$0$4$i = 0, $header$sroa$0$5$i = 0, $header$sroa$0$6$i = 0, $header$sroa$0$7$i = 0, $header$sroa$0$8$i = 0, $header$sroa$0$9$i = 0, $header$sroa$31$2$i = 0, $header$sroa$31$3$i = 0, $header$sroa$31$4$i = 0, $header$sroa$31$5$i = 0, $header$sroa$31$6$i = 0, $header$sroa$50$10$i = 0, $header$sroa$50$11$i = 0, $header$sroa$50$4$i = 0, $header$sroa$50$5$i = 0, $header$sroa$50$6$i = 0, $header$sroa$50$7$i = 0, $header$sroa$50$8$i = 0, $header$sroa$50$9$i = 0;
 var $i$01$i = 0, $i$01$i35 = 0, $i$01$i64 = 0, $i$01$i8 = 0, $i$01$i94 = 0, $i$0150 = 0, $i$024$i = 0, $i$024$i$lcssa = 0, $i$04$i = 0, $i$04$i48 = 0, $i$085$i = 0, $i$085$i$lcssa = 0, $i$1$i = 0, $i$1148 = 0, $i$119$i = 0, $i$175$i = 0, $i$2147 = 0, $i$265$i = 0, $i$353$i = 0, $i$458$i = 0;
 var $inchunk$02$i = 0, $inchunk$02$i30 = 0, $inchunk$02$i89 = 0, $info = 0, $key$02$i = 0, $key$1$i = 0, $key$1$i$lcssa = 0, $or$cond = 0, $or$cond$i = 0, $or$cond$i11 = 0, $or$cond$i34 = 0, $or$cond$i93 = 0, $or$cond135 = 0, $or$cond167 = 0, $outv = 0, $r$03$i = 0, $r$1$i = 0, $scevgep4$i$i = 0, $strlenfirst = 0, $strlenfirst2 = 0;
 var $zlibdata$i = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $compressed$i = sp + 208|0;
 $zlibdata$i = sp + 196|0;
 $compressed_data$i = sp + 184|0;
 $info = sp + 24|0;
 $outv = sp + 8|0;
 $data = sp + 4|0;
 $datasize = sp;
 HEAP32[$data>>2] = 0; //@line 5551 "libraries/lodepng/lodepng.cpp"
 HEAP32[$datasize>>2] = 0; //@line 5552 "libraries/lodepng/lodepng.cpp"
 HEAP32[$out>>2] = 0; //@line 5555 "libraries/lodepng/lodepng.cpp"
 HEAP32[$outsize>>2] = 0; //@line 5556 "libraries/lodepng/lodepng.cpp"
 $0 = ((($state)) + 292|0); //@line 5557 "libraries/lodepng/lodepng.cpp"
 HEAP32[$0>>2] = 0; //@line 5557 "libraries/lodepng/lodepng.cpp"
 $1 = ((($info)) + 28|0); //@line 2556 "libraries/lodepng/lodepng.cpp"
 $2 = ((($info)) + 12|0); //@line 2558 "libraries/lodepng/lodepng.cpp"
 ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0;HEAP32[$1+12>>2]=0|0; //@line 2557 "libraries/lodepng/lodepng.cpp"
 HEAP32[$2>>2] = 6; //@line 2558 "libraries/lodepng/lodepng.cpp"
 $3 = ((($info)) + 16|0); //@line 2559 "libraries/lodepng/lodepng.cpp"
 HEAP32[$3>>2] = 8; //@line 2559 "libraries/lodepng/lodepng.cpp"
 $4 = ((($info)) + 20|0); //@line 2560 "libraries/lodepng/lodepng.cpp"
 HEAP32[$4>>2] = 0; //@line 2560 "libraries/lodepng/lodepng.cpp"
 $5 = ((($info)) + 24|0); //@line 2561 "libraries/lodepng/lodepng.cpp"
 HEAP32[$5>>2] = 0; //@line 2561 "libraries/lodepng/lodepng.cpp"
 $6 = ((($info)) + 8|0); //@line 2883 "libraries/lodepng/lodepng.cpp"
 HEAP32[$6>>2] = 0; //@line 2883 "libraries/lodepng/lodepng.cpp"
 HEAP32[$info>>2] = 0; //@line 2884 "libraries/lodepng/lodepng.cpp"
 $7 = ((($info)) + 4|0); //@line 2885 "libraries/lodepng/lodepng.cpp"
 HEAP32[$7>>2] = 0; //@line 2885 "libraries/lodepng/lodepng.cpp"
 $8 = ((($info)) + 44|0); //@line 2887 "libraries/lodepng/lodepng.cpp"
 $9 = ((($info)) + 120|0); //@line 2894 "libraries/lodepng/lodepng.cpp"
 HEAP32[$9>>2] = 0; //@line 2894 "libraries/lodepng/lodepng.cpp"
 $scevgep4$i$i = ((($info)) + 136|0); //@line 2700 "libraries/lodepng/lodepng.cpp"
 ;HEAP32[$scevgep4$i$i>>2]=0|0;HEAP32[$scevgep4$i$i+4>>2]=0|0;HEAP32[$scevgep4$i$i+8>>2]=0|0;HEAP32[$scevgep4$i$i+12>>2]=0|0;HEAP32[$scevgep4$i$i+16>>2]=0|0;HEAP32[$scevgep4$i$i+20>>2]=0|0; //@line 2701 "libraries/lodepng/lodepng.cpp"
 dest=$8; stop=dest+52|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 2887 "libraries/lodepng/lodepng.cpp"
 $10 = ((($state)) + 132|0); //@line 5560 "libraries/lodepng/lodepng.cpp"
 (__Z17lodepng_info_copyP11LodePNGInfoPKS_($info,$10)|0); //@line 5560 "libraries/lodepng/lodepng.cpp"
 $11 = ((($info)) + 12|0); //@line 5562 "libraries/lodepng/lodepng.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 5562 "libraries/lodepng/lodepng.cpp"
 $13 = ($12|0)==(3); //@line 5562 "libraries/lodepng/lodepng.cpp"
 if ($13) {
  label = 3;
 } else {
  $14 = ((($state)) + 88|0); //@line 5562 "libraries/lodepng/lodepng.cpp"
  $15 = HEAP32[$14>>2]|0; //@line 5562 "libraries/lodepng/lodepng.cpp"
  $16 = ($15|0)==(0); //@line 5562 "libraries/lodepng/lodepng.cpp"
  if (!($16)) {
   label = 3;
  }
 }
 if ((label|0) == 3) {
  $17 = HEAP32[$5>>2]|0; //@line 5563 "libraries/lodepng/lodepng.cpp"
  $18 = (($17) + -1)|0; //@line 5563 "libraries/lodepng/lodepng.cpp"
  $19 = ($18>>>0)>(255); //@line 5563 "libraries/lodepng/lodepng.cpp"
  if ($19) {
   HEAP32[$0>>2] = 68; //@line 5565 "libraries/lodepng/lodepng.cpp"
   $$0 = 68;
   STACKTOP = sp;return ($$0|0); //@line 5736 "libraries/lodepng/lodepng.cpp"
  }
 }
 $20 = ((($state)) + 36|0); //@line 5569 "libraries/lodepng/lodepng.cpp"
 $21 = ((($state)) + 72|0); //@line 5569 "libraries/lodepng/lodepng.cpp"
 $22 = HEAP32[$21>>2]|0; //@line 5569 "libraries/lodepng/lodepng.cpp"
 $23 = ($22|0)==(0); //@line 5569 "libraries/lodepng/lodepng.cpp"
 if ($23) {
  $$pr = HEAP32[$0>>2]|0; //@line 5573 "libraries/lodepng/lodepng.cpp"
  $27 = $$pr;
 } else {
  $24 = ((($state)) + 100|0); //@line 5571 "libraries/lodepng/lodepng.cpp"
  $25 = (__Z25lodepng_auto_choose_colorP16LodePNGColorModePKhjjPKS_($11,$image,$w,$h,$24)|0); //@line 5571 "libraries/lodepng/lodepng.cpp"
  HEAP32[$0>>2] = $25; //@line 5571 "libraries/lodepng/lodepng.cpp"
  $27 = $25;
 }
 $26 = ($27|0)==(0); //@line 5573 "libraries/lodepng/lodepng.cpp"
 if (!($26)) {
  $$0 = $27;
  STACKTOP = sp;return ($$0|0); //@line 5736 "libraries/lodepng/lodepng.cpp"
 }
 $28 = HEAP32[$20>>2]|0; //@line 5575 "libraries/lodepng/lodepng.cpp"
 $29 = ($28>>>0)>(2); //@line 5575 "libraries/lodepng/lodepng.cpp"
 if ($29) {
  HEAP32[$0>>2] = 61; //@line 5577 "libraries/lodepng/lodepng.cpp"
  $$0 = 61;
  STACKTOP = sp;return ($$0|0); //@line 5736 "libraries/lodepng/lodepng.cpp"
 }
 $30 = ((($state)) + 140|0); //@line 5579 "libraries/lodepng/lodepng.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 5579 "libraries/lodepng/lodepng.cpp"
 $32 = ($31>>>0)>(1); //@line 5579 "libraries/lodepng/lodepng.cpp"
 if ($32) {
  HEAP32[$0>>2] = 71; //@line 5581 "libraries/lodepng/lodepng.cpp"
  $$0 = 71;
  STACKTOP = sp;return ($$0|0); //@line 5736 "libraries/lodepng/lodepng.cpp"
 }
 $33 = HEAP32[$11>>2]|0; //@line 2634 "libraries/lodepng/lodepng.cpp"
 $34 = HEAP32[$3>>2]|0; //@line 2634 "libraries/lodepng/lodepng.cpp"
 switch ($33|0) {
 case 0:  {
  switch ($34|0) {
  case 1: case 2: case 4: case 8: case 16:  {
   break;
  }
  default: {
   $$0$i$ph = 37;
   label = 19;
  }
  }
  break;
 }
 case 2:  {
  switch ($34|0) {
  case 8: case 16:  {
   break;
  }
  default: {
   $$0$i$ph = 37;
   label = 19;
  }
  }
  break;
 }
 case 3:  {
  switch ($34|0) {
  case 1: case 2: case 4: case 8:  {
   break;
  }
  default: {
   $$0$i$ph = 37;
   label = 19;
  }
  }
  break;
 }
 case 4:  {
  switch ($34|0) {
  case 8: case 16:  {
   break;
  }
  default: {
   $$0$i$ph = 37;
   label = 19;
  }
  }
  break;
 }
 case 6:  {
  switch ($34|0) {
  case 8: case 16:  {
   break;
  }
  default: {
   $$0$i$ph = 37;
   label = 19;
  }
  }
  break;
 }
 default: {
  $$0$i$ph = 31;
  label = 19;
 }
 }
 if ((label|0) == 19) {
  HEAP32[$0>>2] = $$0$i$ph; //@line 5584 "libraries/lodepng/lodepng.cpp"
  $$0 = $$0$i$ph;
  STACKTOP = sp;return ($$0|0); //@line 5736 "libraries/lodepng/lodepng.cpp"
 }
 HEAP32[$0>>2] = 0; //@line 5584 "libraries/lodepng/lodepng.cpp"
 $35 = ((($state)) + 100|0); //@line 5586 "libraries/lodepng/lodepng.cpp"
 $36 = HEAP32[$35>>2]|0; //@line 2586 "libraries/lodepng/lodepng.cpp"
 $37 = ((($state)) + 104|0); //@line 5586 "libraries/lodepng/lodepng.cpp"
 $38 = HEAP32[$37>>2]|0; //@line 2587 "libraries/lodepng/lodepng.cpp"
 switch ($36|0) {
 case 0:  {
  switch ($38|0) {
  case 1: case 2: case 4: case 8: case 16:  {
   break;
  }
  default: {
   $$0$i4$ph = 37;
   label = 26;
  }
  }
  break;
 }
 case 2:  {
  switch ($38|0) {
  case 8: case 16:  {
   break;
  }
  default: {
   $$0$i4$ph = 37;
   label = 26;
  }
  }
  break;
 }
 case 3:  {
  switch ($38|0) {
  case 1: case 2: case 4: case 8:  {
   break;
  }
  default: {
   $$0$i4$ph = 37;
   label = 26;
  }
  }
  break;
 }
 case 4:  {
  switch ($38|0) {
  case 8: case 16:  {
   break;
  }
  default: {
   $$0$i4$ph = 37;
   label = 26;
  }
  }
  break;
 }
 case 6:  {
  switch ($38|0) {
  case 8: case 16:  {
   break;
  }
  default: {
   $$0$i4$ph = 37;
   label = 26;
  }
  }
  break;
 }
 default: {
  $$0$i4$ph = 31;
  label = 26;
 }
 }
 if ((label|0) == 26) {
  HEAP32[$0>>2] = $$0$i4$ph; //@line 5586 "libraries/lodepng/lodepng.cpp"
  $$0 = $$0$i4$ph;
  STACKTOP = sp;return ($$0|0); //@line 5736 "libraries/lodepng/lodepng.cpp"
 }
 HEAP32[$0>>2] = 0; //@line 5586 "libraries/lodepng/lodepng.cpp"
 $39 = ($36|0)==($33|0); //@line 2586 "libraries/lodepng/lodepng.cpp"
 $40 = ($38|0)==($34|0); //@line 2587 "libraries/lodepng/lodepng.cpp"
 $or$cond167 = $39 & $40; //@line 2586 "libraries/lodepng/lodepng.cpp"
 L43: do {
  if ($or$cond167) {
   $41 = ((($state)) + 116|0); //@line 2588 "libraries/lodepng/lodepng.cpp"
   $42 = HEAP32[$41>>2]|0; //@line 2588 "libraries/lodepng/lodepng.cpp"
   $43 = HEAP32[$1>>2]|0; //@line 2588 "libraries/lodepng/lodepng.cpp"
   $44 = ($42|0)==($43|0); //@line 2588 "libraries/lodepng/lodepng.cpp"
   if ($44) {
    $45 = ($42|0)==(0); //@line 2589 "libraries/lodepng/lodepng.cpp"
    if (!($45)) {
     $46 = ((($state)) + 120|0); //@line 2591 "libraries/lodepng/lodepng.cpp"
     $47 = HEAP32[$46>>2]|0; //@line 2591 "libraries/lodepng/lodepng.cpp"
     $48 = ((($info)) + 32|0); //@line 2591 "libraries/lodepng/lodepng.cpp"
     $49 = HEAP32[$48>>2]|0; //@line 2591 "libraries/lodepng/lodepng.cpp"
     $50 = ($47|0)==($49|0); //@line 2591 "libraries/lodepng/lodepng.cpp"
     if (!($50)) {
      label = 38;
      break;
     }
     $51 = ((($state)) + 124|0); //@line 2592 "libraries/lodepng/lodepng.cpp"
     $52 = HEAP32[$51>>2]|0; //@line 2592 "libraries/lodepng/lodepng.cpp"
     $53 = ((($info)) + 36|0); //@line 2592 "libraries/lodepng/lodepng.cpp"
     $54 = HEAP32[$53>>2]|0; //@line 2592 "libraries/lodepng/lodepng.cpp"
     $55 = ($52|0)==($54|0); //@line 2592 "libraries/lodepng/lodepng.cpp"
     if (!($55)) {
      label = 38;
      break;
     }
     $56 = ((($state)) + 128|0); //@line 2593 "libraries/lodepng/lodepng.cpp"
     $57 = HEAP32[$56>>2]|0; //@line 2593 "libraries/lodepng/lodepng.cpp"
     $58 = ((($info)) + 40|0); //@line 2593 "libraries/lodepng/lodepng.cpp"
     $59 = HEAP32[$58>>2]|0; //@line 2593 "libraries/lodepng/lodepng.cpp"
     $60 = ($57|0)==($59|0); //@line 2593 "libraries/lodepng/lodepng.cpp"
     if (!($60)) {
      label = 38;
      break;
     }
    }
    $61 = ((($state)) + 112|0); //@line 2595 "libraries/lodepng/lodepng.cpp"
    $62 = HEAP32[$61>>2]|0; //@line 2596 "libraries/lodepng/lodepng.cpp"
    $63 = HEAP32[$5>>2]|0; //@line 2595 "libraries/lodepng/lodepng.cpp"
    $64 = ($62|0)==($63|0); //@line 2595 "libraries/lodepng/lodepng.cpp"
    if ($64) {
     $$mask$i = $62 & 1073741823; //@line 2596 "libraries/lodepng/lodepng.cpp"
     $65 = ($$mask$i|0)==(0); //@line 2596 "libraries/lodepng/lodepng.cpp"
     if (!($65)) {
      $66 = ((($state)) + 108|0); //@line 2598 "libraries/lodepng/lodepng.cpp"
      $67 = HEAP32[$66>>2]|0; //@line 2598 "libraries/lodepng/lodepng.cpp"
      $68 = HEAP32[$4>>2]|0; //@line 2598 "libraries/lodepng/lodepng.cpp"
      $69 = $62 << 2; //@line 2596 "libraries/lodepng/lodepng.cpp"
      $i$01$i = 0;
      while(1) {
       $72 = (($67) + ($i$01$i)|0); //@line 2598 "libraries/lodepng/lodepng.cpp"
       $73 = HEAP8[$72>>0]|0; //@line 2598 "libraries/lodepng/lodepng.cpp"
       $74 = (($68) + ($i$01$i)|0); //@line 2598 "libraries/lodepng/lodepng.cpp"
       $75 = HEAP8[$74>>0]|0; //@line 2598 "libraries/lodepng/lodepng.cpp"
       $76 = ($73<<24>>24)==($75<<24>>24); //@line 2598 "libraries/lodepng/lodepng.cpp"
       $71 = (($i$01$i) + 1)|0; //@line 2596 "libraries/lodepng/lodepng.cpp"
       if (!($76)) {
        label = 38;
        break L43;
       }
       $70 = ($71|0)==($69|0); //@line 2596 "libraries/lodepng/lodepng.cpp"
       if ($70) {
        break;
       } else {
        $i$01$i = $71;
       }
      }
     }
     __ZL19preProcessScanlinesPPhPjPKhjjPK11LodePNGInfoPK22LodePNGEncoderSettings($data,$datasize,$image,$w,$h,$info,$20); //@line 5603 "libraries/lodepng/lodepng.cpp"
    } else {
     label = 38;
    }
   } else {
    label = 38;
   }
  } else {
   label = 38;
  }
 } while(0);
 if ((label|0) == 38) {
  switch ($33|0) {
  case 3: case 0:  {
   $$0$i$i$i = 1;
   break;
  }
  case 2:  {
   $$0$i$i$i = 3;
   break;
  }
  case 4:  {
   $$0$i$i$i = 2;
   break;
  }
  case 6:  {
   $$0$i$i$i = 4;
   break;
  }
  default: {
   $$0$i$i$i = 0;
  }
  }
  $77 = Math_imul($h, $w)|0; //@line 2549 "libraries/lodepng/lodepng.cpp"
  $78 = Math_imul($77, $34)|0; //@line 5592 "libraries/lodepng/lodepng.cpp"
  $79 = Math_imul($78, $$0$i$i$i)|0; //@line 5592 "libraries/lodepng/lodepng.cpp"
  $80 = (($79) + 7)|0; //@line 5592 "libraries/lodepng/lodepng.cpp"
  $81 = $80 >>> 3; //@line 5592 "libraries/lodepng/lodepng.cpp"
  $82 = (_malloc($81)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
  $83 = ($82|0)==(0|0); //@line 5595 "libraries/lodepng/lodepng.cpp"
  $84 = ($81|0)!=(0); //@line 5595 "libraries/lodepng/lodepng.cpp"
  $or$cond = $83 & $84; //@line 5595 "libraries/lodepng/lodepng.cpp"
  if ($or$cond) {
   HEAP32[$0>>2] = 83; //@line 5595 "libraries/lodepng/lodepng.cpp"
  } else {
   (__Z15lodepng_convertPhPKhP16LodePNGColorModePKS2_jj($82,$image,$11,$35,$w,$h)|0); //@line 5598 "libraries/lodepng/lodepng.cpp"
   HEAP32[$0>>2] = 0; //@line 5598 "libraries/lodepng/lodepng.cpp"
   __ZL19preProcessScanlinesPPhPjPKhjjPK11LodePNGInfoPK22LodePNGEncoderSettings($data,$datasize,$82,$w,$h,$info,$20); //@line 5600 "libraries/lodepng/lodepng.cpp"
  }
  _free($82); //@line 76 "libraries/lodepng/lodepng.cpp"
 }
 HEAP32[$outv>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
 $85 = ((($outv)) + 8|0); //@line 259 "libraries/lodepng/lodepng.cpp"
 HEAP32[$85>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
 $86 = ((($outv)) + 4|0); //@line 259 "libraries/lodepng/lodepng.cpp"
 HEAP32[$86>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
 $87 = HEAP32[$0>>2]|0; //@line 5606 "libraries/lodepng/lodepng.cpp"
 $88 = ($87|0)==(0); //@line 5606 "libraries/lodepng/lodepng.cpp"
 L71: do {
  if ($88) {
   $89 = (_realloc(0,1)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
   $90 = ($89|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
   if ($90) {
    $$pre = HEAP32[$86>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
    $$pre155 = HEAP32[$85>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
    $92 = $$pre;$94 = $$pre155;
   } else {
    HEAP32[$85>>2] = 1; //@line 231 "libraries/lodepng/lodepng.cpp"
    HEAP32[$outv>>2] = $89; //@line 232 "libraries/lodepng/lodepng.cpp"
    HEAP32[$86>>2] = 1; //@line 243 "libraries/lodepng/lodepng.cpp"
    HEAP8[$89>>0] = -119; //@line 289 "libraries/lodepng/lodepng.cpp"
    $92 = 1;$94 = 1;
   }
   $91 = (($92) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $93 = ($94>>>0)<($91>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($93) {
    $95 = $94 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $96 = ($95>>>0)<($91>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $97 = ($91*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $98 = $97 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $99 = $96 ? $91 : $98; //@line 227 "libraries/lodepng/lodepng.cpp"
    $100 = HEAP32[$outv>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
    $101 = (_realloc($100,$99)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $102 = ($101|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if (!($102)) {
     HEAP32[$85>>2] = $99; //@line 231 "libraries/lodepng/lodepng.cpp"
     HEAP32[$outv>>2] = $101; //@line 232 "libraries/lodepng/lodepng.cpp"
     $104 = $101;
     label = 56;
    }
   } else {
    $$pre$i2$i = HEAP32[$outv>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
    $104 = $$pre$i2$i;
    label = 56;
   }
   if ((label|0) == 56) {
    HEAP32[$86>>2] = $91; //@line 243 "libraries/lodepng/lodepng.cpp"
    $103 = (($104) + ($92)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$103>>0] = 80; //@line 289 "libraries/lodepng/lodepng.cpp"
   }
   $105 = HEAP32[$86>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $106 = (($105) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $107 = HEAP32[$85>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
   $108 = ($107>>>0)<($106>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($108) {
    $109 = $107 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $110 = ($109>>>0)<($106>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $111 = ($106*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $112 = $111 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $113 = $110 ? $106 : $112; //@line 227 "libraries/lodepng/lodepng.cpp"
    $114 = HEAP32[$outv>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
    $115 = (_realloc($114,$113)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $116 = ($115|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if (!($116)) {
     HEAP32[$85>>2] = $113; //@line 231 "libraries/lodepng/lodepng.cpp"
     HEAP32[$outv>>2] = $115; //@line 232 "libraries/lodepng/lodepng.cpp"
     $118 = $115;
     label = 61;
    }
   } else {
    $$pre$i7$i = HEAP32[$outv>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
    $118 = $$pre$i7$i;
    label = 61;
   }
   if ((label|0) == 61) {
    HEAP32[$86>>2] = $106; //@line 243 "libraries/lodepng/lodepng.cpp"
    $117 = (($118) + ($105)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$117>>0] = 78; //@line 289 "libraries/lodepng/lodepng.cpp"
   }
   $119 = HEAP32[$86>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $120 = (($119) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $121 = HEAP32[$85>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
   $122 = ($121>>>0)<($120>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($122) {
    $123 = $121 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $124 = ($123>>>0)<($120>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $125 = ($120*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $126 = $125 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $127 = $124 ? $120 : $126; //@line 227 "libraries/lodepng/lodepng.cpp"
    $128 = HEAP32[$outv>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
    $129 = (_realloc($128,$127)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $130 = ($129|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if (!($130)) {
     HEAP32[$85>>2] = $127; //@line 231 "libraries/lodepng/lodepng.cpp"
     HEAP32[$outv>>2] = $129; //@line 232 "libraries/lodepng/lodepng.cpp"
     $132 = $129;
     label = 66;
    }
   } else {
    $$pre$i12$i = HEAP32[$outv>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
    $132 = $$pre$i12$i;
    label = 66;
   }
   if ((label|0) == 66) {
    HEAP32[$86>>2] = $120; //@line 243 "libraries/lodepng/lodepng.cpp"
    $131 = (($132) + ($119)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$131>>0] = 71; //@line 289 "libraries/lodepng/lodepng.cpp"
   }
   $133 = HEAP32[$86>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $134 = (($133) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $135 = HEAP32[$85>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
   $136 = ($135>>>0)<($134>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($136) {
    $137 = $135 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $138 = ($137>>>0)<($134>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $139 = ($134*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $140 = $139 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $141 = $138 ? $134 : $140; //@line 227 "libraries/lodepng/lodepng.cpp"
    $142 = HEAP32[$outv>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
    $143 = (_realloc($142,$141)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $144 = ($143|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if (!($144)) {
     HEAP32[$85>>2] = $141; //@line 231 "libraries/lodepng/lodepng.cpp"
     HEAP32[$outv>>2] = $143; //@line 232 "libraries/lodepng/lodepng.cpp"
     $146 = $143;
     label = 71;
    }
   } else {
    $$pre$i17$i = HEAP32[$outv>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
    $146 = $$pre$i17$i;
    label = 71;
   }
   if ((label|0) == 71) {
    HEAP32[$86>>2] = $134; //@line 243 "libraries/lodepng/lodepng.cpp"
    $145 = (($146) + ($133)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$145>>0] = 13; //@line 289 "libraries/lodepng/lodepng.cpp"
   }
   $147 = HEAP32[$86>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $148 = (($147) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $149 = HEAP32[$85>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
   $150 = ($149>>>0)<($148>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($150) {
    $151 = $149 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $152 = ($151>>>0)<($148>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $153 = ($148*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $154 = $153 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $155 = $152 ? $148 : $154; //@line 227 "libraries/lodepng/lodepng.cpp"
    $156 = HEAP32[$outv>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
    $157 = (_realloc($156,$155)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $158 = ($157|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if (!($158)) {
     HEAP32[$85>>2] = $155; //@line 231 "libraries/lodepng/lodepng.cpp"
     HEAP32[$outv>>2] = $157; //@line 232 "libraries/lodepng/lodepng.cpp"
     $160 = $157;
     label = 76;
    }
   } else {
    $$pre$i22$i = HEAP32[$outv>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
    $160 = $$pre$i22$i;
    label = 76;
   }
   if ((label|0) == 76) {
    HEAP32[$86>>2] = $148; //@line 243 "libraries/lodepng/lodepng.cpp"
    $159 = (($160) + ($147)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$159>>0] = 10; //@line 289 "libraries/lodepng/lodepng.cpp"
   }
   $161 = HEAP32[$86>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $162 = (($161) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $163 = HEAP32[$85>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
   $164 = ($163>>>0)<($162>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($164) {
    $165 = $163 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $166 = ($165>>>0)<($162>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $167 = ($162*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $168 = $167 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $169 = $166 ? $162 : $168; //@line 227 "libraries/lodepng/lodepng.cpp"
    $170 = HEAP32[$outv>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
    $171 = (_realloc($170,$169)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $172 = ($171|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if (!($172)) {
     HEAP32[$85>>2] = $169; //@line 231 "libraries/lodepng/lodepng.cpp"
     HEAP32[$outv>>2] = $171; //@line 232 "libraries/lodepng/lodepng.cpp"
     $174 = $171;
     label = 81;
    }
   } else {
    $$pre$i27$i = HEAP32[$outv>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
    $174 = $$pre$i27$i;
    label = 81;
   }
   if ((label|0) == 81) {
    HEAP32[$86>>2] = $162; //@line 243 "libraries/lodepng/lodepng.cpp"
    $173 = (($174) + ($161)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$173>>0] = 26; //@line 289 "libraries/lodepng/lodepng.cpp"
   }
   $175 = HEAP32[$86>>2]|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $176 = (($175) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $177 = HEAP32[$85>>2]|0; //@line 225 "libraries/lodepng/lodepng.cpp"
   $178 = ($177>>>0)<($176>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($178) {
    $179 = $177 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $180 = ($179>>>0)<($176>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $181 = ($176*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $182 = $181 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $183 = $180 ? $176 : $182; //@line 227 "libraries/lodepng/lodepng.cpp"
    $184 = HEAP32[$outv>>2]|0; //@line 228 "libraries/lodepng/lodepng.cpp"
    $185 = (_realloc($184,$183)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $186 = ($185|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if (!($186)) {
     HEAP32[$85>>2] = $183; //@line 231 "libraries/lodepng/lodepng.cpp"
     HEAP32[$outv>>2] = $185; //@line 232 "libraries/lodepng/lodepng.cpp"
     $188 = $185;
     label = 86;
    }
   } else {
    $$pre$i32$i = HEAP32[$outv>>2]|0; //@line 289 "libraries/lodepng/lodepng.cpp"
    $188 = $$pre$i32$i;
    label = 86;
   }
   if ((label|0) == 86) {
    HEAP32[$86>>2] = $176; //@line 243 "libraries/lodepng/lodepng.cpp"
    $187 = (($188) + ($175)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$187>>0] = 10; //@line 289 "libraries/lodepng/lodepng.cpp"
   }
   $189 = HEAP32[$11>>2]|0; //@line 5614 "libraries/lodepng/lodepng.cpp"
   $190 = HEAP32[$3>>2]|0; //@line 5614 "libraries/lodepng/lodepng.cpp"
   $191 = HEAP32[$6>>2]|0; //@line 5614 "libraries/lodepng/lodepng.cpp"
   $192 = (_realloc(0,4)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
   $193 = ($192|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
   $$32$i = $193 ? 0 : 4; //@line 229 "libraries/lodepng/lodepng.cpp"
   $194 = $193 ? -4 : 0; //@line 361 "libraries/lodepng/lodepng.cpp"
   $195 = (($192) + ($194)|0); //@line 361 "libraries/lodepng/lodepng.cpp"
   $196 = $w >>> 24; //@line 350 "libraries/lodepng/lodepng.cpp"
   $197 = $196&255; //@line 350 "libraries/lodepng/lodepng.cpp"
   HEAP8[$195>>0] = $197; //@line 350 "libraries/lodepng/lodepng.cpp"
   $198 = $w >>> 16; //@line 351 "libraries/lodepng/lodepng.cpp"
   $199 = $198&255; //@line 351 "libraries/lodepng/lodepng.cpp"
   $200 = ((($195)) + 1|0); //@line 351 "libraries/lodepng/lodepng.cpp"
   HEAP8[$200>>0] = $199; //@line 351 "libraries/lodepng/lodepng.cpp"
   $201 = $w >>> 8; //@line 352 "libraries/lodepng/lodepng.cpp"
   $202 = $201&255; //@line 352 "libraries/lodepng/lodepng.cpp"
   $203 = ((($195)) + 2|0); //@line 352 "libraries/lodepng/lodepng.cpp"
   HEAP8[$203>>0] = $202; //@line 352 "libraries/lodepng/lodepng.cpp"
   $204 = $w&255; //@line 353 "libraries/lodepng/lodepng.cpp"
   $205 = ((($195)) + 3|0); //@line 353 "libraries/lodepng/lodepng.cpp"
   HEAP8[$205>>0] = $204; //@line 353 "libraries/lodepng/lodepng.cpp"
   $206 = (($$32$i) + 4)|0; //@line 360 "libraries/lodepng/lodepng.cpp"
   $207 = $$32$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
   $208 = ($207>>>0)<($206>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
   $209 = ($206*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
   $210 = $209 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
   $211 = $208 ? $206 : $210; //@line 227 "libraries/lodepng/lodepng.cpp"
   $212 = (_realloc($192,$211)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
   $213 = ($212|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
   $$$i = $213 ? $192 : $212; //@line 229 "libraries/lodepng/lodepng.cpp"
   $$32$$i = $213 ? $$32$i : $211; //@line 229 "libraries/lodepng/lodepng.cpp"
   $$32$36$i = $213 ? $$32$i : $206; //@line 229 "libraries/lodepng/lodepng.cpp"
   $214 = (($$32$36$i) + -4)|0; //@line 361 "libraries/lodepng/lodepng.cpp"
   $215 = (($$$i) + ($214)|0); //@line 361 "libraries/lodepng/lodepng.cpp"
   $216 = $h >>> 24; //@line 350 "libraries/lodepng/lodepng.cpp"
   $217 = $216&255; //@line 350 "libraries/lodepng/lodepng.cpp"
   HEAP8[$215>>0] = $217; //@line 350 "libraries/lodepng/lodepng.cpp"
   $218 = $h >>> 16; //@line 351 "libraries/lodepng/lodepng.cpp"
   $219 = $218&255; //@line 351 "libraries/lodepng/lodepng.cpp"
   $220 = ((($215)) + 1|0); //@line 351 "libraries/lodepng/lodepng.cpp"
   HEAP8[$220>>0] = $219; //@line 351 "libraries/lodepng/lodepng.cpp"
   $221 = $h >>> 8; //@line 352 "libraries/lodepng/lodepng.cpp"
   $222 = $221&255; //@line 352 "libraries/lodepng/lodepng.cpp"
   $223 = ((($215)) + 2|0); //@line 352 "libraries/lodepng/lodepng.cpp"
   HEAP8[$223>>0] = $222; //@line 352 "libraries/lodepng/lodepng.cpp"
   $224 = $h&255; //@line 353 "libraries/lodepng/lodepng.cpp"
   $225 = ((($215)) + 3|0); //@line 353 "libraries/lodepng/lodepng.cpp"
   HEAP8[$225>>0] = $224; //@line 353 "libraries/lodepng/lodepng.cpp"
   $226 = $190&255; //@line 4802 "libraries/lodepng/lodepng.cpp"
   $227 = $$32$36$i | 1; //@line 288 "libraries/lodepng/lodepng.cpp"
   $228 = ($$32$$i>>>0)<($227>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($228) {
    $229 = $$32$$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $230 = ($229>>>0)<($227>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $231 = ($227*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $232 = $231 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $233 = $230 ? $227 : $232; //@line 227 "libraries/lodepng/lodepng.cpp"
    $234 = (_realloc($$$i,$233)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $235 = ($234|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if ($235) {
     $header$sroa$0$5$i = $$$i;$header$sroa$31$2$i = $$32$36$i;$header$sroa$50$5$i = $$32$$i;
    } else {
     $header$sroa$0$4$i = $234;$header$sroa$50$4$i = $233;
     label = 89;
    }
   } else {
    $header$sroa$0$4$i = $$$i;$header$sroa$50$4$i = $$32$$i;
    label = 89;
   }
   if ((label|0) == 89) {
    $236 = (($header$sroa$0$4$i) + ($$32$36$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$236>>0] = $226; //@line 289 "libraries/lodepng/lodepng.cpp"
    $header$sroa$0$5$i = $header$sroa$0$4$i;$header$sroa$31$2$i = $227;$header$sroa$50$5$i = $header$sroa$50$4$i;
   }
   $237 = $189&255; //@line 4803 "libraries/lodepng/lodepng.cpp"
   $238 = (($header$sroa$31$2$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $239 = ($header$sroa$50$5$i>>>0)<($238>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($239) {
    $240 = $header$sroa$50$5$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $241 = ($240>>>0)<($238>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $242 = ($238*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $243 = $242 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $244 = $241 ? $238 : $243; //@line 227 "libraries/lodepng/lodepng.cpp"
    $245 = (_realloc($header$sroa$0$5$i,$244)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $246 = ($245|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if ($246) {
     $header$sroa$0$7$i = $header$sroa$0$5$i;$header$sroa$31$3$i = $header$sroa$31$2$i;$header$sroa$50$7$i = $header$sroa$50$5$i;
    } else {
     $header$sroa$0$6$i = $245;$header$sroa$50$6$i = $244;
     label = 92;
    }
   } else {
    $header$sroa$0$6$i = $header$sroa$0$5$i;$header$sroa$50$6$i = $header$sroa$50$5$i;
    label = 92;
   }
   if ((label|0) == 92) {
    $247 = (($header$sroa$0$6$i) + ($header$sroa$31$2$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$247>>0] = $237; //@line 289 "libraries/lodepng/lodepng.cpp"
    $header$sroa$0$7$i = $header$sroa$0$6$i;$header$sroa$31$3$i = $238;$header$sroa$50$7$i = $header$sroa$50$6$i;
   }
   $248 = (($header$sroa$31$3$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $249 = ($header$sroa$50$7$i>>>0)<($248>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($249) {
    $250 = $header$sroa$50$7$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $251 = ($250>>>0)<($248>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $252 = ($248*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $253 = $252 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $254 = $251 ? $248 : $253; //@line 227 "libraries/lodepng/lodepng.cpp"
    $255 = (_realloc($header$sroa$0$7$i,$254)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $256 = ($255|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if ($256) {
     $header$sroa$0$9$i = $header$sroa$0$7$i;$header$sroa$31$4$i = $header$sroa$31$3$i;$header$sroa$50$9$i = $header$sroa$50$7$i;
    } else {
     $header$sroa$0$8$i = $255;$header$sroa$50$8$i = $254;
     label = 95;
    }
   } else {
    $header$sroa$0$8$i = $header$sroa$0$7$i;$header$sroa$50$8$i = $header$sroa$50$7$i;
    label = 95;
   }
   if ((label|0) == 95) {
    $257 = (($header$sroa$0$8$i) + ($header$sroa$31$3$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$257>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
    $header$sroa$0$9$i = $header$sroa$0$8$i;$header$sroa$31$4$i = $248;$header$sroa$50$9$i = $header$sroa$50$8$i;
   }
   $258 = (($header$sroa$31$4$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $259 = ($header$sroa$50$9$i>>>0)<($258>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($259) {
    $260 = $header$sroa$50$9$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $261 = ($260>>>0)<($258>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $262 = ($258*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $263 = $262 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $264 = $261 ? $258 : $263; //@line 227 "libraries/lodepng/lodepng.cpp"
    $265 = (_realloc($header$sroa$0$9$i,$264)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $266 = ($265|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if ($266) {
     $header$sroa$0$11$i = $header$sroa$0$9$i;$header$sroa$31$5$i = $header$sroa$31$4$i;$header$sroa$50$11$i = $header$sroa$50$9$i;
    } else {
     $header$sroa$0$10$i = $265;$header$sroa$50$10$i = $264;
     label = 98;
    }
   } else {
    $header$sroa$0$10$i = $header$sroa$0$9$i;$header$sroa$50$10$i = $header$sroa$50$9$i;
    label = 98;
   }
   if ((label|0) == 98) {
    $267 = (($header$sroa$0$10$i) + ($header$sroa$31$4$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$267>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
    $header$sroa$0$11$i = $header$sroa$0$10$i;$header$sroa$31$5$i = $258;$header$sroa$50$11$i = $header$sroa$50$10$i;
   }
   $268 = $191&255; //@line 4806 "libraries/lodepng/lodepng.cpp"
   $269 = (($header$sroa$31$5$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $270 = ($header$sroa$50$11$i>>>0)<($269>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($270) {
    $271 = $header$sroa$50$11$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $272 = ($271>>>0)<($269>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $273 = ($269*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $274 = $273 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $275 = $272 ? $269 : $274; //@line 227 "libraries/lodepng/lodepng.cpp"
    $276 = (_realloc($header$sroa$0$11$i,$275)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $277 = ($276|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if ($277) {
     $header$sroa$0$13$i = $header$sroa$0$11$i;$header$sroa$31$6$i = $header$sroa$31$5$i;
    } else {
     $header$sroa$0$12$i = $276;
     label = 101;
    }
   } else {
    $header$sroa$0$12$i = $header$sroa$0$11$i;
    label = 101;
   }
   if ((label|0) == 101) {
    $278 = (($header$sroa$0$12$i) + ($header$sroa$31$5$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$278>>0] = $268; //@line 289 "libraries/lodepng/lodepng.cpp"
    $header$sroa$0$13$i = $header$sroa$0$12$i;$header$sroa$31$6$i = $269;
   }
   $279 = (__Z20lodepng_chunk_createPPhPjjPKcPKh($outv,$86,$header$sroa$31$6$i,502730,$header$sroa$0$13$i)|0); //@line 4775 "libraries/lodepng/lodepng.cpp"
   $280 = ($279|0)==(0); //@line 4775 "libraries/lodepng/lodepng.cpp"
   if ($280) {
    $281 = HEAP32[$86>>2]|0; //@line 4776 "libraries/lodepng/lodepng.cpp"
    HEAP32[$85>>2] = $281; //@line 4776 "libraries/lodepng/lodepng.cpp"
   }
   _free($header$sroa$0$13$i); //@line 76 "libraries/lodepng/lodepng.cpp"
   $282 = HEAP32[$scevgep4$i$i>>2]|0; //@line 5617 "libraries/lodepng/lodepng.cpp"
   $283 = ($282|0)==(0|0); //@line 5617 "libraries/lodepng/lodepng.cpp"
   if (!($283)) {
    $284 = ((($info)) + 148|0); //@line 5619 "libraries/lodepng/lodepng.cpp"
    $285 = HEAP32[$284>>2]|0; //@line 5619 "libraries/lodepng/lodepng.cpp"
    $286 = $282; //@line 5535 "libraries/lodepng/lodepng.cpp"
    $287 = ($285|0)==(0); //@line 5535 "libraries/lodepng/lodepng.cpp"
    L156: do {
     if (!($287)) {
      $$pre156 = HEAP32[$86>>2]|0; //@line 2469 "libraries/lodepng/lodepng.cpp"
      $307 = $$pre156;$inchunk$02$i30 = $282;
      while(1) {
       $288 = HEAP8[$inchunk$02$i30>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $289 = $288&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $290 = $289 << 24; //@line 343 "libraries/lodepng/lodepng.cpp"
       $291 = ((($inchunk$02$i30)) + 1|0); //@line 343 "libraries/lodepng/lodepng.cpp"
       $292 = HEAP8[$291>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $293 = $292&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $294 = $293 << 16; //@line 343 "libraries/lodepng/lodepng.cpp"
       $295 = $294 | $290; //@line 343 "libraries/lodepng/lodepng.cpp"
       $296 = ((($inchunk$02$i30)) + 2|0); //@line 343 "libraries/lodepng/lodepng.cpp"
       $297 = HEAP8[$296>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $298 = $297&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $299 = $298 << 8; //@line 343 "libraries/lodepng/lodepng.cpp"
       $300 = $295 | $299; //@line 343 "libraries/lodepng/lodepng.cpp"
       $301 = ((($inchunk$02$i30)) + 3|0); //@line 343 "libraries/lodepng/lodepng.cpp"
       $302 = HEAP8[$301>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $303 = $302&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $304 = $300 | $303; //@line 343 "libraries/lodepng/lodepng.cpp"
       $305 = (($304) + 12)|0; //@line 2467 "libraries/lodepng/lodepng.cpp"
       $306 = (($305) + ($307))|0; //@line 5538 "libraries/lodepng/lodepng.cpp"
       $308 = ($306>>>0)<($305>>>0); //@line 2470 "libraries/lodepng/lodepng.cpp"
       $309 = ($306>>>0)<($307>>>0); //@line 2470 "libraries/lodepng/lodepng.cpp"
       $or$cond$i34 = $308 | $309; //@line 2470 "libraries/lodepng/lodepng.cpp"
       if ($or$cond$i34) {
        $$2$i31 = 77;
        break;
       }
       $310 = HEAP32[$outv>>2]|0; //@line 2472 "libraries/lodepng/lodepng.cpp"
       $311 = (_realloc($310,$306)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       $312 = ($311|0)==(0|0); //@line 2473 "libraries/lodepng/lodepng.cpp"
       if ($312) {
        $$2$i31 = 83;
        break;
       }
       HEAP32[$outv>>2] = $311; //@line 2474 "libraries/lodepng/lodepng.cpp"
       HEAP32[$86>>2] = $306; //@line 2475 "libraries/lodepng/lodepng.cpp"
       $313 = (($311) + ($307)|0); //@line 2476 "libraries/lodepng/lodepng.cpp"
       $314 = ($305|0)==(0); //@line 2478 "libraries/lodepng/lodepng.cpp"
       if (!($314)) {
        $i$01$i35 = 0;
        while(1) {
         $315 = (($inchunk$02$i30) + ($i$01$i35)|0); //@line 2478 "libraries/lodepng/lodepng.cpp"
         $316 = HEAP8[$315>>0]|0; //@line 2478 "libraries/lodepng/lodepng.cpp"
         $317 = (($313) + ($i$01$i35)|0); //@line 2478 "libraries/lodepng/lodepng.cpp"
         HEAP8[$317>>0] = $316; //@line 2478 "libraries/lodepng/lodepng.cpp"
         $318 = (($i$01$i35) + 1)|0; //@line 2478 "libraries/lodepng/lodepng.cpp"
         $319 = ($318|0)==($305|0); //@line 2478 "libraries/lodepng/lodepng.cpp"
         if ($319) {
          break;
         } else {
          $i$01$i35 = $318;
         }
        }
       }
       HEAP32[$85>>2] = $306; //@line 5538 "libraries/lodepng/lodepng.cpp"
       $320 = HEAP8[$inchunk$02$i30>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $321 = $320&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $322 = $321 << 24; //@line 343 "libraries/lodepng/lodepng.cpp"
       $323 = HEAP8[$291>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $324 = $323&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $325 = $324 << 16; //@line 343 "libraries/lodepng/lodepng.cpp"
       $326 = $325 | $322; //@line 343 "libraries/lodepng/lodepng.cpp"
       $327 = HEAP8[$296>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $328 = $327&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $329 = $328 << 8; //@line 343 "libraries/lodepng/lodepng.cpp"
       $330 = $326 | $329; //@line 343 "libraries/lodepng/lodepng.cpp"
       $331 = HEAP8[$301>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $332 = $331&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $333 = $330 | $332; //@line 343 "libraries/lodepng/lodepng.cpp"
       $334 = (($333) + 12)|0; //@line 2454 "libraries/lodepng/lodepng.cpp"
       $335 = (($inchunk$02$i30) + ($334)|0); //@line 2455 "libraries/lodepng/lodepng.cpp"
       $336 = $335; //@line 5535 "libraries/lodepng/lodepng.cpp"
       $337 = (($336) - ($286))|0; //@line 5535 "libraries/lodepng/lodepng.cpp"
       $338 = ($337>>>0)<($285>>>0); //@line 5535 "libraries/lodepng/lodepng.cpp"
       if ($338) {
        $307 = $306;$inchunk$02$i30 = $335;
       } else {
        break L156;
       }
      }
      HEAP32[$0>>2] = $$2$i31; //@line 5619 "libraries/lodepng/lodepng.cpp"
      break L71;
     }
    } while(0);
    HEAP32[$0>>2] = 0; //@line 5619 "libraries/lodepng/lodepng.cpp"
   }
   $339 = HEAP32[$11>>2]|0; //@line 5624 "libraries/lodepng/lodepng.cpp"
   $340 = ($339|0)==(3); //@line 5624 "libraries/lodepng/lodepng.cpp"
   if ($340) {
    $341 = HEAP32[$5>>2]|0; //@line 4820 "libraries/lodepng/lodepng.cpp"
    $$mask$i40 = $341 & 1073741823; //@line 4820 "libraries/lodepng/lodepng.cpp"
    $342 = ($$mask$i40|0)==(0); //@line 4820 "libraries/lodepng/lodepng.cpp"
    if ($342) {
     $PLTE$sroa$0$0$lcssa$i = 0;$PLTE$sroa$9$0$lcssa$i = 0;
    } else {
     $PLTE$sroa$0$03$i = 0;$PLTE$sroa$14$01$i = 0;$PLTE$sroa$9$02$i = 0;$i$04$i = 0;
     while(1) {
      $343 = $i$04$i & 3; //@line 4823 "libraries/lodepng/lodepng.cpp"
      $344 = ($343|0)==(3); //@line 4823 "libraries/lodepng/lodepng.cpp"
      do {
       if ($344) {
        $PLTE$sroa$0$3$i = $PLTE$sroa$0$03$i;$PLTE$sroa$14$3$i = $PLTE$sroa$14$01$i;$PLTE$sroa$9$2$i = $PLTE$sroa$9$02$i;
       } else {
        $345 = HEAP32[$4>>2]|0; //@line 4823 "libraries/lodepng/lodepng.cpp"
        $346 = (($345) + ($i$04$i)|0); //@line 4823 "libraries/lodepng/lodepng.cpp"
        $347 = HEAP8[$346>>0]|0; //@line 4823 "libraries/lodepng/lodepng.cpp"
        $348 = (($PLTE$sroa$9$02$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
        $349 = ($PLTE$sroa$14$01$i>>>0)<($348>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
        if ($349) {
         $350 = $PLTE$sroa$14$01$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
         $351 = ($350>>>0)<($348>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
         $352 = ($348*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
         $353 = $352 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
         $354 = $351 ? $348 : $353; //@line 227 "libraries/lodepng/lodepng.cpp"
         $355 = (_realloc($PLTE$sroa$0$03$i,$354)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
         $356 = ($355|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
         if ($356) {
          $PLTE$sroa$0$3$i = $PLTE$sroa$0$03$i;$PLTE$sroa$14$3$i = $PLTE$sroa$14$01$i;$PLTE$sroa$9$2$i = $PLTE$sroa$9$02$i;
          break;
         } else {
          $PLTE$sroa$0$1$i = $355;$PLTE$sroa$14$1$i = $354;
         }
        } else {
         $PLTE$sroa$0$1$i = $PLTE$sroa$0$03$i;$PLTE$sroa$14$1$i = $PLTE$sroa$14$01$i;
        }
        $357 = (($PLTE$sroa$0$1$i) + ($PLTE$sroa$9$02$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
        HEAP8[$357>>0] = $347; //@line 289 "libraries/lodepng/lodepng.cpp"
        $PLTE$sroa$0$3$i = $PLTE$sroa$0$1$i;$PLTE$sroa$14$3$i = $PLTE$sroa$14$1$i;$PLTE$sroa$9$2$i = $348;
       }
      } while(0);
      $358 = (($i$04$i) + 1)|0; //@line 4820 "libraries/lodepng/lodepng.cpp"
      $359 = HEAP32[$5>>2]|0; //@line 4820 "libraries/lodepng/lodepng.cpp"
      $360 = $359 << 2; //@line 4820 "libraries/lodepng/lodepng.cpp"
      $361 = ($358|0)==($360|0); //@line 4820 "libraries/lodepng/lodepng.cpp"
      if ($361) {
       $PLTE$sroa$0$0$lcssa$i = $PLTE$sroa$0$3$i;$PLTE$sroa$9$0$lcssa$i = $PLTE$sroa$9$2$i;
       break;
      } else {
       $PLTE$sroa$0$03$i = $PLTE$sroa$0$3$i;$PLTE$sroa$14$01$i = $PLTE$sroa$14$3$i;$PLTE$sroa$9$02$i = $PLTE$sroa$9$2$i;$i$04$i = $358;
      }
     }
    }
    $362 = (__Z20lodepng_chunk_createPPhPjjPKcPKh($outv,$86,$PLTE$sroa$9$0$lcssa$i,502735,$PLTE$sroa$0$0$lcssa$i)|0); //@line 4775 "libraries/lodepng/lodepng.cpp"
    $363 = ($362|0)==(0); //@line 4775 "libraries/lodepng/lodepng.cpp"
    if ($363) {
     $364 = HEAP32[$86>>2]|0; //@line 4776 "libraries/lodepng/lodepng.cpp"
     HEAP32[$85>>2] = $364; //@line 4776 "libraries/lodepng/lodepng.cpp"
    }
    _free($PLTE$sroa$0$0$lcssa$i); //@line 76 "libraries/lodepng/lodepng.cpp"
   }
   $365 = ((($state)) + 88|0); //@line 5628 "libraries/lodepng/lodepng.cpp"
   $366 = HEAP32[$365>>2]|0; //@line 5628 "libraries/lodepng/lodepng.cpp"
   $367 = ($366|0)==(0); //@line 5628 "libraries/lodepng/lodepng.cpp"
   L185: do {
    if ($367) {
     label = 135;
    } else {
     $368 = HEAP32[$11>>2]|0; //@line 5628 "libraries/lodepng/lodepng.cpp"
     switch ($368|0) {
     case 6: case 2:  {
      break;
     }
     default: {
      $394 = $368;
      break L185;
     }
     }
     $369 = HEAP32[$5>>2]|0; //@line 4820 "libraries/lodepng/lodepng.cpp"
     $$mask$i46 = $369 & 1073741823; //@line 4820 "libraries/lodepng/lodepng.cpp"
     $370 = ($$mask$i46|0)==(0); //@line 4820 "libraries/lodepng/lodepng.cpp"
     if ($370) {
      $PLTE$sroa$0$0$lcssa$i59 = 0;$PLTE$sroa$9$0$lcssa$i60 = 0;
     } else {
      $PLTE$sroa$0$03$i49 = 0;$PLTE$sroa$14$01$i51 = 0;$PLTE$sroa$9$02$i50 = 0;$i$04$i48 = 0;
      while(1) {
       $371 = $i$04$i48 & 3; //@line 4823 "libraries/lodepng/lodepng.cpp"
       $372 = ($371|0)==(3); //@line 4823 "libraries/lodepng/lodepng.cpp"
       do {
        if ($372) {
         $PLTE$sroa$0$3$i57 = $PLTE$sroa$0$03$i49;$PLTE$sroa$14$3$i55 = $PLTE$sroa$14$01$i51;$PLTE$sroa$9$2$i56 = $PLTE$sroa$9$02$i50;
        } else {
         $373 = HEAP32[$4>>2]|0; //@line 4823 "libraries/lodepng/lodepng.cpp"
         $374 = (($373) + ($i$04$i48)|0); //@line 4823 "libraries/lodepng/lodepng.cpp"
         $375 = HEAP8[$374>>0]|0; //@line 4823 "libraries/lodepng/lodepng.cpp"
         $376 = (($PLTE$sroa$9$02$i50) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
         $377 = ($PLTE$sroa$14$01$i51>>>0)<($376>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
         if ($377) {
          $378 = $PLTE$sroa$14$01$i51 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
          $379 = ($378>>>0)<($376>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
          $380 = ($376*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
          $381 = $380 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
          $382 = $379 ? $376 : $381; //@line 227 "libraries/lodepng/lodepng.cpp"
          $383 = (_realloc($PLTE$sroa$0$03$i49,$382)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
          $384 = ($383|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
          if ($384) {
           $PLTE$sroa$0$3$i57 = $PLTE$sroa$0$03$i49;$PLTE$sroa$14$3$i55 = $PLTE$sroa$14$01$i51;$PLTE$sroa$9$2$i56 = $PLTE$sroa$9$02$i50;
           break;
          } else {
           $PLTE$sroa$0$1$i53 = $383;$PLTE$sroa$14$1$i52 = $382;
          }
         } else {
          $PLTE$sroa$0$1$i53 = $PLTE$sroa$0$03$i49;$PLTE$sroa$14$1$i52 = $PLTE$sroa$14$01$i51;
         }
         $385 = (($PLTE$sroa$0$1$i53) + ($PLTE$sroa$9$02$i50)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
         HEAP8[$385>>0] = $375; //@line 289 "libraries/lodepng/lodepng.cpp"
         $PLTE$sroa$0$3$i57 = $PLTE$sroa$0$1$i53;$PLTE$sroa$14$3$i55 = $PLTE$sroa$14$1$i52;$PLTE$sroa$9$2$i56 = $376;
        }
       } while(0);
       $386 = (($i$04$i48) + 1)|0; //@line 4820 "libraries/lodepng/lodepng.cpp"
       $387 = HEAP32[$5>>2]|0; //@line 4820 "libraries/lodepng/lodepng.cpp"
       $388 = $387 << 2; //@line 4820 "libraries/lodepng/lodepng.cpp"
       $389 = ($386|0)==($388|0); //@line 4820 "libraries/lodepng/lodepng.cpp"
       if ($389) {
        $PLTE$sroa$0$0$lcssa$i59 = $PLTE$sroa$0$3$i57;$PLTE$sroa$9$0$lcssa$i60 = $PLTE$sroa$9$2$i56;
        break;
       } else {
        $PLTE$sroa$0$03$i49 = $PLTE$sroa$0$3$i57;$PLTE$sroa$14$01$i51 = $PLTE$sroa$14$3$i55;$PLTE$sroa$9$02$i50 = $PLTE$sroa$9$2$i56;$i$04$i48 = $386;
       }
      }
     }
     $390 = (__Z20lodepng_chunk_createPPhPjjPKcPKh($outv,$86,$PLTE$sroa$9$0$lcssa$i60,502735,$PLTE$sroa$0$0$lcssa$i59)|0); //@line 4775 "libraries/lodepng/lodepng.cpp"
     $391 = ($390|0)==(0); //@line 4775 "libraries/lodepng/lodepng.cpp"
     if ($391) {
      $392 = HEAP32[$86>>2]|0; //@line 4776 "libraries/lodepng/lodepng.cpp"
      HEAP32[$85>>2] = $392; //@line 4776 "libraries/lodepng/lodepng.cpp"
     }
     _free($PLTE$sroa$0$0$lcssa$i59); //@line 76 "libraries/lodepng/lodepng.cpp"
     label = 135;
    }
   } while(0);
   if ((label|0) == 135) {
    $$pr132 = HEAP32[$11>>2]|0; //@line 5633 "libraries/lodepng/lodepng.cpp"
    $394 = $$pr132;
   }
   $393 = ($394|0)==(3); //@line 5633 "libraries/lodepng/lodepng.cpp"
   do {
    if ($393) {
     $395 = HEAP32[$4>>2]|0; //@line 5633 "libraries/lodepng/lodepng.cpp"
     $396 = HEAP32[$5>>2]|0; //@line 5633 "libraries/lodepng/lodepng.cpp"
     $397 = ($396|0)==(0); //@line 5516 "libraries/lodepng/lodepng.cpp"
     if ($397) {
      break;
     } else {
      $b$05$i = 0;$g$04$i = 0;$i$01$i64 = 0;$key$02$i = 0;$r$03$i = 0;
     }
     L206: while(1) {
      $398 = ($key$02$i|0)!=(0); //@line 5518 "libraries/lodepng/lodepng.cpp"
      $$pre$i = $i$01$i64 << 2; //@line 5524 "libraries/lodepng/lodepng.cpp"
      $$pre10$i = $$pre$i | 3; //@line 5524 "libraries/lodepng/lodepng.cpp"
      $$pre12$i = (($395) + ($$pre10$i)|0); //@line 5524 "libraries/lodepng/lodepng.cpp"
      $$pr133 = HEAP8[$$pre12$i>>0]|0; //@line 5524 "libraries/lodepng/lodepng.cpp"
      $$not = $398 ^ 1; //@line 5518 "libraries/lodepng/lodepng.cpp"
      $399 = ($$pr133<<24>>24)==(0); //@line 5518 "libraries/lodepng/lodepng.cpp"
      $or$cond135 = $399 & $$not; //@line 5518 "libraries/lodepng/lodepng.cpp"
      do {
       if ($or$cond135) {
        $400 = (($395) + ($$pre$i)|0); //@line 5520 "libraries/lodepng/lodepng.cpp"
        $401 = HEAP8[$400>>0]|0; //@line 5520 "libraries/lodepng/lodepng.cpp"
        $402 = $401&255; //@line 5520 "libraries/lodepng/lodepng.cpp"
        $403 = $$pre$i | 1; //@line 5520 "libraries/lodepng/lodepng.cpp"
        $404 = (($395) + ($403)|0); //@line 5520 "libraries/lodepng/lodepng.cpp"
        $405 = HEAP8[$404>>0]|0; //@line 5520 "libraries/lodepng/lodepng.cpp"
        $406 = $405&255; //@line 5520 "libraries/lodepng/lodepng.cpp"
        $407 = $$pre$i | 2; //@line 5520 "libraries/lodepng/lodepng.cpp"
        $408 = (($395) + ($407)|0); //@line 5520 "libraries/lodepng/lodepng.cpp"
        $409 = HEAP8[$408>>0]|0; //@line 5520 "libraries/lodepng/lodepng.cpp"
        $410 = $409&255; //@line 5520 "libraries/lodepng/lodepng.cpp"
        $b$1$i = $410;$g$1$i = $406;$i$1$i = -1;$key$1$i = 1;$r$1$i = $402;
       } else {
        $411 = ($$pr133<<24>>24)==(-1); //@line 5524 "libraries/lodepng/lodepng.cpp"
        if (!($411)) {
         break L206;
        }
        if (!($398)) {
         $b$1$i = $b$05$i;$g$1$i = $g$04$i;$i$1$i = $i$01$i64;$key$1$i = 0;$r$1$i = $r$03$i;
         break;
        }
        $412 = (($395) + ($$pre$i)|0); //@line 5526 "libraries/lodepng/lodepng.cpp"
        $413 = HEAP8[$412>>0]|0; //@line 5526 "libraries/lodepng/lodepng.cpp"
        $414 = $413&255; //@line 5526 "libraries/lodepng/lodepng.cpp"
        $415 = ($r$03$i|0)==($414|0); //@line 5526 "libraries/lodepng/lodepng.cpp"
        if (!($415)) {
         $b$1$i = $b$05$i;$g$1$i = $g$04$i;$i$1$i = $i$01$i64;$key$1$i = $key$02$i;$r$1$i = $r$03$i;
         break;
        }
        $416 = $$pre$i | 1; //@line 5526 "libraries/lodepng/lodepng.cpp"
        $417 = (($395) + ($416)|0); //@line 5526 "libraries/lodepng/lodepng.cpp"
        $418 = HEAP8[$417>>0]|0; //@line 5526 "libraries/lodepng/lodepng.cpp"
        $419 = $418&255; //@line 5526 "libraries/lodepng/lodepng.cpp"
        $420 = ($g$04$i|0)==($419|0); //@line 5526 "libraries/lodepng/lodepng.cpp"
        if (!($420)) {
         $b$1$i = $b$05$i;$g$1$i = $g$04$i;$i$1$i = $i$01$i64;$key$1$i = $key$02$i;$r$1$i = $r$03$i;
         break;
        }
        $421 = $$pre$i | 2; //@line 5526 "libraries/lodepng/lodepng.cpp"
        $422 = (($395) + ($421)|0); //@line 5526 "libraries/lodepng/lodepng.cpp"
        $423 = HEAP8[$422>>0]|0; //@line 5526 "libraries/lodepng/lodepng.cpp"
        $424 = $423&255; //@line 5526 "libraries/lodepng/lodepng.cpp"
        $425 = ($b$05$i|0)==($424|0); //@line 5526 "libraries/lodepng/lodepng.cpp"
        if ($425) {
         break L206;
        } else {
         $b$1$i = $b$05$i;$g$1$i = $g$04$i;$i$1$i = $i$01$i64;$key$1$i = $key$02$i;$r$1$i = $r$03$i;
        }
       }
      } while(0);
      $426 = (($i$1$i) + 1)|0; //@line 5516 "libraries/lodepng/lodepng.cpp"
      $427 = ($426|0)==($396|0); //@line 5516 "libraries/lodepng/lodepng.cpp"
      if ($427) {
       $key$1$i$lcssa = $key$1$i;
       label = 146;
       break;
      } else {
       $b$05$i = $b$1$i;$g$04$i = $g$1$i;$i$01$i64 = $426;$key$02$i = $key$1$i;$r$03$i = $r$1$i;
      }
     }
     if ((label|0) == 146) {
      $428 = ($key$1$i$lcssa|0)==(0); //@line 5633 "libraries/lodepng/lodepng.cpp"
      if ($428) {
       break;
      }
     }
     __ZL13addChunk_tRNSP8ucvectorPK16LodePNGColorMode($outv,$11); //@line 5635 "libraries/lodepng/lodepng.cpp"
     $$pre157 = HEAP32[$11>>2]|0; //@line 5637 "libraries/lodepng/lodepng.cpp"
     $429 = $$pre157;
     label = 148;
    } else {
     $429 = $394;
     label = 148;
    }
   } while(0);
   L219: do {
    if ((label|0) == 148) {
     switch ($429|0) {
     case 2: case 0:  {
      break;
     }
     default: {
      break L219;
     }
     }
     $430 = HEAP32[$1>>2]|0; //@line 5637 "libraries/lodepng/lodepng.cpp"
     $431 = ($430|0)==(0); //@line 5637 "libraries/lodepng/lodepng.cpp"
     if ($431) {
      break;
     }
     __ZL13addChunk_tRNSP8ucvectorPK16LodePNGColorMode($outv,$11); //@line 5639 "libraries/lodepng/lodepng.cpp"
    }
   } while(0);
   $432 = HEAP32[$8>>2]|0; //@line 5643 "libraries/lodepng/lodepng.cpp"
   $433 = ($432|0)==(0); //@line 5643 "libraries/lodepng/lodepng.cpp"
   if (!($433)) {
    $434 = HEAP32[$2>>2]|0; //@line 4989 "libraries/lodepng/lodepng.cpp"
    L226: do {
     switch ($434|0) {
     case 4: case 0:  {
      $435 = ((($info)) + 48|0); //@line 4991 "libraries/lodepng/lodepng.cpp"
      $436 = HEAP32[$435>>2]|0; //@line 4991 "libraries/lodepng/lodepng.cpp"
      $437 = (_realloc(0,1)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $438 = ($437|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
      if ($438) {
       $443 = 1;$bKGD$sroa$0$144$i = 0;$bKGD$sroa$33$043$i = 0;
      } else {
       $439 = $436 >>> 8; //@line 4991 "libraries/lodepng/lodepng.cpp"
       $440 = $439&255; //@line 4991 "libraries/lodepng/lodepng.cpp"
       HEAP8[$437>>0] = $440; //@line 289 "libraries/lodepng/lodepng.cpp"
       $443 = 2;$bKGD$sroa$0$144$i = $437;$bKGD$sroa$33$043$i = 1;
      }
      $$sink$i = HEAP32[$435>>2]|0; //@line 4992 "libraries/lodepng/lodepng.cpp"
      $441 = $bKGD$sroa$33$043$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
      $442 = ($441>>>0)<($443>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
      $444 = ($443*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
      $445 = $444 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
      $446 = $442 ? $443 : $445; //@line 227 "libraries/lodepng/lodepng.cpp"
      $447 = (_realloc($bKGD$sroa$0$144$i,$446)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $448 = ($447|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
      if ($448) {
       $bKGD$sroa$0$18$i = $bKGD$sroa$0$144$i;$bKGD$sroa$33$9$i = $bKGD$sroa$33$043$i;
       break L226;
      }
      $449 = $$sink$i&255; //@line 4992 "libraries/lodepng/lodepng.cpp"
      $450 = (($447) + ($bKGD$sroa$33$043$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
      HEAP8[$450>>0] = $449; //@line 289 "libraries/lodepng/lodepng.cpp"
      $bKGD$sroa$0$18$i = $447;$bKGD$sroa$33$9$i = $443;
      break;
     }
     case 6: case 2:  {
      $451 = ((($info)) + 48|0); //@line 4996 "libraries/lodepng/lodepng.cpp"
      $452 = HEAP32[$451>>2]|0; //@line 4996 "libraries/lodepng/lodepng.cpp"
      $453 = (_realloc(0,1)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $454 = ($453|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
      if ($454) {
       $459 = 1;$bKGD$sroa$0$548$i = 0;$bKGD$sroa$33$247$i = 0;
      } else {
       $455 = $452 >>> 8; //@line 4996 "libraries/lodepng/lodepng.cpp"
       $456 = $455&255; //@line 4996 "libraries/lodepng/lodepng.cpp"
       HEAP8[$453>>0] = $456; //@line 289 "libraries/lodepng/lodepng.cpp"
       $459 = 2;$bKGD$sroa$0$548$i = $453;$bKGD$sroa$33$247$i = 1;
      }
      $$sink49$i = HEAP32[$451>>2]|0; //@line 4997 "libraries/lodepng/lodepng.cpp"
      $457 = $bKGD$sroa$33$247$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
      $458 = ($457>>>0)<($459>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
      $460 = ($459*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
      $461 = $460 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
      $462 = $458 ? $459 : $461; //@line 227 "libraries/lodepng/lodepng.cpp"
      $463 = (_realloc($bKGD$sroa$0$548$i,$462)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $464 = ($463|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
      if ($464) {
       $bKGD$sroa$0$7$i = $bKGD$sroa$0$548$i;$bKGD$sroa$33$3$i = $bKGD$sroa$33$247$i;$bKGD$sroa$54$5$i = $bKGD$sroa$33$247$i;
      } else {
       $465 = $$sink49$i&255; //@line 4997 "libraries/lodepng/lodepng.cpp"
       $466 = (($463) + ($bKGD$sroa$33$247$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
       HEAP8[$466>>0] = $465; //@line 289 "libraries/lodepng/lodepng.cpp"
       $bKGD$sroa$0$7$i = $463;$bKGD$sroa$33$3$i = $459;$bKGD$sroa$54$5$i = $462;
      }
      $467 = ((($info)) + 52|0); //@line 4998 "libraries/lodepng/lodepng.cpp"
      $468 = HEAP32[$467>>2]|0; //@line 4998 "libraries/lodepng/lodepng.cpp"
      $469 = $468 >>> 8; //@line 4998 "libraries/lodepng/lodepng.cpp"
      $470 = $469&255; //@line 4998 "libraries/lodepng/lodepng.cpp"
      $471 = (($bKGD$sroa$33$3$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
      $472 = ($bKGD$sroa$54$5$i>>>0)<($471>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
      if ($472) {
       $473 = $bKGD$sroa$54$5$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
       $474 = ($473>>>0)<($471>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
       $475 = ($471*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
       $476 = $475 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
       $477 = $474 ? $471 : $476; //@line 227 "libraries/lodepng/lodepng.cpp"
       $478 = (_realloc($bKGD$sroa$0$7$i,$477)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       $479 = ($478|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
       if ($479) {
        $bKGD$sroa$0$9$i = $bKGD$sroa$0$7$i;$bKGD$sroa$33$4$i = $bKGD$sroa$33$3$i;$bKGD$sroa$54$7$i = $bKGD$sroa$54$5$i;
       } else {
        $bKGD$sroa$0$8$i = $478;$bKGD$sroa$54$6$i = $477;
        label = 163;
       }
      } else {
       $bKGD$sroa$0$8$i = $bKGD$sroa$0$7$i;$bKGD$sroa$54$6$i = $bKGD$sroa$54$5$i;
       label = 163;
      }
      if ((label|0) == 163) {
       $480 = (($bKGD$sroa$0$8$i) + ($bKGD$sroa$33$3$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
       HEAP8[$480>>0] = $470; //@line 289 "libraries/lodepng/lodepng.cpp"
       $bKGD$sroa$0$9$i = $bKGD$sroa$0$8$i;$bKGD$sroa$33$4$i = $471;$bKGD$sroa$54$7$i = $bKGD$sroa$54$6$i;
      }
      $481 = HEAP32[$467>>2]|0; //@line 4999 "libraries/lodepng/lodepng.cpp"
      $482 = $481&255; //@line 4999 "libraries/lodepng/lodepng.cpp"
      $483 = (($bKGD$sroa$33$4$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
      $484 = ($bKGD$sroa$54$7$i>>>0)<($483>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
      if ($484) {
       $485 = $bKGD$sroa$54$7$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
       $486 = ($485>>>0)<($483>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
       $487 = ($483*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
       $488 = $487 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
       $489 = $486 ? $483 : $488; //@line 227 "libraries/lodepng/lodepng.cpp"
       $490 = (_realloc($bKGD$sroa$0$9$i,$489)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       $491 = ($490|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
       if ($491) {
        $bKGD$sroa$0$11$i = $bKGD$sroa$0$9$i;$bKGD$sroa$33$5$i = $bKGD$sroa$33$4$i;$bKGD$sroa$54$9$i = $bKGD$sroa$54$7$i;
       } else {
        $bKGD$sroa$0$10$i = $490;$bKGD$sroa$54$8$i = $489;
        label = 166;
       }
      } else {
       $bKGD$sroa$0$10$i = $bKGD$sroa$0$9$i;$bKGD$sroa$54$8$i = $bKGD$sroa$54$7$i;
       label = 166;
      }
      if ((label|0) == 166) {
       $492 = (($bKGD$sroa$0$10$i) + ($bKGD$sroa$33$4$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
       HEAP8[$492>>0] = $482; //@line 289 "libraries/lodepng/lodepng.cpp"
       $bKGD$sroa$0$11$i = $bKGD$sroa$0$10$i;$bKGD$sroa$33$5$i = $483;$bKGD$sroa$54$9$i = $bKGD$sroa$54$8$i;
      }
      $493 = ((($info)) + 56|0); //@line 5000 "libraries/lodepng/lodepng.cpp"
      $494 = HEAP32[$493>>2]|0; //@line 5000 "libraries/lodepng/lodepng.cpp"
      $495 = $494 >>> 8; //@line 5000 "libraries/lodepng/lodepng.cpp"
      $496 = $495&255; //@line 5000 "libraries/lodepng/lodepng.cpp"
      $497 = (($bKGD$sroa$33$5$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
      $498 = ($bKGD$sroa$54$9$i>>>0)<($497>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
      if ($498) {
       $499 = $bKGD$sroa$54$9$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
       $500 = ($499>>>0)<($497>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
       $501 = ($497*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
       $502 = $501 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
       $503 = $500 ? $497 : $502; //@line 227 "libraries/lodepng/lodepng.cpp"
       $504 = (_realloc($bKGD$sroa$0$11$i,$503)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       $505 = ($504|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
       if ($505) {
        $bKGD$sroa$0$13$i = $bKGD$sroa$0$11$i;$bKGD$sroa$33$6$i = $bKGD$sroa$33$5$i;$bKGD$sroa$54$11$i = $bKGD$sroa$54$9$i;
       } else {
        $bKGD$sroa$0$12$i = $504;$bKGD$sroa$54$10$i = $503;
        label = 169;
       }
      } else {
       $bKGD$sroa$0$12$i = $bKGD$sroa$0$11$i;$bKGD$sroa$54$10$i = $bKGD$sroa$54$9$i;
       label = 169;
      }
      if ((label|0) == 169) {
       $506 = (($bKGD$sroa$0$12$i) + ($bKGD$sroa$33$5$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
       HEAP8[$506>>0] = $496; //@line 289 "libraries/lodepng/lodepng.cpp"
       $bKGD$sroa$0$13$i = $bKGD$sroa$0$12$i;$bKGD$sroa$33$6$i = $497;$bKGD$sroa$54$11$i = $bKGD$sroa$54$10$i;
      }
      $507 = HEAP32[$493>>2]|0; //@line 5001 "libraries/lodepng/lodepng.cpp"
      $508 = $507&255; //@line 5001 "libraries/lodepng/lodepng.cpp"
      $509 = (($bKGD$sroa$33$6$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
      $510 = ($bKGD$sroa$54$11$i>>>0)<($509>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
      if ($510) {
       $511 = $bKGD$sroa$54$11$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
       $512 = ($511>>>0)<($509>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
       $513 = ($509*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
       $514 = $513 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
       $515 = $512 ? $509 : $514; //@line 227 "libraries/lodepng/lodepng.cpp"
       $516 = (_realloc($bKGD$sroa$0$13$i,$515)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       $517 = ($516|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
       if ($517) {
        $bKGD$sroa$0$18$i = $bKGD$sroa$0$13$i;$bKGD$sroa$33$9$i = $bKGD$sroa$33$6$i;
        break L226;
       } else {
        $bKGD$sroa$0$14$i = $516;
       }
      } else {
       $bKGD$sroa$0$14$i = $bKGD$sroa$0$13$i;
      }
      $518 = (($bKGD$sroa$0$14$i) + ($bKGD$sroa$33$6$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
      HEAP8[$518>>0] = $508; //@line 289 "libraries/lodepng/lodepng.cpp"
      $bKGD$sroa$0$18$i = $bKGD$sroa$0$14$i;$bKGD$sroa$33$9$i = $509;
      break;
     }
     case 3:  {
      $519 = ((($info)) + 48|0); //@line 5005 "libraries/lodepng/lodepng.cpp"
      $520 = HEAP32[$519>>2]|0; //@line 5005 "libraries/lodepng/lodepng.cpp"
      $521 = (_realloc(0,1)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $522 = ($521|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
      if ($522) {
       $bKGD$sroa$0$18$i = 0;$bKGD$sroa$33$9$i = 0;
       break L226;
      }
      $523 = $520&255; //@line 5005 "libraries/lodepng/lodepng.cpp"
      HEAP8[$521>>0] = $523; //@line 289 "libraries/lodepng/lodepng.cpp"
      $bKGD$sroa$0$18$i = $521;$bKGD$sroa$33$9$i = 1;
      break;
     }
     default: {
      $bKGD$sroa$0$18$i = 0;$bKGD$sroa$33$9$i = 0;
     }
     }
    } while(0);
    $524 = (__Z20lodepng_chunk_createPPhPjjPKcPKh($outv,$86,$bKGD$sroa$33$9$i,502745,$bKGD$sroa$0$18$i)|0); //@line 4775 "libraries/lodepng/lodepng.cpp"
    $525 = ($524|0)==(0); //@line 4775 "libraries/lodepng/lodepng.cpp"
    if ($525) {
     $526 = HEAP32[$86>>2]|0; //@line 4776 "libraries/lodepng/lodepng.cpp"
     HEAP32[$85>>2] = $526; //@line 4776 "libraries/lodepng/lodepng.cpp"
    }
    _free($bKGD$sroa$0$18$i); //@line 76 "libraries/lodepng/lodepng.cpp"
   }
   $527 = HEAP32[$9>>2]|0; //@line 5645 "libraries/lodepng/lodepng.cpp"
   $528 = ($527|0)==(0); //@line 5645 "libraries/lodepng/lodepng.cpp"
   if (!($528)) {
    $529 = ((($info)) + 124|0); //@line 5037 "libraries/lodepng/lodepng.cpp"
    $530 = HEAP32[$529>>2]|0; //@line 5037 "libraries/lodepng/lodepng.cpp"
    $531 = (_realloc(0,4)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $532 = ($531|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    $$12$i = $532 ? 0 : 4; //@line 229 "libraries/lodepng/lodepng.cpp"
    $533 = $532 ? -4 : 0; //@line 361 "libraries/lodepng/lodepng.cpp"
    $534 = (($531) + ($533)|0); //@line 361 "libraries/lodepng/lodepng.cpp"
    $535 = $530 >>> 24; //@line 350 "libraries/lodepng/lodepng.cpp"
    $536 = $535&255; //@line 350 "libraries/lodepng/lodepng.cpp"
    HEAP8[$534>>0] = $536; //@line 350 "libraries/lodepng/lodepng.cpp"
    $537 = $530 >>> 16; //@line 351 "libraries/lodepng/lodepng.cpp"
    $538 = $537&255; //@line 351 "libraries/lodepng/lodepng.cpp"
    $539 = ((($534)) + 1|0); //@line 351 "libraries/lodepng/lodepng.cpp"
    HEAP8[$539>>0] = $538; //@line 351 "libraries/lodepng/lodepng.cpp"
    $540 = $530 >>> 8; //@line 352 "libraries/lodepng/lodepng.cpp"
    $541 = $540&255; //@line 352 "libraries/lodepng/lodepng.cpp"
    $542 = ((($534)) + 2|0); //@line 352 "libraries/lodepng/lodepng.cpp"
    HEAP8[$542>>0] = $541; //@line 352 "libraries/lodepng/lodepng.cpp"
    $543 = $530&255; //@line 353 "libraries/lodepng/lodepng.cpp"
    $544 = ((($534)) + 3|0); //@line 353 "libraries/lodepng/lodepng.cpp"
    HEAP8[$544>>0] = $543; //@line 353 "libraries/lodepng/lodepng.cpp"
    $545 = ((($info)) + 128|0); //@line 5038 "libraries/lodepng/lodepng.cpp"
    $546 = HEAP32[$545>>2]|0; //@line 5038 "libraries/lodepng/lodepng.cpp"
    $547 = (($$12$i) + 4)|0; //@line 360 "libraries/lodepng/lodepng.cpp"
    $548 = $$12$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $549 = ($548>>>0)<($547>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $550 = ($547*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $551 = $550 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $552 = $549 ? $547 : $551; //@line 227 "libraries/lodepng/lodepng.cpp"
    $553 = (_realloc($531,$552)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $554 = ($553|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    $$$i82 = $554 ? $531 : $553; //@line 229 "libraries/lodepng/lodepng.cpp"
    $$12$$i = $554 ? $$12$i : $552; //@line 229 "libraries/lodepng/lodepng.cpp"
    $$12$16$i = $554 ? $$12$i : $547; //@line 229 "libraries/lodepng/lodepng.cpp"
    $555 = (($$12$16$i) + -4)|0; //@line 361 "libraries/lodepng/lodepng.cpp"
    $556 = (($$$i82) + ($555)|0); //@line 361 "libraries/lodepng/lodepng.cpp"
    $557 = $546 >>> 24; //@line 350 "libraries/lodepng/lodepng.cpp"
    $558 = $557&255; //@line 350 "libraries/lodepng/lodepng.cpp"
    HEAP8[$556>>0] = $558; //@line 350 "libraries/lodepng/lodepng.cpp"
    $559 = $546 >>> 16; //@line 351 "libraries/lodepng/lodepng.cpp"
    $560 = $559&255; //@line 351 "libraries/lodepng/lodepng.cpp"
    $561 = ((($556)) + 1|0); //@line 351 "libraries/lodepng/lodepng.cpp"
    HEAP8[$561>>0] = $560; //@line 351 "libraries/lodepng/lodepng.cpp"
    $562 = $546 >>> 8; //@line 352 "libraries/lodepng/lodepng.cpp"
    $563 = $562&255; //@line 352 "libraries/lodepng/lodepng.cpp"
    $564 = ((($556)) + 2|0); //@line 352 "libraries/lodepng/lodepng.cpp"
    HEAP8[$564>>0] = $563; //@line 352 "libraries/lodepng/lodepng.cpp"
    $565 = $546&255; //@line 353 "libraries/lodepng/lodepng.cpp"
    $566 = ((($556)) + 3|0); //@line 353 "libraries/lodepng/lodepng.cpp"
    HEAP8[$566>>0] = $565; //@line 353 "libraries/lodepng/lodepng.cpp"
    $567 = ((($info)) + 132|0); //@line 5039 "libraries/lodepng/lodepng.cpp"
    $568 = HEAP32[$567>>2]|0; //@line 5039 "libraries/lodepng/lodepng.cpp"
    $569 = $568&255; //@line 5039 "libraries/lodepng/lodepng.cpp"
    $570 = $$12$16$i | 1; //@line 288 "libraries/lodepng/lodepng.cpp"
    $571 = ($$12$$i>>>0)<($570>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
    if ($571) {
     $572 = $$12$$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
     $573 = ($572>>>0)<($570>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
     $574 = ($570*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
     $575 = $574 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
     $576 = $573 ? $570 : $575; //@line 227 "libraries/lodepng/lodepng.cpp"
     $577 = (_realloc($$$i82,$576)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
     $578 = ($577|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
     if ($578) {
      $data$sroa$0$5$i85 = $$$i82;$data$sroa$19$2$i = $$12$16$i;
     } else {
      $data$sroa$0$4$i83 = $577;
      label = 181;
     }
    } else {
     $data$sroa$0$4$i83 = $$$i82;
     label = 181;
    }
    if ((label|0) == 181) {
     $579 = (($data$sroa$0$4$i83) + ($$12$16$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
     HEAP8[$579>>0] = $569; //@line 289 "libraries/lodepng/lodepng.cpp"
     $data$sroa$0$5$i85 = $data$sroa$0$4$i83;$data$sroa$19$2$i = $570;
    }
    $580 = (__Z20lodepng_chunk_createPPhPjjPKcPKh($outv,$86,$data$sroa$19$2$i,502750,$data$sroa$0$5$i85)|0); //@line 4775 "libraries/lodepng/lodepng.cpp"
    $581 = ($580|0)==(0); //@line 4775 "libraries/lodepng/lodepng.cpp"
    if ($581) {
     $582 = HEAP32[$86>>2]|0; //@line 4776 "libraries/lodepng/lodepng.cpp"
     HEAP32[$85>>2] = $582; //@line 4776 "libraries/lodepng/lodepng.cpp"
    }
    _free($data$sroa$0$5$i85); //@line 76 "libraries/lodepng/lodepng.cpp"
   }
   $583 = ((($info)) + 140|0); //@line 5648 "libraries/lodepng/lodepng.cpp"
   $584 = HEAP32[$583>>2]|0; //@line 5648 "libraries/lodepng/lodepng.cpp"
   $585 = ($584|0)==(0|0); //@line 5648 "libraries/lodepng/lodepng.cpp"
   if (!($585)) {
    $586 = ((($info)) + 152|0); //@line 5650 "libraries/lodepng/lodepng.cpp"
    $587 = HEAP32[$586>>2]|0; //@line 5650 "libraries/lodepng/lodepng.cpp"
    $588 = $584; //@line 5535 "libraries/lodepng/lodepng.cpp"
    $589 = ($587|0)==(0); //@line 5535 "libraries/lodepng/lodepng.cpp"
    L277: do {
     if (!($589)) {
      $$pre158 = HEAP32[$86>>2]|0; //@line 2469 "libraries/lodepng/lodepng.cpp"
      $609 = $$pre158;$inchunk$02$i89 = $584;
      while(1) {
       $590 = HEAP8[$inchunk$02$i89>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $591 = $590&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $592 = $591 << 24; //@line 343 "libraries/lodepng/lodepng.cpp"
       $593 = ((($inchunk$02$i89)) + 1|0); //@line 343 "libraries/lodepng/lodepng.cpp"
       $594 = HEAP8[$593>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $595 = $594&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $596 = $595 << 16; //@line 343 "libraries/lodepng/lodepng.cpp"
       $597 = $596 | $592; //@line 343 "libraries/lodepng/lodepng.cpp"
       $598 = ((($inchunk$02$i89)) + 2|0); //@line 343 "libraries/lodepng/lodepng.cpp"
       $599 = HEAP8[$598>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $600 = $599&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $601 = $600 << 8; //@line 343 "libraries/lodepng/lodepng.cpp"
       $602 = $597 | $601; //@line 343 "libraries/lodepng/lodepng.cpp"
       $603 = ((($inchunk$02$i89)) + 3|0); //@line 343 "libraries/lodepng/lodepng.cpp"
       $604 = HEAP8[$603>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $605 = $604&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $606 = $602 | $605; //@line 343 "libraries/lodepng/lodepng.cpp"
       $607 = (($606) + 12)|0; //@line 2467 "libraries/lodepng/lodepng.cpp"
       $608 = (($607) + ($609))|0; //@line 5538 "libraries/lodepng/lodepng.cpp"
       $610 = ($608>>>0)<($607>>>0); //@line 2470 "libraries/lodepng/lodepng.cpp"
       $611 = ($608>>>0)<($609>>>0); //@line 2470 "libraries/lodepng/lodepng.cpp"
       $or$cond$i93 = $610 | $611; //@line 2470 "libraries/lodepng/lodepng.cpp"
       if ($or$cond$i93) {
        $$2$i90 = 77;
        break;
       }
       $612 = HEAP32[$outv>>2]|0; //@line 2472 "libraries/lodepng/lodepng.cpp"
       $613 = (_realloc($612,$608)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       $614 = ($613|0)==(0|0); //@line 2473 "libraries/lodepng/lodepng.cpp"
       if ($614) {
        $$2$i90 = 83;
        break;
       }
       HEAP32[$outv>>2] = $613; //@line 2474 "libraries/lodepng/lodepng.cpp"
       HEAP32[$86>>2] = $608; //@line 2475 "libraries/lodepng/lodepng.cpp"
       $615 = (($613) + ($609)|0); //@line 2476 "libraries/lodepng/lodepng.cpp"
       $616 = ($607|0)==(0); //@line 2478 "libraries/lodepng/lodepng.cpp"
       if (!($616)) {
        $i$01$i94 = 0;
        while(1) {
         $617 = (($inchunk$02$i89) + ($i$01$i94)|0); //@line 2478 "libraries/lodepng/lodepng.cpp"
         $618 = HEAP8[$617>>0]|0; //@line 2478 "libraries/lodepng/lodepng.cpp"
         $619 = (($615) + ($i$01$i94)|0); //@line 2478 "libraries/lodepng/lodepng.cpp"
         HEAP8[$619>>0] = $618; //@line 2478 "libraries/lodepng/lodepng.cpp"
         $620 = (($i$01$i94) + 1)|0; //@line 2478 "libraries/lodepng/lodepng.cpp"
         $621 = ($620|0)==($607|0); //@line 2478 "libraries/lodepng/lodepng.cpp"
         if ($621) {
          break;
         } else {
          $i$01$i94 = $620;
         }
        }
       }
       HEAP32[$85>>2] = $608; //@line 5538 "libraries/lodepng/lodepng.cpp"
       $622 = HEAP8[$inchunk$02$i89>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $623 = $622&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $624 = $623 << 24; //@line 343 "libraries/lodepng/lodepng.cpp"
       $625 = HEAP8[$593>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $626 = $625&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $627 = $626 << 16; //@line 343 "libraries/lodepng/lodepng.cpp"
       $628 = $627 | $624; //@line 343 "libraries/lodepng/lodepng.cpp"
       $629 = HEAP8[$598>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $630 = $629&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $631 = $630 << 8; //@line 343 "libraries/lodepng/lodepng.cpp"
       $632 = $628 | $631; //@line 343 "libraries/lodepng/lodepng.cpp"
       $633 = HEAP8[$603>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $634 = $633&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $635 = $632 | $634; //@line 343 "libraries/lodepng/lodepng.cpp"
       $636 = (($635) + 12)|0; //@line 2454 "libraries/lodepng/lodepng.cpp"
       $637 = (($inchunk$02$i89) + ($636)|0); //@line 2455 "libraries/lodepng/lodepng.cpp"
       $638 = $637; //@line 5535 "libraries/lodepng/lodepng.cpp"
       $639 = (($638) - ($588))|0; //@line 5535 "libraries/lodepng/lodepng.cpp"
       $640 = ($639>>>0)<($587>>>0); //@line 5535 "libraries/lodepng/lodepng.cpp"
       if ($640) {
        $609 = $608;$inchunk$02$i89 = $637;
       } else {
        break L277;
       }
      }
      HEAP32[$0>>2] = $$2$i90; //@line 5650 "libraries/lodepng/lodepng.cpp"
      break L71;
     }
    } while(0);
    HEAP32[$0>>2] = 0; //@line 5650 "libraries/lodepng/lodepng.cpp"
   }
   $641 = HEAP32[$data>>2]|0; //@line 5655 "libraries/lodepng/lodepng.cpp"
   $642 = HEAP32[$datasize>>2]|0; //@line 5655 "libraries/lodepng/lodepng.cpp"
   HEAP32[$zlibdata$i>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
   $643 = ((($zlibdata$i)) + 8|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   HEAP32[$643>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
   $644 = ((($zlibdata$i)) + 4|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   HEAP32[$644>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
   $645 = ((($state)) + 60|0); //@line 2209 "libraries/lodepng/lodepng.cpp"
   $646 = HEAP32[$645>>2]|0; //@line 2209 "libraries/lodepng/lodepng.cpp"
   $647 = ($646|0)==(0|0); //@line 2209 "libraries/lodepng/lodepng.cpp"
   if ($647) {
    $649 = (__Z21lodepng_zlib_compressPPhPjPKhjPK23LodePNGCompressSettings($zlibdata$i,$644,$641,$642,$20)|0); //@line 2215 "libraries/lodepng/lodepng.cpp"
    $$0$i$i99 = $649;
   } else {
    $648 = (FUNCTION_TABLE_iiiiii[$646 & 8191]($zlibdata$i,$644,$641,$642,$20)|0); //@line 2211 "libraries/lodepng/lodepng.cpp"
    $$0$i$i99 = $648;
   }
   $650 = ($$0$i$i99|0)==(0); //@line 4885 "libraries/lodepng/lodepng.cpp"
   do {
    if ($650) {
     $651 = HEAP32[$zlibdata$i>>2]|0; //@line 4885 "libraries/lodepng/lodepng.cpp"
     $652 = HEAP32[$644>>2]|0; //@line 4885 "libraries/lodepng/lodepng.cpp"
     $653 = (__Z20lodepng_chunk_createPPhPjjPKcPKh($outv,$86,$652,502725,$651)|0); //@line 4775 "libraries/lodepng/lodepng.cpp"
     $654 = ($653|0)==(0); //@line 4775 "libraries/lodepng/lodepng.cpp"
     if (!($654)) {
      $error$0$i = $653;
      break;
     }
     $655 = HEAP32[$86>>2]|0; //@line 4776 "libraries/lodepng/lodepng.cpp"
     HEAP32[$85>>2] = $655; //@line 4776 "libraries/lodepng/lodepng.cpp"
     $error$0$i = 0;
    } else {
     $error$0$i = $$0$i$i99;
    }
   } while(0);
   HEAP32[$643>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
   HEAP32[$644>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
   $656 = HEAP32[$zlibdata$i>>2]|0; //@line 252 "libraries/lodepng/lodepng.cpp"
   _free($656); //@line 76 "libraries/lodepng/lodepng.cpp"
   HEAP32[$0>>2] = $error$0$i; //@line 5655 "libraries/lodepng/lodepng.cpp"
   $657 = ($error$0$i|0)==(0); //@line 5656 "libraries/lodepng/lodepng.cpp"
   if (!($657)) {
    break;
   }
   $658 = ((($info)) + 92|0); //@line 5659 "libraries/lodepng/lodepng.cpp"
   $659 = HEAP32[$658>>2]|0; //@line 5659 "libraries/lodepng/lodepng.cpp"
   $660 = ($659|0)==(0); //@line 5659 "libraries/lodepng/lodepng.cpp"
   do {
    if (!($660)) {
     $661 = (_malloc(7)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
     $662 = ($661|0)==(0|0); //@line 5018 "libraries/lodepng/lodepng.cpp"
     if ($662) {
      break;
     }
     $663 = ((($info)) + 96|0); //@line 5019 "libraries/lodepng/lodepng.cpp"
     $664 = HEAP32[$663>>2]|0; //@line 5019 "libraries/lodepng/lodepng.cpp"
     $665 = $664 >>> 8; //@line 5019 "libraries/lodepng/lodepng.cpp"
     $666 = $665&255; //@line 5019 "libraries/lodepng/lodepng.cpp"
     HEAP8[$661>>0] = $666; //@line 5019 "libraries/lodepng/lodepng.cpp"
     $667 = $664&255; //@line 5020 "libraries/lodepng/lodepng.cpp"
     $668 = ((($661)) + 1|0); //@line 5020 "libraries/lodepng/lodepng.cpp"
     HEAP8[$668>>0] = $667; //@line 5020 "libraries/lodepng/lodepng.cpp"
     $669 = ((($info)) + 100|0); //@line 5021 "libraries/lodepng/lodepng.cpp"
     $670 = HEAP32[$669>>2]|0; //@line 5021 "libraries/lodepng/lodepng.cpp"
     $671 = $670&255; //@line 5021 "libraries/lodepng/lodepng.cpp"
     $672 = ((($661)) + 2|0); //@line 5021 "libraries/lodepng/lodepng.cpp"
     HEAP8[$672>>0] = $671; //@line 5021 "libraries/lodepng/lodepng.cpp"
     $673 = ((($info)) + 104|0); //@line 5022 "libraries/lodepng/lodepng.cpp"
     $674 = HEAP32[$673>>2]|0; //@line 5022 "libraries/lodepng/lodepng.cpp"
     $675 = $674&255; //@line 5022 "libraries/lodepng/lodepng.cpp"
     $676 = ((($661)) + 3|0); //@line 5022 "libraries/lodepng/lodepng.cpp"
     HEAP8[$676>>0] = $675; //@line 5022 "libraries/lodepng/lodepng.cpp"
     $677 = ((($info)) + 108|0); //@line 5023 "libraries/lodepng/lodepng.cpp"
     $678 = HEAP32[$677>>2]|0; //@line 5023 "libraries/lodepng/lodepng.cpp"
     $679 = $678&255; //@line 5023 "libraries/lodepng/lodepng.cpp"
     $680 = ((($661)) + 4|0); //@line 5023 "libraries/lodepng/lodepng.cpp"
     HEAP8[$680>>0] = $679; //@line 5023 "libraries/lodepng/lodepng.cpp"
     $681 = ((($info)) + 112|0); //@line 5024 "libraries/lodepng/lodepng.cpp"
     $682 = HEAP32[$681>>2]|0; //@line 5024 "libraries/lodepng/lodepng.cpp"
     $683 = $682&255; //@line 5024 "libraries/lodepng/lodepng.cpp"
     $684 = ((($661)) + 5|0); //@line 5024 "libraries/lodepng/lodepng.cpp"
     HEAP8[$684>>0] = $683; //@line 5024 "libraries/lodepng/lodepng.cpp"
     $685 = ((($info)) + 116|0); //@line 5025 "libraries/lodepng/lodepng.cpp"
     $686 = HEAP32[$685>>2]|0; //@line 5025 "libraries/lodepng/lodepng.cpp"
     $687 = $686&255; //@line 5025 "libraries/lodepng/lodepng.cpp"
     $688 = ((($661)) + 6|0); //@line 5025 "libraries/lodepng/lodepng.cpp"
     HEAP8[$688>>0] = $687; //@line 5025 "libraries/lodepng/lodepng.cpp"
     $689 = (__Z20lodepng_chunk_createPPhPjjPKcPKh($outv,$86,7,502755,$661)|0); //@line 4775 "libraries/lodepng/lodepng.cpp"
     $690 = ($689|0)==(0); //@line 4775 "libraries/lodepng/lodepng.cpp"
     if ($690) {
      $691 = HEAP32[$86>>2]|0; //@line 4776 "libraries/lodepng/lodepng.cpp"
      HEAP32[$85>>2] = $691; //@line 4776 "libraries/lodepng/lodepng.cpp"
     }
     _free($661); //@line 76 "libraries/lodepng/lodepng.cpp"
    }
   } while(0);
   $692 = ((($info)) + 60|0); //@line 5661 "libraries/lodepng/lodepng.cpp"
   $693 = HEAP32[$692>>2]|0; //@line 5661 "libraries/lodepng/lodepng.cpp"
   $694 = ($693|0)==(0); //@line 5661 "libraries/lodepng/lodepng.cpp"
   L306: do {
    if ($694) {
     $784 = 0;
    } else {
     $695 = ((($info)) + 64|0); //@line 5663 "libraries/lodepng/lodepng.cpp"
     $696 = ((($state)) + 96|0); //@line 5673 "libraries/lodepng/lodepng.cpp"
     $697 = ((($info)) + 68|0); //@line 5675 "libraries/lodepng/lodepng.cpp"
     $698 = ((($compressed$i)) + 8|0); //@line 259 "libraries/lodepng/lodepng.cpp"
     $699 = ((($compressed$i)) + 4|0); //@line 259 "libraries/lodepng/lodepng.cpp"
     $1036 = $693;$i$0150 = 0;
     while(1) {
      $700 = HEAP32[$695>>2]|0; //@line 5663 "libraries/lodepng/lodepng.cpp"
      $701 = (($700) + ($i$0150<<2)|0); //@line 5663 "libraries/lodepng/lodepng.cpp"
      $702 = HEAP32[$701>>2]|0; //@line 5663 "libraries/lodepng/lodepng.cpp"
      $703 = (_strlen($702)|0); //@line 5663 "libraries/lodepng/lodepng.cpp"
      $704 = ($703>>>0)>(79); //@line 5663 "libraries/lodepng/lodepng.cpp"
      if ($704) {
       $$lcssa185 = $1036;
       label = 210;
       break;
      }
      $strlenfirst2 = HEAP8[$702>>0]|0; //@line 5668 "libraries/lodepng/lodepng.cpp"
      $705 = ($strlenfirst2<<24>>24)==(0); //@line 5668 "libraries/lodepng/lodepng.cpp"
      if ($705) {
       $$lcssa186 = $1036;
       label = 212;
       break;
      }
      $706 = HEAP32[$696>>2]|0; //@line 5673 "libraries/lodepng/lodepng.cpp"
      $707 = ($706|0)==(0); //@line 5673 "libraries/lodepng/lodepng.cpp"
      $708 = HEAP32[$697>>2]|0; //@line 5675 "libraries/lodepng/lodepng.cpp"
      $709 = (($708) + ($i$0150<<2)|0); //@line 5675 "libraries/lodepng/lodepng.cpp"
      $710 = HEAP32[$709>>2]|0; //@line 5675 "libraries/lodepng/lodepng.cpp"
      if ($707) {
       __ZL13addChunk_tEXtP8ucvectorPKcS2_($outv,$702,$710); //@line 5679 "libraries/lodepng/lodepng.cpp"
      } else {
       $711 = (_strlen($710)|0); //@line 4921 "libraries/lodepng/lodepng.cpp"
       HEAP32[$compressed$i>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
       HEAP32[$698>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
       HEAP32[$699>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
       $712 = HEAP8[$702>>0]|0; //@line 4925 "libraries/lodepng/lodepng.cpp"
       $713 = ($712<<24>>24)==(0); //@line 4925 "libraries/lodepng/lodepng.cpp"
       do {
        if (!($713)) {
         $724 = $712;$data$sroa$0$023$i = 0;$data$sroa$18$022$i = 0;$data$sroa$29$021$i = 0;$i$024$i = 0;
         while(1) {
          $714 = (($data$sroa$18$022$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
          $715 = ($data$sroa$29$021$i>>>0)<($714>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
          if ($715) {
           $716 = $data$sroa$29$021$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
           $717 = ($716>>>0)<($714>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
           $718 = ($714*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
           $719 = $718 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
           $720 = $717 ? $714 : $719; //@line 227 "libraries/lodepng/lodepng.cpp"
           $721 = (_realloc($data$sroa$0$023$i,$720)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
           $722 = ($721|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
           if ($722) {
            $data$sroa$0$2$i105 = $data$sroa$0$023$i;$data$sroa$18$1$i = $data$sroa$18$022$i;$data$sroa$29$2$i = $data$sroa$29$021$i;
           } else {
            $data$sroa$0$1$i103 = $721;$data$sroa$29$1$i = $720;
            label = 217;
           }
          } else {
           $data$sroa$0$1$i103 = $data$sroa$0$023$i;$data$sroa$29$1$i = $data$sroa$29$021$i;
           label = 217;
          }
          if ((label|0) == 217) {
           label = 0;
           $723 = (($data$sroa$0$1$i103) + ($data$sroa$18$022$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
           HEAP8[$723>>0] = $724; //@line 289 "libraries/lodepng/lodepng.cpp"
           $data$sroa$0$2$i105 = $data$sroa$0$1$i103;$data$sroa$18$1$i = $714;$data$sroa$29$2$i = $data$sroa$29$1$i;
          }
          $725 = (($i$024$i) + 1)|0; //@line 4925 "libraries/lodepng/lodepng.cpp"
          $726 = (($702) + ($725)|0); //@line 4925 "libraries/lodepng/lodepng.cpp"
          $727 = HEAP8[$726>>0]|0; //@line 4925 "libraries/lodepng/lodepng.cpp"
          $728 = ($727<<24>>24)==(0); //@line 4925 "libraries/lodepng/lodepng.cpp"
          if ($728) {
           $data$sroa$0$2$i105$lcssa = $data$sroa$0$2$i105;$data$sroa$18$1$i$lcssa = $data$sroa$18$1$i;$data$sroa$29$2$i$lcssa = $data$sroa$29$2$i;$i$024$i$lcssa = $i$024$i;
           break;
          } else {
           $724 = $727;$data$sroa$0$023$i = $data$sroa$0$2$i105;$data$sroa$18$022$i = $data$sroa$18$1$i;$data$sroa$29$021$i = $data$sroa$29$2$i;$i$024$i = $725;
          }
         }
         $729 = ($i$024$i$lcssa>>>0)>(78); //@line 4926 "libraries/lodepng/lodepng.cpp"
         if ($729) {
          break;
         }
         $730 = (($data$sroa$18$1$i$lcssa) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
         $731 = ($data$sroa$29$2$i$lcssa>>>0)<($730>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
         if ($731) {
          $732 = $data$sroa$29$2$i$lcssa << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
          $733 = ($732>>>0)<($730>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
          $734 = ($730*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
          $735 = $734 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
          $736 = $733 ? $730 : $735; //@line 227 "libraries/lodepng/lodepng.cpp"
          $737 = (_realloc($data$sroa$0$2$i105$lcssa,$736)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
          $738 = ($737|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
          if ($738) {
           $data$sroa$0$4$i109 = $data$sroa$0$2$i105$lcssa;$data$sroa$18$2$i = $data$sroa$18$1$i$lcssa;$data$sroa$29$4$i = $data$sroa$29$2$i$lcssa;
          } else {
           $data$sroa$0$3$i107 = $737;$data$sroa$29$3$i = $736;
           label = 222;
          }
         } else {
          $data$sroa$0$3$i107 = $data$sroa$0$2$i105$lcssa;$data$sroa$29$3$i = $data$sroa$29$2$i$lcssa;
          label = 222;
         }
         if ((label|0) == 222) {
          label = 0;
          $739 = (($data$sroa$0$3$i107) + ($data$sroa$18$1$i$lcssa)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
          HEAP8[$739>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
          $data$sroa$0$4$i109 = $data$sroa$0$3$i107;$data$sroa$18$2$i = $730;$data$sroa$29$4$i = $data$sroa$29$3$i;
         }
         $740 = (($data$sroa$18$2$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
         $741 = ($data$sroa$29$4$i>>>0)<($740>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
         if ($741) {
          $742 = $data$sroa$29$4$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
          $743 = ($742>>>0)<($740>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
          $744 = ($740*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
          $745 = $744 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
          $746 = $743 ? $740 : $745; //@line 227 "libraries/lodepng/lodepng.cpp"
          $747 = (_realloc($data$sroa$0$4$i109,$746)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
          $748 = ($747|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
          if ($748) {
           $data$sroa$0$6$i113 = $data$sroa$0$4$i109;$data$sroa$18$3$i = $data$sroa$18$2$i;$data$sroa$29$6$i = $data$sroa$29$4$i;
          } else {
           $data$sroa$0$5$i111 = $747;$data$sroa$29$5$i = $746;
           label = 225;
          }
         } else {
          $data$sroa$0$5$i111 = $data$sroa$0$4$i109;$data$sroa$29$5$i = $data$sroa$29$4$i;
          label = 225;
         }
         if ((label|0) == 225) {
          label = 0;
          $749 = (($data$sroa$0$5$i111) + ($data$sroa$18$2$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
          HEAP8[$749>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
          $data$sroa$0$6$i113 = $data$sroa$0$5$i111;$data$sroa$18$3$i = $740;$data$sroa$29$6$i = $data$sroa$29$5$i;
         }
         $750 = HEAP32[$645>>2]|0; //@line 2209 "libraries/lodepng/lodepng.cpp"
         $751 = ($750|0)==(0|0); //@line 2209 "libraries/lodepng/lodepng.cpp"
         if ($751) {
          $753 = (__Z21lodepng_zlib_compressPPhPjPKhjPK23LodePNGCompressSettings($compressed$i,$699,$710,$711,$20)|0); //@line 2215 "libraries/lodepng/lodepng.cpp"
          $$0$i$i115 = $753;
         } else {
          $752 = (FUNCTION_TABLE_iiiiii[$750 & 8191]($compressed$i,$699,$710,$711,$20)|0); //@line 2211 "libraries/lodepng/lodepng.cpp"
          $$0$i$i115 = $752;
         }
         $754 = ($$0$i$i115|0)==(0); //@line 4932 "libraries/lodepng/lodepng.cpp"
         do {
          if ($754) {
           $755 = HEAP32[$699>>2]|0; //@line 4934 "libraries/lodepng/lodepng.cpp"
           $756 = ($755|0)==(0); //@line 4934 "libraries/lodepng/lodepng.cpp"
           if ($756) {
            $data$sroa$0$7$lcssa$i = $data$sroa$0$6$i113;$data$sroa$18$4$lcssa$i = $data$sroa$18$3$i;
           } else {
            $data$sroa$0$718$i = $data$sroa$0$6$i113;$data$sroa$18$417$i = $data$sroa$18$3$i;$data$sroa$29$716$i = $data$sroa$29$6$i;$i$119$i = 0;
            while(1) {
             $757 = HEAP32[$compressed$i>>2]|0; //@line 4934 "libraries/lodepng/lodepng.cpp"
             $758 = (($757) + ($i$119$i)|0); //@line 4934 "libraries/lodepng/lodepng.cpp"
             $759 = HEAP8[$758>>0]|0; //@line 4934 "libraries/lodepng/lodepng.cpp"
             $760 = (($data$sroa$18$417$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
             $761 = ($data$sroa$29$716$i>>>0)<($760>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
             if ($761) {
              $762 = $data$sroa$29$716$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
              $763 = ($762>>>0)<($760>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
              $764 = ($760*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
              $765 = $764 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
              $766 = $763 ? $760 : $765; //@line 227 "libraries/lodepng/lodepng.cpp"
              $767 = (_realloc($data$sroa$0$718$i,$766)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
              $768 = ($767|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
              if ($768) {
               $data$sroa$0$9$i = $data$sroa$0$718$i;$data$sroa$18$5$i = $data$sroa$18$417$i;$data$sroa$29$9$i = $data$sroa$29$716$i;
              } else {
               $data$sroa$0$8$i = $767;$data$sroa$29$8$i = $766;
               label = 233;
              }
             } else {
              $data$sroa$0$8$i = $data$sroa$0$718$i;$data$sroa$29$8$i = $data$sroa$29$716$i;
              label = 233;
             }
             if ((label|0) == 233) {
              label = 0;
              $769 = (($data$sroa$0$8$i) + ($data$sroa$18$417$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
              HEAP8[$769>>0] = $759; //@line 289 "libraries/lodepng/lodepng.cpp"
              $data$sroa$0$9$i = $data$sroa$0$8$i;$data$sroa$18$5$i = $760;$data$sroa$29$9$i = $data$sroa$29$8$i;
             }
             $770 = (($i$119$i) + 1)|0; //@line 4934 "libraries/lodepng/lodepng.cpp"
             $771 = HEAP32[$699>>2]|0; //@line 4934 "libraries/lodepng/lodepng.cpp"
             $772 = ($770|0)==($771|0); //@line 4934 "libraries/lodepng/lodepng.cpp"
             if ($772) {
              $data$sroa$0$7$lcssa$i = $data$sroa$0$9$i;$data$sroa$18$4$lcssa$i = $data$sroa$18$5$i;
              break;
             } else {
              $data$sroa$0$718$i = $data$sroa$0$9$i;$data$sroa$18$417$i = $data$sroa$18$5$i;$data$sroa$29$716$i = $data$sroa$29$9$i;$i$119$i = $770;
             }
            }
           }
           $773 = (__Z20lodepng_chunk_createPPhPjjPKcPKh($outv,$86,$data$sroa$18$4$lcssa$i,502760,$data$sroa$0$7$lcssa$i)|0); //@line 4775 "libraries/lodepng/lodepng.cpp"
           $774 = ($773|0)==(0); //@line 4775 "libraries/lodepng/lodepng.cpp"
           if (!($774)) {
            $data$sroa$0$10$i122 = $data$sroa$0$7$lcssa$i;
            break;
           }
           $775 = HEAP32[$86>>2]|0; //@line 4776 "libraries/lodepng/lodepng.cpp"
           HEAP32[$85>>2] = $775; //@line 4776 "libraries/lodepng/lodepng.cpp"
           $data$sroa$0$10$i122 = $data$sroa$0$7$lcssa$i;
          } else {
           $data$sroa$0$10$i122 = $data$sroa$0$6$i113;
          }
         } while(0);
         HEAP32[$698>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
         HEAP32[$699>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
         $776 = HEAP32[$compressed$i>>2]|0; //@line 252 "libraries/lodepng/lodepng.cpp"
         _free($776); //@line 76 "libraries/lodepng/lodepng.cpp"
         HEAP32[$compressed$i>>2] = 0; //@line 253 "libraries/lodepng/lodepng.cpp"
         _free($data$sroa$0$10$i122); //@line 76 "libraries/lodepng/lodepng.cpp"
        }
       } while(0);
      }
      $777 = (($i$0150) + 1)|0; //@line 5661 "libraries/lodepng/lodepng.cpp"
      $778 = HEAP32[$692>>2]|0; //@line 5661 "libraries/lodepng/lodepng.cpp"
      $779 = ($777|0)==($778|0); //@line 5661 "libraries/lodepng/lodepng.cpp"
      if ($779) {
       $784 = $777;
       break L306;
      } else {
       $1036 = $778;$i$0150 = $777;
      }
     }
     if ((label|0) == 210) {
      HEAP32[$0>>2] = 66; //@line 5665 "libraries/lodepng/lodepng.cpp"
      $784 = $$lcssa185;
      break;
     }
     else if ((label|0) == 212) {
      HEAP32[$0>>2] = 67; //@line 5670 "libraries/lodepng/lodepng.cpp"
      $784 = $$lcssa186;
      break;
     }
    }
   } while(0);
   $780 = ((($state)) + 92|0); //@line 5683 "libraries/lodepng/lodepng.cpp"
   $781 = HEAP32[$780>>2]|0; //@line 5683 "libraries/lodepng/lodepng.cpp"
   $782 = ($781|0)==(0); //@line 5683 "libraries/lodepng/lodepng.cpp"
   L358: do {
    if (!($782)) {
     $783 = ($784|0)==(0); //@line 5686 "libraries/lodepng/lodepng.cpp"
     if (!($783)) {
      $785 = ((($info)) + 64|0); //@line 5688 "libraries/lodepng/lodepng.cpp"
      $786 = HEAP32[$785>>2]|0; //@line 5688 "libraries/lodepng/lodepng.cpp"
      $i$1148 = 0;
      while(1) {
       $789 = (($786) + ($i$1148<<2)|0); //@line 5688 "libraries/lodepng/lodepng.cpp"
       $790 = HEAP32[$789>>2]|0; //@line 5688 "libraries/lodepng/lodepng.cpp"
       $791 = (_strcmp($790,502770)|0); //@line 5688 "libraries/lodepng/lodepng.cpp"
       $792 = ($791|0)==(0); //@line 5688 "libraries/lodepng/lodepng.cpp"
       $788 = (($i$1148) + 1)|0; //@line 5686 "libraries/lodepng/lodepng.cpp"
       if ($792) {
        break L358;
       }
       $787 = ($788|0)==($784|0); //@line 5686 "libraries/lodepng/lodepng.cpp"
       if ($787) {
        break;
       } else {
        $i$1148 = $788;
       }
      }
     }
     __ZL13addChunk_tEXtP8ucvectorPKcS2_($outv,502770,502778); //@line 5696 "libraries/lodepng/lodepng.cpp"
    }
   } while(0);
   $793 = ((($info)) + 72|0); //@line 5700 "libraries/lodepng/lodepng.cpp"
   $794 = HEAP32[$793>>2]|0; //@line 5700 "libraries/lodepng/lodepng.cpp"
   $795 = ($794|0)==(0); //@line 5700 "libraries/lodepng/lodepng.cpp"
   L367: do {
    if (!($795)) {
     $796 = ((($info)) + 76|0); //@line 5702 "libraries/lodepng/lodepng.cpp"
     $797 = ((($state)) + 96|0); //@line 5712 "libraries/lodepng/lodepng.cpp"
     $798 = ((($info)) + 80|0); //@line 5713 "libraries/lodepng/lodepng.cpp"
     $799 = ((($info)) + 84|0); //@line 5713 "libraries/lodepng/lodepng.cpp"
     $800 = ((($info)) + 88|0); //@line 5713 "libraries/lodepng/lodepng.cpp"
     $801 = ((($compressed_data$i)) + 8|0); //@line 259 "libraries/lodepng/lodepng.cpp"
     $802 = ((($compressed_data$i)) + 4|0); //@line 259 "libraries/lodepng/lodepng.cpp"
     $i$2147 = 0;
     while(1) {
      $803 = HEAP32[$796>>2]|0; //@line 5702 "libraries/lodepng/lodepng.cpp"
      $804 = (($803) + ($i$2147<<2)|0); //@line 5702 "libraries/lodepng/lodepng.cpp"
      $805 = HEAP32[$804>>2]|0; //@line 5702 "libraries/lodepng/lodepng.cpp"
      $806 = (_strlen($805)|0); //@line 5702 "libraries/lodepng/lodepng.cpp"
      $807 = ($806>>>0)>(79); //@line 5702 "libraries/lodepng/lodepng.cpp"
      if ($807) {
       label = 250;
       break;
      }
      $strlenfirst = HEAP8[$805>>0]|0; //@line 5707 "libraries/lodepng/lodepng.cpp"
      $808 = ($strlenfirst<<24>>24)==(0); //@line 5707 "libraries/lodepng/lodepng.cpp"
      if ($808) {
       label = 252;
       break;
      }
      $809 = HEAP32[$797>>2]|0; //@line 5712 "libraries/lodepng/lodepng.cpp"
      $810 = HEAP32[$798>>2]|0; //@line 5713 "libraries/lodepng/lodepng.cpp"
      $811 = (($810) + ($i$2147<<2)|0); //@line 5713 "libraries/lodepng/lodepng.cpp"
      $812 = HEAP32[$811>>2]|0; //@line 5713 "libraries/lodepng/lodepng.cpp"
      $813 = HEAP32[$799>>2]|0; //@line 5713 "libraries/lodepng/lodepng.cpp"
      $814 = (($813) + ($i$2147<<2)|0); //@line 5713 "libraries/lodepng/lodepng.cpp"
      $815 = HEAP32[$814>>2]|0; //@line 5713 "libraries/lodepng/lodepng.cpp"
      $816 = HEAP32[$800>>2]|0; //@line 5713 "libraries/lodepng/lodepng.cpp"
      $817 = (($816) + ($i$2147<<2)|0); //@line 5713 "libraries/lodepng/lodepng.cpp"
      $818 = HEAP32[$817>>2]|0; //@line 5713 "libraries/lodepng/lodepng.cpp"
      $819 = (_strlen($818)|0); //@line 4948 "libraries/lodepng/lodepng.cpp"
      $830 = $strlenfirst;$data$sroa$0$084$i = 0;$data$sroa$36$083$i = 0;$data$sroa$59$082$i = 0;$i$085$i = 0;
      while(1) {
       $820 = (($data$sroa$36$083$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
       $821 = ($data$sroa$59$082$i>>>0)<($820>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
       if ($821) {
        $822 = $data$sroa$59$082$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $823 = ($822>>>0)<($820>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
        $824 = ($820*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
        $825 = $824 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $826 = $823 ? $820 : $825; //@line 227 "libraries/lodepng/lodepng.cpp"
        $827 = (_realloc($data$sroa$0$084$i,$826)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
        $828 = ($827|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
        if ($828) {
         $data$sroa$0$2$i = $data$sroa$0$084$i;$data$sroa$36$1$i = $data$sroa$36$083$i;$data$sroa$59$2$i = $data$sroa$59$082$i;
        } else {
         $data$sroa$0$1$i = $827;$data$sroa$59$1$i = $826;
         label = 256;
        }
       } else {
        $data$sroa$0$1$i = $data$sroa$0$084$i;$data$sroa$59$1$i = $data$sroa$59$082$i;
        label = 256;
       }
       if ((label|0) == 256) {
        label = 0;
        $829 = (($data$sroa$0$1$i) + ($data$sroa$36$083$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
        HEAP8[$829>>0] = $830; //@line 289 "libraries/lodepng/lodepng.cpp"
        $data$sroa$0$2$i = $data$sroa$0$1$i;$data$sroa$36$1$i = $820;$data$sroa$59$2$i = $data$sroa$59$1$i;
       }
       $831 = (($i$085$i) + 1)|0; //@line 4952 "libraries/lodepng/lodepng.cpp"
       $832 = (($805) + ($831)|0); //@line 4952 "libraries/lodepng/lodepng.cpp"
       $833 = HEAP8[$832>>0]|0; //@line 4952 "libraries/lodepng/lodepng.cpp"
       $834 = ($833<<24>>24)==(0); //@line 4952 "libraries/lodepng/lodepng.cpp"
       if ($834) {
        $data$sroa$0$2$i$lcssa = $data$sroa$0$2$i;$data$sroa$36$1$i$lcssa = $data$sroa$36$1$i;$data$sroa$59$2$i$lcssa = $data$sroa$59$2$i;$i$085$i$lcssa = $i$085$i;
        break;
       } else {
        $830 = $833;$data$sroa$0$084$i = $data$sroa$0$2$i;$data$sroa$36$083$i = $data$sroa$36$1$i;$data$sroa$59$082$i = $data$sroa$59$2$i;$i$085$i = $831;
       }
      }
      $835 = ($i$085$i$lcssa>>>0)>(78); //@line 4953 "libraries/lodepng/lodepng.cpp"
      if (!($835)) {
       $836 = (($data$sroa$36$1$i$lcssa) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
       $837 = ($data$sroa$59$2$i$lcssa>>>0)<($836>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
       if ($837) {
        $838 = $data$sroa$59$2$i$lcssa << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $839 = ($838>>>0)<($836>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
        $840 = ($836*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
        $841 = $840 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $842 = $839 ? $836 : $841; //@line 227 "libraries/lodepng/lodepng.cpp"
        $843 = (_realloc($data$sroa$0$2$i$lcssa,$842)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
        $844 = ($843|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
        if ($844) {
         $data$sroa$0$4$i = $data$sroa$0$2$i$lcssa;$data$sroa$36$2$i = $data$sroa$36$1$i$lcssa;$data$sroa$59$4$i = $data$sroa$59$2$i$lcssa;
        } else {
         $data$sroa$0$3$i = $843;$data$sroa$59$3$i = $842;
         label = 261;
        }
       } else {
        $data$sroa$0$3$i = $data$sroa$0$2$i$lcssa;$data$sroa$59$3$i = $data$sroa$59$2$i$lcssa;
        label = 261;
       }
       if ((label|0) == 261) {
        label = 0;
        $845 = (($data$sroa$0$3$i) + ($data$sroa$36$1$i$lcssa)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
        HEAP8[$845>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
        $data$sroa$0$4$i = $data$sroa$0$3$i;$data$sroa$36$2$i = $836;$data$sroa$59$4$i = $data$sroa$59$3$i;
       }
       $846 = ($809|0)!=(0); //@line 4955 "libraries/lodepng/lodepng.cpp"
       $847 = $846&1; //@line 4955 "libraries/lodepng/lodepng.cpp"
       $848 = (($data$sroa$36$2$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
       $849 = ($data$sroa$59$4$i>>>0)<($848>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
       if ($849) {
        $850 = $data$sroa$59$4$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $851 = ($850>>>0)<($848>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
        $852 = ($848*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
        $853 = $852 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $854 = $851 ? $848 : $853; //@line 227 "libraries/lodepng/lodepng.cpp"
        $855 = (_realloc($data$sroa$0$4$i,$854)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
        $856 = ($855|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
        if ($856) {
         $data$sroa$0$6$i = $data$sroa$0$4$i;$data$sroa$36$3$i = $data$sroa$36$2$i;$data$sroa$59$6$i = $data$sroa$59$4$i;
        } else {
         $data$sroa$0$5$i = $855;$data$sroa$59$5$i = $854;
         label = 264;
        }
       } else {
        $data$sroa$0$5$i = $data$sroa$0$4$i;$data$sroa$59$5$i = $data$sroa$59$4$i;
        label = 264;
       }
       if ((label|0) == 264) {
        label = 0;
        $857 = (($data$sroa$0$5$i) + ($data$sroa$36$2$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
        HEAP8[$857>>0] = $847; //@line 289 "libraries/lodepng/lodepng.cpp"
        $data$sroa$0$6$i = $data$sroa$0$5$i;$data$sroa$36$3$i = $848;$data$sroa$59$6$i = $data$sroa$59$5$i;
       }
       $858 = (($data$sroa$36$3$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
       $859 = ($data$sroa$59$6$i>>>0)<($858>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
       if ($859) {
        $860 = $data$sroa$59$6$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $861 = ($860>>>0)<($858>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
        $862 = ($858*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
        $863 = $862 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $864 = $861 ? $858 : $863; //@line 227 "libraries/lodepng/lodepng.cpp"
        $865 = (_realloc($data$sroa$0$6$i,$864)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
        $866 = ($865|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
        if ($866) {
         $data$sroa$0$9$ph$i = $data$sroa$0$6$i;$data$sroa$36$5$ph$i = $data$sroa$36$3$i;$data$sroa$59$9$ph$i = $data$sroa$59$6$i;
        } else {
         $data$sroa$0$7$i = $865;$data$sroa$59$7$i = $864;
         label = 267;
        }
       } else {
        $data$sroa$0$7$i = $data$sroa$0$6$i;$data$sroa$59$7$i = $data$sroa$59$6$i;
        label = 267;
       }
       if ((label|0) == 267) {
        label = 0;
        $867 = (($data$sroa$0$7$i) + ($data$sroa$36$3$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
        HEAP8[$867>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
        $data$sroa$0$9$ph$i = $data$sroa$0$7$i;$data$sroa$36$5$ph$i = $858;$data$sroa$59$9$ph$i = $data$sroa$59$7$i;
       }
       $868 = HEAP8[$812>>0]|0; //@line 4957 "libraries/lodepng/lodepng.cpp"
       $869 = ($868<<24>>24)==(0); //@line 4957 "libraries/lodepng/lodepng.cpp"
       $870 = (($data$sroa$36$5$ph$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
       $871 = ($data$sroa$59$9$ph$i>>>0)<($870>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
       if ($869) {
        $$lcssa48$i = $870;$$lcssa49$i = $871;$data$sroa$0$9$lcssa$i = $data$sroa$0$9$ph$i;$data$sroa$36$5$lcssa$i = $data$sroa$36$5$ph$i;$data$sroa$59$9$lcssa$i = $data$sroa$59$9$ph$i;
       } else {
        $1037 = $871;$874 = $870;$881 = $868;$data$sroa$0$974$i = $data$sroa$0$9$ph$i;$data$sroa$36$573$i = $data$sroa$36$5$ph$i;$data$sroa$59$972$i = $data$sroa$59$9$ph$i;$i$175$i = 0;
        while(1) {
         if ($1037) {
          $872 = $data$sroa$59$972$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
          $873 = ($872>>>0)<($874>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
          $875 = ($874*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
          $876 = $875 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
          $877 = $873 ? $874 : $876; //@line 227 "libraries/lodepng/lodepng.cpp"
          $878 = (_realloc($data$sroa$0$974$i,$877)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
          $879 = ($878|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
          if ($879) {
           $data$sroa$0$11$i = $data$sroa$0$974$i;$data$sroa$36$6$i = $data$sroa$36$573$i;$data$sroa$59$11$i = $data$sroa$59$972$i;
          } else {
           $data$sroa$0$10$i = $878;$data$sroa$59$10$i = $877;
           label = 271;
          }
         } else {
          $data$sroa$0$10$i = $data$sroa$0$974$i;$data$sroa$59$10$i = $data$sroa$59$972$i;
          label = 271;
         }
         if ((label|0) == 271) {
          label = 0;
          $880 = (($data$sroa$0$10$i) + ($data$sroa$36$573$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
          HEAP8[$880>>0] = $881; //@line 289 "libraries/lodepng/lodepng.cpp"
          $data$sroa$0$11$i = $data$sroa$0$10$i;$data$sroa$36$6$i = $874;$data$sroa$59$11$i = $data$sroa$59$10$i;
         }
         $882 = (($i$175$i) + 1)|0; //@line 4957 "libraries/lodepng/lodepng.cpp"
         $883 = (($812) + ($882)|0); //@line 4957 "libraries/lodepng/lodepng.cpp"
         $884 = HEAP8[$883>>0]|0; //@line 4957 "libraries/lodepng/lodepng.cpp"
         $885 = ($884<<24>>24)==(0); //@line 4957 "libraries/lodepng/lodepng.cpp"
         $886 = (($data$sroa$36$6$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
         $887 = ($data$sroa$59$11$i>>>0)<($886>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
         if ($885) {
          $$lcssa48$i = $886;$$lcssa49$i = $887;$data$sroa$0$9$lcssa$i = $data$sroa$0$11$i;$data$sroa$36$5$lcssa$i = $data$sroa$36$6$i;$data$sroa$59$9$lcssa$i = $data$sroa$59$11$i;
          break;
         } else {
          $1037 = $887;$874 = $886;$881 = $884;$data$sroa$0$974$i = $data$sroa$0$11$i;$data$sroa$36$573$i = $data$sroa$36$6$i;$data$sroa$59$972$i = $data$sroa$59$11$i;$i$175$i = $882;
         }
        }
       }
       if ($$lcssa49$i) {
        $888 = $data$sroa$59$9$lcssa$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $889 = ($888>>>0)<($$lcssa48$i>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
        $890 = ($$lcssa48$i*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
        $891 = $890 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $892 = $889 ? $$lcssa48$i : $891; //@line 227 "libraries/lodepng/lodepng.cpp"
        $893 = (_realloc($data$sroa$0$9$lcssa$i,$892)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
        $894 = ($893|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
        if ($894) {
         $data$sroa$0$14$ph$i = $data$sroa$0$9$lcssa$i;$data$sroa$36$8$ph$i = $data$sroa$36$5$lcssa$i;$data$sroa$59$14$ph$i = $data$sroa$59$9$lcssa$i;
        } else {
         $data$sroa$0$12$i = $893;$data$sroa$59$12$i = $892;
         label = 275;
        }
       } else {
        $data$sroa$0$12$i = $data$sroa$0$9$lcssa$i;$data$sroa$59$12$i = $data$sroa$59$9$lcssa$i;
        label = 275;
       }
       if ((label|0) == 275) {
        label = 0;
        $895 = (($data$sroa$0$12$i) + ($data$sroa$36$5$lcssa$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
        HEAP8[$895>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
        $data$sroa$0$14$ph$i = $data$sroa$0$12$i;$data$sroa$36$8$ph$i = $$lcssa48$i;$data$sroa$59$14$ph$i = $data$sroa$59$12$i;
       }
       $896 = HEAP8[$815>>0]|0; //@line 4959 "libraries/lodepng/lodepng.cpp"
       $897 = ($896<<24>>24)==(0); //@line 4959 "libraries/lodepng/lodepng.cpp"
       $898 = (($data$sroa$36$8$ph$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
       $899 = ($data$sroa$59$14$ph$i>>>0)<($898>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
       if ($897) {
        $$lcssa$i = $898;$$lcssa47$i = $899;$data$sroa$0$14$lcssa$i = $data$sroa$0$14$ph$i;$data$sroa$36$8$lcssa$i = $data$sroa$36$8$ph$i;$data$sroa$59$14$lcssa$i = $data$sroa$59$14$ph$i;
       } else {
        $1038 = $899;$902 = $898;$909 = $896;$data$sroa$0$1464$i = $data$sroa$0$14$ph$i;$data$sroa$36$863$i = $data$sroa$36$8$ph$i;$data$sroa$59$1462$i = $data$sroa$59$14$ph$i;$i$265$i = 0;
        while(1) {
         if ($1038) {
          $900 = $data$sroa$59$1462$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
          $901 = ($900>>>0)<($902>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
          $903 = ($902*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
          $904 = $903 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
          $905 = $901 ? $902 : $904; //@line 227 "libraries/lodepng/lodepng.cpp"
          $906 = (_realloc($data$sroa$0$1464$i,$905)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
          $907 = ($906|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
          if ($907) {
           $data$sroa$0$16$i = $data$sroa$0$1464$i;$data$sroa$36$9$i = $data$sroa$36$863$i;$data$sroa$59$16$i = $data$sroa$59$1462$i;
          } else {
           $data$sroa$0$15$i = $906;$data$sroa$59$15$i = $905;
           label = 279;
          }
         } else {
          $data$sroa$0$15$i = $data$sroa$0$1464$i;$data$sroa$59$15$i = $data$sroa$59$1462$i;
          label = 279;
         }
         if ((label|0) == 279) {
          label = 0;
          $908 = (($data$sroa$0$15$i) + ($data$sroa$36$863$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
          HEAP8[$908>>0] = $909; //@line 289 "libraries/lodepng/lodepng.cpp"
          $data$sroa$0$16$i = $data$sroa$0$15$i;$data$sroa$36$9$i = $902;$data$sroa$59$16$i = $data$sroa$59$15$i;
         }
         $910 = (($i$265$i) + 1)|0; //@line 4959 "libraries/lodepng/lodepng.cpp"
         $911 = (($815) + ($910)|0); //@line 4959 "libraries/lodepng/lodepng.cpp"
         $912 = HEAP8[$911>>0]|0; //@line 4959 "libraries/lodepng/lodepng.cpp"
         $913 = ($912<<24>>24)==(0); //@line 4959 "libraries/lodepng/lodepng.cpp"
         $914 = (($data$sroa$36$9$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
         $915 = ($data$sroa$59$16$i>>>0)<($914>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
         if ($913) {
          $$lcssa$i = $914;$$lcssa47$i = $915;$data$sroa$0$14$lcssa$i = $data$sroa$0$16$i;$data$sroa$36$8$lcssa$i = $data$sroa$36$9$i;$data$sroa$59$14$lcssa$i = $data$sroa$59$16$i;
          break;
         } else {
          $1038 = $915;$902 = $914;$909 = $912;$data$sroa$0$1464$i = $data$sroa$0$16$i;$data$sroa$36$863$i = $data$sroa$36$9$i;$data$sroa$59$1462$i = $data$sroa$59$16$i;$i$265$i = $910;
         }
        }
       }
       if ($$lcssa47$i) {
        $916 = $data$sroa$59$14$lcssa$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $917 = ($916>>>0)<($$lcssa$i>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
        $918 = ($$lcssa$i*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
        $919 = $918 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
        $920 = $917 ? $$lcssa$i : $919; //@line 227 "libraries/lodepng/lodepng.cpp"
        $921 = (_realloc($data$sroa$0$14$lcssa$i,$920)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
        $922 = ($921|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
        if ($922) {
         $data$sroa$0$18$i = $data$sroa$0$14$lcssa$i;$data$sroa$36$10$i = $data$sroa$36$8$lcssa$i;$data$sroa$59$18$i = $data$sroa$59$14$lcssa$i;
        } else {
         $data$sroa$0$17$i = $921;$data$sroa$59$17$i = $920;
         label = 283;
        }
       } else {
        $data$sroa$0$17$i = $data$sroa$0$14$lcssa$i;$data$sroa$59$17$i = $data$sroa$59$14$lcssa$i;
        label = 283;
       }
       if ((label|0) == 283) {
        label = 0;
        $923 = (($data$sroa$0$17$i) + ($data$sroa$36$8$lcssa$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
        HEAP8[$923>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
        $data$sroa$0$18$i = $data$sroa$0$17$i;$data$sroa$36$10$i = $$lcssa$i;$data$sroa$59$18$i = $data$sroa$59$17$i;
       }
       do {
        if ($846) {
         HEAP32[$compressed_data$i>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
         HEAP32[$801>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
         HEAP32[$802>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
         $926 = HEAP32[$645>>2]|0; //@line 2209 "libraries/lodepng/lodepng.cpp"
         $927 = ($926|0)==(0|0); //@line 2209 "libraries/lodepng/lodepng.cpp"
         if ($927) {
          $929 = (__Z21lodepng_zlib_compressPPhPjPKhjPK23LodePNGCompressSettings($compressed_data$i,$802,$818,$819,$20)|0); //@line 2215 "libraries/lodepng/lodepng.cpp"
          $$0$i$i = $929;
         } else {
          $928 = (FUNCTION_TABLE_iiiiii[$926 & 8191]($compressed_data$i,$802,$818,$819,$20)|0); //@line 2211 "libraries/lodepng/lodepng.cpp"
          $$0$i$i = $928;
         }
         $930 = ($$0$i$i|0)==(0); //@line 4968 "libraries/lodepng/lodepng.cpp"
         $$not$i = $930 ^ 1; //@line 4968 "libraries/lodepng/lodepng.cpp"
         $931 = HEAP32[$802>>2]|0; //@line 4970 "libraries/lodepng/lodepng.cpp"
         $932 = ($931|0)==(0); //@line 4970 "libraries/lodepng/lodepng.cpp"
         $or$cond$i11 = $932 | $$not$i; //@line 4968 "libraries/lodepng/lodepng.cpp"
         if ($or$cond$i11) {
          $data$sroa$0$22$i = $data$sroa$0$18$i;$data$sroa$36$13$i = $data$sroa$36$10$i;
         } else {
          $data$sroa$0$1952$i = $data$sroa$0$18$i;$data$sroa$36$1151$i = $data$sroa$36$10$i;$data$sroa$59$1950$i = $data$sroa$59$18$i;$i$353$i = 0;
          while(1) {
           $933 = HEAP32[$compressed_data$i>>2]|0; //@line 4970 "libraries/lodepng/lodepng.cpp"
           $934 = (($933) + ($i$353$i)|0); //@line 4970 "libraries/lodepng/lodepng.cpp"
           $935 = HEAP8[$934>>0]|0; //@line 4970 "libraries/lodepng/lodepng.cpp"
           $936 = (($data$sroa$36$1151$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
           $937 = ($data$sroa$59$1950$i>>>0)<($936>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
           if ($937) {
            $938 = $data$sroa$59$1950$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
            $939 = ($938>>>0)<($936>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
            $940 = ($936*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
            $941 = $940 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
            $942 = $939 ? $936 : $941; //@line 227 "libraries/lodepng/lodepng.cpp"
            $943 = (_realloc($data$sroa$0$1952$i,$942)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
            $944 = ($943|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
            if ($944) {
             $data$sroa$0$21$i = $data$sroa$0$1952$i;$data$sroa$36$12$i = $data$sroa$36$1151$i;$data$sroa$59$21$i = $data$sroa$59$1950$i;
            } else {
             $data$sroa$0$20$i = $943;$data$sroa$59$20$i = $942;
             label = 292;
            }
           } else {
            $data$sroa$0$20$i = $data$sroa$0$1952$i;$data$sroa$59$20$i = $data$sroa$59$1950$i;
            label = 292;
           }
           if ((label|0) == 292) {
            label = 0;
            $945 = (($data$sroa$0$20$i) + ($data$sroa$36$1151$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
            HEAP8[$945>>0] = $935; //@line 289 "libraries/lodepng/lodepng.cpp"
            $data$sroa$0$21$i = $data$sroa$0$20$i;$data$sroa$36$12$i = $936;$data$sroa$59$21$i = $data$sroa$59$20$i;
           }
           $946 = (($i$353$i) + 1)|0; //@line 4970 "libraries/lodepng/lodepng.cpp"
           $947 = HEAP32[$802>>2]|0; //@line 4970 "libraries/lodepng/lodepng.cpp"
           $948 = ($946|0)==($947|0); //@line 4970 "libraries/lodepng/lodepng.cpp"
           if ($948) {
            $data$sroa$0$22$i = $data$sroa$0$21$i;$data$sroa$36$13$i = $data$sroa$36$12$i;
            break;
           } else {
            $data$sroa$0$1952$i = $data$sroa$0$21$i;$data$sroa$36$1151$i = $data$sroa$36$12$i;$data$sroa$59$1950$i = $data$sroa$59$21$i;$i$353$i = $946;
           }
          }
         }
         HEAP32[$801>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
         HEAP32[$802>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
         $949 = HEAP32[$compressed_data$i>>2]|0; //@line 252 "libraries/lodepng/lodepng.cpp"
         _free($949); //@line 76 "libraries/lodepng/lodepng.cpp"
         if ($930) {
          $data$sroa$0$26$i = $data$sroa$0$22$i;$data$sroa$36$16$i = $data$sroa$36$13$i;
          label = 299;
         } else {
          $data$sroa$0$27$i = $data$sroa$0$22$i;
         }
        } else {
         $924 = HEAP8[$818>>0]|0; //@line 4976 "libraries/lodepng/lodepng.cpp"
         $925 = ($924<<24>>24)==(0); //@line 4976 "libraries/lodepng/lodepng.cpp"
         if ($925) {
          $data$sroa$0$26$i = $data$sroa$0$18$i;$data$sroa$36$16$i = $data$sroa$36$10$i;
          label = 299;
          break;
         } else {
          $960 = $924;$data$sroa$0$2357$i = $data$sroa$0$18$i;$data$sroa$36$1456$i = $data$sroa$36$10$i;$data$sroa$59$2255$i = $data$sroa$59$18$i;$i$458$i = 0;
         }
         while(1) {
          $950 = (($data$sroa$36$1456$i) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
          $951 = ($data$sroa$59$2255$i>>>0)<($950>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
          if ($951) {
           $952 = $data$sroa$59$2255$i << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
           $953 = ($952>>>0)<($950>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
           $954 = ($950*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
           $955 = $954 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
           $956 = $953 ? $950 : $955; //@line 227 "libraries/lodepng/lodepng.cpp"
           $957 = (_realloc($data$sroa$0$2357$i,$956)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
           $958 = ($957|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
           if ($958) {
            $data$sroa$0$25$i = $data$sroa$0$2357$i;$data$sroa$36$15$i = $data$sroa$36$1456$i;$data$sroa$59$24$i = $data$sroa$59$2255$i;
           } else {
            $data$sroa$0$24$i = $957;$data$sroa$59$23$i = $956;
            label = 297;
           }
          } else {
           $data$sroa$0$24$i = $data$sroa$0$2357$i;$data$sroa$59$23$i = $data$sroa$59$2255$i;
           label = 297;
          }
          if ((label|0) == 297) {
           label = 0;
           $959 = (($data$sroa$0$24$i) + ($data$sroa$36$1456$i)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
           HEAP8[$959>>0] = $960; //@line 289 "libraries/lodepng/lodepng.cpp"
           $data$sroa$0$25$i = $data$sroa$0$24$i;$data$sroa$36$15$i = $950;$data$sroa$59$24$i = $data$sroa$59$23$i;
          }
          $961 = (($i$458$i) + 1)|0; //@line 4976 "libraries/lodepng/lodepng.cpp"
          $962 = (($818) + ($961)|0); //@line 4976 "libraries/lodepng/lodepng.cpp"
          $963 = HEAP8[$962>>0]|0; //@line 4976 "libraries/lodepng/lodepng.cpp"
          $964 = ($963<<24>>24)==(0); //@line 4976 "libraries/lodepng/lodepng.cpp"
          if ($964) {
           $data$sroa$0$26$i = $data$sroa$0$25$i;$data$sroa$36$16$i = $data$sroa$36$15$i;
           label = 299;
           break;
          } else {
           $960 = $963;$data$sroa$0$2357$i = $data$sroa$0$25$i;$data$sroa$36$1456$i = $data$sroa$36$15$i;$data$sroa$59$2255$i = $data$sroa$59$24$i;$i$458$i = $961;
          }
         }
        }
       } while(0);
       do {
        if ((label|0) == 299) {
         label = 0;
         $965 = (__Z20lodepng_chunk_createPPhPjjPKcPKh($outv,$86,$data$sroa$36$16$i,502787,$data$sroa$0$26$i)|0); //@line 4775 "libraries/lodepng/lodepng.cpp"
         $966 = ($965|0)==(0); //@line 4775 "libraries/lodepng/lodepng.cpp"
         if (!($966)) {
          $data$sroa$0$27$i = $data$sroa$0$26$i;
          break;
         }
         $967 = HEAP32[$86>>2]|0; //@line 4776 "libraries/lodepng/lodepng.cpp"
         HEAP32[$85>>2] = $967; //@line 4776 "libraries/lodepng/lodepng.cpp"
         $data$sroa$0$27$i = $data$sroa$0$26$i;
        }
       } while(0);
       _free($data$sroa$0$27$i); //@line 76 "libraries/lodepng/lodepng.cpp"
      }
      $968 = (($i$2147) + 1)|0; //@line 5700 "libraries/lodepng/lodepng.cpp"
      $969 = HEAP32[$793>>2]|0; //@line 5700 "libraries/lodepng/lodepng.cpp"
      $970 = ($968|0)==($969|0); //@line 5700 "libraries/lodepng/lodepng.cpp"
      if ($970) {
       break L367;
      } else {
       $i$2147 = $968;
      }
     }
     if ((label|0) == 250) {
      HEAP32[$0>>2] = 66; //@line 5704 "libraries/lodepng/lodepng.cpp"
      break;
     }
     else if ((label|0) == 252) {
      HEAP32[$0>>2] = 67; //@line 5709 "libraries/lodepng/lodepng.cpp"
      break;
     }
    }
   } while(0);
   $971 = ((($info)) + 144|0); //@line 5718 "libraries/lodepng/lodepng.cpp"
   $972 = HEAP32[$971>>2]|0; //@line 5718 "libraries/lodepng/lodepng.cpp"
   $973 = ($972|0)==(0|0); //@line 5718 "libraries/lodepng/lodepng.cpp"
   if (!($973)) {
    $974 = ((($info)) + 156|0); //@line 5720 "libraries/lodepng/lodepng.cpp"
    $975 = HEAP32[$974>>2]|0; //@line 5720 "libraries/lodepng/lodepng.cpp"
    $976 = $972; //@line 5535 "libraries/lodepng/lodepng.cpp"
    $977 = ($975|0)==(0); //@line 5535 "libraries/lodepng/lodepng.cpp"
    L460: do {
     if (!($977)) {
      $$pre159 = HEAP32[$86>>2]|0; //@line 2469 "libraries/lodepng/lodepng.cpp"
      $997 = $$pre159;$inchunk$02$i = $972;
      while(1) {
       $978 = HEAP8[$inchunk$02$i>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $979 = $978&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $980 = $979 << 24; //@line 343 "libraries/lodepng/lodepng.cpp"
       $981 = ((($inchunk$02$i)) + 1|0); //@line 343 "libraries/lodepng/lodepng.cpp"
       $982 = HEAP8[$981>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $983 = $982&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $984 = $983 << 16; //@line 343 "libraries/lodepng/lodepng.cpp"
       $985 = $984 | $980; //@line 343 "libraries/lodepng/lodepng.cpp"
       $986 = ((($inchunk$02$i)) + 2|0); //@line 343 "libraries/lodepng/lodepng.cpp"
       $987 = HEAP8[$986>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $988 = $987&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $989 = $988 << 8; //@line 343 "libraries/lodepng/lodepng.cpp"
       $990 = $985 | $989; //@line 343 "libraries/lodepng/lodepng.cpp"
       $991 = ((($inchunk$02$i)) + 3|0); //@line 343 "libraries/lodepng/lodepng.cpp"
       $992 = HEAP8[$991>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $993 = $992&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $994 = $990 | $993; //@line 343 "libraries/lodepng/lodepng.cpp"
       $995 = (($994) + 12)|0; //@line 2467 "libraries/lodepng/lodepng.cpp"
       $996 = (($995) + ($997))|0; //@line 5538 "libraries/lodepng/lodepng.cpp"
       $998 = ($996>>>0)<($995>>>0); //@line 2470 "libraries/lodepng/lodepng.cpp"
       $999 = ($996>>>0)<($997>>>0); //@line 2470 "libraries/lodepng/lodepng.cpp"
       $or$cond$i = $998 | $999; //@line 2470 "libraries/lodepng/lodepng.cpp"
       if ($or$cond$i) {
        $$2$i = 77;
        break;
       }
       $1000 = HEAP32[$outv>>2]|0; //@line 2472 "libraries/lodepng/lodepng.cpp"
       $1001 = (_realloc($1000,$996)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       $1002 = ($1001|0)==(0|0); //@line 2473 "libraries/lodepng/lodepng.cpp"
       if ($1002) {
        $$2$i = 83;
        break;
       }
       HEAP32[$outv>>2] = $1001; //@line 2474 "libraries/lodepng/lodepng.cpp"
       HEAP32[$86>>2] = $996; //@line 2475 "libraries/lodepng/lodepng.cpp"
       $1003 = (($1001) + ($997)|0); //@line 2476 "libraries/lodepng/lodepng.cpp"
       $1004 = ($995|0)==(0); //@line 2478 "libraries/lodepng/lodepng.cpp"
       if (!($1004)) {
        $i$01$i8 = 0;
        while(1) {
         $1005 = (($inchunk$02$i) + ($i$01$i8)|0); //@line 2478 "libraries/lodepng/lodepng.cpp"
         $1006 = HEAP8[$1005>>0]|0; //@line 2478 "libraries/lodepng/lodepng.cpp"
         $1007 = (($1003) + ($i$01$i8)|0); //@line 2478 "libraries/lodepng/lodepng.cpp"
         HEAP8[$1007>>0] = $1006; //@line 2478 "libraries/lodepng/lodepng.cpp"
         $1008 = (($i$01$i8) + 1)|0; //@line 2478 "libraries/lodepng/lodepng.cpp"
         $1009 = ($1008|0)==($995|0); //@line 2478 "libraries/lodepng/lodepng.cpp"
         if ($1009) {
          break;
         } else {
          $i$01$i8 = $1008;
         }
        }
       }
       HEAP32[$85>>2] = $996; //@line 5538 "libraries/lodepng/lodepng.cpp"
       $1010 = HEAP8[$inchunk$02$i>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $1011 = $1010&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $1012 = $1011 << 24; //@line 343 "libraries/lodepng/lodepng.cpp"
       $1013 = HEAP8[$981>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $1014 = $1013&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $1015 = $1014 << 16; //@line 343 "libraries/lodepng/lodepng.cpp"
       $1016 = $1015 | $1012; //@line 343 "libraries/lodepng/lodepng.cpp"
       $1017 = HEAP8[$986>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $1018 = $1017&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $1019 = $1018 << 8; //@line 343 "libraries/lodepng/lodepng.cpp"
       $1020 = $1016 | $1019; //@line 343 "libraries/lodepng/lodepng.cpp"
       $1021 = HEAP8[$991>>0]|0; //@line 343 "libraries/lodepng/lodepng.cpp"
       $1022 = $1021&255; //@line 343 "libraries/lodepng/lodepng.cpp"
       $1023 = $1020 | $1022; //@line 343 "libraries/lodepng/lodepng.cpp"
       $1024 = (($1023) + 12)|0; //@line 2454 "libraries/lodepng/lodepng.cpp"
       $1025 = (($inchunk$02$i) + ($1024)|0); //@line 2455 "libraries/lodepng/lodepng.cpp"
       $1026 = $1025; //@line 5535 "libraries/lodepng/lodepng.cpp"
       $1027 = (($1026) - ($976))|0; //@line 5535 "libraries/lodepng/lodepng.cpp"
       $1028 = ($1027>>>0)<($975>>>0); //@line 5535 "libraries/lodepng/lodepng.cpp"
       if ($1028) {
        $997 = $996;$inchunk$02$i = $1025;
       } else {
        break L460;
       }
      }
      HEAP32[$0>>2] = $$2$i; //@line 5720 "libraries/lodepng/lodepng.cpp"
      break L71;
     }
    } while(0);
    HEAP32[$0>>2] = 0; //@line 5720 "libraries/lodepng/lodepng.cpp"
   }
   $1029 = (__Z20lodepng_chunk_createPPhPjjPKcPKh($outv,$86,0,502792,0)|0); //@line 4775 "libraries/lodepng/lodepng.cpp"
   $1030 = ($1029|0)==(0); //@line 4775 "libraries/lodepng/lodepng.cpp"
   if (!($1030)) {
    break;
   }
   $1031 = HEAP32[$86>>2]|0; //@line 4776 "libraries/lodepng/lodepng.cpp"
   HEAP32[$85>>2] = $1031; //@line 4776 "libraries/lodepng/lodepng.cpp"
  }
 } while(0);
 __Z20lodepng_info_cleanupP11LodePNGInfo($info); //@line 5729 "libraries/lodepng/lodepng.cpp"
 $1032 = HEAP32[$data>>2]|0; //@line 5730 "libraries/lodepng/lodepng.cpp"
 _free($1032); //@line 76 "libraries/lodepng/lodepng.cpp"
 $1033 = HEAP32[$outv>>2]|0; //@line 5732 "libraries/lodepng/lodepng.cpp"
 HEAP32[$out>>2] = $1033; //@line 5732 "libraries/lodepng/lodepng.cpp"
 $1034 = HEAP32[$86>>2]|0; //@line 5733 "libraries/lodepng/lodepng.cpp"
 HEAP32[$outsize>>2] = $1034; //@line 5733 "libraries/lodepng/lodepng.cpp"
 $1035 = HEAP32[$0>>2]|0; //@line 5735 "libraries/lodepng/lodepng.cpp"
 $$0 = $1035;
 STACKTOP = sp;return ($$0|0); //@line 5736 "libraries/lodepng/lodepng.cpp"
}
function __ZL19preProcessScanlinesPPhPjPKhjjPK11LodePNGInfoPK22LodePNGEncoderSettings($out,$outsize,$in,$w,$h,$info_png,$settings) {
 $out = $out|0;
 $outsize = $outsize|0;
 $in = $in|0;
 $w = $w|0;
 $h = $h|0;
 $info_png = $info_png|0;
 $settings = $settings|0;
 var $$0$i$i$i = 0, $$idx$val = 0, $$idx10$val = 0, $$idx4$val = 0, $$idx5$val$us = 0, $$idx6$val$us = 0, $$idx8$val = 0, $$idx9$val = 0, $$old1 = 0, $$pre = 0, $$pre$i = 0, $$pre38 = 0, $$pre41$i = 0, $$pre42$i = 0, $$pre43$i = 0, $$pre44$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $b3$018$us$i = 0, $exitcond32 = 0, $exitcond33 = 0, $exitcond34 = 0, $exitcond35 = 0, $exitcond39$i = 0;
 var $filter_passstart = 0, $filter_passstart$i = 0, $i$015 = 0, $i$015$i = 0, $i$015$us = 0, $i$128$i = 0, $ibp$017$us$i = 0, $not$ = 0, $not$13 = 0, $obp$016$us$i = 0, $or$cond = 0, $or$cond3 = 0, $or$cond41 = 0, $padded_passstart = 0, $padded_passstart$i = 0, $passh = 0, $passh$i = 0, $passstart = 0, $passstart$i = 0, $passw = 0;
 var $passw$i = 0, $scevgep = 0, $scevgep37 = 0, $tmp$us$i = 0, $tmp1$us$i = 0, $tmp2$us$i = 0, $tmp3$us$i = 0, $x$09$us$us$i = 0, $x1$022$us$i = 0, $y$012$us$i = 0, $y2$025$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $passw$i = sp + 276|0;
 $passh$i = sp + 248|0;
 $filter_passstart$i = sp + 216|0;
 $padded_passstart$i = sp + 184|0;
 $passstart$i = sp + 152|0;
 $passw = sp + 124|0;
 $passh = sp + 96|0;
 $filter_passstart = sp + 64|0;
 $padded_passstart = sp + 32|0;
 $passstart = sp;
 $0 = ((($info_png)) + 12|0); //@line 2634 "libraries/lodepng/lodepng.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 2634 "libraries/lodepng/lodepng.cpp"
 $2 = ((($info_png)) + 16|0); //@line 2634 "libraries/lodepng/lodepng.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 2634 "libraries/lodepng/lodepng.cpp"
 switch ($1|0) {
 case 3: case 0:  {
  $$0$i$i$i = 1;
  break;
 }
 case 2:  {
  $$0$i$i$i = 3;
  break;
 }
 case 4:  {
  $$0$i$i$i = 2;
  break;
 }
 case 6:  {
  $$0$i$i$i = 4;
  break;
 }
 default: {
  $$0$i$i$i = 0;
 }
 }
 $4 = Math_imul($$0$i$i$i, $3)|0; //@line 2549 "libraries/lodepng/lodepng.cpp"
 $5 = ((($info_png)) + 8|0); //@line 5430 "libraries/lodepng/lodepng.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 5430 "libraries/lodepng/lodepng.cpp"
 $7 = ($6|0)==(0); //@line 5430 "libraries/lodepng/lodepng.cpp"
 if ($7) {
  $8 = Math_imul($4, $w)|0; //@line 5432 "libraries/lodepng/lodepng.cpp"
  $9 = (($8) + 7)|0; //@line 5432 "libraries/lodepng/lodepng.cpp"
  $10 = $9 >>> 3; //@line 5432 "libraries/lodepng/lodepng.cpp"
  $11 = Math_imul($10, $h)|0; //@line 5432 "libraries/lodepng/lodepng.cpp"
  $12 = (($11) + ($h))|0; //@line 5434 "libraries/lodepng/lodepng.cpp"
  HEAP32[$outsize>>2] = $12; //@line 5432 "libraries/lodepng/lodepng.cpp"
  $13 = (_malloc($12)|0); //@line 5453 "libraries/lodepng/lodepng.cpp"
  HEAP32[$out>>2] = $13; //@line 5433 "libraries/lodepng/lodepng.cpp"
  $14 = ($13|0)!=(0|0); //@line 5434 "libraries/lodepng/lodepng.cpp"
  $15 = ($12|0)==(0); //@line 5434 "libraries/lodepng/lodepng.cpp"
  $or$cond41 = $14 | $15; //@line 5434 "libraries/lodepng/lodepng.cpp"
  if (!($or$cond41)) {
   STACKTOP = sp;return;
  }
  $16 = ($4>>>0)<(8); //@line 5439 "libraries/lodepng/lodepng.cpp"
  if ($16) {
   $17 = $10 << 3; //@line 5439 "libraries/lodepng/lodepng.cpp"
   $18 = ($8|0)==($17|0); //@line 5439 "libraries/lodepng/lodepng.cpp"
   if (!($18)) {
    $19 = (_malloc($11)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
    $not$13 = ($19|0)==(0|0); //@line 5443 "libraries/lodepng/lodepng.cpp"
    if (!($not$13)) {
     __ZL14addPaddingBitsPhPKhjjj($19,$in,$17,$8,$h); //@line 5445 "libraries/lodepng/lodepng.cpp"
     $20 = HEAP32[$out>>2]|0; //@line 5446 "libraries/lodepng/lodepng.cpp"
     $$idx9$val = HEAP32[$0>>2]|0;
     $$idx10$val = HEAP32[$2>>2]|0;
     (__ZL6filterPhPKhjjPK16LodePNGColorModePK22LodePNGEncoderSettings($20,$19,$w,$h,$$idx9$val,$$idx10$val,$settings)|0); //@line 5446 "libraries/lodepng/lodepng.cpp"
    }
    _free($19); //@line 76 "libraries/lodepng/lodepng.cpp"
    STACKTOP = sp;return;
   }
  }
  $$idx8$val = HEAP32[$2>>2]|0;
  (__ZL6filterPhPKhjjPK16LodePNGColorModePK22LodePNGEncoderSettings($13,$in,$w,$h,$1,$$idx8$val,$settings)|0); //@line 5453 "libraries/lodepng/lodepng.cpp"
  STACKTOP = sp;return;
 }
 __ZL19Adam7_getpassvaluesPjS_S_S_S_jjj($passw,$passh,$filter_passstart,$padded_passstart,$passstart,$w,$h,$4); //@line 5463 "libraries/lodepng/lodepng.cpp"
 $21 = ((($filter_passstart)) + 28|0); //@line 5465 "libraries/lodepng/lodepng.cpp"
 $22 = HEAP32[$21>>2]|0; //@line 5465 "libraries/lodepng/lodepng.cpp"
 HEAP32[$outsize>>2] = $22; //@line 5465 "libraries/lodepng/lodepng.cpp"
 $23 = (_malloc($22)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
 HEAP32[$out>>2] = $23; //@line 5466 "libraries/lodepng/lodepng.cpp"
 $24 = ((($passstart)) + 28|0); //@line 5469 "libraries/lodepng/lodepng.cpp"
 $25 = HEAP32[$24>>2]|0; //@line 5469 "libraries/lodepng/lodepng.cpp"
 $26 = (_malloc($25)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
 $27 = ($26|0)!=(0|0); //@line 5470 "libraries/lodepng/lodepng.cpp"
 $28 = ($25|0)==(0); //@line 5470 "libraries/lodepng/lodepng.cpp"
 $or$cond3 = $28 | $27; //@line 5470 "libraries/lodepng/lodepng.cpp"
 $not$ = ($23|0)!=(0|0); //@line 5472 "libraries/lodepng/lodepng.cpp"
 $29 = $not$ & $or$cond3; //@line 5472 "libraries/lodepng/lodepng.cpp"
 L22: do {
  if ($29) {
   __ZL19Adam7_getpassvaluesPjS_S_S_S_jjj($passw$i,$passh$i,$filter_passstart$i,$padded_passstart$i,$passstart$i,$w,$h,$4); //@line 5373 "libraries/lodepng/lodepng.cpp"
   $30 = ($4>>>0)>(7); //@line 5375 "libraries/lodepng/lodepng.cpp"
   if ($30) {
    $32 = $4 >>> 3; //@line 5380 "libraries/lodepng/lodepng.cpp"
    $33 = ($32|0)==(0); //@line 5386 "libraries/lodepng/lodepng.cpp"
    if (!($33)) {
     $34 = Math_imul($32, $w)|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
     $i$015$i = 0;
     while(1) {
      $35 = (($passh$i) + ($i$015$i<<2)|0); //@line 5381 "libraries/lodepng/lodepng.cpp"
      $36 = HEAP32[$35>>2]|0; //@line 5381 "libraries/lodepng/lodepng.cpp"
      $37 = ($36|0)==(0); //@line 5381 "libraries/lodepng/lodepng.cpp"
      if (!($37)) {
       $38 = (($passw$i) + ($i$015$i<<2)|0); //@line 5382 "libraries/lodepng/lodepng.cpp"
       $39 = HEAP32[$38>>2]|0; //@line 5382 "libraries/lodepng/lodepng.cpp"
       $40 = ($39|0)==(0); //@line 5382 "libraries/lodepng/lodepng.cpp"
       if (!($40)) {
        $41 = (($passstart$i) + ($i$015$i<<2)|0); //@line 5385 "libraries/lodepng/lodepng.cpp"
        $42 = (34884 + ($i$015$i<<2)|0); //@line 5384 "libraries/lodepng/lodepng.cpp"
        $43 = (34912 + ($i$015$i<<2)|0); //@line 5384 "libraries/lodepng/lodepng.cpp"
        $44 = (34940 + ($i$015$i<<2)|0); //@line 5384 "libraries/lodepng/lodepng.cpp"
        $45 = (34968 + ($i$015$i<<2)|0); //@line 5384 "libraries/lodepng/lodepng.cpp"
        $$pre$i = HEAP32[$45>>2]|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
        $$pre41$i = HEAP32[$44>>2]|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
        $$pre42$i = HEAP32[$43>>2]|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
        $$pre43$i = HEAP32[$42>>2]|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
        $$pre44$i = HEAP32[$41>>2]|0; //@line 5385 "libraries/lodepng/lodepng.cpp"
        $46 = Math_imul($32, $39)|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
        $47 = Math_imul($$pre$i, $w)|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
        $48 = (($$pre42$i) + ($47))|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
        $49 = Math_imul($32, $48)|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
        $50 = Math_imul($34, $$pre41$i)|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
        $51 = Math_imul($32, $$pre43$i)|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
        $y$012$us$i = 0;
        while(1) {
         $52 = Math_imul($46, $y$012$us$i)|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
         $53 = (($$pre44$i) + ($52))|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
         $54 = Math_imul($50, $y$012$us$i)|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
         $55 = (($49) + ($54))|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
         $x$09$us$us$i = 0;
         while(1) {
          $57 = Math_imul($32, $x$09$us$us$i)|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
          $58 = (($53) + ($57))|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
          $scevgep = (($26) + ($58)|0); //@line 5384 "libraries/lodepng/lodepng.cpp"
          $59 = Math_imul($51, $x$09$us$us$i)|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
          $60 = (($55) + ($59))|0; //@line 5384 "libraries/lodepng/lodepng.cpp"
          $scevgep37 = (($in) + ($60)|0); //@line 5384 "libraries/lodepng/lodepng.cpp"
          _memcpy(($scevgep|0),($scevgep37|0),($32|0))|0; //@line 5388 "libraries/lodepng/lodepng.cpp"
          $61 = (($x$09$us$us$i) + 1)|0; //@line 5382 "libraries/lodepng/lodepng.cpp"
          $exitcond32 = ($61|0)==($39|0); //@line 5382 "libraries/lodepng/lodepng.cpp"
          if ($exitcond32) {
           break;
          } else {
           $x$09$us$us$i = $61;
          }
         }
         $56 = (($y$012$us$i) + 1)|0; //@line 5381 "libraries/lodepng/lodepng.cpp"
         $exitcond33 = ($56|0)==($36|0); //@line 5381 "libraries/lodepng/lodepng.cpp"
         if ($exitcond33) {
          break;
         } else {
          $y$012$us$i = $56;
         }
        }
       }
      }
      $62 = (($i$015$i) + 1)|0; //@line 5377 "libraries/lodepng/lodepng.cpp"
      $63 = ($62|0)==(7); //@line 5377 "libraries/lodepng/lodepng.cpp"
      if ($63) {
       break;
      } else {
       $i$015$i = $62;
      }
     }
    }
   } else {
    $31 = ($4|0)==(0); //@line 5406 "libraries/lodepng/lodepng.cpp"
    $i$128$i = 0;
    while(1) {
     $64 = (($passw$i) + ($i$128$i<<2)|0); //@line 5398 "libraries/lodepng/lodepng.cpp"
     $65 = HEAP32[$64>>2]|0; //@line 5402 "libraries/lodepng/lodepng.cpp"
     $66 = (($passh$i) + ($i$128$i<<2)|0); //@line 5401 "libraries/lodepng/lodepng.cpp"
     $67 = HEAP32[$66>>2]|0; //@line 5401 "libraries/lodepng/lodepng.cpp"
     $68 = ($67|0)==(0); //@line 5401 "libraries/lodepng/lodepng.cpp"
     if (!($68)) {
      $69 = ($65|0)==(0); //@line 5402 "libraries/lodepng/lodepng.cpp"
      $70 = (34968 + ($i$128$i<<2)|0); //@line 5404 "libraries/lodepng/lodepng.cpp"
      $71 = (34940 + ($i$128$i<<2)|0); //@line 5404 "libraries/lodepng/lodepng.cpp"
      $72 = (34912 + ($i$128$i<<2)|0); //@line 5404 "libraries/lodepng/lodepng.cpp"
      if (!($69)) {
       $73 = (($passstart$i) + ($i$128$i<<2)|0); //@line 5405 "libraries/lodepng/lodepng.cpp"
       $74 = (34884 + ($i$128$i<<2)|0); //@line 5404 "libraries/lodepng/lodepng.cpp"
       $$pre = HEAP32[$74>>2]|0; //@line 5404 "libraries/lodepng/lodepng.cpp"
       $$pre38 = HEAP32[$73>>2]|0; //@line 5405 "libraries/lodepng/lodepng.cpp"
       $75 = $$pre38 << 3; //@line 5405 "libraries/lodepng/lodepng.cpp"
       $y2$025$i = 0;
       while(1) {
        $76 = Math_imul($y2$025$i, $65)|0; //@line 5405 "libraries/lodepng/lodepng.cpp"
        if (!($31)) {
         $77 = HEAP32[$72>>2]|0; //@line 5404 "libraries/lodepng/lodepng.cpp"
         $78 = HEAP32[$71>>2]|0; //@line 5404 "libraries/lodepng/lodepng.cpp"
         $79 = Math_imul($78, $y2$025$i)|0; //@line 5404 "libraries/lodepng/lodepng.cpp"
         $80 = HEAP32[$70>>2]|0; //@line 5404 "libraries/lodepng/lodepng.cpp"
         $81 = (($79) + ($80))|0; //@line 5404 "libraries/lodepng/lodepng.cpp"
         $82 = Math_imul($81, $w)|0; //@line 5404 "libraries/lodepng/lodepng.cpp"
         $83 = (($82) + ($77))|0; //@line 5404 "libraries/lodepng/lodepng.cpp"
         $x1$022$us$i = 0;
         while(1) {
          $84 = Math_imul($x1$022$us$i, $$pre)|0; //@line 5404 "libraries/lodepng/lodepng.cpp"
          $tmp$us$i = (($83) + ($84))|0; //@line 5476 "libraries/lodepng/lodepng.cpp"
          $tmp1$us$i = Math_imul($tmp$us$i, $4)|0; //@line 5404 "libraries/lodepng/lodepng.cpp"
          $tmp2$us$i = (($x1$022$us$i) + ($76))|0; //@line 5476 "libraries/lodepng/lodepng.cpp"
          $tmp3$us$i = Math_imul($tmp2$us$i, $4)|0; //@line 5476 "libraries/lodepng/lodepng.cpp"
          $85 = (($tmp3$us$i) + ($75))|0; //@line 5405 "libraries/lodepng/lodepng.cpp"
          $b3$018$us$i = 0;$ibp$017$us$i = $tmp1$us$i;$obp$016$us$i = $85;
          while(1) {
           $86 = $ibp$017$us$i >>> 3; //@line 2351 "libraries/lodepng/lodepng.cpp"
           $87 = (($in) + ($86)|0); //@line 2351 "libraries/lodepng/lodepng.cpp"
           $88 = HEAP8[$87>>0]|0; //@line 2351 "libraries/lodepng/lodepng.cpp"
           $89 = $88&255; //@line 2351 "libraries/lodepng/lodepng.cpp"
           $90 = $ibp$017$us$i & 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
           $91 = $90 ^ 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
           $92 = (($ibp$017$us$i) + 1)|0; //@line 2352 "libraries/lodepng/lodepng.cpp"
           $93 = 1 << $91; //@line 2383 "libraries/lodepng/lodepng.cpp"
           $94 = $89 & $93; //@line 2383 "libraries/lodepng/lodepng.cpp"
           $95 = ($94|0)==(0); //@line 2383 "libraries/lodepng/lodepng.cpp"
           $96 = $obp$016$us$i & 7; //@line 2383 "libraries/lodepng/lodepng.cpp"
           $97 = $96 ^ 7; //@line 2383 "libraries/lodepng/lodepng.cpp"
           $98 = 1 << $97; //@line 2383 "libraries/lodepng/lodepng.cpp"
           if ($95) {
            $105 = $98 ^ 255; //@line 2383 "libraries/lodepng/lodepng.cpp"
            $106 = $obp$016$us$i >>> 3; //@line 2383 "libraries/lodepng/lodepng.cpp"
            $107 = (($26) + ($106)|0); //@line 2383 "libraries/lodepng/lodepng.cpp"
            $108 = HEAP8[$107>>0]|0; //@line 2383 "libraries/lodepng/lodepng.cpp"
            $109 = $108&255; //@line 2383 "libraries/lodepng/lodepng.cpp"
            $110 = $109 & $105; //@line 2383 "libraries/lodepng/lodepng.cpp"
            $111 = $110&255; //@line 2383 "libraries/lodepng/lodepng.cpp"
            HEAP8[$107>>0] = $111; //@line 2383 "libraries/lodepng/lodepng.cpp"
           } else {
            $99 = $obp$016$us$i >>> 3; //@line 2384 "libraries/lodepng/lodepng.cpp"
            $100 = (($26) + ($99)|0); //@line 2384 "libraries/lodepng/lodepng.cpp"
            $101 = HEAP8[$100>>0]|0; //@line 2384 "libraries/lodepng/lodepng.cpp"
            $102 = $101&255; //@line 2384 "libraries/lodepng/lodepng.cpp"
            $103 = $102 | $98; //@line 2384 "libraries/lodepng/lodepng.cpp"
            $104 = $103&255; //@line 2384 "libraries/lodepng/lodepng.cpp"
            HEAP8[$100>>0] = $104; //@line 2384 "libraries/lodepng/lodepng.cpp"
           }
           $112 = (($obp$016$us$i) + 1)|0; //@line 2385 "libraries/lodepng/lodepng.cpp"
           $113 = (($b3$018$us$i) + 1)|0; //@line 5406 "libraries/lodepng/lodepng.cpp"
           $exitcond39$i = ($113|0)==($4|0); //@line 5406 "libraries/lodepng/lodepng.cpp"
           if ($exitcond39$i) {
            break;
           } else {
            $b3$018$us$i = $113;$ibp$017$us$i = $92;$obp$016$us$i = $112;
           }
          }
          $114 = (($x1$022$us$i) + 1)|0; //@line 5402 "libraries/lodepng/lodepng.cpp"
          $exitcond34 = ($114|0)==($65|0); //@line 5402 "libraries/lodepng/lodepng.cpp"
          if ($exitcond34) {
           break;
          } else {
           $x1$022$us$i = $114;
          }
         }
        }
        $115 = (($y2$025$i) + 1)|0; //@line 5401 "libraries/lodepng/lodepng.cpp"
        $exitcond35 = ($115|0)==($67|0); //@line 5401 "libraries/lodepng/lodepng.cpp"
        if ($exitcond35) {
         break;
        } else {
         $y2$025$i = $115;
        }
       }
      }
     }
     $116 = (($i$128$i) + 1)|0; //@line 5395 "libraries/lodepng/lodepng.cpp"
     $117 = ($116|0)==(7); //@line 5395 "libraries/lodepng/lodepng.cpp"
     if ($117) {
      break;
     } else {
      $i$128$i = $116;
     }
    }
   }
   $118 = ($4>>>0)<(8); //@line 5479 "libraries/lodepng/lodepng.cpp"
   if ($118) {
    $i$015$us = 0;
    while(1) {
     $119 = (($i$015$us) + 1)|0; //@line 5481 "libraries/lodepng/lodepng.cpp"
     $120 = (($padded_passstart) + ($119<<2)|0); //@line 5481 "libraries/lodepng/lodepng.cpp"
     $121 = HEAP32[$120>>2]|0; //@line 5481 "libraries/lodepng/lodepng.cpp"
     $122 = (($padded_passstart) + ($i$015$us<<2)|0); //@line 5481 "libraries/lodepng/lodepng.cpp"
     $123 = HEAP32[$122>>2]|0; //@line 5481 "libraries/lodepng/lodepng.cpp"
     $124 = (($121) - ($123))|0; //@line 5481 "libraries/lodepng/lodepng.cpp"
     $125 = (_malloc($124)|0); //@line 66 "libraries/lodepng/lodepng.cpp"
     $126 = ($125|0)==(0|0); //@line 5482 "libraries/lodepng/lodepng.cpp"
     if ($126) {
      break L22;
     }
     $127 = (($passstart) + ($i$015$us<<2)|0); //@line 5483 "libraries/lodepng/lodepng.cpp"
     $128 = HEAP32[$127>>2]|0; //@line 5483 "libraries/lodepng/lodepng.cpp"
     $129 = (($26) + ($128)|0); //@line 5483 "libraries/lodepng/lodepng.cpp"
     $130 = (($passw) + ($i$015$us<<2)|0); //@line 5484 "libraries/lodepng/lodepng.cpp"
     $131 = HEAP32[$130>>2]|0; //@line 5484 "libraries/lodepng/lodepng.cpp"
     $132 = Math_imul($131, $4)|0; //@line 5484 "libraries/lodepng/lodepng.cpp"
     $133 = (($132) + 7)|0; //@line 5484 "libraries/lodepng/lodepng.cpp"
     $134 = $133 & -8; //@line 5484 "libraries/lodepng/lodepng.cpp"
     $135 = (($passh) + ($i$015$us<<2)|0); //@line 5484 "libraries/lodepng/lodepng.cpp"
     $136 = HEAP32[$135>>2]|0; //@line 5484 "libraries/lodepng/lodepng.cpp"
     __ZL14addPaddingBitsPhPKhjjj($125,$129,$134,$132,$136); //@line 5483 "libraries/lodepng/lodepng.cpp"
     $137 = (($filter_passstart) + ($i$015$us<<2)|0); //@line 5485 "libraries/lodepng/lodepng.cpp"
     $138 = HEAP32[$137>>2]|0; //@line 5485 "libraries/lodepng/lodepng.cpp"
     $139 = HEAP32[$out>>2]|0; //@line 5485 "libraries/lodepng/lodepng.cpp"
     $140 = (($139) + ($138)|0); //@line 5485 "libraries/lodepng/lodepng.cpp"
     $$idx5$val$us = HEAP32[$0>>2]|0;
     $$idx6$val$us = HEAP32[$2>>2]|0;
     $141 = (__ZL6filterPhPKhjjPK16LodePNGColorModePK22LodePNGEncoderSettings($140,$125,$131,$136,$$idx5$val$us,$$idx6$val$us,$settings)|0); //@line 5485 "libraries/lodepng/lodepng.cpp"
     _free($125); //@line 76 "libraries/lodepng/lodepng.cpp"
     $142 = ($141|0)!=(0); //@line 5495 "libraries/lodepng/lodepng.cpp"
     $143 = ($119|0)==(7); //@line 5477 "libraries/lodepng/lodepng.cpp"
     $or$cond = $142 | $143;
     if ($or$cond) {
      break;
     } else {
      $i$015$us = $119;
     }
    }
   } else {
    $i$015 = 0;
    while(1) {
     $144 = (($filter_passstart) + ($i$015<<2)|0); //@line 5491 "libraries/lodepng/lodepng.cpp"
     $145 = HEAP32[$144>>2]|0; //@line 5491 "libraries/lodepng/lodepng.cpp"
     $146 = HEAP32[$out>>2]|0; //@line 5491 "libraries/lodepng/lodepng.cpp"
     $147 = (($146) + ($145)|0); //@line 5491 "libraries/lodepng/lodepng.cpp"
     $148 = (($padded_passstart) + ($i$015<<2)|0); //@line 5491 "libraries/lodepng/lodepng.cpp"
     $149 = HEAP32[$148>>2]|0; //@line 5491 "libraries/lodepng/lodepng.cpp"
     $150 = (($26) + ($149)|0); //@line 5491 "libraries/lodepng/lodepng.cpp"
     $151 = (($passw) + ($i$015<<2)|0); //@line 5492 "libraries/lodepng/lodepng.cpp"
     $152 = HEAP32[$151>>2]|0; //@line 5492 "libraries/lodepng/lodepng.cpp"
     $153 = (($passh) + ($i$015<<2)|0); //@line 5492 "libraries/lodepng/lodepng.cpp"
     $154 = HEAP32[$153>>2]|0; //@line 5492 "libraries/lodepng/lodepng.cpp"
     $$idx$val = HEAP32[$0>>2]|0;
     $$idx4$val = HEAP32[$2>>2]|0;
     $155 = (__ZL6filterPhPKhjjPK16LodePNGColorModePK22LodePNGEncoderSettings($147,$150,$152,$154,$$idx$val,$$idx4$val,$settings)|0); //@line 5491 "libraries/lodepng/lodepng.cpp"
     $$old1 = ($155|0)==(0); //@line 5495 "libraries/lodepng/lodepng.cpp"
     if (!($$old1)) {
      break L22;
     }
     $156 = (($i$015) + 1)|0; //@line 5477 "libraries/lodepng/lodepng.cpp"
     $157 = ($156|0)==(7); //@line 5477 "libraries/lodepng/lodepng.cpp"
     if ($157) {
      break;
     } else {
      $i$015 = $156;
     }
    }
   }
  }
 } while(0);
 _free($26); //@line 76 "libraries/lodepng/lodepng.cpp"
 STACKTOP = sp;return;
}
function __ZL14addPaddingBitsPhPKhjjj($out,$in,$olinebits,$ilinebits,$h) {
 $out = $out|0;
 $in = $in|0;
 $olinebits = $olinebits|0;
 $ilinebits = $ilinebits|0;
 $h = $h|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $ibp$06$us = 0, $ibp$12$us = 0, $obp = 0, $obp$promoted = 0, $obp$promoted$us = 0, $x$03$us = 0, $x$14 = 0, $x$14$us = 0;
 var $y$07 = 0, $y$07$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $obp = sp;
 $0 = (($olinebits) - ($ilinebits))|0; //@line 5340 "libraries/lodepng/lodepng.cpp"
 HEAP32[$obp>>2] = 0; //@line 5341 "libraries/lodepng/lodepng.cpp"
 $1 = ($h|0)==(0); //@line 5342 "libraries/lodepng/lodepng.cpp"
 if ($1) {
  STACKTOP = sp;return; //@line 5354 "libraries/lodepng/lodepng.cpp"
 }
 $2 = ($ilinebits|0)==(0); //@line 5345 "libraries/lodepng/lodepng.cpp"
 $3 = ($olinebits|0)==($ilinebits|0); //@line 5352 "libraries/lodepng/lodepng.cpp"
 if ($2) {
  if ($3) {
   STACKTOP = sp;return; //@line 5354 "libraries/lodepng/lodepng.cpp"
  } else {
   $obp$promoted = 0;$y$07 = 0;
  }
  while(1) {
   $34 = $obp$promoted;$x$14 = 0;
   while(1) {
    $33 = $34 & 7; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $35 = $33 ^ 7; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $36 = 1 << $35; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $37 = $36 ^ 255; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $38 = $34 >>> 3; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $39 = (($out) + ($38)|0); //@line 2383 "libraries/lodepng/lodepng.cpp"
    $40 = HEAP8[$39>>0]|0; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $41 = $40&255; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $42 = $37 & $41; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $43 = $42&255; //@line 2383 "libraries/lodepng/lodepng.cpp"
    HEAP8[$39>>0] = $43; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $44 = (($34) + 1)|0; //@line 2385 "libraries/lodepng/lodepng.cpp"
    $45 = (($x$14) + 1)|0; //@line 5352 "libraries/lodepng/lodepng.cpp"
    $46 = ($45|0)==($0|0); //@line 5352 "libraries/lodepng/lodepng.cpp"
    if ($46) {
     break;
    } else {
     $34 = $44;$x$14 = $45;
    }
   }
   $47 = (($obp$promoted) + ($olinebits))|0; //@line 5352 "libraries/lodepng/lodepng.cpp"
   $48 = (($y$07) + 1)|0; //@line 5342 "libraries/lodepng/lodepng.cpp"
   $49 = ($48|0)==($h|0); //@line 5342 "libraries/lodepng/lodepng.cpp"
   if ($49) {
    $$lcssa = $47;
    break;
   } else {
    $obp$promoted = $47;$y$07 = $48;
   }
  }
  HEAP32[$obp>>2] = $$lcssa; //@line 2385 "libraries/lodepng/lodepng.cpp"
  STACKTOP = sp;return; //@line 5354 "libraries/lodepng/lodepng.cpp"
 } else {
  $ibp$06$us = 0;$y$07$us = 0;
 }
 while(1) {
  $ibp$12$us = $ibp$06$us;$x$03$us = 0;
  while(1) {
   $20 = $ibp$12$us >>> 3; //@line 2351 "libraries/lodepng/lodepng.cpp"
   $21 = (($in) + ($20)|0); //@line 2351 "libraries/lodepng/lodepng.cpp"
   $22 = HEAP8[$21>>0]|0; //@line 2351 "libraries/lodepng/lodepng.cpp"
   $23 = $22&255; //@line 2351 "libraries/lodepng/lodepng.cpp"
   $24 = $ibp$12$us & 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
   $25 = $24 ^ 7; //@line 2351 "libraries/lodepng/lodepng.cpp"
   $26 = $23 >>> $25; //@line 2351 "libraries/lodepng/lodepng.cpp"
   $27 = $26 & 1; //@line 2351 "libraries/lodepng/lodepng.cpp"
   $28 = $27&255; //@line 2351 "libraries/lodepng/lodepng.cpp"
   $29 = (($ibp$12$us) + 1)|0; //@line 2352 "libraries/lodepng/lodepng.cpp"
   __ZL22setBitOfReversedStreamPjPhh($obp,$out,$28); //@line 5348 "libraries/lodepng/lodepng.cpp"
   $30 = (($x$03$us) + 1)|0; //@line 5345 "libraries/lodepng/lodepng.cpp"
   $exitcond = ($30|0)==($ilinebits|0); //@line 5345 "libraries/lodepng/lodepng.cpp"
   if ($exitcond) {
    break;
   } else {
    $ibp$12$us = $29;$x$03$us = $30;
   }
  }
  $31 = (($ibp$06$us) + ($ilinebits))|0; //@line 5345 "libraries/lodepng/lodepng.cpp"
  if (!($3)) {
   $obp$promoted$us = HEAP32[$obp>>2]|0; //@line 2385 "libraries/lodepng/lodepng.cpp"
   $5 = $obp$promoted$us;$x$14$us = 0;
   while(1) {
    $4 = $5 & 7; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $6 = $4 ^ 7; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $7 = 1 << $6; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $8 = $7 ^ 255; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $9 = $5 >>> 3; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $10 = (($out) + ($9)|0); //@line 2383 "libraries/lodepng/lodepng.cpp"
    $11 = HEAP8[$10>>0]|0; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $12 = $11&255; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $13 = $8 & $12; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $14 = $13&255; //@line 2383 "libraries/lodepng/lodepng.cpp"
    HEAP8[$10>>0] = $14; //@line 2383 "libraries/lodepng/lodepng.cpp"
    $15 = (($5) + 1)|0; //@line 2385 "libraries/lodepng/lodepng.cpp"
    $16 = (($x$14$us) + 1)|0; //@line 5352 "libraries/lodepng/lodepng.cpp"
    $17 = ($16|0)==($0|0); //@line 5352 "libraries/lodepng/lodepng.cpp"
    if ($17) {
     break;
    } else {
     $5 = $15;$x$14$us = $16;
    }
   }
   $32 = (($0) + ($obp$promoted$us))|0; //@line 5352 "libraries/lodepng/lodepng.cpp"
   HEAP32[$obp>>2] = $32; //@line 2385 "libraries/lodepng/lodepng.cpp"
  }
  $18 = (($y$07$us) + 1)|0; //@line 5342 "libraries/lodepng/lodepng.cpp"
  $19 = ($18|0)==($h|0); //@line 5342 "libraries/lodepng/lodepng.cpp"
  if ($19) {
   break;
  } else {
   $ibp$06$us = $31;$y$07$us = $18;
  }
 }
 STACKTOP = sp;return; //@line 5354 "libraries/lodepng/lodepng.cpp"
}
function __ZL22setBitOfReversedStreamPjPhh($bitpointer,$bitstream,$bit) {
 $bitpointer = $bitpointer|0;
 $bitstream = $bitstream|0;
 $bit = $bit|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($bit<<24>>24)==(0); //@line 2383 "libraries/lodepng/lodepng.cpp"
 $1 = HEAP32[$bitpointer>>2]|0; //@line 2383 "libraries/lodepng/lodepng.cpp"
 $2 = $1 & 7; //@line 2383 "libraries/lodepng/lodepng.cpp"
 $3 = $2 ^ 7; //@line 2383 "libraries/lodepng/lodepng.cpp"
 $4 = 1 << $3; //@line 2383 "libraries/lodepng/lodepng.cpp"
 if ($0) {
  $5 = $4 ^ 255; //@line 2383 "libraries/lodepng/lodepng.cpp"
  $6 = $1 >>> 3; //@line 2383 "libraries/lodepng/lodepng.cpp"
  $7 = (($bitstream) + ($6)|0); //@line 2383 "libraries/lodepng/lodepng.cpp"
  $8 = HEAP8[$7>>0]|0; //@line 2383 "libraries/lodepng/lodepng.cpp"
  $9 = $8&255; //@line 2383 "libraries/lodepng/lodepng.cpp"
  $10 = $9 & $5; //@line 2383 "libraries/lodepng/lodepng.cpp"
  $11 = $10&255; //@line 2383 "libraries/lodepng/lodepng.cpp"
  HEAP8[$7>>0] = $11; //@line 2383 "libraries/lodepng/lodepng.cpp"
  $18 = HEAP32[$bitpointer>>2]|0; //@line 2385 "libraries/lodepng/lodepng.cpp"
  $19 = (($18) + 1)|0; //@line 2385 "libraries/lodepng/lodepng.cpp"
  HEAP32[$bitpointer>>2] = $19; //@line 2385 "libraries/lodepng/lodepng.cpp"
  return; //@line 2386 "libraries/lodepng/lodepng.cpp"
 } else {
  $12 = $1 >>> 3; //@line 2384 "libraries/lodepng/lodepng.cpp"
  $13 = (($bitstream) + ($12)|0); //@line 2384 "libraries/lodepng/lodepng.cpp"
  $14 = HEAP8[$13>>0]|0; //@line 2384 "libraries/lodepng/lodepng.cpp"
  $15 = $14&255; //@line 2384 "libraries/lodepng/lodepng.cpp"
  $16 = $15 | $4; //@line 2384 "libraries/lodepng/lodepng.cpp"
  $17 = $16&255; //@line 2384 "libraries/lodepng/lodepng.cpp"
  HEAP8[$13>>0] = $17; //@line 2384 "libraries/lodepng/lodepng.cpp"
  $18 = HEAP32[$bitpointer>>2]|0; //@line 2385 "libraries/lodepng/lodepng.cpp"
  $19 = (($18) + 1)|0; //@line 2385 "libraries/lodepng/lodepng.cpp"
  HEAP32[$bitpointer>>2] = $19; //@line 2385 "libraries/lodepng/lodepng.cpp"
  return; //@line 2386 "libraries/lodepng/lodepng.cpp"
 }
}
function __ZL6filterPhPKhjjPK16LodePNGColorModePK22LodePNGEncoderSettings($out,$in,$w,$h,$info$0$0$val,$info$0$1$val,$settings) {
 $out = $out|0;
 $in = $in|0;
 $w = $w|0;
 $h = $h|0;
 $info$0$0$val = $info$0$0$val|0;
 $info$0$1$val = $info$0$1$val|0;
 $settings = $settings|0;
 var $$ = 0, $$0$i$i$i = 0, $$0$lcssa$i = 0.0, $$04$i = 0.0, $$1$lcssa$i = 0.0, $$11$i = 0.0, $$2 = 0, $$lcssa140 = 0.0, $$pre = 0, $$pre$phiZ2D = 0, $$pre99 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0.0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0.0, $143 = 0.0, $144 = 0, $145 = 0.0, $146 = 0, $147 = 0, $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0, $151 = 0.0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0;
 var $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0;
 var $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0;
 var $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0;
 var $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0;
 var $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0;
 var $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0;
 var $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $attempt = 0, $attempt10 = 0, $attempt2 = 0, $bestType$032 = 0, $bestType$1$lcssa = 0, $bestType$128 = 0, $bestType$2 = 0, $bestType13$068 = 0, $bestType13$160 = 0, $bestType13$2 = 0, $bestType13$2$lcssa = 0;
 var $bestType5$050 = 0, $bestType5$144 = 0, $bestType5$2 = 0, $bestType5$2$lcssa = 0, $count = 0, $dummy = 0, $i$021$i = 0, $indvars$iv$next92 = 0, $indvars$iv91 = 0, $or$cond = 0, $or$cond112 = 0, $or$cond113 = 0, $or$cond114 = 0, $prevline$134 = 0, $prevline$252 = 0, $prevline$355 = 0, $prevline$466 = 0, $result$0$lcssa$i = 0.0, $result$05$i = 0.0, $result$1$lcssa$i = 0.0;
 var $result$12$i = 0.0, $size = 0, $smallest$031 = 0, $smallest$1$lcssa = 0, $smallest$126 = 0, $smallest$2 = 0, $smallest11$067 = 0, $smallest11$158 = 0, $smallest11$2 = 0, $smallest11$2$lcssa = 0, $smallest3$049 = 0.0, $smallest3$142 = 0.0, $smallest3$2 = 0.0, $smallest3$2$lcssa = 0.0, $strategy$0 = 0, $type12$159 = 0, $type4$143 = 0, $x$024 = 0, $x$122 = 0, $x$229 = 0;
 var $x$438 = 0, $x$541 = 0, $x$645 = 0, $x$761 = 0, $y$020 = 0, $y$020$us = 0, $y$133 = 0, $y$251 = 0, $y$354 = 0, $y$465 = 0, $zlibsettings = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1264|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $attempt = sp + 1204|0;
 $attempt2 = sp + 1144|0;
 $count = sp + 120|0;
 $size = sp + 100|0;
 $attempt10 = sp + 40|0;
 $dummy = sp + 36|0;
 $zlibsettings = sp;
 switch ($info$0$0$val|0) {
 case 3: case 0:  {
  $$0$i$i$i = 1;
  break;
 }
 case 2:  {
  $$0$i$i$i = 3;
  break;
 }
 case 4:  {
  $$0$i$i$i = 2;
  break;
 }
 case 6:  {
  $$0$i$i$i = 4;
  break;
 }
 default: {
  $$0$i$i$i = 0;
 }
 }
 $0 = Math_imul($$0$i$i$i, $info$0$1$val)|0; //@line 2549 "libraries/lodepng/lodepng.cpp"
 $1 = Math_imul($0, $w)|0; //@line 5125 "libraries/lodepng/lodepng.cpp"
 $2 = (($1) + 7)|0; //@line 5125 "libraries/lodepng/lodepng.cpp"
 $3 = $2 >>> 3; //@line 5125 "libraries/lodepng/lodepng.cpp"
 $4 = (($0) + 7)|0; //@line 5127 "libraries/lodepng/lodepng.cpp"
 $5 = $4 >>> 3; //@line 5127 "libraries/lodepng/lodepng.cpp"
 $6 = ((($settings)) + 44|0); //@line 5131 "libraries/lodepng/lodepng.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 5131 "libraries/lodepng/lodepng.cpp"
 $8 = ((($settings)) + 40|0); //@line 5146 "libraries/lodepng/lodepng.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 5146 "libraries/lodepng/lodepng.cpp"
 $10 = ($9|0)==(0); //@line 5146 "libraries/lodepng/lodepng.cpp"
 if ($10) {
  $strategy$0 = $7;
 } else {
  $11 = ($info$0$0$val|0)==(3); //@line 5147 "libraries/lodepng/lodepng.cpp"
  $12 = ($info$0$1$val>>>0)<(8); //@line 5147 "libraries/lodepng/lodepng.cpp"
  $or$cond = $11 | $12; //@line 5147 "libraries/lodepng/lodepng.cpp"
  $$ = $or$cond ? 0 : $7; //@line 5147 "libraries/lodepng/lodepng.cpp"
  $strategy$0 = $$;
 }
 $13 = ($0|0)==(0); //@line 5149 "libraries/lodepng/lodepng.cpp"
 if ($13) {
  $$2 = 31;
  STACKTOP = sp;return ($$2|0); //@line 5332 "libraries/lodepng/lodepng.cpp"
 }
 switch ($strategy$0|0) {
 case 0:  {
  $17 = ($h|0)==(0); //@line 5153 "libraries/lodepng/lodepng.cpp"
  if ($17) {
   $$2 = 0;
   STACKTOP = sp;return ($$2|0); //@line 5332 "libraries/lodepng/lodepng.cpp"
  }
  $18 = (($3) + 1)|0; //@line 5155 "libraries/lodepng/lodepng.cpp"
  $19 = ($3|0)==(0); //@line 5056 "libraries/lodepng/lodepng.cpp"
  if ($19) {
   $y$020$us = 0;
   while(1) {
    $20 = Math_imul($y$020$us, $18)|0; //@line 5155 "libraries/lodepng/lodepng.cpp"
    $21 = (($out) + ($20)|0); //@line 5157 "libraries/lodepng/lodepng.cpp"
    HEAP8[$21>>0] = 0; //@line 5157 "libraries/lodepng/lodepng.cpp"
    $22 = (($y$020$us) + 1)|0; //@line 5153 "libraries/lodepng/lodepng.cpp"
    $23 = ($22|0)==($h|0); //@line 5153 "libraries/lodepng/lodepng.cpp"
    if ($23) {
     $$2 = 0;
     break;
    } else {
     $y$020$us = $22;
    }
   }
   STACKTOP = sp;return ($$2|0); //@line 5332 "libraries/lodepng/lodepng.cpp"
  } else {
   $y$020 = 0;
  }
  while(1) {
   $24 = Math_imul($y$020, $18)|0; //@line 5155 "libraries/lodepng/lodepng.cpp"
   $25 = Math_imul($y$020, $3)|0; //@line 5156 "libraries/lodepng/lodepng.cpp"
   $26 = (($out) + ($24)|0); //@line 5157 "libraries/lodepng/lodepng.cpp"
   HEAP8[$26>>0] = 0; //@line 5157 "libraries/lodepng/lodepng.cpp"
   $27 = (($24) + 1)|0; //@line 5158 "libraries/lodepng/lodepng.cpp"
   $28 = (($out) + ($27)|0); //@line 5158 "libraries/lodepng/lodepng.cpp"
   $29 = (($in) + ($25)|0); //@line 5158 "libraries/lodepng/lodepng.cpp"
   $i$021$i = 0;
   while(1) {
    $30 = (($29) + ($i$021$i)|0); //@line 5056 "libraries/lodepng/lodepng.cpp"
    $31 = HEAP8[$30>>0]|0; //@line 5056 "libraries/lodepng/lodepng.cpp"
    $32 = (($28) + ($i$021$i)|0); //@line 5056 "libraries/lodepng/lodepng.cpp"
    HEAP8[$32>>0] = $31; //@line 5056 "libraries/lodepng/lodepng.cpp"
    $33 = (($i$021$i) + 1)|0; //@line 5056 "libraries/lodepng/lodepng.cpp"
    $34 = ($33|0)==($3|0); //@line 5056 "libraries/lodepng/lodepng.cpp"
    if ($34) {
     break;
    } else {
     $i$021$i = $33;
    }
   }
   $35 = (($y$020) + 1)|0; //@line 5153 "libraries/lodepng/lodepng.cpp"
   $36 = ($35|0)==($h|0); //@line 5153 "libraries/lodepng/lodepng.cpp"
   if ($36) {
    $$2 = 0;
    break;
   } else {
    $y$020 = $35;
   }
  }
  STACKTOP = sp;return ($$2|0); //@line 5332 "libraries/lodepng/lodepng.cpp"
  break;
 }
 case 1:  {
  $37 = ($3|0)==(0); //@line 225 "libraries/lodepng/lodepng.cpp"
  $38 = ($3|0)!=(0); //@line 227 "libraries/lodepng/lodepng.cpp"
  $39 = ($3*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
  $40 = $39 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
  $41 = $38 ? $3 : $40; //@line 227 "libraries/lodepng/lodepng.cpp"
  do {
   if ($37) {
    dest=$attempt; stop=dest+60|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 259 "libraries/lodepng/lodepng.cpp"
    $102 = 0;$103 = 0;$104 = 0;$105 = 0;$106 = 0;
   } else {
    HEAP32[$attempt>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
    $42 = ((($attempt)) + 8|0); //@line 259 "libraries/lodepng/lodepng.cpp"
    HEAP32[$42>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
    $43 = ((($attempt)) + 4|0); //@line 259 "libraries/lodepng/lodepng.cpp"
    HEAP32[$43>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
    $44 = (_realloc(0,$41)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $45 = ($44|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if (!($45)) {
     HEAP32[$42>>2] = $41; //@line 231 "libraries/lodepng/lodepng.cpp"
     HEAP32[$attempt>>2] = $44; //@line 232 "libraries/lodepng/lodepng.cpp"
     HEAP32[$43>>2] = $3; //@line 243 "libraries/lodepng/lodepng.cpp"
     $52 = ((($attempt)) + 12|0); //@line 258 "libraries/lodepng/lodepng.cpp"
     HEAP32[$52>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
     $53 = ((($attempt)) + 20|0); //@line 259 "libraries/lodepng/lodepng.cpp"
     HEAP32[$53>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
     $54 = ((($attempt)) + 16|0); //@line 259 "libraries/lodepng/lodepng.cpp"
     HEAP32[$54>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
     $55 = (_realloc(0,$41)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
     $56 = ($55|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
     if (!($56)) {
      HEAP32[$53>>2] = $41; //@line 231 "libraries/lodepng/lodepng.cpp"
      HEAP32[$52>>2] = $55; //@line 232 "libraries/lodepng/lodepng.cpp"
      HEAP32[$54>>2] = $3; //@line 243 "libraries/lodepng/lodepng.cpp"
      $271 = ((($attempt)) + 24|0); //@line 258 "libraries/lodepng/lodepng.cpp"
      HEAP32[$271>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
      $272 = ((($attempt)) + 32|0); //@line 259 "libraries/lodepng/lodepng.cpp"
      HEAP32[$272>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
      $273 = ((($attempt)) + 28|0); //@line 259 "libraries/lodepng/lodepng.cpp"
      HEAP32[$273>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
      $274 = (_realloc(0,$41)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $275 = ($274|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
      if (!($275)) {
       HEAP32[$272>>2] = $41; //@line 231 "libraries/lodepng/lodepng.cpp"
       HEAP32[$271>>2] = $274; //@line 232 "libraries/lodepng/lodepng.cpp"
       HEAP32[$273>>2] = $3; //@line 243 "libraries/lodepng/lodepng.cpp"
       $276 = ((($attempt)) + 36|0); //@line 258 "libraries/lodepng/lodepng.cpp"
       HEAP32[$276>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
       $277 = ((($attempt)) + 44|0); //@line 259 "libraries/lodepng/lodepng.cpp"
       HEAP32[$277>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
       $278 = ((($attempt)) + 40|0); //@line 259 "libraries/lodepng/lodepng.cpp"
       HEAP32[$278>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
       $279 = (_realloc(0,$41)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       $280 = ($279|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
       if (!($280)) {
        HEAP32[$277>>2] = $41; //@line 231 "libraries/lodepng/lodepng.cpp"
        HEAP32[$276>>2] = $279; //@line 232 "libraries/lodepng/lodepng.cpp"
        HEAP32[$278>>2] = $3; //@line 243 "libraries/lodepng/lodepng.cpp"
        $281 = ((($attempt)) + 48|0); //@line 258 "libraries/lodepng/lodepng.cpp"
        HEAP32[$281>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
        $282 = ((($attempt)) + 56|0); //@line 259 "libraries/lodepng/lodepng.cpp"
        HEAP32[$282>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
        $283 = ((($attempt)) + 52|0); //@line 259 "libraries/lodepng/lodepng.cpp"
        HEAP32[$283>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
        $284 = (_realloc(0,$41)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
        $285 = ($284|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
        if (!($285)) {
         HEAP32[$282>>2] = $41; //@line 231 "libraries/lodepng/lodepng.cpp"
         HEAP32[$281>>2] = $284; //@line 232 "libraries/lodepng/lodepng.cpp"
         HEAP32[$283>>2] = $3; //@line 243 "libraries/lodepng/lodepng.cpp"
         $102 = $44;$103 = $55;$104 = $274;$105 = $279;$106 = $284;
         break;
        }
       }
      }
     }
    }
    $$2 = 83;
    STACKTOP = sp;return ($$2|0); //@line 5332 "libraries/lodepng/lodepng.cpp"
   }
  } while(0);
  $46 = ($h|0)==(0); //@line 5178 "libraries/lodepng/lodepng.cpp"
  if (!($46)) {
   $47 = (($3) + 1)|0; //@line 5214 "libraries/lodepng/lodepng.cpp"
   $48 = ((($attempt)) + 12|0); //@line 5183 "libraries/lodepng/lodepng.cpp"
   $49 = ((($attempt)) + 24|0); //@line 5183 "libraries/lodepng/lodepng.cpp"
   $50 = ((($attempt)) + 36|0); //@line 5183 "libraries/lodepng/lodepng.cpp"
   $51 = ((($attempt)) + 48|0); //@line 5183 "libraries/lodepng/lodepng.cpp"
   $bestType$032 = 0;$prevline$134 = 0;$smallest$031 = 0;$y$133 = 0;
   while(1) {
    $57 = Math_imul($y$133, $3)|0; //@line 5183 "libraries/lodepng/lodepng.cpp"
    $58 = (($in) + ($57)|0); //@line 5183 "libraries/lodepng/lodepng.cpp"
    if ($37) {
     $59 = HEAP32[$attempt>>2]|0; //@line 5183 "libraries/lodepng/lodepng.cpp"
     __ZL14filterScanlinePhPKhS1_jjh($59,$58,$prevline$134,0,$5,0); //@line 5183 "libraries/lodepng/lodepng.cpp"
     $60 = HEAP32[$48>>2]|0; //@line 5183 "libraries/lodepng/lodepng.cpp"
     __ZL14filterScanlinePhPKhS1_jjh($60,$58,$prevline$134,0,$5,1); //@line 5183 "libraries/lodepng/lodepng.cpp"
     $61 = HEAP32[$49>>2]|0; //@line 5183 "libraries/lodepng/lodepng.cpp"
     __ZL14filterScanlinePhPKhS1_jjh($61,$58,$prevline$134,0,$5,2); //@line 5183 "libraries/lodepng/lodepng.cpp"
     $62 = HEAP32[$50>>2]|0; //@line 5183 "libraries/lodepng/lodepng.cpp"
     __ZL14filterScanlinePhPKhS1_jjh($62,$58,$prevline$134,0,$5,3); //@line 5183 "libraries/lodepng/lodepng.cpp"
     $63 = HEAP32[$51>>2]|0; //@line 5183 "libraries/lodepng/lodepng.cpp"
     __ZL14filterScanlinePhPKhS1_jjh($63,$58,$prevline$134,0,$5,4); //@line 5183 "libraries/lodepng/lodepng.cpp"
     $bestType$1$lcssa = 0;$smallest$1$lcssa = 0;
    } else {
     $bestType$128 = $bestType$032;$indvars$iv91 = 0;$smallest$126 = $smallest$031;
     while(1) {
      $64 = (($attempt) + (($indvars$iv91*12)|0)|0); //@line 5183 "libraries/lodepng/lodepng.cpp"
      $65 = HEAP32[$64>>2]|0; //@line 5189 "libraries/lodepng/lodepng.cpp"
      $66 = $indvars$iv91&255; //@line 5183 "libraries/lodepng/lodepng.cpp"
      __ZL14filterScanlinePhPKhS1_jjh($65,$58,$prevline$134,$3,$5,$66); //@line 5183 "libraries/lodepng/lodepng.cpp"
      $67 = ($indvars$iv91|0)==(0); //@line 5187 "libraries/lodepng/lodepng.cpp"
      if ($67) {
       $72 = 0;$x$024 = 0;
       while(1) {
        $68 = (($65) + ($x$024)|0); //@line 5189 "libraries/lodepng/lodepng.cpp"
        $69 = HEAP8[$68>>0]|0; //@line 5189 "libraries/lodepng/lodepng.cpp"
        $70 = $69&255; //@line 5189 "libraries/lodepng/lodepng.cpp"
        $71 = (($72) + ($70))|0; //@line 5189 "libraries/lodepng/lodepng.cpp"
        $73 = (($x$024) + 1)|0; //@line 5189 "libraries/lodepng/lodepng.cpp"
        $74 = ($73|0)==($3|0); //@line 5189 "libraries/lodepng/lodepng.cpp"
        if ($74) {
         $86 = $71;
         break;
        } else {
         $72 = $71;$x$024 = $73;
        }
       }
      } else {
       $82 = 0;$x$122 = 0;
       while(1) {
        $75 = (($65) + ($x$122)|0); //@line 5198 "libraries/lodepng/lodepng.cpp"
        $76 = HEAP8[$75>>0]|0; //@line 5198 "libraries/lodepng/lodepng.cpp"
        $77 = $76&255; //@line 5199 "libraries/lodepng/lodepng.cpp"
        $78 = ($76<<24>>24)>(-1); //@line 5199 "libraries/lodepng/lodepng.cpp"
        $79 = $77 ^ 255; //@line 5199 "libraries/lodepng/lodepng.cpp"
        $80 = $78 ? $77 : $79; //@line 5199 "libraries/lodepng/lodepng.cpp"
        $81 = (($80) + ($82))|0; //@line 5199 "libraries/lodepng/lodepng.cpp"
        $83 = (($x$122) + 1)|0; //@line 5193 "libraries/lodepng/lodepng.cpp"
        $84 = ($83|0)==($3|0); //@line 5193 "libraries/lodepng/lodepng.cpp"
        if ($84) {
         $86 = $81;
         break;
        } else {
         $82 = $81;$x$122 = $83;
        }
       }
      }
      $85 = ($86>>>0)<($smallest$126>>>0); //@line 5204 "libraries/lodepng/lodepng.cpp"
      $or$cond112 = $67 | $85; //@line 5204 "libraries/lodepng/lodepng.cpp"
      $smallest$2 = $or$cond112 ? $86 : $smallest$126; //@line 5204 "libraries/lodepng/lodepng.cpp"
      $bestType$2 = $or$cond112 ? $66 : $bestType$128; //@line 5204 "libraries/lodepng/lodepng.cpp"
      $indvars$iv$next92 = (($indvars$iv91) + 1)|0; //@line 5181 "libraries/lodepng/lodepng.cpp"
      $87 = ($indvars$iv$next92|0)==(5); //@line 5181 "libraries/lodepng/lodepng.cpp"
      if ($87) {
       $bestType$1$lcssa = $bestType$2;$smallest$1$lcssa = $smallest$2;
       break;
      } else {
       $bestType$128 = $bestType$2;$indvars$iv91 = $indvars$iv$next92;$smallest$126 = $smallest$2;
      }
     }
    }
    $88 = Math_imul($y$133, $47)|0; //@line 5214 "libraries/lodepng/lodepng.cpp"
    $89 = (($out) + ($88)|0); //@line 5214 "libraries/lodepng/lodepng.cpp"
    HEAP8[$89>>0] = $bestType$1$lcssa; //@line 5214 "libraries/lodepng/lodepng.cpp"
    if (!($37)) {
     $90 = $bestType$1$lcssa&255; //@line 5215 "libraries/lodepng/lodepng.cpp"
     $91 = (($attempt) + (($90*12)|0)|0); //@line 5215 "libraries/lodepng/lodepng.cpp"
     $92 = HEAP32[$91>>2]|0; //@line 5215 "libraries/lodepng/lodepng.cpp"
     $93 = (($88) + 1)|0; //@line 5215 "libraries/lodepng/lodepng.cpp"
     $x$229 = 0;
     while(1) {
      $94 = (($92) + ($x$229)|0); //@line 5215 "libraries/lodepng/lodepng.cpp"
      $95 = HEAP8[$94>>0]|0; //@line 5215 "libraries/lodepng/lodepng.cpp"
      $96 = (($93) + ($x$229))|0; //@line 5215 "libraries/lodepng/lodepng.cpp"
      $97 = (($out) + ($96)|0); //@line 5215 "libraries/lodepng/lodepng.cpp"
      HEAP8[$97>>0] = $95; //@line 5215 "libraries/lodepng/lodepng.cpp"
      $98 = (($x$229) + 1)|0; //@line 5215 "libraries/lodepng/lodepng.cpp"
      $99 = ($98|0)==($3|0); //@line 5215 "libraries/lodepng/lodepng.cpp"
      if ($99) {
       break;
      } else {
       $x$229 = $98;
      }
     }
    }
    $100 = (($y$133) + 1)|0; //@line 5178 "libraries/lodepng/lodepng.cpp"
    $101 = ($100|0)==($h|0); //@line 5178 "libraries/lodepng/lodepng.cpp"
    if ($101) {
     break;
    } else {
     $bestType$032 = $bestType$1$lcssa;$prevline$134 = $58;$smallest$031 = $smallest$1$lcssa;$y$133 = $100;
    }
   }
  }
  _free($102); //@line 76 "libraries/lodepng/lodepng.cpp"
  _free($103); //@line 76 "libraries/lodepng/lodepng.cpp"
  _free($104); //@line 76 "libraries/lodepng/lodepng.cpp"
  _free($105); //@line 76 "libraries/lodepng/lodepng.cpp"
  _free($106); //@line 76 "libraries/lodepng/lodepng.cpp"
  $$2 = 0;
  STACKTOP = sp;return ($$2|0); //@line 5332 "libraries/lodepng/lodepng.cpp"
  break;
 }
 case 2:  {
  $107 = ($3|0)==(0); //@line 225 "libraries/lodepng/lodepng.cpp"
  $108 = ($3|0)!=(0); //@line 227 "libraries/lodepng/lodepng.cpp"
  $109 = ($3*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
  $110 = $109 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
  $111 = $108 ? $3 : $110; //@line 227 "libraries/lodepng/lodepng.cpp"
  do {
   if ($107) {
    dest=$attempt2; stop=dest+60|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 259 "libraries/lodepng/lodepng.cpp"
    $189 = 0;$190 = 0;$191 = 0;$192 = 0;$193 = 0;
   } else {
    HEAP32[$attempt2>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
    $112 = ((($attempt2)) + 8|0); //@line 259 "libraries/lodepng/lodepng.cpp"
    HEAP32[$112>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
    $113 = ((($attempt2)) + 4|0); //@line 259 "libraries/lodepng/lodepng.cpp"
    HEAP32[$113>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
    $114 = (_realloc(0,$111)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $115 = ($114|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if (!($115)) {
     HEAP32[$112>>2] = $111; //@line 231 "libraries/lodepng/lodepng.cpp"
     HEAP32[$attempt2>>2] = $114; //@line 232 "libraries/lodepng/lodepng.cpp"
     HEAP32[$113>>2] = $3; //@line 243 "libraries/lodepng/lodepng.cpp"
     $119 = ((($attempt2)) + 12|0); //@line 258 "libraries/lodepng/lodepng.cpp"
     HEAP32[$119>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
     $120 = ((($attempt2)) + 20|0); //@line 259 "libraries/lodepng/lodepng.cpp"
     HEAP32[$120>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
     $121 = ((($attempt2)) + 16|0); //@line 259 "libraries/lodepng/lodepng.cpp"
     HEAP32[$121>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
     $122 = (_realloc(0,$111)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
     $123 = ($122|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
     if (!($123)) {
      HEAP32[$120>>2] = $111; //@line 231 "libraries/lodepng/lodepng.cpp"
      HEAP32[$119>>2] = $122; //@line 232 "libraries/lodepng/lodepng.cpp"
      HEAP32[$121>>2] = $3; //@line 243 "libraries/lodepng/lodepng.cpp"
      $286 = ((($attempt2)) + 24|0); //@line 258 "libraries/lodepng/lodepng.cpp"
      HEAP32[$286>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
      $287 = ((($attempt2)) + 32|0); //@line 259 "libraries/lodepng/lodepng.cpp"
      HEAP32[$287>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
      $288 = ((($attempt2)) + 28|0); //@line 259 "libraries/lodepng/lodepng.cpp"
      HEAP32[$288>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
      $289 = (_realloc(0,$111)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $290 = ($289|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
      if (!($290)) {
       HEAP32[$287>>2] = $111; //@line 231 "libraries/lodepng/lodepng.cpp"
       HEAP32[$286>>2] = $289; //@line 232 "libraries/lodepng/lodepng.cpp"
       HEAP32[$288>>2] = $3; //@line 243 "libraries/lodepng/lodepng.cpp"
       $291 = ((($attempt2)) + 36|0); //@line 258 "libraries/lodepng/lodepng.cpp"
       HEAP32[$291>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
       $292 = ((($attempt2)) + 44|0); //@line 259 "libraries/lodepng/lodepng.cpp"
       HEAP32[$292>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
       $293 = ((($attempt2)) + 40|0); //@line 259 "libraries/lodepng/lodepng.cpp"
       HEAP32[$293>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
       $294 = (_realloc(0,$111)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
       $295 = ($294|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
       if (!($295)) {
        HEAP32[$292>>2] = $111; //@line 231 "libraries/lodepng/lodepng.cpp"
        HEAP32[$291>>2] = $294; //@line 232 "libraries/lodepng/lodepng.cpp"
        HEAP32[$293>>2] = $3; //@line 243 "libraries/lodepng/lodepng.cpp"
        $296 = ((($attempt2)) + 48|0); //@line 258 "libraries/lodepng/lodepng.cpp"
        HEAP32[$296>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
        $297 = ((($attempt2)) + 56|0); //@line 259 "libraries/lodepng/lodepng.cpp"
        HEAP32[$297>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
        $298 = ((($attempt2)) + 52|0); //@line 259 "libraries/lodepng/lodepng.cpp"
        HEAP32[$298>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
        $299 = (_realloc(0,$111)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
        $300 = ($299|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
        if (!($300)) {
         HEAP32[$297>>2] = $111; //@line 231 "libraries/lodepng/lodepng.cpp"
         HEAP32[$296>>2] = $299; //@line 232 "libraries/lodepng/lodepng.cpp"
         HEAP32[$298>>2] = $3; //@line 243 "libraries/lodepng/lodepng.cpp"
         $189 = $114;$190 = $122;$191 = $289;$192 = $294;$193 = $299;
         break;
        }
       }
      }
     }
    }
    $$2 = 83;
    STACKTOP = sp;return ($$2|0); //@line 5332 "libraries/lodepng/lodepng.cpp"
   }
  } while(0);
  $116 = ($h|0)==(0); //@line 5235 "libraries/lodepng/lodepng.cpp"
  if (!($116)) {
   $117 = (($3) + 1)|0; //@line 5247 "libraries/lodepng/lodepng.cpp"
   $118 = (+($117>>>0)); //@line 5247 "libraries/lodepng/lodepng.cpp"
   $bestType5$050 = 0;$prevline$252 = 0;$smallest3$049 = 0.0;$y$251 = 0;
   while(1) {
    $124 = Math_imul($y$251, $3)|0; //@line 5240 "libraries/lodepng/lodepng.cpp"
    $125 = (($in) + ($124)|0); //@line 5240 "libraries/lodepng/lodepng.cpp"
    $bestType5$144 = $bestType5$050;$smallest3$142 = $smallest3$049;$type4$143 = 0;
    while(1) {
     $126 = (($attempt2) + (($type4$143*12)|0)|0); //@line 5240 "libraries/lodepng/lodepng.cpp"
     $127 = HEAP32[$126>>2]|0; //@line 5242 "libraries/lodepng/lodepng.cpp"
     $128 = $type4$143&255; //@line 5240 "libraries/lodepng/lodepng.cpp"
     __ZL14filterScanlinePhPKhS1_jjh($127,$125,$prevline$252,$3,$5,$128); //@line 5240 "libraries/lodepng/lodepng.cpp"
     _memset(($count|0),0,1024)|0; //@line 5241 "libraries/lodepng/lodepng.cpp"
     if (!($107)) {
      $x$438 = 0;
      while(1) {
       $129 = (($127) + ($x$438)|0); //@line 5242 "libraries/lodepng/lodepng.cpp"
       $130 = HEAP8[$129>>0]|0; //@line 5242 "libraries/lodepng/lodepng.cpp"
       $131 = $130&255; //@line 5242 "libraries/lodepng/lodepng.cpp"
       $132 = (($count) + ($131<<2)|0); //@line 5242 "libraries/lodepng/lodepng.cpp"
       $133 = HEAP32[$132>>2]|0; //@line 5242 "libraries/lodepng/lodepng.cpp"
       $134 = (($133) + 1)|0; //@line 5242 "libraries/lodepng/lodepng.cpp"
       HEAP32[$132>>2] = $134; //@line 5242 "libraries/lodepng/lodepng.cpp"
       $135 = (($x$438) + 1)|0; //@line 5242 "libraries/lodepng/lodepng.cpp"
       $136 = ($135|0)==($3|0); //@line 5242 "libraries/lodepng/lodepng.cpp"
       if ($136) {
        break;
       } else {
        $x$438 = $135;
       }
      }
     }
     $137 = (($count) + ($type4$143<<2)|0); //@line 5243 "libraries/lodepng/lodepng.cpp"
     $138 = HEAP32[$137>>2]|0; //@line 5243 "libraries/lodepng/lodepng.cpp"
     $139 = (($138) + 1)|0; //@line 5243 "libraries/lodepng/lodepng.cpp"
     HEAP32[$137>>2] = $139; //@line 5243 "libraries/lodepng/lodepng.cpp"
     $168 = 0.0;$x$541 = 0;
     while(1) {
      $140 = (($count) + ($x$541<<2)|0); //@line 5247 "libraries/lodepng/lodepng.cpp"
      $141 = HEAP32[$140>>2]|0; //@line 5247 "libraries/lodepng/lodepng.cpp"
      $142 = (+($141>>>0)); //@line 5247 "libraries/lodepng/lodepng.cpp"
      $143 = $142 / $118; //@line 5247 "libraries/lodepng/lodepng.cpp"
      $144 = ($141|0)==(0); //@line 5248 "libraries/lodepng/lodepng.cpp"
      if ($144) {
       $167 = 0.0;
      } else {
       $145 = 1.0 / $143; //@line 5248 "libraries/lodepng/lodepng.cpp"
       $146 = $145 > 32.0; //@line 5109 "libraries/lodepng/lodepng.cpp"
       if ($146) {
        $$04$i = $145;$result$05$i = 0.0;
        while(1) {
         $148 = $result$05$i + 4.0; //@line 5109 "libraries/lodepng/lodepng.cpp"
         $149 = $$04$i * 0.0625; //@line 5109 "libraries/lodepng/lodepng.cpp"
         $150 = $149 > 32.0; //@line 5109 "libraries/lodepng/lodepng.cpp"
         if ($150) {
          $$04$i = $149;$result$05$i = $148;
         } else {
          $$0$lcssa$i = $149;$result$0$lcssa$i = $148;
          break;
         }
        }
       } else {
        $$0$lcssa$i = $145;$result$0$lcssa$i = 0.0;
       }
       $147 = $$0$lcssa$i > 2.0; //@line 5110 "libraries/lodepng/lodepng.cpp"
       if ($147) {
        $$11$i = $$0$lcssa$i;$result$12$i = $result$0$lcssa$i;
        while(1) {
         $151 = $result$12$i + 1.0; //@line 5110 "libraries/lodepng/lodepng.cpp"
         $152 = $$11$i * 0.5; //@line 5110 "libraries/lodepng/lodepng.cpp"
         $153 = $152 > 2.0; //@line 5110 "libraries/lodepng/lodepng.cpp"
         if ($153) {
          $$11$i = $152;$result$12$i = $151;
         } else {
          $$1$lcssa$i = $152;$result$1$lcssa$i = $151;
          break;
         }
        }
       } else {
        $$1$lcssa$i = $$0$lcssa$i;$result$1$lcssa$i = $result$0$lcssa$i;
       }
       $154 = $$1$lcssa$i * $$1$lcssa$i; //@line 5111 "libraries/lodepng/lodepng.cpp"
       $155 = $$1$lcssa$i * $154; //@line 5111 "libraries/lodepng/lodepng.cpp"
       $156 = $155 / 3.0; //@line 5111 "libraries/lodepng/lodepng.cpp"
       $157 = $$1$lcssa$i * 3.0; //@line 5111 "libraries/lodepng/lodepng.cpp"
       $158 = $$1$lcssa$i * $157; //@line 5111 "libraries/lodepng/lodepng.cpp"
       $159 = $158 * 0.5; //@line 5111 "libraries/lodepng/lodepng.cpp"
       $160 = $156 - $159; //@line 5111 "libraries/lodepng/lodepng.cpp"
       $161 = $157 + $160; //@line 5111 "libraries/lodepng/lodepng.cpp"
       $162 = $161 + -1.8333300352096558; //@line 5111 "libraries/lodepng/lodepng.cpp"
       $163 = $162 * 1.4426950216293335; //@line 5111 "libraries/lodepng/lodepng.cpp"
       $164 = $result$1$lcssa$i + $163; //@line 5111 "libraries/lodepng/lodepng.cpp"
       $165 = $143 * $164; //@line 5248 "libraries/lodepng/lodepng.cpp"
       $167 = $165;
      }
      $166 = $167 + $168; //@line 5254 "libraries/lodepng/lodepng.cpp"
      $169 = (($x$541) + 1)|0; //@line 5245 "libraries/lodepng/lodepng.cpp"
      $170 = ($169|0)==(256); //@line 5245 "libraries/lodepng/lodepng.cpp"
      if ($170) {
       $$lcssa140 = $166;
       break;
      } else {
       $168 = $166;$x$541 = $169;
      }
     }
     $171 = ($type4$143|0)==(0); //@line 5251 "libraries/lodepng/lodepng.cpp"
     $172 = $$lcssa140 < $smallest3$142; //@line 5251 "libraries/lodepng/lodepng.cpp"
     $or$cond113 = $171 | $172; //@line 5251 "libraries/lodepng/lodepng.cpp"
     $smallest3$2 = $or$cond113 ? $$lcssa140 : $smallest3$142; //@line 5251 "libraries/lodepng/lodepng.cpp"
     $bestType5$2 = $or$cond113 ? $type4$143 : $bestType5$144; //@line 5251 "libraries/lodepng/lodepng.cpp"
     $173 = (($type4$143) + 1)|0; //@line 5238 "libraries/lodepng/lodepng.cpp"
     $174 = ($173|0)==(5); //@line 5238 "libraries/lodepng/lodepng.cpp"
     if ($174) {
      $bestType5$2$lcssa = $bestType5$2;$smallest3$2$lcssa = $smallest3$2;
      break;
     } else {
      $bestType5$144 = $bestType5$2;$smallest3$142 = $smallest3$2;$type4$143 = $173;
     }
    }
    $175 = $bestType5$2$lcssa&255; //@line 5261 "libraries/lodepng/lodepng.cpp"
    $176 = Math_imul($y$251, $117)|0; //@line 5261 "libraries/lodepng/lodepng.cpp"
    $177 = (($out) + ($176)|0); //@line 5261 "libraries/lodepng/lodepng.cpp"
    HEAP8[$177>>0] = $175; //@line 5261 "libraries/lodepng/lodepng.cpp"
    if (!($107)) {
     $178 = (($attempt2) + (($bestType5$2$lcssa*12)|0)|0); //@line 5262 "libraries/lodepng/lodepng.cpp"
     $179 = HEAP32[$178>>2]|0; //@line 5262 "libraries/lodepng/lodepng.cpp"
     $180 = (($176) + 1)|0; //@line 5262 "libraries/lodepng/lodepng.cpp"
     $x$645 = 0;
     while(1) {
      $181 = (($179) + ($x$645)|0); //@line 5262 "libraries/lodepng/lodepng.cpp"
      $182 = HEAP8[$181>>0]|0; //@line 5262 "libraries/lodepng/lodepng.cpp"
      $183 = (($180) + ($x$645))|0; //@line 5262 "libraries/lodepng/lodepng.cpp"
      $184 = (($out) + ($183)|0); //@line 5262 "libraries/lodepng/lodepng.cpp"
      HEAP8[$184>>0] = $182; //@line 5262 "libraries/lodepng/lodepng.cpp"
      $185 = (($x$645) + 1)|0; //@line 5262 "libraries/lodepng/lodepng.cpp"
      $186 = ($185|0)==($3|0); //@line 5262 "libraries/lodepng/lodepng.cpp"
      if ($186) {
       break;
      } else {
       $x$645 = $185;
      }
     }
    }
    $187 = (($y$251) + 1)|0; //@line 5235 "libraries/lodepng/lodepng.cpp"
    $188 = ($187|0)==($h|0); //@line 5235 "libraries/lodepng/lodepng.cpp"
    if ($188) {
     break;
    } else {
     $bestType5$050 = $bestType5$2$lcssa;$prevline$252 = $125;$smallest3$049 = $smallest3$2$lcssa;$y$251 = $187;
    }
   }
  }
  _free($189); //@line 76 "libraries/lodepng/lodepng.cpp"
  _free($190); //@line 76 "libraries/lodepng/lodepng.cpp"
  _free($191); //@line 76 "libraries/lodepng/lodepng.cpp"
  _free($192); //@line 76 "libraries/lodepng/lodepng.cpp"
  _free($193); //@line 76 "libraries/lodepng/lodepng.cpp"
  $$2 = 0;
  STACKTOP = sp;return ($$2|0); //@line 5332 "libraries/lodepng/lodepng.cpp"
  break;
 }
 case 4:  {
  $14 = ($h|0)==(0); //@line 5269 "libraries/lodepng/lodepng.cpp"
  if ($14) {
   $$2 = 0;
   STACKTOP = sp;return ($$2|0); //@line 5332 "libraries/lodepng/lodepng.cpp"
  }
  $15 = (($3) + 1)|0; //@line 5271 "libraries/lodepng/lodepng.cpp"
  $16 = ((($settings)) + 48|0); //@line 5273 "libraries/lodepng/lodepng.cpp"
  $prevline$355 = 0;$y$354 = 0;
  while(1) {
   $194 = Math_imul($y$354, $15)|0; //@line 5271 "libraries/lodepng/lodepng.cpp"
   $195 = Math_imul($y$354, $3)|0; //@line 5272 "libraries/lodepng/lodepng.cpp"
   $196 = HEAP32[$16>>2]|0; //@line 5273 "libraries/lodepng/lodepng.cpp"
   $197 = (($196) + ($y$354)|0); //@line 5273 "libraries/lodepng/lodepng.cpp"
   $198 = HEAP8[$197>>0]|0; //@line 5273 "libraries/lodepng/lodepng.cpp"
   $199 = (($out) + ($194)|0); //@line 5274 "libraries/lodepng/lodepng.cpp"
   HEAP8[$199>>0] = $198; //@line 5274 "libraries/lodepng/lodepng.cpp"
   $200 = (($194) + 1)|0; //@line 5275 "libraries/lodepng/lodepng.cpp"
   $201 = (($out) + ($200)|0); //@line 5275 "libraries/lodepng/lodepng.cpp"
   $202 = (($in) + ($195)|0); //@line 5275 "libraries/lodepng/lodepng.cpp"
   __ZL14filterScanlinePhPKhS1_jjh($201,$202,$prevline$355,$3,$5,$198); //@line 5275 "libraries/lodepng/lodepng.cpp"
   $203 = (($y$354) + 1)|0; //@line 5269 "libraries/lodepng/lodepng.cpp"
   $204 = ($203|0)==($h|0); //@line 5269 "libraries/lodepng/lodepng.cpp"
   if ($204) {
    $$2 = 0;
    break;
   } else {
    $prevline$355 = $202;$y$354 = $203;
   }
  }
  STACKTOP = sp;return ($$2|0); //@line 5332 "libraries/lodepng/lodepng.cpp"
  break;
 }
 case 3:  {
  dest=$zlibsettings; src=$settings; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 5289 "libraries/lodepng/lodepng.cpp"
  HEAP32[$zlibsettings>>2] = 1; //@line 5294 "libraries/lodepng/lodepng.cpp"
  $205 = ((($zlibsettings)) + 24|0); //@line 5297 "libraries/lodepng/lodepng.cpp"
  HEAP32[$205>>2] = 0; //@line 5297 "libraries/lodepng/lodepng.cpp"
  $206 = ((($zlibsettings)) + 28|0); //@line 5298 "libraries/lodepng/lodepng.cpp"
  HEAP32[$206>>2] = 0; //@line 5298 "libraries/lodepng/lodepng.cpp"
  $207 = ($3|0)==(0); //@line 225 "libraries/lodepng/lodepng.cpp"
  $208 = ($3|0)!=(0); //@line 227 "libraries/lodepng/lodepng.cpp"
  $209 = ($3*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
  $210 = $209 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
  $211 = $208 ? $3 : $210; //@line 227 "libraries/lodepng/lodepng.cpp"
  if ($207) {
   dest=$attempt10; stop=dest+60|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 259 "libraries/lodepng/lodepng.cpp"
   $317 = 0;
  } else {
   HEAP32[$attempt10>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
   $212 = ((($attempt10)) + 8|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   HEAP32[$212>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
   $213 = ((($attempt10)) + 4|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   HEAP32[$213>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
   $214 = (_realloc(0,$211)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
   $215 = ($214|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
   if ($215) {
    $316 = 0;
   } else {
    HEAP32[$212>>2] = $211; //@line 231 "libraries/lodepng/lodepng.cpp"
    HEAP32[$attempt10>>2] = $214; //@line 232 "libraries/lodepng/lodepng.cpp"
    HEAP32[$213>>2] = $3; //@line 243 "libraries/lodepng/lodepng.cpp"
    $316 = $214;
   }
   $218 = ((($attempt10)) + 12|0); //@line 258 "libraries/lodepng/lodepng.cpp"
   HEAP32[$218>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
   $219 = ((($attempt10)) + 20|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   HEAP32[$219>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
   $220 = ((($attempt10)) + 16|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   HEAP32[$220>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
   $221 = (_realloc(0,$211)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
   $222 = ($221|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
   if (!($222)) {
    HEAP32[$219>>2] = $211; //@line 231 "libraries/lodepng/lodepng.cpp"
    HEAP32[$218>>2] = $221; //@line 232 "libraries/lodepng/lodepng.cpp"
    HEAP32[$220>>2] = $3; //@line 243 "libraries/lodepng/lodepng.cpp"
   }
   $301 = ((($attempt10)) + 24|0); //@line 258 "libraries/lodepng/lodepng.cpp"
   HEAP32[$301>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
   $302 = ((($attempt10)) + 32|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   HEAP32[$302>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
   $303 = ((($attempt10)) + 28|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   HEAP32[$303>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
   $304 = (_realloc(0,$211)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
   $305 = ($304|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
   if (!($305)) {
    HEAP32[$302>>2] = $211; //@line 231 "libraries/lodepng/lodepng.cpp"
    HEAP32[$301>>2] = $304; //@line 232 "libraries/lodepng/lodepng.cpp"
    HEAP32[$303>>2] = $3; //@line 243 "libraries/lodepng/lodepng.cpp"
   }
   $306 = ((($attempt10)) + 36|0); //@line 258 "libraries/lodepng/lodepng.cpp"
   HEAP32[$306>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
   $307 = ((($attempt10)) + 44|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   HEAP32[$307>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
   $308 = ((($attempt10)) + 40|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   HEAP32[$308>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
   $309 = (_realloc(0,$211)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
   $310 = ($309|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
   if (!($310)) {
    HEAP32[$307>>2] = $211; //@line 231 "libraries/lodepng/lodepng.cpp"
    HEAP32[$306>>2] = $309; //@line 232 "libraries/lodepng/lodepng.cpp"
    HEAP32[$308>>2] = $3; //@line 243 "libraries/lodepng/lodepng.cpp"
   }
   $311 = ((($attempt10)) + 48|0); //@line 258 "libraries/lodepng/lodepng.cpp"
   HEAP32[$311>>2] = 0; //@line 258 "libraries/lodepng/lodepng.cpp"
   $312 = ((($attempt10)) + 56|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   HEAP32[$312>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
   $313 = ((($attempt10)) + 52|0); //@line 259 "libraries/lodepng/lodepng.cpp"
   HEAP32[$313>>2] = 0; //@line 259 "libraries/lodepng/lodepng.cpp"
   $314 = (_realloc(0,$211)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
   $315 = ($314|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
   if ($315) {
    $317 = $316;
   } else {
    HEAP32[$312>>2] = $211; //@line 231 "libraries/lodepng/lodepng.cpp"
    HEAP32[$311>>2] = $314; //@line 232 "libraries/lodepng/lodepng.cpp"
    HEAP32[$313>>2] = $3; //@line 243 "libraries/lodepng/lodepng.cpp"
    $317 = $316;
   }
  }
  $216 = ($h|0)==(0); //@line 5304 "libraries/lodepng/lodepng.cpp"
  if ($216) {
   $$pre$phiZ2D = $attempt10;$254 = $317;
  } else {
   $217 = (($3) + 1)|0; //@line 5324 "libraries/lodepng/lodepng.cpp"
   $bestType13$068 = 0;$prevline$466 = 0;$smallest11$067 = 0;$y$465 = 0;
   while(1) {
    $223 = Math_imul($y$465, $3)|0; //@line 5311 "libraries/lodepng/lodepng.cpp"
    $224 = (($in) + ($223)|0); //@line 5311 "libraries/lodepng/lodepng.cpp"
    $bestType13$160 = $bestType13$068;$smallest11$158 = $smallest11$067;$type12$159 = 0;
    while(1) {
     $225 = (((($attempt10) + (($type12$159*12)|0)|0)) + 4|0); //@line 5308 "libraries/lodepng/lodepng.cpp"
     $226 = HEAP32[$225>>2]|0; //@line 5308 "libraries/lodepng/lodepng.cpp"
     $227 = (($attempt10) + (($type12$159*12)|0)|0); //@line 5311 "libraries/lodepng/lodepng.cpp"
     $228 = HEAP32[$227>>2]|0; //@line 5311 "libraries/lodepng/lodepng.cpp"
     $229 = $type12$159&255; //@line 5311 "libraries/lodepng/lodepng.cpp"
     __ZL14filterScanlinePhPKhS1_jjh($228,$224,$prevline$466,$3,$5,$229); //@line 5311 "libraries/lodepng/lodepng.cpp"
     $230 = (($size) + ($type12$159<<2)|0); //@line 5312 "libraries/lodepng/lodepng.cpp"
     HEAP32[$230>>2] = 0; //@line 5312 "libraries/lodepng/lodepng.cpp"
     HEAP32[$dummy>>2] = 0; //@line 5313 "libraries/lodepng/lodepng.cpp"
     $231 = HEAP32[$205>>2]|0; //@line 2209 "libraries/lodepng/lodepng.cpp"
     $232 = ($231|0)==(0|0); //@line 2209 "libraries/lodepng/lodepng.cpp"
     if ($232) {
      (__Z21lodepng_zlib_compressPPhPjPKhjPK23LodePNGCompressSettings($dummy,$230,$228,$226,$zlibsettings)|0); //@line 2215 "libraries/lodepng/lodepng.cpp"
     } else {
      (FUNCTION_TABLE_iiiiii[$231 & 8191]($dummy,$230,$228,$226,$zlibsettings)|0); //@line 2211 "libraries/lodepng/lodepng.cpp"
     }
     $233 = HEAP32[$dummy>>2]|0; //@line 5315 "libraries/lodepng/lodepng.cpp"
     _free($233); //@line 76 "libraries/lodepng/lodepng.cpp"
     $234 = ($type12$159|0)==(0); //@line 5317 "libraries/lodepng/lodepng.cpp"
     $$pre = HEAP32[$230>>2]|0; //@line 5320 "libraries/lodepng/lodepng.cpp"
     $235 = ($$pre>>>0)<($smallest11$158>>>0); //@line 5317 "libraries/lodepng/lodepng.cpp"
     $or$cond114 = $234 | $235; //@line 5317 "libraries/lodepng/lodepng.cpp"
     $smallest11$2 = $or$cond114 ? $$pre : $smallest11$158; //@line 5317 "libraries/lodepng/lodepng.cpp"
     $bestType13$2 = $or$cond114 ? $type12$159 : $bestType13$160; //@line 5317 "libraries/lodepng/lodepng.cpp"
     $236 = (($type12$159) + 1)|0; //@line 5306 "libraries/lodepng/lodepng.cpp"
     $237 = ($236|0)==(5); //@line 5306 "libraries/lodepng/lodepng.cpp"
     if ($237) {
      $bestType13$2$lcssa = $bestType13$2;$smallest11$2$lcssa = $smallest11$2;
      break;
     } else {
      $bestType13$160 = $bestType13$2;$smallest11$158 = $smallest11$2;$type12$159 = $236;
     }
    }
    $238 = $bestType13$2$lcssa&255; //@line 5324 "libraries/lodepng/lodepng.cpp"
    $239 = Math_imul($y$465, $217)|0; //@line 5324 "libraries/lodepng/lodepng.cpp"
    $240 = (($out) + ($239)|0); //@line 5324 "libraries/lodepng/lodepng.cpp"
    HEAP8[$240>>0] = $238; //@line 5324 "libraries/lodepng/lodepng.cpp"
    if (!($207)) {
     $241 = (($attempt10) + (($bestType13$2$lcssa*12)|0)|0); //@line 5325 "libraries/lodepng/lodepng.cpp"
     $242 = HEAP32[$241>>2]|0; //@line 5325 "libraries/lodepng/lodepng.cpp"
     $243 = (($239) + 1)|0; //@line 5325 "libraries/lodepng/lodepng.cpp"
     $x$761 = 0;
     while(1) {
      $244 = (($242) + ($x$761)|0); //@line 5325 "libraries/lodepng/lodepng.cpp"
      $245 = HEAP8[$244>>0]|0; //@line 5325 "libraries/lodepng/lodepng.cpp"
      $246 = (($243) + ($x$761))|0; //@line 5325 "libraries/lodepng/lodepng.cpp"
      $247 = (($out) + ($246)|0); //@line 5325 "libraries/lodepng/lodepng.cpp"
      HEAP8[$247>>0] = $245; //@line 5325 "libraries/lodepng/lodepng.cpp"
      $248 = (($x$761) + 1)|0; //@line 5325 "libraries/lodepng/lodepng.cpp"
      $249 = ($248|0)==($3|0); //@line 5325 "libraries/lodepng/lodepng.cpp"
      if ($249) {
       break;
      } else {
       $x$761 = $248;
      }
     }
    }
    $250 = (($y$465) + 1)|0; //@line 5304 "libraries/lodepng/lodepng.cpp"
    $251 = ($250|0)==($h|0); //@line 5304 "libraries/lodepng/lodepng.cpp"
    if ($251) {
     break;
    } else {
     $bestType13$068 = $bestType13$2$lcssa;$prevline$466 = $224;$smallest11$067 = $smallest11$2$lcssa;$y$465 = $250;
    }
   }
   $$pre99 = HEAP32[$attempt10>>2]|0; //@line 252 "libraries/lodepng/lodepng.cpp"
   $$pre$phiZ2D = $attempt10;$254 = $$pre99;
  }
  $252 = ((($attempt10)) + 8|0); //@line 251 "libraries/lodepng/lodepng.cpp"
  HEAP32[$252>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
  $253 = ((($attempt10)) + 4|0); //@line 251 "libraries/lodepng/lodepng.cpp"
  HEAP32[$253>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
  _free($254); //@line 76 "libraries/lodepng/lodepng.cpp"
  HEAP32[$$pre$phiZ2D>>2] = 0; //@line 253 "libraries/lodepng/lodepng.cpp"
  $255 = ((($attempt10)) + 20|0); //@line 251 "libraries/lodepng/lodepng.cpp"
  HEAP32[$255>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
  $256 = ((($attempt10)) + 16|0); //@line 251 "libraries/lodepng/lodepng.cpp"
  HEAP32[$256>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
  $257 = ((($attempt10)) + 12|0); //@line 252 "libraries/lodepng/lodepng.cpp"
  $258 = HEAP32[$257>>2]|0; //@line 252 "libraries/lodepng/lodepng.cpp"
  _free($258); //@line 76 "libraries/lodepng/lodepng.cpp"
  HEAP32[$257>>2] = 0; //@line 253 "libraries/lodepng/lodepng.cpp"
  $259 = ((($attempt10)) + 32|0); //@line 251 "libraries/lodepng/lodepng.cpp"
  HEAP32[$259>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
  $260 = ((($attempt10)) + 28|0); //@line 251 "libraries/lodepng/lodepng.cpp"
  HEAP32[$260>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
  $261 = ((($attempt10)) + 24|0); //@line 252 "libraries/lodepng/lodepng.cpp"
  $262 = HEAP32[$261>>2]|0; //@line 252 "libraries/lodepng/lodepng.cpp"
  _free($262); //@line 76 "libraries/lodepng/lodepng.cpp"
  HEAP32[$261>>2] = 0; //@line 253 "libraries/lodepng/lodepng.cpp"
  $263 = ((($attempt10)) + 44|0); //@line 251 "libraries/lodepng/lodepng.cpp"
  HEAP32[$263>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
  $264 = ((($attempt10)) + 40|0); //@line 251 "libraries/lodepng/lodepng.cpp"
  HEAP32[$264>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
  $265 = ((($attempt10)) + 36|0); //@line 252 "libraries/lodepng/lodepng.cpp"
  $266 = HEAP32[$265>>2]|0; //@line 252 "libraries/lodepng/lodepng.cpp"
  _free($266); //@line 76 "libraries/lodepng/lodepng.cpp"
  HEAP32[$265>>2] = 0; //@line 253 "libraries/lodepng/lodepng.cpp"
  $267 = ((($attempt10)) + 56|0); //@line 251 "libraries/lodepng/lodepng.cpp"
  HEAP32[$267>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
  $268 = ((($attempt10)) + 52|0); //@line 251 "libraries/lodepng/lodepng.cpp"
  HEAP32[$268>>2] = 0; //@line 251 "libraries/lodepng/lodepng.cpp"
  $269 = ((($attempt10)) + 48|0); //@line 252 "libraries/lodepng/lodepng.cpp"
  $270 = HEAP32[$269>>2]|0; //@line 252 "libraries/lodepng/lodepng.cpp"
  _free($270); //@line 76 "libraries/lodepng/lodepng.cpp"
  $$2 = 0;
  STACKTOP = sp;return ($$2|0); //@line 5332 "libraries/lodepng/lodepng.cpp"
  break;
 }
 default: {
  $$2 = 88;
  STACKTOP = sp;return ($$2|0); //@line 5332 "libraries/lodepng/lodepng.cpp"
 }
 }
 return (0)|0;
}
function __ZL14filterScanlinePhPKhS1_jjh($out,$scanline,$prevline,$length,$bytewidth,$filterType) {
 $out = $out|0;
 $scanline = $scanline|0;
 $prevline = $prevline|0;
 $length = $length|0;
 $bytewidth = $bytewidth|0;
 $filterType = $filterType|0;
 var $$ = 0, $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $div = 0, $div1 = 0, $exitcond = 0, $exitcond52 = 0, $exitcond53 = 0, $exitcond54 = 0, $exitcond55 = 0, $factor$i = 0;
 var $i$021 = 0, $i$1042 = 0, $i$1140 = 0, $i$1238 = 0, $i$124 = 0, $i$222 = 0, $i$328 = 0, $i$426 = 0, $i$536 = 0, $i$634 = 0, $i$732 = 0, $i$830 = 0, $i$943 = 0, $ispos$i = 0, $ispos1$i = 0, $ispos3$i = 0, $neg$i = 0, $neg2$i = 0, $neg4$i = 0, $or$cond$i = 0;
 var $sext$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $filterType&255; //@line 5053 "libraries/lodepng/lodepng.cpp"
 switch ($0|0) {
 case 0:  {
  $2 = ($length|0)==(0); //@line 5056 "libraries/lodepng/lodepng.cpp"
  if ($2) {
   return; //@line 5103 "libraries/lodepng/lodepng.cpp"
  } else {
   $i$021 = 0;
  }
  while(1) {
   $3 = (($scanline) + ($i$021)|0); //@line 5056 "libraries/lodepng/lodepng.cpp"
   $4 = HEAP8[$3>>0]|0; //@line 5056 "libraries/lodepng/lodepng.cpp"
   $5 = (($out) + ($i$021)|0); //@line 5056 "libraries/lodepng/lodepng.cpp"
   HEAP8[$5>>0] = $4; //@line 5056 "libraries/lodepng/lodepng.cpp"
   $6 = (($i$021) + 1)|0; //@line 5056 "libraries/lodepng/lodepng.cpp"
   $7 = ($6|0)==($length|0); //@line 5056 "libraries/lodepng/lodepng.cpp"
   if ($7) {
    break;
   } else {
    $i$021 = $6;
   }
  }
  return; //@line 5103 "libraries/lodepng/lodepng.cpp"
  break;
 }
 case 1:  {
  $1 = ($bytewidth|0)==(0); //@line 5059 "libraries/lodepng/lodepng.cpp"
  if (!($1)) {
   $i$124 = 0;
   while(1) {
    $9 = (($scanline) + ($i$124)|0); //@line 5059 "libraries/lodepng/lodepng.cpp"
    $10 = HEAP8[$9>>0]|0; //@line 5059 "libraries/lodepng/lodepng.cpp"
    $11 = (($out) + ($i$124)|0); //@line 5059 "libraries/lodepng/lodepng.cpp"
    HEAP8[$11>>0] = $10; //@line 5059 "libraries/lodepng/lodepng.cpp"
    $12 = (($i$124) + 1)|0; //@line 5059 "libraries/lodepng/lodepng.cpp"
    $13 = ($12|0)==($bytewidth|0); //@line 5059 "libraries/lodepng/lodepng.cpp"
    if ($13) {
     break;
    } else {
     $i$124 = $12;
    }
   }
  }
  $8 = ($bytewidth>>>0)<($length>>>0); //@line 5060 "libraries/lodepng/lodepng.cpp"
  if ($8) {
   $i$222 = $bytewidth;
  } else {
   return; //@line 5103 "libraries/lodepng/lodepng.cpp"
  }
  while(1) {
   $14 = (($scanline) + ($i$222)|0); //@line 5060 "libraries/lodepng/lodepng.cpp"
   $15 = HEAP8[$14>>0]|0; //@line 5060 "libraries/lodepng/lodepng.cpp"
   $16 = $15&255; //@line 5060 "libraries/lodepng/lodepng.cpp"
   $17 = (($i$222) - ($bytewidth))|0; //@line 5060 "libraries/lodepng/lodepng.cpp"
   $18 = (($scanline) + ($17)|0); //@line 5060 "libraries/lodepng/lodepng.cpp"
   $19 = HEAP8[$18>>0]|0; //@line 5060 "libraries/lodepng/lodepng.cpp"
   $20 = $19&255; //@line 5060 "libraries/lodepng/lodepng.cpp"
   $21 = (($16) - ($20))|0; //@line 5060 "libraries/lodepng/lodepng.cpp"
   $22 = $21&255; //@line 5060 "libraries/lodepng/lodepng.cpp"
   $23 = (($out) + ($i$222)|0); //@line 5060 "libraries/lodepng/lodepng.cpp"
   HEAP8[$23>>0] = $22; //@line 5060 "libraries/lodepng/lodepng.cpp"
   $24 = (($i$222) + 1)|0; //@line 5060 "libraries/lodepng/lodepng.cpp"
   $exitcond = ($24|0)==($length|0); //@line 5060 "libraries/lodepng/lodepng.cpp"
   if ($exitcond) {
    break;
   } else {
    $i$222 = $24;
   }
  }
  return; //@line 5103 "libraries/lodepng/lodepng.cpp"
  break;
 }
 case 2:  {
  $25 = ($prevline|0)==(0|0); //@line 5063 "libraries/lodepng/lodepng.cpp"
  $26 = ($length|0)==(0); //@line 5069 "libraries/lodepng/lodepng.cpp"
  if ($25) {
   if ($26) {
    return; //@line 5103 "libraries/lodepng/lodepng.cpp"
   } else {
    $i$426 = 0;
   }
   while(1) {
    $38 = (($scanline) + ($i$426)|0); //@line 5069 "libraries/lodepng/lodepng.cpp"
    $39 = HEAP8[$38>>0]|0; //@line 5069 "libraries/lodepng/lodepng.cpp"
    $40 = (($out) + ($i$426)|0); //@line 5069 "libraries/lodepng/lodepng.cpp"
    HEAP8[$40>>0] = $39; //@line 5069 "libraries/lodepng/lodepng.cpp"
    $41 = (($i$426) + 1)|0; //@line 5069 "libraries/lodepng/lodepng.cpp"
    $42 = ($41|0)==($length|0); //@line 5069 "libraries/lodepng/lodepng.cpp"
    if ($42) {
     break;
    } else {
     $i$426 = $41;
    }
   }
   return; //@line 5103 "libraries/lodepng/lodepng.cpp"
  } else {
   if ($26) {
    return; //@line 5103 "libraries/lodepng/lodepng.cpp"
   } else {
    $i$328 = 0;
   }
   while(1) {
    $27 = (($scanline) + ($i$328)|0); //@line 5065 "libraries/lodepng/lodepng.cpp"
    $28 = HEAP8[$27>>0]|0; //@line 5065 "libraries/lodepng/lodepng.cpp"
    $29 = $28&255; //@line 5065 "libraries/lodepng/lodepng.cpp"
    $30 = (($prevline) + ($i$328)|0); //@line 5065 "libraries/lodepng/lodepng.cpp"
    $31 = HEAP8[$30>>0]|0; //@line 5065 "libraries/lodepng/lodepng.cpp"
    $32 = $31&255; //@line 5065 "libraries/lodepng/lodepng.cpp"
    $33 = (($29) - ($32))|0; //@line 5065 "libraries/lodepng/lodepng.cpp"
    $34 = $33&255; //@line 5065 "libraries/lodepng/lodepng.cpp"
    $35 = (($out) + ($i$328)|0); //@line 5065 "libraries/lodepng/lodepng.cpp"
    HEAP8[$35>>0] = $34; //@line 5065 "libraries/lodepng/lodepng.cpp"
    $36 = (($i$328) + 1)|0; //@line 5065 "libraries/lodepng/lodepng.cpp"
    $37 = ($36|0)==($length|0); //@line 5065 "libraries/lodepng/lodepng.cpp"
    if ($37) {
     break;
    } else {
     $i$328 = $36;
    }
   }
   return; //@line 5103 "libraries/lodepng/lodepng.cpp"
  }
  break;
 }
 case 3:  {
  $43 = ($prevline|0)==(0|0); //@line 5073 "libraries/lodepng/lodepng.cpp"
  $44 = ($bytewidth|0)==(0); //@line 5080 "libraries/lodepng/lodepng.cpp"
  if ($43) {
   if (!($44)) {
    $i$732 = 0;
    while(1) {
     $74 = (($scanline) + ($i$732)|0); //@line 5080 "libraries/lodepng/lodepng.cpp"
     $75 = HEAP8[$74>>0]|0; //@line 5080 "libraries/lodepng/lodepng.cpp"
     $76 = (($out) + ($i$732)|0); //@line 5080 "libraries/lodepng/lodepng.cpp"
     HEAP8[$76>>0] = $75; //@line 5080 "libraries/lodepng/lodepng.cpp"
     $77 = (($i$732) + 1)|0; //@line 5080 "libraries/lodepng/lodepng.cpp"
     $78 = ($77|0)==($bytewidth|0); //@line 5080 "libraries/lodepng/lodepng.cpp"
     if ($78) {
      break;
     } else {
      $i$732 = $77;
     }
    }
   }
   $73 = ($bytewidth>>>0)<($length>>>0); //@line 5081 "libraries/lodepng/lodepng.cpp"
   if ($73) {
    $i$830 = $bytewidth;
   } else {
    return; //@line 5103 "libraries/lodepng/lodepng.cpp"
   }
   while(1) {
    $79 = (($scanline) + ($i$830)|0); //@line 5081 "libraries/lodepng/lodepng.cpp"
    $80 = HEAP8[$79>>0]|0; //@line 5081 "libraries/lodepng/lodepng.cpp"
    $81 = $80&255; //@line 5081 "libraries/lodepng/lodepng.cpp"
    $82 = (($i$830) - ($bytewidth))|0; //@line 5081 "libraries/lodepng/lodepng.cpp"
    $83 = (($scanline) + ($82)|0); //@line 5081 "libraries/lodepng/lodepng.cpp"
    $84 = HEAP8[$83>>0]|0; //@line 5081 "libraries/lodepng/lodepng.cpp"
    $div = ($84&255) >>> 1; //@line 5081 "libraries/lodepng/lodepng.cpp"
    $85 = $div&255; //@line 5081 "libraries/lodepng/lodepng.cpp"
    $86 = (($81) - ($85))|0; //@line 5081 "libraries/lodepng/lodepng.cpp"
    $87 = $86&255; //@line 5081 "libraries/lodepng/lodepng.cpp"
    $88 = (($out) + ($i$830)|0); //@line 5081 "libraries/lodepng/lodepng.cpp"
    HEAP8[$88>>0] = $87; //@line 5081 "libraries/lodepng/lodepng.cpp"
    $89 = (($i$830) + 1)|0; //@line 5081 "libraries/lodepng/lodepng.cpp"
    $exitcond52 = ($89|0)==($length|0); //@line 5081 "libraries/lodepng/lodepng.cpp"
    if ($exitcond52) {
     break;
    } else {
     $i$830 = $89;
    }
   }
   return; //@line 5103 "libraries/lodepng/lodepng.cpp"
  } else {
   if (!($44)) {
    $i$536 = 0;
    while(1) {
     $46 = (($scanline) + ($i$536)|0); //@line 5075 "libraries/lodepng/lodepng.cpp"
     $47 = HEAP8[$46>>0]|0; //@line 5075 "libraries/lodepng/lodepng.cpp"
     $48 = $47&255; //@line 5075 "libraries/lodepng/lodepng.cpp"
     $49 = (($prevline) + ($i$536)|0); //@line 5075 "libraries/lodepng/lodepng.cpp"
     $50 = HEAP8[$49>>0]|0; //@line 5075 "libraries/lodepng/lodepng.cpp"
     $div1 = ($50&255) >>> 1; //@line 5075 "libraries/lodepng/lodepng.cpp"
     $51 = $div1&255; //@line 5075 "libraries/lodepng/lodepng.cpp"
     $52 = (($48) - ($51))|0; //@line 5075 "libraries/lodepng/lodepng.cpp"
     $53 = $52&255; //@line 5075 "libraries/lodepng/lodepng.cpp"
     $54 = (($out) + ($i$536)|0); //@line 5075 "libraries/lodepng/lodepng.cpp"
     HEAP8[$54>>0] = $53; //@line 5075 "libraries/lodepng/lodepng.cpp"
     $55 = (($i$536) + 1)|0; //@line 5075 "libraries/lodepng/lodepng.cpp"
     $56 = ($55|0)==($bytewidth|0); //@line 5075 "libraries/lodepng/lodepng.cpp"
     if ($56) {
      break;
     } else {
      $i$536 = $55;
     }
    }
   }
   $45 = ($bytewidth>>>0)<($length>>>0); //@line 5076 "libraries/lodepng/lodepng.cpp"
   if ($45) {
    $i$634 = $bytewidth;
   } else {
    return; //@line 5103 "libraries/lodepng/lodepng.cpp"
   }
   while(1) {
    $57 = (($scanline) + ($i$634)|0); //@line 5076 "libraries/lodepng/lodepng.cpp"
    $58 = HEAP8[$57>>0]|0; //@line 5076 "libraries/lodepng/lodepng.cpp"
    $59 = $58&255; //@line 5076 "libraries/lodepng/lodepng.cpp"
    $60 = (($i$634) - ($bytewidth))|0; //@line 5076 "libraries/lodepng/lodepng.cpp"
    $61 = (($scanline) + ($60)|0); //@line 5076 "libraries/lodepng/lodepng.cpp"
    $62 = HEAP8[$61>>0]|0; //@line 5076 "libraries/lodepng/lodepng.cpp"
    $63 = $62&255; //@line 5076 "libraries/lodepng/lodepng.cpp"
    $64 = (($prevline) + ($i$634)|0); //@line 5076 "libraries/lodepng/lodepng.cpp"
    $65 = HEAP8[$64>>0]|0; //@line 5076 "libraries/lodepng/lodepng.cpp"
    $66 = $65&255; //@line 5076 "libraries/lodepng/lodepng.cpp"
    $67 = (($66) + ($63))|0; //@line 5076 "libraries/lodepng/lodepng.cpp"
    $68 = $67 >>> 1; //@line 5076 "libraries/lodepng/lodepng.cpp"
    $69 = (($59) - ($68))|0; //@line 5076 "libraries/lodepng/lodepng.cpp"
    $70 = $69&255; //@line 5076 "libraries/lodepng/lodepng.cpp"
    $71 = (($out) + ($i$634)|0); //@line 5076 "libraries/lodepng/lodepng.cpp"
    HEAP8[$71>>0] = $70; //@line 5076 "libraries/lodepng/lodepng.cpp"
    $72 = (($i$634) + 1)|0; //@line 5076 "libraries/lodepng/lodepng.cpp"
    $exitcond53 = ($72|0)==($length|0); //@line 5076 "libraries/lodepng/lodepng.cpp"
    if ($exitcond53) {
     break;
    } else {
     $i$634 = $72;
    }
   }
   return; //@line 5103 "libraries/lodepng/lodepng.cpp"
  }
  break;
 }
 case 4:  {
  $90 = ($prevline|0)==(0|0); //@line 5085 "libraries/lodepng/lodepng.cpp"
  $91 = ($bytewidth|0)==(0); //@line 5096 "libraries/lodepng/lodepng.cpp"
  if ($90) {
   if (!($91)) {
    $i$1140 = 0;
    while(1) {
     $134 = (($scanline) + ($i$1140)|0); //@line 5096 "libraries/lodepng/lodepng.cpp"
     $135 = HEAP8[$134>>0]|0; //@line 5096 "libraries/lodepng/lodepng.cpp"
     $136 = (($out) + ($i$1140)|0); //@line 5096 "libraries/lodepng/lodepng.cpp"
     HEAP8[$136>>0] = $135; //@line 5096 "libraries/lodepng/lodepng.cpp"
     $137 = (($i$1140) + 1)|0; //@line 5096 "libraries/lodepng/lodepng.cpp"
     $138 = ($137|0)==($bytewidth|0); //@line 5096 "libraries/lodepng/lodepng.cpp"
     if ($138) {
      break;
     } else {
      $i$1140 = $137;
     }
    }
   }
   $133 = ($bytewidth>>>0)<($length>>>0); //@line 5098 "libraries/lodepng/lodepng.cpp"
   if ($133) {
    $i$1238 = $bytewidth;
   } else {
    return; //@line 5103 "libraries/lodepng/lodepng.cpp"
   }
   while(1) {
    $139 = (($scanline) + ($i$1238)|0); //@line 5098 "libraries/lodepng/lodepng.cpp"
    $140 = HEAP8[$139>>0]|0; //@line 5098 "libraries/lodepng/lodepng.cpp"
    $141 = $140&255; //@line 5098 "libraries/lodepng/lodepng.cpp"
    $142 = (($i$1238) - ($bytewidth))|0; //@line 5098 "libraries/lodepng/lodepng.cpp"
    $143 = (($scanline) + ($142)|0); //@line 5098 "libraries/lodepng/lodepng.cpp"
    $144 = HEAP8[$143>>0]|0; //@line 5098 "libraries/lodepng/lodepng.cpp"
    $145 = $144&255; //@line 5098 "libraries/lodepng/lodepng.cpp"
    $146 = (($141) - ($145))|0; //@line 5098 "libraries/lodepng/lodepng.cpp"
    $147 = $146&255; //@line 5098 "libraries/lodepng/lodepng.cpp"
    $148 = (($out) + ($i$1238)|0); //@line 5098 "libraries/lodepng/lodepng.cpp"
    HEAP8[$148>>0] = $147; //@line 5098 "libraries/lodepng/lodepng.cpp"
    $149 = (($i$1238) + 1)|0; //@line 5098 "libraries/lodepng/lodepng.cpp"
    $exitcond54 = ($149|0)==($length|0); //@line 5098 "libraries/lodepng/lodepng.cpp"
    if ($exitcond54) {
     break;
    } else {
     $i$1238 = $149;
    }
   }
   return; //@line 5103 "libraries/lodepng/lodepng.cpp"
  } else {
   if (!($91)) {
    $i$943 = 0;
    while(1) {
     $93 = (($scanline) + ($i$943)|0); //@line 5088 "libraries/lodepng/lodepng.cpp"
     $94 = HEAP8[$93>>0]|0; //@line 5088 "libraries/lodepng/lodepng.cpp"
     $95 = $94&255; //@line 5088 "libraries/lodepng/lodepng.cpp"
     $96 = (($prevline) + ($i$943)|0); //@line 5088 "libraries/lodepng/lodepng.cpp"
     $97 = HEAP8[$96>>0]|0; //@line 5088 "libraries/lodepng/lodepng.cpp"
     $98 = $97&255; //@line 5088 "libraries/lodepng/lodepng.cpp"
     $99 = (($95) - ($98))|0; //@line 5088 "libraries/lodepng/lodepng.cpp"
     $100 = $99&255; //@line 5088 "libraries/lodepng/lodepng.cpp"
     $101 = (($out) + ($i$943)|0); //@line 5088 "libraries/lodepng/lodepng.cpp"
     HEAP8[$101>>0] = $100; //@line 5088 "libraries/lodepng/lodepng.cpp"
     $102 = (($i$943) + 1)|0; //@line 5088 "libraries/lodepng/lodepng.cpp"
     $103 = ($102|0)==($bytewidth|0); //@line 5088 "libraries/lodepng/lodepng.cpp"
     if ($103) {
      break;
     } else {
      $i$943 = $102;
     }
    }
   }
   $92 = ($bytewidth>>>0)<($length>>>0); //@line 5089 "libraries/lodepng/lodepng.cpp"
   if ($92) {
    $i$1042 = $bytewidth;
   } else {
    return; //@line 5103 "libraries/lodepng/lodepng.cpp"
   }
   while(1) {
    $104 = (($scanline) + ($i$1042)|0); //@line 5091 "libraries/lodepng/lodepng.cpp"
    $105 = HEAP8[$104>>0]|0; //@line 5091 "libraries/lodepng/lodepng.cpp"
    $106 = $105&255; //@line 5091 "libraries/lodepng/lodepng.cpp"
    $107 = (($i$1042) - ($bytewidth))|0; //@line 5091 "libraries/lodepng/lodepng.cpp"
    $108 = (($scanline) + ($107)|0); //@line 5091 "libraries/lodepng/lodepng.cpp"
    $109 = HEAP8[$108>>0]|0; //@line 5091 "libraries/lodepng/lodepng.cpp"
    $110 = (($prevline) + ($i$1042)|0); //@line 5091 "libraries/lodepng/lodepng.cpp"
    $111 = HEAP8[$110>>0]|0; //@line 5091 "libraries/lodepng/lodepng.cpp"
    $112 = (($prevline) + ($107)|0); //@line 5091 "libraries/lodepng/lodepng.cpp"
    $113 = HEAP8[$112>>0]|0; //@line 5091 "libraries/lodepng/lodepng.cpp"
    $114 = $111&255; //@line 3749 "libraries/lodepng/lodepng.cpp"
    $115 = $113&255; //@line 3749 "libraries/lodepng/lodepng.cpp"
    $116 = (($114) - ($115))|0; //@line 3749 "libraries/lodepng/lodepng.cpp"
    $ispos$i = ($116|0)>(-1); //@line 3749 "libraries/lodepng/lodepng.cpp"
    $neg$i = (0 - ($116))|0; //@line 3749 "libraries/lodepng/lodepng.cpp"
    $117 = $ispos$i ? $116 : $neg$i; //@line 3749 "libraries/lodepng/lodepng.cpp"
    $118 = $109&255; //@line 3750 "libraries/lodepng/lodepng.cpp"
    $119 = (($118) - ($115))|0; //@line 3750 "libraries/lodepng/lodepng.cpp"
    $ispos1$i = ($119|0)>(-1); //@line 3750 "libraries/lodepng/lodepng.cpp"
    $neg2$i = (0 - ($119))|0; //@line 3750 "libraries/lodepng/lodepng.cpp"
    $120 = $ispos1$i ? $119 : $neg2$i; //@line 3750 "libraries/lodepng/lodepng.cpp"
    $factor$i = Math_imul($115, -2)|0; //@line 5091 "libraries/lodepng/lodepng.cpp"
    $121 = (($114) + ($118))|0; //@line 3751 "libraries/lodepng/lodepng.cpp"
    $122 = (($121) + ($factor$i))|0; //@line 3751 "libraries/lodepng/lodepng.cpp"
    $ispos3$i = ($122|0)>(-1); //@line 3751 "libraries/lodepng/lodepng.cpp"
    $neg4$i = (0 - ($122))|0; //@line 3751 "libraries/lodepng/lodepng.cpp"
    $123 = $ispos3$i ? $122 : $neg4$i; //@line 3751 "libraries/lodepng/lodepng.cpp"
    $sext$i = $123 << 16; //@line 3753 "libraries/lodepng/lodepng.cpp"
    $124 = $sext$i >> 16; //@line 3753 "libraries/lodepng/lodepng.cpp"
    $125 = ($124|0)<($117|0); //@line 3753 "libraries/lodepng/lodepng.cpp"
    $126 = ($124|0)<($120|0); //@line 3753 "libraries/lodepng/lodepng.cpp"
    $or$cond$i = $125 & $126; //@line 3753 "libraries/lodepng/lodepng.cpp"
    $127 = ($120|0)<($117|0); //@line 3754 "libraries/lodepng/lodepng.cpp"
    $$ = $127 ? $111 : $109; //@line 3754 "libraries/lodepng/lodepng.cpp"
    $$0$i = $or$cond$i ? $113 : $$; //@line 3753 "libraries/lodepng/lodepng.cpp"
    $128 = $$0$i&255; //@line 5091 "libraries/lodepng/lodepng.cpp"
    $129 = (($106) - ($128))|0; //@line 5091 "libraries/lodepng/lodepng.cpp"
    $130 = $129&255; //@line 5091 "libraries/lodepng/lodepng.cpp"
    $131 = (($out) + ($i$1042)|0); //@line 5091 "libraries/lodepng/lodepng.cpp"
    HEAP8[$131>>0] = $130; //@line 5091 "libraries/lodepng/lodepng.cpp"
    $132 = (($i$1042) + 1)|0; //@line 5089 "libraries/lodepng/lodepng.cpp"
    $exitcond55 = ($132|0)==($length|0); //@line 5089 "libraries/lodepng/lodepng.cpp"
    if ($exitcond55) {
     break;
    } else {
     $i$1042 = $132;
    }
   }
   return; //@line 5103 "libraries/lodepng/lodepng.cpp"
  }
  break;
 }
 default: {
  return; //@line 5103 "libraries/lodepng/lodepng.cpp"
 }
 }
}
function __ZL13addChunk_tRNSP8ucvectorPK16LodePNGColorMode($out,$info) {
 $out = $out|0;
 $info = $info|0;
 var $$sink = 0, $$sink49 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $amount$056 = 0, $amount$056$lcssa = 0, $i$153 = 0, $tRNS$sroa$0$052 = 0, $tRNS$sroa$0$1 = 0, $tRNS$sroa$0$10 = 0, $tRNS$sroa$0$11 = 0, $tRNS$sroa$0$12 = 0, $tRNS$sroa$0$13 = 0, $tRNS$sroa$0$14 = 0, $tRNS$sroa$0$15 = 0, $tRNS$sroa$0$16 = 0;
 var $tRNS$sroa$0$17 = 0, $tRNS$sroa$0$19 = 0, $tRNS$sroa$0$2 = 0, $tRNS$sroa$0$444 = 0, $tRNS$sroa$0$848 = 0, $tRNS$sroa$33$051 = 0, $tRNS$sroa$33$1 = 0, $tRNS$sroa$33$10 = 0, $tRNS$sroa$33$243 = 0, $tRNS$sroa$33$447 = 0, $tRNS$sroa$33$5 = 0, $tRNS$sroa$33$6 = 0, $tRNS$sroa$33$7 = 0, $tRNS$sroa$33$8 = 0, $tRNS$sroa$54$050 = 0, $tRNS$sroa$54$1 = 0, $tRNS$sroa$54$10 = 0, $tRNS$sroa$54$11 = 0, $tRNS$sroa$54$12 = 0, $tRNS$sroa$54$13 = 0;
 var $tRNS$sroa$54$14 = 0, $tRNS$sroa$54$2 = 0, $tRNS$sroa$54$8 = 0, $tRNS$sroa$54$9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$info>>2]|0; //@line 4837 "libraries/lodepng/lodepng.cpp"
 L1: do {
  switch ($0|0) {
  case 3:  {
   $1 = ((($info)) + 12|0); //@line 4839 "libraries/lodepng/lodepng.cpp"
   $2 = HEAP32[$1>>2]|0; //@line 4839 "libraries/lodepng/lodepng.cpp"
   $3 = ($2|0)==(0); //@line 4841 "libraries/lodepng/lodepng.cpp"
   if ($3) {
    $tRNS$sroa$0$19 = 0;$tRNS$sroa$33$10 = 0;
   } else {
    $4 = ((($info)) + 8|0); //@line 4843 "libraries/lodepng/lodepng.cpp"
    $5 = HEAP32[$4>>2]|0; //@line 4843 "libraries/lodepng/lodepng.cpp"
    $amount$056 = $2;
    while(1) {
     $6 = $amount$056 << 2; //@line 4843 "libraries/lodepng/lodepng.cpp"
     $7 = (($6) + -1)|0; //@line 4843 "libraries/lodepng/lodepng.cpp"
     $8 = (($5) + ($7)|0); //@line 4843 "libraries/lodepng/lodepng.cpp"
     $9 = HEAP8[$8>>0]|0; //@line 4843 "libraries/lodepng/lodepng.cpp"
     $10 = ($9<<24>>24)==(-1); //@line 4843 "libraries/lodepng/lodepng.cpp"
     if (!($10)) {
      $amount$056$lcssa = $amount$056;
      break;
     }
     $12 = (($amount$056) + -1)|0; //@line 4843 "libraries/lodepng/lodepng.cpp"
     $13 = ($12|0)==(0); //@line 4841 "libraries/lodepng/lodepng.cpp"
     if ($13) {
      $tRNS$sroa$0$19 = 0;$tRNS$sroa$33$10 = 0;
      break L1;
     } else {
      $amount$056 = $12;
     }
    }
    $11 = ((($info)) + 8|0); //@line 4847 "libraries/lodepng/lodepng.cpp"
    $i$153 = 0;$tRNS$sroa$0$052 = 0;$tRNS$sroa$33$051 = 0;$tRNS$sroa$54$050 = 0;
    while(1) {
     $14 = $i$153 << 2; //@line 4847 "libraries/lodepng/lodepng.cpp"
     $15 = $14 | 3; //@line 4847 "libraries/lodepng/lodepng.cpp"
     $16 = HEAP32[$11>>2]|0; //@line 4847 "libraries/lodepng/lodepng.cpp"
     $17 = (($16) + ($15)|0); //@line 4847 "libraries/lodepng/lodepng.cpp"
     $18 = HEAP8[$17>>0]|0; //@line 4847 "libraries/lodepng/lodepng.cpp"
     $19 = (($tRNS$sroa$33$051) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
     $20 = ($tRNS$sroa$54$050>>>0)<($19>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
     if ($20) {
      $21 = $tRNS$sroa$54$050 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
      $22 = ($21>>>0)<($19>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
      $23 = ($19*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
      $24 = $23 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
      $25 = $22 ? $19 : $24; //@line 227 "libraries/lodepng/lodepng.cpp"
      $26 = (_realloc($tRNS$sroa$0$052,$25)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
      $27 = ($26|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
      if ($27) {
       $tRNS$sroa$0$2 = $tRNS$sroa$0$052;$tRNS$sroa$33$1 = $tRNS$sroa$33$051;$tRNS$sroa$54$2 = $tRNS$sroa$54$050;
      } else {
       $tRNS$sroa$0$1 = $26;$tRNS$sroa$54$1 = $25;
       label = 9;
      }
     } else {
      $tRNS$sroa$0$1 = $tRNS$sroa$0$052;$tRNS$sroa$54$1 = $tRNS$sroa$54$050;
      label = 9;
     }
     if ((label|0) == 9) {
      label = 0;
      $28 = (($tRNS$sroa$0$1) + ($tRNS$sroa$33$051)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
      HEAP8[$28>>0] = $18; //@line 289 "libraries/lodepng/lodepng.cpp"
      $tRNS$sroa$0$2 = $tRNS$sroa$0$1;$tRNS$sroa$33$1 = $19;$tRNS$sroa$54$2 = $tRNS$sroa$54$1;
     }
     $29 = (($i$153) + 1)|0; //@line 4847 "libraries/lodepng/lodepng.cpp"
     $30 = ($29|0)==($amount$056$lcssa|0); //@line 4847 "libraries/lodepng/lodepng.cpp"
     if ($30) {
      $tRNS$sroa$0$19 = $tRNS$sroa$0$2;$tRNS$sroa$33$10 = $tRNS$sroa$33$1;
      break;
     } else {
      $i$153 = $29;$tRNS$sroa$0$052 = $tRNS$sroa$0$2;$tRNS$sroa$33$051 = $tRNS$sroa$33$1;$tRNS$sroa$54$050 = $tRNS$sroa$54$2;
     }
    }
   }
   break;
  }
  case 0:  {
   $31 = ((($info)) + 16|0); //@line 4851 "libraries/lodepng/lodepng.cpp"
   $32 = HEAP32[$31>>2]|0; //@line 4851 "libraries/lodepng/lodepng.cpp"
   $33 = ($32|0)==(0); //@line 4851 "libraries/lodepng/lodepng.cpp"
   if ($33) {
    $tRNS$sroa$0$19 = 0;$tRNS$sroa$33$10 = 0;
   } else {
    $34 = ((($info)) + 20|0); //@line 4853 "libraries/lodepng/lodepng.cpp"
    $35 = HEAP32[$34>>2]|0; //@line 4853 "libraries/lodepng/lodepng.cpp"
    $36 = (_realloc(0,1)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $37 = ($36|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if ($37) {
     $42 = 1;$tRNS$sroa$0$444 = 0;$tRNS$sroa$33$243 = 0;
    } else {
     $38 = $35 >>> 8; //@line 4853 "libraries/lodepng/lodepng.cpp"
     $39 = $38&255; //@line 4853 "libraries/lodepng/lodepng.cpp"
     HEAP8[$36>>0] = $39; //@line 289 "libraries/lodepng/lodepng.cpp"
     $42 = 2;$tRNS$sroa$0$444 = $36;$tRNS$sroa$33$243 = 1;
    }
    $$sink = HEAP32[$34>>2]|0; //@line 4854 "libraries/lodepng/lodepng.cpp"
    $40 = $tRNS$sroa$33$243 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $41 = ($40>>>0)<($42>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $43 = ($42*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $44 = $43 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $45 = $41 ? $42 : $44; //@line 227 "libraries/lodepng/lodepng.cpp"
    $46 = (_realloc($tRNS$sroa$0$444,$45)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $47 = ($46|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if ($47) {
     $tRNS$sroa$0$19 = $tRNS$sroa$0$444;$tRNS$sroa$33$10 = $tRNS$sroa$33$243;
    } else {
     $48 = $$sink&255; //@line 4854 "libraries/lodepng/lodepng.cpp"
     $49 = (($46) + ($tRNS$sroa$33$243)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
     HEAP8[$49>>0] = $48; //@line 289 "libraries/lodepng/lodepng.cpp"
     $tRNS$sroa$0$19 = $46;$tRNS$sroa$33$10 = $42;
    }
   }
   break;
  }
  case 2:  {
   $50 = ((($info)) + 16|0); //@line 4859 "libraries/lodepng/lodepng.cpp"
   $51 = HEAP32[$50>>2]|0; //@line 4859 "libraries/lodepng/lodepng.cpp"
   $52 = ($51|0)==(0); //@line 4859 "libraries/lodepng/lodepng.cpp"
   if ($52) {
    $tRNS$sroa$0$19 = 0;$tRNS$sroa$33$10 = 0;
   } else {
    $53 = ((($info)) + 20|0); //@line 4861 "libraries/lodepng/lodepng.cpp"
    $54 = HEAP32[$53>>2]|0; //@line 4861 "libraries/lodepng/lodepng.cpp"
    $55 = (_realloc(0,1)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $56 = ($55|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if ($56) {
     $61 = 1;$tRNS$sroa$0$848 = 0;$tRNS$sroa$33$447 = 0;
    } else {
     $57 = $54 >>> 8; //@line 4861 "libraries/lodepng/lodepng.cpp"
     $58 = $57&255; //@line 4861 "libraries/lodepng/lodepng.cpp"
     HEAP8[$55>>0] = $58; //@line 289 "libraries/lodepng/lodepng.cpp"
     $61 = 2;$tRNS$sroa$0$848 = $55;$tRNS$sroa$33$447 = 1;
    }
    $$sink49 = HEAP32[$53>>2]|0; //@line 4862 "libraries/lodepng/lodepng.cpp"
    $59 = $tRNS$sroa$33$447 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $60 = ($59>>>0)<($61>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $62 = ($61*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $63 = $62 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $64 = $60 ? $61 : $63; //@line 227 "libraries/lodepng/lodepng.cpp"
    $65 = (_realloc($tRNS$sroa$0$848,$64)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $66 = ($65|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if ($66) {
     $tRNS$sroa$0$10 = $tRNS$sroa$0$848;$tRNS$sroa$33$5 = $tRNS$sroa$33$447;$tRNS$sroa$54$8 = $tRNS$sroa$33$447;
    } else {
     $67 = $$sink49&255; //@line 4862 "libraries/lodepng/lodepng.cpp"
     $68 = (($65) + ($tRNS$sroa$33$447)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
     HEAP8[$68>>0] = $67; //@line 289 "libraries/lodepng/lodepng.cpp"
     $tRNS$sroa$0$10 = $65;$tRNS$sroa$33$5 = $61;$tRNS$sroa$54$8 = $64;
    }
    $69 = ((($info)) + 24|0); //@line 4863 "libraries/lodepng/lodepng.cpp"
    $70 = HEAP32[$69>>2]|0; //@line 4863 "libraries/lodepng/lodepng.cpp"
    $71 = $70 >>> 8; //@line 4863 "libraries/lodepng/lodepng.cpp"
    $72 = $71&255; //@line 4863 "libraries/lodepng/lodepng.cpp"
    $73 = (($tRNS$sroa$33$5) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
    $74 = ($tRNS$sroa$54$8>>>0)<($73>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
    if ($74) {
     $75 = $tRNS$sroa$54$8 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
     $76 = ($75>>>0)<($73>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
     $77 = ($73*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
     $78 = $77 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
     $79 = $76 ? $73 : $78; //@line 227 "libraries/lodepng/lodepng.cpp"
     $80 = (_realloc($tRNS$sroa$0$10,$79)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
     $81 = ($80|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
     if ($81) {
      $tRNS$sroa$0$12 = $tRNS$sroa$0$10;$tRNS$sroa$33$6 = $tRNS$sroa$33$5;$tRNS$sroa$54$10 = $tRNS$sroa$54$8;
     } else {
      $tRNS$sroa$0$11 = $80;$tRNS$sroa$54$9 = $79;
      label = 23;
     }
    } else {
     $tRNS$sroa$0$11 = $tRNS$sroa$0$10;$tRNS$sroa$54$9 = $tRNS$sroa$54$8;
     label = 23;
    }
    if ((label|0) == 23) {
     $82 = (($tRNS$sroa$0$11) + ($tRNS$sroa$33$5)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
     HEAP8[$82>>0] = $72; //@line 289 "libraries/lodepng/lodepng.cpp"
     $tRNS$sroa$0$12 = $tRNS$sroa$0$11;$tRNS$sroa$33$6 = $73;$tRNS$sroa$54$10 = $tRNS$sroa$54$9;
    }
    $83 = HEAP32[$69>>2]|0; //@line 4864 "libraries/lodepng/lodepng.cpp"
    $84 = $83&255; //@line 4864 "libraries/lodepng/lodepng.cpp"
    $85 = (($tRNS$sroa$33$6) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
    $86 = ($tRNS$sroa$54$10>>>0)<($85>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
    if ($86) {
     $87 = $tRNS$sroa$54$10 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
     $88 = ($87>>>0)<($85>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
     $89 = ($85*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
     $90 = $89 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
     $91 = $88 ? $85 : $90; //@line 227 "libraries/lodepng/lodepng.cpp"
     $92 = (_realloc($tRNS$sroa$0$12,$91)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
     $93 = ($92|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
     if ($93) {
      $tRNS$sroa$0$14 = $tRNS$sroa$0$12;$tRNS$sroa$33$7 = $tRNS$sroa$33$6;$tRNS$sroa$54$12 = $tRNS$sroa$54$10;
     } else {
      $tRNS$sroa$0$13 = $92;$tRNS$sroa$54$11 = $91;
      label = 26;
     }
    } else {
     $tRNS$sroa$0$13 = $tRNS$sroa$0$12;$tRNS$sroa$54$11 = $tRNS$sroa$54$10;
     label = 26;
    }
    if ((label|0) == 26) {
     $94 = (($tRNS$sroa$0$13) + ($tRNS$sroa$33$6)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
     HEAP8[$94>>0] = $84; //@line 289 "libraries/lodepng/lodepng.cpp"
     $tRNS$sroa$0$14 = $tRNS$sroa$0$13;$tRNS$sroa$33$7 = $85;$tRNS$sroa$54$12 = $tRNS$sroa$54$11;
    }
    $95 = ((($info)) + 28|0); //@line 4865 "libraries/lodepng/lodepng.cpp"
    $96 = HEAP32[$95>>2]|0; //@line 4865 "libraries/lodepng/lodepng.cpp"
    $97 = $96 >>> 8; //@line 4865 "libraries/lodepng/lodepng.cpp"
    $98 = $97&255; //@line 4865 "libraries/lodepng/lodepng.cpp"
    $99 = (($tRNS$sroa$33$7) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
    $100 = ($tRNS$sroa$54$12>>>0)<($99>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
    if ($100) {
     $101 = $tRNS$sroa$54$12 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
     $102 = ($101>>>0)<($99>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
     $103 = ($99*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
     $104 = $103 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
     $105 = $102 ? $99 : $104; //@line 227 "libraries/lodepng/lodepng.cpp"
     $106 = (_realloc($tRNS$sroa$0$14,$105)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
     $107 = ($106|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
     if ($107) {
      $tRNS$sroa$0$16 = $tRNS$sroa$0$14;$tRNS$sroa$33$8 = $tRNS$sroa$33$7;$tRNS$sroa$54$14 = $tRNS$sroa$54$12;
     } else {
      $tRNS$sroa$0$15 = $106;$tRNS$sroa$54$13 = $105;
      label = 29;
     }
    } else {
     $tRNS$sroa$0$15 = $tRNS$sroa$0$14;$tRNS$sroa$54$13 = $tRNS$sroa$54$12;
     label = 29;
    }
    if ((label|0) == 29) {
     $108 = (($tRNS$sroa$0$15) + ($tRNS$sroa$33$7)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
     HEAP8[$108>>0] = $98; //@line 289 "libraries/lodepng/lodepng.cpp"
     $tRNS$sroa$0$16 = $tRNS$sroa$0$15;$tRNS$sroa$33$8 = $99;$tRNS$sroa$54$14 = $tRNS$sroa$54$13;
    }
    $109 = HEAP32[$95>>2]|0; //@line 4866 "libraries/lodepng/lodepng.cpp"
    $110 = $109&255; //@line 4866 "libraries/lodepng/lodepng.cpp"
    $111 = (($tRNS$sroa$33$8) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
    $112 = ($tRNS$sroa$54$14>>>0)<($111>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
    if ($112) {
     $113 = $tRNS$sroa$54$14 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
     $114 = ($113>>>0)<($111>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
     $115 = ($111*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
     $116 = $115 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
     $117 = $114 ? $111 : $116; //@line 227 "libraries/lodepng/lodepng.cpp"
     $118 = (_realloc($tRNS$sroa$0$16,$117)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
     $119 = ($118|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
     if ($119) {
      $tRNS$sroa$0$19 = $tRNS$sroa$0$16;$tRNS$sroa$33$10 = $tRNS$sroa$33$8;
      break L1;
     } else {
      $tRNS$sroa$0$17 = $118;
     }
    } else {
     $tRNS$sroa$0$17 = $tRNS$sroa$0$16;
    }
    $120 = (($tRNS$sroa$0$17) + ($tRNS$sroa$33$8)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$120>>0] = $110; //@line 289 "libraries/lodepng/lodepng.cpp"
    $tRNS$sroa$0$19 = $tRNS$sroa$0$17;$tRNS$sroa$33$10 = $111;
   }
   break;
  }
  default: {
   $tRNS$sroa$0$19 = 0;$tRNS$sroa$33$10 = 0;
  }
  }
 } while(0);
 $121 = ((($out)) + 4|0); //@line 4775 "libraries/lodepng/lodepng.cpp"
 $122 = (__Z20lodepng_chunk_createPPhPjjPKcPKh($out,$121,$tRNS$sroa$33$10,502740,$tRNS$sroa$0$19)|0); //@line 4775 "libraries/lodepng/lodepng.cpp"
 $123 = ($122|0)==(0); //@line 4775 "libraries/lodepng/lodepng.cpp"
 if (!($123)) {
  _free($tRNS$sroa$0$19); //@line 76 "libraries/lodepng/lodepng.cpp"
  return;
 }
 $124 = HEAP32[$121>>2]|0; //@line 4776 "libraries/lodepng/lodepng.cpp"
 $125 = ((($out)) + 8|0); //@line 4776 "libraries/lodepng/lodepng.cpp"
 HEAP32[$125>>2] = $124; //@line 4776 "libraries/lodepng/lodepng.cpp"
 _free($tRNS$sroa$0$19); //@line 76 "libraries/lodepng/lodepng.cpp"
 return;
}
function __ZL13addChunk_tEXtP8ucvectorPKcS2_($out,$keyword,$textstring) {
 $out = $out|0;
 $keyword = $keyword|0;
 $textstring = $textstring|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$019 = 0, $i$019$lcssa = 0, $i$114 = 0, $text$sroa$0$018 = 0, $text$sroa$0$1 = 0, $text$sroa$0$2 = 0, $text$sroa$0$2$lcssa = 0, $text$sroa$0$3 = 0, $text$sroa$0$5$lcssa = 0, $text$sroa$0$5$ph = 0;
 var $text$sroa$0$513 = 0, $text$sroa$0$6 = 0, $text$sroa$0$7 = 0, $text$sroa$15$017 = 0, $text$sroa$15$1 = 0, $text$sroa$15$1$lcssa = 0, $text$sroa$15$3$lcssa = 0, $text$sroa$15$3$ph = 0, $text$sroa$15$312 = 0, $text$sroa$15$4 = 0, $text$sroa$24$016 = 0, $text$sroa$24$1 = 0, $text$sroa$24$2 = 0, $text$sroa$24$2$lcssa = 0, $text$sroa$24$3 = 0, $text$sroa$24$5$ph = 0, $text$sroa$24$511 = 0, $text$sroa$24$6 = 0, $text$sroa$24$7 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$keyword>>0]|0; //@line 4906 "libraries/lodepng/lodepng.cpp"
 $1 = ($0<<24>>24)==(0); //@line 4906 "libraries/lodepng/lodepng.cpp"
 if ($1) {
  return;
 } else {
  $12 = $0;$i$019 = 0;$text$sroa$0$018 = 0;$text$sroa$15$017 = 0;$text$sroa$24$016 = 0;
 }
 while(1) {
  $2 = (($text$sroa$15$017) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
  $3 = ($text$sroa$24$016>>>0)<($2>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
  if ($3) {
   $4 = $text$sroa$24$016 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
   $5 = ($4>>>0)<($2>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
   $6 = ($2*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
   $7 = $6 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
   $8 = $5 ? $2 : $7; //@line 227 "libraries/lodepng/lodepng.cpp"
   $9 = (_realloc($text$sroa$0$018,$8)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
   $10 = ($9|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
   if ($10) {
    $text$sroa$0$2 = $text$sroa$0$018;$text$sroa$15$1 = $text$sroa$15$017;$text$sroa$24$2 = $text$sroa$24$016;
   } else {
    $text$sroa$0$1 = $9;$text$sroa$24$1 = $8;
    label = 4;
   }
  } else {
   $text$sroa$0$1 = $text$sroa$0$018;$text$sroa$24$1 = $text$sroa$24$016;
   label = 4;
  }
  if ((label|0) == 4) {
   label = 0;
   $11 = (($text$sroa$0$1) + ($text$sroa$15$017)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
   HEAP8[$11>>0] = $12; //@line 289 "libraries/lodepng/lodepng.cpp"
   $text$sroa$0$2 = $text$sroa$0$1;$text$sroa$15$1 = $2;$text$sroa$24$2 = $text$sroa$24$1;
  }
  $13 = (($i$019) + 1)|0; //@line 4906 "libraries/lodepng/lodepng.cpp"
  $14 = (($keyword) + ($13)|0); //@line 4906 "libraries/lodepng/lodepng.cpp"
  $15 = HEAP8[$14>>0]|0; //@line 4906 "libraries/lodepng/lodepng.cpp"
  $16 = ($15<<24>>24)==(0); //@line 4906 "libraries/lodepng/lodepng.cpp"
  if ($16) {
   $i$019$lcssa = $i$019;$text$sroa$0$2$lcssa = $text$sroa$0$2;$text$sroa$15$1$lcssa = $text$sroa$15$1;$text$sroa$24$2$lcssa = $text$sroa$24$2;
   break;
  } else {
   $12 = $15;$i$019 = $13;$text$sroa$0$018 = $text$sroa$0$2;$text$sroa$15$017 = $text$sroa$15$1;$text$sroa$24$016 = $text$sroa$24$2;
  }
 }
 $17 = ($i$019$lcssa>>>0)>(78); //@line 4907 "libraries/lodepng/lodepng.cpp"
 if ($17) {
  return;
 }
 $18 = (($text$sroa$15$1$lcssa) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
 $19 = ($text$sroa$24$2$lcssa>>>0)<($18>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
 if ($19) {
  $20 = $text$sroa$24$2$lcssa << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
  $21 = ($20>>>0)<($18>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
  $22 = ($18*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
  $23 = $22 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
  $24 = $21 ? $18 : $23; //@line 227 "libraries/lodepng/lodepng.cpp"
  $25 = (_realloc($text$sroa$0$2$lcssa,$24)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
  $26 = ($25|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
  if ($26) {
   $text$sroa$0$5$ph = $text$sroa$0$2$lcssa;$text$sroa$15$3$ph = $text$sroa$15$1$lcssa;$text$sroa$24$5$ph = $text$sroa$24$2$lcssa;
  } else {
   $text$sroa$0$3 = $25;$text$sroa$24$3 = $24;
   label = 9;
  }
 } else {
  $text$sroa$0$3 = $text$sroa$0$2$lcssa;$text$sroa$24$3 = $text$sroa$24$2$lcssa;
  label = 9;
 }
 if ((label|0) == 9) {
  $27 = (($text$sroa$0$3) + ($text$sroa$15$1$lcssa)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
  HEAP8[$27>>0] = 0; //@line 289 "libraries/lodepng/lodepng.cpp"
  $text$sroa$0$5$ph = $text$sroa$0$3;$text$sroa$15$3$ph = $18;$text$sroa$24$5$ph = $text$sroa$24$3;
 }
 $28 = HEAP8[$textstring>>0]|0; //@line 4909 "libraries/lodepng/lodepng.cpp"
 $29 = ($28<<24>>24)==(0); //@line 4909 "libraries/lodepng/lodepng.cpp"
 if ($29) {
  $text$sroa$0$5$lcssa = $text$sroa$0$5$ph;$text$sroa$15$3$lcssa = $text$sroa$15$3$ph;
 } else {
  $40 = $28;$i$114 = 0;$text$sroa$0$513 = $text$sroa$0$5$ph;$text$sroa$15$312 = $text$sroa$15$3$ph;$text$sroa$24$511 = $text$sroa$24$5$ph;
  while(1) {
   $30 = (($text$sroa$15$312) + 1)|0; //@line 288 "libraries/lodepng/lodepng.cpp"
   $31 = ($text$sroa$24$511>>>0)<($30>>>0); //@line 225 "libraries/lodepng/lodepng.cpp"
   if ($31) {
    $32 = $text$sroa$24$511 << 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $33 = ($32>>>0)<($30>>>0); //@line 227 "libraries/lodepng/lodepng.cpp"
    $34 = ($30*3)|0; //@line 227 "libraries/lodepng/lodepng.cpp"
    $35 = $34 >>> 1; //@line 227 "libraries/lodepng/lodepng.cpp"
    $36 = $33 ? $30 : $35; //@line 227 "libraries/lodepng/lodepng.cpp"
    $37 = (_realloc($text$sroa$0$513,$36)|0); //@line 71 "libraries/lodepng/lodepng.cpp"
    $38 = ($37|0)==(0|0); //@line 229 "libraries/lodepng/lodepng.cpp"
    if ($38) {
     $text$sroa$0$7 = $text$sroa$0$513;$text$sroa$15$4 = $text$sroa$15$312;$text$sroa$24$7 = $text$sroa$24$511;
    } else {
     $text$sroa$0$6 = $37;$text$sroa$24$6 = $36;
     label = 13;
    }
   } else {
    $text$sroa$0$6 = $text$sroa$0$513;$text$sroa$24$6 = $text$sroa$24$511;
    label = 13;
   }
   if ((label|0) == 13) {
    label = 0;
    $39 = (($text$sroa$0$6) + ($text$sroa$15$312)|0); //@line 289 "libraries/lodepng/lodepng.cpp"
    HEAP8[$39>>0] = $40; //@line 289 "libraries/lodepng/lodepng.cpp"
    $text$sroa$0$7 = $text$sroa$0$6;$text$sroa$15$4 = $30;$text$sroa$24$7 = $text$sroa$24$6;
   }
   $41 = (($i$114) + 1)|0; //@line 4909 "libraries/lodepng/lodepng.cpp"
   $42 = (($textstring) + ($41)|0); //@line 4909 "libraries/lodepng/lodepng.cpp"
   $43 = HEAP8[$42>>0]|0; //@line 4909 "libraries/lodepng/lodepng.cpp"
   $44 = ($43<<24>>24)==(0); //@line 4909 "libraries/lodepng/lodepng.cpp"
   if ($44) {
    $text$sroa$0$5$lcssa = $text$sroa$0$7;$text$sroa$15$3$lcssa = $text$sroa$15$4;
    break;
   } else {
    $40 = $43;$i$114 = $41;$text$sroa$0$513 = $text$sroa$0$7;$text$sroa$15$312 = $text$sroa$15$4;$text$sroa$24$511 = $text$sroa$24$7;
   }
  }
 }
 $45 = ((($out)) + 4|0); //@line 4775 "libraries/lodepng/lodepng.cpp"
 $46 = (__Z20lodepng_chunk_createPPhPjjPKcPKh($out,$45,$text$sroa$15$3$lcssa,502765,$text$sroa$0$5$lcssa)|0); //@line 4775 "libraries/lodepng/lodepng.cpp"
 $47 = ($46|0)==(0); //@line 4775 "libraries/lodepng/lodepng.cpp"
 if ($47) {
  $48 = HEAP32[$45>>2]|0; //@line 4776 "libraries/lodepng/lodepng.cpp"
  $49 = ((($out)) + 8|0); //@line 4776 "libraries/lodepng/lodepng.cpp"
  HEAP32[$49>>2] = $48; //@line 4776 "libraries/lodepng/lodepng.cpp"
 }
 _free($text$sroa$0$5$lcssa); //@line 76 "libraries/lodepng/lodepng.cpp"
 return;
}
function __Z18lodepng_error_textj($code) {
 $code = $code|0;
 var $$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 do {
  switch ($code|0) {
  case 0:  {
   $$0 = 506491;
   break;
  }
  case 1:  {
   $$0 = 506474;
   break;
  }
  case 10:  {
   $$0 = 506421;
   break;
  }
  case 11:  {
   $$0 = 506365;
   break;
  }
  case 13:  {
   $$0 = 506318;
   break;
  }
  case 14:  {
   $$0 = 506318;
   break;
  }
  case 15:  {
   $$0 = 506318;
   break;
  }
  case 16:  {
   $$0 = 506263;
   break;
  }
  case 17:  {
   $$0 = 506067;
   break;
  }
  case 18:  {
   $$0 = 506225;
   break;
  }
  case 19:  {
   $$0 = 506067;
   break;
  }
  case 20:  {
   $$0 = 506170;
   break;
  }
  case 21:  {
   $$0 = 506116;
   break;
  }
  case 22:  {
   $$0 = 506067;
   break;
  }
  case 23:  {
   $$0 = 506019;
   break;
  }
  case 24:  {
   $$0 = 505989;
   break;
  }
  case 25:  {
   $$0 = 505947;
   break;
  }
  case 26:  {
   $$0 = 505886;
   break;
  }
  case 27:  {
   $$0 = 505848;
   break;
  }
  case 28:  {
   $$0 = 505798;
   break;
  }
  case 29:  {
   $$0 = 505762;
   break;
  }
  case 30:  {
   $$0 = 505706;
   break;
  }
  case 31:  {
   $$0 = 505676;
   break;
  }
  case 32:  {
   $$0 = 505645;
   break;
  }
  case 33:  {
   $$0 = 505619;
   break;
  }
  case 34:  {
   $$0 = 505590;
   break;
  }
  case 35:  {
   $$0 = 505530;
   break;
  }
  case 36:  {
   $$0 = 505494;
   break;
  }
  case 37:  {
   $$0 = 505450;
   break;
  }
  case 38:  {
   $$0 = 505427;
   break;
  }
  case 39:  {
   $$0 = 505344;
   break;
  }
  case 40:  {
   $$0 = 505298;
   break;
  }
  case 41:  {
   $$0 = 505258;
   break;
  }
  case 42:  {
   $$0 = 505193;
   break;
  }
  case 43:  {
   $$0 = 505149;
   break;
  }
  case 44:  {
   $$0 = 505103;
   break;
  }
  case 45:  {
   $$0 = 505063;
   break;
  }
  case 48:  {
   $$0 = 505029;
   break;
  }
  case 49:  {
   $$0 = 504972;
   break;
  }
  case 50:  {
   $$0 = 504972;
   break;
  }
  case 51:  {
   $$0 = 504923;
   break;
  }
  case 52:  {
   $$0 = 504888;
   break;
  }
  case 53:  {
   $$0 = 504860;
   break;
  }
  case 54:  {
   $$0 = 504802;
   break;
  }
  case 55:  {
   $$0 = 504755;
   break;
  }
  case 56:  {
   $$0 = 504679;
   break;
  }
  case 57:  {
   $$0 = 504624;
   break;
  }
  case 58:  {
   $$0 = 504561;
   break;
  }
  case 59:  {
   $$0 = 504520;
   break;
  }
  case 60:  {
   $$0 = 504445;
   break;
  }
  case 61:  {
   $$0 = 504364;
   break;
  }
  case 62:  {
   $$0 = 504315;
   break;
  }
  case 63:  {
   $$0 = 504237;
   break;
  }
  case 64:  {
   $$0 = 504179;
   break;
  }
  case 66:  {
   $$0 = 504084;
   break;
  }
  case 67:  {
   $$0 = 503990;
   break;
  }
  case 68:  {
   $$0 = 503901;
   break;
  }
  case 69:  {
   $$0 = 503834;
   break;
  }
  case 71:  {
   $$0 = 503774;
   break;
  }
  case 72:  {
   $$0 = 503678;
   break;
  }
  case 73:  {
   $$0 = 503654;
   break;
  }
  case 74:  {
   $$0 = 503630;
   break;
  }
  case 75:  {
   $$0 = 503573;
   break;
  }
  case 76:  {
   $$0 = 503526;
   break;
  }
  case 77:  {
   $$0 = 503494;
   break;
  }
  case 78:  {
   $$0 = 503462;
   break;
  }
  case 79:  {
   $$0 = 503430;
   break;
  }
  case 80:  {
   $$0 = 503395;
   break;
  }
  case 81:  {
   $$0 = 503358;
   break;
  }
  case 82:  {
   $$0 = 503289;
   break;
  }
  case 83:  {
   $$0 = 503264;
   break;
  }
  case 84:  {
   $$0 = 503206;
   break;
  }
  case 86:  {
   $$0 = 503156;
   break;
  }
  case 87:  {
   $$0 = 503075;
   break;
  }
  case 88:  {
   $$0 = 503002;
   break;
  }
  case 89:  {
   $$0 = 502944;
   break;
  }
  case 90:  {
   $$0 = 502910;
   break;
  }
  case 91:  {
   $$0 = 502879;
   break;
  }
  case 92:  {
   $$0 = 502848;
   break;
  }
  case 93:  {
   $$0 = 502816;
   break;
  }
  default: {
   $$0 = 502797;
  }
  }
 } while(0);
 return ($$0|0); //@line 5908 "libraries/lodepng/lodepng.cpp"
}
function __ZN7lodepng5StateC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep4$i$i$i = 0, dest = 0, label = 0, sp = 0;
 var stop = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (33380); //@line 5994 "libraries/lodepng/lodepng.cpp"
 $0 = ((($this)) + 4|0); //@line 4738 "libraries/lodepng/lodepng.cpp"
 $1 = ((($this)) + 24|0); //@line 4722 "libraries/lodepng/lodepng.cpp"
 HEAP32[$1>>2] = 1; //@line 4722 "libraries/lodepng/lodepng.cpp"
 $2 = ((($this)) + 28|0); //@line 4724 "libraries/lodepng/lodepng.cpp"
 HEAP32[$2>>2] = 1; //@line 4724 "libraries/lodepng/lodepng.cpp"
 $3 = ((($this)) + 32|0); //@line 4725 "libraries/lodepng/lodepng.cpp"
 HEAP32[$3>>2] = 0; //@line 4725 "libraries/lodepng/lodepng.cpp"
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;HEAP32[$0+12>>2]=0|0;HEAP32[$0+16>>2]=0|0; //@line 4727 "libraries/lodepng/lodepng.cpp"
 $4 = ((($this)) + 36|0); //@line 2252 "libraries/lodepng/lodepng.cpp"
 HEAP32[$4>>2] = 2; //@line 2252 "libraries/lodepng/lodepng.cpp"
 $5 = ((($this)) + 40|0); //@line 2253 "libraries/lodepng/lodepng.cpp"
 HEAP32[$5>>2] = 1; //@line 2253 "libraries/lodepng/lodepng.cpp"
 $6 = ((($this)) + 44|0); //@line 2254 "libraries/lodepng/lodepng.cpp"
 HEAP32[$6>>2] = 2048; //@line 2254 "libraries/lodepng/lodepng.cpp"
 $7 = ((($this)) + 48|0); //@line 2255 "libraries/lodepng/lodepng.cpp"
 HEAP32[$7>>2] = 3; //@line 2255 "libraries/lodepng/lodepng.cpp"
 $8 = ((($this)) + 52|0); //@line 2256 "libraries/lodepng/lodepng.cpp"
 HEAP32[$8>>2] = 128; //@line 2256 "libraries/lodepng/lodepng.cpp"
 $9 = ((($this)) + 56|0); //@line 2257 "libraries/lodepng/lodepng.cpp"
 HEAP32[$9>>2] = 1; //@line 2257 "libraries/lodepng/lodepng.cpp"
 $10 = ((($this)) + 60|0); //@line 2259 "libraries/lodepng/lodepng.cpp"
 HEAP32[$10>>2] = 0; //@line 2259 "libraries/lodepng/lodepng.cpp"
 $11 = ((($this)) + 64|0); //@line 2260 "libraries/lodepng/lodepng.cpp"
 HEAP32[$11>>2] = 0; //@line 2260 "libraries/lodepng/lodepng.cpp"
 $12 = ((($this)) + 68|0); //@line 2261 "libraries/lodepng/lodepng.cpp"
 HEAP32[$12>>2] = 0; //@line 2261 "libraries/lodepng/lodepng.cpp"
 $13 = ((($this)) + 76|0); //@line 5790 "libraries/lodepng/lodepng.cpp"
 HEAP32[$13>>2] = 1; //@line 5790 "libraries/lodepng/lodepng.cpp"
 $14 = ((($this)) + 80|0); //@line 5791 "libraries/lodepng/lodepng.cpp"
 HEAP32[$14>>2] = 1; //@line 5791 "libraries/lodepng/lodepng.cpp"
 $15 = ((($this)) + 72|0); //@line 5792 "libraries/lodepng/lodepng.cpp"
 HEAP32[$15>>2] = 1; //@line 5792 "libraries/lodepng/lodepng.cpp"
 $16 = ((($this)) + 88|0); //@line 5793 "libraries/lodepng/lodepng.cpp"
 HEAP32[$16>>2] = 0; //@line 5793 "libraries/lodepng/lodepng.cpp"
 $17 = ((($this)) + 84|0); //@line 5794 "libraries/lodepng/lodepng.cpp"
 HEAP32[$17>>2] = 0; //@line 5794 "libraries/lodepng/lodepng.cpp"
 $18 = ((($this)) + 92|0); //@line 5796 "libraries/lodepng/lodepng.cpp"
 HEAP32[$18>>2] = 0; //@line 5796 "libraries/lodepng/lodepng.cpp"
 $19 = ((($this)) + 96|0); //@line 5797 "libraries/lodepng/lodepng.cpp"
 HEAP32[$19>>2] = 1; //@line 5797 "libraries/lodepng/lodepng.cpp"
 $20 = ((($this)) + 116|0); //@line 2556 "libraries/lodepng/lodepng.cpp"
 $21 = ((($this)) + 100|0); //@line 2558 "libraries/lodepng/lodepng.cpp"
 ;HEAP32[$20>>2]=0|0;HEAP32[$20+4>>2]=0|0;HEAP32[$20+8>>2]=0|0;HEAP32[$20+12>>2]=0|0; //@line 2557 "libraries/lodepng/lodepng.cpp"
 HEAP32[$21>>2] = 6; //@line 2558 "libraries/lodepng/lodepng.cpp"
 $22 = ((($this)) + 104|0); //@line 2559 "libraries/lodepng/lodepng.cpp"
 HEAP32[$22>>2] = 8; //@line 2559 "libraries/lodepng/lodepng.cpp"
 $23 = ((($this)) + 108|0); //@line 2560 "libraries/lodepng/lodepng.cpp"
 HEAP32[$23>>2] = 0; //@line 2560 "libraries/lodepng/lodepng.cpp"
 $24 = ((($this)) + 112|0); //@line 2561 "libraries/lodepng/lodepng.cpp"
 HEAP32[$24>>2] = 0; //@line 2561 "libraries/lodepng/lodepng.cpp"
 $25 = ((($this)) + 160|0); //@line 2556 "libraries/lodepng/lodepng.cpp"
 $26 = ((($this)) + 144|0); //@line 2558 "libraries/lodepng/lodepng.cpp"
 ;HEAP32[$25>>2]=0|0;HEAP32[$25+4>>2]=0|0;HEAP32[$25+8>>2]=0|0;HEAP32[$25+12>>2]=0|0; //@line 2557 "libraries/lodepng/lodepng.cpp"
 HEAP32[$26>>2] = 6; //@line 2558 "libraries/lodepng/lodepng.cpp"
 $27 = ((($this)) + 148|0); //@line 2559 "libraries/lodepng/lodepng.cpp"
 HEAP32[$27>>2] = 8; //@line 2559 "libraries/lodepng/lodepng.cpp"
 $28 = ((($this)) + 152|0); //@line 2560 "libraries/lodepng/lodepng.cpp"
 HEAP32[$28>>2] = 0; //@line 2560 "libraries/lodepng/lodepng.cpp"
 $29 = ((($this)) + 156|0); //@line 2561 "libraries/lodepng/lodepng.cpp"
 HEAP32[$29>>2] = 0; //@line 2561 "libraries/lodepng/lodepng.cpp"
 $30 = ((($this)) + 140|0); //@line 2883 "libraries/lodepng/lodepng.cpp"
 HEAP32[$30>>2] = 0; //@line 2883 "libraries/lodepng/lodepng.cpp"
 $31 = ((($this)) + 132|0); //@line 2884 "libraries/lodepng/lodepng.cpp"
 HEAP32[$31>>2] = 0; //@line 2884 "libraries/lodepng/lodepng.cpp"
 $32 = ((($this)) + 136|0); //@line 2885 "libraries/lodepng/lodepng.cpp"
 HEAP32[$32>>2] = 0; //@line 2885 "libraries/lodepng/lodepng.cpp"
 $33 = ((($this)) + 176|0); //@line 2887 "libraries/lodepng/lodepng.cpp"
 $34 = ((($this)) + 252|0); //@line 2894 "libraries/lodepng/lodepng.cpp"
 HEAP32[$34>>2] = 0; //@line 2894 "libraries/lodepng/lodepng.cpp"
 $scevgep4$i$i$i = ((($this)) + 268|0); //@line 2700 "libraries/lodepng/lodepng.cpp"
 ;HEAP32[$scevgep4$i$i$i>>2]=0|0;HEAP32[$scevgep4$i$i$i+4>>2]=0|0;HEAP32[$scevgep4$i$i$i+8>>2]=0|0;HEAP32[$scevgep4$i$i$i+12>>2]=0|0;HEAP32[$scevgep4$i$i$i+16>>2]=0|0;HEAP32[$scevgep4$i$i$i+20>>2]=0|0; //@line 2701 "libraries/lodepng/lodepng.cpp"
 dest=$33; stop=dest+52|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 2887 "libraries/lodepng/lodepng.cpp"
 $35 = ((($this)) + 292|0); //@line 4745 "libraries/lodepng/lodepng.cpp"
 HEAP32[$35>>2] = 1; //@line 4745 "libraries/lodepng/lodepng.cpp"
 return; //@line 5996 "libraries/lodepng/lodepng.cpp"
}
function _auxiliar_open($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 STACKTOP = sp;return 0;
}
function _auxiliar_newclass($L,$classname,$func) {
 $L = $L|0;
 $classname = $classname|0;
 $func = $func|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $classname;
 $2 = $func;
 $3 = $0;
 $4 = $1;
 (_luaL_newmetatable($3,$4)|0);
 $5 = $0;
 _lua_pushstring($5,523215);
 $6 = $0;
 _lua_createtable($6,0,0);
 $7 = $0;
 _lua_pushstring($7,506520);
 $8 = $0;
 $9 = $1;
 _lua_pushstring($8,$9);
 $10 = $0;
 _lua_rawset($10,-3);
 while(1) {
  $11 = $2;
  $12 = HEAP32[$11>>2]|0;
  $13 = ($12|0)!=(0|0);
  $14 = $0;
  if (!($13)) {
   break;
  }
  $15 = $2;
  $16 = HEAP32[$15>>2]|0;
  _lua_pushstring($14,$16);
  $17 = $0;
  $18 = $2;
  $19 = ((($18)) + 4|0);
  $20 = HEAP32[$19>>2]|0;
  _lua_pushcclosure($17,$20,0);
  $21 = $0;
  $22 = $2;
  $23 = HEAP32[$22>>2]|0;
  $24 = HEAP8[$23>>0]|0;
  $25 = $24 << 24 >> 24;
  $26 = ($25|0)==(95);
  $27 = $26 ? -5 : -3;
  _lua_rawset($21,$27);
  $28 = $2;
  $29 = ((($28)) + 8|0);
  $2 = $29;
 }
 _lua_rawset($14,-3);
 $30 = $0;
 _lua_settop($30,-2);
 STACKTOP = sp;return;
}
function _auxiliar_tostring($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $buf = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_ptr4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $buf = sp + 24|0;
 $1 = $L;
 $2 = $1;
 $3 = (_lua_getmetatable($2,1)|0);
 $4 = ($3|0)!=(0);
 if ($4) {
  $5 = $1;
  _lua_pushstring($5,523215);
  $6 = $1;
  _lua_gettable($6,-2);
  $7 = $1;
  $8 = (_lua_type($7,-1)|0);
  $9 = ($8|0)==(5);
  if ($9) {
   $10 = $1;
   _lua_pushstring($10,506520);
   $11 = $1;
   _lua_gettable($11,-2);
   $12 = $1;
   $13 = (_lua_isstring($12,-1)|0);
   $14 = ($13|0)!=(0);
   if ($14) {
    $15 = $1;
    $16 = (_lua_touserdata($15,1)|0);
    HEAP32[$vararg_buffer>>2] = $16;
    (_sprintf($buf,620521,$vararg_buffer)|0);
    $17 = $1;
    $18 = $1;
    $19 = (_lua_tolstring($18,-1,0)|0);
    HEAP32[$vararg_buffer1>>2] = $19;
    $vararg_ptr4 = ((($vararg_buffer1)) + 4|0);
    HEAP32[$vararg_ptr4>>2] = $buf;
    (_lua_pushfstring($17,522142,$vararg_buffer1)|0);
    $0 = 1;
    $22 = $0;
    STACKTOP = sp;return ($22|0);
   }
  }
 }
 $20 = $1;
 _lua_pushstring($20,506526);
 $21 = $1;
 (_lua_error($21)|0);
 $0 = 1;
 $22 = $0;
 STACKTOP = sp;return ($22|0);
}
function _auxiliar_add2group($L,$classname,$groupname) {
 $L = $L|0;
 $classname = $classname|0;
 $groupname = $groupname|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $classname;
 $2 = $groupname;
 $3 = $0;
 $4 = $1;
 _lua_getfield($3,-10000,$4);
 $5 = $0;
 $6 = $2;
 _lua_pushstring($5,$6);
 $7 = $0;
 _lua_pushboolean($7,1);
 $8 = $0;
 _lua_rawset($8,-3);
 $9 = $0;
 _lua_settop($9,-2);
 STACKTOP = sp;return;
}
function _auxiliar_checkboolean($L,$objidx) {
 $L = $L|0;
 $objidx = $objidx|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $objidx;
 $2 = $0;
 $3 = $1;
 $4 = (_lua_type($2,$3)|0);
 $5 = ($4|0)==(1);
 if (!($5)) {
  $6 = $0;
  $7 = $1;
  $8 = $0;
  $9 = (_lua_typename($8,1)|0);
  (_luaL_typerror($6,$7,$9)|0);
 }
 $10 = $0;
 $11 = $1;
 $12 = (_lua_toboolean($10,$11)|0);
 STACKTOP = sp;return ($12|0);
}
function _auxiliar_checkclass($L,$classname,$objidx) {
 $L = $L|0;
 $classname = $classname|0;
 $objidx = $objidx|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $data = 0, $msg = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $msg = sp + 20|0;
 $0 = $L;
 $1 = $classname;
 $2 = $objidx;
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = (_auxiliar_getclassudata($3,$4,$5)|0);
 $data = $6;
 $7 = $data;
 $8 = ($7|0)!=(0|0);
 if ($8) {
  $12 = $data;
  STACKTOP = sp;return ($12|0);
 }
 $9 = $1;
 HEAP32[$vararg_buffer>>2] = $9;
 (_sprintf($msg,506575,$vararg_buffer)|0);
 $10 = $0;
 $11 = $2;
 (_luaL_argerror($10,$11,$msg)|0);
 $12 = $data;
 STACKTOP = sp;return ($12|0);
}
function _auxiliar_getclassudata($L,$classname,$objidx) {
 $L = $L|0;
 $classname = $classname|0;
 $objidx = $objidx|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $classname;
 $2 = $objidx;
 $3 = $0;
 $4 = $2;
 $5 = $1;
 $6 = (_luaL_checkudata($3,$4,$5)|0);
 STACKTOP = sp;return ($6|0);
}
function _auxiliar_checkgroup($L,$groupname,$objidx) {
 $L = $L|0;
 $groupname = $groupname|0;
 $objidx = $objidx|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $data = 0, $msg = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $msg = sp + 20|0;
 $0 = $L;
 $1 = $groupname;
 $2 = $objidx;
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = (_auxiliar_getgroupudata($3,$4,$5)|0);
 $data = $6;
 $7 = $data;
 $8 = ($7|0)!=(0|0);
 if ($8) {
  $12 = $data;
  STACKTOP = sp;return ($12|0);
 }
 $9 = $1;
 HEAP32[$vararg_buffer>>2] = $9;
 (_sprintf($msg,506575,$vararg_buffer)|0);
 $10 = $0;
 $11 = $2;
 (_luaL_argerror($10,$11,$msg)|0);
 $12 = $data;
 STACKTOP = sp;return ($12|0);
}
function _auxiliar_getgroupudata($L,$groupname,$objidx) {
 $L = $L|0;
 $groupname = $groupname|0;
 $objidx = $objidx|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $L;
 $2 = $groupname;
 $3 = $objidx;
 $4 = $1;
 $5 = $3;
 $6 = (_lua_getmetatable($4,$5)|0);
 $7 = ($6|0)!=(0);
 do {
  if ($7) {
   $8 = $1;
   $9 = $2;
   _lua_pushstring($8,$9);
   $10 = $1;
   _lua_rawget($10,-2);
   $11 = $1;
   $12 = (_lua_type($11,-1)|0);
   $13 = ($12|0)==(0);
   $14 = $1;
   _lua_settop($14,-3);
   if ($13) {
    $0 = 0;
    break;
   } else {
    $15 = $1;
    $16 = $3;
    $17 = (_lua_touserdata($15,$16)|0);
    $0 = $17;
    break;
   }
  } else {
   $0 = 0;
  }
 } while(0);
 $18 = $0;
 STACKTOP = sp;return ($18|0);
}
function _auxiliar_setclass($L,$classname,$objidx) {
 $L = $L|0;
 $classname = $classname|0;
 $objidx = $objidx|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $classname;
 $2 = $objidx;
 $3 = $0;
 $4 = $1;
 _lua_getfield($3,-10000,$4);
 $5 = $2;
 $6 = ($5|0)<(0);
 if ($6) {
  $7 = $2;
  $8 = (($7) + -1)|0;
  $2 = $8;
 }
 $9 = $0;
 $10 = $2;
 (_lua_setmetatable($9,$10)|0);
 STACKTOP = sp;return;
}
function _buffer_open($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 STACKTOP = sp;return 0;
}
function _buffer_init($buf,$io,$tm) {
 $buf = $buf|0;
 $io = $io|0;
 $tm = $tm|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $buf;
 $1 = $io;
 $2 = $tm;
 $3 = $0;
 $4 = ((($3)) + 28|0);
 HEAP32[$4>>2] = 0;
 $5 = $0;
 $6 = ((($5)) + 24|0);
 HEAP32[$6>>2] = 0;
 $7 = $1;
 $8 = $0;
 $9 = ((($8)) + 16|0);
 HEAP32[$9>>2] = $7;
 $10 = $2;
 $11 = $0;
 $12 = ((($11)) + 20|0);
 HEAP32[$12>>2] = $10;
 $13 = $0;
 $14 = ((($13)) + 8|0);
 HEAP32[$14>>2] = 0;
 $15 = $0;
 $16 = ((($15)) + 12|0);
 HEAP32[$16>>2] = 0;
 $17 = (+_timeout_gettime());
 $18 = $0;
 HEAPF64[$18>>3] = $17;
 STACKTOP = sp;return;
}
function _buffer_meth_getstats($L,$buf) {
 $L = $L|0;
 $buf = $buf|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $buf;
 $2 = $0;
 $3 = $1;
 $4 = ((($3)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (+($5>>>0));
 _lua_pushnumber($2,$6);
 $7 = $0;
 $8 = $1;
 $9 = ((($8)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (+($10>>>0));
 _lua_pushnumber($7,$11);
 $12 = $0;
 $13 = (+_timeout_gettime());
 $14 = $1;
 $15 = +HEAPF64[$14>>3];
 $16 = $13 - $15;
 _lua_pushnumber($12,$16);
 STACKTOP = sp;return 3;
}
function _buffer_meth_setstats($L,$buf) {
 $L = $L|0;
 $buf = $buf|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0;
 var $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $buf;
 $2 = $0;
 $3 = $1;
 $4 = ((($3)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (+($5>>>0));
 $7 = (+_luaL_optnumber($2,2,$6));
 $8 = (~~(($7)));
 $9 = $1;
 $10 = ((($9)) + 12|0);
 HEAP32[$10>>2] = $8;
 $11 = $0;
 $12 = $1;
 $13 = ((($12)) + 8|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (+($14>>>0));
 $16 = (+_luaL_optnumber($11,3,$15));
 $17 = (~~(($16)));
 $18 = $1;
 $19 = ((($18)) + 8|0);
 HEAP32[$19>>2] = $17;
 $20 = $0;
 $21 = (_lua_isnumber($20,4)|0);
 $22 = ($21|0)!=(0);
 if (!($22)) {
  $28 = $0;
  _lua_pushnumber($28,1.0);
  STACKTOP = sp;return 1;
 }
 $23 = (+_timeout_gettime());
 $24 = $0;
 $25 = (+_lua_tonumber($24,4));
 $26 = $23 - $25;
 $27 = $1;
 HEAPF64[$27>>3] = $26;
 $28 = $0;
 _lua_pushnumber($28,1.0);
 STACKTOP = sp;return 1;
}
function _buffer_meth_send($L,$buf) {
 $L = $L|0;
 $buf = $buf|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0.0, $68 = 0, $69 = 0, $7 = 0.0, $70 = 0, $71 = 0, $72 = 0.0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $data = 0;
 var $end = 0, $err = 0, $sent = 0, $size = 0, $start = 0, $tm = 0, $top = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $size = sp + 20|0;
 $sent = sp + 16|0;
 $0 = $L;
 $1 = $buf;
 $2 = $0;
 $3 = (_lua_gettop($2)|0);
 $top = $3;
 $err = 0;
 HEAP32[$size>>2] = 0;
 HEAP32[$sent>>2] = 0;
 $4 = $0;
 $5 = (_luaL_checklstring($4,2,$size)|0);
 $data = $5;
 $6 = $0;
 $7 = (+_luaL_optnumber($6,3,1.0));
 $8 = (~~(($7)));
 $start = $8;
 $9 = $0;
 $10 = (+_luaL_optnumber($9,4,-1.0));
 $11 = (~~(($10)));
 $end = $11;
 $12 = $1;
 $13 = ((($12)) + 20|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (_timeout_markstart($14)|0);
 $tm = $15;
 $16 = $start;
 $17 = ($16|0)<(0);
 if ($17) {
  $18 = HEAP32[$size>>2]|0;
  $19 = $start;
  $20 = (($18) + ($19))|0;
  $21 = (($20) + 1)|0;
  $start = $21;
 }
 $22 = $end;
 $23 = ($22|0)<(0);
 if ($23) {
  $24 = HEAP32[$size>>2]|0;
  $25 = $end;
  $26 = (($24) + ($25))|0;
  $27 = (($26) + 1)|0;
  $end = $27;
 }
 $28 = $start;
 $29 = ($28|0)<(1);
 if ($29) {
  $start = 1;
 }
 $30 = $end;
 $31 = HEAP32[$size>>2]|0;
 $32 = ($30|0)>($31|0);
 if ($32) {
  $33 = HEAP32[$size>>2]|0;
  $end = $33;
 }
 $34 = $start;
 $35 = $end;
 $36 = ($34|0)<=($35|0);
 if ($36) {
  $37 = $1;
  $38 = $data;
  $39 = $start;
  $40 = (($38) + ($39)|0);
  $41 = ((($40)) + -1|0);
  $42 = $end;
  $43 = $start;
  $44 = (($42) - ($43))|0;
  $45 = (($44) + 1)|0;
  $46 = (_sendraw($37,$41,$45,$sent)|0);
  $err = $46;
 }
 $47 = $err;
 $48 = ($47|0)!=(0);
 $49 = $0;
 if ($48) {
  _lua_pushnil($49);
  $50 = $0;
  $51 = $1;
  $52 = ((($51)) + 16|0);
  $53 = HEAP32[$52>>2]|0;
  $54 = ((($53)) + 12|0);
  $55 = HEAP32[$54>>2]|0;
  $56 = $1;
  $57 = ((($56)) + 16|0);
  $58 = HEAP32[$57>>2]|0;
  $59 = HEAP32[$58>>2]|0;
  $60 = $err;
  $61 = (FUNCTION_TABLE_iii[$55 & 8191]($59,$60)|0);
  _lua_pushstring($50,$61);
  $62 = $0;
  $63 = HEAP32[$sent>>2]|0;
  $64 = $start;
  $65 = (($63) + ($64))|0;
  $66 = (($65) - 1)|0;
  $67 = (+($66>>>0));
  _lua_pushnumber($62,$67);
  $75 = $0;
  $76 = (_lua_gettop($75)|0);
  $77 = $top;
  $78 = (($76) - ($77))|0;
  STACKTOP = sp;return ($78|0);
 } else {
  $68 = HEAP32[$sent>>2]|0;
  $69 = $start;
  $70 = (($68) + ($69))|0;
  $71 = (($70) - 1)|0;
  $72 = (+($71>>>0));
  _lua_pushnumber($49,$72);
  $73 = $0;
  _lua_pushnil($73);
  $74 = $0;
  _lua_pushnil($74);
  $75 = $0;
  $76 = (_lua_gettop($75)|0);
  $77 = $top;
  $78 = (($76) - ($77))|0;
  STACKTOP = sp;return ($78|0);
 }
 return (0)|0;
}
function _sendraw($buf,$data,$count,$sent) {
 $buf = $buf|0;
 $data = $data|0;
 $count = $count|0;
 $sent = $sent|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $done = 0, $err = 0, $io = 0, $step = 0, $tm = 0, $total = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $done = sp + 4|0;
 $0 = $buf;
 $1 = $data;
 $2 = $count;
 $3 = $sent;
 $4 = $0;
 $5 = ((($4)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $io = $6;
 $7 = $0;
 $8 = ((($7)) + 20|0);
 $9 = HEAP32[$8>>2]|0;
 $tm = $9;
 $total = 0;
 $err = 0;
 while(1) {
  $10 = $total;
  $11 = $2;
  $12 = ($10>>>0)<($11>>>0);
  $13 = $err;
  $14 = ($13|0)==(0);
  $15 = $12 ? $14 : 0;
  if (!($15)) {
   break;
  }
  $16 = $2;
  $17 = $total;
  $18 = (($16) - ($17))|0;
  $19 = ($18>>>0)<=(8192);
  if ($19) {
   $20 = $2;
   $21 = $total;
   $22 = (($20) - ($21))|0;
   $23 = $22;
  } else {
   $23 = 8192;
  }
  $step = $23;
  $24 = $io;
  $25 = ((($24)) + 4|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = $io;
  $28 = HEAP32[$27>>2]|0;
  $29 = $1;
  $30 = $total;
  $31 = (($29) + ($30)|0);
  $32 = $step;
  $33 = $tm;
  $34 = (FUNCTION_TABLE_iiiiii[$26 & 8191]($28,$31,$32,$done,$33)|0);
  $err = $34;
  $35 = HEAP32[$done>>2]|0;
  $36 = $total;
  $37 = (($36) + ($35))|0;
  $total = $37;
 }
 $38 = $total;
 $39 = $3;
 HEAP32[$39>>2] = $38;
 $40 = $total;
 $41 = $0;
 $42 = ((($41)) + 8|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = (($43) + ($40))|0;
 HEAP32[$42>>2] = $44;
 $45 = $err;
 STACKTOP = sp;return ($45|0);
}
function _buffer_meth_receive($L,$buf) {
 $L = $L|0;
 $buf = $buf|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0.0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $8 = 0, $9 = 0, $b = 0, $err = 0, $p = 0, $part = 0, $size = 0, $tm = 0, $top = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1072|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $b = sp + 16|0;
 $size = sp + 12|0;
 $0 = $L;
 $1 = $buf;
 $err = 0;
 $2 = $0;
 $3 = (_lua_gettop($2)|0);
 $top = $3;
 $4 = $0;
 $5 = (_luaL_optlstring($4,3,1282466,$size)|0);
 $part = $5;
 $6 = $1;
 $7 = ((($6)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (_timeout_markstart($8)|0);
 $tm = $9;
 $10 = $0;
 _luaL_buffinit($10,$b);
 $11 = $part;
 $12 = HEAP32[$size>>2]|0;
 _luaL_addlstring($b,$11,$12);
 $13 = $0;
 $14 = (_lua_isnumber($13,2)|0);
 $15 = ($14|0)!=(0);
 do {
  if ($15) {
   $41 = $1;
   $42 = $0;
   $43 = (+_lua_tonumber($42,2));
   $44 = (~~(($43))>>>0);
   $45 = HEAP32[$size>>2]|0;
   $46 = (($44) - ($45))|0;
   $47 = (_recvraw($41,$46,$b)|0);
   $err = $47;
  } else {
   $16 = $0;
   $17 = (_luaL_optlstring($16,2,506590,0)|0);
   $p = $17;
   $18 = $p;
   $19 = HEAP8[$18>>0]|0;
   $20 = $19 << 24 >> 24;
   $21 = ($20|0)==(42);
   if ($21) {
    $22 = $p;
    $23 = ((($22)) + 1|0);
    $24 = HEAP8[$23>>0]|0;
    $25 = $24 << 24 >> 24;
    $26 = ($25|0)==(108);
    if ($26) {
     $27 = $1;
     $28 = (_recvline($27,$b)|0);
     $err = $28;
     break;
    }
   }
   $29 = $p;
   $30 = HEAP8[$29>>0]|0;
   $31 = $30 << 24 >> 24;
   $32 = ($31|0)==(42);
   if ($32) {
    $33 = $p;
    $34 = ((($33)) + 1|0);
    $35 = HEAP8[$34>>0]|0;
    $36 = $35 << 24 >> 24;
    $37 = ($36|0)==(97);
    if ($37) {
     $38 = $1;
     $39 = (_recvall($38,$b)|0);
     $err = $39;
     break;
    }
   }
   $40 = $0;
   (_luaL_argerror($40,2,506593)|0);
  }
 } while(0);
 $48 = $err;
 $49 = ($48|0)!=(0);
 _luaL_pushresult($b);
 $50 = $0;
 if ($49) {
  $51 = $1;
  $52 = ((($51)) + 16|0);
  $53 = HEAP32[$52>>2]|0;
  $54 = ((($53)) + 12|0);
  $55 = HEAP32[$54>>2]|0;
  $56 = $1;
  $57 = ((($56)) + 16|0);
  $58 = HEAP32[$57>>2]|0;
  $59 = HEAP32[$58>>2]|0;
  $60 = $err;
  $61 = (FUNCTION_TABLE_iii[$55 & 8191]($59,$60)|0);
  _lua_pushstring($50,$61);
  $62 = $0;
  _lua_pushvalue($62,-2);
  $63 = $0;
  _lua_pushnil($63);
  $64 = $0;
  _lua_replace($64,-4);
  $66 = $0;
  $67 = (_lua_gettop($66)|0);
  $68 = $top;
  $69 = (($67) - ($68))|0;
  STACKTOP = sp;return ($69|0);
 } else {
  _lua_pushnil($50);
  $65 = $0;
  _lua_pushnil($65);
  $66 = $0;
  $67 = (_lua_gettop($66)|0);
  $68 = $top;
  $69 = (($67) - ($68))|0;
  STACKTOP = sp;return ($69|0);
 }
 return (0)|0;
}
function _recvline($buf,$b) {
 $buf = $buf|0;
 $b = $b|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $count = 0, $data = 0, $err = 0, $pos = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $count = sp + 8|0;
 $data = sp;
 $0 = $buf;
 $1 = $b;
 $err = 0;
 while(1) {
  $2 = $err;
  $3 = ($2|0)==(0);
  if (!($3)) {
   label = 15;
   break;
  }
  $4 = $0;
  $5 = (_buffer_get($4,$data,$count)|0);
  $err = $5;
  $pos = 0;
  while(1) {
   $6 = $pos;
   $7 = HEAP32[$count>>2]|0;
   $8 = ($6>>>0)<($7>>>0);
   if ($8) {
    $9 = $pos;
    $10 = HEAP32[$data>>2]|0;
    $11 = (($10) + ($9)|0);
    $12 = HEAP8[$11>>0]|0;
    $13 = $12 << 24 >> 24;
    $14 = ($13|0)!=(10);
    $43 = $14;
   } else {
    $43 = 0;
   }
   $15 = $pos;
   if (!($43)) {
    break;
   }
   $16 = HEAP32[$data>>2]|0;
   $17 = (($16) + ($15)|0);
   $18 = HEAP8[$17>>0]|0;
   $19 = $18 << 24 >> 24;
   $20 = ($19|0)!=(13);
   if ($20) {
    $21 = $1;
    $22 = HEAP32[$21>>2]|0;
    $23 = $1;
    $24 = ((($23)) + 12|0);
    $25 = ((($24)) + 1024|0);
    $26 = ($22>>>0)<($25>>>0);
    if (!($26)) {
     $27 = $1;
     (_luaL_prepbuffer($27)|0);
    }
    $28 = $pos;
    $29 = HEAP32[$data>>2]|0;
    $30 = (($29) + ($28)|0);
    $31 = HEAP8[$30>>0]|0;
    $32 = $1;
    $33 = HEAP32[$32>>2]|0;
    $34 = ((($33)) + 1|0);
    HEAP32[$32>>2] = $34;
    HEAP8[$33>>0] = $31;
   }
   $35 = $pos;
   $36 = (($35) + 1)|0;
   $pos = $36;
  }
  $37 = HEAP32[$count>>2]|0;
  $38 = ($15>>>0)<($37>>>0);
  $39 = $0;
  $40 = $pos;
  if ($38) {
   break;
  }
  _buffer_skip($39,$40);
 }
 if ((label|0) == 15) {
  $42 = $err;
  STACKTOP = sp;return ($42|0);
 }
 $41 = (($40) + 1)|0;
 _buffer_skip($39,$41);
 $42 = $err;
 STACKTOP = sp;return ($42|0);
}
function _buffer_get($buf,$data,$count) {
 $buf = $buf|0;
 $data = $data|0;
 $count = $count|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $err = 0, $got = 0, $io = 0, $tm = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $got = sp;
 $0 = $buf;
 $1 = $data;
 $2 = $count;
 $err = 0;
 $3 = $0;
 $4 = ((($3)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $io = $5;
 $6 = $0;
 $7 = ((($6)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $tm = $8;
 $9 = $0;
 $10 = (_buffer_isempty($9)|0);
 $11 = ($10|0)!=(0);
 if ($11) {
  $12 = $io;
  $13 = ((($12)) + 8|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = $io;
  $16 = HEAP32[$15>>2]|0;
  $17 = $0;
  $18 = ((($17)) + 32|0);
  $19 = $tm;
  $20 = (FUNCTION_TABLE_iiiiii[$14 & 8191]($16,$18,8192,$got,$19)|0);
  $err = $20;
  $21 = $0;
  $22 = ((($21)) + 24|0);
  HEAP32[$22>>2] = 0;
  $23 = HEAP32[$got>>2]|0;
  $24 = $0;
  $25 = ((($24)) + 28|0);
  HEAP32[$25>>2] = $23;
 }
 $26 = $0;
 $27 = ((($26)) + 28|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = $0;
 $30 = ((($29)) + 24|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = (($28) - ($31))|0;
 $33 = $2;
 HEAP32[$33>>2] = $32;
 $34 = $0;
 $35 = ((($34)) + 32|0);
 $36 = $0;
 $37 = ((($36)) + 24|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = (($35) + ($38)|0);
 $40 = $1;
 HEAP32[$40>>2] = $39;
 $41 = $err;
 STACKTOP = sp;return ($41|0);
}
function _buffer_isempty($buf) {
 $buf = $buf|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $buf;
 $1 = $0;
 $2 = ((($1)) + 24|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $0;
 $5 = ((($4)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($3>>>0)>=($6>>>0);
 $8 = $7&1;
 STACKTOP = sp;return ($8|0);
}
function _buffer_skip($buf,$count) {
 $buf = $buf|0;
 $count = $count|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $buf;
 $1 = $count;
 $2 = $1;
 $3 = $0;
 $4 = ((($3)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($5) + ($2))|0;
 HEAP32[$4>>2] = $6;
 $7 = $1;
 $8 = $0;
 $9 = ((($8)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + ($7))|0;
 HEAP32[$9>>2] = $11;
 $12 = $0;
 $13 = (_buffer_isempty($12)|0);
 $14 = ($13|0)!=(0);
 if (!($14)) {
  STACKTOP = sp;return;
 }
 $15 = $0;
 $16 = ((($15)) + 28|0);
 HEAP32[$16>>2] = 0;
 $17 = $0;
 $18 = ((($17)) + 24|0);
 HEAP32[$18>>2] = 0;
 STACKTOP = sp;return;
}
function _recvall($buf,$b) {
 $buf = $buf|0;
 $b = $b|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $count = 0, $data = 0, $err = 0, $total = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $data = sp + 4|0;
 $count = sp;
 $1 = $buf;
 $2 = $b;
 $err = 0;
 $total = 0;
 while(1) {
  $3 = $err;
  $4 = ($3|0)==(0);
  if (!($4)) {
   break;
  }
  $5 = $1;
  $6 = (_buffer_get($5,$data,$count)|0);
  $err = $6;
  $7 = HEAP32[$count>>2]|0;
  $8 = $total;
  $9 = (($8) + ($7))|0;
  $total = $9;
  $10 = $2;
  $11 = HEAP32[$data>>2]|0;
  $12 = HEAP32[$count>>2]|0;
  _luaL_addlstring($10,$11,$12);
  $13 = $1;
  $14 = HEAP32[$count>>2]|0;
  _buffer_skip($13,$14);
 }
 $15 = $err;
 $16 = ($15|0)==(-2);
 if (!($16)) {
  $19 = $err;
  $0 = $19;
  $20 = $0;
  STACKTOP = sp;return ($20|0);
 }
 $17 = $total;
 $18 = ($17>>>0)>(0);
 if ($18) {
  $0 = 0;
  $20 = $0;
  STACKTOP = sp;return ($20|0);
 } else {
  $0 = -2;
  $20 = $0;
  STACKTOP = sp;return ($20|0);
 }
 return (0)|0;
}
function _recvraw($buf,$wanted,$b) {
 $buf = $buf|0;
 $wanted = $wanted|0;
 $b = $b|0;
 var $$old = 0, $$old1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $count = 0, $data = 0, $err = 0, $or$cond = 0, $total = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $count = sp + 4|0;
 $data = sp;
 $0 = $buf;
 $1 = $wanted;
 $2 = $b;
 $err = 0;
 $total = 0;
 $$old = $err;
 $$old1 = ($$old|0)==(0);
 if (!($$old1)) {
  $28 = $err;
  STACKTOP = sp;return ($28|0);
 }
 while(1) {
  $3 = $0;
  $4 = (_buffer_get($3,$data,$count)|0);
  $err = $4;
  $5 = HEAP32[$count>>2]|0;
  $6 = $1;
  $7 = $total;
  $8 = (($6) - ($7))|0;
  $9 = ($5>>>0)<($8>>>0);
  if ($9) {
   $10 = HEAP32[$count>>2]|0;
   $14 = $10;
  } else {
   $11 = $1;
   $12 = $total;
   $13 = (($11) - ($12))|0;
   $14 = $13;
  }
  HEAP32[$count>>2] = $14;
  $15 = $2;
  $16 = HEAP32[$data>>2]|0;
  $17 = HEAP32[$count>>2]|0;
  _luaL_addlstring($15,$16,$17);
  $18 = $0;
  $19 = HEAP32[$count>>2]|0;
  _buffer_skip($18,$19);
  $20 = HEAP32[$count>>2]|0;
  $21 = $total;
  $22 = (($21) + ($20))|0;
  $total = $22;
  $23 = $total;
  $24 = $1;
  $25 = ($23>>>0)<($24>>>0);
  $26 = $err;
  $27 = ($26|0)==(0);
  $or$cond = $25 & $27;
  if (!($or$cond)) {
   break;
  }
 }
 $28 = $err;
 STACKTOP = sp;return ($28|0);
}
function _except_open($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 _luax_register($1,0,34996);
 STACKTOP = sp;return 0;
}
function _global_newtry($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 _lua_settop($1,1);
 $2 = $0;
 $3 = (_lua_type($2,1)|0);
 $4 = ($3|0)==(0);
 if ($4) {
  $5 = $0;
  _lua_pushcclosure($5,4247,0);
 }
 $6 = $0;
 _lua_pushcclosure($6,4248,1);
 STACKTOP = sp;return 1;
}
function _do_nothing($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 STACKTOP = sp;return 0;
}
function _finalize($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $L;
 $2 = $1;
 $3 = (_lua_toboolean($2,1)|0);
 $4 = ($3|0)!=(0);
 $5 = $1;
 if ($4) {
  $10 = (_lua_gettop($5)|0);
  $0 = $10;
  $11 = $0;
  STACKTOP = sp;return ($11|0);
 } else {
  _lua_pushvalue($5,-10003);
  $6 = $1;
  (_lua_pcall($6,0,0,0)|0);
  $7 = $1;
  _lua_settop($7,2);
  $8 = $1;
  _wrap($8);
  $9 = $1;
  (_lua_error($9)|0);
  $0 = 0;
  $11 = $0;
  STACKTOP = sp;return ($11|0);
 }
 return (0)|0;
}
function _wrap($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 _lua_createtable($1,0,0);
 $2 = $0;
 _lua_pushnumber($2,1.0);
 $3 = $0;
 _lua_pushvalue($3,-3);
 $4 = $0;
 _lua_settable($4,-3);
 $5 = $0;
 _lua_insert($5,-2);
 $6 = $0;
 _lua_settop($6,-2);
 STACKTOP = sp;return;
}
function _global_protect($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 _lua_pushcclosure($1,4249,1);
 STACKTOP = sp;return 1;
}
function _protected_($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $L;
 $2 = $1;
 _lua_pushvalue($2,-10003);
 $3 = $1;
 _lua_insert($3,1);
 $4 = $1;
 $5 = $1;
 $6 = (_lua_gettop($5)|0);
 $7 = (($6) - 1)|0;
 $8 = (_lua_pcall($4,$7,-1,0)|0);
 $9 = ($8|0)!=(0);
 $10 = $1;
 do {
  if ($9) {
   $11 = (_unwrap($10)|0);
   $12 = ($11|0)!=(0);
   if ($12) {
    $0 = 2;
    break;
   } else {
    $13 = $1;
    (_lua_error($13)|0);
    $0 = 0;
    break;
   }
  } else {
   $14 = (_lua_gettop($10)|0);
   $0 = $14;
  }
 } while(0);
 $15 = $0;
 STACKTOP = sp;return ($15|0);
}
function _unwrap($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $L;
 $2 = $1;
 $3 = (_lua_type($2,-1)|0);
 $4 = ($3|0)==(5);
 if ($4) {
  $5 = $1;
  _lua_pushnumber($5,1.0);
  $6 = $1;
  _lua_gettable($6,-2);
  $7 = $1;
  _lua_pushnil($7);
  $8 = $1;
  _lua_insert($8,-2);
  $0 = 1;
  $9 = $0;
  STACKTOP = sp;return ($9|0);
 } else {
  $0 = 0;
  $9 = $0;
  STACKTOP = sp;return ($9|0);
 }
 return (0)|0;
}
function _inet_open($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $0;
 _lua_pushstring($1,506632);
 $2 = $0;
 _lua_createtable($2,0,0);
 $3 = $0;
 _luax_register($3,0,35020);
 $4 = $0;
 _lua_settable($4,-3);
 STACKTOP = sp;return 0;
}
function _inet_global_toip($L) {
 $L = $L|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $address = 0, $err = 0, $hp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 20|0;
 $hp = sp + 4|0;
 $1 = $L;
 $2 = $1;
 $3 = (_luaL_checklstring($2,1,0)|0);
 $address = $3;
 HEAP32[$hp>>2] = 0;
 $4 = $address;
 $5 = (_inet_gethost($4,$hp)|0);
 $err = $5;
 $6 = $err;
 $7 = ($6|0)!=(0);
 $8 = $1;
 if ($7) {
  _lua_pushnil($8);
  $9 = $1;
  $10 = $err;
  $11 = (_socket_hoststrerror($10)|0);
  _lua_pushstring($9,$11);
  $0 = 2;
  $19 = $0;
  STACKTOP = sp;return ($19|0);
 } else {
  $12 = HEAP32[$hp>>2]|0;
  $13 = ((($12)) + 16|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = HEAP32[$14>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$15>>2]|0;
  $16 = (_inet_ntoa($$byval_copy)|0);
  _lua_pushstring($8,$16);
  $17 = $1;
  $18 = HEAP32[$hp>>2]|0;
  _inet_pushresolved($17,$18);
  $0 = 2;
  $19 = $0;
  STACKTOP = sp;return ($19|0);
 }
 return (0)|0;
}
function _inet_gethost($address,$hp) {
 $address = $address|0;
 $hp = $hp|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $addr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $addr = sp;
 $1 = $address;
 $2 = $hp;
 $3 = $1;
 $4 = (_inet_aton($3,$addr)|0);
 $5 = ($4|0)!=(0);
 if ($5) {
  $6 = $2;
  $7 = (_socket_gethostbyaddr($addr,4,$6)|0);
  $0 = $7;
  $11 = $0;
  STACKTOP = sp;return ($11|0);
 } else {
  $8 = $1;
  $9 = $2;
  $10 = (_socket_gethostbyname($8,$9)|0);
  $0 = $10;
  $11 = $0;
  STACKTOP = sp;return ($11|0);
 }
 return (0)|0;
}
function _inet_pushresolved($L,$hp) {
 $L = $L|0;
 $hp = $hp|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0.0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0.0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $addr = 0;
 var $alias = 0, $i = 0, $resolved = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 24|0;
 $0 = $L;
 $1 = $hp;
 $2 = $0;
 _lua_createtable($2,0,0);
 $3 = $0;
 $4 = (_lua_gettop($3)|0);
 $resolved = $4;
 $5 = $0;
 _lua_pushstring($5,610375);
 $6 = $0;
 $7 = $1;
 $8 = HEAP32[$7>>2]|0;
 _lua_pushstring($6,$8);
 $9 = $0;
 $10 = $resolved;
 _lua_settable($9,$10);
 $11 = $0;
 _lua_pushstring($11,506641);
 $12 = $0;
 _lua_pushstring($12,506644);
 $i = 1;
 $13 = $1;
 $14 = ((($13)) + 4|0);
 $15 = HEAP32[$14>>2]|0;
 $alias = $15;
 $16 = $0;
 _lua_createtable($16,0,0);
 $17 = $alias;
 $18 = ($17|0)!=(0|0);
 L1: do {
  if ($18) {
   while(1) {
    $19 = $alias;
    $20 = HEAP32[$19>>2]|0;
    $21 = ($20|0)!=(0|0);
    if (!($21)) {
     break L1;
    }
    $22 = $0;
    $23 = $i;
    $24 = (+($23|0));
    _lua_pushnumber($22,$24);
    $25 = $0;
    $26 = $alias;
    $27 = HEAP32[$26>>2]|0;
    _lua_pushstring($25,$27);
    $28 = $0;
    _lua_settable($28,-3);
    $29 = $i;
    $30 = (($29) + 1)|0;
    $i = $30;
    $31 = $alias;
    $32 = ((($31)) + 4|0);
    $alias = $32;
   }
  }
 } while(0);
 $33 = $0;
 $34 = $resolved;
 _lua_settable($33,$34);
 $i = 1;
 $35 = $0;
 _lua_createtable($35,0,0);
 $36 = $1;
 $37 = ((($36)) + 16|0);
 $38 = HEAP32[$37>>2]|0;
 $addr = $38;
 $39 = $addr;
 $40 = ($39|0)!=(0|0);
 if (!($40)) {
  $56 = $0;
  $57 = $resolved;
  _lua_settable($56,$57);
  STACKTOP = sp;return;
 }
 while(1) {
  $41 = $addr;
  $42 = HEAP32[$41>>2]|0;
  $43 = ($42|0)!=(0|0);
  if (!($43)) {
   break;
  }
  $44 = $0;
  $45 = $i;
  $46 = (+($45|0));
  _lua_pushnumber($44,$46);
  $47 = $0;
  $48 = $addr;
  $49 = HEAP32[$48>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$49>>2]|0;
  $50 = (_inet_ntoa($$byval_copy)|0);
  _lua_pushstring($47,$50);
  $51 = $0;
  _lua_settable($51,-3);
  $52 = $i;
  $53 = (($52) + 1)|0;
  $i = $53;
  $54 = $addr;
  $55 = ((($54)) + 4|0);
  $addr = $55;
 }
 $56 = $0;
 $57 = $resolved;
 _lua_settable($56,$57);
 STACKTOP = sp;return;
}
function _inet_global_tohostname($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $address = 0, $err = 0, $hp = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $hp = sp + 4|0;
 $1 = $L;
 $2 = $1;
 $3 = (_luaL_checklstring($2,1,0)|0);
 $address = $3;
 HEAP32[$hp>>2] = 0;
 $4 = $address;
 $5 = (_inet_gethost($4,$hp)|0);
 $err = $5;
 $6 = $err;
 $7 = ($6|0)!=(0);
 $8 = $1;
 if ($7) {
  _lua_pushnil($8);
  $9 = $1;
  $10 = $err;
  $11 = (_socket_hoststrerror($10)|0);
  _lua_pushstring($9,$11);
  $0 = 2;
  $16 = $0;
  STACKTOP = sp;return ($16|0);
 } else {
  $12 = HEAP32[$hp>>2]|0;
  $13 = HEAP32[$12>>2]|0;
  _lua_pushstring($8,$13);
  $14 = $1;
  $15 = HEAP32[$hp>>2]|0;
  _inet_pushresolved($14,$15);
  $0 = 2;
  $16 = $0;
  STACKTOP = sp;return ($16|0);
 }
 return (0)|0;
}
function _inet_global_gethostname($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $name = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 272|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp + 8|0;
 $1 = $L;
 $2 = ((($name)) + 256|0);
 HEAP8[$2>>0] = 0;
 $3 = (_gethostname($name,256)|0);
 $4 = ($3|0)<(0);
 $5 = $1;
 if ($4) {
  _lua_pushnil($5);
  $6 = $1;
  _lua_pushstring($6,506673);
  $0 = 2;
  $7 = $0;
  STACKTOP = sp;return ($7|0);
 } else {
  _lua_pushstring($5,$name);
  $0 = 1;
  $7 = $0;
  STACKTOP = sp;return ($7|0);
 }
 return (0)|0;
}
function _inet_meth_getpeername($L,$ps) {
 $L = $L|0;
 $ps = $ps|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $peer = 0, $peer_len = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 32|0;
 $peer = sp + 8|0;
 $peer_len = sp;
 $0 = $L;
 $1 = $ps;
 HEAP32[$peer_len>>2] = 16;
 $2 = $1;
 $3 = HEAP32[$2>>2]|0;
 $4 = (_getpeername($3,$peer,$peer_len)|0);
 $5 = ($4|0)<(0);
 $6 = $0;
 if ($5) {
  _lua_pushnil($6);
  $7 = $0;
  _lua_pushstring($7,506692);
  STACKTOP = sp;return 2;
 } else {
  $8 = ((($peer)) + 4|0);
  ;HEAP32[$$byval_copy>>2]=HEAP32[$8>>2]|0;
  $9 = (_inet_ntoa($$byval_copy)|0);
  _lua_pushstring($6,$9);
  $10 = $0;
  $11 = ((($peer)) + 2|0);
  $12 = HEAP16[$11>>1]|0;
  $13 = (_ntohs($12)|0);
  $14 = (+($13&65535));
  _lua_pushnumber($10,$14);
  STACKTOP = sp;return 2;
 }
 return (0)|0;
}
function _inet_meth_getsockname($L,$ps) {
 $L = $L|0;
 $ps = $ps|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $local = 0, $local_len = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 32|0;
 $local = sp + 8|0;
 $local_len = sp;
 $0 = $L;
 $1 = $ps;
 HEAP32[$local_len>>2] = 16;
 $2 = $1;
 $3 = HEAP32[$2>>2]|0;
 $4 = (_getsockname($3,$local,$local_len)|0);
 $5 = ($4|0)<(0);
 $6 = $0;
 if ($5) {
  _lua_pushnil($6);
  $7 = $0;
  _lua_pushstring($7,506711);
  STACKTOP = sp;return 2;
 } else {
  $8 = ((($local)) + 4|0);
  ;HEAP32[$$byval_copy>>2]=HEAP32[$8>>2]|0;
  $9 = (_inet_ntoa($$byval_copy)|0);
  _lua_pushstring($6,$9);
  $10 = $0;
  $11 = ((($local)) + 2|0);
  $12 = HEAP16[$11>>1]|0;
  $13 = (_ntohs($12)|0);
  $14 = (+($13&65535));
  _lua_pushnumber($10,$14);
  STACKTOP = sp;return 2;
 }
 return (0)|0;
}
function _inet_trycreate($ps,$type) {
 $ps = $ps|0;
 $type = $type|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $ps;
 $1 = $type;
 $2 = $0;
 $3 = $1;
 $4 = (_socket_create($2,2,$3,0)|0);
 $5 = (_socket_strerror($4)|0);
 STACKTOP = sp;return ($5|0);
}
function _inet_tryconnect($ps,$address,$port,$tm) {
 $ps = $ps|0;
 $address = $address|0;
 $port = $port|0;
 $tm = $tm|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $addr = 0, $err = 0, $hp = 0, $remote = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $remote = sp + 16|0;
 $hp = sp + 4|0;
 $1 = $ps;
 $2 = $address;
 $3 = $port;
 $4 = $tm;
 ;HEAP32[$remote>>2]=0|0;HEAP32[$remote+4>>2]=0|0;HEAP32[$remote+8>>2]=0|0;HEAP32[$remote+12>>2]=0|0;
 HEAP16[$remote>>1] = 2;
 $5 = $3;
 $6 = (_htons($5)|0);
 $7 = ((($remote)) + 2|0);
 HEAP16[$7>>1] = $6;
 $8 = $2;
 $9 = (_strcmp($8,506969)|0);
 $10 = ($9|0)!=(0);
 do {
  if ($10) {
   $11 = $2;
   $12 = ((($remote)) + 4|0);
   $13 = (_inet_aton($11,$12)|0);
   $14 = ($13|0)!=(0);
   if (!($14)) {
    HEAP32[$hp>>2] = 0;
    $15 = $2;
    $16 = (_socket_gethostbyname($15,$hp)|0);
    $err = $16;
    $17 = $err;
    $18 = ($17|0)!=(0);
    if (!($18)) {
     $21 = HEAP32[$hp>>2]|0;
     $22 = ((($21)) + 16|0);
     $23 = HEAP32[$22>>2]|0;
     $addr = $23;
     $24 = ((($remote)) + 4|0);
     $25 = $addr;
     $26 = HEAP32[$25>>2]|0;
     ;HEAP32[$24>>2]=HEAP32[$26>>2]|0;
     break;
    }
    $19 = $err;
    $20 = (_socket_hoststrerror($19)|0);
    $0 = $20;
    $32 = $0;
    STACKTOP = sp;return ($32|0);
   }
  } else {
   HEAP16[$remote>>1] = 0;
  }
 } while(0);
 $27 = $1;
 $28 = $4;
 $29 = (_socket_connect($27,$remote,16,$28)|0);
 $err = $29;
 $30 = $err;
 $31 = (_socket_strerror($30)|0);
 $0 = $31;
 $32 = $0;
 STACKTOP = sp;return ($32|0);
}
function _inet_trybind($ps,$address,$port) {
 $ps = $ps|0;
 $address = $address|0;
 $port = $port|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $addr = 0, $err = 0, $hp = 0, $local = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $local = sp + 16|0;
 $hp = sp + 4|0;
 $1 = $ps;
 $2 = $address;
 $3 = $port;
 ;HEAP32[$local>>2]=0|0;HEAP32[$local+4>>2]=0|0;HEAP32[$local+8>>2]=0|0;HEAP32[$local+12>>2]=0|0;
 $4 = (_htonl(0)|0);
 $5 = ((($local)) + 4|0);
 HEAP32[$5>>2] = $4;
 $6 = $3;
 $7 = (_htons($6)|0);
 $8 = ((($local)) + 2|0);
 HEAP16[$8>>1] = $7;
 HEAP16[$local>>1] = 2;
 $9 = $2;
 $10 = (_strcmp($9,506969)|0);
 $11 = ($10|0)!=(0);
 do {
  if ($11) {
   $12 = $2;
   $13 = ((($local)) + 4|0);
   $14 = (_inet_aton($12,$13)|0);
   $15 = ($14|0)!=(0);
   if (!($15)) {
    HEAP32[$hp>>2] = 0;
    $16 = $2;
    $17 = (_socket_gethostbyname($16,$hp)|0);
    $err = $17;
    $18 = $err;
    $19 = ($18|0)!=(0);
    if (!($19)) {
     $22 = HEAP32[$hp>>2]|0;
     $23 = ((($22)) + 16|0);
     $24 = HEAP32[$23>>2]|0;
     $addr = $24;
     $25 = ((($local)) + 4|0);
     $26 = $addr;
     $27 = HEAP32[$26>>2]|0;
     ;HEAP32[$25>>2]=HEAP32[$27>>2]|0;
     break;
    }
    $20 = $err;
    $21 = (_socket_hoststrerror($20)|0);
    $0 = $21;
    $35 = $0;
    STACKTOP = sp;return ($35|0);
   }
  }
 } while(0);
 $28 = $1;
 $29 = (_socket_bind($28,$local,16)|0);
 $err = $29;
 $30 = $err;
 $31 = ($30|0)!=(0);
 if ($31) {
  $32 = $1;
  _socket_destroy($32);
 }
 $33 = $err;
 $34 = (_socket_strerror($33)|0);
 $0 = $34;
 $35 = $0;
 STACKTOP = sp;return ($35|0);
}
function _io_init($io,$send,$recv,$error,$ctx) {
 $io = $io|0;
 $send = $send|0;
 $recv = $recv|0;
 $error = $error|0;
 $ctx = $ctx|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $io;
 $1 = $send;
 $2 = $recv;
 $3 = $error;
 $4 = $ctx;
 $5 = $1;
 $6 = $0;
 $7 = ((($6)) + 4|0);
 HEAP32[$7>>2] = $5;
 $8 = $2;
 $9 = $0;
 $10 = ((($9)) + 8|0);
 HEAP32[$10>>2] = $8;
 $11 = $3;
 $12 = $0;
 $13 = ((($12)) + 12|0);
 HEAP32[$13>>2] = $11;
 $14 = $4;
 $15 = $0;
 HEAP32[$15>>2] = $14;
 STACKTOP = sp;return;
}
function _io_strerror($err) {
 $err = $err|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $err;
 $2 = $1;
 switch ($2|0) {
 case 0:  {
  $0 = 0;
  break;
 }
 case -2:  {
  $0 = 506730;
  break;
 }
 case -1:  {
  $0 = 506810;
  break;
 }
 default: {
  $0 = 517474;
 }
 }
 $3 = $0;
 STACKTOP = sp;return ($3|0);
}
function _opt_meth_setoption($L,$opt,$ps) {
 $L = $L|0;
 $opt = $opt|0;
 $ps = $ps|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $msg = 0, $name = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $msg = sp + 20|0;
 $0 = $L;
 $1 = $opt;
 $2 = $ps;
 $3 = $0;
 $4 = (_luaL_checklstring($3,2,0)|0);
 $name = $4;
 while(1) {
  $5 = $1;
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6|0)!=(0|0);
  if ($7) {
   $8 = $name;
   $9 = $1;
   $10 = HEAP32[$9>>2]|0;
   $11 = (_strcmp($8,$10)|0);
   $12 = ($11|0)!=(0);
   $26 = $12;
  } else {
   $26 = 0;
  }
  $13 = $1;
  if (!($26)) {
   break;
  }
  $14 = ((($13)) + 8|0);
  $1 = $14;
 }
 $15 = ((($13)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ($16|0)!=(0|0);
 if (!($17)) {
  $18 = $name;
  HEAP32[$vararg_buffer>>2] = $18;
  (_sprintf($msg,506737,$vararg_buffer)|0);
  $19 = $0;
  (_luaL_argerror($19,2,$msg)|0);
 }
 $20 = $1;
 $21 = ((($20)) + 4|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = $0;
 $24 = $2;
 $25 = (FUNCTION_TABLE_iii[$22 & 8191]($23,$24)|0);
 STACKTOP = sp;return ($25|0);
}
function _opt_reuseaddr($L,$ps) {
 $L = $L|0;
 $ps = $ps|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $ps;
 $2 = $0;
 $3 = $1;
 $4 = (_opt_setboolean($2,$3,1,2)|0);
 STACKTOP = sp;return ($4|0);
}
function _opt_setboolean($L,$ps,$level,$name) {
 $L = $L|0;
 $ps = $ps|0;
 $level = $level|0;
 $name = $name|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $val = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $val = sp;
 $0 = $L;
 $1 = $ps;
 $2 = $level;
 $3 = $name;
 $4 = $0;
 $5 = (_auxiliar_checkboolean($4,3)|0);
 HEAP32[$val>>2] = $5;
 $6 = $0;
 $7 = $1;
 $8 = $2;
 $9 = $3;
 $10 = (_opt_set($6,$7,$8,$9,$val,4)|0);
 STACKTOP = sp;return ($10|0);
}
function _opt_set($L,$ps,$level,$name,$val,$len) {
 $L = $L|0;
 $ps = $ps|0;
 $level = $level|0;
 $name = $name|0;
 $val = $val|0;
 $len = $len|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $L;
 $2 = $ps;
 $3 = $level;
 $4 = $name;
 $5 = $val;
 $6 = $len;
 $7 = $2;
 $8 = HEAP32[$7>>2]|0;
 $9 = $3;
 $10 = $4;
 $11 = $5;
 $12 = $6;
 $13 = (_setsockopt($8,$9,$10,$11,$12)|0);
 $14 = ($13|0)<(0);
 $15 = $1;
 if ($14) {
  _lua_pushnil($15);
  $16 = $1;
  _lua_pushstring($16,506764);
  $0 = 2;
  $17 = $0;
  STACKTOP = sp;return ($17|0);
 } else {
  _lua_pushnumber($15,1.0);
  $0 = 1;
  $17 = $0;
  STACKTOP = sp;return ($17|0);
 }
 return (0)|0;
}
function _opt_tcp_nodelay($L,$ps) {
 $L = $L|0;
 $ps = $ps|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $ps;
 $2 = $0;
 $3 = $1;
 $4 = (_opt_setboolean($2,$3,6,1)|0);
 STACKTOP = sp;return ($4|0);
}
function _opt_keepalive($L,$ps) {
 $L = $L|0;
 $ps = $ps|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $ps;
 $2 = $0;
 $3 = $1;
 $4 = (_opt_setboolean($2,$3,1,9)|0);
 STACKTOP = sp;return ($4|0);
}
function _opt_dontroute($L,$ps) {
 $L = $L|0;
 $ps = $ps|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $ps;
 $2 = $0;
 $3 = $1;
 $4 = (_opt_setboolean($2,$3,1,5)|0);
 STACKTOP = sp;return ($4|0);
}
function _opt_broadcast($L,$ps) {
 $L = $L|0;
 $ps = $ps|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $ps;
 $2 = $0;
 $3 = $1;
 $4 = (_opt_setboolean($2,$3,1,6)|0);
 STACKTOP = sp;return ($4|0);
}
function _opt_ip_multicast_loop($L,$ps) {
 $L = $L|0;
 $ps = $ps|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $ps;
 $2 = $0;
 $3 = $1;
 $4 = (_opt_setboolean($2,$3,0,34)|0);
 STACKTOP = sp;return ($4|0);
}
function _opt_linger($L,$ps) {
 $L = $L|0;
 $ps = $ps|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0.0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $li = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $li = sp;
 $0 = $L;
 $1 = $ps;
 $2 = $0;
 $3 = (_lua_type($2,3)|0);
 $4 = ($3|0)==(5);
 if (!($4)) {
  $5 = $0;
  $6 = $0;
  $7 = (_lua_typename($6,5)|0);
  (_luaL_typerror($5,3,$7)|0);
 }
 $8 = $0;
 _lua_pushstring($8,526513);
 $9 = $0;
 _lua_gettable($9,3);
 $10 = $0;
 $11 = (_lua_type($10,-1)|0);
 $12 = ($11|0)==(1);
 if (!($12)) {
  $13 = $0;
  (_luaL_argerror($13,3,506782)|0);
 }
 $14 = $0;
 $15 = (_lua_toboolean($14,-1)|0);
 $16 = $15&65535;
 $17 = $16&65535;
 HEAP32[$li>>2] = $17;
 $18 = $0;
 _lua_pushstring($18,506810);
 $19 = $0;
 _lua_gettable($19,3);
 $20 = $0;
 $21 = (_lua_isnumber($20,-1)|0);
 $22 = ($21|0)!=(0);
 if (!($22)) {
  $23 = $0;
  (_luaL_argerror($23,3,506818)|0);
 }
 $24 = $0;
 $25 = (+_lua_tonumber($24,-1));
 $26 = (~~(($25))&65535);
 $27 = $26&65535;
 $28 = ((($li)) + 4|0);
 HEAP32[$28>>2] = $27;
 $29 = $0;
 $30 = $1;
 $31 = (_opt_set($29,$30,1,13,$li,8)|0);
 STACKTOP = sp;return ($31|0);
}
function _opt_ip_multicast_ttl($L,$ps) {
 $L = $L|0;
 $ps = $ps|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $val = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $val = sp;
 $0 = $L;
 $1 = $ps;
 $2 = $0;
 $3 = (+_luaL_checknumber($2,3));
 $4 = (~~(($3)));
 HEAP32[$val>>2] = $4;
 $5 = $0;
 $6 = $1;
 $7 = (_opt_set($5,$6,1,13,$val,4)|0);
 STACKTOP = sp;return ($7|0);
}
function _opt_ip_add_membership($L,$ps) {
 $L = $L|0;
 $ps = $ps|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $ps;
 $2 = $0;
 $3 = $1;
 $4 = (_opt_setmembership($2,$3,0,35)|0);
 STACKTOP = sp;return ($4|0);
}
function _opt_setmembership($L,$ps,$level,$name) {
 $L = $L|0;
 $ps = $ps|0;
 $level = $level|0;
 $name = $name|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $val = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $val = sp;
 $0 = $L;
 $1 = $ps;
 $2 = $level;
 $3 = $name;
 $4 = $0;
 $5 = (_lua_type($4,3)|0);
 $6 = ($5|0)==(5);
 if (!($6)) {
  $7 = $0;
  $8 = $0;
  $9 = (_lua_typename($8,5)|0);
  (_luaL_typerror($7,3,$9)|0);
 }
 $10 = $0;
 _lua_pushstring($10,506850);
 $11 = $0;
 _lua_gettable($11,3);
 $12 = $0;
 $13 = (_lua_isstring($12,-1)|0);
 $14 = ($13|0)!=(0);
 if (!($14)) {
  $15 = $0;
  (_luaL_argerror($15,3,506860)|0);
 }
 $16 = $0;
 $17 = (_lua_tolstring($16,-1,0)|0);
 $18 = (_inet_aton($17,$val)|0);
 $19 = ($18|0)!=(0);
 if (!($19)) {
  $20 = $0;
  (_luaL_argerror($20,3,506894)|0);
 }
 $21 = $0;
 _lua_pushstring($21,506925);
 $22 = $0;
 _lua_gettable($22,3);
 $23 = $0;
 $24 = (_lua_isstring($23,-1)|0);
 $25 = ($24|0)!=(0);
 if (!($25)) {
  $26 = $0;
  (_luaL_argerror($26,3,506935)|0);
 }
 $27 = (_htonl(0)|0);
 $28 = ((($val)) + 4|0);
 HEAP32[$28>>2] = $27;
 $29 = $0;
 $30 = (_lua_tolstring($29,-1,0)|0);
 $31 = (_strcmp($30,506969)|0);
 $32 = ($31|0)!=(0);
 if ($32) {
  $33 = $0;
  $34 = (_lua_tolstring($33,-1,0)|0);
  $35 = ((($val)) + 4|0);
  $36 = (_inet_aton($34,$35)|0);
  $37 = ($36|0)!=(0);
  if (!($37)) {
   $38 = $0;
   (_luaL_argerror($38,3,506971)|0);
  }
 }
 $39 = $0;
 $40 = $1;
 $41 = $2;
 $42 = $3;
 $43 = (_opt_set($39,$40,$41,$42,$val,8)|0);
 STACKTOP = sp;return ($43|0);
}
function _opt_ip_drop_membersip($L,$ps) {
 $L = $L|0;
 $ps = $ps|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $L;
 $1 = $ps;
 $2 = $0;
 $3 = $1;
 $4 = (_opt_setmembership($2,$3,0,36)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN4love10b64_decodeEPKciRi($src,$slen,$size) {
 $src = $src|0;
 $slen = $slen|0;
 $size = $size|0;
 var $$ = 0, $$lcssa38 = 0, $$lcssa53 = 0, $$off = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $7 = 0, $8 = 0, $9 = 0, $d$0$ph$ph40 = 0, $i$011 = 0, $in = 0, $len$012 = 0, $len$1 = 0, $len$1$lcssa = 0, $or$cond = 0, $out = 0;
 var $pos$0$ph$ph41 = 0, $pos$113 = 0, $pos$2$lcssa28 = 0, $pos$2$lcssa28$lcssa = 0, $pos$2$lcssa29 = 0, $pos$26 = 0, $v$0$be = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $in = sp;
 $out = sp + 4|0;
 $0 = (($slen|0) / 4)&-1; //@line 37 "common/b64.cpp"
 $1 = ($0*3)|0; //@line 37 "common/b64.cpp"
 HEAP32[$size>>2] = $1; //@line 37 "common/b64.cpp"
 $2 = ($slen|0)<(-3); //@line 39 "common/b64.cpp"
 $3 = $2 ? -1 : $1; //@line 39 "common/b64.cpp"
 $4 = (__Znaj($3)|0); //@line 39 "common/b64.cpp"
 HEAP32[$in>>2] = 0; //@line 42 "common/b64.cpp"
 $5 = ((($out)) + 1|0); //@line 31 "common/b64.cpp"
 $6 = ((($out)) + 2|0); //@line 32 "common/b64.cpp"
 $7 = ($slen|0)<(0); //@line 45 "common/b64.cpp"
 if ($7) {
  STACKTOP = sp;return ($4|0); //@line 77 "common/b64.cpp"
 } else {
  $d$0$ph$ph40 = $4;$pos$0$ph$ph41 = 0;
 }
 L3: while(1) {
  $i$011 = 0;$len$012 = 0;$pos$113 = $pos$0$ph$ph41;
  while(1) {
   $9 = ($pos$113|0)>($slen|0); //@line 51 "common/b64.cpp"
   if ($9) {
    $pos$2$lcssa29 = $pos$113;
    label = 13;
   } else {
    $pos$26 = $pos$113;
    while(1) {
     $10 = (($pos$26) + 1)|0; //@line 53 "common/b64.cpp"
     $11 = (($src) + ($pos$26)|0); //@line 53 "common/b64.cpp"
     $12 = HEAP8[$11>>0]|0; //@line 53 "common/b64.cpp"
     $$off = (($12) + -43)<<24>>24; //@line 54 "common/b64.cpp"
     $13 = ($$off&255)>(79); //@line 54 "common/b64.cpp"
     if ($13) {
      $20 = 0;
     } else {
      $14 = $12 << 24 >> 24; //@line 54 "common/b64.cpp"
      $15 = (($14) + -43)|0; //@line 54 "common/b64.cpp"
      $16 = (507002 + ($15)|0); //@line 54 "common/b64.cpp"
      $17 = HEAP8[$16>>0]|0; //@line 54 "common/b64.cpp"
      $18 = $17 << 24 >> 24; //@line 54 "common/b64.cpp"
      $20 = $18;
     }
     $19 = $20&255; //@line 54 "common/b64.cpp"
     $21 = ($19<<24>>24)==(0); //@line 55 "common/b64.cpp"
     if ($21) {
      $v$0$be = 0;
     } else {
      $22 = ($20|0)==(36); //@line 56 "common/b64.cpp"
      $23 = (($20) + 195)|0; //@line 56 "common/b64.cpp"
      $24 = $23&255; //@line 56 "common/b64.cpp"
      $25 = $22 ? 0 : $24; //@line 56 "common/b64.cpp"
      $v$0$be = $25;
     }
     $26 = ($pos$26|0)<($slen|0); //@line 51 "common/b64.cpp"
     $27 = ($v$0$be<<24>>24)==(0); //@line 51 "common/b64.cpp"
     $$ = $27 & $26; //@line 51 "common/b64.cpp"
     if ($$) {
      $pos$26 = $10;
     } else {
      break;
     }
    }
    if ($26) {
     $28 = (($len$012) + 1)|0; //@line 61 "common/b64.cpp"
     if ($27) {
      $len$1 = $28;$pos$2$lcssa28 = $10;
     } else {
      $29 = $v$0$be&255; //@line 63 "common/b64.cpp"
      $30 = (($29) + 255)|0; //@line 63 "common/b64.cpp"
      $31 = $30&255; //@line 63 "common/b64.cpp"
      $32 = (($in) + ($i$011)|0); //@line 63 "common/b64.cpp"
      HEAP8[$32>>0] = $31; //@line 63 "common/b64.cpp"
      $len$1 = $28;$pos$2$lcssa28 = $10;
     }
    } else {
     $pos$2$lcssa29 = $10;
     label = 13;
    }
   }
   if ((label|0) == 13) {
    label = 0;
    $33 = (($in) + ($i$011)|0); //@line 66 "common/b64.cpp"
    HEAP8[$33>>0] = 0; //@line 66 "common/b64.cpp"
    $len$1 = $len$012;$pos$2$lcssa28 = $pos$2$lcssa29;
   }
   $34 = (($i$011) + 1)|0; //@line 47 "common/b64.cpp"
   $35 = ($i$011|0)>(2); //@line 47 "common/b64.cpp"
   $36 = ($pos$2$lcssa28|0)>($slen|0); //@line 47 "common/b64.cpp"
   $or$cond = $35 | $36; //@line 47 "common/b64.cpp"
   if (!($or$cond)) {
    $i$011 = $34;$len$012 = $len$1;$pos$113 = $pos$2$lcssa28;
    continue;
   }
   $37 = ($len$1|0)==(0); //@line 69 "common/b64.cpp"
   if (!($37)) {
    $39 = HEAP32[$in>>2]|0; //@line 30 "common/b64.cpp"
    $40 = $39 << 2; //@line 30 "common/b64.cpp"
    $41 = $39 >>> 8; //@line 30 "common/b64.cpp"
    $42 = $41 << 24; //@line 30 "common/b64.cpp"
    $43 = $42 >> 24; //@line 30 "common/b64.cpp"
    $44 = $43 >>> 4; //@line 30 "common/b64.cpp"
    $45 = $44 | $40; //@line 30 "common/b64.cpp"
    $46 = $45&255; //@line 30 "common/b64.cpp"
    HEAP8[$out>>0] = $46; //@line 30 "common/b64.cpp"
    $47 = $41 << 4; //@line 31 "common/b64.cpp"
    $48 = $39 << 8; //@line 31 "common/b64.cpp"
    $49 = $48 >> 24; //@line 31 "common/b64.cpp"
    $50 = $49 >>> 2; //@line 31 "common/b64.cpp"
    $51 = $50 | $47; //@line 31 "common/b64.cpp"
    $52 = $51&255; //@line 31 "common/b64.cpp"
    HEAP8[$5>>0] = $52; //@line 31 "common/b64.cpp"
    $53 = $49 << 6; //@line 32 "common/b64.cpp"
    $54 = $39 >>> 24; //@line 32 "common/b64.cpp"
    $55 = $53 | $54; //@line 32 "common/b64.cpp"
    $56 = $55&255; //@line 32 "common/b64.cpp"
    HEAP8[$6>>0] = $56; //@line 32 "common/b64.cpp"
    $57 = ($len$1|0)>(1); //@line 72 "common/b64.cpp"
    if ($57) {
     $$lcssa53 = $46;$len$1$lcssa = $len$1;$pos$2$lcssa28$lcssa = $pos$2$lcssa28;
     break;
    }
   }
   $38 = ($pos$2$lcssa28|0)>($slen|0); //@line 45 "common/b64.cpp"
   if ($38) {
    label = 20;
    break L3;
   } else {
    $i$011 = 0;$len$012 = 0;$pos$113 = $pos$2$lcssa28;
   }
  }
  $58 = (($len$1$lcssa) + -1)|0; //@line 72 "common/b64.cpp"
  $59 = ((($d$0$ph$ph40)) + 1|0); //@line 73 "common/b64.cpp"
  HEAP8[$d$0$ph$ph40>>0] = $$lcssa53; //@line 73 "common/b64.cpp"
  $60 = ($58|0)>(1); //@line 72 "common/b64.cpp"
  if ($60) {
   $61 = 1;$63 = $59;
   while(1) {
    $$phi$trans$insert = (($out) + ($61)|0); //@line 73 "common/b64.cpp"
    $$pre = HEAP8[$$phi$trans$insert>>0]|0; //@line 73 "common/b64.cpp"
    $62 = ((($63)) + 1|0); //@line 73 "common/b64.cpp"
    HEAP8[$63>>0] = $$pre; //@line 73 "common/b64.cpp"
    $64 = (($61) + 1)|0; //@line 72 "common/b64.cpp"
    $65 = ($64|0)<($58|0); //@line 72 "common/b64.cpp"
    if ($65) {
     $61 = $64;$63 = $62;
    } else {
     $$lcssa38 = $62;
     break;
    }
   }
  } else {
   $$lcssa38 = $59;
  }
  $8 = ($pos$2$lcssa28$lcssa|0)>($slen|0); //@line 45 "common/b64.cpp"
  if ($8) {
   label = 20;
   break;
  } else {
   $d$0$ph$ph40 = $$lcssa38;$pos$0$ph$ph41 = $pos$2$lcssa28$lcssa;
  }
 }
 if ((label|0) == 20) {
  STACKTOP = sp;return ($4|0); //@line 77 "common/b64.cpp"
 }
 return (0)|0;
}
function __ZN4love7Matrix4C2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 dest=$this; stop=dest+60|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 98 "common/Matrix.cpp"
 $0 = ((($this)) + 60|0); //@line 99 "common/Matrix.cpp"
 HEAPF32[$0>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 $1 = ((($this)) + 40|0); //@line 99 "common/Matrix.cpp"
 HEAPF32[$1>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 $2 = ((($this)) + 20|0); //@line 99 "common/Matrix.cpp"
 HEAPF32[$2>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 HEAPF32[$this>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 return; //@line 38 "common/Matrix.cpp"
}
function __ZN4love7Matrix411setIdentityEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 dest=$this; stop=dest+60|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 98 "common/Matrix.cpp"
 $0 = ((($this)) + 60|0); //@line 99 "common/Matrix.cpp"
 HEAPF32[$0>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 $1 = ((($this)) + 40|0); //@line 99 "common/Matrix.cpp"
 HEAPF32[$1>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 $2 = ((($this)) + 20|0); //@line 99 "common/Matrix.cpp"
 HEAPF32[$2>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 HEAPF32[$this>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 return; //@line 100 "common/Matrix.cpp"
}
function __ZN4love7Matrix4C2Efffffffff($this,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $0 = 0.0, $1 = 0.0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0.0;
 var $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 dest=$this; stop=dest+60|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 135 "common/Matrix.cpp"
 $0 = (+Math_cos((+$angle))); //@line 136 "common/Matrix.cpp"
 $1 = (+Math_sin((+$angle))); //@line 136 "common/Matrix.cpp"
 $2 = ((($this)) + 60|0); //@line 143 "common/Matrix.cpp"
 HEAPF32[$2>>2] = 1.0; //@line 143 "common/Matrix.cpp"
 $3 = ((($this)) + 40|0); //@line 143 "common/Matrix.cpp"
 HEAPF32[$3>>2] = 1.0; //@line 143 "common/Matrix.cpp"
 $4 = $0 * $sx; //@line 144 "common/Matrix.cpp"
 $5 = $1 * $ky; //@line 144 "common/Matrix.cpp"
 $6 = $5 * $sy; //@line 144 "common/Matrix.cpp"
 $7 = $4 - $6; //@line 144 "common/Matrix.cpp"
 HEAPF32[$this>>2] = $7; //@line 144 "common/Matrix.cpp"
 $8 = $1 * $sx; //@line 145 "common/Matrix.cpp"
 $9 = $0 * $ky; //@line 145 "common/Matrix.cpp"
 $10 = $9 * $sy; //@line 145 "common/Matrix.cpp"
 $11 = $8 + $10; //@line 145 "common/Matrix.cpp"
 $12 = ((($this)) + 4|0); //@line 145 "common/Matrix.cpp"
 HEAPF32[$12>>2] = $11; //@line 145 "common/Matrix.cpp"
 $13 = $0 * $kx; //@line 146 "common/Matrix.cpp"
 $14 = $13 * $sx; //@line 146 "common/Matrix.cpp"
 $15 = $1 * $sy; //@line 146 "common/Matrix.cpp"
 $16 = $14 - $15; //@line 146 "common/Matrix.cpp"
 $17 = ((($this)) + 16|0); //@line 146 "common/Matrix.cpp"
 HEAPF32[$17>>2] = $16; //@line 146 "common/Matrix.cpp"
 $18 = $1 * $kx; //@line 147 "common/Matrix.cpp"
 $19 = $18 * $sx; //@line 147 "common/Matrix.cpp"
 $20 = $0 * $sy; //@line 147 "common/Matrix.cpp"
 $21 = $20 + $19; //@line 147 "common/Matrix.cpp"
 $22 = ((($this)) + 20|0); //@line 147 "common/Matrix.cpp"
 HEAPF32[$22>>2] = $21; //@line 147 "common/Matrix.cpp"
 $23 = $7 * $ox; //@line 148 "common/Matrix.cpp"
 $24 = $x - $23; //@line 148 "common/Matrix.cpp"
 $25 = $16 * $oy; //@line 148 "common/Matrix.cpp"
 $26 = $24 - $25; //@line 148 "common/Matrix.cpp"
 $27 = ((($this)) + 48|0); //@line 148 "common/Matrix.cpp"
 HEAPF32[$27>>2] = $26; //@line 148 "common/Matrix.cpp"
 $28 = $11 * $ox; //@line 149 "common/Matrix.cpp"
 $29 = $y - $28; //@line 149 "common/Matrix.cpp"
 $30 = $21 * $oy; //@line 149 "common/Matrix.cpp"
 $31 = $29 - $30; //@line 149 "common/Matrix.cpp"
 $32 = ((($this)) + 52|0); //@line 149 "common/Matrix.cpp"
 HEAPF32[$32>>2] = $31; //@line 149 "common/Matrix.cpp"
 return; //@line 43 "common/Matrix.cpp"
}
function __ZN4love7Matrix4D2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 47 "common/Matrix.cpp"
}
function __ZNK4love7Matrix4mlERKS0_($agg$result,$this,$m) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $m = $m|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $100 = 0.0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0.0, $110 = 0, $111 = 0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0;
 var $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0.0, $130 = 0, $131 = 0.0, $132 = 0.0, $133 = 0.0;
 var $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0.0, $148 = 0.0, $149 = 0.0, $15 = 0.0, $150 = 0.0, $151 = 0.0;
 var $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0.0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0, $166 = 0, $167 = 0.0, $168 = 0.0, $169 = 0, $17 = 0.0;
 var $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0, $182 = 0.0, $183 = 0.0, $184 = 0.0, $185 = 0.0, $186 = 0.0, $187 = 0.0, $188 = 0.0;
 var $189 = 0, $19 = 0.0, $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0.0, $194 = 0.0, $195 = 0.0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0.0, $201 = 0.0, $202 = 0.0, $203 = 0.0, $204 = 0.0, $21 = 0.0;
 var $22 = 0.0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0, $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0.0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0.0;
 var $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0.0, $50 = 0.0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0, $56 = 0, $57 = 0.0, $58 = 0.0;
 var $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0, $68 = 0.0, $69 = 0.0, $7 = 0.0, $70 = 0.0, $71 = 0, $72 = 0, $73 = 0.0, $74 = 0.0, $75 = 0, $76 = 0.0;
 var $77 = 0.0, $78 = 0.0, $79 = 0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0, $88 = 0.0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0;
 var $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 dest=$agg$result; stop=dest+60|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 98 "common/Matrix.cpp"
 $0 = ((($agg$result)) + 60|0); //@line 99 "common/Matrix.cpp"
 HEAPF32[$0>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 $1 = ((($agg$result)) + 40|0); //@line 99 "common/Matrix.cpp"
 HEAPF32[$1>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 $2 = ((($agg$result)) + 20|0); //@line 99 "common/Matrix.cpp"
 HEAPF32[$2>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 $3 = +HEAPF32[$this>>2]; //@line 62 "common/Matrix.cpp"
 $4 = +HEAPF32[$m>>2]; //@line 62 "common/Matrix.cpp"
 $5 = $3 * $4; //@line 62 "common/Matrix.cpp"
 $6 = ((($this)) + 16|0); //@line 62 "common/Matrix.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 62 "common/Matrix.cpp"
 $8 = ((($m)) + 4|0); //@line 62 "common/Matrix.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 62 "common/Matrix.cpp"
 $10 = $7 * $9; //@line 62 "common/Matrix.cpp"
 $11 = $5 + $10; //@line 62 "common/Matrix.cpp"
 $12 = ((($this)) + 32|0); //@line 62 "common/Matrix.cpp"
 $13 = +HEAPF32[$12>>2]; //@line 62 "common/Matrix.cpp"
 $14 = ((($m)) + 8|0); //@line 62 "common/Matrix.cpp"
 $15 = +HEAPF32[$14>>2]; //@line 62 "common/Matrix.cpp"
 $16 = $13 * $15; //@line 62 "common/Matrix.cpp"
 $17 = $11 + $16; //@line 62 "common/Matrix.cpp"
 $18 = ((($this)) + 48|0); //@line 62 "common/Matrix.cpp"
 $19 = +HEAPF32[$18>>2]; //@line 62 "common/Matrix.cpp"
 $20 = ((($m)) + 12|0); //@line 62 "common/Matrix.cpp"
 $21 = +HEAPF32[$20>>2]; //@line 62 "common/Matrix.cpp"
 $22 = $19 * $21; //@line 62 "common/Matrix.cpp"
 $23 = $17 + $22; //@line 62 "common/Matrix.cpp"
 HEAPF32[$agg$result>>2] = $23; //@line 62 "common/Matrix.cpp"
 $24 = ((($m)) + 16|0); //@line 63 "common/Matrix.cpp"
 $25 = +HEAPF32[$24>>2]; //@line 63 "common/Matrix.cpp"
 $26 = $3 * $25; //@line 63 "common/Matrix.cpp"
 $27 = ((($m)) + 20|0); //@line 63 "common/Matrix.cpp"
 $28 = +HEAPF32[$27>>2]; //@line 63 "common/Matrix.cpp"
 $29 = $7 * $28; //@line 63 "common/Matrix.cpp"
 $30 = $26 + $29; //@line 63 "common/Matrix.cpp"
 $31 = ((($m)) + 24|0); //@line 63 "common/Matrix.cpp"
 $32 = +HEAPF32[$31>>2]; //@line 63 "common/Matrix.cpp"
 $33 = $13 * $32; //@line 63 "common/Matrix.cpp"
 $34 = $30 + $33; //@line 63 "common/Matrix.cpp"
 $35 = ((($m)) + 28|0); //@line 63 "common/Matrix.cpp"
 $36 = +HEAPF32[$35>>2]; //@line 63 "common/Matrix.cpp"
 $37 = $19 * $36; //@line 63 "common/Matrix.cpp"
 $38 = $34 + $37; //@line 63 "common/Matrix.cpp"
 $39 = ((($agg$result)) + 16|0); //@line 63 "common/Matrix.cpp"
 HEAPF32[$39>>2] = $38; //@line 63 "common/Matrix.cpp"
 $40 = ((($m)) + 32|0); //@line 64 "common/Matrix.cpp"
 $41 = +HEAPF32[$40>>2]; //@line 64 "common/Matrix.cpp"
 $42 = $3 * $41; //@line 64 "common/Matrix.cpp"
 $43 = ((($m)) + 36|0); //@line 64 "common/Matrix.cpp"
 $44 = +HEAPF32[$43>>2]; //@line 64 "common/Matrix.cpp"
 $45 = $7 * $44; //@line 64 "common/Matrix.cpp"
 $46 = $42 + $45; //@line 64 "common/Matrix.cpp"
 $47 = ((($m)) + 40|0); //@line 64 "common/Matrix.cpp"
 $48 = +HEAPF32[$47>>2]; //@line 64 "common/Matrix.cpp"
 $49 = $13 * $48; //@line 64 "common/Matrix.cpp"
 $50 = $46 + $49; //@line 64 "common/Matrix.cpp"
 $51 = ((($m)) + 44|0); //@line 64 "common/Matrix.cpp"
 $52 = +HEAPF32[$51>>2]; //@line 64 "common/Matrix.cpp"
 $53 = $19 * $52; //@line 64 "common/Matrix.cpp"
 $54 = $50 + $53; //@line 64 "common/Matrix.cpp"
 $55 = ((($agg$result)) + 32|0); //@line 64 "common/Matrix.cpp"
 HEAPF32[$55>>2] = $54; //@line 64 "common/Matrix.cpp"
 $56 = ((($m)) + 48|0); //@line 65 "common/Matrix.cpp"
 $57 = +HEAPF32[$56>>2]; //@line 65 "common/Matrix.cpp"
 $58 = $3 * $57; //@line 65 "common/Matrix.cpp"
 $59 = ((($m)) + 52|0); //@line 65 "common/Matrix.cpp"
 $60 = +HEAPF32[$59>>2]; //@line 65 "common/Matrix.cpp"
 $61 = $7 * $60; //@line 65 "common/Matrix.cpp"
 $62 = $58 + $61; //@line 65 "common/Matrix.cpp"
 $63 = ((($m)) + 56|0); //@line 65 "common/Matrix.cpp"
 $64 = +HEAPF32[$63>>2]; //@line 65 "common/Matrix.cpp"
 $65 = $13 * $64; //@line 65 "common/Matrix.cpp"
 $66 = $62 + $65; //@line 65 "common/Matrix.cpp"
 $67 = ((($m)) + 60|0); //@line 65 "common/Matrix.cpp"
 $68 = +HEAPF32[$67>>2]; //@line 65 "common/Matrix.cpp"
 $69 = $19 * $68; //@line 65 "common/Matrix.cpp"
 $70 = $66 + $69; //@line 65 "common/Matrix.cpp"
 $71 = ((($agg$result)) + 48|0); //@line 65 "common/Matrix.cpp"
 HEAPF32[$71>>2] = $70; //@line 65 "common/Matrix.cpp"
 $72 = ((($this)) + 4|0); //@line 67 "common/Matrix.cpp"
 $73 = +HEAPF32[$72>>2]; //@line 67 "common/Matrix.cpp"
 $74 = $73 * $4; //@line 67 "common/Matrix.cpp"
 $75 = ((($this)) + 20|0); //@line 67 "common/Matrix.cpp"
 $76 = +HEAPF32[$75>>2]; //@line 67 "common/Matrix.cpp"
 $77 = $76 * $9; //@line 67 "common/Matrix.cpp"
 $78 = $74 + $77; //@line 67 "common/Matrix.cpp"
 $79 = ((($this)) + 36|0); //@line 67 "common/Matrix.cpp"
 $80 = +HEAPF32[$79>>2]; //@line 67 "common/Matrix.cpp"
 $81 = $80 * $15; //@line 67 "common/Matrix.cpp"
 $82 = $78 + $81; //@line 67 "common/Matrix.cpp"
 $83 = ((($this)) + 52|0); //@line 67 "common/Matrix.cpp"
 $84 = +HEAPF32[$83>>2]; //@line 67 "common/Matrix.cpp"
 $85 = $84 * $21; //@line 67 "common/Matrix.cpp"
 $86 = $82 + $85; //@line 67 "common/Matrix.cpp"
 $87 = ((($agg$result)) + 4|0); //@line 67 "common/Matrix.cpp"
 HEAPF32[$87>>2] = $86; //@line 67 "common/Matrix.cpp"
 $88 = $73 * $25; //@line 68 "common/Matrix.cpp"
 $89 = $76 * $28; //@line 68 "common/Matrix.cpp"
 $90 = $88 + $89; //@line 68 "common/Matrix.cpp"
 $91 = $80 * $32; //@line 68 "common/Matrix.cpp"
 $92 = $90 + $91; //@line 68 "common/Matrix.cpp"
 $93 = $84 * $36; //@line 68 "common/Matrix.cpp"
 $94 = $92 + $93; //@line 68 "common/Matrix.cpp"
 HEAPF32[$2>>2] = $94; //@line 68 "common/Matrix.cpp"
 $95 = $73 * $41; //@line 69 "common/Matrix.cpp"
 $96 = $76 * $44; //@line 69 "common/Matrix.cpp"
 $97 = $95 + $96; //@line 69 "common/Matrix.cpp"
 $98 = $80 * $48; //@line 69 "common/Matrix.cpp"
 $99 = $97 + $98; //@line 69 "common/Matrix.cpp"
 $100 = $84 * $52; //@line 69 "common/Matrix.cpp"
 $101 = $99 + $100; //@line 69 "common/Matrix.cpp"
 $102 = ((($agg$result)) + 36|0); //@line 69 "common/Matrix.cpp"
 HEAPF32[$102>>2] = $101; //@line 69 "common/Matrix.cpp"
 $103 = $73 * $57; //@line 70 "common/Matrix.cpp"
 $104 = $76 * $60; //@line 70 "common/Matrix.cpp"
 $105 = $103 + $104; //@line 70 "common/Matrix.cpp"
 $106 = $80 * $64; //@line 70 "common/Matrix.cpp"
 $107 = $105 + $106; //@line 70 "common/Matrix.cpp"
 $108 = $84 * $68; //@line 70 "common/Matrix.cpp"
 $109 = $107 + $108; //@line 70 "common/Matrix.cpp"
 $110 = ((($agg$result)) + 52|0); //@line 70 "common/Matrix.cpp"
 HEAPF32[$110>>2] = $109; //@line 70 "common/Matrix.cpp"
 $111 = ((($this)) + 8|0); //@line 72 "common/Matrix.cpp"
 $112 = +HEAPF32[$111>>2]; //@line 72 "common/Matrix.cpp"
 $113 = +HEAPF32[$m>>2]; //@line 72 "common/Matrix.cpp"
 $114 = $112 * $113; //@line 72 "common/Matrix.cpp"
 $115 = ((($this)) + 24|0); //@line 72 "common/Matrix.cpp"
 $116 = +HEAPF32[$115>>2]; //@line 72 "common/Matrix.cpp"
 $117 = +HEAPF32[$8>>2]; //@line 72 "common/Matrix.cpp"
 $118 = $116 * $117; //@line 72 "common/Matrix.cpp"
 $119 = $114 + $118; //@line 72 "common/Matrix.cpp"
 $120 = ((($this)) + 40|0); //@line 72 "common/Matrix.cpp"
 $121 = +HEAPF32[$120>>2]; //@line 72 "common/Matrix.cpp"
 $122 = +HEAPF32[$14>>2]; //@line 72 "common/Matrix.cpp"
 $123 = $121 * $122; //@line 72 "common/Matrix.cpp"
 $124 = $119 + $123; //@line 72 "common/Matrix.cpp"
 $125 = ((($this)) + 56|0); //@line 72 "common/Matrix.cpp"
 $126 = +HEAPF32[$125>>2]; //@line 72 "common/Matrix.cpp"
 $127 = +HEAPF32[$20>>2]; //@line 72 "common/Matrix.cpp"
 $128 = $126 * $127; //@line 72 "common/Matrix.cpp"
 $129 = $124 + $128; //@line 72 "common/Matrix.cpp"
 $130 = ((($agg$result)) + 8|0); //@line 72 "common/Matrix.cpp"
 HEAPF32[$130>>2] = $129; //@line 72 "common/Matrix.cpp"
 $131 = +HEAPF32[$24>>2]; //@line 73 "common/Matrix.cpp"
 $132 = $112 * $131; //@line 73 "common/Matrix.cpp"
 $133 = +HEAPF32[$27>>2]; //@line 73 "common/Matrix.cpp"
 $134 = $116 * $133; //@line 73 "common/Matrix.cpp"
 $135 = $132 + $134; //@line 73 "common/Matrix.cpp"
 $136 = +HEAPF32[$31>>2]; //@line 73 "common/Matrix.cpp"
 $137 = $121 * $136; //@line 73 "common/Matrix.cpp"
 $138 = $135 + $137; //@line 73 "common/Matrix.cpp"
 $139 = +HEAPF32[$35>>2]; //@line 73 "common/Matrix.cpp"
 $140 = $126 * $139; //@line 73 "common/Matrix.cpp"
 $141 = $138 + $140; //@line 73 "common/Matrix.cpp"
 $142 = ((($agg$result)) + 24|0); //@line 73 "common/Matrix.cpp"
 HEAPF32[$142>>2] = $141; //@line 73 "common/Matrix.cpp"
 $143 = +HEAPF32[$40>>2]; //@line 74 "common/Matrix.cpp"
 $144 = $112 * $143; //@line 74 "common/Matrix.cpp"
 $145 = +HEAPF32[$43>>2]; //@line 74 "common/Matrix.cpp"
 $146 = $116 * $145; //@line 74 "common/Matrix.cpp"
 $147 = $144 + $146; //@line 74 "common/Matrix.cpp"
 $148 = +HEAPF32[$47>>2]; //@line 74 "common/Matrix.cpp"
 $149 = $121 * $148; //@line 74 "common/Matrix.cpp"
 $150 = $147 + $149; //@line 74 "common/Matrix.cpp"
 $151 = +HEAPF32[$51>>2]; //@line 74 "common/Matrix.cpp"
 $152 = $126 * $151; //@line 74 "common/Matrix.cpp"
 $153 = $150 + $152; //@line 74 "common/Matrix.cpp"
 HEAPF32[$1>>2] = $153; //@line 74 "common/Matrix.cpp"
 $154 = +HEAPF32[$56>>2]; //@line 75 "common/Matrix.cpp"
 $155 = $112 * $154; //@line 75 "common/Matrix.cpp"
 $156 = +HEAPF32[$59>>2]; //@line 75 "common/Matrix.cpp"
 $157 = $116 * $156; //@line 75 "common/Matrix.cpp"
 $158 = $155 + $157; //@line 75 "common/Matrix.cpp"
 $159 = +HEAPF32[$63>>2]; //@line 75 "common/Matrix.cpp"
 $160 = $121 * $159; //@line 75 "common/Matrix.cpp"
 $161 = $158 + $160; //@line 75 "common/Matrix.cpp"
 $162 = +HEAPF32[$67>>2]; //@line 75 "common/Matrix.cpp"
 $163 = $126 * $162; //@line 75 "common/Matrix.cpp"
 $164 = $161 + $163; //@line 75 "common/Matrix.cpp"
 $165 = ((($agg$result)) + 56|0); //@line 75 "common/Matrix.cpp"
 HEAPF32[$165>>2] = $164; //@line 75 "common/Matrix.cpp"
 $166 = ((($this)) + 12|0); //@line 77 "common/Matrix.cpp"
 $167 = +HEAPF32[$166>>2]; //@line 77 "common/Matrix.cpp"
 $168 = $167 * $113; //@line 77 "common/Matrix.cpp"
 $169 = ((($this)) + 28|0); //@line 77 "common/Matrix.cpp"
 $170 = +HEAPF32[$169>>2]; //@line 77 "common/Matrix.cpp"
 $171 = $170 * $117; //@line 77 "common/Matrix.cpp"
 $172 = $168 + $171; //@line 77 "common/Matrix.cpp"
 $173 = ((($this)) + 44|0); //@line 77 "common/Matrix.cpp"
 $174 = +HEAPF32[$173>>2]; //@line 77 "common/Matrix.cpp"
 $175 = $174 * $122; //@line 77 "common/Matrix.cpp"
 $176 = $172 + $175; //@line 77 "common/Matrix.cpp"
 $177 = ((($this)) + 60|0); //@line 77 "common/Matrix.cpp"
 $178 = +HEAPF32[$177>>2]; //@line 77 "common/Matrix.cpp"
 $179 = $178 * $127; //@line 77 "common/Matrix.cpp"
 $180 = $176 + $179; //@line 77 "common/Matrix.cpp"
 $181 = ((($agg$result)) + 12|0); //@line 77 "common/Matrix.cpp"
 HEAPF32[$181>>2] = $180; //@line 77 "common/Matrix.cpp"
 $182 = $167 * $131; //@line 78 "common/Matrix.cpp"
 $183 = $170 * $133; //@line 78 "common/Matrix.cpp"
 $184 = $182 + $183; //@line 78 "common/Matrix.cpp"
 $185 = $174 * $136; //@line 78 "common/Matrix.cpp"
 $186 = $184 + $185; //@line 78 "common/Matrix.cpp"
 $187 = $178 * $139; //@line 78 "common/Matrix.cpp"
 $188 = $186 + $187; //@line 78 "common/Matrix.cpp"
 $189 = ((($agg$result)) + 28|0); //@line 78 "common/Matrix.cpp"
 HEAPF32[$189>>2] = $188; //@line 78 "common/Matrix.cpp"
 $190 = $167 * $143; //@line 79 "common/Matrix.cpp"
 $191 = $170 * $145; //@line 79 "common/Matrix.cpp"
 $192 = $190 + $191; //@line 79 "common/Matrix.cpp"
 $193 = $174 * $148; //@line 79 "common/Matrix.cpp"
 $194 = $192 + $193; //@line 79 "common/Matrix.cpp"
 $195 = $178 * $151; //@line 79 "common/Matrix.cpp"
 $196 = $194 + $195; //@line 79 "common/Matrix.cpp"
 $197 = ((($agg$result)) + 44|0); //@line 79 "common/Matrix.cpp"
 HEAPF32[$197>>2] = $196; //@line 79 "common/Matrix.cpp"
 $198 = $167 * $154; //@line 80 "common/Matrix.cpp"
 $199 = $170 * $156; //@line 80 "common/Matrix.cpp"
 $200 = $198 + $199; //@line 80 "common/Matrix.cpp"
 $201 = $174 * $159; //@line 80 "common/Matrix.cpp"
 $202 = $200 + $201; //@line 80 "common/Matrix.cpp"
 $203 = $178 * $162; //@line 80 "common/Matrix.cpp"
 $204 = $202 + $203; //@line 80 "common/Matrix.cpp"
 HEAPF32[$0>>2] = $204; //@line 80 "common/Matrix.cpp"
 return; //@line 83 "common/Matrix.cpp"
}
function __ZN4love7Matrix4mLERKS0_($this,$m) {
 $this = $this|0;
 $m = $m|0;
 var $t = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $t = sp;
 __ZNK4love7Matrix4mlERKS0_($t,$this,$m); //@line 87 "common/Matrix.cpp"
 dest=$this; src=$t; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 88 "common/Matrix.cpp"
 STACKTOP = sp;return; //@line 89 "common/Matrix.cpp"
}
function __ZNK4love7Matrix411getElementsEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ($this|0); //@line 93 "common/Matrix.cpp"
}
function __ZN4love7Matrix414setTranslationEff($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 dest=$this; stop=dest+60|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 98 "common/Matrix.cpp"
 $0 = ((($this)) + 60|0); //@line 99 "common/Matrix.cpp"
 HEAPF32[$0>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 $1 = ((($this)) + 40|0); //@line 99 "common/Matrix.cpp"
 HEAPF32[$1>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 $2 = ((($this)) + 20|0); //@line 99 "common/Matrix.cpp"
 HEAPF32[$2>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 HEAPF32[$this>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 $3 = ((($this)) + 48|0); //@line 105 "common/Matrix.cpp"
 HEAPF32[$3>>2] = $x; //@line 105 "common/Matrix.cpp"
 $4 = ((($this)) + 52|0); //@line 106 "common/Matrix.cpp"
 HEAPF32[$4>>2] = $y; //@line 106 "common/Matrix.cpp"
 return; //@line 107 "common/Matrix.cpp"
}
function __ZN4love7Matrix49translateEff($this,$x,$y) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $t = 0, $t$i = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $t$i = sp + 64|0;
 $t = sp;
 $0 = ((($t)) + 60|0); //@line 99 "common/Matrix.cpp"
 $1 = ((($t)) + 40|0); //@line 99 "common/Matrix.cpp"
 $2 = ((($t)) + 20|0); //@line 99 "common/Matrix.cpp"
 dest=$t; stop=dest+60|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 98 "common/Matrix.cpp"
 HEAPF32[$0>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 HEAPF32[$1>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 HEAPF32[$2>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 HEAPF32[$t>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 $3 = ((($t)) + 48|0); //@line 105 "common/Matrix.cpp"
 HEAPF32[$3>>2] = $x; //@line 105 "common/Matrix.cpp"
 $4 = ((($t)) + 52|0); //@line 106 "common/Matrix.cpp"
 HEAPF32[$4>>2] = $y; //@line 106 "common/Matrix.cpp"
 __ZNK4love7Matrix4mlERKS0_($t$i,$this,$t); //@line 87 "common/Matrix.cpp"
 dest=$this; src=$t$i; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 88 "common/Matrix.cpp"
 STACKTOP = sp;return; //@line 157 "common/Matrix.cpp"
}
function __ZN4love7Matrix46rotateEf($this,$rad) {
 $this = $this|0;
 $rad = +$rad;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $t = 0, $t$i = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $t$i = sp + 64|0;
 $t = sp;
 $0 = ((($t)) + 60|0); //@line 99 "common/Matrix.cpp"
 $1 = ((($t)) + 40|0); //@line 99 "common/Matrix.cpp"
 $2 = ((($t)) + 20|0); //@line 99 "common/Matrix.cpp"
 dest=$t; stop=dest+60|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 98 "common/Matrix.cpp"
 HEAPF32[$0>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 HEAPF32[$1>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 $3 = (+Math_cos((+$rad))); //@line 112 "common/Matrix.cpp"
 $4 = (+Math_sin((+$rad))); //@line 112 "common/Matrix.cpp"
 HEAPF32[$t>>2] = $3; //@line 113 "common/Matrix.cpp"
 $5 = -$4; //@line 114 "common/Matrix.cpp"
 $6 = ((($t)) + 16|0); //@line 114 "common/Matrix.cpp"
 HEAPF32[$6>>2] = $5; //@line 114 "common/Matrix.cpp"
 $7 = ((($t)) + 4|0); //@line 115 "common/Matrix.cpp"
 HEAPF32[$7>>2] = $4; //@line 115 "common/Matrix.cpp"
 HEAPF32[$2>>2] = $3; //@line 116 "common/Matrix.cpp"
 __ZNK4love7Matrix4mlERKS0_($t$i,$this,$t); //@line 87 "common/Matrix.cpp"
 dest=$this; src=$t$i; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 88 "common/Matrix.cpp"
 STACKTOP = sp;return; //@line 164 "common/Matrix.cpp"
}
function __ZN4love7Matrix45scaleEff($this,$sx,$sy) {
 $this = $this|0;
 $sx = +$sx;
 $sy = +$sy;
 var $0 = 0, $1 = 0, $2 = 0, $t = 0, $t$i = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $t$i = sp + 64|0;
 $t = sp;
 $0 = ((($t)) + 60|0); //@line 99 "common/Matrix.cpp"
 $1 = ((($t)) + 40|0); //@line 99 "common/Matrix.cpp"
 $2 = ((($t)) + 20|0); //@line 99 "common/Matrix.cpp"
 dest=$t; stop=dest+60|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 98 "common/Matrix.cpp"
 HEAPF32[$0>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 HEAPF32[$1>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 HEAPF32[$t>>2] = $sx; //@line 122 "common/Matrix.cpp"
 HEAPF32[$2>>2] = $sy; //@line 123 "common/Matrix.cpp"
 __ZNK4love7Matrix4mlERKS0_($t$i,$this,$t); //@line 87 "common/Matrix.cpp"
 dest=$this; src=$t$i; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 88 "common/Matrix.cpp"
 STACKTOP = sp;return; //@line 171 "common/Matrix.cpp"
}
function __ZN4love7Matrix45shearEff($this,$kx,$ky) {
 $this = $this|0;
 $kx = +$kx;
 $ky = +$ky;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $t = 0, $t$i = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $t$i = sp + 64|0;
 $t = sp;
 $0 = ((($t)) + 60|0); //@line 99 "common/Matrix.cpp"
 $1 = ((($t)) + 40|0); //@line 99 "common/Matrix.cpp"
 $2 = ((($t)) + 20|0); //@line 99 "common/Matrix.cpp"
 dest=$t; stop=dest+60|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 98 "common/Matrix.cpp"
 HEAPF32[$0>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 HEAPF32[$1>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 HEAPF32[$2>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 HEAPF32[$t>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 $3 = ((($t)) + 4|0); //@line 129 "common/Matrix.cpp"
 HEAPF32[$3>>2] = $ky; //@line 129 "common/Matrix.cpp"
 $4 = ((($t)) + 16|0); //@line 130 "common/Matrix.cpp"
 HEAPF32[$4>>2] = $kx; //@line 130 "common/Matrix.cpp"
 __ZNK4love7Matrix4mlERKS0_($t$i,$this,$t); //@line 87 "common/Matrix.cpp"
 dest=$this; src=$t$i; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 88 "common/Matrix.cpp"
 STACKTOP = sp;return; //@line 178 "common/Matrix.cpp"
}
function __ZN4love7Matrix45orthoEffff($agg$result,$left,$right,$bottom,$top) {
 $agg$result = $agg$result|0;
 $left = +$left;
 $right = +$right;
 $bottom = +$bottom;
 $top = +$top;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 dest=$agg$result; stop=dest+60|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 98 "common/Matrix.cpp"
 $0 = ((($agg$result)) + 60|0); //@line 99 "common/Matrix.cpp"
 HEAPF32[$0>>2] = 1.0; //@line 99 "common/Matrix.cpp"
 $1 = ((($agg$result)) + 40|0); //@line 99 "common/Matrix.cpp"
 $2 = ((($agg$result)) + 20|0); //@line 99 "common/Matrix.cpp"
 $3 = $right - $left; //@line 184 "common/Matrix.cpp"
 $4 = 2.0 / $3; //@line 184 "common/Matrix.cpp"
 HEAPF32[$agg$result>>2] = $4; //@line 184 "common/Matrix.cpp"
 $5 = $top - $bottom; //@line 185 "common/Matrix.cpp"
 $6 = 2.0 / $5; //@line 185 "common/Matrix.cpp"
 HEAPF32[$2>>2] = $6; //@line 185 "common/Matrix.cpp"
 HEAPF32[$1>>2] = -1.0; //@line 186 "common/Matrix.cpp"
 $7 = $left + $right; //@line 188 "common/Matrix.cpp"
 $8 = -$7; //@line 188 "common/Matrix.cpp"
 $9 = $8 / $3; //@line 188 "common/Matrix.cpp"
 $10 = ((($agg$result)) + 48|0); //@line 188 "common/Matrix.cpp"
 HEAPF32[$10>>2] = $9; //@line 188 "common/Matrix.cpp"
 $11 = $bottom + $top; //@line 189 "common/Matrix.cpp"
 $12 = -$11; //@line 189 "common/Matrix.cpp"
 $13 = $12 / $5; //@line 189 "common/Matrix.cpp"
 $14 = ((($agg$result)) + 52|0); //@line 189 "common/Matrix.cpp"
 HEAPF32[$14>>2] = $13; //@line 189 "common/Matrix.cpp"
 return; //@line 192 "common/Matrix.cpp"
}
function __ZN4love7Matrix3C2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$this>>2]=0|0;HEAP32[$this+4>>2]=0|0;HEAP32[$this+8>>2]=0|0;HEAP32[$this+12>>2]=0|0;HEAP32[$this+16>>2]=0|0;HEAP32[$this+20>>2]=0|0;HEAP32[$this+24>>2]=0|0;HEAP32[$this+28>>2]=0|0; //@line 235 "common/Matrix.cpp"
 HEAPF32[$this>>2] = 1.0; //@line 236 "common/Matrix.cpp"
 $0 = ((($this)) + 16|0); //@line 236 "common/Matrix.cpp"
 HEAPF32[$0>>2] = 1.0; //@line 236 "common/Matrix.cpp"
 $1 = ((($this)) + 32|0); //@line 236 "common/Matrix.cpp"
 HEAPF32[$1>>2] = 1.0; //@line 236 "common/Matrix.cpp"
 return; //@line 202 "common/Matrix.cpp"
}
function __ZN4love7Matrix3C2ERKNS_7Matrix4E($this,$mat4) {
 $this = $this|0;
 $mat4 = $mat4|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$mat4>>2]|0; //@line 209 "common/Matrix.cpp"
 HEAP32[$this>>2] = $0; //@line 209 "common/Matrix.cpp"
 $1 = ((($mat4)) + 4|0); //@line 210 "common/Matrix.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 210 "common/Matrix.cpp"
 $3 = ((($this)) + 4|0); //@line 210 "common/Matrix.cpp"
 HEAP32[$3>>2] = $2; //@line 210 "common/Matrix.cpp"
 $4 = ((($mat4)) + 8|0); //@line 211 "common/Matrix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 211 "common/Matrix.cpp"
 $6 = ((($this)) + 8|0); //@line 211 "common/Matrix.cpp"
 HEAP32[$6>>2] = $5; //@line 211 "common/Matrix.cpp"
 $7 = ((($mat4)) + 16|0); //@line 214 "common/Matrix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 214 "common/Matrix.cpp"
 $9 = ((($this)) + 12|0); //@line 214 "common/Matrix.cpp"
 HEAP32[$9>>2] = $8; //@line 214 "common/Matrix.cpp"
 $10 = ((($mat4)) + 20|0); //@line 215 "common/Matrix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 215 "common/Matrix.cpp"
 $12 = ((($this)) + 16|0); //@line 215 "common/Matrix.cpp"
 HEAP32[$12>>2] = $11; //@line 215 "common/Matrix.cpp"
 $13 = ((($mat4)) + 24|0); //@line 216 "common/Matrix.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 216 "common/Matrix.cpp"
 $15 = ((($this)) + 20|0); //@line 216 "common/Matrix.cpp"
 HEAP32[$15>>2] = $14; //@line 216 "common/Matrix.cpp"
 $16 = ((($mat4)) + 32|0); //@line 219 "common/Matrix.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 219 "common/Matrix.cpp"
 $18 = ((($this)) + 24|0); //@line 219 "common/Matrix.cpp"
 HEAP32[$18>>2] = $17; //@line 219 "common/Matrix.cpp"
 $19 = ((($mat4)) + 36|0); //@line 220 "common/Matrix.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 220 "common/Matrix.cpp"
 $21 = ((($this)) + 28|0); //@line 220 "common/Matrix.cpp"
 HEAP32[$21>>2] = $20; //@line 220 "common/Matrix.cpp"
 $22 = ((($mat4)) + 40|0); //@line 221 "common/Matrix.cpp"
 $23 = HEAP32[$22>>2]|0; //@line 221 "common/Matrix.cpp"
 $24 = ((($this)) + 32|0); //@line 221 "common/Matrix.cpp"
 HEAP32[$24>>2] = $23; //@line 221 "common/Matrix.cpp"
 return; //@line 222 "common/Matrix.cpp"
}
function __ZN4love7Matrix3C2Efffffffff($this,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $0 = 0.0, $1 = 0.0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0.0;
 var $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+Math_cos((+$angle))); //@line 298 "common/Matrix.cpp"
 $1 = (+Math_sin((+$angle))); //@line 298 "common/Matrix.cpp"
 $2 = $0 * $sx; //@line 304 "common/Matrix.cpp"
 $3 = $1 * $ky; //@line 304 "common/Matrix.cpp"
 $4 = $3 * $sy; //@line 304 "common/Matrix.cpp"
 $5 = $2 - $4; //@line 304 "common/Matrix.cpp"
 HEAPF32[$this>>2] = $5; //@line 304 "common/Matrix.cpp"
 $6 = $1 * $sx; //@line 305 "common/Matrix.cpp"
 $7 = $0 * $ky; //@line 305 "common/Matrix.cpp"
 $8 = $7 * $sy; //@line 305 "common/Matrix.cpp"
 $9 = $6 + $8; //@line 305 "common/Matrix.cpp"
 $10 = ((($this)) + 4|0); //@line 305 "common/Matrix.cpp"
 HEAPF32[$10>>2] = $9; //@line 305 "common/Matrix.cpp"
 $11 = $0 * $kx; //@line 306 "common/Matrix.cpp"
 $12 = $11 * $sx; //@line 306 "common/Matrix.cpp"
 $13 = $1 * $sy; //@line 306 "common/Matrix.cpp"
 $14 = $12 - $13; //@line 306 "common/Matrix.cpp"
 $15 = ((($this)) + 12|0); //@line 306 "common/Matrix.cpp"
 HEAPF32[$15>>2] = $14; //@line 306 "common/Matrix.cpp"
 $16 = $1 * $kx; //@line 307 "common/Matrix.cpp"
 $17 = $16 * $sx; //@line 307 "common/Matrix.cpp"
 $18 = $0 * $sy; //@line 307 "common/Matrix.cpp"
 $19 = $18 + $17; //@line 307 "common/Matrix.cpp"
 $20 = ((($this)) + 16|0); //@line 307 "common/Matrix.cpp"
 HEAPF32[$20>>2] = $19; //@line 307 "common/Matrix.cpp"
 $21 = $5 * $ox; //@line 308 "common/Matrix.cpp"
 $22 = $x - $21; //@line 308 "common/Matrix.cpp"
 $23 = $14 * $oy; //@line 308 "common/Matrix.cpp"
 $24 = $22 - $23; //@line 308 "common/Matrix.cpp"
 $25 = ((($this)) + 24|0); //@line 308 "common/Matrix.cpp"
 HEAPF32[$25>>2] = $24; //@line 308 "common/Matrix.cpp"
 $26 = $9 * $ox; //@line 309 "common/Matrix.cpp"
 $27 = $y - $26; //@line 309 "common/Matrix.cpp"
 $28 = $19 * $oy; //@line 309 "common/Matrix.cpp"
 $29 = $27 - $28; //@line 309 "common/Matrix.cpp"
 $30 = ((($this)) + 28|0); //@line 309 "common/Matrix.cpp"
 HEAPF32[$30>>2] = $29; //@line 309 "common/Matrix.cpp"
 $31 = ((($this)) + 20|0); //@line 311 "common/Matrix.cpp"
 HEAPF32[$31>>2] = 0.0; //@line 311 "common/Matrix.cpp"
 $32 = ((($this)) + 8|0); //@line 311 "common/Matrix.cpp"
 HEAPF32[$32>>2] = 0.0; //@line 311 "common/Matrix.cpp"
 $33 = ((($this)) + 32|0); //@line 312 "common/Matrix.cpp"
 HEAPF32[$33>>2] = 1.0; //@line 312 "common/Matrix.cpp"
 return; //@line 227 "common/Matrix.cpp"
}
function __ZN4love7Matrix317setTransformationEfffffffff($this,$x,$y,$angle,$sx,$sy,$ox,$oy,$kx,$ky) {
 $this = $this|0;
 $x = +$x;
 $y = +$y;
 $angle = +$angle;
 $sx = +$sx;
 $sy = +$sy;
 $ox = +$ox;
 $oy = +$oy;
 $kx = +$kx;
 $ky = +$ky;
 var $0 = 0.0, $1 = 0.0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0.0;
 var $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+Math_cos((+$angle))); //@line 298 "common/Matrix.cpp"
 $1 = (+Math_sin((+$angle))); //@line 298 "common/Matrix.cpp"
 $2 = $0 * $sx; //@line 304 "common/Matrix.cpp"
 $3 = $1 * $ky; //@line 304 "common/Matrix.cpp"
 $4 = $3 * $sy; //@line 304 "common/Matrix.cpp"
 $5 = $2 - $4; //@line 304 "common/Matrix.cpp"
 HEAPF32[$this>>2] = $5; //@line 304 "common/Matrix.cpp"
 $6 = $1 * $sx; //@line 305 "common/Matrix.cpp"
 $7 = $0 * $ky; //@line 305 "common/Matrix.cpp"
 $8 = $7 * $sy; //@line 305 "common/Matrix.cpp"
 $9 = $6 + $8; //@line 305 "common/Matrix.cpp"
 $10 = ((($this)) + 4|0); //@line 305 "common/Matrix.cpp"
 HEAPF32[$10>>2] = $9; //@line 305 "common/Matrix.cpp"
 $11 = $0 * $kx; //@line 306 "common/Matrix.cpp"
 $12 = $11 * $sx; //@line 306 "common/Matrix.cpp"
 $13 = $1 * $sy; //@line 306 "common/Matrix.cpp"
 $14 = $12 - $13; //@line 306 "common/Matrix.cpp"
 $15 = ((($this)) + 12|0); //@line 306 "common/Matrix.cpp"
 HEAPF32[$15>>2] = $14; //@line 306 "common/Matrix.cpp"
 $16 = $1 * $kx; //@line 307 "common/Matrix.cpp"
 $17 = $16 * $sx; //@line 307 "common/Matrix.cpp"
 $18 = $0 * $sy; //@line 307 "common/Matrix.cpp"
 $19 = $18 + $17; //@line 307 "common/Matrix.cpp"
 $20 = ((($this)) + 16|0); //@line 307 "common/Matrix.cpp"
 HEAPF32[$20>>2] = $19; //@line 307 "common/Matrix.cpp"
 $21 = $5 * $ox; //@line 308 "common/Matrix.cpp"
 $22 = $x - $21; //@line 308 "common/Matrix.cpp"
 $23 = $14 * $oy; //@line 308 "common/Matrix.cpp"
 $24 = $22 - $23; //@line 308 "common/Matrix.cpp"
 $25 = ((($this)) + 24|0); //@line 308 "common/Matrix.cpp"
 HEAPF32[$25>>2] = $24; //@line 308 "common/Matrix.cpp"
 $26 = $9 * $ox; //@line 309 "common/Matrix.cpp"
 $27 = $y - $26; //@line 309 "common/Matrix.cpp"
 $28 = $19 * $oy; //@line 309 "common/Matrix.cpp"
 $29 = $27 - $28; //@line 309 "common/Matrix.cpp"
 $30 = ((($this)) + 28|0); //@line 309 "common/Matrix.cpp"
 HEAPF32[$30>>2] = $29; //@line 309 "common/Matrix.cpp"
 $31 = ((($this)) + 20|0); //@line 311 "common/Matrix.cpp"
 HEAPF32[$31>>2] = 0.0; //@line 311 "common/Matrix.cpp"
 $32 = ((($this)) + 8|0); //@line 311 "common/Matrix.cpp"
 HEAPF32[$32>>2] = 0.0; //@line 311 "common/Matrix.cpp"
 $33 = ((($this)) + 32|0); //@line 312 "common/Matrix.cpp"
 HEAPF32[$33>>2] = 1.0; //@line 312 "common/Matrix.cpp"
 return; //@line 313 "common/Matrix.cpp"
}
function __ZN4love7Matrix3D2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 231 "common/Matrix.cpp"
}
function __ZNK4love7Matrix311getElementsEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ($this|0); //@line 266 "common/Matrix.cpp"
}
function __ZNK4love7Matrix317transposedInverseEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0.0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0.0, $2 = 0.0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0;
 var $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0.0, $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0;
 var $45 = 0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0;
 var $63 = 0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $7 = 0.0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = +HEAPF32[$this>>2]; //@line 275 "common/Matrix.cpp"
 $1 = ((($this)) + 16|0); //@line 275 "common/Matrix.cpp"
 $2 = +HEAPF32[$1>>2]; //@line 275 "common/Matrix.cpp"
 $3 = ((($this)) + 32|0); //@line 275 "common/Matrix.cpp"
 $4 = +HEAPF32[$3>>2]; //@line 275 "common/Matrix.cpp"
 $5 = $2 * $4; //@line 275 "common/Matrix.cpp"
 $6 = ((($this)) + 28|0); //@line 275 "common/Matrix.cpp"
 $7 = +HEAPF32[$6>>2]; //@line 275 "common/Matrix.cpp"
 $8 = ((($this)) + 20|0); //@line 275 "common/Matrix.cpp"
 $9 = +HEAPF32[$8>>2]; //@line 275 "common/Matrix.cpp"
 $10 = $7 * $9; //@line 275 "common/Matrix.cpp"
 $11 = $5 - $10; //@line 275 "common/Matrix.cpp"
 $12 = $0 * $11; //@line 275 "common/Matrix.cpp"
 $13 = ((($this)) + 4|0); //@line 276 "common/Matrix.cpp"
 $14 = +HEAPF32[$13>>2]; //@line 276 "common/Matrix.cpp"
 $15 = ((($this)) + 12|0); //@line 276 "common/Matrix.cpp"
 $16 = +HEAPF32[$15>>2]; //@line 276 "common/Matrix.cpp"
 $17 = $4 * $16; //@line 276 "common/Matrix.cpp"
 $18 = ((($this)) + 24|0); //@line 276 "common/Matrix.cpp"
 $19 = +HEAPF32[$18>>2]; //@line 276 "common/Matrix.cpp"
 $20 = $9 * $19; //@line 276 "common/Matrix.cpp"
 $21 = $17 - $20; //@line 276 "common/Matrix.cpp"
 $22 = $14 * $21; //@line 276 "common/Matrix.cpp"
 $23 = $12 - $22; //@line 276 "common/Matrix.cpp"
 $24 = ((($this)) + 8|0); //@line 277 "common/Matrix.cpp"
 $25 = +HEAPF32[$24>>2]; //@line 277 "common/Matrix.cpp"
 $26 = $7 * $16; //@line 277 "common/Matrix.cpp"
 $27 = $2 * $19; //@line 277 "common/Matrix.cpp"
 $28 = $26 - $27; //@line 277 "common/Matrix.cpp"
 $29 = $25 * $28; //@line 277 "common/Matrix.cpp"
 $30 = $29 + $23; //@line 277 "common/Matrix.cpp"
 $31 = 1.0 / $30; //@line 279 "common/Matrix.cpp"
 ;HEAP32[$agg$result>>2]=0|0;HEAP32[$agg$result+4>>2]=0|0;HEAP32[$agg$result+8>>2]=0|0;HEAP32[$agg$result+12>>2]=0|0;HEAP32[$agg$result+16>>2]=0|0;HEAP32[$agg$result+20>>2]=0|0;HEAP32[$agg$result+24>>2]=0|0; //@line 235 "common/Matrix.cpp"
 $32 = ((($agg$result)) + 16|0); //@line 236 "common/Matrix.cpp"
 $33 = ((($agg$result)) + 32|0); //@line 236 "common/Matrix.cpp"
 $34 = $31 * $11; //@line 283 "common/Matrix.cpp"
 HEAPF32[$agg$result>>2] = $34; //@line 283 "common/Matrix.cpp"
 $35 = -$31; //@line 284 "common/Matrix.cpp"
 $36 = $14 * $4; //@line 284 "common/Matrix.cpp"
 $37 = $25 * $7; //@line 284 "common/Matrix.cpp"
 $38 = $36 - $37; //@line 284 "common/Matrix.cpp"
 $39 = $38 * $35; //@line 284 "common/Matrix.cpp"
 $40 = ((($agg$result)) + 12|0); //@line 284 "common/Matrix.cpp"
 HEAPF32[$40>>2] = $39; //@line 284 "common/Matrix.cpp"
 $41 = $14 * $9; //@line 285 "common/Matrix.cpp"
 $42 = $25 * $2; //@line 285 "common/Matrix.cpp"
 $43 = $41 - $42; //@line 285 "common/Matrix.cpp"
 $44 = $31 * $43; //@line 285 "common/Matrix.cpp"
 $45 = ((($agg$result)) + 24|0); //@line 285 "common/Matrix.cpp"
 HEAPF32[$45>>2] = $44; //@line 285 "common/Matrix.cpp"
 $46 = $21 * $35; //@line 286 "common/Matrix.cpp"
 $47 = ((($agg$result)) + 4|0); //@line 286 "common/Matrix.cpp"
 HEAPF32[$47>>2] = $46; //@line 286 "common/Matrix.cpp"
 $48 = $0 * $4; //@line 287 "common/Matrix.cpp"
 $49 = $25 * $19; //@line 287 "common/Matrix.cpp"
 $50 = $48 - $49; //@line 287 "common/Matrix.cpp"
 $51 = $31 * $50; //@line 287 "common/Matrix.cpp"
 HEAPF32[$32>>2] = $51; //@line 287 "common/Matrix.cpp"
 $52 = $0 * $9; //@line 288 "common/Matrix.cpp"
 $53 = $16 * $25; //@line 288 "common/Matrix.cpp"
 $54 = $52 - $53; //@line 288 "common/Matrix.cpp"
 $55 = $54 * $35; //@line 288 "common/Matrix.cpp"
 $56 = ((($agg$result)) + 28|0); //@line 288 "common/Matrix.cpp"
 HEAPF32[$56>>2] = $55; //@line 288 "common/Matrix.cpp"
 $57 = $31 * $28; //@line 289 "common/Matrix.cpp"
 $58 = ((($agg$result)) + 8|0); //@line 289 "common/Matrix.cpp"
 HEAPF32[$58>>2] = $57; //@line 289 "common/Matrix.cpp"
 $59 = $0 * $7; //@line 290 "common/Matrix.cpp"
 $60 = $19 * $14; //@line 290 "common/Matrix.cpp"
 $61 = $59 - $60; //@line 290 "common/Matrix.cpp"
 $62 = $61 * $35; //@line 290 "common/Matrix.cpp"
 $63 = ((($agg$result)) + 20|0); //@line 290 "common/Matrix.cpp"
 HEAPF32[$63>>2] = $62; //@line 290 "common/Matrix.cpp"
 $64 = $0 * $2; //@line 291 "common/Matrix.cpp"
 $65 = $16 * $14; //@line 291 "common/Matrix.cpp"
 $66 = $64 - $65; //@line 291 "common/Matrix.cpp"
 $67 = $31 * $66; //@line 291 "common/Matrix.cpp"
 HEAPF32[$33>>2] = $67; //@line 291 "common/Matrix.cpp"
 return; //@line 294 "common/Matrix.cpp"
}
function __ZN4love10filesystem6physfs4FileD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (35060); //@line 55 "modules/filesystem/physfs/File.cpp"
 $0 = ((($this)) + 24|0); //@line 56 "modules/filesystem/physfs/File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 56 "modules/filesystem/physfs/File.cpp"
 $2 = ($1|0)==(0); //@line 56 "modules/filesystem/physfs/File.cpp"
 if (!($2)) {
  $3 = ((($this)) + 20|0); //@line 122 "modules/filesystem/physfs/File.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 122 "modules/filesystem/physfs/File.cpp"
  $5 = ($4|0)==(0|0); //@line 122 "modules/filesystem/physfs/File.cpp"
  if (!($5)) {
   __THREW__ = 0;
   $6 = (invoke_ii(4250,($4|0))|0); //@line 122 "modules/filesystem/physfs/File.cpp"
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if ($8) {
    $10 = ___cxa_find_matching_catch_3(0|0)|0;
    $11 = tempRet0;
    $12 = ((($this)) + 8|0); //@line 58 "modules/filesystem/physfs/File.cpp"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12); //@line 58 "modules/filesystem/physfs/File.cpp"
    __ZN4love10filesystem4FileD2Ev($this); //@line 58 "modules/filesystem/physfs/File.cpp"
    ___clang_call_terminate($10); //@line 58 "modules/filesystem/physfs/File.cpp"
    // unreachable; //@line 58 "modules/filesystem/physfs/File.cpp"
   }
   $9 = ($6|0)==(0); //@line 122 "modules/filesystem/physfs/File.cpp"
   if (!($9)) {
    HEAP32[$0>>2] = 0; //@line 125 "modules/filesystem/physfs/File.cpp"
    HEAP32[$3>>2] = 0; //@line 126 "modules/filesystem/physfs/File.cpp"
   }
  }
 }
 $13 = ((($this)) + 8|0); //@line 58 "modules/filesystem/physfs/File.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13); //@line 58 "modules/filesystem/physfs/File.cpp"
 __ZN4love10filesystem4FileD2Ev($this); //@line 58 "modules/filesystem/physfs/File.cpp"
 return; //@line 58 "modules/filesystem/physfs/File.cpp"
}
function __ZN4love10filesystem6physfs4FileD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (35060); //@line 55 "modules/filesystem/physfs/File.cpp"
 $0 = ((($this)) + 24|0); //@line 56 "modules/filesystem/physfs/File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 56 "modules/filesystem/physfs/File.cpp"
 $2 = ($1|0)==(0); //@line 56 "modules/filesystem/physfs/File.cpp"
 if (!($2)) {
  $3 = ((($this)) + 20|0); //@line 122 "modules/filesystem/physfs/File.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 122 "modules/filesystem/physfs/File.cpp"
  $5 = ($4|0)==(0|0); //@line 122 "modules/filesystem/physfs/File.cpp"
  if (!($5)) {
   __THREW__ = 0;
   $6 = (invoke_ii(4250,($4|0))|0); //@line 122 "modules/filesystem/physfs/File.cpp"
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if ($8) {
    $10 = ___cxa_find_matching_catch_3(0|0)|0;
    $11 = tempRet0;
    $12 = ((($this)) + 8|0); //@line 58 "modules/filesystem/physfs/File.cpp"
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12); //@line 58 "modules/filesystem/physfs/File.cpp"
    __ZN4love10filesystem4FileD2Ev($this); //@line 58 "modules/filesystem/physfs/File.cpp"
    ___clang_call_terminate($10); //@line 58 "modules/filesystem/physfs/File.cpp"
    // unreachable; //@line 58 "modules/filesystem/physfs/File.cpp"
   }
   $9 = ($6|0)==(0); //@line 122 "modules/filesystem/physfs/File.cpp"
   if (!($9)) {
    HEAP32[$0>>2] = 0; //@line 125 "modules/filesystem/physfs/File.cpp"
    HEAP32[$3>>2] = 0; //@line 126 "modules/filesystem/physfs/File.cpp"
   }
  }
 }
 $13 = ((($this)) + 8|0); //@line 58 "modules/filesystem/physfs/File.cpp"
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13); //@line 58 "modules/filesystem/physfs/File.cpp"
 __ZN4love10filesystem4FileD2Ev($this); //@line 58 "modules/filesystem/physfs/File.cpp"
 __ZdlPv($this); //@line 55 "modules/filesystem/physfs/File.cpp"
 return; //@line 58 "modules/filesystem/physfs/File.cpp"
}
function __ZN4love10filesystem6physfs4File4openENS0_4File4ModeE($this,$mode) {
 $this = $this|0;
 $mode = $mode|0;
 var $$ = 0, $$0 = 0, $$02 = 0, $$03 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $handle$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_ptr8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($mode|0)==(0); //@line 62 "modules/filesystem/physfs/File.cpp"
 if ($0) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 118 "modules/filesystem/physfs/File.cpp"
 }
 $1 = (_PHYSFS_isInit()|0); //@line 65 "modules/filesystem/physfs/File.cpp"
 $2 = ($1|0)==(0); //@line 65 "modules/filesystem/physfs/File.cpp"
 if ($2) {
  $3 = (___cxa_allocate_exception(16)|0); //@line 66 "modules/filesystem/physfs/File.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($3|0),(507115|0),($vararg_buffer|0)); //@line 66 "modules/filesystem/physfs/File.cpp"
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if (!($5)) {
   ___cxa_throw(($3|0),(96|0),(344|0)); //@line 66 "modules/filesystem/physfs/File.cpp"
   // unreachable; //@line 66 "modules/filesystem/physfs/File.cpp"
  }
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  ___cxa_free_exception(($3|0)); //@line 66 "modules/filesystem/physfs/File.cpp"
  $$02 = $7;$$03 = $6;
  ___resumeException($$03|0);
  // unreachable;
 }
 $8 = ($mode|0)==(1); //@line 69 "modules/filesystem/physfs/File.cpp"
 if ($8) {
  $9 = ((($this)) + 8|0); //@line 69 "modules/filesystem/physfs/File.cpp"
  $10 = HEAP8[$9>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $11 = $10 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $12 = ($11<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($12) {
   $15 = ((($9)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $16 = $15;
  } else {
   $13 = ((($this)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $14 = HEAP32[$13>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $16 = $14;
  }
  $17 = (_PHYSFS_exists($16)|0); //@line 69 "modules/filesystem/physfs/File.cpp"
  $18 = ($17|0)==(0); //@line 69 "modules/filesystem/physfs/File.cpp"
  if ($18) {
   $19 = (___cxa_allocate_exception(16)|0); //@line 70 "modules/filesystem/physfs/File.cpp"
   $20 = HEAP8[$9>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $21 = $20 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $22 = ($21<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   if ($22) {
    $25 = ((($9)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $26 = $25;
   } else {
    $23 = ((($this)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $24 = HEAP32[$23>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
    $26 = $24;
   }
   __THREW__ = 0;
   HEAP32[$vararg_buffer1>>2] = $26; //@line 70 "modules/filesystem/physfs/File.cpp"
   invoke_viii(3349,($19|0),(507142|0),($vararg_buffer1|0)); //@line 70 "modules/filesystem/physfs/File.cpp"
   $27 = __THREW__; __THREW__ = 0;
   $28 = $27&1;
   if (!($28)) {
    ___cxa_throw(($19|0),(96|0),(344|0)); //@line 70 "modules/filesystem/physfs/File.cpp"
    // unreachable; //@line 70 "modules/filesystem/physfs/File.cpp"
   }
   $29 = ___cxa_find_matching_catch_2()|0;
   $30 = tempRet0;
   ___cxa_free_exception(($19|0)); //@line 70 "modules/filesystem/physfs/File.cpp"
   $$02 = $30;$$03 = $29;
   ___resumeException($$03|0);
   // unreachable;
  }
 } else {
  $31 = $mode | 1; //@line 73 "modules/filesystem/physfs/File.cpp"
  $32 = ($31|0)==(3); //@line 73 "modules/filesystem/physfs/File.cpp"
  if ($32) {
   $33 = (_PHYSFS_getWriteDir()|0); //@line 73 "modules/filesystem/physfs/File.cpp"
   $34 = ($33|0)==(0|0); //@line 73 "modules/filesystem/physfs/File.cpp"
   if ($34) {
    $35 = (__ZN4love10filesystem24hack_setupWriteDirectoryEv()|0); //@line 73 "modules/filesystem/physfs/File.cpp"
    if (!($35)) {
     $36 = (___cxa_allocate_exception(16)|0); //@line 74 "modules/filesystem/physfs/File.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($36|0),(507182|0),($vararg_buffer3|0)); //@line 74 "modules/filesystem/physfs/File.cpp"
     $37 = __THREW__; __THREW__ = 0;
     $38 = $37&1;
     if (!($38)) {
      ___cxa_throw(($36|0),(96|0),(344|0)); //@line 74 "modules/filesystem/physfs/File.cpp"
      // unreachable; //@line 74 "modules/filesystem/physfs/File.cpp"
     }
     $39 = ___cxa_find_matching_catch_2()|0;
     $40 = tempRet0;
     ___cxa_free_exception(($36|0)); //@line 74 "modules/filesystem/physfs/File.cpp"
     $$02 = $40;$$03 = $39;
     ___resumeException($$03|0);
     // unreachable;
    }
   }
  }
 }
 $41 = ((($this)) + 20|0); //@line 77 "modules/filesystem/physfs/File.cpp"
 $42 = HEAP32[$41>>2]|0; //@line 77 "modules/filesystem/physfs/File.cpp"
 $43 = ($42|0)==(0|0); //@line 77 "modules/filesystem/physfs/File.cpp"
 if (!($43)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 118 "modules/filesystem/physfs/File.cpp"
 }
 (_PHYSFS_getLastError()|0); //@line 80 "modules/filesystem/physfs/File.cpp"
 switch ($mode|0) {
 case 1:  {
  $44 = ((($this)) + 8|0); //@line 86 "modules/filesystem/physfs/File.cpp"
  $45 = HEAP8[$44>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $46 = $45 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $47 = ($46<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($47) {
   $50 = ((($44)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $51 = $50;
  } else {
   $48 = ((($this)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $49 = HEAP32[$48>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $51 = $49;
  }
  $52 = (_PHYSFS_openRead($51)|0); //@line 86 "modules/filesystem/physfs/File.cpp"
  $handle$0 = $52;
  label = 37;
  break;
 }
 case 3:  {
  $53 = ((($this)) + 8|0); //@line 89 "modules/filesystem/physfs/File.cpp"
  $54 = HEAP8[$53>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $55 = $54 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $56 = ($55<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($56) {
   $59 = ((($53)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $60 = $59;
  } else {
   $57 = ((($this)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $58 = HEAP32[$57>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $60 = $58;
  }
  $61 = (_PHYSFS_openAppend($60)|0); //@line 89 "modules/filesystem/physfs/File.cpp"
  $handle$0 = $61;
  label = 37;
  break;
 }
 case 2:  {
  $62 = ((($this)) + 8|0); //@line 92 "modules/filesystem/physfs/File.cpp"
  $63 = HEAP8[$62>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $64 = $63 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $65 = ($64<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  if ($65) {
   $68 = ((($62)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $69 = $68;
  } else {
   $66 = ((($this)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $67 = HEAP32[$66>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
   $69 = $67;
  }
  $70 = (_PHYSFS_openWrite($69)|0); //@line 92 "modules/filesystem/physfs/File.cpp"
  $handle$0 = $70;
  label = 37;
  break;
 }
 default: {
 }
 }
 if ((label|0) == 37) {
  $71 = ($handle$0|0)==(0|0); //@line 98 "modules/filesystem/physfs/File.cpp"
  if (!($71)) {
   HEAP32[$41>>2] = $handle$0; //@line 106 "modules/filesystem/physfs/File.cpp"
   $87 = ((($this)) + 24|0); //@line 108 "modules/filesystem/physfs/File.cpp"
   HEAP32[$87>>2] = $mode; //@line 108 "modules/filesystem/physfs/File.cpp"
   $88 = HEAP32[$this>>2]|0; //@line 110 "modules/filesystem/physfs/File.cpp"
   $89 = ((($88)) + 64|0); //@line 110 "modules/filesystem/physfs/File.cpp"
   $90 = HEAP32[$89>>2]|0; //@line 110 "modules/filesystem/physfs/File.cpp"
   $91 = ((($this)) + 28|0); //@line 110 "modules/filesystem/physfs/File.cpp"
   $92 = HEAP32[$91>>2]|0; //@line 110 "modules/filesystem/physfs/File.cpp"
   $93 = ((($this)) + 32|0); //@line 110 "modules/filesystem/physfs/File.cpp"
   $94 = $93; //@line 110 "modules/filesystem/physfs/File.cpp"
   $95 = $94; //@line 110 "modules/filesystem/physfs/File.cpp"
   $96 = HEAP32[$95>>2]|0; //@line 110 "modules/filesystem/physfs/File.cpp"
   $97 = (($94) + 4)|0; //@line 110 "modules/filesystem/physfs/File.cpp"
   $98 = $97; //@line 110 "modules/filesystem/physfs/File.cpp"
   $99 = HEAP32[$98>>2]|0; //@line 110 "modules/filesystem/physfs/File.cpp"
   $100 = (FUNCTION_TABLE_iiiii[$90 & 8191]($this,$92,$96,$99)|0); //@line 110 "modules/filesystem/physfs/File.cpp"
   if (!($100)) {
    HEAP32[$91>>2] = 0; //@line 113 "modules/filesystem/physfs/File.cpp"
    $101 = $93; //@line 114 "modules/filesystem/physfs/File.cpp"
    $102 = $101; //@line 114 "modules/filesystem/physfs/File.cpp"
    HEAP32[$102>>2] = 0; //@line 114 "modules/filesystem/physfs/File.cpp"
    $103 = (($101) + 4)|0; //@line 114 "modules/filesystem/physfs/File.cpp"
    $104 = $103; //@line 114 "modules/filesystem/physfs/File.cpp"
    HEAP32[$104>>2] = 0; //@line 114 "modules/filesystem/physfs/File.cpp"
   }
   $105 = HEAP32[$41>>2]|0; //@line 117 "modules/filesystem/physfs/File.cpp"
   $106 = ($105|0)!=(0|0); //@line 117 "modules/filesystem/physfs/File.cpp"
   $$0 = $106;
   STACKTOP = sp;return ($$0|0); //@line 118 "modules/filesystem/physfs/File.cpp"
  }
 }
 $72 = (_PHYSFS_getLastError()|0); //@line 100 "modules/filesystem/physfs/File.cpp"
 $73 = ($72|0)==(0|0); //@line 101 "modules/filesystem/physfs/File.cpp"
 $$ = $73 ? 517474 : $72; //@line 101 "modules/filesystem/physfs/File.cpp"
 $74 = (___cxa_allocate_exception(16)|0); //@line 103 "modules/filesystem/physfs/File.cpp"
 $75 = ((($this)) + 8|0); //@line 103 "modules/filesystem/physfs/File.cpp"
 $76 = HEAP8[$75>>0]|0; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $77 = $76 & 1; //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 $78 = ($77<<24>>24)==(0); //@line 1664 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
 if ($78) {
  $81 = ((($75)) + 1|0); //@line 1752 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $82 = $81;
 } else {
  $79 = ((($this)) + 16|0); //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $80 = HEAP32[$79>>2]|0; //@line 1746 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/string"
  $82 = $80;
 }
 __THREW__ = 0;
 HEAP32[$vararg_buffer5>>2] = $82; //@line 103 "modules/filesystem/physfs/File.cpp"
 $vararg_ptr8 = ((($vararg_buffer5)) + 4|0); //@line 103 "modules/filesystem/physfs/File.cpp"
 HEAP32[$vararg_ptr8>>2] = $$; //@line 103 "modules/filesystem/physfs/File.cpp"
 invoke_viii(3349,($74|0),(507213|0),($vararg_buffer5|0)); //@line 103 "modules/filesystem/physfs/File.cpp"
 $83 = __THREW__; __THREW__ = 0;
 $84 = $83&1;
 if (!($84)) {
  ___cxa_throw(($74|0),(96|0),(344|0)); //@line 103 "modules/filesystem/physfs/File.cpp"
  // unreachable; //@line 103 "modules/filesystem/physfs/File.cpp"
 }
 $85 = ___cxa_find_matching_catch_2()|0;
 $86 = tempRet0;
 ___cxa_free_exception(($74|0)); //@line 103 "modules/filesystem/physfs/File.cpp"
 $$02 = $86;$$03 = $85;
 ___resumeException($$03|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love10filesystem6physfs4File5closeEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 122 "modules/filesystem/physfs/File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 122 "modules/filesystem/physfs/File.cpp"
 $2 = ($1|0)==(0|0); //@line 122 "modules/filesystem/physfs/File.cpp"
 if ($2) {
  $$0 = 0;
 } else {
  $3 = (_PHYSFS_close($1)|0); //@line 122 "modules/filesystem/physfs/File.cpp"
  $4 = ($3|0)==(0); //@line 122 "modules/filesystem/physfs/File.cpp"
  if ($4) {
   $$0 = 0;
  } else {
   $5 = ((($this)) + 24|0); //@line 125 "modules/filesystem/physfs/File.cpp"
   HEAP32[$5>>2] = 0; //@line 125 "modules/filesystem/physfs/File.cpp"
   HEAP32[$0>>2] = 0; //@line 126 "modules/filesystem/physfs/File.cpp"
   $$0 = 1;
  }
 }
 return ($$0|0); //@line 129 "modules/filesystem/physfs/File.cpp"
}
function __ZNK4love10filesystem6physfs4File6isOpenEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 133 "modules/filesystem/physfs/File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 133 "modules/filesystem/physfs/File.cpp"
 $2 = ($1|0)==(0); //@line 133 "modules/filesystem/physfs/File.cpp"
 if ($2) {
  $6 = 0;
  return ($6|0); //@line 133 "modules/filesystem/physfs/File.cpp"
 }
 $3 = ((($this)) + 20|0); //@line 133 "modules/filesystem/physfs/File.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 133 "modules/filesystem/physfs/File.cpp"
 $5 = ($4|0)!=(0|0); //@line 133 "modules/filesystem/physfs/File.cpp"
 $6 = $5;
 return ($6|0); //@line 133 "modules/filesystem/physfs/File.cpp"
}
function __ZN4love10filesystem6physfs4File7getSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 140 "modules/filesystem/physfs/File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 140 "modules/filesystem/physfs/File.cpp"
 $2 = ($1|0)==(0|0); //@line 140 "modules/filesystem/physfs/File.cpp"
 if ($2) {
  $3 = HEAP32[$this>>2]|0; //@line 142 "modules/filesystem/physfs/File.cpp"
  $4 = ((($3)) + 16|0); //@line 142 "modules/filesystem/physfs/File.cpp"
  $5 = HEAP32[$4>>2]|0; //@line 142 "modules/filesystem/physfs/File.cpp"
  (FUNCTION_TABLE_iii[$5 & 8191]($this,1)|0); //@line 142 "modules/filesystem/physfs/File.cpp"
  $6 = HEAP32[$0>>2]|0; //@line 143 "modules/filesystem/physfs/File.cpp"
  $7 = (_PHYSFS_fileLength($6)|0); //@line 143 "modules/filesystem/physfs/File.cpp"
  $8 = tempRet0; //@line 143 "modules/filesystem/physfs/File.cpp"
  $9 = HEAP32[$this>>2]|0; //@line 144 "modules/filesystem/physfs/File.cpp"
  $10 = ((($9)) + 20|0); //@line 144 "modules/filesystem/physfs/File.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 144 "modules/filesystem/physfs/File.cpp"
  (FUNCTION_TABLE_ii[$11 & 8191]($this)|0); //@line 144 "modules/filesystem/physfs/File.cpp"
  $14 = $8;$15 = $7;
  tempRet0 = ($14); //@line 149 "modules/filesystem/physfs/File.cpp"
  return ($15|0); //@line 149 "modules/filesystem/physfs/File.cpp"
 } else {
  $12 = (_PHYSFS_fileLength($1)|0); //@line 148 "modules/filesystem/physfs/File.cpp"
  $13 = tempRet0; //@line 148 "modules/filesystem/physfs/File.cpp"
  $14 = $13;$15 = $12;
  tempRet0 = ($14); //@line 149 "modules/filesystem/physfs/File.cpp"
  return ($15|0); //@line 149 "modules/filesystem/physfs/File.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem6physfs4File4readEPvx($this,$dst,$0,$1) {
 $this = $this|0;
 $dst = $dst|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$01 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $2 = ((($this)) + 20|0); //@line 153 "modules/filesystem/physfs/File.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 153 "modules/filesystem/physfs/File.cpp"
 $4 = ($3|0)==(0|0); //@line 153 "modules/filesystem/physfs/File.cpp"
 if (!($4)) {
  $5 = ((($this)) + 24|0); //@line 153 "modules/filesystem/physfs/File.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 153 "modules/filesystem/physfs/File.cpp"
  $7 = ($6|0)==(1); //@line 153 "modules/filesystem/physfs/File.cpp"
  if ($7) {
   $13 = (_PHYSFS_fileLength($3)|0); //@line 156 "modules/filesystem/physfs/File.cpp"
   $14 = tempRet0; //@line 156 "modules/filesystem/physfs/File.cpp"
   $15 = ($0|0)==(-1); //@line 157 "modules/filesystem/physfs/File.cpp"
   $16 = ($1|0)==(-1); //@line 157 "modules/filesystem/physfs/File.cpp"
   $17 = $15 & $16; //@line 157 "modules/filesystem/physfs/File.cpp"
   $18 = $17 ? $13 : $0; //@line 157 "modules/filesystem/physfs/File.cpp"
   $19 = $17 ? $14 : $1; //@line 157 "modules/filesystem/physfs/File.cpp"
   $20 = ($19|0)>($14|0); //@line 158 "modules/filesystem/physfs/File.cpp"
   $21 = ($18>>>0)>($13>>>0); //@line 158 "modules/filesystem/physfs/File.cpp"
   $22 = ($19|0)==($14|0); //@line 158 "modules/filesystem/physfs/File.cpp"
   $23 = $22 & $21; //@line 158 "modules/filesystem/physfs/File.cpp"
   $24 = $20 | $23; //@line 158 "modules/filesystem/physfs/File.cpp"
   $25 = $24 ? $13 : $18; //@line 158 "modules/filesystem/physfs/File.cpp"
   $26 = $24 ? $14 : $19; //@line 158 "modules/filesystem/physfs/File.cpp"
   $27 = ($26|0)>(0); //@line 160 "modules/filesystem/physfs/File.cpp"
   $28 = ($25>>>0)>(4294967295); //@line 160 "modules/filesystem/physfs/File.cpp"
   $29 = ($26|0)==(0); //@line 160 "modules/filesystem/physfs/File.cpp"
   $30 = $29 & $28; //@line 160 "modules/filesystem/physfs/File.cpp"
   $31 = $27 | $30; //@line 160 "modules/filesystem/physfs/File.cpp"
   $32 = $31 ? -1 : $25; //@line 160 "modules/filesystem/physfs/File.cpp"
   $33 = $31 ? 0 : $26; //@line 160 "modules/filesystem/physfs/File.cpp"
   $34 = ($33|0)<(0); //@line 162 "modules/filesystem/physfs/File.cpp"
   if (!($34)) {
    $40 = HEAP32[$2>>2]|0; //@line 168 "modules/filesystem/physfs/File.cpp"
    $41 = (_PHYSFS_read($40,$dst,1,$32)|0); //@line 168 "modules/filesystem/physfs/File.cpp"
    $42 = tempRet0; //@line 168 "modules/filesystem/physfs/File.cpp"
    tempRet0 = ($42); //@line 171 "modules/filesystem/physfs/File.cpp"
    STACKTOP = sp;return ($41|0); //@line 171 "modules/filesystem/physfs/File.cpp"
   }
   $35 = (___cxa_allocate_exception(16)|0); //@line 163 "modules/filesystem/physfs/File.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($35|0),(507273|0),($vararg_buffer1|0)); //@line 163 "modules/filesystem/physfs/File.cpp"
   $36 = __THREW__; __THREW__ = 0;
   $37 = $36&1;
   if (!($37)) {
    ___cxa_throw(($35|0),(96|0),(344|0)); //@line 163 "modules/filesystem/physfs/File.cpp"
    // unreachable; //@line 163 "modules/filesystem/physfs/File.cpp"
   }
   $38 = ___cxa_find_matching_catch_2()|0;
   $39 = tempRet0;
   ___cxa_free_exception(($35|0)); //@line 163 "modules/filesystem/physfs/File.cpp"
   $$0 = $38;$$01 = $39;
   ___resumeException($$0|0);
   // unreachable;
  }
 }
 $8 = (___cxa_allocate_exception(16)|0); //@line 154 "modules/filesystem/physfs/File.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($8|0),(507241|0),($vararg_buffer|0)); //@line 154 "modules/filesystem/physfs/File.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if (!($10)) {
  ___cxa_throw(($8|0),(96|0),(344|0)); //@line 154 "modules/filesystem/physfs/File.cpp"
  // unreachable; //@line 154 "modules/filesystem/physfs/File.cpp"
 }
 $11 = ___cxa_find_matching_catch_2()|0;
 $12 = tempRet0;
 ___cxa_free_exception(($8|0)); //@line 154 "modules/filesystem/physfs/File.cpp"
 $$0 = $11;$$01 = $12;
 ___resumeException($$0|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love10filesystem6physfs4File5writeEPKvx($this,$data,$0,$1) {
 $this = $this|0;
 $data = $data|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $switch = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $2 = ((($this)) + 20|0); //@line 176 "modules/filesystem/physfs/File.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 176 "modules/filesystem/physfs/File.cpp"
 $4 = ($3|0)==(0|0); //@line 176 "modules/filesystem/physfs/File.cpp"
 if (!($4)) {
  $5 = ((($this)) + 24|0); //@line 176 "modules/filesystem/physfs/File.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 176 "modules/filesystem/physfs/File.cpp"
  $7 = $6 & -2; //@line 176 "modules/filesystem/physfs/File.cpp"
  $switch = ($7|0)==(2); //@line 176 "modules/filesystem/physfs/File.cpp"
  if ($switch) {
   $13 = ($1|0)>(0); //@line 180 "modules/filesystem/physfs/File.cpp"
   $14 = ($0>>>0)>(4294967295); //@line 180 "modules/filesystem/physfs/File.cpp"
   $15 = ($1|0)==(0); //@line 180 "modules/filesystem/physfs/File.cpp"
   $16 = $15 & $14; //@line 180 "modules/filesystem/physfs/File.cpp"
   $17 = $13 | $16; //@line 180 "modules/filesystem/physfs/File.cpp"
   $18 = $17 ? -1 : $0; //@line 180 "modules/filesystem/physfs/File.cpp"
   $19 = $17 ? 0 : $1; //@line 180 "modules/filesystem/physfs/File.cpp"
   $20 = ($19|0)<(0); //@line 182 "modules/filesystem/physfs/File.cpp"
   if ($20) {
    $21 = (___cxa_allocate_exception(16)|0); //@line 183 "modules/filesystem/physfs/File.cpp"
    __THREW__ = 0;
    invoke_viii(3349,($21|0),(507324|0),($vararg_buffer1|0)); //@line 183 "modules/filesystem/physfs/File.cpp"
    $22 = __THREW__; __THREW__ = 0;
    $23 = $22&1;
    if (!($23)) {
     ___cxa_throw(($21|0),(96|0),(344|0)); //@line 183 "modules/filesystem/physfs/File.cpp"
     // unreachable; //@line 183 "modules/filesystem/physfs/File.cpp"
    }
    $24 = ___cxa_find_matching_catch_2()|0;
    $25 = tempRet0;
    ___cxa_free_exception(($21|0)); //@line 183 "modules/filesystem/physfs/File.cpp"
    $$01 = $25;$$02 = $24;
    ___resumeException($$02|0);
    // unreachable;
   }
   $26 = (_PHYSFS_write($3,$data,1,$18)|0); //@line 189 "modules/filesystem/physfs/File.cpp"
   $27 = tempRet0; //@line 189 "modules/filesystem/physfs/File.cpp"
   $28 = ($26|0)==($18|0); //@line 193 "modules/filesystem/physfs/File.cpp"
   $29 = ($27|0)==($19|0); //@line 193 "modules/filesystem/physfs/File.cpp"
   $30 = $28 & $29; //@line 193 "modules/filesystem/physfs/File.cpp"
   if (!($30)) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0); //@line 214 "modules/filesystem/physfs/File.cpp"
   }
   $31 = ((($this)) + 28|0); //@line 197 "modules/filesystem/physfs/File.cpp"
   $32 = HEAP32[$31>>2]|0; //@line 197 "modules/filesystem/physfs/File.cpp"
   $33 = ($32|0)==(1); //@line 197 "modules/filesystem/physfs/File.cpp"
   if ($33) {
    $34 = ((($this)) + 32|0); //@line 197 "modules/filesystem/physfs/File.cpp"
    $35 = $34; //@line 197 "modules/filesystem/physfs/File.cpp"
    $36 = $35; //@line 197 "modules/filesystem/physfs/File.cpp"
    $37 = HEAP32[$36>>2]|0; //@line 197 "modules/filesystem/physfs/File.cpp"
    $38 = (($35) + 4)|0; //@line 197 "modules/filesystem/physfs/File.cpp"
    $39 = $38; //@line 197 "modules/filesystem/physfs/File.cpp"
    $40 = HEAP32[$39>>2]|0; //@line 197 "modules/filesystem/physfs/File.cpp"
    $41 = ($40|0)>($19|0); //@line 197 "modules/filesystem/physfs/File.cpp"
    $42 = ($37>>>0)>($18>>>0); //@line 197 "modules/filesystem/physfs/File.cpp"
    $43 = ($40|0)==($19|0); //@line 197 "modules/filesystem/physfs/File.cpp"
    $44 = $43 & $42; //@line 197 "modules/filesystem/physfs/File.cpp"
    $45 = $41 | $44; //@line 197 "modules/filesystem/physfs/File.cpp"
    if ($45) {
     $46 = (_memchr($data,10,$18)|0); //@line 199 "modules/filesystem/physfs/File.cpp"
     $47 = ($46|0)==(0|0); //@line 199 "modules/filesystem/physfs/File.cpp"
     if (!($47)) {
      $48 = HEAP32[$this>>2]|0; //@line 200 "modules/filesystem/physfs/File.cpp"
      $49 = ((($48)) + 48|0); //@line 200 "modules/filesystem/physfs/File.cpp"
      $50 = HEAP32[$49>>2]|0; //@line 200 "modules/filesystem/physfs/File.cpp"
      (FUNCTION_TABLE_ii[$50 & 8191]($this)|0); //@line 200 "modules/filesystem/physfs/File.cpp"
     }
    }
   }
   _emscripten_asm_const_v(0); //@line 204 "modules/filesystem/physfs/File.cpp"
   $$0 = 1;
   STACKTOP = sp;return ($$0|0); //@line 214 "modules/filesystem/physfs/File.cpp"
  }
 }
 $8 = (___cxa_allocate_exception(16)|0); //@line 177 "modules/filesystem/physfs/File.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($8|0),(507292|0),($vararg_buffer|0)); //@line 177 "modules/filesystem/physfs/File.cpp"
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if (!($10)) {
  ___cxa_throw(($8|0),(96|0),(344|0)); //@line 177 "modules/filesystem/physfs/File.cpp"
  // unreachable; //@line 177 "modules/filesystem/physfs/File.cpp"
 }
 $11 = ___cxa_find_matching_catch_2()|0;
 $12 = tempRet0;
 ___cxa_free_exception(($8|0)); //@line 177 "modules/filesystem/physfs/File.cpp"
 $$01 = $12;$$02 = $11;
 ___resumeException($$02|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love10filesystem6physfs4File5flushEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $switch = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($this)) + 20|0); //@line 218 "modules/filesystem/physfs/File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 218 "modules/filesystem/physfs/File.cpp"
 $2 = ($1|0)==(0|0); //@line 218 "modules/filesystem/physfs/File.cpp"
 if (!($2)) {
  $3 = ((($this)) + 24|0); //@line 218 "modules/filesystem/physfs/File.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 218 "modules/filesystem/physfs/File.cpp"
  $5 = $4 & -2; //@line 218 "modules/filesystem/physfs/File.cpp"
  $switch = ($5|0)==(2); //@line 218 "modules/filesystem/physfs/File.cpp"
  if ($switch) {
   $11 = (_PHYSFS_flush($1)|0); //@line 221 "modules/filesystem/physfs/File.cpp"
   $12 = ($11|0)!=(0); //@line 221 "modules/filesystem/physfs/File.cpp"
   STACKTOP = sp;return ($12|0); //@line 221 "modules/filesystem/physfs/File.cpp"
  }
 }
 $6 = (___cxa_allocate_exception(16)|0); //@line 219 "modules/filesystem/physfs/File.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($6|0),(507292|0),($vararg_buffer|0)); //@line 219 "modules/filesystem/physfs/File.cpp"
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  ___cxa_free_exception(($6|0)); //@line 219 "modules/filesystem/physfs/File.cpp"
  ___resumeException($9|0);
  // unreachable;
 } else {
  ___cxa_throw(($6|0),(96|0),(344|0)); //@line 219 "modules/filesystem/physfs/File.cpp"
  // unreachable; //@line 219 "modules/filesystem/physfs/File.cpp"
 }
 return (0)|0;
}
function __ZN4love10filesystem6physfs4File5isEOFEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 243 "modules/filesystem/physfs/File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 243 "modules/filesystem/physfs/File.cpp"
 $2 = ($1|0)==(0|0); //@line 243 "modules/filesystem/physfs/File.cpp"
 if ($2) {
  $5 = 1;
  return ($5|0); //@line 243 "modules/filesystem/physfs/File.cpp"
 }
 $3 = (_PHYSFS_eof($1)|0); //@line 237 "modules/filesystem/physfs/File.cpp"
 $4 = ($3|0)!=(0); //@line 237 "modules/filesystem/physfs/File.cpp"
 $5 = $4;
 return ($5|0); //@line 243 "modules/filesystem/physfs/File.cpp"
}
function __ZN4love10filesystem6physfs4File4tellEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0); //@line 248 "modules/filesystem/physfs/File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 248 "modules/filesystem/physfs/File.cpp"
 $2 = ($1|0)==(0|0); //@line 248 "modules/filesystem/physfs/File.cpp"
 if ($2) {
  $5 = -1;$6 = -1;
 } else {
  $3 = (_PHYSFS_tell($1)|0); //@line 251 "modules/filesystem/physfs/File.cpp"
  $4 = tempRet0; //@line 251 "modules/filesystem/physfs/File.cpp"
  $5 = $4;$6 = $3;
 }
 tempRet0 = ($5); //@line 252 "modules/filesystem/physfs/File.cpp"
 return ($6|0); //@line 252 "modules/filesystem/physfs/File.cpp"
}
function __ZN4love10filesystem6physfs4File4seekEy($this,$0,$1) {
 $this = $this|0;
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($this)) + 20|0); //@line 256 "modules/filesystem/physfs/File.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 256 "modules/filesystem/physfs/File.cpp"
 $4 = ($3|0)==(0|0); //@line 256 "modules/filesystem/physfs/File.cpp"
 if ($4) {
  $7 = 0;
  return ($7|0); //@line 256 "modules/filesystem/physfs/File.cpp"
 }
 $5 = (_PHYSFS_seek($3,$0,$1)|0); //@line 256 "modules/filesystem/physfs/File.cpp"
 $6 = ($5|0)!=(0); //@line 256 "modules/filesystem/physfs/File.cpp"
 $7 = $6;
 return ($7|0); //@line 256 "modules/filesystem/physfs/File.cpp"
}
function __ZN4love10filesystem6physfs4File9setBufferENS0_4File10BufferModeEx($this,$bufmode,$0,$1) {
 $this = $this|0;
 $bufmode = $bufmode|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$1 = 0, $$off = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ret$0 = 0, $switch = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($1|0)<(0); //@line 261 "modules/filesystem/physfs/File.cpp"
 if ($2) {
  $$1 = 0;
  return ($$1|0); //@line 295 "modules/filesystem/physfs/File.cpp"
 }
 $3 = HEAP32[$this>>2]|0; //@line 266 "modules/filesystem/physfs/File.cpp"
 $4 = ((($3)) + 24|0); //@line 266 "modules/filesystem/physfs/File.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 266 "modules/filesystem/physfs/File.cpp"
 $6 = (FUNCTION_TABLE_ii[$5 & 8191]($this)|0); //@line 266 "modules/filesystem/physfs/File.cpp"
 if (!($6)) {
  $7 = ((($this)) + 28|0); //@line 268 "modules/filesystem/physfs/File.cpp"
  HEAP32[$7>>2] = $bufmode; //@line 268 "modules/filesystem/physfs/File.cpp"
  $8 = ((($this)) + 32|0); //@line 269 "modules/filesystem/physfs/File.cpp"
  $9 = $8; //@line 269 "modules/filesystem/physfs/File.cpp"
  $10 = $9; //@line 269 "modules/filesystem/physfs/File.cpp"
  HEAP32[$10>>2] = $0; //@line 269 "modules/filesystem/physfs/File.cpp"
  $11 = (($9) + 4)|0; //@line 269 "modules/filesystem/physfs/File.cpp"
  $12 = $11; //@line 269 "modules/filesystem/physfs/File.cpp"
  HEAP32[$12>>2] = $1; //@line 269 "modules/filesystem/physfs/File.cpp"
  $$1 = 1;
  return ($$1|0); //@line 295 "modules/filesystem/physfs/File.cpp"
 }
 $$off = (($bufmode) + -1)|0; //@line 275 "modules/filesystem/physfs/File.cpp"
 $switch = ($$off>>>0)<(2); //@line 275 "modules/filesystem/physfs/File.cpp"
 $13 = ((($this)) + 20|0); //@line 284 "modules/filesystem/physfs/File.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 284 "modules/filesystem/physfs/File.cpp"
 if ($switch) {
  $16 = (_PHYSFS_setBuffer($14,$0,$1)|0); //@line 284 "modules/filesystem/physfs/File.cpp"
  $22 = $0;$25 = $1;$ret$0 = $16;
 } else {
  $15 = (_PHYSFS_setBuffer($14,0,0)|0); //@line 279 "modules/filesystem/physfs/File.cpp"
  $22 = 0;$25 = 0;$ret$0 = $15;
 }
 $17 = ($ret$0|0)==(0); //@line 288 "modules/filesystem/physfs/File.cpp"
 if ($17) {
  $$1 = 0;
  return ($$1|0); //@line 295 "modules/filesystem/physfs/File.cpp"
 }
 $18 = ((($this)) + 28|0); //@line 291 "modules/filesystem/physfs/File.cpp"
 HEAP32[$18>>2] = $bufmode; //@line 291 "modules/filesystem/physfs/File.cpp"
 $19 = ((($this)) + 32|0); //@line 292 "modules/filesystem/physfs/File.cpp"
 $20 = $19; //@line 292 "modules/filesystem/physfs/File.cpp"
 $21 = $20; //@line 292 "modules/filesystem/physfs/File.cpp"
 HEAP32[$21>>2] = $22; //@line 292 "modules/filesystem/physfs/File.cpp"
 $23 = (($20) + 4)|0; //@line 292 "modules/filesystem/physfs/File.cpp"
 $24 = $23; //@line 292 "modules/filesystem/physfs/File.cpp"
 HEAP32[$24>>2] = $25; //@line 292 "modules/filesystem/physfs/File.cpp"
 $$1 = 1;
 return ($$1|0); //@line 295 "modules/filesystem/physfs/File.cpp"
}
function __ZNK4love10filesystem6physfs4File9getBufferERx($this,$size) {
 $this = $this|0;
 $size = $size|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0); //@line 299 "modules/filesystem/physfs/File.cpp"
 $1 = $0; //@line 299 "modules/filesystem/physfs/File.cpp"
 $2 = $1; //@line 299 "modules/filesystem/physfs/File.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 299 "modules/filesystem/physfs/File.cpp"
 $4 = (($1) + 4)|0; //@line 299 "modules/filesystem/physfs/File.cpp"
 $5 = $4; //@line 299 "modules/filesystem/physfs/File.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 299 "modules/filesystem/physfs/File.cpp"
 $7 = $size; //@line 299 "modules/filesystem/physfs/File.cpp"
 $8 = $7; //@line 299 "modules/filesystem/physfs/File.cpp"
 HEAP32[$8>>2] = $3; //@line 299 "modules/filesystem/physfs/File.cpp"
 $9 = (($7) + 4)|0; //@line 299 "modules/filesystem/physfs/File.cpp"
 $10 = $9; //@line 299 "modules/filesystem/physfs/File.cpp"
 HEAP32[$10>>2] = $6; //@line 299 "modules/filesystem/physfs/File.cpp"
 $11 = ((($this)) + 28|0); //@line 300 "modules/filesystem/physfs/File.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 300 "modules/filesystem/physfs/File.cpp"
 return ($12|0); //@line 300 "modules/filesystem/physfs/File.cpp"
}
function __ZNK4love10filesystem6physfs4File7getModeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0); //@line 310 "modules/filesystem/physfs/File.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 310 "modules/filesystem/physfs/File.cpp"
 return ($1|0); //@line 310 "modules/filesystem/physfs/File.cpp"
}
function __ZNK4love10filesystem6physfs4File11getFilenameEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0); //@line 305 "modules/filesystem/physfs/File.cpp"
 return ($0|0); //@line 305 "modules/filesystem/physfs/File.cpp"
}
function __ZN4love10filesystem6physfs4FileC2ERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE($this,$filename) {
 $this = $this|0;
 $filename = $filename|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 43 "./modules/filesystem/File.h"
 HEAP32[$this>>2] = (35060); //@line 51 "modules/filesystem/physfs/File.cpp"
 $0 = ((($this)) + 8|0); //@line 46 "modules/filesystem/physfs/File.cpp"
 __THREW__ = 0;
 invoke_vii(3351,($0|0),($filename|0)); //@line 46 "modules/filesystem/physfs/File.cpp"
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  __ZN4love10filesystem4FileD2Ev($this); //@line 52 "modules/filesystem/physfs/File.cpp"
  ___resumeException($4|0);
  // unreachable;
 } else {
  $3 = ((($this)) + 20|0); //@line 47 "modules/filesystem/physfs/File.cpp"
  ;HEAP32[$3>>2]=0|0;HEAP32[$3+4>>2]=0|0;HEAP32[$3+8>>2]=0|0;HEAP32[$3+12>>2]=0|0;HEAP32[$3+16>>2]=0|0; //@line 48 "modules/filesystem/physfs/File.cpp"
  return; //@line 52 "modules/filesystem/physfs/File.cpp"
 }
}
function __ZN4love5image6magpie11ASTCHandlerD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 42 "modules/image/magpie/ASTCHandler.h"
 return; //@line 42 "modules/image/magpie/ASTCHandler.h"
}
function __ZN4love5image6magpie11ASTCHandler8canParseEPKNS_10filesystem8FileDataE($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$data>>2]|0; //@line 91 "modules/image/magpie/ASTCHandler.cpp"
 $1 = ((($0)) + 20|0); //@line 91 "modules/image/magpie/ASTCHandler.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 91 "modules/image/magpie/ASTCHandler.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($data)|0); //@line 91 "modules/image/magpie/ASTCHandler.cpp"
 $4 = ($3>>>0)<(17); //@line 91 "modules/image/magpie/ASTCHandler.cpp"
 if ($4) {
  $$1 = 0;
  return ($$1|0); //@line 105 "modules/image/magpie/ASTCHandler.cpp"
 }
 $5 = HEAP32[$data>>2]|0; //@line 94 "modules/image/magpie/ASTCHandler.cpp"
 $6 = ((($5)) + 16|0); //@line 94 "modules/image/magpie/ASTCHandler.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 94 "modules/image/magpie/ASTCHandler.cpp"
 $8 = (FUNCTION_TABLE_ii[$7 & 8191]($data)|0); //@line 94 "modules/image/magpie/ASTCHandler.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 96 "modules/image/magpie/ASTCHandler.cpp"
 $10 = $9&255; //@line 96 "modules/image/magpie/ASTCHandler.cpp"
 $11 = ((($8)) + 1|0); //@line 97 "modules/image/magpie/ASTCHandler.cpp"
 $12 = HEAP8[$11>>0]|0; //@line 97 "modules/image/magpie/ASTCHandler.cpp"
 $13 = $12&255; //@line 97 "modules/image/magpie/ASTCHandler.cpp"
 $14 = $13 << 8; //@line 97 "modules/image/magpie/ASTCHandler.cpp"
 $15 = $14 | $10; //@line 97 "modules/image/magpie/ASTCHandler.cpp"
 $16 = ((($8)) + 2|0); //@line 98 "modules/image/magpie/ASTCHandler.cpp"
 $17 = HEAP8[$16>>0]|0; //@line 98 "modules/image/magpie/ASTCHandler.cpp"
 $18 = $17&255; //@line 98 "modules/image/magpie/ASTCHandler.cpp"
 $19 = $18 << 16; //@line 98 "modules/image/magpie/ASTCHandler.cpp"
 $20 = $15 | $19; //@line 98 "modules/image/magpie/ASTCHandler.cpp"
 $21 = ((($8)) + 3|0); //@line 99 "modules/image/magpie/ASTCHandler.cpp"
 $22 = HEAP8[$21>>0]|0; //@line 99 "modules/image/magpie/ASTCHandler.cpp"
 $23 = $22&255; //@line 99 "modules/image/magpie/ASTCHandler.cpp"
 $24 = $23 << 24; //@line 99 "modules/image/magpie/ASTCHandler.cpp"
 $25 = $20 | $24; //@line 99 "modules/image/magpie/ASTCHandler.cpp"
 $26 = ($25|0)==(1554098963); //@line 101 "modules/image/magpie/ASTCHandler.cpp"
 $$1 = $26;
 return ($$1|0); //@line 105 "modules/image/magpie/ASTCHandler.cpp"
}
function __ZN4love5image6magpie11ASTCHandler5parseEPNS_10filesystem8FileDataERNSt3__16vectorINS0_19CompressedImageData8SubImageENS6_9allocatorIS9_EEEERjRNS8_6FormatERb($this,$filedata,$images,$dataSize,$format,$sRGB) {
 $this = $this|0;
 $filedata = $filedata|0;
 $images = $images|0;
 $dataSize = $dataSize|0;
 $format = $format|0;
 $sRGB = $sRGB|0;
 var $$0 = 0, $$0$i$ph = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $mip = 0, $or$cond$i = 0, $or$cond11$i = 0, $or$cond13$i = 0, $or$cond15$i = 0, $or$cond17$i = 0, $or$cond19$i = 0, $or$cond21$i = 0, $or$cond23$i = 0, $or$cond25$i = 0, $or$cond27$i = 0, $or$cond3$i = 0, $or$cond5$i = 0, $or$cond7$i = 0, $or$cond9$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0;
 var $vararg_ptr3 = 0, $vararg_ptr4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $mip = sp + 40|0;
 $0 = HEAP32[$this>>2]|0; //@line 109 "modules/image/magpie/ASTCHandler.cpp"
 $1 = ((($0)) + 16|0); //@line 109 "modules/image/magpie/ASTCHandler.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 109 "modules/image/magpie/ASTCHandler.cpp"
 $3 = (FUNCTION_TABLE_iii[$2 & 8191]($this,$filedata)|0); //@line 109 "modules/image/magpie/ASTCHandler.cpp"
 if (!($3)) {
  $4 = (___cxa_allocate_exception(16)|0); //@line 110 "modules/image/magpie/ASTCHandler.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($4|0),(507447|0),($vararg_buffer|0)); //@line 110 "modules/image/magpie/ASTCHandler.cpp"
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   ___cxa_throw(($4|0),(96|0),(344|0)); //@line 110 "modules/image/magpie/ASTCHandler.cpp"
   // unreachable; //@line 110 "modules/image/magpie/ASTCHandler.cpp"
  }
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  ___cxa_free_exception(($4|0)); //@line 110 "modules/image/magpie/ASTCHandler.cpp"
  $$1 = $7;$$12 = $8;
  ___resumeException($$1|0);
  // unreachable;
 }
 $9 = HEAP32[$filedata>>2]|0; //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $10 = ((($9)) + 16|0); //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($filedata)|0); //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $13 = ((($12)) + 4|0); //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $14 = HEAP8[$13>>0]|0; //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $15 = ((($12)) + 5|0); //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $16 = HEAP8[$15>>0]|0; //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $17 = ((($12)) + 6|0); //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $18 = HEAP8[$17>>0]|0; //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $19 = ((($12)) + 7|0); //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $20 = HEAP8[$19>>0]|0; //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $21 = ((($12)) + 8|0); //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $22 = HEAP8[$21>>0]|0; //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $23 = ((($12)) + 9|0); //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $24 = HEAP8[$23>>0]|0; //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $25 = ((($12)) + 10|0); //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $26 = HEAP8[$25>>0]|0; //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $27 = ((($12)) + 11|0); //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $28 = HEAP8[$27>>0]|0; //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $29 = ((($12)) + 12|0); //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $30 = HEAP8[$29>>0]|0; //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $31 = ((($12)) + 13|0); //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $32 = HEAP8[$31>>0]|0; //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $33 = ((($12)) + 14|0); //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $34 = HEAP8[$33>>0]|0; //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $35 = ((($12)) + 15|0); //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $36 = HEAP8[$35>>0]|0; //@line 112 "modules/image/magpie/ASTCHandler.cpp"
 $37 = $14&255; //@line 114 "modules/image/magpie/ASTCHandler.cpp"
 $38 = $16&255; //@line 114 "modules/image/magpie/ASTCHandler.cpp"
 $39 = $18&255; //@line 114 "modules/image/magpie/ASTCHandler.cpp"
 $40 = ($18&255)>(1); //@line 52 "modules/image/magpie/ASTCHandler.cpp"
 do {
  if (!($40)) {
   $41 = ($14<<24>>24)==(4); //@line 55 "modules/image/magpie/ASTCHandler.cpp"
   $42 = ($16<<24>>24)==(4); //@line 55 "modules/image/magpie/ASTCHandler.cpp"
   $or$cond$i = $41 & $42; //@line 55 "modules/image/magpie/ASTCHandler.cpp"
   if ($or$cond$i) {
    $$0$i$ph = 23;
   } else {
    $43 = ($14<<24>>24)==(5); //@line 57 "modules/image/magpie/ASTCHandler.cpp"
    $or$cond3$i = $43 & $42; //@line 57 "modules/image/magpie/ASTCHandler.cpp"
    if ($or$cond3$i) {
     $$0$i$ph = 24;
    } else {
     $44 = ($16<<24>>24)==(5); //@line 59 "modules/image/magpie/ASTCHandler.cpp"
     $or$cond5$i = $43 & $44; //@line 59 "modules/image/magpie/ASTCHandler.cpp"
     if ($or$cond5$i) {
      $$0$i$ph = 25;
     } else {
      $45 = ($14<<24>>24)==(6); //@line 61 "modules/image/magpie/ASTCHandler.cpp"
      $or$cond7$i = $45 & $44; //@line 61 "modules/image/magpie/ASTCHandler.cpp"
      if ($or$cond7$i) {
       $$0$i$ph = 26;
      } else {
       $46 = ($16<<24>>24)==(6); //@line 63 "modules/image/magpie/ASTCHandler.cpp"
       $or$cond9$i = $45 & $46; //@line 63 "modules/image/magpie/ASTCHandler.cpp"
       if ($or$cond9$i) {
        $$0$i$ph = 27;
       } else {
        $47 = ($14<<24>>24)==(8); //@line 65 "modules/image/magpie/ASTCHandler.cpp"
        $or$cond11$i = $47 & $44; //@line 65 "modules/image/magpie/ASTCHandler.cpp"
        if ($or$cond11$i) {
         $$0$i$ph = 28;
        } else {
         $or$cond13$i = $47 & $46; //@line 67 "modules/image/magpie/ASTCHandler.cpp"
         if ($or$cond13$i) {
          $$0$i$ph = 29;
         } else {
          $48 = ($16<<24>>24)==(8); //@line 69 "modules/image/magpie/ASTCHandler.cpp"
          $or$cond15$i = $47 & $48; //@line 69 "modules/image/magpie/ASTCHandler.cpp"
          if ($or$cond15$i) {
           $$0$i$ph = 30;
          } else {
           $49 = ($14<<24>>24)==(10); //@line 71 "modules/image/magpie/ASTCHandler.cpp"
           $or$cond17$i = $49 & $44; //@line 71 "modules/image/magpie/ASTCHandler.cpp"
           if ($or$cond17$i) {
            $$0$i$ph = 31;
           } else {
            $or$cond19$i = $49 & $46; //@line 73 "modules/image/magpie/ASTCHandler.cpp"
            if ($or$cond19$i) {
             $$0$i$ph = 32;
            } else {
             $or$cond21$i = $49 & $48; //@line 75 "modules/image/magpie/ASTCHandler.cpp"
             if ($or$cond21$i) {
              $$0$i$ph = 33;
             } else {
              $50 = ($16<<24>>24)==(10); //@line 77 "modules/image/magpie/ASTCHandler.cpp"
              $or$cond23$i = $49 & $50; //@line 77 "modules/image/magpie/ASTCHandler.cpp"
              if ($or$cond23$i) {
               $$0$i$ph = 34;
              } else {
               $51 = ($14<<24>>24)==(12); //@line 79 "modules/image/magpie/ASTCHandler.cpp"
               $or$cond25$i = $51 & $50; //@line 79 "modules/image/magpie/ASTCHandler.cpp"
               if ($or$cond25$i) {
                $$0$i$ph = 35;
               } else {
                $52 = ($16<<24>>24)==(12); //@line 81 "modules/image/magpie/ASTCHandler.cpp"
                $or$cond27$i = $51 & $52; //@line 81 "modules/image/magpie/ASTCHandler.cpp"
                if ($or$cond27$i) {
                 $$0$i$ph = 36;
                } else {
                 break;
                }
               }
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    }
   }
   $58 = $20&255; //@line 119 "modules/image/magpie/ASTCHandler.cpp"
   $59 = $22&255; //@line 119 "modules/image/magpie/ASTCHandler.cpp"
   $60 = $59 << 8; //@line 119 "modules/image/magpie/ASTCHandler.cpp"
   $61 = $60 | $58; //@line 119 "modules/image/magpie/ASTCHandler.cpp"
   $62 = $24&255; //@line 119 "modules/image/magpie/ASTCHandler.cpp"
   $63 = $62 << 16; //@line 119 "modules/image/magpie/ASTCHandler.cpp"
   $64 = $61 | $63; //@line 119 "modules/image/magpie/ASTCHandler.cpp"
   $65 = $26&255; //@line 120 "modules/image/magpie/ASTCHandler.cpp"
   $66 = $28&255; //@line 120 "modules/image/magpie/ASTCHandler.cpp"
   $67 = $66 << 8; //@line 120 "modules/image/magpie/ASTCHandler.cpp"
   $68 = $67 | $65; //@line 120 "modules/image/magpie/ASTCHandler.cpp"
   $69 = $30&255; //@line 120 "modules/image/magpie/ASTCHandler.cpp"
   $70 = $69 << 16; //@line 120 "modules/image/magpie/ASTCHandler.cpp"
   $71 = $68 | $70; //@line 120 "modules/image/magpie/ASTCHandler.cpp"
   $72 = $32&255; //@line 121 "modules/image/magpie/ASTCHandler.cpp"
   $73 = $34&255; //@line 121 "modules/image/magpie/ASTCHandler.cpp"
   $74 = $73 << 8; //@line 121 "modules/image/magpie/ASTCHandler.cpp"
   $75 = $74 | $72; //@line 121 "modules/image/magpie/ASTCHandler.cpp"
   $76 = $36&255; //@line 121 "modules/image/magpie/ASTCHandler.cpp"
   $77 = $76 << 16; //@line 121 "modules/image/magpie/ASTCHandler.cpp"
   $78 = $75 | $77; //@line 121 "modules/image/magpie/ASTCHandler.cpp"
   $79 = (($37) + -1)|0; //@line 123 "modules/image/magpie/ASTCHandler.cpp"
   $80 = (($79) + ($64))|0; //@line 123 "modules/image/magpie/ASTCHandler.cpp"
   $81 = (($80>>>0) / ($37>>>0))&-1; //@line 123 "modules/image/magpie/ASTCHandler.cpp"
   $82 = (($38) + -1)|0; //@line 124 "modules/image/magpie/ASTCHandler.cpp"
   $83 = (($82) + ($71))|0; //@line 124 "modules/image/magpie/ASTCHandler.cpp"
   $84 = (($83>>>0) / ($38>>>0))&-1; //@line 124 "modules/image/magpie/ASTCHandler.cpp"
   $85 = (($39) + -1)|0; //@line 125 "modules/image/magpie/ASTCHandler.cpp"
   $86 = (($85) + ($78))|0; //@line 125 "modules/image/magpie/ASTCHandler.cpp"
   $87 = (($86>>>0) / ($39>>>0))&-1; //@line 125 "modules/image/magpie/ASTCHandler.cpp"
   $88 = $81 << 4; //@line 127 "modules/image/magpie/ASTCHandler.cpp"
   $89 = Math_imul($88, $84)|0; //@line 127 "modules/image/magpie/ASTCHandler.cpp"
   $90 = Math_imul($89, $87)|0; //@line 127 "modules/image/magpie/ASTCHandler.cpp"
   $91 = (($90) + 16)|0; //@line 129 "modules/image/magpie/ASTCHandler.cpp"
   $92 = HEAP32[$filedata>>2]|0; //@line 129 "modules/image/magpie/ASTCHandler.cpp"
   $93 = ((($92)) + 20|0); //@line 129 "modules/image/magpie/ASTCHandler.cpp"
   $94 = HEAP32[$93>>2]|0; //@line 129 "modules/image/magpie/ASTCHandler.cpp"
   $95 = (FUNCTION_TABLE_ii[$94 & 8191]($filedata)|0); //@line 129 "modules/image/magpie/ASTCHandler.cpp"
   $96 = ($91>>>0)>($95>>>0); //@line 129 "modules/image/magpie/ASTCHandler.cpp"
   if ($96) {
    $97 = (___cxa_allocate_exception(16)|0); //@line 130 "modules/image/magpie/ASTCHandler.cpp"
    __THREW__ = 0;
    invoke_viii(3349,($97|0),(507563|0),($vararg_buffer5|0)); //@line 130 "modules/image/magpie/ASTCHandler.cpp"
    $98 = __THREW__; __THREW__ = 0;
    $99 = $98&1;
    if (!($99)) {
     ___cxa_throw(($97|0),(96|0),(344|0)); //@line 130 "modules/image/magpie/ASTCHandler.cpp"
     // unreachable; //@line 130 "modules/image/magpie/ASTCHandler.cpp"
    }
    $100 = ___cxa_find_matching_catch_2()|0;
    $101 = tempRet0;
    ___cxa_free_exception(($97|0)); //@line 130 "modules/image/magpie/ASTCHandler.cpp"
    $$1 = $100;$$12 = $101;
    ___resumeException($$1|0);
    // unreachable;
   }
   __THREW__ = 0;
   $102 = (invoke_ii(3492,($90|0))|0); //@line 136 "modules/image/magpie/ASTCHandler.cpp"
   $103 = __THREW__; __THREW__ = 0;
   $104 = $103&1;
   if (!($104)) {
    $105 = HEAP32[$filedata>>2]|0; //@line 144 "modules/image/magpie/ASTCHandler.cpp"
    $106 = ((($105)) + 16|0); //@line 144 "modules/image/magpie/ASTCHandler.cpp"
    $107 = HEAP32[$106>>2]|0; //@line 144 "modules/image/magpie/ASTCHandler.cpp"
    $108 = (FUNCTION_TABLE_ii[$107 & 8191]($filedata)|0); //@line 144 "modules/image/magpie/ASTCHandler.cpp"
    $109 = ((($108)) + 16|0); //@line 144 "modules/image/magpie/ASTCHandler.cpp"
    _memcpy(($102|0),($109|0),($90|0))|0; //@line 144 "modules/image/magpie/ASTCHandler.cpp"
    HEAP32[$mip>>2] = $64; //@line 148 "modules/image/magpie/ASTCHandler.cpp"
    $110 = ((($mip)) + 4|0); //@line 149 "modules/image/magpie/ASTCHandler.cpp"
    HEAP32[$110>>2] = $71; //@line 149 "modules/image/magpie/ASTCHandler.cpp"
    $111 = ((($mip)) + 8|0); //@line 151 "modules/image/magpie/ASTCHandler.cpp"
    HEAP32[$111>>2] = $90; //@line 151 "modules/image/magpie/ASTCHandler.cpp"
    $112 = ((($mip)) + 12|0); //@line 152 "modules/image/magpie/ASTCHandler.cpp"
    HEAP32[$112>>2] = $102; //@line 152 "modules/image/magpie/ASTCHandler.cpp"
    $113 = ((($images)) + 4|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $114 = HEAP32[$113>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $115 = ((($images)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
    $116 = HEAP32[$115>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    $117 = ($114|0)==($116|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
    if ($117) {
     __ZNSt3__16vectorIN4love5image19CompressedImageData8SubImageENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_($images,$mip); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$dataSize>>2] = $90; //@line 156 "modules/image/magpie/ASTCHandler.cpp"
     HEAP32[$format>>2] = $$0$i$ph; //@line 157 "modules/image/magpie/ASTCHandler.cpp"
     HEAP8[$sRGB>>0] = 0; //@line 158 "modules/image/magpie/ASTCHandler.cpp"
     STACKTOP = sp;return ($102|0); //@line 160 "modules/image/magpie/ASTCHandler.cpp"
    } else {
     ;HEAP32[$114>>2]=HEAP32[$mip>>2]|0;HEAP32[$114+4>>2]=HEAP32[$mip+4>>2]|0;HEAP32[$114+8>>2]=HEAP32[$mip+8>>2]|0;HEAP32[$114+12>>2]=HEAP32[$mip+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
     $118 = HEAP32[$113>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $119 = ((($118)) + 16|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$113>>2] = $119; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     HEAP32[$dataSize>>2] = $90; //@line 156 "modules/image/magpie/ASTCHandler.cpp"
     HEAP32[$format>>2] = $$0$i$ph; //@line 157 "modules/image/magpie/ASTCHandler.cpp"
     HEAP8[$sRGB>>0] = 0; //@line 158 "modules/image/magpie/ASTCHandler.cpp"
     STACKTOP = sp;return ($102|0); //@line 160 "modules/image/magpie/ASTCHandler.cpp"
    }
   }
   $120 = ___cxa_find_matching_catch_3(9056|0)|0;
   $121 = tempRet0;
   $122 = (_llvm_eh_typeid_for((9056|0))|0); //@line 137 "modules/image/magpie/ASTCHandler.cpp"
   $123 = ($121|0)==($122|0); //@line 137 "modules/image/magpie/ASTCHandler.cpp"
   if (!($123)) {
    $$1 = $120;$$12 = $121;
    ___resumeException($$1|0);
    // unreachable;
   }
   (___cxa_begin_catch(($120|0))|0); //@line 137 "modules/image/magpie/ASTCHandler.cpp"
   $124 = (___cxa_allocate_exception(16)|0); //@line 140 "modules/image/magpie/ASTCHandler.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($124|0),(507610|0),($vararg_buffer7|0)); //@line 140 "modules/image/magpie/ASTCHandler.cpp"
   $125 = __THREW__; __THREW__ = 0;
   $126 = $125&1;
   if ($126) {
    $130 = ___cxa_find_matching_catch_2()|0;
    $131 = tempRet0;
    ___cxa_free_exception(($124|0)); //@line 140 "modules/image/magpie/ASTCHandler.cpp"
    $$0 = $130;$$01 = $131;
   } else {
    __THREW__ = 0;
    invoke_viii(3350,($124|0),(96|0),(344|0)); //@line 140 "modules/image/magpie/ASTCHandler.cpp"
    $127 = __THREW__; __THREW__ = 0;
    $128 = ___cxa_find_matching_catch_2()|0;
    $129 = tempRet0;
    $$0 = $128;$$01 = $129;
   }
   __THREW__ = 0;
   invoke_v(3378); //@line 141 "modules/image/magpie/ASTCHandler.cpp"
   $132 = __THREW__; __THREW__ = 0;
   $133 = $132&1;
   if ($133) {
    $134 = ___cxa_find_matching_catch_3(0|0)|0;
    $135 = tempRet0;
    ___clang_call_terminate($134); //@line 141 "modules/image/magpie/ASTCHandler.cpp"
    // unreachable; //@line 141 "modules/image/magpie/ASTCHandler.cpp"
   } else {
    $$1 = $$0;$$12 = $$01;
    ___resumeException($$1|0);
    // unreachable;
   }
  }
 } while(0);
 $53 = (___cxa_allocate_exception(16)|0); //@line 117 "modules/image/magpie/ASTCHandler.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer1>>2] = $37; //@line 117 "modules/image/magpie/ASTCHandler.cpp"
 $vararg_ptr3 = ((($vararg_buffer1)) + 4|0); //@line 117 "modules/image/magpie/ASTCHandler.cpp"
 HEAP32[$vararg_ptr3>>2] = $38; //@line 117 "modules/image/magpie/ASTCHandler.cpp"
 $vararg_ptr4 = ((($vararg_buffer1)) + 8|0); //@line 117 "modules/image/magpie/ASTCHandler.cpp"
 HEAP32[$vararg_ptr4>>2] = $39; //@line 117 "modules/image/magpie/ASTCHandler.cpp"
 invoke_viii(3349,($53|0),(507501|0),($vararg_buffer1|0)); //@line 117 "modules/image/magpie/ASTCHandler.cpp"
 $54 = __THREW__; __THREW__ = 0;
 $55 = $54&1;
 if (!($55)) {
  ___cxa_throw(($53|0),(96|0),(344|0)); //@line 117 "modules/image/magpie/ASTCHandler.cpp"
  // unreachable; //@line 117 "modules/image/magpie/ASTCHandler.cpp"
 }
 $56 = ___cxa_find_matching_catch_2()|0;
 $57 = tempRet0;
 ___cxa_free_exception(($53|0)); //@line 117 "modules/image/magpie/ASTCHandler.cpp"
 $$1 = $56;$$12 = $57;
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love5image6magpie19CompressedImageDataD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (35184); //@line 65 "modules/image/magpie/CompressedImageData.cpp"
 $0 = ((($this)) + 16|0); //@line 66 "modules/image/magpie/CompressedImageData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 66 "modules/image/magpie/CompressedImageData.cpp"
 $2 = ($1|0)==(0|0); //@line 66 "modules/image/magpie/CompressedImageData.cpp"
 if (!($2)) {
  __ZdaPv($1); //@line 66 "modules/image/magpie/CompressedImageData.cpp"
 }
 __ZN4love5image19CompressedImageDataD2Ev($this); //@line 67 "modules/image/magpie/CompressedImageData.cpp"
 return; //@line 67 "modules/image/magpie/CompressedImageData.cpp"
}
function __ZN4love5image6magpie19CompressedImageDataD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (35184); //@line 65 "modules/image/magpie/CompressedImageData.cpp"
 $0 = ((($this)) + 16|0); //@line 66 "modules/image/magpie/CompressedImageData.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 66 "modules/image/magpie/CompressedImageData.cpp"
 $2 = ($1|0)==(0|0); //@line 66 "modules/image/magpie/CompressedImageData.cpp"
 if (!($2)) {
  __ZdaPv($1); //@line 66 "modules/image/magpie/CompressedImageData.cpp"
 }
 __ZN4love5image19CompressedImageDataD2Ev($this); //@line 67 "modules/image/magpie/CompressedImageData.cpp"
 __ZdlPv($this); //@line 65 "modules/image/magpie/CompressedImageData.cpp"
 return; //@line 67 "modules/image/magpie/CompressedImageData.cpp"
}
function __ZN4love5image6magpie19CompressedImageDataC2ENSt3__14listIPNS1_23CompressedFormatHandlerENS3_9allocatorIS6_EEEEPNS_10filesystem8FileDataE($this,$formats,$filedata) {
 $this = $this|0;
 $formats = $formats|0;
 $filedata = $filedata|0;
 var $$01 = 0, $$02 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $__begin$sroa$0$0 = 0, $__begin$sroa$0$04 = 0, $__begin$sroa$0$06 = 0, $lpad$phi$index = 0, $lpad$phi$index8 = 0, $parser$2 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 __ZN4love5image19CompressedImageDataC2Ev($this); //@line 30 "modules/image/magpie/CompressedImageData.cpp"
 HEAP32[$this>>2] = (35184); //@line 31 "modules/image/magpie/CompressedImageData.cpp"
 $0 = ((($formats)) + 4|0); //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $__begin$sroa$0$04 = HEAP32[$0>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $1 = $__begin$sroa$0$04; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 $2 = ($1|0)==($formats|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
 do {
  if ($2) {
   label = 9;
  } else {
   $8 = $1;$__begin$sroa$0$06 = $__begin$sroa$0$04;
   while(1) {
    $7 = ((($8)) + 8|0); //@line 318 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $9 = HEAP32[$7>>2]|0; //@line 34 "modules/image/magpie/CompressedImageData.cpp"
    $10 = HEAP32[$9>>2]|0; //@line 36 "modules/image/magpie/CompressedImageData.cpp"
    $11 = ((($10)) + 16|0); //@line 36 "modules/image/magpie/CompressedImageData.cpp"
    $12 = HEAP32[$11>>2]|0; //@line 36 "modules/image/magpie/CompressedImageData.cpp"
    __THREW__ = 0;
    $13 = (invoke_iii($12|0,($9|0),($filedata|0))|0); //@line 36 "modules/image/magpie/CompressedImageData.cpp"
    $14 = __THREW__; __THREW__ = 0;
    $15 = $14&1;
    if ($15) {
     label = 2;
     break;
    }
    if ($13) {
     $parser$2 = $9;
     break;
    }
    $16 = $__begin$sroa$0$06; //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $17 = ((($16)) + 4|0); //@line 337 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $__begin$sroa$0$0 = HEAP32[$17>>2]|0; //@line 572 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $18 = $__begin$sroa$0$0; //@line 359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    $19 = ($18|0)==($formats|0); //@line 363 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/list"
    if ($19) {
     $parser$2 = 0;
     break;
    } else {
     $8 = $18;$__begin$sroa$0$06 = $__begin$sroa$0$0;
    }
   }
   if ((label|0) == 2) {
    $3 = ___cxa_find_matching_catch_2()|0;
    $4 = tempRet0;
    $lpad$phi$index = $3;$lpad$phi$index8 = $4;
    break;
   }
   $20 = ($parser$2|0)==(0|0); //@line 43 "modules/image/magpie/CompressedImageData.cpp"
   if ($20) {
    label = 9;
   } else {
    $27 = HEAP32[$parser$2>>2]|0; //@line 46 "modules/image/magpie/CompressedImageData.cpp"
    $28 = ((($27)) + 20|0); //@line 46 "modules/image/magpie/CompressedImageData.cpp"
    $29 = HEAP32[$28>>2]|0; //@line 46 "modules/image/magpie/CompressedImageData.cpp"
    $30 = ((($this)) + 24|0); //@line 46 "modules/image/magpie/CompressedImageData.cpp"
    $31 = ((($this)) + 20|0); //@line 46 "modules/image/magpie/CompressedImageData.cpp"
    $32 = ((($this)) + 8|0); //@line 46 "modules/image/magpie/CompressedImageData.cpp"
    $33 = ((($this)) + 12|0); //@line 46 "modules/image/magpie/CompressedImageData.cpp"
    __THREW__ = 0;
    $34 = (invoke_iiiiiii($29|0,($parser$2|0),($filedata|0),($30|0),($31|0),($32|0),($33|0))|0); //@line 59 "modules/image/magpie/CompressedImageData.cpp"
    $35 = __THREW__; __THREW__ = 0;
    $36 = $35&1;
    if ($36) {
     label = 3;
    } else {
     $37 = ((($this)) + 16|0); //@line 46 "modules/image/magpie/CompressedImageData.cpp"
     HEAP32[$37>>2] = $34; //@line 46 "modules/image/magpie/CompressedImageData.cpp"
     $38 = ($34|0)==(0|0); //@line 48 "modules/image/magpie/CompressedImageData.cpp"
     if ($38) {
      $39 = (___cxa_allocate_exception(16)|0); //@line 49 "modules/image/magpie/CompressedImageData.cpp"
      __THREW__ = 0;
      invoke_viii(3349,($39|0),(507716|0),($vararg_buffer1|0)); //@line 49 "modules/image/magpie/CompressedImageData.cpp"
      $40 = __THREW__; __THREW__ = 0;
      $41 = $40&1;
      if (!($41)) {
       __THREW__ = 0;
       invoke_viii(3350,($39|0),(96|0),(344|0)); //@line 49 "modules/image/magpie/CompressedImageData.cpp"
       $42 = __THREW__; __THREW__ = 0;
       label = 3;
       break;
      }
      $43 = ___cxa_find_matching_catch_2()|0;
      $44 = tempRet0;
      ___cxa_free_exception(($39|0)); //@line 49 "modules/image/magpie/CompressedImageData.cpp"
      $$01 = $44;$$02 = $43;
      __ZN4love5image19CompressedImageDataD2Ev($this); //@line 62 "modules/image/magpie/CompressedImageData.cpp"
      ___resumeException($$02|0);
      // unreachable;
     }
     $45 = HEAP32[$32>>2]|0; //@line 51 "modules/image/magpie/CompressedImageData.cpp"
     $46 = ($45|0)==(0); //@line 51 "modules/image/magpie/CompressedImageData.cpp"
     if ($46) {
      __ZdaPv($34); //@line 53 "modules/image/magpie/CompressedImageData.cpp"
      $47 = (___cxa_allocate_exception(16)|0); //@line 54 "modules/image/magpie/CompressedImageData.cpp"
      __THREW__ = 0;
      invoke_viii(3349,($47|0),(507667|0),($vararg_buffer3|0)); //@line 54 "modules/image/magpie/CompressedImageData.cpp"
      $48 = __THREW__; __THREW__ = 0;
      $49 = $48&1;
      if (!($49)) {
       __THREW__ = 0;
       invoke_viii(3350,($47|0),(96|0),(344|0)); //@line 54 "modules/image/magpie/CompressedImageData.cpp"
       $50 = __THREW__; __THREW__ = 0;
       label = 3;
       break;
      }
      $51 = ___cxa_find_matching_catch_2()|0;
      $52 = tempRet0;
      ___cxa_free_exception(($47|0)); //@line 54 "modules/image/magpie/CompressedImageData.cpp"
      $$01 = $52;$$02 = $51;
      __ZN4love5image19CompressedImageDataD2Ev($this); //@line 62 "modules/image/magpie/CompressedImageData.cpp"
      ___resumeException($$02|0);
      // unreachable;
     }
     $53 = ((($this)) + 28|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $54 = HEAP32[$53>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $55 = HEAP32[$30>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
     $56 = ($54|0)==($55|0); //@line 57 "modules/image/magpie/CompressedImageData.cpp"
     if (!($56)) {
      $57 = HEAP32[$31>>2]|0; //@line 57 "modules/image/magpie/CompressedImageData.cpp"
      $58 = ($57|0)==(0); //@line 57 "modules/image/magpie/CompressedImageData.cpp"
      if (!($58)) {
       STACKTOP = sp;return; //@line 62 "modules/image/magpie/CompressedImageData.cpp"
      }
     }
     __ZdaPv($34); //@line 59 "modules/image/magpie/CompressedImageData.cpp"
     $59 = (___cxa_allocate_exception(16)|0); //@line 60 "modules/image/magpie/CompressedImageData.cpp"
     __THREW__ = 0;
     invoke_viii(3349,($59|0),(507749|0),($vararg_buffer5|0)); //@line 60 "modules/image/magpie/CompressedImageData.cpp"
     $60 = __THREW__; __THREW__ = 0;
     $61 = $60&1;
     if (!($61)) {
      __THREW__ = 0;
      invoke_viii(3350,($59|0),(96|0),(344|0)); //@line 60 "modules/image/magpie/CompressedImageData.cpp"
      $62 = __THREW__; __THREW__ = 0;
      label = 3;
      break;
     }
     $63 = ___cxa_find_matching_catch_2()|0;
     $64 = tempRet0;
     ___cxa_free_exception(($59|0)); //@line 60 "modules/image/magpie/CompressedImageData.cpp"
     $$01 = $64;$$02 = $63;
     __ZN4love5image19CompressedImageDataD2Ev($this); //@line 62 "modules/image/magpie/CompressedImageData.cpp"
     ___resumeException($$02|0);
     // unreachable;
    }
   }
  }
 } while(0);
 do {
  if ((label|0) == 9) {
   $21 = (___cxa_allocate_exception(16)|0); //@line 44 "modules/image/magpie/CompressedImageData.cpp"
   __THREW__ = 0;
   invoke_viii(3349,($21|0),(507667|0),($vararg_buffer|0)); //@line 44 "modules/image/magpie/CompressedImageData.cpp"
   $22 = __THREW__; __THREW__ = 0;
   $23 = $22&1;
   if (!($23)) {
    __THREW__ = 0;
    invoke_viii(3350,($21|0),(96|0),(344|0)); //@line 44 "modules/image/magpie/CompressedImageData.cpp"
    $24 = __THREW__; __THREW__ = 0;
    label = 3;
    break;
   }
   $25 = ___cxa_find_matching_catch_2()|0;
   $26 = tempRet0;
   ___cxa_free_exception(($21|0)); //@line 44 "modules/image/magpie/CompressedImageData.cpp"
   $$01 = $26;$$02 = $25;
   __ZN4love5image19CompressedImageDataD2Ev($this); //@line 62 "modules/image/magpie/CompressedImageData.cpp"
   ___resumeException($$02|0);
   // unreachable;
  }
 } while(0);
 if ((label|0) == 3) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  $lpad$phi$index = $5;$lpad$phi$index8 = $6;
 }
 $$01 = $lpad$phi$index8;$$02 = $lpad$phi$index;
 __ZN4love5image19CompressedImageDataD2Ev($this); //@line 62 "modules/image/magpie/CompressedImageData.cpp"
 ___resumeException($$02|0);
 // unreachable;
}
function __ZN4love5image6magpie10DDSHandlerD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 47 "modules/image/magpie/ddsHandler.h"
 return; //@line 47 "modules/image/magpie/ddsHandler.h"
}
function __ZN4love5image6magpie10DDSHandler8canParseEPKNS_10filesystem8FileDataE($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$data>>2]|0; //@line 32 "modules/image/magpie/ddsHandler.cpp"
 $1 = ((($0)) + 16|0); //@line 32 "modules/image/magpie/ddsHandler.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 32 "modules/image/magpie/ddsHandler.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($data)|0); //@line 32 "modules/image/magpie/ddsHandler.cpp"
 $4 = HEAP32[$data>>2]|0; //@line 32 "modules/image/magpie/ddsHandler.cpp"
 $5 = ((($4)) + 20|0); //@line 32 "modules/image/magpie/ddsHandler.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 32 "modules/image/magpie/ddsHandler.cpp"
 $7 = (FUNCTION_TABLE_ii[$6 & 8191]($data)|0); //@line 32 "modules/image/magpie/ddsHandler.cpp"
 $8 = (__ZN3dds15isCompressedDDSEPKvj($3,$7)|0); //@line 32 "modules/image/magpie/ddsHandler.cpp"
 return ($8|0); //@line 32 "modules/image/magpie/ddsHandler.cpp"
}
function __ZN4love5image6magpie10DDSHandler5parseEPNS_10filesystem8FileDataERNSt3__16vectorINS0_19CompressedImageData8SubImageENS6_9allocatorIS9_EEEERjRNS8_6FormatERb($this,$filedata,$images,$dataSize,$format,$sRGB) {
 $this = $this|0;
 $filedata = $filedata|0;
 $images = $images|0;
 $dataSize = $dataSize|0;
 $format = $format|0;
 $sRGB = $sRGB|0;
 var $$0 = 0, $$0$i$ph = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$3 = 0, $$34 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $__new_last2$i$i$i = 0, $__new_last2$i$i$i5 = 0, $data$0 = 0, $data$1 = 0, $data$2 = 0, $dataOffset$0 = 0, $i$0 = 0, $i1$0 = 0, $isSRGB$0$ph = 0, $lpad$phi$index = 0, $lpad$phi$index9 = 0, $mip = 0, $parser = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i6 = 0, $scevgep4$i$i$i = 0, $scevgep4$i$i$i7 = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $parser = sp + 48|0;
 $mip = sp + 32|0;
 $0 = HEAP32[$filedata>>2]|0; //@line 37 "modules/image/magpie/ddsHandler.cpp"
 $1 = ((($0)) + 16|0); //@line 37 "modules/image/magpie/ddsHandler.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 37 "modules/image/magpie/ddsHandler.cpp"
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($filedata)|0); //@line 37 "modules/image/magpie/ddsHandler.cpp"
 $4 = HEAP32[$filedata>>2]|0; //@line 37 "modules/image/magpie/ddsHandler.cpp"
 $5 = ((($4)) + 20|0); //@line 37 "modules/image/magpie/ddsHandler.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 37 "modules/image/magpie/ddsHandler.cpp"
 $7 = (FUNCTION_TABLE_ii[$6 & 8191]($filedata)|0); //@line 37 "modules/image/magpie/ddsHandler.cpp"
 $8 = (__ZN3dds5isDDSEPKvj($3,$7)|0); //@line 37 "modules/image/magpie/ddsHandler.cpp"
 if (!($8)) {
  $9 = (___cxa_allocate_exception(16)|0); //@line 38 "modules/image/magpie/ddsHandler.cpp"
  __THREW__ = 0;
  invoke_viii(3349,($9|0),(507830|0),($vararg_buffer|0)); //@line 38 "modules/image/magpie/ddsHandler.cpp"
  $10 = __THREW__; __THREW__ = 0;
  $11 = $10&1;
  if (!($11)) {
   ___cxa_throw(($9|0),(96|0),(344|0)); //@line 38 "modules/image/magpie/ddsHandler.cpp"
   // unreachable; //@line 38 "modules/image/magpie/ddsHandler.cpp"
  }
  $12 = ___cxa_find_matching_catch_2()|0;
  $13 = tempRet0;
  ___cxa_free_exception(($9|0)); //@line 38 "modules/image/magpie/ddsHandler.cpp"
  $$3 = $12;$$34 = $13;
  ___resumeException($$3|0);
  // unreachable;
 }
 HEAP32[$dataSize>>2] = 0; //@line 44 "modules/image/magpie/ddsHandler.cpp"
 $14 = HEAP32[$images>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $15 = ((($images)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $16 = HEAP32[$15>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $17 = ($16|0)==($14|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($17)) {
  $__new_last2$i$i$i = $14; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $scevgep$i$i$i = ((($16)) + -16|0);
  $18 = $scevgep$i$i$i;
  $19 = (($18) - ($__new_last2$i$i$i))|0;
  $20 = $19 >>> 4;
  $21 = $20 ^ -1;
  $scevgep4$i$i$i = (($16) + ($21<<4)|0);
  HEAP32[$15>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $22 = HEAP32[$filedata>>2]|0; //@line 50 "modules/image/magpie/ddsHandler.cpp"
 $23 = ((($22)) + 16|0); //@line 50 "modules/image/magpie/ddsHandler.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 50 "modules/image/magpie/ddsHandler.cpp"
 __THREW__ = 0;
 $25 = (invoke_ii($24|0,($filedata|0))|0); //@line 50 "modules/image/magpie/ddsHandler.cpp"
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  label = 24;
 } else {
  $28 = HEAP32[$filedata>>2]|0; //@line 50 "modules/image/magpie/ddsHandler.cpp"
  $29 = ((($28)) + 20|0); //@line 50 "modules/image/magpie/ddsHandler.cpp"
  $30 = HEAP32[$29>>2]|0; //@line 50 "modules/image/magpie/ddsHandler.cpp"
  __THREW__ = 0;
  $31 = (invoke_ii($30|0,($filedata|0))|0); //@line 50 "modules/image/magpie/ddsHandler.cpp"
  $32 = __THREW__; __THREW__ = 0;
  $33 = $32&1;
  if ($33) {
   label = 24;
  } else {
   __THREW__ = 0;
   invoke_viii(4251,($parser|0),($25|0),($31|0)); //@line 50 "modules/image/magpie/ddsHandler.cpp"
   $34 = __THREW__; __THREW__ = 0;
   $35 = $34&1;
   if ($35) {
    label = 24;
   } else {
    __THREW__ = 0;
    $36 = (invoke_ii(4252,($parser|0))|0); //@line 52 "modules/image/magpie/ddsHandler.cpp"
    $37 = __THREW__; __THREW__ = 0;
    $38 = $37&1;
    L15: do {
     if ($38) {
      label = 27;
     } else {
      do {
       switch ($36|0) {
       case 0:  {
        $$0$i$ph = 1;$isSRGB$0$ph = 0;
        break;
       }
       case 1:  {
        $$0$i$ph = 2;$isSRGB$0$ph = 0;
        break;
       }
       case 2:  {
        $$0$i$ph = 3;$isSRGB$0$ph = 0;
        break;
       }
       case 3:  {
        $$0$i$ph = 4;$isSRGB$0$ph = 0;
        break;
       }
       case 4:  {
        $$0$i$ph = 5;$isSRGB$0$ph = 0;
        break;
       }
       case 5:  {
        $$0$i$ph = 6;$isSRGB$0$ph = 0;
        break;
       }
       case 6:  {
        $$0$i$ph = 7;$isSRGB$0$ph = 0;
        break;
       }
       case 7:  {
        $$0$i$ph = 8;$isSRGB$0$ph = 0;
        break;
       }
       case 8:  {
        $$0$i$ph = 9;$isSRGB$0$ph = 0;
        break;
       }
       case 9:  {
        $$0$i$ph = 10;$isSRGB$0$ph = 0;
        break;
       }
       case 10:  {
        $$0$i$ph = 10;$isSRGB$0$ph = 1;
        break;
       }
       default: {
        $39 = (___cxa_allocate_exception(16)|0); //@line 55 "modules/image/magpie/ddsHandler.cpp"
        __THREW__ = 0;
        invoke_viii(3349,($39|0),(507881|0),($vararg_buffer1|0)); //@line 55 "modules/image/magpie/ddsHandler.cpp"
        $40 = __THREW__; __THREW__ = 0;
        $41 = $40&1;
        if ($41) {
         $51 = ___cxa_find_matching_catch_3(8992|0)|0;
         $52 = tempRet0;
         ___cxa_free_exception(($39|0)); //@line 55 "modules/image/magpie/ddsHandler.cpp"
         $$0 = $51;$$01 = $52;$data$1 = 0;
         break L15;
        } else {
         __THREW__ = 0;
         invoke_viii(3350,($39|0),(96|0),(344|0)); //@line 55 "modules/image/magpie/ddsHandler.cpp"
         $42 = __THREW__; __THREW__ = 0;
         label = 27;
         break L15;
        }
       }
       }
      } while(0);
      __THREW__ = 0;
      $53 = (invoke_ii(4253,($parser|0))|0); //@line 57 "modules/image/magpie/ddsHandler.cpp"
      $54 = __THREW__; __THREW__ = 0;
      $55 = $54&1;
      if ($55) {
       label = 27;
      } else {
       $56 = ($53|0)==(0); //@line 57 "modules/image/magpie/ddsHandler.cpp"
       if ($56) {
        $57 = (___cxa_allocate_exception(16)|0); //@line 58 "modules/image/magpie/ddsHandler.cpp"
        __THREW__ = 0;
        invoke_viii(3349,($57|0),(507934|0),($vararg_buffer3|0)); //@line 58 "modules/image/magpie/ddsHandler.cpp"
        $58 = __THREW__; __THREW__ = 0;
        $59 = $58&1;
        if ($59) {
         $61 = ___cxa_find_matching_catch_3(8992|0)|0;
         $62 = tempRet0;
         ___cxa_free_exception(($57|0)); //@line 58 "modules/image/magpie/ddsHandler.cpp"
         $$0 = $61;$$01 = $62;$data$1 = 0;
         break;
        } else {
         __THREW__ = 0;
         invoke_viii(3350,($57|0),(96|0),(344|0)); //@line 58 "modules/image/magpie/ddsHandler.cpp"
         $60 = __THREW__; __THREW__ = 0;
         label = 27;
         break;
        }
       } else {
        $i$0 = 0;
       }
       while(1) {
        __THREW__ = 0;
        $63 = (invoke_ii(4253,($parser|0))|0); //@line 61 "modules/image/magpie/ddsHandler.cpp"
        $64 = __THREW__; __THREW__ = 0;
        $65 = $64&1;
        if ($65) {
         label = 26;
         break;
        }
        $66 = ($i$0>>>0)<($63>>>0); //@line 61 "modules/image/magpie/ddsHandler.cpp"
        if (!($66)) {
         break;
        }
        __THREW__ = 0;
        $75 = (invoke_iii(4254,($parser|0),($i$0|0))|0); //@line 63 "modules/image/magpie/ddsHandler.cpp"
        $76 = __THREW__; __THREW__ = 0;
        $77 = $76&1;
        if ($77) {
         label = 26;
         break;
        }
        $78 = ((($75)) + 8|0); //@line 64 "modules/image/magpie/ddsHandler.cpp"
        $79 = HEAP32[$78>>2]|0; //@line 64 "modules/image/magpie/ddsHandler.cpp"
        $80 = HEAP32[$dataSize>>2]|0; //@line 64 "modules/image/magpie/ddsHandler.cpp"
        $81 = (($80) + ($79))|0; //@line 64 "modules/image/magpie/ddsHandler.cpp"
        HEAP32[$dataSize>>2] = $81; //@line 64 "modules/image/magpie/ddsHandler.cpp"
        $82 = (($i$0) + 1)|0; //@line 61 "modules/image/magpie/ddsHandler.cpp"
        $i$0 = $82;
       }
       if ((label|0) == 26) {
        $47 = ___cxa_find_matching_catch_3(8992|0)|0;
        $48 = tempRet0;
        $data$0 = 0;$lpad$phi$index = $47;$lpad$phi$index9 = $48;
        label = 28;
        break;
       }
       $67 = HEAP32[$dataSize>>2]|0; //@line 67 "modules/image/magpie/ddsHandler.cpp"
       __THREW__ = 0;
       $68 = (invoke_ii(3492,($67|0))|0); //@line 67 "modules/image/magpie/ddsHandler.cpp"
       $69 = __THREW__; __THREW__ = 0;
       $70 = $69&1;
       if ($70) {
        label = 27;
       } else {
        $71 = ((($mip)) + 4|0); //@line 80 "modules/image/magpie/ddsHandler.cpp"
        $72 = ((($mip)) + 8|0); //@line 81 "modules/image/magpie/ddsHandler.cpp"
        $73 = ((($mip)) + 12|0); //@line 85 "modules/image/magpie/ddsHandler.cpp"
        $74 = ((($images)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
        $dataOffset$0 = 0;$i1$0 = 0;
        while(1) {
         __THREW__ = 0;
         $83 = (invoke_ii(4253,($parser|0))|0); //@line 72 "modules/image/magpie/ddsHandler.cpp"
         $84 = __THREW__; __THREW__ = 0;
         $85 = $84&1;
         if ($85) {
          label = 25;
          break;
         }
         $86 = ($i1$0>>>0)<($83>>>0); //@line 72 "modules/image/magpie/ddsHandler.cpp"
         if (!($86)) {
          label = 43;
          break;
         }
         __THREW__ = 0;
         $87 = (invoke_iii(4254,($parser|0),($i1$0|0))|0); //@line 75 "modules/image/magpie/ddsHandler.cpp"
         $88 = __THREW__; __THREW__ = 0;
         $89 = $88&1;
         if ($89) {
          label = 25;
          break;
         }
         $90 = HEAP32[$87>>2]|0; //@line 79 "modules/image/magpie/ddsHandler.cpp"
         HEAP32[$mip>>2] = $90; //@line 79 "modules/image/magpie/ddsHandler.cpp"
         $91 = ((($87)) + 4|0); //@line 80 "modules/image/magpie/ddsHandler.cpp"
         $92 = HEAP32[$91>>2]|0; //@line 80 "modules/image/magpie/ddsHandler.cpp"
         HEAP32[$71>>2] = $92; //@line 80 "modules/image/magpie/ddsHandler.cpp"
         $93 = ((($87)) + 8|0); //@line 81 "modules/image/magpie/ddsHandler.cpp"
         $94 = HEAP32[$93>>2]|0; //@line 81 "modules/image/magpie/ddsHandler.cpp"
         HEAP32[$72>>2] = $94; //@line 81 "modules/image/magpie/ddsHandler.cpp"
         $95 = (($68) + ($dataOffset$0)|0); //@line 84 "modules/image/magpie/ddsHandler.cpp"
         $96 = ((($87)) + 12|0); //@line 84 "modules/image/magpie/ddsHandler.cpp"
         $97 = HEAP32[$96>>2]|0; //@line 84 "modules/image/magpie/ddsHandler.cpp"
         _memcpy(($95|0),($97|0),($94|0))|0; //@line 84 "modules/image/magpie/ddsHandler.cpp"
         HEAP32[$73>>2] = $95; //@line 85 "modules/image/magpie/ddsHandler.cpp"
         $98 = HEAP32[$15>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $99 = HEAP32[$74>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         $100 = ($98|0)==($99|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         if ($100) {
          __THREW__ = 0;
          invoke_vii(4255,($images|0),($mip|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $103 = __THREW__; __THREW__ = 0;
          $104 = $103&1;
          if ($104) {
           label = 25;
           break;
          }
         } else {
          ;HEAP32[$98>>2]=HEAP32[$mip>>2]|0;HEAP32[$98+4>>2]=HEAP32[$mip+4>>2]|0;HEAP32[$98+8>>2]=HEAP32[$mip+8>>2]|0;HEAP32[$98+12>>2]=HEAP32[$mip+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
          $101 = HEAP32[$15>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          $102 = ((($101)) + 16|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
          HEAP32[$15>>2] = $102; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
         }
         $105 = (($94) + ($dataOffset$0))|0; //@line 87 "modules/image/magpie/ddsHandler.cpp"
         $106 = (($i1$0) + 1)|0; //@line 72 "modules/image/magpie/ddsHandler.cpp"
         $dataOffset$0 = $105;$i1$0 = $106;
        }
        if ((label|0) == 25) {
         $45 = ___cxa_find_matching_catch_3(8992|0)|0;
         $46 = tempRet0;
         $data$0 = $68;$lpad$phi$index = $45;$lpad$phi$index9 = $46;
         label = 28;
         break;
        }
        else if ((label|0) == 43) {
         __ZN3dds6ParserD2Ev($parser); //@line 91 "modules/image/magpie/ddsHandler.cpp"
         HEAP32[$format>>2] = $$0$i$ph; //@line 99 "modules/image/magpie/ddsHandler.cpp"
         HEAP8[$sRGB>>0] = $isSRGB$0$ph; //@line 100 "modules/image/magpie/ddsHandler.cpp"
         STACKTOP = sp;return ($68|0); //@line 101 "modules/image/magpie/ddsHandler.cpp"
        }
       }
      }
     }
    } while(0);
    if ((label|0) == 27) {
     $49 = ___cxa_find_matching_catch_3(8992|0)|0;
     $50 = tempRet0;
     $data$0 = 0;$lpad$phi$index = $49;$lpad$phi$index9 = $50;
     label = 28;
    }
    if ((label|0) == 28) {
     $$0 = $lpad$phi$index;$$01 = $lpad$phi$index9;$data$1 = $data$0;
    }
    __ZN3dds6ParserD2Ev($parser); //@line 91 "modules/image/magpie/ddsHandler.cpp"
    $$1 = $$0;$$12 = $$01;$data$2 = $data$1;
   }
  }
 }
 if ((label|0) == 24) {
  $43 = ___cxa_find_matching_catch_3(8992|0)|0;
  $44 = tempRet0;
  $$1 = $43;$$12 = $44;$data$2 = 0;
 }
 $107 = (_llvm_eh_typeid_for((8992|0))|0); //@line 91 "modules/image/magpie/ddsHandler.cpp"
 $108 = ($$12|0)==($107|0); //@line 91 "modules/image/magpie/ddsHandler.cpp"
 if (!($108)) {
  $$3 = $$1;$$34 = $$12;
  ___resumeException($$3|0);
  // unreachable;
 }
 $109 = (___cxa_begin_catch(($$1|0))|0); //@line 91 "modules/image/magpie/ddsHandler.cpp"
 $110 = ($data$2|0)==(0|0); //@line 94 "modules/image/magpie/ddsHandler.cpp"
 if (!($110)) {
  __ZdaPv($data$2); //@line 94 "modules/image/magpie/ddsHandler.cpp"
 }
 $111 = HEAP32[$images>>2]|0; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $112 = HEAP32[$15>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $113 = ($112|0)==($111|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($113)) {
  $__new_last2$i$i$i5 = $111; //@line 368 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $scevgep$i$i$i6 = ((($112)) + -16|0);
  $114 = $scevgep$i$i$i6;
  $115 = (($114) - ($__new_last2$i$i$i5))|0;
  $116 = $115 >>> 4;
  $117 = $116 ^ -1;
  $scevgep4$i$i$i7 = (($112) + ($117<<4)|0);
  HEAP32[$15>>2] = $scevgep4$i$i$i7; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $118 = (___cxa_allocate_exception(16)|0); //@line 96 "modules/image/magpie/ddsHandler.cpp"
 $119 = HEAP32[$109>>2]|0; //@line 96 "modules/image/magpie/ddsHandler.cpp"
 $120 = ((($119)) + 8|0); //@line 96 "modules/image/magpie/ddsHandler.cpp"
 $121 = HEAP32[$120>>2]|0; //@line 96 "modules/image/magpie/ddsHandler.cpp"
 $122 = (FUNCTION_TABLE_ii[$121 & 8191]($109)|0); //@line 96 "modules/image/magpie/ddsHandler.cpp"
 __THREW__ = 0;
 HEAP32[$vararg_buffer5>>2] = $122; //@line 96 "modules/image/magpie/ddsHandler.cpp"
 invoke_viii(3349,($118|0),(595867|0),($vararg_buffer5|0)); //@line 96 "modules/image/magpie/ddsHandler.cpp"
 $123 = __THREW__; __THREW__ = 0;
 $124 = $123&1;
 if ($124) {
  $128 = ___cxa_find_matching_catch_2()|0;
  $129 = tempRet0;
  ___cxa_free_exception(($118|0)); //@line 96 "modules/image/magpie/ddsHandler.cpp"
  $$2 = $128;$$23 = $129;
 } else {
  __THREW__ = 0;
  invoke_viii(3350,($118|0),(96|0),(344|0)); //@line 96 "modules/image/magpie/ddsHandler.cpp"
  $125 = __THREW__; __THREW__ = 0;
  $126 = ___cxa_find_matching_catch_2()|0;
  $127 = tempRet0;
  $$2 = $126;$$23 = $127;
 }
 __THREW__ = 0;
 invoke_v(3378); //@line 97 "modules/image/magpie/ddsHandler.cpp"
 $130 = __THREW__; __THREW__ = 0;
 $131 = $130&1;
 if ($131) {
  $132 = ___cxa_find_matching_catch_3(0|0)|0;
  $133 = tempRet0;
  ___clang_call_terminate($132); //@line 97 "modules/image/magpie/ddsHandler.cpp"
  // unreachable; //@line 97 "modules/image/magpie/ddsHandler.cpp"
 } else {
  $$3 = $$2;$$34 = $$23;
  ___resumeException($$3|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN4love5image6magpie13FormatHandlerD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 38 "modules/image/magpie/FormatHandler.cpp"
}
function __ZN4love5image6magpie13FormatHandlerD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this); //@line 37 "modules/image/magpie/FormatHandler.cpp"
 return; //@line 38 "modules/image/magpie/FormatHandler.cpp"
}
function __ZN4love5image6magpie13FormatHandler9canDecodeEPNS_10filesystem8FileDataE($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0; //@line 42 "modules/image/magpie/FormatHandler.cpp"
}
function __ZN4love5image6magpie13FormatHandler9canEncodeENS0_9ImageData13EncodedFormatE($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0; //@line 47 "modules/image/magpie/FormatHandler.cpp"
}
function __ZN4love5image6magpie13FormatHandler6decodeEPNS_10filesystem8FileDataE($agg$result,$this,$0) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $1 = (___cxa_allocate_exception(16)|0); //@line 52 "modules/image/magpie/FormatHandler.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($1|0),(508029|0),($vararg_buffer|0)); //@line 52 "modules/image/magpie/FormatHandler.cpp"
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  ___cxa_free_exception(($1|0)); //@line 52 "modules/image/magpie/FormatHandler.cpp"
  ___resumeException($4|0);
  // unreachable;
 } else {
  ___cxa_throw(($1|0),(96|0),(344|0)); //@line 52 "modules/image/magpie/FormatHandler.cpp"
  // unreachable; //@line 52 "modules/image/magpie/FormatHandler.cpp"
 }
}
function __ZN4love5image6magpie13FormatHandler6encodeERKNS2_12DecodedImageENS0_9ImageData13EncodedFormatE($agg$result,$this,$0,$1) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $2 = (___cxa_allocate_exception(16)|0); //@line 57 "modules/image/magpie/FormatHandler.cpp"
 __THREW__ = 0;
 invoke_viii(3349,($2|0),(508088|0),($vararg_buffer|0)); //@line 57 "modules/image/magpie/FormatHandler.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  ___cxa_free_exception(($2|0)); //@line 57 "modules/image/magpie/FormatHandler.cpp"
  ___resumeException($5|0);
  // unreachable;
 } else {
  ___cxa_throw(($2|0),(96|0),(344|0)); //@line 57 "modules/image/magpie/FormatHandler.cpp"
  // unreachable; //@line 57 "modules/image/magpie/FormatHandler.cpp"
 }
}
function __ZN4love5image6magpie13FormatHandler4freeEPh($this,$mem) {
 $this = $this|0;
 $mem = $mem|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mem|0)==(0|0); //@line 62 "modules/image/magpie/FormatHandler.cpp"
 if (!($0)) {
  __ZdaPv($mem); //@line 62 "modules/image/magpie/FormatHandler.cpp"
 }
 return; //@line 63 "modules/image/magpie/FormatHandler.cpp"
}
function __ZN4love5image6magpie13FormatHandlerC2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6ObjectC2Ev($this); //@line 32 "modules/image/magpie/FormatHandler.cpp"
 HEAP32[$this>>2] = (35248); //@line 33 "modules/image/magpie/FormatHandler.cpp"
 return; //@line 34 "modules/image/magpie/FormatHandler.cpp"
}
function __ZN4love6thread3sdl6ThreadD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $l$i = 0, $l1$i = 0, $not$ = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l$i = sp + 4|0;
 $l1$i = sp;
 HEAP32[$this>>2] = (35292); //@line 37 "modules/thread/sdl/Thread.cpp"
 $0 = ((($this)) + 8|0); //@line 38 "modules/thread/sdl/Thread.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 38 "modules/thread/sdl/Thread.cpp"
 $2 = ($1<<24>>24)==(0); //@line 38 "modules/thread/sdl/Thread.cpp"
 do {
  if ($2) {
   $3 = ((($this)) + 16|0); //@line 62 "modules/thread/sdl/Thread.cpp"
   __THREW__ = 0;
   invoke_vii(4256,($l$i|0),($3|0)); //@line 62 "modules/thread/sdl/Thread.cpp"
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   if (!($5)) {
    $6 = ((($this)) + 12|0); //@line 63 "modules/thread/sdl/Thread.cpp"
    $7 = HEAP32[$6>>2]|0; //@line 63 "modules/thread/sdl/Thread.cpp"
    $not$ = ($7|0)==(0|0); //@line 64 "modules/thread/sdl/Thread.cpp"
    __ZN4love6thread4LockD2Ev($l$i); //@line 65 "modules/thread/sdl/Thread.cpp"
    if ($not$) {
     break;
    }
    $8 = HEAP32[$6>>2]|0; //@line 66 "modules/thread/sdl/Thread.cpp"
    __THREW__ = 0;
    invoke_vii(4257,($8|0),(0|0)); //@line 66 "modules/thread/sdl/Thread.cpp"
    $9 = __THREW__; __THREW__ = 0;
    $10 = $9&1;
    if (!($10)) {
     __THREW__ = 0;
     invoke_vii(4256,($l1$i|0),($3|0)); //@line 67 "modules/thread/sdl/Thread.cpp"
     $11 = __THREW__; __THREW__ = 0;
     $12 = $11&1;
     if (!($12)) {
      HEAP8[$0>>0] = 0; //@line 68 "modules/thread/sdl/Thread.cpp"
      HEAP32[$6>>2] = 0; //@line 69 "modules/thread/sdl/Thread.cpp"
      __ZN4love6thread4LockD2Ev($l1$i); //@line 70 "modules/thread/sdl/Thread.cpp"
      break;
     }
    }
   }
   $13 = ___cxa_find_matching_catch_3(0|0)|0;
   $14 = tempRet0;
   $15 = ((($this)) + 16|0); //@line 45 "modules/thread/sdl/Thread.cpp"
   __ZN4love6thread3sdl5MutexD2Ev($15); //@line 45 "modules/thread/sdl/Thread.cpp"
   ___clang_call_terminate($13); //@line 45 "modules/thread/sdl/Thread.cpp"
   // unreachable; //@line 45 "modules/thread/sdl/Thread.cpp"
  }
 } while(0);
 $16 = ((($this)) + 16|0); //@line 45 "modules/thread/sdl/Thread.cpp"
 __ZN4love6thread3sdl5MutexD2Ev($16); //@line 45 "modules/thread/sdl/Thread.cpp"
 STACKTOP = sp;return; //@line 45 "modules/thread/sdl/Thread.cpp"
}
function __ZN4love6thread3sdl6ThreadD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN4love6thread3sdl6ThreadD2Ev($this); //@line 37 "modules/thread/sdl/Thread.cpp"
 __ZdlPv($this); //@line 37 "modules/thread/sdl/Thread.cpp"
 return; //@line 45 "modules/thread/sdl/Thread.cpp"
}
function __ZN4love6thread3sdl6Thread5startEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = ((($this)) + 16|0); //@line 49 "modules/thread/sdl/Thread.cpp"
 __ZN4love6thread4LockC2ERNS0_5MutexE($l,$0); //@line 49 "modules/thread/sdl/Thread.cpp"
 $1 = ((($this)) + 8|0); //@line 50 "modules/thread/sdl/Thread.cpp"
 $2 = HEAP8[$1>>0]|0; //@line 50 "modules/thread/sdl/Thread.cpp"
 $3 = ($2<<24>>24)==(0); //@line 50 "modules/thread/sdl/Thread.cpp"
 if (!($3)) {
  $$0 = 0;
  __ZN4love6thread4LockD2Ev($l); //@line 57 "modules/thread/sdl/Thread.cpp"
  STACKTOP = sp;return ($$0|0); //@line 57 "modules/thread/sdl/Thread.cpp"
 }
 $4 = ((($this)) + 12|0); //@line 52 "modules/thread/sdl/Thread.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 52 "modules/thread/sdl/Thread.cpp"
 $6 = ($5|0)==(0|0); //@line 52 "modules/thread/sdl/Thread.cpp"
 if ($6) {
  label = 5;
 } else {
  __THREW__ = 0;
  invoke_vii(4257,($5|0),(0|0)); //@line 53 "modules/thread/sdl/Thread.cpp"
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if (!($8)) {
   label = 5;
  }
 }
 if ((label|0) == 5) {
  $11 = ((($this)) + 4|0); //@line 54 "modules/thread/sdl/Thread.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 54 "modules/thread/sdl/Thread.cpp"
  __THREW__ = 0;
  $13 = (invoke_ii(4258,($12|0))|0); //@line 54 "modules/thread/sdl/Thread.cpp"
  $14 = __THREW__; __THREW__ = 0;
  $15 = $14&1;
  if (!($15)) {
   __THREW__ = 0;
   $16 = (invoke_iiii(4259,(4260|0),($13|0),($this|0))|0); //@line 54 "modules/thread/sdl/Thread.cpp"
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if (!($18)) {
    HEAP32[$4>>2] = $16; //@line 54 "modules/thread/sdl/Thread.cpp"
    $19 = ($16|0)!=(0|0); //@line 55 "modules/thread/sdl/Thread.cpp"
    $20 = $19&1; //@line 55 "modules/thread/sdl/Thread.cpp"
    HEAP8[$1>>0] = $20; //@line 55 "modules/thread/sdl/Thread.cpp"
    $$0 = $19;
    __ZN4love6thread4LockD2Ev($l); //@line 57 "modules/thread/sdl/Thread.cpp"
    STACKTOP = sp;return ($$0|0); //@line 57 "modules/thread/sdl/Thread.cpp"
   }
  }
 }
 $9 = ___cxa_find_matching_catch_2()|0;
 $10 = tempRet0;
 __ZN4love6thread4LockD2Ev($l); //@line 57 "modules/thread/sdl/Thread.cpp"
 ___resumeException($9|0);
 // unreachable;
 return (0)|0;
}
function __ZN4love6thread3sdl6Thread13thread_runnerEPv($data) {
 $data = $data|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = ((($data)) + 4|0); //@line 81 "modules/thread/sdl/Thread.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 81 "modules/thread/sdl/Thread.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 81 "modules/thread/sdl/Thread.cpp"
 $3 = ((($2)) + 8|0); //@line 81 "modules/thread/sdl/Thread.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 81 "modules/thread/sdl/Thread.cpp"
 FUNCTION_TABLE_vi[$4 & 8191]($1); //@line 81 "modules/thread/sdl/Thread.cpp"
 $5 = ((($data)) + 16|0); //@line 82 "modules/thread/sdl/Thread.cpp"
 __ZN4love6thread4LockC2ERNS0_5MutexE($l,$5); //@line 82 "modules/thread/sdl/Thread.cpp"
 $6 = ((($data)) + 8|0); //@line 83 "modules/thread/sdl/Thread.cpp"
 HEAP8[$6>>0] = 0; //@line 83 "modules/thread/sdl/Thread.cpp"
 __ZN4love6thread4LockD2Ev($l); //@line 85 "modules/thread/sdl/Thread.cpp"
 STACKTOP = sp;return 0; //@line 85 "modules/thread/sdl/Thread.cpp"
}
function __ZN4love6thread3sdl6Thread4waitEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $l = 0, $l1 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp + 4|0;
 $l1 = sp;
 $0 = ((($this)) + 16|0); //@line 62 "modules/thread/sdl/Thread.cpp"
 __ZN4love6thread4LockC2ERNS0_5MutexE($l,$0); //@line 62 "modules/thread/sdl/Thread.cpp"
 $1 = ((($this)) + 12|0); //@line 63 "modules/thread/sdl/Thread.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 63 "modules/thread/sdl/Thread.cpp"
 $not$ = ($2|0)==(0|0); //@line 64 "modules/thread/sdl/Thread.cpp"
 __ZN4love6thread4LockD2Ev($l); //@line 65 "modules/thread/sdl/Thread.cpp"
 if ($not$) {
  STACKTOP = sp;return; //@line 70 "modules/thread/sdl/Thread.cpp"
 }
 $3 = HEAP32[$1>>2]|0; //@line 66 "modules/thread/sdl/Thread.cpp"
 _SDL_WaitThread($3,0); //@line 66 "modules/thread/sdl/Thread.cpp"
 __ZN4love6thread4LockC2ERNS0_5MutexE($l1,$0); //@line 67 "modules/thread/sdl/Thread.cpp"
 $4 = ((($this)) + 8|0); //@line 68 "modules/thread/sdl/Thread.cpp"
 HEAP8[$4>>0] = 0; //@line 68 "modules/thread/sdl/Thread.cpp"
 HEAP32[$1>>2] = 0; //@line 69 "modules/thread/sdl/Thread.cpp"
 __ZN4love6thread4LockD2Ev($l1); //@line 70 "modules/thread/sdl/Thread.cpp"
 STACKTOP = sp;return; //@line 70 "modules/thread/sdl/Thread.cpp"
}
function __ZN4love6thread3sdl6Thread9isRunningEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = ((($this)) + 16|0); //@line 74 "modules/thread/sdl/Thread.cpp"
 __ZN4love6thread4LockC2ERNS0_5MutexE($l,$0); //@line 74 "modules/thread/sdl/Thread.cpp"
 $1 = ((($this)) + 8|0); //@line 75 "modules/thread/sdl/Thread.cpp"
 $2 = HEAP8[$1>>0]|0; //@line 75 "modules/thread/sdl/Thread.cpp"
 $3 = ($2<<24>>24)!=(0); //@line 75 "modules/thread/sdl/Thread.cpp"
 __ZN4love6thread4LockD2Ev($l); //@line 76 "modules/thread/sdl/Thread.cpp"
 STACKTOP = sp;return ($3|0); //@line 76 "modules/thread/sdl/Thread.cpp"
}
function __ZN4love6thread3sdl6ThreadC2EPNS0_10ThreadableE($this,$t) {
 $this = $this|0;
 $t = $t|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (35292); //@line 33 "modules/thread/sdl/Thread.cpp"
 $0 = ((($this)) + 4|0); //@line 30 "modules/thread/sdl/Thread.cpp"
 HEAP32[$0>>2] = $t; //@line 30 "modules/thread/sdl/Thread.cpp"
 $1 = ((($this)) + 8|0); //@line 31 "modules/thread/sdl/Thread.cpp"
 HEAP8[$1>>0] = 0; //@line 31 "modules/thread/sdl/Thread.cpp"
 $2 = ((($this)) + 12|0); //@line 32 "modules/thread/sdl/Thread.cpp"
 HEAP32[$2>>2] = 0; //@line 32 "modules/thread/sdl/Thread.cpp"
 $3 = ((($this)) + 16|0); //@line 29 "modules/thread/sdl/Thread.cpp"
 __ZN4love6thread3sdl5MutexC2Ev($3); //@line 29 "modules/thread/sdl/Thread.cpp"
 return; //@line 34 "modules/thread/sdl/Thread.cpp"
}
function __ZN3dds5isDDSEPKvj($data,$dataSize) {
 $data = $data|0;
 $dataSize = $dataSize|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ($dataSize>>>0)<(128); //@line 141 "libraries/ddsparse/ddsparse.cpp"
 do {
  if ($0) {
   $$1 = 0;
  } else {
   $1 = HEAP32[$data>>2]|0; //@line 145 "libraries/ddsparse/ddsparse.cpp"
   $2 = ($1|0)==(542327876); //@line 145 "libraries/ddsparse/ddsparse.cpp"
   if ($2) {
    $3 = ((($data)) + 4|0); //@line 150 "libraries/ddsparse/ddsparse.cpp"
    $4 = HEAP32[$3>>2]|0; //@line 154 "libraries/ddsparse/ddsparse.cpp"
    $5 = ($4|0)==(124); //@line 154 "libraries/ddsparse/ddsparse.cpp"
    if ($5) {
     $6 = ((($data)) + 76|0); //@line 154 "libraries/ddsparse/ddsparse.cpp"
     $7 = HEAP32[$6>>2]|0; //@line 154 "libraries/ddsparse/ddsparse.cpp"
     $8 = ($7|0)==(32); //@line 154 "libraries/ddsparse/ddsparse.cpp"
     if ($8) {
      $9 = ((($data)) + 80|0); //@line 160 "libraries/ddsparse/ddsparse.cpp"
      $10 = HEAP32[$9>>2]|0; //@line 160 "libraries/ddsparse/ddsparse.cpp"
      $11 = $10 & 4; //@line 160 "libraries/ddsparse/ddsparse.cpp"
      $12 = ($11|0)==(0); //@line 160 "libraries/ddsparse/ddsparse.cpp"
      if (!($12)) {
       $13 = ((($data)) + 84|0); //@line 160 "libraries/ddsparse/ddsparse.cpp"
       $14 = HEAP32[$13>>2]|0; //@line 160 "libraries/ddsparse/ddsparse.cpp"
       $15 = ($14|0)==(808540228); //@line 160 "libraries/ddsparse/ddsparse.cpp"
       $16 = ($dataSize>>>0)<(148); //@line 163 "libraries/ddsparse/ddsparse.cpp"
       $or$cond = $16 & $15; //@line 160 "libraries/ddsparse/ddsparse.cpp"
       if ($or$cond) {
        $$1 = 0;
        break;
       }
      }
      $$1 = 1;
     } else {
      $$1 = 0;
     }
    } else {
     $$1 = 0;
    }
   } else {
    $$1 = 0;
   }
  }
 } while(0);
 return ($$1|0); //@line 168 "libraries/ddsparse/ddsparse.cpp"
}
function __ZN3dds15isCompressedDDSEPKvj($data,$dataSize) {
 $data = $data|0;
 $dataSize = $dataSize|0;
 var $$1 = 0, $$clear = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond$i = 0, $switch$cast = 0, $switch$cast$clear = 0, $switch$downshift = 0, $switch$masked = 0, $switch$split102D = 0, $switch$split12D = 0, $switch$split132D = 0, $switch$split162D = 0, $switch$split192D = 0, $switch$split2D = 0, $switch$split42D = 0;
 var $switch$split72D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($dataSize>>>0)<(128); //@line 141 "libraries/ddsparse/ddsparse.cpp"
 L1: do {
  if ($0) {
   $$1 = 0;
  } else {
   $1 = HEAP32[$data>>2]|0; //@line 145 "libraries/ddsparse/ddsparse.cpp"
   $2 = ($1|0)==(542327876); //@line 145 "libraries/ddsparse/ddsparse.cpp"
   if ($2) {
    $3 = ((($data)) + 4|0); //@line 150 "libraries/ddsparse/ddsparse.cpp"
    $4 = HEAP32[$3>>2]|0; //@line 154 "libraries/ddsparse/ddsparse.cpp"
    $5 = ($4|0)==(124); //@line 154 "libraries/ddsparse/ddsparse.cpp"
    if ($5) {
     $6 = ((($data)) + 76|0); //@line 154 "libraries/ddsparse/ddsparse.cpp"
     $7 = HEAP32[$6>>2]|0; //@line 154 "libraries/ddsparse/ddsparse.cpp"
     $8 = ($7|0)==(32); //@line 154 "libraries/ddsparse/ddsparse.cpp"
     if ($8) {
      $9 = ((($data)) + 80|0); //@line 160 "libraries/ddsparse/ddsparse.cpp"
      $10 = HEAP32[$9>>2]|0; //@line 41 "libraries/ddsparse/ddsparse.cpp"
      $11 = $10 & 4; //@line 160 "libraries/ddsparse/ddsparse.cpp"
      $12 = ($11|0)==(0); //@line 160 "libraries/ddsparse/ddsparse.cpp"
      if ($12) {
       $$1 = 0;
      } else {
       $13 = ((($data)) + 84|0); //@line 160 "libraries/ddsparse/ddsparse.cpp"
       $14 = HEAP32[$13>>2]|0; //@line 160 "libraries/ddsparse/ddsparse.cpp"
       $15 = ($14|0)==(808540228); //@line 160 "libraries/ddsparse/ddsparse.cpp"
       $16 = ($dataSize>>>0)<(148); //@line 163 "libraries/ddsparse/ddsparse.cpp"
       $or$cond$i = $16 & $15; //@line 160 "libraries/ddsparse/ddsparse.cpp"
       if ($or$cond$i) {
        $$1 = 0;
       } else {
        $17 = ((($data)) + 84|0); //@line 182 "libraries/ddsparse/ddsparse.cpp"
        $18 = HEAP32[$17>>2]|0; //@line 182 "libraries/ddsparse/ddsparse.cpp"
        $19 = ($18|0)==(808540228); //@line 182 "libraries/ddsparse/ddsparse.cpp"
        if ($19) {
         $20 = ((($data)) + 128|0); //@line 184 "libraries/ddsparse/ddsparse.cpp"
         $21 = HEAP32[$20>>2]|0; //@line 185 "libraries/ddsparse/ddsparse.cpp"
         $switch$cast = (($21) + -70)|0; //@line 83 "libraries/ddsparse/ddsparse.cpp"
         $22 = ($switch$cast>>>0)<(30); //@line 83 "libraries/ddsparse/ddsparse.cpp"
         if (!($22)) {
          $$1 = 0;
          break;
         }
         $switch$cast$clear = $switch$cast & 1073741823; //@line 83 "libraries/ddsparse/ddsparse.cpp"
         $switch$downshift = 1056997375 >>> $switch$cast$clear;
         $25 = $switch$downshift & 1; //@line 83 "libraries/ddsparse/ddsparse.cpp"
         $$clear = $25 & 1073741823; //@line 83 "libraries/ddsparse/ddsparse.cpp"
         $switch$masked = ($$clear|0)!=(0); //@line 83 "libraries/ddsparse/ddsparse.cpp"
         $$1 = $switch$masked;
         break;
        }
        $23 = ((($data)) + 84|0); //@line 46 "libraries/ddsparse/ddsparse.cpp"
        $24 = HEAP32[$23>>2]|0; //@line 46 "libraries/ddsparse/ddsparse.cpp"
        $switch$split2D = ($24|0)<(894720068);
        L12: do {
         if ($switch$split2D) {
          $switch$split12D = ($24|0)<(843666497);
          if ($switch$split12D) {
           $switch$split72D = ($24|0)<(827611204);
           if ($switch$split72D) {
            switch ($24|0) {
            case 826889281:  {
             label = 13;
             break L12;
             break;
            }
            default: {
             $$1 = 0;
             break L1;
            }
            }
           }
           switch ($24|0) {
           case 827611204:  {
            break;
           }
           default: {
            $$1 = 0;
            break L1;
           }
           }
           $$1 = 1;
           break L1;
          } else {
           $switch$split102D = ($24|0)<(861165636);
           if ($switch$split102D) {
            switch ($24|0) {
            case 843666497:  {
             label = 15;
             break L12;
             break;
            }
            default: {
             $$1 = 0;
             break L1;
            }
            }
           }
           switch ($24|0) {
           case 861165636:  {
            break;
           }
           default: {
            $$1 = 0;
            break L1;
           }
           }
           $$1 = 1;
           break L1;
          }
         } else {
          $switch$split42D = ($24|0)<(1395999554);
          if ($switch$split42D) {
           $switch$split132D = ($24|0)<(1395934018);
           if ($switch$split132D) {
            switch ($24|0) {
            case 894720068:  {
             break;
            }
            default: {
             $$1 = 0;
             break L1;
            }
            }
            $$1 = 1;
            break L1;
           } else {
            switch ($24|0) {
            case 1395934018:  {
             break;
            }
            default: {
             $$1 = 0;
             break L1;
            }
            }
            $$1 = 1;
            break L1;
           }
          }
          $switch$split162D = ($24|0)<(1429488450);
          if ($switch$split162D) {
           switch ($24|0) {
           case 1395999554:  {
            break;
           }
           default: {
            $$1 = 0;
            break L1;
           }
           }
           $$1 = 1;
           break L1;
          }
          $switch$split192D = ($24|0)<(1429553986);
          if ($switch$split192D) {
           switch ($24|0) {
           case 1429488450:  {
            label = 13;
            break L12;
            break;
           }
           default: {
            $$1 = 0;
            break L1;
           }
           }
          } else {
           switch ($24|0) {
           case 1429553986:  {
            label = 15;
            break L12;
            break;
           }
           default: {
            $$1 = 0;
            break L1;
           }
           }
          }
         }
        } while(0);
        if ((label|0) == 13) {
         $$1 = 1;
         break;
        }
        else if ((label|0) == 15) {
         $$1 = 1;
         break;
        }
       }
      }
     } else {
      $$1 = 0;
     }
    } else {
     $$1 = 0;
    }
   } else {
    $$1 = 0;
   }
  }
 } while(0);
 return ($$1|0); //@line 189 "libraries/ddsparse/ddsparse.cpp"
}
function __ZN3dds6ParserC2EPKvj($this,$data,$dataSize) {
 $this = $this|0;
 $data = $data|0;
 $dataSize = $dataSize|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $0 = ((($this)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($this)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ((($this)) + 12|0); //@line 192 "libraries/ddsparse/ddsparse.cpp"
 HEAP32[$2>>2] = 11; //@line 192 "libraries/ddsparse/ddsparse.cpp"
 __THREW__ = 0;
 (invoke_iiii(4261,($this|0),($data|0),($dataSize|0))|0); //@line 194 "libraries/ddsparse/ddsparse.cpp"
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  return; //@line 195 "libraries/ddsparse/ddsparse.cpp"
 }
 $5 = ___cxa_find_matching_catch_2()|0;
 $6 = tempRet0;
 $7 = HEAP32[$this>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = ($7|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = $7; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($8) {
  ___resumeException($5|0);
  // unreachable;
 }
 $10 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10|0)==($7|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($11)) {
  $scevgep$i$i$i = ((($10)) + -16|0);
  $12 = $scevgep$i$i$i;
  $13 = (($12) - ($9))|0;
  $14 = $13 >>> 4;
  $15 = $14 ^ -1;
  $scevgep4$i$i$i = (($10) + ($15<<4)|0);
  HEAP32[$0>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($7); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 ___resumeException($5|0);
 // unreachable;
}
function __ZN3dds6Parser9parseDataEPKvj($this,$data,$dataSize) {
 $this = $this|0;
 $data = $data|0;
 $dataSize = $dataSize|0;
 var $$0$i = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f$0$i = 0, $offset$0 = 0, $or$cond$i = 0, $switch$split102D = 0, $switch$split12D = 0, $switch$split132D = 0, $switch$split162D = 0, $switch$split192D = 0, $switch$split2D = 0, $switch$split42D = 0, $switch$split72D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($dataSize>>>0)<(128); //@line 141 "libraries/ddsparse/ddsparse.cpp"
 if ($0) {
  $$2 = 0;
  return ($$2|0); //@line 355 "libraries/ddsparse/ddsparse.cpp"
 }
 $1 = HEAP32[$data>>2]|0; //@line 145 "libraries/ddsparse/ddsparse.cpp"
 $2 = ($1|0)==(542327876); //@line 145 "libraries/ddsparse/ddsparse.cpp"
 if (!($2)) {
  $$2 = 0;
  return ($$2|0); //@line 355 "libraries/ddsparse/ddsparse.cpp"
 }
 $3 = ((($data)) + 4|0); //@line 150 "libraries/ddsparse/ddsparse.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 154 "libraries/ddsparse/ddsparse.cpp"
 $5 = ($4|0)==(124); //@line 154 "libraries/ddsparse/ddsparse.cpp"
 if (!($5)) {
  $$2 = 0;
  return ($$2|0); //@line 355 "libraries/ddsparse/ddsparse.cpp"
 }
 $6 = ((($data)) + 76|0); //@line 154 "libraries/ddsparse/ddsparse.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 154 "libraries/ddsparse/ddsparse.cpp"
 $8 = ($7|0)==(32); //@line 154 "libraries/ddsparse/ddsparse.cpp"
 if (!($8)) {
  $$2 = 0;
  return ($$2|0); //@line 355 "libraries/ddsparse/ddsparse.cpp"
 }
 $9 = ((($data)) + 80|0); //@line 160 "libraries/ddsparse/ddsparse.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 41 "libraries/ddsparse/ddsparse.cpp"
 $11 = $10 & 4; //@line 160 "libraries/ddsparse/ddsparse.cpp"
 $12 = ($11|0)==(0); //@line 160 "libraries/ddsparse/ddsparse.cpp"
 L13: do {
  if ($12) {
   $$0$i = 11;
   label = 31;
  } else {
   $13 = ((($data)) + 84|0); //@line 160 "libraries/ddsparse/ddsparse.cpp"
   $14 = HEAP32[$13>>2]|0; //@line 160 "libraries/ddsparse/ddsparse.cpp"
   $15 = ($14|0)==(808540228); //@line 160 "libraries/ddsparse/ddsparse.cpp"
   $16 = ($dataSize>>>0)<(148); //@line 163 "libraries/ddsparse/ddsparse.cpp"
   $or$cond$i = $16 & $15; //@line 160 "libraries/ddsparse/ddsparse.cpp"
   if ($or$cond$i) {
    $$2 = 0;
    return ($$2|0); //@line 355 "libraries/ddsparse/ddsparse.cpp"
   }
   $17 = ((($data)) + 84|0); //@line 320 "libraries/ddsparse/ddsparse.cpp"
   $18 = HEAP32[$17>>2]|0; //@line 320 "libraries/ddsparse/ddsparse.cpp"
   $19 = ($18|0)==(808540228); //@line 320 "libraries/ddsparse/ddsparse.cpp"
   if ($19) {
    $20 = ((($data)) + 128|0); //@line 322 "libraries/ddsparse/ddsparse.cpp"
    $21 = ((($data)) + 132|0); //@line 326 "libraries/ddsparse/ddsparse.cpp"
    $22 = HEAP32[$21>>2]|0; //@line 326 "libraries/ddsparse/ddsparse.cpp"
    switch ($22|0) {
    case 0: case 3:  {
     break;
    }
    default: {
     $$2 = 0;
     return ($$2|0); //@line 355 "libraries/ddsparse/ddsparse.cpp"
    }
    }
    $23 = ((($data)) + 140|0); //@line 336 "libraries/ddsparse/ddsparse.cpp"
    $24 = HEAP32[$23>>2]|0; //@line 336 "libraries/ddsparse/ddsparse.cpp"
    $25 = ($24>>>0)>(1); //@line 336 "libraries/ddsparse/ddsparse.cpp"
    if ($25) {
     $$2 = 0;
     return ($$2|0); //@line 355 "libraries/ddsparse/ddsparse.cpp"
    }
    $26 = HEAP32[$20>>2]|0; //@line 339 "libraries/ddsparse/ddsparse.cpp"
    do {
     switch ($26|0) {
     case 72: case 71: case 70:  {
      $f$0$i = 0;
      break;
     }
     case 75: case 74: case 73:  {
      $f$0$i = 1;
      break;
     }
     case 78: case 77: case 76:  {
      $f$0$i = 2;
      break;
     }
     case 80: case 79:  {
      $f$0$i = 3;
      break;
     }
     case 81:  {
      $f$0$i = 4;
      break;
     }
     case 83: case 82:  {
      $f$0$i = 5;
      break;
     }
     case 84:  {
      $f$0$i = 6;
      break;
     }
     case 95: case 94:  {
      $f$0$i = 7;
      break;
     }
     case 96:  {
      $f$0$i = 8;
      break;
     }
     case 98: case 97:  {
      $f$0$i = 9;
      break;
     }
     case 99:  {
      $f$0$i = 10;
      break;
     }
     default: {
      $f$0$i = 11;
     }
     }
    } while(0);
    $27 = ((($this)) + 12|0); //@line 339 "libraries/ddsparse/ddsparse.cpp"
    HEAP32[$27>>2] = $f$0$i; //@line 339 "libraries/ddsparse/ddsparse.cpp"
    $32 = $f$0$i;$offset$0 = 148;
    break;
   }
   $28 = ((($data)) + 84|0); //@line 46 "libraries/ddsparse/ddsparse.cpp"
   $29 = HEAP32[$28>>2]|0; //@line 46 "libraries/ddsparse/ddsparse.cpp"
   $switch$split2D = ($29|0)<(894720068);
   L40: do {
    if ($switch$split2D) {
     $switch$split12D = ($29|0)<(843666497);
     if ($switch$split12D) {
      $switch$split72D = ($29|0)<(827611204);
      if ($switch$split72D) {
       switch ($29|0) {
       case 826889281:  {
        label = 27;
        break L40;
        break;
       }
       default: {
        $$0$i = 11;
        label = 31;
        break L13;
       }
       }
      }
      switch ($29|0) {
      case 827611204:  {
       break;
      }
      default: {
       $$0$i = 11;
       label = 31;
       break L13;
      }
      }
      $$0$i = 0;
      label = 31;
      break L13;
     } else {
      $switch$split102D = ($29|0)<(861165636);
      if ($switch$split102D) {
       switch ($29|0) {
       case 843666497:  {
        label = 29;
        break L40;
        break;
       }
       default: {
        $$0$i = 11;
        label = 31;
        break L13;
       }
       }
      }
      switch ($29|0) {
      case 861165636:  {
       break;
      }
      default: {
       $$0$i = 11;
       label = 31;
       break L13;
      }
      }
      $$0$i = 1;
      label = 31;
      break L13;
     }
    } else {
     $switch$split42D = ($29|0)<(1395999554);
     if ($switch$split42D) {
      $switch$split132D = ($29|0)<(1395934018);
      if ($switch$split132D) {
       switch ($29|0) {
       case 894720068:  {
        break;
       }
       default: {
        $$0$i = 11;
        label = 31;
        break L13;
       }
       }
       $$0$i = 2;
       label = 31;
       break L13;
      } else {
       switch ($29|0) {
       case 1395934018:  {
        break;
       }
       default: {
        $$0$i = 11;
        label = 31;
        break L13;
       }
       }
       $$0$i = 4;
       label = 31;
       break L13;
      }
     }
     $switch$split162D = ($29|0)<(1429488450);
     if ($switch$split162D) {
      switch ($29|0) {
      case 1395999554:  {
       break;
      }
      default: {
       $$0$i = 11;
       label = 31;
       break L13;
      }
      }
      $$0$i = 6;
      label = 31;
      break L13;
     }
     $switch$split192D = ($29|0)<(1429553986);
     if ($switch$split192D) {
      switch ($29|0) {
      case 1429488450:  {
       label = 27;
       break L40;
       break;
      }
      default: {
       $$0$i = 11;
       label = 31;
       break L13;
      }
      }
     } else {
      switch ($29|0) {
      case 1429553986:  {
       label = 29;
       break L40;
       break;
      }
      default: {
       $$0$i = 11;
       label = 31;
       break L13;
      }
      }
     }
    }
   } while(0);
   if ((label|0) == 27) {
    $$0$i = 3;
    label = 31;
    break;
   }
   else if ((label|0) == 29) {
    $$0$i = 5;
    label = 31;
    break;
   }
  }
 } while(0);
 if ((label|0) == 31) {
  $30 = ((($this)) + 12|0); //@line 342 "libraries/ddsparse/ddsparse.cpp"
  HEAP32[$30>>2] = $$0$i; //@line 342 "libraries/ddsparse/ddsparse.cpp"
  $32 = $$0$i;$offset$0 = 128;
 }
 $31 = ($32|0)==(11); //@line 345 "libraries/ddsparse/ddsparse.cpp"
 if ($31) {
  $$2 = 0;
  return ($$2|0); //@line 355 "libraries/ddsparse/ddsparse.cpp"
 }
 $33 = ((($data)) + 16|0); //@line 349 "libraries/ddsparse/ddsparse.cpp"
 $34 = HEAP32[$33>>2]|0; //@line 349 "libraries/ddsparse/ddsparse.cpp"
 $35 = ((($data)) + 12|0); //@line 350 "libraries/ddsparse/ddsparse.cpp"
 $36 = HEAP32[$35>>2]|0; //@line 350 "libraries/ddsparse/ddsparse.cpp"
 $37 = ((($data)) + 28|0); //@line 352 "libraries/ddsparse/ddsparse.cpp"
 $38 = HEAP32[$37>>2]|0; //@line 352 "libraries/ddsparse/ddsparse.cpp"
 $39 = ($38|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
 $40 = $39 ? 1 : $38; //@line 352 "libraries/ddsparse/ddsparse.cpp"
 $41 = (($data) + ($offset$0)|0); //@line 354 "libraries/ddsparse/ddsparse.cpp"
 $42 = (($dataSize) - ($offset$0))|0; //@line 354 "libraries/ddsparse/ddsparse.cpp"
 $43 = (__ZN3dds6Parser12parseTexDataEPKhjNS_6FormatEiii($this,$41,$42,$32,$34,$36,$40)|0); //@line 354 "libraries/ddsparse/ddsparse.cpp"
 $$2 = $43;
 return ($$2|0); //@line 355 "libraries/ddsparse/ddsparse.cpp"
}
function __ZN3dds6Parser12parseTexDataEPKhjNS_6FormatEiii($this,$data,$dataSize,$fmt,$w,$h,$mips) {
 $this = $this|0;
 $data = $data|0;
 $dataSize = $dataSize|0;
 $fmt = $fmt|0;
 $w = $w|0;
 $h = $h|0;
 $mips = $mips|0;
 var $$0224 = 0, $$0224$us = 0, $$0621 = 0, $$0621$us = 0, $$3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $i$023 = 0, $i$023$us = 0, $img = 0, $lpad$phi$index = 0, $lpad$phi$index3 = 0, $newTexData = 0, $numBlocksHigh$0$i = 0, $numBlocksHigh$0$i$us = 0, $numBlocksWide$0$i = 0, $numBlocksWide$0$i$us = 0, $numBytesPerBlock$0$i = 0, $offset$022 = 0, $offset$022$us = 0, $or$cond = 0, $or$cond$us = 0, $scevgep$i$i$i = 0, $scevgep$i$i$i10 = 0, $scevgep4$i$i$i = 0;
 var $scevgep4$i$i$i11 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $newTexData = sp + 16|0;
 $img = sp;
 HEAP32[$newTexData>>2] = 0; //@line 451 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $0 = ((($newTexData)) + 4|0); //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$0>>2] = 0; //@line 452 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ((($newTexData)) + 8|0); //@line 453 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$1>>2] = 0; //@line 2138 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $2 = ($mips|0)>(0); //@line 278 "libraries/ddsparse/ddsparse.cpp"
 L1: do {
  if ($2) {
   $3 = ((($img)) + 4|0); //@line 282 "libraries/ddsparse/ddsparse.cpp"
   $4 = ((($img)) + 8|0); //@line 284 "libraries/ddsparse/ddsparse.cpp"
   $5 = ((($img)) + 12|0); //@line 291 "libraries/ddsparse/ddsparse.cpp"
   $6 = ((($newTexData)) + 8|0); //@line 2198 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $7 = ($fmt|0)==(0);
   L3: do {
    if ($7) {
     $$0224$us = $w;$$0621$us = $h;$i$023$us = 0;$offset$022$us = 0;
     while(1) {
      ;HEAP32[$img>>2]=0|0;HEAP32[$img+4>>2]=0|0;HEAP32[$img+8>>2]=0|0;HEAP32[$img+12>>2]=0|0; //@line 62 "libraries/ddsparse/ddsparse.h"
      HEAP32[$img>>2] = $$0224$us; //@line 281 "libraries/ddsparse/ddsparse.cpp"
      HEAP32[$3>>2] = $$0621$us; //@line 282 "libraries/ddsparse/ddsparse.cpp"
      $8 = ($$0224$us|0)>(0); //@line 264 "libraries/ddsparse/ddsparse.cpp"
      if ($8) {
       $9 = (($$0224$us) + 3)|0; //@line 265 "libraries/ddsparse/ddsparse.cpp"
       $10 = (($9|0) / 4)&-1; //@line 265 "libraries/ddsparse/ddsparse.cpp"
       $11 = ($10|0)>(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $12 = $11 ? $10 : 1; //@line 265 "libraries/ddsparse/ddsparse.cpp"
       $numBlocksWide$0$i$us = $12;
      } else {
       $numBlocksWide$0$i$us = 0;
      }
      $13 = ($$0621$us|0)>(0); //@line 267 "libraries/ddsparse/ddsparse.cpp"
      if ($13) {
       $14 = (($$0621$us) + 3)|0; //@line 268 "libraries/ddsparse/ddsparse.cpp"
       $15 = (($14|0) / 4)&-1; //@line 268 "libraries/ddsparse/ddsparse.cpp"
       $16 = ($15|0)>(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $17 = $16 ? $15 : 1; //@line 268 "libraries/ddsparse/ddsparse.cpp"
       $numBlocksHigh$0$i$us = $17;
      } else {
       $numBlocksHigh$0$i$us = 0;
      }
      $18 = $numBlocksWide$0$i$us << 3; //@line 270 "libraries/ddsparse/ddsparse.cpp"
      $19 = Math_imul($18, $numBlocksHigh$0$i$us)|0; //@line 296 "libraries/ddsparse/ddsparse.cpp"
      HEAP32[$4>>2] = $19; //@line 284 "libraries/ddsparse/ddsparse.cpp"
      $20 = ($19|0)==(0); //@line 287 "libraries/ddsparse/ddsparse.cpp"
      $21 = (($19) + ($offset$022$us))|0; //@line 287 "libraries/ddsparse/ddsparse.cpp"
      $22 = ($21>>>0)>($dataSize>>>0); //@line 287 "libraries/ddsparse/ddsparse.cpp"
      $or$cond$us = $20 | $22; //@line 287 "libraries/ddsparse/ddsparse.cpp"
      if ($or$cond$us) {
       break L3;
      }
      $23 = (($data) + ($offset$022$us)|0); //@line 291 "libraries/ddsparse/ddsparse.cpp"
      HEAP32[$5>>2] = $23; //@line 291 "libraries/ddsparse/ddsparse.cpp"
      $24 = HEAP32[$0>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $25 = HEAP32[$6>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $26 = ($24|0)==($25|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($26) {
       __THREW__ = 0;
       invoke_vii(4262,($newTexData|0),($img|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $28 = __THREW__; __THREW__ = 0;
       $29 = $28&1;
       if ($29) {
        break;
       }
      } else {
       ;HEAP32[$24>>2]=HEAP32[$img>>2]|0;HEAP32[$24+4>>2]=HEAP32[$img+4>>2]|0;HEAP32[$24+8>>2]=HEAP32[$img+8>>2]|0;HEAP32[$24+12>>2]=HEAP32[$img+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $27 = ((($24)) + 16|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$0>>2] = $27; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      $30 = (($$0224$us|0) / 2)&-1; //@line 298 "libraries/ddsparse/ddsparse.cpp"
      $31 = ($30|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $32 = $31 ? 1 : $30; //@line 298 "libraries/ddsparse/ddsparse.cpp"
      $33 = (($$0621$us|0) / 2)&-1; //@line 299 "libraries/ddsparse/ddsparse.cpp"
      $34 = ($33|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $35 = $34 ? 1 : $33; //@line 299 "libraries/ddsparse/ddsparse.cpp"
      $36 = (($i$023$us) + 1)|0; //@line 278 "libraries/ddsparse/ddsparse.cpp"
      $37 = ($36|0)<($mips|0); //@line 278 "libraries/ddsparse/ddsparse.cpp"
      if ($37) {
       $$0224$us = $32;$$0621$us = $35;$i$023$us = $36;$offset$022$us = $21;
      } else {
       label = 33;
       break L1;
      }
     }
     $38 = ___cxa_find_matching_catch_2()|0;
     $39 = tempRet0;
     $lpad$phi$index = $38;$lpad$phi$index3 = $39;
     label = 23;
     break L1;
    } else {
     $$0224 = $w;$$0621 = $h;$i$023 = 0;$offset$022 = 0;
     while(1) {
      ;HEAP32[$img>>2]=0|0;HEAP32[$img+4>>2]=0|0;HEAP32[$img+8>>2]=0|0;HEAP32[$img+12>>2]=0|0; //@line 62 "libraries/ddsparse/ddsparse.h"
      HEAP32[$img>>2] = $$0224; //@line 281 "libraries/ddsparse/ddsparse.cpp"
      HEAP32[$3>>2] = $$0621; //@line 282 "libraries/ddsparse/ddsparse.cpp"
      switch ($fmt|0) {
      case 4: case 3: case 0:  {
       $numBytesPerBlock$0$i = 8;
       break;
      }
      case 10: case 9: case 7: case 5: case 6: case 2: case 1:  {
       $numBytesPerBlock$0$i = 16;
       break;
      }
      default: {
       $numBytesPerBlock$0$i = 0;
      }
      }
      $40 = ($$0224|0)>(0); //@line 264 "libraries/ddsparse/ddsparse.cpp"
      if ($40) {
       $41 = (($$0224) + 3)|0; //@line 265 "libraries/ddsparse/ddsparse.cpp"
       $42 = (($41|0) / 4)&-1; //@line 265 "libraries/ddsparse/ddsparse.cpp"
       $43 = ($42|0)>(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $44 = $43 ? $42 : 1; //@line 265 "libraries/ddsparse/ddsparse.cpp"
       $numBlocksWide$0$i = $44;
      } else {
       $numBlocksWide$0$i = 0;
      }
      $45 = ($$0621|0)>(0); //@line 267 "libraries/ddsparse/ddsparse.cpp"
      if ($45) {
       $46 = (($$0621) + 3)|0; //@line 268 "libraries/ddsparse/ddsparse.cpp"
       $47 = (($46|0) / 4)&-1; //@line 268 "libraries/ddsparse/ddsparse.cpp"
       $48 = ($47|0)>(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
       $49 = $48 ? $47 : 1; //@line 268 "libraries/ddsparse/ddsparse.cpp"
       $numBlocksHigh$0$i = $49;
      } else {
       $numBlocksHigh$0$i = 0;
      }
      $50 = Math_imul($numBlocksWide$0$i, $numBytesPerBlock$0$i)|0; //@line 270 "libraries/ddsparse/ddsparse.cpp"
      $51 = Math_imul($50, $numBlocksHigh$0$i)|0; //@line 296 "libraries/ddsparse/ddsparse.cpp"
      HEAP32[$4>>2] = $51; //@line 284 "libraries/ddsparse/ddsparse.cpp"
      $52 = ($51|0)==(0); //@line 287 "libraries/ddsparse/ddsparse.cpp"
      $53 = (($51) + ($offset$022))|0; //@line 287 "libraries/ddsparse/ddsparse.cpp"
      $54 = ($53>>>0)>($dataSize>>>0); //@line 287 "libraries/ddsparse/ddsparse.cpp"
      $or$cond = $52 | $54; //@line 287 "libraries/ddsparse/ddsparse.cpp"
      if ($or$cond) {
       break L3;
      }
      $68 = (($data) + ($offset$022)|0); //@line 291 "libraries/ddsparse/ddsparse.cpp"
      HEAP32[$5>>2] = $68; //@line 291 "libraries/ddsparse/ddsparse.cpp"
      $69 = HEAP32[$0>>2]|0; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $70 = HEAP32[$6>>2]|0; //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      $71 = ($69|0)==($70|0); //@line 1591 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      if ($71) {
       __THREW__ = 0;
       invoke_vii(4262,($newTexData|0),($img|0)); //@line 1600 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       $73 = __THREW__; __THREW__ = 0;
       $74 = $73&1;
       if ($74) {
        break;
       }
      } else {
       ;HEAP32[$69>>2]=HEAP32[$img>>2]|0;HEAP32[$69+4>>2]=HEAP32[$img+4>>2]|0;HEAP32[$69+8>>2]=HEAP32[$img+8>>2]|0;HEAP32[$69+12>>2]=HEAP32[$img+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
       $72 = ((($69)) + 16|0); //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
       HEAP32[$0>>2] = $72; //@line 1597 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
      }
      $75 = (($$0224|0) / 2)&-1; //@line 298 "libraries/ddsparse/ddsparse.cpp"
      $76 = ($75|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $77 = $76 ? 1 : $75; //@line 298 "libraries/ddsparse/ddsparse.cpp"
      $78 = (($$0621|0) / 2)&-1; //@line 299 "libraries/ddsparse/ddsparse.cpp"
      $79 = ($78|0)<(1); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
      $80 = $79 ? 1 : $78; //@line 299 "libraries/ddsparse/ddsparse.cpp"
      $81 = (($i$023) + 1)|0; //@line 278 "libraries/ddsparse/ddsparse.cpp"
      $82 = ($81|0)<($mips|0); //@line 278 "libraries/ddsparse/ddsparse.cpp"
      if ($82) {
       $$0224 = $77;$$0621 = $80;$i$023 = $81;$offset$022 = $53;
      } else {
       label = 33;
       break L1;
      }
     }
     $55 = ___cxa_find_matching_catch_2()|0;
     $56 = tempRet0;
     $lpad$phi$index = $55;$lpad$phi$index3 = $56;
     label = 23;
     break L1;
    }
   } while(0);
   $$3 = 0;
   label = 35;
  } else {
   label = 33;
  }
 } while(0);
 if ((label|0) == 33) {
  $83 = ($this|0)==($newTexData|0); //@line 1359 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($83) {
   $$3 = 1;
   label = 35;
  } else {
   $84 = HEAP32[$newTexData>>2]|0; //@line 1362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $85 = HEAP32[$0>>2]|0; //@line 1362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   __THREW__ = 0;
   invoke_viii(4263,($this|0),($84|0),($85|0)); //@line 1362 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $86 = __THREW__; __THREW__ = 0;
   $87 = $86&1;
   if ($87) {
    $57 = ___cxa_find_matching_catch_2()|0;
    $58 = tempRet0;
    $lpad$phi$index = $57;$lpad$phi$index3 = $58;
    label = 23;
   } else {
    $$3 = 1;
    label = 35;
   }
  }
 }
 if ((label|0) == 23) {
  $59 = HEAP32[$newTexData>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $60 = ($59|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $61 = $59; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($60) {
   ___resumeException($lpad$phi$index|0);
   // unreachable;
  }
  $62 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $63 = ($62|0)==($59|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($63)) {
   $scevgep$i$i$i10 = ((($62)) + -16|0);
   $64 = $scevgep$i$i$i10;
   $65 = (($64) - ($61))|0;
   $66 = $65 >>> 4;
   $67 = $66 ^ -1;
   $scevgep4$i$i$i11 = (($62) + ($67<<4)|0);
   HEAP32[$0>>2] = $scevgep4$i$i$i11; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($59); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  ___resumeException($lpad$phi$index|0);
  // unreachable;
 }
 else if ((label|0) == 35) {
  $88 = HEAP32[$newTexData>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $89 = ($88|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $90 = $88; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($89) {
   STACKTOP = sp;return ($$3|0); //@line 305 "libraries/ddsparse/ddsparse.cpp"
  }
  $91 = HEAP32[$0>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $92 = ($91|0)==($88|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($92)) {
   $scevgep$i$i$i = ((($91)) + -16|0);
   $93 = $scevgep$i$i$i;
   $94 = (($93) - ($90))|0;
   $95 = $94 >>> 4;
   $96 = $95 ^ -1;
   $scevgep4$i$i$i = (($91) + ($96<<4)|0);
   HEAP32[$0>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($88); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  STACKTOP = sp;return ($$3|0); //@line 305 "libraries/ddsparse/ddsparse.cpp"
 }
 return (0)|0;
}
function __ZNSt3__16vectorIN3dds5ImageENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_($this,$__x) {
 $this = $this|0;
 $__x = $__x|0;
 var $$0$i1 = 0, $$0$i2 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = (($4) + 1)|0; //@line 1579 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = ($5>>>0)>(268435455); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($6) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$in = $$pre;
 } else {
  $$in = $2;
 }
 $7 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $8 = HEAP32[$7>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = (($8) - ($$in))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $9 >> 4; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $11 = ($10>>>0)<(134217727); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($11) {
  $15 = $9 >> 3; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = ($15>>>0)<($5>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $17 = $16 ? $5 : $15; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $18 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $19 = (($18) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $20 = $19 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $21 = ($17|0)==(0); //@line 325 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
  if ($21) {
   $$0$i2 = 0;$25 = 0;$26 = $20;$32 = $18;
  } else {
   $$0$i1 = $17;$40 = $18;$41 = $20;
   label = 6;
  }
 } else {
  $12 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($$in))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$0$i1 = 268435455;$40 = $12;$41 = $14;
  label = 6;
 }
 if ((label|0) == 6) {
  $22 = $$0$i1 << 4; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $23 = (__Znwj($22)|0); //@line 156 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  $$0$i2 = $$0$i1;$25 = $23;$26 = $41;$32 = $40;
 }
 $24 = (($25) + ($26<<4)|0); //@line 326 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $27 = (($25) + ($$0$i2<<4)|0); //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 $28 = $27; //@line 327 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 ;HEAP32[$24>>2]=HEAP32[$__x>>2]|0;HEAP32[$24+4>>2]=HEAP32[$__x+4>>2]|0;HEAP32[$24+8>>2]=HEAP32[$__x+8>>2]|0;HEAP32[$24+12>>2]=HEAP32[$__x+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $29 = ((($24)) + 16|0); //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $30 = $29; //@line 1582 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $31 = (($32) - ($$in))|0; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $33 = $31 >> 4; //@line 1552 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $34 = (0 - ($33))|0; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $35 = (($24) + ($34<<4)|0); //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $36 = $35; //@line 1553 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $37 = $$in; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 _memcpy(($35|0),($37|0),($31|0))|0; //@line 1554 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 HEAP32[$this>>2] = $36; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$0>>2] = $30; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 HEAP32[$7>>2] = $28; //@line 3537 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/type_traits"
 $38 = ($$in|0)==(0); //@line 356 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/__split_buffer"
 if ($38) {
  return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $39 = $$in; //@line 1636 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 __ZdlPv($39); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 1584 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZNSt3__16vectorIN3dds5ImageENS_9allocatorIS2_EEE6assignIPS2_EENS_9enable_ifIXaasr21__is_forward_iteratorIT_EE5valuesr16is_constructibleIS2_NS_15iterator_traitsIS9_E9referenceEEE5valueEvE4typeES9_S9_($this,$__first,$__last) {
 $this = $this|0;
 $__first = $__first|0;
 $__last = $__last|0;
 var $$0$i5 = 0, $$01$i = 0, $$01$i3 = 0, $$__last = 0, $$pre = 0, $$pre$i1 = 0, $$pre8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, $__new_last2$i$i = 0, $scevgep$i$i = 0, $scevgep$i$i$i$i = 0, $scevgep4$i$i = 0, $scevgep4$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $__last; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $1 = $__first; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $2 = (($0) - ($1))|0; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $3 = $2 >> 4; //@line 494 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
 $4 = ((($this)) + 8|0); //@line 2199 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $5 = HEAP32[$4>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $6 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = (($5) - ($6))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $8 = $7 >> 4; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $9 = ($3>>>0)>($8>>>0); //@line 1398 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $10 = $6; //@line 1408 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($9)) {
  $11 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $12 = HEAP32[$11>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $13 = (($12) - ($6))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $14 = $13 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $15 = ($3>>>0)>($14>>>0); //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $16 = (($__first) + ($14<<4)|0); //@line 467 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/iterator"
  $$__last = $15 ? $16 : $__last; //@line 1402 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $17 = $$__last; //@line 1763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $18 = (($17) - ($1))|0; //@line 1763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $19 = $18 >> 4; //@line 1763 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $20 = $6; //@line 1764 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  _memmove(($20|0),($__first|0),($18|0))|0; //@line 1764 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $21 = (($10) + ($19<<4)|0); //@line 1765 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  if (!($15)) {
   $28 = HEAP32[$11>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $29 = ($28|0)==($21|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($29) {
    return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   }
   $__new_last2$i$i = $21; //@line 818 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $scevgep$i$i = ((($28)) + -16|0);
   $30 = $scevgep$i$i;
   $31 = (($30) - ($__new_last2$i$i))|0;
   $32 = $31 >>> 4;
   $33 = $32 ^ -1;
   $scevgep4$i$i = (($28) + ($33<<4)|0);
   HEAP32[$11>>2] = $scevgep4$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  $22 = ($$__last|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($22) {
   return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  $$pre$i1 = HEAP32[$11>>2]|0; //@line 1030 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$01$i3 = $16;$23 = $$pre$i1;
  while(1) {
   ;HEAP32[$23>>2]=HEAP32[$$01$i3>>2]|0;HEAP32[$23+4>>2]=HEAP32[$$01$i3+4>>2]|0;HEAP32[$23+8>>2]=HEAP32[$$01$i3+8>>2]|0;HEAP32[$23+12>>2]=HEAP32[$$01$i3+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
   $24 = HEAP32[$11>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $25 = ((($24)) + 16|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   HEAP32[$11>>2] = $25; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $26 = ((($$01$i3)) + 16|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $27 = ($26|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   if ($27) {
    break;
   } else {
    $$01$i3 = $26;$23 = $25;
   }
  }
  return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 $34 = ($6|0)==(0); //@line 945 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $35 = $6; //@line 948 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($34) {
  $63 = $5;
 } else {
  $36 = ((($this)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $37 = HEAP32[$36>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $38 = ($37|0)==($10|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if (!($38)) {
   $scevgep$i$i$i$i = ((($37)) + -16|0);
   $39 = $scevgep$i$i$i$i;
   $40 = (($39) - ($6))|0;
   $41 = $40 >>> 4;
   $42 = $41 ^ -1;
   $scevgep4$i$i$i$i = (($37) + ($42<<4)|0);
   HEAP32[$36>>2] = $scevgep4$i$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  }
  __ZdlPv($35); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
  HEAP32[$4>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$36>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$this>>2] = 0; //@line 949 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $63 = 0;
 }
 $43 = ($3>>>0)>(268435455); //@line 967 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($43) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 968 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre = HEAP32[$4>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $$pre8 = HEAP32[$this>>2]|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $45 = $$pre;$46 = $$pre8;
 } else {
  $45 = $63;$46 = 0;
 }
 $44 = (($45) - ($46))|0; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $47 = $44 >> 4; //@line 371 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $48 = ($47>>>0)<(134217727); //@line 970 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($48) {
  $49 = $44 >> 3; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $50 = ($49>>>0)<($3>>>0); //@line 702 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/algorithm"
  $51 = $50 ? $3 : $49; //@line 972 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $52 = ($51>>>0)>(268435455); //@line 934 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($52) {
   __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this); //@line 935 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
   $$0$i5 = $51;
  } else {
   $$0$i5 = $51;
  }
 } else {
  $$0$i5 = 268435455;
 }
 $53 = $$0$i5 << 4; //@line 1634 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
 $54 = (__Znwj($53)|0); //@line 1030 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $55 = ((($this)) + 4|0); //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$55>>2] = $54; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$this>>2] = $54; //@line 936 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $56 = (($54) + ($$0$i5<<4)|0); //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 HEAP32[$4>>2] = $56; //@line 937 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $57 = ($__first|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($57) {
  return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 } else {
  $$01$i = $__first;$58 = $54;
 }
 while(1) {
  ;HEAP32[$58>>2]=HEAP32[$$01$i>>2]|0;HEAP32[$58+4>>2]=HEAP32[$$01$i+4>>2]|0;HEAP32[$58+8>>2]=HEAP32[$$01$i+8>>2]|0;HEAP32[$58+12>>2]=HEAP32[$$01$i+12>>2]|0; //@line 1645 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/memory"
  $59 = HEAP32[$55>>2]|0; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $60 = ((($59)) + 16|0); //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  HEAP32[$55>>2] = $60; //@line 1032 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $61 = ((($$01$i)) + 16|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  $62 = ($61|0)==($__last|0); //@line 1027 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
  if ($62) {
   break;
  } else {
   $$01$i = $61;$58 = $60;
  }
 }
 return; //@line 1420 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
}
function __ZN3dds6ParserD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep$i$i$i = 0, $scevgep4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = ($0|0)==(0|0); //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = $0; //@line 469 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if ($1) {
  return; //@line 218 "libraries/ddsparse/ddsparse.cpp"
 }
 $3 = ((($this)) + 4|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = HEAP32[$3>>2]|0; //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($4|0)==($0|0); //@line 443 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 if (!($5)) {
  $scevgep$i$i$i = ((($4)) + -16|0);
  $6 = $scevgep$i$i$i;
  $7 = (($6) - ($2))|0;
  $8 = $7 >>> 4;
  $9 = $8 ^ -1;
  $scevgep4$i$i$i = (($4) + ($9<<4)|0);
  HEAP32[$3>>2] = $scevgep4$i$i$i; //@line 444 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 }
 __ZdlPv($0); //@line 164 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/new"
 return; //@line 218 "libraries/ddsparse/ddsparse.cpp"
}
function __ZNK3dds6Parser9getFormatEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0); //@line 222 "libraries/ddsparse/ddsparse.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 222 "libraries/ddsparse/ddsparse.cpp"
 return ($1|0); //@line 222 "libraries/ddsparse/ddsparse.cpp"
}
function __ZNK3dds6Parser12getImageDataEj($this,$miplevel) {
 $this = $this|0;
 $miplevel = $miplevel|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $5 = ($4>>>0)>($miplevel>>>0); //@line 227 "libraries/ddsparse/ddsparse.cpp"
 $6 = $2; //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $7 = (($6) + ($miplevel<<4)|0); //@line 1514 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $$0 = $5 ? $7 : 0; //@line 227 "libraries/ddsparse/ddsparse.cpp"
 return ($$0|0); //@line 231 "libraries/ddsparse/ddsparse.cpp"
}
function __ZNK3dds6Parser14getMipmapCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0); //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $1 = HEAP32[$0>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $2 = HEAP32[$this>>2]|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $3 = (($1) - ($2))|0; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 $4 = $3 >> 4; //@line 653 "/Users/tanner/code/c/emsdk_portable/emscripten/incoming/system/include/libcxx/vector"
 return ($4|0); //@line 235 "libraries/ddsparse/ddsparse.cpp"
}
function _INT123_safe_realloc($ptr,$size) {
 $ptr = $ptr|0;
 $size = $size|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($ptr|0)==(0|0);
 if ($0) {
  $1 = (_malloc($size)|0);
  $$0 = $1;
  return ($$0|0);
 } else {
  $2 = (_realloc($ptr,$size)|0);
  $$0 = $2;
  return ($$0|0);
 }
 return (0)|0;
}
function _INT123_compat_close($infd) {
 $infd = $infd|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_close($infd)|0);
 return ($0|0);
}
function _INT123_frame_freq($fr) {
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 19704|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (35716 + ($1<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 return ($3|0);
}
function _INT123_read_frame($fr) {
 $fr = $fr|0;
 var $$ = 0, $$$i = 0, $$$i5 = 0, $$0 = 0, $$0$i27$ph = 0, $$0$i6 = 0, $$0$i6$ph32 = 0, $$1$i23 = 0, $$14$1$i = 0.0, $$14$i = 0.0, $$lcssa$i = 0, $$lcssa$i$lcssa = 0, $$lcssa11$i = 0, $$lcssa521 = 0, $$lcssa547 = 0, $$lcssa559 = 0, $$lcssa560 = 0, $$lcssa565 = 0, $$lcssa569 = 0, $$lcssa574 = 0;
 var $$lcssa575 = 0, $$lcssa576 = 0, $$lcssa6$i = 0, $$lcssa7$i = 0, $$masked$i = 0, $$masked5$i = 0, $$masked6$1$i = 0, $$masked6$i = 0, $$not$i = 0, $$not$i14 = 0, $$phi$trans$insert$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i12 = 0, $$pre23$i = 0, $$pre24$i = 0, $$pre25$i = 0, $$pre26$i = 0, $$pre280 = 0, $$pre281 = 0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0;
 var $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0;
 var $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0;
 var $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0;
 var $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0;
 var $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0;
 var $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0;
 var $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0;
 var $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0;
 var $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0;
 var $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0;
 var $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0;
 var $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0.0, $54 = 0, $540 = 0.0, $541 = 0, $542 = 0.0, $543 = 0.0, $544 = 0, $545 = 0.0, $546 = 0, $547 = 0, $548 = 0;
 var $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0;
 var $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0;
 var $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0;
 var $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0;
 var $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0;
 var $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0.0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0;
 var $657 = 0.0, $658 = 0.0, $659 = 0, $66 = 0, $660 = 0, $661 = 0.0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0.0, $671 = 0.0, $672 = 0.0, $673 = 0.0, $674 = 0;
 var $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0;
 var $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0;
 var $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0;
 var $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0.0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0.0;
 var $747 = 0.0, $748 = 0, $749 = 0, $75 = 0, $750 = 0.0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0;
 var $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0.0, $772 = 0, $773 = 0.0, $774 = 0.0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0.0, $782 = 0.0;
 var $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0.0, $789 = 0.0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0.0, $797 = 0.0, $798 = 0, $799 = 0, $8 = 0, $80 = 0;
 var $800 = 0, $801 = 0, $802 = 0, $803 = 0.0, $804 = 0.0, $805 = 0.0, $806 = 0.0, $807 = 0.0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0;
 var $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0;
 var $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0;
 var $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0;
 var $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0;
 var $891 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bpf$0$i = 0.0, $diff$0$i = 0.0, $freeformat_count = 0, $freeformat_count$i = 0, $gain_offset$0$i = 0.0, $gain_offset$1$i = 0.0, $gt$0$1$i = 0, $gt$0$i = 0;
 var $headcount$0 = 0, $headcount$1 = 0, $headcount$1$lcssa = 0, $headcount$1$lcssa536 = 0, $headcount$1546 = 0, $headcount$2204 = 0, $headcount$3 = 0, $headcount$4 = 0, $headcount$5 = 0, $headcount$5$lcssa = 0, $headcount$6 = 0, $headcount$7 = 0, $headcount$7$lcssa = 0, $headcount$7$lcssa552 = 0, $headcount$7$lcssa553 = 0, $headcount$7558 = 0, $i$0$lcssa$i = 0, $i$021$i = 0, $lame_offset$0$i = 0, $lame_offset$1$i = 0;
 var $lame_offset$2$i = 0, $lame_offset$3$i = 0, $lame_offset$5$1$i = 0, $lame_offset$5$i = 0, $major$i = 0, $minor$i = 0, $nb$i = 0, $newhead = 0, $newhead$i = 0, $newhead$i11 = 0, $nexthead$i = 0, $or$cond = 0, $or$cond$i = 0, $or$cond$i$i = 0, $or$cond$i$i19 = 0, $or$cond$i$i24 = 0, $or$cond$i$us$i = 0, $or$cond$i$us$i16 = 0, $or$cond$i13 = 0, $or$cond$i4 = 0;
 var $or$cond$i7 = 0, $or$cond$lcssa$i = 0, $or$cond112$i = 0, $or$cond13$i = 0, $or$cond3$i = 0, $or$cond3$i$i = 0, $or$cond3$i$i20 = 0, $or$cond3$i$i25 = 0, $or$cond3$i$us$i = 0, $or$cond3$i$us$i17 = 0, $or$cond3$i8 = 0, $or$cond48 = 0, $or$cond49 = 0, $or$cond5$i = 0, $or$cond5$i$i = 0, $or$cond5$i$i21 = 0, $or$cond5$i$i26 = 0, $or$cond5$i$us$i = 0, $or$cond5$i$us$i18 = 0, $or$cond5$i9 = 0;
 var $replay_gain$i = 0, $rest$i = 0, $ret$1 = 0, $switch$split2D = 0, $tmp = 0, $try$0$i = 0, $try$0$us$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer12 = 0, $vararg_buffer16 = 0, $vararg_buffer19 = 0, $vararg_buffer22 = 0, $vararg_buffer25 = 0, $vararg_buffer28 = 0, $vararg_buffer31 = 0, $vararg_buffer34 = 0, $vararg_buffer38 = 0, $vararg_buffer40 = 0, $vararg_buffer43 = 0;
 var $vararg_buffer46 = 0, $vararg_buffer49 = 0, $vararg_buffer5 = 0, $vararg_buffer54 = 0, $vararg_buffer57 = 0, $vararg_buffer60 = 0, $vararg_buffer63 = 0, $vararg_buffer66 = 0, $vararg_buffer69 = 0, $vararg_buffer72 = 0, $vararg_buffer75 = 0, $vararg_buffer79 = 0, $vararg_buffer9 = 0, $vararg_ptr15 = 0, $vararg_ptr37 = 0, $vararg_ptr4 = 0, $vararg_ptr52 = 0, $vararg_ptr53 = 0, $vararg_ptr78 = 0, $vararg_ptr8 = 0;
 var $vararg_ptr82 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 288|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer79 = sp + 224|0;
 $vararg_buffer75 = sp + 216|0;
 $vararg_buffer72 = sp + 208|0;
 $vararg_buffer69 = sp + 200|0;
 $vararg_buffer66 = sp + 192|0;
 $vararg_buffer63 = sp + 184|0;
 $vararg_buffer60 = sp + 176|0;
 $vararg_buffer57 = sp + 168|0;
 $vararg_buffer54 = sp + 160|0;
 $vararg_buffer49 = sp + 144|0;
 $vararg_buffer46 = sp + 136|0;
 $vararg_buffer43 = sp + 128|0;
 $vararg_buffer40 = sp + 120|0;
 $vararg_buffer38 = sp + 112|0;
 $vararg_buffer34 = sp + 96|0;
 $vararg_buffer31 = sp + 88|0;
 $vararg_buffer28 = sp + 80|0;
 $vararg_buffer25 = sp + 72|0;
 $vararg_buffer22 = sp + 64|0;
 $vararg_buffer19 = sp + 56|0;
 $vararg_buffer16 = sp + 48|0;
 $vararg_buffer12 = sp + 40|0;
 $vararg_buffer9 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer1 = sp + 16|0;
 $vararg_buffer = sp + 8|0;
 $nexthead$i = sp + 260|0;
 $newhead$i11 = sp + 256|0;
 $freeformat_count$i = sp + 252|0;
 $newhead$i = sp + 248|0;
 $replay_gain$i = sp;
 $nb$i = sp + 270|0;
 $major$i = sp + 244|0;
 $minor$i = sp + 240|0;
 $rest$i = sp + 264|0;
 $freeformat_count = sp + 236|0;
 $newhead = sp + 232|0;
 HEAP32[$freeformat_count>>2] = 0;
 $0 = ((($fr)) + 19736|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 9584|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($fr)) + 19860|0);
 HEAP32[$4>>2] = $1;
 $5 = ((($fr)) + 28084|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 do {
  if (!($7)) {
   $8 = ($3|0)==(0);
   if ($8) {
    $48 = (($6) + -1)|0;
    HEAP32[$2>>2] = $48;
    break;
   }
   $9 = ((($fr)) + 27904|0);
   HEAP32[$9>>2] = 1;
   $10 = ((($fr)) + 27900|0);
   HEAP32[$10>>2] = 1;
   $11 = (($3) + -1)|0;
   HEAP32[$2>>2] = $11;
   $12 = ((($fr)) + 19784|0);
   HEAP32[$12>>2] = 0;
   $13 = ((($fr)) + 27808|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ((($fr)) + 19788|0);
   HEAP32[$15>>2] = $14;
   $16 = ((($fr)) + 19688|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ($17|0)==(3);
   if ($18) {
    $19 = $14;
    $20 = ((($fr)) + 9548|0);
    $21 = ((($fr)) + 19864|0);
    $22 = HEAP32[$21>>2]|0;
    _memcpy(($19|0),($20|0),($22|0))|0;
   }
   $23 = ((($fr)) + 19696|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = ($24|0)==(0);
   if ($25) {
    $$0 = 1;
    STACKTOP = sp;return ($$0|0);
   }
   $26 = HEAP32[$15>>2]|0;
   $27 = HEAP8[$26>>0]|0;
   $28 = $27&255;
   $29 = $28 << 8;
   $30 = ((($26)) + 1|0);
   $31 = HEAP8[$30>>0]|0;
   $32 = $31&255;
   $33 = $29 | $32;
   $34 = $33 << 8;
   $35 = ((($26)) + 2|0);
   $36 = HEAP8[$35>>0]|0;
   $37 = $36&255;
   $38 = $34 | $37;
   $39 = HEAP32[$12>>2]|0;
   $40 = $38 << $39;
   $41 = (($39) + 16)|0;
   $42 = $40 >>> 8;
   $43 = $42 & 65535;
   $44 = $41 >> 3;
   $45 = (($26) + ($44)|0);
   HEAP32[$15>>2] = $45;
   $46 = $41 & 7;
   HEAP32[$12>>2] = $46;
   $47 = ((($fr)) + 27952|0);
   HEAP32[$47>>2] = $43;
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 $49 = ((($fr)) + 27956|0);
 $50 = ((($fr)) + 19684|0);
 $51 = ((($fr)) + 27820|0);
 $52 = ((($fr)) + 27824|0);
 $53 = ((($fr)) + 28068|0);
 $54 = ((($fr)) + 28064|0);
 $55 = ((($fr)) + 19768|0);
 $56 = HEAP32[37544]|0;
 $57 = ((($fr)) + 28352|0);
 $58 = ((($fr)) + 19868|0);
 $59 = ((($fr)) + 28392|0);
 $60 = ((($fr)) + 28396|0);
 $61 = ((($fr)) + 28397|0);
 $62 = ((($fr)) + 28398|0);
 $63 = ((($fr)) + 28399|0);
 $64 = ((($fr)) + 28400|0);
 $65 = ((($fr)) + 27976|0);
 $66 = ((($fr)) + 28064|0);
 $67 = ((($fr)) + 27816|0);
 $68 = ((($fr)) + 27808|0);
 $69 = ((($fr)) + 27812|0);
 $70 = ((($fr)) + 19748|0);
 $71 = ((($fr)) + 19760|0);
 $72 = ((($fr)) + 19688|0);
 $73 = ((($fr)) + 19652|0);
 $74 = ((($fr)) + 19672|0);
 $75 = ((($nb$i)) + 9|0);
 $76 = ((($nb$i)) + 4|0);
 $77 = ((($fr)) + 19744|0);
 $78 = ((($replay_gain$i)) + 4|0);
 $79 = ((($fr)) + 19816|0);
 $80 = ((($fr)) + 19832|0);
 $81 = ((($fr)) + 19824|0);
 $82 = ((($fr)) + 19820|0);
 $83 = ((($fr)) + 19836|0);
 $84 = ((($fr)) + 19828|0);
 $85 = ((($fr)) + 27828|0);
 $86 = ((($fr)) + 19840|0);
 $87 = ((($fr)) + 27960|0);
 $headcount$0 = 0;
 L14: while(1) {
  $88 = HEAP32[$49>>2]|0;
  $89 = ((($88)) + 44|0);
  $90 = HEAP32[$89>>2]|0;
  $91 = ($90|0)==(0|0);
  if ($91) {
   $93 = $88;
  } else {
   FUNCTION_TABLE_vi[$90 & 8191]($fr);
   $$pre = HEAP32[$49>>2]|0;
   $93 = $$pre;
  }
  $92 = ((($93)) + 12|0);
  $94 = HEAP32[$92>>2]|0;
  $95 = (FUNCTION_TABLE_iii[$94 & 8191]($fr,$newhead)|0);
  $96 = ($95|0)<(1);
  if ($96) {
   $ret$1 = $95;
   break;
  } else {
   $headcount$1 = $headcount$0;
  }
  L19: while(1) {
   HEAP32[$50>>2] = 2;
   $97 = HEAP32[$51>>2]|0;
   $98 = ($97|0)==(0);
   do {
    if (!($98)) {
     $99 = HEAP32[$newhead>>2]|0;
     $100 = ($97|0)==($99|0);
     if ($100) {
      HEAP32[$50>>2] = 0;
      break;
     }
     $101 = $99 ^ $97;
     $102 = $101 & 1575936;
     $103 = ($102|0)==(0);
     if ($103) {
      $104 = $97 & 192;
      $105 = ($104|0)==(192);
      $106 = $99 & 192;
      $107 = ($106|0)==(192);
      $tmp = $105 ^ $107;
      if (!($tmp)) {
       HEAP32[$50>>2] = 1;
      }
     }
    }
   } while(0);
   $108 = HEAP32[$52>>2]|0;
   $109 = ($108|0)==(0);
   L29: do {
    if ($109) {
     $110 = HEAP32[$newhead>>2]|0;
     $111 = ($110>>>0)<(4292870144);
     $112 = $110 & 393216;
     $113 = ($112|0)==(0);
     $or$cond$i = $111 | $113;
     $114 = $110 & 61440;
     $115 = ($114|0)==(61440);
     $or$cond3$i = $115 | $or$cond$i;
     $116 = $110 & 3072;
     $117 = ($116|0)==(3072);
     $or$cond5$i = $117 | $or$cond3$i;
     if ($or$cond5$i) {
      HEAP32[$freeformat_count$i>>2] = 0;
      HEAP32[$newhead$i>>2] = $110;
      $118 = $110 & -256;
      $119 = ($118|0)==(1229206272);
      if ($119) {
       $$lcssa547 = $110;$headcount$1$lcssa = $headcount$1;
       label = 21;
       break L19;
      }
      $125 = HEAP32[$53>>2]|0;
      $126 = $125 & 32;
      $127 = ($126|0)==(0);
      if ($127) {
       $128 = HEAP32[$54>>2]|0;
       $129 = ($128|0)>(1);
       if ($129) {
        $130 = HEAP8[$55>>0]|0;
        $131 = ($130<<24>>24)==(0);
        if ($131) {
         HEAP32[$vararg_buffer>>2] = $110;
         (_fprintf($56,510983,$vararg_buffer)|0);
        }
       }
      }
      $132 = ($110|0)==(1380533830);
      L38: do {
       if ($132) {
        $133 = HEAP32[$53>>2]|0;
        $134 = $133 & 32;
        $135 = ($134|0)==(0);
        if ($135) {
         $136 = HEAP32[$54>>2]|0;
         $137 = ($136|0)>(1);
         if ($137) {
          $138 = HEAP8[$55>>0]|0;
          $139 = ($138<<24>>24)==(0);
          if ($139) {
           (_fwrite(511022,32,1,$56)|0);
          }
         }
        }
        $140 = HEAP32[$49>>2]|0;
        $141 = ((($140)) + 12|0);
        $142 = HEAP32[$141>>2]|0;
        $143 = (FUNCTION_TABLE_iii[$142 & 8191]($fr,$newhead$i)|0);
        $144 = ($143|0)<(1);
        if ($144) {
         $$0$i6 = $143;$headcount$6 = $headcount$1;
        } else {
         while(1) {
          $145 = HEAP32[$newhead$i>>2]|0;
          $146 = ($145|0)==(1684108385);
          $147 = HEAP32[$49>>2]|0;
          if ($146) {
           $$lcssa521 = $147;
           break;
          }
          $148 = ((($147)) + 16|0);
          $149 = HEAP32[$148>>2]|0;
          $150 = (FUNCTION_TABLE_iii[$149 & 8191]($fr,$newhead$i)|0);
          $151 = ($150|0)<(1);
          if ($151) {
           $$0$i6 = $150;$headcount$6 = $headcount$1;
           break L38;
          }
         }
         $152 = ((($$lcssa521)) + 12|0);
         $153 = HEAP32[$152>>2]|0;
         $154 = (FUNCTION_TABLE_iii[$153 & 8191]($fr,$newhead$i)|0);
         $155 = ($154|0)<(1);
         if ($155) {
          $$0$i6 = $154;$headcount$6 = $headcount$1;
         } else {
          $headcount$1$lcssa536 = $headcount$1;
          label = 37;
          break L19;
         }
        }
       } else {
        HEAP32[$newhead>>2] = 0;
        $164 = HEAP32[$57>>2]|0;
        $165 = ($164>>>0)>(65536);
        $$$i5 = $165 ? $164 : 65536;
        $166 = (($headcount$1) + 1)|0;
        $167 = ($$$i5|0)>(-1);
        $$not$i = $167 ^ 1;
        $168 = ($166|0)<($$$i5|0);
        $or$cond112$i = $168 | $$not$i;
        L50: do {
         if ($or$cond112$i) {
          L52: do {
           if ($167) {
            $headcount$3 = $166;
            while(1) {
             $190 = HEAP32[$49>>2]|0;
             $191 = ((($190)) + 16|0);
             $192 = HEAP32[$191>>2]|0;
             $193 = (FUNCTION_TABLE_iii[$192 & 8191]($fr,$newhead$i)|0);
             $194 = ($193|0)<(1);
             if ($194) {
              $$0$i6 = $193;$headcount$6 = $headcount$3;
              break L38;
             }
             $195 = HEAP32[$newhead$i>>2]|0;
             $196 = ($195>>>0)<(4292870144);
             $197 = $195 & 393216;
             $198 = ($197|0)==(0);
             $or$cond$i$i = $196 | $198;
             $199 = $195 & 61440;
             $200 = ($199|0)==(61440);
             $or$cond3$i$i = $200 | $or$cond$i$i;
             $201 = $195 & 3072;
             $202 = ($201|0)==(3072);
             $or$cond5$i$i = $202 | $or$cond3$i$i;
             if (!($or$cond5$i$i)) {
              $205 = (_decode_header($fr,$195,$freeformat_count$i)|0);
              $206 = ($205|0)==(0);
              if (!($206)) {
               $$lcssa11$i = $205;$headcount$4 = $headcount$3;
               break L52;
              }
             }
             $203 = (($headcount$3) + 1)|0;
             $204 = ($203|0)<($$$i5|0);
             if ($204) {
              $headcount$3 = $203;
             } else {
              $headcount$5 = $203;
              break L50;
             }
            }
           } else {
            $169 = HEAP32[$49>>2]|0;
            $170 = ((($169)) + 16|0);
            $171 = HEAP32[$170>>2]|0;
            $172 = (FUNCTION_TABLE_iii[$171 & 8191]($fr,$newhead$i)|0);
            $173 = ($172|0)<(1);
            if ($173) {
             $$0$i6 = $172;$headcount$6 = $166;
             break L38;
            } else {
             $headcount$2204 = $166;
            }
            while(1) {
             $174 = HEAP32[$newhead$i>>2]|0;
             $175 = ($174>>>0)<(4292870144);
             $176 = $174 & 393216;
             $177 = ($176|0)==(0);
             $or$cond$i$us$i = $175 | $177;
             $178 = $174 & 61440;
             $179 = ($178|0)==(61440);
             $or$cond3$i$us$i = $179 | $or$cond$i$us$i;
             $180 = $174 & 3072;
             $181 = ($180|0)==(3072);
             $or$cond5$i$us$i = $181 | $or$cond3$i$us$i;
             if (!($or$cond5$i$us$i)) {
              $182 = (_decode_header($fr,$174,$freeformat_count$i)|0);
              $183 = ($182|0)==(0);
              if (!($183)) {
               $$lcssa11$i = $182;$headcount$4 = $headcount$2204;
               break L52;
              }
             }
             $184 = (($headcount$2204) + 1)|0;
             $185 = HEAP32[$49>>2]|0;
             $186 = ((($185)) + 16|0);
             $187 = HEAP32[$186>>2]|0;
             $188 = (FUNCTION_TABLE_iii[$187 & 8191]($fr,$newhead$i)|0);
             $189 = ($188|0)<(1);
             if ($189) {
              $$0$i6 = $188;$headcount$6 = $184;
              break L38;
             } else {
              $headcount$2204 = $184;
             }
            }
           }
          } while(0);
          $207 = ($$lcssa11$i|0)<(0);
          if ($207) {
           $$0$i6$ph32 = $$lcssa11$i;
           label = 56;
           break L14;
          } else {
           $headcount$5 = $headcount$4;
          }
         } else {
          $headcount$5 = $166;
         }
        } while(0);
        $208 = ($headcount$5|0)<($$$i5|0);
        $or$cond48 = $208 | $$not$i;
        if (!($or$cond48)) {
         $headcount$5$lcssa = $headcount$5;
         label = 54;
         break L14;
        }
        $213 = HEAP32[$newhead$i>>2]|0;
        HEAP32[$newhead>>2] = $213;
        $headcount$7 = $headcount$5;
        break L29;
       }
      } while(0);
      $212 = ($$0$i6|0)<(0);
      if ($212) {
       $ret$1 = $$0$i6;
       break L14;
      } else {
       $headcount$7 = $headcount$6;
      }
     } else {
      $headcount$7 = $headcount$1;
     }
    } else {
     $headcount$7 = $headcount$1;
    }
   } while(0);
   $214 = HEAP32[$newhead>>2]|0;
   $215 = ($214>>>0)<(4292870144);
   $216 = $214 & 393216;
   $217 = ($216|0)==(0);
   $or$cond$i7 = $215 | $217;
   $218 = $214 & 61440;
   $219 = ($218|0)==(61440);
   $or$cond3$i8 = $219 | $or$cond$i7;
   $220 = $214 & 3072;
   $221 = ($220|0)==(3072);
   $or$cond5$i9 = $221 | $or$cond3$i8;
   if ($or$cond5$i9) {
    $224 = $214;
   } else {
    $222 = (_decode_header($fr,$214,$freeformat_count)|0);
    $223 = ($222|0)<(0);
    if ($223) {
     $ret$1 = $222;
     break L14;
    }
    switch ($222|0) {
    case 3:  {
     $headcount$0 = $headcount$7;
     continue L14;
     break;
    }
    case 10:  {
     $ret$1 = 10;
     break L14;
     break;
    }
    case 2:  {
     $headcount$1 = $headcount$7;
     continue L19;
     break;
    }
    case 0:  {
     break;
    }
    default: {
     $headcount$7$lcssa = $headcount$7;
     label = 105;
     break L19;
    }
    }
    $$pre281 = HEAP32[$newhead>>2]|0;
    $224 = $$pre281;
   }
   HEAP32[$newhead$i11>>2] = $224;
   HEAP32[$newhead>>2] = 0;
   $225 = $224 & -256;
   $switch$split2D = ($225|0)<(1413564160);
   if ($switch$split2D) {
    switch ($225|0) {
    case 1229206272:  {
     $$lcssa560 = $224;$headcount$7$lcssa553 = $headcount$7;
     label = 70;
     break L19;
     break;
    }
    default: {
    }
    }
   } else {
    switch ($225|0) {
    case 1413564160:  {
     $$lcssa559 = $224;$headcount$7$lcssa552 = $headcount$7;
     label = 66;
     break L19;
     break;
    }
    default: {
    }
    }
   }
   $252 = HEAP32[$53>>2]|0;
   $253 = $252 & 32;
   $254 = ($253|0)==(0);
   if ($254) {
    $255 = HEAP8[$55>>0]|0;
    $256 = ($255<<24>>24)==(0);
    if ($256) {
     $257 = HEAP32[$49>>2]|0;
     $258 = ((($257)) + 36|0);
     $259 = HEAP32[$258>>2]|0;
     $260 = (FUNCTION_TABLE_ii[$259 & 8191]($fr)|0);
     $261 = (($260) + -4)|0;
     HEAP32[$vararg_buffer5>>2] = $224;
     $vararg_ptr8 = ((($vararg_buffer5)) + 4|0);
     HEAP32[$vararg_ptr8>>2] = $261;
     (_fprintf($56,512154,$vararg_buffer5)|0);
     $$pre$i12 = HEAP32[$53>>2]|0;
     $263 = $$pre$i12;
    } else {
     $263 = $252;
    }
   } else {
    $263 = $252;
   }
   $262 = $263 & 32;
   $264 = ($262|0)==(0);
   $265 = ($225|0)==(1651339264);
   $or$cond$i13 = $265 & $264;
   if ($or$cond$i13) {
    (_fwrite(512210,32,1,$56)|0);
    $$pre23$i = HEAP32[$53>>2]|0;
    $267 = $$pre23$i;
   } else {
    $267 = $263;
   }
   $266 = $267 & 128;
   $268 = ($266|0)==(0);
   if (!($268)) {
    $$lcssa565 = $267;
    label = 99;
    break L14;
   }
   $269 = HEAP32[$57>>2]|0;
   HEAP32[$58>>2] = 0;
   $270 = $267 & 32;
   $271 = ($270|0)==(0);
   if ($271) {
    $272 = HEAP8[$55>>0]|0;
    $273 = ($272<<24>>24)==(0);
    if ($273) {
     (_fwrite(512243,26,1,$56)|0);
    }
   }
   $$not$i14 = ($269|0)<(0);
   L91: do {
    if ($$not$i14) {
     $try$0$us$i = 0;
     while(1) {
      $274 = (($try$0$us$i) + 1)|0;
      $275 = HEAP32[$49>>2]|0;
      $276 = ((($275)) + 16|0);
      $277 = HEAP32[$276>>2]|0;
      $278 = (FUNCTION_TABLE_iii[$277 & 8191]($fr,$newhead$i11)|0);
      $279 = ($278|0)<(1);
      $280 = HEAP32[$newhead$i11>>2]|0;
      if ($279) {
       $$lcssa6$i = $278;$$lcssa7$i = $280;
       label = 87;
       break L14;
      }
      $281 = ($280>>>0)<(4292870144);
      $282 = $280 & 393216;
      $283 = ($282|0)==(0);
      $or$cond$i$us$i16 = $281 | $283;
      $284 = $280 & 61440;
      $285 = ($284|0)==(61440);
      $or$cond3$i$us$i17 = $285 | $or$cond$i$us$i16;
      $286 = $280 & 3072;
      $287 = ($286|0)==(3072);
      $or$cond5$i$us$i18 = $287 | $or$cond3$i$us$i17;
      if ($or$cond5$i$us$i18) {
       $try$0$us$i = $274;
      } else {
       $$lcssa$i = $274;$307 = $280;$or$cond$lcssa$i = 1;
       break;
      }
     }
    } else {
     $891 = $224;$try$0$i = 0;
     while(1) {
      $288 = (($try$0$i) + 1)|0;
      $289 = ($288|0)<($269|0);
      if (!($289)) {
       $$lcssa$i = $288;$307 = $891;$or$cond$lcssa$i = 0;
       break L91;
      }
      $290 = HEAP32[$49>>2]|0;
      $291 = ((($290)) + 16|0);
      $292 = HEAP32[$291>>2]|0;
      $293 = (FUNCTION_TABLE_iii[$292 & 8191]($fr,$newhead$i11)|0);
      $294 = ($293|0)<(1);
      $295 = HEAP32[$newhead$i11>>2]|0;
      if ($294) {
       $$lcssa6$i = $293;$$lcssa7$i = $295;
       label = 87;
       break L14;
      }
      $300 = ($295>>>0)<(4292870144);
      $301 = $295 & 393216;
      $302 = ($301|0)==(0);
      $or$cond$i$i19 = $300 | $302;
      $303 = $295 & 61440;
      $304 = ($303|0)==(61440);
      $or$cond3$i$i20 = $304 | $or$cond$i$i19;
      $305 = $295 & 3072;
      $306 = ($305|0)==(3072);
      $or$cond5$i$i21 = $306 | $or$cond3$i$i20;
      if ($or$cond5$i$i21) {
       $891 = $295;$try$0$i = $288;
      } else {
       $$lcssa$i = $288;$307 = $295;$or$cond$lcssa$i = 1;
       break;
      }
     }
    }
   } while(0);
   HEAP32[$newhead>>2] = $307;
   $308 = HEAP32[$53>>2]|0;
   $309 = $308 & 32;
   $310 = ($309|0)==(0);
   if ($310) {
    $311 = HEAP8[$55>>0]|0;
    $312 = ($311<<24>>24)==(0);
    if ($312) {
     HEAP32[$vararg_buffer9>>2] = $$lcssa$i;
     (_fprintf($56,512320,$vararg_buffer9)|0);
    }
   }
   if (!($or$cond$lcssa$i)) {
    $$lcssa$i$lcssa = $$lcssa$i;
    label = 96;
    break L14;
   }
   HEAP32[$52>>2] = 0;
   $headcount$1 = $headcount$7;
  }
  if ((label|0) == 21) {
   label = 0;
   HEAP32[$51>>2] = 0;
   $120 = (_INT123_parse_new_id3($fr,$$lcssa547)|0);
   $121 = ($120|0)<(0);
   if ($121) {
    $$0$i6$ph32 = $120;
    label = 56;
    break;
   }
   $122 = ($120|0)>(0);
   if ($122) {
    $123 = HEAP32[$59>>2]|0;
    $124 = $123 | 3;
    HEAP32[$59>>2] = $124;
    $headcount$1546 = $headcount$1$lcssa;
    label = 58;
   } else {
    $headcount$1546 = $headcount$1$lcssa;
    label = 58;
   }
  }
  else if ((label|0) == 37) {
   label = 0;
   $156 = HEAP32[$53>>2]|0;
   $157 = $156 & 32;
   $158 = ($157|0)==(0);
   if ($158) {
    $159 = HEAP32[$54>>2]|0;
    $160 = ($159|0)>(1);
    if ($160) {
     $161 = HEAP8[$55>>0]|0;
     $162 = ($161<<24>>24)==(0);
     if ($162) {
      (_fwrite(511055,27,1,$56)|0);
     }
    }
   }
   HEAP32[$51>>2] = 0;
   $163 = HEAP32[$newhead$i>>2]|0;
   HEAP32[$newhead>>2] = $163;
   $headcount$1546 = $headcount$1$lcssa536;
   label = 58;
  }
  else if ((label|0) == 66) {
   label = 0;
   $226 = $$lcssa559 >>> 24;
   $227 = $226&255;
   HEAP8[$60>>0] = $227;
   $228 = $$lcssa559 >>> 16;
   $229 = $228&255;
   HEAP8[$61>>0] = $229;
   $230 = $$lcssa559 >>> 8;
   $231 = $230&255;
   HEAP8[$62>>0] = $231;
   $232 = $$lcssa559&255;
   HEAP8[$63>>0] = $232;
   $233 = HEAP32[$49>>2]|0;
   $234 = ((($233)) + 8|0);
   $235 = HEAP32[$234>>2]|0;
   $236 = (FUNCTION_TABLE_iiii[$235 & 8191]($fr,$64,124)|0);
   $237 = ($236|0)<(0);
   if ($237) {
    $$1$i23 = $236;
    label = 102;
    break;
   }
   $238 = HEAP32[$59>>2]|0;
   $239 = $238 | 3;
   HEAP32[$59>>2] = $239;
   $240 = HEAP32[$65>>2]|0;
   $241 = $240 | 2;
   HEAP32[$65>>2] = $241;
   $242 = HEAP32[$53>>2]|0;
   $243 = $242 & 32;
   $244 = ($243|0)==(0);
   if ($244) {
    $245 = HEAP32[$66>>2]|0;
    $246 = ($245|0)>(2);
    if ($246) {
     (_fwrite(512128,25,1,$56)|0);
     $headcount$7558 = $headcount$7$lcssa552;
     label = 103;
    } else {
     $headcount$7558 = $headcount$7$lcssa552;
     label = 103;
    }
   } else {
    $headcount$7558 = $headcount$7$lcssa552;
    label = 103;
   }
  }
  else if ((label|0) == 70) {
   label = 0;
   HEAP32[$51>>2] = 0;
   $247 = (_INT123_parse_new_id3($fr,$$lcssa560)|0);
   $248 = ($247|0)<(0);
   if ($248) {
    $$1$i23 = $247;
    label = 102;
    break;
   }
   $249 = ($247|0)>(0);
   if ($249) {
    $250 = HEAP32[$59>>2]|0;
    $251 = $250 | 3;
    HEAP32[$59>>2] = $251;
    $headcount$7558 = $headcount$7$lcssa553;
    label = 103;
   } else {
    $headcount$7558 = $headcount$7$lcssa553;
    label = 103;
   }
  }
  else if ((label|0) == 105) {
   label = 0;
   $320 = HEAP32[$52>>2]|0;
   $321 = ($320|0)==(0);
   if ($321) {
    $322 = HEAP32[$newhead>>2]|0;
    HEAP32[$nexthead$i>>2] = 0;
    $323 = HEAP32[$65>>2]|0;
    $324 = $323 & 12;
    $325 = ($324|0)==(0);
    do {
     if (!($325)) {
      $326 = HEAP32[$49>>2]|0;
      $327 = ((($326)) + 36|0);
      $328 = HEAP32[$327>>2]|0;
      $329 = (FUNCTION_TABLE_ii[$328 & 8191]($fr)|0);
      $330 = HEAP32[$49>>2]|0;
      $331 = ((($330)) + 20|0);
      $332 = HEAP32[$331>>2]|0;
      $333 = HEAP32[$0>>2]|0;
      $334 = (FUNCTION_TABLE_iii[$332 & 8191]($fr,$333)|0);
      $335 = ($334|0)<(0);
      if ($335) {
       $$lcssa569 = $334;
       label = 108;
       break L14;
      }
      $342 = HEAP32[$49>>2]|0;
      $343 = ((($342)) + 12|0);
      $344 = HEAP32[$343>>2]|0;
      $345 = (FUNCTION_TABLE_iii[$344 & 8191]($fr,$nexthead$i)|0);
      $346 = HEAP32[$49>>2]|0;
      $347 = ((($346)) + 28|0);
      $348 = HEAP32[$347>>2]|0;
      $349 = ((($346)) + 36|0);
      $350 = HEAP32[$349>>2]|0;
      $351 = (FUNCTION_TABLE_ii[$350 & 8191]($fr)|0);
      $352 = (($351) - ($329))|0;
      $353 = (FUNCTION_TABLE_iii[$348 & 8191]($fr,$352)|0);
      $354 = ($353|0)<(0);
      if ($354) {
       label = 113;
       break L14;
      }
      switch ($345|0) {
      case -10:  {
       $$0$i27$ph = -10;
       label = 123;
       break L14;
       break;
      }
      case 0:  {
       label = 116;
       break L14;
       break;
      }
      default: {
      }
      }
      $361 = HEAP32[$nexthead$i>>2]|0;
      $362 = ($361>>>0)<(4292870144);
      $363 = $361 & 393216;
      $364 = ($363|0)==(0);
      $or$cond$i$i24 = $362 | $364;
      $365 = $361 & 61440;
      $366 = ($365|0)==(61440);
      $or$cond3$i$i25 = $366 | $or$cond$i$i24;
      $367 = $361 & 3072;
      $368 = ($367|0)==(3072);
      $or$cond5$i$i26 = $368 | $or$cond3$i$i25;
      if (!($or$cond5$i$i26)) {
       $369 = $361 ^ $322;
       $370 = $369 & -128000;
       $371 = ($370|0)==(0);
       if ($371) {
        break;
       }
      }
      HEAP32[$51>>2] = 0;
      $372 = HEAP32[$49>>2]|0;
      $373 = ((($372)) + 28|0);
      $374 = HEAP32[$373>>2]|0;
      $375 = (FUNCTION_TABLE_iii[$374 & 8191]($fr,3)|0);
      $376 = ($375|0)<(0);
      if ($376) {
       label = 121;
       break L14;
      }
      $headcount$0 = $headcount$7$lcssa;
      continue L14;
     }
    } while(0);
   }
   $383 = HEAP32[$49>>2]|0;
   $384 = ((($383)) + 36|0);
   $385 = HEAP32[$384>>2]|0;
   $386 = (FUNCTION_TABLE_ii[$385 & 8191]($fr)|0);
   $387 = (($386) + -4)|0;
   $388 = HEAP32[$67>>2]|0;
   $389 = (((((($fr)) + 19872|0) + (($388*3968)|0)|0)) + 512|0);
   $390 = HEAP32[$49>>2]|0;
   $391 = ((($390)) + 24|0);
   $392 = HEAP32[$391>>2]|0;
   $393 = HEAP32[$0>>2]|0;
   $394 = (FUNCTION_TABLE_iiii[$392 & 8191]($fr,$389,$393)|0);
   $395 = ($394|0)<(0);
   if ($395) {
    $ret$1 = $394;
    break;
   }
   $396 = HEAP32[$68>>2]|0;
   HEAP32[$69>>2] = $396;
   HEAP32[$68>>2] = $389;
   $397 = HEAP32[$67>>2]|0;
   $398 = $397 & 1;
   $399 = $398 ^ 1;
   HEAP32[$67>>2] = $399;
   $400 = HEAP32[$52>>2]|0;
   $401 = ($400|0)==(0);
   if (!($401)) {
    $$lcssa574 = $387;$$lcssa576 = $389;
    label = 237;
    break;
   }
   $402 = HEAP32[$newhead>>2]|0;
   HEAP32[$52>>2] = $402;
   $403 = HEAP32[$70>>2]|0;
   $404 = ($403|0)<(0);
   if (!($404)) {
    $$lcssa574 = $387;$$lcssa576 = $389;
    label = 237;
    break;
   }
   HEAP32[$71>>2] = $387;
   $405 = HEAP32[$72>>2]|0;
   $406 = ($405|0)==(3);
   if (!($406)) {
    $$lcssa575 = $387;
    label = 236;
    break;
   }
   $407 = HEAP32[$73>>2]|0;
   $408 = ($407|0)==(2);
   $409 = HEAP32[$74>>2]|0;
   $410 = ($409|0)!=(0);
   $411 = $410 ? 17 : 32;
   $412 = $410 ? 9 : 17;
   $413 = $408 ? $411 : $412;
   $414 = HEAP32[$53>>2]|0;
   $415 = $414 & 16384;
   $416 = ($415|0)==(0);
   if (!($416)) {
    $$lcssa575 = $387;
    label = 236;
    break;
   }
   $417 = HEAP32[$0>>2]|0;
   $418 = (($413) + 120)|0;
   $419 = ($417|0)<($418|0);
   if ($419) {
    $$lcssa575 = $387;
    label = 236;
    break;
   } else {
    $i$021$i = 2;
   }
   while(1) {
    $420 = (($389) + ($i$021$i)|0);
    $421 = HEAP8[$420>>0]|0;
    $422 = ($421<<24>>24)==(0);
    if (!($422)) {
     $i$0$lcssa$i = $i$021$i;
     break;
    }
    $423 = (($i$021$i) + 1)|0;
    $424 = ($423|0)<($413|0);
    if ($424) {
     $i$021$i = $423;
    } else {
     $i$0$lcssa$i = $423;
     break;
    }
   }
   $425 = ($i$0$lcssa$i|0)==($413|0);
   if (!($425)) {
    $$lcssa575 = $387;
    label = 236;
    break;
   }
   $426 = (($389) + ($413)|0);
   $427 = HEAP8[$426>>0]|0;
   switch ($427<<24>>24) {
   case 73:  {
    $428 = (($413) + 1)|0;
    $429 = (($389) + ($428)|0);
    $430 = HEAP8[$429>>0]|0;
    $431 = ($430<<24>>24)==(110);
    if (!($431)) {
     $$lcssa575 = $387;
     label = 236;
     break L14;
    }
    $432 = $413 | 2;
    $433 = (($389) + ($432)|0);
    $434 = HEAP8[$433>>0]|0;
    $435 = ($434<<24>>24)==(102);
    if (!($435)) {
     $$lcssa575 = $387;
     label = 236;
     break L14;
    }
    $436 = (($413) + 3)|0;
    $437 = (($389) + ($436)|0);
    $438 = HEAP8[$437>>0]|0;
    $439 = ($438<<24>>24)==(111);
    if (!($439)) {
     $$lcssa575 = $387;
     label = 236;
     break L14;
    }
    break;
   }
   case 88:  {
    $440 = (($413) + 1)|0;
    $441 = (($389) + ($440)|0);
    $442 = HEAP8[$441>>0]|0;
    $443 = ($442<<24>>24)==(105);
    if (!($443)) {
     $$lcssa575 = $387;
     label = 236;
     break L14;
    }
    $444 = $413 | 2;
    $445 = (($389) + ($444)|0);
    $446 = HEAP8[$445>>0]|0;
    $447 = ($446<<24>>24)==(110);
    if (!($447)) {
     $$lcssa575 = $387;
     label = 236;
     break L14;
    }
    $448 = (($413) + 3)|0;
    $449 = (($389) + ($448)|0);
    $450 = HEAP8[$449>>0]|0;
    $451 = ($450<<24>>24)==(103);
    if (!($451)) {
     $$lcssa575 = $387;
     label = 236;
     break L14;
    }
    HEAP32[$77>>2] = 1;
    break;
   }
   default: {
    $$lcssa575 = $387;
    label = 236;
    break L14;
   }
   }
   $452 = $414 & 32;
   $453 = ($452|0)==(0);
   if ($453) {
    $454 = HEAP32[$54>>2]|0;
    $455 = ($454|0)>(1);
    if ($455) {
     (_fwrite(512726,37,1,$56)|0);
     $$pre$i = HEAP32[$68>>2]|0;
     $459 = $$pre$i;
    } else {
     $459 = $389;
    }
   } else {
    $459 = $389;
   }
   $456 = $413 | 4;
   $457 = (($456) + 3)|0;
   $458 = (($459) + ($457)|0);
   $460 = HEAP8[$458>>0]|0;
   $461 = $460&255;
   $462 = (($456) + 4)|0;
   $463 = $461 & 1;
   $464 = ($463|0)==(0);
   if ($464) {
    $lame_offset$0$i = $462;
   } else {
    $465 = HEAP32[$53>>2]|0;
    $466 = $465 & 4096;
    $467 = ($466|0)==(0);
    do {
     if ($467) {
      $472 = (($459) + ($462)|0);
      $473 = HEAP8[$472>>0]|0;
      $474 = $473&255;
      $475 = $474 << 24;
      $476 = (($456) + 5)|0;
      $477 = (($459) + ($476)|0);
      $478 = HEAP8[$477>>0]|0;
      $479 = $478&255;
      $480 = $479 << 16;
      $481 = $480 | $475;
      $482 = (($456) + 6)|0;
      $483 = (($459) + ($482)|0);
      $484 = HEAP8[$483>>0]|0;
      $485 = $484&255;
      $486 = $485 << 8;
      $487 = $481 | $486;
      $488 = (($456) + 7)|0;
      $489 = (($459) + ($488)|0);
      $490 = HEAP8[$489>>0]|0;
      $491 = $490&255;
      $492 = $487 | $491;
      $493 = ($492>>>0)>(932067);
      $$$i = $493 ? 0 : $492;
      HEAP32[$86>>2] = $$$i;
      $494 = $465 & 64;
      $495 = ($494|0)==(0);
      if ($495) {
       $497 = $465;
      } else {
       _INT123_frame_gapless_init($fr,$$$i,0,0);
       $$pre24$i = HEAP32[$53>>2]|0;
       $497 = $$pre24$i;
      }
      $496 = $497 & 32;
      $498 = ($496|0)==(0);
      if ($498) {
       $499 = HEAP32[$54>>2]|0;
       $500 = ($499|0)>(2);
       if (!($500)) {
        break;
       }
       $501 = HEAP32[$86>>2]|0;
       HEAP32[$vararg_buffer31>>2] = $501;
       (_fprintf($56,512830,$vararg_buffer31)|0);
      }
     } else {
      $468 = $465 & 32;
      $469 = ($468|0)==(0);
      if ($469) {
       $470 = HEAP32[$54>>2]|0;
       $471 = ($470|0)>(2);
       if ($471) {
        (_fwrite(512764,65,1,$56)|0);
       }
      }
     }
    } while(0);
    $502 = (($456) + 8)|0;
    $lame_offset$0$i = $502;
   }
   $503 = $461 & 2;
   $504 = ($503|0)==(0);
   if ($504) {
    $lame_offset$1$i = $lame_offset$0$i;
   } else {
    $505 = HEAP32[$53>>2]|0;
    $506 = $505 & 4096;
    $507 = ($506|0)==(0);
    do {
     if ($507) {
      $512 = HEAP32[$68>>2]|0;
      $513 = (($512) + ($lame_offset$0$i)|0);
      $514 = HEAP8[$513>>0]|0;
      $515 = $514&255;
      $516 = $515 << 24;
      $517 = (($lame_offset$0$i) + 1)|0;
      $518 = (($512) + ($517)|0);
      $519 = HEAP8[$518>>0]|0;
      $520 = $519&255;
      $521 = $520 << 16;
      $522 = $521 | $516;
      $523 = (($lame_offset$0$i) + 2)|0;
      $524 = (($512) + ($523)|0);
      $525 = HEAP8[$524>>0]|0;
      $526 = $525&255;
      $527 = $526 << 8;
      $528 = $522 | $527;
      $529 = (($lame_offset$0$i) + 3)|0;
      $530 = (($512) + ($529)|0);
      $531 = HEAP8[$530>>0]|0;
      $532 = $531&255;
      $533 = $528 | $532;
      $534 = HEAP32[$87>>2]|0;
      $535 = ($534|0)<(1);
      do {
       if ($535) {
        HEAP32[$87>>2] = $533;
       } else {
        $536 = ($534|0)!=($533|0);
        $537 = $505 & 32;
        $538 = ($537|0)==(0);
        $or$cond13$i = $538 & $536;
        if (!($or$cond13$i)) {
         break;
        }
        $539 = (+($534|0));
        $540 = 1.0 / $539;
        $541 = (($534) - ($533))|0;
        $542 = (+($541|0));
        $543 = $542 * $540;
        $544 = $543 < 0.0;
        $545 = -$543;
        $diff$0$i = $544 ? $545 : $543;
        $546 = HEAP32[$54>>2]|0;
        $547 = ($546|0)>(2);
        if ($547) {
         HEAP32[$vararg_buffer34>>2] = $533;
         $vararg_ptr37 = ((($vararg_buffer34)) + 8|0);
         HEAPF64[$vararg_ptr37>>3] = $diff$0$i;
         (_fprintf($56,512919,$vararg_buffer34)|0);
        }
        $548 = $diff$0$i > 1.0;
        if (!($548)) {
         break;
        }
        (_fprintf($56,512996,$vararg_buffer38)|0);
       }
      } while(0);
      $549 = HEAP32[$53>>2]|0;
      $550 = $549 & 32;
      $551 = ($550|0)==(0);
      if (!($551)) {
       break;
      }
      $552 = HEAP32[$54>>2]|0;
      $553 = ($552|0)>(2);
      if (!($553)) {
       break;
      }
      HEAP32[$vararg_buffer40>>2] = $533;
      (_fprintf($56,513098,$vararg_buffer40)|0);
     } else {
      $508 = $505 & 32;
      $509 = ($508|0)==(0);
      if ($509) {
       $510 = HEAP32[$54>>2]|0;
       $511 = ($510|0)>(2);
       if (!($511)) {
        break;
       }
       (_fwrite(512854,64,1,$56)|0);
      }
     }
    } while(0);
    $554 = (($lame_offset$0$i) + 4)|0;
    $lame_offset$1$i = $554;
   }
   $555 = $461 & 4;
   $556 = ($555|0)==(0);
   if ($556) {
    $lame_offset$2$i = $lame_offset$1$i;
   } else {
    $557 = HEAP32[$68>>2]|0;
    $558 = (($557) + ($lame_offset$1$i)|0);
    (_INT123_frame_fill_toc($fr,$558)|0);
    $559 = (($lame_offset$1$i) + 100)|0;
    $lame_offset$2$i = $559;
   }
   $560 = $461 & 8;
   $561 = ($560|0)==(0);
   if ($561) {
    $lame_offset$3$i = $lame_offset$2$i;
   } else {
    $562 = HEAP32[$53>>2]|0;
    $563 = $562 & 32;
    $564 = ($563|0)==(0);
    do {
     if ($564) {
      $565 = HEAP32[$54>>2]|0;
      $566 = ($565|0)>(2);
      if (!($566)) {
       break;
      }
      $567 = HEAP32[$68>>2]|0;
      $568 = (($567) + ($lame_offset$2$i)|0);
      $569 = HEAP8[$568>>0]|0;
      $570 = $569&255;
      $571 = $570 << 24;
      $572 = (($lame_offset$2$i) + 1)|0;
      $573 = (($567) + ($572)|0);
      $574 = HEAP8[$573>>0]|0;
      $575 = $574&255;
      $576 = $575 << 16;
      $577 = $576 | $571;
      $578 = (($lame_offset$2$i) + 2)|0;
      $579 = (($567) + ($578)|0);
      $580 = HEAP8[$579>>0]|0;
      $581 = $580&255;
      $582 = $581 << 8;
      $583 = $577 | $582;
      $584 = (($lame_offset$2$i) + 3)|0;
      $585 = (($567) + ($584)|0);
      $586 = HEAP8[$585>>0]|0;
      $587 = $586&255;
      $588 = $583 | $587;
      HEAP32[$vararg_buffer43>>2] = $588;
      (_fprintf($56,513121,$vararg_buffer43)|0);
     }
    } while(0);
    $589 = (($lame_offset$2$i) + 4)|0;
    $lame_offset$3$i = $589;
   }
   $590 = HEAP32[$68>>2]|0;
   $591 = (($590) + ($lame_offset$3$i)|0);
   $592 = HEAP8[$591>>0]|0;
   $593 = ($592<<24>>24)==(0);
   if (!($593)) {
    $594 = $replay_gain$i;
    $595 = $594;
    HEAP32[$595>>2] = 0;
    $596 = (($594) + 4)|0;
    $597 = $596;
    HEAP32[$597>>2] = 0;
    dest=$nb$i; src=$591; stop=dest+9|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
    HEAP8[$75>>0] = 0;
    $598 = HEAP32[$53>>2]|0;
    $599 = $598 & 32;
    $600 = ($599|0)==(0);
    do {
     if ($600) {
      $601 = HEAP32[$54>>2]|0;
      $602 = ($601|0)>(2);
      if (!($602)) {
       break;
      }
      HEAP32[$vararg_buffer46>>2] = $nb$i;
      (_fprintf($56,513148,$vararg_buffer46)|0);
     }
    } while(0);
    $603 = (_strncmp(513173,$nb$i,4)|0);
    $604 = ($603|0)==(0);
    if ($604) {
     HEAP8[$rest$i>>0] = 0;
     HEAP32[$vararg_buffer49>>2] = $major$i;
     $vararg_ptr52 = ((($vararg_buffer49)) + 4|0);
     HEAP32[$vararg_ptr52>>2] = $minor$i;
     $vararg_ptr53 = ((($vararg_buffer49)) + 8|0);
     HEAP32[$vararg_ptr53>>2] = $rest$i;
     $605 = (_sscanf($76,513178,$vararg_buffer49)|0);
     $606 = ($605|0)>(1);
     do {
      if ($606) {
       $607 = HEAP32[$major$i>>2]|0;
       $608 = ($607>>>0)<(3);
       if (!($608)) {
        $609 = ($607|0)==(3);
        $610 = HEAP32[$minor$i>>2]|0;
        $611 = ($610>>>0)<(95);
        $or$cond$i4 = $609 & $611;
        if (!($or$cond$i4)) {
         $gain_offset$0$i = 0.0;
         break;
        }
       }
       $612 = HEAP32[$53>>2]|0;
       $613 = $612 & 32;
       $614 = ($613|0)==(0);
       if (!($614)) {
        $gain_offset$0$i = 6.0;
        break;
       }
       $615 = HEAP32[$54>>2]|0;
       $616 = ($615|0)>(2);
       if (!($616)) {
        $gain_offset$0$i = 6.0;
        break;
       }
       HEAPF64[$vararg_buffer54>>3] = 6.0;
       (_fprintf($56,513186,$vararg_buffer54)|0);
       $gain_offset$0$i = 6.0;
      } else {
       $617 = HEAP32[$53>>2]|0;
       $618 = $617 & 32;
       $619 = ($618|0)==(0);
       if (!($619)) {
        $gain_offset$0$i = 0.0;
        break;
       }
       $620 = HEAP32[$54>>2]|0;
       $621 = ($620|0)>(2);
       if (!($621)) {
        $gain_offset$0$i = 0.0;
        break;
       }
       (_fwrite(513252,43,1,$56)|0);
       $gain_offset$0$i = 0.0;
      }
     } while(0);
     $gain_offset$1$i = $gain_offset$0$i;
    } else {
     $gain_offset$1$i = 0.0;
    }
    $622 = (($lame_offset$3$i) + 9)|0;
    $623 = HEAP32[$68>>2]|0;
    $624 = (($623) + ($622)|0);
    $625 = HEAP8[$624>>0]|0;
    $626 = $625&255;
    $627 = $626 & 15;
    $628 = HEAP32[$53>>2]|0;
    $629 = $628 & 32;
    $630 = ($629|0)==(0);
    do {
     if ($630) {
      $631 = HEAP32[$54>>2]|0;
      $632 = ($631|0)>(2);
      if (!($632)) {
       break;
      }
      $633 = $626 >>> 4;
      HEAP32[$vararg_buffer57>>2] = $633;
      (_fprintf($56,513296,$vararg_buffer57)|0);
      HEAP32[$vararg_buffer60>>2] = $627;
      (_fprintf($56,513316,$vararg_buffer60)|0);
     }
    } while(0);
    switch ($627|0) {
    case 8: case 1:  {
     HEAP32[$77>>2] = 0;
     break;
    }
    case 9: case 2:  {
     HEAP32[$77>>2] = 2;
     break;
    }
    default: {
     HEAP32[$77>>2] = 1;
    }
    }
    $634 = HEAP32[$53>>2]|0;
    $635 = $634 & 32;
    $636 = ($635|0)==(0);
    do {
     if ($636) {
      $637 = HEAP32[$54>>2]|0;
      $638 = ($637|0)>(2);
      if (!($638)) {
       break;
      }
      HEAPF64[$vararg_buffer63>>3] = 0.0;
      (_fprintf($56,513341,$vararg_buffer63)|0);
     }
    } while(0);
    $639 = (($lame_offset$3$i) + 15)|0;
    $640 = HEAP32[$68>>2]|0;
    $641 = (($640) + ($639)|0);
    $642 = HEAP8[$641>>0]|0;
    $643 = $642&255;
    $644 = $643 >>> 2;
    $645 = $644 & 7;
    $646 = ($645|0)==(0);
    L220: do {
     if ($646) {
      label = 210;
     } else {
      $647 = ($642&255) >>> 5;
      switch ($647<<24>>24) {
      case 1:  {
       $gt$0$i = 0;
       break;
      }
      case 2:  {
       $gt$0$i = 1;
       break;
      }
      default: {
       $664 = $642;$lame_offset$5$i = $639;
       break L220;
      }
      }
      $648 = $642 & 2;
      $649 = ($648<<24>>24)!=(0);
      $650 = $649 ? -0.10000000149011612 : 0.10000000149011612;
      $651 = $643 << 8;
      $652 = (($lame_offset$3$i) + 16)|0;
      $653 = (($640) + ($652)|0);
      $654 = HEAP8[$653>>0]|0;
      $655 = $654&255;
      $$masked6$i = $651 & 256;
      $656 = $655 | $$masked6$i;
      $657 = (+($656|0));
      $658 = $657 * $650;
      $659 = (($replay_gain$i) + ($gt$0$i<<2)|0);
      $660 = ($645|0)==(3);
      $661 = $658 + $gain_offset$1$i;
      $$14$i = $660 ? $661 : $658;
      HEAPF32[$659>>2] = $$14$i;
      label = 210;
     }
    } while(0);
    if ((label|0) == 210) {
     label = 0;
     $662 = (($lame_offset$3$i) + 17)|0;
     $$phi$trans$insert$i = (($640) + ($662)|0);
     $$pre25$i = HEAP8[$$phi$trans$insert$i>>0]|0;
     $664 = $$pre25$i;$lame_offset$5$i = $662;
    }
    $663 = $664&255;
    $665 = $663 >>> 2;
    $666 = $665 & 7;
    $667 = ($666|0)==(0);
    L228: do {
     if ($667) {
      label = 231;
     } else {
      $736 = ($664&255) >>> 5;
      switch ($736<<24>>24) {
      case 1:  {
       $gt$0$1$i = 0;
       break;
      }
      case 2:  {
       $gt$0$1$i = 1;
       break;
      }
      default: {
       $lame_offset$5$1$i = $lame_offset$5$i;
       break L228;
      }
      }
      $737 = $664 & 2;
      $738 = ($737<<24>>24)!=(0);
      $739 = $738 ? -0.10000000149011612 : 0.10000000149011612;
      $740 = $663 << 8;
      $741 = (($lame_offset$5$i) + 1)|0;
      $742 = (($640) + ($741)|0);
      $743 = HEAP8[$742>>0]|0;
      $744 = $743&255;
      $$masked6$1$i = $740 & 256;
      $745 = $744 | $$masked6$1$i;
      $746 = (+($745|0));
      $747 = $746 * $739;
      $748 = (($replay_gain$i) + ($gt$0$1$i<<2)|0);
      $749 = ($666|0)==(3);
      $750 = $747 + $gain_offset$1$i;
      $$14$1$i = $749 ? $750 : $747;
      HEAPF32[$748>>2] = $$14$1$i;
      label = 231;
     }
    } while(0);
    if ((label|0) == 231) {
     label = 0;
     $751 = (($lame_offset$5$i) + 2)|0;
     $lame_offset$5$1$i = $751;
    }
    $752 = HEAP32[$53>>2]|0;
    $753 = $752 & 32;
    $754 = ($753|0)==(0);
    do {
     if ($754) {
      $668 = HEAP32[$54>>2]|0;
      $669 = ($668|0)>(2);
      if (!($669)) {
       break;
      }
      $670 = +HEAPF32[$replay_gain$i>>2];
      $671 = $670;
      HEAPF64[$vararg_buffer66>>3] = $671;
      (_fprintf($56,513383,$vararg_buffer66)|0);
      $672 = +HEAPF32[$78>>2];
      $673 = $672;
      HEAPF64[$vararg_buffer69>>3] = $673;
      (_fprintf($56,513418,$vararg_buffer69)|0);
     }
    } while(0);
    $674 = HEAP32[$79>>2]|0;
    $675 = ($674|0)<(1);
    if ($675) {
     HEAPF32[$80>>2] = 0.0;
     $676 = HEAP32[$replay_gain$i>>2]|0;
     HEAP32[$81>>2] = $676;
     HEAP32[$79>>2] = 0;
    }
    $677 = HEAP32[$82>>2]|0;
    $678 = ($677|0)<(1);
    if ($678) {
     HEAPF32[$83>>2] = 0.0;
     $733 = HEAP32[$78>>2]|0;
     HEAP32[$84>>2] = $733;
     HEAP32[$82>>2] = 0;
    }
    $734 = HEAP32[$77>>2]|0;
    $735 = ($734|0)==(2);
    do {
     if ($735) {
      $679 = (($lame_offset$5$1$i) + 1)|0;
      $680 = HEAP32[$68>>2]|0;
      $681 = (($680) + ($679)|0);
      $682 = HEAP8[$681>>0]|0;
      $683 = $682&255;
      HEAP32[$85>>2] = $683;
      $684 = HEAP32[$53>>2]|0;
      $685 = $684 & 32;
      $686 = ($685|0)==(0);
      if (!($686)) {
       break;
      }
      $687 = HEAP32[$54>>2]|0;
      $688 = ($687|0)>(2);
      if (!($688)) {
       break;
      }
      HEAP32[$vararg_buffer72>>2] = $683;
      (_fprintf($56,513458,$vararg_buffer72)|0);
     }
    } while(0);
    $689 = (($lame_offset$5$1$i) + 2)|0;
    $690 = HEAP32[$53>>2]|0;
    $691 = $690 & 32;
    $692 = ($691|0)==(0);
    do {
     if ($692) {
      $693 = HEAP32[$54>>2]|0;
      $694 = ($693|0)>(2);
      if (!($694)) {
       $713 = $690;
       break;
      }
      $695 = HEAP32[$68>>2]|0;
      $696 = (($695) + ($689)|0);
      $697 = HEAP8[$696>>0]|0;
      $698 = $697&255;
      $699 = $698 << 4;
      $700 = (($lame_offset$5$1$i) + 3)|0;
      $701 = (($695) + ($700)|0);
      $702 = HEAP8[$701>>0]|0;
      $703 = $702&255;
      $704 = $703 >>> 4;
      $705 = $704 | $699;
      $706 = $703 << 8;
      $707 = (($lame_offset$5$1$i) + 4)|0;
      $708 = (($695) + ($707)|0);
      $709 = HEAP8[$708>>0]|0;
      $710 = $709&255;
      $$masked5$i = $706 & 3840;
      $711 = $$masked5$i | $710;
      HEAP32[$vararg_buffer75>>2] = $705;
      $vararg_ptr78 = ((($vararg_buffer75)) + 4|0);
      HEAP32[$vararg_ptr78>>2] = $711;
      (_fprintf($56,513485,$vararg_buffer75)|0);
      $$pre26$i = HEAP32[$53>>2]|0;
      $713 = $$pre26$i;
     } else {
      $713 = $690;
     }
    } while(0);
    $712 = $713 & 64;
    $714 = ($712|0)==(0);
    if (!($714)) {
     $715 = HEAP32[$68>>2]|0;
     $716 = (($715) + ($689)|0);
     $717 = HEAP8[$716>>0]|0;
     $718 = $717&255;
     $719 = $718 << 4;
     $720 = (($lame_offset$5$1$i) + 3)|0;
     $721 = (($715) + ($720)|0);
     $722 = HEAP8[$721>>0]|0;
     $723 = $722&255;
     $724 = $723 >>> 4;
     $725 = $724 | $719;
     $726 = $723 << 8;
     $727 = (($lame_offset$5$1$i) + 4)|0;
     $728 = (($715) + ($727)|0);
     $729 = HEAP8[$728>>0]|0;
     $730 = $729&255;
     $$masked$i = $726 & 3840;
     $731 = $$masked$i | $730;
     $732 = HEAP32[$86>>2]|0;
     _INT123_frame_gapless_init($fr,$732,$725,$731);
    }
   }
   $755 = HEAP32[$67>>2]|0;
   $756 = (((((($fr)) + 19872|0) + (($755*3968)|0)|0)) + 512|0);
   HEAP32[$68>>2] = $756;
   $757 = $755 & 1;
   $758 = $757 ^ 1;
   HEAP32[$67>>2] = $758;
   $759 = HEAP32[$49>>2]|0;
   $760 = ((($759)) + 44|0);
   $761 = HEAP32[$760>>2]|0;
   $762 = ($761|0)==(0|0);
   if (!($762)) {
    FUNCTION_TABLE_vi[$761 & 8191]($fr);
   }
   HEAP32[$51>>2] = 0;
   $headcount$0 = $headcount$7$lcssa;
   continue;
  }
  if ((label|0) == 58) {
   label = 0;
   $headcount$0 = $headcount$1546;
   continue;
  }
  else if ((label|0) == 103) {
   label = 0;
   $headcount$0 = $headcount$7558;
   continue;
  }
 }
 switch (label|0) {
  case 54: {
   $209 = HEAP32[$53>>2]|0;
   $210 = $209 & 32;
   $211 = ($210|0)==(0);
   if ($211) {
    HEAP32[$vararg_buffer1>>2] = 1078;
    $vararg_ptr4 = ((($vararg_buffer1)) + 4|0);
    HEAP32[$vararg_ptr4>>2] = $headcount$5$lcssa;
    (_fprintf($56,512044,$vararg_buffer1)|0);
   }
   $ret$1 = 10;
   break;
  }
  case 56: {
   $ret$1 = $$0$i6$ph32;
   break;
  }
  case 87: {
   HEAP32[$newhead>>2] = $$lcssa7$i;
   $296 = HEAP32[$53>>2]|0;
   $297 = $296 & 32;
   $298 = ($297|0)==(0);
   if ($298) {
    (_fwrite(512270,49,1,$56)|0);
   }
   $299 = ($$lcssa6$i|0)==(0);
   if ($299) {
    $ret$1 = 10;
   } else {
    $$1$i23 = $$lcssa6$i;
    label = 102;
   }
   break;
  }
  case 96: {
   $313 = HEAP32[$53>>2]|0;
   $314 = $313 & 32;
   $315 = ($314|0)==(0);
   if ($315) {
    HEAP32[$vararg_buffer12>>2] = 1159;
    $vararg_ptr15 = ((($vararg_buffer12)) + 4|0);
    HEAP32[$vararg_ptr15>>2] = $$lcssa$i$lcssa;
    (_fprintf($56,512355,$vararg_buffer12)|0);
   }
   $316 = ((($fr)) + 28376|0);
   HEAP32[$316>>2] = 28;
   $$1$i23 = -1;
   label = 102;
   break;
  }
  case 99: {
   $317 = $$lcssa565 & 32;
   $318 = ($317|0)==(0);
   if ($318) {
    HEAP32[$vararg_buffer16>>2] = 1173;
    (_fprintf($56,512482,$vararg_buffer16)|0);
   }
   $319 = ((($fr)) + 28376|0);
   HEAP32[$319>>2] = 27;
   $$1$i23 = -1;
   label = 102;
   break;
  }
  case 108: {
   $336 = ($$lcssa569|0)==(-1);
   if ($336) {
    $337 = HEAP32[$53>>2]|0;
    $338 = $337 & 32;
    $339 = ($338|0)==(0);
    if ($339) {
     HEAP32[$vararg_buffer19>>2] = 966;
     (_fprintf($56,512531,$vararg_buffer19)|0);
    }
   }
   $340 = ($$lcssa569|0)==(-10);
   $341 = $340 ? -10 : -1;
   $$0$i27$ph = $341;
   label = 123;
   break;
  }
  case 113: {
   $355 = HEAP32[$53>>2]|0;
   $356 = $355 & 32;
   $357 = ($356|0)==(0);
   if ($357) {
    HEAP32[$vararg_buffer22>>2] = 975;
    (_fprintf($56,512565,$vararg_buffer22)|0);
    $$0$i27$ph = -1;
    label = 123;
   } else {
    $$0$i27$ph = -1;
    label = 123;
   }
   break;
  }
  case 116: {
   $358 = HEAP32[$53>>2]|0;
   $359 = $358 & 32;
   $360 = ($359|0)==(0);
   if ($360) {
    HEAP32[$vararg_buffer25>>2] = 984;
    (_fprintf($56,512604,$vararg_buffer25)|0);
   }
   $ret$1 = 10;
   break;
  }
  case 121: {
   $377 = HEAP32[$53>>2]|0;
   $378 = $377 & 32;
   $379 = ($378|0)==(0);
   if ($379) {
    HEAP32[$vararg_buffer28>>2] = 996;
    (_fprintf($56,512679,$vararg_buffer28)|0);
    $$0$i27$ph = -1;
    label = 123;
   } else {
    $$0$i27$ph = -1;
    label = 123;
   }
   break;
  }
  case 236: {
   _INT123_do_rva($fr);
   $$pre280 = HEAP32[$68>>2]|0;
   $766 = $$pre280;$837 = $$lcssa575;
   label = 238;
   break;
  }
  case 237: {
   $763 = $$lcssa576;
   $766 = $763;$837 = $$lcssa574;
   label = 238;
   break;
  }
 }
 if ((label|0) == 102) {
  $ret$1 = $$1$i23;
 }
 else if ((label|0) == 123) {
  $380 = HEAP32[$49>>2]|0;
  $381 = ((($380)) + 28|0);
  $382 = HEAP32[$381>>2]|0;
  (FUNCTION_TABLE_iii[$382 & 8191]($fr,4)|0);
  $ret$1 = $$0$i27$ph;
 }
 else if ((label|0) == 238) {
  $764 = ((($fr)) + 19784|0);
  HEAP32[$764>>2] = 0;
  $765 = ((($fr)) + 19788|0);
  HEAP32[$765>>2] = $766;
  $767 = ((($fr)) + 19856|0);
  $768 = HEAP32[$767>>2]|0;
  $769 = (($768) + 1)|0;
  HEAP32[$767>>2] = $769;
  $770 = ($769|0)==(0);
  if (!($770)) {
   $771 = (+($768|0));
   $772 = ((($fr)) + 19848|0);
   $773 = +HEAPF64[$772>>3];
   $774 = $773 * $771;
   $775 = HEAP32[$72>>2]|0;
   switch ($775|0) {
   case 1:  {
    $776 = ((($fr)) + 19700|0);
    $777 = HEAP32[$776>>2]|0;
    $778 = HEAP32[$74>>2]|0;
    $779 = ((35332 + (($778*192)|0)|0) + ($777<<2)|0);
    $780 = HEAP32[$779>>2]|0;
    $781 = (+($780|0));
    $782 = $781 * 48000.0;
    $783 = ((($fr)) + 19704|0);
    $784 = HEAP32[$783>>2]|0;
    $785 = (35716 + ($784<<2)|0);
    $786 = HEAP32[$785>>2]|0;
    $787 = $786 << $778;
    $788 = (+($787|0));
    $789 = $782 / $788;
    $bpf$0$i = $789;
    break;
   }
   case 3: case 2:  {
    $790 = ((($fr)) + 19700|0);
    $791 = HEAP32[$790>>2]|0;
    $792 = (($775) + -1)|0;
    $793 = HEAP32[$74>>2]|0;
    $794 = (((35332 + (($793*192)|0)|0) + ($792<<6)|0) + ($791<<2)|0);
    $795 = HEAP32[$794>>2]|0;
    $796 = (+($795|0));
    $797 = $796 * 144000.0;
    $798 = ((($fr)) + 19704|0);
    $799 = HEAP32[$798>>2]|0;
    $800 = (35716 + ($799<<2)|0);
    $801 = HEAP32[$800>>2]|0;
    $802 = $801 << $793;
    $803 = (+($802|0));
    $804 = $797 / $803;
    $bpf$0$i = $804;
    break;
   }
   default: {
    $bpf$0$i = 1.0;
   }
   }
   $805 = $bpf$0$i + $774;
   $806 = (+($769|0));
   $807 = $805 / $806;
   HEAPF64[$772>>3] = $807;
  }
  $808 = HEAP32[$70>>2]|0;
  $809 = (($808) + 1)|0;
  HEAP32[$70>>2] = $809;
  $810 = ((($fr)) + 19764|0);
  $811 = HEAP32[$810>>2]|0;
  $812 = $811 & 2;
  $813 = ($812|0)==(0);
  do {
   if ($813) {
    $814 = HEAP32[$86>>2]|0;
    $815 = ($814|0)<(1);
    $816 = ($809|0)<($814|0);
    $or$cond = $815 | $816;
    if ($or$cond) {
     $817 = ((($fr)) + 27920|0);
     $818 = HEAP32[$817>>2]|0;
     $819 = ($818|0)<(1);
     $820 = ($809|0)<($818|0);
     $or$cond49 = $819 | $820;
     if ($or$cond49) {
      break;
     }
    }
    $821 = $811 | 2;
    HEAP32[$810>>2] = $821;
    $822 = HEAP32[$53>>2]|0;
    $823 = $822 & 32;
    $824 = ($823|0)==(0);
    if ($824) {
     $825 = ((($fr)) + 27920|0);
     $826 = HEAP32[$825>>2]|0;
     $827 = ($826|0)>(0);
     $$ = $827 ? $826 : $814;
     HEAP32[$vararg_buffer79>>2] = $809;
     $vararg_ptr82 = ((($vararg_buffer79)) + 4|0);
     HEAP32[$vararg_ptr82>>2] = $$;
     (_fprintf($56,513658,$vararg_buffer79)|0);
    }
   }
  } while(0);
  $828 = HEAP32[$5>>2]|0;
  $829 = ($828|0)==(0);
  if (!($829)) {
   $830 = HEAP32[$72>>2]|0;
   $831 = ($830|0)==(3);
   if ($831) {
    $832 = ((($fr)) + 9548|0);
    $833 = HEAP32[$68>>2]|0;
    $834 = ((($fr)) + 19864|0);
    $835 = HEAP32[$834>>2]|0;
    _memcpy(($832|0),($833|0),($835|0))|0;
   }
  }
  $836 = ((($fr)) + 19752|0);
  HEAP32[$836>>2] = $837;
  $838 = HEAP32[$810>>2]|0;
  $839 = $838 & 1;
  $840 = ($839|0)==(0);
  if (!($840)) {
   $841 = ((($fr)) + 27832|0);
   $842 = ((($fr)) + 27844|0);
   $843 = HEAP32[$842>>2]|0;
   $844 = ($843|0)==(0);
   if (!($844)) {
    $845 = HEAP32[$70>>2]|0;
    $846 = ((($fr)) + 27840|0);
    $847 = HEAP32[$846>>2]|0;
    $848 = ($845|0)==($847|0);
    if ($848) {
     _INT123_fi_add($841,$837);
    }
   }
  }
  $849 = HEAP8[$55>>0]|0;
  $850 = ($849<<24>>24)>(0);
  if ($850) {
   $851 = (($849) + -1)<<24>>24;
   HEAP8[$55>>0] = $851;
  }
  $852 = HEAP32[$49>>2]|0;
  $853 = ((($852)) + 44|0);
  $854 = HEAP32[$853>>2]|0;
  $855 = ($854|0)==(0|0);
  if (!($855)) {
   FUNCTION_TABLE_vi[$854 & 8191]($fr);
  }
  $856 = ((($fr)) + 27904|0);
  HEAP32[$856>>2] = 1;
  $857 = ((($fr)) + 27900|0);
  HEAP32[$857>>2] = 1;
  $858 = ((($fr)) + 19696|0);
  $859 = HEAP32[$858>>2]|0;
  $860 = ($859|0)==(0);
  if (!($860)) {
   $861 = HEAP32[$765>>2]|0;
   $862 = HEAP8[$861>>0]|0;
   $863 = $862&255;
   $864 = $863 << 8;
   $865 = ((($861)) + 1|0);
   $866 = HEAP8[$865>>0]|0;
   $867 = $866&255;
   $868 = $864 | $867;
   $869 = $868 << 8;
   $870 = ((($861)) + 2|0);
   $871 = HEAP8[$870>>0]|0;
   $872 = $871&255;
   $873 = $869 | $872;
   $874 = HEAP32[$764>>2]|0;
   $875 = $873 << $874;
   $876 = (($874) + 16)|0;
   $877 = $875 >>> 8;
   $878 = $877 & 65535;
   $879 = $876 >> 3;
   $880 = (($861) + ($879)|0);
   HEAP32[$765>>2] = $880;
   $881 = $876 & 7;
   HEAP32[$764>>2] = $881;
   $882 = ((($fr)) + 27952|0);
   HEAP32[$882>>2] = $878;
  }
  $883 = HEAP32[$newhead>>2]|0;
  HEAP32[$51>>2] = $883;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 $884 = HEAP32[$49>>2]|0;
 $885 = ((($884)) + 44|0);
 $886 = HEAP32[$885>>2]|0;
 $887 = ($886|0)==(0|0);
 if (!($887)) {
  FUNCTION_TABLE_vi[$886 & 8191]($fr);
 }
 HEAP8[$55>>0] = 0;
 $888 = ((($fr)) + 28376|0);
 $889 = HEAP32[$888>>2]|0;
 $890 = ($889|0)==(0);
 if ($890) {
  HEAP32[$888>>2] = 18;
 }
 HEAP32[$0>>2] = $1;
 HEAP32[$2>>2] = $3;
 $$0 = $ret$1;
 STACKTOP = sp;return ($$0|0);
}
function _INT123_parse_new_id3($fr,$first4bytes) {
 $fr = $fr|0;
 $first4bytes = $first4bytes|0;
 var $$ = 0, $$0 = 0, $$0$i$i = 0, $$0$i$i47 = 0, $$0$i1$i = 0, $$0$i6$i = 0, $$01$i$i = 0, $$16 = 0, $$187 = 0, $$lcssa = 0, $$lcssa253 = 0, $$lcssa254 = 0, $$lcssa258 = 0, $$lcssa262 = 0, $$lcssa266 = 0, $$lobit$i = 0, $$lobit$i41 = 0, $$lobit$i50 = 0, $$lobit1$i = 0, $$lobit1$i42 = 0;
 var $$lobit2$i = 0, $$lobit2$i43 = 0, $$lobit3$i = 0, $$lobit3$i46 = 0, $$not = 0, $$off = 0, $$off63 = 0, $$off64 = 0, $$pre = 0, $$pre$i = 0, $$pre$i44 = 0, $$pre$phi$i32Z2D = 0, $$pre$phi$iZ2D = 0, $$pre$phi28$iZ2D = 0, $$pre$phi31$iZ2D = 0, $$pre27$i = 0, $$pre28$i = 0, $$pre29$i = 0, $$pre30$i = 0, $$ret$3 = 0;
 var $$tt$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0;
 var $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0;
 var $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0;
 var $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0;
 var $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0;
 var $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0;
 var $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0;
 var $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0;
 var $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0;
 var $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0;
 var $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0;
 var $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0;
 var $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0;
 var $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0;
 var $476 = 0.0, $477 = 0.0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0.0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0;
 var $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0;
 var $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0;
 var $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0;
 var $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0;
 var $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0;
 var $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0;
 var $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0;
 var $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0.0, $633 = 0.0, $634 = 0, $635 = 0, $636 = 0, $637 = 0;
 var $638 = 0, $639 = 0, $64 = 0, $640 = 0.0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0.0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0;
 var $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0;
 var $674 = 0, $675 = 0, $676 = 0, $677 = 0.0, $678 = 0.0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0.0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0;
 var $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0;
 var $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0;
 var $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0;
 var $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0;
 var $764 = 0, $765 = 0, $766 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $buf = 0, $encoding$off$i$i = 0, $encoding$off$i$i19 = 0, $fflags$0 = 0, $framesize$0 = 0, $i$0$lcssa$us$i$i = 0, $i$0$lcssa$us$i$i30 = 0, $i$02$i = 0, $i$02$i$lcssa = 0, $i$03$us$i$i = 0, $i$03$us$i$i28 = 0, $i$099 = 0;
 var $id = 0, $ipos$0 = 0, $ipos$0$in = 0, $ipos$0$in$ph = 0, $ipos$0$lcssa252 = 0, $is_peak$0$ph$i = 0, $localcom$i = 0, $localex$i = 0, $opos$0$ph = 0, $or$cond = 0, $or$cond$i = 0, $or$cond115 = 0, $or$cond18 = 0, $or$cond5 = 0, $or$cond66 = 0, $pos$0 = 0, $realdata$0 = 0, $realsize$1 = 0, $ret$1 = 0, $ret$3 = 0;
 var $rva_mode$0 = 0, $rva_mode$0$ph$i = 0, $rva_mode$0$ph$i45 = 0, $sext = 0, $sext153 = 0, $tagpos$0$ph = 0, $tagpos$1101 = 0, $tagpos$2 = 0, $text$06$i$i = 0, $text$06$i$i33 = 0, $text$06$i$us$i = 0, $text$06$i$us$i$lcssa = 0, $text$06$i$us$i22 = 0, $text$06$i$us$i22$lcssa = 0, $text$06$i$us12$i = 0, $text$06$i$us12$i$lcssa = 0, $text$06$i$us13$i = 0, $text$06$i$us13$i$lcssa = 0, $text$06$us$i$i = 0, $text$06$us$i$i25 = 0;
 var $text$3$$i$i = 0, $text$3$i$i = 0, $text$3$i$i37 = 0, $tt$0 = 0, $vararg_buffer = 0, $vararg_buffer10 = 0, $vararg_buffer101 = 0, $vararg_buffer108 = 0, $vararg_buffer112 = 0, $vararg_buffer115 = 0, $vararg_buffer15 = 0, $vararg_buffer22 = 0, $vararg_buffer26 = 0, $vararg_buffer30 = 0, $vararg_buffer33 = 0, $vararg_buffer37 = 0, $vararg_buffer42 = 0, $vararg_buffer45 = 0, $vararg_buffer48 = 0, $vararg_buffer5 = 0;
 var $vararg_buffer52 = 0, $vararg_buffer55 = 0, $vararg_buffer58 = 0, $vararg_buffer61 = 0, $vararg_buffer64 = 0, $vararg_buffer67 = 0, $vararg_buffer70 = 0, $vararg_buffer74 = 0, $vararg_buffer77 = 0, $vararg_buffer80 = 0, $vararg_buffer83 = 0, $vararg_buffer86 = 0, $vararg_buffer89 = 0, $vararg_buffer92 = 0, $vararg_buffer95 = 0, $vararg_buffer98 = 0, $vararg_ptr1 = 0, $vararg_ptr104 = 0, $vararg_ptr105 = 0, $vararg_ptr106 = 0;
 var $vararg_ptr107 = 0, $vararg_ptr111 = 0, $vararg_ptr118 = 0, $vararg_ptr13 = 0, $vararg_ptr14 = 0, $vararg_ptr18 = 0, $vararg_ptr19 = 0, $vararg_ptr2 = 0, $vararg_ptr20 = 0, $vararg_ptr21 = 0, $vararg_ptr25 = 0, $vararg_ptr29 = 0, $vararg_ptr3 = 0, $vararg_ptr36 = 0, $vararg_ptr4 = 0, $vararg_ptr40 = 0, $vararg_ptr41 = 0, $vararg_ptr51 = 0, $vararg_ptr73 = 0, $vararg_ptr8 = 0;
 var $vararg_ptr9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 416|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer115 = sp + 320|0;
 $vararg_buffer112 = sp + 312|0;
 $vararg_buffer108 = sp + 304|0;
 $vararg_buffer101 = sp + 280|0;
 $vararg_buffer98 = sp + 272|0;
 $vararg_buffer95 = sp + 264|0;
 $vararg_buffer92 = sp + 256|0;
 $vararg_buffer89 = sp + 248|0;
 $vararg_buffer86 = sp + 240|0;
 $vararg_buffer83 = sp + 232|0;
 $vararg_buffer80 = sp + 224|0;
 $vararg_buffer77 = sp + 216|0;
 $vararg_buffer74 = sp + 208|0;
 $vararg_buffer70 = sp + 200|0;
 $vararg_buffer67 = sp + 192|0;
 $vararg_buffer64 = sp + 184|0;
 $vararg_buffer61 = sp + 176|0;
 $vararg_buffer58 = sp + 168|0;
 $vararg_buffer55 = sp + 160|0;
 $vararg_buffer52 = sp + 152|0;
 $vararg_buffer48 = sp + 144|0;
 $vararg_buffer45 = sp + 136|0;
 $vararg_buffer42 = sp + 128|0;
 $vararg_buffer37 = sp + 112|0;
 $vararg_buffer33 = sp + 104|0;
 $vararg_buffer30 = sp + 96|0;
 $vararg_buffer26 = sp + 88|0;
 $vararg_buffer22 = sp + 80|0;
 $vararg_buffer15 = sp + 56|0;
 $vararg_buffer10 = sp + 40|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer = sp;
 $localex$i = sp + 368|0;
 $localcom$i = sp + 336|0;
 $buf = sp + 400|0;
 $id = sp + 328|0;
 $0 = $first4bytes&255;
 $1 = $first4bytes & 255;
 $2 = ($1|0)==(255);
 if ($2) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ((($fr)) + 27956|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 24|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (FUNCTION_TABLE_iiii[$6 & 8191]($fr,$buf,6)|0);
 $8 = ($7|0)<(0);
 if ($8) {
  $$0 = $7;
  STACKTOP = sp;return ($$0|0);
 }
 $9 = HEAP8[$buf>>0]|0;
 $10 = $9&255;
 $11 = ($9<<24>>24)==(-1);
 if ($11) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $12 = ((($buf)) + 1|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = ((($buf)) + 2|0);
 $15 = HEAP8[$14>>0]|0;
 $16 = ((($buf)) + 3|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = $17 | $15;
 $19 = ((($buf)) + 4|0);
 $20 = HEAP8[$19>>0]|0;
 $21 = $18 | $20;
 $22 = ((($buf)) + 5|0);
 $23 = HEAP8[$22>>0]|0;
 $24 = $21 | $23;
 $25 = ($24<<24>>24)<(0);
 if ($25) {
  $41 = ((($fr)) + 28068|0);
  $42 = HEAP32[$41>>2]|0;
  $43 = $42 & 32;
  $44 = ($43|0)==(0);
  if (!($44)) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
  $45 = HEAP32[37544]|0;
  $46 = $15&255;
  $47 = $17&255;
  $48 = $20&255;
  $49 = $23&255;
  HEAP32[$vararg_buffer>>2] = 583;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $46;
  $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
  HEAP32[$vararg_ptr2>>2] = $47;
  $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
  HEAP32[$vararg_ptr3>>2] = $48;
  $vararg_ptr4 = ((($vararg_buffer)) + 16|0);
  HEAP32[$vararg_ptr4>>2] = $49;
  (_fprintf($45,508195,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $26 = $23&255;
 $27 = $20&255;
 $28 = $15&255;
 $29 = $17&255;
 $30 = $28 << 21;
 $31 = $29 << 14;
 $32 = $31 | $30;
 $33 = $27 << 7;
 $34 = $32 | $33;
 $35 = $34 | $26;
 $36 = ((($fr)) + 28064|0);
 $37 = ((($fr)) + 28068|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = $38 & 32;
 $40 = ($39|0)==(0);
 if ($40) {
  $50 = HEAP32[$36>>2]|0;
  $51 = ($50|0)>(1);
  if ($51) {
   $52 = HEAP32[37544]|0;
   HEAP32[$vararg_buffer5>>2] = $1;
   $vararg_ptr8 = ((($vararg_buffer5)) + 4|0);
   HEAP32[$vararg_ptr8>>2] = $10;
   $vararg_ptr9 = ((($vararg_buffer5)) + 8|0);
   HEAP32[$vararg_ptr9>>2] = $35;
   (_fprintf($52,508294,$vararg_buffer5)|0);
   $$pre = HEAP32[$37>>2]|0;
   $54 = $$pre;
  } else {
   $54 = $38;
  }
 } else {
  $54 = $38;
 }
 $53 = $54 & 8192;
 $55 = ($53|0)==(0);
 if ($55) {
  $56 = $13&255;
  $57 = $56 & 15;
  $58 = ($57|0)!=(0);
  $$off = (($1) + -2)|0;
  $59 = ($$off>>>0)>(2);
  $60 = $59 | $58;
  if ($60) {
   label = 12;
  } else {
   $73 = ((($fr)) + 28524|0);
   HEAP8[$73>>0] = $0;
   $74 = (($35) + 1)|0;
   $75 = (_malloc($74)|0);
   $76 = ($75|0)==(0|0);
   if ($76) {
    $743 = $54 & 32;
    $744 = ($743|0)==(0);
    if ($744) {
     $745 = HEAP32[37544]|0;
     HEAP32[$vararg_buffer115>>2] = 816;
     $vararg_ptr118 = ((($vararg_buffer115)) + 4|0);
     HEAP32[$vararg_ptr118>>2] = $35;
     (_fprintf($745,510868,$vararg_buffer115)|0);
    }
    $746 = HEAP32[$3>>2]|0;
    $747 = ((($746)) + 20|0);
    $748 = HEAP32[$747>>2]|0;
    $749 = (FUNCTION_TABLE_iii[$748 & 8191]($fr,$35)|0);
    $750 = ($749|0)<(0);
    $$16 = $750 ? $749 : 0;
    $$0 = $$16;
    STACKTOP = sp;return ($$0|0);
   }
   $77 = HEAP32[$3>>2]|0;
   $78 = ((($77)) + 24|0);
   $79 = HEAP32[$78>>2]|0;
   $80 = (FUNCTION_TABLE_iiii[$79 & 8191]($fr,$75,$35)|0);
   $81 = ($80|0)>(0);
   L31: do {
    if ($81) {
     $82 = (($75) + ($35)|0);
     HEAP8[$82>>0] = 0;
     $83 = $56 & 64;
     $84 = ($83|0)==(0);
     do {
      if ($84) {
       $tagpos$0$ph = 0;
      } else {
       $85 = ($1|0)==(3);
       $86 = HEAP8[$75>>0]|0;
       $87 = $86&255;
       if ($85) {
        $88 = $87 << 24;
        $89 = ((($75)) + 1|0);
        $90 = HEAP8[$89>>0]|0;
        $91 = $90&255;
        $92 = $91 << 16;
        $93 = $92 | $88;
        $94 = ((($75)) + 2|0);
        $95 = HEAP8[$94>>0]|0;
        $96 = $95&255;
        $97 = $96 << 8;
        $98 = $93 | $97;
        $99 = ((($75)) + 3|0);
        $100 = HEAP8[$99>>0]|0;
        $101 = $100&255;
        $102 = $98 | $101;
        $tagpos$0$ph = $102;
        break;
       }
       $103 = ((($75)) + 1|0);
       $104 = HEAP8[$103>>0]|0;
       $105 = $104&255;
       $106 = $104 | $86;
       $107 = ((($75)) + 2|0);
       $108 = HEAP8[$107>>0]|0;
       $109 = $108&255;
       $110 = $106 | $108;
       $111 = ((($75)) + 3|0);
       $112 = HEAP8[$111>>0]|0;
       $113 = $112&255;
       $114 = $110 | $112;
       $115 = ($114<<24>>24)<(0);
       if (!($115)) {
        $116 = $87 << 21;
        $117 = $105 << 14;
        $118 = $117 | $116;
        $119 = $109 << 7;
        $120 = $118 | $119;
        $121 = $120 | $113;
        $tagpos$0$ph = $121;
        break;
       }
       $122 = HEAP32[$37>>2]|0;
       $123 = $122 & 32;
       $124 = ($123|0)==(0);
       if (!($124)) {
        $ret$1 = 0;
        break L31;
       }
       $125 = HEAP32[37544]|0;
       HEAP32[$vararg_buffer15>>2] = 626;
       $vararg_ptr18 = ((($vararg_buffer15)) + 4|0);
       HEAP32[$vararg_ptr18>>2] = $87;
       $vararg_ptr19 = ((($vararg_buffer15)) + 8|0);
       HEAP32[$vararg_ptr19>>2] = $105;
       $vararg_ptr20 = ((($vararg_buffer15)) + 12|0);
       HEAP32[$vararg_ptr20>>2] = $109;
       $vararg_ptr21 = ((($vararg_buffer15)) + 16|0);
       HEAP32[$vararg_ptr21>>2] = $113;
       (_fprintf($125,508502,$vararg_buffer15)|0);
       $ret$1 = 0;
       break L31;
      }
     } while(0);
     $126 = ((($id)) + 4|0);
     HEAP8[$126>>0] = 0;
     $127 = (($35) + -10)|0;
     $128 = ($tagpos$0$ph>>>0)<($127>>>0);
     L43: do {
      if ($128) {
       $129 = HEAP32[37544]|0;
       $130 = $56 & 128;
       $131 = ((($fr)) + 28560|0);
       $132 = ((($fr)) + 28564|0);
       $133 = ((($fr)) + 28552|0);
       $134 = ((($fr)) + 28556|0);
       $135 = ((($localcom$i)) + 20|0);
       $136 = ((($localcom$i)) + 8|0);
       $137 = ((($localcom$i)) + 16|0);
       $138 = ((($localcom$i)) + 8|0);
       $139 = ((($localcom$i)) + 28|0);
       $140 = ((($fr)) + 28568|0);
       $141 = ((($fr)) + 28572|0);
       $142 = ((($localex$i)) + 20|0);
       $143 = ((($localex$i)) + 8|0);
       $144 = ((($localex$i)) + 16|0);
       $145 = ((($localex$i)) + 28|0);
       $146 = ((($fr)) + 28064|0);
       $147 = ($1|0)==(3);
       $tagpos$1101 = $tagpos$0$ph;
       L45: while(1) {
        $148 = HEAP8[$73>>0]|0;
        $149 = ($148<<24>>24)==(2);
        $150 = $149 ? 3 : 4;
        $i$099 = 0;
        while(1) {
         $153 = (($i$099) + ($tagpos$1101))|0;
         $154 = (($75) + ($153)|0);
         $155 = HEAP8[$154>>0]|0;
         $$off63 = (($155) + -48)<<24>>24;
         $156 = ($$off63&255)<(10);
         $$off64 = (($155) + -65)<<24>>24;
         $157 = ($$off64&255)<(26);
         $or$cond66 = $156 | $157;
         $152 = (($i$099) + 1)|0;
         if (!($or$cond66)) {
          break L43;
         }
         $151 = ($152|0)<($150|0);
         if ($151) {
          $i$099 = $152;
         } else {
          break;
         }
        }
        $158 = (($75) + ($tagpos$1101)|0);
        (_strncpy($id,$158,$150)|0);
        $159 = (($id) + ($150)|0);
        HEAP8[$159>>0] = 0;
        $160 = (($150) + ($tagpos$1101))|0;
        $161 = HEAP8[$73>>0]|0;
        $162 = ($161<<24>>24)==(2);
        $163 = (($75) + ($160)|0);
        $164 = HEAP8[$163>>0]|0;
        $165 = $164&255;
        do {
         if ($162) {
          $166 = $165 << 16;
          $167 = ((($163)) + 1|0);
          $168 = HEAP8[$167>>0]|0;
          $169 = $168&255;
          $170 = $169 << 8;
          $171 = $170 | $166;
          $172 = ((($163)) + 2|0);
          $173 = HEAP8[$172>>0]|0;
          $174 = $173&255;
          $175 = $171 | $174;
          $framesize$0 = $175;
         } else {
          if ($147) {
           $176 = $165 << 24;
           $177 = ((($163)) + 1|0);
           $178 = HEAP8[$177>>0]|0;
           $179 = $178&255;
           $180 = $179 << 16;
           $181 = $180 | $176;
           $182 = ((($163)) + 2|0);
           $183 = HEAP8[$182>>0]|0;
           $184 = $183&255;
           $185 = $184 << 8;
           $186 = $181 | $185;
           $187 = ((($163)) + 3|0);
           $188 = HEAP8[$187>>0]|0;
           $189 = $188&255;
           $190 = $186 | $189;
           $framesize$0 = $190;
           break;
          }
          $191 = ((($163)) + 1|0);
          $192 = HEAP8[$191>>0]|0;
          $193 = $192 | $164;
          $194 = ((($163)) + 2|0);
          $195 = HEAP8[$194>>0]|0;
          $196 = $193 | $195;
          $197 = ((($163)) + 3|0);
          $198 = HEAP8[$197>>0]|0;
          $199 = $196 | $198;
          $200 = ($199<<24>>24)<(0);
          if ($200) {
           label = 38;
           break L45;
          }
          $201 = $198&255;
          $202 = $195&255;
          $203 = $192&255;
          $204 = $165 << 21;
          $205 = $203 << 14;
          $206 = $205 | $204;
          $207 = $202 << 7;
          $208 = $206 | $207;
          $209 = $208 | $201;
          $framesize$0 = $209;
         }
        } while(0);
        $213 = HEAP32[$37>>2]|0;
        $214 = $213 & 32;
        $215 = ($214|0)==(0);
        if ($215) {
         $216 = HEAP32[$36>>2]|0;
         $217 = ($216|0)>(2);
         if ($217) {
          HEAP32[$vararg_buffer26>>2] = $id;
          $vararg_ptr29 = ((($vararg_buffer26)) + 4|0);
          HEAP32[$vararg_ptr29>>2] = $framesize$0;
          (_fprintf($129,508659,$vararg_buffer26)|0);
         }
        }
        $218 = (($160) + ($150))|0;
        $219 = (($218) + ($framesize$0))|0;
        $220 = ($219>>>0)>($35>>>0);
        if ($220) {
         label = 44;
         break;
        }
        $224 = HEAP8[$73>>0]|0;
        $225 = ($224&255)>(2);
        if ($225) {
         $226 = (($75) + ($218)|0);
         $227 = HEAP8[$226>>0]|0;
         $228 = $227&255;
         $229 = $228 << 8;
         $230 = (($218) + 1)|0;
         $231 = (($75) + ($230)|0);
         $232 = HEAP8[$231>>0]|0;
         $233 = $232&255;
         $234 = $229 | $233;
         $235 = (($218) + 2)|0;
         $236 = (($219) + 2)|0;
         $fflags$0 = $234;$pos$0 = $235;$tagpos$2 = $236;
        } else {
         $fflags$0 = 0;$pos$0 = $218;$tagpos$2 = $219;
        }
        do {
         if ($149) {
          $i$02$i = 0;
          while(1) {
           $237 = (35752 + ($i$02$i<<2)|0);
           $238 = HEAP32[$237>>2]|0;
           $239 = (_strncmp($id,$238,3)|0);
           $240 = ($239|0)==(0);
           if ($240) {
            $$lcssa = $238;$i$02$i$lcssa = $i$02$i;
            label = 50;
            break;
           }
           $249 = (($i$02$i) + 1)|0;
           $250 = ($249>>>0)<(37);
           if ($250) {
            $i$02$i = $249;
           } else {
            label = 54;
            break;
           }
          }
          if ((label|0) == 50) {
           label = 0;
           $241 = (35900 + ($i$02$i$lcssa<<2)|0);
           $242 = HEAP32[$241>>2]|0;
           $243 = HEAPU8[$242>>0]|(HEAPU8[$242+1>>0]<<8)|(HEAPU8[$242+2>>0]<<16)|(HEAPU8[$242+3>>0]<<24);
           HEAP32[$id>>2] = $243;
           $244 = HEAP32[$37>>2]|0;
           $245 = $244 & 32;
           $246 = ($245|0)==(0);
           if (!($246)) {
            label = 57;
            break;
           }
           $247 = HEAP32[$146>>2]|0;
           $248 = ($247|0)>(2);
           if (!($248)) {
            label = 57;
            break;
           }
           HEAP32[$vararg_buffer33>>2] = $$lcssa;
           $vararg_ptr36 = ((($vararg_buffer33)) + 4|0);
           HEAP32[$vararg_ptr36>>2] = $242;
           (_fprintf($129,509111,$vararg_buffer33)|0);
           label = 57;
           break;
          }
          else if ((label|0) == 54) {
           label = 0;
           $251 = HEAP32[$37>>2]|0;
           $252 = $251 & 32;
           $253 = ($252|0)==(0);
           if (!($253)) {
            break;
           }
           $254 = HEAP32[$146>>2]|0;
           $255 = ($254|0)>(2);
           if (!($255)) {
            break;
           }
           $256 = HEAP32[$id>>2]|0;
           $sext153 = $256 << 24;
           $257 = $sext153 >> 24;
           $258 = $256 >>> 8;
           $259 = $258 << 24;
           $260 = $259 >> 24;
           $261 = $256 << 8;
           $262 = $261 >> 24;
           HEAP32[$vararg_buffer37>>2] = $257;
           $vararg_ptr40 = ((($vararg_buffer37)) + 4|0);
           HEAP32[$vararg_ptr40>>2] = $260;
           $vararg_ptr41 = ((($vararg_buffer37)) + 8|0);
           HEAP32[$vararg_ptr41>>2] = $262;
           (_fprintf($129,509146,$vararg_buffer37)|0);
           break;
          }
         } else {
          label = 57;
         }
        } while(0);
        L78: do {
         if ((label|0) == 57) {
          label = 0;
          $263 = $fflags$0 & 36796;
          $264 = ($263|0)==(0);
          if (!($264)) {
           $267 = HEAP32[$37>>2]|0;
           $268 = $267 & 32;
           $269 = ($268|0)==(0);
           if (!($269)) {
            break;
           }
           HEAP32[$vararg_buffer42>>2] = 701;
           (_fprintf($129,509210,$vararg_buffer42)|0);
           break;
          }
          $265 = (_strncmp(509190,$id,4)|0);
          $266 = ($265|0)==(0);
          do {
           if ($266) {
            $tt$0 = 0;
           } else {
            $271 = (_strncmp((509195),$id,4)|0);
            $272 = ($271|0)==(0);
            if ($272) {
             $tt$0 = 1;
             break;
            }
            $759 = (_strncmp((509200),$id,4)|0);
            $760 = ($759|0)==(0);
            if ($760) {
             $tt$0 = 2;
             break;
            }
            $761 = (_strncmp((509205),$id,4)|0);
            $762 = ($761|0)==(0);
            $$187 = $762 ? 3 : -2;
            $tt$0 = $$187;
           }
          } while(0);
          $273 = HEAP8[$id>>0]|0;
          $274 = ($273<<24>>24)==(84);
          $275 = ($tt$0|0)!=(1);
          $or$cond = $275 & $274;
          $$tt$0 = $or$cond ? -1 : $tt$0;
          $276 = ($$tt$0|0)==(-2);
          if ($276) {
           break;
          }
          $277 = (($75) + ($pos$0)|0);
          $278 = $fflags$0 & 2;
          $279 = $278 | $130;
          $280 = ($279|0)==(0);
          L90: do {
           if ($280) {
            $realdata$0 = $277;$realsize$1 = $framesize$0;
           } else {
            $281 = (_malloc($framesize$0)|0);
            $282 = ($281|0)==(0|0);
            if ($282) {
             $283 = HEAP32[$37>>2]|0;
             $284 = $283 & 32;
             $285 = ($284|0)==(0);
             if (!($285)) {
              break L78;
             }
             HEAP32[$vararg_buffer45>>2] = 726;
             (_fprintf($129,509273,$vararg_buffer45)|0);
             break L78;
            }
            $286 = HEAP8[$277>>0]|0;
            HEAP8[$281>>0] = $286;
            $287 = (($pos$0) + ($framesize$0))|0;
            $ipos$0$in$ph = $pos$0;$opos$0$ph = 1;
            while(1) {
             $ipos$0$in = $ipos$0$in$ph;
             while(1) {
              $ipos$0 = (($ipos$0$in) + 1)|0;
              $288 = ($ipos$0>>>0)<($287>>>0);
              if (!($288)) {
               $realdata$0 = $281;$realsize$1 = $opos$0$ph;
               break L90;
              }
              $289 = (($75) + ($ipos$0)|0);
              $290 = HEAP8[$289>>0]|0;
              $291 = ($290<<24>>24)==(0);
              if (!($291)) {
               $$lcssa253 = $290;$ipos$0$lcssa252 = $ipos$0;
               break;
              }
              $292 = (($75) + ($ipos$0$in)|0);
              $293 = HEAP8[$292>>0]|0;
              $294 = ($293<<24>>24)==(-1);
              if ($294) {
               $ipos$0$in = $ipos$0;
              } else {
               $$lcssa253 = $290;$ipos$0$lcssa252 = $ipos$0;
               break;
              }
             }
             $295 = (($opos$0$ph) + 1)|0;
             $296 = (($281) + ($opos$0$ph)|0);
             HEAP8[$296>>0] = $$lcssa253;
             $ipos$0$in$ph = $ipos$0$lcssa252;$opos$0$ph = $295;
            }
           }
          } while(0);
          L104: do {
           switch ($$tt$0|0) {
           case 3: case 0:  {
            $297 = HEAP8[$realdata$0>>0]|0;
            $298 = ((($realdata$0)) + 1|0);
            $299 = ((($realdata$0)) + 4|0);
            $300 = $299;
            $301 = $realdata$0;
            $302 = ($realsize$1|0)<(4);
            L108: do {
             if ($302) {
              $303 = HEAP32[$37>>2]|0;
              $304 = $303 & 32;
              $305 = ($304|0)==(0);
              if (!($305)) {
               break;
              }
              HEAP32[$vararg_buffer48>>2] = 327;
              $vararg_ptr51 = ((($vararg_buffer48)) + 4|0);
              HEAP32[$vararg_ptr51>>2] = $realsize$1;
              (_fprintf($129,509347,$vararg_buffer48)|0);
             } else {
              $306 = ($$tt$0|0)==(3);
              do {
               if ($306) {
                $307 = HEAP32[$131>>2]|0;
                $308 = HEAP32[$132>>2]|0;
                $309 = $308 << 5;
                $310 = (($309) + 32)|0;
                $311 = (_INT123_safe_realloc($307,$310)|0);
                $312 = ($311|0)==(0|0);
                if ($312) {
                 $345 = 0;
                 break;
                }
                HEAP32[$131>>2] = $311;
                $313 = HEAP32[$132>>2]|0;
                $314 = (($313) + 1)|0;
                HEAP32[$132>>2] = $314;
                $315 = (((($311) + ($313<<5)|0)) + 20|0);
                _mpg123_init_string($315);
                $316 = (((($311) + ($313<<5)|0)) + 8|0);
                _mpg123_init_string($316);
                $317 = (($311) + ($313<<5)|0);
                ;HEAP32[$317>>2]=0|0;HEAP16[$317+4>>1]=0|0;HEAP8[$317+6>>0]=0|0;
                $318 = HEAP32[$132>>2]|0;
                $319 = (($318) + -1)|0;
                $320 = HEAP32[$131>>2]|0;
                $321 = (($320) + ($319<<5)|0);
                $345 = $321;
               } else {
                $322 = HEAP32[$133>>2]|0;
                $323 = HEAP32[$134>>2]|0;
                $324 = $323 << 5;
                $325 = (($324) + 32)|0;
                $326 = (_INT123_safe_realloc($322,$325)|0);
                $327 = ($326|0)==(0|0);
                if ($327) {
                 $345 = 0;
                 break;
                }
                HEAP32[$133>>2] = $326;
                $328 = HEAP32[$134>>2]|0;
                $329 = (($328) + 1)|0;
                HEAP32[$134>>2] = $329;
                $330 = (((($326) + ($328<<5)|0)) + 20|0);
                _mpg123_init_string($330);
                $331 = (((($326) + ($328<<5)|0)) + 8|0);
                _mpg123_init_string($331);
                $332 = (($326) + ($328<<5)|0);
                ;HEAP32[$332>>2]=0|0;HEAP16[$332+4>>1]=0|0;HEAP8[$332+6>>0]=0|0;
                $333 = HEAP32[$134>>2]|0;
                $334 = (($333) + -1)|0;
                $335 = HEAP32[$133>>2]|0;
                $336 = (($335) + ($334<<5)|0);
                $345 = $336;
               }
              } while(0);
              $337 = HEAP32[$37>>2]|0;
              $338 = $337 & 32;
              $339 = ($338|0)==(0);
              do {
               if ($339) {
                $340 = HEAP32[$36>>2]|0;
                $341 = ($340|0)>(3);
                if (!($341)) {
                 break;
                }
                $342 = HEAP8[$realdata$0>>0]|0;
                $343 = $342 << 24 >> 24;
                switch ($343|0) {
                case 0:  {
                 $$0$i6$i = 509454;
                 break;
                }
                case 1:  {
                 $$0$i6$i = 509443;
                 break;
                }
                case 2:  {
                 $$0$i6$i = 509433;
                 break;
                }
                case 3:  {
                 $$0$i6$i = 509427;
                 break;
                }
                default: {
                 $$0$i6$i = 509418;
                }
                }
                HEAP32[$vararg_buffer52>>2] = $$0$i6$i;
                (_fprintf($129,509462,$vararg_buffer52)|0);
               }
              } while(0);
              $344 = ($345|0)==(0|0);
              if ($344) {
               $346 = HEAP32[$37>>2]|0;
               $347 = $346 & 32;
               $348 = ($347|0)==(0);
               if (!($348)) {
                break;
               }
               HEAP32[$vararg_buffer55>>2] = 334;
               (_fprintf($129,509502,$vararg_buffer55)|0);
               break;
              }
              ;HEAP8[$345>>0]=HEAP8[$298>>0]|0;HEAP8[$345+1>>0]=HEAP8[$298+1>>0]|0;HEAP8[$345+2>>0]=HEAP8[$298+2>>0]|0;
              $349 = ((($345)) + 3|0);
              $350 = HEAP32[$id>>2]|0;
              HEAP8[$349>>0]=$350&255;HEAP8[$349+1>>0]=($350>>8)&255;HEAP8[$349+2>>0]=($350>>16)&255;HEAP8[$349+3>>0]=$350>>24;
              $351 = ((($realdata$0)) + 3|0);
              HEAP8[$351>>0] = $297;
              $352 = $297 << 24 >> 24;
              $353 = (($realsize$1) + -4)|0;
              $354 = (36048 + ($352<<2)|0);
              $355 = HEAP32[$354>>2]|0;
              $356 = ($353|0)>(0);
              L132: do {
               if ($356) {
                $encoding$off$i$i = (($352) + -1)|0;
                $357 = ($encoding$off$i$i>>>0)<(2);
                if ($357) {
                 $379 = 0;$text$06$us$i$i = $299;
                 while(1) {
                  $376 = HEAP8[$text$06$us$i$i>>0]|0;
                  $377 = ($376<<24>>24)==(0);
                  if ($377) {
                   $378 = (($353) - ($379))|0;
                   $380 = ($355>>>0)>($378>>>0);
                   if ($380) {
                    break L132;
                   } else {
                    $i$03$us$i$i = 1;
                   }
                   while(1) {
                    $381 = (($text$06$us$i$i) + ($i$03$us$i$i)|0);
                    $382 = HEAP8[$381>>0]|0;
                    $383 = ($382<<24>>24)==(0);
                    if (!($383)) {
                     $i$0$lcssa$us$i$i = $i$03$us$i$i;
                     break;
                    }
                    $389 = (($i$03$us$i$i) + 1)|0;
                    $390 = ($389>>>0)<($355>>>0);
                    if ($390) {
                     $i$03$us$i$i = $389;
                    } else {
                     $i$0$lcssa$us$i$i = $389;
                     break;
                    }
                   }
                   $384 = ($i$0$lcssa$us$i$i|0)==($355|0);
                   $385 = (($text$06$us$i$i) + ($355)|0);
                   if ($384) {
                    $text$3$i$i = $385;
                    label = 117;
                    break L132;
                   } else {
                    $$pre$phi$iZ2D = $385;
                   }
                  } else {
                   $$pre$i = (($text$06$us$i$i) + ($355)|0);
                   $$pre$phi$iZ2D = $$pre$i;
                  }
                  $386 = $$pre$phi$iZ2D;
                  $387 = (($386) - ($300))|0;
                  $388 = ($387|0)<($353|0);
                  if ($388) {
                   $379 = $387;$text$06$us$i$i = $$pre$phi$iZ2D;
                  } else {
                   $text$3$i$i = $$pre$phi$iZ2D;
                   label = 117;
                   break L132;
                  }
                 }
                }
                switch ($352|0) {
                case 3:  {
                 $763 = 0;$text$06$i$us$i = $299;
                 while(1) {
                  $358 = HEAP8[$text$06$i$us$i>>0]|0;
                  $359 = ($358<<24>>24)==(0);
                  if ($359) {
                   $$lcssa266 = $763;$text$06$i$us$i$lcssa = $text$06$i$us$i;
                   break;
                  }
                  $363 = (($text$06$i$us$i) + ($355)|0);
                  $364 = $363;
                  $365 = (($364) - ($300))|0;
                  $366 = ($365|0)<($353|0);
                  if ($366) {
                   $763 = $365;$text$06$i$us$i = $363;
                  } else {
                   $text$3$i$i = $363;
                   label = 117;
                   break L132;
                  }
                 }
                 $360 = (($353) - ($$lcssa266))|0;
                 $361 = ($355>>>0)>($360>>>0);
                 if ($361) {
                  break L132;
                 }
                 $362 = (($text$06$i$us$i$lcssa) + ($355)|0);
                 $text$3$i$i = $362;
                 label = 117;
                 break L132;
                 break;
                }
                case 0:  {
                 $764 = 0;$text$06$i$us12$i = $299;
                 while(1) {
                  $367 = HEAP8[$text$06$i$us12$i>>0]|0;
                  $368 = ($367<<24>>24)==(0);
                  if ($368) {
                   $$lcssa262 = $764;$text$06$i$us12$i$lcssa = $text$06$i$us12$i;
                   break;
                  }
                  $372 = (($text$06$i$us12$i) + ($355)|0);
                  $373 = $372;
                  $374 = (($373) - ($300))|0;
                  $375 = ($374|0)<($353|0);
                  if ($375) {
                   $764 = $374;$text$06$i$us12$i = $372;
                  } else {
                   $text$3$i$i = $372;
                   label = 117;
                   break L132;
                  }
                 }
                 $369 = (($353) - ($$lcssa262))|0;
                 $370 = ($355>>>0)>($369>>>0);
                 if ($370) {
                  break L132;
                 }
                 $371 = (($text$06$i$us12$i$lcssa) + ($355)|0);
                 $text$3$i$i = $371;
                 label = 117;
                 break L132;
                 break;
                }
                default: {
                 $394 = 0;$text$06$i$i = $299;
                 while(1) {
                  $391 = HEAP8[$text$06$i$i>>0]|0;
                  $392 = ($391<<24>>24)==(0);
                  if ($392) {
                   $393 = (($353) - ($394))|0;
                   $395 = ($355>>>0)>($393>>>0);
                   if ($395) {
                    break L132;
                   }
                   $396 = (($text$06$i$i) + ($355)|0);
                   switch ($352|0) {
                   case 0: case 3:  {
                    $text$3$i$i = $396;
                    label = 117;
                    break L132;
                    break;
                   }
                   default: {
                    $$pre$phi28$iZ2D = $396;
                   }
                   }
                  } else {
                   $$pre27$i = (($text$06$i$i) + ($355)|0);
                   $$pre$phi28$iZ2D = $$pre27$i;
                  }
                  $397 = $$pre$phi28$iZ2D;
                  $398 = (($397) - ($300))|0;
                  $399 = ($398|0)<($353|0);
                  if ($399) {
                   $394 = $398;$text$06$i$i = $$pre$phi28$iZ2D;
                  } else {
                   $text$3$i$i = $$pre$phi28$iZ2D;
                   label = 117;
                   break;
                  }
                 }
                }
                }
               } else {
                $text$3$i$i = $299;
                label = 117;
               }
              } while(0);
              do {
               if ((label|0) == 117) {
                label = 0;
                $400 = $text$3$i$i;
                $401 = (($400) - ($300))|0;
                $402 = ($401>>>0)>=($353>>>0);
                $403 = ($text$3$i$i|0)==(0|0);
                $or$cond$i = $403 | $402;
                if ($or$cond$i) {
                 break;
                }
                _mpg123_init_string($135);
                _mpg123_init_string($136);
                ;HEAP32[$localcom$i>>2]=0|0;HEAP16[$localcom$i+4>>1]=0|0;HEAP8[$localcom$i+6>>0]=0|0;
                $422 = ((($345)) + 8|0);
                $423 = (($401) + 1)|0;
                $424 = HEAP32[$37>>2]|0;
                $425 = $424 >>> 5;
                $$lobit$i = $425 & 1;
                $426 = $$lobit$i ^ 1;
                $427 = $424 & 2048;
                _store_id3_text($422,$351,$423,$426,$427);
                $428 = ($$tt$0|0)==(0);
                if ($428) {
                 $429 = HEAP32[$37>>2]|0;
                 $430 = $429 >>> 5;
                 $$lobit3$i = $430 & 1;
                 $431 = $$lobit3$i ^ 1;
                 _store_id3_text($136,$351,$423,$431,0);
                }
                $432 = ((($text$3$i$i)) + -1|0);
                HEAP8[$432>>0] = $297;
                $433 = ((($345)) + 20|0);
                $434 = (($realsize$1) + 1)|0;
                $435 = (($301) - ($400))|0;
                $436 = (($434) + ($435))|0;
                $437 = HEAP32[$37>>2]|0;
                $438 = $437 >>> 5;
                $$lobit1$i = $438 & 1;
                $439 = $$lobit1$i ^ 1;
                $440 = $437 & 2048;
                _store_id3_text($433,$432,$436,$439,$440);
                $441 = HEAP32[$37>>2]|0;
                $442 = $441 & 32;
                $443 = ($442|0)==(0);
                do {
                 if ($443) {
                  $444 = HEAP32[$36>>2]|0;
                  $445 = ($444|0)>(3);
                  if (!($445)) {
                   break;
                  }
                  $446 = ((($345)) + 16|0);
                  $447 = HEAP32[$446>>2]|0;
                  HEAP32[$vararg_buffer61>>2] = $447;
                  (_fprintf($129,510022,$vararg_buffer61)|0);
                  $448 = ((($345)) + 28|0);
                  $449 = HEAP32[$448>>2]|0;
                  HEAP32[$vararg_buffer64>>2] = $449;
                  (_fprintf($129,510063,$vararg_buffer64)|0);
                 }
                } while(0);
                $$not = $428 ^ 1;
                $450 = HEAP32[$137>>2]|0;
                $451 = ($450|0)==(0);
                $or$cond115 = $451 | $$not;
                L174: do {
                 if (!($or$cond115)) {
                  $452 = HEAP32[$138>>2]|0;
                  $453 = (_strcasecmp($452,510104)|0);
                  $454 = ($453|0)==(0);
                  do {
                   if ($454) {
                    $rva_mode$0$ph$i = 0;
                   } else {
                    $455 = (_strcasecmp($452,510108)|0);
                    $456 = ($455|0)==(0);
                    if ($456) {
                     $rva_mode$0$ph$i = 0;
                     break;
                    }
                    $457 = (_strcasecmp($452,510116)|0);
                    $458 = ($457|0)==(0);
                    if ($458) {
                     $rva_mode$0$ph$i = 0;
                     break;
                    }
                    $459 = (_strcasecmp($452,510126)|0);
                    $460 = ($459|0)==(0);
                    if ($460) {
                     $rva_mode$0$ph$i = 0;
                     break;
                    }
                    $461 = (_strcasecmp($452,510136)|0);
                    $462 = ($461|0)==(0);
                    if ($462) {
                     $rva_mode$0$ph$i = 1;
                     break;
                    }
                    $463 = (_strcasecmp($452,510146)|0);
                    $464 = ($463|0)==(0);
                    if ($464) {
                     $rva_mode$0$ph$i = 1;
                     break;
                    }
                    $465 = (_strcasecmp($452,510161)|0);
                    $466 = ($465|0)==(0);
                    if ($466) {
                     $rva_mode$0$ph$i = 1;
                    } else {
                     break L174;
                    }
                   }
                  } while(0);
                  $467 = (((($fr)) + 19816|0) + ($rva_mode$0$ph$i<<2)|0);
                  $468 = HEAP32[$467>>2]|0;
                  $469 = ($468|0)<(2);
                  if (!($469)) {
                   break;
                  }
                  $470 = HEAP32[$37>>2]|0;
                  $471 = $470 >>> 5;
                  $$lobit2$i = $471 & 1;
                  $472 = $$lobit2$i ^ 1;
                  _store_id3_text($135,$432,$436,$472,0);
                  $473 = HEAP32[$139>>2]|0;
                  $474 = ($473|0)==(0);
                  if ($474) {
                   break;
                  }
                  $475 = HEAP32[$135>>2]|0;
                  $476 = (+_atof($475));
                  $477 = $476;
                  $478 = (((($fr)) + 19824|0) + ($rva_mode$0$ph$i<<2)|0);
                  HEAPF32[$478>>2] = $477;
                  $479 = HEAP32[$37>>2]|0;
                  $480 = $479 & 32;
                  $481 = ($480|0)==(0);
                  do {
                   if ($481) {
                    $482 = HEAP32[$36>>2]|0;
                    $483 = ($482|0)>(2);
                    if (!($483)) {
                     break;
                    }
                    $484 = $477;
                    HEAPF64[$vararg_buffer67>>3] = $484;
                    (_fprintf($129,510170,$vararg_buffer67)|0);
                   }
                  } while(0);
                  $485 = (((($fr)) + 19832|0) + ($rva_mode$0$ph$i<<2)|0);
                  HEAPF32[$485>>2] = 0.0;
                  HEAP32[$467>>2] = 1;
                 }
                } while(0);
                _mpg123_free_string($135);
                _mpg123_free_string($136);
                break L108;
               }
              } while(0);
              $404 = HEAP32[$37>>2]|0;
              $405 = $404 & 32;
              $406 = ($405|0)==(0);
              if ($406) {
               HEAP32[$vararg_buffer58>>2] = 345;
               (_fprintf($129,509551,$vararg_buffer58)|0);
              }
              $407 = HEAP32[$134>>2]|0;
              $408 = ($407|0)==(0);
              if ($408) {
               break;
              }
              $409 = (($407) + -1)|0;
              $410 = HEAP32[$133>>2]|0;
              $411 = (((($410) + ($409<<5)|0)) + 20|0);
              _mpg123_free_string($411);
              $412 = (((($410) + ($409<<5)|0)) + 8|0);
              _mpg123_free_string($412);
              $413 = HEAP32[$134>>2]|0;
              $414 = ($413>>>0)>(1);
              $415 = HEAP32[$133>>2]|0;
              if (!($414)) {
               _free($415);
               HEAP32[$133>>2] = 0;
               HEAP32[$134>>2] = 0;
               break;
              }
              $416 = $413 << 5;
              $417 = (($416) + -32)|0;
              $418 = (_INT123_safe_realloc($415,$417)|0);
              $419 = ($418|0)==(0|0);
              if ($419) {
               break;
              }
              HEAP32[$133>>2] = $418;
              $420 = HEAP32[$134>>2]|0;
              $421 = (($420) + -1)|0;
              HEAP32[$134>>2] = $421;
             }
            } while(0);
            break;
           }
           case 1:  {
            $486 = HEAP8[$realdata$0>>0]|0;
            $487 = ((($realdata$0)) + 1|0);
            $488 = $487;
            $489 = $realdata$0;
            $490 = ($realsize$1|0)<(1);
            L202: do {
             if ($490) {
              $491 = HEAP32[$37>>2]|0;
              $492 = $491 & 32;
              $493 = ($492|0)==(0);
              if (!($493)) {
               break;
              }
              HEAP32[$vararg_buffer70>>2] = 409;
              $vararg_ptr73 = ((($vararg_buffer70)) + 4|0);
              HEAP32[$vararg_ptr73>>2] = $realsize$1;
              (_fprintf($129,509347,$vararg_buffer70)|0);
             } else {
              $494 = $486 << 24 >> 24;
              $495 = (($realsize$1) + -1)|0;
              $496 = (36048 + ($494<<2)|0);
              $497 = HEAP32[$496>>2]|0;
              $498 = ($495|0)>(0);
              L206: do {
               if ($498) {
                $encoding$off$i$i19 = (($494) + -1)|0;
                $499 = ($encoding$off$i$i19>>>0)<(2);
                if ($499) {
                 $521 = 0;$text$06$us$i$i25 = $487;
                 while(1) {
                  $518 = HEAP8[$text$06$us$i$i25>>0]|0;
                  $519 = ($518<<24>>24)==(0);
                  if ($519) {
                   $520 = (($495) - ($521))|0;
                   $522 = ($497>>>0)>($520>>>0);
                   if ($522) {
                    $$01$i$i = 0;
                    break L206;
                   } else {
                    $i$03$us$i$i28 = 1;
                   }
                   while(1) {
                    $523 = (($text$06$us$i$i25) + ($i$03$us$i$i28)|0);
                    $524 = HEAP8[$523>>0]|0;
                    $525 = ($524<<24>>24)==(0);
                    if (!($525)) {
                     $i$0$lcssa$us$i$i30 = $i$03$us$i$i28;
                     break;
                    }
                    $531 = (($i$03$us$i$i28) + 1)|0;
                    $532 = ($531>>>0)<($497>>>0);
                    if ($532) {
                     $i$03$us$i$i28 = $531;
                    } else {
                     $i$0$lcssa$us$i$i30 = $531;
                     break;
                    }
                   }
                   $526 = ($i$0$lcssa$us$i$i30|0)==($497|0);
                   $527 = (($text$06$us$i$i25) + ($497)|0);
                   if ($526) {
                    $text$3$i$i37 = $527;
                    label = 172;
                    break L206;
                   } else {
                    $$pre$phi$i32Z2D = $527;
                   }
                  } else {
                   $$pre29$i = (($text$06$us$i$i25) + ($497)|0);
                   $$pre$phi$i32Z2D = $$pre29$i;
                  }
                  $528 = $$pre$phi$i32Z2D;
                  $529 = (($528) - ($488))|0;
                  $530 = ($529|0)<($495|0);
                  if ($530) {
                   $521 = $529;$text$06$us$i$i25 = $$pre$phi$i32Z2D;
                  } else {
                   $text$3$i$i37 = $$pre$phi$i32Z2D;
                   label = 172;
                   break L206;
                  }
                 }
                }
                switch ($494|0) {
                case 3:  {
                 $765 = 0;$text$06$i$us$i22 = $487;
                 while(1) {
                  $500 = HEAP8[$text$06$i$us$i22>>0]|0;
                  $501 = ($500<<24>>24)==(0);
                  if ($501) {
                   $$lcssa258 = $765;$text$06$i$us$i22$lcssa = $text$06$i$us$i22;
                   break;
                  }
                  $505 = (($text$06$i$us$i22) + ($497)|0);
                  $506 = $505;
                  $507 = (($506) - ($488))|0;
                  $508 = ($507|0)<($495|0);
                  if ($508) {
                   $765 = $507;$text$06$i$us$i22 = $505;
                  } else {
                   $text$3$i$i37 = $505;
                   label = 172;
                   break L206;
                  }
                 }
                 $502 = (($495) - ($$lcssa258))|0;
                 $503 = ($497>>>0)>($502>>>0);
                 if ($503) {
                  $$01$i$i = 0;
                  break L206;
                 }
                 $504 = (($text$06$i$us$i22$lcssa) + ($497)|0);
                 $text$3$i$i37 = $504;
                 label = 172;
                 break L206;
                 break;
                }
                case 0:  {
                 $766 = 0;$text$06$i$us13$i = $487;
                 while(1) {
                  $509 = HEAP8[$text$06$i$us13$i>>0]|0;
                  $510 = ($509<<24>>24)==(0);
                  if ($510) {
                   $$lcssa254 = $766;$text$06$i$us13$i$lcssa = $text$06$i$us13$i;
                   break;
                  }
                  $514 = (($text$06$i$us13$i) + ($497)|0);
                  $515 = $514;
                  $516 = (($515) - ($488))|0;
                  $517 = ($516|0)<($495|0);
                  if ($517) {
                   $766 = $516;$text$06$i$us13$i = $514;
                  } else {
                   $text$3$i$i37 = $514;
                   label = 172;
                   break L206;
                  }
                 }
                 $511 = (($495) - ($$lcssa254))|0;
                 $512 = ($497>>>0)>($511>>>0);
                 if ($512) {
                  $$01$i$i = 0;
                  break L206;
                 }
                 $513 = (($text$06$i$us13$i$lcssa) + ($497)|0);
                 $text$3$i$i37 = $513;
                 label = 172;
                 break L206;
                 break;
                }
                default: {
                 $536 = 0;$text$06$i$i33 = $487;
                 while(1) {
                  $533 = HEAP8[$text$06$i$i33>>0]|0;
                  $534 = ($533<<24>>24)==(0);
                  if ($534) {
                   $535 = (($495) - ($536))|0;
                   $537 = ($497>>>0)>($535>>>0);
                   if ($537) {
                    $$01$i$i = 0;
                    break L206;
                   }
                   $538 = (($text$06$i$i33) + ($497)|0);
                   switch ($494|0) {
                   case 0: case 3:  {
                    $text$3$i$i37 = $538;
                    label = 172;
                    break L206;
                    break;
                   }
                   default: {
                    $$pre$phi31$iZ2D = $538;
                   }
                   }
                  } else {
                   $$pre30$i = (($text$06$i$i33) + ($497)|0);
                   $$pre$phi31$iZ2D = $$pre30$i;
                  }
                  $539 = $$pre$phi31$iZ2D;
                  $540 = (($539) - ($488))|0;
                  $541 = ($540|0)<($495|0);
                  if ($541) {
                   $536 = $540;$text$06$i$i33 = $$pre$phi31$iZ2D;
                  } else {
                   $text$3$i$i37 = $$pre$phi31$iZ2D;
                   label = 172;
                   break;
                  }
                 }
                }
                }
               } else {
                $text$3$i$i37 = $487;
                label = 172;
               }
              } while(0);
              if ((label|0) == 172) {
               label = 0;
               $542 = $text$3$i$i37;
               $543 = (($542) - ($488))|0;
               $544 = ($543>>>0)<($495>>>0);
               $text$3$$i$i = $544 ? $text$3$i$i37 : 0;
               $$01$i$i = $text$3$$i$i;
              }
              $545 = HEAP32[$37>>2]|0;
              $546 = $545 & 32;
              $547 = ($546|0)==(0);
              do {
               if ($547) {
                $548 = HEAP32[$36>>2]|0;
                $549 = ($548|0)>(3);
                if (!($549)) {
                 break;
                }
                switch ($494|0) {
                case 0:  {
                 $$0$i$i = 509454;
                 break;
                }
                case 1:  {
                 $$0$i$i = 509443;
                 break;
                }
                case 2:  {
                 $$0$i$i = 509433;
                 break;
                }
                case 3:  {
                 $$0$i$i = 509427;
                 break;
                }
                default: {
                 $$0$i$i = 509418;
                }
                }
                HEAP32[$vararg_buffer74>>2] = $$0$i$i;
                (_fprintf($129,510192,$vararg_buffer74)|0);
               }
              } while(0);
              $550 = ($$01$i$i|0)==(0|0);
              if ($550) {
               $551 = HEAP32[$37>>2]|0;
               $552 = $551 & 32;
               $553 = ($552|0)==(0);
               if (!($553)) {
                break;
               }
               HEAP32[$vararg_buffer77>>2] = 416;
               (_fprintf($129,510230,$vararg_buffer77)|0);
               break;
              }
              $554 = HEAP32[$140>>2]|0;
              $555 = HEAP32[$141>>2]|0;
              $556 = $555 << 5;
              $557 = (($556) + 32)|0;
              $558 = (_INT123_safe_realloc($554,$557)|0);
              $559 = ($558|0)==(0|0);
              do {
               if (!($559)) {
                HEAP32[$140>>2] = $558;
                $560 = HEAP32[$141>>2]|0;
                $561 = (($560) + 1)|0;
                HEAP32[$141>>2] = $561;
                $562 = (((($558) + ($560<<5)|0)) + 20|0);
                _mpg123_init_string($562);
                $563 = (((($558) + ($560<<5)|0)) + 8|0);
                _mpg123_init_string($563);
                $564 = (($558) + ($560<<5)|0);
                ;HEAP32[$564>>2]=0|0;HEAP16[$564+4>>1]=0|0;HEAP8[$564+6>>0]=0|0;
                $565 = HEAP32[$141>>2]|0;
                $566 = (($565) + -1)|0;
                $567 = HEAP32[$140>>2]|0;
                $568 = (($567) + ($566<<5)|0);
                $569 = ($568|0)==(0|0);
                if ($569) {
                 break;
                }
                $573 = (((($567) + ($566<<5)|0)) + 3|0);
                $574 = HEAP32[$id>>2]|0;
                HEAP8[$573>>0]=$574&255;HEAP8[$573+1>>0]=($574>>8)&255;HEAP8[$573+2>>0]=($574>>16)&255;HEAP8[$573+3>>0]=$574>>24;
                _mpg123_init_string($142);
                _mpg123_init_string($143);
                ;HEAP32[$localex$i>>2]=0|0;HEAP16[$localex$i+4>>1]=0|0;HEAP8[$localex$i+6>>0]=0|0;
                $575 = (((($567) + ($566<<5)|0)) + 8|0);
                $576 = $$01$i$i;
                $577 = (($576) - ($488))|0;
                $578 = (($577) + 1)|0;
                $579 = HEAP32[$37>>2]|0;
                $580 = $579 >>> 5;
                $$lobit$i41 = $580 & 1;
                $581 = $$lobit$i41 ^ 1;
                $582 = $579 & 2048;
                _store_id3_text($575,$realdata$0,$578,$581,$582);
                $583 = HEAP32[$37>>2]|0;
                $584 = $583 >>> 5;
                $$lobit1$i42 = $584 & 1;
                $585 = $$lobit1$i42 ^ 1;
                _store_id3_text($143,$realdata$0,$578,$585,0);
                $586 = ((($$01$i$i)) + -1|0);
                HEAP8[$586>>0] = $486;
                $587 = (((($567) + ($566<<5)|0)) + 20|0);
                $588 = (($realsize$1) + 1)|0;
                $589 = (($588) + ($489))|0;
                $590 = (($589) - ($576))|0;
                $591 = HEAP32[$37>>2]|0;
                $592 = $591 >>> 5;
                $$lobit2$i43 = $592 & 1;
                $593 = $$lobit2$i43 ^ 1;
                $594 = $591 & 2048;
                _store_id3_text($587,$586,$590,$593,$594);
                $595 = HEAP32[$144>>2]|0;
                $596 = ($595|0)==(0);
                L258: do {
                 if (!($596)) {
                  $597 = HEAP32[$143>>2]|0;
                  $598 = (_strncasecmp($597,510342,17)|0);
                  $599 = ($598|0)==(0);
                  do {
                   if ($599) {
                    $600 = HEAP32[$37>>2]|0;
                    $601 = $600 & 32;
                    $602 = ($601|0)==(0);
                    do {
                     if ($602) {
                      $603 = HEAP32[$36>>2]|0;
                      $604 = ($603|0)>(2);
                      if (!($604)) {
                       $605 = $597;
                       break;
                      }
                      (_fwrite(510360,37,1,$129)|0);
                      $$pre28$i = HEAP32[$143>>2]|0;
                      $605 = $$pre28$i;
                     } else {
                      $605 = $597;
                     }
                    } while(0);
                    $606 = (_strcasecmp($605,510398)|0);
                    $607 = ($606|0)==(0);
                    if ($607) {
                     $is_peak$0$ph$i = 1;$rva_mode$0$ph$i45 = 0;
                     break;
                    }
                    $608 = (_strcasecmp($605,510420)|0);
                    $609 = ($608|0)==(0);
                    if ($609) {
                     $is_peak$0$ph$i = 0;$rva_mode$0$ph$i45 = 0;
                    } else {
                     break L258;
                    }
                   } else {
                    $610 = (_strncasecmp($597,510442,17)|0);
                    $611 = ($610|0)==(0);
                    if (!($611)) {
                     break L258;
                    }
                    $612 = HEAP32[$37>>2]|0;
                    $613 = $612 & 32;
                    $614 = ($613|0)==(0);
                    do {
                     if ($614) {
                      $615 = HEAP32[$36>>2]|0;
                      $616 = ($615|0)>(2);
                      if (!($616)) {
                       $617 = $597;
                       break;
                      }
                      (_fwrite(510460,37,1,$129)|0);
                      $$pre$i44 = HEAP32[$143>>2]|0;
                      $617 = $$pre$i44;
                     } else {
                      $617 = $597;
                     }
                    } while(0);
                    $618 = (_strcasecmp($617,510498)|0);
                    $619 = ($618|0)==(0);
                    if ($619) {
                     $is_peak$0$ph$i = 1;$rva_mode$0$ph$i45 = 1;
                     break;
                    }
                    $620 = (_strcasecmp($617,510520)|0);
                    $621 = ($620|0)==(0);
                    if ($621) {
                     $is_peak$0$ph$i = 0;$rva_mode$0$ph$i45 = 1;
                    } else {
                     break L258;
                    }
                   }
                  } while(0);
                  $622 = (((($fr)) + 19816|0) + ($rva_mode$0$ph$i45<<2)|0);
                  $623 = HEAP32[$622>>2]|0;
                  $624 = ($623|0)<(3);
                  if (!($624)) {
                   break;
                  }
                  $625 = HEAP32[$37>>2]|0;
                  $626 = $625 >>> 5;
                  $$lobit3$i46 = $626 & 1;
                  $627 = $$lobit3$i46 ^ 1;
                  _store_id3_text($142,$586,$590,$627,0);
                  $628 = HEAP32[$145>>2]|0;
                  $629 = ($628|0)==(0);
                  if ($629) {
                   break;
                  }
                  $630 = ($is_peak$0$ph$i|0)==(0);
                  $631 = HEAP32[$142>>2]|0;
                  $632 = (+_atof($631));
                  $633 = $632;
                  do {
                   if ($630) {
                    $641 = (((($fr)) + 19824|0) + ($rva_mode$0$ph$i45<<2)|0);
                    HEAPF32[$641>>2] = $633;
                    $642 = HEAP32[$37>>2]|0;
                    $643 = $642 & 32;
                    $644 = ($643|0)==(0);
                    if (!($644)) {
                     break;
                    }
                    $645 = HEAP32[$36>>2]|0;
                    $646 = ($645|0)>(2);
                    if (!($646)) {
                     break;
                    }
                    $647 = $633;
                    HEAPF64[$vararg_buffer86>>3] = $647;
                    (_fprintf($129,510561,$vararg_buffer86)|0);
                   } else {
                    $634 = (((($fr)) + 19832|0) + ($rva_mode$0$ph$i45<<2)|0);
                    HEAPF32[$634>>2] = $633;
                    $635 = HEAP32[$37>>2]|0;
                    $636 = $635 & 32;
                    $637 = ($636|0)==(0);
                    if (!($637)) {
                     break;
                    }
                    $638 = HEAP32[$36>>2]|0;
                    $639 = ($638|0)>(2);
                    if (!($639)) {
                     break;
                    }
                    $640 = $633;
                    HEAPF64[$vararg_buffer83>>3] = $640;
                    (_fprintf($129,510542,$vararg_buffer83)|0);
                   }
                  } while(0);
                  HEAP32[$622>>2] = 2;
                 }
                } while(0);
                _mpg123_free_string($142);
                _mpg123_free_string($143);
                break L202;
               }
              } while(0);
              $570 = HEAP32[$37>>2]|0;
              $571 = $570 & 32;
              $572 = ($571|0)==(0);
              if (!($572)) {
               break;
              }
              HEAP32[$vararg_buffer80>>2] = 422;
              (_fprintf($129,510290,$vararg_buffer80)|0);
             }
            } while(0);
            break;
           }
           case 2:  {
            $648 = HEAP32[$37>>2]|0;
            $649 = $648 & 32;
            $650 = ($649|0)==(0);
            do {
             if ($650) {
              $651 = HEAP32[$36>>2]|0;
              $652 = ($651|0)>(2);
              if (!($652)) {
               break;
              }
              HEAP32[$vararg_buffer89>>2] = $realdata$0;
              (_fprintf($129,510582,$vararg_buffer89)|0);
             }
            } while(0);
            $653 = (_strncasecmp($realdata$0,510614,5)|0);
            $654 = ($653|0)==(0);
            do {
             if ($654) {
              label = 219;
             } else {
              $655 = (_strncasecmp($realdata$0,510620,10)|0);
              $656 = ($655|0)==(0);
              if ($656) {
               label = 219;
               break;
              }
              $657 = (_strncasecmp($realdata$0,510631,4)|0);
              $658 = ($657|0)==(0);
              if ($658) {
               label = 219;
              } else {
               $rva_mode$0 = 0;
              }
             }
            } while(0);
            if ((label|0) == 219) {
             label = 0;
             $rva_mode$0 = 1;
            }
            $659 = (((($fr)) + 19816|0) + ($rva_mode$0<<2)|0);
            $660 = HEAP32[$659>>2]|0;
            $661 = ($660|0)<(4);
            if (!($661)) {
             break L104;
            }
            $662 = (_strlen($realdata$0)|0);
            $663 = (($662) + 1)|0;
            $664 = (($realdata$0) + ($663)|0);
            $665 = HEAP8[$664>>0]|0;
            $666 = ($665<<24>>24)==(1);
            if (!($666)) {
             break L104;
            }
            $667 = (($662) + 2)|0;
            $668 = (($realdata$0) + ($667)|0);
            $669 = HEAP8[$668>>0]|0;
            $670 = $669 << 24 >> 24;
            $671 = $670 << 8;
            $672 = (($662) + 3)|0;
            $673 = (($realdata$0) + ($672)|0);
            $674 = HEAP8[$673>>0]|0;
            $675 = $674&255;
            $676 = $671 | $675;
            $677 = (+($676|0));
            $678 = $677 * 0.001953125;
            $679 = (((($fr)) + 19824|0) + ($rva_mode$0<<2)|0);
            HEAPF32[$679>>2] = $678;
            $680 = HEAP32[$37>>2]|0;
            $681 = $680 & 32;
            $682 = ($681|0)==(0);
            do {
             if ($682) {
              $683 = HEAP32[$36>>2]|0;
              $684 = ($683|0)>(2);
              if (!($684)) {
               break;
              }
              $685 = $678;
              HEAPF64[$vararg_buffer92>>3] = $685;
              (_fprintf($129,510170,$vararg_buffer92)|0);
             }
            } while(0);
            $686 = (((($fr)) + 19832|0) + ($rva_mode$0<<2)|0);
            HEAPF32[$686>>2] = 0.0;
            HEAP32[$659>>2] = 3;
            break;
           }
           case -1:  {
            $687 = HEAP32[$131>>2]|0;
            $688 = HEAP32[$132>>2]|0;
            $689 = $688 << 5;
            $690 = (($689) + 32)|0;
            $691 = (_INT123_safe_realloc($687,$690)|0);
            $692 = ($691|0)==(0|0);
            if ($692) {
             $$0$i$i47 = 0;
            } else {
             HEAP32[$131>>2] = $691;
             $693 = HEAP32[$132>>2]|0;
             $694 = (($693) + 1)|0;
             HEAP32[$132>>2] = $694;
             $695 = (((($691) + ($693<<5)|0)) + 20|0);
             _mpg123_init_string($695);
             $696 = (((($691) + ($693<<5)|0)) + 8|0);
             _mpg123_init_string($696);
             $697 = (($691) + ($693<<5)|0);
             ;HEAP32[$697>>2]=0|0;HEAP16[$697+4>>1]=0|0;HEAP8[$697+6>>0]=0|0;
             $698 = HEAP32[$132>>2]|0;
             $699 = (($698) + -1)|0;
             $700 = HEAP32[$131>>2]|0;
             $701 = (($700) + ($699<<5)|0);
             $$0$i$i47 = $701;
            }
            $702 = HEAP32[$37>>2]|0;
            $703 = $702 & 32;
            $704 = ($703|0)==(0);
            do {
             if ($704) {
              $705 = HEAP32[$36>>2]|0;
              $706 = ($705|0)>(3);
              if (!($706)) {
               break;
              }
              $707 = HEAP8[$realdata$0>>0]|0;
              $708 = $707 << 24 >> 24;
              switch ($708|0) {
              case 0:  {
               $$0$i1$i = 509454;
               break;
              }
              case 1:  {
               $$0$i1$i = 509443;
               break;
              }
              case 2:  {
               $$0$i1$i = 509433;
               break;
              }
              case 3:  {
               $$0$i1$i = 509427;
               break;
              }
              default: {
               $$0$i1$i = 509418;
              }
              }
              HEAP32[$vararg_buffer95>>2] = $$0$i1$i;
              (_fprintf($129,510636,$vararg_buffer95)|0);
             }
            } while(0);
            $709 = ($$0$i$i47|0)==(0|0);
            if ($709) {
             $710 = HEAP32[$37>>2]|0;
             $711 = $710 & 32;
             $712 = ($711|0)==(0);
             if (!($712)) {
              break L104;
             }
             HEAP32[$vararg_buffer98>>2] = 302;
             (_fprintf($129,510673,$vararg_buffer98)|0);
             break L104;
            }
            $713 = ((($$0$i$i47)) + 3|0);
            $714 = HEAP32[$id>>2]|0;
            HEAP8[$713>>0]=$714&255;HEAP8[$713+1>>0]=($714>>8)&255;HEAP8[$713+2>>0]=($714>>16)&255;HEAP8[$713+3>>0]=$714>>24;
            $715 = ((($$0$i$i47)) + 20|0);
            $716 = HEAP32[$37>>2]|0;
            $717 = $716 >>> 5;
            $$lobit$i50 = $717 & 1;
            $718 = $$lobit$i50 ^ 1;
            $719 = $716 & 2048;
            _store_id3_text($715,$realdata$0,$realsize$1,$718,$719);
            $720 = HEAP32[$37>>2]|0;
            $721 = $720 & 32;
            $722 = ($721|0)==(0);
            if (!($722)) {
             break L104;
            }
            $723 = HEAP32[$36>>2]|0;
            $724 = ($723|0)>(3);
            if (!($724)) {
             break L104;
            }
            $725 = HEAP32[$id>>2]|0;
            $sext = $725 << 24;
            $726 = $sext >> 24;
            $727 = $725 >>> 8;
            $728 = $727 << 24;
            $729 = $728 >> 24;
            $730 = $725 << 8;
            $731 = $730 >> 24;
            $732 = $725 >> 24;
            $733 = HEAP32[$715>>2]|0;
            HEAP32[$vararg_buffer101>>2] = $726;
            $vararg_ptr104 = ((($vararg_buffer101)) + 4|0);
            HEAP32[$vararg_ptr104>>2] = $729;
            $vararg_ptr105 = ((($vararg_buffer101)) + 8|0);
            HEAP32[$vararg_ptr105>>2] = $731;
            $vararg_ptr106 = ((($vararg_buffer101)) + 12|0);
            HEAP32[$vararg_ptr106>>2] = $732;
            $vararg_ptr107 = ((($vararg_buffer101)) + 16|0);
            HEAP32[$vararg_ptr107>>2] = $733;
            (_fprintf($129,510719,$vararg_buffer101)|0);
            break;
           }
           default: {
            $734 = HEAP32[$37>>2]|0;
            $735 = $734 & 32;
            $736 = ($735|0)==(0);
            if (!($736)) {
             break L104;
            }
            HEAP32[$vararg_buffer108>>2] = 786;
            $vararg_ptr111 = ((($vararg_buffer108)) + 4|0);
            HEAP32[$vararg_ptr111>>2] = $$tt$0;
            (_fprintf($129,510756,$vararg_buffer108)|0);
           }
           }
          } while(0);
          if ($280) {
           break;
          }
          _free($realdata$0);
         }
        } while(0);
        $270 = ($tagpos$2>>>0)<($127>>>0);
        if ($270) {
         $tagpos$1101 = $tagpos$2;
        } else {
         break L43;
        }
       }
       if ((label|0) == 38) {
        $210 = HEAP32[$37>>2]|0;
        $211 = $210 & 32;
        $212 = ($211|0)==(0);
        if (!($212)) {
         break;
        }
        HEAP32[$vararg_buffer22>>2] = 666;
        $vararg_ptr25 = ((($vararg_buffer22)) + 4|0);
        HEAP32[$vararg_ptr25>>2] = $id;
        (_fprintf($129,508572,$vararg_buffer22)|0);
        break;
       }
       else if ((label|0) == 44) {
        $221 = HEAP32[$37>>2]|0;
        $222 = $221 & 32;
        $223 = ($222|0)==(0);
        if (!($223)) {
         break;
        }
        HEAP32[$vararg_buffer30>>2] = 673;
        (_fprintf($129,508693,$vararg_buffer30)|0);
        break;
       }
      }
     } while(0);
     $ret$1 = 1;
    } else {
     $737 = ($35|0)==(0);
     if ($737) {
      $ret$1 = $80;
     } else {
      $738 = HEAP32[$37>>2]|0;
      $739 = $738 & 32;
      $740 = ($739|0)==(0);
      $741 = ($80|0)!=(-10);
      $or$cond5 = $741 & $740;
      if ($or$cond5) {
       $742 = HEAP32[37544]|0;
       HEAP32[$vararg_buffer112>>2] = 808;
       (_fprintf($742,510804,$vararg_buffer112)|0);
       $ret$1 = $80;
      } else {
       $ret$1 = $80;
      }
     }
    }
   } while(0);
   _free($75);
   $ret$3 = $ret$1;
  }
 } else {
  label = 12;
 }
 if ((label|0) == 12) {
  $61 = $54 & 32;
  $62 = ($61|0)==(0);
  do {
   if ($62) {
    if ($55) {
     $66 = HEAP32[37544]|0;
     $67 = $13&255;
     HEAP32[$vararg_buffer10>>2] = 599;
     $vararg_ptr13 = ((($vararg_buffer10)) + 4|0);
     HEAP32[$vararg_ptr13>>2] = $1;
     $vararg_ptr14 = ((($vararg_buffer10)) + 8|0);
     HEAP32[$vararg_ptr14>>2] = $67;
     (_fprintf($66,508378,$vararg_buffer10)|0);
     break;
    }
    $63 = HEAP32[$36>>2]|0;
    $64 = ($63|0)>(2);
    if ($64) {
     $65 = HEAP32[37544]|0;
     (_fwrite(508334,43,1,$65)|0);
    }
   }
  } while(0);
  $68 = HEAP32[$3>>2]|0;
  $69 = ((($68)) + 20|0);
  $70 = HEAP32[$69>>2]|0;
  $71 = (FUNCTION_TABLE_iii[$70 & 8191]($fr,$35)|0);
  $72 = ($71|0)<(0);
  $$ = $72 ? $71 : 1;
  $ret$3 = $$;
 }
 $751 = ($ret$3|0)<(1);
 $752 = $13 & 16;
 $753 = ($752<<24>>24)==(0);
 $or$cond18 = $753 | $751;
 if ($or$cond18) {
  $$0 = $ret$3;
  STACKTOP = sp;return ($$0|0);
 }
 $754 = HEAP32[$3>>2]|0;
 $755 = ((($754)) + 20|0);
 $756 = HEAP32[$755>>2]|0;
 $757 = (FUNCTION_TABLE_iii[$756 & 8191]($fr,$35)|0);
 $758 = ($757|0)<(0);
 $$ret$3 = $758 ? $757 : $ret$3;
 $$0 = $$ret$3;
 STACKTOP = sp;return ($$0|0);
}
function _mpg123_init_string($sb) {
 $sb = $sb|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$sb>>2] = 0;
 $0 = ((($sb)) + 4|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($sb)) + 8|0);
 HEAP32[$1>>2] = 0;
 return;
}
function _mpg123_free_string($sb) {
 $sb = $sb|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$sb>>2]|0;
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  _free($0);
 }
 HEAP32[$sb>>2] = 0;
 $2 = ((($sb)) + 4|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($sb)) + 8|0);
 HEAP32[$3>>2] = 0;
 return;
}
function _store_id3_text($sb,$source,$source_size,$noquiet,$notranslate) {
 $sb = $sb|0;
 $source = $source|0;
 $source_size = $source_size|0;
 $noquiet = $noquiet|0;
 $notranslate = $notranslate|0;
 var $$013$i = 0, $$04$i = 0, $$1$i = 0, $$12$i = 0, $$2$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $or$cond = 0, $or$cond$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer10 = 0, $vararg_buffer5 = 0, $vararg_ptr4 = 0, $vararg_ptr8 = 0, $vararg_ptr9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer10 = sp + 32|0;
 $vararg_buffer5 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($source_size|0)==(0);
 if ($0) {
  STACKTOP = sp;return;
 }
 $1 = ($notranslate|0)==(0);
 if (!($1)) {
  $2 = (_mpg123_resize_string($sb,$source_size)|0);
  $3 = ($2|0)==(0);
  if (!($3)) {
   $6 = HEAP32[$sb>>2]|0;
   _memcpy(($6|0),($source|0),($source_size|0))|0;
   $7 = ((($sb)) + 8|0);
   HEAP32[$7>>2] = $source_size;
   STACKTOP = sp;return;
  }
  $4 = ($noquiet|0)==(0);
  if ($4) {
   STACKTOP = sp;return;
  }
  $5 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer>>2] = 203;
  (_fprintf($5,509607,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $8 = HEAP8[$source>>0]|0;
 $9 = ((($source)) + 1|0);
 $10 = (($source_size) + -1)|0;
 $11 = $8&255;
 $12 = ($8&255)>(3);
 if ($12) {
  $13 = ($noquiet|0)==(0);
  if (!($13)) {
   $14 = HEAP32[37544]|0;
   HEAP32[$vararg_buffer1>>2] = 227;
   $vararg_ptr4 = ((($vararg_buffer1)) + 4|0);
   HEAP32[$vararg_ptr4>>2] = $11;
   (_fprintf($14,509670,$vararg_buffer1)|0);
  }
  _mpg123_free_string($sb);
 } else {
  $15 = (36048 + ($11<<2)|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ($8<<24>>24)!=(2);
  $18 = ($16>>>0)<($10>>>0);
  $or$cond$i = $17 & $18;
  L21: do {
   if ($or$cond$i) {
    $$013$i = $10;$$04$i = $9;
    while(1) {
     $19 = HEAP8[$$04$i>>0]|0;
     $20 = ($19<<24>>24)==(0);
     if (!($20)) {
      $$1$i = $$04$i;$$12$i = $$013$i;
      break L21;
     }
     $21 = (($$013$i) + -1)|0;
     $22 = ((($$04$i)) + 1|0);
     $23 = ($21>>>0)>($16>>>0);
     if ($23) {
      $$013$i = $21;$$04$i = $22;
     } else {
      $$1$i = $22;$$12$i = $21;
      break;
     }
    }
   } else {
    $$1$i = $9;$$12$i = $10;
   }
  } while(0);
  $24 = (($$12$i>>>0) % ($16>>>0))&-1;
  $25 = ($24|0)==(0);
  if ($25) {
   $$2$i = $$12$i;
  } else {
   $26 = ($noquiet|0)==(0);
   if (!($26)) {
    $27 = HEAP32[37544]|0;
    HEAP32[$vararg_buffer5>>2] = 244;
    $vararg_ptr8 = ((($vararg_buffer5)) + 4|0);
    HEAP32[$vararg_ptr8>>2] = $$12$i;
    $vararg_ptr9 = ((($vararg_buffer5)) + 8|0);
    HEAP32[$vararg_ptr9>>2] = $11;
    (_fprintf($27,509741,$vararg_buffer5)|0);
   }
   $28 = (($$12$i) - ($24))|0;
   $$2$i = $28;
  }
  $29 = (36064 + ($11<<2)|0);
  $30 = HEAP32[$29>>2]|0;
  FUNCTION_TABLE_viiii[$30 & 8191]($sb,$$1$i,$$2$i,$noquiet);
 }
 $31 = ((($sb)) + 8|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ($32|0)==(0);
 $34 = ($noquiet|0)!=(0);
 $or$cond = $34 & $33;
 if (!($or$cond)) {
  STACKTOP = sp;return;
 }
 $35 = HEAP32[37544]|0;
 HEAP32[$vararg_buffer10>>2] = 215;
 (_fprintf($35,509939,$vararg_buffer10)|0);
 STACKTOP = sp;return;
}
function _mpg123_resize_string($sb,$new) {
 $sb = $sb|0;
 $new = $new|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($new|0)==(0);
 $1 = ((($sb)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 if ($0) {
  $3 = ($2|0)==(0);
  if (!($3)) {
   $4 = HEAP32[$sb>>2]|0;
   $5 = ($4|0)==(0|0);
   if (!($5)) {
    _free($4);
   }
  }
  HEAP32[$sb>>2] = 0;
  HEAP32[$1>>2] = 0;
  $6 = ((($sb)) + 8|0);
  HEAP32[$6>>2] = 0;
  $$1 = 1;
  return ($$1|0);
 }
 $7 = ($2|0)==($new|0);
 if ($7) {
  $$1 = 1;
  return ($$1|0);
 }
 $8 = HEAP32[$sb>>2]|0;
 $9 = (_INT123_safe_realloc($8,$new)|0);
 $10 = ($9|0)==(0|0);
 if ($10) {
  $$1 = 0;
  return ($$1|0);
 }
 HEAP32[$sb>>2] = $9;
 HEAP32[$1>>2] = $new;
 $$1 = 1;
 return ($$1|0);
}
function _convert_latin1($sb,$s,$l,$noquiet) {
 $sb = $sb|0;
 $s = $s|0;
 $l = $l|0;
 $noquiet = $noquiet|0;
 var $$length$0 = 0, $$lobit = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $exitcond8 = 0, $i$04 = 0, $i$11 = 0, $length$0$lcssa = 0, $length$03 = 0, $p$02 = 0, $p$1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ($l|0)==(0);
 if ($0) {
  $length$0$lcssa = 0;
 } else {
  $i$04 = 0;$length$03 = $l;
  while(1) {
   $1 = (($s) + ($i$04)|0);
   $2 = HEAP8[$1>>0]|0;
   $$lobit = ($2&255) >>> 7;
   $3 = $$lobit&255;
   $$length$0 = (($3) + ($length$03))|0;
   $4 = (($i$04) + 1)|0;
   $exitcond8 = ($4|0)==($l|0);
   if ($exitcond8) {
    $length$0$lcssa = $$length$0;
    break;
   } else {
    $i$04 = $4;$length$03 = $$length$0;
   }
  }
 }
 $5 = (($length$0$lcssa) + 1)|0;
 $6 = (_mpg123_resize_string($sb,$5)|0);
 $7 = ($6|0)==(0);
 if ($7) {
  _mpg123_free_string($sb);
  return;
 }
 $8 = HEAP32[$sb>>2]|0;
 if ($0) {
  $26 = $8;
 } else {
  $i$11 = 0;$p$02 = $8;
  while(1) {
   $9 = (($s) + ($i$11)|0);
   $10 = HEAP8[$9>>0]|0;
   $11 = ($10<<24>>24)>(-1);
   if ($11) {
    HEAP8[$p$02>>0] = $10;
    $12 = ((($p$02)) + 1|0);
    $p$1 = $12;
   } else {
    $13 = $10&255;
    $14 = $13 >>> 6;
    $15 = $14 | 192;
    $16 = $15&255;
    HEAP8[$p$02>>0] = $16;
    $17 = HEAP8[$9>>0]|0;
    $18 = $17&255;
    $19 = $18 & 63;
    $20 = $19 | 128;
    $21 = $20&255;
    $22 = ((($p$02)) + 1|0);
    HEAP8[$22>>0] = $21;
    $23 = ((($p$02)) + 2|0);
    $p$1 = $23;
   }
   $24 = (($i$11) + 1)|0;
   $exitcond = ($24|0)==($l|0);
   if ($exitcond) {
    break;
   } else {
    $i$11 = $24;$p$02 = $p$1;
   }
  }
  $$pre = HEAP32[$sb>>2]|0;
  $26 = $$pre;
 }
 $25 = (($26) + ($length$0$lcssa)|0);
 HEAP8[$25>>0] = 0;
 $27 = ((($sb)) + 8|0);
 HEAP32[$27>>2] = $5;
 return;
}
function _convert_utf16bom($sb,$s,$l,$noquiet) {
 $sb = $sb|0;
 $s = $s|0;
 $l = $l|0;
 $noquiet = $noquiet|0;
 var $$ = 0, $$2 = 0, $$3 = 0, $$lcssa = 0, $$masked = 0, $$pn = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$028 = 0, $i$028$lcssa = 0, $i$3$ph = 0, $i$426 = 0, $i$514 = 0, $length$023 = 0, $length$027 = 0, $length$027$lcssa = 0, $length$3$ph = 0, $n$4 = 0, $p$025 = 0, $p$1 = 0, $vararg_buffer = 0;
 var $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 16|0;
 $1 = sp + 12|0;
 HEAP32[$0>>2] = $s;
 HEAP32[$1>>2] = $l;
 $2 = (_check_bom($0,$1)|0);
 $3 = ($2|0)==(-1);
 $$ = $3&1;
 $$2 = $$ ^ 1;
 $4 = HEAP32[$1>>2]|0;
 $5 = $4 & -2;
 $6 = ($5|0)==(0);
 L1: do {
  if ($6) {
   $length$023 = 0;$n$4 = 0;
  } else {
   $7 = HEAP32[$0>>2]|0;
   $8 = $3 ? 3 : 2;
   $i$028 = 0;$length$027 = 0;
   while(1) {
    $9 = (($i$028) + ($$))|0;
    $10 = (($7) + ($9)|0);
    $11 = HEAP8[$10>>0]|0;
    $12 = $11&255;
    $13 = $12 << 8;
    $14 = (($i$028) + ($$2))|0;
    $15 = (($7) + ($14)|0);
    $16 = HEAP8[$15>>0]|0;
    $17 = $16&255;
    $18 = $13 | $17;
    $19 = $13 & 55296;
    $20 = ($19|0)==(55296);
    if ($20) {
     $21 = (($i$028) + 3)|0;
     $22 = ($21>>>0)<($4>>>0);
     if (!($22)) {
      $$lcssa = $18;$i$028$lcssa = $i$028;$length$027$lcssa = $length$027;
      break;
     }
     $23 = (($8) + ($i$028))|0;
     $24 = (($7) + ($23)|0);
     $25 = HEAP8[$24>>0]|0;
     $26 = $25 & -36;
     $27 = ($26<<24>>24)==(-36);
     if (!($27)) {
      $$lcssa = $18;$i$028$lcssa = $i$028;$length$027$lcssa = $length$027;
      break;
     }
     $30 = (($i$028) + 2)|0;
     $$pn = 4;$i$3$ph = $30;
    } else {
     $31 = ($18>>>0)<(128);
     $32 = ($18>>>0)<(2048);
     $$3 = $32 ? 2 : 3;
     $33 = $31 ? 1 : $$3;
     $$pn = $33;$i$3$ph = $i$028;
    }
    $length$3$ph = (($$pn) + ($length$027))|0;
    $34 = (($i$3$ph) + 2)|0;
    $35 = ($34>>>0)<($5>>>0);
    if ($35) {
     $i$028 = $34;$length$027 = $length$3$ph;
    } else {
     $length$023 = $length$3$ph;$n$4 = $5;
     break L1;
    }
   }
   $28 = ($noquiet|0)==(0);
   if ($28) {
    $length$023 = $length$027$lcssa;$n$4 = $i$028$lcssa;
   } else {
    $29 = HEAP32[37544]|0;
    HEAP32[$vararg_buffer>>2] = 941;
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
    HEAP32[$vararg_ptr1>>2] = $i$028$lcssa;
    $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
    HEAP32[$vararg_ptr2>>2] = $$lcssa;
    (_fprintf($29,509873,$vararg_buffer)|0);
    $length$023 = $length$027$lcssa;$n$4 = $i$028$lcssa;
   }
  }
 } while(0);
 $36 = (($length$023) + 1)|0;
 $37 = (_mpg123_resize_string($sb,$36)|0);
 $38 = ($37|0)==(0);
 if ($38) {
  _mpg123_free_string($sb);
  STACKTOP = sp;return;
 }
 $39 = HEAP32[$sb>>2]|0;
 $40 = ($n$4|0)==(0);
 if ($40) {
  $117 = $39;
 } else {
  $i$426 = 0;$p$025 = $39;
  while(1) {
   $41 = (($i$426) + ($$))|0;
   $42 = HEAP32[$0>>2]|0;
   $43 = (($42) + ($41)|0);
   $44 = HEAP8[$43>>0]|0;
   $45 = $44&255;
   $46 = $45 << 8;
   $47 = (($i$426) + ($$2))|0;
   $48 = (($42) + ($47)|0);
   $49 = HEAP8[$48>>0]|0;
   $50 = $49&255;
   $51 = $46 | $50;
   $52 = $46 & 55296;
   $53 = ($52|0)==(55296);
   do {
    if ($53) {
     $78 = (($i$426) + 2)|0;
     $79 = (($78) + ($$))|0;
     $80 = (($42) + ($79)|0);
     $81 = HEAP8[$80>>0]|0;
     $82 = $81&255;
     $83 = $82 << 8;
     $84 = (($78) + ($$2))|0;
     $85 = (($42) + ($84)|0);
     $86 = HEAP8[$85>>0]|0;
     $87 = $86&255;
     $88 = $51 << 10;
     $89 = $88 & 1047552;
     $$masked = $83 & 768;
     $90 = $87 | $89;
     $91 = $90 | $$masked;
     $92 = (($91) + 65536)|0;
     $93 = $92 >>> 18;
     $94 = $93 | 240;
     $95 = $94&255;
     $96 = ((($p$025)) + 1|0);
     HEAP8[$p$025>>0] = $95;
     $97 = $92 >>> 12;
     $98 = $97 & 63;
     $99 = $98 | 128;
     $100 = $99&255;
     $101 = ((($p$025)) + 2|0);
     HEAP8[$96>>0] = $100;
     $102 = $92 >>> 6;
     $103 = $102 & 63;
     $104 = $103 | 128;
     $105 = $104&255;
     $106 = ((($p$025)) + 3|0);
     HEAP8[$101>>0] = $105;
     $107 = $92 & 63;
     $108 = $107 | 128;
     $109 = $108&255;
     $110 = ((($p$025)) + 4|0);
     HEAP8[$106>>0] = $109;
     $i$514 = $78;$p$1 = $110;
    } else {
     $54 = ($51>>>0)<(128);
     if ($54) {
      $55 = ((($p$025)) + 1|0);
      HEAP8[$p$025>>0] = $49;
      $i$514 = $i$426;$p$1 = $55;
      break;
     }
     $56 = ($51>>>0)<(2048);
     if ($56) {
      $57 = $51 >>> 6;
      $58 = $57 | 192;
      $59 = $58&255;
      $60 = ((($p$025)) + 1|0);
      HEAP8[$p$025>>0] = $59;
      $61 = $50 & 63;
      $62 = $61 | 128;
      $63 = $62&255;
      $64 = ((($p$025)) + 2|0);
      HEAP8[$60>>0] = $63;
      $i$514 = $i$426;$p$1 = $64;
      break;
     } else {
      $65 = $45 >>> 4;
      $66 = $65 | 224;
      $67 = $66&255;
      $68 = ((($p$025)) + 1|0);
      HEAP8[$p$025>>0] = $67;
      $69 = $51 >>> 6;
      $70 = $69 & 63;
      $71 = $70 | 128;
      $72 = $71&255;
      $73 = ((($p$025)) + 2|0);
      HEAP8[$68>>0] = $72;
      $74 = $50 & 63;
      $75 = $74 | 128;
      $76 = $75&255;
      $77 = ((($p$025)) + 3|0);
      HEAP8[$73>>0] = $76;
      $i$514 = $i$426;$p$1 = $77;
      break;
     }
    }
   } while(0);
   $111 = (($i$514) + 2)|0;
   $112 = ($111>>>0)<($n$4>>>0);
   if ($112) {
    $i$426 = $111;$p$025 = $p$1;
   } else {
    break;
   }
  }
  $$pre = HEAP32[$sb>>2]|0;
  $117 = $$pre;
 }
 $113 = ((($sb)) + 4|0);
 $114 = HEAP32[$113>>2]|0;
 $115 = (($114) + -1)|0;
 $116 = (($117) + ($115)|0);
 HEAP8[$116>>0] = 0;
 $118 = HEAP32[$113>>2]|0;
 $119 = ((($sb)) + 8|0);
 HEAP32[$119>>2] = $118;
 STACKTOP = sp;return;
}
function _check_bom($source,$len) {
 $source = $source|0;
 $len = $len|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $this_bom$1$ = 0, $this_bom$11 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$len>>2]|0;
 $1 = ($0>>>0)<(2);
 if ($1) {
  return 0;
 }
 $2 = HEAP32[$source>>2]|0;
 $3 = HEAP8[$2>>0]|0;
 switch ($3<<24>>24) {
 case -1:  {
  $7 = ((($2)) + 1|0);
  $8 = HEAP8[$7>>0]|0;
  $9 = ($8<<24>>24)==(-2);
  $$ = $9 << 31 >> 31;
  if ($9) {
   $this_bom$11 = $$;
  } else {
   return 0;
  }
  break;
 }
 case -2:  {
  $4 = ((($2)) + 1|0);
  $5 = HEAP8[$4>>0]|0;
  $6 = ($5<<24>>24)==(-1);
  if ($6) {
   $this_bom$11 = 1;
  } else {
   return 0;
  }
  break;
 }
 default: {
  return 0;
 }
 }
 $10 = ((($2)) + 2|0);
 HEAP32[$source>>2] = $10;
 $11 = (($0) + -2)|0;
 HEAP32[$len>>2] = $11;
 $12 = (_check_bom($source,$len)|0);
 $13 = ($12|0)==(0);
 $this_bom$1$ = $13 ? $this_bom$11 : $12;
 return ($this_bom$1$|0);
}
function _convert_utf8($sb,$source,$len,$noquiet) {
 $sb = $sb|0;
 $source = $source|0;
 $len = $len|0;
 $noquiet = $noquiet|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($len) + 1)|0;
 $1 = (_mpg123_resize_string($sb,$0)|0);
 $2 = ($1|0)==(0);
 if ($2) {
  _mpg123_free_string($sb);
  return;
 } else {
  $3 = HEAP32[$sb>>2]|0;
  _memcpy(($3|0),($source|0),($len|0))|0;
  $4 = HEAP32[$sb>>2]|0;
  $5 = (($4) + ($len)|0);
  HEAP8[$5>>0] = 0;
  $6 = ((($sb)) + 8|0);
  HEAP32[$6>>2] = $0;
  return;
 }
}
function _decode_header($fr,$newhead,$freeformat_count) {
 $fr = $fr|0;
 $newhead = $newhead|0;
 $freeformat_count = $freeformat_count|0;
 var $$0$i = 0, $$0$i2 = 0, $$1 = 0, $$lcssa40 = 0, $$lcssa41 = 0, $$lobit = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0;
 var $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $head$i = 0, $i$07$i = 0;
 var $i$07$i$lcssa36 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer10 = 0, $vararg_buffer14 = 0, $vararg_buffer4 = 0, $vararg_buffer7 = 0, $vararg_ptr13 = 0, $vararg_ptr17 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer14 = sp + 40|0;
 $vararg_buffer10 = sp + 32|0;
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $head$i = sp + 48|0;
 $0 = $newhead >>> 19;
 $1 = $0 & 2;
 $2 = ($1|0)==(0);
 $3 = $0 & 1;
 $4 = $3 ^ 1;
 $5 = $2 ? 1 : $4;
 $$lobit = $1 >>> 1;
 $6 = $$lobit ^ 1;
 $7 = ((($fr)) + 19672|0);
 HEAP32[$7>>2] = $5;
 $8 = ((($fr)) + 19676|0);
 HEAP32[$8>>2] = $6;
 $9 = ((($fr)) + 28064|0);
 $10 = ((($fr)) + 28068|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 & 128;
 $13 = ($12|0)==(0);
 if ($13) {
  $14 = ((($fr)) + 27820|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ($15|0)==(0);
  if ($16) {
   label = 4;
  } else {
   $17 = $15 >>> 19;
   $18 = $17 ^ $0;
   $19 = $18 & 3;
   $20 = ($19|0)==(0);
   if (!($20)) {
    label = 4;
   }
  }
 } else {
  label = 4;
 }
 do {
  if ((label|0) == 4) {
   $21 = $newhead >>> 17;
   $22 = $21 & 3;
   $23 = (4 - ($22))|0;
   $24 = ((($fr)) + 19688|0);
   HEAP32[$24>>2] = $23;
   $25 = ($$lobit|0)==(1);
   $26 = $newhead >>> 10;
   $27 = $26 & 3;
   if ($25) {
    $30 = (0 - ($5))|0;
    $31 = $30 & 3;
    $32 = (($31) + ($27))|0;
    $33 = ((($fr)) + 19704|0);
    HEAP32[$33>>2] = $32;
    break;
   } else {
    $28 = (($27) + 6)|0;
    $29 = ((($fr)) + 19704|0);
    HEAP32[$29>>2] = $28;
    break;
   }
  }
 } while(0);
 $34 = $newhead >>> 16;
 $35 = $34 & 1;
 $36 = $35 ^ 1;
 $37 = ((($fr)) + 19696|0);
 HEAP32[$37>>2] = $36;
 $38 = $newhead & 61440;
 $39 = $38 >>> 12;
 $40 = ((($fr)) + 19700|0);
 HEAP32[$40>>2] = $39;
 $41 = $newhead >>> 9;
 $42 = $41 & 1;
 $43 = ((($fr)) + 19708|0);
 HEAP32[$43>>2] = $42;
 $44 = $newhead >>> 8;
 $45 = $44 & 1;
 $46 = ((($fr)) + 19712|0);
 HEAP32[$46>>2] = $45;
 $47 = $newhead >>> 6;
 $48 = $47 & 3;
 $49 = ((($fr)) + 19716|0);
 HEAP32[$49>>2] = $48;
 $50 = $newhead >>> 4;
 $51 = $50 & 3;
 $52 = ((($fr)) + 19720|0);
 HEAP32[$52>>2] = $51;
 $53 = $newhead >>> 3;
 $54 = $53 & 1;
 $55 = ((($fr)) + 19724|0);
 HEAP32[$55>>2] = $54;
 $56 = $newhead >>> 2;
 $57 = $56 & 1;
 $58 = ((($fr)) + 19728|0);
 HEAP32[$58>>2] = $57;
 $59 = $newhead & 3;
 $60 = ((($fr)) + 19732|0);
 HEAP32[$60>>2] = $59;
 $61 = ($38|0)==(0);
 $62 = $61&1;
 $63 = ((($fr)) + 19776|0);
 HEAP32[$63>>2] = $62;
 $64 = ($48|0)==(3);
 $65 = $64 ? 1 : 2;
 $66 = ((($fr)) + 19652|0);
 HEAP32[$66>>2] = $65;
 L10: do {
  if ($61) {
   $67 = ((($fr)) + 19780|0);
   $68 = HEAP32[$67>>2]|0;
   $69 = ($68|0)<(0);
   if (!($69)) {
    $129 = (($42) + ($68))|0;
    $130 = ((($fr)) + 19736|0);
    HEAP32[$130>>2] = $129;
    break;
   }
   $70 = HEAP32[$freeformat_count>>2]|0;
   $71 = (($70) + 1)|0;
   HEAP32[$freeformat_count>>2] = $71;
   $72 = ($70|0)>(4);
   if ($72) {
    $73 = $11 & 32;
    $74 = ($73|0)==(0);
    if (!($74)) {
     $$1 = 0;
     STACKTOP = sp;return ($$1|0);
    }
    $75 = HEAP32[$9>>2]|0;
    $76 = ($75|0)>(2);
    if (!($76)) {
     $$1 = 0;
     STACKTOP = sp;return ($$1|0);
    }
    $77 = HEAP32[37544]|0;
    HEAP32[$vararg_buffer>>2] = 728;
    (_fprintf($77,511083,$vararg_buffer)|0);
    $$1 = 0;
    STACKTOP = sp;return ($$1|0);
   }
   $78 = ((($fr)) + 27976|0);
   $79 = HEAP32[$78>>2]|0;
   $80 = $79 & 12;
   $81 = ($80|0)==(0);
   L25: do {
    if ($81) {
     $82 = $11 & 32;
     $83 = ($82|0)==(0);
     if ($83) {
      $84 = HEAP32[37544]|0;
      HEAP32[$vararg_buffer1>>2] = 634;
      (_fprintf($84,511179,$vararg_buffer1)|0);
      label = 23;
     } else {
      label = 23;
     }
    } else {
     $85 = ((($fr)) + 27956|0);
     $86 = HEAP32[$85>>2]|0;
     $87 = ((($86)) + 12|0);
     $88 = HEAP32[$87>>2]|0;
     $89 = (FUNCTION_TABLE_iii[$88 & 8191]($fr,$head$i)|0);
     $90 = ($89|0)<(1);
     L29: do {
      if ($90) {
       $$0$i = $89;
       label = 24;
      } else {
       $91 = HEAP32[$85>>2]|0;
       $92 = ((($fr)) + 27820|0);
       $94 = $91;$i$07$i = 4;
       while(1) {
        $93 = ((($94)) + 16|0);
        $95 = HEAP32[$93>>2]|0;
        $96 = (FUNCTION_TABLE_iii[$95 & 8191]($fr,$head$i)|0);
        $97 = ($96|0)<(1);
        if ($97) {
         $$0$i = $96;
         label = 24;
         break L29;
        }
        $98 = HEAP32[$head$i>>2]|0;
        $99 = HEAP32[$92>>2]|0;
        $100 = $99 ^ $98;
        $101 = $100 & -66320;
        $102 = ($101|0)==(0);
        if ($102) {
         $i$07$i$lcssa36 = $i$07$i;
         label = 20;
         break;
        }
        $109 = (($i$07$i) + 1)|0;
        $110 = ($109|0)<(3460);
        $111 = HEAP32[$85>>2]|0;
        if ($110) {
         $94 = $111;$i$07$i = $109;
        } else {
         $$lcssa40 = $109;$$lcssa41 = $111;
         label = 22;
         break;
        }
       }
       if ((label|0) == 20) {
        $103 = HEAP32[$85>>2]|0;
        $104 = ((($103)) + 28|0);
        $105 = HEAP32[$104>>2]|0;
        $106 = (($i$07$i$lcssa36) + 1)|0;
        (FUNCTION_TABLE_iii[$105 & 8191]($fr,$106)|0);
        $107 = (($i$07$i$lcssa36) + -3)|0;
        $108 = ((($fr)) + 19736|0);
        HEAP32[$108>>2] = $107;
        $116 = $107;
        break;
       }
       else if ((label|0) == 22) {
        $112 = ((($$lcssa41)) + 28|0);
        $113 = HEAP32[$112>>2]|0;
        (FUNCTION_TABLE_iii[$113 & 8191]($fr,$$lcssa40)|0);
        label = 23;
        break L25;
       }
      }
     } while(0);
     L38: do {
      if ((label|0) == 24) {
       switch ($$0$i|0) {
       case 1:  {
        $$phi$trans$insert = ((($fr)) + 19736|0);
        $$pre = HEAP32[$$phi$trans$insert>>2]|0;
        $116 = $$pre;
        break L38;
        break;
       }
       case -10:  {
        $$1 = -10;
        STACKTOP = sp;return ($$1|0);
        break;
       }
       default: {
        $$0$i2 = $$0$i;
        break L25;
       }
       }
      }
     } while(0);
     $114 = HEAP32[$43>>2]|0;
     $115 = (($116) - ($114))|0;
     HEAP32[$67>>2] = $115;
     $117 = HEAP32[$10>>2]|0;
     $118 = $117 & 32;
     $119 = ($118|0)==(0);
     if (!($119)) {
      break L10;
     }
     $120 = HEAP32[$9>>2]|0;
     $121 = ($120|0)>(1);
     if (!($121)) {
      break L10;
     }
     $122 = HEAP32[37544]|0;
     HEAP32[$vararg_buffer4>>2] = $115;
     (_fprintf($122,511281,$vararg_buffer4)|0);
     break L10;
    }
   } while(0);
   if ((label|0) == 23) {
    $$0$i2 = 0;
   }
   $123 = HEAP32[$10>>2]|0;
   $124 = $123 & 32;
   $125 = ($124|0)==(0);
   if (!($125)) {
    $$1 = $$0$i2;
    STACKTOP = sp;return ($$1|0);
   }
   $126 = HEAP32[$9>>2]|0;
   $127 = ($126|0)>(2);
   if (!($127)) {
    $$1 = $$0$i2;
    STACKTOP = sp;return ($$1|0);
   }
   $128 = HEAP32[37544]|0;
   HEAP32[$vararg_buffer7>>2] = 743;
   (_fprintf($128,511315,$vararg_buffer7)|0);
   $$1 = $$0$i2;
   STACKTOP = sp;return ($$1|0);
  }
 } while(0);
 $131 = ((($fr)) + 19688|0);
 $132 = HEAP32[$131>>2]|0;
 switch ($132|0) {
 case 1:  {
  $133 = ((($fr)) + 19692|0);
  HEAP32[$133>>2] = 4264;
  $134 = HEAP32[$63>>2]|0;
  $135 = ($134|0)==(0);
  if ($135) {
   $136 = HEAP32[$40>>2]|0;
   $137 = HEAP32[$7>>2]|0;
   $138 = ((35332 + (($137*192)|0)|0) + ($136<<2)|0);
   $139 = HEAP32[$138>>2]|0;
   $140 = ($139*12000)|0;
   $141 = ((($fr)) + 19736|0);
   $142 = ((($fr)) + 19704|0);
   $143 = HEAP32[$142>>2]|0;
   $144 = (35716 + ($143<<2)|0);
   $145 = HEAP32[$144>>2]|0;
   $146 = (($140|0) / ($145|0))&-1;
   $147 = HEAP32[$43>>2]|0;
   $148 = (($147) + ($146))|0;
   $149 = $148 << 2;
   $150 = (($149) + -4)|0;
   HEAP32[$141>>2] = $150;
  }
  break;
 }
 case 2:  {
  $151 = ((($fr)) + 19692|0);
  HEAP32[$151>>2] = 4265;
  $152 = HEAP32[$63>>2]|0;
  $153 = ($152|0)==(0);
  if ($153) {
   $154 = HEAP32[$40>>2]|0;
   $155 = HEAP32[$7>>2]|0;
   $156 = ((((35332 + (($155*192)|0)|0)) + 64|0) + ($154<<2)|0);
   $157 = HEAP32[$156>>2]|0;
   $158 = ($157*144000)|0;
   $159 = ((($fr)) + 19736|0);
   $160 = ((($fr)) + 19704|0);
   $161 = HEAP32[$160>>2]|0;
   $162 = (35716 + ($161<<2)|0);
   $163 = HEAP32[$162>>2]|0;
   $164 = (($158|0) / ($163|0))&-1;
   $165 = HEAP32[$43>>2]|0;
   $166 = (($164) + -4)|0;
   $167 = (($166) + ($165))|0;
   HEAP32[$159>>2] = $167;
  }
  break;
 }
 case 3:  {
  $168 = ((($fr)) + 19692|0);
  HEAP32[$168>>2] = 4266;
  $169 = HEAP32[$7>>2]|0;
  $170 = ($169|0)==(0);
  $171 = HEAP32[$66>>2]|0;
  $172 = ($171|0)==(1);
  $173 = $172 ? 17 : 32;
  $174 = $172 ? 9 : 17;
  $175 = $170 ? $173 : $174;
  $176 = ((($fr)) + 19864|0);
  HEAP32[$176>>2] = $175;
  $177 = HEAP32[$37>>2]|0;
  $178 = ($177|0)==(0);
  if (!($178)) {
   $179 = $175 | 2;
   HEAP32[$176>>2] = $179;
  }
  $180 = HEAP32[$63>>2]|0;
  $181 = ($180|0)==(0);
  if ($181) {
   $182 = HEAP32[$40>>2]|0;
   $183 = ((((35332 + (($169*192)|0)|0)) + 128|0) + ($182<<2)|0);
   $184 = HEAP32[$183>>2]|0;
   $185 = ($184*144000)|0;
   $186 = ((($fr)) + 19736|0);
   $187 = ((($fr)) + 19704|0);
   $188 = HEAP32[$187>>2]|0;
   $189 = (35716 + ($188<<2)|0);
   $190 = HEAP32[$189>>2]|0;
   $191 = $190 << $169;
   $192 = (($185|0) / ($191|0))&-1;
   $193 = HEAP32[$43>>2]|0;
   $194 = (($192) + -4)|0;
   $195 = (($194) + ($193))|0;
   HEAP32[$186>>2] = $195;
  }
  break;
 }
 default: {
  $196 = HEAP32[$10>>2]|0;
  $197 = $196 & 32;
  $198 = ($197|0)==(0);
  if (!($198)) {
   $$1 = 0;
   STACKTOP = sp;return ($$1|0);
  }
  $199 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer10>>2] = 800;
  $vararg_ptr13 = ((($vararg_buffer10)) + 4|0);
  HEAP32[$vararg_ptr13>>2] = $132;
  (_fprintf($199,511936,$vararg_buffer10)|0);
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 }
 $200 = ((($fr)) + 19736|0);
 $201 = HEAP32[$200>>2]|0;
 $202 = ($201|0)>(3456);
 if (!($202)) {
  $$1 = 1;
  STACKTOP = sp;return ($$1|0);
 }
 $203 = HEAP32[$10>>2]|0;
 $204 = $203 & 32;
 $205 = ($204|0)==(0);
 if (!($205)) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $206 = HEAP32[37544]|0;
 $207 = (($201) + 4)|0;
 $208 = HEAP32[$43>>2]|0;
 $209 = (($207) - ($208))|0;
 HEAP32[$vararg_buffer14>>2] = 806;
 $vararg_ptr17 = ((($vararg_buffer14)) + 4|0);
 HEAP32[$vararg_ptr17>>2] = $209;
 (_fprintf($206,512000,$vararg_buffer14)|0);
 $$1 = 0;
 STACKTOP = sp;return ($$1|0);
}
function _INT123_do_layer1($fr) {
 $fr = $fr|0;
 var $$ = 0, $$lcssa = 0, $$lcssa40 = 0, $$pn = 0, $$pre = 0, $$pre$i = 0, $$pre$phi22Z2D = 0, $$pre$phi24Z2D = 0, $$pre$phi26Z2D = 0, $$pre34$i = 0, $$pre35$i = 0, $$pre35$i$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0;
 var $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0;
 var $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0;
 var $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0;
 var $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0;
 var $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0;
 var $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0;
 var $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0;
 var $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0;
 var $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0.0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0.0, $354 = 0.0, $355 = 0, $356 = 0, $357 = 0;
 var $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0.0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0.0, $369 = 0.0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0;
 var $376 = 0, $377 = 0, $378 = 0, $379 = 0.0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0.0, $385 = 0.0, $386 = 0, $387 = 0, $388 = 0, $389 = 0.0, $39 = 0, $390 = 0.0, $391 = 0, $392 = 0, $393 = 0;
 var $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0;
 var $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0;
 var $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0.0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0.0, $439 = 0.0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0;
 var $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $ba$0$lcssa36$i = 0, $ba$018$i = 0, $ba$031$i = 0, $ba$115$i = 0, $ba$127$i = 0, $ba$2$lcssa$i = 0, $ba$2$lcssa$i6 = 0, $ba$211$i = 0, $ba$216$i = 0, $ba$38$i = 0, $ba$39$i = 0;
 var $ba$424$i = 0, $ba$446$i = 0, $ba$522$i = 0, $ba$541$i = 0, $balloc = 0, $clip$016 = 0, $clip$1 = 0, $clip$1$lcssa = 0, $exitcond = 0, $exitcond$i = 0, $exitcond$i11 = 0, $exitcond28$i = 0, $exitcond29$i = 0, $exitcond31$i = 0, $exitcond32$i = 0, $exitcond33$i = 0, $exitcond50$i = 0, $exitcond53$i = 0, $exitcond54$i = 0, $exitcond56$i = 0;
 var $exitcond58$i = 0, $exitcond59$i = 0, $f0$0$lcssa$i = 0, $f0$018$i = 0, $f0$1$i = 0, $f0$211$i = 0, $f0$3$i = 0, $f01$044$i = 0, $f01$1$i = 0, $f1$0$lcssa$i = 0, $f1$019$i = 0, $f1$1$i = 0, $f1$212$i = 0, $f1$3$i = 0, $i$015 = 0, $i$017$i = 0, $i$032$i = 0, $i$114$i = 0, $i$128$i = 0, $i$210$i = 0;
 var $i$220$i = 0, $i$313$i = 0, $i$37$i = 0, $i$47$i = 0, $i$547$i = 0, $i$643$i = 0, $i1$025$i = 0, $i1$123$i = 0, $or$cond = 0, $sample$030$i = 0, $sample$1$i = 0, $sample$1045$i = 0, $sample$11$i = 0, $sample$1240$i = 0, $sample$13$i = 0, $sample$2$i = 0, $sample$2$i$lcssa = 0, $sample$326$i = 0, $sample$4$i = 0, $sample$5$lcssa$i = 0;
 var $sample$515$i = 0, $sample$6$i = 0, $sample$7$i = 0, $sample$7$i$lcssa = 0, $sample$88$i = 0, $sample$9$i = 0, $sca$0$lcssa$i = 0, $sca$0$lcssa$i5 = 0, $sca$017$i = 0, $sca$09$i = 0, $sca$1$i = 0, $sca$1$i8 = 0, $sca$2$i = 0, $sca$2$i$lcssa = 0, $sca$2$i9 = 0, $sca$2$i9$lcssa = 0, $sca$310$i = 0, $sca$36$i = 0, $sca$4$i = 0, $sca$4$i10 = 0;
 var $sca$521$i = 0, $sca$542$i = 0, $sca$6$i = 0, $sca$6$i13 = 0, $scale_index = 0, $scevgep$i = 0, $scevgep$i3 = 0, $scevgep30$i = 0, $scevgep51$i = 0, $scevgep52$i = 0, $scevgep55$i = 0, $scevgep64$i = 0, $smpb$i = 0, $storemerge$i = 0.0, $storemerge1$i = 0.0, $storemerge2$i = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 768|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $smpb$i = sp + 512|0;
 $balloc = sp + 256|0;
 $scale_index = sp;
 $0 = ((($fr)) + 19652|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 28592|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($fr)) + 19660|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($fr)) + 19716|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(1);
 if ($8) {
  $9 = ((($fr)) + 19720|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = $10 << 2;
  $12 = (($11) + 4)|0;
  $14 = $12;
 } else {
  $14 = 32;
 }
 $13 = ((($fr)) + 19656|0);
 HEAP32[$13>>2] = $14;
 $15 = ($1|0)==(1);
 $16 = ($5|0)==(3);
 $or$cond = $15 | $16;
 $$ = $or$cond ? 0 : $5;
 $17 = ($1|0)==(2);
 if ($17) {
  $20 = ($14|0)>(0);
  $21 = ((($fr)) + 19788|0);
  if ($20) {
   $22 = ((($fr)) + 19784|0);
   $$pre34$i = HEAP32[$21>>2]|0;
   $26 = $$pre34$i;$ba$018$i = $balloc;$i$017$i = 0;
   while(1) {
    $25 = HEAP8[$26>>0]|0;
    $27 = $25&255;
    $28 = $27 << 8;
    $29 = ((($26)) + 1|0);
    $30 = HEAP8[$29>>0]|0;
    $31 = $30&255;
    $32 = $28 | $31;
    $33 = $32 << 8;
    $34 = ((($26)) + 2|0);
    $35 = HEAP8[$34>>0]|0;
    $36 = $35&255;
    $37 = $33 | $36;
    $38 = HEAP32[$22>>2]|0;
    $39 = $37 << $38;
    $40 = (($38) + 4)|0;
    $41 = $39 >>> 20;
    $42 = $41 & 15;
    $43 = $40 >> 3;
    $44 = (($26) + ($43)|0);
    HEAP32[$21>>2] = $44;
    $45 = $40 & 7;
    HEAP32[$22>>2] = $45;
    $46 = ((($ba$018$i)) + 4|0);
    HEAP32[$ba$018$i>>2] = $42;
    $47 = HEAP8[$44>>0]|0;
    $48 = $47&255;
    $49 = $48 << 8;
    $50 = ((($44)) + 1|0);
    $51 = HEAP8[$50>>0]|0;
    $52 = $51&255;
    $53 = $49 | $52;
    $54 = $53 << 8;
    $55 = ((($44)) + 2|0);
    $56 = HEAP8[$55>>0]|0;
    $57 = $56&255;
    $58 = $54 | $57;
    $59 = HEAP32[$22>>2]|0;
    $60 = $58 << $59;
    $61 = (($59) + 4)|0;
    $62 = $60 >>> 20;
    $63 = $62 & 15;
    $64 = $61 >> 3;
    $65 = (($44) + ($64)|0);
    HEAP32[$21>>2] = $65;
    $66 = $61 & 7;
    HEAP32[$22>>2] = $66;
    $67 = ((($ba$018$i)) + 8|0);
    HEAP32[$46>>2] = $63;
    $68 = (($i$017$i) + 1)|0;
    $exitcond31$i = ($68|0)==($14|0);
    if ($exitcond31$i) {
     $$lcssa40 = $65;
     break;
    } else {
     $26 = $65;$ba$018$i = $67;$i$017$i = $68;
    }
   }
   $23 = $14 << 1;
   $scevgep30$i = (($balloc) + ($23<<2)|0);
   $24 = ($14|0)<(32);
   if ($24) {
    $$pre$phi22Z2D = $22;$$pre35$i = $$lcssa40;$ba$0$lcssa36$i = $scevgep30$i;
    label = 9;
   } else {
    $450 = $$lcssa40;$451 = 0;
   }
  } else {
   $$pre35$i$pre = HEAP32[$21>>2]|0;
   $$pre = ((($fr)) + 19784|0);
   $$pre$phi22Z2D = $$pre;$$pre35$i = $$pre35$i$pre;$ba$0$lcssa36$i = $balloc;
   label = 9;
  }
  if ((label|0) == 9) {
   $72 = $$pre35$i;$ba$115$i = $ba$0$lcssa36$i;$i$114$i = $14;
   while(1) {
    $71 = HEAP8[$72>>0]|0;
    $73 = $71&255;
    $74 = $73 << 8;
    $75 = ((($72)) + 1|0);
    $76 = HEAP8[$75>>0]|0;
    $77 = $76&255;
    $78 = $74 | $77;
    $79 = $78 << 8;
    $80 = ((($72)) + 2|0);
    $81 = HEAP8[$80>>0]|0;
    $82 = $81&255;
    $83 = $79 | $82;
    $84 = HEAP32[$$pre$phi22Z2D>>2]|0;
    $85 = $83 << $84;
    $86 = (($84) + 4)|0;
    $87 = $85 >>> 20;
    $88 = $87 & 15;
    $89 = $86 >> 3;
    $90 = (($72) + ($89)|0);
    HEAP32[$21>>2] = $90;
    $91 = $86 & 7;
    HEAP32[$$pre$phi22Z2D>>2] = $91;
    $92 = ((($ba$115$i)) + 4|0);
    HEAP32[$ba$115$i>>2] = $88;
    $93 = (($i$114$i) + 1)|0;
    $exitcond29$i = ($93|0)==(32);
    if ($exitcond29$i) {
     $450 = $90;$451 = 1;
     break;
    } else {
     $72 = $90;$ba$115$i = $92;$i$114$i = $93;
    }
   }
  }
  if ($20) {
   $69 = ((($fr)) + 19788|0);
   $70 = ((($fr)) + 19784|0);
   $101 = $450;$ba$211$i = $balloc;$i$210$i = 0;$sca$09$i = $scale_index;
   while(1) {
    $97 = ((($ba$211$i)) + 4|0);
    $98 = HEAP32[$ba$211$i>>2]|0;
    $99 = ($98|0)==(0);
    if ($99) {
     $126 = $101;$sca$1$i = $sca$09$i;
    } else {
     $100 = HEAP8[$101>>0]|0;
     $102 = $100&255;
     $103 = $102 << 8;
     $104 = ((($101)) + 1|0);
     $105 = HEAP8[$104>>0]|0;
     $106 = $105&255;
     $107 = $103 | $106;
     $108 = $107 << 8;
     $109 = ((($101)) + 2|0);
     $110 = HEAP8[$109>>0]|0;
     $111 = $110&255;
     $112 = $108 | $111;
     $113 = HEAP32[$70>>2]|0;
     $114 = $112 << $113;
     $115 = (($113) + 6)|0;
     $116 = $114 >>> 18;
     $117 = $116 & 63;
     $118 = $115 >> 3;
     $119 = (($101) + ($118)|0);
     HEAP32[$69>>2] = $119;
     $120 = $115 & 7;
     HEAP32[$70>>2] = $120;
     $121 = ((($sca$09$i)) + 4|0);
     HEAP32[$sca$09$i>>2] = $117;
     $126 = $119;$sca$1$i = $121;
    }
    $122 = ((($ba$211$i)) + 8|0);
    $123 = HEAP32[$97>>2]|0;
    $124 = ($123|0)==(0);
    if ($124) {
     $453 = $126;$sca$2$i = $sca$1$i;
    } else {
     $125 = HEAP8[$126>>0]|0;
     $127 = $125&255;
     $128 = $127 << 8;
     $129 = ((($126)) + 1|0);
     $130 = HEAP8[$129>>0]|0;
     $131 = $130&255;
     $132 = $128 | $131;
     $133 = $132 << 8;
     $134 = ((($126)) + 2|0);
     $135 = HEAP8[$134>>0]|0;
     $136 = $135&255;
     $137 = $133 | $136;
     $138 = HEAP32[$70>>2]|0;
     $139 = $137 << $138;
     $140 = (($138) + 6)|0;
     $141 = $139 >>> 18;
     $142 = $141 & 63;
     $143 = $140 >> 3;
     $144 = (($126) + ($143)|0);
     HEAP32[$69>>2] = $144;
     $145 = $140 & 7;
     HEAP32[$70>>2] = $145;
     $146 = ((($sca$1$i)) + 4|0);
     HEAP32[$sca$1$i>>2] = $142;
     $453 = $144;$sca$2$i = $146;
    }
    $147 = (($i$210$i) + 1)|0;
    $exitcond28$i = ($147|0)==($14|0);
    if ($exitcond28$i) {
     $$lcssa = $453;$sca$2$i$lcssa = $sca$2$i;
     break;
    } else {
     $101 = $453;$ba$211$i = $122;$i$210$i = $147;$sca$09$i = $sca$2$i;
    }
   }
   $94 = $14 << 1;
   $scevgep$i = (($balloc) + ($94<<2)|0);
   $452 = $$lcssa;$ba$2$lcssa$i = $scevgep$i;$sca$0$lcssa$i = $sca$2$i$lcssa;
  } else {
   $452 = $450;$ba$2$lcssa$i = $balloc;$sca$0$lcssa$i = $scale_index;
  }
  $95 = ((($fr)) + 19788|0);
  $96 = ((($fr)) + 19784|0);
  if ($451) {
   $152 = $452;$ba$38$i = $ba$2$lcssa$i;$i$37$i = $14;$sca$36$i = $sca$0$lcssa$i;
   while(1) {
    $148 = ((($ba$38$i)) + 4|0);
    $149 = HEAP32[$ba$38$i>>2]|0;
    $150 = ($149|0)==(0);
    if ($150) {
     $454 = $152;$sca$4$i = $sca$36$i;
    } else {
     $151 = HEAP8[$152>>0]|0;
     $153 = $151&255;
     $154 = $153 << 8;
     $155 = ((($152)) + 1|0);
     $156 = HEAP8[$155>>0]|0;
     $157 = $156&255;
     $158 = $154 | $157;
     $159 = $158 << 8;
     $160 = ((($152)) + 2|0);
     $161 = HEAP8[$160>>0]|0;
     $162 = $161&255;
     $163 = $159 | $162;
     $164 = HEAP32[$96>>2]|0;
     $165 = $163 << $164;
     $166 = (($164) + 6)|0;
     $167 = $165 >>> 18;
     $168 = $167 & 63;
     $169 = $166 >> 3;
     $170 = (($152) + ($169)|0);
     HEAP32[$95>>2] = $170;
     $171 = $166 & 7;
     HEAP32[$96>>2] = $171;
     $172 = ((($sca$36$i)) + 4|0);
     HEAP32[$sca$36$i>>2] = $168;
     $173 = HEAP8[$170>>0]|0;
     $174 = $173&255;
     $175 = $174 << 8;
     $176 = ((($170)) + 1|0);
     $177 = HEAP8[$176>>0]|0;
     $178 = $177&255;
     $179 = $175 | $178;
     $180 = $179 << 8;
     $181 = ((($170)) + 2|0);
     $182 = HEAP8[$181>>0]|0;
     $183 = $182&255;
     $184 = $180 | $183;
     $185 = HEAP32[$96>>2]|0;
     $186 = $184 << $185;
     $187 = (($185) + 6)|0;
     $188 = $186 >>> 18;
     $189 = $188 & 63;
     $190 = $187 >> 3;
     $191 = (($170) + ($190)|0);
     HEAP32[$95>>2] = $191;
     $192 = $187 & 7;
     HEAP32[$96>>2] = $192;
     $193 = ((($sca$36$i)) + 8|0);
     HEAP32[$172>>2] = $189;
     $454 = $191;$sca$4$i = $193;
    }
    $194 = (($i$37$i) + 1)|0;
    $exitcond$i = ($194|0)==(32);
    if ($exitcond$i) {
     $$pre$phi24Z2D = $95;$$pre$phi26Z2D = $96;
     break;
    } else {
     $152 = $454;$ba$38$i = $148;$i$37$i = $194;$sca$36$i = $sca$4$i;
    }
   }
  } else {
   $$pre$phi24Z2D = $95;$$pre$phi26Z2D = $96;
  }
 } else {
  $18 = ((($fr)) + 19788|0);
  $19 = ((($fr)) + 19784|0);
  $$pre$i = HEAP32[$18>>2]|0;
  $196 = $$pre$i;$ba$424$i = $balloc;$i1$025$i = 0;
  while(1) {
   $195 = HEAP8[$196>>0]|0;
   $197 = $195&255;
   $198 = $197 << 8;
   $199 = ((($196)) + 1|0);
   $200 = HEAP8[$199>>0]|0;
   $201 = $200&255;
   $202 = $198 | $201;
   $203 = $202 << 8;
   $204 = ((($196)) + 2|0);
   $205 = HEAP8[$204>>0]|0;
   $206 = $205&255;
   $207 = $203 | $206;
   $208 = HEAP32[$19>>2]|0;
   $209 = $207 << $208;
   $210 = (($208) + 4)|0;
   $211 = $209 >>> 20;
   $212 = $211 & 15;
   $213 = $210 >> 3;
   $214 = (($196) + ($213)|0);
   HEAP32[$18>>2] = $214;
   $215 = $210 & 7;
   HEAP32[$19>>2] = $215;
   $216 = ((($ba$424$i)) + 4|0);
   HEAP32[$ba$424$i>>2] = $212;
   $217 = (($i1$025$i) + 1)|0;
   $exitcond33$i = ($217|0)==(32);
   if ($exitcond33$i) {
    $222 = $214;$ba$522$i = $balloc;$i1$123$i = 0;$sca$521$i = $scale_index;
    break;
   } else {
    $196 = $214;$ba$424$i = $216;$i1$025$i = $217;
   }
  }
  while(1) {
   $218 = ((($ba$522$i)) + 4|0);
   $219 = HEAP32[$ba$522$i>>2]|0;
   $220 = ($219|0)==(0);
   if ($220) {
    $455 = $222;$sca$6$i = $sca$521$i;
   } else {
    $221 = HEAP8[$222>>0]|0;
    $223 = $221&255;
    $224 = $223 << 8;
    $225 = ((($222)) + 1|0);
    $226 = HEAP8[$225>>0]|0;
    $227 = $226&255;
    $228 = $224 | $227;
    $229 = $228 << 8;
    $230 = ((($222)) + 2|0);
    $231 = HEAP8[$230>>0]|0;
    $232 = $231&255;
    $233 = $229 | $232;
    $234 = HEAP32[$19>>2]|0;
    $235 = $233 << $234;
    $236 = (($234) + 6)|0;
    $237 = $235 >>> 18;
    $238 = $237 & 63;
    $239 = $236 >> 3;
    $240 = (($222) + ($239)|0);
    HEAP32[$18>>2] = $240;
    $241 = $236 & 7;
    HEAP32[$19>>2] = $241;
    $242 = ((($sca$521$i)) + 4|0);
    HEAP32[$sca$521$i>>2] = $238;
    $455 = $240;$sca$6$i = $242;
   }
   $243 = (($i1$123$i) + 1)|0;
   $exitcond32$i = ($243|0)==(32);
   if ($exitcond32$i) {
    $$pre$phi24Z2D = $18;$$pre$phi26Z2D = $19;
    break;
   } else {
    $222 = $455;$ba$522$i = $218;$i1$123$i = $243;$sca$521$i = $sca$6$i;
   }
  }
 }
 $244 = ((($3)) + 128|0);
 $245 = ((($fr)) + 19668|0);
 $246 = ($$|0)==(-1);
 $247 = ((($fr)) + 19640|0);
 $248 = ((($fr)) + 19644|0);
 $249 = (($3) + ($$<<7)|0);
 $clip$016 = 0;$i$015 = 0;
 while(1) {
  $250 = HEAP32[$0>>2]|0;
  $251 = ($250|0)==(2);
  if ($251) {
   $252 = HEAP32[$13>>2]|0;
   $253 = ($252|0)>(0);
   if ($253) {
    $ba$031$i = $balloc;$i$032$i = 0;$sample$030$i = $smpb$i;
    while(1) {
     $256 = ((($ba$031$i)) + 4|0);
     $257 = HEAP32[$ba$031$i>>2]|0;
     $258 = ($257|0)==(0);
     if ($258) {
      $sample$1$i = $sample$030$i;
     } else {
      $259 = (($257) + 1)|0;
      $260 = HEAP32[$$pre$phi24Z2D>>2]|0;
      $261 = HEAP8[$260>>0]|0;
      $262 = $261&255;
      $263 = $262 << 8;
      $264 = ((($260)) + 1|0);
      $265 = HEAP8[$264>>0]|0;
      $266 = $265&255;
      $267 = $263 | $266;
      $268 = $267 << 8;
      $269 = ((($260)) + 2|0);
      $270 = HEAP8[$269>>0]|0;
      $271 = $270&255;
      $272 = $268 | $271;
      $273 = HEAP32[$$pre$phi26Z2D>>2]|0;
      $274 = $272 << $273;
      $275 = $274 & 16777215;
      $276 = (($259) + ($273))|0;
      $277 = (23 - ($257))|0;
      $278 = $275 >>> $277;
      $279 = $276 >> 3;
      $280 = (($260) + ($279)|0);
      HEAP32[$$pre$phi24Z2D>>2] = $280;
      $281 = $276 & 7;
      HEAP32[$$pre$phi26Z2D>>2] = $281;
      $282 = ((($sample$030$i)) + 4|0);
      HEAP32[$sample$030$i>>2] = $278;
      $sample$1$i = $282;
     }
     $283 = ((($ba$031$i)) + 8|0);
     $284 = HEAP32[$256>>2]|0;
     $285 = ($284|0)==(0);
     if ($285) {
      $sample$2$i = $sample$1$i;
     } else {
      $286 = (($284) + 1)|0;
      $287 = HEAP32[$$pre$phi24Z2D>>2]|0;
      $288 = HEAP8[$287>>0]|0;
      $289 = $288&255;
      $290 = $289 << 8;
      $291 = ((($287)) + 1|0);
      $292 = HEAP8[$291>>0]|0;
      $293 = $292&255;
      $294 = $290 | $293;
      $295 = $294 << 8;
      $296 = ((($287)) + 2|0);
      $297 = HEAP8[$296>>0]|0;
      $298 = $297&255;
      $299 = $295 | $298;
      $300 = HEAP32[$$pre$phi26Z2D>>2]|0;
      $301 = $299 << $300;
      $302 = $301 & 16777215;
      $303 = (($286) + ($300))|0;
      $304 = (23 - ($284))|0;
      $305 = $302 >>> $304;
      $306 = $303 >> 3;
      $307 = (($287) + ($306)|0);
      HEAP32[$$pre$phi24Z2D>>2] = $307;
      $308 = $303 & 7;
      HEAP32[$$pre$phi26Z2D>>2] = $308;
      $309 = ((($sample$1$i)) + 4|0);
      HEAP32[$sample$1$i>>2] = $305;
      $sample$2$i = $309;
     }
     $310 = (($i$032$i) + 1)|0;
     $exitcond56$i = ($310|0)==($252|0);
     if ($exitcond56$i) {
      $sample$2$i$lcssa = $sample$2$i;
      break;
     } else {
      $ba$031$i = $283;$i$032$i = $310;$sample$030$i = $sample$2$i;
     }
    }
    $254 = $252 << 1;
    $scevgep55$i = (($balloc) + ($254<<2)|0);
    $255 = ($252|0)<(32);
    if ($255) {
     $ba$127$i = $scevgep55$i;$i$128$i = $252;$sample$326$i = $sample$2$i$lcssa;
     label = 39;
    } else {
     $456 = 0;
    }
   } else {
    $ba$127$i = $balloc;$i$128$i = $252;$sample$326$i = $smpb$i;
    label = 39;
   }
   if ((label|0) == 39) {
    while(1) {
     label = 0;
     $312 = ((($ba$127$i)) + 4|0);
     $313 = HEAP32[$ba$127$i>>2]|0;
     $314 = ($313|0)==(0);
     if ($314) {
      $sample$4$i = $sample$326$i;
     } else {
      $315 = (($313) + 1)|0;
      $316 = HEAP32[$$pre$phi24Z2D>>2]|0;
      $317 = HEAP8[$316>>0]|0;
      $318 = $317&255;
      $319 = $318 << 8;
      $320 = ((($316)) + 1|0);
      $321 = HEAP8[$320>>0]|0;
      $322 = $321&255;
      $323 = $319 | $322;
      $324 = $323 << 8;
      $325 = ((($316)) + 2|0);
      $326 = HEAP8[$325>>0]|0;
      $327 = $326&255;
      $328 = $324 | $327;
      $329 = HEAP32[$$pre$phi26Z2D>>2]|0;
      $330 = $328 << $329;
      $331 = $330 & 16777215;
      $332 = (($315) + ($329))|0;
      $333 = (23 - ($313))|0;
      $334 = $331 >>> $333;
      $335 = $332 >> 3;
      $336 = (($316) + ($335)|0);
      HEAP32[$$pre$phi24Z2D>>2] = $336;
      $337 = $332 & 7;
      HEAP32[$$pre$phi26Z2D>>2] = $337;
      $338 = ((($sample$326$i)) + 4|0);
      HEAP32[$sample$326$i>>2] = $334;
      $sample$4$i = $338;
     }
     $339 = (($i$128$i) + 1)|0;
     $exitcond54$i = ($339|0)==(32);
     if ($exitcond54$i) {
      $456 = 1;
      break;
     } else {
      $ba$127$i = $312;$i$128$i = $339;$sample$326$i = $sample$4$i;
      label = 39;
     }
    }
   }
   if ($253) {
    $311 = $252 << 1;
    $ba$216$i = $balloc;$f0$018$i = $3;$f1$019$i = $244;$i$220$i = 0;$sample$515$i = $smpb$i;$sca$017$i = $scale_index;
    while(1) {
     $340 = ((($ba$216$i)) + 4|0);
     $341 = HEAP32[$ba$216$i>>2]|0;
     $342 = ($341|0)==(0);
     if ($342) {
      $sample$6$i = $sample$515$i;$sca$1$i8 = $sca$017$i;$storemerge1$i = 0.0;
     } else {
      $343 = -1 << $341;
      $344 = ((($sample$515$i)) + 4|0);
      $345 = HEAP32[$sample$515$i>>2]|0;
      $346 = (($343) + 1)|0;
      $347 = (($346) + ($345))|0;
      $348 = (+($347|0));
      $349 = ((($sca$017$i)) + 4|0);
      $350 = HEAP32[$sca$017$i>>2]|0;
      $351 = (($341) + 1)|0;
      $352 = ((((($fr)) + 12440|0) + ($351<<8)|0) + ($350<<2)|0);
      $353 = +HEAPF32[$352>>2];
      $354 = $348 * $353;
      $sample$6$i = $344;$sca$1$i8 = $349;$storemerge1$i = $354;
     }
     $f0$1$i = ((($f0$018$i)) + 4|0);
     HEAPF32[$f0$018$i>>2] = $storemerge1$i;
     $355 = ((($ba$216$i)) + 8|0);
     $356 = HEAP32[$340>>2]|0;
     $357 = ($356|0)==(0);
     if ($357) {
      $sample$7$i = $sample$6$i;$sca$2$i9 = $sca$1$i8;$storemerge2$i = 0.0;
     } else {
      $358 = -1 << $356;
      $359 = ((($sample$6$i)) + 4|0);
      $360 = HEAP32[$sample$6$i>>2]|0;
      $361 = (($358) + 1)|0;
      $362 = (($361) + ($360))|0;
      $363 = (+($362|0));
      $364 = ((($sca$1$i8)) + 4|0);
      $365 = HEAP32[$sca$1$i8>>2]|0;
      $366 = (($356) + 1)|0;
      $367 = ((((($fr)) + 12440|0) + ($366<<8)|0) + ($365<<2)|0);
      $368 = +HEAPF32[$367>>2];
      $369 = $363 * $368;
      $sample$7$i = $359;$sca$2$i9 = $364;$storemerge2$i = $369;
     }
     $f1$1$i = ((($f1$019$i)) + 4|0);
     HEAPF32[$f1$019$i>>2] = $storemerge2$i;
     $370 = (($i$220$i) + 1)|0;
     $exitcond53$i = ($370|0)==($252|0);
     if ($exitcond53$i) {
      $sample$7$i$lcssa = $sample$7$i;$sca$2$i9$lcssa = $sca$2$i9;
      break;
     } else {
      $ba$216$i = $355;$f0$018$i = $f0$1$i;$f1$019$i = $f1$1$i;$i$220$i = $370;$sample$515$i = $sample$7$i;$sca$017$i = $sca$2$i9;
     }
    }
    $scevgep51$i = (($balloc) + ($311<<2)|0);
    $scevgep$i3 = (((($3)) + 128|0) + ($252<<2)|0);
    $scevgep52$i = (($3) + ($252<<2)|0);
    $ba$2$lcssa$i6 = $scevgep51$i;$f0$0$lcssa$i = $scevgep52$i;$f1$0$lcssa$i = $scevgep$i3;$sample$5$lcssa$i = $sample$7$i$lcssa;$sca$0$lcssa$i5 = $sca$2$i9$lcssa;
   } else {
    $ba$2$lcssa$i6 = $balloc;$f0$0$lcssa$i = $3;$f1$0$lcssa$i = $244;$sample$5$lcssa$i = $smpb$i;$sca$0$lcssa$i5 = $scale_index;
   }
   if ($456) {
    $ba$39$i = $ba$2$lcssa$i6;$f0$211$i = $f0$0$lcssa$i;$f1$212$i = $f1$0$lcssa$i;$i$313$i = $252;$sample$88$i = $sample$5$lcssa$i;$sca$310$i = $sca$0$lcssa$i5;
    while(1) {
     $371 = ((($ba$39$i)) + 4|0);
     $372 = HEAP32[$ba$39$i>>2]|0;
     $373 = ($372|0)==(0);
     if ($373) {
      HEAPF32[$f1$212$i>>2] = 0.0;
      HEAPF32[$f0$211$i>>2] = 0.0;
      $sample$9$i = $sample$88$i;$sca$4$i10 = $sca$310$i;
     } else {
      $374 = -1 << $372;
      $375 = ((($sample$88$i)) + 4|0);
      $376 = HEAP32[$sample$88$i>>2]|0;
      $377 = (($374) + 1)|0;
      $378 = (($377) + ($376))|0;
      $379 = (+($378|0));
      $380 = ((($sca$310$i)) + 4|0);
      $381 = HEAP32[$sca$310$i>>2]|0;
      $382 = (($372) + 1)|0;
      $383 = ((((($fr)) + 12440|0) + ($382<<8)|0) + ($381<<2)|0);
      $384 = +HEAPF32[$383>>2];
      $385 = $379 * $384;
      HEAPF32[$f0$211$i>>2] = $385;
      $386 = ((($sca$310$i)) + 8|0);
      $387 = HEAP32[$380>>2]|0;
      $388 = ((((($fr)) + 12440|0) + ($382<<8)|0) + ($387<<2)|0);
      $389 = +HEAPF32[$388>>2];
      $390 = $389 * $379;
      HEAPF32[$f1$212$i>>2] = $390;
      $sample$9$i = $375;$sca$4$i10 = $386;
     }
     $f1$3$i = ((($f1$212$i)) + 4|0);
     $f0$3$i = ((($f0$211$i)) + 4|0);
     $391 = (($i$313$i) + 1)|0;
     $exitcond50$i = ($391|0)==(32);
     if ($exitcond50$i) {
      break;
     } else {
      $ba$39$i = $371;$f0$211$i = $f0$3$i;$f1$212$i = $f1$3$i;$i$313$i = $391;$sample$88$i = $sample$9$i;$sca$310$i = $sca$4$i10;
     }
    }
   }
   $392 = HEAP32[$245>>2]|0;
   $393 = ($392|0)<(32);
   if ($393) {
    $i$47$i = $392;
    while(1) {
     $394 = (((($3)) + 128|0) + ($i$47$i<<2)|0);
     HEAPF32[$394>>2] = 0.0;
     $395 = (($3) + ($i$47$i<<2)|0);
     HEAPF32[$395>>2] = 0.0;
     $396 = (($i$47$i) + 1)|0;
     $exitcond$i11 = ($396|0)==(32);
     if ($exitcond$i11) {
      break;
     } else {
      $i$47$i = $396;
     }
    }
   }
  } else {
   $ba$446$i = $balloc;$i$547$i = 0;$sample$1045$i = $smpb$i;
   while(1) {
    $397 = ((($ba$446$i)) + 4|0);
    $398 = HEAP32[$ba$446$i>>2]|0;
    $399 = ($398|0)==(0);
    if ($399) {
     $sample$11$i = $sample$1045$i;
    } else {
     $400 = (($398) + 1)|0;
     $401 = HEAP32[$$pre$phi24Z2D>>2]|0;
     $402 = HEAP8[$401>>0]|0;
     $403 = $402&255;
     $404 = $403 << 8;
     $405 = ((($401)) + 1|0);
     $406 = HEAP8[$405>>0]|0;
     $407 = $406&255;
     $408 = $404 | $407;
     $409 = $408 << 8;
     $410 = ((($401)) + 2|0);
     $411 = HEAP8[$410>>0]|0;
     $412 = $411&255;
     $413 = $409 | $412;
     $414 = HEAP32[$$pre$phi26Z2D>>2]|0;
     $415 = $413 << $414;
     $416 = $415 & 16777215;
     $417 = (($400) + ($414))|0;
     $418 = (23 - ($398))|0;
     $419 = $416 >>> $418;
     $420 = $417 >> 3;
     $421 = (($401) + ($420)|0);
     HEAP32[$$pre$phi24Z2D>>2] = $421;
     $422 = $417 & 7;
     HEAP32[$$pre$phi26Z2D>>2] = $422;
     $423 = ((($sample$1045$i)) + 4|0);
     HEAP32[$sample$1045$i>>2] = $419;
     $sample$11$i = $423;
    }
    $424 = (($i$547$i) + 1)|0;
    $exitcond59$i = ($424|0)==(32);
    if ($exitcond59$i) {
     $ba$541$i = $balloc;$f01$044$i = $3;$i$643$i = 0;$sample$1240$i = $smpb$i;$sca$542$i = $scale_index;
     break;
    } else {
     $ba$446$i = $397;$i$547$i = $424;$sample$1045$i = $sample$11$i;
    }
   }
   while(1) {
    $425 = ((($ba$541$i)) + 4|0);
    $426 = HEAP32[$ba$541$i>>2]|0;
    $427 = ($426|0)==(0);
    if ($427) {
     $sample$13$i = $sample$1240$i;$sca$6$i13 = $sca$542$i;$storemerge$i = 0.0;
    } else {
     $428 = -1 << $426;
     $429 = ((($sample$1240$i)) + 4|0);
     $430 = HEAP32[$sample$1240$i>>2]|0;
     $431 = (($428) + 1)|0;
     $432 = (($431) + ($430))|0;
     $433 = (+($432|0));
     $434 = ((($sca$542$i)) + 4|0);
     $435 = HEAP32[$sca$542$i>>2]|0;
     $436 = (($426) + 1)|0;
     $437 = ((((($fr)) + 12440|0) + ($436<<8)|0) + ($435<<2)|0);
     $438 = +HEAPF32[$437>>2];
     $439 = $433 * $438;
     $sample$13$i = $429;$sca$6$i13 = $434;$storemerge$i = $439;
    }
    $f01$1$i = ((($f01$044$i)) + 4|0);
    HEAPF32[$f01$044$i>>2] = $storemerge$i;
    $440 = (($i$643$i) + 1)|0;
    $exitcond58$i = ($440|0)==(32);
    if ($exitcond58$i) {
     break;
    } else {
     $ba$541$i = $425;$f01$044$i = $f01$1$i;$i$643$i = $440;$sample$1240$i = $sample$13$i;$sca$542$i = $sca$6$i13;
    }
   }
   $441 = HEAP32[$245>>2]|0;
   $442 = ($441|0)<(32);
   if ($442) {
    $scevgep64$i = (($3) + ($441<<2)|0);
    $443 = $441 << 2;
    $444 = (128 - ($443))|0;
    _memset(($scevgep64$i|0),0,($444|0))|0;
   }
  }
  if ($246) {
   $447 = HEAP32[$247>>2]|0;
   $448 = (FUNCTION_TABLE_iiii[$447 & 8191]($3,$244,$fr)|0);
   $$pn = $448;
  } else {
   $445 = HEAP32[$248>>2]|0;
   $446 = (FUNCTION_TABLE_iii[$445 & 8191]($249,$fr)|0);
   $$pn = $446;
  }
  $clip$1 = (($$pn) + ($clip$016))|0;
  $449 = (($i$015) + 1)|0;
  $exitcond = ($449|0)==(12);
  if ($exitcond) {
   $clip$1$lcssa = $clip$1;
   break;
  } else {
   $clip$016 = $clip$1;$i$015 = $449;
  }
 }
 STACKTOP = sp;return ($clip$1$lcssa|0);
}
function _INT123_do_layer2($fr) {
 $fr = $fr|0;
 var $$ = 0, $$$i = 0, $$0$lcssa$i = 0, $$024$i = 0, $$08$i = 0, $$1$i = 0, $$1$lcssa$i = 0, $$114$i = 0, $$2$i = 0, $$2$i$lcssa = 0, $$39$i = 0, $$4$i = 0, $$lcssa29 = 0, $$lcssa30 = 0, $$pre = 0, $$pre$i = 0, $$pre18 = 0, $$pre19 = 0, $$pre41$i = 0, $$pre42$i = 0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0;
 var $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0;
 var $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0;
 var $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0;
 var $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0;
 var $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0.0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0;
 var $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0;
 var $440 = 0, $441 = 0.0, $442 = 0.0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0;
 var $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0.0, $465 = 0.0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0;
 var $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0.0, $488 = 0.0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0;
 var $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0;
 var $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0;
 var $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0;
 var $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0.0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0;
 var $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0.0, $582 = 0, $583 = 0, $584 = 0;
 var $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0;
 var $602 = 0, $603 = 0.0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0;
 var $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0.0, $626 = 0, $627 = 0.0, $628 = 0, $629 = 0.0, $63 = 0, $630 = 0, $631 = 0.0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0.0, $637 = 0.0, $638 = 0.0;
 var $639 = 0.0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0;
 var $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0;
 var $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0;
 var $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0;
 var $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0;
 var $729 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $alloc1$0$lcssa$i = 0, $alloc1$0$lcssa$i3 = 0, $alloc1$020$i = 0, $alloc1$033$i = 0, $alloc1$127$i = 0, $alloc1$17$i = 0, $alloc1$217$i = 0, $bit_alloc = 0, $bita$0$lcssa$i = 0, $bita$0$lcssa$i4 = 0;
 var $bita$019$i = 0, $bita$035$i = 0, $bita$1$lcssa$i = 0, $bita$112$i = 0, $bita$129$i = 0, $bita$224$i = 0, $bita$26$i = 0, $bita$319$i = 0, $bita$414$i = 0, $bita$511$i = 0, $clip$012 = 0, $clip$1$lcssa = 0, $clip$1$lcssa$lcssa = 0, $clip$2 = 0, $clip$2$1 = 0, $clip$2$2 = 0, $clip$2$us = 0, $clip$2$us$1 = 0, $clip$2$us$2 = 0, $exitcond$i = 0;
 var $exitcond17 = 0, $exitcond32$i = 0, $exitcond33$i = 0, $exitcond34$i = 0, $fold$i = 0, $fold35$i = 0, $fold36$i = 0, $fold37$i = 0, $i$011 = 0, $i$022$i = 0, $i$034$i = 0, $i$128$i = 0, $i$18$i = 0, $i$222$i = 0, $i$23$us$i = 0, $i$318$i = 0, $i$412$i = 0, $i$59$i = 0, $j$013$i = 0, $j$12$us$i = 0;
 var $or$cond = 0, $or$cond$i = 0, $scale = 0, $scevgep$i = 0, $scevgep$i6 = 0, $scfsi$023$i = 0, $scfsi$1$i = 0, $scfsi$213$i = 0, $scfsi$3$i = 0, $scfsi$410$i = 0, $scfsi$5$i = 0, $scfsi_buf$i = 0, $sext$i = 0, $sext1$i = 0, $sext2$i = 0, $sext3$i = 0, $sext4$i = 0, $sext5$i = 0, $table$0$i = 0, $table$i = 0;
 var $table2$i = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1360|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $table$i = sp + 1320|0;
 $table2$i = sp + 1280|0;
 $scfsi_buf$i = sp + 1024|0;
 $bit_alloc = sp + 768|0;
 $scale = sp;
 $0 = ((($fr)) + 19652|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 28596|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($fr)) + 19660|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($fr)) + 19704|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)>(2);
 if ($8) {
  $table$0$i = 4;
 } else {
  $9 = ((($fr)) + 19700|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = (2 - ($1))|0;
  $12 = (((36080 + ($7<<7)|0) + ($11<<6)|0) + ($10<<2)|0);
  $13 = HEAP32[$12>>2]|0;
  $table$0$i = $13;
 }
 $14 = (36464 + ($table$0$i<<2)|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = (36484 + ($table$0$i<<2)|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ((($fr)) + 19632|0);
 HEAP32[$18>>2] = $17;
 $19 = ((($fr)) + 19664|0);
 HEAP32[$19>>2] = $15;
 $20 = ((($fr)) + 19716|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ($21|0)==(1);
 $23 = $17;
 if ($22) {
  $25 = ((($fr)) + 19720|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = $26 << 2;
  $28 = (($27) + 4)|0;
  $29 = ((($fr)) + 19656|0);
  HEAP32[$29>>2] = $28;
  $30 = ($28|0)>($15|0);
  if ($30) {
   $31 = HEAP32[37544]|0;
   (_fwrite(511400,46,1,$31)|0);
   $32 = HEAP32[$19>>2]|0;
   HEAP32[$29>>2] = $32;
   $$pre = HEAP32[$0>>2]|0;
   $$pre18 = HEAP32[$18>>2]|0;
   $36 = $$pre;$38 = $32;$397 = $29;$41 = $32;$727 = $$pre18;
  } else {
   $36 = $1;$38 = $15;$397 = $29;$41 = $28;$727 = $23;
  }
 } else {
  $24 = ((($fr)) + 19656|0);
  HEAP32[$24>>2] = $15;
  $36 = $1;$38 = $15;$397 = $24;$41 = $15;$727 = $23;
 }
 $33 = ($1|0)==(1);
 $34 = ($5|0)==(3);
 $or$cond = $33 | $34;
 $$ = $or$cond ? 0 : $5;
 $35 = (($36) + -1)|0;
 $37 = $38 << $35;
 $39 = ($35|0)==(0);
 if ($39) {
  $44 = ($38|0)==(0);
  if ($44) {
   label = 27;
  } else {
   $45 = ((($fr)) + 19788|0);
   $46 = ((($fr)) + 19784|0);
   $$pre42$i = HEAP32[$45>>2]|0;
   $166 = $$pre42$i;$alloc1$217$i = $727;$bita$319$i = $bit_alloc;$i$318$i = $38;
   while(1) {
    $163 = HEAP16[$alloc1$217$i>>1]|0;
    $164 = $163 << 16 >> 16;
    $165 = HEAP8[$166>>0]|0;
    $167 = $165&255;
    $168 = $167 << 8;
    $169 = ((($166)) + 1|0);
    $170 = HEAP8[$169>>0]|0;
    $171 = $170&255;
    $172 = $168 | $171;
    $173 = $172 << 8;
    $174 = ((($166)) + 2|0);
    $175 = HEAP8[$174>>0]|0;
    $176 = $175&255;
    $177 = $173 | $176;
    $178 = HEAP32[$46>>2]|0;
    $179 = $177 << $178;
    $180 = $179 & 16777215;
    $181 = (($178) + ($164))|0;
    $182 = (24 - ($164))|0;
    $183 = $180 >>> $182;
    $184 = $181 >> 3;
    $185 = (($166) + ($184)|0);
    HEAP32[$45>>2] = $185;
    $186 = $181 & 7;
    HEAP32[$46>>2] = $186;
    $sext1$i = $183 << 24;
    $187 = $sext1$i >> 24;
    $188 = ((($bita$319$i)) + 4|0);
    HEAP32[$bita$319$i>>2] = $187;
    $189 = (($i$318$i) + -1)|0;
    $190 = 1 << $164;
    $191 = (($alloc1$217$i) + ($190<<2)|0);
    $192 = ($189|0)==(0);
    if ($192) {
     $$lcssa29 = $185;
     break;
    } else {
     $166 = $185;$alloc1$217$i = $191;$bita$319$i = $188;$i$318$i = $189;
    }
   }
   $193 = ((($fr)) + 19792|0);
   $198 = $$lcssa29;$bita$414$i = $bit_alloc;$i$412$i = $38;$scfsi$213$i = $scfsi_buf$i;
   while(1) {
    $194 = ((($bita$414$i)) + 4|0);
    $195 = HEAP32[$bita$414$i>>2]|0;
    $196 = ($195|0)==(0);
    if ($196) {
     $728 = $198;$scfsi$3$i = $scfsi$213$i;
    } else {
     $197 = HEAP8[$198>>0]|0;
     $199 = $197&255;
     $200 = HEAP32[$46>>2]|0;
     $201 = $199 << $200;
     $202 = $201 & 255;
     HEAP32[$193>>2] = $202;
     $203 = ((($198)) + 1|0);
     $204 = HEAP8[$203>>0]|0;
     $205 = $204&255;
     $206 = $205 << $200;
     $207 = $206 >>> 8;
     $208 = $207 | $202;
     $209 = $208 >>> 6;
     HEAP32[$193>>2] = $209;
     $210 = (($200) + 2)|0;
     $211 = $210 >> 3;
     $212 = (($198) + ($211)|0);
     HEAP32[$45>>2] = $212;
     $213 = $210 & 7;
     HEAP32[$46>>2] = $213;
     $sext$i = $209 << 24;
     $214 = $sext$i >> 24;
     $215 = ((($scfsi$213$i)) + 4|0);
     HEAP32[$scfsi$213$i>>2] = $214;
     $728 = $212;$scfsi$3$i = $215;
    }
    $216 = (($i$412$i) + -1)|0;
    $217 = ($216|0)==(0);
    if ($217) {
     label = 27;
     break;
    } else {
     $198 = $728;$bita$414$i = $194;$i$412$i = $216;$scfsi$213$i = $scfsi$3$i;
    }
   }
  }
 } else {
  $40 = ($41|0)==(0);
  if ($40) {
   $alloc1$0$lcssa$i = $727;$bita$0$lcssa$i = $bit_alloc;
  } else {
   $42 = ((($fr)) + 19788|0);
   $43 = ((($fr)) + 19784|0);
   $$pre$i = HEAP32[$42>>2]|0;
   $50 = $$pre$i;$alloc1$033$i = $727;$bita$035$i = $bit_alloc;$i$034$i = $41;
   while(1) {
    $47 = HEAP16[$alloc1$033$i>>1]|0;
    $48 = $47 << 16 >> 16;
    $49 = HEAP8[$50>>0]|0;
    $51 = $49&255;
    $52 = $51 << 8;
    $53 = ((($50)) + 1|0);
    $54 = HEAP8[$53>>0]|0;
    $55 = $54&255;
    $56 = $52 | $55;
    $57 = $56 << 8;
    $58 = ((($50)) + 2|0);
    $59 = HEAP8[$58>>0]|0;
    $60 = $59&255;
    $61 = $57 | $60;
    $62 = HEAP32[$43>>2]|0;
    $63 = $61 << $62;
    $64 = $63 & 16777215;
    $65 = (($62) + ($48))|0;
    $66 = (24 - ($48))|0;
    $67 = $64 >>> $66;
    $68 = $65 >> 3;
    $69 = (($50) + ($68)|0);
    HEAP32[$42>>2] = $69;
    $70 = $65 & 7;
    HEAP32[$43>>2] = $70;
    $sext4$i = $67 << 24;
    $71 = $sext4$i >> 24;
    $72 = ((($bita$035$i)) + 4|0);
    HEAP32[$bita$035$i>>2] = $71;
    $73 = HEAP8[$69>>0]|0;
    $74 = $73&255;
    $75 = $74 << 8;
    $76 = ((($69)) + 1|0);
    $77 = HEAP8[$76>>0]|0;
    $78 = $77&255;
    $79 = $75 | $78;
    $80 = $79 << 8;
    $81 = ((($69)) + 2|0);
    $82 = HEAP8[$81>>0]|0;
    $83 = $82&255;
    $84 = $80 | $83;
    $85 = HEAP32[$43>>2]|0;
    $86 = $84 << $85;
    $87 = $86 & 16777215;
    $88 = (($85) + ($48))|0;
    $89 = $87 >>> $66;
    $90 = $88 >> 3;
    $91 = (($69) + ($90)|0);
    HEAP32[$42>>2] = $91;
    $92 = $88 & 7;
    HEAP32[$43>>2] = $92;
    $sext5$i = $89 << 24;
    $93 = $sext5$i >> 24;
    $94 = ((($bita$035$i)) + 8|0);
    HEAP32[$72>>2] = $93;
    $95 = (($i$034$i) + -1)|0;
    $96 = 1 << $48;
    $97 = (($alloc1$033$i) + ($96<<2)|0);
    $98 = ($95|0)==(0);
    if ($98) {
     $$lcssa30 = $97;
     break;
    } else {
     $50 = $91;$alloc1$033$i = $97;$bita$035$i = $94;$i$034$i = $95;
    }
   }
   $99 = $41 << 1;
   $scevgep$i = (($bit_alloc) + ($99<<2)|0);
   $alloc1$0$lcssa$i = $$lcssa30;$bita$0$lcssa$i = $scevgep$i;
  }
  $100 = ($38|0)==($41|0);
  if (!($100)) {
   $101 = (($38) - ($41))|0;
   $102 = ((($fr)) + 19788|0);
   $103 = ((($fr)) + 19784|0);
   $$pre41$i = HEAP32[$102>>2]|0;
   $107 = $$pre41$i;$alloc1$127$i = $alloc1$0$lcssa$i;$bita$129$i = $bita$0$lcssa$i;$i$128$i = $101;
   while(1) {
    $104 = HEAP16[$alloc1$127$i>>1]|0;
    $105 = $104 << 16 >> 16;
    $106 = HEAP8[$107>>0]|0;
    $108 = $106&255;
    $109 = $108 << 8;
    $110 = ((($107)) + 1|0);
    $111 = HEAP8[$110>>0]|0;
    $112 = $111&255;
    $113 = $109 | $112;
    $114 = $113 << 8;
    $115 = ((($107)) + 2|0);
    $116 = HEAP8[$115>>0]|0;
    $117 = $116&255;
    $118 = $114 | $117;
    $119 = HEAP32[$103>>2]|0;
    $120 = $118 << $119;
    $121 = $120 & 16777215;
    $122 = (($119) + ($105))|0;
    $123 = (24 - ($105))|0;
    $124 = $121 >>> $123;
    $125 = $122 >> 3;
    $126 = (($107) + ($125)|0);
    HEAP32[$102>>2] = $126;
    $127 = $122 & 7;
    HEAP32[$103>>2] = $127;
    $sext3$i = $124 << 24;
    $128 = $sext3$i >> 24;
    HEAP32[$bita$129$i>>2] = $128;
    $129 = ((($bita$129$i)) + 4|0);
    HEAP32[$129>>2] = $128;
    $130 = ((($bita$129$i)) + 8|0);
    $131 = (($i$128$i) + -1)|0;
    $132 = 1 << $105;
    $133 = (($alloc1$127$i) + ($132<<2)|0);
    $134 = ($131|0)==(0);
    if ($134) {
     break;
    } else {
     $107 = $126;$alloc1$127$i = $133;$bita$129$i = $130;$i$128$i = $131;
    }
   }
  }
  $135 = ($37|0)==(0);
  if (!($135)) {
   $136 = ((($fr)) + 19788|0);
   $137 = ((($fr)) + 19784|0);
   $138 = ((($fr)) + 19792|0);
   $bita$224$i = $bit_alloc;$i$222$i = $37;$scfsi$023$i = $scfsi_buf$i;
   while(1) {
    $139 = ((($bita$224$i)) + 4|0);
    $140 = HEAP32[$bita$224$i>>2]|0;
    $141 = ($140|0)==(0);
    if ($141) {
     $scfsi$1$i = $scfsi$023$i;
    } else {
     $142 = HEAP32[$136>>2]|0;
     $143 = HEAP8[$142>>0]|0;
     $144 = $143&255;
     $145 = HEAP32[$137>>2]|0;
     $146 = $144 << $145;
     $147 = $146 & 255;
     HEAP32[$138>>2] = $147;
     $148 = ((($142)) + 1|0);
     $149 = HEAP8[$148>>0]|0;
     $150 = $149&255;
     $151 = $150 << $145;
     $152 = $151 >>> 8;
     $153 = $152 | $147;
     $154 = $153 >>> 6;
     HEAP32[$138>>2] = $154;
     $155 = (($145) + 2)|0;
     $156 = $155 >> 3;
     $157 = (($142) + ($156)|0);
     HEAP32[$136>>2] = $157;
     $158 = $155 & 7;
     HEAP32[$137>>2] = $158;
     $sext2$i = $154 << 24;
     $159 = $sext2$i >> 24;
     $160 = ((($scfsi$023$i)) + 4|0);
     HEAP32[$scfsi$023$i>>2] = $159;
     $scfsi$1$i = $160;
    }
    $161 = (($i$222$i) + -1)|0;
    $162 = ($161|0)==(0);
    if ($162) {
     label = 27;
     break;
    } else {
     $bita$224$i = $139;$i$222$i = $161;$scfsi$023$i = $scfsi$1$i;
    }
   }
  }
 }
 if ((label|0) == 27) {
  $218 = ($37|0)==(0);
  if (!($218)) {
   $219 = ((($fr)) + 19788|0);
   $220 = ((($fr)) + 19784|0);
   $221 = ((($fr)) + 19792|0);
   $$08$i = $scale;$bita$511$i = $bit_alloc;$i$59$i = $37;$scfsi$410$i = $scfsi_buf$i;
   while(1) {
    $222 = ((($bita$511$i)) + 4|0);
    $223 = HEAP32[$bita$511$i>>2]|0;
    $224 = ($223|0)==(0);
    L43: do {
     if ($224) {
      $$1$i = $$08$i;$scfsi$5$i = $scfsi$410$i;
     } else {
      $225 = ((($scfsi$410$i)) + 4|0);
      $226 = HEAP32[$scfsi$410$i>>2]|0;
      switch ($226|0) {
      case 0:  {
       $227 = HEAP32[$219>>2]|0;
       $228 = HEAP8[$227>>0]|0;
       $229 = $228&255;
       $230 = HEAP32[$220>>2]|0;
       $231 = $229 << $230;
       $232 = $231 & 255;
       HEAP32[$221>>2] = $232;
       $233 = ((($227)) + 1|0);
       $234 = HEAP8[$233>>0]|0;
       $235 = $234&255;
       $236 = $235 << $230;
       $237 = $236 >>> 8;
       $238 = $237 | $232;
       $239 = $238 >>> 2;
       HEAP32[$221>>2] = $239;
       $240 = (($230) + 6)|0;
       $241 = $240 >> 3;
       $242 = (($227) + ($241)|0);
       HEAP32[$219>>2] = $242;
       $243 = $240 & 7;
       HEAP32[$220>>2] = $243;
       $244 = ((($$08$i)) + 4|0);
       HEAP32[$$08$i>>2] = $239;
       $245 = HEAP8[$242>>0]|0;
       $246 = $245&255;
       $247 = HEAP32[$220>>2]|0;
       $248 = $246 << $247;
       $249 = $248 & 255;
       HEAP32[$221>>2] = $249;
       $250 = ((($242)) + 1|0);
       $251 = HEAP8[$250>>0]|0;
       $252 = $251&255;
       $253 = $252 << $247;
       $254 = $253 >>> 8;
       $255 = $254 | $249;
       $256 = $255 >>> 2;
       HEAP32[$221>>2] = $256;
       $257 = (($247) + 6)|0;
       $258 = $257 >> 3;
       $259 = (($242) + ($258)|0);
       HEAP32[$219>>2] = $259;
       $260 = $257 & 7;
       HEAP32[$220>>2] = $260;
       $261 = ((($$08$i)) + 8|0);
       HEAP32[$244>>2] = $256;
       $262 = HEAP8[$259>>0]|0;
       $263 = $262&255;
       $264 = HEAP32[$220>>2]|0;
       $265 = $263 << $264;
       $266 = $265 & 255;
       HEAP32[$221>>2] = $266;
       $267 = ((($259)) + 1|0);
       $268 = HEAP8[$267>>0]|0;
       $269 = $268&255;
       $270 = $269 << $264;
       $271 = $270 >>> 8;
       $272 = $271 | $266;
       $273 = $272 >>> 2;
       HEAP32[$221>>2] = $273;
       $274 = (($264) + 6)|0;
       $275 = $274 >> 3;
       $276 = (($259) + ($275)|0);
       HEAP32[$219>>2] = $276;
       $277 = $274 & 7;
       HEAP32[$220>>2] = $277;
       $278 = ((($$08$i)) + 12|0);
       HEAP32[$261>>2] = $273;
       $$1$i = $278;$scfsi$5$i = $225;
       break L43;
       break;
      }
      case 1:  {
       $279 = HEAP32[$219>>2]|0;
       $280 = HEAP8[$279>>0]|0;
       $281 = $280&255;
       $282 = HEAP32[$220>>2]|0;
       $283 = $281 << $282;
       $284 = $283 & 255;
       HEAP32[$221>>2] = $284;
       $285 = ((($279)) + 1|0);
       $286 = HEAP8[$285>>0]|0;
       $287 = $286&255;
       $288 = $287 << $282;
       $289 = $288 >>> 8;
       $290 = $289 | $284;
       $291 = $290 >>> 2;
       HEAP32[$221>>2] = $291;
       $292 = (($282) + 6)|0;
       $293 = $292 >> 3;
       $294 = (($279) + ($293)|0);
       HEAP32[$219>>2] = $294;
       $295 = $292 & 7;
       HEAP32[$220>>2] = $295;
       $296 = ((($$08$i)) + 4|0);
       HEAP32[$$08$i>>2] = $291;
       $297 = ((($$08$i)) + 8|0);
       HEAP32[$296>>2] = $291;
       $298 = HEAP8[$294>>0]|0;
       $299 = $298&255;
       $300 = HEAP32[$220>>2]|0;
       $301 = $299 << $300;
       $302 = $301 & 255;
       HEAP32[$221>>2] = $302;
       $303 = ((($294)) + 1|0);
       $304 = HEAP8[$303>>0]|0;
       $305 = $304&255;
       $306 = $305 << $300;
       $307 = $306 >>> 8;
       $308 = $307 | $302;
       $309 = $308 >>> 2;
       HEAP32[$221>>2] = $309;
       $310 = (($300) + 6)|0;
       $311 = $310 >> 3;
       $312 = (($294) + ($311)|0);
       HEAP32[$219>>2] = $312;
       $313 = $310 & 7;
       HEAP32[$220>>2] = $313;
       $314 = ((($$08$i)) + 12|0);
       HEAP32[$297>>2] = $309;
       $$1$i = $314;$scfsi$5$i = $225;
       break L43;
       break;
      }
      case 2:  {
       $315 = HEAP32[$219>>2]|0;
       $316 = HEAP8[$315>>0]|0;
       $317 = $316&255;
       $318 = HEAP32[$220>>2]|0;
       $319 = $317 << $318;
       $320 = $319 & 255;
       HEAP32[$221>>2] = $320;
       $321 = ((($315)) + 1|0);
       $322 = HEAP8[$321>>0]|0;
       $323 = $322&255;
       $324 = $323 << $318;
       $325 = $324 >>> 8;
       $326 = $325 | $320;
       $327 = $326 >>> 2;
       HEAP32[$221>>2] = $327;
       $328 = (($318) + 6)|0;
       $329 = $328 >> 3;
       $330 = (($315) + ($329)|0);
       HEAP32[$219>>2] = $330;
       $331 = $328 & 7;
       HEAP32[$220>>2] = $331;
       $332 = ((($$08$i)) + 4|0);
       HEAP32[$$08$i>>2] = $327;
       $333 = ((($$08$i)) + 8|0);
       HEAP32[$332>>2] = $327;
       $334 = ((($$08$i)) + 12|0);
       HEAP32[$333>>2] = $327;
       $$1$i = $334;$scfsi$5$i = $225;
       break L43;
       break;
      }
      default: {
       $335 = HEAP32[$219>>2]|0;
       $336 = HEAP8[$335>>0]|0;
       $337 = $336&255;
       $338 = HEAP32[$220>>2]|0;
       $339 = $337 << $338;
       $340 = $339 & 255;
       HEAP32[$221>>2] = $340;
       $341 = ((($335)) + 1|0);
       $342 = HEAP8[$341>>0]|0;
       $343 = $342&255;
       $344 = $343 << $338;
       $345 = $344 >>> 8;
       $346 = $345 | $340;
       $347 = $346 >>> 2;
       HEAP32[$221>>2] = $347;
       $348 = (($338) + 6)|0;
       $349 = $348 >> 3;
       $350 = (($335) + ($349)|0);
       HEAP32[$219>>2] = $350;
       $351 = $348 & 7;
       HEAP32[$220>>2] = $351;
       $352 = ((($$08$i)) + 4|0);
       HEAP32[$$08$i>>2] = $347;
       $353 = HEAP8[$350>>0]|0;
       $354 = $353&255;
       $355 = HEAP32[$220>>2]|0;
       $356 = $354 << $355;
       $357 = $356 & 255;
       HEAP32[$221>>2] = $357;
       $358 = ((($350)) + 1|0);
       $359 = HEAP8[$358>>0]|0;
       $360 = $359&255;
       $361 = $360 << $355;
       $362 = $361 >>> 8;
       $363 = $362 | $357;
       $364 = $363 >>> 2;
       HEAP32[$221>>2] = $364;
       $365 = (($355) + 6)|0;
       $366 = $365 >> 3;
       $367 = (($350) + ($366)|0);
       HEAP32[$219>>2] = $367;
       $368 = $365 & 7;
       HEAP32[$220>>2] = $368;
       $369 = ((($$08$i)) + 8|0);
       HEAP32[$352>>2] = $364;
       $370 = ((($$08$i)) + 12|0);
       HEAP32[$369>>2] = $364;
       $$1$i = $370;$scfsi$5$i = $225;
       break L43;
      }
      }
     }
    } while(0);
    $371 = (($i$59$i) + -1)|0;
    $372 = ($371|0)==(0);
    if ($372) {
     break;
    } else {
     $$08$i = $$1$i;$bita$511$i = $222;$i$59$i = $371;$scfsi$410$i = $scfsi$5$i;
    }
   }
  }
 }
 $373 = ((($fr)) + 19788|0);
 $374 = ((($fr)) + 19784|0);
 $375 = ((($table$i)) + 12|0);
 $376 = ((($table$i)) + 20|0);
 $377 = ((($table$i)) + 36|0);
 $378 = ((($table2$i)) + 12|0);
 $379 = ((($table2$i)) + 20|0);
 $380 = ((($table2$i)) + 36|0);
 $381 = ((($fr)) + 19668|0);
 $382 = ($$|0)==(-1);
 $383 = ((($fr)) + 19640|0);
 $384 = ((($fr)) + 19644|0);
 $385 = ((($3)) + 512|0);
 $386 = ((($3)) + 128|0);
 $387 = ((($3)) + 640|0);
 $388 = ((($3)) + 256|0);
 $389 = ((($3)) + 768|0);
 $390 = (($3) + ($$<<9)|0);
 $391 = (((($3) + ($$<<9)|0)) + 128|0);
 $392 = (((($3) + ($$<<9)|0)) + 256|0);
 $729 = $727;$clip$012 = 0;$i$011 = 0;
 while(1) {
  $393 = $i$011 >> 2;
  $394 = HEAP32[$0>>2]|0;
  $395 = HEAP32[$19>>2]|0;
  $396 = HEAP32[$397>>2]|0;
  $398 = ($396|0)>(0);
  if ($398) {
   $399 = ($394|0)>(0);
   $$024$i = $scale;$alloc1$020$i = $729;$bita$019$i = $bit_alloc;$i$022$i = 0;
   while(1) {
    $402 = HEAP16[$alloc1$020$i>>1]|0;
    $403 = $402 << 16 >> 16;
    if ($399) {
     $$114$i = $$024$i;$bita$112$i = $bita$019$i;$j$013$i = 0;
     while(1) {
      $404 = ((($bita$112$i)) + 4|0);
      $405 = HEAP32[$bita$112$i>>2]|0;
      $406 = ($405|0)==(0);
      if ($406) {
       $534 = ((((($3) + ($j$013$i<<9)|0)) + 256|0) + ($i$022$i<<2)|0);
       HEAPF32[$534>>2] = 0.0;
       $535 = ((((($3) + ($j$013$i<<9)|0)) + 128|0) + ($i$022$i<<2)|0);
       HEAPF32[$535>>2] = 0.0;
       $536 = ((($3) + ($j$013$i<<9)|0) + ($i$022$i<<2)|0);
       HEAPF32[$536>>2] = 0.0;
       $$2$i = $$114$i;
      } else {
       $407 = (($alloc1$020$i) + ($405<<2)|0);
       $408 = HEAP16[$407>>1]|0;
       $409 = $408 << 16 >> 16;
       $410 = (((($alloc1$020$i) + ($405<<2)|0)) + 2|0);
       $411 = HEAP16[$410>>1]|0;
       $412 = $411 << 16 >> 16;
       $413 = ($411<<16>>16)<(0);
       if ($413) {
        $414 = (($$114$i) + ($393<<2)|0);
        $415 = HEAP32[$414>>2]|0;
        $416 = ((((($fr)) + 12440|0) + ($409<<8)|0) + ($415<<2)|0);
        $417 = +HEAPF32[$416>>2];
        $418 = HEAP32[$373>>2]|0;
        $419 = HEAP8[$418>>0]|0;
        $420 = $419&255;
        $421 = $420 << 8;
        $422 = ((($418)) + 1|0);
        $423 = HEAP8[$422>>0]|0;
        $424 = $423&255;
        $425 = $421 | $424;
        $426 = $425 << 8;
        $427 = ((($418)) + 2|0);
        $428 = HEAP8[$427>>0]|0;
        $429 = $428&255;
        $430 = $426 | $429;
        $431 = HEAP32[$374>>2]|0;
        $432 = $430 << $431;
        $433 = $432 & 16777215;
        $434 = (($431) + ($409))|0;
        $435 = (24 - ($409))|0;
        $436 = $433 >>> $435;
        $437 = $434 >> 3;
        $438 = (($418) + ($437)|0);
        HEAP32[$373>>2] = $438;
        $439 = $434 & 7;
        HEAP32[$374>>2] = $439;
        $440 = (($436) + ($412))|0;
        $441 = (+($440|0));
        $442 = $441 * $417;
        $443 = ((($3) + ($j$013$i<<9)|0) + ($i$022$i<<2)|0);
        HEAPF32[$443>>2] = $442;
        $444 = HEAP8[$438>>0]|0;
        $445 = $444&255;
        $446 = $445 << 8;
        $447 = ((($438)) + 1|0);
        $448 = HEAP8[$447>>0]|0;
        $449 = $448&255;
        $450 = $446 | $449;
        $451 = $450 << 8;
        $452 = ((($438)) + 2|0);
        $453 = HEAP8[$452>>0]|0;
        $454 = $453&255;
        $455 = $451 | $454;
        $456 = $455 << $439;
        $457 = $456 & 16777215;
        $458 = (($439) + ($409))|0;
        $459 = $457 >>> $435;
        $460 = $458 >> 3;
        $461 = (($438) + ($460)|0);
        HEAP32[$373>>2] = $461;
        $fold36$i = (($434) + ($409))|0;
        $462 = $fold36$i & 7;
        HEAP32[$374>>2] = $462;
        $463 = (($459) + ($412))|0;
        $464 = (+($463|0));
        $465 = $464 * $417;
        $466 = ((((($3) + ($j$013$i<<9)|0)) + 128|0) + ($i$022$i<<2)|0);
        HEAPF32[$466>>2] = $465;
        $467 = HEAP8[$461>>0]|0;
        $468 = $467&255;
        $469 = $468 << 8;
        $470 = ((($461)) + 1|0);
        $471 = HEAP8[$470>>0]|0;
        $472 = $471&255;
        $473 = $469 | $472;
        $474 = $473 << 8;
        $475 = ((($461)) + 2|0);
        $476 = HEAP8[$475>>0]|0;
        $477 = $476&255;
        $478 = $474 | $477;
        $479 = $478 << $462;
        $480 = $479 & 16777215;
        $481 = (($462) + ($409))|0;
        $482 = $480 >>> $435;
        $483 = $481 >> 3;
        $484 = (($461) + ($483)|0);
        HEAP32[$373>>2] = $484;
        $fold37$i = (($fold36$i) + ($409))|0;
        $485 = $fold37$i & 7;
        HEAP32[$374>>2] = $485;
        $486 = (($482) + ($412))|0;
        $487 = (+($486|0));
        $488 = $487 * $417;
        $489 = ((((($3) + ($j$013$i<<9)|0)) + 256|0) + ($i$022$i<<2)|0);
        HEAPF32[$489>>2] = $488;
       } else {
        dest=$table$i; stop=dest+36|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
        HEAP32[$375>>2] = 667216;
        HEAP32[$376>>2] = 667600;
        HEAP32[$377>>2] = 669136;
        $490 = (($$114$i) + ($393<<2)|0);
        $491 = HEAP32[$490>>2]|0;
        $492 = HEAP32[$373>>2]|0;
        $493 = HEAP8[$492>>0]|0;
        $494 = $493&255;
        $495 = $494 << 8;
        $496 = ((($492)) + 1|0);
        $497 = HEAP8[$496>>0]|0;
        $498 = $497&255;
        $499 = $495 | $498;
        $500 = $499 << 8;
        $501 = ((($492)) + 2|0);
        $502 = HEAP8[$501>>0]|0;
        $503 = $502&255;
        $504 = $500 | $503;
        $505 = HEAP32[$374>>2]|0;
        $506 = $504 << $505;
        $507 = $506 & 16777215;
        $508 = (($505) + ($409))|0;
        $509 = (24 - ($409))|0;
        $510 = $507 >>> $509;
        $511 = $508 >> 3;
        $512 = (($492) + ($511)|0);
        HEAP32[$373>>2] = $512;
        $513 = $508 & 7;
        HEAP32[$374>>2] = $513;
        $514 = (($table$i) + ($412<<2)|0);
        $515 = HEAP32[$514>>2]|0;
        $516 = (($515) + ($510<<2)|0);
        $517 = (($516) + ($510<<2)|0);
        $518 = (($517) + ($510<<2)|0);
        $519 = ((($518)) + 4|0);
        $520 = HEAP32[$518>>2]|0;
        $521 = ((((($fr)) + 12440|0) + ($520<<8)|0) + ($491<<2)|0);
        $522 = HEAP32[$521>>2]|0;
        $523 = ((($3) + ($j$013$i<<9)|0) + ($i$022$i<<2)|0);
        HEAP32[$523>>2] = $522;
        $524 = ((($519)) + 4|0);
        $525 = HEAP32[$519>>2]|0;
        $526 = ((((($fr)) + 12440|0) + ($525<<8)|0) + ($491<<2)|0);
        $527 = HEAP32[$526>>2]|0;
        $528 = ((((($3) + ($j$013$i<<9)|0)) + 128|0) + ($i$022$i<<2)|0);
        HEAP32[$528>>2] = $527;
        $529 = HEAP32[$524>>2]|0;
        $530 = ((((($fr)) + 12440|0) + ($529<<8)|0) + ($491<<2)|0);
        $531 = HEAP32[$530>>2]|0;
        $532 = ((((($3) + ($j$013$i<<9)|0)) + 256|0) + ($i$022$i<<2)|0);
        HEAP32[$532>>2] = $531;
       }
       $533 = ((($$114$i)) + 12|0);
       $$2$i = $533;
      }
      $537 = (($j$013$i) + 1)|0;
      $exitcond33$i = ($537|0)==($394|0);
      if ($exitcond33$i) {
       $$2$i$lcssa = $$2$i;
       break;
      } else {
       $$114$i = $$2$i;$bita$112$i = $404;$j$013$i = $537;
      }
     }
     $scevgep$i6 = (($bita$019$i) + ($394<<2)|0);
     $$1$lcssa$i = $$2$i$lcssa;$bita$1$lcssa$i = $scevgep$i6;
    } else {
     $$1$lcssa$i = $$024$i;$bita$1$lcssa$i = $bita$019$i;
    }
    $538 = (($i$022$i) + 1)|0;
    $539 = 1 << $403;
    $540 = (($alloc1$020$i) + ($539<<2)|0);
    $exitcond34$i = ($538|0)==($396|0);
    if ($exitcond34$i) {
     $$0$lcssa$i = $$1$lcssa$i;$alloc1$0$lcssa$i3 = $540;$bita$0$lcssa$i4 = $bita$1$lcssa$i;
     break;
    } else {
     $$024$i = $$1$lcssa$i;$alloc1$020$i = $540;$bita$019$i = $bita$1$lcssa$i;$i$022$i = $538;
    }
   }
  } else {
   $$0$lcssa$i = $scale;$alloc1$0$lcssa$i3 = $729;$bita$0$lcssa$i4 = $bit_alloc;
  }
  $400 = ($396|0)<($395|0);
  if ($400) {
   $401 = (($393) + 3)|0;
   $$39$i = $$0$lcssa$i;$alloc1$17$i = $alloc1$0$lcssa$i3;$bita$26$i = $bita$0$lcssa$i4;$i$18$i = $396;
   while(1) {
    $541 = HEAP16[$alloc1$17$i>>1]|0;
    $542 = $541 << 16 >> 16;
    $543 = ((($bita$26$i)) + 4|0);
    $544 = ((($bita$26$i)) + 8|0);
    $545 = HEAP32[$543>>2]|0;
    $546 = ($545|0)==(0);
    if ($546) {
     $695 = (((($3)) + 768|0) + ($i$18$i<<2)|0);
     HEAPF32[$695>>2] = 0.0;
     $696 = (((($3)) + 640|0) + ($i$18$i<<2)|0);
     HEAPF32[$696>>2] = 0.0;
     $697 = (((($3)) + 512|0) + ($i$18$i<<2)|0);
     HEAPF32[$697>>2] = 0.0;
     $698 = (((($3)) + 256|0) + ($i$18$i<<2)|0);
     HEAPF32[$698>>2] = 0.0;
     $699 = (((($3)) + 128|0) + ($i$18$i<<2)|0);
     HEAPF32[$699>>2] = 0.0;
     $700 = (($3) + ($i$18$i<<2)|0);
     HEAPF32[$700>>2] = 0.0;
     $$4$i = $$39$i;
    } else {
     $547 = (($alloc1$17$i) + ($545<<2)|0);
     $548 = HEAP16[$547>>1]|0;
     $549 = $548 << 16 >> 16;
     $550 = (((($alloc1$17$i) + ($545<<2)|0)) + 2|0);
     $551 = HEAP16[$550>>1]|0;
     $552 = $551 << 16 >> 16;
     $553 = ($551<<16>>16)<(0);
     if ($553) {
      $554 = (($$39$i) + ($401<<2)|0);
      $555 = HEAP32[$554>>2]|0;
      $556 = ((((($fr)) + 12440|0) + ($549<<8)|0) + ($555<<2)|0);
      $557 = +HEAPF32[$556>>2];
      $558 = HEAP32[$373>>2]|0;
      $559 = HEAP8[$558>>0]|0;
      $560 = $559&255;
      $561 = $560 << 8;
      $562 = ((($558)) + 1|0);
      $563 = HEAP8[$562>>0]|0;
      $564 = $563&255;
      $565 = $561 | $564;
      $566 = $565 << 8;
      $567 = ((($558)) + 2|0);
      $568 = HEAP8[$567>>0]|0;
      $569 = $568&255;
      $570 = $566 | $569;
      $571 = HEAP32[$374>>2]|0;
      $572 = $570 << $571;
      $573 = $572 & 16777215;
      $574 = (($571) + ($549))|0;
      $575 = (24 - ($549))|0;
      $576 = $573 >>> $575;
      $577 = $574 >> 3;
      $578 = (($558) + ($577)|0);
      HEAP32[$373>>2] = $578;
      $579 = $574 & 7;
      HEAP32[$374>>2] = $579;
      $580 = (($576) + ($552))|0;
      $581 = (+($580|0));
      $582 = (($3) + ($i$18$i<<2)|0);
      HEAPF32[$582>>2] = $581;
      $583 = HEAP8[$578>>0]|0;
      $584 = $583&255;
      $585 = $584 << 8;
      $586 = ((($578)) + 1|0);
      $587 = HEAP8[$586>>0]|0;
      $588 = $587&255;
      $589 = $585 | $588;
      $590 = $589 << 8;
      $591 = ((($578)) + 2|0);
      $592 = HEAP8[$591>>0]|0;
      $593 = $592&255;
      $594 = $590 | $593;
      $595 = $594 << $579;
      $596 = $595 & 16777215;
      $597 = (($579) + ($549))|0;
      $598 = $596 >>> $575;
      $599 = $597 >> 3;
      $600 = (($578) + ($599)|0);
      HEAP32[$373>>2] = $600;
      $fold$i = (($574) + ($549))|0;
      $601 = $fold$i & 7;
      HEAP32[$374>>2] = $601;
      $602 = (($598) + ($552))|0;
      $603 = (+($602|0));
      $604 = (((($3)) + 128|0) + ($i$18$i<<2)|0);
      HEAPF32[$604>>2] = $603;
      $605 = HEAP8[$600>>0]|0;
      $606 = $605&255;
      $607 = $606 << 8;
      $608 = ((($600)) + 1|0);
      $609 = HEAP8[$608>>0]|0;
      $610 = $609&255;
      $611 = $607 | $610;
      $612 = $611 << 8;
      $613 = ((($600)) + 2|0);
      $614 = HEAP8[$613>>0]|0;
      $615 = $614&255;
      $616 = $612 | $615;
      $617 = $616 << $601;
      $618 = $617 & 16777215;
      $619 = (($601) + ($549))|0;
      $620 = $618 >>> $575;
      $621 = $619 >> 3;
      $622 = (($600) + ($621)|0);
      HEAP32[$373>>2] = $622;
      $fold35$i = (($fold$i) + ($549))|0;
      $623 = $fold35$i & 7;
      HEAP32[$374>>2] = $623;
      $624 = (($620) + ($552))|0;
      $625 = (+($624|0));
      $626 = (((($3)) + 256|0) + ($i$18$i<<2)|0);
      HEAPF32[$626>>2] = $625;
      $627 = $581 * $557;
      $628 = (((($3)) + 512|0) + ($i$18$i<<2)|0);
      HEAPF32[$628>>2] = $627;
      $629 = $603 * $557;
      $630 = (((($3)) + 640|0) + ($i$18$i<<2)|0);
      HEAPF32[$630>>2] = $629;
      $631 = $625 * $557;
      $632 = (((($3)) + 768|0) + ($i$18$i<<2)|0);
      HEAPF32[$632>>2] = $631;
      $633 = (($$39$i) + ($393<<2)|0);
      $634 = HEAP32[$633>>2]|0;
      $635 = ((((($fr)) + 12440|0) + ($549<<8)|0) + ($634<<2)|0);
      $636 = +HEAPF32[$635>>2];
      $637 = $636 * $581;
      HEAPF32[$582>>2] = $637;
      $638 = $636 * $603;
      HEAPF32[$604>>2] = $638;
      $639 = $625 * $636;
      HEAPF32[$626>>2] = $639;
     } else {
      dest=$table2$i; stop=dest+36|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
      HEAP32[$378>>2] = 667216;
      HEAP32[$379>>2] = 667600;
      HEAP32[$380>>2] = 669136;
      $640 = (($$39$i) + ($393<<2)|0);
      $641 = HEAP32[$640>>2]|0;
      $642 = (($$39$i) + ($401<<2)|0);
      $643 = HEAP32[$642>>2]|0;
      $644 = HEAP32[$373>>2]|0;
      $645 = HEAP8[$644>>0]|0;
      $646 = $645&255;
      $647 = $646 << 8;
      $648 = ((($644)) + 1|0);
      $649 = HEAP8[$648>>0]|0;
      $650 = $649&255;
      $651 = $647 | $650;
      $652 = $651 << 8;
      $653 = ((($644)) + 2|0);
      $654 = HEAP8[$653>>0]|0;
      $655 = $654&255;
      $656 = $652 | $655;
      $657 = HEAP32[$374>>2]|0;
      $658 = $656 << $657;
      $659 = $658 & 16777215;
      $660 = (($657) + ($549))|0;
      $661 = (24 - ($549))|0;
      $662 = $659 >>> $661;
      $663 = $660 >> 3;
      $664 = (($644) + ($663)|0);
      HEAP32[$373>>2] = $664;
      $665 = $660 & 7;
      HEAP32[$374>>2] = $665;
      $666 = (($table2$i) + ($552<<2)|0);
      $667 = HEAP32[$666>>2]|0;
      $668 = (($667) + ($662<<2)|0);
      $669 = (($668) + ($662<<2)|0);
      $670 = (($669) + ($662<<2)|0);
      $671 = HEAP32[$670>>2]|0;
      $672 = ((((($fr)) + 12440|0) + ($671<<8)|0) + ($641<<2)|0);
      $673 = HEAP32[$672>>2]|0;
      $674 = (($3) + ($i$18$i<<2)|0);
      HEAP32[$674>>2] = $673;
      $675 = ((($670)) + 4|0);
      $676 = ((((($fr)) + 12440|0) + ($671<<8)|0) + ($643<<2)|0);
      $677 = HEAP32[$676>>2]|0;
      $678 = (((($3)) + 512|0) + ($i$18$i<<2)|0);
      HEAP32[$678>>2] = $677;
      $679 = HEAP32[$675>>2]|0;
      $680 = ((((($fr)) + 12440|0) + ($679<<8)|0) + ($641<<2)|0);
      $681 = HEAP32[$680>>2]|0;
      $682 = (((($3)) + 128|0) + ($i$18$i<<2)|0);
      HEAP32[$682>>2] = $681;
      $683 = ((($675)) + 4|0);
      $684 = ((((($fr)) + 12440|0) + ($679<<8)|0) + ($643<<2)|0);
      $685 = HEAP32[$684>>2]|0;
      $686 = (((($3)) + 640|0) + ($i$18$i<<2)|0);
      HEAP32[$686>>2] = $685;
      $687 = HEAP32[$683>>2]|0;
      $688 = ((((($fr)) + 12440|0) + ($687<<8)|0) + ($641<<2)|0);
      $689 = HEAP32[$688>>2]|0;
      $690 = (((($3)) + 256|0) + ($i$18$i<<2)|0);
      HEAP32[$690>>2] = $689;
      $691 = ((((($fr)) + 12440|0) + ($687<<8)|0) + ($643<<2)|0);
      $692 = HEAP32[$691>>2]|0;
      $693 = (((($3)) + 768|0) + ($i$18$i<<2)|0);
      HEAP32[$693>>2] = $692;
     }
     $694 = ((($$39$i)) + 24|0);
     $$4$i = $694;
    }
    $701 = (($i$18$i) + 1)|0;
    $702 = 1 << $542;
    $703 = (($alloc1$17$i) + ($702<<2)|0);
    $exitcond32$i = ($701|0)==($395|0);
    if ($exitcond32$i) {
     break;
    } else {
     $$39$i = $$4$i;$alloc1$17$i = $703;$bita$26$i = $544;$i$18$i = $701;
    }
   }
  }
  $704 = HEAP32[$381>>2]|0;
  $705 = ($395|0)>($704|0);
  $$$i = $705 ? $704 : $395;
  $706 = ($$$i|0)<(32);
  $707 = ($394|0)>(0);
  $or$cond$i = $707 & $706;
  if ($or$cond$i) {
   $i$23$us$i = $$$i;
   while(1) {
    $j$12$us$i = 0;
    while(1) {
     $708 = ((((($3) + ($j$12$us$i<<9)|0)) + 256|0) + ($i$23$us$i<<2)|0);
     HEAPF32[$708>>2] = 0.0;
     $709 = ((((($3) + ($j$12$us$i<<9)|0)) + 128|0) + ($i$23$us$i<<2)|0);
     HEAPF32[$709>>2] = 0.0;
     $710 = ((($3) + ($j$12$us$i<<9)|0) + ($i$23$us$i<<2)|0);
     HEAPF32[$710>>2] = 0.0;
     $711 = (($j$12$us$i) + 1)|0;
     $exitcond$i = ($711|0)==($394|0);
     if ($exitcond$i) {
      break;
     } else {
      $j$12$us$i = $711;
     }
    }
    $712 = (($i$23$us$i) + 1)|0;
    $713 = ($712|0)<(32);
    if ($713) {
     $i$23$us$i = $712;
    } else {
     break;
    }
   }
  }
  if ($382) {
   $720 = HEAP32[$383>>2]|0;
   $721 = (FUNCTION_TABLE_iiii[$720 & 8191]($3,$385,$fr)|0);
   $clip$2$us = (($721) + ($clip$012))|0;
   $722 = HEAP32[$383>>2]|0;
   $723 = (FUNCTION_TABLE_iiii[$722 & 8191]($386,$387,$fr)|0);
   $clip$2$us$1 = (($723) + ($clip$2$us))|0;
   $724 = HEAP32[$383>>2]|0;
   $725 = (FUNCTION_TABLE_iiii[$724 & 8191]($388,$389,$fr)|0);
   $clip$2$us$2 = (($725) + ($clip$2$us$1))|0;
   $clip$1$lcssa = $clip$2$us$2;
  } else {
   $714 = HEAP32[$384>>2]|0;
   $715 = (FUNCTION_TABLE_iii[$714 & 8191]($390,$fr)|0);
   $clip$2 = (($715) + ($clip$012))|0;
   $716 = HEAP32[$384>>2]|0;
   $717 = (FUNCTION_TABLE_iii[$716 & 8191]($391,$fr)|0);
   $clip$2$1 = (($717) + ($clip$2))|0;
   $718 = HEAP32[$384>>2]|0;
   $719 = (FUNCTION_TABLE_iii[$718 & 8191]($392,$fr)|0);
   $clip$2$2 = (($719) + ($clip$2$1))|0;
   $clip$1$lcssa = $clip$2$2;
  }
  $726 = (($i$011) + 1)|0;
  $exitcond17 = ($726|0)==(12);
  if ($exitcond17) {
   $clip$1$lcssa$lcssa = $clip$1$lcssa;
   break;
  }
  $$pre19 = HEAP32[$18>>2]|0;
  $729 = $$pre19;$clip$012 = $clip$1$lcssa;$i$011 = $726;
 }
 STACKTOP = sp;return ($clip$1$lcssa$lcssa|0);
}
function _INT123_do_layer3($fr) {
 $fr = $fr|0;
 var $$ = 0, $$$i = 0, $$$i10 = 0, $$4$i = 0, $$5 = 0, $$9 = 0, $$do_l$0$i = 0, $$do_l$0$i$lcssa = 0, $$neg9$i = 0, $$pn$i = 0, $$pre = 0, $$pre$i = 0, $$pre$phi$iZ2D = 0, $$pre24$i = 0, $$pre25$i = 0, $$pre26$i = 0, $$pre27$i = 0, $$pre28$i = 0, $$pre31$i = 0, $$pre31$i87 = 0;
 var $$pre31$i88 = 0, $$pre32$i = 0, $$pre32$i90 = 0, $$pre32$i91 = 0, $$pre33$i = 0, $$pre33$pre$i = 0, $$pre34$i = 0, $$pre34$pre$i = 0, $$pre38$i = 0, $$pre44$i = 0, $$sink23$i = 0, $$sink59$i = 0, $$sink59$i$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0;
 var $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0;
 var $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0;
 var $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0;
 var $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0;
 var $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0;
 var $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0;
 var $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0;
 var $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0;
 var $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0;
 var $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0;
 var $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0;
 var $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0;
 var $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0;
 var $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0;
 var $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0;
 var $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0;
 var $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0;
 var $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0;
 var $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0.0, $601 = 0;
 var $602 = 0.0, $603 = 0.0, $604 = 0.0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0;
 var $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0.0, $638 = 0;
 var $639 = 0.0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0.0, $648 = 0.0, $649 = 0.0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0;
 var $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0.0, $661 = 0, $662 = 0.0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0.0, $67 = 0, $670 = 0.0, $671 = 0.0, $672 = 0, $673 = 0, $674 = 0;
 var $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0.0, $692 = 0;
 var $693 = 0.0, $694 = 0, $695 = 0, $696 = 0.0, $697 = 0.0, $698 = 0.0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0;
 var $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0.0, $721 = 0, $722 = 0.0, $723 = 0, $724 = 0, $725 = 0.0, $726 = 0.0, $727 = 0.0, $728 = 0;
 var $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0.0, $738 = 0, $739 = 0.0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0.0, $745 = 0.0, $746 = 0.0;
 var $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0.0, $761 = 0, $762 = 0.0, $763 = 0.0, $764 = 0;
 var $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0;
 var $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0;
 var $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0;
 var $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0;
 var $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0.0, $846 = 0.0, $847 = 0.0, $848 = 0.0, $849 = 0.0, $85 = 0, $850 = 0.0, $851 = 0.0, $852 = 0.0, $853 = 0.0, $854 = 0.0;
 var $855 = 0.0, $856 = 0.0, $857 = 0.0, $858 = 0.0, $859 = 0.0, $86 = 0, $860 = 0.0, $861 = 0, $862 = 0.0, $863 = 0.0, $864 = 0.0, $865 = 0.0, $866 = 0.0, $867 = 0.0, $868 = 0.0, $869 = 0.0, $87 = 0, $870 = 0, $871 = 0, $872 = 0.0;
 var $873 = 0.0, $874 = 0.0, $875 = 0.0, $876 = 0.0, $877 = 0.0, $878 = 0.0, $879 = 0.0, $88 = 0, $880 = 0, $881 = 0, $882 = 0.0, $883 = 0.0, $884 = 0.0, $885 = 0.0, $886 = 0.0, $887 = 0.0, $888 = 0.0, $889 = 0.0, $89 = 0, $890 = 0;
 var $891 = 0, $892 = 0.0, $893 = 0.0, $894 = 0.0, $895 = 0.0, $896 = 0.0, $897 = 0.0, $898 = 0.0, $899 = 0.0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0.0, $903 = 0.0, $904 = 0.0, $905 = 0.0, $906 = 0.0, $907 = 0.0, $908 = 0.0;
 var $909 = 0.0, $91 = 0, $910 = 0, $911 = 0, $912 = 0.0, $913 = 0.0, $914 = 0.0, $915 = 0.0, $916 = 0.0, $917 = 0.0, $918 = 0.0, $919 = 0.0, $92 = 0, $920 = 0, $921 = 0, $922 = 0.0, $923 = 0.0, $924 = 0.0, $925 = 0.0, $926 = 0.0;
 var $927 = 0.0, $928 = 0.0, $929 = 0.0, $93 = 0, $930 = 0, $931 = 0, $932 = 0.0, $933 = 0.0, $934 = 0.0, $935 = 0.0, $936 = 0.0, $937 = 0.0, $938 = 0.0, $939 = 0.0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0;
 var $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0;
 var $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0;
 var $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0;
 var $ch$018$i = 0, $ch$054 = 0, $ch$114$i = 0, $clip$060 = 0, $clip$060$lcssa = 0, $clip$060$lcssa160 = 0, $clip$1$lcssa = 0, $clip$158$us = 0, $clip$2 = 0, $clip$2$1 = 0, $clip$2$10 = 0, $clip$2$11 = 0, $clip$2$12 = 0, $clip$2$13 = 0, $clip$2$14 = 0, $clip$2$15 = 0, $clip$2$16 = 0, $clip$2$17 = 0, $clip$2$2 = 0, $clip$2$3 = 0;
 var $clip$2$4 = 0, $clip$2$5 = 0, $clip$2$6 = 0, $clip$2$7 = 0, $clip$2$8 = 0, $clip$2$9 = 0, $clip$2$us = 0, $do_l$023$i = 0, $exitcond$i13 = 0, $exitcond22$i = 0, $exitcond39$i = 0, $exitcond42$i = 0, $exitcond43$i = 0, $exitcond44$i = 0, $exitcond85 = 0, $fold = 0, $fold$i = 0, $fold45$i = 0, $fold46$i = 0, $fold47$i = 0;
 var $fold48$i = 0, $fold49$i = 0, $fold50$i = 0, $fold52$i = 0, $fold53$i = 0, $fold65 = 0, $fold93 = 0, $fold94 = 0, $fold95 = 0, $fold96 = 0, $gr$016$i = 0, $gr$061 = 0, $i$045 = 0, $i3$050 = 0, $i4$046 = 0, $i_stereo$0 = 0, $idx$014$i = 0, $idx$119$i = 0, $idx14$0$lcssa$i = 0, $idx14$035$i = 0;
 var $idx14$129$i = 0, $idx14$2$i = 0, $idx14$325$i = 0, $idx5$011$i = 0, $idx5$110$i = 0, $idx5$2$i = 0, $in0$051 = 0, $in05$047 = 0, $in1$052 = 0, $in16$048 = 0, $lwin$024$i = 0, $ms_stereo$0 = 0, $or$cond$i = 0, $or$cond20$i = 0, $or$cond5 = 0, $part2bits$0 = 0, $part2bits2$0 = 0, $r0c$0$i = 0, $rawout1$0$i = 0, $rawout1$112$i = 0;
 var $rawout1$219$i = 0, $rawout1$3$ph$i = 0, $rawout1$38$i = 0, $rawout2$0$i = 0, $rawout2$113$i = 0, $rawout2$220$i = 0, $rawout2$3$ph$i = 0, $rawout2$39$i = 0, $sb$0$i = 0, $sb$015$i = 0, $sb$06$i = 0, $sb$114$i = 0, $sb$120$i = 0, $sb$221$i = 0, $sb$3$ph$i = 0, $sb$310$i = 0, $sb15$030$i = 0, $sb19$026$i = 0, $sb6$09$i = 0, $sblim$08$i = 0;
 var $scalefacs = 0, $scevgep$i = 0, $scevgep38$i = 0, $sfb$017$i = 0, $sfb12$034$i = 0, $sfb4$012$i = 0, $sideinfo = 0, $single$0 = 0, $ss$059$us = 0, $stereo1$0 = 0, $storemerge$i = 0, $storemerge54$i = 0, $tspnt$0$i = 0, $tspnt$115$i = 0, $tspnt$222$i = 0, $tspnt$3$ph$i = 0, $tspnt$311$i = 0, $vararg_buffer = 0, $vararg_buffer11 = 0, $vararg_buffer14 = 0;
 var $vararg_buffer2 = 0, $vararg_buffer5 = 0, $vararg_buffer8 = 0, $vararg_ptr1 = 0, $xr1$07$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 736|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer14 = sp + 40|0;
 $vararg_buffer11 = sp + 32|0;
 $vararg_buffer8 = sp + 24|0;
 $vararg_buffer5 = sp + 16|0;
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $scalefacs = sp + 424|0;
 $sideinfo = sp + 48|0;
 $0 = ((($fr)) + 19652|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 19660|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($fr)) + 19704|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($1|0)==(1);
 $7 = ($3|0)==(-1);
 $$ = $7 ? 2 : 1;
 $stereo1$0 = $6 ? 1 : $$;
 $single$0 = $6 ? 0 : $3;
 $8 = ((($fr)) + 19716|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)==(1);
 if ($10) {
  $11 = ((($fr)) + 19720|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = $12 >>> 1;
  $14 = $13 & 1;
  $15 = $12 & 1;
  $i_stereo$0 = $15;$ms_stereo$0 = $14;
 } else {
  $i_stereo$0 = 0;$ms_stereo$0 = 0;
 }
 $16 = ((($fr)) + 19672|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ($17|0)!=(0);
 $19 = $18 ? 1 : 2;
 $20 = ($single$0|0)==(3);
 $21 = $20 ? 4 : 0;
 $22 = (36504 + (($17*20)|0)|0);
 $23 = (((36504 + (($17*20)|0)|0)) + 4|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($fr)) + 19788|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = HEAP8[$26>>0]|0;
 $28 = $27&255;
 $29 = $28 << 8;
 $30 = ((($26)) + 1|0);
 $31 = HEAP8[$30>>0]|0;
 $32 = $31&255;
 $33 = $29 | $32;
 $34 = $33 << 8;
 $35 = ((($26)) + 2|0);
 $36 = HEAP8[$35>>0]|0;
 $37 = $36&255;
 $38 = $34 | $37;
 $39 = ((($fr)) + 19784|0);
 $40 = HEAP32[$39>>2]|0;
 $41 = $38 << $40;
 $42 = $41 & 16777215;
 $43 = (($40) + ($24))|0;
 $44 = (24 - ($24))|0;
 $45 = $42 >>> $44;
 $46 = $43 >> 3;
 $47 = (($26) + ($46)|0);
 HEAP32[$25>>2] = $47;
 $48 = $43 & 7;
 HEAP32[$39>>2] = $48;
 HEAP32[$sideinfo>>2] = $45;
 $49 = ((($fr)) + 19868|0);
 $50 = HEAP32[$49>>2]|0;
 $51 = ($45>>>0)>($50>>>0);
 if ($51) {
  $52 = ((($fr)) + 27904|0);
  $53 = HEAP32[$52>>2]|0;
  $54 = ($53|0)==(0);
  if ($54) {
   $55 = ((($fr)) + 28068|0);
   $56 = HEAP32[$55>>2]|0;
   $57 = $56 & 32;
   $58 = ($57|0)==(0);
   if ($58) {
    $59 = ((($fr)) + 28064|0);
    $60 = HEAP32[$59>>2]|0;
    $61 = ($60|0)>(1);
    if ($61) {
     $62 = HEAP32[37544]|0;
     $63 = (($45) - ($50))|0;
     $64 = ((($fr)) + 19748|0);
     $65 = HEAP32[$64>>2]|0;
     HEAP32[$vararg_buffer>>2] = $63;
     $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
     HEAP32[$vararg_ptr1>>2] = $65;
     (_fprintf($62,511447,$vararg_buffer)|0);
     $$pre$i = HEAP32[$39>>2]|0;
     $$pre24$i = HEAP32[$25>>2]|0;
     $$pre25$i = HEAP32[$16>>2]|0;
     $$pre26$i = HEAP32[$49>>2]|0;
     $67 = $$pre$i;$70 = $$pre24$i;$73 = $$pre25$i;$75 = $$pre26$i;
    } else {
     $67 = $48;$70 = $47;$73 = $17;$75 = $50;
    }
   } else {
    $67 = $48;$70 = $47;$73 = $17;$75 = $50;
   }
  } else {
   $67 = $48;$70 = $47;$73 = $17;$75 = $50;
  }
  $66 = (($67) - ($24))|0;
  $68 = $66 >> 3;
  $69 = (($70) + ($68)|0);
  HEAP32[$25>>2] = $69;
  $71 = $66 & 7;
  HEAP32[$39>>2] = $71;
  $72 = ($73|0)==(0);
  if ($72) {
   $74 = $75 >>> 1;
   $76 = $74&255;
   HEAP8[$69>>0] = $76;
   $77 = HEAP32[$49>>2]|0;
   $78 = $77 << 7;
   $79 = $78&255;
   $80 = HEAP32[$25>>2]|0;
   $81 = ((($80)) + 1|0);
   HEAP8[$81>>0] = $79;
  } else {
   $82 = $75&255;
   HEAP8[$69>>0] = $82;
  }
  $83 = HEAP32[$25>>2]|0;
  $84 = ((($83)) + 2|0);
  $85 = ((($fr)) + 19864|0);
  $86 = HEAP32[$85>>2]|0;
  $87 = (($86) + -2)|0;
  _memset(($84|0),0,($87|0))|0;
  $88 = HEAP32[$25>>2]|0;
  $89 = HEAP8[$88>>0]|0;
  $90 = $89&255;
  $91 = $90 << 8;
  $92 = ((($88)) + 1|0);
  $93 = HEAP8[$92>>0]|0;
  $94 = $93&255;
  $95 = $91 | $94;
  $96 = $95 << 8;
  $97 = ((($88)) + 2|0);
  $98 = HEAP8[$97>>0]|0;
  $99 = $98&255;
  $100 = $96 | $99;
  $101 = HEAP32[$39>>2]|0;
  $102 = $100 << $101;
  $103 = $102 & 16777215;
  $104 = (($101) + ($24))|0;
  $105 = $103 >>> $44;
  $106 = $104 >> 3;
  $107 = (($88) + ($106)|0);
  HEAP32[$25>>2] = $107;
  $108 = $104 & 7;
  HEAP32[$39>>2] = $108;
  HEAP32[$sideinfo>>2] = $105;
  $$pre27$i = HEAP32[$49>>2]|0;
  $$pre28$i = HEAP32[$16>>2]|0;
  $$pre$phi$iZ2D = $85;$112 = $$pre27$i;$120 = $$pre28$i;$124 = $107;$127 = $108;
 } else {
  $$pre44$i = ((($fr)) + 19864|0);
  $$pre$phi$iZ2D = $$pre44$i;$112 = $50;$120 = $17;$124 = $47;$127 = $48;
 }
 $109 = ((($fr)) + 19736|0);
 $110 = HEAP32[$109>>2]|0;
 $111 = (($110) + ($112))|0;
 $113 = HEAP32[$$pre$phi$iZ2D>>2]|0;
 $114 = (($111) - ($113))|0;
 $115 = ((($fr)) + 19696|0);
 $116 = HEAP32[$115>>2]|0;
 $117 = ($116|0)!=(0);
 $$neg9$i = $117 ? -2 : 0;
 $118 = (($114) + ($$neg9$i))|0;
 $119 = ($120|0)==(0);
 $121 = $119 ? 511 : 255;
 $122 = ($118>>>0)>($121>>>0);
 $$$i = $122 ? $121 : $118;
 HEAP32[$49>>2] = $$$i;
 $123 = HEAP8[$124>>0]|0;
 $125 = $123&255;
 $126 = $125 << $127;
 $128 = $126 & 255;
 $129 = ((($fr)) + 19792|0);
 HEAP32[$129>>2] = $128;
 $130 = ((($124)) + 1|0);
 $131 = HEAP8[$130>>0]|0;
 $132 = $131&255;
 $133 = $132 << $127;
 $134 = $133 >>> 8;
 $135 = $134 | $128;
 $136 = (((36504 + (($17*20)|0)|0)) + 12|0);
 $137 = (((36504 + (($17*20)|0)|0)) + 8|0);
 $$sink59$i$in = $6 ? $137 : $136;
 $$sink59$i = HEAP32[$$sink59$i$in>>2]|0;
 $138 = $135 << $$sink59$i;
 $139 = $138 >>> 8;
 HEAP32[$129>>2] = $139;
 $140 = (($$sink59$i) + ($127))|0;
 $$pn$i = $140 >> 3;
 $storemerge54$i = (($124) + ($$pn$i)|0);
 HEAP32[$25>>2] = $storemerge54$i;
 $storemerge$i = $140 & 7;
 HEAP32[$39>>2] = $storemerge$i;
 $141 = ((($sideinfo)) + 4|0);
 HEAP32[$141>>2] = $139;
 $142 = ($1|0)>(0);
 $or$cond20$i = $142 & $119;
 if ($or$cond20$i) {
  $155 = $storemerge54$i;$158 = $storemerge$i;$ch$018$i = 0;
  while(1) {
   $153 = (((($sideinfo)) + 8|0) + (($ch$018$i*184)|0)|0);
   HEAP32[$153>>2] = -1;
   $154 = HEAP8[$155>>0]|0;
   $156 = $154&255;
   $157 = $156 << $158;
   $159 = $157 & 255;
   HEAP32[$129>>2] = $159;
   $160 = ((($155)) + 1|0);
   $161 = HEAP8[$160>>0]|0;
   $162 = $161&255;
   $163 = $162 << $158;
   $164 = $163 >>> 8;
   $165 = $164 | $159;
   $166 = $165 >>> 4;
   HEAP32[$129>>2] = $166;
   $167 = (($158) + 4)|0;
   $168 = $167 >>> 3;
   $169 = (($155) + ($168)|0);
   HEAP32[$25>>2] = $169;
   $170 = $167 & 7;
   HEAP32[$39>>2] = $170;
   $171 = (((((($sideinfo)) + 8|0) + (($ch$018$i*184)|0)|0)) + 92|0);
   HEAP32[$171>>2] = $166;
   $172 = (($ch$018$i) + 1)|0;
   $exitcond22$i = ($172|0)==($1|0);
   if ($exitcond22$i) {
    $$pre31$i88 = $169;$$pre32$i91 = $170;
    break;
   } else {
    $155 = $169;$158 = $170;$ch$018$i = $172;
   }
  }
 } else {
  $$pre31$i88 = $storemerge54$i;$$pre32$i91 = $storemerge$i;
 }
 $143 = $17 | 1;
 $144 = ($143|0)==(1);
 L21: do {
  if ($144) {
   $145 = ((($fr)) + 28068|0);
   $146 = HEAP32[37544]|0;
   $147 = ((($fr)) + 11952|0);
   $148 = ($ms_stereo$0|0)==(0);
   $149 = (((36504 + (($17*20)|0)|0)) + 16|0);
   $150 = ((($fr)) + 19796|0);
   $151 = ((($fr)) + 19676|0);
   $152 = (((161928 + (($5*110)|0)|0)) + 44|0);
   $$pre31$i = $$pre31$i88;$$pre32$i = $$pre32$i91;$gr$016$i = 0;
   L23: while(1) {
    if ($142) {
     $173 = HEAP32[$149>>2]|0;
     $174 = (24 - ($173))|0;
     $176 = $$pre31$i;$189 = $$pre32$i;$ch$114$i = 0;
     while(1) {
      $175 = HEAP8[$176>>0]|0;
      $177 = $175&255;
      $178 = $177 << 8;
      $179 = ((($176)) + 1|0);
      $180 = HEAP8[$179>>0]|0;
      $181 = $180&255;
      $182 = $178 | $181;
      $183 = $182 << 8;
      $184 = ((($176)) + 2|0);
      $185 = HEAP8[$184>>0]|0;
      $186 = $185&255;
      $187 = $183 | $186;
      $188 = $187 << $189;
      $190 = (($189) + 12)|0;
      $191 = $188 >>> 12;
      $192 = $191 & 4095;
      $193 = $190 >> 3;
      $194 = (($176) + ($193)|0);
      HEAP32[$25>>2] = $194;
      $195 = $190 & 7;
      HEAP32[$39>>2] = $195;
      $196 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 4|0);
      HEAP32[$196>>2] = $192;
      $197 = HEAP8[$194>>0]|0;
      $198 = $197&255;
      $199 = $198 << 8;
      $200 = ((($194)) + 1|0);
      $201 = HEAP8[$200>>0]|0;
      $202 = $201&255;
      $203 = $199 | $202;
      $204 = $203 << 8;
      $205 = ((($194)) + 2|0);
      $206 = HEAP8[$205>>0]|0;
      $207 = $206&255;
      $208 = $204 | $207;
      $209 = $208 << $195;
      $210 = (($195) + 9)|0;
      $211 = $209 >>> 15;
      $212 = $211 & 511;
      $213 = $210 >>> 3;
      $214 = (($194) + ($213)|0);
      HEAP32[$25>>2] = $214;
      $fold$i = (($189) + 5)|0;
      $215 = $fold$i & 7;
      HEAP32[$39>>2] = $215;
      $216 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 8|0);
      HEAP32[$216>>2] = $212;
      $217 = ($212>>>0)>(288);
      if ($217) {
       $218 = HEAP32[$145>>2]|0;
       $219 = $218 & 32;
       $220 = ($219|0)==(0);
       if ($220) {
        HEAP32[$vararg_buffer2>>2] = 454;
        (_fprintf($146,511502,$vararg_buffer2)|0);
        $$pre33$pre$i = HEAP32[$25>>2]|0;
        $$pre34$pre$i = HEAP32[$39>>2]|0;
        $$pre33$i = $$pre33$pre$i;$$pre34$i = $$pre34$pre$i;
       } else {
        $$pre33$i = $214;$$pre34$i = $215;
       }
       HEAP32[$216>>2] = 288;
       $222 = $$pre33$i;$225 = $$pre34$i;
      } else {
       $222 = $214;$225 = $215;
      }
      $221 = HEAP8[$222>>0]|0;
      $223 = $221&255;
      $224 = $223 << $225;
      $226 = $224 & 255;
      HEAP32[$129>>2] = $226;
      $227 = ((($222)) + 1|0);
      $228 = HEAP8[$227>>0]|0;
      $229 = $228&255;
      $230 = $229 << $225;
      $231 = $226 << 8;
      $232 = $230 | $231;
      $233 = $232 >>> 8;
      HEAP32[$129>>2] = $233;
      $234 = (($225) + 8)|0;
      $235 = $234 >> 3;
      $236 = (($222) + ($235)|0);
      HEAP32[$25>>2] = $236;
      $237 = $234 & 7;
      HEAP32[$39>>2] = $237;
      $238 = (0 - ($233))|0;
      $239 = (($147) + ($238<<2)|0);
      $240 = (($239) + ($21<<2)|0);
      $241 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 88|0);
      $242 = ((($240)) + 8|0);
      $$4$i = $148 ? $240 : $242;
      HEAP32[$241>>2] = $$4$i;
      $243 = HEAP8[$236>>0]|0;
      $244 = $243&255;
      $245 = $244 << 8;
      $246 = ((($236)) + 1|0);
      $247 = HEAP8[$246>>0]|0;
      $248 = $247&255;
      $249 = $245 | $248;
      $250 = $249 << 8;
      $251 = ((($236)) + 2|0);
      $252 = HEAP8[$251>>0]|0;
      $253 = $252&255;
      $254 = $250 | $253;
      $255 = $254 << $237;
      $256 = $255 & 16777215;
      $257 = (($237) + ($173))|0;
      $258 = $256 >>> $174;
      $259 = $257 >> 3;
      $260 = (($236) + ($259)|0);
      HEAP32[$25>>2] = $260;
      $fold45$i = (($234) + ($173))|0;
      $261 = $fold45$i & 7;
      HEAP32[$39>>2] = $261;
      $262 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 12|0);
      HEAP32[$262>>2] = $258;
      $263 = HEAP8[$260>>0]|0;
      $264 = $263&255;
      $265 = $264 << $261;
      $266 = $265&255;
      HEAP8[$150>>0] = $266;
      $267 = (($261) + 1)|0;
      $268 = $267 >>> 3;
      $269 = (($260) + ($268)|0);
      HEAP32[$25>>2] = $269;
      $fold46$i = (($fold45$i) + 1)|0;
      $270 = $fold46$i & 7;
      HEAP32[$39>>2] = $270;
      $271 = ($266<<24>>24)<(0);
      $272 = HEAP8[$269>>0]|0;
      $273 = $272&255;
      $274 = $273 << $270;
      $275 = $274 & 255;
      HEAP32[$129>>2] = $275;
      $276 = ((($269)) + 1|0);
      $277 = HEAP8[$276>>0]|0;
      $278 = $277&255;
      $279 = $278 << $270;
      $280 = $279 >>> 8;
      $281 = $280 | $275;
      do {
       if ($271) {
        $358 = $281 >>> 6;
        HEAP32[$129>>2] = $358;
        $359 = (($270) + 2)|0;
        $360 = $359 >>> 3;
        $361 = (($269) + ($360)|0);
        HEAP32[$25>>2] = $361;
        $fold49$i = (($fold45$i) + 3)|0;
        $362 = $fold49$i & 7;
        HEAP32[$39>>2] = $362;
        $363 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 16|0);
        HEAP32[$363>>2] = $358;
        $364 = HEAP8[$361>>0]|0;
        $365 = $364&255;
        $366 = $365 << $362;
        $367 = $366&255;
        HEAP8[$150>>0] = $367;
        $368 = (($362) + 1)|0;
        $369 = $368 >>> 3;
        $370 = (($361) + ($369)|0);
        HEAP32[$25>>2] = $370;
        $fold50$i = (($fold45$i) + 4)|0;
        $371 = $fold50$i & 7;
        HEAP32[$39>>2] = $371;
        $372 = $366 >>> 7;
        $373 = $372 & 1;
        $374 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 20|0);
        HEAP32[$374>>2] = $373;
        $375 = HEAP8[$370>>0]|0;
        $376 = $375&255;
        $377 = $376 << $371;
        $378 = $377 & 255;
        HEAP32[$129>>2] = $378;
        $379 = ((($370)) + 1|0);
        $380 = HEAP8[$379>>0]|0;
        $381 = $380&255;
        $382 = $381 << $371;
        $383 = $382 >>> 8;
        $384 = $383 | $378;
        $385 = $384 >>> 3;
        HEAP32[$129>>2] = $385;
        $386 = (($371) + 5)|0;
        $387 = $386 >>> 3;
        $388 = (($370) + ($387)|0);
        HEAP32[$25>>2] = $388;
        HEAP32[$39>>2] = $270;
        $389 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 24|0);
        HEAP32[$389>>2] = $385;
        $390 = HEAP8[$388>>0]|0;
        $391 = $390&255;
        $392 = $391 << $270;
        $393 = $392 & 255;
        HEAP32[$129>>2] = $393;
        $394 = ((($388)) + 1|0);
        $395 = HEAP8[$394>>0]|0;
        $396 = $395&255;
        $397 = $396 << $270;
        $398 = $397 >>> 8;
        $399 = $398 | $393;
        $400 = $399 >>> 3;
        HEAP32[$129>>2] = $400;
        $401 = (($270) + 5)|0;
        $402 = $401 >>> 3;
        $403 = (($388) + ($402)|0);
        HEAP32[$25>>2] = $403;
        $fold = (($fold45$i) + 6)|0;
        $404 = $fold & 7;
        HEAP32[$39>>2] = $404;
        $405 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 28|0);
        HEAP32[$405>>2] = $400;
        $406 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 32|0);
        HEAP32[$406>>2] = 0;
        $407 = HEAP8[$403>>0]|0;
        $408 = $407&255;
        $409 = $408 << $404;
        $410 = $409 & 255;
        HEAP32[$129>>2] = $410;
        $411 = ((($403)) + 1|0);
        $412 = HEAP8[$411>>0]|0;
        $413 = $412&255;
        $414 = $413 << $404;
        $415 = $414 >>> 8;
        $416 = $415 | $410;
        $417 = $416 >>> 5;
        HEAP32[$129>>2] = $417;
        $418 = (($404) + 3)|0;
        $419 = $418 >>> 3;
        $420 = (($403) + ($419)|0);
        HEAP32[$25>>2] = $420;
        $fold96 = (($fold45$i) + 1)|0;
        $421 = $fold96 & 7;
        HEAP32[$39>>2] = $421;
        $422 = $417 << 3;
        $423 = (($$4$i) + ($422<<2)|0);
        $424 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 76|0);
        HEAP32[$424>>2] = $423;
        $425 = HEAP8[$420>>0]|0;
        $426 = $425&255;
        $427 = $426 << $421;
        $428 = $427 & 255;
        HEAP32[$129>>2] = $428;
        $429 = ((($420)) + 1|0);
        $430 = HEAP8[$429>>0]|0;
        $431 = $430&255;
        $432 = $431 << $421;
        $433 = $432 >>> 8;
        $434 = $433 | $428;
        $435 = $434 >>> 5;
        HEAP32[$129>>2] = $435;
        $436 = (($421) + 3)|0;
        $437 = $436 >>> 3;
        $438 = (($420) + ($437)|0);
        HEAP32[$25>>2] = $438;
        $fold52$i = (($fold45$i) + 4)|0;
        $439 = $fold52$i & 7;
        HEAP32[$39>>2] = $439;
        $440 = $435 << 3;
        $441 = (($$4$i) + ($440<<2)|0);
        $442 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 80|0);
        HEAP32[$442>>2] = $441;
        $443 = HEAP8[$438>>0]|0;
        $444 = $443&255;
        $445 = $444 << $439;
        $446 = $445 & 255;
        HEAP32[$129>>2] = $446;
        $447 = ((($438)) + 1|0);
        $448 = HEAP8[$447>>0]|0;
        $449 = $448&255;
        $450 = $449 << $439;
        $451 = $450 >>> 8;
        $452 = $451 | $446;
        $453 = $452 >>> 5;
        HEAP32[$129>>2] = $453;
        $454 = (($439) + 3)|0;
        $455 = $454 >>> 3;
        $456 = (($438) + ($455)|0);
        HEAP32[$25>>2] = $456;
        $fold53$i = (($fold45$i) + 7)|0;
        $457 = $fold53$i & 7;
        HEAP32[$39>>2] = $457;
        $458 = $453 << 3;
        $459 = (($$4$i) + ($458<<2)|0);
        $460 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 84|0);
        HEAP32[$460>>2] = $459;
        $461 = HEAP32[$363>>2]|0;
        $462 = ($461|0)==(0);
        if ($462) {
         break L23;
        }
        $467 = HEAP32[$16>>2]|0;
        $468 = ($467|0)==(0);
        $469 = ($461|0)==(2);
        $or$cond$i = $469 | $468;
        $470 = HEAP32[$151>>2]|0;
        $471 = ($470|0)==(0);
        if ($or$cond$i) {
         if ($471) {
          $472 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 56|0);
          HEAP32[$472>>2] = 18;
          $473 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 60|0);
          HEAP32[$473>>2] = 288;
          $496 = $467;$498 = $456;$501 = $457;
          break;
         }
        } else {
         if ($471) {
          $485 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 56|0);
          HEAP32[$485>>2] = 27;
          $486 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 60|0);
          HEAP32[$486>>2] = 288;
          $496 = $467;$498 = $456;$501 = $457;
          break;
         }
        }
        if ($469) {
         $474 = HEAP32[$374>>2]|0;
         $475 = ($474|0)==(0);
         if ($475) {
          $r0c$0$i = 6;
         } else {
          label = 34;
         }
        } else {
         label = 34;
        }
        if ((label|0) == 34) {
         label = 0;
         $r0c$0$i = 8;
        }
        $476 = ((161928 + (($5*110)|0)|0) + ($r0c$0$i<<1)|0);
        $477 = HEAP16[$476>>1]|0;
        $478 = $477&65535;
        $479 = $478 >>> 1;
        $480 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 56|0);
        HEAP32[$480>>2] = $479;
        $481 = HEAP16[$152>>1]|0;
        $482 = $481&65535;
        $483 = $482 >>> 1;
        $484 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 60|0);
        HEAP32[$484>>2] = $483;
        $496 = $467;$498 = $456;$501 = $457;
       } else {
        $282 = $281 >>> 3;
        HEAP32[$129>>2] = $282;
        $283 = (($270) + 5)|0;
        $284 = $283 >>> 3;
        $285 = (($269) + ($284)|0);
        HEAP32[$25>>2] = $285;
        $fold93 = (($fold45$i) + 6)|0;
        $286 = $fold93 & 7;
        HEAP32[$39>>2] = $286;
        $287 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 24|0);
        HEAP32[$287>>2] = $282;
        $288 = HEAP8[$285>>0]|0;
        $289 = $288&255;
        $290 = $289 << $286;
        $291 = $290 & 255;
        HEAP32[$129>>2] = $291;
        $292 = ((($285)) + 1|0);
        $293 = HEAP8[$292>>0]|0;
        $294 = $293&255;
        $295 = $294 << $286;
        $296 = $295 >>> 8;
        $297 = $296 | $291;
        $298 = $297 >>> 3;
        HEAP32[$129>>2] = $298;
        $299 = (($286) + 5)|0;
        $300 = $299 >>> 3;
        $301 = (($285) + ($300)|0);
        HEAP32[$25>>2] = $301;
        $fold94 = (($fold45$i) + 3)|0;
        $302 = $fold94 & 7;
        HEAP32[$39>>2] = $302;
        $303 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 28|0);
        HEAP32[$303>>2] = $298;
        $304 = HEAP8[$301>>0]|0;
        $305 = $304&255;
        $306 = $305 << $302;
        $307 = $306 & 255;
        HEAP32[$129>>2] = $307;
        $308 = ((($301)) + 1|0);
        $309 = HEAP8[$308>>0]|0;
        $310 = $309&255;
        $311 = $310 << $302;
        $312 = $311 >>> 8;
        $313 = $312 | $307;
        $314 = $313 >>> 3;
        HEAP32[$129>>2] = $314;
        $315 = (($302) + 5)|0;
        $316 = $315 >>> 3;
        $317 = (($301) + ($316)|0);
        HEAP32[$25>>2] = $317;
        $fold95 = (($234) + ($173))|0;
        $318 = $fold95 & 7;
        HEAP32[$39>>2] = $318;
        $319 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 32|0);
        HEAP32[$319>>2] = $314;
        $320 = HEAP8[$317>>0]|0;
        $321 = $320&255;
        $322 = $321 << $318;
        $323 = $322 & 255;
        HEAP32[$129>>2] = $323;
        $324 = ((($317)) + 1|0);
        $325 = HEAP8[$324>>0]|0;
        $326 = $325&255;
        $327 = $326 << $318;
        $328 = $327 >>> 8;
        $329 = $328 | $323;
        $330 = $329 >>> 4;
        HEAP32[$129>>2] = $330;
        $331 = (($318) + 4)|0;
        $332 = $331 >>> 3;
        $333 = (($317) + ($332)|0);
        HEAP32[$25>>2] = $333;
        $fold65 = (($fold45$i) + 4)|0;
        $334 = $fold65 & 7;
        HEAP32[$39>>2] = $334;
        $335 = HEAP8[$333>>0]|0;
        $336 = $335&255;
        $337 = $336 << $334;
        $338 = $337 & 255;
        HEAP32[$129>>2] = $338;
        $339 = ((($333)) + 1|0);
        $340 = HEAP8[$339>>0]|0;
        $341 = $340&255;
        $342 = $341 << $334;
        $343 = $342 >>> 8;
        $344 = $343 | $338;
        $345 = $344 >>> 5;
        HEAP32[$129>>2] = $345;
        $346 = (($334) + 3)|0;
        $347 = $346 >>> 3;
        $348 = (($333) + ($347)|0);
        HEAP32[$25>>2] = $348;
        $fold47$i = (($fold45$i) + 7)|0;
        $349 = $fold47$i & 7;
        HEAP32[$39>>2] = $349;
        $350 = (($330) + 1)|0;
        $351 = ((161928 + (($5*110)|0)|0) + ($350<<1)|0);
        $352 = HEAP16[$351>>1]|0;
        $353 = $352&65535;
        $354 = $353 >>> 1;
        $355 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 56|0);
        HEAP32[$355>>2] = $354;
        $356 = (($345) + ($350))|0;
        $357 = ($356>>>0)>(21);
        if ($357) {
         $$sink23$i = 288;
        } else {
         $487 = (($356) + 1)|0;
         $488 = ((161928 + (($5*110)|0)|0) + ($487<<1)|0);
         $489 = HEAP16[$488>>1]|0;
         $490 = $489&65535;
         $491 = $490 >>> 1;
         $$sink23$i = $491;
        }
        $492 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 60|0);
        HEAP32[$492>>2] = $$sink23$i;
        $493 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 16|0);
        HEAP32[$493>>2] = 0;
        $494 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 20|0);
        HEAP32[$494>>2] = 0;
        $$pre38$i = HEAP32[$16>>2]|0;
        $496 = $$pre38$i;$498 = $348;$501 = $349;
       }
      } while(0);
      $495 = ($496|0)==(0);
      if ($495) {
       $497 = HEAP8[$498>>0]|0;
       $499 = $497&255;
       $500 = $499 << $501;
       $502 = $500&255;
       HEAP8[$150>>0] = $502;
       $503 = (($501) + 1)|0;
       $504 = $503 >>> 3;
       $505 = (($498) + ($504)|0);
       HEAP32[$25>>2] = $505;
       $506 = $503 & 7;
       HEAP32[$39>>2] = $506;
       $507 = $500 >>> 7;
       $508 = $507 & 1;
       $509 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 64|0);
       HEAP32[$509>>2] = $508;
       $511 = $505;$514 = $506;
      } else {
       $511 = $498;$514 = $501;
      }
      $510 = HEAP8[$511>>0]|0;
      $512 = $510&255;
      $513 = $512 << $514;
      $515 = $513&255;
      HEAP8[$150>>0] = $515;
      $516 = (($514) + 1)|0;
      $517 = $516 >> 3;
      $518 = (($511) + ($517)|0);
      HEAP32[$25>>2] = $518;
      $519 = $516 & 7;
      HEAP32[$39>>2] = $519;
      $520 = $513 >>> 7;
      $521 = $520 & 1;
      $522 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 68|0);
      HEAP32[$522>>2] = $521;
      $523 = HEAP8[$518>>0]|0;
      $524 = $523&255;
      $525 = $524 << $519;
      $526 = $525&255;
      HEAP8[$150>>0] = $526;
      $527 = (($519) + 1)|0;
      $528 = $527 >>> 3;
      $529 = (($518) + ($528)|0);
      HEAP32[$25>>2] = $529;
      $fold48$i = (($514) + 2)|0;
      $530 = $fold48$i & 7;
      HEAP32[$39>>2] = $530;
      $531 = $525 >>> 7;
      $532 = $531 & 1;
      $533 = ((((((($sideinfo)) + 8|0) + (($ch$114$i*184)|0)|0) + (($gr$016$i*92)|0)|0)) + 72|0);
      HEAP32[$533>>2] = $532;
      $534 = (($ch$114$i) + 1)|0;
      $535 = ($534|0)<($1|0);
      if ($535) {
       $176 = $529;$189 = $530;$ch$114$i = $534;
      } else {
       $$pre31$i87 = $529;$$pre32$i90 = $530;
       break;
      }
     }
    } else {
     $$pre31$i87 = $$pre31$i;$$pre32$i90 = $$pre32$i;
    }
    $536 = (($gr$016$i) + 1)|0;
    $537 = HEAP32[$22>>2]|0;
    $538 = ($536|0)<($537|0);
    if ($538) {
     $$pre31$i = $$pre31$i87;$$pre32$i = $$pre32$i90;$gr$016$i = $536;
    } else {
     break L21;
    }
   }
   $463 = HEAP32[$145>>2]|0;
   $464 = $463 & 32;
   $465 = ($464|0)==(0);
   if ($465) {
    $466 = HEAP32[37544]|0;
    HEAP32[$vararg_buffer5>>2] = 479;
    (_fprintf($466,511546,$vararg_buffer5)|0);
    $$pre = HEAP32[$145>>2]|0;
    $540 = $$pre;
   } else {
    $540 = $463;
   }
   $539 = $540 & 32;
   $541 = ($539|0)==(0);
   if (!($541)) {
    $$5 = 0;
    STACKTOP = sp;return ($$5|0);
   }
   $542 = HEAP32[37544]|0;
   HEAP32[$vararg_buffer8>>2] = 1923;
   (_fprintf($542,511622,$vararg_buffer8)|0);
   $$5 = 0;
   STACKTOP = sp;return ($$5|0);
  }
 } while(0);
 $543 = HEAP32[$sideinfo>>2]|0;
 _INT123_set_pointer($fr,$543);
 $544 = ((($fr)) + 28600|0);
 $545 = ((($fr)) + 28604|0);
 $546 = ($1|0)==(2);
 $547 = ((($scalefacs)) + 156|0);
 $548 = ($ms_stereo$0|0)==(0);
 $549 = ($i_stereo$0|0)==(0);
 $550 = $ms_stereo$0 | $i_stereo$0;
 $551 = ($550|0)!=(0);
 $or$cond5 = $20 | $551;
 $552 = (((161928 + (($5*110)|0)|0)) + 92|0);
 $553 = (((161928 + (($5*110)|0)|0)) + 108|0);
 $554 = ((($scalefacs)) + 236|0);
 $555 = (((161928 + (($5*110)|0)|0)) + 67|0);
 $556 = ($single$0|0)==(-1);
 $557 = ((($fr)) + 19640|0);
 $558 = ((($fr)) + 19644|0);
 $clip$060 = 0;$gr$061 = 0;
 L67: while(1) {
  $559 = HEAP32[$544>>2]|0;
  $560 = HEAP32[$545>>2]|0;
  $561 = (((($sideinfo)) + 8|0) + (($gr$061*92)|0)|0);
  $562 = HEAP32[$16>>2]|0;
  $563 = ($562|0)==(0);
  if ($563) {
   $565 = (_III_get_scale_factors_1($fr,$scalefacs,$561)|0);
   $part2bits$0 = $565;
  } else {
   $564 = (_III_get_scale_factors_2($fr,$scalefacs,$561,0)|0);
   $part2bits$0 = $564;
  }
  $566 = (_III_dequantize_sample($fr,$559,$scalefacs,$561,$5,$part2bits$0)|0);
  $567 = ($566|0)==(0);
  if (!($567)) {
   $clip$060$lcssa = $clip$060;
   label = 50;
   break;
  }
  L74: do {
   if ($546) {
    $576 = (((($sideinfo)) + 192|0) + (($gr$061*92)|0)|0);
    $577 = HEAP32[$16>>2]|0;
    $578 = ($577|0)==(0);
    if ($578) {
     $580 = (_III_get_scale_factors_1($fr,$547,$576)|0);
     $part2bits2$0 = $580;
    } else {
     $579 = (_III_get_scale_factors_2($fr,$547,$576,$i_stereo$0)|0);
     $part2bits2$0 = $579;
    }
    $581 = ((($559)) + 2304|0);
    $582 = (_III_dequantize_sample($fr,$581,$547,$576,$5,$part2bits2$0)|0);
    $583 = ($582|0)==(0);
    if (!($583)) {
     $clip$060$lcssa160 = $clip$060;
     label = 58;
     break L67;
    }
    if (!($548)) {
     $592 = (((((($sideinfo)) + 8|0) + (($gr$061*92)|0)|0)) + 52|0);
     $593 = HEAP32[$592>>2]|0;
     $594 = (((((($sideinfo)) + 192|0) + (($gr$061*92)|0)|0)) + 52|0);
     $595 = HEAP32[$594>>2]|0;
     $596 = ($595>>>0)>($593>>>0);
     $$9 = $596 ? $595 : $593;
     $597 = ($$9*18)|0;
     $598 = ($$9|0)>(0);
     if ($598) {
      $i$045 = 0;
      while(1) {
       $599 = (($559) + ($i$045<<2)|0);
       $600 = +HEAPF32[$599>>2];
       $601 = (((($559)) + 2304|0) + ($i$045<<2)|0);
       $602 = +HEAPF32[$601>>2];
       $603 = $602 + $600;
       HEAPF32[$599>>2] = $603;
       $604 = $600 - $602;
       HEAPF32[$601>>2] = $604;
       $605 = (($i$045) + 1)|0;
       $606 = ($605|0)<($597|0);
       if ($606) {
        $i$045 = $605;
       } else {
        break;
       }
      }
     }
    }
    L86: do {
     if (!($549)) {
      $607 = HEAP32[$16>>2]|0;
      $608 = (((((($sideinfo)) + 192|0) + (($gr$061*92)|0)|0)) + 12|0);
      $609 = HEAP32[$608>>2]|0;
      $610 = $609 & $607;
      $611 = (($610) + ($607))|0;
      $612 = ((36816 + ($611<<4)|0) + ($ms_stereo$0<<3)|0);
      $613 = HEAP32[$612>>2]|0;
      $614 = ((((36816 + ($611<<4)|0) + ($ms_stereo$0<<3)|0)) + 4|0);
      $615 = HEAP32[$614>>2]|0;
      $616 = (((((($sideinfo)) + 192|0) + (($gr$061*92)|0)|0)) + 16|0);
      $617 = HEAP32[$616>>2]|0;
      $618 = ($617|0)==(2);
      if (!($618)) {
       $705 = (((((($sideinfo)) + 192|0) + (($gr$061*92)|0)|0)) + 48|0);
       $706 = HEAP32[$705>>2]|0;
       $707 = ($706|0)>(21);
       if ($707) {
        break;
       }
       $708 = ((161928 + (($5*110)|0)|0) + ($706<<1)|0);
       $709 = HEAP16[$708>>1]|0;
       $710 = $709&65535;
       $711 = ($706|0)<(21);
       if ($711) {
        $712 = ((($559)) + 2304|0);
        $idx14$035$i = $710;$sfb12$034$i = $706;
        while(1) {
         $713 = ((((161928 + (($5*110)|0)|0)) + 46|0) + ($sfb12$034$i)|0);
         $714 = HEAP8[$713>>0]|0;
         $715 = $714&255;
         $716 = (((($scalefacs)) + 156|0) + ($sfb12$034$i<<2)|0);
         $717 = HEAP32[$716>>2]|0;
         $718 = ($717|0)==(7);
         do {
          if ($718) {
           $732 = (($715) + ($idx14$035$i))|0;
           $idx14$2$i = $732;
          } else {
           $719 = (($613) + ($717<<2)|0);
           $720 = +HEAPF32[$719>>2];
           $721 = (($615) + ($717<<2)|0);
           $722 = +HEAPF32[$721>>2];
           $723 = ($714<<24>>24)==(0);
           if ($723) {
            $idx14$2$i = $idx14$035$i;
            break;
           } else {
            $idx14$129$i = $idx14$035$i;$sb15$030$i = $715;
           }
           while(1) {
            $724 = (($559) + ($idx14$129$i<<2)|0);
            $725 = +HEAPF32[$724>>2];
            $726 = $725 * $720;
            HEAPF32[$724>>2] = $726;
            $727 = $725 * $722;
            $728 = (($712) + ($idx14$129$i<<2)|0);
            HEAPF32[$728>>2] = $727;
            $729 = (($sb15$030$i) + -1)|0;
            $730 = (($idx14$129$i) + 1)|0;
            $731 = ($sb15$030$i|0)>(1);
            if ($731) {
             $idx14$129$i = $730;$sb15$030$i = $729;
            } else {
             $idx14$2$i = $730;
             break;
            }
           }
          }
         } while(0);
         $733 = (($sfb12$034$i) + 1)|0;
         $exitcond44$i = ($733|0)==(21);
         if ($exitcond44$i) {
          $idx14$0$lcssa$i = $idx14$2$i;
          break;
         } else {
          $idx14$035$i = $idx14$2$i;$sfb12$034$i = $733;
         }
        }
       } else {
        $idx14$0$lcssa$i = $710;
       }
       $734 = HEAP32[$554>>2]|0;
       $735 = ($734|0)==(7);
       if ($735) {
        break;
       }
       $736 = (($613) + ($734<<2)|0);
       $737 = +HEAPF32[$736>>2];
       $738 = (($615) + ($734<<2)|0);
       $739 = +HEAPF32[$738>>2];
       $740 = HEAP8[$555>>0]|0;
       $741 = $740&255;
       $742 = ((($559)) + 2304|0);
       $idx14$325$i = $idx14$0$lcssa$i;$sb19$026$i = $741;
       while(1) {
        $743 = (($559) + ($idx14$325$i<<2)|0);
        $744 = +HEAPF32[$743>>2];
        $745 = $744 * $737;
        HEAPF32[$743>>2] = $745;
        $746 = $744 * $739;
        $747 = (($742) + ($idx14$325$i<<2)|0);
        HEAPF32[$747>>2] = $746;
        $748 = (($sb19$026$i) + -1)|0;
        $749 = (($idx14$325$i) + 1)|0;
        $750 = ($sb19$026$i|0)>(1);
        if ($750) {
         $idx14$325$i = $749;$sb19$026$i = $748;
        } else {
         break L86;
        }
       }
      }
      $619 = (((((($sideinfo)) + 192|0) + (($gr$061*92)|0)|0)) + 20|0);
      $620 = HEAP32[$619>>2]|0;
      $621 = ($620|0)!=(0);
      $$$i10 = $621&1;
      $622 = ((($559)) + 2304|0);
      $623 = (33 - ($620))|0;
      $do_l$023$i = $$$i10;$lwin$024$i = 0;
      while(1) {
       $624 = ((((((($sideinfo)) + 192|0) + (($gr$061*92)|0)|0)) + 36|0) + ($lwin$024$i<<2)|0);
       $625 = HEAP32[$624>>2]|0;
       $626 = ($625|0)>(3);
       $$do_l$0$i = $626 ? 0 : $do_l$023$i;
       $627 = ($625|0)<(12);
       if ($627) {
        $628 = (($lwin$024$i) - ($620))|0;
        $sfb$017$i = $625;
        while(1) {
         $629 = ($sfb$017$i*3)|0;
         $630 = (($628) + ($629))|0;
         $631 = (((($scalefacs)) + 156|0) + ($630<<2)|0);
         $632 = HEAP32[$631>>2]|0;
         $633 = ($632|0)==(7);
         do {
          if (!($633)) {
           $634 = ((((161928 + (($5*110)|0)|0)) + 96|0) + ($sfb$017$i)|0);
           $635 = HEAP8[$634>>0]|0;
           $636 = (($613) + ($632<<2)|0);
           $637 = +HEAPF32[$636>>2];
           $638 = (($615) + ($632<<2)|0);
           $639 = +HEAPF32[$638>>2];
           $640 = ($635<<24>>24)==(0);
           if ($640) {
            break;
           }
           $641 = ((((161928 + (($5*110)|0)|0)) + 68|0) + ($sfb$017$i<<1)|0);
           $642 = HEAP16[$641>>1]|0;
           $643 = $642&65535;
           $644 = (($643) + ($lwin$024$i))|0;
           $645 = $635&255;
           $idx$014$i = $644;$sb$015$i = $645;
           while(1) {
            $646 = (($559) + ($idx$014$i<<2)|0);
            $647 = +HEAPF32[$646>>2];
            $648 = $647 * $637;
            HEAPF32[$646>>2] = $648;
            $649 = $647 * $639;
            $650 = (($622) + ($idx$014$i<<2)|0);
            HEAPF32[$650>>2] = $649;
            $651 = (($sb$015$i) + -1)|0;
            $652 = (($idx$014$i) + 3)|0;
            $653 = ($sb$015$i|0)>(1);
            if ($653) {
             $idx$014$i = $652;$sb$015$i = $651;
            } else {
             break;
            }
           }
          }
         } while(0);
         $654 = (($sfb$017$i) + 1)|0;
         $exitcond42$i = ($654|0)==(12);
         if ($exitcond42$i) {
          break;
         } else {
          $sfb$017$i = $654;
         }
        }
       }
       $655 = (($623) + ($lwin$024$i))|0;
       $656 = (((($scalefacs)) + 156|0) + ($655<<2)|0);
       $657 = HEAP32[$656>>2]|0;
       $658 = ($657|0)==(7);
       if (!($658)) {
        $659 = (($613) + ($657<<2)|0);
        $660 = +HEAPF32[$659>>2];
        $661 = (($615) + ($657<<2)|0);
        $662 = +HEAPF32[$661>>2];
        $663 = HEAP8[$553>>0]|0;
        $664 = $663&255;
        $665 = HEAP16[$552>>1]|0;
        $666 = $665&65535;
        $667 = (($666) + ($lwin$024$i))|0;
        $idx$119$i = $667;$sb$120$i = $664;
        while(1) {
         $668 = (($559) + ($idx$119$i<<2)|0);
         $669 = +HEAPF32[$668>>2];
         $670 = $669 * $660;
         HEAPF32[$668>>2] = $670;
         $671 = $669 * $662;
         $672 = (($622) + ($idx$119$i<<2)|0);
         HEAPF32[$672>>2] = $671;
         $673 = (($sb$120$i) + -1)|0;
         $674 = (($idx$119$i) + 3)|0;
         $675 = ($sb$120$i|0)>(1);
         if ($675) {
          $idx$119$i = $674;$sb$120$i = $673;
         } else {
          break;
         }
        }
       }
       $676 = (($lwin$024$i) + 1)|0;
       $exitcond43$i = ($676|0)==(3);
       if ($exitcond43$i) {
        $$do_l$0$i$lcssa = $$do_l$0$i;
        break;
       } else {
        $do_l$023$i = $$do_l$0$i;$lwin$024$i = $676;
       }
      }
      $677 = ($$do_l$0$i$lcssa|0)==(0);
      if (!($677)) {
       $678 = (((((($sideinfo)) + 192|0) + (($gr$061*92)|0)|0)) + 48|0);
       $679 = HEAP32[$678>>2]|0;
       $680 = ($679|0)<(8);
       if ($680) {
        $681 = ((161928 + (($5*110)|0)|0) + ($679<<1)|0);
        $682 = HEAP16[$681>>1]|0;
        $683 = $682&65535;
        $idx5$011$i = $683;$sfb4$012$i = $679;
        while(1) {
         $684 = ((((161928 + (($5*110)|0)|0)) + 46|0) + ($sfb4$012$i)|0);
         $685 = HEAP8[$684>>0]|0;
         $686 = $685&255;
         $687 = (((($scalefacs)) + 156|0) + ($sfb4$012$i<<2)|0);
         $688 = HEAP32[$687>>2]|0;
         $689 = ($688|0)==(7);
         do {
          if ($689) {
           $703 = (($686) + ($idx5$011$i))|0;
           $idx5$2$i = $703;
          } else {
           $690 = (($613) + ($688<<2)|0);
           $691 = +HEAPF32[$690>>2];
           $692 = (($615) + ($688<<2)|0);
           $693 = +HEAPF32[$692>>2];
           $694 = ($685<<24>>24)==(0);
           if ($694) {
            $idx5$2$i = $idx5$011$i;
            break;
           } else {
            $idx5$110$i = $idx5$011$i;$sb6$09$i = $686;
           }
           while(1) {
            $695 = (($559) + ($idx5$110$i<<2)|0);
            $696 = +HEAPF32[$695>>2];
            $697 = $696 * $691;
            HEAPF32[$695>>2] = $697;
            $698 = $696 * $693;
            $699 = (($622) + ($idx5$110$i<<2)|0);
            HEAPF32[$699>>2] = $698;
            $700 = (($sb6$09$i) + -1)|0;
            $701 = (($idx5$110$i) + 1)|0;
            $702 = ($sb6$09$i|0)>(1);
            if ($702) {
             $idx5$110$i = $701;$sb6$09$i = $700;
            } else {
             $idx5$2$i = $701;
             break;
            }
           }
          }
         } while(0);
         $704 = (($sfb4$012$i) + 1)|0;
         $exitcond$i13 = ($704|0)==(8);
         if ($exitcond$i13) {
          break;
         } else {
          $idx5$011$i = $idx5$2$i;$sfb4$012$i = $704;
         }
        }
       }
      }
     }
    } while(0);
    do {
     if ($or$cond5) {
      $751 = (((((($sideinfo)) + 192|0) + (($gr$061*92)|0)|0)) + 52|0);
      $752 = HEAP32[$751>>2]|0;
      $753 = (((((($sideinfo)) + 8|0) + (($gr$061*92)|0)|0)) + 52|0);
      $754 = HEAP32[$753>>2]|0;
      $755 = ($752>>>0)>($754>>>0);
      if ($755) {
       HEAP32[$753>>2] = $752;
       break;
      } else {
       HEAP32[$751>>2] = $754;
       break;
      }
     }
    } while(0);
    switch ($single$0|0) {
    case 3:  {
     $756 = (((((($sideinfo)) + 192|0) + (($gr$061*92)|0)|0)) + 52|0);
     $757 = HEAP32[$756>>2]|0;
     $758 = ($757|0)>(0);
     if (!($758)) {
      $ch$054 = 0;
      break L74;
     }
     $759 = ($757*18)|0;
     $i3$050 = 0;$in0$051 = $559;$in1$052 = $581;
     while(1) {
      $760 = +HEAPF32[$in0$051>>2];
      $761 = ((($in1$052)) + 4|0);
      $762 = +HEAPF32[$in1$052>>2];
      $763 = $762 + $760;
      HEAPF32[$in0$051>>2] = $763;
      $764 = (($i3$050) + 1)|0;
      $765 = ((($in0$051)) + 4|0);
      $766 = ($764|0)<($759|0);
      if ($766) {
       $i3$050 = $764;$in0$051 = $765;$in1$052 = $761;
      } else {
       $ch$054 = 0;
       break;
      }
     }
     break;
    }
    case 1:  {
     $767 = (((((($sideinfo)) + 192|0) + (($gr$061*92)|0)|0)) + 52|0);
     $768 = HEAP32[$767>>2]|0;
     $769 = ($768|0)>(0);
     if (!($769)) {
      $ch$054 = 0;
      break L74;
     }
     $770 = ($768*18)|0;
     $i4$046 = 0;$in05$047 = $559;$in16$048 = $581;
     while(1) {
      $771 = ((($in16$048)) + 4|0);
      $772 = HEAP32[$in16$048>>2]|0;
      $773 = ((($in05$047)) + 4|0);
      HEAP32[$in05$047>>2] = $772;
      $774 = (($i4$046) + 1)|0;
      $775 = ($774|0)<($770|0);
      if ($775) {
       $i4$046 = $774;$in05$047 = $773;$in16$048 = $771;
      } else {
       $ch$054 = 0;
       break;
      }
     }
     break;
    }
    default: {
     $ch$054 = 0;
     break L74;
    }
    }
   } else {
    $ch$054 = 0;
   }
  } while(0);
  while(1) {
   $834 = ((((((($sideinfo)) + 8|0) + (($ch$054*184)|0)|0) + (($gr$061*92)|0)|0)) + 16|0);
   $835 = HEAP32[$834>>2]|0;
   $836 = ($835|0)==(2);
   if ($836) {
    $837 = ((((((($sideinfo)) + 8|0) + (($ch$054*184)|0)|0) + (($gr$061*92)|0)|0)) + 20|0);
    $838 = HEAP32[$837>>2]|0;
    $839 = ($838|0)==(0);
    if (!($839)) {
     $sblim$08$i = 1;
     label = 114;
    }
   } else {
    $840 = ((((((($sideinfo)) + 8|0) + (($ch$054*184)|0)|0) + (($gr$061*92)|0)|0)) + 52|0);
    $841 = HEAP32[$840>>2]|0;
    $842 = (($841) + -1)|0;
    $843 = ($842|0)==(0);
    if (!($843)) {
     $sblim$08$i = $842;
     label = 114;
    }
   }
   if ((label|0) == 114) {
    label = 0;
    $844 = (((($559) + (($ch$054*2304)|0)|0)) + 72|0);
    $845 = +HEAPF32[179577];
    $846 = +HEAPF32[179585];
    $847 = +HEAPF32[(718312)>>2];
    $848 = +HEAPF32[(718344)>>2];
    $849 = +HEAPF32[(718316)>>2];
    $850 = +HEAPF32[(718348)>>2];
    $851 = +HEAPF32[(718320)>>2];
    $852 = +HEAPF32[(718352)>>2];
    $853 = +HEAPF32[(718324)>>2];
    $854 = +HEAPF32[(718356)>>2];
    $855 = +HEAPF32[(718328)>>2];
    $856 = +HEAPF32[(718360)>>2];
    $857 = +HEAPF32[(718332)>>2];
    $858 = +HEAPF32[(718364)>>2];
    $859 = +HEAPF32[(718336)>>2];
    $860 = +HEAPF32[(718368)>>2];
    $sb$06$i = $sblim$08$i;$xr1$07$i = $844;
    while(1) {
     $861 = ((($xr1$07$i)) + -4|0);
     $862 = +HEAPF32[$861>>2];
     $863 = +HEAPF32[$xr1$07$i>>2];
     $864 = $845 * $862;
     $865 = $846 * $863;
     $866 = $864 - $865;
     HEAPF32[$861>>2] = $866;
     $867 = $845 * $863;
     $868 = $846 * $862;
     $869 = $868 + $867;
     $870 = ((($xr1$07$i)) + 4|0);
     HEAPF32[$xr1$07$i>>2] = $869;
     $871 = ((($xr1$07$i)) + -8|0);
     $872 = +HEAPF32[$871>>2];
     $873 = +HEAPF32[$870>>2];
     $874 = $847 * $872;
     $875 = $848 * $873;
     $876 = $874 - $875;
     HEAPF32[$871>>2] = $876;
     $877 = $847 * $873;
     $878 = $848 * $872;
     $879 = $878 + $877;
     $880 = ((($xr1$07$i)) + 8|0);
     HEAPF32[$870>>2] = $879;
     $881 = ((($xr1$07$i)) + -12|0);
     $882 = +HEAPF32[$881>>2];
     $883 = +HEAPF32[$880>>2];
     $884 = $849 * $882;
     $885 = $850 * $883;
     $886 = $884 - $885;
     HEAPF32[$881>>2] = $886;
     $887 = $849 * $883;
     $888 = $850 * $882;
     $889 = $888 + $887;
     $890 = ((($xr1$07$i)) + 12|0);
     HEAPF32[$880>>2] = $889;
     $891 = ((($xr1$07$i)) + -16|0);
     $892 = +HEAPF32[$891>>2];
     $893 = +HEAPF32[$890>>2];
     $894 = $851 * $892;
     $895 = $852 * $893;
     $896 = $894 - $895;
     HEAPF32[$891>>2] = $896;
     $897 = $851 * $893;
     $898 = $852 * $892;
     $899 = $898 + $897;
     $900 = ((($xr1$07$i)) + 16|0);
     HEAPF32[$890>>2] = $899;
     $901 = ((($xr1$07$i)) + -20|0);
     $902 = +HEAPF32[$901>>2];
     $903 = +HEAPF32[$900>>2];
     $904 = $853 * $902;
     $905 = $854 * $903;
     $906 = $904 - $905;
     HEAPF32[$901>>2] = $906;
     $907 = $853 * $903;
     $908 = $854 * $902;
     $909 = $908 + $907;
     $910 = ((($xr1$07$i)) + 20|0);
     HEAPF32[$900>>2] = $909;
     $911 = ((($xr1$07$i)) + -24|0);
     $912 = +HEAPF32[$911>>2];
     $913 = +HEAPF32[$910>>2];
     $914 = $855 * $912;
     $915 = $856 * $913;
     $916 = $914 - $915;
     HEAPF32[$911>>2] = $916;
     $917 = $855 * $913;
     $918 = $856 * $912;
     $919 = $918 + $917;
     $920 = ((($xr1$07$i)) + 24|0);
     HEAPF32[$910>>2] = $919;
     $921 = ((($xr1$07$i)) + -28|0);
     $922 = +HEAPF32[$921>>2];
     $923 = +HEAPF32[$920>>2];
     $924 = $857 * $922;
     $925 = $858 * $923;
     $926 = $924 - $925;
     HEAPF32[$921>>2] = $926;
     $927 = $857 * $923;
     $928 = $858 * $922;
     $929 = $928 + $927;
     $930 = ((($xr1$07$i)) + 28|0);
     HEAPF32[$920>>2] = $929;
     $931 = ((($xr1$07$i)) + -32|0);
     $932 = +HEAPF32[$931>>2];
     $933 = +HEAPF32[$930>>2];
     $934 = $859 * $932;
     $935 = $860 * $933;
     $936 = $934 - $935;
     HEAPF32[$931>>2] = $936;
     $937 = $859 * $933;
     $938 = $860 * $932;
     $939 = $938 + $937;
     HEAPF32[$930>>2] = $939;
     $940 = (($sb$06$i) + -1)|0;
     $941 = ((($xr1$07$i)) + 72|0);
     $942 = ($940|0)==(0);
     if ($942) {
      break;
     } else {
      $sb$06$i = $940;$xr1$07$i = $941;
     }
    }
   }
   $943 = (($560) + (($ch$054*2304)|0)|0);
   $944 = (((($fr)) + 9224|0) + ($ch$054<<2)|0);
   $945 = HEAP32[$944>>2]|0;
   $946 = ((((($fr)) + 8|0) + (($945*4608)|0)|0) + (($ch$054*2304)|0)|0);
   $947 = (1 - ($945))|0;
   $948 = ((((($fr)) + 8|0) + (($947*4608)|0)|0) + (($ch$054*2304)|0)|0);
   HEAP32[$944>>2] = $947;
   $949 = ((((((($sideinfo)) + 8|0) + (($ch$054*184)|0)|0) + (($gr$061*92)|0)|0)) + 20|0);
   $950 = HEAP32[$949>>2]|0;
   $951 = ($950|0)==(0);
   if ($951) {
    $rawout1$0$i = $946;$rawout2$0$i = $948;$sb$0$i = 0;$tspnt$0$i = $943;
   } else {
    $952 = (($559) + (($ch$054*2304)|0)|0);
    _INT123_dct36($952,$946,$948,718372,$943);
    $953 = (((($559) + (($ch$054*2304)|0)|0)) + 72|0);
    $954 = ((((((($fr)) + 8|0) + (($945*4608)|0)|0) + (($ch$054*2304)|0)|0)) + 72|0);
    $955 = ((((((($fr)) + 8|0) + (($947*4608)|0)|0) + (($ch$054*2304)|0)|0)) + 72|0);
    $956 = (((($560) + (($ch$054*2304)|0)|0)) + 4|0);
    _INT123_dct36($953,$954,$955,718980,$956);
    $957 = ((((((($fr)) + 8|0) + (($945*4608)|0)|0) + (($ch$054*2304)|0)|0)) + 144|0);
    $958 = ((((((($fr)) + 8|0) + (($947*4608)|0)|0) + (($ch$054*2304)|0)|0)) + 144|0);
    $959 = (((($560) + (($ch$054*2304)|0)|0)) + 8|0);
    $rawout1$0$i = $957;$rawout2$0$i = $958;$sb$0$i = 2;$tspnt$0$i = $959;
   }
   $960 = HEAP32[$834>>2]|0;
   $961 = ($960|0)==(2);
   $962 = ((((((($sideinfo)) + 8|0) + (($ch$054*184)|0)|0) + (($gr$061*92)|0)|0)) + 52|0);
   $963 = HEAP32[$962>>2]|0;
   $964 = ($sb$0$i>>>0)<($963>>>0);
   if ($961) {
    if ($964) {
     $rawout1$112$i = $rawout1$0$i;$rawout2$113$i = $rawout2$0$i;$sb$114$i = $sb$0$i;$tspnt$115$i = $tspnt$0$i;
     while(1) {
      $968 = ((($559) + (($ch$054*2304)|0)|0) + (($sb$114$i*72)|0)|0);
      _dct12($968,$rawout1$112$i,$rawout2$113$i,(718660),$tspnt$115$i);
      $969 = $sb$114$i | 1;
      $970 = ((($559) + (($ch$054*2304)|0)|0) + (($969*72)|0)|0);
      $971 = ((($rawout1$112$i)) + 72|0);
      $972 = ((($rawout2$113$i)) + 72|0);
      $973 = ((($tspnt$115$i)) + 4|0);
      _dct12($970,$971,$972,(719268),$973);
      $974 = (($sb$114$i) + 2)|0;
      $975 = ((($tspnt$115$i)) + 8|0);
      $976 = ((($rawout1$112$i)) + 144|0);
      $977 = ((($rawout2$113$i)) + 144|0);
      $978 = HEAP32[$962>>2]|0;
      $979 = ($974>>>0)<($978>>>0);
      if ($979) {
       $rawout1$112$i = $976;$rawout2$113$i = $977;$sb$114$i = $974;$tspnt$115$i = $975;
      } else {
       $rawout1$3$ph$i = $976;$rawout2$3$ph$i = $977;$sb$3$ph$i = $974;$tspnt$3$ph$i = $975;
       label = 122;
       break;
      }
     }
    } else {
     $rawout1$38$i = $rawout1$0$i;$rawout2$39$i = $rawout2$0$i;$sb$310$i = $sb$0$i;$tspnt$311$i = $tspnt$0$i;
     label = 125;
    }
   } else {
    if ($964) {
     $965 = (718372 + (($960*144)|0)|0);
     $966 = (718980 + (($960*144)|0)|0);
     $rawout1$219$i = $rawout1$0$i;$rawout2$220$i = $rawout2$0$i;$sb$221$i = $sb$0$i;$tspnt$222$i = $tspnt$0$i;
     while(1) {
      $980 = ((($559) + (($ch$054*2304)|0)|0) + (($sb$221$i*72)|0)|0);
      _INT123_dct36($980,$rawout1$219$i,$rawout2$220$i,$965,$tspnt$222$i);
      $981 = $sb$221$i | 1;
      $982 = ((($559) + (($ch$054*2304)|0)|0) + (($981*72)|0)|0);
      $983 = ((($rawout1$219$i)) + 72|0);
      $984 = ((($rawout2$220$i)) + 72|0);
      $985 = ((($tspnt$222$i)) + 4|0);
      _INT123_dct36($982,$983,$984,$966,$985);
      $986 = (($sb$221$i) + 2)|0;
      $987 = ((($tspnt$222$i)) + 8|0);
      $988 = ((($rawout1$219$i)) + 144|0);
      $989 = ((($rawout2$220$i)) + 144|0);
      $990 = HEAP32[$962>>2]|0;
      $991 = ($986>>>0)<($990>>>0);
      if ($991) {
       $rawout1$219$i = $988;$rawout2$220$i = $989;$sb$221$i = $986;$tspnt$222$i = $987;
      } else {
       $rawout1$3$ph$i = $988;$rawout2$3$ph$i = $989;$sb$3$ph$i = $986;$tspnt$3$ph$i = $987;
       label = 122;
       break;
      }
     }
    } else {
     $rawout1$38$i = $rawout1$0$i;$rawout2$39$i = $rawout2$0$i;$sb$310$i = $sb$0$i;$tspnt$311$i = $tspnt$0$i;
     label = 125;
    }
   }
   if ((label|0) == 122) {
    label = 0;
    $967 = ($sb$3$ph$i>>>0)<(32);
    if ($967) {
     $rawout1$38$i = $rawout1$3$ph$i;$rawout2$39$i = $rawout2$3$ph$i;$sb$310$i = $sb$3$ph$i;$tspnt$311$i = $tspnt$3$ph$i;
     label = 125;
    }
   }
   if ((label|0) == 125) {
    while(1) {
     label = 0;
     $scevgep$i = ((($rawout2$39$i)) + 72|0);
     $992 = ((($rawout1$38$i)) + 4|0);
     $993 = HEAP32[$rawout1$38$i>>2]|0;
     HEAP32[$tspnt$311$i>>2] = $993;
     $994 = ((($rawout2$39$i)) + 4|0);
     HEAPF32[$rawout2$39$i>>2] = 0.0;
     $995 = ((($rawout1$38$i)) + 8|0);
     $996 = HEAP32[$992>>2]|0;
     $997 = ((($tspnt$311$i)) + 128|0);
     HEAP32[$997>>2] = $996;
     $998 = ((($rawout2$39$i)) + 8|0);
     HEAPF32[$994>>2] = 0.0;
     $999 = ((($rawout1$38$i)) + 12|0);
     $1000 = HEAP32[$995>>2]|0;
     $1001 = ((($tspnt$311$i)) + 256|0);
     HEAP32[$1001>>2] = $1000;
     $1002 = ((($rawout2$39$i)) + 12|0);
     HEAPF32[$998>>2] = 0.0;
     $1003 = ((($rawout1$38$i)) + 16|0);
     $1004 = HEAP32[$999>>2]|0;
     $1005 = ((($tspnt$311$i)) + 384|0);
     HEAP32[$1005>>2] = $1004;
     $1006 = ((($rawout2$39$i)) + 16|0);
     HEAPF32[$1002>>2] = 0.0;
     $1007 = ((($rawout1$38$i)) + 20|0);
     $1008 = HEAP32[$1003>>2]|0;
     $1009 = ((($tspnt$311$i)) + 512|0);
     HEAP32[$1009>>2] = $1008;
     $1010 = ((($rawout2$39$i)) + 20|0);
     HEAPF32[$1006>>2] = 0.0;
     $1011 = ((($rawout1$38$i)) + 24|0);
     $1012 = HEAP32[$1007>>2]|0;
     $1013 = ((($tspnt$311$i)) + 640|0);
     HEAP32[$1013>>2] = $1012;
     $1014 = ((($rawout2$39$i)) + 24|0);
     HEAPF32[$1010>>2] = 0.0;
     $1015 = ((($rawout1$38$i)) + 28|0);
     $1016 = HEAP32[$1011>>2]|0;
     $1017 = ((($tspnt$311$i)) + 768|0);
     HEAP32[$1017>>2] = $1016;
     $1018 = ((($rawout2$39$i)) + 28|0);
     HEAPF32[$1014>>2] = 0.0;
     $1019 = ((($rawout1$38$i)) + 32|0);
     $1020 = HEAP32[$1015>>2]|0;
     $1021 = ((($tspnt$311$i)) + 896|0);
     HEAP32[$1021>>2] = $1020;
     $1022 = ((($rawout2$39$i)) + 32|0);
     HEAPF32[$1018>>2] = 0.0;
     $1023 = ((($rawout1$38$i)) + 36|0);
     $1024 = HEAP32[$1019>>2]|0;
     $1025 = ((($tspnt$311$i)) + 1024|0);
     HEAP32[$1025>>2] = $1024;
     $1026 = ((($rawout2$39$i)) + 36|0);
     HEAPF32[$1022>>2] = 0.0;
     $1027 = ((($rawout1$38$i)) + 40|0);
     $1028 = HEAP32[$1023>>2]|0;
     $1029 = ((($tspnt$311$i)) + 1152|0);
     HEAP32[$1029>>2] = $1028;
     $1030 = ((($rawout2$39$i)) + 40|0);
     HEAPF32[$1026>>2] = 0.0;
     $1031 = ((($rawout1$38$i)) + 44|0);
     $1032 = HEAP32[$1027>>2]|0;
     $1033 = ((($tspnt$311$i)) + 1280|0);
     HEAP32[$1033>>2] = $1032;
     $1034 = ((($rawout2$39$i)) + 44|0);
     HEAPF32[$1030>>2] = 0.0;
     $1035 = ((($rawout1$38$i)) + 48|0);
     $1036 = HEAP32[$1031>>2]|0;
     $1037 = ((($tspnt$311$i)) + 1408|0);
     HEAP32[$1037>>2] = $1036;
     $1038 = ((($rawout2$39$i)) + 48|0);
     HEAPF32[$1034>>2] = 0.0;
     $1039 = ((($rawout1$38$i)) + 52|0);
     $1040 = HEAP32[$1035>>2]|0;
     $1041 = ((($tspnt$311$i)) + 1536|0);
     HEAP32[$1041>>2] = $1040;
     $1042 = ((($rawout2$39$i)) + 52|0);
     HEAPF32[$1038>>2] = 0.0;
     $1043 = ((($rawout1$38$i)) + 56|0);
     $1044 = HEAP32[$1039>>2]|0;
     $1045 = ((($tspnt$311$i)) + 1664|0);
     HEAP32[$1045>>2] = $1044;
     $1046 = ((($rawout2$39$i)) + 56|0);
     HEAPF32[$1042>>2] = 0.0;
     $1047 = ((($rawout1$38$i)) + 60|0);
     $1048 = HEAP32[$1043>>2]|0;
     $1049 = ((($tspnt$311$i)) + 1792|0);
     HEAP32[$1049>>2] = $1048;
     $1050 = ((($rawout2$39$i)) + 60|0);
     HEAPF32[$1046>>2] = 0.0;
     $1051 = ((($rawout1$38$i)) + 64|0);
     $1052 = HEAP32[$1047>>2]|0;
     $1053 = ((($tspnt$311$i)) + 1920|0);
     HEAP32[$1053>>2] = $1052;
     $1054 = ((($rawout2$39$i)) + 64|0);
     HEAPF32[$1050>>2] = 0.0;
     $1055 = ((($rawout1$38$i)) + 68|0);
     $1056 = HEAP32[$1051>>2]|0;
     $1057 = ((($tspnt$311$i)) + 2048|0);
     HEAP32[$1057>>2] = $1056;
     $1058 = ((($rawout2$39$i)) + 68|0);
     HEAPF32[$1054>>2] = 0.0;
     $1059 = HEAP32[$1055>>2]|0;
     $1060 = ((($tspnt$311$i)) + 2176|0);
     HEAP32[$1060>>2] = $1059;
     HEAPF32[$1058>>2] = 0.0;
     $scevgep38$i = ((($rawout1$38$i)) + 72|0);
     $1061 = (($sb$310$i) + 1)|0;
     $1062 = ((($tspnt$311$i)) + 4|0);
     $exitcond39$i = ($1061|0)==(32);
     if ($exitcond39$i) {
      break;
     } else {
      $rawout1$38$i = $scevgep38$i;$rawout2$39$i = $scevgep$i;$sb$310$i = $1061;$tspnt$311$i = $1062;
      label = 125;
     }
    }
   }
   $1063 = (($ch$054) + 1)|0;
   $1064 = ($1063|0)<($stereo1$0|0);
   if ($1064) {
    $ch$054 = $1063;
   } else {
    break;
   }
  }
  if ($556) {
   $clip$158$us = $clip$060;$ss$059$us = 0;
   while(1) {
    $829 = HEAP32[$557>>2]|0;
    $830 = (($560) + ($ss$059$us<<7)|0);
    $831 = (((($560)) + 2304|0) + ($ss$059$us<<7)|0);
    $832 = (FUNCTION_TABLE_iiii[$829 & 8191]($830,$831,$fr)|0);
    $clip$2$us = (($832) + ($clip$158$us))|0;
    $833 = (($ss$059$us) + 1)|0;
    $exitcond85 = ($833|0)==(18);
    if ($exitcond85) {
     $clip$1$lcssa = $clip$2$us;
     break;
    } else {
     $clip$158$us = $clip$2$us;$ss$059$us = $833;
    }
   }
  } else {
   $776 = HEAP32[$558>>2]|0;
   $777 = (FUNCTION_TABLE_iii[$776 & 8191]($560,$fr)|0);
   $clip$2 = (($777) + ($clip$060))|0;
   $778 = HEAP32[$558>>2]|0;
   $779 = ((($560)) + 128|0);
   $780 = (FUNCTION_TABLE_iii[$778 & 8191]($779,$fr)|0);
   $clip$2$1 = (($780) + ($clip$2))|0;
   $781 = HEAP32[$558>>2]|0;
   $782 = ((($560)) + 256|0);
   $783 = (FUNCTION_TABLE_iii[$781 & 8191]($782,$fr)|0);
   $clip$2$2 = (($783) + ($clip$2$1))|0;
   $784 = HEAP32[$558>>2]|0;
   $785 = ((($560)) + 384|0);
   $786 = (FUNCTION_TABLE_iii[$784 & 8191]($785,$fr)|0);
   $clip$2$3 = (($786) + ($clip$2$2))|0;
   $787 = HEAP32[$558>>2]|0;
   $788 = ((($560)) + 512|0);
   $789 = (FUNCTION_TABLE_iii[$787 & 8191]($788,$fr)|0);
   $clip$2$4 = (($789) + ($clip$2$3))|0;
   $790 = HEAP32[$558>>2]|0;
   $791 = ((($560)) + 640|0);
   $792 = (FUNCTION_TABLE_iii[$790 & 8191]($791,$fr)|0);
   $clip$2$5 = (($792) + ($clip$2$4))|0;
   $793 = HEAP32[$558>>2]|0;
   $794 = ((($560)) + 768|0);
   $795 = (FUNCTION_TABLE_iii[$793 & 8191]($794,$fr)|0);
   $clip$2$6 = (($795) + ($clip$2$5))|0;
   $796 = HEAP32[$558>>2]|0;
   $797 = ((($560)) + 896|0);
   $798 = (FUNCTION_TABLE_iii[$796 & 8191]($797,$fr)|0);
   $clip$2$7 = (($798) + ($clip$2$6))|0;
   $799 = HEAP32[$558>>2]|0;
   $800 = ((($560)) + 1024|0);
   $801 = (FUNCTION_TABLE_iii[$799 & 8191]($800,$fr)|0);
   $clip$2$8 = (($801) + ($clip$2$7))|0;
   $802 = HEAP32[$558>>2]|0;
   $803 = ((($560)) + 1152|0);
   $804 = (FUNCTION_TABLE_iii[$802 & 8191]($803,$fr)|0);
   $clip$2$9 = (($804) + ($clip$2$8))|0;
   $805 = HEAP32[$558>>2]|0;
   $806 = ((($560)) + 1280|0);
   $807 = (FUNCTION_TABLE_iii[$805 & 8191]($806,$fr)|0);
   $clip$2$10 = (($807) + ($clip$2$9))|0;
   $808 = HEAP32[$558>>2]|0;
   $809 = ((($560)) + 1408|0);
   $810 = (FUNCTION_TABLE_iii[$808 & 8191]($809,$fr)|0);
   $clip$2$11 = (($810) + ($clip$2$10))|0;
   $811 = HEAP32[$558>>2]|0;
   $812 = ((($560)) + 1536|0);
   $813 = (FUNCTION_TABLE_iii[$811 & 8191]($812,$fr)|0);
   $clip$2$12 = (($813) + ($clip$2$11))|0;
   $814 = HEAP32[$558>>2]|0;
   $815 = ((($560)) + 1664|0);
   $816 = (FUNCTION_TABLE_iii[$814 & 8191]($815,$fr)|0);
   $clip$2$13 = (($816) + ($clip$2$12))|0;
   $817 = HEAP32[$558>>2]|0;
   $818 = ((($560)) + 1792|0);
   $819 = (FUNCTION_TABLE_iii[$817 & 8191]($818,$fr)|0);
   $clip$2$14 = (($819) + ($clip$2$13))|0;
   $820 = HEAP32[$558>>2]|0;
   $821 = ((($560)) + 1920|0);
   $822 = (FUNCTION_TABLE_iii[$820 & 8191]($821,$fr)|0);
   $clip$2$15 = (($822) + ($clip$2$14))|0;
   $823 = HEAP32[$558>>2]|0;
   $824 = ((($560)) + 2048|0);
   $825 = (FUNCTION_TABLE_iii[$823 & 8191]($824,$fr)|0);
   $clip$2$16 = (($825) + ($clip$2$15))|0;
   $826 = HEAP32[$558>>2]|0;
   $827 = ((($560)) + 2176|0);
   $828 = (FUNCTION_TABLE_iii[$826 & 8191]($827,$fr)|0);
   $clip$2$17 = (($828) + ($clip$2$16))|0;
   $clip$1$lcssa = $clip$2$17;
  }
  $1065 = (($gr$061) + 1)|0;
  $1066 = ($1065|0)<($19|0);
  if ($1066) {
   $clip$060 = $clip$1$lcssa;$gr$061 = $1065;
  } else {
   $$5 = $clip$1$lcssa;
   label = 128;
   break;
  }
 }
 if ((label|0) == 50) {
  $568 = ((($fr)) + 28068|0);
  $569 = HEAP32[$568>>2]|0;
  $570 = $569 & 32;
  $571 = ($570|0)==(0);
  if (!($571)) {
   $$5 = $clip$060$lcssa;
   STACKTOP = sp;return ($$5|0);
  }
  $572 = ((($fr)) + 28064|0);
  $573 = HEAP32[$572>>2]|0;
  $574 = ($573|0)>(1);
  if (!($574)) {
   $$5 = $clip$060$lcssa;
   STACKTOP = sp;return ($$5|0);
  }
  $575 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer11>>2] = 1946;
  (_fprintf($575,511891,$vararg_buffer11)|0);
  $$5 = $clip$060$lcssa;
  STACKTOP = sp;return ($$5|0);
 }
 else if ((label|0) == 58) {
  $584 = ((($fr)) + 28068|0);
  $585 = HEAP32[$584>>2]|0;
  $586 = $585 & 32;
  $587 = ($586|0)==(0);
  if (!($587)) {
   $$5 = $clip$060$lcssa160;
   STACKTOP = sp;return ($$5|0);
  }
  $588 = ((($fr)) + 28064|0);
  $589 = HEAP32[$588>>2]|0;
  $590 = ($589|0)>(1);
  if (!($590)) {
   $$5 = $clip$060$lcssa160;
   STACKTOP = sp;return ($$5|0);
  }
  $591 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer14>>2] = 1962;
  (_fprintf($591,511891,$vararg_buffer14)|0);
  $$5 = $clip$060$lcssa160;
  STACKTOP = sp;return ($$5|0);
 }
 else if ((label|0) == 128) {
  STACKTOP = sp;return ($$5|0);
 }
 return (0)|0;
}
function _INT123_set_pointer($fr,$backstep) {
 $fr = $fr|0;
 $backstep = $backstep|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27808|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 19864|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = (0 - ($backstep))|0;
 $6 = (($4) + ($5)|0);
 $7 = ((($fr)) + 19788|0);
 HEAP32[$7>>2] = $6;
 $8 = ($backstep|0)==(0);
 if ($8) {
  $15 = ((($fr)) + 19784|0);
  HEAP32[$15>>2] = 0;
  return;
 }
 $9 = ((($fr)) + 27812|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($fr)) + 19860|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10) + ($12)|0);
 $14 = (($13) + ($5)|0);
 _memcpy(($6|0),($14|0),($backstep|0))|0;
 $15 = ((($fr)) + 19784|0);
 HEAP32[$15>>2] = 0;
 return;
}
function _III_get_scale_factors_2($fr,$scf,$gr_info,$i_stereo) {
 $fr = $fr|0;
 $scf = $scf|0;
 $gr_info = $gr_info|0;
 $i_stereo = $i_stereo|0;
 var $$ = 0, $$014 = 0, $$1$lcssa = 0, $$15 = 0, $$28 = 0, $$3 = 0, $$3$lcssa = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $fold = 0, $i$011 = 0, $j$04 = 0, $j$17 = 0, $n$0 = 0, $numbits$012 = 0, $numbits$1 = 0, $numbits$1$lcssa = 0, $slen$0 = 0, $slen$0$in = 0, $slen$113 = 0, $smax19 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($i_stereo|0)==(0);
 $1 = ((($gr_info)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2 >>> 1;
 $4 = (681424 + ($3<<2)|0);
 $5 = (682448 + ($2<<2)|0);
 $slen$0$in = $0 ? $5 : $4;
 $slen$0 = HEAP32[$slen$0$in>>2]|0;
 $6 = $slen$0 >>> 15;
 $7 = $6 & 1;
 $8 = ((($gr_info)) + 64|0);
 HEAP32[$8>>2] = $7;
 $9 = ((($gr_info)) + 16|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10|0)==(2);
 if ($11) {
  $12 = ((($gr_info)) + 20|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ($13|0)==(0);
  $$ = $14 ? 1 : 2;
  $n$0 = $$;
 } else {
  $n$0 = 0;
 }
 $15 = $slen$0 >>> 12;
 $16 = $15 & 7;
 $17 = ((($fr)) + 19788|0);
 $18 = ((($fr)) + 19784|0);
 $19 = ((($fr)) + 19792|0);
 $$014 = $scf;$i$011 = 0;$numbits$012 = 0;$slen$113 = $slen$0;
 while(1) {
  $20 = $slen$113 & 7;
  $21 = $slen$113 >>> 3;
  $22 = ($20|0)==(0);
  $23 = (((511685 + (($n$0*24)|0)|0) + ($16<<2)|0) + ($i$011)|0);
  $24 = HEAP8[$23>>0]|0;
  $25 = ($24<<24>>24)!=(0);
  if ($22) {
   if ($25) {
    $27 = $24&255;
    $28 = ($27>>>0)>(1);
    $smax19 = $28 ? $27 : 1;
    $29 = $smax19 << 2;
    _memset(($$014|0),0,($29|0))|0;
    $$28 = $$014;$j$17 = 0;
    while(1) {
     $53 = ((($$28)) + 4|0);
     $54 = (($j$17) + 1)|0;
     $55 = ($54|0)<($27|0);
     if ($55) {
      $$28 = $53;$j$17 = $54;
     } else {
      $$3 = $53;$numbits$1 = $numbits$012;
      break;
     }
    }
   } else {
    $$3 = $$014;$numbits$1 = $numbits$012;
   }
  } else {
   $26 = $24&255;
   if ($25) {
    $$pre = HEAP32[$17>>2]|0;
    $$15 = $$014;$31 = $$pre;$j$04 = 0;
    while(1) {
     $30 = HEAP8[$31>>0]|0;
     $32 = $30&255;
     $33 = HEAP32[$18>>2]|0;
     $34 = $32 << $33;
     $35 = $34 & 255;
     HEAP32[$19>>2] = $35;
     $36 = ((($31)) + 1|0);
     $37 = HEAP8[$36>>0]|0;
     $38 = $37&255;
     $39 = $38 << $33;
     $40 = $39 >>> 8;
     $41 = $40 | $35;
     $42 = $41 << $20;
     $43 = $42 >>> 8;
     HEAP32[$19>>2] = $43;
     $44 = (($33) + ($20))|0;
     $45 = $44 >> 3;
     $46 = (($31) + ($45)|0);
     HEAP32[$17>>2] = $46;
     $fold = (($33) + ($slen$113))|0;
     $47 = $fold & 7;
     HEAP32[$18>>2] = $47;
     $48 = ((($$15)) + 4|0);
     HEAP32[$$15>>2] = $43;
     $49 = (($j$04) + 1)|0;
     $50 = ($49|0)<($26|0);
     if ($50) {
      $$15 = $48;$31 = $46;$j$04 = $49;
     } else {
      $$1$lcssa = $48;
      break;
     }
    }
   } else {
    $$1$lcssa = $$014;
   }
   $51 = Math_imul($26, $20)|0;
   $52 = (($51) + ($numbits$012))|0;
   $$3 = $$1$lcssa;$numbits$1 = $52;
  }
  $56 = (($i$011) + 1)|0;
  $exitcond = ($56|0)==(4);
  if ($exitcond) {
   $$3$lcssa = $$3;$numbits$1$lcssa = $numbits$1;
   break;
  } else {
   $$014 = $$3;$i$011 = $56;$numbits$012 = $numbits$1;$slen$113 = $21;
  }
 }
 $57 = $n$0 << 3;
 $58 = $57 | 4;
 _memset(($$3$lcssa|0),0,($58|0))|0;
 return ($numbits$1$lcssa|0);
}
function _III_get_scale_factors_1($fr,$scf,$gr_info) {
 $fr = $fr|0;
 $scf = $scf|0;
 $gr_info = $gr_info|0;
 var $$014 = 0, $$1$ph = 0, $$10 = 0, $$112 = 0, $$12 = 0, $$210 = 0, $$318 = 0, $$416 = 0, $$6 = 0, $$8 = 0, $$lcssa = 0, $$lcssa61 = 0, $$lcssa62 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phi52Z2D = 0, $$pre$phi54Z2D = 0, $$pre47 = 0, $$pre48 = 0, $$pre49 = 0;
 var $$pre49$pre = 0, $$pre51 = 0, $$pre53 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0;
 var $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0;
 var $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0;
 var $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0;
 var $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0;
 var $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0;
 var $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0;
 var $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0;
 var $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0;
 var $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0;
 var $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0;
 var $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0;
 var $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0;
 var $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0;
 var $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0;
 var $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0;
 var $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $i$013 = 0, $i$1$ph = 0, $i$111 = 0, $i$29 = 0, $i1$017 = 0, $i1$115 = 0, $numbits$0$ph = 0, $numbits$1 = 0, $numbits$2 = 0, $numbits$3 = 0, $numbits$4 = 0, $numbits$6 = 0, $scevgep = 0, $scevgep35 = 0, $scevgep36 = 0, $scevgep37 = 0, $scevgep38 = 0, $scevgep39 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($gr_info)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (511757 + ($1)|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = (511773 + ($1)|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = $6&255;
 $8 = ((($gr_info)) + 16|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)==(2);
 if ($10) {
  $11 = (($7) + ($4))|0;
  $12 = ($11*18)|0;
  $13 = ((($gr_info)) + 20|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ($14|0)==(0);
  $$phi$trans$insert = ((($fr)) + 19788|0);
  if ($15) {
   $$pre49$pre = HEAP32[$$phi$trans$insert>>2]|0;
   $$pre51 = ((($fr)) + 19784|0);
   $$pre53 = ((($fr)) + 19792|0);
   $$1$ph = $scf;$$pre$phi52Z2D = $$pre51;$$pre$phi54Z2D = $$pre53;$$pre49 = $$pre49$pre;$i$1$ph = 18;$numbits$0$ph = $12;
  } else {
   $16 = ((($fr)) + 19784|0);
   $17 = ((($fr)) + 19792|0);
   $$pre48 = HEAP32[$$phi$trans$insert>>2]|0;
   $$014 = $scf;$19 = $$pre48;$i$013 = 8;
   while(1) {
    $18 = HEAP8[$19>>0]|0;
    $20 = $18&255;
    $21 = HEAP32[$16>>2]|0;
    $22 = $20 << $21;
    $23 = $22 & 255;
    HEAP32[$17>>2] = $23;
    $24 = ((($19)) + 1|0);
    $25 = HEAP8[$24>>0]|0;
    $26 = $25&255;
    $27 = $26 << $21;
    $28 = $27 >>> 8;
    $29 = $28 | $23;
    $30 = $29 << $4;
    $31 = $30 >>> 8;
    HEAP32[$17>>2] = $31;
    $32 = (($21) + ($4))|0;
    $33 = $32 >> 3;
    $34 = (($19) + ($33)|0);
    HEAP32[$$phi$trans$insert>>2] = $34;
    $35 = $32 & 7;
    HEAP32[$16>>2] = $35;
    $36 = ((($$014)) + 4|0);
    HEAP32[$$014>>2] = $31;
    $37 = (($i$013) + -1)|0;
    $38 = ($37|0)==(0);
    if ($38) {
     $$lcssa61 = $34;
     break;
    } else {
     $$014 = $36;$19 = $34;$i$013 = $37;
    }
   }
   $scevgep37 = ((($scf)) + 32|0);
   $39 = (($12) - ($4))|0;
   $$1$ph = $scevgep37;$$pre$phi52Z2D = $16;$$pre$phi54Z2D = $17;$$pre49 = $$lcssa61;$i$1$ph = 9;$numbits$0$ph = $39;
  }
  $$112 = $$1$ph;$41 = $$pre49;$i$111 = $i$1$ph;
  while(1) {
   $40 = HEAP8[$41>>0]|0;
   $42 = $40&255;
   $43 = HEAP32[$$pre$phi52Z2D>>2]|0;
   $44 = $42 << $43;
   $45 = $44 & 255;
   HEAP32[$$pre$phi54Z2D>>2] = $45;
   $46 = ((($41)) + 1|0);
   $47 = HEAP8[$46>>0]|0;
   $48 = $47&255;
   $49 = $48 << $43;
   $50 = $49 >>> 8;
   $51 = $50 | $45;
   $52 = $51 << $4;
   $53 = $52 >>> 8;
   HEAP32[$$pre$phi54Z2D>>2] = $53;
   $54 = (($43) + ($4))|0;
   $55 = $54 >> 3;
   $56 = (($41) + ($55)|0);
   HEAP32[$$phi$trans$insert>>2] = $56;
   $57 = $54 & 7;
   HEAP32[$$pre$phi52Z2D>>2] = $57;
   $58 = ((($$112)) + 4|0);
   HEAP32[$$112>>2] = $53;
   $59 = (($i$111) + -1)|0;
   $60 = ($59|0)==(0);
   if ($60) {
    $$lcssa = $56;
    break;
   } else {
    $$112 = $58;$41 = $56;$i$111 = $59;
   }
  }
  $scevgep36 = (($$1$ph) + ($i$1$ph<<2)|0);
  $scevgep = ((($scevgep36)) + 72|0);
  $$210 = $scevgep36;$62 = $$lcssa;$i$29 = 18;
  while(1) {
   $61 = HEAP8[$62>>0]|0;
   $63 = $61&255;
   $64 = HEAP32[$$pre$phi52Z2D>>2]|0;
   $65 = $63 << $64;
   $66 = $65 & 255;
   HEAP32[$$pre$phi54Z2D>>2] = $66;
   $67 = ((($62)) + 1|0);
   $68 = HEAP8[$67>>0]|0;
   $69 = $68&255;
   $70 = $69 << $64;
   $71 = $70 >>> 8;
   $72 = $71 | $66;
   $73 = $72 << $7;
   $74 = $73 >>> 8;
   HEAP32[$$pre$phi54Z2D>>2] = $74;
   $75 = (($64) + ($7))|0;
   $76 = $75 >> 3;
   $77 = (($62) + ($76)|0);
   HEAP32[$$phi$trans$insert>>2] = $77;
   $78 = $75 & 7;
   HEAP32[$$pre$phi52Z2D>>2] = $78;
   $79 = ((($$210)) + 4|0);
   HEAP32[$$210>>2] = $74;
   $80 = (($i$29) + -1)|0;
   $81 = ($80|0)==(0);
   if ($81) {
    break;
   } else {
    $$210 = $79;$62 = $77;$i$29 = $80;
   }
  }
  $scevgep35 = ((($scevgep36)) + 68|0);
  $82 = ((($scevgep35)) + 8|0);
  HEAP32[$scevgep>>2] = 0;
  $83 = ((($scevgep35)) + 12|0);
  HEAP32[$82>>2] = 0;
  HEAP32[$83>>2] = 0;
  $numbits$6 = $numbits$0$ph;
  return ($numbits$6|0);
 }
 $84 = HEAP32[$gr_info>>2]|0;
 $85 = ($84|0)<(0);
 if ($85) {
  $86 = ((($fr)) + 19788|0);
  $87 = ((($fr)) + 19784|0);
  $88 = ((($fr)) + 19792|0);
  $$pre47 = HEAP32[$86>>2]|0;
  $$318 = $scf;$90 = $$pre47;$i1$017 = 11;
  while(1) {
   $89 = HEAP8[$90>>0]|0;
   $91 = $89&255;
   $92 = HEAP32[$87>>2]|0;
   $93 = $91 << $92;
   $94 = $93 & 255;
   HEAP32[$88>>2] = $94;
   $95 = ((($90)) + 1|0);
   $96 = HEAP8[$95>>0]|0;
   $97 = $96&255;
   $98 = $97 << $92;
   $99 = $98 >>> 8;
   $100 = $99 | $94;
   $101 = $100 << $4;
   $102 = $101 >>> 8;
   HEAP32[$88>>2] = $102;
   $103 = (($92) + ($4))|0;
   $104 = $103 >> 3;
   $105 = (($90) + ($104)|0);
   HEAP32[$86>>2] = $105;
   $106 = $103 & 7;
   HEAP32[$87>>2] = $106;
   $107 = ((($$318)) + 4|0);
   HEAP32[$$318>>2] = $102;
   $108 = (($i1$017) + -1)|0;
   $109 = ($108|0)==(0);
   if ($109) {
    $$lcssa62 = $105;
    break;
   } else {
    $$318 = $107;$90 = $105;$i1$017 = $108;
   }
  }
  $scevgep39 = ((($scf)) + 44|0);
  $$416 = $scevgep39;$111 = $$lcssa62;$i1$115 = 10;
  while(1) {
   $110 = HEAP8[$111>>0]|0;
   $112 = $110&255;
   $113 = HEAP32[$87>>2]|0;
   $114 = $112 << $113;
   $115 = $114 & 255;
   HEAP32[$88>>2] = $115;
   $116 = ((($111)) + 1|0);
   $117 = HEAP8[$116>>0]|0;
   $118 = $117&255;
   $119 = $118 << $113;
   $120 = $119 >>> 8;
   $121 = $120 | $115;
   $122 = $121 << $7;
   $123 = $122 >>> 8;
   HEAP32[$88>>2] = $123;
   $124 = (($113) + ($7))|0;
   $125 = $124 >> 3;
   $126 = (($111) + ($125)|0);
   HEAP32[$86>>2] = $126;
   $127 = $124 & 7;
   HEAP32[$87>>2] = $127;
   $128 = ((($$416)) + 4|0);
   HEAP32[$$416>>2] = $123;
   $129 = (($i1$115) + -1)|0;
   $130 = ($129|0)==(0);
   if ($130) {
    break;
   } else {
    $$416 = $128;$111 = $126;$i1$115 = $129;
   }
  }
  $scevgep38 = ((($scf)) + 84|0);
  $131 = (($7) + ($4))|0;
  $132 = ($131*10)|0;
  $133 = (($132) + ($4))|0;
  HEAP32[$scevgep38>>2] = 0;
  $numbits$6 = $133;
  return ($numbits$6|0);
 }
 $134 = $84 & 8;
 $135 = ($134|0)==(0);
 if ($135) {
  $136 = ((($fr)) + 19788|0);
  $137 = ((($fr)) + 19784|0);
  $138 = ((($fr)) + 19792|0);
  $$pre = HEAP32[$136>>2]|0;
  $139 = HEAP8[$$pre>>0]|0;
  $140 = $139&255;
  $141 = HEAP32[$137>>2]|0;
  $142 = $140 << $141;
  $143 = $142 & 255;
  HEAP32[$138>>2] = $143;
  $144 = ((($$pre)) + 1|0);
  $145 = HEAP8[$144>>0]|0;
  $146 = $145&255;
  $147 = $146 << $141;
  $148 = $147 >>> 8;
  $149 = $148 | $143;
  $150 = $149 << $4;
  $151 = $150 >>> 8;
  HEAP32[$138>>2] = $151;
  $152 = (($141) + ($4))|0;
  $153 = $152 >> 3;
  $154 = (($$pre) + ($153)|0);
  HEAP32[$136>>2] = $154;
  $155 = $152 & 7;
  HEAP32[$137>>2] = $155;
  $156 = ((($scf)) + 4|0);
  HEAP32[$scf>>2] = $151;
  $157 = HEAP8[$154>>0]|0;
  $158 = $157&255;
  $159 = HEAP32[$137>>2]|0;
  $160 = $158 << $159;
  $161 = $160 & 255;
  HEAP32[$138>>2] = $161;
  $162 = ((($154)) + 1|0);
  $163 = HEAP8[$162>>0]|0;
  $164 = $163&255;
  $165 = $164 << $159;
  $166 = $165 >>> 8;
  $167 = $166 | $161;
  $168 = $167 << $4;
  $169 = $168 >>> 8;
  HEAP32[$138>>2] = $169;
  $170 = (($159) + ($4))|0;
  $171 = $170 >> 3;
  $172 = (($154) + ($171)|0);
  HEAP32[$136>>2] = $172;
  $173 = $170 & 7;
  HEAP32[$137>>2] = $173;
  $174 = ((($scf)) + 8|0);
  HEAP32[$156>>2] = $169;
  $175 = HEAP8[$172>>0]|0;
  $176 = $175&255;
  $177 = HEAP32[$137>>2]|0;
  $178 = $176 << $177;
  $179 = $178 & 255;
  HEAP32[$138>>2] = $179;
  $180 = ((($172)) + 1|0);
  $181 = HEAP8[$180>>0]|0;
  $182 = $181&255;
  $183 = $182 << $177;
  $184 = $183 >>> 8;
  $185 = $184 | $179;
  $186 = $185 << $4;
  $187 = $186 >>> 8;
  HEAP32[$138>>2] = $187;
  $188 = (($177) + ($4))|0;
  $189 = $188 >> 3;
  $190 = (($172) + ($189)|0);
  HEAP32[$136>>2] = $190;
  $191 = $188 & 7;
  HEAP32[$137>>2] = $191;
  $192 = ((($scf)) + 12|0);
  HEAP32[$174>>2] = $187;
  $193 = HEAP8[$190>>0]|0;
  $194 = $193&255;
  $195 = HEAP32[$137>>2]|0;
  $196 = $194 << $195;
  $197 = $196 & 255;
  HEAP32[$138>>2] = $197;
  $198 = ((($190)) + 1|0);
  $199 = HEAP8[$198>>0]|0;
  $200 = $199&255;
  $201 = $200 << $195;
  $202 = $201 >>> 8;
  $203 = $202 | $197;
  $204 = $203 << $4;
  $205 = $204 >>> 8;
  HEAP32[$138>>2] = $205;
  $206 = (($195) + ($4))|0;
  $207 = $206 >> 3;
  $208 = (($190) + ($207)|0);
  HEAP32[$136>>2] = $208;
  $209 = $206 & 7;
  HEAP32[$137>>2] = $209;
  $210 = ((($scf)) + 16|0);
  HEAP32[$192>>2] = $205;
  $211 = HEAP8[$208>>0]|0;
  $212 = $211&255;
  $213 = HEAP32[$137>>2]|0;
  $214 = $212 << $213;
  $215 = $214 & 255;
  HEAP32[$138>>2] = $215;
  $216 = ((($208)) + 1|0);
  $217 = HEAP8[$216>>0]|0;
  $218 = $217&255;
  $219 = $218 << $213;
  $220 = $219 >>> 8;
  $221 = $220 | $215;
  $222 = $221 << $4;
  $223 = $222 >>> 8;
  HEAP32[$138>>2] = $223;
  $224 = (($213) + ($4))|0;
  $225 = $224 >> 3;
  $226 = (($208) + ($225)|0);
  HEAP32[$136>>2] = $226;
  $227 = $224 & 7;
  HEAP32[$137>>2] = $227;
  $228 = ((($scf)) + 20|0);
  HEAP32[$210>>2] = $223;
  $229 = HEAP8[$226>>0]|0;
  $230 = $229&255;
  $231 = HEAP32[$137>>2]|0;
  $232 = $230 << $231;
  $233 = $232 & 255;
  HEAP32[$138>>2] = $233;
  $234 = ((($226)) + 1|0);
  $235 = HEAP8[$234>>0]|0;
  $236 = $235&255;
  $237 = $236 << $231;
  $238 = $237 >>> 8;
  $239 = $238 | $233;
  $240 = $239 << $4;
  $241 = $240 >>> 8;
  HEAP32[$138>>2] = $241;
  $242 = (($231) + ($4))|0;
  $243 = $242 >> 3;
  $244 = (($226) + ($243)|0);
  HEAP32[$136>>2] = $244;
  $245 = $242 & 7;
  HEAP32[$137>>2] = $245;
  HEAP32[$228>>2] = $241;
  $246 = ($4*6)|0;
  $numbits$1 = $246;
 } else {
  $numbits$1 = 0;
 }
 $247 = $84 & 4;
 $248 = ($247|0)==(0);
 if ($248) {
  $$6 = ((($scf)) + 24|0);
  $249 = ((($fr)) + 19788|0);
  $250 = ((($fr)) + 19784|0);
  $251 = ((($fr)) + 19792|0);
  $252 = HEAP32[$249>>2]|0;
  $253 = HEAP8[$252>>0]|0;
  $254 = $253&255;
  $255 = HEAP32[$250>>2]|0;
  $256 = $254 << $255;
  $257 = $256 & 255;
  HEAP32[$251>>2] = $257;
  $258 = ((($252)) + 1|0);
  $259 = HEAP8[$258>>0]|0;
  $260 = $259&255;
  $261 = $260 << $255;
  $262 = $261 >>> 8;
  $263 = $262 | $257;
  $264 = $263 << $4;
  $265 = $264 >>> 8;
  HEAP32[$251>>2] = $265;
  $266 = (($255) + ($4))|0;
  $267 = $266 >> 3;
  $268 = (($252) + ($267)|0);
  HEAP32[$249>>2] = $268;
  $269 = $266 & 7;
  HEAP32[$250>>2] = $269;
  $270 = ((($scf)) + 28|0);
  HEAP32[$$6>>2] = $265;
  $271 = HEAP8[$268>>0]|0;
  $272 = $271&255;
  $273 = HEAP32[$250>>2]|0;
  $274 = $272 << $273;
  $275 = $274 & 255;
  HEAP32[$251>>2] = $275;
  $276 = ((($268)) + 1|0);
  $277 = HEAP8[$276>>0]|0;
  $278 = $277&255;
  $279 = $278 << $273;
  $280 = $279 >>> 8;
  $281 = $280 | $275;
  $282 = $281 << $4;
  $283 = $282 >>> 8;
  HEAP32[$251>>2] = $283;
  $284 = (($273) + ($4))|0;
  $285 = $284 >> 3;
  $286 = (($268) + ($285)|0);
  HEAP32[$249>>2] = $286;
  $287 = $284 & 7;
  HEAP32[$250>>2] = $287;
  $288 = ((($scf)) + 32|0);
  HEAP32[$270>>2] = $283;
  $289 = HEAP8[$286>>0]|0;
  $290 = $289&255;
  $291 = HEAP32[$250>>2]|0;
  $292 = $290 << $291;
  $293 = $292 & 255;
  HEAP32[$251>>2] = $293;
  $294 = ((($286)) + 1|0);
  $295 = HEAP8[$294>>0]|0;
  $296 = $295&255;
  $297 = $296 << $291;
  $298 = $297 >>> 8;
  $299 = $298 | $293;
  $300 = $299 << $4;
  $301 = $300 >>> 8;
  HEAP32[$251>>2] = $301;
  $302 = (($291) + ($4))|0;
  $303 = $302 >> 3;
  $304 = (($286) + ($303)|0);
  HEAP32[$249>>2] = $304;
  $305 = $302 & 7;
  HEAP32[$250>>2] = $305;
  $306 = ((($scf)) + 36|0);
  HEAP32[$288>>2] = $301;
  $307 = HEAP8[$304>>0]|0;
  $308 = $307&255;
  $309 = HEAP32[$250>>2]|0;
  $310 = $308 << $309;
  $311 = $310 & 255;
  HEAP32[$251>>2] = $311;
  $312 = ((($304)) + 1|0);
  $313 = HEAP8[$312>>0]|0;
  $314 = $313&255;
  $315 = $314 << $309;
  $316 = $315 >>> 8;
  $317 = $316 | $311;
  $318 = $317 << $4;
  $319 = $318 >>> 8;
  HEAP32[$251>>2] = $319;
  $320 = (($309) + ($4))|0;
  $321 = $320 >> 3;
  $322 = (($304) + ($321)|0);
  HEAP32[$249>>2] = $322;
  $323 = $320 & 7;
  HEAP32[$250>>2] = $323;
  $324 = ((($scf)) + 40|0);
  HEAP32[$306>>2] = $319;
  $325 = HEAP8[$322>>0]|0;
  $326 = $325&255;
  $327 = HEAP32[$250>>2]|0;
  $328 = $326 << $327;
  $329 = $328 & 255;
  HEAP32[$251>>2] = $329;
  $330 = ((($322)) + 1|0);
  $331 = HEAP8[$330>>0]|0;
  $332 = $331&255;
  $333 = $332 << $327;
  $334 = $333 >>> 8;
  $335 = $334 | $329;
  $336 = $335 << $4;
  $337 = $336 >>> 8;
  HEAP32[$251>>2] = $337;
  $338 = (($327) + ($4))|0;
  $339 = $338 >> 3;
  $340 = (($322) + ($339)|0);
  HEAP32[$249>>2] = $340;
  $341 = $338 & 7;
  HEAP32[$250>>2] = $341;
  HEAP32[$324>>2] = $337;
  $342 = ($4*5)|0;
  $343 = (($numbits$1) + ($342))|0;
  $numbits$2 = $343;
 } else {
  $numbits$2 = $numbits$1;
 }
 $344 = $84 & 2;
 $345 = ($344|0)==(0);
 if ($345) {
  $$8 = ((($scf)) + 44|0);
  $346 = ((($fr)) + 19788|0);
  $347 = ((($fr)) + 19784|0);
  $348 = ((($fr)) + 19792|0);
  $349 = HEAP32[$346>>2]|0;
  $350 = HEAP8[$349>>0]|0;
  $351 = $350&255;
  $352 = HEAP32[$347>>2]|0;
  $353 = $351 << $352;
  $354 = $353 & 255;
  HEAP32[$348>>2] = $354;
  $355 = ((($349)) + 1|0);
  $356 = HEAP8[$355>>0]|0;
  $357 = $356&255;
  $358 = $357 << $352;
  $359 = $358 >>> 8;
  $360 = $359 | $354;
  $361 = $360 << $7;
  $362 = $361 >>> 8;
  HEAP32[$348>>2] = $362;
  $363 = (($352) + ($7))|0;
  $364 = $363 >> 3;
  $365 = (($349) + ($364)|0);
  HEAP32[$346>>2] = $365;
  $366 = $363 & 7;
  HEAP32[$347>>2] = $366;
  $367 = ((($scf)) + 48|0);
  HEAP32[$$8>>2] = $362;
  $368 = HEAP8[$365>>0]|0;
  $369 = $368&255;
  $370 = HEAP32[$347>>2]|0;
  $371 = $369 << $370;
  $372 = $371 & 255;
  HEAP32[$348>>2] = $372;
  $373 = ((($365)) + 1|0);
  $374 = HEAP8[$373>>0]|0;
  $375 = $374&255;
  $376 = $375 << $370;
  $377 = $376 >>> 8;
  $378 = $377 | $372;
  $379 = $378 << $7;
  $380 = $379 >>> 8;
  HEAP32[$348>>2] = $380;
  $381 = (($370) + ($7))|0;
  $382 = $381 >> 3;
  $383 = (($365) + ($382)|0);
  HEAP32[$346>>2] = $383;
  $384 = $381 & 7;
  HEAP32[$347>>2] = $384;
  $385 = ((($scf)) + 52|0);
  HEAP32[$367>>2] = $380;
  $386 = HEAP8[$383>>0]|0;
  $387 = $386&255;
  $388 = HEAP32[$347>>2]|0;
  $389 = $387 << $388;
  $390 = $389 & 255;
  HEAP32[$348>>2] = $390;
  $391 = ((($383)) + 1|0);
  $392 = HEAP8[$391>>0]|0;
  $393 = $392&255;
  $394 = $393 << $388;
  $395 = $394 >>> 8;
  $396 = $395 | $390;
  $397 = $396 << $7;
  $398 = $397 >>> 8;
  HEAP32[$348>>2] = $398;
  $399 = (($388) + ($7))|0;
  $400 = $399 >> 3;
  $401 = (($383) + ($400)|0);
  HEAP32[$346>>2] = $401;
  $402 = $399 & 7;
  HEAP32[$347>>2] = $402;
  $403 = ((($scf)) + 56|0);
  HEAP32[$385>>2] = $398;
  $404 = HEAP8[$401>>0]|0;
  $405 = $404&255;
  $406 = HEAP32[$347>>2]|0;
  $407 = $405 << $406;
  $408 = $407 & 255;
  HEAP32[$348>>2] = $408;
  $409 = ((($401)) + 1|0);
  $410 = HEAP8[$409>>0]|0;
  $411 = $410&255;
  $412 = $411 << $406;
  $413 = $412 >>> 8;
  $414 = $413 | $408;
  $415 = $414 << $7;
  $416 = $415 >>> 8;
  HEAP32[$348>>2] = $416;
  $417 = (($406) + ($7))|0;
  $418 = $417 >> 3;
  $419 = (($401) + ($418)|0);
  HEAP32[$346>>2] = $419;
  $420 = $417 & 7;
  HEAP32[$347>>2] = $420;
  $421 = ((($scf)) + 60|0);
  HEAP32[$403>>2] = $416;
  $422 = HEAP8[$419>>0]|0;
  $423 = $422&255;
  $424 = HEAP32[$347>>2]|0;
  $425 = $423 << $424;
  $426 = $425 & 255;
  HEAP32[$348>>2] = $426;
  $427 = ((($419)) + 1|0);
  $428 = HEAP8[$427>>0]|0;
  $429 = $428&255;
  $430 = $429 << $424;
  $431 = $430 >>> 8;
  $432 = $431 | $426;
  $433 = $432 << $7;
  $434 = $433 >>> 8;
  HEAP32[$348>>2] = $434;
  $435 = (($424) + ($7))|0;
  $436 = $435 >> 3;
  $437 = (($419) + ($436)|0);
  HEAP32[$346>>2] = $437;
  $438 = $435 & 7;
  HEAP32[$347>>2] = $438;
  HEAP32[$421>>2] = $434;
  $439 = ($7*5)|0;
  $440 = (($numbits$2) + ($439))|0;
  $numbits$3 = $440;
 } else {
  $numbits$3 = $numbits$2;
 }
 $441 = $84 & 1;
 $442 = ($441|0)==(0);
 if ($442) {
  $$10 = ((($scf)) + 64|0);
  $443 = ((($fr)) + 19788|0);
  $444 = ((($fr)) + 19784|0);
  $445 = ((($fr)) + 19792|0);
  $446 = HEAP32[$443>>2]|0;
  $447 = HEAP8[$446>>0]|0;
  $448 = $447&255;
  $449 = HEAP32[$444>>2]|0;
  $450 = $448 << $449;
  $451 = $450 & 255;
  HEAP32[$445>>2] = $451;
  $452 = ((($446)) + 1|0);
  $453 = HEAP8[$452>>0]|0;
  $454 = $453&255;
  $455 = $454 << $449;
  $456 = $455 >>> 8;
  $457 = $456 | $451;
  $458 = $457 << $7;
  $459 = $458 >>> 8;
  HEAP32[$445>>2] = $459;
  $460 = (($449) + ($7))|0;
  $461 = $460 >> 3;
  $462 = (($446) + ($461)|0);
  HEAP32[$443>>2] = $462;
  $463 = $460 & 7;
  HEAP32[$444>>2] = $463;
  $464 = ((($scf)) + 68|0);
  HEAP32[$$10>>2] = $459;
  $465 = HEAP8[$462>>0]|0;
  $466 = $465&255;
  $467 = HEAP32[$444>>2]|0;
  $468 = $466 << $467;
  $469 = $468 & 255;
  HEAP32[$445>>2] = $469;
  $470 = ((($462)) + 1|0);
  $471 = HEAP8[$470>>0]|0;
  $472 = $471&255;
  $473 = $472 << $467;
  $474 = $473 >>> 8;
  $475 = $474 | $469;
  $476 = $475 << $7;
  $477 = $476 >>> 8;
  HEAP32[$445>>2] = $477;
  $478 = (($467) + ($7))|0;
  $479 = $478 >> 3;
  $480 = (($462) + ($479)|0);
  HEAP32[$443>>2] = $480;
  $481 = $478 & 7;
  HEAP32[$444>>2] = $481;
  $482 = ((($scf)) + 72|0);
  HEAP32[$464>>2] = $477;
  $483 = HEAP8[$480>>0]|0;
  $484 = $483&255;
  $485 = HEAP32[$444>>2]|0;
  $486 = $484 << $485;
  $487 = $486 & 255;
  HEAP32[$445>>2] = $487;
  $488 = ((($480)) + 1|0);
  $489 = HEAP8[$488>>0]|0;
  $490 = $489&255;
  $491 = $490 << $485;
  $492 = $491 >>> 8;
  $493 = $492 | $487;
  $494 = $493 << $7;
  $495 = $494 >>> 8;
  HEAP32[$445>>2] = $495;
  $496 = (($485) + ($7))|0;
  $497 = $496 >> 3;
  $498 = (($480) + ($497)|0);
  HEAP32[$443>>2] = $498;
  $499 = $496 & 7;
  HEAP32[$444>>2] = $499;
  $500 = ((($scf)) + 76|0);
  HEAP32[$482>>2] = $495;
  $501 = HEAP8[$498>>0]|0;
  $502 = $501&255;
  $503 = HEAP32[$444>>2]|0;
  $504 = $502 << $503;
  $505 = $504 & 255;
  HEAP32[$445>>2] = $505;
  $506 = ((($498)) + 1|0);
  $507 = HEAP8[$506>>0]|0;
  $508 = $507&255;
  $509 = $508 << $503;
  $510 = $509 >>> 8;
  $511 = $510 | $505;
  $512 = $511 << $7;
  $513 = $512 >>> 8;
  HEAP32[$445>>2] = $513;
  $514 = (($503) + ($7))|0;
  $515 = $514 >> 3;
  $516 = (($498) + ($515)|0);
  HEAP32[$443>>2] = $516;
  $517 = $514 & 7;
  HEAP32[$444>>2] = $517;
  $518 = ((($scf)) + 80|0);
  HEAP32[$500>>2] = $513;
  $519 = HEAP8[$516>>0]|0;
  $520 = $519&255;
  $521 = HEAP32[$444>>2]|0;
  $522 = $520 << $521;
  $523 = $522 & 255;
  HEAP32[$445>>2] = $523;
  $524 = ((($516)) + 1|0);
  $525 = HEAP8[$524>>0]|0;
  $526 = $525&255;
  $527 = $526 << $521;
  $528 = $527 >>> 8;
  $529 = $528 | $523;
  $530 = $529 << $7;
  $531 = $530 >>> 8;
  HEAP32[$445>>2] = $531;
  $532 = (($521) + ($7))|0;
  $533 = $532 >> 3;
  $534 = (($516) + ($533)|0);
  HEAP32[$443>>2] = $534;
  $535 = $532 & 7;
  HEAP32[$444>>2] = $535;
  HEAP32[$518>>2] = $531;
  $536 = ($7*5)|0;
  $537 = (($numbits$3) + ($536))|0;
  $numbits$4 = $537;
 } else {
  $numbits$4 = $numbits$3;
 }
 $$12 = ((($scf)) + 84|0);
 HEAP32[$$12>>2] = 0;
 $numbits$6 = $numbits$4;
 return ($numbits$6|0);
}
function _III_dequantize_sample($fr,$xr,$scf,$gr_info,$sfreq,$part2bits) {
 $fr = $fr|0;
 $xr = $xr|0;
 $scf = $scf|0;
 $gr_info = $gr_info|0;
 $sfreq = $sfreq|0;
 $part2bits = $part2bits|0;
 var $$02226 = 0, $$10$lcssa = 0, $$10$lcssa$lcssa = 0, $$10313 = 0, $$11 = 0, $$11$lcssa = 0, $$12265 = 0, $$13$lcssa = 0, $$13$lcssa$lcssa = 0, $$13200 = 0, $$14 = 0, $$14$2 = 0, $$15$lcssa = 0, $$24 = 0, $$35148 = 0, $$4 = 0, $$46135 = 0, $$5 = 0, $$6 = 0, $$6$lcssa = 0;
 var $$9339 = 0, $$in = 0, $$lcssa = 0, $$lcssa100 = 0, $$lcssa573 = 0, $$lcssa574 = 0, $$lcssa576 = 0, $$lcssa577 = 0, $$lcssa589 = 0, $$lcssa592 = 0, $$lcssa593 = 0, $$lcssa597 = 0, $$lcssa598 = 0, $$lcssa599 = 0, $$lcssa600 = 0, $$lcssa604 = 0, $$lcssa605 = 0, $$lcssa606 = 0, $$lcssa612 = 0, $$lcssa615 = 0;
 var $$lcssa616 = 0, $$lcssa620 = 0, $$lcssa621 = 0, $$lcssa622 = 0, $$lcssa623 = 0, $$lcssa624 = 0, $$lcssa625 = 0, $$lcssa626 = 0, $$lcssa92 = 0, $$lcssa93 = 0, $$neg56 = 0, $$neg58 = 0, $$neg60 = 0, $$neg62 = 0, $$phi$trans$insert = 0, $$phi$trans$insert405 = 0, $$pn = 0.0, $$pn13 = 0.0, $$pn15 = 0.0, $$pn17 = 0.0;
 var $$pn20 = 0.0, $$pn22 = 0.0, $$pn24 = 0.0, $$pn26 = 0.0, $$pre = 0, $$pre406 = 0, $$pre407 = 0, $$pre408 = 0, $$pre409 = 0, $$sink403 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0.0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0.0, $16 = 0, $160 = 0.0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0.0, $166 = 0.0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0;
 var $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0.0, $196 = 0.0, $197 = 0;
 var $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0.0, $202 = 0.0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0;
 var $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0;
 var $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0.0;
 var $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0.0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0.0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0;
 var $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0;
 var $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0;
 var $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0;
 var $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0;
 var $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0.0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0;
 var $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0;
 var $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0;
 var $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0.0, $421 = 0.0, $422 = 0, $423 = 0, $424 = 0, $425 = 0.0, $426 = 0.0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0;
 var $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0;
 var $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0.0, $454 = 0.0, $455 = 0, $456 = 0, $457 = 0, $458 = 0.0, $459 = 0.0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0;
 var $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0;
 var $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0;
 var $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0.0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0.0, $52 = 0, $520 = 0;
 var $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0;
 var $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0;
 var $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0.0, $573 = 0, $574 = 0, $575 = 0;
 var $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0.0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0;
 var $594 = 0, $595 = 0, $596 = 0.0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0.0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0;
 var $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0.0, $99 = 0;
 var $cb$0220 = 0, $cb$1$lcssa = 0, $cb$1$lcssa$lcssa = 0, $cb$1193 = 0, $cb$2 = 0, $cb$3141 = 0, $cb$4130 = 0, $cb$5 = 0, $cb$6 = 0, $cb$6$lcssa = 0, $cb6$0334 = 0, $cb6$1$lcssa = 0, $cb6$1$lcssa$lcssa = 0, $cb6$1307 = 0, $cb6$2 = 0, $cb6$2$lcssa = 0, $cb6$3260 = 0, $cb6$5 = 0, $cb6$5$2 = 0, $cb6$6$lcssa = 0;
 var $exitcond = 0, $exitcond394 = 0, $i$0217 = 0, $i$1127 = 0, $i4$0332 = 0, $l = 0, $l3$0146 = 0, $l3$1257 = 0, $lp$0197 = 0, $lp10$0311 = 0, $lwin$0219 = 0, $lwin$1$lcssa = 0, $lwin$1$lcssa$lcssa = 0, $lwin$1192 = 0, $lwin$2 = 0, $lwin$3140 = 0, $lwin$3140$lcssa587 = 0, $lwin$378 = 0, $lwin$4129 = 0, $lwin$5 = 0;
 var $lwin$6 = 0, $lwin$6$lcssa = 0, $m$0 = 0, $m$0$in = 0, $m$11$ph = 0, $m$1222 = 0, $m$2$lcssa = 0, $m$2$lcssa$lcssa = 0, $m$2195 = 0, $m$3 = 0, $m$4143 = 0, $m$4143$lcssa585 = 0, $m$481 = 0, $m$5132 = 0, $m$6 = 0, $m$7 = 0, $m$7$lcssa = 0, $m7$0335 = 0, $m7$1$lcssa = 0, $m7$1$lcssa$lcssa = 0;
 var $m7$1308 = 0, $m7$2 = 0, $m7$2$lcssa = 0, $m7$3261 = 0, $m7$5 = 0, $m7$5$2 = 0, $m7$6$lcssa = 0, $mask$0215 = 0, $mask$1$lcssa = 0, $mask$1$lcssa$lcssa = 0, $mask$10$lcssa = 0, $mask$10118 = 0, $mask$11$lcssa = 0, $mask$11125 = 0, $mask$1189 = 0, $mask$12 = 0, $mask$14328 = 0, $mask$15$lcssa = 0, $mask$15$lcssa$lcssa = 0, $mask$15302 = 0;
 var $mask$16$lcssa = 0, $mask$16271 = 0, $mask$17$lcssa = 0, $mask$17279 = 0, $mask$18$lcssa = 0, $mask$18286 = 0, $mask$19 = 0, $mask$2$lcssa = 0, $mask$20$lcssa = 0, $mask$20294 = 0, $mask$21 = 0, $mask$21$lcssa = 0, $mask$2156 = 0, $mask$22254 = 0, $mask$23$lcssa = 0, $mask$23227 = 0, $mask$24$lcssa = 0, $mask$24235 = 0, $mask$25$lcssa = 0, $mask$26 = 0;
 var $mask$26$1 = 0, $mask$26$2 = 0, $mask$26$3 = 0, $mask$3$lcssa = 0, $mask$3164 = 0, $mask$4$lcssa = 0, $mask$4171 = 0, $mask$5 = 0, $mask$6$lcssa = 0, $mask$6180 = 0, $mask$7 = 0, $mask$8137 = 0, $mask$9$lcssa = 0, $mask$9110 = 0, $max = 0, $max5$0333 = 0, $max5$1$lcssa = 0, $max5$1$lcssa$lcssa = 0, $max5$1306 = 0, $max5$2 = 0;
 var $max5$3 = 0, $max5$3$lcssa = 0, $max5$4259 = 0, $max5$4259$lcssa = 0, $max5$497 = 0, $max5$6 = 0, $max5$6$1 = 0, $max5$6$2 = 0, $max5$6$3 = 0, $max5$8$ph = 0, $mc$0223 = 0, $mc$1$lcssa = 0, $mc$1$lcssa$lcssa = 0, $mc$10$ph = 0, $mc$10105 = 0, $mc$1196 = 0, $mc$2 = 0, $mc$3144 = 0, $mc$3144$lcssa583 = 0, $mc$384 = 0;
 var $mc$4133 = 0, $mc$5 = 0, $mc$6 = 0, $mc$6$lcssa = 0, $mc9$0337 = 0, $mc9$1$lcssa = 0, $mc9$1$lcssa$lcssa = 0, $mc9$1310 = 0, $mc9$2 = 0, $mc9$3263 = 0, $mc9$5 = 0, $mc9$5$2 = 0, $mc9$6$lcssa = 0, $me$0 = 0, $me$0$in = 0, $num$0216 = 0, $num$1$lcssa = 0, $num$1$lcssa$lcssa = 0, $num$10$lcssa = 0, $num$10119 = 0;
 var $num$11$lcssa = 0, $num$11126 = 0, $num$1190 = 0, $num$12 = 0, $num$14 = 0, $num$16329 = 0, $num$17$lcssa = 0, $num$17$lcssa$lcssa = 0, $num$17303 = 0, $num$18$lcssa = 0, $num$18272 = 0, $num$19$lcssa = 0, $num$19280 = 0, $num$2$lcssa = 0, $num$20$lcssa = 0, $num$20287 = 0, $num$21 = 0, $num$2157 = 0, $num$22$lcssa = 0, $num$22295 = 0;
 var $num$23 = 0, $num$23$lcssa = 0, $num$24255 = 0, $num$25$lcssa = 0, $num$25228 = 0, $num$26$lcssa = 0, $num$26236 = 0, $num$27$lcssa = 0, $num$28 = 0, $num$28$1 = 0, $num$28$2 = 0, $num$28$3 = 0, $num$3$lcssa = 0, $num$30 = 0, $num$31 = 0, $num$3165 = 0, $num$4$lcssa = 0, $num$4172 = 0, $num$5 = 0, $num$6$lcssa = 0;
 var $num$6181 = 0, $num$7 = 0, $num$8138 = 0, $num$9$lcssa = 0, $num$9111 = 0, $or$cond = 0, $or$cond136 = 0, $or$cond29 = 0, $or$cond29253 = 0, $part2remain$0224 = 0, $part2remain$1$lcssa = 0, $part2remain$1$lcssa$lcssa = 0, $part2remain$10 = 0, $part2remain$1198 = 0, $part2remain$12330 = 0, $part2remain$13$lcssa = 0, $part2remain$13$lcssa$lcssa = 0, $part2remain$13304 = 0, $part2remain$14$lcssa = 0, $part2remain$15$lcssa = 0;
 var $part2remain$16 = 0, $part2remain$17$lcssa = 0, $part2remain$18 = 0, $part2remain$18$lcssa = 0, $part2remain$19256 = 0, $part2remain$2$lcssa = 0, $part2remain$20$lcssa = 0, $part2remain$20$lcssa$lcssa = 0, $part2remain$21 = 0, $part2remain$22 = 0, $part2remain$23$lcssa = 0, $part2remain$23101 = 0, $part2remain$3$lcssa = 0, $part2remain$4 = 0, $part2remain$5$lcssa = 0, $part2remain$6 = 0, $part2remain$7145 = 0, $part2remain$8$lcssa = 0, $part2remain$8$lcssa$lcssa = 0, $pretab$0331 = 0;
 var $pretab$1$lcssa = 0, $pretab$1$lcssa$lcssa = 0, $pretab$1305 = 0, $pretab$2 = 0, $pretab$2$lcssa = 0, $pretab$3258 = 0, $pretab$5 = 0, $pretab$5$2 = 0, $pretab$6$lcssa = 0, $scevgep = 0, $scevgep396 = 0, $step$0218 = 0, $step$1$lcssa = 0, $step$1$lcssa$lcssa = 0, $step$1191 = 0, $step$2 = 0, $step$3139 = 0, $step$4128 = 0, $step$5 = 0, $step$6 = 0;
 var $step$6$lcssa = 0, $storemerge = 0.0, $storemerge$1 = 0.0, $storemerge$2 = 0.0, $storemerge$3 = 0.0, $storemerge11 = 0.0, $storemerge12 = 0.0, $storemerge14 = 0.0, $storemerge16 = 0.0, $storemerge18 = 0.0, $storemerge19 = 0.0, $storemerge21 = 0.0, $storemerge23 = 0.0, $storemerge25 = 0.0, $uglygep = 0, $uglygep399 = 0, $v$0221 = 0.0, $v$1$lcssa = 0.0, $v$1$lcssa$lcssa = 0.0, $v$1194 = 0.0;
 var $v$2 = 0.0, $v$3142 = 0.0, $v$4131 = 0.0, $v$5 = 0.0, $v$6 = 0.0, $v$6$lcssa = 0.0, $v8$0336 = 0.0, $v8$1$lcssa = 0.0, $v8$1$lcssa$lcssa = 0.0, $v8$1309 = 0.0, $v8$2 = 0.0, $v8$2$lcssa = 0.0, $v8$3262 = 0.0, $v8$5 = 0.0, $v8$5$2 = 0.0, $v8$6$lcssa = 0.0, $val$1 = 0, $val$1$pn = 0, $val14$1 = 0, $val14$1$pn = 0;
 var $val16$1 = 0, $val16$1$pn = 0, $val2$1 = 0, $val2$1$pn = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $xrpnt$0225 = 0, $xrpnt$1$lcssa = 0, $xrpnt$1$lcssa$lcssa = 0, $xrpnt$10$lcssa = 0, $xrpnt$10$ph = 0, $xrpnt$10106 = 0, $xrpnt$11 = 0, $xrpnt$1199 = 0, $xrpnt$13338 = 0, $xrpnt$14$lcssa = 0, $xrpnt$14$lcssa$lcssa = 0, $xrpnt$14312 = 0, $xrpnt$17 = 0;
 var $xrpnt$2 = 0, $xrpnt$20 = 0, $xrpnt$21264 = 0, $xrpnt$21264$lcssa = 0, $xrpnt$2199 = 0, $xrpnt$22$lcssa = 0, $xrpnt$24 = 0, $xrpnt$24$1 = 0, $xrpnt$24$2 = 0, $xrpnt$24$3 = 0, $xrpnt$27 = 0, $xrpnt$27398 = 0, $xrpnt$3147 = 0, $xrpnt$3147$lcssa = 0, $xrpnt$3147$lcssa581 = 0, $xrpnt$389 = 0, $xrpnt$4134 = 0, $xrpnt$5 = 0, $xrpnt$6 = 0, $xrpnt$8$ph = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $l = sp + 32|0;
 $max = sp + 16|0;
 $0 = ((($gr_info)) + 68|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + 1)|0;
 $3 = ((($gr_info)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) - ($part2bits))|0;
 $6 = ((($fr)) + 19784|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (0 - ($7))|0;
 $9 = $8 & 7;
 $10 = ((($fr)) + 19788|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = HEAP8[$11>>0]|0;
 $13 = $12&255;
 $14 = $13 << 8;
 $15 = ((($11)) + 1|0);
 $16 = HEAP8[$15>>0]|0;
 $17 = $16&255;
 $18 = $14 | $17;
 $19 = $18 << 8;
 $20 = ((($11)) + 2|0);
 $21 = HEAP8[$20>>0]|0;
 $22 = $21&255;
 $23 = $19 | $22;
 $24 = $23 << $7;
 $25 = $24 & 16777215;
 $26 = (($9) + ($7))|0;
 $27 = (24 - ($9))|0;
 $28 = $25 >>> $27;
 $29 = $26 >> 3;
 $30 = (($11) + ($29)|0);
 HEAP32[$10>>2] = $30;
 HEAP32[$6>>2] = 0;
 $31 = $28 << 24;
 $32 = (8 - ($9))|0;
 $33 = $31 << $32;
 $34 = (($5) - ($9))|0;
 $35 = ((($gr_info)) + 8|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = ((($gr_info)) + 56|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = ((($gr_info)) + 60|0);
 $40 = HEAP32[$39>>2]|0;
 $41 = (288 - ($36))|0;
 $42 = $41 >> 1;
 $43 = ($36|0)>($38|0);
 do {
  if ($43) {
   HEAP32[$l>>2] = $38;
   $46 = ($36|0)>($40|0);
   if ($46) {
    $50 = (($40) - ($38))|0;
    $51 = ((($l)) + 4|0);
    HEAP32[$51>>2] = $50;
    $52 = (($36) - ($40))|0;
    $53 = ((($l)) + 8|0);
    HEAP32[$53>>2] = $52;
    $607 = $38;
    break;
   } else {
    $47 = (($36) - ($38))|0;
    $48 = ((($l)) + 4|0);
    HEAP32[$48>>2] = $47;
    $49 = ((($l)) + 8|0);
    HEAP32[$49>>2] = 0;
    $607 = $38;
    break;
   }
  } else {
   HEAP32[$l>>2] = $36;
   $44 = ((($l)) + 4|0);
   HEAP32[$44>>2] = 0;
   $45 = ((($l)) + 8|0);
   HEAP32[$45>>2] = 0;
   $607 = $36;
  }
 } while(0);
 $54 = ((($gr_info)) + 16|0);
 $55 = HEAP32[$54>>2]|0;
 $56 = ($55|0)==(2);
 if ($56) {
  $57 = ((($gr_info)) + 20|0);
  $58 = HEAP32[$57>>2]|0;
  $59 = ($58|0)==(0);
  $60 = ((($max)) + 12|0);
  HEAP32[$60>>2] = -1;
  $61 = ((($max)) + 8|0);
  if ($59) {
   HEAP32[$61>>2] = -1;
   $64 = (((684496 + (($sfreq*12)|0)|0)) + 4|0);
   $65 = (((684604 + (($sfreq*12)|0)|0)) + 4|0);
   $$sink403 = -1;$m$0$in = $64;$me$0$in = $65;
  } else {
   HEAP32[$61>>2] = 2;
   $62 = (684496 + (($sfreq*12)|0)|0);
   $63 = (684604 + (($sfreq*12)|0)|0);
   $$sink403 = 2;$m$0$in = $62;$me$0$in = $63;
  }
  $66 = ((($max)) + 4|0);
  HEAP32[$66>>2] = $$sink403;
  HEAP32[$max>>2] = $$sink403;
  $me$0 = HEAP32[$me$0$in>>2]|0;
  $m$0 = HEAP32[$m$0$in>>2]|0;
  $67 = ((($gr_info)) + 88|0);
  $$02226 = $scf;$608 = $30;$76 = $607;$cb$0220 = 0;$i$0217 = 0;$lwin$0219 = 3;$m$1222 = $m$0;$mask$0215 = $33;$mc$0223 = 0;$num$0216 = $9;$part2remain$0224 = $34;$step$0218 = 0;$v$0221 = 0.0;$xrpnt$0225 = $xr;
  while(1) {
   $73 = (((($gr_info)) + 24|0) + ($i$0217<<2)|0);
   $74 = HEAP32[$73>>2]|0;
   $75 = ($76|0)==(0);
   L16: do {
    if ($75) {
     $$13$lcssa = $$02226;$611 = $608;$cb$1$lcssa = $cb$0220;$lwin$1$lcssa = $lwin$0219;$m$2$lcssa = $m$1222;$mask$1$lcssa = $mask$0215;$mc$1$lcssa = $mc$0223;$num$1$lcssa = $num$0216;$part2remain$1$lcssa = $part2remain$0224;$step$1$lcssa = $step$0218;$v$1$lcssa = $v$0221;$xrpnt$1$lcssa = $xrpnt$0225;
    } else {
     $77 = (((36544 + ($74<<3)|0)) + 4|0);
     $78 = HEAP32[$77>>2]|0;
     $79 = (36544 + ($74<<3)|0);
     $80 = ($74>>>0)<(16);
     $81 = ($74>>>0)<(16);
     $$13200 = $$02226;$612 = $608;$cb$1193 = $cb$0220;$lp$0197 = $76;$lwin$1192 = $lwin$0219;$m$2195 = $m$1222;$mask$1189 = $mask$0215;$mc$1196 = $mc$0223;$num$1190 = $num$0216;$part2remain$1198 = $part2remain$0224;$step$1191 = $step$0218;$v$1194 = $v$0221;$xrpnt$1199 = $xrpnt$0225;
     while(1) {
      $82 = ($mc$1196|0)==(0);
      do {
       if ($82) {
        $83 = ((($m$2195)) + 4|0);
        $84 = HEAP32[$m$2195>>2]|0;
        $85 = ((($m$2195)) + 8|0);
        $86 = HEAP32[$83>>2]|0;
        $87 = (($xr) + ($86<<2)|0);
        $88 = ((($m$2195)) + 12|0);
        $89 = HEAP32[$85>>2]|0;
        $90 = ((($m$2195)) + 16|0);
        $91 = HEAP32[$88>>2]|0;
        $92 = ($89|0)==(3);
        $93 = ((($$13200)) + 4|0);
        $94 = HEAP32[$$13200>>2]|0;
        $95 = $94 << $2;
        if ($92) {
         $96 = HEAP32[$67>>2]|0;
         $97 = (($96) + ($95<<2)|0);
         $98 = +HEAPF32[$97>>2];
         $$24 = $93;$cb$2 = $91;$lwin$2 = 3;$m$3 = $90;$mc$2 = $84;$step$2 = 1;$v$2 = $98;$xrpnt$2 = $87;
         break;
        } else {
         $99 = (((($gr_info)) + 76|0) + ($89<<2)|0);
         $100 = HEAP32[$99>>2]|0;
         $101 = (($100) + ($95<<2)|0);
         $102 = +HEAPF32[$101>>2];
         $$24 = $93;$cb$2 = $91;$lwin$2 = $89;$m$3 = $90;$mc$2 = $84;$step$2 = 3;$v$2 = $102;$xrpnt$2 = $87;
         break;
        }
       } else {
        $$24 = $$13200;$cb$2 = $cb$1193;$lwin$2 = $lwin$1192;$m$3 = $m$2195;$mc$2 = $mc$1196;$step$2 = $step$1191;$v$2 = $v$1194;$xrpnt$2 = $xrpnt$1199;
       }
      } while(0);
      $103 = ($num$1190>>>0)<(24);
      if ($103) {
       $104 = (23 - ($num$1190))|0;
       $105 = $104 & -8;
       $106 = (($num$1190) + 8)|0;
       $114 = $612;$mask$2156 = $mask$1189;$num$2157 = $num$1190;
       while(1) {
        $113 = ((($114)) + 1|0);
        HEAP32[$10>>2] = $113;
        $115 = HEAP8[$114>>0]|0;
        $116 = $115&255;
        $117 = (24 - ($num$2157))|0;
        $118 = $116 << $117;
        $119 = $118 | $mask$2156;
        $120 = (($num$2157) + 8)|0;
        $121 = ($120>>>0)<(24);
        if ($121) {
         $114 = $113;$mask$2156 = $119;$num$2157 = $120;
        } else {
         $$lcssa592 = $113;$$lcssa593 = $119;
         break;
        }
       }
       $107 = (($part2remain$1198) + -8)|0;
       $108 = (($107) - ($105))|0;
       $109 = (($106) + ($105))|0;
       $613 = $$lcssa592;$mask$2$lcssa = $$lcssa593;$num$2$lcssa = $109;$part2remain$2$lcssa = $108;
      } else {
       $613 = $612;$mask$2$lcssa = $mask$1189;$num$2$lcssa = $num$1190;$part2remain$2$lcssa = $part2remain$1198;
      }
      $110 = HEAP16[$78>>1]|0;
      $111 = $110 << 16 >> 16;
      $112 = ($110<<16>>16)<(0);
      if ($112) {
       $125 = $111;$mask$3164 = $mask$2$lcssa;$num$3165 = $num$2$lcssa;$val$1$pn = $78;
       while(1) {
        $122 = ((($val$1$pn)) + 2|0);
        $123 = ($mask$3164|0)<(0);
        $124 = (0 - ($125))|0;
        $126 = (($122) + ($124<<1)|0);
        $val$1 = $123 ? $126 : $122;
        $127 = (($num$3165) + -1)|0;
        $128 = $mask$3164 << 1;
        $129 = HEAP16[$val$1>>1]|0;
        $130 = $129 << 16 >> 16;
        $131 = ($129<<16>>16)<(0);
        if ($131) {
         $125 = $130;$mask$3164 = $128;$num$3165 = $127;$val$1$pn = $val$1;
        } else {
         $$lcssa92 = $130;$mask$3$lcssa = $128;$num$3$lcssa = $127;
         break;
        }
       }
      } else {
       $$lcssa92 = $111;$mask$3$lcssa = $mask$2$lcssa;$num$3$lcssa = $num$2$lcssa;
      }
      $132 = $$lcssa92 >> 4;
      $133 = $$lcssa92 & 15;
      switch ($132|0) {
      case 15:  {
       $134 = HEAP32[$79>>2]|0;
       if ($81) {
        label = 32;
       } else {
        $135 = (($max) + ($lwin$2<<2)|0);
        HEAP32[$135>>2] = $cb$2;
        $136 = ($num$3$lcssa>>>0)<(24);
        if ($136) {
         $137 = (23 - ($num$3$lcssa))|0;
         $138 = $137 & -8;
         $139 = (($num$3$lcssa) + 8)|0;
         $141 = $613;$mask$4171 = $mask$3$lcssa;$num$4172 = $num$3$lcssa;
         while(1) {
          $140 = ((($141)) + 1|0);
          HEAP32[$10>>2] = $140;
          $142 = HEAP8[$141>>0]|0;
          $143 = $142&255;
          $144 = (24 - ($num$4172))|0;
          $145 = $143 << $144;
          $146 = $145 | $mask$4171;
          $147 = (($num$4172) + 8)|0;
          $148 = ($147>>>0)<(24);
          if ($148) {
           $141 = $140;$mask$4171 = $146;$num$4172 = $147;
          } else {
           $$lcssa597 = $140;$$lcssa598 = $146;
           break;
          }
         }
         $149 = (($part2remain$2$lcssa) + -8)|0;
         $150 = (($149) - ($138))|0;
         $151 = (($139) + ($138))|0;
         $614 = $$lcssa597;$mask$4$lcssa = $$lcssa598;$num$4$lcssa = $151;$part2remain$3$lcssa = $150;
        } else {
         $614 = $613;$mask$4$lcssa = $mask$3$lcssa;$num$4$lcssa = $num$3$lcssa;$part2remain$3$lcssa = $part2remain$2$lcssa;
        }
        $152 = (32 - ($134))|0;
        $153 = $mask$4$lcssa >>> $152;
        $154 = (($153) + 15)|0;
        $$neg56 = (($num$4$lcssa) + -1)|0;
        $155 = (($$neg56) - ($134))|0;
        $156 = $mask$4$lcssa << $134;
        $157 = ($156|0)<(0);
        $158 = (684712 + ($154<<2)|0);
        $159 = +HEAPF32[$158>>2];
        $160 = -$159;
        $$pn26 = $157 ? $160 : $159;
        $storemerge25 = $v$2 * $$pn26;
        HEAPF32[$xrpnt$2>>2] = $storemerge25;
        $161 = $156 << 1;
        $615 = $614;$mask$5 = $161;$num$5 = $155;$part2remain$4 = $part2remain$3$lcssa;
       }
       break;
      }
      case 0:  {
       HEAPF32[$xrpnt$2>>2] = 0.0;
       $615 = $613;$mask$5 = $mask$3$lcssa;$num$5 = $num$3$lcssa;$part2remain$4 = $part2remain$2$lcssa;
       break;
      }
      default: {
       label = 32;
      }
      }
      if ((label|0) == 32) {
       label = 0;
       $162 = (($max) + ($lwin$2<<2)|0);
       HEAP32[$162>>2] = $cb$2;
       $163 = ($mask$3$lcssa|0)<(0);
       $164 = (684712 + ($132<<2)|0);
       $165 = +HEAPF32[$164>>2];
       $166 = -$165;
       $$pn24 = $163 ? $166 : $165;
       $storemerge23 = $v$2 * $$pn24;
       HEAPF32[$xrpnt$2>>2] = $storemerge23;
       $167 = (($num$3$lcssa) + -1)|0;
       $168 = $mask$3$lcssa << 1;
       $615 = $613;$mask$5 = $168;$num$5 = $167;$part2remain$4 = $part2remain$2$lcssa;
      }
      $169 = (($xrpnt$2) + ($step$2<<2)|0);
      switch ($133|0) {
      case 15:  {
       $170 = HEAP32[$79>>2]|0;
       if ($80) {
        label = 41;
       } else {
        $171 = (($max) + ($lwin$2<<2)|0);
        HEAP32[$171>>2] = $cb$2;
        $172 = ($num$5>>>0)<(24);
        if ($172) {
         $173 = (23 - ($num$5))|0;
         $174 = $173 & -8;
         $175 = (($num$5) + 8)|0;
         $177 = $615;$mask$6180 = $mask$5;$num$6181 = $num$5;
         while(1) {
          $176 = ((($177)) + 1|0);
          HEAP32[$10>>2] = $176;
          $178 = HEAP8[$177>>0]|0;
          $179 = $178&255;
          $180 = (24 - ($num$6181))|0;
          $181 = $179 << $180;
          $182 = $181 | $mask$6180;
          $183 = (($num$6181) + 8)|0;
          $184 = ($183>>>0)<(24);
          if ($184) {
           $177 = $176;$mask$6180 = $182;$num$6181 = $183;
          } else {
           $$lcssa599 = $176;$$lcssa600 = $182;
           break;
          }
         }
         $185 = (($part2remain$4) + -8)|0;
         $186 = (($185) - ($174))|0;
         $187 = (($175) + ($174))|0;
         $616 = $$lcssa599;$mask$6$lcssa = $$lcssa600;$num$6$lcssa = $187;$part2remain$5$lcssa = $186;
        } else {
         $616 = $615;$mask$6$lcssa = $mask$5;$num$6$lcssa = $num$5;$part2remain$5$lcssa = $part2remain$4;
        }
        $188 = (32 - ($170))|0;
        $189 = $mask$6$lcssa >>> $188;
        $190 = (($189) + 15)|0;
        $$neg58 = (($num$6$lcssa) + -1)|0;
        $191 = (($$neg58) - ($170))|0;
        $192 = $mask$6$lcssa << $170;
        $193 = ($192|0)<(0);
        $194 = (684712 + ($190<<2)|0);
        $195 = +HEAPF32[$194>>2];
        $196 = -$195;
        $$pn22 = $193 ? $196 : $195;
        $storemerge21 = $v$2 * $$pn22;
        HEAPF32[$169>>2] = $storemerge21;
        $197 = $192 << 1;
        $617 = $616;$mask$7 = $197;$num$7 = $191;$part2remain$6 = $part2remain$5$lcssa;
       }
       break;
      }
      case 0:  {
       HEAPF32[$169>>2] = 0.0;
       $617 = $615;$mask$7 = $mask$5;$num$7 = $num$5;$part2remain$6 = $part2remain$4;
       break;
      }
      default: {
       label = 41;
      }
      }
      if ((label|0) == 41) {
       label = 0;
       $198 = (($max) + ($lwin$2<<2)|0);
       HEAP32[$198>>2] = $cb$2;
       $199 = ($mask$5|0)<(0);
       $200 = (684712 + ($133<<2)|0);
       $201 = +HEAPF32[$200>>2];
       $202 = -$201;
       $$pn20 = $199 ? $202 : $201;
       $storemerge19 = $v$2 * $$pn20;
       HEAPF32[$169>>2] = $storemerge19;
       $203 = (($num$5) + -1)|0;
       $204 = $mask$5 << 1;
       $617 = $615;$mask$7 = $204;$num$7 = $203;$part2remain$6 = $part2remain$4;
      }
      $205 = (($169) + ($step$2<<2)|0);
      $206 = (($lp$0197) + -1)|0;
      $207 = (($mc$2) + -1)|0;
      $208 = ($206|0)==(0);
      if ($208) {
       $$13$lcssa = $$24;$611 = $617;$cb$1$lcssa = $cb$2;$lwin$1$lcssa = $lwin$2;$m$2$lcssa = $m$3;$mask$1$lcssa = $mask$7;$mc$1$lcssa = $207;$num$1$lcssa = $num$7;$part2remain$1$lcssa = $part2remain$6;$step$1$lcssa = $step$2;$v$1$lcssa = $v$2;$xrpnt$1$lcssa = $205;
       break L16;
      } else {
       $$13200 = $$24;$612 = $617;$cb$1193 = $cb$2;$lp$0197 = $206;$lwin$1192 = $lwin$2;$m$2195 = $m$3;$mask$1189 = $mask$7;$mc$1196 = $207;$num$1190 = $num$7;$part2remain$1198 = $part2remain$6;$step$1191 = $step$2;$v$1194 = $v$2;$xrpnt$1199 = $205;
      }
     }
    }
   } while(0);
   $209 = (($i$0217) + 1)|0;
   $exitcond = ($209|0)==(2);
   if ($exitcond) {
    $$13$lcssa$lcssa = $$13$lcssa;$$lcssa604 = $611;$cb$1$lcssa$lcssa = $cb$1$lcssa;$lwin$1$lcssa$lcssa = $lwin$1$lcssa;$m$2$lcssa$lcssa = $m$2$lcssa;$mask$1$lcssa$lcssa = $mask$1$lcssa;$mc$1$lcssa$lcssa = $mc$1$lcssa;$num$1$lcssa$lcssa = $num$1$lcssa;$part2remain$1$lcssa$lcssa = $part2remain$1$lcssa;$step$1$lcssa$lcssa = $step$1$lcssa;$v$1$lcssa$lcssa = $v$1$lcssa;$xrpnt$1$lcssa$lcssa = $xrpnt$1$lcssa;
    break;
   }
   $$phi$trans$insert405 = (($l) + ($209<<2)|0);
   $$pre406 = HEAP32[$$phi$trans$insert405>>2]|0;
   $$02226 = $$13$lcssa;$608 = $611;$76 = $$pre406;$cb$0220 = $cb$1$lcssa;$i$0217 = $209;$lwin$0219 = $lwin$1$lcssa;$m$1222 = $m$2$lcssa;$mask$0215 = $mask$1$lcssa;$mc$0223 = $mc$1$lcssa;$num$0216 = $num$1$lcssa;$part2remain$0224 = $part2remain$1$lcssa;$step$0218 = $step$1$lcssa;$v$0221 = $v$1$lcssa;$xrpnt$0225 = $xrpnt$1$lcssa;
  }
  $68 = ($42|0)!=(0);
  $69 = (($part2remain$1$lcssa$lcssa) + ($num$1$lcssa$lcssa))|0;
  $70 = ($69|0)>(0);
  $or$cond136 = $68 & $70;
  L65: do {
   if ($or$cond136) {
    $71 = ((($xr)) + 2324|0);
    $72 = ((($gr_info)) + 72|0);
    $$35148 = $$13$lcssa$lcssa;$610 = $$lcssa604;$cb$3141 = $cb$1$lcssa$lcssa;$l3$0146 = $42;$lwin$3140 = $lwin$1$lcssa$lcssa;$m$4143 = $m$2$lcssa$lcssa;$mask$8137 = $mask$1$lcssa$lcssa;$mc$3144 = $mc$1$lcssa$lcssa;$num$8138 = $num$1$lcssa$lcssa;$part2remain$7145 = $part2remain$1$lcssa$lcssa;$step$3139 = $step$1$lcssa$lcssa;$v$3142 = $v$1$lcssa$lcssa;$xrpnt$3147 = $xrpnt$1$lcssa$lcssa;
    while(1) {
     $210 = ($xrpnt$3147>>>0)<($71>>>0);
     if (!($210)) {
      $xrpnt$3147$lcssa = $xrpnt$3147;
      break;
     }
     $217 = HEAP32[$72>>2]|0;
     $218 = (((36800 + ($217<<3)|0)) + 4|0);
     $219 = HEAP32[$218>>2]|0;
     $220 = ($num$8138>>>0)<(24);
     if ($220) {
      $221 = (23 - ($num$8138))|0;
      $222 = $221 & -8;
      $223 = (($num$8138) + 8)|0;
      $231 = $610;$mask$9110 = $mask$8137;$num$9111 = $num$8138;
      while(1) {
       $230 = ((($231)) + 1|0);
       HEAP32[$10>>2] = $230;
       $232 = HEAP8[$231>>0]|0;
       $233 = $232&255;
       $234 = (24 - ($num$9111))|0;
       $235 = $233 << $234;
       $236 = $235 | $mask$9110;
       $237 = (($num$9111) + 8)|0;
       $238 = ($237>>>0)<(24);
       if ($238) {
        $231 = $230;$mask$9110 = $236;$num$9111 = $237;
       } else {
        $$lcssa576 = $230;$$lcssa577 = $236;
        break;
       }
      }
      $224 = (($part2remain$7145) + -8)|0;
      $225 = (($224) - ($222))|0;
      $226 = (($223) + ($222))|0;
      $618 = $$lcssa576;$mask$9$lcssa = $$lcssa577;$num$9$lcssa = $226;$part2remain$8$lcssa = $225;
     } else {
      $618 = $610;$mask$9$lcssa = $mask$8137;$num$9$lcssa = $num$8138;$part2remain$8$lcssa = $part2remain$7145;
     }
     $227 = HEAP16[$219>>1]|0;
     $228 = $227 << 16 >> 16;
     $229 = ($227<<16>>16)<(0);
     if ($229) {
      $242 = $228;$mask$10118 = $mask$9$lcssa;$num$10119 = $num$9$lcssa;$val2$1$pn = $219;
      while(1) {
       $239 = ((($val2$1$pn)) + 2|0);
       $240 = ($mask$10118|0)<(0);
       $241 = (0 - ($242))|0;
       $243 = (($239) + ($241<<1)|0);
       $val2$1 = $240 ? $243 : $239;
       $244 = (($num$10119) + -1)|0;
       $245 = $mask$10118 << 1;
       $246 = HEAP16[$val2$1>>1]|0;
       $247 = $246 << 16 >> 16;
       $248 = ($246<<16>>16)<(0);
       if ($248) {
        $242 = $247;$mask$10118 = $245;$num$10119 = $244;$val2$1$pn = $val2$1;
       } else {
        $$lcssa = $247;$mask$10$lcssa = $245;$num$10$lcssa = $244;
        break;
       }
      }
     } else {
      $$lcssa = $228;$mask$10$lcssa = $mask$9$lcssa;$num$10$lcssa = $num$9$lcssa;
     }
     $249 = (($num$10$lcssa) + ($part2remain$8$lcssa))|0;
     $250 = ($249|0)<(1);
     if ($250) {
      $$lcssa589 = $618;$lwin$3140$lcssa587 = $lwin$3140;$m$4143$lcssa585 = $m$4143;$mc$3144$lcssa583 = $mc$3144;$part2remain$8$lcssa$lcssa = $part2remain$8$lcssa;$xrpnt$3147$lcssa581 = $xrpnt$3147;
      label = 67;
      break;
     } else {
      $$46135 = $$35148;$cb$4130 = $cb$3141;$i$1127 = 0;$lwin$4129 = $lwin$3140;$m$5132 = $m$4143;$mask$11125 = $mask$10$lcssa;$mc$4133 = $mc$3144;$num$11126 = $num$10$lcssa;$step$4128 = $step$3139;$v$4131 = $v$3142;$xrpnt$4134 = $xrpnt$3147;
     }
     while(1) {
      $251 = $i$1127 & 1;
      $252 = ($251|0)==(0);
      if ($252) {
       $253 = ($mc$4133|0)==(0);
       do {
        if ($253) {
         $254 = ((($m$5132)) + 4|0);
         $255 = HEAP32[$m$5132>>2]|0;
         $256 = ((($m$5132)) + 8|0);
         $257 = HEAP32[$254>>2]|0;
         $258 = (($xr) + ($257<<2)|0);
         $259 = ((($m$5132)) + 12|0);
         $260 = HEAP32[$256>>2]|0;
         $261 = ((($m$5132)) + 16|0);
         $262 = HEAP32[$259>>2]|0;
         $263 = ($260|0)==(3);
         $264 = ((($$46135)) + 4|0);
         $265 = HEAP32[$$46135>>2]|0;
         $266 = $265 << $2;
         if ($263) {
          $267 = HEAP32[$67>>2]|0;
          $268 = (($267) + ($266<<2)|0);
          $269 = +HEAPF32[$268>>2];
          $$5 = $264;$cb$5 = $262;$lwin$5 = 3;$m$6 = $261;$mc$5 = $255;$step$5 = 1;$v$5 = $269;$xrpnt$5 = $258;
          break;
         } else {
          $270 = (((($gr_info)) + 76|0) + ($260<<2)|0);
          $271 = HEAP32[$270>>2]|0;
          $272 = (($271) + ($266<<2)|0);
          $273 = +HEAPF32[$272>>2];
          $$5 = $264;$cb$5 = $262;$lwin$5 = $260;$m$6 = $261;$mc$5 = $255;$step$5 = 3;$v$5 = $273;$xrpnt$5 = $258;
          break;
         }
        } else {
         $$5 = $$46135;$cb$5 = $cb$4130;$lwin$5 = $lwin$4129;$m$6 = $m$5132;$mc$5 = $mc$4133;$step$5 = $step$4128;$v$5 = $v$4131;$xrpnt$5 = $xrpnt$4134;
        }
       } while(0);
       $274 = (($mc$5) + -1)|0;
       $$6 = $$5;$cb$6 = $cb$5;$lwin$6 = $lwin$5;$m$7 = $m$6;$mc$6 = $274;$step$6 = $step$5;$v$6 = $v$5;$xrpnt$6 = $xrpnt$5;
      } else {
       $$6 = $$46135;$cb$6 = $cb$4130;$lwin$6 = $lwin$4129;$m$7 = $m$5132;$mc$6 = $mc$4133;$step$6 = $step$4128;$v$6 = $v$4131;$xrpnt$6 = $xrpnt$4134;
      }
      $275 = 8 >>> $i$1127;
      $276 = $275 & $$lcssa;
      $277 = ($276|0)==(0);
      if ($277) {
       HEAPF32[$xrpnt$6>>2] = 0.0;
       $mask$12 = $mask$11125;$num$12 = $num$11126;
      } else {
       $278 = (($max) + ($lwin$6<<2)|0);
       HEAP32[$278>>2] = $cb$6;
       $279 = (($num$11126) + ($part2remain$8$lcssa))|0;
       $280 = ($279|0)<(1);
       if ($280) {
        $$6$lcssa = $$6;$cb$6$lcssa = $cb$6;$lwin$6$lcssa = $lwin$6;$m$7$lcssa = $m$7;$mask$11$lcssa = $mask$11125;$mc$6$lcssa = $mc$6;$num$11$lcssa = $num$11126;$step$6$lcssa = $step$6;$v$6$lcssa = $v$6;$xrpnt$8$ph = $xrpnt$6;
        break;
       }
       $281 = ($mask$11125|0)<(0);
       $282 = -$v$6;
       $storemerge18 = $281 ? $282 : $v$6;
       HEAPF32[$xrpnt$6>>2] = $storemerge18;
       $283 = (($num$11126) + -1)|0;
       $284 = $mask$11125 << 1;
       $mask$12 = $284;$num$12 = $283;
      }
      $285 = (($xrpnt$6) + ($step$6<<2)|0);
      $286 = (($i$1127) + 1)|0;
      $287 = ($286|0)<(4);
      if ($287) {
       $$46135 = $$6;$cb$4130 = $cb$6;$i$1127 = $286;$lwin$4129 = $lwin$6;$m$5132 = $m$7;$mask$11125 = $mask$12;$mc$4133 = $mc$6;$num$11126 = $num$12;$step$4128 = $step$6;$v$4131 = $v$6;$xrpnt$4134 = $285;
      } else {
       $$6$lcssa = $$6;$cb$6$lcssa = $cb$6;$lwin$6$lcssa = $lwin$6;$m$7$lcssa = $m$7;$mask$11$lcssa = $mask$12;$mc$6$lcssa = $mc$6;$num$11$lcssa = $num$12;$step$6$lcssa = $step$6;$v$6$lcssa = $v$6;$xrpnt$8$ph = $285;
       break;
      }
     }
     $289 = (($l3$0146) + -1)|0;
     $290 = ($289|0)!=(0);
     $291 = (($part2remain$8$lcssa) + ($num$11$lcssa))|0;
     $292 = ($291|0)>(0);
     $or$cond = $290 & $292;
     if ($or$cond) {
      $$35148 = $$6$lcssa;$610 = $618;$cb$3141 = $cb$6$lcssa;$l3$0146 = $289;$lwin$3140 = $lwin$6$lcssa;$m$4143 = $m$7$lcssa;$mask$8137 = $mask$11$lcssa;$mc$3144 = $mc$6$lcssa;$num$8138 = $num$11$lcssa;$part2remain$7145 = $part2remain$8$lcssa;$step$3139 = $step$6$lcssa;$v$3142 = $v$6$lcssa;$xrpnt$3147 = $xrpnt$8$ph;
     } else {
      $609 = $618;$lwin$378 = $lwin$6$lcssa;$m$481 = $m$7$lcssa;$mc$384 = $mc$6$lcssa;$num$14 = $num$11$lcssa;$part2remain$10 = $part2remain$8$lcssa;$xrpnt$389 = $xrpnt$8$ph;
      break L65;
     }
    }
    if ((label|0) == 67) {
     $288 = (0 - ($part2remain$8$lcssa$lcssa))|0;
     $609 = $$lcssa589;$lwin$378 = $lwin$3140$lcssa587;$m$481 = $m$4143$lcssa585;$mc$384 = $mc$3144$lcssa583;$num$14 = $288;$part2remain$10 = $part2remain$8$lcssa$lcssa;$xrpnt$389 = $xrpnt$3147$lcssa581;
     break;
    }
    $211 = ((($fr)) + 28068|0);
    $212 = HEAP32[$211>>2]|0;
    $213 = $212 & 32;
    $214 = ($213|0)==(0);
    if ($214) {
     $215 = ((($xr)) + 2304|0);
     $216 = HEAP32[37544]|0;
     HEAP32[$vararg_buffer>>2] = 884;
     $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
     HEAP32[$vararg_ptr1>>2] = $xrpnt$3147$lcssa;
     $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
     HEAP32[$vararg_ptr2>>2] = $215;
     (_fprintf($216,511789,$vararg_buffer)|0);
    }
    $$4 = 2;
    STACKTOP = sp;return ($$4|0);
   } else {
    $609 = $$lcssa604;$lwin$378 = $lwin$1$lcssa$lcssa;$m$481 = $m$2$lcssa$lcssa;$mc$384 = $mc$1$lcssa$lcssa;$num$14 = $num$1$lcssa$lcssa;$part2remain$10 = $part2remain$1$lcssa$lcssa;$xrpnt$389 = $xrpnt$1$lcssa$lcssa;
   }
  } while(0);
  $293 = ($lwin$378|0)<(3);
  L105: do {
   if ($293) {
    $m$11$ph = $m$481;$mc$10$ph = $mc$384;$xrpnt$10$ph = $xrpnt$389;
    while(1) {
     $294 = ($mc$10$ph|0)>(0);
     if ($294) {
      $mc$10105 = $mc$10$ph;$xrpnt$10106 = $xrpnt$10$ph;
      while(1) {
       HEAPF32[$xrpnt$10106>>2] = 0.0;
       $295 = ((($xrpnt$10106)) + 12|0);
       HEAPF32[$295>>2] = 0.0;
       $296 = ((($xrpnt$10106)) + 24|0);
       $297 = (($mc$10105) + -1)|0;
       $298 = ($mc$10105|0)>(1);
       if ($298) {
        $mc$10105 = $297;$xrpnt$10106 = $296;
       } else {
        $xrpnt$10$lcssa = $296;
        break;
       }
      }
     } else {
      $xrpnt$10$lcssa = $xrpnt$10$ph;
     }
     $299 = ($m$11$ph>>>0)<($me$0>>>0);
     if (!($299)) {
      $xrpnt$11 = $xrpnt$10$lcssa;
      break L105;
     }
     $300 = ((($m$11$ph)) + 4|0);
     $301 = ((($m$11$ph)) + 8|0);
     $302 = HEAP32[$300>>2]|0;
     $303 = (($xr) + ($302<<2)|0);
     $304 = HEAP32[$301>>2]|0;
     $305 = ($304|0)==(0);
     if ($305) {
      $xrpnt$11 = $303;
      break L105;
     }
     $306 = HEAP32[$m$11$ph>>2]|0;
     $307 = ((($m$11$ph)) + 16|0);
     $m$11$ph = $307;$mc$10$ph = $306;$xrpnt$10$ph = $303;
    }
   } else {
    $xrpnt$11 = $xrpnt$389;
   }
  } while(0);
  $308 = HEAP32[$max>>2]|0;
  $309 = (($308) + 1)|0;
  $310 = ((($gr_info)) + 36|0);
  HEAP32[$310>>2] = $309;
  $311 = HEAP32[$66>>2]|0;
  $312 = (($311) + 1)|0;
  $313 = ((($gr_info)) + 40|0);
  HEAP32[$313>>2] = $312;
  $314 = HEAP32[$61>>2]|0;
  $315 = (($314) + 1)|0;
  $316 = ((($gr_info)) + 44|0);
  HEAP32[$316>>2] = $315;
  $317 = HEAP32[$60>>2]|0;
  $318 = (($317) + 1)|0;
  $319 = ((($gr_info)) + 48|0);
  HEAP32[$319>>2] = $318;
  $320 = ($308|0)>($311|0);
  $321 = $320 ? $308 : $311;
  $322 = ($321|0)>($314|0);
  $323 = $322 ? $321 : $314;
  $324 = (($323) + 1)|0;
  $325 = ($324|0)==(0);
  $326 = ((((($fr)) + 10424|0) + (($sfreq*56)|0)|0) + ($324<<2)|0);
  $327 = ((((($fr)) + 9596|0) + (($sfreq*92)|0)|0) + ($318<<2)|0);
  $$in = $325 ? $327 : $326;
  $328 = HEAP32[$$in>>2]|0;
  $329 = ((($gr_info)) + 52|0);
  HEAP32[$329>>2] = $328;
  $$pre407 = HEAP32[$6>>2]|0;
  $534 = $$pre407;$537 = $609;$num$31 = $num$14;$part2remain$22 = $part2remain$10;$xrpnt$27 = $xrpnt$11;
 } else {
  $330 = ((($gr_info)) + 64|0);
  $331 = HEAP32[$330>>2]|0;
  $332 = (511847 + (($331*22)|0)|0);
  $333 = (((684496 + (($sfreq*12)|0)|0)) + 8|0);
  $334 = HEAP32[$333>>2]|0;
  $335 = ((($gr_info)) + 88|0);
  $$9339 = $scf;$343 = $607;$619 = $30;$cb6$0334 = 0;$i4$0332 = 0;$m7$0335 = $334;$mask$14328 = $33;$max5$0333 = -1;$mc9$0337 = 0;$num$16329 = $9;$part2remain$12330 = $34;$pretab$0331 = $332;$v8$0336 = 0.0;$xrpnt$13338 = $xr;
  while(1) {
   $340 = (((($gr_info)) + 24|0) + ($i4$0332<<2)|0);
   $341 = HEAP32[$340>>2]|0;
   $342 = ($343|0)==(0);
   if ($342) {
    $$10$lcssa = $$9339;$622 = $619;$cb6$1$lcssa = $cb6$0334;$m7$1$lcssa = $m7$0335;$mask$15$lcssa = $mask$14328;$max5$1$lcssa = $max5$0333;$mc9$1$lcssa = $mc9$0337;$num$17$lcssa = $num$16329;$part2remain$13$lcssa = $part2remain$12330;$pretab$1$lcssa = $pretab$0331;$v8$1$lcssa = $v8$0336;$xrpnt$14$lcssa = $xrpnt$13338;
   } else {
    $344 = (((36544 + ($341<<3)|0)) + 4|0);
    $345 = HEAP32[$344>>2]|0;
    $346 = (36544 + ($341<<3)|0);
    $347 = $343 << 1;
    $348 = ($341>>>0)<(16);
    $349 = ($341>>>0)<(16);
    $$10313 = $$9339;$623 = $619;$cb6$1307 = $cb6$0334;$lp10$0311 = $343;$m7$1308 = $m7$0335;$mask$15302 = $mask$14328;$max5$1306 = $max5$0333;$mc9$1310 = $mc9$0337;$num$17303 = $num$16329;$part2remain$13304 = $part2remain$12330;$pretab$1305 = $pretab$0331;$v8$1309 = $v8$0336;$xrpnt$14312 = $xrpnt$13338;
    while(1) {
     $350 = ($mc9$1310|0)==(0);
     if ($350) {
      $351 = ((($m7$1308)) + 4|0);
      $352 = HEAP32[$m7$1308>>2]|0;
      $353 = ((($m7$1308)) + 8|0);
      $354 = HEAP32[$351>>2]|0;
      $355 = ((($$10313)) + 4|0);
      $356 = HEAP32[$$10313>>2]|0;
      $357 = ((($pretab$1305)) + 1|0);
      $358 = HEAP8[$pretab$1305>>0]|0;
      $359 = $358&255;
      $360 = (($359) + ($356))|0;
      $361 = $360 << $2;
      $362 = HEAP32[$335>>2]|0;
      $363 = (($362) + ($361<<2)|0);
      $364 = +HEAPF32[$363>>2];
      $$11 = $355;$cb6$2 = $354;$m7$2 = $353;$mc9$2 = $352;$pretab$2 = $357;$v8$2 = $364;
     } else {
      $$11 = $$10313;$cb6$2 = $cb6$1307;$m7$2 = $m7$1308;$mc9$2 = $mc9$1310;$pretab$2 = $pretab$1305;$v8$2 = $v8$1309;
     }
     $365 = ($num$17303>>>0)<(24);
     if ($365) {
      $366 = (23 - ($num$17303))|0;
      $367 = $366 & -8;
      $368 = (($num$17303) + 8)|0;
      $376 = $623;$mask$16271 = $mask$15302;$num$18272 = $num$17303;
      while(1) {
       $375 = ((($376)) + 1|0);
       HEAP32[$10>>2] = $375;
       $377 = HEAP8[$376>>0]|0;
       $378 = $377&255;
       $379 = (24 - ($num$18272))|0;
       $380 = $378 << $379;
       $381 = $380 | $mask$16271;
       $382 = (($num$18272) + 8)|0;
       $383 = ($382>>>0)<(24);
       if ($383) {
        $376 = $375;$mask$16271 = $381;$num$18272 = $382;
       } else {
        $$lcssa615 = $375;$$lcssa616 = $381;
        break;
       }
      }
      $369 = (($part2remain$13304) + -8)|0;
      $370 = (($369) - ($367))|0;
      $371 = (($368) + ($367))|0;
      $624 = $$lcssa615;$mask$16$lcssa = $$lcssa616;$num$18$lcssa = $371;$part2remain$14$lcssa = $370;
     } else {
      $624 = $623;$mask$16$lcssa = $mask$15302;$num$18$lcssa = $num$17303;$part2remain$14$lcssa = $part2remain$13304;
     }
     $372 = HEAP16[$345>>1]|0;
     $373 = $372 << 16 >> 16;
     $374 = ($372<<16>>16)<(0);
     if ($374) {
      $387 = $373;$mask$17279 = $mask$16$lcssa;$num$19280 = $num$18$lcssa;$val14$1$pn = $345;
      while(1) {
       $384 = ((($val14$1$pn)) + 2|0);
       $385 = ($mask$17279|0)<(0);
       $386 = (0 - ($387))|0;
       $388 = (($384) + ($386<<1)|0);
       $val14$1 = $385 ? $388 : $384;
       $389 = (($num$19280) + -1)|0;
       $390 = $mask$17279 << 1;
       $391 = HEAP16[$val14$1>>1]|0;
       $392 = $391 << 16 >> 16;
       $393 = ($391<<16>>16)<(0);
       if ($393) {
        $387 = $392;$mask$17279 = $390;$num$19280 = $389;$val14$1$pn = $val14$1;
       } else {
        $$lcssa100 = $392;$mask$17$lcssa = $390;$num$19$lcssa = $389;
        break;
       }
      }
     } else {
      $$lcssa100 = $373;$mask$17$lcssa = $mask$16$lcssa;$num$19$lcssa = $num$18$lcssa;
     }
     $394 = $$lcssa100 >> 4;
     $395 = $$lcssa100 & 15;
     switch ($394|0) {
     case 15:  {
      $396 = HEAP32[$346>>2]|0;
      if ($349) {
       label = 97;
      } else {
       $397 = ($num$19$lcssa>>>0)<(24);
       if ($397) {
        $398 = (23 - ($num$19$lcssa))|0;
        $399 = $398 & -8;
        $400 = (($num$19$lcssa) + 8)|0;
        $402 = $624;$mask$18286 = $mask$17$lcssa;$num$20287 = $num$19$lcssa;
        while(1) {
         $401 = ((($402)) + 1|0);
         HEAP32[$10>>2] = $401;
         $403 = HEAP8[$402>>0]|0;
         $404 = $403&255;
         $405 = (24 - ($num$20287))|0;
         $406 = $404 << $405;
         $407 = $406 | $mask$18286;
         $408 = (($num$20287) + 8)|0;
         $409 = ($408>>>0)<(24);
         if ($409) {
          $402 = $401;$mask$18286 = $407;$num$20287 = $408;
         } else {
          $$lcssa620 = $401;$$lcssa621 = $407;
          break;
         }
        }
        $410 = (($part2remain$14$lcssa) + -8)|0;
        $411 = (($410) - ($399))|0;
        $412 = (($400) + ($399))|0;
        $625 = $$lcssa620;$mask$18$lcssa = $$lcssa621;$num$20$lcssa = $412;$part2remain$15$lcssa = $411;
       } else {
        $625 = $624;$mask$18$lcssa = $mask$17$lcssa;$num$20$lcssa = $num$19$lcssa;$part2remain$15$lcssa = $part2remain$14$lcssa;
       }
       $413 = (32 - ($396))|0;
       $414 = $mask$18$lcssa >>> $413;
       $415 = (($414) + 15)|0;
       $$neg60 = (($num$20$lcssa) + -1)|0;
       $416 = (($$neg60) - ($396))|0;
       $417 = $mask$18$lcssa << $396;
       $418 = ($417|0)<(0);
       $419 = (684712 + ($415<<2)|0);
       $420 = +HEAPF32[$419>>2];
       $421 = -$420;
       $$pn17 = $418 ? $421 : $420;
       $storemerge16 = $v8$2 * $$pn17;
       HEAPF32[$xrpnt$14312>>2] = $storemerge16;
       $422 = $417 << 1;
       $626 = $625;$mask$19 = $422;$max5$2 = $cb6$2;$num$21 = $416;$part2remain$16 = $part2remain$15$lcssa;
      }
      break;
     }
     case 0:  {
      HEAPF32[$xrpnt$14312>>2] = 0.0;
      $626 = $624;$mask$19 = $mask$17$lcssa;$max5$2 = $max5$1306;$num$21 = $num$19$lcssa;$part2remain$16 = $part2remain$14$lcssa;
      break;
     }
     default: {
      label = 97;
     }
     }
     if ((label|0) == 97) {
      label = 0;
      $423 = ($mask$17$lcssa|0)<(0);
      $424 = (684712 + ($394<<2)|0);
      $425 = +HEAPF32[$424>>2];
      $426 = -$425;
      $$pn15 = $423 ? $426 : $425;
      $storemerge14 = $v8$2 * $$pn15;
      HEAPF32[$xrpnt$14312>>2] = $storemerge14;
      $427 = (($num$19$lcssa) + -1)|0;
      $428 = $mask$17$lcssa << 1;
      $626 = $624;$mask$19 = $428;$max5$2 = $cb6$2;$num$21 = $427;$part2remain$16 = $part2remain$14$lcssa;
     }
     $xrpnt$17 = ((($xrpnt$14312)) + 4|0);
     switch ($395|0) {
     case 15:  {
      $429 = HEAP32[$346>>2]|0;
      if ($348) {
       label = 106;
      } else {
       $430 = ($num$21>>>0)<(24);
       if ($430) {
        $431 = (23 - ($num$21))|0;
        $432 = $431 & -8;
        $433 = (($num$21) + 8)|0;
        $435 = $626;$mask$20294 = $mask$19;$num$22295 = $num$21;
        while(1) {
         $434 = ((($435)) + 1|0);
         HEAP32[$10>>2] = $434;
         $436 = HEAP8[$435>>0]|0;
         $437 = $436&255;
         $438 = (24 - ($num$22295))|0;
         $439 = $437 << $438;
         $440 = $439 | $mask$20294;
         $441 = (($num$22295) + 8)|0;
         $442 = ($441>>>0)<(24);
         if ($442) {
          $435 = $434;$mask$20294 = $440;$num$22295 = $441;
         } else {
          $$lcssa622 = $434;$$lcssa623 = $440;
          break;
         }
        }
        $443 = (($part2remain$16) + -8)|0;
        $444 = (($443) - ($432))|0;
        $445 = (($433) + ($432))|0;
        $627 = $$lcssa622;$mask$20$lcssa = $$lcssa623;$num$22$lcssa = $445;$part2remain$17$lcssa = $444;
       } else {
        $627 = $626;$mask$20$lcssa = $mask$19;$num$22$lcssa = $num$21;$part2remain$17$lcssa = $part2remain$16;
       }
       $446 = (32 - ($429))|0;
       $447 = $mask$20$lcssa >>> $446;
       $448 = (($447) + 15)|0;
       $$neg62 = (($num$22$lcssa) + -1)|0;
       $449 = (($$neg62) - ($429))|0;
       $450 = $mask$20$lcssa << $429;
       $451 = ($450|0)<(0);
       $452 = (684712 + ($448<<2)|0);
       $453 = +HEAPF32[$452>>2];
       $454 = -$453;
       $$pn13 = $451 ? $454 : $453;
       $storemerge12 = $v8$2 * $$pn13;
       HEAPF32[$xrpnt$17>>2] = $storemerge12;
       $455 = $450 << 1;
       $628 = $627;$mask$21 = $455;$max5$3 = $cb6$2;$num$23 = $449;$part2remain$18 = $part2remain$17$lcssa;
      }
      break;
     }
     case 0:  {
      HEAPF32[$xrpnt$17>>2] = 0.0;
      $628 = $626;$mask$21 = $mask$19;$max5$3 = $max5$2;$num$23 = $num$21;$part2remain$18 = $part2remain$16;
      break;
     }
     default: {
      label = 106;
     }
     }
     if ((label|0) == 106) {
      label = 0;
      $456 = ($mask$19|0)<(0);
      $457 = (684712 + ($395<<2)|0);
      $458 = +HEAPF32[$457>>2];
      $459 = -$458;
      $$pn = $456 ? $459 : $458;
      $storemerge11 = $v8$2 * $$pn;
      HEAPF32[$xrpnt$17>>2] = $storemerge11;
      $460 = (($num$21) + -1)|0;
      $461 = $mask$19 << 1;
      $628 = $626;$mask$21 = $461;$max5$3 = $cb6$2;$num$23 = $460;$part2remain$18 = $part2remain$16;
     }
     $xrpnt$20 = ((($xrpnt$14312)) + 8|0);
     $462 = (($lp10$0311) + -1)|0;
     $463 = (($mc9$2) + -1)|0;
     $464 = ($462|0)==(0);
     if ($464) {
      $$11$lcssa = $$11;$$lcssa624 = $628;$$lcssa625 = $463;$cb6$2$lcssa = $cb6$2;$m7$2$lcssa = $m7$2;$mask$21$lcssa = $mask$21;$max5$3$lcssa = $max5$3;$num$23$lcssa = $num$23;$part2remain$18$lcssa = $part2remain$18;$pretab$2$lcssa = $pretab$2;$v8$2$lcssa = $v8$2;
      break;
     } else {
      $$10313 = $$11;$623 = $628;$cb6$1307 = $cb6$2;$lp10$0311 = $462;$m7$1308 = $m7$2;$mask$15302 = $mask$21;$max5$1306 = $max5$3;$mc9$1310 = $463;$num$17303 = $num$23;$part2remain$13304 = $part2remain$18;$pretab$1305 = $pretab$2;$v8$1309 = $v8$2;$xrpnt$14312 = $xrpnt$20;
     }
    }
    $scevgep = (($xrpnt$13338) + ($347<<2)|0);
    $$10$lcssa = $$11$lcssa;$622 = $$lcssa624;$cb6$1$lcssa = $cb6$2$lcssa;$m7$1$lcssa = $m7$2$lcssa;$mask$15$lcssa = $mask$21$lcssa;$max5$1$lcssa = $max5$3$lcssa;$mc9$1$lcssa = $$lcssa625;$num$17$lcssa = $num$23$lcssa;$part2remain$13$lcssa = $part2remain$18$lcssa;$pretab$1$lcssa = $pretab$2$lcssa;$v8$1$lcssa = $v8$2$lcssa;$xrpnt$14$lcssa = $scevgep;
   }
   $465 = (($i4$0332) + 1)|0;
   $exitcond394 = ($465|0)==(3);
   if ($exitcond394) {
    $$10$lcssa$lcssa = $$10$lcssa;$$lcssa626 = $622;$cb6$1$lcssa$lcssa = $cb6$1$lcssa;$m7$1$lcssa$lcssa = $m7$1$lcssa;$mask$15$lcssa$lcssa = $mask$15$lcssa;$max5$1$lcssa$lcssa = $max5$1$lcssa;$mc9$1$lcssa$lcssa = $mc9$1$lcssa;$num$17$lcssa$lcssa = $num$17$lcssa;$part2remain$13$lcssa$lcssa = $part2remain$13$lcssa;$pretab$1$lcssa$lcssa = $pretab$1$lcssa;$v8$1$lcssa$lcssa = $v8$1$lcssa;$xrpnt$14$lcssa$lcssa = $xrpnt$14$lcssa;
    break;
   }
   $$phi$trans$insert = (($l) + ($465<<2)|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0;
   $$9339 = $$10$lcssa;$343 = $$pre;$619 = $622;$cb6$0334 = $cb6$1$lcssa;$i4$0332 = $465;$m7$0335 = $m7$1$lcssa;$mask$14328 = $mask$15$lcssa;$max5$0333 = $max5$1$lcssa;$mc9$0337 = $mc9$1$lcssa;$num$16329 = $num$17$lcssa;$part2remain$12330 = $part2remain$13$lcssa;$pretab$0331 = $pretab$1$lcssa;$v8$0336 = $v8$1$lcssa;$xrpnt$13338 = $xrpnt$14$lcssa;
  }
  $336 = ($42|0)!=(0);
  $337 = (($part2remain$13$lcssa$lcssa) + ($num$17$lcssa$lcssa))|0;
  $338 = ($337|0)>(0);
  $or$cond29253 = $336 & $338;
  L165: do {
   if ($or$cond29253) {
    $339 = ((($gr_info)) + 72|0);
    $$12265 = $$10$lcssa$lcssa;$621 = $$lcssa626;$cb6$3260 = $cb6$1$lcssa$lcssa;$l3$1257 = $42;$m7$3261 = $m7$1$lcssa$lcssa;$mask$22254 = $mask$15$lcssa$lcssa;$max5$4259 = $max5$1$lcssa$lcssa;$mc9$3263 = $mc9$1$lcssa$lcssa;$num$24255 = $num$17$lcssa$lcssa;$part2remain$19256 = $part2remain$13$lcssa$lcssa;$pretab$3258 = $pretab$1$lcssa$lcssa;$v8$3262 = $v8$1$lcssa$lcssa;$xrpnt$21264 = $xrpnt$14$lcssa$lcssa;
    while(1) {
     $466 = HEAP32[$339>>2]|0;
     $467 = (((36800 + ($466<<3)|0)) + 4|0);
     $468 = HEAP32[$467>>2]|0;
     $469 = ($num$24255>>>0)<(24);
     if ($469) {
      $470 = (23 - ($num$24255))|0;
      $471 = $470 & -8;
      $472 = (($num$24255) + 8)|0;
      $480 = $621;$mask$23227 = $mask$22254;$num$25228 = $num$24255;
      while(1) {
       $479 = ((($480)) + 1|0);
       HEAP32[$10>>2] = $479;
       $481 = HEAP8[$480>>0]|0;
       $482 = $481&255;
       $483 = (24 - ($num$25228))|0;
       $484 = $482 << $483;
       $485 = $484 | $mask$23227;
       $486 = (($num$25228) + 8)|0;
       $487 = ($486>>>0)<(24);
       if ($487) {
        $480 = $479;$mask$23227 = $485;$num$25228 = $486;
       } else {
        $$lcssa605 = $479;$$lcssa606 = $485;
        break;
       }
      }
      $473 = (($part2remain$19256) + -8)|0;
      $474 = (($473) - ($471))|0;
      $475 = (($472) + ($471))|0;
      $629 = $$lcssa605;$mask$23$lcssa = $$lcssa606;$num$25$lcssa = $475;$part2remain$20$lcssa = $474;
     } else {
      $629 = $621;$mask$23$lcssa = $mask$22254;$num$25$lcssa = $num$24255;$part2remain$20$lcssa = $part2remain$19256;
     }
     $476 = HEAP16[$468>>1]|0;
     $477 = $476 << 16 >> 16;
     $478 = ($476<<16>>16)<(0);
     if ($478) {
      $491 = $477;$mask$24235 = $mask$23$lcssa;$num$26236 = $num$25$lcssa;$val16$1$pn = $468;
      while(1) {
       $488 = ((($val16$1$pn)) + 2|0);
       $489 = ($mask$24235|0)<(0);
       $490 = (0 - ($491))|0;
       $492 = (($488) + ($490<<1)|0);
       $val16$1 = $489 ? $492 : $488;
       $493 = (($num$26236) + -1)|0;
       $494 = $mask$24235 << 1;
       $495 = HEAP16[$val16$1>>1]|0;
       $496 = $495 << 16 >> 16;
       $497 = ($495<<16>>16)<(0);
       if ($497) {
        $491 = $496;$mask$24235 = $494;$num$26236 = $493;$val16$1$pn = $val16$1;
       } else {
        $$lcssa93 = $496;$mask$24$lcssa = $494;$num$26$lcssa = $493;
        break;
       }
      }
     } else {
      $$lcssa93 = $477;$mask$24$lcssa = $mask$23$lcssa;$num$26$lcssa = $num$25$lcssa;
     }
     $498 = (($num$26$lcssa) + ($part2remain$20$lcssa))|0;
     $499 = ($498|0)<(1);
     if ($499) {
      $$lcssa612 = $629;$max5$4259$lcssa = $max5$4259;$part2remain$20$lcssa$lcssa = $part2remain$20$lcssa;$xrpnt$21264$lcssa = $xrpnt$21264;
      break;
     }
     $500 = ($mc9$3263|0)==(0);
     if ($500) {
      $501 = ((($m7$3261)) + 4|0);
      $502 = HEAP32[$m7$3261>>2]|0;
      $503 = ((($m7$3261)) + 8|0);
      $504 = HEAP32[$501>>2]|0;
      $505 = ((($$12265)) + 4|0);
      $506 = HEAP32[$$12265>>2]|0;
      $507 = ((($pretab$3258)) + 1|0);
      $508 = HEAP8[$pretab$3258>>0]|0;
      $509 = $508&255;
      $510 = (($509) + ($506))|0;
      $511 = $510 << $2;
      $512 = HEAP32[$335>>2]|0;
      $513 = (($512) + ($511<<2)|0);
      $514 = +HEAPF32[$513>>2];
      $$14 = $505;$cb6$5 = $504;$m7$5 = $503;$mc9$5 = $502;$pretab$5 = $507;$v8$5 = $514;
     } else {
      $$14 = $$12265;$cb6$5 = $cb6$3260;$m7$5 = $m7$3261;$mc9$5 = $mc9$3263;$pretab$5 = $pretab$3258;$v8$5 = $v8$3262;
     }
     $515 = (($mc9$5) + -1)|0;
     $516 = $$lcssa93 & 8;
     $517 = ($516|0)==(0);
     if ($517) {
      HEAPF32[$xrpnt$21264>>2] = 0.0;
      $mask$26 = $mask$24$lcssa;$max5$6 = $max5$4259;$num$28 = $num$26$lcssa;
     } else {
      $518 = ($mask$24$lcssa|0)<(0);
      $519 = -$v8$5;
      $storemerge = $518 ? $519 : $v8$5;
      HEAPF32[$xrpnt$21264>>2] = $storemerge;
      $520 = (($num$26$lcssa) + -1)|0;
      $521 = $mask$24$lcssa << 1;
      $mask$26 = $521;$max5$6 = $cb6$5;$num$28 = $520;
     }
     $xrpnt$24 = ((($xrpnt$21264)) + 4|0);
     $567 = $$lcssa93 & 4;
     $568 = ($567|0)==(0);
     if ($568) {
      HEAPF32[$xrpnt$24>>2] = 0.0;
      $mask$26$1 = $mask$26;$max5$6$1 = $max5$6;$num$28$1 = $num$28;
      label = 142;
     } else {
      $569 = (($num$28) + ($part2remain$20$lcssa))|0;
      $570 = ($569|0)<(1);
      if ($570) {
       $$15$lcssa = $$14;$cb6$6$lcssa = $cb6$5;$m7$6$lcssa = $m7$5;$mask$25$lcssa = $mask$26;$max5$8$ph = $cb6$5;$mc9$6$lcssa = $515;$num$27$lcssa = $num$28;$pretab$6$lcssa = $pretab$5;$v8$6$lcssa = $v8$5;$xrpnt$22$lcssa = $xrpnt$24;
      } else {
       $571 = ($mask$26|0)<(0);
       $572 = -$v8$5;
       $storemerge$1 = $571 ? $572 : $v8$5;
       HEAPF32[$xrpnt$24>>2] = $storemerge$1;
       $573 = (($num$28) + -1)|0;
       $574 = $mask$26 << 1;
       $mask$26$1 = $574;$max5$6$1 = $cb6$5;$num$28$1 = $573;
       label = 142;
      }
     }
     do {
      if ((label|0) == 142) {
       label = 0;
       $xrpnt$24$1 = ((($xrpnt$21264)) + 8|0);
       $575 = ($515|0)==(0);
       if ($575) {
        $576 = ((($m7$5)) + 4|0);
        $577 = HEAP32[$m7$5>>2]|0;
        $578 = ((($m7$5)) + 8|0);
        $579 = HEAP32[$576>>2]|0;
        $580 = ((($$14)) + 4|0);
        $581 = HEAP32[$$14>>2]|0;
        $582 = ((($pretab$5)) + 1|0);
        $583 = HEAP8[$pretab$5>>0]|0;
        $584 = $583&255;
        $585 = (($584) + ($581))|0;
        $586 = $585 << $2;
        $587 = HEAP32[$335>>2]|0;
        $588 = (($587) + ($586<<2)|0);
        $589 = +HEAPF32[$588>>2];
        $$14$2 = $580;$cb6$5$2 = $579;$m7$5$2 = $578;$mc9$5$2 = $577;$pretab$5$2 = $582;$v8$5$2 = $589;
       } else {
        $$14$2 = $$14;$cb6$5$2 = $cb6$5;$m7$5$2 = $m7$5;$mc9$5$2 = $515;$pretab$5$2 = $pretab$5;$v8$5$2 = $v8$5;
       }
       $590 = (($mc9$5$2) + -1)|0;
       $591 = $$lcssa93 & 2;
       $592 = ($591|0)==(0);
       if ($592) {
        HEAPF32[$xrpnt$24$1>>2] = 0.0;
        $mask$26$2 = $mask$26$1;$max5$6$2 = $max5$6$1;$num$28$2 = $num$28$1;
       } else {
        $593 = (($num$28$1) + ($part2remain$20$lcssa))|0;
        $594 = ($593|0)<(1);
        if ($594) {
         $$15$lcssa = $$14$2;$cb6$6$lcssa = $cb6$5$2;$m7$6$lcssa = $m7$5$2;$mask$25$lcssa = $mask$26$1;$max5$8$ph = $cb6$5$2;$mc9$6$lcssa = $590;$num$27$lcssa = $num$28$1;$pretab$6$lcssa = $pretab$5$2;$v8$6$lcssa = $v8$5$2;$xrpnt$22$lcssa = $xrpnt$24$1;
         break;
        }
        $595 = ($mask$26$1|0)<(0);
        $596 = -$v8$5$2;
        $storemerge$2 = $595 ? $596 : $v8$5$2;
        HEAPF32[$xrpnt$24$1>>2] = $storemerge$2;
        $597 = (($num$28$1) + -1)|0;
        $598 = $mask$26$1 << 1;
        $mask$26$2 = $598;$max5$6$2 = $cb6$5$2;$num$28$2 = $597;
       }
       $xrpnt$24$2 = ((($xrpnt$21264)) + 12|0);
       $599 = $$lcssa93 & 1;
       $600 = ($599|0)==(0);
       if ($600) {
        HEAPF32[$xrpnt$24$2>>2] = 0.0;
        $mask$26$3 = $mask$26$2;$max5$6$3 = $max5$6$2;$num$28$3 = $num$28$2;
       } else {
        $601 = (($num$28$2) + ($part2remain$20$lcssa))|0;
        $602 = ($601|0)<(1);
        if ($602) {
         $$15$lcssa = $$14$2;$cb6$6$lcssa = $cb6$5$2;$m7$6$lcssa = $m7$5$2;$mask$25$lcssa = $mask$26$2;$max5$8$ph = $cb6$5$2;$mc9$6$lcssa = $590;$num$27$lcssa = $num$28$2;$pretab$6$lcssa = $pretab$5$2;$v8$6$lcssa = $v8$5$2;$xrpnt$22$lcssa = $xrpnt$24$2;
         break;
        }
        $603 = ($mask$26$2|0)<(0);
        $604 = -$v8$5$2;
        $storemerge$3 = $603 ? $604 : $v8$5$2;
        HEAPF32[$xrpnt$24$2>>2] = $storemerge$3;
        $605 = (($num$28$2) + -1)|0;
        $606 = $mask$26$2 << 1;
        $mask$26$3 = $606;$max5$6$3 = $cb6$5$2;$num$28$3 = $605;
       }
       $xrpnt$24$3 = ((($xrpnt$21264)) + 16|0);
       $$15$lcssa = $$14$2;$cb6$6$lcssa = $cb6$5$2;$m7$6$lcssa = $m7$5$2;$mask$25$lcssa = $mask$26$3;$max5$8$ph = $max5$6$3;$mc9$6$lcssa = $590;$num$27$lcssa = $num$28$3;$pretab$6$lcssa = $pretab$5$2;$v8$6$lcssa = $v8$5$2;$xrpnt$22$lcssa = $xrpnt$24$3;
      }
     } while(0);
     $523 = (($l3$1257) + -1)|0;
     $524 = ($523|0)!=(0);
     $525 = (($part2remain$20$lcssa) + ($num$27$lcssa))|0;
     $526 = ($525|0)>(0);
     $or$cond29 = $524 & $526;
     if ($or$cond29) {
      $$12265 = $$15$lcssa;$621 = $629;$cb6$3260 = $cb6$6$lcssa;$l3$1257 = $523;$m7$3261 = $m7$6$lcssa;$mask$22254 = $mask$25$lcssa;$max5$4259 = $max5$8$ph;$mc9$3263 = $mc9$6$lcssa;$num$24255 = $num$27$lcssa;$part2remain$19256 = $part2remain$20$lcssa;$pretab$3258 = $pretab$6$lcssa;$v8$3262 = $v8$6$lcssa;$xrpnt$21264 = $xrpnt$22$lcssa;
     } else {
      $620 = $629;$max5$497 = $max5$8$ph;$num$30 = $num$27$lcssa;$part2remain$21 = $part2remain$20$lcssa;$xrpnt$2199 = $xrpnt$22$lcssa;
      break L165;
     }
    }
    $522 = (0 - ($part2remain$20$lcssa$lcssa))|0;
    $620 = $$lcssa612;$max5$497 = $max5$4259$lcssa;$num$30 = $522;$part2remain$21 = $part2remain$20$lcssa$lcssa;$xrpnt$2199 = $xrpnt$21264$lcssa;
   } else {
    $620 = $$lcssa626;$max5$497 = $max5$1$lcssa$lcssa;$num$30 = $num$17$lcssa$lcssa;$part2remain$21 = $part2remain$13$lcssa$lcssa;$xrpnt$2199 = $xrpnt$14$lcssa$lcssa;
   }
  } while(0);
  $527 = (($max5$497) + 1)|0;
  $528 = ((($gr_info)) + 48|0);
  HEAP32[$528>>2] = $527;
  $529 = ((((($fr)) + 9596|0) + (($sfreq*92)|0)|0) + ($527<<2)|0);
  $530 = HEAP32[$529>>2]|0;
  $531 = ((($gr_info)) + 52|0);
  HEAP32[$531>>2] = $530;
  $534 = 0;$537 = $620;$num$31 = $num$30;$part2remain$22 = $part2remain$21;$xrpnt$27 = $xrpnt$2199;
 }
 $532 = (($part2remain$22) + ($num$31))|0;
 $533 = (($534) - ($num$31))|0;
 $535 = $533 >> 3;
 $536 = (($537) + ($535)|0);
 HEAP32[$10>>2] = $536;
 $538 = $533 & 7;
 HEAP32[$6>>2] = $538;
 $539 = ((($xr)) + 2304|0);
 $540 = ($xrpnt$27>>>0)<($539>>>0);
 if ($540) {
  $xrpnt$27398 = $xrpnt$27;
  $scevgep396 = ((($xr)) + 2300|0);
  $541 = (3 - ($xrpnt$27398))|0;
  $uglygep = (($scevgep396) + ($541)|0);
  $uglygep399 = $uglygep;
  $542 = (($uglygep399) + 4)|0;
  $543 = $542 & -4;
  _memset(($xrpnt$27|0),0,($543|0))|0;
 }
 $544 = ($532|0)>(16);
 if ($544) {
  $545 = ((($fr)) + 19792|0);
  $546 = (($532) + -17)|0;
  $547 = $546 & -16;
  $$pre408 = HEAP32[$10>>2]|0;
  $$pre409 = HEAP32[$6>>2]|0;
  $549 = $$pre409;$552 = $$pre408;$part2remain$23101 = $532;
  while(1) {
   $548 = (($549) + 16)|0;
   $550 = $548 >> 3;
   $551 = (($552) + ($550)|0);
   $553 = $548 & 7;
   $554 = (($part2remain$23101) + -16)|0;
   $555 = ($554|0)>(16);
   if ($555) {
    $549 = $553;$552 = $551;$part2remain$23101 = $554;
   } else {
    $$lcssa573 = $551;$$lcssa574 = $553;
    break;
   }
  }
  HEAP32[$10>>2] = $$lcssa573;
  HEAP32[$6>>2] = $$lcssa574;
  HEAP32[$545>>2] = $$lcssa574;
  $556 = (($532) + -16)|0;
  $557 = (($556) - ($547))|0;
  $part2remain$23$lcssa = $557;
 } else {
  $part2remain$23$lcssa = $532;
 }
 $558 = ($part2remain$23$lcssa|0)>(0);
 if ($558) {
  $559 = HEAP32[$10>>2]|0;
  $560 = ((($fr)) + 19792|0);
  $561 = HEAP32[$6>>2]|0;
  $562 = (($561) + ($part2remain$23$lcssa))|0;
  $563 = $562 >> 3;
  $564 = (($559) + ($563)|0);
  HEAP32[$10>>2] = $564;
  $565 = $562 & 7;
  HEAP32[$6>>2] = $565;
  HEAP32[$560>>2] = $565;
 } else {
  $566 = ($part2remain$23$lcssa|0)<(0);
  if ($566) {
   $$4 = 1;
   STACKTOP = sp;return ($$4|0);
  }
 }
 $$4 = 0;
 STACKTOP = sp;return ($$4|0);
}
function _INT123_dct36($inbuf,$o1,$o2,$wintab,$tsbuf) {
 $inbuf = $inbuf|0;
 $o1 = $o1|0;
 $o2 = $o2|0;
 $wintab = $wintab|0;
 $tsbuf = $tsbuf|0;
 var $$neg11 = 0.0, $$neg12 = 0.0, $0 = 0, $1 = 0.0, $10 = 0.0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0;
 var $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0.0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0.0, $130 = 0.0, $131 = 0.0;
 var $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0.0, $148 = 0.0, $149 = 0.0, $15 = 0.0;
 var $150 = 0.0, $151 = 0.0, $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0.0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0;
 var $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0, $177 = 0.0, $178 = 0.0, $179 = 0, $18 = 0.0, $180 = 0, $181 = 0.0, $182 = 0.0, $183 = 0, $184 = 0.0, $185 = 0, $186 = 0.0;
 var $187 = 0, $188 = 0.0, $189 = 0.0, $19 = 0.0, $190 = 0.0, $191 = 0, $192 = 0, $193 = 0.0, $194 = 0, $195 = 0.0, $196 = 0.0, $197 = 0.0, $198 = 0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0, $201 = 0.0, $202 = 0.0, $203 = 0;
 var $204 = 0, $205 = 0.0, $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0, $21 = 0.0, $210 = 0.0, $211 = 0, $212 = 0.0, $213 = 0.0, $214 = 0.0, $215 = 0, $216 = 0, $217 = 0.0, $218 = 0, $219 = 0.0, $22 = 0.0, $220 = 0.0, $221 = 0.0;
 var $222 = 0, $223 = 0.0, $224 = 0, $225 = 0.0, $226 = 0.0, $227 = 0, $228 = 0, $229 = 0.0, $23 = 0, $230 = 0.0, $231 = 0, $232 = 0.0, $233 = 0, $234 = 0.0, $235 = 0, $236 = 0.0, $237 = 0.0, $238 = 0.0, $239 = 0, $24 = 0.0;
 var $240 = 0, $241 = 0.0, $242 = 0, $243 = 0.0, $244 = 0.0, $245 = 0.0, $246 = 0, $247 = 0.0, $248 = 0, $249 = 0.0, $25 = 0.0, $250 = 0.0, $251 = 0, $252 = 0, $253 = 0.0, $254 = 0.0, $255 = 0, $256 = 0.0, $257 = 0, $258 = 0.0;
 var $259 = 0, $26 = 0, $260 = 0.0, $261 = 0.0, $262 = 0.0, $263 = 0, $264 = 0, $265 = 0.0, $266 = 0, $267 = 0.0, $268 = 0.0, $269 = 0.0, $27 = 0.0, $270 = 0, $271 = 0.0, $272 = 0, $273 = 0.0, $274 = 0.0, $275 = 0, $276 = 0;
 var $277 = 0.0, $278 = 0.0, $279 = 0, $28 = 0.0, $280 = 0.0, $281 = 0, $282 = 0.0, $283 = 0, $284 = 0.0, $285 = 0.0, $286 = 0.0, $287 = 0, $288 = 0, $289 = 0.0, $29 = 0, $290 = 0, $291 = 0.0, $292 = 0.0, $293 = 0.0, $294 = 0;
 var $295 = 0.0, $296 = 0, $297 = 0.0, $298 = 0.0, $299 = 0, $3 = 0.0, $30 = 0.0, $300 = 0, $301 = 0.0, $302 = 0.0, $303 = 0, $304 = 0.0, $305 = 0, $306 = 0.0, $307 = 0, $308 = 0.0, $309 = 0.0, $31 = 0.0, $310 = 0.0, $311 = 0;
 var $312 = 0, $313 = 0.0, $314 = 0, $315 = 0.0, $316 = 0.0, $317 = 0.0, $318 = 0, $319 = 0.0, $32 = 0, $320 = 0, $321 = 0.0, $322 = 0.0, $323 = 0, $324 = 0, $325 = 0.0, $326 = 0.0, $327 = 0, $328 = 0.0, $329 = 0, $33 = 0.0;
 var $330 = 0.0, $331 = 0, $332 = 0.0, $333 = 0.0, $334 = 0.0, $335 = 0, $336 = 0, $337 = 0.0, $338 = 0, $339 = 0.0, $34 = 0.0, $340 = 0.0, $341 = 0.0, $342 = 0, $343 = 0.0, $344 = 0, $345 = 0.0, $346 = 0.0, $347 = 0, $348 = 0;
 var $349 = 0.0, $35 = 0, $350 = 0.0, $351 = 0, $352 = 0.0, $353 = 0, $354 = 0.0, $355 = 0, $356 = 0.0, $357 = 0.0, $358 = 0.0, $359 = 0, $36 = 0.0, $360 = 0, $361 = 0.0, $362 = 0, $363 = 0.0, $364 = 0.0, $365 = 0.0, $366 = 0;
 var $367 = 0.0, $368 = 0, $369 = 0.0, $37 = 0.0, $370 = 0.0, $371 = 0, $372 = 0, $373 = 0.0, $374 = 0.0, $375 = 0.0, $376 = 0.0, $377 = 0.0, $378 = 0.0, $379 = 0.0, $38 = 0, $380 = 0, $381 = 0.0, $382 = 0, $383 = 0.0, $384 = 0.0;
 var $385 = 0.0, $386 = 0, $39 = 0.0, $4 = 0.0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0, $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0;
 var $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0.0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0.0, $70 = 0.0, $71 = 0.0, $72 = 0.0;
 var $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0.0, $90 = 0.0;
 var $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $factor = 0.0, $factor7 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($inbuf)) + 64|0);
 $1 = +HEAPF32[$0>>2];
 $2 = ((($inbuf)) + 68|0);
 $3 = +HEAPF32[$2>>2];
 $4 = $3 + $1;
 $5 = ((($inbuf)) + 60|0);
 $6 = +HEAPF32[$5>>2];
 $7 = $1 + $6;
 HEAPF32[$0>>2] = $7;
 $8 = ((($inbuf)) + 56|0);
 $9 = +HEAPF32[$8>>2];
 $10 = $9 + $6;
 $11 = ((($inbuf)) + 52|0);
 $12 = +HEAPF32[$11>>2];
 $13 = $12 + $9;
 HEAPF32[$8>>2] = $13;
 $14 = ((($inbuf)) + 48|0);
 $15 = +HEAPF32[$14>>2];
 $16 = $15 + $12;
 $17 = ((($inbuf)) + 44|0);
 $18 = +HEAPF32[$17>>2];
 $19 = $18 + $15;
 HEAPF32[$14>>2] = $19;
 $20 = ((($inbuf)) + 40|0);
 $21 = +HEAPF32[$20>>2];
 $22 = $21 + $18;
 $23 = ((($inbuf)) + 36|0);
 $24 = +HEAPF32[$23>>2];
 $25 = $24 + $21;
 HEAPF32[$20>>2] = $25;
 $26 = ((($inbuf)) + 32|0);
 $27 = +HEAPF32[$26>>2];
 $28 = $27 + $24;
 $29 = ((($inbuf)) + 28|0);
 $30 = +HEAPF32[$29>>2];
 $31 = $30 + $27;
 HEAPF32[$26>>2] = $31;
 $32 = ((($inbuf)) + 24|0);
 $33 = +HEAPF32[$32>>2];
 $34 = $33 + $30;
 $35 = ((($inbuf)) + 20|0);
 $36 = +HEAPF32[$35>>2];
 $37 = $36 + $33;
 HEAPF32[$32>>2] = $37;
 $38 = ((($inbuf)) + 16|0);
 $39 = +HEAPF32[$38>>2];
 $40 = $39 + $36;
 $41 = ((($inbuf)) + 12|0);
 $42 = +HEAPF32[$41>>2];
 $43 = $42 + $39;
 HEAPF32[$38>>2] = $43;
 $44 = ((($inbuf)) + 8|0);
 $45 = +HEAPF32[$44>>2];
 $46 = $45 + $42;
 $47 = ((($inbuf)) + 4|0);
 $48 = +HEAPF32[$47>>2];
 $49 = $48 + $45;
 HEAPF32[$44>>2] = $49;
 $50 = +HEAPF32[$inbuf>>2];
 $51 = $50 + $48;
 HEAPF32[$47>>2] = $51;
 $52 = $4 + $10;
 HEAPF32[$2>>2] = $52;
 $53 = $16 + $10;
 HEAPF32[$5>>2] = $53;
 $54 = $22 + $16;
 HEAPF32[$11>>2] = $54;
 $55 = $28 + $22;
 HEAPF32[$17>>2] = $55;
 $56 = $34 + $28;
 HEAPF32[$23>>2] = $56;
 $57 = $40 + $34;
 HEAPF32[$29>>2] = $57;
 $58 = $46 + $40;
 HEAPF32[$35>>2] = $58;
 $59 = $51 + $46;
 HEAPF32[$41>>2] = $59;
 $60 = +HEAPF32[179737];
 $61 = $7 + $31;
 $62 = $61 - $43;
 $63 = $62 * $60;
 $64 = $19 * $60;
 $65 = $50 - $64;
 $66 = $65 - $64;
 $67 = $66 - $63;
 $factor = $63 * 2.0;
 $68 = $66 + $factor;
 $69 = $50 + $64;
 $70 = +HEAPF32[179738];
 $71 = $13 + $25;
 $72 = $71 - $49;
 $73 = $72 * $70;
 $74 = $67 - $73;
 $75 = $73 + $67;
 $76 = +HEAPF32[179739];
 $77 = $43 + $31;
 $78 = $76 * $77;
 $79 = +HEAPF32[179740];
 $80 = $31 - $7;
 $81 = $79 * $80;
 $82 = +HEAPF32[179741];
 $83 = $43 + $7;
 $84 = $82 * $83;
 $85 = $69 - $78;
 $86 = $85 - $84;
 $87 = $78 + $69;
 $88 = $87 + $81;
 $89 = $69 - $81;
 $90 = $89 + $84;
 $91 = +HEAPF32[179742];
 $92 = $49 + $25;
 $93 = $91 * $92;
 $94 = +HEAPF32[179743];
 $95 = $25 - $13;
 $96 = $94 * $95;
 $97 = $37 * $70;
 $98 = $96 + $93;
 $99 = $98 + $97;
 $100 = $99 + $88;
 $101 = $88 - $99;
 $102 = $96 - $97;
 $103 = $93 - $97;
 $104 = +HEAPF32[179744];
 $105 = $49 + $13;
 $106 = $104 * $105;
 $107 = $103 + $106;
 $108 = $107 + $90;
 $109 = $90 - $107;
 $110 = $102 - $106;
 $111 = $110 + $86;
 $112 = $86 - $110;
 $113 = $54 * $60;
 $114 = $52 + $56;
 $115 = $114 - $58;
 $116 = $115 * $60;
 $117 = $51 + $113;
 $118 = $51 - $113;
 $119 = $118 - $113;
 $120 = $119 - $116;
 $121 = $58 + $56;
 $122 = $121 * $76;
 $123 = $56 - $52;
 $124 = $123 * $79;
 $factor7 = $116 * 2.0;
 $125 = $factor7 + $119;
 $126 = +HEAPF32[(667196)>>2];
 $127 = $125 * $126;
 $128 = $58 + $52;
 $129 = $128 * $82;
 $130 = $117 - $122;
 $131 = $130 - $129;
 $132 = $117 + $124;
 $133 = $132 + $122;
 $134 = $129 - $124;
 $135 = $134 + $117;
 $136 = $55 + $59;
 $137 = $136 * $91;
 $138 = $55 - $53;
 $139 = $138 * $94;
 $140 = $57 * $70;
 $141 = $139 + $137;
 $142 = $141 + $140;
 $143 = $142 + $133;
 $144 = +HEAPF32[166795];
 $145 = $143 * $144;
 $146 = $133 - $142;
 $147 = +HEAPF32[(667212)>>2];
 $148 = $146 * $147;
 $149 = $53 + $59;
 $150 = $149 * $104;
 $151 = $150 - $140;
 $152 = $151 + $137;
 $153 = $152 + $135;
 $154 = +HEAPF32[(667192)>>2];
 $155 = $153 * $154;
 $156 = $55 - $59;
 $157 = $156 + $53;
 $158 = $157 * $70;
 $159 = $135 - $152;
 $160 = +HEAPF32[(667200)>>2];
 $161 = $159 * $160;
 $$neg11 = -$140;
 $$neg12 = $$neg11 - $150;
 $162 = $$neg12 + $139;
 $163 = $120 - $158;
 $164 = +HEAPF32[(667184)>>2];
 $165 = $164 * $163;
 $166 = $158 + $120;
 $167 = +HEAPF32[(667208)>>2];
 $168 = $167 * $166;
 $169 = $162 + $131;
 $170 = +HEAPF32[(667188)>>2];
 $171 = $170 * $169;
 $172 = $131 - $162;
 $173 = +HEAPF32[(667204)>>2];
 $174 = $173 * $172;
 $175 = $145 + $100;
 $176 = ((($wintab)) + 108|0);
 $177 = +HEAPF32[$176>>2];
 $178 = $177 * $175;
 $179 = ((($o2)) + 36|0);
 HEAPF32[$179>>2] = $178;
 $180 = ((($wintab)) + 104|0);
 $181 = +HEAPF32[$180>>2];
 $182 = $181 * $175;
 $183 = ((($o2)) + 32|0);
 HEAPF32[$183>>2] = $182;
 $184 = $100 - $145;
 $185 = ((($o1)) + 32|0);
 $186 = +HEAPF32[$185>>2];
 $187 = ((($wintab)) + 32|0);
 $188 = +HEAPF32[$187>>2];
 $189 = $188 * $184;
 $190 = $189 + $186;
 $191 = ((($tsbuf)) + 1024|0);
 HEAPF32[$191>>2] = $190;
 $192 = ((($o1)) + 36|0);
 $193 = +HEAPF32[$192>>2];
 $194 = ((($wintab)) + 36|0);
 $195 = +HEAPF32[$194>>2];
 $196 = $195 * $184;
 $197 = $196 + $193;
 $198 = ((($tsbuf)) + 1152|0);
 HEAPF32[$198>>2] = $197;
 $199 = $165 + $74;
 $200 = ((($wintab)) + 112|0);
 $201 = +HEAPF32[$200>>2];
 $202 = $201 * $199;
 $203 = ((($o2)) + 40|0);
 HEAPF32[$203>>2] = $202;
 $204 = ((($wintab)) + 100|0);
 $205 = +HEAPF32[$204>>2];
 $206 = $205 * $199;
 $207 = ((($o2)) + 28|0);
 HEAPF32[$207>>2] = $206;
 $208 = $74 - $165;
 $209 = ((($o1)) + 28|0);
 $210 = +HEAPF32[$209>>2];
 $211 = ((($wintab)) + 28|0);
 $212 = +HEAPF32[$211>>2];
 $213 = $212 * $208;
 $214 = $213 + $210;
 $215 = ((($tsbuf)) + 896|0);
 HEAPF32[$215>>2] = $214;
 $216 = ((($o1)) + 40|0);
 $217 = +HEAPF32[$216>>2];
 $218 = ((($wintab)) + 40|0);
 $219 = +HEAPF32[$218>>2];
 $220 = $219 * $208;
 $221 = $220 + $217;
 $222 = ((($tsbuf)) + 1280|0);
 HEAPF32[$222>>2] = $221;
 $223 = $171 + $111;
 $224 = ((($wintab)) + 116|0);
 $225 = +HEAPF32[$224>>2];
 $226 = $225 * $223;
 $227 = ((($o2)) + 44|0);
 HEAPF32[$227>>2] = $226;
 $228 = ((($wintab)) + 96|0);
 $229 = +HEAPF32[$228>>2];
 $230 = $229 * $223;
 $231 = ((($o2)) + 24|0);
 HEAPF32[$231>>2] = $230;
 $232 = $111 - $171;
 $233 = ((($o1)) + 24|0);
 $234 = +HEAPF32[$233>>2];
 $235 = ((($wintab)) + 24|0);
 $236 = +HEAPF32[$235>>2];
 $237 = $236 * $232;
 $238 = $237 + $234;
 $239 = ((($tsbuf)) + 768|0);
 HEAPF32[$239>>2] = $238;
 $240 = ((($o1)) + 44|0);
 $241 = +HEAPF32[$240>>2];
 $242 = ((($wintab)) + 44|0);
 $243 = +HEAPF32[$242>>2];
 $244 = $243 * $232;
 $245 = $244 + $241;
 $246 = ((($tsbuf)) + 1408|0);
 HEAPF32[$246>>2] = $245;
 $247 = $155 + $108;
 $248 = ((($wintab)) + 120|0);
 $249 = +HEAPF32[$248>>2];
 $250 = $249 * $247;
 $251 = ((($o2)) + 48|0);
 HEAPF32[$251>>2] = $250;
 $252 = ((($wintab)) + 92|0);
 $253 = +HEAPF32[$252>>2];
 $254 = $253 * $247;
 $255 = ((($o2)) + 20|0);
 HEAPF32[$255>>2] = $254;
 $256 = $108 - $155;
 $257 = ((($o1)) + 20|0);
 $258 = +HEAPF32[$257>>2];
 $259 = ((($wintab)) + 20|0);
 $260 = +HEAPF32[$259>>2];
 $261 = $260 * $256;
 $262 = $261 + $258;
 $263 = ((($tsbuf)) + 640|0);
 HEAPF32[$263>>2] = $262;
 $264 = ((($o1)) + 48|0);
 $265 = +HEAPF32[$264>>2];
 $266 = ((($wintab)) + 48|0);
 $267 = +HEAPF32[$266>>2];
 $268 = $267 * $256;
 $269 = $268 + $265;
 $270 = ((($tsbuf)) + 1536|0);
 HEAPF32[$270>>2] = $269;
 $271 = $127 + $68;
 $272 = ((($wintab)) + 124|0);
 $273 = +HEAPF32[$272>>2];
 $274 = $273 * $271;
 $275 = ((($o2)) + 52|0);
 HEAPF32[$275>>2] = $274;
 $276 = ((($wintab)) + 88|0);
 $277 = +HEAPF32[$276>>2];
 $278 = $277 * $271;
 $279 = ((($o2)) + 16|0);
 HEAPF32[$279>>2] = $278;
 $280 = $68 - $127;
 $281 = ((($o1)) + 16|0);
 $282 = +HEAPF32[$281>>2];
 $283 = ((($wintab)) + 16|0);
 $284 = +HEAPF32[$283>>2];
 $285 = $284 * $280;
 $286 = $285 + $282;
 $287 = ((($tsbuf)) + 512|0);
 HEAPF32[$287>>2] = $286;
 $288 = ((($o1)) + 52|0);
 $289 = +HEAPF32[$288>>2];
 $290 = ((($wintab)) + 52|0);
 $291 = +HEAPF32[$290>>2];
 $292 = $291 * $280;
 $293 = $292 + $289;
 $294 = ((($tsbuf)) + 1664|0);
 HEAPF32[$294>>2] = $293;
 $295 = $161 + $109;
 $296 = ((($wintab)) + 128|0);
 $297 = +HEAPF32[$296>>2];
 $298 = $297 * $295;
 $299 = ((($o2)) + 56|0);
 HEAPF32[$299>>2] = $298;
 $300 = ((($wintab)) + 84|0);
 $301 = +HEAPF32[$300>>2];
 $302 = $301 * $295;
 $303 = ((($o2)) + 12|0);
 HEAPF32[$303>>2] = $302;
 $304 = $109 - $161;
 $305 = ((($o1)) + 12|0);
 $306 = +HEAPF32[$305>>2];
 $307 = ((($wintab)) + 12|0);
 $308 = +HEAPF32[$307>>2];
 $309 = $308 * $304;
 $310 = $309 + $306;
 $311 = ((($tsbuf)) + 384|0);
 HEAPF32[$311>>2] = $310;
 $312 = ((($o1)) + 56|0);
 $313 = +HEAPF32[$312>>2];
 $314 = ((($wintab)) + 56|0);
 $315 = +HEAPF32[$314>>2];
 $316 = $315 * $304;
 $317 = $316 + $313;
 $318 = ((($tsbuf)) + 1792|0);
 HEAPF32[$318>>2] = $317;
 $319 = $174 + $112;
 $320 = ((($wintab)) + 132|0);
 $321 = +HEAPF32[$320>>2];
 $322 = $321 * $319;
 $323 = ((($o2)) + 60|0);
 HEAPF32[$323>>2] = $322;
 $324 = ((($wintab)) + 80|0);
 $325 = +HEAPF32[$324>>2];
 $326 = $325 * $319;
 $327 = ((($o2)) + 8|0);
 HEAPF32[$327>>2] = $326;
 $328 = $112 - $174;
 $329 = ((($o1)) + 8|0);
 $330 = +HEAPF32[$329>>2];
 $331 = ((($wintab)) + 8|0);
 $332 = +HEAPF32[$331>>2];
 $333 = $332 * $328;
 $334 = $333 + $330;
 $335 = ((($tsbuf)) + 256|0);
 HEAPF32[$335>>2] = $334;
 $336 = ((($o1)) + 60|0);
 $337 = +HEAPF32[$336>>2];
 $338 = ((($wintab)) + 60|0);
 $339 = +HEAPF32[$338>>2];
 $340 = $339 * $328;
 $341 = $340 + $337;
 $342 = ((($tsbuf)) + 1920|0);
 HEAPF32[$342>>2] = $341;
 $343 = $168 + $75;
 $344 = ((($wintab)) + 136|0);
 $345 = +HEAPF32[$344>>2];
 $346 = $345 * $343;
 $347 = ((($o2)) + 64|0);
 HEAPF32[$347>>2] = $346;
 $348 = ((($wintab)) + 76|0);
 $349 = +HEAPF32[$348>>2];
 $350 = $349 * $343;
 $351 = ((($o2)) + 4|0);
 HEAPF32[$351>>2] = $350;
 $352 = $75 - $168;
 $353 = ((($o1)) + 4|0);
 $354 = +HEAPF32[$353>>2];
 $355 = ((($wintab)) + 4|0);
 $356 = +HEAPF32[$355>>2];
 $357 = $356 * $352;
 $358 = $357 + $354;
 $359 = ((($tsbuf)) + 128|0);
 HEAPF32[$359>>2] = $358;
 $360 = ((($o1)) + 64|0);
 $361 = +HEAPF32[$360>>2];
 $362 = ((($wintab)) + 64|0);
 $363 = +HEAPF32[$362>>2];
 $364 = $363 * $352;
 $365 = $364 + $361;
 $366 = ((($tsbuf)) + 2048|0);
 HEAPF32[$366>>2] = $365;
 $367 = $148 + $101;
 $368 = ((($wintab)) + 140|0);
 $369 = +HEAPF32[$368>>2];
 $370 = $369 * $367;
 $371 = ((($o2)) + 68|0);
 HEAPF32[$371>>2] = $370;
 $372 = ((($wintab)) + 72|0);
 $373 = +HEAPF32[$372>>2];
 $374 = $373 * $367;
 HEAPF32[$o2>>2] = $374;
 $375 = $101 - $148;
 $376 = +HEAPF32[$o1>>2];
 $377 = +HEAPF32[$wintab>>2];
 $378 = $377 * $375;
 $379 = $378 + $376;
 HEAPF32[$tsbuf>>2] = $379;
 $380 = ((($o1)) + 68|0);
 $381 = +HEAPF32[$380>>2];
 $382 = ((($wintab)) + 68|0);
 $383 = +HEAPF32[$382>>2];
 $384 = $383 * $375;
 $385 = $384 + $381;
 $386 = ((($tsbuf)) + 2176|0);
 HEAPF32[$386>>2] = $385;
 return;
}
function _dct12($in,$rawout1,$rawout2,$wi,$ts) {
 $in = $in|0;
 $rawout1 = $rawout1|0;
 $rawout2 = $rawout2|0;
 $wi = $wi|0;
 $ts = $ts|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0, $109 = 0, $11 = 0, $110 = 0.0, $111 = 0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0;
 var $116 = 0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0, $123 = 0.0, $124 = 0, $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0, $129 = 0, $13 = 0, $130 = 0.0, $131 = 0, $132 = 0.0, $133 = 0.0;
 var $134 = 0.0, $135 = 0, $136 = 0, $137 = 0.0, $138 = 0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0, $143 = 0, $144 = 0, $145 = 0.0, $146 = 0, $147 = 0.0, $148 = 0.0, $149 = 0, $15 = 0, $150 = 0.0, $151 = 0.0;
 var $152 = 0, $153 = 0.0, $154 = 0.0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0.0, $17 = 0.0;
 var $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0, $174 = 0.0, $175 = 0.0, $176 = 0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0.0, $182 = 0.0, $183 = 0.0, $184 = 0.0, $185 = 0.0, $186 = 0.0, $187 = 0.0, $188 = 0.0;
 var $189 = 0.0, $19 = 0.0, $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0.0, $194 = 0.0, $195 = 0.0, $196 = 0.0, $197 = 0.0, $198 = 0.0, $199 = 0.0, $2 = 0, $20 = 0.0, $200 = 0.0, $201 = 0.0, $202 = 0.0, $203 = 0.0, $204 = 0, $205 = 0.0;
 var $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0, $21 = 0, $210 = 0.0, $211 = 0.0, $212 = 0, $213 = 0.0, $214 = 0.0, $215 = 0.0, $216 = 0.0, $217 = 0.0, $218 = 0.0, $219 = 0.0, $22 = 0.0, $220 = 0.0, $221 = 0.0, $222 = 0.0, $223 = 0.0;
 var $224 = 0.0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0.0, $229 = 0, $23 = 0.0, $230 = 0, $231 = 0, $232 = 0.0, $233 = 0, $234 = 0.0, $235 = 0.0, $236 = 0, $237 = 0.0, $238 = 0.0, $239 = 0, $24 = 0, $240 = 0.0, $241 = 0.0;
 var $242 = 0, $243 = 0.0, $244 = 0.0, $245 = 0.0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0.0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0.0, $254 = 0.0, $255 = 0.0, $256 = 0.0, $257 = 0.0, $258 = 0.0, $259 = 0.0, $26 = 0.0;
 var $260 = 0, $261 = 0.0, $262 = 0.0, $263 = 0, $264 = 0.0, $265 = 0.0, $266 = 0.0, $267 = 0.0, $268 = 0.0, $269 = 0.0, $27 = 0, $270 = 0.0, $271 = 0.0, $272 = 0.0, $273 = 0.0, $274 = 0.0, $275 = 0.0, $276 = 0.0, $277 = 0.0, $278 = 0.0;
 var $279 = 0.0, $28 = 0.0, $280 = 0.0, $281 = 0.0, $282 = 0.0, $283 = 0.0, $284 = 0.0, $285 = 0.0, $286 = 0.0, $287 = 0.0, $288 = 0.0, $289 = 0.0, $29 = 0.0, $290 = 0.0, $291 = 0, $292 = 0.0, $293 = 0.0, $294 = 0, $295 = 0.0, $296 = 0.0;
 var $297 = 0, $298 = 0.0, $299 = 0.0, $3 = 0, $30 = 0.0, $300 = 0, $301 = 0.0, $302 = 0.0, $303 = 0.0, $304 = 0.0, $305 = 0.0, $306 = 0.0, $307 = 0.0, $308 = 0.0, $309 = 0.0, $31 = 0.0, $310 = 0.0, $311 = 0.0, $312 = 0.0, $313 = 0.0;
 var $314 = 0.0, $315 = 0.0, $316 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0.0, $45 = 0, $46 = 0.0, $47 = 0.0;
 var $48 = 0.0, $49 = 0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0, $64 = 0, $65 = 0.0;
 var $66 = 0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0;
 var $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0, $89 = 0.0, $9 = 0, $90 = 0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0, $95 = 0, $96 = 0.0, $97 = 0, $98 = 0.0, $99 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$rawout1>>2]|0;
 HEAP32[$ts>>2] = $0;
 $1 = ((($rawout1)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($ts)) + 128|0);
 HEAP32[$3>>2] = $2;
 $4 = ((($rawout1)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($ts)) + 256|0);
 HEAP32[$6>>2] = $5;
 $7 = ((($rawout1)) + 12|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($ts)) + 384|0);
 HEAP32[$9>>2] = $8;
 $10 = ((($rawout1)) + 16|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($ts)) + 512|0);
 HEAP32[$12>>2] = $11;
 $13 = ((($rawout1)) + 20|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($ts)) + 640|0);
 HEAP32[$15>>2] = $14;
 $16 = ((($in)) + 60|0);
 $17 = +HEAPF32[$16>>2];
 $18 = ((($in)) + 48|0);
 $19 = +HEAPF32[$18>>2];
 $20 = $19 + $17;
 $21 = ((($in)) + 36|0);
 $22 = +HEAPF32[$21>>2];
 $23 = $22 + $19;
 $24 = ((($in)) + 24|0);
 $25 = +HEAPF32[$24>>2];
 $26 = $25 + $22;
 $27 = ((($in)) + 12|0);
 $28 = +HEAPF32[$27>>2];
 $29 = $28 + $25;
 $30 = +HEAPF32[$in>>2];
 $31 = $30 + $28;
 $32 = $20 + $26;
 $33 = $31 + $26;
 $34 = +HEAPF32[179738];
 $35 = $34 * $29;
 $36 = $33 * $34;
 $37 = $30 - $23;
 $38 = $31 - $32;
 $39 = +HEAPF32[(719560)>>2];
 $40 = $38 * $39;
 $41 = $40 + $37;
 $42 = $37 - $40;
 $43 = ((($rawout1)) + 64|0);
 $44 = +HEAPF32[$43>>2];
 $45 = ((($wi)) + 40|0);
 $46 = +HEAPF32[$45>>2];
 $47 = $41 * $46;
 $48 = $47 + $44;
 $49 = ((($ts)) + 2048|0);
 HEAPF32[$49>>2] = $48;
 $50 = ((($rawout1)) + 52|0);
 $51 = +HEAPF32[$50>>2];
 $52 = ((($wi)) + 28|0);
 $53 = +HEAPF32[$52>>2];
 $54 = $53 * $41;
 $55 = $54 + $51;
 $56 = ((($ts)) + 1664|0);
 HEAPF32[$56>>2] = $55;
 $57 = ((($rawout1)) + 28|0);
 $58 = +HEAPF32[$57>>2];
 $59 = ((($wi)) + 4|0);
 $60 = +HEAPF32[$59>>2];
 $61 = $60 * $42;
 $62 = $61 + $58;
 $63 = ((($ts)) + 896|0);
 HEAPF32[$63>>2] = $62;
 $64 = ((($rawout1)) + 40|0);
 $65 = +HEAPF32[$64>>2];
 $66 = ((($wi)) + 16|0);
 $67 = +HEAPF32[$66>>2];
 $68 = $67 * $42;
 $69 = $68 + $65;
 $70 = ((($ts)) + 1280|0);
 HEAPF32[$70>>2] = $69;
 $71 = +HEAPF32[179737];
 $72 = $71 * $23;
 $73 = $72 + $30;
 $74 = $73 + $35;
 $75 = $73 - $35;
 $76 = $71 * $32;
 $77 = $76 + $31;
 $78 = $77 + $36;
 $79 = +HEAPF32[179889];
 $80 = $78 * $79;
 $81 = $77 - $36;
 $82 = +HEAPF32[(719564)>>2];
 $83 = $81 * $82;
 $84 = $80 + $74;
 $85 = $74 - $80;
 $86 = $83 + $75;
 $87 = $75 - $83;
 $88 = ((($rawout1)) + 68|0);
 $89 = +HEAPF32[$88>>2];
 $90 = ((($wi)) + 44|0);
 $91 = +HEAPF32[$90>>2];
 $92 = $86 * $91;
 $93 = $92 + $89;
 $94 = ((($ts)) + 2176|0);
 HEAPF32[$94>>2] = $93;
 $95 = ((($rawout1)) + 48|0);
 $96 = +HEAPF32[$95>>2];
 $97 = ((($wi)) + 24|0);
 $98 = +HEAPF32[$97>>2];
 $99 = $98 * $86;
 $100 = $99 + $96;
 $101 = ((($ts)) + 1536|0);
 HEAPF32[$101>>2] = $100;
 $102 = ((($rawout1)) + 56|0);
 $103 = +HEAPF32[$102>>2];
 $104 = ((($wi)) + 32|0);
 $105 = +HEAPF32[$104>>2];
 $106 = $105 * $84;
 $107 = $106 + $103;
 $108 = ((($ts)) + 1792|0);
 HEAPF32[$108>>2] = $107;
 $109 = ((($rawout1)) + 60|0);
 $110 = +HEAPF32[$109>>2];
 $111 = ((($wi)) + 36|0);
 $112 = +HEAPF32[$111>>2];
 $113 = $112 * $84;
 $114 = $113 + $110;
 $115 = ((($ts)) + 1920|0);
 HEAPF32[$115>>2] = $114;
 $116 = ((($rawout1)) + 24|0);
 $117 = +HEAPF32[$116>>2];
 $118 = +HEAPF32[$wi>>2];
 $119 = $118 * $87;
 $120 = $119 + $117;
 $121 = ((($ts)) + 768|0);
 HEAPF32[$121>>2] = $120;
 $122 = ((($rawout1)) + 44|0);
 $123 = +HEAPF32[$122>>2];
 $124 = ((($wi)) + 20|0);
 $125 = +HEAPF32[$124>>2];
 $126 = $125 * $87;
 $127 = $126 + $123;
 $128 = ((($ts)) + 1408|0);
 HEAPF32[$128>>2] = $127;
 $129 = ((($rawout1)) + 32|0);
 $130 = +HEAPF32[$129>>2];
 $131 = ((($wi)) + 8|0);
 $132 = +HEAPF32[$131>>2];
 $133 = $132 * $85;
 $134 = $133 + $130;
 $135 = ((($ts)) + 1024|0);
 HEAPF32[$135>>2] = $134;
 $136 = ((($rawout1)) + 36|0);
 $137 = +HEAPF32[$136>>2];
 $138 = ((($wi)) + 12|0);
 $139 = +HEAPF32[$138>>2];
 $140 = $139 * $85;
 $141 = $140 + $137;
 $142 = ((($ts)) + 1152|0);
 HEAPF32[$142>>2] = $141;
 $143 = ((($in)) + 4|0);
 $144 = ((($in)) + 64|0);
 $145 = +HEAPF32[$144>>2];
 $146 = ((($in)) + 52|0);
 $147 = +HEAPF32[$146>>2];
 $148 = $147 + $145;
 $149 = ((($in)) + 40|0);
 $150 = +HEAPF32[$149>>2];
 $151 = $150 + $147;
 $152 = ((($in)) + 28|0);
 $153 = +HEAPF32[$152>>2];
 $154 = $153 + $150;
 $155 = ((($in)) + 16|0);
 $156 = +HEAPF32[$155>>2];
 $157 = $156 + $153;
 $158 = +HEAPF32[$143>>2];
 $159 = $158 + $156;
 $160 = $148 + $154;
 $161 = $159 + $154;
 $162 = +HEAPF32[179738];
 $163 = $162 * $157;
 $164 = $161 * $162;
 $165 = $158 - $151;
 $166 = $159 - $160;
 $167 = +HEAPF32[(719560)>>2];
 $168 = $166 * $167;
 $169 = $168 + $165;
 $170 = $165 - $168;
 $171 = +HEAPF32[$45>>2];
 $172 = $169 * $171;
 $173 = ((($rawout2)) + 16|0);
 HEAPF32[$173>>2] = $172;
 $174 = +HEAPF32[$52>>2];
 $175 = $174 * $169;
 $176 = ((($rawout2)) + 4|0);
 HEAPF32[$176>>2] = $175;
 $177 = +HEAPF32[$59>>2];
 $178 = $177 * $170;
 $179 = +HEAPF32[$56>>2];
 $180 = $179 + $178;
 HEAPF32[$56>>2] = $180;
 $181 = +HEAPF32[$66>>2];
 $182 = $181 * $170;
 $183 = +HEAPF32[$49>>2];
 $184 = $183 + $182;
 HEAPF32[$49>>2] = $184;
 $185 = +HEAPF32[179737];
 $186 = $185 * $151;
 $187 = $186 + $158;
 $188 = $187 + $163;
 $189 = $187 - $163;
 $190 = $185 * $160;
 $191 = $190 + $159;
 $192 = $191 + $164;
 $193 = +HEAPF32[179889];
 $194 = $192 * $193;
 $195 = $191 - $164;
 $196 = +HEAPF32[(719564)>>2];
 $197 = $195 * $196;
 $198 = $194 + $188;
 $199 = $188 - $194;
 $200 = $197 + $189;
 $201 = $189 - $197;
 $202 = +HEAPF32[$90>>2];
 $203 = $200 * $202;
 $204 = ((($rawout2)) + 20|0);
 HEAPF32[$204>>2] = $203;
 $205 = +HEAPF32[$97>>2];
 $206 = $200 * $205;
 HEAPF32[$rawout2>>2] = $206;
 $207 = +HEAPF32[$104>>2];
 $208 = $207 * $198;
 $209 = ((($rawout2)) + 8|0);
 HEAPF32[$209>>2] = $208;
 $210 = +HEAPF32[$111>>2];
 $211 = $210 * $198;
 $212 = ((($rawout2)) + 12|0);
 HEAPF32[$212>>2] = $211;
 $213 = +HEAPF32[$wi>>2];
 $214 = $213 * $201;
 $215 = +HEAPF32[$101>>2];
 $216 = $215 + $214;
 HEAPF32[$101>>2] = $216;
 $217 = +HEAPF32[$124>>2];
 $218 = $217 * $201;
 $219 = +HEAPF32[$94>>2];
 $220 = $219 + $218;
 HEAPF32[$94>>2] = $220;
 $221 = +HEAPF32[$131>>2];
 $222 = $221 * $199;
 $223 = +HEAPF32[$108>>2];
 $224 = $223 + $222;
 HEAPF32[$108>>2] = $224;
 $225 = +HEAPF32[$138>>2];
 $226 = $225 * $199;
 $227 = +HEAPF32[$115>>2];
 $228 = $227 + $226;
 HEAPF32[$115>>2] = $228;
 $229 = ((($in)) + 8|0);
 $230 = ((($rawout2)) + 48|0);
 $231 = ((($in)) + 68|0);
 ;HEAP32[$230>>2]=0|0;HEAP32[$230+4>>2]=0|0;HEAP32[$230+8>>2]=0|0;HEAP32[$230+12>>2]=0|0;HEAP32[$230+16>>2]=0|0;HEAP32[$230+20>>2]=0|0;
 $232 = +HEAPF32[$231>>2];
 $233 = ((($in)) + 56|0);
 $234 = +HEAPF32[$233>>2];
 $235 = $234 + $232;
 $236 = ((($in)) + 44|0);
 $237 = +HEAPF32[$236>>2];
 $238 = $237 + $234;
 $239 = ((($in)) + 32|0);
 $240 = +HEAPF32[$239>>2];
 $241 = $240 + $237;
 $242 = ((($in)) + 20|0);
 $243 = +HEAPF32[$242>>2];
 $244 = $243 + $240;
 $245 = +HEAPF32[$229>>2];
 $246 = $245 + $243;
 $247 = $235 + $241;
 $248 = $246 + $241;
 $249 = +HEAPF32[179738];
 $250 = $249 * $244;
 $251 = $248 * $249;
 $252 = $245 - $238;
 $253 = $246 - $247;
 $254 = +HEAPF32[(719560)>>2];
 $255 = $253 * $254;
 $256 = $255 + $252;
 $257 = $252 - $255;
 $258 = +HEAPF32[$45>>2];
 $259 = $256 * $258;
 $260 = ((($rawout2)) + 40|0);
 HEAPF32[$260>>2] = $259;
 $261 = +HEAPF32[$52>>2];
 $262 = $261 * $256;
 $263 = ((($rawout2)) + 28|0);
 HEAPF32[$263>>2] = $262;
 $264 = +HEAPF32[$59>>2];
 $265 = $264 * $257;
 $266 = +HEAPF32[$176>>2];
 $267 = $266 + $265;
 HEAPF32[$176>>2] = $267;
 $268 = +HEAPF32[$66>>2];
 $269 = $268 * $257;
 $270 = +HEAPF32[$173>>2];
 $271 = $270 + $269;
 HEAPF32[$173>>2] = $271;
 $272 = +HEAPF32[179737];
 $273 = $272 * $238;
 $274 = $273 + $245;
 $275 = $274 + $250;
 $276 = $274 - $250;
 $277 = $272 * $247;
 $278 = $277 + $246;
 $279 = $278 + $251;
 $280 = +HEAPF32[179889];
 $281 = $279 * $280;
 $282 = $278 - $251;
 $283 = +HEAPF32[(719564)>>2];
 $284 = $282 * $283;
 $285 = $281 + $275;
 $286 = $275 - $281;
 $287 = $284 + $276;
 $288 = $276 - $284;
 $289 = +HEAPF32[$90>>2];
 $290 = $287 * $289;
 $291 = ((($rawout2)) + 44|0);
 HEAPF32[$291>>2] = $290;
 $292 = +HEAPF32[$97>>2];
 $293 = $287 * $292;
 $294 = ((($rawout2)) + 24|0);
 HEAPF32[$294>>2] = $293;
 $295 = +HEAPF32[$104>>2];
 $296 = $295 * $285;
 $297 = ((($rawout2)) + 32|0);
 HEAPF32[$297>>2] = $296;
 $298 = +HEAPF32[$111>>2];
 $299 = $298 * $285;
 $300 = ((($rawout2)) + 36|0);
 HEAPF32[$300>>2] = $299;
 $301 = +HEAPF32[$wi>>2];
 $302 = $301 * $288;
 $303 = +HEAPF32[$rawout2>>2];
 $304 = $303 + $302;
 HEAPF32[$rawout2>>2] = $304;
 $305 = +HEAPF32[$124>>2];
 $306 = $305 * $288;
 $307 = +HEAPF32[$204>>2];
 $308 = $307 + $306;
 HEAPF32[$204>>2] = $308;
 $309 = +HEAPF32[$131>>2];
 $310 = $309 * $286;
 $311 = +HEAPF32[$209>>2];
 $312 = $311 + $310;
 HEAPF32[$209>>2] = $312;
 $313 = +HEAPF32[$138>>2];
 $314 = $313 * $286;
 $315 = +HEAPF32[$212>>2];
 $316 = $315 + $314;
 HEAPF32[$212>>2] = $316;
 return;
}
function _INT123_frame_gapless_init($fr,$framecount,$bskip,$eskip) {
 $fr = $fr|0;
 $framecount = $framecount|0;
 $bskip = $bskip|0;
 $eskip = $eskip|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27920|0);
 HEAP32[$0>>2] = $framecount;
 $1 = ($framecount|0)>(0);
 if (!($1)) {
  $19 = ((($fr)) + 27940|0);
  HEAP32[$19>>2] = 0;
  $20 = ((($fr)) + 27932|0);
  HEAP32[$20>>2] = 0;
  $21 = ((($fr)) + 27936|0);
  HEAP32[$21>>2] = 0;
  $22 = ((($fr)) + 27944|0);
  HEAP32[$22>>2] = 0;
  $23 = ((($fr)) + 27948|0);
  HEAP32[$23>>2] = 0;
  return;
 }
 $2 = (($bskip) + 529)|0;
 $3 = ((($fr)) + 27932|0);
 HEAP32[$3>>2] = $2;
 $4 = ((($fr)) + 19688|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(1);
 if ($6) {
  $15 = 384;
 } else {
  $7 = ($5|0)==(2);
  if ($7) {
   $15 = 1152;
  } else {
   $8 = ((($fr)) + 19672|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = ($9|0)==(0);
   if ($10) {
    $11 = ((($fr)) + 19676|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = ($12|0)!=(0);
    $phitmp = $13 ? 576 : 1152;
    $15 = $phitmp;
   } else {
    $15 = 576;
   }
  }
 }
 $14 = Math_imul($15, $framecount)|0;
 $16 = (529 - ($eskip))|0;
 $17 = (($16) + ($14))|0;
 $18 = ((($fr)) + 27940|0);
 HEAP32[$18>>2] = $17;
 $21 = ((($fr)) + 27936|0);
 HEAP32[$21>>2] = 0;
 $22 = ((($fr)) + 27944|0);
 HEAP32[$22>>2] = 0;
 $23 = ((($fr)) + 27948|0);
 HEAP32[$23>>2] = 0;
 return;
}
function _INT123_frame_fill_toc($fr,$in) {
 $fr = $fr|0;
 $in = $in|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 19772|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $3 = (_malloc(100)|0);
  HEAP32[$0>>2] = $3;
  $4 = ($3|0)==(0|0);
  if ($4) {
   $$0 = 0;
   return ($$0|0);
  } else {
   $5 = $3;
  }
 } else {
  $5 = $1;
 }
 dest=$5; src=$in; stop=dest+100|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
 $$0 = 1;
 return ($$0|0);
}
function _INT123_do_rva($fr) {
 $fr = $fr|0;
 var $$$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0;
 var $26 = 0.0, $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0.0;
 var $newscale$0 = 0.0, $not$$i = 0, $p$1$i1 = 0.0, $rt$0$i = 0, $rvafact$0 = 0.0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_ptr4 = 0, $vararg_ptr5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($fr)) + 28080|0);
 $1 = HEAP32[$0>>2]|0;
 switch ($1|0) {
 case 0:  {
  $p$1$i1 = -1.0;$rvafact$0 = 1.0;
  break;
 }
 case 2:  {
  $2 = ((($fr)) + 19820|0);
  $3 = HEAP32[$2>>2]|0;
  $not$$i = ($3|0)!=(-1);
  $$$i = $not$$i&1;
  $rt$0$i = $$$i;
  label = 3;
  break;
 }
 default: {
  $rt$0$i = 0;
  label = 3;
 }
 }
 if ((label|0) == 3) {
  $4 = (((($fr)) + 19816|0) + ($rt$0$i<<2)|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ($5|0)==(-1);
  if ($6) {
   $p$1$i1 = -1.0;$rvafact$0 = 1.0;
  } else {
   $7 = (((($fr)) + 19832|0) + ($rt$0$i<<2)|0);
   $8 = +HEAPF32[$7>>2];
   $9 = $8;
   $10 = (((($fr)) + 19824|0) + ($rt$0$i<<2)|0);
   $11 = +HEAPF32[$10>>2];
   $12 = $11;
   $13 = ((($fr)) + 28068|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = $14 & 32;
   $16 = ($15|0)==(0);
   if ($16) {
    $17 = ((($fr)) + 28064|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = ($18|0)>(1);
    if ($19) {
     $20 = HEAP32[37544]|0;
     HEAPF64[$vararg_buffer>>3] = $12;
     (_fprintf($20,513525,$vararg_buffer)|0);
    }
   }
   $21 = $12 * 0.050000000000000003;
   $22 = (+Math_pow(10.0,(+$21)));
   $p$1$i1 = $9;$rvafact$0 = $22;
  }
 }
 $23 = ((($fr)) + 28344|0);
 $24 = +HEAPF64[$23>>3];
 $25 = $24 * $rvafact$0;
 $26 = $25 * $p$1$i1;
 $27 = $26 > 1.0;
 if ($27) {
  $28 = 1.0 / $p$1$i1;
  $29 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer1>>2] = 1019;
  $vararg_ptr4 = ((($vararg_buffer1)) + 8|0);
  HEAPF64[$vararg_ptr4>>3] = $28;
  $vararg_ptr5 = ((($vararg_buffer1)) + 16|0);
  HEAPF64[$vararg_ptr5>>3] = $p$1$i1;
  (_fprintf($29,513555,$vararg_buffer1)|0);
  $newscale$0 = $28;
 } else {
  $newscale$0 = $25;
 }
 $30 = ((($fr)) + 19808|0);
 $31 = +HEAPF64[$30>>3];
 $32 = $newscale$0 != $31;
 if (!($32)) {
  $33 = ((($fr)) + 28380|0);
  $34 = HEAP32[$33>>2]|0;
  $35 = ($34|0)==(0);
  if ($35) {
   STACKTOP = sp;return;
  }
 }
 HEAPF64[$30>>3] = $newscale$0;
 $36 = ((($fr)) + 19648|0);
 $37 = HEAP32[$36>>2]|0;
 $38 = ($37|0)==(0|0);
 if ($38) {
  STACKTOP = sp;return;
 }
 FUNCTION_TABLE_vi[$37 & 8191]($fr);
 STACKTOP = sp;return;
}
function _INT123_fi_add($fi,$pos) {
 $fi = $fi|0;
 $pos = $pos|0;
 var $$lcssa = 0, $$pre = 0, $$pre2 = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $c$02$i = 0, $fi$idx = 0, $fi$idx$val = 0, $fi$idx$val$i = 0, $fi$idx$val$pre$i = 0, $fi$idx1$val$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fi)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fi)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==($3|0);
 do {
  if ($4) {
   $5 = ((($fi)) + 4|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = Math_imul($6, $1)|0;
   $8 = ((($fi)) + 20|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = ($9|0)==(0);
   if ($10) {
    $15 = $1;
    label = 5;
   } else {
    $11 = (($9) + ($1))|0;
    $12 = (_INT123_fi_resize($fi,$11)|0);
    $13 = ($12|0)==(0);
    if (!($13)) {
     $$pre = HEAP32[$0>>2]|0;
     $15 = $$pre;
     label = 5;
    }
   }
   if ((label|0) == 5) {
    $14 = ($15>>>0)<(2);
    if (!($14)) {
     $16 = HEAP32[$5>>2]|0;
     $17 = $16 << 1;
     HEAP32[$5>>2] = $17;
     $18 = $15 >>> 1;
     HEAP32[$0>>2] = $18;
     $19 = ($18|0)==(0);
     if ($19) {
      $fi$idx$val$i = $17;$fi$idx1$val$i = 0;
     } else {
      $20 = HEAP32[$fi>>2]|0;
      $c$02$i = 0;
      while(1) {
       $21 = $c$02$i << 1;
       $22 = (($20) + ($21<<2)|0);
       $23 = HEAP32[$22>>2]|0;
       $24 = (($20) + ($c$02$i<<2)|0);
       HEAP32[$24>>2] = $23;
       $25 = (($c$02$i) + 1)|0;
       $26 = HEAP32[$0>>2]|0;
       $27 = ($25>>>0)<($26>>>0);
       if ($27) {
        $c$02$i = $25;
       } else {
        $$lcssa = $26;
        break;
       }
      }
      $fi$idx$val$pre$i = HEAP32[$5>>2]|0;
      $fi$idx$val$i = $fi$idx$val$pre$i;$fi$idx1$val$i = $$lcssa;
     }
     $28 = Math_imul($fi$idx$val$i, $fi$idx1$val$i)|0;
     $29 = ((($fi)) + 8|0);
     HEAP32[$29>>2] = $28;
    }
   }
   $30 = ((($fi)) + 8|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = ($31|0)==($7|0);
   if ($32) {
    $$pre2 = HEAP32[$0>>2]|0;
    $$pre3 = HEAP32[$2>>2]|0;
    $34 = $$pre2;$35 = $$pre3;
    break;
   } else {
    return;
   }
  } else {
   $34 = $1;$35 = $3;
  }
 } while(0);
 $33 = ($34>>>0)<($35>>>0);
 if (!($33)) {
  return;
 }
 $36 = HEAP32[$fi>>2]|0;
 $37 = (($36) + ($34<<2)|0);
 HEAP32[$37>>2] = $pos;
 $38 = HEAP32[$0>>2]|0;
 $39 = (($38) + 1)|0;
 HEAP32[$0>>2] = $39;
 $fi$idx = ((($fi)) + 4|0);
 $fi$idx$val = HEAP32[$fi$idx>>2]|0;
 $40 = Math_imul($fi$idx$val, $39)|0;
 $41 = ((($fi)) + 8|0);
 HEAP32[$41>>2] = $40;
 return;
}
function _INT123_fi_resize($fi,$newsize) {
 $fi = $fi|0;
 $newsize = $newsize|0;
 var $$0 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $c$02$i = 0, $fi$idx = 0, $fi$idx$val = 0, $fi$idx$val$i = 0, $fi$idx$val$pre$i = 0, $fi$idx3$val = 0, $or$cond = 0, $or$cond2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($fi)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==($newsize|0);
 if ($2) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ($newsize|0)!=(0);
 $4 = ($1>>>0)>($newsize>>>0);
 $or$cond2 = $3 & $4;
 if ($or$cond2) {
  $5 = ((($fi)) + 16|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6>>>0)>($newsize>>>0);
  if ($7) {
   $8 = ((($fi)) + 4|0);
   $9 = ((($fi)) + 8|0);
   $11 = $6;
   while(1) {
    $10 = ($11>>>0)<(2);
    if ($10) {
     $13 = $11;
    } else {
     $14 = HEAP32[$8>>2]|0;
     $15 = $14 << 1;
     HEAP32[$8>>2] = $15;
     $16 = $11 >>> 1;
     HEAP32[$5>>2] = $16;
     $17 = ($16|0)==(0);
     if ($17) {
      $27 = 0;$fi$idx$val$i = $15;
     } else {
      $18 = HEAP32[$fi>>2]|0;
      $c$02$i = 0;
      while(1) {
       $19 = $c$02$i << 1;
       $20 = (($18) + ($19<<2)|0);
       $21 = HEAP32[$20>>2]|0;
       $22 = (($18) + ($c$02$i<<2)|0);
       HEAP32[$22>>2] = $21;
       $23 = (($c$02$i) + 1)|0;
       $24 = HEAP32[$5>>2]|0;
       $25 = ($23>>>0)<($24>>>0);
       if ($25) {
        $c$02$i = $23;
       } else {
        $$lcssa = $24;
        break;
       }
      }
      $fi$idx$val$pre$i = HEAP32[$8>>2]|0;
      $27 = $$lcssa;$fi$idx$val$i = $fi$idx$val$pre$i;
     }
     $26 = Math_imul($fi$idx$val$i, $27)|0;
     HEAP32[$9>>2] = $26;
     $13 = $27;
    }
    $12 = ($13>>>0)>($newsize>>>0);
    if ($12) {
     $11 = $13;
    } else {
     break;
    }
   }
  }
 }
 $28 = HEAP32[$fi>>2]|0;
 $29 = $newsize << 2;
 $30 = (_INT123_safe_realloc($28,$29)|0);
 $31 = ($newsize|0)==(0);
 $32 = ($30|0)!=(0|0);
 $or$cond = $31 | $32;
 if (!($or$cond)) {
  $38 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer>>2] = 78;
  (_fprintf($38,513751,$vararg_buffer)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$fi>>2] = $30;
 HEAP32[$0>>2] = $newsize;
 $33 = ((($fi)) + 16|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = ($34>>>0)>($newsize>>>0);
 if ($35) {
  HEAP32[$33>>2] = $newsize;
  $fi$idx3$val = $newsize;
 } else {
  $fi$idx3$val = $34;
 }
 $fi$idx = ((($fi)) + 4|0);
 $fi$idx$val = HEAP32[$fi$idx>>2]|0;
 $36 = Math_imul($fi$idx3$val, $fi$idx$val)|0;
 $37 = ((($fi)) + 8|0);
 HEAP32[$37>>2] = $36;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _INT123_compute_bpf($fr) {
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0.0, $bpf$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 19688|0);
 $1 = HEAP32[$0>>2]|0;
 switch ($1|0) {
 case 1:  {
  $2 = ((($fr)) + 19700|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($fr)) + 19672|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ((35332 + (($5*192)|0)|0) + ($3<<2)|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (+($7|0));
  $9 = $8 * 48000.0;
  $10 = ((($fr)) + 19704|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (35716 + ($11<<2)|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = $13 << $5;
  $15 = (+($14|0));
  $16 = $9 / $15;
  $bpf$0 = $16;
  return (+$bpf$0);
  break;
 }
 case 3: case 2:  {
  $17 = ((($fr)) + 19700|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = (($1) + -1)|0;
  $20 = ((($fr)) + 19672|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = (((35332 + (($21*192)|0)|0) + ($19<<6)|0) + ($18<<2)|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = (+($23|0));
  $25 = $24 * 144000.0;
  $26 = ((($fr)) + 19704|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = (35716 + ($27<<2)|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = $29 << $21;
  $31 = (+($30|0));
  $32 = $25 / $31;
  $bpf$0 = $32;
  return (+$bpf$0);
  break;
 }
 default: {
  $bpf$0 = 1.0;
  return (+$bpf$0);
 }
 }
 return +(0.0);
}
function _INT123_frame_init_par($fr,$mp) {
 $fr = $fr|0;
 $mp = $mp|0;
 var $$phi$trans$insert = 0, $$phi$trans$insert1 = 0, $$pre = 0, $$pre$phi5Z2D = 0, $$pre2 = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i = 0, $i$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27892|0);
 HEAP32[$0>>2] = 1;
 $1 = ((($fr)) + 27856|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($fr)) + 27872|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($fr)) + 27864|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($fr)) + 27868|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($fr)) + 9264|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($fr)) + 9268|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($fr)) + 9276|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($fr)) + 9280|0);
 HEAP32[$8>>2] = 0;
 $9 = ((($fr)) + 9588|0);
 HEAP32[$9>>2] = 0;
 $10 = ((($fr)) + 28588|0);
 HEAP32[$10>>2] = 0;
 $11 = ((($fr)) + 19772|0);
 HEAP32[$11>>2] = 0;
 $12 = (_INT123_defdec()|0);
 $13 = ((($fr)) + 19620|0);
 HEAP32[$13>>2] = $12;
 $14 = (_INT123_decclass($12)|0);
 $15 = ((($fr)) + 19624|0);
 HEAP32[$15>>2] = $14;
 $16 = ((($fr)) + 19352|0);
 HEAP32[$16>>2] = 16384;
 $17 = ((($fr)) + 19356|0);
 HEAP32[$17>>2] = 16384;
 $18 = ((($fr)) + 19360|0);
 HEAP32[$18>>2] = 32768;
 $19 = ((($fr)) + 9288|0);
 HEAP32[$19>>2] = 0;
 $i$01$i = 0;
 while(1) {
  $20 = (((($fr)) + 9420|0) + ($i$01$i<<2)|0);
  HEAPF32[$20>>2] = 1.0;
  $21 = (((($fr)) + 9292|0) + ($i$01$i<<2)|0);
  HEAPF32[$21>>2] = 1.0;
  $22 = (($i$01$i) + 1)|0;
  $exitcond$i = ($22|0)==(32);
  if ($exitcond$i) {
   break;
  } else {
   $i$01$i = $22;
  }
 }
 $23 = ((($fr)) + 28576|0);
 _INT123_init_icy($23);
 _INT123_init_id3($fr);
 $24 = ((($fr)) + 27876|0);
 _INT123_invalidate_format($24);
 $25 = ((($fr)) + 27988|0);
 $26 = ((($fr)) + 27972|0);
 HEAP32[$26>>2] = 0;
 $27 = ((($fr)) + 28608|0);
 HEAP32[$27>>2] = 0;
 $28 = ((($fr)) + 28612|0);
 HEAP32[$28>>2] = 0;
 $29 = ((($fr)) + 28380|0);
 ;HEAP32[$25>>2]=0|0;HEAP32[$25+4>>2]=0|0;HEAP32[$25+8>>2]=0|0;HEAP32[$25+12>>2]=0|0;HEAP32[$25+16>>2]=0|0;
 HEAP32[$29>>2] = 1;
 $30 = ((($fr)) + 28376|0);
 HEAP32[$30>>2] = 0;
 $31 = ($mp|0)==(0|0);
 $32 = ((($fr)) + 28064|0);
 if ($31) {
  $33 = ((($fr)) + 28344|0);
  HEAPF64[$33>>3] = 1.0;
  $34 = ((($fr)) + 28068|0);
  HEAP32[$34>>2] = 32832;
  $35 = ((($fr)) + 28072|0);
  HEAP32[$32>>2] = 0;
  $36 = ((($fr)) + 28336|0);
  HEAP32[$36>>2] = 0;
  $37 = ((($fr)) + 28352|0);
  ;HEAP32[$35>>2]=0|0;HEAP32[$35+4>>2]=0|0;HEAP32[$35+8>>2]=0|0;HEAP32[$35+12>>2]=0|0;HEAP32[$35+16>>2]=0|0;HEAP32[$35+20>>2]=0|0;
  HEAP32[$37>>2] = 1024;
  $38 = ((($fr)) + 28356|0);
  HEAP32[$38>>2] = 1000;
  $39 = ((($fr)) + 28360|0);
  HEAP32[$39>>2] = 4;
  (_mpg123_fmt_all($32)|0);
  $40 = ((($fr)) + 28364|0);
  HEAP32[$40>>2] = 5;
  $41 = ((($fr)) + 28368|0);
  HEAP32[$41>>2] = 4096;
  $$pre$phi5Z2D = $38;$43 = 5;$44 = 4096;
 } else {
  _memcpy(($32|0),($mp|0),312)|0;
  $$phi$trans$insert = ((($fr)) + 28364|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $$phi$trans$insert1 = ((($fr)) + 28368|0);
  $$pre2 = HEAP32[$$phi$trans$insert1>>2]|0;
  $$pre4 = ((($fr)) + 28356|0);
  $$pre$phi5Z2D = $$pre4;$43 = $$pre;$44 = $$pre2;
 }
 $42 = ((($fr)) + 28020|0);
 _bc_prepare($42,$43,$44);
 $45 = ((($fr)) + 19680|0);
 HEAP32[$45>>2] = 0;
 _frame_fixed_reset($fr);
 $46 = ((($fr)) + 19636|0);
 HEAP32[$46>>2] = 0;
 $47 = ((($fr)) + 19644|0);
 HEAP32[$47>>2] = 0;
 $48 = ((($fr)) + 19648|0);
 HEAP32[$48>>2] = 0;
 $49 = ((($fr)) + 27832|0);
 _INT123_fi_init($49);
 $50 = HEAP32[$$pre$phi5Z2D>>2]|0;
 $51 = ($50|0)>(-1);
 if ($51) {
  $52 = ((($fr)) + 27852|0);
  HEAP32[$52>>2] = 0;
  (_INT123_fi_resize($49,$50)|0);
  return;
 }
 $53 = (0 - ($50))|0;
 $54 = ((($fr)) + 27852|0);
 HEAP32[$54>>2] = $53;
 $55 = ((($fr)) + 27844|0);
 $56 = HEAP32[$55>>2]|0;
 $57 = ($56>>>0)<($53>>>0);
 if (!($57)) {
  return;
 }
 (_INT123_fi_resize($49,$53)|0);
 return;
}
function _INT123_defdec() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1;
}
function _INT123_decclass($type) {
 $type = $type|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($type|0)==(7);
 $1 = $type | 2;
 $2 = ($1|0)==(11);
 $3 = $0 | $2;
 $4 = ($1|0)==(14);
 $5 = $4 | $3;
 $6 = $5 ? 2 : 1;
 return ($6|0);
}
function _INT123_init_icy($icy) {
 $icy = $icy|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$icy>>2] = 0;
 return;
}
function _INT123_init_id3($fr) {
 $fr = $fr|0;
 var $0 = 0, $1 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 28524|0);
 HEAP8[$0>>0] = 0;
 $1 = ((($fr)) + 28528|0);
 dest=$1; stop=dest+48|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 return;
}
function _INT123_invalidate_format($af) {
 $af = $af|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$af>>2] = 0;
 $0 = ((($af)) + 12|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($af)) + 8|0);
 HEAP32[$1>>2] = 0;
 return;
}
function _mpg123_fmt_all($mp) {
 $mp = $mp|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond6 = 0, $exitcond6$1 = 0, $rate$04 = 0, $rate$04$1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ($mp|0)==(0|0);
 if ($0) {
  $$0 = 25;
  return ($$0|0);
 }
 $1 = ((($mp)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2 & 32;
 $4 = ($3|0)==(0);
 if ($4) {
  $5 = HEAP32[$mp>>2]|0;
  $6 = ($5|0)>(2);
  if ($6) {
   $7 = HEAP32[37544]|0;
   (_fwrite(513796,28,1,$7)|0);
   $rate$04 = 0;
  } else {
   $rate$04 = 0;
  }
 } else {
  $rate$04 = 0;
 }
 while(1) {
  $8 = (((($mp)) + 32|0) + (($rate$04*12)|0)|0);
  HEAP8[$8>>0] = 1;
  $9 = (((((($mp)) + 32|0) + (($rate$04*12)|0)|0)) + 1|0);
  HEAP8[$9>>0] = 1;
  $10 = (((((($mp)) + 32|0) + (($rate$04*12)|0)|0)) + 2|0);
  HEAP8[$10>>0] = 1;
  $11 = (((((($mp)) + 32|0) + (($rate$04*12)|0)|0)) + 3|0);
  HEAP8[$11>>0] = 1;
  $12 = (((((($mp)) + 32|0) + (($rate$04*12)|0)|0)) + 4|0);
  HEAP8[$12>>0] = 1;
  $13 = (((((($mp)) + 32|0) + (($rate$04*12)|0)|0)) + 5|0);
  HEAP8[$13>>0] = 1;
  $14 = (((((($mp)) + 32|0) + (($rate$04*12)|0)|0)) + 6|0);
  HEAP8[$14>>0] = 1;
  $15 = (((((($mp)) + 32|0) + (($rate$04*12)|0)|0)) + 7|0);
  HEAP8[$15>>0] = 0;
  $16 = (((((($mp)) + 32|0) + (($rate$04*12)|0)|0)) + 8|0);
  HEAP8[$16>>0] = 1;
  $17 = (((((($mp)) + 32|0) + (($rate$04*12)|0)|0)) + 9|0);
  HEAP8[$17>>0] = 1;
  $18 = (((((($mp)) + 32|0) + (($rate$04*12)|0)|0)) + 10|0);
  HEAP8[$18>>0] = 1;
  $19 = (((((($mp)) + 32|0) + (($rate$04*12)|0)|0)) + 11|0);
  HEAP8[$19>>0] = 1;
  $20 = (($rate$04) + 1)|0;
  $exitcond6 = ($20|0)==(10);
  if ($exitcond6) {
   $rate$04$1 = 0;
   break;
  } else {
   $rate$04 = $20;
  }
 }
 while(1) {
  $21 = (((($mp)) + 152|0) + (($rate$04$1*12)|0)|0);
  HEAP8[$21>>0] = 1;
  $22 = (((((($mp)) + 152|0) + (($rate$04$1*12)|0)|0)) + 1|0);
  HEAP8[$22>>0] = 1;
  $23 = (((((($mp)) + 152|0) + (($rate$04$1*12)|0)|0)) + 2|0);
  HEAP8[$23>>0] = 1;
  $24 = (((((($mp)) + 152|0) + (($rate$04$1*12)|0)|0)) + 3|0);
  HEAP8[$24>>0] = 1;
  $25 = (((((($mp)) + 152|0) + (($rate$04$1*12)|0)|0)) + 4|0);
  HEAP8[$25>>0] = 1;
  $26 = (((((($mp)) + 152|0) + (($rate$04$1*12)|0)|0)) + 5|0);
  HEAP8[$26>>0] = 1;
  $27 = (((((($mp)) + 152|0) + (($rate$04$1*12)|0)|0)) + 6|0);
  HEAP8[$27>>0] = 1;
  $28 = (((((($mp)) + 152|0) + (($rate$04$1*12)|0)|0)) + 7|0);
  HEAP8[$28>>0] = 0;
  $29 = (((((($mp)) + 152|0) + (($rate$04$1*12)|0)|0)) + 8|0);
  HEAP8[$29>>0] = 1;
  $30 = (((((($mp)) + 152|0) + (($rate$04$1*12)|0)|0)) + 9|0);
  HEAP8[$30>>0] = 1;
  $31 = (((((($mp)) + 152|0) + (($rate$04$1*12)|0)|0)) + 10|0);
  HEAP8[$31>>0] = 1;
  $32 = (((((($mp)) + 152|0) + (($rate$04$1*12)|0)|0)) + 11|0);
  HEAP8[$32>>0] = 1;
  $33 = (($rate$04$1) + 1)|0;
  $exitcond6$1 = ($33|0)==(10);
  if ($exitcond6$1) {
   $$0 = 0;
   break;
  } else {
   $rate$04$1 = $33;
  }
 }
 return ($$0|0);
}
function _bc_prepare($bc,$pool_size,$bufblock) {
 $bc = $bc|0;
 $pool_size = $pool_size|0;
 $bufblock = $bufblock|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($bc)) + 28|0);
 HEAP32[$0>>2] = $pool_size;
 $1 = ((($bc)) + 24|0);
 HEAP32[$1>>2] = $bufblock;
 $2 = ((($bc)) + 36|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($bc)) + 32|0);
 HEAP32[$3>>2] = 0;
 ;HEAP32[$bc>>2]=0|0;HEAP32[$bc+4>>2]=0|0;HEAP32[$bc+8>>2]=0|0;HEAP32[$bc+12>>2]=0|0;HEAP32[$bc+16>>2]=0|0;HEAP32[$bc+20>>2]=0|0;
 return;
}
function _frame_fixed_reset($fr) {
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 28576|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  _free($1);
 }
 HEAP32[$0>>2] = 0;
 $3 = ((($fr)) + 28580|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($fr)) + 28584|0);
 HEAP32[$4>>2] = 0;
 _INT123_open_bad($fr);
 $5 = ((($fr)) + 27900|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($fr)) + 27904|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($fr)) + 28392|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($fr)) + 27896|0);
 HEAP32[$8>>2] = 0;
 $9 = ((($fr)) + 19748|0);
 HEAP32[$9>>2] = -1;
 $10 = ((($fr)) + 19752|0);
 HEAP32[$10>>2] = -1;
 $11 = ((($fr)) + 19756|0);
 HEAP32[$11>>2] = -1;
 $12 = ((($fr)) + 19764|0);
 HEAP32[$12>>2] = 1;
 $13 = ((($fr)) + 19768|0);
 HEAP8[$13>>0] = 0;
 $14 = ((($fr)) + 19760|0);
 HEAP32[$14>>2] = 0;
 $15 = ((($fr)) + 28388|0);
 HEAP32[$15>>2] = 0;
 $16 = ((($fr)) + 27820|0);
 HEAP32[$16>>2] = 0;
 $17 = ((($fr)) + 27824|0);
 HEAP32[$17>>2] = 0;
 $18 = ((($fr)) + 19744|0);
 HEAP32[$18>>2] = 0;
 $19 = ((($fr)) + 27828|0);
 HEAP32[$19>>2] = 0;
 $20 = ((($fr)) + 19840|0);
 HEAP32[$20>>2] = 0;
 $21 = ((($fr)) + 19844|0);
 HEAP32[$21>>2] = -1;
 $22 = ((($fr)) + 19736|0);
 HEAP32[$22>>2] = 0;
 $23 = ((($fr)) + 19856|0);
 HEAP32[$23>>2] = 0;
 $24 = ((($fr)) + 19848|0);
 HEAPF64[$24>>3] = 0.0;
 $25 = ((($fr)) + 19740|0);
 HEAP32[$25>>2] = 0;
 $26 = ((($fr)) + 19808|0);
 HEAPF64[$26>>3] = -1.0;
 $27 = ((($fr)) + 19816|0);
 HEAP32[$27>>2] = -1;
 $28 = ((($fr)) + 19820|0);
 HEAP32[$28>>2] = -1;
 $29 = ((($fr)) + 19824|0);
 $30 = ((($fr)) + 19860|0);
 HEAP32[$30>>2] = 0;
 $31 = ((($fr)) + 27908|0);
 HEAP32[$31>>2] = 0;
 $32 = ((($fr)) + 28360|0);
 ;HEAP32[$29>>2]=0|0;HEAP32[$29+4>>2]=0|0;HEAP32[$29+8>>2]=0|0;HEAP32[$29+12>>2]=0|0;
 $33 = HEAP32[$32>>2]|0;
 $34 = (0 - ($33))|0;
 $35 = ((($fr)) + 27916|0);
 HEAP32[$35>>2] = $34;
 $36 = ((($fr)) + 27912|0);
 HEAP32[$36>>2] = -1;
 HEAP32[$fr>>2] = 1;
 $37 = ((($fr)) + 4|0);
 HEAP32[$37>>2] = 0;
 $38 = ((($fr)) + 27920|0);
 HEAP32[$38>>2] = -1;
 $39 = ((($fr)) + 27924|0);
 $40 = ((($fr)) + 9272|0);
 ;HEAP32[$39>>2]=0|0;HEAP32[$39+4>>2]=0|0;HEAP32[$39+8>>2]=0|0;HEAP32[$39+12>>2]=0|0;HEAP32[$39+16>>2]=0|0;HEAP32[$39+20>>2]=0|0;HEAP32[$39+24>>2]=0|0;
 HEAP32[$40>>2] = 1;
 _INT123_reset_id3($fr);
 $41 = ((($fr)) + 28576|0);
 _INT123_reset_icy($41);
 HEAP32[$3>>2] = 0;
 HEAP32[$4>>2] = 0;
 $42 = ((($fr)) + 9584|0);
 HEAP32[$42>>2] = 0;
 $43 = ((($fr)) + 19696|0);
 HEAP32[$43>>2] = 0;
 $44 = ((($fr)) + 19780|0);
 HEAP32[$44>>2] = -1;
 return;
}
function _INT123_open_bad($mh) {
 $mh = $mh|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($mh)) + 28576|0);
 _INT123_clear_icy($0);
 $1 = ((($mh)) + 27956|0);
 HEAP32[$1>>2] = 36864;
 $2 = ((($mh)) + 27976|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($mh)) + 28020|0);
 ;HEAP32[$3>>2]=0|0;HEAP32[$3+4>>2]=0|0;HEAP32[$3+8>>2]=0|0;HEAP32[$3+12>>2]=0|0;HEAP32[$3+16>>2]=0|0;HEAP32[$3+20>>2]=0|0;
 $4 = ((($mh)) + 27960|0);
 HEAP32[$4>>2] = -1;
 return;
}
function _INT123_clear_icy($icy) {
 $icy = $icy|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$icy>>2]|0;
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  _free($0);
 }
 HEAP32[$icy>>2] = 0;
 return;
}
function _bad_init($mh) {
 $mh = $mh|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($mh)) + 28376|0);
 HEAP32[$0>>2] = 24;
 return -1;
}
function _bad_close($mh) {
 $mh = $mh|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _bad_fullread($mh,$data,$count) {
 $mh = $mh|0;
 $data = $data|0;
 $count = $count|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($mh)) + 28376|0);
 HEAP32[$0>>2] = 24;
 return -1;
}
function _bad_head_read($mh,$newhead) {
 $mh = $mh|0;
 $newhead = $newhead|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($mh)) + 28376|0);
 HEAP32[$0>>2] = 24;
 return -1;
}
function _bad_head_shift($mh,$head) {
 $mh = $mh|0;
 $head = $head|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($mh)) + 28376|0);
 HEAP32[$0>>2] = 24;
 return -1;
}
function _bad_skip_bytes($mh,$len) {
 $mh = $mh|0;
 $len = $len|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($mh)) + 28376|0);
 HEAP32[$0>>2] = 24;
 return -1;
}
function _bad_read_frame_body($mh,$data,$size) {
 $mh = $mh|0;
 $data = $data|0;
 $size = $size|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($mh)) + 28376|0);
 HEAP32[$0>>2] = 24;
 return -1;
}
function _bad_back_bytes($mh,$bytes) {
 $mh = $mh|0;
 $bytes = $bytes|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($mh)) + 28376|0);
 HEAP32[$0>>2] = 24;
 return -1;
}
function _bad_seek_frame($mh,$num) {
 $mh = $mh|0;
 $num = $num|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($mh)) + 28376|0);
 HEAP32[$0>>2] = 24;
 return -1;
}
function _bad_tell($mh) {
 $mh = $mh|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($mh)) + 28376|0);
 HEAP32[$0>>2] = 24;
 return -1;
}
function _bad_rewind($mh) {
 $mh = $mh|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _INT123_reset_id3($fr) {
 $fr = $fr|0;
 var $0 = 0, $1 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 _INT123_exit_id3($fr);
 $0 = ((($fr)) + 28524|0);
 HEAP8[$0>>0] = 0;
 $1 = ((($fr)) + 28528|0);
 dest=$1; stop=dest+48|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 return;
}
function _INT123_exit_id3($fr) {
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01$i = 0, $i$01$i1 = 0, $i$01$i4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 28552|0);
 $1 = ((($fr)) + 28556|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0);
 if (!($3)) {
  $i$01$i = 0;
  while(1) {
   $4 = HEAP32[$0>>2]|0;
   $5 = (((($4) + ($i$01$i<<5)|0)) + 20|0);
   _mpg123_free_string($5);
   $6 = (((($4) + ($i$01$i<<5)|0)) + 8|0);
   _mpg123_free_string($6);
   $7 = (($i$01$i) + 1)|0;
   $8 = HEAP32[$1>>2]|0;
   $9 = ($7>>>0)<($8>>>0);
   if ($9) {
    $i$01$i = $7;
   } else {
    break;
   }
  }
 }
 $10 = HEAP32[$0>>2]|0;
 _free($10);
 HEAP32[$0>>2] = 0;
 HEAP32[$1>>2] = 0;
 $11 = ((($fr)) + 28568|0);
 $12 = ((($fr)) + 28572|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13|0)==(0);
 if (!($14)) {
  $i$01$i1 = 0;
  while(1) {
   $15 = HEAP32[$11>>2]|0;
   $16 = (((($15) + ($i$01$i1<<5)|0)) + 20|0);
   _mpg123_free_string($16);
   $17 = (((($15) + ($i$01$i1<<5)|0)) + 8|0);
   _mpg123_free_string($17);
   $18 = (($i$01$i1) + 1)|0;
   $19 = HEAP32[$12>>2]|0;
   $20 = ($18>>>0)<($19>>>0);
   if ($20) {
    $i$01$i1 = $18;
   } else {
    break;
   }
  }
 }
 $21 = HEAP32[$11>>2]|0;
 _free($21);
 HEAP32[$11>>2] = 0;
 HEAP32[$12>>2] = 0;
 $22 = ((($fr)) + 28560|0);
 $23 = ((($fr)) + 28564|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ($24|0)==(0);
 if ($25) {
  $32 = HEAP32[$22>>2]|0;
  _free($32);
  HEAP32[$22>>2] = 0;
  HEAP32[$23>>2] = 0;
  return;
 } else {
  $i$01$i4 = 0;
 }
 while(1) {
  $26 = HEAP32[$22>>2]|0;
  $27 = (((($26) + ($i$01$i4<<5)|0)) + 20|0);
  _mpg123_free_string($27);
  $28 = (((($26) + ($i$01$i4<<5)|0)) + 8|0);
  _mpg123_free_string($28);
  $29 = (($i$01$i4) + 1)|0;
  $30 = HEAP32[$23>>2]|0;
  $31 = ($29>>>0)<($30>>>0);
  if ($31) {
   $i$01$i4 = $29;
  } else {
   break;
  }
 }
 $32 = HEAP32[$22>>2]|0;
 _free($32);
 HEAP32[$22>>2] = 0;
 HEAP32[$23>>2] = 0;
 return;
}
function _INT123_reset_icy($icy) {
 $icy = $icy|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$icy>>2]|0;
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  _free($0);
 }
 HEAP32[$icy>>2] = 0;
 return;
}
function _INT123_fi_init($fi) {
 $fi = $fi|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$fi>>2] = 0;
 $0 = ((($fi)) + 4|0);
 HEAP32[$0>>2] = 1;
 $1 = ((($fi)) + 8|0);
 ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0;HEAP32[$1+12>>2]=0|0;
 return;
}
function _INT123_frame_index_setup($fr) {
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ret$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 28356|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(-1);
 if ($2) {
  $3 = ((($fr)) + 27832|0);
  $4 = ((($fr)) + 27852|0);
  HEAP32[$4>>2] = 0;
  $5 = (_INT123_fi_resize($3,$1)|0);
  $ret$0 = $5;
  return ($ret$0|0);
 }
 $6 = (0 - ($1))|0;
 $7 = ((($fr)) + 27852|0);
 HEAP32[$7>>2] = $6;
 $8 = ((($fr)) + 27844|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9>>>0)<($6>>>0);
 if (!($10)) {
  $ret$0 = 0;
  return ($ret$0|0);
 }
 $11 = ((($fr)) + 27832|0);
 $12 = (_INT123_fi_resize($11,$6)|0);
 $ret$0 = $12;
 return ($ret$0|0);
}
function _INT123_frame_outbuffer($fr) {
 $fr = $fr|0;
 var $$0 = 0, $$0$i = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($fr)) + 27896|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27892|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0);
 if ($4) {
  $5 = ((($fr)) + 27868|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6>>>0)<($1>>>0);
  if ($7) {
   $8 = ((($fr)) + 28376|0);
   HEAP32[$8>>2] = 6;
   $9 = ((($fr)) + 28068|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = $10 & 32;
   $12 = ($11|0)==(0);
   if (!($12)) {
    $$0 = -1;
    STACKTOP = sp;return ($$0|0);
   }
   $13 = HEAP32[37544]|0;
   HEAP32[$vararg_buffer>>2] = 184;
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
   HEAP32[$vararg_ptr1>>2] = $6;
   $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
   HEAP32[$vararg_ptr2>>2] = $1;
   (_fprintf($13,513825,$vararg_buffer)|0);
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $14 = ((($fr)) + 27872|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)==(0|0);
 $$pre = ((($fr)) + 27868|0);
 do {
  if ($16) {
   label = 9;
  } else {
   $17 = HEAP32[$$pre>>2]|0;
   $18 = ($17|0)==($1|0);
   if ($18) {
    HEAP32[$$pre>>2] = $1;
    $19 = ((($fr)) + 27856|0);
    HEAP32[$19>>2] = 0;
    $26 = $15;$32 = $19;
    break;
   } else {
    _free($15);
    HEAP32[$14>>2] = 0;
    label = 9;
    break;
   }
  }
 } while(0);
 if ((label|0) == 9) {
  HEAP32[$$pre>>2] = $1;
  $20 = ((($fr)) + 27856|0);
  HEAP32[$20>>2] = 0;
  $21 = (($1) + 15)|0;
  $22 = (_malloc($21)|0);
  HEAP32[$14>>2] = $22;
  $23 = ($22|0)==(0|0);
  if ($23) {
   $24 = ((($fr)) + 28376|0);
   HEAP32[$24>>2] = 7;
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  } else {
   $26 = $22;$32 = $20;
  }
 }
 $25 = $26;
 $27 = $25 & 15;
 $28 = ($27|0)==(0);
 $29 = ((($26)) + 16|0);
 $30 = (0 - ($27))|0;
 $31 = (($29) + ($30)|0);
 $$0$i = $28 ? $26 : $31;
 HEAP32[$32>>2] = $$0$i;
 HEAP32[$2>>2] = 1;
 $33 = ((($fr)) + 27864|0);
 HEAP32[$33>>2] = 0;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _INT123_frame_buffers($fr) {
 $fr = $fr|0;
 var $$0$i = 0, $$0$i6 = 0, $$2 = 0, $$buffssize$0 = 0, $$in = 0, $$pre = 0, $$pre$phi14Z2D = 0, $$pre$phiZ2D = 0, $$pre13 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0, $9 = 0, $buffssize$0 = 0, $fr$idx$val = 0, $fr$idx5$val = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 19620|0);
 $1 = HEAP32[$0>>2]|0;
 switch ($1|0) {
 case 10:  {
  $buffssize$0 = 17408;
  break;
 }
 case 6: case 5:  {
  label = 3;
  break;
 }
 default: {
  $2 = ($1|0)==(8);
  if ($2) {
   label = 3;
  } else {
   $buffssize$0 = 0;
  }
 }
 }
 if ((label|0) == 3) {
  $buffssize$0 = 4352;
 }
 $3 = ($buffssize$0>>>0)<(4352);
 $$buffssize$0 = $3 ? 4352 : $buffssize$0;
 $4 = $$buffssize$0 | 15;
 $5 = ((($fr)) + 9264|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0|0);
 if ($7) {
  label = 7;
 } else {
  $8 = ((($fr)) + 9268|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)==($4|0);
  if ($10) {
   $$pre$phiZ2D = $8;$14 = $6;
  } else {
   _free($6);
   HEAP32[$5>>2] = 0;
   label = 7;
  }
 }
 do {
  if ((label|0) == 7) {
   $11 = (_malloc($4)|0);
   HEAP32[$5>>2] = $11;
   $12 = ($11|0)==(0|0);
   if ($12) {
    $$2 = -1;
    return ($$2|0);
   } else {
    $$pre = ((($fr)) + 9268|0);
    $$pre$phiZ2D = $$pre;$14 = $11;
    break;
   }
  }
 } while(0);
 HEAP32[$$pre$phiZ2D>>2] = $4;
 $13 = $14;
 $15 = $13 & 15;
 $16 = ($15|0)==(0);
 $17 = ((($14)) + 16|0);
 $18 = (0 - ($15))|0;
 $19 = (($17) + ($18)|0);
 $$0$i = $16 ? $14 : $19;
 $20 = ((($fr)) + 9232|0);
 HEAP32[$20>>2] = $$0$i;
 $21 = ((($$0$i)) + 544|0);
 $22 = ((($fr)) + 9236|0);
 HEAP32[$22>>2] = $21;
 $23 = ((($$0$i)) + 1088|0);
 $24 = ((($fr)) + 9240|0);
 HEAP32[$24>>2] = $23;
 $25 = ((($$0$i)) + 1632|0);
 $26 = ((($fr)) + 9244|0);
 HEAP32[$26>>2] = $25;
 $27 = ((($fr)) + 9248|0);
 HEAP32[$27>>2] = $$0$i;
 $28 = ((($fr)) + 9252|0);
 HEAP32[$28>>2] = $23;
 $29 = ((($$0$i)) + 2176|0);
 $30 = ((($fr)) + 9256|0);
 HEAP32[$30>>2] = $29;
 $31 = ((($$0$i)) + 3264|0);
 $32 = ((($fr)) + 9260|0);
 HEAP32[$32>>2] = $31;
 $33 = ((($fr)) + 9276|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = ($34|0)==(0|0);
 if ($35) {
  label = 12;
 } else {
  $36 = ((($fr)) + 9280|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = ($37|0)==(2176);
  if ($38) {
   $$in = $34;$$pre$phi14Z2D = $36;
  } else {
   _free($34);
   HEAP32[$33>>2] = 0;
   label = 12;
  }
 }
 do {
  if ((label|0) == 12) {
   $39 = (_malloc(2176)|0);
   HEAP32[$33>>2] = $39;
   $40 = ($39|0)==(0|0);
   if ($40) {
    $$2 = -1;
    return ($$2|0);
   } else {
    $$pre13 = ((($fr)) + 9280|0);
    $$in = $39;$$pre$phi14Z2D = $$pre13;
    break;
   }
  }
 } while(0);
 $41 = $$in;
 HEAP32[$$pre$phi14Z2D>>2] = 2176;
 $42 = ((($fr)) + 9284|0);
 HEAP32[$42>>2] = $41;
 $43 = ((($fr)) + 28588|0);
 $44 = HEAP32[$43>>2]|0;
 $45 = ($44|0)==(0|0);
 do {
  if ($45) {
   $46 = (_malloc(10559)|0);
   HEAP32[$43>>2] = $46;
   $47 = ($46|0)==(0|0);
   if ($47) {
    $$2 = -1;
    return ($$2|0);
   } else {
    $48 = $46;
    $49 = $48 & 63;
    $50 = ($49|0)==(0);
    $51 = ((($46)) + 64|0);
    $52 = (0 - ($49))|0;
    $53 = (($51) + ($52)|0);
    $$0$i6 = $50 ? $46 : $53;
    $54 = ((($fr)) + 28592|0);
    HEAP32[$54>>2] = $$0$i6;
    $55 = ((($$0$i6)) + 256|0);
    $56 = ((($fr)) + 28596|0);
    HEAP32[$56>>2] = $55;
    $57 = ((($$0$i6)) + 1280|0);
    $58 = ((($fr)) + 28600|0);
    HEAP32[$58>>2] = $57;
    $59 = ((($$0$i6)) + 5888|0);
    $60 = ((($fr)) + 28604|0);
    HEAP32[$60>>2] = $59;
    break;
   }
  }
 } while(0);
 $fr$idx$val = HEAP32[$5>>2]|0;
 $fr$idx5$val = HEAP32[$$pre$phiZ2D>>2]|0;
 _memset(($fr$idx$val|0),0,($fr$idx5$val|0))|0;
 $$2 = 0;
 return ($$2|0);
}
function _INT123_frame_buffers_reset($fr) {
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $fr$idx = 0, $fr$idx$val = 0, $fr$idx1 = 0, $fr$idx1$val = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27864|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($fr)) + 27816|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($fr)) + 23840|0);
 $3 = ((($fr)) + 27808|0);
 HEAP32[$3>>2] = $2;
 $4 = ((($fr)) + 27812|0);
 HEAP32[$4>>2] = $2;
 $5 = ((($fr)) + 19868|0);
 HEAP32[$5>>2] = 0;
 $fr$idx = ((($fr)) + 9264|0);
 $fr$idx$val = HEAP32[$fr$idx>>2]|0;
 $fr$idx1 = ((($fr)) + 9268|0);
 $fr$idx1$val = HEAP32[$fr$idx1>>2]|0;
 _memset(($fr$idx$val|0),0,($fr$idx1$val|0))|0;
 $6 = ((($fr)) + 19872|0);
 _memset(($6|0),0,7936)|0;
 $7 = ((($fr)) + 9548|0);
 ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;HEAP32[$7+12>>2]=0|0;HEAP32[$7+16>>2]=0|0;HEAP32[$7+20>>2]=0|0;HEAP32[$7+24>>2]=0|0;HEAP32[$7+28>>2]=0|0;HEAP16[$7+32>>1]=0|0;
 $8 = ((($fr)) + 8|0);
 _memset(($8|0),0,9224)|0;
 return 0;
}
function _INT123_frame_reset($fr) {
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fr$idx$i = 0, $fr$idx$val$i = 0, $fr$idx1$i = 0, $fr$idx1$val$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27864|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($fr)) + 27816|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($fr)) + 23840|0);
 $3 = ((($fr)) + 27808|0);
 HEAP32[$3>>2] = $2;
 $4 = ((($fr)) + 27812|0);
 HEAP32[$4>>2] = $2;
 $5 = ((($fr)) + 19868|0);
 HEAP32[$5>>2] = 0;
 $fr$idx$i = ((($fr)) + 9264|0);
 $fr$idx$val$i = HEAP32[$fr$idx$i>>2]|0;
 $fr$idx1$i = ((($fr)) + 9268|0);
 $fr$idx1$val$i = HEAP32[$fr$idx1$i>>2]|0;
 _memset(($fr$idx$val$i|0),0,($fr$idx1$val$i|0))|0;
 $6 = ((($fr)) + 19872|0);
 _memset(($6|0),0,7936)|0;
 $7 = ((($fr)) + 9548|0);
 ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;HEAP32[$7+12>>2]=0|0;HEAP32[$7+16>>2]=0|0;HEAP32[$7+20>>2]=0|0;HEAP32[$7+24>>2]=0|0;HEAP32[$7+28>>2]=0|0;HEAP16[$7+32>>1]=0|0;
 $8 = ((($fr)) + 8|0);
 _memset(($8|0),0,9224)|0;
 _frame_fixed_reset($fr);
 $9 = ((($fr)) + 19772|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10|0)==(0|0);
 if ($11) {
  $12 = ((($fr)) + 27832|0);
  _INT123_fi_reset($12);
  return 0;
 }
 _free($10);
 HEAP32[$9>>2] = 0;
 $12 = ((($fr)) + 27832|0);
 _INT123_fi_reset($12);
 return 0;
}
function _INT123_fi_reset($fi) {
 $fi = $fi|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fi)) + 16|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($fi)) + 4|0);
 HEAP32[$1>>2] = 1;
 $2 = ((($fi)) + 8|0);
 HEAP32[$2>>2] = 0;
 return;
}
function _INT123_frame_exit($fr) {
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27872|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  _free($1);
 }
 HEAP32[$0>>2] = 0;
 $3 = ((($fr)) + 9264|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if (!($5)) {
  _free($4);
 }
 HEAP32[$3>>2] = 0;
 $6 = ((($fr)) + 9268|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($fr)) + 9276|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 if (!($9)) {
  _free($8);
 }
 HEAP32[$7>>2] = 0;
 $10 = ((($fr)) + 9280|0);
 HEAP32[$10>>2] = 0;
 $11 = ((($fr)) + 9588|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)==(0|0);
 if (!($13)) {
  _free($12);
 }
 HEAP32[$11>>2] = 0;
 $14 = ((($fr)) + 28588|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)==(0|0);
 if (!($16)) {
  _free($15);
 }
 $17 = ((($fr)) + 19772|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ($18|0)==(0|0);
 if (!($19)) {
  _free($18);
  HEAP32[$17>>2] = 0;
 }
 $20 = ((($fr)) + 27832|0);
 _INT123_fi_exit($20);
 _INT123_exit_id3($fr);
 $21 = ((($fr)) + 28576|0);
 _INT123_clear_icy($21);
 $22 = ((($fr)) + 28612|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = ($23|0)==(0|0);
 if ($24) {
  $27 = ((($fr)) + 28020|0);
  _bc_cleanup($27);
  return;
 }
 $25 = ((($fr)) + 28608|0);
 $26 = HEAP32[$25>>2]|0;
 FUNCTION_TABLE_vi[$23 & 8191]($26);
 HEAP32[$25>>2] = 0;
 $27 = ((($fr)) + 28020|0);
 _bc_cleanup($27);
 return;
}
function _INT123_fi_exit($fi) {
 $fi = $fi|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fi)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  $3 = HEAP32[$fi>>2]|0;
  $4 = ($3|0)==(0|0);
  if (!($4)) {
   _free($3);
  }
 }
 HEAP32[$fi>>2] = 0;
 $5 = ((($fi)) + 4|0);
 HEAP32[$5>>2] = 1;
 $6 = ((($fi)) + 8|0);
 ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;HEAP32[$6+12>>2]=0|0;
 return;
}
function _bc_cleanup($bc) {
 $bc = $bc|0;
 var $$01$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($bc)) + 36|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $$01$i = $1;
  while(1) {
   $3 = ((($$01$i)) + 12|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = HEAP32[$$01$i>>2]|0;
   _free($5);
   _free($$01$i);
   $6 = ($4|0)==(0|0);
   if ($6) {
    break;
   } else {
    $$01$i = $4;
   }
  }
 }
 HEAP32[$0>>2] = 0;
 $7 = ((($bc)) + 32|0);
 HEAP32[$7>>2] = 0;
 return;
}
function _INT123_frame_index_find($fr,$want_frame,$get_frame) {
 $fr = $fr|0;
 $want_frame = $want_frame|0;
 $get_frame = $get_frame|0;
 var $$$$i = 0, $$$$i3 = 0, $$$i = 0, $$$i2 = 0, $$1 = 0, $$pre = 0, $$pre6 = 0, $$pre6$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0.0, $108 = 0.0;
 var $109 = 0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0, $113 = 0, $114 = 0.0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0, $126 = 0;
 var $127 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0, $95 = 0, $96 = 0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $fi$0 = 0, $ret$0$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$get_frame>>2] = 0;
 $0 = ((($fr)) + 27848|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  $75 = ((($fr)) + 28068|0);
  $76 = HEAP32[$75>>2]|0;
  $77 = $76 & 512;
  $78 = ($77|0)==(0);
  if ($78) {
   $126 = ((($fr)) + 27824|0);
   HEAP32[$126>>2] = 0;
   $127 = ((($fr)) + 27820|0);
   HEAP32[$127>>2] = 0;
   $$1 = 0;
   return ($$1|0);
  }
  $79 = ((($fr)) + 19760|0);
  $80 = HEAP32[$79>>2]|0;
  HEAP32[$get_frame>>2] = 0;
  $81 = ((($fr)) + 19772|0);
  $82 = HEAP32[$81>>2]|0;
  $83 = ($82|0)==(0|0);
  if (!($83)) {
   $84 = ((($fr)) + 19840|0);
   $85 = HEAP32[$84>>2]|0;
   $86 = ($85|0)>(0);
   if ($86) {
    $87 = ((($fr)) + 27960|0);
    $88 = HEAP32[$87>>2]|0;
    $89 = ($88|0)>(0);
    if ($89) {
     $90 = (+($want_frame|0));
     $91 = $90 * 100.0;
     $92 = (+($85|0));
     $93 = $91 / $92;
     $94 = (~~(($93)));
     $95 = ($94|0)<(0);
     $$$i2 = $95 ? 0 : $94;
     $96 = ($$$i2|0)>(99);
     $$$$i3 = $96 ? 99 : $$$i2;
     $97 = (+($$$$i3|0));
     $98 = $92 * 0.01;
     $99 = $98 * $97;
     $100 = (~~(($99)));
     HEAP32[$get_frame>>2] = $100;
     $101 = ((($fr)) + 19764|0);
     $102 = HEAP32[$101>>2]|0;
     $103 = $102 & -2;
     HEAP32[$101>>2] = $103;
     $104 = ((($fr)) + 19768|0);
     HEAP8[$104>>0] = 1;
     $105 = (($82) + ($$$$i3)|0);
     $106 = HEAP8[$105>>0]|0;
     $107 = (+($106&255));
     $108 = $107 * 0.00390625;
     $109 = HEAP32[$87>>2]|0;
     $110 = (+($109|0));
     $111 = $108 * $110;
     $112 = (~~(($111)));
     $$1 = $112;
     return ($$1|0);
    }
   }
  }
  $113 = ((($fr)) + 19848|0);
  $114 = +HEAPF64[$113>>3];
  $115 = $114 > 0.0;
  if (!($115)) {
   $$1 = $80;
   return ($$1|0);
  }
  $116 = ((($fr)) + 19764|0);
  $117 = HEAP32[$116>>2]|0;
  $118 = $117 & -2;
  HEAP32[$116>>2] = $118;
  $119 = ((($fr)) + 19768|0);
  HEAP8[$119>>0] = 1;
  HEAP32[$get_frame>>2] = $want_frame;
  $120 = HEAP32[$79>>2]|0;
  $121 = (+($120|0));
  $122 = (+($want_frame|0));
  $123 = $114 * $122;
  $124 = $121 + $123;
  $125 = (~~(($124)));
  $$1 = $125;
  return ($$1|0);
 }
 $3 = ((($fr)) + 27836|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($want_frame|0) / ($4|0))&-1;
 $6 = ($5>>>0)<($1>>>0);
 if ($6) {
  $67 = $4;$fi$0 = $5;
 } else {
  $7 = ((($fr)) + 28068|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = $8 & 512;
  $10 = ($9|0)==(0);
  do {
   if ($10) {
    $$pre6 = $4;$65 = $1;
   } else {
    $11 = (($1) + -1)|0;
    $12 = Math_imul($4, $11)|0;
    $13 = (($want_frame) - ($12))|0;
    $14 = ($13>>>0)>(10);
    if ($14) {
     $15 = ((($fr)) + 19760|0);
     $16 = HEAP32[$15>>2]|0;
     HEAP32[$get_frame>>2] = 0;
     $17 = ((($fr)) + 19772|0);
     $18 = HEAP32[$17>>2]|0;
     $19 = ($18|0)==(0|0);
     if ($19) {
      label = 9;
     } else {
      $20 = ((($fr)) + 19840|0);
      $21 = HEAP32[$20>>2]|0;
      $22 = ($21|0)>(0);
      if ($22) {
       $23 = ((($fr)) + 27960|0);
       $24 = HEAP32[$23>>2]|0;
       $25 = ($24|0)>(0);
       if ($25) {
        $26 = (+($want_frame|0));
        $27 = $26 * 100.0;
        $28 = (+($21|0));
        $29 = $27 / $28;
        $30 = (~~(($29)));
        $31 = ($30|0)<(0);
        $$$i = $31 ? 0 : $30;
        $32 = ($$$i|0)>(99);
        $$$$i = $32 ? 99 : $$$i;
        $33 = (+($$$$i|0));
        $34 = $28 * 0.01;
        $35 = $34 * $33;
        $36 = (~~(($35)));
        HEAP32[$get_frame>>2] = $36;
        $37 = ((($fr)) + 19764|0);
        $38 = HEAP32[$37>>2]|0;
        $39 = $38 & -2;
        HEAP32[$37>>2] = $39;
        $40 = ((($fr)) + 19768|0);
        HEAP8[$40>>0] = 1;
        $41 = (($18) + ($$$$i)|0);
        $42 = HEAP8[$41>>0]|0;
        $43 = (+($42&255));
        $44 = $43 * 0.00390625;
        $45 = HEAP32[$23>>2]|0;
        $46 = (+($45|0));
        $47 = $44 * $46;
        $48 = (~~(($47)));
        $ret$0$i = $48;
       } else {
        label = 9;
       }
      } else {
       label = 9;
      }
     }
     if ((label|0) == 9) {
      $49 = ((($fr)) + 19848|0);
      $50 = +HEAPF64[$49>>3];
      $51 = $50 > 0.0;
      if ($51) {
       $52 = ((($fr)) + 19764|0);
       $53 = HEAP32[$52>>2]|0;
       $54 = $53 & -2;
       HEAP32[$52>>2] = $54;
       $55 = ((($fr)) + 19768|0);
       HEAP8[$55>>0] = 1;
       HEAP32[$get_frame>>2] = $want_frame;
       $56 = HEAP32[$15>>2]|0;
       $57 = (+($56|0));
       $58 = (+($want_frame|0));
       $59 = $50 * $58;
       $60 = $57 + $59;
       $61 = (~~(($60)));
       $ret$0$i = $61;
      } else {
       $ret$0$i = $16;
      }
     }
     $62 = HEAP32[$15>>2]|0;
     $63 = ($ret$0$i|0)>($62|0);
     if ($63) {
      $$1 = $ret$0$i;
      return ($$1|0);
     } else {
      $$pre = HEAP32[$0>>2]|0;
      $$pre6$pre = HEAP32[$3>>2]|0;
      $$pre6 = $$pre6$pre;$65 = $$pre;
      break;
     }
    } else {
     $$pre6 = $4;$65 = $1;
    }
   }
  } while(0);
  $64 = (($65) + -1)|0;
  $67 = $$pre6;$fi$0 = $64;
 }
 $66 = Math_imul($67, $fi$0)|0;
 HEAP32[$get_frame>>2] = $66;
 $68 = ((($fr)) + 27832|0);
 $69 = HEAP32[$68>>2]|0;
 $70 = (($69) + ($fi$0<<2)|0);
 $71 = HEAP32[$70>>2]|0;
 $72 = ((($fr)) + 19764|0);
 $73 = HEAP32[$72>>2]|0;
 $74 = $73 | 1;
 HEAP32[$72>>2] = $74;
 $$1 = $71;
 return ($$1|0);
}
function _INT123_frame_ins2outs($fr,$ins) {
 $fr = $fr|0;
 $ins = $ins|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $outs$0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($fr)) + 19680|0);
 $1 = HEAP32[$0>>2]|0;
 switch ($1|0) {
 case 2: case 1: case 0:  {
  $2 = $ins >> $1;
  $outs$0 = $2;
  break;
 }
 case 3:  {
  $3 = (_INT123_ntom_ins2outs($fr,$ins)|0);
  $outs$0 = $3;
  break;
 }
 default: {
  $4 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer>>2] = 769;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $1;
  (_fprintf($4,513889,$vararg_buffer)|0);
  $outs$0 = 0;
 }
 }
 STACKTOP = sp;return ($outs$0|0);
}
function _INT123_ntom_ins2outs($fr,$ins) {
 $fr = $fr|0;
 $ins = $ins|0;
 var $$02 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ntm$0 = 0, $phitmp = 0, $soff$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 19688|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(1);
 if ($2) {
  $14 = 384;
 } else {
  $3 = ($1|0)==(2);
  if ($3) {
   $14 = 1152;
  } else {
   $4 = ((($fr)) + 19672|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = ($5|0)==(0);
   if ($6) {
    $7 = ((($fr)) + 19676|0);
    $8 = HEAP32[$7>>2]|0;
    $9 = ($8|0)!=(0);
    $phitmp = $9 ? 576 : 1152;
    $14 = $phitmp;
   } else {
    $14 = 576;
   }
  }
 }
 $10 = ($ins|0)<(1);
 if ($10) {
  $24 = 0;
  return ($24|0);
 }
 $11 = ((($fr)) + 19360|0);
 $12 = HEAP32[$11>>2]|0;
 $$02 = $ins;$ntm$0 = 16384;$soff$0 = 0;
 while(1) {
  $13 = ($$02|0)>($14|0);
  $15 = $13 ? $14 : $$02;
  $16 = Math_imul($15, $12)|0;
  $17 = (($16) + ($ntm$0))|0;
  $18 = (($17|0) / 32768)&-1;
  $19 = (($18) + ($soff$0))|0;
  $20 = $18 << 15;
  $21 = (($17) - ($20))|0;
  $22 = (($$02) - ($15))|0;
  $23 = ($22|0)>(0);
  if ($23) {
   $$02 = $22;$ntm$0 = $21;$soff$0 = $19;
  } else {
   $24 = $19;
   break;
  }
 }
 return ($24|0);
}
function _INT123_frame_outs($fr,$num) {
 $fr = $fr|0;
 $num = $num|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $outs$0 = 0, $phitmp = 0, $vararg_buffer = 0;
 var $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($fr)) + 19680|0);
 $1 = HEAP32[$0>>2]|0;
 switch ($1|0) {
 case 2: case 1: case 0:  {
  $2 = ((($fr)) + 19688|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ($3|0)==(1);
  if ($4) {
   $13 = 384;
  } else {
   $5 = ($3|0)==(2);
   if ($5) {
    $13 = 1152;
   } else {
    $6 = ((($fr)) + 19672|0);
    $7 = HEAP32[$6>>2]|0;
    $8 = ($7|0)==(0);
    if ($8) {
     $9 = ((($fr)) + 19676|0);
     $10 = HEAP32[$9>>2]|0;
     $11 = ($10|0)!=(0);
     $phitmp = $11 ? 576 : 1152;
     $13 = $phitmp;
    } else {
     $13 = 576;
    }
   }
  }
  $12 = $13 >>> $1;
  $14 = Math_imul($12, $num)|0;
  $outs$0 = $14;
  STACKTOP = sp;return ($outs$0|0);
  break;
 }
 case 3:  {
  $15 = (_INT123_ntom_frmouts($fr,$num)|0);
  $outs$0 = $15;
  STACKTOP = sp;return ($outs$0|0);
  break;
 }
 default: {
  $16 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer>>2] = 789;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $1;
  (_fprintf($16,513889,$vararg_buffer)|0);
  $outs$0 = 0;
  STACKTOP = sp;return ($outs$0|0);
 }
 }
 return (0)|0;
}
function _INT123_ntom_frmouts($fr,$frame) {
 $fr = $fr|0;
 $frame = $frame|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond14 = 0, $exitcond15 = 0, $exitcond16 = 0, $f$01 = 0, $f$01$us = 0, $f$01$us6 = 0;
 var $ntm$03 = 0, $ntm$03$us = 0, $ntm$03$us4 = 0, $soff$02 = 0, $soff$02$us = 0, $soff$02$us5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($frame|0)>(0);
 if (!($0)) {
  $$0 = 0;
  return ($$0|0);
 }
 $1 = ((($fr)) + 19688|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($fr)) + 19360|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($fr)) + 19676|0);
 switch ($2|0) {
 case 1:  {
  $6 = ($4*384)|0;
  $f$01$us = 0;$ntm$03$us = 16384;$soff$02$us = 0;
  while(1) {
   $7 = (($6) + ($ntm$03$us))|0;
   $8 = (($7|0) / 32768)&-1;
   $9 = (($8) + ($soff$02$us))|0;
   $10 = $8 << 15;
   $11 = (($7) - ($10))|0;
   $12 = (($f$01$us) + 1)|0;
   $exitcond14 = ($12|0)==($frame|0);
   if ($exitcond14) {
    $$0 = $9;
    break;
   } else {
    $f$01$us = $12;$ntm$03$us = $11;$soff$02$us = $9;
   }
  }
  return ($$0|0);
  break;
 }
 case 2:  {
  $15 = ($4*1152)|0;
  $f$01$us6 = 0;$ntm$03$us4 = 16384;$soff$02$us5 = 0;
  while(1) {
   $16 = (($15) + ($ntm$03$us4))|0;
   $17 = (($16|0) / 32768)&-1;
   $18 = (($17) + ($soff$02$us5))|0;
   $19 = $17 << 15;
   $20 = (($16) - ($19))|0;
   $21 = (($f$01$us6) + 1)|0;
   $exitcond15 = ($21|0)==($frame|0);
   if ($exitcond15) {
    $$0 = $18;
    break;
   } else {
    $f$01$us6 = $21;$ntm$03$us4 = $20;$soff$02$us5 = $18;
   }
  }
  return ($$0|0);
  break;
 }
 default: {
  $13 = ((($fr)) + 19672|0);
  $$pre = HEAP32[$13>>2]|0;
  $14 = ($$pre|0)==(0);
  $f$01 = 0;$ntm$03 = 16384;$soff$02 = 0;
  while(1) {
   if ($14) {
    $22 = HEAP32[$5>>2]|0;
    $23 = ($22|0)==(0);
    if ($23) {
     $25 = 1152;
    } else {
     label = 10;
    }
   } else {
    label = 10;
   }
   if ((label|0) == 10) {
    label = 0;
    $25 = 576;
   }
   $24 = Math_imul($4, $25)|0;
   $26 = (($24) + ($ntm$03))|0;
   $27 = (($26|0) / 32768)&-1;
   $28 = (($27) + ($soff$02))|0;
   $29 = $27 << 15;
   $30 = (($26) - ($29))|0;
   $31 = (($f$01) + 1)|0;
   $exitcond16 = ($31|0)==($frame|0);
   if ($exitcond16) {
    $$0 = $28;
    break;
   } else {
    $f$01 = $31;$ntm$03 = $30;$soff$02 = $28;
   }
  }
  return ($$0|0);
 }
 }
 return (0)|0;
}
function _INT123_frame_expect_outsamples($fr) {
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $outs$0 = 0, $phitmp = 0, $vararg_buffer = 0, $vararg_ptr1 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($fr)) + 19680|0);
 $1 = HEAP32[$0>>2]|0;
 switch ($1|0) {
 case 2: case 1: case 0:  {
  $2 = ((($fr)) + 19688|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ($3|0)==(1);
  if ($4) {
   $13 = 384;
  } else {
   $5 = ($3|0)==(2);
   if ($5) {
    $13 = 1152;
   } else {
    $6 = ((($fr)) + 19672|0);
    $7 = HEAP32[$6>>2]|0;
    $8 = ($7|0)==(0);
    if ($8) {
     $9 = ((($fr)) + 19676|0);
     $10 = HEAP32[$9>>2]|0;
     $11 = ($10|0)!=(0);
     $phitmp = $11 ? 576 : 1152;
     $13 = $phitmp;
    } else {
     $13 = 576;
    }
   }
  }
  $12 = $13 >>> $1;
  $outs$0 = $12;
  STACKTOP = sp;return ($outs$0|0);
  break;
 }
 case 3:  {
  $14 = (_INT123_ntom_frame_outsamples($fr)|0);
  $outs$0 = $14;
  STACKTOP = sp;return ($outs$0|0);
  break;
 }
 default: {
  $15 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer>>2] = 811;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $1;
  (_fprintf($15,513889,$vararg_buffer)|0);
  $outs$0 = 0;
  STACKTOP = sp;return ($outs$0|0);
 }
 }
 return (0)|0;
}
function _INT123_ntom_frame_outsamples($fr) {
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 19352|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 19688|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(1);
 if ($4) {
  $15 = 384;
 } else {
  $5 = ($3|0)==(2);
  if ($5) {
   $15 = 1152;
  } else {
   $6 = ((($fr)) + 19672|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0);
   if ($8) {
    $9 = ((($fr)) + 19676|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = ($10|0)!=(0);
    $phitmp = $11 ? 576 : 1152;
    $15 = $phitmp;
   } else {
    $15 = 576;
   }
  }
 }
 $12 = ((($fr)) + 19360|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = Math_imul($13, $15)|0;
 $16 = (($14) + ($1))|0;
 $17 = (($16|0) / 32768)&-1;
 return ($17|0);
}
function _INT123_ntom_frameoff($fr,$soff) {
 $fr = $fr|0;
 $soff = $soff|0;
 var $$0 = 0, $$0$us13 = 0, $$0$us215 = 0, $$01 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ioff$0 = 0, $ioff$0$us14 = 0, $ioff$0$us316 = 0, $ntm$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($soff|0)<(1);
 if ($0) {
  $$01 = 0;
  return ($$01|0);
 }
 $1 = ((($fr)) + 19688|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($fr)) + 19360|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($fr)) + 19676|0);
 switch ($2|0) {
 case 1:  {
  $6 = ($4*384)|0;
  $7 = (($6) + 16384)|0;
  $8 = (($7|0) / 32768)&-1;
  $9 = ($8|0)>($soff|0);
  if ($9) {
   $$01 = 0;
   return ($$01|0);
  } else {
   $$0$us13 = $soff;$11 = $8;$14 = $7;$ioff$0$us14 = 0;
  }
  while(1) {
   $10 = (($$0$us13) - ($11))|0;
   $12 = $11 << 15;
   $13 = (($14) - ($12))|0;
   $15 = (($ioff$0$us14) + 1)|0;
   $16 = (($6) + ($13))|0;
   $17 = (($16|0) / 32768)&-1;
   $18 = ($10|0)<($17|0);
   if ($18) {
    $$01 = $15;
    break;
   } else {
    $$0$us13 = $10;$11 = $17;$14 = $16;$ioff$0$us14 = $15;
   }
  }
  return ($$01|0);
  break;
 }
 case 2:  {
  $19 = ($4*1152)|0;
  $20 = (($19) + 16384)|0;
  $21 = (($20|0) / 32768)&-1;
  $22 = ($21|0)>($soff|0);
  if ($22) {
   $$01 = 0;
   return ($$01|0);
  } else {
   $$0$us215 = $soff;$26 = $21;$29 = $20;$ioff$0$us316 = 0;
  }
  while(1) {
   $25 = (($$0$us215) - ($26))|0;
   $27 = $26 << 15;
   $28 = (($29) - ($27))|0;
   $30 = (($ioff$0$us316) + 1)|0;
   $31 = (($19) + ($28))|0;
   $32 = (($31|0) / 32768)&-1;
   $33 = ($25|0)<($32|0);
   if ($33) {
    $$01 = $30;
    break;
   } else {
    $$0$us215 = $25;$26 = $32;$29 = $31;$ioff$0$us316 = $30;
   }
  }
  return ($$01|0);
  break;
 }
 default: {
  $23 = ((($fr)) + 19672|0);
  $$pre = HEAP32[$23>>2]|0;
  $24 = ($$pre|0)==(0);
  $$0 = $soff;$ioff$0 = 0;$ntm$0 = 16384;
  while(1) {
   if ($24) {
    $34 = HEAP32[$5>>2]|0;
    $35 = ($34|0)==(0);
    if ($35) {
     $37 = 1152;
    } else {
     label = 10;
    }
   } else {
    label = 10;
   }
   if ((label|0) == 10) {
    label = 0;
    $37 = 576;
   }
   $36 = Math_imul($4, $37)|0;
   $38 = (($36) + ($ntm$0))|0;
   $39 = (($38|0) / 32768)&-1;
   $40 = ($$0|0)<($39|0);
   if ($40) {
    $$01 = $ioff$0;
    break;
   }
   $41 = (($$0) - ($39))|0;
   $42 = $39 << 15;
   $43 = (($38) - ($42))|0;
   $44 = (($ioff$0) + 1)|0;
   $$0 = $41;$ioff$0 = $44;$ntm$0 = $43;
  }
  return ($$01|0);
 }
 }
 return (0)|0;
}
function _INT123_frame_gapless_realinit($fr) {
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $outs$0$i = 0, $outs$0$i1 = 0, $outs$0$i3 = 0, $phitmp = 0, $vararg_buffer = 0, $vararg_buffer2 = 0;
 var $vararg_buffer6 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, $vararg_ptr9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer6 = sp + 16|0;
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($fr)) + 27932|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 19680|0);
 $3 = HEAP32[$2>>2]|0;
 switch ($3|0) {
 case 2: case 1: case 0:  {
  $4 = $1 >> $3;
  $outs$0$i = $4;
  break;
 }
 case 3:  {
  $5 = (_INT123_ntom_ins2outs($fr,$1)|0);
  $outs$0$i = $5;
  break;
 }
 default: {
  $6 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer>>2] = 769;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $3;
  (_fprintf($6,513889,$vararg_buffer)|0);
  $outs$0$i = 0;
 }
 }
 $7 = ((($fr)) + 27936|0);
 HEAP32[$7>>2] = $outs$0$i;
 $8 = ((($fr)) + 27940|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = HEAP32[$2>>2]|0;
 switch ($10|0) {
 case 2: case 1: case 0:  {
  $11 = $9 >> $10;
  $outs$0$i1 = $11;
  break;
 }
 case 3:  {
  $12 = (_INT123_ntom_ins2outs($fr,$9)|0);
  $outs$0$i1 = $12;
  break;
 }
 default: {
  $13 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer2>>2] = 769;
  $vararg_ptr5 = ((($vararg_buffer2)) + 4|0);
  HEAP32[$vararg_ptr5>>2] = $10;
  (_fprintf($13,513889,$vararg_buffer2)|0);
  $outs$0$i1 = 0;
 }
 }
 $14 = ((($fr)) + 27944|0);
 HEAP32[$14>>2] = $outs$0$i1;
 $15 = ((($fr)) + 27920|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($fr)) + 19688|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ($18|0)==(1);
 if ($19) {
  $28 = 384;
 } else {
  $20 = ($18|0)==(2);
  if ($20) {
   $28 = 1152;
  } else {
   $21 = ((($fr)) + 19672|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = ($22|0)==(0);
   if ($23) {
    $24 = ((($fr)) + 19676|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = ($25|0)!=(0);
    $phitmp = $26 ? 576 : 1152;
    $28 = $phitmp;
   } else {
    $28 = 576;
   }
  }
 }
 $27 = Math_imul($28, $16)|0;
 $29 = HEAP32[$2>>2]|0;
 switch ($29|0) {
 case 2: case 1: case 0:  {
  $30 = $27 >> $29;
  $outs$0$i3 = $30;
  $33 = ((($fr)) + 27948|0);
  HEAP32[$33>>2] = $outs$0$i3;
  STACKTOP = sp;return;
  break;
 }
 case 3:  {
  $31 = (_INT123_ntom_ins2outs($fr,$27)|0);
  $outs$0$i3 = $31;
  $33 = ((($fr)) + 27948|0);
  HEAP32[$33>>2] = $outs$0$i3;
  STACKTOP = sp;return;
  break;
 }
 default: {
  $32 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer6>>2] = 769;
  $vararg_ptr9 = ((($vararg_buffer6)) + 4|0);
  HEAP32[$vararg_ptr9>>2] = $29;
  (_fprintf($32,513889,$vararg_buffer6)|0);
  $outs$0$i3 = 0;
  $33 = ((($fr)) + 27948|0);
  HEAP32[$33>>2] = $outs$0$i3;
  STACKTOP = sp;return;
 }
 }
}
function _INT123_frame_gapless_update($fr,$total_samples) {
 $fr = $fr|0;
 $total_samples = $total_samples|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $phitmp = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($fr)) + 27920|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 19688|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(1);
 if ($4) {
  $13 = 384;
 } else {
  $5 = ($3|0)==(2);
  if ($5) {
   $13 = 1152;
  } else {
   $6 = ((($fr)) + 19672|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0);
   if ($8) {
    $9 = ((($fr)) + 19676|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = ($10|0)!=(0);
    $phitmp = $11 ? 576 : 1152;
    $13 = $phitmp;
   } else {
    $13 = 576;
   }
  }
 }
 $12 = Math_imul($13, $1)|0;
 $14 = ((($fr)) + 28068|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = $15 & 32;
 $17 = ($16|0)!=(0);
 $18 = ($12|0)==($total_samples|0);
 $or$cond = $18 | $17;
 if (!($or$cond)) {
  $19 = HEAP32[37544]|0;
  (_fwrite(514026,90,1,$19)|0);
 }
 $20 = ($12|0)>($total_samples|0);
 if (!($20)) {
  STACKTOP = sp;return;
 }
 $21 = HEAP32[$14>>2]|0;
 $22 = $21 & 32;
 $23 = ($22|0)==(0);
 if ($23) {
  $24 = HEAP32[37544]|0;
  $25 = ((($fr)) + 27940|0);
  $26 = HEAP32[$25>>2]|0;
  HEAP32[$vararg_buffer>>2] = 873;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $total_samples;
  $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
  HEAP32[$vararg_ptr2>>2] = $26;
  (_fprintf($24,514117,$vararg_buffer)|0);
  $$pre$phiZ2D = $25;
 } else {
  $$pre = ((($fr)) + 27940|0);
  $$pre$phiZ2D = $$pre;
 }
 HEAP32[$0>>2] = -1;
 HEAP32[$$pre$phiZ2D>>2] = 0;
 $27 = ((($fr)) + 27932|0);
 HEAP32[$27>>2] = 0;
 $28 = ((($fr)) + 27936|0);
 HEAP32[$28>>2] = 0;
 $29 = ((($fr)) + 27944|0);
 HEAP32[$29>>2] = 0;
 $30 = ((($fr)) + 27948|0);
 HEAP32[$30>>2] = 0;
 _INT123_frame_gapless_realinit($fr);
 $31 = ((($fr)) + 27912|0);
 HEAP32[$31>>2] = -1;
 $32 = ((($fr)) + 27928|0);
 HEAP32[$32>>2] = 0;
 STACKTOP = sp;return;
}
function _INT123_frame_set_frameseek($fr,$fe) {
 $fr = $fr|0;
 $fe = $fe|0;
 var $$$$i = 0, $$$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $fr$idx = 0, $fr$idx$val = 0, $fr$idx1$val = 0, $fr$idx2 = 0, $fr$idx2$val = 0, $num$0$i = 0, $num$0$i5 = 0, $or$cond$i = 0, $or$cond3$i = 0, $outs$0$i = 0, $outs$0$i8 = 0, $phitmp$i = 0, $phitmp$i3 = 0, $phitmp$i4 = 0, $phitmp$i7 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer5 = 0, $vararg_buffer8 = 0, $vararg_ptr11 = 0;
 var $vararg_ptr4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer8 = sp + 24|0;
 $vararg_buffer5 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($fr)) + 27908|0);
 HEAP32[$0>>2] = $fe;
 $1 = ((($fr)) + 28068|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2 & 64;
 $4 = ($3|0)==(0);
 do {
  if ($4) {
   label = 41;
  } else {
   $5 = ((($fr)) + 27920|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($6|0)>(0);
   if ($7) {
    $8 = ((($fr)) + 27936|0);
    $9 = HEAP32[$8>>2]|0;
    $10 = ((($fr)) + 19680|0);
    $11 = HEAP32[$10>>2]|0;
    switch ($11|0) {
    case 2: case 1: case 0:  {
     $12 = ((($fr)) + 19688|0);
     $13 = HEAP32[$12>>2]|0;
     $14 = ($13|0)==(1);
     if ($14) {
      $23 = 384;
     } else {
      $15 = ($13|0)==(2);
      if ($15) {
       $23 = 1152;
      } else {
       $16 = ((($fr)) + 19672|0);
       $17 = HEAP32[$16>>2]|0;
       $18 = ($17|0)==(0);
       if ($18) {
        $19 = ((($fr)) + 19676|0);
        $20 = HEAP32[$19>>2]|0;
        $21 = ($20|0)!=(0);
        $phitmp$i = $21 ? 576 : 1152;
        $23 = $phitmp$i;
       } else {
        $23 = 576;
       }
      }
     }
     $22 = $23 >>> $11;
     $24 = (($9|0) / ($22|0))&-1;
     $num$0$i = $24;
     break;
    }
    case 3:  {
     $25 = (_INT123_ntom_frameoff($fr,$9)|0);
     $num$0$i = $25;
     break;
    }
    default: {
     $26 = HEAP32[37544]|0;
     HEAP32[$vararg_buffer>>2] = 831;
     (_fprintf($26,513960,$vararg_buffer)|0);
     $num$0$i = 0;
    }
    }
    $27 = ($num$0$i|0)<($fe|0);
    if ($27) {
     $47 = ((($fr)) + 27924|0);
     HEAP32[$47>>2] = 0;
    } else {
     HEAP32[$0>>2] = $num$0$i;
     $28 = HEAP32[$8>>2]|0;
     $29 = HEAP32[$10>>2]|0;
     switch ($29|0) {
     case 2: case 1: case 0:  {
      $30 = ((($fr)) + 19688|0);
      $31 = HEAP32[$30>>2]|0;
      $32 = ($31|0)==(1);
      if ($32) {
       $41 = 384;
      } else {
       $33 = ($31|0)==(2);
       if ($33) {
        $41 = 1152;
       } else {
        $34 = ((($fr)) + 19672|0);
        $35 = HEAP32[$34>>2]|0;
        $36 = ($35|0)==(0);
        if ($36) {
         $37 = ((($fr)) + 19676|0);
         $38 = HEAP32[$37>>2]|0;
         $39 = ($38|0)!=(0);
         $phitmp$i3 = $39 ? 576 : 1152;
         $41 = $phitmp$i3;
        } else {
         $41 = 576;
        }
       }
      }
      $40 = $41 >>> $29;
      $42 = Math_imul($40, $num$0$i)|0;
      $outs$0$i = $42;
      break;
     }
     case 3:  {
      $43 = (_INT123_ntom_frmouts($fr,$num$0$i)|0);
      $outs$0$i = $43;
      break;
     }
     default: {
      $44 = HEAP32[37544]|0;
      HEAP32[$vararg_buffer1>>2] = 789;
      $vararg_ptr4 = ((($vararg_buffer1)) + 4|0);
      HEAP32[$vararg_ptr4>>2] = $29;
      (_fprintf($44,513889,$vararg_buffer1)|0);
      $outs$0$i = 0;
     }
     }
     $45 = (($28) - ($outs$0$i))|0;
     $46 = ((($fr)) + 27924|0);
     HEAP32[$46>>2] = $45;
    }
    $48 = ((($fr)) + 27944|0);
    $49 = HEAP32[$48>>2]|0;
    $50 = ($49|0)>(0);
    if (!($50)) {
     $87 = ((($fr)) + 27912|0);
     HEAP32[$87>>2] = -1;
     $88 = ((($fr)) + 27928|0);
     HEAP32[$88>>2] = 0;
     break;
    }
    $51 = HEAP32[$10>>2]|0;
    switch ($51|0) {
    case 2: case 1: case 0:  {
     $52 = ((($fr)) + 19688|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ($53|0)==(1);
     if ($54) {
      $63 = 384;
     } else {
      $55 = ($53|0)==(2);
      if ($55) {
       $63 = 1152;
      } else {
       $56 = ((($fr)) + 19672|0);
       $57 = HEAP32[$56>>2]|0;
       $58 = ($57|0)==(0);
       if ($58) {
        $59 = ((($fr)) + 19676|0);
        $60 = HEAP32[$59>>2]|0;
        $61 = ($60|0)!=(0);
        $phitmp$i4 = $61 ? 576 : 1152;
        $63 = $phitmp$i4;
       } else {
        $63 = 576;
       }
      }
     }
     $62 = $63 >>> $51;
     $64 = (($49|0) / ($62|0))&-1;
     $num$0$i5 = $64;
     break;
    }
    case 3:  {
     $65 = (_INT123_ntom_frameoff($fr,$49)|0);
     $num$0$i5 = $65;
     break;
    }
    default: {
     $66 = HEAP32[37544]|0;
     HEAP32[$vararg_buffer5>>2] = 831;
     (_fprintf($66,513960,$vararg_buffer5)|0);
     $num$0$i5 = 0;
    }
    }
    $67 = ((($fr)) + 27912|0);
    HEAP32[$67>>2] = $num$0$i5;
    $68 = HEAP32[$48>>2]|0;
    $69 = HEAP32[$10>>2]|0;
    switch ($69|0) {
    case 2: case 1: case 0:  {
     $70 = ((($fr)) + 19688|0);
     $71 = HEAP32[$70>>2]|0;
     $72 = ($71|0)==(1);
     if ($72) {
      $81 = 384;
     } else {
      $73 = ($71|0)==(2);
      if ($73) {
       $81 = 1152;
      } else {
       $74 = ((($fr)) + 19672|0);
       $75 = HEAP32[$74>>2]|0;
       $76 = ($75|0)==(0);
       if ($76) {
        $77 = ((($fr)) + 19676|0);
        $78 = HEAP32[$77>>2]|0;
        $79 = ($78|0)!=(0);
        $phitmp$i7 = $79 ? 576 : 1152;
        $81 = $phitmp$i7;
       } else {
        $81 = 576;
       }
      }
     }
     $80 = $81 >>> $69;
     $82 = Math_imul($80, $num$0$i5)|0;
     $outs$0$i8 = $82;
     break;
    }
    case 3:  {
     $83 = (_INT123_ntom_frmouts($fr,$num$0$i5)|0);
     $outs$0$i8 = $83;
     break;
    }
    default: {
     $84 = HEAP32[37544]|0;
     HEAP32[$vararg_buffer8>>2] = 789;
     $vararg_ptr11 = ((($vararg_buffer8)) + 4|0);
     HEAP32[$vararg_ptr11>>2] = $69;
     (_fprintf($84,513889,$vararg_buffer8)|0);
     $outs$0$i8 = 0;
    }
    }
    $85 = (($68) - ($outs$0$i8))|0;
    $86 = ((($fr)) + 27928|0);
    HEAP32[$86>>2] = $85;
   } else {
    label = 41;
   }
  }
 } while(0);
 if ((label|0) == 41) {
  $89 = ((($fr)) + 27928|0);
  HEAP32[$89>>2] = 0;
  $90 = ((($fr)) + 27924|0);
  HEAP32[$90>>2] = 0;
  $91 = ((($fr)) + 27912|0);
  HEAP32[$91>>2] = -1;
 }
 $fr$idx = ((($fr)) + 19688|0);
 $fr$idx$val = HEAP32[$fr$idx>>2]|0;
 $fr$idx1$val = HEAP32[$0>>2]|0;
 $fr$idx2 = ((($fr)) + 28360|0);
 $fr$idx2$val = HEAP32[$fr$idx2>>2]|0;
 $92 = ($fr$idx$val|0)==(3);
 $93 = ($fr$idx2$val|0)<(1);
 $or$cond$i = $92 & $93;
 $$$i = $or$cond$i ? 1 : $fr$idx2$val;
 $94 = ($fr$idx$val|0)!=(3);
 $95 = ($$$i|0)>(2);
 $or$cond3$i = $94 & $95;
 $$$$i = $or$cond3$i ? 2 : $$$i;
 $96 = (($fr$idx1$val) - ($$$$i))|0;
 $97 = ((($fr)) + 27916|0);
 HEAP32[$97>>2] = $96;
 STACKTOP = sp;return;
}
function _INT123_frame_skip($fr) {
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 19688|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(3);
 if (!($2)) {
  return;
 }
 _INT123_set_pointer($fr,512);
 return;
}
function _INT123_frame_set_seek($fr,$sp) {
 $fr = $fr|0;
 $sp = $sp|0;
 var $$$$i = 0, $$$i = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fr$idx = 0, $fr$idx$val = 0, $fr$idx1$val$pre = 0, $fr$idx2 = 0, $fr$idx2$val = 0, $num$0$i = 0, $or$cond$i = 0, $or$cond3$i = 0, $outs$0$i = 0, $phitmp$i = 0, $phitmp$i3 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_ptr4 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($fr)) + 19680|0);
 $1 = HEAP32[$0>>2]|0;
 switch ($1|0) {
 case 2: case 1: case 0:  {
  $2 = ((($fr)) + 19688|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ($3|0)==(1);
  if ($4) {
   $13 = 384;
  } else {
   $5 = ($3|0)==(2);
   if ($5) {
    $13 = 1152;
   } else {
    $6 = ((($fr)) + 19672|0);
    $7 = HEAP32[$6>>2]|0;
    $8 = ($7|0)==(0);
    if ($8) {
     $9 = ((($fr)) + 19676|0);
     $10 = HEAP32[$9>>2]|0;
     $11 = ($10|0)!=(0);
     $phitmp$i = $11 ? 576 : 1152;
     $13 = $phitmp$i;
    } else {
     $13 = 576;
    }
   }
  }
  $12 = $13 >>> $1;
  $14 = (($sp|0) / ($12|0))&-1;
  $num$0$i = $14;
  break;
 }
 case 3:  {
  $15 = (_INT123_ntom_frameoff($fr,$sp)|0);
  $num$0$i = $15;
  break;
 }
 default: {
  $16 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer>>2] = 831;
  (_fprintf($16,513960,$vararg_buffer)|0);
  $num$0$i = 0;
 }
 }
 $17 = ((($fr)) + 27908|0);
 HEAP32[$17>>2] = $num$0$i;
 $18 = HEAP32[$0>>2]|0;
 $19 = ($18|0)==(3);
 if ($19) {
  _INT123_ntom_set_ntom($fr,$num$0$i);
  $fr$idx1$val$pre = HEAP32[$17>>2]|0;
  $$pre = HEAP32[$0>>2]|0;
  $25 = $fr$idx1$val$pre;$27 = $$pre;
 } else {
  $25 = $num$0$i;$27 = $18;
 }
 $fr$idx = ((($fr)) + 19688|0);
 $fr$idx$val = HEAP32[$fr$idx>>2]|0;
 $fr$idx2 = ((($fr)) + 28360|0);
 $fr$idx2$val = HEAP32[$fr$idx2>>2]|0;
 $20 = ($fr$idx$val|0)==(3);
 $21 = ($fr$idx2$val|0)<(1);
 $or$cond$i = $20 & $21;
 $$$i = $or$cond$i ? 1 : $fr$idx2$val;
 $22 = ($fr$idx$val|0)!=(3);
 $23 = ($$$i|0)>(2);
 $or$cond3$i = $22 & $23;
 $$$$i = $or$cond3$i ? 2 : $$$i;
 $24 = (($25) - ($$$$i))|0;
 $26 = ((($fr)) + 27916|0);
 HEAP32[$26>>2] = $24;
 switch ($27|0) {
 case 2: case 1: case 0:  {
  $28 = ($fr$idx$val|0)==(1);
  if ($28) {
   $37 = 384;
  } else {
   $29 = ($fr$idx$val|0)==(2);
   if ($29) {
    $37 = 1152;
   } else {
    $30 = ((($fr)) + 19672|0);
    $31 = HEAP32[$30>>2]|0;
    $32 = ($31|0)==(0);
    if ($32) {
     $33 = ((($fr)) + 19676|0);
     $34 = HEAP32[$33>>2]|0;
     $35 = ($34|0)!=(0);
     $phitmp$i3 = $35 ? 576 : 1152;
     $37 = $phitmp$i3;
    } else {
     $37 = 576;
    }
   }
  }
  $36 = $37 >>> $27;
  $38 = Math_imul($36, $25)|0;
  $outs$0$i = $38;
  $41 = (($sp) - ($outs$0$i))|0;
  $42 = ((($fr)) + 27924|0);
  HEAP32[$42>>2] = $41;
  STACKTOP = sp;return;
  break;
 }
 case 3:  {
  $39 = (_INT123_ntom_frmouts($fr,$25)|0);
  $outs$0$i = $39;
  $41 = (($sp) - ($outs$0$i))|0;
  $42 = ((($fr)) + 27924|0);
  HEAP32[$42>>2] = $41;
  STACKTOP = sp;return;
  break;
 }
 default: {
  $40 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer1>>2] = 789;
  $vararg_ptr4 = ((($vararg_buffer1)) + 4|0);
  HEAP32[$vararg_ptr4>>2] = $27;
  (_fprintf($40,513889,$vararg_buffer1)|0);
  $outs$0$i = 0;
  $41 = (($sp) - ($outs$0$i))|0;
  $42 = ((($fr)) + 27924|0);
  HEAP32[$42>>2] = $41;
  STACKTOP = sp;return;
 }
 }
}
function _INT123_ntom_set_ntom($fr,$num) {
 $fr = $fr|0;
 $num = $num|0;
 var $$ = 0, $$pre = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond12$i = 0, $exitcond13$i = 0;
 var $exitcond14$i = 0, $exitcond14$i$us = 0, $f$02$i = 0, $f$02$i$us = 0, $f$02$us$i = 0, $f$02$us3$i = 0, $ntm$0$lcssa$i = 0, $ntm$01$i = 0, $ntm$01$i$us = 0, $ntm$01$us$i = 0, $ntm$01$us4$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($num|0)>(0);
 if ($0) {
  $1 = ((($fr)) + 19688|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ((($fr)) + 19360|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ((($fr)) + 19676|0);
  switch ($2|0) {
  case 1:  {
   $6 = ($4*384)|0;
   $f$02$us$i = 0;$ntm$01$us$i = 16384;
   while(1) {
    $7 = (($ntm$01$us$i) + ($6))|0;
    $8 = (($7|0) / 32768)&-1;
    $9 = $8 << 15;
    $10 = (($7) - ($9))|0;
    $11 = (($f$02$us$i) + 1)|0;
    $exitcond12$i = ($11|0)==($num|0);
    if ($exitcond12$i) {
     $ntm$0$lcssa$i = $10;
     break;
    } else {
     $f$02$us$i = $11;$ntm$01$us$i = $10;
    }
   }
   break;
  }
  case 2:  {
   $22 = ($4*1152)|0;
   $f$02$us3$i = 0;$ntm$01$us4$i = 16384;
   while(1) {
    $23 = (($ntm$01$us4$i) + ($22))|0;
    $24 = (($23|0) / 32768)&-1;
    $25 = $24 << 15;
    $26 = (($23) - ($25))|0;
    $27 = (($f$02$us3$i) + 1)|0;
    $exitcond13$i = ($27|0)==($num|0);
    if ($exitcond13$i) {
     $ntm$0$lcssa$i = $26;
     break;
    } else {
     $f$02$us3$i = $27;$ntm$01$us4$i = $26;
    }
   }
   break;
  }
  default: {
   $12 = ((($fr)) + 19672|0);
   $$pre$i = HEAP32[$12>>2]|0;
   $13 = ($$pre$i|0)==(0);
   if ($13) {
    $$pre = HEAP32[$5>>2]|0;
    $15 = ($$pre|0)==(0);
    $$ = $15 ? 1152 : 576;
    $16 = Math_imul($$, $4)|0;
    $f$02$i$us = 0;$ntm$01$i$us = 16384;
    while(1) {
     $17 = (($16) + ($ntm$01$i$us))|0;
     $18 = (($17|0) / 32768)&-1;
     $19 = $18 << 15;
     $20 = (($17) - ($19))|0;
     $21 = (($f$02$i$us) + 1)|0;
     $exitcond14$i$us = ($21|0)==($num|0);
     if ($exitcond14$i$us) {
      $ntm$0$lcssa$i = $20;
      break;
     } else {
      $f$02$i$us = $21;$ntm$01$i$us = $20;
     }
    }
   } else {
    $14 = ($4*576)|0;
    $f$02$i = 0;$ntm$01$i = 16384;
    while(1) {
     $28 = (($14) + ($ntm$01$i))|0;
     $29 = (($28|0) / 32768)&-1;
     $30 = $29 << 15;
     $31 = (($28) - ($30))|0;
     $32 = (($f$02$i) + 1)|0;
     $exitcond14$i = ($32|0)==($num|0);
     if ($exitcond14$i) {
      $ntm$0$lcssa$i = $31;
      break;
     } else {
      $f$02$i = $32;$ntm$01$i = $31;
     }
    }
   }
  }
  }
 } else {
  $ntm$0$lcssa$i = 16384;
 }
 $33 = ((($fr)) + 19352|0);
 HEAP32[$33>>2] = $ntm$0$lcssa$i;
 $34 = ((($fr)) + 19356|0);
 HEAP32[$34>>2] = $ntm$0$lcssa$i;
 return;
}
function _INT123_frame_output_format($fr) {
 $fr = $fr|0;
 var $$ = 0, $$0 = 0, $$0$i$ph$i = 0, $$0$i$ph$i10 = 0, $$0$i$ph$i17 = 0, $$0$i$ph$i3 = 0, $$0$i1$ph = 0, $$41 = 0, $$42 = 0, $$43 = 0, $$44 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $f0$1 = 0, $i$03$i = 0, $i$03$i$lcssa = 0, $i$03$i12 = 0, $i$03$i12$lcssa = 0, $i$03$i19 = 0, $i$03$i19$lcssa = 0, $i$03$i5 = 0, $i$03$i5$lcssa = 0, $nf = 0, $or$cond$i = 0, $vararg_buffer = 0, $vararg_buffer12 = 0, $vararg_buffer4 = 0, $vararg_ptr1 = 0, $vararg_ptr10 = 0, $vararg_ptr11 = 0, $vararg_ptr15 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0;
 var $vararg_ptr7 = 0, $vararg_ptr8 = 0, $vararg_ptr9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer12 = sp + 40|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer = sp;
 $nf = sp + 48|0;
 $0 = ((($fr)) + 19652|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($nf)) + 8|0);
 HEAP32[$2>>2] = $1;
 $3 = ((($fr)) + 28068|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 >>> 1;
 $6 = $5 & 8;
 $7 = $4 & 1024;
 $8 = ($7|0)==(0);
 $f0$1 = $8 ? $6 : 6;
 $9 = $7 >>> 8;
 $10 = $9 ^ 12;
 $11 = $4 & 7;
 $12 = ($11|0)==(0);
 $13 = $12 ? $1 : 1;
 $14 = $4 & 8;
 $15 = ($14|0)==(0);
 $$ = $15 ? $13 : 2;
 $16 = $4 & 15;
 $17 = ($16|0)==(0);
 if (!($17)) {
  HEAP32[$2>>2] = $$;
 }
 $18 = ((($fr)) + 28072|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($19|0)==(0);
 L4: do {
  if ($20) {
   $102 = (_freq_fit($fr,$nf,$f0$1,2)|0);
   $103 = ($102|0)==(0);
   if ($103) {
    $104 = ($f0$1>>>0)>(2);
    $105 = $104 ? $f0$1 : 2;
    $106 = (_freq_fit($fr,$nf,$105,$10)|0);
    $107 = ($106|0)==(0);
    if ($107) {
     $108 = HEAP32[$2>>2]|0;
     switch ($108|0) {
     case 2:  {
      $109 = HEAP32[$3>>2]|0;
      $110 = $109 & 8;
      $111 = ($110|0)==(0);
      if ($111) {
       HEAP32[$2>>2] = 1;
      }
      break;
     }
     case 1:  {
      $112 = HEAP32[$3>>2]|0;
      $113 = $112 & 7;
      $114 = ($113|0)==(0);
      if ($114) {
       HEAP32[$2>>2] = 2;
      }
      break;
     }
     default: {
     }
     }
     $115 = (_freq_fit($fr,$nf,$f0$1,2)|0);
     $116 = ($115|0)==(0);
     if ($116) {
      $117 = (_freq_fit($fr,$nf,$105,$10)|0);
      $118 = ($117|0)==(0);
      if ($118) {
       $119 = HEAP32[$3>>2]|0;
       $120 = $119 & 32;
       $121 = ($120|0)==(0);
       if ($121) {
        $122 = HEAP32[37544]|0;
        $123 = $119 & 8;
        $124 = ($123|0)==(0);
        if ($124) {
         $125 = $119 & 7;
         $126 = ($125|0)!=(0);
         $127 = $126 ? 514230 : 1282466;
         $136 = $127;
        } else {
         $136 = 514237;
        }
        $128 = $119 & 16;
        $129 = ($128|0)!=(0);
        $130 = $129 ? 514246 : 1282466;
        $131 = (_INT123_frame_freq($fr)|0);
        $132 = (_INT123_frame_freq($fr)|0);
        $133 = $132 >> 1;
        $134 = (_INT123_frame_freq($fr)|0);
        $135 = $134 >> 2;
        HEAP32[$vararg_buffer4>>2] = 269;
        $vararg_ptr7 = ((($vararg_buffer4)) + 4|0);
        HEAP32[$vararg_ptr7>>2] = $136;
        $vararg_ptr8 = ((($vararg_buffer4)) + 8|0);
        HEAP32[$vararg_ptr8>>2] = $130;
        $vararg_ptr9 = ((($vararg_buffer4)) + 12|0);
        HEAP32[$vararg_ptr9>>2] = $131;
        $vararg_ptr10 = ((($vararg_buffer4)) + 16|0);
        HEAP32[$vararg_ptr10>>2] = $133;
        $vararg_ptr11 = ((($vararg_buffer4)) + 20|0);
        HEAP32[$vararg_ptr11>>2] = $135;
        (_fprintf($122,514331,$vararg_buffer4)|0);
       }
       $137 = ((($fr)) + 28376|0);
       HEAP32[$137>>2] = 1;
       $$0 = -1;
       STACKTOP = sp;return ($$0|0);
      }
     }
    }
   }
  } else {
   $21 = ((($nf)) + 12|0);
   HEAP32[$21>>2] = $19;
   $22 = (($$) + -1)|0;
   $23 = ($19|0)==(8000);
   if ($23) {
    $$0$i$ph$i = 0;
   } else {
    $24 = ($19|0)==(11025);
    if ($24) {
     $$0$i$ph$i = 1;
    } else {
     $25 = ($19|0)==(12000);
     if ($25) {
      $$0$i$ph$i = 2;
     } else {
      $26 = ($19|0)==(16000);
      if ($26) {
       $$0$i$ph$i = 3;
      } else {
       $27 = ($19|0)==(22050);
       if ($27) {
        $$0$i$ph$i = 4;
       } else {
        $28 = ($19|0)==(24000);
        if ($28) {
         $$0$i$ph$i = 5;
        } else {
         $29 = ($19|0)==(32000);
         if ($29) {
          $$0$i$ph$i = 6;
         } else {
          $30 = ($19|0)==(44100);
          if ($30) {
           $$0$i$ph$i = 7;
          } else {
           $31 = ($19|0)==(48000);
           $$41 = $31 ? 8 : 9;
           $$0$i$ph$i = $$41;
          }
         }
        }
       }
      }
     }
    }
   }
   $32 = ($f0$1>>>0)<(2);
   L34: do {
    if ($32) {
     $i$03$i = $6;
     while(1) {
      $33 = (((((($fr)) + 28096|0) + (($22*120)|0)|0) + (($$0$i$ph$i*12)|0)|0) + ($i$03$i)|0);
      $34 = HEAP8[$33>>0]|0;
      $35 = ($34<<24>>24)==(0);
      if (!($35)) {
       $i$03$i$lcssa = $i$03$i;
       break;
      }
      $36 = (($i$03$i) + 1)|0;
      $37 = ($36|0)<(2);
      if ($37) {
       $i$03$i = $36;
      } else {
       break L34;
      }
     }
     $38 = (36948 + ($i$03$i$lcssa<<2)|0);
     $39 = HEAP32[$38>>2]|0;
     HEAP32[$nf>>2] = $39;
     break L4;
    }
   } while(0);
   $40 = ($f0$1>>>0)>(2);
   $41 = $40 ? $f0$1 : 2;
   if ($23) {
    $$0$i$ph$i3 = 0;
   } else {
    $42 = ($19|0)==(11025);
    if ($42) {
     $$0$i$ph$i3 = 1;
    } else {
     $43 = ($19|0)==(12000);
     if ($43) {
      $$0$i$ph$i3 = 2;
     } else {
      $44 = ($19|0)==(16000);
      if ($44) {
       $$0$i$ph$i3 = 3;
      } else {
       $45 = ($19|0)==(22050);
       if ($45) {
        $$0$i$ph$i3 = 4;
       } else {
        $46 = ($19|0)==(24000);
        if ($46) {
         $$0$i$ph$i3 = 5;
        } else {
         $47 = ($19|0)==(32000);
         if ($47) {
          $$0$i$ph$i3 = 6;
         } else {
          $48 = ($19|0)==(44100);
          if ($48) {
           $$0$i$ph$i3 = 7;
          } else {
           $49 = ($19|0)==(48000);
           $$42 = $49 ? 8 : 9;
           $$0$i$ph$i3 = $$42;
          }
         }
        }
       }
      }
     }
    }
   }
   $50 = ($41|0)<($10|0);
   L50: do {
    if ($50) {
     $i$03$i5 = $41;
     while(1) {
      $51 = (((((($fr)) + 28096|0) + (($22*120)|0)|0) + (($$0$i$ph$i3*12)|0)|0) + ($i$03$i5)|0);
      $52 = HEAP8[$51>>0]|0;
      $53 = ($52<<24>>24)==(0);
      if (!($53)) {
       $i$03$i5$lcssa = $i$03$i5;
       break;
      }
      $54 = (($i$03$i5) + 1)|0;
      $55 = ($54|0)<($10|0);
      if ($55) {
       $i$03$i5 = $54;
      } else {
       break L50;
      }
     }
     $56 = (36948 + ($i$03$i5$lcssa<<2)|0);
     $57 = HEAP32[$56>>2]|0;
     HEAP32[$nf>>2] = $57;
     break L4;
    }
   } while(0);
   switch ($$|0) {
   case 2:  {
    if ($15) {
     HEAP32[$2>>2] = 1;
     $59 = 1;
    } else {
     $59 = 2;
    }
    break;
   }
   case 1:  {
    if ($12) {
     HEAP32[$2>>2] = 2;
     $59 = 2;
    } else {
     $59 = 1;
    }
    break;
   }
   default: {
    $59 = $$;
   }
   }
   $58 = (($59) + -1)|0;
   if ($23) {
    $$0$i$ph$i10 = 0;
   } else {
    $60 = ($19|0)==(11025);
    if ($60) {
     $$0$i$ph$i10 = 1;
    } else {
     $61 = ($19|0)==(12000);
     if ($61) {
      $$0$i$ph$i10 = 2;
     } else {
      $62 = ($19|0)==(16000);
      if ($62) {
       $$0$i$ph$i10 = 3;
      } else {
       $63 = ($19|0)==(22050);
       if ($63) {
        $$0$i$ph$i10 = 4;
       } else {
        $64 = ($19|0)==(24000);
        if ($64) {
         $$0$i$ph$i10 = 5;
        } else {
         $65 = ($19|0)==(32000);
         if ($65) {
          $$0$i$ph$i10 = 6;
         } else {
          $66 = ($19|0)==(44100);
          if ($66) {
           $$0$i$ph$i10 = 7;
          } else {
           $67 = ($19|0)==(48000);
           $$43 = $67 ? 8 : 9;
           $$0$i$ph$i10 = $$43;
          }
         }
        }
       }
      }
     }
    }
   }
   $68 = ($f0$1>>>0)<(2);
   L72: do {
    if ($68) {
     $i$03$i12 = $6;
     while(1) {
      $69 = (((((($fr)) + 28096|0) + (($58*120)|0)|0) + (($$0$i$ph$i10*12)|0)|0) + ($i$03$i12)|0);
      $70 = HEAP8[$69>>0]|0;
      $71 = ($70<<24>>24)==(0);
      if (!($71)) {
       $i$03$i12$lcssa = $i$03$i12;
       break;
      }
      $72 = (($i$03$i12) + 1)|0;
      $73 = ($72|0)<(2);
      if ($73) {
       $i$03$i12 = $72;
      } else {
       break L72;
      }
     }
     $74 = (36948 + ($i$03$i12$lcssa<<2)|0);
     $75 = HEAP32[$74>>2]|0;
     HEAP32[$nf>>2] = $75;
     break L4;
    }
   } while(0);
   if ($23) {
    $$0$i$ph$i17 = 0;
   } else {
    $76 = ($19|0)==(11025);
    if ($76) {
     $$0$i$ph$i17 = 1;
    } else {
     $77 = ($19|0)==(12000);
     if ($77) {
      $$0$i$ph$i17 = 2;
     } else {
      $78 = ($19|0)==(16000);
      if ($78) {
       $$0$i$ph$i17 = 3;
      } else {
       $79 = ($19|0)==(22050);
       if ($79) {
        $$0$i$ph$i17 = 4;
       } else {
        $80 = ($19|0)==(24000);
        if ($80) {
         $$0$i$ph$i17 = 5;
        } else {
         $81 = ($19|0)==(32000);
         if ($81) {
          $$0$i$ph$i17 = 6;
         } else {
          $82 = ($19|0)==(44100);
          if ($82) {
           $$0$i$ph$i17 = 7;
          } else {
           $83 = ($19|0)==(48000);
           $$44 = $83 ? 8 : 9;
           $$0$i$ph$i17 = $$44;
          }
         }
        }
       }
      }
     }
    }
   }
   $84 = ($41|0)<($10|0);
   L88: do {
    if ($84) {
     $i$03$i19 = $41;
     while(1) {
      $85 = (((((($fr)) + 28096|0) + (($58*120)|0)|0) + (($$0$i$ph$i17*12)|0)|0) + ($i$03$i19)|0);
      $86 = HEAP8[$85>>0]|0;
      $87 = ($86<<24>>24)==(0);
      if (!($87)) {
       $i$03$i19$lcssa = $i$03$i19;
       break;
      }
      $88 = (($i$03$i19) + 1)|0;
      $89 = ($88|0)<($10|0);
      if ($89) {
       $i$03$i19 = $88;
      } else {
       break L88;
      }
     }
     $90 = (36948 + ($i$03$i19$lcssa<<2)|0);
     $91 = HEAP32[$90>>2]|0;
     HEAP32[$nf>>2] = $91;
     break L4;
    }
   } while(0);
   $92 = $4 & 32;
   $93 = ($92|0)==(0);
   if ($93) {
    $94 = HEAP32[37544]|0;
    $95 = ($11|0)!=(0);
    $96 = $95 ? 514230 : 1282466;
    $97 = $15 ? $96 : 514237;
    $98 = $4 & 16;
    $99 = ($98|0)!=(0);
    $100 = $99 ? 514246 : 1282466;
    HEAP32[$vararg_buffer>>2] = 244;
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
    HEAP32[$vararg_ptr1>>2] = $97;
    $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
    HEAP32[$vararg_ptr2>>2] = $100;
    $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
    HEAP32[$vararg_ptr3>>2] = $19;
    (_fprintf($94,514253,$vararg_buffer)|0);
   }
   $101 = ((($fr)) + 28376|0);
   HEAP32[$101>>2] = 1;
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 $138 = ((($nf)) + 12|0);
 $139 = HEAP32[$138>>2]|0;
 $140 = ((($fr)) + 27876|0);
 $141 = ((($fr)) + 27888|0);
 $142 = HEAP32[$141>>2]|0;
 $143 = ($139|0)==($142|0);
 $144 = HEAP32[$2>>2]|0;
 $145 = ((($fr)) + 27884|0);
 if ($143) {
  $146 = HEAP32[$145>>2]|0;
  $147 = ($144|0)==($146|0);
  if ($147) {
   $148 = HEAP32[$nf>>2]|0;
   $149 = HEAP32[$140>>2]|0;
   $150 = ($148|0)==($149|0);
   if ($150) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
  }
 }
 HEAP32[$141>>2] = $139;
 HEAP32[$145>>2] = $144;
 $151 = HEAP32[$nf>>2]|0;
 HEAP32[$140>>2] = $151;
 $152 = $151 & 15;
 $153 = ($152|0)==(0);
 if ($153) {
  $154 = $151 & 64;
  $155 = ($154|0)==(0);
  if ($155) {
   $156 = $151 & 16384;
   $157 = ($156|0)==(0);
   if ($157) {
    $158 = $151 & 256;
    $159 = ($158|0)!=(0);
    $160 = ($151|0)==(512);
    $or$cond$i = $160 | $159;
    if ($or$cond$i) {
     $$0$i1$ph = 4;
    } else {
     $161 = ($151|0)==(1024);
     if ($161) {
      $$0$i1$ph = 8;
     } else {
      $163 = ((($fr)) + 27880|0);
      HEAP32[$163>>2] = 0;
      $164 = HEAP32[$3>>2]|0;
      $165 = $164 & 32;
      $166 = ($165|0)==(0);
      if ($166) {
       $167 = HEAP32[37544]|0;
       HEAP32[$vararg_buffer12>>2] = 293;
       $vararg_ptr15 = ((($vararg_buffer12)) + 4|0);
       HEAP32[$vararg_ptr15>>2] = $151;
       (_fprintf($167,514421,$vararg_buffer12)|0);
      }
      $168 = ((($fr)) + 28376|0);
      HEAP32[$168>>2] = 1;
      $$0 = -1;
      STACKTOP = sp;return ($$0|0);
     }
    }
   } else {
    $$0$i1$ph = 3;
   }
  } else {
   $$0$i1$ph = 2;
  }
 } else {
  $$0$i1$ph = 1;
 }
 $162 = ((($fr)) + 27880|0);
 HEAP32[$162>>2] = $$0$i1$ph;
 $$0 = 1;
 STACKTOP = sp;return ($$0|0);
}
function _freq_fit($fr,$nf,$f0,$f2) {
 $fr = $fr|0;
 $nf = $nf|0;
 $f0 = $f0|0;
 $f2 = $f2|0;
 var $$0$i$ph$i = 0, $$0$i$ph$i11 = 0, $$0$i$ph$i4 = 0, $$0$i2$ph = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $i$03$i = 0, $i$03$i$lcssa = 0, $i$03$i13 = 0, $i$03$i13$lcssa = 0, $i$03$i6 = 0, $i$03$i6$lcssa = 0, $i$031 = 0, $i$031$lcssa63 = 0, $i$130 = 0, $i$130$lcssa61 = 0, $or$cond$i = 0, $or$cond$i$i = 0, $or$cond$i$i10 = 0, $or$cond$i$i3 = 0, $rrn$0 = 0, $rrn$0$in = 0, $rrn$0$lcssa62 = 0, $rrn$1 = 0, $rrn$1$in = 0, $rrn$1$lcssa60 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_INT123_frame_freq($fr)|0);
 $1 = ((($fr)) + 28076|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $0 >> $2;
 $4 = ((($nf)) + 12|0);
 HEAP32[$4>>2] = $3;
 $5 = ((($nf)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + -1)|0;
 $8 = ($3|0)==(8000);
 if ($8) {
  $$0$i$ph$i = 0;
  label = 11;
 } else {
  $9 = ($3|0)==(11025);
  if ($9) {
   $$0$i$ph$i = 1;
   label = 11;
  } else {
   $10 = ($3|0)==(12000);
   if ($10) {
    $$0$i$ph$i = 2;
    label = 11;
   } else {
    $11 = ($3|0)==(16000);
    if ($11) {
     $$0$i$ph$i = 3;
     label = 11;
    } else {
     $12 = ($3|0)==(22050);
     if ($12) {
      $$0$i$ph$i = 4;
      label = 11;
     } else {
      $13 = ($3|0)==(24000);
      if ($13) {
       $$0$i$ph$i = 5;
       label = 11;
      } else {
       $14 = ($3|0)==(32000);
       if ($14) {
        $$0$i$ph$i = 6;
        label = 11;
       } else {
        $15 = ($3|0)==(44100);
        if ($15) {
         $$0$i$ph$i = 7;
         label = 11;
        } else {
         $16 = ($3|0)==(48000);
         if ($16) {
          $$0$i$ph$i = 8;
          label = 11;
         } else {
          $17 = ((($fr)) + 28072|0);
          $18 = HEAP32[$17>>2]|0;
          $19 = ($18|0)!=(0);
          $20 = ($18|0)==($3|0);
          $or$cond$i$i = $19 & $20;
          if ($or$cond$i$i) {
           $$0$i$ph$i = 9;
           label = 11;
          }
         }
        }
       }
      }
     }
    }
   }
  }
 }
 L11: do {
  if ((label|0) == 11) {
   $21 = ($f0|0)<($f2|0);
   if ($21) {
    $i$03$i = $f0;
    while(1) {
     $22 = (((((($fr)) + 28096|0) + (($7*120)|0)|0) + (($$0$i$ph$i*12)|0)|0) + ($i$03$i)|0);
     $23 = HEAP8[$22>>0]|0;
     $24 = ($23<<24>>24)==(0);
     if (!($24)) {
      $i$03$i$lcssa = $i$03$i;
      break;
     }
     $25 = (($i$03$i) + 1)|0;
     $26 = ($25|0)<($f2|0);
     if ($26) {
      $i$03$i = $25;
     } else {
      break L11;
     }
    }
    $27 = (36948 + ($i$03$i$lcssa<<2)|0);
    $28 = HEAP32[$27>>2]|0;
    HEAP32[$nf>>2] = $28;
    $$1 = 1;
    return ($$1|0);
   }
  }
 } while(0);
 $29 = ((($fr)) + 28068|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = $30 & 32768;
 $32 = ($31|0)==(0);
 if ($32) {
  $$1 = 0;
  return ($$1|0);
 }
 $33 = $3 >> 1;
 HEAP32[$4>>2] = $33;
 $34 = ($33|0)==(8000);
 if ($34) {
  $$0$i$ph$i11 = 0;
  label = 26;
 } else {
  $35 = ($33|0)==(11025);
  if ($35) {
   $$0$i$ph$i11 = 1;
   label = 26;
  } else {
   $36 = ($33|0)==(12000);
   if ($36) {
    $$0$i$ph$i11 = 2;
    label = 26;
   } else {
    $37 = ($33|0)==(16000);
    if ($37) {
     $$0$i$ph$i11 = 3;
     label = 26;
    } else {
     $38 = ($33|0)==(22050);
     if ($38) {
      $$0$i$ph$i11 = 4;
      label = 26;
     } else {
      $39 = ($33|0)==(24000);
      if ($39) {
       $$0$i$ph$i11 = 5;
       label = 26;
      } else {
       $40 = ($33|0)==(32000);
       if ($40) {
        $$0$i$ph$i11 = 6;
        label = 26;
       } else {
        $41 = ($33|0)==(44100);
        if ($41) {
         $$0$i$ph$i11 = 7;
         label = 26;
        } else {
         $42 = ($33|0)==(48000);
         if ($42) {
          $$0$i$ph$i11 = 8;
          label = 26;
         } else {
          $43 = ((($fr)) + 28072|0);
          $44 = HEAP32[$43>>2]|0;
          $45 = ($44|0)!=(0);
          $46 = ($44|0)==($33|0);
          $or$cond$i$i10 = $45 & $46;
          if ($or$cond$i$i10) {
           $$0$i$ph$i11 = 9;
           label = 26;
          }
         }
        }
       }
      }
     }
    }
   }
  }
 }
 L32: do {
  if ((label|0) == 26) {
   $47 = ($f0|0)<($f2|0);
   if ($47) {
    $i$03$i13 = $f0;
    while(1) {
     $48 = (((((($fr)) + 28096|0) + (($7*120)|0)|0) + (($$0$i$ph$i11*12)|0)|0) + ($i$03$i13)|0);
     $49 = HEAP8[$48>>0]|0;
     $50 = ($49<<24>>24)==(0);
     if (!($50)) {
      $i$03$i13$lcssa = $i$03$i13;
      break;
     }
     $51 = (($i$03$i13) + 1)|0;
     $52 = ($51|0)<($f2|0);
     if ($52) {
      $i$03$i13 = $51;
     } else {
      break L32;
     }
    }
    $53 = (36948 + ($i$03$i13$lcssa<<2)|0);
    $54 = HEAP32[$53>>2]|0;
    HEAP32[$nf>>2] = $54;
    $$1 = 1;
    return ($$1|0);
   }
  }
 } while(0);
 $55 = $3 >> 2;
 HEAP32[$4>>2] = $55;
 $56 = ($55|0)==(8000);
 if ($56) {
  $$0$i$ph$i4 = 0;
  label = 40;
 } else {
  $57 = ($55|0)==(11025);
  if ($57) {
   $$0$i$ph$i4 = 1;
   label = 40;
  } else {
   $58 = ($55|0)==(12000);
   if ($58) {
    $$0$i$ph$i4 = 2;
    label = 40;
   } else {
    $59 = ($55|0)==(16000);
    if ($59) {
     $$0$i$ph$i4 = 3;
     label = 40;
    } else {
     $60 = ($55|0)==(22050);
     if ($60) {
      $$0$i$ph$i4 = 4;
      label = 40;
     } else {
      $61 = ($55|0)==(24000);
      if ($61) {
       $$0$i$ph$i4 = 5;
       label = 40;
      } else {
       $62 = ($55|0)==(32000);
       if ($62) {
        $$0$i$ph$i4 = 6;
        label = 40;
       } else {
        $63 = ($55|0)==(44100);
        if ($63) {
         $$0$i$ph$i4 = 7;
         label = 40;
        } else {
         $64 = ($55|0)==(48000);
         if ($64) {
          $$0$i$ph$i4 = 8;
          label = 40;
         } else {
          $65 = ((($fr)) + 28072|0);
          $66 = HEAP32[$65>>2]|0;
          $67 = ($66|0)!=(0);
          $68 = ($66|0)==($55|0);
          $or$cond$i$i3 = $67 & $68;
          if ($or$cond$i$i3) {
           $$0$i$ph$i4 = 9;
           label = 40;
          }
         }
        }
       }
      }
     }
    }
   }
  }
 }
 L50: do {
  if ((label|0) == 40) {
   $69 = ($f0|0)<($f2|0);
   if ($69) {
    $i$03$i6 = $f0;
    while(1) {
     $70 = (((((($fr)) + 28096|0) + (($7*120)|0)|0) + (($$0$i$ph$i4*12)|0)|0) + ($i$03$i6)|0);
     $71 = HEAP8[$70>>0]|0;
     $72 = ($71<<24>>24)==(0);
     if (!($72)) {
      $i$03$i6$lcssa = $i$03$i6;
      break;
     }
     $73 = (($i$03$i6) + 1)|0;
     $74 = ($73|0)<($f2|0);
     if ($74) {
      $i$03$i6 = $73;
     } else {
      break L50;
     }
    }
    $75 = (36948 + ($i$03$i6$lcssa<<2)|0);
    $76 = HEAP32[$75>>2]|0;
    HEAP32[$nf>>2] = $76;
    $$1 = 1;
    return ($$1|0);
   }
  }
 } while(0);
 if ($32) {
  $$1 = 0;
  return ($$1|0);
 }
 $77 = ((($fr)) + 28072|0);
 $78 = HEAP32[$77>>2]|0;
 $79 = $78 | $2;
 $80 = ($79|0)==(0);
 if (!($80)) {
  $$1 = 0;
  return ($$1|0);
 }
 $81 = (_INT123_frame_freq($fr)|0);
 $82 = ($81|0)==(8000);
 if ($82) {
  $$0$i2$ph = 0;
 } else {
  $83 = ($81|0)==(11025);
  if ($83) {
   $$0$i2$ph = 1;
  } else {
   $84 = ($81|0)==(12000);
   if ($84) {
    $$0$i2$ph = 2;
   } else {
    $85 = ($81|0)==(16000);
    if ($85) {
     $$0$i2$ph = 3;
    } else {
     $86 = ($81|0)==(22050);
     if ($86) {
      $$0$i2$ph = 4;
     } else {
      $87 = ($81|0)==(24000);
      if ($87) {
       $$0$i2$ph = 5;
      } else {
       $88 = ($81|0)==(32000);
       if ($88) {
        $$0$i2$ph = 6;
       } else {
        $89 = ($81|0)==(44100);
        if ($89) {
         $$0$i2$ph = 7;
        } else {
         $90 = ($81|0)==(48000);
         if ($90) {
          $$0$i2$ph = 8;
         } else {
          $91 = HEAP32[$77>>2]|0;
          $92 = ($91|0)!=(0);
          $93 = ($91|0)==($81|0);
          $or$cond$i = $92 & $93;
          if ($or$cond$i) {
           $$0$i2$ph = 9;
          } else {
           $$1 = 0;
           return ($$1|0);
          }
         }
        }
       }
      }
     }
    }
   }
  }
 }
 $94 = ($f0|0)<($f2|0);
 if ($94) {
  $i$031 = $f0;
 } else {
  $$1 = 0;
  return ($$1|0);
 }
 L78: while(1) {
  $rrn$0$in = $$0$i2$ph;
  while(1) {
   $rrn$0 = (($rrn$0$in) + 1)|0;
   $95 = ($rrn$0|0)<(9);
   if (!($95)) {
    break;
   }
   $96 = (((((($fr)) + 28096|0) + (($7*120)|0)|0) + (($rrn$0*12)|0)|0) + ($i$031)|0);
   $97 = HEAP8[$96>>0]|0;
   $98 = ($97<<24>>24)==(0);
   if ($98) {
    $rrn$0$in = $rrn$0;
   } else {
    $i$031$lcssa63 = $i$031;$rrn$0$lcssa62 = $rrn$0;
    label = 61;
    break L78;
   }
  }
  $103 = (($i$031) + 1)|0;
  $104 = ($103|0)<($f2|0);
  if ($104) {
   $i$031 = $103;
  } else {
   break;
  }
 }
 if ((label|0) == 61) {
  $99 = (36912 + ($rrn$0$lcssa62<<2)|0);
  $100 = HEAP32[$99>>2]|0;
  HEAP32[$4>>2] = $100;
  $101 = (36948 + ($i$031$lcssa63<<2)|0);
  $102 = HEAP32[$101>>2]|0;
  HEAP32[$nf>>2] = $102;
  $$1 = 1;
  return ($$1|0);
 }
 if ($94) {
  $i$130 = $f0;
 } else {
  $$1 = 0;
  return ($$1|0);
 }
 L90: while(1) {
  $rrn$1$in = $$0$i2$ph;
  while(1) {
   $rrn$1 = (($rrn$1$in) + -1)|0;
   $105 = ($rrn$1$in|0)>(0);
   if (!($105)) {
    break;
   }
   $106 = (((((($fr)) + 28096|0) + (($7*120)|0)|0) + (($rrn$1*12)|0)|0) + ($i$130)|0);
   $107 = HEAP8[$106>>0]|0;
   $108 = ($107<<24>>24)==(0);
   if ($108) {
    $rrn$1$in = $rrn$1;
   } else {
    $i$130$lcssa61 = $i$130;$rrn$1$lcssa60 = $rrn$1;
    break L90;
   }
  }
  $113 = (($i$130) + 1)|0;
  $114 = ($113|0)<($f2|0);
  if ($114) {
   $i$130 = $113;
  } else {
   $$1 = 0;
   label = 68;
   break;
  }
 }
 if ((label|0) == 68) {
  return ($$1|0);
 }
 $109 = (36912 + ($rrn$1$lcssa60<<2)|0);
 $110 = HEAP32[$109>>2]|0;
 HEAP32[$4>>2] = $110;
 $111 = (36948 + ($i$130$lcssa61<<2)|0);
 $112 = HEAP32[$111>>2]|0;
 HEAP32[$nf>>2] = $112;
 $$1 = 1;
 return ($$1|0);
}
function _mpg123_format_none($mh) {
 $mh = $mh|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mh|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ((($mh)) + 28068|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2 & 32;
 $4 = ($3|0)==(0);
 if ($4) {
  $5 = ((($mh)) + 28064|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6|0)>(2);
  if ($7) {
   $8 = HEAP32[37544]|0;
   (_fwrite(514473,29,1,$8)|0);
  }
 }
 $9 = ((($mh)) + 28096|0);
 _memset(($9|0),0,240)|0;
 $$0 = 0;
 return ($$0|0);
}
function _mpg123_format($mh,$rate,$channels,$encodings) {
 $mh = $mh|0;
 $rate = $rate|0;
 $channels = $channels|0;
 $encodings = $encodings|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mh|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ((($mh)) + 28064|0);
 $2 = (_mpg123_fmt($1,$rate,$channels,$encodings)|0);
 $3 = ($2|0)==(0);
 if ($3) {
  $$0 = 0;
  return ($$0|0);
 }
 $4 = ((($mh)) + 28376|0);
 HEAP32[$4>>2] = $2;
 $$0 = -1;
 return ($$0|0);
}
function _mpg123_fmt($mp,$rate,$channels,$encodings) {
 $mp = $mp|0;
 $rate = $rate|0;
 $channels = $channels|0;
 $encodings = $encodings|0;
 var $$0 = 0, $$0$i$ph = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $ch = 0, $or$cond$i = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp + 8|0;
 $ch = sp;
 $0 = $ch;
 $1 = $0;
 HEAP32[$1>>2] = 0;
 $2 = (($0) + 4)|0;
 $3 = $2;
 HEAP32[$3>>2] = 1;
 $4 = ($mp|0)==(0|0);
 if ($4) {
  $$0 = 25;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = $channels & 3;
 $6 = ($5|0)==(0);
 if ($6) {
  $$0 = 2;
  STACKTOP = sp;return ($$0|0);
 }
 $7 = ((($mp)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $8 & 32;
 $10 = ($9|0)==(0);
 if ($10) {
  $11 = HEAP32[$mp>>2]|0;
  $12 = ($11|0)>(2);
  if ($12) {
   $13 = HEAP32[37544]|0;
   HEAP32[$vararg_buffer>>2] = $rate;
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
   HEAP32[$vararg_ptr1>>2] = $channels;
   $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
   HEAP32[$vararg_ptr2>>2] = $encodings;
   (_fprintf($13,514503,$vararg_buffer)|0);
  }
 }
 $14 = $channels & 2;
 $15 = ($14|0)==(0);
 if ($15) {
  $16 = ((($ch)) + 4|0);
  HEAP32[$16>>2] = 0;
  $33 = 0;$34 = 0;
 } else {
  $17 = $channels & 1;
  $18 = ($17|0)==(0);
  if ($18) {
   HEAP32[$ch>>2] = 1;
   $33 = 1;$34 = 1;
  } else {
   $33 = 0;$34 = 1;
  }
 }
 $19 = ($rate|0)==(8000);
 if ($19) {
  $$0$i$ph = 0;
 } else {
  $20 = ($rate|0)==(11025);
  if ($20) {
   $$0$i$ph = 1;
  } else {
   $25 = ($rate|0)==(12000);
   if ($25) {
    $$0$i$ph = 2;
   } else {
    $26 = ($rate|0)==(16000);
    if ($26) {
     $$0$i$ph = 3;
    } else {
     $27 = ($rate|0)==(22050);
     if ($27) {
      $$0$i$ph = 4;
     } else {
      $28 = ($rate|0)==(24000);
      if ($28) {
       $$0$i$ph = 5;
      } else {
       $29 = ($rate|0)==(32000);
       if ($29) {
        $$0$i$ph = 6;
       } else {
        $30 = ($rate|0)==(44100);
        if ($30) {
         $$0$i$ph = 7;
        } else {
         $31 = ($rate|0)==(48000);
         if ($31) {
          $$0$i$ph = 8;
         } else {
          $21 = ((($mp)) + 8|0);
          $22 = HEAP32[$21>>2]|0;
          $23 = ($22|0)!=(0);
          $24 = ($22|0)==($rate|0);
          $or$cond$i = $23 & $24;
          if ($or$cond$i) {
           $$0$i$ph = 9;
          } else {
           $$0 = 3;
           STACKTOP = sp;return ($$0|0);
          }
         }
        }
       }
      }
     }
    }
   }
  }
 }
 $32 = ($33|0)==($34|0);
 $35 = $encodings & 208;
 $36 = ($35|0)==(208);
 if ($32) {
  if ($36) {
   $57 = ((((($mp)) + 32|0) + (($34*120)|0)|0) + (($$0$i$ph*12)|0)|0);
   HEAP8[$57>>0] = 1;
  }
  $60 = $encodings & 96;
  $61 = ($60|0)==(96);
  if ($61) {
   $62 = ((((((($mp)) + 32|0) + (($34*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 1|0);
   HEAP8[$62>>0] = 1;
  }
  $63 = $encodings & 4480;
  $64 = ($63|0)==(4480);
  if ($64) {
   $65 = ((((((($mp)) + 32|0) + (($34*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 2|0);
   HEAP8[$65>>0] = 1;
  }
  $66 = $encodings & 8448;
  $67 = ($66|0)==(8448);
  if ($67) {
   $68 = ((((((($mp)) + 32|0) + (($34*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 3|0);
   HEAP8[$68>>0] = 1;
  }
  $69 = $encodings & 20608;
  $70 = ($69|0)==(20608);
  if ($70) {
   $71 = ((((((($mp)) + 32|0) + (($34*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 4|0);
   HEAP8[$71>>0] = 1;
  }
  $72 = $encodings & 24576;
  $73 = ($72|0)==(24576);
  if ($73) {
   $74 = ((((((($mp)) + 32|0) + (($34*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 5|0);
   HEAP8[$74>>0] = 1;
  }
  $75 = $encodings & 512;
  $76 = ($75|0)==(0);
  if (!($76)) {
   $77 = ((((((($mp)) + 32|0) + (($34*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 6|0);
   HEAP8[$77>>0] = 1;
  }
  $78 = $encodings & 130;
  $79 = ($78|0)==(130);
  if ($79) {
   $80 = ((((((($mp)) + 32|0) + (($34*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 8|0);
   HEAP8[$80>>0] = 1;
  }
  $81 = $encodings & 1;
  $82 = ($81|0)==(0);
  if (!($82)) {
   $83 = ((((((($mp)) + 32|0) + (($34*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 9|0);
   HEAP8[$83>>0] = 1;
  }
  $84 = $encodings & 4;
  $85 = ($84|0)==(0);
  if (!($85)) {
   $86 = ((((((($mp)) + 32|0) + (($34*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 10|0);
   HEAP8[$86>>0] = 1;
  }
  $87 = $encodings & 8;
  $88 = ($87|0)==(0);
  if ($88) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
  $89 = ((((((($mp)) + 32|0) + (($34*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 11|0);
  HEAP8[$89>>0] = 1;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $37 = $encodings & 96;
 $38 = ($37|0)==(96);
 $39 = $encodings & 4480;
 $40 = ($39|0)==(4480);
 $41 = $encodings & 8448;
 $42 = ($41|0)==(8448);
 $43 = $encodings & 20608;
 $44 = ($43|0)==(20608);
 $45 = $encodings & 24576;
 $46 = ($45|0)==(24576);
 $47 = $encodings & 512;
 $48 = ($47|0)==(0);
 $49 = $encodings & 130;
 $50 = ($49|0)==(130);
 $51 = $encodings & 1;
 $52 = ($51|0)==(0);
 $53 = $encodings & 4;
 $54 = ($53|0)==(0);
 $55 = $encodings & 8;
 $56 = ($55|0)==(0);
 if ($36) {
  $58 = HEAP32[$ch>>2]|0;
  $59 = ((((($mp)) + 32|0) + (($58*120)|0)|0) + (($$0$i$ph*12)|0)|0);
  HEAP8[$59>>0] = 1;
 }
 if ($38) {
  $90 = HEAP32[$ch>>2]|0;
  $91 = ((((((($mp)) + 32|0) + (($90*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 1|0);
  HEAP8[$91>>0] = 1;
 }
 if ($40) {
  $92 = HEAP32[$ch>>2]|0;
  $93 = ((((((($mp)) + 32|0) + (($92*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 2|0);
  HEAP8[$93>>0] = 1;
 }
 if ($42) {
  $94 = HEAP32[$ch>>2]|0;
  $95 = ((((((($mp)) + 32|0) + (($94*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 3|0);
  HEAP8[$95>>0] = 1;
 }
 if ($44) {
  $96 = HEAP32[$ch>>2]|0;
  $97 = ((((((($mp)) + 32|0) + (($96*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 4|0);
  HEAP8[$97>>0] = 1;
 }
 if ($46) {
  $98 = HEAP32[$ch>>2]|0;
  $99 = ((((((($mp)) + 32|0) + (($98*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 5|0);
  HEAP8[$99>>0] = 1;
 }
 if (!($48)) {
  $100 = HEAP32[$ch>>2]|0;
  $101 = ((((((($mp)) + 32|0) + (($100*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 6|0);
  HEAP8[$101>>0] = 1;
 }
 if ($50) {
  $102 = HEAP32[$ch>>2]|0;
  $103 = ((((((($mp)) + 32|0) + (($102*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 8|0);
  HEAP8[$103>>0] = 1;
 }
 if (!($52)) {
  $104 = HEAP32[$ch>>2]|0;
  $105 = ((((((($mp)) + 32|0) + (($104*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 9|0);
  HEAP8[$105>>0] = 1;
 }
 if (!($54)) {
  $106 = HEAP32[$ch>>2]|0;
  $107 = ((((((($mp)) + 32|0) + (($106*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 10|0);
  HEAP8[$107>>0] = 1;
 }
 if (!($56)) {
  $108 = HEAP32[$ch>>2]|0;
  $109 = ((((((($mp)) + 32|0) + (($108*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 11|0);
  HEAP8[$109>>0] = 1;
 }
 $110 = ((($ch)) + 4|0);
 if ($36) {
  $111 = HEAP32[$110>>2]|0;
  $112 = ((((($mp)) + 32|0) + (($111*120)|0)|0) + (($$0$i$ph*12)|0)|0);
  HEAP8[$112>>0] = 1;
 }
 if ($38) {
  $113 = HEAP32[$110>>2]|0;
  $114 = ((((((($mp)) + 32|0) + (($113*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 1|0);
  HEAP8[$114>>0] = 1;
 }
 if ($40) {
  $115 = HEAP32[$110>>2]|0;
  $116 = ((((((($mp)) + 32|0) + (($115*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 2|0);
  HEAP8[$116>>0] = 1;
 }
 if ($42) {
  $117 = HEAP32[$110>>2]|0;
  $118 = ((((((($mp)) + 32|0) + (($117*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 3|0);
  HEAP8[$118>>0] = 1;
 }
 if ($44) {
  $119 = HEAP32[$110>>2]|0;
  $120 = ((((((($mp)) + 32|0) + (($119*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 4|0);
  HEAP8[$120>>0] = 1;
 }
 if ($46) {
  $121 = HEAP32[$110>>2]|0;
  $122 = ((((((($mp)) + 32|0) + (($121*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 5|0);
  HEAP8[$122>>0] = 1;
 }
 if (!($48)) {
  $123 = HEAP32[$110>>2]|0;
  $124 = ((((((($mp)) + 32|0) + (($123*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 6|0);
  HEAP8[$124>>0] = 1;
 }
 if ($50) {
  $125 = HEAP32[$110>>2]|0;
  $126 = ((((((($mp)) + 32|0) + (($125*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 8|0);
  HEAP8[$126>>0] = 1;
 }
 if (!($52)) {
  $127 = HEAP32[$110>>2]|0;
  $128 = ((((((($mp)) + 32|0) + (($127*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 9|0);
  HEAP8[$128>>0] = 1;
 }
 if (!($54)) {
  $129 = HEAP32[$110>>2]|0;
  $130 = ((((((($mp)) + 32|0) + (($129*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 10|0);
  HEAP8[$130>>0] = 1;
 }
 if ($56) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $131 = HEAP32[$110>>2]|0;
 $132 = ((((((($mp)) + 32|0) + (($131*120)|0)|0) + (($$0$i$ph*12)|0)|0)) + 11|0);
 HEAP8[$132>>0] = 1;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _samples_to_storage($fr,$s) {
 $fr = $fr|0;
 $s = $s|0;
 var $$0 = 0, $$pn = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27876|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 16384;
 $3 = ($2|0)==(0);
 $4 = ((($fr)) + 27884|0);
 $5 = HEAP32[$4>>2]|0;
 if ($3) {
  $7 = ((($fr)) + 27880|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = Math_imul($8, $s)|0;
  $$pn = $9;
  $$0 = Math_imul($$pn, $5)|0;
  return ($$0|0);
 } else {
  $6 = $s << 2;
  $$pn = $6;
  $$0 = Math_imul($$pn, $5)|0;
  return ($$0|0);
 }
 return (0)|0;
}
function _INT123_samples_to_bytes($fr,$s) {
 $fr = $fr|0;
 $s = $s|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27880|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = Math_imul($1, $s)|0;
 $3 = ((($fr)) + 27884|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = Math_imul($2, $4)|0;
 return ($5|0);
}
function _INT123_bytes_to_samples($fr,$b) {
 $fr = $fr|0;
 $b = $b|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27880|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($b|0) / ($1|0))&-1;
 $3 = ((($fr)) + 27884|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 return ($5|0);
}
function _postprocess_buffer($fr) {
 $fr = $fr|0;
 var $$cast$i = 0, $$cast$i1 = 0, $$lcssa$i = 0, $$lcssa$i5 = 0, $$lcssa31 = 0, $$pr = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $8 = 0, $9 = 0, $i$013 = 0, $i1$012 = 0, $rpos$04$i = 0, $rpos$04$i2 = 0, $switch$split12D = 0, $switch$split2D = 0, $wpos$0$lcssa$i = 0, $wpos$0$lcssa$i6 = 0, $wpos$03$i = 0, $wpos$03$i3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27876|0);
 $1 = HEAP32[$0>>2]|0;
 $switch$split2D = ($1|0)<(20608);
 L1: do {
  if ($switch$split2D) {
   switch ($1|0) {
   case 8448:  {
    label = 2;
    break;
   }
   default: {
    $63 = $1;
   }
   }
  } else {
   $switch$split12D = ($1|0)<(24576);
   if (!($switch$split12D)) {
    switch ($1|0) {
    case 24576:  {
     label = 2;
     break L1;
     break;
    }
    default: {
     $63 = $1;
     break L1;
    }
    }
   }
   switch ($1|0) {
   case 20608:  {
    break;
   }
   default: {
    $63 = $1;
    break L1;
   }
   }
   $39 = ((($fr)) + 27856|0);
   $40 = HEAP32[$39>>2]|0;
   $$cast$i1 = $40;
   $41 = ((($fr)) + 27864|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = ($42>>>0)<(4);
   if ($43) {
    $$lcssa$i5 = $$cast$i1;$wpos$0$lcssa$i6 = $40;
   } else {
    $rpos$04$i2 = $40;$wpos$03$i3 = $40;
    while(1) {
     $44 = ((($rpos$04$i2)) + 1|0);
     $45 = HEAP8[$44>>0]|0;
     HEAP8[$wpos$03$i3>>0] = $45;
     $46 = ((($wpos$03$i3)) + 1|0);
     $47 = ((($rpos$04$i2)) + 2|0);
     $48 = HEAP8[$47>>0]|0;
     HEAP8[$46>>0] = $48;
     $49 = ((($wpos$03$i3)) + 2|0);
     $50 = ((($rpos$04$i2)) + 3|0);
     $51 = HEAP8[$50>>0]|0;
     HEAP8[$49>>0] = $51;
     $52 = ((($wpos$03$i3)) + 3|0);
     $53 = ((($rpos$04$i2)) + 4|0);
     $54 = HEAP32[$39>>2]|0;
     $55 = $53;
     $56 = (($55) + 4)|0;
     $57 = (($56) - ($54))|0;
     $58 = HEAP32[$41>>2]|0;
     $59 = ($57>>>0)>($58>>>0);
     if ($59) {
      $$lcssa$i5 = $54;$wpos$0$lcssa$i6 = $52;
      break;
     } else {
      $rpos$04$i2 = $53;$wpos$03$i3 = $52;
     }
    }
   }
   $60 = $wpos$0$lcssa$i6;
   $61 = (($60) - ($$lcssa$i5))|0;
   HEAP32[$41>>2] = $61;
   label = 17;
  }
 } while(0);
 if ((label|0) == 2) {
  $2 = ((($fr)) + 27856|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($fr)) + 27864|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ($5>>>0)>(3);
  if ($6) {
   $i$013 = 0;
   while(1) {
    $7 = (($3) + ($i$013<<2)|0);
    $8 = HEAP32[$7>>2]|0;
    $9 = ($8|0)>(-1);
    do {
     if ($9) {
      $10 = $8 ^ -2147483648;
      HEAP32[$7>>2] = $10;
     } else {
      $11 = ($8|0)==(-2147483648);
      if ($11) {
       HEAP32[$7>>2] = 0;
       break;
      } else {
       $12 = $8 ^ -2147483648;
       HEAP32[$7>>2] = $12;
       break;
      }
     }
    } while(0);
    $13 = (($i$013) + 1)|0;
    $14 = HEAP32[$4>>2]|0;
    $15 = $14 >>> 2;
    $16 = ($13>>>0)<($15>>>0);
    if ($16) {
     $i$013 = $13;
    } else {
     $$lcssa31 = $14;
     break;
    }
   }
   $$pre = HEAP32[$0>>2]|0;
   $18 = $$pre;$20 = $$lcssa31;
  } else {
   $18 = $1;$20 = $5;
  }
  $17 = ($18|0)==(24576);
  if ($17) {
   $$cast$i = $3;
   $19 = ($20>>>0)<(4);
   if ($19) {
    $$lcssa$i = $$cast$i;$wpos$0$lcssa$i = $3;
   } else {
    $rpos$04$i = $3;$wpos$03$i = $3;
    while(1) {
     $21 = ((($rpos$04$i)) + 1|0);
     $22 = HEAP8[$21>>0]|0;
     HEAP8[$wpos$03$i>>0] = $22;
     $23 = ((($wpos$03$i)) + 1|0);
     $24 = ((($rpos$04$i)) + 2|0);
     $25 = HEAP8[$24>>0]|0;
     HEAP8[$23>>0] = $25;
     $26 = ((($wpos$03$i)) + 2|0);
     $27 = ((($rpos$04$i)) + 3|0);
     $28 = HEAP8[$27>>0]|0;
     HEAP8[$26>>0] = $28;
     $29 = ((($wpos$03$i)) + 3|0);
     $30 = ((($rpos$04$i)) + 4|0);
     $31 = HEAP32[$2>>2]|0;
     $32 = $30;
     $33 = (($32) + 4)|0;
     $34 = (($33) - ($31))|0;
     $35 = HEAP32[$4>>2]|0;
     $36 = ($34>>>0)>($35>>>0);
     if ($36) {
      $$lcssa$i = $31;$wpos$0$lcssa$i = $29;
      break;
     } else {
      $rpos$04$i = $30;$wpos$03$i = $29;
     }
    }
   }
   $37 = $wpos$0$lcssa$i;
   $38 = (($37) - ($$lcssa$i))|0;
   HEAP32[$4>>2] = $38;
   label = 17;
  } else {
   $63 = $18;
  }
 }
 if ((label|0) == 17) {
  $$pr = HEAP32[$0>>2]|0;
  $63 = $$pr;
 }
 $62 = ($63|0)==(96);
 if (!($62)) {
  return;
 }
 $64 = ((($fr)) + 27856|0);
 $65 = HEAP32[$64>>2]|0;
 $66 = ((($fr)) + 27864|0);
 $67 = HEAP32[$66>>2]|0;
 $68 = ($67>>>0)>(1);
 if (!($68)) {
  return;
 }
 $69 = $67 >>> 1;
 $i1$012 = 0;
 while(1) {
  $70 = (($65) + ($i1$012<<1)|0);
  $71 = HEAP16[$70>>1]|0;
  $72 = $71&65535;
  $73 = (($72) + 32768)|0;
  $74 = $73&65535;
  HEAP16[$70>>1] = $74;
  $75 = (($i1$012) + 1)|0;
  $76 = ($75>>>0)<($69>>>0);
  if ($76) {
   $i1$012 = $75;
  } else {
   break;
  }
 }
 return;
}
function _INT123_dct64($out0,$out1,$samples) {
 $out0 = $out0|0;
 $out1 = $out1|0;
 $samples = $samples|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $100 = 0, $101 = 0.0, $102 = 0.0, $103 = 0, $104 = 0, $105 = 0.0, $106 = 0, $107 = 0.0, $108 = 0.0, $109 = 0, $11 = 0.0, $110 = 0.0, $111 = 0.0, $112 = 0, $113 = 0, $114 = 0.0, $115 = 0;
 var $116 = 0.0, $117 = 0.0, $118 = 0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0, $123 = 0.0, $124 = 0, $125 = 0.0, $126 = 0.0, $127 = 0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0, $131 = 0, $132 = 0.0, $133 = 0;
 var $134 = 0.0, $135 = 0.0, $136 = 0, $137 = 0.0, $138 = 0.0, $139 = 0, $14 = 0.0, $140 = 0, $141 = 0.0, $142 = 0, $143 = 0.0, $144 = 0.0, $145 = 0, $146 = 0.0, $147 = 0.0, $148 = 0, $149 = 0, $15 = 0, $150 = 0.0, $151 = 0;
 var $152 = 0.0, $153 = 0.0, $154 = 0, $155 = 0.0, $156 = 0.0, $157 = 0, $158 = 0, $159 = 0.0, $16 = 0.0, $160 = 0, $161 = 0.0, $162 = 0.0, $163 = 0, $164 = 0.0, $165 = 0.0, $166 = 0, $167 = 0, $168 = 0.0, $169 = 0, $17 = 0.0;
 var $170 = 0.0, $171 = 0.0, $172 = 0, $173 = 0.0, $174 = 0.0, $175 = 0, $176 = 0, $177 = 0.0, $178 = 0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0, $182 = 0.0, $183 = 0.0, $184 = 0, $185 = 0, $186 = 0.0, $187 = 0, $188 = 0.0;
 var $189 = 0.0, $19 = 0, $190 = 0, $191 = 0.0, $192 = 0.0, $193 = 0, $194 = 0, $195 = 0.0, $196 = 0, $197 = 0.0, $198 = 0.0, $199 = 0, $2 = 0.0, $20 = 0.0, $200 = 0.0, $201 = 0.0, $202 = 0, $203 = 0, $204 = 0.0, $205 = 0;
 var $206 = 0.0, $207 = 0.0, $208 = 0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0, $213 = 0.0, $214 = 0, $215 = 0.0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0.0, $22 = 0.0, $220 = 0, $221 = 0, $222 = 0.0, $223 = 0;
 var $224 = 0.0, $225 = 0.0, $226 = 0, $227 = 0.0, $228 = 0.0, $229 = 0, $23 = 0.0, $230 = 0.0, $231 = 0.0, $232 = 0.0, $233 = 0.0, $234 = 0.0, $235 = 0, $236 = 0.0, $237 = 0, $238 = 0.0, $239 = 0.0, $24 = 0, $240 = 0, $241 = 0.0;
 var $242 = 0, $243 = 0.0, $244 = 0.0, $245 = 0, $246 = 0.0, $247 = 0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0, $251 = 0.0, $252 = 0, $253 = 0.0, $254 = 0.0, $255 = 0, $256 = 0.0, $257 = 0, $258 = 0.0, $259 = 0.0, $26 = 0.0;
 var $260 = 0, $261 = 0.0, $262 = 0, $263 = 0.0, $264 = 0.0, $265 = 0, $266 = 0.0, $267 = 0, $268 = 0.0, $269 = 0.0, $27 = 0, $270 = 0, $271 = 0.0, $272 = 0, $273 = 0.0, $274 = 0.0, $275 = 0, $276 = 0.0, $277 = 0, $278 = 0.0;
 var $279 = 0.0, $28 = 0.0, $280 = 0, $281 = 0.0, $282 = 0.0, $283 = 0, $284 = 0, $285 = 0.0, $286 = 0, $287 = 0.0, $288 = 0.0, $289 = 0, $29 = 0.0, $290 = 0.0, $291 = 0.0, $292 = 0, $293 = 0, $294 = 0.0, $295 = 0, $296 = 0.0;
 var $297 = 0.0, $298 = 0, $299 = 0.0, $3 = 0, $30 = 0, $300 = 0.0, $301 = 0, $302 = 0, $303 = 0.0, $304 = 0, $305 = 0.0, $306 = 0.0, $307 = 0, $308 = 0.0, $309 = 0.0, $31 = 0, $310 = 0, $311 = 0, $312 = 0.0, $313 = 0;
 var $314 = 0.0, $315 = 0.0, $316 = 0, $317 = 0.0, $318 = 0.0, $319 = 0, $32 = 0.0, $320 = 0, $321 = 0.0, $322 = 0, $323 = 0.0, $324 = 0.0, $325 = 0, $326 = 0.0, $327 = 0.0, $328 = 0, $329 = 0, $33 = 0, $330 = 0.0, $331 = 0;
 var $332 = 0.0, $333 = 0.0, $334 = 0, $335 = 0.0, $336 = 0.0, $337 = 0, $338 = 0.0, $339 = 0.0, $34 = 0.0, $340 = 0.0, $341 = 0.0, $342 = 0.0, $343 = 0, $344 = 0.0, $345 = 0, $346 = 0.0, $347 = 0.0, $348 = 0, $349 = 0, $35 = 0.0;
 var $350 = 0.0, $351 = 0, $352 = 0.0, $353 = 0.0, $354 = 0, $355 = 0, $356 = 0.0, $357 = 0, $358 = 0.0, $359 = 0.0, $36 = 0, $360 = 0, $361 = 0, $362 = 0.0, $363 = 0, $364 = 0.0, $365 = 0.0, $366 = 0, $367 = 0, $368 = 0.0;
 var $369 = 0, $37 = 0, $370 = 0.0, $371 = 0.0, $372 = 0, $373 = 0, $374 = 0.0, $375 = 0, $376 = 0.0, $377 = 0.0, $378 = 0, $379 = 0, $38 = 0.0, $380 = 0.0, $381 = 0, $382 = 0.0, $383 = 0.0, $384 = 0, $385 = 0.0, $386 = 0;
 var $387 = 0.0, $388 = 0.0, $389 = 0, $39 = 0, $390 = 0.0, $391 = 0, $392 = 0.0, $393 = 0.0, $394 = 0, $395 = 0.0, $396 = 0.0, $397 = 0, $398 = 0, $399 = 0.0, $4 = 0.0, $40 = 0.0, $400 = 0, $401 = 0.0, $402 = 0.0, $403 = 0;
 var $404 = 0.0, $405 = 0.0, $406 = 0, $407 = 0, $408 = 0.0, $409 = 0, $41 = 0.0, $410 = 0.0, $411 = 0.0, $412 = 0, $413 = 0.0, $414 = 0.0, $415 = 0, $416 = 0, $417 = 0.0, $418 = 0, $419 = 0.0, $42 = 0, $420 = 0.0, $421 = 0;
 var $422 = 0.0, $423 = 0.0, $424 = 0, $425 = 0, $426 = 0.0, $427 = 0, $428 = 0.0, $429 = 0.0, $43 = 0, $430 = 0, $431 = 0.0, $432 = 0.0, $433 = 0, $434 = 0, $435 = 0.0, $436 = 0, $437 = 0.0, $438 = 0.0, $439 = 0, $44 = 0.0;
 var $440 = 0.0, $441 = 0.0, $442 = 0, $443 = 0, $444 = 0.0, $445 = 0, $446 = 0.0, $447 = 0.0, $448 = 0, $449 = 0.0, $45 = 0, $450 = 0.0, $451 = 0, $452 = 0.0, $453 = 0, $454 = 0.0, $455 = 0.0, $456 = 0.0, $457 = 0.0, $458 = 0;
 var $459 = 0, $46 = 0.0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0.0, $465 = 0, $466 = 0.0, $467 = 0.0, $468 = 0, $469 = 0, $47 = 0.0, $470 = 0.0, $471 = 0, $472 = 0.0, $473 = 0.0, $474 = 0, $475 = 0, $476 = 0.0;
 var $477 = 0, $478 = 0.0, $479 = 0.0, $48 = 0, $480 = 0, $481 = 0.0, $482 = 0, $483 = 0.0, $484 = 0.0, $485 = 0, $486 = 0.0, $487 = 0, $488 = 0.0, $489 = 0.0, $49 = 0, $490 = 0.0, $491 = 0.0, $492 = 0, $493 = 0, $494 = 0.0;
 var $495 = 0, $496 = 0.0, $497 = 0.0, $498 = 0.0, $499 = 0.0, $5 = 0.0, $50 = 0.0, $500 = 0, $501 = 0, $502 = 0.0, $503 = 0, $504 = 0.0, $505 = 0.0, $506 = 0.0, $507 = 0.0, $508 = 0, $509 = 0, $51 = 0, $510 = 0.0, $511 = 0.0;
 var $512 = 0.0, $513 = 0.0, $514 = 0.0, $515 = 0, $516 = 0.0, $517 = 0, $518 = 0.0, $519 = 0.0, $52 = 0.0, $520 = 0, $521 = 0, $522 = 0.0, $523 = 0, $524 = 0.0, $525 = 0.0, $526 = 0, $527 = 0, $528 = 0.0, $529 = 0, $53 = 0.0;
 var $530 = 0.0, $531 = 0.0, $532 = 0, $533 = 0.0, $534 = 0, $535 = 0.0, $536 = 0.0, $537 = 0, $538 = 0.0, $539 = 0, $54 = 0, $540 = 0.0, $541 = 0.0, $542 = 0.0, $543 = 0.0, $544 = 0, $545 = 0, $546 = 0.0, $547 = 0, $548 = 0.0;
 var $549 = 0.0, $55 = 0, $550 = 0.0, $551 = 0.0, $552 = 0, $553 = 0, $554 = 0.0, $555 = 0, $556 = 0.0, $557 = 0.0, $558 = 0.0, $559 = 0.0, $56 = 0.0, $560 = 0, $561 = 0.0, $562 = 0.0, $563 = 0.0, $564 = 0.0, $565 = 0.0, $566 = 0;
 var $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0.0, $574 = 0, $575 = 0.0, $576 = 0.0, $577 = 0, $578 = 0, $579 = 0.0, $58 = 0.0, $580 = 0, $581 = 0.0, $582 = 0.0, $583 = 0, $584 = 0;
 var $585 = 0.0, $586 = 0, $587 = 0.0, $588 = 0.0, $589 = 0.0, $59 = 0.0, $590 = 0.0, $591 = 0, $592 = 0, $593 = 0.0, $594 = 0, $595 = 0.0, $596 = 0.0, $597 = 0.0, $598 = 0.0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0.0;
 var $602 = 0, $603 = 0.0, $604 = 0.0, $605 = 0, $606 = 0, $607 = 0.0, $608 = 0, $609 = 0.0, $61 = 0, $610 = 0.0, $611 = 0, $612 = 0, $613 = 0.0, $614 = 0, $615 = 0.0, $616 = 0.0, $617 = 0.0, $618 = 0.0, $619 = 0, $62 = 0.0;
 var $620 = 0, $621 = 0.0, $622 = 0.0, $623 = 0.0, $624 = 0.0, $625 = 0.0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0.0, $633 = 0, $634 = 0.0, $635 = 0.0, $636 = 0.0, $637 = 0.0, $638 = 0.0;
 var $639 = 0, $64 = 0.0, $640 = 0, $641 = 0.0, $642 = 0, $643 = 0.0, $644 = 0.0, $645 = 0, $646 = 0.0, $647 = 0.0, $648 = 0.0, $649 = 0, $65 = 0.0, $650 = 0, $651 = 0.0, $652 = 0, $653 = 0.0, $654 = 0.0, $655 = 0, $656 = 0.0;
 var $657 = 0.0, $658 = 0.0, $659 = 0, $66 = 0, $660 = 0, $661 = 0.0, $662 = 0, $663 = 0.0, $664 = 0.0, $665 = 0, $666 = 0.0, $667 = 0.0, $668 = 0.0, $669 = 0, $67 = 0, $670 = 0, $671 = 0.0, $672 = 0, $673 = 0.0, $674 = 0.0;
 var $675 = 0, $676 = 0.0, $677 = 0.0, $678 = 0.0, $679 = 0, $68 = 0.0, $680 = 0, $681 = 0.0, $682 = 0, $683 = 0.0, $684 = 0.0, $685 = 0, $686 = 0.0, $687 = 0.0, $688 = 0.0, $689 = 0, $69 = 0, $690 = 0, $691 = 0.0, $692 = 0;
 var $693 = 0.0, $694 = 0.0, $695 = 0, $696 = 0.0, $697 = 0.0, $698 = 0.0, $699 = 0, $7 = 0, $70 = 0.0, $700 = 0, $701 = 0.0, $702 = 0, $703 = 0.0, $704 = 0.0, $705 = 0, $706 = 0.0, $707 = 0.0, $708 = 0.0, $709 = 0, $71 = 0.0;
 var $710 = 0, $711 = 0.0, $712 = 0, $713 = 0.0, $714 = 0.0, $715 = 0, $716 = 0.0, $717 = 0.0, $718 = 0.0, $719 = 0, $72 = 0, $720 = 0, $721 = 0.0, $722 = 0, $723 = 0.0, $724 = 0.0, $725 = 0, $726 = 0.0, $727 = 0.0, $728 = 0.0;
 var $729 = 0, $73 = 0, $730 = 0, $731 = 0.0, $732 = 0, $733 = 0.0, $734 = 0.0, $735 = 0, $736 = 0.0, $737 = 0.0, $738 = 0.0, $739 = 0, $74 = 0.0, $740 = 0, $741 = 0.0, $742 = 0, $743 = 0.0, $744 = 0.0, $745 = 0, $746 = 0.0;
 var $747 = 0.0, $748 = 0.0, $749 = 0, $75 = 0, $750 = 0, $751 = 0.0, $752 = 0, $753 = 0.0, $754 = 0.0, $755 = 0, $756 = 0.0, $757 = 0.0, $758 = 0.0, $759 = 0, $76 = 0.0, $760 = 0, $761 = 0.0, $762 = 0, $763 = 0.0, $764 = 0.0;
 var $765 = 0, $766 = 0.0, $767 = 0.0, $768 = 0.0, $769 = 0, $77 = 0.0, $770 = 0, $771 = 0.0, $772 = 0, $773 = 0.0, $774 = 0.0, $775 = 0, $776 = 0.0, $777 = 0.0, $778 = 0.0, $779 = 0, $78 = 0, $780 = 0, $781 = 0.0, $782 = 0.0;
 var $783 = 0.0, $784 = 0, $785 = 0.0, $786 = 0.0, $787 = 0.0, $788 = 0.0, $789 = 0.0, $79 = 0, $790 = 0.0, $791 = 0.0, $792 = 0.0, $793 = 0.0, $794 = 0.0, $795 = 0.0, $796 = 0.0, $797 = 0.0, $798 = 0.0, $799 = 0.0, $8 = 0.0, $80 = 0.0;
 var $800 = 0.0, $801 = 0.0, $802 = 0.0, $803 = 0.0, $804 = 0.0, $805 = 0.0, $806 = 0.0, $807 = 0.0, $808 = 0.0, $809 = 0.0, $81 = 0, $810 = 0.0, $811 = 0.0, $812 = 0, $813 = 0.0, $814 = 0.0, $815 = 0.0, $816 = 0, $817 = 0.0, $818 = 0.0;
 var $819 = 0, $82 = 0.0, $820 = 0.0, $821 = 0.0, $822 = 0, $823 = 0, $824 = 0.0, $825 = 0, $826 = 0.0, $827 = 0.0, $828 = 0, $829 = 0.0, $83 = 0.0, $830 = 0.0, $831 = 0, $832 = 0.0, $833 = 0.0, $834 = 0, $835 = 0.0, $836 = 0;
 var $837 = 0.0, $838 = 0.0, $839 = 0, $84 = 0, $840 = 0.0, $841 = 0.0, $842 = 0, $843 = 0.0, $844 = 0.0, $845 = 0, $846 = 0.0, $847 = 0.0, $848 = 0, $849 = 0.0, $85 = 0, $850 = 0.0, $851 = 0, $852 = 0.0, $853 = 0.0, $854 = 0;
 var $855 = 0.0, $856 = 0.0, $857 = 0, $858 = 0.0, $859 = 0.0, $86 = 0.0, $860 = 0, $861 = 0.0, $862 = 0, $863 = 0.0, $864 = 0.0, $865 = 0, $866 = 0.0, $867 = 0, $868 = 0.0, $869 = 0.0, $87 = 0, $870 = 0.0, $871 = 0, $872 = 0;
 var $873 = 0.0, $874 = 0.0, $875 = 0, $876 = 0.0, $877 = 0.0, $878 = 0, $879 = 0.0, $88 = 0.0, $880 = 0, $881 = 0.0, $882 = 0.0, $883 = 0, $884 = 0.0, $885 = 0, $886 = 0.0, $887 = 0.0, $888 = 0.0, $889 = 0, $89 = 0.0, $890 = 0;
 var $891 = 0.0, $892 = 0.0, $893 = 0, $894 = 0, $895 = 0, $896 = 0.0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0.0, $901 = 0.0, $902 = 0, $903 = 0, $904 = 0, $905 = 0.0, $906 = 0.0, $907 = 0, $908 = 0;
 var $909 = 0, $91 = 0.0, $910 = 0, $911 = 0.0, $912 = 0, $913 = 0, $914 = 0, $915 = 0.0, $916 = 0.0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0.0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0;
 var $927 = 0.0, $928 = 0.0, $929 = 0, $93 = 0.0, $930 = 0, $931 = 0, $932 = 0, $933 = 0.0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0.0, $939 = 0, $94 = 0.0, $940 = 0.0, $941 = 0.0, $942 = 0, $943 = 0, $944 = 0;
 var $945 = 0, $946 = 0, $947 = 0.0, $948 = 0.0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0.0, $955 = 0.0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0.0, $960 = 0, $961 = 0.0, $962 = 0.0;
 var $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0.0, $969 = 0.0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0.0, $976 = 0.0, $977 = 0, $978 = 0, $979 = 0, $98 = 0.0, $980 = 0;
 var $981 = 0, $99 = 0.0, $b1$1118 = 0, $b1$638 = 0, $b2$1119 = 0, $b2$639 = 0, $bs$1120 = 0, $bs$640 = 0, $bufs = 0, $j$041 = 0, $j$121 = 0, $scevgep100 = 0, $scevgep101 = 0, $scevgep102 = 0, $scevgep104 = 0, $scevgep105 = 0, $scevgep107 = 0, $scevgep109 = 0, $scevgep110 = 0, $scevgep111 = 0;
 var $scevgep114 = 0, $scevgep115 = 0, $scevgep116 = 0, $scevgep117 = 0, $scevgep118 = 0, $scevgep92 = 0, $scevgep94 = 0, $scevgep95 = 0, $scevgep96 = 0, $scevgep98 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $bufs = sp;
 $0 = HEAP32[8828]|0;
 $scevgep117 = ((($bufs)) + 64|0);
 $1 = ((($samples)) + 4|0);
 $2 = +HEAPF32[$samples>>2];
 $3 = ((($samples)) + 124|0);
 $4 = +HEAPF32[$3>>2];
 $5 = $4 + $2;
 $6 = ((($bufs)) + 4|0);
 HEAPF32[$bufs>>2] = $5;
 $7 = ((($samples)) + 8|0);
 $8 = +HEAPF32[$1>>2];
 $9 = ((($samples)) + 120|0);
 $10 = +HEAPF32[$9>>2];
 $11 = $10 + $8;
 $12 = ((($bufs)) + 8|0);
 HEAPF32[$6>>2] = $11;
 $13 = ((($samples)) + 12|0);
 $14 = +HEAPF32[$7>>2];
 $15 = ((($samples)) + 116|0);
 $16 = +HEAPF32[$15>>2];
 $17 = $16 + $14;
 $18 = ((($bufs)) + 12|0);
 HEAPF32[$12>>2] = $17;
 $19 = ((($samples)) + 16|0);
 $20 = +HEAPF32[$13>>2];
 $21 = ((($samples)) + 112|0);
 $22 = +HEAPF32[$21>>2];
 $23 = $22 + $20;
 $24 = ((($bufs)) + 16|0);
 HEAPF32[$18>>2] = $23;
 $25 = ((($samples)) + 20|0);
 $26 = +HEAPF32[$19>>2];
 $27 = ((($samples)) + 108|0);
 $28 = +HEAPF32[$27>>2];
 $29 = $28 + $26;
 $30 = ((($bufs)) + 20|0);
 HEAPF32[$24>>2] = $29;
 $31 = ((($samples)) + 24|0);
 $32 = +HEAPF32[$25>>2];
 $33 = ((($samples)) + 104|0);
 $34 = +HEAPF32[$33>>2];
 $35 = $34 + $32;
 $36 = ((($bufs)) + 24|0);
 HEAPF32[$30>>2] = $35;
 $37 = ((($samples)) + 28|0);
 $38 = +HEAPF32[$31>>2];
 $39 = ((($samples)) + 100|0);
 $40 = +HEAPF32[$39>>2];
 $41 = $40 + $38;
 $42 = ((($bufs)) + 28|0);
 HEAPF32[$36>>2] = $41;
 $43 = ((($samples)) + 32|0);
 $44 = +HEAPF32[$37>>2];
 $45 = ((($samples)) + 96|0);
 $46 = +HEAPF32[$45>>2];
 $47 = $46 + $44;
 $48 = ((($bufs)) + 32|0);
 HEAPF32[$42>>2] = $47;
 $49 = ((($samples)) + 36|0);
 $50 = +HEAPF32[$43>>2];
 $51 = ((($samples)) + 92|0);
 $52 = +HEAPF32[$51>>2];
 $53 = $52 + $50;
 $54 = ((($bufs)) + 36|0);
 HEAPF32[$48>>2] = $53;
 $55 = ((($samples)) + 40|0);
 $56 = +HEAPF32[$49>>2];
 $57 = ((($samples)) + 88|0);
 $58 = +HEAPF32[$57>>2];
 $59 = $58 + $56;
 $60 = ((($bufs)) + 40|0);
 HEAPF32[$54>>2] = $59;
 $61 = ((($samples)) + 44|0);
 $62 = +HEAPF32[$55>>2];
 $63 = ((($samples)) + 84|0);
 $64 = +HEAPF32[$63>>2];
 $65 = $64 + $62;
 $66 = ((($bufs)) + 44|0);
 HEAPF32[$60>>2] = $65;
 $67 = ((($samples)) + 48|0);
 $68 = +HEAPF32[$61>>2];
 $69 = ((($samples)) + 80|0);
 $70 = +HEAPF32[$69>>2];
 $71 = $70 + $68;
 $72 = ((($bufs)) + 48|0);
 HEAPF32[$66>>2] = $71;
 $73 = ((($samples)) + 52|0);
 $74 = +HEAPF32[$67>>2];
 $75 = ((($samples)) + 76|0);
 $76 = +HEAPF32[$75>>2];
 $77 = $76 + $74;
 $78 = ((($bufs)) + 52|0);
 HEAPF32[$72>>2] = $77;
 $79 = ((($samples)) + 56|0);
 $80 = +HEAPF32[$73>>2];
 $81 = ((($samples)) + 72|0);
 $82 = +HEAPF32[$81>>2];
 $83 = $82 + $80;
 $84 = ((($bufs)) + 56|0);
 HEAPF32[$78>>2] = $83;
 $85 = ((($samples)) + 60|0);
 $86 = +HEAPF32[$79>>2];
 $87 = ((($samples)) + 68|0);
 $88 = +HEAPF32[$87>>2];
 $89 = $88 + $86;
 $90 = ((($bufs)) + 60|0);
 HEAPF32[$84>>2] = $89;
 $91 = +HEAPF32[$85>>2];
 $92 = ((($samples)) + 64|0);
 $93 = +HEAPF32[$92>>2];
 $94 = $93 + $91;
 HEAPF32[$90>>2] = $94;
 $scevgep118 = ((($samples)) + 64|0);
 $95 = ((($samples)) + 60|0);
 $96 = +HEAPF32[$95>>2];
 $97 = ((($samples)) + 68|0);
 $98 = +HEAPF32[$scevgep118>>2];
 $99 = $96 - $98;
 $100 = ((($0)) + 60|0);
 $101 = +HEAPF32[$100>>2];
 $102 = $99 * $101;
 $103 = ((($bufs)) + 68|0);
 HEAPF32[$scevgep117>>2] = $102;
 $104 = ((($samples)) + 56|0);
 $105 = +HEAPF32[$104>>2];
 $106 = ((($samples)) + 72|0);
 $107 = +HEAPF32[$97>>2];
 $108 = $105 - $107;
 $109 = ((($0)) + 56|0);
 $110 = +HEAPF32[$109>>2];
 $111 = $108 * $110;
 $112 = ((($bufs)) + 72|0);
 HEAPF32[$103>>2] = $111;
 $113 = ((($samples)) + 52|0);
 $114 = +HEAPF32[$113>>2];
 $115 = ((($samples)) + 76|0);
 $116 = +HEAPF32[$106>>2];
 $117 = $114 - $116;
 $118 = ((($0)) + 52|0);
 $119 = +HEAPF32[$118>>2];
 $120 = $117 * $119;
 $121 = ((($bufs)) + 76|0);
 HEAPF32[$112>>2] = $120;
 $122 = ((($samples)) + 48|0);
 $123 = +HEAPF32[$122>>2];
 $124 = ((($samples)) + 80|0);
 $125 = +HEAPF32[$115>>2];
 $126 = $123 - $125;
 $127 = ((($0)) + 48|0);
 $128 = +HEAPF32[$127>>2];
 $129 = $126 * $128;
 $130 = ((($bufs)) + 80|0);
 HEAPF32[$121>>2] = $129;
 $131 = ((($samples)) + 44|0);
 $132 = +HEAPF32[$131>>2];
 $133 = ((($samples)) + 84|0);
 $134 = +HEAPF32[$124>>2];
 $135 = $132 - $134;
 $136 = ((($0)) + 44|0);
 $137 = +HEAPF32[$136>>2];
 $138 = $135 * $137;
 $139 = ((($bufs)) + 84|0);
 HEAPF32[$130>>2] = $138;
 $140 = ((($samples)) + 40|0);
 $141 = +HEAPF32[$140>>2];
 $142 = ((($samples)) + 88|0);
 $143 = +HEAPF32[$133>>2];
 $144 = $141 - $143;
 $145 = ((($0)) + 40|0);
 $146 = +HEAPF32[$145>>2];
 $147 = $144 * $146;
 $148 = ((($bufs)) + 88|0);
 HEAPF32[$139>>2] = $147;
 $149 = ((($samples)) + 36|0);
 $150 = +HEAPF32[$149>>2];
 $151 = ((($samples)) + 92|0);
 $152 = +HEAPF32[$142>>2];
 $153 = $150 - $152;
 $154 = ((($0)) + 36|0);
 $155 = +HEAPF32[$154>>2];
 $156 = $153 * $155;
 $157 = ((($bufs)) + 92|0);
 HEAPF32[$148>>2] = $156;
 $158 = ((($samples)) + 32|0);
 $159 = +HEAPF32[$158>>2];
 $160 = ((($samples)) + 96|0);
 $161 = +HEAPF32[$151>>2];
 $162 = $159 - $161;
 $163 = ((($0)) + 32|0);
 $164 = +HEAPF32[$163>>2];
 $165 = $162 * $164;
 $166 = ((($bufs)) + 96|0);
 HEAPF32[$157>>2] = $165;
 $167 = ((($samples)) + 28|0);
 $168 = +HEAPF32[$167>>2];
 $169 = ((($samples)) + 100|0);
 $170 = +HEAPF32[$160>>2];
 $171 = $168 - $170;
 $172 = ((($0)) + 28|0);
 $173 = +HEAPF32[$172>>2];
 $174 = $171 * $173;
 $175 = ((($bufs)) + 100|0);
 HEAPF32[$166>>2] = $174;
 $176 = ((($samples)) + 24|0);
 $177 = +HEAPF32[$176>>2];
 $178 = ((($samples)) + 104|0);
 $179 = +HEAPF32[$169>>2];
 $180 = $177 - $179;
 $181 = ((($0)) + 24|0);
 $182 = +HEAPF32[$181>>2];
 $183 = $180 * $182;
 $184 = ((($bufs)) + 104|0);
 HEAPF32[$175>>2] = $183;
 $185 = ((($samples)) + 20|0);
 $186 = +HEAPF32[$185>>2];
 $187 = ((($samples)) + 108|0);
 $188 = +HEAPF32[$178>>2];
 $189 = $186 - $188;
 $190 = ((($0)) + 20|0);
 $191 = +HEAPF32[$190>>2];
 $192 = $189 * $191;
 $193 = ((($bufs)) + 108|0);
 HEAPF32[$184>>2] = $192;
 $194 = ((($samples)) + 16|0);
 $195 = +HEAPF32[$194>>2];
 $196 = ((($samples)) + 112|0);
 $197 = +HEAPF32[$187>>2];
 $198 = $195 - $197;
 $199 = ((($0)) + 16|0);
 $200 = +HEAPF32[$199>>2];
 $201 = $198 * $200;
 $202 = ((($bufs)) + 112|0);
 HEAPF32[$193>>2] = $201;
 $203 = ((($samples)) + 12|0);
 $204 = +HEAPF32[$203>>2];
 $205 = ((($samples)) + 116|0);
 $206 = +HEAPF32[$196>>2];
 $207 = $204 - $206;
 $208 = ((($0)) + 12|0);
 $209 = +HEAPF32[$208>>2];
 $210 = $207 * $209;
 $211 = ((($bufs)) + 116|0);
 HEAPF32[$202>>2] = $210;
 $212 = ((($samples)) + 8|0);
 $213 = +HEAPF32[$212>>2];
 $214 = ((($samples)) + 120|0);
 $215 = +HEAPF32[$205>>2];
 $216 = $213 - $215;
 $217 = ((($0)) + 8|0);
 $218 = +HEAPF32[$217>>2];
 $219 = $216 * $218;
 $220 = ((($bufs)) + 120|0);
 HEAPF32[$211>>2] = $219;
 $221 = ((($samples)) + 4|0);
 $222 = +HEAPF32[$221>>2];
 $223 = ((($samples)) + 124|0);
 $224 = +HEAPF32[$214>>2];
 $225 = $222 - $224;
 $226 = ((($0)) + 4|0);
 $227 = +HEAPF32[$226>>2];
 $228 = $225 * $227;
 $229 = ((($bufs)) + 124|0);
 HEAPF32[$220>>2] = $228;
 $230 = +HEAPF32[$samples>>2];
 $231 = +HEAPF32[$223>>2];
 $232 = $230 - $231;
 $233 = +HEAPF32[$0>>2];
 $234 = $232 * $233;
 HEAPF32[$229>>2] = $234;
 $scevgep116 = ((($bufs)) + 128|0);
 $235 = HEAP32[(35316)>>2]|0;
 $scevgep114 = ((($bufs)) + 160|0);
 $236 = +HEAPF32[$bufs>>2];
 $237 = ((($bufs)) + 60|0);
 $238 = +HEAPF32[$237>>2];
 $239 = $238 + $236;
 $240 = ((($bufs)) + 132|0);
 HEAPF32[$scevgep116>>2] = $239;
 $241 = +HEAPF32[$6>>2];
 $242 = ((($bufs)) + 56|0);
 $243 = +HEAPF32[$242>>2];
 $244 = $243 + $241;
 $245 = ((($bufs)) + 136|0);
 HEAPF32[$240>>2] = $244;
 $246 = +HEAPF32[$12>>2];
 $247 = ((($bufs)) + 52|0);
 $248 = +HEAPF32[$247>>2];
 $249 = $248 + $246;
 $250 = ((($bufs)) + 140|0);
 HEAPF32[$245>>2] = $249;
 $251 = +HEAPF32[$18>>2];
 $252 = ((($bufs)) + 48|0);
 $253 = +HEAPF32[$252>>2];
 $254 = $253 + $251;
 $255 = ((($bufs)) + 144|0);
 HEAPF32[$250>>2] = $254;
 $256 = +HEAPF32[$24>>2];
 $257 = ((($bufs)) + 44|0);
 $258 = +HEAPF32[$257>>2];
 $259 = $258 + $256;
 $260 = ((($bufs)) + 148|0);
 HEAPF32[$255>>2] = $259;
 $261 = +HEAPF32[$30>>2];
 $262 = ((($bufs)) + 40|0);
 $263 = +HEAPF32[$262>>2];
 $264 = $263 + $261;
 $265 = ((($bufs)) + 152|0);
 HEAPF32[$260>>2] = $264;
 $266 = +HEAPF32[$36>>2];
 $267 = ((($bufs)) + 36|0);
 $268 = +HEAPF32[$267>>2];
 $269 = $268 + $266;
 $270 = ((($bufs)) + 156|0);
 HEAPF32[$265>>2] = $269;
 $271 = +HEAPF32[$42>>2];
 $272 = ((($bufs)) + 32|0);
 $273 = +HEAPF32[$272>>2];
 $274 = $273 + $271;
 HEAPF32[$270>>2] = $274;
 $scevgep115 = ((($bufs)) + 32|0);
 $scevgep110 = ((($bufs)) + 192|0);
 $275 = ((($bufs)) + 28|0);
 $276 = +HEAPF32[$275>>2];
 $277 = ((($bufs)) + 36|0);
 $278 = +HEAPF32[$scevgep115>>2];
 $279 = $276 - $278;
 $280 = ((($235)) + 28|0);
 $281 = +HEAPF32[$280>>2];
 $282 = $279 * $281;
 $283 = ((($bufs)) + 164|0);
 HEAPF32[$scevgep114>>2] = $282;
 $284 = ((($bufs)) + 24|0);
 $285 = +HEAPF32[$284>>2];
 $286 = ((($bufs)) + 40|0);
 $287 = +HEAPF32[$277>>2];
 $288 = $285 - $287;
 $289 = ((($235)) + 24|0);
 $290 = +HEAPF32[$289>>2];
 $291 = $288 * $290;
 $292 = ((($bufs)) + 168|0);
 HEAPF32[$283>>2] = $291;
 $293 = ((($bufs)) + 20|0);
 $294 = +HEAPF32[$293>>2];
 $295 = ((($bufs)) + 44|0);
 $296 = +HEAPF32[$286>>2];
 $297 = $294 - $296;
 $298 = ((($235)) + 20|0);
 $299 = +HEAPF32[$298>>2];
 $300 = $297 * $299;
 $301 = ((($bufs)) + 172|0);
 HEAPF32[$292>>2] = $300;
 $302 = ((($bufs)) + 16|0);
 $303 = +HEAPF32[$302>>2];
 $304 = ((($bufs)) + 48|0);
 $305 = +HEAPF32[$295>>2];
 $306 = $303 - $305;
 $307 = ((($235)) + 16|0);
 $308 = +HEAPF32[$307>>2];
 $309 = $306 * $308;
 $310 = ((($bufs)) + 176|0);
 HEAPF32[$301>>2] = $309;
 $311 = ((($bufs)) + 12|0);
 $312 = +HEAPF32[$311>>2];
 $313 = ((($bufs)) + 52|0);
 $314 = +HEAPF32[$304>>2];
 $315 = $312 - $314;
 $316 = ((($235)) + 12|0);
 $317 = +HEAPF32[$316>>2];
 $318 = $315 * $317;
 $319 = ((($bufs)) + 180|0);
 HEAPF32[$310>>2] = $318;
 $320 = ((($bufs)) + 8|0);
 $321 = +HEAPF32[$320>>2];
 $322 = ((($bufs)) + 56|0);
 $323 = +HEAPF32[$313>>2];
 $324 = $321 - $323;
 $325 = ((($235)) + 8|0);
 $326 = +HEAPF32[$325>>2];
 $327 = $324 * $326;
 $328 = ((($bufs)) + 184|0);
 HEAPF32[$319>>2] = $327;
 $329 = ((($bufs)) + 4|0);
 $330 = +HEAPF32[$329>>2];
 $331 = ((($bufs)) + 60|0);
 $332 = +HEAPF32[$322>>2];
 $333 = $330 - $332;
 $334 = ((($235)) + 4|0);
 $335 = +HEAPF32[$334>>2];
 $336 = $333 * $335;
 $337 = ((($bufs)) + 188|0);
 HEAPF32[$328>>2] = $336;
 $338 = +HEAPF32[$bufs>>2];
 $339 = +HEAPF32[$331>>2];
 $340 = $338 - $339;
 $341 = +HEAPF32[$235>>2];
 $342 = $340 * $341;
 HEAPF32[$337>>2] = $342;
 $scevgep111 = ((($bufs)) + 64|0);
 $343 = ((($bufs)) + 68|0);
 $344 = +HEAPF32[$scevgep111>>2];
 $345 = ((($bufs)) + 124|0);
 $346 = +HEAPF32[$345>>2];
 $347 = $346 + $344;
 $348 = ((($bufs)) + 196|0);
 HEAPF32[$scevgep110>>2] = $347;
 $349 = ((($bufs)) + 72|0);
 $350 = +HEAPF32[$343>>2];
 $351 = ((($bufs)) + 120|0);
 $352 = +HEAPF32[$351>>2];
 $353 = $352 + $350;
 $354 = ((($bufs)) + 200|0);
 HEAPF32[$348>>2] = $353;
 $355 = ((($bufs)) + 76|0);
 $356 = +HEAPF32[$349>>2];
 $357 = ((($bufs)) + 116|0);
 $358 = +HEAPF32[$357>>2];
 $359 = $358 + $356;
 $360 = ((($bufs)) + 204|0);
 HEAPF32[$354>>2] = $359;
 $361 = ((($bufs)) + 80|0);
 $362 = +HEAPF32[$355>>2];
 $363 = ((($bufs)) + 112|0);
 $364 = +HEAPF32[$363>>2];
 $365 = $364 + $362;
 $366 = ((($bufs)) + 208|0);
 HEAPF32[$360>>2] = $365;
 $367 = ((($bufs)) + 84|0);
 $368 = +HEAPF32[$361>>2];
 $369 = ((($bufs)) + 108|0);
 $370 = +HEAPF32[$369>>2];
 $371 = $370 + $368;
 $372 = ((($bufs)) + 212|0);
 HEAPF32[$366>>2] = $371;
 $373 = ((($bufs)) + 88|0);
 $374 = +HEAPF32[$367>>2];
 $375 = ((($bufs)) + 104|0);
 $376 = +HEAPF32[$375>>2];
 $377 = $376 + $374;
 $378 = ((($bufs)) + 216|0);
 HEAPF32[$372>>2] = $377;
 $379 = ((($bufs)) + 92|0);
 $380 = +HEAPF32[$373>>2];
 $381 = ((($bufs)) + 100|0);
 $382 = +HEAPF32[$381>>2];
 $383 = $382 + $380;
 $384 = ((($bufs)) + 220|0);
 HEAPF32[$378>>2] = $383;
 $385 = +HEAPF32[$379>>2];
 $386 = ((($bufs)) + 96|0);
 $387 = +HEAPF32[$386>>2];
 $388 = $387 + $385;
 HEAPF32[$384>>2] = $388;
 $scevgep107 = ((($bufs)) + 224|0);
 $scevgep109 = ((($bufs)) + 96|0);
 $scevgep105 = ((($bufs)) + 128|0);
 $389 = ((($bufs)) + 100|0);
 $390 = +HEAPF32[$scevgep109>>2];
 $391 = ((($bufs)) + 92|0);
 $392 = +HEAPF32[$391>>2];
 $393 = $390 - $392;
 $394 = ((($235)) + 28|0);
 $395 = +HEAPF32[$394>>2];
 $396 = $393 * $395;
 $397 = ((($bufs)) + 228|0);
 HEAPF32[$scevgep107>>2] = $396;
 $398 = ((($bufs)) + 104|0);
 $399 = +HEAPF32[$389>>2];
 $400 = ((($bufs)) + 88|0);
 $401 = +HEAPF32[$400>>2];
 $402 = $399 - $401;
 $403 = ((($235)) + 24|0);
 $404 = +HEAPF32[$403>>2];
 $405 = $402 * $404;
 $406 = ((($bufs)) + 232|0);
 HEAPF32[$397>>2] = $405;
 $407 = ((($bufs)) + 108|0);
 $408 = +HEAPF32[$398>>2];
 $409 = ((($bufs)) + 84|0);
 $410 = +HEAPF32[$409>>2];
 $411 = $408 - $410;
 $412 = ((($235)) + 20|0);
 $413 = +HEAPF32[$412>>2];
 $414 = $411 * $413;
 $415 = ((($bufs)) + 236|0);
 HEAPF32[$406>>2] = $414;
 $416 = ((($bufs)) + 112|0);
 $417 = +HEAPF32[$407>>2];
 $418 = ((($bufs)) + 80|0);
 $419 = +HEAPF32[$418>>2];
 $420 = $417 - $419;
 $421 = ((($235)) + 16|0);
 $422 = +HEAPF32[$421>>2];
 $423 = $420 * $422;
 $424 = ((($bufs)) + 240|0);
 HEAPF32[$415>>2] = $423;
 $425 = ((($bufs)) + 116|0);
 $426 = +HEAPF32[$416>>2];
 $427 = ((($bufs)) + 76|0);
 $428 = +HEAPF32[$427>>2];
 $429 = $426 - $428;
 $430 = ((($235)) + 12|0);
 $431 = +HEAPF32[$430>>2];
 $432 = $429 * $431;
 $433 = ((($bufs)) + 244|0);
 HEAPF32[$424>>2] = $432;
 $434 = ((($bufs)) + 120|0);
 $435 = +HEAPF32[$425>>2];
 $436 = ((($bufs)) + 72|0);
 $437 = +HEAPF32[$436>>2];
 $438 = $435 - $437;
 $439 = ((($235)) + 8|0);
 $440 = +HEAPF32[$439>>2];
 $441 = $438 * $440;
 $442 = ((($bufs)) + 248|0);
 HEAPF32[$433>>2] = $441;
 $443 = ((($bufs)) + 124|0);
 $444 = +HEAPF32[$434>>2];
 $445 = ((($bufs)) + 68|0);
 $446 = +HEAPF32[$445>>2];
 $447 = $444 - $446;
 $448 = ((($235)) + 4|0);
 $449 = +HEAPF32[$448>>2];
 $450 = $447 * $449;
 $451 = ((($bufs)) + 252|0);
 HEAPF32[$442>>2] = $450;
 $452 = +HEAPF32[$443>>2];
 $453 = ((($bufs)) + 64|0);
 $454 = +HEAPF32[$453>>2];
 $455 = $452 - $454;
 $456 = +HEAPF32[$235>>2];
 $457 = $455 * $456;
 HEAPF32[$451>>2] = $457;
 $458 = HEAP32[(35320)>>2]|0;
 $459 = ((($bufs)) + 160|0);
 $460 = ((($458)) + 12|0);
 $461 = ((($458)) + 8|0);
 $462 = ((($458)) + 4|0);
 $b1$638 = $scevgep105;$b2$639 = $459;$bs$640 = $bufs;$j$041 = 2;
 while(1) {
  $463 = ((($b1$638)) + 4|0);
  $464 = +HEAPF32[$b1$638>>2];
  $465 = ((($b2$639)) + -4|0);
  $466 = +HEAPF32[$465>>2];
  $467 = $466 + $464;
  $468 = ((($bs$640)) + 4|0);
  HEAPF32[$bs$640>>2] = $467;
  $469 = ((($b1$638)) + 8|0);
  $470 = +HEAPF32[$463>>2];
  $471 = ((($b2$639)) + -8|0);
  $472 = +HEAPF32[$471>>2];
  $473 = $472 + $470;
  $474 = ((($bs$640)) + 8|0);
  HEAPF32[$468>>2] = $473;
  $475 = ((($b1$638)) + 12|0);
  $476 = +HEAPF32[$469>>2];
  $477 = ((($b2$639)) + -12|0);
  $478 = +HEAPF32[$477>>2];
  $479 = $478 + $476;
  $480 = ((($bs$640)) + 12|0);
  HEAPF32[$474>>2] = $479;
  $481 = +HEAPF32[$475>>2];
  $482 = ((($b2$639)) + -16|0);
  $483 = +HEAPF32[$482>>2];
  $484 = $483 + $481;
  HEAPF32[$480>>2] = $484;
  $scevgep92 = ((($bs$640)) + 16|0);
  $scevgep94 = ((($b1$638)) + 16|0);
  $scevgep96 = ((($b1$638)) + 32|0);
  $485 = ((($b2$639)) + -20|0);
  $486 = +HEAPF32[$485>>2];
  $487 = ((($b1$638)) + 20|0);
  $488 = +HEAPF32[$scevgep94>>2];
  $489 = $486 - $488;
  $490 = +HEAPF32[$460>>2];
  $491 = $489 * $490;
  $492 = ((($bs$640)) + 20|0);
  HEAPF32[$scevgep92>>2] = $491;
  $493 = ((($b2$639)) + -24|0);
  $494 = +HEAPF32[$493>>2];
  $495 = ((($b1$638)) + 24|0);
  $496 = +HEAPF32[$487>>2];
  $497 = $494 - $496;
  $498 = +HEAPF32[$461>>2];
  $499 = $497 * $498;
  $500 = ((($bs$640)) + 24|0);
  HEAPF32[$492>>2] = $499;
  $501 = ((($b2$639)) + -28|0);
  $502 = +HEAPF32[$501>>2];
  $503 = ((($b1$638)) + 28|0);
  $504 = +HEAPF32[$495>>2];
  $505 = $502 - $504;
  $506 = +HEAPF32[$462>>2];
  $507 = $505 * $506;
  $508 = ((($bs$640)) + 28|0);
  HEAPF32[$500>>2] = $507;
  $509 = ((($b2$639)) + -32|0);
  $510 = +HEAPF32[$509>>2];
  $511 = +HEAPF32[$503>>2];
  $512 = $510 - $511;
  $513 = +HEAPF32[$458>>2];
  $514 = $512 * $513;
  HEAPF32[$508>>2] = $514;
  $scevgep95 = ((($bs$640)) + 32|0);
  $515 = ((($b1$638)) + 36|0);
  $516 = +HEAPF32[$scevgep96>>2];
  $517 = ((($b2$639)) + 28|0);
  $518 = +HEAPF32[$517>>2];
  $519 = $518 + $516;
  $520 = ((($bs$640)) + 36|0);
  HEAPF32[$scevgep95>>2] = $519;
  $521 = ((($b1$638)) + 40|0);
  $522 = +HEAPF32[$515>>2];
  $523 = ((($b2$639)) + 24|0);
  $524 = +HEAPF32[$523>>2];
  $525 = $524 + $522;
  $526 = ((($bs$640)) + 40|0);
  HEAPF32[$520>>2] = $525;
  $527 = ((($b1$638)) + 44|0);
  $528 = +HEAPF32[$521>>2];
  $529 = ((($b2$639)) + 20|0);
  $530 = +HEAPF32[$529>>2];
  $531 = $530 + $528;
  $532 = ((($bs$640)) + 44|0);
  HEAPF32[$526>>2] = $531;
  $533 = +HEAPF32[$527>>2];
  $534 = ((($b2$639)) + 16|0);
  $535 = +HEAPF32[$534>>2];
  $536 = $535 + $533;
  HEAPF32[$532>>2] = $536;
  $scevgep98 = ((($bs$640)) + 48|0);
  $scevgep100 = ((($b1$638)) + 48|0);
  $scevgep102 = ((($b1$638)) + 64|0);
  $537 = ((($b1$638)) + 52|0);
  $538 = +HEAPF32[$scevgep100>>2];
  $539 = ((($b2$639)) + 12|0);
  $540 = +HEAPF32[$539>>2];
  $541 = $538 - $540;
  $542 = +HEAPF32[$460>>2];
  $543 = $541 * $542;
  $544 = ((($bs$640)) + 52|0);
  HEAPF32[$scevgep98>>2] = $543;
  $545 = ((($b1$638)) + 56|0);
  $546 = +HEAPF32[$537>>2];
  $547 = ((($b2$639)) + 8|0);
  $548 = +HEAPF32[$547>>2];
  $549 = $546 - $548;
  $550 = +HEAPF32[$461>>2];
  $551 = $549 * $550;
  $552 = ((($bs$640)) + 56|0);
  HEAPF32[$544>>2] = $551;
  $553 = ((($b1$638)) + 60|0);
  $554 = +HEAPF32[$545>>2];
  $555 = ((($b2$639)) + 4|0);
  $556 = +HEAPF32[$555>>2];
  $557 = $554 - $556;
  $558 = +HEAPF32[$462>>2];
  $559 = $557 * $558;
  $560 = ((($bs$640)) + 60|0);
  HEAPF32[$552>>2] = $559;
  $561 = +HEAPF32[$553>>2];
  $562 = +HEAPF32[$b2$639>>2];
  $563 = $561 - $562;
  $564 = +HEAPF32[$458>>2];
  $565 = $563 * $564;
  HEAPF32[$560>>2] = $565;
  $scevgep101 = ((($bs$640)) + 64|0);
  $566 = ((($b2$639)) + 64|0);
  $567 = (($j$041) + -1)|0;
  $568 = ($567|0)==(0);
  if ($568) {
   break;
  } else {
   $b1$638 = $scevgep102;$b2$639 = $566;$bs$640 = $scevgep101;$j$041 = $567;
  }
 }
 $scevgep104 = ((($bufs)) + 128|0);
 $569 = HEAP32[(35324)>>2]|0;
 $570 = ((($bufs)) + 16|0);
 $571 = ((($569)) + 4|0);
 $b1$1118 = $bufs;$b2$1119 = $570;$bs$1120 = $scevgep104;$j$121 = 4;
 while(1) {
  $572 = ((($b1$1118)) + 4|0);
  $573 = +HEAPF32[$b1$1118>>2];
  $574 = ((($b2$1119)) + -4|0);
  $575 = +HEAPF32[$574>>2];
  $576 = $575 + $573;
  $577 = ((($bs$1120)) + 4|0);
  HEAPF32[$bs$1120>>2] = $576;
  $578 = ((($b1$1118)) + 8|0);
  $579 = +HEAPF32[$572>>2];
  $580 = ((($b2$1119)) + -8|0);
  $581 = +HEAPF32[$580>>2];
  $582 = $581 + $579;
  $583 = ((($bs$1120)) + 8|0);
  HEAPF32[$577>>2] = $582;
  $584 = ((($b2$1119)) + -12|0);
  $585 = +HEAPF32[$584>>2];
  $586 = ((($b1$1118)) + 12|0);
  $587 = +HEAPF32[$578>>2];
  $588 = $585 - $587;
  $589 = +HEAPF32[$571>>2];
  $590 = $588 * $589;
  $591 = ((($bs$1120)) + 12|0);
  HEAPF32[$583>>2] = $590;
  $592 = ((($b2$1119)) + -16|0);
  $593 = +HEAPF32[$592>>2];
  $594 = ((($b1$1118)) + 16|0);
  $595 = +HEAPF32[$586>>2];
  $596 = $593 - $595;
  $597 = +HEAPF32[$569>>2];
  $598 = $596 * $597;
  $599 = ((($bs$1120)) + 16|0);
  HEAPF32[$591>>2] = $598;
  $600 = ((($b1$1118)) + 20|0);
  $601 = +HEAPF32[$594>>2];
  $602 = ((($b2$1119)) + 12|0);
  $603 = +HEAPF32[$602>>2];
  $604 = $603 + $601;
  $605 = ((($bs$1120)) + 20|0);
  HEAPF32[$599>>2] = $604;
  $606 = ((($b1$1118)) + 24|0);
  $607 = +HEAPF32[$600>>2];
  $608 = ((($b2$1119)) + 8|0);
  $609 = +HEAPF32[$608>>2];
  $610 = $609 + $607;
  $611 = ((($bs$1120)) + 24|0);
  HEAPF32[$605>>2] = $610;
  $612 = ((($b1$1118)) + 28|0);
  $613 = +HEAPF32[$606>>2];
  $614 = ((($b2$1119)) + 4|0);
  $615 = +HEAPF32[$614>>2];
  $616 = $613 - $615;
  $617 = +HEAPF32[$571>>2];
  $618 = $616 * $617;
  $619 = ((($bs$1120)) + 28|0);
  HEAPF32[$611>>2] = $618;
  $620 = ((($b1$1118)) + 32|0);
  $621 = +HEAPF32[$612>>2];
  $622 = +HEAPF32[$b2$1119>>2];
  $623 = $621 - $622;
  $624 = +HEAPF32[$569>>2];
  $625 = $623 * $624;
  $626 = ((($bs$1120)) + 32|0);
  HEAPF32[$619>>2] = $625;
  $627 = ((($b2$1119)) + 32|0);
  $628 = (($j$121) + -1)|0;
  $629 = ($628|0)==(0);
  if ($629) {
   break;
  } else {
   $b1$1118 = $620;$b2$1119 = $627;$bs$1120 = $626;$j$121 = $628;
  }
 }
 $630 = HEAP32[(35328)>>2]|0;
 $631 = ((($bufs)) + 132|0);
 $632 = +HEAPF32[$scevgep104>>2];
 $633 = ((($bufs)) + 136|0);
 $634 = +HEAPF32[$631>>2];
 $635 = $634 + $632;
 HEAPF32[$bufs>>2] = $635;
 $636 = $632 - $634;
 $637 = +HEAPF32[$630>>2];
 $638 = $637 * $636;
 $639 = ((($bufs)) + 8|0);
 HEAPF32[$6>>2] = $638;
 $640 = ((($bufs)) + 140|0);
 $641 = +HEAPF32[$633>>2];
 $642 = ((($bufs)) + 144|0);
 $643 = +HEAPF32[$640>>2];
 $644 = $643 + $641;
 $645 = ((($bufs)) + 12|0);
 HEAPF32[$639>>2] = $644;
 $646 = $643 - $641;
 $647 = +HEAPF32[$630>>2];
 $648 = $647 * $646;
 $649 = ((($bufs)) + 16|0);
 HEAPF32[$645>>2] = $648;
 $650 = ((($bufs)) + 148|0);
 $651 = +HEAPF32[$642>>2];
 $652 = ((($bufs)) + 152|0);
 $653 = +HEAPF32[$650>>2];
 $654 = $653 + $651;
 $655 = ((($bufs)) + 20|0);
 HEAPF32[$649>>2] = $654;
 $656 = $651 - $653;
 $657 = +HEAPF32[$630>>2];
 $658 = $657 * $656;
 $659 = ((($bufs)) + 24|0);
 HEAPF32[$655>>2] = $658;
 $660 = ((($bufs)) + 156|0);
 $661 = +HEAPF32[$652>>2];
 $662 = ((($bufs)) + 160|0);
 $663 = +HEAPF32[$660>>2];
 $664 = $663 + $661;
 $665 = ((($bufs)) + 28|0);
 HEAPF32[$659>>2] = $664;
 $666 = $663 - $661;
 $667 = +HEAPF32[$630>>2];
 $668 = $667 * $666;
 $669 = ((($bufs)) + 32|0);
 HEAPF32[$665>>2] = $668;
 $670 = ((($bufs)) + 164|0);
 $671 = +HEAPF32[$662>>2];
 $672 = ((($bufs)) + 168|0);
 $673 = +HEAPF32[$670>>2];
 $674 = $673 + $671;
 $675 = ((($bufs)) + 36|0);
 HEAPF32[$669>>2] = $674;
 $676 = $671 - $673;
 $677 = +HEAPF32[$630>>2];
 $678 = $677 * $676;
 $679 = ((($bufs)) + 40|0);
 HEAPF32[$675>>2] = $678;
 $680 = ((($bufs)) + 172|0);
 $681 = +HEAPF32[$672>>2];
 $682 = ((($bufs)) + 176|0);
 $683 = +HEAPF32[$680>>2];
 $684 = $683 + $681;
 $685 = ((($bufs)) + 44|0);
 HEAPF32[$679>>2] = $684;
 $686 = $683 - $681;
 $687 = +HEAPF32[$630>>2];
 $688 = $687 * $686;
 $689 = ((($bufs)) + 48|0);
 HEAPF32[$685>>2] = $688;
 $690 = ((($bufs)) + 180|0);
 $691 = +HEAPF32[$682>>2];
 $692 = ((($bufs)) + 184|0);
 $693 = +HEAPF32[$690>>2];
 $694 = $693 + $691;
 $695 = ((($bufs)) + 52|0);
 HEAPF32[$689>>2] = $694;
 $696 = $691 - $693;
 $697 = +HEAPF32[$630>>2];
 $698 = $697 * $696;
 $699 = ((($bufs)) + 56|0);
 HEAPF32[$695>>2] = $698;
 $700 = ((($bufs)) + 188|0);
 $701 = +HEAPF32[$692>>2];
 $702 = ((($bufs)) + 192|0);
 $703 = +HEAPF32[$700>>2];
 $704 = $703 + $701;
 $705 = ((($bufs)) + 60|0);
 HEAPF32[$699>>2] = $704;
 $706 = $703 - $701;
 $707 = +HEAPF32[$630>>2];
 $708 = $707 * $706;
 $709 = ((($bufs)) + 64|0);
 HEAPF32[$705>>2] = $708;
 $710 = ((($bufs)) + 196|0);
 $711 = +HEAPF32[$702>>2];
 $712 = ((($bufs)) + 200|0);
 $713 = +HEAPF32[$710>>2];
 $714 = $713 + $711;
 $715 = ((($bufs)) + 68|0);
 HEAPF32[$709>>2] = $714;
 $716 = $711 - $713;
 $717 = +HEAPF32[$630>>2];
 $718 = $717 * $716;
 $719 = ((($bufs)) + 72|0);
 HEAPF32[$715>>2] = $718;
 $720 = ((($bufs)) + 204|0);
 $721 = +HEAPF32[$712>>2];
 $722 = ((($bufs)) + 208|0);
 $723 = +HEAPF32[$720>>2];
 $724 = $723 + $721;
 $725 = ((($bufs)) + 76|0);
 HEAPF32[$719>>2] = $724;
 $726 = $723 - $721;
 $727 = +HEAPF32[$630>>2];
 $728 = $727 * $726;
 $729 = ((($bufs)) + 80|0);
 HEAPF32[$725>>2] = $728;
 $730 = ((($bufs)) + 212|0);
 $731 = +HEAPF32[$722>>2];
 $732 = ((($bufs)) + 216|0);
 $733 = +HEAPF32[$730>>2];
 $734 = $733 + $731;
 $735 = ((($bufs)) + 84|0);
 HEAPF32[$729>>2] = $734;
 $736 = $731 - $733;
 $737 = +HEAPF32[$630>>2];
 $738 = $737 * $736;
 $739 = ((($bufs)) + 88|0);
 HEAPF32[$735>>2] = $738;
 $740 = ((($bufs)) + 220|0);
 $741 = +HEAPF32[$732>>2];
 $742 = ((($bufs)) + 224|0);
 $743 = +HEAPF32[$740>>2];
 $744 = $743 + $741;
 $745 = ((($bufs)) + 92|0);
 HEAPF32[$739>>2] = $744;
 $746 = $743 - $741;
 $747 = +HEAPF32[$630>>2];
 $748 = $747 * $746;
 $749 = ((($bufs)) + 96|0);
 HEAPF32[$745>>2] = $748;
 $750 = ((($bufs)) + 228|0);
 $751 = +HEAPF32[$742>>2];
 $752 = ((($bufs)) + 232|0);
 $753 = +HEAPF32[$750>>2];
 $754 = $753 + $751;
 $755 = ((($bufs)) + 100|0);
 HEAPF32[$749>>2] = $754;
 $756 = $751 - $753;
 $757 = +HEAPF32[$630>>2];
 $758 = $757 * $756;
 $759 = ((($bufs)) + 104|0);
 HEAPF32[$755>>2] = $758;
 $760 = ((($bufs)) + 236|0);
 $761 = +HEAPF32[$752>>2];
 $762 = ((($bufs)) + 240|0);
 $763 = +HEAPF32[$760>>2];
 $764 = $763 + $761;
 $765 = ((($bufs)) + 108|0);
 HEAPF32[$759>>2] = $764;
 $766 = $763 - $761;
 $767 = +HEAPF32[$630>>2];
 $768 = $767 * $766;
 $769 = ((($bufs)) + 112|0);
 HEAPF32[$765>>2] = $768;
 $770 = ((($bufs)) + 244|0);
 $771 = +HEAPF32[$762>>2];
 $772 = ((($bufs)) + 248|0);
 $773 = +HEAPF32[$770>>2];
 $774 = $773 + $771;
 $775 = ((($bufs)) + 116|0);
 HEAPF32[$769>>2] = $774;
 $776 = $771 - $773;
 $777 = +HEAPF32[$630>>2];
 $778 = $777 * $776;
 $779 = ((($bufs)) + 120|0);
 HEAPF32[$775>>2] = $778;
 $780 = ((($bufs)) + 252|0);
 $781 = +HEAPF32[$772>>2];
 $782 = +HEAPF32[$780>>2];
 $783 = $782 + $781;
 $784 = ((($bufs)) + 124|0);
 HEAPF32[$779>>2] = $783;
 $785 = $782 - $781;
 $786 = +HEAPF32[$630>>2];
 $787 = $786 * $785;
 HEAPF32[$784>>2] = $787;
 $788 = +HEAPF32[$645>>2];
 $789 = +HEAPF32[$639>>2];
 $790 = $789 + $788;
 HEAPF32[$639>>2] = $790;
 $791 = +HEAPF32[$665>>2];
 $792 = +HEAPF32[$659>>2];
 $793 = $792 + $791;
 HEAPF32[$659>>2] = $793;
 $794 = +HEAPF32[$685>>2];
 $795 = +HEAPF32[$679>>2];
 $796 = $795 + $794;
 HEAPF32[$679>>2] = $796;
 $797 = +HEAPF32[$705>>2];
 $798 = +HEAPF32[$699>>2];
 $799 = $798 + $797;
 HEAPF32[$699>>2] = $799;
 $800 = +HEAPF32[$725>>2];
 $801 = +HEAPF32[$719>>2];
 $802 = $801 + $800;
 HEAPF32[$719>>2] = $802;
 $803 = +HEAPF32[$745>>2];
 $804 = +HEAPF32[$739>>2];
 $805 = $804 + $803;
 HEAPF32[$739>>2] = $805;
 $806 = +HEAPF32[$765>>2];
 $807 = +HEAPF32[$759>>2];
 $808 = $807 + $806;
 HEAPF32[$759>>2] = $808;
 $809 = +HEAPF32[$784>>2];
 $810 = +HEAPF32[$779>>2];
 $811 = $810 + $809;
 $812 = ((($bufs)) + 24|0);
 $813 = +HEAPF32[$812>>2];
 $814 = +HEAPF32[$649>>2];
 $815 = $814 + $813;
 HEAPF32[$649>>2] = $815;
 $816 = ((($bufs)) + 20|0);
 $817 = +HEAPF32[$816>>2];
 $818 = $813 + $817;
 HEAPF32[$812>>2] = $818;
 $819 = ((($bufs)) + 28|0);
 $820 = +HEAPF32[$819>>2];
 $821 = $820 + $817;
 HEAPF32[$816>>2] = $821;
 $822 = ((($bufs)) + 32|0);
 $823 = ((($bufs)) + 56|0);
 $824 = +HEAPF32[$823>>2];
 $825 = ((($bufs)) + 48|0);
 $826 = +HEAPF32[$825>>2];
 $827 = $826 + $824;
 $828 = ((($bufs)) + 52|0);
 $829 = +HEAPF32[$828>>2];
 $830 = $824 + $829;
 $831 = ((($bufs)) + 60|0);
 $832 = +HEAPF32[$831>>2];
 $833 = $832 + $829;
 $834 = ((($bufs)) + 88|0);
 $835 = +HEAPF32[$834>>2];
 $836 = ((($bufs)) + 80|0);
 $837 = +HEAPF32[$836>>2];
 $838 = $837 + $835;
 HEAPF32[$836>>2] = $838;
 $839 = ((($bufs)) + 84|0);
 $840 = +HEAPF32[$839>>2];
 $841 = $835 + $840;
 HEAPF32[$834>>2] = $841;
 $842 = ((($bufs)) + 92|0);
 $843 = +HEAPF32[$842>>2];
 $844 = $843 + $840;
 HEAPF32[$839>>2] = $844;
 $845 = ((($bufs)) + 112|0);
 $846 = +HEAPF32[$845>>2];
 $847 = $846 + $811;
 $848 = ((($bufs)) + 116|0);
 $849 = +HEAPF32[$848>>2];
 $850 = $811 + $849;
 $851 = ((($bufs)) + 124|0);
 $852 = +HEAPF32[$851>>2];
 $853 = $852 + $849;
 $854 = ((($bufs)) + 48|0);
 $855 = +HEAPF32[$822>>2];
 $856 = $855 + $827;
 HEAPF32[$822>>2] = $856;
 $857 = ((($bufs)) + 40|0);
 $858 = +HEAPF32[$857>>2];
 $859 = $827 + $858;
 HEAPF32[$854>>2] = $859;
 $860 = ((($bufs)) + 56|0);
 $861 = $830 + $858;
 HEAPF32[$857>>2] = $861;
 $862 = ((($bufs)) + 36|0);
 $863 = +HEAPF32[$862>>2];
 $864 = $863 + $830;
 HEAPF32[$860>>2] = $864;
 $865 = ((($bufs)) + 52|0);
 $866 = $833 + $863;
 HEAPF32[$862>>2] = $866;
 $867 = ((($bufs)) + 44|0);
 $868 = +HEAPF32[$867>>2];
 $869 = $868 + $833;
 HEAPF32[$865>>2] = $869;
 $870 = $832 + $868;
 HEAPF32[$867>>2] = $870;
 $871 = ((($bufs)) + 112|0);
 $872 = ((($bufs)) + 96|0);
 $873 = +HEAPF32[$872>>2];
 $874 = $873 + $847;
 HEAPF32[$872>>2] = $874;
 $875 = ((($bufs)) + 104|0);
 $876 = +HEAPF32[$875>>2];
 $877 = $847 + $876;
 HEAPF32[$871>>2] = $877;
 $878 = ((($bufs)) + 120|0);
 $879 = $850 + $876;
 HEAPF32[$875>>2] = $879;
 $880 = ((($bufs)) + 100|0);
 $881 = +HEAPF32[$880>>2];
 $882 = $881 + $850;
 HEAPF32[$878>>2] = $882;
 $883 = ((($bufs)) + 116|0);
 $884 = $853 + $881;
 HEAPF32[$880>>2] = $884;
 $885 = ((($bufs)) + 108|0);
 $886 = +HEAPF32[$885>>2];
 $887 = $886 + $853;
 HEAPF32[$883>>2] = $887;
 $888 = $852 + $886;
 HEAPF32[$885>>2] = $888;
 $889 = HEAP32[$bufs>>2]|0;
 $890 = ((($out0)) + 1024|0);
 HEAP32[$890>>2] = $889;
 $891 = +HEAPF32[$scevgep117>>2];
 $892 = $874 + $891;
 $893 = ((($out0)) + 960|0);
 HEAPF32[$893>>2] = $892;
 $894 = ((($out0)) + 896|0);
 HEAPF32[$894>>2] = $856;
 $895 = ((($bufs)) + 80|0);
 $896 = $838 + $874;
 $897 = ((($out0)) + 832|0);
 HEAPF32[$897>>2] = $896;
 $898 = HEAP32[$570>>2]|0;
 $899 = ((($out0)) + 768|0);
 HEAP32[$899>>2] = $898;
 $900 = +HEAPF32[$895>>2];
 $901 = $877 + $900;
 $902 = ((($out0)) + 704|0);
 HEAPF32[$902>>2] = $901;
 $903 = ((($out0)) + 640|0);
 HEAPF32[$903>>2] = $859;
 $904 = ((($bufs)) + 72|0);
 $905 = +HEAPF32[$904>>2];
 $906 = $905 + $877;
 $907 = ((($out0)) + 576|0);
 HEAPF32[$907>>2] = $906;
 $908 = ((($bufs)) + 8|0);
 $909 = HEAP32[$908>>2]|0;
 $910 = ((($out0)) + 512|0);
 HEAP32[$910>>2] = $909;
 $911 = $879 + $905;
 $912 = ((($out0)) + 448|0);
 HEAPF32[$912>>2] = $911;
 $913 = ((($out0)) + 384|0);
 HEAPF32[$913>>2] = $861;
 $914 = ((($bufs)) + 88|0);
 $915 = +HEAPF32[$914>>2];
 $916 = $915 + $879;
 $917 = ((($out0)) + 320|0);
 HEAPF32[$917>>2] = $916;
 $918 = ((($bufs)) + 24|0);
 $919 = HEAP32[$918>>2]|0;
 $920 = ((($out0)) + 256|0);
 HEAP32[$920>>2] = $919;
 $921 = $882 + $915;
 $922 = ((($out0)) + 192|0);
 HEAPF32[$922>>2] = $921;
 $923 = ((($bufs)) + 56|0);
 $924 = HEAP32[$923>>2]|0;
 $925 = ((($out0)) + 128|0);
 HEAP32[$925>>2] = $924;
 $926 = ((($bufs)) + 68|0);
 $927 = +HEAPF32[$926>>2];
 $928 = $927 + $882;
 $929 = ((($out0)) + 64|0);
 HEAPF32[$929>>2] = $928;
 $930 = ((($bufs)) + 4|0);
 $931 = HEAP32[$930>>2]|0;
 HEAP32[$out0>>2] = $931;
 HEAP32[$out1>>2] = $931;
 $932 = ((($bufs)) + 100|0);
 $933 = $884 + $927;
 $934 = ((($out1)) + 64|0);
 HEAPF32[$934>>2] = $933;
 $935 = ((($bufs)) + 36|0);
 $936 = HEAP32[$935>>2]|0;
 $937 = ((($out1)) + 128|0);
 HEAP32[$937>>2] = $936;
 $938 = +HEAPF32[$932>>2];
 $939 = ((($bufs)) + 84|0);
 $940 = +HEAPF32[$939>>2];
 $941 = $940 + $938;
 $942 = ((($out1)) + 192|0);
 HEAPF32[$942>>2] = $941;
 $943 = ((($bufs)) + 20|0);
 $944 = HEAP32[$943>>2]|0;
 $945 = ((($out1)) + 256|0);
 HEAP32[$945>>2] = $944;
 $946 = ((($bufs)) + 116|0);
 $947 = +HEAPF32[$946>>2];
 $948 = $947 + $940;
 $949 = ((($out1)) + 320|0);
 HEAPF32[$949>>2] = $948;
 $950 = ((($bufs)) + 52|0);
 $951 = HEAP32[$950>>2]|0;
 $952 = ((($out1)) + 384|0);
 HEAP32[$952>>2] = $951;
 $953 = ((($bufs)) + 76|0);
 $954 = +HEAPF32[$953>>2];
 $955 = $954 + $947;
 $956 = ((($out1)) + 448|0);
 HEAPF32[$956>>2] = $955;
 $957 = ((($bufs)) + 12|0);
 $958 = HEAP32[$957>>2]|0;
 $959 = ((($out1)) + 512|0);
 HEAP32[$959>>2] = $958;
 $960 = ((($bufs)) + 108|0);
 $961 = +HEAPF32[$960>>2];
 $962 = $961 + $954;
 $963 = ((($out1)) + 576|0);
 HEAPF32[$963>>2] = $962;
 $964 = ((($bufs)) + 44|0);
 $965 = HEAP32[$964>>2]|0;
 $966 = ((($out1)) + 640|0);
 HEAP32[$966>>2] = $965;
 $967 = ((($bufs)) + 92|0);
 $968 = +HEAPF32[$967>>2];
 $969 = $968 + $961;
 $970 = ((($out1)) + 704|0);
 HEAPF32[$970>>2] = $969;
 $971 = ((($bufs)) + 28|0);
 $972 = HEAP32[$971>>2]|0;
 $973 = ((($out1)) + 768|0);
 HEAP32[$973>>2] = $972;
 $974 = ((($bufs)) + 124|0);
 $975 = +HEAPF32[$974>>2];
 $976 = $975 + $968;
 $977 = ((($out1)) + 832|0);
 HEAPF32[$977>>2] = $976;
 $978 = ((($bufs)) + 60|0);
 $979 = HEAP32[$978>>2]|0;
 $980 = ((($out1)) + 896|0);
 HEAP32[$980>>2] = $979;
 $981 = ((($out1)) + 960|0);
 HEAPF32[$981>>2] = $975;
 STACKTOP = sp;return;
}
function _INT123_do_equalizer($bandPtr,$channel,$equalizer) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $equalizer = $equalizer|0;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0.0, $4 = 0.0, $5 = 0, $exitcond = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $i$01 = 0;
 while(1) {
  $0 = (($bandPtr) + ($i$01<<2)|0);
  $1 = +HEAPF32[$0>>2];
  $2 = ((($equalizer) + ($channel<<7)|0) + ($i$01<<2)|0);
  $3 = +HEAPF32[$2>>2];
  $4 = $3 * $1;
  HEAPF32[$0>>2] = $4;
  $5 = (($i$01) + 1)|0;
  $exitcond = ($5|0)==(32);
  if ($exitcond) {
   break;
  } else {
   $i$01 = $5;
  }
 }
 return;
}
function _INT123_set_synth_functions($fr) {
 $fr = $fr|0;
 var $$$i = 0, $$0 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $9 = 0, $basic_format$0$ph = 0, $basic_synth$0$i = 0, $or$cond$i$i = 0, $or$cond10$i$i = 0, $or$cond11$i$i = 0, $or$cond12$i$i = 0, $or$cond13$i$i = 0, $or$cond14$i$i = 0, $or$cond15$i$i = 0, $or$cond16$i$i = 0, $or$cond17$i$i = 0, $or$cond18$i$i = 0, $or$cond19$i$i = 0, $or$cond20$i$i = 0;
 var $or$cond21$i$i = 0, $or$cond22$i$i = 0, $or$cond9$i$i = 0, $switch = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer10 = 0, $vararg_buffer4 = 0, $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer10 = sp + 32|0;
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($fr)) + 27876|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 64;
 $3 = ($2|0)==(0);
 if ($3) {
  $4 = $1 & 15;
  $5 = ($4|0)==(0);
  if ($5) {
   $6 = $1 & 3584;
   $7 = ($6|0)==(0);
   if ($7) {
    $8 = $1 & 16640;
    $9 = ($8|0)==(0);
    if ($9) {
     $10 = ((($fr)) + 28068|0);
     $11 = HEAP32[$10>>2]|0;
     $12 = $11 & 32;
     $13 = ($12|0)==(0);
     if (!($13)) {
      $$0 = -1;
      STACKTOP = sp;return ($$0|0);
     }
     $14 = HEAP32[37544]|0;
     HEAP32[$vararg_buffer>>2] = 339;
     (_fprintf($14,514559,$vararg_buffer)|0);
     $$0 = -1;
     STACKTOP = sp;return ($$0|0);
    } else {
     $basic_format$0$ph = 3;
    }
   } else {
    $basic_format$0$ph = 2;
   }
  } else {
   $basic_format$0$ph = 1;
  }
 } else {
  $basic_format$0$ph = 0;
 }
 $15 = ((($fr)) + 19680|0);
 $16 = HEAP32[$15>>2]|0;
 $switch = ($16>>>0)<(4);
 if (!($switch)) {
  $17 = ((($fr)) + 28068|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = $18 & 32;
  $20 = ($19|0)==(0);
  if (!($20)) {
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  }
  $21 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer1>>2] = 359;
  (_fprintf($21,514650,$vararg_buffer1)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $22 = ((((($fr)) + 19364|0) + ($16<<4)|0) + ($basic_format$0$ph<<2)|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = ((($fr)) + 19636|0);
 HEAP32[$24>>2] = $23;
 $25 = ((((($fr)) + 19428|0) + ($16<<4)|0) + ($basic_format$0$ph<<2)|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ((($fr)) + 19640|0);
 HEAP32[$27>>2] = $26;
 $28 = ((($fr)) + 27884|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = ($29|0)==(2);
 $31 = ((((($fr)) + 19492|0) + ($16<<4)|0) + ($basic_format$0$ph<<2)|0);
 $32 = ((((($fr)) + 19556|0) + ($16<<4)|0) + ($basic_format$0$ph<<2)|0);
 $$in = $30 ? $31 : $32;
 $33 = HEAP32[$$in>>2]|0;
 $34 = ((($fr)) + 19644|0);
 HEAP32[$34>>2] = $33;
 $35 = $23;
 $36 = ($35|0)==(4267|0);
 if ($36) {
  $37 = ((($fr)) + 19364|0);
  $38 = HEAP32[$37>>2]|0;
  $basic_synth$0$i = $38;
 } else {
  $basic_synth$0$i = $35;
 }
 $39 = ($basic_synth$0$i|0)==(2094|0);
 $40 = ($basic_synth$0$i|0)==(2095|0);
 $or$cond$i$i = $39 | $40;
 $41 = ($basic_synth$0$i|0)==(2096|0);
 $or$cond9$i$i = $41 | $or$cond$i$i;
 $42 = ($basic_synth$0$i|0)==(2097|0);
 $or$cond10$i$i = $42 | $or$cond9$i$i;
 $43 = ($basic_synth$0$i|0)==(2098|0);
 $or$cond11$i$i = $43 | $or$cond10$i$i;
 $44 = ($basic_synth$0$i|0)==(2099|0);
 $or$cond12$i$i = $44 | $or$cond11$i$i;
 $45 = ($basic_synth$0$i|0)==(2100|0);
 $or$cond13$i$i = $45 | $or$cond12$i$i;
 $46 = ($basic_synth$0$i|0)==(2101|0);
 $or$cond14$i$i = $46 | $or$cond13$i$i;
 $47 = ($basic_synth$0$i|0)==(2102|0);
 $or$cond15$i$i = $47 | $or$cond14$i$i;
 $48 = ($basic_synth$0$i|0)==(2103|0);
 $or$cond16$i$i = $48 | $or$cond15$i$i;
 $49 = ($basic_synth$0$i|0)==(2104|0);
 $or$cond17$i$i = $49 | $or$cond16$i$i;
 $50 = ($basic_synth$0$i|0)==(2105|0);
 $or$cond18$i$i = $50 | $or$cond17$i$i;
 $51 = ($basic_synth$0$i|0)==(2106|0);
 $or$cond19$i$i = $51 | $or$cond18$i$i;
 $52 = ($basic_synth$0$i|0)==(2107|0);
 $or$cond20$i$i = $52 | $or$cond19$i$i;
 $53 = ($basic_synth$0$i|0)==(2108|0);
 $or$cond21$i$i = $53 | $or$cond20$i$i;
 $54 = ($basic_synth$0$i|0)==(2109|0);
 $or$cond22$i$i = $54 | $or$cond21$i$i;
 $$$i = $or$cond22$i$i ? 1 : 15;
 if (!($or$cond22$i$i)) {
  $55 = ((($fr)) + 28068|0);
  $56 = HEAP32[$55>>2]|0;
  $57 = $56 & 32;
  $58 = ($57|0)==(0);
  if ($58) {
   $59 = HEAP32[37544]|0;
   HEAP32[$vararg_buffer4>>2] = 303;
   (_fprintf($59,514748,$vararg_buffer4)|0);
  }
  $60 = ((($fr)) + 28376|0);
  HEAP32[$60>>2] = 37;
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $61 = ((($fr)) + 19620|0);
 HEAP32[$61>>2] = $$$i;
 $62 = $$$i | 2;
 $63 = ($62|0)==(11);
 $64 = $63 ? 2 : 1;
 $65 = ((($fr)) + 19624|0);
 HEAP32[$65>>2] = $64;
 $66 = (_INT123_frame_buffers($fr)|0);
 $67 = ($66|0)==(0);
 if (!($67)) {
  $68 = ((($fr)) + 28376|0);
  HEAP32[$68>>2] = 11;
  $69 = ((($fr)) + 28068|0);
  $70 = HEAP32[$69>>2]|0;
  $71 = $70 & 32;
  $72 = ($71|0)==(0);
  if (!($72)) {
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  }
  $73 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer7>>2] = 381;
  (_fprintf($73,514840,$vararg_buffer7)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $74 = ($basic_format$0$ph|0)==(1);
 if ($74) {
  $75 = (_INT123_make_conv16to8_table($fr)|0);
  $76 = ($75|0)==(0);
  if (!($76)) {
   $77 = ((($fr)) + 28068|0);
   $78 = HEAP32[$77>>2]|0;
   $79 = $78 & 32;
   $80 = ($79|0)==(0);
   if (!($80)) {
    $$0 = -1;
    STACKTOP = sp;return ($$0|0);
   }
   $81 = HEAP32[37544]|0;
   HEAP32[$vararg_buffer10>>2] = 391;
   (_fprintf($81,515000,$vararg_buffer10)|0);
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  }
 }
 _INT123_init_layer3_stuff($fr,4268);
 _INT123_init_layer12_stuff($fr,4269);
 $82 = ((($fr)) + 19648|0);
 HEAP32[$82>>2] = 4270;
 _INT123_make_decode_tables($fr);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _INT123_synth_1to1_8bit_wrap($bandPtr,$channel,$fr,$final) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $fr = $fr|0;
 $final = $final|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, $samples$02 = 0, $samples_tmp = 0, $tmp1$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = (($samples_tmp) + ($channel<<1)|0);
 $1 = ((($fr)) + 27856|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($fr)) + 27864|0);
 $4 = HEAP32[$3>>2]|0;
 HEAP32[$1>>2] = $samples_tmp;
 HEAP32[$3>>2] = 0;
 $5 = ((($fr)) + 19364|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (FUNCTION_TABLE_iiiii[$6 & 8191]($bandPtr,$channel,$fr,0)|0);
 HEAP32[$1>>2] = $2;
 $8 = (($4) + ($channel))|0;
 $9 = (($2) + ($8)|0);
 $10 = ((($fr)) + 9592|0);
 $i$01 = 0;$samples$02 = $9;$tmp1$03 = $0;
 while(1) {
  $11 = HEAP16[$tmp1$03>>1]|0;
  $12 = $11 << 16 >> 16;
  $13 = $12 >> 3;
  $14 = HEAP32[$10>>2]|0;
  $15 = (($14) + ($13)|0);
  $16 = HEAP8[$15>>0]|0;
  HEAP8[$samples$02>>0] = $16;
  $17 = ((($samples$02)) + 2|0);
  $18 = ((($tmp1$03)) + 4|0);
  $19 = (($i$01) + 1)|0;
  $exitcond = ($19|0)==(32);
  if ($exitcond) {
   break;
  } else {
   $i$01 = $19;$samples$02 = $17;$tmp1$03 = $18;
  }
 }
 $20 = ($final|0)!=(0);
 $21 = $20 ? 64 : 0;
 $22 = (($4) + ($21))|0;
 HEAP32[$3>>2] = $22;
 STACKTOP = sp;return ($7|0);
}
function _INT123_synth_1to1($bandPtr,$channel,$fr,$final) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $fr = $fr|0;
 $final = $final|0;
 var $$neg = 0.0, $$neg1 = 0.0, $$neg15 = 0.0, $$neg16 = 0.0, $$neg17 = 0.0, $$neg18 = 0.0, $$neg19 = 0.0, $$neg2 = 0.0, $$neg20 = 0.0, $$neg21 = 0.0, $$neg22 = 0.0, $$neg23 = 0.0, $$neg24 = 0.0, $$neg25 = 0.0, $$neg26 = 0.0, $$neg27 = 0.0, $$neg28 = 0.0, $$neg29 = 0.0, $$neg3 = 0.0, $$neg4 = 0.0;
 var $$neg5 = 0.0, $$neg6 = 0.0, $$neg7 = 0.0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0, $109 = 0.0, $11 = 0, $110 = 0.0;
 var $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0.0, $131 = 0.0, $132 = 0.0, $133 = 0, $134 = 0.0, $135 = 0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0, $146 = 0.0, $147 = 0;
 var $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0, $158 = 0.0, $159 = 0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0, $164 = 0.0, $165 = 0;
 var $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0, $17 = 0, $170 = 0.0, $171 = 0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0.0;
 var $184 = 0, $185 = 0.0, $186 = 0.0, $187 = 0, $188 = 0.0, $189 = 0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0, $200 = 0.0;
 var $201 = 0, $202 = 0.0, $203 = 0, $204 = 0.0, $205 = 0, $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0.0, $213 = 0, $214 = 0.0, $215 = 0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0;
 var $22 = 0, $220 = 0.0, $221 = 0, $222 = 0.0, $223 = 0, $224 = 0.0, $225 = 0, $226 = 0.0, $227 = 0, $228 = 0.0, $229 = 0, $23 = 0, $230 = 0.0, $231 = 0, $232 = 0.0, $233 = 0, $234 = 0.0, $235 = 0, $236 = 0.0, $237 = 0;
 var $238 = 0.0, $239 = 0, $24 = 0, $240 = 0.0, $241 = 0, $242 = 0.0, $243 = 0, $244 = 0.0, $245 = 0.0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0, $254 = 0, $255 = 0;
 var $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0, $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0;
 var $55 = 0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0.0;
 var $73 = 0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0, $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0, $90 = 0.0;
 var $91 = 0, $92 = 0.0, $93 = 0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0, $99 = 0.0, $b0$0 = 0, $b0$1$pn48 = 0, $b0$1$pn48$phi = 0, $b0$155 = 0, $b0$2 = 0, $b0$252 = 0, $bo1$0 = 0, $buf$0 = 0, $clip$056 = 0, $clip$1 = 0, $clip$1$lcssa = 0;
 var $clip$2 = 0, $clip$349 = 0, $clip$4 = 0, $clip$4$lcssa = 0, $j$057 = 0, $j$150 = 0, $samples$0$in = 0, $samples$154 = 0, $samples$2 = 0, $samples$247 = 0, $samples$253 = 0, $scevgep = 0, $scevgep62 = 0, $scevgep63 = 0, $scevgep66 = 0, $scevgep67 = 0, $tmp = 0.0, $tmp11 = 0.0, $tmp13 = 0.0, $tmp30 = 0.0;
 var $tmp32 = 0.0, $tmp34 = 0.0, $tmp36 = 0.0, $tmp38 = 0.0, $tmp40 = 0.0, $tmp42 = 0.0, $tmp44 = 0.0, $tmp45 = 0.0, $tmp9 = 0.0, $window$058 = 0, $window$151 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($fr)) + 9288|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = ((($fr)) + 9292|0);
  _INT123_do_equalizer($bandPtr,$channel,$8);
 }
 $9 = ($channel|0)==(0);
 if ($9) {
  $10 = ((($fr)) + 9272|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + 15)|0;
  $13 = $12 & 15;
  HEAP32[$10>>2] = $13;
  $14 = ((($fr)) + 9248|0);
  $18 = $13;$buf$0 = $14;$samples$0$in = $4;
 } else {
  $15 = ((($4)) + 2|0);
  $16 = ((($fr)) + 9256|0);
  $$phi$trans$insert = ((($fr)) + 9272|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $18 = $$pre;$buf$0 = $16;$samples$0$in = $15;
 }
 $17 = $18 & 1;
 $19 = ($17|0)==(0);
 $20 = ((($buf$0)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 if ($19) {
  $27 = (($18) + 1)|0;
  $28 = HEAP32[$buf$0>>2]|0;
  $29 = (($28) + ($18<<2)|0);
  $30 = (($21) + ($18<<2)|0);
  $31 = ((($30)) + 4|0);
  _INT123_dct64($29,$31,$bandPtr);
  $b0$0 = $21;$bo1$0 = $27;
 } else {
  $22 = HEAP32[$buf$0>>2]|0;
  $23 = (($18) + 1)|0;
  $24 = $23 & 15;
  $25 = (($21) + ($24<<2)|0);
  $26 = (($22) + ($18<<2)|0);
  _INT123_dct64($25,$26,$bandPtr);
  $b0$0 = $22;$bo1$0 = $18;
 }
 $32 = ((($fr)) + 9284|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($33)) + 64|0);
 $35 = (0 - ($bo1$0))|0;
 $36 = (($34) + ($35<<2)|0);
 $scevgep = ((($samples$0$in)) + 64|0);
 $37 = (528 - ($bo1$0))|0;
 $scevgep62 = (($33) + ($37<<2)|0);
 $38 = (496 - ($bo1$0))|0;
 $b0$155 = $b0$0;$clip$056 = 0;$j$057 = 16;$samples$154 = $samples$0$in;$window$058 = $36;
 while(1) {
  $39 = ((($window$058)) + 4|0);
  $40 = +HEAPF32[$window$058>>2];
  $41 = ((($b0$155)) + 4|0);
  $42 = +HEAPF32[$b0$155>>2];
  $43 = $42 * $40;
  $44 = ((($window$058)) + 8|0);
  $45 = +HEAPF32[$39>>2];
  $46 = ((($b0$155)) + 8|0);
  $47 = +HEAPF32[$41>>2];
  $$neg = $47 * $45;
  $48 = ((($window$058)) + 12|0);
  $49 = +HEAPF32[$44>>2];
  $50 = ((($b0$155)) + 12|0);
  $51 = +HEAPF32[$46>>2];
  $52 = $51 * $49;
  $53 = ((($window$058)) + 16|0);
  $54 = +HEAPF32[$48>>2];
  $55 = ((($b0$155)) + 16|0);
  $56 = +HEAPF32[$50>>2];
  $$neg1 = $56 * $54;
  $57 = ((($window$058)) + 20|0);
  $58 = +HEAPF32[$53>>2];
  $59 = ((($b0$155)) + 20|0);
  $60 = +HEAPF32[$55>>2];
  $61 = $60 * $58;
  $62 = ((($window$058)) + 24|0);
  $63 = +HEAPF32[$57>>2];
  $64 = ((($b0$155)) + 24|0);
  $65 = +HEAPF32[$59>>2];
  $$neg2 = $65 * $63;
  $66 = ((($window$058)) + 28|0);
  $67 = +HEAPF32[$62>>2];
  $68 = ((($b0$155)) + 28|0);
  $69 = +HEAPF32[$64>>2];
  $70 = $69 * $67;
  $71 = ((($window$058)) + 32|0);
  $72 = +HEAPF32[$66>>2];
  $73 = ((($b0$155)) + 32|0);
  $74 = +HEAPF32[$68>>2];
  $$neg3 = $74 * $72;
  $75 = ((($window$058)) + 36|0);
  $76 = +HEAPF32[$71>>2];
  $77 = ((($b0$155)) + 36|0);
  $78 = +HEAPF32[$73>>2];
  $79 = $78 * $76;
  $80 = ((($window$058)) + 40|0);
  $81 = +HEAPF32[$75>>2];
  $82 = ((($b0$155)) + 40|0);
  $83 = +HEAPF32[$77>>2];
  $$neg4 = $83 * $81;
  $84 = ((($window$058)) + 44|0);
  $85 = +HEAPF32[$80>>2];
  $86 = ((($b0$155)) + 44|0);
  $87 = +HEAPF32[$82>>2];
  $88 = $87 * $85;
  $89 = ((($window$058)) + 48|0);
  $90 = +HEAPF32[$84>>2];
  $91 = ((($b0$155)) + 48|0);
  $92 = +HEAPF32[$86>>2];
  $$neg5 = $92 * $90;
  $93 = ((($window$058)) + 52|0);
  $94 = +HEAPF32[$89>>2];
  $95 = ((($b0$155)) + 52|0);
  $96 = +HEAPF32[$91>>2];
  $97 = $96 * $94;
  $98 = ((($window$058)) + 56|0);
  $99 = +HEAPF32[$93>>2];
  $100 = ((($b0$155)) + 56|0);
  $101 = +HEAPF32[$95>>2];
  $$neg6 = $101 * $99;
  $102 = ((($window$058)) + 60|0);
  $103 = +HEAPF32[$98>>2];
  $104 = ((($b0$155)) + 60|0);
  $105 = +HEAPF32[$100>>2];
  $106 = $105 * $103;
  $107 = +HEAPF32[$102>>2];
  $108 = ((($b0$155)) + 64|0);
  $109 = +HEAPF32[$104>>2];
  $$neg7 = $109 * $107;
  $tmp = $$neg7 + $$neg6;
  $tmp9 = $$neg5 + $$neg4;
  $tmp11 = $$neg3 + $$neg2;
  $tmp13 = $$neg1 + $$neg;
  $110 = $52 + $43;
  $111 = $110 + $61;
  $112 = $111 + $70;
  $113 = $112 + $79;
  $114 = $113 + $88;
  $115 = $114 + $97;
  $116 = $115 + $106;
  $117 = $116 - $tmp13;
  $118 = $117 - $tmp11;
  $119 = $118 - $tmp9;
  $120 = $119 - $tmp;
  $121 = $120 > 32767.0;
  do {
   if ($121) {
    HEAP16[$samples$154>>1] = 32767;
    $122 = (($clip$056) + 1)|0;
    $clip$1 = $122;
   } else {
    $123 = $120 < -32768.0;
    if ($123) {
     HEAP16[$samples$154>>1] = -32768;
     $124 = (($clip$056) + 1)|0;
     $clip$1 = $124;
     break;
    } else {
     $125 = (~~(($120)));
     HEAP16[$samples$154>>1] = $125;
     $clip$1 = $clip$056;
     break;
    }
   }
  } while(0);
  $126 = (($j$057) + -1)|0;
  $127 = ((($window$058)) + 128|0);
  $128 = ((($samples$154)) + 4|0);
  $129 = ($126|0)==(0);
  if ($129) {
   $clip$1$lcssa = $clip$1;
   break;
  } else {
   $b0$155 = $108;$clip$056 = $clip$1;$j$057 = $126;$samples$154 = $128;$window$058 = $127;
  }
 }
 $scevgep63 = ((($b0$0)) + 1024|0);
 $scevgep66 = ((($b0$0)) + 960|0);
 $scevgep67 = (($33) + ($38<<2)|0);
 $130 = +HEAPF32[$scevgep62>>2];
 $131 = +HEAPF32[$scevgep63>>2];
 $132 = $131 * $130;
 $133 = ((($scevgep67)) + 136|0);
 $134 = +HEAPF32[$133>>2];
 $135 = ((($b0$0)) + 1032|0);
 $136 = +HEAPF32[$135>>2];
 $137 = $136 * $134;
 $138 = $137 + $132;
 $139 = ((($scevgep67)) + 144|0);
 $140 = +HEAPF32[$139>>2];
 $141 = ((($b0$0)) + 1040|0);
 $142 = +HEAPF32[$141>>2];
 $143 = $142 * $140;
 $144 = $138 + $143;
 $145 = ((($scevgep67)) + 152|0);
 $146 = +HEAPF32[$145>>2];
 $147 = ((($b0$0)) + 1048|0);
 $148 = +HEAPF32[$147>>2];
 $149 = $148 * $146;
 $150 = $144 + $149;
 $151 = ((($scevgep67)) + 160|0);
 $152 = +HEAPF32[$151>>2];
 $153 = ((($b0$0)) + 1056|0);
 $154 = +HEAPF32[$153>>2];
 $155 = $154 * $152;
 $156 = $150 + $155;
 $157 = ((($scevgep67)) + 168|0);
 $158 = +HEAPF32[$157>>2];
 $159 = ((($b0$0)) + 1064|0);
 $160 = +HEAPF32[$159>>2];
 $161 = $160 * $158;
 $162 = $156 + $161;
 $163 = ((($scevgep67)) + 176|0);
 $164 = +HEAPF32[$163>>2];
 $165 = ((($b0$0)) + 1072|0);
 $166 = +HEAPF32[$165>>2];
 $167 = $166 * $164;
 $168 = $162 + $167;
 $169 = ((($scevgep67)) + 184|0);
 $170 = +HEAPF32[$169>>2];
 $171 = ((($b0$0)) + 1080|0);
 $172 = +HEAPF32[$171>>2];
 $173 = $172 * $170;
 $174 = $168 + $173;
 $175 = $174 > 32767.0;
 do {
  if ($175) {
   HEAP16[$scevgep>>1] = 32767;
   $176 = (($clip$1$lcssa) + 1)|0;
   $clip$2 = $176;
  } else {
   $177 = $174 < -32768.0;
   if ($177) {
    HEAP16[$scevgep>>1] = -32768;
    $178 = (($clip$1$lcssa) + 1)|0;
    $clip$2 = $178;
    break;
   } else {
    $179 = (~~(($174)));
    HEAP16[$scevgep>>1] = $179;
    $clip$2 = $clip$1$lcssa;
    break;
   }
  }
 } while(0);
 $180 = $bo1$0 << 1;
 $181 = (($scevgep67) + ($180<<2)|0);
 $samples$247 = ((($samples$0$in)) + 68|0);
 $b0$1$pn48 = $scevgep63;$b0$252 = $scevgep66;$clip$349 = $clip$2;$j$150 = 15;$samples$253 = $samples$247;$window$151 = $181;
 while(1) {
  $182 = ((($window$151)) + -4|0);
  $183 = +HEAPF32[$182>>2];
  $184 = ((($b0$1$pn48)) + -60|0);
  $185 = +HEAPF32[$b0$252>>2];
  $186 = $185 * $183;
  $187 = ((($window$151)) + -8|0);
  $188 = +HEAPF32[$187>>2];
  $189 = ((($b0$1$pn48)) + -56|0);
  $190 = +HEAPF32[$184>>2];
  $$neg15 = $190 * $188;
  $191 = ((($window$151)) + -12|0);
  $192 = +HEAPF32[$191>>2];
  $193 = ((($b0$1$pn48)) + -52|0);
  $194 = +HEAPF32[$189>>2];
  $$neg16 = $194 * $192;
  $195 = ((($window$151)) + -16|0);
  $196 = +HEAPF32[$195>>2];
  $197 = ((($b0$1$pn48)) + -48|0);
  $198 = +HEAPF32[$193>>2];
  $$neg17 = $198 * $196;
  $199 = ((($window$151)) + -20|0);
  $200 = +HEAPF32[$199>>2];
  $201 = ((($b0$1$pn48)) + -44|0);
  $202 = +HEAPF32[$197>>2];
  $$neg18 = $202 * $200;
  $203 = ((($window$151)) + -24|0);
  $204 = +HEAPF32[$203>>2];
  $205 = ((($b0$1$pn48)) + -40|0);
  $206 = +HEAPF32[$201>>2];
  $$neg19 = $206 * $204;
  $207 = ((($window$151)) + -28|0);
  $208 = +HEAPF32[$207>>2];
  $209 = ((($b0$1$pn48)) + -36|0);
  $210 = +HEAPF32[$205>>2];
  $$neg20 = $210 * $208;
  $211 = ((($window$151)) + -32|0);
  $212 = +HEAPF32[$211>>2];
  $213 = ((($b0$1$pn48)) + -32|0);
  $214 = +HEAPF32[$209>>2];
  $$neg21 = $214 * $212;
  $215 = ((($window$151)) + -36|0);
  $216 = +HEAPF32[$215>>2];
  $217 = ((($b0$1$pn48)) + -28|0);
  $218 = +HEAPF32[$213>>2];
  $$neg22 = $218 * $216;
  $219 = ((($window$151)) + -40|0);
  $220 = +HEAPF32[$219>>2];
  $221 = ((($b0$1$pn48)) + -24|0);
  $222 = +HEAPF32[$217>>2];
  $$neg23 = $222 * $220;
  $223 = ((($window$151)) + -44|0);
  $224 = +HEAPF32[$223>>2];
  $225 = ((($b0$1$pn48)) + -20|0);
  $226 = +HEAPF32[$221>>2];
  $$neg24 = $226 * $224;
  $227 = ((($window$151)) + -48|0);
  $228 = +HEAPF32[$227>>2];
  $229 = ((($b0$1$pn48)) + -16|0);
  $230 = +HEAPF32[$225>>2];
  $$neg25 = $230 * $228;
  $231 = ((($window$151)) + -52|0);
  $232 = +HEAPF32[$231>>2];
  $233 = ((($b0$1$pn48)) + -12|0);
  $234 = +HEAPF32[$229>>2];
  $$neg26 = $234 * $232;
  $235 = ((($window$151)) + -56|0);
  $236 = +HEAPF32[$235>>2];
  $237 = ((($b0$1$pn48)) + -8|0);
  $238 = +HEAPF32[$233>>2];
  $$neg27 = $238 * $236;
  $239 = ((($window$151)) + -60|0);
  $240 = +HEAPF32[$239>>2];
  $241 = ((($b0$1$pn48)) + -4|0);
  $242 = +HEAPF32[$237>>2];
  $$neg28 = $242 * $240;
  $243 = ((($window$151)) + -64|0);
  $244 = +HEAPF32[$243>>2];
  $245 = +HEAPF32[$241>>2];
  $$neg29 = $245 * $244;
  $tmp30 = $$neg29 + $186;
  $tmp32 = $$neg28 + $$neg27;
  $tmp34 = $$neg26 + $$neg25;
  $tmp36 = $$neg24 + $$neg23;
  $tmp38 = $$neg22 + $$neg21;
  $tmp40 = $$neg20 + $$neg19;
  $tmp42 = $$neg18 + $$neg17;
  $tmp44 = $$neg16 + $$neg15;
  $tmp45 = -$tmp44;
  $246 = $tmp45 - $tmp42;
  $247 = $246 - $tmp40;
  $248 = $247 - $tmp38;
  $249 = $248 - $tmp36;
  $250 = $249 - $tmp34;
  $251 = $250 - $tmp32;
  $252 = $251 - $tmp30;
  $253 = $252 > 32767.0;
  do {
   if ($253) {
    HEAP16[$samples$253>>1] = 32767;
    $254 = (($clip$349) + 1)|0;
    $clip$4 = $254;
   } else {
    $255 = $252 < -32768.0;
    if ($255) {
     HEAP16[$samples$253>>1] = -32768;
     $256 = (($clip$349) + 1)|0;
     $clip$4 = $256;
     break;
    } else {
     $257 = (~~(($252)));
     HEAP16[$samples$253>>1] = $257;
     $clip$4 = $clip$349;
     break;
    }
   }
  } while(0);
  $258 = (($j$150) + -1)|0;
  $259 = ((($window$151)) + -128|0);
  $b0$2 = ((($b0$252)) + -64|0);
  $samples$2 = ((($samples$253)) + 4|0);
  $260 = ($258|0)==(0);
  if ($260) {
   $clip$4$lcssa = $clip$4;
   break;
  } else {
   $b0$1$pn48$phi = $b0$252;$b0$252 = $b0$2;$clip$349 = $clip$4;$j$150 = $258;$samples$253 = $samples$2;$window$151 = $259;$b0$1$pn48 = $b0$1$pn48$phi;
  }
 }
 $261 = ($final|0)==(0);
 if ($261) {
  return ($clip$4$lcssa|0);
 }
 $262 = HEAP32[$2>>2]|0;
 $263 = (($262) + 128)|0;
 HEAP32[$2>>2] = $263;
 return ($clip$4$lcssa|0);
}
function _INT123_synth_1to1_8bit($bandPtr,$channel,$fr,$final) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $fr = $fr|0;
 $final = $final|0;
 var $$neg = 0.0, $$neg1 = 0.0, $$neg15 = 0.0, $$neg16 = 0.0, $$neg17 = 0.0, $$neg18 = 0.0, $$neg19 = 0.0, $$neg2 = 0.0, $$neg20 = 0.0, $$neg21 = 0.0, $$neg22 = 0.0, $$neg23 = 0.0, $$neg24 = 0.0, $$neg25 = 0.0, $$neg26 = 0.0, $$neg27 = 0.0, $$neg28 = 0.0, $$neg29 = 0.0, $$neg3 = 0.0, $$neg4 = 0.0;
 var $$neg5 = 0.0, $$neg6 = 0.0, $$neg7 = 0.0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0, $102 = 0.0, $103 = 0, $104 = 0.0, $105 = 0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0, $11 = 0, $110 = 0.0;
 var $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0, $146 = 0.0, $147 = 0;
 var $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0, $158 = 0.0, $159 = 0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0, $164 = 0.0, $165 = 0;
 var $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0, $17 = 0, $170 = 0.0, $171 = 0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0, $176 = 0.0, $177 = 0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0.0, $198 = 0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0;
 var $201 = 0.0, $202 = 0, $203 = 0.0, $204 = 0, $205 = 0.0, $206 = 0, $207 = 0.0, $208 = 0, $209 = 0.0, $21 = 0, $210 = 0, $211 = 0.0, $212 = 0, $213 = 0.0, $214 = 0, $215 = 0.0, $216 = 0, $217 = 0.0, $218 = 0, $219 = 0.0;
 var $22 = 0, $220 = 0, $221 = 0.0, $222 = 0, $223 = 0.0, $224 = 0, $225 = 0.0, $226 = 0, $227 = 0.0, $228 = 0, $229 = 0.0, $23 = 0, $230 = 0, $231 = 0.0, $232 = 0, $233 = 0.0, $234 = 0, $235 = 0.0, $236 = 0, $237 = 0.0;
 var $238 = 0, $239 = 0.0, $24 = 0, $240 = 0, $241 = 0.0, $242 = 0, $243 = 0.0, $244 = 0, $245 = 0.0, $246 = 0, $247 = 0.0, $248 = 0, $249 = 0.0, $25 = 0, $250 = 0, $251 = 0.0, $252 = 0, $253 = 0.0, $254 = 0, $255 = 0.0;
 var $256 = 0.0, $257 = 0.0, $258 = 0.0, $259 = 0.0, $26 = 0, $260 = 0.0, $261 = 0.0, $262 = 0.0, $263 = 0.0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0;
 var $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0.0, $42 = 0, $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0;
 var $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0.0, $63 = 0, $64 = 0.0, $65 = 0, $66 = 0.0, $67 = 0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0, $73 = 0.0, $74 = 0, $75 = 0.0, $76 = 0;
 var $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0, $86 = 0.0, $87 = 0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0, $91 = 0.0, $92 = 0, $93 = 0.0, $94 = 0;
 var $95 = 0.0, $96 = 0, $97 = 0.0, $98 = 0.0, $99 = 0, $b0$0 = 0, $b0$1$pn48 = 0, $b0$1$pn48$phi = 0, $b0$155 = 0, $b0$2 = 0, $b0$252 = 0, $bo1$0 = 0, $buf$0 = 0, $clip$056 = 0, $clip$1 = 0, $clip$1$lcssa = 0, $clip$2 = 0, $clip$349 = 0, $clip$4 = 0, $clip$4$lcssa = 0;
 var $j$057 = 0, $j$150 = 0, $samples$0 = 0, $samples$154 = 0, $samples$2 = 0, $samples$247 = 0, $samples$253 = 0, $scevgep = 0, $scevgep61 = 0, $scevgep62 = 0, $scevgep64 = 0, $scevgep65 = 0, $tmp = 0.0, $tmp11 = 0.0, $tmp13 = 0.0, $tmp30 = 0.0, $tmp32 = 0.0, $tmp34 = 0.0, $tmp36 = 0.0, $tmp38 = 0.0;
 var $tmp40 = 0.0, $tmp42 = 0.0, $tmp44 = 0.0, $tmp45 = 0.0, $tmp9 = 0.0, $window$058 = 0, $window$151 = 0, $write_8bit_tmp$0 = 0, $write_8bit_tmp2$0 = 0, $write_8bit_tmp4$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($fr)) + 9288|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = ((($fr)) + 9292|0);
  _INT123_do_equalizer($bandPtr,$channel,$8);
 }
 $9 = ($channel|0)==(0);
 if ($9) {
  $10 = ((($fr)) + 9272|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + 15)|0;
  $13 = $12 & 15;
  HEAP32[$10>>2] = $13;
  $14 = ((($fr)) + 9248|0);
  $18 = $13;$buf$0 = $14;$samples$0 = $4;
 } else {
  $15 = ((($4)) + 1|0);
  $16 = ((($fr)) + 9256|0);
  $$phi$trans$insert = ((($fr)) + 9272|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $18 = $$pre;$buf$0 = $16;$samples$0 = $15;
 }
 $17 = $18 & 1;
 $19 = ($17|0)==(0);
 $20 = ((($buf$0)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 if ($19) {
  $27 = (($18) + 1)|0;
  $28 = HEAP32[$buf$0>>2]|0;
  $29 = (($28) + ($18<<2)|0);
  $30 = (($21) + ($18<<2)|0);
  $31 = ((($30)) + 4|0);
  _INT123_dct64($29,$31,$bandPtr);
  $b0$0 = $21;$bo1$0 = $27;
 } else {
  $22 = HEAP32[$buf$0>>2]|0;
  $23 = (($18) + 1)|0;
  $24 = $23 & 15;
  $25 = (($21) + ($24<<2)|0);
  $26 = (($22) + ($18<<2)|0);
  _INT123_dct64($25,$26,$bandPtr);
  $b0$0 = $22;$bo1$0 = $18;
 }
 $32 = ((($fr)) + 9284|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($33)) + 64|0);
 $35 = (0 - ($bo1$0))|0;
 $36 = (($34) + ($35<<2)|0);
 $37 = ((($fr)) + 9592|0);
 $38 = (528 - ($bo1$0))|0;
 $scevgep61 = (($33) + ($38<<2)|0);
 $39 = (496 - ($bo1$0))|0;
 $b0$155 = $b0$0;$clip$056 = 0;$j$057 = 16;$samples$154 = $samples$0;$window$058 = $36;
 while(1) {
  $40 = ((($window$058)) + 4|0);
  $41 = +HEAPF32[$window$058>>2];
  $42 = ((($b0$155)) + 4|0);
  $43 = +HEAPF32[$b0$155>>2];
  $44 = $43 * $41;
  $45 = ((($window$058)) + 8|0);
  $46 = +HEAPF32[$40>>2];
  $47 = ((($b0$155)) + 8|0);
  $48 = +HEAPF32[$42>>2];
  $$neg = $48 * $46;
  $49 = ((($window$058)) + 12|0);
  $50 = +HEAPF32[$45>>2];
  $51 = ((($b0$155)) + 12|0);
  $52 = +HEAPF32[$47>>2];
  $53 = $52 * $50;
  $54 = ((($window$058)) + 16|0);
  $55 = +HEAPF32[$49>>2];
  $56 = ((($b0$155)) + 16|0);
  $57 = +HEAPF32[$51>>2];
  $$neg1 = $57 * $55;
  $58 = ((($window$058)) + 20|0);
  $59 = +HEAPF32[$54>>2];
  $60 = ((($b0$155)) + 20|0);
  $61 = +HEAPF32[$56>>2];
  $62 = $61 * $59;
  $63 = ((($window$058)) + 24|0);
  $64 = +HEAPF32[$58>>2];
  $65 = ((($b0$155)) + 24|0);
  $66 = +HEAPF32[$60>>2];
  $$neg2 = $66 * $64;
  $67 = ((($window$058)) + 28|0);
  $68 = +HEAPF32[$63>>2];
  $69 = ((($b0$155)) + 28|0);
  $70 = +HEAPF32[$65>>2];
  $71 = $70 * $68;
  $72 = ((($window$058)) + 32|0);
  $73 = +HEAPF32[$67>>2];
  $74 = ((($b0$155)) + 32|0);
  $75 = +HEAPF32[$69>>2];
  $$neg3 = $75 * $73;
  $76 = ((($window$058)) + 36|0);
  $77 = +HEAPF32[$72>>2];
  $78 = ((($b0$155)) + 36|0);
  $79 = +HEAPF32[$74>>2];
  $80 = $79 * $77;
  $81 = ((($window$058)) + 40|0);
  $82 = +HEAPF32[$76>>2];
  $83 = ((($b0$155)) + 40|0);
  $84 = +HEAPF32[$78>>2];
  $$neg4 = $84 * $82;
  $85 = ((($window$058)) + 44|0);
  $86 = +HEAPF32[$81>>2];
  $87 = ((($b0$155)) + 44|0);
  $88 = +HEAPF32[$83>>2];
  $89 = $88 * $86;
  $90 = ((($window$058)) + 48|0);
  $91 = +HEAPF32[$85>>2];
  $92 = ((($b0$155)) + 48|0);
  $93 = +HEAPF32[$87>>2];
  $$neg5 = $93 * $91;
  $94 = ((($window$058)) + 52|0);
  $95 = +HEAPF32[$90>>2];
  $96 = ((($b0$155)) + 52|0);
  $97 = +HEAPF32[$92>>2];
  $98 = $97 * $95;
  $99 = ((($window$058)) + 56|0);
  $100 = +HEAPF32[$94>>2];
  $101 = ((($b0$155)) + 56|0);
  $102 = +HEAPF32[$96>>2];
  $$neg6 = $102 * $100;
  $103 = ((($window$058)) + 60|0);
  $104 = +HEAPF32[$99>>2];
  $105 = ((($b0$155)) + 60|0);
  $106 = +HEAPF32[$101>>2];
  $107 = $106 * $104;
  $108 = +HEAPF32[$103>>2];
  $109 = ((($b0$155)) + 64|0);
  $110 = +HEAPF32[$105>>2];
  $$neg7 = $110 * $108;
  $tmp = $$neg7 + $$neg6;
  $tmp9 = $$neg5 + $$neg4;
  $tmp11 = $$neg3 + $$neg2;
  $tmp13 = $$neg1 + $$neg;
  $111 = $53 + $44;
  $112 = $111 + $62;
  $113 = $112 + $71;
  $114 = $113 + $80;
  $115 = $114 + $89;
  $116 = $115 + $98;
  $117 = $116 + $107;
  $118 = $117 - $tmp13;
  $119 = $118 - $tmp11;
  $120 = $119 - $tmp9;
  $121 = $120 - $tmp;
  $122 = $121 > 32767.0;
  do {
   if ($122) {
    $123 = (($clip$056) + 1)|0;
    $clip$1 = $123;$write_8bit_tmp$0 = 32767;
   } else {
    $124 = $121 < -32768.0;
    if ($124) {
     $125 = (($clip$056) + 1)|0;
     $clip$1 = $125;$write_8bit_tmp$0 = -32768;
     break;
    } else {
     $126 = (~~(($121)));
     $clip$1 = $clip$056;$write_8bit_tmp$0 = $126;
     break;
    }
   }
  } while(0);
  $127 = $write_8bit_tmp$0 << 16 >> 16;
  $128 = $127 >> 3;
  $129 = HEAP32[$37>>2]|0;
  $130 = (($129) + ($128)|0);
  $131 = HEAP8[$130>>0]|0;
  HEAP8[$samples$154>>0] = $131;
  $132 = (($j$057) + -1)|0;
  $133 = ((($window$058)) + 128|0);
  $134 = ((($samples$154)) + 2|0);
  $135 = ($132|0)==(0);
  if ($135) {
   $clip$1$lcssa = $clip$1;
   break;
  } else {
   $b0$155 = $109;$clip$056 = $clip$1;$j$057 = $132;$samples$154 = $134;$window$058 = $133;
  }
 }
 $scevgep = ((($samples$0)) + 32|0);
 $scevgep62 = ((($b0$0)) + 1024|0);
 $scevgep64 = ((($b0$0)) + 960|0);
 $scevgep65 = (($33) + ($39<<2)|0);
 $136 = +HEAPF32[$scevgep61>>2];
 $137 = +HEAPF32[$scevgep62>>2];
 $138 = $137 * $136;
 $139 = ((($scevgep65)) + 136|0);
 $140 = +HEAPF32[$139>>2];
 $141 = ((($b0$0)) + 1032|0);
 $142 = +HEAPF32[$141>>2];
 $143 = $142 * $140;
 $144 = $143 + $138;
 $145 = ((($scevgep65)) + 144|0);
 $146 = +HEAPF32[$145>>2];
 $147 = ((($b0$0)) + 1040|0);
 $148 = +HEAPF32[$147>>2];
 $149 = $148 * $146;
 $150 = $144 + $149;
 $151 = ((($scevgep65)) + 152|0);
 $152 = +HEAPF32[$151>>2];
 $153 = ((($b0$0)) + 1048|0);
 $154 = +HEAPF32[$153>>2];
 $155 = $154 * $152;
 $156 = $150 + $155;
 $157 = ((($scevgep65)) + 160|0);
 $158 = +HEAPF32[$157>>2];
 $159 = ((($b0$0)) + 1056|0);
 $160 = +HEAPF32[$159>>2];
 $161 = $160 * $158;
 $162 = $156 + $161;
 $163 = ((($scevgep65)) + 168|0);
 $164 = +HEAPF32[$163>>2];
 $165 = ((($b0$0)) + 1064|0);
 $166 = +HEAPF32[$165>>2];
 $167 = $166 * $164;
 $168 = $162 + $167;
 $169 = ((($scevgep65)) + 176|0);
 $170 = +HEAPF32[$169>>2];
 $171 = ((($b0$0)) + 1072|0);
 $172 = +HEAPF32[$171>>2];
 $173 = $172 * $170;
 $174 = $168 + $173;
 $175 = ((($scevgep65)) + 184|0);
 $176 = +HEAPF32[$175>>2];
 $177 = ((($b0$0)) + 1080|0);
 $178 = +HEAPF32[$177>>2];
 $179 = $178 * $176;
 $180 = $174 + $179;
 $181 = $180 > 32767.0;
 do {
  if ($181) {
   $182 = (($clip$1$lcssa) + 1)|0;
   $clip$2 = $182;$write_8bit_tmp2$0 = 32767;
  } else {
   $183 = $180 < -32768.0;
   if ($183) {
    $184 = (($clip$1$lcssa) + 1)|0;
    $clip$2 = $184;$write_8bit_tmp2$0 = -32768;
    break;
   } else {
    $185 = (~~(($180)));
    $clip$2 = $clip$1$lcssa;$write_8bit_tmp2$0 = $185;
    break;
   }
  }
 } while(0);
 $186 = $write_8bit_tmp2$0 << 16 >> 16;
 $187 = $186 >> 3;
 $188 = HEAP32[$37>>2]|0;
 $189 = (($188) + ($187)|0);
 $190 = HEAP8[$189>>0]|0;
 HEAP8[$scevgep>>0] = $190;
 $191 = $bo1$0 << 1;
 $192 = (($scevgep65) + ($191<<2)|0);
 $samples$247 = ((($samples$0)) + 34|0);
 $b0$1$pn48 = $scevgep62;$b0$252 = $scevgep64;$clip$349 = $clip$2;$j$150 = 15;$samples$253 = $samples$247;$window$151 = $192;
 while(1) {
  $193 = ((($window$151)) + -4|0);
  $194 = +HEAPF32[$193>>2];
  $195 = ((($b0$1$pn48)) + -60|0);
  $196 = +HEAPF32[$b0$252>>2];
  $197 = $196 * $194;
  $198 = ((($window$151)) + -8|0);
  $199 = +HEAPF32[$198>>2];
  $200 = ((($b0$1$pn48)) + -56|0);
  $201 = +HEAPF32[$195>>2];
  $$neg15 = $201 * $199;
  $202 = ((($window$151)) + -12|0);
  $203 = +HEAPF32[$202>>2];
  $204 = ((($b0$1$pn48)) + -52|0);
  $205 = +HEAPF32[$200>>2];
  $$neg16 = $205 * $203;
  $206 = ((($window$151)) + -16|0);
  $207 = +HEAPF32[$206>>2];
  $208 = ((($b0$1$pn48)) + -48|0);
  $209 = +HEAPF32[$204>>2];
  $$neg17 = $209 * $207;
  $210 = ((($window$151)) + -20|0);
  $211 = +HEAPF32[$210>>2];
  $212 = ((($b0$1$pn48)) + -44|0);
  $213 = +HEAPF32[$208>>2];
  $$neg18 = $213 * $211;
  $214 = ((($window$151)) + -24|0);
  $215 = +HEAPF32[$214>>2];
  $216 = ((($b0$1$pn48)) + -40|0);
  $217 = +HEAPF32[$212>>2];
  $$neg19 = $217 * $215;
  $218 = ((($window$151)) + -28|0);
  $219 = +HEAPF32[$218>>2];
  $220 = ((($b0$1$pn48)) + -36|0);
  $221 = +HEAPF32[$216>>2];
  $$neg20 = $221 * $219;
  $222 = ((($window$151)) + -32|0);
  $223 = +HEAPF32[$222>>2];
  $224 = ((($b0$1$pn48)) + -32|0);
  $225 = +HEAPF32[$220>>2];
  $$neg21 = $225 * $223;
  $226 = ((($window$151)) + -36|0);
  $227 = +HEAPF32[$226>>2];
  $228 = ((($b0$1$pn48)) + -28|0);
  $229 = +HEAPF32[$224>>2];
  $$neg22 = $229 * $227;
  $230 = ((($window$151)) + -40|0);
  $231 = +HEAPF32[$230>>2];
  $232 = ((($b0$1$pn48)) + -24|0);
  $233 = +HEAPF32[$228>>2];
  $$neg23 = $233 * $231;
  $234 = ((($window$151)) + -44|0);
  $235 = +HEAPF32[$234>>2];
  $236 = ((($b0$1$pn48)) + -20|0);
  $237 = +HEAPF32[$232>>2];
  $$neg24 = $237 * $235;
  $238 = ((($window$151)) + -48|0);
  $239 = +HEAPF32[$238>>2];
  $240 = ((($b0$1$pn48)) + -16|0);
  $241 = +HEAPF32[$236>>2];
  $$neg25 = $241 * $239;
  $242 = ((($window$151)) + -52|0);
  $243 = +HEAPF32[$242>>2];
  $244 = ((($b0$1$pn48)) + -12|0);
  $245 = +HEAPF32[$240>>2];
  $$neg26 = $245 * $243;
  $246 = ((($window$151)) + -56|0);
  $247 = +HEAPF32[$246>>2];
  $248 = ((($b0$1$pn48)) + -8|0);
  $249 = +HEAPF32[$244>>2];
  $$neg27 = $249 * $247;
  $250 = ((($window$151)) + -60|0);
  $251 = +HEAPF32[$250>>2];
  $252 = ((($b0$1$pn48)) + -4|0);
  $253 = +HEAPF32[$248>>2];
  $$neg28 = $253 * $251;
  $254 = ((($window$151)) + -64|0);
  $255 = +HEAPF32[$254>>2];
  $256 = +HEAPF32[$252>>2];
  $$neg29 = $256 * $255;
  $tmp30 = $$neg29 + $197;
  $tmp32 = $$neg28 + $$neg27;
  $tmp34 = $$neg26 + $$neg25;
  $tmp36 = $$neg24 + $$neg23;
  $tmp38 = $$neg22 + $$neg21;
  $tmp40 = $$neg20 + $$neg19;
  $tmp42 = $$neg18 + $$neg17;
  $tmp44 = $$neg16 + $$neg15;
  $tmp45 = -$tmp44;
  $257 = $tmp45 - $tmp42;
  $258 = $257 - $tmp40;
  $259 = $258 - $tmp38;
  $260 = $259 - $tmp36;
  $261 = $260 - $tmp34;
  $262 = $261 - $tmp32;
  $263 = $262 - $tmp30;
  $264 = $263 > 32767.0;
  do {
   if ($264) {
    $265 = (($clip$349) + 1)|0;
    $clip$4 = $265;$write_8bit_tmp4$0 = 32767;
   } else {
    $266 = $263 < -32768.0;
    if ($266) {
     $267 = (($clip$349) + 1)|0;
     $clip$4 = $267;$write_8bit_tmp4$0 = -32768;
     break;
    } else {
     $268 = (~~(($263)));
     $clip$4 = $clip$349;$write_8bit_tmp4$0 = $268;
     break;
    }
   }
  } while(0);
  $269 = $write_8bit_tmp4$0 << 16 >> 16;
  $270 = $269 >> 3;
  $271 = HEAP32[$37>>2]|0;
  $272 = (($271) + ($270)|0);
  $273 = HEAP8[$272>>0]|0;
  HEAP8[$samples$253>>0] = $273;
  $274 = (($j$150) + -1)|0;
  $275 = ((($window$151)) + -128|0);
  $b0$2 = ((($b0$252)) + -64|0);
  $samples$2 = ((($samples$253)) + 2|0);
  $276 = ($274|0)==(0);
  if ($276) {
   $clip$4$lcssa = $clip$4;
   break;
  } else {
   $b0$1$pn48$phi = $b0$252;$b0$252 = $b0$2;$clip$349 = $clip$4;$j$150 = $274;$samples$253 = $samples$2;$window$151 = $275;$b0$1$pn48 = $b0$1$pn48$phi;
  }
 }
 $277 = ($final|0)==(0);
 if ($277) {
  return ($clip$4$lcssa|0);
 }
 $278 = HEAP32[$2>>2]|0;
 $279 = (($278) + 64)|0;
 HEAP32[$2>>2] = $279;
 return ($clip$4$lcssa|0);
}
function _INT123_synth_1to1_real($bandPtr,$channel,$fr,$final) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $fr = $fr|0;
 $final = $final|0;
 var $$neg = 0.0, $$neg1 = 0.0, $$neg15 = 0.0, $$neg16 = 0.0, $$neg17 = 0.0, $$neg18 = 0.0, $$neg19 = 0.0, $$neg2 = 0.0, $$neg20 = 0.0, $$neg21 = 0.0, $$neg22 = 0.0, $$neg23 = 0.0, $$neg24 = 0.0, $$neg25 = 0.0, $$neg26 = 0.0, $$neg27 = 0.0, $$neg28 = 0.0, $$neg29 = 0.0, $$neg3 = 0.0, $$neg4 = 0.0;
 var $$neg5 = 0.0, $$neg6 = 0.0, $$neg7 = 0.0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0, $109 = 0.0, $11 = 0, $110 = 0.0;
 var $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0;
 var $13 = 0, $130 = 0.0, $131 = 0, $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0, $136 = 0.0, $137 = 0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0;
 var $148 = 0.0, $149 = 0, $15 = 0, $150 = 0.0, $151 = 0.0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0, $16 = 0, $160 = 0.0, $161 = 0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0;
 var $166 = 0.0, $167 = 0, $168 = 0.0, $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0.0, $172 = 0, $173 = 0, $174 = 0, $175 = 0.0, $176 = 0, $177 = 0.0, $178 = 0.0, $179 = 0, $18 = 0, $180 = 0.0, $181 = 0, $182 = 0.0, $183 = 0;
 var $184 = 0.0, $185 = 0, $186 = 0.0, $187 = 0, $188 = 0.0, $189 = 0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0, $200 = 0.0;
 var $201 = 0, $202 = 0.0, $203 = 0, $204 = 0.0, $205 = 0, $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0.0, $213 = 0, $214 = 0.0, $215 = 0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0;
 var $22 = 0, $220 = 0.0, $221 = 0, $222 = 0.0, $223 = 0, $224 = 0.0, $225 = 0, $226 = 0.0, $227 = 0, $228 = 0.0, $229 = 0, $23 = 0, $230 = 0.0, $231 = 0, $232 = 0.0, $233 = 0, $234 = 0.0, $235 = 0, $236 = 0.0, $237 = 0.0;
 var $238 = 0.0, $239 = 0.0, $24 = 0, $240 = 0.0, $241 = 0.0, $242 = 0.0, $243 = 0.0, $244 = 0.0, $245 = 0.0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0, $45 = 0.0, $46 = 0, $47 = 0.0;
 var $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0.0;
 var $66 = 0, $67 = 0.0, $68 = 0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0, $83 = 0.0;
 var $84 = 0, $85 = 0.0, $86 = 0, $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0, $90 = 0.0, $91 = 0, $92 = 0.0, $93 = 0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0, $99 = 0.0, $b0$0 = 0, $b0$1$pn50 = 0, $b0$1$pn50$phi = 0;
 var $b0$155 = 0, $b0$2 = 0, $b0$251 = 0, $bo1$0 = 0, $buf$0 = 0, $j$053 = 0, $j$148 = 0, $samples$0$in = 0, $samples$156 = 0, $samples$2 = 0, $samples$247 = 0, $samples$252 = 0, $scevgep = 0, $scevgep60 = 0, $scevgep61 = 0, $scevgep62 = 0, $scevgep63 = 0, $tmp = 0.0, $tmp11 = 0.0, $tmp13 = 0.0;
 var $tmp30 = 0.0, $tmp32 = 0.0, $tmp34 = 0.0, $tmp36 = 0.0, $tmp38 = 0.0, $tmp40 = 0.0, $tmp42 = 0.0, $tmp44 = 0.0, $tmp45 = 0.0, $tmp9 = 0.0, $window$054 = 0, $window$149 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($fr)) + 9288|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = ((($fr)) + 9292|0);
  _INT123_do_equalizer($bandPtr,$channel,$8);
 }
 $9 = ($channel|0)==(0);
 if ($9) {
  $10 = ((($fr)) + 9272|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + 15)|0;
  $13 = $12 & 15;
  HEAP32[$10>>2] = $13;
  $14 = ((($fr)) + 9248|0);
  $18 = $13;$buf$0 = $14;$samples$0$in = $4;
 } else {
  $15 = ((($4)) + 4|0);
  $16 = ((($fr)) + 9256|0);
  $$phi$trans$insert = ((($fr)) + 9272|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $18 = $$pre;$buf$0 = $16;$samples$0$in = $15;
 }
 $17 = $18 & 1;
 $19 = ($17|0)==(0);
 $20 = ((($buf$0)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 if ($19) {
  $27 = (($18) + 1)|0;
  $28 = HEAP32[$buf$0>>2]|0;
  $29 = (($28) + ($18<<2)|0);
  $30 = (($21) + ($18<<2)|0);
  $31 = ((($30)) + 4|0);
  _INT123_dct64($29,$31,$bandPtr);
  $b0$0 = $21;$bo1$0 = $27;
 } else {
  $22 = HEAP32[$buf$0>>2]|0;
  $23 = (($18) + 1)|0;
  $24 = $23 & 15;
  $25 = (($21) + ($24<<2)|0);
  $26 = (($22) + ($18<<2)|0);
  _INT123_dct64($25,$26,$bandPtr);
  $b0$0 = $22;$bo1$0 = $18;
 }
 $32 = ((($fr)) + 9284|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($33)) + 64|0);
 $35 = (0 - ($bo1$0))|0;
 $36 = (($34) + ($35<<2)|0);
 $scevgep = ((($samples$0$in)) + 128|0);
 $37 = (528 - ($bo1$0))|0;
 $scevgep60 = (($33) + ($37<<2)|0);
 $38 = (496 - ($bo1$0))|0;
 $scevgep62 = (($33) + ($38<<2)|0);
 $b0$155 = $b0$0;$j$053 = 16;$samples$156 = $samples$0$in;$window$054 = $36;
 while(1) {
  $39 = ((($window$054)) + 4|0);
  $40 = +HEAPF32[$window$054>>2];
  $41 = ((($b0$155)) + 4|0);
  $42 = +HEAPF32[$b0$155>>2];
  $43 = $42 * $40;
  $44 = ((($window$054)) + 8|0);
  $45 = +HEAPF32[$39>>2];
  $46 = ((($b0$155)) + 8|0);
  $47 = +HEAPF32[$41>>2];
  $$neg = $47 * $45;
  $48 = ((($window$054)) + 12|0);
  $49 = +HEAPF32[$44>>2];
  $50 = ((($b0$155)) + 12|0);
  $51 = +HEAPF32[$46>>2];
  $52 = $51 * $49;
  $53 = ((($window$054)) + 16|0);
  $54 = +HEAPF32[$48>>2];
  $55 = ((($b0$155)) + 16|0);
  $56 = +HEAPF32[$50>>2];
  $$neg1 = $56 * $54;
  $57 = ((($window$054)) + 20|0);
  $58 = +HEAPF32[$53>>2];
  $59 = ((($b0$155)) + 20|0);
  $60 = +HEAPF32[$55>>2];
  $61 = $60 * $58;
  $62 = ((($window$054)) + 24|0);
  $63 = +HEAPF32[$57>>2];
  $64 = ((($b0$155)) + 24|0);
  $65 = +HEAPF32[$59>>2];
  $$neg2 = $65 * $63;
  $66 = ((($window$054)) + 28|0);
  $67 = +HEAPF32[$62>>2];
  $68 = ((($b0$155)) + 28|0);
  $69 = +HEAPF32[$64>>2];
  $70 = $69 * $67;
  $71 = ((($window$054)) + 32|0);
  $72 = +HEAPF32[$66>>2];
  $73 = ((($b0$155)) + 32|0);
  $74 = +HEAPF32[$68>>2];
  $$neg3 = $74 * $72;
  $75 = ((($window$054)) + 36|0);
  $76 = +HEAPF32[$71>>2];
  $77 = ((($b0$155)) + 36|0);
  $78 = +HEAPF32[$73>>2];
  $79 = $78 * $76;
  $80 = ((($window$054)) + 40|0);
  $81 = +HEAPF32[$75>>2];
  $82 = ((($b0$155)) + 40|0);
  $83 = +HEAPF32[$77>>2];
  $$neg4 = $83 * $81;
  $84 = ((($window$054)) + 44|0);
  $85 = +HEAPF32[$80>>2];
  $86 = ((($b0$155)) + 44|0);
  $87 = +HEAPF32[$82>>2];
  $88 = $87 * $85;
  $89 = ((($window$054)) + 48|0);
  $90 = +HEAPF32[$84>>2];
  $91 = ((($b0$155)) + 48|0);
  $92 = +HEAPF32[$86>>2];
  $$neg5 = $92 * $90;
  $93 = ((($window$054)) + 52|0);
  $94 = +HEAPF32[$89>>2];
  $95 = ((($b0$155)) + 52|0);
  $96 = +HEAPF32[$91>>2];
  $97 = $96 * $94;
  $98 = ((($window$054)) + 56|0);
  $99 = +HEAPF32[$93>>2];
  $100 = ((($b0$155)) + 56|0);
  $101 = +HEAPF32[$95>>2];
  $$neg6 = $101 * $99;
  $102 = ((($window$054)) + 60|0);
  $103 = +HEAPF32[$98>>2];
  $104 = ((($b0$155)) + 60|0);
  $105 = +HEAPF32[$100>>2];
  $106 = $105 * $103;
  $107 = +HEAPF32[$102>>2];
  $108 = ((($b0$155)) + 64|0);
  $109 = +HEAPF32[$104>>2];
  $$neg7 = $109 * $107;
  $tmp = $$neg7 + $$neg6;
  $tmp9 = $$neg5 + $$neg4;
  $tmp11 = $$neg3 + $$neg2;
  $tmp13 = $$neg1 + $$neg;
  $110 = $52 + $43;
  $111 = $110 + $61;
  $112 = $111 + $70;
  $113 = $112 + $79;
  $114 = $113 + $88;
  $115 = $114 + $97;
  $116 = $115 + $106;
  $117 = $116 - $tmp13;
  $118 = $117 - $tmp11;
  $119 = $118 - $tmp9;
  $120 = $119 - $tmp;
  $121 = $120 * 3.0517578125E-5;
  HEAPF32[$samples$156>>2] = $121;
  $122 = (($j$053) + -1)|0;
  $123 = ((($window$054)) + 128|0);
  $124 = ((($samples$156)) + 8|0);
  $125 = ($122|0)==(0);
  if ($125) {
   break;
  } else {
   $b0$155 = $108;$j$053 = $122;$samples$156 = $124;$window$054 = $123;
  }
 }
 $scevgep61 = ((($b0$0)) + 1024|0);
 $scevgep63 = ((($b0$0)) + 960|0);
 $126 = +HEAPF32[$scevgep60>>2];
 $127 = +HEAPF32[$scevgep61>>2];
 $128 = $127 * $126;
 $129 = ((($scevgep62)) + 136|0);
 $130 = +HEAPF32[$129>>2];
 $131 = ((($b0$0)) + 1032|0);
 $132 = +HEAPF32[$131>>2];
 $133 = $132 * $130;
 $134 = $133 + $128;
 $135 = ((($scevgep62)) + 144|0);
 $136 = +HEAPF32[$135>>2];
 $137 = ((($b0$0)) + 1040|0);
 $138 = +HEAPF32[$137>>2];
 $139 = $138 * $136;
 $140 = $134 + $139;
 $141 = ((($scevgep62)) + 152|0);
 $142 = +HEAPF32[$141>>2];
 $143 = ((($b0$0)) + 1048|0);
 $144 = +HEAPF32[$143>>2];
 $145 = $144 * $142;
 $146 = $140 + $145;
 $147 = ((($scevgep62)) + 160|0);
 $148 = +HEAPF32[$147>>2];
 $149 = ((($b0$0)) + 1056|0);
 $150 = +HEAPF32[$149>>2];
 $151 = $150 * $148;
 $152 = $146 + $151;
 $153 = ((($scevgep62)) + 168|0);
 $154 = +HEAPF32[$153>>2];
 $155 = ((($b0$0)) + 1064|0);
 $156 = +HEAPF32[$155>>2];
 $157 = $156 * $154;
 $158 = $152 + $157;
 $159 = ((($scevgep62)) + 176|0);
 $160 = +HEAPF32[$159>>2];
 $161 = ((($b0$0)) + 1072|0);
 $162 = +HEAPF32[$161>>2];
 $163 = $162 * $160;
 $164 = $158 + $163;
 $165 = ((($scevgep62)) + 184|0);
 $166 = +HEAPF32[$165>>2];
 $167 = ((($b0$0)) + 1080|0);
 $168 = +HEAPF32[$167>>2];
 $169 = $168 * $166;
 $170 = $164 + $169;
 $171 = $170 * 3.0517578125E-5;
 HEAPF32[$scevgep>>2] = $171;
 $172 = $bo1$0 << 1;
 $173 = (($scevgep62) + ($172<<2)|0);
 $samples$247 = ((($samples$0$in)) + 136|0);
 $b0$1$pn50 = $scevgep61;$b0$251 = $scevgep63;$j$148 = 15;$samples$252 = $samples$247;$window$149 = $173;
 while(1) {
  $174 = ((($window$149)) + -4|0);
  $175 = +HEAPF32[$174>>2];
  $176 = ((($b0$1$pn50)) + -60|0);
  $177 = +HEAPF32[$b0$251>>2];
  $178 = $177 * $175;
  $179 = ((($window$149)) + -8|0);
  $180 = +HEAPF32[$179>>2];
  $181 = ((($b0$1$pn50)) + -56|0);
  $182 = +HEAPF32[$176>>2];
  $$neg15 = $182 * $180;
  $183 = ((($window$149)) + -12|0);
  $184 = +HEAPF32[$183>>2];
  $185 = ((($b0$1$pn50)) + -52|0);
  $186 = +HEAPF32[$181>>2];
  $$neg16 = $186 * $184;
  $187 = ((($window$149)) + -16|0);
  $188 = +HEAPF32[$187>>2];
  $189 = ((($b0$1$pn50)) + -48|0);
  $190 = +HEAPF32[$185>>2];
  $$neg17 = $190 * $188;
  $191 = ((($window$149)) + -20|0);
  $192 = +HEAPF32[$191>>2];
  $193 = ((($b0$1$pn50)) + -44|0);
  $194 = +HEAPF32[$189>>2];
  $$neg18 = $194 * $192;
  $195 = ((($window$149)) + -24|0);
  $196 = +HEAPF32[$195>>2];
  $197 = ((($b0$1$pn50)) + -40|0);
  $198 = +HEAPF32[$193>>2];
  $$neg19 = $198 * $196;
  $199 = ((($window$149)) + -28|0);
  $200 = +HEAPF32[$199>>2];
  $201 = ((($b0$1$pn50)) + -36|0);
  $202 = +HEAPF32[$197>>2];
  $$neg20 = $202 * $200;
  $203 = ((($window$149)) + -32|0);
  $204 = +HEAPF32[$203>>2];
  $205 = ((($b0$1$pn50)) + -32|0);
  $206 = +HEAPF32[$201>>2];
  $$neg21 = $206 * $204;
  $207 = ((($window$149)) + -36|0);
  $208 = +HEAPF32[$207>>2];
  $209 = ((($b0$1$pn50)) + -28|0);
  $210 = +HEAPF32[$205>>2];
  $$neg22 = $210 * $208;
  $211 = ((($window$149)) + -40|0);
  $212 = +HEAPF32[$211>>2];
  $213 = ((($b0$1$pn50)) + -24|0);
  $214 = +HEAPF32[$209>>2];
  $$neg23 = $214 * $212;
  $215 = ((($window$149)) + -44|0);
  $216 = +HEAPF32[$215>>2];
  $217 = ((($b0$1$pn50)) + -20|0);
  $218 = +HEAPF32[$213>>2];
  $$neg24 = $218 * $216;
  $219 = ((($window$149)) + -48|0);
  $220 = +HEAPF32[$219>>2];
  $221 = ((($b0$1$pn50)) + -16|0);
  $222 = +HEAPF32[$217>>2];
  $$neg25 = $222 * $220;
  $223 = ((($window$149)) + -52|0);
  $224 = +HEAPF32[$223>>2];
  $225 = ((($b0$1$pn50)) + -12|0);
  $226 = +HEAPF32[$221>>2];
  $$neg26 = $226 * $224;
  $227 = ((($window$149)) + -56|0);
  $228 = +HEAPF32[$227>>2];
  $229 = ((($b0$1$pn50)) + -8|0);
  $230 = +HEAPF32[$225>>2];
  $$neg27 = $230 * $228;
  $231 = ((($window$149)) + -60|0);
  $232 = +HEAPF32[$231>>2];
  $233 = ((($b0$1$pn50)) + -4|0);
  $234 = +HEAPF32[$229>>2];
  $$neg28 = $234 * $232;
  $235 = ((($window$149)) + -64|0);
  $236 = +HEAPF32[$235>>2];
  $237 = +HEAPF32[$233>>2];
  $$neg29 = $237 * $236;
  $tmp30 = $$neg29 + $178;
  $tmp32 = $$neg28 + $$neg27;
  $tmp34 = $$neg26 + $$neg25;
  $tmp36 = $$neg24 + $$neg23;
  $tmp38 = $$neg22 + $$neg21;
  $tmp40 = $$neg20 + $$neg19;
  $tmp42 = $$neg18 + $$neg17;
  $tmp44 = $$neg16 + $$neg15;
  $tmp45 = -$tmp44;
  $238 = $tmp45 - $tmp42;
  $239 = $238 - $tmp40;
  $240 = $239 - $tmp38;
  $241 = $240 - $tmp36;
  $242 = $241 - $tmp34;
  $243 = $242 - $tmp32;
  $244 = $243 - $tmp30;
  $245 = $244 * 3.0517578125E-5;
  HEAPF32[$samples$252>>2] = $245;
  $246 = (($j$148) + -1)|0;
  $247 = ((($window$149)) + -128|0);
  $b0$2 = ((($b0$251)) + -64|0);
  $samples$2 = ((($samples$252)) + 8|0);
  $248 = ($246|0)==(0);
  if ($248) {
   break;
  } else {
   $b0$1$pn50$phi = $b0$251;$b0$251 = $b0$2;$j$148 = $246;$samples$252 = $samples$2;$window$149 = $247;$b0$1$pn50 = $b0$1$pn50$phi;
  }
 }
 $249 = ($final|0)==(0);
 if ($249) {
  return 0;
 }
 $250 = HEAP32[$2>>2]|0;
 $251 = (($250) + 256)|0;
 HEAP32[$2>>2] = $251;
 return 0;
}
function _INT123_synth_1to1_s32($bandPtr,$channel,$fr,$final) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $fr = $fr|0;
 $final = $final|0;
 var $$neg = 0.0, $$neg1 = 0.0, $$neg15 = 0.0, $$neg16 = 0.0, $$neg17 = 0.0, $$neg18 = 0.0, $$neg19 = 0.0, $$neg2 = 0.0, $$neg20 = 0.0, $$neg21 = 0.0, $$neg22 = 0.0, $$neg23 = 0.0, $$neg24 = 0.0, $$neg25 = 0.0, $$neg26 = 0.0, $$neg27 = 0.0, $$neg28 = 0.0, $$neg29 = 0.0, $$neg3 = 0.0, $$neg4 = 0.0;
 var $$neg5 = 0.0, $$neg6 = 0.0, $$neg7 = 0.0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0, $109 = 0.0, $11 = 0, $110 = 0.0;
 var $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0, $136 = 0.0, $137 = 0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0;
 var $148 = 0.0, $149 = 0, $15 = 0, $150 = 0.0, $151 = 0.0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0, $16 = 0, $160 = 0.0, $161 = 0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0;
 var $166 = 0.0, $167 = 0, $168 = 0.0, $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0, $172 = 0.0, $173 = 0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0.0, $188 = 0, $189 = 0.0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0, $200 = 0.0;
 var $201 = 0, $202 = 0.0, $203 = 0, $204 = 0.0, $205 = 0, $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0.0, $213 = 0, $214 = 0.0, $215 = 0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0;
 var $22 = 0, $220 = 0.0, $221 = 0, $222 = 0.0, $223 = 0, $224 = 0.0, $225 = 0, $226 = 0.0, $227 = 0, $228 = 0.0, $229 = 0, $23 = 0, $230 = 0.0, $231 = 0, $232 = 0.0, $233 = 0, $234 = 0.0, $235 = 0, $236 = 0.0, $237 = 0;
 var $238 = 0.0, $239 = 0, $24 = 0, $240 = 0.0, $241 = 0, $242 = 0.0, $243 = 0, $244 = 0.0, $245 = 0, $246 = 0.0, $247 = 0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0.0, $254 = 0.0, $255 = 0.0;
 var $256 = 0.0, $257 = 0.0, $258 = 0.0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0, $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0, $49 = 0.0;
 var $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0;
 var $68 = 0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0, $85 = 0.0;
 var $86 = 0, $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0, $90 = 0.0, $91 = 0, $92 = 0.0, $93 = 0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0, $99 = 0.0, $b0$0 = 0, $b0$1$pn51 = 0, $b0$1$pn51$phi = 0, $b0$157 = 0, $b0$2 = 0;
 var $b0$252 = 0, $bo1$0 = 0, $buf$0 = 0, $clip$056 = 0, $clip$1 = 0, $clip$1$lcssa = 0, $clip$2 = 0, $clip$350 = 0, $clip$4 = 0, $clip$4$lcssa = 0, $j$055 = 0, $j$149 = 0, $samples$0$in = 0, $samples$158 = 0, $samples$2 = 0, $samples$247 = 0, $samples$253 = 0, $scevgep = 0, $scevgep62 = 0, $scevgep63 = 0;
 var $scevgep64 = 0, $scevgep65 = 0, $tmp = 0.0, $tmp11 = 0.0, $tmp13 = 0.0, $tmp30 = 0.0, $tmp32 = 0.0, $tmp34 = 0.0, $tmp36 = 0.0, $tmp38 = 0.0, $tmp40 = 0.0, $tmp42 = 0.0, $tmp44 = 0.0, $tmp45 = 0.0, $tmp9 = 0.0, $window$054 = 0, $window$148 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($fr)) + 9288|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = ((($fr)) + 9292|0);
  _INT123_do_equalizer($bandPtr,$channel,$8);
 }
 $9 = ($channel|0)==(0);
 if ($9) {
  $10 = ((($fr)) + 9272|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + 15)|0;
  $13 = $12 & 15;
  HEAP32[$10>>2] = $13;
  $14 = ((($fr)) + 9248|0);
  $18 = $13;$buf$0 = $14;$samples$0$in = $4;
 } else {
  $15 = ((($4)) + 4|0);
  $16 = ((($fr)) + 9256|0);
  $$phi$trans$insert = ((($fr)) + 9272|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $18 = $$pre;$buf$0 = $16;$samples$0$in = $15;
 }
 $17 = $18 & 1;
 $19 = ($17|0)==(0);
 $20 = ((($buf$0)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 if ($19) {
  $27 = (($18) + 1)|0;
  $28 = HEAP32[$buf$0>>2]|0;
  $29 = (($28) + ($18<<2)|0);
  $30 = (($21) + ($18<<2)|0);
  $31 = ((($30)) + 4|0);
  _INT123_dct64($29,$31,$bandPtr);
  $b0$0 = $21;$bo1$0 = $27;
 } else {
  $22 = HEAP32[$buf$0>>2]|0;
  $23 = (($18) + 1)|0;
  $24 = $23 & 15;
  $25 = (($21) + ($24<<2)|0);
  $26 = (($22) + ($18<<2)|0);
  _INT123_dct64($25,$26,$bandPtr);
  $b0$0 = $22;$bo1$0 = $18;
 }
 $32 = ((($fr)) + 9284|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($33)) + 64|0);
 $35 = (0 - ($bo1$0))|0;
 $36 = (($34) + ($35<<2)|0);
 $scevgep = ((($samples$0$in)) + 128|0);
 $37 = (528 - ($bo1$0))|0;
 $scevgep62 = (($33) + ($37<<2)|0);
 $38 = (496 - ($bo1$0))|0;
 $scevgep64 = (($33) + ($38<<2)|0);
 $b0$157 = $b0$0;$clip$056 = 0;$j$055 = 16;$samples$158 = $samples$0$in;$window$054 = $36;
 while(1) {
  $39 = ((($window$054)) + 4|0);
  $40 = +HEAPF32[$window$054>>2];
  $41 = ((($b0$157)) + 4|0);
  $42 = +HEAPF32[$b0$157>>2];
  $43 = $42 * $40;
  $44 = ((($window$054)) + 8|0);
  $45 = +HEAPF32[$39>>2];
  $46 = ((($b0$157)) + 8|0);
  $47 = +HEAPF32[$41>>2];
  $$neg = $47 * $45;
  $48 = ((($window$054)) + 12|0);
  $49 = +HEAPF32[$44>>2];
  $50 = ((($b0$157)) + 12|0);
  $51 = +HEAPF32[$46>>2];
  $52 = $51 * $49;
  $53 = ((($window$054)) + 16|0);
  $54 = +HEAPF32[$48>>2];
  $55 = ((($b0$157)) + 16|0);
  $56 = +HEAPF32[$50>>2];
  $$neg1 = $56 * $54;
  $57 = ((($window$054)) + 20|0);
  $58 = +HEAPF32[$53>>2];
  $59 = ((($b0$157)) + 20|0);
  $60 = +HEAPF32[$55>>2];
  $61 = $60 * $58;
  $62 = ((($window$054)) + 24|0);
  $63 = +HEAPF32[$57>>2];
  $64 = ((($b0$157)) + 24|0);
  $65 = +HEAPF32[$59>>2];
  $$neg2 = $65 * $63;
  $66 = ((($window$054)) + 28|0);
  $67 = +HEAPF32[$62>>2];
  $68 = ((($b0$157)) + 28|0);
  $69 = +HEAPF32[$64>>2];
  $70 = $69 * $67;
  $71 = ((($window$054)) + 32|0);
  $72 = +HEAPF32[$66>>2];
  $73 = ((($b0$157)) + 32|0);
  $74 = +HEAPF32[$68>>2];
  $$neg3 = $74 * $72;
  $75 = ((($window$054)) + 36|0);
  $76 = +HEAPF32[$71>>2];
  $77 = ((($b0$157)) + 36|0);
  $78 = +HEAPF32[$73>>2];
  $79 = $78 * $76;
  $80 = ((($window$054)) + 40|0);
  $81 = +HEAPF32[$75>>2];
  $82 = ((($b0$157)) + 40|0);
  $83 = +HEAPF32[$77>>2];
  $$neg4 = $83 * $81;
  $84 = ((($window$054)) + 44|0);
  $85 = +HEAPF32[$80>>2];
  $86 = ((($b0$157)) + 44|0);
  $87 = +HEAPF32[$82>>2];
  $88 = $87 * $85;
  $89 = ((($window$054)) + 48|0);
  $90 = +HEAPF32[$84>>2];
  $91 = ((($b0$157)) + 48|0);
  $92 = +HEAPF32[$86>>2];
  $$neg5 = $92 * $90;
  $93 = ((($window$054)) + 52|0);
  $94 = +HEAPF32[$89>>2];
  $95 = ((($b0$157)) + 52|0);
  $96 = +HEAPF32[$91>>2];
  $97 = $96 * $94;
  $98 = ((($window$054)) + 56|0);
  $99 = +HEAPF32[$93>>2];
  $100 = ((($b0$157)) + 56|0);
  $101 = +HEAPF32[$95>>2];
  $$neg6 = $101 * $99;
  $102 = ((($window$054)) + 60|0);
  $103 = +HEAPF32[$98>>2];
  $104 = ((($b0$157)) + 60|0);
  $105 = +HEAPF32[$100>>2];
  $106 = $105 * $103;
  $107 = +HEAPF32[$102>>2];
  $108 = ((($b0$157)) + 64|0);
  $109 = +HEAPF32[$104>>2];
  $$neg7 = $109 * $107;
  $tmp = $$neg7 + $$neg6;
  $tmp9 = $$neg5 + $$neg4;
  $tmp11 = $$neg3 + $$neg2;
  $tmp13 = $$neg1 + $$neg;
  $110 = $52 + $43;
  $111 = $110 + $61;
  $112 = $111 + $70;
  $113 = $112 + $79;
  $114 = $113 + $88;
  $115 = $114 + $97;
  $116 = $115 + $106;
  $117 = $116 - $tmp13;
  $118 = $117 - $tmp11;
  $119 = $118 - $tmp9;
  $120 = $119 - $tmp;
  $121 = $120 * 65536.0;
  $122 = $121;
  $123 = $122 > 2147483647.0;
  do {
   if ($123) {
    HEAP32[$samples$158>>2] = 2147483647;
    $124 = (($clip$056) + 1)|0;
    $clip$1 = $124;
   } else {
    $125 = $121 < -2147483648.0;
    if ($125) {
     HEAP32[$samples$158>>2] = -2147483648;
     $126 = (($clip$056) + 1)|0;
     $clip$1 = $126;
     break;
    } else {
     $127 = (~~(($121)));
     HEAP32[$samples$158>>2] = $127;
     $clip$1 = $clip$056;
     break;
    }
   }
  } while(0);
  $128 = (($j$055) + -1)|0;
  $129 = ((($window$054)) + 128|0);
  $130 = ((($samples$158)) + 8|0);
  $131 = ($128|0)==(0);
  if ($131) {
   $clip$1$lcssa = $clip$1;
   break;
  } else {
   $b0$157 = $108;$clip$056 = $clip$1;$j$055 = $128;$samples$158 = $130;$window$054 = $129;
  }
 }
 $scevgep63 = ((($b0$0)) + 1024|0);
 $scevgep65 = ((($b0$0)) + 960|0);
 $132 = +HEAPF32[$scevgep62>>2];
 $133 = +HEAPF32[$scevgep63>>2];
 $134 = $133 * $132;
 $135 = ((($scevgep64)) + 136|0);
 $136 = +HEAPF32[$135>>2];
 $137 = ((($b0$0)) + 1032|0);
 $138 = +HEAPF32[$137>>2];
 $139 = $138 * $136;
 $140 = $139 + $134;
 $141 = ((($scevgep64)) + 144|0);
 $142 = +HEAPF32[$141>>2];
 $143 = ((($b0$0)) + 1040|0);
 $144 = +HEAPF32[$143>>2];
 $145 = $144 * $142;
 $146 = $140 + $145;
 $147 = ((($scevgep64)) + 152|0);
 $148 = +HEAPF32[$147>>2];
 $149 = ((($b0$0)) + 1048|0);
 $150 = +HEAPF32[$149>>2];
 $151 = $150 * $148;
 $152 = $146 + $151;
 $153 = ((($scevgep64)) + 160|0);
 $154 = +HEAPF32[$153>>2];
 $155 = ((($b0$0)) + 1056|0);
 $156 = +HEAPF32[$155>>2];
 $157 = $156 * $154;
 $158 = $152 + $157;
 $159 = ((($scevgep64)) + 168|0);
 $160 = +HEAPF32[$159>>2];
 $161 = ((($b0$0)) + 1064|0);
 $162 = +HEAPF32[$161>>2];
 $163 = $162 * $160;
 $164 = $158 + $163;
 $165 = ((($scevgep64)) + 176|0);
 $166 = +HEAPF32[$165>>2];
 $167 = ((($b0$0)) + 1072|0);
 $168 = +HEAPF32[$167>>2];
 $169 = $168 * $166;
 $170 = $164 + $169;
 $171 = ((($scevgep64)) + 184|0);
 $172 = +HEAPF32[$171>>2];
 $173 = ((($b0$0)) + 1080|0);
 $174 = +HEAPF32[$173>>2];
 $175 = $174 * $172;
 $176 = $170 + $175;
 $177 = $176 * 65536.0;
 $178 = $177;
 $179 = $178 > 2147483647.0;
 do {
  if ($179) {
   HEAP32[$scevgep>>2] = 2147483647;
   $180 = (($clip$1$lcssa) + 1)|0;
   $clip$2 = $180;
  } else {
   $181 = $177 < -2147483648.0;
   if ($181) {
    HEAP32[$scevgep>>2] = -2147483648;
    $182 = (($clip$1$lcssa) + 1)|0;
    $clip$2 = $182;
    break;
   } else {
    $183 = (~~(($177)));
    HEAP32[$scevgep>>2] = $183;
    $clip$2 = $clip$1$lcssa;
    break;
   }
  }
 } while(0);
 $184 = $bo1$0 << 1;
 $185 = (($scevgep64) + ($184<<2)|0);
 $samples$247 = ((($samples$0$in)) + 136|0);
 $b0$1$pn51 = $scevgep63;$b0$252 = $scevgep65;$clip$350 = $clip$2;$j$149 = 15;$samples$253 = $samples$247;$window$148 = $185;
 while(1) {
  $186 = ((($window$148)) + -4|0);
  $187 = +HEAPF32[$186>>2];
  $188 = ((($b0$1$pn51)) + -60|0);
  $189 = +HEAPF32[$b0$252>>2];
  $190 = $189 * $187;
  $191 = ((($window$148)) + -8|0);
  $192 = +HEAPF32[$191>>2];
  $193 = ((($b0$1$pn51)) + -56|0);
  $194 = +HEAPF32[$188>>2];
  $$neg15 = $194 * $192;
  $195 = ((($window$148)) + -12|0);
  $196 = +HEAPF32[$195>>2];
  $197 = ((($b0$1$pn51)) + -52|0);
  $198 = +HEAPF32[$193>>2];
  $$neg16 = $198 * $196;
  $199 = ((($window$148)) + -16|0);
  $200 = +HEAPF32[$199>>2];
  $201 = ((($b0$1$pn51)) + -48|0);
  $202 = +HEAPF32[$197>>2];
  $$neg17 = $202 * $200;
  $203 = ((($window$148)) + -20|0);
  $204 = +HEAPF32[$203>>2];
  $205 = ((($b0$1$pn51)) + -44|0);
  $206 = +HEAPF32[$201>>2];
  $$neg18 = $206 * $204;
  $207 = ((($window$148)) + -24|0);
  $208 = +HEAPF32[$207>>2];
  $209 = ((($b0$1$pn51)) + -40|0);
  $210 = +HEAPF32[$205>>2];
  $$neg19 = $210 * $208;
  $211 = ((($window$148)) + -28|0);
  $212 = +HEAPF32[$211>>2];
  $213 = ((($b0$1$pn51)) + -36|0);
  $214 = +HEAPF32[$209>>2];
  $$neg20 = $214 * $212;
  $215 = ((($window$148)) + -32|0);
  $216 = +HEAPF32[$215>>2];
  $217 = ((($b0$1$pn51)) + -32|0);
  $218 = +HEAPF32[$213>>2];
  $$neg21 = $218 * $216;
  $219 = ((($window$148)) + -36|0);
  $220 = +HEAPF32[$219>>2];
  $221 = ((($b0$1$pn51)) + -28|0);
  $222 = +HEAPF32[$217>>2];
  $$neg22 = $222 * $220;
  $223 = ((($window$148)) + -40|0);
  $224 = +HEAPF32[$223>>2];
  $225 = ((($b0$1$pn51)) + -24|0);
  $226 = +HEAPF32[$221>>2];
  $$neg23 = $226 * $224;
  $227 = ((($window$148)) + -44|0);
  $228 = +HEAPF32[$227>>2];
  $229 = ((($b0$1$pn51)) + -20|0);
  $230 = +HEAPF32[$225>>2];
  $$neg24 = $230 * $228;
  $231 = ((($window$148)) + -48|0);
  $232 = +HEAPF32[$231>>2];
  $233 = ((($b0$1$pn51)) + -16|0);
  $234 = +HEAPF32[$229>>2];
  $$neg25 = $234 * $232;
  $235 = ((($window$148)) + -52|0);
  $236 = +HEAPF32[$235>>2];
  $237 = ((($b0$1$pn51)) + -12|0);
  $238 = +HEAPF32[$233>>2];
  $$neg26 = $238 * $236;
  $239 = ((($window$148)) + -56|0);
  $240 = +HEAPF32[$239>>2];
  $241 = ((($b0$1$pn51)) + -8|0);
  $242 = +HEAPF32[$237>>2];
  $$neg27 = $242 * $240;
  $243 = ((($window$148)) + -60|0);
  $244 = +HEAPF32[$243>>2];
  $245 = ((($b0$1$pn51)) + -4|0);
  $246 = +HEAPF32[$241>>2];
  $$neg28 = $246 * $244;
  $247 = ((($window$148)) + -64|0);
  $248 = +HEAPF32[$247>>2];
  $249 = +HEAPF32[$245>>2];
  $$neg29 = $249 * $248;
  $tmp30 = $$neg29 + $190;
  $tmp32 = $$neg28 + $$neg27;
  $tmp34 = $$neg26 + $$neg25;
  $tmp36 = $$neg24 + $$neg23;
  $tmp38 = $$neg22 + $$neg21;
  $tmp40 = $$neg20 + $$neg19;
  $tmp42 = $$neg18 + $$neg17;
  $tmp44 = $$neg16 + $$neg15;
  $tmp45 = -$tmp44;
  $250 = $tmp45 - $tmp42;
  $251 = $250 - $tmp40;
  $252 = $251 - $tmp38;
  $253 = $252 - $tmp36;
  $254 = $253 - $tmp34;
  $255 = $254 - $tmp32;
  $256 = $255 - $tmp30;
  $257 = $256 * 65536.0;
  $258 = $257;
  $259 = $258 > 2147483647.0;
  do {
   if ($259) {
    HEAP32[$samples$253>>2] = 2147483647;
    $260 = (($clip$350) + 1)|0;
    $clip$4 = $260;
   } else {
    $261 = $257 < -2147483648.0;
    if ($261) {
     HEAP32[$samples$253>>2] = -2147483648;
     $262 = (($clip$350) + 1)|0;
     $clip$4 = $262;
     break;
    } else {
     $263 = (~~(($257)));
     HEAP32[$samples$253>>2] = $263;
     $clip$4 = $clip$350;
     break;
    }
   }
  } while(0);
  $264 = (($j$149) + -1)|0;
  $265 = ((($window$148)) + -128|0);
  $b0$2 = ((($b0$252)) + -64|0);
  $samples$2 = ((($samples$253)) + 8|0);
  $266 = ($264|0)==(0);
  if ($266) {
   $clip$4$lcssa = $clip$4;
   break;
  } else {
   $b0$1$pn51$phi = $b0$252;$b0$252 = $b0$2;$clip$350 = $clip$4;$j$149 = $264;$samples$253 = $samples$2;$window$148 = $265;$b0$1$pn51 = $b0$1$pn51$phi;
  }
 }
 $267 = ($final|0)==(0);
 if ($267) {
  return ($clip$4$lcssa|0);
 }
 $268 = HEAP32[$2>>2]|0;
 $269 = (($268) + 256)|0;
 HEAP32[$2>>2] = $269;
 return ($clip$4$lcssa|0);
}
function _INT123_synth_2to1($bandPtr,$channel,$fr,$final) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $fr = $fr|0;
 $final = $final|0;
 var $$neg = 0.0, $$neg1 = 0.0, $$neg15 = 0.0, $$neg16 = 0.0, $$neg17 = 0.0, $$neg18 = 0.0, $$neg19 = 0.0, $$neg2 = 0.0, $$neg20 = 0.0, $$neg21 = 0.0, $$neg22 = 0.0, $$neg23 = 0.0, $$neg24 = 0.0, $$neg25 = 0.0, $$neg26 = 0.0, $$neg27 = 0.0, $$neg28 = 0.0, $$neg29 = 0.0, $$neg3 = 0.0, $$neg4 = 0.0;
 var $$neg5 = 0.0, $$neg6 = 0.0, $$neg7 = 0.0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0;
 var $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0.0, $131 = 0.0, $132 = 0.0, $133 = 0, $134 = 0.0, $135 = 0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0, $146 = 0.0, $147 = 0;
 var $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0, $158 = 0.0, $159 = 0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0, $164 = 0.0, $165 = 0;
 var $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0, $17 = 0, $170 = 0.0, $171 = 0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0.0;
 var $184 = 0, $185 = 0.0, $186 = 0.0, $187 = 0, $188 = 0.0, $189 = 0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0, $200 = 0.0;
 var $201 = 0, $202 = 0.0, $203 = 0, $204 = 0.0, $205 = 0, $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0.0, $213 = 0, $214 = 0.0, $215 = 0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0;
 var $22 = 0, $220 = 0.0, $221 = 0, $222 = 0.0, $223 = 0, $224 = 0.0, $225 = 0, $226 = 0.0, $227 = 0, $228 = 0.0, $229 = 0, $23 = 0, $230 = 0.0, $231 = 0, $232 = 0.0, $233 = 0, $234 = 0.0, $235 = 0, $236 = 0.0, $237 = 0;
 var $238 = 0.0, $239 = 0, $24 = 0, $240 = 0.0, $241 = 0, $242 = 0.0, $243 = 0, $244 = 0.0, $245 = 0.0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0, $254 = 0, $255 = 0;
 var $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0, $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0;
 var $55 = 0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0.0;
 var $73 = 0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0, $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0, $90 = 0.0;
 var $91 = 0, $92 = 0.0, $93 = 0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0, $99 = 0.0, $b0$0 = 0, $b0$1$pn48 = 0, $b0$1$pn48$phi = 0, $b0$155 = 0, $b0$2 = 0, $b0$252 = 0, $bo1$0 = 0, $buf$0 = 0, $clip$056 = 0, $clip$1 = 0, $clip$1$lcssa = 0;
 var $clip$2 = 0, $clip$349 = 0, $clip$4 = 0, $clip$4$lcssa = 0, $j$057 = 0, $j$150 = 0, $samples$0$in = 0, $samples$154 = 0, $samples$2 = 0, $samples$247 = 0, $samples$253 = 0, $scevgep = 0, $scevgep62 = 0, $scevgep63 = 0, $scevgep66 = 0, $scevgep67 = 0, $tmp = 0.0, $tmp11 = 0.0, $tmp13 = 0.0, $tmp30 = 0.0;
 var $tmp32 = 0.0, $tmp34 = 0.0, $tmp36 = 0.0, $tmp38 = 0.0, $tmp40 = 0.0, $tmp42 = 0.0, $tmp44 = 0.0, $tmp45 = 0.0, $tmp9 = 0.0, $window$058 = 0, $window$151 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($fr)) + 9288|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = ((($fr)) + 9292|0);
  _INT123_do_equalizer($bandPtr,$channel,$8);
 }
 $9 = ($channel|0)==(0);
 if ($9) {
  $10 = ((($fr)) + 9272|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + 15)|0;
  $13 = $12 & 15;
  HEAP32[$10>>2] = $13;
  $14 = ((($fr)) + 9248|0);
  $18 = $13;$buf$0 = $14;$samples$0$in = $4;
 } else {
  $15 = ((($4)) + 2|0);
  $16 = ((($fr)) + 9256|0);
  $$phi$trans$insert = ((($fr)) + 9272|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $18 = $$pre;$buf$0 = $16;$samples$0$in = $15;
 }
 $17 = $18 & 1;
 $19 = ($17|0)==(0);
 $20 = ((($buf$0)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 if ($19) {
  $27 = (($18) + 1)|0;
  $28 = HEAP32[$buf$0>>2]|0;
  $29 = (($28) + ($18<<2)|0);
  $30 = (($21) + ($18<<2)|0);
  $31 = ((($30)) + 4|0);
  _INT123_dct64($29,$31,$bandPtr);
  $b0$0 = $21;$bo1$0 = $27;
 } else {
  $22 = HEAP32[$buf$0>>2]|0;
  $23 = (($18) + 1)|0;
  $24 = $23 & 15;
  $25 = (($21) + ($24<<2)|0);
  $26 = (($22) + ($18<<2)|0);
  _INT123_dct64($25,$26,$bandPtr);
  $b0$0 = $22;$bo1$0 = $18;
 }
 $32 = ((($fr)) + 9284|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($33)) + 64|0);
 $35 = (0 - ($bo1$0))|0;
 $36 = (($34) + ($35<<2)|0);
 $scevgep = ((($samples$0$in)) + 32|0);
 $37 = (528 - ($bo1$0))|0;
 $scevgep62 = (($33) + ($37<<2)|0);
 $38 = (464 - ($bo1$0))|0;
 $b0$155 = $b0$0;$clip$056 = 0;$j$057 = 8;$samples$154 = $samples$0$in;$window$058 = $36;
 while(1) {
  $39 = ((($window$058)) + 4|0);
  $40 = +HEAPF32[$window$058>>2];
  $41 = ((($b0$155)) + 4|0);
  $42 = +HEAPF32[$b0$155>>2];
  $43 = $42 * $40;
  $44 = ((($window$058)) + 8|0);
  $45 = +HEAPF32[$39>>2];
  $46 = ((($b0$155)) + 8|0);
  $47 = +HEAPF32[$41>>2];
  $$neg = $47 * $45;
  $48 = ((($window$058)) + 12|0);
  $49 = +HEAPF32[$44>>2];
  $50 = ((($b0$155)) + 12|0);
  $51 = +HEAPF32[$46>>2];
  $52 = $51 * $49;
  $53 = ((($window$058)) + 16|0);
  $54 = +HEAPF32[$48>>2];
  $55 = ((($b0$155)) + 16|0);
  $56 = +HEAPF32[$50>>2];
  $$neg1 = $56 * $54;
  $57 = ((($window$058)) + 20|0);
  $58 = +HEAPF32[$53>>2];
  $59 = ((($b0$155)) + 20|0);
  $60 = +HEAPF32[$55>>2];
  $61 = $60 * $58;
  $62 = ((($window$058)) + 24|0);
  $63 = +HEAPF32[$57>>2];
  $64 = ((($b0$155)) + 24|0);
  $65 = +HEAPF32[$59>>2];
  $$neg2 = $65 * $63;
  $66 = ((($window$058)) + 28|0);
  $67 = +HEAPF32[$62>>2];
  $68 = ((($b0$155)) + 28|0);
  $69 = +HEAPF32[$64>>2];
  $70 = $69 * $67;
  $71 = ((($window$058)) + 32|0);
  $72 = +HEAPF32[$66>>2];
  $73 = ((($b0$155)) + 32|0);
  $74 = +HEAPF32[$68>>2];
  $$neg3 = $74 * $72;
  $75 = ((($window$058)) + 36|0);
  $76 = +HEAPF32[$71>>2];
  $77 = ((($b0$155)) + 36|0);
  $78 = +HEAPF32[$73>>2];
  $79 = $78 * $76;
  $80 = ((($window$058)) + 40|0);
  $81 = +HEAPF32[$75>>2];
  $82 = ((($b0$155)) + 40|0);
  $83 = +HEAPF32[$77>>2];
  $$neg4 = $83 * $81;
  $84 = ((($window$058)) + 44|0);
  $85 = +HEAPF32[$80>>2];
  $86 = ((($b0$155)) + 44|0);
  $87 = +HEAPF32[$82>>2];
  $88 = $87 * $85;
  $89 = ((($window$058)) + 48|0);
  $90 = +HEAPF32[$84>>2];
  $91 = ((($b0$155)) + 48|0);
  $92 = +HEAPF32[$86>>2];
  $$neg5 = $92 * $90;
  $93 = ((($window$058)) + 52|0);
  $94 = +HEAPF32[$89>>2];
  $95 = ((($b0$155)) + 52|0);
  $96 = +HEAPF32[$91>>2];
  $97 = $96 * $94;
  $98 = ((($window$058)) + 56|0);
  $99 = +HEAPF32[$93>>2];
  $100 = ((($b0$155)) + 56|0);
  $101 = +HEAPF32[$95>>2];
  $$neg6 = $101 * $99;
  $102 = ((($window$058)) + 60|0);
  $103 = +HEAPF32[$98>>2];
  $104 = ((($b0$155)) + 60|0);
  $105 = +HEAPF32[$100>>2];
  $106 = $105 * $103;
  $107 = +HEAPF32[$102>>2];
  $108 = +HEAPF32[$104>>2];
  $$neg7 = $108 * $107;
  $tmp = $$neg7 + $$neg6;
  $tmp9 = $$neg5 + $$neg4;
  $tmp11 = $$neg3 + $$neg2;
  $tmp13 = $$neg1 + $$neg;
  $109 = $52 + $43;
  $110 = $109 + $61;
  $111 = $110 + $70;
  $112 = $111 + $79;
  $113 = $112 + $88;
  $114 = $113 + $97;
  $115 = $114 + $106;
  $116 = $115 - $tmp13;
  $117 = $116 - $tmp11;
  $118 = $117 - $tmp9;
  $119 = $118 - $tmp;
  $120 = $119 > 32767.0;
  do {
   if ($120) {
    HEAP16[$samples$154>>1] = 32767;
    $121 = (($clip$056) + 1)|0;
    $clip$1 = $121;
   } else {
    $122 = $119 < -32768.0;
    if ($122) {
     HEAP16[$samples$154>>1] = -32768;
     $123 = (($clip$056) + 1)|0;
     $clip$1 = $123;
     break;
    } else {
     $124 = (~~(($119)));
     HEAP16[$samples$154>>1] = $124;
     $clip$1 = $clip$056;
     break;
    }
   }
  } while(0);
  $125 = (($j$057) + -1)|0;
  $126 = ((($b0$155)) + 128|0);
  $127 = ((($window$058)) + 256|0);
  $128 = ((($samples$154)) + 4|0);
  $129 = ($125|0)==(0);
  if ($129) {
   $clip$1$lcssa = $clip$1;
   break;
  } else {
   $b0$155 = $126;$clip$056 = $clip$1;$j$057 = $125;$samples$154 = $128;$window$058 = $127;
  }
 }
 $scevgep63 = ((($b0$0)) + 1024|0);
 $scevgep66 = ((($b0$0)) + 896|0);
 $scevgep67 = (($33) + ($38<<2)|0);
 $130 = +HEAPF32[$scevgep62>>2];
 $131 = +HEAPF32[$scevgep63>>2];
 $132 = $131 * $130;
 $133 = ((($scevgep67)) + 264|0);
 $134 = +HEAPF32[$133>>2];
 $135 = ((($b0$0)) + 1032|0);
 $136 = +HEAPF32[$135>>2];
 $137 = $136 * $134;
 $138 = $137 + $132;
 $139 = ((($scevgep67)) + 272|0);
 $140 = +HEAPF32[$139>>2];
 $141 = ((($b0$0)) + 1040|0);
 $142 = +HEAPF32[$141>>2];
 $143 = $142 * $140;
 $144 = $138 + $143;
 $145 = ((($scevgep67)) + 280|0);
 $146 = +HEAPF32[$145>>2];
 $147 = ((($b0$0)) + 1048|0);
 $148 = +HEAPF32[$147>>2];
 $149 = $148 * $146;
 $150 = $144 + $149;
 $151 = ((($scevgep67)) + 288|0);
 $152 = +HEAPF32[$151>>2];
 $153 = ((($b0$0)) + 1056|0);
 $154 = +HEAPF32[$153>>2];
 $155 = $154 * $152;
 $156 = $150 + $155;
 $157 = ((($scevgep67)) + 296|0);
 $158 = +HEAPF32[$157>>2];
 $159 = ((($b0$0)) + 1064|0);
 $160 = +HEAPF32[$159>>2];
 $161 = $160 * $158;
 $162 = $156 + $161;
 $163 = ((($scevgep67)) + 304|0);
 $164 = +HEAPF32[$163>>2];
 $165 = ((($b0$0)) + 1072|0);
 $166 = +HEAPF32[$165>>2];
 $167 = $166 * $164;
 $168 = $162 + $167;
 $169 = ((($scevgep67)) + 312|0);
 $170 = +HEAPF32[$169>>2];
 $171 = ((($b0$0)) + 1080|0);
 $172 = +HEAPF32[$171>>2];
 $173 = $172 * $170;
 $174 = $168 + $173;
 $175 = $174 > 32767.0;
 do {
  if ($175) {
   HEAP16[$scevgep>>1] = 32767;
   $176 = (($clip$1$lcssa) + 1)|0;
   $clip$2 = $176;
  } else {
   $177 = $174 < -32768.0;
   if ($177) {
    HEAP16[$scevgep>>1] = -32768;
    $178 = (($clip$1$lcssa) + 1)|0;
    $clip$2 = $178;
    break;
   } else {
    $179 = (~~(($174)));
    HEAP16[$scevgep>>1] = $179;
    $clip$2 = $clip$1$lcssa;
    break;
   }
  }
 } while(0);
 $180 = $bo1$0 << 1;
 $181 = (($scevgep67) + ($180<<2)|0);
 $samples$247 = ((($samples$0$in)) + 36|0);
 $b0$1$pn48 = $scevgep63;$b0$252 = $scevgep66;$clip$349 = $clip$2;$j$150 = 7;$samples$253 = $samples$247;$window$151 = $181;
 while(1) {
  $182 = ((($window$151)) + -4|0);
  $183 = +HEAPF32[$182>>2];
  $184 = ((($b0$1$pn48)) + -124|0);
  $185 = +HEAPF32[$b0$252>>2];
  $186 = $185 * $183;
  $187 = ((($window$151)) + -8|0);
  $188 = +HEAPF32[$187>>2];
  $189 = ((($b0$1$pn48)) + -120|0);
  $190 = +HEAPF32[$184>>2];
  $$neg15 = $190 * $188;
  $191 = ((($window$151)) + -12|0);
  $192 = +HEAPF32[$191>>2];
  $193 = ((($b0$1$pn48)) + -116|0);
  $194 = +HEAPF32[$189>>2];
  $$neg16 = $194 * $192;
  $195 = ((($window$151)) + -16|0);
  $196 = +HEAPF32[$195>>2];
  $197 = ((($b0$1$pn48)) + -112|0);
  $198 = +HEAPF32[$193>>2];
  $$neg17 = $198 * $196;
  $199 = ((($window$151)) + -20|0);
  $200 = +HEAPF32[$199>>2];
  $201 = ((($b0$1$pn48)) + -108|0);
  $202 = +HEAPF32[$197>>2];
  $$neg18 = $202 * $200;
  $203 = ((($window$151)) + -24|0);
  $204 = +HEAPF32[$203>>2];
  $205 = ((($b0$1$pn48)) + -104|0);
  $206 = +HEAPF32[$201>>2];
  $$neg19 = $206 * $204;
  $207 = ((($window$151)) + -28|0);
  $208 = +HEAPF32[$207>>2];
  $209 = ((($b0$1$pn48)) + -100|0);
  $210 = +HEAPF32[$205>>2];
  $$neg20 = $210 * $208;
  $211 = ((($window$151)) + -32|0);
  $212 = +HEAPF32[$211>>2];
  $213 = ((($b0$1$pn48)) + -96|0);
  $214 = +HEAPF32[$209>>2];
  $$neg21 = $214 * $212;
  $215 = ((($window$151)) + -36|0);
  $216 = +HEAPF32[$215>>2];
  $217 = ((($b0$1$pn48)) + -92|0);
  $218 = +HEAPF32[$213>>2];
  $$neg22 = $218 * $216;
  $219 = ((($window$151)) + -40|0);
  $220 = +HEAPF32[$219>>2];
  $221 = ((($b0$1$pn48)) + -88|0);
  $222 = +HEAPF32[$217>>2];
  $$neg23 = $222 * $220;
  $223 = ((($window$151)) + -44|0);
  $224 = +HEAPF32[$223>>2];
  $225 = ((($b0$1$pn48)) + -84|0);
  $226 = +HEAPF32[$221>>2];
  $$neg24 = $226 * $224;
  $227 = ((($window$151)) + -48|0);
  $228 = +HEAPF32[$227>>2];
  $229 = ((($b0$1$pn48)) + -80|0);
  $230 = +HEAPF32[$225>>2];
  $$neg25 = $230 * $228;
  $231 = ((($window$151)) + -52|0);
  $232 = +HEAPF32[$231>>2];
  $233 = ((($b0$1$pn48)) + -76|0);
  $234 = +HEAPF32[$229>>2];
  $$neg26 = $234 * $232;
  $235 = ((($window$151)) + -56|0);
  $236 = +HEAPF32[$235>>2];
  $237 = ((($b0$1$pn48)) + -72|0);
  $238 = +HEAPF32[$233>>2];
  $$neg27 = $238 * $236;
  $239 = ((($window$151)) + -60|0);
  $240 = +HEAPF32[$239>>2];
  $241 = ((($b0$1$pn48)) + -68|0);
  $242 = +HEAPF32[$237>>2];
  $$neg28 = $242 * $240;
  $243 = ((($window$151)) + -64|0);
  $244 = +HEAPF32[$243>>2];
  $245 = +HEAPF32[$241>>2];
  $$neg29 = $245 * $244;
  $tmp30 = $$neg29 + $186;
  $tmp32 = $$neg28 + $$neg27;
  $tmp34 = $$neg26 + $$neg25;
  $tmp36 = $$neg24 + $$neg23;
  $tmp38 = $$neg22 + $$neg21;
  $tmp40 = $$neg20 + $$neg19;
  $tmp42 = $$neg18 + $$neg17;
  $tmp44 = $$neg16 + $$neg15;
  $tmp45 = -$tmp44;
  $246 = $tmp45 - $tmp42;
  $247 = $246 - $tmp40;
  $248 = $247 - $tmp38;
  $249 = $248 - $tmp36;
  $250 = $249 - $tmp34;
  $251 = $250 - $tmp32;
  $252 = $251 - $tmp30;
  $253 = $252 > 32767.0;
  do {
   if ($253) {
    HEAP16[$samples$253>>1] = 32767;
    $254 = (($clip$349) + 1)|0;
    $clip$4 = $254;
   } else {
    $255 = $252 < -32768.0;
    if ($255) {
     HEAP16[$samples$253>>1] = -32768;
     $256 = (($clip$349) + 1)|0;
     $clip$4 = $256;
     break;
    } else {
     $257 = (~~(($252)));
     HEAP16[$samples$253>>1] = $257;
     $clip$4 = $clip$349;
     break;
    }
   }
  } while(0);
  $258 = (($j$150) + -1)|0;
  $259 = ((($window$151)) + -256|0);
  $b0$2 = ((($b0$252)) + -128|0);
  $samples$2 = ((($samples$253)) + 4|0);
  $260 = ($258|0)==(0);
  if ($260) {
   $clip$4$lcssa = $clip$4;
   break;
  } else {
   $b0$1$pn48$phi = $b0$252;$b0$252 = $b0$2;$clip$349 = $clip$4;$j$150 = $258;$samples$253 = $samples$2;$window$151 = $259;$b0$1$pn48 = $b0$1$pn48$phi;
  }
 }
 $261 = ($final|0)==(0);
 if ($261) {
  return ($clip$4$lcssa|0);
 }
 $262 = HEAP32[$2>>2]|0;
 $263 = (($262) + 64)|0;
 HEAP32[$2>>2] = $263;
 return ($clip$4$lcssa|0);
}
function _INT123_synth_2to1_8bit($bandPtr,$channel,$fr,$final) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $fr = $fr|0;
 $final = $final|0;
 var $$neg = 0.0, $$neg1 = 0.0, $$neg15 = 0.0, $$neg16 = 0.0, $$neg17 = 0.0, $$neg18 = 0.0, $$neg19 = 0.0, $$neg2 = 0.0, $$neg20 = 0.0, $$neg21 = 0.0, $$neg22 = 0.0, $$neg23 = 0.0, $$neg24 = 0.0, $$neg25 = 0.0, $$neg26 = 0.0, $$neg27 = 0.0, $$neg28 = 0.0, $$neg29 = 0.0, $$neg3 = 0.0, $$neg4 = 0.0;
 var $$neg5 = 0.0, $$neg6 = 0.0, $$neg7 = 0.0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0, $102 = 0.0, $103 = 0, $104 = 0.0, $105 = 0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0;
 var $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0, $146 = 0.0, $147 = 0;
 var $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0, $158 = 0.0, $159 = 0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0, $164 = 0.0, $165 = 0;
 var $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0, $17 = 0, $170 = 0.0, $171 = 0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0, $176 = 0.0, $177 = 0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0.0, $198 = 0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0;
 var $201 = 0.0, $202 = 0, $203 = 0.0, $204 = 0, $205 = 0.0, $206 = 0, $207 = 0.0, $208 = 0, $209 = 0.0, $21 = 0, $210 = 0, $211 = 0.0, $212 = 0, $213 = 0.0, $214 = 0, $215 = 0.0, $216 = 0, $217 = 0.0, $218 = 0, $219 = 0.0;
 var $22 = 0, $220 = 0, $221 = 0.0, $222 = 0, $223 = 0.0, $224 = 0, $225 = 0.0, $226 = 0, $227 = 0.0, $228 = 0, $229 = 0.0, $23 = 0, $230 = 0, $231 = 0.0, $232 = 0, $233 = 0.0, $234 = 0, $235 = 0.0, $236 = 0, $237 = 0.0;
 var $238 = 0, $239 = 0.0, $24 = 0, $240 = 0, $241 = 0.0, $242 = 0, $243 = 0.0, $244 = 0, $245 = 0.0, $246 = 0, $247 = 0.0, $248 = 0, $249 = 0.0, $25 = 0, $250 = 0, $251 = 0.0, $252 = 0, $253 = 0.0, $254 = 0, $255 = 0.0;
 var $256 = 0.0, $257 = 0.0, $258 = 0.0, $259 = 0.0, $26 = 0, $260 = 0.0, $261 = 0.0, $262 = 0.0, $263 = 0.0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0;
 var $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0.0, $42 = 0, $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0;
 var $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0.0, $63 = 0, $64 = 0.0, $65 = 0, $66 = 0.0, $67 = 0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0, $73 = 0.0, $74 = 0, $75 = 0.0, $76 = 0;
 var $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0, $86 = 0.0, $87 = 0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0, $91 = 0.0, $92 = 0, $93 = 0.0, $94 = 0;
 var $95 = 0.0, $96 = 0, $97 = 0.0, $98 = 0.0, $99 = 0, $b0$0 = 0, $b0$1$pn48 = 0, $b0$1$pn48$phi = 0, $b0$155 = 0, $b0$2 = 0, $b0$252 = 0, $bo1$0 = 0, $buf$0 = 0, $clip$056 = 0, $clip$1 = 0, $clip$1$lcssa = 0, $clip$2 = 0, $clip$349 = 0, $clip$4 = 0, $clip$4$lcssa = 0;
 var $j$057 = 0, $j$150 = 0, $samples$0 = 0, $samples$154 = 0, $samples$2 = 0, $samples$247 = 0, $samples$253 = 0, $scevgep = 0, $scevgep61 = 0, $scevgep62 = 0, $scevgep64 = 0, $scevgep65 = 0, $tmp = 0.0, $tmp11 = 0.0, $tmp13 = 0.0, $tmp30 = 0.0, $tmp32 = 0.0, $tmp34 = 0.0, $tmp36 = 0.0, $tmp38 = 0.0;
 var $tmp40 = 0.0, $tmp42 = 0.0, $tmp44 = 0.0, $tmp45 = 0.0, $tmp9 = 0.0, $window$058 = 0, $window$151 = 0, $write_8bit_tmp$0 = 0, $write_8bit_tmp2$0 = 0, $write_8bit_tmp4$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($fr)) + 9288|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = ((($fr)) + 9292|0);
  _INT123_do_equalizer($bandPtr,$channel,$8);
 }
 $9 = ($channel|0)==(0);
 if ($9) {
  $10 = ((($fr)) + 9272|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + 15)|0;
  $13 = $12 & 15;
  HEAP32[$10>>2] = $13;
  $14 = ((($fr)) + 9248|0);
  $18 = $13;$buf$0 = $14;$samples$0 = $4;
 } else {
  $15 = ((($4)) + 1|0);
  $16 = ((($fr)) + 9256|0);
  $$phi$trans$insert = ((($fr)) + 9272|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $18 = $$pre;$buf$0 = $16;$samples$0 = $15;
 }
 $17 = $18 & 1;
 $19 = ($17|0)==(0);
 $20 = ((($buf$0)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 if ($19) {
  $27 = (($18) + 1)|0;
  $28 = HEAP32[$buf$0>>2]|0;
  $29 = (($28) + ($18<<2)|0);
  $30 = (($21) + ($18<<2)|0);
  $31 = ((($30)) + 4|0);
  _INT123_dct64($29,$31,$bandPtr);
  $b0$0 = $21;$bo1$0 = $27;
 } else {
  $22 = HEAP32[$buf$0>>2]|0;
  $23 = (($18) + 1)|0;
  $24 = $23 & 15;
  $25 = (($21) + ($24<<2)|0);
  $26 = (($22) + ($18<<2)|0);
  _INT123_dct64($25,$26,$bandPtr);
  $b0$0 = $22;$bo1$0 = $18;
 }
 $32 = ((($fr)) + 9284|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($33)) + 64|0);
 $35 = (0 - ($bo1$0))|0;
 $36 = (($34) + ($35<<2)|0);
 $37 = ((($fr)) + 9592|0);
 $38 = (528 - ($bo1$0))|0;
 $scevgep61 = (($33) + ($38<<2)|0);
 $39 = (464 - ($bo1$0))|0;
 $b0$155 = $b0$0;$clip$056 = 0;$j$057 = 8;$samples$154 = $samples$0;$window$058 = $36;
 while(1) {
  $40 = ((($window$058)) + 4|0);
  $41 = +HEAPF32[$window$058>>2];
  $42 = ((($b0$155)) + 4|0);
  $43 = +HEAPF32[$b0$155>>2];
  $44 = $43 * $41;
  $45 = ((($window$058)) + 8|0);
  $46 = +HEAPF32[$40>>2];
  $47 = ((($b0$155)) + 8|0);
  $48 = +HEAPF32[$42>>2];
  $$neg = $48 * $46;
  $49 = ((($window$058)) + 12|0);
  $50 = +HEAPF32[$45>>2];
  $51 = ((($b0$155)) + 12|0);
  $52 = +HEAPF32[$47>>2];
  $53 = $52 * $50;
  $54 = ((($window$058)) + 16|0);
  $55 = +HEAPF32[$49>>2];
  $56 = ((($b0$155)) + 16|0);
  $57 = +HEAPF32[$51>>2];
  $$neg1 = $57 * $55;
  $58 = ((($window$058)) + 20|0);
  $59 = +HEAPF32[$54>>2];
  $60 = ((($b0$155)) + 20|0);
  $61 = +HEAPF32[$56>>2];
  $62 = $61 * $59;
  $63 = ((($window$058)) + 24|0);
  $64 = +HEAPF32[$58>>2];
  $65 = ((($b0$155)) + 24|0);
  $66 = +HEAPF32[$60>>2];
  $$neg2 = $66 * $64;
  $67 = ((($window$058)) + 28|0);
  $68 = +HEAPF32[$63>>2];
  $69 = ((($b0$155)) + 28|0);
  $70 = +HEAPF32[$65>>2];
  $71 = $70 * $68;
  $72 = ((($window$058)) + 32|0);
  $73 = +HEAPF32[$67>>2];
  $74 = ((($b0$155)) + 32|0);
  $75 = +HEAPF32[$69>>2];
  $$neg3 = $75 * $73;
  $76 = ((($window$058)) + 36|0);
  $77 = +HEAPF32[$72>>2];
  $78 = ((($b0$155)) + 36|0);
  $79 = +HEAPF32[$74>>2];
  $80 = $79 * $77;
  $81 = ((($window$058)) + 40|0);
  $82 = +HEAPF32[$76>>2];
  $83 = ((($b0$155)) + 40|0);
  $84 = +HEAPF32[$78>>2];
  $$neg4 = $84 * $82;
  $85 = ((($window$058)) + 44|0);
  $86 = +HEAPF32[$81>>2];
  $87 = ((($b0$155)) + 44|0);
  $88 = +HEAPF32[$83>>2];
  $89 = $88 * $86;
  $90 = ((($window$058)) + 48|0);
  $91 = +HEAPF32[$85>>2];
  $92 = ((($b0$155)) + 48|0);
  $93 = +HEAPF32[$87>>2];
  $$neg5 = $93 * $91;
  $94 = ((($window$058)) + 52|0);
  $95 = +HEAPF32[$90>>2];
  $96 = ((($b0$155)) + 52|0);
  $97 = +HEAPF32[$92>>2];
  $98 = $97 * $95;
  $99 = ((($window$058)) + 56|0);
  $100 = +HEAPF32[$94>>2];
  $101 = ((($b0$155)) + 56|0);
  $102 = +HEAPF32[$96>>2];
  $$neg6 = $102 * $100;
  $103 = ((($window$058)) + 60|0);
  $104 = +HEAPF32[$99>>2];
  $105 = ((($b0$155)) + 60|0);
  $106 = +HEAPF32[$101>>2];
  $107 = $106 * $104;
  $108 = +HEAPF32[$103>>2];
  $109 = +HEAPF32[$105>>2];
  $$neg7 = $109 * $108;
  $tmp = $$neg7 + $$neg6;
  $tmp9 = $$neg5 + $$neg4;
  $tmp11 = $$neg3 + $$neg2;
  $tmp13 = $$neg1 + $$neg;
  $110 = $53 + $44;
  $111 = $110 + $62;
  $112 = $111 + $71;
  $113 = $112 + $80;
  $114 = $113 + $89;
  $115 = $114 + $98;
  $116 = $115 + $107;
  $117 = $116 - $tmp13;
  $118 = $117 - $tmp11;
  $119 = $118 - $tmp9;
  $120 = $119 - $tmp;
  $121 = $120 > 32767.0;
  do {
   if ($121) {
    $122 = (($clip$056) + 1)|0;
    $clip$1 = $122;$write_8bit_tmp$0 = 32767;
   } else {
    $123 = $120 < -32768.0;
    if ($123) {
     $124 = (($clip$056) + 1)|0;
     $clip$1 = $124;$write_8bit_tmp$0 = -32768;
     break;
    } else {
     $125 = (~~(($120)));
     $clip$1 = $clip$056;$write_8bit_tmp$0 = $125;
     break;
    }
   }
  } while(0);
  $126 = $write_8bit_tmp$0 << 16 >> 16;
  $127 = $126 >> 3;
  $128 = HEAP32[$37>>2]|0;
  $129 = (($128) + ($127)|0);
  $130 = HEAP8[$129>>0]|0;
  HEAP8[$samples$154>>0] = $130;
  $131 = (($j$057) + -1)|0;
  $132 = ((($b0$155)) + 128|0);
  $133 = ((($window$058)) + 256|0);
  $134 = ((($samples$154)) + 2|0);
  $135 = ($131|0)==(0);
  if ($135) {
   $clip$1$lcssa = $clip$1;
   break;
  } else {
   $b0$155 = $132;$clip$056 = $clip$1;$j$057 = $131;$samples$154 = $134;$window$058 = $133;
  }
 }
 $scevgep = ((($samples$0)) + 16|0);
 $scevgep62 = ((($b0$0)) + 1024|0);
 $scevgep64 = ((($b0$0)) + 896|0);
 $scevgep65 = (($33) + ($39<<2)|0);
 $136 = +HEAPF32[$scevgep61>>2];
 $137 = +HEAPF32[$scevgep62>>2];
 $138 = $137 * $136;
 $139 = ((($scevgep65)) + 264|0);
 $140 = +HEAPF32[$139>>2];
 $141 = ((($b0$0)) + 1032|0);
 $142 = +HEAPF32[$141>>2];
 $143 = $142 * $140;
 $144 = $143 + $138;
 $145 = ((($scevgep65)) + 272|0);
 $146 = +HEAPF32[$145>>2];
 $147 = ((($b0$0)) + 1040|0);
 $148 = +HEAPF32[$147>>2];
 $149 = $148 * $146;
 $150 = $144 + $149;
 $151 = ((($scevgep65)) + 280|0);
 $152 = +HEAPF32[$151>>2];
 $153 = ((($b0$0)) + 1048|0);
 $154 = +HEAPF32[$153>>2];
 $155 = $154 * $152;
 $156 = $150 + $155;
 $157 = ((($scevgep65)) + 288|0);
 $158 = +HEAPF32[$157>>2];
 $159 = ((($b0$0)) + 1056|0);
 $160 = +HEAPF32[$159>>2];
 $161 = $160 * $158;
 $162 = $156 + $161;
 $163 = ((($scevgep65)) + 296|0);
 $164 = +HEAPF32[$163>>2];
 $165 = ((($b0$0)) + 1064|0);
 $166 = +HEAPF32[$165>>2];
 $167 = $166 * $164;
 $168 = $162 + $167;
 $169 = ((($scevgep65)) + 304|0);
 $170 = +HEAPF32[$169>>2];
 $171 = ((($b0$0)) + 1072|0);
 $172 = +HEAPF32[$171>>2];
 $173 = $172 * $170;
 $174 = $168 + $173;
 $175 = ((($scevgep65)) + 312|0);
 $176 = +HEAPF32[$175>>2];
 $177 = ((($b0$0)) + 1080|0);
 $178 = +HEAPF32[$177>>2];
 $179 = $178 * $176;
 $180 = $174 + $179;
 $181 = $180 > 32767.0;
 do {
  if ($181) {
   $182 = (($clip$1$lcssa) + 1)|0;
   $clip$2 = $182;$write_8bit_tmp2$0 = 32767;
  } else {
   $183 = $180 < -32768.0;
   if ($183) {
    $184 = (($clip$1$lcssa) + 1)|0;
    $clip$2 = $184;$write_8bit_tmp2$0 = -32768;
    break;
   } else {
    $185 = (~~(($180)));
    $clip$2 = $clip$1$lcssa;$write_8bit_tmp2$0 = $185;
    break;
   }
  }
 } while(0);
 $186 = $write_8bit_tmp2$0 << 16 >> 16;
 $187 = $186 >> 3;
 $188 = HEAP32[$37>>2]|0;
 $189 = (($188) + ($187)|0);
 $190 = HEAP8[$189>>0]|0;
 HEAP8[$scevgep>>0] = $190;
 $191 = $bo1$0 << 1;
 $192 = (($scevgep65) + ($191<<2)|0);
 $samples$247 = ((($samples$0)) + 18|0);
 $b0$1$pn48 = $scevgep62;$b0$252 = $scevgep64;$clip$349 = $clip$2;$j$150 = 7;$samples$253 = $samples$247;$window$151 = $192;
 while(1) {
  $193 = ((($window$151)) + -4|0);
  $194 = +HEAPF32[$193>>2];
  $195 = ((($b0$1$pn48)) + -124|0);
  $196 = +HEAPF32[$b0$252>>2];
  $197 = $196 * $194;
  $198 = ((($window$151)) + -8|0);
  $199 = +HEAPF32[$198>>2];
  $200 = ((($b0$1$pn48)) + -120|0);
  $201 = +HEAPF32[$195>>2];
  $$neg15 = $201 * $199;
  $202 = ((($window$151)) + -12|0);
  $203 = +HEAPF32[$202>>2];
  $204 = ((($b0$1$pn48)) + -116|0);
  $205 = +HEAPF32[$200>>2];
  $$neg16 = $205 * $203;
  $206 = ((($window$151)) + -16|0);
  $207 = +HEAPF32[$206>>2];
  $208 = ((($b0$1$pn48)) + -112|0);
  $209 = +HEAPF32[$204>>2];
  $$neg17 = $209 * $207;
  $210 = ((($window$151)) + -20|0);
  $211 = +HEAPF32[$210>>2];
  $212 = ((($b0$1$pn48)) + -108|0);
  $213 = +HEAPF32[$208>>2];
  $$neg18 = $213 * $211;
  $214 = ((($window$151)) + -24|0);
  $215 = +HEAPF32[$214>>2];
  $216 = ((($b0$1$pn48)) + -104|0);
  $217 = +HEAPF32[$212>>2];
  $$neg19 = $217 * $215;
  $218 = ((($window$151)) + -28|0);
  $219 = +HEAPF32[$218>>2];
  $220 = ((($b0$1$pn48)) + -100|0);
  $221 = +HEAPF32[$216>>2];
  $$neg20 = $221 * $219;
  $222 = ((($window$151)) + -32|0);
  $223 = +HEAPF32[$222>>2];
  $224 = ((($b0$1$pn48)) + -96|0);
  $225 = +HEAPF32[$220>>2];
  $$neg21 = $225 * $223;
  $226 = ((($window$151)) + -36|0);
  $227 = +HEAPF32[$226>>2];
  $228 = ((($b0$1$pn48)) + -92|0);
  $229 = +HEAPF32[$224>>2];
  $$neg22 = $229 * $227;
  $230 = ((($window$151)) + -40|0);
  $231 = +HEAPF32[$230>>2];
  $232 = ((($b0$1$pn48)) + -88|0);
  $233 = +HEAPF32[$228>>2];
  $$neg23 = $233 * $231;
  $234 = ((($window$151)) + -44|0);
  $235 = +HEAPF32[$234>>2];
  $236 = ((($b0$1$pn48)) + -84|0);
  $237 = +HEAPF32[$232>>2];
  $$neg24 = $237 * $235;
  $238 = ((($window$151)) + -48|0);
  $239 = +HEAPF32[$238>>2];
  $240 = ((($b0$1$pn48)) + -80|0);
  $241 = +HEAPF32[$236>>2];
  $$neg25 = $241 * $239;
  $242 = ((($window$151)) + -52|0);
  $243 = +HEAPF32[$242>>2];
  $244 = ((($b0$1$pn48)) + -76|0);
  $245 = +HEAPF32[$240>>2];
  $$neg26 = $245 * $243;
  $246 = ((($window$151)) + -56|0);
  $247 = +HEAPF32[$246>>2];
  $248 = ((($b0$1$pn48)) + -72|0);
  $249 = +HEAPF32[$244>>2];
  $$neg27 = $249 * $247;
  $250 = ((($window$151)) + -60|0);
  $251 = +HEAPF32[$250>>2];
  $252 = ((($b0$1$pn48)) + -68|0);
  $253 = +HEAPF32[$248>>2];
  $$neg28 = $253 * $251;
  $254 = ((($window$151)) + -64|0);
  $255 = +HEAPF32[$254>>2];
  $256 = +HEAPF32[$252>>2];
  $$neg29 = $256 * $255;
  $tmp30 = $$neg29 + $197;
  $tmp32 = $$neg28 + $$neg27;
  $tmp34 = $$neg26 + $$neg25;
  $tmp36 = $$neg24 + $$neg23;
  $tmp38 = $$neg22 + $$neg21;
  $tmp40 = $$neg20 + $$neg19;
  $tmp42 = $$neg18 + $$neg17;
  $tmp44 = $$neg16 + $$neg15;
  $tmp45 = -$tmp44;
  $257 = $tmp45 - $tmp42;
  $258 = $257 - $tmp40;
  $259 = $258 - $tmp38;
  $260 = $259 - $tmp36;
  $261 = $260 - $tmp34;
  $262 = $261 - $tmp32;
  $263 = $262 - $tmp30;
  $264 = $263 > 32767.0;
  do {
   if ($264) {
    $265 = (($clip$349) + 1)|0;
    $clip$4 = $265;$write_8bit_tmp4$0 = 32767;
   } else {
    $266 = $263 < -32768.0;
    if ($266) {
     $267 = (($clip$349) + 1)|0;
     $clip$4 = $267;$write_8bit_tmp4$0 = -32768;
     break;
    } else {
     $268 = (~~(($263)));
     $clip$4 = $clip$349;$write_8bit_tmp4$0 = $268;
     break;
    }
   }
  } while(0);
  $269 = $write_8bit_tmp4$0 << 16 >> 16;
  $270 = $269 >> 3;
  $271 = HEAP32[$37>>2]|0;
  $272 = (($271) + ($270)|0);
  $273 = HEAP8[$272>>0]|0;
  HEAP8[$samples$253>>0] = $273;
  $274 = (($j$150) + -1)|0;
  $275 = ((($window$151)) + -256|0);
  $b0$2 = ((($b0$252)) + -128|0);
  $samples$2 = ((($samples$253)) + 2|0);
  $276 = ($274|0)==(0);
  if ($276) {
   $clip$4$lcssa = $clip$4;
   break;
  } else {
   $b0$1$pn48$phi = $b0$252;$b0$252 = $b0$2;$clip$349 = $clip$4;$j$150 = $274;$samples$253 = $samples$2;$window$151 = $275;$b0$1$pn48 = $b0$1$pn48$phi;
  }
 }
 $277 = ($final|0)==(0);
 if ($277) {
  return ($clip$4$lcssa|0);
 }
 $278 = HEAP32[$2>>2]|0;
 $279 = (($278) + 32)|0;
 HEAP32[$2>>2] = $279;
 return ($clip$4$lcssa|0);
}
function _INT123_synth_2to1_real($bandPtr,$channel,$fr,$final) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $fr = $fr|0;
 $final = $final|0;
 var $$neg = 0.0, $$neg1 = 0.0, $$neg15 = 0.0, $$neg16 = 0.0, $$neg17 = 0.0, $$neg18 = 0.0, $$neg19 = 0.0, $$neg2 = 0.0, $$neg20 = 0.0, $$neg21 = 0.0, $$neg22 = 0.0, $$neg23 = 0.0, $$neg24 = 0.0, $$neg25 = 0.0, $$neg26 = 0.0, $$neg27 = 0.0, $$neg28 = 0.0, $$neg29 = 0.0, $$neg3 = 0.0, $$neg4 = 0.0;
 var $$neg5 = 0.0, $$neg6 = 0.0, $$neg7 = 0.0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0;
 var $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0;
 var $13 = 0, $130 = 0.0, $131 = 0, $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0, $136 = 0.0, $137 = 0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0;
 var $148 = 0.0, $149 = 0, $15 = 0, $150 = 0.0, $151 = 0.0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0, $16 = 0, $160 = 0.0, $161 = 0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0;
 var $166 = 0.0, $167 = 0, $168 = 0.0, $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0.0, $172 = 0, $173 = 0, $174 = 0, $175 = 0.0, $176 = 0, $177 = 0.0, $178 = 0.0, $179 = 0, $18 = 0, $180 = 0.0, $181 = 0, $182 = 0.0, $183 = 0;
 var $184 = 0.0, $185 = 0, $186 = 0.0, $187 = 0, $188 = 0.0, $189 = 0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0, $200 = 0.0;
 var $201 = 0, $202 = 0.0, $203 = 0, $204 = 0.0, $205 = 0, $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0.0, $213 = 0, $214 = 0.0, $215 = 0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0;
 var $22 = 0, $220 = 0.0, $221 = 0, $222 = 0.0, $223 = 0, $224 = 0.0, $225 = 0, $226 = 0.0, $227 = 0, $228 = 0.0, $229 = 0, $23 = 0, $230 = 0.0, $231 = 0, $232 = 0.0, $233 = 0, $234 = 0.0, $235 = 0, $236 = 0.0, $237 = 0.0;
 var $238 = 0.0, $239 = 0.0, $24 = 0, $240 = 0.0, $241 = 0.0, $242 = 0.0, $243 = 0.0, $244 = 0.0, $245 = 0.0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0, $45 = 0.0, $46 = 0, $47 = 0.0;
 var $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0.0;
 var $66 = 0, $67 = 0.0, $68 = 0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0, $83 = 0.0;
 var $84 = 0, $85 = 0.0, $86 = 0, $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0, $90 = 0.0, $91 = 0, $92 = 0.0, $93 = 0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0, $99 = 0.0, $b0$0 = 0, $b0$1$pn50 = 0, $b0$1$pn50$phi = 0;
 var $b0$155 = 0, $b0$2 = 0, $b0$251 = 0, $bo1$0 = 0, $buf$0 = 0, $j$053 = 0, $j$148 = 0, $samples$0$in = 0, $samples$156 = 0, $samples$2 = 0, $samples$247 = 0, $samples$252 = 0, $scevgep = 0, $scevgep60 = 0, $scevgep61 = 0, $scevgep62 = 0, $scevgep63 = 0, $tmp = 0.0, $tmp11 = 0.0, $tmp13 = 0.0;
 var $tmp30 = 0.0, $tmp32 = 0.0, $tmp34 = 0.0, $tmp36 = 0.0, $tmp38 = 0.0, $tmp40 = 0.0, $tmp42 = 0.0, $tmp44 = 0.0, $tmp45 = 0.0, $tmp9 = 0.0, $window$054 = 0, $window$149 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($fr)) + 9288|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = ((($fr)) + 9292|0);
  _INT123_do_equalizer($bandPtr,$channel,$8);
 }
 $9 = ($channel|0)==(0);
 if ($9) {
  $10 = ((($fr)) + 9272|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + 15)|0;
  $13 = $12 & 15;
  HEAP32[$10>>2] = $13;
  $14 = ((($fr)) + 9248|0);
  $18 = $13;$buf$0 = $14;$samples$0$in = $4;
 } else {
  $15 = ((($4)) + 4|0);
  $16 = ((($fr)) + 9256|0);
  $$phi$trans$insert = ((($fr)) + 9272|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $18 = $$pre;$buf$0 = $16;$samples$0$in = $15;
 }
 $17 = $18 & 1;
 $19 = ($17|0)==(0);
 $20 = ((($buf$0)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 if ($19) {
  $27 = (($18) + 1)|0;
  $28 = HEAP32[$buf$0>>2]|0;
  $29 = (($28) + ($18<<2)|0);
  $30 = (($21) + ($18<<2)|0);
  $31 = ((($30)) + 4|0);
  _INT123_dct64($29,$31,$bandPtr);
  $b0$0 = $21;$bo1$0 = $27;
 } else {
  $22 = HEAP32[$buf$0>>2]|0;
  $23 = (($18) + 1)|0;
  $24 = $23 & 15;
  $25 = (($21) + ($24<<2)|0);
  $26 = (($22) + ($18<<2)|0);
  _INT123_dct64($25,$26,$bandPtr);
  $b0$0 = $22;$bo1$0 = $18;
 }
 $32 = ((($fr)) + 9284|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($33)) + 64|0);
 $35 = (0 - ($bo1$0))|0;
 $36 = (($34) + ($35<<2)|0);
 $scevgep = ((($samples$0$in)) + 64|0);
 $37 = (528 - ($bo1$0))|0;
 $scevgep60 = (($33) + ($37<<2)|0);
 $38 = (464 - ($bo1$0))|0;
 $scevgep62 = (($33) + ($38<<2)|0);
 $b0$155 = $b0$0;$j$053 = 8;$samples$156 = $samples$0$in;$window$054 = $36;
 while(1) {
  $39 = ((($window$054)) + 4|0);
  $40 = +HEAPF32[$window$054>>2];
  $41 = ((($b0$155)) + 4|0);
  $42 = +HEAPF32[$b0$155>>2];
  $43 = $42 * $40;
  $44 = ((($window$054)) + 8|0);
  $45 = +HEAPF32[$39>>2];
  $46 = ((($b0$155)) + 8|0);
  $47 = +HEAPF32[$41>>2];
  $$neg = $47 * $45;
  $48 = ((($window$054)) + 12|0);
  $49 = +HEAPF32[$44>>2];
  $50 = ((($b0$155)) + 12|0);
  $51 = +HEAPF32[$46>>2];
  $52 = $51 * $49;
  $53 = ((($window$054)) + 16|0);
  $54 = +HEAPF32[$48>>2];
  $55 = ((($b0$155)) + 16|0);
  $56 = +HEAPF32[$50>>2];
  $$neg1 = $56 * $54;
  $57 = ((($window$054)) + 20|0);
  $58 = +HEAPF32[$53>>2];
  $59 = ((($b0$155)) + 20|0);
  $60 = +HEAPF32[$55>>2];
  $61 = $60 * $58;
  $62 = ((($window$054)) + 24|0);
  $63 = +HEAPF32[$57>>2];
  $64 = ((($b0$155)) + 24|0);
  $65 = +HEAPF32[$59>>2];
  $$neg2 = $65 * $63;
  $66 = ((($window$054)) + 28|0);
  $67 = +HEAPF32[$62>>2];
  $68 = ((($b0$155)) + 28|0);
  $69 = +HEAPF32[$64>>2];
  $70 = $69 * $67;
  $71 = ((($window$054)) + 32|0);
  $72 = +HEAPF32[$66>>2];
  $73 = ((($b0$155)) + 32|0);
  $74 = +HEAPF32[$68>>2];
  $$neg3 = $74 * $72;
  $75 = ((($window$054)) + 36|0);
  $76 = +HEAPF32[$71>>2];
  $77 = ((($b0$155)) + 36|0);
  $78 = +HEAPF32[$73>>2];
  $79 = $78 * $76;
  $80 = ((($window$054)) + 40|0);
  $81 = +HEAPF32[$75>>2];
  $82 = ((($b0$155)) + 40|0);
  $83 = +HEAPF32[$77>>2];
  $$neg4 = $83 * $81;
  $84 = ((($window$054)) + 44|0);
  $85 = +HEAPF32[$80>>2];
  $86 = ((($b0$155)) + 44|0);
  $87 = +HEAPF32[$82>>2];
  $88 = $87 * $85;
  $89 = ((($window$054)) + 48|0);
  $90 = +HEAPF32[$84>>2];
  $91 = ((($b0$155)) + 48|0);
  $92 = +HEAPF32[$86>>2];
  $$neg5 = $92 * $90;
  $93 = ((($window$054)) + 52|0);
  $94 = +HEAPF32[$89>>2];
  $95 = ((($b0$155)) + 52|0);
  $96 = +HEAPF32[$91>>2];
  $97 = $96 * $94;
  $98 = ((($window$054)) + 56|0);
  $99 = +HEAPF32[$93>>2];
  $100 = ((($b0$155)) + 56|0);
  $101 = +HEAPF32[$95>>2];
  $$neg6 = $101 * $99;
  $102 = ((($window$054)) + 60|0);
  $103 = +HEAPF32[$98>>2];
  $104 = ((($b0$155)) + 60|0);
  $105 = +HEAPF32[$100>>2];
  $106 = $105 * $103;
  $107 = +HEAPF32[$102>>2];
  $108 = +HEAPF32[$104>>2];
  $$neg7 = $108 * $107;
  $tmp = $$neg7 + $$neg6;
  $tmp9 = $$neg5 + $$neg4;
  $tmp11 = $$neg3 + $$neg2;
  $tmp13 = $$neg1 + $$neg;
  $109 = $52 + $43;
  $110 = $109 + $61;
  $111 = $110 + $70;
  $112 = $111 + $79;
  $113 = $112 + $88;
  $114 = $113 + $97;
  $115 = $114 + $106;
  $116 = $115 - $tmp13;
  $117 = $116 - $tmp11;
  $118 = $117 - $tmp9;
  $119 = $118 - $tmp;
  $120 = $119 * 3.0517578125E-5;
  HEAPF32[$samples$156>>2] = $120;
  $121 = (($j$053) + -1)|0;
  $122 = ((($b0$155)) + 128|0);
  $123 = ((($window$054)) + 256|0);
  $124 = ((($samples$156)) + 8|0);
  $125 = ($121|0)==(0);
  if ($125) {
   break;
  } else {
   $b0$155 = $122;$j$053 = $121;$samples$156 = $124;$window$054 = $123;
  }
 }
 $scevgep61 = ((($b0$0)) + 1024|0);
 $scevgep63 = ((($b0$0)) + 896|0);
 $126 = +HEAPF32[$scevgep60>>2];
 $127 = +HEAPF32[$scevgep61>>2];
 $128 = $127 * $126;
 $129 = ((($scevgep62)) + 264|0);
 $130 = +HEAPF32[$129>>2];
 $131 = ((($b0$0)) + 1032|0);
 $132 = +HEAPF32[$131>>2];
 $133 = $132 * $130;
 $134 = $133 + $128;
 $135 = ((($scevgep62)) + 272|0);
 $136 = +HEAPF32[$135>>2];
 $137 = ((($b0$0)) + 1040|0);
 $138 = +HEAPF32[$137>>2];
 $139 = $138 * $136;
 $140 = $134 + $139;
 $141 = ((($scevgep62)) + 280|0);
 $142 = +HEAPF32[$141>>2];
 $143 = ((($b0$0)) + 1048|0);
 $144 = +HEAPF32[$143>>2];
 $145 = $144 * $142;
 $146 = $140 + $145;
 $147 = ((($scevgep62)) + 288|0);
 $148 = +HEAPF32[$147>>2];
 $149 = ((($b0$0)) + 1056|0);
 $150 = +HEAPF32[$149>>2];
 $151 = $150 * $148;
 $152 = $146 + $151;
 $153 = ((($scevgep62)) + 296|0);
 $154 = +HEAPF32[$153>>2];
 $155 = ((($b0$0)) + 1064|0);
 $156 = +HEAPF32[$155>>2];
 $157 = $156 * $154;
 $158 = $152 + $157;
 $159 = ((($scevgep62)) + 304|0);
 $160 = +HEAPF32[$159>>2];
 $161 = ((($b0$0)) + 1072|0);
 $162 = +HEAPF32[$161>>2];
 $163 = $162 * $160;
 $164 = $158 + $163;
 $165 = ((($scevgep62)) + 312|0);
 $166 = +HEAPF32[$165>>2];
 $167 = ((($b0$0)) + 1080|0);
 $168 = +HEAPF32[$167>>2];
 $169 = $168 * $166;
 $170 = $164 + $169;
 $171 = $170 * 3.0517578125E-5;
 HEAPF32[$scevgep>>2] = $171;
 $172 = $bo1$0 << 1;
 $173 = (($scevgep62) + ($172<<2)|0);
 $samples$247 = ((($samples$0$in)) + 72|0);
 $b0$1$pn50 = $scevgep61;$b0$251 = $scevgep63;$j$148 = 7;$samples$252 = $samples$247;$window$149 = $173;
 while(1) {
  $174 = ((($window$149)) + -4|0);
  $175 = +HEAPF32[$174>>2];
  $176 = ((($b0$1$pn50)) + -124|0);
  $177 = +HEAPF32[$b0$251>>2];
  $178 = $177 * $175;
  $179 = ((($window$149)) + -8|0);
  $180 = +HEAPF32[$179>>2];
  $181 = ((($b0$1$pn50)) + -120|0);
  $182 = +HEAPF32[$176>>2];
  $$neg15 = $182 * $180;
  $183 = ((($window$149)) + -12|0);
  $184 = +HEAPF32[$183>>2];
  $185 = ((($b0$1$pn50)) + -116|0);
  $186 = +HEAPF32[$181>>2];
  $$neg16 = $186 * $184;
  $187 = ((($window$149)) + -16|0);
  $188 = +HEAPF32[$187>>2];
  $189 = ((($b0$1$pn50)) + -112|0);
  $190 = +HEAPF32[$185>>2];
  $$neg17 = $190 * $188;
  $191 = ((($window$149)) + -20|0);
  $192 = +HEAPF32[$191>>2];
  $193 = ((($b0$1$pn50)) + -108|0);
  $194 = +HEAPF32[$189>>2];
  $$neg18 = $194 * $192;
  $195 = ((($window$149)) + -24|0);
  $196 = +HEAPF32[$195>>2];
  $197 = ((($b0$1$pn50)) + -104|0);
  $198 = +HEAPF32[$193>>2];
  $$neg19 = $198 * $196;
  $199 = ((($window$149)) + -28|0);
  $200 = +HEAPF32[$199>>2];
  $201 = ((($b0$1$pn50)) + -100|0);
  $202 = +HEAPF32[$197>>2];
  $$neg20 = $202 * $200;
  $203 = ((($window$149)) + -32|0);
  $204 = +HEAPF32[$203>>2];
  $205 = ((($b0$1$pn50)) + -96|0);
  $206 = +HEAPF32[$201>>2];
  $$neg21 = $206 * $204;
  $207 = ((($window$149)) + -36|0);
  $208 = +HEAPF32[$207>>2];
  $209 = ((($b0$1$pn50)) + -92|0);
  $210 = +HEAPF32[$205>>2];
  $$neg22 = $210 * $208;
  $211 = ((($window$149)) + -40|0);
  $212 = +HEAPF32[$211>>2];
  $213 = ((($b0$1$pn50)) + -88|0);
  $214 = +HEAPF32[$209>>2];
  $$neg23 = $214 * $212;
  $215 = ((($window$149)) + -44|0);
  $216 = +HEAPF32[$215>>2];
  $217 = ((($b0$1$pn50)) + -84|0);
  $218 = +HEAPF32[$213>>2];
  $$neg24 = $218 * $216;
  $219 = ((($window$149)) + -48|0);
  $220 = +HEAPF32[$219>>2];
  $221 = ((($b0$1$pn50)) + -80|0);
  $222 = +HEAPF32[$217>>2];
  $$neg25 = $222 * $220;
  $223 = ((($window$149)) + -52|0);
  $224 = +HEAPF32[$223>>2];
  $225 = ((($b0$1$pn50)) + -76|0);
  $226 = +HEAPF32[$221>>2];
  $$neg26 = $226 * $224;
  $227 = ((($window$149)) + -56|0);
  $228 = +HEAPF32[$227>>2];
  $229 = ((($b0$1$pn50)) + -72|0);
  $230 = +HEAPF32[$225>>2];
  $$neg27 = $230 * $228;
  $231 = ((($window$149)) + -60|0);
  $232 = +HEAPF32[$231>>2];
  $233 = ((($b0$1$pn50)) + -68|0);
  $234 = +HEAPF32[$229>>2];
  $$neg28 = $234 * $232;
  $235 = ((($window$149)) + -64|0);
  $236 = +HEAPF32[$235>>2];
  $237 = +HEAPF32[$233>>2];
  $$neg29 = $237 * $236;
  $tmp30 = $$neg29 + $178;
  $tmp32 = $$neg28 + $$neg27;
  $tmp34 = $$neg26 + $$neg25;
  $tmp36 = $$neg24 + $$neg23;
  $tmp38 = $$neg22 + $$neg21;
  $tmp40 = $$neg20 + $$neg19;
  $tmp42 = $$neg18 + $$neg17;
  $tmp44 = $$neg16 + $$neg15;
  $tmp45 = -$tmp44;
  $238 = $tmp45 - $tmp42;
  $239 = $238 - $tmp40;
  $240 = $239 - $tmp38;
  $241 = $240 - $tmp36;
  $242 = $241 - $tmp34;
  $243 = $242 - $tmp32;
  $244 = $243 - $tmp30;
  $245 = $244 * 3.0517578125E-5;
  HEAPF32[$samples$252>>2] = $245;
  $246 = (($j$148) + -1)|0;
  $247 = ((($window$149)) + -256|0);
  $b0$2 = ((($b0$251)) + -128|0);
  $samples$2 = ((($samples$252)) + 8|0);
  $248 = ($246|0)==(0);
  if ($248) {
   break;
  } else {
   $b0$1$pn50$phi = $b0$251;$b0$251 = $b0$2;$j$148 = $246;$samples$252 = $samples$2;$window$149 = $247;$b0$1$pn50 = $b0$1$pn50$phi;
  }
 }
 $249 = ($final|0)==(0);
 if ($249) {
  return 0;
 }
 $250 = HEAP32[$2>>2]|0;
 $251 = (($250) + 128)|0;
 HEAP32[$2>>2] = $251;
 return 0;
}
function _INT123_synth_2to1_s32($bandPtr,$channel,$fr,$final) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $fr = $fr|0;
 $final = $final|0;
 var $$neg = 0.0, $$neg1 = 0.0, $$neg15 = 0.0, $$neg16 = 0.0, $$neg17 = 0.0, $$neg18 = 0.0, $$neg19 = 0.0, $$neg2 = 0.0, $$neg20 = 0.0, $$neg21 = 0.0, $$neg22 = 0.0, $$neg23 = 0.0, $$neg24 = 0.0, $$neg25 = 0.0, $$neg26 = 0.0, $$neg27 = 0.0, $$neg28 = 0.0, $$neg29 = 0.0, $$neg3 = 0.0, $$neg4 = 0.0;
 var $$neg5 = 0.0, $$neg6 = 0.0, $$neg7 = 0.0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0;
 var $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0, $136 = 0.0, $137 = 0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0;
 var $148 = 0.0, $149 = 0, $15 = 0, $150 = 0.0, $151 = 0.0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0, $16 = 0, $160 = 0.0, $161 = 0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0;
 var $166 = 0.0, $167 = 0, $168 = 0.0, $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0, $172 = 0.0, $173 = 0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0.0, $188 = 0, $189 = 0.0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0, $200 = 0.0;
 var $201 = 0, $202 = 0.0, $203 = 0, $204 = 0.0, $205 = 0, $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0.0, $213 = 0, $214 = 0.0, $215 = 0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0;
 var $22 = 0, $220 = 0.0, $221 = 0, $222 = 0.0, $223 = 0, $224 = 0.0, $225 = 0, $226 = 0.0, $227 = 0, $228 = 0.0, $229 = 0, $23 = 0, $230 = 0.0, $231 = 0, $232 = 0.0, $233 = 0, $234 = 0.0, $235 = 0, $236 = 0.0, $237 = 0;
 var $238 = 0.0, $239 = 0, $24 = 0, $240 = 0.0, $241 = 0, $242 = 0.0, $243 = 0, $244 = 0.0, $245 = 0, $246 = 0.0, $247 = 0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0.0, $254 = 0.0, $255 = 0.0;
 var $256 = 0.0, $257 = 0.0, $258 = 0.0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0, $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0, $49 = 0.0;
 var $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0;
 var $68 = 0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0, $85 = 0.0;
 var $86 = 0, $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0, $90 = 0.0, $91 = 0, $92 = 0.0, $93 = 0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0, $99 = 0.0, $b0$0 = 0, $b0$1$pn51 = 0, $b0$1$pn51$phi = 0, $b0$157 = 0, $b0$2 = 0;
 var $b0$252 = 0, $bo1$0 = 0, $buf$0 = 0, $clip$056 = 0, $clip$1 = 0, $clip$1$lcssa = 0, $clip$2 = 0, $clip$350 = 0, $clip$4 = 0, $clip$4$lcssa = 0, $j$055 = 0, $j$149 = 0, $samples$0$in = 0, $samples$158 = 0, $samples$2 = 0, $samples$247 = 0, $samples$253 = 0, $scevgep = 0, $scevgep62 = 0, $scevgep63 = 0;
 var $scevgep64 = 0, $scevgep65 = 0, $tmp = 0.0, $tmp11 = 0.0, $tmp13 = 0.0, $tmp30 = 0.0, $tmp32 = 0.0, $tmp34 = 0.0, $tmp36 = 0.0, $tmp38 = 0.0, $tmp40 = 0.0, $tmp42 = 0.0, $tmp44 = 0.0, $tmp45 = 0.0, $tmp9 = 0.0, $window$054 = 0, $window$148 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($fr)) + 9288|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = ((($fr)) + 9292|0);
  _INT123_do_equalizer($bandPtr,$channel,$8);
 }
 $9 = ($channel|0)==(0);
 if ($9) {
  $10 = ((($fr)) + 9272|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + 15)|0;
  $13 = $12 & 15;
  HEAP32[$10>>2] = $13;
  $14 = ((($fr)) + 9248|0);
  $18 = $13;$buf$0 = $14;$samples$0$in = $4;
 } else {
  $15 = ((($4)) + 4|0);
  $16 = ((($fr)) + 9256|0);
  $$phi$trans$insert = ((($fr)) + 9272|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $18 = $$pre;$buf$0 = $16;$samples$0$in = $15;
 }
 $17 = $18 & 1;
 $19 = ($17|0)==(0);
 $20 = ((($buf$0)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 if ($19) {
  $27 = (($18) + 1)|0;
  $28 = HEAP32[$buf$0>>2]|0;
  $29 = (($28) + ($18<<2)|0);
  $30 = (($21) + ($18<<2)|0);
  $31 = ((($30)) + 4|0);
  _INT123_dct64($29,$31,$bandPtr);
  $b0$0 = $21;$bo1$0 = $27;
 } else {
  $22 = HEAP32[$buf$0>>2]|0;
  $23 = (($18) + 1)|0;
  $24 = $23 & 15;
  $25 = (($21) + ($24<<2)|0);
  $26 = (($22) + ($18<<2)|0);
  _INT123_dct64($25,$26,$bandPtr);
  $b0$0 = $22;$bo1$0 = $18;
 }
 $32 = ((($fr)) + 9284|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($33)) + 64|0);
 $35 = (0 - ($bo1$0))|0;
 $36 = (($34) + ($35<<2)|0);
 $scevgep = ((($samples$0$in)) + 64|0);
 $37 = (528 - ($bo1$0))|0;
 $scevgep62 = (($33) + ($37<<2)|0);
 $38 = (464 - ($bo1$0))|0;
 $scevgep64 = (($33) + ($38<<2)|0);
 $b0$157 = $b0$0;$clip$056 = 0;$j$055 = 8;$samples$158 = $samples$0$in;$window$054 = $36;
 while(1) {
  $39 = ((($window$054)) + 4|0);
  $40 = +HEAPF32[$window$054>>2];
  $41 = ((($b0$157)) + 4|0);
  $42 = +HEAPF32[$b0$157>>2];
  $43 = $42 * $40;
  $44 = ((($window$054)) + 8|0);
  $45 = +HEAPF32[$39>>2];
  $46 = ((($b0$157)) + 8|0);
  $47 = +HEAPF32[$41>>2];
  $$neg = $47 * $45;
  $48 = ((($window$054)) + 12|0);
  $49 = +HEAPF32[$44>>2];
  $50 = ((($b0$157)) + 12|0);
  $51 = +HEAPF32[$46>>2];
  $52 = $51 * $49;
  $53 = ((($window$054)) + 16|0);
  $54 = +HEAPF32[$48>>2];
  $55 = ((($b0$157)) + 16|0);
  $56 = +HEAPF32[$50>>2];
  $$neg1 = $56 * $54;
  $57 = ((($window$054)) + 20|0);
  $58 = +HEAPF32[$53>>2];
  $59 = ((($b0$157)) + 20|0);
  $60 = +HEAPF32[$55>>2];
  $61 = $60 * $58;
  $62 = ((($window$054)) + 24|0);
  $63 = +HEAPF32[$57>>2];
  $64 = ((($b0$157)) + 24|0);
  $65 = +HEAPF32[$59>>2];
  $$neg2 = $65 * $63;
  $66 = ((($window$054)) + 28|0);
  $67 = +HEAPF32[$62>>2];
  $68 = ((($b0$157)) + 28|0);
  $69 = +HEAPF32[$64>>2];
  $70 = $69 * $67;
  $71 = ((($window$054)) + 32|0);
  $72 = +HEAPF32[$66>>2];
  $73 = ((($b0$157)) + 32|0);
  $74 = +HEAPF32[$68>>2];
  $$neg3 = $74 * $72;
  $75 = ((($window$054)) + 36|0);
  $76 = +HEAPF32[$71>>2];
  $77 = ((($b0$157)) + 36|0);
  $78 = +HEAPF32[$73>>2];
  $79 = $78 * $76;
  $80 = ((($window$054)) + 40|0);
  $81 = +HEAPF32[$75>>2];
  $82 = ((($b0$157)) + 40|0);
  $83 = +HEAPF32[$77>>2];
  $$neg4 = $83 * $81;
  $84 = ((($window$054)) + 44|0);
  $85 = +HEAPF32[$80>>2];
  $86 = ((($b0$157)) + 44|0);
  $87 = +HEAPF32[$82>>2];
  $88 = $87 * $85;
  $89 = ((($window$054)) + 48|0);
  $90 = +HEAPF32[$84>>2];
  $91 = ((($b0$157)) + 48|0);
  $92 = +HEAPF32[$86>>2];
  $$neg5 = $92 * $90;
  $93 = ((($window$054)) + 52|0);
  $94 = +HEAPF32[$89>>2];
  $95 = ((($b0$157)) + 52|0);
  $96 = +HEAPF32[$91>>2];
  $97 = $96 * $94;
  $98 = ((($window$054)) + 56|0);
  $99 = +HEAPF32[$93>>2];
  $100 = ((($b0$157)) + 56|0);
  $101 = +HEAPF32[$95>>2];
  $$neg6 = $101 * $99;
  $102 = ((($window$054)) + 60|0);
  $103 = +HEAPF32[$98>>2];
  $104 = ((($b0$157)) + 60|0);
  $105 = +HEAPF32[$100>>2];
  $106 = $105 * $103;
  $107 = +HEAPF32[$102>>2];
  $108 = +HEAPF32[$104>>2];
  $$neg7 = $108 * $107;
  $tmp = $$neg7 + $$neg6;
  $tmp9 = $$neg5 + $$neg4;
  $tmp11 = $$neg3 + $$neg2;
  $tmp13 = $$neg1 + $$neg;
  $109 = $52 + $43;
  $110 = $109 + $61;
  $111 = $110 + $70;
  $112 = $111 + $79;
  $113 = $112 + $88;
  $114 = $113 + $97;
  $115 = $114 + $106;
  $116 = $115 - $tmp13;
  $117 = $116 - $tmp11;
  $118 = $117 - $tmp9;
  $119 = $118 - $tmp;
  $120 = $119 * 65536.0;
  $121 = $120;
  $122 = $121 > 2147483647.0;
  do {
   if ($122) {
    HEAP32[$samples$158>>2] = 2147483647;
    $123 = (($clip$056) + 1)|0;
    $clip$1 = $123;
   } else {
    $124 = $120 < -2147483648.0;
    if ($124) {
     HEAP32[$samples$158>>2] = -2147483648;
     $125 = (($clip$056) + 1)|0;
     $clip$1 = $125;
     break;
    } else {
     $126 = (~~(($120)));
     HEAP32[$samples$158>>2] = $126;
     $clip$1 = $clip$056;
     break;
    }
   }
  } while(0);
  $127 = (($j$055) + -1)|0;
  $128 = ((($b0$157)) + 128|0);
  $129 = ((($window$054)) + 256|0);
  $130 = ((($samples$158)) + 8|0);
  $131 = ($127|0)==(0);
  if ($131) {
   $clip$1$lcssa = $clip$1;
   break;
  } else {
   $b0$157 = $128;$clip$056 = $clip$1;$j$055 = $127;$samples$158 = $130;$window$054 = $129;
  }
 }
 $scevgep63 = ((($b0$0)) + 1024|0);
 $scevgep65 = ((($b0$0)) + 896|0);
 $132 = +HEAPF32[$scevgep62>>2];
 $133 = +HEAPF32[$scevgep63>>2];
 $134 = $133 * $132;
 $135 = ((($scevgep64)) + 264|0);
 $136 = +HEAPF32[$135>>2];
 $137 = ((($b0$0)) + 1032|0);
 $138 = +HEAPF32[$137>>2];
 $139 = $138 * $136;
 $140 = $139 + $134;
 $141 = ((($scevgep64)) + 272|0);
 $142 = +HEAPF32[$141>>2];
 $143 = ((($b0$0)) + 1040|0);
 $144 = +HEAPF32[$143>>2];
 $145 = $144 * $142;
 $146 = $140 + $145;
 $147 = ((($scevgep64)) + 280|0);
 $148 = +HEAPF32[$147>>2];
 $149 = ((($b0$0)) + 1048|0);
 $150 = +HEAPF32[$149>>2];
 $151 = $150 * $148;
 $152 = $146 + $151;
 $153 = ((($scevgep64)) + 288|0);
 $154 = +HEAPF32[$153>>2];
 $155 = ((($b0$0)) + 1056|0);
 $156 = +HEAPF32[$155>>2];
 $157 = $156 * $154;
 $158 = $152 + $157;
 $159 = ((($scevgep64)) + 296|0);
 $160 = +HEAPF32[$159>>2];
 $161 = ((($b0$0)) + 1064|0);
 $162 = +HEAPF32[$161>>2];
 $163 = $162 * $160;
 $164 = $158 + $163;
 $165 = ((($scevgep64)) + 304|0);
 $166 = +HEAPF32[$165>>2];
 $167 = ((($b0$0)) + 1072|0);
 $168 = +HEAPF32[$167>>2];
 $169 = $168 * $166;
 $170 = $164 + $169;
 $171 = ((($scevgep64)) + 312|0);
 $172 = +HEAPF32[$171>>2];
 $173 = ((($b0$0)) + 1080|0);
 $174 = +HEAPF32[$173>>2];
 $175 = $174 * $172;
 $176 = $170 + $175;
 $177 = $176 * 65536.0;
 $178 = $177;
 $179 = $178 > 2147483647.0;
 do {
  if ($179) {
   HEAP32[$scevgep>>2] = 2147483647;
   $180 = (($clip$1$lcssa) + 1)|0;
   $clip$2 = $180;
  } else {
   $181 = $177 < -2147483648.0;
   if ($181) {
    HEAP32[$scevgep>>2] = -2147483648;
    $182 = (($clip$1$lcssa) + 1)|0;
    $clip$2 = $182;
    break;
   } else {
    $183 = (~~(($177)));
    HEAP32[$scevgep>>2] = $183;
    $clip$2 = $clip$1$lcssa;
    break;
   }
  }
 } while(0);
 $184 = $bo1$0 << 1;
 $185 = (($scevgep64) + ($184<<2)|0);
 $samples$247 = ((($samples$0$in)) + 72|0);
 $b0$1$pn51 = $scevgep63;$b0$252 = $scevgep65;$clip$350 = $clip$2;$j$149 = 7;$samples$253 = $samples$247;$window$148 = $185;
 while(1) {
  $186 = ((($window$148)) + -4|0);
  $187 = +HEAPF32[$186>>2];
  $188 = ((($b0$1$pn51)) + -124|0);
  $189 = +HEAPF32[$b0$252>>2];
  $190 = $189 * $187;
  $191 = ((($window$148)) + -8|0);
  $192 = +HEAPF32[$191>>2];
  $193 = ((($b0$1$pn51)) + -120|0);
  $194 = +HEAPF32[$188>>2];
  $$neg15 = $194 * $192;
  $195 = ((($window$148)) + -12|0);
  $196 = +HEAPF32[$195>>2];
  $197 = ((($b0$1$pn51)) + -116|0);
  $198 = +HEAPF32[$193>>2];
  $$neg16 = $198 * $196;
  $199 = ((($window$148)) + -16|0);
  $200 = +HEAPF32[$199>>2];
  $201 = ((($b0$1$pn51)) + -112|0);
  $202 = +HEAPF32[$197>>2];
  $$neg17 = $202 * $200;
  $203 = ((($window$148)) + -20|0);
  $204 = +HEAPF32[$203>>2];
  $205 = ((($b0$1$pn51)) + -108|0);
  $206 = +HEAPF32[$201>>2];
  $$neg18 = $206 * $204;
  $207 = ((($window$148)) + -24|0);
  $208 = +HEAPF32[$207>>2];
  $209 = ((($b0$1$pn51)) + -104|0);
  $210 = +HEAPF32[$205>>2];
  $$neg19 = $210 * $208;
  $211 = ((($window$148)) + -28|0);
  $212 = +HEAPF32[$211>>2];
  $213 = ((($b0$1$pn51)) + -100|0);
  $214 = +HEAPF32[$209>>2];
  $$neg20 = $214 * $212;
  $215 = ((($window$148)) + -32|0);
  $216 = +HEAPF32[$215>>2];
  $217 = ((($b0$1$pn51)) + -96|0);
  $218 = +HEAPF32[$213>>2];
  $$neg21 = $218 * $216;
  $219 = ((($window$148)) + -36|0);
  $220 = +HEAPF32[$219>>2];
  $221 = ((($b0$1$pn51)) + -92|0);
  $222 = +HEAPF32[$217>>2];
  $$neg22 = $222 * $220;
  $223 = ((($window$148)) + -40|0);
  $224 = +HEAPF32[$223>>2];
  $225 = ((($b0$1$pn51)) + -88|0);
  $226 = +HEAPF32[$221>>2];
  $$neg23 = $226 * $224;
  $227 = ((($window$148)) + -44|0);
  $228 = +HEAPF32[$227>>2];
  $229 = ((($b0$1$pn51)) + -84|0);
  $230 = +HEAPF32[$225>>2];
  $$neg24 = $230 * $228;
  $231 = ((($window$148)) + -48|0);
  $232 = +HEAPF32[$231>>2];
  $233 = ((($b0$1$pn51)) + -80|0);
  $234 = +HEAPF32[$229>>2];
  $$neg25 = $234 * $232;
  $235 = ((($window$148)) + -52|0);
  $236 = +HEAPF32[$235>>2];
  $237 = ((($b0$1$pn51)) + -76|0);
  $238 = +HEAPF32[$233>>2];
  $$neg26 = $238 * $236;
  $239 = ((($window$148)) + -56|0);
  $240 = +HEAPF32[$239>>2];
  $241 = ((($b0$1$pn51)) + -72|0);
  $242 = +HEAPF32[$237>>2];
  $$neg27 = $242 * $240;
  $243 = ((($window$148)) + -60|0);
  $244 = +HEAPF32[$243>>2];
  $245 = ((($b0$1$pn51)) + -68|0);
  $246 = +HEAPF32[$241>>2];
  $$neg28 = $246 * $244;
  $247 = ((($window$148)) + -64|0);
  $248 = +HEAPF32[$247>>2];
  $249 = +HEAPF32[$245>>2];
  $$neg29 = $249 * $248;
  $tmp30 = $$neg29 + $190;
  $tmp32 = $$neg28 + $$neg27;
  $tmp34 = $$neg26 + $$neg25;
  $tmp36 = $$neg24 + $$neg23;
  $tmp38 = $$neg22 + $$neg21;
  $tmp40 = $$neg20 + $$neg19;
  $tmp42 = $$neg18 + $$neg17;
  $tmp44 = $$neg16 + $$neg15;
  $tmp45 = -$tmp44;
  $250 = $tmp45 - $tmp42;
  $251 = $250 - $tmp40;
  $252 = $251 - $tmp38;
  $253 = $252 - $tmp36;
  $254 = $253 - $tmp34;
  $255 = $254 - $tmp32;
  $256 = $255 - $tmp30;
  $257 = $256 * 65536.0;
  $258 = $257;
  $259 = $258 > 2147483647.0;
  do {
   if ($259) {
    HEAP32[$samples$253>>2] = 2147483647;
    $260 = (($clip$350) + 1)|0;
    $clip$4 = $260;
   } else {
    $261 = $257 < -2147483648.0;
    if ($261) {
     HEAP32[$samples$253>>2] = -2147483648;
     $262 = (($clip$350) + 1)|0;
     $clip$4 = $262;
     break;
    } else {
     $263 = (~~(($257)));
     HEAP32[$samples$253>>2] = $263;
     $clip$4 = $clip$350;
     break;
    }
   }
  } while(0);
  $264 = (($j$149) + -1)|0;
  $265 = ((($window$148)) + -256|0);
  $b0$2 = ((($b0$252)) + -128|0);
  $samples$2 = ((($samples$253)) + 8|0);
  $266 = ($264|0)==(0);
  if ($266) {
   $clip$4$lcssa = $clip$4;
   break;
  } else {
   $b0$1$pn51$phi = $b0$252;$b0$252 = $b0$2;$clip$350 = $clip$4;$j$149 = $264;$samples$253 = $samples$2;$window$148 = $265;$b0$1$pn51 = $b0$1$pn51$phi;
  }
 }
 $267 = ($final|0)==(0);
 if ($267) {
  return ($clip$4$lcssa|0);
 }
 $268 = HEAP32[$2>>2]|0;
 $269 = (($268) + 128)|0;
 HEAP32[$2>>2] = $269;
 return ($clip$4$lcssa|0);
}
function _INT123_synth_4to1($bandPtr,$channel,$fr,$final) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $fr = $fr|0;
 $final = $final|0;
 var $$neg = 0.0, $$neg1 = 0.0, $$neg15 = 0.0, $$neg16 = 0.0, $$neg17 = 0.0, $$neg18 = 0.0, $$neg19 = 0.0, $$neg2 = 0.0, $$neg20 = 0.0, $$neg21 = 0.0, $$neg22 = 0.0, $$neg23 = 0.0, $$neg24 = 0.0, $$neg25 = 0.0, $$neg26 = 0.0, $$neg27 = 0.0, $$neg28 = 0.0, $$neg29 = 0.0, $$neg3 = 0.0, $$neg4 = 0.0;
 var $$neg5 = 0.0, $$neg6 = 0.0, $$neg7 = 0.0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0;
 var $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0.0, $131 = 0.0, $132 = 0.0, $133 = 0, $134 = 0.0, $135 = 0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0, $146 = 0.0, $147 = 0;
 var $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0, $158 = 0.0, $159 = 0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0, $164 = 0.0, $165 = 0;
 var $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0, $17 = 0, $170 = 0.0, $171 = 0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0.0;
 var $184 = 0, $185 = 0.0, $186 = 0.0, $187 = 0, $188 = 0.0, $189 = 0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0, $200 = 0.0;
 var $201 = 0, $202 = 0.0, $203 = 0, $204 = 0.0, $205 = 0, $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0.0, $213 = 0, $214 = 0.0, $215 = 0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0;
 var $22 = 0, $220 = 0.0, $221 = 0, $222 = 0.0, $223 = 0, $224 = 0.0, $225 = 0, $226 = 0.0, $227 = 0, $228 = 0.0, $229 = 0, $23 = 0, $230 = 0.0, $231 = 0, $232 = 0.0, $233 = 0, $234 = 0.0, $235 = 0, $236 = 0.0, $237 = 0;
 var $238 = 0.0, $239 = 0, $24 = 0, $240 = 0.0, $241 = 0, $242 = 0.0, $243 = 0, $244 = 0.0, $245 = 0.0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0, $254 = 0, $255 = 0;
 var $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0, $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0;
 var $55 = 0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0.0;
 var $73 = 0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0, $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0, $90 = 0.0;
 var $91 = 0, $92 = 0.0, $93 = 0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0, $99 = 0.0, $b0$0 = 0, $b0$1$pn48 = 0, $b0$1$pn48$phi = 0, $b0$155 = 0, $b0$2 = 0, $b0$252 = 0, $bo1$0 = 0, $buf$0 = 0, $clip$056 = 0, $clip$1 = 0, $clip$1$lcssa = 0;
 var $clip$2 = 0, $clip$349 = 0, $clip$4 = 0, $clip$4$lcssa = 0, $j$057 = 0, $j$150 = 0, $samples$0$in = 0, $samples$154 = 0, $samples$2 = 0, $samples$247 = 0, $samples$253 = 0, $scevgep = 0, $scevgep62 = 0, $scevgep63 = 0, $scevgep66 = 0, $scevgep67 = 0, $tmp = 0.0, $tmp11 = 0.0, $tmp13 = 0.0, $tmp30 = 0.0;
 var $tmp32 = 0.0, $tmp34 = 0.0, $tmp36 = 0.0, $tmp38 = 0.0, $tmp40 = 0.0, $tmp42 = 0.0, $tmp44 = 0.0, $tmp45 = 0.0, $tmp9 = 0.0, $window$058 = 0, $window$151 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($fr)) + 9288|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = ((($fr)) + 9292|0);
  _INT123_do_equalizer($bandPtr,$channel,$8);
 }
 $9 = ($channel|0)==(0);
 if ($9) {
  $10 = ((($fr)) + 9272|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + 15)|0;
  $13 = $12 & 15;
  HEAP32[$10>>2] = $13;
  $14 = ((($fr)) + 9248|0);
  $18 = $13;$buf$0 = $14;$samples$0$in = $4;
 } else {
  $15 = ((($4)) + 2|0);
  $16 = ((($fr)) + 9256|0);
  $$phi$trans$insert = ((($fr)) + 9272|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $18 = $$pre;$buf$0 = $16;$samples$0$in = $15;
 }
 $17 = $18 & 1;
 $19 = ($17|0)==(0);
 $20 = ((($buf$0)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 if ($19) {
  $27 = (($18) + 1)|0;
  $28 = HEAP32[$buf$0>>2]|0;
  $29 = (($28) + ($18<<2)|0);
  $30 = (($21) + ($18<<2)|0);
  $31 = ((($30)) + 4|0);
  _INT123_dct64($29,$31,$bandPtr);
  $b0$0 = $21;$bo1$0 = $27;
 } else {
  $22 = HEAP32[$buf$0>>2]|0;
  $23 = (($18) + 1)|0;
  $24 = $23 & 15;
  $25 = (($21) + ($24<<2)|0);
  $26 = (($22) + ($18<<2)|0);
  _INT123_dct64($25,$26,$bandPtr);
  $b0$0 = $22;$bo1$0 = $18;
 }
 $32 = ((($fr)) + 9284|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($33)) + 64|0);
 $35 = (0 - ($bo1$0))|0;
 $36 = (($34) + ($35<<2)|0);
 $scevgep = ((($samples$0$in)) + 16|0);
 $37 = (528 - ($bo1$0))|0;
 $scevgep62 = (($33) + ($37<<2)|0);
 $38 = (400 - ($bo1$0))|0;
 $b0$155 = $b0$0;$clip$056 = 0;$j$057 = 4;$samples$154 = $samples$0$in;$window$058 = $36;
 while(1) {
  $39 = ((($window$058)) + 4|0);
  $40 = +HEAPF32[$window$058>>2];
  $41 = ((($b0$155)) + 4|0);
  $42 = +HEAPF32[$b0$155>>2];
  $43 = $42 * $40;
  $44 = ((($window$058)) + 8|0);
  $45 = +HEAPF32[$39>>2];
  $46 = ((($b0$155)) + 8|0);
  $47 = +HEAPF32[$41>>2];
  $$neg = $47 * $45;
  $48 = ((($window$058)) + 12|0);
  $49 = +HEAPF32[$44>>2];
  $50 = ((($b0$155)) + 12|0);
  $51 = +HEAPF32[$46>>2];
  $52 = $51 * $49;
  $53 = ((($window$058)) + 16|0);
  $54 = +HEAPF32[$48>>2];
  $55 = ((($b0$155)) + 16|0);
  $56 = +HEAPF32[$50>>2];
  $$neg1 = $56 * $54;
  $57 = ((($window$058)) + 20|0);
  $58 = +HEAPF32[$53>>2];
  $59 = ((($b0$155)) + 20|0);
  $60 = +HEAPF32[$55>>2];
  $61 = $60 * $58;
  $62 = ((($window$058)) + 24|0);
  $63 = +HEAPF32[$57>>2];
  $64 = ((($b0$155)) + 24|0);
  $65 = +HEAPF32[$59>>2];
  $$neg2 = $65 * $63;
  $66 = ((($window$058)) + 28|0);
  $67 = +HEAPF32[$62>>2];
  $68 = ((($b0$155)) + 28|0);
  $69 = +HEAPF32[$64>>2];
  $70 = $69 * $67;
  $71 = ((($window$058)) + 32|0);
  $72 = +HEAPF32[$66>>2];
  $73 = ((($b0$155)) + 32|0);
  $74 = +HEAPF32[$68>>2];
  $$neg3 = $74 * $72;
  $75 = ((($window$058)) + 36|0);
  $76 = +HEAPF32[$71>>2];
  $77 = ((($b0$155)) + 36|0);
  $78 = +HEAPF32[$73>>2];
  $79 = $78 * $76;
  $80 = ((($window$058)) + 40|0);
  $81 = +HEAPF32[$75>>2];
  $82 = ((($b0$155)) + 40|0);
  $83 = +HEAPF32[$77>>2];
  $$neg4 = $83 * $81;
  $84 = ((($window$058)) + 44|0);
  $85 = +HEAPF32[$80>>2];
  $86 = ((($b0$155)) + 44|0);
  $87 = +HEAPF32[$82>>2];
  $88 = $87 * $85;
  $89 = ((($window$058)) + 48|0);
  $90 = +HEAPF32[$84>>2];
  $91 = ((($b0$155)) + 48|0);
  $92 = +HEAPF32[$86>>2];
  $$neg5 = $92 * $90;
  $93 = ((($window$058)) + 52|0);
  $94 = +HEAPF32[$89>>2];
  $95 = ((($b0$155)) + 52|0);
  $96 = +HEAPF32[$91>>2];
  $97 = $96 * $94;
  $98 = ((($window$058)) + 56|0);
  $99 = +HEAPF32[$93>>2];
  $100 = ((($b0$155)) + 56|0);
  $101 = +HEAPF32[$95>>2];
  $$neg6 = $101 * $99;
  $102 = ((($window$058)) + 60|0);
  $103 = +HEAPF32[$98>>2];
  $104 = ((($b0$155)) + 60|0);
  $105 = +HEAPF32[$100>>2];
  $106 = $105 * $103;
  $107 = +HEAPF32[$102>>2];
  $108 = +HEAPF32[$104>>2];
  $$neg7 = $108 * $107;
  $tmp = $$neg7 + $$neg6;
  $tmp9 = $$neg5 + $$neg4;
  $tmp11 = $$neg3 + $$neg2;
  $tmp13 = $$neg1 + $$neg;
  $109 = $52 + $43;
  $110 = $109 + $61;
  $111 = $110 + $70;
  $112 = $111 + $79;
  $113 = $112 + $88;
  $114 = $113 + $97;
  $115 = $114 + $106;
  $116 = $115 - $tmp13;
  $117 = $116 - $tmp11;
  $118 = $117 - $tmp9;
  $119 = $118 - $tmp;
  $120 = $119 > 32767.0;
  do {
   if ($120) {
    HEAP16[$samples$154>>1] = 32767;
    $121 = (($clip$056) + 1)|0;
    $clip$1 = $121;
   } else {
    $122 = $119 < -32768.0;
    if ($122) {
     HEAP16[$samples$154>>1] = -32768;
     $123 = (($clip$056) + 1)|0;
     $clip$1 = $123;
     break;
    } else {
     $124 = (~~(($119)));
     HEAP16[$samples$154>>1] = $124;
     $clip$1 = $clip$056;
     break;
    }
   }
  } while(0);
  $125 = (($j$057) + -1)|0;
  $126 = ((($b0$155)) + 256|0);
  $127 = ((($window$058)) + 512|0);
  $128 = ((($samples$154)) + 4|0);
  $129 = ($125|0)==(0);
  if ($129) {
   $clip$1$lcssa = $clip$1;
   break;
  } else {
   $b0$155 = $126;$clip$056 = $clip$1;$j$057 = $125;$samples$154 = $128;$window$058 = $127;
  }
 }
 $scevgep63 = ((($b0$0)) + 1024|0);
 $scevgep66 = ((($b0$0)) + 768|0);
 $scevgep67 = (($33) + ($38<<2)|0);
 $130 = +HEAPF32[$scevgep62>>2];
 $131 = +HEAPF32[$scevgep63>>2];
 $132 = $131 * $130;
 $133 = ((($scevgep67)) + 520|0);
 $134 = +HEAPF32[$133>>2];
 $135 = ((($b0$0)) + 1032|0);
 $136 = +HEAPF32[$135>>2];
 $137 = $136 * $134;
 $138 = $137 + $132;
 $139 = ((($scevgep67)) + 528|0);
 $140 = +HEAPF32[$139>>2];
 $141 = ((($b0$0)) + 1040|0);
 $142 = +HEAPF32[$141>>2];
 $143 = $142 * $140;
 $144 = $138 + $143;
 $145 = ((($scevgep67)) + 536|0);
 $146 = +HEAPF32[$145>>2];
 $147 = ((($b0$0)) + 1048|0);
 $148 = +HEAPF32[$147>>2];
 $149 = $148 * $146;
 $150 = $144 + $149;
 $151 = ((($scevgep67)) + 544|0);
 $152 = +HEAPF32[$151>>2];
 $153 = ((($b0$0)) + 1056|0);
 $154 = +HEAPF32[$153>>2];
 $155 = $154 * $152;
 $156 = $150 + $155;
 $157 = ((($scevgep67)) + 552|0);
 $158 = +HEAPF32[$157>>2];
 $159 = ((($b0$0)) + 1064|0);
 $160 = +HEAPF32[$159>>2];
 $161 = $160 * $158;
 $162 = $156 + $161;
 $163 = ((($scevgep67)) + 560|0);
 $164 = +HEAPF32[$163>>2];
 $165 = ((($b0$0)) + 1072|0);
 $166 = +HEAPF32[$165>>2];
 $167 = $166 * $164;
 $168 = $162 + $167;
 $169 = ((($scevgep67)) + 568|0);
 $170 = +HEAPF32[$169>>2];
 $171 = ((($b0$0)) + 1080|0);
 $172 = +HEAPF32[$171>>2];
 $173 = $172 * $170;
 $174 = $168 + $173;
 $175 = $174 > 32767.0;
 do {
  if ($175) {
   HEAP16[$scevgep>>1] = 32767;
   $176 = (($clip$1$lcssa) + 1)|0;
   $clip$2 = $176;
  } else {
   $177 = $174 < -32768.0;
   if ($177) {
    HEAP16[$scevgep>>1] = -32768;
    $178 = (($clip$1$lcssa) + 1)|0;
    $clip$2 = $178;
    break;
   } else {
    $179 = (~~(($174)));
    HEAP16[$scevgep>>1] = $179;
    $clip$2 = $clip$1$lcssa;
    break;
   }
  }
 } while(0);
 $180 = $bo1$0 << 1;
 $181 = (($scevgep67) + ($180<<2)|0);
 $samples$247 = ((($samples$0$in)) + 20|0);
 $b0$1$pn48 = $scevgep63;$b0$252 = $scevgep66;$clip$349 = $clip$2;$j$150 = 3;$samples$253 = $samples$247;$window$151 = $181;
 while(1) {
  $182 = ((($window$151)) + -4|0);
  $183 = +HEAPF32[$182>>2];
  $184 = ((($b0$1$pn48)) + -252|0);
  $185 = +HEAPF32[$b0$252>>2];
  $186 = $185 * $183;
  $187 = ((($window$151)) + -8|0);
  $188 = +HEAPF32[$187>>2];
  $189 = ((($b0$1$pn48)) + -248|0);
  $190 = +HEAPF32[$184>>2];
  $$neg15 = $190 * $188;
  $191 = ((($window$151)) + -12|0);
  $192 = +HEAPF32[$191>>2];
  $193 = ((($b0$1$pn48)) + -244|0);
  $194 = +HEAPF32[$189>>2];
  $$neg16 = $194 * $192;
  $195 = ((($window$151)) + -16|0);
  $196 = +HEAPF32[$195>>2];
  $197 = ((($b0$1$pn48)) + -240|0);
  $198 = +HEAPF32[$193>>2];
  $$neg17 = $198 * $196;
  $199 = ((($window$151)) + -20|0);
  $200 = +HEAPF32[$199>>2];
  $201 = ((($b0$1$pn48)) + -236|0);
  $202 = +HEAPF32[$197>>2];
  $$neg18 = $202 * $200;
  $203 = ((($window$151)) + -24|0);
  $204 = +HEAPF32[$203>>2];
  $205 = ((($b0$1$pn48)) + -232|0);
  $206 = +HEAPF32[$201>>2];
  $$neg19 = $206 * $204;
  $207 = ((($window$151)) + -28|0);
  $208 = +HEAPF32[$207>>2];
  $209 = ((($b0$1$pn48)) + -228|0);
  $210 = +HEAPF32[$205>>2];
  $$neg20 = $210 * $208;
  $211 = ((($window$151)) + -32|0);
  $212 = +HEAPF32[$211>>2];
  $213 = ((($b0$1$pn48)) + -224|0);
  $214 = +HEAPF32[$209>>2];
  $$neg21 = $214 * $212;
  $215 = ((($window$151)) + -36|0);
  $216 = +HEAPF32[$215>>2];
  $217 = ((($b0$1$pn48)) + -220|0);
  $218 = +HEAPF32[$213>>2];
  $$neg22 = $218 * $216;
  $219 = ((($window$151)) + -40|0);
  $220 = +HEAPF32[$219>>2];
  $221 = ((($b0$1$pn48)) + -216|0);
  $222 = +HEAPF32[$217>>2];
  $$neg23 = $222 * $220;
  $223 = ((($window$151)) + -44|0);
  $224 = +HEAPF32[$223>>2];
  $225 = ((($b0$1$pn48)) + -212|0);
  $226 = +HEAPF32[$221>>2];
  $$neg24 = $226 * $224;
  $227 = ((($window$151)) + -48|0);
  $228 = +HEAPF32[$227>>2];
  $229 = ((($b0$1$pn48)) + -208|0);
  $230 = +HEAPF32[$225>>2];
  $$neg25 = $230 * $228;
  $231 = ((($window$151)) + -52|0);
  $232 = +HEAPF32[$231>>2];
  $233 = ((($b0$1$pn48)) + -204|0);
  $234 = +HEAPF32[$229>>2];
  $$neg26 = $234 * $232;
  $235 = ((($window$151)) + -56|0);
  $236 = +HEAPF32[$235>>2];
  $237 = ((($b0$1$pn48)) + -200|0);
  $238 = +HEAPF32[$233>>2];
  $$neg27 = $238 * $236;
  $239 = ((($window$151)) + -60|0);
  $240 = +HEAPF32[$239>>2];
  $241 = ((($b0$1$pn48)) + -196|0);
  $242 = +HEAPF32[$237>>2];
  $$neg28 = $242 * $240;
  $243 = ((($window$151)) + -64|0);
  $244 = +HEAPF32[$243>>2];
  $245 = +HEAPF32[$241>>2];
  $$neg29 = $245 * $244;
  $tmp30 = $$neg29 + $186;
  $tmp32 = $$neg28 + $$neg27;
  $tmp34 = $$neg26 + $$neg25;
  $tmp36 = $$neg24 + $$neg23;
  $tmp38 = $$neg22 + $$neg21;
  $tmp40 = $$neg20 + $$neg19;
  $tmp42 = $$neg18 + $$neg17;
  $tmp44 = $$neg16 + $$neg15;
  $tmp45 = -$tmp44;
  $246 = $tmp45 - $tmp42;
  $247 = $246 - $tmp40;
  $248 = $247 - $tmp38;
  $249 = $248 - $tmp36;
  $250 = $249 - $tmp34;
  $251 = $250 - $tmp32;
  $252 = $251 - $tmp30;
  $253 = $252 > 32767.0;
  do {
   if ($253) {
    HEAP16[$samples$253>>1] = 32767;
    $254 = (($clip$349) + 1)|0;
    $clip$4 = $254;
   } else {
    $255 = $252 < -32768.0;
    if ($255) {
     HEAP16[$samples$253>>1] = -32768;
     $256 = (($clip$349) + 1)|0;
     $clip$4 = $256;
     break;
    } else {
     $257 = (~~(($252)));
     HEAP16[$samples$253>>1] = $257;
     $clip$4 = $clip$349;
     break;
    }
   }
  } while(0);
  $258 = (($j$150) + -1)|0;
  $259 = ((($window$151)) + -512|0);
  $b0$2 = ((($b0$252)) + -256|0);
  $samples$2 = ((($samples$253)) + 4|0);
  $260 = ($258|0)==(0);
  if ($260) {
   $clip$4$lcssa = $clip$4;
   break;
  } else {
   $b0$1$pn48$phi = $b0$252;$b0$252 = $b0$2;$clip$349 = $clip$4;$j$150 = $258;$samples$253 = $samples$2;$window$151 = $259;$b0$1$pn48 = $b0$1$pn48$phi;
  }
 }
 $261 = ($final|0)==(0);
 if ($261) {
  return ($clip$4$lcssa|0);
 }
 $262 = HEAP32[$2>>2]|0;
 $263 = (($262) + 32)|0;
 HEAP32[$2>>2] = $263;
 return ($clip$4$lcssa|0);
}
function _INT123_synth_4to1_8bit($bandPtr,$channel,$fr,$final) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $fr = $fr|0;
 $final = $final|0;
 var $$neg = 0.0, $$neg1 = 0.0, $$neg15 = 0.0, $$neg16 = 0.0, $$neg17 = 0.0, $$neg18 = 0.0, $$neg19 = 0.0, $$neg2 = 0.0, $$neg20 = 0.0, $$neg21 = 0.0, $$neg22 = 0.0, $$neg23 = 0.0, $$neg24 = 0.0, $$neg25 = 0.0, $$neg26 = 0.0, $$neg27 = 0.0, $$neg28 = 0.0, $$neg29 = 0.0, $$neg3 = 0.0, $$neg4 = 0.0;
 var $$neg5 = 0.0, $$neg6 = 0.0, $$neg7 = 0.0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0, $102 = 0.0, $103 = 0, $104 = 0.0, $105 = 0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0;
 var $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0, $146 = 0.0, $147 = 0;
 var $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0, $158 = 0.0, $159 = 0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0, $164 = 0.0, $165 = 0;
 var $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0, $17 = 0, $170 = 0.0, $171 = 0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0, $176 = 0.0, $177 = 0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0.0, $198 = 0, $199 = 0.0, $2 = 0, $20 = 0, $200 = 0;
 var $201 = 0.0, $202 = 0, $203 = 0.0, $204 = 0, $205 = 0.0, $206 = 0, $207 = 0.0, $208 = 0, $209 = 0.0, $21 = 0, $210 = 0, $211 = 0.0, $212 = 0, $213 = 0.0, $214 = 0, $215 = 0.0, $216 = 0, $217 = 0.0, $218 = 0, $219 = 0.0;
 var $22 = 0, $220 = 0, $221 = 0.0, $222 = 0, $223 = 0.0, $224 = 0, $225 = 0.0, $226 = 0, $227 = 0.0, $228 = 0, $229 = 0.0, $23 = 0, $230 = 0, $231 = 0.0, $232 = 0, $233 = 0.0, $234 = 0, $235 = 0.0, $236 = 0, $237 = 0.0;
 var $238 = 0, $239 = 0.0, $24 = 0, $240 = 0, $241 = 0.0, $242 = 0, $243 = 0.0, $244 = 0, $245 = 0.0, $246 = 0, $247 = 0.0, $248 = 0, $249 = 0.0, $25 = 0, $250 = 0, $251 = 0.0, $252 = 0, $253 = 0.0, $254 = 0, $255 = 0.0;
 var $256 = 0.0, $257 = 0.0, $258 = 0.0, $259 = 0.0, $26 = 0, $260 = 0.0, $261 = 0.0, $262 = 0.0, $263 = 0.0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0;
 var $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0.0, $42 = 0, $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0, $55 = 0.0, $56 = 0, $57 = 0.0, $58 = 0;
 var $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0.0, $63 = 0, $64 = 0.0, $65 = 0, $66 = 0.0, $67 = 0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0, $73 = 0.0, $74 = 0, $75 = 0.0, $76 = 0;
 var $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0, $86 = 0.0, $87 = 0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0, $91 = 0.0, $92 = 0, $93 = 0.0, $94 = 0;
 var $95 = 0.0, $96 = 0, $97 = 0.0, $98 = 0.0, $99 = 0, $b0$0 = 0, $b0$1$pn48 = 0, $b0$1$pn48$phi = 0, $b0$155 = 0, $b0$2 = 0, $b0$252 = 0, $bo1$0 = 0, $buf$0 = 0, $clip$056 = 0, $clip$1 = 0, $clip$1$lcssa = 0, $clip$2 = 0, $clip$349 = 0, $clip$4 = 0, $clip$4$lcssa = 0;
 var $j$057 = 0, $j$150 = 0, $samples$0 = 0, $samples$154 = 0, $samples$2 = 0, $samples$247 = 0, $samples$253 = 0, $scevgep = 0, $scevgep61 = 0, $scevgep62 = 0, $scevgep64 = 0, $scevgep65 = 0, $tmp = 0.0, $tmp11 = 0.0, $tmp13 = 0.0, $tmp30 = 0.0, $tmp32 = 0.0, $tmp34 = 0.0, $tmp36 = 0.0, $tmp38 = 0.0;
 var $tmp40 = 0.0, $tmp42 = 0.0, $tmp44 = 0.0, $tmp45 = 0.0, $tmp9 = 0.0, $window$058 = 0, $window$151 = 0, $write_8bit_tmp$0 = 0, $write_8bit_tmp2$0 = 0, $write_8bit_tmp4$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($fr)) + 9288|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = ((($fr)) + 9292|0);
  _INT123_do_equalizer($bandPtr,$channel,$8);
 }
 $9 = ($channel|0)==(0);
 if ($9) {
  $10 = ((($fr)) + 9272|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + 15)|0;
  $13 = $12 & 15;
  HEAP32[$10>>2] = $13;
  $14 = ((($fr)) + 9248|0);
  $18 = $13;$buf$0 = $14;$samples$0 = $4;
 } else {
  $15 = ((($4)) + 1|0);
  $16 = ((($fr)) + 9256|0);
  $$phi$trans$insert = ((($fr)) + 9272|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $18 = $$pre;$buf$0 = $16;$samples$0 = $15;
 }
 $17 = $18 & 1;
 $19 = ($17|0)==(0);
 $20 = ((($buf$0)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 if ($19) {
  $27 = (($18) + 1)|0;
  $28 = HEAP32[$buf$0>>2]|0;
  $29 = (($28) + ($18<<2)|0);
  $30 = (($21) + ($18<<2)|0);
  $31 = ((($30)) + 4|0);
  _INT123_dct64($29,$31,$bandPtr);
  $b0$0 = $21;$bo1$0 = $27;
 } else {
  $22 = HEAP32[$buf$0>>2]|0;
  $23 = (($18) + 1)|0;
  $24 = $23 & 15;
  $25 = (($21) + ($24<<2)|0);
  $26 = (($22) + ($18<<2)|0);
  _INT123_dct64($25,$26,$bandPtr);
  $b0$0 = $22;$bo1$0 = $18;
 }
 $32 = ((($fr)) + 9284|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($33)) + 64|0);
 $35 = (0 - ($bo1$0))|0;
 $36 = (($34) + ($35<<2)|0);
 $37 = ((($fr)) + 9592|0);
 $38 = (528 - ($bo1$0))|0;
 $scevgep61 = (($33) + ($38<<2)|0);
 $39 = (400 - ($bo1$0))|0;
 $b0$155 = $b0$0;$clip$056 = 0;$j$057 = 4;$samples$154 = $samples$0;$window$058 = $36;
 while(1) {
  $40 = ((($window$058)) + 4|0);
  $41 = +HEAPF32[$window$058>>2];
  $42 = ((($b0$155)) + 4|0);
  $43 = +HEAPF32[$b0$155>>2];
  $44 = $43 * $41;
  $45 = ((($window$058)) + 8|0);
  $46 = +HEAPF32[$40>>2];
  $47 = ((($b0$155)) + 8|0);
  $48 = +HEAPF32[$42>>2];
  $$neg = $48 * $46;
  $49 = ((($window$058)) + 12|0);
  $50 = +HEAPF32[$45>>2];
  $51 = ((($b0$155)) + 12|0);
  $52 = +HEAPF32[$47>>2];
  $53 = $52 * $50;
  $54 = ((($window$058)) + 16|0);
  $55 = +HEAPF32[$49>>2];
  $56 = ((($b0$155)) + 16|0);
  $57 = +HEAPF32[$51>>2];
  $$neg1 = $57 * $55;
  $58 = ((($window$058)) + 20|0);
  $59 = +HEAPF32[$54>>2];
  $60 = ((($b0$155)) + 20|0);
  $61 = +HEAPF32[$56>>2];
  $62 = $61 * $59;
  $63 = ((($window$058)) + 24|0);
  $64 = +HEAPF32[$58>>2];
  $65 = ((($b0$155)) + 24|0);
  $66 = +HEAPF32[$60>>2];
  $$neg2 = $66 * $64;
  $67 = ((($window$058)) + 28|0);
  $68 = +HEAPF32[$63>>2];
  $69 = ((($b0$155)) + 28|0);
  $70 = +HEAPF32[$65>>2];
  $71 = $70 * $68;
  $72 = ((($window$058)) + 32|0);
  $73 = +HEAPF32[$67>>2];
  $74 = ((($b0$155)) + 32|0);
  $75 = +HEAPF32[$69>>2];
  $$neg3 = $75 * $73;
  $76 = ((($window$058)) + 36|0);
  $77 = +HEAPF32[$72>>2];
  $78 = ((($b0$155)) + 36|0);
  $79 = +HEAPF32[$74>>2];
  $80 = $79 * $77;
  $81 = ((($window$058)) + 40|0);
  $82 = +HEAPF32[$76>>2];
  $83 = ((($b0$155)) + 40|0);
  $84 = +HEAPF32[$78>>2];
  $$neg4 = $84 * $82;
  $85 = ((($window$058)) + 44|0);
  $86 = +HEAPF32[$81>>2];
  $87 = ((($b0$155)) + 44|0);
  $88 = +HEAPF32[$83>>2];
  $89 = $88 * $86;
  $90 = ((($window$058)) + 48|0);
  $91 = +HEAPF32[$85>>2];
  $92 = ((($b0$155)) + 48|0);
  $93 = +HEAPF32[$87>>2];
  $$neg5 = $93 * $91;
  $94 = ((($window$058)) + 52|0);
  $95 = +HEAPF32[$90>>2];
  $96 = ((($b0$155)) + 52|0);
  $97 = +HEAPF32[$92>>2];
  $98 = $97 * $95;
  $99 = ((($window$058)) + 56|0);
  $100 = +HEAPF32[$94>>2];
  $101 = ((($b0$155)) + 56|0);
  $102 = +HEAPF32[$96>>2];
  $$neg6 = $102 * $100;
  $103 = ((($window$058)) + 60|0);
  $104 = +HEAPF32[$99>>2];
  $105 = ((($b0$155)) + 60|0);
  $106 = +HEAPF32[$101>>2];
  $107 = $106 * $104;
  $108 = +HEAPF32[$103>>2];
  $109 = +HEAPF32[$105>>2];
  $$neg7 = $109 * $108;
  $tmp = $$neg7 + $$neg6;
  $tmp9 = $$neg5 + $$neg4;
  $tmp11 = $$neg3 + $$neg2;
  $tmp13 = $$neg1 + $$neg;
  $110 = $53 + $44;
  $111 = $110 + $62;
  $112 = $111 + $71;
  $113 = $112 + $80;
  $114 = $113 + $89;
  $115 = $114 + $98;
  $116 = $115 + $107;
  $117 = $116 - $tmp13;
  $118 = $117 - $tmp11;
  $119 = $118 - $tmp9;
  $120 = $119 - $tmp;
  $121 = $120 > 32767.0;
  do {
   if ($121) {
    $122 = (($clip$056) + 1)|0;
    $clip$1 = $122;$write_8bit_tmp$0 = 32767;
   } else {
    $123 = $120 < -32768.0;
    if ($123) {
     $124 = (($clip$056) + 1)|0;
     $clip$1 = $124;$write_8bit_tmp$0 = -32768;
     break;
    } else {
     $125 = (~~(($120)));
     $clip$1 = $clip$056;$write_8bit_tmp$0 = $125;
     break;
    }
   }
  } while(0);
  $126 = $write_8bit_tmp$0 << 16 >> 16;
  $127 = $126 >> 3;
  $128 = HEAP32[$37>>2]|0;
  $129 = (($128) + ($127)|0);
  $130 = HEAP8[$129>>0]|0;
  HEAP8[$samples$154>>0] = $130;
  $131 = (($j$057) + -1)|0;
  $132 = ((($b0$155)) + 256|0);
  $133 = ((($window$058)) + 512|0);
  $134 = ((($samples$154)) + 2|0);
  $135 = ($131|0)==(0);
  if ($135) {
   $clip$1$lcssa = $clip$1;
   break;
  } else {
   $b0$155 = $132;$clip$056 = $clip$1;$j$057 = $131;$samples$154 = $134;$window$058 = $133;
  }
 }
 $scevgep = ((($samples$0)) + 8|0);
 $scevgep62 = ((($b0$0)) + 1024|0);
 $scevgep64 = ((($b0$0)) + 768|0);
 $scevgep65 = (($33) + ($39<<2)|0);
 $136 = +HEAPF32[$scevgep61>>2];
 $137 = +HEAPF32[$scevgep62>>2];
 $138 = $137 * $136;
 $139 = ((($scevgep65)) + 520|0);
 $140 = +HEAPF32[$139>>2];
 $141 = ((($b0$0)) + 1032|0);
 $142 = +HEAPF32[$141>>2];
 $143 = $142 * $140;
 $144 = $143 + $138;
 $145 = ((($scevgep65)) + 528|0);
 $146 = +HEAPF32[$145>>2];
 $147 = ((($b0$0)) + 1040|0);
 $148 = +HEAPF32[$147>>2];
 $149 = $148 * $146;
 $150 = $144 + $149;
 $151 = ((($scevgep65)) + 536|0);
 $152 = +HEAPF32[$151>>2];
 $153 = ((($b0$0)) + 1048|0);
 $154 = +HEAPF32[$153>>2];
 $155 = $154 * $152;
 $156 = $150 + $155;
 $157 = ((($scevgep65)) + 544|0);
 $158 = +HEAPF32[$157>>2];
 $159 = ((($b0$0)) + 1056|0);
 $160 = +HEAPF32[$159>>2];
 $161 = $160 * $158;
 $162 = $156 + $161;
 $163 = ((($scevgep65)) + 552|0);
 $164 = +HEAPF32[$163>>2];
 $165 = ((($b0$0)) + 1064|0);
 $166 = +HEAPF32[$165>>2];
 $167 = $166 * $164;
 $168 = $162 + $167;
 $169 = ((($scevgep65)) + 560|0);
 $170 = +HEAPF32[$169>>2];
 $171 = ((($b0$0)) + 1072|0);
 $172 = +HEAPF32[$171>>2];
 $173 = $172 * $170;
 $174 = $168 + $173;
 $175 = ((($scevgep65)) + 568|0);
 $176 = +HEAPF32[$175>>2];
 $177 = ((($b0$0)) + 1080|0);
 $178 = +HEAPF32[$177>>2];
 $179 = $178 * $176;
 $180 = $174 + $179;
 $181 = $180 > 32767.0;
 do {
  if ($181) {
   $182 = (($clip$1$lcssa) + 1)|0;
   $clip$2 = $182;$write_8bit_tmp2$0 = 32767;
  } else {
   $183 = $180 < -32768.0;
   if ($183) {
    $184 = (($clip$1$lcssa) + 1)|0;
    $clip$2 = $184;$write_8bit_tmp2$0 = -32768;
    break;
   } else {
    $185 = (~~(($180)));
    $clip$2 = $clip$1$lcssa;$write_8bit_tmp2$0 = $185;
    break;
   }
  }
 } while(0);
 $186 = $write_8bit_tmp2$0 << 16 >> 16;
 $187 = $186 >> 3;
 $188 = HEAP32[$37>>2]|0;
 $189 = (($188) + ($187)|0);
 $190 = HEAP8[$189>>0]|0;
 HEAP8[$scevgep>>0] = $190;
 $191 = $bo1$0 << 1;
 $192 = (($scevgep65) + ($191<<2)|0);
 $samples$247 = ((($samples$0)) + 10|0);
 $b0$1$pn48 = $scevgep62;$b0$252 = $scevgep64;$clip$349 = $clip$2;$j$150 = 3;$samples$253 = $samples$247;$window$151 = $192;
 while(1) {
  $193 = ((($window$151)) + -4|0);
  $194 = +HEAPF32[$193>>2];
  $195 = ((($b0$1$pn48)) + -252|0);
  $196 = +HEAPF32[$b0$252>>2];
  $197 = $196 * $194;
  $198 = ((($window$151)) + -8|0);
  $199 = +HEAPF32[$198>>2];
  $200 = ((($b0$1$pn48)) + -248|0);
  $201 = +HEAPF32[$195>>2];
  $$neg15 = $201 * $199;
  $202 = ((($window$151)) + -12|0);
  $203 = +HEAPF32[$202>>2];
  $204 = ((($b0$1$pn48)) + -244|0);
  $205 = +HEAPF32[$200>>2];
  $$neg16 = $205 * $203;
  $206 = ((($window$151)) + -16|0);
  $207 = +HEAPF32[$206>>2];
  $208 = ((($b0$1$pn48)) + -240|0);
  $209 = +HEAPF32[$204>>2];
  $$neg17 = $209 * $207;
  $210 = ((($window$151)) + -20|0);
  $211 = +HEAPF32[$210>>2];
  $212 = ((($b0$1$pn48)) + -236|0);
  $213 = +HEAPF32[$208>>2];
  $$neg18 = $213 * $211;
  $214 = ((($window$151)) + -24|0);
  $215 = +HEAPF32[$214>>2];
  $216 = ((($b0$1$pn48)) + -232|0);
  $217 = +HEAPF32[$212>>2];
  $$neg19 = $217 * $215;
  $218 = ((($window$151)) + -28|0);
  $219 = +HEAPF32[$218>>2];
  $220 = ((($b0$1$pn48)) + -228|0);
  $221 = +HEAPF32[$216>>2];
  $$neg20 = $221 * $219;
  $222 = ((($window$151)) + -32|0);
  $223 = +HEAPF32[$222>>2];
  $224 = ((($b0$1$pn48)) + -224|0);
  $225 = +HEAPF32[$220>>2];
  $$neg21 = $225 * $223;
  $226 = ((($window$151)) + -36|0);
  $227 = +HEAPF32[$226>>2];
  $228 = ((($b0$1$pn48)) + -220|0);
  $229 = +HEAPF32[$224>>2];
  $$neg22 = $229 * $227;
  $230 = ((($window$151)) + -40|0);
  $231 = +HEAPF32[$230>>2];
  $232 = ((($b0$1$pn48)) + -216|0);
  $233 = +HEAPF32[$228>>2];
  $$neg23 = $233 * $231;
  $234 = ((($window$151)) + -44|0);
  $235 = +HEAPF32[$234>>2];
  $236 = ((($b0$1$pn48)) + -212|0);
  $237 = +HEAPF32[$232>>2];
  $$neg24 = $237 * $235;
  $238 = ((($window$151)) + -48|0);
  $239 = +HEAPF32[$238>>2];
  $240 = ((($b0$1$pn48)) + -208|0);
  $241 = +HEAPF32[$236>>2];
  $$neg25 = $241 * $239;
  $242 = ((($window$151)) + -52|0);
  $243 = +HEAPF32[$242>>2];
  $244 = ((($b0$1$pn48)) + -204|0);
  $245 = +HEAPF32[$240>>2];
  $$neg26 = $245 * $243;
  $246 = ((($window$151)) + -56|0);
  $247 = +HEAPF32[$246>>2];
  $248 = ((($b0$1$pn48)) + -200|0);
  $249 = +HEAPF32[$244>>2];
  $$neg27 = $249 * $247;
  $250 = ((($window$151)) + -60|0);
  $251 = +HEAPF32[$250>>2];
  $252 = ((($b0$1$pn48)) + -196|0);
  $253 = +HEAPF32[$248>>2];
  $$neg28 = $253 * $251;
  $254 = ((($window$151)) + -64|0);
  $255 = +HEAPF32[$254>>2];
  $256 = +HEAPF32[$252>>2];
  $$neg29 = $256 * $255;
  $tmp30 = $$neg29 + $197;
  $tmp32 = $$neg28 + $$neg27;
  $tmp34 = $$neg26 + $$neg25;
  $tmp36 = $$neg24 + $$neg23;
  $tmp38 = $$neg22 + $$neg21;
  $tmp40 = $$neg20 + $$neg19;
  $tmp42 = $$neg18 + $$neg17;
  $tmp44 = $$neg16 + $$neg15;
  $tmp45 = -$tmp44;
  $257 = $tmp45 - $tmp42;
  $258 = $257 - $tmp40;
  $259 = $258 - $tmp38;
  $260 = $259 - $tmp36;
  $261 = $260 - $tmp34;
  $262 = $261 - $tmp32;
  $263 = $262 - $tmp30;
  $264 = $263 > 32767.0;
  do {
   if ($264) {
    $265 = (($clip$349) + 1)|0;
    $clip$4 = $265;$write_8bit_tmp4$0 = 32767;
   } else {
    $266 = $263 < -32768.0;
    if ($266) {
     $267 = (($clip$349) + 1)|0;
     $clip$4 = $267;$write_8bit_tmp4$0 = -32768;
     break;
    } else {
     $268 = (~~(($263)));
     $clip$4 = $clip$349;$write_8bit_tmp4$0 = $268;
     break;
    }
   }
  } while(0);
  $269 = $write_8bit_tmp4$0 << 16 >> 16;
  $270 = $269 >> 3;
  $271 = HEAP32[$37>>2]|0;
  $272 = (($271) + ($270)|0);
  $273 = HEAP8[$272>>0]|0;
  HEAP8[$samples$253>>0] = $273;
  $274 = (($j$150) + -1)|0;
  $275 = ((($window$151)) + -512|0);
  $b0$2 = ((($b0$252)) + -256|0);
  $samples$2 = ((($samples$253)) + 2|0);
  $276 = ($274|0)==(0);
  if ($276) {
   $clip$4$lcssa = $clip$4;
   break;
  } else {
   $b0$1$pn48$phi = $b0$252;$b0$252 = $b0$2;$clip$349 = $clip$4;$j$150 = $274;$samples$253 = $samples$2;$window$151 = $275;$b0$1$pn48 = $b0$1$pn48$phi;
  }
 }
 $277 = ($final|0)==(0);
 if ($277) {
  return ($clip$4$lcssa|0);
 }
 $278 = HEAP32[$2>>2]|0;
 $279 = (($278) + 16)|0;
 HEAP32[$2>>2] = $279;
 return ($clip$4$lcssa|0);
}
function _INT123_synth_4to1_real($bandPtr,$channel,$fr,$final) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $fr = $fr|0;
 $final = $final|0;
 var $$neg = 0.0, $$neg1 = 0.0, $$neg15 = 0.0, $$neg16 = 0.0, $$neg17 = 0.0, $$neg18 = 0.0, $$neg19 = 0.0, $$neg2 = 0.0, $$neg20 = 0.0, $$neg21 = 0.0, $$neg22 = 0.0, $$neg23 = 0.0, $$neg24 = 0.0, $$neg25 = 0.0, $$neg26 = 0.0, $$neg27 = 0.0, $$neg28 = 0.0, $$neg29 = 0.0, $$neg3 = 0.0, $$neg4 = 0.0;
 var $$neg5 = 0.0, $$neg6 = 0.0, $$neg7 = 0.0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0;
 var $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0;
 var $13 = 0, $130 = 0.0, $131 = 0, $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0, $136 = 0.0, $137 = 0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0;
 var $148 = 0.0, $149 = 0, $15 = 0, $150 = 0.0, $151 = 0.0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0, $16 = 0, $160 = 0.0, $161 = 0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0;
 var $166 = 0.0, $167 = 0, $168 = 0.0, $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0.0, $172 = 0, $173 = 0, $174 = 0, $175 = 0.0, $176 = 0, $177 = 0.0, $178 = 0.0, $179 = 0, $18 = 0, $180 = 0.0, $181 = 0, $182 = 0.0, $183 = 0;
 var $184 = 0.0, $185 = 0, $186 = 0.0, $187 = 0, $188 = 0.0, $189 = 0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0, $200 = 0.0;
 var $201 = 0, $202 = 0.0, $203 = 0, $204 = 0.0, $205 = 0, $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0.0, $213 = 0, $214 = 0.0, $215 = 0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0;
 var $22 = 0, $220 = 0.0, $221 = 0, $222 = 0.0, $223 = 0, $224 = 0.0, $225 = 0, $226 = 0.0, $227 = 0, $228 = 0.0, $229 = 0, $23 = 0, $230 = 0.0, $231 = 0, $232 = 0.0, $233 = 0, $234 = 0.0, $235 = 0, $236 = 0.0, $237 = 0.0;
 var $238 = 0.0, $239 = 0.0, $24 = 0, $240 = 0.0, $241 = 0.0, $242 = 0.0, $243 = 0.0, $244 = 0.0, $245 = 0.0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0, $45 = 0.0, $46 = 0, $47 = 0.0;
 var $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0.0;
 var $66 = 0, $67 = 0.0, $68 = 0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0, $83 = 0.0;
 var $84 = 0, $85 = 0.0, $86 = 0, $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0, $90 = 0.0, $91 = 0, $92 = 0.0, $93 = 0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0, $99 = 0.0, $b0$0 = 0, $b0$1$pn50 = 0, $b0$1$pn50$phi = 0;
 var $b0$155 = 0, $b0$2 = 0, $b0$251 = 0, $bo1$0 = 0, $buf$0 = 0, $j$053 = 0, $j$148 = 0, $samples$0$in = 0, $samples$156 = 0, $samples$2 = 0, $samples$247 = 0, $samples$252 = 0, $scevgep = 0, $scevgep60 = 0, $scevgep61 = 0, $scevgep62 = 0, $scevgep63 = 0, $tmp = 0.0, $tmp11 = 0.0, $tmp13 = 0.0;
 var $tmp30 = 0.0, $tmp32 = 0.0, $tmp34 = 0.0, $tmp36 = 0.0, $tmp38 = 0.0, $tmp40 = 0.0, $tmp42 = 0.0, $tmp44 = 0.0, $tmp45 = 0.0, $tmp9 = 0.0, $window$054 = 0, $window$149 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($fr)) + 9288|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = ((($fr)) + 9292|0);
  _INT123_do_equalizer($bandPtr,$channel,$8);
 }
 $9 = ($channel|0)==(0);
 if ($9) {
  $10 = ((($fr)) + 9272|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + 15)|0;
  $13 = $12 & 15;
  HEAP32[$10>>2] = $13;
  $14 = ((($fr)) + 9248|0);
  $18 = $13;$buf$0 = $14;$samples$0$in = $4;
 } else {
  $15 = ((($4)) + 4|0);
  $16 = ((($fr)) + 9256|0);
  $$phi$trans$insert = ((($fr)) + 9272|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $18 = $$pre;$buf$0 = $16;$samples$0$in = $15;
 }
 $17 = $18 & 1;
 $19 = ($17|0)==(0);
 $20 = ((($buf$0)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 if ($19) {
  $27 = (($18) + 1)|0;
  $28 = HEAP32[$buf$0>>2]|0;
  $29 = (($28) + ($18<<2)|0);
  $30 = (($21) + ($18<<2)|0);
  $31 = ((($30)) + 4|0);
  _INT123_dct64($29,$31,$bandPtr);
  $b0$0 = $21;$bo1$0 = $27;
 } else {
  $22 = HEAP32[$buf$0>>2]|0;
  $23 = (($18) + 1)|0;
  $24 = $23 & 15;
  $25 = (($21) + ($24<<2)|0);
  $26 = (($22) + ($18<<2)|0);
  _INT123_dct64($25,$26,$bandPtr);
  $b0$0 = $22;$bo1$0 = $18;
 }
 $32 = ((($fr)) + 9284|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($33)) + 64|0);
 $35 = (0 - ($bo1$0))|0;
 $36 = (($34) + ($35<<2)|0);
 $scevgep = ((($samples$0$in)) + 32|0);
 $37 = (528 - ($bo1$0))|0;
 $scevgep60 = (($33) + ($37<<2)|0);
 $38 = (400 - ($bo1$0))|0;
 $scevgep62 = (($33) + ($38<<2)|0);
 $b0$155 = $b0$0;$j$053 = 4;$samples$156 = $samples$0$in;$window$054 = $36;
 while(1) {
  $39 = ((($window$054)) + 4|0);
  $40 = +HEAPF32[$window$054>>2];
  $41 = ((($b0$155)) + 4|0);
  $42 = +HEAPF32[$b0$155>>2];
  $43 = $42 * $40;
  $44 = ((($window$054)) + 8|0);
  $45 = +HEAPF32[$39>>2];
  $46 = ((($b0$155)) + 8|0);
  $47 = +HEAPF32[$41>>2];
  $$neg = $47 * $45;
  $48 = ((($window$054)) + 12|0);
  $49 = +HEAPF32[$44>>2];
  $50 = ((($b0$155)) + 12|0);
  $51 = +HEAPF32[$46>>2];
  $52 = $51 * $49;
  $53 = ((($window$054)) + 16|0);
  $54 = +HEAPF32[$48>>2];
  $55 = ((($b0$155)) + 16|0);
  $56 = +HEAPF32[$50>>2];
  $$neg1 = $56 * $54;
  $57 = ((($window$054)) + 20|0);
  $58 = +HEAPF32[$53>>2];
  $59 = ((($b0$155)) + 20|0);
  $60 = +HEAPF32[$55>>2];
  $61 = $60 * $58;
  $62 = ((($window$054)) + 24|0);
  $63 = +HEAPF32[$57>>2];
  $64 = ((($b0$155)) + 24|0);
  $65 = +HEAPF32[$59>>2];
  $$neg2 = $65 * $63;
  $66 = ((($window$054)) + 28|0);
  $67 = +HEAPF32[$62>>2];
  $68 = ((($b0$155)) + 28|0);
  $69 = +HEAPF32[$64>>2];
  $70 = $69 * $67;
  $71 = ((($window$054)) + 32|0);
  $72 = +HEAPF32[$66>>2];
  $73 = ((($b0$155)) + 32|0);
  $74 = +HEAPF32[$68>>2];
  $$neg3 = $74 * $72;
  $75 = ((($window$054)) + 36|0);
  $76 = +HEAPF32[$71>>2];
  $77 = ((($b0$155)) + 36|0);
  $78 = +HEAPF32[$73>>2];
  $79 = $78 * $76;
  $80 = ((($window$054)) + 40|0);
  $81 = +HEAPF32[$75>>2];
  $82 = ((($b0$155)) + 40|0);
  $83 = +HEAPF32[$77>>2];
  $$neg4 = $83 * $81;
  $84 = ((($window$054)) + 44|0);
  $85 = +HEAPF32[$80>>2];
  $86 = ((($b0$155)) + 44|0);
  $87 = +HEAPF32[$82>>2];
  $88 = $87 * $85;
  $89 = ((($window$054)) + 48|0);
  $90 = +HEAPF32[$84>>2];
  $91 = ((($b0$155)) + 48|0);
  $92 = +HEAPF32[$86>>2];
  $$neg5 = $92 * $90;
  $93 = ((($window$054)) + 52|0);
  $94 = +HEAPF32[$89>>2];
  $95 = ((($b0$155)) + 52|0);
  $96 = +HEAPF32[$91>>2];
  $97 = $96 * $94;
  $98 = ((($window$054)) + 56|0);
  $99 = +HEAPF32[$93>>2];
  $100 = ((($b0$155)) + 56|0);
  $101 = +HEAPF32[$95>>2];
  $$neg6 = $101 * $99;
  $102 = ((($window$054)) + 60|0);
  $103 = +HEAPF32[$98>>2];
  $104 = ((($b0$155)) + 60|0);
  $105 = +HEAPF32[$100>>2];
  $106 = $105 * $103;
  $107 = +HEAPF32[$102>>2];
  $108 = +HEAPF32[$104>>2];
  $$neg7 = $108 * $107;
  $tmp = $$neg7 + $$neg6;
  $tmp9 = $$neg5 + $$neg4;
  $tmp11 = $$neg3 + $$neg2;
  $tmp13 = $$neg1 + $$neg;
  $109 = $52 + $43;
  $110 = $109 + $61;
  $111 = $110 + $70;
  $112 = $111 + $79;
  $113 = $112 + $88;
  $114 = $113 + $97;
  $115 = $114 + $106;
  $116 = $115 - $tmp13;
  $117 = $116 - $tmp11;
  $118 = $117 - $tmp9;
  $119 = $118 - $tmp;
  $120 = $119 * 3.0517578125E-5;
  HEAPF32[$samples$156>>2] = $120;
  $121 = (($j$053) + -1)|0;
  $122 = ((($b0$155)) + 256|0);
  $123 = ((($window$054)) + 512|0);
  $124 = ((($samples$156)) + 8|0);
  $125 = ($121|0)==(0);
  if ($125) {
   break;
  } else {
   $b0$155 = $122;$j$053 = $121;$samples$156 = $124;$window$054 = $123;
  }
 }
 $scevgep61 = ((($b0$0)) + 1024|0);
 $scevgep63 = ((($b0$0)) + 768|0);
 $126 = +HEAPF32[$scevgep60>>2];
 $127 = +HEAPF32[$scevgep61>>2];
 $128 = $127 * $126;
 $129 = ((($scevgep62)) + 520|0);
 $130 = +HEAPF32[$129>>2];
 $131 = ((($b0$0)) + 1032|0);
 $132 = +HEAPF32[$131>>2];
 $133 = $132 * $130;
 $134 = $133 + $128;
 $135 = ((($scevgep62)) + 528|0);
 $136 = +HEAPF32[$135>>2];
 $137 = ((($b0$0)) + 1040|0);
 $138 = +HEAPF32[$137>>2];
 $139 = $138 * $136;
 $140 = $134 + $139;
 $141 = ((($scevgep62)) + 536|0);
 $142 = +HEAPF32[$141>>2];
 $143 = ((($b0$0)) + 1048|0);
 $144 = +HEAPF32[$143>>2];
 $145 = $144 * $142;
 $146 = $140 + $145;
 $147 = ((($scevgep62)) + 544|0);
 $148 = +HEAPF32[$147>>2];
 $149 = ((($b0$0)) + 1056|0);
 $150 = +HEAPF32[$149>>2];
 $151 = $150 * $148;
 $152 = $146 + $151;
 $153 = ((($scevgep62)) + 552|0);
 $154 = +HEAPF32[$153>>2];
 $155 = ((($b0$0)) + 1064|0);
 $156 = +HEAPF32[$155>>2];
 $157 = $156 * $154;
 $158 = $152 + $157;
 $159 = ((($scevgep62)) + 560|0);
 $160 = +HEAPF32[$159>>2];
 $161 = ((($b0$0)) + 1072|0);
 $162 = +HEAPF32[$161>>2];
 $163 = $162 * $160;
 $164 = $158 + $163;
 $165 = ((($scevgep62)) + 568|0);
 $166 = +HEAPF32[$165>>2];
 $167 = ((($b0$0)) + 1080|0);
 $168 = +HEAPF32[$167>>2];
 $169 = $168 * $166;
 $170 = $164 + $169;
 $171 = $170 * 3.0517578125E-5;
 HEAPF32[$scevgep>>2] = $171;
 $172 = $bo1$0 << 1;
 $173 = (($scevgep62) + ($172<<2)|0);
 $samples$247 = ((($samples$0$in)) + 40|0);
 $b0$1$pn50 = $scevgep61;$b0$251 = $scevgep63;$j$148 = 3;$samples$252 = $samples$247;$window$149 = $173;
 while(1) {
  $174 = ((($window$149)) + -4|0);
  $175 = +HEAPF32[$174>>2];
  $176 = ((($b0$1$pn50)) + -252|0);
  $177 = +HEAPF32[$b0$251>>2];
  $178 = $177 * $175;
  $179 = ((($window$149)) + -8|0);
  $180 = +HEAPF32[$179>>2];
  $181 = ((($b0$1$pn50)) + -248|0);
  $182 = +HEAPF32[$176>>2];
  $$neg15 = $182 * $180;
  $183 = ((($window$149)) + -12|0);
  $184 = +HEAPF32[$183>>2];
  $185 = ((($b0$1$pn50)) + -244|0);
  $186 = +HEAPF32[$181>>2];
  $$neg16 = $186 * $184;
  $187 = ((($window$149)) + -16|0);
  $188 = +HEAPF32[$187>>2];
  $189 = ((($b0$1$pn50)) + -240|0);
  $190 = +HEAPF32[$185>>2];
  $$neg17 = $190 * $188;
  $191 = ((($window$149)) + -20|0);
  $192 = +HEAPF32[$191>>2];
  $193 = ((($b0$1$pn50)) + -236|0);
  $194 = +HEAPF32[$189>>2];
  $$neg18 = $194 * $192;
  $195 = ((($window$149)) + -24|0);
  $196 = +HEAPF32[$195>>2];
  $197 = ((($b0$1$pn50)) + -232|0);
  $198 = +HEAPF32[$193>>2];
  $$neg19 = $198 * $196;
  $199 = ((($window$149)) + -28|0);
  $200 = +HEAPF32[$199>>2];
  $201 = ((($b0$1$pn50)) + -228|0);
  $202 = +HEAPF32[$197>>2];
  $$neg20 = $202 * $200;
  $203 = ((($window$149)) + -32|0);
  $204 = +HEAPF32[$203>>2];
  $205 = ((($b0$1$pn50)) + -224|0);
  $206 = +HEAPF32[$201>>2];
  $$neg21 = $206 * $204;
  $207 = ((($window$149)) + -36|0);
  $208 = +HEAPF32[$207>>2];
  $209 = ((($b0$1$pn50)) + -220|0);
  $210 = +HEAPF32[$205>>2];
  $$neg22 = $210 * $208;
  $211 = ((($window$149)) + -40|0);
  $212 = +HEAPF32[$211>>2];
  $213 = ((($b0$1$pn50)) + -216|0);
  $214 = +HEAPF32[$209>>2];
  $$neg23 = $214 * $212;
  $215 = ((($window$149)) + -44|0);
  $216 = +HEAPF32[$215>>2];
  $217 = ((($b0$1$pn50)) + -212|0);
  $218 = +HEAPF32[$213>>2];
  $$neg24 = $218 * $216;
  $219 = ((($window$149)) + -48|0);
  $220 = +HEAPF32[$219>>2];
  $221 = ((($b0$1$pn50)) + -208|0);
  $222 = +HEAPF32[$217>>2];
  $$neg25 = $222 * $220;
  $223 = ((($window$149)) + -52|0);
  $224 = +HEAPF32[$223>>2];
  $225 = ((($b0$1$pn50)) + -204|0);
  $226 = +HEAPF32[$221>>2];
  $$neg26 = $226 * $224;
  $227 = ((($window$149)) + -56|0);
  $228 = +HEAPF32[$227>>2];
  $229 = ((($b0$1$pn50)) + -200|0);
  $230 = +HEAPF32[$225>>2];
  $$neg27 = $230 * $228;
  $231 = ((($window$149)) + -60|0);
  $232 = +HEAPF32[$231>>2];
  $233 = ((($b0$1$pn50)) + -196|0);
  $234 = +HEAPF32[$229>>2];
  $$neg28 = $234 * $232;
  $235 = ((($window$149)) + -64|0);
  $236 = +HEAPF32[$235>>2];
  $237 = +HEAPF32[$233>>2];
  $$neg29 = $237 * $236;
  $tmp30 = $$neg29 + $178;
  $tmp32 = $$neg28 + $$neg27;
  $tmp34 = $$neg26 + $$neg25;
  $tmp36 = $$neg24 + $$neg23;
  $tmp38 = $$neg22 + $$neg21;
  $tmp40 = $$neg20 + $$neg19;
  $tmp42 = $$neg18 + $$neg17;
  $tmp44 = $$neg16 + $$neg15;
  $tmp45 = -$tmp44;
  $238 = $tmp45 - $tmp42;
  $239 = $238 - $tmp40;
  $240 = $239 - $tmp38;
  $241 = $240 - $tmp36;
  $242 = $241 - $tmp34;
  $243 = $242 - $tmp32;
  $244 = $243 - $tmp30;
  $245 = $244 * 3.0517578125E-5;
  HEAPF32[$samples$252>>2] = $245;
  $246 = (($j$148) + -1)|0;
  $247 = ((($window$149)) + -512|0);
  $b0$2 = ((($b0$251)) + -256|0);
  $samples$2 = ((($samples$252)) + 8|0);
  $248 = ($246|0)==(0);
  if ($248) {
   break;
  } else {
   $b0$1$pn50$phi = $b0$251;$b0$251 = $b0$2;$j$148 = $246;$samples$252 = $samples$2;$window$149 = $247;$b0$1$pn50 = $b0$1$pn50$phi;
  }
 }
 $249 = ($final|0)==(0);
 if ($249) {
  return 0;
 }
 $250 = HEAP32[$2>>2]|0;
 $251 = (($250) + 64)|0;
 HEAP32[$2>>2] = $251;
 return 0;
}
function _INT123_synth_4to1_s32($bandPtr,$channel,$fr,$final) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $fr = $fr|0;
 $final = $final|0;
 var $$neg = 0.0, $$neg1 = 0.0, $$neg15 = 0.0, $$neg16 = 0.0, $$neg17 = 0.0, $$neg18 = 0.0, $$neg19 = 0.0, $$neg2 = 0.0, $$neg20 = 0.0, $$neg21 = 0.0, $$neg22 = 0.0, $$neg23 = 0.0, $$neg24 = 0.0, $$neg25 = 0.0, $$neg26 = 0.0, $$neg27 = 0.0, $$neg28 = 0.0, $$neg29 = 0.0, $$neg3 = 0.0, $$neg4 = 0.0;
 var $$neg5 = 0.0, $$neg6 = 0.0, $$neg7 = 0.0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0;
 var $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0, $136 = 0.0, $137 = 0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0, $142 = 0.0, $143 = 0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0;
 var $148 = 0.0, $149 = 0, $15 = 0, $150 = 0.0, $151 = 0.0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0, $16 = 0, $160 = 0.0, $161 = 0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0;
 var $166 = 0.0, $167 = 0, $168 = 0.0, $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0, $172 = 0.0, $173 = 0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0.0, $188 = 0, $189 = 0.0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0, $200 = 0.0;
 var $201 = 0, $202 = 0.0, $203 = 0, $204 = 0.0, $205 = 0, $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0.0, $213 = 0, $214 = 0.0, $215 = 0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0;
 var $22 = 0, $220 = 0.0, $221 = 0, $222 = 0.0, $223 = 0, $224 = 0.0, $225 = 0, $226 = 0.0, $227 = 0, $228 = 0.0, $229 = 0, $23 = 0, $230 = 0.0, $231 = 0, $232 = 0.0, $233 = 0, $234 = 0.0, $235 = 0, $236 = 0.0, $237 = 0;
 var $238 = 0.0, $239 = 0, $24 = 0, $240 = 0.0, $241 = 0, $242 = 0.0, $243 = 0, $244 = 0.0, $245 = 0, $246 = 0.0, $247 = 0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0.0, $254 = 0.0, $255 = 0.0;
 var $256 = 0.0, $257 = 0.0, $258 = 0.0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0, $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0, $49 = 0.0;
 var $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0;
 var $68 = 0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0, $85 = 0.0;
 var $86 = 0, $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0, $90 = 0.0, $91 = 0, $92 = 0.0, $93 = 0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0, $99 = 0.0, $b0$0 = 0, $b0$1$pn51 = 0, $b0$1$pn51$phi = 0, $b0$157 = 0, $b0$2 = 0;
 var $b0$252 = 0, $bo1$0 = 0, $buf$0 = 0, $clip$056 = 0, $clip$1 = 0, $clip$1$lcssa = 0, $clip$2 = 0, $clip$350 = 0, $clip$4 = 0, $clip$4$lcssa = 0, $j$055 = 0, $j$149 = 0, $samples$0$in = 0, $samples$158 = 0, $samples$2 = 0, $samples$247 = 0, $samples$253 = 0, $scevgep = 0, $scevgep62 = 0, $scevgep63 = 0;
 var $scevgep64 = 0, $scevgep65 = 0, $tmp = 0.0, $tmp11 = 0.0, $tmp13 = 0.0, $tmp30 = 0.0, $tmp32 = 0.0, $tmp34 = 0.0, $tmp36 = 0.0, $tmp38 = 0.0, $tmp40 = 0.0, $tmp42 = 0.0, $tmp44 = 0.0, $tmp45 = 0.0, $tmp9 = 0.0, $window$054 = 0, $window$148 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($fr)) + 9288|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = ((($fr)) + 9292|0);
  _INT123_do_equalizer($bandPtr,$channel,$8);
 }
 $9 = ($channel|0)==(0);
 if ($9) {
  $10 = ((($fr)) + 9272|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + 15)|0;
  $13 = $12 & 15;
  HEAP32[$10>>2] = $13;
  $14 = ((($fr)) + 9248|0);
  $18 = $13;$buf$0 = $14;$samples$0$in = $4;
 } else {
  $15 = ((($4)) + 4|0);
  $16 = ((($fr)) + 9256|0);
  $$phi$trans$insert = ((($fr)) + 9272|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $18 = $$pre;$buf$0 = $16;$samples$0$in = $15;
 }
 $17 = $18 & 1;
 $19 = ($17|0)==(0);
 $20 = ((($buf$0)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 if ($19) {
  $27 = (($18) + 1)|0;
  $28 = HEAP32[$buf$0>>2]|0;
  $29 = (($28) + ($18<<2)|0);
  $30 = (($21) + ($18<<2)|0);
  $31 = ((($30)) + 4|0);
  _INT123_dct64($29,$31,$bandPtr);
  $b0$0 = $21;$bo1$0 = $27;
 } else {
  $22 = HEAP32[$buf$0>>2]|0;
  $23 = (($18) + 1)|0;
  $24 = $23 & 15;
  $25 = (($21) + ($24<<2)|0);
  $26 = (($22) + ($18<<2)|0);
  _INT123_dct64($25,$26,$bandPtr);
  $b0$0 = $22;$bo1$0 = $18;
 }
 $32 = ((($fr)) + 9284|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($33)) + 64|0);
 $35 = (0 - ($bo1$0))|0;
 $36 = (($34) + ($35<<2)|0);
 $scevgep = ((($samples$0$in)) + 32|0);
 $37 = (528 - ($bo1$0))|0;
 $scevgep62 = (($33) + ($37<<2)|0);
 $38 = (400 - ($bo1$0))|0;
 $scevgep64 = (($33) + ($38<<2)|0);
 $b0$157 = $b0$0;$clip$056 = 0;$j$055 = 4;$samples$158 = $samples$0$in;$window$054 = $36;
 while(1) {
  $39 = ((($window$054)) + 4|0);
  $40 = +HEAPF32[$window$054>>2];
  $41 = ((($b0$157)) + 4|0);
  $42 = +HEAPF32[$b0$157>>2];
  $43 = $42 * $40;
  $44 = ((($window$054)) + 8|0);
  $45 = +HEAPF32[$39>>2];
  $46 = ((($b0$157)) + 8|0);
  $47 = +HEAPF32[$41>>2];
  $$neg = $47 * $45;
  $48 = ((($window$054)) + 12|0);
  $49 = +HEAPF32[$44>>2];
  $50 = ((($b0$157)) + 12|0);
  $51 = +HEAPF32[$46>>2];
  $52 = $51 * $49;
  $53 = ((($window$054)) + 16|0);
  $54 = +HEAPF32[$48>>2];
  $55 = ((($b0$157)) + 16|0);
  $56 = +HEAPF32[$50>>2];
  $$neg1 = $56 * $54;
  $57 = ((($window$054)) + 20|0);
  $58 = +HEAPF32[$53>>2];
  $59 = ((($b0$157)) + 20|0);
  $60 = +HEAPF32[$55>>2];
  $61 = $60 * $58;
  $62 = ((($window$054)) + 24|0);
  $63 = +HEAPF32[$57>>2];
  $64 = ((($b0$157)) + 24|0);
  $65 = +HEAPF32[$59>>2];
  $$neg2 = $65 * $63;
  $66 = ((($window$054)) + 28|0);
  $67 = +HEAPF32[$62>>2];
  $68 = ((($b0$157)) + 28|0);
  $69 = +HEAPF32[$64>>2];
  $70 = $69 * $67;
  $71 = ((($window$054)) + 32|0);
  $72 = +HEAPF32[$66>>2];
  $73 = ((($b0$157)) + 32|0);
  $74 = +HEAPF32[$68>>2];
  $$neg3 = $74 * $72;
  $75 = ((($window$054)) + 36|0);
  $76 = +HEAPF32[$71>>2];
  $77 = ((($b0$157)) + 36|0);
  $78 = +HEAPF32[$73>>2];
  $79 = $78 * $76;
  $80 = ((($window$054)) + 40|0);
  $81 = +HEAPF32[$75>>2];
  $82 = ((($b0$157)) + 40|0);
  $83 = +HEAPF32[$77>>2];
  $$neg4 = $83 * $81;
  $84 = ((($window$054)) + 44|0);
  $85 = +HEAPF32[$80>>2];
  $86 = ((($b0$157)) + 44|0);
  $87 = +HEAPF32[$82>>2];
  $88 = $87 * $85;
  $89 = ((($window$054)) + 48|0);
  $90 = +HEAPF32[$84>>2];
  $91 = ((($b0$157)) + 48|0);
  $92 = +HEAPF32[$86>>2];
  $$neg5 = $92 * $90;
  $93 = ((($window$054)) + 52|0);
  $94 = +HEAPF32[$89>>2];
  $95 = ((($b0$157)) + 52|0);
  $96 = +HEAPF32[$91>>2];
  $97 = $96 * $94;
  $98 = ((($window$054)) + 56|0);
  $99 = +HEAPF32[$93>>2];
  $100 = ((($b0$157)) + 56|0);
  $101 = +HEAPF32[$95>>2];
  $$neg6 = $101 * $99;
  $102 = ((($window$054)) + 60|0);
  $103 = +HEAPF32[$98>>2];
  $104 = ((($b0$157)) + 60|0);
  $105 = +HEAPF32[$100>>2];
  $106 = $105 * $103;
  $107 = +HEAPF32[$102>>2];
  $108 = +HEAPF32[$104>>2];
  $$neg7 = $108 * $107;
  $tmp = $$neg7 + $$neg6;
  $tmp9 = $$neg5 + $$neg4;
  $tmp11 = $$neg3 + $$neg2;
  $tmp13 = $$neg1 + $$neg;
  $109 = $52 + $43;
  $110 = $109 + $61;
  $111 = $110 + $70;
  $112 = $111 + $79;
  $113 = $112 + $88;
  $114 = $113 + $97;
  $115 = $114 + $106;
  $116 = $115 - $tmp13;
  $117 = $116 - $tmp11;
  $118 = $117 - $tmp9;
  $119 = $118 - $tmp;
  $120 = $119 * 65536.0;
  $121 = $120;
  $122 = $121 > 2147483647.0;
  do {
   if ($122) {
    HEAP32[$samples$158>>2] = 2147483647;
    $123 = (($clip$056) + 1)|0;
    $clip$1 = $123;
   } else {
    $124 = $120 < -2147483648.0;
    if ($124) {
     HEAP32[$samples$158>>2] = -2147483648;
     $125 = (($clip$056) + 1)|0;
     $clip$1 = $125;
     break;
    } else {
     $126 = (~~(($120)));
     HEAP32[$samples$158>>2] = $126;
     $clip$1 = $clip$056;
     break;
    }
   }
  } while(0);
  $127 = (($j$055) + -1)|0;
  $128 = ((($b0$157)) + 256|0);
  $129 = ((($window$054)) + 512|0);
  $130 = ((($samples$158)) + 8|0);
  $131 = ($127|0)==(0);
  if ($131) {
   $clip$1$lcssa = $clip$1;
   break;
  } else {
   $b0$157 = $128;$clip$056 = $clip$1;$j$055 = $127;$samples$158 = $130;$window$054 = $129;
  }
 }
 $scevgep63 = ((($b0$0)) + 1024|0);
 $scevgep65 = ((($b0$0)) + 768|0);
 $132 = +HEAPF32[$scevgep62>>2];
 $133 = +HEAPF32[$scevgep63>>2];
 $134 = $133 * $132;
 $135 = ((($scevgep64)) + 520|0);
 $136 = +HEAPF32[$135>>2];
 $137 = ((($b0$0)) + 1032|0);
 $138 = +HEAPF32[$137>>2];
 $139 = $138 * $136;
 $140 = $139 + $134;
 $141 = ((($scevgep64)) + 528|0);
 $142 = +HEAPF32[$141>>2];
 $143 = ((($b0$0)) + 1040|0);
 $144 = +HEAPF32[$143>>2];
 $145 = $144 * $142;
 $146 = $140 + $145;
 $147 = ((($scevgep64)) + 536|0);
 $148 = +HEAPF32[$147>>2];
 $149 = ((($b0$0)) + 1048|0);
 $150 = +HEAPF32[$149>>2];
 $151 = $150 * $148;
 $152 = $146 + $151;
 $153 = ((($scevgep64)) + 544|0);
 $154 = +HEAPF32[$153>>2];
 $155 = ((($b0$0)) + 1056|0);
 $156 = +HEAPF32[$155>>2];
 $157 = $156 * $154;
 $158 = $152 + $157;
 $159 = ((($scevgep64)) + 552|0);
 $160 = +HEAPF32[$159>>2];
 $161 = ((($b0$0)) + 1064|0);
 $162 = +HEAPF32[$161>>2];
 $163 = $162 * $160;
 $164 = $158 + $163;
 $165 = ((($scevgep64)) + 560|0);
 $166 = +HEAPF32[$165>>2];
 $167 = ((($b0$0)) + 1072|0);
 $168 = +HEAPF32[$167>>2];
 $169 = $168 * $166;
 $170 = $164 + $169;
 $171 = ((($scevgep64)) + 568|0);
 $172 = +HEAPF32[$171>>2];
 $173 = ((($b0$0)) + 1080|0);
 $174 = +HEAPF32[$173>>2];
 $175 = $174 * $172;
 $176 = $170 + $175;
 $177 = $176 * 65536.0;
 $178 = $177;
 $179 = $178 > 2147483647.0;
 do {
  if ($179) {
   HEAP32[$scevgep>>2] = 2147483647;
   $180 = (($clip$1$lcssa) + 1)|0;
   $clip$2 = $180;
  } else {
   $181 = $177 < -2147483648.0;
   if ($181) {
    HEAP32[$scevgep>>2] = -2147483648;
    $182 = (($clip$1$lcssa) + 1)|0;
    $clip$2 = $182;
    break;
   } else {
    $183 = (~~(($177)));
    HEAP32[$scevgep>>2] = $183;
    $clip$2 = $clip$1$lcssa;
    break;
   }
  }
 } while(0);
 $184 = $bo1$0 << 1;
 $185 = (($scevgep64) + ($184<<2)|0);
 $samples$247 = ((($samples$0$in)) + 40|0);
 $b0$1$pn51 = $scevgep63;$b0$252 = $scevgep65;$clip$350 = $clip$2;$j$149 = 3;$samples$253 = $samples$247;$window$148 = $185;
 while(1) {
  $186 = ((($window$148)) + -4|0);
  $187 = +HEAPF32[$186>>2];
  $188 = ((($b0$1$pn51)) + -252|0);
  $189 = +HEAPF32[$b0$252>>2];
  $190 = $189 * $187;
  $191 = ((($window$148)) + -8|0);
  $192 = +HEAPF32[$191>>2];
  $193 = ((($b0$1$pn51)) + -248|0);
  $194 = +HEAPF32[$188>>2];
  $$neg15 = $194 * $192;
  $195 = ((($window$148)) + -12|0);
  $196 = +HEAPF32[$195>>2];
  $197 = ((($b0$1$pn51)) + -244|0);
  $198 = +HEAPF32[$193>>2];
  $$neg16 = $198 * $196;
  $199 = ((($window$148)) + -16|0);
  $200 = +HEAPF32[$199>>2];
  $201 = ((($b0$1$pn51)) + -240|0);
  $202 = +HEAPF32[$197>>2];
  $$neg17 = $202 * $200;
  $203 = ((($window$148)) + -20|0);
  $204 = +HEAPF32[$203>>2];
  $205 = ((($b0$1$pn51)) + -236|0);
  $206 = +HEAPF32[$201>>2];
  $$neg18 = $206 * $204;
  $207 = ((($window$148)) + -24|0);
  $208 = +HEAPF32[$207>>2];
  $209 = ((($b0$1$pn51)) + -232|0);
  $210 = +HEAPF32[$205>>2];
  $$neg19 = $210 * $208;
  $211 = ((($window$148)) + -28|0);
  $212 = +HEAPF32[$211>>2];
  $213 = ((($b0$1$pn51)) + -228|0);
  $214 = +HEAPF32[$209>>2];
  $$neg20 = $214 * $212;
  $215 = ((($window$148)) + -32|0);
  $216 = +HEAPF32[$215>>2];
  $217 = ((($b0$1$pn51)) + -224|0);
  $218 = +HEAPF32[$213>>2];
  $$neg21 = $218 * $216;
  $219 = ((($window$148)) + -36|0);
  $220 = +HEAPF32[$219>>2];
  $221 = ((($b0$1$pn51)) + -220|0);
  $222 = +HEAPF32[$217>>2];
  $$neg22 = $222 * $220;
  $223 = ((($window$148)) + -40|0);
  $224 = +HEAPF32[$223>>2];
  $225 = ((($b0$1$pn51)) + -216|0);
  $226 = +HEAPF32[$221>>2];
  $$neg23 = $226 * $224;
  $227 = ((($window$148)) + -44|0);
  $228 = +HEAPF32[$227>>2];
  $229 = ((($b0$1$pn51)) + -212|0);
  $230 = +HEAPF32[$225>>2];
  $$neg24 = $230 * $228;
  $231 = ((($window$148)) + -48|0);
  $232 = +HEAPF32[$231>>2];
  $233 = ((($b0$1$pn51)) + -208|0);
  $234 = +HEAPF32[$229>>2];
  $$neg25 = $234 * $232;
  $235 = ((($window$148)) + -52|0);
  $236 = +HEAPF32[$235>>2];
  $237 = ((($b0$1$pn51)) + -204|0);
  $238 = +HEAPF32[$233>>2];
  $$neg26 = $238 * $236;
  $239 = ((($window$148)) + -56|0);
  $240 = +HEAPF32[$239>>2];
  $241 = ((($b0$1$pn51)) + -200|0);
  $242 = +HEAPF32[$237>>2];
  $$neg27 = $242 * $240;
  $243 = ((($window$148)) + -60|0);
  $244 = +HEAPF32[$243>>2];
  $245 = ((($b0$1$pn51)) + -196|0);
  $246 = +HEAPF32[$241>>2];
  $$neg28 = $246 * $244;
  $247 = ((($window$148)) + -64|0);
  $248 = +HEAPF32[$247>>2];
  $249 = +HEAPF32[$245>>2];
  $$neg29 = $249 * $248;
  $tmp30 = $$neg29 + $190;
  $tmp32 = $$neg28 + $$neg27;
  $tmp34 = $$neg26 + $$neg25;
  $tmp36 = $$neg24 + $$neg23;
  $tmp38 = $$neg22 + $$neg21;
  $tmp40 = $$neg20 + $$neg19;
  $tmp42 = $$neg18 + $$neg17;
  $tmp44 = $$neg16 + $$neg15;
  $tmp45 = -$tmp44;
  $250 = $tmp45 - $tmp42;
  $251 = $250 - $tmp40;
  $252 = $251 - $tmp38;
  $253 = $252 - $tmp36;
  $254 = $253 - $tmp34;
  $255 = $254 - $tmp32;
  $256 = $255 - $tmp30;
  $257 = $256 * 65536.0;
  $258 = $257;
  $259 = $258 > 2147483647.0;
  do {
   if ($259) {
    HEAP32[$samples$253>>2] = 2147483647;
    $260 = (($clip$350) + 1)|0;
    $clip$4 = $260;
   } else {
    $261 = $257 < -2147483648.0;
    if ($261) {
     HEAP32[$samples$253>>2] = -2147483648;
     $262 = (($clip$350) + 1)|0;
     $clip$4 = $262;
     break;
    } else {
     $263 = (~~(($257)));
     HEAP32[$samples$253>>2] = $263;
     $clip$4 = $clip$350;
     break;
    }
   }
  } while(0);
  $264 = (($j$149) + -1)|0;
  $265 = ((($window$148)) + -512|0);
  $b0$2 = ((($b0$252)) + -256|0);
  $samples$2 = ((($samples$253)) + 8|0);
  $266 = ($264|0)==(0);
  if ($266) {
   $clip$4$lcssa = $clip$4;
   break;
  } else {
   $b0$1$pn51$phi = $b0$252;$b0$252 = $b0$2;$clip$350 = $clip$4;$j$149 = $264;$samples$253 = $samples$2;$window$148 = $265;$b0$1$pn51 = $b0$1$pn51$phi;
  }
 }
 $267 = ($final|0)==(0);
 if ($267) {
  return ($clip$4$lcssa|0);
 }
 $268 = HEAP32[$2>>2]|0;
 $269 = (($268) + 64)|0;
 HEAP32[$2>>2] = $269;
 return ($clip$4$lcssa|0);
}
function _INT123_synth_ntom($bandPtr,$channel,$fr,$final) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $fr = $fr|0;
 $final = $final|0;
 var $$lcssa = 0, $$neg = 0.0, $$neg16 = 0.0, $$neg17 = 0.0, $$neg18 = 0.0, $$neg19 = 0.0, $$neg2 = 0.0, $$neg20 = 0.0, $$neg21 = 0.0, $$neg22 = 0.0, $$neg23 = 0.0, $$neg24 = 0.0, $$neg25 = 0.0, $$neg26 = 0.0, $$neg27 = 0.0, $$neg28 = 0.0, $$neg29 = 0.0, $$neg3 = 0.0, $$neg30 = 0.0, $$neg4 = 0.0;
 var $$neg48 = 0, $$neg5 = 0.0, $$neg6 = 0.0, $$neg7 = 0.0, $$neg8 = 0.0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0, $102 = 0.0, $103 = 0, $104 = 0.0, $105 = 0, $106 = 0.0, $107 = 0.0, $108 = 0, $109 = 0.0;
 var $11 = 0, $110 = 0, $111 = 0.0, $112 = 0, $113 = 0.0, $114 = 0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0;
 var $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0;
 var $164 = 0.0, $165 = 0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0, $17 = 0, $170 = 0.0, $171 = 0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0, $176 = 0.0, $177 = 0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0;
 var $182 = 0.0, $183 = 0, $184 = 0.0, $185 = 0.0, $186 = 0.0, $187 = 0, $188 = 0.0, $189 = 0, $19 = 0, $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0.0, $198 = 0.0, $199 = 0, $2 = 0;
 var $20 = 0, $200 = 0.0, $201 = 0, $202 = 0.0, $203 = 0.0, $204 = 0.0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0;
 var $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0;
 var $236 = 0, $237 = 0.0, $238 = 0, $239 = 0.0, $24 = 0, $240 = 0.0, $241 = 0, $242 = 0.0, $243 = 0, $244 = 0.0, $245 = 0, $246 = 0.0, $247 = 0, $248 = 0.0, $249 = 0, $25 = 0, $250 = 0.0, $251 = 0, $252 = 0.0, $253 = 0;
 var $254 = 0.0, $255 = 0, $256 = 0.0, $257 = 0, $258 = 0.0, $259 = 0, $26 = 0, $260 = 0.0, $261 = 0, $262 = 0.0, $263 = 0, $264 = 0.0, $265 = 0, $266 = 0.0, $267 = 0, $268 = 0.0, $269 = 0, $27 = 0, $270 = 0.0, $271 = 0;
 var $272 = 0.0, $273 = 0, $274 = 0.0, $275 = 0, $276 = 0.0, $277 = 0, $278 = 0.0, $279 = 0, $28 = 0, $280 = 0.0, $281 = 0, $282 = 0.0, $283 = 0, $284 = 0.0, $285 = 0, $286 = 0.0, $287 = 0, $288 = 0.0, $289 = 0, $29 = 0;
 var $290 = 0.0, $291 = 0, $292 = 0.0, $293 = 0, $294 = 0.0, $295 = 0, $296 = 0.0, $297 = 0, $298 = 0.0, $299 = 0.0, $3 = 0, $30 = 0, $300 = 0.0, $301 = 0.0, $302 = 0.0, $303 = 0.0, $304 = 0.0, $305 = 0.0, $306 = 0.0, $307 = 0;
 var $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0;
 var $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0, $55 = 0.0;
 var $56 = 0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0.0, $63 = 0, $64 = 0.0, $65 = 0, $66 = 0.0, $67 = 0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0.0, $71 = 0.0, $72 = 0, $73 = 0.0;
 var $74 = 0, $75 = 0.0, $76 = 0, $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0, $86 = 0.0, $87 = 0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0, $91 = 0.0;
 var $92 = 0, $93 = 0.0, $94 = 0, $95 = 0.0, $96 = 0, $97 = 0.0, $98 = 0.0, $99 = 0, $b0$0 = 0, $b0$1$pn59 = 0, $b0$1$pn59$phi = 0, $b0$180 = 0, $b0$2 = 0, $b0$3 = 0, $b0$364 = 0, $bo1$0 = 0, $buf$0 = 0, $clip$081 = 0, $clip$10 = 0, $clip$10$lcssa = 0;
 var $clip$3 = 0, $clip$3$lcssa = 0, $clip$466$us = 0, $clip$466$us$us = 0, $clip$466$us89 = 0, $clip$6 = 0, $clip$760 = 0, $exitcond = 0, $exitcond148 = 0, $exitcond150 = 0, $j$082 = 0, $j$162 = 0, $ntom$0 = 0, $ntom$274 = 0, $ntom$274$us = 0, $ntom$274$us$us = 0, $ntom$274$us92 = 0, $ntom$3 = 0, $ntom$467 = 0, $ntom$5 = 0;
 var $ntom$661 = 0, $ntom$754 = 0, $ntom$754$us = 0, $ntom$754$us$us = 0, $ntom$754$us84 = 0, $ntom$8 = 0, $ntom$8$lcssa = 0, $samples$0$in = 0, $samples$179 = 0, $samples$272 = 0, $samples$272$us = 0, $samples$272$us$us = 0, $samples$272$us94 = 0, $samples$3 = 0, $samples$3$lcssa = 0, $samples$465 = 0, $samples$465$us = 0, $samples$465$us$us = 0, $samples$465$us90 = 0, $samples$5 = 0;
 var $samples$658 = 0, $samples$752 = 0, $samples$752$us = 0, $samples$752$us$us = 0, $samples$752$us86 = 0, $samples$8 = 0, $samples$8$lcssa = 0, $scevgep = 0, $scevgep139 = 0, $scevgep140 = 0, $scevgep141 = 0, $scevgep142 = 0, $scevgep143 = 0, $scevgep144 = 0, $scevgep145 = 0, $scevgep146 = 0, $scevgep147 = 0, $scevgep149 = 0, $scevgep151 = 0, $scevgep152 = 0;
 var $scevgep153 = 0, $scevgep154 = 0, $scevgep155 = 0, $scevgep156 = 0, $scevgep157 = 0, $scevgep158 = 0, $scevgep159 = 0, $scevgep160 = 0, $scevgep161 = 0, $scevgep162 = 0, $scevgep163 = 0, $tmp = 0.0, $tmp10 = 0.0, $tmp12 = 0.0, $tmp14 = 0.0, $tmp31 = 0.0, $tmp33 = 0.0, $tmp35 = 0.0, $tmp37 = 0.0, $tmp39 = 0.0;
 var $tmp41 = 0.0, $tmp43 = 0.0, $tmp45 = 0.0, $tmp46 = 0.0, $window$083 = 0, $window$263 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($fr)) + 9288|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = ((($fr)) + 9292|0);
  _INT123_do_equalizer($bandPtr,$channel,$8);
 }
 $9 = ($channel|0)!=(0);
 if ($9) {
  $18 = ((($4)) + 2|0);
  $19 = ((($fr)) + 9256|0);
  $20 = ((($fr)) + 19356|0);
  $21 = HEAP32[$20>>2]|0;
  $$phi$trans$insert = ((($fr)) + 9272|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $23 = $$pre;$buf$0 = $19;$ntom$0 = $21;$samples$0$in = $18;
 } else {
  $10 = ((($fr)) + 9272|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + 15)|0;
  $13 = $12 & 15;
  HEAP32[$10>>2] = $13;
  $14 = ((($fr)) + 9248|0);
  $15 = ((($fr)) + 19352|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ((($fr)) + 19356|0);
  HEAP32[$17>>2] = $16;
  $23 = $13;$buf$0 = $14;$ntom$0 = $16;$samples$0$in = $4;
 }
 $22 = $23 & 1;
 $24 = ($22|0)==(0);
 $25 = ((($buf$0)) + 4|0);
 $26 = HEAP32[$25>>2]|0;
 if ($24) {
  $32 = (($23) + 1)|0;
  $33 = HEAP32[$buf$0>>2]|0;
  $34 = (($33) + ($23<<2)|0);
  $35 = (($26) + ($23<<2)|0);
  $36 = ((($35)) + 4|0);
  _INT123_dct64($34,$36,$bandPtr);
  $b0$0 = $26;$bo1$0 = $32;
 } else {
  $27 = HEAP32[$buf$0>>2]|0;
  $28 = (($23) + 1)|0;
  $29 = $28 & 15;
  $30 = (($26) + ($29<<2)|0);
  $31 = (($27) + ($23<<2)|0);
  _INT123_dct64($30,$31,$bandPtr);
  $b0$0 = $27;$bo1$0 = $23;
 }
 $37 = ((($fr)) + 9284|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = ((($38)) + 64|0);
 $40 = (0 - ($bo1$0))|0;
 $41 = (($39) + ($40<<2)|0);
 $42 = ((($fr)) + 19360|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = (($43) + ($ntom$0))|0;
 $45 = (528 - ($bo1$0))|0;
 $scevgep161 = ((($b0$0)) + 1024|0);
 $46 = (496 - ($bo1$0))|0;
 $48 = $44;$b0$180 = $b0$0;$clip$081 = 0;$j$082 = 16;$samples$179 = $samples$0$in;$window$083 = $41;
 while(1) {
  $47 = ($48|0)<(32768);
  do {
   if ($47) {
    $clip$3 = $clip$081;$ntom$3 = $48;$samples$3 = $samples$179;
   } else {
    $49 = ((($window$083)) + 4|0);
    $50 = +HEAPF32[$window$083>>2];
    $51 = ((($b0$180)) + 4|0);
    $52 = +HEAPF32[$b0$180>>2];
    $53 = $52 * $50;
    $54 = ((($window$083)) + 8|0);
    $55 = +HEAPF32[$49>>2];
    $56 = ((($b0$180)) + 8|0);
    $57 = +HEAPF32[$51>>2];
    $$neg = $57 * $55;
    $58 = ((($window$083)) + 12|0);
    $59 = +HEAPF32[$54>>2];
    $60 = ((($b0$180)) + 12|0);
    $61 = +HEAPF32[$56>>2];
    $62 = $61 * $59;
    $63 = ((($window$083)) + 16|0);
    $64 = +HEAPF32[$58>>2];
    $65 = ((($b0$180)) + 16|0);
    $66 = +HEAPF32[$60>>2];
    $$neg2 = $66 * $64;
    $67 = ((($window$083)) + 20|0);
    $68 = +HEAPF32[$63>>2];
    $69 = ((($b0$180)) + 20|0);
    $70 = +HEAPF32[$65>>2];
    $71 = $70 * $68;
    $72 = ((($window$083)) + 24|0);
    $73 = +HEAPF32[$67>>2];
    $74 = ((($b0$180)) + 24|0);
    $75 = +HEAPF32[$69>>2];
    $$neg3 = $75 * $73;
    $76 = ((($window$083)) + 28|0);
    $77 = +HEAPF32[$72>>2];
    $78 = ((($b0$180)) + 28|0);
    $79 = +HEAPF32[$74>>2];
    $80 = $79 * $77;
    $81 = ((($window$083)) + 32|0);
    $82 = +HEAPF32[$76>>2];
    $83 = ((($b0$180)) + 32|0);
    $84 = +HEAPF32[$78>>2];
    $$neg4 = $84 * $82;
    $85 = ((($window$083)) + 36|0);
    $86 = +HEAPF32[$81>>2];
    $87 = ((($b0$180)) + 36|0);
    $88 = +HEAPF32[$83>>2];
    $89 = $88 * $86;
    $90 = ((($window$083)) + 40|0);
    $91 = +HEAPF32[$85>>2];
    $92 = ((($b0$180)) + 40|0);
    $93 = +HEAPF32[$87>>2];
    $$neg5 = $93 * $91;
    $94 = ((($window$083)) + 44|0);
    $95 = +HEAPF32[$90>>2];
    $96 = ((($b0$180)) + 44|0);
    $97 = +HEAPF32[$92>>2];
    $98 = $97 * $95;
    $99 = ((($window$083)) + 48|0);
    $100 = +HEAPF32[$94>>2];
    $101 = ((($b0$180)) + 48|0);
    $102 = +HEAPF32[$96>>2];
    $$neg6 = $102 * $100;
    $103 = ((($window$083)) + 52|0);
    $104 = +HEAPF32[$99>>2];
    $105 = ((($b0$180)) + 52|0);
    $106 = +HEAPF32[$101>>2];
    $107 = $106 * $104;
    $108 = ((($window$083)) + 56|0);
    $109 = +HEAPF32[$103>>2];
    $110 = ((($b0$180)) + 56|0);
    $111 = +HEAPF32[$105>>2];
    $$neg7 = $111 * $109;
    $112 = ((($window$083)) + 60|0);
    $113 = +HEAPF32[$108>>2];
    $114 = ((($b0$180)) + 60|0);
    $115 = +HEAPF32[$110>>2];
    $116 = $115 * $113;
    $117 = +HEAPF32[$112>>2];
    $118 = +HEAPF32[$114>>2];
    $$neg8 = $118 * $117;
    $tmp = $$neg8 + $$neg7;
    $tmp10 = $$neg6 + $$neg5;
    $tmp12 = $$neg4 + $$neg3;
    $tmp14 = $$neg2 + $$neg;
    $119 = $62 + $53;
    $120 = $119 + $71;
    $121 = $120 + $80;
    $122 = $121 + $89;
    $123 = $122 + $98;
    $124 = $123 + $107;
    $125 = $124 + $116;
    $126 = $125 - $tmp14;
    $127 = $126 - $tmp12;
    $128 = $127 - $tmp10;
    $129 = $128 - $tmp;
    $130 = $129 > 32767.0;
    $131 = $129 < -32768.0;
    $132 = (~~(($129)));
    $133 = (($48) + -32768)|0;
    $134 = $133 >>> 15;
    $135 = $134 << 15;
    $136 = (($133) - ($135))|0;
    $137 = $134 << 1;
    if ($130) {
     $138 = (($clip$081) + 1)|0;
     if ($131) {
      $ntom$274$us$us = $48;$samples$272$us$us = $samples$179;
      while(1) {
       HEAP16[$samples$272$us$us>>1] = 32767;
       $139 = ((($samples$272$us$us)) + 4|0);
       $140 = (($ntom$274$us$us) + -32768)|0;
       $141 = ($140|0)>(32767);
       if ($141) {
        $ntom$274$us$us = $140;$samples$272$us$us = $139;
       } else {
        break;
       }
      }
      $scevgep158 = ((($samples$179)) + 4|0);
      $scevgep159 = (($scevgep158) + ($137<<1)|0);
      $152 = (($138) + ($134))|0;
      $clip$3 = $152;$ntom$3 = $136;$samples$3 = $scevgep159;
      break;
     } else {
      $ntom$274$us = $48;$samples$272$us = $samples$179;
      while(1) {
       HEAP16[$samples$272$us>>1] = 32767;
       $142 = ((($samples$272$us)) + 4|0);
       $143 = (($ntom$274$us) + -32768)|0;
       $144 = ($143|0)>(32767);
       if ($144) {
        $ntom$274$us = $143;$samples$272$us = $142;
       } else {
        break;
       }
      }
      $scevgep156 = ((($samples$179)) + 4|0);
      $scevgep157 = (($scevgep156) + ($137<<1)|0);
      $153 = (($138) + ($134))|0;
      $clip$3 = $153;$ntom$3 = $136;$samples$3 = $scevgep157;
      break;
     }
    }
    if (!($131)) {
     $ntom$274 = $48;$samples$272 = $samples$179;
     while(1) {
      HEAP16[$samples$272>>1] = $132;
      $149 = ((($samples$272)) + 4|0);
      $150 = (($ntom$274) + -32768)|0;
      $151 = ($150|0)>(32767);
      if ($151) {
       $ntom$274 = $150;$samples$272 = $149;
      } else {
       break;
      }
     }
     $scevgep152 = ((($samples$179)) + 4|0);
     $scevgep153 = (($scevgep152) + ($137<<1)|0);
     $clip$3 = $clip$081;$ntom$3 = $136;$samples$3 = $scevgep153;
     break;
    }
    $145 = (($clip$081) + 1)|0;
    $ntom$274$us92 = $48;$samples$272$us94 = $samples$179;
    while(1) {
     HEAP16[$samples$272$us94>>1] = -32768;
     $146 = ((($samples$272$us94)) + 4|0);
     $147 = (($ntom$274$us92) + -32768)|0;
     $148 = ($147|0)>(32767);
     if ($148) {
      $ntom$274$us92 = $147;$samples$272$us94 = $146;
     } else {
      break;
     }
    }
    $scevgep154 = ((($samples$179)) + 4|0);
    $scevgep155 = (($scevgep154) + ($137<<1)|0);
    $154 = (($145) + ($134))|0;
    $clip$3 = $154;$ntom$3 = $136;$samples$3 = $scevgep155;
   }
  } while(0);
  $b0$2 = ((($b0$180)) + 64|0);
  $155 = (($j$082) + -1)|0;
  $156 = ((($window$083)) + 128|0);
  $157 = ($155|0)==(0);
  $158 = (($43) + ($ntom$3))|0;
  if ($157) {
   $$lcssa = $158;$clip$3$lcssa = $clip$3;$samples$3$lcssa = $samples$3;
   break;
  } else {
   $48 = $158;$b0$180 = $b0$2;$clip$081 = $clip$3;$j$082 = $155;$samples$179 = $samples$3;$window$083 = $156;
  }
 }
 $scevgep162 = ((($b0$0)) + 960|0);
 $scevgep163 = (($38) + ($46<<2)|0);
 $159 = ($$lcssa|0)>(32767);
 do {
  if ($159) {
   $scevgep160 = (($38) + ($45<<2)|0);
   $160 = +HEAPF32[$scevgep160>>2];
   $161 = +HEAPF32[$scevgep161>>2];
   $162 = $161 * $160;
   $163 = ((($scevgep163)) + 136|0);
   $164 = +HEAPF32[$163>>2];
   $165 = ((($b0$0)) + 1032|0);
   $166 = +HEAPF32[$165>>2];
   $167 = $166 * $164;
   $168 = $167 + $162;
   $169 = ((($scevgep163)) + 144|0);
   $170 = +HEAPF32[$169>>2];
   $171 = ((($b0$0)) + 1040|0);
   $172 = +HEAPF32[$171>>2];
   $173 = $172 * $170;
   $174 = $168 + $173;
   $175 = ((($scevgep163)) + 152|0);
   $176 = +HEAPF32[$175>>2];
   $177 = ((($b0$0)) + 1048|0);
   $178 = +HEAPF32[$177>>2];
   $179 = $178 * $176;
   $180 = $174 + $179;
   $181 = ((($scevgep163)) + 160|0);
   $182 = +HEAPF32[$181>>2];
   $183 = ((($b0$0)) + 1056|0);
   $184 = +HEAPF32[$183>>2];
   $185 = $184 * $182;
   $186 = $180 + $185;
   $187 = ((($scevgep163)) + 168|0);
   $188 = +HEAPF32[$187>>2];
   $189 = ((($b0$0)) + 1064|0);
   $190 = +HEAPF32[$189>>2];
   $191 = $190 * $188;
   $192 = $186 + $191;
   $193 = ((($scevgep163)) + 176|0);
   $194 = +HEAPF32[$193>>2];
   $195 = ((($b0$0)) + 1072|0);
   $196 = +HEAPF32[$195>>2];
   $197 = $196 * $194;
   $198 = $192 + $197;
   $199 = ((($scevgep163)) + 184|0);
   $200 = +HEAPF32[$199>>2];
   $201 = ((($b0$0)) + 1080|0);
   $202 = +HEAPF32[$201>>2];
   $203 = $202 * $200;
   $204 = $198 + $203;
   $205 = $204 > 32767.0;
   $206 = $204 < -32768.0;
   $207 = (~~(($204)));
   $208 = (($$lcssa) + -32768)|0;
   $209 = $208 >>> 15;
   $210 = $209 << 15;
   $211 = $209 << 1;
   $212 = (($211) + 2)|0;
   if ($205) {
    $213 = (($clip$3$lcssa) + ($209))|0;
    $214 = (($213) + 1)|0;
    if ($206) {
     $clip$466$us$us = $clip$3$lcssa;$samples$465$us$us = $samples$3$lcssa;
     while(1) {
      HEAP16[$samples$465$us$us>>1] = 32767;
      $215 = (($clip$466$us$us) + 1)|0;
      $216 = ((($samples$465$us$us)) + 4|0);
      $exitcond = ($215|0)==($214|0);
      if ($exitcond) {
       break;
      } else {
       $clip$466$us$us = $215;$samples$465$us$us = $216;
      }
     }
     $226 = (($208) - ($210))|0;
     $scevgep146 = (($samples$3$lcssa) + ($212<<1)|0);
     $clip$6 = $214;$ntom$5 = $226;$samples$5 = $scevgep146;
     break;
    } else {
     $clip$466$us = $clip$3$lcssa;$samples$465$us = $samples$3$lcssa;
     while(1) {
      HEAP16[$samples$465$us>>1] = 32767;
      $217 = (($clip$466$us) + 1)|0;
      $218 = ((($samples$465$us)) + 4|0);
      $exitcond148 = ($217|0)==($214|0);
      if ($exitcond148) {
       break;
      } else {
       $clip$466$us = $217;$samples$465$us = $218;
      }
     }
     $227 = (($208) - ($210))|0;
     $scevgep147 = (($samples$3$lcssa) + ($212<<1)|0);
     $clip$6 = $214;$ntom$5 = $227;$samples$5 = $scevgep147;
     break;
    }
   }
   if (!($206)) {
    $ntom$467 = $$lcssa;$samples$465 = $samples$3$lcssa;
    while(1) {
     HEAP16[$samples$465>>1] = $207;
     $223 = ((($samples$465)) + 4|0);
     $224 = (($ntom$467) + -32768)|0;
     $225 = ($224|0)>(32767);
     if ($225) {
      $ntom$467 = $224;$samples$465 = $223;
     } else {
      break;
     }
    }
    $229 = (($208) - ($210))|0;
    $scevgep151 = (($samples$3$lcssa) + ($212<<1)|0);
    $clip$6 = $clip$3$lcssa;$ntom$5 = $229;$samples$5 = $scevgep151;
    break;
   }
   $219 = (($clip$3$lcssa) + ($209))|0;
   $220 = (($219) + 1)|0;
   $clip$466$us89 = $clip$3$lcssa;$samples$465$us90 = $samples$3$lcssa;
   while(1) {
    HEAP16[$samples$465$us90>>1] = -32768;
    $221 = (($clip$466$us89) + 1)|0;
    $222 = ((($samples$465$us90)) + 4|0);
    $exitcond150 = ($221|0)==($220|0);
    if ($exitcond150) {
     break;
    } else {
     $clip$466$us89 = $221;$samples$465$us90 = $222;
    }
   }
   $228 = (($208) - ($210))|0;
   $scevgep149 = (($samples$3$lcssa) + ($212<<1)|0);
   $clip$6 = $220;$ntom$5 = $228;$samples$5 = $scevgep149;
  } else {
   $clip$6 = $clip$3$lcssa;$ntom$5 = $$lcssa;$samples$5 = $samples$3$lcssa;
  }
 } while(0);
 $230 = $bo1$0 << 1;
 $231 = (($scevgep163) + ($230<<2)|0);
 $232 = HEAP32[$42>>2]|0;
 $233 = (($232) + -32768)|0;
 $b0$1$pn59 = $scevgep161;$b0$364 = $scevgep162;$clip$760 = $clip$6;$j$162 = 15;$ntom$661 = $ntom$5;$samples$658 = $samples$5;$window$263 = $231;
 while(1) {
  $234 = (($232) + ($ntom$661))|0;
  $235 = ($234|0)<(32768);
  do {
   if ($235) {
    $clip$10 = $clip$760;$ntom$8 = $234;$samples$8 = $samples$658;
   } else {
    $236 = ((($window$263)) + -4|0);
    $237 = +HEAPF32[$236>>2];
    $238 = ((($b0$1$pn59)) + -60|0);
    $239 = +HEAPF32[$b0$364>>2];
    $240 = $239 * $237;
    $241 = ((($window$263)) + -8|0);
    $242 = +HEAPF32[$241>>2];
    $243 = ((($b0$1$pn59)) + -56|0);
    $244 = +HEAPF32[$238>>2];
    $$neg16 = $244 * $242;
    $245 = ((($window$263)) + -12|0);
    $246 = +HEAPF32[$245>>2];
    $247 = ((($b0$1$pn59)) + -52|0);
    $248 = +HEAPF32[$243>>2];
    $$neg17 = $248 * $246;
    $249 = ((($window$263)) + -16|0);
    $250 = +HEAPF32[$249>>2];
    $251 = ((($b0$1$pn59)) + -48|0);
    $252 = +HEAPF32[$247>>2];
    $$neg18 = $252 * $250;
    $253 = ((($window$263)) + -20|0);
    $254 = +HEAPF32[$253>>2];
    $255 = ((($b0$1$pn59)) + -44|0);
    $256 = +HEAPF32[$251>>2];
    $$neg19 = $256 * $254;
    $257 = ((($window$263)) + -24|0);
    $258 = +HEAPF32[$257>>2];
    $259 = ((($b0$1$pn59)) + -40|0);
    $260 = +HEAPF32[$255>>2];
    $$neg20 = $260 * $258;
    $261 = ((($window$263)) + -28|0);
    $262 = +HEAPF32[$261>>2];
    $263 = ((($b0$1$pn59)) + -36|0);
    $264 = +HEAPF32[$259>>2];
    $$neg21 = $264 * $262;
    $265 = ((($window$263)) + -32|0);
    $266 = +HEAPF32[$265>>2];
    $267 = ((($b0$1$pn59)) + -32|0);
    $268 = +HEAPF32[$263>>2];
    $$neg22 = $268 * $266;
    $269 = ((($window$263)) + -36|0);
    $270 = +HEAPF32[$269>>2];
    $271 = ((($b0$1$pn59)) + -28|0);
    $272 = +HEAPF32[$267>>2];
    $$neg23 = $272 * $270;
    $273 = ((($window$263)) + -40|0);
    $274 = +HEAPF32[$273>>2];
    $275 = ((($b0$1$pn59)) + -24|0);
    $276 = +HEAPF32[$271>>2];
    $$neg24 = $276 * $274;
    $277 = ((($window$263)) + -44|0);
    $278 = +HEAPF32[$277>>2];
    $279 = ((($b0$1$pn59)) + -20|0);
    $280 = +HEAPF32[$275>>2];
    $$neg25 = $280 * $278;
    $281 = ((($window$263)) + -48|0);
    $282 = +HEAPF32[$281>>2];
    $283 = ((($b0$1$pn59)) + -16|0);
    $284 = +HEAPF32[$279>>2];
    $$neg26 = $284 * $282;
    $285 = ((($window$263)) + -52|0);
    $286 = +HEAPF32[$285>>2];
    $287 = ((($b0$1$pn59)) + -12|0);
    $288 = +HEAPF32[$283>>2];
    $$neg27 = $288 * $286;
    $289 = ((($window$263)) + -56|0);
    $290 = +HEAPF32[$289>>2];
    $291 = ((($b0$1$pn59)) + -8|0);
    $292 = +HEAPF32[$287>>2];
    $$neg28 = $292 * $290;
    $293 = ((($window$263)) + -60|0);
    $294 = +HEAPF32[$293>>2];
    $295 = ((($b0$1$pn59)) + -4|0);
    $296 = +HEAPF32[$291>>2];
    $$neg29 = $296 * $294;
    $297 = ((($window$263)) + -64|0);
    $298 = +HEAPF32[$297>>2];
    $299 = +HEAPF32[$295>>2];
    $$neg30 = $299 * $298;
    $tmp31 = $$neg30 + $240;
    $tmp33 = $$neg29 + $$neg28;
    $tmp35 = $$neg27 + $$neg26;
    $tmp37 = $$neg25 + $$neg24;
    $tmp39 = $$neg23 + $$neg22;
    $tmp41 = $$neg21 + $$neg20;
    $tmp43 = $$neg19 + $$neg18;
    $tmp45 = $$neg17 + $$neg16;
    $tmp46 = -$tmp45;
    $300 = $tmp46 - $tmp43;
    $301 = $300 - $tmp41;
    $302 = $301 - $tmp39;
    $303 = $302 - $tmp37;
    $304 = $303 - $tmp35;
    $305 = $304 - $tmp33;
    $306 = $305 - $tmp31;
    $307 = $306 > 32767.0;
    $308 = $306 < -32768.0;
    $309 = (~~(($306)));
    $310 = (($233) + ($ntom$661))|0;
    $311 = $310 >>> 15;
    $312 = $311 << 15;
    $313 = (($310) - ($312))|0;
    $314 = $311 << 1;
    if ($307) {
     $315 = (($clip$760) + 1)|0;
     if ($308) {
      $ntom$754$us$us = $234;$samples$752$us$us = $samples$658;
      while(1) {
       HEAP16[$samples$752$us$us>>1] = 32767;
       $316 = ((($samples$752$us$us)) + 4|0);
       $317 = (($ntom$754$us$us) + -32768)|0;
       $318 = ($317|0)>(32767);
       if ($318) {
        $ntom$754$us$us = $317;$samples$752$us$us = $316;
       } else {
        break;
       }
      }
      $scevgep144 = ((($samples$658)) + 4|0);
      $scevgep145 = (($scevgep144) + ($314<<1)|0);
      $329 = (($315) + ($311))|0;
      $clip$10 = $329;$ntom$8 = $313;$samples$8 = $scevgep145;
      break;
     } else {
      $ntom$754$us = $234;$samples$752$us = $samples$658;
      while(1) {
       HEAP16[$samples$752$us>>1] = 32767;
       $319 = ((($samples$752$us)) + 4|0);
       $320 = (($ntom$754$us) + -32768)|0;
       $321 = ($320|0)>(32767);
       if ($321) {
        $ntom$754$us = $320;$samples$752$us = $319;
       } else {
        break;
       }
      }
      $scevgep142 = ((($samples$658)) + 4|0);
      $scevgep143 = (($scevgep142) + ($314<<1)|0);
      $330 = (($315) + ($311))|0;
      $clip$10 = $330;$ntom$8 = $313;$samples$8 = $scevgep143;
      break;
     }
    }
    if (!($308)) {
     $ntom$754 = $234;$samples$752 = $samples$658;
     while(1) {
      HEAP16[$samples$752>>1] = $309;
      $326 = ((($samples$752)) + 4|0);
      $327 = (($ntom$754) + -32768)|0;
      $328 = ($327|0)>(32767);
      if ($328) {
       $ntom$754 = $327;$samples$752 = $326;
      } else {
       break;
      }
     }
     $scevgep = ((($samples$658)) + 4|0);
     $scevgep139 = (($scevgep) + ($314<<1)|0);
     $clip$10 = $clip$760;$ntom$8 = $313;$samples$8 = $scevgep139;
     break;
    }
    $322 = (($clip$760) + 1)|0;
    $ntom$754$us84 = $234;$samples$752$us86 = $samples$658;
    while(1) {
     HEAP16[$samples$752$us86>>1] = -32768;
     $323 = ((($samples$752$us86)) + 4|0);
     $324 = (($ntom$754$us84) + -32768)|0;
     $325 = ($324|0)>(32767);
     if ($325) {
      $ntom$754$us84 = $324;$samples$752$us86 = $323;
     } else {
      break;
     }
    }
    $scevgep140 = ((($samples$658)) + 4|0);
    $scevgep141 = (($scevgep140) + ($314<<1)|0);
    $331 = (($322) + ($311))|0;
    $clip$10 = $331;$ntom$8 = $313;$samples$8 = $scevgep141;
   }
  } while(0);
  $332 = (($j$162) + -1)|0;
  $333 = ((($window$263)) + -128|0);
  $b0$3 = ((($b0$364)) + -64|0);
  $334 = ($332|0)==(0);
  if ($334) {
   $clip$10$lcssa = $clip$10;$ntom$8$lcssa = $ntom$8;$samples$8$lcssa = $samples$8;
   break;
  } else {
   $b0$1$pn59$phi = $b0$364;$b0$364 = $b0$3;$clip$760 = $clip$10;$j$162 = $332;$ntom$661 = $ntom$8;$samples$658 = $samples$8;$window$263 = $333;$b0$1$pn59 = $b0$1$pn59$phi;
  }
 }
 $335 = (((($fr)) + 19352|0) + ($channel<<2)|0);
 HEAP32[$335>>2] = $ntom$8$lcssa;
 $336 = ($final|0)==(0);
 if ($336) {
  return ($clip$10$lcssa|0);
 }
 $337 = HEAP32[$0>>2]|0;
 $338 = $samples$8$lcssa;
 $$neg48 = $9 ? -2 : 0;
 $339 = (($338) + ($$neg48))|0;
 $340 = (($339) - ($337))|0;
 HEAP32[$2>>2] = $340;
 return ($clip$10$lcssa|0);
}
function _INT123_synth_ntom_8bit($bandPtr,$channel,$fr,$final) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $fr = $fr|0;
 $final = $final|0;
 var $$lcssa = 0, $$neg = 0.0, $$neg16 = 0.0, $$neg17 = 0.0, $$neg18 = 0.0, $$neg19 = 0.0, $$neg2 = 0.0, $$neg20 = 0.0, $$neg21 = 0.0, $$neg22 = 0.0, $$neg23 = 0.0, $$neg24 = 0.0, $$neg25 = 0.0, $$neg26 = 0.0, $$neg27 = 0.0, $$neg28 = 0.0, $$neg29 = 0.0, $$neg3 = 0.0, $$neg30 = 0.0, $$neg4 = 0.0;
 var $$neg48 = 0, $$neg5 = 0.0, $$neg6 = 0.0, $$neg7 = 0.0, $$neg8 = 0.0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0, $107 = 0.0, $108 = 0.0, $109 = 0;
 var $11 = 0, $110 = 0.0, $111 = 0, $112 = 0.0, $113 = 0, $114 = 0.0, $115 = 0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0;
 var $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0, $18 = 0, $180 = 0.0, $181 = 0;
 var $182 = 0.0, $183 = 0.0, $184 = 0.0, $185 = 0, $186 = 0.0, $187 = 0, $188 = 0.0, $189 = 0.0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0.0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0, $2 = 0;
 var $20 = 0, $200 = 0.0, $201 = 0.0, $202 = 0.0, $203 = 0, $204 = 0.0, $205 = 0, $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0.0, $213 = 0.0, $214 = 0.0, $215 = 0, $216 = 0.0, $217 = 0;
 var $218 = 0.0, $219 = 0.0, $22 = 0, $220 = 0.0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0;
 var $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0;
 var $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0.0, $267 = 0, $268 = 0.0, $269 = 0.0, $27 = 0, $270 = 0, $271 = 0.0;
 var $272 = 0, $273 = 0.0, $274 = 0, $275 = 0.0, $276 = 0, $277 = 0.0, $278 = 0, $279 = 0.0, $28 = 0, $280 = 0, $281 = 0.0, $282 = 0, $283 = 0.0, $284 = 0, $285 = 0.0, $286 = 0, $287 = 0.0, $288 = 0, $289 = 0.0, $29 = 0;
 var $290 = 0, $291 = 0.0, $292 = 0, $293 = 0.0, $294 = 0, $295 = 0.0, $296 = 0, $297 = 0.0, $298 = 0, $299 = 0.0, $3 = 0, $30 = 0, $300 = 0, $301 = 0.0, $302 = 0, $303 = 0.0, $304 = 0, $305 = 0.0, $306 = 0, $307 = 0.0;
 var $308 = 0, $309 = 0.0, $31 = 0, $310 = 0, $311 = 0.0, $312 = 0, $313 = 0.0, $314 = 0, $315 = 0.0, $316 = 0, $317 = 0.0, $318 = 0, $319 = 0.0, $32 = 0, $320 = 0, $321 = 0.0, $322 = 0, $323 = 0.0, $324 = 0, $325 = 0.0;
 var $326 = 0, $327 = 0.0, $328 = 0.0, $329 = 0.0, $33 = 0, $330 = 0.0, $331 = 0.0, $332 = 0.0, $333 = 0.0, $334 = 0.0, $335 = 0.0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0;
 var $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0;
 var $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0;
 var $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0.0;
 var $52 = 0, $53 = 0.0, $54 = 0.0, $55 = 0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0, $62 = 0.0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0, $69 = 0.0, $7 = 0;
 var $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0, $87 = 0.0, $88 = 0;
 var $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0, $92 = 0.0, $93 = 0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0, $98 = 0.0, $99 = 0.0, $b0$0 = 0, $b0$1$pn59 = 0, $b0$1$pn59$phi = 0, $b0$180 = 0, $b0$2 = 0, $b0$3 = 0, $b0$364 = 0, $bo1$0 = 0;
 var $buf$0 = 0, $clip$081 = 0, $clip$10 = 0, $clip$10$lcssa = 0, $clip$3 = 0, $clip$3$lcssa = 0, $clip$466$us = 0, $clip$466$us$us = 0, $clip$466$us90 = 0, $clip$6 = 0, $clip$760 = 0, $exitcond = 0, $exitcond151 = 0, $exitcond153 = 0, $j$082 = 0, $j$162 = 0, $ntom$0 = 0, $ntom$274 = 0, $ntom$274$us = 0, $ntom$274$us$us = 0;
 var $ntom$274$us94 = 0, $ntom$3 = 0, $ntom$467 = 0, $ntom$5 = 0, $ntom$661 = 0, $ntom$754 = 0, $ntom$754$us = 0, $ntom$754$us$us = 0, $ntom$754$us84 = 0, $ntom$8 = 0, $ntom$8$lcssa = 0, $samples$0 = 0, $samples$179 = 0, $samples$272 = 0, $samples$272$us = 0, $samples$272$us$us = 0, $samples$272$us96 = 0, $samples$3 = 0, $samples$3$lcssa = 0, $samples$465 = 0;
 var $samples$465$us = 0, $samples$465$us$us = 0, $samples$465$us91 = 0, $samples$5 = 0, $samples$658 = 0, $samples$752 = 0, $samples$752$us = 0, $samples$752$us$us = 0, $samples$752$us86 = 0, $samples$8 = 0, $samples$8$lcssa = 0, $scevgep = 0, $scevgep142 = 0, $scevgep143 = 0, $scevgep144 = 0, $scevgep145 = 0, $scevgep146 = 0, $scevgep147 = 0, $scevgep148 = 0, $scevgep149 = 0;
 var $scevgep150 = 0, $scevgep152 = 0, $scevgep154 = 0, $scevgep155 = 0, $scevgep156 = 0, $scevgep157 = 0, $scevgep158 = 0, $scevgep159 = 0, $scevgep160 = 0, $scevgep161 = 0, $scevgep162 = 0, $scevgep163 = 0, $scevgep164 = 0, $scevgep165 = 0, $scevgep166 = 0, $tmp = 0.0, $tmp10 = 0.0, $tmp12 = 0.0, $tmp14 = 0.0, $tmp31 = 0.0;
 var $tmp33 = 0.0, $tmp35 = 0.0, $tmp37 = 0.0, $tmp39 = 0.0, $tmp41 = 0.0, $tmp43 = 0.0, $tmp45 = 0.0, $tmp46 = 0.0, $window$083 = 0, $window$263 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($fr)) + 9288|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = ((($fr)) + 9292|0);
  _INT123_do_equalizer($bandPtr,$channel,$8);
 }
 $9 = ($channel|0)!=(0);
 if ($9) {
  $18 = ((($4)) + 1|0);
  $19 = ((($fr)) + 9256|0);
  $20 = ((($fr)) + 19356|0);
  $21 = HEAP32[$20>>2]|0;
  $$phi$trans$insert = ((($fr)) + 9272|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $23 = $$pre;$buf$0 = $19;$ntom$0 = $21;$samples$0 = $18;
 } else {
  $10 = ((($fr)) + 9272|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + 15)|0;
  $13 = $12 & 15;
  HEAP32[$10>>2] = $13;
  $14 = ((($fr)) + 9248|0);
  $15 = ((($fr)) + 19352|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ((($fr)) + 19356|0);
  HEAP32[$17>>2] = $16;
  $23 = $13;$buf$0 = $14;$ntom$0 = $16;$samples$0 = $4;
 }
 $22 = $23 & 1;
 $24 = ($22|0)==(0);
 $25 = ((($buf$0)) + 4|0);
 $26 = HEAP32[$25>>2]|0;
 if ($24) {
  $32 = (($23) + 1)|0;
  $33 = HEAP32[$buf$0>>2]|0;
  $34 = (($33) + ($23<<2)|0);
  $35 = (($26) + ($23<<2)|0);
  $36 = ((($35)) + 4|0);
  _INT123_dct64($34,$36,$bandPtr);
  $b0$0 = $26;$bo1$0 = $32;
 } else {
  $27 = HEAP32[$buf$0>>2]|0;
  $28 = (($23) + 1)|0;
  $29 = $28 & 15;
  $30 = (($26) + ($29<<2)|0);
  $31 = (($27) + ($23<<2)|0);
  _INT123_dct64($30,$31,$bandPtr);
  $b0$0 = $27;$bo1$0 = $23;
 }
 $37 = ((($fr)) + 9284|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = ((($38)) + 64|0);
 $40 = (0 - ($bo1$0))|0;
 $41 = (($39) + ($40<<2)|0);
 $42 = ((($fr)) + 19360|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = (($43) + ($ntom$0))|0;
 $45 = ((($fr)) + 9592|0);
 $46 = (528 - ($bo1$0))|0;
 $scevgep164 = ((($b0$0)) + 1024|0);
 $47 = (496 - ($bo1$0))|0;
 $49 = $44;$b0$180 = $b0$0;$clip$081 = 0;$j$082 = 16;$samples$179 = $samples$0;$window$083 = $41;
 while(1) {
  $48 = ($49|0)<(32768);
  do {
   if ($48) {
    $clip$3 = $clip$081;$ntom$3 = $49;$samples$3 = $samples$179;
   } else {
    $50 = ((($window$083)) + 4|0);
    $51 = +HEAPF32[$window$083>>2];
    $52 = ((($b0$180)) + 4|0);
    $53 = +HEAPF32[$b0$180>>2];
    $54 = $53 * $51;
    $55 = ((($window$083)) + 8|0);
    $56 = +HEAPF32[$50>>2];
    $57 = ((($b0$180)) + 8|0);
    $58 = +HEAPF32[$52>>2];
    $$neg = $58 * $56;
    $59 = ((($window$083)) + 12|0);
    $60 = +HEAPF32[$55>>2];
    $61 = ((($b0$180)) + 12|0);
    $62 = +HEAPF32[$57>>2];
    $63 = $62 * $60;
    $64 = ((($window$083)) + 16|0);
    $65 = +HEAPF32[$59>>2];
    $66 = ((($b0$180)) + 16|0);
    $67 = +HEAPF32[$61>>2];
    $$neg2 = $67 * $65;
    $68 = ((($window$083)) + 20|0);
    $69 = +HEAPF32[$64>>2];
    $70 = ((($b0$180)) + 20|0);
    $71 = +HEAPF32[$66>>2];
    $72 = $71 * $69;
    $73 = ((($window$083)) + 24|0);
    $74 = +HEAPF32[$68>>2];
    $75 = ((($b0$180)) + 24|0);
    $76 = +HEAPF32[$70>>2];
    $$neg3 = $76 * $74;
    $77 = ((($window$083)) + 28|0);
    $78 = +HEAPF32[$73>>2];
    $79 = ((($b0$180)) + 28|0);
    $80 = +HEAPF32[$75>>2];
    $81 = $80 * $78;
    $82 = ((($window$083)) + 32|0);
    $83 = +HEAPF32[$77>>2];
    $84 = ((($b0$180)) + 32|0);
    $85 = +HEAPF32[$79>>2];
    $$neg4 = $85 * $83;
    $86 = ((($window$083)) + 36|0);
    $87 = +HEAPF32[$82>>2];
    $88 = ((($b0$180)) + 36|0);
    $89 = +HEAPF32[$84>>2];
    $90 = $89 * $87;
    $91 = ((($window$083)) + 40|0);
    $92 = +HEAPF32[$86>>2];
    $93 = ((($b0$180)) + 40|0);
    $94 = +HEAPF32[$88>>2];
    $$neg5 = $94 * $92;
    $95 = ((($window$083)) + 44|0);
    $96 = +HEAPF32[$91>>2];
    $97 = ((($b0$180)) + 44|0);
    $98 = +HEAPF32[$93>>2];
    $99 = $98 * $96;
    $100 = ((($window$083)) + 48|0);
    $101 = +HEAPF32[$95>>2];
    $102 = ((($b0$180)) + 48|0);
    $103 = +HEAPF32[$97>>2];
    $$neg6 = $103 * $101;
    $104 = ((($window$083)) + 52|0);
    $105 = +HEAPF32[$100>>2];
    $106 = ((($b0$180)) + 52|0);
    $107 = +HEAPF32[$102>>2];
    $108 = $107 * $105;
    $109 = ((($window$083)) + 56|0);
    $110 = +HEAPF32[$104>>2];
    $111 = ((($b0$180)) + 56|0);
    $112 = +HEAPF32[$106>>2];
    $$neg7 = $112 * $110;
    $113 = ((($window$083)) + 60|0);
    $114 = +HEAPF32[$109>>2];
    $115 = ((($b0$180)) + 60|0);
    $116 = +HEAPF32[$111>>2];
    $117 = $116 * $114;
    $118 = +HEAPF32[$113>>2];
    $119 = +HEAPF32[$115>>2];
    $$neg8 = $119 * $118;
    $tmp = $$neg8 + $$neg7;
    $tmp10 = $$neg6 + $$neg5;
    $tmp12 = $$neg4 + $$neg3;
    $tmp14 = $$neg2 + $$neg;
    $120 = $63 + $54;
    $121 = $120 + $72;
    $122 = $121 + $81;
    $123 = $122 + $90;
    $124 = $123 + $99;
    $125 = $124 + $108;
    $126 = $125 + $117;
    $127 = $126 - $tmp14;
    $128 = $127 - $tmp12;
    $129 = $128 - $tmp10;
    $130 = $129 - $tmp;
    $131 = $130 > 32767.0;
    $132 = $130 < -32768.0;
    $133 = (~~(($130)));
    $134 = (($49) + -32768)|0;
    $135 = $134 >>> 15;
    $136 = $135 << 15;
    $137 = (($134) - ($136))|0;
    $138 = $135 << 1;
    if ($131) {
     $139 = (($clip$081) + 1)|0;
     if ($132) {
      $ntom$274$us$us = $49;$samples$272$us$us = $samples$179;
      while(1) {
       $140 = HEAP32[$45>>2]|0;
       $141 = ((($140)) + 4095|0);
       $142 = HEAP8[$141>>0]|0;
       HEAP8[$samples$272$us$us>>0] = $142;
       $143 = ((($samples$272$us$us)) + 2|0);
       $144 = (($ntom$274$us$us) + -32768)|0;
       $145 = ($144|0)>(32767);
       if ($145) {
        $ntom$274$us$us = $144;$samples$272$us$us = $143;
       } else {
        break;
       }
      }
      $scevgep161 = ((($samples$179)) + 2|0);
      $scevgep162 = (($scevgep161) + ($138)|0);
      $167 = (($139) + ($135))|0;
      $clip$3 = $167;$ntom$3 = $137;$samples$3 = $scevgep162;
      break;
     } else {
      $ntom$274$us = $49;$samples$272$us = $samples$179;
      while(1) {
       $146 = HEAP32[$45>>2]|0;
       $147 = ((($146)) + 4095|0);
       $148 = HEAP8[$147>>0]|0;
       HEAP8[$samples$272$us>>0] = $148;
       $149 = ((($samples$272$us)) + 2|0);
       $150 = (($ntom$274$us) + -32768)|0;
       $151 = ($150|0)>(32767);
       if ($151) {
        $ntom$274$us = $150;$samples$272$us = $149;
       } else {
        break;
       }
      }
      $scevgep159 = ((($samples$179)) + 2|0);
      $scevgep160 = (($scevgep159) + ($138)|0);
      $168 = (($139) + ($135))|0;
      $clip$3 = $168;$ntom$3 = $137;$samples$3 = $scevgep160;
      break;
     }
    }
    if ($132) {
     $154 = (($clip$081) + 1)|0;
     $ntom$274$us94 = $49;$samples$272$us96 = $samples$179;
     while(1) {
      $155 = HEAP32[$45>>2]|0;
      $156 = ((($155)) + -4096|0);
      $157 = HEAP8[$156>>0]|0;
      HEAP8[$samples$272$us96>>0] = $157;
      $158 = ((($samples$272$us96)) + 2|0);
      $159 = (($ntom$274$us94) + -32768)|0;
      $160 = ($159|0)>(32767);
      if ($160) {
       $ntom$274$us94 = $159;$samples$272$us96 = $158;
      } else {
       break;
      }
     }
     $scevgep157 = ((($samples$179)) + 2|0);
     $scevgep158 = (($scevgep157) + ($138)|0);
     $169 = (($154) + ($135))|0;
     $clip$3 = $169;$ntom$3 = $137;$samples$3 = $scevgep158;
     break;
    } else {
     $152 = $133 << 16 >> 16;
     $153 = $152 >> 3;
     $ntom$274 = $49;$samples$272 = $samples$179;
     while(1) {
      $161 = HEAP32[$45>>2]|0;
      $162 = (($161) + ($153)|0);
      $163 = HEAP8[$162>>0]|0;
      HEAP8[$samples$272>>0] = $163;
      $164 = ((($samples$272)) + 2|0);
      $165 = (($ntom$274) + -32768)|0;
      $166 = ($165|0)>(32767);
      if ($166) {
       $ntom$274 = $165;$samples$272 = $164;
      } else {
       break;
      }
     }
     $scevgep155 = ((($samples$179)) + 2|0);
     $scevgep156 = (($scevgep155) + ($138)|0);
     $clip$3 = $clip$081;$ntom$3 = $137;$samples$3 = $scevgep156;
     break;
    }
   }
  } while(0);
  $b0$2 = ((($b0$180)) + 64|0);
  $170 = (($j$082) + -1)|0;
  $171 = ((($window$083)) + 128|0);
  $172 = ($170|0)==(0);
  $173 = HEAP32[$42>>2]|0;
  $174 = (($173) + ($ntom$3))|0;
  if ($172) {
   $$lcssa = $174;$clip$3$lcssa = $clip$3;$samples$3$lcssa = $samples$3;
   break;
  } else {
   $49 = $174;$b0$180 = $b0$2;$clip$081 = $clip$3;$j$082 = $170;$samples$179 = $samples$3;$window$083 = $171;
  }
 }
 $scevgep165 = ((($b0$0)) + 960|0);
 $scevgep166 = (($38) + ($47<<2)|0);
 $175 = ($$lcssa|0)>(32767);
 do {
  if ($175) {
   $scevgep163 = (($38) + ($46<<2)|0);
   $176 = +HEAPF32[$scevgep163>>2];
   $177 = +HEAPF32[$scevgep164>>2];
   $178 = $177 * $176;
   $179 = ((($scevgep166)) + 136|0);
   $180 = +HEAPF32[$179>>2];
   $181 = ((($b0$0)) + 1032|0);
   $182 = +HEAPF32[$181>>2];
   $183 = $182 * $180;
   $184 = $183 + $178;
   $185 = ((($scevgep166)) + 144|0);
   $186 = +HEAPF32[$185>>2];
   $187 = ((($b0$0)) + 1040|0);
   $188 = +HEAPF32[$187>>2];
   $189 = $188 * $186;
   $190 = $184 + $189;
   $191 = ((($scevgep166)) + 152|0);
   $192 = +HEAPF32[$191>>2];
   $193 = ((($b0$0)) + 1048|0);
   $194 = +HEAPF32[$193>>2];
   $195 = $194 * $192;
   $196 = $190 + $195;
   $197 = ((($scevgep166)) + 160|0);
   $198 = +HEAPF32[$197>>2];
   $199 = ((($b0$0)) + 1056|0);
   $200 = +HEAPF32[$199>>2];
   $201 = $200 * $198;
   $202 = $196 + $201;
   $203 = ((($scevgep166)) + 168|0);
   $204 = +HEAPF32[$203>>2];
   $205 = ((($b0$0)) + 1064|0);
   $206 = +HEAPF32[$205>>2];
   $207 = $206 * $204;
   $208 = $202 + $207;
   $209 = ((($scevgep166)) + 176|0);
   $210 = +HEAPF32[$209>>2];
   $211 = ((($b0$0)) + 1072|0);
   $212 = +HEAPF32[$211>>2];
   $213 = $212 * $210;
   $214 = $208 + $213;
   $215 = ((($scevgep166)) + 184|0);
   $216 = +HEAPF32[$215>>2];
   $217 = ((($b0$0)) + 1080|0);
   $218 = +HEAPF32[$217>>2];
   $219 = $218 * $216;
   $220 = $214 + $219;
   $221 = $220 > 32767.0;
   $222 = $220 < -32768.0;
   $223 = (~~(($220)));
   $224 = (($$lcssa) + -32768)|0;
   $225 = $224 >>> 15;
   $226 = $225 << 15;
   $227 = $225 << 1;
   $228 = (($227) + 2)|0;
   if ($221) {
    $229 = (($clip$3$lcssa) + ($225))|0;
    $230 = (($229) + 1)|0;
    if ($222) {
     $clip$466$us$us = $clip$3$lcssa;$samples$465$us$us = $samples$3$lcssa;
     while(1) {
      $231 = (($clip$466$us$us) + 1)|0;
      $232 = HEAP32[$45>>2]|0;
      $233 = ((($232)) + 4095|0);
      $234 = HEAP8[$233>>0]|0;
      HEAP8[$samples$465$us$us>>0] = $234;
      $235 = ((($samples$465$us$us)) + 2|0);
      $exitcond = ($231|0)==($230|0);
      if ($exitcond) {
       break;
      } else {
       $clip$466$us$us = $231;$samples$465$us$us = $235;
      }
     }
     $256 = (($224) - ($226))|0;
     $scevgep149 = (($samples$3$lcssa) + ($228)|0);
     $clip$6 = $230;$ntom$5 = $256;$samples$5 = $scevgep149;
     break;
    } else {
     $clip$466$us = $clip$3$lcssa;$samples$465$us = $samples$3$lcssa;
     while(1) {
      $236 = (($clip$466$us) + 1)|0;
      $237 = HEAP32[$45>>2]|0;
      $238 = ((($237)) + 4095|0);
      $239 = HEAP8[$238>>0]|0;
      HEAP8[$samples$465$us>>0] = $239;
      $240 = ((($samples$465$us)) + 2|0);
      $exitcond151 = ($236|0)==($230|0);
      if ($exitcond151) {
       break;
      } else {
       $clip$466$us = $236;$samples$465$us = $240;
      }
     }
     $257 = (($224) - ($226))|0;
     $scevgep150 = (($samples$3$lcssa) + ($228)|0);
     $clip$6 = $230;$ntom$5 = $257;$samples$5 = $scevgep150;
     break;
    }
   }
   if ($222) {
    $243 = (($clip$3$lcssa) + ($225))|0;
    $244 = (($243) + 1)|0;
    $clip$466$us90 = $clip$3$lcssa;$samples$465$us91 = $samples$3$lcssa;
    while(1) {
     $245 = (($clip$466$us90) + 1)|0;
     $246 = HEAP32[$45>>2]|0;
     $247 = ((($246)) + -4096|0);
     $248 = HEAP8[$247>>0]|0;
     HEAP8[$samples$465$us91>>0] = $248;
     $249 = ((($samples$465$us91)) + 2|0);
     $exitcond153 = ($245|0)==($244|0);
     if ($exitcond153) {
      break;
     } else {
      $clip$466$us90 = $245;$samples$465$us91 = $249;
     }
    }
    $258 = (($224) - ($226))|0;
    $scevgep152 = (($samples$3$lcssa) + ($228)|0);
    $clip$6 = $244;$ntom$5 = $258;$samples$5 = $scevgep152;
    break;
   } else {
    $241 = $223 << 16 >> 16;
    $242 = $241 >> 3;
    $ntom$467 = $$lcssa;$samples$465 = $samples$3$lcssa;
    while(1) {
     $250 = HEAP32[$45>>2]|0;
     $251 = (($250) + ($242)|0);
     $252 = HEAP8[$251>>0]|0;
     HEAP8[$samples$465>>0] = $252;
     $253 = ((($samples$465)) + 2|0);
     $254 = (($ntom$467) + -32768)|0;
     $255 = ($254|0)>(32767);
     if ($255) {
      $ntom$467 = $254;$samples$465 = $253;
     } else {
      break;
     }
    }
    $259 = (($224) - ($226))|0;
    $scevgep154 = (($samples$3$lcssa) + ($228)|0);
    $clip$6 = $clip$3$lcssa;$ntom$5 = $259;$samples$5 = $scevgep154;
    break;
   }
  } else {
   $clip$6 = $clip$3$lcssa;$ntom$5 = $$lcssa;$samples$5 = $samples$3$lcssa;
  }
 } while(0);
 $260 = $bo1$0 << 1;
 $261 = (($scevgep166) + ($260<<2)|0);
 $b0$1$pn59 = $scevgep164;$b0$364 = $scevgep165;$clip$760 = $clip$6;$j$162 = 15;$ntom$661 = $ntom$5;$samples$658 = $samples$5;$window$263 = $261;
 while(1) {
  $262 = HEAP32[$42>>2]|0;
  $263 = (($262) + ($ntom$661))|0;
  $264 = ($263|0)<(32768);
  do {
   if ($264) {
    $clip$10 = $clip$760;$ntom$8 = $263;$samples$8 = $samples$658;
   } else {
    $265 = ((($window$263)) + -4|0);
    $266 = +HEAPF32[$265>>2];
    $267 = ((($b0$1$pn59)) + -60|0);
    $268 = +HEAPF32[$b0$364>>2];
    $269 = $268 * $266;
    $270 = ((($window$263)) + -8|0);
    $271 = +HEAPF32[$270>>2];
    $272 = ((($b0$1$pn59)) + -56|0);
    $273 = +HEAPF32[$267>>2];
    $$neg16 = $273 * $271;
    $274 = ((($window$263)) + -12|0);
    $275 = +HEAPF32[$274>>2];
    $276 = ((($b0$1$pn59)) + -52|0);
    $277 = +HEAPF32[$272>>2];
    $$neg17 = $277 * $275;
    $278 = ((($window$263)) + -16|0);
    $279 = +HEAPF32[$278>>2];
    $280 = ((($b0$1$pn59)) + -48|0);
    $281 = +HEAPF32[$276>>2];
    $$neg18 = $281 * $279;
    $282 = ((($window$263)) + -20|0);
    $283 = +HEAPF32[$282>>2];
    $284 = ((($b0$1$pn59)) + -44|0);
    $285 = +HEAPF32[$280>>2];
    $$neg19 = $285 * $283;
    $286 = ((($window$263)) + -24|0);
    $287 = +HEAPF32[$286>>2];
    $288 = ((($b0$1$pn59)) + -40|0);
    $289 = +HEAPF32[$284>>2];
    $$neg20 = $289 * $287;
    $290 = ((($window$263)) + -28|0);
    $291 = +HEAPF32[$290>>2];
    $292 = ((($b0$1$pn59)) + -36|0);
    $293 = +HEAPF32[$288>>2];
    $$neg21 = $293 * $291;
    $294 = ((($window$263)) + -32|0);
    $295 = +HEAPF32[$294>>2];
    $296 = ((($b0$1$pn59)) + -32|0);
    $297 = +HEAPF32[$292>>2];
    $$neg22 = $297 * $295;
    $298 = ((($window$263)) + -36|0);
    $299 = +HEAPF32[$298>>2];
    $300 = ((($b0$1$pn59)) + -28|0);
    $301 = +HEAPF32[$296>>2];
    $$neg23 = $301 * $299;
    $302 = ((($window$263)) + -40|0);
    $303 = +HEAPF32[$302>>2];
    $304 = ((($b0$1$pn59)) + -24|0);
    $305 = +HEAPF32[$300>>2];
    $$neg24 = $305 * $303;
    $306 = ((($window$263)) + -44|0);
    $307 = +HEAPF32[$306>>2];
    $308 = ((($b0$1$pn59)) + -20|0);
    $309 = +HEAPF32[$304>>2];
    $$neg25 = $309 * $307;
    $310 = ((($window$263)) + -48|0);
    $311 = +HEAPF32[$310>>2];
    $312 = ((($b0$1$pn59)) + -16|0);
    $313 = +HEAPF32[$308>>2];
    $$neg26 = $313 * $311;
    $314 = ((($window$263)) + -52|0);
    $315 = +HEAPF32[$314>>2];
    $316 = ((($b0$1$pn59)) + -12|0);
    $317 = +HEAPF32[$312>>2];
    $$neg27 = $317 * $315;
    $318 = ((($window$263)) + -56|0);
    $319 = +HEAPF32[$318>>2];
    $320 = ((($b0$1$pn59)) + -8|0);
    $321 = +HEAPF32[$316>>2];
    $$neg28 = $321 * $319;
    $322 = ((($window$263)) + -60|0);
    $323 = +HEAPF32[$322>>2];
    $324 = ((($b0$1$pn59)) + -4|0);
    $325 = +HEAPF32[$320>>2];
    $$neg29 = $325 * $323;
    $326 = ((($window$263)) + -64|0);
    $327 = +HEAPF32[$326>>2];
    $328 = +HEAPF32[$324>>2];
    $$neg30 = $328 * $327;
    $tmp31 = $$neg30 + $269;
    $tmp33 = $$neg29 + $$neg28;
    $tmp35 = $$neg27 + $$neg26;
    $tmp37 = $$neg25 + $$neg24;
    $tmp39 = $$neg23 + $$neg22;
    $tmp41 = $$neg21 + $$neg20;
    $tmp43 = $$neg19 + $$neg18;
    $tmp45 = $$neg17 + $$neg16;
    $tmp46 = -$tmp45;
    $329 = $tmp46 - $tmp43;
    $330 = $329 - $tmp41;
    $331 = $330 - $tmp39;
    $332 = $331 - $tmp37;
    $333 = $332 - $tmp35;
    $334 = $333 - $tmp33;
    $335 = $334 - $tmp31;
    $336 = $335 > 32767.0;
    $337 = $335 < -32768.0;
    $338 = (~~(($335)));
    $339 = (($ntom$661) + -32768)|0;
    $340 = (($339) + ($262))|0;
    $341 = $340 >>> 15;
    $342 = $341 << 15;
    $343 = (($340) - ($342))|0;
    $344 = $341 << 1;
    if ($336) {
     $345 = (($clip$760) + 1)|0;
     if ($337) {
      $ntom$754$us$us = $263;$samples$752$us$us = $samples$658;
      while(1) {
       $346 = HEAP32[$45>>2]|0;
       $347 = ((($346)) + 4095|0);
       $348 = HEAP8[$347>>0]|0;
       HEAP8[$samples$752$us$us>>0] = $348;
       $349 = ((($samples$752$us$us)) + 2|0);
       $350 = (($ntom$754$us$us) + -32768)|0;
       $351 = ($350|0)>(32767);
       if ($351) {
        $ntom$754$us$us = $350;$samples$752$us$us = $349;
       } else {
        break;
       }
      }
      $scevgep147 = ((($samples$658)) + 2|0);
      $scevgep148 = (($scevgep147) + ($344)|0);
      $373 = (($345) + ($341))|0;
      $clip$10 = $373;$ntom$8 = $343;$samples$8 = $scevgep148;
      break;
     } else {
      $ntom$754$us = $263;$samples$752$us = $samples$658;
      while(1) {
       $352 = HEAP32[$45>>2]|0;
       $353 = ((($352)) + 4095|0);
       $354 = HEAP8[$353>>0]|0;
       HEAP8[$samples$752$us>>0] = $354;
       $355 = ((($samples$752$us)) + 2|0);
       $356 = (($ntom$754$us) + -32768)|0;
       $357 = ($356|0)>(32767);
       if ($357) {
        $ntom$754$us = $356;$samples$752$us = $355;
       } else {
        break;
       }
      }
      $scevgep145 = ((($samples$658)) + 2|0);
      $scevgep146 = (($scevgep145) + ($344)|0);
      $374 = (($345) + ($341))|0;
      $clip$10 = $374;$ntom$8 = $343;$samples$8 = $scevgep146;
      break;
     }
    }
    if ($337) {
     $360 = (($clip$760) + 1)|0;
     $ntom$754$us84 = $263;$samples$752$us86 = $samples$658;
     while(1) {
      $361 = HEAP32[$45>>2]|0;
      $362 = ((($361)) + -4096|0);
      $363 = HEAP8[$362>>0]|0;
      HEAP8[$samples$752$us86>>0] = $363;
      $364 = ((($samples$752$us86)) + 2|0);
      $365 = (($ntom$754$us84) + -32768)|0;
      $366 = ($365|0)>(32767);
      if ($366) {
       $ntom$754$us84 = $365;$samples$752$us86 = $364;
      } else {
       break;
      }
     }
     $scevgep143 = ((($samples$658)) + 2|0);
     $scevgep144 = (($scevgep143) + ($344)|0);
     $375 = (($360) + ($341))|0;
     $clip$10 = $375;$ntom$8 = $343;$samples$8 = $scevgep144;
     break;
    } else {
     $358 = $338 << 16 >> 16;
     $359 = $358 >> 3;
     $ntom$754 = $263;$samples$752 = $samples$658;
     while(1) {
      $367 = HEAP32[$45>>2]|0;
      $368 = (($367) + ($359)|0);
      $369 = HEAP8[$368>>0]|0;
      HEAP8[$samples$752>>0] = $369;
      $370 = ((($samples$752)) + 2|0);
      $371 = (($ntom$754) + -32768)|0;
      $372 = ($371|0)>(32767);
      if ($372) {
       $ntom$754 = $371;$samples$752 = $370;
      } else {
       break;
      }
     }
     $scevgep = ((($samples$658)) + 2|0);
     $scevgep142 = (($scevgep) + ($344)|0);
     $clip$10 = $clip$760;$ntom$8 = $343;$samples$8 = $scevgep142;
     break;
    }
   }
  } while(0);
  $376 = (($j$162) + -1)|0;
  $377 = ((($window$263)) + -128|0);
  $b0$3 = ((($b0$364)) + -64|0);
  $378 = ($376|0)==(0);
  if ($378) {
   $clip$10$lcssa = $clip$10;$ntom$8$lcssa = $ntom$8;$samples$8$lcssa = $samples$8;
   break;
  } else {
   $b0$1$pn59$phi = $b0$364;$b0$364 = $b0$3;$clip$760 = $clip$10;$j$162 = $376;$ntom$661 = $ntom$8;$samples$658 = $samples$8;$window$263 = $377;$b0$1$pn59 = $b0$1$pn59$phi;
  }
 }
 $379 = (((($fr)) + 19352|0) + ($channel<<2)|0);
 HEAP32[$379>>2] = $ntom$8$lcssa;
 $380 = ($final|0)==(0);
 if ($380) {
  return ($clip$10$lcssa|0);
 }
 $381 = HEAP32[$0>>2]|0;
 $382 = $samples$8$lcssa;
 $$neg48 = $9 << 31 >> 31;
 $383 = (($382) + ($$neg48))|0;
 $384 = (($383) - ($381))|0;
 HEAP32[$2>>2] = $384;
 return ($clip$10$lcssa|0);
}
function _INT123_synth_ntom_real($bandPtr,$channel,$fr,$final) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $fr = $fr|0;
 $final = $final|0;
 var $$lcssa = 0, $$neg = 0.0, $$neg16 = 0.0, $$neg17 = 0.0, $$neg18 = 0.0, $$neg19 = 0.0, $$neg2 = 0.0, $$neg20 = 0.0, $$neg21 = 0.0, $$neg22 = 0.0, $$neg23 = 0.0, $$neg24 = 0.0, $$neg25 = 0.0, $$neg26 = 0.0, $$neg27 = 0.0, $$neg28 = 0.0, $$neg29 = 0.0, $$neg3 = 0.0, $$neg30 = 0.0, $$neg4 = 0.0;
 var $$neg48 = 0, $$neg5 = 0.0, $$neg6 = 0.0, $$neg7 = 0.0, $$neg8 = 0.0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0.0, $103 = 0, $104 = 0.0, $105 = 0, $106 = 0.0, $107 = 0, $108 = 0.0, $109 = 0;
 var $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0, $113 = 0.0, $114 = 0, $115 = 0.0, $116 = 0, $117 = 0.0, $118 = 0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0.0, $123 = 0, $124 = 0.0, $125 = 0, $126 = 0.0, $127 = 0;
 var $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0.0;
 var $146 = 0, $147 = 0.0, $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0.0, $152 = 0.0, $153 = 0, $154 = 0.0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0, $16 = 0, $160 = 0.0, $161 = 0, $162 = 0.0, $163 = 0.0;
 var $164 = 0.0, $165 = 0, $166 = 0.0, $167 = 0, $168 = 0.0, $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0, $172 = 0.0, $173 = 0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0, $178 = 0.0, $179 = 0, $18 = 0, $180 = 0.0, $181 = 0.0;
 var $182 = 0.0, $183 = 0, $184 = 0.0, $185 = 0, $186 = 0.0, $187 = 0.0, $188 = 0.0, $189 = 0.0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0;
 var $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0.0, $207 = 0, $208 = 0.0, $209 = 0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0.0, $213 = 0, $214 = 0.0, $215 = 0, $216 = 0.0, $217 = 0;
 var $218 = 0.0, $219 = 0, $22 = 0, $220 = 0.0, $221 = 0.0, $222 = 0, $223 = 0.0, $224 = 0, $225 = 0.0, $226 = 0, $227 = 0.0, $228 = 0, $229 = 0.0, $23 = 0, $230 = 0.0, $231 = 0, $232 = 0.0, $233 = 0, $234 = 0.0, $235 = 0;
 var $236 = 0.0, $237 = 0, $238 = 0.0, $239 = 0.0, $24 = 0, $240 = 0, $241 = 0.0, $242 = 0, $243 = 0.0, $244 = 0, $245 = 0.0, $246 = 0, $247 = 0.0, $248 = 0.0, $249 = 0, $25 = 0, $250 = 0.0, $251 = 0, $252 = 0.0, $253 = 0;
 var $254 = 0.0, $255 = 0, $256 = 0.0, $257 = 0.0, $258 = 0, $259 = 0.0, $26 = 0, $260 = 0, $261 = 0.0, $262 = 0, $263 = 0.0, $264 = 0, $265 = 0.0, $266 = 0.0, $267 = 0, $268 = 0.0, $269 = 0, $27 = 0, $270 = 0.0, $271 = 0.0;
 var $272 = 0, $273 = 0.0, $274 = 0.0, $275 = 0.0, $276 = 0.0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0;
 var $290 = 0, $291 = 0, $292 = 0, $293 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0;
 var $62 = 0.0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0;
 var $80 = 0.0, $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0, $89 = 0.0, $9 = 0, $90 = 0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0, $95 = 0.0, $96 = 0, $97 = 0.0, $98 = 0;
 var $99 = 0.0, $b0$0 = 0, $b0$1$pn56 = 0, $b0$1$pn56$phi = 0, $b0$172 = 0, $b0$2 = 0, $b0$3 = 0, $b0$361 = 0, $bo1$0 = 0, $buf$0 = 0, $j$074 = 0, $j$159 = 0, $ntom$0 = 0, $ntom$268 = 0, $ntom$3 = 0, $ntom$463 = 0, $ntom$5 = 0, $ntom$658 = 0, $ntom$753 = 0, $ntom$8 = 0;
 var $ntom$8$lcssa = 0, $samples$0$in = 0, $samples$173 = 0, $samples$267 = 0, $samples$3 = 0, $samples$3$lcssa = 0, $samples$462 = 0, $samples$5 = 0, $samples$657 = 0, $samples$752 = 0, $samples$8 = 0, $samples$8$lcssa = 0, $scevgep = 0, $scevgep83 = 0, $scevgep84 = 0, $scevgep85 = 0, $scevgep86 = 0, $scevgep87 = 0, $scevgep88 = 0, $scevgep89 = 0;
 var $scevgep90 = 0, $tmp = 0.0, $tmp10 = 0.0, $tmp12 = 0.0, $tmp14 = 0.0, $tmp31 = 0.0, $tmp33 = 0.0, $tmp35 = 0.0, $tmp37 = 0.0, $tmp39 = 0.0, $tmp41 = 0.0, $tmp43 = 0.0, $tmp45 = 0.0, $tmp46 = 0.0, $window$075 = 0, $window$260 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($fr)) + 9288|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = ((($fr)) + 9292|0);
  _INT123_do_equalizer($bandPtr,$channel,$8);
 }
 $9 = ($channel|0)!=(0);
 if ($9) {
  $18 = ((($4)) + 4|0);
  $19 = ((($fr)) + 9256|0);
  $20 = ((($fr)) + 19356|0);
  $21 = HEAP32[$20>>2]|0;
  $$phi$trans$insert = ((($fr)) + 9272|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $23 = $$pre;$buf$0 = $19;$ntom$0 = $21;$samples$0$in = $18;
 } else {
  $10 = ((($fr)) + 9272|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + 15)|0;
  $13 = $12 & 15;
  HEAP32[$10>>2] = $13;
  $14 = ((($fr)) + 9248|0);
  $15 = ((($fr)) + 19352|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ((($fr)) + 19356|0);
  HEAP32[$17>>2] = $16;
  $23 = $13;$buf$0 = $14;$ntom$0 = $16;$samples$0$in = $4;
 }
 $22 = $23 & 1;
 $24 = ($22|0)==(0);
 $25 = ((($buf$0)) + 4|0);
 $26 = HEAP32[$25>>2]|0;
 if ($24) {
  $32 = (($23) + 1)|0;
  $33 = HEAP32[$buf$0>>2]|0;
  $34 = (($33) + ($23<<2)|0);
  $35 = (($26) + ($23<<2)|0);
  $36 = ((($35)) + 4|0);
  _INT123_dct64($34,$36,$bandPtr);
  $b0$0 = $26;$bo1$0 = $32;
 } else {
  $27 = HEAP32[$buf$0>>2]|0;
  $28 = (($23) + 1)|0;
  $29 = $28 & 15;
  $30 = (($26) + ($29<<2)|0);
  $31 = (($27) + ($23<<2)|0);
  _INT123_dct64($30,$31,$bandPtr);
  $b0$0 = $27;$bo1$0 = $23;
 }
 $37 = ((($fr)) + 9284|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = ((($38)) + 64|0);
 $40 = (0 - ($bo1$0))|0;
 $41 = (($39) + ($40<<2)|0);
 $42 = ((($fr)) + 19360|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = (($43) + ($ntom$0))|0;
 $45 = (528 - ($bo1$0))|0;
 $scevgep88 = ((($b0$0)) + 1024|0);
 $46 = (496 - ($bo1$0))|0;
 $48 = $44;$b0$172 = $b0$0;$j$074 = 16;$samples$173 = $samples$0$in;$window$075 = $41;
 while(1) {
  $47 = ($48|0)<(32768);
  if ($47) {
   $ntom$3 = $48;$samples$3 = $samples$173;
  } else {
   $49 = ((($b0$172)) + 8|0);
   $50 = +HEAPF32[$49>>2];
   $51 = ((($window$075)) + 8|0);
   $52 = +HEAPF32[$51>>2];
   $53 = $50 * $52;
   $54 = +HEAPF32[$b0$172>>2];
   $55 = +HEAPF32[$window$075>>2];
   $56 = $54 * $55;
   $57 = $53 + $56;
   $58 = ((($b0$172)) + 16|0);
   $59 = +HEAPF32[$58>>2];
   $60 = ((($window$075)) + 16|0);
   $61 = +HEAPF32[$60>>2];
   $62 = $59 * $61;
   $63 = $57 + $62;
   $64 = ((($b0$172)) + 24|0);
   $65 = +HEAPF32[$64>>2];
   $66 = ((($window$075)) + 24|0);
   $67 = +HEAPF32[$66>>2];
   $68 = $65 * $67;
   $69 = $63 + $68;
   $70 = ((($b0$172)) + 32|0);
   $71 = +HEAPF32[$70>>2];
   $72 = ((($window$075)) + 32|0);
   $73 = +HEAPF32[$72>>2];
   $74 = $71 * $73;
   $75 = $69 + $74;
   $76 = ((($b0$172)) + 40|0);
   $77 = +HEAPF32[$76>>2];
   $78 = ((($window$075)) + 40|0);
   $79 = +HEAPF32[$78>>2];
   $80 = $77 * $79;
   $81 = $75 + $80;
   $82 = ((($b0$172)) + 48|0);
   $83 = +HEAPF32[$82>>2];
   $84 = ((($window$075)) + 48|0);
   $85 = +HEAPF32[$84>>2];
   $86 = $83 * $85;
   $87 = $81 + $86;
   $88 = ((($b0$172)) + 56|0);
   $89 = +HEAPF32[$88>>2];
   $90 = ((($window$075)) + 56|0);
   $91 = +HEAPF32[$90>>2];
   $92 = $89 * $91;
   $93 = $87 + $92;
   $94 = ((($b0$172)) + 12|0);
   $95 = +HEAPF32[$94>>2];
   $96 = ((($window$075)) + 12|0);
   $97 = +HEAPF32[$96>>2];
   $$neg2 = $95 * $97;
   $98 = ((($b0$172)) + 4|0);
   $99 = +HEAPF32[$98>>2];
   $100 = ((($window$075)) + 4|0);
   $101 = +HEAPF32[$100>>2];
   $$neg = $99 * $101;
   $tmp14 = $$neg2 + $$neg;
   $102 = $93 - $tmp14;
   $103 = ((($b0$172)) + 28|0);
   $104 = +HEAPF32[$103>>2];
   $105 = ((($window$075)) + 28|0);
   $106 = +HEAPF32[$105>>2];
   $$neg4 = $104 * $106;
   $107 = ((($b0$172)) + 20|0);
   $108 = +HEAPF32[$107>>2];
   $109 = ((($window$075)) + 20|0);
   $110 = +HEAPF32[$109>>2];
   $$neg3 = $108 * $110;
   $tmp12 = $$neg4 + $$neg3;
   $111 = $102 - $tmp12;
   $112 = ((($b0$172)) + 44|0);
   $113 = +HEAPF32[$112>>2];
   $114 = ((($window$075)) + 44|0);
   $115 = +HEAPF32[$114>>2];
   $$neg6 = $113 * $115;
   $116 = ((($b0$172)) + 36|0);
   $117 = +HEAPF32[$116>>2];
   $118 = ((($window$075)) + 36|0);
   $119 = +HEAPF32[$118>>2];
   $$neg5 = $117 * $119;
   $tmp10 = $$neg6 + $$neg5;
   $120 = $111 - $tmp10;
   $121 = ((($b0$172)) + 60|0);
   $122 = +HEAPF32[$121>>2];
   $123 = ((($window$075)) + 60|0);
   $124 = +HEAPF32[$123>>2];
   $$neg8 = $122 * $124;
   $125 = ((($b0$172)) + 52|0);
   $126 = +HEAPF32[$125>>2];
   $127 = ((($window$075)) + 52|0);
   $128 = +HEAPF32[$127>>2];
   $$neg7 = $126 * $128;
   $tmp = $$neg8 + $$neg7;
   $129 = $120 - $tmp;
   $130 = $129 * 3.0517578125E-5;
   $131 = (($48) + -32768)|0;
   $132 = $131 >>> 15;
   $133 = $132 << 15;
   $134 = (($131) - ($133))|0;
   $135 = $132 << 1;
   $ntom$268 = $48;$samples$267 = $samples$173;
   while(1) {
    HEAPF32[$samples$267>>2] = $130;
    $136 = ((($samples$267)) + 8|0);
    $137 = (($ntom$268) + -32768)|0;
    $138 = ($137|0)>(32767);
    if ($138) {
     $ntom$268 = $137;$samples$267 = $136;
    } else {
     break;
    }
   }
   $scevgep85 = ((($samples$173)) + 8|0);
   $scevgep86 = (($scevgep85) + ($135<<2)|0);
   $ntom$3 = $134;$samples$3 = $scevgep86;
  }
  $b0$2 = ((($b0$172)) + 64|0);
  $139 = (($j$074) + -1)|0;
  $140 = ((($window$075)) + 128|0);
  $141 = ($139|0)==(0);
  $142 = (($43) + ($ntom$3))|0;
  if ($141) {
   $$lcssa = $142;$samples$3$lcssa = $samples$3;
   break;
  } else {
   $48 = $142;$b0$172 = $b0$2;$j$074 = $139;$samples$173 = $samples$3;$window$075 = $140;
  }
 }
 $scevgep89 = ((($b0$0)) + 960|0);
 $scevgep90 = (($38) + ($46<<2)|0);
 $143 = ($$lcssa|0)>(32767);
 if ($143) {
  $scevgep87 = (($38) + ($45<<2)|0);
  $144 = ((($b0$0)) + 1032|0);
  $145 = +HEAPF32[$144>>2];
  $146 = ((($scevgep90)) + 136|0);
  $147 = +HEAPF32[$146>>2];
  $148 = $145 * $147;
  $149 = +HEAPF32[$scevgep88>>2];
  $150 = +HEAPF32[$scevgep87>>2];
  $151 = $149 * $150;
  $152 = $148 + $151;
  $153 = ((($b0$0)) + 1040|0);
  $154 = +HEAPF32[$153>>2];
  $155 = ((($scevgep90)) + 144|0);
  $156 = +HEAPF32[$155>>2];
  $157 = $154 * $156;
  $158 = $152 + $157;
  $159 = ((($b0$0)) + 1048|0);
  $160 = +HEAPF32[$159>>2];
  $161 = ((($scevgep90)) + 152|0);
  $162 = +HEAPF32[$161>>2];
  $163 = $160 * $162;
  $164 = $158 + $163;
  $165 = ((($b0$0)) + 1056|0);
  $166 = +HEAPF32[$165>>2];
  $167 = ((($scevgep90)) + 160|0);
  $168 = +HEAPF32[$167>>2];
  $169 = $166 * $168;
  $170 = $164 + $169;
  $171 = ((($b0$0)) + 1064|0);
  $172 = +HEAPF32[$171>>2];
  $173 = ((($scevgep90)) + 168|0);
  $174 = +HEAPF32[$173>>2];
  $175 = $172 * $174;
  $176 = $170 + $175;
  $177 = ((($b0$0)) + 1072|0);
  $178 = +HEAPF32[$177>>2];
  $179 = ((($scevgep90)) + 176|0);
  $180 = +HEAPF32[$179>>2];
  $181 = $178 * $180;
  $182 = $176 + $181;
  $183 = ((($b0$0)) + 1080|0);
  $184 = +HEAPF32[$183>>2];
  $185 = ((($scevgep90)) + 184|0);
  $186 = +HEAPF32[$185>>2];
  $187 = $184 * $186;
  $188 = $182 + $187;
  $189 = $188 * 3.0517578125E-5;
  $190 = (($$lcssa) + -32768)|0;
  $191 = $190 >>> 15;
  $192 = $191 << 15;
  $193 = $191 << 1;
  $194 = (($193) + 2)|0;
  $ntom$463 = $$lcssa;$samples$462 = $samples$3$lcssa;
  while(1) {
   HEAPF32[$samples$462>>2] = $189;
   $195 = ((($samples$462)) + 8|0);
   $196 = (($ntom$463) + -32768)|0;
   $197 = ($196|0)>(32767);
   if ($197) {
    $ntom$463 = $196;$samples$462 = $195;
   } else {
    break;
   }
  }
  $198 = (($190) - ($192))|0;
  $scevgep84 = (($samples$3$lcssa) + ($194<<2)|0);
  $ntom$5 = $198;$samples$5 = $scevgep84;
 } else {
  $ntom$5 = $$lcssa;$samples$5 = $samples$3$lcssa;
 }
 $199 = $bo1$0 << 1;
 $200 = (($scevgep90) + ($199<<2)|0);
 $201 = HEAP32[$42>>2]|0;
 $202 = (($201) + -32768)|0;
 $b0$1$pn56 = $scevgep88;$b0$361 = $scevgep89;$j$159 = 15;$ntom$658 = $ntom$5;$samples$657 = $samples$5;$window$260 = $200;
 while(1) {
  $203 = (($201) + ($ntom$658))|0;
  $204 = ($203|0)<(32768);
  if ($204) {
   $ntom$8 = $203;$samples$8 = $samples$657;
  } else {
   $205 = ((($b0$1$pn56)) + -56|0);
   $206 = +HEAPF32[$205>>2];
   $207 = ((($window$260)) + -12|0);
   $208 = +HEAPF32[$207>>2];
   $$neg17 = $206 * $208;
   $209 = ((($b0$1$pn56)) + -60|0);
   $210 = +HEAPF32[$209>>2];
   $211 = ((($window$260)) + -8|0);
   $212 = +HEAPF32[$211>>2];
   $$neg16 = $210 * $212;
   $tmp45 = $$neg17 + $$neg16;
   $tmp46 = -$tmp45;
   $213 = ((($b0$1$pn56)) + -48|0);
   $214 = +HEAPF32[$213>>2];
   $215 = ((($window$260)) + -20|0);
   $216 = +HEAPF32[$215>>2];
   $$neg19 = $214 * $216;
   $217 = ((($b0$1$pn56)) + -52|0);
   $218 = +HEAPF32[$217>>2];
   $219 = ((($window$260)) + -16|0);
   $220 = +HEAPF32[$219>>2];
   $$neg18 = $218 * $220;
   $tmp43 = $$neg19 + $$neg18;
   $221 = $tmp46 - $tmp43;
   $222 = ((($b0$1$pn56)) + -40|0);
   $223 = +HEAPF32[$222>>2];
   $224 = ((($window$260)) + -28|0);
   $225 = +HEAPF32[$224>>2];
   $$neg21 = $223 * $225;
   $226 = ((($b0$1$pn56)) + -44|0);
   $227 = +HEAPF32[$226>>2];
   $228 = ((($window$260)) + -24|0);
   $229 = +HEAPF32[$228>>2];
   $$neg20 = $227 * $229;
   $tmp41 = $$neg21 + $$neg20;
   $230 = $221 - $tmp41;
   $231 = ((($b0$1$pn56)) + -32|0);
   $232 = +HEAPF32[$231>>2];
   $233 = ((($window$260)) + -36|0);
   $234 = +HEAPF32[$233>>2];
   $$neg23 = $232 * $234;
   $235 = ((($b0$1$pn56)) + -36|0);
   $236 = +HEAPF32[$235>>2];
   $237 = ((($window$260)) + -32|0);
   $238 = +HEAPF32[$237>>2];
   $$neg22 = $236 * $238;
   $tmp39 = $$neg23 + $$neg22;
   $239 = $230 - $tmp39;
   $240 = ((($b0$1$pn56)) + -24|0);
   $241 = +HEAPF32[$240>>2];
   $242 = ((($window$260)) + -44|0);
   $243 = +HEAPF32[$242>>2];
   $$neg25 = $241 * $243;
   $244 = ((($b0$1$pn56)) + -28|0);
   $245 = +HEAPF32[$244>>2];
   $246 = ((($window$260)) + -40|0);
   $247 = +HEAPF32[$246>>2];
   $$neg24 = $245 * $247;
   $tmp37 = $$neg25 + $$neg24;
   $248 = $239 - $tmp37;
   $249 = ((($b0$1$pn56)) + -16|0);
   $250 = +HEAPF32[$249>>2];
   $251 = ((($window$260)) + -52|0);
   $252 = +HEAPF32[$251>>2];
   $$neg27 = $250 * $252;
   $253 = ((($b0$1$pn56)) + -20|0);
   $254 = +HEAPF32[$253>>2];
   $255 = ((($window$260)) + -48|0);
   $256 = +HEAPF32[$255>>2];
   $$neg26 = $254 * $256;
   $tmp35 = $$neg27 + $$neg26;
   $257 = $248 - $tmp35;
   $258 = ((($b0$1$pn56)) + -8|0);
   $259 = +HEAPF32[$258>>2];
   $260 = ((($window$260)) + -60|0);
   $261 = +HEAPF32[$260>>2];
   $$neg29 = $259 * $261;
   $262 = ((($b0$1$pn56)) + -12|0);
   $263 = +HEAPF32[$262>>2];
   $264 = ((($window$260)) + -56|0);
   $265 = +HEAPF32[$264>>2];
   $$neg28 = $263 * $265;
   $tmp33 = $$neg29 + $$neg28;
   $266 = $257 - $tmp33;
   $267 = ((($b0$1$pn56)) + -4|0);
   $268 = +HEAPF32[$267>>2];
   $269 = ((($window$260)) + -64|0);
   $270 = +HEAPF32[$269>>2];
   $$neg30 = $268 * $270;
   $271 = +HEAPF32[$b0$361>>2];
   $272 = ((($window$260)) + -4|0);
   $273 = +HEAPF32[$272>>2];
   $274 = $271 * $273;
   $tmp31 = $$neg30 + $274;
   $275 = $266 - $tmp31;
   $276 = $275 * 3.0517578125E-5;
   $277 = (($202) + ($ntom$658))|0;
   $278 = $277 >>> 15;
   $279 = $278 << 15;
   $280 = (($277) - ($279))|0;
   $281 = $278 << 1;
   $ntom$753 = $203;$samples$752 = $samples$657;
   while(1) {
    HEAPF32[$samples$752>>2] = $276;
    $282 = ((($samples$752)) + 8|0);
    $283 = (($ntom$753) + -32768)|0;
    $284 = ($283|0)>(32767);
    if ($284) {
     $ntom$753 = $283;$samples$752 = $282;
    } else {
     break;
    }
   }
   $scevgep = ((($samples$657)) + 8|0);
   $scevgep83 = (($scevgep) + ($281<<2)|0);
   $ntom$8 = $280;$samples$8 = $scevgep83;
  }
  $285 = (($j$159) + -1)|0;
  $286 = ((($window$260)) + -128|0);
  $b0$3 = ((($b0$361)) + -64|0);
  $287 = ($285|0)==(0);
  if ($287) {
   $ntom$8$lcssa = $ntom$8;$samples$8$lcssa = $samples$8;
   break;
  } else {
   $b0$1$pn56$phi = $b0$361;$b0$361 = $b0$3;$j$159 = $285;$ntom$658 = $ntom$8;$samples$657 = $samples$8;$window$260 = $286;$b0$1$pn56 = $b0$1$pn56$phi;
  }
 }
 $288 = (((($fr)) + 19352|0) + ($channel<<2)|0);
 HEAP32[$288>>2] = $ntom$8$lcssa;
 $289 = ($final|0)==(0);
 if ($289) {
  return 0;
 }
 $290 = HEAP32[$0>>2]|0;
 $291 = $samples$8$lcssa;
 $$neg48 = $9 ? -4 : 0;
 $292 = (($291) + ($$neg48))|0;
 $293 = (($292) - ($290))|0;
 HEAP32[$2>>2] = $293;
 return 0;
}
function _INT123_synth_ntom_s32($bandPtr,$channel,$fr,$final) {
 $bandPtr = $bandPtr|0;
 $channel = $channel|0;
 $fr = $fr|0;
 $final = $final|0;
 var $$lcssa = 0, $$neg = 0.0, $$neg16 = 0.0, $$neg17 = 0.0, $$neg18 = 0.0, $$neg19 = 0.0, $$neg2 = 0.0, $$neg20 = 0.0, $$neg21 = 0.0, $$neg22 = 0.0, $$neg23 = 0.0, $$neg24 = 0.0, $$neg25 = 0.0, $$neg26 = 0.0, $$neg27 = 0.0, $$neg28 = 0.0, $$neg29 = 0.0, $$neg3 = 0.0, $$neg30 = 0.0, $$neg4 = 0.0;
 var $$neg48 = 0, $$neg5 = 0.0, $$neg6 = 0.0, $$neg7 = 0.0, $$neg8 = 0.0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0.0, $102 = 0.0, $103 = 0, $104 = 0.0, $105 = 0, $106 = 0.0, $107 = 0, $108 = 0.0, $109 = 0;
 var $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0, $113 = 0.0, $114 = 0, $115 = 0.0, $116 = 0, $117 = 0.0, $118 = 0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0.0, $123 = 0, $124 = 0.0, $125 = 0, $126 = 0.0, $127 = 0;
 var $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0.0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0.0;
 var $164 = 0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0, $172 = 0.0, $173 = 0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0, $178 = 0.0, $179 = 0, $18 = 0, $180 = 0.0, $181 = 0.0;
 var $182 = 0.0, $183 = 0, $184 = 0.0, $185 = 0, $186 = 0.0, $187 = 0.0, $188 = 0.0, $189 = 0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0.0, $194 = 0.0, $195 = 0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0.0, $2 = 0;
 var $20 = 0, $200 = 0.0, $201 = 0, $202 = 0.0, $203 = 0, $204 = 0.0, $205 = 0.0, $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0;
 var $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0;
 var $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0.0, $242 = 0, $243 = 0.0, $244 = 0, $245 = 0.0, $246 = 0, $247 = 0.0, $248 = 0, $249 = 0.0, $25 = 0, $250 = 0, $251 = 0.0, $252 = 0, $253 = 0.0;
 var $254 = 0, $255 = 0.0, $256 = 0.0, $257 = 0, $258 = 0.0, $259 = 0, $26 = 0, $260 = 0.0, $261 = 0, $262 = 0.0, $263 = 0, $264 = 0.0, $265 = 0.0, $266 = 0, $267 = 0.0, $268 = 0, $269 = 0.0, $27 = 0, $270 = 0, $271 = 0.0;
 var $272 = 0, $273 = 0.0, $274 = 0.0, $275 = 0, $276 = 0.0, $277 = 0, $278 = 0.0, $279 = 0, $28 = 0, $280 = 0.0, $281 = 0, $282 = 0.0, $283 = 0.0, $284 = 0, $285 = 0.0, $286 = 0, $287 = 0.0, $288 = 0, $289 = 0.0, $29 = 0;
 var $290 = 0, $291 = 0.0, $292 = 0.0, $293 = 0, $294 = 0.0, $295 = 0, $296 = 0.0, $297 = 0, $298 = 0.0, $299 = 0, $3 = 0, $30 = 0, $300 = 0.0, $301 = 0.0, $302 = 0, $303 = 0.0, $304 = 0, $305 = 0.0, $306 = 0.0, $307 = 0;
 var $308 = 0.0, $309 = 0.0, $31 = 0, $310 = 0.0, $311 = 0.0, $312 = 0.0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0;
 var $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0;
 var $344 = 0, $345 = 0, $346 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0.0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0.0;
 var $69 = 0.0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0.0;
 var $87 = 0.0, $88 = 0, $89 = 0.0, $9 = 0, $90 = 0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0, $95 = 0.0, $96 = 0, $97 = 0.0, $98 = 0, $99 = 0.0, $b0$0 = 0, $b0$1$pn62 = 0, $b0$1$pn62$phi = 0, $b0$182 = 0, $b0$2 = 0, $b0$3 = 0;
 var $b0$364 = 0, $bo1$0 = 0, $buf$0 = 0, $clip$079 = 0, $clip$10 = 0, $clip$10$lcssa = 0, $clip$3 = 0, $clip$3$lcssa = 0, $clip$465$us = 0, $clip$465$us$us = 0, $clip$465$us90 = 0, $clip$6 = 0, $clip$758 = 0, $exitcond = 0, $exitcond148 = 0, $exitcond150 = 0, $j$080 = 0, $j$160 = 0, $ntom$0 = 0, $ntom$273 = 0;
 var $ntom$273$us = 0, $ntom$273$us$us = 0, $ntom$273$us93 = 0, $ntom$3 = 0, $ntom$466 = 0, $ntom$5 = 0, $ntom$659 = 0, $ntom$753 = 0, $ntom$753$us = 0, $ntom$753$us$us = 0, $ntom$753$us85 = 0, $ntom$8 = 0, $ntom$8$lcssa = 0, $samples$0$in = 0, $samples$183 = 0, $samples$274 = 0, $samples$274$us = 0, $samples$274$us$us = 0, $samples$274$us92 = 0, $samples$3 = 0;
 var $samples$3$lcssa = 0, $samples$467 = 0, $samples$467$us = 0, $samples$467$us$us = 0, $samples$467$us88 = 0, $samples$5 = 0, $samples$663 = 0, $samples$754 = 0, $samples$754$us = 0, $samples$754$us$us = 0, $samples$754$us84 = 0, $samples$8 = 0, $samples$8$lcssa = 0, $scevgep = 0, $scevgep139 = 0, $scevgep140 = 0, $scevgep141 = 0, $scevgep142 = 0, $scevgep143 = 0, $scevgep144 = 0;
 var $scevgep145 = 0, $scevgep146 = 0, $scevgep147 = 0, $scevgep149 = 0, $scevgep151 = 0, $scevgep152 = 0, $scevgep153 = 0, $scevgep154 = 0, $scevgep155 = 0, $scevgep156 = 0, $scevgep157 = 0, $scevgep158 = 0, $scevgep159 = 0, $scevgep160 = 0, $scevgep161 = 0, $scevgep162 = 0, $scevgep163 = 0, $tmp = 0.0, $tmp10 = 0.0, $tmp12 = 0.0;
 var $tmp14 = 0.0, $tmp31 = 0.0, $tmp33 = 0.0, $tmp35 = 0.0, $tmp37 = 0.0, $tmp39 = 0.0, $tmp41 = 0.0, $tmp43 = 0.0, $tmp45 = 0.0, $tmp46 = 0.0, $window$081 = 0, $window$261 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($fr)) + 9288|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = ((($fr)) + 9292|0);
  _INT123_do_equalizer($bandPtr,$channel,$8);
 }
 $9 = ($channel|0)!=(0);
 if ($9) {
  $18 = ((($4)) + 4|0);
  $19 = ((($fr)) + 9256|0);
  $20 = ((($fr)) + 19356|0);
  $21 = HEAP32[$20>>2]|0;
  $$phi$trans$insert = ((($fr)) + 9272|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $23 = $$pre;$buf$0 = $19;$ntom$0 = $21;$samples$0$in = $18;
 } else {
  $10 = ((($fr)) + 9272|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + 15)|0;
  $13 = $12 & 15;
  HEAP32[$10>>2] = $13;
  $14 = ((($fr)) + 9248|0);
  $15 = ((($fr)) + 19352|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ((($fr)) + 19356|0);
  HEAP32[$17>>2] = $16;
  $23 = $13;$buf$0 = $14;$ntom$0 = $16;$samples$0$in = $4;
 }
 $22 = $23 & 1;
 $24 = ($22|0)==(0);
 $25 = ((($buf$0)) + 4|0);
 $26 = HEAP32[$25>>2]|0;
 if ($24) {
  $32 = (($23) + 1)|0;
  $33 = HEAP32[$buf$0>>2]|0;
  $34 = (($33) + ($23<<2)|0);
  $35 = (($26) + ($23<<2)|0);
  $36 = ((($35)) + 4|0);
  _INT123_dct64($34,$36,$bandPtr);
  $b0$0 = $26;$bo1$0 = $32;
 } else {
  $27 = HEAP32[$buf$0>>2]|0;
  $28 = (($23) + 1)|0;
  $29 = $28 & 15;
  $30 = (($26) + ($29<<2)|0);
  $31 = (($27) + ($23<<2)|0);
  _INT123_dct64($30,$31,$bandPtr);
  $b0$0 = $27;$bo1$0 = $23;
 }
 $37 = ((($fr)) + 9284|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = ((($38)) + 64|0);
 $40 = (0 - ($bo1$0))|0;
 $41 = (($39) + ($40<<2)|0);
 $42 = ((($fr)) + 19360|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = (($43) + ($ntom$0))|0;
 $45 = (528 - ($bo1$0))|0;
 $scevgep160 = (($38) + ($45<<2)|0);
 $46 = (496 - ($bo1$0))|0;
 $scevgep162 = (($38) + ($46<<2)|0);
 $48 = $44;$b0$182 = $b0$0;$clip$079 = 0;$j$080 = 16;$samples$183 = $samples$0$in;$window$081 = $41;
 while(1) {
  $47 = ($48|0)<(32768);
  do {
   if ($47) {
    $clip$3 = $clip$079;$ntom$3 = $48;$samples$3 = $samples$183;
   } else {
    $49 = ((($b0$182)) + 8|0);
    $50 = +HEAPF32[$49>>2];
    $51 = ((($window$081)) + 8|0);
    $52 = +HEAPF32[$51>>2];
    $53 = $50 * $52;
    $54 = +HEAPF32[$b0$182>>2];
    $55 = +HEAPF32[$window$081>>2];
    $56 = $54 * $55;
    $57 = $53 + $56;
    $58 = ((($b0$182)) + 16|0);
    $59 = +HEAPF32[$58>>2];
    $60 = ((($window$081)) + 16|0);
    $61 = +HEAPF32[$60>>2];
    $62 = $59 * $61;
    $63 = $57 + $62;
    $64 = ((($b0$182)) + 24|0);
    $65 = +HEAPF32[$64>>2];
    $66 = ((($window$081)) + 24|0);
    $67 = +HEAPF32[$66>>2];
    $68 = $65 * $67;
    $69 = $63 + $68;
    $70 = ((($b0$182)) + 32|0);
    $71 = +HEAPF32[$70>>2];
    $72 = ((($window$081)) + 32|0);
    $73 = +HEAPF32[$72>>2];
    $74 = $71 * $73;
    $75 = $69 + $74;
    $76 = ((($b0$182)) + 40|0);
    $77 = +HEAPF32[$76>>2];
    $78 = ((($window$081)) + 40|0);
    $79 = +HEAPF32[$78>>2];
    $80 = $77 * $79;
    $81 = $75 + $80;
    $82 = ((($b0$182)) + 48|0);
    $83 = +HEAPF32[$82>>2];
    $84 = ((($window$081)) + 48|0);
    $85 = +HEAPF32[$84>>2];
    $86 = $83 * $85;
    $87 = $81 + $86;
    $88 = ((($b0$182)) + 56|0);
    $89 = +HEAPF32[$88>>2];
    $90 = ((($window$081)) + 56|0);
    $91 = +HEAPF32[$90>>2];
    $92 = $89 * $91;
    $93 = $87 + $92;
    $94 = ((($b0$182)) + 12|0);
    $95 = +HEAPF32[$94>>2];
    $96 = ((($window$081)) + 12|0);
    $97 = +HEAPF32[$96>>2];
    $$neg2 = $95 * $97;
    $98 = ((($b0$182)) + 4|0);
    $99 = +HEAPF32[$98>>2];
    $100 = ((($window$081)) + 4|0);
    $101 = +HEAPF32[$100>>2];
    $$neg = $99 * $101;
    $tmp14 = $$neg2 + $$neg;
    $102 = $93 - $tmp14;
    $103 = ((($b0$182)) + 28|0);
    $104 = +HEAPF32[$103>>2];
    $105 = ((($window$081)) + 28|0);
    $106 = +HEAPF32[$105>>2];
    $$neg4 = $104 * $106;
    $107 = ((($b0$182)) + 20|0);
    $108 = +HEAPF32[$107>>2];
    $109 = ((($window$081)) + 20|0);
    $110 = +HEAPF32[$109>>2];
    $$neg3 = $108 * $110;
    $tmp12 = $$neg4 + $$neg3;
    $111 = $102 - $tmp12;
    $112 = ((($b0$182)) + 44|0);
    $113 = +HEAPF32[$112>>2];
    $114 = ((($window$081)) + 44|0);
    $115 = +HEAPF32[$114>>2];
    $$neg6 = $113 * $115;
    $116 = ((($b0$182)) + 36|0);
    $117 = +HEAPF32[$116>>2];
    $118 = ((($window$081)) + 36|0);
    $119 = +HEAPF32[$118>>2];
    $$neg5 = $117 * $119;
    $tmp10 = $$neg6 + $$neg5;
    $120 = $111 - $tmp10;
    $121 = ((($b0$182)) + 60|0);
    $122 = +HEAPF32[$121>>2];
    $123 = ((($window$081)) + 60|0);
    $124 = +HEAPF32[$123>>2];
    $$neg8 = $122 * $124;
    $125 = ((($b0$182)) + 52|0);
    $126 = +HEAPF32[$125>>2];
    $127 = ((($window$081)) + 52|0);
    $128 = +HEAPF32[$127>>2];
    $$neg7 = $126 * $128;
    $tmp = $$neg8 + $$neg7;
    $129 = $120 - $tmp;
    $130 = $129 * 65536.0;
    $131 = $130;
    $132 = $131 > 2147483647.0;
    $133 = $130 < -2147483648.0;
    $134 = (~~(($130)));
    $135 = (($48) + -32768)|0;
    $136 = $135 >>> 15;
    $137 = $136 << 15;
    $138 = (($135) - ($137))|0;
    $139 = $136 << 1;
    if ($132) {
     $140 = (($clip$079) + 1)|0;
     if ($133) {
      $ntom$273$us$us = $48;$samples$274$us$us = $samples$183;
      while(1) {
       HEAP32[$samples$274$us$us>>2] = 2147483647;
       $141 = ((($samples$274$us$us)) + 8|0);
       $142 = (($ntom$273$us$us) + -32768)|0;
       $143 = ($142|0)>(32767);
       if ($143) {
        $ntom$273$us$us = $142;$samples$274$us$us = $141;
       } else {
        break;
       }
      }
      $scevgep158 = ((($samples$183)) + 8|0);
      $scevgep159 = (($scevgep158) + ($139<<2)|0);
      $154 = (($140) + ($136))|0;
      $clip$3 = $154;$ntom$3 = $138;$samples$3 = $scevgep159;
      break;
     } else {
      $ntom$273$us = $48;$samples$274$us = $samples$183;
      while(1) {
       HEAP32[$samples$274$us>>2] = 2147483647;
       $144 = ((($samples$274$us)) + 8|0);
       $145 = (($ntom$273$us) + -32768)|0;
       $146 = ($145|0)>(32767);
       if ($146) {
        $ntom$273$us = $145;$samples$274$us = $144;
       } else {
        break;
       }
      }
      $scevgep156 = ((($samples$183)) + 8|0);
      $scevgep157 = (($scevgep156) + ($139<<2)|0);
      $155 = (($140) + ($136))|0;
      $clip$3 = $155;$ntom$3 = $138;$samples$3 = $scevgep157;
      break;
     }
    }
    if (!($133)) {
     $ntom$273 = $48;$samples$274 = $samples$183;
     while(1) {
      HEAP32[$samples$274>>2] = $134;
      $151 = ((($samples$274)) + 8|0);
      $152 = (($ntom$273) + -32768)|0;
      $153 = ($152|0)>(32767);
      if ($153) {
       $ntom$273 = $152;$samples$274 = $151;
      } else {
       break;
      }
     }
     $scevgep152 = ((($samples$183)) + 8|0);
     $scevgep153 = (($scevgep152) + ($139<<2)|0);
     $clip$3 = $clip$079;$ntom$3 = $138;$samples$3 = $scevgep153;
     break;
    }
    $147 = (($clip$079) + 1)|0;
    $ntom$273$us93 = $48;$samples$274$us92 = $samples$183;
    while(1) {
     HEAP32[$samples$274$us92>>2] = -2147483648;
     $148 = ((($samples$274$us92)) + 8|0);
     $149 = (($ntom$273$us93) + -32768)|0;
     $150 = ($149|0)>(32767);
     if ($150) {
      $ntom$273$us93 = $149;$samples$274$us92 = $148;
     } else {
      break;
     }
    }
    $scevgep154 = ((($samples$183)) + 8|0);
    $scevgep155 = (($scevgep154) + ($139<<2)|0);
    $156 = (($147) + ($136))|0;
    $clip$3 = $156;$ntom$3 = $138;$samples$3 = $scevgep155;
   }
  } while(0);
  $b0$2 = ((($b0$182)) + 64|0);
  $157 = (($j$080) + -1)|0;
  $158 = ((($window$081)) + 128|0);
  $159 = ($157|0)==(0);
  $160 = (($43) + ($ntom$3))|0;
  if ($159) {
   $$lcssa = $160;$clip$3$lcssa = $clip$3;$samples$3$lcssa = $samples$3;
   break;
  } else {
   $48 = $160;$b0$182 = $b0$2;$clip$079 = $clip$3;$j$080 = $157;$samples$183 = $samples$3;$window$081 = $158;
  }
 }
 $scevgep161 = ((($b0$0)) + 1024|0);
 $scevgep163 = ((($b0$0)) + 960|0);
 $161 = ($$lcssa|0)>(32767);
 do {
  if ($161) {
   $162 = ((($b0$0)) + 1032|0);
   $163 = +HEAPF32[$162>>2];
   $164 = ((($scevgep162)) + 136|0);
   $165 = +HEAPF32[$164>>2];
   $166 = $163 * $165;
   $167 = +HEAPF32[$scevgep161>>2];
   $168 = +HEAPF32[$scevgep160>>2];
   $169 = $167 * $168;
   $170 = $166 + $169;
   $171 = ((($b0$0)) + 1040|0);
   $172 = +HEAPF32[$171>>2];
   $173 = ((($scevgep162)) + 144|0);
   $174 = +HEAPF32[$173>>2];
   $175 = $172 * $174;
   $176 = $170 + $175;
   $177 = ((($b0$0)) + 1048|0);
   $178 = +HEAPF32[$177>>2];
   $179 = ((($scevgep162)) + 152|0);
   $180 = +HEAPF32[$179>>2];
   $181 = $178 * $180;
   $182 = $176 + $181;
   $183 = ((($b0$0)) + 1056|0);
   $184 = +HEAPF32[$183>>2];
   $185 = ((($scevgep162)) + 160|0);
   $186 = +HEAPF32[$185>>2];
   $187 = $184 * $186;
   $188 = $182 + $187;
   $189 = ((($b0$0)) + 1064|0);
   $190 = +HEAPF32[$189>>2];
   $191 = ((($scevgep162)) + 168|0);
   $192 = +HEAPF32[$191>>2];
   $193 = $190 * $192;
   $194 = $188 + $193;
   $195 = ((($b0$0)) + 1072|0);
   $196 = +HEAPF32[$195>>2];
   $197 = ((($scevgep162)) + 176|0);
   $198 = +HEAPF32[$197>>2];
   $199 = $196 * $198;
   $200 = $194 + $199;
   $201 = ((($b0$0)) + 1080|0);
   $202 = +HEAPF32[$201>>2];
   $203 = ((($scevgep162)) + 184|0);
   $204 = +HEAPF32[$203>>2];
   $205 = $202 * $204;
   $206 = $200 + $205;
   $207 = $206 * 65536.0;
   $208 = $207;
   $209 = $208 > 2147483647.0;
   $210 = $207 < -2147483648.0;
   $211 = (~~(($207)));
   $212 = (($$lcssa) + -32768)|0;
   $213 = $212 >>> 15;
   $214 = $213 << 15;
   $215 = $213 << 1;
   $216 = (($215) + 2)|0;
   if ($209) {
    $217 = (($clip$3$lcssa) + ($213))|0;
    $218 = (($217) + 1)|0;
    if ($210) {
     $clip$465$us$us = $clip$3$lcssa;$samples$467$us$us = $samples$3$lcssa;
     while(1) {
      HEAP32[$samples$467$us$us>>2] = 2147483647;
      $219 = (($clip$465$us$us) + 1)|0;
      $220 = ((($samples$467$us$us)) + 8|0);
      $exitcond = ($219|0)==($218|0);
      if ($exitcond) {
       break;
      } else {
       $clip$465$us$us = $219;$samples$467$us$us = $220;
      }
     }
     $230 = (($212) - ($214))|0;
     $scevgep146 = (($samples$3$lcssa) + ($216<<2)|0);
     $clip$6 = $218;$ntom$5 = $230;$samples$5 = $scevgep146;
     break;
    } else {
     $clip$465$us = $clip$3$lcssa;$samples$467$us = $samples$3$lcssa;
     while(1) {
      HEAP32[$samples$467$us>>2] = 2147483647;
      $221 = (($clip$465$us) + 1)|0;
      $222 = ((($samples$467$us)) + 8|0);
      $exitcond148 = ($221|0)==($218|0);
      if ($exitcond148) {
       break;
      } else {
       $clip$465$us = $221;$samples$467$us = $222;
      }
     }
     $231 = (($212) - ($214))|0;
     $scevgep147 = (($samples$3$lcssa) + ($216<<2)|0);
     $clip$6 = $218;$ntom$5 = $231;$samples$5 = $scevgep147;
     break;
    }
   }
   if (!($210)) {
    $ntom$466 = $$lcssa;$samples$467 = $samples$3$lcssa;
    while(1) {
     HEAP32[$samples$467>>2] = $211;
     $227 = ((($samples$467)) + 8|0);
     $228 = (($ntom$466) + -32768)|0;
     $229 = ($228|0)>(32767);
     if ($229) {
      $ntom$466 = $228;$samples$467 = $227;
     } else {
      break;
     }
    }
    $233 = (($212) - ($214))|0;
    $scevgep151 = (($samples$3$lcssa) + ($216<<2)|0);
    $clip$6 = $clip$3$lcssa;$ntom$5 = $233;$samples$5 = $scevgep151;
    break;
   }
   $223 = (($clip$3$lcssa) + ($213))|0;
   $224 = (($223) + 1)|0;
   $clip$465$us90 = $clip$3$lcssa;$samples$467$us88 = $samples$3$lcssa;
   while(1) {
    HEAP32[$samples$467$us88>>2] = -2147483648;
    $225 = (($clip$465$us90) + 1)|0;
    $226 = ((($samples$467$us88)) + 8|0);
    $exitcond150 = ($225|0)==($224|0);
    if ($exitcond150) {
     break;
    } else {
     $clip$465$us90 = $225;$samples$467$us88 = $226;
    }
   }
   $232 = (($212) - ($214))|0;
   $scevgep149 = (($samples$3$lcssa) + ($216<<2)|0);
   $clip$6 = $224;$ntom$5 = $232;$samples$5 = $scevgep149;
  } else {
   $clip$6 = $clip$3$lcssa;$ntom$5 = $$lcssa;$samples$5 = $samples$3$lcssa;
  }
 } while(0);
 $234 = $bo1$0 << 1;
 $235 = (($scevgep162) + ($234<<2)|0);
 $236 = HEAP32[$42>>2]|0;
 $237 = (($236) + -32768)|0;
 $b0$1$pn62 = $scevgep161;$b0$364 = $scevgep163;$clip$758 = $clip$6;$j$160 = 15;$ntom$659 = $ntom$5;$samples$663 = $samples$5;$window$261 = $235;
 while(1) {
  $238 = (($236) + ($ntom$659))|0;
  $239 = ($238|0)<(32768);
  do {
   if ($239) {
    $clip$10 = $clip$758;$ntom$8 = $238;$samples$8 = $samples$663;
   } else {
    $240 = ((($b0$1$pn62)) + -56|0);
    $241 = +HEAPF32[$240>>2];
    $242 = ((($window$261)) + -12|0);
    $243 = +HEAPF32[$242>>2];
    $$neg17 = $241 * $243;
    $244 = ((($b0$1$pn62)) + -60|0);
    $245 = +HEAPF32[$244>>2];
    $246 = ((($window$261)) + -8|0);
    $247 = +HEAPF32[$246>>2];
    $$neg16 = $245 * $247;
    $tmp45 = $$neg17 + $$neg16;
    $tmp46 = -$tmp45;
    $248 = ((($b0$1$pn62)) + -48|0);
    $249 = +HEAPF32[$248>>2];
    $250 = ((($window$261)) + -20|0);
    $251 = +HEAPF32[$250>>2];
    $$neg19 = $249 * $251;
    $252 = ((($b0$1$pn62)) + -52|0);
    $253 = +HEAPF32[$252>>2];
    $254 = ((($window$261)) + -16|0);
    $255 = +HEAPF32[$254>>2];
    $$neg18 = $253 * $255;
    $tmp43 = $$neg19 + $$neg18;
    $256 = $tmp46 - $tmp43;
    $257 = ((($b0$1$pn62)) + -40|0);
    $258 = +HEAPF32[$257>>2];
    $259 = ((($window$261)) + -28|0);
    $260 = +HEAPF32[$259>>2];
    $$neg21 = $258 * $260;
    $261 = ((($b0$1$pn62)) + -44|0);
    $262 = +HEAPF32[$261>>2];
    $263 = ((($window$261)) + -24|0);
    $264 = +HEAPF32[$263>>2];
    $$neg20 = $262 * $264;
    $tmp41 = $$neg21 + $$neg20;
    $265 = $256 - $tmp41;
    $266 = ((($b0$1$pn62)) + -32|0);
    $267 = +HEAPF32[$266>>2];
    $268 = ((($window$261)) + -36|0);
    $269 = +HEAPF32[$268>>2];
    $$neg23 = $267 * $269;
    $270 = ((($b0$1$pn62)) + -36|0);
    $271 = +HEAPF32[$270>>2];
    $272 = ((($window$261)) + -32|0);
    $273 = +HEAPF32[$272>>2];
    $$neg22 = $271 * $273;
    $tmp39 = $$neg23 + $$neg22;
    $274 = $265 - $tmp39;
    $275 = ((($b0$1$pn62)) + -24|0);
    $276 = +HEAPF32[$275>>2];
    $277 = ((($window$261)) + -44|0);
    $278 = +HEAPF32[$277>>2];
    $$neg25 = $276 * $278;
    $279 = ((($b0$1$pn62)) + -28|0);
    $280 = +HEAPF32[$279>>2];
    $281 = ((($window$261)) + -40|0);
    $282 = +HEAPF32[$281>>2];
    $$neg24 = $280 * $282;
    $tmp37 = $$neg25 + $$neg24;
    $283 = $274 - $tmp37;
    $284 = ((($b0$1$pn62)) + -16|0);
    $285 = +HEAPF32[$284>>2];
    $286 = ((($window$261)) + -52|0);
    $287 = +HEAPF32[$286>>2];
    $$neg27 = $285 * $287;
    $288 = ((($b0$1$pn62)) + -20|0);
    $289 = +HEAPF32[$288>>2];
    $290 = ((($window$261)) + -48|0);
    $291 = +HEAPF32[$290>>2];
    $$neg26 = $289 * $291;
    $tmp35 = $$neg27 + $$neg26;
    $292 = $283 - $tmp35;
    $293 = ((($b0$1$pn62)) + -8|0);
    $294 = +HEAPF32[$293>>2];
    $295 = ((($window$261)) + -60|0);
    $296 = +HEAPF32[$295>>2];
    $$neg29 = $294 * $296;
    $297 = ((($b0$1$pn62)) + -12|0);
    $298 = +HEAPF32[$297>>2];
    $299 = ((($window$261)) + -56|0);
    $300 = +HEAPF32[$299>>2];
    $$neg28 = $298 * $300;
    $tmp33 = $$neg29 + $$neg28;
    $301 = $292 - $tmp33;
    $302 = ((($b0$1$pn62)) + -4|0);
    $303 = +HEAPF32[$302>>2];
    $304 = ((($window$261)) + -64|0);
    $305 = +HEAPF32[$304>>2];
    $$neg30 = $303 * $305;
    $306 = +HEAPF32[$b0$364>>2];
    $307 = ((($window$261)) + -4|0);
    $308 = +HEAPF32[$307>>2];
    $309 = $306 * $308;
    $tmp31 = $$neg30 + $309;
    $310 = $301 - $tmp31;
    $311 = $310 * 65536.0;
    $312 = $311;
    $313 = $312 > 2147483647.0;
    $314 = $311 < -2147483648.0;
    $315 = (~~(($311)));
    $316 = (($237) + ($ntom$659))|0;
    $317 = $316 >>> 15;
    $318 = $317 << 15;
    $319 = (($316) - ($318))|0;
    $320 = $317 << 1;
    if ($313) {
     $321 = (($clip$758) + 1)|0;
     if ($314) {
      $ntom$753$us$us = $238;$samples$754$us$us = $samples$663;
      while(1) {
       HEAP32[$samples$754$us$us>>2] = 2147483647;
       $322 = ((($samples$754$us$us)) + 8|0);
       $323 = (($ntom$753$us$us) + -32768)|0;
       $324 = ($323|0)>(32767);
       if ($324) {
        $ntom$753$us$us = $323;$samples$754$us$us = $322;
       } else {
        break;
       }
      }
      $scevgep144 = ((($samples$663)) + 8|0);
      $scevgep145 = (($scevgep144) + ($320<<2)|0);
      $335 = (($321) + ($317))|0;
      $clip$10 = $335;$ntom$8 = $319;$samples$8 = $scevgep145;
      break;
     } else {
      $ntom$753$us = $238;$samples$754$us = $samples$663;
      while(1) {
       HEAP32[$samples$754$us>>2] = 2147483647;
       $325 = ((($samples$754$us)) + 8|0);
       $326 = (($ntom$753$us) + -32768)|0;
       $327 = ($326|0)>(32767);
       if ($327) {
        $ntom$753$us = $326;$samples$754$us = $325;
       } else {
        break;
       }
      }
      $scevgep142 = ((($samples$663)) + 8|0);
      $scevgep143 = (($scevgep142) + ($320<<2)|0);
      $336 = (($321) + ($317))|0;
      $clip$10 = $336;$ntom$8 = $319;$samples$8 = $scevgep143;
      break;
     }
    }
    if (!($314)) {
     $ntom$753 = $238;$samples$754 = $samples$663;
     while(1) {
      HEAP32[$samples$754>>2] = $315;
      $332 = ((($samples$754)) + 8|0);
      $333 = (($ntom$753) + -32768)|0;
      $334 = ($333|0)>(32767);
      if ($334) {
       $ntom$753 = $333;$samples$754 = $332;
      } else {
       break;
      }
     }
     $scevgep = ((($samples$663)) + 8|0);
     $scevgep139 = (($scevgep) + ($320<<2)|0);
     $clip$10 = $clip$758;$ntom$8 = $319;$samples$8 = $scevgep139;
     break;
    }
    $328 = (($clip$758) + 1)|0;
    $ntom$753$us85 = $238;$samples$754$us84 = $samples$663;
    while(1) {
     HEAP32[$samples$754$us84>>2] = -2147483648;
     $329 = ((($samples$754$us84)) + 8|0);
     $330 = (($ntom$753$us85) + -32768)|0;
     $331 = ($330|0)>(32767);
     if ($331) {
      $ntom$753$us85 = $330;$samples$754$us84 = $329;
     } else {
      break;
     }
    }
    $scevgep140 = ((($samples$663)) + 8|0);
    $scevgep141 = (($scevgep140) + ($320<<2)|0);
    $337 = (($328) + ($317))|0;
    $clip$10 = $337;$ntom$8 = $319;$samples$8 = $scevgep141;
   }
  } while(0);
  $338 = (($j$160) + -1)|0;
  $339 = ((($window$261)) + -128|0);
  $b0$3 = ((($b0$364)) + -64|0);
  $340 = ($338|0)==(0);
  if ($340) {
   $clip$10$lcssa = $clip$10;$ntom$8$lcssa = $ntom$8;$samples$8$lcssa = $samples$8;
   break;
  } else {
   $b0$1$pn62$phi = $b0$364;$b0$364 = $b0$3;$clip$758 = $clip$10;$j$160 = $338;$ntom$659 = $ntom$8;$samples$663 = $samples$8;$window$261 = $339;$b0$1$pn62 = $b0$1$pn62$phi;
  }
 }
 $341 = (((($fr)) + 19352|0) + ($channel<<2)|0);
 HEAP32[$341>>2] = $ntom$8$lcssa;
 $342 = ($final|0)==(0);
 if ($342) {
  return ($clip$10$lcssa|0);
 }
 $343 = HEAP32[$0>>2]|0;
 $344 = $samples$8$lcssa;
 $$neg48 = $9 ? -4 : 0;
 $345 = (($344) + ($$neg48))|0;
 $346 = (($345) - ($343))|0;
 HEAP32[$2>>2] = $346;
 return ($clip$10$lcssa|0);
}
function _INT123_make_conv16to8_table($fr) {
 $fr = $fr|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0, $25 = 0;
 var $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $c1$0 = 0;
 var $exitcond = 0, $exitcond15 = 0, $exitcond16 = 0, $exitcond17 = 0, $i$08 = 0, $i$110 = 0, $i$211 = 0, $i$39 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_ptr4 = 0, $vararg_ptr5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($fr)) + 27876|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 9588|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 do {
  if ($4) {
   $5 = (_malloc(8192)|0);
   HEAP32[$2>>2] = $5;
   $6 = ($5|0)==(0|0);
   if (!($6)) {
    $13 = ((($5)) + 4096|0);
    $14 = ((($fr)) + 9592|0);
    HEAP32[$14>>2] = $13;
    break;
   }
   $7 = ((($fr)) + 28376|0);
   HEAP32[$7>>2] = 4;
   $8 = ((($fr)) + 28068|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = $9 & 32;
   $11 = ($10|0)==(0);
   if (!($11)) {
    $$0 = -1;
    STACKTOP = sp;return ($$0|0);
   }
   $12 = HEAP32[37544]|0;
   HEAP32[$vararg_buffer>>2] = 253;
   (_fprintf($12,514898,$vararg_buffer)|0);
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 switch ($1|0) {
 case 4:  {
  $15 = ((($fr)) + 28068|0);
  $16 = HEAP32[37544]|0;
  $17 = ((($fr)) + 9592|0);
  $i$08 = -4096;
  while(1) {
   $18 = ($i$08|0)<(0);
   $19 = (+($i$08|0));
   $20 = $19 * 0.062255859375;
   if ($18) {
    $21 = 1.0 - $20;
    $22 = (+Math_log((+$21)));
    $23 = $22 * 22.902783774112294;
    $24 = (~~(($23)));
    $25 = (127 - ($24))|0;
    $c1$0 = $25;
   } else {
    $26 = $20 + 1.0;
    $27 = (+Math_log((+$26)));
    $28 = $27 * 22.902783774112294;
    $29 = (~~(($28)));
    $30 = (255 - ($29))|0;
    $c1$0 = $30;
   }
   $31 = ($c1$0>>>0)>(255);
   if ($31) {
    $32 = HEAP32[$15>>2]|0;
    $33 = $32 & 32;
    $34 = ($33|0)==(0);
    if ($34) {
     HEAP32[$vararg_buffer1>>2] = 269;
     $vararg_ptr4 = ((($vararg_buffer1)) + 4|0);
     HEAP32[$vararg_ptr4>>2] = $i$08;
     $vararg_ptr5 = ((($vararg_buffer1)) + 8|0);
     HEAP32[$vararg_ptr5>>2] = $c1$0;
     (_fprintf($16,514961,$vararg_buffer1)|0);
    }
   }
   $35 = ($c1$0|0)==(0);
   $36 = $c1$0&255;
   $37 = $35 ? 2 : $36;
   $38 = HEAP32[$17>>2]|0;
   $39 = (($38) + ($i$08)|0);
   HEAP8[$39>>0] = $37;
   $40 = (($i$08) + 1)|0;
   $exitcond = ($40|0)==(4096);
   if ($exitcond) {
    $$0 = 0;
    break;
   } else {
    $i$08 = $40;
   }
  }
  STACKTOP = sp;return ($$0|0);
  break;
 }
 case 130:  {
  $42 = ((($fr)) + 9592|0);
  $i$110 = -4096;
  while(1) {
   $44 = $i$110 >>> 5;
   $45 = $44&255;
   $46 = HEAP32[$42>>2]|0;
   $47 = (($46) + ($i$110)|0);
   HEAP8[$47>>0] = $45;
   $48 = (($i$110) + 1)|0;
   $exitcond16 = ($48|0)==(4096);
   if ($exitcond16) {
    $$0 = 0;
    break;
   } else {
    $i$110 = $48;
   }
  }
  STACKTOP = sp;return ($$0|0);
  break;
 }
 case 1:  {
  $41 = ((($fr)) + 9592|0);
  $i$211 = -4096;
  while(1) {
   $49 = $i$211 >>> 5;
   $50 = (($49) + 128)|0;
   $51 = $50&255;
   $52 = HEAP32[$41>>2]|0;
   $53 = (($52) + ($i$211)|0);
   HEAP8[$53>>0] = $51;
   $54 = (($i$211) + 1)|0;
   $exitcond17 = ($54|0)==(4096);
   if ($exitcond17) {
    $$0 = 0;
    break;
   } else {
    $i$211 = $54;
   }
  }
  STACKTOP = sp;return ($$0|0);
  break;
 }
 default: {
  $43 = ((($fr)) + 9592|0);
  $i$39 = -4096;
  while(1) {
   $55 = HEAP32[$43>>2]|0;
   $56 = (($55) + ($i$39)|0);
   HEAP8[$56>>0] = 0;
   $57 = (($i$39) + 1)|0;
   $exitcond15 = ($57|0)==(4096);
   if ($exitcond15) {
    $$0 = 0;
    break;
   } else {
    $i$39 = $57;
   }
  }
  STACKTOP = sp;return ($$0|0);
 }
 }
 return (0)|0;
}
function _INT123_init_layer3_gainpow2($fr,$i) {
 $fr = $fr|0;
 $i = $i|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $exp2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($i) + 210)|0;
 $1 = (+($0|0));
 $2 = $1 * -0.25;
 $exp2 = (+_exp2($2));
 $3 = $exp2;
 return (+$3);
}
function _INT123_init_layer3_stuff($fr,$gainpow2) {
 $fr = $fr|0;
 $gainpow2 = $gainpow2|0;
 var $$ = 0, $$1 = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $exitcond10 = 0, $exitcond8 = 0, $exitcond9 = 0, $i$07 = 0, $i$14 = 0, $i$25 = 0, $j$06 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $i$07 = -256;
 while(1) {
  $1 = (+FUNCTION_TABLE_dii[$gainpow2 & 8191]($fr,$i$07));
  $2 = (($i$07) + 256)|0;
  $3 = (((($fr)) + 10928|0) + ($2<<2)|0);
  HEAPF32[$3>>2] = $1;
  $4 = (($i$07) + 1)|0;
  $exitcond10 = ($4|0)==(122);
  if ($exitcond10) {
   break;
  } else {
   $i$07 = $4;
  }
 }
 $0 = ((($fr)) + 19668|0);
 $j$06 = 0;
 while(1) {
  $i$14 = 0;
  while(1) {
   $5 = ((161928 + (($j$06*110)|0)|0) + ($i$14<<1)|0);
   $6 = HEAP16[$5>>1]|0;
   $7 = $6&65535;
   $8 = (($7) + 7)|0;
   $9 = (($8>>>0) / 18)&-1;
   $10 = (($9) + 1)|0;
   $11 = ((((($fr)) + 9596|0) + (($j$06*92)|0)|0) + ($i$14<<2)|0);
   HEAP32[$11>>2] = $10;
   $12 = HEAP32[$0>>2]|0;
   $13 = ($9|0)<($12|0);
   $$ = $13 ? $10 : $12;
   HEAP32[$11>>2] = $$;
   $14 = (($i$14) + 1)|0;
   $exitcond = ($14|0)==(23);
   if ($exitcond) {
    $i$25 = 0;
    break;
   } else {
    $i$14 = $14;
   }
  }
  while(1) {
   $15 = ((((161928 + (($j$06*110)|0)|0)) + 68|0) + ($i$25<<1)|0);
   $16 = HEAP16[$15>>1]|0;
   $17 = $16&65535;
   $18 = (($17) + -1)|0;
   $19 = (($18|0) / 18)&-1;
   $20 = (($19) + 1)|0;
   $21 = ((((($fr)) + 10424|0) + (($j$06*56)|0)|0) + ($i$25<<2)|0);
   HEAP32[$21>>2] = $20;
   $22 = HEAP32[$0>>2]|0;
   $23 = ($19|0)<($22|0);
   $$1 = $23 ? $20 : $22;
   HEAP32[$21>>2] = $$1;
   $24 = (($i$25) + 1)|0;
   $exitcond8 = ($24|0)==(14);
   if ($exitcond8) {
    break;
   } else {
    $i$25 = $24;
   }
  }
  $25 = (($j$06) + 1)|0;
  $exitcond9 = ($25|0)==(9);
  if ($exitcond9) {
   break;
  } else {
   $j$06 = $25;
  }
 }
 return;
}
function _INT123_init_layer12_table($fr,$table,$m) {
 $fr = $fr|0;
 $table = $table|0;
 $m = $m|0;
 var $$01 = 0, $0 = 0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $exitcond = 0, $exp2 = 0.0, $i$02 = 0, $j$03 = 0, $scevgep = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (3024 + ($m<<3)|0);
 $1 = +HEAPF64[$0>>3];
 $$01 = $table;$i$02 = 0;$j$03 = 3;
 while(1) {
  $2 = (+($j$03|0));
  $3 = $2 * 0.33333333333333331;
  $exp2 = (+_exp2($3));
  $4 = $exp2 * $1;
  $5 = $4;
  $6 = ((($$01)) + 4|0);
  HEAPF32[$$01>>2] = $5;
  $7 = (($i$02) + 1)|0;
  $8 = (($j$03) + -1)|0;
  $exitcond = ($7|0)==(63);
  if ($exitcond) {
   break;
  } else {
   $$01 = $6;$i$02 = $7;$j$03 = $8;
  }
 }
 $scevgep = ((($table)) + 252|0);
 return ($scevgep|0);
}
function _INT123_init_layer12_stuff($fr,$init_table) {
 $fr = $fr|0;
 $init_table = $init_table|0;
 var $0 = 0, $1 = 0, $2 = 0, $exitcond = 0, $k$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $k$01 = 0;
 while(1) {
  $0 = (((($fr)) + 12440|0) + ($k$01<<8)|0);
  $1 = (FUNCTION_TABLE_iiii[$init_table & 8191]($fr,$0,$k$01)|0);
  HEAPF32[$1>>2] = 0.0;
  $2 = (($k$01) + 1)|0;
  $exitcond = ($2|0)==(27);
  if ($exitcond) {
   break;
  } else {
   $k$01 = $2;
  }
 }
 return;
}
function _INT123_make_decode_tables($fr) {
 $fr = $fr|0;
 var $$idx$0 = 0, $$idx$2 = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0.0, $42 = 0;
 var $43 = 0, $44 = 0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $exitcond11 = 0, $i$11 = 0, $idx$07 = 0, $idx$23 = 0, $j$06 = 0, $j$12 = 0, $scaleval$08 = 0.0, $scaleval$1 = 0.0, $scaleval$24 = 0.0, $scaleval$3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 19808|0);
 $1 = +HEAPF64[$0>>3];
 $2 = $1 < 0.0;
 if ($2) {
  $3 = ((($fr)) + 28344|0);
  $4 = +HEAPF64[$3>>3];
  $6 = $4;
 } else {
  $6 = $1;
 }
 $5 = $6 * -0.5;
 $7 = ((($fr)) + 9284|0);
 $idx$07 = 0;$j$06 = 0;$scaleval$08 = $5;
 while(1) {
  $8 = ($idx$07|0)<(528);
  if ($8) {
   $9 = (36996 + ($j$06<<2)|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = (+($10|0));
   $12 = $11 * $scaleval$08;
   $13 = $12;
   $14 = HEAP32[$7>>2]|0;
   $15 = (($14) + ($idx$07<<2)|0);
   HEAPF32[$15>>2] = $13;
   $16 = (($idx$07) + 16)|0;
   $17 = (($14) + ($16<<2)|0);
   HEAPF32[$17>>2] = $13;
  }
  $18 = (($j$06|0) % 32)&-1;
  $19 = ($18|0)==(31);
  $20 = (($idx$07) + -1023)|0;
  $$idx$0 = $19 ? $20 : $idx$07;
  $21 = (($j$06|0) % 64)&-1;
  $22 = ($21|0)==(63);
  $23 = -$scaleval$08;
  $scaleval$1 = $22 ? $23 : $scaleval$08;
  $24 = (($j$06) + 1)|0;
  $25 = (($$idx$0) + 32)|0;
  $exitcond11 = ($24|0)==(256);
  if ($exitcond11) {
   $i$11 = 256;$idx$23 = $25;$j$12 = 256;$scaleval$24 = $scaleval$1;
   break;
  } else {
   $idx$07 = $25;$j$06 = $24;$scaleval$08 = $scaleval$1;
  }
 }
 while(1) {
  $26 = ($idx$23|0)<(528);
  if ($26) {
   $27 = (36996 + ($j$12<<2)|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = (+($28|0));
   $30 = $29 * $scaleval$24;
   $31 = $30;
   $32 = HEAP32[$7>>2]|0;
   $33 = (($32) + ($idx$23<<2)|0);
   HEAPF32[$33>>2] = $31;
   $34 = (($idx$23) + 16)|0;
   $35 = (($32) + ($34<<2)|0);
   HEAPF32[$35>>2] = $31;
  }
  $36 = (($i$11|0) % 32)&-1;
  $37 = ($36|0)==(31);
  $38 = (($idx$23) + -1023)|0;
  $$idx$2 = $37 ? $38 : $idx$23;
  $39 = (($i$11|0) % 64)&-1;
  $40 = ($39|0)==(63);
  $41 = -$scaleval$24;
  $scaleval$3 = $40 ? $41 : $scaleval$24;
  $42 = (($i$11) + 1)|0;
  $43 = (($j$12) + -1)|0;
  $44 = (($$idx$2) + 32)|0;
  $exitcond = ($42|0)==(512);
  if ($exitcond) {
   break;
  } else {
   $i$11 = $42;$idx$23 = $44;$j$12 = $43;$scaleval$24 = $scaleval$3;
  }
 }
 return;
}
function _INT123_frame_cpu_opt($fr,$cpu) {
 $fr = $fr|0;
 $cpu = $cpu|0;
 var $$cmp = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, $vararg_buffer = 0, $vararg_buffer3 = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $0 = (_INT123_dectype($cpu)|0);
 $1 = ((($fr)) + 19364|0);
 _memcpy(($1|0),(38024|0),256)|0;
 $$cmp = ($0>>>0)>(1);
 if ($$cmp) {
  $2 = ((($fr)) + 28068|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = $3 & 32;
  $5 = ($4|0)==(0);
  if ($5) {
   $8 = HEAP32[37544]|0;
   HEAP32[$vararg_buffer>>2] = 460;
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
   HEAP32[$vararg_ptr1>>2] = $0;
   $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
   HEAP32[$vararg_ptr2>>2] = 1;
   (_fprintf($8,515152,$vararg_buffer)|0);
   $$pre = HEAP32[$1>>2]|0;
   $phitmp = ($$pre|0)==(2094|0);
   $9 = ((($fr)) + 19620|0);
   HEAP32[$9>>2] = 1;
   $10 = ((($fr)) + 19624|0);
   HEAP32[$10>>2] = 1;
   if (!($phitmp)) {
    $11 = ((($fr)) + 19368|0);
    HEAP32[$11>>2] = 4267;
    $12 = ((($fr)) + 19560|0);
    HEAP32[$12>>2] = 4271;
    $13 = ((($fr)) + 19496|0);
    HEAP32[$13>>2] = 4272;
   }
  } else {
   label = 3;
  }
 } else {
  label = 3;
 }
 if ((label|0) == 3) {
  $6 = ((($fr)) + 19620|0);
  HEAP32[$6>>2] = 1;
  $7 = ((($fr)) + 19624|0);
  HEAP32[$7>>2] = 1;
 }
 $14 = ((($fr)) + 28068|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = $15 & 32;
 $17 = ($16|0)==(0);
 if (!($17)) {
  STACKTOP = sp;return 1;
 }
 $18 = ((($fr)) + 28064|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($19|0)==(0);
 if ($20) {
  STACKTOP = sp;return 1;
 }
 $21 = HEAP32[37544]|0;
 HEAP32[$vararg_buffer3>>2] = 597716;
 (_fprintf($21,515219,$vararg_buffer3)|0);
 STACKTOP = sp;return 1;
}
function _INT123_dectype($decoder) {
 $decoder = $decoder|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($decoder|0)==(0|0);
 if ($0) {
  $$0 = 0;
 } else {
  $1 = HEAP8[$decoder>>0]|0;
  $2 = ($1<<24>>24)==(0);
  if ($2) {
   $$0 = 0;
  } else {
   $3 = (_strcasecmp($decoder,515058)|0);
   $4 = ($3|0)==(0);
   if ($4) {
    $$0 = 0;
   } else {
    $5 = (_strcasecmp($decoder,597716)|0);
    $6 = ($5|0)==(0);
    if ($6) {
     $$0 = 1;
    } else {
     $7 = (_strcasecmp($decoder,515063)|0);
     $8 = ($7|0)==(0);
     if ($8) {
      $$0 = 2;
     } else {
      $9 = (_strcasecmp($decoder,515078)|0);
      $10 = ($9|0)==(0);
      if ($10) {
       $$0 = 3;
      } else {
       $11 = (_strcasecmp($decoder,515083)|0);
       $12 = ($11|0)==(0);
       if ($12) {
        $$0 = 4;
       } else {
        $13 = (_strcasecmp($decoder,515088)|0);
        $14 = ($13|0)==(0);
        if ($14) {
         $$0 = 5;
        } else {
         $15 = (_strcasecmp($decoder,515093)|0);
         $16 = ($15|0)==(0);
         if ($16) {
          $$0 = 6;
         } else {
          $17 = (_strcasecmp($decoder,515105)|0);
          $18 = ($17|0)==(0);
          if ($18) {
           $$0 = 7;
          } else {
           $19 = (_strcasecmp($decoder,515109)|0);
           $20 = ($19|0)==(0);
           if ($20) {
            $$0 = 8;
           } else {
            $21 = (_strcasecmp($decoder,515115)|0);
            $22 = ($21|0)==(0);
            if ($22) {
             $$0 = 9;
            } else {
             $23 = (_strcasecmp($decoder,515124)|0);
             $24 = ($23|0)==(0);
             if ($24) {
              $$0 = 10;
             } else {
              $25 = (_strcasecmp($decoder,515132)|0);
              $26 = ($25|0)==(0);
              if ($26) {
               $$0 = 11;
              } else {
               $27 = (_strcasecmp($decoder,515136)|0);
               $28 = ($27|0)==(0);
               if ($28) {
                $$0 = 12;
               } else {
                $29 = (_strcasecmp($decoder,515143)|0);
                $30 = ($29|0)==(0);
                if ($30) {
                 $$0 = 13;
                } else {
                 $31 = (_strcasecmp($decoder,515147)|0);
                 $32 = ($31|0)==(0);
                 $$ = $32 ? 14 : 15;
                 return ($$|0);
                }
               }
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    }
   }
  }
 }
 return ($$0|0);
}
function _synth_stereo_wrap($bandPtr_l,$bandPtr_r,$fr) {
 $bandPtr_l = $bandPtr_l|0;
 $bandPtr_r = $bandPtr_r|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 19636|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (FUNCTION_TABLE_iiiii[$1 & 8191]($bandPtr_l,0,$fr,0)|0);
 $3 = HEAP32[$0>>2]|0;
 $4 = (FUNCTION_TABLE_iiiii[$3 & 8191]($bandPtr_r,1,$fr,1)|0);
 $5 = (($4) + ($2))|0;
 return ($5|0);
}
function _INT123_synth_1to1_m2s($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 19364|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiiii[$3 & 8191]($bandPtr,0,$fr,1)|0);
 $5 = ((($fr)) + 27864|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + -128)|0;
 $8 = (($1) + ($7)|0);
 $9 = HEAP16[$8>>1]|0;
 $10 = ((($8)) + 2|0);
 HEAP16[$10>>1] = $9;
 $11 = ((($8)) + 4|0);
 $12 = HEAP16[$11>>1]|0;
 $13 = ((($11)) + 2|0);
 HEAP16[$13>>1] = $12;
 $14 = ((($11)) + 4|0);
 $15 = HEAP16[$14>>1]|0;
 $16 = ((($14)) + 2|0);
 HEAP16[$16>>1] = $15;
 $17 = ((($14)) + 4|0);
 $18 = HEAP16[$17>>1]|0;
 $19 = ((($17)) + 2|0);
 HEAP16[$19>>1] = $18;
 $20 = ((($17)) + 4|0);
 $21 = HEAP16[$20>>1]|0;
 $22 = ((($20)) + 2|0);
 HEAP16[$22>>1] = $21;
 $23 = ((($20)) + 4|0);
 $24 = HEAP16[$23>>1]|0;
 $25 = ((($23)) + 2|0);
 HEAP16[$25>>1] = $24;
 $26 = ((($23)) + 4|0);
 $27 = HEAP16[$26>>1]|0;
 $28 = ((($26)) + 2|0);
 HEAP16[$28>>1] = $27;
 $29 = ((($26)) + 4|0);
 $30 = HEAP16[$29>>1]|0;
 $31 = ((($29)) + 2|0);
 HEAP16[$31>>1] = $30;
 $32 = ((($29)) + 4|0);
 $33 = HEAP16[$32>>1]|0;
 $34 = ((($32)) + 2|0);
 HEAP16[$34>>1] = $33;
 $35 = ((($32)) + 4|0);
 $36 = HEAP16[$35>>1]|0;
 $37 = ((($35)) + 2|0);
 HEAP16[$37>>1] = $36;
 $38 = ((($35)) + 4|0);
 $39 = HEAP16[$38>>1]|0;
 $40 = ((($38)) + 2|0);
 HEAP16[$40>>1] = $39;
 $41 = ((($38)) + 4|0);
 $42 = HEAP16[$41>>1]|0;
 $43 = ((($41)) + 2|0);
 HEAP16[$43>>1] = $42;
 $44 = ((($41)) + 4|0);
 $45 = HEAP16[$44>>1]|0;
 $46 = ((($44)) + 2|0);
 HEAP16[$46>>1] = $45;
 $47 = ((($44)) + 4|0);
 $48 = HEAP16[$47>>1]|0;
 $49 = ((($47)) + 2|0);
 HEAP16[$49>>1] = $48;
 $50 = ((($47)) + 4|0);
 $51 = HEAP16[$50>>1]|0;
 $52 = ((($50)) + 2|0);
 HEAP16[$52>>1] = $51;
 $53 = ((($50)) + 4|0);
 $54 = HEAP16[$53>>1]|0;
 $55 = ((($53)) + 2|0);
 HEAP16[$55>>1] = $54;
 $56 = ((($53)) + 4|0);
 $57 = HEAP16[$56>>1]|0;
 $58 = ((($56)) + 2|0);
 HEAP16[$58>>1] = $57;
 $59 = ((($56)) + 4|0);
 $60 = HEAP16[$59>>1]|0;
 $61 = ((($59)) + 2|0);
 HEAP16[$61>>1] = $60;
 $62 = ((($59)) + 4|0);
 $63 = HEAP16[$62>>1]|0;
 $64 = ((($62)) + 2|0);
 HEAP16[$64>>1] = $63;
 $65 = ((($62)) + 4|0);
 $66 = HEAP16[$65>>1]|0;
 $67 = ((($65)) + 2|0);
 HEAP16[$67>>1] = $66;
 $68 = ((($65)) + 4|0);
 $69 = HEAP16[$68>>1]|0;
 $70 = ((($68)) + 2|0);
 HEAP16[$70>>1] = $69;
 $71 = ((($68)) + 4|0);
 $72 = HEAP16[$71>>1]|0;
 $73 = ((($71)) + 2|0);
 HEAP16[$73>>1] = $72;
 $74 = ((($71)) + 4|0);
 $75 = HEAP16[$74>>1]|0;
 $76 = ((($74)) + 2|0);
 HEAP16[$76>>1] = $75;
 $77 = ((($74)) + 4|0);
 $78 = HEAP16[$77>>1]|0;
 $79 = ((($77)) + 2|0);
 HEAP16[$79>>1] = $78;
 $80 = ((($77)) + 4|0);
 $81 = HEAP16[$80>>1]|0;
 $82 = ((($80)) + 2|0);
 HEAP16[$82>>1] = $81;
 $83 = ((($80)) + 4|0);
 $84 = HEAP16[$83>>1]|0;
 $85 = ((($83)) + 2|0);
 HEAP16[$85>>1] = $84;
 $86 = ((($83)) + 4|0);
 $87 = HEAP16[$86>>1]|0;
 $88 = ((($86)) + 2|0);
 HEAP16[$88>>1] = $87;
 $89 = ((($86)) + 4|0);
 $90 = HEAP16[$89>>1]|0;
 $91 = ((($89)) + 2|0);
 HEAP16[$91>>1] = $90;
 $92 = ((($89)) + 4|0);
 $93 = HEAP16[$92>>1]|0;
 $94 = ((($92)) + 2|0);
 HEAP16[$94>>1] = $93;
 $95 = ((($92)) + 4|0);
 $96 = HEAP16[$95>>1]|0;
 $97 = ((($95)) + 2|0);
 HEAP16[$97>>1] = $96;
 $98 = ((($95)) + 4|0);
 $99 = HEAP16[$98>>1]|0;
 $100 = ((($98)) + 2|0);
 HEAP16[$100>>1] = $99;
 $101 = ((($98)) + 4|0);
 $102 = HEAP16[$101>>1]|0;
 $103 = ((($101)) + 2|0);
 HEAP16[$103>>1] = $102;
 return ($4|0);
}
function _INT123_synth_1to1_8bit_m2s($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 19368|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiiii[$3 & 8191]($bandPtr,0,$fr,1)|0);
 $5 = ((($fr)) + 27864|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + -64)|0;
 $8 = (($1) + ($7)|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = ((($8)) + 1|0);
 HEAP8[$10>>0] = $9;
 $11 = ((($8)) + 2|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = ((($11)) + 1|0);
 HEAP8[$13>>0] = $12;
 $14 = ((($11)) + 2|0);
 $15 = HEAP8[$14>>0]|0;
 $16 = ((($14)) + 1|0);
 HEAP8[$16>>0] = $15;
 $17 = ((($14)) + 2|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = ((($17)) + 1|0);
 HEAP8[$19>>0] = $18;
 $20 = ((($17)) + 2|0);
 $21 = HEAP8[$20>>0]|0;
 $22 = ((($20)) + 1|0);
 HEAP8[$22>>0] = $21;
 $23 = ((($20)) + 2|0);
 $24 = HEAP8[$23>>0]|0;
 $25 = ((($23)) + 1|0);
 HEAP8[$25>>0] = $24;
 $26 = ((($23)) + 2|0);
 $27 = HEAP8[$26>>0]|0;
 $28 = ((($26)) + 1|0);
 HEAP8[$28>>0] = $27;
 $29 = ((($26)) + 2|0);
 $30 = HEAP8[$29>>0]|0;
 $31 = ((($29)) + 1|0);
 HEAP8[$31>>0] = $30;
 $32 = ((($29)) + 2|0);
 $33 = HEAP8[$32>>0]|0;
 $34 = ((($32)) + 1|0);
 HEAP8[$34>>0] = $33;
 $35 = ((($32)) + 2|0);
 $36 = HEAP8[$35>>0]|0;
 $37 = ((($35)) + 1|0);
 HEAP8[$37>>0] = $36;
 $38 = ((($35)) + 2|0);
 $39 = HEAP8[$38>>0]|0;
 $40 = ((($38)) + 1|0);
 HEAP8[$40>>0] = $39;
 $41 = ((($38)) + 2|0);
 $42 = HEAP8[$41>>0]|0;
 $43 = ((($41)) + 1|0);
 HEAP8[$43>>0] = $42;
 $44 = ((($41)) + 2|0);
 $45 = HEAP8[$44>>0]|0;
 $46 = ((($44)) + 1|0);
 HEAP8[$46>>0] = $45;
 $47 = ((($44)) + 2|0);
 $48 = HEAP8[$47>>0]|0;
 $49 = ((($47)) + 1|0);
 HEAP8[$49>>0] = $48;
 $50 = ((($47)) + 2|0);
 $51 = HEAP8[$50>>0]|0;
 $52 = ((($50)) + 1|0);
 HEAP8[$52>>0] = $51;
 $53 = ((($50)) + 2|0);
 $54 = HEAP8[$53>>0]|0;
 $55 = ((($53)) + 1|0);
 HEAP8[$55>>0] = $54;
 $56 = ((($53)) + 2|0);
 $57 = HEAP8[$56>>0]|0;
 $58 = ((($56)) + 1|0);
 HEAP8[$58>>0] = $57;
 $59 = ((($56)) + 2|0);
 $60 = HEAP8[$59>>0]|0;
 $61 = ((($59)) + 1|0);
 HEAP8[$61>>0] = $60;
 $62 = ((($59)) + 2|0);
 $63 = HEAP8[$62>>0]|0;
 $64 = ((($62)) + 1|0);
 HEAP8[$64>>0] = $63;
 $65 = ((($62)) + 2|0);
 $66 = HEAP8[$65>>0]|0;
 $67 = ((($65)) + 1|0);
 HEAP8[$67>>0] = $66;
 $68 = ((($65)) + 2|0);
 $69 = HEAP8[$68>>0]|0;
 $70 = ((($68)) + 1|0);
 HEAP8[$70>>0] = $69;
 $71 = ((($68)) + 2|0);
 $72 = HEAP8[$71>>0]|0;
 $73 = ((($71)) + 1|0);
 HEAP8[$73>>0] = $72;
 $74 = ((($71)) + 2|0);
 $75 = HEAP8[$74>>0]|0;
 $76 = ((($74)) + 1|0);
 HEAP8[$76>>0] = $75;
 $77 = ((($74)) + 2|0);
 $78 = HEAP8[$77>>0]|0;
 $79 = ((($77)) + 1|0);
 HEAP8[$79>>0] = $78;
 $80 = ((($77)) + 2|0);
 $81 = HEAP8[$80>>0]|0;
 $82 = ((($80)) + 1|0);
 HEAP8[$82>>0] = $81;
 $83 = ((($80)) + 2|0);
 $84 = HEAP8[$83>>0]|0;
 $85 = ((($83)) + 1|0);
 HEAP8[$85>>0] = $84;
 $86 = ((($83)) + 2|0);
 $87 = HEAP8[$86>>0]|0;
 $88 = ((($86)) + 1|0);
 HEAP8[$88>>0] = $87;
 $89 = ((($86)) + 2|0);
 $90 = HEAP8[$89>>0]|0;
 $91 = ((($89)) + 1|0);
 HEAP8[$91>>0] = $90;
 $92 = ((($89)) + 2|0);
 $93 = HEAP8[$92>>0]|0;
 $94 = ((($92)) + 1|0);
 HEAP8[$94>>0] = $93;
 $95 = ((($92)) + 2|0);
 $96 = HEAP8[$95>>0]|0;
 $97 = ((($95)) + 1|0);
 HEAP8[$97>>0] = $96;
 $98 = ((($95)) + 2|0);
 $99 = HEAP8[$98>>0]|0;
 $100 = ((($98)) + 1|0);
 HEAP8[$100>>0] = $99;
 $101 = ((($98)) + 2|0);
 $102 = HEAP8[$101>>0]|0;
 $103 = ((($101)) + 1|0);
 HEAP8[$103>>0] = $102;
 return ($4|0);
}
function _INT123_synth_1to1_real_m2s($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 19372|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiiii[$3 & 8191]($bandPtr,0,$fr,1)|0);
 $5 = ((($fr)) + 27864|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + -256)|0;
 $8 = (($1) + ($7)|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($8)) + 4|0);
 HEAP32[$10>>2] = $9;
 $11 = ((($8)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($11)) + 4|0);
 HEAP32[$13>>2] = $12;
 $14 = ((($11)) + 8|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($14)) + 4|0);
 HEAP32[$16>>2] = $15;
 $17 = ((($14)) + 8|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($17)) + 4|0);
 HEAP32[$19>>2] = $18;
 $20 = ((($17)) + 8|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ((($20)) + 4|0);
 HEAP32[$22>>2] = $21;
 $23 = ((($20)) + 8|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($23)) + 4|0);
 HEAP32[$25>>2] = $24;
 $26 = ((($23)) + 8|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ((($26)) + 4|0);
 HEAP32[$28>>2] = $27;
 $29 = ((($26)) + 8|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ((($29)) + 4|0);
 HEAP32[$31>>2] = $30;
 $32 = ((($29)) + 8|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($32)) + 4|0);
 HEAP32[$34>>2] = $33;
 $35 = ((($32)) + 8|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = ((($35)) + 4|0);
 HEAP32[$37>>2] = $36;
 $38 = ((($35)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ((($38)) + 4|0);
 HEAP32[$40>>2] = $39;
 $41 = ((($38)) + 8|0);
 $42 = HEAP32[$41>>2]|0;
 $43 = ((($41)) + 4|0);
 HEAP32[$43>>2] = $42;
 $44 = ((($41)) + 8|0);
 $45 = HEAP32[$44>>2]|0;
 $46 = ((($44)) + 4|0);
 HEAP32[$46>>2] = $45;
 $47 = ((($44)) + 8|0);
 $48 = HEAP32[$47>>2]|0;
 $49 = ((($47)) + 4|0);
 HEAP32[$49>>2] = $48;
 $50 = ((($47)) + 8|0);
 $51 = HEAP32[$50>>2]|0;
 $52 = ((($50)) + 4|0);
 HEAP32[$52>>2] = $51;
 $53 = ((($50)) + 8|0);
 $54 = HEAP32[$53>>2]|0;
 $55 = ((($53)) + 4|0);
 HEAP32[$55>>2] = $54;
 $56 = ((($53)) + 8|0);
 $57 = HEAP32[$56>>2]|0;
 $58 = ((($56)) + 4|0);
 HEAP32[$58>>2] = $57;
 $59 = ((($56)) + 8|0);
 $60 = HEAP32[$59>>2]|0;
 $61 = ((($59)) + 4|0);
 HEAP32[$61>>2] = $60;
 $62 = ((($59)) + 8|0);
 $63 = HEAP32[$62>>2]|0;
 $64 = ((($62)) + 4|0);
 HEAP32[$64>>2] = $63;
 $65 = ((($62)) + 8|0);
 $66 = HEAP32[$65>>2]|0;
 $67 = ((($65)) + 4|0);
 HEAP32[$67>>2] = $66;
 $68 = ((($65)) + 8|0);
 $69 = HEAP32[$68>>2]|0;
 $70 = ((($68)) + 4|0);
 HEAP32[$70>>2] = $69;
 $71 = ((($68)) + 8|0);
 $72 = HEAP32[$71>>2]|0;
 $73 = ((($71)) + 4|0);
 HEAP32[$73>>2] = $72;
 $74 = ((($71)) + 8|0);
 $75 = HEAP32[$74>>2]|0;
 $76 = ((($74)) + 4|0);
 HEAP32[$76>>2] = $75;
 $77 = ((($74)) + 8|0);
 $78 = HEAP32[$77>>2]|0;
 $79 = ((($77)) + 4|0);
 HEAP32[$79>>2] = $78;
 $80 = ((($77)) + 8|0);
 $81 = HEAP32[$80>>2]|0;
 $82 = ((($80)) + 4|0);
 HEAP32[$82>>2] = $81;
 $83 = ((($80)) + 8|0);
 $84 = HEAP32[$83>>2]|0;
 $85 = ((($83)) + 4|0);
 HEAP32[$85>>2] = $84;
 $86 = ((($83)) + 8|0);
 $87 = HEAP32[$86>>2]|0;
 $88 = ((($86)) + 4|0);
 HEAP32[$88>>2] = $87;
 $89 = ((($86)) + 8|0);
 $90 = HEAP32[$89>>2]|0;
 $91 = ((($89)) + 4|0);
 HEAP32[$91>>2] = $90;
 $92 = ((($89)) + 8|0);
 $93 = HEAP32[$92>>2]|0;
 $94 = ((($92)) + 4|0);
 HEAP32[$94>>2] = $93;
 $95 = ((($92)) + 8|0);
 $96 = HEAP32[$95>>2]|0;
 $97 = ((($95)) + 4|0);
 HEAP32[$97>>2] = $96;
 $98 = ((($95)) + 8|0);
 $99 = HEAP32[$98>>2]|0;
 $100 = ((($98)) + 4|0);
 HEAP32[$100>>2] = $99;
 $101 = ((($98)) + 8|0);
 $102 = HEAP32[$101>>2]|0;
 $103 = ((($101)) + 4|0);
 HEAP32[$103>>2] = $102;
 return ($4|0);
}
function _INT123_synth_1to1_s32_m2s($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 19376|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiiii[$3 & 8191]($bandPtr,0,$fr,1)|0);
 $5 = ((($fr)) + 27864|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + -256)|0;
 $8 = (($1) + ($7)|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($8)) + 4|0);
 HEAP32[$10>>2] = $9;
 $11 = ((($8)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($11)) + 4|0);
 HEAP32[$13>>2] = $12;
 $14 = ((($11)) + 8|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($14)) + 4|0);
 HEAP32[$16>>2] = $15;
 $17 = ((($14)) + 8|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($17)) + 4|0);
 HEAP32[$19>>2] = $18;
 $20 = ((($17)) + 8|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ((($20)) + 4|0);
 HEAP32[$22>>2] = $21;
 $23 = ((($20)) + 8|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($23)) + 4|0);
 HEAP32[$25>>2] = $24;
 $26 = ((($23)) + 8|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ((($26)) + 4|0);
 HEAP32[$28>>2] = $27;
 $29 = ((($26)) + 8|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ((($29)) + 4|0);
 HEAP32[$31>>2] = $30;
 $32 = ((($29)) + 8|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($32)) + 4|0);
 HEAP32[$34>>2] = $33;
 $35 = ((($32)) + 8|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = ((($35)) + 4|0);
 HEAP32[$37>>2] = $36;
 $38 = ((($35)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ((($38)) + 4|0);
 HEAP32[$40>>2] = $39;
 $41 = ((($38)) + 8|0);
 $42 = HEAP32[$41>>2]|0;
 $43 = ((($41)) + 4|0);
 HEAP32[$43>>2] = $42;
 $44 = ((($41)) + 8|0);
 $45 = HEAP32[$44>>2]|0;
 $46 = ((($44)) + 4|0);
 HEAP32[$46>>2] = $45;
 $47 = ((($44)) + 8|0);
 $48 = HEAP32[$47>>2]|0;
 $49 = ((($47)) + 4|0);
 HEAP32[$49>>2] = $48;
 $50 = ((($47)) + 8|0);
 $51 = HEAP32[$50>>2]|0;
 $52 = ((($50)) + 4|0);
 HEAP32[$52>>2] = $51;
 $53 = ((($50)) + 8|0);
 $54 = HEAP32[$53>>2]|0;
 $55 = ((($53)) + 4|0);
 HEAP32[$55>>2] = $54;
 $56 = ((($53)) + 8|0);
 $57 = HEAP32[$56>>2]|0;
 $58 = ((($56)) + 4|0);
 HEAP32[$58>>2] = $57;
 $59 = ((($56)) + 8|0);
 $60 = HEAP32[$59>>2]|0;
 $61 = ((($59)) + 4|0);
 HEAP32[$61>>2] = $60;
 $62 = ((($59)) + 8|0);
 $63 = HEAP32[$62>>2]|0;
 $64 = ((($62)) + 4|0);
 HEAP32[$64>>2] = $63;
 $65 = ((($62)) + 8|0);
 $66 = HEAP32[$65>>2]|0;
 $67 = ((($65)) + 4|0);
 HEAP32[$67>>2] = $66;
 $68 = ((($65)) + 8|0);
 $69 = HEAP32[$68>>2]|0;
 $70 = ((($68)) + 4|0);
 HEAP32[$70>>2] = $69;
 $71 = ((($68)) + 8|0);
 $72 = HEAP32[$71>>2]|0;
 $73 = ((($71)) + 4|0);
 HEAP32[$73>>2] = $72;
 $74 = ((($71)) + 8|0);
 $75 = HEAP32[$74>>2]|0;
 $76 = ((($74)) + 4|0);
 HEAP32[$76>>2] = $75;
 $77 = ((($74)) + 8|0);
 $78 = HEAP32[$77>>2]|0;
 $79 = ((($77)) + 4|0);
 HEAP32[$79>>2] = $78;
 $80 = ((($77)) + 8|0);
 $81 = HEAP32[$80>>2]|0;
 $82 = ((($80)) + 4|0);
 HEAP32[$82>>2] = $81;
 $83 = ((($80)) + 8|0);
 $84 = HEAP32[$83>>2]|0;
 $85 = ((($83)) + 4|0);
 HEAP32[$85>>2] = $84;
 $86 = ((($83)) + 8|0);
 $87 = HEAP32[$86>>2]|0;
 $88 = ((($86)) + 4|0);
 HEAP32[$88>>2] = $87;
 $89 = ((($86)) + 8|0);
 $90 = HEAP32[$89>>2]|0;
 $91 = ((($89)) + 4|0);
 HEAP32[$91>>2] = $90;
 $92 = ((($89)) + 8|0);
 $93 = HEAP32[$92>>2]|0;
 $94 = ((($92)) + 4|0);
 HEAP32[$94>>2] = $93;
 $95 = ((($92)) + 8|0);
 $96 = HEAP32[$95>>2]|0;
 $97 = ((($95)) + 4|0);
 HEAP32[$97>>2] = $96;
 $98 = ((($95)) + 8|0);
 $99 = HEAP32[$98>>2]|0;
 $100 = ((($98)) + 4|0);
 HEAP32[$100>>2] = $99;
 $101 = ((($98)) + 8|0);
 $102 = HEAP32[$101>>2]|0;
 $103 = ((($101)) + 4|0);
 HEAP32[$103>>2] = $102;
 return ($4|0);
}
function _INT123_synth_2to1_m2s($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 19380|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiiii[$3 & 8191]($bandPtr,0,$fr,1)|0);
 $5 = ((($fr)) + 27864|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + -64)|0;
 $8 = (($1) + ($7)|0);
 $9 = HEAP16[$8>>1]|0;
 $10 = ((($8)) + 2|0);
 HEAP16[$10>>1] = $9;
 $11 = ((($8)) + 4|0);
 $12 = HEAP16[$11>>1]|0;
 $13 = ((($11)) + 2|0);
 HEAP16[$13>>1] = $12;
 $14 = ((($11)) + 4|0);
 $15 = HEAP16[$14>>1]|0;
 $16 = ((($14)) + 2|0);
 HEAP16[$16>>1] = $15;
 $17 = ((($14)) + 4|0);
 $18 = HEAP16[$17>>1]|0;
 $19 = ((($17)) + 2|0);
 HEAP16[$19>>1] = $18;
 $20 = ((($17)) + 4|0);
 $21 = HEAP16[$20>>1]|0;
 $22 = ((($20)) + 2|0);
 HEAP16[$22>>1] = $21;
 $23 = ((($20)) + 4|0);
 $24 = HEAP16[$23>>1]|0;
 $25 = ((($23)) + 2|0);
 HEAP16[$25>>1] = $24;
 $26 = ((($23)) + 4|0);
 $27 = HEAP16[$26>>1]|0;
 $28 = ((($26)) + 2|0);
 HEAP16[$28>>1] = $27;
 $29 = ((($26)) + 4|0);
 $30 = HEAP16[$29>>1]|0;
 $31 = ((($29)) + 2|0);
 HEAP16[$31>>1] = $30;
 $32 = ((($29)) + 4|0);
 $33 = HEAP16[$32>>1]|0;
 $34 = ((($32)) + 2|0);
 HEAP16[$34>>1] = $33;
 $35 = ((($32)) + 4|0);
 $36 = HEAP16[$35>>1]|0;
 $37 = ((($35)) + 2|0);
 HEAP16[$37>>1] = $36;
 $38 = ((($35)) + 4|0);
 $39 = HEAP16[$38>>1]|0;
 $40 = ((($38)) + 2|0);
 HEAP16[$40>>1] = $39;
 $41 = ((($38)) + 4|0);
 $42 = HEAP16[$41>>1]|0;
 $43 = ((($41)) + 2|0);
 HEAP16[$43>>1] = $42;
 $44 = ((($41)) + 4|0);
 $45 = HEAP16[$44>>1]|0;
 $46 = ((($44)) + 2|0);
 HEAP16[$46>>1] = $45;
 $47 = ((($44)) + 4|0);
 $48 = HEAP16[$47>>1]|0;
 $49 = ((($47)) + 2|0);
 HEAP16[$49>>1] = $48;
 $50 = ((($47)) + 4|0);
 $51 = HEAP16[$50>>1]|0;
 $52 = ((($50)) + 2|0);
 HEAP16[$52>>1] = $51;
 $53 = ((($50)) + 4|0);
 $54 = HEAP16[$53>>1]|0;
 $55 = ((($53)) + 2|0);
 HEAP16[$55>>1] = $54;
 return ($4|0);
}
function _INT123_synth_2to1_8bit_m2s($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 19384|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiiii[$3 & 8191]($bandPtr,0,$fr,1)|0);
 $5 = ((($fr)) + 27864|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + -32)|0;
 $8 = (($1) + ($7)|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = ((($8)) + 1|0);
 HEAP8[$10>>0] = $9;
 $11 = ((($8)) + 2|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = ((($11)) + 1|0);
 HEAP8[$13>>0] = $12;
 $14 = ((($11)) + 2|0);
 $15 = HEAP8[$14>>0]|0;
 $16 = ((($14)) + 1|0);
 HEAP8[$16>>0] = $15;
 $17 = ((($14)) + 2|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = ((($17)) + 1|0);
 HEAP8[$19>>0] = $18;
 $20 = ((($17)) + 2|0);
 $21 = HEAP8[$20>>0]|0;
 $22 = ((($20)) + 1|0);
 HEAP8[$22>>0] = $21;
 $23 = ((($20)) + 2|0);
 $24 = HEAP8[$23>>0]|0;
 $25 = ((($23)) + 1|0);
 HEAP8[$25>>0] = $24;
 $26 = ((($23)) + 2|0);
 $27 = HEAP8[$26>>0]|0;
 $28 = ((($26)) + 1|0);
 HEAP8[$28>>0] = $27;
 $29 = ((($26)) + 2|0);
 $30 = HEAP8[$29>>0]|0;
 $31 = ((($29)) + 1|0);
 HEAP8[$31>>0] = $30;
 $32 = ((($29)) + 2|0);
 $33 = HEAP8[$32>>0]|0;
 $34 = ((($32)) + 1|0);
 HEAP8[$34>>0] = $33;
 $35 = ((($32)) + 2|0);
 $36 = HEAP8[$35>>0]|0;
 $37 = ((($35)) + 1|0);
 HEAP8[$37>>0] = $36;
 $38 = ((($35)) + 2|0);
 $39 = HEAP8[$38>>0]|0;
 $40 = ((($38)) + 1|0);
 HEAP8[$40>>0] = $39;
 $41 = ((($38)) + 2|0);
 $42 = HEAP8[$41>>0]|0;
 $43 = ((($41)) + 1|0);
 HEAP8[$43>>0] = $42;
 $44 = ((($41)) + 2|0);
 $45 = HEAP8[$44>>0]|0;
 $46 = ((($44)) + 1|0);
 HEAP8[$46>>0] = $45;
 $47 = ((($44)) + 2|0);
 $48 = HEAP8[$47>>0]|0;
 $49 = ((($47)) + 1|0);
 HEAP8[$49>>0] = $48;
 $50 = ((($47)) + 2|0);
 $51 = HEAP8[$50>>0]|0;
 $52 = ((($50)) + 1|0);
 HEAP8[$52>>0] = $51;
 $53 = ((($50)) + 2|0);
 $54 = HEAP8[$53>>0]|0;
 $55 = ((($53)) + 1|0);
 HEAP8[$55>>0] = $54;
 return ($4|0);
}
function _INT123_synth_2to1_real_m2s($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 19388|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiiii[$3 & 8191]($bandPtr,0,$fr,1)|0);
 $5 = ((($fr)) + 27864|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + -128)|0;
 $8 = (($1) + ($7)|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($8)) + 4|0);
 HEAP32[$10>>2] = $9;
 $11 = ((($8)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($11)) + 4|0);
 HEAP32[$13>>2] = $12;
 $14 = ((($11)) + 8|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($14)) + 4|0);
 HEAP32[$16>>2] = $15;
 $17 = ((($14)) + 8|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($17)) + 4|0);
 HEAP32[$19>>2] = $18;
 $20 = ((($17)) + 8|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ((($20)) + 4|0);
 HEAP32[$22>>2] = $21;
 $23 = ((($20)) + 8|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($23)) + 4|0);
 HEAP32[$25>>2] = $24;
 $26 = ((($23)) + 8|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ((($26)) + 4|0);
 HEAP32[$28>>2] = $27;
 $29 = ((($26)) + 8|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ((($29)) + 4|0);
 HEAP32[$31>>2] = $30;
 $32 = ((($29)) + 8|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($32)) + 4|0);
 HEAP32[$34>>2] = $33;
 $35 = ((($32)) + 8|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = ((($35)) + 4|0);
 HEAP32[$37>>2] = $36;
 $38 = ((($35)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ((($38)) + 4|0);
 HEAP32[$40>>2] = $39;
 $41 = ((($38)) + 8|0);
 $42 = HEAP32[$41>>2]|0;
 $43 = ((($41)) + 4|0);
 HEAP32[$43>>2] = $42;
 $44 = ((($41)) + 8|0);
 $45 = HEAP32[$44>>2]|0;
 $46 = ((($44)) + 4|0);
 HEAP32[$46>>2] = $45;
 $47 = ((($44)) + 8|0);
 $48 = HEAP32[$47>>2]|0;
 $49 = ((($47)) + 4|0);
 HEAP32[$49>>2] = $48;
 $50 = ((($47)) + 8|0);
 $51 = HEAP32[$50>>2]|0;
 $52 = ((($50)) + 4|0);
 HEAP32[$52>>2] = $51;
 $53 = ((($50)) + 8|0);
 $54 = HEAP32[$53>>2]|0;
 $55 = ((($53)) + 4|0);
 HEAP32[$55>>2] = $54;
 return ($4|0);
}
function _INT123_synth_2to1_s32_m2s($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 19392|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiiii[$3 & 8191]($bandPtr,0,$fr,1)|0);
 $5 = ((($fr)) + 27864|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + -128)|0;
 $8 = (($1) + ($7)|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($8)) + 4|0);
 HEAP32[$10>>2] = $9;
 $11 = ((($8)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($11)) + 4|0);
 HEAP32[$13>>2] = $12;
 $14 = ((($11)) + 8|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($14)) + 4|0);
 HEAP32[$16>>2] = $15;
 $17 = ((($14)) + 8|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($17)) + 4|0);
 HEAP32[$19>>2] = $18;
 $20 = ((($17)) + 8|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ((($20)) + 4|0);
 HEAP32[$22>>2] = $21;
 $23 = ((($20)) + 8|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($23)) + 4|0);
 HEAP32[$25>>2] = $24;
 $26 = ((($23)) + 8|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ((($26)) + 4|0);
 HEAP32[$28>>2] = $27;
 $29 = ((($26)) + 8|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ((($29)) + 4|0);
 HEAP32[$31>>2] = $30;
 $32 = ((($29)) + 8|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($32)) + 4|0);
 HEAP32[$34>>2] = $33;
 $35 = ((($32)) + 8|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = ((($35)) + 4|0);
 HEAP32[$37>>2] = $36;
 $38 = ((($35)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ((($38)) + 4|0);
 HEAP32[$40>>2] = $39;
 $41 = ((($38)) + 8|0);
 $42 = HEAP32[$41>>2]|0;
 $43 = ((($41)) + 4|0);
 HEAP32[$43>>2] = $42;
 $44 = ((($41)) + 8|0);
 $45 = HEAP32[$44>>2]|0;
 $46 = ((($44)) + 4|0);
 HEAP32[$46>>2] = $45;
 $47 = ((($44)) + 8|0);
 $48 = HEAP32[$47>>2]|0;
 $49 = ((($47)) + 4|0);
 HEAP32[$49>>2] = $48;
 $50 = ((($47)) + 8|0);
 $51 = HEAP32[$50>>2]|0;
 $52 = ((($50)) + 4|0);
 HEAP32[$52>>2] = $51;
 $53 = ((($50)) + 8|0);
 $54 = HEAP32[$53>>2]|0;
 $55 = ((($53)) + 4|0);
 HEAP32[$55>>2] = $54;
 return ($4|0);
}
function _INT123_synth_4to1_m2s($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 19396|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiiii[$3 & 8191]($bandPtr,0,$fr,1)|0);
 $5 = ((($fr)) + 27864|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + -32)|0;
 $8 = (($1) + ($7)|0);
 $9 = HEAP16[$8>>1]|0;
 $10 = ((($8)) + 2|0);
 HEAP16[$10>>1] = $9;
 $11 = ((($8)) + 4|0);
 $12 = HEAP16[$11>>1]|0;
 $13 = ((($11)) + 2|0);
 HEAP16[$13>>1] = $12;
 $14 = ((($11)) + 4|0);
 $15 = HEAP16[$14>>1]|0;
 $16 = ((($14)) + 2|0);
 HEAP16[$16>>1] = $15;
 $17 = ((($14)) + 4|0);
 $18 = HEAP16[$17>>1]|0;
 $19 = ((($17)) + 2|0);
 HEAP16[$19>>1] = $18;
 $20 = ((($17)) + 4|0);
 $21 = HEAP16[$20>>1]|0;
 $22 = ((($20)) + 2|0);
 HEAP16[$22>>1] = $21;
 $23 = ((($20)) + 4|0);
 $24 = HEAP16[$23>>1]|0;
 $25 = ((($23)) + 2|0);
 HEAP16[$25>>1] = $24;
 $26 = ((($23)) + 4|0);
 $27 = HEAP16[$26>>1]|0;
 $28 = ((($26)) + 2|0);
 HEAP16[$28>>1] = $27;
 $29 = ((($26)) + 4|0);
 $30 = HEAP16[$29>>1]|0;
 $31 = ((($29)) + 2|0);
 HEAP16[$31>>1] = $30;
 return ($4|0);
}
function _INT123_synth_4to1_8bit_m2s($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 19400|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiiii[$3 & 8191]($bandPtr,0,$fr,1)|0);
 $5 = ((($fr)) + 27864|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + -16)|0;
 $8 = (($1) + ($7)|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = ((($8)) + 1|0);
 HEAP8[$10>>0] = $9;
 $11 = ((($8)) + 2|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = ((($11)) + 1|0);
 HEAP8[$13>>0] = $12;
 $14 = ((($11)) + 2|0);
 $15 = HEAP8[$14>>0]|0;
 $16 = ((($14)) + 1|0);
 HEAP8[$16>>0] = $15;
 $17 = ((($14)) + 2|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = ((($17)) + 1|0);
 HEAP8[$19>>0] = $18;
 $20 = ((($17)) + 2|0);
 $21 = HEAP8[$20>>0]|0;
 $22 = ((($20)) + 1|0);
 HEAP8[$22>>0] = $21;
 $23 = ((($20)) + 2|0);
 $24 = HEAP8[$23>>0]|0;
 $25 = ((($23)) + 1|0);
 HEAP8[$25>>0] = $24;
 $26 = ((($23)) + 2|0);
 $27 = HEAP8[$26>>0]|0;
 $28 = ((($26)) + 1|0);
 HEAP8[$28>>0] = $27;
 $29 = ((($26)) + 2|0);
 $30 = HEAP8[$29>>0]|0;
 $31 = ((($29)) + 1|0);
 HEAP8[$31>>0] = $30;
 return ($4|0);
}
function _INT123_synth_4to1_real_m2s($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 19404|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiiii[$3 & 8191]($bandPtr,0,$fr,1)|0);
 $5 = ((($fr)) + 27864|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + -64)|0;
 $8 = (($1) + ($7)|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($8)) + 4|0);
 HEAP32[$10>>2] = $9;
 $11 = ((($8)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($11)) + 4|0);
 HEAP32[$13>>2] = $12;
 $14 = ((($11)) + 8|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($14)) + 4|0);
 HEAP32[$16>>2] = $15;
 $17 = ((($14)) + 8|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($17)) + 4|0);
 HEAP32[$19>>2] = $18;
 $20 = ((($17)) + 8|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ((($20)) + 4|0);
 HEAP32[$22>>2] = $21;
 $23 = ((($20)) + 8|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($23)) + 4|0);
 HEAP32[$25>>2] = $24;
 $26 = ((($23)) + 8|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ((($26)) + 4|0);
 HEAP32[$28>>2] = $27;
 $29 = ((($26)) + 8|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ((($29)) + 4|0);
 HEAP32[$31>>2] = $30;
 return ($4|0);
}
function _INT123_synth_4to1_s32_m2s($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 19408|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiiii[$3 & 8191]($bandPtr,0,$fr,1)|0);
 $5 = ((($fr)) + 27864|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + -64)|0;
 $8 = (($1) + ($7)|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($8)) + 4|0);
 HEAP32[$10>>2] = $9;
 $11 = ((($8)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($11)) + 4|0);
 HEAP32[$13>>2] = $12;
 $14 = ((($11)) + 8|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($14)) + 4|0);
 HEAP32[$16>>2] = $15;
 $17 = ((($14)) + 8|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($17)) + 4|0);
 HEAP32[$19>>2] = $18;
 $20 = ((($17)) + 8|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ((($20)) + 4|0);
 HEAP32[$22>>2] = $21;
 $23 = ((($20)) + 8|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($23)) + 4|0);
 HEAP32[$25>>2] = $24;
 $26 = ((($23)) + 8|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ((($26)) + 4|0);
 HEAP32[$28>>2] = $27;
 $29 = ((($26)) + 8|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ((($29)) + 4|0);
 HEAP32[$31>>2] = $30;
 return ($4|0);
}
function _INT123_synth_ntom_m2s($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $samples$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27864|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27856|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_INT123_synth_ntom($bandPtr,0,$fr,1)|0);
 $5 = HEAP32[$0>>2]|0;
 $6 = (($5) - ($1))|0;
 $7 = ($6>>>0)>(3);
 if (!($7)) {
  return ($4|0);
 }
 $8 = (($3) + ($1)|0);
 $9 = $6 >>> 2;
 $i$01 = 0;$samples$02 = $8;
 while(1) {
  $10 = HEAP16[$samples$02>>1]|0;
  $11 = ((($samples$02)) + 2|0);
  HEAP16[$11>>1] = $10;
  $12 = ((($samples$02)) + 4|0);
  $13 = (($i$01) + 1)|0;
  $14 = ($13>>>0)<($9>>>0);
  if ($14) {
   $i$01 = $13;$samples$02 = $12;
  } else {
   break;
  }
 }
 return ($4|0);
}
function _INT123_synth_ntom_8bit_m2s($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $samples$02 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27864|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27856|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_INT123_synth_ntom_8bit($bandPtr,0,$fr,1)|0);
 $5 = HEAP32[$0>>2]|0;
 $6 = (($5) - ($1))|0;
 $7 = ($6>>>0)>(1);
 if (!($7)) {
  return ($4|0);
 }
 $8 = (($3) + ($1)|0);
 $i$01 = 0;$samples$02 = $8;
 while(1) {
  $9 = HEAP8[$samples$02>>0]|0;
  $10 = ((($samples$02)) + 1|0);
  HEAP8[$10>>0] = $9;
  $11 = ((($samples$02)) + 2|0);
  $12 = (($i$01) + 1)|0;
  $13 = HEAP32[$0>>2]|0;
  $14 = (($13) - ($1))|0;
  $15 = $14 >>> 1;
  $16 = ($12>>>0)<($15>>>0);
  if ($16) {
   $i$01 = $12;$samples$02 = $11;
  } else {
   break;
  }
 }
 return ($4|0);
}
function _INT123_synth_ntom_real_m2s($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $samples$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27864|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27856|0);
 $3 = HEAP32[$2>>2]|0;
 (_INT123_synth_ntom_real($bandPtr,0,$fr,1)|0);
 $4 = HEAP32[$0>>2]|0;
 $5 = (($4) - ($1))|0;
 $6 = ($5>>>0)>(7);
 if (!($6)) {
  return 0;
 }
 $7 = (($3) + ($1)|0);
 $8 = $5 >>> 3;
 $i$01 = 0;$samples$02 = $7;
 while(1) {
  $9 = HEAP32[$samples$02>>2]|0;
  $10 = ((($samples$02)) + 4|0);
  HEAP32[$10>>2] = $9;
  $11 = ((($samples$02)) + 8|0);
  $12 = (($i$01) + 1)|0;
  $13 = ($12>>>0)<($8>>>0);
  if ($13) {
   $i$01 = $12;$samples$02 = $11;
  } else {
   break;
  }
 }
 return 0;
}
function _INT123_synth_ntom_s32_m2s($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $samples$02 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27864|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27856|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_INT123_synth_ntom_s32($bandPtr,0,$fr,1)|0);
 $5 = HEAP32[$0>>2]|0;
 $6 = (($5) - ($1))|0;
 $7 = ($6>>>0)>(7);
 if (!($7)) {
  return ($4|0);
 }
 $8 = (($3) + ($1)|0);
 $i$01 = 0;$samples$02 = $8;
 while(1) {
  $9 = HEAP32[$samples$02>>2]|0;
  $10 = ((($samples$02)) + 4|0);
  HEAP32[$10>>2] = $9;
  $11 = ((($samples$02)) + 8|0);
  $12 = (($i$01) + 1)|0;
  $13 = HEAP32[$0>>2]|0;
  $14 = (($13) - ($1))|0;
  $15 = $14 >>> 3;
  $16 = ($12>>>0)<($15>>>0);
  if ($16) {
   $i$01 = $12;$samples$02 = $11;
  } else {
   break;
  }
 }
 return ($4|0);
}
function _INT123_synth_1to1_mono($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $samples_tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$0>>2] = $samples_tmp;
 HEAP32[$2>>2] = 0;
 $4 = ((($fr)) + 19364|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiiii[$5 & 8191]($bandPtr,0,$fr,0)|0);
 HEAP32[$0>>2] = $1;
 $7 = (($1) + ($3)|0);
 $8 = HEAP16[$samples_tmp>>1]|0;
 HEAP16[$7>>1] = $8;
 $9 = ((($7)) + 2|0);
 $10 = ((($samples_tmp)) + 4|0);
 $11 = HEAP16[$10>>1]|0;
 HEAP16[$9>>1] = $11;
 $12 = ((($9)) + 2|0);
 $13 = ((($samples_tmp)) + 8|0);
 $14 = HEAP16[$13>>1]|0;
 HEAP16[$12>>1] = $14;
 $15 = ((($12)) + 2|0);
 $16 = ((($samples_tmp)) + 12|0);
 $17 = HEAP16[$16>>1]|0;
 HEAP16[$15>>1] = $17;
 $18 = ((($15)) + 2|0);
 $19 = ((($samples_tmp)) + 16|0);
 $20 = HEAP16[$19>>1]|0;
 HEAP16[$18>>1] = $20;
 $21 = ((($18)) + 2|0);
 $22 = ((($samples_tmp)) + 20|0);
 $23 = HEAP16[$22>>1]|0;
 HEAP16[$21>>1] = $23;
 $24 = ((($21)) + 2|0);
 $25 = ((($samples_tmp)) + 24|0);
 $26 = HEAP16[$25>>1]|0;
 HEAP16[$24>>1] = $26;
 $27 = ((($24)) + 2|0);
 $28 = ((($samples_tmp)) + 28|0);
 $29 = HEAP16[$28>>1]|0;
 HEAP16[$27>>1] = $29;
 $30 = ((($27)) + 2|0);
 $31 = ((($samples_tmp)) + 32|0);
 $32 = HEAP16[$31>>1]|0;
 HEAP16[$30>>1] = $32;
 $33 = ((($30)) + 2|0);
 $34 = ((($samples_tmp)) + 36|0);
 $35 = HEAP16[$34>>1]|0;
 HEAP16[$33>>1] = $35;
 $36 = ((($33)) + 2|0);
 $37 = ((($samples_tmp)) + 40|0);
 $38 = HEAP16[$37>>1]|0;
 HEAP16[$36>>1] = $38;
 $39 = ((($36)) + 2|0);
 $40 = ((($samples_tmp)) + 44|0);
 $41 = HEAP16[$40>>1]|0;
 HEAP16[$39>>1] = $41;
 $42 = ((($39)) + 2|0);
 $43 = ((($samples_tmp)) + 48|0);
 $44 = HEAP16[$43>>1]|0;
 HEAP16[$42>>1] = $44;
 $45 = ((($42)) + 2|0);
 $46 = ((($samples_tmp)) + 52|0);
 $47 = HEAP16[$46>>1]|0;
 HEAP16[$45>>1] = $47;
 $48 = ((($45)) + 2|0);
 $49 = ((($samples_tmp)) + 56|0);
 $50 = HEAP16[$49>>1]|0;
 HEAP16[$48>>1] = $50;
 $51 = ((($48)) + 2|0);
 $52 = ((($samples_tmp)) + 60|0);
 $53 = HEAP16[$52>>1]|0;
 HEAP16[$51>>1] = $53;
 $54 = ((($51)) + 2|0);
 $55 = ((($samples_tmp)) + 64|0);
 $56 = HEAP16[$55>>1]|0;
 HEAP16[$54>>1] = $56;
 $57 = ((($54)) + 2|0);
 $58 = ((($samples_tmp)) + 68|0);
 $59 = HEAP16[$58>>1]|0;
 HEAP16[$57>>1] = $59;
 $60 = ((($57)) + 2|0);
 $61 = ((($samples_tmp)) + 72|0);
 $62 = HEAP16[$61>>1]|0;
 HEAP16[$60>>1] = $62;
 $63 = ((($60)) + 2|0);
 $64 = ((($samples_tmp)) + 76|0);
 $65 = HEAP16[$64>>1]|0;
 HEAP16[$63>>1] = $65;
 $66 = ((($63)) + 2|0);
 $67 = ((($samples_tmp)) + 80|0);
 $68 = HEAP16[$67>>1]|0;
 HEAP16[$66>>1] = $68;
 $69 = ((($66)) + 2|0);
 $70 = ((($samples_tmp)) + 84|0);
 $71 = HEAP16[$70>>1]|0;
 HEAP16[$69>>1] = $71;
 $72 = ((($69)) + 2|0);
 $73 = ((($samples_tmp)) + 88|0);
 $74 = HEAP16[$73>>1]|0;
 HEAP16[$72>>1] = $74;
 $75 = ((($72)) + 2|0);
 $76 = ((($samples_tmp)) + 92|0);
 $77 = HEAP16[$76>>1]|0;
 HEAP16[$75>>1] = $77;
 $78 = ((($75)) + 2|0);
 $79 = ((($samples_tmp)) + 96|0);
 $80 = HEAP16[$79>>1]|0;
 HEAP16[$78>>1] = $80;
 $81 = ((($78)) + 2|0);
 $82 = ((($samples_tmp)) + 100|0);
 $83 = HEAP16[$82>>1]|0;
 HEAP16[$81>>1] = $83;
 $84 = ((($81)) + 2|0);
 $85 = ((($samples_tmp)) + 104|0);
 $86 = HEAP16[$85>>1]|0;
 HEAP16[$84>>1] = $86;
 $87 = ((($84)) + 2|0);
 $88 = ((($samples_tmp)) + 108|0);
 $89 = HEAP16[$88>>1]|0;
 HEAP16[$87>>1] = $89;
 $90 = ((($87)) + 2|0);
 $91 = ((($samples_tmp)) + 112|0);
 $92 = HEAP16[$91>>1]|0;
 HEAP16[$90>>1] = $92;
 $93 = ((($90)) + 2|0);
 $94 = ((($samples_tmp)) + 116|0);
 $95 = HEAP16[$94>>1]|0;
 HEAP16[$93>>1] = $95;
 $96 = ((($93)) + 2|0);
 $97 = ((($samples_tmp)) + 120|0);
 $98 = HEAP16[$97>>1]|0;
 HEAP16[$96>>1] = $98;
 $99 = ((($96)) + 2|0);
 $100 = ((($samples_tmp)) + 124|0);
 $101 = HEAP16[$100>>1]|0;
 HEAP16[$99>>1] = $101;
 $102 = (($3) + 64)|0;
 HEAP32[$2>>2] = $102;
 STACKTOP = sp;return ($6|0);
}
function _INT123_synth_1to1_8bit_mono($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $samples_tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$0>>2] = $samples_tmp;
 HEAP32[$2>>2] = 0;
 $4 = ((($fr)) + 19368|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiiii[$5 & 8191]($bandPtr,0,$fr,0)|0);
 HEAP32[$0>>2] = $1;
 $7 = (($1) + ($3)|0);
 $8 = HEAP8[$samples_tmp>>0]|0;
 HEAP8[$7>>0] = $8;
 $9 = ((($7)) + 1|0);
 $10 = ((($samples_tmp)) + 2|0);
 $11 = HEAP8[$10>>0]|0;
 HEAP8[$9>>0] = $11;
 $12 = ((($9)) + 1|0);
 $13 = ((($samples_tmp)) + 4|0);
 $14 = HEAP8[$13>>0]|0;
 HEAP8[$12>>0] = $14;
 $15 = ((($12)) + 1|0);
 $16 = ((($samples_tmp)) + 6|0);
 $17 = HEAP8[$16>>0]|0;
 HEAP8[$15>>0] = $17;
 $18 = ((($15)) + 1|0);
 $19 = ((($samples_tmp)) + 8|0);
 $20 = HEAP8[$19>>0]|0;
 HEAP8[$18>>0] = $20;
 $21 = ((($18)) + 1|0);
 $22 = ((($samples_tmp)) + 10|0);
 $23 = HEAP8[$22>>0]|0;
 HEAP8[$21>>0] = $23;
 $24 = ((($21)) + 1|0);
 $25 = ((($samples_tmp)) + 12|0);
 $26 = HEAP8[$25>>0]|0;
 HEAP8[$24>>0] = $26;
 $27 = ((($24)) + 1|0);
 $28 = ((($samples_tmp)) + 14|0);
 $29 = HEAP8[$28>>0]|0;
 HEAP8[$27>>0] = $29;
 $30 = ((($27)) + 1|0);
 $31 = ((($samples_tmp)) + 16|0);
 $32 = HEAP8[$31>>0]|0;
 HEAP8[$30>>0] = $32;
 $33 = ((($30)) + 1|0);
 $34 = ((($samples_tmp)) + 18|0);
 $35 = HEAP8[$34>>0]|0;
 HEAP8[$33>>0] = $35;
 $36 = ((($33)) + 1|0);
 $37 = ((($samples_tmp)) + 20|0);
 $38 = HEAP8[$37>>0]|0;
 HEAP8[$36>>0] = $38;
 $39 = ((($36)) + 1|0);
 $40 = ((($samples_tmp)) + 22|0);
 $41 = HEAP8[$40>>0]|0;
 HEAP8[$39>>0] = $41;
 $42 = ((($39)) + 1|0);
 $43 = ((($samples_tmp)) + 24|0);
 $44 = HEAP8[$43>>0]|0;
 HEAP8[$42>>0] = $44;
 $45 = ((($42)) + 1|0);
 $46 = ((($samples_tmp)) + 26|0);
 $47 = HEAP8[$46>>0]|0;
 HEAP8[$45>>0] = $47;
 $48 = ((($45)) + 1|0);
 $49 = ((($samples_tmp)) + 28|0);
 $50 = HEAP8[$49>>0]|0;
 HEAP8[$48>>0] = $50;
 $51 = ((($48)) + 1|0);
 $52 = ((($samples_tmp)) + 30|0);
 $53 = HEAP8[$52>>0]|0;
 HEAP8[$51>>0] = $53;
 $54 = ((($51)) + 1|0);
 $55 = ((($samples_tmp)) + 32|0);
 $56 = HEAP8[$55>>0]|0;
 HEAP8[$54>>0] = $56;
 $57 = ((($54)) + 1|0);
 $58 = ((($samples_tmp)) + 34|0);
 $59 = HEAP8[$58>>0]|0;
 HEAP8[$57>>0] = $59;
 $60 = ((($57)) + 1|0);
 $61 = ((($samples_tmp)) + 36|0);
 $62 = HEAP8[$61>>0]|0;
 HEAP8[$60>>0] = $62;
 $63 = ((($60)) + 1|0);
 $64 = ((($samples_tmp)) + 38|0);
 $65 = HEAP8[$64>>0]|0;
 HEAP8[$63>>0] = $65;
 $66 = ((($63)) + 1|0);
 $67 = ((($samples_tmp)) + 40|0);
 $68 = HEAP8[$67>>0]|0;
 HEAP8[$66>>0] = $68;
 $69 = ((($66)) + 1|0);
 $70 = ((($samples_tmp)) + 42|0);
 $71 = HEAP8[$70>>0]|0;
 HEAP8[$69>>0] = $71;
 $72 = ((($69)) + 1|0);
 $73 = ((($samples_tmp)) + 44|0);
 $74 = HEAP8[$73>>0]|0;
 HEAP8[$72>>0] = $74;
 $75 = ((($72)) + 1|0);
 $76 = ((($samples_tmp)) + 46|0);
 $77 = HEAP8[$76>>0]|0;
 HEAP8[$75>>0] = $77;
 $78 = ((($75)) + 1|0);
 $79 = ((($samples_tmp)) + 48|0);
 $80 = HEAP8[$79>>0]|0;
 HEAP8[$78>>0] = $80;
 $81 = ((($78)) + 1|0);
 $82 = ((($samples_tmp)) + 50|0);
 $83 = HEAP8[$82>>0]|0;
 HEAP8[$81>>0] = $83;
 $84 = ((($81)) + 1|0);
 $85 = ((($samples_tmp)) + 52|0);
 $86 = HEAP8[$85>>0]|0;
 HEAP8[$84>>0] = $86;
 $87 = ((($84)) + 1|0);
 $88 = ((($samples_tmp)) + 54|0);
 $89 = HEAP8[$88>>0]|0;
 HEAP8[$87>>0] = $89;
 $90 = ((($87)) + 1|0);
 $91 = ((($samples_tmp)) + 56|0);
 $92 = HEAP8[$91>>0]|0;
 HEAP8[$90>>0] = $92;
 $93 = ((($90)) + 1|0);
 $94 = ((($samples_tmp)) + 58|0);
 $95 = HEAP8[$94>>0]|0;
 HEAP8[$93>>0] = $95;
 $96 = ((($93)) + 1|0);
 $97 = ((($samples_tmp)) + 60|0);
 $98 = HEAP8[$97>>0]|0;
 HEAP8[$96>>0] = $98;
 $99 = ((($96)) + 1|0);
 $100 = ((($samples_tmp)) + 62|0);
 $101 = HEAP8[$100>>0]|0;
 HEAP8[$99>>0] = $101;
 $102 = (($3) + 32)|0;
 HEAP32[$2>>2] = $102;
 STACKTOP = sp;return ($6|0);
}
function _INT123_synth_1to1_real_mono($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $samples_tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$0>>2] = $samples_tmp;
 HEAP32[$2>>2] = 0;
 $4 = ((($fr)) + 19372|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiiii[$5 & 8191]($bandPtr,0,$fr,0)|0);
 HEAP32[$0>>2] = $1;
 $7 = (($1) + ($3)|0);
 $8 = HEAP32[$samples_tmp>>2]|0;
 HEAP32[$7>>2] = $8;
 $9 = ((($7)) + 4|0);
 $10 = ((($samples_tmp)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 HEAP32[$9>>2] = $11;
 $12 = ((($9)) + 4|0);
 $13 = ((($samples_tmp)) + 16|0);
 $14 = HEAP32[$13>>2]|0;
 HEAP32[$12>>2] = $14;
 $15 = ((($12)) + 4|0);
 $16 = ((($samples_tmp)) + 24|0);
 $17 = HEAP32[$16>>2]|0;
 HEAP32[$15>>2] = $17;
 $18 = ((($15)) + 4|0);
 $19 = ((($samples_tmp)) + 32|0);
 $20 = HEAP32[$19>>2]|0;
 HEAP32[$18>>2] = $20;
 $21 = ((($18)) + 4|0);
 $22 = ((($samples_tmp)) + 40|0);
 $23 = HEAP32[$22>>2]|0;
 HEAP32[$21>>2] = $23;
 $24 = ((($21)) + 4|0);
 $25 = ((($samples_tmp)) + 48|0);
 $26 = HEAP32[$25>>2]|0;
 HEAP32[$24>>2] = $26;
 $27 = ((($24)) + 4|0);
 $28 = ((($samples_tmp)) + 56|0);
 $29 = HEAP32[$28>>2]|0;
 HEAP32[$27>>2] = $29;
 $30 = ((($27)) + 4|0);
 $31 = ((($samples_tmp)) + 64|0);
 $32 = HEAP32[$31>>2]|0;
 HEAP32[$30>>2] = $32;
 $33 = ((($30)) + 4|0);
 $34 = ((($samples_tmp)) + 72|0);
 $35 = HEAP32[$34>>2]|0;
 HEAP32[$33>>2] = $35;
 $36 = ((($33)) + 4|0);
 $37 = ((($samples_tmp)) + 80|0);
 $38 = HEAP32[$37>>2]|0;
 HEAP32[$36>>2] = $38;
 $39 = ((($36)) + 4|0);
 $40 = ((($samples_tmp)) + 88|0);
 $41 = HEAP32[$40>>2]|0;
 HEAP32[$39>>2] = $41;
 $42 = ((($39)) + 4|0);
 $43 = ((($samples_tmp)) + 96|0);
 $44 = HEAP32[$43>>2]|0;
 HEAP32[$42>>2] = $44;
 $45 = ((($42)) + 4|0);
 $46 = ((($samples_tmp)) + 104|0);
 $47 = HEAP32[$46>>2]|0;
 HEAP32[$45>>2] = $47;
 $48 = ((($45)) + 4|0);
 $49 = ((($samples_tmp)) + 112|0);
 $50 = HEAP32[$49>>2]|0;
 HEAP32[$48>>2] = $50;
 $51 = ((($48)) + 4|0);
 $52 = ((($samples_tmp)) + 120|0);
 $53 = HEAP32[$52>>2]|0;
 HEAP32[$51>>2] = $53;
 $54 = ((($51)) + 4|0);
 $55 = ((($samples_tmp)) + 128|0);
 $56 = HEAP32[$55>>2]|0;
 HEAP32[$54>>2] = $56;
 $57 = ((($54)) + 4|0);
 $58 = ((($samples_tmp)) + 136|0);
 $59 = HEAP32[$58>>2]|0;
 HEAP32[$57>>2] = $59;
 $60 = ((($57)) + 4|0);
 $61 = ((($samples_tmp)) + 144|0);
 $62 = HEAP32[$61>>2]|0;
 HEAP32[$60>>2] = $62;
 $63 = ((($60)) + 4|0);
 $64 = ((($samples_tmp)) + 152|0);
 $65 = HEAP32[$64>>2]|0;
 HEAP32[$63>>2] = $65;
 $66 = ((($63)) + 4|0);
 $67 = ((($samples_tmp)) + 160|0);
 $68 = HEAP32[$67>>2]|0;
 HEAP32[$66>>2] = $68;
 $69 = ((($66)) + 4|0);
 $70 = ((($samples_tmp)) + 168|0);
 $71 = HEAP32[$70>>2]|0;
 HEAP32[$69>>2] = $71;
 $72 = ((($69)) + 4|0);
 $73 = ((($samples_tmp)) + 176|0);
 $74 = HEAP32[$73>>2]|0;
 HEAP32[$72>>2] = $74;
 $75 = ((($72)) + 4|0);
 $76 = ((($samples_tmp)) + 184|0);
 $77 = HEAP32[$76>>2]|0;
 HEAP32[$75>>2] = $77;
 $78 = ((($75)) + 4|0);
 $79 = ((($samples_tmp)) + 192|0);
 $80 = HEAP32[$79>>2]|0;
 HEAP32[$78>>2] = $80;
 $81 = ((($78)) + 4|0);
 $82 = ((($samples_tmp)) + 200|0);
 $83 = HEAP32[$82>>2]|0;
 HEAP32[$81>>2] = $83;
 $84 = ((($81)) + 4|0);
 $85 = ((($samples_tmp)) + 208|0);
 $86 = HEAP32[$85>>2]|0;
 HEAP32[$84>>2] = $86;
 $87 = ((($84)) + 4|0);
 $88 = ((($samples_tmp)) + 216|0);
 $89 = HEAP32[$88>>2]|0;
 HEAP32[$87>>2] = $89;
 $90 = ((($87)) + 4|0);
 $91 = ((($samples_tmp)) + 224|0);
 $92 = HEAP32[$91>>2]|0;
 HEAP32[$90>>2] = $92;
 $93 = ((($90)) + 4|0);
 $94 = ((($samples_tmp)) + 232|0);
 $95 = HEAP32[$94>>2]|0;
 HEAP32[$93>>2] = $95;
 $96 = ((($93)) + 4|0);
 $97 = ((($samples_tmp)) + 240|0);
 $98 = HEAP32[$97>>2]|0;
 HEAP32[$96>>2] = $98;
 $99 = ((($96)) + 4|0);
 $100 = ((($samples_tmp)) + 248|0);
 $101 = HEAP32[$100>>2]|0;
 HEAP32[$99>>2] = $101;
 $102 = (($3) + 128)|0;
 HEAP32[$2>>2] = $102;
 STACKTOP = sp;return ($6|0);
}
function _INT123_synth_1to1_s32_mono($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $samples_tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$0>>2] = $samples_tmp;
 HEAP32[$2>>2] = 0;
 $4 = ((($fr)) + 19376|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiiii[$5 & 8191]($bandPtr,0,$fr,0)|0);
 HEAP32[$0>>2] = $1;
 $7 = (($1) + ($3)|0);
 $8 = HEAP32[$samples_tmp>>2]|0;
 HEAP32[$7>>2] = $8;
 $9 = ((($7)) + 4|0);
 $10 = ((($samples_tmp)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 HEAP32[$9>>2] = $11;
 $12 = ((($9)) + 4|0);
 $13 = ((($samples_tmp)) + 16|0);
 $14 = HEAP32[$13>>2]|0;
 HEAP32[$12>>2] = $14;
 $15 = ((($12)) + 4|0);
 $16 = ((($samples_tmp)) + 24|0);
 $17 = HEAP32[$16>>2]|0;
 HEAP32[$15>>2] = $17;
 $18 = ((($15)) + 4|0);
 $19 = ((($samples_tmp)) + 32|0);
 $20 = HEAP32[$19>>2]|0;
 HEAP32[$18>>2] = $20;
 $21 = ((($18)) + 4|0);
 $22 = ((($samples_tmp)) + 40|0);
 $23 = HEAP32[$22>>2]|0;
 HEAP32[$21>>2] = $23;
 $24 = ((($21)) + 4|0);
 $25 = ((($samples_tmp)) + 48|0);
 $26 = HEAP32[$25>>2]|0;
 HEAP32[$24>>2] = $26;
 $27 = ((($24)) + 4|0);
 $28 = ((($samples_tmp)) + 56|0);
 $29 = HEAP32[$28>>2]|0;
 HEAP32[$27>>2] = $29;
 $30 = ((($27)) + 4|0);
 $31 = ((($samples_tmp)) + 64|0);
 $32 = HEAP32[$31>>2]|0;
 HEAP32[$30>>2] = $32;
 $33 = ((($30)) + 4|0);
 $34 = ((($samples_tmp)) + 72|0);
 $35 = HEAP32[$34>>2]|0;
 HEAP32[$33>>2] = $35;
 $36 = ((($33)) + 4|0);
 $37 = ((($samples_tmp)) + 80|0);
 $38 = HEAP32[$37>>2]|0;
 HEAP32[$36>>2] = $38;
 $39 = ((($36)) + 4|0);
 $40 = ((($samples_tmp)) + 88|0);
 $41 = HEAP32[$40>>2]|0;
 HEAP32[$39>>2] = $41;
 $42 = ((($39)) + 4|0);
 $43 = ((($samples_tmp)) + 96|0);
 $44 = HEAP32[$43>>2]|0;
 HEAP32[$42>>2] = $44;
 $45 = ((($42)) + 4|0);
 $46 = ((($samples_tmp)) + 104|0);
 $47 = HEAP32[$46>>2]|0;
 HEAP32[$45>>2] = $47;
 $48 = ((($45)) + 4|0);
 $49 = ((($samples_tmp)) + 112|0);
 $50 = HEAP32[$49>>2]|0;
 HEAP32[$48>>2] = $50;
 $51 = ((($48)) + 4|0);
 $52 = ((($samples_tmp)) + 120|0);
 $53 = HEAP32[$52>>2]|0;
 HEAP32[$51>>2] = $53;
 $54 = ((($51)) + 4|0);
 $55 = ((($samples_tmp)) + 128|0);
 $56 = HEAP32[$55>>2]|0;
 HEAP32[$54>>2] = $56;
 $57 = ((($54)) + 4|0);
 $58 = ((($samples_tmp)) + 136|0);
 $59 = HEAP32[$58>>2]|0;
 HEAP32[$57>>2] = $59;
 $60 = ((($57)) + 4|0);
 $61 = ((($samples_tmp)) + 144|0);
 $62 = HEAP32[$61>>2]|0;
 HEAP32[$60>>2] = $62;
 $63 = ((($60)) + 4|0);
 $64 = ((($samples_tmp)) + 152|0);
 $65 = HEAP32[$64>>2]|0;
 HEAP32[$63>>2] = $65;
 $66 = ((($63)) + 4|0);
 $67 = ((($samples_tmp)) + 160|0);
 $68 = HEAP32[$67>>2]|0;
 HEAP32[$66>>2] = $68;
 $69 = ((($66)) + 4|0);
 $70 = ((($samples_tmp)) + 168|0);
 $71 = HEAP32[$70>>2]|0;
 HEAP32[$69>>2] = $71;
 $72 = ((($69)) + 4|0);
 $73 = ((($samples_tmp)) + 176|0);
 $74 = HEAP32[$73>>2]|0;
 HEAP32[$72>>2] = $74;
 $75 = ((($72)) + 4|0);
 $76 = ((($samples_tmp)) + 184|0);
 $77 = HEAP32[$76>>2]|0;
 HEAP32[$75>>2] = $77;
 $78 = ((($75)) + 4|0);
 $79 = ((($samples_tmp)) + 192|0);
 $80 = HEAP32[$79>>2]|0;
 HEAP32[$78>>2] = $80;
 $81 = ((($78)) + 4|0);
 $82 = ((($samples_tmp)) + 200|0);
 $83 = HEAP32[$82>>2]|0;
 HEAP32[$81>>2] = $83;
 $84 = ((($81)) + 4|0);
 $85 = ((($samples_tmp)) + 208|0);
 $86 = HEAP32[$85>>2]|0;
 HEAP32[$84>>2] = $86;
 $87 = ((($84)) + 4|0);
 $88 = ((($samples_tmp)) + 216|0);
 $89 = HEAP32[$88>>2]|0;
 HEAP32[$87>>2] = $89;
 $90 = ((($87)) + 4|0);
 $91 = ((($samples_tmp)) + 224|0);
 $92 = HEAP32[$91>>2]|0;
 HEAP32[$90>>2] = $92;
 $93 = ((($90)) + 4|0);
 $94 = ((($samples_tmp)) + 232|0);
 $95 = HEAP32[$94>>2]|0;
 HEAP32[$93>>2] = $95;
 $96 = ((($93)) + 4|0);
 $97 = ((($samples_tmp)) + 240|0);
 $98 = HEAP32[$97>>2]|0;
 HEAP32[$96>>2] = $98;
 $99 = ((($96)) + 4|0);
 $100 = ((($samples_tmp)) + 248|0);
 $101 = HEAP32[$100>>2]|0;
 HEAP32[$99>>2] = $101;
 $102 = (($3) + 128)|0;
 HEAP32[$2>>2] = $102;
 STACKTOP = sp;return ($6|0);
}
function _INT123_synth_2to1_mono($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $samples_tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$0>>2] = $samples_tmp;
 HEAP32[$2>>2] = 0;
 $4 = ((($fr)) + 19380|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiiii[$5 & 8191]($bandPtr,0,$fr,0)|0);
 HEAP32[$0>>2] = $1;
 $7 = (($1) + ($3)|0);
 $8 = HEAP16[$samples_tmp>>1]|0;
 HEAP16[$7>>1] = $8;
 $9 = ((($7)) + 2|0);
 $10 = ((($samples_tmp)) + 4|0);
 $11 = HEAP16[$10>>1]|0;
 HEAP16[$9>>1] = $11;
 $12 = ((($9)) + 2|0);
 $13 = ((($samples_tmp)) + 8|0);
 $14 = HEAP16[$13>>1]|0;
 HEAP16[$12>>1] = $14;
 $15 = ((($12)) + 2|0);
 $16 = ((($samples_tmp)) + 12|0);
 $17 = HEAP16[$16>>1]|0;
 HEAP16[$15>>1] = $17;
 $18 = ((($15)) + 2|0);
 $19 = ((($samples_tmp)) + 16|0);
 $20 = HEAP16[$19>>1]|0;
 HEAP16[$18>>1] = $20;
 $21 = ((($18)) + 2|0);
 $22 = ((($samples_tmp)) + 20|0);
 $23 = HEAP16[$22>>1]|0;
 HEAP16[$21>>1] = $23;
 $24 = ((($21)) + 2|0);
 $25 = ((($samples_tmp)) + 24|0);
 $26 = HEAP16[$25>>1]|0;
 HEAP16[$24>>1] = $26;
 $27 = ((($24)) + 2|0);
 $28 = ((($samples_tmp)) + 28|0);
 $29 = HEAP16[$28>>1]|0;
 HEAP16[$27>>1] = $29;
 $30 = ((($27)) + 2|0);
 $31 = ((($samples_tmp)) + 32|0);
 $32 = HEAP16[$31>>1]|0;
 HEAP16[$30>>1] = $32;
 $33 = ((($30)) + 2|0);
 $34 = ((($samples_tmp)) + 36|0);
 $35 = HEAP16[$34>>1]|0;
 HEAP16[$33>>1] = $35;
 $36 = ((($33)) + 2|0);
 $37 = ((($samples_tmp)) + 40|0);
 $38 = HEAP16[$37>>1]|0;
 HEAP16[$36>>1] = $38;
 $39 = ((($36)) + 2|0);
 $40 = ((($samples_tmp)) + 44|0);
 $41 = HEAP16[$40>>1]|0;
 HEAP16[$39>>1] = $41;
 $42 = ((($39)) + 2|0);
 $43 = ((($samples_tmp)) + 48|0);
 $44 = HEAP16[$43>>1]|0;
 HEAP16[$42>>1] = $44;
 $45 = ((($42)) + 2|0);
 $46 = ((($samples_tmp)) + 52|0);
 $47 = HEAP16[$46>>1]|0;
 HEAP16[$45>>1] = $47;
 $48 = ((($45)) + 2|0);
 $49 = ((($samples_tmp)) + 56|0);
 $50 = HEAP16[$49>>1]|0;
 HEAP16[$48>>1] = $50;
 $51 = ((($48)) + 2|0);
 $52 = ((($samples_tmp)) + 60|0);
 $53 = HEAP16[$52>>1]|0;
 HEAP16[$51>>1] = $53;
 $54 = (($3) + 32)|0;
 HEAP32[$2>>2] = $54;
 STACKTOP = sp;return ($6|0);
}
function _INT123_synth_2to1_8bit_mono($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $samples_tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$0>>2] = $samples_tmp;
 HEAP32[$2>>2] = 0;
 $4 = ((($fr)) + 19384|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiiii[$5 & 8191]($bandPtr,0,$fr,0)|0);
 HEAP32[$0>>2] = $1;
 $7 = (($1) + ($3)|0);
 $8 = HEAP8[$samples_tmp>>0]|0;
 HEAP8[$7>>0] = $8;
 $9 = ((($7)) + 1|0);
 $10 = ((($samples_tmp)) + 2|0);
 $11 = HEAP8[$10>>0]|0;
 HEAP8[$9>>0] = $11;
 $12 = ((($9)) + 1|0);
 $13 = ((($samples_tmp)) + 4|0);
 $14 = HEAP8[$13>>0]|0;
 HEAP8[$12>>0] = $14;
 $15 = ((($12)) + 1|0);
 $16 = ((($samples_tmp)) + 6|0);
 $17 = HEAP8[$16>>0]|0;
 HEAP8[$15>>0] = $17;
 $18 = ((($15)) + 1|0);
 $19 = ((($samples_tmp)) + 8|0);
 $20 = HEAP8[$19>>0]|0;
 HEAP8[$18>>0] = $20;
 $21 = ((($18)) + 1|0);
 $22 = ((($samples_tmp)) + 10|0);
 $23 = HEAP8[$22>>0]|0;
 HEAP8[$21>>0] = $23;
 $24 = ((($21)) + 1|0);
 $25 = ((($samples_tmp)) + 12|0);
 $26 = HEAP8[$25>>0]|0;
 HEAP8[$24>>0] = $26;
 $27 = ((($24)) + 1|0);
 $28 = ((($samples_tmp)) + 14|0);
 $29 = HEAP8[$28>>0]|0;
 HEAP8[$27>>0] = $29;
 $30 = ((($27)) + 1|0);
 $31 = ((($samples_tmp)) + 16|0);
 $32 = HEAP8[$31>>0]|0;
 HEAP8[$30>>0] = $32;
 $33 = ((($30)) + 1|0);
 $34 = ((($samples_tmp)) + 18|0);
 $35 = HEAP8[$34>>0]|0;
 HEAP8[$33>>0] = $35;
 $36 = ((($33)) + 1|0);
 $37 = ((($samples_tmp)) + 20|0);
 $38 = HEAP8[$37>>0]|0;
 HEAP8[$36>>0] = $38;
 $39 = ((($36)) + 1|0);
 $40 = ((($samples_tmp)) + 22|0);
 $41 = HEAP8[$40>>0]|0;
 HEAP8[$39>>0] = $41;
 $42 = ((($39)) + 1|0);
 $43 = ((($samples_tmp)) + 24|0);
 $44 = HEAP8[$43>>0]|0;
 HEAP8[$42>>0] = $44;
 $45 = ((($42)) + 1|0);
 $46 = ((($samples_tmp)) + 26|0);
 $47 = HEAP8[$46>>0]|0;
 HEAP8[$45>>0] = $47;
 $48 = ((($45)) + 1|0);
 $49 = ((($samples_tmp)) + 28|0);
 $50 = HEAP8[$49>>0]|0;
 HEAP8[$48>>0] = $50;
 $51 = ((($48)) + 1|0);
 $52 = ((($samples_tmp)) + 30|0);
 $53 = HEAP8[$52>>0]|0;
 HEAP8[$51>>0] = $53;
 $54 = (($3) + 16)|0;
 HEAP32[$2>>2] = $54;
 STACKTOP = sp;return ($6|0);
}
function _INT123_synth_2to1_real_mono($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $samples_tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$0>>2] = $samples_tmp;
 HEAP32[$2>>2] = 0;
 $4 = ((($fr)) + 19388|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiiii[$5 & 8191]($bandPtr,0,$fr,0)|0);
 HEAP32[$0>>2] = $1;
 $7 = (($1) + ($3)|0);
 $8 = HEAP32[$samples_tmp>>2]|0;
 HEAP32[$7>>2] = $8;
 $9 = ((($7)) + 4|0);
 $10 = ((($samples_tmp)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 HEAP32[$9>>2] = $11;
 $12 = ((($9)) + 4|0);
 $13 = ((($samples_tmp)) + 16|0);
 $14 = HEAP32[$13>>2]|0;
 HEAP32[$12>>2] = $14;
 $15 = ((($12)) + 4|0);
 $16 = ((($samples_tmp)) + 24|0);
 $17 = HEAP32[$16>>2]|0;
 HEAP32[$15>>2] = $17;
 $18 = ((($15)) + 4|0);
 $19 = ((($samples_tmp)) + 32|0);
 $20 = HEAP32[$19>>2]|0;
 HEAP32[$18>>2] = $20;
 $21 = ((($18)) + 4|0);
 $22 = ((($samples_tmp)) + 40|0);
 $23 = HEAP32[$22>>2]|0;
 HEAP32[$21>>2] = $23;
 $24 = ((($21)) + 4|0);
 $25 = ((($samples_tmp)) + 48|0);
 $26 = HEAP32[$25>>2]|0;
 HEAP32[$24>>2] = $26;
 $27 = ((($24)) + 4|0);
 $28 = ((($samples_tmp)) + 56|0);
 $29 = HEAP32[$28>>2]|0;
 HEAP32[$27>>2] = $29;
 $30 = ((($27)) + 4|0);
 $31 = ((($samples_tmp)) + 64|0);
 $32 = HEAP32[$31>>2]|0;
 HEAP32[$30>>2] = $32;
 $33 = ((($30)) + 4|0);
 $34 = ((($samples_tmp)) + 72|0);
 $35 = HEAP32[$34>>2]|0;
 HEAP32[$33>>2] = $35;
 $36 = ((($33)) + 4|0);
 $37 = ((($samples_tmp)) + 80|0);
 $38 = HEAP32[$37>>2]|0;
 HEAP32[$36>>2] = $38;
 $39 = ((($36)) + 4|0);
 $40 = ((($samples_tmp)) + 88|0);
 $41 = HEAP32[$40>>2]|0;
 HEAP32[$39>>2] = $41;
 $42 = ((($39)) + 4|0);
 $43 = ((($samples_tmp)) + 96|0);
 $44 = HEAP32[$43>>2]|0;
 HEAP32[$42>>2] = $44;
 $45 = ((($42)) + 4|0);
 $46 = ((($samples_tmp)) + 104|0);
 $47 = HEAP32[$46>>2]|0;
 HEAP32[$45>>2] = $47;
 $48 = ((($45)) + 4|0);
 $49 = ((($samples_tmp)) + 112|0);
 $50 = HEAP32[$49>>2]|0;
 HEAP32[$48>>2] = $50;
 $51 = ((($48)) + 4|0);
 $52 = ((($samples_tmp)) + 120|0);
 $53 = HEAP32[$52>>2]|0;
 HEAP32[$51>>2] = $53;
 $54 = (($3) + 64)|0;
 HEAP32[$2>>2] = $54;
 STACKTOP = sp;return ($6|0);
}
function _INT123_synth_2to1_s32_mono($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $samples_tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$0>>2] = $samples_tmp;
 HEAP32[$2>>2] = 0;
 $4 = ((($fr)) + 19392|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiiii[$5 & 8191]($bandPtr,0,$fr,0)|0);
 HEAP32[$0>>2] = $1;
 $7 = (($1) + ($3)|0);
 $8 = HEAP32[$samples_tmp>>2]|0;
 HEAP32[$7>>2] = $8;
 $9 = ((($7)) + 4|0);
 $10 = ((($samples_tmp)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 HEAP32[$9>>2] = $11;
 $12 = ((($9)) + 4|0);
 $13 = ((($samples_tmp)) + 16|0);
 $14 = HEAP32[$13>>2]|0;
 HEAP32[$12>>2] = $14;
 $15 = ((($12)) + 4|0);
 $16 = ((($samples_tmp)) + 24|0);
 $17 = HEAP32[$16>>2]|0;
 HEAP32[$15>>2] = $17;
 $18 = ((($15)) + 4|0);
 $19 = ((($samples_tmp)) + 32|0);
 $20 = HEAP32[$19>>2]|0;
 HEAP32[$18>>2] = $20;
 $21 = ((($18)) + 4|0);
 $22 = ((($samples_tmp)) + 40|0);
 $23 = HEAP32[$22>>2]|0;
 HEAP32[$21>>2] = $23;
 $24 = ((($21)) + 4|0);
 $25 = ((($samples_tmp)) + 48|0);
 $26 = HEAP32[$25>>2]|0;
 HEAP32[$24>>2] = $26;
 $27 = ((($24)) + 4|0);
 $28 = ((($samples_tmp)) + 56|0);
 $29 = HEAP32[$28>>2]|0;
 HEAP32[$27>>2] = $29;
 $30 = ((($27)) + 4|0);
 $31 = ((($samples_tmp)) + 64|0);
 $32 = HEAP32[$31>>2]|0;
 HEAP32[$30>>2] = $32;
 $33 = ((($30)) + 4|0);
 $34 = ((($samples_tmp)) + 72|0);
 $35 = HEAP32[$34>>2]|0;
 HEAP32[$33>>2] = $35;
 $36 = ((($33)) + 4|0);
 $37 = ((($samples_tmp)) + 80|0);
 $38 = HEAP32[$37>>2]|0;
 HEAP32[$36>>2] = $38;
 $39 = ((($36)) + 4|0);
 $40 = ((($samples_tmp)) + 88|0);
 $41 = HEAP32[$40>>2]|0;
 HEAP32[$39>>2] = $41;
 $42 = ((($39)) + 4|0);
 $43 = ((($samples_tmp)) + 96|0);
 $44 = HEAP32[$43>>2]|0;
 HEAP32[$42>>2] = $44;
 $45 = ((($42)) + 4|0);
 $46 = ((($samples_tmp)) + 104|0);
 $47 = HEAP32[$46>>2]|0;
 HEAP32[$45>>2] = $47;
 $48 = ((($45)) + 4|0);
 $49 = ((($samples_tmp)) + 112|0);
 $50 = HEAP32[$49>>2]|0;
 HEAP32[$48>>2] = $50;
 $51 = ((($48)) + 4|0);
 $52 = ((($samples_tmp)) + 120|0);
 $53 = HEAP32[$52>>2]|0;
 HEAP32[$51>>2] = $53;
 $54 = (($3) + 64)|0;
 HEAP32[$2>>2] = $54;
 STACKTOP = sp;return ($6|0);
}
function _INT123_synth_4to1_mono($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $samples_tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$0>>2] = $samples_tmp;
 HEAP32[$2>>2] = 0;
 $4 = ((($fr)) + 19396|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiiii[$5 & 8191]($bandPtr,0,$fr,0)|0);
 HEAP32[$0>>2] = $1;
 $7 = (($1) + ($3)|0);
 $8 = HEAP16[$samples_tmp>>1]|0;
 HEAP16[$7>>1] = $8;
 $9 = ((($7)) + 2|0);
 $10 = ((($samples_tmp)) + 4|0);
 $11 = HEAP16[$10>>1]|0;
 HEAP16[$9>>1] = $11;
 $12 = ((($9)) + 2|0);
 $13 = ((($samples_tmp)) + 8|0);
 $14 = HEAP16[$13>>1]|0;
 HEAP16[$12>>1] = $14;
 $15 = ((($12)) + 2|0);
 $16 = ((($samples_tmp)) + 12|0);
 $17 = HEAP16[$16>>1]|0;
 HEAP16[$15>>1] = $17;
 $18 = ((($15)) + 2|0);
 $19 = ((($samples_tmp)) + 16|0);
 $20 = HEAP16[$19>>1]|0;
 HEAP16[$18>>1] = $20;
 $21 = ((($18)) + 2|0);
 $22 = ((($samples_tmp)) + 20|0);
 $23 = HEAP16[$22>>1]|0;
 HEAP16[$21>>1] = $23;
 $24 = ((($21)) + 2|0);
 $25 = ((($samples_tmp)) + 24|0);
 $26 = HEAP16[$25>>1]|0;
 HEAP16[$24>>1] = $26;
 $27 = ((($24)) + 2|0);
 $28 = ((($samples_tmp)) + 28|0);
 $29 = HEAP16[$28>>1]|0;
 HEAP16[$27>>1] = $29;
 $30 = (($3) + 16)|0;
 HEAP32[$2>>2] = $30;
 STACKTOP = sp;return ($6|0);
}
function _INT123_synth_4to1_8bit_mono($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $samples_tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$0>>2] = $samples_tmp;
 HEAP32[$2>>2] = 0;
 $4 = ((($fr)) + 19400|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiiii[$5 & 8191]($bandPtr,0,$fr,0)|0);
 HEAP32[$0>>2] = $1;
 $7 = (($1) + ($3)|0);
 $8 = HEAP8[$samples_tmp>>0]|0;
 HEAP8[$7>>0] = $8;
 $9 = ((($7)) + 1|0);
 $10 = ((($samples_tmp)) + 2|0);
 $11 = HEAP8[$10>>0]|0;
 HEAP8[$9>>0] = $11;
 $12 = ((($9)) + 1|0);
 $13 = ((($samples_tmp)) + 4|0);
 $14 = HEAP8[$13>>0]|0;
 HEAP8[$12>>0] = $14;
 $15 = ((($12)) + 1|0);
 $16 = ((($samples_tmp)) + 6|0);
 $17 = HEAP8[$16>>0]|0;
 HEAP8[$15>>0] = $17;
 $18 = ((($15)) + 1|0);
 $19 = ((($samples_tmp)) + 8|0);
 $20 = HEAP8[$19>>0]|0;
 HEAP8[$18>>0] = $20;
 $21 = ((($18)) + 1|0);
 $22 = ((($samples_tmp)) + 10|0);
 $23 = HEAP8[$22>>0]|0;
 HEAP8[$21>>0] = $23;
 $24 = ((($21)) + 1|0);
 $25 = ((($samples_tmp)) + 12|0);
 $26 = HEAP8[$25>>0]|0;
 HEAP8[$24>>0] = $26;
 $27 = ((($24)) + 1|0);
 $28 = ((($samples_tmp)) + 14|0);
 $29 = HEAP8[$28>>0]|0;
 HEAP8[$27>>0] = $29;
 $30 = (($3) + 8)|0;
 HEAP32[$2>>2] = $30;
 STACKTOP = sp;return ($6|0);
}
function _INT123_synth_4to1_real_mono($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $samples_tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$0>>2] = $samples_tmp;
 HEAP32[$2>>2] = 0;
 $4 = ((($fr)) + 19404|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiiii[$5 & 8191]($bandPtr,0,$fr,0)|0);
 HEAP32[$0>>2] = $1;
 $7 = (($1) + ($3)|0);
 $8 = HEAP32[$samples_tmp>>2]|0;
 HEAP32[$7>>2] = $8;
 $9 = ((($7)) + 4|0);
 $10 = ((($samples_tmp)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 HEAP32[$9>>2] = $11;
 $12 = ((($9)) + 4|0);
 $13 = ((($samples_tmp)) + 16|0);
 $14 = HEAP32[$13>>2]|0;
 HEAP32[$12>>2] = $14;
 $15 = ((($12)) + 4|0);
 $16 = ((($samples_tmp)) + 24|0);
 $17 = HEAP32[$16>>2]|0;
 HEAP32[$15>>2] = $17;
 $18 = ((($15)) + 4|0);
 $19 = ((($samples_tmp)) + 32|0);
 $20 = HEAP32[$19>>2]|0;
 HEAP32[$18>>2] = $20;
 $21 = ((($18)) + 4|0);
 $22 = ((($samples_tmp)) + 40|0);
 $23 = HEAP32[$22>>2]|0;
 HEAP32[$21>>2] = $23;
 $24 = ((($21)) + 4|0);
 $25 = ((($samples_tmp)) + 48|0);
 $26 = HEAP32[$25>>2]|0;
 HEAP32[$24>>2] = $26;
 $27 = ((($24)) + 4|0);
 $28 = ((($samples_tmp)) + 56|0);
 $29 = HEAP32[$28>>2]|0;
 HEAP32[$27>>2] = $29;
 $30 = (($3) + 32)|0;
 HEAP32[$2>>2] = $30;
 STACKTOP = sp;return ($6|0);
}
function _INT123_synth_4to1_s32_mono($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $samples_tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$0>>2] = $samples_tmp;
 HEAP32[$2>>2] = 0;
 $4 = ((($fr)) + 19408|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiiii[$5 & 8191]($bandPtr,0,$fr,0)|0);
 HEAP32[$0>>2] = $1;
 $7 = (($1) + ($3)|0);
 $8 = HEAP32[$samples_tmp>>2]|0;
 HEAP32[$7>>2] = $8;
 $9 = ((($7)) + 4|0);
 $10 = ((($samples_tmp)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 HEAP32[$9>>2] = $11;
 $12 = ((($9)) + 4|0);
 $13 = ((($samples_tmp)) + 16|0);
 $14 = HEAP32[$13>>2]|0;
 HEAP32[$12>>2] = $14;
 $15 = ((($12)) + 4|0);
 $16 = ((($samples_tmp)) + 24|0);
 $17 = HEAP32[$16>>2]|0;
 HEAP32[$15>>2] = $17;
 $18 = ((($15)) + 4|0);
 $19 = ((($samples_tmp)) + 32|0);
 $20 = HEAP32[$19>>2]|0;
 HEAP32[$18>>2] = $20;
 $21 = ((($18)) + 4|0);
 $22 = ((($samples_tmp)) + 40|0);
 $23 = HEAP32[$22>>2]|0;
 HEAP32[$21>>2] = $23;
 $24 = ((($21)) + 4|0);
 $25 = ((($samples_tmp)) + 48|0);
 $26 = HEAP32[$25>>2]|0;
 HEAP32[$24>>2] = $26;
 $27 = ((($24)) + 4|0);
 $28 = ((($samples_tmp)) + 56|0);
 $29 = HEAP32[$28>>2]|0;
 HEAP32[$27>>2] = $29;
 $30 = (($3) + 32)|0;
 HEAP32[$2>>2] = $30;
 STACKTOP = sp;return ($6|0);
}
function _INT123_synth_ntom_mono($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $samples$03 = 0, $samples_tmp = 0, $tmp1$02 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1024|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27864|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27856|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$2>>2] = $samples_tmp;
 HEAP32[$0>>2] = 0;
 $4 = (_INT123_synth_ntom($bandPtr,0,$fr,1)|0);
 HEAP32[$2>>2] = $3;
 $5 = HEAP32[$0>>2]|0;
 $6 = ($5>>>0)>(3);
 if (!($6)) {
  $14 = $5 >>> 1;
  $15 = (($14) + ($1))|0;
  HEAP32[$0>>2] = $15;
  STACKTOP = sp;return ($4|0);
 }
 $7 = (($3) + ($1)|0);
 $8 = $5 >>> 2;
 $i$01 = 0;$samples$03 = $7;$tmp1$02 = $samples_tmp;
 while(1) {
  $9 = HEAP16[$tmp1$02>>1]|0;
  HEAP16[$samples$03>>1] = $9;
  $10 = ((($samples$03)) + 2|0);
  $11 = ((($tmp1$02)) + 4|0);
  $12 = (($i$01) + 1)|0;
  $13 = ($12>>>0)<($8>>>0);
  if ($13) {
   $i$01 = $12;$samples$03 = $10;$tmp1$02 = $11;
  } else {
   break;
  }
 }
 $14 = $5 >>> 1;
 $15 = (($14) + ($1))|0;
 HEAP32[$0>>2] = $15;
 STACKTOP = sp;return ($4|0);
}
function _INT123_synth_ntom_8bit_mono($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $samples$03 = 0, $samples_tmp = 0;
 var $tmp1$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 512|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27864|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27856|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$2>>2] = $samples_tmp;
 HEAP32[$0>>2] = 0;
 $4 = (_INT123_synth_ntom_8bit($bandPtr,0,$fr,1)|0);
 HEAP32[$2>>2] = $3;
 $5 = HEAP32[$0>>2]|0;
 $6 = ($5>>>0)<(2);
 if ($6) {
  $$lcssa = 0;
  $15 = (($$lcssa) + ($1))|0;
  HEAP32[$0>>2] = $15;
  STACKTOP = sp;return ($4|0);
 }
 $7 = (($3) + ($1)|0);
 $i$01 = 0;$samples$03 = $7;$tmp1$02 = $samples_tmp;
 while(1) {
  $8 = HEAP8[$tmp1$02>>0]|0;
  HEAP8[$samples$03>>0] = $8;
  $9 = ((($samples$03)) + 1|0);
  $10 = ((($tmp1$02)) + 2|0);
  $11 = (($i$01) + 1)|0;
  $12 = HEAP32[$0>>2]|0;
  $13 = $12 >>> 1;
  $14 = ($11>>>0)<($13>>>0);
  if ($14) {
   $i$01 = $11;$samples$03 = $9;$tmp1$02 = $10;
  } else {
   $$lcssa = $13;
   break;
  }
 }
 $15 = (($$lcssa) + ($1))|0;
 HEAP32[$0>>2] = $15;
 STACKTOP = sp;return ($4|0);
}
function _INT123_synth_ntom_real_mono($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $samples$03 = 0, $samples_tmp = 0, $tmp1$02 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 2048|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27864|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27856|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$2>>2] = $samples_tmp;
 HEAP32[$0>>2] = 0;
 (_INT123_synth_ntom_real($bandPtr,0,$fr,1)|0);
 HEAP32[$2>>2] = $3;
 $4 = HEAP32[$0>>2]|0;
 $5 = ($4>>>0)>(7);
 if (!($5)) {
  $13 = $4 >>> 1;
  $14 = (($13) + ($1))|0;
  HEAP32[$0>>2] = $14;
  STACKTOP = sp;return 0;
 }
 $6 = (($3) + ($1)|0);
 $7 = $4 >>> 3;
 $i$01 = 0;$samples$03 = $6;$tmp1$02 = $samples_tmp;
 while(1) {
  $8 = HEAP32[$tmp1$02>>2]|0;
  HEAP32[$samples$03>>2] = $8;
  $9 = ((($samples$03)) + 4|0);
  $10 = ((($tmp1$02)) + 8|0);
  $11 = (($i$01) + 1)|0;
  $12 = ($11>>>0)<($7>>>0);
  if ($12) {
   $i$01 = $11;$samples$03 = $9;$tmp1$02 = $10;
  } else {
   break;
  }
 }
 $13 = $4 >>> 1;
 $14 = (($13) + ($1))|0;
 HEAP32[$0>>2] = $14;
 STACKTOP = sp;return 0;
}
function _INT123_synth_ntom_s32_mono($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $samples$03 = 0;
 var $samples_tmp = 0, $tmp1$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 2048|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27864|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27856|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$2>>2] = $samples_tmp;
 HEAP32[$0>>2] = 0;
 $4 = (_INT123_synth_ntom_s32($bandPtr,0,$fr,1)|0);
 HEAP32[$2>>2] = $3;
 $5 = HEAP32[$0>>2]|0;
 $6 = ($5>>>0)>(7);
 if (!($6)) {
  $$lcssa = $5;
  $15 = $$lcssa >>> 1;
  $16 = (($15) + ($1))|0;
  HEAP32[$0>>2] = $16;
  STACKTOP = sp;return ($4|0);
 }
 $7 = (($3) + ($1)|0);
 $i$01 = 0;$samples$03 = $7;$tmp1$02 = $samples_tmp;
 while(1) {
  $8 = HEAP32[$tmp1$02>>2]|0;
  HEAP32[$samples$03>>2] = $8;
  $9 = ((($samples$03)) + 4|0);
  $10 = ((($tmp1$02)) + 8|0);
  $11 = (($i$01) + 1)|0;
  $12 = HEAP32[$0>>2]|0;
  $13 = $12 >>> 3;
  $14 = ($11>>>0)<($13>>>0);
  if ($14) {
   $i$01 = $11;$samples$03 = $9;$tmp1$02 = $10;
  } else {
   $$lcssa = $12;
   break;
  }
 }
 $15 = $$lcssa >>> 1;
 $16 = (($15) + ($1))|0;
 HEAP32[$0>>2] = $16;
 STACKTOP = sp;return ($4|0);
}
function _INT123_synth_1to1_8bit_wrap_mono($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0;
 var $i$01 = 0, $samples$02 = 0, $samples_tmp = 0, $tmp1$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$0>>2] = $samples_tmp;
 HEAP32[$2>>2] = 0;
 $4 = ((($fr)) + 19364|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiiii[$5 & 8191]($bandPtr,0,$fr,0)|0);
 HEAP32[$0>>2] = $1;
 $7 = (($1) + ($3)|0);
 $8 = ((($fr)) + 9592|0);
 $i$01 = 0;$samples$02 = $7;$tmp1$03 = $samples_tmp;
 while(1) {
  $9 = HEAP16[$tmp1$03>>1]|0;
  $10 = $9 << 16 >> 16;
  $11 = $10 >> 3;
  $12 = HEAP32[$8>>2]|0;
  $13 = (($12) + ($11)|0);
  $14 = HEAP8[$13>>0]|0;
  $15 = ((($samples$02)) + 1|0);
  HEAP8[$samples$02>>0] = $14;
  $16 = ((($tmp1$03)) + 4|0);
  $17 = (($i$01) + 1)|0;
  $exitcond = ($17|0)==(32);
  if ($exitcond) {
   break;
  } else {
   $i$01 = $17;$samples$02 = $15;$tmp1$03 = $16;
  }
 }
 $18 = (($3) + 32)|0;
 HEAP32[$2>>2] = $18;
 STACKTOP = sp;return ($6|0);
}
function _INT123_synth_1to1_8bit_wrap_m2s($bandPtr,$fr) {
 $bandPtr = $bandPtr|0;
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, $samples$02 = 0, $samples_tmp = 0, $tmp1$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samples_tmp = sp;
 $0 = ((($fr)) + 27856|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 27864|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$0>>2] = $samples_tmp;
 HEAP32[$2>>2] = 0;
 $4 = ((($fr)) + 19364|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiiii[$5 & 8191]($bandPtr,0,$fr,0)|0);
 HEAP32[$0>>2] = $1;
 $7 = (($1) + ($3)|0);
 $8 = ((($fr)) + 9592|0);
 $i$01 = 0;$samples$02 = $7;$tmp1$03 = $samples_tmp;
 while(1) {
  $9 = HEAP16[$tmp1$03>>1]|0;
  $10 = $9 << 16 >> 16;
  $11 = $10 >> 3;
  $12 = HEAP32[$8>>2]|0;
  $13 = (($12) + ($11)|0);
  $14 = HEAP8[$13>>0]|0;
  $15 = ((($samples$02)) + 1|0);
  HEAP8[$samples$02>>0] = $14;
  $16 = HEAP16[$tmp1$03>>1]|0;
  $17 = $16 << 16 >> 16;
  $18 = $17 >> 3;
  $19 = HEAP32[$8>>2]|0;
  $20 = (($19) + ($18)|0);
  $21 = HEAP8[$20>>0]|0;
  $22 = ((($samples$02)) + 2|0);
  HEAP8[$15>>0] = $21;
  $23 = ((($tmp1$03)) + 4|0);
  $24 = (($i$01) + 1)|0;
  $exitcond = ($24|0)==(32);
  if ($exitcond) {
   break;
  } else {
   $i$01 = $24;$samples$02 = $22;$tmp1$03 = $23;
  }
 }
 $25 = (($3) + 64)|0;
 HEAP32[$2>>2] = $25;
 STACKTOP = sp;return ($6|0);
}
function _INT123_check_decoders() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _bc_poolsize($bc,$pool_size,$bufblock) {
 $bc = $bc|0;
 $pool_size = $pool_size|0;
 $bufblock = $bufblock|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($bc)) + 28|0);
 HEAP32[$0>>2] = $pool_size;
 $1 = ((($bc)) + 24|0);
 HEAP32[$1>>2] = $bufblock;
 return;
}
function _INT123_feed_more($fr,$in,$count) {
 $fr = $fr|0;
 $in = $in|0;
 $count = $count|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $ret$0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($fr)) + 28020|0);
 $1 = (_bc_add($0,$in,$count)|0);
 $2 = ($1|0)==(0);
 if ($2) {
  $ret$0 = 0;
  STACKTOP = sp;return ($ret$0|0);
 }
 $3 = ((($fr)) + 28068|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 & 32;
 $6 = ($5|0)==(0);
 if (!($6)) {
  $ret$0 = -1;
  STACKTOP = sp;return ($ret$0|0);
 }
 $7 = HEAP32[37544]|0;
 HEAP32[$vararg_buffer>>2] = 767;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = -1;
 (_fprintf($7,515232,$vararg_buffer)|0);
 $ret$0 = -1;
 STACKTOP = sp;return ($ret$0|0);
}
function _bc_add($bc,$data,$size) {
 $bc = $bc|0;
 $data = $data|0;
 $size = $size|0;
 var $$0$i$i = 0, $$1 = 0, $$1$ = 0, $$13 = 0, $$2 = 0, $$24 = 0, $$lcssa = 0, $$old111 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ret$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$old111 = ($size|0)>(0);
 if (!($$old111)) {
  $ret$2 = 0;
  return ($ret$2|0);
 }
 $0 = ((($bc)) + 4|0);
 $1 = ((($bc)) + 36|0);
 $2 = ((($bc)) + 24|0);
 $3 = ((($bc)) + 32|0);
 $4 = ((($bc)) + 8|0);
 $$pre = HEAP32[$0>>2]|0;
 $$1 = $size;$$13 = $data;$6 = $$pre;
 while(1) {
  $5 = ($6|0)==(0|0);
  if ($5) {
   $$2 = $$1;$$24 = $$13;$43 = $6;
  } else {
   $7 = ((($6)) + 4|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ((($6)) + 8|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($10|0)>($8|0);
   if ($11) {
    $12 = (($10) - ($8))|0;
    $13 = ($12|0)>($$1|0);
    $$1$ = $13 ? $$1 : $12;
    $14 = HEAP32[$6>>2]|0;
    $15 = (($14) + ($8)|0);
    _memcpy(($15|0),($$13|0),($$1$|0))|0;
    $16 = HEAP32[$0>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = (($18) + ($$1$))|0;
    HEAP32[$17>>2] = $19;
    $20 = (($$1) - ($$1$))|0;
    $21 = HEAP32[$4>>2]|0;
    $22 = (($21) + ($$1$))|0;
    HEAP32[$4>>2] = $22;
    $23 = (($$13) + ($$1$)|0);
    $$2 = $20;$$24 = $23;$43 = $16;
   } else {
    $$2 = $$1;$$24 = $$13;$43 = $6;
   }
  }
  $24 = ($$2|0)>(0);
  if (!($24)) {
   $ret$2 = 0;
   label = 18;
   break;
  }
  $25 = HEAP32[$1>>2]|0;
  $26 = ($25|0)==(0|0);
  if ($26) {
   $32 = HEAP32[$2>>2]|0;
   $33 = (_malloc(16)|0);
   $34 = ($33|0)==(0|0);
   if ($34) {
    $ret$2 = -2;
    label = 18;
    break;
   }
   $35 = ($32>>>0)<($$2>>>0);
   $36 = $35 ? $$2 : $32;
   $37 = ((($33)) + 8|0);
   HEAP32[$37>>2] = $36;
   $38 = (_malloc($36)|0);
   HEAP32[$33>>2] = $38;
   $39 = ($38|0)==(0|0);
   if ($39) {
    $$lcssa = $33;
    label = 11;
    break;
   }
   $40 = ((($33)) + 4|0);
   HEAP32[$40>>2] = 0;
   $41 = ((($33)) + 12|0);
   HEAP32[$41>>2] = 0;
   $$0$i$i = $33;
  } else {
   $27 = ((($25)) + 12|0);
   $28 = HEAP32[$27>>2]|0;
   HEAP32[$1>>2] = $28;
   HEAP32[$27>>2] = 0;
   $29 = ((($25)) + 4|0);
   HEAP32[$29>>2] = 0;
   $30 = HEAP32[$3>>2]|0;
   $31 = (($30) + -1)|0;
   HEAP32[$3>>2] = $31;
   $$0$i$i = $25;
  }
  $42 = ($43|0)==(0|0);
  if ($42) {
   $45 = HEAP32[$bc>>2]|0;
   $46 = ($45|0)==(0|0);
   if ($46) {
    HEAP32[$bc>>2] = $$0$i$i;
   }
  } else {
   $44 = ((($43)) + 12|0);
   HEAP32[$44>>2] = $$0$i$i;
  }
  HEAP32[$0>>2] = $$0$i$i;
  $$1 = $$2;$$13 = $$24;$6 = $$0$i$i;
 }
 if ((label|0) == 11) {
  _free($$lcssa);
  $ret$2 = -2;
  return ($ret$2|0);
 }
 else if ((label|0) == 18) {
  return ($ret$2|0);
 }
 return (0)|0;
}
function _bc_fill_pool($bc) {
 $bc = $bc|0;
 var $$lcssa15 = 0, $$pre = 0, $$pre7 = 0, $$pre8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($bc)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($bc)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1>>>0)>($3>>>0);
 if ($4) {
  $5 = ((($bc)) + 36|0);
  $31 = $3;$32 = $1;
  while(1) {
   $11 = HEAP32[$5>>2]|0;
   $12 = ((($11)) + 12|0);
   $13 = HEAP32[$12>>2]|0;
   HEAP32[$5>>2] = $13;
   $14 = ($11|0)==(0|0);
   if ($14) {
    $17 = $32;$19 = $31;
   } else {
    $15 = HEAP32[$11>>2]|0;
    _free($15);
    _free($11);
    $$pre = HEAP32[$0>>2]|0;
    $$pre7 = HEAP32[$2>>2]|0;
    $17 = $$pre;$19 = $$pre7;
   }
   $16 = (($17) + -1)|0;
   HEAP32[$0>>2] = $16;
   $18 = ($16>>>0)>($19>>>0);
   if ($18) {
    $31 = $19;$32 = $16;
   } else {
    $7 = $16;$8 = $19;
    break;
   }
  }
 } else {
  $7 = $1;$8 = $3;
 }
 $6 = ($7>>>0)<($8>>>0);
 if (!($6)) {
  return;
 }
 $9 = ((($bc)) + 24|0);
 $10 = ((($bc)) + 36|0);
 $$pre8 = HEAP32[$9>>2]|0;
 $29 = $7;
 while(1) {
  $20 = (_malloc(16)|0);
  $21 = ($20|0)==(0|0);
  if ($21) {
   label = 12;
   break;
  }
  $22 = ((($20)) + 8|0);
  HEAP32[$22>>2] = $$pre8;
  $23 = (_malloc($$pre8)|0);
  HEAP32[$20>>2] = $23;
  $24 = ($23|0)==(0|0);
  if ($24) {
   $$lcssa15 = $20;
   label = 10;
   break;
  }
  $25 = ((($20)) + 4|0);
  HEAP32[$25>>2] = 0;
  $26 = ((($20)) + 12|0);
  $27 = HEAP32[$10>>2]|0;
  HEAP32[$26>>2] = $27;
  HEAP32[$10>>2] = $20;
  $28 = (($29) + 1)|0;
  HEAP32[$0>>2] = $28;
  $30 = ($28>>>0)<($8>>>0);
  if ($30) {
   $29 = $28;
  } else {
   label = 12;
   break;
  }
 }
 if ((label|0) == 10) {
  _free($$lcssa15);
  return;
 }
 else if ((label|0) == 12) {
  return;
 }
}
function _default_init($fr) {
 $fr = $fr|0;
 var $$$i = 0, $$0$i$i = 0, $$0$i1$i = 0, $$0$i3$i = 0, $$1 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phi3Z2D = 0, $$pre$phiZ2D = 0, $$pre2 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $nix_lseek$ = 0, $or$cond = 0, $posix_read$ = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($fr)) + 28092|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(0);
 do {
  if ($2) {
   $3 = ((($fr)) + 27988|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==(0|0);
   if ($5) {
    $7 = ((($fr)) + 27968|0);
    $8 = HEAP32[$7>>2]|0;
    $9 = (_fcntl($8,3,$vararg_buffer1)|0);
    $10 = $9 | 2048;
    $11 = HEAP32[$7>>2]|0;
    HEAP32[$vararg_buffer3>>2] = $10;
    (_fcntl($11,4,$vararg_buffer3)|0);
    $12 = ((($fr)) + 27984|0);
    HEAP32[$12>>2] = 4273;
    $13 = HEAP32[$0>>2]|0;
    $14 = ((($fr)) + 27980|0);
    HEAP32[$14>>2] = $13;
    $15 = ((($fr)) + 27976|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = $16 | 32;
    HEAP32[$15>>2] = $17;
    $$pre$phi3Z2D = $15;$$pre$phiZ2D = $3;$31 = $17;
    break;
   }
   $6 = HEAP32[37544]|0;
   HEAP32[$vararg_buffer>>2] = 1031;
   (_fprintf($6,515288,$vararg_buffer)|0);
   $$1 = -1;
   STACKTOP = sp;return ($$1|0);
  } else {
   $18 = ((($fr)) + 27984|0);
   HEAP32[$18>>2] = 4274;
   $$phi$trans$insert = ((($fr)) + 27976|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0;
   $$pre2 = ((($fr)) + 27988|0);
   $$pre$phi3Z2D = $$phi$trans$insert;$$pre$phiZ2D = $$pre2;$31 = $$pre;
  }
 } while(0);
 $19 = HEAP32[$$pre$phiZ2D>>2]|0;
 $20 = ($19|0)==(0|0);
 $posix_read$ = $20 ? 4275 : $19;
 $21 = ((($fr)) + 28008|0);
 HEAP32[$21>>2] = $posix_read$;
 $22 = ((($fr)) + 27992|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = ($23|0)==(0|0);
 $25 = $24 ? 4276 : $23;
 $26 = ((($fr)) + 28012|0);
 $27 = ((($fr)) + 28336|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ($28|0)>(0);
 $nix_lseek$ = $29 ? 4277 : $25;
 HEAP32[$26>>2] = $nix_lseek$;
 $30 = $31 & 64;
 $32 = ($30|0)==(0);
 if ($32) {
  $39 = ((($fr)) + 27968|0);
  $40 = HEAP32[$39>>2]|0;
  $41 = (FUNCTION_TABLE_iiii[$nix_lseek$ & 8191]($40,0,2)|0);
  $$0$i$i = $41;
  label = 10;
 } else {
  $33 = ((($fr)) + 28000|0);
  $34 = HEAP32[$33>>2]|0;
  $35 = ($34|0)==(0|0);
  if (!($35)) {
   $36 = ((($fr)) + 27972|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = (FUNCTION_TABLE_iiii[$34 & 8191]($37,0,2)|0);
   $$0$i$i = $38;
   label = 10;
  }
 }
 do {
  if ((label|0) == 10) {
   $42 = ($$0$i$i|0)<(0);
   if (!($42)) {
    $43 = HEAP32[$$pre$phi3Z2D>>2]|0;
    $44 = $43 & 64;
    $45 = ($44|0)==(0);
    if ($45) {
     $52 = HEAP32[$26>>2]|0;
     $53 = ((($fr)) + 27968|0);
     $54 = HEAP32[$53>>2]|0;
     $55 = (FUNCTION_TABLE_iiii[$52 & 8191]($54,-128,2)|0);
     $$0$i1$i = $55;
    } else {
     $46 = ((($fr)) + 28000|0);
     $47 = HEAP32[$46>>2]|0;
     $48 = ($47|0)==(0|0);
     if ($48) {
      break;
     }
     $49 = ((($fr)) + 27972|0);
     $50 = HEAP32[$49>>2]|0;
     $51 = (FUNCTION_TABLE_iiii[$47 & 8191]($50,-128,2)|0);
     $$0$i1$i = $51;
    }
    $56 = ($$0$i1$i|0)<(0);
    if (!($56)) {
     $57 = ((($fr)) + 27956|0);
     $58 = HEAP32[$57>>2]|0;
     $59 = ((($58)) + 8|0);
     $60 = HEAP32[$59>>2]|0;
     $61 = ((($fr)) + 28396|0);
     $62 = (FUNCTION_TABLE_iiii[$60 & 8191]($fr,$61,128)|0);
     $63 = ($62|0)==(128);
     if ($63) {
      $64 = (_strncmp($61,515438,3)|0);
      $65 = ($64|0)==(0);
      $66 = (($$0$i$i) + -128)|0;
      $$$i = $65 ? $66 : $$0$i$i;
      $67 = HEAP32[$$pre$phi3Z2D>>2]|0;
      $68 = $67 & 64;
      $69 = ($68|0)==(0);
      if ($69) {
       $76 = HEAP32[$26>>2]|0;
       $77 = ((($fr)) + 27968|0);
       $78 = HEAP32[$77>>2]|0;
       $79 = (FUNCTION_TABLE_iiii[$76 & 8191]($78,0,0)|0);
       $$0$i3$i = $79;
      } else {
       $70 = ((($fr)) + 28000|0);
       $71 = HEAP32[$70>>2]|0;
       $72 = ($71|0)==(0|0);
       if ($72) {
        break;
       }
       $73 = ((($fr)) + 27972|0);
       $74 = HEAP32[$73>>2]|0;
       $75 = (FUNCTION_TABLE_iiii[$71 & 8191]($74,0,0)|0);
       $$0$i3$i = $75;
      }
      $80 = ($$0$i3$i|0)<(0);
      $81 = ($$$i|0)<(1);
      $or$cond = $81 | $80;
      if (!($or$cond)) {
       $82 = ((($fr)) + 27960|0);
       HEAP32[$82>>2] = $$$i;
       $83 = ((($fr)) + 27964|0);
       HEAP32[$83>>2] = 0;
       $84 = HEAP32[$$pre$phi3Z2D>>2]|0;
       $85 = $84 | 4;
       HEAP32[$$pre$phi3Z2D>>2] = $85;
       $86 = (_strncmp($61,515438,3)|0);
       $87 = ($86|0)==(0);
       if (!($87)) {
        $$1 = 0;
        STACKTOP = sp;return ($$1|0);
       }
       $88 = $84 | 6;
       HEAP32[$$pre$phi3Z2D>>2] = $88;
       $89 = ((($fr)) + 28392|0);
       $90 = HEAP32[$89>>2]|0;
       $91 = $90 | 1;
       HEAP32[$89>>2] = $91;
       $$1 = 0;
       STACKTOP = sp;return ($$1|0);
      }
     }
    }
   }
  }
 } while(0);
 $92 = ((($fr)) + 27960|0);
 HEAP32[$92>>2] = -1;
 $93 = ((($fr)) + 27964|0);
 HEAP32[$93>>2] = 0;
 $94 = ((($fr)) + 28068|0);
 $95 = HEAP32[$94>>2]|0;
 $96 = $95 & 256;
 $97 = ($96|0)==(0);
 if ($97) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $98 = ((($fr)) + 27956|0);
 $99 = HEAP32[$98>>2]|0;
 $100 = ($99|0)==(38280|0);
 do {
  if ($100) {
   HEAP32[$98>>2] = (38424);
   $101 = ((($fr)) + 28016|0);
   HEAP32[$101>>2] = 2145;
  } else {
   $102 = ($99|0)==((38328)|0);
   if ($102) {
    HEAP32[$98>>2] = (38472);
    $103 = ((($fr)) + 28016|0);
    HEAP32[$103>>2] = 2154;
    break;
   }
   $104 = $95 & 32;
   $105 = ($104|0)==(0);
   if (!($105)) {
    $$1 = -1;
    STACKTOP = sp;return ($$1|0);
   }
   $106 = HEAP32[37544]|0;
   HEAP32[$vararg_buffer6>>2] = 1090;
   (_fprintf($106,515799,$vararg_buffer6)|0);
   $$1 = -1;
   STACKTOP = sp;return ($$1|0);
  }
 } while(0);
 $107 = ((($fr)) + 28020|0);
 ;HEAP32[$107>>2]=0|0;HEAP32[$107+4>>2]=0|0;HEAP32[$107+8>>2]=0|0;HEAP32[$107+12>>2]=0|0;HEAP32[$107+16>>2]=0|0;HEAP32[$107+20>>2]=0|0;
 HEAP32[$92>>2] = 0;
 $108 = HEAP32[$$pre$phi3Z2D>>2]|0;
 $109 = $108 | 8;
 HEAP32[$$pre$phi3Z2D>>2] = $109;
 $$1 = 0;
 STACKTOP = sp;return ($$1|0);
}
function _timeout_read($fr,$buf,$count) {
 $fr = $fr|0;
 $buf = $buf|0;
 $count = $count|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $fds = 0, $ret$0 = 0, $tv = 0, $vararg_buffer = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $tv = sp + 136|0;
 $fds = sp + 8|0;
 $0 = ((($fr)) + 27980|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$tv>>2] = $1;
 $2 = ((($tv)) + 4|0);
 HEAP32[$2>>2] = 0;
 dest=$fds; stop=dest+128|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $3 = ((($fr)) + 27968|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 & 31;
 $6 = 1 << $5;
 $7 = $4 >>> 5;
 $8 = (($fds) + ($7<<2)|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $6 | $9;
 HEAP32[$8>>2] = $10;
 $11 = (($4) + 1)|0;
 $12 = (_select($11,$fds,0,0,$tv)|0);
 $13 = ($12|0)>(0);
 if ($13) {
  $14 = HEAP32[$3>>2]|0;
  $15 = (_read($14,$buf,$count)|0);
  $ret$0 = $15;
  STACKTOP = sp;return ($ret$0|0);
 }
 $16 = ((($fr)) + 28068|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 & 32;
 $19 = ($18|0)==(0);
 if (!($19)) {
  $ret$0 = -1;
  STACKTOP = sp;return ($ret$0|0);
 }
 $20 = HEAP32[37544]|0;
 HEAP32[$vararg_buffer>>2] = 83;
 (_fprintf($20,515398,$vararg_buffer)|0);
 $ret$0 = -1;
 STACKTOP = sp;return ($ret$0|0);
}
function _plain_read($fr,$buf,$count) {
 $fr = $fr|0;
 $buf = $buf|0;
 $count = $count|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27976|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 64;
 $3 = ($2|0)==(0);
 if ($3) {
  $10 = ((($fr)) + 28008|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ((($fr)) + 27968|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = (FUNCTION_TABLE_iiii[$11 & 8191]($13,$buf,$count)|0);
  $$0$i = $14;
  return ($$0$i|0);
 }
 $4 = ((($fr)) + 27996|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if ($6) {
  $$0$i = -1;
  return ($$0$i|0);
 }
 $7 = ((($fr)) + 27972|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (FUNCTION_TABLE_iiii[$5 & 8191]($8,$buf,$count)|0);
 $$0$i = $9;
 return ($$0$i|0);
}
function _posix_read($fd,$buf,$count) {
 $fd = $fd|0;
 $buf = $buf|0;
 $count = $count|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_read($fd,$buf,$count)|0);
 return ($0|0);
}
function _posix_lseek($fd,$offset,$whence) {
 $fd = $fd|0;
 $offset = $offset|0;
 $whence = $whence|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lseek($fd,$offset,$whence)|0);
 return ($0|0);
}
function _nix_lseek($fd,$offset,$whence) {
 $fd = $fd|0;
 $offset = $offset|0;
 $whence = $whence|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return -1;
}
function _plain_fullread($fr,$buf,$count) {
 $fr = $fr|0;
 $buf = $buf|0;
 $count = $count|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cnt$01 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27984|0);
 $1 = ($count|0)>(0);
 if (!($1)) {
  $$0 = 0;
  return ($$0|0);
 }
 $2 = ((($fr)) + 27976|0);
 $3 = ((($fr)) + 27964|0);
 $cnt$01 = 0;
 while(1) {
  $4 = HEAP32[$0>>2]|0;
  $5 = (($buf) + ($cnt$01)|0);
  $6 = (($count) - ($cnt$01))|0;
  $7 = (FUNCTION_TABLE_iiii[$4 & 8191]($fr,$5,$6)|0);
  $8 = ($7|0)<(0);
  if ($8) {
   $$0 = -1;
   label = 8;
   break;
  }
  $9 = ($7|0)==(0);
  if ($9) {
   $$0 = $cnt$01;
   label = 8;
   break;
  }
  $10 = HEAP32[$2>>2]|0;
  $11 = $10 & 8;
  $12 = ($11|0)==(0);
  if ($12) {
   $13 = HEAP32[$3>>2]|0;
   $14 = (($13) + ($7))|0;
   HEAP32[$3>>2] = $14;
  }
  $15 = (($7) + ($cnt$01))|0;
  $16 = ($15|0)<($count|0);
  if ($16) {
   $cnt$01 = $15;
  } else {
   $$0 = $15;
   label = 8;
   break;
  }
 }
 if ((label|0) == 8) {
  return ($$0|0);
 }
 return (0)|0;
}
function _icy_fullread($fr,$buf,$count) {
 $fr = $fr|0;
 $buf = $buf|0;
 $count = $count|0;
 var $$0$i = 0, $$6 = 0, $$lcssa = 0, $$lcssa106 = 0, $$pre = 0, $$pre56 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cnt$0$be = 0, $cnt$01$i = 0;
 var $cnt$033 = 0, $cnt$033$lcssa109 = 0, $cnt$1 = 0, $cnt$2 = 0, $cnt$2$ph10 = 0, $left$032 = 0, $temp_buff = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer10 = 0, $vararg_buffer14 = 0, $vararg_buffer4 = 0, $vararg_buffer7 = 0, $vararg_ptr13 = 0, $vararg_ptr17 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer14 = sp + 40|0;
 $vararg_buffer10 = sp + 32|0;
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $temp_buff = sp + 48|0;
 $0 = ((($fr)) + 27976|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 4;
 $3 = ($2|0)==(0);
 if (!($3)) {
  $14 = ((($fr)) + 28068|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = $15 & 32;
  $17 = ($16|0)==(0);
  if (!($17)) {
   $$6 = -1;
   STACKTOP = sp;return ($$6|0);
  }
  $18 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer>>2] = 97;
  (_fprintf($18,515442,$vararg_buffer)|0);
  $$6 = -1;
  STACKTOP = sp;return ($$6|0);
 }
 $4 = ($count|0)>(0);
 if (!($4)) {
  $$6 = 0;
  STACKTOP = sp;return ($$6|0);
 }
 $5 = ((($fr)) + 28584|0);
 $6 = ((($fr)) + 27984|0);
 $7 = ((($fr)) + 27964|0);
 $8 = ((($fr)) + 28580|0);
 $9 = ((($fr)) + 28068|0);
 $10 = HEAP32[37544]|0;
 $11 = ((($fr)) + 27956|0);
 $12 = ((($fr)) + 28576|0);
 $13 = ((($fr)) + 28392|0);
 $$pre = HEAP32[$5>>2]|0;
 $21 = $$pre;$cnt$033 = 0;
 L11: while(1) {
  $19 = (($count) - ($cnt$033))|0;
  $20 = ($21|0)<($19|0);
  if ($20) {
   $22 = ($21|0)>(0);
   if ($22) {
    $23 = HEAP32[$6>>2]|0;
    $24 = (($buf) + ($cnt$033)|0);
    $25 = (FUNCTION_TABLE_iiii[$23 & 8191]($fr,$24,$21)|0);
    $26 = ($25|0)<(1);
    if ($26) {
     $$lcssa = $25;$cnt$033$lcssa109 = $cnt$033;
     label = 9;
     break;
    }
    $31 = HEAP32[$0>>2]|0;
    $32 = $31 & 8;
    $33 = ($32|0)==(0);
    if ($33) {
     $34 = HEAP32[$7>>2]|0;
     $35 = (($34) + ($25))|0;
     HEAP32[$7>>2] = $35;
    }
    $36 = (($25) + ($cnt$033))|0;
    $37 = HEAP32[$5>>2]|0;
    $38 = (($37) - ($25))|0;
    HEAP32[$5>>2] = $38;
    $39 = ($38|0)>(0);
    if ($39) {
     $110 = $38;$cnt$2$ph10 = $36;
    } else {
     $cnt$1 = $36;
     label = 15;
    }
   } else {
    $cnt$1 = $cnt$033;
    label = 15;
   }
   if ((label|0) == 15) {
    label = 0;
    $40 = HEAP32[$6>>2]|0;
    $41 = (FUNCTION_TABLE_iiii[$40 & 8191]($fr,$temp_buff,1)|0);
    $42 = ($41|0)<(0);
    if ($42) {
     label = 16;
     break;
    }
    $46 = ($41|0)==(0);
    if ($46) {
     $cnt$2 = $cnt$1;
     label = 39;
     break;
    }
    $47 = HEAP32[$0>>2]|0;
    $48 = $47 & 8;
    $49 = ($48|0)==(0);
    if ($49) {
     $50 = HEAP32[$7>>2]|0;
     $51 = (($50) + ($41))|0;
     HEAP32[$7>>2] = $51;
    }
    $52 = HEAP8[$temp_buff>>0]|0;
    $53 = $52&255;
    $54 = $53 << 4;
    $55 = ($52<<24>>24)==(0);
    do {
     if (!($55)) {
      $56 = $54 | 1;
      $57 = (_malloc($56)|0);
      $58 = ($57|0)==(0|0);
      if ($58) {
       $78 = HEAP32[$9>>2]|0;
       $79 = $78 & 32;
       $80 = ($79|0)==(0);
       if ($80) {
        HEAP32[$vararg_buffer10>>2] = 177;
        $vararg_ptr13 = ((($vararg_buffer10)) + 4|0);
        HEAP32[$vararg_ptr13>>2] = $54;
        (_fprintf($10,515647,$vararg_buffer10)|0);
       }
       $81 = HEAP32[$11>>2]|0;
       $82 = ((($81)) + 20|0);
       $83 = HEAP32[$82>>2]|0;
       (FUNCTION_TABLE_iii[$83 & 8191]($fr,$54)|0);
       break;
      }
      $59 = (($57) + ($54)|0);
      $left$032 = $54;
      while(1) {
       $62 = HEAP32[$6>>2]|0;
       $63 = (0 - ($left$032))|0;
       $64 = (($59) + ($63)|0);
       $65 = (FUNCTION_TABLE_iiii[$62 & 8191]($fr,$64,$left$032)|0);
       $66 = ($65|0)<(1);
       $61 = (($left$032) - ($65))|0;
       if ($66) {
        label = 26;
        break L11;
       }
       $60 = ($61|0)>(0);
       if ($60) {
        $left$032 = $61;
       } else {
        $$lcssa106 = $65;
        break;
       }
      }
      $$pre56 = HEAP32[$0>>2]|0;
      HEAP8[$59>>0] = 0;
      $70 = $$pre56 & 8;
      $71 = ($70|0)==(0);
      if ($71) {
       $72 = HEAP32[$7>>2]|0;
       $73 = (($72) + ($$lcssa106))|0;
       HEAP32[$7>>2] = $73;
      }
      $74 = HEAP32[$12>>2]|0;
      $75 = ($74|0)==(0|0);
      if (!($75)) {
       _free($74);
      }
      HEAP32[$12>>2] = $57;
      $76 = HEAP32[$13>>2]|0;
      $77 = $76 | 4;
      HEAP32[$13>>2] = $77;
     }
    } while(0);
    $84 = HEAP32[$8>>2]|0;
    HEAP32[$5>>2] = $84;
    $110 = $84;$cnt$2$ph10 = $cnt$1;
   }
   $111 = $110;$cnt$0$be = $cnt$2$ph10;
  } else {
   $85 = (($buf) + ($cnt$033)|0);
   $86 = ($19|0)>(0);
   if ($86) {
    $cnt$01$i = 0;
   } else {
    $$6 = $cnt$033;
    label = 52;
    break;
   }
   while(1) {
    $87 = HEAP32[$6>>2]|0;
    $88 = (($85) + ($cnt$01$i)|0);
    $89 = (($19) - ($cnt$01$i))|0;
    $90 = (FUNCTION_TABLE_iiii[$87 & 8191]($fr,$88,$89)|0);
    $91 = ($90|0)<(0);
    if ($91) {
     $104 = $19;
     label = 47;
     break L11;
    }
    $92 = ($90|0)==(0);
    if ($92) {
     $$0$i = $cnt$01$i;
     break;
    }
    $93 = HEAP32[$0>>2]|0;
    $94 = $93 & 8;
    $95 = ($94|0)==(0);
    if ($95) {
     $96 = HEAP32[$7>>2]|0;
     $97 = (($96) + ($90))|0;
     HEAP32[$7>>2] = $97;
    }
    $98 = (($90) + ($cnt$01$i))|0;
    $99 = ($98|0)<($19|0);
    if ($99) {
     $cnt$01$i = $98;
    } else {
     $$0$i = $98;
     break;
    }
   }
   $100 = ($$0$i|0)<(0);
   if ($100) {
    $104 = $19;
    label = 47;
    break;
   }
   $105 = ($$0$i|0)==(0);
   if ($105) {
    $$6 = $cnt$033;
    label = 52;
    break;
   }
   $106 = (($$0$i) + ($cnt$033))|0;
   $107 = HEAP32[$5>>2]|0;
   $108 = (($107) - ($$0$i))|0;
   HEAP32[$5>>2] = $108;
   $111 = $108;$cnt$0$be = $106;
  }
  $109 = ($cnt$0$be|0)<($count|0);
  if ($109) {
   $21 = $111;$cnt$033 = $cnt$0$be;
  } else {
   $$6 = $cnt$0$be;
   label = 52;
   break;
  }
 }
 if ((label|0) == 9) {
  $27 = ($$lcssa|0)==(0);
  if ($27) {
   $cnt$2 = $cnt$033$lcssa109;
   label = 39;
  } else {
   $28 = HEAP32[$9>>2]|0;
   $29 = $28 & 32;
   $30 = ($29|0)==(0);
   if ($30) {
    HEAP32[$vararg_buffer1>>2] = 126;
    (_fprintf($10,515526,$vararg_buffer1)|0);
    label = 37;
   } else {
    label = 37;
   }
  }
 }
 else if ((label|0) == 16) {
  $43 = HEAP32[$9>>2]|0;
  $44 = $43 & 32;
  $45 = ($44|0)==(0);
  if ($45) {
   HEAP32[$vararg_buffer4>>2] = 145;
   (_fprintf($10,515567,$vararg_buffer4)|0);
   label = 37;
  } else {
   label = 37;
  }
 }
 else if ((label|0) == 26) {
  $67 = HEAP32[$9>>2]|0;
  $68 = $67 & 32;
  $69 = ($68|0)==(0);
  if ($69) {
   HEAP32[$vararg_buffer7>>2] = 164;
   (_fprintf($10,515607,$vararg_buffer7)|0);
   label = 37;
  } else {
   label = 37;
  }
 }
 else if ((label|0) == 47) {
  $101 = HEAP32[$9>>2]|0;
  $102 = $101 & 32;
  $103 = ($102|0)==(0);
  if (!($103)) {
   $$6 = -1;
   STACKTOP = sp;return ($$6|0);
  }
  HEAP32[$vararg_buffer14>>2] = 186;
  $vararg_ptr17 = ((($vararg_buffer14)) + 4|0);
  HEAP32[$vararg_ptr17>>2] = $104;
  (_fprintf($10,515752,$vararg_buffer14)|0);
  $$6 = -1;
  STACKTOP = sp;return ($$6|0);
 }
 else if ((label|0) == 52) {
  STACKTOP = sp;return ($$6|0);
 }
 if ((label|0) == 37) {
  $$6 = -1;
  STACKTOP = sp;return ($$6|0);
 }
 else if ((label|0) == 39) {
  $$6 = $cnt$2;
  STACKTOP = sp;return ($$6|0);
 }
 return (0)|0;
}
function _stream_close($fr) {
 $fr = $fr|0;
 var $$pre = 0, $$pre$i = 0, $$pre1 = 0, $$pre3 = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27976|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 1;
 $3 = ($2|0)==(0);
 $$pre3 = ((($fr)) + 27968|0);
 if ($3) {
  $6 = $1;
 } else {
  $4 = HEAP32[$$pre3>>2]|0;
  (_INT123_compat_close($4)|0);
  $$pre = HEAP32[$0>>2]|0;
  $6 = $$pre;
 }
 HEAP32[$$pre3>>2] = 0;
 $5 = $6 & 8;
 $7 = ($5|0)==(0);
 if ($7) {
  $27 = $6;
 } else {
  $8 = ((($fr)) + 28020|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)==(0|0);
  if (!($10)) {
   $11 = ((($fr)) + 28052|0);
   $12 = ((($fr)) + 28048|0);
   $13 = ((($fr)) + 28056|0);
   $15 = $9;
   while(1) {
    $14 = ((($15)) + 12|0);
    $16 = HEAP32[$14>>2]|0;
    HEAP32[$8>>2] = $16;
    $17 = HEAP32[$11>>2]|0;
    $18 = HEAP32[$12>>2]|0;
    $19 = ($17>>>0)<($18>>>0);
    if ($19) {
     $20 = $16;
     $21 = HEAP32[$13>>2]|0;
     HEAP32[$14>>2] = $21;
     HEAP32[$13>>2] = $15;
     $22 = (($17) + 1)|0;
     HEAP32[$11>>2] = $22;
     $24 = $20;
    } else {
     $25 = HEAP32[$15>>2]|0;
     _free($25);
     _free($15);
     $$pre$i = HEAP32[$8>>2]|0;
     $24 = $$pre$i;
    }
    $23 = ($24|0)==(0|0);
    if ($23) {
     break;
    } else {
     $15 = $24;
    }
   }
  }
  _bc_fill_pool($8);
  ;HEAP32[$8>>2]=0|0;HEAP32[$8+4>>2]=0|0;HEAP32[$8+8>>2]=0|0;HEAP32[$8+12>>2]=0|0;HEAP32[$8+16>>2]=0|0;HEAP32[$8+20>>2]=0|0;
  $$pre1 = HEAP32[$0>>2]|0;
  $27 = $$pre1;
 }
 $26 = $27 & 64;
 $28 = ($26|0)==(0);
 if ($28) {
  return;
 }
 $29 = ((($fr)) + 28004|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ($30|0)==(0|0);
 $$pre4 = ((($fr)) + 27972|0);
 if (!($31)) {
  $32 = HEAP32[$$pre4>>2]|0;
  FUNCTION_TABLE_vi[$30 & 8191]($32);
 }
 HEAP32[$$pre4>>2] = 0;
 return;
}
function _generic_head_read($fr,$newhead) {
 $fr = $fr|0;
 $newhead = $newhead|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $hbuf = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $hbuf = sp;
 $0 = ((($fr)) + 27956|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiii[$3 & 8191]($fr,$hbuf,4)|0);
 switch ($4|0) {
 case 4:  {
  $5 = HEAP8[$hbuf>>0]|0;
  $6 = $5&255;
  $7 = $6 << 24;
  $8 = ((($hbuf)) + 1|0);
  $9 = HEAP8[$8>>0]|0;
  $10 = $9&255;
  $11 = $10 << 16;
  $12 = $11 | $7;
  $13 = ((($hbuf)) + 2|0);
  $14 = HEAP8[$13>>0]|0;
  $15 = $14&255;
  $16 = $15 << 8;
  $17 = $12 | $16;
  $18 = ((($hbuf)) + 3|0);
  $19 = HEAP8[$18>>0]|0;
  $20 = $19&255;
  $21 = $17 | $20;
  HEAP32[$newhead>>2] = $21;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 case -10:  {
  $$0 = -10;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 default: {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 }
 return (0)|0;
}
function _generic_head_shift($fr,$head) {
 $fr = $fr|0;
 $head = $head|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hbuf = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $hbuf = sp;
 $0 = ((($fr)) + 27956|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiii[$3 & 8191]($fr,$hbuf,1)|0);
 switch ($4|0) {
 case 1:  {
  $5 = HEAP32[$head>>2]|0;
  $6 = $5 << 8;
  $7 = HEAP8[$hbuf>>0]|0;
  $8 = $7&255;
  $9 = $8 | $6;
  HEAP32[$head>>2] = $9;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 case -10:  {
  $$0 = -10;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 default: {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 }
 return (0)|0;
}
function _stream_skip_bytes($fr,$len) {
 $fr = $fr|0;
 $len = $len|0;
 var $$0$i$i = 0, $$02 = 0, $$02$ = 0, $$2 = 0, $$3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buf = 0, $ret$0$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1024|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $buf = sp;
 $0 = ((($fr)) + 27976|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 4;
 $3 = ($2|0)==(0);
 if (!($3)) {
  $4 = $1 & 64;
  $5 = ($4|0)==(0);
  if ($5) {
   $12 = ((($fr)) + 28012|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = ((($fr)) + 27968|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = (FUNCTION_TABLE_iiii[$13 & 8191]($15,$len,1)|0);
   $$0$i$i = $16;
   label = 6;
  } else {
   $6 = ((($fr)) + 28000|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0|0);
   if ($8) {
    label = 8;
   } else {
    $9 = ((($fr)) + 27972|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (FUNCTION_TABLE_iiii[$7 & 8191]($10,$len,1)|0);
    $$0$i$i = $11;
    label = 6;
   }
  }
  if ((label|0) == 6) {
   $17 = ($$0$i$i|0)>(-1);
   if ($17) {
    $18 = ((($fr)) + 27964|0);
    HEAP32[$18>>2] = $$0$i$i;
    $ret$0$i = $$0$i$i;
   } else {
    label = 8;
   }
  }
  if ((label|0) == 8) {
   $19 = ((($fr)) + 28376|0);
   HEAP32[$19>>2] = 40;
   $ret$0$i = -1;
  }
  $20 = ($ret$0$i|0)>(-1);
  $21 = $20 ? $ret$0$i : -1;
  $$3 = $21;
  STACKTOP = sp;return ($$3|0);
 }
 $22 = ($len|0)>(-1);
 if ($22) {
  $23 = ((($fr)) + 27956|0);
  $$02 = $len;
  while(1) {
   $24 = ($$02|0)>(0);
   if (!($24)) {
    label = 15;
    break;
   }
   $25 = ($$02|0)<(1024);
   $26 = $25 ? $$02 : 1024;
   $27 = HEAP32[$23>>2]|0;
   $28 = ((($27)) + 8|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = (FUNCTION_TABLE_iiii[$29 & 8191]($fr,$buf,$26)|0);
   $31 = ($30|0)<(0);
   if ($31) {
    $$2 = $30;
    break;
   }
   $32 = ($30|0)==(0);
   $$02$ = (($$02) - ($30))|0;
   if ($32) {
    label = 15;
    break;
   } else {
    $$02 = $$02$;
   }
  }
  if ((label|0) == 15) {
   $33 = HEAP32[$23>>2]|0;
   $34 = ((($33)) + 36|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = (FUNCTION_TABLE_ii[$35 & 8191]($fr)|0);
   $$2 = $36;
  }
  $$3 = $$2;
  STACKTOP = sp;return ($$3|0);
 }
 $37 = $1 & 8;
 $38 = ($37|0)==(0);
 if ($38) {
  $50 = ((($fr)) + 28376|0);
  HEAP32[$50>>2] = 23;
  $$3 = -1;
  STACKTOP = sp;return ($$3|0);
 }
 $39 = ((($fr)) + 28032|0);
 $40 = HEAP32[$39>>2]|0;
 $41 = (0 - ($len))|0;
 $42 = ($40|0)<($41|0);
 if ($42) {
  $49 = ((($fr)) + 28376|0);
  HEAP32[$49>>2] = 23;
  $$3 = -1;
  STACKTOP = sp;return ($$3|0);
 } else {
  $43 = (($40) + ($len))|0;
  HEAP32[$39>>2] = $43;
  $44 = ((($fr)) + 27956|0);
  $45 = HEAP32[$44>>2]|0;
  $46 = ((($45)) + 36|0);
  $47 = HEAP32[$46>>2]|0;
  $48 = (FUNCTION_TABLE_ii[$47 & 8191]($fr)|0);
  $$3 = $48;
  STACKTOP = sp;return ($$3|0);
 }
 return (0)|0;
}
function _generic_read_frame_body($fr,$buf,$size) {
 $fr = $fr|0;
 $buf = $buf|0;
 $size = $size|0;
 var $$ = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27956|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiii[$3 & 8191]($fr,$buf,$size)|0);
 $5 = ($4|0)==($size|0);
 $$ = $5 ? $4 : -10;
 return ($$|0);
}
function _stream_back_bytes($fr,$bytes) {
 $fr = $fr|0;
 $bytes = $bytes|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27956|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 36|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($fr)|0);
 $5 = (($4) - ($bytes))|0;
 $6 = ($5|0)<(0);
 if ($6) {
  $$0 = -1;
  return ($$0|0);
 }
 $7 = (0 - ($bytes))|0;
 $8 = (_stream_skip_bytes($fr,$7)|0);
 $not$ = ($8|0)!=($5|0);
 $$ = $not$ << 31 >> 31;
 $$0 = $$;
 return ($$0|0);
}
function _stream_seek_frame($fr,$newframe) {
 $fr = $fr|0;
 $newframe = $newframe|0;
 var $$0 = 0, $$1 = 0, $$ph = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $preframe = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $preframe = sp;
 $0 = ((($fr)) + 27976|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 4;
 $3 = ($2|0)==(0);
 $4 = ((($fr)) + 19748|0);
 if ($3) {
  $5 = HEAP32[$4>>2]|0;
  $6 = ($5|0)>($newframe|0);
  if ($6) {
   $29 = ((($fr)) + 28376|0);
   HEAP32[$29>>2] = 23;
   $$1 = -1;
   STACKTOP = sp;return ($$1|0);
  }
 }
 $7 = (_INT123_frame_index_find($fr,$newframe,$preframe)|0);
 $8 = HEAP32[$4>>2]|0;
 $9 = ($8|0)>=($newframe|0);
 $10 = HEAP32[$preframe>>2]|0;
 $11 = ($8|0)<($10|0);
 $or$cond = $9 | $11;
 if ($or$cond) {
  $12 = ((($fr)) + 27956|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ((($13)) + 36|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = (FUNCTION_TABLE_ii[$15 & 8191]($fr)|0);
  $17 = (($7) - ($16))|0;
  $18 = HEAP32[$12>>2]|0;
  $19 = ((($18)) + 20|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = (FUNCTION_TABLE_iii[$20 & 8191]($fr,$17)|0);
  $22 = ($21|0)==($7|0);
  if ($22) {
   $23 = HEAP32[$preframe>>2]|0;
   $24 = (($23) + -1)|0;
   HEAP32[$4>>2] = $24;
   $$ph = $24;
   label = 6;
  } else {
   $$0 = -1;
  }
 } else {
  $$ph = $8;
  label = 6;
 }
 L9: do {
  if ((label|0) == 6) {
   $25 = ($$ph|0)<($newframe|0);
   if ($25) {
    while(1) {
     $26 = (_INT123_read_frame($fr)|0);
     $27 = ($26|0)==(0);
     if ($27) {
      $$0 = 0;
      break L9;
     }
     $$pre = HEAP32[$4>>2]|0;
     $28 = ($$pre|0)<($newframe|0);
     if (!($28)) {
      $$0 = 0;
      break;
     }
    }
   } else {
    $$0 = 0;
   }
  }
 } while(0);
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0);
}
function _generic_tell($fr) {
 $fr = $fr|0;
 var $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27976|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 8;
 $3 = ($2|0)==(0);
 if ($3) {
  $$phi$trans$insert = ((($fr)) + 27964|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $10 = $$pre;
  return ($10|0);
 } else {
  $4 = ((($fr)) + 28040|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ((($fr)) + 28032|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (($7) + ($5))|0;
  $9 = ((($fr)) + 27964|0);
  HEAP32[$9>>2] = $8;
  $10 = $8;
  return ($10|0);
 }
 return (0)|0;
}
function _stream_rewind($fr) {
 $fr = $fr|0;
 var $$0$i$i = 0, $$pre = 0, $$pre$phiZ2D = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ret$0$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 27976|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 4;
 $3 = ($2|0)==(0);
 if ($3) {
  $22 = $1;
 } else {
  $4 = $1 & 64;
  $5 = ($4|0)==(0);
  if ($5) {
   $12 = ((($fr)) + 28012|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = ((($fr)) + 27968|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = (FUNCTION_TABLE_iiii[$13 & 8191]($15,0,0)|0);
   $$0$i$i = $16;
   label = 6;
  } else {
   $6 = ((($fr)) + 28000|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0|0);
   if ($8) {
    label = 8;
   } else {
    $9 = ((($fr)) + 27972|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (FUNCTION_TABLE_iiii[$7 & 8191]($10,0,0)|0);
    $$0$i$i = $11;
    label = 6;
   }
  }
  if ((label|0) == 6) {
   $17 = ($$0$i$i|0)>(-1);
   if ($17) {
    $18 = ((($fr)) + 27964|0);
    HEAP32[$18>>2] = $$0$i$i;
    $$pre$phiZ2D = $18;$ret$0$i = $$0$i$i;
   } else {
    label = 8;
   }
  }
  if ((label|0) == 8) {
   $19 = ((($fr)) + 28376|0);
   HEAP32[$19>>2] = 40;
   $$pre1 = ((($fr)) + 27964|0);
   $$pre$phiZ2D = $$pre1;$ret$0$i = -1;
  }
  HEAP32[$$pre$phiZ2D>>2] = $ret$0$i;
  $20 = ((($fr)) + 28040|0);
  HEAP32[$20>>2] = $ret$0$i;
  $$pre = HEAP32[$0>>2]|0;
  $22 = $$pre;
 }
 $21 = $22 & 8;
 $23 = ($21|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($fr)) + 28032|0);
 HEAP32[$24>>2] = 0;
 $25 = ((($fr)) + 28036|0);
 HEAP32[$25>>2] = 0;
 $26 = ((($fr)) + 28040|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ((($fr)) + 27964|0);
 HEAP32[$28>>2] = $27;
 return;
}
function _feed_init($fr) {
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 28020|0);
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;HEAP32[$0+12>>2]=0|0;HEAP32[$0+16>>2]=0|0;HEAP32[$0+20>>2]=0|0;
 _bc_fill_pool($0);
 $1 = ((($fr)) + 27960|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($fr)) + 27964|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($fr)) + 27976|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 | 8;
 HEAP32[$3>>2] = $5;
 return 0;
}
function _feed_read($fr,$out,$count) {
 $fr = $fr|0;
 $out = $out|0;
 $count = $count|0;
 var $$$i = 0, $$$i17 = 0, $$0 = 0, $$0$i = 0, $$lcssa27 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b$0$i = 0, $b$04$i = 0;
 var $b$06$i = 0, $b$06$i$lcssa = 0, $b$11$i19 = 0, $b$11$i19$phi = 0, $offset$05$i = 0, $offset$05$i$lcssa = 0, $offset$13$i18 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 28028|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 28032|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) - ($3))|0;
 $5 = ($4|0)<($count|0);
 L1: do {
  if ($5) {
   $8 = ((($fr)) + 28036|0);
   $9 = HEAP32[$8>>2]|0;
   HEAP32[$2>>2] = $9;
   $$0$i = -10;
  } else {
   $6 = ((($fr)) + 28020|0);
   $b$04$i = HEAP32[$6>>2]|0;
   $7 = ($b$04$i|0)==(0|0);
   if ($7) {
    $$0$i = 0;
   } else {
    $b$06$i = $b$04$i;$offset$05$i = 0;
    while(1) {
     $10 = ((($b$06$i)) + 4|0);
     $11 = HEAP32[$10>>2]|0;
     $12 = (($11) + ($offset$05$i))|0;
     $13 = ($12|0)>($3|0);
     if ($13) {
      $$lcssa27 = $11;$b$06$i$lcssa = $b$06$i;$offset$05$i$lcssa = $offset$05$i;
      break;
     }
     $26 = ((($b$06$i)) + 12|0);
     $b$0$i = HEAP32[$26>>2]|0;
     $27 = ($b$0$i|0)==(0|0);
     if ($27) {
      $$0$i = 0;
      break L1;
     } else {
      $b$06$i = $b$0$i;$offset$05$i = $12;
     }
    }
    $14 = ($count|0)>(0);
    if ($14) {
     $15 = (($3) - ($offset$05$i$lcssa))|0;
     $16 = (($$lcssa27) - ($15))|0;
     $17 = ($16|0)<($count|0);
     $$$i17 = $17 ? $16 : $count;
     $18 = HEAP32[$b$06$i$lcssa>>2]|0;
     $19 = (($18) + ($15)|0);
     _memcpy(($out|0),($19|0),($$$i17|0))|0;
     $20 = HEAP32[$2>>2]|0;
     $21 = (($20) + ($$$i17))|0;
     HEAP32[$2>>2] = $21;
     $22 = ((($b$06$i$lcssa)) + 12|0);
     $23 = HEAP32[$22>>2]|0;
     $24 = ($23|0)!=(0|0);
     $25 = $17 & $24;
     if ($25) {
      $31 = $23;$33 = $21;$35 = $$$i17;$b$11$i19 = $b$06$i$lcssa;$offset$13$i18 = $offset$05$i$lcssa;
      while(1) {
       $28 = ((($b$11$i19)) + 4|0);
       $29 = HEAP32[$28>>2]|0;
       $30 = (($29) + ($offset$13$i18))|0;
       $$phi$trans$insert = ((($31)) + 4|0);
       $$pre = HEAP32[$$phi$trans$insert>>2]|0;
       $32 = (($33) - ($30))|0;
       $34 = (($count) - ($35))|0;
       $36 = (($$pre) - ($32))|0;
       $37 = ($34|0)>($36|0);
       $$$i = $37 ? $36 : $34;
       $38 = (($out) + ($35)|0);
       $39 = HEAP32[$31>>2]|0;
       $40 = (($39) + ($32)|0);
       _memcpy(($38|0),($40|0),($$$i|0))|0;
       $41 = (($$$i) + ($35))|0;
       $42 = HEAP32[$2>>2]|0;
       $43 = (($42) + ($$$i))|0;
       HEAP32[$2>>2] = $43;
       $44 = ((($31)) + 12|0);
       $45 = HEAP32[$44>>2]|0;
       $46 = ($41|0)<($count|0);
       $47 = ($45|0)!=(0|0);
       $48 = $46 & $47;
       if ($48) {
        $b$11$i19$phi = $31;$31 = $45;$33 = $43;$35 = $41;$offset$13$i18 = $30;$b$11$i19 = $b$11$i19$phi;
       } else {
        $$0$i = $41;
        break;
       }
      }
     } else {
      $$0$i = $$$i17;
     }
    } else {
     $$0$i = 0;
    }
   }
  }
 } while(0);
 $49 = ($$0$i|0)<(0);
 $50 = ($$0$i|0)==($count|0);
 $or$cond = $49 | $50;
 $$0 = $or$cond ? $$0$i : -1;
 return ($$0|0);
}
function _feed_skip_bytes($fr,$len) {
 $fr = $fr|0;
 $len = $len|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($len|0)>(-1);
 if (!($0)) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ((($fr)) + 28028|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($fr)) + 28032|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2) - ($4))|0;
 $6 = ($5|0)<($len|0);
 if ($6) {
  $7 = ((($fr)) + 28036|0);
  $8 = HEAP32[$7>>2]|0;
  HEAP32[$3>>2] = $8;
  $$0 = -10;
  return ($$0|0);
 }
 $9 = (($4) + ($len))|0;
 HEAP32[$3>>2] = $9;
 $10 = ($9|0)<(0);
 if ($10) {
  $$0 = $9;
  return ($$0|0);
 }
 $11 = ((($fr)) + 28040|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($12) + ($9))|0;
 $$0 = $13;
 return ($$0|0);
}
function _feed_back_bytes($fr,$bytes) {
 $fr = $fr|0;
 $bytes = $bytes|0;
 var $$0 = 0, $$0$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($bytes|0)>(-1);
 do {
  if ($0) {
   $1 = ((($fr)) + 28032|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = ($2|0)<($bytes|0);
   if ($3) {
    $$0$in = -1;
   } else {
    $4 = (($2) - ($bytes))|0;
    HEAP32[$1>>2] = $4;
    $$0$in = $4;
   }
  } else {
   $5 = (0 - ($bytes))|0;
   $6 = ((($fr)) + 28028|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ((($fr)) + 28032|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (($7) - ($9))|0;
   $11 = ($10|0)<($5|0);
   if ($11) {
    $12 = ((($fr)) + 28036|0);
    $13 = HEAP32[$12>>2]|0;
    HEAP32[$8>>2] = $13;
    $$0$in = -10;
    break;
   }
   $14 = (($9) - ($bytes))|0;
   HEAP32[$8>>2] = $14;
   $15 = ($14|0)<(0);
   if ($15) {
    $$0$in = $14;
   } else {
    $16 = ((($fr)) + 28040|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = (($17) + ($14))|0;
    $$0$in = $18;
   }
  }
 } while(0);
 $$0 = $$0$in >> 31;
 return ($$0|0);
}
function _feed_seek_frame($fr,$num) {
 $fr = $fr|0;
 $num = $num|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return -1;
}
function _buffered_forget($fr) {
 $fr = $fr|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b$0$lcssa$i = 0, $b$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 28020|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fr)) + 28032|0);
 $3 = ($1|0)==(0|0);
 L1: do {
  if ($3) {
   $$pre = ((($fr)) + 28040|0);
   $$pre$phiZ2D = $$pre;$b$0$lcssa$i = 0;
  } else {
   $4 = ((($fr)) + 28024|0);
   $5 = ((($fr)) + 28040|0);
   $6 = ((($fr)) + 28028|0);
   $7 = ((($fr)) + 28052|0);
   $8 = ((($fr)) + 28048|0);
   $9 = ((($fr)) + 28056|0);
   $b$01$i = $1;
   while(1) {
    $10 = HEAP32[$2>>2]|0;
    $11 = ((($b$01$i)) + 4|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = ($10|0)<($12|0);
    if ($13) {
     $$pre$phiZ2D = $5;$b$0$lcssa$i = $b$01$i;
     break L1;
    }
    $14 = ((($b$01$i)) + 12|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = ($15|0)==(0|0);
    if ($16) {
     HEAP32[$4>>2] = 0;
    }
    $17 = HEAP32[$5>>2]|0;
    $18 = (($17) + ($12))|0;
    HEAP32[$5>>2] = $18;
    $19 = (($10) - ($12))|0;
    HEAP32[$2>>2] = $19;
    $20 = HEAP32[$6>>2]|0;
    $21 = (($20) - ($12))|0;
    HEAP32[$6>>2] = $21;
    $22 = HEAP32[$7>>2]|0;
    $23 = HEAP32[$8>>2]|0;
    $24 = ($22>>>0)<($23>>>0);
    if ($24) {
     $25 = HEAP32[$9>>2]|0;
     HEAP32[$14>>2] = $25;
     HEAP32[$9>>2] = $b$01$i;
     $26 = (($22) + 1)|0;
     HEAP32[$7>>2] = $26;
    } else {
     $27 = HEAP32[$b$01$i>>2]|0;
     _free($27);
     _free($b$01$i);
    }
    if ($16) {
     $$pre$phiZ2D = $5;$b$0$lcssa$i = 0;
     break;
    } else {
     $b$01$i = $15;
    }
   }
  }
 } while(0);
 HEAP32[$0>>2] = $b$0$lcssa$i;
 $28 = HEAP32[$2>>2]|0;
 $29 = ((($fr)) + 28036|0);
 HEAP32[$29>>2] = $28;
 $30 = HEAP32[$$pre$phiZ2D>>2]|0;
 $31 = (($30) + ($28))|0;
 $32 = ((($fr)) + 27964|0);
 HEAP32[$32>>2] = $31;
 return;
}
function _buffered_fullread($fr,$out,$count) {
 $fr = $fr|0;
 $out = $out|0;
 $count = $count|0;
 var $$$i = 0, $$$i34 = 0, $$0$i = 0, $$25 = 0, $$4 = 0, $$count = 0, $$lcssa45 = 0, $$lcssa47 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $b$0$i = 0, $b$04$i = 0, $b$06$i = 0, $b$06$i$lcssa = 0, $b$11$i36 = 0, $b$11$i36$phi = 0, $need$0 = 0, $offset$05$i = 0, $offset$05$i$lcssa = 0, $offset$13$i35 = 0, $readbuf = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer5 = 0, $vararg_ptr4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 4128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $readbuf = sp + 24|0;
 $0 = ((($fr)) + 28020|0);
 $1 = ((($fr)) + 28028|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($fr)) + 28032|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2) - ($4))|0;
 $6 = ($5|0)<($count|0);
 do {
  if ($6) {
   $7 = (($count) - ($2))|0;
   $8 = (($7) + ($4))|0;
   $9 = ((($fr)) + 28016|0);
   $need$0 = $8;
   while(1) {
    $10 = ($need$0|0)>(0);
    if (!($10)) {
     label = 15;
     break;
    }
    $11 = HEAP32[$9>>2]|0;
    $12 = (FUNCTION_TABLE_iiii[$11 & 8191]($fr,$readbuf,4096)|0);
    $13 = ($12|0)<(0);
    if ($13) {
     label = 5;
     break;
    }
    $19 = ($12|0)>(0);
    if (!($19)) {
     label = 12;
     break;
    }
    $20 = (_bc_add($0,$readbuf,$12)|0);
    $21 = ($20|0)==(0);
    if (!($21)) {
     $$lcssa47 = $20;
     label = 9;
     break;
    }
    $27 = (($need$0) - ($12))|0;
    $28 = ($12>>>0)<(4096);
    if ($28) {
     label = 12;
     break;
    } else {
     $need$0 = $27;
    }
   }
   if ((label|0) == 5) {
    $14 = ((($fr)) + 28068|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = $15 & 32;
    $17 = ($16|0)==(0);
    if ($17) {
     $18 = HEAP32[37544]|0;
     HEAP32[$vararg_buffer>>2] = 844;
     (_fprintf($18,515864,$vararg_buffer)|0);
    }
   }
   else if ((label|0) == 9) {
    $22 = ((($fr)) + 28068|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = $23 & 32;
    $25 = ($24|0)==(0);
    if ($25) {
     $26 = HEAP32[37544]|0;
     HEAP32[$vararg_buffer1>>2] = 851;
     $vararg_ptr4 = ((($vararg_buffer1)) + 4|0);
     HEAP32[$vararg_ptr4>>2] = $$lcssa47;
     (_fprintf($26,515902,$vararg_buffer1)|0);
    }
   }
   else if ((label|0) == 12) {
    $29 = ((($fr)) + 28068|0);
    $30 = HEAP32[$29>>2]|0;
    $31 = $30 & 32;
    $32 = ($31|0)==(0);
    if ($32) {
     $33 = ((($fr)) + 28064|0);
     $34 = HEAP32[$33>>2]|0;
     $35 = ($34|0)>(2);
     if ($35) {
      $36 = HEAP32[37544]|0;
      (_fwrite(515960,22,1,$36)|0);
      label = 15;
     } else {
      label = 15;
     }
    } else {
     label = 15;
    }
   }
   if ((label|0) == 15) {
    $37 = HEAP32[$1>>2]|0;
    $38 = HEAP32[$3>>2]|0;
    $39 = (($37) - ($38))|0;
    $40 = ($39|0)<($count|0);
    $$count = $40 ? $39 : $count;
    $$25 = $$count;$42 = $37;$43 = $38;
    break;
   }
   $$4 = -1;
   STACKTOP = sp;return ($$4|0);
  } else {
   $$25 = $count;$42 = $2;$43 = $4;
  }
 } while(0);
 $41 = (($42) - ($43))|0;
 $44 = ($41|0)<($$25|0);
 L22: do {
  if ($44) {
   $46 = ((($fr)) + 28036|0);
   $47 = HEAP32[$46>>2]|0;
   HEAP32[$3>>2] = $47;
   $$0$i = -10;
  } else {
   $b$04$i = HEAP32[$0>>2]|0;
   $45 = ($b$04$i|0)==(0|0);
   if ($45) {
    $$0$i = 0;
   } else {
    $b$06$i = $b$04$i;$offset$05$i = 0;
    while(1) {
     $48 = ((($b$06$i)) + 4|0);
     $49 = HEAP32[$48>>2]|0;
     $50 = (($49) + ($offset$05$i))|0;
     $51 = ($50|0)>($43|0);
     if ($51) {
      $$lcssa45 = $49;$b$06$i$lcssa = $b$06$i;$offset$05$i$lcssa = $offset$05$i;
      break;
     }
     $64 = ((($b$06$i)) + 12|0);
     $b$0$i = HEAP32[$64>>2]|0;
     $65 = ($b$0$i|0)==(0|0);
     if ($65) {
      $$0$i = 0;
      break L22;
     } else {
      $b$06$i = $b$0$i;$offset$05$i = $50;
     }
    }
    $52 = ($$25|0)>(0);
    if ($52) {
     $53 = (($43) - ($offset$05$i$lcssa))|0;
     $54 = (($$lcssa45) - ($53))|0;
     $55 = ($$25|0)>($54|0);
     $$$i34 = $55 ? $54 : $$25;
     $56 = HEAP32[$b$06$i$lcssa>>2]|0;
     $57 = (($56) + ($53)|0);
     _memcpy(($out|0),($57|0),($$$i34|0))|0;
     $58 = HEAP32[$3>>2]|0;
     $59 = (($58) + ($$$i34))|0;
     HEAP32[$3>>2] = $59;
     $60 = ((($b$06$i$lcssa)) + 12|0);
     $61 = HEAP32[$60>>2]|0;
     $62 = ($61|0)!=(0|0);
     $63 = $55 & $62;
     if ($63) {
      $69 = $61;$71 = $59;$73 = $$$i34;$b$11$i36 = $b$06$i$lcssa;$offset$13$i35 = $offset$05$i$lcssa;
      while(1) {
       $66 = ((($b$11$i36)) + 4|0);
       $67 = HEAP32[$66>>2]|0;
       $68 = (($67) + ($offset$13$i35))|0;
       $$phi$trans$insert = ((($69)) + 4|0);
       $$pre = HEAP32[$$phi$trans$insert>>2]|0;
       $70 = (($71) - ($68))|0;
       $72 = (($$25) - ($73))|0;
       $74 = (($$pre) - ($70))|0;
       $75 = ($72|0)>($74|0);
       $$$i = $75 ? $74 : $72;
       $76 = (($out) + ($73)|0);
       $77 = HEAP32[$69>>2]|0;
       $78 = (($77) + ($70)|0);
       _memcpy(($76|0),($78|0),($$$i|0))|0;
       $79 = (($$$i) + ($73))|0;
       $80 = HEAP32[$3>>2]|0;
       $81 = (($80) + ($$$i))|0;
       HEAP32[$3>>2] = $81;
       $82 = ((($69)) + 12|0);
       $83 = HEAP32[$82>>2]|0;
       $84 = ($79|0)<($$25|0);
       $85 = ($83|0)!=(0|0);
       $86 = $84 & $85;
       if ($86) {
        $b$11$i36$phi = $69;$69 = $83;$71 = $81;$73 = $79;$offset$13$i35 = $68;$b$11$i36 = $b$11$i36$phi;
       } else {
        $$0$i = $79;
        break;
       }
      }
     } else {
      $$0$i = $$$i34;
     }
    } else {
     $$0$i = 0;
    }
   }
  }
 } while(0);
 $87 = ($$0$i|0)==($$25|0);
 if ($87) {
  $$4 = $$25;
  STACKTOP = sp;return ($$4|0);
 }
 $88 = ((($fr)) + 28068|0);
 $89 = HEAP32[$88>>2]|0;
 $90 = $89 & 32;
 $91 = ($90|0)==(0);
 if (!($91)) {
  $$4 = -1;
  STACKTOP = sp;return ($$4|0);
 }
 $92 = HEAP32[37544]|0;
 HEAP32[$vararg_buffer5>>2] = 869;
 (_fprintf($92,515983,$vararg_buffer5)|0);
 $$4 = -1;
 STACKTOP = sp;return ($$4|0);
}
function _INT123_open_stream_handle($fr,$iohandle) {
 $fr = $fr|0;
 $iohandle = $iohandle|0;
 var $$lobit$i = 0, $$sink$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fr)) + 28576|0);
 _INT123_clear_icy($0);
 $1 = ((($fr)) + 27960|0);
 HEAP32[$1>>2] = -1;
 $2 = ((($fr)) + 27968|0);
 HEAP32[$2>>2] = -1;
 $3 = ((($fr)) + 27972|0);
 HEAP32[$3>>2] = $iohandle;
 $4 = ((($fr)) + 27976|0);
 HEAP32[$4>>2] = 64;
 $5 = ((($fr)) + 28336|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)>(0);
 if ($7) {
  $8 = ((($fr)) + 28580|0);
  HEAP32[$8>>2] = $6;
  $9 = ((($fr)) + 28584|0);
  HEAP32[$9>>2] = $6;
  $$sink$i = (38328);
 } else {
  $$sink$i = 38280;
 }
 $10 = ((($fr)) + 27956|0);
 HEAP32[$10>>2] = $$sink$i;
 $11 = HEAP32[$$sink$i>>2]|0;
 $12 = (FUNCTION_TABLE_ii[$11 & 8191]($fr)|0);
 $$lobit$i = $12 >> 31;
 return ($$lobit$i|0);
}
function _INT123_prepare_decode_tables() {
 var $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0.0, $4 = 0.0, $5 = 0.0;
 var $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $k$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[8828]|0;
 $k$01 = 0;
 while(1) {
  $1 = (+($k$01|0));
  $2 = $1 * 0.098174770424681035;
  $3 = $2 + 0.049087385212340517;
  $4 = (+Math_cos((+$3)));
  $5 = 0.5 / $4;
  $6 = $5;
  $7 = (($0) + ($k$01<<2)|0);
  HEAPF32[$7>>2] = $6;
  $8 = (($k$01) + 1)|0;
  $exitcond = ($8|0)==(16);
  if ($exitcond) {
   break;
  } else {
   $k$01 = $8;
  }
 }
 $9 = HEAP32[(35316)>>2]|0;
 HEAPF32[$9>>2] = 0.50241929292678833;
 $10 = ((($9)) + 4|0);
 HEAPF32[$10>>2] = 0.52249860763549805;
 $11 = ((($9)) + 8|0);
 HEAPF32[$11>>2] = 0.56694406270980835;
 $12 = ((($9)) + 12|0);
 HEAPF32[$12>>2] = 0.64682179689407349;
 $13 = ((($9)) + 16|0);
 HEAPF32[$13>>2] = 0.78815460205078125;
 $14 = ((($9)) + 20|0);
 HEAPF32[$14>>2] = 1.0606776475906372;
 $15 = ((($9)) + 24|0);
 HEAPF32[$15>>2] = 1.7224471569061279;
 $16 = ((($9)) + 28|0);
 HEAPF32[$16>>2] = 5.1011486053466797;
 $17 = HEAP32[(35320)>>2]|0;
 HEAPF32[$17>>2] = 0.50979560613632202;
 $18 = ((($17)) + 4|0);
 HEAPF32[$18>>2] = 0.60134488344192505;
 $19 = ((($17)) + 8|0);
 HEAPF32[$19>>2] = 0.89997619390487671;
 $20 = ((($17)) + 12|0);
 HEAPF32[$20>>2] = 2.562915563583374;
 $21 = HEAP32[(35324)>>2]|0;
 HEAPF32[$21>>2] = 0.54119610786437988;
 $22 = ((($21)) + 4|0);
 HEAPF32[$22>>2] = 1.3065630197525024;
 $23 = HEAP32[(35328)>>2]|0;
 HEAPF32[$23>>2] = 0.70710676908493042;
 return;
}
function _mpg123_init() {
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[179892]|0;
 $1 = ($0|0)==(0);
 if (!($1)) {
  return 0;
 }
 _INT123_init_layer12();
 _INT123_init_layer3();
 _INT123_prepare_decode_tables();
 HEAP32[179892] = 1;
 return 0;
}
function _INT123_init_layer12() {
 var $$lcssa = 0, $$pre64 = 0, $$pre65 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $12 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $itable$011$us$1 = 0, $itable$011$us$2 = 0, $itable$16$us$us$2 = 0, $j$09$us$1 = 0, $j$09$us$2 = 0;
 var $k$04$us$us$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[166804] = 1;
 HEAP32[(667220)>>2] = 1;
 HEAP32[(667224)>>2] = 1;
 HEAP32[(667228)>>2] = 0;
 HEAP32[(667232)>>2] = 1;
 HEAP32[(667236)>>2] = 1;
 HEAP32[(667240)>>2] = 2;
 HEAP32[(667244)>>2] = 1;
 HEAP32[(667248)>>2] = 1;
 HEAP32[(667252)>>2] = 1;
 HEAP32[(667256)>>2] = 0;
 HEAP32[(667260)>>2] = 1;
 HEAP32[(667264)>>2] = 0;
 HEAP32[(667268)>>2] = 0;
 HEAP32[(667272)>>2] = 1;
 HEAP32[(667276)>>2] = 2;
 HEAP32[(667280)>>2] = 0;
 HEAP32[(667284)>>2] = 1;
 HEAP32[(667288)>>2] = 1;
 HEAP32[(667292)>>2] = 2;
 HEAP32[(667296)>>2] = 1;
 HEAP32[(667300)>>2] = 0;
 HEAP32[(667304)>>2] = 2;
 HEAP32[(667308)>>2] = 1;
 HEAP32[(667312)>>2] = 2;
 HEAP32[(667316)>>2] = 2;
 HEAP32[(667320)>>2] = 1;
 HEAP32[(667324)>>2] = 1;
 HEAP32[(667328)>>2] = 1;
 HEAP32[(667332)>>2] = 0;
 HEAP32[(667336)>>2] = 0;
 HEAP32[(667340)>>2] = 1;
 HEAP32[(667344)>>2] = 0;
 HEAP32[(667348)>>2] = 2;
 HEAP32[(667352)>>2] = 1;
 HEAP32[(667356)>>2] = 0;
 HEAP32[(667360)>>2] = 1;
 HEAP32[(667364)>>2] = 0;
 HEAP32[(667368)>>2] = 0;
 HEAP32[(667372)>>2] = 0;
 HEAP32[(667376)>>2] = 0;
 HEAP32[(667380)>>2] = 0;
 HEAP32[(667384)>>2] = 2;
 HEAP32[(667388)>>2] = 0;
 HEAP32[(667392)>>2] = 0;
 HEAP32[(667396)>>2] = 1;
 HEAP32[(667400)>>2] = 2;
 HEAP32[(667404)>>2] = 0;
 HEAP32[(667408)>>2] = 0;
 HEAP32[(667412)>>2] = 2;
 HEAP32[(667416)>>2] = 0;
 HEAP32[(667420)>>2] = 2;
 HEAP32[(667424)>>2] = 2;
 HEAP32[(667428)>>2] = 0;
 HEAP32[(667432)>>2] = 1;
 HEAP32[(667436)>>2] = 1;
 HEAP32[(667440)>>2] = 2;
 HEAP32[(667444)>>2] = 0;
 HEAP32[(667448)>>2] = 1;
 HEAP32[(667452)>>2] = 2;
 HEAP32[(667456)>>2] = 2;
 HEAP32[(667460)>>2] = 1;
 HEAP32[(667464)>>2] = 2;
 HEAP32[(667468)>>2] = 1;
 HEAP32[(667472)>>2] = 0;
 HEAP32[(667476)>>2] = 2;
 HEAP32[(667480)>>2] = 0;
 HEAP32[(667484)>>2] = 0;
 HEAP32[(667488)>>2] = 2;
 HEAP32[(667492)>>2] = 2;
 HEAP32[(667496)>>2] = 0;
 HEAP32[(667500)>>2] = 2;
 HEAP32[(667504)>>2] = 1;
 HEAP32[(667508)>>2] = 2;
 HEAP32[(667512)>>2] = 2;
 HEAP32[(667516)>>2] = 0;
 HEAP32[(667520)>>2] = 2;
 HEAP32[(667524)>>2] = 2;
 HEAP32[(667528)>>2] = 2;
 HEAP32[(667532)>>2] = 2;
 HEAP32[(667536)>>2] = 2;
 $itable$011$us$1 = 667600;$j$09$us$1 = 0;
 while(1) {
  $0 = (38556 + ($j$09$us$1<<2)|0);
  $$pre64 = HEAP32[$0>>2]|0;
  $1 = ((($itable$011$us$1)) + 4|0);
  HEAP32[$itable$011$us$1>>2] = 17;
  $2 = ((($itable$011$us$1)) + 8|0);
  HEAP32[$1>>2] = 17;
  $3 = ((($itable$011$us$1)) + 12|0);
  HEAP32[$2>>2] = $$pre64;
  $4 = ((($itable$011$us$1)) + 16|0);
  HEAP32[$3>>2] = 18;
  $5 = ((($itable$011$us$1)) + 20|0);
  HEAP32[$4>>2] = 17;
  $6 = ((($itable$011$us$1)) + 24|0);
  HEAP32[$5>>2] = $$pre64;
  $7 = ((($itable$011$us$1)) + 28|0);
  HEAP32[$6>>2] = 0;
  $8 = ((($itable$011$us$1)) + 32|0);
  HEAP32[$7>>2] = 17;
  $9 = ((($itable$011$us$1)) + 36|0);
  HEAP32[$8>>2] = $$pre64;
  $10 = ((($itable$011$us$1)) + 40|0);
  HEAP32[$9>>2] = 19;
  $11 = ((($itable$011$us$1)) + 44|0);
  HEAP32[$10>>2] = 17;
  $12 = ((($itable$011$us$1)) + 48|0);
  HEAP32[$11>>2] = $$pre64;
  $13 = ((($itable$011$us$1)) + 52|0);
  HEAP32[$12>>2] = 20;
  $14 = ((($itable$011$us$1)) + 56|0);
  HEAP32[$13>>2] = 17;
  HEAP32[$14>>2] = $$pre64;
  $15 = ((($itable$011$us$1)) + 60|0);
  $16 = ((($itable$011$us$1)) + 64|0);
  HEAP32[$15>>2] = 17;
  $17 = ((($itable$011$us$1)) + 68|0);
  HEAP32[$16>>2] = 18;
  $18 = ((($itable$011$us$1)) + 72|0);
  HEAP32[$17>>2] = $$pre64;
  $19 = ((($itable$011$us$1)) + 76|0);
  HEAP32[$18>>2] = 18;
  $20 = ((($itable$011$us$1)) + 80|0);
  HEAP32[$19>>2] = 18;
  $21 = ((($itable$011$us$1)) + 84|0);
  HEAP32[$20>>2] = $$pre64;
  $22 = ((($itable$011$us$1)) + 88|0);
  HEAP32[$21>>2] = 0;
  $23 = ((($itable$011$us$1)) + 92|0);
  HEAP32[$22>>2] = 18;
  $24 = ((($itable$011$us$1)) + 96|0);
  HEAP32[$23>>2] = $$pre64;
  $25 = ((($itable$011$us$1)) + 100|0);
  HEAP32[$24>>2] = 19;
  $26 = ((($itable$011$us$1)) + 104|0);
  HEAP32[$25>>2] = 18;
  $27 = ((($itable$011$us$1)) + 108|0);
  HEAP32[$26>>2] = $$pre64;
  $28 = ((($itable$011$us$1)) + 112|0);
  HEAP32[$27>>2] = 20;
  $29 = ((($itable$011$us$1)) + 116|0);
  HEAP32[$28>>2] = 18;
  HEAP32[$29>>2] = $$pre64;
  $30 = ((($itable$011$us$1)) + 120|0);
  $31 = ((($itable$011$us$1)) + 124|0);
  HEAP32[$30>>2] = 17;
  $32 = ((($itable$011$us$1)) + 128|0);
  HEAP32[$31>>2] = 0;
  $33 = ((($itable$011$us$1)) + 132|0);
  HEAP32[$32>>2] = $$pre64;
  $34 = ((($itable$011$us$1)) + 136|0);
  HEAP32[$33>>2] = 18;
  $35 = ((($itable$011$us$1)) + 140|0);
  HEAP32[$34>>2] = 0;
  $36 = ((($itable$011$us$1)) + 144|0);
  HEAP32[$35>>2] = $$pre64;
  $37 = ((($itable$011$us$1)) + 148|0);
  HEAP32[$36>>2] = 0;
  $38 = ((($itable$011$us$1)) + 152|0);
  HEAP32[$37>>2] = 0;
  $39 = ((($itable$011$us$1)) + 156|0);
  HEAP32[$38>>2] = $$pre64;
  $40 = ((($itable$011$us$1)) + 160|0);
  HEAP32[$39>>2] = 19;
  $41 = ((($itable$011$us$1)) + 164|0);
  HEAP32[$40>>2] = 0;
  $42 = ((($itable$011$us$1)) + 168|0);
  HEAP32[$41>>2] = $$pre64;
  $43 = ((($itable$011$us$1)) + 172|0);
  HEAP32[$42>>2] = 20;
  $44 = ((($itable$011$us$1)) + 176|0);
  HEAP32[$43>>2] = 0;
  HEAP32[$44>>2] = $$pre64;
  $45 = ((($itable$011$us$1)) + 180|0);
  $46 = ((($itable$011$us$1)) + 184|0);
  HEAP32[$45>>2] = 17;
  $47 = ((($itable$011$us$1)) + 188|0);
  HEAP32[$46>>2] = 19;
  $48 = ((($itable$011$us$1)) + 192|0);
  HEAP32[$47>>2] = $$pre64;
  $49 = ((($itable$011$us$1)) + 196|0);
  HEAP32[$48>>2] = 18;
  $50 = ((($itable$011$us$1)) + 200|0);
  HEAP32[$49>>2] = 19;
  $51 = ((($itable$011$us$1)) + 204|0);
  HEAP32[$50>>2] = $$pre64;
  $52 = ((($itable$011$us$1)) + 208|0);
  HEAP32[$51>>2] = 0;
  $53 = ((($itable$011$us$1)) + 212|0);
  HEAP32[$52>>2] = 19;
  $54 = ((($itable$011$us$1)) + 216|0);
  HEAP32[$53>>2] = $$pre64;
  $55 = ((($itable$011$us$1)) + 220|0);
  HEAP32[$54>>2] = 19;
  $56 = ((($itable$011$us$1)) + 224|0);
  HEAP32[$55>>2] = 19;
  $57 = ((($itable$011$us$1)) + 228|0);
  HEAP32[$56>>2] = $$pre64;
  $58 = ((($itable$011$us$1)) + 232|0);
  HEAP32[$57>>2] = 20;
  $59 = ((($itable$011$us$1)) + 236|0);
  HEAP32[$58>>2] = 19;
  HEAP32[$59>>2] = $$pre64;
  $60 = ((($itable$011$us$1)) + 240|0);
  $61 = ((($itable$011$us$1)) + 244|0);
  HEAP32[$60>>2] = 17;
  $62 = ((($itable$011$us$1)) + 248|0);
  HEAP32[$61>>2] = 20;
  $63 = ((($itable$011$us$1)) + 252|0);
  HEAP32[$62>>2] = $$pre64;
  $64 = ((($itable$011$us$1)) + 256|0);
  HEAP32[$63>>2] = 18;
  $65 = ((($itable$011$us$1)) + 260|0);
  HEAP32[$64>>2] = 20;
  $66 = ((($itable$011$us$1)) + 264|0);
  HEAP32[$65>>2] = $$pre64;
  $67 = ((($itable$011$us$1)) + 268|0);
  HEAP32[$66>>2] = 0;
  $68 = ((($itable$011$us$1)) + 272|0);
  HEAP32[$67>>2] = 20;
  $69 = ((($itable$011$us$1)) + 276|0);
  HEAP32[$68>>2] = $$pre64;
  $70 = ((($itable$011$us$1)) + 280|0);
  HEAP32[$69>>2] = 19;
  $71 = ((($itable$011$us$1)) + 284|0);
  HEAP32[$70>>2] = 20;
  $72 = ((($itable$011$us$1)) + 288|0);
  HEAP32[$71>>2] = $$pre64;
  $73 = ((($itable$011$us$1)) + 292|0);
  HEAP32[$72>>2] = 20;
  $74 = ((($itable$011$us$1)) + 296|0);
  HEAP32[$73>>2] = 20;
  $75 = ((($itable$011$us$1)) + 300|0);
  HEAP32[$74>>2] = $$pre64;
  $76 = (($j$09$us$1) + 1)|0;
  $77 = ($76|0)<(5);
  if ($77) {
   $itable$011$us$1 = $75;$j$09$us$1 = $76;
  } else {
   $itable$011$us$2 = 669136;$j$09$us$2 = 0;
   break;
  }
 }
 while(1) {
  $78 = (38592 + ($j$09$us$2<<2)|0);
  $$pre65 = HEAP32[$78>>2]|0;
  $itable$16$us$us$2 = $itable$011$us$2;$k$04$us$us$2 = 0;
  while(1) {
   $79 = (38592 + ($k$04$us$us$2<<2)|0);
   $80 = HEAP32[$79>>2]|0;
   $81 = ((($itable$16$us$us$2)) + 4|0);
   HEAP32[$itable$16$us$us$2>>2] = 21;
   $82 = ((($itable$16$us$us$2)) + 8|0);
   HEAP32[$81>>2] = $80;
   $83 = ((($itable$16$us$us$2)) + 12|0);
   HEAP32[$82>>2] = $$pre65;
   $84 = ((($itable$16$us$us$2)) + 16|0);
   HEAP32[$83>>2] = 1;
   $85 = ((($itable$16$us$us$2)) + 20|0);
   HEAP32[$84>>2] = $80;
   $86 = ((($itable$16$us$us$2)) + 24|0);
   HEAP32[$85>>2] = $$pre65;
   $87 = ((($itable$16$us$us$2)) + 28|0);
   HEAP32[$86>>2] = 22;
   $88 = ((($itable$16$us$us$2)) + 32|0);
   HEAP32[$87>>2] = $80;
   $89 = ((($itable$16$us$us$2)) + 36|0);
   HEAP32[$88>>2] = $$pre65;
   $90 = ((($itable$16$us$us$2)) + 40|0);
   HEAP32[$89>>2] = 23;
   $91 = ((($itable$16$us$us$2)) + 44|0);
   HEAP32[$90>>2] = $80;
   $92 = ((($itable$16$us$us$2)) + 48|0);
   HEAP32[$91>>2] = $$pre65;
   $93 = ((($itable$16$us$us$2)) + 52|0);
   HEAP32[$92>>2] = 0;
   $94 = ((($itable$16$us$us$2)) + 56|0);
   HEAP32[$93>>2] = $80;
   $95 = ((($itable$16$us$us$2)) + 60|0);
   HEAP32[$94>>2] = $$pre65;
   $96 = ((($itable$16$us$us$2)) + 64|0);
   HEAP32[$95>>2] = 24;
   $97 = ((($itable$16$us$us$2)) + 68|0);
   HEAP32[$96>>2] = $80;
   $98 = ((($itable$16$us$us$2)) + 72|0);
   HEAP32[$97>>2] = $$pre65;
   $99 = ((($itable$16$us$us$2)) + 76|0);
   HEAP32[$98>>2] = 25;
   $100 = ((($itable$16$us$us$2)) + 80|0);
   HEAP32[$99>>2] = $80;
   $101 = ((($itable$16$us$us$2)) + 84|0);
   HEAP32[$100>>2] = $$pre65;
   $102 = ((($itable$16$us$us$2)) + 88|0);
   HEAP32[$101>>2] = 2;
   $103 = ((($itable$16$us$us$2)) + 92|0);
   HEAP32[$102>>2] = $80;
   $104 = ((($itable$16$us$us$2)) + 96|0);
   HEAP32[$103>>2] = $$pre65;
   $105 = ((($itable$16$us$us$2)) + 100|0);
   HEAP32[$104>>2] = 26;
   $106 = ((($itable$16$us$us$2)) + 104|0);
   HEAP32[$105>>2] = $80;
   $107 = ((($itable$16$us$us$2)) + 108|0);
   HEAP32[$106>>2] = $$pre65;
   $108 = (($k$04$us$us$2) + 1)|0;
   $109 = ($108|0)<(9);
   if ($109) {
    $itable$16$us$us$2 = $107;$k$04$us$us$2 = $108;
   } else {
    $$lcssa = $107;
    break;
   }
  }
  $110 = (($j$09$us$2) + 1)|0;
  $111 = ($110|0)<(9);
  if ($111) {
   $itable$011$us$2 = $$lcssa;$j$09$us$2 = $110;
  } else {
   break;
  }
 }
 return;
}
function _INT123_init_layer3() {
 var $0 = 0.0, $1 = 0.0, $10 = 0.0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0.0, $110 = 0, $111 = 0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0;
 var $116 = 0.0, $117 = 0.0, $118 = 0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0, $126 = 0.0, $127 = 0.0, $128 = 0, $129 = 0, $13 = 0.0, $130 = 0.0, $131 = 0.0, $132 = 0.0, $133 = 0;
 var $134 = 0, $135 = 0.0, $136 = 0.0, $137 = 0, $138 = 0, $139 = 0.0, $14 = 0.0, $140 = 0.0, $141 = 0, $142 = 0, $143 = 0.0, $144 = 0.0, $145 = 0, $146 = 0, $147 = 0.0, $148 = 0.0, $149 = 0, $15 = 0, $150 = 0, $151 = 0.0;
 var $152 = 0, $153 = 0.0, $154 = 0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0, $159 = 0.0, $16 = 0, $160 = 0, $161 = 0.0, $162 = 0, $163 = 0, $164 = 0.0, $165 = 0.0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0.0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0.0, $195 = 0.0, $196 = 0.0, $197 = 0.0, $198 = 0.0, $199 = 0.0, $2 = 0.0, $20 = 0.0, $200 = 0.0, $201 = 0.0, $202 = 0.0, $203 = 0.0, $204 = 0.0, $205 = 0.0;
 var $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0.0, $21 = 0.0, $210 = 0.0, $211 = 0.0, $212 = 0.0, $213 = 0.0, $214 = 0.0, $215 = 0.0, $216 = 0.0, $217 = 0.0, $218 = 0.0, $219 = 0.0, $22 = 0.0, $220 = 0.0, $221 = 0.0, $222 = 0.0, $223 = 0.0;
 var $224 = 0.0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0.0, $229 = 0.0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0.0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0.0, $249 = 0.0, $25 = 0.0, $250 = 0.0, $251 = 0.0, $252 = 0.0, $253 = 0.0, $254 = 0.0, $255 = 0.0, $256 = 0.0, $257 = 0.0, $258 = 0.0, $259 = 0.0, $26 = 0.0;
 var $260 = 0.0, $261 = 0.0, $262 = 0.0, $263 = 0.0, $264 = 0.0, $265 = 0.0, $266 = 0.0, $267 = 0.0, $268 = 0.0, $269 = 0.0, $27 = 0.0, $270 = 0.0, $271 = 0.0, $272 = 0.0, $273 = 0.0, $274 = 0.0, $275 = 0.0, $276 = 0.0, $277 = 0.0, $278 = 0.0;
 var $279 = 0.0, $28 = 0.0, $280 = 0.0, $281 = 0.0, $282 = 0.0, $283 = 0.0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0.0, $291 = 0.0, $292 = 0.0, $293 = 0.0, $294 = 0.0, $295 = 0.0, $296 = 0.0;
 var $297 = 0.0, $298 = 0.0, $299 = 0.0, $3 = 0, $30 = 0, $300 = 0.0, $301 = 0.0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0.0, $321 = 0.0, $322 = 0.0, $323 = 0.0, $324 = 0.0, $325 = 0.0, $326 = 0.0, $327 = 0.0, $328 = 0.0, $329 = 0.0, $33 = 0, $330 = 0.0, $331 = 0.0;
 var $332 = 0.0, $333 = 0.0, $334 = 0.0, $335 = 0.0, $336 = 0.0, $337 = 0.0, $338 = 0.0, $339 = 0.0, $34 = 0.0, $340 = 0.0, $341 = 0.0, $342 = 0.0, $343 = 0.0, $344 = 0.0, $345 = 0.0, $346 = 0.0, $347 = 0.0, $348 = 0.0, $349 = 0.0, $35 = 0.0;
 var $350 = 0.0, $351 = 0.0, $352 = 0.0, $353 = 0.0, $354 = 0.0, $355 = 0.0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0.0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0;
 var $369 = 0, $37 = 0.0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0;
 var $387 = 0, $388 = 0, $389 = 0, $39 = 0.0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0.0, $400 = 0, $401 = 0, $402 = 0, $403 = 0;
 var $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0.0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0.0, $420 = 0, $421 = 0;
 var $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0.0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0;
 var $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0;
 var $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0;
 var $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0;
 var $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0.0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0.0, $510 = 0, $511 = 0;
 var $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0.0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0.0;
 var $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0.0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0;
 var $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0;
 var $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0;
 var $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0.0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0.0, $60 = 0.0, $600 = 0, $601 = 0;
 var $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0.0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0.0;
 var $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0.0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0;
 var $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0;
 var $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0.0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0;
 var $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0.0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0.0, $690 = 0, $691 = 0, $692 = 0;
 var $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0.0, $70 = 0.0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0;
 var $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0;
 var $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0.0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0;
 var $747 = 0, $748 = 0, $749 = 0, $75 = 0.0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0.0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0;
 var $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0.0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $78 = 0.0, $79 = 0, $8 = 0.0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0.0, $85 = 0.0;
 var $86 = 0.0, $87 = 0.0, $88 = 0, $89 = 0, $9 = 0.0, $90 = 0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bdf$148 = 0, $bdf$254 = 0, $cb$147 = 0, $cb$253 = 0, $exitcond100 = 0;
 var $exitcond101 = 0, $exitcond109 = 0, $exitcond115 = 0, $exitcond120 = 0, $exitcond126 = 0, $exitcond127 = 0, $exitcond131 = 0, $exitcond132 = 0, $exitcond133 = 0, $exitcond134 = 0, $exitcond89 = 0, $exitcond92 = 0, $exitcond93 = 0, $exitcond98 = 0, $i$076 = 0, $i$1250 = 0, $i$1356 = 0, $i$1440 = 0, $i$1537 = 0, $i$1732 = 0;
 var $i$175 = 0, $i$1828 = 0, $i$274 = 0, $i$373 = 0, $i$769 = 0, $i$868 = 0, $indvars$iv = 0, $indvars$iv106 = 0, $indvars$iv112 = 0, $indvars$iv116 = 0, $j$260 = 0, $j$339 = 0, $j$631 = 0, $lftr$limit = 0, $mp$149 = 0, $mp$355 = 0, $scevgep = 0, $scevgep104 = 0, $scevgep107 = 0, $scevgep110 = 0;
 var $scevgep113 = 0, $scevgep117 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $i$076 = 0;
 while(1) {
  $0 = (+($i$076|0));
  $1 = (+Math_pow((+$0),1.3333333333333333));
  $2 = $1;
  $3 = (684712 + ($i$076<<2)|0);
  HEAPF32[$3>>2] = $2;
  $4 = (($i$076) + 1)|0;
  $exitcond134 = ($4|0)==(8207);
  if ($exitcond134) {
   $i$175 = 0;
   break;
  } else {
   $i$076 = $4;
  }
 }
 while(1) {
  $5 = (3240 + ($i$175<<3)|0);
  $6 = +HEAPF64[$5>>3];
  $7 = $6 * $6;
  $8 = $7 + 1.0;
  $9 = (+Math_sqrt((+$8)));
  $10 = 1.0 / $9;
  $11 = $10;
  $12 = (718308 + ($i$175<<2)|0);
  HEAPF32[$12>>2] = $11;
  $13 = $6 / $9;
  $14 = $13;
  $15 = (718340 + ($i$175<<2)|0);
  HEAPF32[$15>>2] = $14;
  $16 = (($i$175) + 1)|0;
  $exitcond133 = ($16|0)==(8);
  if ($exitcond133) {
   $i$274 = 0;
   break;
  } else {
   $i$175 = $16;
  }
 }
 while(1) {
  $17 = $i$274 << 1;
  $18 = $17 | 1;
  $19 = (+($18|0));
  $20 = $19 * 0.043633231299858237;
  $21 = (+Math_sin((+$20)));
  $22 = $21 * 0.5;
  $23 = (($17) + 19)|0;
  $24 = (+($23|0));
  $25 = $24 * 0.043633231299858237;
  $26 = (+Math_cos((+$25)));
  $27 = $22 / $26;
  $28 = $27;
  $29 = (718516 + ($i$274<<2)|0);
  HEAPF32[$29>>2] = $28;
  $30 = (718372 + ($i$274<<2)|0);
  HEAPF32[$30>>2] = $28;
  $31 = (($i$274) + 18)|0;
  $32 = $31 << 1;
  $33 = $32 | 1;
  $34 = (+($33|0));
  $35 = $34 * 0.043633231299858237;
  $36 = (+Math_sin((+$35)));
  $37 = $36 * 0.5;
  $38 = (($32) + 19)|0;
  $39 = (+($38|0));
  $40 = $39 * 0.043633231299858237;
  $41 = (+Math_cos((+$40)));
  $42 = $37 / $41;
  $43 = $42;
  $44 = (718804 + ($31<<2)|0);
  HEAPF32[$44>>2] = $43;
  $45 = (718372 + ($31<<2)|0);
  HEAPF32[$45>>2] = $43;
  $46 = (($i$274) + 1)|0;
  $exitcond132 = ($46|0)==(18);
  if ($exitcond132) {
   $i$373 = 0;
   break;
  } else {
   $i$274 = $46;
  }
 }
 while(1) {
  $47 = (($i$373) + 18)|0;
  $48 = $47 << 1;
  $49 = (($48) + 19)|0;
  $50 = (+($49|0));
  $51 = $50 * 0.043633231299858237;
  $52 = (+Math_cos((+$51)));
  $53 = 0.5 / $52;
  $54 = $53;
  $55 = (718516 + ($47<<2)|0);
  HEAPF32[$55>>2] = $54;
  $56 = (($i$373) + 12)|0;
  $57 = $56 << 1;
  $58 = (($57) + 19)|0;
  $59 = (+($58|0));
  $60 = $59 * 0.043633231299858237;
  $61 = (+Math_cos((+$60)));
  $62 = 0.5 / $61;
  $63 = $62;
  $64 = (718804 + ($56<<2)|0);
  HEAPF32[$64>>2] = $63;
  $65 = $i$373 << 1;
  $66 = (($65) + 13)|0;
  $67 = (+($66|0));
  $68 = $67 * 0.1308996938995747;
  $69 = (+Math_sin((+$68)));
  $70 = $69 * 0.5;
  $71 = (($i$373) + 24)|0;
  $72 = $71 << 1;
  $73 = (($72) + 19)|0;
  $74 = (+($73|0));
  $75 = $74 * 0.043633231299858237;
  $76 = (+Math_cos((+$75)));
  $77 = $70 / $76;
  $78 = $77;
  $79 = (718516 + ($71<<2)|0);
  HEAPF32[$79>>2] = $78;
  $80 = (718804 + ($i$373<<2)|0);
  HEAPF32[$80>>2] = 0.0;
  $81 = (($i$373) + 30)|0;
  $82 = (718516 + ($81<<2)|0);
  HEAPF32[$82>>2] = 0.0;
  $83 = $65 | 1;
  $84 = (+($83|0));
  $85 = $84 * 0.1308996938995747;
  $86 = (+Math_sin((+$85)));
  $87 = $86 * 0.5;
  $88 = (($i$373) + 6)|0;
  $89 = $88 << 1;
  $90 = (($89) + 19)|0;
  $91 = (+($90|0));
  $92 = $91 * 0.043633231299858237;
  $93 = (+Math_cos((+$92)));
  $94 = $87 / $93;
  $95 = $94;
  $96 = (718804 + ($88<<2)|0);
  HEAPF32[$96>>2] = $95;
  $97 = (($i$373) + 1)|0;
  $exitcond131 = ($97|0)==(6);
  if ($exitcond131) {
   break;
  } else {
   $i$373 = $97;
  }
 }
 HEAPF32[166786] = 1.0;
 HEAPF32[(667148)>>2] = 0.98480772972106933;
 HEAPF32[(667152)>>2] = 0.93969261646270751;
 HEAPF32[(667156)>>2] = 0.86602538824081421;
 HEAPF32[(667160)>>2] = 0.76604443788528442;
 HEAPF32[(667164)>>2] = 0.6427876353263855;
 HEAPF32[(667168)>>2] = 0.5;
 HEAPF32[(667172)>>2] = 0.34202015399932861;
 HEAPF32[(667176)>>2] = 0.1736481785774231;
 HEAPF32[166795] = 0.50190991163253784;
 HEAPF32[(667184)>>2] = 0.51763808727264404;
 HEAPF32[(667188)>>2] = 0.55168896913528442;
 HEAPF32[(667192)>>2] = 0.61038726568222046;
 HEAPF32[(667196)>>2] = 0.70710676908493042;
 HEAPF32[(667200)>>2] = 0.87172341346740723;
 HEAPF32[(667204)>>2] = 1.1831008195877075;
 HEAPF32[(667208)>>2] = 1.9318516254425049;
 HEAPF32[(667212)>>2] = 5.7368564605712891;
 HEAPF32[179889] = 0.51763808727264404;
 HEAPF32[(719560)>>2] = 0.70710676908493042;
 HEAPF32[(719564)>>2] = 1.9318516254425049;
 HEAPF32[179738] = 0.86602538824081421;
 HEAPF32[179737] = 0.5;
 HEAPF32[179739] = 0.93969261646270751;
 HEAPF32[179740] = -0.1736481785774231;
 HEAPF32[179741] = -0.76604443788528442;
 HEAPF32[179742] = 0.98480772972106933;
 HEAPF32[179743] = -0.34202015399932861;
 HEAPF32[179744] = -0.6427876353263855;
 $i$769 = 0;
 while(1) {
  $98 = $i$769 << 1;
  $99 = $98 | 1;
  $100 = (+($99|0));
  $101 = $100 * 0.1308996938995747;
  $102 = (+Math_sin((+$101)));
  $103 = $102 * 0.5;
  $104 = (($98) + 7)|0;
  $105 = (+($104|0));
  $106 = $105 * 0.1308996938995747;
  $107 = (+Math_cos((+$106)));
  $108 = $103 / $107;
  $109 = $108;
  $110 = (718660 + ($i$769<<2)|0);
  HEAPF32[$110>>2] = $109;
  $111 = (($i$769) + 1)|0;
  $exitcond127 = ($111|0)==(12);
  if ($exitcond127) {
   $i$868 = 0;
   break;
  } else {
   $i$769 = $111;
  }
 }
 while(1) {
  $112 = (+($i$868|0));
  $113 = $112 * 0.26179938779914941;
  $114 = (+Math_tan((+$113)));
  $115 = $114 + 1.0;
  $116 = $114 / $115;
  $117 = $116;
  $118 = (717540 + ($i$868<<2)|0);
  HEAPF32[$118>>2] = $117;
  $119 = 1.0 / $115;
  $120 = $119;
  $121 = (717604 + ($i$868<<2)|0);
  HEAPF32[$121>>2] = $120;
  $122 = $114 * 1.4142135623730951;
  $123 = $122 / $115;
  $124 = $123;
  $125 = (717668 + ($i$868<<2)|0);
  HEAPF32[$125>>2] = $124;
  $126 = 1.4142135623730951 / $115;
  $127 = $126;
  $128 = (717732 + ($i$868<<2)|0);
  HEAPF32[$128>>2] = $127;
  $129 = ($i$868|0)>(0);
  $130 = $112 * 0.5;
  $131 = $112 + 1.0;
  $132 = $131 * 0.5;
  do {
   if ($129) {
    $133 = $i$868 & 1;
    $134 = ($133|0)==(0);
    if ($134) {
     $151 = (+Math_pow(0.8408964152537145,(+$130)));
     $152 = (717796 + ($i$868<<2)|0);
     HEAPF32[$152>>2] = 1.0;
     $153 = $151;
     $154 = (717924 + ($i$868<<2)|0);
     HEAPF32[$154>>2] = $153;
     $155 = (718052 + ($i$868<<2)|0);
     HEAPF32[$155>>2] = 1.4142135381698608;
     $156 = $151 * 1.4142135623730951;
     $157 = $156;
     $158 = (718180 + ($i$868<<2)|0);
     HEAPF32[$158>>2] = $157;
     $159 = (+Math_pow(0.70710678118654757,(+$130)));
     $160 = (717860 + ($i$868<<2)|0);
     HEAPF32[$160>>2] = 1.0;
     $161 = $159;
     $162 = (717988 + ($i$868<<2)|0);
     HEAPF32[$162>>2] = $161;
     $163 = (718116 + ($i$868<<2)|0);
     HEAPF32[$163>>2] = 1.4142135381698608;
     $164 = $159 * 1.4142135623730951;
     $165 = $164;
     $166 = (718244 + ($i$868<<2)|0);
     HEAPF32[$166>>2] = $165;
     break;
    } else {
     $135 = (+Math_pow(0.8408964152537145,(+$132)));
     $136 = $135;
     $137 = (717796 + ($i$868<<2)|0);
     HEAPF32[$137>>2] = $136;
     $138 = (717924 + ($i$868<<2)|0);
     HEAPF32[$138>>2] = 1.0;
     $139 = $135 * 1.4142135623730951;
     $140 = $139;
     $141 = (718052 + ($i$868<<2)|0);
     HEAPF32[$141>>2] = $140;
     $142 = (718180 + ($i$868<<2)|0);
     HEAPF32[$142>>2] = 1.4142135381698608;
     $143 = (+Math_pow(0.70710678118654757,(+$132)));
     $144 = $143;
     $145 = (717860 + ($i$868<<2)|0);
     HEAPF32[$145>>2] = $144;
     $146 = (717988 + ($i$868<<2)|0);
     HEAPF32[$146>>2] = 1.0;
     $147 = $143 * 1.4142135623730951;
     $148 = $147;
     $149 = (718116 + ($i$868<<2)|0);
     HEAPF32[$149>>2] = $148;
     $150 = (718244 + ($i$868<<2)|0);
     HEAPF32[$150>>2] = 1.4142135381698608;
     break;
    }
   } else {
    $167 = (717796 + ($i$868<<2)|0);
    HEAPF32[$167>>2] = 1.0;
    $168 = (717924 + ($i$868<<2)|0);
    HEAPF32[$168>>2] = 1.0;
    $169 = (718052 + ($i$868<<2)|0);
    HEAPF32[$169>>2] = 1.4142135381698608;
    $170 = (718180 + ($i$868<<2)|0);
    HEAPF32[$170>>2] = 1.4142135381698608;
    $171 = (717860 + ($i$868<<2)|0);
    HEAPF32[$171>>2] = 1.0;
    $172 = (717988 + ($i$868<<2)|0);
    HEAPF32[$172>>2] = 1.0;
    $173 = (718116 + ($i$868<<2)|0);
    HEAPF32[$173>>2] = 1.4142135381698608;
    $174 = (718244 + ($i$868<<2)|0);
    HEAPF32[$174>>2] = 1.4142135381698608;
   }
  } while(0);
  $175 = (($i$868) + 1)|0;
  $exitcond126 = ($175|0)==(16);
  if ($exitcond126) {
   break;
  } else {
   $i$868 = $175;
  }
 }
 $176 = HEAP32[179593]|0;
 HEAP32[179745] = $176;
 $177 = HEAP32[(718380)>>2]|0;
 HEAP32[(718988)>>2] = $177;
 $178 = HEAP32[(718388)>>2]|0;
 HEAP32[(718996)>>2] = $178;
 $179 = HEAP32[(718396)>>2]|0;
 HEAP32[(719004)>>2] = $179;
 $180 = HEAP32[(718404)>>2]|0;
 HEAP32[(719012)>>2] = $180;
 $181 = HEAP32[(718412)>>2]|0;
 HEAP32[(719020)>>2] = $181;
 $182 = HEAP32[(718420)>>2]|0;
 HEAP32[(719028)>>2] = $182;
 $183 = HEAP32[(718428)>>2]|0;
 HEAP32[(719036)>>2] = $183;
 $184 = HEAP32[(718436)>>2]|0;
 HEAP32[(719044)>>2] = $184;
 $185 = HEAP32[(718444)>>2]|0;
 HEAP32[(719052)>>2] = $185;
 $186 = HEAP32[(718452)>>2]|0;
 HEAP32[(719060)>>2] = $186;
 $187 = HEAP32[(718460)>>2]|0;
 HEAP32[(719068)>>2] = $187;
 $188 = HEAP32[(718468)>>2]|0;
 HEAP32[(719076)>>2] = $188;
 $189 = HEAP32[(718476)>>2]|0;
 HEAP32[(719084)>>2] = $189;
 $190 = HEAP32[(718484)>>2]|0;
 HEAP32[(719092)>>2] = $190;
 $191 = HEAP32[(718492)>>2]|0;
 HEAP32[(719100)>>2] = $191;
 $192 = HEAP32[(718500)>>2]|0;
 HEAP32[(719108)>>2] = $192;
 $193 = HEAP32[(718508)>>2]|0;
 HEAP32[(719116)>>2] = $193;
 $194 = +HEAPF32[(718376)>>2];
 $195 = -$194;
 HEAPF32[(718984)>>2] = $195;
 $196 = +HEAPF32[(718384)>>2];
 $197 = -$196;
 HEAPF32[(718992)>>2] = $197;
 $198 = +HEAPF32[(718392)>>2];
 $199 = -$198;
 HEAPF32[(719000)>>2] = $199;
 $200 = +HEAPF32[(718400)>>2];
 $201 = -$200;
 HEAPF32[(719008)>>2] = $201;
 $202 = +HEAPF32[(718408)>>2];
 $203 = -$202;
 HEAPF32[(719016)>>2] = $203;
 $204 = +HEAPF32[(718416)>>2];
 $205 = -$204;
 HEAPF32[(719024)>>2] = $205;
 $206 = +HEAPF32[(718424)>>2];
 $207 = -$206;
 HEAPF32[(719032)>>2] = $207;
 $208 = +HEAPF32[(718432)>>2];
 $209 = -$208;
 HEAPF32[(719040)>>2] = $209;
 $210 = +HEAPF32[(718440)>>2];
 $211 = -$210;
 HEAPF32[(719048)>>2] = $211;
 $212 = +HEAPF32[(718448)>>2];
 $213 = -$212;
 HEAPF32[(719056)>>2] = $213;
 $214 = +HEAPF32[(718456)>>2];
 $215 = -$214;
 HEAPF32[(719064)>>2] = $215;
 $216 = +HEAPF32[(718464)>>2];
 $217 = -$216;
 HEAPF32[(719072)>>2] = $217;
 $218 = +HEAPF32[(718472)>>2];
 $219 = -$218;
 HEAPF32[(719080)>>2] = $219;
 $220 = +HEAPF32[(718480)>>2];
 $221 = -$220;
 HEAPF32[(719088)>>2] = $221;
 $222 = +HEAPF32[(718488)>>2];
 $223 = -$222;
 HEAPF32[(719096)>>2] = $223;
 $224 = +HEAPF32[(718496)>>2];
 $225 = -$224;
 HEAPF32[(719104)>>2] = $225;
 $226 = +HEAPF32[(718504)>>2];
 $227 = -$226;
 HEAPF32[(719112)>>2] = $227;
 $228 = +HEAPF32[(718512)>>2];
 $229 = -$228;
 HEAPF32[(719120)>>2] = $229;
 $230 = HEAP32[(718516)>>2]|0;
 HEAP32[(719124)>>2] = $230;
 $231 = HEAP32[(718524)>>2]|0;
 HEAP32[(719132)>>2] = $231;
 $232 = HEAP32[(718532)>>2]|0;
 HEAP32[(719140)>>2] = $232;
 $233 = HEAP32[(718540)>>2]|0;
 HEAP32[(719148)>>2] = $233;
 $234 = HEAP32[(718548)>>2]|0;
 HEAP32[(719156)>>2] = $234;
 $235 = HEAP32[(718556)>>2]|0;
 HEAP32[(719164)>>2] = $235;
 $236 = HEAP32[(718564)>>2]|0;
 HEAP32[(719172)>>2] = $236;
 $237 = HEAP32[(718572)>>2]|0;
 HEAP32[(719180)>>2] = $237;
 $238 = HEAP32[(718580)>>2]|0;
 HEAP32[(719188)>>2] = $238;
 $239 = HEAP32[(718588)>>2]|0;
 HEAP32[(719196)>>2] = $239;
 $240 = HEAP32[(718596)>>2]|0;
 HEAP32[(719204)>>2] = $240;
 $241 = HEAP32[(718604)>>2]|0;
 HEAP32[(719212)>>2] = $241;
 $242 = HEAP32[(718612)>>2]|0;
 HEAP32[(719220)>>2] = $242;
 $243 = HEAP32[(718620)>>2]|0;
 HEAP32[(719228)>>2] = $243;
 $244 = HEAP32[(718628)>>2]|0;
 HEAP32[(719236)>>2] = $244;
 $245 = HEAP32[(718636)>>2]|0;
 HEAP32[(719244)>>2] = $245;
 $246 = HEAP32[(718644)>>2]|0;
 HEAP32[(719252)>>2] = $246;
 $247 = HEAP32[(718652)>>2]|0;
 HEAP32[(719260)>>2] = $247;
 $248 = +HEAPF32[(718520)>>2];
 $249 = -$248;
 HEAPF32[(719128)>>2] = $249;
 $250 = +HEAPF32[(718528)>>2];
 $251 = -$250;
 HEAPF32[(719136)>>2] = $251;
 $252 = +HEAPF32[(718536)>>2];
 $253 = -$252;
 HEAPF32[(719144)>>2] = $253;
 $254 = +HEAPF32[(718544)>>2];
 $255 = -$254;
 HEAPF32[(719152)>>2] = $255;
 $256 = +HEAPF32[(718552)>>2];
 $257 = -$256;
 HEAPF32[(719160)>>2] = $257;
 $258 = +HEAPF32[(718560)>>2];
 $259 = -$258;
 HEAPF32[(719168)>>2] = $259;
 $260 = +HEAPF32[(718568)>>2];
 $261 = -$260;
 HEAPF32[(719176)>>2] = $261;
 $262 = +HEAPF32[(718576)>>2];
 $263 = -$262;
 HEAPF32[(719184)>>2] = $263;
 $264 = +HEAPF32[(718584)>>2];
 $265 = -$264;
 HEAPF32[(719192)>>2] = $265;
 $266 = +HEAPF32[(718592)>>2];
 $267 = -$266;
 HEAPF32[(719200)>>2] = $267;
 $268 = +HEAPF32[(718600)>>2];
 $269 = -$268;
 HEAPF32[(719208)>>2] = $269;
 $270 = +HEAPF32[(718608)>>2];
 $271 = -$270;
 HEAPF32[(719216)>>2] = $271;
 $272 = +HEAPF32[(718616)>>2];
 $273 = -$272;
 HEAPF32[(719224)>>2] = $273;
 $274 = +HEAPF32[(718624)>>2];
 $275 = -$274;
 HEAPF32[(719232)>>2] = $275;
 $276 = +HEAPF32[(718632)>>2];
 $277 = -$276;
 HEAPF32[(719240)>>2] = $277;
 $278 = +HEAPF32[(718640)>>2];
 $279 = -$278;
 HEAPF32[(719248)>>2] = $279;
 $280 = +HEAPF32[(718648)>>2];
 $281 = -$280;
 HEAPF32[(719256)>>2] = $281;
 $282 = +HEAPF32[(718656)>>2];
 $283 = -$282;
 HEAPF32[(719264)>>2] = $283;
 $284 = HEAP32[(718660)>>2]|0;
 HEAP32[(719268)>>2] = $284;
 $285 = HEAP32[(718668)>>2]|0;
 HEAP32[(719276)>>2] = $285;
 $286 = HEAP32[(718676)>>2]|0;
 HEAP32[(719284)>>2] = $286;
 $287 = HEAP32[(718684)>>2]|0;
 HEAP32[(719292)>>2] = $287;
 $288 = HEAP32[(718692)>>2]|0;
 HEAP32[(719300)>>2] = $288;
 $289 = HEAP32[(718700)>>2]|0;
 HEAP32[(719308)>>2] = $289;
 $290 = +HEAPF32[(718664)>>2];
 $291 = -$290;
 HEAPF32[(719272)>>2] = $291;
 $292 = +HEAPF32[(718672)>>2];
 $293 = -$292;
 HEAPF32[(719280)>>2] = $293;
 $294 = +HEAPF32[(718680)>>2];
 $295 = -$294;
 HEAPF32[(719288)>>2] = $295;
 $296 = +HEAPF32[(718688)>>2];
 $297 = -$296;
 HEAPF32[(719296)>>2] = $297;
 $298 = +HEAPF32[(718696)>>2];
 $299 = -$298;
 HEAPF32[(719304)>>2] = $299;
 $300 = +HEAPF32[(718704)>>2];
 $301 = -$300;
 HEAPF32[(719312)>>2] = $301;
 $302 = HEAP32[(718804)>>2]|0;
 HEAP32[(719412)>>2] = $302;
 $303 = HEAP32[(718812)>>2]|0;
 HEAP32[(719420)>>2] = $303;
 $304 = HEAP32[(718820)>>2]|0;
 HEAP32[(719428)>>2] = $304;
 $305 = HEAP32[(718828)>>2]|0;
 HEAP32[(719436)>>2] = $305;
 $306 = HEAP32[(718836)>>2]|0;
 HEAP32[(719444)>>2] = $306;
 $307 = HEAP32[(718844)>>2]|0;
 HEAP32[(719452)>>2] = $307;
 $308 = HEAP32[(718852)>>2]|0;
 HEAP32[(719460)>>2] = $308;
 $309 = HEAP32[(718860)>>2]|0;
 HEAP32[(719468)>>2] = $309;
 $310 = HEAP32[(718868)>>2]|0;
 HEAP32[(719476)>>2] = $310;
 $311 = HEAP32[(718876)>>2]|0;
 HEAP32[(719484)>>2] = $311;
 $312 = HEAP32[(718884)>>2]|0;
 HEAP32[(719492)>>2] = $312;
 $313 = HEAP32[(718892)>>2]|0;
 HEAP32[(719500)>>2] = $313;
 $314 = HEAP32[(718900)>>2]|0;
 HEAP32[(719508)>>2] = $314;
 $315 = HEAP32[(718908)>>2]|0;
 HEAP32[(719516)>>2] = $315;
 $316 = HEAP32[(718916)>>2]|0;
 HEAP32[(719524)>>2] = $316;
 $317 = HEAP32[(718924)>>2]|0;
 HEAP32[(719532)>>2] = $317;
 $318 = HEAP32[(718932)>>2]|0;
 HEAP32[(719540)>>2] = $318;
 $319 = HEAP32[(718940)>>2]|0;
 HEAP32[(719548)>>2] = $319;
 $320 = +HEAPF32[(718808)>>2];
 $321 = -$320;
 HEAPF32[(719416)>>2] = $321;
 $322 = +HEAPF32[(718816)>>2];
 $323 = -$322;
 HEAPF32[(719424)>>2] = $323;
 $324 = +HEAPF32[(718824)>>2];
 $325 = -$324;
 HEAPF32[(719432)>>2] = $325;
 $326 = +HEAPF32[(718832)>>2];
 $327 = -$326;
 HEAPF32[(719440)>>2] = $327;
 $328 = +HEAPF32[(718840)>>2];
 $329 = -$328;
 HEAPF32[(719448)>>2] = $329;
 $330 = +HEAPF32[(718848)>>2];
 $331 = -$330;
 HEAPF32[(719456)>>2] = $331;
 $332 = +HEAPF32[(718856)>>2];
 $333 = -$332;
 HEAPF32[(719464)>>2] = $333;
 $334 = +HEAPF32[(718864)>>2];
 $335 = -$334;
 HEAPF32[(719472)>>2] = $335;
 $336 = +HEAPF32[(718872)>>2];
 $337 = -$336;
 HEAPF32[(719480)>>2] = $337;
 $338 = +HEAPF32[(718880)>>2];
 $339 = -$338;
 HEAPF32[(719488)>>2] = $339;
 $340 = +HEAPF32[(718888)>>2];
 $341 = -$340;
 HEAPF32[(719496)>>2] = $341;
 $342 = +HEAPF32[(718896)>>2];
 $343 = -$342;
 HEAPF32[(719504)>>2] = $343;
 $344 = +HEAPF32[(718904)>>2];
 $345 = -$344;
 HEAPF32[(719512)>>2] = $345;
 $346 = +HEAPF32[(718912)>>2];
 $347 = -$346;
 HEAPF32[(719520)>>2] = $347;
 $348 = +HEAPF32[(718920)>>2];
 $349 = -$348;
 HEAPF32[(719528)>>2] = $349;
 $350 = +HEAPF32[(718928)>>2];
 $351 = -$350;
 HEAPF32[(719536)>>2] = $351;
 $352 = +HEAPF32[(718936)>>2];
 $353 = -$352;
 HEAPF32[(719544)>>2] = $353;
 $354 = +HEAPF32[(718944)>>2];
 $355 = -$354;
 HEAPF32[(719552)>>2] = $355;
 $indvars$iv = (726900);$indvars$iv106 = (727380);$indvars$iv112 = (721780);$indvars$iv116 = (719748);$j$260 = 0;
 while(1) {
  $356 = (726772 + (($j$260*608)|0)|0);
  $357 = (684496 + (($j$260*12)|0)|0);
  HEAP32[$357>>2] = $356;
  $358 = (((161928 + (($j$260*110)|0)|0)) + 46|0);
  $359 = HEAP8[$358>>0]|0;
  $360 = $359&255;
  $361 = $360 >>> 1;
  $362 = (((726772 + (($j$260*608)|0)|0)) + 4|0);
  HEAP32[$356>>2] = $361;
  $363 = (((726772 + (($j$260*608)|0)|0)) + 8|0);
  HEAP32[$362>>2] = 0;
  $364 = (((726772 + (($j$260*608)|0)|0)) + 12|0);
  HEAP32[$363>>2] = 3;
  $365 = (((726772 + (($j$260*608)|0)|0)) + 16|0);
  HEAP32[$364>>2] = 0;
  $366 = (((161928 + (($j$260*110)|0)|0)) + 47|0);
  $367 = HEAP8[$366>>0]|0;
  $368 = $367&255;
  $369 = $368 >>> 1;
  $370 = (((726772 + (($j$260*608)|0)|0)) + 20|0);
  HEAP32[$365>>2] = $369;
  $371 = (((726772 + (($j$260*608)|0)|0)) + 24|0);
  HEAP32[$370>>2] = $360;
  $372 = (((726772 + (($j$260*608)|0)|0)) + 28|0);
  HEAP32[$371>>2] = 3;
  $373 = (((726772 + (($j$260*608)|0)|0)) + 32|0);
  HEAP32[$372>>2] = 1;
  $374 = (((161928 + (($j$260*110)|0)|0)) + 48|0);
  $375 = (($368) + ($360))|0;
  $376 = HEAP8[$374>>0]|0;
  $377 = $376&255;
  $378 = $377 >>> 1;
  $379 = (((726772 + (($j$260*608)|0)|0)) + 36|0);
  HEAP32[$373>>2] = $378;
  $380 = (((726772 + (($j$260*608)|0)|0)) + 40|0);
  HEAP32[$379>>2] = $375;
  $381 = (((726772 + (($j$260*608)|0)|0)) + 44|0);
  HEAP32[$380>>2] = 3;
  $382 = (((726772 + (($j$260*608)|0)|0)) + 48|0);
  HEAP32[$381>>2] = 2;
  $383 = (((161928 + (($j$260*110)|0)|0)) + 49|0);
  $384 = (($377) + ($375))|0;
  $385 = HEAP8[$383>>0]|0;
  $386 = $385&255;
  $387 = $386 >>> 1;
  $388 = (((726772 + (($j$260*608)|0)|0)) + 52|0);
  HEAP32[$382>>2] = $387;
  $389 = (((726772 + (($j$260*608)|0)|0)) + 56|0);
  HEAP32[$388>>2] = $384;
  $390 = (((726772 + (($j$260*608)|0)|0)) + 60|0);
  HEAP32[$389>>2] = 3;
  $391 = (((726772 + (($j$260*608)|0)|0)) + 64|0);
  HEAP32[$390>>2] = 3;
  $392 = (((161928 + (($j$260*110)|0)|0)) + 50|0);
  $393 = (($386) + ($384))|0;
  $394 = HEAP8[$392>>0]|0;
  $395 = $394&255;
  $396 = $395 >>> 1;
  $397 = (((726772 + (($j$260*608)|0)|0)) + 68|0);
  HEAP32[$391>>2] = $396;
  $398 = (((726772 + (($j$260*608)|0)|0)) + 72|0);
  HEAP32[$397>>2] = $393;
  $399 = (((726772 + (($j$260*608)|0)|0)) + 76|0);
  HEAP32[$398>>2] = 3;
  $400 = (((726772 + (($j$260*608)|0)|0)) + 80|0);
  HEAP32[$399>>2] = 4;
  $401 = (((161928 + (($j$260*110)|0)|0)) + 51|0);
  $402 = (($395) + ($393))|0;
  $403 = HEAP8[$401>>0]|0;
  $404 = $403&255;
  $405 = $404 >>> 1;
  $406 = (((726772 + (($j$260*608)|0)|0)) + 84|0);
  HEAP32[$400>>2] = $405;
  $407 = (((726772 + (($j$260*608)|0)|0)) + 88|0);
  HEAP32[$406>>2] = $402;
  $408 = (((726772 + (($j$260*608)|0)|0)) + 92|0);
  HEAP32[$407>>2] = 3;
  $409 = (((726772 + (($j$260*608)|0)|0)) + 96|0);
  HEAP32[$408>>2] = 5;
  $410 = (((161928 + (($j$260*110)|0)|0)) + 52|0);
  $411 = (($404) + ($402))|0;
  $412 = HEAP8[$410>>0]|0;
  $413 = $412&255;
  $414 = $413 >>> 1;
  $415 = (((726772 + (($j$260*608)|0)|0)) + 100|0);
  HEAP32[$409>>2] = $414;
  $416 = (((726772 + (($j$260*608)|0)|0)) + 104|0);
  HEAP32[$415>>2] = $411;
  $417 = (((726772 + (($j$260*608)|0)|0)) + 108|0);
  HEAP32[$416>>2] = 3;
  $418 = (((726772 + (($j$260*608)|0)|0)) + 112|0);
  HEAP32[$417>>2] = 6;
  $419 = (((161928 + (($j$260*110)|0)|0)) + 53|0);
  $420 = (($413) + ($411))|0;
  $421 = HEAP8[$419>>0]|0;
  $422 = $421&255;
  $423 = $422 >>> 1;
  $424 = (((726772 + (($j$260*608)|0)|0)) + 116|0);
  HEAP32[$418>>2] = $423;
  $425 = (((726772 + (($j$260*608)|0)|0)) + 120|0);
  HEAP32[$424>>2] = $420;
  $426 = (((726772 + (($j$260*608)|0)|0)) + 124|0);
  HEAP32[$425>>2] = 3;
  HEAP32[$426>>2] = 7;
  $427 = (($422) + ($420))|0;
  $428 = (((161928 + (($j$260*110)|0)|0)) + 99|0);
  $lftr$limit = (((161928 + (($j$260*110)|0)|0)) + 109|0);
  $bdf$148 = $428;$cb$147 = 3;$i$1250 = $427;$mp$149 = $indvars$iv;
  while(1) {
   $429 = HEAP8[$bdf$148>>0]|0;
   $430 = $429&255;
   $431 = $430 >>> 1;
   $432 = ((($mp$149)) + 4|0);
   HEAP32[$mp$149>>2] = $431;
   $433 = ((($mp$149)) + 8|0);
   HEAP32[$432>>2] = $i$1250;
   $434 = ((($mp$149)) + 12|0);
   HEAP32[$433>>2] = 0;
   $435 = ((($mp$149)) + 16|0);
   HEAP32[$434>>2] = $cb$147;
   $436 = ((($mp$149)) + 20|0);
   HEAP32[$435>>2] = $431;
   $437 = (($i$1250) + 1)|0;
   $438 = ((($mp$149)) + 24|0);
   HEAP32[$436>>2] = $437;
   $439 = ((($mp$149)) + 28|0);
   HEAP32[$438>>2] = 1;
   $440 = ((($mp$149)) + 32|0);
   HEAP32[$439>>2] = $cb$147;
   $441 = ((($mp$149)) + 36|0);
   HEAP32[$440>>2] = $431;
   $442 = (($i$1250) + 2)|0;
   $443 = ((($mp$149)) + 40|0);
   HEAP32[$441>>2] = $442;
   $444 = ((($mp$149)) + 44|0);
   HEAP32[$443>>2] = 2;
   HEAP32[$444>>2] = $cb$147;
   $445 = ((($bdf$148)) + 1|0);
   $scevgep104 = ((($mp$149)) + 48|0);
   $446 = ($431*6)|0;
   $447 = (($446) + ($i$1250))|0;
   $448 = (($cb$147) + 1)|0;
   $exitcond109 = ($445|0)==($lftr$limit|0);
   if ($exitcond109) {
    break;
   } else {
    $bdf$148 = $445;$cb$147 = $448;$i$1250 = $447;$mp$149 = $scevgep104;
   }
  }
  $449 = (((161928 + (($j$260*110)|0)|0)) + 96|0);
  $450 = (684604 + (($j$260*12)|0)|0);
  HEAP32[$450>>2] = $indvars$iv106;
  $451 = (721156 + (($j$260*624)|0)|0);
  $452 = (((684496 + (($j$260*12)|0)|0)) + 4|0);
  HEAP32[$452>>2] = $451;
  $bdf$254 = $449;$cb$253 = 0;$i$1356 = 0;$mp$355 = $451;
  while(1) {
   $453 = HEAP8[$bdf$254>>0]|0;
   $454 = $453&255;
   $455 = $454 >>> 1;
   $456 = ((($mp$355)) + 4|0);
   HEAP32[$mp$355>>2] = $455;
   $457 = ((($mp$355)) + 8|0);
   HEAP32[$456>>2] = $i$1356;
   $458 = ((($mp$355)) + 12|0);
   HEAP32[$457>>2] = 0;
   $459 = ((($mp$355)) + 16|0);
   HEAP32[$458>>2] = $cb$253;
   $460 = ((($mp$355)) + 20|0);
   HEAP32[$459>>2] = $455;
   $461 = $i$1356 | 1;
   $462 = ((($mp$355)) + 24|0);
   HEAP32[$460>>2] = $461;
   $463 = ((($mp$355)) + 28|0);
   HEAP32[$462>>2] = 1;
   $464 = ((($mp$355)) + 32|0);
   HEAP32[$463>>2] = $cb$253;
   $465 = ((($mp$355)) + 36|0);
   HEAP32[$464>>2] = $455;
   $466 = (($i$1356) + 2)|0;
   $467 = ((($mp$355)) + 40|0);
   HEAP32[$465>>2] = $466;
   $468 = ((($mp$355)) + 44|0);
   HEAP32[$467>>2] = 2;
   HEAP32[$468>>2] = $cb$253;
   $469 = ((($bdf$254)) + 1|0);
   $scevgep110 = ((($mp$355)) + 48|0);
   $470 = ($455*6)|0;
   $471 = (($470) + ($i$1356))|0;
   $472 = (($cb$253) + 1)|0;
   $exitcond115 = ($472|0)==(13);
   if ($exitcond115) {
    break;
   } else {
    $bdf$254 = $469;$cb$253 = $472;$i$1356 = $471;$mp$355 = $scevgep110;
   }
  }
  $473 = (((684604 + (($j$260*12)|0)|0)) + 4|0);
  HEAP32[$473>>2] = $indvars$iv112;
  $474 = (719572 + (($j$260*176)|0)|0);
  $475 = (((684496 + (($j$260*12)|0)|0)) + 8|0);
  HEAP32[$475>>2] = $474;
  $476 = HEAP8[$358>>0]|0;
  $477 = $476&255;
  $478 = $477 >>> 1;
  $479 = (((719572 + (($j$260*176)|0)|0)) + 4|0);
  HEAP32[$474>>2] = $478;
  $480 = (((719572 + (($j$260*176)|0)|0)) + 8|0);
  HEAP32[$479>>2] = 0;
  $481 = (((719572 + (($j$260*176)|0)|0)) + 12|0);
  HEAP32[$480>>2] = $369;
  $482 = (((719572 + (($j$260*176)|0)|0)) + 16|0);
  HEAP32[$481>>2] = 1;
  $483 = (((719572 + (($j$260*176)|0)|0)) + 20|0);
  HEAP32[$482>>2] = $378;
  $484 = (((719572 + (($j$260*176)|0)|0)) + 24|0);
  HEAP32[$483>>2] = 2;
  $485 = (((719572 + (($j$260*176)|0)|0)) + 28|0);
  HEAP32[$484>>2] = $387;
  $486 = (((719572 + (($j$260*176)|0)|0)) + 32|0);
  HEAP32[$485>>2] = 3;
  $487 = (((719572 + (($j$260*176)|0)|0)) + 36|0);
  HEAP32[$486>>2] = $396;
  $488 = (((719572 + (($j$260*176)|0)|0)) + 40|0);
  HEAP32[$487>>2] = 4;
  $489 = (((719572 + (($j$260*176)|0)|0)) + 44|0);
  HEAP32[$488>>2] = $405;
  $490 = (((719572 + (($j$260*176)|0)|0)) + 48|0);
  HEAP32[$489>>2] = 5;
  $491 = HEAP8[$410>>0]|0;
  $492 = $491&255;
  $493 = $492 >>> 1;
  $494 = (((719572 + (($j$260*176)|0)|0)) + 52|0);
  HEAP32[$490>>2] = $493;
  $495 = (((719572 + (($j$260*176)|0)|0)) + 56|0);
  HEAP32[$494>>2] = 6;
  $496 = (((161928 + (($j$260*110)|0)|0)) + 54|0);
  $497 = HEAP8[$419>>0]|0;
  $498 = $497&255;
  $499 = $498 >>> 1;
  $500 = (((719572 + (($j$260*176)|0)|0)) + 60|0);
  HEAP32[$495>>2] = $499;
  $501 = (((719572 + (($j$260*176)|0)|0)) + 64|0);
  HEAP32[$500>>2] = 7;
  $502 = (((161928 + (($j$260*110)|0)|0)) + 55|0);
  $503 = HEAP8[$496>>0]|0;
  $504 = $503&255;
  $505 = $504 >>> 1;
  $506 = (((719572 + (($j$260*176)|0)|0)) + 68|0);
  HEAP32[$501>>2] = $505;
  $507 = (((719572 + (($j$260*176)|0)|0)) + 72|0);
  HEAP32[$506>>2] = 8;
  $508 = (((161928 + (($j$260*110)|0)|0)) + 56|0);
  $509 = HEAP8[$502>>0]|0;
  $510 = $509&255;
  $511 = $510 >>> 1;
  $512 = (((719572 + (($j$260*176)|0)|0)) + 76|0);
  HEAP32[$507>>2] = $511;
  $513 = (((719572 + (($j$260*176)|0)|0)) + 80|0);
  HEAP32[$512>>2] = 9;
  $514 = (((161928 + (($j$260*110)|0)|0)) + 57|0);
  $515 = HEAP8[$508>>0]|0;
  $516 = $515&255;
  $517 = $516 >>> 1;
  $518 = (((719572 + (($j$260*176)|0)|0)) + 84|0);
  HEAP32[$513>>2] = $517;
  $519 = (((719572 + (($j$260*176)|0)|0)) + 88|0);
  HEAP32[$518>>2] = 10;
  $520 = (((161928 + (($j$260*110)|0)|0)) + 58|0);
  $521 = HEAP8[$514>>0]|0;
  $522 = $521&255;
  $523 = $522 >>> 1;
  $524 = (((719572 + (($j$260*176)|0)|0)) + 92|0);
  HEAP32[$519>>2] = $523;
  $525 = (((719572 + (($j$260*176)|0)|0)) + 96|0);
  HEAP32[$524>>2] = 11;
  $526 = (((161928 + (($j$260*110)|0)|0)) + 59|0);
  $527 = HEAP8[$520>>0]|0;
  $528 = $527&255;
  $529 = $528 >>> 1;
  $530 = (((719572 + (($j$260*176)|0)|0)) + 100|0);
  HEAP32[$525>>2] = $529;
  $531 = (((719572 + (($j$260*176)|0)|0)) + 104|0);
  HEAP32[$530>>2] = 12;
  $532 = (((161928 + (($j$260*110)|0)|0)) + 60|0);
  $533 = HEAP8[$526>>0]|0;
  $534 = $533&255;
  $535 = $534 >>> 1;
  $536 = (((719572 + (($j$260*176)|0)|0)) + 108|0);
  HEAP32[$531>>2] = $535;
  $537 = (((719572 + (($j$260*176)|0)|0)) + 112|0);
  HEAP32[$536>>2] = 13;
  $538 = (((161928 + (($j$260*110)|0)|0)) + 61|0);
  $539 = HEAP8[$532>>0]|0;
  $540 = $539&255;
  $541 = $540 >>> 1;
  $542 = (((719572 + (($j$260*176)|0)|0)) + 116|0);
  HEAP32[$537>>2] = $541;
  $543 = (((719572 + (($j$260*176)|0)|0)) + 120|0);
  HEAP32[$542>>2] = 14;
  $544 = (((161928 + (($j$260*110)|0)|0)) + 62|0);
  $545 = HEAP8[$538>>0]|0;
  $546 = $545&255;
  $547 = $546 >>> 1;
  $548 = (((719572 + (($j$260*176)|0)|0)) + 124|0);
  HEAP32[$543>>2] = $547;
  $549 = (((719572 + (($j$260*176)|0)|0)) + 128|0);
  HEAP32[$548>>2] = 15;
  $550 = (((161928 + (($j$260*110)|0)|0)) + 63|0);
  $551 = HEAP8[$544>>0]|0;
  $552 = $551&255;
  $553 = $552 >>> 1;
  $554 = (((719572 + (($j$260*176)|0)|0)) + 132|0);
  HEAP32[$549>>2] = $553;
  $555 = (((719572 + (($j$260*176)|0)|0)) + 136|0);
  HEAP32[$554>>2] = 16;
  $556 = (((161928 + (($j$260*110)|0)|0)) + 64|0);
  $557 = HEAP8[$550>>0]|0;
  $558 = $557&255;
  $559 = $558 >>> 1;
  $560 = (((719572 + (($j$260*176)|0)|0)) + 140|0);
  HEAP32[$555>>2] = $559;
  $561 = (((719572 + (($j$260*176)|0)|0)) + 144|0);
  HEAP32[$560>>2] = 17;
  $562 = (((161928 + (($j$260*110)|0)|0)) + 65|0);
  $563 = HEAP8[$556>>0]|0;
  $564 = $563&255;
  $565 = $564 >>> 1;
  $566 = (((719572 + (($j$260*176)|0)|0)) + 148|0);
  HEAP32[$561>>2] = $565;
  $567 = (((719572 + (($j$260*176)|0)|0)) + 152|0);
  HEAP32[$566>>2] = 18;
  $568 = (((161928 + (($j$260*110)|0)|0)) + 66|0);
  $569 = HEAP8[$562>>0]|0;
  $570 = $569&255;
  $571 = $570 >>> 1;
  $572 = (((719572 + (($j$260*176)|0)|0)) + 156|0);
  HEAP32[$567>>2] = $571;
  $573 = (((719572 + (($j$260*176)|0)|0)) + 160|0);
  HEAP32[$572>>2] = 19;
  $574 = (((161928 + (($j$260*110)|0)|0)) + 67|0);
  $575 = HEAP8[$568>>0]|0;
  $576 = $575&255;
  $577 = $576 >>> 1;
  $578 = (((719572 + (($j$260*176)|0)|0)) + 164|0);
  HEAP32[$573>>2] = $577;
  $579 = (((719572 + (($j$260*176)|0)|0)) + 168|0);
  HEAP32[$578>>2] = 20;
  $580 = HEAP8[$574>>0]|0;
  $581 = $580&255;
  $582 = $581 >>> 1;
  $583 = (((719572 + (($j$260*176)|0)|0)) + 172|0);
  HEAP32[$579>>2] = $582;
  HEAP32[$583>>2] = 21;
  $584 = (((684604 + (($j$260*12)|0)|0)) + 8|0);
  HEAP32[$584>>2] = $indvars$iv116;
  $585 = (($j$260) + 1)|0;
  $scevgep = ((($indvars$iv)) + 608|0);
  $scevgep107 = ((($indvars$iv106)) + 608|0);
  $scevgep113 = ((($indvars$iv112)) + 624|0);
  $scevgep117 = ((($indvars$iv116)) + 176|0);
  $exitcond120 = ($585|0)==(9);
  if ($exitcond120) {
   $i$1440 = 0;
   break;
  } else {
   $indvars$iv = $scevgep;$indvars$iv106 = $scevgep107;$indvars$iv112 = $scevgep113;$indvars$iv116 = $scevgep117;$j$260 = $585;
  }
 }
 while(1) {
  $586 = ($i$1440*36)|0;
  $j$339 = 0;
  while(1) {
   $587 = ($j$339*6)|0;
   $588 = (($587) + ($586))|0;
   $589 = $j$339 << 3;
   $590 = $i$1440 | $589;
   $591 = $590 | 12288;
   $592 = (681424 + ($588<<2)|0);
   HEAP32[$592>>2] = $591;
   $593 = $588 | 1;
   $594 = $590 | 12352;
   $595 = (681424 + ($593<<2)|0);
   HEAP32[$595>>2] = $594;
   $596 = (($588) + 2)|0;
   $597 = $590 | 12416;
   $598 = (681424 + ($596<<2)|0);
   HEAP32[$598>>2] = $597;
   $599 = (($588) + 3)|0;
   $600 = $590 | 12480;
   $601 = (681424 + ($599<<2)|0);
   HEAP32[$601>>2] = $600;
   $602 = (($588) + 4)|0;
   $603 = $590 | 12544;
   $604 = (681424 + ($602<<2)|0);
   HEAP32[$604>>2] = $603;
   $605 = (($588) + 5)|0;
   $606 = $590 | 12608;
   $607 = (681424 + ($605<<2)|0);
   HEAP32[$607>>2] = $606;
   $608 = (($j$339) + 1)|0;
   $exitcond100 = ($608|0)==(6);
   if ($exitcond100) {
    break;
   } else {
    $j$339 = $608;
   }
  }
  $609 = (($i$1440) + 1)|0;
  $exitcond101 = ($609|0)==(5);
  if ($exitcond101) {
   $i$1537 = 0;
   break;
  } else {
   $i$1440 = $609;
  }
 }
 while(1) {
  $610 = $i$1537 << 4;
  $611 = (($610) + 180)|0;
  $612 = $i$1537 | 16384;
  $613 = (681424 + ($611<<2)|0);
  HEAP32[$613>>2] = $612;
  $614 = $i$1537 | 16448;
  $615 = (($610) + 181)|0;
  $616 = (681424 + ($615<<2)|0);
  HEAP32[$616>>2] = $614;
  $617 = $i$1537 | 16512;
  $618 = (($610) + 182)|0;
  $619 = (681424 + ($618<<2)|0);
  HEAP32[$619>>2] = $617;
  $620 = $i$1537 | 16576;
  $621 = (($610) + 183)|0;
  $622 = (681424 + ($621<<2)|0);
  HEAP32[$622>>2] = $620;
  $623 = (($610) + 184)|0;
  $624 = $i$1537 | 16392;
  $625 = (681424 + ($623<<2)|0);
  HEAP32[$625>>2] = $624;
  $626 = $i$1537 | 16456;
  $627 = (($610) + 185)|0;
  $628 = (681424 + ($627<<2)|0);
  HEAP32[$628>>2] = $626;
  $629 = $i$1537 | 16520;
  $630 = (($610) + 186)|0;
  $631 = (681424 + ($630<<2)|0);
  HEAP32[$631>>2] = $629;
  $632 = $i$1537 | 16584;
  $633 = (($610) + 187)|0;
  $634 = (681424 + ($633<<2)|0);
  HEAP32[$634>>2] = $632;
  $635 = (($610) + 188)|0;
  $636 = $i$1537 | 16400;
  $637 = (681424 + ($635<<2)|0);
  HEAP32[$637>>2] = $636;
  $638 = $i$1537 | 16464;
  $639 = (($610) + 189)|0;
  $640 = (681424 + ($639<<2)|0);
  HEAP32[$640>>2] = $638;
  $641 = $i$1537 | 16528;
  $642 = (($610) + 190)|0;
  $643 = (681424 + ($642<<2)|0);
  HEAP32[$643>>2] = $641;
  $644 = $i$1537 | 16592;
  $645 = (($610) + 191)|0;
  $646 = (681424 + ($645<<2)|0);
  HEAP32[$646>>2] = $644;
  $647 = (($610) + 192)|0;
  $648 = $i$1537 | 16408;
  $649 = (681424 + ($647<<2)|0);
  HEAP32[$649>>2] = $648;
  $650 = $i$1537 | 16472;
  $651 = (($610) + 193)|0;
  $652 = (681424 + ($651<<2)|0);
  HEAP32[$652>>2] = $650;
  $653 = $i$1537 | 16536;
  $654 = (($610) + 194)|0;
  $655 = (681424 + ($654<<2)|0);
  HEAP32[$655>>2] = $653;
  $656 = $i$1537 | 16600;
  $657 = (($610) + 195)|0;
  $658 = (681424 + ($657<<2)|0);
  HEAP32[$658>>2] = $656;
  $659 = (($i$1537) + 1)|0;
  $exitcond98 = ($659|0)==(4);
  if ($exitcond98) {
   break;
  } else {
   $i$1537 = $659;
  }
 }
 HEAP32[(682400)>>2] = 20480;
 HEAP32[(684448)>>2] = 40960;
 HEAP32[(682404)>>2] = 20488;
 HEAP32[(684452)>>2] = 40968;
 HEAP32[(682408)>>2] = 20496;
 HEAP32[(684456)>>2] = 40976;
 HEAP32[(682412)>>2] = 20481;
 HEAP32[(684460)>>2] = 40961;
 HEAP32[(682416)>>2] = 20489;
 HEAP32[(684464)>>2] = 40969;
 HEAP32[(682420)>>2] = 20497;
 HEAP32[(684468)>>2] = 40977;
 HEAP32[(682424)>>2] = 20482;
 HEAP32[(684472)>>2] = 40962;
 HEAP32[(682428)>>2] = 20490;
 HEAP32[(684476)>>2] = 40970;
 HEAP32[(682432)>>2] = 20498;
 HEAP32[(684480)>>2] = 40978;
 HEAP32[(682436)>>2] = 20483;
 HEAP32[(684484)>>2] = 40963;
 HEAP32[(682440)>>2] = 20491;
 HEAP32[(684488)>>2] = 40971;
 HEAP32[(682444)>>2] = 20499;
 HEAP32[(684492)>>2] = 40979;
 $i$1732 = 0;
 while(1) {
  $660 = ($i$1732*80)|0;
  $j$631 = 0;
  while(1) {
   $661 = $j$631 << 4;
   $662 = (($661) + ($660))|0;
   $663 = $j$631 << 3;
   $664 = $663 | $i$1732;
   $665 = (682448 + ($662<<2)|0);
   HEAP32[$665>>2] = $664;
   $666 = $662 | 1;
   $667 = $664 | 512;
   $668 = (682448 + ($666<<2)|0);
   HEAP32[$668>>2] = $667;
   $669 = $662 | 2;
   $670 = $664 | 1024;
   $671 = (682448 + ($669<<2)|0);
   HEAP32[$671>>2] = $670;
   $672 = $662 | 3;
   $673 = $664 | 1536;
   $674 = (682448 + ($672<<2)|0);
   HEAP32[$674>>2] = $673;
   $675 = $662 | 4;
   $676 = $664 | 64;
   $677 = (682448 + ($675<<2)|0);
   HEAP32[$677>>2] = $676;
   $678 = $662 | 5;
   $679 = $664 | 576;
   $680 = (682448 + ($678<<2)|0);
   HEAP32[$680>>2] = $679;
   $681 = $662 | 6;
   $682 = $664 | 1088;
   $683 = (682448 + ($681<<2)|0);
   HEAP32[$683>>2] = $682;
   $684 = $662 | 7;
   $685 = $664 | 1600;
   $686 = (682448 + ($684<<2)|0);
   HEAP32[$686>>2] = $685;
   $687 = $662 | 8;
   $688 = $664 | 128;
   $689 = (682448 + ($687<<2)|0);
   HEAP32[$689>>2] = $688;
   $690 = $662 | 9;
   $691 = $664 | 640;
   $692 = (682448 + ($690<<2)|0);
   HEAP32[$692>>2] = $691;
   $693 = $662 | 10;
   $694 = $664 | 1152;
   $695 = (682448 + ($693<<2)|0);
   HEAP32[$695>>2] = $694;
   $696 = $662 | 11;
   $697 = $664 | 1664;
   $698 = (682448 + ($696<<2)|0);
   HEAP32[$698>>2] = $697;
   $699 = $662 | 12;
   $700 = $664 | 192;
   $701 = (682448 + ($699<<2)|0);
   HEAP32[$701>>2] = $700;
   $702 = $662 | 13;
   $703 = $664 | 704;
   $704 = (682448 + ($702<<2)|0);
   HEAP32[$704>>2] = $703;
   $705 = $662 | 14;
   $706 = $664 | 1216;
   $707 = (682448 + ($705<<2)|0);
   HEAP32[$707>>2] = $706;
   $708 = $662 | 15;
   $709 = $664 | 1728;
   $710 = (682448 + ($708<<2)|0);
   HEAP32[$710>>2] = $709;
   $711 = (($j$631) + 1)|0;
   $exitcond92 = ($711|0)==(5);
   if ($exitcond92) {
    break;
   } else {
    $j$631 = $711;
   }
  }
  $712 = (($i$1732) + 1)|0;
  $exitcond93 = ($712|0)==(5);
  if ($exitcond93) {
   $i$1828 = 0;
   break;
  } else {
   $i$1732 = $712;
  }
 }
 while(1) {
  $713 = ($i$1828*20)|0;
  $714 = (($713) + 400)|0;
  $715 = $i$1828 | 4096;
  $716 = (682448 + ($714<<2)|0);
  HEAP32[$716>>2] = $715;
  $717 = $i$1828 | 4160;
  $718 = (($713) + 401)|0;
  $719 = (682448 + ($718<<2)|0);
  HEAP32[$719>>2] = $717;
  $720 = $i$1828 | 4224;
  $721 = (($713) + 402)|0;
  $722 = (682448 + ($721<<2)|0);
  HEAP32[$722>>2] = $720;
  $723 = $i$1828 | 4288;
  $724 = (($713) + 403)|0;
  $725 = (682448 + ($724<<2)|0);
  HEAP32[$725>>2] = $723;
  $726 = (($713) + 404)|0;
  $727 = $i$1828 | 4104;
  $728 = (682448 + ($726<<2)|0);
  HEAP32[$728>>2] = $727;
  $729 = $i$1828 | 4168;
  $730 = (($713) + 405)|0;
  $731 = (682448 + ($730<<2)|0);
  HEAP32[$731>>2] = $729;
  $732 = $i$1828 | 4232;
  $733 = (($713) + 406)|0;
  $734 = (682448 + ($733<<2)|0);
  HEAP32[$734>>2] = $732;
  $735 = $i$1828 | 4296;
  $736 = (($713) + 407)|0;
  $737 = (682448 + ($736<<2)|0);
  HEAP32[$737>>2] = $735;
  $738 = (($713) + 408)|0;
  $739 = $i$1828 | 4112;
  $740 = (682448 + ($738<<2)|0);
  HEAP32[$740>>2] = $739;
  $741 = $i$1828 | 4176;
  $742 = (($713) + 409)|0;
  $743 = (682448 + ($742<<2)|0);
  HEAP32[$743>>2] = $741;
  $744 = $i$1828 | 4240;
  $745 = (($713) + 410)|0;
  $746 = (682448 + ($745<<2)|0);
  HEAP32[$746>>2] = $744;
  $747 = $i$1828 | 4304;
  $748 = (($713) + 411)|0;
  $749 = (682448 + ($748<<2)|0);
  HEAP32[$749>>2] = $747;
  $750 = (($713) + 412)|0;
  $751 = $i$1828 | 4120;
  $752 = (682448 + ($750<<2)|0);
  HEAP32[$752>>2] = $751;
  $753 = $i$1828 | 4184;
  $754 = (($713) + 413)|0;
  $755 = (682448 + ($754<<2)|0);
  HEAP32[$755>>2] = $753;
  $756 = $i$1828 | 4248;
  $757 = (($713) + 414)|0;
  $758 = (682448 + ($757<<2)|0);
  HEAP32[$758>>2] = $756;
  $759 = $i$1828 | 4312;
  $760 = (($713) + 415)|0;
  $761 = (682448 + ($760<<2)|0);
  HEAP32[$761>>2] = $759;
  $762 = (($713) + 416)|0;
  $763 = $i$1828 | 4128;
  $764 = (682448 + ($762<<2)|0);
  HEAP32[$764>>2] = $763;
  $765 = $i$1828 | 4192;
  $766 = (($713) + 417)|0;
  $767 = (682448 + ($766<<2)|0);
  HEAP32[$767>>2] = $765;
  $768 = $i$1828 | 4256;
  $769 = (($713) + 418)|0;
  $770 = (682448 + ($769<<2)|0);
  HEAP32[$770>>2] = $768;
  $771 = $i$1828 | 4320;
  $772 = (($713) + 419)|0;
  $773 = (682448 + ($772<<2)|0);
  HEAP32[$773>>2] = $771;
  $774 = (($i$1828) + 1)|0;
  $exitcond89 = ($774|0)==(5);
  if ($exitcond89) {
   break;
  } else {
   $i$1828 = $774;
  }
 }
 return;
}
function _mpg123_exit() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _mpg123_new($decoder,$error) {
 $decoder = $decoder|0;
 $error = $error|0;
 var $$err$1$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $err$1$ph$i = 0, $err$2$i = 0, $fr$12$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[179892]|0;
 $1 = ($0|0)==(0);
 do {
  if ($1) {
   $err$1$ph$i = 8;
   label = 6;
  } else {
   $2 = (_malloc(28616)|0);
   $3 = ($2|0)==(0|0);
   if ($3) {
    $err$1$ph$i = 0;
    label = 6;
   } else {
    _INT123_frame_init_par($2,0);
    $4 = (_INT123_frame_cpu_opt($2,$decoder)|0);
    $5 = ($4|0)==(1);
    if ($5) {
     $6 = ((($2)) + 28380|0);
     HEAP32[$6>>2] = 1;
     $err$2$i = 0;$fr$12$i = $2;
     break;
    } else {
     _INT123_frame_exit($2);
     _free($2);
     $err$1$ph$i = 9;
     label = 6;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 6) {
  $7 = ($err$1$ph$i|0)==(0);
  $$err$1$i = $7 ? 7 : $err$1$ph$i;
  $err$2$i = $$err$1$i;$fr$12$i = 0;
 }
 $8 = ($error|0)==(0|0);
 if ($8) {
  return ($fr$12$i|0);
 }
 HEAP32[$error>>2] = $err$2$i;
 return ($fr$12$i|0);
}
function _mpg123_param($mh,$key,$val,$fval) {
 $mh = $mh|0;
 $key = $key|0;
 $val = $val|0;
 $fval = +$fval;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $key$off = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mh|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ((($mh)) + 28064|0);
 $2 = (_mpg123_par($1,$key,$val,$fval)|0);
 $3 = ($2|0)==(0);
 if (!($3)) {
  $4 = ((($mh)) + 28376|0);
  HEAP32[$4>>2] = $2;
  $$0 = -1;
  return ($$0|0);
 }
 $5 = ($key|0)==(15);
 if ($5) {
  $6 = (_INT123_frame_index_setup($mh)|0);
  $7 = ($6|0)==(0);
  if ($7) {
   $$0 = 0;
   return ($$0|0);
  }
  $8 = ((($mh)) + 28376|0);
  HEAP32[$8>>2] = 36;
  $$0 = $6;
  return ($$0|0);
 } else {
  $key$off = (($key) + -17)|0;
  $9 = ($key$off>>>0)<(2);
  if (!($9)) {
   $$0 = 0;
   return ($$0|0);
  }
  $10 = ((($mh)) + 28020|0);
  $11 = ((($mh)) + 28364|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ((($mh)) + 28368|0);
  $14 = HEAP32[$13>>2]|0;
  _bc_poolsize($10,$12,$14);
  $$0 = 0;
  return ($$0|0);
 }
 return (0)|0;
}
function _mpg123_par($mp,$key,$val,$fval) {
 $mp = $mp|0;
 $key = $key|0;
 $val = $val|0;
 $fval = +$fval;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mp|0)==(0|0);
 L1: do {
  if ($0) {
   $$0 = 25;
  } else {
   do {
    switch ($key|0) {
    case 0:  {
     HEAP32[$mp>>2] = $val;
     $$0 = 0;
     break L1;
     break;
    }
    case 1:  {
     $1 = ((($mp)) + 4|0);
     HEAP32[$1>>2] = $val;
     $$0 = 0;
     break L1;
     break;
    }
    case 2:  {
     $2 = ((($mp)) + 4|0);
     $3 = HEAP32[$2>>2]|0;
     $4 = $3 | $val;
     HEAP32[$2>>2] = $4;
     $$0 = 0;
     break L1;
     break;
    }
    case 13:  {
     $5 = $val ^ -1;
     $6 = ((($mp)) + 4|0);
     $7 = HEAP32[$6>>2]|0;
     $8 = $7 & $5;
     HEAP32[$6>>2] = $8;
     $$0 = 0;
     break L1;
     break;
    }
    case 3:  {
     $9 = ($val|0)>(96000);
     if ($9) {
      $$0 = 3;
      break L1;
     }
     $10 = ($val|0)<(0);
     $11 = $10 ? 0 : $val;
     $12 = ((($mp)) + 8|0);
     HEAP32[$12>>2] = $11;
     $$0 = 0;
     break L1;
     break;
    }
    case 4:  {
     $13 = ($val>>>0)>(2);
     if ($13) {
      $$0 = 3;
      break L1;
     }
     $14 = ((($mp)) + 12|0);
     HEAP32[$14>>2] = $val;
     $$0 = 0;
     break L1;
     break;
    }
    case 5:  {
     $15 = ($val>>>0)>(2);
     if ($15) {
      $$0 = 12;
      break L1;
     }
     $16 = ((($mp)) + 16|0);
     HEAP32[$16>>2] = $val;
     $$0 = 0;
     break L1;
     break;
    }
    case 6:  {
     $17 = ($val|0)<(0);
     $18 = $17 ? 0 : $val;
     $19 = ((($mp)) + 20|0);
     HEAP32[$19>>2] = $18;
     $$0 = 0;
     break L1;
     break;
    }
    case 7:  {
     $20 = ($val|0)<(0);
     $21 = $20 ? 0 : $val;
     $22 = ((($mp)) + 24|0);
     HEAP32[$22>>2] = $21;
     $$0 = 0;
     break L1;
     break;
    }
    case 10:  {
     $23 = ($val|0)>(0);
     $24 = $23 ? $val : 0;
     $25 = ((($mp)) + 272|0);
     HEAP32[$25>>2] = $24;
     $$0 = 0;
     break L1;
     break;
    }
    case 11:  {
     $26 = ($val|0)==(0);
     $27 = (+($val|0));
     $28 = $27 * 3.0517578125E-5;
     $29 = $26 ? $fval : $28;
     $30 = ((($mp)) + 280|0);
     HEAPF64[$30>>3] = $29;
     $$0 = 0;
     break L1;
     break;
    }
    case 12:  {
     $31 = ($val|0)<(0);
     $32 = $31 ? 0 : $val;
     $33 = ((($mp)) + 28|0);
     HEAP32[$33>>2] = $32;
     $$0 = 0;
     break L1;
     break;
    }
    case 14:  {
     $34 = ((($mp)) + 288|0);
     HEAP32[$34>>2] = $val;
     $$0 = 0;
     break L1;
     break;
    }
    case 15:  {
     $35 = ((($mp)) + 292|0);
     HEAP32[$35>>2] = $val;
     $$0 = 0;
     break L1;
     break;
    }
    case 16:  {
     $36 = ($val|0)>(-1);
     if (!($36)) {
      $$0 = 39;
      break L1;
     }
     $37 = ((($mp)) + 296|0);
     HEAP32[$37>>2] = $val;
     $$0 = 0;
     break L1;
     break;
    }
    case 17:  {
     $38 = ($val|0)>(-1);
     if (!($38)) {
      $$0 = 39;
      break L1;
     }
     $39 = ((($mp)) + 300|0);
     HEAP32[$39>>2] = $val;
     $$0 = 0;
     break L1;
     break;
    }
    case 18:  {
     $40 = ($val|0)>(0);
     if (!($40)) {
      $$0 = 39;
      break L1;
     }
     $41 = ((($mp)) + 304|0);
     HEAP32[$41>>2] = $val;
     $$0 = 0;
     break L1;
     break;
    }
    default: {
     $$0 = 5;
     break L1;
    }
    }
   } while(0);
  }
 } while(0);
 return ($$0|0);
}
function _mpg123_open_handle($mh,$iohandle) {
 $mh = $mh|0;
 $iohandle = $iohandle|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mh|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ((($mh)) + 27956|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if (!($5)) {
  FUNCTION_TABLE_vi[$4 & 8191]($mh);
 }
 $6 = ((($mh)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0);
 if (!($8)) {
  $9 = ((($mh)) + 27876|0);
  _INT123_invalidate_format($9);
  HEAP32[$6>>2] = 0;
 }
 (_INT123_frame_reset($mh)|0);
 $10 = ((($mh)) + 27996|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0|0);
 if ($12) {
  $13 = ((($mh)) + 28376|0);
  HEAP32[$13>>2] = 41;
  $$0 = -1;
  return ($$0|0);
 } else {
  $14 = (_INT123_open_stream_handle($mh,$iohandle)|0);
  $$0 = $14;
  return ($$0|0);
 }
 return (0)|0;
}
function _mpg123_replace_reader_handle($mh,$r_read,$r_lseek,$cleanup) {
 $mh = $mh|0;
 $r_read = $r_read|0;
 $r_lseek = $r_lseek|0;
 $cleanup = $cleanup|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mh|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ((($mh)) + 27956|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if (!($5)) {
  FUNCTION_TABLE_vi[$4 & 8191]($mh);
 }
 $6 = ((($mh)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0);
 if (!($8)) {
  $9 = ((($mh)) + 27876|0);
  _INT123_invalidate_format($9);
  HEAP32[$6>>2] = 0;
 }
 (_INT123_frame_reset($mh)|0);
 $10 = ((($mh)) + 27996|0);
 HEAP32[$10>>2] = $r_read;
 $11 = ((($mh)) + 28000|0);
 HEAP32[$11>>2] = $r_lseek;
 $12 = ((($mh)) + 28004|0);
 HEAP32[$12>>2] = $cleanup;
 $$0 = 0;
 return ($$0|0);
}
function _INT123_decode_update($mh) {
 $mh = $mh|0;
 var $$0 = 0, $$ph = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $9 = 0, $phitmp = 0, $phitmp1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($mh)) + 19748|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)<(0);
 if ($2) {
  $3 = ((($mh)) + 28068|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = $4 & 32;
  $6 = ($5|0)==(0);
  if ($6) {
   $7 = HEAP32[37544]|0;
   HEAP32[$vararg_buffer>>2] = 498;
   (_fprintf($7,516024,$vararg_buffer)|0);
  }
  $8 = ((($mh)) + 28376|0);
  HEAP32[$8>>2] = 37;
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $9 = (_INT123_frame_freq($mh)|0);
 $10 = (_INT123_frame_output_format($mh)|0);
 $11 = ($10|0)<(0);
 if ($11) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $12 = ($10|0)==(1);
 if ($12) {
  $13 = ((($mh)) + 4|0);
  HEAP32[$13>>2] = 1;
 }
 $14 = ((($mh)) + 27888|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)==($9|0);
 do {
  if ($16) {
   $17 = ((($mh)) + 19680|0);
   HEAP32[$17>>2] = 0;
   $$ph = 0;
   label = 14;
  } else {
   $18 = $9 >> 1;
   $19 = ($15|0)==($18|0);
   if ($19) {
    $20 = ((($mh)) + 19680|0);
    HEAP32[$20>>2] = 1;
    $$ph = 1;
    label = 14;
    break;
   }
   $21 = $9 >> 2;
   $22 = ($15|0)==($21|0);
   $23 = ((($mh)) + 19680|0);
   if ($22) {
    HEAP32[$23>>2] = 2;
    $$ph = 2;
    label = 14;
    break;
   }
   HEAP32[$23>>2] = 3;
   $40 = (_INT123_synth_ntom_set_step($mh)|0);
   $41 = ($40|0)==(0);
   if (!($41)) {
    $$0 = -1;
    STACKTOP = sp;return ($$0|0);
   }
   $42 = (_INT123_frame_freq($mh)|0);
   $43 = HEAP32[$14>>2]|0;
   $44 = ($42|0)>($43|0);
   if ($44) {
    $45 = $43 << 5;
    $46 = ((($mh)) + 19668|0);
    HEAP32[$46>>2] = $45;
    $47 = (_INT123_frame_freq($mh)|0);
    $48 = HEAP32[$46>>2]|0;
    $49 = (($48|0) / ($47|0))&-1;
    HEAP32[$46>>2] = $49;
   } else {
    $50 = ((($mh)) + 19668|0);
    HEAP32[$50>>2] = 32;
   }
   $51 = ((($mh)) + 19688|0);
   $52 = HEAP32[$51>>2]|0;
   $53 = ($52|0)==(1);
   if ($53) {
    $66 = 384;
   } else {
    $54 = ($52|0)==(2);
    if ($54) {
     $66 = 1152;
    } else {
     $55 = ((($mh)) + 19672|0);
     $56 = HEAP32[$55>>2]|0;
     $57 = ($56|0)==(0);
     if ($57) {
      $58 = ((($mh)) + 19676|0);
      $59 = HEAP32[$58>>2]|0;
      $60 = ($59|0)!=(0);
      $phitmp = $60 ? 576 : 1152;
      $66 = $phitmp;
     } else {
      $66 = 576;
     }
    }
   }
   $61 = HEAP32[$14>>2]|0;
   $62 = $61 << 15;
   $63 = (_INT123_frame_freq($mh)|0);
   $64 = (($62>>>0) / ($63>>>0))&-1;
   $65 = Math_imul($64, $66)|0;
   $67 = (($65) + 32767)|0;
   $68 = $67 >>> 15;
   $69 = (_samples_to_storage($mh,$68)|0);
   $70 = ((($mh)) + 27896|0);
   HEAP32[$70>>2] = $69;
  }
 } while(0);
 if ((label|0) == 14) {
  $24 = 32 >>> $$ph;
  $25 = ((($mh)) + 19668|0);
  HEAP32[$25>>2] = $24;
  $26 = ((($mh)) + 19688|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = ($27|0)==(1);
  if ($28) {
   $37 = 384;
  } else {
   $29 = ($27|0)==(2);
   if ($29) {
    $37 = 1152;
   } else {
    $30 = ((($mh)) + 19672|0);
    $31 = HEAP32[$30>>2]|0;
    $32 = ($31|0)==(0);
    if ($32) {
     $33 = ((($mh)) + 19676|0);
     $34 = HEAP32[$33>>2]|0;
     $35 = ($34|0)!=(0);
     $phitmp1 = $35 ? 576 : 1152;
     $37 = $phitmp1;
    } else {
     $37 = 576;
    }
   }
  }
  $36 = $37 >>> $$ph;
  $38 = (_samples_to_storage($mh,$36)|0);
  $39 = ((($mh)) + 27896|0);
  HEAP32[$39>>2] = $38;
 }
 $71 = ((($mh)) + 28068|0);
 $72 = HEAP32[$71>>2]|0;
 $73 = $72 & 7;
 $74 = ($73|0)==(0);
 do {
  if ($74) {
   $75 = ((($mh)) + 27884|0);
   $76 = HEAP32[$75>>2]|0;
   $77 = ($76|0)==(1);
   $78 = ((($mh)) + 19660|0);
   if ($77) {
    HEAP32[$78>>2] = 3;
    break;
   } else {
    HEAP32[$78>>2] = -1;
    break;
   }
  } else {
   $79 = (($73) + -1)|0;
   $80 = ((($mh)) + 19660|0);
   HEAP32[$80>>2] = $79;
  }
 } while(0);
 $81 = (_INT123_set_synth_functions($mh)|0);
 $82 = ($81|0)==(0);
 if (!($82)) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $83 = (_INT123_frame_outbuffer($mh)|0);
 $84 = ($83|0)==(0);
 if (!($84)) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 _INT123_do_rva($mh);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _INT123_synth_ntom_set_step($fr) {
 $fr = $fr|0;
 var $$ = 0, $$0 = 0, $$off = 0, $$off6 = 0, $$pre = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond12$i = 0, $exitcond13$i = 0, $exitcond14$i = 0, $exitcond14$i$us = 0, $f$02$i = 0, $f$02$i$us = 0, $f$02$us$i = 0, $f$02$us3$i = 0, $ntm$0$lcssa$i = 0, $ntm$01$i = 0, $ntm$01$i$us = 0;
 var $ntm$01$us$i = 0, $ntm$01$us4$i = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_buffer5 = 0, $vararg_ptr1 = 0, $vararg_ptr10 = 0, $vararg_ptr8 = 0, $vararg_ptr9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 16|0;
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_INT123_frame_freq($fr)|0);
 $1 = ((($fr)) + 27888|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($fr)) + 28068|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 & 32;
 $6 = ($5|0)==(0);
 if ($6) {
  $7 = ((($fr)) + 28064|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8|0)>(1);
  if ($9) {
   $10 = HEAP32[37544]|0;
   HEAP32[$vararg_buffer>>2] = $0;
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
   HEAP32[$vararg_ptr1>>2] = $2;
   (_fprintf($10,516146,$vararg_buffer)|0);
  }
 }
 $$off = (($0) + -1)|0;
 $11 = ($$off>>>0)>(95999);
 $$off6 = (($2) + -1)|0;
 $12 = ($$off6>>>0)>(95999);
 $13 = $11 | $12;
 if ($13) {
  $14 = HEAP32[$3>>2]|0;
  $15 = $14 & 32;
  $16 = ($15|0)==(0);
  if ($16) {
   $17 = HEAP32[37544]|0;
   HEAP32[$vararg_buffer2>>2] = 22;
   (_fprintf($17,516177,$vararg_buffer2)|0);
  }
  $18 = ((($fr)) + 28376|0);
  HEAP32[$18>>2] = 3;
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $19 = $2 << 15;
 $20 = (($19>>>0) / ($0>>>0))&-1;
 $21 = ((($fr)) + 19360|0);
 HEAP32[$21>>2] = $20;
 $22 = ($20>>>0)>(262144);
 if ($22) {
  $23 = HEAP32[$3>>2]|0;
  $24 = $23 & 32;
  $25 = ($24|0)==(0);
  if ($25) {
   $26 = HEAP32[37544]|0;
   HEAP32[$vararg_buffer5>>2] = 31;
   $vararg_ptr8 = ((($vararg_buffer5)) + 4|0);
   HEAP32[$vararg_ptr8>>2] = 8;
   $vararg_ptr9 = ((($vararg_buffer5)) + 8|0);
   HEAP32[$vararg_ptr9>>2] = $20;
   $vararg_ptr10 = ((($vararg_buffer5)) + 12|0);
   HEAP32[$vararg_ptr10>>2] = 262144;
   (_fprintf($26,516227,$vararg_buffer5)|0);
  }
  $27 = ((($fr)) + 28376|0);
  HEAP32[$27>>2] = 3;
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $28 = ((($fr)) + 19748|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = ($29|0)>(0);
 if ($30) {
  $31 = ((($fr)) + 19688|0);
  $32 = HEAP32[$31>>2]|0;
  $33 = ((($fr)) + 19676|0);
  switch ($32|0) {
  case 1:  {
   $34 = ($20*384)|0;
   $f$02$us$i = 0;$ntm$01$us$i = 16384;
   while(1) {
    $35 = (($ntm$01$us$i) + ($34))|0;
    $36 = (($35|0) / 32768)&-1;
    $37 = $36 << 15;
    $38 = (($35) - ($37))|0;
    $39 = (($f$02$us$i) + 1)|0;
    $exitcond12$i = ($39|0)==($29|0);
    if ($exitcond12$i) {
     $ntm$0$lcssa$i = $38;
     break;
    } else {
     $f$02$us$i = $39;$ntm$01$us$i = $38;
    }
   }
   break;
  }
  case 2:  {
   $50 = ($20*1152)|0;
   $f$02$us3$i = 0;$ntm$01$us4$i = 16384;
   while(1) {
    $51 = (($ntm$01$us4$i) + ($50))|0;
    $52 = (($51|0) / 32768)&-1;
    $53 = $52 << 15;
    $54 = (($51) - ($53))|0;
    $55 = (($f$02$us3$i) + 1)|0;
    $exitcond13$i = ($55|0)==($29|0);
    if ($exitcond13$i) {
     $ntm$0$lcssa$i = $54;
     break;
    } else {
     $f$02$us3$i = $55;$ntm$01$us4$i = $54;
    }
   }
   break;
  }
  default: {
   $40 = ((($fr)) + 19672|0);
   $$pre$i = HEAP32[$40>>2]|0;
   $41 = ($$pre$i|0)==(0);
   if ($41) {
    $$pre = HEAP32[$33>>2]|0;
    $43 = ($$pre|0)==(0);
    $$ = $43 ? 1152 : 576;
    $44 = Math_imul($$, $20)|0;
    $f$02$i$us = 0;$ntm$01$i$us = 16384;
    while(1) {
     $45 = (($44) + ($ntm$01$i$us))|0;
     $46 = (($45|0) / 32768)&-1;
     $47 = $46 << 15;
     $48 = (($45) - ($47))|0;
     $49 = (($f$02$i$us) + 1)|0;
     $exitcond14$i$us = ($49|0)==($29|0);
     if ($exitcond14$i$us) {
      $ntm$0$lcssa$i = $48;
      break;
     } else {
      $f$02$i$us = $49;$ntm$01$i$us = $48;
     }
    }
   } else {
    $42 = ($20*576)|0;
    $f$02$i = 0;$ntm$01$i = 16384;
    while(1) {
     $56 = (($42) + ($ntm$01$i))|0;
     $57 = (($56|0) / 32768)&-1;
     $58 = $57 << 15;
     $59 = (($56) - ($58))|0;
     $60 = (($f$02$i) + 1)|0;
     $exitcond14$i = ($60|0)==($29|0);
     if ($exitcond14$i) {
      $ntm$0$lcssa$i = $59;
      break;
     } else {
      $f$02$i = $60;$ntm$01$i = $59;
     }
    }
   }
  }
  }
 } else {
  $ntm$0$lcssa$i = 16384;
 }
 $61 = ((($fr)) + 19356|0);
 HEAP32[$61>>2] = $ntm$0$lcssa$i;
 $62 = ((($fr)) + 19352|0);
 HEAP32[$62>>2] = $ntm$0$lcssa$i;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _decode_the_frame($fr) {
 $fr = $fr|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0;
 var $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_INT123_frame_expect_outsamples($fr)|0);
 $1 = (_samples_to_storage($fr,$0)|0);
 $2 = ((($fr)) + 19692|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($fr)|0);
 $5 = ((($fr)) + 28388|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + ($4))|0;
 HEAP32[$5>>2] = $7;
 $8 = ((($fr)) + 27864|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($1>>>0)>($9>>>0);
 if (!($10)) {
  _postprocess_buffer($fr);
  STACKTOP = sp;return;
 }
 $11 = ((($fr)) + 28068|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = $12 & 32;
 $14 = ($13|0)==(0);
 if ($14) {
  $15 = ((($fr)) + 28064|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ($16|0)>(1);
  if ($17) {
   $18 = HEAP32[37544]|0;
   $19 = ((($fr)) + 19748|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = (($1) - ($9))|0;
   HEAP32[$vararg_buffer>>2] = $20;
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
   HEAP32[$vararg_ptr1>>2] = $21;
   $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
   HEAP32[$vararg_ptr2>>2] = $9;
   (_fprintf($18,516290,$vararg_buffer)|0);
   $$pre = HEAP32[$8>>2]|0;
   $25 = $$pre;
  } else {
   $25 = $9;
  }
 } else {
  $25 = $9;
 }
 $22 = ((($fr)) + 27856|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = (($23) + ($25)|0);
 $26 = ((($fr)) + 27876|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = $27 & 15;
 $29 = ($28|0)==(0);
 if ($29) {
  $34 = 0;
 } else {
  $30 = ((($fr)) + 9592|0);
  $31 = HEAP32[$30>>2]|0;
  $32 = HEAP8[$31>>0]|0;
  $34 = $32;
 }
 $33 = (($1) - ($25))|0;
 _memset(($24|0),($34|0),($33|0))|0;
 HEAP32[$8>>2] = $1;
 $35 = ((($fr)) + 19748|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = (($36) + 1)|0;
 _INT123_ntom_set_ntom($fr,$37);
 _postprocess_buffer($fr);
 STACKTOP = sp;return;
}
function _frame_buffercheck($fr) {
 $fr = $fr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer3 = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr6 = 0, $vararg_ptr7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $0 = ((($fr)) + 19764|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 1;
 $3 = ($2|0)==(0);
 if ($3) {
  STACKTOP = sp;return;
 }
 $4 = ((($fr)) + 27920|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)>(0);
 if ($6) {
  $7 = ((($fr)) + 19748|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8|0)<($5|0);
  if (!($9)) {
   STACKTOP = sp;return;
  }
 }
 $10 = ((($fr)) + 27912|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)>(-1);
 if ($12) {
  $13 = ((($fr)) + 19748|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ($14|0)<($11|0);
  if (!($15)) {
   $16 = ($14|0)==($11|0);
   if ($16) {
    $17 = ((($fr)) + 27928|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = (_INT123_samples_to_bytes($fr,$18)|0);
    $23 = $19;
   } else {
    $23 = 0;
   }
   $20 = ((($fr)) + 27864|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = ($21|0)>($23|0);
   if ($22) {
    HEAP32[$20>>2] = $23;
    $38 = $23;
   } else {
    $38 = $21;
   }
   $24 = ((($fr)) + 28068|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = $25 & 32;
   $27 = ($26|0)==(0);
   if ($27) {
    $28 = ((($fr)) + 28064|0);
    $29 = HEAP32[$28>>2]|0;
    $30 = ($29|0)>(2);
    if ($30) {
     $31 = HEAP32[37544]|0;
     $32 = HEAP32[$13>>2]|0;
     $33 = HEAP32[$10>>2]|0;
     $34 = ($32|0)==($33|0);
     if ($34) {
      $35 = ((($fr)) + 27928|0);
      $36 = HEAP32[$35>>2]|0;
      $37 = $36;
     } else {
      $37 = 0;
     }
     HEAP32[$vararg_buffer>>2] = $32;
     $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
     HEAP32[$vararg_ptr1>>2] = $37;
     $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
     HEAP32[$vararg_ptr2>>2] = $38;
     (_fprintf($31,516352,$vararg_buffer)|0);
    }
   }
  }
 }
 $39 = ((($fr)) + 27924|0);
 $40 = HEAP32[$39>>2]|0;
 $41 = ($40|0)==(0);
 if ($41) {
  STACKTOP = sp;return;
 }
 $42 = ((($fr)) + 19748|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = ((($fr)) + 27908|0);
 $45 = HEAP32[$44>>2]|0;
 $46 = ($43|0)==($45|0);
 if (!($46)) {
  STACKTOP = sp;return;
 }
 $47 = (_INT123_samples_to_bytes($fr,$40)|0);
 $48 = ((($fr)) + 27864|0);
 $49 = HEAP32[$48>>2]|0;
 $50 = ($49|0)>($47|0);
 do {
  if ($50) {
   $51 = (($49) - ($47))|0;
   HEAP32[$48>>2] = $51;
   $52 = ((($fr)) + 27892|0);
   $53 = HEAP32[$52>>2]|0;
   $54 = ($53|0)==(0);
   $55 = ((($fr)) + 27856|0);
   $56 = HEAP32[$55>>2]|0;
   $57 = (($56) + ($47)|0);
   if ($54) {
    _memmove(($56|0),($57|0),($51|0))|0;
    break;
   } else {
    $58 = ((($fr)) + 27860|0);
    HEAP32[$58>>2] = $57;
    break;
   }
  } else {
   HEAP32[$48>>2] = 0;
  }
 } while(0);
 $59 = ((($fr)) + 28068|0);
 $60 = HEAP32[$59>>2]|0;
 $61 = $60 & 32;
 $62 = ($61|0)==(0);
 if ($62) {
  $63 = ((($fr)) + 28064|0);
  $64 = HEAP32[$63>>2]|0;
  $65 = ($64|0)>(2);
  if ($65) {
   $66 = HEAP32[37544]|0;
   $67 = HEAP32[$42>>2]|0;
   $68 = HEAP32[$39>>2]|0;
   $69 = HEAP32[$48>>2]|0;
   HEAP32[$vararg_buffer3>>2] = $67;
   $vararg_ptr6 = ((($vararg_buffer3)) + 4|0);
   HEAP32[$vararg_ptr6>>2] = $68;
   $vararg_ptr7 = ((($vararg_buffer3)) + 8|0);
   HEAP32[$vararg_ptr7>>2] = $69;
   (_fprintf($66,516434,$vararg_buffer3)|0);
  }
 }
 HEAP32[$39>>2] = 0;
 STACKTOP = sp;return;
}
function _get_next_frame($mh) {
 $mh = $mh|0;
 var $$3 = 0, $$change$0 = 0, $$change$0$lcssa = 0, $$lcssa18 = 0, $$not = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $7 = 0, $8 = 0, $9 = 0, $brmerge = 0, $change$0 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($mh)) + 28380|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($mh)) + 27904|0);
 $3 = ((($mh)) + 27900|0);
 $4 = ((($mh)) + 19684|0);
 $5 = ((($mh)) + 19756|0);
 $6 = ((($mh)) + 19748|0);
 $7 = ((($mh)) + 27908|0);
 $8 = ((($mh)) + 27916|0);
 $9 = ((($mh)) + 28088|0);
 $10 = ((($mh)) + 19692|0);
 $11 = ((($mh)) + 27864|0);
 $12 = ((($mh)) + 19680|0);
 $change$0 = $1;
 while(1) {
  $13 = HEAP32[$2>>2]|0;
  $14 = ($13|0)==(0);
  if (!($14)) {
   $15 = HEAP32[$6>>2]|0;
   $16 = HEAP32[$7>>2]|0;
   $17 = ($15|0)<($16|0);
   if ($17) {
    $18 = HEAP32[$8>>2]|0;
    $19 = ($15|0)<($18|0);
    if (!($19)) {
     $20 = HEAP32[$10>>2]|0;
     (FUNCTION_TABLE_ii[$20 & 8191]($mh)|0);
     HEAP32[$11>>2] = 0;
     $21 = HEAP32[$12>>2]|0;
     $22 = ($21|0)==(3);
     if ($22) {
      $23 = HEAP32[$6>>2]|0;
      $24 = (($23) + 1)|0;
      _INT123_ntom_set_ntom($mh,$24);
     }
     HEAP32[$3>>2] = 0;
     HEAP32[$2>>2] = 0;
    }
   }
  }
  HEAP32[$3>>2] = 0;
  $25 = (_INT123_read_frame($mh)|0);
  $26 = ($25|0)==(-10);
  if ($26) {
   $$3 = -10;
   label = 26;
   break;
  }
  $27 = ($25|0)<(1);
  if ($27) {
   $$lcssa18 = $25;
   label = 10;
   break;
  }
  $38 = HEAP32[$4>>2]|0;
  $39 = ($38|0)>(1);
  $$change$0 = $39 ? 1 : $change$0;
  $40 = HEAP32[$5>>2]|0;
  $41 = (($40) + 1)|0;
  HEAP32[$5>>2] = $41;
  $42 = HEAP32[$6>>2]|0;
  $43 = HEAP32[$7>>2]|0;
  $44 = ($42|0)<($43|0);
  if (!($44)) {
   $45 = HEAP32[$9>>2]|0;
   $46 = ($45|0)==(0);
   if ($46) {
    $$change$0$lcssa = $$change$0;
    label = 20;
    break;
   }
   $47 = (($41|0) % ($45|0))&-1;
   $48 = ($47|0)==(0);
   if ($48) {
    $$change$0$lcssa = $$change$0;
    label = 20;
    break;
   }
  }
  $49 = HEAP32[$2>>2]|0;
  $50 = ($49|0)==(0);
  $$not = $44 ^ 1;
  $brmerge = $50 | $$not;
  if (!($brmerge)) {
   $51 = HEAP32[$8>>2]|0;
   $52 = ($42|0)<($51|0);
   if (!($52)) {
    $change$0 = $$change$0;
    continue;
   }
  }
  _INT123_frame_skip($mh);
  $change$0 = $$change$0;
 }
 if ((label|0) == 10) {
  $28 = ($$lcssa18|0)==(0);
  if (!($28)) {
   $29 = ((($mh)) + 27960|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = ($30|0)>(-1);
   if (!($31)) {
    $$3 = -1;
    return ($$3|0);
   }
   $32 = ((($mh)) + 27964|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ($33|0)==($30|0);
   if (!($34)) {
    $$3 = -1;
    return ($$3|0);
   }
  }
  $35 = HEAP32[$6>>2]|0;
  $36 = (($35) + 1)|0;
  $37 = ((($mh)) + 19840|0);
  HEAP32[$37>>2] = $36;
  $$3 = -12;
  return ($$3|0);
 }
 else if ((label|0) == 20) {
  $53 = ($$change$0$lcssa|0)==(0);
  if (!($53)) {
   $54 = (_INT123_decode_update($mh)|0);
   $55 = ($54|0)<(0);
   if ($55) {
    $$3 = -1;
    return ($$3|0);
   }
   HEAP32[$0>>2] = 0;
   $56 = HEAP32[$mh>>2]|0;
   $57 = ($56|0)==(0);
   if (!($57)) {
    _INT123_frame_gapless_realinit($mh);
    $58 = HEAP32[$6>>2]|0;
    _INT123_frame_set_frameseek($mh,$58);
    HEAP32[$mh>>2] = 0;
    $59 = HEAP32[$6>>2]|0;
    $60 = HEAP32[$7>>2]|0;
    $61 = ($59|0)<($60|0);
    if ($61) {
     $62 = (_get_next_frame($mh)|0);
     $not$ = ($62|0)>(-1);
     if (!($not$)) {
      $$3 = $62;
      return ($$3|0);
     }
    }
   }
  }
  $$3 = 0;
  return ($$3|0);
 }
 else if ((label|0) == 26) {
  return ($$3|0);
 }
 return (0)|0;
}
function _mpg123_read($mh,$out,$size,$done) {
 $mh = $mh|0;
 $out = $out|0;
 $size = $size|0;
 $done = $done|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_mpg123_decode($mh,0,0,$out,$size,$done)|0);
 return ($0|0);
}
function _mpg123_decode($mh,$inmemory,$inmemsize,$outmemory,$outmemsize,$done) {
 $mh = $mh|0;
 $inmemory = $inmemory|0;
 $inmemsize = $inmemsize|0;
 $outmemory = $outmemory|0;
 $outmemsize = $outmemsize|0;
 $done = $done|0;
 var $$ = 0, $$0 = 0, $$01$ph = 0, $$lcssa = 0, $$outmemsize = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $mdone$0$ph = 0, $mdone$0$ph$lcssa = 0, $mdone$1 = 0, $ret$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($done|0)!=(0|0);
 if ($0) {
  HEAP32[$done>>2] = 0;
 }
 $1 = ($mh|0)==(0|0);
 if ($1) {
  $$0 = -1;
  return ($$0|0);
 }
 $2 = ($inmemsize|0)==(0);
 do {
  if ($2) {
   label = 10;
  } else {
   $3 = ($inmemory|0)==(0|0);
   if ($3) {
    $9 = ((($mh)) + 28376|0);
    HEAP32[$9>>2] = 31;
    $mdone$1 = 0;$ret$2 = -1;
    break;
   }
   $4 = (_INT123_feed_more($mh,$inmemory,$inmemsize)|0);
   $5 = ($4|0)==(0);
   if ($5) {
    $6 = ((($mh)) + 28376|0);
    $7 = HEAP32[$6>>2]|0;
    $8 = ($7|0)==(18);
    if ($8) {
     HEAP32[$6>>2] = 0;
     label = 10;
    } else {
     label = 10;
    }
   } else {
    $mdone$1 = 0;$ret$2 = -1;
   }
  }
 } while(0);
 L14: do {
  if ((label|0) == 10) {
   $10 = ($outmemory|0)==(0|0);
   $$outmemsize = $10 ? 0 : $outmemsize;
   $11 = ((($mh)) + 27900|0);
   $12 = ((($mh)) + 27864|0);
   $13 = ((($mh)) + 4|0);
   $14 = ((($mh)) + 27868|0);
   $15 = ((($mh)) + 27896|0);
   $16 = ((($mh)) + 27856|0);
   $17 = ((($mh)) + 27904|0);
   $18 = ((($mh)) + 27860|0);
   $$01$ph = $outmemory;$mdone$0$ph = 0;
   L16: while(1) {
    while(1) {
     $19 = HEAP32[$11>>2]|0;
     $20 = ($19|0)==(0);
     if (!($20)) {
      $21 = HEAP32[$13>>2]|0;
      $22 = ($21|0)==(0);
      if (!($22)) {
       $mdone$0$ph$lcssa = $mdone$0$ph;
       break L16;
      }
      $23 = HEAP32[$14>>2]|0;
      $24 = HEAP32[$12>>2]|0;
      $25 = (($23) - ($24))|0;
      $26 = HEAP32[$15>>2]|0;
      $27 = ($25>>>0)<($26>>>0);
      if ($27) {
       $mdone$1 = $mdone$0$ph;$ret$2 = 14;
       break L14;
      }
      _decode_the_frame($mh);
      HEAP32[$17>>2] = 0;
      HEAP32[$11>>2] = 0;
      $28 = HEAP32[$16>>2]|0;
      HEAP32[$18>>2] = $28;
      _frame_buffercheck($mh);
     }
     $29 = HEAP32[$12>>2]|0;
     $30 = ($29|0)==(0);
     if (!($30)) {
      $$lcssa = $29;
      break;
     }
     $41 = (_get_next_frame($mh)|0);
     $42 = ($41|0)<(0);
     if ($42) {
      $mdone$1 = $mdone$0$ph;$ret$2 = $41;
      break L14;
     }
    }
    $31 = (($$outmemsize) - ($mdone$0$ph))|0;
    $32 = ($$lcssa>>>0)>($31>>>0);
    $$ = $32 ? $31 : $$lcssa;
    $33 = HEAP32[$18>>2]|0;
    _memcpy(($$01$ph|0),($33|0),($$|0))|0;
    $34 = HEAP32[$12>>2]|0;
    $35 = (($34) - ($$))|0;
    HEAP32[$12>>2] = $35;
    $36 = (($$01$ph) + ($$)|0);
    $37 = (($$) + ($mdone$0$ph))|0;
    $38 = HEAP32[$18>>2]|0;
    $39 = (($38) + ($$)|0);
    HEAP32[$18>>2] = $39;
    $40 = ($$outmemsize>>>0)>($37>>>0);
    if ($40) {
     $$01$ph = $36;$mdone$0$ph = $37;
    } else {
     $mdone$1 = $37;$ret$2 = 0;
     break L14;
    }
   }
   HEAP32[$13>>2] = 0;
   $mdone$1 = $mdone$0$ph$lcssa;$ret$2 = -11;
  }
 } while(0);
 if (!($0)) {
  $$0 = $ret$2;
  return ($$0|0);
 }
 HEAP32[$done>>2] = $mdone$1;
 $$0 = $ret$2;
 return ($$0|0);
}
function _mpg123_getformat($mh,$rate,$channels,$encoding) {
 $mh = $mh|0;
 $rate = $rate|0;
 $channels = $channels|0;
 $encoding = $encoding|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $not$$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mh|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ((($mh)) + 19748|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)<(0);
 if ($3) {
  $4 = (_get_next_frame($mh)|0);
  $not$$i = ($4|0)>(-1);
  if (!($not$$i)) {
   $$0 = $4;
   return ($$0|0);
  }
 }
 $5 = ($rate|0)==(0|0);
 if (!($5)) {
  $6 = ((($mh)) + 27888|0);
  $7 = HEAP32[$6>>2]|0;
  HEAP32[$rate>>2] = $7;
 }
 $8 = ($channels|0)==(0|0);
 if (!($8)) {
  $9 = ((($mh)) + 27884|0);
  $10 = HEAP32[$9>>2]|0;
  HEAP32[$channels>>2] = $10;
 }
 $11 = ($encoding|0)==(0|0);
 if (!($11)) {
  $12 = ((($mh)) + 27876|0);
  $13 = HEAP32[$12>>2]|0;
  HEAP32[$encoding>>2] = $13;
 }
 $14 = ((($mh)) + 4|0);
 HEAP32[$14>>2] = 0;
 $$0 = 0;
 return ($$0|0);
}
function _mpg123_tell($mh) {
 $mh = $mh|0;
 var $$0 = 0, $$neg2$i = 0, $$neg4$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $pos$0 = 0, $s$0$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mh|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ((($mh)) + 19748|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)<(0);
 if ($3) {
  $$0 = 0;
  return ($$0|0);
 }
 $4 = ((($mh)) + 27908|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($2|0)<($5|0);
 do {
  if ($6) {
   label = 6;
  } else {
   $7 = ($2|0)==($5|0);
   $8 = ((($mh)) + 27900|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = ($9|0)==(0);
   if ($7) {
    if (!($10)) {
     label = 6;
     break;
    }
   } else {
    if (!($10)) {
     $15 = (_INT123_frame_outs($mh,$2)|0);
     $16 = ((($mh)) + 27864|0);
     $17 = HEAP32[$16>>2]|0;
     $18 = (_INT123_bytes_to_samples($mh,$17)|0);
     $19 = (($15) - ($18))|0;
     $pos$0 = $19;
     break;
    }
   }
   $20 = (($2) + 1)|0;
   $21 = (_INT123_frame_outs($mh,$20)|0);
   $22 = ((($mh)) + 27864|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = (_INT123_bytes_to_samples($mh,$23)|0);
   $25 = (($21) - ($24))|0;
   $pos$0 = $25;
  }
 } while(0);
 if ((label|0) == 6) {
  $11 = (_INT123_frame_outs($mh,$5)|0);
  $12 = ((($mh)) + 27924|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = (($13) + ($11))|0;
  $pos$0 = $14;
 }
 $26 = ((($mh)) + 28068|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = $27 & 64;
 $29 = ($28|0)==(0);
 do {
  if ($29) {
   $s$0$i = $pos$0;
  } else {
   $30 = ((($mh)) + 27944|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = ($31|0)<($pos$0|0);
   if (!($32)) {
    $40 = ((($mh)) + 27936|0);
    $41 = HEAP32[$40>>2]|0;
    $42 = (($pos$0) - ($41))|0;
    $s$0$i = $42;
    break;
   }
   $33 = ((($mh)) + 27948|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = ($34|0)>($pos$0|0);
   $36 = ((($mh)) + 27936|0);
   $37 = HEAP32[$36>>2]|0;
   if ($35) {
    $38 = (($31) - ($37))|0;
    $s$0$i = $38;
    break;
   } else {
    $$neg2$i = (($31) + ($pos$0))|0;
    $$neg4$i = (($$neg2$i) - ($34))|0;
    $39 = (($$neg4$i) - ($37))|0;
    $s$0$i = $39;
    break;
   }
  }
 } while(0);
 $43 = ($s$0$i|0)>(0);
 $44 = $43 ? $s$0$i : 0;
 $$0 = $44;
 return ($$0|0);
}
function _mpg123_seek($mh,$sampleoff,$whence) {
 $mh = $mh|0;
 $sampleoff = $sampleoff|0;
 $whence = $whence|0;
 var $$$i = 0, $$0 = 0, $$neg2$i = 0, $$neg4$i = 0, $$pos$0 = 0, $$pr$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $not$$i = 0, $or$cond = 0, $or$cond$i = 0, $pos$0 = 0, $s$0$i = 0, $s$0$i4 = 0, $s$0$i6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_mpg123_tell($mh)|0);
 $1 = ($0|0)<(0);
 $2 = ($whence|0)!=(0);
 $or$cond = $2 & $1;
 if ($or$cond) {
  $3 = ($mh|0)==(0|0);
  if ($3) {
   $$0 = -1;
   return ($$0|0);
  }
  $4 = ((($mh)) + 28376|0);
  HEAP32[$4>>2] = 32;
  $$0 = -1;
  return ($$0|0);
 }
 $5 = ((($mh)) + 19748|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)<(0);
 if ($7) {
  $8 = (_get_next_frame($mh)|0);
  $not$$i = ($8|0)>(-1);
  if (!($not$$i)) {
   $$0 = $8;
   return ($$0|0);
  }
 }
 L12: do {
  switch ($whence|0) {
  case 1:  {
   $9 = (($0) + ($sampleoff))|0;
   $pos$0 = $9;
   break;
  }
  case 0:  {
   $pos$0 = $sampleoff;
   break;
  }
  case 2:  {
   $10 = ((($mh)) + 19840|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = ($11|0)<(1);
   do {
    if ($12) {
     $13 = ((($mh)) + 27976|0);
     $14 = HEAP32[$13>>2]|0;
     $15 = $14 & 4;
     $16 = ($15|0)==(0);
     if (!($16)) {
      (_mpg123_scan($mh)|0);
      $$pr$pre = HEAP32[$10>>2]|0;
      $17 = ($$pr$pre|0)>(0);
      if ($17) {
       $18 = $$pr$pre;
       break;
      }
     }
     $38 = ((($mh)) + 27944|0);
     $39 = HEAP32[$38>>2]|0;
     $40 = ($39|0)>(0);
     if (!($40)) {
      $49 = ((($mh)) + 28376|0);
      HEAP32[$49>>2] = 19;
      $$0 = -1;
      return ($$0|0);
     }
     $41 = ((($mh)) + 28068|0);
     $42 = HEAP32[$41>>2]|0;
     $43 = $42 & 64;
     $44 = ($43|0)==(0);
     if ($44) {
      $s$0$i4 = $39;
     } else {
      $45 = ((($mh)) + 27936|0);
      $46 = HEAP32[$45>>2]|0;
      $47 = (($39) - ($46))|0;
      $s$0$i4 = $47;
     }
     $48 = (($s$0$i4) - ($sampleoff))|0;
     $pos$0 = $48;
     break L12;
    } else {
     $18 = $11;
    }
   } while(0);
   $19 = (_INT123_frame_outs($mh,$18)|0);
   $20 = ((($mh)) + 28068|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = $21 & 64;
   $23 = ($22|0)==(0);
   do {
    if ($23) {
     $s$0$i = $19;
    } else {
     $24 = ((($mh)) + 27944|0);
     $25 = HEAP32[$24>>2]|0;
     $26 = ($25|0)<($19|0);
     if (!($26)) {
      $34 = ((($mh)) + 27936|0);
      $35 = HEAP32[$34>>2]|0;
      $36 = (($19) - ($35))|0;
      $s$0$i = $36;
      break;
     }
     $27 = ((($mh)) + 27948|0);
     $28 = HEAP32[$27>>2]|0;
     $29 = ($28|0)>($19|0);
     $30 = ((($mh)) + 27936|0);
     $31 = HEAP32[$30>>2]|0;
     if ($29) {
      $32 = (($25) - ($31))|0;
      $s$0$i = $32;
      break;
     } else {
      $$neg2$i = (($25) + ($19))|0;
      $$neg4$i = (($$neg2$i) - ($28))|0;
      $33 = (($$neg4$i) - ($31))|0;
      $s$0$i = $33;
      break;
     }
    }
   } while(0);
   $37 = (($s$0$i) - ($sampleoff))|0;
   $pos$0 = $37;
   break;
  }
  default: {
   $50 = ((($mh)) + 28376|0);
   HEAP32[$50>>2] = 20;
   $$0 = -1;
   return ($$0|0);
  }
  }
 } while(0);
 $51 = ($pos$0|0)<(0);
 $$pos$0 = $51 ? 0 : $pos$0;
 $52 = ((($mh)) + 28068|0);
 $53 = HEAP32[$52>>2]|0;
 $54 = $53 & 64;
 $55 = ($54|0)==(0);
 if ($55) {
  $s$0$i6 = $$pos$0;
 } else {
  $56 = ((($mh)) + 27936|0);
  $57 = HEAP32[$56>>2]|0;
  $58 = (($57) + ($$pos$0))|0;
  $59 = ((($mh)) + 27944|0);
  $60 = HEAP32[$59>>2]|0;
  $61 = ($58|0)<($60|0);
  if ($61) {
   $s$0$i6 = $58;
  } else {
   $62 = ((($mh)) + 27948|0);
   $63 = HEAP32[$62>>2]|0;
   $64 = (($58) - ($60))|0;
   $65 = (($64) + ($63))|0;
   $s$0$i6 = $65;
  }
 }
 _INT123_frame_set_seek($mh,$s$0$i6);
 $66 = ((($mh)) + 27916|0);
 $67 = HEAP32[$66>>2]|0;
 $68 = ($67|0)<(0);
 $$$i = $68 ? 0 : $67;
 $69 = ((($mh)) + 27864|0);
 HEAP32[$69>>2] = 0;
 $70 = HEAP32[$5>>2]|0;
 $71 = ((($mh)) + 27908|0);
 $72 = HEAP32[$71>>2]|0;
 $73 = ($70|0)<($72|0);
 if ($73) {
  $74 = ((($mh)) + 27900|0);
  HEAP32[$74>>2] = 0;
  $75 = ($70|0)>($$$i|0);
  if (!($75)) {
   label = 29;
  }
 } else {
  label = 29;
 }
 do {
  if ((label|0) == 29) {
   $76 = ($70|0)==($$$i|0);
   if ($76) {
    $77 = ((($mh)) + 27900|0);
    $78 = HEAP32[$77>>2]|0;
    $79 = ($78|0)!=(0);
    $80 = ($$$i|0)<($72|0);
    $or$cond$i = $80 | $79;
    if ($or$cond$i) {
     break;
    }
   }
   $81 = (($$$i) + -1)|0;
   $82 = ($70|0)==($81|0);
   if ($82) {
    $83 = ((($mh)) + 27900|0);
    HEAP32[$83>>2] = 0;
    break;
   }
   (_INT123_frame_buffers_reset($mh)|0);
   $84 = ((($mh)) + 19680|0);
   $85 = HEAP32[$84>>2]|0;
   $86 = ($85|0)==(3);
   if ($86) {
    _INT123_ntom_set_ntom($mh,$$$i);
   }
   $87 = ((($mh)) + 27956|0);
   $88 = HEAP32[$87>>2]|0;
   $89 = ((($88)) + 32|0);
   $90 = HEAP32[$89>>2]|0;
   $91 = (FUNCTION_TABLE_iii[$90 & 8191]($mh,$$$i)|0);
   $92 = ($91|0)<(0);
   if ($92) {
    $$0 = $91;
    return ($$0|0);
   }
   $93 = HEAP32[$5>>2]|0;
   $94 = HEAP32[$71>>2]|0;
   $95 = ($93|0)<($94|0);
   if ($95) {
    $96 = ((($mh)) + 27900|0);
    HEAP32[$96>>2] = 0;
   }
   $97 = ((($mh)) + 19756|0);
   HEAP32[$97>>2] = $93;
  }
 } while(0);
 $98 = (_mpg123_tell($mh)|0);
 $$0 = $98;
 return ($$0|0);
}
function _mpg123_scan($mh) {
 $mh = $mh|0;
 var $$ = 0, $$0 = 0, $$lobit = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $not$ = 0, $not$$i = 0, $phitmp = 0, $track_frames$0$lcssa = 0, $track_frames$02 = 0, $track_samples$0$lcssa = 0, $track_samples$0$ph = 0, $track_samples$03 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mh|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ((($mh)) + 27976|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2 & 4;
 $4 = ($3|0)==(0);
 if ($4) {
  $5 = ((($mh)) + 28376|0);
  HEAP32[$5>>2] = 23;
  $$0 = -1;
  return ($$0|0);
 }
 $6 = ((($mh)) + 19748|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)<(0);
 if ($8) {
  $9 = (_get_next_frame($mh)|0);
  $not$$i = ($9|0)>(-1);
  if (!($not$$i)) {
   $not$ = ($9|0)!=(-12);
   $$ = $not$ << 31 >> 31;
   $$0 = $$;
   return ($$0|0);
  }
 }
 $10 = (_mpg123_tell($mh)|0);
 $11 = ((($mh)) + 27956|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($12)) + 32|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (FUNCTION_TABLE_iii[$14 & 8191]($mh,0)|0);
 $16 = ($15|0)<(0);
 if ($16) {
  $$0 = -1;
  return ($$0|0);
 }
 $17 = HEAP32[$6>>2]|0;
 $18 = ($17|0)==(0);
 if (!($18)) {
  $$0 = -1;
  return ($$0|0);
 }
 $19 = ((($mh)) + 19688|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ($20|0)==(1);
 if ($21) {
  $track_samples$0$ph = 384;
 } else {
  $22 = ($20|0)==(2);
  if ($22) {
   $track_samples$0$ph = 1152;
  } else {
   $23 = ((($mh)) + 19672|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = ($24|0)==(0);
   if ($25) {
    $26 = ((($mh)) + 19676|0);
    $27 = HEAP32[$26>>2]|0;
    $28 = ($27|0)!=(0);
    $phitmp = $28 ? 576 : 1152;
    $track_samples$0$ph = $phitmp;
   } else {
    $track_samples$0$ph = 576;
   }
  }
 }
 $29 = (_INT123_read_frame($mh)|0);
 $30 = ($29|0)==(1);
 if ($30) {
  $31 = ((($mh)) + 19672|0);
  $32 = ((($mh)) + 19676|0);
  $track_frames$02 = 1;$track_samples$03 = $track_samples$0$ph;
  while(1) {
   $33 = (($track_frames$02) + 1)|0;
   $34 = HEAP32[$19>>2]|0;
   $35 = ($34|0)==(1);
   do {
    if ($35) {
     $42 = 384;
    } else {
     $36 = ($34|0)==(2);
     if ($36) {
      $42 = 1152;
     } else {
      $37 = HEAP32[$31>>2]|0;
      $38 = ($37|0)==(0);
      if ($38) {
       $39 = HEAP32[$32>>2]|0;
       $40 = ($39|0)==(0);
       if ($40) {
        $42 = 1152;
        break;
       }
      }
      $42 = 576;
     }
    }
   } while(0);
   $41 = (($42) + ($track_samples$03))|0;
   $43 = (_INT123_read_frame($mh)|0);
   $44 = ($43|0)==(1);
   if ($44) {
    $track_frames$02 = $33;$track_samples$03 = $41;
   } else {
    $track_frames$0$lcssa = $33;$track_samples$0$lcssa = $41;
    break;
   }
  }
 } else {
  $track_frames$0$lcssa = 1;$track_samples$0$lcssa = $track_samples$0$ph;
 }
 $45 = ((($mh)) + 19840|0);
 HEAP32[$45>>2] = $track_frames$0$lcssa;
 $46 = ((($mh)) + 19844|0);
 HEAP32[$46>>2] = $track_samples$0$lcssa;
 (_mpg123_seek_frame($mh,0,$track_frames$0$lcssa)|0);
 $47 = HEAP32[$46>>2]|0;
 _INT123_frame_gapless_update($mh,$47);
 $48 = (_mpg123_seek($mh,$10,0)|0);
 $$lobit = $48 >> 31;
 $$0 = $$lobit;
 return ($$0|0);
}
function _mpg123_seek_frame($mh,$offset,$whence) {
 $mh = $mh|0;
 $offset = $offset|0;
 $whence = $whence|0;
 var $$$i = 0, $$0 = 0, $$pos$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $not$$i = 0, $not$$i1 = 0;
 var $or$cond$i = 0, $pos$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mh|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ((($mh)) + 19748|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)<(0);
 if ($3) {
  $4 = (_get_next_frame($mh)|0);
  $not$$i = ($4|0)>(-1);
  if (!($not$$i)) {
   $$0 = $4;
   return ($$0|0);
  }
 }
 L8: do {
  switch ($whence|0) {
  case 1:  {
   $5 = HEAP32[$1>>2]|0;
   $6 = (($5) + ($offset))|0;
   $pos$0 = $6;
   break;
  }
  case 0:  {
   $pos$0 = $offset;
   break;
  }
  case 2:  {
   $7 = ((($mh)) + 19840|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ($8|0)>(0);
   if ($9) {
    $10 = (($8) - ($offset))|0;
    $pos$0 = $10;
    break L8;
   }
   $11 = ((($mh)) + 28376|0);
   HEAP32[$11>>2] = 19;
   $$0 = -1;
   return ($$0|0);
   break;
  }
  default: {
   $12 = ((($mh)) + 28376|0);
   HEAP32[$12>>2] = 20;
   $$0 = -1;
   return ($$0|0);
  }
  }
 } while(0);
 $13 = ($pos$0|0)<(0);
 $$pos$0 = $13 ? 0 : $pos$0;
 _INT123_frame_set_frameseek($mh,$$pos$0);
 $14 = ((($mh)) + 27916|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)<(0);
 $$$i = $16 ? 0 : $15;
 $17 = ((($mh)) + 27864|0);
 HEAP32[$17>>2] = 0;
 $18 = HEAP32[$1>>2]|0;
 $19 = ((($mh)) + 27908|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ($18|0)<($20|0);
 if ($21) {
  $22 = ((($mh)) + 27900|0);
  HEAP32[$22>>2] = 0;
  $23 = ($18|0)>($$$i|0);
  if ($23) {
   $47 = $18;$48 = $20;
  } else {
   label = 12;
  }
 } else {
  label = 12;
 }
 do {
  if ((label|0) == 12) {
   $24 = ($18|0)==($$$i|0);
   if ($24) {
    $25 = ((($mh)) + 27900|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ($26|0)!=(0);
    $28 = ($$$i|0)<($20|0);
    $or$cond$i = $28 | $27;
    if ($or$cond$i) {
     $47 = $$$i;$48 = $20;
     break;
    }
   }
   $29 = (($$$i) + -1)|0;
   $30 = ($18|0)==($29|0);
   if ($30) {
    $31 = ((($mh)) + 27900|0);
    HEAP32[$31>>2] = 0;
    $47 = $18;$48 = $20;
    break;
   }
   (_INT123_frame_buffers_reset($mh)|0);
   $32 = ((($mh)) + 19680|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ($33|0)==(3);
   if ($34) {
    _INT123_ntom_set_ntom($mh,$$$i);
   }
   $35 = ((($mh)) + 27956|0);
   $36 = HEAP32[$35>>2]|0;
   $37 = ((($36)) + 32|0);
   $38 = HEAP32[$37>>2]|0;
   $39 = (FUNCTION_TABLE_iii[$38 & 8191]($mh,$$$i)|0);
   $40 = ($39|0)<(0);
   if ($40) {
    $$0 = $39;
    return ($$0|0);
   }
   $41 = HEAP32[$1>>2]|0;
   $42 = HEAP32[$19>>2]|0;
   $43 = ($41|0)<($42|0);
   if ($43) {
    $44 = ((($mh)) + 27900|0);
    HEAP32[$44>>2] = 0;
   }
   $45 = ((($mh)) + 19756|0);
   HEAP32[$45>>2] = $41;
   $47 = $41;$48 = $42;
  }
 } while(0);
 $46 = ($47|0)<($48|0);
 if ($46) {
  $$0 = $48;
  return ($$0|0);
 }
 $49 = ((($mh)) + 27900|0);
 $50 = HEAP32[$49>>2]|0;
 $51 = ($50|0)==(0);
 if (!($51)) {
  $$0 = $47;
  return ($$0|0);
 }
 $52 = HEAP32[$17>>2]|0;
 $not$$i1 = ($52|0)==(0);
 $53 = $not$$i1&1;
 $54 = (($53) + ($47))|0;
 $$0 = $54;
 return ($$0|0);
}
function _mpg123_length($mh) {
 $mh = $mh|0;
 var $$0 = 0, $$neg2$i = 0, $$neg4$i = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0.0, $45 = 0.0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, $length$0 = 0, $not$$i = 0, $phitmp1 = 0, $phitmp2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mh|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ((($mh)) + 19748|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)<(0);
 if ($3) {
  $4 = (_get_next_frame($mh)|0);
  $not$$i = ($4|0)>(-1);
  if (!($not$$i)) {
   $$0 = $4;
   return ($$0|0);
  }
 }
 $5 = ((($mh)) + 19844|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)>(-1);
 do {
  if ($7) {
   $length$0 = $6;
  } else {
   $8 = ((($mh)) + 19840|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = ($9|0)>(0);
   if ($10) {
    $11 = ((($mh)) + 19688|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = ($12|0)==(1);
    if ($13) {
     $22 = 384;
    } else {
     $14 = ($12|0)==(2);
     if ($14) {
      $22 = 1152;
     } else {
      $15 = ((($mh)) + 19672|0);
      $16 = HEAP32[$15>>2]|0;
      $17 = ($16|0)==(0);
      if ($17) {
       $18 = ((($mh)) + 19676|0);
       $19 = HEAP32[$18>>2]|0;
       $20 = ($19|0)!=(0);
       $phitmp1 = $20 ? 576 : 1152;
       $22 = $phitmp1;
      } else {
       $22 = 576;
      }
     }
    }
    $21 = Math_imul($22, $9)|0;
    $length$0 = $21;
    break;
   }
   $23 = ((($mh)) + 27960|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = ($24|0)>(0);
   if (!($25)) {
    $47 = ($24|0)==(0);
    if (!($47)) {
     $$0 = -1;
     return ($$0|0);
    }
    $48 = (_mpg123_tell($mh)|0);
    $$0 = $48;
    return ($$0|0);
   }
   $26 = ((($mh)) + 19848|0);
   $27 = +HEAPF64[$26>>3];
   $28 = $27 != 0.0;
   if ($28) {
    $31 = $24;$33 = $27;
   } else {
    $29 = (+_INT123_compute_bpf($mh));
    $$pre = HEAP32[$23>>2]|0;
    $31 = $$pre;$33 = $29;
   }
   $30 = (+($31|0));
   $32 = $30 / $33;
   $34 = ((($mh)) + 19688|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = ($35|0)==(1);
   if ($36) {
    $45 = 384.0;
   } else {
    $37 = ($35|0)==(2);
    if ($37) {
     $45 = 1152.0;
    } else {
     $38 = ((($mh)) + 19672|0);
     $39 = HEAP32[$38>>2]|0;
     $40 = ($39|0)==(0);
     if ($40) {
      $41 = ((($mh)) + 19676|0);
      $42 = HEAP32[$41>>2]|0;
      $43 = ($42|0)!=(0);
      $phitmp2 = $43 ? 576.0 : 1152.0;
      $45 = $phitmp2;
     } else {
      $45 = 576.0;
     }
    }
   }
   $44 = $45 * $32;
   $46 = (~~(($44)));
   $length$0 = $46;
  }
 } while(0);
 $49 = (_INT123_frame_ins2outs($mh,$length$0)|0);
 $50 = ((($mh)) + 28068|0);
 $51 = HEAP32[$50>>2]|0;
 $52 = $51 & 64;
 $53 = ($52|0)==(0);
 if ($53) {
  $$0 = $49;
  return ($$0|0);
 }
 $54 = ((($mh)) + 27944|0);
 $55 = HEAP32[$54>>2]|0;
 $56 = ($55|0)<($49|0);
 if (!($56)) {
  $64 = ((($mh)) + 27936|0);
  $65 = HEAP32[$64>>2]|0;
  $66 = (($49) - ($65))|0;
  $$0 = $66;
  return ($$0|0);
 }
 $57 = ((($mh)) + 27948|0);
 $58 = HEAP32[$57>>2]|0;
 $59 = ($58|0)>($49|0);
 $60 = ((($mh)) + 27936|0);
 $61 = HEAP32[$60>>2]|0;
 if ($59) {
  $62 = (($55) - ($61))|0;
  $$0 = $62;
  return ($$0|0);
 } else {
  $$neg2$i = (($55) + ($49))|0;
  $$neg4$i = (($$neg2$i) - ($58))|0;
  $63 = (($$neg4$i) - ($61))|0;
  $$0 = $63;
  return ($$0|0);
 }
 return (0)|0;
}
function _mpg123_delete($mh) {
 $mh = $mh|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mh|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($mh)) + 27956|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if (!($5)) {
  FUNCTION_TABLE_vi[$4 & 8191]($mh);
 }
 $6 = ((($mh)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0);
 if (!($8)) {
  $9 = ((($mh)) + 27876|0);
  _INT123_invalidate_format($9);
  HEAP32[$6>>2] = 0;
 }
 (_INT123_frame_reset($mh)|0);
 _INT123_frame_exit($mh);
 _free($mh);
 return;
}
function ___PHYSFS_sort($entries,$max,$cmpfn,$swapfn) {
 $entries = $entries|0;
 $max = $max|0;
 $cmpfn = $cmpfn|0;
 $swapfn = $swapfn|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $entries;
 $1 = $max;
 $2 = $cmpfn;
 $3 = $swapfn;
 $4 = $1;
 $5 = ($4>>>0)>(0);
 if (!($5)) {
  STACKTOP = sp;return;
 }
 $6 = $0;
 $7 = $1;
 $8 = (($7) - 1)|0;
 $9 = $2;
 $10 = $3;
 ___PHYSFS_quick_sort($6,0,$8,$9,$10);
 STACKTOP = sp;return;
}
function ___PHYSFS_quick_sort($a,$lo,$hi,$cmpfn,$swapfn) {
 $a = $a|0;
 $lo = $lo|0;
 $hi = $hi|0;
 $cmpfn = $cmpfn|0;
 $swapfn = $swapfn|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $i = 0, $j = 0, $v = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $a;
 $1 = $lo;
 $2 = $hi;
 $3 = $cmpfn;
 $4 = $swapfn;
 $5 = $2;
 $6 = $1;
 $7 = (($5) - ($6))|0;
 $8 = ($7>>>0)<=(4);
 if ($8) {
  $9 = $0;
  $10 = $1;
  $11 = $2;
  $12 = $3;
  $13 = $4;
  ___PHYSFS_bubble_sort($9,$10,$11,$12,$13);
  STACKTOP = sp;return;
 }
 $14 = $2;
 $15 = $1;
 $16 = (($14) + ($15))|0;
 $17 = (($16>>>0) / 2)&-1;
 $i = $17;
 $18 = $3;
 $19 = $0;
 $20 = $1;
 $21 = $i;
 $22 = (FUNCTION_TABLE_iiii[$18 & 8191]($19,$20,$21)|0);
 $23 = ($22|0)>(0);
 if ($23) {
  $24 = $4;
  $25 = $0;
  $26 = $1;
  $27 = $i;
  FUNCTION_TABLE_viii[$24 & 8191]($25,$26,$27);
 }
 $28 = $3;
 $29 = $0;
 $30 = $1;
 $31 = $2;
 $32 = (FUNCTION_TABLE_iiii[$28 & 8191]($29,$30,$31)|0);
 $33 = ($32|0)>(0);
 if ($33) {
  $34 = $4;
  $35 = $0;
  $36 = $1;
  $37 = $2;
  FUNCTION_TABLE_viii[$34 & 8191]($35,$36,$37);
 }
 $38 = $3;
 $39 = $0;
 $40 = $i;
 $41 = $2;
 $42 = (FUNCTION_TABLE_iiii[$38 & 8191]($39,$40,$41)|0);
 $43 = ($42|0)>(0);
 if ($43) {
  $44 = $4;
  $45 = $0;
  $46 = $i;
  $47 = $2;
  FUNCTION_TABLE_viii[$44 & 8191]($45,$46,$47);
 }
 $48 = $2;
 $49 = (($48) - 1)|0;
 $j = $49;
 $50 = $4;
 $51 = $0;
 $52 = $i;
 $53 = $j;
 FUNCTION_TABLE_viii[$50 & 8191]($51,$52,$53);
 $54 = $1;
 $i = $54;
 $55 = $j;
 $v = $55;
 while(1) {
  $56 = $3;
  $57 = $0;
  $58 = $i;
  $59 = (($58) + 1)|0;
  $i = $59;
  $60 = $v;
  $61 = (FUNCTION_TABLE_iiii[$56 & 8191]($57,$59,$60)|0);
  $62 = ($61|0)<(0);
  if ($62) {
   continue;
  }
  while(1) {
   $63 = $3;
   $64 = $0;
   $65 = $j;
   $66 = (($65) + -1)|0;
   $j = $66;
   $67 = $v;
   $68 = (FUNCTION_TABLE_iiii[$63 & 8191]($64,$66,$67)|0);
   $69 = ($68|0)>(0);
   if (!($69)) {
    break;
   }
  }
  $70 = $j;
  $71 = $i;
  $72 = ($70>>>0)<($71>>>0);
  if ($72) {
   break;
  }
  $73 = $4;
  $74 = $0;
  $75 = $i;
  $76 = $j;
  FUNCTION_TABLE_viii[$73 & 8191]($74,$75,$76);
 }
 $77 = $i;
 $78 = $2;
 $79 = (($78) - 1)|0;
 $80 = ($77|0)!=($79|0);
 if ($80) {
  $81 = $4;
  $82 = $0;
  $83 = $i;
  $84 = $2;
  $85 = (($84) - 1)|0;
  FUNCTION_TABLE_viii[$81 & 8191]($82,$83,$85);
 }
 $86 = $0;
 $87 = $1;
 $88 = $j;
 $89 = $3;
 $90 = $4;
 ___PHYSFS_quick_sort($86,$87,$88,$89,$90);
 $91 = $0;
 $92 = $i;
 $93 = (($92) + 1)|0;
 $94 = $2;
 $95 = $3;
 $96 = $4;
 ___PHYSFS_quick_sort($91,$93,$94,$95,$96);
 STACKTOP = sp;return;
}
function ___PHYSFS_bubble_sort($a,$lo,$hi,$cmpfn,$swapfn) {
 $a = $a|0;
 $lo = $lo|0;
 $hi = $hi|0;
 $cmpfn = $cmpfn|0;
 $swapfn = $swapfn|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0, $sorted = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $a;
 $1 = $lo;
 $2 = $hi;
 $3 = $cmpfn;
 $4 = $swapfn;
 while(1) {
  $sorted = 1;
  $5 = $1;
  $i = $5;
  while(1) {
   $6 = $i;
   $7 = $2;
   $8 = ($6>>>0)<($7>>>0);
   if (!($8)) {
    break;
   }
   $9 = $3;
   $10 = $0;
   $11 = $i;
   $12 = $i;
   $13 = (($12) + 1)|0;
   $14 = (FUNCTION_TABLE_iiii[$9 & 8191]($10,$11,$13)|0);
   $15 = ($14|0)>(0);
   if ($15) {
    $16 = $4;
    $17 = $0;
    $18 = $i;
    $19 = $i;
    $20 = (($19) + 1)|0;
    FUNCTION_TABLE_viii[$16 & 8191]($17,$18,$20);
    $sorted = 0;
   }
   $21 = $i;
   $22 = (($21) + 1)|0;
   $i = $22;
  }
  $23 = $sorted;
  $24 = ($23|0)!=(0);
  $25 = $24 ^ 1;
  if (!($25)) {
   break;
  }
 }
 STACKTOP = sp;return;
}
function ___PHYSFS_setError($str) {
 $str = $str|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $err = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $str;
 $1 = $0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  STACKTOP = sp;return;
 }
 $3 = (_findErrorForCurrentThread()|0);
 $err = $3;
 $4 = $err;
 $5 = ($4|0)==(0|0);
 if ($5) {
  $6 = HEAP32[(732252)>>2]|0;
  $7 = (FUNCTION_TABLE_iii[$6 & 8191](92,0)|0);
  $err = $7;
  $8 = $err;
  $9 = ($8|0)==(0|0);
  if ($9) {
   STACKTOP = sp;return;
  }
  $10 = $err;
  dest=$10; stop=dest+92|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0));
  $11 = (___PHYSFS_platformGetThreadID()|0);
  $12 = $err;
  HEAP32[$12>>2] = $11;
  $13 = HEAP32[183066]|0;
  $14 = ($13|0)!=(0|0);
  if ($14) {
   $15 = HEAP32[183066]|0;
   (___PHYSFS_platformGrabMutex($15)|0);
  }
  $16 = HEAP32[183067]|0;
  $17 = $err;
  $18 = ((($17)) + 88|0);
  HEAP32[$18>>2] = $16;
  $19 = $err;
  HEAP32[183067] = $19;
  $20 = HEAP32[183066]|0;
  $21 = ($20|0)!=(0|0);
  if ($21) {
   $22 = HEAP32[183066]|0;
   ___PHYSFS_platformReleaseMutex($22);
  }
 }
 $23 = $err;
 $24 = ((($23)) + 4|0);
 HEAP32[$24>>2] = 1;
 $25 = $err;
 $26 = ((($25)) + 8|0);
 $27 = $0;
 (_strncpy($26,$27,80)|0);
 $28 = $err;
 $29 = ((($28)) + 8|0);
 $30 = ((($29)) + 79|0);
 HEAP8[$30>>0] = 0;
 STACKTOP = sp;return;
}
function _findErrorForCurrentThread() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0, $tid = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = HEAP32[183066]|0;
 $2 = ($1|0)!=(0|0);
 if ($2) {
  $3 = HEAP32[183066]|0;
  (___PHYSFS_platformGrabMutex($3)|0);
 }
 $4 = HEAP32[183067]|0;
 $5 = ($4|0)!=(0|0);
 L4: do {
  if ($5) {
   $6 = (___PHYSFS_platformGetThreadID()|0);
   $tid = $6;
   $7 = HEAP32[183067]|0;
   $i = $7;
   while(1) {
    $8 = $i;
    $9 = ($8|0)!=(0|0);
    if (!($9)) {
     break L4;
    }
    $10 = $i;
    $11 = HEAP32[$10>>2]|0;
    $12 = $tid;
    $13 = ($11|0)==($12|0);
    if ($13) {
     break;
    }
    $18 = $i;
    $19 = ((($18)) + 88|0);
    $20 = HEAP32[$19>>2]|0;
    $i = $20;
   }
   $14 = HEAP32[183066]|0;
   $15 = ($14|0)!=(0|0);
   if ($15) {
    $16 = HEAP32[183066]|0;
    ___PHYSFS_platformReleaseMutex($16);
   }
   $17 = $i;
   $0 = $17;
   $24 = $0;
   STACKTOP = sp;return ($24|0);
  }
 } while(0);
 $21 = HEAP32[183066]|0;
 $22 = ($21|0)!=(0|0);
 if ($22) {
  $23 = HEAP32[183066]|0;
  ___PHYSFS_platformReleaseMutex($23);
 }
 $0 = 0;
 $24 = $0;
 STACKTOP = sp;return ($24|0);
}
function _PHYSFS_getLastError() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $err = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = (_findErrorForCurrentThread()|0);
 $err = $1;
 $2 = $err;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
  $4 = $err;
  $5 = ((($4)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6|0)!=(0);
  if ($7) {
   $8 = $err;
   $9 = ((($8)) + 4|0);
   HEAP32[$9>>2] = 0;
   $10 = $err;
   $11 = ((($10)) + 8|0);
   $0 = $11;
   $12 = $0;
   STACKTOP = sp;return ($12|0);
  }
 }
 $0 = 0;
 $12 = $0;
 STACKTOP = sp;return ($12|0);
}
function _PHYSFS_getLinkedVersion($ver) {
 $ver = $ver|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $ver;
 $1 = $0;
 $2 = ($1|0)!=(0|0);
 if (!($2)) {
  STACKTOP = sp;return;
 }
 $3 = $0;
 HEAP8[$3>>0] = 2;
 $4 = $0;
 $5 = ((($4)) + 1|0);
 HEAP8[$5>>0] = 0;
 $6 = $0;
 $7 = ((($6)) + 2|0);
 HEAP8[$7>>0] = 3;
 STACKTOP = sp;return;
}
function _PHYSFS_init($argv0) {
 $argv0 = $argv0|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ptr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $argv0;
 $2 = HEAP32[183068]|0;
 $3 = ($2|0)!=(0);
 do {
  if ($3) {
   ___PHYSFS_setError(516522);
   $0 = 0;
  } else {
   $4 = HEAP32[183069]|0;
   $5 = ($4|0)!=(0);
   if (!($5)) {
    _setDefaultAllocator();
   }
   $6 = HEAP32[183061]|0;
   $7 = ($6|0)!=(0|0);
   if ($7) {
    $8 = HEAP32[183061]|0;
    $9 = (FUNCTION_TABLE_i[$8 & 8191]()|0);
    $10 = ($9|0)!=(0);
    if (!($10)) {
     ___PHYSFS_setError(0);
     $0 = 0;
     break;
    }
   }
   $11 = (___PHYSFS_platformInit()|0);
   $12 = ($11|0)!=(0);
   if (!($12)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $13 = (_initializeMutexes()|0);
   $14 = ($13|0)!=(0);
   if (!($14)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $15 = $1;
   $16 = (_calculateBaseDir($15)|0);
   HEAP32[183071] = $16;
   $17 = HEAP32[183071]|0;
   $18 = ($17|0)==(0|0);
   if ($18) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $19 = HEAP32[183071]|0;
   $20 = (___PHYSFS_platformRealPath($19)|0);
   $ptr = $20;
   $21 = HEAP32[(732260)>>2]|0;
   $22 = HEAP32[183071]|0;
   FUNCTION_TABLE_vi[$21 & 8191]($22);
   $23 = $ptr;
   $24 = ($23|0)==(0|0);
   if ($24) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $25 = $ptr;
   HEAP32[183071] = $25;
   $26 = (_appendDirSep(732284)|0);
   $27 = ($26|0)!=(0);
   if (!($27)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $28 = (_calculateUserDir()|0);
   HEAP32[183072] = $28;
   $29 = HEAP32[183072]|0;
   $30 = ($29|0)==(0|0);
   if (!($30)) {
    $31 = (_appendDirSep(732288)|0);
    $32 = ($31|0)!=(0);
    if ($32) {
     HEAP32[183068] = 1;
     $35 = (_PHYSFS_getLastError()|0);
     ___PHYSFS_setError($35);
     $0 = 1;
     break;
    }
   }
   $33 = HEAP32[(732260)>>2]|0;
   $34 = HEAP32[183071]|0;
   FUNCTION_TABLE_vi[$33 & 8191]($34);
   HEAP32[183071] = 0;
   $0 = 0;
  }
 } while(0);
 $36 = $0;
 STACKTOP = sp;return ($36|0);
}
function _setDefaultAllocator() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[183069]|0;
 $1 = ($0|0)!=(0);
 if ($1) {
  ___assert_fail((516542|0),(516561|0),2178,(516625|0));
  // unreachable;
 }
 $2 = (___PHYSFS_platformSetDefaultAllocator(732244)|0);
 $3 = ($2|0)!=(0);
 if ($3) {
  return;
 }
 HEAP32[183061] = 0;
 HEAP32[(732248)>>2] = 0;
 HEAP32[(732252)>>2] = 4278;
 HEAP32[(732256)>>2] = 4279;
 HEAP32[(732260)>>2] = 4280;
 return;
}
function _mallocAllocatorMalloc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $3 = sp;
 $4 = $3;
 $5 = $4;
 HEAP32[$5>>2] = $0;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = $1;
 $8 = $3;
 $9 = $8;
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8) + 4)|0;
 $12 = $11;
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13>>>0)>(0);
 $15 = ($10>>>0)>(4294967295);
 $16 = ($13|0)==(0);
 $17 = $16 & $15;
 $18 = $14 | $17;
 if ($18) {
  ___PHYSFS_setError(596704);
  $2 = 0;
  $26 = $2;
  STACKTOP = sp;return ($26|0);
 } else {
  $19 = $3;
  $20 = $19;
  $21 = HEAP32[$20>>2]|0;
  $22 = (($19) + 4)|0;
  $23 = $22;
  $24 = HEAP32[$23>>2]|0;
  $25 = (_malloc($21)|0);
  $2 = $25;
  $26 = $2;
  STACKTOP = sp;return ($26|0);
 }
 return (0)|0;
}
function _mallocAllocatorRealloc($ptr,$0,$1) {
 $ptr = $ptr|0;
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $4 = sp;
 $3 = $ptr;
 $5 = $4;
 $6 = $5;
 HEAP32[$6>>2] = $0;
 $7 = (($5) + 4)|0;
 $8 = $7;
 HEAP32[$8>>2] = $1;
 $9 = $4;
 $10 = $9;
 $11 = HEAP32[$10>>2]|0;
 $12 = (($9) + 4)|0;
 $13 = $12;
 $14 = HEAP32[$13>>2]|0;
 $15 = ($14>>>0)>(0);
 $16 = ($11>>>0)>(4294967295);
 $17 = ($14|0)==(0);
 $18 = $17 & $16;
 $19 = $15 | $18;
 if ($19) {
  ___PHYSFS_setError(596704);
  $2 = 0;
  $28 = $2;
  STACKTOP = sp;return ($28|0);
 } else {
  $20 = $3;
  $21 = $4;
  $22 = $21;
  $23 = HEAP32[$22>>2]|0;
  $24 = (($21) + 4)|0;
  $25 = $24;
  $26 = HEAP32[$25>>2]|0;
  $27 = (_realloc($20,$23)|0);
  $2 = $27;
  $28 = $2;
  STACKTOP = sp;return ($28|0);
 }
 return (0)|0;
}
function _mallocAllocatorFree($ptr) {
 $ptr = $ptr|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $ptr;
 $1 = $0;
 _free($1);
 STACKTOP = sp;return;
}
function _initializeMutexes() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = (___PHYSFS_platformCreateMutex()|0);
 HEAP32[183066] = $1;
 $2 = HEAP32[183066]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
  $4 = (___PHYSFS_platformCreateMutex()|0);
  HEAP32[183070] = $4;
  $5 = HEAP32[183070]|0;
  $6 = ($5|0)==(0|0);
  if (!($6)) {
   $0 = 1;
   $13 = $0;
   STACKTOP = sp;return ($13|0);
  }
 }
 $7 = HEAP32[183066]|0;
 $8 = ($7|0)!=(0|0);
 if ($8) {
  $9 = HEAP32[183066]|0;
  ___PHYSFS_platformDestroyMutex($9);
 }
 $10 = HEAP32[183070]|0;
 $11 = ($10|0)!=(0|0);
 if ($11) {
  $12 = HEAP32[183070]|0;
  ___PHYSFS_platformDestroyMutex($12);
 }
 HEAP32[183070] = 0;
 HEAP32[183066] = 0;
 $0 = 0;
 $13 = $0;
 STACKTOP = sp;return ($13|0);
}
function _calculateBaseDir($argv0) {
 $argv0 = $argv0|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dirsep = 0, $p = 0, $ptr = 0, $retval = 0, $size = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $argv0;
 $retval = 0;
 $dirsep = 0;
 $ptr = 0;
 $2 = $1;
 $3 = (___PHYSFS_platformCalcBaseDir($2)|0);
 $retval = $3;
 $4 = $retval;
 $5 = ($4|0)!=(0|0);
 if ($5) {
  $6 = $retval;
  $0 = $6;
  $49 = $0;
  STACKTOP = sp;return ($49|0);
 }
 $7 = $1;
 $8 = ($7|0)==(0|0);
 if ($8) {
  ___PHYSFS_setError(516645);
  $0 = 0;
  $49 = $0;
  STACKTOP = sp;return ($49|0);
 }
 $9 = (_PHYSFS_getDirSeparator()|0);
 $dirsep = $9;
 $10 = $dirsep;
 $11 = (_strlen($10)|0);
 $12 = ($11|0)==(1);
 $13 = $1;
 $14 = $dirsep;
 L9: do {
  if ($12) {
   $15 = HEAP8[$14>>0]|0;
   $16 = $15 << 24 >> 24;
   $17 = (_strrchr($13,$16)|0);
   $ptr = $17;
  } else {
   $18 = (_strstr($13,$14)|0);
   $ptr = $18;
   $19 = $ptr;
   $20 = ($19|0)!=(0|0);
   if ($20) {
    $21 = $ptr;
    $p = $21;
    while(1) {
     $22 = $p;
     $23 = ($22|0)!=(0|0);
     if (!($23)) {
      break L9;
     }
     $24 = $p;
     $ptr = $24;
     $25 = $p;
     $26 = ((($25)) + 1|0);
     $27 = $dirsep;
     $28 = (_strstr($26,$27)|0);
     $p = $28;
    }
   }
  }
 } while(0);
 $29 = $ptr;
 $30 = ($29|0)!=(0|0);
 if (!($30)) {
  ___PHYSFS_setError(517147);
  $0 = 0;
  $49 = $0;
  STACKTOP = sp;return ($49|0);
 }
 $31 = $ptr;
 $32 = $1;
 $33 = $31;
 $34 = $32;
 $35 = (($33) - ($34))|0;
 $size = $35;
 $36 = HEAP32[(732252)>>2]|0;
 $37 = $size;
 $38 = (($37) + 1)|0;
 $39 = (FUNCTION_TABLE_iii[$36 & 8191]($38,0)|0);
 $retval = $39;
 $40 = $retval;
 $41 = ($40|0)==(0|0);
 if ($41) {
  ___PHYSFS_setError(596704);
  $0 = 0;
  $49 = $0;
  STACKTOP = sp;return ($49|0);
 } else {
  $42 = $retval;
  $43 = $1;
  $44 = $size;
  _memcpy(($42|0),($43|0),($44|0))|0;
  $45 = $size;
  $46 = $retval;
  $47 = (($46) + ($45)|0);
  HEAP8[$47>>0] = 0;
  $48 = $retval;
  $0 = $48;
  $49 = $0;
  STACKTOP = sp;return ($49|0);
 }
 return (0)|0;
}
function _PHYSFS_getDirSeparator() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[9660]|0;
 return ($0|0);
}
function _appendDirSep($dir) {
 $dir = $dir|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dirsep = 0, $ptr = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $dir;
 $2 = (_PHYSFS_getDirSeparator()|0);
 $dirsep = $2;
 $3 = $1;
 $4 = HEAP32[$3>>2]|0;
 $5 = $1;
 $6 = HEAP32[$5>>2]|0;
 $7 = (_strlen($6)|0);
 $8 = (($4) + ($7)|0);
 $9 = $dirsep;
 $10 = (_strlen($9)|0);
 $11 = (0 - ($10))|0;
 $12 = (($8) + ($11)|0);
 $13 = $dirsep;
 $14 = (_strcmp($12,$13)|0);
 $15 = ($14|0)==(0);
 if ($15) {
  $0 = 1;
  $36 = $0;
  STACKTOP = sp;return ($36|0);
 }
 $16 = HEAP32[(732256)>>2]|0;
 $17 = $1;
 $18 = HEAP32[$17>>2]|0;
 $19 = $1;
 $20 = HEAP32[$19>>2]|0;
 $21 = (_strlen($20)|0);
 $22 = $dirsep;
 $23 = (_strlen($22)|0);
 $24 = (($21) + ($23))|0;
 $25 = (($24) + 1)|0;
 $26 = (FUNCTION_TABLE_iiii[$16 & 8191]($18,$25,0)|0);
 $ptr = $26;
 $27 = $ptr;
 $28 = ($27|0)!=(0|0);
 if ($28) {
  $32 = $ptr;
  $33 = $dirsep;
  (_strcat($32,$33)|0);
  $34 = $ptr;
  $35 = $1;
  HEAP32[$35>>2] = $34;
  $0 = 1;
  $36 = $0;
  STACKTOP = sp;return ($36|0);
 } else {
  $29 = HEAP32[(732260)>>2]|0;
  $30 = $1;
  $31 = HEAP32[$30>>2]|0;
  FUNCTION_TABLE_vi[$29 & 8191]($31);
  $0 = 0;
  $36 = $0;
  STACKTOP = sp;return ($36|0);
 }
 return (0)|0;
}
function _calculateUserDir() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dirsep = 0, $ptr = 0, $retval = 0, $str = 0;
 var $uname = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (___PHYSFS_platformGetUserDir()|0);
 $retval = $0;
 $1 = $retval;
 $2 = ($1|0)!=(0|0);
 if ($2) {
  $3 = $retval;
  $4 = (___PHYSFS_platformRealPath($3)|0);
  $ptr = $4;
  $5 = HEAP32[(732260)>>2]|0;
  $6 = $retval;
  FUNCTION_TABLE_vi[$5 & 8191]($6);
  $7 = $ptr;
  $retval = $7;
 }
 $8 = $retval;
 $9 = ($8|0)==(0|0);
 if (!($9)) {
  $35 = $retval;
  STACKTOP = sp;return ($35|0);
 }
 $10 = (_PHYSFS_getDirSeparator()|0);
 $dirsep = $10;
 $11 = (___PHYSFS_platformGetUserName()|0);
 $uname = $11;
 $12 = $uname;
 $13 = ($12|0)!=(0|0);
 $14 = $uname;
 $15 = $13 ? $14 : 604635;
 $str = $15;
 $16 = HEAP32[(732252)>>2]|0;
 $17 = HEAP32[183071]|0;
 $18 = (_strlen($17)|0);
 $19 = $str;
 $20 = (_strlen($19)|0);
 $21 = (($18) + ($20))|0;
 $22 = $dirsep;
 $23 = (_strlen($22)|0);
 $24 = (($21) + ($23))|0;
 $25 = (($24) + 6)|0;
 $26 = (FUNCTION_TABLE_iii[$16 & 8191]($25,0)|0);
 $retval = $26;
 $27 = $retval;
 $28 = ($27|0)==(0|0);
 if ($28) {
  ___PHYSFS_setError(596704);
 } else {
  $29 = $retval;
  $30 = HEAP32[183071]|0;
  $31 = $dirsep;
  $32 = $str;
  HEAP32[$vararg_buffer>>2] = $30;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $31;
  $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
  HEAP32[$vararg_ptr2>>2] = $32;
  (_sprintf($29,516659,$vararg_buffer)|0);
 }
 $33 = HEAP32[(732260)>>2]|0;
 $34 = $uname;
 FUNCTION_TABLE_vi[$33 & 8191]($34);
 $35 = $retval;
 STACKTOP = sp;return ($35|0);
}
function _PHYSFS_deinit() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = HEAP32[183068]|0;
 $2 = ($1|0)!=(0);
 if (!($2)) {
  ___PHYSFS_setError(516671);
  $0 = 0;
  $20 = $0;
  STACKTOP = sp;return ($20|0);
 }
 $3 = (___PHYSFS_platformDeinit()|0);
 $4 = ($3|0)!=(0);
 if (!($4)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $20 = $0;
  STACKTOP = sp;return ($20|0);
 }
 (_closeFileHandleList(732292)|0);
 $5 = (_PHYSFS_setWriteDir(0)|0);
 $6 = ($5|0)!=(0);
 if (!($6)) {
  ___PHYSFS_setError(516687);
  $0 = 0;
  $20 = $0;
  STACKTOP = sp;return ($20|0);
 }
 _freeSearchPath();
 _freeErrorMessages();
 $7 = HEAP32[183071]|0;
 $8 = ($7|0)!=(0|0);
 if ($8) {
  $9 = HEAP32[(732260)>>2]|0;
  $10 = HEAP32[183071]|0;
  FUNCTION_TABLE_vi[$9 & 8191]($10);
  HEAP32[183071] = 0;
 }
 $11 = HEAP32[183072]|0;
 $12 = ($11|0)!=(0|0);
 if ($12) {
  $13 = HEAP32[(732260)>>2]|0;
  $14 = HEAP32[183072]|0;
  FUNCTION_TABLE_vi[$13 & 8191]($14);
  HEAP32[183072] = 0;
 }
 HEAP32[183077] = 0;
 HEAP32[183068] = 0;
 $15 = HEAP32[183066]|0;
 ___PHYSFS_platformDestroyMutex($15);
 $16 = HEAP32[183070]|0;
 ___PHYSFS_platformDestroyMutex($16);
 $17 = HEAP32[(732248)>>2]|0;
 $18 = ($17|0)!=(0|0);
 if ($18) {
  $19 = HEAP32[(732248)>>2]|0;
  FUNCTION_TABLE_v[$19 & 8191]();
 }
 HEAP32[183070] = 0;
 HEAP32[183066] = 0;
 $0 = 1;
 $20 = $0;
 STACKTOP = sp;return ($20|0);
}
function _closeFileHandleList($list) {
 $list = $list|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0, $next = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $list;
 $next = 0;
 $2 = $1;
 $3 = HEAP32[$2>>2]|0;
 $i = $3;
 while(1) {
  $4 = $i;
  $5 = ($4|0)!=(0|0);
  if (!($5)) {
   label = 6;
   break;
  }
  $6 = $i;
  $7 = ((($6)) + 32|0);
  $8 = HEAP32[$7>>2]|0;
  $next = $8;
  $9 = $i;
  $10 = ((($9)) + 12|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ((($11)) + 80|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = $i;
  $15 = HEAP32[$14>>2]|0;
  $16 = (FUNCTION_TABLE_ii[$13 & 8191]($15)|0);
  $17 = ($16|0)!=(0);
  if (!($17)) {
   label = 4;
   break;
  }
  $20 = HEAP32[(732260)>>2]|0;
  $21 = $i;
  FUNCTION_TABLE_vi[$20 & 8191]($21);
  $22 = $next;
  $i = $22;
 }
 if ((label|0) == 4) {
  $18 = $i;
  $19 = $1;
  HEAP32[$19>>2] = $18;
  $0 = 0;
  $24 = $0;
  STACKTOP = sp;return ($24|0);
 }
 else if ((label|0) == 6) {
  $23 = $1;
  HEAP32[$23>>2] = 0;
  $0 = 1;
  $24 = $0;
  STACKTOP = sp;return ($24|0);
 }
 return (0)|0;
}
function _PHYSFS_setWriteDir($newDir) {
 $newDir = $newDir|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $newDir;
 $retval = 1;
 $2 = HEAP32[183070]|0;
 (___PHYSFS_platformGrabMutex($2)|0);
 $3 = HEAP32[183074]|0;
 $4 = ($3|0)!=(0|0);
 do {
  if ($4) {
   $5 = HEAP32[183074]|0;
   $6 = HEAP32[183073]|0;
   $7 = (_freeDirHandle($5,$6)|0);
   $8 = ($7|0)!=(0);
   if ($8) {
    HEAP32[183074] = 0;
    break;
   }
   ___PHYSFS_setError(0);
   $9 = HEAP32[183070]|0;
   ___PHYSFS_platformReleaseMutex($9);
   $0 = 0;
   $19 = $0;
   STACKTOP = sp;return ($19|0);
  }
 } while(0);
 $10 = $1;
 $11 = ($10|0)!=(0|0);
 if ($11) {
  $12 = $1;
  $13 = (_createDirHandle($12,0,1)|0);
  HEAP32[183074] = $13;
  $14 = HEAP32[183074]|0;
  $15 = ($14|0)!=(0|0);
  $16 = $15&1;
  $retval = $16;
 }
 $17 = HEAP32[183070]|0;
 ___PHYSFS_platformReleaseMutex($17);
 $18 = $retval;
 $0 = $18;
 $19 = $0;
 STACKTOP = sp;return ($19|0);
}
function _freeDirHandle($dh,$openList) {
 $dh = $dh|0;
 $openList = $openList|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $dh;
 $2 = $openList;
 $3 = $1;
 $4 = ($3|0)==(0|0);
 if ($4) {
  $0 = 1;
  $33 = $0;
  STACKTOP = sp;return ($33|0);
 }
 $5 = $2;
 $i = $5;
 while(1) {
  $6 = $i;
  $7 = ($6|0)!=(0|0);
  if (!($7)) {
   label = 8;
   break;
  }
  $8 = $i;
  $9 = ((($8)) + 8|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = $1;
  $12 = ($10|0)==($11|0);
  if ($12) {
   label = 6;
   break;
  }
  $13 = $i;
  $14 = ((($13)) + 32|0);
  $15 = HEAP32[$14>>2]|0;
  $i = $15;
 }
 if ((label|0) == 6) {
  ___PHYSFS_setError(516687);
  $0 = 0;
  $33 = $0;
  STACKTOP = sp;return ($33|0);
 }
 else if ((label|0) == 8) {
  $16 = $1;
  $17 = ((($16)) + 12|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = ((($18)) + 52|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = $1;
  $22 = HEAP32[$21>>2]|0;
  FUNCTION_TABLE_vi[$20 & 8191]($22);
  $23 = HEAP32[(732260)>>2]|0;
  $24 = $1;
  $25 = ((($24)) + 4|0);
  $26 = HEAP32[$25>>2]|0;
  FUNCTION_TABLE_vi[$23 & 8191]($26);
  $27 = HEAP32[(732260)>>2]|0;
  $28 = $1;
  $29 = ((($28)) + 8|0);
  $30 = HEAP32[$29>>2]|0;
  FUNCTION_TABLE_vi[$27 & 8191]($30);
  $31 = HEAP32[(732260)>>2]|0;
  $32 = $1;
  FUNCTION_TABLE_vi[$31 & 8191]($32);
  $0 = 1;
  $33 = $0;
  STACKTOP = sp;return ($33|0);
 }
 return (0)|0;
}
function _createDirHandle($newDir,$mountPoint,$forWriting) {
 $newDir = $newDir|0;
 $mountPoint = $mountPoint|0;
 $forWriting = $forWriting|0;
 var $$alloca_mul = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $dirHandle = 0, $len = 0, $tmpmntpnt = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $newDir;
 $2 = $mountPoint;
 $3 = $forWriting;
 $dirHandle = 0;
 $tmpmntpnt = 0;
 $4 = $1;
 $5 = ($4|0)!=(0|0);
 L1: do {
  if ($5) {
   $6 = $2;
   $7 = ($6|0)!=(0|0);
   if ($7) {
    $8 = $2;
    $9 = (_strlen($8)|0);
    $10 = (($9) + 1)|0;
    $len = $10;
    $11 = $len;
    $12 = ($11>>>0)<(128);
    if ($12) {
     $13 = $len;
     $14 = (($13) + 1)|0;
     $$alloca_mul = $14;
     $15 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
     $17 = $15;
    } else {
     $17 = 0;
    }
    $16 = $len;
    $18 = (___PHYSFS_initSmallAlloc($17,$16,0)|0);
    $tmpmntpnt = $18;
    $19 = $tmpmntpnt;
    $20 = ($19|0)!=(0|0);
    if (!($20)) {
     ___PHYSFS_setError(596704);
     break;
    }
    $21 = $2;
    $22 = $tmpmntpnt;
    $23 = (_sanitizePlatformIndependentPath($21,$22)|0);
    $24 = ($23|0)!=(0);
    if (!($24)) {
     break;
    }
    $25 = $tmpmntpnt;
    $2 = $25;
   }
   $26 = $1;
   $27 = $3;
   $28 = (_openDirectory($26,$27)|0);
   $dirHandle = $28;
   $29 = $dirHandle;
   $30 = ($29|0)!=(0|0);
   if (!($30)) {
    ___PHYSFS_setError(0);
    break;
   }
   $31 = HEAP32[(732252)>>2]|0;
   $32 = $1;
   $33 = (_strlen($32)|0);
   $34 = (($33) + 1)|0;
   $35 = (FUNCTION_TABLE_iii[$31 & 8191]($34,0)|0);
   $36 = $dirHandle;
   $37 = ((($36)) + 4|0);
   HEAP32[$37>>2] = $35;
   $38 = $dirHandle;
   $39 = ((($38)) + 4|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = ($40|0)!=(0|0);
   if (!($41)) {
    ___PHYSFS_setError(596704);
    break;
   }
   $42 = $dirHandle;
   $43 = ((($42)) + 4|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = $1;
   (_strcpy($44,$45)|0);
   $46 = $2;
   $47 = ($46|0)!=(0|0);
   do {
    if ($47) {
     $48 = $2;
     $49 = HEAP8[$48>>0]|0;
     $50 = $49 << 24 >> 24;
     $51 = ($50|0)!=(0);
     if ($51) {
      $52 = HEAP32[(732252)>>2]|0;
      $53 = $2;
      $54 = (_strlen($53)|0);
      $55 = (($54) + 2)|0;
      $56 = (FUNCTION_TABLE_iii[$52 & 8191]($55,0)|0);
      $57 = $dirHandle;
      $58 = ((($57)) + 8|0);
      HEAP32[$58>>2] = $56;
      $59 = $dirHandle;
      $60 = ((($59)) + 8|0);
      $61 = HEAP32[$60>>2]|0;
      $62 = ($61|0)!=(0|0);
      if ($62) {
       $63 = $dirHandle;
       $64 = ((($63)) + 8|0);
       $65 = HEAP32[$64>>2]|0;
       $66 = $2;
       (_strcpy($65,$66)|0);
       $67 = $dirHandle;
       $68 = ((($67)) + 8|0);
       $69 = HEAP32[$68>>2]|0;
       (_strcat($69,523303)|0);
       break;
      } else {
       ___PHYSFS_setError(596704);
       break L1;
      }
     }
    }
   } while(0);
   $70 = $tmpmntpnt;
   ___PHYSFS_smallFree($70);
   $71 = $dirHandle;
   $0 = $71;
   $92 = $0;
   STACKTOP = sp;return ($92|0);
  } else {
   ___PHYSFS_setError(517147);
  }
 } while(0);
 $72 = $dirHandle;
 $73 = ($72|0)!=(0|0);
 if ($73) {
  $74 = $dirHandle;
  $75 = ((($74)) + 12|0);
  $76 = HEAP32[$75>>2]|0;
  $77 = ((($76)) + 52|0);
  $78 = HEAP32[$77>>2]|0;
  $79 = $dirHandle;
  $80 = HEAP32[$79>>2]|0;
  FUNCTION_TABLE_vi[$78 & 8191]($80);
  $81 = HEAP32[(732260)>>2]|0;
  $82 = $dirHandle;
  $83 = ((($82)) + 4|0);
  $84 = HEAP32[$83>>2]|0;
  FUNCTION_TABLE_vi[$81 & 8191]($84);
  $85 = HEAP32[(732260)>>2]|0;
  $86 = $dirHandle;
  $87 = ((($86)) + 8|0);
  $88 = HEAP32[$87>>2]|0;
  FUNCTION_TABLE_vi[$85 & 8191]($88);
  $89 = HEAP32[(732260)>>2]|0;
  $90 = $dirHandle;
  FUNCTION_TABLE_vi[$89 & 8191]($90);
 }
 $91 = $tmpmntpnt;
 ___PHYSFS_smallFree($91);
 $0 = 0;
 $92 = $0;
 STACKTOP = sp;return ($92|0);
}
function ___PHYSFS_initSmallAlloc($ptr,$0,$1) {
 $ptr = $ptr|0;
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $retval = 0, $useHeap = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $4 = sp;
 $3 = $ptr;
 $5 = $4;
 $6 = $5;
 HEAP32[$6>>2] = $0;
 $7 = (($5) + 4)|0;
 $8 = $7;
 HEAP32[$8>>2] = $1;
 $9 = $3;
 $10 = ($9|0)==(0|0);
 $11 = $10 ? 1 : 0;
 $12 = $11&255;
 $useHeap = $12;
 $13 = $useHeap;
 $14 = ($13<<24>>24)!=(0);
 if ($14) {
  $15 = HEAP32[(732252)>>2]|0;
  $16 = $4;
  $17 = $16;
  $18 = HEAP32[$17>>2]|0;
  $19 = (($16) + 4)|0;
  $20 = $19;
  $21 = HEAP32[$20>>2]|0;
  $22 = (_i64Add(($18|0),($21|0),1,0)|0);
  $23 = tempRet0;
  $24 = (FUNCTION_TABLE_iii[$15 & 8191]($22,$23)|0);
  $3 = $24;
 }
 $25 = $3;
 $26 = ($25|0)!=(0|0);
 if ($26) {
  $27 = $3;
  $retval = $27;
  $28 = $useHeap;
  $29 = $retval;
  HEAP8[$29>>0] = $28;
  $30 = $retval;
  $31 = ((($30)) + 1|0);
  $2 = $31;
  $32 = $2;
  STACKTOP = sp;return ($32|0);
 } else {
  $2 = 0;
  $32 = $2;
  STACKTOP = sp;return ($32|0);
 }
 return (0)|0;
}
function _sanitizePlatformIndependentPath($src,$dst) {
 $src = $src|0;
 $dst = $dst|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ch = 0, $prev = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $src;
 $2 = $dst;
 while(1) {
  $3 = $1;
  $4 = HEAP8[$3>>0]|0;
  $5 = $4 << 24 >> 24;
  $6 = ($5|0)==(47);
  if (!($6)) {
   break;
  }
  $7 = $1;
  $8 = ((($7)) + 1|0);
  $1 = $8;
 }
 $9 = $2;
 $prev = $9;
 while(1) {
  $10 = $1;
  $11 = ((($10)) + 1|0);
  $1 = $11;
  $12 = HEAP8[$10>>0]|0;
  $ch = $12;
  $13 = $ch;
  $14 = $13 << 24 >> 24;
  $15 = ($14|0)==(58);
  if ($15) {
   label = 7;
   break;
  }
  $16 = $ch;
  $17 = $16 << 24 >> 24;
  $18 = ($17|0)==(92);
  if ($18) {
   label = 7;
   break;
  }
  $19 = $ch;
  $20 = $19 << 24 >> 24;
  $21 = ($20|0)==(47);
  if ($21) {
   $22 = $2;
   HEAP8[$22>>0] = 0;
   $23 = $prev;
   $24 = (_strcmp($23,619449)|0);
   $25 = ($24|0)==(0);
   if ($25) {
    label = 11;
    break;
   }
   $26 = $prev;
   $27 = (_strcmp($26,519838)|0);
   $28 = ($27|0)==(0);
   if ($28) {
    label = 11;
    break;
   }
   while(1) {
    $29 = $1;
    $30 = HEAP8[$29>>0]|0;
    $31 = $30 << 24 >> 24;
    $32 = ($31|0)==(47);
    $33 = $1;
    if (!($32)) {
     break;
    }
    $34 = ((($33)) + 1|0);
    $1 = $34;
   }
   $35 = HEAP8[$33>>0]|0;
   $36 = $35 << 24 >> 24;
   $37 = ($36|0)==(0);
   if ($37) {
    label = 17;
    break;
   }
   $38 = $2;
   $39 = ((($38)) + 1|0);
   $prev = $39;
  }
  $40 = $ch;
  $41 = $2;
  $42 = ((($41)) + 1|0);
  $2 = $42;
  HEAP8[$41>>0] = $40;
  $43 = $ch;
  $44 = $43 << 24 >> 24;
  $45 = ($44|0)!=(0);
  if (!($45)) {
   label = 17;
   break;
  }
 }
 if ((label|0) == 7) {
  ___PHYSFS_setError(516704);
  $0 = 0;
  $46 = $0;
  STACKTOP = sp;return ($46|0);
 }
 else if ((label|0) == 11) {
  ___PHYSFS_setError(516704);
  $0 = 0;
  $46 = $0;
  STACKTOP = sp;return ($46|0);
 }
 else if ((label|0) == 17) {
  $0 = 1;
  $46 = $0;
  STACKTOP = sp;return ($46|0);
 }
 return (0)|0;
}
function _openDirectory($d,$forWriting) {
 $d = $d|0;
 $forWriting = $forWriting|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, $ext = 0, $i = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $d;
 $2 = $forWriting;
 $retval = 0;
 $3 = $1;
 $4 = (___PHYSFS_platformExists($3)|0);
 $5 = ($4|0)!=(0);
 if (!($5)) {
  ___PHYSFS_setError(517404);
  $0 = 0;
  $66 = $0;
  STACKTOP = sp;return ($66|0);
 }
 $6 = $1;
 $7 = (_find_filename_extension($6)|0);
 $ext = $7;
 $8 = $ext;
 $9 = ($8|0)!=(0|0);
 $i = 38628;
 L5: do {
  if ($9) {
   while(1) {
    $10 = $i;
    $11 = HEAP32[$10>>2]|0;
    $12 = ($11|0)!=(0|0);
    $13 = $retval;
    $14 = ($13|0)==(0|0);
    $15 = $12 ? $14 : 0;
    if (!($15)) {
     break;
    }
    $16 = $ext;
    $17 = $i;
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = HEAP32[$19>>2]|0;
    $21 = (___PHYSFS_stricmpASCII($16,$20)|0);
    $22 = ($21|0)==(0);
    if ($22) {
     $23 = $i;
     $24 = HEAP32[$23>>2]|0;
     $25 = $1;
     $26 = $2;
     $27 = (_tryOpenDir($24,$25,$26)|0);
     $retval = $27;
    }
    $28 = $i;
    $29 = ((($28)) + 4|0);
    $i = $29;
   }
   $i = 38628;
   while(1) {
    $30 = $i;
    $31 = HEAP32[$30>>2]|0;
    $32 = ($31|0)!=(0|0);
    $33 = $retval;
    $34 = ($33|0)==(0|0);
    $35 = $32 ? $34 : 0;
    if (!($35)) {
     break L5;
    }
    $36 = $ext;
    $37 = $i;
    $38 = HEAP32[$37>>2]|0;
    $39 = HEAP32[$38>>2]|0;
    $40 = HEAP32[$39>>2]|0;
    $41 = (___PHYSFS_stricmpASCII($36,$40)|0);
    $42 = ($41|0)!=(0);
    if ($42) {
     $43 = $i;
     $44 = HEAP32[$43>>2]|0;
     $45 = $1;
     $46 = $2;
     $47 = (_tryOpenDir($44,$45,$46)|0);
     $retval = $47;
    }
    $48 = $i;
    $49 = ((($48)) + 4|0);
    $i = $49;
   }
  } else {
   while(1) {
    $50 = $i;
    $51 = HEAP32[$50>>2]|0;
    $52 = ($51|0)!=(0|0);
    $53 = $retval;
    $54 = ($53|0)==(0|0);
    $55 = $52 ? $54 : 0;
    if (!($55)) {
     break L5;
    }
    $56 = $i;
    $57 = HEAP32[$56>>2]|0;
    $58 = $1;
    $59 = $2;
    $60 = (_tryOpenDir($57,$58,$59)|0);
    $retval = $60;
    $61 = $i;
    $62 = ((($61)) + 4|0);
    $i = $62;
   }
  }
 } while(0);
 $63 = $retval;
 $64 = ($63|0)==(0|0);
 if ($64) {
  ___PHYSFS_setError(517361);
  $0 = 0;
  $66 = $0;
  STACKTOP = sp;return ($66|0);
 } else {
  $65 = $retval;
  $0 = $65;
  $66 = $0;
  STACKTOP = sp;return ($66|0);
 }
 return (0)|0;
}
function _find_filename_extension($fname) {
 $fname = $fname|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p = 0, $retval = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $fname;
 $1 = $0;
 $2 = (_strchr($1,46)|0);
 $retval = $2;
 $3 = $retval;
 $p = $3;
 while(1) {
  $4 = $p;
  $5 = ($4|0)!=(0|0);
  if (!($5)) {
   break;
  }
  $6 = $p;
  $7 = ((($6)) + 1|0);
  $8 = (_strchr($7,46)|0);
  $p = $8;
  $9 = $p;
  $10 = ($9|0)!=(0|0);
  if (!($10)) {
   continue;
  }
  $11 = $p;
  $retval = $11;
 }
 $12 = $retval;
 $13 = ($12|0)!=(0|0);
 if (!($13)) {
  $16 = $retval;
  STACKTOP = sp;return ($16|0);
 }
 $14 = $retval;
 $15 = ((($14)) + 1|0);
 $retval = $15;
 $16 = $retval;
 STACKTOP = sp;return ($16|0);
}
function _tryOpenDir($funcs,$d,$forWriting) {
 $funcs = $funcs|0;
 $d = $d|0;
 $forWriting = $forWriting|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $opaque = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $funcs;
 $1 = $d;
 $2 = $forWriting;
 $retval = 0;
 $3 = $0;
 $4 = ((($3)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $1;
 $7 = $2;
 $8 = (FUNCTION_TABLE_iii[$5 & 8191]($6,$7)|0);
 $9 = ($8|0)!=(0);
 if (!($9)) {
  $34 = $retval;
  STACKTOP = sp;return ($34|0);
 }
 $10 = $0;
 $11 = ((($10)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = $1;
 $14 = $2;
 $15 = (FUNCTION_TABLE_iii[$12 & 8191]($13,$14)|0);
 $opaque = $15;
 $16 = $opaque;
 $17 = ($16|0)!=(0|0);
 if (!($17)) {
  $34 = $retval;
  STACKTOP = sp;return ($34|0);
 }
 $18 = HEAP32[(732252)>>2]|0;
 $19 = (FUNCTION_TABLE_iii[$18 & 8191](20,0)|0);
 $retval = $19;
 $20 = $retval;
 $21 = ($20|0)==(0|0);
 if ($21) {
  $22 = $0;
  $23 = ((($22)) + 52|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = $opaque;
  FUNCTION_TABLE_vi[$24 & 8191]($25);
  $34 = $retval;
  STACKTOP = sp;return ($34|0);
 } else {
  $26 = $retval;
  ;HEAP32[$26>>2]=0|0;HEAP32[$26+4>>2]=0|0;HEAP32[$26+8>>2]=0|0;HEAP32[$26+12>>2]=0|0;HEAP32[$26+16>>2]=0|0;
  $27 = $retval;
  $28 = ((($27)) + 8|0);
  HEAP32[$28>>2] = 0;
  $29 = $0;
  $30 = $retval;
  $31 = ((($30)) + 12|0);
  HEAP32[$31>>2] = $29;
  $32 = $opaque;
  $33 = $retval;
  HEAP32[$33>>2] = $32;
  $34 = $retval;
  STACKTOP = sp;return ($34|0);
 }
 return (0)|0;
}
function ___PHYSFS_smallFree($ptr) {
 $ptr = $ptr|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $block = 0, $useHeap = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $ptr;
 $1 = $0;
 $2 = ($1|0)!=(0|0);
 if (!($2)) {
  STACKTOP = sp;return;
 }
 $3 = $0;
 $4 = ((($3)) + -1|0);
 $block = $4;
 $5 = $block;
 $6 = HEAP8[$5>>0]|0;
 $useHeap = $6;
 $7 = $useHeap;
 $8 = ($7<<24>>24)!=(0);
 if (!($8)) {
  STACKTOP = sp;return;
 }
 $9 = HEAP32[(732260)>>2]|0;
 $10 = $block;
 FUNCTION_TABLE_vi[$9 & 8191]($10);
 STACKTOP = sp;return;
}
function _freeSearchPath() {
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0, $next = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $next = 0;
 (_closeFileHandleList(732300)|0);
 $0 = HEAP32[183076]|0;
 $1 = ($0|0)!=(0|0);
 if (!($1)) {
  STACKTOP = sp;return;
 }
 $2 = HEAP32[183076]|0;
 $i = $2;
 while(1) {
  $3 = $i;
  $4 = ($3|0)!=(0|0);
  if (!($4)) {
   break;
  }
  $5 = $i;
  $6 = ((($5)) + 16|0);
  $7 = HEAP32[$6>>2]|0;
  $next = $7;
  $8 = $i;
  $9 = HEAP32[183075]|0;
  (_freeDirHandle($8,$9)|0);
  $10 = $next;
  $i = $10;
 }
 HEAP32[183076] = 0;
 STACKTOP = sp;return;
}
function _freeErrorMessages() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $i = 0, $next = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = HEAP32[183067]|0;
 $i = $0;
 while(1) {
  $1 = $i;
  $2 = ($1|0)!=(0|0);
  if (!($2)) {
   break;
  }
  $3 = $i;
  $4 = ((($3)) + 88|0);
  $5 = HEAP32[$4>>2]|0;
  $next = $5;
  $6 = HEAP32[(732260)>>2]|0;
  $7 = $i;
  FUNCTION_TABLE_vi[$6 & 8191]($7);
  $8 = $next;
  $i = $8;
 }
 HEAP32[183067] = 0;
 STACKTOP = sp;return;
}
function _PHYSFS_isInit() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[183068]|0;
 return ($0|0);
}
function _PHYSFS_freeList($list) {
 $list = $list|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $list;
 $1 = $0;
 $i = $1;
 while(1) {
  $2 = $i;
  $3 = HEAP32[$2>>2]|0;
  $4 = ($3|0)!=(0|0);
  $5 = HEAP32[(732260)>>2]|0;
  if (!($4)) {
   break;
  }
  $6 = $i;
  $7 = HEAP32[$6>>2]|0;
  FUNCTION_TABLE_vi[$5 & 8191]($7);
  $8 = $i;
  $9 = ((($8)) + 4|0);
  $i = $9;
 }
 $10 = $0;
 FUNCTION_TABLE_vi[$5 & 8191]($10);
 STACKTOP = sp;return;
}
function _PHYSFS_getUserDir() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[183072]|0;
 return ($0|0);
}
function _PHYSFS_getWriteDir() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $retval = 0;
 $0 = HEAP32[183070]|0;
 (___PHYSFS_platformGrabMutex($0)|0);
 $1 = HEAP32[183074]|0;
 $2 = ($1|0)!=(0|0);
 if ($2) {
  $3 = HEAP32[183074]|0;
  $4 = ((($3)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  $retval = $5;
 }
 $6 = HEAP32[183070]|0;
 ___PHYSFS_platformReleaseMutex($6);
 $7 = $retval;
 STACKTOP = sp;return ($7|0);
}
function _PHYSFS_mount($newDir,$mountPoint,$appendToPath) {
 $newDir = $newDir|0;
 $mountPoint = $mountPoint|0;
 $appendToPath = $appendToPath|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $dh = 0, $i = 0, $prev = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $newDir;
 $2 = $mountPoint;
 $3 = $appendToPath;
 $prev = 0;
 $4 = $1;
 $5 = ($4|0)==(0|0);
 if ($5) {
  ___PHYSFS_setError(517147);
  $0 = 0;
  $40 = $0;
  STACKTOP = sp;return ($40|0);
 }
 $6 = $2;
 $7 = ($6|0)==(0|0);
 if ($7) {
  $2 = 523303;
 }
 $8 = HEAP32[183070]|0;
 (___PHYSFS_platformGrabMutex($8)|0);
 $9 = HEAP32[183076]|0;
 $i = $9;
 while(1) {
  $10 = $i;
  $11 = ($10|0)!=(0|0);
  $12 = $1;
  if (!($11)) {
   break;
  }
  $13 = $i;
  $14 = ((($13)) + 4|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = (_strcmp($12,$15)|0);
  $17 = ($16|0)==(0);
  if ($17) {
   label = 8;
   break;
  }
  $19 = $i;
  $prev = $19;
  $20 = $i;
  $21 = ((($20)) + 16|0);
  $22 = HEAP32[$21>>2]|0;
  $i = $22;
 }
 if ((label|0) == 8) {
  ___PHYSFS_setError(0);
  $18 = HEAP32[183070]|0;
  ___PHYSFS_platformReleaseMutex($18);
  $0 = 1;
  $40 = $0;
  STACKTOP = sp;return ($40|0);
 }
 $23 = $2;
 $24 = (_createDirHandle($12,$23,0)|0);
 $dh = $24;
 $25 = $dh;
 $26 = ($25|0)==(0|0);
 if ($26) {
  ___PHYSFS_setError(0);
  $27 = HEAP32[183070]|0;
  ___PHYSFS_platformReleaseMutex($27);
  $0 = 0;
  $40 = $0;
  STACKTOP = sp;return ($40|0);
 }
 $28 = $3;
 $29 = ($28|0)!=(0);
 do {
  if ($29) {
   $30 = $prev;
   $31 = ($30|0)==(0|0);
   $32 = $dh;
   if ($31) {
    HEAP32[183076] = $32;
    break;
   } else {
    $33 = $prev;
    $34 = ((($33)) + 16|0);
    HEAP32[$34>>2] = $32;
    break;
   }
  } else {
   $35 = HEAP32[183076]|0;
   $36 = $dh;
   $37 = ((($36)) + 16|0);
   HEAP32[$37>>2] = $35;
   $38 = $dh;
   HEAP32[183076] = $38;
  }
 } while(0);
 $39 = HEAP32[183070]|0;
 ___PHYSFS_platformReleaseMutex($39);
 $0 = 1;
 $40 = $0;
 STACKTOP = sp;return ($40|0);
}
function _PHYSFS_removeFromSearchPath($oldDir) {
 $oldDir = $oldDir|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0, $next = 0, $prev = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $oldDir;
 $prev = 0;
 $next = 0;
 $2 = $1;
 $3 = ($2|0)==(0|0);
 if ($3) {
  ___PHYSFS_setError(517147);
  $0 = 0;
  $32 = $0;
  STACKTOP = sp;return ($32|0);
 }
 $4 = HEAP32[183070]|0;
 (___PHYSFS_platformGrabMutex($4)|0);
 $5 = HEAP32[183076]|0;
 $i = $5;
 while(1) {
  $6 = $i;
  $7 = ($6|0)!=(0|0);
  if (!($7)) {
   label = 13;
   break;
  }
  $8 = $i;
  $9 = ((($8)) + 4|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = $1;
  $12 = (_strcmp($10,$11)|0);
  $13 = ($12|0)==(0);
  $14 = $i;
  if ($13) {
   break;
  }
  $prev = $14;
  $28 = $i;
  $29 = ((($28)) + 16|0);
  $30 = HEAP32[$29>>2]|0;
  $i = $30;
 }
 if ((label|0) == 13) {
  ___PHYSFS_setError(516722);
  $31 = HEAP32[183070]|0;
  ___PHYSFS_platformReleaseMutex($31);
  $0 = 0;
  $32 = $0;
  STACKTOP = sp;return ($32|0);
 }
 $15 = ((($14)) + 16|0);
 $16 = HEAP32[$15>>2]|0;
 $next = $16;
 $17 = $i;
 $18 = HEAP32[183075]|0;
 $19 = (_freeDirHandle($17,$18)|0);
 $20 = ($19|0)!=(0);
 if (!($20)) {
  ___PHYSFS_setError(0);
  $21 = HEAP32[183070]|0;
  ___PHYSFS_platformReleaseMutex($21);
  $0 = 0;
  $32 = $0;
  STACKTOP = sp;return ($32|0);
 }
 $22 = $prev;
 $23 = ($22|0)==(0|0);
 $24 = $next;
 if ($23) {
  HEAP32[183076] = $24;
 } else {
  $25 = $prev;
  $26 = ((($25)) + 16|0);
  HEAP32[$26>>2] = $24;
 }
 ___PHYSFS_setError(0);
 $27 = HEAP32[183070]|0;
 ___PHYSFS_platformReleaseMutex($27);
 $0 = 1;
 $32 = $0;
 STACKTOP = sp;return ($32|0);
}
function _PHYSFS_getMountPoint($dir) {
 $dir = $dir|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $dir;
 $2 = HEAP32[183070]|0;
 (___PHYSFS_platformGrabMutex($2)|0);
 $3 = HEAP32[183076]|0;
 $i = $3;
 while(1) {
  $4 = $i;
  $5 = ($4|0)!=(0|0);
  if (!($5)) {
   label = 8;
   break;
  }
  $6 = $i;
  $7 = ((($6)) + 4|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = $1;
  $10 = (_strcmp($8,$9)|0);
  $11 = ($10|0)==(0);
  $12 = $i;
  if ($11) {
   break;
  }
  $22 = ((($12)) + 16|0);
  $23 = HEAP32[$22>>2]|0;
  $i = $23;
 }
 if ((label|0) == 8) {
  $24 = HEAP32[183070]|0;
  ___PHYSFS_platformReleaseMutex($24);
  ___PHYSFS_setError(516722);
  $0 = 0;
  $25 = $0;
  STACKTOP = sp;return ($25|0);
 }
 $13 = ((($12)) + 8|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ($14|0)!=(0|0);
 if ($15) {
  $16 = $i;
  $17 = ((($16)) + 8|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = $18;
 } else {
  $19 = 523303;
 }
 $retval = $19;
 $20 = HEAP32[183070]|0;
 ___PHYSFS_platformReleaseMutex($20);
 $21 = $retval;
 $0 = $21;
 $25 = $0;
 STACKTOP = sp;return ($25|0);
}
function _PHYSFS_mkdir($_dname) {
 $_dname = $_dname|0;
 var $$alloca_mul = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dname = 0, $len = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $_dname;
 $retval = 0;
 $2 = $1;
 $3 = ($2|0)==(0|0);
 if ($3) {
  ___PHYSFS_setError(517147);
  $0 = 0;
  $22 = $0;
  STACKTOP = sp;return ($22|0);
 }
 $4 = $1;
 $5 = (_strlen($4)|0);
 $6 = (($5) + 1)|0;
 $len = $6;
 $7 = $len;
 $8 = ($7>>>0)<(128);
 if ($8) {
  $9 = $len;
  $10 = (($9) + 1)|0;
  $$alloca_mul = $10;
  $11 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
  $13 = $11;
 } else {
  $13 = 0;
 }
 $12 = $len;
 $14 = (___PHYSFS_initSmallAlloc($13,$12,0)|0);
 $dname = $14;
 $15 = $dname;
 $16 = ($15|0)==(0|0);
 if ($16) {
  ___PHYSFS_setError(596704);
  $0 = 0;
  $22 = $0;
  STACKTOP = sp;return ($22|0);
 } else {
  $17 = $1;
  $18 = $dname;
  $19 = (_doMkdir($17,$18)|0);
  $retval = $19;
  $20 = $dname;
  ___PHYSFS_smallFree($20);
  $21 = $retval;
  $0 = $21;
  $22 = $0;
  STACKTOP = sp;return ($22|0);
 }
 return (0)|0;
}
function _doMkdir($_dname,$dname) {
 $_dname = $_dname|0;
 $dname = $dname|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $end = 0, $exists = 0, $h = 0, $or$cond = 0, $retval = 0, $start = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $2 = sp + 20|0;
 $exists = sp;
 $1 = $_dname;
 HEAP32[$2>>2] = $dname;
 $retval = 0;
 HEAP32[$exists>>2] = 1;
 $3 = $1;
 $4 = HEAP32[$2>>2]|0;
 $5 = (_sanitizePlatformIndependentPath($3,$4)|0);
 $6 = ($5|0)!=(0);
 if (!($6)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $53 = $0;
  STACKTOP = sp;return ($53|0);
 }
 $7 = HEAP32[183070]|0;
 (___PHYSFS_platformGrabMutex($7)|0);
 $8 = HEAP32[183074]|0;
 $9 = ($8|0)==(0|0);
 if ($9) {
  ___PHYSFS_setError(516751);
  $10 = HEAP32[183070]|0;
  ___PHYSFS_platformReleaseMutex($10);
  $0 = 0;
  $53 = $0;
  STACKTOP = sp;return ($53|0);
 }
 $11 = HEAP32[183074]|0;
 $h = $11;
 $12 = $h;
 $13 = (_verifyPath($12,$2,1)|0);
 $14 = ($13|0)!=(0);
 if (!($14)) {
  ___PHYSFS_setError(0);
  $15 = HEAP32[183070]|0;
  ___PHYSFS_platformReleaseMutex($15);
  $0 = 0;
  $53 = $0;
  STACKTOP = sp;return ($53|0);
 }
 $16 = HEAP32[$2>>2]|0;
 $start = $16;
 while(1) {
  $17 = $start;
  $18 = (_strchr($17,47)|0);
  $end = $18;
  $19 = $end;
  $20 = ($19|0)!=(0|0);
  if ($20) {
   $21 = $end;
   HEAP8[$21>>0] = 0;
  }
  $22 = HEAP32[$exists>>2]|0;
  $23 = ($22|0)!=(0);
  if ($23) {
   $24 = $h;
   $25 = ((($24)) + 12|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = ((($26)) + 20|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = $h;
   $30 = HEAP32[$29>>2]|0;
   $31 = HEAP32[$2>>2]|0;
   $32 = (FUNCTION_TABLE_iiii[$28 & 8191]($30,$31,$exists)|0);
   $retval = $32;
  }
  $33 = HEAP32[$exists>>2]|0;
  $34 = ($33|0)!=(0);
  if (!($34)) {
   $35 = $h;
   $36 = ((($35)) + 12|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = ((($37)) + 48|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = $h;
   $41 = HEAP32[$40>>2]|0;
   $42 = HEAP32[$2>>2]|0;
   $43 = (FUNCTION_TABLE_iii[$39 & 8191]($41,$42)|0);
   $retval = $43;
  }
  $44 = $retval;
  $45 = ($44|0)==(0);
  $46 = $end;
  $47 = ($46|0)==(0|0);
  $or$cond = $45 | $47;
  if ($or$cond) {
   break;
  }
  $48 = $end;
  HEAP8[$48>>0] = 47;
  $49 = $end;
  $50 = ((($49)) + 1|0);
  $start = $50;
 }
 $51 = HEAP32[183070]|0;
 ___PHYSFS_platformReleaseMutex($51);
 $52 = $retval;
 $0 = $52;
 $53 = $0;
 STACKTOP = sp;return ($53|0);
}
function _verifyPath($h,$_fname,$allowMissing) {
 $h = $h|0;
 $_fname = $_fname|0;
 $allowMissing = $allowMissing|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $end = 0, $fname = 0, $len = 0, $mntpntlen = 0, $or$cond = 0, $rc = 0, $retval = 0, $start = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $retval = sp + 20|0;
 $1 = $h;
 $2 = $_fname;
 $3 = $allowMissing;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $fname = $5;
 HEAP32[$retval>>2] = 1;
 $6 = $fname;
 $7 = HEAP8[$6>>0]|0;
 $8 = $7 << 24 >> 24;
 $9 = ($8|0)==(0);
 if ($9) {
  $0 = 1;
  $89 = $0;
  STACKTOP = sp;return ($89|0);
 }
 $10 = $1;
 $11 = ((($10)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)!=(0|0);
 if ($13) {
  $14 = $1;
  $15 = ((($14)) + 8|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = (_strlen($16)|0);
  $mntpntlen = $17;
  $18 = $fname;
  $19 = (_strlen($18)|0);
  $len = $19;
  $20 = $mntpntlen;
  $21 = ($20>>>0)>(1);
  if (!($21)) {
   ___assert_fail((516778|0),(516561|0),1286,(516792|0));
   // unreachable;
  }
  $22 = $len;
  $23 = $mntpntlen;
  $24 = (($23) - 1)|0;
  $25 = ($22>>>0)<($24>>>0);
  if ($25) {
   ___PHYSFS_setError(516803);
   $0 = 0;
   $89 = $0;
   STACKTOP = sp;return ($89|0);
  }
  $26 = $1;
  $27 = ((($26)) + 8|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = $fname;
  $30 = $mntpntlen;
  $31 = (($30) - 1)|0;
  $32 = (_strncmp($28,$29,$31)|0);
  HEAP32[$retval>>2] = $32;
  $33 = HEAP32[$retval>>2]|0;
  $34 = ($33|0)!=(0);
  if ($34) {
   ___PHYSFS_setError(516803);
   $0 = 0;
   $89 = $0;
   STACKTOP = sp;return ($89|0);
  }
  $35 = $len;
  $36 = $mntpntlen;
  $37 = (($36) - 1)|0;
  $38 = ($35>>>0)>($37>>>0);
  if ($38) {
   $39 = $mntpntlen;
   $40 = (($39) - 1)|0;
   $41 = $fname;
   $42 = (($41) + ($40)|0);
   $43 = HEAP8[$42>>0]|0;
   $44 = $43 << 24 >> 24;
   $45 = ($44|0)!=(47);
   if ($45) {
    ___PHYSFS_setError(516803);
    $0 = 0;
    $89 = $0;
    STACKTOP = sp;return ($89|0);
   }
  }
  $46 = $mntpntlen;
  $47 = (($46) - 1)|0;
  $48 = $fname;
  $49 = (($48) + ($47)|0);
  $fname = $49;
  $50 = $fname;
  $51 = HEAP8[$50>>0]|0;
  $52 = $51 << 24 >> 24;
  $53 = ($52|0)==(47);
  if ($53) {
   $54 = $fname;
   $55 = ((($54)) + 1|0);
   $fname = $55;
  }
  $56 = $fname;
  $57 = $2;
  HEAP32[$57>>2] = $56;
  HEAP32[$retval>>2] = 1;
 }
 $58 = $fname;
 $start = $58;
 $59 = HEAP32[183077]|0;
 $60 = ($59|0)!=(0);
 L27: do {
  if (!($60)) {
   while(1) {
    $rc = 0;
    $61 = $start;
    $62 = (_strchr($61,47)|0);
    $end = $62;
    $63 = $end;
    $64 = ($63|0)!=(0|0);
    if ($64) {
     $65 = $end;
     HEAP8[$65>>0] = 0;
    }
    $66 = $1;
    $67 = ((($66)) + 12|0);
    $68 = HEAP32[$67>>2]|0;
    $69 = ((($68)) + 24|0);
    $70 = HEAP32[$69>>2]|0;
    $71 = $1;
    $72 = HEAP32[$71>>2]|0;
    $73 = $fname;
    $74 = (FUNCTION_TABLE_iiii[$70 & 8191]($72,$73,$retval)|0);
    $rc = $74;
    $75 = $end;
    $76 = ($75|0)!=(0|0);
    if ($76) {
     $77 = $end;
     HEAP8[$77>>0] = 47;
    }
    $78 = $rc;
    $79 = ($78|0)!=(0);
    if ($79) {
     label = 22;
     break;
    }
    $80 = HEAP32[$retval>>2]|0;
    $81 = ($80|0)!=(0);
    $82 = $end;
    $83 = ($82|0)==(0|0);
    if (!($81)) {
     label = 24;
     break;
    }
    if ($83) {
     break L27;
    }
    $86 = $end;
    $87 = ((($86)) + 1|0);
    $start = $87;
   }
   if ((label|0) == 22) {
    ___PHYSFS_setError(516818);
    $0 = 0;
    $89 = $0;
    STACKTOP = sp;return ($89|0);
   }
   else if ((label|0) == 24) {
    $84 = $3;
    $85 = ($84|0)!=(0);
    $or$cond = $83 | $85;
    if (!($or$cond)) {
     break;
    }
    HEAP32[$retval>>2] = 1;
    break;
   }
  }
 } while(0);
 $88 = HEAP32[$retval>>2]|0;
 $0 = $88;
 $89 = $0;
 STACKTOP = sp;return ($89|0);
}
function _PHYSFS_enumerateFiles($path) {
 $path = $path|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ecd = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ecd = sp;
 $1 = $path;
 ;HEAP32[$ecd>>2]=0|0;HEAP32[$ecd+4>>2]=0|0;HEAP32[$ecd+8>>2]=0|0;
 $2 = HEAP32[(732252)>>2]|0;
 $3 = (FUNCTION_TABLE_iii[$2 & 8191](4,0)|0);
 HEAP32[$ecd>>2] = $3;
 $4 = HEAP32[$ecd>>2]|0;
 $5 = ($4|0)==(0|0);
 if ($5) {
  ___PHYSFS_setError(596704);
  $0 = 0;
  $12 = $0;
  STACKTOP = sp;return ($12|0);
 } else {
  $6 = $1;
  _PHYSFS_enumerateFilesCallback($6,4281,$ecd);
  $7 = ((($ecd)) + 4|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = HEAP32[$ecd>>2]|0;
  $10 = (($9) + ($8<<2)|0);
  HEAP32[$10>>2] = 0;
  $11 = HEAP32[$ecd>>2]|0;
  $0 = $11;
  $12 = $0;
  STACKTOP = sp;return ($12|0);
 }
 return (0)|0;
}
function _enumFilesCallback($data,$origdir,$str) {
 $data = $data|0;
 $origdir = $origdir|0;
 $str = $str|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $newstr = 0, $or$cond = 0, $pecd = 0, $pos = 0, $ptr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $pos = sp + 12|0;
 $0 = $data;
 $1 = $origdir;
 $2 = $str;
 $3 = $0;
 $pecd = $3;
 $4 = $pecd;
 $5 = ((($4)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 HEAP32[$pos>>2] = $6;
 $7 = $2;
 $8 = $pecd;
 $9 = HEAP32[$8>>2]|0;
 $10 = (_locateInStringList($7,$9,$pos)|0);
 $11 = ($10|0)!=(0);
 if ($11) {
  STACKTOP = sp;return;
 }
 $12 = HEAP32[(732256)>>2]|0;
 $13 = $pecd;
 $14 = HEAP32[$13>>2]|0;
 $15 = $pecd;
 $16 = ((($15)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = (($17) + 2)|0;
 $19 = $18<<2;
 $20 = (FUNCTION_TABLE_iiii[$12 & 8191]($14,$19,0)|0);
 $ptr = $20;
 $21 = HEAP32[(732252)>>2]|0;
 $22 = $2;
 $23 = (_strlen($22)|0);
 $24 = (($23) + 1)|0;
 $25 = (FUNCTION_TABLE_iii[$21 & 8191]($24,0)|0);
 $newstr = $25;
 $26 = $ptr;
 $27 = ($26|0)!=(0|0);
 if ($27) {
  $28 = $ptr;
  $29 = $pecd;
  HEAP32[$29>>2] = $28;
 }
 $30 = $ptr;
 $31 = ($30|0)==(0|0);
 $32 = $newstr;
 $33 = ($32|0)==(0|0);
 $or$cond = $31 | $33;
 if ($or$cond) {
  STACKTOP = sp;return;
 }
 $34 = $newstr;
 $35 = $2;
 (_strcpy($34,$35)|0);
 $36 = HEAP32[$pos>>2]|0;
 $37 = $pecd;
 $38 = ((($37)) + 4|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ($36|0)!=($39|0);
 if ($40) {
  $41 = HEAP32[$pos>>2]|0;
  $42 = (($41) + 1)|0;
  $43 = $pecd;
  $44 = HEAP32[$43>>2]|0;
  $45 = (($44) + ($42<<2)|0);
  $46 = HEAP32[$pos>>2]|0;
  $47 = $pecd;
  $48 = HEAP32[$47>>2]|0;
  $49 = (($48) + ($46<<2)|0);
  $50 = $pecd;
  $51 = ((($50)) + 4|0);
  $52 = HEAP32[$51>>2]|0;
  $53 = HEAP32[$pos>>2]|0;
  $54 = (($52) - ($53))|0;
  $55 = $54<<2;
  _memmove(($45|0),($49|0),($55|0))|0;
 }
 $56 = $newstr;
 $57 = HEAP32[$pos>>2]|0;
 $58 = $pecd;
 $59 = HEAP32[$58>>2]|0;
 $60 = (($59) + ($57<<2)|0);
 HEAP32[$60>>2] = $56;
 $61 = $pecd;
 $62 = ((($61)) + 4|0);
 $63 = HEAP32[$62>>2]|0;
 $64 = (($63) + 1)|0;
 HEAP32[$62>>2] = $64;
 STACKTOP = sp;return;
}
function _locateInStringList($str,$list,$pos) {
 $str = $str|0;
 $list = $list|0;
 $pos = $pos|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cmp = 0, $half_len = 0, $len = 0, $lo = 0, $middle = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $str;
 $2 = $list;
 $3 = $pos;
 $4 = $3;
 $5 = HEAP32[$4>>2]|0;
 $len = $5;
 $lo = 0;
 while(1) {
  $6 = $len;
  $7 = ($6>>>0)>(0);
  if (!($7)) {
   label = 8;
   break;
  }
  $8 = $len;
  $9 = $8 >>> 1;
  $half_len = $9;
  $10 = $lo;
  $11 = $half_len;
  $12 = (($10) + ($11))|0;
  $middle = $12;
  $13 = $middle;
  $14 = $2;
  $15 = (($14) + ($13<<2)|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = $1;
  $18 = (_strcmp($16,$17)|0);
  $cmp = $18;
  $19 = $cmp;
  $20 = ($19|0)==(0);
  if ($20) {
   label = 4;
   break;
  }
  $21 = $cmp;
  $22 = ($21|0)>(0);
  if ($22) {
   $23 = $half_len;
   $len = $23;
   continue;
  } else {
   $24 = $middle;
   $25 = (($24) + 1)|0;
   $lo = $25;
   $26 = $half_len;
   $27 = (($26) + 1)|0;
   $28 = $len;
   $29 = (($28) - ($27))|0;
   $len = $29;
   continue;
  }
 }
 if ((label|0) == 4) {
  $0 = 1;
  $32 = $0;
  STACKTOP = sp;return ($32|0);
 }
 else if ((label|0) == 8) {
  $30 = $lo;
  $31 = $3;
  HEAP32[$31>>2] = $30;
  $0 = 0;
  $32 = $0;
  STACKTOP = sp;return ($32|0);
 }
 return (0)|0;
}
function _PHYSFS_enumerateFilesCallback($_fname,$callback,$data) {
 $_fname = $_fname|0;
 $callback = $callback|0;
 $data = $data|0;
 var $$alloca_mul = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $arcfname = 0, $fname = 0, $i = 0, $len = 0, $noSyms = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $arcfname = sp;
 $0 = $_fname;
 $1 = $callback;
 $2 = $data;
 $3 = $0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  ___PHYSFS_setError(517147);
  STACKTOP = sp;return;
 }
 $5 = $1;
 $6 = ($5|0)==(0|0);
 if ($6) {
  ___PHYSFS_setError(517147);
  STACKTOP = sp;return;
 }
 $7 = $0;
 $8 = (_strlen($7)|0);
 $9 = (($8) + 1)|0;
 $len = $9;
 $10 = $len;
 $11 = ($10>>>0)<(128);
 if ($11) {
  $12 = $len;
  $13 = (($12) + 1)|0;
  $$alloca_mul = $13;
  $14 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
  $16 = $14;
 } else {
  $16 = 0;
 }
 $15 = $len;
 $17 = (___PHYSFS_initSmallAlloc($16,$15,0)|0);
 $fname = $17;
 $18 = $fname;
 $19 = ($18|0)==(0|0);
 if ($19) {
  ___PHYSFS_setError(596704);
  STACKTOP = sp;return;
 }
 $20 = $0;
 $21 = $fname;
 $22 = (_sanitizePlatformIndependentPath($20,$21)|0);
 $23 = ($22|0)!=(0);
 if ($23) {
  $24 = HEAP32[183070]|0;
  (___PHYSFS_platformGrabMutex($24)|0);
  $25 = HEAP32[183077]|0;
  $26 = ($25|0)!=(0);
  $27 = $26 ^ 1;
  $28 = $27&1;
  $noSyms = $28;
  $29 = HEAP32[183076]|0;
  $i = $29;
  while(1) {
   $30 = $i;
   $31 = ($30|0)!=(0|0);
   if (!($31)) {
    break;
   }
   $32 = $fname;
   HEAP32[$arcfname>>2] = $32;
   $33 = $i;
   $34 = HEAP32[$arcfname>>2]|0;
   $35 = (_partOfMountPoint($33,$34)|0);
   $36 = ($35|0)!=(0);
   $37 = $i;
   if ($36) {
    $38 = HEAP32[$arcfname>>2]|0;
    $39 = $1;
    $40 = $0;
    $41 = $2;
    _enumerateFromMountPoint($37,$38,$39,$40,$41);
   } else {
    $42 = (_verifyPath($37,$arcfname,0)|0);
    $43 = ($42|0)!=(0);
    if ($43) {
     $44 = $i;
     $45 = ((($44)) + 12|0);
     $46 = HEAP32[$45>>2]|0;
     $47 = ((($46)) + 12|0);
     $48 = HEAP32[$47>>2]|0;
     $49 = $i;
     $50 = HEAP32[$49>>2]|0;
     $51 = HEAP32[$arcfname>>2]|0;
     $52 = $noSyms;
     $53 = $1;
     $54 = $0;
     $55 = $2;
     FUNCTION_TABLE_viiiiii[$48 & 8191]($50,$51,$52,$53,$54,$55);
    }
   }
   $56 = $i;
   $57 = ((($56)) + 16|0);
   $58 = HEAP32[$57>>2]|0;
   $i = $58;
  }
  $59 = HEAP32[183070]|0;
  ___PHYSFS_platformReleaseMutex($59);
 }
 $60 = $fname;
 ___PHYSFS_smallFree($60);
 STACKTOP = sp;return;
}
function _partOfMountPoint($h,$fname) {
 $h = $h|0;
 $fname = $fname|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $len = 0, $mntpntlen = 0, $rc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $h;
 $2 = $fname;
 $3 = $1;
 $4 = ((($3)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 do {
  if ($6) {
   $0 = 0;
  } else {
   $7 = $2;
   $8 = HEAP8[$7>>0]|0;
   $9 = $8 << 24 >> 24;
   $10 = ($9|0)==(0);
   if ($10) {
    $0 = 1;
    break;
   }
   $11 = $2;
   $12 = (_strlen($11)|0);
   $len = $12;
   $13 = $1;
   $14 = ((($13)) + 8|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = (_strlen($15)|0);
   $mntpntlen = $16;
   $17 = $len;
   $18 = $mntpntlen;
   $19 = ($17>>>0)>($18>>>0);
   if ($19) {
    $0 = 0;
    break;
   }
   $20 = $len;
   $21 = (($20) + 1)|0;
   $22 = $mntpntlen;
   $23 = ($21|0)==($22|0);
   if ($23) {
    $0 = 0;
    break;
   }
   $24 = $2;
   $25 = $1;
   $26 = ((($25)) + 8|0);
   $27 = HEAP32[$26>>2]|0;
   $28 = $len;
   $29 = (_strncmp($24,$27,$28)|0);
   $rc = $29;
   $30 = $rc;
   $31 = ($30|0)!=(0);
   if ($31) {
    $0 = 0;
    break;
   } else {
    $32 = $len;
    $33 = $1;
    $34 = ((($33)) + 8|0);
    $35 = HEAP32[$34>>2]|0;
    $36 = (($35) + ($32)|0);
    $37 = HEAP8[$36>>0]|0;
    $38 = $37 << 24 >> 24;
    $39 = ($38|0)==(47);
    $40 = $39&1;
    $0 = $40;
    break;
   }
  }
 } while(0);
 $41 = $0;
 STACKTOP = sp;return ($41|0);
}
function _enumerateFromMountPoint($i,$arcfname,$callback,$_fname,$data) {
 $i = $i|0;
 $arcfname = $arcfname|0;
 $callback = $callback|0;
 $_fname = $_fname|0;
 $data = $data|0;
 var $$alloca_mul = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $end = 0, $len = 0, $mountPoint = 0, $ptr = 0, $slen = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $i;
 $1 = $arcfname;
 $2 = $callback;
 $3 = $_fname;
 $4 = $data;
 $5 = $1;
 $6 = (_strlen($5)|0);
 $len = $6;
 $ptr = 0;
 $end = 0;
 $7 = $0;
 $8 = ((($7)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (_strlen($9)|0);
 $11 = (($10) + 1)|0;
 $slen = $11;
 $12 = $slen;
 $13 = ($12>>>0)<(128);
 if ($13) {
  $14 = $slen;
  $15 = (($14) + 1)|0;
  $$alloca_mul = $15;
  $16 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
  $18 = $16;
 } else {
  $18 = 0;
 }
 $17 = $slen;
 $19 = (___PHYSFS_initSmallAlloc($18,$17,0)|0);
 $mountPoint = $19;
 $20 = $mountPoint;
 $21 = ($20|0)==(0|0);
 if ($21) {
  STACKTOP = sp;return;
 }
 $22 = $mountPoint;
 $23 = $0;
 $24 = ((($23)) + 8|0);
 $25 = HEAP32[$24>>2]|0;
 (_strcpy($22,$25)|0);
 $26 = $mountPoint;
 $27 = $len;
 $28 = ($27|0)!=(0);
 $29 = $len;
 $30 = (($29) + 1)|0;
 $31 = $28 ? $30 : 0;
 $32 = (($26) + ($31)|0);
 $ptr = $32;
 $33 = $ptr;
 $34 = (_strchr($33,47)|0);
 $end = $34;
 $35 = $end;
 $36 = ($35|0)!=(0|0);
 if (!($36)) {
  ___assert_fail((521632|0),(516561|0),1565,(516846|0));
  // unreachable;
 }
 $37 = $end;
 HEAP8[$37>>0] = 0;
 $38 = $2;
 $39 = $4;
 $40 = $3;
 $41 = $ptr;
 FUNCTION_TABLE_viii[$38 & 8191]($39,$40,$41);
 $42 = $mountPoint;
 ___PHYSFS_smallFree($42);
 STACKTOP = sp;return;
}
function _PHYSFS_getRealDir($_fname) {
 $_fname = $_fname|0;
 var $$alloca_mul = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $arcfname = 0;
 var $fname = 0, $i = 0, $len = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $arcfname = sp;
 $1 = $_fname;
 $retval = 0;
 $fname = 0;
 $2 = $1;
 $3 = ($2|0)==(0|0);
 if ($3) {
  ___PHYSFS_setError(517147);
  $0 = 0;
  $57 = $0;
  STACKTOP = sp;return ($57|0);
 }
 $4 = $1;
 $5 = (_strlen($4)|0);
 $6 = (($5) + 1)|0;
 $len = $6;
 $7 = $len;
 $8 = ($7>>>0)<(128);
 if ($8) {
  $9 = $len;
  $10 = (($9) + 1)|0;
  $$alloca_mul = $10;
  $11 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
  $13 = $11;
 } else {
  $13 = 0;
 }
 $12 = $len;
 $14 = (___PHYSFS_initSmallAlloc($13,$12,0)|0);
 $fname = $14;
 $15 = $fname;
 $16 = ($15|0)==(0|0);
 if ($16) {
  ___PHYSFS_setError(596704);
  $0 = 0;
  $57 = $0;
  STACKTOP = sp;return ($57|0);
 }
 $17 = $1;
 $18 = $fname;
 $19 = (_sanitizePlatformIndependentPath($17,$18)|0);
 $20 = ($19|0)!=(0);
 if ($20) {
  $21 = HEAP32[183070]|0;
  (___PHYSFS_platformGrabMutex($21)|0);
  $22 = HEAP32[183076]|0;
  $i = $22;
  while(1) {
   $23 = $i;
   $24 = ($23|0)!=(0|0);
   $25 = $retval;
   $26 = ($25|0)==(0|0);
   $27 = $24 ? $26 : 0;
   if (!($27)) {
    break;
   }
   $28 = $fname;
   HEAP32[$arcfname>>2] = $28;
   $29 = $i;
   $30 = HEAP32[$arcfname>>2]|0;
   $31 = (_partOfMountPoint($29,$30)|0);
   $32 = ($31|0)!=(0);
   $33 = $i;
   if ($32) {
    $34 = ((($33)) + 4|0);
    $35 = HEAP32[$34>>2]|0;
    $retval = $35;
   } else {
    $36 = (_verifyPath($33,$arcfname,0)|0);
    $37 = ($36|0)!=(0);
    if ($37) {
     $38 = $i;
     $39 = ((($38)) + 12|0);
     $40 = HEAP32[$39>>2]|0;
     $41 = ((($40)) + 16|0);
     $42 = HEAP32[$41>>2]|0;
     $43 = $i;
     $44 = HEAP32[$43>>2]|0;
     $45 = HEAP32[$arcfname>>2]|0;
     $46 = (FUNCTION_TABLE_iii[$42 & 8191]($44,$45)|0);
     $47 = ($46|0)!=(0);
     if ($47) {
      $48 = $i;
      $49 = ((($48)) + 4|0);
      $50 = HEAP32[$49>>2]|0;
      $retval = $50;
     }
    }
   }
   $51 = $i;
   $52 = ((($51)) + 16|0);
   $53 = HEAP32[$52>>2]|0;
   $i = $53;
  }
  $54 = HEAP32[183070]|0;
  ___PHYSFS_platformReleaseMutex($54);
 }
 $55 = $fname;
 ___PHYSFS_smallFree($55);
 $56 = $retval;
 $0 = $56;
 $57 = $0;
 STACKTOP = sp;return ($57|0);
}
function _PHYSFS_permitSymbolicLinks($allow) {
 $allow = $allow|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $allow;
 $1 = $0;
 HEAP32[183077] = $1;
 STACKTOP = sp;return;
}
function _PHYSFS_symbolicLinksPermitted() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[183077]|0;
 return ($0|0);
}
function _PHYSFS_delete($_fname) {
 $_fname = $_fname|0;
 var $$alloca_mul = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fname = 0, $len = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $_fname;
 $2 = $1;
 $3 = ($2|0)==(0|0);
 if ($3) {
  ___PHYSFS_setError(517147);
  $0 = 0;
  $22 = $0;
  STACKTOP = sp;return ($22|0);
 }
 $4 = $1;
 $5 = (_strlen($4)|0);
 $6 = (($5) + 1)|0;
 $len = $6;
 $7 = $len;
 $8 = ($7>>>0)<(128);
 if ($8) {
  $9 = $len;
  $10 = (($9) + 1)|0;
  $$alloca_mul = $10;
  $11 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
  $13 = $11;
 } else {
  $13 = 0;
 }
 $12 = $len;
 $14 = (___PHYSFS_initSmallAlloc($13,$12,0)|0);
 $fname = $14;
 $15 = $fname;
 $16 = ($15|0)==(0|0);
 if ($16) {
  ___PHYSFS_setError(596704);
  $0 = 0;
  $22 = $0;
  STACKTOP = sp;return ($22|0);
 } else {
  $17 = $1;
  $18 = $fname;
  $19 = (_doDelete($17,$18)|0);
  $retval = $19;
  $20 = $fname;
  ___PHYSFS_smallFree($20);
  $21 = $retval;
  $0 = $21;
  $22 = $0;
  STACKTOP = sp;return ($22|0);
 }
 return (0)|0;
}
function _doDelete($_fname,$fname) {
 $_fname = $_fname|0;
 $fname = $fname|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $h = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $2 = sp + 8|0;
 $1 = $_fname;
 HEAP32[$2>>2] = $fname;
 $3 = $1;
 $4 = HEAP32[$2>>2]|0;
 $5 = (_sanitizePlatformIndependentPath($3,$4)|0);
 $6 = ($5|0)!=(0);
 if (!($6)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $27 = $0;
  STACKTOP = sp;return ($27|0);
 }
 $7 = HEAP32[183070]|0;
 (___PHYSFS_platformGrabMutex($7)|0);
 $8 = HEAP32[183074]|0;
 $9 = ($8|0)==(0|0);
 if ($9) {
  ___PHYSFS_setError(516751);
  $10 = HEAP32[183070]|0;
  ___PHYSFS_platformReleaseMutex($10);
  $0 = 0;
  $27 = $0;
  STACKTOP = sp;return ($27|0);
 }
 $11 = HEAP32[183074]|0;
 $h = $11;
 $12 = $h;
 $13 = (_verifyPath($12,$2,0)|0);
 $14 = ($13|0)!=(0);
 if ($14) {
  $16 = $h;
  $17 = ((($16)) + 12|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = ((($18)) + 44|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = $h;
  $22 = HEAP32[$21>>2]|0;
  $23 = HEAP32[$2>>2]|0;
  $24 = (FUNCTION_TABLE_iii[$20 & 8191]($22,$23)|0);
  $retval = $24;
  $25 = HEAP32[183070]|0;
  ___PHYSFS_platformReleaseMutex($25);
  $26 = $retval;
  $0 = $26;
  $27 = $0;
  STACKTOP = sp;return ($27|0);
 } else {
  ___PHYSFS_setError(0);
  $15 = HEAP32[183070]|0;
  ___PHYSFS_platformReleaseMutex($15);
  $0 = 0;
  $27 = $0;
  STACKTOP = sp;return ($27|0);
 }
 return (0)|0;
}
function _PHYSFS_exists($fname) {
 $fname = $fname|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $fname;
 $1 = $0;
 $2 = (_PHYSFS_getRealDir($1)|0);
 $3 = ($2|0)!=(0|0);
 $4 = $3&1;
 STACKTOP = sp;return ($4|0);
}
function _PHYSFS_getLastModTime($_fname) {
 $_fname = $_fname|0;
 var $$alloca_mul = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $arcfname = 0, $exists = 0, $fname = 0;
 var $i = 0, $len = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 8|0;
 $retval = sp;
 $exists = sp + 20|0;
 $arcfname = sp + 16|0;
 $1 = $_fname;
 $2 = $retval;
 $3 = $2;
 HEAP32[$3>>2] = -1;
 $4 = (($2) + 4)|0;
 $5 = $4;
 HEAP32[$5>>2] = -1;
 $6 = $1;
 $7 = ($6|0)==(0|0);
 if ($7) {
  ___PHYSFS_setError(517147);
  $8 = $0;
  $9 = $8;
  HEAP32[$9>>2] = -1;
  $10 = (($8) + 4)|0;
  $11 = $10;
  HEAP32[$11>>2] = -1;
  $90 = $0;
  $91 = $90;
  $92 = HEAP32[$91>>2]|0;
  $93 = (($90) + 4)|0;
  $94 = $93;
  $95 = HEAP32[$94>>2]|0;
  tempRet0 = ($95);
  STACKTOP = sp;return ($92|0);
 }
 $12 = $1;
 $13 = (_strlen($12)|0);
 $14 = (($13) + 1)|0;
 $len = $14;
 $15 = $len;
 $16 = ($15>>>0)<(128);
 if ($16) {
  $17 = $len;
  $18 = (($17) + 1)|0;
  $$alloca_mul = $18;
  $19 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
  $21 = $19;
 } else {
  $21 = 0;
 }
 $20 = $len;
 $22 = (___PHYSFS_initSmallAlloc($21,$20,0)|0);
 $fname = $22;
 $23 = $fname;
 $24 = ($23|0)==(0|0);
 if ($24) {
  ___PHYSFS_setError(596704);
  $25 = $0;
  $26 = $25;
  HEAP32[$26>>2] = -1;
  $27 = (($25) + 4)|0;
  $28 = $27;
  HEAP32[$28>>2] = -1;
  $90 = $0;
  $91 = $90;
  $92 = HEAP32[$91>>2]|0;
  $93 = (($90) + 4)|0;
  $94 = $93;
  $95 = HEAP32[$94>>2]|0;
  tempRet0 = ($95);
  STACKTOP = sp;return ($92|0);
 }
 $29 = $1;
 $30 = $fname;
 $31 = (_sanitizePlatformIndependentPath($29,$30)|0);
 $32 = ($31|0)!=(0);
 do {
  if ($32) {
   $33 = $fname;
   $34 = HEAP8[$33>>0]|0;
   $35 = $34 << 24 >> 24;
   $36 = ($35|0)==(0);
   if ($36) {
    $37 = $retval;
    $38 = $37;
    HEAP32[$38>>2] = 1;
    $39 = (($37) + 4)|0;
    $40 = $39;
    HEAP32[$40>>2] = 0;
    break;
   }
   HEAP32[$exists>>2] = 0;
   $41 = HEAP32[183070]|0;
   (___PHYSFS_platformGrabMutex($41)|0);
   $42 = HEAP32[183076]|0;
   $i = $42;
   while(1) {
    $43 = $i;
    $44 = ($43|0)!=(0|0);
    if (!($44)) {
     break;
    }
    $45 = HEAP32[$exists>>2]|0;
    $46 = ($45|0)!=(0);
    $47 = $46 ^ 1;
    if (!($47)) {
     break;
    }
    $48 = $fname;
    HEAP32[$arcfname>>2] = $48;
    $49 = $i;
    $50 = HEAP32[$arcfname>>2]|0;
    $51 = (_partOfMountPoint($49,$50)|0);
    HEAP32[$exists>>2] = $51;
    $52 = HEAP32[$exists>>2]|0;
    $53 = ($52|0)!=(0);
    if ($53) {
     $54 = $retval;
     $55 = $54;
     HEAP32[$55>>2] = 1;
     $56 = (($54) + 4)|0;
     $57 = $56;
     HEAP32[$57>>2] = 0;
    } else {
     $58 = $i;
     $59 = (_verifyPath($58,$arcfname,0)|0);
     $60 = ($59|0)!=(0);
     if ($60) {
      $61 = $i;
      $62 = ((($61)) + 12|0);
      $63 = HEAP32[$62>>2]|0;
      $64 = ((($63)) + 28|0);
      $65 = HEAP32[$64>>2]|0;
      $66 = $i;
      $67 = HEAP32[$66>>2]|0;
      $68 = HEAP32[$arcfname>>2]|0;
      $69 = (FUNCTION_TABLE_iiii[$65 & 8191]($67,$68,$exists)|0);
      $70 = tempRet0;
      $71 = $retval;
      $72 = $71;
      HEAP32[$72>>2] = $69;
      $73 = (($71) + 4)|0;
      $74 = $73;
      HEAP32[$74>>2] = $70;
     }
    }
    $75 = $i;
    $76 = ((($75)) + 16|0);
    $77 = HEAP32[$76>>2]|0;
    $i = $77;
   }
   $78 = HEAP32[183070]|0;
   ___PHYSFS_platformReleaseMutex($78);
  }
 } while(0);
 $79 = $fname;
 ___PHYSFS_smallFree($79);
 $80 = $retval;
 $81 = $80;
 $82 = HEAP32[$81>>2]|0;
 $83 = (($80) + 4)|0;
 $84 = $83;
 $85 = HEAP32[$84>>2]|0;
 $86 = $0;
 $87 = $86;
 HEAP32[$87>>2] = $82;
 $88 = (($86) + 4)|0;
 $89 = $88;
 HEAP32[$89>>2] = $85;
 $90 = $0;
 $91 = $90;
 $92 = HEAP32[$91>>2]|0;
 $93 = (($90) + 4)|0;
 $94 = $93;
 $95 = HEAP32[$94>>2]|0;
 tempRet0 = ($95);
 STACKTOP = sp;return ($92|0);
}
function _PHYSFS_isDirectory($_fname) {
 $_fname = $_fname|0;
 var $$alloca_mul = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $arcfname = 0, $exists = 0, $fname = 0;
 var $i = 0, $len = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $exists = sp + 4|0;
 $arcfname = sp;
 $1 = $_fname;
 $retval = 0;
 $2 = $1;
 $3 = ($2|0)==(0|0);
 if ($3) {
  ___PHYSFS_setError(517147);
  $0 = 0;
  $55 = $0;
  STACKTOP = sp;return ($55|0);
 }
 $4 = $1;
 $5 = (_strlen($4)|0);
 $6 = (($5) + 1)|0;
 $len = $6;
 $7 = $len;
 $8 = ($7>>>0)<(128);
 if ($8) {
  $9 = $len;
  $10 = (($9) + 1)|0;
  $$alloca_mul = $10;
  $11 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
  $13 = $11;
 } else {
  $13 = 0;
 }
 $12 = $len;
 $14 = (___PHYSFS_initSmallAlloc($13,$12,0)|0);
 $fname = $14;
 $15 = $fname;
 $16 = ($15|0)==(0|0);
 if ($16) {
  ___PHYSFS_setError(596704);
  $0 = 0;
  $55 = $0;
  STACKTOP = sp;return ($55|0);
 }
 $17 = $1;
 $18 = $fname;
 $19 = (_sanitizePlatformIndependentPath($17,$18)|0);
 $20 = ($19|0)!=(0);
 do {
  if ($20) {
   $21 = $fname;
   $22 = HEAP8[$21>>0]|0;
   $23 = $22 << 24 >> 24;
   $24 = ($23|0)==(0);
   if ($24) {
    $retval = 1;
    break;
   }
   HEAP32[$exists>>2] = 0;
   $25 = HEAP32[183070]|0;
   (___PHYSFS_platformGrabMutex($25)|0);
   $26 = HEAP32[183076]|0;
   $i = $26;
   while(1) {
    $27 = $i;
    $28 = ($27|0)!=(0|0);
    if (!($28)) {
     break;
    }
    $29 = HEAP32[$exists>>2]|0;
    $30 = ($29|0)!=(0);
    $31 = $30 ^ 1;
    if (!($31)) {
     break;
    }
    $32 = $fname;
    HEAP32[$arcfname>>2] = $32;
    $33 = $i;
    $34 = HEAP32[$arcfname>>2]|0;
    $35 = (_partOfMountPoint($33,$34)|0);
    HEAP32[$exists>>2] = $35;
    $36 = ($35|0)!=(0);
    if ($36) {
     $retval = 1;
    } else {
     $37 = $i;
     $38 = (_verifyPath($37,$arcfname,0)|0);
     $39 = ($38|0)!=(0);
     if ($39) {
      $40 = $i;
      $41 = ((($40)) + 12|0);
      $42 = HEAP32[$41>>2]|0;
      $43 = ((($42)) + 20|0);
      $44 = HEAP32[$43>>2]|0;
      $45 = $i;
      $46 = HEAP32[$45>>2]|0;
      $47 = HEAP32[$arcfname>>2]|0;
      $48 = (FUNCTION_TABLE_iiii[$44 & 8191]($46,$47,$exists)|0);
      $retval = $48;
     }
    }
    $49 = $i;
    $50 = ((($49)) + 16|0);
    $51 = HEAP32[$50>>2]|0;
    $i = $51;
   }
   $52 = HEAP32[183070]|0;
   ___PHYSFS_platformReleaseMutex($52);
  } else {
   $retval = 0;
  }
 } while(0);
 $53 = $fname;
 ___PHYSFS_smallFree($53);
 $54 = $retval;
 $0 = $54;
 $55 = $0;
 STACKTOP = sp;return ($55|0);
}
function _PHYSFS_isSymbolicLink($_fname) {
 $_fname = $_fname|0;
 var $$alloca_mul = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $arcfname = 0;
 var $fileExists = 0, $fname = 0, $i = 0, $len = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $fileExists = sp + 4|0;
 $arcfname = sp;
 $1 = $_fname;
 $retval = 0;
 $2 = HEAP32[183077]|0;
 $3 = ($2|0)!=(0);
 if (!($3)) {
  ___PHYSFS_setError(516818);
  $0 = 0;
  $57 = $0;
  STACKTOP = sp;return ($57|0);
 }
 $4 = $1;
 $5 = ($4|0)==(0|0);
 if ($5) {
  ___PHYSFS_setError(517147);
  $0 = 0;
  $57 = $0;
  STACKTOP = sp;return ($57|0);
 }
 $6 = $1;
 $7 = (_strlen($6)|0);
 $8 = (($7) + 1)|0;
 $len = $8;
 $9 = $len;
 $10 = ($9>>>0)<(128);
 if ($10) {
  $11 = $len;
  $12 = (($11) + 1)|0;
  $$alloca_mul = $12;
  $13 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
  $15 = $13;
 } else {
  $15 = 0;
 }
 $14 = $len;
 $16 = (___PHYSFS_initSmallAlloc($15,$14,0)|0);
 $fname = $16;
 $17 = $fname;
 $18 = ($17|0)==(0|0);
 if ($18) {
  ___PHYSFS_setError(596704);
  $0 = 0;
  $57 = $0;
  STACKTOP = sp;return ($57|0);
 }
 $19 = $1;
 $20 = $fname;
 $21 = (_sanitizePlatformIndependentPath($19,$20)|0);
 $22 = ($21|0)!=(0);
 do {
  if ($22) {
   $23 = $fname;
   $24 = HEAP8[$23>>0]|0;
   $25 = $24 << 24 >> 24;
   $26 = ($25|0)==(0);
   if ($26) {
    $retval = 1;
    break;
   }
   HEAP32[$fileExists>>2] = 0;
   $27 = HEAP32[183070]|0;
   (___PHYSFS_platformGrabMutex($27)|0);
   $28 = HEAP32[183076]|0;
   $i = $28;
   while(1) {
    $29 = $i;
    $30 = ($29|0)!=(0|0);
    if (!($30)) {
     break;
    }
    $31 = HEAP32[$fileExists>>2]|0;
    $32 = ($31|0)!=(0);
    $33 = $32 ^ 1;
    if (!($33)) {
     break;
    }
    $34 = $fname;
    HEAP32[$arcfname>>2] = $34;
    $35 = $i;
    $36 = HEAP32[$arcfname>>2]|0;
    $37 = (_partOfMountPoint($35,$36)|0);
    HEAP32[$fileExists>>2] = $37;
    $38 = ($37|0)!=(0);
    if ($38) {
     $retval = 0;
    } else {
     $39 = $i;
     $40 = (_verifyPath($39,$arcfname,0)|0);
     $41 = ($40|0)!=(0);
     if ($41) {
      $42 = $i;
      $43 = ((($42)) + 12|0);
      $44 = HEAP32[$43>>2]|0;
      $45 = ((($44)) + 24|0);
      $46 = HEAP32[$45>>2]|0;
      $47 = $i;
      $48 = HEAP32[$47>>2]|0;
      $49 = HEAP32[$arcfname>>2]|0;
      $50 = (FUNCTION_TABLE_iiii[$46 & 8191]($48,$49,$fileExists)|0);
      $retval = $50;
     }
    }
    $51 = $i;
    $52 = ((($51)) + 16|0);
    $53 = HEAP32[$52>>2]|0;
    $i = $53;
   }
   $54 = HEAP32[183070]|0;
   ___PHYSFS_platformReleaseMutex($54);
  } else {
   $retval = 0;
  }
 } while(0);
 $55 = $fname;
 ___PHYSFS_smallFree($55);
 $56 = $retval;
 $0 = $56;
 $57 = $0;
 STACKTOP = sp;return ($57|0);
}
function _PHYSFS_openWrite($filename) {
 $filename = $filename|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $filename;
 $1 = $0;
 $2 = (_doOpenWrite($1,0)|0);
 STACKTOP = sp;return ($2|0);
}
function _doOpenWrite($_fname,$appending) {
 $_fname = $_fname|0;
 $appending = $appending|0;
 var $$alloca_mul = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $8 = 0, $9 = 0, $f = 0, $fh = 0, $fname = 0;
 var $h = 0, $len = 0, $opaque = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $fname = sp + 12|0;
 $1 = $_fname;
 $2 = $appending;
 $fh = 0;
 $3 = $1;
 $4 = ($3|0)==(0|0);
 if ($4) {
  ___PHYSFS_setError(517147);
  $0 = 0;
  $75 = $0;
  STACKTOP = sp;return ($75|0);
 }
 $5 = $1;
 $6 = (_strlen($5)|0);
 $7 = (($6) + 1)|0;
 $len = $7;
 $8 = $len;
 $9 = ($8>>>0)<(128);
 if ($9) {
  $10 = $len;
  $11 = (($10) + 1)|0;
  $$alloca_mul = $11;
  $12 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
  $14 = $12;
 } else {
  $14 = 0;
 }
 $13 = $len;
 $15 = (___PHYSFS_initSmallAlloc($14,$13,0)|0);
 HEAP32[$fname>>2] = $15;
 $16 = HEAP32[$fname>>2]|0;
 $17 = ($16|0)==(0|0);
 if ($17) {
  ___PHYSFS_setError(596704);
  $0 = 0;
  $75 = $0;
  STACKTOP = sp;return ($75|0);
 }
 $18 = $1;
 $19 = HEAP32[$fname>>2]|0;
 $20 = (_sanitizePlatformIndependentPath($18,$19)|0);
 $21 = ($20|0)!=(0);
 if ($21) {
  $opaque = 0;
  $h = 0;
  $22 = HEAP32[183070]|0;
  (___PHYSFS_platformGrabMutex($22)|0);
  $23 = HEAP32[183074]|0;
  $24 = ($23|0)!=(0|0);
  do {
   if ($24) {
    $25 = HEAP32[183074]|0;
    $h = $25;
    $26 = $h;
    $27 = (_verifyPath($26,$fname,0)|0);
    $28 = ($27|0)!=(0);
    if (!($28)) {
     ___PHYSFS_setError(0);
     break;
    }
    $29 = $h;
    $30 = ((($29)) + 12|0);
    $31 = HEAP32[$30>>2]|0;
    $f = $31;
    $32 = $2;
    $33 = ($32|0)!=(0);
    $34 = $f;
    if ($33) {
     $35 = ((($34)) + 40|0);
     $36 = HEAP32[$35>>2]|0;
     $37 = $h;
     $38 = HEAP32[$37>>2]|0;
     $39 = HEAP32[$fname>>2]|0;
     $40 = (FUNCTION_TABLE_iii[$36 & 8191]($38,$39)|0);
     $opaque = $40;
    } else {
     $41 = ((($34)) + 36|0);
     $42 = HEAP32[$41>>2]|0;
     $43 = $h;
     $44 = HEAP32[$43>>2]|0;
     $45 = HEAP32[$fname>>2]|0;
     $46 = (FUNCTION_TABLE_iii[$42 & 8191]($44,$45)|0);
     $opaque = $46;
    }
    $47 = $opaque;
    $48 = ($47|0)==(0|0);
    if ($48) {
     ___PHYSFS_setError(0);
     break;
    }
    $49 = HEAP32[(732252)>>2]|0;
    $50 = (FUNCTION_TABLE_iii[$49 & 8191](36,0)|0);
    $fh = $50;
    $51 = $fh;
    $52 = ($51|0)==(0|0);
    if ($52) {
     $53 = $f;
     $54 = ((($53)) + 80|0);
     $55 = HEAP32[$54>>2]|0;
     $56 = $opaque;
     (FUNCTION_TABLE_ii[$55 & 8191]($56)|0);
     ___PHYSFS_setError(596704);
     break;
    } else {
     $57 = $fh;
     dest=$57; stop=dest+36|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
     $58 = $opaque;
     $59 = $fh;
     HEAP32[$59>>2] = $58;
     $60 = $h;
     $61 = $fh;
     $62 = ((($61)) + 8|0);
     HEAP32[$62>>2] = $60;
     $63 = $h;
     $64 = ((($63)) + 12|0);
     $65 = HEAP32[$64>>2]|0;
     $66 = $fh;
     $67 = ((($66)) + 12|0);
     HEAP32[$67>>2] = $65;
     $68 = HEAP32[183073]|0;
     $69 = $fh;
     $70 = ((($69)) + 32|0);
     HEAP32[$70>>2] = $68;
     $71 = $fh;
     HEAP32[183073] = $71;
     break;
    }
   } else {
    ___PHYSFS_setError(516751);
   }
  } while(0);
  $72 = HEAP32[183070]|0;
  ___PHYSFS_platformReleaseMutex($72);
 }
 $73 = HEAP32[$fname>>2]|0;
 ___PHYSFS_smallFree($73);
 $74 = $fh;
 $0 = $74;
 $75 = $0;
 STACKTOP = sp;return ($75|0);
}
function _PHYSFS_openAppend($filename) {
 $filename = $filename|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $filename;
 $1 = $0;
 $2 = (_doOpenWrite($1,1)|0);
 STACKTOP = sp;return ($2|0);
}
function _PHYSFS_openRead($_fname) {
 $_fname = $_fname|0;
 var $$alloca_mul = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $9 = 0, $arcfname = 0, $fh = 0, $fileExists = 0, $fname = 0, $i = 0, $len = 0, $opaque = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $fileExists = sp + 12|0;
 $arcfname = sp;
 $1 = $_fname;
 $fh = 0;
 $2 = $1;
 $3 = ($2|0)==(0|0);
 if ($3) {
  ___PHYSFS_setError(517147);
  $0 = 0;
  $80 = $0;
  STACKTOP = sp;return ($80|0);
 }
 $4 = $1;
 $5 = (_strlen($4)|0);
 $6 = (($5) + 1)|0;
 $len = $6;
 $7 = $len;
 $8 = ($7>>>0)<(128);
 if ($8) {
  $9 = $len;
  $10 = (($9) + 1)|0;
  $$alloca_mul = $10;
  $11 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
  $13 = $11;
 } else {
  $13 = 0;
 }
 $12 = $len;
 $14 = (___PHYSFS_initSmallAlloc($13,$12,0)|0);
 $fname = $14;
 $15 = $fname;
 $16 = ($15|0)==(0|0);
 if ($16) {
  ___PHYSFS_setError(596704);
  $0 = 0;
  $80 = $0;
  STACKTOP = sp;return ($80|0);
 }
 $17 = $1;
 $18 = $fname;
 $19 = (_sanitizePlatformIndependentPath($17,$18)|0);
 $20 = ($19|0)!=(0);
 if ($20) {
  HEAP32[$fileExists>>2] = 0;
  $i = 0;
  $opaque = 0;
  $21 = HEAP32[183070]|0;
  (___PHYSFS_platformGrabMutex($21)|0);
  $22 = HEAP32[183076]|0;
  $23 = ($22|0)!=(0|0);
  do {
   if ($23) {
    $24 = HEAP32[183076]|0;
    $i = $24;
    while(1) {
     $25 = $fname;
     HEAP32[$arcfname>>2] = $25;
     $26 = $i;
     $27 = (_verifyPath($26,$arcfname,0)|0);
     $28 = ($27|0)!=(0);
     if ($28) {
      $29 = $i;
      $30 = ((($29)) + 12|0);
      $31 = HEAP32[$30>>2]|0;
      $32 = ((($31)) + 32|0);
      $33 = HEAP32[$32>>2]|0;
      $34 = $i;
      $35 = HEAP32[$34>>2]|0;
      $36 = HEAP32[$arcfname>>2]|0;
      $37 = (FUNCTION_TABLE_iiii[$33 & 8191]($35,$36,$fileExists)|0);
      $opaque = $37;
      $38 = $opaque;
      $39 = ($38|0)!=(0|0);
      if ($39) {
       break;
      }
     }
     $40 = $i;
     $41 = ((($40)) + 16|0);
     $42 = HEAP32[$41>>2]|0;
     $i = $42;
     $43 = $i;
     $44 = ($43|0)!=(0|0);
     if (!($44)) {
      break;
     }
     $45 = HEAP32[$fileExists>>2]|0;
     $46 = ($45|0)!=(0);
     $47 = $46 ^ 1;
     if (!($47)) {
      break;
     }
    }
    $48 = $opaque;
    $49 = ($48|0)==(0|0);
    if ($49) {
     ___PHYSFS_setError(0);
     break;
    }
    $50 = HEAP32[(732252)>>2]|0;
    $51 = (FUNCTION_TABLE_iii[$50 & 8191](36,0)|0);
    $fh = $51;
    $52 = $fh;
    $53 = ($52|0)==(0|0);
    if ($53) {
     $54 = $i;
     $55 = ((($54)) + 12|0);
     $56 = HEAP32[$55>>2]|0;
     $57 = ((($56)) + 80|0);
     $58 = HEAP32[$57>>2]|0;
     $59 = $opaque;
     (FUNCTION_TABLE_ii[$58 & 8191]($59)|0);
     ___PHYSFS_setError(596704);
     break;
    } else {
     $60 = $fh;
     dest=$60; stop=dest+36|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
     $61 = $opaque;
     $62 = $fh;
     HEAP32[$62>>2] = $61;
     $63 = $fh;
     $64 = ((($63)) + 4|0);
     HEAP8[$64>>0] = 1;
     $65 = $i;
     $66 = $fh;
     $67 = ((($66)) + 8|0);
     HEAP32[$67>>2] = $65;
     $68 = $i;
     $69 = ((($68)) + 12|0);
     $70 = HEAP32[$69>>2]|0;
     $71 = $fh;
     $72 = ((($71)) + 12|0);
     HEAP32[$72>>2] = $70;
     $73 = HEAP32[183075]|0;
     $74 = $fh;
     $75 = ((($74)) + 32|0);
     HEAP32[$75>>2] = $73;
     $76 = $fh;
     HEAP32[183075] = $76;
     break;
    }
   } else {
    ___PHYSFS_setError(516803);
   }
  } while(0);
  $77 = HEAP32[183070]|0;
  ___PHYSFS_platformReleaseMutex($77);
 }
 $78 = $fname;
 ___PHYSFS_smallFree($78);
 $79 = $fh;
 $0 = $79;
 $80 = $0;
 STACKTOP = sp;return ($80|0);
}
function _PHYSFS_close($_handle) {
 $_handle = $_handle|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $handle = 0, $rc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $_handle;
 $2 = $1;
 $handle = $2;
 $3 = HEAP32[183070]|0;
 (___PHYSFS_platformGrabMutex($3)|0);
 $4 = $handle;
 $5 = (_closeHandleInOpenList(732300,$4)|0);
 $rc = $5;
 $6 = $rc;
 $7 = ($6|0)==(-1);
 if ($7) {
  ___PHYSFS_setError(0);
  $8 = HEAP32[183070]|0;
  ___PHYSFS_platformReleaseMutex($8);
  $0 = 0;
  $19 = $0;
  STACKTOP = sp;return ($19|0);
 }
 $9 = $rc;
 $10 = ($9|0)!=(0);
 if (!($10)) {
  $11 = $handle;
  $12 = (_closeHandleInOpenList(732292,$11)|0);
  $rc = $12;
  $13 = $rc;
  $14 = ($13|0)==(-1);
  if ($14) {
   ___PHYSFS_setError(0);
   $15 = HEAP32[183070]|0;
   ___PHYSFS_platformReleaseMutex($15);
   $0 = 0;
   $19 = $0;
   STACKTOP = sp;return ($19|0);
  }
 }
 $16 = HEAP32[183070]|0;
 ___PHYSFS_platformReleaseMutex($16);
 $17 = $rc;
 $18 = ($17|0)!=(0);
 if ($18) {
  $0 = 1;
  $19 = $0;
  STACKTOP = sp;return ($19|0);
 } else {
  ___PHYSFS_setError(516870);
  $0 = 0;
  $19 = $0;
  STACKTOP = sp;return ($19|0);
 }
 return (0)|0;
}
function _closeHandleInOpenList($list,$handle) {
 $list = $list|0;
 $handle = $handle|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0, $prev = 0, $rc = 0, $tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $list;
 $2 = $handle;
 $prev = 0;
 $rc = 1;
 $3 = $1;
 $4 = HEAP32[$3>>2]|0;
 $i = $4;
 while(1) {
  $5 = $i;
  $6 = ($5|0)!=(0|0);
  if (!($6)) {
   label = 15;
   break;
  }
  $7 = $i;
  $8 = $2;
  $9 = ($7|0)==($8|0);
  if ($9) {
   break;
  }
  $41 = $i;
  $prev = $41;
  $42 = $i;
  $43 = ((($42)) + 32|0);
  $44 = HEAP32[$43>>2]|0;
  $i = $44;
 }
 if ((label|0) == 15) {
  $0 = 0;
  $45 = $0;
  STACKTOP = sp;return ($45|0);
 }
 $10 = $2;
 $11 = ((($10)) + 16|0);
 $12 = HEAP32[$11>>2]|0;
 $tmp = $12;
 $13 = $2;
 $14 = (_PHYSFS_flush($13)|0);
 $rc = $14;
 $15 = $rc;
 $16 = ($15|0)!=(0);
 if ($16) {
  $17 = $2;
  $18 = ((($17)) + 12|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = ((($19)) + 80|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = $2;
  $23 = HEAP32[$22>>2]|0;
  $24 = (FUNCTION_TABLE_ii[$21 & 8191]($23)|0);
  $rc = $24;
 }
 $25 = $rc;
 $26 = ($25|0)!=(0);
 if (!($26)) {
  $0 = -1;
  $45 = $0;
  STACKTOP = sp;return ($45|0);
 }
 $27 = $tmp;
 $28 = ($27|0)!=(0|0);
 if ($28) {
  $29 = HEAP32[(732260)>>2]|0;
  $30 = $tmp;
  FUNCTION_TABLE_vi[$29 & 8191]($30);
 }
 $31 = $prev;
 $32 = ($31|0)==(0|0);
 $33 = $2;
 $34 = ((($33)) + 32|0);
 $35 = HEAP32[$34>>2]|0;
 if ($32) {
  $36 = $1;
  HEAP32[$36>>2] = $35;
 } else {
  $37 = $prev;
  $38 = ((($37)) + 32|0);
  HEAP32[$38>>2] = $35;
 }
 $39 = HEAP32[(732260)>>2]|0;
 $40 = $2;
 FUNCTION_TABLE_vi[$39 & 8191]($40);
 $0 = 1;
 $45 = $0;
 STACKTOP = sp;return ($45|0);
}
function _PHYSFS_flush($handle) {
 $handle = $handle|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fh = 0, $rc = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $rc = sp;
 $1 = $handle;
 $2 = $1;
 $fh = $2;
 $3 = $fh;
 $4 = ((($3)) + 4|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = ($6|0)!=(0);
 if (!($7)) {
  $8 = $fh;
  $9 = ((($8)) + 28|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = $fh;
  $12 = ((($11)) + 24|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ($10|0)==($13|0);
  if (!($14)) {
   $15 = $fh;
   $16 = ((($15)) + 12|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($17)) + 60|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = $fh;
   $21 = HEAP32[$20>>2]|0;
   $22 = $fh;
   $23 = ((($22)) + 16|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = $fh;
   $26 = ((($25)) + 28|0);
   $27 = HEAP32[$26>>2]|0;
   $28 = (($24) + ($27)|0);
   $29 = $fh;
   $30 = ((($29)) + 24|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = $fh;
   $33 = ((($32)) + 28|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = (($31) - ($34))|0;
   $36 = (FUNCTION_TABLE_iiiii[$19 & 8191]($21,$28,$35,1)|0);
   $37 = tempRet0;
   $38 = $rc;
   $39 = $38;
   HEAP32[$39>>2] = $36;
   $40 = (($38) + 4)|0;
   $41 = $40;
   HEAP32[$41>>2] = $37;
   $42 = $rc;
   $43 = $42;
   $44 = HEAP32[$43>>2]|0;
   $45 = (($42) + 4)|0;
   $46 = $45;
   $47 = HEAP32[$46>>2]|0;
   $48 = ($47|0)<(0);
   $49 = ($44>>>0)<=(0);
   $50 = ($47|0)==(0);
   $51 = $50 & $49;
   $52 = $48 | $51;
   if ($52) {
    ___PHYSFS_setError(0);
    $0 = 0;
    $57 = $0;
    STACKTOP = sp;return ($57|0);
   } else {
    $53 = $fh;
    $54 = ((($53)) + 24|0);
    HEAP32[$54>>2] = 0;
    $55 = $fh;
    $56 = ((($55)) + 28|0);
    HEAP32[$56>>2] = 0;
    $0 = 1;
    $57 = $0;
    STACKTOP = sp;return ($57|0);
   }
  }
 }
 $0 = 1;
 $57 = $0;
 STACKTOP = sp;return ($57|0);
}
function _PHYSFS_read($handle,$buffer,$objSize,$objCount) {
 $handle = $handle|0;
 $buffer = $buffer|0;
 $objSize = $objSize|0;
 $objCount = $objCount|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $fh = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = $handle;
 $2 = $buffer;
 $3 = $objSize;
 $4 = $objCount;
 $5 = $1;
 $fh = $5;
 $6 = $fh;
 $7 = ((($6)) + 4|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = ($8<<24>>24)!=(0);
 do {
  if ($9) {
   $14 = $3;
   $15 = ($14|0)==(0);
   if ($15) {
    ___PHYSFS_setError(0);
    $16 = $0;
    $17 = $16;
    HEAP32[$17>>2] = 0;
    $18 = (($16) + 4)|0;
    $19 = $18;
    HEAP32[$19>>2] = 0;
    break;
   }
   $20 = $4;
   $21 = ($20|0)==(0);
   if ($21) {
    ___PHYSFS_setError(0);
    $22 = $0;
    $23 = $22;
    HEAP32[$23>>2] = 0;
    $24 = (($22) + 4)|0;
    $25 = $24;
    HEAP32[$25>>2] = 0;
    break;
   }
   $26 = $fh;
   $27 = ((($26)) + 16|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = ($28|0)!=(0|0);
   $30 = $fh;
   if ($29) {
    $31 = $2;
    $32 = $3;
    $33 = $4;
    $34 = (_doBufferedRead($30,$31,$32,$33)|0);
    $35 = tempRet0;
    $36 = $0;
    $37 = $36;
    HEAP32[$37>>2] = $34;
    $38 = (($36) + 4)|0;
    $39 = $38;
    HEAP32[$39>>2] = $35;
    break;
   } else {
    $40 = ((($30)) + 12|0);
    $41 = HEAP32[$40>>2]|0;
    $42 = ((($41)) + 56|0);
    $43 = HEAP32[$42>>2]|0;
    $44 = $fh;
    $45 = HEAP32[$44>>2]|0;
    $46 = $2;
    $47 = $3;
    $48 = $4;
    $49 = (FUNCTION_TABLE_iiiii[$43 & 8191]($45,$46,$47,$48)|0);
    $50 = tempRet0;
    $51 = $0;
    $52 = $51;
    HEAP32[$52>>2] = $49;
    $53 = (($51) + 4)|0;
    $54 = $53;
    HEAP32[$54>>2] = $50;
    break;
   }
  } else {
   ___PHYSFS_setError(516888);
   $10 = $0;
   $11 = $10;
   HEAP32[$11>>2] = -1;
   $12 = (($10) + 4)|0;
   $13 = $12;
   HEAP32[$13>>2] = -1;
  }
 } while(0);
 $55 = $0;
 $56 = $55;
 $57 = HEAP32[$56>>2]|0;
 $58 = (($55) + 4)|0;
 $59 = $58;
 $60 = HEAP32[$59>>2]|0;
 tempRet0 = ($60);
 STACKTOP = sp;return ($57|0);
}
function _doBufferedRead($fh,$buffer,$objSize,$objCount) {
 $fh = $fh|0;
 $buffer = $buffer|0;
 $objSize = $objSize|0;
 $objCount = $objCount|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $buffered = 0, $copied = 0, $mustread = 0, $or$cond = 0, $rc = 0, $remainder = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 24|0;
 $retval = sp + 16|0;
 $mustread = sp + 8|0;
 $rc = sp;
 $1 = $fh;
 $2 = $buffer;
 $3 = $objSize;
 $4 = $objCount;
 $5 = $retval;
 $6 = $5;
 HEAP32[$6>>2] = 0;
 $7 = (($5) + 4)|0;
 $8 = $7;
 HEAP32[$8>>2] = 0;
 $remainder = 0;
 while(1) {
  $9 = $4;
  $10 = ($9>>>0)>(0);
  if (!($10)) {
   label = 10;
   break;
  }
  $11 = $1;
  $12 = ((($11)) + 24|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = $1;
  $15 = ((($14)) + 28|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = (($13) - ($16))|0;
  $buffered = $17;
  $18 = $3;
  $19 = $4;
  $20 = Math_imul($18, $19)|0;
  $21 = $remainder;
  $22 = (($20) - ($21))|0;
  $23 = $mustread;
  $24 = $23;
  HEAP32[$24>>2] = $22;
  $25 = (($23) + 4)|0;
  $26 = $25;
  HEAP32[$26>>2] = 0;
  $27 = $buffered;
  $28 = ($27|0)==(0);
  if ($28) {
   $29 = $1;
   $30 = ((($29)) + 12|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = ((($31)) + 56|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = $1;
   $35 = HEAP32[$34>>2]|0;
   $36 = $1;
   $37 = ((($36)) + 16|0);
   $38 = HEAP32[$37>>2]|0;
   $39 = $1;
   $40 = ((($39)) + 20|0);
   $41 = HEAP32[$40>>2]|0;
   $42 = (FUNCTION_TABLE_iiiii[$33 & 8191]($35,$38,1,$41)|0);
   $43 = tempRet0;
   $44 = $rc;
   $45 = $44;
   HEAP32[$45>>2] = $42;
   $46 = (($44) + 4)|0;
   $47 = $46;
   HEAP32[$47>>2] = $43;
   $48 = $rc;
   $49 = $48;
   $50 = HEAP32[$49>>2]|0;
   $51 = (($48) + 4)|0;
   $52 = $51;
   $53 = HEAP32[$52>>2]|0;
   $54 = ($53|0)<(0);
   $55 = ($50>>>0)<=(0);
   $56 = ($53|0)==(0);
   $57 = $56 & $55;
   $58 = $54 | $57;
   if ($58) {
    label = 5;
    break;
   }
   $94 = $rc;
   $95 = $94;
   $96 = HEAP32[$95>>2]|0;
   $97 = (($94) + 4)|0;
   $98 = $97;
   $99 = HEAP32[$98>>2]|0;
   $100 = $1;
   $101 = ((($100)) + 24|0);
   HEAP32[$101>>2] = $96;
   $buffered = $96;
   $102 = $1;
   $103 = ((($102)) + 28|0);
   HEAP32[$103>>2] = 0;
  }
  $104 = $buffered;
  $105 = $mustread;
  $106 = $105;
  $107 = HEAP32[$106>>2]|0;
  $108 = (($105) + 4)|0;
  $109 = $108;
  $110 = HEAP32[$109>>2]|0;
  $111 = (0)>($110>>>0);
  $112 = ($104>>>0)>($107>>>0);
  $113 = (0)==($110|0);
  $114 = $113 & $112;
  $115 = $111 | $114;
  if ($115) {
   $116 = $mustread;
   $117 = $116;
   $118 = HEAP32[$117>>2]|0;
   $119 = (($116) + 4)|0;
   $120 = $119;
   $121 = HEAP32[$120>>2]|0;
   $buffered = $118;
  }
  $122 = $2;
  $123 = $1;
  $124 = ((($123)) + 16|0);
  $125 = HEAP32[$124>>2]|0;
  $126 = $1;
  $127 = ((($126)) + 28|0);
  $128 = HEAP32[$127>>2]|0;
  $129 = (($125) + ($128)|0);
  $130 = $buffered;
  _memcpy(($122|0),($129|0),($130|0))|0;
  $131 = $2;
  $132 = $buffered;
  $133 = (($131) + ($132)|0);
  $2 = $133;
  $134 = $buffered;
  $135 = $1;
  $136 = ((($135)) + 28|0);
  $137 = HEAP32[$136>>2]|0;
  $138 = (($137) + ($134))|0;
  HEAP32[$136>>2] = $138;
  $139 = $remainder;
  $140 = $buffered;
  $141 = (($140) + ($139))|0;
  $buffered = $141;
  $142 = $buffered;
  $143 = $3;
  $144 = (($142>>>0) / ($143>>>0))&-1;
  $copied = $144;
  $145 = $buffered;
  $146 = $3;
  $147 = (($145>>>0) % ($146>>>0))&-1;
  $remainder = $147;
  $148 = $copied;
  $149 = $retval;
  $150 = $149;
  $151 = HEAP32[$150>>2]|0;
  $152 = (($149) + 4)|0;
  $153 = $152;
  $154 = HEAP32[$153>>2]|0;
  $155 = (_i64Add(($151|0),($154|0),($148|0),0)|0);
  $156 = tempRet0;
  $157 = $retval;
  $158 = $157;
  HEAP32[$158>>2] = $155;
  $159 = (($157) + 4)|0;
  $160 = $159;
  HEAP32[$160>>2] = $156;
  $161 = $copied;
  $162 = $4;
  $163 = (($162) - ($161))|0;
  $4 = $163;
 }
 if ((label|0) == 5) {
  $59 = $remainder;
  $60 = $1;
  $61 = ((($60)) + 28|0);
  $62 = HEAP32[$61>>2]|0;
  $63 = (($62) - ($59))|0;
  HEAP32[$61>>2] = $63;
  $64 = $rc;
  $65 = $64;
  $66 = HEAP32[$65>>2]|0;
  $67 = (($64) + 4)|0;
  $68 = $67;
  $69 = HEAP32[$68>>2]|0;
  $70 = ($66|0)==(-1);
  $71 = ($69|0)==(-1);
  $72 = $70 & $71;
  $73 = $retval;
  $74 = $73;
  $75 = HEAP32[$74>>2]|0;
  $76 = (($73) + 4)|0;
  $77 = $76;
  $78 = HEAP32[$77>>2]|0;
  $79 = ($75|0)==(0);
  $80 = ($78|0)==(0);
  $81 = $79 & $80;
  $or$cond = $72 & $81;
  $82 = $retval;
  $83 = $82;
  $84 = HEAP32[$83>>2]|0;
  $85 = (($82) + 4)|0;
  $86 = $85;
  $87 = HEAP32[$86>>2]|0;
  $88 = $or$cond ? -1 : $84;
  $89 = $or$cond ? -1 : $87;
  $90 = $0;
  $91 = $90;
  HEAP32[$91>>2] = $88;
  $92 = (($90) + 4)|0;
  $93 = $92;
  HEAP32[$93>>2] = $89;
  $174 = $0;
  $175 = $174;
  $176 = HEAP32[$175>>2]|0;
  $177 = (($174) + 4)|0;
  $178 = $177;
  $179 = HEAP32[$178>>2]|0;
  tempRet0 = ($179);
  STACKTOP = sp;return ($176|0);
 }
 else if ((label|0) == 10) {
  $164 = $retval;
  $165 = $164;
  $166 = HEAP32[$165>>2]|0;
  $167 = (($164) + 4)|0;
  $168 = $167;
  $169 = HEAP32[$168>>2]|0;
  $170 = $0;
  $171 = $170;
  HEAP32[$171>>2] = $166;
  $172 = (($170) + 4)|0;
  $173 = $172;
  HEAP32[$173>>2] = $169;
  $174 = $0;
  $175 = $174;
  $176 = HEAP32[$175>>2]|0;
  $177 = (($174) + 4)|0;
  $178 = $177;
  $179 = HEAP32[$178>>2]|0;
  tempRet0 = ($179);
  STACKTOP = sp;return ($176|0);
 }
 return (0)|0;
}
function _PHYSFS_write($handle,$buffer,$objSize,$objCount) {
 $handle = $handle|0;
 $buffer = $buffer|0;
 $objSize = $objSize|0;
 $objCount = $objCount|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $fh = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = $handle;
 $2 = $buffer;
 $3 = $objSize;
 $4 = $objCount;
 $5 = $1;
 $fh = $5;
 $6 = $fh;
 $7 = ((($6)) + 4|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = ($8<<24>>24)!=(0);
 do {
  if ($9) {
   ___PHYSFS_setError(516918);
   $10 = $0;
   $11 = $10;
   HEAP32[$11>>2] = -1;
   $12 = (($10) + 4)|0;
   $13 = $12;
   HEAP32[$13>>2] = -1;
  } else {
   $14 = $3;
   $15 = ($14|0)==(0);
   if ($15) {
    ___PHYSFS_setError(0);
    $16 = $0;
    $17 = $16;
    HEAP32[$17>>2] = 0;
    $18 = (($16) + 4)|0;
    $19 = $18;
    HEAP32[$19>>2] = 0;
    break;
   }
   $20 = $4;
   $21 = ($20|0)==(0);
   if ($21) {
    ___PHYSFS_setError(0);
    $22 = $0;
    $23 = $22;
    HEAP32[$23>>2] = 0;
    $24 = (($22) + 4)|0;
    $25 = $24;
    HEAP32[$25>>2] = 0;
    break;
   }
   $26 = $fh;
   $27 = ((($26)) + 16|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = ($28|0)!=(0|0);
   if ($29) {
    $30 = $1;
    $31 = $2;
    $32 = $3;
    $33 = $4;
    $34 = (_doBufferedWrite($30,$31,$32,$33)|0);
    $35 = tempRet0;
    $36 = $0;
    $37 = $36;
    HEAP32[$37>>2] = $34;
    $38 = (($36) + 4)|0;
    $39 = $38;
    HEAP32[$39>>2] = $35;
    break;
   } else {
    $40 = $fh;
    $41 = ((($40)) + 12|0);
    $42 = HEAP32[$41>>2]|0;
    $43 = ((($42)) + 60|0);
    $44 = HEAP32[$43>>2]|0;
    $45 = $fh;
    $46 = HEAP32[$45>>2]|0;
    $47 = $2;
    $48 = $3;
    $49 = $4;
    $50 = (FUNCTION_TABLE_iiiii[$44 & 8191]($46,$47,$48,$49)|0);
    $51 = tempRet0;
    $52 = $0;
    $53 = $52;
    HEAP32[$53>>2] = $50;
    $54 = (($52) + 4)|0;
    $55 = $54;
    HEAP32[$55>>2] = $51;
    break;
   }
  }
 } while(0);
 $56 = $0;
 $57 = $56;
 $58 = HEAP32[$57>>2]|0;
 $59 = (($56) + 4)|0;
 $60 = $59;
 $61 = HEAP32[$60>>2]|0;
 tempRet0 = ($61);
 STACKTOP = sp;return ($58|0);
}
function _doBufferedWrite($handle,$buffer,$objSize,$objCount) {
 $handle = $handle|0;
 $buffer = $buffer|0;
 $objSize = $objSize|0;
 $objCount = $objCount|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $fh = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = $handle;
 $2 = $buffer;
 $3 = $objSize;
 $4 = $objCount;
 $5 = $1;
 $fh = $5;
 $6 = $fh;
 $7 = ((($6)) + 24|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $3;
 $10 = $4;
 $11 = Math_imul($9, $10)|0;
 $12 = (($8) + ($11))|0;
 $13 = $fh;
 $14 = ((($13)) + 20|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($12>>>0)<($15>>>0);
 do {
  if ($16) {
   $17 = $fh;
   $18 = ((($17)) + 16|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = $fh;
   $21 = ((($20)) + 24|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = (($19) + ($22)|0);
   $24 = $2;
   $25 = $3;
   $26 = $4;
   $27 = Math_imul($25, $26)|0;
   _memcpy(($23|0),($24|0),($27|0))|0;
   $28 = $3;
   $29 = $4;
   $30 = Math_imul($28, $29)|0;
   $31 = $fh;
   $32 = ((($31)) + 24|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = (($33) + ($30))|0;
   HEAP32[$32>>2] = $34;
   $35 = $4;
   $36 = $0;
   $37 = $36;
   HEAP32[$37>>2] = $35;
   $38 = (($36) + 4)|0;
   $39 = $38;
   HEAP32[$39>>2] = 0;
  } else {
   $40 = $1;
   $41 = (_PHYSFS_flush($40)|0);
   $42 = ($41|0)!=(0);
   if ($42) {
    $47 = $fh;
    $48 = ((($47)) + 12|0);
    $49 = HEAP32[$48>>2]|0;
    $50 = ((($49)) + 60|0);
    $51 = HEAP32[$50>>2]|0;
    $52 = $fh;
    $53 = HEAP32[$52>>2]|0;
    $54 = $2;
    $55 = $3;
    $56 = $4;
    $57 = (FUNCTION_TABLE_iiiii[$51 & 8191]($53,$54,$55,$56)|0);
    $58 = tempRet0;
    $59 = $0;
    $60 = $59;
    HEAP32[$60>>2] = $57;
    $61 = (($59) + 4)|0;
    $62 = $61;
    HEAP32[$62>>2] = $58;
    break;
   } else {
    ___PHYSFS_setError(0);
    $43 = $0;
    $44 = $43;
    HEAP32[$44>>2] = -1;
    $45 = (($43) + 4)|0;
    $46 = $45;
    HEAP32[$46>>2] = -1;
    break;
   }
  }
 } while(0);
 $63 = $0;
 $64 = $63;
 $65 = HEAP32[$64>>2]|0;
 $66 = (($63) + 4)|0;
 $67 = $66;
 $68 = HEAP32[$67>>2]|0;
 tempRet0 = ($68);
 STACKTOP = sp;return ($65|0);
}
function _PHYSFS_eof($handle) {
 $handle = $handle|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fh = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $handle;
 $2 = $1;
 $fh = $2;
 $3 = $fh;
 $4 = ((($3)) + 4|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ($5<<24>>24)!=(0);
 if (!($6)) {
  $0 = 0;
  $25 = $0;
  STACKTOP = sp;return ($25|0);
 }
 $7 = $fh;
 $8 = ((($7)) + 28|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $fh;
 $11 = ((($10)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($9|0)==($12|0);
 if ($13) {
  $14 = $fh;
  $15 = ((($14)) + 12|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ((($16)) + 64|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = $fh;
  $20 = HEAP32[$19>>2]|0;
  $21 = (FUNCTION_TABLE_ii[$18 & 8191]($20)|0);
  $22 = ($21|0)!=(0);
  $24 = $22;
 } else {
  $24 = 0;
 }
 $23 = $24&1;
 $0 = $23;
 $25 = $0;
 STACKTOP = sp;return ($25|0);
}
function _PHYSFS_tell($handle) {
 $handle = $handle|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fh = 0, $pos = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $pos = sp + 8|0;
 $retval = sp;
 $0 = $handle;
 $1 = $0;
 $fh = $1;
 $2 = $fh;
 $3 = ((($2)) + 12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 68|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $fh;
 $8 = HEAP32[$7>>2]|0;
 $9 = (FUNCTION_TABLE_ii[$6 & 8191]($8)|0);
 $10 = tempRet0;
 $11 = $pos;
 $12 = $11;
 HEAP32[$12>>2] = $9;
 $13 = (($11) + 4)|0;
 $14 = $13;
 HEAP32[$14>>2] = $10;
 $15 = $fh;
 $16 = ((($15)) + 4|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = $17&255;
 $19 = ($18|0)!=(0);
 $20 = $pos;
 $21 = $20;
 $22 = HEAP32[$21>>2]|0;
 $23 = (($20) + 4)|0;
 $24 = $23;
 $25 = HEAP32[$24>>2]|0;
 $26 = $fh;
 $27 = ((($26)) + 24|0);
 $28 = HEAP32[$27>>2]|0;
 if ($19) {
  $29 = (_i64Subtract(($22|0),($25|0),($28|0),0)|0);
  $30 = tempRet0;
  $31 = $fh;
  $32 = ((($31)) + 28|0);
  $33 = HEAP32[$32>>2]|0;
  $34 = (_i64Add(($29|0),($30|0),($33|0),0)|0);
  $35 = tempRet0;
  $40 = $34;$43 = $35;
 } else {
  $36 = (_i64Add(($22|0),($25|0),($28|0),0)|0);
  $37 = tempRet0;
  $40 = $36;$43 = $37;
 }
 $38 = $retval;
 $39 = $38;
 HEAP32[$39>>2] = $40;
 $41 = (($38) + 4)|0;
 $42 = $41;
 HEAP32[$42>>2] = $43;
 $44 = $retval;
 $45 = $44;
 $46 = HEAP32[$45>>2]|0;
 $47 = (($44) + 4)|0;
 $48 = $47;
 $49 = HEAP32[$48>>2]|0;
 tempRet0 = ($49);
 STACKTOP = sp;return ($46|0);
}
function _PHYSFS_seek($handle,$0,$1) {
 $handle = $handle|0;
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fh = 0, $offset = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $4 = sp + 8|0;
 $offset = sp;
 $3 = $handle;
 $5 = $4;
 $6 = $5;
 HEAP32[$6>>2] = $0;
 $7 = (($5) + 4)|0;
 $8 = $7;
 HEAP32[$8>>2] = $1;
 $9 = $3;
 $fh = $9;
 $10 = $3;
 $11 = (_PHYSFS_flush($10)|0);
 $12 = ($11|0)!=(0);
 if (!($12)) {
  ___PHYSFS_setError(0);
  $2 = 0;
  $117 = $2;
  STACKTOP = sp;return ($117|0);
 }
 $13 = $fh;
 $14 = ((($13)) + 16|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)!=(0|0);
 do {
  if ($16) {
   $17 = $fh;
   $18 = ((($17)) + 4|0);
   $19 = HEAP8[$18>>0]|0;
   $20 = $19&255;
   $21 = ($20|0)!=(0);
   if ($21) {
    $22 = $4;
    $23 = $22;
    $24 = HEAP32[$23>>2]|0;
    $25 = (($22) + 4)|0;
    $26 = $25;
    $27 = HEAP32[$26>>2]|0;
    $28 = $3;
    $29 = (_PHYSFS_tell($28)|0);
    $30 = tempRet0;
    $31 = (_i64Subtract(($24|0),($27|0),($29|0),($30|0))|0);
    $32 = tempRet0;
    $33 = $offset;
    $34 = $33;
    HEAP32[$34>>2] = $31;
    $35 = (($33) + 4)|0;
    $36 = $35;
    HEAP32[$36>>2] = $32;
    $37 = $offset;
    $38 = $37;
    $39 = HEAP32[$38>>2]|0;
    $40 = (($37) + 4)|0;
    $41 = $40;
    $42 = HEAP32[$41>>2]|0;
    $43 = ($42|0)>(0);
    $44 = ($39>>>0)>=(0);
    $45 = ($42|0)==(0);
    $46 = $45 & $44;
    $47 = $43 | $46;
    if ($47) {
     $48 = $offset;
     $49 = $48;
     $50 = HEAP32[$49>>2]|0;
     $51 = (($48) + 4)|0;
     $52 = $51;
     $53 = HEAP32[$52>>2]|0;
     $54 = $fh;
     $55 = ((($54)) + 24|0);
     $56 = HEAP32[$55>>2]|0;
     $57 = $fh;
     $58 = ((($57)) + 28|0);
     $59 = HEAP32[$58>>2]|0;
     $60 = (($56) - ($59))|0;
     $61 = ($53|0)<(0);
     $62 = ($50>>>0)<=($60>>>0);
     $63 = ($53|0)==(0);
     $64 = $63 & $62;
     $65 = $61 | $64;
     if (!($65)) {
      label = 7;
     }
    } else {
     label = 7;
    }
    if ((label|0) == 7) {
     $66 = $offset;
     $67 = $66;
     $68 = HEAP32[$67>>2]|0;
     $69 = (($66) + 4)|0;
     $70 = $69;
     $71 = HEAP32[$70>>2]|0;
     $72 = ($71|0)<(0);
     if (!($72)) {
      break;
     }
     $73 = $offset;
     $74 = $73;
     $75 = HEAP32[$74>>2]|0;
     $76 = (($73) + 4)|0;
     $77 = $76;
     $78 = HEAP32[$77>>2]|0;
     $79 = (_i64Subtract(0,0,($75|0),($78|0))|0);
     $80 = tempRet0;
     $81 = $fh;
     $82 = ((($81)) + 28|0);
     $83 = HEAP32[$82>>2]|0;
     $84 = ($80|0)<(0);
     $85 = ($79>>>0)<=($83>>>0);
     $86 = ($80|0)==(0);
     $87 = $86 & $85;
     $88 = $84 | $87;
     if (!($88)) {
      break;
     }
    }
    $89 = $offset;
    $90 = $89;
    $91 = HEAP32[$90>>2]|0;
    $92 = (($89) + 4)|0;
    $93 = $92;
    $94 = HEAP32[$93>>2]|0;
    $95 = $fh;
    $96 = ((($95)) + 28|0);
    $97 = HEAP32[$96>>2]|0;
    $98 = (($97) + ($91))|0;
    HEAP32[$96>>2] = $98;
    $2 = 1;
    $117 = $2;
    STACKTOP = sp;return ($117|0);
   }
  }
 } while(0);
 $99 = $fh;
 $100 = ((($99)) + 28|0);
 HEAP32[$100>>2] = 0;
 $101 = $fh;
 $102 = ((($101)) + 24|0);
 HEAP32[$102>>2] = 0;
 $103 = $fh;
 $104 = ((($103)) + 12|0);
 $105 = HEAP32[$104>>2]|0;
 $106 = ((($105)) + 72|0);
 $107 = HEAP32[$106>>2]|0;
 $108 = $fh;
 $109 = HEAP32[$108>>2]|0;
 $110 = $4;
 $111 = $110;
 $112 = HEAP32[$111>>2]|0;
 $113 = (($110) + 4)|0;
 $114 = $113;
 $115 = HEAP32[$114>>2]|0;
 $116 = (FUNCTION_TABLE_iiii[$107 & 8191]($109,$112,$115)|0);
 $2 = $116;
 $117 = $2;
 STACKTOP = sp;return ($117|0);
}
function _PHYSFS_fileLength($handle) {
 $handle = $handle|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fh = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $handle;
 $1 = $0;
 $fh = $1;
 $2 = $fh;
 $3 = ((($2)) + 12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 76|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $fh;
 $8 = HEAP32[$7>>2]|0;
 $9 = (FUNCTION_TABLE_ii[$6 & 8191]($8)|0);
 $10 = tempRet0;
 tempRet0 = ($10);
 STACKTOP = sp;return ($9|0);
}
function _PHYSFS_setBuffer($handle,$0,$1) {
 $handle = $handle|0;
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bufsize = 0, $curpos = 0, $fh = 0, $newbuf = 0, $pos = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $4 = sp + 16|0;
 $pos = sp + 8|0;
 $curpos = sp;
 $3 = $handle;
 $5 = $4;
 $6 = $5;
 HEAP32[$6>>2] = $0;
 $7 = (($5) + 4)|0;
 $8 = $7;
 HEAP32[$8>>2] = $1;
 $9 = $3;
 $fh = $9;
 $10 = $4;
 $11 = $10;
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10) + 4)|0;
 $14 = $13;
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15>>>0)>(0);
 $17 = ($12>>>0)>(4294967295);
 $18 = ($15|0)==(0);
 $19 = $18 & $17;
 $20 = $16 | $19;
 if ($20) {
  ___PHYSFS_setError(516948);
  $2 = 0;
  $129 = $2;
  STACKTOP = sp;return ($129|0);
 }
 $21 = $4;
 $22 = $21;
 $23 = HEAP32[$22>>2]|0;
 $24 = (($21) + 4)|0;
 $25 = $24;
 $26 = HEAP32[$25>>2]|0;
 $bufsize = $23;
 $27 = $3;
 $28 = (_PHYSFS_flush($27)|0);
 $29 = ($28|0)!=(0);
 if (!($29)) {
  ___PHYSFS_setError(0);
  $2 = 0;
  $129 = $2;
  STACKTOP = sp;return ($129|0);
 }
 $30 = $fh;
 $31 = ((($30)) + 4|0);
 $32 = HEAP8[$31>>0]|0;
 $33 = $32&255;
 $34 = ($33|0)!=(0);
 if ($34) {
  $35 = $fh;
  $36 = ((($35)) + 24|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = $fh;
  $39 = ((($38)) + 28|0);
  $40 = HEAP32[$39>>2]|0;
  $41 = ($37|0)!=($40|0);
  if ($41) {
   $42 = $fh;
   $43 = ((($42)) + 12|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = ((($44)) + 68|0);
   $46 = HEAP32[$45>>2]|0;
   $47 = $fh;
   $48 = HEAP32[$47>>2]|0;
   $49 = (FUNCTION_TABLE_ii[$46 & 8191]($48)|0);
   $50 = tempRet0;
   $51 = $curpos;
   $52 = $51;
   HEAP32[$52>>2] = $49;
   $53 = (($51) + 4)|0;
   $54 = $53;
   HEAP32[$54>>2] = $50;
   $55 = $curpos;
   $56 = $55;
   $57 = HEAP32[$56>>2]|0;
   $58 = (($55) + 4)|0;
   $59 = $58;
   $60 = HEAP32[$59>>2]|0;
   $61 = ($57|0)==(-1);
   $62 = ($60|0)==(-1);
   $63 = $61 & $62;
   if ($63) {
    ___PHYSFS_setError(0);
    $2 = 0;
    $129 = $2;
    STACKTOP = sp;return ($129|0);
   }
   $64 = $curpos;
   $65 = $64;
   $66 = HEAP32[$65>>2]|0;
   $67 = (($64) + 4)|0;
   $68 = $67;
   $69 = HEAP32[$68>>2]|0;
   $70 = $fh;
   $71 = ((($70)) + 24|0);
   $72 = HEAP32[$71>>2]|0;
   $73 = (_i64Subtract(($66|0),($69|0),($72|0),0)|0);
   $74 = tempRet0;
   $75 = $fh;
   $76 = ((($75)) + 28|0);
   $77 = HEAP32[$76>>2]|0;
   $78 = (_i64Add(($73|0),($74|0),($77|0),0)|0);
   $79 = tempRet0;
   $80 = $pos;
   $81 = $80;
   HEAP32[$81>>2] = $78;
   $82 = (($80) + 4)|0;
   $83 = $82;
   HEAP32[$83>>2] = $79;
   $84 = $fh;
   $85 = ((($84)) + 12|0);
   $86 = HEAP32[$85>>2]|0;
   $87 = ((($86)) + 72|0);
   $88 = HEAP32[$87>>2]|0;
   $89 = $fh;
   $90 = HEAP32[$89>>2]|0;
   $91 = $pos;
   $92 = $91;
   $93 = HEAP32[$92>>2]|0;
   $94 = (($91) + 4)|0;
   $95 = $94;
   $96 = HEAP32[$95>>2]|0;
   $97 = (FUNCTION_TABLE_iiii[$88 & 8191]($90,$93,$96)|0);
   $98 = ($97|0)!=(0);
   if (!($98)) {
    ___PHYSFS_setError(0);
    $2 = 0;
    $129 = $2;
    STACKTOP = sp;return ($129|0);
   }
  }
 }
 $99 = $bufsize;
 $100 = ($99|0)==(0);
 do {
  if ($100) {
   $101 = $fh;
   $102 = ((($101)) + 16|0);
   $103 = HEAP32[$102>>2]|0;
   $104 = ($103|0)!=(0|0);
   if ($104) {
    $105 = HEAP32[(732260)>>2]|0;
    $106 = $fh;
    $107 = ((($106)) + 16|0);
    $108 = HEAP32[$107>>2]|0;
    FUNCTION_TABLE_vi[$105 & 8191]($108);
    $109 = $fh;
    $110 = ((($109)) + 16|0);
    HEAP32[$110>>2] = 0;
   }
  } else {
   $111 = HEAP32[(732256)>>2]|0;
   $112 = $fh;
   $113 = ((($112)) + 16|0);
   $114 = HEAP32[$113>>2]|0;
   $115 = $bufsize;
   $116 = (FUNCTION_TABLE_iiii[$111 & 8191]($114,$115,0)|0);
   $newbuf = $116;
   $117 = $newbuf;
   $118 = ($117|0)==(0|0);
   if (!($118)) {
    $119 = $newbuf;
    $120 = $fh;
    $121 = ((($120)) + 16|0);
    HEAP32[$121>>2] = $119;
    break;
   }
   ___PHYSFS_setError(596704);
   $2 = 0;
   $129 = $2;
   STACKTOP = sp;return ($129|0);
  }
 } while(0);
 $122 = $bufsize;
 $123 = $fh;
 $124 = ((($123)) + 20|0);
 HEAP32[$124>>2] = $122;
 $125 = $fh;
 $126 = ((($125)) + 28|0);
 HEAP32[$126>>2] = 0;
 $127 = $fh;
 $128 = ((($127)) + 24|0);
 HEAP32[$128>>2] = 0;
 $2 = 1;
 $129 = $2;
 STACKTOP = sp;return ($129|0);
}
function ___PHYSFS_stricmpASCII($str1,$str2) {
 $str1 = $str1|0;
 $str2 = $str2|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ch1 = 0, $ch2 = 0, $cp1 = 0, $cp2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $str1;
 $2 = $str2;
 while(1) {
  $3 = $1;
  $4 = ((($3)) + 1|0);
  $1 = $4;
  $5 = HEAP8[$3>>0]|0;
  $ch1 = $5;
  $6 = $2;
  $7 = ((($6)) + 1|0);
  $2 = $7;
  $8 = HEAP8[$6>>0]|0;
  $ch2 = $8;
  $9 = $ch1;
  $10 = $9 << 24 >> 24;
  $11 = ($10|0)>=(65);
  if ($11) {
   $12 = $ch1;
   $13 = $12 << 24 >> 24;
   $14 = ($13|0)<=(90);
   if ($14) {
    $15 = $ch1;
    $16 = $15 << 24 >> 24;
    $17 = (($16) + 32)|0;
    $21 = $17;
   } else {
    label = 5;
   }
  } else {
   label = 5;
  }
  if ((label|0) == 5) {
   label = 0;
   $18 = $ch1;
   $19 = $18 << 24 >> 24;
   $21 = $19;
  }
  $20 = $21&255;
  $cp1 = $20;
  $22 = $ch2;
  $23 = $22 << 24 >> 24;
  $24 = ($23|0)>=(65);
  if ($24) {
   $25 = $ch2;
   $26 = $25 << 24 >> 24;
   $27 = ($26|0)<=(90);
   if ($27) {
    $28 = $ch2;
    $29 = $28 << 24 >> 24;
    $30 = (($29) + 32)|0;
    $34 = $30;
   } else {
    label = 9;
   }
  } else {
   label = 9;
  }
  if ((label|0) == 9) {
   label = 0;
   $31 = $ch2;
   $32 = $31 << 24 >> 24;
   $34 = $32;
  }
  $33 = $34&255;
  $cp2 = $33;
  $35 = $cp1;
  $36 = $35 << 24 >> 24;
  $37 = $cp2;
  $38 = $37 << 24 >> 24;
  $39 = ($36|0)<($38|0);
  if ($39) {
   label = 11;
   break;
  }
  $40 = $cp1;
  $41 = $40 << 24 >> 24;
  $42 = $cp2;
  $43 = $42 << 24 >> 24;
  $44 = ($41|0)>($43|0);
  if ($44) {
   label = 13;
   break;
  }
  $45 = $cp1;
  $46 = $45 << 24 >> 24;
  $47 = ($46|0)==(0);
  if ($47) {
   label = 15;
   break;
  }
 }
 if ((label|0) == 11) {
  $0 = -1;
  $48 = $0;
  STACKTOP = sp;return ($48|0);
 }
 else if ((label|0) == 13) {
  $0 = 1;
  $48 = $0;
  STACKTOP = sp;return ($48|0);
 }
 else if ((label|0) == 15) {
  $0 = 0;
  $48 = $0;
  STACKTOP = sp;return ($48|0);
 }
 return (0)|0;
}
function ___PHYSFS_platformCopyEnvironmentVariable($varname) {
 $varname = $varname|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $envr = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $varname;
 $1 = $0;
 $2 = (_getenv(($1|0))|0);
 $envr = $2;
 $retval = 0;
 $3 = $envr;
 $4 = ($3|0)!=(0|0);
 if ($4) {
  $5 = HEAP32[(732252)>>2]|0;
  $6 = $envr;
  $7 = (_strlen($6)|0);
  $8 = (($7) + 1)|0;
  $9 = (FUNCTION_TABLE_iii[$5 & 8191]($8,0)|0);
  $retval = $9;
  $10 = $retval;
  $11 = ($10|0)!=(0|0);
  if ($11) {
   $12 = $retval;
   $13 = $envr;
   (_strcpy($12,$13)|0);
  }
 }
 $14 = $retval;
 STACKTOP = sp;return ($14|0);
}
function ___PHYSFS_platformGetUserName() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = (_getUserNameByUID()|0);
 $retval = $0;
 $1 = $retval;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $3 = (___PHYSFS_platformCopyEnvironmentVariable(516975)|0);
  $retval = $3;
 }
 $4 = $retval;
 STACKTOP = sp;return ($4|0);
}
function _getUserNameByUID() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $pw = 0, $retval = 0, $uid = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = (_getuid()|0);
 $uid = $0;
 $retval = 0;
 $1 = $uid;
 $2 = (_getpwuid(($1|0))|0);
 $pw = $2;
 $3 = $pw;
 $4 = ($3|0)!=(0|0);
 if ($4) {
  $5 = $pw;
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6|0)!=(0|0);
  if ($7) {
   $8 = HEAP32[(732252)>>2]|0;
   $9 = $pw;
   $10 = HEAP32[$9>>2]|0;
   $11 = (_strlen($10)|0);
   $12 = (($11) + 1)|0;
   $13 = (FUNCTION_TABLE_iii[$8 & 8191]($12,0)|0);
   $retval = $13;
   $14 = $retval;
   $15 = ($14|0)!=(0|0);
   if ($15) {
    $16 = $retval;
    $17 = $pw;
    $18 = HEAP32[$17>>2]|0;
    (_strcpy($16,$18)|0);
   }
  }
 }
 $19 = $retval;
 STACKTOP = sp;return ($19|0);
}
function ___PHYSFS_platformGetUserDir() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $retval = 0, $statbuf = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $statbuf = sp;
 $0 = (___PHYSFS_platformCopyEnvironmentVariable(516980)|0);
 $retval = $0;
 $1 = $retval;
 $2 = ($1|0)!=(0|0);
 do {
  if ($2) {
   $3 = $retval;
   $4 = (_stat($3,$statbuf)|0);
   $5 = ($4|0)==(-1);
   if (!($5)) {
    $6 = ((($statbuf)) + 12|0);
    $7 = HEAP32[$6>>2]|0;
    $8 = $7 & 61440;
    $9 = ($8|0)==(16384);
    $10 = $9&1;
    $11 = ($10|0)==(0);
    if (!($11)) {
     break;
    }
   }
   $12 = HEAP32[(732260)>>2]|0;
   $13 = $retval;
   FUNCTION_TABLE_vi[$12 & 8191]($13);
   $retval = 0;
  }
 } while(0);
 $14 = $retval;
 $15 = ($14|0)==(0|0);
 if (!($15)) {
  $17 = $retval;
  STACKTOP = sp;return ($17|0);
 }
 $16 = (_getUserDirByUID()|0);
 $retval = $16;
 $17 = $retval;
 STACKTOP = sp;return ($17|0);
}
function _getUserDirByUID() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $pw = 0, $retval = 0, $uid = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = (_getuid()|0);
 $uid = $0;
 $retval = 0;
 $1 = $uid;
 $2 = (_getpwuid(($1|0))|0);
 $pw = $2;
 $3 = $pw;
 $4 = ($3|0)!=(0|0);
 if ($4) {
  $5 = $pw;
  $6 = ((($5)) + 20|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)!=(0|0);
  if ($8) {
   $9 = HEAP32[(732252)>>2]|0;
   $10 = $pw;
   $11 = ((($10)) + 20|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = (_strlen($12)|0);
   $14 = (($13) + 1)|0;
   $15 = (FUNCTION_TABLE_iii[$9 & 8191]($14,0)|0);
   $retval = $15;
   $16 = $retval;
   $17 = ($16|0)!=(0|0);
   if ($17) {
    $18 = $retval;
    $19 = $pw;
    $20 = ((($19)) + 20|0);
    $21 = HEAP32[$20>>2]|0;
    (_strcpy($18,$21)|0);
   }
  }
 }
 $22 = $retval;
 STACKTOP = sp;return ($22|0);
}
function ___PHYSFS_platformExists($fname) {
 $fname = $fname|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $statbuf = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $statbuf = sp;
 $1 = $fname;
 $2 = $1;
 $3 = (_lstat($2,$statbuf)|0);
 $4 = ($3|0)==(-1);
 if ($4) {
  $5 = (___errno_location()|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (_strerror($6)|0);
  ___PHYSFS_setError($7);
  $0 = 0;
  $8 = $0;
  STACKTOP = sp;return ($8|0);
 } else {
  $0 = 1;
  $8 = $0;
  STACKTOP = sp;return ($8|0);
 }
 return (0)|0;
}
function ___PHYSFS_platformIsSymLink($fname) {
 $fname = $fname|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $statbuf = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $statbuf = sp;
 $1 = $fname;
 $2 = $1;
 $3 = (_lstat($2,$statbuf)|0);
 $4 = ($3|0)==(-1);
 if ($4) {
  $5 = (___errno_location()|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (_strerror($6)|0);
  ___PHYSFS_setError($7);
  $0 = 0;
  $13 = $0;
  STACKTOP = sp;return ($13|0);
 } else {
  $8 = ((($statbuf)) + 12|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = $9 & 61440;
  $11 = ($10|0)==(40960);
  $12 = $11 ? 1 : 0;
  $0 = $12;
  $13 = $0;
  STACKTOP = sp;return ($13|0);
 }
 return (0)|0;
}
function ___PHYSFS_platformIsDirectory($fname) {
 $fname = $fname|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $statbuf = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $statbuf = sp;
 $1 = $fname;
 $2 = $1;
 $3 = (_stat($2,$statbuf)|0);
 $4 = ($3|0)==(-1);
 if ($4) {
  $5 = (___errno_location()|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (_strerror($6)|0);
  ___PHYSFS_setError($7);
  $0 = 0;
  $13 = $0;
  STACKTOP = sp;return ($13|0);
 } else {
  $8 = ((($statbuf)) + 12|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = $9 & 61440;
  $11 = ($10|0)==(16384);
  $12 = $11 ? 1 : 0;
  $0 = $12;
  $13 = $0;
  STACKTOP = sp;return ($13|0);
 }
 return (0)|0;
}
function ___PHYSFS_platformCvtToDependent($prepend,$dirName,$append) {
 $prepend = $prepend|0;
 $dirName = $dirName|0;
 $append = $append|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $len = 0, $retval = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $prepend;
 $2 = $dirName;
 $3 = $append;
 $4 = $1;
 $5 = ($4|0)!=(0|0);
 if ($5) {
  $6 = $1;
  $7 = (_strlen($6)|0);
  $13 = $7;
 } else {
  $13 = 0;
 }
 $8 = $3;
 $9 = ($8|0)!=(0|0);
 if ($9) {
  $10 = $3;
  $11 = (_strlen($10)|0);
  $14 = $11;
 } else {
  $14 = 0;
 }
 $12 = (($13) + ($14))|0;
 $15 = $2;
 $16 = (_strlen($15)|0);
 $17 = (($12) + ($16))|0;
 $18 = (($17) + 1)|0;
 $len = $18;
 $19 = HEAP32[(732252)>>2]|0;
 $20 = $len;
 $21 = ($20|0)<(0);
 $22 = $21 << 31 >> 31;
 $23 = (FUNCTION_TABLE_iii[$19 & 8191]($20,$22)|0);
 $retval = $23;
 $24 = $retval;
 $25 = ($24|0)==(0|0);
 if ($25) {
  ___PHYSFS_setError(596704);
  $0 = 0;
  $37 = $0;
  STACKTOP = sp;return ($37|0);
 }
 $26 = $1;
 $27 = ($26|0)!=(0|0);
 $28 = $retval;
 if ($27) {
  $29 = $1;
  (_strcpy($28,$29)|0);
 } else {
  HEAP8[$28>>0] = 0;
 }
 $30 = $retval;
 $31 = $2;
 (_strcat($30,$31)|0);
 $32 = $3;
 $33 = ($32|0)!=(0|0);
 if ($33) {
  $34 = $retval;
  $35 = $3;
  (_strcat($34,$35)|0);
 }
 $36 = $retval;
 $0 = $36;
 $37 = $0;
 STACKTOP = sp;return ($37|0);
}
function ___PHYSFS_platformEnumerateFiles($dirname,$omitSymLinks,$callback,$origdir,$callbackdata) {
 $dirname = $dirname|0;
 $omitSymLinks = $omitSymLinks|0;
 $callback = $callback|0;
 $origdir = $origdir|0;
 $callbackdata = $callbackdata|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $9 = 0, $buf = 0, $bufsize = 0, $dir = 0, $dlen = 0, $ent = 0, $len = 0, $p = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $dirname;
 $1 = $omitSymLinks;
 $2 = $callback;
 $3 = $origdir;
 $4 = $callbackdata;
 $bufsize = 0;
 $buf = 0;
 $dlen = 0;
 $5 = $1;
 $6 = ($5|0)!=(0);
 if ($6) {
  $7 = $0;
  $8 = (_strlen($7)|0);
  $dlen = $8;
  $9 = $dlen;
  $10 = (($9) + 256)|0;
  $bufsize = $10;
  $11 = HEAP32[(732252)>>2]|0;
  $12 = $bufsize;
  $13 = ($12|0)<(0);
  $14 = $13 << 31 >> 31;
  $15 = (FUNCTION_TABLE_iii[$11 & 8191]($12,$14)|0);
  $buf = $15;
  $16 = $buf;
  $17 = ($16|0)==(0|0);
  if ($17) {
   STACKTOP = sp;return;
  }
  $18 = $buf;
  $19 = $0;
  (_strcpy($18,$19)|0);
  $20 = $dlen;
  $21 = (($20) - 1)|0;
  $22 = $buf;
  $23 = (($22) + ($21)|0);
  $24 = HEAP8[$23>>0]|0;
  $25 = $24 << 24 >> 24;
  $26 = ($25|0)!=(47);
  if ($26) {
   $27 = $dlen;
   $28 = (($27) + 1)|0;
   $dlen = $28;
   $29 = $buf;
   $30 = (($29) + ($27)|0);
   HEAP8[$30>>0] = 47;
   $31 = $dlen;
   $32 = $buf;
   $33 = (($32) + ($31)|0);
   HEAP8[$33>>0] = 0;
  }
 }
 $34 = (___errno_location()|0);
 HEAP32[$34>>2] = 0;
 $35 = $0;
 $36 = (_opendir($35)|0);
 $dir = $36;
 $37 = $dir;
 $38 = ($37|0)==(0|0);
 if ($38) {
  $39 = HEAP32[(732260)>>2]|0;
  $40 = $buf;
  FUNCTION_TABLE_vi[$39 & 8191]($40);
  STACKTOP = sp;return;
 }
 while(1) {
  $41 = $dir;
  $42 = (_readdir($41)|0);
  $ent = $42;
  $43 = ($42|0)!=(0|0);
  if (!($43)) {
   break;
  }
  $44 = $ent;
  $45 = ((($44)) + 11|0);
  $46 = (_strcmp($45,619449)|0);
  $47 = ($46|0)==(0);
  if ($47) {
   continue;
  }
  $48 = $ent;
  $49 = ((($48)) + 11|0);
  $50 = (_strcmp($49,519838)|0);
  $51 = ($50|0)==(0);
  if ($51) {
   continue;
  }
  $52 = $1;
  $53 = ($52|0)!=(0);
  if ($53) {
   $54 = $ent;
   $55 = ((($54)) + 11|0);
   $56 = (_strlen($55)|0);
   $57 = $dlen;
   $58 = (($56) + ($57))|0;
   $59 = (($58) + 1)|0;
   $len = $59;
   $60 = $len;
   $61 = $bufsize;
   $62 = ($60|0)>($61|0);
   if ($62) {
    $63 = HEAP32[(732256)>>2]|0;
    $64 = $buf;
    $65 = $len;
    $66 = ($65|0)<(0);
    $67 = $66 << 31 >> 31;
    $68 = (FUNCTION_TABLE_iiii[$63 & 8191]($64,$65,$67)|0);
    $p = $68;
    $69 = $p;
    $70 = ($69|0)==(0|0);
    if ($70) {
     continue;
    }
    $71 = $p;
    $buf = $71;
    $72 = $len;
    $bufsize = $72;
   }
   $73 = $buf;
   $74 = $dlen;
   $75 = (($73) + ($74)|0);
   $76 = $ent;
   $77 = ((($76)) + 11|0);
   (_strcpy($75,$77)|0);
   $78 = $buf;
   $79 = (___PHYSFS_platformIsSymLink($78)|0);
   $80 = ($79|0)!=(0);
   if ($80) {
    continue;
   }
  }
  $81 = $2;
  $82 = $4;
  $83 = $3;
  $84 = $ent;
  $85 = ((($84)) + 11|0);
  FUNCTION_TABLE_viii[$81 & 8191]($82,$83,$85);
 }
 $86 = HEAP32[(732260)>>2]|0;
 $87 = $buf;
 FUNCTION_TABLE_vi[$86 & 8191]($87);
 $88 = $dir;
 (_closedir($88)|0);
 STACKTOP = sp;return;
}
function ___PHYSFS_platformMkDir($path) {
 $path = $path|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $rc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $path;
 $2 = (___errno_location()|0);
 HEAP32[$2>>2] = 0;
 $3 = $1;
 $4 = (_mkdir($3,448)|0);
 $rc = $4;
 $5 = $rc;
 $6 = ($5|0)==(-1);
 if ($6) {
  $7 = (___errno_location()|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = (_strerror($8)|0);
  ___PHYSFS_setError($9);
  $0 = 0;
  $10 = $0;
  STACKTOP = sp;return ($10|0);
 } else {
  $0 = 1;
  $10 = $0;
  STACKTOP = sp;return ($10|0);
 }
 return (0)|0;
}
function ___PHYSFS_platformOpenRead($filename) {
 $filename = $filename|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $filename;
 $1 = $0;
 $2 = (_doOpen($1,0)|0);
 STACKTOP = sp;return ($2|0);
}
function _doOpen($filename,$mode) {
 $filename = $filename|0;
 $mode = $mode|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $appending = 0, $fd = 0, $retval = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $1 = $filename;
 $2 = $mode;
 $3 = $2;
 $4 = $3 & 1024;
 $appending = $4;
 $5 = (___errno_location()|0);
 HEAP32[$5>>2] = 0;
 $6 = $2;
 $7 = $6 & -1025;
 $2 = $7;
 $8 = $1;
 $9 = $2;
 HEAP32[$vararg_buffer>>2] = 384;
 $10 = (_open($8,$9,$vararg_buffer)|0);
 $fd = $10;
 $11 = $fd;
 $12 = ($11|0)<(0);
 if ($12) {
  $13 = (___errno_location()|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = (_strerror($14)|0);
  ___PHYSFS_setError($15);
  $0 = 0;
  $32 = $0;
  STACKTOP = sp;return ($32|0);
 }
 $16 = $appending;
 $17 = ($16|0)!=(0);
 if ($17) {
  $18 = $fd;
  $19 = (_lseek($18,0,2)|0);
  $20 = ($19|0)<(0);
  if ($20) {
   $21 = $fd;
   (_close($21)|0);
   $22 = (___errno_location()|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = (_strerror($23)|0);
   ___PHYSFS_setError($24);
   $0 = 0;
   $32 = $0;
   STACKTOP = sp;return ($32|0);
  }
 }
 $25 = HEAP32[(732252)>>2]|0;
 $26 = (FUNCTION_TABLE_iii[$25 & 8191](4,0)|0);
 $retval = $26;
 $27 = $retval;
 $28 = ($27|0)==(0|0);
 $29 = $fd;
 if ($28) {
  (_close($29)|0);
  ___PHYSFS_setError(596704);
  $0 = 0;
  $32 = $0;
  STACKTOP = sp;return ($32|0);
 } else {
  $30 = $retval;
  HEAP32[$30>>2] = $29;
  $31 = $retval;
  $0 = $31;
  $32 = $0;
  STACKTOP = sp;return ($32|0);
 }
 return (0)|0;
}
function ___PHYSFS_platformOpenWrite($filename) {
 $filename = $filename|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $filename;
 $1 = $0;
 $2 = (_doOpen($1,577)|0);
 STACKTOP = sp;return ($2|0);
}
function ___PHYSFS_platformOpenAppend($filename) {
 $filename = $filename|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $filename;
 $1 = $0;
 $2 = (_doOpen($1,1089)|0);
 STACKTOP = sp;return ($2|0);
}
function ___PHYSFS_platformRead($opaque,$buffer,$size,$count) {
 $opaque = $opaque|0;
 $buffer = $buffer|0;
 $size = $size|0;
 $count = $count|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fd = 0, $max = 0, $or$cond = 0, $rc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = $opaque;
 $2 = $buffer;
 $3 = $size;
 $4 = $count;
 $5 = $1;
 $6 = HEAP32[$5>>2]|0;
 $fd = $6;
 $7 = $3;
 $8 = $4;
 $9 = Math_imul($7, $8)|0;
 $max = $9;
 $10 = $fd;
 $11 = $2;
 $12 = $max;
 $13 = (_read($10,$11,$12)|0);
 $rc = $13;
 $14 = $rc;
 $15 = ($14|0)==(-1);
 if ($15) {
  $16 = (___errno_location()|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = (_strerror($17)|0);
  ___PHYSFS_setError($18);
  $19 = $rc;
  $20 = ($19|0)<(0);
  $21 = $20 << 31 >> 31;
  $22 = $0;
  $23 = $22;
  HEAP32[$23>>2] = $19;
  $24 = (($22) + 4)|0;
  $25 = $24;
  HEAP32[$25>>2] = $21;
  $46 = $0;
  $47 = $46;
  $48 = HEAP32[$47>>2]|0;
  $49 = (($46) + 4)|0;
  $50 = $49;
  $51 = HEAP32[$50>>2]|0;
  tempRet0 = ($51);
  STACKTOP = sp;return ($48|0);
 }
 $26 = $rc;
 $27 = $max;
 $28 = ($26|0)<=($27|0);
 if (!($28)) {
  ___assert_fail((516985|0),(516995|0),310,(517067|0));
  // unreachable;
 }
 $29 = $rc;
 $30 = $max;
 $31 = ($29|0)<($30|0);
 $32 = $3;
 $33 = ($32>>>0)>(1);
 $or$cond = $31 & $33;
 if ($or$cond) {
  $34 = $fd;
  $35 = $rc;
  $36 = $3;
  $37 = (($35>>>0) % ($36>>>0))&-1;
  $38 = (0 - ($37))|0;
  (_lseek($34,$38,1)|0);
 }
 $39 = $rc;
 $40 = $3;
 $41 = (($39>>>0) / ($40>>>0))&-1;
 $42 = $0;
 $43 = $42;
 HEAP32[$43>>2] = $41;
 $44 = (($42) + 4)|0;
 $45 = $44;
 HEAP32[$45>>2] = 0;
 $46 = $0;
 $47 = $46;
 $48 = HEAP32[$47>>2]|0;
 $49 = (($46) + 4)|0;
 $50 = $49;
 $51 = HEAP32[$50>>2]|0;
 tempRet0 = ($51);
 STACKTOP = sp;return ($48|0);
}
function ___PHYSFS_platformWrite($opaque,$buffer,$size,$count) {
 $opaque = $opaque|0;
 $buffer = $buffer|0;
 $size = $size|0;
 $count = $count|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fd = 0, $max = 0, $or$cond = 0, $rc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = $opaque;
 $2 = $buffer;
 $3 = $size;
 $4 = $count;
 $5 = $1;
 $6 = HEAP32[$5>>2]|0;
 $fd = $6;
 $7 = $3;
 $8 = $4;
 $9 = Math_imul($7, $8)|0;
 $max = $9;
 $10 = $fd;
 $11 = $2;
 $12 = $max;
 $13 = (_write($10,$11,$12)|0);
 $rc = $13;
 $14 = $rc;
 $15 = ($14|0)==(-1);
 if ($15) {
  $16 = (___errno_location()|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = (_strerror($17)|0);
  ___PHYSFS_setError($18);
  $19 = $rc;
  $20 = ($19|0)<(0);
  $21 = $20 << 31 >> 31;
  $22 = $0;
  $23 = $22;
  HEAP32[$23>>2] = $19;
  $24 = (($22) + 4)|0;
  $25 = $24;
  HEAP32[$25>>2] = $21;
  $46 = $0;
  $47 = $46;
  $48 = HEAP32[$47>>2]|0;
  $49 = (($46) + 4)|0;
  $50 = $49;
  $51 = HEAP32[$50>>2]|0;
  tempRet0 = ($51);
  STACKTOP = sp;return ($48|0);
 }
 $26 = $rc;
 $27 = $max;
 $28 = ($26|0)<=($27|0);
 if (!($28)) {
  ___assert_fail((516985|0),(516995|0),327,(517089|0));
  // unreachable;
 }
 $29 = $rc;
 $30 = $max;
 $31 = ($29|0)<($30|0);
 $32 = $3;
 $33 = ($32>>>0)>(1);
 $or$cond = $31 & $33;
 if ($or$cond) {
  $34 = $fd;
  $35 = $rc;
  $36 = $3;
  $37 = (($35>>>0) % ($36>>>0))&-1;
  $38 = (0 - ($37))|0;
  (_lseek($34,$38,1)|0);
 }
 $39 = $rc;
 $40 = $3;
 $41 = (($39>>>0) / ($40>>>0))&-1;
 $42 = $0;
 $43 = $42;
 HEAP32[$43>>2] = $41;
 $44 = (($42) + 4)|0;
 $45 = $44;
 HEAP32[$45>>2] = 0;
 $46 = $0;
 $47 = $46;
 $48 = HEAP32[$47>>2]|0;
 $49 = (($46) + 4)|0;
 $50 = $49;
 $51 = HEAP32[$50>>2]|0;
 tempRet0 = ($51);
 STACKTOP = sp;return ($48|0);
}
function ___PHYSFS_platformSeek($opaque,$0,$1) {
 $opaque = $opaque|0;
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $fd = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $4 = sp;
 $3 = $opaque;
 $5 = $4;
 $6 = $5;
 HEAP32[$6>>2] = $0;
 $7 = (($5) + 4)|0;
 $8 = $7;
 HEAP32[$8>>2] = $1;
 $9 = $3;
 $10 = HEAP32[$9>>2]|0;
 $fd = $10;
 $11 = $fd;
 $12 = $4;
 $13 = $12;
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12) + 4)|0;
 $16 = $15;
 $17 = HEAP32[$16>>2]|0;
 $18 = (_lseek($11,$14,0)|0);
 $19 = ($18|0)==(-1);
 if ($19) {
  $20 = (___errno_location()|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = (_strerror($21)|0);
  ___PHYSFS_setError($22);
  $2 = 0;
  $23 = $2;
  STACKTOP = sp;return ($23|0);
 } else {
  $2 = 1;
  $23 = $2;
  STACKTOP = sp;return ($23|0);
 }
 return (0)|0;
}
function ___PHYSFS_platformTell($opaque) {
 $opaque = $opaque|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fd = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 8|0;
 $retval = sp;
 $1 = $opaque;
 $2 = $1;
 $3 = HEAP32[$2>>2]|0;
 $fd = $3;
 $4 = $fd;
 $5 = (_lseek($4,0,1)|0);
 $6 = ($5|0)<(0);
 $7 = $6 << 31 >> 31;
 $8 = $retval;
 $9 = $8;
 HEAP32[$9>>2] = $5;
 $10 = (($8) + 4)|0;
 $11 = $10;
 HEAP32[$11>>2] = $7;
 $12 = $retval;
 $13 = $12;
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12) + 4)|0;
 $16 = $15;
 $17 = HEAP32[$16>>2]|0;
 $18 = ($14|0)==(-1);
 $19 = ($17|0)==(-1);
 $20 = $18 & $19;
 if ($20) {
  $21 = (___errno_location()|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = (_strerror($22)|0);
  ___PHYSFS_setError($23);
  $24 = $0;
  $25 = $24;
  HEAP32[$25>>2] = -1;
  $26 = (($24) + 4)|0;
  $27 = $26;
  HEAP32[$27>>2] = -1;
  $38 = $0;
  $39 = $38;
  $40 = HEAP32[$39>>2]|0;
  $41 = (($38) + 4)|0;
  $42 = $41;
  $43 = HEAP32[$42>>2]|0;
  tempRet0 = ($43);
  STACKTOP = sp;return ($40|0);
 } else {
  $28 = $retval;
  $29 = $28;
  $30 = HEAP32[$29>>2]|0;
  $31 = (($28) + 4)|0;
  $32 = $31;
  $33 = HEAP32[$32>>2]|0;
  $34 = $0;
  $35 = $34;
  HEAP32[$35>>2] = $30;
  $36 = (($34) + 4)|0;
  $37 = $36;
  HEAP32[$37>>2] = $33;
  $38 = $0;
  $39 = $38;
  $40 = HEAP32[$39>>2]|0;
  $41 = (($38) + 4)|0;
  $42 = $41;
  $43 = HEAP32[$42>>2]|0;
  tempRet0 = ($43);
  STACKTOP = sp;return ($40|0);
 }
 return (0)|0;
}
function ___PHYSFS_platformFileLength($opaque) {
 $opaque = $opaque|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fd = 0, $statbuf = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $statbuf = sp + 8|0;
 $1 = $opaque;
 $2 = $1;
 $3 = HEAP32[$2>>2]|0;
 $fd = $3;
 $4 = $fd;
 $5 = (_fstat($4,$statbuf)|0);
 $6 = ($5|0)==(-1);
 if ($6) {
  $7 = (___errno_location()|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = (_strerror($8)|0);
  ___PHYSFS_setError($9);
  $10 = $0;
  $11 = $10;
  HEAP32[$11>>2] = -1;
  $12 = (($10) + 4)|0;
  $13 = $12;
  HEAP32[$13>>2] = -1;
 } else {
  $14 = ((($statbuf)) + 36|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ($15|0)<(0);
  $17 = $16 << 31 >> 31;
  $18 = $0;
  $19 = $18;
  HEAP32[$19>>2] = $15;
  $20 = (($18) + 4)|0;
  $21 = $20;
  HEAP32[$21>>2] = $17;
 }
 $22 = $0;
 $23 = $22;
 $24 = HEAP32[$23>>2]|0;
 $25 = (($22) + 4)|0;
 $26 = $25;
 $27 = HEAP32[$26>>2]|0;
 tempRet0 = ($27);
 STACKTOP = sp;return ($24|0);
}
function ___PHYSFS_platformEOF($opaque) {
 $opaque = $opaque|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $len = 0, $or$cond = 0, $pos = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $pos = sp + 8|0;
 $len = sp;
 $0 = $opaque;
 $1 = $0;
 $2 = (___PHYSFS_platformTell($1)|0);
 $3 = tempRet0;
 $4 = $pos;
 $5 = $4;
 HEAP32[$5>>2] = $2;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = $3;
 $8 = $0;
 $9 = (___PHYSFS_platformFileLength($8)|0);
 $10 = tempRet0;
 $11 = $len;
 $12 = $11;
 HEAP32[$12>>2] = $9;
 $13 = (($11) + 4)|0;
 $14 = $13;
 HEAP32[$14>>2] = $10;
 $15 = $pos;
 $16 = $15;
 $17 = HEAP32[$16>>2]|0;
 $18 = (($15) + 4)|0;
 $19 = $18;
 $20 = HEAP32[$19>>2]|0;
 $21 = ($20|0)<(0);
 $22 = $len;
 $23 = $22;
 $24 = HEAP32[$23>>2]|0;
 $25 = (($22) + 4)|0;
 $26 = $25;
 $27 = HEAP32[$26>>2]|0;
 $28 = ($27|0)<(0);
 $or$cond = $21 | $28;
 if ($or$cond) {
  $47 = 1;
  $46 = $47&1;
  STACKTOP = sp;return ($46|0);
 }
 $29 = $pos;
 $30 = $29;
 $31 = HEAP32[$30>>2]|0;
 $32 = (($29) + 4)|0;
 $33 = $32;
 $34 = HEAP32[$33>>2]|0;
 $35 = $len;
 $36 = $35;
 $37 = HEAP32[$36>>2]|0;
 $38 = (($35) + 4)|0;
 $39 = $38;
 $40 = HEAP32[$39>>2]|0;
 $41 = ($34|0)>($40|0);
 $42 = ($31>>>0)>=($37>>>0);
 $43 = ($34|0)==($40|0);
 $44 = $43 & $42;
 $45 = $41 | $44;
 $47 = $45;
 $46 = $47&1;
 STACKTOP = sp;return ($46|0);
}
function ___PHYSFS_platformFlush($opaque) {
 $opaque = $opaque|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fd = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $opaque;
 $2 = $1;
 $3 = HEAP32[$2>>2]|0;
 $fd = $3;
 $4 = $fd;
 $5 = (_fsync($4)|0);
 $6 = ($5|0)==(-1);
 if ($6) {
  $7 = (___errno_location()|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = (_strerror($8)|0);
  ___PHYSFS_setError($9);
  $0 = 0;
  $10 = $0;
  STACKTOP = sp;return ($10|0);
 } else {
  $0 = 1;
  $10 = $0;
  STACKTOP = sp;return ($10|0);
 }
 return (0)|0;
}
function ___PHYSFS_platformClose($opaque) {
 $opaque = $opaque|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fd = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $opaque;
 $2 = $1;
 $3 = HEAP32[$2>>2]|0;
 $fd = $3;
 $4 = $fd;
 $5 = (_close($4)|0);
 $6 = ($5|0)==(-1);
 if ($6) {
  $7 = (___errno_location()|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = (_strerror($8)|0);
  ___PHYSFS_setError($9);
  $0 = 0;
  $12 = $0;
  STACKTOP = sp;return ($12|0);
 } else {
  $10 = HEAP32[(732260)>>2]|0;
  $11 = $1;
  FUNCTION_TABLE_vi[$10 & 8191]($11);
  $0 = 1;
  $12 = $0;
  STACKTOP = sp;return ($12|0);
 }
 return (0)|0;
}
function ___PHYSFS_platformDelete($path) {
 $path = $path|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $path;
 $2 = $1;
 $3 = (_remove($2)|0);
 $4 = ($3|0)==(-1);
 if ($4) {
  $5 = (___errno_location()|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (_strerror($6)|0);
  ___PHYSFS_setError($7);
  $0 = 0;
  $8 = $0;
  STACKTOP = sp;return ($8|0);
 } else {
  $0 = 1;
  $8 = $0;
  STACKTOP = sp;return ($8|0);
 }
 return (0)|0;
}
function ___PHYSFS_platformGetLastModTime($fname) {
 $fname = $fname|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $statbuf = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $statbuf = sp + 8|0;
 $1 = $fname;
 $2 = $1;
 $3 = (_stat($2,$statbuf)|0);
 $4 = ($3|0)<(0);
 if ($4) {
  $5 = (___errno_location()|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (_strerror($6)|0);
  ___PHYSFS_setError($7);
  $8 = $0;
  $9 = $8;
  HEAP32[$9>>2] = -1;
  $10 = (($8) + 4)|0;
  $11 = $10;
  HEAP32[$11>>2] = -1;
 } else {
  $12 = ((($statbuf)) + 56|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ($13|0)<(0);
  $15 = $14 << 31 >> 31;
  $16 = $0;
  $17 = $16;
  HEAP32[$17>>2] = $13;
  $18 = (($16) + 4)|0;
  $19 = $18;
  HEAP32[$19>>2] = $15;
 }
 $20 = $0;
 $21 = $20;
 $22 = HEAP32[$21>>2]|0;
 $23 = (($20) + 4)|0;
 $24 = $23;
 $25 = HEAP32[$24>>2]|0;
 tempRet0 = ($25);
 STACKTOP = sp;return ($22|0);
}
function ___PHYSFS_platformInit() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1;
}
function ___PHYSFS_platformDeinit() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1;
}
function ___PHYSFS_platformCalcBaseDir($argv0) {
 $argv0 = $argv0|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $envr = 0, $or$cond = 0, $or$cond3 = 0, $path = 0, $pid = 0, $ptr = 0, $ptr1 = 0, $rc = 0, $retval = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp + 8|0;
 $pid = sp;
 $path = sp + 48|0;
 $1 = $argv0;
 $retval = 0;
 $envr = 0;
 $2 = $1;
 $3 = ($2|0)!=(0|0);
 if ($3) {
  $4 = $1;
  $5 = (_strchr($4,47)|0);
  $6 = ($5|0)!=(0|0);
  if ($6) {
   $0 = 0;
   $64 = $0;
   STACKTOP = sp;return ($64|0);
  }
 }
 $7 = (_readSymLink(517112)|0);
 $retval = $7;
 $8 = $retval;
 $9 = ($8|0)==(0|0);
 if ($9) {
  $10 = (_getpid()|0);
  $11 = ($10|0)<(0);
  $12 = $11 << 31 >> 31;
  $13 = $pid;
  $14 = $13;
  HEAP32[$14>>2] = $10;
  $15 = (($13) + 4)|0;
  $16 = $15;
  HEAP32[$16>>2] = $12;
  $17 = $pid;
  $18 = $17;
  $19 = HEAP32[$18>>2]|0;
  $20 = (($17) + 4)|0;
  $21 = $20;
  $22 = HEAP32[$21>>2]|0;
  $23 = $vararg_buffer;
  $24 = $23;
  HEAP32[$24>>2] = $19;
  $25 = (($23) + 4)|0;
  $26 = $25;
  HEAP32[$26>>2] = $22;
  $27 = (_snprintf($path,64,517127,$vararg_buffer)|0);
  $rc = $27;
  $28 = $rc;
  $29 = ($28|0)>(0);
  $30 = $rc;
  $31 = ($30>>>0)<(64);
  $or$cond = $29 & $31;
  if ($or$cond) {
   $32 = (_readSymLink($path)|0);
   $retval = $32;
  }
 }
 $33 = $retval;
 $34 = ($33|0)!=(0|0);
 if ($34) {
  $35 = $retval;
  $36 = (_strrchr($35,47)|0);
  $ptr = $36;
  $37 = $ptr;
  $38 = ($37|0)!=(0|0);
  if ($38) {
   $39 = $ptr;
   HEAP8[$39>>0] = 0;
  }
 }
 $40 = $retval;
 $41 = ($40|0)==(0|0);
 $42 = $1;
 $43 = ($42|0)!=(0|0);
 $or$cond3 = $41 & $43;
 do {
  if ($or$cond3) {
   $44 = (___PHYSFS_platformCopyEnvironmentVariable(517142)|0);
   $envr = $44;
   $45 = $envr;
   $46 = ($45|0)!=(0|0);
   if ($46) {
    $47 = $1;
    $48 = $envr;
    $49 = (_findBinaryInPath($47,$48)|0);
    $retval = $49;
    $50 = HEAP32[(732260)>>2]|0;
    $51 = $envr;
    FUNCTION_TABLE_vi[$50 & 8191]($51);
    break;
   }
   ___PHYSFS_setError(0);
   $0 = 0;
   $64 = $0;
   STACKTOP = sp;return ($64|0);
  }
 } while(0);
 $52 = $retval;
 $53 = ($52|0)!=(0|0);
 if ($53) {
  $54 = HEAP32[(732256)>>2]|0;
  $55 = $retval;
  $56 = $retval;
  $57 = (_strlen($56)|0);
  $58 = (($57) + 1)|0;
  $59 = (FUNCTION_TABLE_iiii[$54 & 8191]($55,$58,0)|0);
  $ptr1 = $59;
  $60 = $ptr1;
  $61 = ($60|0)!=(0|0);
  if ($61) {
   $62 = $ptr1;
   $retval = $62;
  }
 }
 $63 = $retval;
 $0 = $63;
 $64 = $0;
 STACKTOP = sp;return ($64|0);
}
function _readSymLink($path) {
 $path = $path|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $len = 0, $ptr = 0, $rc = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $path;
 $len = 64;
 $rc = -1;
 $retval = 0;
 while(1) {
  $2 = HEAP32[(732256)>>2]|0;
  $3 = $retval;
  $4 = $len;
  $5 = (FUNCTION_TABLE_iiii[$2 & 8191]($3,$4,0)|0);
  $ptr = $5;
  $6 = $ptr;
  $7 = ($6|0)==(0|0);
  if ($7) {
   break;
  }
  $8 = $ptr;
  $retval = $8;
  $9 = $1;
  $10 = $retval;
  $11 = $len;
  $12 = (_readlink($9,$10,$11)|0);
  $rc = $12;
  $13 = $rc;
  $14 = ($13|0)==(-1);
  if ($14) {
   break;
  }
  $15 = $rc;
  $16 = $len;
  $17 = ($15|0)<($16|0);
  if ($17) {
   label = 5;
   break;
  }
  $22 = $len;
  $23 = $22<<1;
  $len = $23;
 }
 if ((label|0) == 5) {
  $18 = $rc;
  $19 = $retval;
  $20 = (($19) + ($18)|0);
  HEAP8[$20>>0] = 0;
  $21 = $retval;
  $0 = $21;
  $28 = $0;
  STACKTOP = sp;return ($28|0);
 }
 $24 = $retval;
 $25 = ($24|0)!=(0|0);
 if ($25) {
  $26 = HEAP32[(732260)>>2]|0;
  $27 = $retval;
  FUNCTION_TABLE_vi[$26 & 8191]($27);
 }
 $0 = 0;
 $28 = $0;
 STACKTOP = sp;return ($28|0);
}
function _findBinaryInPath($bin,$envr) {
 $bin = $bin|0;
 $envr = $envr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $7 = 0, $8 = 0, $9 = 0, $alloc_size = 0, $exe = 0, $ptr = 0, $size = 0, $start = 0, $x = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $bin;
 $2 = $envr;
 $alloc_size = 0;
 $exe = 0;
 $3 = $2;
 $start = $3;
 $4 = $1;
 $5 = ($4|0)==(0|0);
 if ($5) {
  ___PHYSFS_setError(517147);
  $0 = 0;
  $65 = $0;
  STACKTOP = sp;return ($65|0);
 }
 $6 = $2;
 $7 = ($6|0)==(0|0);
 if ($7) {
  ___PHYSFS_setError(517147);
  $0 = 0;
  $65 = $0;
  STACKTOP = sp;return ($65|0);
 }
 while(1) {
  $8 = $start;
  $9 = (_strchr($8,58)|0);
  $ptr = $9;
  $10 = $ptr;
  $11 = ($10|0)!=(0|0);
  if ($11) {
   $12 = $ptr;
   HEAP8[$12>>0] = 0;
  }
  $13 = $start;
  $14 = (_strlen($13)|0);
  $15 = $1;
  $16 = (_strlen($15)|0);
  $17 = (($14) + ($16))|0;
  $18 = (($17) + 2)|0;
  $size = $18;
  $19 = $size;
  $20 = $alloc_size;
  $21 = ($19>>>0)>($20>>>0);
  if ($21) {
   $22 = HEAP32[(732256)>>2]|0;
   $23 = $exe;
   $24 = $size;
   $25 = (FUNCTION_TABLE_iiii[$22 & 8191]($23,$24,0)|0);
   $x = $25;
   $26 = $x;
   $27 = ($26|0)==(0|0);
   if ($27) {
    label = 9;
    break;
   }
   $32 = $size;
   $alloc_size = $32;
   $33 = $x;
   $exe = $33;
  }
  $34 = $exe;
  $35 = $start;
  (_strcpy($34,$35)|0);
  $36 = $exe;
  $37 = HEAP8[$36>>0]|0;
  $38 = $37 << 24 >> 24;
  $39 = ($38|0)==(0);
  if ($39) {
   label = 15;
  } else {
   $40 = $exe;
   $41 = (_strlen($40)|0);
   $42 = (($41) - 1)|0;
   $43 = $exe;
   $44 = (($43) + ($42)|0);
   $45 = HEAP8[$44>>0]|0;
   $46 = $45 << 24 >> 24;
   $47 = ($46|0)!=(47);
   if ($47) {
    label = 15;
   }
  }
  if ((label|0) == 15) {
   label = 0;
   $48 = $exe;
   (_strcat($48,523303)|0);
  }
  $49 = $exe;
  $50 = $1;
  (_strcat($49,$50)|0);
  $51 = $exe;
  $52 = (_access($51,1)|0);
  $53 = ($52|0)==(0);
  if ($53) {
   label = 17;
   break;
  }
  $57 = $ptr;
  $58 = ((($57)) + 1|0);
  $start = $58;
  $59 = $ptr;
  $60 = ($59|0)!=(0|0);
  if (!($60)) {
   label = 19;
   break;
  }
 }
 if ((label|0) == 9) {
  $28 = $exe;
  $29 = ($28|0)!=(0|0);
  if ($29) {
   $30 = HEAP32[(732260)>>2]|0;
   $31 = $exe;
   FUNCTION_TABLE_vi[$30 & 8191]($31);
  }
  ___PHYSFS_setError(596704);
  $0 = 0;
  $65 = $0;
  STACKTOP = sp;return ($65|0);
 }
 else if ((label|0) == 17) {
  $54 = $exe;
  $55 = $start;
  (_strcpy($54,$55)|0);
  $56 = $exe;
  $0 = $56;
  $65 = $0;
  STACKTOP = sp;return ($65|0);
 }
 else if ((label|0) == 19) {
  $61 = $exe;
  $62 = ($61|0)!=(0|0);
  if ($62) {
   $63 = HEAP32[(732260)>>2]|0;
   $64 = $exe;
   FUNCTION_TABLE_vi[$63 & 8191]($64);
  }
  $0 = 0;
  $65 = $0;
  STACKTOP = sp;return ($65|0);
 }
 return (0)|0;
}
function ___PHYSFS_platformRealPath($path) {
 $path = $path|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $resolved_path = 0, $retval = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 4112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $resolved_path = sp + 16|0;
 $1 = $path;
 $retval = 0;
 $2 = (___errno_location()|0);
 HEAP32[$2>>2] = 0;
 $3 = $1;
 $4 = (_realpath($3,$resolved_path)|0);
 $5 = ($4|0)!=(0|0);
 if (!($5)) {
  $6 = (___errno_location()|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (_strerror($7)|0);
  ___PHYSFS_setError($8);
  $0 = 0;
  $17 = $0;
  STACKTOP = sp;return ($17|0);
 }
 $9 = HEAP32[(732252)>>2]|0;
 $10 = (_strlen($resolved_path)|0);
 $11 = (($10) + 1)|0;
 $12 = (FUNCTION_TABLE_iii[$9 & 8191]($11,0)|0);
 $retval = $12;
 $13 = $retval;
 $14 = ($13|0)==(0|0);
 if ($14) {
  ___PHYSFS_setError(596704);
  $0 = 0;
  $17 = $0;
  STACKTOP = sp;return ($17|0);
 } else {
  $15 = $retval;
  (_strcpy($15,$resolved_path)|0);
  $16 = $retval;
  $0 = $16;
  $17 = $0;
  STACKTOP = sp;return ($17|0);
 }
 return (0)|0;
}
function ___PHYSFS_platformSetDefaultAllocator($a) {
 $a = $a|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $a;
 STACKTOP = sp;return 0;
}
function ___PHYSFS_platformGetThreadID() {
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_pthread_self()|0);
 $1 = $0;
 $2 = $1;
 return ($2|0);
}
function ___PHYSFS_platformCreateMutex() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $m = 0;
 var $rc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = HEAP32[(732252)>>2]|0;
 $2 = (FUNCTION_TABLE_iii[$1 & 8191](36,0)|0);
 $m = $2;
 $3 = $m;
 $4 = ($3|0)==(0|0);
 if ($4) {
  ___PHYSFS_setError(596704);
  $0 = 0;
  $18 = $0;
  STACKTOP = sp;return ($18|0);
 }
 $5 = $m;
 $6 = (_pthread_mutex_init(($5|0),(0|0))|0);
 $rc = $6;
 $7 = $rc;
 $8 = ($7|0)!=(0);
 if ($8) {
  $9 = HEAP32[(732260)>>2]|0;
  $10 = $m;
  FUNCTION_TABLE_vi[$9 & 8191]($10);
  $11 = $rc;
  $12 = (_strerror($11)|0);
  ___PHYSFS_setError($12);
  $0 = 0;
  $18 = $0;
  STACKTOP = sp;return ($18|0);
 } else {
  $13 = $m;
  $14 = ((($13)) + 32|0);
  HEAP32[$14>>2] = 0;
  $15 = $m;
  $16 = ((($15)) + 28|0);
  HEAP32[$16>>2] = (-559038737);
  $17 = $m;
  $0 = $17;
  $18 = $0;
  STACKTOP = sp;return ($18|0);
 }
 return (0)|0;
}
function ___PHYSFS_platformDestroyMutex($mutex) {
 $mutex = $mutex|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $m = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $mutex;
 $1 = $0;
 $m = $1;
 $2 = $m;
 $3 = ((($2)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (_pthread_self()|0);
 $6 = ($4|0)==($5|0);
 if ($6) {
  $7 = $m;
  $8 = ((($7)) + 32|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9>>>0)>(0);
  if ($10) {
   $11 = $m;
   (_pthread_mutex_unlock(($11|0))|0);
  }
 }
 $12 = $m;
 (_pthread_mutex_destroy(($12|0))|0);
 $13 = HEAP32[(732260)>>2]|0;
 $14 = $m;
 FUNCTION_TABLE_vi[$13 & 8191]($14);
 STACKTOP = sp;return;
}
function ___PHYSFS_platformGrabMutex($mutex) {
 $mutex = $mutex|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $m = 0, $tid = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $mutex;
 $2 = $1;
 $m = $2;
 $3 = (_pthread_self()|0);
 $tid = $3;
 $4 = $m;
 $5 = ((($4)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $tid;
 $8 = ($6|0)!=($7|0);
 do {
  if ($8) {
   $9 = $m;
   $10 = (_pthread_mutex_lock(($9|0))|0);
   $11 = ($10|0)!=(0);
   if (!($11)) {
    $12 = $tid;
    $13 = $m;
    $14 = ((($13)) + 28|0);
    HEAP32[$14>>2] = $12;
    break;
   }
   $0 = 0;
   $19 = $0;
   STACKTOP = sp;return ($19|0);
  }
 } while(0);
 $15 = $m;
 $16 = ((($15)) + 32|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = (($17) + 1)|0;
 HEAP32[$16>>2] = $18;
 $0 = 1;
 $19 = $0;
 STACKTOP = sp;return ($19|0);
}
function ___PHYSFS_platformReleaseMutex($mutex) {
 $mutex = $mutex|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $m = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $mutex;
 $1 = $0;
 $m = $1;
 $2 = $m;
 $3 = ((($2)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (_pthread_self()|0);
 $6 = ($4|0)==($5|0);
 if (!($6)) {
  STACKTOP = sp;return;
 }
 $7 = $m;
 $8 = ((($7)) + 32|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (($9) + -1)|0;
 HEAP32[$8>>2] = $10;
 $11 = ($10|0)==(0);
 if (!($11)) {
  STACKTOP = sp;return;
 }
 $12 = $m;
 $13 = ((($12)) + 28|0);
 HEAP32[$13>>2] = (-559038737);
 $14 = $m;
 (_pthread_mutex_unlock(($14|0))|0);
 STACKTOP = sp;return;
}
function _DIR_isArchive($filename,$forWriting) {
 $filename = $filename|0;
 $forWriting = $forWriting|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $filename;
 $1 = $forWriting;
 $2 = $0;
 $3 = (___PHYSFS_platformIsDirectory($2)|0);
 STACKTOP = sp;return ($3|0);
}
function _DIR_openArchive($name,$forWriting) {
 $name = $name|0;
 $forWriting = $forWriting|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dirsep = 0, $namelen = 0, $retval = 0, $seplen = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $name;
 $2 = $forWriting;
 $3 = (_PHYSFS_getDirSeparator()|0);
 $dirsep = $3;
 $retval = 0;
 $4 = $1;
 $5 = (_strlen($4)|0);
 $namelen = $5;
 $6 = $dirsep;
 $7 = (_strlen($6)|0);
 $seplen = $7;
 $8 = $1;
 $9 = $2;
 $10 = (_DIR_isArchive($8,$9)|0);
 $11 = ($10|0)!=(0);
 if (!($11)) {
  ___PHYSFS_setError(517361);
  $0 = 0;
  $34 = $0;
  STACKTOP = sp;return ($34|0);
 }
 $12 = HEAP32[(732252)>>2]|0;
 $13 = $namelen;
 $14 = $seplen;
 $15 = (($13) + ($14))|0;
 $16 = (($15) + 1)|0;
 $17 = (FUNCTION_TABLE_iii[$12 & 8191]($16,0)|0);
 $retval = $17;
 $18 = $retval;
 $19 = ($18|0)==(0|0);
 if ($19) {
  ___PHYSFS_setError(596704);
  $0 = 0;
  $34 = $0;
  STACKTOP = sp;return ($34|0);
 }
 $20 = $retval;
 $21 = $1;
 (_strcpy($20,$21)|0);
 $22 = $1;
 $23 = $namelen;
 $24 = (($22) + ($23)|0);
 $25 = $seplen;
 $26 = (0 - ($25))|0;
 $27 = (($24) + ($26)|0);
 $28 = $dirsep;
 $29 = (_strcmp($27,$28)|0);
 $30 = ($29|0)!=(0);
 if ($30) {
  $31 = $retval;
  $32 = $dirsep;
  (_strcat($31,$32)|0);
 }
 $33 = $retval;
 $0 = $33;
 $34 = $0;
 STACKTOP = sp;return ($34|0);
}
function _DIR_enumerateFiles($opaque,$dname,$omitSymLinks,$cb,$origdir,$callbackdata) {
 $opaque = $opaque|0;
 $dname = $dname|0;
 $omitSymLinks = $omitSymLinks|0;
 $cb = $cb|0;
 $origdir = $origdir|0;
 $callbackdata = $callbackdata|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $d = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $dname;
 $2 = $omitSymLinks;
 $3 = $cb;
 $4 = $origdir;
 $5 = $callbackdata;
 $6 = $0;
 $7 = $1;
 $8 = (___PHYSFS_platformCvtToDependent($6,$7,0)|0);
 $d = $8;
 $9 = $d;
 $10 = ($9|0)!=(0|0);
 if (!($10)) {
  STACKTOP = sp;return;
 }
 $11 = $d;
 $12 = $2;
 $13 = $3;
 $14 = $4;
 $15 = $5;
 ___PHYSFS_platformEnumerateFiles($11,$12,$13,$14,$15);
 $16 = HEAP32[(732260)>>2]|0;
 $17 = $d;
 FUNCTION_TABLE_vi[$16 & 8191]($17);
 STACKTOP = sp;return;
}
function _DIR_exists($opaque,$name) {
 $opaque = $opaque|0;
 $name = $name|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $opaque;
 $2 = $name;
 $3 = $1;
 $4 = $2;
 $5 = (___PHYSFS_platformCvtToDependent($3,$4,0)|0);
 $f = $5;
 $6 = $f;
 $7 = ($6|0)==(0|0);
 if ($7) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $13 = $0;
  STACKTOP = sp;return ($13|0);
 } else {
  $8 = $f;
  $9 = (___PHYSFS_platformExists($8)|0);
  $retval = $9;
  $10 = HEAP32[(732260)>>2]|0;
  $11 = $f;
  FUNCTION_TABLE_vi[$10 & 8191]($11);
  $12 = $retval;
  $0 = $12;
  $13 = $0;
  STACKTOP = sp;return ($13|0);
 }
 return (0)|0;
}
function _DIR_isDirectory($opaque,$name,$fileExists) {
 $opaque = $opaque|0;
 $name = $name|0;
 $fileExists = $fileExists|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $d = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $opaque;
 $2 = $name;
 $3 = $fileExists;
 $4 = $1;
 $5 = $2;
 $6 = (___PHYSFS_platformCvtToDependent($4,$5,0)|0);
 $d = $6;
 $retval = 0;
 $7 = $d;
 $8 = ($7|0)==(0|0);
 if ($8) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $20 = $0;
  STACKTOP = sp;return ($20|0);
 }
 $9 = $d;
 $10 = (___PHYSFS_platformExists($9)|0);
 $11 = $3;
 HEAP32[$11>>2] = $10;
 $12 = $3;
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13|0)!=(0);
 if ($14) {
  $15 = $d;
  $16 = (___PHYSFS_platformIsDirectory($15)|0);
  $retval = $16;
 }
 $17 = HEAP32[(732260)>>2]|0;
 $18 = $d;
 FUNCTION_TABLE_vi[$17 & 8191]($18);
 $19 = $retval;
 $0 = $19;
 $20 = $0;
 STACKTOP = sp;return ($20|0);
}
function _DIR_isSymLink($opaque,$name,$fileExists) {
 $opaque = $opaque|0;
 $name = $name|0;
 $fileExists = $fileExists|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $f = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $opaque;
 $2 = $name;
 $3 = $fileExists;
 $4 = $1;
 $5 = $2;
 $6 = (___PHYSFS_platformCvtToDependent($4,$5,0)|0);
 $f = $6;
 $retval = 0;
 $7 = $f;
 $8 = ($7|0)==(0|0);
 if ($8) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $20 = $0;
  STACKTOP = sp;return ($20|0);
 }
 $9 = $f;
 $10 = (___PHYSFS_platformExists($9)|0);
 $11 = $3;
 HEAP32[$11>>2] = $10;
 $12 = $3;
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13|0)!=(0);
 if ($14) {
  $15 = $f;
  $16 = (___PHYSFS_platformIsSymLink($15)|0);
  $retval = $16;
 }
 $17 = HEAP32[(732260)>>2]|0;
 $18 = $f;
 FUNCTION_TABLE_vi[$17 & 8191]($18);
 $19 = $retval;
 $0 = $19;
 $20 = $0;
 STACKTOP = sp;return ($20|0);
}
function _DIR_getLastModTime($opaque,$name,$fileExists) {
 $opaque = $opaque|0;
 $name = $name|0;
 $fileExists = $fileExists|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $d = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 8|0;
 $retval = sp;
 $1 = $opaque;
 $2 = $name;
 $3 = $fileExists;
 $4 = $1;
 $5 = $2;
 $6 = (___PHYSFS_platformCvtToDependent($4,$5,0)|0);
 $d = $6;
 $7 = $retval;
 $8 = $7;
 HEAP32[$8>>2] = -1;
 $9 = (($7) + 4)|0;
 $10 = $9;
 HEAP32[$10>>2] = -1;
 $11 = $d;
 $12 = ($11|0)==(0|0);
 if ($12) {
  ___PHYSFS_setError(0);
  $13 = $0;
  $14 = $13;
  HEAP32[$14>>2] = 0;
  $15 = (($13) + 4)|0;
  $16 = $15;
  HEAP32[$16>>2] = 0;
  $42 = $0;
  $43 = $42;
  $44 = HEAP32[$43>>2]|0;
  $45 = (($42) + 4)|0;
  $46 = $45;
  $47 = HEAP32[$46>>2]|0;
  tempRet0 = ($47);
  STACKTOP = sp;return ($44|0);
 }
 $17 = $d;
 $18 = (___PHYSFS_platformExists($17)|0);
 $19 = $3;
 HEAP32[$19>>2] = $18;
 $20 = $3;
 $21 = HEAP32[$20>>2]|0;
 $22 = ($21|0)!=(0);
 if ($22) {
  $23 = $d;
  $24 = (___PHYSFS_platformGetLastModTime($23)|0);
  $25 = tempRet0;
  $26 = $retval;
  $27 = $26;
  HEAP32[$27>>2] = $24;
  $28 = (($26) + 4)|0;
  $29 = $28;
  HEAP32[$29>>2] = $25;
 }
 $30 = HEAP32[(732260)>>2]|0;
 $31 = $d;
 FUNCTION_TABLE_vi[$30 & 8191]($31);
 $32 = $retval;
 $33 = $32;
 $34 = HEAP32[$33>>2]|0;
 $35 = (($32) + 4)|0;
 $36 = $35;
 $37 = HEAP32[$36>>2]|0;
 $38 = $0;
 $39 = $38;
 HEAP32[$39>>2] = $34;
 $40 = (($38) + 4)|0;
 $41 = $40;
 HEAP32[$41>>2] = $37;
 $42 = $0;
 $43 = $42;
 $44 = HEAP32[$43>>2]|0;
 $45 = (($42) + 4)|0;
 $46 = $45;
 $47 = HEAP32[$46>>2]|0;
 tempRet0 = ($47);
 STACKTOP = sp;return ($44|0);
}
function _DIR_openRead($opaque,$fnm,$exist) {
 $opaque = $opaque|0;
 $fnm = $fnm|0;
 $exist = $exist|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $fnm;
 $2 = $exist;
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = (_doOpen_6761($3,$4,4282,$5)|0);
 STACKTOP = sp;return ($6|0);
}
function _doOpen_6761($opaque,$name,$openFunc,$fileExists) {
 $opaque = $opaque|0;
 $name = $name|0;
 $openFunc = $openFunc|0;
 $fileExists = $fileExists|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f = 0, $rc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $opaque;
 $2 = $name;
 $3 = $openFunc;
 $4 = $fileExists;
 $5 = $1;
 $6 = $2;
 $7 = (___PHYSFS_platformCvtToDependent($5,$6,0)|0);
 $f = $7;
 $rc = 0;
 $8 = $f;
 $9 = ($8|0)==(0|0);
 if ($9) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $26 = $0;
  STACKTOP = sp;return ($26|0);
 }
 $10 = $4;
 $11 = ($10|0)!=(0|0);
 if ($11) {
  $12 = $f;
  $13 = (___PHYSFS_platformExists($12)|0);
  $14 = $4;
  HEAP32[$14>>2] = $13;
  $15 = $4;
  $16 = HEAP32[$15>>2]|0;
  $17 = ($16|0)!=(0);
  if (!($17)) {
   $18 = HEAP32[(732260)>>2]|0;
   $19 = $f;
   FUNCTION_TABLE_vi[$18 & 8191]($19);
   $0 = 0;
   $26 = $0;
   STACKTOP = sp;return ($26|0);
  }
 }
 $20 = $3;
 $21 = $f;
 $22 = (FUNCTION_TABLE_ii[$20 & 8191]($21)|0);
 $rc = $22;
 $23 = HEAP32[(732260)>>2]|0;
 $24 = $f;
 FUNCTION_TABLE_vi[$23 & 8191]($24);
 $25 = $rc;
 $0 = $25;
 $26 = $0;
 STACKTOP = sp;return ($26|0);
}
function _DIR_openWrite($opaque,$filename) {
 $opaque = $opaque|0;
 $filename = $filename|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $filename;
 $2 = $0;
 $3 = $1;
 $4 = (_doOpen_6761($2,$3,4283,0)|0);
 STACKTOP = sp;return ($4|0);
}
function _DIR_openAppend($opaque,$filename) {
 $opaque = $opaque|0;
 $filename = $filename|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $filename;
 $2 = $0;
 $3 = $1;
 $4 = (_doOpen_6761($2,$3,4284,0)|0);
 STACKTOP = sp;return ($4|0);
}
function _DIR_remove($opaque,$name) {
 $opaque = $opaque|0;
 $name = $name|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $opaque;
 $2 = $name;
 $3 = $1;
 $4 = $2;
 $5 = (___PHYSFS_platformCvtToDependent($3,$4,0)|0);
 $f = $5;
 $6 = $f;
 $7 = ($6|0)==(0|0);
 if ($7) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $13 = $0;
  STACKTOP = sp;return ($13|0);
 } else {
  $8 = $f;
  $9 = (___PHYSFS_platformDelete($8)|0);
  $retval = $9;
  $10 = HEAP32[(732260)>>2]|0;
  $11 = $f;
  FUNCTION_TABLE_vi[$10 & 8191]($11);
  $12 = $retval;
  $0 = $12;
  $13 = $0;
  STACKTOP = sp;return ($13|0);
 }
 return (0)|0;
}
function _DIR_mkdir($opaque,$name) {
 $opaque = $opaque|0;
 $name = $name|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $opaque;
 $2 = $name;
 $3 = $1;
 $4 = $2;
 $5 = (___PHYSFS_platformCvtToDependent($3,$4,0)|0);
 $f = $5;
 $6 = $f;
 $7 = ($6|0)==(0|0);
 if ($7) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $13 = $0;
  STACKTOP = sp;return ($13|0);
 } else {
  $8 = $f;
  $9 = (___PHYSFS_platformMkDir($8)|0);
  $retval = $9;
  $10 = HEAP32[(732260)>>2]|0;
  $11 = $f;
  FUNCTION_TABLE_vi[$10 & 8191]($11);
  $12 = $retval;
  $0 = $12;
  $13 = $0;
  STACKTOP = sp;return ($13|0);
 }
 return (0)|0;
}
function _DIR_dirClose($opaque) {
 $opaque = $opaque|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = HEAP32[(732260)>>2]|0;
 $2 = $0;
 FUNCTION_TABLE_vi[$1 & 8191]($2);
 STACKTOP = sp;return;
}
function _DIR_read($opaque,$buffer,$objSize,$objCount) {
 $opaque = $opaque|0;
 $buffer = $buffer|0;
 $objSize = $objSize|0;
 $objCount = $objCount|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $retval = sp;
 $0 = $opaque;
 $1 = $buffer;
 $2 = $objSize;
 $3 = $objCount;
 $4 = $0;
 $5 = $1;
 $6 = $2;
 $7 = $3;
 $8 = (___PHYSFS_platformRead($4,$5,$6,$7)|0);
 $9 = tempRet0;
 $10 = $retval;
 $11 = $10;
 HEAP32[$11>>2] = $8;
 $12 = (($10) + 4)|0;
 $13 = $12;
 HEAP32[$13>>2] = $9;
 $14 = $retval;
 $15 = $14;
 $16 = HEAP32[$15>>2]|0;
 $17 = (($14) + 4)|0;
 $18 = $17;
 $19 = HEAP32[$18>>2]|0;
 tempRet0 = ($19);
 STACKTOP = sp;return ($16|0);
}
function _DIR_write($opaque,$buffer,$objSize,$objCount) {
 $opaque = $opaque|0;
 $buffer = $buffer|0;
 $objSize = $objSize|0;
 $objCount = $objCount|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $retval = sp;
 $0 = $opaque;
 $1 = $buffer;
 $2 = $objSize;
 $3 = $objCount;
 $4 = $0;
 $5 = $1;
 $6 = $2;
 $7 = $3;
 $8 = (___PHYSFS_platformWrite($4,$5,$6,$7)|0);
 $9 = tempRet0;
 $10 = $retval;
 $11 = $10;
 HEAP32[$11>>2] = $8;
 $12 = (($10) + 4)|0;
 $13 = $12;
 HEAP32[$13>>2] = $9;
 $14 = $retval;
 $15 = $14;
 $16 = HEAP32[$15>>2]|0;
 $17 = (($14) + 4)|0;
 $18 = $17;
 $19 = HEAP32[$18>>2]|0;
 tempRet0 = ($19);
 STACKTOP = sp;return ($16|0);
}
function _DIR_eof($opaque) {
 $opaque = $opaque|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $0;
 $2 = (___PHYSFS_platformEOF($1)|0);
 STACKTOP = sp;return ($2|0);
}
function _DIR_tell($opaque) {
 $opaque = $opaque|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $0;
 $2 = (___PHYSFS_platformTell($1)|0);
 $3 = tempRet0;
 tempRet0 = ($3);
 STACKTOP = sp;return ($2|0);
}
function _DIR_seek($opaque,$0,$1) {
 $opaque = $opaque|0;
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $3 = sp;
 $2 = $opaque;
 $4 = $3;
 $5 = $4;
 HEAP32[$5>>2] = $0;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = $1;
 $8 = $2;
 $9 = $3;
 $10 = $9;
 $11 = HEAP32[$10>>2]|0;
 $12 = (($9) + 4)|0;
 $13 = $12;
 $14 = HEAP32[$13>>2]|0;
 $15 = (___PHYSFS_platformSeek($8,$11,$14)|0);
 STACKTOP = sp;return ($15|0);
}
function _DIR_fileLength($opaque) {
 $opaque = $opaque|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $0;
 $2 = (___PHYSFS_platformFileLength($1)|0);
 $3 = tempRet0;
 tempRet0 = ($3);
 STACKTOP = sp;return ($2|0);
}
function _DIR_fileClose($opaque) {
 $opaque = $opaque|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $opaque;
 $2 = $1;
 $3 = (___PHYSFS_platformFlush($2)|0);
 $4 = ($3|0)!=(0);
 do {
  if ($4) {
   $5 = $1;
   $6 = (___PHYSFS_platformClose($5)|0);
   $7 = ($6|0)!=(0);
   if ($7) {
    $0 = 1;
    break;
   } else {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
  } else {
   ___PHYSFS_setError(0);
   $0 = 0;
  }
 } while(0);
 $8 = $0;
 STACKTOP = sp;return ($8|0);
}
function _ZIP_isArchive($filename,$forWriting) {
 $filename = $filename|0;
 $forWriting = $forWriting|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $in = 0, $retval = 0, $sig = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $sig = sp + 8|0;
 $1 = $filename;
 $2 = $forWriting;
 $retval = 0;
 $3 = $1;
 $4 = (___PHYSFS_platformOpenRead($3)|0);
 $in = $4;
 $5 = $in;
 $6 = ($5|0)==(0|0);
 if ($6) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $24 = $0;
  STACKTOP = sp;return ($24|0);
 }
 $7 = $in;
 $8 = (_readui32($7,$sig)|0);
 $9 = ($8|0)!=(0);
 if ($9) {
  $10 = HEAP32[$sig>>2]|0;
  $11 = ($10|0)==(67324752);
  $12 = $11&1;
  $retval = $12;
  $13 = $retval;
  $14 = ($13|0)!=(0);
  if (!($14)) {
   $15 = $in;
   $16 = (_zip_find_end_of_central_dir($15,0)|0);
   $17 = tempRet0;
   $18 = ($16|0)!=(-1);
   $19 = ($17|0)!=(-1);
   $20 = $18 | $19;
   $21 = $20&1;
   $retval = $21;
  }
 }
 $22 = $in;
 (___PHYSFS_platformClose($22)|0);
 $23 = $retval;
 $0 = $23;
 $24 = $0;
 STACKTOP = sp;return ($24|0);
}
function _readui32($in,$val) {
 $in = $in|0;
 $val = $val|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp;
 $1 = $in;
 $2 = $val;
 $3 = $1;
 $4 = (___PHYSFS_platformRead($3,$v,4,1)|0);
 $5 = tempRet0;
 $6 = ($4|0)!=(1);
 $7 = ($5|0)!=(0);
 $8 = $6 | $7;
 if ($8) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $12 = $0;
  STACKTOP = sp;return ($12|0);
 } else {
  $9 = HEAP32[$v>>2]|0;
  $10 = (_PHYSFS_swapULE32($9)|0);
  $11 = $2;
  HEAP32[$11>>2] = $10;
  $0 = 1;
  $12 = $0;
  STACKTOP = sp;return ($12|0);
 }
 return (0)|0;
}
function _zip_find_end_of_central_dir($in,$len) {
 $in = $in|0;
 $len = $len|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $buf = 0, $extra = 0, $filelen = 0, $filepos = 0, $found = 0;
 var $i = 0, $maxread = 0, $totalread = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 320|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 16|0;
 $buf = sp + 56|0;
 $extra = sp + 48|0;
 $filelen = sp + 8|0;
 $filepos = sp;
 $1 = $in;
 $2 = $len;
 ;HEAP8[$extra>>0]=0|0;HEAP8[$extra+1>>0]=0|0;HEAP8[$extra+2>>0]=0|0;HEAP8[$extra+3>>0]=0|0;
 $i = 0;
 $totalread = 0;
 $found = 0;
 $3 = $1;
 $4 = (___PHYSFS_platformFileLength($3)|0);
 $5 = tempRet0;
 $6 = $filelen;
 $7 = $6;
 HEAP32[$7>>2] = $4;
 $8 = (($6) + 4)|0;
 $9 = $8;
 HEAP32[$9>>2] = $5;
 $10 = $filelen;
 $11 = $10;
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10) + 4)|0;
 $14 = $13;
 $15 = HEAP32[$14>>2]|0;
 $16 = ($12|0)==(-1);
 $17 = ($15|0)==(-1);
 $18 = $16 & $17;
 if ($18) {
  ___PHYSFS_setError(0);
  $19 = $0;
  $20 = $19;
  HEAP32[$20>>2] = 0;
  $21 = (($19) + 4)|0;
  $22 = $21;
  HEAP32[$22>>2] = 0;
  $229 = $0;
  $230 = $229;
  $231 = HEAP32[$230>>2]|0;
  $232 = (($229) + 4)|0;
  $233 = $232;
  $234 = HEAP32[$233>>2]|0;
  tempRet0 = ($234);
  STACKTOP = sp;return ($231|0);
 }
 $23 = $filelen;
 $24 = $23;
 $25 = HEAP32[$24>>2]|0;
 $26 = (($23) + 4)|0;
 $27 = $26;
 $28 = HEAP32[$27>>2]|0;
 $29 = ($28|0)>(0);
 $30 = ($25>>>0)>(4294967295);
 $31 = ($28|0)==(0);
 $32 = $31 & $30;
 $33 = $29 | $32;
 if ($33) {
  ___PHYSFS_setError(517300);
  $34 = $0;
  $35 = $34;
  HEAP32[$35>>2] = 0;
  $36 = (($34) + 4)|0;
  $37 = $36;
  HEAP32[$37>>2] = 0;
  $229 = $0;
  $230 = $229;
  $231 = HEAP32[$230>>2]|0;
  $232 = (($229) + 4)|0;
  $233 = $232;
  $234 = HEAP32[$233>>2]|0;
  tempRet0 = ($234);
  STACKTOP = sp;return ($231|0);
 }
 $38 = $filelen;
 $39 = $38;
 $40 = HEAP32[$39>>2]|0;
 $41 = (($38) + 4)|0;
 $42 = $41;
 $43 = HEAP32[$42>>2]|0;
 $44 = (0)<($43|0);
 $45 = (256)<($40>>>0);
 $46 = (0)==($43|0);
 $47 = $46 & $45;
 $48 = $44 | $47;
 if ($48) {
  $49 = $filelen;
  $50 = $49;
  $51 = HEAP32[$50>>2]|0;
  $52 = (($49) + 4)|0;
  $53 = $52;
  $54 = HEAP32[$53>>2]|0;
  $55 = (_i64Subtract(($51|0),($54|0),256,0)|0);
  $56 = tempRet0;
  $57 = $filepos;
  $58 = $57;
  HEAP32[$58>>2] = $55;
  $59 = (($57) + 4)|0;
  $60 = $59;
  HEAP32[$60>>2] = $56;
  $maxread = 256;
 } else {
  $61 = $filepos;
  $62 = $61;
  HEAP32[$62>>2] = 0;
  $63 = (($61) + 4)|0;
  $64 = $63;
  HEAP32[$64>>2] = 0;
  $65 = $filelen;
  $66 = $65;
  $67 = HEAP32[$66>>2]|0;
  $68 = (($65) + 4)|0;
  $69 = $68;
  $70 = HEAP32[$69>>2]|0;
  $maxread = $67;
 }
 while(1) {
  $71 = $totalread;
  $72 = ($71|0)<(0);
  $73 = $72 << 31 >> 31;
  $74 = $filelen;
  $75 = $74;
  $76 = HEAP32[$75>>2]|0;
  $77 = (($74) + 4)|0;
  $78 = $77;
  $79 = HEAP32[$78>>2]|0;
  $80 = ($73|0)<($79|0);
  $81 = ($71>>>0)<($76>>>0);
  $82 = ($73|0)==($79|0);
  $83 = $82 & $81;
  $84 = $80 | $83;
  $85 = $totalread;
  $86 = ($85|0)<(65557);
  $87 = $84 ? $86 : 0;
  if (!($87)) {
   label = 28;
   break;
  }
  $88 = $1;
  $89 = $filepos;
  $90 = $89;
  $91 = HEAP32[$90>>2]|0;
  $92 = (($89) + 4)|0;
  $93 = $92;
  $94 = HEAP32[$93>>2]|0;
  $95 = (___PHYSFS_platformSeek($88,$91,$94)|0);
  $96 = ($95|0)!=(0);
  if (!($96)) {
   label = 10;
   break;
  }
  $101 = $totalread;
  $102 = ($101|0)!=(0);
  $103 = $1;
  $104 = $maxread;
  if ($102) {
   $105 = (($104) - 4)|0;
   $106 = (___PHYSFS_platformRead($103,$buf,$105,1)|0);
   $107 = tempRet0;
   $108 = ($106|0)!=(1);
   $109 = ($107|0)!=(0);
   $110 = $108 | $109;
   if ($110) {
    label = 13;
    break;
   }
   $115 = $maxread;
   $116 = (($115) - 4)|0;
   $117 = (($buf) + ($116)|0);
   ;HEAP8[$117>>0]=HEAP8[$extra>>0]|0;HEAP8[$117+1>>0]=HEAP8[$extra+1>>0]|0;HEAP8[$117+2>>0]=HEAP8[$extra+2>>0]|0;HEAP8[$117+3>>0]=HEAP8[$extra+3>>0]|0;
   $118 = $maxread;
   $119 = (($118) - 4)|0;
   $120 = $totalread;
   $121 = (($120) + ($119))|0;
   $totalread = $121;
  } else {
   $122 = (___PHYSFS_platformRead($103,$buf,$104,1)|0);
   $123 = tempRet0;
   $124 = ($122|0)!=(1);
   $125 = ($123|0)!=(0);
   $126 = $124 | $125;
   if ($126) {
    label = 16;
    break;
   }
   $131 = $maxread;
   $132 = $totalread;
   $133 = (($132) + ($131))|0;
   $totalread = $133;
  }
  ;HEAP8[$extra>>0]=HEAP8[$buf>>0]|0;HEAP8[$extra+1>>0]=HEAP8[$buf+1>>0]|0;HEAP8[$extra+2>>0]=HEAP8[$buf+2>>0]|0;HEAP8[$extra+3>>0]=HEAP8[$buf+3>>0]|0;
  $134 = $maxread;
  $135 = (($134) - 4)|0;
  $i = $135;
  while(1) {
   $136 = $i;
   $137 = ($136|0)>(0);
   if (!($137)) {
    break;
   }
   $138 = $i;
   $139 = (($138) + 0)|0;
   $140 = (($buf) + ($139)|0);
   $141 = HEAP8[$140>>0]|0;
   $142 = $141&255;
   $143 = ($142|0)==(80);
   if ($143) {
    $144 = $i;
    $145 = (($144) + 1)|0;
    $146 = (($buf) + ($145)|0);
    $147 = HEAP8[$146>>0]|0;
    $148 = $147&255;
    $149 = ($148|0)==(75);
    if ($149) {
     $150 = $i;
     $151 = (($150) + 2)|0;
     $152 = (($buf) + ($151)|0);
     $153 = HEAP8[$152>>0]|0;
     $154 = $153&255;
     $155 = ($154|0)==(5);
     if ($155) {
      $156 = $i;
      $157 = (($156) + 3)|0;
      $158 = (($buf) + ($157)|0);
      $159 = HEAP8[$158>>0]|0;
      $160 = $159&255;
      $161 = ($160|0)==(6);
      if ($161) {
       label = 24;
       break;
      }
     }
    }
   }
   $162 = $i;
   $163 = (($162) + -1)|0;
   $i = $163;
  }
  if ((label|0) == 24) {
   label = 0;
   $found = 1;
  }
  $164 = $found;
  $165 = ($164|0)!=(0);
  if ($165) {
   label = 28;
   break;
  }
  $166 = $maxread;
  $167 = (($166) - 4)|0;
  $168 = ($167|0)<(0);
  $169 = $168 << 31 >> 31;
  $170 = $filepos;
  $171 = $170;
  $172 = HEAP32[$171>>2]|0;
  $173 = (($170) + 4)|0;
  $174 = $173;
  $175 = HEAP32[$174>>2]|0;
  $176 = (_i64Subtract(($172|0),($175|0),($167|0),($169|0))|0);
  $177 = tempRet0;
  $178 = $filepos;
  $179 = $178;
  HEAP32[$179>>2] = $176;
  $180 = (($178) + 4)|0;
  $181 = $180;
  HEAP32[$181>>2] = $177;
  $182 = $filepos;
  $183 = $182;
  $184 = HEAP32[$183>>2]|0;
  $185 = (($182) + 4)|0;
  $186 = $185;
  $187 = HEAP32[$186>>2]|0;
  $188 = ($187|0)<(0);
  $189 = $188 ? 0 : $176;
  $190 = $188 ? 0 : $177;
  $191 = $filepos;
  $192 = $191;
  HEAP32[$192>>2] = $189;
  $193 = (($191) + 4)|0;
  $194 = $193;
  HEAP32[$194>>2] = $190;
 }
 if ((label|0) == 10) {
  ___PHYSFS_setError(0);
  $97 = $0;
  $98 = $97;
  HEAP32[$98>>2] = -1;
  $99 = (($97) + 4)|0;
  $100 = $99;
  HEAP32[$100>>2] = -1;
  $229 = $0;
  $230 = $229;
  $231 = HEAP32[$230>>2]|0;
  $232 = (($229) + 4)|0;
  $233 = $232;
  $234 = HEAP32[$233>>2]|0;
  tempRet0 = ($234);
  STACKTOP = sp;return ($231|0);
 }
 else if ((label|0) == 13) {
  $111 = $0;
  $112 = $111;
  HEAP32[$112>>2] = -1;
  $113 = (($111) + 4)|0;
  $114 = $113;
  HEAP32[$114>>2] = -1;
  $229 = $0;
  $230 = $229;
  $231 = HEAP32[$230>>2]|0;
  $232 = (($229) + 4)|0;
  $233 = $232;
  $234 = HEAP32[$233>>2]|0;
  tempRet0 = ($234);
  STACKTOP = sp;return ($231|0);
 }
 else if ((label|0) == 16) {
  $127 = $0;
  $128 = $127;
  HEAP32[$128>>2] = -1;
  $129 = (($127) + 4)|0;
  $130 = $129;
  HEAP32[$130>>2] = -1;
  $229 = $0;
  $230 = $229;
  $231 = HEAP32[$230>>2]|0;
  $232 = (($229) + 4)|0;
  $233 = $232;
  $234 = HEAP32[$233>>2]|0;
  tempRet0 = ($234);
  STACKTOP = sp;return ($231|0);
 }
 else if ((label|0) == 28) {
  $195 = $found;
  $196 = ($195|0)!=(0);
  if (!($196)) {
   ___PHYSFS_setError(517325);
   $197 = $0;
   $198 = $197;
   HEAP32[$198>>2] = -1;
   $199 = (($197) + 4)|0;
   $200 = $199;
   HEAP32[$200>>2] = -1;
   $229 = $0;
   $230 = $229;
   $231 = HEAP32[$230>>2]|0;
   $232 = (($229) + 4)|0;
   $233 = $232;
   $234 = HEAP32[$233>>2]|0;
   tempRet0 = ($234);
   STACKTOP = sp;return ($231|0);
  }
  $201 = $2;
  $202 = ($201|0)!=(0|0);
  if ($202) {
   $203 = $filelen;
   $204 = $203;
   $205 = HEAP32[$204>>2]|0;
   $206 = (($203) + 4)|0;
   $207 = $206;
   $208 = HEAP32[$207>>2]|0;
   $209 = $2;
   $210 = $209;
   $211 = $210;
   HEAP32[$211>>2] = $205;
   $212 = (($210) + 4)|0;
   $213 = $212;
   HEAP32[$213>>2] = $208;
  }
  $214 = $filepos;
  $215 = $214;
  $216 = HEAP32[$215>>2]|0;
  $217 = (($214) + 4)|0;
  $218 = $217;
  $219 = HEAP32[$218>>2]|0;
  $220 = $i;
  $221 = ($220|0)<(0);
  $222 = $221 << 31 >> 31;
  $223 = (_i64Add(($216|0),($219|0),($220|0),($222|0))|0);
  $224 = tempRet0;
  $225 = $0;
  $226 = $225;
  HEAP32[$226>>2] = $223;
  $227 = (($225) + 4)|0;
  $228 = $227;
  HEAP32[$228>>2] = $224;
  $229 = $0;
  $230 = $229;
  $231 = HEAP32[$230>>2]|0;
  $232 = (($229) + 4)|0;
  $233 = $232;
  $234 = HEAP32[$233>>2]|0;
  tempRet0 = ($234);
  STACKTOP = sp;return ($231|0);
 }
 return (0)|0;
}
function _ZIP_openArchive($name,$forWriting) {
 $name = $name|0;
 $forWriting = $forWriting|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cent_dir_ofs = 0, $data_start = 0, $in = 0;
 var $info = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $data_start = sp + 4|0;
 $cent_dir_ofs = sp;
 $1 = $name;
 $2 = $forWriting;
 $in = 0;
 $info = 0;
 $3 = $2;
 $4 = ($3|0)!=(0);
 if ($4) {
  ___PHYSFS_setError(517340);
  $0 = 0;
  $36 = $0;
  STACKTOP = sp;return ($36|0);
 }
 $5 = $1;
 $6 = (___PHYSFS_platformOpenRead($5)|0);
 $in = $6;
 $7 = ($6|0)==(0|0);
 if (!($7)) {
  $8 = $1;
  $9 = (_zip_create_zipinfo($8)|0);
  $info = $9;
  $10 = ($9|0)==(0|0);
  if (!($10)) {
   $11 = $in;
   $12 = $info;
   $13 = (_zip_parse_end_of_central_dir($11,$12,$data_start,$cent_dir_ofs)|0);
   $14 = ($13|0)!=(0);
   if ($14) {
    $15 = $in;
    $16 = $info;
    $17 = HEAP32[$data_start>>2]|0;
    $18 = HEAP32[$cent_dir_ofs>>2]|0;
    $19 = (_zip_load_entries($15,$16,$17,$18)|0);
    $20 = ($19|0)!=(0);
    if ($20) {
     $21 = $in;
     (___PHYSFS_platformClose($21)|0);
     $22 = $info;
     $0 = $22;
     $36 = $0;
     STACKTOP = sp;return ($36|0);
    }
   }
  }
 }
 $23 = $info;
 $24 = ($23|0)!=(0|0);
 if ($24) {
  $25 = $info;
  $26 = HEAP32[$25>>2]|0;
  $27 = ($26|0)!=(0|0);
  if ($27) {
   $28 = HEAP32[(732260)>>2]|0;
   $29 = $info;
   $30 = HEAP32[$29>>2]|0;
   FUNCTION_TABLE_vi[$28 & 8191]($30);
  }
  $31 = HEAP32[(732260)>>2]|0;
  $32 = $info;
  FUNCTION_TABLE_vi[$31 & 8191]($32);
 }
 $33 = $in;
 $34 = ($33|0)!=(0|0);
 if ($34) {
  $35 = $in;
  (___PHYSFS_platformClose($35)|0);
 }
 $0 = 0;
 $36 = $0;
 STACKTOP = sp;return ($36|0);
}
function _zip_create_zipinfo($name) {
 $name = $name|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $info = 0, $ptr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $name;
 $2 = HEAP32[(732252)>>2]|0;
 $3 = (FUNCTION_TABLE_iii[$2 & 8191](12,0)|0);
 $info = $3;
 $4 = $info;
 $5 = ($4|0)==(0|0);
 if ($5) {
  ___PHYSFS_setError(596704);
  $0 = 0;
  $22 = $0;
  STACKTOP = sp;return ($22|0);
 }
 $6 = $info;
 ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;
 $7 = HEAP32[(732252)>>2]|0;
 $8 = $1;
 $9 = (_strlen($8)|0);
 $10 = (($9) + 1)|0;
 $11 = (FUNCTION_TABLE_iii[$7 & 8191]($10,0)|0);
 $ptr = $11;
 $12 = $ptr;
 $13 = ($12|0)==(0|0);
 if ($13) {
  $14 = HEAP32[(732260)>>2]|0;
  $15 = $info;
  FUNCTION_TABLE_vi[$14 & 8191]($15);
  ___PHYSFS_setError(596704);
  $0 = 0;
  $22 = $0;
  STACKTOP = sp;return ($22|0);
 } else {
  $16 = $ptr;
  $17 = $info;
  HEAP32[$17>>2] = $16;
  $18 = $info;
  $19 = HEAP32[$18>>2]|0;
  $20 = $1;
  (_strcpy($19,$20)|0);
  $21 = $info;
  $0 = $21;
  $22 = $0;
  STACKTOP = sp;return ($22|0);
 }
 return (0)|0;
}
function _zip_parse_end_of_central_dir($in,$info,$data_start,$central_dir_ofs) {
 $in = $in|0;
 $info = $info|0;
 $data_start = $data_start|0;
 $central_dir_ofs = $central_dir_ofs|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $len = 0, $pos = 0, $ui16 = 0, $ui32 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ui32 = sp + 16|0;
 $ui16 = sp + 40|0;
 $len = sp + 8|0;
 $pos = sp;
 $1 = $in;
 $2 = $info;
 $3 = $data_start;
 $4 = $central_dir_ofs;
 $5 = $1;
 $6 = (_zip_find_end_of_central_dir($5,$len)|0);
 $7 = tempRet0;
 $8 = $pos;
 $9 = $8;
 HEAP32[$9>>2] = $6;
 $10 = (($8) + 4)|0;
 $11 = $10;
 HEAP32[$11>>2] = $7;
 $12 = $pos;
 $13 = $12;
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12) + 4)|0;
 $16 = $15;
 $17 = HEAP32[$16>>2]|0;
 $18 = ($14|0)==(-1);
 $19 = ($17|0)==(-1);
 $20 = $18 & $19;
 do {
  if ($20) {
   ___PHYSFS_setError(0);
   $0 = 0;
  } else {
   $21 = $1;
   $22 = $pos;
   $23 = $22;
   $24 = HEAP32[$23>>2]|0;
   $25 = (($22) + 4)|0;
   $26 = $25;
   $27 = HEAP32[$26>>2]|0;
   $28 = (___PHYSFS_platformSeek($21,$24,$27)|0);
   $29 = ($28|0)!=(0);
   if (!($29)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $30 = $1;
   $31 = (_readui32($30,$ui32)|0);
   $32 = ($31|0)!=(0);
   if (!($32)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $33 = HEAP32[$ui32>>2]|0;
   $34 = ($33|0)!=(101010256);
   if ($34) {
    ___PHYSFS_setError(517325);
    $0 = 0;
    break;
   }
   $35 = $1;
   $36 = (_readui16($35,$ui16)|0);
   $37 = ($36|0)!=(0);
   if (!($37)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $38 = HEAP16[$ui16>>1]|0;
   $39 = $38&65535;
   $40 = ($39|0)!=(0);
   if ($40) {
    ___PHYSFS_setError(517361);
    $0 = 0;
    break;
   }
   $41 = $1;
   $42 = (_readui16($41,$ui16)|0);
   $43 = ($42|0)!=(0);
   if (!($43)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $44 = HEAP16[$ui16>>1]|0;
   $45 = $44&65535;
   $46 = ($45|0)!=(0);
   if ($46) {
    ___PHYSFS_setError(517361);
    $0 = 0;
    break;
   }
   $47 = $1;
   $48 = (_readui16($47,$ui16)|0);
   $49 = ($48|0)!=(0);
   if (!($49)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $50 = $1;
   $51 = $2;
   $52 = ((($51)) + 4|0);
   $53 = (_readui16($50,$52)|0);
   $54 = ($53|0)!=(0);
   if (!($54)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $55 = HEAP16[$ui16>>1]|0;
   $56 = $55&65535;
   $57 = $2;
   $58 = ((($57)) + 4|0);
   $59 = HEAP16[$58>>1]|0;
   $60 = $59&65535;
   $61 = ($56|0)!=($60|0);
   if ($61) {
    ___PHYSFS_setError(517361);
    $0 = 0;
    break;
   }
   $62 = $1;
   $63 = (_readui32($62,$ui32)|0);
   $64 = ($63|0)!=(0);
   if (!($64)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $65 = $1;
   $66 = $4;
   $67 = (_readui32($65,$66)|0);
   $68 = ($67|0)!=(0);
   if (!($68)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $69 = $pos;
   $70 = $69;
   $71 = HEAP32[$70>>2]|0;
   $72 = (($69) + 4)|0;
   $73 = $72;
   $74 = HEAP32[$73>>2]|0;
   $75 = $4;
   $76 = HEAP32[$75>>2]|0;
   $77 = HEAP32[$ui32>>2]|0;
   $78 = (($76) + ($77))|0;
   $79 = ($74|0)<(0);
   $80 = ($71>>>0)<($78>>>0);
   $81 = ($74|0)==(0);
   $82 = $81 & $80;
   $83 = $79 | $82;
   if ($83) {
    ___PHYSFS_setError(517361);
    $0 = 0;
    break;
   }
   $84 = $pos;
   $85 = $84;
   $86 = HEAP32[$85>>2]|0;
   $87 = (($84) + 4)|0;
   $88 = $87;
   $89 = HEAP32[$88>>2]|0;
   $90 = $4;
   $91 = HEAP32[$90>>2]|0;
   $92 = HEAP32[$ui32>>2]|0;
   $93 = (($91) + ($92))|0;
   $94 = (_i64Subtract(($86|0),($89|0),($93|0),0)|0);
   $95 = tempRet0;
   $96 = $3;
   HEAP32[$96>>2] = $94;
   $97 = $3;
   $98 = HEAP32[$97>>2]|0;
   $99 = $4;
   $100 = HEAP32[$99>>2]|0;
   $101 = (($100) + ($98))|0;
   HEAP32[$99>>2] = $101;
   $102 = $1;
   $103 = (_readui16($102,$ui16)|0);
   $104 = ($103|0)!=(0);
   if (!($104)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $105 = $pos;
   $106 = $105;
   $107 = HEAP32[$106>>2]|0;
   $108 = (($105) + 4)|0;
   $109 = $108;
   $110 = HEAP32[$109>>2]|0;
   $111 = (_i64Add(($107|0),($110|0),22,0)|0);
   $112 = tempRet0;
   $113 = HEAP16[$ui16>>1]|0;
   $114 = $113&65535;
   $115 = (_i64Add(($111|0),($112|0),($114|0),0)|0);
   $116 = tempRet0;
   $117 = $len;
   $118 = $117;
   $119 = HEAP32[$118>>2]|0;
   $120 = (($117) + 4)|0;
   $121 = $120;
   $122 = HEAP32[$121>>2]|0;
   $123 = ($115|0)!=($119|0);
   $124 = ($116|0)!=($122|0);
   $125 = $123 | $124;
   if ($125) {
    ___PHYSFS_setError(517361);
    $0 = 0;
    break;
   } else {
    $0 = 1;
    break;
   }
  }
 } while(0);
 $126 = $0;
 STACKTOP = sp;return ($126|0);
}
function _readui16($in,$val) {
 $in = $in|0;
 $val = $val|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp + 12|0;
 $1 = $in;
 $2 = $val;
 $3 = $1;
 $4 = (___PHYSFS_platformRead($3,$v,2,1)|0);
 $5 = tempRet0;
 $6 = ($4|0)!=(1);
 $7 = ($5|0)!=(0);
 $8 = $6 | $7;
 if ($8) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $12 = $0;
  STACKTOP = sp;return ($12|0);
 } else {
  $9 = HEAP16[$v>>1]|0;
  $10 = (_PHYSFS_swapULE16($9)|0);
  $11 = $2;
  HEAP16[$11>>1] = $10;
  $0 = 1;
  $12 = $0;
  STACKTOP = sp;return ($12|0);
 }
 return (0)|0;
}
function _zip_load_entries($in,$info,$data_ofs,$central_ofs) {
 $in = $in|0;
 $info = $info|0;
 $data_ofs = $data_ofs|0;
 $central_ofs = $central_ofs|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i = 0, $max = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $in;
 $2 = $info;
 $3 = $data_ofs;
 $4 = $central_ofs;
 $5 = $2;
 $6 = ((($5)) + 4|0);
 $7 = HEAP16[$6>>1]|0;
 $8 = $7&65535;
 $max = $8;
 $9 = $1;
 $10 = $4;
 $11 = (___PHYSFS_platformSeek($9,$10,0)|0);
 $12 = ($11|0)!=(0);
 if (!($12)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $45 = $0;
  STACKTOP = sp;return ($45|0);
 }
 $13 = HEAP32[(732252)>>2]|0;
 $14 = $max;
 $15 = ($14*48)|0;
 $16 = (FUNCTION_TABLE_iii[$13 & 8191]($15,0)|0);
 $17 = $2;
 $18 = ((($17)) + 8|0);
 HEAP32[$18>>2] = $16;
 $19 = $2;
 $20 = ((($19)) + 8|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ($21|0)==(0|0);
 if ($22) {
  ___PHYSFS_setError(596704);
  $0 = 0;
  $45 = $0;
  STACKTOP = sp;return ($45|0);
 }
 $i = 0;
 while(1) {
  $23 = $i;
  $24 = $max;
  $25 = ($23>>>0)<($24>>>0);
  if (!($25)) {
   label = 10;
   break;
  }
  $26 = $1;
  $27 = $i;
  $28 = $2;
  $29 = ((($28)) + 8|0);
  $30 = HEAP32[$29>>2]|0;
  $31 = (($30) + (($27*48)|0)|0);
  $32 = $3;
  $33 = (_zip_load_entry($26,$31,$32)|0);
  $34 = ($33|0)!=(0);
  if (!($34)) {
   label = 8;
   break;
  }
  $39 = $i;
  $40 = (($39) + 1)|0;
  $i = $40;
 }
 if ((label|0) == 8) {
  $35 = $2;
  $36 = ((($35)) + 8|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = $i;
  _zip_free_entries($37,$38);
  $0 = 0;
  $45 = $0;
  STACKTOP = sp;return ($45|0);
 }
 else if ((label|0) == 10) {
  $41 = $2;
  $42 = ((($41)) + 8|0);
  $43 = HEAP32[$42>>2]|0;
  $44 = $max;
  ___PHYSFS_sort($43,$44,4285,4286);
  $0 = 1;
  $45 = $0;
  STACKTOP = sp;return ($45|0);
 }
 return (0)|0;
}
function _zip_load_entry($in,$entry,$ofs_fixup) {
 $in = $in|0;
 $entry = $entry|0;
 $ofs_fixup = $ofs_fixup|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $commentlen = 0, $external_attr = 0, $extralen = 0;
 var $fnamelen = 0, $si64 = 0, $ui16 = 0, $ui32 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $fnamelen = sp + 38|0;
 $extralen = sp + 36|0;
 $commentlen = sp + 34|0;
 $external_attr = sp + 12|0;
 $ui16 = sp + 32|0;
 $ui32 = sp + 8|0;
 $si64 = sp;
 $1 = $in;
 $2 = $entry;
 $3 = $ofs_fixup;
 $4 = $1;
 $5 = (_readui32($4,$ui32)|0);
 $6 = ($5|0)!=(0);
 if (!($6)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $7 = HEAP32[$ui32>>2]|0;
 $8 = ($7|0)!=(33639248);
 if ($8) {
  ___PHYSFS_setError(517386);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $9 = $1;
 $10 = $2;
 $11 = ((($10)) + 16|0);
 $12 = (_readui16($9,$11)|0);
 $13 = ($12|0)!=(0);
 if (!($13)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $14 = $1;
 $15 = $2;
 $16 = ((($15)) + 18|0);
 $17 = (_readui16($14,$16)|0);
 $18 = ($17|0)!=(0);
 if (!($18)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $19 = $1;
 $20 = (_readui16($19,$ui16)|0);
 $21 = ($20|0)!=(0);
 if (!($21)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $22 = $1;
 $23 = $2;
 $24 = ((($23)) + 20|0);
 $25 = (_readui16($22,$24)|0);
 $26 = ($25|0)!=(0);
 if (!($26)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $27 = $1;
 $28 = (_readui32($27,$ui32)|0);
 $29 = ($28|0)!=(0);
 if (!($29)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $30 = HEAP32[$ui32>>2]|0;
 $31 = (_zip_dos_time_to_physfs_time($30)|0);
 $32 = tempRet0;
 $33 = $2;
 $34 = ((($33)) + 40|0);
 $35 = $34;
 $36 = $35;
 HEAP32[$36>>2] = $31;
 $37 = (($35) + 4)|0;
 $38 = $37;
 HEAP32[$38>>2] = $32;
 $39 = $1;
 $40 = $2;
 $41 = ((($40)) + 24|0);
 $42 = (_readui32($39,$41)|0);
 $43 = ($42|0)!=(0);
 if (!($43)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $44 = $1;
 $45 = $2;
 $46 = ((($45)) + 28|0);
 $47 = (_readui32($44,$46)|0);
 $48 = ($47|0)!=(0);
 if (!($48)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $49 = $1;
 $50 = $2;
 $51 = ((($50)) + 32|0);
 $52 = (_readui32($49,$51)|0);
 $53 = ($52|0)!=(0);
 if (!($53)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $54 = $1;
 $55 = (_readui16($54,$fnamelen)|0);
 $56 = ($55|0)!=(0);
 if (!($56)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $57 = $1;
 $58 = (_readui16($57,$extralen)|0);
 $59 = ($58|0)!=(0);
 if (!($59)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $60 = $1;
 $61 = (_readui16($60,$commentlen)|0);
 $62 = ($61|0)!=(0);
 if (!($62)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $63 = $1;
 $64 = (_readui16($63,$ui16)|0);
 $65 = ($64|0)!=(0);
 if (!($65)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $66 = $1;
 $67 = (_readui16($66,$ui16)|0);
 $68 = ($67|0)!=(0);
 if (!($68)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $69 = $1;
 $70 = (_readui32($69,$external_attr)|0);
 $71 = ($70|0)!=(0);
 if (!($71)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $72 = $1;
 $73 = $2;
 $74 = ((($73)) + 12|0);
 $75 = (_readui32($72,$74)|0);
 $76 = ($75|0)!=(0);
 if (!($76)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $77 = $3;
 $78 = $2;
 $79 = ((($78)) + 12|0);
 $80 = HEAP32[$79>>2]|0;
 $81 = (($80) + ($77))|0;
 HEAP32[$79>>2] = $81;
 $82 = $2;
 $83 = ((($82)) + 4|0);
 HEAP32[$83>>2] = 0;
 $84 = $2;
 $85 = HEAP32[$external_attr>>2]|0;
 $86 = (_zip_has_symlink_attr($84,$85)|0);
 $87 = ($86|0)!=(0);
 $88 = $87 ? 1 : 0;
 $89 = $2;
 $90 = ((($89)) + 8|0);
 HEAP32[$90>>2] = $88;
 $91 = HEAP32[(732252)>>2]|0;
 $92 = HEAP16[$fnamelen>>1]|0;
 $93 = $92&65535;
 $94 = (($93) + 1)|0;
 $95 = ($94|0)<(0);
 $96 = $95 << 31 >> 31;
 $97 = (FUNCTION_TABLE_iii[$91 & 8191]($94,$96)|0);
 $98 = $2;
 HEAP32[$98>>2] = $97;
 $99 = $2;
 $100 = HEAP32[$99>>2]|0;
 $101 = ($100|0)==(0|0);
 if ($101) {
  ___PHYSFS_setError(596704);
  $0 = 0;
  $156 = $0;
  STACKTOP = sp;return ($156|0);
 }
 $102 = $1;
 $103 = $2;
 $104 = HEAP32[$103>>2]|0;
 $105 = HEAP16[$fnamelen>>1]|0;
 $106 = $105&65535;
 $107 = (___PHYSFS_platformRead($102,$104,$106,1)|0);
 $108 = tempRet0;
 $109 = ($107|0)!=(1);
 $110 = ($108|0)!=(0);
 $111 = $109 | $110;
 do {
  if (!($111)) {
   $112 = HEAP16[$fnamelen>>1]|0;
   $113 = $112&65535;
   $114 = $2;
   $115 = HEAP32[$114>>2]|0;
   $116 = (($115) + ($113)|0);
   HEAP8[$116>>0] = 0;
   $117 = $2;
   $118 = $2;
   $119 = HEAP32[$118>>2]|0;
   _zip_convert_dos_path($117,$119);
   $120 = $1;
   $121 = (___PHYSFS_platformTell($120)|0);
   $122 = tempRet0;
   $123 = $si64;
   $124 = $123;
   HEAP32[$124>>2] = $121;
   $125 = (($123) + 4)|0;
   $126 = $125;
   HEAP32[$126>>2] = $122;
   $127 = $si64;
   $128 = $127;
   $129 = HEAP32[$128>>2]|0;
   $130 = (($127) + 4)|0;
   $131 = $130;
   $132 = HEAP32[$131>>2]|0;
   $133 = ($129|0)==(-1);
   $134 = ($132|0)==(-1);
   $135 = $133 & $134;
   if ($135) {
    break;
   }
   $136 = $1;
   $137 = $si64;
   $138 = $137;
   $139 = HEAP32[$138>>2]|0;
   $140 = (($137) + 4)|0;
   $141 = $140;
   $142 = HEAP32[$141>>2]|0;
   $143 = HEAP16[$extralen>>1]|0;
   $144 = $143&65535;
   $145 = (_i64Add(($139|0),($142|0),($144|0),0)|0);
   $146 = tempRet0;
   $147 = HEAP16[$commentlen>>1]|0;
   $148 = $147&65535;
   $149 = (_i64Add(($145|0),($146|0),($148|0),0)|0);
   $150 = tempRet0;
   $151 = (___PHYSFS_platformSeek($136,$149,$150)|0);
   $152 = ($151|0)!=(0);
   if (!($152)) {
    break;
   }
   $0 = 1;
   $156 = $0;
   STACKTOP = sp;return ($156|0);
  }
 } while(0);
 $153 = HEAP32[(732260)>>2]|0;
 $154 = $2;
 $155 = HEAP32[$154>>2]|0;
 FUNCTION_TABLE_vi[$153 & 8191]($155);
 $0 = 0;
 $156 = $0;
 STACKTOP = sp;return ($156|0);
}
function _zip_dos_time_to_physfs_time($dostime) {
 $dostime = $dostime|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dosdate = 0, $unixtime = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $unixtime = sp;
 $0 = $dostime;
 dest=$unixtime; stop=dest+44|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $1 = $0;
 $2 = $1 >>> 16;
 $3 = $2 & 65535;
 $dosdate = $3;
 $4 = $0;
 $5 = $4 & 65535;
 $0 = $5;
 $6 = $dosdate;
 $7 = $6 >>> 9;
 $8 = $7 & 127;
 $9 = (($8) + 80)|0;
 $10 = ((($unixtime)) + 20|0);
 HEAP32[$10>>2] = $9;
 $11 = $dosdate;
 $12 = $11 >>> 5;
 $13 = $12 & 15;
 $14 = (($13) - 1)|0;
 $15 = ((($unixtime)) + 16|0);
 HEAP32[$15>>2] = $14;
 $16 = $dosdate;
 $17 = $16 & 31;
 $18 = ((($unixtime)) + 12|0);
 HEAP32[$18>>2] = $17;
 $19 = $0;
 $20 = $19 >>> 11;
 $21 = $20 & 31;
 $22 = ((($unixtime)) + 8|0);
 HEAP32[$22>>2] = $21;
 $23 = $0;
 $24 = $23 >>> 5;
 $25 = $24 & 63;
 $26 = ((($unixtime)) + 4|0);
 HEAP32[$26>>2] = $25;
 $27 = $0;
 $28 = $27 << 1;
 $29 = $28 & 62;
 HEAP32[$unixtime>>2] = $29;
 $30 = ((($unixtime)) + 32|0);
 HEAP32[$30>>2] = -1;
 $31 = (_mktime(($unixtime|0))|0);
 $32 = ($31|0)<(0);
 $33 = $32 << 31 >> 31;
 tempRet0 = ($33);
 STACKTOP = sp;return ($31|0);
}
function _zip_has_symlink_attr($entry,$extern_attr) {
 $entry = $entry|0;
 $extern_attr = $extern_attr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $xattr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $entry;
 $1 = $extern_attr;
 $2 = $1;
 $3 = $2 >>> 16;
 $4 = $3 & 65535;
 $5 = $4&65535;
 $xattr = $5;
 $6 = $0;
 $7 = ((($6)) + 16|0);
 $8 = HEAP16[$7>>1]|0;
 $9 = $8&65535;
 $10 = (_zip_version_does_symlinks($9)|0);
 $11 = ($10|0)!=(0);
 if ($11) {
  $12 = $0;
  $13 = ((($12)) + 32|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ($14>>>0)>(0);
  if ($15) {
   $16 = $xattr;
   $17 = $16&65535;
   $18 = $17 & 61440;
   $19 = ($18|0)==(40960);
   $21 = $19;
  } else {
   $21 = 0;
  }
 } else {
  $21 = 0;
 }
 $20 = $21&1;
 STACKTOP = sp;return ($20|0);
}
function _zip_version_does_symlinks($version) {
 $version = $version|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $hosttype = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $version;
 $retval = 0;
 $1 = $0;
 $2 = $1 >>> 8;
 $3 = $2 & 255;
 $4 = $3&255;
 $hosttype = $4;
 $5 = $hosttype;
 $6 = $5&255;
 switch ($6|0) {
 case 18: case 15: case 13: case 14: case 11: case 6: case 4: case 2: case 1: case 0:  {
  break;
 }
 default: {
  $retval = 1;
 }
 }
 $7 = $retval;
 STACKTOP = sp;return ($7|0);
}
function _zip_convert_dos_path($entry,$path) {
 $entry = $entry|0;
 $path = $path|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $hosttype = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $entry;
 $1 = $path;
 $2 = $0;
 $3 = ((($2)) + 16|0);
 $4 = HEAP16[$3>>1]|0;
 $5 = $4&65535;
 $6 = $5 >> 8;
 $7 = $6 & 255;
 $8 = $7&255;
 $hosttype = $8;
 $9 = $hosttype;
 $10 = $9&255;
 $11 = ($10|0)==(0);
 if (!($11)) {
  STACKTOP = sp;return;
 }
 while(1) {
  $12 = $1;
  $13 = HEAP8[$12>>0]|0;
  $14 = ($13<<24>>24)!=(0);
  if (!($14)) {
   break;
  }
  $15 = $1;
  $16 = HEAP8[$15>>0]|0;
  $17 = $16 << 24 >> 24;
  $18 = ($17|0)==(92);
  if ($18) {
   $19 = $1;
   HEAP8[$19>>0] = 47;
  }
  $20 = $1;
  $21 = ((($20)) + 1|0);
  $1 = $21;
 }
 STACKTOP = sp;return;
}
function _zip_free_entries($entries,$max) {
 $entries = $entries|0;
 $max = $max|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $entry = 0, $i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $entries;
 $1 = $max;
 $i = 0;
 while(1) {
  $2 = $i;
  $3 = $1;
  $4 = ($2>>>0)<($3>>>0);
  if (!($4)) {
   break;
  }
  $5 = $i;
  $6 = $0;
  $7 = (($6) + (($5*48)|0)|0);
  $entry = $7;
  $8 = $entry;
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)!=(0|0);
  if ($10) {
   $11 = HEAP32[(732260)>>2]|0;
   $12 = $entry;
   $13 = HEAP32[$12>>2]|0;
   FUNCTION_TABLE_vi[$11 & 8191]($13);
  }
  $14 = $i;
  $15 = (($14) + 1)|0;
  $i = $15;
 }
 $16 = HEAP32[(732260)>>2]|0;
 $17 = $0;
 FUNCTION_TABLE_vi[$16 & 8191]($17);
 STACKTOP = sp;return;
}
function _zip_entry_cmp($_a,$one,$two) {
 $_a = $_a|0;
 $one = $one|0;
 $two = $two|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $a = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $_a;
 $2 = $one;
 $3 = $two;
 $4 = $2;
 $5 = $3;
 $6 = ($4|0)!=($5|0);
 if ($6) {
  $7 = $1;
  $a = $7;
  $8 = $2;
  $9 = $a;
  $10 = (($9) + (($8*48)|0)|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = $3;
  $13 = $a;
  $14 = (($13) + (($12*48)|0)|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = (_strcmp($11,$15)|0);
  $0 = $16;
  $17 = $0;
  STACKTOP = sp;return ($17|0);
 } else {
  $0 = 0;
  $17 = $0;
  STACKTOP = sp;return ($17|0);
 }
 return (0)|0;
}
function _zip_entry_swap($_a,$one,$two) {
 $_a = $_a|0;
 $one = $one|0;
 $two = $two|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $first = 0, $second = 0, $tmp = 0, dest = 0;
 var label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tmp = sp;
 $0 = $_a;
 $1 = $one;
 $2 = $two;
 $3 = $1;
 $4 = $2;
 $5 = ($3|0)!=($4|0);
 if (!($5)) {
  STACKTOP = sp;return;
 }
 $6 = $1;
 $7 = $0;
 $8 = (($7) + (($6*48)|0)|0);
 $first = $8;
 $9 = $2;
 $10 = $0;
 $11 = (($10) + (($9*48)|0)|0);
 $second = $11;
 $12 = $first;
 dest=$tmp; src=$12; stop=dest+48|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 $13 = $first;
 $14 = $second;
 dest=$13; src=$14; stop=dest+48|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 $15 = $second;
 dest=$15; src=$tmp; stop=dest+48|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 STACKTOP = sp;return;
}
function _ZIP_enumerateFiles($opaque,$dname,$omitSymLinks,$cb,$origdir,$callbackdata) {
 $opaque = $opaque|0;
 $dname = $dname|0;
 $omitSymLinks = $omitSymLinks|0;
 $cb = $cb|0;
 $origdir = $origdir|0;
 $callbackdata = $callbackdata|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $add = 0, $dlen = 0, $dlen_inc = 0, $e = 0, $e_new = 0, $i = 0;
 var $info = 0, $ln = 0, $max = 0, $ptr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $dname;
 $2 = $omitSymLinks;
 $3 = $cb;
 $4 = $origdir;
 $5 = $callbackdata;
 $6 = $0;
 $info = $6;
 $7 = $info;
 $8 = $1;
 $9 = (_zip_find_start_of_dir($7,$8,0)|0);
 $i = $9;
 $10 = $i;
 $11 = ($10|0)==(-1);
 if ($11) {
  STACKTOP = sp;return;
 }
 $12 = $1;
 $13 = (_strlen($12)|0);
 $dlen = $13;
 $14 = $dlen;
 $15 = ($14|0)>(0);
 if ($15) {
  $16 = $dlen;
  $17 = (($16) - 1)|0;
  $18 = $1;
  $19 = (($18) + ($17)|0);
  $20 = HEAP8[$19>>0]|0;
  $21 = $20 << 24 >> 24;
  $22 = ($21|0)==(47);
  if ($22) {
   $23 = $dlen;
   $24 = (($23) + -1)|0;
   $dlen = $24;
  }
 }
 $25 = $dlen;
 $26 = ($25|0)>(0);
 $27 = $26 ? 1 : 0;
 $28 = $dlen;
 $29 = (($27) + ($28))|0;
 $dlen_inc = $29;
 $30 = $info;
 $31 = ((($30)) + 4|0);
 $32 = HEAP16[$31>>1]|0;
 $33 = $32&65535;
 $max = $33;
 L8: while(1) {
  $34 = $i;
  $35 = $max;
  $36 = ($34|0)<($35|0);
  if (!($36)) {
   label = 20;
   break;
  }
  $37 = $i;
  $38 = $info;
  $39 = ((($38)) + 8|0);
  $40 = HEAP32[$39>>2]|0;
  $41 = (($40) + (($37*48)|0)|0);
  $42 = HEAP32[$41>>2]|0;
  $e = $42;
  $43 = $dlen;
  $44 = ($43|0)!=(0);
  if ($44) {
   $45 = $e;
   $46 = $1;
   $47 = $dlen;
   $48 = (_strncmp($45,$46,$47)|0);
   $49 = ($48|0)!=(0);
   if ($49) {
    label = 20;
    break;
   }
   $50 = $dlen;
   $51 = $e;
   $52 = (($51) + ($50)|0);
   $53 = HEAP8[$52>>0]|0;
   $54 = $53 << 24 >> 24;
   $55 = ($54|0)!=(47);
   if ($55) {
    label = 20;
    break;
   }
  }
  $56 = $2;
  $57 = ($56|0)!=(0);
  if ($57) {
   $58 = $i;
   $59 = $info;
   $60 = ((($59)) + 8|0);
   $61 = HEAP32[$60>>2]|0;
   $62 = (($61) + (($58*48)|0)|0);
   $63 = (_zip_entry_is_symlink($62)|0);
   $64 = ($63|0)!=(0);
   if ($64) {
    $65 = $i;
    $66 = (($65) + 1)|0;
    $i = $66;
    continue;
   }
  }
  $67 = $e;
  $68 = $dlen_inc;
  $69 = (($67) + ($68)|0);
  $add = $69;
  $70 = $add;
  $71 = (_strchr($70,47)|0);
  $ptr = $71;
  $72 = $ptr;
  $73 = ($72|0)!=(0|0);
  if ($73) {
   $74 = $ptr;
   $75 = $add;
   $76 = $74;
   $77 = $75;
   $78 = (($76) - ($77))|0;
   $81 = $78;
  } else {
   $79 = $add;
   $80 = (_strlen($79)|0);
   $81 = $80;
  }
  $ln = $81;
  $82 = $3;
  $83 = $5;
  $84 = $4;
  $85 = $add;
  $86 = $ln;
  _doEnumCallback($82,$83,$84,$85,$86);
  $87 = $dlen_inc;
  $88 = $ln;
  $89 = (($88) + ($87))|0;
  $ln = $89;
  while(1) {
   $90 = $i;
   $91 = (($90) + 1)|0;
   $i = $91;
   $92 = $max;
   $93 = ($91|0)<($92|0);
   $94 = $ptr;
   $95 = ($94|0)!=(0|0);
   $96 = $93 ? $95 : 0;
   if (!($96)) {
    continue L8;
   }
   $97 = $i;
   $98 = $info;
   $99 = ((($98)) + 8|0);
   $100 = HEAP32[$99>>2]|0;
   $101 = (($100) + (($97*48)|0)|0);
   $102 = HEAP32[$101>>2]|0;
   $e_new = $102;
   $103 = $e;
   $104 = $e_new;
   $105 = $ln;
   $106 = (_strncmp($103,$104,$105)|0);
   $107 = ($106|0)!=(0);
   if ($107) {
    continue L8;
   }
   $108 = $ln;
   $109 = $e_new;
   $110 = (($109) + ($108)|0);
   $111 = HEAP8[$110>>0]|0;
   $112 = $111 << 24 >> 24;
   $113 = ($112|0)!=(47);
   if ($113) {
    continue L8;
   }
  }
 }
 if ((label|0) == 20) {
  STACKTOP = sp;return;
 }
}
function _zip_find_start_of_dir($info,$path,$stop_on_first_find) {
 $info = $info|0;
 $path = $path|0;
 $stop_on_first_find = $stop_on_first_find|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $8 = 0, $9 = 0, $ch = 0, $dlen = 0;
 var $hi = 0, $lo = 0, $middle = 0, $name = 0, $rc = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $info;
 $2 = $path;
 $3 = $stop_on_first_find;
 $lo = 0;
 $4 = $1;
 $5 = ((($4)) + 4|0);
 $6 = HEAP16[$5>>1]|0;
 $7 = $6&65535;
 $8 = (($7) - 1)|0;
 $hi = $8;
 $9 = $2;
 $10 = (_strlen($9)|0);
 $dlen = $10;
 $retval = -1;
 $11 = $2;
 $12 = HEAP8[$11>>0]|0;
 $13 = $12 << 24 >> 24;
 $14 = ($13|0)==(0);
 if ($14) {
  $0 = 0;
  $77 = $0;
  STACKTOP = sp;return ($77|0);
 }
 $15 = $dlen;
 $16 = ($15>>>0)>(0);
 if ($16) {
  $17 = $dlen;
  $18 = (($17) - 1)|0;
  $19 = $2;
  $20 = (($19) + ($18)|0);
  $21 = HEAP8[$20>>0]|0;
  $22 = $21 << 24 >> 24;
  $23 = ($22|0)==(47);
  if ($23) {
   $24 = $dlen;
   $25 = (($24) + -1)|0;
   $dlen = $25;
  }
 }
 L8: while(1) {
  $26 = $lo;
  $27 = $hi;
  $28 = ($26|0)<=($27|0);
  if (!($28)) {
   label = 20;
   break;
  }
  $29 = $lo;
  $30 = $hi;
  $31 = $lo;
  $32 = (($30) - ($31))|0;
  $33 = (($32|0) / 2)&-1;
  $34 = (($29) + ($33))|0;
  $middle = $34;
  $35 = $middle;
  $36 = $1;
  $37 = ((($36)) + 8|0);
  $38 = HEAP32[$37>>2]|0;
  $39 = (($38) + (($35*48)|0)|0);
  $40 = HEAP32[$39>>2]|0;
  $name = $40;
  $41 = $2;
  $42 = $name;
  $43 = $dlen;
  $44 = (_strncmp($41,$42,$43)|0);
  $rc = $44;
  $45 = $rc;
  $46 = ($45|0)==(0);
  do {
   if ($46) {
    $47 = $dlen;
    $48 = $name;
    $49 = (($48) + ($47)|0);
    $50 = HEAP8[$49>>0]|0;
    $ch = $50;
    $51 = $ch;
    $52 = $51 << 24 >> 24;
    $53 = (47)<($52|0);
    if ($53) {
     $rc = -1;
     break;
    }
    $54 = $ch;
    $55 = $54 << 24 >> 24;
    $56 = (47)>($55|0);
    if ($56) {
     $rc = 1;
     break;
    }
    $57 = $3;
    $58 = ($57|0)!=(0);
    if ($58) {
     label = 13;
     break L8;
    }
    $60 = $dlen;
    $61 = (($60) + 1)|0;
    $62 = $name;
    $63 = (($62) + ($61)|0);
    $64 = HEAP8[$63>>0]|0;
    $65 = $64 << 24 >> 24;
    $66 = ($65|0)==(0);
    $67 = $middle;
    if ($66) {
     label = 15;
     break L8;
    }
    $retval = $67;
    $69 = $middle;
    $70 = (($69) - 1)|0;
    $hi = $70;
   }
  } while(0);
  $71 = $rc;
  $72 = ($71|0)>(0);
  $73 = $middle;
  if ($72) {
   $74 = (($73) + 1)|0;
   $lo = $74;
   continue;
  } else {
   $75 = (($73) - 1)|0;
   $hi = $75;
   continue;
  }
 }
 if ((label|0) == 13) {
  $59 = $middle;
  $0 = $59;
  $77 = $0;
  STACKTOP = sp;return ($77|0);
 }
 else if ((label|0) == 15) {
  $68 = (($67) + 1)|0;
  $0 = $68;
  $77 = $0;
  STACKTOP = sp;return ($77|0);
 }
 else if ((label|0) == 20) {
  $76 = $retval;
  $0 = $76;
  $77 = $0;
  STACKTOP = sp;return ($77|0);
 }
 return (0)|0;
}
function _zip_entry_is_symlink($entry) {
 $entry = $entry|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $entry;
 $1 = $0;
 $2 = ((($1)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(1);
 if ($4) {
  $14 = 1;
 } else {
  $5 = $0;
  $6 = ((($5)) + 8|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)==(5);
  if ($8) {
   $14 = 1;
  } else {
   $9 = $0;
   $10 = ((($9)) + 4|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = ($11|0)!=(0|0);
   $14 = $12;
  }
 }
 $13 = $14&1;
 STACKTOP = sp;return ($13|0);
}
function _doEnumCallback($cb,$callbackdata,$odir,$str,$ln) {
 $cb = $cb|0;
 $callbackdata = $callbackdata|0;
 $odir = $odir|0;
 $str = $str|0;
 $ln = $ln|0;
 var $$alloca_mul = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $newstr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $cb;
 $1 = $callbackdata;
 $2 = $odir;
 $3 = $str;
 $4 = $ln;
 $5 = $4;
 $6 = (($5) + 1)|0;
 $7 = ($6|0)<(128);
 if ($7) {
  $8 = $4;
  $9 = (($8) + 1)|0;
  $10 = (($9) + 1)|0;
  $$alloca_mul = $10;
  $11 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
  $16 = $11;
 } else {
  $16 = 0;
 }
 $12 = $4;
 $13 = (($12) + 1)|0;
 $14 = ($13|0)<(0);
 $15 = $14 << 31 >> 31;
 $17 = (___PHYSFS_initSmallAlloc($16,$13,$15)|0);
 $newstr = $17;
 $18 = $newstr;
 $19 = ($18|0)==(0|0);
 if ($19) {
  STACKTOP = sp;return;
 }
 $20 = $newstr;
 $21 = $3;
 $22 = $4;
 _memcpy(($20|0),($21|0),($22|0))|0;
 $23 = $4;
 $24 = $newstr;
 $25 = (($24) + ($23)|0);
 HEAP8[$25>>0] = 0;
 $26 = $0;
 $27 = $1;
 $28 = $2;
 $29 = $newstr;
 FUNCTION_TABLE_viii[$26 & 8191]($27,$28,$29);
 $30 = $newstr;
 ___PHYSFS_smallFree($30);
 STACKTOP = sp;return;
}
function _ZIP_exists($opaque,$name) {
 $opaque = $opaque|0;
 $name = $name|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $entry = 0, $info = 0, $isDir = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $isDir = sp + 8|0;
 $0 = $opaque;
 $1 = $name;
 $2 = $0;
 $info = $2;
 $3 = $info;
 $4 = $1;
 $5 = (_zip_find_entry($3,$4,$isDir)|0);
 $entry = $5;
 $6 = $entry;
 $7 = ($6|0)!=(0|0);
 $8 = HEAP32[$isDir>>2]|0;
 $9 = ($8|0)!=(0);
 $10 = $7 ? 1 : $9;
 $11 = $10&1;
 STACKTOP = sp;return ($11|0);
}
function _zip_find_entry($info,$path,$isDir) {
 $info = $info|0;
 $path = $path|0;
 $isDir = $isDir|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $a = 0, $hi = 0, $i = 0, $lo = 0, $middle = 0, $pathlen = 0, $rc = 0, $thispath = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $info;
 $2 = $path;
 $3 = $isDir;
 $4 = $1;
 $5 = ((($4)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $a = $6;
 $7 = $2;
 $8 = (_strlen($7)|0);
 $pathlen = $8;
 $lo = 0;
 $9 = $1;
 $10 = ((($9)) + 4|0);
 $11 = HEAP16[$10>>1]|0;
 $12 = $11&65535;
 $13 = (($12) - 1)|0;
 $hi = $13;
 $thispath = 0;
 while(1) {
  $14 = $lo;
  $15 = $hi;
  $16 = ($14|0)<=($15|0);
  if (!($16)) {
   break;
  }
  $17 = $lo;
  $18 = $hi;
  $19 = $lo;
  $20 = (($18) - ($19))|0;
  $21 = (($20|0) / 2)&-1;
  $22 = (($17) + ($21))|0;
  $middle = $22;
  $23 = $middle;
  $24 = $a;
  $25 = (($24) + (($23*48)|0)|0);
  $26 = HEAP32[$25>>2]|0;
  $thispath = $26;
  $27 = $2;
  $28 = $thispath;
  $29 = $pathlen;
  $30 = (_strncmp($27,$28,$29)|0);
  $rc = $30;
  $31 = $rc;
  $32 = ($31|0)>(0);
  if ($32) {
   $33 = $middle;
   $34 = (($33) + 1)|0;
   $lo = $34;
   continue;
  }
  $35 = $rc;
  $36 = ($35|0)<(0);
  if (!($36)) {
   label = 7;
   break;
  }
  $37 = $middle;
  $38 = (($37) - 1)|0;
  $hi = $38;
 }
 L8: do {
  if ((label|0) == 7) {
   $39 = $3;
   $40 = ($39|0)!=(0|0);
   if ($40) {
    $41 = $pathlen;
    $42 = $thispath;
    $43 = (($42) + ($41)|0);
    $44 = HEAP8[$43>>0]|0;
    $45 = $44 << 24 >> 24;
    $46 = ($45|0)==(47);
    $47 = $46&1;
    $48 = $3;
    HEAP32[$48>>2] = $47;
    $49 = $3;
    $50 = HEAP32[$49>>2]|0;
    $51 = ($50|0)!=(0);
    if ($51) {
     $0 = 0;
     $101 = $0;
     STACKTOP = sp;return ($101|0);
    }
   }
   $52 = $pathlen;
   $53 = $thispath;
   $54 = (($53) + ($52)|0);
   $55 = HEAP8[$54>>0]|0;
   $56 = $55 << 24 >> 24;
   $57 = ($56|0)==(0);
   if ($57) {
    $58 = $middle;
    $59 = $a;
    $60 = (($59) + (($58*48)|0)|0);
    $0 = $60;
    $101 = $0;
    STACKTOP = sp;return ($101|0);
   }
   $61 = $lo;
   $i = $61;
   while(1) {
    $62 = $i;
    $63 = $hi;
    $64 = ($62|0)<($63|0);
    if (!($64)) {
     break L8;
    }
    $65 = $i;
    $66 = $a;
    $67 = (($66) + (($65*48)|0)|0);
    $68 = HEAP32[$67>>2]|0;
    $thispath = $68;
    $69 = $2;
    $70 = $thispath;
    $71 = $pathlen;
    $72 = (_strncmp($69,$70,$71)|0);
    $73 = ($72|0)==(0);
    if ($73) {
     $74 = $3;
     $75 = ($74|0)!=(0|0);
     if ($75) {
      $76 = $pathlen;
      $77 = $thispath;
      $78 = (($77) + ($76)|0);
      $79 = HEAP8[$78>>0]|0;
      $80 = $79 << 24 >> 24;
      $81 = ($80|0)==(47);
      $82 = $81&1;
      $83 = $3;
      HEAP32[$83>>2] = $82;
      $84 = $3;
      $85 = HEAP32[$84>>2]|0;
      $86 = ($85|0)!=(0);
      if ($86) {
       label = 17;
       break;
      }
     }
     $87 = $pathlen;
     $88 = $thispath;
     $89 = (($88) + ($87)|0);
     $90 = HEAP8[$89>>0]|0;
     $91 = $90 << 24 >> 24;
     $92 = ($91|0)==(0);
     if ($92) {
      label = 19;
      break;
     }
    }
    $96 = $i;
    $97 = (($96) + 1)|0;
    $i = $97;
   }
   if ((label|0) == 17) {
    $0 = 0;
    $101 = $0;
    STACKTOP = sp;return ($101|0);
   }
   else if ((label|0) == 19) {
    $93 = $i;
    $94 = $a;
    $95 = (($94) + (($93*48)|0)|0);
    $0 = $95;
    $101 = $0;
    STACKTOP = sp;return ($101|0);
   }
  }
 } while(0);
 $98 = $3;
 $99 = ($98|0)!=(0|0);
 if ($99) {
  $100 = $3;
  HEAP32[$100>>2] = 0;
 }
 ___PHYSFS_setError(517404);
 $0 = 0;
 $101 = $0;
 STACKTOP = sp;return ($101|0);
}
function _ZIP_isDirectory($opaque,$name,$fileExists) {
 $opaque = $opaque|0;
 $name = $name|0;
 $fileExists = $fileExists|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $entry = 0, $in = 0, $info = 0, $isDir = 0, $rc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $isDir = sp + 12|0;
 $1 = $opaque;
 $2 = $name;
 $3 = $fileExists;
 $4 = $1;
 $info = $4;
 $5 = $info;
 $6 = $2;
 $7 = (_zip_find_entry($5,$6,$isDir)|0);
 $entry = $7;
 $8 = HEAP32[$isDir>>2]|0;
 $9 = ($8|0)!=(0);
 $10 = $entry;
 $11 = ($10|0)!=(0|0);
 $12 = $9 ? 1 : $11;
 $13 = $12&1;
 $14 = $3;
 HEAP32[$14>>2] = $13;
 $15 = HEAP32[$isDir>>2]|0;
 $16 = ($15|0)!=(0);
 if ($16) {
  $0 = 1;
  $51 = $0;
  STACKTOP = sp;return ($51|0);
 }
 $17 = $entry;
 $18 = ($17|0)==(0|0);
 if ($18) {
  ___PHYSFS_setError(517404);
  $0 = 0;
  $51 = $0;
  STACKTOP = sp;return ($51|0);
 }
 $19 = $entry;
 $20 = ((($19)) + 8|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ($21|0)==(1);
 if ($22) {
  $23 = $info;
  $24 = HEAP32[$23>>2]|0;
  $25 = (___PHYSFS_platformOpenRead($24)|0);
  $in = $25;
  $26 = $in;
  $27 = ($26|0)==(0|0);
  if ($27) {
   ___PHYSFS_setError(0);
   $0 = 0;
   $51 = $0;
   STACKTOP = sp;return ($51|0);
  }
  $28 = $in;
  $29 = $info;
  $30 = $entry;
  $31 = (_zip_resolve($28,$29,$30)|0);
  $rc = $31;
  $32 = $in;
  (___PHYSFS_platformClose($32)|0);
  $33 = $rc;
  $34 = ($33|0)!=(0);
  if (!($34)) {
   $0 = 0;
   $51 = $0;
   STACKTOP = sp;return ($51|0);
  }
 }
 $35 = $entry;
 $36 = ((($35)) + 8|0);
 $37 = HEAP32[$36>>2]|0;
 $38 = ($37|0)==(5);
 if ($38) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $51 = $0;
  STACKTOP = sp;return ($51|0);
 }
 $39 = $entry;
 $40 = ((($39)) + 4|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = ($41|0)==(0|0);
 if ($42) {
  ___PHYSFS_setError(517488);
  $0 = 0;
  $51 = $0;
  STACKTOP = sp;return ($51|0);
 } else {
  $43 = $info;
  $44 = $entry;
  $45 = ((($44)) + 4|0);
  $46 = HEAP32[$45>>2]|0;
  $47 = HEAP32[$46>>2]|0;
  $48 = (_zip_find_start_of_dir($43,$47,1)|0);
  $49 = ($48|0)>=(0);
  $50 = $49&1;
  $0 = $50;
  $51 = $0;
  STACKTOP = sp;return ($51|0);
 }
 return (0)|0;
}
function _zip_resolve($in,$info,$entry) {
 $in = $in|0;
 $info = $info|0;
 $entry = $entry|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $resolve_type = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $in;
 $2 = $info;
 $3 = $entry;
 $retval = 1;
 $4 = $3;
 $5 = ((($4)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $resolve_type = $6;
 $7 = $resolve_type;
 $8 = ($7|0)==(4);
 if ($8) {
  ___PHYSFS_setError(517386);
  $0 = 0;
  $43 = $0;
  STACKTOP = sp;return ($43|0);
 }
 $9 = $resolve_type;
 $10 = ($9|0)==(5);
 if ($10) {
  ___PHYSFS_setError(517386);
  $0 = 0;
  $43 = $0;
  STACKTOP = sp;return ($43|0);
 }
 $11 = $resolve_type;
 $12 = ($11|0)==(2);
 if ($12) {
  ___PHYSFS_setError(517419);
  $0 = 0;
  $43 = $0;
  STACKTOP = sp;return ($43|0);
 }
 $13 = $resolve_type;
 $14 = ($13|0)!=(3);
 do {
  if ($14) {
   $15 = $3;
   $16 = ((($15)) + 8|0);
   HEAP32[$16>>2] = 2;
   $17 = $1;
   $18 = $3;
   $19 = (_zip_parse_local($17,$18)|0);
   $retval = $19;
   $20 = $retval;
   $21 = ($20|0)!=(0);
   $22 = $resolve_type;
   $23 = ($22|0)==(1);
   $or$cond = $21 & $23;
   if ($or$cond) {
    $24 = $1;
    $25 = $2;
    $26 = $3;
    $27 = (_zip_resolve_symlink($24,$25,$26)|0);
    $retval = $27;
   }
   $28 = $resolve_type;
   $29 = ($28|0)==(1);
   if ($29) {
    $30 = $retval;
    $31 = ($30|0)!=(0);
    $32 = $31 ? 3 : 5;
    $33 = $3;
    $34 = ((($33)) + 8|0);
    HEAP32[$34>>2] = $32;
    break;
   }
   $35 = $resolve_type;
   $36 = ($35|0)==(0);
   if ($36) {
    $37 = $retval;
    $38 = ($37|0)!=(0);
    $39 = $38 ? 3 : 4;
    $40 = $3;
    $41 = ((($40)) + 8|0);
    HEAP32[$41>>2] = $39;
   }
  }
 } while(0);
 $42 = $retval;
 $0 = $42;
 $43 = $0;
 STACKTOP = sp;return ($43|0);
}
function _zip_parse_local($in,$entry) {
 $in = $in|0;
 $entry = $entry|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $extralen = 0, $fnamelen = 0, $ui16 = 0, $ui32 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ui32 = sp;
 $ui16 = sp + 20|0;
 $fnamelen = sp + 18|0;
 $extralen = sp + 16|0;
 $1 = $in;
 $2 = $entry;
 $3 = $1;
 $4 = $2;
 $5 = ((($4)) + 12|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (___PHYSFS_platformSeek($3,$6,0)|0);
 $8 = ($7|0)!=(0);
 do {
  if ($8) {
   $9 = $1;
   $10 = (_readui32($9,$ui32)|0);
   $11 = ($10|0)!=(0);
   if (!($11)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $12 = HEAP32[$ui32>>2]|0;
   $13 = ($12|0)!=(67324752);
   if ($13) {
    ___PHYSFS_setError(517386);
    $0 = 0;
    break;
   }
   $14 = $1;
   $15 = (_readui16($14,$ui16)|0);
   $16 = ($15|0)!=(0);
   if (!($16)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $17 = HEAP16[$ui16>>1]|0;
   $18 = $17&65535;
   $19 = $2;
   $20 = ((($19)) + 18|0);
   $21 = HEAP16[$20>>1]|0;
   $22 = $21&65535;
   $23 = ($18|0)!=($22|0);
   if ($23) {
    ___PHYSFS_setError(517386);
    $0 = 0;
    break;
   }
   $24 = $1;
   $25 = (_readui16($24,$ui16)|0);
   $26 = ($25|0)!=(0);
   if (!($26)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $27 = $1;
   $28 = (_readui16($27,$ui16)|0);
   $29 = ($28|0)!=(0);
   if (!($29)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $30 = HEAP16[$ui16>>1]|0;
   $31 = $30&65535;
   $32 = $2;
   $33 = ((($32)) + 20|0);
   $34 = HEAP16[$33>>1]|0;
   $35 = $34&65535;
   $36 = ($31|0)!=($35|0);
   if ($36) {
    ___PHYSFS_setError(517386);
    $0 = 0;
    break;
   }
   $37 = $1;
   $38 = (_readui32($37,$ui32)|0);
   $39 = ($38|0)!=(0);
   if (!($39)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $40 = $1;
   $41 = (_readui32($40,$ui32)|0);
   $42 = ($41|0)!=(0);
   if (!($42)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $43 = HEAP32[$ui32>>2]|0;
   $44 = ($43|0)!=(0);
   if ($44) {
    $45 = HEAP32[$ui32>>2]|0;
    $46 = $2;
    $47 = ((($46)) + 24|0);
    $48 = HEAP32[$47>>2]|0;
    $49 = ($45|0)!=($48|0);
    if ($49) {
     ___PHYSFS_setError(517386);
     $0 = 0;
     break;
    }
   }
   $50 = $1;
   $51 = (_readui32($50,$ui32)|0);
   $52 = ($51|0)!=(0);
   if (!($52)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $53 = HEAP32[$ui32>>2]|0;
   $54 = ($53|0)!=(0);
   if ($54) {
    $55 = HEAP32[$ui32>>2]|0;
    $56 = $2;
    $57 = ((($56)) + 28|0);
    $58 = HEAP32[$57>>2]|0;
    $59 = ($55|0)!=($58|0);
    if ($59) {
     ___PHYSFS_setError(517386);
     $0 = 0;
     break;
    }
   }
   $60 = $1;
   $61 = (_readui32($60,$ui32)|0);
   $62 = ($61|0)!=(0);
   if (!($62)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $63 = HEAP32[$ui32>>2]|0;
   $64 = ($63|0)!=(0);
   if ($64) {
    $65 = HEAP32[$ui32>>2]|0;
    $66 = $2;
    $67 = ((($66)) + 32|0);
    $68 = HEAP32[$67>>2]|0;
    $69 = ($65|0)!=($68|0);
    if ($69) {
     ___PHYSFS_setError(517386);
     $0 = 0;
     break;
    }
   }
   $70 = $1;
   $71 = (_readui16($70,$fnamelen)|0);
   $72 = ($71|0)!=(0);
   if (!($72)) {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
   $73 = $1;
   $74 = (_readui16($73,$extralen)|0);
   $75 = ($74|0)!=(0);
   if ($75) {
    $76 = HEAP16[$fnamelen>>1]|0;
    $77 = $76&65535;
    $78 = HEAP16[$extralen>>1]|0;
    $79 = $78&65535;
    $80 = (($77) + ($79))|0;
    $81 = (($80) + 30)|0;
    $82 = $2;
    $83 = ((($82)) + 12|0);
    $84 = HEAP32[$83>>2]|0;
    $85 = (($84) + ($81))|0;
    HEAP32[$83>>2] = $85;
    $0 = 1;
    break;
   } else {
    ___PHYSFS_setError(0);
    $0 = 0;
    break;
   }
  } else {
   ___PHYSFS_setError(0);
   $0 = 0;
  }
 } while(0);
 $86 = $0;
 STACKTOP = sp;return ($86|0);
}
function _zip_resolve_symlink($in,$info,$entry) {
 $in = $in|0;
 $info = $info|0;
 $entry = $entry|0;
 var $$alloca_mul = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $complen = 0, $compressed = 0, $path = 0;
 var $rc = 0, $size = 0, $stream = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $stream = sp + 8|0;
 $1 = $in;
 $2 = $info;
 $3 = $entry;
 $4 = $3;
 $5 = ((($4)) + 32|0);
 $6 = HEAP32[$5>>2]|0;
 $size = $6;
 $rc = 0;
 $7 = $1;
 $8 = $3;
 $9 = ((($8)) + 12|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (___PHYSFS_platformSeek($7,$10,0)|0);
 $12 = ($11|0)!=(0);
 if (!($12)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $95 = $0;
  STACKTOP = sp;return ($95|0);
 }
 $13 = HEAP32[(732252)>>2]|0;
 $14 = $size;
 $15 = (($14) + 1)|0;
 $16 = (FUNCTION_TABLE_iii[$13 & 8191]($15,0)|0);
 $path = $16;
 $17 = $path;
 $18 = ($17|0)==(0|0);
 if ($18) {
  ___PHYSFS_setError(596704);
  $0 = 0;
  $95 = $0;
  STACKTOP = sp;return ($95|0);
 }
 $19 = $3;
 $20 = ((($19)) + 20|0);
 $21 = HEAP16[$20>>1]|0;
 $22 = $21&65535;
 $23 = ($22|0)==(0);
 if ($23) {
  $24 = $1;
  $25 = $path;
  $26 = $size;
  $27 = (___PHYSFS_platformRead($24,$25,$26,1)|0);
  $28 = tempRet0;
  $29 = ($27|0)==(1);
  $30 = ($28|0)==(0);
  $31 = $29 & $30;
  $32 = $31&1;
  $rc = $32;
 } else {
  $33 = $3;
  $34 = ((($33)) + 28|0);
  $35 = HEAP32[$34>>2]|0;
  $complen = $35;
  $36 = $complen;
  $37 = ($36>>>0)<(128);
  if ($37) {
   $38 = $complen;
   $39 = (($38) + 1)|0;
   $$alloca_mul = $39;
   $40 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
   $42 = $40;
  } else {
   $42 = 0;
  }
  $41 = $complen;
  $43 = (___PHYSFS_initSmallAlloc($42,$41,0)|0);
  $compressed = $43;
  $44 = $compressed;
  $45 = ($44|0)!=(0|0);
  if ($45) {
   $46 = $1;
   $47 = $compressed;
   $48 = $complen;
   $49 = (___PHYSFS_platformRead($46,$47,$48,1)|0);
   $50 = tempRet0;
   $51 = ($49|0)==(1);
   $52 = ($50|0)==(0);
   $53 = $51 & $52;
   if ($53) {
    _initializeZStream($stream);
    $54 = $compressed;
    HEAP32[$stream>>2] = $54;
    $55 = $complen;
    $56 = ((($stream)) + 4|0);
    HEAP32[$56>>2] = $55;
    $57 = $path;
    $58 = ((($stream)) + 12|0);
    HEAP32[$58>>2] = $57;
    $59 = $size;
    $60 = ((($stream)) + 16|0);
    HEAP32[$60>>2] = $59;
    $61 = (_inflateInit2_($stream,-15,613577,56)|0);
    $62 = (_zlib_err($61)|0);
    $63 = ($62|0)==(0);
    if ($63) {
     $64 = (_inflate($stream,4)|0);
     $65 = (_zlib_err($64)|0);
     $rc = $65;
     (_inflateEnd($stream)|0);
     $66 = $rc;
     $67 = ($66|0)==(0);
     $68 = $rc;
     $69 = ($68|0)==(1);
     $70 = $67 ? 1 : $69;
     $71 = $70&1;
     $rc = $71;
    }
   }
   $72 = $compressed;
   ___PHYSFS_smallFree($72);
  }
 }
 $73 = $rc;
 $74 = ($73|0)!=(0);
 if ($74) {
  $77 = $3;
  $78 = ((($77)) + 32|0);
  $79 = HEAP32[$78>>2]|0;
  $80 = $path;
  $81 = (($80) + ($79)|0);
  HEAP8[$81>>0] = 0;
  $82 = $3;
  $83 = $path;
  _zip_convert_dos_path($82,$83);
  $84 = $1;
  $85 = $2;
  $86 = $path;
  $87 = (_zip_follow_symlink($84,$85,$86)|0);
  $88 = $3;
  $89 = ((($88)) + 4|0);
  HEAP32[$89>>2] = $87;
 } else {
  $75 = HEAP32[(732260)>>2]|0;
  $76 = $path;
  FUNCTION_TABLE_vi[$75 & 8191]($76);
 }
 $90 = $3;
 $91 = ((($90)) + 4|0);
 $92 = HEAP32[$91>>2]|0;
 $93 = ($92|0)!=(0|0);
 $94 = $93&1;
 $0 = $94;
 $95 = $0;
 STACKTOP = sp;return ($95|0);
}
function _initializeZStream($pstr) {
 $pstr = $pstr|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $pstr;
 $1 = $0;
 dest=$1; stop=dest+56|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $2 = $0;
 $3 = ((($2)) + 32|0);
 HEAP32[$3>>2] = 4287;
 $4 = $0;
 $5 = ((($4)) + 36|0);
 HEAP32[$5>>2] = 4288;
 $6 = $0;
 $7 = ((($6)) + 40|0);
 HEAP32[$7>>2] = 732244;
 STACKTOP = sp;return;
}
function _zlibPhysfsAlloc($opaque,$items,$size) {
 $opaque = $opaque|0;
 $items = $items|0;
 $size = $size|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $items;
 $2 = $size;
 $3 = $0;
 $4 = ((($3)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $1;
 $7 = $2;
 $8 = Math_imul($6, $7)|0;
 $9 = (FUNCTION_TABLE_iii[$5 & 8191]($8,0)|0);
 STACKTOP = sp;return ($9|0);
}
function _zlibPhysfsFree($opaque,$address) {
 $opaque = $opaque|0;
 $address = $address|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $address;
 $2 = $0;
 $3 = ((($2)) + 16|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $1;
 FUNCTION_TABLE_vi[$4 & 8191]($5);
 STACKTOP = sp;return;
}
function _zlib_err($rc) {
 $rc = $rc|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $str = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $rc;
 $1 = $0;
 $2 = (_zlib_error_string($1)|0);
 $str = $2;
 $3 = $str;
 $4 = ($3|0)!=(0|0);
 if ($4) {
  $5 = $str;
  ___PHYSFS_setError($5);
 }
 $6 = $0;
 STACKTOP = sp;return ($6|0);
}
function _zlib_error_string($rc) {
 $rc = $rc|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $rc;
 $2 = $1;
 switch ($2|0) {
 case 0:  {
  $0 = 0;
  break;
 }
 case 1:  {
  $0 = 0;
  break;
 }
 case -1:  {
  $3 = (___errno_location()|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = (_strerror($4)|0);
  $0 = $5;
  break;
 }
 case 2:  {
  $0 = 613583;
  break;
 }
 case -3:  {
  $0 = 613634;
  break;
 }
 case -4:  {
  $0 = 517447;
  break;
 }
 case -5:  {
  $0 = 613665;
  break;
 }
 case -6:  {
  $0 = 517460;
  break;
 }
 default: {
  $0 = 517474;
 }
 }
 $6 = $0;
 STACKTOP = sp;return ($6|0);
}
function _zip_follow_symlink($in,$info,$path) {
 $in = $in|0;
 $info = $info|0;
 $path = $path|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $entry = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $in;
 $1 = $info;
 $2 = $path;
 $3 = $2;
 _zip_expand_symlink_path($3);
 $4 = $1;
 $5 = $2;
 $6 = (_zip_find_entry($4,$5,0)|0);
 $entry = $6;
 $7 = $entry;
 $8 = ($7|0)!=(0|0);
 do {
  if ($8) {
   $9 = $0;
   $10 = $1;
   $11 = $entry;
   $12 = (_zip_resolve($9,$10,$11)|0);
   $13 = ($12|0)!=(0);
   if (!($13)) {
    $entry = 0;
    break;
   }
   $14 = $entry;
   $15 = ((($14)) + 4|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = ($16|0)!=(0|0);
   if ($17) {
    $18 = $entry;
    $19 = ((($18)) + 4|0);
    $20 = HEAP32[$19>>2]|0;
    $entry = $20;
   }
  }
 } while(0);
 $21 = HEAP32[(732260)>>2]|0;
 $22 = $2;
 FUNCTION_TABLE_vi[$21 & 8191]($22);
 $23 = $entry;
 STACKTOP = sp;return ($23|0);
}
function _zip_expand_symlink_path($path) {
 $path = $path|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $prevptr = 0, $ptr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $path;
 $1 = $0;
 $ptr = $1;
 $2 = $0;
 $prevptr = $2;
 while(1) {
  $3 = $ptr;
  $4 = (_strchr($3,47)|0);
  $ptr = $4;
  $5 = $ptr;
  $6 = ($5|0)==(0|0);
  if ($6) {
   break;
  }
  $7 = $ptr;
  $8 = ((($7)) + 1|0);
  $9 = HEAP8[$8>>0]|0;
  $10 = $9 << 24 >> 24;
  $11 = ($10|0)==(46);
  $12 = $ptr;
  if (!($11)) {
   $prevptr = $12;
   $63 = $ptr;
   $64 = ((($63)) + 1|0);
   $ptr = $64;
   continue;
  }
  $13 = ((($12)) + 2|0);
  $14 = HEAP8[$13>>0]|0;
  $15 = $14 << 24 >> 24;
  $16 = ($15|0)==(47);
  $17 = $ptr;
  if ($16) {
   $18 = $ptr;
   $19 = ((($18)) + 2|0);
   $20 = $ptr;
   $21 = ((($20)) + 2|0);
   $22 = (_strlen($21)|0);
   $23 = (($22) + 1)|0;
   _memmove(($17|0),($19|0),($23|0))|0;
   continue;
  }
  $24 = ((($17)) + 2|0);
  $25 = HEAP8[$24>>0]|0;
  $26 = $25 << 24 >> 24;
  $27 = ($26|0)==(0);
  $28 = $ptr;
  if ($27) {
   HEAP8[$28>>0] = 0;
   continue;
  }
  $29 = ((($28)) + 2|0);
  $30 = HEAP8[$29>>0]|0;
  $31 = $30 << 24 >> 24;
  $32 = ($31|0)==(46);
  if (!($32)) {
   continue;
  }
  $33 = $ptr;
  $34 = ((($33)) + 3|0);
  $35 = HEAP8[$34>>0]|0;
  $36 = $35 << 24 >> 24;
  $37 = ($36|0)==(47);
  L14: do {
   if ($37) {
    $38 = $prevptr;
    $39 = $ptr;
    $40 = ((($39)) + 4|0);
    $41 = $ptr;
    $42 = ((($41)) + 4|0);
    $43 = (_strlen($42)|0);
    $44 = (($43) + 1)|0;
    _memmove(($38|0),($40|0),($44|0))|0;
    $45 = $prevptr;
    $ptr = $45;
    while(1) {
     $46 = $prevptr;
     $47 = $0;
     $48 = ($46|0)!=($47|0);
     if (!($48)) {
      break L14;
     }
     $49 = $prevptr;
     $50 = ((($49)) + -1|0);
     $prevptr = $50;
     $51 = $prevptr;
     $52 = HEAP8[$51>>0]|0;
     $53 = $52 << 24 >> 24;
     $54 = ($53|0)==(47);
     if ($54) {
      break;
     }
    }
    $55 = $prevptr;
    $56 = ((($55)) + 1|0);
    $prevptr = $56;
   }
  } while(0);
  $57 = $ptr;
  $58 = ((($57)) + 3|0);
  $59 = HEAP8[$58>>0]|0;
  $60 = $59 << 24 >> 24;
  $61 = ($60|0)==(0);
  if (!($61)) {
   continue;
  }
  $62 = $prevptr;
  HEAP8[$62>>0] = 0;
 }
 STACKTOP = sp;return;
}
function _ZIP_isSymLink($opaque,$name,$fileExists) {
 $opaque = $opaque|0;
 $name = $name|0;
 $fileExists = $fileExists|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $entry = 0;
 var $isDir = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $isDir = sp + 4|0;
 $1 = $opaque;
 $2 = $name;
 $3 = $fileExists;
 $4 = $1;
 $5 = $2;
 $6 = (_zip_find_entry($4,$5,$isDir)|0);
 $entry = $6;
 $7 = HEAP32[$isDir>>2]|0;
 $8 = ($7|0)!=(0);
 $9 = $entry;
 $10 = ($9|0)!=(0|0);
 $11 = $8 ? 1 : $10;
 $12 = $11&1;
 $13 = $3;
 HEAP32[$13>>2] = $12;
 $14 = $entry;
 $15 = ($14|0)==(0|0);
 if ($15) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $18 = $0;
  STACKTOP = sp;return ($18|0);
 } else {
  $16 = $entry;
  $17 = (_zip_entry_is_symlink($16)|0);
  $0 = $17;
  $18 = $0;
  STACKTOP = sp;return ($18|0);
 }
 return (0)|0;
}
function _ZIP_getLastModTime($opaque,$name,$fileExists) {
 $opaque = $opaque|0;
 $name = $name|0;
 $fileExists = $fileExists|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $entry = 0, $info = 0, $isDir = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $isDir = sp + 16|0;
 $1 = $opaque;
 $2 = $name;
 $3 = $fileExists;
 $4 = $1;
 $info = $4;
 $5 = $info;
 $6 = $2;
 $7 = (_zip_find_entry($5,$6,$isDir)|0);
 $entry = $7;
 $8 = HEAP32[$isDir>>2]|0;
 $9 = ($8|0)!=(0);
 $10 = $entry;
 $11 = ($10|0)!=(0|0);
 $12 = $9 ? 1 : $11;
 $13 = $12&1;
 $14 = $3;
 HEAP32[$14>>2] = $13;
 $15 = HEAP32[$isDir>>2]|0;
 $16 = ($15|0)!=(0);
 do {
  if ($16) {
   $17 = $0;
   $18 = $17;
   HEAP32[$18>>2] = 1;
   $19 = (($17) + 4)|0;
   $20 = $19;
   HEAP32[$20>>2] = 0;
  } else {
   $21 = $entry;
   $22 = ($21|0)==(0|0);
   if ($22) {
    ___PHYSFS_setError(0);
    $23 = $0;
    $24 = $23;
    HEAP32[$24>>2] = -1;
    $25 = (($23) + 4)|0;
    $26 = $25;
    HEAP32[$26>>2] = -1;
    break;
   } else {
    $27 = $entry;
    $28 = ((($27)) + 40|0);
    $29 = $28;
    $30 = $29;
    $31 = HEAP32[$30>>2]|0;
    $32 = (($29) + 4)|0;
    $33 = $32;
    $34 = HEAP32[$33>>2]|0;
    $35 = $0;
    $36 = $35;
    HEAP32[$36>>2] = $31;
    $37 = (($35) + 4)|0;
    $38 = $37;
    HEAP32[$38>>2] = $34;
    break;
   }
  }
 } while(0);
 $39 = $0;
 $40 = $39;
 $41 = HEAP32[$40>>2]|0;
 $42 = (($39) + 4)|0;
 $43 = $42;
 $44 = HEAP32[$43>>2]|0;
 tempRet0 = ($44);
 STACKTOP = sp;return ($41|0);
}
function _ZIP_openRead($opaque,$fnm,$fileExists) {
 $opaque = $opaque|0;
 $fnm = $fnm|0;
 $fileExists = $fileExists|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $7 = 0, $8 = 0, $9 = 0, $entry = 0, $finfo = 0, $in = 0, $info = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $opaque;
 $2 = $fnm;
 $3 = $fileExists;
 $4 = $1;
 $info = $4;
 $5 = $info;
 $6 = $2;
 $7 = (_zip_find_entry($5,$6,0)|0);
 $entry = $7;
 $finfo = 0;
 $8 = $entry;
 $9 = ($8|0)!=(0|0);
 $10 = $9&1;
 $11 = $3;
 HEAP32[$11>>2] = $10;
 $12 = $entry;
 $13 = ($12|0)==(0|0);
 if ($13) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $63 = $0;
  STACKTOP = sp;return ($63|0);
 }
 $14 = $info;
 $15 = HEAP32[$14>>2]|0;
 $16 = $info;
 $17 = $entry;
 $18 = (_zip_get_file_handle($15,$16,$17)|0);
 $in = $18;
 $19 = $in;
 $20 = ($19|0)==(0|0);
 if ($20) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $63 = $0;
  STACKTOP = sp;return ($63|0);
 }
 $21 = HEAP32[(732252)>>2]|0;
 $22 = (FUNCTION_TABLE_iii[$21 & 8191](76,0)|0);
 $finfo = $22;
 $23 = $finfo;
 $24 = ($23|0)==(0|0);
 if ($24) {
  $25 = $in;
  (___PHYSFS_platformClose($25)|0);
  ___PHYSFS_setError(596704);
  $0 = 0;
  $63 = $0;
  STACKTOP = sp;return ($63|0);
 }
 $26 = $finfo;
 dest=$26; stop=dest+76|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $27 = $in;
 $28 = $finfo;
 $29 = ((($28)) + 4|0);
 HEAP32[$29>>2] = $27;
 $30 = $entry;
 $31 = ((($30)) + 4|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ($32|0)!=(0|0);
 $34 = $entry;
 if ($33) {
  $35 = ((($34)) + 4|0);
  $36 = HEAP32[$35>>2]|0;
  $38 = $36;
 } else {
  $38 = $34;
 }
 $37 = $finfo;
 HEAP32[$37>>2] = $38;
 $39 = $finfo;
 $40 = ((($39)) + 20|0);
 _initializeZStream($40);
 $41 = $finfo;
 $42 = HEAP32[$41>>2]|0;
 $43 = ((($42)) + 20|0);
 $44 = HEAP16[$43>>1]|0;
 $45 = $44&65535;
 $46 = ($45|0)!=(0);
 if ($46) {
  $47 = $finfo;
  $48 = ((($47)) + 20|0);
  $49 = (_inflateInit2_($48,-15,613577,56)|0);
  $50 = (_zlib_err($49)|0);
  $51 = ($50|0)!=(0);
  if ($51) {
   $52 = $finfo;
   (_ZIP_fileClose($52)|0);
   $0 = 0;
   $63 = $0;
   STACKTOP = sp;return ($63|0);
  }
  $53 = HEAP32[(732252)>>2]|0;
  $54 = (FUNCTION_TABLE_iii[$53 & 8191](16384,0)|0);
  $55 = $finfo;
  $56 = ((($55)) + 16|0);
  HEAP32[$56>>2] = $54;
  $57 = $finfo;
  $58 = ((($57)) + 16|0);
  $59 = HEAP32[$58>>2]|0;
  $60 = ($59|0)==(0|0);
  if ($60) {
   $61 = $finfo;
   (_ZIP_fileClose($61)|0);
   ___PHYSFS_setError(596704);
   $0 = 0;
   $63 = $0;
   STACKTOP = sp;return ($63|0);
  }
 }
 $62 = $finfo;
 $0 = $62;
 $63 = $0;
 STACKTOP = sp;return ($63|0);
}
function _zip_get_file_handle($fn,$inf,$entry) {
 $fn = $fn|0;
 $inf = $inf|0;
 $entry = $entry|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $offset = 0, $retval = 0, $success = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $offset = sp;
 $1 = $fn;
 $2 = $inf;
 $3 = $entry;
 $4 = $1;
 $5 = (___PHYSFS_platformOpenRead($4)|0);
 $retval = $5;
 $6 = $retval;
 $7 = ($6|0)==(0|0);
 if ($7) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $42 = $0;
  STACKTOP = sp;return ($42|0);
 }
 $8 = $retval;
 $9 = $2;
 $10 = $3;
 $11 = (_zip_resolve($8,$9,$10)|0);
 $success = $11;
 $12 = $success;
 $13 = ($12|0)!=(0);
 if ($13) {
  $14 = $3;
  $15 = ((($14)) + 4|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ($16|0)!=(0|0);
  $18 = $3;
  if ($17) {
   $19 = ((($18)) + 4|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = ((($20)) + 12|0);
   $22 = HEAP32[$21>>2]|0;
   $27 = $22;
  } else {
   $23 = ((($18)) + 12|0);
   $24 = HEAP32[$23>>2]|0;
   $27 = $24;
  }
  $25 = $offset;
  $26 = $25;
  HEAP32[$26>>2] = $27;
  $28 = (($25) + 4)|0;
  $29 = $28;
  HEAP32[$29>>2] = 0;
  $30 = $retval;
  $31 = $offset;
  $32 = $31;
  $33 = HEAP32[$32>>2]|0;
  $34 = (($31) + 4)|0;
  $35 = $34;
  $36 = HEAP32[$35>>2]|0;
  $37 = (___PHYSFS_platformSeek($30,$33,$36)|0);
  $success = $37;
 }
 $38 = $success;
 $39 = ($38|0)!=(0);
 if (!($39)) {
  $40 = $retval;
  (___PHYSFS_platformClose($40)|0);
  $retval = 0;
 }
 $41 = $retval;
 $0 = $41;
 $42 = $0;
 STACKTOP = sp;return ($42|0);
}
function _ZIP_fileClose($opaque) {
 $opaque = $opaque|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $finfo = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $opaque;
 $2 = $1;
 $finfo = $2;
 $3 = $finfo;
 $4 = ((($3)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (___PHYSFS_platformClose($5)|0);
 $7 = ($6|0)!=(0);
 if (!($7)) {
  ___PHYSFS_setError(0);
  $0 = 0;
  $26 = $0;
  STACKTOP = sp;return ($26|0);
 }
 $8 = $finfo;
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 20|0);
 $11 = HEAP16[$10>>1]|0;
 $12 = $11&65535;
 $13 = ($12|0)!=(0);
 if ($13) {
  $14 = $finfo;
  $15 = ((($14)) + 20|0);
  (_inflateEnd($15)|0);
 }
 $16 = $finfo;
 $17 = ((($16)) + 16|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ($18|0)!=(0|0);
 if ($19) {
  $20 = HEAP32[(732260)>>2]|0;
  $21 = $finfo;
  $22 = ((($21)) + 16|0);
  $23 = HEAP32[$22>>2]|0;
  FUNCTION_TABLE_vi[$20 & 8191]($23);
 }
 $24 = HEAP32[(732260)>>2]|0;
 $25 = $finfo;
 FUNCTION_TABLE_vi[$24 & 8191]($25);
 $0 = 1;
 $26 = $0;
 STACKTOP = sp;return ($26|0);
}
function _ZIP_openWrite($opaque,$filename) {
 $opaque = $opaque|0;
 $filename = $filename|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $filename;
 ___PHYSFS_setError(517504);
 STACKTOP = sp;return (0|0);
}
function _ZIP_openAppend($opaque,$filename) {
 $opaque = $opaque|0;
 $filename = $filename|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $filename;
 ___PHYSFS_setError(517504);
 STACKTOP = sp;return (0|0);
}
function _ZIP_remove($opaque,$name) {
 $opaque = $opaque|0;
 $name = $name|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $name;
 ___PHYSFS_setError(517504);
 STACKTOP = sp;return 0;
}
function _ZIP_mkdir($opaque,$name) {
 $opaque = $opaque|0;
 $name = $name|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $name;
 ___PHYSFS_setError(517504);
 STACKTOP = sp;return 0;
}
function _ZIP_dirClose($opaque) {
 $opaque = $opaque|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $zi = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $0;
 $zi = $1;
 $2 = $zi;
 $3 = ((($2)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $zi;
 $6 = ((($5)) + 4|0);
 $7 = HEAP16[$6>>1]|0;
 $8 = $7&65535;
 _zip_free_entries($4,$8);
 $9 = HEAP32[(732260)>>2]|0;
 $10 = $zi;
 $11 = HEAP32[$10>>2]|0;
 FUNCTION_TABLE_vi[$9 & 8191]($11);
 $12 = HEAP32[(732260)>>2]|0;
 $13 = $zi;
 FUNCTION_TABLE_vi[$12 & 8191]($13);
 STACKTOP = sp;return;
}
function _ZIP_read($opaque,$buf,$objSize,$objCount) {
 $opaque = $opaque|0;
 $buf = $buf|0;
 $objSize = $objSize|0;
 $objCount = $objCount|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $avail = 0, $before = 0, $br = 0;
 var $entry = 0, $finfo = 0, $maxread = 0, $rc = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 32|0;
 $retval = sp + 24|0;
 $maxread = sp + 16|0;
 $avail = sp + 8|0;
 $br = sp;
 $1 = $opaque;
 $2 = $buf;
 $3 = $objSize;
 $4 = $objCount;
 $5 = $1;
 $finfo = $5;
 $6 = $finfo;
 $7 = HEAP32[$6>>2]|0;
 $entry = $7;
 $8 = $retval;
 $9 = $8;
 HEAP32[$9>>2] = 0;
 $10 = (($8) + 4)|0;
 $11 = $10;
 HEAP32[$11>>2] = 0;
 $12 = $3;
 $13 = $4;
 $14 = (___muldi3(($12|0),0,($13|0),0)|0);
 $15 = tempRet0;
 $16 = $maxread;
 $17 = $16;
 HEAP32[$17>>2] = $14;
 $18 = (($16) + 4)|0;
 $19 = $18;
 HEAP32[$19>>2] = $15;
 $20 = $entry;
 $21 = ((($20)) + 32|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = $finfo;
 $24 = ((($23)) + 12|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = (($22) - ($25))|0;
 $27 = $avail;
 $28 = $27;
 HEAP32[$28>>2] = $26;
 $29 = (($27) + 4)|0;
 $30 = $29;
 HEAP32[$30>>2] = 0;
 $31 = $maxread;
 $32 = $31;
 $33 = HEAP32[$32>>2]|0;
 $34 = (($31) + 4)|0;
 $35 = $34;
 $36 = HEAP32[$35>>2]|0;
 $37 = ($33|0)==(0);
 $38 = ($36|0)==(0);
 $39 = $37 & $38;
 if ($39) {
  ___PHYSFS_setError(0);
  $40 = $0;
  $41 = $40;
  HEAP32[$41>>2] = 0;
  $42 = (($40) + 4)|0;
  $43 = $42;
  HEAP32[$43>>2] = 0;
  $311 = $0;
  $312 = $311;
  $313 = HEAP32[$312>>2]|0;
  $314 = (($311) + 4)|0;
  $315 = $314;
  $316 = HEAP32[$315>>2]|0;
  tempRet0 = ($316);
  STACKTOP = sp;return ($313|0);
 }
 $44 = $avail;
 $45 = $44;
 $46 = HEAP32[$45>>2]|0;
 $47 = (($44) + 4)|0;
 $48 = $47;
 $49 = HEAP32[$48>>2]|0;
 $50 = $maxread;
 $51 = $50;
 $52 = HEAP32[$51>>2]|0;
 $53 = (($50) + 4)|0;
 $54 = $53;
 $55 = HEAP32[$54>>2]|0;
 $56 = ($49|0)<($55|0);
 $57 = ($46>>>0)<($52>>>0);
 $58 = ($49|0)==($55|0);
 $59 = $58 & $57;
 $60 = $56 | $59;
 if ($60) {
  $61 = $avail;
  $62 = $61;
  $63 = HEAP32[$62>>2]|0;
  $64 = (($61) + 4)|0;
  $65 = $64;
  $66 = HEAP32[$65>>2]|0;
  $67 = $avail;
  $68 = $67;
  $69 = HEAP32[$68>>2]|0;
  $70 = (($67) + 4)|0;
  $71 = $70;
  $72 = HEAP32[$71>>2]|0;
  $73 = $3;
  $74 = (___remdi3(($69|0),($72|0),($73|0),0)|0);
  $75 = tempRet0;
  $76 = (_i64Subtract(($63|0),($66|0),($74|0),($75|0))|0);
  $77 = tempRet0;
  $78 = $maxread;
  $79 = $78;
  HEAP32[$79>>2] = $76;
  $80 = (($78) + 4)|0;
  $81 = $80;
  HEAP32[$81>>2] = $77;
  $82 = $maxread;
  $83 = $82;
  $84 = HEAP32[$83>>2]|0;
  $85 = (($82) + 4)|0;
  $86 = $85;
  $87 = HEAP32[$86>>2]|0;
  $88 = $3;
  $89 = (___divdi3(($84|0),($87|0),($88|0),0)|0);
  $90 = tempRet0;
  $4 = $89;
  $91 = $4;
  $92 = ($91|0)==(0);
  ___PHYSFS_setError(517528);
  if ($92) {
   $93 = $0;
   $94 = $93;
   HEAP32[$94>>2] = 0;
   $95 = (($93) + 4)|0;
   $96 = $95;
   HEAP32[$96>>2] = 0;
   $311 = $0;
   $312 = $311;
   $313 = HEAP32[$312>>2]|0;
   $314 = (($311) + 4)|0;
   $315 = $314;
   $316 = HEAP32[$315>>2]|0;
   tempRet0 = ($316);
   STACKTOP = sp;return ($313|0);
  }
 }
 $97 = $entry;
 $98 = ((($97)) + 20|0);
 $99 = HEAP16[$98>>1]|0;
 $100 = $99&65535;
 $101 = ($100|0)==(0);
 if ($101) {
  $102 = $finfo;
  $103 = ((($102)) + 4|0);
  $104 = HEAP32[$103>>2]|0;
  $105 = $2;
  $106 = $3;
  $107 = $4;
  $108 = (___PHYSFS_platformRead($104,$105,$106,$107)|0);
  $109 = tempRet0;
  $110 = $retval;
  $111 = $110;
  HEAP32[$111>>2] = $108;
  $112 = (($110) + 4)|0;
  $113 = $112;
  HEAP32[$113>>2] = $109;
 } else {
  $114 = $2;
  $115 = $finfo;
  $116 = ((($115)) + 20|0);
  $117 = ((($116)) + 12|0);
  HEAP32[$117>>2] = $114;
  $118 = $3;
  $119 = $4;
  $120 = Math_imul($118, $119)|0;
  $121 = $finfo;
  $122 = ((($121)) + 20|0);
  $123 = ((($122)) + 16|0);
  HEAP32[$123>>2] = $120;
  while(1) {
   $124 = $retval;
   $125 = $124;
   $126 = HEAP32[$125>>2]|0;
   $127 = (($124) + 4)|0;
   $128 = $127;
   $129 = HEAP32[$128>>2]|0;
   $130 = $maxread;
   $131 = $130;
   $132 = HEAP32[$131>>2]|0;
   $133 = (($130) + 4)|0;
   $134 = $133;
   $135 = HEAP32[$134>>2]|0;
   $136 = ($129|0)<($135|0);
   $137 = ($126>>>0)<($132>>>0);
   $138 = ($129|0)==($135|0);
   $139 = $138 & $137;
   $140 = $136 | $139;
   if (!($140)) {
    break;
   }
   $141 = $finfo;
   $142 = ((($141)) + 20|0);
   $143 = ((($142)) + 20|0);
   $144 = HEAP32[$143>>2]|0;
   $before = $144;
   $145 = $finfo;
   $146 = ((($145)) + 20|0);
   $147 = ((($146)) + 4|0);
   $148 = HEAP32[$147>>2]|0;
   $149 = ($148|0)==(0);
   if ($149) {
    $150 = $entry;
    $151 = ((($150)) + 28|0);
    $152 = HEAP32[$151>>2]|0;
    $153 = $finfo;
    $154 = ((($153)) + 8|0);
    $155 = HEAP32[$154>>2]|0;
    $156 = (($152) - ($155))|0;
    $157 = $br;
    $158 = $157;
    HEAP32[$158>>2] = $156;
    $159 = (($157) + 4)|0;
    $160 = $159;
    HEAP32[$160>>2] = 0;
    $161 = $br;
    $162 = $161;
    $163 = HEAP32[$162>>2]|0;
    $164 = (($161) + 4)|0;
    $165 = $164;
    $166 = HEAP32[$165>>2]|0;
    $167 = ($166|0)>(0);
    $168 = ($163>>>0)>(0);
    $169 = ($166|0)==(0);
    $170 = $169 & $168;
    $171 = $167 | $170;
    if ($171) {
     $172 = $br;
     $173 = $172;
     $174 = HEAP32[$173>>2]|0;
     $175 = (($172) + 4)|0;
     $176 = $175;
     $177 = HEAP32[$176>>2]|0;
     $178 = ($177|0)>(0);
     $179 = ($174>>>0)>(16384);
     $180 = ($177|0)==(0);
     $181 = $180 & $179;
     $182 = $178 | $181;
     if ($182) {
      $183 = $br;
      $184 = $183;
      HEAP32[$184>>2] = 16384;
      $185 = (($183) + 4)|0;
      $186 = $185;
      HEAP32[$186>>2] = 0;
     }
     $187 = $finfo;
     $188 = ((($187)) + 4|0);
     $189 = HEAP32[$188>>2]|0;
     $190 = $finfo;
     $191 = ((($190)) + 16|0);
     $192 = HEAP32[$191>>2]|0;
     $193 = $br;
     $194 = $193;
     $195 = HEAP32[$194>>2]|0;
     $196 = (($193) + 4)|0;
     $197 = $196;
     $198 = HEAP32[$197>>2]|0;
     $199 = (___PHYSFS_platformRead($189,$192,1,$195)|0);
     $200 = tempRet0;
     $201 = $br;
     $202 = $201;
     HEAP32[$202>>2] = $199;
     $203 = (($201) + 4)|0;
     $204 = $203;
     HEAP32[$204>>2] = $200;
     $205 = $br;
     $206 = $205;
     $207 = HEAP32[$206>>2]|0;
     $208 = (($205) + 4)|0;
     $209 = $208;
     $210 = HEAP32[$209>>2]|0;
     $211 = ($210|0)<(0);
     $212 = ($207>>>0)<=(0);
     $213 = ($210|0)==(0);
     $214 = $213 & $212;
     $215 = $211 | $214;
     if ($215) {
      break;
     }
     $216 = $br;
     $217 = $216;
     $218 = HEAP32[$217>>2]|0;
     $219 = (($216) + 4)|0;
     $220 = $219;
     $221 = HEAP32[$220>>2]|0;
     $222 = $finfo;
     $223 = ((($222)) + 8|0);
     $224 = HEAP32[$223>>2]|0;
     $225 = (($224) + ($218))|0;
     HEAP32[$223>>2] = $225;
     $226 = $finfo;
     $227 = ((($226)) + 16|0);
     $228 = HEAP32[$227>>2]|0;
     $229 = $finfo;
     $230 = ((($229)) + 20|0);
     HEAP32[$230>>2] = $228;
     $231 = $br;
     $232 = $231;
     $233 = HEAP32[$232>>2]|0;
     $234 = (($231) + 4)|0;
     $235 = $234;
     $236 = HEAP32[$235>>2]|0;
     $237 = $finfo;
     $238 = ((($237)) + 20|0);
     $239 = ((($238)) + 4|0);
     HEAP32[$239>>2] = $233;
    }
   }
   $240 = $finfo;
   $241 = ((($240)) + 20|0);
   $242 = (_inflate($241,2)|0);
   $243 = (_zlib_err($242)|0);
   $rc = $243;
   $244 = $finfo;
   $245 = ((($244)) + 20|0);
   $246 = ((($245)) + 20|0);
   $247 = HEAP32[$246>>2]|0;
   $248 = $before;
   $249 = (($247) - ($248))|0;
   $250 = $retval;
   $251 = $250;
   $252 = HEAP32[$251>>2]|0;
   $253 = (($250) + 4)|0;
   $254 = $253;
   $255 = HEAP32[$254>>2]|0;
   $256 = (_i64Add(($252|0),($255|0),($249|0),0)|0);
   $257 = tempRet0;
   $258 = $retval;
   $259 = $258;
   HEAP32[$259>>2] = $256;
   $260 = (($258) + 4)|0;
   $261 = $260;
   HEAP32[$261>>2] = $257;
   $262 = $rc;
   $263 = ($262|0)!=(0);
   if ($263) {
    break;
   }
  }
  $264 = $3;
  $265 = $retval;
  $266 = $265;
  $267 = HEAP32[$266>>2]|0;
  $268 = (($265) + 4)|0;
  $269 = $268;
  $270 = HEAP32[$269>>2]|0;
  $271 = (___divdi3(($267|0),($270|0),($264|0),0)|0);
  $272 = tempRet0;
  $273 = $retval;
  $274 = $273;
  HEAP32[$274>>2] = $271;
  $275 = (($273) + 4)|0;
  $276 = $275;
  HEAP32[$276>>2] = $272;
 }
 $277 = $retval;
 $278 = $277;
 $279 = HEAP32[$278>>2]|0;
 $280 = (($277) + 4)|0;
 $281 = $280;
 $282 = HEAP32[$281>>2]|0;
 $283 = ($282|0)>(0);
 $284 = ($279>>>0)>(0);
 $285 = ($282|0)==(0);
 $286 = $285 & $284;
 $287 = $283 | $286;
 if ($287) {
  $288 = $retval;
  $289 = $288;
  $290 = HEAP32[$289>>2]|0;
  $291 = (($288) + 4)|0;
  $292 = $291;
  $293 = HEAP32[$292>>2]|0;
  $294 = $3;
  $295 = (___muldi3(($290|0),($293|0),($294|0),0)|0);
  $296 = tempRet0;
  $297 = $finfo;
  $298 = ((($297)) + 12|0);
  $299 = HEAP32[$298>>2]|0;
  $300 = (($299) + ($295))|0;
  HEAP32[$298>>2] = $300;
 }
 $301 = $retval;
 $302 = $301;
 $303 = HEAP32[$302>>2]|0;
 $304 = (($301) + 4)|0;
 $305 = $304;
 $306 = HEAP32[$305>>2]|0;
 $307 = $0;
 $308 = $307;
 HEAP32[$308>>2] = $303;
 $309 = (($307) + 4)|0;
 $310 = $309;
 HEAP32[$310>>2] = $306;
 $311 = $0;
 $312 = $311;
 $313 = HEAP32[$312>>2]|0;
 $314 = (($311) + 4)|0;
 $315 = $314;
 $316 = HEAP32[$315>>2]|0;
 tempRet0 = ($316);
 STACKTOP = sp;return ($313|0);
}
function _ZIP_write($opaque,$buf,$objSize,$objCount) {
 $opaque = $opaque|0;
 $buf = $buf|0;
 $objSize = $objSize|0;
 $objCount = $objCount|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $buf;
 $2 = $objSize;
 $3 = $objCount;
 ___PHYSFS_setError(517504);
 tempRet0 = (-1);
 STACKTOP = sp;return -1;
}
function _ZIP_eof($opaque) {
 $opaque = $opaque|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $finfo = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $0;
 $finfo = $1;
 $2 = $finfo;
 $3 = ((($2)) + 12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $finfo;
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($6)) + 32|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($4>>>0)>=($8>>>0);
 $10 = $9&1;
 STACKTOP = sp;return ($10|0);
}
function _ZIP_tell($opaque) {
 $opaque = $opaque|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $0;
 $2 = ((($1)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 tempRet0 = (0);
 STACKTOP = sp;return ($3|0);
}
function _ZIP_seek($opaque,$0,$1) {
 $opaque = $opaque|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$ = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $buf = 0, $entry = 0, $finfo = 0, $in = 0, $maxread = 0, $newpos = 0, $str = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 624|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $4 = sp + 8|0;
 $newpos = sp;
 $str = sp + 24|0;
 $buf = sp + 104|0;
 $3 = $opaque;
 $5 = $4;
 $6 = $5;
 HEAP32[$6>>2] = $0;
 $7 = (($5) + 4)|0;
 $8 = $7;
 HEAP32[$8>>2] = $1;
 $9 = $3;
 $finfo = $9;
 $10 = $finfo;
 $11 = HEAP32[$10>>2]|0;
 $entry = $11;
 $12 = $finfo;
 $13 = ((($12)) + 4|0);
 $14 = HEAP32[$13>>2]|0;
 $in = $14;
 $15 = $4;
 $16 = $15;
 $17 = HEAP32[$16>>2]|0;
 $18 = (($15) + 4)|0;
 $19 = $18;
 $20 = HEAP32[$19>>2]|0;
 $21 = $entry;
 $22 = ((($21)) + 32|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = ($20>>>0)>(0);
 $25 = ($17>>>0)>($23>>>0);
 $26 = ($20|0)==(0);
 $27 = $26 & $25;
 $28 = $24 | $27;
 if ($28) {
  ___PHYSFS_setError(517528);
  $2 = 0;
  $123 = $2;
  STACKTOP = sp;return ($123|0);
 }
 $29 = $entry;
 $30 = ((($29)) + 20|0);
 $31 = HEAP16[$30>>1]|0;
 $32 = $31&65535;
 $33 = ($32|0)==(0);
 $34 = $4;
 $35 = $34;
 $36 = HEAP32[$35>>2]|0;
 $37 = (($34) + 4)|0;
 $38 = $37;
 $39 = HEAP32[$38>>2]|0;
 L5: do {
  if ($33) {
   $40 = $entry;
   $41 = ((($40)) + 12|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = (_i64Add(($36|0),($39|0),($42|0),0)|0);
   $44 = tempRet0;
   $45 = $newpos;
   $46 = $45;
   HEAP32[$46>>2] = $43;
   $47 = (($45) + 4)|0;
   $48 = $47;
   HEAP32[$48>>2] = $44;
   $49 = $in;
   $50 = $newpos;
   $51 = $50;
   $52 = HEAP32[$51>>2]|0;
   $53 = (($50) + 4)|0;
   $54 = $53;
   $55 = HEAP32[$54>>2]|0;
   $56 = (___PHYSFS_platformSeek($49,$52,$55)|0);
   $57 = ($56|0)!=(0);
   if ($57) {
    $58 = $4;
    $59 = $58;
    $60 = HEAP32[$59>>2]|0;
    $61 = (($58) + 4)|0;
    $62 = $61;
    $63 = HEAP32[$62>>2]|0;
    $64 = $finfo;
    $65 = ((($64)) + 12|0);
    HEAP32[$65>>2] = $60;
    break;
   }
   ___PHYSFS_setError(0);
   $2 = 0;
   $123 = $2;
   STACKTOP = sp;return ($123|0);
  } else {
   $66 = $finfo;
   $67 = ((($66)) + 12|0);
   $68 = HEAP32[$67>>2]|0;
   $69 = ($39>>>0)<(0);
   $70 = ($36>>>0)<($68>>>0);
   $71 = ($39|0)==(0);
   $72 = $71 & $70;
   $73 = $69 | $72;
   do {
    if ($73) {
     _initializeZStream($str);
     $74 = (_inflateInit2_($str,-15,613577,56)|0);
     $75 = (_zlib_err($74)|0);
     $76 = ($75|0)!=(0);
     if ($76) {
      $2 = 0;
      $123 = $2;
      STACKTOP = sp;return ($123|0);
     }
     $77 = $in;
     $78 = $entry;
     $79 = ((($78)) + 12|0);
     $80 = HEAP32[$79>>2]|0;
     $81 = (___PHYSFS_platformSeek($77,$80,0)|0);
     $82 = ($81|0)!=(0);
     if ($82) {
      $83 = $finfo;
      $84 = ((($83)) + 20|0);
      (_inflateEnd($84)|0);
      $85 = $finfo;
      $86 = ((($85)) + 20|0);
      dest=$86; src=$str; stop=dest+56|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
      $87 = $finfo;
      $88 = ((($87)) + 8|0);
      HEAP32[$88>>2] = 0;
      $89 = $finfo;
      $90 = ((($89)) + 12|0);
      HEAP32[$90>>2] = 0;
      break;
     }
     $2 = 0;
     $123 = $2;
     STACKTOP = sp;return ($123|0);
    }
   } while(0);
   while(1) {
    $91 = $finfo;
    $92 = ((($91)) + 12|0);
    $93 = HEAP32[$92>>2]|0;
    $94 = $4;
    $95 = $94;
    $96 = HEAP32[$95>>2]|0;
    $97 = (($94) + 4)|0;
    $98 = $97;
    $99 = HEAP32[$98>>2]|0;
    $100 = ($93|0)!=($96|0);
    $101 = (0)!=($99|0);
    $102 = $100 | $101;
    if (!($102)) {
     break L5;
    }
    $103 = $4;
    $104 = $103;
    $105 = HEAP32[$104>>2]|0;
    $106 = (($103) + 4)|0;
    $107 = $106;
    $108 = HEAP32[$107>>2]|0;
    $109 = $finfo;
    $110 = ((($109)) + 12|0);
    $111 = HEAP32[$110>>2]|0;
    $112 = (_i64Subtract(($105|0),($108|0),($111|0),0)|0);
    $113 = tempRet0;
    $maxread = $112;
    $114 = $maxread;
    $115 = ($114>>>0)>(512);
    $$ = $115 ? 512 : $112;
    $maxread = $$;
    $116 = $finfo;
    $117 = $maxread;
    $118 = (_ZIP_read($116,$buf,$117,1)|0);
    $119 = tempRet0;
    $120 = ($118|0)!=(1);
    $121 = ($119|0)!=(0);
    $122 = $120 | $121;
    if ($122) {
     break;
    }
   }
   $2 = 0;
   $123 = $2;
   STACKTOP = sp;return ($123|0);
  }
 } while(0);
 $2 = 1;
 $123 = $2;
 STACKTOP = sp;return ($123|0);
}
function _ZIP_fileLength($opaque) {
 $opaque = $opaque|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $finfo = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $opaque;
 $1 = $0;
 $finfo = $1;
 $2 = $finfo;
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 32|0);
 $5 = HEAP32[$4>>2]|0;
 tempRet0 = (0);
 STACKTOP = sp;return ($5|0);
}
function _PHYSFS_swapULE16($x) {
 $x = $x|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $x;
 $1 = $0;
 STACKTOP = sp;return ($1|0);
}
function _PHYSFS_swapULE32($x) {
 $x = $x|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = $x;
 $1 = $0;
 STACKTOP = sp;return ($1|0);
}
function _luaA_pushobject($L,$o) {
 $L = $L|0;
 $o = $o|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $o;
 $3 = $2;
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2) + 4)|0;
 $6 = $5;
 $7 = HEAP32[$6>>2]|0;
 $8 = $1;
 $9 = $8;
 HEAP32[$9>>2] = $4;
 $10 = (($8) + 4)|0;
 $11 = $10;
 HEAP32[$11>>2] = $7;
 $12 = ((($o)) + 8|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($1)) + 8|0);
 HEAP32[$14>>2] = $13;
 $15 = HEAP32[$0>>2]|0;
 $16 = ((($15)) + 16|0);
 HEAP32[$0>>2] = $16;
 return;
}
function _lua_checkstack($L,$size) {
 $L = $L|0;
 $size = $size|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $res$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($size|0)>(8000);
 if ($0) {
  $res$0 = 0;
  return ($res$0|0);
 }
 $1 = ((($L)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($L)) + 12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2) - ($4))|0;
 $6 = $5 >> 4;
 $7 = (($6) + ($size))|0;
 $8 = ($7|0)>(8000);
 $9 = $2;
 if ($8) {
  $res$0 = 0;
  return ($res$0|0);
 }
 $10 = ($size|0)>(0);
 if (!($10)) {
  $res$0 = 1;
  return ($res$0|0);
 }
 $11 = ((($L)) + 28|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($12) - ($2))|0;
 $14 = $size << 4;
 $15 = ($13|0)>($14|0);
 if ($15) {
  $21 = $9;
 } else {
  _luaD_growstack($L,$size);
  $$pre = HEAP32[$1>>2]|0;
  $21 = $$pre;
 }
 $16 = ((($L)) + 20|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ((($17)) + 8|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($21) + ($size<<4)|0);
 $22 = ($19>>>0)<($20>>>0);
 if (!($22)) {
  $res$0 = 1;
  return ($res$0|0);
 }
 HEAP32[$18>>2] = $20;
 $res$0 = 1;
 return ($res$0|0);
}
function _lua_xmove($from,$to,$n) {
 $from = $from|0;
 $to = $to|0;
 $n = $n|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $exitcond = 0, $exitcond2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($from|0)==($to|0);
 if ($0) {
  return;
 }
 $1 = ((($from)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (0 - ($n))|0;
 $4 = (($2) + ($3<<4)|0);
 HEAP32[$1>>2] = $4;
 $5 = ($n|0)>(0);
 if (!($5)) {
  return;
 }
 $6 = ((($to)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($7)) + 16|0);
 HEAP32[$6>>2] = $8;
 $9 = $4;
 $10 = $9;
 $11 = HEAP32[$10>>2]|0;
 $12 = (($9) + 4)|0;
 $13 = $12;
 $14 = HEAP32[$13>>2]|0;
 $15 = $7;
 $16 = $15;
 HEAP32[$16>>2] = $11;
 $17 = (($15) + 4)|0;
 $18 = $17;
 HEAP32[$18>>2] = $14;
 $19 = (((($2) + ($3<<4)|0)) + 8|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ((($7)) + 8|0);
 HEAP32[$21>>2] = $20;
 $exitcond2 = ($n|0)==(1);
 if ($exitcond2) {
  return;
 } else {
  $23 = 1;
 }
 while(1) {
  $$pre = HEAP32[$1>>2]|0;
  $22 = (($$pre) + ($23<<4)|0);
  $24 = HEAP32[$6>>2]|0;
  $25 = ((($24)) + 16|0);
  HEAP32[$6>>2] = $25;
  $26 = $22;
  $27 = $26;
  $28 = HEAP32[$27>>2]|0;
  $29 = (($26) + 4)|0;
  $30 = $29;
  $31 = HEAP32[$30>>2]|0;
  $32 = $24;
  $33 = $32;
  HEAP32[$33>>2] = $28;
  $34 = (($32) + 4)|0;
  $35 = $34;
  HEAP32[$35>>2] = $31;
  $36 = (((($$pre) + ($23<<4)|0)) + 8|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = ((($24)) + 8|0);
  HEAP32[$38>>2] = $37;
  $39 = (($23) + 1)|0;
  $exitcond = ($39|0)==($n|0);
  if ($exitcond) {
   break;
  } else {
   $23 = $39;
  }
 }
 return;
}
function _lua_setlevel($from,$to) {
 $from = $from|0;
 $to = $to|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($from)) + 52|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = ((($to)) + 52|0);
 HEAP16[$2>>1] = $1;
 return;
}
function _lua_atpanic($L,$panicf) {
 $L = $L|0;
 $panicf = $panicf|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 88|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$2>>2] = $panicf;
 return ($3|0);
}
function _lua_newthread($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 68|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 64|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($3>>>0)<($5>>>0);
 if (!($6)) {
  _luaC_step($L);
 }
 $7 = (_luaE_newthread($L)|0);
 $8 = ((($L)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 HEAP32[$9>>2] = $7;
 $10 = ((($9)) + 8|0);
 HEAP32[$10>>2] = 8;
 $11 = HEAP32[$8>>2]|0;
 $12 = ((($11)) + 16|0);
 HEAP32[$8>>2] = $12;
 return ($7|0);
}
function _lua_gettop($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($L)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) - ($3))|0;
 $5 = $4 >> 4;
 return ($5|0);
}
function _lua_settop($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $scevgep4 = 0, $scevgep7 = 0, $uglygep = 0, $uglygep6 = 0, $umax = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(-1);
 if (!($0)) {
  $16 = (($idx) + 1)|0;
  $17 = ((($L)) + 8|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = (($18) + ($16<<4)|0);
  HEAP32[$17>>2] = $19;
  return;
 }
 $1 = ((($L)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($L)) + 12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + ($idx<<4)|0);
 $6 = ($2>>>0)<($5>>>0);
 if ($6) {
  $7 = $2;
  $scevgep4 = ((($2)) + 16|0);
  $8 = ($5>>>0)>($scevgep4>>>0);
  $umax = $8 ? $5 : $scevgep4;
  $9 = $7 ^ -1;
  $uglygep = (($umax) + ($9)|0);
  $uglygep6 = $uglygep;
  $10 = $uglygep6 >>> 4;
  $11 = (($10) + 1)|0;
  $13 = $2;
  while(1) {
   $12 = ((($13)) + 16|0);
   $14 = ((($13)) + 8|0);
   HEAP32[$14>>2] = 0;
   $15 = ($12>>>0)<($5>>>0);
   if ($15) {
    $13 = $12;
   } else {
    break;
   }
  }
  $scevgep7 = (($2) + ($11<<4)|0);
  HEAP32[$1>>2] = $scevgep7;
 }
 HEAP32[$1>>2] = $5;
 return;
}
function _lua_remove($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$lcssa1 = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $p$0$ph = 0, $p$02 = 0, $p$02$phi = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $p$0$ph = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $p$0$ph = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $p$0$ph = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $p$0$ph = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $p$0$ph = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $p$0$ph = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($p$0$ph)) + 16|0);
 $39 = ((($L)) + 8|0);
 $40 = HEAP32[$39>>2]|0;
 $41 = ($38>>>0)<($40>>>0);
 if ($41) {
  $43 = $38;$p$02 = $p$0$ph;
 } else {
  $$lcssa1 = $40;
  $59 = ((($$lcssa1)) + -16|0);
  HEAP32[$39>>2] = $59;
  return;
 }
 while(1) {
  $42 = $43;
  $44 = $42;
  $45 = HEAP32[$44>>2]|0;
  $46 = (($42) + 4)|0;
  $47 = $46;
  $48 = HEAP32[$47>>2]|0;
  $49 = $p$02;
  $50 = $49;
  HEAP32[$50>>2] = $45;
  $51 = (($49) + 4)|0;
  $52 = $51;
  HEAP32[$52>>2] = $48;
  $53 = ((($p$02)) + 24|0);
  $54 = HEAP32[$53>>2]|0;
  $55 = ((($p$02)) + 8|0);
  HEAP32[$55>>2] = $54;
  $56 = ((($43)) + 16|0);
  $57 = HEAP32[$39>>2]|0;
  $58 = ($56>>>0)<($57>>>0);
  if ($58) {
   $p$02$phi = $43;$43 = $56;$p$02 = $p$02$phi;
  } else {
   $$lcssa1 = $57;
   break;
  }
 }
 $59 = ((($$lcssa1)) + -16|0);
 HEAP32[$39>>2] = $59;
 return;
}
function _lua_insert($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $8 = 0, $9 = 0, $q$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($L)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ($39>>>0)>($$1$i>>>0);
 if ($40) {
  $q$01 = $39;
 } else {
  $57 = $39;
  $56 = $57;
  $58 = $56;
  $59 = HEAP32[$58>>2]|0;
  $60 = (($56) + 4)|0;
  $61 = $60;
  $62 = HEAP32[$61>>2]|0;
  $63 = $$1$i;
  $64 = $63;
  HEAP32[$64>>2] = $59;
  $65 = (($63) + 4)|0;
  $66 = $65;
  HEAP32[$66>>2] = $62;
  $67 = ((($57)) + 8|0);
  $68 = HEAP32[$67>>2]|0;
  $69 = ((($$1$i)) + 8|0);
  HEAP32[$69>>2] = $68;
  return;
 }
 while(1) {
  $41 = ((($q$01)) + -16|0);
  $42 = $41;
  $43 = $42;
  $44 = HEAP32[$43>>2]|0;
  $45 = (($42) + 4)|0;
  $46 = $45;
  $47 = HEAP32[$46>>2]|0;
  $48 = $q$01;
  $49 = $48;
  HEAP32[$49>>2] = $44;
  $50 = (($48) + 4)|0;
  $51 = $50;
  HEAP32[$51>>2] = $47;
  $52 = ((($q$01)) + -8|0);
  $53 = HEAP32[$52>>2]|0;
  $54 = ((($q$01)) + 8|0);
  HEAP32[$54>>2] = $53;
  $55 = ($41>>>0)>($$1$i>>>0);
  if ($55) {
   $q$01 = $41;
  } else {
   break;
  }
 }
 $$pre = HEAP32[$38>>2]|0;
 $57 = $$pre;
 $56 = $57;
 $58 = $56;
 $59 = HEAP32[$58>>2]|0;
 $60 = (($56) + 4)|0;
 $61 = $60;
 $62 = HEAP32[$61>>2]|0;
 $63 = $$1$i;
 $64 = $63;
 HEAP32[$64>>2] = $59;
 $65 = (($63) + 4)|0;
 $66 = $65;
 HEAP32[$66>>2] = $62;
 $67 = ((($57)) + 8|0);
 $68 = HEAP32[$67>>2]|0;
 $69 = ((($$1$i)) + 8|0);
 HEAP32[$69>>2] = $68;
 return;
}
function _lua_replace($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$1$i$ph = 0, $$luaO_nilobject_$i = 0, $$pre = 0, $$pre$phi4Z2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($idx|0)==(-10001);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 20|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = ((($L)) + 40|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($2|0)==($4|0);
   if ($5) {
    _luaG_runerror($L,517545,$vararg_buffer);
    $$pre$phi4Z2D = $1;
    label = 13;
   } else {
    $$pre$phi4Z2D = $1;
    label = 13;
   }
  } else {
   $6 = ($idx|0)>(0);
   L5: do {
    if ($6) {
     $7 = ((($L)) + 12|0);
     $8 = HEAP32[$7>>2]|0;
     $9 = (($idx) + -1)|0;
     $10 = (($8) + ($9<<4)|0);
     $11 = ((($L)) + 8|0);
     $12 = HEAP32[$11>>2]|0;
     $13 = ($10>>>0)<($12>>>0);
     $$luaO_nilobject_$i = $13 ? $10 : 624096;
     $$1$i$ph = $$luaO_nilobject_$i;
    } else {
     $14 = ($idx|0)>(-10000);
     if ($14) {
      $15 = ((($L)) + 8|0);
      $16 = HEAP32[$15>>2]|0;
      $17 = (($16) + ($idx<<4)|0);
      $$1$i$ph = $17;
      break;
     }
     switch ($idx|0) {
     case -10000:  {
      $18 = ((($L)) + 16|0);
      $19 = HEAP32[$18>>2]|0;
      $20 = ((($19)) + 96|0);
      $$1$i$ph = $20;
      break L5;
      break;
     }
     case -10001:  {
      $$pre = ((($L)) + 20|0);
      $$pre$phi4Z2D = $$pre;
      label = 13;
      break L1;
      break;
     }
     case -10002:  {
      $21 = ((($L)) + 72|0);
      $$1$i$ph = $21;
      break L5;
      break;
     }
     default: {
      $22 = ((($L)) + 20|0);
      $23 = HEAP32[$22>>2]|0;
      $24 = ((($23)) + 4|0);
      $25 = HEAP32[$24>>2]|0;
      $26 = HEAP32[$25>>2]|0;
      $27 = (-10002 - ($idx))|0;
      $28 = ((($26)) + 7|0);
      $29 = HEAP8[$28>>0]|0;
      $30 = $29&255;
      $31 = ($27|0)>($30|0);
      $32 = (($27) + -1)|0;
      $33 = (((($26)) + 24|0) + ($32<<4)|0);
      $34 = $31 ? 624096 : $33;
      $$1$i$ph = $34;
      break L5;
     }
     }
    }
   } while(0);
   $62 = ((($L)) + 8|0);
   $63 = HEAP32[$62>>2]|0;
   $64 = ((($63)) + -16|0);
   $65 = $64;
   $66 = $65;
   $67 = HEAP32[$66>>2]|0;
   $68 = (($65) + 4)|0;
   $69 = $68;
   $70 = HEAP32[$69>>2]|0;
   $71 = $$1$i$ph;
   $72 = $71;
   HEAP32[$72>>2] = $67;
   $73 = (($71) + 4)|0;
   $74 = $73;
   HEAP32[$74>>2] = $70;
   $75 = ((($63)) + -8|0);
   $76 = HEAP32[$75>>2]|0;
   $77 = ((($$1$i$ph)) + 8|0);
   HEAP32[$77>>2] = $76;
   $78 = ($idx|0)<(-10002);
   if ($78) {
    $79 = HEAP32[$62>>2]|0;
    $80 = ((($79)) + -8|0);
    $81 = HEAP32[$80>>2]|0;
    $82 = ($81|0)>(3);
    if ($82) {
     $83 = ((($79)) + -16|0);
     $84 = HEAP32[$83>>2]|0;
     $85 = ((($84)) + 5|0);
     $86 = HEAP8[$85>>0]|0;
     $87 = $86 & 3;
     $88 = ($87<<24>>24)==(0);
     if ($88) {
      $$pre$phiZ2D = $62;
     } else {
      $89 = ((($L)) + 20|0);
      $90 = HEAP32[$89>>2]|0;
      $91 = ((($90)) + 4|0);
      $92 = HEAP32[$91>>2]|0;
      $93 = HEAP32[$92>>2]|0;
      $94 = ((($93)) + 5|0);
      $95 = HEAP8[$94>>0]|0;
      $96 = $95 & 4;
      $97 = ($96<<24>>24)==(0);
      if ($97) {
       $$pre$phiZ2D = $62;
      } else {
       _luaC_barrierf($L,$93,$84);
       $$pre$phiZ2D = $62;
      }
     }
    } else {
     $$pre$phiZ2D = $62;
    }
   } else {
    $$pre$phiZ2D = $62;
   }
  }
 } while(0);
 if ((label|0) == 13) {
  $35 = HEAP32[$$pre$phi4Z2D>>2]|0;
  $36 = ((($35)) + 4|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = HEAP32[$37>>2]|0;
  $39 = ((($L)) + 88|0);
  $40 = ((($38)) + 12|0);
  $41 = HEAP32[$40>>2]|0;
  HEAP32[$39>>2] = $41;
  $42 = ((($L)) + 96|0);
  HEAP32[$42>>2] = 5;
  $43 = HEAP32[$36>>2]|0;
  $44 = HEAP32[$43>>2]|0;
  $45 = ((($L)) + 8|0);
  $46 = HEAP32[$45>>2]|0;
  $47 = ((($46)) + -16|0);
  $48 = HEAP32[$47>>2]|0;
  $49 = ((($44)) + 12|0);
  HEAP32[$49>>2] = $48;
  $50 = ((($46)) + -8|0);
  $51 = HEAP32[$50>>2]|0;
  $52 = ($51|0)>(3);
  if ($52) {
   $53 = HEAP32[$47>>2]|0;
   $54 = ((($53)) + 5|0);
   $55 = HEAP8[$54>>0]|0;
   $56 = $55 & 3;
   $57 = ($56<<24>>24)==(0);
   if ($57) {
    $$pre$phiZ2D = $45;
   } else {
    $58 = ((($44)) + 5|0);
    $59 = HEAP8[$58>>0]|0;
    $60 = $59 & 4;
    $61 = ($60<<24>>24)==(0);
    if ($61) {
     $$pre$phiZ2D = $45;
    } else {
     _luaC_barrierf($L,$44,$53);
     $$pre$phiZ2D = $45;
    }
   }
  } else {
   $$pre$phiZ2D = $45;
  }
 }
 $98 = HEAP32[$$pre$phiZ2D>>2]|0;
 $99 = ((($98)) + -16|0);
 HEAP32[$$pre$phiZ2D>>2] = $99;
 STACKTOP = sp;return;
}
function _lua_pushvalue($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($L)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = $$1$i;
 $41 = $40;
 $42 = HEAP32[$41>>2]|0;
 $43 = (($40) + 4)|0;
 $44 = $43;
 $45 = HEAP32[$44>>2]|0;
 $46 = $39;
 $47 = $46;
 HEAP32[$47>>2] = $42;
 $48 = (($46) + 4)|0;
 $49 = $48;
 HEAP32[$49>>2] = $45;
 $50 = ((($$1$i)) + 8|0);
 $51 = HEAP32[$50>>2]|0;
 $52 = ((($39)) + 8|0);
 HEAP32[$52>>2] = $51;
 $53 = HEAP32[$38>>2]|0;
 $54 = ((($53)) + 16|0);
 HEAP32[$38>>2] = $54;
 return;
}
function _lua_type($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$1$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   if ($7) {
    $$1$i = $4;
   } else {
    $40 = -1;
    return ($40|0);
   }
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    if ($34) {
     $40 = -1;
    } else {
     $$1$i = $36;
     break L1;
    }
    return ($40|0);
   }
   }
  }
 } while(0);
 $37 = ($$1$i|0)==(624096|0);
 if ($37) {
  $40 = -1;
  return ($40|0);
 }
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = $39;
 return ($40|0);
}
function _lua_typename($L,$t) {
 $L = $L|0;
 $t = $t|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($t|0)==(-1);
 if ($0) {
  $3 = 522797;
  return ($3|0);
 }
 $1 = (38844 + ($t<<2)|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2;
 return ($3|0);
}
function _lua_iscfunction($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ($39|0)==(6);
 if (!($40)) {
  $46 = 0;
  $45 = $46&1;
  return ($45|0);
 }
 $41 = HEAP32[$$1$i>>2]|0;
 $42 = ((($41)) + 6|0);
 $43 = HEAP8[$42>>0]|0;
 $44 = ($43<<24>>24)!=(0);
 $46 = $44;
 $45 = $46&1;
 return ($45|0);
}
function _lua_isnumber($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $n = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $n = sp;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ($39|0)==(3);
 if ($40) {
  $44 = 1;
  $43 = $44&1;
  STACKTOP = sp;return ($43|0);
 }
 $41 = (_luaV_tonumber($$1$i,$n)|0);
 $42 = ($41|0)!=(0|0);
 $44 = $42;
 $43 = $44&1;
 STACKTOP = sp;return ($43|0);
}
function _lua_isstring($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$1$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, $phitmp1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   if ($7) {
    $$1$i$i = $4;
   } else {
    $41 = 0;
    $40 = $41&1;
    return ($40|0);
   }
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    if ($34) {
     $41 = 0;
    } else {
     $$1$i$i = $36;
     break L1;
    }
    $40 = $41&1;
    return ($40|0);
   }
   }
  }
 } while(0);
 $37 = ($$1$i$i|0)==(624096|0);
 if ($37) {
  $41 = 0;
  $40 = $41&1;
  return ($40|0);
 }
 $38 = ((($$1$i$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $phitmp = (($39) + -3)|0;
 $phitmp1 = ($phitmp>>>0)<(2);
 $41 = $phitmp1;
 $40 = $41&1;
 return ($40|0);
}
function _lua_isuserdata($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$ = 0, $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ($39|0)==(7);
 $41 = ($39|0)==(2);
 $$ = $40 | $41;
 $42 = $$&1;
 return ($42|0);
}
function _lua_rawequal($L,$index1,$index2) {
 $L = $L|0;
 $index1 = $index1|0;
 $index2 = $index2|0;
 var $$1$i = 0, $$1$i3 = 0, $$luaO_nilobject_$i = 0, $$luaO_nilobject_$i2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($index1|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($index1) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($index1|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($index1<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($index1|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($index1))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ($index2|0)>(0);
 L13: do {
  if ($38) {
   $39 = ((($L)) + 12|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = (($index2) + -1)|0;
   $42 = (($40) + ($41<<4)|0);
   $43 = ((($L)) + 8|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = ($42>>>0)<($44>>>0);
   $$luaO_nilobject_$i2 = $45 ? $42 : 624096;
   $$1$i3 = $$luaO_nilobject_$i2;
  } else {
   $46 = ($index2|0)>(-10000);
   if ($46) {
    $47 = ((($L)) + 8|0);
    $48 = HEAP32[$47>>2]|0;
    $49 = (($48) + ($index2<<4)|0);
    $$1$i3 = $49;
    break;
   }
   switch ($index2|0) {
   case -10000:  {
    $50 = ((($L)) + 16|0);
    $51 = HEAP32[$50>>2]|0;
    $52 = ((($51)) + 96|0);
    $$1$i3 = $52;
    break L13;
    break;
   }
   case -10001:  {
    $53 = ((($L)) + 20|0);
    $54 = HEAP32[$53>>2]|0;
    $55 = ((($54)) + 4|0);
    $56 = HEAP32[$55>>2]|0;
    $57 = HEAP32[$56>>2]|0;
    $58 = ((($L)) + 88|0);
    $59 = ((($57)) + 12|0);
    $60 = HEAP32[$59>>2]|0;
    HEAP32[$58>>2] = $60;
    $61 = ((($L)) + 96|0);
    HEAP32[$61>>2] = 5;
    $$1$i3 = $58;
    break L13;
    break;
   }
   case -10002:  {
    $62 = ((($L)) + 72|0);
    $$1$i3 = $62;
    break L13;
    break;
   }
   default: {
    $63 = ((($L)) + 20|0);
    $64 = HEAP32[$63>>2]|0;
    $65 = ((($64)) + 4|0);
    $66 = HEAP32[$65>>2]|0;
    $67 = HEAP32[$66>>2]|0;
    $68 = (-10002 - ($index2))|0;
    $69 = ((($67)) + 7|0);
    $70 = HEAP8[$69>>0]|0;
    $71 = $70&255;
    $72 = ($68|0)>($71|0);
    $73 = (($68) + -1)|0;
    $74 = (((($67)) + 24|0) + ($73<<4)|0);
    $75 = $72 ? 624096 : $74;
    $$1$i3 = $75;
    break L13;
   }
   }
  }
 } while(0);
 $76 = ($$1$i|0)==(624096|0);
 $77 = ($$1$i3|0)==(624096|0);
 $or$cond = $76 | $77;
 if ($or$cond) {
  $79 = 0;
  return ($79|0);
 }
 $78 = (_luaO_rawequalObj($$1$i,$$1$i3)|0);
 $79 = $78;
 return ($79|0);
}
function _lua_lessthan($L,$index1,$index2) {
 $L = $L|0;
 $index1 = $index1|0;
 $index2 = $index2|0;
 var $$1$i = 0, $$1$i3 = 0, $$luaO_nilobject_$i = 0, $$luaO_nilobject_$i2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($index1|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($index1) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($index1|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($index1<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($index1|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($index1))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ($index2|0)>(0);
 L13: do {
  if ($38) {
   $39 = ((($L)) + 12|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = (($index2) + -1)|0;
   $42 = (($40) + ($41<<4)|0);
   $43 = ((($L)) + 8|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = ($42>>>0)<($44>>>0);
   $$luaO_nilobject_$i2 = $45 ? $42 : 624096;
   $$1$i3 = $$luaO_nilobject_$i2;
  } else {
   $46 = ($index2|0)>(-10000);
   if ($46) {
    $47 = ((($L)) + 8|0);
    $48 = HEAP32[$47>>2]|0;
    $49 = (($48) + ($index2<<4)|0);
    $$1$i3 = $49;
    break;
   }
   switch ($index2|0) {
   case -10000:  {
    $50 = ((($L)) + 16|0);
    $51 = HEAP32[$50>>2]|0;
    $52 = ((($51)) + 96|0);
    $$1$i3 = $52;
    break L13;
    break;
   }
   case -10001:  {
    $53 = ((($L)) + 20|0);
    $54 = HEAP32[$53>>2]|0;
    $55 = ((($54)) + 4|0);
    $56 = HEAP32[$55>>2]|0;
    $57 = HEAP32[$56>>2]|0;
    $58 = ((($L)) + 88|0);
    $59 = ((($57)) + 12|0);
    $60 = HEAP32[$59>>2]|0;
    HEAP32[$58>>2] = $60;
    $61 = ((($L)) + 96|0);
    HEAP32[$61>>2] = 5;
    $$1$i3 = $58;
    break L13;
    break;
   }
   case -10002:  {
    $62 = ((($L)) + 72|0);
    $$1$i3 = $62;
    break L13;
    break;
   }
   default: {
    $63 = ((($L)) + 20|0);
    $64 = HEAP32[$63>>2]|0;
    $65 = ((($64)) + 4|0);
    $66 = HEAP32[$65>>2]|0;
    $67 = HEAP32[$66>>2]|0;
    $68 = (-10002 - ($index2))|0;
    $69 = ((($67)) + 7|0);
    $70 = HEAP8[$69>>0]|0;
    $71 = $70&255;
    $72 = ($68|0)>($71|0);
    $73 = (($68) + -1)|0;
    $74 = (((($67)) + 24|0) + ($73<<4)|0);
    $75 = $72 ? 624096 : $74;
    $$1$i3 = $75;
    break L13;
   }
   }
  }
 } while(0);
 $76 = ($$1$i|0)==(624096|0);
 $77 = ($$1$i3|0)==(624096|0);
 $or$cond = $76 | $77;
 if ($or$cond) {
  $79 = 0;
  return ($79|0);
 }
 $78 = (_luaV_lessthan($L,$$1$i,$$1$i3)|0);
 $79 = $78;
 return ($79|0);
}
function _lua_tonumber($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$0 = 0.0, $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $n = 0, $o$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $n = sp;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ($39|0)==(3);
 if ($40) {
  $o$0 = $$1$i;
 } else {
  $41 = (_luaV_tonumber($$1$i,$n)|0);
  $42 = ($41|0)==(0|0);
  if ($42) {
   $$0 = 0.0;
   STACKTOP = sp;return (+$$0);
  } else {
   $o$0 = $41;
  }
 }
 $43 = +HEAPF64[$o$0>>3];
 $$0 = $43;
 STACKTOP = sp;return (+$$0);
}
function _lua_tointeger($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$0 = 0, $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0.0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $n = 0, $o$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $n = sp;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ($39|0)==(3);
 if ($40) {
  $o$0 = $$1$i;
 } else {
  $41 = (_luaV_tonumber($$1$i,$n)|0);
  $42 = ($41|0)==(0|0);
  if ($42) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  } else {
   $o$0 = $41;
  }
 }
 $43 = +HEAPF64[$o$0>>3];
 $44 = (~~(($43)));
 $$0 = $44;
 STACKTOP = sp;return ($$0|0);
}
function _lua_toboolean($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ($39|0)==(0);
 if ($40) {
  $44 = 0;
  $43 = $44&1;
  return ($43|0);
 }
 $41 = ($39|0)==(1);
 if (!($41)) {
  $44 = 1;
  $43 = $44&1;
  return ($43|0);
 }
 $42 = HEAP32[$$1$i>>2]|0;
 $phitmp = ($42|0)!=(0);
 $44 = $phitmp;
 $43 = $44&1;
 return ($43|0);
}
function _lua_tolstring($L,$idx,$len) {
 $L = $L|0;
 $idx = $idx|0;
 $len = $len|0;
 var $$0 = 0, $$1$i = 0, $$luaO_nilobject_$i = 0, $$luaO_nilobject_$i1 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $o$0 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ($39|0)==(4);
 L13: do {
  if ($40) {
   $o$0 = $$1$i;
  } else {
   $41 = (_luaV_tostring($L,$$1$i)|0);
   $42 = ($41|0)==(0);
   if ($42) {
    $43 = ($len|0)==(0|0);
    if ($43) {
     $$0 = 0;
     return ($$0|0);
    }
    HEAP32[$len>>2] = 0;
    $$0 = 0;
    return ($$0|0);
   }
   $44 = ((($L)) + 16|0);
   $45 = HEAP32[$44>>2]|0;
   $46 = ((($45)) + 68|0);
   $47 = HEAP32[$46>>2]|0;
   $48 = ((($45)) + 64|0);
   $49 = HEAP32[$48>>2]|0;
   $50 = ($47>>>0)<($49>>>0);
   if (!($50)) {
    _luaC_step($L);
   }
   if ($0) {
    $51 = ((($L)) + 12|0);
    $52 = HEAP32[$51>>2]|0;
    $53 = (($idx) + -1)|0;
    $54 = (($52) + ($53<<4)|0);
    $55 = ((($L)) + 8|0);
    $56 = HEAP32[$55>>2]|0;
    $57 = ($54>>>0)<($56>>>0);
    $$luaO_nilobject_$i1 = $57 ? $54 : 624096;
    $o$0 = $$luaO_nilobject_$i1;
    break;
   }
   $58 = ($idx|0)>(-10000);
   if ($58) {
    $59 = ((($L)) + 8|0);
    $60 = HEAP32[$59>>2]|0;
    $61 = (($60) + ($idx<<4)|0);
    $o$0 = $61;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $62 = HEAP32[$44>>2]|0;
    $63 = ((($62)) + 96|0);
    $o$0 = $63;
    break L13;
    break;
   }
   case -10001:  {
    $64 = ((($L)) + 20|0);
    $65 = HEAP32[$64>>2]|0;
    $66 = ((($65)) + 4|0);
    $67 = HEAP32[$66>>2]|0;
    $68 = HEAP32[$67>>2]|0;
    $69 = ((($L)) + 88|0);
    $70 = ((($68)) + 12|0);
    $71 = HEAP32[$70>>2]|0;
    HEAP32[$69>>2] = $71;
    $72 = ((($L)) + 96|0);
    HEAP32[$72>>2] = 5;
    $o$0 = $69;
    break L13;
    break;
   }
   case -10002:  {
    $73 = ((($L)) + 72|0);
    $o$0 = $73;
    break L13;
    break;
   }
   default: {
    $74 = ((($L)) + 20|0);
    $75 = HEAP32[$74>>2]|0;
    $76 = ((($75)) + 4|0);
    $77 = HEAP32[$76>>2]|0;
    $78 = HEAP32[$77>>2]|0;
    $79 = (-10002 - ($idx))|0;
    $80 = ((($78)) + 7|0);
    $81 = HEAP8[$80>>0]|0;
    $82 = $81&255;
    $83 = ($79|0)>($82|0);
    $84 = (($79) + -1)|0;
    $85 = (((($78)) + 24|0) + ($84<<4)|0);
    $86 = $83 ? 624096 : $85;
    $o$0 = $86;
    break L13;
   }
   }
  }
 } while(0);
 $87 = ($len|0)==(0|0);
 if ($87) {
  $$pre = HEAP32[$o$0>>2]|0;
  $92 = $$pre;
 } else {
  $88 = HEAP32[$o$0>>2]|0;
  $89 = ((($88)) + 12|0);
  $90 = HEAP32[$89>>2]|0;
  HEAP32[$len>>2] = $90;
  $92 = $88;
 }
 $91 = ((($92)) + 16|0);
 $$0 = $91;
 return ($$0|0);
}
function _lua_objlen($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$0 = 0, $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 switch ($39|0) {
 case 4:  {
  $40 = HEAP32[$$1$i>>2]|0;
  $41 = ((($40)) + 12|0);
  $42 = HEAP32[$41>>2]|0;
  $$0 = $42;
  return ($$0|0);
  break;
 }
 case 7:  {
  $43 = HEAP32[$$1$i>>2]|0;
  $44 = ((($43)) + 16|0);
  $45 = HEAP32[$44>>2]|0;
  $$0 = $45;
  return ($$0|0);
  break;
 }
 case 5:  {
  $46 = HEAP32[$$1$i>>2]|0;
  $47 = (_luaH_getn($46)|0);
  $$0 = $47;
  return ($$0|0);
  break;
 }
 case 3:  {
  $48 = (_luaV_tostring($L,$$1$i)|0);
  $49 = ($48|0)==(0);
  if ($49) {
   $$0 = 0;
   return ($$0|0);
  }
  $50 = HEAP32[$$1$i>>2]|0;
  $51 = ((($50)) + 12|0);
  $52 = HEAP32[$51>>2]|0;
  $$0 = $52;
  return ($$0|0);
  break;
 }
 default: {
  $$0 = 0;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function _lua_tocfunction($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ($39|0)==(6);
 if (!($40)) {
  $47 = 0;
  return ($47|0);
 }
 $41 = HEAP32[$$1$i>>2]|0;
 $42 = ((($41)) + 6|0);
 $43 = HEAP8[$42>>0]|0;
 $44 = ($43<<24>>24)==(0);
 if ($44) {
  $47 = 0;
  return ($47|0);
 }
 $45 = ((($41)) + 16|0);
 $46 = HEAP32[$45>>2]|0;
 $47 = $46;
 return ($47|0);
}
function _lua_touserdata($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$0 = 0, $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 switch ($39|0) {
 case 7:  {
  $40 = HEAP32[$$1$i>>2]|0;
  $41 = ((($40)) + 24|0);
  $$0 = $41;
  return ($$0|0);
  break;
 }
 case 2:  {
  $42 = HEAP32[$$1$i>>2]|0;
  $$0 = $42;
  return ($$0|0);
  break;
 }
 default: {
  $$0 = 0;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function _lua_tothread($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ($39|0)==(8);
 if (!($40)) {
  $42 = 0;
  return ($42|0);
 }
 $41 = HEAP32[$$1$i>>2]|0;
 $42 = $41;
 return ($42|0);
}
function _lua_topointer($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$0 = 0, $$1$i = 0, $$1$i$i = 0, $$luaO_nilobject_$i = 0, $$luaO_nilobject_$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 switch ($39|0) {
 case 5:  {
  $40 = HEAP32[$$1$i>>2]|0;
  $$0 = $40;
  return ($$0|0);
  break;
 }
 case 6:  {
  $41 = HEAP32[$$1$i>>2]|0;
  $$0 = $41;
  return ($$0|0);
  break;
 }
 case 8:  {
  $42 = HEAP32[$$1$i>>2]|0;
  $$0 = $42;
  return ($$0|0);
  break;
 }
 case 2: case 7:  {
  L21: do {
   if ($0) {
    $43 = ((($L)) + 12|0);
    $44 = HEAP32[$43>>2]|0;
    $45 = (($idx) + -1)|0;
    $46 = (($44) + ($45<<4)|0);
    $47 = ((($L)) + 8|0);
    $48 = HEAP32[$47>>2]|0;
    $49 = ($46>>>0)<($48>>>0);
    $$luaO_nilobject_$i$i = $49 ? $46 : 624096;
    $$1$i$i = $$luaO_nilobject_$i$i;
   } else {
    $50 = ($idx|0)>(-10000);
    if ($50) {
     $51 = ((($L)) + 8|0);
     $52 = HEAP32[$51>>2]|0;
     $53 = (($52) + ($idx<<4)|0);
     $$1$i$i = $53;
     break;
    }
    switch ($idx|0) {
    case -10000:  {
     $54 = ((($L)) + 16|0);
     $55 = HEAP32[$54>>2]|0;
     $56 = ((($55)) + 96|0);
     $$1$i$i = $56;
     break L21;
     break;
    }
    case -10001:  {
     $57 = ((($L)) + 20|0);
     $58 = HEAP32[$57>>2]|0;
     $59 = ((($58)) + 4|0);
     $60 = HEAP32[$59>>2]|0;
     $61 = HEAP32[$60>>2]|0;
     $62 = ((($L)) + 88|0);
     $63 = ((($61)) + 12|0);
     $64 = HEAP32[$63>>2]|0;
     HEAP32[$62>>2] = $64;
     $65 = ((($L)) + 96|0);
     HEAP32[$65>>2] = 5;
     $$1$i$i = $62;
     break L21;
     break;
    }
    case -10002:  {
     $66 = ((($L)) + 72|0);
     $$1$i$i = $66;
     break L21;
     break;
    }
    default: {
     $67 = ((($L)) + 20|0);
     $68 = HEAP32[$67>>2]|0;
     $69 = ((($68)) + 4|0);
     $70 = HEAP32[$69>>2]|0;
     $71 = HEAP32[$70>>2]|0;
     $72 = (-10002 - ($idx))|0;
     $73 = ((($71)) + 7|0);
     $74 = HEAP8[$73>>0]|0;
     $75 = $74&255;
     $76 = ($72|0)>($75|0);
     $77 = (($72) + -1)|0;
     $78 = (((($71)) + 24|0) + ($77<<4)|0);
     $79 = $76 ? 624096 : $78;
     $$1$i$i = $79;
     break L21;
    }
    }
   }
  } while(0);
  $80 = ((($$1$i$i)) + 8|0);
  $81 = HEAP32[$80>>2]|0;
  switch ($81|0) {
  case 7:  {
   $82 = HEAP32[$$1$i$i>>2]|0;
   $83 = ((($82)) + 24|0);
   $$0 = $83;
   return ($$0|0);
   break;
  }
  case 2:  {
   $84 = HEAP32[$$1$i$i>>2]|0;
   $$0 = $84;
   return ($$0|0);
   break;
  }
  default: {
   $$0 = 0;
   return ($$0|0);
  }
  }
  break;
 }
 default: {
  $$0 = 0;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function _lua_pushnil($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($1)) + 16|0);
 HEAP32[$0>>2] = $3;
 return;
}
function _lua_pushnumber($L,$n) {
 $L = $L|0;
 $n = +$n;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 HEAPF64[$1>>3] = $n;
 $2 = ((($1)) + 8|0);
 HEAP32[$2>>2] = 3;
 $3 = ((($1)) + 16|0);
 HEAP32[$0>>2] = $3;
 return;
}
function _lua_pushinteger($L,$n) {
 $L = $L|0;
 $n = $n|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (+($n|0));
 HEAPF64[$1>>3] = $2;
 $3 = ((($1)) + 8|0);
 HEAP32[$3>>2] = 3;
 $4 = ((($1)) + 16|0);
 HEAP32[$0>>2] = $4;
 return;
}
function _lua_pushlstring($L,$s,$len) {
 $L = $L|0;
 $s = $s|0;
 $len = $len|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 68|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 64|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($3>>>0)<($5>>>0);
 if (!($6)) {
  _luaC_step($L);
 }
 $7 = ((($L)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (_luaS_newlstr($L,$s,$len)|0);
 HEAP32[$8>>2] = $9;
 $10 = ((($8)) + 8|0);
 HEAP32[$10>>2] = 4;
 $11 = HEAP32[$7>>2]|0;
 $12 = ((($11)) + 16|0);
 HEAP32[$7>>2] = $12;
 return;
}
function _lua_pushstring($L,$s) {
 $L = $L|0;
 $s = $s|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ($s|0)==(0|0);
 if ($0) {
  $1 = ((($L)) + 8|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ((($2)) + 8|0);
  HEAP32[$3>>2] = 0;
  $4 = ((($2)) + 16|0);
  HEAP32[$1>>2] = $4;
  return;
 }
 $5 = (_strlen($s)|0);
 $6 = ((($L)) + 16|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($7)) + 68|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($7)) + 64|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($9>>>0)<($11>>>0);
 if (!($12)) {
  _luaC_step($L);
 }
 $13 = ((($L)) + 8|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (_luaS_newlstr($L,$s,$5)|0);
 HEAP32[$14>>2] = $15;
 $16 = ((($14)) + 8|0);
 HEAP32[$16>>2] = 4;
 $17 = HEAP32[$13>>2]|0;
 $18 = ((($17)) + 16|0);
 HEAP32[$13>>2] = $18;
 return;
}
function _lua_pushvfstring($L,$fmt,$argp) {
 $L = $L|0;
 $fmt = $fmt|0;
 $argp = $argp|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 68|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 64|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($3>>>0)<($5>>>0);
 if (!($6)) {
  _luaC_step($L);
 }
 $7 = (_luaO_pushvfstring($L,$fmt,$argp)|0);
 return ($7|0);
}
function _lua_pushfstring($L,$fmt,$varargs) {
 $L = $L|0;
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $argp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $argp = sp;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 68|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 64|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($3>>>0)<($5>>>0);
 if (!($6)) {
  _luaC_step($L);
 }
 HEAP32[$argp>>2] = $varargs;
 $7 = (_luaO_pushvfstring($L,$fmt,$argp)|0);
 STACKTOP = sp;return ($7|0);
}
function _lua_pushcclosure($L,$fn,$n) {
 $L = $L|0;
 $fn = $fn|0;
 $n = $n|0;
 var $$0$i = 0, $$0$in$i = 0, $$in = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 68|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 64|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($3>>>0)<($5>>>0);
 if (!($6)) {
  _luaC_step($L);
 }
 $7 = ((($L)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($L)) + 40|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($8|0)==($10|0);
 if ($11) {
  $12 = ((($L)) + 72|0);
  $$0$in$i = $12;
 } else {
  $13 = ((($8)) + 4|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = HEAP32[$14>>2]|0;
  $16 = ((($15)) + 12|0);
  $$0$in$i = $16;
 }
 $$0$i = HEAP32[$$0$in$i>>2]|0;
 $17 = (_luaF_newCclosure($L,$n,$$0$i)|0);
 $18 = ((($17)) + 16|0);
 HEAP32[$18>>2] = $fn;
 $19 = ((($L)) + 8|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (0 - ($n))|0;
 $22 = (($20) + ($21<<4)|0);
 HEAP32[$19>>2] = $22;
 $23 = ($n|0)==(0);
 if ($23) {
  $$lcssa = $22;
  HEAP32[$$lcssa>>2] = $17;
  $43 = ((($$lcssa)) + 8|0);
  HEAP32[$43>>2] = 6;
  $44 = HEAP32[$19>>2]|0;
  $45 = ((($44)) + 16|0);
  HEAP32[$19>>2] = $45;
  return;
 } else {
  $$in = $n;$26 = $22;
 }
 while(1) {
  $24 = (($$in) + -1)|0;
  $25 = (($26) + ($24<<4)|0);
  $27 = (((($17)) + 24|0) + ($24<<4)|0);
  $28 = $25;
  $29 = $28;
  $30 = HEAP32[$29>>2]|0;
  $31 = (($28) + 4)|0;
  $32 = $31;
  $33 = HEAP32[$32>>2]|0;
  $34 = $27;
  $35 = $34;
  HEAP32[$35>>2] = $30;
  $36 = (($34) + 4)|0;
  $37 = $36;
  HEAP32[$37>>2] = $33;
  $38 = (((($26) + ($24<<4)|0)) + 8|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = (((((($17)) + 24|0) + ($24<<4)|0)) + 8|0);
  HEAP32[$40>>2] = $39;
  $41 = ($24|0)==(0);
  $42 = HEAP32[$19>>2]|0;
  if ($41) {
   $$lcssa = $42;
   break;
  } else {
   $$in = $24;$26 = $42;
  }
 }
 HEAP32[$$lcssa>>2] = $17;
 $43 = ((($$lcssa)) + 8|0);
 HEAP32[$43>>2] = 6;
 $44 = HEAP32[$19>>2]|0;
 $45 = ((($44)) + 16|0);
 HEAP32[$19>>2] = $45;
 return;
}
function _lua_pushboolean($L,$b) {
 $L = $L|0;
 $b = $b|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($b|0)!=(0);
 $3 = $2&1;
 HEAP32[$1>>2] = $3;
 $4 = ((($1)) + 8|0);
 HEAP32[$4>>2] = 1;
 $5 = ((($1)) + 16|0);
 HEAP32[$0>>2] = $5;
 return;
}
function _lua_pushlightuserdata($L,$p) {
 $L = $L|0;
 $p = $p|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$1>>2] = $p;
 $2 = ((($1)) + 8|0);
 HEAP32[$2>>2] = 2;
 $3 = HEAP32[$0>>2]|0;
 $4 = ((($3)) + 16|0);
 HEAP32[$0>>2] = $4;
 return;
}
function _lua_pushthread($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$1>>2] = $L;
 $2 = ((($1)) + 8|0);
 HEAP32[$2>>2] = 8;
 $3 = HEAP32[$0>>2]|0;
 $4 = ((($3)) + 16|0);
 HEAP32[$0>>2] = $4;
 $5 = ((($L)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($6)) + 112|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==($L|0);
 $10 = $9&1;
 return ($10|0);
}
function _lua_gettable($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($L)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ((($39)) + -16|0);
 _luaV_gettable($L,$$1$i,$40,$40);
 return;
}
function _lua_getfield($L,$idx,$k) {
 $L = $L|0;
 $idx = $idx|0;
 $k = $k|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $key = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $key = sp;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = (_strlen($k)|0);
 $39 = (_luaS_newlstr($L,$k,$38)|0);
 HEAP32[$key>>2] = $39;
 $40 = ((($key)) + 8|0);
 HEAP32[$40>>2] = 4;
 $41 = ((($L)) + 8|0);
 $42 = HEAP32[$41>>2]|0;
 _luaV_gettable($L,$$1$i,$key,$42);
 $43 = HEAP32[$41>>2]|0;
 $44 = ((($43)) + 16|0);
 HEAP32[$41>>2] = $44;
 STACKTOP = sp;return;
}
function _lua_rawget($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = HEAP32[$$1$i>>2]|0;
 $39 = ((($L)) + 8|0);
 $40 = HEAP32[$39>>2]|0;
 $41 = ((($40)) + -16|0);
 $42 = (_luaH_get($38,$41)|0);
 $43 = HEAP32[$39>>2]|0;
 $44 = ((($43)) + -16|0);
 $45 = $42;
 $46 = $45;
 $47 = HEAP32[$46>>2]|0;
 $48 = (($45) + 4)|0;
 $49 = $48;
 $50 = HEAP32[$49>>2]|0;
 $51 = $44;
 $52 = $51;
 HEAP32[$52>>2] = $47;
 $53 = (($51) + 4)|0;
 $54 = $53;
 HEAP32[$54>>2] = $50;
 $55 = ((($42)) + 8|0);
 $56 = HEAP32[$55>>2]|0;
 $57 = ((($43)) + -8|0);
 HEAP32[$57>>2] = $56;
 return;
}
function _lua_rawgeti($L,$idx,$n) {
 $L = $L|0;
 $idx = $idx|0;
 $n = $n|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = HEAP32[$$1$i>>2]|0;
 $39 = (_luaH_getnum($38,$n)|0);
 $40 = ((($L)) + 8|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = $39;
 $43 = $42;
 $44 = HEAP32[$43>>2]|0;
 $45 = (($42) + 4)|0;
 $46 = $45;
 $47 = HEAP32[$46>>2]|0;
 $48 = $41;
 $49 = $48;
 HEAP32[$49>>2] = $44;
 $50 = (($48) + 4)|0;
 $51 = $50;
 HEAP32[$51>>2] = $47;
 $52 = ((($39)) + 8|0);
 $53 = HEAP32[$52>>2]|0;
 $54 = ((($41)) + 8|0);
 HEAP32[$54>>2] = $53;
 $55 = HEAP32[$40>>2]|0;
 $56 = ((($55)) + 16|0);
 HEAP32[$40>>2] = $56;
 return;
}
function _lua_createtable($L,$narray,$nrec) {
 $L = $L|0;
 $narray = $narray|0;
 $nrec = $nrec|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 68|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 64|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($3>>>0)<($5>>>0);
 if (!($6)) {
  _luaC_step($L);
 }
 $7 = ((($L)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (_luaH_new($L,$narray,$nrec)|0);
 HEAP32[$8>>2] = $9;
 $10 = ((($8)) + 8|0);
 HEAP32[$10>>2] = 5;
 $11 = HEAP32[$7>>2]|0;
 $12 = ((($11)) + 16|0);
 HEAP32[$7>>2] = $12;
 return;
}
function _lua_getmetatable($L,$objindex) {
 $L = $L|0;
 $objindex = $objindex|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $mt$0 = 0, $mt$0$in = 0, $res$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($objindex|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($objindex) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($objindex|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($objindex<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($objindex|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($objindex))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 switch ($39|0) {
 case 5:  {
  $40 = HEAP32[$$1$i>>2]|0;
  $41 = ((($40)) + 8|0);
  $mt$0$in = $41;
  break;
 }
 case 7:  {
  $42 = HEAP32[$$1$i>>2]|0;
  $43 = ((($42)) + 8|0);
  $mt$0$in = $43;
  break;
 }
 default: {
  $44 = ((($L)) + 16|0);
  $45 = HEAP32[$44>>2]|0;
  $46 = (((($45)) + 152|0) + ($39<<2)|0);
  $mt$0$in = $46;
 }
 }
 $mt$0 = HEAP32[$mt$0$in>>2]|0;
 $47 = ($mt$0|0)==(0|0);
 if ($47) {
  $res$0 = 0;
  return ($res$0|0);
 }
 $48 = ((($L)) + 8|0);
 $49 = HEAP32[$48>>2]|0;
 HEAP32[$49>>2] = $mt$0;
 $50 = ((($49)) + 8|0);
 HEAP32[$50>>2] = 5;
 $51 = HEAP32[$48>>2]|0;
 $52 = ((($51)) + 16|0);
 HEAP32[$48>>2] = $52;
 $res$0 = 1;
 return ($res$0|0);
}
function _lua_getfenv($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 switch ($39|0) {
 case 6:  {
  $40 = ((($L)) + 8|0);
  $41 = HEAP32[$40>>2]|0;
  $42 = HEAP32[$$1$i>>2]|0;
  $43 = ((($42)) + 12|0);
  $44 = HEAP32[$43>>2]|0;
  HEAP32[$41>>2] = $44;
  $45 = ((($41)) + 8|0);
  HEAP32[$45>>2] = 5;
  $$pre$phiZ2D = $40;
  $72 = HEAP32[$$pre$phiZ2D>>2]|0;
  $73 = ((($72)) + 16|0);
  HEAP32[$$pre$phiZ2D>>2] = $73;
  return;
  break;
 }
 case 7:  {
  $46 = ((($L)) + 8|0);
  $47 = HEAP32[$46>>2]|0;
  $48 = HEAP32[$$1$i>>2]|0;
  $49 = ((($48)) + 12|0);
  $50 = HEAP32[$49>>2]|0;
  HEAP32[$47>>2] = $50;
  $51 = ((($47)) + 8|0);
  HEAP32[$51>>2] = 5;
  $$pre$phiZ2D = $46;
  $72 = HEAP32[$$pre$phiZ2D>>2]|0;
  $73 = ((($72)) + 16|0);
  HEAP32[$$pre$phiZ2D>>2] = $73;
  return;
  break;
 }
 case 8:  {
  $52 = HEAP32[$$1$i>>2]|0;
  $53 = ((($52)) + 72|0);
  $54 = ((($L)) + 8|0);
  $55 = HEAP32[$54>>2]|0;
  $56 = $53;
  $57 = $56;
  $58 = HEAP32[$57>>2]|0;
  $59 = (($56) + 4)|0;
  $60 = $59;
  $61 = HEAP32[$60>>2]|0;
  $62 = $55;
  $63 = $62;
  HEAP32[$63>>2] = $58;
  $64 = (($62) + 4)|0;
  $65 = $64;
  HEAP32[$65>>2] = $61;
  $66 = ((($52)) + 80|0);
  $67 = HEAP32[$66>>2]|0;
  $68 = ((($55)) + 8|0);
  HEAP32[$68>>2] = $67;
  $$pre$phiZ2D = $54;
  $72 = HEAP32[$$pre$phiZ2D>>2]|0;
  $73 = ((($72)) + 16|0);
  HEAP32[$$pre$phiZ2D>>2] = $73;
  return;
  break;
 }
 default: {
  $69 = ((($L)) + 8|0);
  $70 = HEAP32[$69>>2]|0;
  $71 = ((($70)) + 8|0);
  HEAP32[$71>>2] = 0;
  $$pre$phiZ2D = $69;
  $72 = HEAP32[$$pre$phiZ2D>>2]|0;
  $73 = ((($72)) + 16|0);
  HEAP32[$$pre$phiZ2D>>2] = $73;
  return;
 }
 }
}
function _lua_settable($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($L)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ((($39)) + -32|0);
 $41 = ((($39)) + -16|0);
 _luaV_settable($L,$$1$i,$40,$41);
 $42 = HEAP32[$38>>2]|0;
 $43 = ((($42)) + -32|0);
 HEAP32[$38>>2] = $43;
 return;
}
function _lua_setfield($L,$idx,$k) {
 $L = $L|0;
 $idx = $idx|0;
 $k = $k|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $key = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $key = sp;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = (_strlen($k)|0);
 $39 = (_luaS_newlstr($L,$k,$38)|0);
 HEAP32[$key>>2] = $39;
 $40 = ((($key)) + 8|0);
 HEAP32[$40>>2] = 4;
 $41 = ((($L)) + 8|0);
 $42 = HEAP32[$41>>2]|0;
 $43 = ((($42)) + -16|0);
 _luaV_settable($L,$$1$i,$key,$43);
 $44 = HEAP32[$41>>2]|0;
 $45 = ((($44)) + -16|0);
 HEAP32[$41>>2] = $45;
 STACKTOP = sp;return;
}
function _lua_rawset($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($L)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ((($39)) + -16|0);
 $41 = HEAP32[$$1$i>>2]|0;
 $42 = ((($39)) + -32|0);
 $43 = (_luaH_set($L,$41,$42)|0);
 $44 = $40;
 $45 = $44;
 $46 = HEAP32[$45>>2]|0;
 $47 = (($44) + 4)|0;
 $48 = $47;
 $49 = HEAP32[$48>>2]|0;
 $50 = $43;
 $51 = $50;
 HEAP32[$51>>2] = $46;
 $52 = (($50) + 4)|0;
 $53 = $52;
 HEAP32[$53>>2] = $49;
 $54 = ((($39)) + -8|0);
 $55 = HEAP32[$54>>2]|0;
 $56 = ((($43)) + 8|0);
 HEAP32[$56>>2] = $55;
 $57 = HEAP32[$38>>2]|0;
 $58 = ((($57)) + -8|0);
 $59 = HEAP32[$58>>2]|0;
 $60 = ($59|0)>(3);
 if (!($60)) {
  $73 = $57;
  $72 = ((($73)) + -32|0);
  HEAP32[$38>>2] = $72;
  return;
 }
 $61 = ((($57)) + -16|0);
 $62 = HEAP32[$61>>2]|0;
 $63 = ((($62)) + 5|0);
 $64 = HEAP8[$63>>0]|0;
 $65 = $64 & 3;
 $66 = ($65<<24>>24)==(0);
 if ($66) {
  $73 = $57;
  $72 = ((($73)) + -32|0);
  HEAP32[$38>>2] = $72;
  return;
 }
 $67 = HEAP32[$$1$i>>2]|0;
 $68 = ((($67)) + 5|0);
 $69 = HEAP8[$68>>0]|0;
 $70 = $69 & 4;
 $71 = ($70<<24>>24)==(0);
 if ($71) {
  $73 = $57;
  $72 = ((($73)) + -32|0);
  HEAP32[$38>>2] = $72;
  return;
 }
 _luaC_barrierback($L,$67);
 $$pre = HEAP32[$38>>2]|0;
 $73 = $$pre;
 $72 = ((($73)) + -32|0);
 HEAP32[$38>>2] = $72;
 return;
}
function _lua_rawseti($L,$idx,$n) {
 $L = $L|0;
 $idx = $idx|0;
 $n = $n|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($L)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ((($39)) + -16|0);
 $41 = HEAP32[$$1$i>>2]|0;
 $42 = (_luaH_setnum($L,$41,$n)|0);
 $43 = $40;
 $44 = $43;
 $45 = HEAP32[$44>>2]|0;
 $46 = (($43) + 4)|0;
 $47 = $46;
 $48 = HEAP32[$47>>2]|0;
 $49 = $42;
 $50 = $49;
 HEAP32[$50>>2] = $45;
 $51 = (($49) + 4)|0;
 $52 = $51;
 HEAP32[$52>>2] = $48;
 $53 = ((($39)) + -8|0);
 $54 = HEAP32[$53>>2]|0;
 $55 = ((($42)) + 8|0);
 HEAP32[$55>>2] = $54;
 $56 = HEAP32[$38>>2]|0;
 $57 = ((($56)) + -8|0);
 $58 = HEAP32[$57>>2]|0;
 $59 = ($58|0)>(3);
 if (!($59)) {
  $72 = $56;
  $71 = ((($72)) + -16|0);
  HEAP32[$38>>2] = $71;
  return;
 }
 $60 = ((($56)) + -16|0);
 $61 = HEAP32[$60>>2]|0;
 $62 = ((($61)) + 5|0);
 $63 = HEAP8[$62>>0]|0;
 $64 = $63 & 3;
 $65 = ($64<<24>>24)==(0);
 if ($65) {
  $72 = $56;
  $71 = ((($72)) + -16|0);
  HEAP32[$38>>2] = $71;
  return;
 }
 $66 = HEAP32[$$1$i>>2]|0;
 $67 = ((($66)) + 5|0);
 $68 = HEAP8[$67>>0]|0;
 $69 = $68 & 4;
 $70 = ($69<<24>>24)==(0);
 if ($70) {
  $72 = $56;
  $71 = ((($72)) + -16|0);
  HEAP32[$38>>2] = $71;
  return;
 }
 _luaC_barrierback($L,$66);
 $$pre = HEAP32[$38>>2]|0;
 $72 = $$pre;
 $71 = ((($72)) + -16|0);
 HEAP32[$38>>2] = $71;
 return;
}
function _lua_setmetatable($L,$objindex) {
 $L = $L|0;
 $objindex = $objindex|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $8 = 0, $9 = 0, $mt$0 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ($objindex|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($objindex) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($objindex|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($objindex<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($objindex|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($objindex))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($L)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ((($39)) + -8|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = ($41|0)==(0);
 if ($42) {
  $mt$0 = 0;
 } else {
  $43 = ((($39)) + -16|0);
  $44 = HEAP32[$43>>2]|0;
  $mt$0 = $44;
 }
 $45 = ((($$1$i)) + 8|0);
 $46 = HEAP32[$45>>2]|0;
 switch ($46|0) {
 case 5:  {
  $47 = HEAP32[$$1$i>>2]|0;
  $48 = ((($47)) + 8|0);
  HEAP32[$48>>2] = $mt$0;
  $49 = ($mt$0|0)==(0|0);
  if (!($49)) {
   $50 = ((($mt$0)) + 5|0);
   $51 = HEAP8[$50>>0]|0;
   $52 = $51 & 3;
   $53 = ($52<<24>>24)==(0);
   if (!($53)) {
    $54 = HEAP32[$$1$i>>2]|0;
    $55 = ((($54)) + 5|0);
    $56 = HEAP8[$55>>0]|0;
    $57 = $56 & 4;
    $58 = ($57<<24>>24)==(0);
    if (!($58)) {
     _luaC_barrierback($L,$54);
    }
   }
  }
  break;
 }
 case 7:  {
  $59 = HEAP32[$$1$i>>2]|0;
  $60 = ((($59)) + 8|0);
  HEAP32[$60>>2] = $mt$0;
  $61 = ($mt$0|0)==(0|0);
  if (!($61)) {
   $62 = ((($mt$0)) + 5|0);
   $63 = HEAP8[$62>>0]|0;
   $64 = $63 & 3;
   $65 = ($64<<24>>24)==(0);
   if (!($65)) {
    $66 = HEAP32[$$1$i>>2]|0;
    $67 = ((($66)) + 5|0);
    $68 = HEAP8[$67>>0]|0;
    $69 = $68 & 4;
    $70 = ($69<<24>>24)==(0);
    if (!($70)) {
     _luaC_barrierf($L,$66,$mt$0);
    }
   }
  }
  break;
 }
 default: {
  $71 = ((($L)) + 16|0);
  $72 = HEAP32[$71>>2]|0;
  $73 = (((($72)) + 152|0) + ($46<<2)|0);
  HEAP32[$73>>2] = $mt$0;
 }
 }
 $74 = HEAP32[$38>>2]|0;
 $75 = ((($74)) + -16|0);
 HEAP32[$38>>2] = $75;
 return 1;
}
function _lua_setfenv($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $$pre = 0, $$pre$phi4Z2D = 0, $$pre$phiZ2D = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0, $res$02 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 switch ($39|0) {
 case 6:  {
  $40 = ((($L)) + 8|0);
  $41 = HEAP32[$40>>2]|0;
  $42 = ((($41)) + -16|0);
  $43 = HEAP32[$42>>2]|0;
  $44 = HEAP32[$$1$i>>2]|0;
  $45 = ((($44)) + 12|0);
  HEAP32[$45>>2] = $43;
  $$pre$phiZ2D = $40;$60 = $41;
  break;
 }
 case 7:  {
  $46 = ((($L)) + 8|0);
  $47 = HEAP32[$46>>2]|0;
  $48 = ((($47)) + -16|0);
  $49 = HEAP32[$48>>2]|0;
  $50 = HEAP32[$$1$i>>2]|0;
  $51 = ((($50)) + 12|0);
  HEAP32[$51>>2] = $49;
  $$pre$phiZ2D = $46;$60 = $47;
  break;
 }
 case 8:  {
  $52 = HEAP32[$$1$i>>2]|0;
  $53 = ((($52)) + 72|0);
  $54 = ((($L)) + 8|0);
  $55 = HEAP32[$54>>2]|0;
  $56 = ((($55)) + -16|0);
  $57 = HEAP32[$56>>2]|0;
  HEAP32[$53>>2] = $57;
  $58 = ((($52)) + 80|0);
  HEAP32[$58>>2] = 5;
  $$pre = HEAP32[$54>>2]|0;
  $$pre$phiZ2D = $54;$60 = $$pre;
  break;
 }
 default: {
  $$pre3 = ((($L)) + 8|0);
  $$pre$phi4Z2D = $$pre3;$res$02 = 0;
  $71 = HEAP32[$$pre$phi4Z2D>>2]|0;
  $72 = ((($71)) + -16|0);
  HEAP32[$$pre$phi4Z2D>>2] = $72;
  return ($res$02|0);
 }
 }
 $59 = ((($60)) + -16|0);
 $61 = HEAP32[$59>>2]|0;
 $62 = ((($61)) + 5|0);
 $63 = HEAP8[$62>>0]|0;
 $64 = $63 & 3;
 $65 = ($64<<24>>24)==(0);
 if ($65) {
  $$pre$phi4Z2D = $$pre$phiZ2D;$res$02 = 1;
  $71 = HEAP32[$$pre$phi4Z2D>>2]|0;
  $72 = ((($71)) + -16|0);
  HEAP32[$$pre$phi4Z2D>>2] = $72;
  return ($res$02|0);
 }
 $66 = HEAP32[$$1$i>>2]|0;
 $67 = ((($66)) + 5|0);
 $68 = HEAP8[$67>>0]|0;
 $69 = $68 & 4;
 $70 = ($69<<24>>24)==(0);
 if ($70) {
  $$pre$phi4Z2D = $$pre$phiZ2D;$res$02 = 1;
  $71 = HEAP32[$$pre$phi4Z2D>>2]|0;
  $72 = ((($71)) + -16|0);
  HEAP32[$$pre$phi4Z2D>>2] = $72;
  return ($res$02|0);
 }
 _luaC_barrierf($L,$66,$61);
 $$pre$phi4Z2D = $$pre$phiZ2D;$res$02 = 1;
 $71 = HEAP32[$$pre$phi4Z2D>>2]|0;
 $72 = ((($71)) + -16|0);
 HEAP32[$$pre$phi4Z2D>>2] = $72;
 return ($res$02|0);
}
function _lua_call($L,$nargs,$nresults) {
 $L = $L|0;
 $nargs = $nargs|0;
 $nresults = $nresults|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $nargs ^ -1;
 $3 = (($1) + ($2<<4)|0);
 _luaD_call($L,$3,$nresults);
 $4 = ($nresults|0)==(-1);
 if (!($4)) {
  return;
 }
 $5 = HEAP32[$0>>2]|0;
 $6 = ((($L)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($7)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($5>>>0)<($9>>>0);
 if ($10) {
  return;
 }
 HEAP32[$8>>2] = $5;
 return;
}
function _lua_pcall($L,$nargs,$nresults,$errfunc) {
 $L = $L|0;
 $nargs = $nargs|0;
 $nresults = $nresults|0;
 $errfunc = $errfunc|0;
 var $$1$i = 0, $$cast = 0, $$luaO_nilobject_$i = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $c = 0, $func$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $c = sp;
 $0 = ($errfunc|0)==(0);
 if ($0) {
  $$phi$trans$insert = ((($L)) + 32|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $49 = $$pre;$func$0 = 0;
 } else {
  $1 = ($errfunc|0)>(0);
  L4: do {
   if ($1) {
    $2 = ((($L)) + 12|0);
    $3 = HEAP32[$2>>2]|0;
    $4 = (($errfunc) + -1)|0;
    $5 = (($3) + ($4<<4)|0);
    $6 = ((($L)) + 8|0);
    $7 = HEAP32[$6>>2]|0;
    $8 = ($5>>>0)<($7>>>0);
    $$luaO_nilobject_$i = $8 ? $5 : 624096;
    $$1$i = $$luaO_nilobject_$i;
   } else {
    $9 = ($errfunc|0)>(-10000);
    if ($9) {
     $10 = ((($L)) + 8|0);
     $11 = HEAP32[$10>>2]|0;
     $12 = (($11) + ($errfunc<<4)|0);
     $$1$i = $12;
     break;
    }
    switch ($errfunc|0) {
    case -10000:  {
     $13 = ((($L)) + 16|0);
     $14 = HEAP32[$13>>2]|0;
     $15 = ((($14)) + 96|0);
     $$1$i = $15;
     break L4;
     break;
    }
    case -10001:  {
     $16 = ((($L)) + 20|0);
     $17 = HEAP32[$16>>2]|0;
     $18 = ((($17)) + 4|0);
     $19 = HEAP32[$18>>2]|0;
     $20 = HEAP32[$19>>2]|0;
     $21 = ((($L)) + 88|0);
     $22 = ((($20)) + 12|0);
     $23 = HEAP32[$22>>2]|0;
     HEAP32[$21>>2] = $23;
     $24 = ((($L)) + 96|0);
     HEAP32[$24>>2] = 5;
     $$1$i = $21;
     break L4;
     break;
    }
    case -10002:  {
     $25 = ((($L)) + 72|0);
     $$1$i = $25;
     break L4;
     break;
    }
    default: {
     $26 = ((($L)) + 20|0);
     $27 = HEAP32[$26>>2]|0;
     $28 = ((($27)) + 4|0);
     $29 = HEAP32[$28>>2]|0;
     $30 = HEAP32[$29>>2]|0;
     $31 = (-10002 - ($errfunc))|0;
     $32 = ((($30)) + 7|0);
     $33 = HEAP8[$32>>0]|0;
     $34 = $33&255;
     $35 = ($31|0)>($34|0);
     $36 = (($31) + -1)|0;
     $37 = (((($30)) + 24|0) + ($36<<4)|0);
     $38 = $35 ? 624096 : $37;
     $$1$i = $38;
     break L4;
    }
    }
   }
  } while(0);
  $39 = ((($L)) + 32|0);
  $40 = HEAP32[$39>>2]|0;
  $41 = $$1$i;
  $42 = (($41) - ($40))|0;
  $49 = $40;$func$0 = $42;
 }
 $43 = ((($L)) + 8|0);
 $44 = HEAP32[$43>>2]|0;
 $45 = $nargs ^ -1;
 $46 = (($44) + ($45<<4)|0);
 HEAP32[$c>>2] = $46;
 $47 = ((($c)) + 4|0);
 HEAP32[$47>>2] = $nresults;
 $$cast = $46;
 $48 = (($$cast) - ($49))|0;
 $50 = (_luaD_pcall($L,4289,$c,$48,$func$0)|0);
 $51 = ($nresults|0)==(-1);
 if (!($51)) {
  STACKTOP = sp;return ($50|0);
 }
 $52 = HEAP32[$43>>2]|0;
 $53 = ((($L)) + 20|0);
 $54 = HEAP32[$53>>2]|0;
 $55 = ((($54)) + 8|0);
 $56 = HEAP32[$55>>2]|0;
 $57 = ($52>>>0)<($56>>>0);
 if ($57) {
  STACKTOP = sp;return ($50|0);
 }
 HEAP32[$55>>2] = $52;
 STACKTOP = sp;return ($50|0);
}
function _f_call($L,$ud) {
 $L = $L|0;
 $ud = $ud|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$ud>>2]|0;
 $1 = ((($ud)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 _luaD_call($L,$0,$2);
 return;
}
function _lua_load($L,$reader,$data,$chunkname) {
 $L = $L|0;
 $reader = $reader|0;
 $data = $data|0;
 $chunkname = $chunkname|0;
 var $$chunkname = 0, $0 = 0, $1 = 0, $z = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $z = sp;
 $0 = ($chunkname|0)==(0|0);
 $$chunkname = $0 ? 523335 : $chunkname;
 _luaZ_init($L,$z,$reader,$data);
 $1 = (_luaD_protectedparser($L,$z,$$chunkname)|0);
 STACKTOP = sp;return ($1|0);
}
function _lua_dump($L,$writer,$data) {
 $L = $L|0;
 $writer = $writer|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $status$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + -8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(6);
 if (!($4)) {
  $status$0 = 1;
  return ($status$0|0);
 }
 $5 = ((($1)) + -16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($6)) + 6|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = ($8<<24>>24)==(0);
 if (!($9)) {
  $status$0 = 1;
  return ($status$0|0);
 }
 $10 = ((($6)) + 16|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (_luaU_dump($L,$11,$writer,$data,0)|0);
 $status$0 = $12;
 return ($status$0|0);
}
function _lua_status($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 6|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = $1&255;
 return ($2|0);
}
function _lua_gc($L,$what,$data) {
 $L = $L|0;
 $what = $what|0;
 $data = $data|0;
 var $$ = 0, $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $res$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 L1: do {
  switch ($what|0) {
  case 0:  {
   $2 = ((($1)) + 64|0);
   HEAP32[$2>>2] = -3;
   $res$1 = 0;
   break;
  }
  case 1:  {
   $3 = ((($1)) + 68|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ((($1)) + 64|0);
   HEAP32[$5>>2] = $4;
   $res$1 = 0;
   break;
  }
  case 2:  {
   _luaC_fullgc($L);
   $res$1 = 0;
   break;
  }
  case 3:  {
   $6 = ((($1)) + 68|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = $7 >>> 10;
   $res$1 = $8;
   break;
  }
  case 4:  {
   $9 = ((($1)) + 68|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = $10 & 1023;
   $res$1 = $11;
   break;
  }
  case 5:  {
   $12 = $data << 10;
   $13 = ((($1)) + 68|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ($14>>>0)<($12>>>0);
   $16 = (($14) - ($12))|0;
   $$ = $15 ? 0 : $16;
   $17 = ((($1)) + 64|0);
   HEAP32[$17>>2] = $$;
   $18 = ((($1)) + 21|0);
   $19 = ($$>>>0)>($14>>>0);
   if ($19) {
    $res$1 = 0;
   } else {
    while(1) {
     _luaC_step($L);
     $20 = HEAP8[$18>>0]|0;
     $21 = ($20<<24>>24)==(0);
     if ($21) {
      $res$1 = 1;
      break L1;
     }
     $$pre = HEAP32[$17>>2]|0;
     $$pre1 = HEAP32[$13>>2]|0;
     $22 = ($$pre>>>0)>($$pre1>>>0);
     if ($22) {
      $res$1 = 0;
      break;
     }
    }
   }
   break;
  }
  case 6:  {
   $23 = ((($1)) + 80|0);
   $24 = HEAP32[$23>>2]|0;
   HEAP32[$23>>2] = $data;
   $res$1 = $24;
   break;
  }
  case 7:  {
   $25 = ((($1)) + 84|0);
   $26 = HEAP32[$25>>2]|0;
   HEAP32[$25>>2] = $data;
   $res$1 = $26;
   break;
  }
  default: {
   $res$1 = -1;
  }
  }
 } while(0);
 return ($res$1|0);
}
function _lua_error($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _luaG_errormsg($L);
 return 0;
}
function _lua_next($L,$idx) {
 $L = $L|0;
 $idx = $idx|0;
 var $$1$i = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($idx|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($idx) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($idx|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($idx<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($idx|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($idx))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = HEAP32[$$1$i>>2]|0;
 $39 = ((($L)) + 8|0);
 $40 = HEAP32[$39>>2]|0;
 $41 = ((($40)) + -16|0);
 $42 = (_luaH_next($L,$38,$41)|0);
 $43 = ($42|0)==(0);
 $44 = HEAP32[$39>>2]|0;
 $45 = ((($44)) + -16|0);
 $46 = ((($44)) + 16|0);
 $storemerge = $43 ? $45 : $46;
 HEAP32[$39>>2] = $storemerge;
 return ($42|0);
}
function _lua_concat($L,$n) {
 $L = $L|0;
 $n = $n|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)>(1);
 if ($0) {
  $1 = ((($L)) + 16|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ((($2)) + 68|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ((($2)) + 64|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($4>>>0)<($6>>>0);
  if (!($7)) {
   _luaC_step($L);
  }
  $8 = ((($L)) + 8|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($L)) + 12|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($9) - ($11))|0;
  $13 = $12 >> 4;
  $14 = (($13) + -1)|0;
  _luaV_concat($L,$n,$14);
  $15 = HEAP32[$8>>2]|0;
  $16 = (1 - ($n))|0;
  $17 = (($15) + ($16<<4)|0);
  HEAP32[$8>>2] = $17;
  return;
 } else {
  $18 = ($n|0)==(0);
  if (!($18)) {
   return;
  }
  $19 = ((($L)) + 8|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = (_luaS_newlstr($L,1282466,0)|0);
  HEAP32[$20>>2] = $21;
  $22 = ((($20)) + 8|0);
  HEAP32[$22>>2] = 4;
  $23 = HEAP32[$19>>2]|0;
  $24 = ((($23)) + 16|0);
  HEAP32[$19>>2] = $24;
  return;
 }
}
function _lua_newuserdata($L,$size) {
 $L = $L|0;
 $size = $size|0;
 var $$0$i = 0, $$0$in$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 68|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 64|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($3>>>0)<($5>>>0);
 if (!($6)) {
  _luaC_step($L);
 }
 $7 = ((($L)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($L)) + 40|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($8|0)==($10|0);
 if ($11) {
  $12 = ((($L)) + 72|0);
  $$0$in$i = $12;
 } else {
  $13 = ((($8)) + 4|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = HEAP32[$14>>2]|0;
  $16 = ((($15)) + 12|0);
  $$0$in$i = $16;
 }
 $$0$i = HEAP32[$$0$in$i>>2]|0;
 $17 = (_luaS_newudata($L,$size,$$0$i)|0);
 $18 = ((($L)) + 8|0);
 $19 = HEAP32[$18>>2]|0;
 HEAP32[$19>>2] = $17;
 $20 = ((($19)) + 8|0);
 HEAP32[$20>>2] = 7;
 $21 = HEAP32[$18>>2]|0;
 $22 = ((($21)) + 16|0);
 HEAP32[$18>>2] = $22;
 $23 = ((($17)) + 24|0);
 return ($23|0);
}
function _lua_getupvalue($L,$funcindex,$n) {
 $L = $L|0;
 $funcindex = $funcindex|0;
 $n = $n|0;
 var $$1$i = 0, $$1$i12 = 0, $$1$i15 = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $9 = 0, $or$cond$i = 0, $val$sroa$0$04 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($funcindex|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($funcindex) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($funcindex|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($funcindex<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($funcindex|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($funcindex))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ($39|0)==(6);
 if (!($40)) {
  $$1$i12 = 0;
  return ($$1$i12|0);
 }
 $41 = HEAP32[$$1$i>>2]|0;
 $42 = ((($41)) + 6|0);
 $43 = HEAP16[$42>>1]|0;
 $44 = $43&255;
 $45 = ($44<<24>>24)==(0);
 do {
  if ($45) {
   $53 = ((($41)) + 16|0);
   $54 = HEAP32[$53>>2]|0;
   $55 = ($n|0)>(0);
   if (!($55)) {
    $$1$i12 = 0;
    return ($$1$i12|0);
   }
   $56 = ((($54)) + 36|0);
   $57 = HEAP32[$56>>2]|0;
   $58 = ($57|0)<($n|0);
   if ($58) {
    $$1$i12 = 0;
    return ($$1$i12|0);
   } else {
    $59 = (($n) + -1)|0;
    $60 = (((($41)) + 20|0) + ($59<<2)|0);
    $61 = HEAP32[$60>>2]|0;
    $62 = ((($61)) + 8|0);
    $63 = HEAP32[$62>>2]|0;
    $64 = ((($54)) + 28|0);
    $65 = HEAP32[$64>>2]|0;
    $66 = (($65) + ($59<<2)|0);
    $67 = HEAP32[$66>>2]|0;
    $68 = ((($67)) + 16|0);
    $$1$i15 = $68;$val$sroa$0$04 = $63;
    break;
   }
  } else {
   $46 = ($43&65535) >>> 8;
   $47 = $46&65535;
   $48 = ($n|0)<(1);
   $49 = ($47|0)<($n|0);
   $or$cond$i = $48 | $49;
   if ($or$cond$i) {
    $$1$i12 = 0;
    return ($$1$i12|0);
   } else {
    $50 = (($n) + -1)|0;
    $51 = (((($41)) + 24|0) + ($50<<4)|0);
    $52 = $51;
    $$1$i15 = 1282466;$val$sroa$0$04 = $52;
    break;
   }
  }
 } while(0);
 $69 = $val$sroa$0$04;
 $70 = ((($L)) + 8|0);
 $71 = HEAP32[$70>>2]|0;
 $72 = $val$sroa$0$04;
 $73 = $72;
 $74 = $73;
 $75 = HEAP32[$74>>2]|0;
 $76 = (($73) + 4)|0;
 $77 = $76;
 $78 = HEAP32[$77>>2]|0;
 $79 = $71;
 $80 = $79;
 HEAP32[$80>>2] = $75;
 $81 = (($79) + 4)|0;
 $82 = $81;
 HEAP32[$82>>2] = $78;
 $83 = ((($69)) + 8|0);
 $84 = HEAP32[$83>>2]|0;
 $85 = ((($71)) + 8|0);
 HEAP32[$85>>2] = $84;
 $86 = HEAP32[$70>>2]|0;
 $87 = ((($86)) + 16|0);
 HEAP32[$70>>2] = $87;
 $$1$i12 = $$1$i15;
 return ($$1$i12|0);
}
function _lua_setupvalue($L,$funcindex,$n) {
 $L = $L|0;
 $funcindex = $funcindex|0;
 $n = $n|0;
 var $$1$i = 0, $$1$i12 = 0, $$1$i15 = 0, $$luaO_nilobject_$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond$i = 0, $val$sroa$0$04 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($funcindex|0)>(0);
 L1: do {
  if ($0) {
   $1 = ((($L)) + 12|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($funcindex) + -1)|0;
   $4 = (($2) + ($3<<4)|0);
   $5 = ((($L)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   $$luaO_nilobject_$i = $7 ? $4 : 624096;
   $$1$i = $$luaO_nilobject_$i;
  } else {
   $8 = ($funcindex|0)>(-10000);
   if ($8) {
    $9 = ((($L)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + ($funcindex<<4)|0);
    $$1$i = $11;
    break;
   }
   switch ($funcindex|0) {
   case -10000:  {
    $12 = ((($L)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + 96|0);
    $$1$i = $14;
    break L1;
    break;
   }
   case -10001:  {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($L)) + 88|0);
    $21 = ((($19)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$20>>2] = $22;
    $23 = ((($L)) + 96|0);
    HEAP32[$23>>2] = 5;
    $$1$i = $20;
    break L1;
    break;
   }
   case -10002:  {
    $24 = ((($L)) + 72|0);
    $$1$i = $24;
    break L1;
    break;
   }
   default: {
    $25 = ((($L)) + 20|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($26)) + 4|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = HEAP32[$28>>2]|0;
    $30 = (-10002 - ($funcindex))|0;
    $31 = ((($29)) + 7|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = ($30|0)>($33|0);
    $35 = (($30) + -1)|0;
    $36 = (((($29)) + 24|0) + ($35<<4)|0);
    $37 = $34 ? 624096 : $36;
    $$1$i = $37;
    break L1;
   }
   }
  }
 } while(0);
 $38 = ((($$1$i)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ($39|0)==(6);
 if (!($40)) {
  $$1$i12 = 0;
  return ($$1$i12|0);
 }
 $41 = HEAP32[$$1$i>>2]|0;
 $42 = ((($41)) + 6|0);
 $43 = HEAP16[$42>>1]|0;
 $44 = $43&255;
 $45 = ($44<<24>>24)==(0);
 do {
  if ($45) {
   $53 = ((($41)) + 16|0);
   $54 = HEAP32[$53>>2]|0;
   $55 = ($n|0)>(0);
   if (!($55)) {
    $$1$i12 = 0;
    return ($$1$i12|0);
   }
   $56 = ((($54)) + 36|0);
   $57 = HEAP32[$56>>2]|0;
   $58 = ($57|0)<($n|0);
   if ($58) {
    $$1$i12 = 0;
    return ($$1$i12|0);
   } else {
    $59 = (($n) + -1)|0;
    $60 = (((($41)) + 20|0) + ($59<<2)|0);
    $61 = HEAP32[$60>>2]|0;
    $62 = ((($61)) + 8|0);
    $63 = HEAP32[$62>>2]|0;
    $64 = ((($54)) + 28|0);
    $65 = HEAP32[$64>>2]|0;
    $66 = (($65) + ($59<<2)|0);
    $67 = HEAP32[$66>>2]|0;
    $68 = ((($67)) + 16|0);
    $$1$i15 = $68;$val$sroa$0$04 = $63;
    break;
   }
  } else {
   $46 = ($43&65535) >>> 8;
   $47 = $46&65535;
   $48 = ($n|0)<(1);
   $49 = ($47|0)<($n|0);
   $or$cond$i = $48 | $49;
   if ($or$cond$i) {
    $$1$i12 = 0;
    return ($$1$i12|0);
   } else {
    $50 = (($n) + -1)|0;
    $51 = (((($41)) + 24|0) + ($50<<4)|0);
    $52 = $51;
    $$1$i15 = 1282466;$val$sroa$0$04 = $52;
    break;
   }
  }
 } while(0);
 $69 = ((($L)) + 8|0);
 $70 = HEAP32[$69>>2]|0;
 $71 = ((($70)) + -16|0);
 HEAP32[$69>>2] = $71;
 $72 = $val$sroa$0$04;
 $73 = $val$sroa$0$04;
 $74 = $71;
 $75 = $74;
 $76 = HEAP32[$75>>2]|0;
 $77 = (($74) + 4)|0;
 $78 = $77;
 $79 = HEAP32[$78>>2]|0;
 $80 = $73;
 $81 = $80;
 HEAP32[$81>>2] = $76;
 $82 = (($80) + 4)|0;
 $83 = $82;
 HEAP32[$83>>2] = $79;
 $84 = ((($70)) + -8|0);
 $85 = HEAP32[$84>>2]|0;
 $86 = ((($72)) + 8|0);
 HEAP32[$86>>2] = $85;
 $87 = HEAP32[$69>>2]|0;
 $88 = ((($87)) + 8|0);
 $89 = HEAP32[$88>>2]|0;
 $90 = ($89|0)>(3);
 if (!($90)) {
  $$1$i12 = $$1$i15;
  return ($$1$i12|0);
 }
 $91 = HEAP32[$87>>2]|0;
 $92 = ((($91)) + 5|0);
 $93 = HEAP8[$92>>0]|0;
 $94 = $93 & 3;
 $95 = ($94<<24>>24)==(0);
 if ($95) {
  $$1$i12 = $$1$i15;
  return ($$1$i12|0);
 }
 $96 = HEAP32[$$1$i>>2]|0;
 $97 = ((($96)) + 5|0);
 $98 = HEAP8[$97>>0]|0;
 $99 = $98 & 4;
 $100 = ($99<<24>>24)==(0);
 if ($100) {
  $$1$i12 = $$1$i15;
  return ($$1$i12|0);
 }
 _luaC_barrierf($L,$96,$91);
 $$1$i12 = $$1$i15;
 return ($$1$i12|0);
}
function _lua_sethook($L,$func,$mask,$count) {
 $L = $L|0;
 $func = $func|0;
 $mask = $mask|0;
 $count = $count|0;
 var $$func = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($func|0)==(0|0);
 $1 = ($mask|0)==(0);
 $or$cond = $0 | $1;
 $$func = $or$cond ? 0 : $func;
 $2 = ((($L)) + 68|0);
 HEAP32[$2>>2] = $$func;
 $3 = ((($L)) + 60|0);
 HEAP32[$3>>2] = $count;
 $4 = ((($L)) + 64|0);
 HEAP32[$4>>2] = $count;
 $5 = $mask&255;
 $6 = $or$cond ? 0 : $5;
 $7 = ((($L)) + 56|0);
 HEAP8[$7>>0] = $6;
 return 1;
}
function _lua_gethook($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 68|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function _lua_gethookmask($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 56|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = $1&255;
 return ($2|0);
}
function _lua_gethookcount($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 60|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function _lua_getstack($L,$level,$ar) {
 $L = $L|0;
 $level = $level|0;
 $ar = $ar|0;
 var $$0$lcssa = 0, $$03 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ci$0$lcssa = 0, $ci$04 = 0, $status$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($level|0)>(0);
 L1: do {
  if ($2) {
   $3 = ((($L)) + 40|0);
   $4 = HEAP32[$3>>2]|0;
   $$03 = $level;$ci$04 = $1;
   while(1) {
    $5 = ($ci$04>>>0)>($4>>>0);
    if (!($5)) {
     $status$0 = 0;
     break;
    }
    $6 = (($$03) + -1)|0;
    $7 = ((($ci$04)) + 4|0);
    $8 = HEAP32[$7>>2]|0;
    $9 = HEAP32[$8>>2]|0;
    $10 = ((($9)) + 6|0);
    $11 = HEAP8[$10>>0]|0;
    $12 = ($11<<24>>24)==(0);
    if ($12) {
     $13 = ((($ci$04)) + 20|0);
     $14 = HEAP32[$13>>2]|0;
     $15 = (($6) - ($14))|0;
     $$1 = $15;
    } else {
     $$1 = $6;
    }
    $16 = ((($ci$04)) + -24|0);
    $17 = ($$1|0)>(0);
    if ($17) {
     $$03 = $$1;$ci$04 = $16;
    } else {
     $$0$lcssa = $$1;$ci$0$lcssa = $16;
     break L1;
    }
   }
   return ($status$0|0);
  } else {
   $$0$lcssa = $level;$ci$0$lcssa = $1;
  }
 } while(0);
 $18 = ($$0$lcssa|0)==(0);
 if (!($18)) {
  $27 = ((($ar)) + 96|0);
  HEAP32[$27>>2] = 0;
  $status$0 = 1;
  return ($status$0|0);
 }
 $19 = ((($L)) + 40|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ($ci$0$lcssa>>>0)>($20>>>0);
 if (!($21)) {
  $status$0 = 0;
  return ($status$0|0);
 }
 $22 = $ci$0$lcssa;
 $23 = $20;
 $24 = (($22) - ($23))|0;
 $25 = (($24|0) / 24)&-1;
 $26 = ((($ar)) + 96|0);
 HEAP32[$26>>2] = $25;
 $status$0 = 1;
 return ($status$0|0);
}
function _lua_getlocal($L,$ar,$n) {
 $L = $L|0;
 $ar = $ar|0;
 $n = $n|0;
 var $$1$i$ph = 0, $$1$i2 = 0, $$in$i = 0, $$phi$trans$insert$i$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ci$idx$i = 0, $ci$idx$val$i = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 40|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($ar)) + 96|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + (($3*24)|0)|0);
 $ci$idx$i = (((($1) + (($3*24)|0)|0)) + 4|0);
 $ci$idx$val$i = HEAP32[$ci$idx$i>>2]|0;
 $5 = ((($ci$idx$val$i)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(6);
 if ($7) {
  $8 = HEAP32[$ci$idx$val$i>>2]|0;
  $9 = ((($8)) + 6|0);
  $10 = HEAP8[$9>>0]|0;
  $11 = ($10<<24>>24)==(0);
  if ($11) {
   $12 = ((($8)) + 16|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = ($13|0)==(0|0);
   if ($14) {
    label = 9;
   } else {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ($16|0)==($4|0);
    if ($17) {
     $18 = ((($L)) + 24|0);
     $19 = HEAP32[$18>>2]|0;
     $20 = (((($1) + (($3*24)|0)|0)) + 12|0);
     HEAP32[$20>>2] = $19;
     $$pre$i = HEAP32[$ci$idx$val$i>>2]|0;
     $22 = $$pre$i;$27 = $19;
    } else {
     $$phi$trans$insert$i$i = (((($1) + (($3*24)|0)|0)) + 12|0);
     $$pre$i$i = HEAP32[$$phi$trans$insert$i$i>>2]|0;
     $22 = $8;$27 = $$pre$i$i;
    }
    $21 = ((($22)) + 16|0);
    $23 = HEAP32[$21>>2]|0;
    $24 = ((($23)) + 12|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = (($27) - ($25))|0;
    $28 = $26 >> 2;
    $29 = (($28) + -1)|0;
    $30 = (_luaF_getlocalname($13,$n,$29)|0);
    $31 = ($30|0)==(0|0);
    if ($31) {
     label = 9;
    } else {
     $$pre = HEAP32[$4>>2]|0;
     $$1$i$ph = $30;$46 = $$pre;
    }
   }
  } else {
   label = 9;
  }
 } else {
  label = 9;
 }
 do {
  if ((label|0) == 9) {
   $32 = ((($L)) + 20|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ($33|0)==($4|0);
   $35 = ((($L)) + 8|0);
   $36 = ((($4)) + 28|0);
   $$in$i = $34 ? $35 : $36;
   $37 = HEAP32[$$in$i>>2]|0;
   $38 = HEAP32[$4>>2]|0;
   $39 = (($37) - ($38))|0;
   $40 = $39 >> 4;
   $41 = ($40|0)>=($n|0);
   $42 = ($n|0)>(0);
   $or$cond$i = $42 & $41;
   if ($or$cond$i) {
    $43 = $38;
    $$1$i$ph = 517568;$46 = $43;
    break;
   } else {
    $$1$i2 = 0;
    return ($$1$i2|0);
   }
  }
 } while(0);
 $44 = (($n) + -1)|0;
 $45 = (($46) + ($44<<4)|0);
 _luaA_pushobject($L,$45);
 $$1$i2 = $$1$i$ph;
 return ($$1$i2|0);
}
function _lua_setlocal($L,$ar,$n) {
 $L = $L|0;
 $ar = $ar|0;
 $n = $n|0;
 var $$1$i$ph = 0, $$1$i2 = 0, $$in$i = 0, $$phi$trans$insert$i$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$phi4Z2D = 0, $$pre$phiZ2D = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, $ci$idx$i = 0, $ci$idx$val$i = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 40|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($ar)) + 96|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + (($3*24)|0)|0);
 $ci$idx$i = (((($1) + (($3*24)|0)|0)) + 4|0);
 $ci$idx$val$i = HEAP32[$ci$idx$i>>2]|0;
 $5 = ((($ci$idx$val$i)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(6);
 if ($7) {
  $8 = HEAP32[$ci$idx$val$i>>2]|0;
  $9 = ((($8)) + 6|0);
  $10 = HEAP8[$9>>0]|0;
  $11 = ($10<<24>>24)==(0);
  if ($11) {
   $12 = ((($8)) + 16|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = ($13|0)==(0|0);
   if ($14) {
    label = 9;
   } else {
    $15 = ((($L)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ($16|0)==($4|0);
    if ($17) {
     $18 = ((($L)) + 24|0);
     $19 = HEAP32[$18>>2]|0;
     $20 = (((($1) + (($3*24)|0)|0)) + 12|0);
     HEAP32[$20>>2] = $19;
     $$pre$i = HEAP32[$ci$idx$val$i>>2]|0;
     $22 = $$pre$i;$27 = $19;
    } else {
     $$phi$trans$insert$i$i = (((($1) + (($3*24)|0)|0)) + 12|0);
     $$pre$i$i = HEAP32[$$phi$trans$insert$i$i>>2]|0;
     $22 = $8;$27 = $$pre$i$i;
    }
    $21 = ((($22)) + 16|0);
    $23 = HEAP32[$21>>2]|0;
    $24 = ((($23)) + 12|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = (($27) - ($25))|0;
    $28 = $26 >> 2;
    $29 = (($28) + -1)|0;
    $30 = (_luaF_getlocalname($13,$n,$29)|0);
    $31 = ($30|0)==(0|0);
    if ($31) {
     label = 9;
    } else {
     $$pre = HEAP32[$4>>2]|0;
     $$pre3 = ((($L)) + 8|0);
     $$1$i$ph = $30;$$pre$phiZ2D = $$pre3;$48 = $$pre;
    }
   }
  } else {
   label = 9;
  }
 } else {
  label = 9;
 }
 do {
  if ((label|0) == 9) {
   $32 = ((($L)) + 20|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ($33|0)==($4|0);
   $35 = ((($L)) + 8|0);
   $36 = ((($4)) + 28|0);
   $$in$i = $34 ? $35 : $36;
   $37 = HEAP32[$$in$i>>2]|0;
   $38 = HEAP32[$4>>2]|0;
   $39 = (($37) - ($38))|0;
   $40 = $39 >> 4;
   $41 = ($40|0)>=($n|0);
   $42 = ($n|0)>(0);
   $or$cond$i = $42 & $41;
   if ($or$cond$i) {
    $43 = $38;
    $$1$i$ph = 517568;$$pre$phiZ2D = $35;$48 = $43;
    break;
   } else {
    $$1$i2 = 0;$$pre$phi4Z2D = $35;
    $62 = HEAP32[$$pre$phi4Z2D>>2]|0;
    $63 = ((($62)) + -16|0);
    HEAP32[$$pre$phi4Z2D>>2] = $63;
    return ($$1$i2|0);
   }
  }
 } while(0);
 $44 = HEAP32[$$pre$phiZ2D>>2]|0;
 $45 = ((($44)) + -16|0);
 $46 = (($n) + -1)|0;
 $47 = (($48) + ($46<<4)|0);
 $49 = $45;
 $50 = $49;
 $51 = HEAP32[$50>>2]|0;
 $52 = (($49) + 4)|0;
 $53 = $52;
 $54 = HEAP32[$53>>2]|0;
 $55 = $47;
 $56 = $55;
 HEAP32[$56>>2] = $51;
 $57 = (($55) + 4)|0;
 $58 = $57;
 HEAP32[$58>>2] = $54;
 $59 = ((($44)) + -8|0);
 $60 = HEAP32[$59>>2]|0;
 $61 = (((($48) + ($46<<4)|0)) + 8|0);
 HEAP32[$61>>2] = $60;
 $$1$i2 = $$1$i$ph;$$pre$phi4Z2D = $$pre$phiZ2D;
 $62 = HEAP32[$$pre$phi4Z2D>>2]|0;
 $63 = ((($62)) + -16|0);
 HEAP32[$$pre$phi4Z2D>>2] = $63;
 return ($$1$i2|0);
}
function _lua_getinfo($L,$what,$ar) {
 $L = $L|0;
 $what = $what|0;
 $ar = $ar|0;
 var $$0 = 0, $$0$i = 0, $$019$i = 0, $$04 = 0, $$05 = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i$i = 0, $$pre$i$i4$i = 0, $$pre$i6$i = 0, $$pre$phi$iZ2D = 0, $$pre3$i = 0, $$sink = 0, $$sink$i$i = 0, $$sink1$i$i = 0, $$sink2$i$i = 0, $0 = 0, $1 = 0, $10 = 0;
 var $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0;
 var $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0;
 var $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0;
 var $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0;
 var $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $ci$0 = 0, $f$0 = 0, $f$02 = 0, $i$01$i = 0, $status$010$i = 0, $status$1$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$what>>0]|0;
 $1 = ($0<<24>>24)==(62);
 if ($1) {
  $2 = ((($L)) + 8|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($3)) + -16|0);
  $5 = ((($what)) + 1|0);
  $6 = HEAP32[$4>>2]|0;
  HEAP32[$2>>2] = $4;
  $$0 = $5;$ci$0 = 0;$f$0 = $6;
  label = 5;
 } else {
  $7 = ((($ar)) + 96|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8|0)==(0);
  if ($9) {
   $$05 = $what;
   label = 8;
  } else {
   $10 = ((($L)) + 40|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = (($11) + (($8*24)|0)|0);
   $13 = (((($11) + (($8*24)|0)|0)) + 4|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = HEAP32[$14>>2]|0;
   $$0 = $what;$ci$0 = $12;$f$0 = $15;
   label = 5;
  }
 }
 L5: do {
  if ((label|0) == 5) {
   $16 = ($f$0|0)==(0|0);
   if ($16) {
    $$05 = $$0;
    label = 8;
   } else {
    $17 = HEAP8[$$0>>0]|0;
    $18 = ($17<<24>>24)==(0);
    if ($18) {
     $$0$i = 1;$$04 = $$0;$184 = 0;$f$02 = $f$0;
    } else {
     $19 = ((($f$0)) + 6|0);
     $20 = ((($f$0)) + 16|0);
     $21 = ((($ar)) + 16|0);
     $22 = ((($ar)) + 28|0);
     $23 = ((($ar)) + 32|0);
     $24 = ((($ar)) + 12|0);
     $25 = ((($ar)) + 36|0);
     $26 = ($ci$0|0)==(0|0);
     $27 = ((($ar)) + 20|0);
     $28 = ((($ci$0)) + 4|0);
     $29 = ((($L)) + 20|0);
     $30 = ((($L)) + 24|0);
     $31 = ((($ci$0)) + 12|0);
     $32 = ((($f$0)) + 7|0);
     $33 = ((($ar)) + 24|0);
     $34 = ((($ar)) + 8|0);
     $35 = ((($ar)) + 4|0);
     $36 = ((($ci$0)) + 20|0);
     $37 = ((($ci$0)) + -24|0);
     $38 = ((($ci$0)) + -20|0);
     $39 = ((($ci$0)) + -12|0);
     $$019$i = $$0;$50 = $17;$status$010$i = 1;
     while(1) {
      $49 = $50 << 24 >> 24;
      L11: do {
       switch ($49|0) {
       case 83:  {
        $51 = HEAP8[$19>>0]|0;
        $52 = ($51<<24>>24)==(0);
        if ($52) {
         $53 = HEAP32[$20>>2]|0;
         $54 = ((($53)) + 32|0);
         $55 = HEAP32[$54>>2]|0;
         $56 = ((($55)) + 16|0);
         HEAP32[$21>>2] = $56;
         $57 = ((($53)) + 60|0);
         $58 = HEAP32[$57>>2]|0;
         $59 = ((($53)) + 64|0);
         $60 = HEAP32[$59>>2]|0;
         $61 = ($58|0)==(0);
         $62 = $61 ? 517604 : 517609;
         $$sink$i$i = $62;$$sink1$i$i = $60;$$sink2$i$i = $58;$63 = $56;
        } else {
         HEAP32[$21>>2] = 517599;
         $$sink$i$i = 620519;$$sink1$i$i = -1;$$sink2$i$i = -1;$63 = 517599;
        }
        HEAP32[$22>>2] = $$sink2$i$i;
        HEAP32[$23>>2] = $$sink1$i$i;
        HEAP32[$24>>2] = $$sink$i$i;
        _luaO_chunkid($25,$63,60);
        $status$1$i = $status$010$i;
        break;
       }
       case 108:  {
        if ($26) {
         $90 = -1;
        } else {
         $64 = HEAP32[$28>>2]|0;
         $65 = ((($64)) + 8|0);
         $66 = HEAP32[$65>>2]|0;
         $67 = ($66|0)==(6);
         if ($67) {
          $68 = HEAP32[$64>>2]|0;
          $69 = ((($68)) + 6|0);
          $70 = HEAP8[$69>>0]|0;
          $71 = ($70<<24>>24)==(0);
          if ($71) {
           $72 = HEAP32[$29>>2]|0;
           $73 = ($72|0)==($ci$0|0);
           if ($73) {
            $74 = HEAP32[$30>>2]|0;
            HEAP32[$31>>2] = $74;
            $$pre$i$i = HEAP32[$64>>2]|0;
            $76 = $$pre$i$i;$81 = $74;
           } else {
            $$pre$i$i$i = HEAP32[$31>>2]|0;
            $76 = $68;$81 = $$pre$i$i$i;
           }
           $75 = ((($76)) + 16|0);
           $77 = HEAP32[$75>>2]|0;
           $78 = ((($77)) + 12|0);
           $79 = HEAP32[$78>>2]|0;
           $80 = (($81) - ($79))|0;
           $82 = $80 >> 2;
           $83 = (($82) + -1)|0;
           $84 = ($80|0)<(4);
           if ($84) {
            $90 = -1;
           } else {
            $85 = ((($77)) + 20|0);
            $86 = HEAP32[$85>>2]|0;
            $87 = ($86|0)==(0|0);
            if ($87) {
             $90 = 0;
            } else {
             $88 = (($86) + ($83<<2)|0);
             $89 = HEAP32[$88>>2]|0;
             $90 = $89;
            }
           }
          } else {
           $90 = -1;
          }
         } else {
          $90 = -1;
         }
        }
        HEAP32[$27>>2] = $90;
        $status$1$i = $status$010$i;
        break;
       }
       case 117:  {
        $91 = HEAP8[$32>>0]|0;
        $92 = $91&255;
        HEAP32[$33>>2] = $92;
        $status$1$i = $status$010$i;
        break;
       }
       case 110:  {
        L32: do {
         if ($26) {
          label = 35;
         } else {
          $93 = HEAP32[$28>>2]|0;
          $94 = ((($93)) + 8|0);
          $95 = HEAP32[$94>>2]|0;
          $96 = ($95|0)==(6);
          if ($96) {
           $97 = HEAP32[$93>>2]|0;
           $98 = ((($97)) + 6|0);
           $99 = HEAP8[$98>>0]|0;
           $100 = ($99<<24>>24)==(0);
           if ($100) {
            $101 = HEAP32[$36>>2]|0;
            $102 = ($101|0)>(0);
            if ($102) {
             label = 35;
             break;
            }
           }
          }
          $103 = HEAP32[$38>>2]|0;
          $104 = ((($103)) + 8|0);
          $105 = HEAP32[$104>>2]|0;
          $106 = ($105|0)==(6);
          if ($106) {
           $107 = HEAP32[$103>>2]|0;
           $108 = ((($107)) + 6|0);
           $109 = HEAP8[$108>>0]|0;
           $110 = ($109<<24>>24)==(0);
           if ($110) {
            $111 = HEAP32[$29>>2]|0;
            $112 = ($111|0)==($37|0);
            if ($112) {
             $113 = HEAP32[$30>>2]|0;
             HEAP32[$39>>2] = $113;
             $$pre$i6$i = HEAP32[$103>>2]|0;
             $115 = $$pre$i6$i;$120 = $113;
            } else {
             $$pre$i$i4$i = HEAP32[$39>>2]|0;
             $115 = $107;$120 = $$pre$i$i4$i;
            }
            $114 = ((($115)) + 16|0);
            $116 = HEAP32[$114>>2]|0;
            $117 = ((($116)) + 12|0);
            $118 = HEAP32[$117>>2]|0;
            $119 = (($120) - ($118))|0;
            $121 = $119 >> 2;
            $122 = (($121) + -1)|0;
            $123 = $118;
            $124 = (($123) + ($122<<2)|0);
            $125 = HEAP32[$124>>2]|0;
            $126 = $125 & 63;
            switch ($126|0) {
            case 33: case 29: case 28:  {
             break;
            }
            default: {
             label = 35;
             break L32;
            }
            }
            $127 = $125 >>> 6;
            $128 = $127 & 255;
            $129 = (_getobjname($L,$37,$128,$35)|0);
            HEAP32[$34>>2] = $129;
            $130 = ($129|0)==(0|0);
            if (!($130)) {
             $status$1$i = $status$010$i;
             break L11;
            }
           } else {
            label = 35;
           }
          } else {
           label = 35;
          }
         }
        } while(0);
        if ((label|0) == 35) {
         label = 0;
         HEAP32[$34>>2] = 0;
        }
        HEAP32[$34>>2] = 1282466;
        HEAP32[$35>>2] = 0;
        $status$1$i = $status$010$i;
        break;
       }
       case 102: case 76:  {
        $status$1$i = $status$010$i;
        break;
       }
       default: {
        $status$1$i = 0;
       }
       }
      } while(0);
      $131 = ((($$019$i)) + 1|0);
      $132 = HEAP8[$131>>0]|0;
      $133 = ($132<<24>>24)==(0);
      if ($133) {
       $$0$i = $status$1$i;$$04 = $$0;$184 = 0;$f$02 = $f$0;
       break L5;
      } else {
       $$019$i = $131;$50 = $132;$status$010$i = $status$1$i;
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $40 = ((($ar)) + 8|0);
  HEAP32[$40>>2] = 1282466;
  $41 = ((($ar)) + 4|0);
  HEAP32[$41>>2] = 1282466;
  $42 = ((($ar)) + 12|0);
  HEAP32[$42>>2] = 517581;
  $43 = ((($ar)) + 20|0);
  HEAP32[$43>>2] = -1;
  $44 = ((($ar)) + 28|0);
  HEAP32[$44>>2] = -1;
  $45 = ((($ar)) + 32|0);
  HEAP32[$45>>2] = -1;
  $46 = ((($ar)) + 16|0);
  HEAP32[$46>>2] = 517586;
  $47 = ((($ar)) + 36|0);
  _luaO_chunkid($47,517586,60);
  $48 = ((($ar)) + 24|0);
  HEAP32[$48>>2] = 0;
  $$0$i = 1;$$04 = $$05;$184 = 1;$f$02 = 0;
 }
 $134 = (_strchr($$04,102)|0);
 $135 = ($134|0)==(0|0);
 if (!($135)) {
  $136 = ((($L)) + 8|0);
  $137 = HEAP32[$136>>2]|0;
  if ($184) {
   $$sink = 0;
  } else {
   HEAP32[$137>>2] = $f$02;
   $$sink = 6;
  }
  $138 = ((($137)) + 8|0);
  HEAP32[$138>>2] = $$sink;
  $139 = ((($L)) + 28|0);
  $140 = HEAP32[$139>>2]|0;
  $141 = HEAP32[$136>>2]|0;
  $142 = (($140) - ($141))|0;
  $143 = ($142|0)<(17);
  $144 = $141;
  if ($143) {
   _luaD_growstack($L,1);
   $$pre = HEAP32[$136>>2]|0;
   $146 = $$pre;
  } else {
   $146 = $144;
  }
  $145 = ((($146)) + 16|0);
  HEAP32[$136>>2] = $145;
 }
 $147 = (_strchr($$04,76)|0);
 $148 = ($147|0)==(0|0);
 if ($148) {
  return ($$0$i|0);
 }
 if ($184) {
  label = 49;
 } else {
  $149 = ((($f$02)) + 6|0);
  $150 = HEAP8[$149>>0]|0;
  $151 = ($150<<24>>24)==(0);
  if ($151) {
   $156 = (_luaH_new($L,0,0)|0);
   $157 = ((($f$02)) + 16|0);
   $158 = HEAP32[$157>>2]|0;
   $159 = ((($158)) + 20|0);
   $160 = HEAP32[$159>>2]|0;
   $161 = ((($158)) + 48|0);
   $162 = HEAP32[$161>>2]|0;
   $163 = ($162|0)>(0);
   if ($163) {
    $i$01$i = 0;
    while(1) {
     $164 = (($160) + ($i$01$i<<2)|0);
     $165 = HEAP32[$164>>2]|0;
     $166 = (_luaH_setnum($L,$156,$165)|0);
     HEAP32[$166>>2] = 1;
     $167 = ((($166)) + 8|0);
     HEAP32[$167>>2] = 1;
     $168 = (($i$01$i) + 1)|0;
     $169 = HEAP32[$157>>2]|0;
     $170 = ((($169)) + 48|0);
     $171 = HEAP32[$170>>2]|0;
     $172 = ($168|0)<($171|0);
     if ($172) {
      $i$01$i = $168;
     } else {
      break;
     }
    }
   }
   $173 = ((($L)) + 8|0);
   $174 = HEAP32[$173>>2]|0;
   HEAP32[$174>>2] = $156;
   $175 = ((($174)) + 8|0);
   HEAP32[$175>>2] = 5;
   $$pre$i = HEAP32[$173>>2]|0;
   $176 = $$pre$i;
   $$pre$phi$iZ2D = $173;$180 = $$pre$i;$185 = $176;
  } else {
   label = 49;
  }
 }
 if ((label|0) == 49) {
  $152 = ((($L)) + 8|0);
  $153 = HEAP32[$152>>2]|0;
  $154 = ((($153)) + 8|0);
  HEAP32[$154>>2] = 0;
  $155 = $153;
  $$pre$phi$iZ2D = $152;$180 = $155;$185 = $153;
 }
 $177 = ((($L)) + 28|0);
 $178 = HEAP32[$177>>2]|0;
 $179 = (($178) - ($180))|0;
 $181 = ($179|0)<(17);
 if ($181) {
  _luaD_growstack($L,1);
  $$pre3$i = HEAP32[$$pre$phi$iZ2D>>2]|0;
  $183 = $$pre3$i;
 } else {
  $183 = $185;
 }
 $182 = ((($183)) + 16|0);
 HEAP32[$$pre$phi$iZ2D>>2] = $182;
 return ($$0$i|0);
}
function _getobjname($L,$ci,$stackpos,$name) {
 $L = $L|0;
 $ci = $ci|0;
 $stackpos = $stackpos|0;
 $name = $name|0;
 var $$0$i = 0, $$0$i3 = 0, $$0$i4 = 0, $$3 = 0, $$lcssa = 0, $$lcssa44 = 0, $$lcssa45 = 0, $$lcssa46 = 0, $$lcssa47 = 0, $$lcssa48 = 0, $$lcssa49 = 0, $$lcssa50 = 0, $$pre = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $8 = 0, $9 = 0, $stackpos$tr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($ci)) + 4|0);
 $1 = ((($L)) + 20|0);
 $2 = ((($L)) + 24|0);
 $3 = ((($ci)) + 12|0);
 $stackpos$tr = $stackpos;
 L1: while(1) {
  $4 = HEAP32[$0>>2]|0;
  $5 = ((($4)) + 8|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6|0)==(6);
  if (!($7)) {
   $$3 = 0;
   label = 24;
   break;
  }
  $8 = HEAP32[$4>>2]|0;
  $9 = ((($8)) + 6|0);
  $10 = HEAP8[$9>>0]|0;
  $11 = ($10<<24>>24)==(0);
  if (!($11)) {
   $$3 = 0;
   label = 24;
   break;
  }
  $12 = ((($8)) + 16|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ((($8)) + 6|0);
  $15 = HEAP8[$14>>0]|0;
  $16 = ($15<<24>>24)==(0);
  if ($16) {
   $17 = HEAP32[$1>>2]|0;
   $18 = ($17|0)==($ci|0);
   if ($18) {
    $19 = HEAP32[$2>>2]|0;
    HEAP32[$3>>2] = $19;
    $$pre = HEAP32[$4>>2]|0;
    $21 = $$pre;$26 = $19;
   } else {
    $$pre$i = HEAP32[$3>>2]|0;
    $21 = $8;$26 = $$pre$i;
   }
   $20 = ((($21)) + 16|0);
   $22 = HEAP32[$20>>2]|0;
   $23 = ((($22)) + 12|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = (($26) - ($24))|0;
   $27 = $25 >> 2;
   $28 = (($27) + -1)|0;
   $$0$i = $28;
  } else {
   $$0$i = -1;
  }
  $29 = (($stackpos$tr) + 1)|0;
  $30 = (_luaF_getlocalname($13,$29,$$0$i)|0);
  HEAP32[$name>>2] = $30;
  $31 = ($30|0)==(0|0);
  if (!($31)) {
   $$3 = 519801;
   label = 24;
   break;
  }
  $32 = (_symbexec($13,$$0$i,$stackpos$tr)|0);
  $33 = $32 & 63;
  switch ($33|0) {
  case 5:  {
   $$lcssa = $13;$$lcssa47 = $32;
   label = 11;
   break L1;
   break;
  }
  case 6:  {
   $$lcssa44 = $13;$$lcssa48 = $32;
   label = 13;
   break L1;
   break;
  }
  case 4:  {
   $$lcssa45 = $13;$$lcssa49 = $32;
   label = 17;
   break L1;
   break;
  }
  case 11:  {
   $$lcssa46 = $13;$$lcssa50 = $32;
   label = 20;
   break L1;
   break;
  }
  case 0:  {
   break;
  }
  default: {
   $$3 = 0;
   label = 24;
   break L1;
  }
  }
  $40 = $32 >>> 6;
  $41 = $40 & 255;
  $42 = $32 >>> 23;
  $43 = ($42>>>0)<($41>>>0);
  if ($43) {
   $stackpos$tr = $42;
  } else {
   $$3 = 0;
   label = 24;
   break;
  }
 }
 if ((label|0) == 11) {
  $34 = $$lcssa47 >>> 14;
  $35 = ((($$lcssa)) + 8|0);
  $36 = HEAP32[$35>>2]|0;
  $37 = (($36) + ($34<<4)|0);
  $38 = HEAP32[$37>>2]|0;
  $39 = ((($38)) + 16|0);
  HEAP32[$name>>2] = $39;
  $$3 = 517613;
  return ($$3|0);
 }
 else if ((label|0) == 13) {
  $44 = $$lcssa48 >>> 14;
  $45 = $44 & 256;
  $46 = ($45|0)==(0);
  if ($46) {
   $$0$i4 = 523335;
  } else {
   $47 = $44 & 255;
   $48 = ((($$lcssa44)) + 8|0);
   $49 = HEAP32[$48>>2]|0;
   $50 = (((($49) + ($47<<4)|0)) + 8|0);
   $51 = HEAP32[$50>>2]|0;
   $52 = ($51|0)==(4);
   if ($52) {
    $53 = (($49) + ($47<<4)|0);
    $54 = HEAP32[$53>>2]|0;
    $55 = ((($54)) + 16|0);
    $$0$i4 = $55;
   } else {
    $$0$i4 = 523335;
   }
  }
  HEAP32[$name>>2] = $$0$i4;
  $$3 = 517620;
  return ($$3|0);
 }
 else if ((label|0) == 17) {
  $56 = ((($$lcssa45)) + 28|0);
  $57 = HEAP32[$56>>2]|0;
  $58 = ($57|0)==(0|0);
  if ($58) {
   $63 = 523335;
  } else {
   $59 = $$lcssa49 >>> 23;
   $60 = (($57) + ($59<<2)|0);
   $61 = HEAP32[$60>>2]|0;
   $62 = ((($61)) + 16|0);
   $63 = $62;
  }
  HEAP32[$name>>2] = $63;
  $$3 = 517626;
  return ($$3|0);
 }
 else if ((label|0) == 20) {
  $64 = $$lcssa50 >>> 14;
  $65 = $64 & 256;
  $66 = ($65|0)==(0);
  if ($66) {
   $$0$i3 = 523335;
  } else {
   $67 = $64 & 255;
   $68 = ((($$lcssa46)) + 8|0);
   $69 = HEAP32[$68>>2]|0;
   $70 = (((($69) + ($67<<4)|0)) + 8|0);
   $71 = HEAP32[$70>>2]|0;
   $72 = ($71|0)==(4);
   if ($72) {
    $73 = (($69) + ($67<<4)|0);
    $74 = HEAP32[$73>>2]|0;
    $75 = ((($74)) + 16|0);
    $$0$i3 = $75;
   } else {
    $$0$i3 = 523335;
   }
  }
  HEAP32[$name>>2] = $$0$i3;
  $$3 = 519425;
  return ($$3|0);
 }
 else if ((label|0) == 24) {
  return ($$3|0);
 }
 return (0)|0;
}
function _symbexec($pt,$lastpc,$reg) {
 $pt = $pt|0;
 $lastpc = $lastpc|0;
 $reg = $reg|0;
 var $$10 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $b$0 = 0;
 var $c$0 = 0, $j$0$lcssa$ph = 0, $j$051 = 0, $j4$054 = 0, $last$057 = 0, $last$1 = 0, $last$1$pc$0 = 0, $last$1$pc$019 = 0, $last$1$pc$023 = 0, $last$342 = 0, $last$342$lcssa = 0, $notlhs = 0, $notrhs = 0, $or$cond = 0, $or$cond$i = 0, $or$cond1$i = 0, $or$cond13 = 0, $or$cond16 = 0, $or$cond17$not = 0, $or$cond18 = 0;
 var $or$cond2$i = 0, $or$cond68 = 0, $or$cond70 = 0, $or$cond72 = 0, $or$cond74 = 0, $pc$0$ = 0, $pc$0$last$1 = 0, $pc$058 = 0, $pc$1 = 0, $pc$543 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pt)) + 44|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($pt)) + 75|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = ($3&255)<(251);
 if (!($4)) {
  $$10 = 0;
  return ($$10|0);
 }
 $5 = $3&255;
 $6 = ((($pt)) + 73|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $9 = ((($pt)) + 74|0);
 $10 = HEAP8[$9>>0]|0;
 $11 = $10&255;
 $12 = $11 & 1;
 $13 = (($12) + ($8))|0;
 $14 = ($13>>>0)>($5>>>0);
 $15 = $11 & 5;
 $16 = ($15|0)==(4);
 $or$cond$i = $16 | $14;
 if ($or$cond$i) {
  $$10 = 0;
  return ($$10|0);
 }
 $17 = ((($pt)) + 36|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($pt)) + 72|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = $20 & 255;
 $22 = ($18|0)>($21|0);
 $23 = $20 >>> 24;
 if ($22) {
  $$10 = 0;
  return ($$10|0);
 }
 $24 = ((($pt)) + 48|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = ($25|0)==($1|0);
 $27 = ($25|0)==(0);
 $or$cond1$i = $27 | $26;
 $28 = ($1|0)>(0);
 $or$cond2$i = $28 & $or$cond1$i;
 if (!($or$cond2$i)) {
  $$10 = 0;
  return ($$10|0);
 }
 $29 = (($1) + -1)|0;
 $30 = ((($pt)) + 12|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = (($31) + ($29<<2)|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = $33 & 63;
 $35 = ($34|0)==(30);
 if (!($35)) {
  $$10 = 0;
  return ($$10|0);
 }
 $36 = ($lastpc|0)>(0);
 if (!($36)) {
  $$10 = $33;
  return ($$10|0);
 }
 $37 = ((($pt)) + 8|0);
 $notlhs = ($reg|0)==(255);
 $38 = ((($pt)) + 52|0);
 $39 = ((($pt)) + 16|0);
 $40 = ((($pt)) + 40|0);
 $41 = $20 & 393216;
 $42 = ($41|0)==(131072);
 $last$057 = $29;$pc$058 = 0;
 L19: while(1) {
  $43 = (($31) + ($pc$058<<2)|0);
  $44 = HEAP32[$43>>2]|0;
  $45 = $44 & 63;
  $46 = $44 >>> 6;
  $47 = $46 & 255;
  $48 = ($45>>>0)<(38);
  $49 = ($47>>>0)<($23>>>0);
  $or$cond = $48 & $49;
  if (!($or$cond)) {
   $$10 = 0;
   label = 73;
   break;
  }
  $50 = (518247 + ($45)|0);
  $51 = HEAP8[$50>>0]|0;
  $52 = $51&255;
  $53 = $52 & 3;
  L22: do {
   switch ($53|0) {
   case 0:  {
    $54 = $44 >>> 23;
    $55 = $44 >>> 14;
    $56 = $55 & 511;
    $57 = $52 >>> 4;
    $58 = $57 & 3;
    L24: do {
     switch ($58|0) {
     case 0:  {
      $59 = ($54|0)==(0);
      if (!($59)) {
       $$10 = 0;
       label = 73;
       break L19;
      }
      break;
     }
     case 3:  {
      $61 = $54 & 256;
      $62 = ($61|0)==(0);
      if ($62) {
       $66 = ($23>>>0)>($54>>>0);
       if ($66) {
        break L24;
       } else {
        $$10 = 0;
        label = 73;
        break L19;
       }
      } else {
       $63 = $54 & 255;
       $64 = HEAP32[$40>>2]|0;
       $65 = ($63|0)<($64|0);
       if ($65) {
        break L24;
       } else {
        $$10 = 0;
        label = 73;
        break L19;
       }
      }
      break;
     }
     case 2:  {
      $60 = ($23>>>0)>($54>>>0);
      if (!($60)) {
       $$10 = 0;
       label = 73;
       break L19;
      }
      break;
     }
     default: {
     }
     }
    } while(0);
    $67 = $52 >>> 2;
    $68 = $67 & 3;
    switch ($68|0) {
    case 0:  {
     $69 = ($56|0)==(0);
     if ($69) {
      $b$0 = $54;$c$0 = 0;
      break L22;
     } else {
      $$10 = 0;
      label = 73;
      break L19;
     }
     break;
    }
    case 3:  {
     $71 = $55 & 256;
     $72 = ($71|0)==(0);
     if ($72) {
      $76 = ($23>>>0)>($56>>>0);
      if ($76) {
       $b$0 = $54;$c$0 = $56;
       break L22;
      } else {
       $$10 = 0;
       label = 73;
       break L19;
      }
     } else {
      $73 = $55 & 255;
      $74 = HEAP32[$40>>2]|0;
      $75 = ($73|0)<($74|0);
      if ($75) {
       $b$0 = $54;$c$0 = $56;
       break L22;
      } else {
       $$10 = 0;
       label = 73;
       break L19;
      }
     }
     break;
    }
    case 2:  {
     $70 = ($23>>>0)>($56>>>0);
     if ($70) {
      $b$0 = $54;$c$0 = $56;
      break L22;
     } else {
      $$10 = 0;
      label = 73;
      break L19;
     }
     break;
    }
    default: {
     $b$0 = $54;$c$0 = $56;
     break L22;
    }
    }
    break;
   }
   case 1:  {
    $77 = $44 >>> 14;
    $78 = $52 & 48;
    $79 = ($78|0)==(48);
    if ($79) {
     $80 = HEAP32[$40>>2]|0;
     $81 = ($77|0)<($80|0);
     if ($81) {
      $b$0 = $77;$c$0 = 0;
     } else {
      $$10 = 0;
      label = 73;
      break L19;
     }
    } else {
     $b$0 = $77;$c$0 = 0;
    }
    break;
   }
   case 2:  {
    $82 = $44 >>> 14;
    $83 = (($82) + -131071)|0;
    $84 = $52 & 48;
    $85 = ($84|0)==(32);
    if ($85) {
     $86 = (($pc$058) + 1)|0;
     $87 = (($86) + ($83))|0;
     $88 = ($87|0)>(-1);
     $89 = ($87|0)<($1|0);
     $or$cond68 = $88 & $89;
     if (!($or$cond68)) {
      $$10 = 0;
      label = 73;
      break L19;
     }
     $90 = ($87|0)>(0);
     if ($90) {
      $91 = (($83) + ($pc$058))|0;
      $j$051 = 0;
      while(1) {
       $92 = (($91) - ($j$051))|0;
       $93 = (($31) + ($92<<2)|0);
       $94 = HEAP32[$93>>2]|0;
       $95 = $94 & 8372287;
       $96 = ($95|0)==(34);
       if (!($96)) {
        $j$0$lcssa$ph = $j$051;
        break;
       }
       $97 = (($j$051) + 1)|0;
       $98 = ($97|0)<($87|0);
       if ($98) {
        $j$051 = $97;
       } else {
        $j$0$lcssa$ph = $97;
        break;
       }
      }
      $99 = $j$0$lcssa$ph & 1;
      $100 = ($99|0)==(0);
      if ($100) {
       $b$0 = $83;$c$0 = 0;
      } else {
       $$10 = 0;
       label = 73;
       break L19;
      }
     } else {
      $b$0 = $83;$c$0 = 0;
     }
    } else {
     $b$0 = $83;$c$0 = 0;
    }
    break;
   }
   default: {
    $b$0 = 0;$c$0 = 0;
   }
   }
  } while(0);
  $101 = $51 & 64;
  $102 = ($101<<24>>24)!=(0);
  $103 = ($47|0)==($reg|0);
  $or$cond13 = $103 & $102;
  $last$1 = $or$cond13 ? $pc$058 : $last$057;
  $104 = ($51<<24>>24)<(0);
  if ($104) {
   $105 = (($pc$058) + 2)|0;
   $106 = ($105|0)<($1|0);
   if (!($106)) {
    $$10 = 0;
    label = 73;
    break;
   }
   $107 = (($pc$058) + 1)|0;
   $108 = (($31) + ($107<<2)|0);
   $109 = HEAP32[$108>>2]|0;
   $110 = $109 & 63;
   $111 = ($110|0)==(22);
   if (!($111)) {
    $$10 = 0;
    label = 73;
    break;
   }
  }
  do {
   switch ($45|0) {
   case 2:  {
    $112 = ($c$0|0)==(1);
    if ($112) {
     $113 = (($pc$058) + 2)|0;
     $114 = ($113|0)<($1|0);
     if (!($114)) {
      $$10 = 0;
      label = 73;
      break L19;
     }
     $115 = (($pc$058) + 1)|0;
     $116 = (($31) + ($115<<2)|0);
     $117 = HEAP32[$116>>2]|0;
     $118 = $117 & 8372287;
     $119 = ($118|0)==(34);
     if ($119) {
      $$10 = 0;
      label = 73;
      break L19;
     } else {
      $last$342 = $last$1;$pc$543 = $pc$058;
     }
    } else {
     $last$342 = $last$1;$pc$543 = $pc$058;
    }
    break;
   }
   case 3:  {
    $120 = ($47|0)>($reg|0);
    $121 = ($b$0|0)<($reg|0);
    $or$cond16 = $120 | $121;
    $last$1$pc$023 = $or$cond16 ? $last$1 : $pc$058;
    $last$342 = $last$1$pc$023;$pc$543 = $pc$058;
    break;
   }
   case 8: case 4:  {
    $122 = ($b$0|0)<($21|0);
    if ($122) {
     $last$342 = $last$1;$pc$543 = $pc$058;
    } else {
     $$10 = 0;
     label = 73;
     break L19;
    }
    break;
   }
   case 7: case 5:  {
    $123 = HEAP32[$37>>2]|0;
    $124 = (((($123) + ($b$0<<4)|0)) + 8|0);
    $125 = HEAP32[$124>>2]|0;
    $126 = ($125|0)==(4);
    if ($126) {
     $last$342 = $last$1;$pc$543 = $pc$058;
    } else {
     $$10 = 0;
     label = 73;
     break L19;
    }
    break;
   }
   case 11:  {
    $127 = (($47) + 1)|0;
    $128 = ($127>>>0)<($23>>>0);
    if (!($128)) {
     $$10 = 0;
     label = 73;
     break L19;
    }
    $129 = ($127|0)==($reg|0);
    $pc$0$last$1 = $129 ? $pc$058 : $last$1;
    $last$342 = $pc$0$last$1;$pc$543 = $pc$058;
    break;
   }
   case 21:  {
    $130 = ($b$0|0)<($c$0|0);
    if ($130) {
     $last$342 = $last$1;$pc$543 = $pc$058;
    } else {
     $$10 = 0;
     label = 73;
     break L19;
    }
    break;
   }
   case 33:  {
    $131 = ($c$0|0)==(0);
    if ($131) {
     $$10 = 0;
     label = 73;
     break L19;
    }
    $132 = (($47) + 2)|0;
    $133 = (($c$0) + ($132))|0;
    $134 = ($133>>>0)<($23>>>0);
    if (!($134)) {
     $$10 = 0;
     label = 73;
     break L19;
    }
    $135 = ($132|0)>($reg|0);
    $last$1$pc$0 = $135 ? $last$1 : $pc$058;
    $last$342 = $last$1$pc$0;$pc$543 = $pc$058;
    break;
   }
   case 32: case 31:  {
    $136 = (($47) + 3)|0;
    $137 = ($136>>>0)<($23>>>0);
    if ($137) {
     label = 48;
    } else {
     $$10 = 0;
     label = 73;
     break L19;
    }
    break;
   }
   case 22:  {
    label = 48;
    break;
   }
   case 29: case 28:  {
    $142 = ($b$0|0)!=(0);
    $143 = (($b$0) + ($47))|0;
    $144 = ($143|0)>($23|0);
    $or$cond70 = $142 & $144;
    if ($or$cond70) {
     $$10 = 0;
     label = 73;
     break L19;
    }
    $145 = (($c$0) + -1)|0;
    $146 = ($c$0|0)==(0);
    if ($146) {
     $147 = (($pc$058) + 1)|0;
     $148 = (($31) + ($147<<2)|0);
     $149 = HEAP32[$148>>2]|0;
     $150 = $149 & 63;
     switch ($150|0) {
     case 34: case 30: case 29: case 28:  {
      break;
     }
     default: {
      $$10 = 0;
      label = 73;
      break L19;
     }
     }
     $151 = ($149>>>0)<(8388608);
     if (!($151)) {
      $$10 = 0;
      label = 73;
      break L19;
     }
    } else {
     $152 = ($145|0)!=(0);
     $153 = (($145) + ($47))|0;
     $154 = ($153|0)>($23|0);
     $or$cond72 = $152 & $154;
     if ($or$cond72) {
      $$10 = 0;
      label = 73;
      break L19;
     }
    }
    $155 = ($47|0)>($reg|0);
    $last$1$pc$019 = $155 ? $last$1 : $pc$058;
    $last$342 = $last$1$pc$019;$pc$543 = $pc$058;
    break;
   }
   case 30:  {
    $156 = ($b$0|0)>(1);
    if ($156) {
     $157 = (($47) + -1)|0;
     $158 = (($157) + ($b$0))|0;
     $159 = ($158|0)>($23|0);
     if ($159) {
      $$10 = 0;
      label = 73;
      break L19;
     } else {
      $last$342 = $last$1;$pc$543 = $pc$058;
     }
    } else {
     $last$342 = $last$1;$pc$543 = $pc$058;
    }
    break;
   }
   case 34:  {
    $160 = ($b$0|0)<(1);
    $161 = (($b$0) + ($47))|0;
    $162 = ($161|0)<($23|0);
    $or$cond74 = $160 | $162;
    if (!($or$cond74)) {
     $$10 = 0;
     label = 73;
     break L19;
    }
    $163 = ($c$0|0)==(0);
    if ($163) {
     $164 = (($pc$058) + 1)|0;
     $165 = ($164|0)<($29|0);
     if ($165) {
      $last$342 = $last$1;$pc$543 = $164;
     } else {
      $$10 = 0;
      label = 73;
      break L19;
     }
    } else {
     $last$342 = $last$1;$pc$543 = $pc$058;
    }
    break;
   }
   case 36:  {
    $166 = HEAP32[$38>>2]|0;
    $167 = ($b$0|0)<($166|0);
    if (!($167)) {
     $$10 = 0;
     label = 73;
     break L19;
    }
    $168 = HEAP32[$39>>2]|0;
    $169 = (($168) + ($b$0<<2)|0);
    $170 = HEAP32[$169>>2]|0;
    $171 = ((($170)) + 72|0);
    $172 = HEAP8[$171>>0]|0;
    $173 = $172&255;
    $174 = (($173) + ($pc$058))|0;
    $175 = ($174|0)<($1|0);
    if (!($175)) {
     $$10 = 0;
     label = 73;
     break L19;
    }
    $176 = ($172<<24>>24)==(0);
    if (!($176)) {
     $j4$054 = 1;
     while(1) {
      $179 = (($j4$054) + ($pc$058))|0;
      $180 = (($31) + ($179<<2)|0);
      $181 = HEAP32[$180>>2]|0;
      $182 = $181 & 59;
      $183 = $182 | 4;
      $184 = ($183|0)==(4);
      if (!($184)) {
       $$10 = 0;
       label = 73;
       break L19;
      }
      $177 = (($j4$054) + 1)|0;
      $178 = ($j4$054|0)<($173|0);
      if ($178) {
       $j4$054 = $177;
      } else {
       break;
      }
     }
    }
    $pc$0$ = $notlhs ? $pc$058 : $174;
    $last$342 = $last$1;$pc$543 = $pc$0$;
    break;
   }
   case 37:  {
    if (!($42)) {
     $$10 = 0;
     label = 73;
     break L19;
    }
    $185 = ($b$0|0)==(0);
    if ($185) {
     $186 = (($pc$058) + 1)|0;
     $187 = (($31) + ($186<<2)|0);
     $188 = HEAP32[$187>>2]|0;
     $189 = $188 & 63;
     switch ($189|0) {
     case 34: case 30: case 29: case 28:  {
      break;
     }
     default: {
      $$10 = 0;
      label = 73;
      break L19;
     }
     }
     $190 = ($188>>>0)<(8388608);
     if (!($190)) {
      $$10 = 0;
      label = 73;
      break L19;
     }
    }
    $191 = (($47) + -1)|0;
    $192 = (($191) + ($b$0))|0;
    $193 = ($192|0)>($23|0);
    if ($193) {
     $$10 = 0;
     label = 73;
     break L19;
    } else {
     $last$342 = $last$1;$pc$543 = $pc$058;
    }
    break;
   }
   default: {
    $last$342 = $last$1;$pc$543 = $pc$058;
   }
   }
  } while(0);
  if ((label|0) == 48) {
   label = 0;
   $138 = (($pc$058) + 1)|0;
   $139 = (($138) + ($b$0))|0;
   $notrhs = ($pc$058|0)>=($139|0);
   $or$cond17$not = $notlhs | $notrhs;
   $140 = ($139|0)>($lastpc|0);
   $or$cond18 = $140 | $or$cond17$not;
   $141 = $or$cond18 ? 0 : $b$0;
   $pc$1 = (($141) + ($pc$058))|0;
   $last$342 = $last$1;$pc$543 = $pc$1;
  }
  $194 = (($pc$543) + 1)|0;
  $195 = ($194|0)<($lastpc|0);
  if ($195) {
   $last$057 = $last$342;$pc$058 = $194;
  } else {
   $last$342$lcssa = $last$342;
   label = 72;
   break;
  }
 }
 if ((label|0) == 72) {
  $$phi$trans$insert = (($31) + ($last$342$lcssa<<2)|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $$10 = $$pre;
  return ($$10|0);
 }
 else if ((label|0) == 73) {
  return ($$10|0);
 }
 return (0)|0;
}
function _luaG_checkcode($pt) {
 $pt = $pt|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pt)) + 44|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_symbexec($pt,$1,255)|0);
 $3 = ($2|0)!=(0);
 $4 = $3&1;
 return ($4|0);
}
function _luaG_typeerror($L,$o,$op) {
 $L = $L|0;
 $o = $o|0;
 $op = $op|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $name = 0, $p$01$i = 0, $vararg_buffer = 0, $vararg_buffer4 = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0, $vararg_ptr7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer = sp;
 $name = sp + 24|0;
 HEAP32[$name>>2] = 0;
 $0 = ((($o)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (38844 + ($1<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($L)) + 20|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($5)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($6>>>0)<($8>>>0);
 L1: do {
  if ($9) {
   $p$01$i = $6;
   while(1) {
    $12 = ($p$01$i|0)==($o|0);
    $11 = ((($p$01$i)) + 16|0);
    if ($12) {
     break;
    }
    $10 = ($11>>>0)<($8>>>0);
    if ($10) {
     $p$01$i = $11;
    } else {
     break L1;
    }
   }
   $13 = ((($L)) + 12|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = $o;
   $16 = (($15) - ($14))|0;
   $17 = $16 >> 4;
   $18 = (_getobjname($L,$5,$17,$name)|0);
   $19 = ($18|0)==(0|0);
   if (!($19)) {
    $20 = HEAP32[$name>>2]|0;
    HEAP32[$vararg_buffer>>2] = $op;
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
    HEAP32[$vararg_ptr1>>2] = $18;
    $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
    HEAP32[$vararg_ptr2>>2] = $20;
    $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
    HEAP32[$vararg_ptr3>>2] = $3;
    _luaG_runerror($L,517634,$vararg_buffer);
    STACKTOP = sp;return;
   }
  }
 } while(0);
 HEAP32[$vararg_buffer4>>2] = $op;
 $vararg_ptr7 = ((($vararg_buffer4)) + 4|0);
 HEAP32[$vararg_ptr7>>2] = $3;
 _luaG_runerror($L,517669,$vararg_buffer4);
 STACKTOP = sp;return;
}
function _luaG_runerror($L,$fmt,$varargs) {
 $L = $L|0;
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $$0$i$i = 0, $$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $argp = 0, $buff$i = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $buff$i = sp + 32|0;
 $argp = sp + 16|0;
 HEAP32[$argp>>2] = $varargs;
 $0 = (_luaO_pushvfstring($L,$fmt,$argp)|0);
 $1 = ((($L)) + 20|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(6);
 if (!($7)) {
  _luaG_errormsg($L);
  STACKTOP = sp;return;
 }
 $8 = HEAP32[$4>>2]|0;
 $9 = ((($8)) + 6|0);
 $10 = HEAP8[$9>>0]|0;
 $11 = ($10<<24>>24)==(0);
 if (!($11)) {
  _luaG_errormsg($L);
  STACKTOP = sp;return;
 }
 $12 = ((($L)) + 24|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($2)) + 12|0);
 HEAP32[$14>>2] = $13;
 $$pre$i$i = HEAP32[$4>>2]|0;
 $15 = ((($$pre$i$i)) + 16|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($16)) + 12|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = (($13) - ($18))|0;
 $20 = $19 >> 2;
 $21 = (($20) + -1)|0;
 $22 = ($19|0)<(4);
 if ($22) {
  $$0$i$i = -1;
 } else {
  $23 = ((($16)) + 20|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = ($24|0)==(0|0);
  if ($25) {
   $$0$i$i = 0;
  } else {
   $26 = (($24) + ($21<<2)|0);
   $27 = HEAP32[$26>>2]|0;
   $$0$i$i = $27;
  }
 }
 $28 = ((($$pre$i$i)) + 16|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = ((($29)) + 32|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = ((($31)) + 16|0);
 _luaO_chunkid($buff$i,$32,60);
 HEAP32[$vararg_buffer>>2] = $buff$i;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $$0$i$i;
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
 HEAP32[$vararg_ptr2>>2] = $0;
 (_luaO_pushfstring($L,519897,$vararg_buffer)|0);
 _luaG_errormsg($L);
 STACKTOP = sp;return;
}
function _luaG_errormsg($L) {
 $L = $L|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 116|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  _luaD_throw($L,2);
  // unreachable;
 }
 $3 = ((($L)) + 32|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + ($1)|0);
 $6 = ((($5)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(6);
 if (!($8)) {
  _luaD_throw($L,5);
  // unreachable;
 }
 $9 = ((($L)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + -16|0);
 $12 = $11;
 $13 = $12;
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12) + 4)|0;
 $16 = $15;
 $17 = HEAP32[$16>>2]|0;
 $18 = $10;
 $19 = $18;
 HEAP32[$19>>2] = $14;
 $20 = (($18) + 4)|0;
 $21 = $20;
 HEAP32[$21>>2] = $17;
 $22 = ((($10)) + -8|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = ((($10)) + 8|0);
 HEAP32[$24>>2] = $23;
 $25 = HEAP32[$9>>2]|0;
 $26 = ((($25)) + -16|0);
 $27 = $5;
 $28 = $27;
 $29 = HEAP32[$28>>2]|0;
 $30 = (($27) + 4)|0;
 $31 = $30;
 $32 = HEAP32[$31>>2]|0;
 $33 = $26;
 $34 = $33;
 HEAP32[$34>>2] = $29;
 $35 = (($33) + 4)|0;
 $36 = $35;
 HEAP32[$36>>2] = $32;
 $37 = HEAP32[$6>>2]|0;
 $38 = ((($25)) + -8|0);
 HEAP32[$38>>2] = $37;
 $39 = ((($L)) + 28|0);
 $40 = HEAP32[$39>>2]|0;
 $41 = HEAP32[$9>>2]|0;
 $42 = (($40) - ($41))|0;
 $43 = ($42|0)<(17);
 $44 = $41;
 if ($43) {
  _luaD_growstack($L,1);
  $$pre = HEAP32[$9>>2]|0;
  $46 = $$pre;
 } else {
  $46 = $44;
 }
 $45 = ((($46)) + 16|0);
 HEAP32[$9>>2] = $45;
 $47 = ((($46)) + -16|0);
 _luaD_call($L,$47,1);
 _luaD_throw($L,2);
 // unreachable;
}
function _luaG_concaterror($L,$p1,$p2) {
 $L = $L|0;
 $p1 = $p1|0;
 $p2 = $p2|0;
 var $$off = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $name$i = 0, $p$01$i$i = 0, $p2$p1 = 0, $switch = 0, $vararg_buffer = 0, $vararg_buffer4 = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0, $vararg_ptr7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer = sp;
 $name$i = sp + 24|0;
 $0 = ((($p1)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $$off = (($1) + -3)|0;
 $switch = ($$off>>>0)<(2);
 $p2$p1 = $switch ? $p2 : $p1;
 HEAP32[$name$i>>2] = 0;
 $2 = ((($p2$p1)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (38844 + ($3<<2)|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($L)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($7)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($8>>>0)<($10>>>0);
 L1: do {
  if ($11) {
   $p$01$i$i = $8;
   while(1) {
    $14 = ($p$01$i$i|0)==($p2$p1|0);
    $13 = ((($p$01$i$i)) + 16|0);
    if ($14) {
     break;
    }
    $12 = ($13>>>0)<($10>>>0);
    if ($12) {
     $p$01$i$i = $13;
    } else {
     break L1;
    }
   }
   $15 = ((($L)) + 12|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = $p2$p1;
   $18 = (($17) - ($16))|0;
   $19 = $18 >> 4;
   $20 = (_getobjname($L,$7,$19,$name$i)|0);
   $21 = ($20|0)==(0|0);
   if (!($21)) {
    $22 = HEAP32[$name$i>>2]|0;
    HEAP32[$vararg_buffer>>2] = 517694;
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
    HEAP32[$vararg_ptr1>>2] = $20;
    $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
    HEAP32[$vararg_ptr2>>2] = $22;
    $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
    HEAP32[$vararg_ptr3>>2] = $5;
    _luaG_runerror($L,517634,$vararg_buffer);
    STACKTOP = sp;return;
   }
  }
 } while(0);
 HEAP32[$vararg_buffer4>>2] = 517694;
 $vararg_ptr7 = ((($vararg_buffer4)) + 4|0);
 HEAP32[$vararg_ptr7>>2] = $5;
 _luaG_runerror($L,517669,$vararg_buffer4);
 STACKTOP = sp;return;
}
function _luaG_aritherror($L,$p1,$p2) {
 $L = $L|0;
 $p1 = $p1|0;
 $p2 = $p2|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $name$i = 0, $p$01$i$i = 0, $p1$p2 = 0, $temp = 0, $vararg_buffer = 0, $vararg_buffer4 = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0, $vararg_ptr7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 32|0;
 $vararg_buffer = sp + 16|0;
 $name$i = sp + 40|0;
 $temp = sp;
 $0 = (_luaV_tonumber($p1,$temp)|0);
 $1 = ($0|0)==(0|0);
 $p1$p2 = $1 ? $p1 : $p2;
 HEAP32[$name$i>>2] = 0;
 $2 = ((($p1$p2)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (38844 + ($3<<2)|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($L)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($7)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($8>>>0)<($10>>>0);
 L1: do {
  if ($11) {
   $p$01$i$i = $8;
   while(1) {
    $14 = ($p$01$i$i|0)==($p1$p2|0);
    $13 = ((($p$01$i$i)) + 16|0);
    if ($14) {
     break;
    }
    $12 = ($13>>>0)<($10>>>0);
    if ($12) {
     $p$01$i$i = $13;
    } else {
     break L1;
    }
   }
   $15 = ((($L)) + 12|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = $p1$p2;
   $18 = (($17) - ($16))|0;
   $19 = $18 >> 4;
   $20 = (_getobjname($L,$7,$19,$name$i)|0);
   $21 = ($20|0)==(0|0);
   if (!($21)) {
    $22 = HEAP32[$name$i>>2]|0;
    HEAP32[$vararg_buffer>>2] = 517706;
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
    HEAP32[$vararg_ptr1>>2] = $20;
    $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
    HEAP32[$vararg_ptr2>>2] = $22;
    $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
    HEAP32[$vararg_ptr3>>2] = $5;
    _luaG_runerror($L,517634,$vararg_buffer);
    STACKTOP = sp;return;
   }
  }
 } while(0);
 HEAP32[$vararg_buffer4>>2] = 517706;
 $vararg_ptr7 = ((($vararg_buffer4)) + 4|0);
 HEAP32[$vararg_ptr7>>2] = $5;
 _luaG_runerror($L,517669,$vararg_buffer4);
 STACKTOP = sp;return;
}
function _luaG_ordererror($L,$p1,$p2) {
 $L = $L|0;
 $p1 = $p1|0;
 $p2 = $p2|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_ptr4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($p1)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (38844 + ($1<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($p2)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (38844 + ($5<<2)|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($3)) + 2|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = ((($7)) + 2|0);
 $11 = HEAP8[$10>>0]|0;
 $12 = ($9<<24>>24)==($11<<24>>24);
 if ($12) {
  HEAP32[$vararg_buffer>>2] = $3;
  _luaG_runerror($L,517728,$vararg_buffer);
  STACKTOP = sp;return 0;
 } else {
  HEAP32[$vararg_buffer1>>2] = $3;
  $vararg_ptr4 = ((($vararg_buffer1)) + 4|0);
  HEAP32[$vararg_ptr4>>2] = $7;
  _luaG_runerror($L,517761,$vararg_buffer1);
  STACKTOP = sp;return 0;
 }
 return (0)|0;
}
function _luaD_throw($L,$errcode) {
 $L = $L|0;
 $errcode = $errcode|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 112|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = ((($1)) + 160|0);
  HEAP32[$3>>2] = $errcode;
  $4 = HEAP32[$0>>2]|0;
  $5 = ((($4)) + 4|0);
  _longjmp(($5|0),1);
  // unreachable;
 }
 $6 = $errcode&255;
 $7 = ((($L)) + 6|0);
 HEAP8[$7>>0] = $6;
 $8 = ((($L)) + 16|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 88|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0|0);
 if ($12) {
  _exit(1);
  // unreachable;
 }
 $13 = ((($L)) + 40|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($L)) + 20|0);
 HEAP32[$15>>2] = $14;
 $16 = HEAP32[$14>>2]|0;
 $17 = ((($L)) + 12|0);
 HEAP32[$17>>2] = $16;
 _luaF_close($L,$16);
 $18 = HEAP32[$17>>2]|0;
 switch ($errcode|0) {
 case 4:  {
  $19 = (_luaS_newlstr($L,518874,17)|0);
  HEAP32[$18>>2] = $19;
  $20 = ((($18)) + 8|0);
  HEAP32[$20>>2] = 4;
  break;
 }
 case 5:  {
  $21 = (_luaS_newlstr($L,517791,23)|0);
  HEAP32[$18>>2] = $21;
  $22 = ((($18)) + 8|0);
  HEAP32[$22>>2] = 4;
  break;
 }
 case 2: case 3:  {
  $23 = ((($L)) + 8|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = ((($24)) + -16|0);
  $26 = $25;
  $27 = $26;
  $28 = HEAP32[$27>>2]|0;
  $29 = (($26) + 4)|0;
  $30 = $29;
  $31 = HEAP32[$30>>2]|0;
  $32 = $18;
  $33 = $32;
  HEAP32[$33>>2] = $28;
  $34 = (($32) + 4)|0;
  $35 = $34;
  HEAP32[$35>>2] = $31;
  $36 = ((($24)) + -8|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = ((($18)) + 8|0);
  HEAP32[$38>>2] = $37;
  break;
 }
 default: {
 }
 }
 $39 = ((($18)) + 16|0);
 $40 = ((($L)) + 8|0);
 HEAP32[$40>>2] = $39;
 $41 = ((($L)) + 54|0);
 $42 = HEAP16[$41>>1]|0;
 $43 = ((($L)) + 52|0);
 HEAP16[$43>>1] = $42;
 $44 = ((($L)) + 57|0);
 HEAP8[$44>>0] = 1;
 $45 = ((($L)) + 48|0);
 $46 = HEAP32[$45>>2]|0;
 $47 = ($46|0)>(20000);
 if ($47) {
  $48 = HEAP32[$15>>2]|0;
  $49 = HEAP32[$13>>2]|0;
  $50 = (($48) - ($49))|0;
  $51 = (($50|0) / 24)&-1;
  $52 = (($51) + 1)|0;
  $53 = ($52|0)<(20000);
  if ($53) {
   $54 = $49;
   $55 = ($46*24)|0;
   $56 = (_luaM_realloc_($L,$54,$55,480000)|0);
   HEAP32[$13>>2] = $56;
   HEAP32[$45>>2] = 20000;
   $57 = HEAP32[$15>>2]|0;
   $58 = (($57) - ($49))|0;
   $59 = (($58|0) / 24)&-1;
   $60 = (($56) + (($59*24)|0)|0);
   HEAP32[$15>>2] = $60;
   $61 = ((($56)) + 479976|0);
   $62 = ((($L)) + 36|0);
   HEAP32[$62>>2] = $61;
  }
 }
 $63 = ((($L)) + 116|0);
 HEAP32[$63>>2] = 0;
 HEAP32[$0>>2] = 0;
 $64 = HEAP32[$8>>2]|0;
 $65 = ((($64)) + 88|0);
 $66 = HEAP32[$65>>2]|0;
 (FUNCTION_TABLE_ii[$66 & 8191]($L)|0);
 _exit(1);
 // unreachable;
}
function _luaD_rawrunprotected($L,$f,$ud) {
 $L = $L|0;
 $f = $f|0;
 $ud = $ud|0;
 var $$reg2mem17$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cond = 0, $cond19 = 0, $lj = 0, _setjmpTable = 0, _setjmpTableSize = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 _setjmpTableSize = 4;_setjmpTable = _malloc(40) | 0;HEAP32[_setjmpTable>>2]=0;
 $lj = sp;
 $0 = ((($lj)) + 160|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($L)) + 112|0);
 $2 = HEAP32[$1>>2]|0;
 HEAP32[$lj>>2] = $2;
 HEAP32[$1>>2] = $lj;
 $3 = ((($lj)) + 4|0);
 _setjmpTable = _saveSetjmp($3,1,_setjmpTable|0,_setjmpTableSize|0)|0;_setjmpTableSize = tempRet0;
 __THREW__ = 0;
 $4 = __THREW__; __THREW__ = 0;
 if ((($4|0) != 0) & ((threwValue|0) != 0)) { $5 = _testSetjmp(HEAP32[$4>>2]|0, _setjmpTable|0, _setjmpTableSize|0)|0; if (($5|0) == 0) { _longjmp($4|0, threwValue|0); } tempRet0 = (threwValue); } else { $5 = -1; };
 $6 = tempRet0;
 $cond = ($5|0)==(1);
 if ($cond) {
  $$reg2mem17$0 = $6;
 } else {
  $$reg2mem17$0 = 0;
 }
 while(1) {
  $7 = ($$reg2mem17$0|0)==(0);
  if (!($7)) {
   label = 6;
   break;
  }
  __THREW__ = 0;
  invoke_vii($f|0,($L|0),($ud|0));
  $8 = __THREW__; __THREW__ = 0;
  if ((($8|0) != 0) & ((threwValue|0) != 0)) { $9 = _testSetjmp(HEAP32[$8>>2]|0, _setjmpTable|0, _setjmpTableSize|0)|0; if (($9|0) == 0) { _longjmp($8|0, threwValue|0); } tempRet0 = (threwValue); } else { $9 = -1; };
  $10 = tempRet0;
  $cond19 = ($9|0)==(1);
  if ($cond19) {
   $$reg2mem17$0 = $10;
  } else {
   break;
  }
 }
 if ((label|0) == 6) {
  $11 = HEAP32[$lj>>2]|0;
  HEAP32[$1>>2] = $11;
  $12 = HEAP32[$0>>2]|0;
  _free(_setjmpTable|0);
  STACKTOP = sp;return ($12|0);
 }
 $11 = HEAP32[$lj>>2]|0;
 HEAP32[$1>>2] = $11;
 $12 = HEAP32[$0>>2]|0;
 _free(_setjmpTable|0);
 STACKTOP = sp;return ($12|0);
}
function _luaD_reallocstack($L,$newsize) {
 $L = $L|0;
 $newsize = $newsize|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ci$01$i = 0, $up$0$i = 0;
 var $up$02$i = 0, $up$03$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($newsize) + 6)|0;
 $3 = (($newsize) + 7)|0;
 $4 = ($3>>>0)<(268435456);
 if ($4) {
  $5 = ((($L)) + 44|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = $6 << 4;
  $8 = $2 << 4;
  $9 = (_luaM_realloc_($L,$1,$7,$8)|0);
  $$pre$phiZ2D = $5;$11 = $9;
 } else {
  $10 = (_luaM_toobig($L)|0);
  $$pre = ((($L)) + 44|0);
  $$pre$phiZ2D = $$pre;$11 = $10;
 }
 HEAP32[$0>>2] = $11;
 HEAP32[$$pre$phiZ2D>>2] = $2;
 $12 = (($11) + ($newsize<<4)|0);
 $13 = ((($L)) + 28|0);
 HEAP32[$13>>2] = $12;
 $14 = ((($L)) + 8|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = $1;
 $17 = (($15) - ($16))|0;
 $18 = $17 >> 4;
 $19 = (($11) + ($18<<4)|0);
 HEAP32[$14>>2] = $19;
 $20 = ((($L)) + 104|0);
 $up$02$i = HEAP32[$20>>2]|0;
 $21 = ($up$02$i|0)==(0|0);
 if (!($21)) {
  $up$03$i = $up$02$i;
  while(1) {
   $22 = ((($up$03$i)) + 8|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = (($23) - ($16))|0;
   $25 = $24 >> 4;
   $26 = (($11) + ($25<<4)|0);
   HEAP32[$22>>2] = $26;
   $up$0$i = HEAP32[$up$03$i>>2]|0;
   $27 = ($up$0$i|0)==(0|0);
   if ($27) {
    break;
   } else {
    $up$03$i = $up$0$i;
   }
  }
 }
 $28 = ((($L)) + 40|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = ((($L)) + 20|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = ($29>>>0)>($31>>>0);
 if ($32) {
  $50 = ((($L)) + 12|0);
  $51 = HEAP32[$50>>2]|0;
  $52 = (($51) - ($16))|0;
  $53 = $52 >> 4;
  $54 = HEAP32[$0>>2]|0;
  $55 = (($54) + ($53<<4)|0);
  HEAP32[$50>>2] = $55;
  return;
 }
 $33 = HEAP32[$0>>2]|0;
 $ci$01$i = $29;
 while(1) {
  $34 = ((($ci$01$i)) + 8|0);
  $35 = HEAP32[$34>>2]|0;
  $36 = (($35) - ($16))|0;
  $37 = $36 >> 4;
  $38 = (($33) + ($37<<4)|0);
  HEAP32[$34>>2] = $38;
  $39 = HEAP32[$ci$01$i>>2]|0;
  $40 = (($39) - ($16))|0;
  $41 = $40 >> 4;
  $42 = (($33) + ($41<<4)|0);
  HEAP32[$ci$01$i>>2] = $42;
  $43 = ((($ci$01$i)) + 4|0);
  $44 = HEAP32[$43>>2]|0;
  $45 = (($44) - ($16))|0;
  $46 = $45 >> 4;
  $47 = (($33) + ($46<<4)|0);
  HEAP32[$43>>2] = $47;
  $48 = ((($ci$01$i)) + 24|0);
  $49 = ($48>>>0)>($31>>>0);
  if ($49) {
   break;
  } else {
   $ci$01$i = $48;
  }
 }
 $50 = ((($L)) + 12|0);
 $51 = HEAP32[$50>>2]|0;
 $52 = (($51) - ($16))|0;
 $53 = $52 >> 4;
 $54 = HEAP32[$0>>2]|0;
 $55 = (($54) + ($53<<4)|0);
 HEAP32[$50>>2] = $55;
 return;
}
function _luaD_reallocCI($L,$newsize) {
 $L = $L|0;
 $newsize = $newsize|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 40|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($newsize) + 1)|0;
 $3 = ($2>>>0)<(178956971);
 if ($3) {
  $4 = ((($L)) + 48|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ($5*24)|0;
  $7 = ($newsize*24)|0;
  $8 = (_luaM_realloc_($L,$1,$6,$7)|0);
  $$pre$phiZ2D = $4;$10 = $8;
 } else {
  $9 = (_luaM_toobig($L)|0);
  $$pre = ((($L)) + 48|0);
  $$pre$phiZ2D = $$pre;$10 = $9;
 }
 HEAP32[$0>>2] = $10;
 HEAP32[$$pre$phiZ2D>>2] = $newsize;
 $11 = ((($L)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = $1;
 $14 = (($12) - ($13))|0;
 $15 = (($14|0) / 24)&-1;
 $16 = (($10) + (($15*24)|0)|0);
 HEAP32[$11>>2] = $16;
 $17 = (($10) + (($newsize*24)|0)|0);
 $18 = ((($17)) + -24|0);
 $19 = ((($L)) + 36|0);
 HEAP32[$19>>2] = $18;
 return;
}
function _luaD_growstack($L,$n) {
 $L = $L|0;
 $n = $n|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 44|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)<($n|0);
 if ($2) {
  $4 = (($1) + ($n))|0;
  _luaD_reallocstack($L,$4);
  return;
 } else {
  $3 = $1 << 1;
  _luaD_reallocstack($L,$3);
  return;
 }
}
function _luaD_callhook($L,$event,$line) {
 $L = $L|0;
 $event = $event|0;
 $line = $line|0;
 var $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $ar = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ar = sp;
 $0 = ((($L)) + 68|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  STACKTOP = sp;return;
 }
 $3 = ((($L)) + 57|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)==(0);
 if ($5) {
  STACKTOP = sp;return;
 }
 $6 = ((($L)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($L)) + 32|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (($7) - ($9))|0;
 $11 = ((($L)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($12)) + 8|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($14) - ($9))|0;
 HEAP32[$ar>>2] = $event;
 $16 = ((($ar)) + 20|0);
 HEAP32[$16>>2] = $line;
 $17 = ($event|0)==(4);
 if ($17) {
  $$sink = 0;
 } else {
  $18 = $12;
  $19 = ((($L)) + 40|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = (($18) - ($20))|0;
  $22 = (($21|0) / 24)&-1;
  $$sink = $22;
 }
 $23 = ((($ar)) + 96|0);
 HEAP32[$23>>2] = $$sink;
 $24 = ((($L)) + 28|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = (($25) - ($7))|0;
 $27 = ($26|0)<(321);
 do {
  if ($27) {
   $28 = ((($L)) + 44|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = ($29|0)<(20);
   if ($30) {
    $32 = (($29) + 20)|0;
    _luaD_reallocstack($L,$32);
    break;
   } else {
    $31 = $29 << 1;
    _luaD_reallocstack($L,$31);
    break;
   }
  }
 } while(0);
 $33 = HEAP32[$6>>2]|0;
 $34 = ((($33)) + 320|0);
 $35 = HEAP32[$11>>2]|0;
 $36 = ((($35)) + 8|0);
 HEAP32[$36>>2] = $34;
 HEAP8[$3>>0] = 0;
 FUNCTION_TABLE_vii[$1 & 8191]($L,$ar);
 HEAP8[$3>>0] = 1;
 $37 = HEAP32[$8>>2]|0;
 $38 = (($37) + ($15)|0);
 $39 = HEAP32[$11>>2]|0;
 $40 = ((($39)) + 8|0);
 HEAP32[$40>>2] = $38;
 $41 = (($37) + ($10)|0);
 HEAP32[$6>>2] = $41;
 STACKTOP = sp;return;
}
function _luaD_precall($L,$func,$nresults) {
 $L = $L|0;
 $func = $func|0;
 $nresults = $nresults|0;
 var $$0$lcssa$i = 0, $$01 = 0, $$05$i = 0, $$1 = 0, $$12 = 0, $$in = 0, $$pre = 0, $$pre$i = 0, $$pre$i4 = 0, $$pre$phi29Z2D = 0, $$pre$phiZ2D = 0, $$pre12$i = 0, $$pre22 = 0, $$pre23 = 0, $$pre24 = 0, $$pre25 = 0, $$pre27 = 0, $$sink = 0, $$sink19 = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0.0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0;
 var $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0;
 var $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0;
 var $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0;
 var $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0;
 var $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0;
 var $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0;
 var $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $ar$i = 0, $ar$i10 = 0, $base$0 = 0, $exitcond$i = 0, $exitcond$i21 = 0, $exitcond10$i = 0, $exitcond11$i = 0, $htab$0$i = 0, $i$02$i = 0, $p$01$i = 0, $scevgep$i = 0, $st$020 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $ar$i10 = sp + 112|0;
 $ar$i = sp + 12|0;
 $0 = ((($func)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(6);
 if ($2) {
  $$pre27 = ((($L)) + 32|0);
  $$01 = $func;$$pre$phi29Z2D = $$pre27;$$pre$phiZ2D = $$pre27;
 } else {
  $3 = (_luaT_gettmbyobj($L,$func,16)|0);
  $4 = ((($L)) + 32|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = $func;
  $7 = (($6) - ($5))|0;
  $8 = ((($3)) + 8|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)==(6);
  if (!($10)) {
   _luaG_typeerror($L,$func,521014);
  }
  $11 = ((($L)) + 8|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ($12>>>0)>($func>>>0);
  if ($13) {
   $p$01$i = $12;
   while(1) {
    $15 = ((($p$01$i)) + -16|0);
    $16 = $15;
    $17 = $16;
    $18 = HEAP32[$17>>2]|0;
    $19 = (($16) + 4)|0;
    $20 = $19;
    $21 = HEAP32[$20>>2]|0;
    $22 = $p$01$i;
    $23 = $22;
    HEAP32[$23>>2] = $18;
    $24 = (($22) + 4)|0;
    $25 = $24;
    HEAP32[$25>>2] = $21;
    $26 = ((($p$01$i)) + -8|0);
    $27 = HEAP32[$26>>2]|0;
    $28 = ((($p$01$i)) + 8|0);
    HEAP32[$28>>2] = $27;
    $29 = ($15>>>0)>($func>>>0);
    if ($29) {
     $p$01$i = $15;
    } else {
     break;
    }
   }
   $$pre$i = HEAP32[$11>>2]|0;
   $33 = $$pre$i;
  } else {
   $14 = $12;
   $33 = $14;
  }
  $30 = ((($L)) + 28|0);
  $31 = HEAP32[$30>>2]|0;
  $32 = (($31) - ($33))|0;
  $34 = ($32|0)<(17);
  do {
   if ($34) {
    $35 = ((($L)) + 44|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($36|0)<(1);
    if ($37) {
     $39 = (($36) + 1)|0;
     _luaD_reallocstack($L,$39);
     break;
    } else {
     $38 = $36 << 1;
     _luaD_reallocstack($L,$38);
     break;
    }
   }
  } while(0);
  $40 = HEAP32[$11>>2]|0;
  $41 = ((($40)) + 16|0);
  HEAP32[$11>>2] = $41;
  $42 = HEAP32[$4>>2]|0;
  $43 = (($42) + ($7)|0);
  $44 = $3;
  $45 = $44;
  $46 = HEAP32[$45>>2]|0;
  $47 = (($44) + 4)|0;
  $48 = $47;
  $49 = HEAP32[$48>>2]|0;
  $50 = $43;
  $51 = $50;
  HEAP32[$51>>2] = $46;
  $52 = (($50) + 4)|0;
  $53 = $52;
  HEAP32[$53>>2] = $49;
  $54 = HEAP32[$8>>2]|0;
  $55 = ((($43)) + 8|0);
  HEAP32[$55>>2] = $54;
  $$01 = $43;$$pre$phi29Z2D = $4;$$pre$phiZ2D = $4;
 }
 $56 = HEAP32[$$pre$phi29Z2D>>2]|0;
 $57 = $$01;
 $58 = (($57) - ($56))|0;
 $59 = HEAP32[$$01>>2]|0;
 $60 = ((($L)) + 24|0);
 $61 = HEAP32[$60>>2]|0;
 $62 = ((($L)) + 20|0);
 $63 = HEAP32[$62>>2]|0;
 $64 = ((($63)) + 12|0);
 HEAP32[$64>>2] = $61;
 $65 = ((($59)) + 6|0);
 $66 = HEAP8[$65>>0]|0;
 $67 = ($66<<24>>24)==(0);
 if (!($67)) {
  $288 = ((($L)) + 28|0);
  $289 = HEAP32[$288>>2]|0;
  $290 = ((($L)) + 8|0);
  $291 = HEAP32[$290>>2]|0;
  $292 = (($289) - ($291))|0;
  $293 = ($292|0)<(321);
  do {
   if ($293) {
    $294 = ((($L)) + 44|0);
    $295 = HEAP32[$294>>2]|0;
    $296 = ($295|0)<(20);
    if ($296) {
     $298 = (($295) + 20)|0;
     _luaD_reallocstack($L,$298);
     break;
    } else {
     $297 = $295 << 1;
     _luaD_reallocstack($L,$297);
     break;
    }
   }
  } while(0);
  $299 = HEAP32[$62>>2]|0;
  $300 = ((($L)) + 36|0);
  $301 = HEAP32[$300>>2]|0;
  $302 = ($299|0)==($301|0);
  if ($302) {
   $303 = ((($L)) + 48|0);
   $304 = HEAP32[$303>>2]|0;
   $305 = ($304|0)>(20000);
   if ($305) {
    _luaD_throw($L,5);
    // unreachable;
   }
   $306 = $304 << 1;
   $307 = ((($L)) + 40|0);
   $308 = HEAP32[$307>>2]|0;
   $309 = $306 | 1;
   $310 = ($309>>>0)<(178956971);
   if ($310) {
    $311 = ($304*24)|0;
    $312 = ($304*48)|0;
    $313 = (_luaM_realloc_($L,$308,$311,$312)|0);
    $315 = $313;
   } else {
    $314 = (_luaM_toobig($L)|0);
    $315 = $314;
   }
   HEAP32[$307>>2] = $315;
   HEAP32[$303>>2] = $306;
   $316 = HEAP32[$62>>2]|0;
   $317 = $308;
   $318 = (($316) - ($317))|0;
   $319 = (($318|0) / 24)&-1;
   $320 = (($315) + (($319*24)|0)|0);
   HEAP32[$62>>2] = $320;
   $321 = (($315) + (($306*24)|0)|0);
   $322 = ((($321)) + -24|0);
   HEAP32[$300>>2] = $322;
   $323 = ($306|0)>(20000);
   if ($323) {
    _luaG_runerror($L,517815,$vararg_buffer1);
    $$pre = HEAP32[$62>>2]|0;
    $$sink19 = $$pre;
   } else {
    $$sink19 = $320;
   }
  } else {
   $$sink19 = $299;
  }
  $324 = ((($$sink19)) + 24|0);
  HEAP32[$62>>2] = $324;
  $325 = HEAP32[$$pre$phiZ2D>>2]|0;
  $326 = (($325) + ($58)|0);
  $327 = ((($$sink19)) + 28|0);
  HEAP32[$327>>2] = $326;
  $328 = ((($326)) + 16|0);
  HEAP32[$324>>2] = $328;
  $329 = ((($L)) + 12|0);
  HEAP32[$329>>2] = $328;
  $330 = HEAP32[$290>>2]|0;
  $331 = ((($330)) + 320|0);
  $332 = ((($$sink19)) + 32|0);
  HEAP32[$332>>2] = $331;
  $333 = ((($$sink19)) + 40|0);
  HEAP32[$333>>2] = $nresults;
  $334 = ((($L)) + 56|0);
  $335 = HEAP16[$334>>1]|0;
  $336 = $335 & 1;
  $337 = ($336<<16>>16)==(0);
  $338 = $330;
  $339 = $325;
  $340 = $331;
  if ($337) {
   $373 = $324;
  } else {
   $341 = ((($L)) + 68|0);
   $342 = HEAP32[$341>>2]|0;
   $343 = ($342|0)==(0|0);
   if ($343) {
    $373 = $324;
   } else {
    $344 = ((($L)) + 57|0);
    $345 = ($335&65535)<(256);
    if ($345) {
     $373 = $324;
    } else {
     $346 = (($338) - ($339))|0;
     $347 = (($340) - ($339))|0;
     HEAP32[$ar$i10>>2] = 0;
     $348 = ((($ar$i10)) + 20|0);
     HEAP32[$348>>2] = -1;
     $349 = $324;
     $350 = ((($L)) + 40|0);
     $351 = HEAP32[$350>>2]|0;
     $352 = (($349) - ($351))|0;
     $353 = (($352|0) / 24)&-1;
     $354 = ((($ar$i10)) + 96|0);
     HEAP32[$354>>2] = $353;
     $355 = HEAP32[$288>>2]|0;
     $356 = (($355) - ($338))|0;
     $357 = ($356|0)<(321);
     do {
      if ($357) {
       $358 = ((($L)) + 44|0);
       $359 = HEAP32[$358>>2]|0;
       $360 = ($359|0)<(20);
       if ($360) {
        $362 = (($359) + 20)|0;
        _luaD_reallocstack($L,$362);
        break;
       } else {
        $361 = $359 << 1;
        _luaD_reallocstack($L,$361);
        break;
       }
      }
     } while(0);
     $363 = HEAP32[$290>>2]|0;
     $364 = ((($363)) + 320|0);
     $365 = HEAP32[$62>>2]|0;
     $366 = ((($365)) + 8|0);
     HEAP32[$366>>2] = $364;
     HEAP8[$344>>0] = 0;
     FUNCTION_TABLE_vii[$342 & 8191]($L,$ar$i10);
     HEAP8[$344>>0] = 1;
     $367 = HEAP32[$$pre$phiZ2D>>2]|0;
     $368 = (($367) + ($347)|0);
     $369 = HEAP32[$62>>2]|0;
     $370 = ((($369)) + 8|0);
     HEAP32[$370>>2] = $368;
     $371 = (($367) + ($346)|0);
     HEAP32[$290>>2] = $371;
     $373 = $369;
    }
   }
  }
  $372 = ((($373)) + 4|0);
  $374 = HEAP32[$372>>2]|0;
  $375 = HEAP32[$374>>2]|0;
  $376 = ((($375)) + 16|0);
  $377 = HEAP32[$376>>2]|0;
  $378 = (FUNCTION_TABLE_ii[$377 & 8191]($L)|0);
  $379 = ($378|0)<(0);
  if ($379) {
   $$1 = 2;
   STACKTOP = sp;return ($$1|0);
  }
  $380 = HEAP32[$290>>2]|0;
  $381 = (0 - ($378))|0;
  $382 = (($380) + ($381<<4)|0);
  (_luaD_poscall($L,$382)|0);
  $$1 = 1;
  STACKTOP = sp;return ($$1|0);
 }
 $68 = ((($59)) + 16|0);
 $69 = HEAP32[$68>>2]|0;
 $70 = ((($L)) + 28|0);
 $71 = HEAP32[$70>>2]|0;
 $72 = ((($L)) + 8|0);
 $73 = HEAP32[$72>>2]|0;
 $74 = (($71) - ($73))|0;
 $75 = ((($69)) + 75|0);
 $76 = HEAP8[$75>>0]|0;
 $77 = $76&255;
 $78 = $77 << 4;
 $79 = ($74|0)>($78|0);
 do {
  if (!($79)) {
   $80 = ((($L)) + 44|0);
   $81 = HEAP32[$80>>2]|0;
   $82 = ($81|0)<($77|0);
   if ($82) {
    $84 = (($81) + ($77))|0;
    _luaD_reallocstack($L,$84);
    break;
   } else {
    $83 = $81 << 1;
    _luaD_reallocstack($L,$83);
    break;
   }
  }
 } while(0);
 $85 = HEAP32[$$pre$phiZ2D>>2]|0;
 $86 = (($85) + ($58)|0);
 $87 = ((($69)) + 74|0);
 $88 = HEAP8[$87>>0]|0;
 $89 = ($88<<24>>24)==(0);
 if ($89) {
  $90 = ((($86)) + 16|0);
  $91 = HEAP32[$72>>2]|0;
  $92 = ((($69)) + 73|0);
  $93 = HEAP8[$92>>0]|0;
  $94 = $93&255;
  $95 = (($90) + ($94<<4)|0);
  $96 = ($91>>>0)>($95>>>0);
  if ($96) {
   HEAP32[$72>>2] = $95;
   $$12 = $86;$base$0 = $90;
  } else {
   $$12 = $86;$base$0 = $90;
  }
 } else {
  $97 = HEAP32[$72>>2]|0;
  $98 = $86;
  $99 = (($97) - ($98))|0;
  $100 = $99 >> 4;
  $101 = (($100) + -1)|0;
  $102 = ((($69)) + 73|0);
  $103 = HEAP8[$102>>0]|0;
  $104 = $103&255;
  $105 = ($104|0)<($100|0);
  $106 = $97;
  if ($105) {
   $$0$lcssa$i = $101;$383 = $106;$384 = $97;
  } else {
   $$05$i = $101;$108 = $106;
   while(1) {
    $107 = ((($108)) + 16|0);
    $109 = ((($108)) + 8|0);
    HEAP32[$109>>2] = 0;
    $110 = (($$05$i) + 1)|0;
    $exitcond11$i = ($110|0)==($104|0);
    if ($exitcond11$i) {
     break;
    } else {
     $$05$i = $110;$108 = $107;
    }
   }
   $111 = (($104) - ($101))|0;
   $scevgep$i = (($106) + ($111<<4)|0);
   HEAP32[$72>>2] = $scevgep$i;
   $112 = $scevgep$i;
   $$0$lcssa$i = $104;$383 = $scevgep$i;$384 = $112;
  }
  $113 = HEAP16[$87>>1]|0;
  $114 = $113 & 4;
  $115 = ($114<<16>>16)==(0);
  if ($115) {
   $165 = $383;$htab$0$i = 0;
  } else {
   $116 = ($113&65535) >>> 8;
   $117 = $116&255;
   $118 = (($$0$lcssa$i) - ($104))|0;
   $119 = ((($L)) + 16|0);
   $120 = HEAP32[$119>>2]|0;
   $121 = ((($120)) + 68|0);
   $122 = HEAP32[$121>>2]|0;
   $123 = ((($120)) + 64|0);
   $124 = HEAP32[$123>>2]|0;
   $125 = ($122>>>0)<($124>>>0);
   if ($125) {
    $128 = $384;$130 = $117;
   } else {
    _luaC_step($L);
    $$pre$i4 = HEAP8[$75>>0]|0;
    $$pre22 = HEAP32[$72>>2]|0;
    $128 = $$pre22;$130 = $$pre$i4;
   }
   $126 = HEAP32[$70>>2]|0;
   $127 = (($126) - ($128))|0;
   $129 = $130&255;
   $131 = $129 << 4;
   $132 = ($127|0)>($131|0);
   do {
    if (!($132)) {
     $133 = ((($L)) + 44|0);
     $134 = HEAP32[$133>>2]|0;
     $135 = ($134|0)<($129|0);
     if ($135) {
      $137 = (($134) + ($129))|0;
      _luaD_reallocstack($L,$137);
      break;
     } else {
      $136 = $134 << 1;
      _luaD_reallocstack($L,$136);
      break;
     }
    }
   } while(0);
   $138 = (_luaH_new($L,$118,1)|0);
   $139 = ($118|0)>(0);
   if ($139) {
    $140 = (0 - ($118))|0;
    $i$02$i = 0;
    while(1) {
     $141 = HEAP32[$72>>2]|0;
     $142 = (($141) + ($140<<4)|0);
     $143 = (($142) + ($i$02$i<<4)|0);
     $144 = (($i$02$i) + 1)|0;
     $145 = (_luaH_setnum($L,$138,$144)|0);
     $146 = $143;
     $147 = $146;
     $148 = HEAP32[$147>>2]|0;
     $149 = (($146) + 4)|0;
     $150 = $149;
     $151 = HEAP32[$150>>2]|0;
     $152 = $145;
     $153 = $152;
     HEAP32[$153>>2] = $148;
     $154 = (($152) + 4)|0;
     $155 = $154;
     HEAP32[$155>>2] = $151;
     $156 = ((($143)) + 8|0);
     $157 = HEAP32[$156>>2]|0;
     $158 = ((($145)) + 8|0);
     HEAP32[$158>>2] = $157;
     $exitcond10$i = ($144|0)==($118|0);
     if ($exitcond10$i) {
      break;
     } else {
      $i$02$i = $144;
     }
    }
   }
   $159 = (_luaS_newlstr($L,519423,1)|0);
   $160 = (_luaH_setstr($L,$138,$159)|0);
   $161 = (+($118|0));
   HEAPF64[$160>>3] = $161;
   $162 = ((($160)) + 8|0);
   HEAP32[$162>>2] = 3;
   $$pre23 = HEAP32[$72>>2]|0;
   $165 = $$pre23;$htab$0$i = $138;
  }
  $163 = (0 - ($$0$lcssa$i))|0;
  $164 = (($165) + ($163<<4)|0);
  $166 = ($103<<24>>24)==(0);
  if (!($166)) {
   $167 = ((($165)) + 16|0);
   HEAP32[$72>>2] = $167;
   $168 = $164;
   $169 = $168;
   $170 = HEAP32[$169>>2]|0;
   $171 = (($168) + 4)|0;
   $172 = $171;
   $173 = HEAP32[$172>>2]|0;
   $174 = $165;
   $175 = $174;
   HEAP32[$175>>2] = $170;
   $176 = (($174) + 4)|0;
   $177 = $176;
   HEAP32[$177>>2] = $173;
   $178 = (((($165) + ($163<<4)|0)) + 8|0);
   $179 = HEAP32[$178>>2]|0;
   $180 = ((($165)) + 8|0);
   HEAP32[$180>>2] = $179;
   HEAP32[$178>>2] = 0;
   $exitcond$i21 = ($103<<24>>24)==(1);
   if (!($exitcond$i21)) {
    $182 = 1;
    while(1) {
     $$pre12$i = HEAP32[$72>>2]|0;
     $181 = (($164) + ($182<<4)|0);
     $183 = ((($$pre12$i)) + 16|0);
     HEAP32[$72>>2] = $183;
     $184 = $181;
     $185 = $184;
     $186 = HEAP32[$185>>2]|0;
     $187 = (($184) + 4)|0;
     $188 = $187;
     $189 = HEAP32[$188>>2]|0;
     $190 = $$pre12$i;
     $191 = $190;
     HEAP32[$191>>2] = $186;
     $192 = (($190) + 4)|0;
     $193 = $192;
     HEAP32[$193>>2] = $189;
     $194 = ((($181)) + 8|0);
     $195 = HEAP32[$194>>2]|0;
     $196 = ((($$pre12$i)) + 8|0);
     HEAP32[$196>>2] = $195;
     HEAP32[$194>>2] = 0;
     $197 = (($182) + 1)|0;
     $exitcond$i = ($197|0)==($104|0);
     if ($exitcond$i) {
      break;
     } else {
      $182 = $197;
     }
    }
   }
  }
  $198 = ($htab$0$i|0)==(0|0);
  if (!($198)) {
   $199 = HEAP32[$72>>2]|0;
   $200 = ((($199)) + 16|0);
   HEAP32[$72>>2] = $200;
   HEAP32[$199>>2] = $htab$0$i;
   $201 = ((($199)) + 8|0);
   HEAP32[$201>>2] = 5;
  }
  $202 = HEAP32[$$pre$phiZ2D>>2]|0;
  $203 = (($202) + ($58)|0);
  $$12 = $203;$base$0 = $165;
 }
 $204 = HEAP32[$62>>2]|0;
 $205 = ((($L)) + 36|0);
 $206 = HEAP32[$205>>2]|0;
 $207 = ($204|0)==($206|0);
 if ($207) {
  $208 = ((($L)) + 48|0);
  $209 = HEAP32[$208>>2]|0;
  $210 = ($209|0)>(20000);
  if ($210) {
   _luaD_throw($L,5);
   // unreachable;
  }
  $211 = $209 << 1;
  $212 = ((($L)) + 40|0);
  $213 = HEAP32[$212>>2]|0;
  $214 = $211 | 1;
  $215 = ($214>>>0)<(178956971);
  if ($215) {
   $216 = ($209*24)|0;
   $217 = ($209*48)|0;
   $218 = (_luaM_realloc_($L,$213,$216,$217)|0);
   $220 = $218;
  } else {
   $219 = (_luaM_toobig($L)|0);
   $220 = $219;
  }
  HEAP32[$212>>2] = $220;
  HEAP32[$208>>2] = $211;
  $221 = HEAP32[$62>>2]|0;
  $222 = $213;
  $223 = (($221) - ($222))|0;
  $224 = (($223|0) / 24)&-1;
  $225 = (($220) + (($224*24)|0)|0);
  HEAP32[$62>>2] = $225;
  $226 = (($220) + (($211*24)|0)|0);
  $227 = ((($226)) + -24|0);
  HEAP32[$205>>2] = $227;
  $228 = ($211|0)>(20000);
  if ($228) {
   _luaG_runerror($L,517815,$vararg_buffer);
   $$pre24 = HEAP32[$62>>2]|0;
   $$sink = $$pre24;
  } else {
   $$sink = $225;
  }
 } else {
  $$sink = $204;
 }
 $229 = ((($$sink)) + 24|0);
 HEAP32[$62>>2] = $229;
 $230 = ((($$sink)) + 28|0);
 HEAP32[$230>>2] = $$12;
 HEAP32[$229>>2] = $base$0;
 $231 = ((($L)) + 12|0);
 HEAP32[$231>>2] = $base$0;
 $232 = HEAP8[$75>>0]|0;
 $233 = $232&255;
 $234 = (($base$0) + ($233<<4)|0);
 $235 = ((($$sink)) + 32|0);
 HEAP32[$235>>2] = $234;
 $236 = ((($69)) + 12|0);
 $237 = HEAP32[$236>>2]|0;
 HEAP32[$60>>2] = $237;
 $238 = ((($$sink)) + 44|0);
 HEAP32[$238>>2] = 0;
 $239 = ((($$sink)) + 40|0);
 HEAP32[$239>>2] = $nresults;
 $240 = HEAP32[$72>>2]|0;
 $241 = ($240>>>0)<($234>>>0);
 $242 = $237;
 if ($241) {
  $243 = HEAP32[$235>>2]|0;
  $st$020 = $240;
  while(1) {
   $244 = ((($st$020)) + 8|0);
   HEAP32[$244>>2] = 0;
   $245 = ((($st$020)) + 16|0);
   $246 = ($245>>>0)<($243>>>0);
   if ($246) {
    $st$020 = $245;
   } else {
    $$in = $243;
    break;
   }
  }
 } else {
  $$in = $234;
 }
 $247 = $$in;
 HEAP32[$72>>2] = $$in;
 $248 = ((($L)) + 56|0);
 $249 = HEAP16[$248>>1]|0;
 $250 = $249 & 1;
 $251 = ($250<<16>>16)==(0);
 $252 = $$in;
 if ($251) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $253 = ((($242)) + 4|0);
 HEAP32[$60>>2] = $253;
 $254 = ((($L)) + 68|0);
 $255 = HEAP32[$254>>2]|0;
 $256 = ($255|0)==(0|0);
 if ($256) {
  $287 = $253;
 } else {
  $257 = ((($L)) + 57|0);
  $258 = ($249&65535)<(256);
  if ($258) {
   $287 = $253;
  } else {
   $259 = HEAP32[$$pre$phi29Z2D>>2]|0;
   $260 = (($252) - ($259))|0;
   $261 = (($247) - ($259))|0;
   HEAP32[$ar$i>>2] = 0;
   $262 = ((($ar$i)) + 20|0);
   HEAP32[$262>>2] = -1;
   $263 = $229;
   $264 = ((($L)) + 40|0);
   $265 = HEAP32[$264>>2]|0;
   $266 = (($263) - ($265))|0;
   $267 = (($266|0) / 24)&-1;
   $268 = ((($ar$i)) + 96|0);
   HEAP32[$268>>2] = $267;
   $269 = HEAP32[$70>>2]|0;
   $270 = (($269) - ($252))|0;
   $271 = ($270|0)<(321);
   do {
    if ($271) {
     $272 = ((($L)) + 44|0);
     $273 = HEAP32[$272>>2]|0;
     $274 = ($273|0)<(20);
     if ($274) {
      $276 = (($273) + 20)|0;
      _luaD_reallocstack($L,$276);
      break;
     } else {
      $275 = $273 << 1;
      _luaD_reallocstack($L,$275);
      break;
     }
    }
   } while(0);
   $277 = HEAP32[$72>>2]|0;
   $278 = ((($277)) + 320|0);
   $279 = HEAP32[$62>>2]|0;
   $280 = ((($279)) + 8|0);
   HEAP32[$280>>2] = $278;
   HEAP8[$257>>0] = 0;
   FUNCTION_TABLE_vii[$255 & 8191]($L,$ar$i);
   HEAP8[$257>>0] = 1;
   $281 = HEAP32[$$pre$phiZ2D>>2]|0;
   $282 = (($281) + ($261)|0);
   $283 = HEAP32[$62>>2]|0;
   $284 = ((($283)) + 8|0);
   HEAP32[$284>>2] = $282;
   $285 = (($281) + ($260)|0);
   HEAP32[$72>>2] = $285;
   $$pre25 = HEAP32[$60>>2]|0;
   $287 = $$pre25;
  }
 }
 $286 = ((($287)) + -4|0);
 HEAP32[$60>>2] = $286;
 $$1 = 0;
 STACKTOP = sp;return ($$1|0);
}
function _luaD_poscall($L,$firstResult) {
 $L = $L|0;
 $firstResult = $firstResult|0;
 var $$0 = 0, $$15 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $ar$i$i = 0, $ar$i1$i = 0, $i$03 = 0, $i$03$lcssa = 0;
 var $i$11 = 0, $or$cond$i = 0, $res$04 = 0, $res$04$lcssa = 0, $res$1$lcssa = 0, $res$12 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 208|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ar$i1$i = sp + 100|0;
 $ar$i$i = sp;
 $0 = ((($L)) + 56|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = $1 & 2;
 $3 = ($2<<16>>16)==(0);
 if ($3) {
  $$phi$trans$insert = ((($L)) + 20|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $$0 = $firstResult;$$pre$phiZ2D = $$phi$trans$insert;$107 = $$pre;
 } else {
  $4 = ((($L)) + 32|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = $firstResult;
  $7 = (($6) - ($5))|0;
  $8 = ((($L)) + 68|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)==(0|0);
  $11 = $5;
  if ($10) {
   $144 = $11;$145 = $5;
  } else {
   $12 = ((($L)) + 57|0);
   $13 = ($1&65535)<(256);
   if ($13) {
    $144 = $11;$145 = $5;
   } else {
    $14 = ((($L)) + 8|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = (($15) - ($5))|0;
    $17 = ((($L)) + 20|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = ((($18)) + 8|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = (($20) - ($5))|0;
    HEAP32[$ar$i$i>>2] = 1;
    $22 = ((($ar$i$i)) + 20|0);
    HEAP32[$22>>2] = -1;
    $23 = $18;
    $24 = ((($L)) + 40|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = (($23) - ($25))|0;
    $27 = (($26|0) / 24)&-1;
    $28 = ((($ar$i$i)) + 96|0);
    HEAP32[$28>>2] = $27;
    $29 = ((($L)) + 28|0);
    $30 = HEAP32[$29>>2]|0;
    $31 = (($30) - ($15))|0;
    $32 = ($31|0)<(321);
    do {
     if ($32) {
      $33 = ((($L)) + 44|0);
      $34 = HEAP32[$33>>2]|0;
      $35 = ($34|0)<(20);
      if ($35) {
       $37 = (($34) + 20)|0;
       _luaD_reallocstack($L,$37);
       break;
      } else {
       $36 = $34 << 1;
       _luaD_reallocstack($L,$36);
       break;
      }
     }
    } while(0);
    $38 = HEAP32[$14>>2]|0;
    $39 = ((($38)) + 320|0);
    $40 = HEAP32[$17>>2]|0;
    $41 = ((($40)) + 8|0);
    HEAP32[$41>>2] = $39;
    HEAP8[$12>>0] = 0;
    FUNCTION_TABLE_vii[$9 & 8191]($L,$ar$i$i);
    HEAP8[$12>>0] = 1;
    $42 = HEAP32[$4>>2]|0;
    $43 = (($42) + ($21)|0);
    $44 = HEAP32[$17>>2]|0;
    $45 = ((($44)) + 8|0);
    HEAP32[$45>>2] = $43;
    $46 = (($42) + ($16)|0);
    HEAP32[$14>>2] = $46;
    $47 = $42;
    $144 = $42;$145 = $47;
   }
  }
  $48 = ((($L)) + 20|0);
  $49 = HEAP32[$48>>2]|0;
  $50 = ((($49)) + 4|0);
  $51 = HEAP32[$50>>2]|0;
  $52 = HEAP32[$51>>2]|0;
  $53 = ((($52)) + 6|0);
  $54 = HEAP8[$53>>0]|0;
  $55 = ($54<<24>>24)==(0);
  L14: do {
   if ($55) {
    $56 = HEAP16[$0>>1]|0;
    $57 = $56 & 2;
    $58 = ($57<<16>>16)==(0);
    if ($58) {
     $105 = $144;$146 = $49;
    } else {
     $59 = ($56&65535) >>> 8;
     $60 = $59&255;
     $61 = ((($L)) + 57|0);
     $62 = ((($L)) + 8|0);
     $63 = ((($ar$i1$i)) + 20|0);
     $64 = ((($ar$i1$i)) + 96|0);
     $65 = ((($L)) + 28|0);
     $66 = ((($L)) + 44|0);
     $147 = $144;$68 = $49;$75 = $60;$83 = $145;
     while(1) {
      $67 = ((($68)) + 20|0);
      $69 = HEAP32[$67>>2]|0;
      $70 = (($69) + -1)|0;
      HEAP32[$67>>2] = $70;
      $71 = ($69|0)==(0);
      if ($71) {
       $105 = $147;$146 = $68;
       break L14;
      }
      $72 = HEAP32[$8>>2]|0;
      $73 = ($72|0)==(0|0);
      $74 = ($75<<24>>24)==(0);
      $or$cond$i = $74 | $73;
      if ($or$cond$i) {
       $148 = $147;$149 = $83;$150 = $68;
      } else {
       $81 = HEAP32[$62>>2]|0;
       $82 = (($81) - ($83))|0;
       $84 = ((($68)) + 8|0);
       $85 = HEAP32[$84>>2]|0;
       $86 = (($85) - ($83))|0;
       HEAP32[$ar$i1$i>>2] = 4;
       HEAP32[$63>>2] = -1;
       HEAP32[$64>>2] = 0;
       $87 = HEAP32[$65>>2]|0;
       $88 = (($87) - ($81))|0;
       $89 = ($88|0)<(321);
       do {
        if ($89) {
         $90 = HEAP32[$66>>2]|0;
         $91 = ($90|0)<(20);
         if ($91) {
          $93 = (($90) + 20)|0;
          _luaD_reallocstack($L,$93);
          break;
         } else {
          $92 = $90 << 1;
          _luaD_reallocstack($L,$92);
          break;
         }
        }
       } while(0);
       $94 = HEAP32[$62>>2]|0;
       $95 = ((($94)) + 320|0);
       $96 = HEAP32[$48>>2]|0;
       $97 = ((($96)) + 8|0);
       HEAP32[$97>>2] = $95;
       HEAP8[$61>>0] = 0;
       FUNCTION_TABLE_vii[$72 & 8191]($L,$ar$i1$i);
       HEAP8[$61>>0] = 1;
       $98 = HEAP32[$4>>2]|0;
       $99 = (($98) + ($86)|0);
       $100 = HEAP32[$48>>2]|0;
       $101 = ((($100)) + 8|0);
       HEAP32[$101>>2] = $99;
       $102 = (($98) + ($82)|0);
       HEAP32[$62>>2] = $102;
       $103 = $98;
       $148 = $98;$149 = $103;$150 = $100;
      }
      $76 = HEAP16[$0>>1]|0;
      $77 = $76 & 2;
      $78 = ($77<<16>>16)==(0);
      $79 = ($76&65535) >>> 8;
      $80 = $79&255;
      if ($78) {
       $105 = $148;$146 = $150;
       break;
      } else {
       $147 = $148;$68 = $150;$75 = $80;$83 = $149;
      }
     }
    }
   } else {
    $105 = $144;$146 = $49;
   }
  } while(0);
  $104 = (($105) + ($7)|0);
  $$0 = $104;$$pre$phiZ2D = $48;$107 = $146;
 }
 $106 = ((($107)) + -24|0);
 HEAP32[$$pre$phiZ2D>>2] = $106;
 $108 = ((($107)) + 4|0);
 $109 = HEAP32[$108>>2]|0;
 $110 = ((($107)) + 16|0);
 $111 = HEAP32[$110>>2]|0;
 $112 = HEAP32[$106>>2]|0;
 $113 = ((($L)) + 12|0);
 HEAP32[$113>>2] = $112;
 $114 = ((($107)) + -12|0);
 $115 = HEAP32[$114>>2]|0;
 $116 = ((($L)) + 24|0);
 HEAP32[$116>>2] = $115;
 $117 = ((($L)) + 8|0);
 $118 = ($111|0)==(0);
 if ($118) {
  $res$1$lcssa = $109;
  HEAP32[$117>>2] = $res$1$lcssa;
  $143 = (($111) + 1)|0;
  STACKTOP = sp;return ($143|0);
 } else {
  $$15 = $$0;$i$03 = $111;$res$04 = $109;
 }
 while(1) {
  $120 = HEAP32[$117>>2]|0;
  $121 = ($$15>>>0)<($120>>>0);
  if (!($121)) {
   $i$03$lcssa = $i$03;$res$04$lcssa = $res$04;
   break;
  }
  $122 = ((($$15)) + 16|0);
  $123 = ((($res$04)) + 16|0);
  $124 = $$15;
  $125 = $124;
  $126 = HEAP32[$125>>2]|0;
  $127 = (($124) + 4)|0;
  $128 = $127;
  $129 = HEAP32[$128>>2]|0;
  $130 = $res$04;
  $131 = $130;
  HEAP32[$131>>2] = $126;
  $132 = (($130) + 4)|0;
  $133 = $132;
  HEAP32[$133>>2] = $129;
  $134 = ((($$15)) + 8|0);
  $135 = HEAP32[$134>>2]|0;
  $136 = ((($res$04)) + 8|0);
  HEAP32[$136>>2] = $135;
  $137 = (($i$03) + -1)|0;
  $138 = ($137|0)==(0);
  if ($138) {
   $res$1$lcssa = $123;
   label = 27;
   break;
  } else {
   $$15 = $122;$i$03 = $137;$res$04 = $123;
  }
 }
 if ((label|0) == 27) {
  HEAP32[$117>>2] = $res$1$lcssa;
  $143 = (($111) + 1)|0;
  STACKTOP = sp;return ($143|0);
 }
 $119 = ($i$03$lcssa|0)>(0);
 if ($119) {
  $i$11 = $i$03$lcssa;$res$12 = $res$04$lcssa;
 } else {
  $res$1$lcssa = $res$04$lcssa;
  HEAP32[$117>>2] = $res$1$lcssa;
  $143 = (($111) + 1)|0;
  STACKTOP = sp;return ($143|0);
 }
 while(1) {
  $139 = (($i$11) + -1)|0;
  $140 = ((($res$12)) + 16|0);
  $141 = ((($res$12)) + 8|0);
  HEAP32[$141>>2] = 0;
  $142 = ($i$11|0)>(1);
  if ($142) {
   $i$11 = $139;$res$12 = $140;
  } else {
   $res$1$lcssa = $140;
   break;
  }
 }
 HEAP32[$117>>2] = $res$1$lcssa;
 $143 = (($111) + 1)|0;
 STACKTOP = sp;return ($143|0);
}
function _luaD_call($L,$func,$nResults) {
 $L = $L|0;
 $func = $func|0;
 $nResults = $nResults|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($L)) + 52|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = (($1) + 1)<<16>>16;
 HEAP16[$0>>1] = $2;
 $3 = ($2&65535)>(199);
 do {
  if ($3) {
   $4 = ($2<<16>>16)==(200);
   if ($4) {
    _luaG_runerror($L,517830,$vararg_buffer);
    break;
   }
   $5 = ($2&65535)>(224);
   if ($5) {
    _luaD_throw($L,5);
    // unreachable;
   }
  }
 } while(0);
 $6 = (_luaD_precall($L,$func,$nResults)|0);
 $7 = ($6|0)==(0);
 if ($7) {
  _luaV_execute($L,1);
 }
 $8 = HEAP16[$0>>1]|0;
 $9 = (($8) + -1)<<16>>16;
 HEAP16[$0>>1] = $9;
 $10 = ((($L)) + 16|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($11)) + 68|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($11)) + 64|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($13>>>0)<($15>>>0);
 if ($16) {
  STACKTOP = sp;return;
 }
 _luaC_step($L);
 STACKTOP = sp;return;
}
function _lua_resume($L,$nargs) {
 $L = $L|0;
 $nargs = $nargs|0;
 var $$0 = 0, $$cast = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $9 = 0, $status$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 6|0);
 $1 = HEAP8[$0>>0]|0;
 switch ($1<<24>>24) {
 case 1:  {
  break;
 }
 case 0:  {
  $2 = ((($L)) + 20|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($L)) + 40|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ($3|0)==($5|0);
  if (!($6)) {
   $8 = $3;
   label = 4;
  }
  break;
 }
 default: {
  $$phi$trans$insert = ((($L)) + 20|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $8 = $$pre;
  label = 4;
 }
 }
 if ((label|0) == 4) {
  $7 = HEAP32[$8>>2]|0;
  $9 = ((($L)) + 8|0);
  HEAP32[$9>>2] = $7;
  $10 = (_luaS_newlstr($L,517847,37)|0);
  HEAP32[$7>>2] = $10;
  $11 = ((($7)) + 8|0);
  HEAP32[$11>>2] = 4;
  $12 = ((($L)) + 28|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = HEAP32[$9>>2]|0;
  $15 = (($13) - ($14))|0;
  $16 = ($15|0)<(17);
  do {
   if ($16) {
    $17 = ((($L)) + 44|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = ($18|0)<(1);
    if ($19) {
     $21 = (($18) + 1)|0;
     _luaD_reallocstack($L,$21);
     break;
    } else {
     $20 = $18 << 1;
     _luaD_reallocstack($L,$20);
     break;
    }
   }
  } while(0);
  $22 = HEAP32[$9>>2]|0;
  $23 = ((($22)) + 16|0);
  HEAP32[$9>>2] = $23;
  $$0 = 2;
  return ($$0|0);
 }
 $24 = ((($L)) + 52|0);
 $25 = HEAP16[$24>>1]|0;
 $26 = ($25&65535)>(199);
 if ($26) {
  $27 = ((($L)) + 20|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = HEAP32[$28>>2]|0;
  $30 = ((($L)) + 8|0);
  HEAP32[$30>>2] = $29;
  $31 = (_luaS_newlstr($L,517830,16)|0);
  HEAP32[$29>>2] = $31;
  $32 = ((($29)) + 8|0);
  HEAP32[$32>>2] = 4;
  $33 = ((($L)) + 28|0);
  $34 = HEAP32[$33>>2]|0;
  $35 = HEAP32[$30>>2]|0;
  $36 = (($34) - ($35))|0;
  $37 = ($36|0)<(17);
  do {
   if ($37) {
    $38 = ((($L)) + 44|0);
    $39 = HEAP32[$38>>2]|0;
    $40 = ($39|0)<(1);
    if ($40) {
     $42 = (($39) + 1)|0;
     _luaD_reallocstack($L,$42);
     break;
    } else {
     $41 = $39 << 1;
     _luaD_reallocstack($L,$41);
     break;
    }
   }
  } while(0);
  $43 = HEAP32[$30>>2]|0;
  $44 = ((($43)) + 16|0);
  HEAP32[$30>>2] = $44;
  $$0 = 2;
  return ($$0|0);
 }
 $45 = (($25) + 1)<<16>>16;
 HEAP16[$24>>1] = $45;
 $46 = ((($L)) + 54|0);
 HEAP16[$46>>1] = $45;
 $47 = ((($L)) + 8|0);
 $48 = HEAP32[$47>>2]|0;
 $49 = (0 - ($nargs))|0;
 $50 = (($48) + ($49<<4)|0);
 $51 = (_luaD_rawrunprotected($L,4290,$50)|0);
 $52 = ($51|0)==(0);
 if ($52) {
  $77 = HEAP8[$0>>0]|0;
  $78 = $77&255;
  $status$0 = $78;
 } else {
  $53 = $51&255;
  HEAP8[$0>>0] = $53;
  $54 = HEAP32[$47>>2]|0;
  switch ($51|0) {
  case 4:  {
   $55 = (_luaS_newlstr($L,518874,17)|0);
   HEAP32[$54>>2] = $55;
   $56 = ((($54)) + 8|0);
   HEAP32[$56>>2] = 4;
   break;
  }
  case 5:  {
   $57 = (_luaS_newlstr($L,517791,23)|0);
   HEAP32[$54>>2] = $57;
   $58 = ((($54)) + 8|0);
   HEAP32[$58>>2] = 4;
   break;
  }
  case 2: case 3:  {
   $59 = ((($54)) + -16|0);
   $60 = $59;
   $61 = $60;
   $62 = HEAP32[$61>>2]|0;
   $63 = (($60) + 4)|0;
   $64 = $63;
   $65 = HEAP32[$64>>2]|0;
   $66 = $54;
   $67 = $66;
   HEAP32[$67>>2] = $62;
   $68 = (($66) + 4)|0;
   $69 = $68;
   HEAP32[$69>>2] = $65;
   $70 = ((($54)) + -8|0);
   $71 = HEAP32[$70>>2]|0;
   $72 = ((($54)) + 8|0);
   HEAP32[$72>>2] = $71;
   break;
  }
  default: {
  }
  }
  $73 = ((($54)) + 16|0);
  HEAP32[$47>>2] = $73;
  $$cast = $73;
  $74 = ((($L)) + 20|0);
  $75 = HEAP32[$74>>2]|0;
  $76 = ((($75)) + 8|0);
  HEAP32[$76>>2] = $$cast;
  $status$0 = $51;
 }
 $79 = HEAP16[$24>>1]|0;
 $80 = (($79) + -1)<<16>>16;
 HEAP16[$24>>1] = $80;
 $$0 = $status$0;
 return ($$0|0);
}
function _resume($L,$ud) {
 $L = $L|0;
 $ud = $ud|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 20|0);
 $1 = ((($L)) + 6|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = ($2<<24>>24)==(0);
 do {
  if ($3) {
   $4 = ((($ud)) + -16|0);
   $5 = (_luaD_precall($L,$4,-1)|0);
   $6 = ($5|0)==(0);
   if (!($6)) {
    return;
   }
  } else {
   $7 = HEAP32[$0>>2]|0;
   HEAP8[$1>>0] = 0;
   $8 = ((($7)) + 4|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = HEAP32[$9>>2]|0;
   $11 = ((($10)) + 6|0);
   $12 = HEAP8[$11>>0]|0;
   $13 = ($12<<24>>24)==(0);
   if ($13) {
    $20 = HEAP32[$7>>2]|0;
    $21 = ((($L)) + 12|0);
    HEAP32[$21>>2] = $20;
    break;
   }
   $14 = (_luaD_poscall($L,$ud)|0);
   $15 = ($14|0)==(0);
   if (!($15)) {
    $16 = HEAP32[$0>>2]|0;
    $17 = ((($16)) + 8|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = ((($L)) + 8|0);
    HEAP32[$19>>2] = $18;
   }
  }
 } while(0);
 $22 = HEAP32[$0>>2]|0;
 $23 = ((($L)) + 40|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = (($22) - ($24))|0;
 $26 = (($25|0) / 24)&-1;
 _luaV_execute($L,$26);
 return;
}
function _lua_yield($L,$nresults) {
 $L = $L|0;
 $nresults = $nresults|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($L)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1&65535;
 $3 = $1 >>> 16;
 $4 = $3&65535;
 $5 = ($2&65535)>($4&65535);
 if ($5) {
  _luaG_runerror($L,517885,$vararg_buffer);
 }
 $6 = ((($L)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (0 - ($nresults))|0;
 $9 = (($7) + ($8<<4)|0);
 $10 = ((($L)) + 12|0);
 HEAP32[$10>>2] = $9;
 $11 = ((($L)) + 6|0);
 HEAP8[$11>>0] = 1;
 STACKTOP = sp;return -1;
}
function _luaD_pcall($L,$func,$u,$old_top,$ef) {
 $L = $L|0;
 $func = $func|0;
 $u = $u|0;
 $old_top = $old_top|0;
 $ef = $ef|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 52|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = ((($L)) + 20|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($L)) + 40|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($L)) + 57|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = ((($L)) + 116|0);
 $9 = HEAP32[$8>>2]|0;
 HEAP32[$8>>2] = $ef;
 $10 = (_luaD_rawrunprotected($L,$func,$u)|0);
 $11 = ($10|0)==(0);
 if ($11) {
  HEAP32[$8>>2] = $9;
  return ($10|0);
 }
 $12 = (($3) - ($5))|0;
 $13 = ((($L)) + 32|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($14) + ($old_top)|0);
 _luaF_close($L,$15);
 switch ($10|0) {
 case 4:  {
  $16 = (_luaS_newlstr($L,518874,17)|0);
  HEAP32[$15>>2] = $16;
  $17 = ((($15)) + 8|0);
  HEAP32[$17>>2] = 4;
  break;
 }
 case 5:  {
  $18 = (_luaS_newlstr($L,517791,23)|0);
  HEAP32[$15>>2] = $18;
  $19 = ((($15)) + 8|0);
  HEAP32[$19>>2] = 4;
  break;
 }
 case 2: case 3:  {
  $20 = ((($L)) + 8|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = ((($21)) + -16|0);
  $23 = $22;
  $24 = $23;
  $25 = HEAP32[$24>>2]|0;
  $26 = (($23) + 4)|0;
  $27 = $26;
  $28 = HEAP32[$27>>2]|0;
  $29 = $15;
  $30 = $29;
  HEAP32[$30>>2] = $25;
  $31 = (($29) + 4)|0;
  $32 = $31;
  HEAP32[$32>>2] = $28;
  $33 = ((($21)) + -8|0);
  $34 = HEAP32[$33>>2]|0;
  $35 = ((($15)) + 8|0);
  HEAP32[$35>>2] = $34;
  break;
 }
 default: {
 }
 }
 $36 = ((($15)) + 16|0);
 $37 = ((($L)) + 8|0);
 HEAP32[$37>>2] = $36;
 HEAP16[$0>>1] = $1;
 $38 = HEAP32[$4>>2]|0;
 $39 = (($38) + ($12)|0);
 HEAP32[$2>>2] = $39;
 $40 = HEAP32[$39>>2]|0;
 $41 = ((($L)) + 12|0);
 HEAP32[$41>>2] = $40;
 $42 = ((($39)) + 12|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = ((($L)) + 24|0);
 HEAP32[$44>>2] = $43;
 HEAP8[$6>>0] = $7;
 $45 = ((($L)) + 48|0);
 $46 = HEAP32[$45>>2]|0;
 $47 = ($46|0)>(20000);
 $48 = $38;
 if (!($47)) {
  HEAP32[$8>>2] = $9;
  return ($10|0);
 }
 $49 = (($12|0) / 24)&-1;
 $50 = (($49) + 1)|0;
 $51 = ($50|0)<(20000);
 if (!($51)) {
  HEAP32[$8>>2] = $9;
  return ($10|0);
 }
 $52 = ($46*24)|0;
 $53 = (_luaM_realloc_($L,$38,$52,480000)|0);
 HEAP32[$4>>2] = $53;
 HEAP32[$45>>2] = 20000;
 $54 = HEAP32[$2>>2]|0;
 $55 = (($54) - ($48))|0;
 $56 = (($55|0) / 24)&-1;
 $57 = (($53) + (($56*24)|0)|0);
 HEAP32[$2>>2] = $57;
 $58 = ((($53)) + 479976|0);
 $59 = ((($L)) + 36|0);
 HEAP32[$59>>2] = $58;
 HEAP32[$8>>2] = $9;
 return ($10|0);
}
function _luaD_protectedparser($L,$z,$name) {
 $L = $L|0;
 $z = $z|0;
 $name = $name|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $p = sp;
 HEAP32[$p>>2] = $z;
 $0 = ((($p)) + 16|0);
 HEAP32[$0>>2] = $name;
 $1 = ((($p)) + 4|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($p)) + 12|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($L)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($L)) + 32|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4) - ($6))|0;
 $8 = ((($L)) + 116|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (_luaD_pcall($L,4291,$p,$7,$9)|0);
 $11 = HEAP32[$1>>2]|0;
 $12 = HEAP32[$2>>2]|0;
 (_luaM_realloc_($L,$11,$12,0)|0);
 STACKTOP = sp;return ($10|0);
}
function _f_parser($L,$ud) {
 $L = $L|0;
 $ud = $ud|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$ud>>2]|0;
 $1 = (_luaZ_lookahead($0)|0);
 $2 = ((($L)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 68|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($3)) + 64|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($5>>>0)<($7>>>0);
 if (!($8)) {
  _luaC_step($L);
 }
 $9 = ($1|0)==(27);
 $10 = $9 ? 4292 : 4293;
 $11 = HEAP32[$ud>>2]|0;
 $12 = ((($ud)) + 4|0);
 $13 = ((($ud)) + 16|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (FUNCTION_TABLE_iiiii[$10 & 8191]($L,$11,$12,$14)|0);
 $16 = ((($15)) + 72|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = $17&255;
 $19 = ((($L)) + 72|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (_luaF_newLclosure($L,$18,$20)|0);
 $22 = ((($21)) + 16|0);
 HEAP32[$22>>2] = $15;
 $23 = HEAP8[$16>>0]|0;
 $24 = ($23<<24>>24)==(0);
 if (!($24)) {
  $i$01 = 0;
  while(1) {
   $25 = (_luaF_newupval($L)|0);
   $26 = (((($21)) + 20|0) + ($i$01<<2)|0);
   HEAP32[$26>>2] = $25;
   $27 = (($i$01) + 1)|0;
   $28 = HEAP8[$16>>0]|0;
   $29 = $28&255;
   $30 = ($27|0)<($29|0);
   if ($30) {
    $i$01 = $27;
   } else {
    break;
   }
  }
 }
 $31 = ((($L)) + 8|0);
 $32 = HEAP32[$31>>2]|0;
 HEAP32[$32>>2] = $21;
 $33 = ((($32)) + 8|0);
 HEAP32[$33>>2] = 6;
 $34 = ((($L)) + 28|0);
 $35 = HEAP32[$34>>2]|0;
 $36 = HEAP32[$31>>2]|0;
 $37 = (($35) - ($36))|0;
 $38 = ($37|0)<(17);
 if (!($38)) {
  $44 = HEAP32[$31>>2]|0;
  $45 = ((($44)) + 16|0);
  HEAP32[$31>>2] = $45;
  return;
 }
 $39 = ((($L)) + 44|0);
 $40 = HEAP32[$39>>2]|0;
 $41 = ($40|0)<(1);
 if ($41) {
  $43 = (($40) + 1)|0;
  _luaD_reallocstack($L,$43);
  $44 = HEAP32[$31>>2]|0;
  $45 = ((($44)) + 16|0);
  HEAP32[$31>>2] = $45;
  return;
 } else {
  $42 = $40 << 1;
  _luaD_reallocstack($L,$42);
  $44 = HEAP32[$31>>2]|0;
  $45 = ((($44)) + 16|0);
  HEAP32[$31>>2] = $45;
  return;
 }
}
function _luaU_dump($L,$f,$w,$data,$strip) {
 $L = $L|0;
 $f = $f|0;
 $w = $w|0;
 $data = $data|0;
 $strip = $strip|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $D = 0, $h$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $h$i = sp + 20|0;
 $D = sp;
 HEAP32[$D>>2] = $L;
 $0 = ((($D)) + 4|0);
 HEAP32[$0>>2] = $w;
 $1 = ((($D)) + 8|0);
 HEAP32[$1>>2] = $data;
 $2 = ((($D)) + 12|0);
 HEAP32[$2>>2] = $strip;
 $3 = ((($D)) + 16|0);
 _luaU_header($h$i);
 $4 = (FUNCTION_TABLE_iiiii[$w & 8191]($L,$h$i,12,$data)|0);
 HEAP32[$3>>2] = $4;
 _DumpFunction($f,0,$D);
 $5 = HEAP32[$3>>2]|0;
 STACKTOP = sp;return ($5|0);
}
function _DumpFunction($f,$p,$D) {
 $f = $f|0;
 $p = $p|0;
 $D = $D|0;
 var $$ph = 0, $$phi$trans$insert = 0, $$phi$trans$insert28 = 0, $$pre = 0, $$pre$i = 0, $$pre$phiZ2D = 0, $$pre29 = 0, $$pre30 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0.0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0;
 var $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0;
 var $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0;
 var $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0;
 var $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0;
 var $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0;
 var $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0;
 var $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0;
 var $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0;
 var $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $exitcond$i = 0, $exitcond17$i = 0;
 var $exitcond27 = 0, $i$0$i25 = 0, $i$014$i = 0, $i$1$i24 = 0, $i$113$i = 0, $or$cond = 0, $size$i = 0, $size$i$i = 0, $size$i$i16 = 0, $size$i7$i = 0, $size1$i = 0, $size1$i$i = 0, $size1$i$i17 = 0, $size1$i8$i = 0, $x$i = 0, $x$i$i = 0, $x$i1$i = 0, $x$i11 = 0, $x$i5 = 0, $x$i8 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 76|0;
 $size$i$i16 = sp + 72|0;
 $size1$i$i17 = sp + 68|0;
 $1 = sp;
 $x$i1$i = sp + 85|0;
 $x$i$i = sp + 84|0;
 $2 = sp + 64|0;
 $3 = sp + 60|0;
 $x$i11 = sp + 83|0;
 $x$i8 = sp + 82|0;
 $x$i5 = sp + 81|0;
 $x$i = sp + 80|0;
 $4 = sp + 56|0;
 $5 = sp + 52|0;
 $size$i = sp + 48|0;
 $size1$i = sp + 44|0;
 $size$i7$i = sp + 40|0;
 $size1$i8$i = sp + 36|0;
 $6 = sp + 32|0;
 $7 = sp + 28|0;
 $8 = sp + 24|0;
 $size$i$i = sp + 20|0;
 $size1$i$i = sp + 16|0;
 $9 = sp + 12|0;
 $10 = sp + 8|0;
 $11 = ((($f)) + 32|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)==($p|0);
 if ($13) {
  label = 3;
 } else {
  $14 = ((($D)) + 12|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ($15|0)!=(0);
  $17 = ($12|0)==(0|0);
  $or$cond = $17 | $16;
  if ($or$cond) {
   label = 3;
  } else {
   $27 = ((($12)) + 16|0);
   $28 = ((($12)) + 12|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = (($29) + 1)|0;
   HEAP32[$size1$i>>2] = $30;
   $31 = ((($D)) + 16|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = ($32|0)==(0);
   if ($33) {
    $34 = ((($D)) + 4|0);
    $35 = HEAP32[$34>>2]|0;
    $36 = HEAP32[$D>>2]|0;
    $37 = ((($D)) + 8|0);
    $38 = HEAP32[$37>>2]|0;
    $39 = (FUNCTION_TABLE_iiiii[$35 & 8191]($36,$size1$i,4,$38)|0);
    HEAP32[$31>>2] = $39;
    $40 = ($39|0)==(0);
    if ($40) {
     $41 = HEAP32[$size1$i>>2]|0;
     $42 = HEAP32[$34>>2]|0;
     $43 = HEAP32[$D>>2]|0;
     $44 = HEAP32[$37>>2]|0;
     $45 = (FUNCTION_TABLE_iiiii[$42 & 8191]($43,$27,$41,$44)|0);
     HEAP32[$31>>2] = $45;
     $333 = $45;
    } else {
     $333 = $39;
    }
   } else {
    $333 = $32;
   }
   $$pre$phiZ2D = $31;$49 = $333;
  }
 }
 if ((label|0) == 3) {
  HEAP32[$size$i>>2] = 0;
  $18 = ((($D)) + 16|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = ($19|0)==(0);
  if ($20) {
   $21 = ((($D)) + 4|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = HEAP32[$D>>2]|0;
   $24 = ((($D)) + 8|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = (FUNCTION_TABLE_iiiii[$22 & 8191]($23,$size$i,4,$25)|0);
   HEAP32[$18>>2] = $26;
   $332 = $26;
  } else {
   $332 = $19;
  }
  $$pre$phiZ2D = $18;$49 = $332;
 }
 $46 = ((($f)) + 60|0);
 $47 = HEAP32[$46>>2]|0;
 HEAP32[$5>>2] = $47;
 $48 = ($49|0)==(0);
 if ($48) {
  $50 = ((($D)) + 4|0);
  $51 = HEAP32[$50>>2]|0;
  $52 = HEAP32[$D>>2]|0;
  $53 = ((($D)) + 8|0);
  $54 = HEAP32[$53>>2]|0;
  $55 = (FUNCTION_TABLE_iiiii[$51 & 8191]($52,$5,4,$54)|0);
  HEAP32[$$pre$phiZ2D>>2] = $55;
  $59 = $55;
 } else {
  $59 = $49;
 }
 $56 = ((($f)) + 64|0);
 $57 = HEAP32[$56>>2]|0;
 HEAP32[$4>>2] = $57;
 $58 = ($59|0)==(0);
 if ($58) {
  $60 = ((($D)) + 4|0);
  $61 = HEAP32[$60>>2]|0;
  $62 = HEAP32[$D>>2]|0;
  $63 = ((($D)) + 8|0);
  $64 = HEAP32[$63>>2]|0;
  $65 = (FUNCTION_TABLE_iiiii[$61 & 8191]($62,$4,4,$64)|0);
  HEAP32[$$pre$phiZ2D>>2] = $65;
  $70 = $65;
 } else {
  $70 = $59;
 }
 $66 = ((($f)) + 72|0);
 $67 = HEAP16[$66>>1]|0;
 $68 = $67&255;
 HEAP8[$x$i>>0] = $68;
 $69 = ($70|0)==(0);
 if ($69) {
  $73 = ((($D)) + 4|0);
  $74 = HEAP32[$73>>2]|0;
  $75 = HEAP32[$D>>2]|0;
  $76 = ((($D)) + 8|0);
  $77 = HEAP32[$76>>2]|0;
  $78 = (FUNCTION_TABLE_iiiii[$74 & 8191]($75,$x$i,1,$77)|0);
  HEAP32[$$pre$phiZ2D>>2] = $78;
  $$phi$trans$insert = ((($f)) + 73|0);
  $$pre = HEAP8[$$phi$trans$insert>>0]|0;
  HEAP8[$x$i5>>0] = $$pre;
  $79 = ($78|0)==(0);
  if ($79) {
   $80 = ((($D)) + 4|0);
   $81 = HEAP32[$80>>2]|0;
   $82 = HEAP32[$D>>2]|0;
   $83 = ((($D)) + 8|0);
   $84 = HEAP32[$83>>2]|0;
   $85 = (FUNCTION_TABLE_iiiii[$81 & 8191]($82,$x$i5,1,$84)|0);
   HEAP32[$$pre$phiZ2D>>2] = $85;
   $90 = $85;
  } else {
   $90 = $78;
  }
 } else {
  $71 = ($67&65535) >>> 8;
  $72 = $71&255;
  HEAP8[$x$i5>>0] = $72;
  $90 = $70;
 }
 $86 = ((($f)) + 74|0);
 $87 = HEAP16[$86>>1]|0;
 $88 = $87&255;
 HEAP8[$x$i8>>0] = $88;
 $89 = ($90|0)==(0);
 if ($89) {
  $93 = ((($D)) + 4|0);
  $94 = HEAP32[$93>>2]|0;
  $95 = HEAP32[$D>>2]|0;
  $96 = ((($D)) + 8|0);
  $97 = HEAP32[$96>>2]|0;
  $98 = (FUNCTION_TABLE_iiiii[$94 & 8191]($95,$x$i8,1,$97)|0);
  HEAP32[$$pre$phiZ2D>>2] = $98;
  $$phi$trans$insert28 = ((($f)) + 75|0);
  $$pre29 = HEAP8[$$phi$trans$insert28>>0]|0;
  HEAP8[$x$i11>>0] = $$pre29;
  $99 = ($98|0)==(0);
  if ($99) {
   $100 = ((($D)) + 4|0);
   $101 = HEAP32[$100>>2]|0;
   $102 = HEAP32[$D>>2]|0;
   $103 = ((($D)) + 8|0);
   $104 = HEAP32[$103>>2]|0;
   $105 = (FUNCTION_TABLE_iiiii[$101 & 8191]($102,$x$i11,1,$104)|0);
   HEAP32[$$pre$phiZ2D>>2] = $105;
   $111 = $105;
  } else {
   $111 = $98;
  }
 } else {
  $91 = ($87&65535) >>> 8;
  $92 = $91&255;
  HEAP8[$x$i11>>0] = $92;
  $111 = $90;
 }
 $106 = ((($f)) + 12|0);
 $107 = HEAP32[$106>>2]|0;
 $108 = ((($f)) + 44|0);
 $109 = HEAP32[$108>>2]|0;
 HEAP32[$3>>2] = $109;
 $110 = ($111|0)==(0);
 if ($110) {
  $112 = ((($D)) + 4|0);
  $113 = HEAP32[$112>>2]|0;
  $114 = HEAP32[$D>>2]|0;
  $115 = ((($D)) + 8|0);
  $116 = HEAP32[$115>>2]|0;
  $117 = (FUNCTION_TABLE_iiiii[$113 & 8191]($114,$3,4,$116)|0);
  HEAP32[$$pre$phiZ2D>>2] = $117;
  $118 = ($117|0)==(0);
  if ($118) {
   $121 = $109 << 2;
   $122 = HEAP32[$112>>2]|0;
   $123 = HEAP32[$D>>2]|0;
   $124 = HEAP32[$115>>2]|0;
   $125 = (FUNCTION_TABLE_iiiii[$122 & 8191]($123,$107,$121,$124)|0);
   HEAP32[$$pre$phiZ2D>>2] = $125;
   $126 = ((($f)) + 40|0);
   $127 = HEAP32[$126>>2]|0;
   HEAP32[$2>>2] = $127;
   $128 = ($125|0)==(0);
   if ($128) {
    $129 = ((($D)) + 4|0);
    $130 = HEAP32[$129>>2]|0;
    $131 = HEAP32[$D>>2]|0;
    $132 = ((($D)) + 8|0);
    $133 = HEAP32[$132>>2]|0;
    $134 = (FUNCTION_TABLE_iiiii[$130 & 8191]($131,$2,4,$133)|0);
    HEAP32[$$pre$phiZ2D>>2] = $134;
    $136 = $127;$334 = $2;$335 = $134;
   } else {
    $136 = $127;$334 = $2;$335 = $125;
   }
  } else {
   $$ph = $117;
   label = 25;
  }
 } else {
  $$ph = $111;
  label = 25;
 }
 if ((label|0) == 25) {
  $119 = ((($f)) + 40|0);
  $120 = HEAP32[$119>>2]|0;
  HEAP32[$2>>2] = $120;
  $136 = $120;$334 = $2;$335 = $$ph;
 }
 $135 = ($136|0)>(0);
 L38: do {
  if ($135) {
   $137 = ((($f)) + 8|0);
   $138 = ((($D)) + 4|0);
   $139 = ((($D)) + 8|0);
   $146 = $335;$i$0$i25 = 0;
   while(1) {
    $140 = HEAP32[$137>>2]|0;
    $141 = (($140) + ($i$0$i25<<4)|0);
    $142 = (((($140) + ($i$0$i25<<4)|0)) + 8|0);
    $143 = HEAP32[$142>>2]|0;
    $144 = $143&255;
    HEAP8[$x$i$i>>0] = $144;
    $145 = ($146|0)==(0);
    if ($145) {
     $147 = HEAP32[$138>>2]|0;
     $148 = HEAP32[$D>>2]|0;
     $149 = HEAP32[$139>>2]|0;
     $150 = (FUNCTION_TABLE_iiiii[$147 & 8191]($148,$x$i$i,1,$149)|0);
     HEAP32[$$pre$phiZ2D>>2] = $150;
     $$pre30 = HEAP32[$142>>2]|0;
     $151 = $$pre30;$155 = $150;
    } else {
     $151 = $143;$155 = $146;
    }
    L45: do {
     switch ($151|0) {
     case 4:  {
      $166 = HEAP32[$141>>2]|0;
      $167 = ($166|0)==(0|0);
      if ($167) {
       HEAP32[$size$i$i16>>2] = 0;
       $168 = ($155|0)==(0);
       if ($168) {
        $169 = HEAP32[$138>>2]|0;
        $170 = HEAP32[$D>>2]|0;
        $171 = HEAP32[$139>>2]|0;
        $172 = (FUNCTION_TABLE_iiiii[$169 & 8191]($170,$size$i$i16,4,$171)|0);
        HEAP32[$$pre$phiZ2D>>2] = $172;
        $339 = $172;
       } else {
        $339 = $155;
       }
       $336 = $339;
       break L45;
      }
      $173 = ((($166)) + 16|0);
      $174 = ((($166)) + 12|0);
      $175 = HEAP32[$174>>2]|0;
      $176 = (($175) + 1)|0;
      HEAP32[$size1$i$i17>>2] = $176;
      $177 = ($155|0)==(0);
      if ($177) {
       $178 = HEAP32[$138>>2]|0;
       $179 = HEAP32[$D>>2]|0;
       $180 = HEAP32[$139>>2]|0;
       $181 = (FUNCTION_TABLE_iiiii[$178 & 8191]($179,$size1$i$i17,4,$180)|0);
       HEAP32[$$pre$phiZ2D>>2] = $181;
       $182 = ($181|0)==(0);
       if ($182) {
        $183 = HEAP32[$size1$i$i17>>2]|0;
        $184 = HEAP32[$138>>2]|0;
        $185 = HEAP32[$D>>2]|0;
        $186 = HEAP32[$139>>2]|0;
        $187 = (FUNCTION_TABLE_iiiii[$184 & 8191]($185,$173,$183,$186)|0);
        HEAP32[$$pre$phiZ2D>>2] = $187;
        $340 = $187;
       } else {
        $340 = $181;
       }
      } else {
       $340 = $155;
      }
      $336 = $340;
      break;
     }
     case 1:  {
      $152 = HEAP32[$141>>2]|0;
      $153 = $152&255;
      HEAP8[$x$i1$i>>0] = $153;
      $154 = ($155|0)==(0);
      if ($154) {
       $156 = HEAP32[$138>>2]|0;
       $157 = HEAP32[$D>>2]|0;
       $158 = HEAP32[$139>>2]|0;
       $159 = (FUNCTION_TABLE_iiiii[$156 & 8191]($157,$x$i1$i,1,$158)|0);
       HEAP32[$$pre$phiZ2D>>2] = $159;
       $337 = $159;
      } else {
       $337 = $155;
      }
      $336 = $337;
      break;
     }
     case 3:  {
      $160 = +HEAPF64[$141>>3];
      HEAPF64[$1>>3] = $160;
      $161 = ($155|0)==(0);
      if ($161) {
       $162 = HEAP32[$138>>2]|0;
       $163 = HEAP32[$D>>2]|0;
       $164 = HEAP32[$139>>2]|0;
       $165 = (FUNCTION_TABLE_iiiii[$162 & 8191]($163,$1,8,$164)|0);
       HEAP32[$$pre$phiZ2D>>2] = $165;
       $338 = $165;
      } else {
       $338 = $155;
      }
      $336 = $338;
      break;
     }
     default: {
      $336 = $155;
     }
     }
    } while(0);
    $188 = (($i$0$i25) + 1)|0;
    $exitcond27 = ($188|0)==($136|0);
    if ($exitcond27) {
     $192 = $336;
     break L38;
    } else {
     $146 = $336;$i$0$i25 = $188;
    }
   }
  } else {
   $192 = $335;
  }
 } while(0);
 $189 = ((($f)) + 52|0);
 $190 = HEAP32[$189>>2]|0;
 HEAP32[$0>>2] = $190;
 $191 = ($192|0)==(0);
 if ($191) {
  $193 = ((($D)) + 4|0);
  $194 = HEAP32[$193>>2]|0;
  $195 = HEAP32[$D>>2]|0;
  $196 = ((($D)) + 8|0);
  $197 = HEAP32[$196>>2]|0;
  $198 = (FUNCTION_TABLE_iiiii[$194 & 8191]($195,$0,4,$197)|0);
  HEAP32[$$pre$phiZ2D>>2] = $198;
 }
 $199 = ($190|0)>(0);
 if ($199) {
  $200 = ((($f)) + 16|0);
  $i$1$i24 = 0;
  while(1) {
   $201 = HEAP32[$200>>2]|0;
   $202 = (($201) + ($i$1$i24<<2)|0);
   $203 = HEAP32[$202>>2]|0;
   $204 = HEAP32[$11>>2]|0;
   _DumpFunction($203,$204,$D);
   $205 = (($i$1$i24) + 1)|0;
   $exitcond = ($205|0)==($190|0);
   if ($exitcond) {
    break;
   } else {
    $i$1$i24 = $205;
   }
  }
 }
 $206 = ((($D)) + 12|0);
 $207 = HEAP32[$206>>2]|0;
 $208 = ($207|0)==(0);
 if ($208) {
  $209 = ((($f)) + 48|0);
  $210 = HEAP32[$209>>2]|0;
  $213 = $210;
 } else {
  $213 = 0;
 }
 $211 = ((($f)) + 20|0);
 $212 = HEAP32[$211>>2]|0;
 HEAP32[$10>>2] = $213;
 $214 = HEAP32[$$pre$phiZ2D>>2]|0;
 $215 = ($214|0)==(0);
 if ($215) {
  $216 = ((($D)) + 4|0);
  $217 = HEAP32[$216>>2]|0;
  $218 = HEAP32[$D>>2]|0;
  $219 = ((($D)) + 8|0);
  $220 = HEAP32[$219>>2]|0;
  $221 = (FUNCTION_TABLE_iiiii[$217 & 8191]($218,$10,4,$220)|0);
  HEAP32[$$pre$phiZ2D>>2] = $221;
  $222 = ($221|0)==(0);
  if ($222) {
   $223 = $213 << 2;
   $224 = HEAP32[$216>>2]|0;
   $225 = HEAP32[$D>>2]|0;
   $226 = HEAP32[$219>>2]|0;
   $227 = (FUNCTION_TABLE_iiiii[$224 & 8191]($225,$212,$223,$226)|0);
   HEAP32[$$pre$phiZ2D>>2] = $227;
   $234 = $227;
  } else {
   $234 = $221;
  }
 } else {
  $234 = $214;
 }
 $228 = HEAP32[$206>>2]|0;
 $229 = ($228|0)==(0);
 if ($229) {
  $230 = ((($f)) + 56|0);
  $231 = HEAP32[$230>>2]|0;
  $232 = $231;
 } else {
  $232 = 0;
 }
 HEAP32[$9>>2] = $232;
 $233 = ($234|0)==(0);
 if ($233) {
  $235 = ((($D)) + 4|0);
  $236 = HEAP32[$235>>2]|0;
  $237 = HEAP32[$D>>2]|0;
  $238 = ((($D)) + 8|0);
  $239 = HEAP32[$238>>2]|0;
  $240 = (FUNCTION_TABLE_iiiii[$236 & 8191]($237,$9,4,$239)|0);
  HEAP32[$$pre$phiZ2D>>2] = $240;
  $341 = $240;
 } else {
  $341 = $234;
 }
 $241 = ($232|0)>(0);
 if ($241) {
  $242 = ((($f)) + 24|0);
  $243 = ((($D)) + 4|0);
  $244 = ((($D)) + 8|0);
  $250 = $341;$i$014$i = 0;
  while(1) {
   $245 = HEAP32[$242>>2]|0;
   $246 = (($245) + (($i$014$i*12)|0)|0);
   $247 = HEAP32[$246>>2]|0;
   $248 = ($247|0)==(0|0);
   if ($248) {
    HEAP32[$size$i$i>>2] = 0;
    $249 = ($250|0)==(0);
    if ($249) {
     $251 = HEAP32[$243>>2]|0;
     $252 = HEAP32[$D>>2]|0;
     $253 = HEAP32[$244>>2]|0;
     $254 = (FUNCTION_TABLE_iiiii[$251 & 8191]($252,$size$i$i,4,$253)|0);
     HEAP32[$$pre$phiZ2D>>2] = $254;
     $342 = $254;
    } else {
     $342 = $250;
    }
    $274 = $342;
   } else {
    $255 = ((($247)) + 16|0);
    $256 = ((($247)) + 12|0);
    $257 = HEAP32[$256>>2]|0;
    $258 = (($257) + 1)|0;
    HEAP32[$size1$i$i>>2] = $258;
    $259 = ($250|0)==(0);
    do {
     if ($259) {
      $260 = HEAP32[$243>>2]|0;
      $261 = HEAP32[$D>>2]|0;
      $262 = HEAP32[$244>>2]|0;
      $263 = (FUNCTION_TABLE_iiiii[$260 & 8191]($261,$size1$i$i,4,$262)|0);
      HEAP32[$$pre$phiZ2D>>2] = $263;
      $264 = ($263|0)==(0);
      if (!($264)) {
       $343 = $263;
       break;
      }
      $265 = HEAP32[$size1$i$i>>2]|0;
      $266 = HEAP32[$243>>2]|0;
      $267 = HEAP32[$D>>2]|0;
      $268 = HEAP32[$244>>2]|0;
      $269 = (FUNCTION_TABLE_iiiii[$266 & 8191]($267,$255,$265,$268)|0);
      HEAP32[$$pre$phiZ2D>>2] = $269;
      $343 = $269;
     } else {
      $343 = $250;
     }
    } while(0);
    $274 = $343;
   }
   $270 = HEAP32[$242>>2]|0;
   $271 = (((($270) + (($i$014$i*12)|0)|0)) + 4|0);
   $272 = HEAP32[$271>>2]|0;
   HEAP32[$8>>2] = $272;
   $273 = ($274|0)==(0);
   if ($273) {
    $275 = HEAP32[$243>>2]|0;
    $276 = HEAP32[$D>>2]|0;
    $277 = HEAP32[$244>>2]|0;
    $278 = (FUNCTION_TABLE_iiiii[$275 & 8191]($276,$8,4,$277)|0);
    HEAP32[$$pre$phiZ2D>>2] = $278;
    $$pre$i = HEAP32[$242>>2]|0;
    $280 = $$pre$i;$283 = $278;
   } else {
    $280 = $270;$283 = $274;
   }
   $279 = (((($280) + (($i$014$i*12)|0)|0)) + 8|0);
   $281 = HEAP32[$279>>2]|0;
   HEAP32[$7>>2] = $281;
   $282 = ($283|0)==(0);
   if ($282) {
    $284 = HEAP32[$243>>2]|0;
    $285 = HEAP32[$D>>2]|0;
    $286 = HEAP32[$244>>2]|0;
    $287 = (FUNCTION_TABLE_iiiii[$284 & 8191]($285,$7,4,$286)|0);
    HEAP32[$$pre$phiZ2D>>2] = $287;
    $344 = $287;
   } else {
    $344 = $283;
   }
   $288 = (($i$014$i) + 1)|0;
   $exitcond17$i = ($288|0)==($232|0);
   if ($exitcond17$i) {
    $295 = $344;
    break;
   } else {
    $250 = $344;$i$014$i = $288;
   }
  }
 } else {
  $295 = $341;
 }
 $289 = HEAP32[$206>>2]|0;
 $290 = ($289|0)==(0);
 if ($290) {
  $291 = ((($f)) + 36|0);
  $292 = HEAP32[$291>>2]|0;
  $293 = $292;
 } else {
  $293 = 0;
 }
 HEAP32[$6>>2] = $293;
 $294 = ($295|0)==(0);
 if ($294) {
  $296 = ((($D)) + 4|0);
  $297 = HEAP32[$296>>2]|0;
  $298 = HEAP32[$D>>2]|0;
  $299 = ((($D)) + 8|0);
  $300 = HEAP32[$299>>2]|0;
  $301 = (FUNCTION_TABLE_iiiii[$297 & 8191]($298,$6,4,$300)|0);
  HEAP32[$$pre$phiZ2D>>2] = $301;
  $345 = $301;
 } else {
  $345 = $295;
 }
 $302 = ($293|0)>(0);
 if (!($302)) {
  STACKTOP = sp;return;
 }
 $303 = ((($f)) + 28|0);
 $304 = ((($D)) + 4|0);
 $305 = ((($D)) + 8|0);
 $311 = $345;$i$113$i = 0;
 while(1) {
  $306 = HEAP32[$303>>2]|0;
  $307 = (($306) + ($i$113$i<<2)|0);
  $308 = HEAP32[$307>>2]|0;
  $309 = ($308|0)==(0|0);
  if ($309) {
   HEAP32[$size$i7$i>>2] = 0;
   $310 = ($311|0)==(0);
   if ($310) {
    $312 = HEAP32[$304>>2]|0;
    $313 = HEAP32[$D>>2]|0;
    $314 = HEAP32[$305>>2]|0;
    $315 = (FUNCTION_TABLE_iiiii[$312 & 8191]($313,$size$i7$i,4,$314)|0);
    HEAP32[$$pre$phiZ2D>>2] = $315;
    $346 = $315;
   } else {
    $346 = $311;
   }
   $347 = $346;
  } else {
   $316 = ((($308)) + 16|0);
   $317 = ((($308)) + 12|0);
   $318 = HEAP32[$317>>2]|0;
   $319 = (($318) + 1)|0;
   HEAP32[$size1$i8$i>>2] = $319;
   $320 = ($311|0)==(0);
   do {
    if ($320) {
     $321 = HEAP32[$304>>2]|0;
     $322 = HEAP32[$D>>2]|0;
     $323 = HEAP32[$305>>2]|0;
     $324 = (FUNCTION_TABLE_iiiii[$321 & 8191]($322,$size1$i8$i,4,$323)|0);
     HEAP32[$$pre$phiZ2D>>2] = $324;
     $325 = ($324|0)==(0);
     if (!($325)) {
      $348 = $324;
      break;
     }
     $326 = HEAP32[$size1$i8$i>>2]|0;
     $327 = HEAP32[$304>>2]|0;
     $328 = HEAP32[$D>>2]|0;
     $329 = HEAP32[$305>>2]|0;
     $330 = (FUNCTION_TABLE_iiiii[$327 & 8191]($328,$316,$326,$329)|0);
     HEAP32[$$pre$phiZ2D>>2] = $330;
     $348 = $330;
    } else {
     $348 = $311;
    }
   } while(0);
   $347 = $348;
  }
  $331 = (($i$113$i) + 1)|0;
  $exitcond$i = ($331|0)==($293|0);
  if ($exitcond$i) {
   break;
  } else {
   $311 = $347;$i$113$i = $331;
  }
 }
 STACKTOP = sp;return;
}
function _luaF_newCclosure($L,$nelems,$e) {
 $L = $L|0;
 $nelems = $nelems|0;
 $e = $e|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $nelems << 4;
 $1 = (($0) + 24)|0;
 $2 = (_luaM_realloc_($L,0,0,$1)|0);
 _luaC_link($L,$2,6);
 $3 = ((($2)) + 6|0);
 HEAP8[$3>>0] = 1;
 $4 = ((($2)) + 12|0);
 HEAP32[$4>>2] = $e;
 $5 = $nelems&255;
 $6 = ((($2)) + 7|0);
 HEAP8[$6>>0] = $5;
 return ($2|0);
}
function _luaF_newLclosure($L,$nelems,$e) {
 $L = $L|0;
 $nelems = $nelems|0;
 $e = $e|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $scevgep = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $nelems << 2;
 $1 = (($0) + 20)|0;
 $2 = (_luaM_realloc_($L,0,0,$1)|0);
 _luaC_link($L,$2,6);
 $3 = ((($2)) + 6|0);
 HEAP8[$3>>0] = 0;
 $4 = ((($2)) + 12|0);
 HEAP32[$4>>2] = $e;
 $5 = $nelems&255;
 $6 = ((($2)) + 7|0);
 HEAP8[$6>>0] = $5;
 $7 = ($nelems|0)==(0);
 if ($7) {
  return ($2|0);
 }
 $scevgep = ((($2)) + 20|0);
 _memset(($scevgep|0),0,($0|0))|0;
 return ($2|0);
}
function _luaF_newupval($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaM_realloc_($L,0,0,32)|0);
 _luaC_link($L,$0,10);
 $1 = ((($0)) + 16|0);
 $2 = ((($0)) + 8|0);
 HEAP32[$2>>2] = $1;
 $3 = ((($0)) + 24|0);
 HEAP32[$3>>2] = 0;
 return ($0|0);
}
function _luaF_findupval($L,$level) {
 $L = $L|0;
 $level = $level|0;
 var $$0 = 0, $$lcssa11 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $pp$0$lcssa = 0, $pp$03 = 0, $pp$03$phi = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($L)) + 104|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 L1: do {
  if ($4) {
   $pp$0$lcssa = $2;
  } else {
   $6 = $3;$pp$03 = $2;
   while(1) {
    $5 = ((($6)) + 8|0);
    $7 = HEAP32[$5>>2]|0;
    $8 = ($7>>>0)<($level>>>0);
    if ($8) {
     $pp$0$lcssa = $pp$03;
     break L1;
    }
    $9 = ($7|0)==($level|0);
    if ($9) {
     $$lcssa11 = $6;
     break;
    }
    $22 = HEAP32[$6>>2]|0;
    $23 = ($22|0)==(0|0);
    if ($23) {
     $pp$0$lcssa = $6;
     break L1;
    } else {
     $pp$03$phi = $6;$6 = $22;$pp$03 = $pp$03$phi;
    }
   }
   $10 = ((($$lcssa11)) + 5|0);
   $11 = HEAP8[$10>>0]|0;
   $12 = $11&255;
   $13 = ((($1)) + 20|0);
   $14 = HEAP8[$13>>0]|0;
   $15 = $14&255;
   $16 = $15 ^ 3;
   $17 = $12 & 3;
   $18 = $17 & $16;
   $19 = ($18|0)==(0);
   if ($19) {
    $$0 = $$lcssa11;
    return ($$0|0);
   }
   $20 = $12 ^ 3;
   $21 = $20&255;
   HEAP8[$10>>0] = $21;
   $$0 = $$lcssa11;
   return ($$0|0);
  }
 } while(0);
 $24 = (_luaM_realloc_($L,0,0,32)|0);
 $25 = ((($24)) + 4|0);
 HEAP8[$25>>0] = 10;
 $26 = ((($1)) + 20|0);
 $27 = HEAP8[$26>>0]|0;
 $28 = $27&255;
 $29 = $28 & 3;
 $30 = $29&255;
 $31 = ((($24)) + 5|0);
 HEAP8[$31>>0] = $30;
 $32 = ((($24)) + 8|0);
 HEAP32[$32>>2] = $level;
 $33 = HEAP32[$pp$0$lcssa>>2]|0;
 HEAP32[$24>>2] = $33;
 HEAP32[$pp$0$lcssa>>2] = $24;
 $34 = ((($1)) + 120|0);
 $35 = ((($24)) + 16|0);
 HEAP32[$35>>2] = $34;
 $36 = ((($1)) + 136|0);
 $37 = ((($36)) + 4|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = ((($24)) + 20|0);
 HEAP32[$39>>2] = $38;
 $40 = ((($38)) + 16|0);
 HEAP32[$40>>2] = $24;
 HEAP32[$37>>2] = $24;
 $$0 = $24;
 return ($$0|0);
}
function _luaF_freeupval($L,$uv) {
 $L = $L|0;
 $uv = $uv|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($uv)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($uv)) + 16|0);
 $3 = ($1|0)==($2|0);
 if ($3) {
  (_luaM_realloc_($L,$uv,32,0)|0);
  return;
 }
 $4 = ((($uv)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($4)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($7)) + 16|0);
 HEAP32[$8>>2] = $5;
 $9 = $7;
 $10 = HEAP32[$4>>2]|0;
 $11 = ((($10)) + 16|0);
 $12 = ((($11)) + 4|0);
 HEAP32[$12>>2] = $9;
 (_luaM_realloc_($L,$uv,32,0)|0);
 return;
}
function _luaF_close($L,$level) {
 $L = $L|0;
 $level = $level|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 104|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 $3 = ((($L)) + 16|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 20|0);
 $7 = $1;
 while(1) {
  $6 = ((($7)) + 8|0);
  $8 = HEAP32[$6>>2]|0;
  $9 = ($8>>>0)<($level>>>0);
  if ($9) {
   label = 10;
   break;
  }
  $10 = HEAP32[$7>>2]|0;
  HEAP32[$0>>2] = $10;
  $11 = ((($7)) + 5|0);
  $12 = HEAP8[$11>>0]|0;
  $13 = $12&255;
  $14 = HEAP8[$5>>0]|0;
  $15 = $14&255;
  $16 = $15 ^ 3;
  $17 = $13 & 3;
  $18 = $17 & $16;
  $19 = ($18|0)==(0);
  $20 = ((($7)) + 16|0);
  if ($19) {
   $30 = HEAP32[$20>>2]|0;
   $31 = ((($20)) + 4|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = ((($32)) + 16|0);
   HEAP32[$33>>2] = $30;
   $34 = $32;
   $35 = HEAP32[$20>>2]|0;
   $36 = ((($35)) + 16|0);
   $37 = ((($36)) + 4|0);
   HEAP32[$37>>2] = $34;
   $38 = $8;
   $39 = $38;
   $40 = HEAP32[$39>>2]|0;
   $41 = (($38) + 4)|0;
   $42 = $41;
   $43 = HEAP32[$42>>2]|0;
   $44 = $20;
   $45 = $44;
   HEAP32[$45>>2] = $40;
   $46 = (($44) + 4)|0;
   $47 = $46;
   HEAP32[$47>>2] = $43;
   $48 = ((($8)) + 8|0);
   $49 = HEAP32[$48>>2]|0;
   $50 = ((($20)) + 8|0);
   HEAP32[$50>>2] = $49;
   HEAP32[$6>>2] = $20;
   _luaC_linkupval($L,$7);
  } else {
   $21 = ($8|0)==($20|0);
   if (!($21)) {
    $22 = HEAP32[$20>>2]|0;
    $23 = ((($20)) + 4|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = ((($24)) + 16|0);
    HEAP32[$25>>2] = $22;
    $26 = $24;
    $27 = HEAP32[$20>>2]|0;
    $28 = ((($27)) + 16|0);
    $29 = ((($28)) + 4|0);
    HEAP32[$29>>2] = $26;
   }
   (_luaM_realloc_($L,$7,32,0)|0);
  }
  $51 = HEAP32[$0>>2]|0;
  $52 = ($51|0)==(0|0);
  if ($52) {
   label = 10;
   break;
  } else {
   $7 = $51;
  }
 }
 if ((label|0) == 10) {
  return;
 }
}
function _luaF_newproto($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = (_luaM_realloc_($L,0,0,76)|0);
 _luaC_link($L,$0,9);
 $1 = ((($0)) + 8|0);
 $2 = ((($0)) + 72|0);
 dest=$1; stop=dest+60|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 HEAP32[$2>>2] = 0;
 return ($0|0);
}
function _luaF_freeproto($L,$f) {
 $L = $L|0;
 $f = $f|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($f)) + 44|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 2;
 (_luaM_realloc_($L,$1,$4,0)|0);
 $5 = ((($f)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($f)) + 52|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $8 << 2;
 (_luaM_realloc_($L,$6,$9,0)|0);
 $10 = ((($f)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($f)) + 40|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = $13 << 4;
 (_luaM_realloc_($L,$11,$14,0)|0);
 $15 = ((($f)) + 20|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($f)) + 48|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = $18 << 2;
 (_luaM_realloc_($L,$16,$19,0)|0);
 $20 = ((($f)) + 24|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ((($f)) + 56|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = ($23*12)|0;
 (_luaM_realloc_($L,$21,$24,0)|0);
 $25 = ((($f)) + 28|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ((($f)) + 36|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = $28 << 2;
 (_luaM_realloc_($L,$26,$29,0)|0);
 (_luaM_realloc_($L,$f,76,0)|0);
 return;
}
function _luaF_freeclosure($L,$c) {
 $L = $L|0;
 $c = $c|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($c)) + 6|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = $1&255;
 $3 = ($2<<24>>24)==(0);
 $4 = ($1&65535) >>> 8;
 $5 = $4&65535;
 $6 = $5 << 4;
 $7 = (($6) + 24)|0;
 $8 = $5 << 2;
 $9 = (($8) + 20)|0;
 $10 = $3 ? $9 : $7;
 (_luaM_realloc_($L,$c,$10,0)|0);
 return;
}
function _luaF_getlocalname($f,$local_number,$pc) {
 $f = $f|0;
 $local_number = $local_number|0;
 $pc = $pc|0;
 var $$0 = 0, $$013 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $i$04 = 0, $i$04$lcssa10 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 56|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(0);
 if (!($2)) {
  $$0 = 0;
  return ($$0|0);
 }
 $3 = ((($f)) + 24|0);
 $4 = HEAP32[$3>>2]|0;
 $$013 = $local_number;$i$04 = 0;
 while(1) {
  $5 = (((($4) + (($i$04*12)|0)|0)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6|0)>($pc|0);
  if ($7) {
   $$0 = 0;
   label = 8;
   break;
  }
  $8 = (((($4) + (($i$04*12)|0)|0)) + 8|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)>($pc|0);
  if ($10) {
   $11 = (($$013) + -1)|0;
   $12 = ($11|0)==(0);
   if ($12) {
    $i$04$lcssa10 = $i$04;
    label = 6;
    break;
   } else {
    $$1 = $11;
   }
  } else {
   $$1 = $$013;
  }
  $16 = (($i$04) + 1)|0;
  $17 = ($16|0)<($1|0);
  if ($17) {
   $$013 = $$1;$i$04 = $16;
  } else {
   $$0 = 0;
   label = 8;
   break;
  }
 }
 if ((label|0) == 6) {
  $13 = (($4) + (($i$04$lcssa10*12)|0)|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ((($14)) + 16|0);
  $$0 = $15;
  return ($$0|0);
 }
 else if ((label|0) == 8) {
  return ($$0|0);
 }
 return (0)|0;
}
function _luaC_separateudata($L,$all) {
 $L = $L|0;
 $all = $all|0;
 var $$lcssa7$sink = 0, $$lcssa8 = 0, $$lcssa9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $8 = 0, $9 = 0, $deadmem$0$lcssa = 0, $deadmem$0$lcssa5 = 0, $deadmem$0$ph13 = 0, $deadmem$010 = 0, $deadmem$010$us = 0, $deadmem$06 = 0, $or$cond3 = 0, $p$0$ph12 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 112|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if ($5) {
  $deadmem$0$lcssa = 0;
  return ($deadmem$0$lcssa|0);
 }
 $6 = ($all|0)==(0);
 $7 = ((($1)) + 48|0);
 $78 = $4;$deadmem$0$ph13 = 0;$p$0$ph12 = $3;
 L4: while(1) {
  L6: do {
   if ($6) {
    $40 = $78;$deadmem$010 = $deadmem$0$ph13;
    while(1) {
     $39 = ((($40)) + 5|0);
     $41 = HEAP8[$39>>0]|0;
     $42 = $41 & 3;
     $43 = ($42<<24>>24)!=(0);
     $44 = $41 & 8;
     $45 = ($44<<24>>24)==(0);
     $or$cond3 = $45 & $43;
     if (!($or$cond3)) {
      $$lcssa7$sink = $40;$deadmem$06 = $deadmem$010;
      break L6;
     }
     $48 = ((($40)) + 8|0);
     $49 = HEAP32[$48>>2]|0;
     $50 = ($49|0)==(0|0);
     if ($50) {
      $$lcssa8 = $40;$$lcssa9 = $39;$deadmem$0$lcssa5 = $deadmem$010;
      label = 17;
      break L6;
     }
     $51 = ((($49)) + 6|0);
     $52 = HEAP8[$51>>0]|0;
     $53 = $52 & 4;
     $54 = ($53<<24>>24)==(0);
     if (!($54)) {
      $$lcssa8 = $40;$$lcssa9 = $39;$deadmem$0$lcssa5 = $deadmem$010;
      label = 17;
      break L6;
     }
     $55 = HEAP32[$0>>2]|0;
     $56 = ((($55)) + 196|0);
     $57 = HEAP32[$56>>2]|0;
     $58 = (_luaT_gettm($49,2,$57)|0);
     $59 = ($58|0)==(0|0);
     if ($59) {
      $$lcssa8 = $40;$$lcssa9 = $39;$deadmem$0$lcssa5 = $deadmem$010;
      label = 17;
      break L6;
     }
     $64 = ((($40)) + 16|0);
     $65 = HEAP32[$64>>2]|0;
     $66 = (($deadmem$010) + 24)|0;
     $67 = (($66) + ($65))|0;
     $68 = HEAP8[$39>>0]|0;
     $69 = $68&255;
     $70 = $69 | 8;
     $71 = $70&255;
     HEAP8[$39>>0] = $71;
     $72 = HEAP32[$40>>2]|0;
     HEAP32[$p$0$ph12>>2] = $72;
     $73 = HEAP32[$7>>2]|0;
     $74 = ($73|0)==(0|0);
     if ($74) {
      HEAP32[$40>>2] = $40;
     } else {
      $75 = HEAP32[$73>>2]|0;
      HEAP32[$40>>2] = $75;
      HEAP32[$73>>2] = $40;
     }
     HEAP32[$7>>2] = $40;
     $76 = HEAP32[$p$0$ph12>>2]|0;
     $77 = ($76|0)==(0|0);
     if ($77) {
      $deadmem$0$lcssa = $67;
      label = 22;
      break L4;
     } else {
      $40 = $76;$deadmem$010 = $67;
     }
    }
   } else {
    $9 = $78;$deadmem$010$us = $deadmem$0$ph13;
    while(1) {
     $8 = ((($9)) + 5|0);
     $10 = HEAP8[$8>>0]|0;
     $11 = $10 & 8;
     $12 = ($11<<24>>24)==(0);
     if (!($12)) {
      $$lcssa7$sink = $9;$deadmem$06 = $deadmem$010$us;
      break L6;
     }
     $13 = ((($9)) + 8|0);
     $14 = HEAP32[$13>>2]|0;
     $15 = ($14|0)==(0|0);
     if ($15) {
      $$lcssa8 = $9;$$lcssa9 = $8;$deadmem$0$lcssa5 = $deadmem$010$us;
      label = 17;
      break L6;
     }
     $16 = ((($14)) + 6|0);
     $17 = HEAP8[$16>>0]|0;
     $18 = $17 & 4;
     $19 = ($18<<24>>24)==(0);
     if (!($19)) {
      $$lcssa8 = $9;$$lcssa9 = $8;$deadmem$0$lcssa5 = $deadmem$010$us;
      label = 17;
      break L6;
     }
     $20 = HEAP32[$0>>2]|0;
     $21 = ((($20)) + 196|0);
     $22 = HEAP32[$21>>2]|0;
     $23 = (_luaT_gettm($14,2,$22)|0);
     $24 = ($23|0)==(0|0);
     if ($24) {
      $$lcssa8 = $9;$$lcssa9 = $8;$deadmem$0$lcssa5 = $deadmem$010$us;
      label = 17;
      break L6;
     }
     $25 = ((($9)) + 16|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = (($deadmem$010$us) + 24)|0;
     $28 = (($27) + ($26))|0;
     $29 = HEAP8[$8>>0]|0;
     $30 = $29&255;
     $31 = $30 | 8;
     $32 = $31&255;
     HEAP8[$8>>0] = $32;
     $33 = HEAP32[$9>>2]|0;
     HEAP32[$p$0$ph12>>2] = $33;
     $34 = HEAP32[$7>>2]|0;
     $35 = ($34|0)==(0|0);
     if ($35) {
      HEAP32[$9>>2] = $9;
     } else {
      $36 = HEAP32[$34>>2]|0;
      HEAP32[$9>>2] = $36;
      HEAP32[$34>>2] = $9;
     }
     HEAP32[$7>>2] = $9;
     $37 = HEAP32[$p$0$ph12>>2]|0;
     $38 = ($37|0)==(0|0);
     if ($38) {
      $deadmem$0$lcssa = $28;
      label = 22;
      break L4;
     } else {
      $9 = $37;$deadmem$010$us = $28;
     }
    }
   }
  } while(0);
  if ((label|0) == 17) {
   label = 0;
   $60 = HEAP8[$$lcssa9>>0]|0;
   $61 = $60&255;
   $62 = $61 | 8;
   $63 = $62&255;
   HEAP8[$$lcssa9>>0] = $63;
   $$lcssa7$sink = $$lcssa8;$deadmem$06 = $deadmem$0$lcssa5;
  }
  $46 = HEAP32[$$lcssa7$sink>>2]|0;
  $47 = ($46|0)==(0|0);
  if ($47) {
   $deadmem$0$lcssa = $deadmem$06;
   label = 22;
   break;
  } else {
   $78 = $46;$deadmem$0$ph13 = $deadmem$06;$p$0$ph12 = $$lcssa7$sink;
  }
 }
 if ((label|0) == 22) {
  return ($deadmem$0$lcssa|0);
 }
 return (0)|0;
}
function _luaC_callGCTM($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 48|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  return;
 }
 while(1) {
  _GCTM($L);
  $5 = HEAP32[$0>>2]|0;
  $6 = ((($5)) + 48|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)==(0|0);
  if ($8) {
   break;
  }
 }
 return;
}
function _GCTM($L) {
 $L = $L|0;
 var $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 48|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==($3|0);
 if ($5) {
  HEAP32[$2>>2] = 0;
  $$pre$phiZ2D = $4;
 } else {
  $6 = HEAP32[$4>>2]|0;
  HEAP32[$3>>2] = $6;
  $$pre$phiZ2D = $4;
 }
 $7 = ((($1)) + 112|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = HEAP32[$8>>2]|0;
 HEAP32[$$pre$phiZ2D>>2] = $9;
 HEAP32[$8>>2] = $4;
 $10 = ((($4)) + 5|0);
 $11 = HEAP8[$10>>0]|0;
 $12 = $11&255;
 $13 = $12 & 248;
 $14 = ((($1)) + 20|0);
 $15 = HEAP8[$14>>0]|0;
 $16 = $15&255;
 $17 = $16 & 3;
 $18 = $17 | $13;
 $19 = $18&255;
 HEAP8[$10>>0] = $19;
 $20 = ((($4)) + 8|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ($21|0)==(0|0);
 if ($22) {
  return;
 }
 $23 = ((($21)) + 6|0);
 $24 = HEAP8[$23>>0]|0;
 $25 = $24 & 4;
 $26 = ($25<<24>>24)==(0);
 if (!($26)) {
  return;
 }
 $27 = ((($1)) + 196|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = (_luaT_gettm($21,2,$28)|0);
 $30 = ($29|0)==(0|0);
 if ($30) {
  return;
 }
 $31 = ((($L)) + 57|0);
 $32 = HEAP8[$31>>0]|0;
 $33 = ((($1)) + 64|0);
 $34 = HEAP32[$33>>2]|0;
 HEAP8[$31>>0] = 0;
 $35 = ((($1)) + 68|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = $36 << 1;
 HEAP32[$33>>2] = $37;
 $38 = ((($L)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = $29;
 $41 = $40;
 $42 = HEAP32[$41>>2]|0;
 $43 = (($40) + 4)|0;
 $44 = $43;
 $45 = HEAP32[$44>>2]|0;
 $46 = $39;
 $47 = $46;
 HEAP32[$47>>2] = $42;
 $48 = (($46) + 4)|0;
 $49 = $48;
 HEAP32[$49>>2] = $45;
 $50 = ((($29)) + 8|0);
 $51 = HEAP32[$50>>2]|0;
 $52 = ((($39)) + 8|0);
 HEAP32[$52>>2] = $51;
 $53 = HEAP32[$38>>2]|0;
 $54 = ((($53)) + 16|0);
 HEAP32[$54>>2] = $4;
 $55 = ((($53)) + 24|0);
 HEAP32[$55>>2] = 7;
 $56 = HEAP32[$38>>2]|0;
 $57 = ((($56)) + 32|0);
 HEAP32[$38>>2] = $57;
 _luaD_call($L,$56,0);
 HEAP8[$31>>0] = $32;
 HEAP32[$33>>2] = $34;
 return;
}
function _luaC_freeall($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 20|0);
 HEAP8[$2>>0] = 67;
 $3 = ((($1)) + 28|0);
 (_sweeplist($L,$3,-3)|0);
 $4 = ((($1)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)>(0);
 if (!($6)) {
  return;
 }
 $i$01 = 0;
 while(1) {
  $7 = HEAP32[$1>>2]|0;
  $8 = (($7) + ($i$01<<2)|0);
  (_sweeplist($L,$8,-3)|0);
  $9 = (($i$01) + 1)|0;
  $10 = HEAP32[$4>>2]|0;
  $11 = ($9|0)<($10|0);
  if ($11) {
   $i$01 = $9;
  } else {
   break;
  }
 }
 return;
}
function _sweeplist($L,$p,$count) {
 $L = $L|0;
 $p = $p|0;
 $count = $count|0;
 var $$0$be = 0, $$0$lcssa = 0, $$012 = 0, $$03 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phiZ2D = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 20|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = $4 ^ 3;
 $6 = HEAP32[$p>>2]|0;
 $7 = ($6|0)==(0|0);
 if ($7) {
  $$0$lcssa = $p;
  return ($$0$lcssa|0);
 }
 $8 = ((($1)) + 28|0);
 $$012 = $count;$$03 = $p;$12 = $6;
 while(1) {
  $9 = (($$012) + -1)|0;
  $10 = ($$012|0)==(0);
  if ($10) {
   $$0$lcssa = $$03;
   label = 20;
   break;
  }
  $11 = ((($12)) + 4|0);
  $13 = HEAP16[$11>>1]|0;
  $14 = $13&255;
  $15 = ($14<<24>>24)==(8);
  if ($15) {
   $18 = ((($12)) + 104|0);
   (_sweeplist($L,$18,-3)|0);
   $$phi$trans$insert = ((($12)) + 5|0);
   $$pre = HEAP8[$$phi$trans$insert>>0]|0;
   $$pre$phiZ2D = $$phi$trans$insert;$20 = $$pre;
  } else {
   $16 = ($13&65535) >>> 8;
   $17 = $16&255;
   $$pre5 = ((($12)) + 5|0);
   $$pre$phiZ2D = $$pre5;$20 = $17;
  }
  $19 = $20&255;
  $21 = $19 ^ 3;
  $22 = $21 & $5;
  $23 = ($22|0)==(0);
  L11: do {
   if ($23) {
    $32 = HEAP32[$12>>2]|0;
    HEAP32[$$03>>2] = $32;
    $33 = HEAP32[$8>>2]|0;
    $34 = ($12|0)==($33|0);
    if ($34) {
     $35 = HEAP32[$12>>2]|0;
     HEAP32[$8>>2] = $35;
    }
    $36 = HEAP8[$11>>0]|0;
    $37 = $36&255;
    switch ($37|0) {
    case 9:  {
     _luaF_freeproto($L,$12);
     $$0$be = $$03;
     break L11;
     break;
    }
    case 6:  {
     _luaF_freeclosure($L,$12);
     $$0$be = $$03;
     break L11;
     break;
    }
    case 10:  {
     _luaF_freeupval($L,$12);
     $$0$be = $$03;
     break L11;
     break;
    }
    case 5:  {
     _luaH_free($L,$12);
     $$0$be = $$03;
     break L11;
     break;
    }
    case 8:  {
     _luaE_freethread($L,$12);
     $$0$be = $$03;
     break L11;
     break;
    }
    case 4:  {
     $38 = HEAP32[$0>>2]|0;
     $39 = ((($38)) + 4|0);
     $40 = HEAP32[$39>>2]|0;
     $41 = (($40) + -1)|0;
     HEAP32[$39>>2] = $41;
     $42 = ((($12)) + 12|0);
     $43 = HEAP32[$42>>2]|0;
     $44 = (($43) + 17)|0;
     (_luaM_realloc_($L,$12,$44,0)|0);
     $$0$be = $$03;
     break L11;
     break;
    }
    case 7:  {
     $45 = ((($12)) + 16|0);
     $46 = HEAP32[$45>>2]|0;
     $47 = (($46) + 24)|0;
     (_luaM_realloc_($L,$12,$47,0)|0);
     $$0$be = $$03;
     break L11;
     break;
    }
    default: {
     $$0$be = $$03;
     break L11;
    }
    }
   } else {
    $24 = $19 & 248;
    $25 = HEAP8[$2>>0]|0;
    $26 = $25&255;
    $27 = $26 & 3;
    $28 = $27 | $24;
    $29 = $28&255;
    HEAP8[$$pre$phiZ2D>>0] = $29;
    $$0$be = $12;
   }
  } while(0);
  $30 = HEAP32[$$0$be>>2]|0;
  $31 = ($30|0)==(0|0);
  if ($31) {
   $$0$lcssa = $$0$be;
   label = 20;
   break;
  } else {
   $$012 = $9;$$03 = $$0$be;$12 = $30;
  }
 }
 if ((label|0) == 20) {
  return ($$0$lcssa|0);
 }
 return (0)|0;
}
function _luaC_step($L) {
 $L = $L|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lim$0 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 84|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3*10)|0;
 $5 = ($4|0)==(0);
 $$ = $5 ? 2147483646 : $4;
 $6 = ((($1)) + 68|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($1)) + 64|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (($7) - ($9))|0;
 $11 = ((($1)) + 76|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10) + ($12))|0;
 HEAP32[$11>>2] = $13;
 $14 = ((($1)) + 21|0);
 $lim$0 = $$;
 while(1) {
  $15 = (_singlestep($L)|0);
  $16 = (($lim$0) - ($15))|0;
  $17 = HEAP8[$14>>0]|0;
  $18 = ($17<<24>>24)!=(0);
  $19 = ($16|0)>(0);
  $or$cond = $18 & $19;
  if ($or$cond) {
   $lim$0 = $16;
  } else {
   break;
  }
 }
 if (!($18)) {
  $26 = ((($1)) + 72|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = (($27>>>0) / 100)&-1;
  $29 = ((($1)) + 80|0);
  $30 = HEAP32[$29>>2]|0;
  $31 = Math_imul($28, $30)|0;
  HEAP32[$8>>2] = $31;
  return;
 }
 $20 = HEAP32[$11>>2]|0;
 $21 = ($20>>>0)<(1024);
 if ($21) {
  $22 = HEAP32[$6>>2]|0;
  $23 = (($22) + 1024)|0;
  HEAP32[$8>>2] = $23;
  return;
 } else {
  $24 = (($20) + -1024)|0;
  HEAP32[$11>>2] = $24;
  $25 = HEAP32[$6>>2]|0;
  HEAP32[$8>>2] = $25;
  return;
 }
}
function _singlestep($L) {
 $L = $L|0;
 var $$0 = 0, $$09$i$i = 0, $$in$i$i = 0, $$in13$i$i = 0, $$pre = 0, $$pre$i$i = 0, $$pre$phi$iZ2D = 0, $$pre14$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0;
 var $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $m$0$lcssa$i12$i = 0, $m$01$i10$i = 0, $not$$i$i = 0, $or$cond$i = 0, $or$cond$i$i = 0, $or$cond15$i$i = 0, $or$cond8$i$i = 0, $sum = 0;
 var $u$0$i$i = 0, $uv$0$i$i = 0, $uv$0$in$i$i = 0, $uv$0$in3$i$i = 0, $uv$04$i$i = 0, $uv$05$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 21|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 switch ($4|0) {
 case 0:  {
  _markroot($L);
  $$0 = 0;
  return ($$0|0);
  break;
 }
 case 1:  {
  $5 = ((($1)) + 36|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6|0)==(0|0);
  if (!($7)) {
   $8 = (_propagatemark($1)|0);
   $$0 = $8;
   return ($$0|0);
  }
  $9 = ((($1)) + 120|0);
  $10 = ((($1)) + 136|0);
  $uv$0$in3$i$i = ((($10)) + 4|0);
  $uv$04$i$i = HEAP32[$uv$0$in3$i$i>>2]|0;
  $11 = ($uv$04$i$i|0)==($9|0);
  if (!($11)) {
   $uv$05$i$i = $uv$04$i$i;
   while(1) {
    $12 = ((($uv$05$i$i)) + 5|0);
    $13 = HEAP8[$12>>0]|0;
    $14 = $13 & 7;
    $15 = ($14<<24>>24)==(0);
    if ($15) {
     $16 = ((($uv$05$i$i)) + 8|0);
     $17 = HEAP32[$16>>2]|0;
     $18 = ((($17)) + 8|0);
     $19 = HEAP32[$18>>2]|0;
     $20 = ($19|0)>(3);
     if ($20) {
      $21 = HEAP32[$17>>2]|0;
      $22 = ((($21)) + 5|0);
      $23 = HEAP8[$22>>0]|0;
      $24 = $23 & 3;
      $25 = ($24<<24>>24)==(0);
      if (!($25)) {
       _reallymarkobject($1,$21);
      }
     }
    }
    $26 = ((($uv$05$i$i)) + 16|0);
    $uv$0$in$i$i = ((($26)) + 4|0);
    $uv$0$i$i = HEAP32[$uv$0$in$i$i>>2]|0;
    $27 = ($uv$0$i$i|0)==($9|0);
    if ($27) {
     break;
    } else {
     $uv$05$i$i = $uv$0$i$i;
    }
   }
   $$pre = HEAP32[$5>>2]|0;
   $28 = ($$pre|0)==(0|0);
   if (!($28)) {
    while(1) {
     (_propagatemark($1)|0);
     $29 = HEAP32[$5>>2]|0;
     $30 = ($29|0)==(0|0);
     if ($30) {
      break;
     }
    }
   }
  }
  $31 = ((($1)) + 44|0);
  $32 = HEAP32[$31>>2]|0;
  HEAP32[$5>>2] = $32;
  HEAP32[$31>>2] = 0;
  $33 = ((($L)) + 5|0);
  $34 = HEAP8[$33>>0]|0;
  $35 = $34 & 3;
  $36 = ($35<<24>>24)==(0);
  if (!($36)) {
   _reallymarkobject($1,$L);
  }
  _markmt($1);
  $37 = HEAP32[$5>>2]|0;
  $38 = ($37|0)==(0|0);
  if (!($38)) {
   while(1) {
    (_propagatemark($1)|0);
    $39 = HEAP32[$5>>2]|0;
    $40 = ($39|0)==(0|0);
    if ($40) {
     break;
    }
   }
  }
  $41 = ((($1)) + 40|0);
  $42 = HEAP32[$41>>2]|0;
  HEAP32[$5>>2] = $42;
  HEAP32[$41>>2] = 0;
  $43 = ($42|0)==(0);
  if (!($43)) {
   while(1) {
    (_propagatemark($1)|0);
    $44 = HEAP32[$5>>2]|0;
    $45 = ($44|0)==(0|0);
    if ($45) {
     break;
    }
   }
  }
  $46 = (_luaC_separateudata($L,0)|0);
  $47 = ((($1)) + 48|0);
  $48 = HEAP32[$47>>2]|0;
  $49 = ($48|0)==(0|0);
  if (!($49)) {
   $50 = ((($1)) + 20|0);
   $u$0$i$i = $48;
   while(1) {
    $51 = HEAP32[$u$0$i$i>>2]|0;
    $52 = ((($51)) + 5|0);
    $53 = HEAP8[$52>>0]|0;
    $54 = $53&255;
    $55 = $54 & 248;
    $56 = HEAP8[$50>>0]|0;
    $57 = $56&255;
    $58 = $57 & 3;
    $59 = $58 | $55;
    $60 = $59&255;
    HEAP8[$52>>0] = $60;
    _reallymarkobject($1,$51);
    $61 = HEAP32[$47>>2]|0;
    $62 = ($51|0)==($61|0);
    if ($62) {
     break;
    } else {
     $u$0$i$i = $51;
    }
   }
  }
  $63 = HEAP32[$5>>2]|0;
  $64 = ($63|0)==(0|0);
  if ($64) {
   $m$0$lcssa$i12$i = 0;
  } else {
   $m$01$i10$i = 0;
   while(1) {
    $65 = (_propagatemark($1)|0);
    $66 = (($65) + ($m$01$i10$i))|0;
    $67 = HEAP32[$5>>2]|0;
    $68 = ($67|0)==(0|0);
    if ($68) {
     $m$0$lcssa$i12$i = $66;
     break;
    } else {
     $m$01$i10$i = $66;
    }
   }
  }
  $69 = HEAP32[$31>>2]|0;
  $70 = ($69|0)==(0|0);
  if (!($70)) {
   $$09$i$i = $69;
   while(1) {
    $71 = ((($$09$i$i)) + 28|0);
    $72 = HEAP32[$71>>2]|0;
    $73 = ((($$09$i$i)) + 5|0);
    $74 = HEAP8[$73>>0]|0;
    $75 = $74 & 16;
    $76 = ($75<<24>>24)==(0);
    $77 = ($72|0)==(0);
    $or$cond15$i$i = $77 | $76;
    if (!($or$cond15$i$i)) {
     $78 = ((($$09$i$i)) + 12|0);
     $$pre$i$i = HEAP32[$78>>2]|0;
     $$in$i$i = $72;
     while(1) {
      $79 = (($$in$i$i) + -1)|0;
      $80 = (((($$pre$i$i) + ($79<<4)|0)) + 8|0);
      $81 = HEAP32[$80>>2]|0;
      $82 = ($81|0)>(3);
      do {
       if ($82) {
        $83 = (($$pre$i$i) + ($79<<4)|0);
        $84 = ($81|0)==(4);
        $85 = HEAP32[$83>>2]|0;
        $86 = ((($85)) + 5|0);
        $87 = HEAP8[$86>>0]|0;
        if ($84) {
         $88 = $87&255;
         $89 = $88 & 252;
         $90 = $89&255;
         HEAP8[$86>>0] = $90;
         break;
        }
        $91 = $87 & 3;
        $92 = ($91<<24>>24)==(0);
        if ($92) {
         $93 = ($81|0)==(7);
         $94 = $87 & 8;
         $95 = ($94<<24>>24)!=(0);
         $or$cond$i$i = $93 & $95;
         if (!($or$cond$i$i)) {
          break;
         }
        }
        HEAP32[$80>>2] = 0;
       }
      } while(0);
      $96 = ($79|0)==(0);
      if ($96) {
       break;
      } else {
       $$in$i$i = $79;
      }
     }
    }
    $97 = ((($$09$i$i)) + 7|0);
    $98 = HEAP8[$97>>0]|0;
    $99 = $98&255;
    $100 = 1 << $99;
    $101 = ((($$09$i$i)) + 16|0);
    $$pre14$i$i = HEAP32[$101>>2]|0;
    $$in13$i$i = $100;
    while(1) {
     $102 = (($$in13$i$i) + -1)|0;
     $103 = (($$pre14$i$i) + ($102<<5)|0);
     $104 = (((($$pre14$i$i) + ($102<<5)|0)) + 8|0);
     $105 = HEAP32[$104>>2]|0;
     $106 = ($105|0)==(0);
     do {
      if (!($106)) {
       $108 = (((($$pre14$i$i) + ($102<<5)|0)) + 24|0);
       $109 = HEAP32[$108>>2]|0;
       $110 = ($109|0)>(3);
       do {
        if ($110) {
         $111 = (((($$pre14$i$i) + ($102<<5)|0)) + 16|0);
         $112 = ($109|0)==(4);
         $113 = HEAP32[$111>>2]|0;
         $114 = ((($113)) + 5|0);
         $115 = HEAP8[$114>>0]|0;
         if ($112) {
          $116 = $115&255;
          $117 = $116 & 252;
          $118 = $117&255;
          HEAP8[$114>>0] = $118;
          label = 42;
          break;
         }
         $119 = $115 & 3;
         $not$$i$i = ($119<<24>>24)==(0);
         if ($not$$i$i) {
          label = 42;
         } else {
          HEAP32[$104>>2] = 0;
         }
        } else {
         label = 42;
        }
       } while(0);
       if ((label|0) == 42) {
        label = 0;
        $120 = ($105|0)>(3);
        if (!($120)) {
         break;
        }
        $121 = ($105|0)==(4);
        $122 = HEAP32[$103>>2]|0;
        $123 = ((($122)) + 5|0);
        $124 = HEAP8[$123>>0]|0;
        if ($121) {
         $125 = $124&255;
         $126 = $125 & 252;
         $127 = $126&255;
         HEAP8[$123>>0] = $127;
         break;
        }
        $128 = $124 & 3;
        $129 = ($128<<24>>24)==(0);
        if ($129) {
         $130 = ($105|0)==(7);
         $131 = $124 & 8;
         $132 = ($131<<24>>24)!=(0);
         $or$cond8$i$i = $130 & $132;
         if (!($or$cond8$i$i)) {
          break;
         }
        }
        HEAP32[$104>>2] = 0;
        if (!($110)) {
         break;
        }
       }
       HEAP32[$108>>2] = 11;
      }
     } while(0);
     $107 = ($102|0)==(0);
     if ($107) {
      break;
     } else {
      $$in13$i$i = $102;
     }
    }
    $133 = ((($$09$i$i)) + 24|0);
    $134 = HEAP32[$133>>2]|0;
    $135 = ($134|0)==(0|0);
    if ($135) {
     break;
    } else {
     $$09$i$i = $134;
    }
   }
  }
  $136 = ((($1)) + 20|0);
  $137 = HEAP8[$136>>0]|0;
  $138 = $137&255;
  $139 = $138 ^ 3;
  $140 = $139&255;
  HEAP8[$136>>0] = $140;
  $141 = ((($1)) + 24|0);
  HEAP32[$141>>2] = 0;
  $142 = ((($1)) + 28|0);
  $143 = ((($1)) + 32|0);
  HEAP32[$143>>2] = $142;
  HEAP8[$2>>0] = 2;
  $144 = ((($1)) + 68|0);
  $145 = HEAP32[$144>>2]|0;
  $sum = (($m$0$lcssa$i12$i) + ($46))|0;
  $146 = (($145) - ($sum))|0;
  $147 = ((($1)) + 72|0);
  HEAP32[$147>>2] = $146;
  $$0 = 0;
  return ($$0|0);
  break;
 }
 case 2:  {
  $148 = ((($1)) + 68|0);
  $149 = HEAP32[$148>>2]|0;
  $150 = ((($1)) + 24|0);
  $151 = HEAP32[$150>>2]|0;
  $152 = (($151) + 1)|0;
  HEAP32[$150>>2] = $152;
  $153 = HEAP32[$1>>2]|0;
  $154 = (($153) + ($151<<2)|0);
  (_sweeplist($L,$154,-3)|0);
  $155 = HEAP32[$150>>2]|0;
  $156 = ((($1)) + 8|0);
  $157 = HEAP32[$156>>2]|0;
  $158 = ($155|0)<($157|0);
  if (!($158)) {
   HEAP8[$2>>0] = 3;
  }
  $159 = HEAP32[$148>>2]|0;
  $160 = ((($1)) + 72|0);
  $161 = HEAP32[$160>>2]|0;
  $162 = (($159) - ($149))|0;
  $163 = (($162) + ($161))|0;
  HEAP32[$160>>2] = $163;
  $$0 = 10;
  return ($$0|0);
  break;
 }
 case 3:  {
  $164 = ((($1)) + 68|0);
  $165 = HEAP32[$164>>2]|0;
  $166 = ((($1)) + 32|0);
  $167 = HEAP32[$166>>2]|0;
  $168 = (_sweeplist($L,$167,40)|0);
  HEAP32[$166>>2] = $168;
  $169 = HEAP32[$168>>2]|0;
  $170 = ($169|0)==(0|0);
  if ($170) {
   $171 = HEAP32[$0>>2]|0;
   $172 = ((($171)) + 4|0);
   $173 = HEAP32[$172>>2]|0;
   $174 = ((($171)) + 8|0);
   $175 = HEAP32[$174>>2]|0;
   $176 = (($175|0) / 4)&-1;
   $177 = ($173>>>0)<($176>>>0);
   $178 = ($175|0)>(64);
   $or$cond$i = $178 & $177;
   if ($or$cond$i) {
    $179 = (($175|0) / 2)&-1;
    _luaS_resize($L,$179);
   }
   $180 = ((($171)) + 52|0);
   $181 = ((($171)) + 60|0);
   $182 = HEAP32[$181>>2]|0;
   $183 = ($182>>>0)>(64);
   if ($183) {
    $184 = $182 >>> 1;
    $185 = (($184) + 1)|0;
    $186 = ($185>>>0)<(4294967294);
    if ($186) {
     $187 = HEAP32[$180>>2]|0;
     $188 = (_luaM_realloc_($L,$187,$182,$184)|0);
     $$pre$phi$iZ2D = $180;$190 = $188;
    } else {
     $189 = (_luaM_toobig($L)|0);
     $$pre$phi$iZ2D = $180;$190 = $189;
    }
    HEAP32[$$pre$phi$iZ2D>>2] = $190;
    HEAP32[$181>>2] = $184;
   }
   HEAP8[$2>>0] = 4;
  }
  $191 = HEAP32[$164>>2]|0;
  $192 = ((($1)) + 72|0);
  $193 = HEAP32[$192>>2]|0;
  $194 = (($191) - ($165))|0;
  $195 = (($194) + ($193))|0;
  HEAP32[$192>>2] = $195;
  $$0 = 400;
  return ($$0|0);
  break;
 }
 case 4:  {
  $196 = ((($1)) + 48|0);
  $197 = HEAP32[$196>>2]|0;
  $198 = ($197|0)==(0|0);
  if ($198) {
   HEAP8[$2>>0] = 0;
   $203 = ((($1)) + 76|0);
   HEAP32[$203>>2] = 0;
   $$0 = 0;
   return ($$0|0);
  }
  _GCTM($L);
  $199 = ((($1)) + 72|0);
  $200 = HEAP32[$199>>2]|0;
  $201 = ($200>>>0)>(100);
  if (!($201)) {
   $$0 = 100;
   return ($$0|0);
  }
  $202 = (($200) + -100)|0;
  HEAP32[$199>>2] = $202;
  $$0 = 100;
  return ($$0|0);
  break;
 }
 default: {
  $$0 = 0;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function _markroot($L) {
 $L = $L|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 36|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($1)) + 40|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($1)) + 44|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($1)) + 112|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($6)) + 5|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = $8 & 3;
 $10 = ($9<<24>>24)==(0);
 if ($10) {
  $12 = $6;
 } else {
  _reallymarkobject($1,$6);
  $$pre = HEAP32[$5>>2]|0;
  $12 = $$pre;
 }
 $11 = ((($12)) + 80|0);
 $13 = HEAP32[$11>>2]|0;
 $14 = ($13|0)>(3);
 if ($14) {
  $15 = ((($12)) + 72|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ((($16)) + 5|0);
  $18 = HEAP8[$17>>0]|0;
  $19 = $18 & 3;
  $20 = ($19<<24>>24)==(0);
  if (!($20)) {
   _reallymarkobject($1,$16);
  }
 }
 $21 = HEAP32[$0>>2]|0;
 $22 = ((($21)) + 104|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = ($23|0)>(3);
 if (!($24)) {
  _markmt($1);
  $31 = ((($1)) + 21|0);
  HEAP8[$31>>0] = 1;
  return;
 }
 $25 = ((($21)) + 96|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ((($26)) + 5|0);
 $28 = HEAP8[$27>>0]|0;
 $29 = $28 & 3;
 $30 = ($29<<24>>24)==(0);
 if ($30) {
  _markmt($1);
  $31 = ((($1)) + 21|0);
  HEAP8[$31>>0] = 1;
  return;
 }
 _reallymarkobject($1,$26);
 _markmt($1);
 $31 = ((($1)) + 21|0);
 HEAP8[$31>>0] = 1;
 return;
}
function _reallymarkobject($g,$o) {
 $g = $g|0;
 $o = $o|0;
 var $$lcssa58 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre21 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $o$tr = 0, $o$tr$lcssa52 = 0, $o$tr$lcssa53 = 0, $o$tr$lcssa54 = 0;
 var $o$tr$lcssa55 = 0, $o$tr$lcssa56 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$phi$trans$insert = ((($o)) + 5|0);
 $$pre = HEAP8[$$phi$trans$insert>>0]|0;
 $2 = $$pre;$o$tr = $o;
 L1: while(1) {
  $0 = ((($o$tr)) + 5|0);
  $1 = $2&255;
  $3 = $1 & 252;
  $4 = $3&255;
  HEAP8[$0>>0] = $4;
  $5 = ((($o$tr)) + 4|0);
  $6 = HEAP8[$5>>0]|0;
  $7 = $6&255;
  switch ($7|0) {
  case 9:  {
   $o$tr$lcssa52 = $o$tr;
   label = 15;
   break L1;
   break;
  }
  case 10:  {
   $$lcssa58 = $0;$o$tr$lcssa53 = $o$tr;
   label = 7;
   break L1;
   break;
  }
  case 6:  {
   $o$tr$lcssa54 = $o$tr;
   label = 12;
   break L1;
   break;
  }
  case 5:  {
   $o$tr$lcssa55 = $o$tr;
   label = 13;
   break L1;
   break;
  }
  case 8:  {
   $o$tr$lcssa56 = $o$tr;
   label = 14;
   break L1;
   break;
  }
  case 7:  {
   break;
  }
  default: {
   label = 16;
   break L1;
  }
  }
  $8 = ((($o$tr)) + 8|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = $3 | 4;
  $11 = $10&255;
  HEAP8[$0>>0] = $11;
  $12 = ($9|0)==(0|0);
  if (!($12)) {
   $13 = ((($9)) + 5|0);
   $14 = HEAP8[$13>>0]|0;
   $15 = $14 & 3;
   $16 = ($15<<24>>24)==(0);
   if (!($16)) {
    _reallymarkobject($g,$9);
   }
  }
  $17 = ((($o$tr)) + 12|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = ((($18)) + 5|0);
  $20 = HEAP8[$19>>0]|0;
  $21 = $20 & 3;
  $22 = ($21<<24>>24)==(0);
  if ($22) {
   label = 16;
   break;
  } else {
   $2 = $20;$o$tr = $18;
  }
 }
 if ((label|0) == 7) {
  $23 = ((($o$tr$lcssa53)) + 8|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = ((($24)) + 8|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = ($26|0)>(3);
  if ($27) {
   $28 = HEAP32[$24>>2]|0;
   $29 = ((($28)) + 5|0);
   $30 = HEAP8[$29>>0]|0;
   $31 = $30 & 3;
   $32 = ($31<<24>>24)==(0);
   if ($32) {
    $35 = $24;
   } else {
    _reallymarkobject($g,$28);
    $$pre21 = HEAP32[$23>>2]|0;
    $35 = $$pre21;
   }
  } else {
   $35 = $24;
  }
  $33 = ((($o$tr$lcssa53)) + 16|0);
  $34 = ($35|0)==($33|0);
  if (!($34)) {
   return;
  }
  $36 = HEAP8[$$lcssa58>>0]|0;
  $37 = $36&255;
  $38 = $37 | 4;
  $39 = $38&255;
  HEAP8[$$lcssa58>>0] = $39;
  return;
 }
 else if ((label|0) == 12) {
  $40 = ((($g)) + 36|0);
  $41 = HEAP32[$40>>2]|0;
  $42 = ((($o$tr$lcssa54)) + 8|0);
  HEAP32[$42>>2] = $41;
  HEAP32[$40>>2] = $o$tr$lcssa54;
  return;
 }
 else if ((label|0) == 13) {
  $43 = ((($g)) + 36|0);
  $44 = HEAP32[$43>>2]|0;
  $45 = ((($o$tr$lcssa55)) + 24|0);
  HEAP32[$45>>2] = $44;
  HEAP32[$43>>2] = $o$tr$lcssa55;
  return;
 }
 else if ((label|0) == 14) {
  $46 = ((($g)) + 36|0);
  $47 = HEAP32[$46>>2]|0;
  $48 = ((($o$tr$lcssa56)) + 108|0);
  HEAP32[$48>>2] = $47;
  HEAP32[$46>>2] = $o$tr$lcssa56;
  return;
 }
 else if ((label|0) == 15) {
  $49 = ((($g)) + 36|0);
  $50 = HEAP32[$49>>2]|0;
  $51 = ((($o$tr$lcssa52)) + 68|0);
  HEAP32[$51>>2] = $50;
  HEAP32[$49>>2] = $o$tr$lcssa52;
  return;
 }
 else if ((label|0) == 16) {
  return;
 }
}
function _markmt($g) {
 $g = $g|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($g)) + 152|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = ((($1)) + 5|0);
  $4 = HEAP8[$3>>0]|0;
  $5 = $4 & 3;
  $6 = ($5<<24>>24)==(0);
  if (!($6)) {
   _reallymarkobject($g,$1);
  }
 }
 $7 = ((($g)) + 156|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 if (!($9)) {
  $10 = ((($8)) + 5|0);
  $11 = HEAP8[$10>>0]|0;
  $12 = $11 & 3;
  $13 = ($12<<24>>24)==(0);
  if (!($13)) {
   _reallymarkobject($g,$8);
  }
 }
 $14 = ((($g)) + 160|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)==(0|0);
 if (!($16)) {
  $17 = ((($15)) + 5|0);
  $18 = HEAP8[$17>>0]|0;
  $19 = $18 & 3;
  $20 = ($19<<24>>24)==(0);
  if (!($20)) {
   _reallymarkobject($g,$15);
  }
 }
 $21 = ((($g)) + 164|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ($22|0)==(0|0);
 if (!($23)) {
  $24 = ((($22)) + 5|0);
  $25 = HEAP8[$24>>0]|0;
  $26 = $25 & 3;
  $27 = ($26<<24>>24)==(0);
  if (!($27)) {
   _reallymarkobject($g,$22);
  }
 }
 $28 = ((($g)) + 168|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = ($29|0)==(0|0);
 if (!($30)) {
  $31 = ((($29)) + 5|0);
  $32 = HEAP8[$31>>0]|0;
  $33 = $32 & 3;
  $34 = ($33<<24>>24)==(0);
  if (!($34)) {
   _reallymarkobject($g,$29);
  }
 }
 $35 = ((($g)) + 172|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = ($36|0)==(0|0);
 if (!($37)) {
  $38 = ((($36)) + 5|0);
  $39 = HEAP8[$38>>0]|0;
  $40 = $39 & 3;
  $41 = ($40<<24>>24)==(0);
  if (!($41)) {
   _reallymarkobject($g,$36);
  }
 }
 $42 = ((($g)) + 176|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = ($43|0)==(0|0);
 if (!($44)) {
  $45 = ((($43)) + 5|0);
  $46 = HEAP8[$45>>0]|0;
  $47 = $46 & 3;
  $48 = ($47<<24>>24)==(0);
  if (!($48)) {
   _reallymarkobject($g,$43);
  }
 }
 $49 = ((($g)) + 180|0);
 $50 = HEAP32[$49>>2]|0;
 $51 = ($50|0)==(0|0);
 if (!($51)) {
  $52 = ((($50)) + 5|0);
  $53 = HEAP8[$52>>0]|0;
  $54 = $53 & 3;
  $55 = ($54<<24>>24)==(0);
  if (!($55)) {
   _reallymarkobject($g,$50);
  }
 }
 $56 = ((($g)) + 184|0);
 $57 = HEAP32[$56>>2]|0;
 $58 = ($57|0)==(0|0);
 if ($58) {
  return;
 }
 $59 = ((($57)) + 5|0);
 $60 = HEAP8[$59>>0]|0;
 $61 = $60 & 3;
 $62 = ($61<<24>>24)==(0);
 if ($62) {
  return;
 }
 _reallymarkobject($g,$57);
 return;
}
function _propagatemark($g) {
 $g = $g|0;
 var $$0 = 0, $$in$i = 0, $$lim$0$i = 0, $$phi$trans$insert$i = 0, $$pr$pre$i = 0, $$pr19$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i1 = 0, $$pre$i5 = 0, $$pre$i8 = 0, $$pre$phi$iZ2D = 0, $$pre$phi28Z2D = 0, $$pre$phiZ2D = 0, $$pre11$i = 0, $$pre25$pre$phiZZ2D = 0, $$pre26 = 0, $$pre27 = 0, $$pre7$i = 0, $$pre8$i = 0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0;
 var $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0;
 var $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0;
 var $369 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $ci$06$i = 0, $exitcond = 0, $exitcond24 = 0, $i$04$i = 0, $i$09$i = 0, $i$17$i = 0, $i$25$i = 0, $i$34$i = 0, $i1$03$i = 0, $lim$0$lcssa$i = 0;
 var $lim$05$i = 0, $o$0$lcssa$i = 0, $o$02$i = 0, $o$11$i = 0, $or$cond$i = 0, $or$cond$i$i = 0, $or$cond1$i$i = 0, $or$cond3$i = 0, $tmp = 0, $tmp13 = 0, $tmp14 = 0, $tmp15 = 0, $weakkey$0813$i = 0, $weakkey$0814$i = 0, $weakvalue$0912$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($g)) + 36|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 5|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = $4 | 4;
 $6 = $5&255;
 HEAP8[$2>>0] = $6;
 $7 = ((($1)) + 4|0);
 $8 = HEAP16[$7>>1]|0;
 $9 = $8&65535;
 $10 = $9 & 255;
 switch ($10|0) {
 case 5:  {
  $11 = ((($1)) + 24|0);
  $12 = HEAP32[$11>>2]|0;
  HEAP32[$0>>2] = $12;
  $13 = ((($1)) + 8|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ($14|0)==(0|0);
  do {
   if ($15) {
    $363 = 0;$weakkey$0813$i = 0;
    label = 13;
   } else {
    $16 = ((($14)) + 5|0);
    $17 = HEAP8[$16>>0]|0;
    $18 = $17 & 3;
    $19 = ($18<<24>>24)==(0);
    if ($19) {
     $$pr19$i = $14;
    } else {
     _reallymarkobject($g,$14);
     $$pr$pre$i = HEAP32[$13>>2]|0;
     $20 = ($$pr$pre$i|0)==(0|0);
     if ($20) {
      $363 = 0;$weakkey$0813$i = 0;
      label = 13;
      break;
     } else {
      $$pr19$i = $$pr$pre$i;
     }
    }
    $21 = ((($$pr19$i)) + 6|0);
    $22 = HEAP8[$21>>0]|0;
    $23 = $22 & 8;
    $24 = ($23<<24>>24)==(0);
    if ($24) {
     $25 = ((($g)) + 200|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = (_luaT_gettm($$pr19$i,3,$26)|0);
     $28 = ($27|0)==(0|0);
     if ($28) {
      $363 = 0;$weakkey$0813$i = 0;
      label = 13;
     } else {
      $29 = ((($27)) + 8|0);
      $30 = HEAP32[$29>>2]|0;
      $31 = ($30|0)==(4);
      if ($31) {
       $32 = HEAP32[$27>>2]|0;
       $33 = ((($32)) + 16|0);
       $34 = (_strchr($33,107)|0);
       $35 = ($34|0)!=(0|0);
       $36 = $35&1;
       $37 = (_strchr($33,118)|0);
       $38 = ($37|0)!=(0|0);
       $39 = $38&1;
       $or$cond$i = $35 | $38;
       if ($or$cond$i) {
        $40 = HEAP8[$2>>0]|0;
        $41 = $40&255;
        $42 = $41 & 231;
        $43 = $36 << 3;
        $44 = $39 << 4;
        $45 = $44 | $43;
        $46 = $45 | $42;
        $47 = $46&255;
        HEAP8[$2>>0] = $47;
        $48 = ((($g)) + 44|0);
        $49 = HEAP32[$48>>2]|0;
        HEAP32[$11>>2] = $49;
        HEAP32[$48>>2] = $1;
       }
       $or$cond3$i = $35 & $38;
       if ($or$cond3$i) {
        $$pre26 = ((($1)) + 7|0);
        $$pre25$pre$phiZZ2D = $$pre26;
        label = 52;
        break;
       } else {
        if ($38) {
         $364 = $35;$365 = 1;$weakkey$0814$i = $36;$weakvalue$0912$i = $39;
         label = 19;
         break;
        } else {
         $363 = $35;$weakkey$0813$i = $36;
         label = 13;
         break;
        }
       }
      } else {
       $363 = 0;$weakkey$0813$i = 0;
       label = 13;
      }
     }
    } else {
     $363 = 0;$weakkey$0813$i = 0;
     label = 13;
    }
   }
  } while(0);
  if ((label|0) == 13) {
   $50 = ((($1)) + 28|0);
   $51 = HEAP32[$50>>2]|0;
   $52 = ($51|0)==(0);
   if ($52) {
    $364 = $363;$365 = 0;$weakkey$0814$i = $weakkey$0813$i;$weakvalue$0912$i = 0;
    label = 19;
   } else {
    $53 = ((($1)) + 12|0);
    $$in$i = $51;
    while(1) {
     $54 = (($$in$i) + -1)|0;
     $55 = HEAP32[$53>>2]|0;
     $56 = (((($55) + ($54<<4)|0)) + 8|0);
     $57 = HEAP32[$56>>2]|0;
     $58 = ($57|0)>(3);
     if ($58) {
      $59 = (($55) + ($54<<4)|0);
      $60 = HEAP32[$59>>2]|0;
      $61 = ((($60)) + 5|0);
      $62 = HEAP8[$61>>0]|0;
      $63 = $62 & 3;
      $64 = ($63<<24>>24)==(0);
      if (!($64)) {
       _reallymarkobject($g,$60);
      }
     }
     $65 = ($54|0)==(0);
     if ($65) {
      $364 = $363;$365 = 0;$weakkey$0814$i = $weakkey$0813$i;$weakvalue$0912$i = 0;
      label = 19;
      break;
     } else {
      $$in$i = $54;
     }
    }
   }
  }
  if ((label|0) == 19) {
   $66 = ((($1)) + 7|0);
   $67 = HEAP8[$66>>0]|0;
   $68 = $67&255;
   $69 = 1 << $68;
   $70 = (($69) + -1)|0;
   $71 = ((($1)) + 16|0);
   L28: do {
    if ($364) {
     if ($365) {
      $$pre = HEAP32[$71>>2]|0;
      $90 = $70;
      while(1) {
       $89 = (((($$pre) + ($90<<5)|0)) + 8|0);
       $91 = HEAP32[$89>>2]|0;
       $92 = ($91|0)==(0);
       if ($92) {
        $93 = (((($$pre) + ($90<<5)|0)) + 24|0);
        $94 = HEAP32[$93>>2]|0;
        $95 = ($94|0)>(3);
        if ($95) {
         HEAP32[$93>>2] = 11;
        }
       }
       $96 = (($90) + -1)|0;
       $97 = ($90|0)==(0);
       if ($97) {
        break L28;
       } else {
        $90 = $96;
       }
      }
     } else {
      $74 = $70;
     }
     while(1) {
      $72 = HEAP32[$71>>2]|0;
      $73 = (($72) + ($74<<5)|0);
      $75 = (((($72) + ($74<<5)|0)) + 8|0);
      $76 = HEAP32[$75>>2]|0;
      $77 = ($76|0)==(0);
      if ($77) {
       $84 = (((($72) + ($74<<5)|0)) + 24|0);
       $85 = HEAP32[$84>>2]|0;
       $86 = ($85|0)>(3);
       if ($86) {
        HEAP32[$84>>2] = 11;
       }
      } else {
       $78 = ($76|0)>(3);
       if ($78) {
        $79 = HEAP32[$73>>2]|0;
        $80 = ((($79)) + 5|0);
        $81 = HEAP8[$80>>0]|0;
        $82 = $81 & 3;
        $83 = ($82<<24>>24)==(0);
        if (!($83)) {
         _reallymarkobject($g,$79);
        }
       }
      }
      $87 = (($74) + -1)|0;
      $88 = ($74|0)==(0);
      if ($88) {
       break;
      } else {
       $74 = $87;
      }
     }
    } else {
     if ($365) {
      $100 = $70;
      while(1) {
       $98 = HEAP32[$71>>2]|0;
       $99 = (((($98) + ($100<<5)|0)) + 8|0);
       $101 = HEAP32[$99>>2]|0;
       $102 = ($101|0)==(0);
       $103 = (((($98) + ($100<<5)|0)) + 24|0);
       $104 = HEAP32[$103>>2]|0;
       $105 = ($104|0)>(3);
       if ($102) {
        if ($105) {
         HEAP32[$103>>2] = 11;
        }
       } else {
        if ($105) {
         $106 = (((($98) + ($100<<5)|0)) + 16|0);
         $107 = HEAP32[$106>>2]|0;
         $108 = ((($107)) + 5|0);
         $109 = HEAP8[$108>>0]|0;
         $110 = $109 & 3;
         $111 = ($110<<24>>24)==(0);
         if (!($111)) {
          _reallymarkobject($g,$107);
         }
        }
       }
       $112 = (($100) + -1)|0;
       $113 = ($100|0)==(0);
       if ($113) {
        break L28;
       } else {
        $100 = $112;
       }
      }
     } else {
      $116 = $70;
     }
     while(1) {
      $114 = HEAP32[$71>>2]|0;
      $115 = (($114) + ($116<<5)|0);
      $117 = (((($114) + ($116<<5)|0)) + 8|0);
      $118 = HEAP32[$117>>2]|0;
      $119 = ($118|0)==(0);
      $120 = (((($114) + ($116<<5)|0)) + 24|0);
      $121 = HEAP32[$120>>2]|0;
      $122 = ($121|0)>(3);
      if ($119) {
       if ($122) {
        HEAP32[$120>>2] = 11;
       }
      } else {
       if ($122) {
        $123 = (((($114) + ($116<<5)|0)) + 16|0);
        $124 = HEAP32[$123>>2]|0;
        $125 = ((($124)) + 5|0);
        $126 = HEAP8[$125>>0]|0;
        $127 = $126 & 3;
        $128 = ($127<<24>>24)==(0);
        if ($128) {
         $130 = $118;
        } else {
         _reallymarkobject($g,$124);
         $$pre$i = HEAP32[$117>>2]|0;
         $130 = $$pre$i;
        }
       } else {
        $130 = $118;
       }
       $129 = ($130|0)>(3);
       if ($129) {
        $131 = HEAP32[$115>>2]|0;
        $132 = ((($131)) + 5|0);
        $133 = HEAP8[$132>>0]|0;
        $134 = $133 & 3;
        $135 = ($134<<24>>24)==(0);
        if (!($135)) {
         _reallymarkobject($g,$131);
        }
       }
      }
      $136 = (($116) + -1)|0;
      $137 = ($116|0)==(0);
      if ($137) {
       break;
      } else {
       $116 = $136;
      }
     }
    }
   } while(0);
   $138 = $weakvalue$0912$i | $weakkey$0814$i;
   $139 = ($138|0)==(0);
   if ($139) {
    $$pre$phiZ2D = $66;
   } else {
    $$pre25$pre$phiZZ2D = $66;
    label = 52;
   }
  }
  if ((label|0) == 52) {
   $140 = HEAP8[$2>>0]|0;
   $141 = $140&255;
   $142 = $141 & 251;
   $143 = $142&255;
   HEAP8[$2>>0] = $143;
   $$pre$phiZ2D = $$pre25$pre$phiZZ2D;
  }
  $144 = ((($1)) + 28|0);
  $145 = HEAP32[$144>>2]|0;
  $146 = $145 << 4;
  $147 = (($146) + 32)|0;
  $148 = HEAP8[$$pre$phiZ2D>>0]|0;
  $149 = $148&255;
  $150 = 32 << $149;
  $151 = (($147) + ($150))|0;
  $$0 = $151;
  return ($$0|0);
  break;
 }
 case 6:  {
  $152 = ((($1)) + 8|0);
  $153 = HEAP32[$152>>2]|0;
  HEAP32[$0>>2] = $153;
  $154 = ((($1)) + 12|0);
  $155 = HEAP32[$154>>2]|0;
  $156 = ((($155)) + 5|0);
  $157 = HEAP8[$156>>0]|0;
  $158 = $157 & 3;
  $159 = ($158<<24>>24)==(0);
  if (!($159)) {
   _reallymarkobject($g,$155);
  }
  $160 = ((($1)) + 6|0);
  $161 = HEAP16[$160>>1]|0;
  $162 = $161&255;
  $163 = ($162<<24>>24)==(0);
  $164 = ($161&65535) >>> 8;
  $165 = $164&255;
  if ($163) {
   $181 = ((($1)) + 16|0);
   $182 = HEAP32[$181>>2]|0;
   $183 = ((($182)) + 5|0);
   $184 = HEAP8[$183>>0]|0;
   $185 = $184 & 3;
   $186 = ($185<<24>>24)==(0);
   if ($186) {
    $$pre8$i = ((($1)) + 7|0);
    $$pre$phi$iZ2D = $$pre8$i;$188 = $165;
   } else {
    _reallymarkobject($g,$182);
    $$phi$trans$insert$i = ((($1)) + 7|0);
    $$pre7$i = HEAP8[$$phi$trans$insert$i>>0]|0;
    $$pre$phi$iZ2D = $$phi$trans$insert$i;$188 = $$pre7$i;
   }
   $187 = ($188<<24>>24)==(0);
   if (!($187)) {
    $i1$03$i = 0;
    while(1) {
     $189 = (((($1)) + 20|0) + ($i1$03$i<<2)|0);
     $190 = HEAP32[$189>>2]|0;
     $191 = ((($190)) + 5|0);
     $192 = HEAP8[$191>>0]|0;
     $193 = $192 & 3;
     $194 = ($193<<24>>24)==(0);
     if (!($194)) {
      _reallymarkobject($g,$190);
     }
     $195 = (($i1$03$i) + 1)|0;
     $196 = HEAP8[$$pre$phi$iZ2D>>0]|0;
     $197 = $196&255;
     $198 = ($195|0)<($197|0);
     if ($198) {
      $i1$03$i = $195;
     } else {
      break;
     }
    }
   }
  } else {
   $166 = ((($1)) + 7|0);
   $167 = ($165<<24>>24)==(0);
   if (!($167)) {
    $366 = $165;$i$04$i = 0;
    while(1) {
     $168 = (((((($1)) + 24|0) + ($i$04$i<<4)|0)) + 8|0);
     $169 = HEAP32[$168>>2]|0;
     $170 = ($169|0)>(3);
     if ($170) {
      $171 = (((($1)) + 24|0) + ($i$04$i<<4)|0);
      $172 = HEAP32[$171>>2]|0;
      $173 = ((($172)) + 5|0);
      $174 = HEAP8[$173>>0]|0;
      $175 = $174 & 3;
      $176 = ($175<<24>>24)==(0);
      if ($176) {
       $179 = $366;
      } else {
       _reallymarkobject($g,$172);
       $$pre$i1 = HEAP8[$166>>0]|0;
       $179 = $$pre$i1;
      }
     } else {
      $179 = $366;
     }
     $177 = (($i$04$i) + 1)|0;
     $178 = $179&255;
     $180 = ($177|0)<($178|0);
     if ($180) {
      $366 = $179;$i$04$i = $177;
     } else {
      break;
     }
    }
   }
  }
  $199 = HEAP16[$160>>1]|0;
  $200 = $199&255;
  $201 = ($200<<24>>24)==(0);
  $202 = ($199&65535) >>> 8;
  $203 = $202&65535;
  if ($201) {
   $206 = $203 << 2;
   $207 = (($206) + 20)|0;
   $$0 = $207;
   return ($$0|0);
  } else {
   $204 = $203 << 4;
   $205 = (($204) + 24)|0;
   $$0 = $205;
   return ($$0|0);
  }
  break;
 }
 case 8:  {
  $208 = ($8&65535) >>> 8;
  $209 = $208&65535;
  $210 = ((($1)) + 108|0);
  $211 = HEAP32[$210>>2]|0;
  HEAP32[$0>>2] = $211;
  $212 = ((($g)) + 40|0);
  $213 = HEAP32[$212>>2]|0;
  HEAP32[$210>>2] = $213;
  HEAP32[$212>>2] = $1;
  $214 = $209 & 251;
  $215 = $214&255;
  HEAP8[$2>>0] = $215;
  $216 = ((($1)) + 80|0);
  $217 = HEAP32[$216>>2]|0;
  $218 = ($217|0)>(3);
  if ($218) {
   $219 = ((($1)) + 72|0);
   $220 = HEAP32[$219>>2]|0;
   $221 = ((($220)) + 5|0);
   $222 = HEAP8[$221>>0]|0;
   $223 = $222 & 3;
   $224 = ($223<<24>>24)==(0);
   if (!($224)) {
    _reallymarkobject($g,$220);
   }
  }
  $225 = ((($1)) + 8|0);
  $226 = HEAP32[$225>>2]|0;
  $227 = ((($1)) + 40|0);
  $228 = HEAP32[$227>>2]|0;
  $229 = ((($1)) + 20|0);
  $230 = HEAP32[$229>>2]|0;
  $231 = ($228>>>0)>($230>>>0);
  if ($231) {
   $lim$0$lcssa$i = $226;
  } else {
   $ci$06$i = $228;$lim$05$i = $226;
   while(1) {
    $232 = ((($ci$06$i)) + 8|0);
    $233 = HEAP32[$232>>2]|0;
    $234 = ($lim$05$i>>>0)<($233>>>0);
    $$lim$0$i = $234 ? $233 : $lim$05$i;
    $235 = ((($ci$06$i)) + 24|0);
    $236 = ($235>>>0)>($230>>>0);
    if ($236) {
     $lim$0$lcssa$i = $$lim$0$i;
     break;
    } else {
     $ci$06$i = $235;$lim$05$i = $$lim$0$i;
    }
   }
  }
  $237 = ((($1)) + 32|0);
  $238 = HEAP32[$237>>2]|0;
  $239 = ($238>>>0)<($226>>>0);
  if ($239) {
   $367 = $226;$o$02$i = $238;
   while(1) {
    $241 = ((($o$02$i)) + 8|0);
    $242 = HEAP32[$241>>2]|0;
    $243 = ($242|0)>(3);
    if ($243) {
     $244 = HEAP32[$o$02$i>>2]|0;
     $245 = ((($244)) + 5|0);
     $246 = HEAP8[$245>>0]|0;
     $247 = $246 & 3;
     $248 = ($247<<24>>24)==(0);
     if ($248) {
      $251 = $367;
     } else {
      _reallymarkobject($g,$244);
      $$pre$i5 = HEAP32[$225>>2]|0;
      $251 = $$pre$i5;
     }
    } else {
     $251 = $367;
    }
    $249 = ((($o$02$i)) + 16|0);
    $250 = ($249>>>0)<($251>>>0);
    if ($250) {
     $367 = $251;$o$02$i = $249;
    } else {
     $o$0$lcssa$i = $249;
     break;
    }
   }
  } else {
   $o$0$lcssa$i = $238;
  }
  $240 = ($o$0$lcssa$i>>>0)>($lim$0$lcssa$i>>>0);
  if (!($240)) {
   $o$11$i = $o$0$lcssa$i;
   while(1) {
    $252 = ((($o$11$i)) + 8|0);
    HEAP32[$252>>2] = 0;
    $253 = ((($o$11$i)) + 16|0);
    $254 = ($253>>>0)>($lim$0$lcssa$i>>>0);
    if ($254) {
     break;
    } else {
     $o$11$i = $253;
    }
   }
  }
  $255 = HEAP32[$237>>2]|0;
  $256 = $lim$0$lcssa$i;
  $257 = (($256) - ($255))|0;
  $258 = ((($1)) + 48|0);
  $259 = HEAP32[$258>>2]|0;
  $260 = ($259|0)>(20000);
  if ($260) {
   $$pre27 = ((($1)) + 44|0);
   $$pre$phi28Z2D = $$pre27;
  } else {
   $261 = HEAP32[$229>>2]|0;
   $262 = HEAP32[$227>>2]|0;
   $263 = (($261) - ($262))|0;
   $264 = (($263|0) / 24)&-1;
   $265 = $264 << 2;
   $266 = ($265|0)<($259|0);
   $267 = ($259|0)>(16);
   $or$cond$i$i = $267 & $266;
   if ($or$cond$i$i) {
    $268 = (($259|0) / 2)&-1;
    _luaD_reallocCI($1,$268);
   }
   $269 = $257 >> 2;
   $270 = ((($1)) + 44|0);
   $271 = HEAP32[$270>>2]|0;
   $272 = ($269|0)<($271|0);
   $273 = ($271|0)>(90);
   $or$cond1$i$i = $272 & $273;
   if ($or$cond1$i$i) {
    $274 = (($271|0) / 2)&-1;
    _luaD_reallocstack($1,$274);
    $$pre$phi28Z2D = $270;
   } else {
    $$pre$phi28Z2D = $270;
   }
  }
  $275 = HEAP32[$$pre$phi28Z2D>>2]|0;
  $276 = $275 << 4;
  $277 = (($276) + 120)|0;
  $278 = HEAP32[$258>>2]|0;
  $279 = ($278*24)|0;
  $280 = (($277) + ($279))|0;
  $$0 = $280;
  return ($$0|0);
  break;
 }
 case 9:  {
  $281 = ((($1)) + 68|0);
  $282 = HEAP32[$281>>2]|0;
  HEAP32[$0>>2] = $282;
  $283 = ((($1)) + 32|0);
  $284 = HEAP32[$283>>2]|0;
  $285 = ($284|0)==(0|0);
  if (!($285)) {
   $286 = ((($284)) + 5|0);
   $287 = HEAP8[$286>>0]|0;
   $288 = $287&255;
   $289 = $288 & 252;
   $290 = $289&255;
   HEAP8[$286>>0] = $290;
  }
  $291 = ((($1)) + 40|0);
  $292 = HEAP32[$291>>2]|0;
  $293 = ($292|0)>(0);
  if ($293) {
   $294 = ((($1)) + 8|0);
   $368 = $292;$i$09$i = 0;
   while(1) {
    $300 = HEAP32[$294>>2]|0;
    $301 = (((($300) + ($i$09$i<<4)|0)) + 8|0);
    $302 = HEAP32[$301>>2]|0;
    $303 = ($302|0)>(3);
    if ($303) {
     $304 = (($300) + ($i$09$i<<4)|0);
     $305 = HEAP32[$304>>2]|0;
     $306 = ((($305)) + 5|0);
     $307 = HEAP8[$306>>0]|0;
     $308 = $307 & 3;
     $309 = ($308<<24>>24)==(0);
     if ($309) {
      $312 = $368;
     } else {
      _reallymarkobject($g,$305);
      $$pre$i8 = HEAP32[$291>>2]|0;
      $312 = $$pre$i8;
     }
    } else {
     $312 = $368;
    }
    $310 = (($i$09$i) + 1)|0;
    $311 = ($310|0)<($312|0);
    if ($311) {
     $368 = $312;$i$09$i = $310;
    } else {
     break;
    }
   }
  }
  $295 = ((($1)) + 36|0);
  $296 = HEAP32[$295>>2]|0;
  $297 = ($296|0)>(0);
  if ($297) {
   $298 = ((($1)) + 28|0);
   $299 = HEAP32[$298>>2]|0;
   $i$17$i = 0;
   while(1) {
    $317 = (($299) + ($i$17$i<<2)|0);
    $318 = HEAP32[$317>>2]|0;
    $319 = ($318|0)==(0|0);
    if (!($319)) {
     $320 = ((($318)) + 5|0);
     $321 = HEAP8[$320>>0]|0;
     $322 = $321&255;
     $323 = $322 & 252;
     $324 = $323&255;
     HEAP8[$320>>0] = $324;
    }
    $325 = (($i$17$i) + 1)|0;
    $exitcond24 = ($325|0)==($296|0);
    if ($exitcond24) {
     break;
    } else {
     $i$17$i = $325;
    }
   }
  }
  $313 = ((($1)) + 52|0);
  $314 = HEAP32[$313>>2]|0;
  $315 = ($314|0)>(0);
  if ($315) {
   $316 = ((($1)) + 16|0);
   $369 = $314;$i$25$i = 0;
   while(1) {
    $331 = HEAP32[$316>>2]|0;
    $332 = (($331) + ($i$25$i<<2)|0);
    $333 = HEAP32[$332>>2]|0;
    $334 = ($333|0)==(0|0);
    if ($334) {
     $341 = $369;
    } else {
     $335 = ((($333)) + 5|0);
     $336 = HEAP8[$335>>0]|0;
     $337 = $336 & 3;
     $338 = ($337<<24>>24)==(0);
     if ($338) {
      $341 = $369;
     } else {
      _reallymarkobject($g,$333);
      $$pre11$i = HEAP32[$313>>2]|0;
      $341 = $$pre11$i;
     }
    }
    $339 = (($i$25$i) + 1)|0;
    $340 = ($339|0)<($341|0);
    if ($340) {
     $369 = $341;$i$25$i = $339;
    } else {
     $359 = $341;
     break;
    }
   }
  } else {
   $359 = $314;
  }
  $326 = ((($1)) + 56|0);
  $327 = HEAP32[$326>>2]|0;
  $328 = ($327|0)>(0);
  if ($328) {
   $329 = ((($1)) + 24|0);
   $330 = HEAP32[$329>>2]|0;
   $i$34$i = 0;
   while(1) {
    $342 = (($330) + (($i$34$i*12)|0)|0);
    $343 = HEAP32[$342>>2]|0;
    $344 = ($343|0)==(0|0);
    if (!($344)) {
     $345 = ((($343)) + 5|0);
     $346 = HEAP8[$345>>0]|0;
     $347 = $346&255;
     $348 = $347 & 252;
     $349 = $348&255;
     HEAP8[$345>>0] = $349;
    }
    $350 = (($i$34$i) + 1)|0;
    $exitcond = ($350|0)==($327|0);
    if ($exitcond) {
     break;
    } else {
     $i$34$i = $350;
    }
   }
  }
  $351 = ((($1)) + 44|0);
  $352 = HEAP32[$351>>2]|0;
  $353 = HEAP32[$291>>2]|0;
  $354 = $353 << 4;
  $355 = ((($1)) + 48|0);
  $356 = HEAP32[$355>>2]|0;
  $357 = ($327*12)|0;
  $358 = HEAP32[$295>>2]|0;
  $tmp = (($359) + ($352))|0;
  $tmp13 = (($tmp) + ($356))|0;
  $tmp14 = (($tmp13) + ($358))|0;
  $tmp15 = $tmp14 << 2;
  $360 = (($354) + 76)|0;
  $361 = (($360) + ($357))|0;
  $362 = (($361) + ($tmp15))|0;
  $$0 = $362;
  return ($$0|0);
  break;
 }
 default: {
  $$0 = 0;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function _luaC_fullgc($L) {
 $L = $L|0;
 var $$pr = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 21|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = ($3&255)<(2);
 if ($4) {
  $5 = ((($1)) + 24|0);
  HEAP32[$5>>2] = 0;
  $6 = ((($1)) + 28|0);
  $7 = ((($1)) + 32|0);
  HEAP32[$7>>2] = $6;
  $8 = ((($1)) + 36|0);
  HEAP32[$8>>2] = 0;
  $9 = ((($1)) + 40|0);
  HEAP32[$9>>2] = 0;
  $10 = ((($1)) + 44|0);
  HEAP32[$10>>2] = 0;
  HEAP8[$2>>0] = 2;
  label = 4;
 } else {
  $11 = ($3<<24>>24)==(4);
  if (!($11)) {
   label = 4;
  }
 }
 if ((label|0) == 4) {
  while(1) {
   label = 0;
   (_singlestep($L)|0);
   $$pr = HEAP8[$2>>0]|0;
   $12 = ($$pr<<24>>24)==(4);
   if ($12) {
    break;
   } else {
    label = 4;
   }
  }
 }
 _markroot($L);
 $13 = HEAP8[$2>>0]|0;
 $14 = ($13<<24>>24)==(0);
 if (!($14)) {
  while(1) {
   (_singlestep($L)|0);
   $15 = HEAP8[$2>>0]|0;
   $16 = ($15<<24>>24)==(0);
   if ($16) {
    break;
   }
  }
 }
 $17 = ((($1)) + 72|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = (($18>>>0) / 100)&-1;
 $20 = ((($1)) + 80|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = Math_imul($19, $21)|0;
 $23 = ((($1)) + 64|0);
 HEAP32[$23>>2] = $22;
 return;
}
function _luaC_barrierf($L,$o,$v) {
 $L = $L|0;
 $o = $o|0;
 $v = $v|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 21|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = ($3<<24>>24)==(1);
 if ($4) {
  _reallymarkobject($1,$v);
  return;
 } else {
  $5 = ((($o)) + 5|0);
  $6 = HEAP8[$5>>0]|0;
  $7 = $6&255;
  $8 = $7 & 248;
  $9 = ((($1)) + 20|0);
  $10 = HEAP8[$9>>0]|0;
  $11 = $10&255;
  $12 = $11 & 3;
  $13 = $12 | $8;
  $14 = $13&255;
  HEAP8[$5>>0] = $14;
  return;
 }
}
function _luaC_barrierback($L,$t) {
 $L = $L|0;
 $t = $t|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($t)) + 5|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = $4 & 251;
 $6 = $5&255;
 HEAP8[$2>>0] = $6;
 $7 = ((($1)) + 40|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($t)) + 24|0);
 HEAP32[$9>>2] = $8;
 HEAP32[$7>>2] = $t;
 return;
}
function _luaC_link($L,$o,$tt) {
 $L = $L|0;
 $o = $o|0;
 $tt = $tt|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$o>>2] = $3;
 HEAP32[$2>>2] = $o;
 $4 = ((($1)) + 20|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = $6 & 3;
 $8 = $7&255;
 $9 = ((($o)) + 5|0);
 HEAP8[$9>>0] = $8;
 $10 = ((($o)) + 4|0);
 HEAP8[$10>>0] = $tt;
 return;
}
function _luaC_linkupval($L,$uv) {
 $L = $L|0;
 $uv = $uv|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$uv>>2] = $3;
 HEAP32[$2>>2] = $uv;
 $4 = ((($uv)) + 5|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = $6 & 7;
 $8 = ($7|0)==(0);
 if (!($8)) {
  return;
 }
 $9 = ((($1)) + 21|0);
 $10 = HEAP8[$9>>0]|0;
 $11 = ($10<<24>>24)==(1);
 if (!($11)) {
  $24 = $6 & 248;
  $25 = ((($1)) + 20|0);
  $26 = HEAP8[$25>>0]|0;
  $27 = $26&255;
  $28 = $27 & 3;
  $29 = $28 | $24;
  $30 = $29&255;
  HEAP8[$4>>0] = $30;
  return;
 }
 $12 = $6 | 4;
 $13 = $12&255;
 HEAP8[$4>>0] = $13;
 $14 = ((($uv)) + 8|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($15)) + 8|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ($17|0)>(3);
 if (!($18)) {
  return;
 }
 $19 = HEAP32[$15>>2]|0;
 $20 = ((($19)) + 5|0);
 $21 = HEAP8[$20>>0]|0;
 $22 = $21 & 3;
 $23 = ($22<<24>>24)==(0);
 if ($23) {
  return;
 }
 _reallymarkobject($1,$19);
 return;
}
function _luaM_growaux_($L,$block,$size,$size_elems,$limit,$errormsg) {
 $L = $L|0;
 $block = $block|0;
 $size = $size|0;
 $size_elems = $size_elems|0;
 $limit = $limit|0;
 $errormsg = $errormsg|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $newsize$0 = 0, $or$cond$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[$size>>2]|0;
 $1 = (($limit|0) / 2)&-1;
 $2 = ($0|0)<($1|0);
 if ($2) {
  $4 = $0 << 1;
  $5 = ($4|0)<(4);
  $$ = $5 ? 4 : $4;
  $newsize$0 = $$;
 } else {
  $3 = ($0|0)<($limit|0);
  if ($3) {
   $newsize$0 = $limit;
  } else {
   _luaG_runerror($L,$errormsg,$vararg_buffer);
   $newsize$0 = $limit;
  }
 }
 $6 = (($newsize$0) + 1)|0;
 $7 = (4294967293 / ($size_elems>>>0))&-1;
 $8 = ($6>>>0)>($7>>>0);
 if ($8) {
  _luaG_runerror($L,517936,$vararg_buffer1);
  $25 = 0;
  HEAP32[$size>>2] = $newsize$0;
  STACKTOP = sp;return ($25|0);
 }
 $9 = HEAP32[$size>>2]|0;
 $10 = Math_imul($9, $size_elems)|0;
 $11 = Math_imul($newsize$0, $size_elems)|0;
 $12 = ((($L)) + 16|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($13)) + 12|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($13)) + 16|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = (FUNCTION_TABLE_iiiii[$15 & 8191]($17,$block,$10,$11)|0);
 $19 = ($18|0)==(0|0);
 $20 = ($11|0)!=(0);
 $or$cond$i = $20 & $19;
 if ($or$cond$i) {
  _luaD_throw($L,4);
  // unreachable;
 }
 $21 = ((($13)) + 68|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = (($11) - ($10))|0;
 $24 = (($23) + ($22))|0;
 HEAP32[$21>>2] = $24;
 $25 = $18;
 HEAP32[$size>>2] = $newsize$0;
 STACKTOP = sp;return ($25|0);
}
function _luaM_realloc_($L,$block,$osize,$nsize) {
 $L = $L|0;
 $block = $block|0;
 $osize = $osize|0;
 $nsize = $nsize|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiiii[$3 & 8191]($5,$block,$osize,$nsize)|0);
 $7 = ($6|0)==(0|0);
 $8 = ($nsize|0)!=(0);
 $or$cond = $8 & $7;
 if ($or$cond) {
  _luaD_throw($L,4);
  // unreachable;
 } else {
  $9 = ((($1)) + 68|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = (($nsize) - ($osize))|0;
  $12 = (($11) + ($10))|0;
  HEAP32[$9>>2] = $12;
  return ($6|0);
 }
 return (0)|0;
}
function _luaM_toobig($L) {
 $L = $L|0;
 var $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _luaG_runerror($L,517936,$vararg_buffer);
 STACKTOP = sp;return (0|0);
}
function _luaO_int2fb($x) {
 $x = $x|0;
 var $$0$lcssa = 0, $$01 = 0, $$02 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $e$0$lcssa = 0, $e$03 = 0, $phitmp = 0, $phitmp$lcssa = 0, $phitmp5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($x>>>0)>(15);
 if ($0) {
  $$02 = $x;$e$03 = 1;
  while(1) {
   $1 = (($$02) + 1)|0;
   $2 = $1 >>> 1;
   $phitmp = (($e$03) + 1)|0;
   $3 = ($1>>>0)>(31);
   if ($3) {
    $$02 = $2;$e$03 = $phitmp;
   } else {
    $$lcssa = $2;$phitmp$lcssa = $phitmp;
    break;
   }
  }
  $phitmp5 = $phitmp$lcssa << 3;
  $$0$lcssa = $$lcssa;$e$0$lcssa = $phitmp5;
 } else {
  $$0$lcssa = $x;$e$0$lcssa = 8;
 }
 $4 = ($$0$lcssa>>>0)<(8);
 $5 = (($$0$lcssa) + -8)|0;
 $6 = $e$0$lcssa | $5;
 $$01 = $4 ? $$0$lcssa : $6;
 return ($$01|0);
}
function _luaO_fb2int($x) {
 $x = $x|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $x >>> 3;
 $1 = $0 & 31;
 $2 = ($1|0)==(0);
 if ($2) {
  $$0 = $x;
  return ($$0|0);
 }
 $3 = $x & 7;
 $4 = $3 | 8;
 $5 = (($1) + -1)|0;
 $6 = $4 << $5;
 $$0 = $6;
 return ($$0|0);
}
function _luaO_log2($x) {
 $x = $x|0;
 var $$0$lcssa = 0, $$01 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $l$0$lcssa = 0, $l$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($x>>>0)>(255);
 if ($0) {
  $$01 = $x;$l$02 = -1;
  while(1) {
   $1 = (($l$02) + 8)|0;
   $2 = $$01 >>> 8;
   $3 = ($$01>>>0)>(65535);
   if ($3) {
    $$01 = $2;$l$02 = $1;
   } else {
    $$0$lcssa = $2;$l$0$lcssa = $1;
    break;
   }
  }
 } else {
  $$0$lcssa = $x;$l$0$lcssa = -1;
 }
 $4 = (517975 + ($$0$lcssa)|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = (($6) + ($l$0$lcssa))|0;
 return ($7|0);
}
function _luaO_rawequalObj($t1,$t2) {
 $t1 = $t1|0;
 $t2 = $t2|0;
 var $$0 = 0, $$0$shrunk = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($t1)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($t2)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==($3|0);
 L1: do {
  if ($4) {
   switch ($1|0) {
   case 0:  {
    $$0$shrunk = 1;
    break L1;
    break;
   }
   case 3:  {
    $5 = +HEAPF64[$t1>>3];
    $6 = +HEAPF64[$t2>>3];
    $7 = $5 == $6;
    $$0$shrunk = $7;
    break L1;
    break;
   }
   case 1:  {
    $8 = HEAP32[$t1>>2]|0;
    $9 = HEAP32[$t2>>2]|0;
    $10 = ($8|0)==($9|0);
    $$0$shrunk = $10;
    break L1;
    break;
   }
   case 2:  {
    $11 = HEAP32[$t1>>2]|0;
    $12 = HEAP32[$t2>>2]|0;
    $13 = ($11|0)==($12|0);
    $$0$shrunk = $13;
    break L1;
    break;
   }
   default: {
    $14 = HEAP32[$t1>>2]|0;
    $15 = HEAP32[$t2>>2]|0;
    $16 = ($14|0)==($15|0);
    $$0$shrunk = $16;
    break L1;
   }
   }
  } else {
   $$0$shrunk = 0;
  }
 } while(0);
 $$0 = $$0$shrunk&1;
 return ($$0|0);
}
function _luaO_str2d($s,$result) {
 $s = $s|0;
 $result = $result|0;
 var $$ = 0, $$0 = 0, $$lcssa = 0, $$lcssa7 = 0, $$pre = 0, $$pre3 = 0, $0 = 0.0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0;
 var $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $endptr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $endptr = sp;
 $0 = (+_strtod($s,$endptr));
 HEAPF64[$result>>3] = $0;
 $1 = HEAP32[$endptr>>2]|0;
 $2 = ($1|0)==($s|0);
 if ($2) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = HEAP8[$1>>0]|0;
 switch ($3<<24>>24) {
 case 88: case 120:  {
  $4 = (_strtoul($s,$endptr,16)|0);
  $5 = (+($4>>>0));
  HEAPF64[$result>>3] = $5;
  $$pre = HEAP32[$endptr>>2]|0;
  $$pre3 = HEAP8[$$pre>>0]|0;
  $19 = $$pre;$7 = $$pre3;
  break;
 }
 default: {
  $19 = $1;$7 = $3;
 }
 }
 $6 = ($7<<24>>24)==(0);
 if ($6) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 $8 = $7&255;
 $9 = (_isspace($8)|0);
 $10 = ($9|0)==(0);
 if ($10) {
  $$lcssa = $19;
 } else {
  $12 = $19;
  while(1) {
   $11 = ((($12)) + 1|0);
   $13 = HEAP8[$11>>0]|0;
   $14 = $13&255;
   $15 = (_isspace($14)|0);
   $16 = ($15|0)==(0);
   if ($16) {
    $$lcssa7 = $11;
    break;
   } else {
    $12 = $11;
   }
  }
  HEAP32[$endptr>>2] = $$lcssa7;
  $$lcssa = $$lcssa7;
 }
 $17 = HEAP8[$$lcssa>>0]|0;
 $18 = ($17<<24>>24)==(0);
 $$ = $18&1;
 $$0 = $$;
 STACKTOP = sp;return ($$0|0);
}
function _luaO_pushvfstring($L,$fmt,$argp) {
 $L = $L|0;
 $fmt = $fmt|0;
 $argp = $argp|0;
 var $$ = 0, $$0$lcssa = 0, $$015 = 0, $$cast = 0, $$lcssa = 0, $$pre = 0, $$pre$i = 0, $$pre$i10 = 0, $$pre$i12 = 0, $$pre$i2 = 0, $$pre$i4 = 0, $$pre$i6 = 0, $$pre$i8 = 0, $$pre21 = 0, $$pre22 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0.0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0.0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $arglist_current = 0, $arglist_current11 = 0, $arglist_current2 = 0, $arglist_current5 = 0, $arglist_current8 = 0, $arglist_next = 0, $arglist_next12 = 0, $arglist_next3 = 0, $arglist_next6 = 0, $arglist_next9 = 0, $buff = 0, $buff3 = 0, $buff4 = 0, $expanded = 0, $expanded13 = 0, $expanded15 = 0, $expanded16 = 0;
 var $expanded17 = 0, $expanded19 = 0, $expanded20 = 0, $expanded22 = 0, $expanded23 = 0, $expanded24 = 0, $expanded26 = 0, $expanded27 = 0, $expanded29 = 0, $expanded30 = 0, $expanded31 = 0, $expanded33 = 0, $expanded34 = 0, $expanded36 = 0, $expanded37 = 0, $expanded38 = 0, $expanded40 = 0, $expanded41 = 0, $expanded43 = 0, $expanded44 = 0;
 var $expanded45 = 0, $n$0$lcssa = 0, $n$014 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $buff = sp + 32|0;
 $buff3 = sp + 8|0;
 $buff4 = sp + 4|0;
 $0 = ((($L)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_luaS_newlstr($L,1282466,0)|0);
 HEAP32[$1>>2] = $2;
 $3 = ((($1)) + 8|0);
 HEAP32[$3>>2] = 4;
 $4 = ((($L)) + 28|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = HEAP32[$0>>2]|0;
 $7 = (($5) - ($6))|0;
 $8 = ($7|0)<(17);
 $9 = $6;
 if ($8) {
  _luaD_growstack($L,1);
  $$pre$i = HEAP32[$0>>2]|0;
  $11 = $$pre$i;
 } else {
  $11 = $9;
 }
 $10 = ((($11)) + 16|0);
 HEAP32[$0>>2] = $10;
 $12 = (_strchr($fmt,37)|0);
 $13 = ($12|0)==(0|0);
 L4: do {
  if ($13) {
   $$0$lcssa = $fmt;$$lcssa = $10;$n$0$lcssa = 1;
  } else {
   $14 = ((($buff4)) + 1|0);
   $15 = ((($buff4)) + 2|0);
   $16 = ((($buff)) + 1|0);
   $$015 = $fmt;$18 = $12;$22 = $10;$n$014 = 1;
   while(1) {
    $17 = $18;
    $19 = $$015;
    $20 = (($17) - ($19))|0;
    $21 = (_luaS_newlstr($L,$$015,$20)|0);
    HEAP32[$22>>2] = $21;
    $23 = ((($22)) + 8|0);
    HEAP32[$23>>2] = 4;
    $24 = HEAP32[$4>>2]|0;
    $25 = HEAP32[$0>>2]|0;
    $26 = (($24) - ($25))|0;
    $27 = ($26|0)<(17);
    $28 = $25;
    if ($27) {
     _luaD_growstack($L,1);
     $$pre = HEAP32[$0>>2]|0;
     $30 = $$pre;
    } else {
     $30 = $28;
    }
    $29 = ((($30)) + 16|0);
    HEAP32[$0>>2] = $29;
    $31 = ((($18)) + 1|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32 << 24 >> 24;
    switch ($33|0) {
    case 115:  {
     $arglist_current = HEAP32[$argp>>2]|0;
     $34 = $arglist_current;
     $35 = ((0) + 4|0);
     $expanded13 = $35;
     $expanded = (($expanded13) - 1)|0;
     $36 = (($34) + ($expanded))|0;
     $37 = ((0) + 4|0);
     $expanded17 = $37;
     $expanded16 = (($expanded17) - 1)|0;
     $expanded15 = $expanded16 ^ -1;
     $38 = $36 & $expanded15;
     $39 = $38;
     $40 = HEAP32[$39>>2]|0;
     $arglist_next = ((($39)) + 4|0);
     HEAP32[$argp>>2] = $arglist_next;
     $41 = ($40|0)==(0|0);
     $$ = $41 ? 619383 : $40;
     $42 = HEAP32[$0>>2]|0;
     $43 = (_strlen($$)|0);
     $44 = (_luaS_newlstr($L,$$,$43)|0);
     HEAP32[$42>>2] = $44;
     $45 = ((($42)) + 8|0);
     HEAP32[$45>>2] = 4;
     $46 = HEAP32[$4>>2]|0;
     $47 = HEAP32[$0>>2]|0;
     $48 = (($46) - ($47))|0;
     $49 = ($48|0)<(17);
     $50 = $47;
     if ($49) {
      _luaD_growstack($L,1);
      $$pre$i2 = HEAP32[$0>>2]|0;
      $52 = $$pre$i2;
     } else {
      $52 = $50;
     }
     $51 = ((($52)) + 16|0);
     HEAP32[$0>>2] = $51;
     $167 = $51;
     break;
    }
    case 99:  {
     $arglist_current2 = HEAP32[$argp>>2]|0;
     $53 = $arglist_current2;
     $54 = ((0) + 4|0);
     $expanded20 = $54;
     $expanded19 = (($expanded20) - 1)|0;
     $55 = (($53) + ($expanded19))|0;
     $56 = ((0) + 4|0);
     $expanded24 = $56;
     $expanded23 = (($expanded24) - 1)|0;
     $expanded22 = $expanded23 ^ -1;
     $57 = $55 & $expanded22;
     $58 = $57;
     $59 = HEAP32[$58>>2]|0;
     $arglist_next3 = ((($58)) + 4|0);
     HEAP32[$argp>>2] = $arglist_next3;
     $60 = $59&255;
     HEAP8[$buff>>0] = $60;
     HEAP8[$16>>0] = 0;
     $61 = HEAP32[$0>>2]|0;
     $62 = (_strlen($buff)|0);
     $63 = (_luaS_newlstr($L,$buff,$62)|0);
     HEAP32[$61>>2] = $63;
     $64 = ((($61)) + 8|0);
     HEAP32[$64>>2] = 4;
     $65 = HEAP32[$4>>2]|0;
     $66 = HEAP32[$0>>2]|0;
     $67 = (($65) - ($66))|0;
     $68 = ($67|0)<(17);
     $69 = $66;
     if ($68) {
      _luaD_growstack($L,1);
      $$pre$i4 = HEAP32[$0>>2]|0;
      $71 = $$pre$i4;
     } else {
      $71 = $69;
     }
     $70 = ((($71)) + 16|0);
     HEAP32[$0>>2] = $70;
     $167 = $70;
     break;
    }
    case 100:  {
     $arglist_current5 = HEAP32[$argp>>2]|0;
     $72 = $arglist_current5;
     $73 = ((0) + 4|0);
     $expanded27 = $73;
     $expanded26 = (($expanded27) - 1)|0;
     $74 = (($72) + ($expanded26))|0;
     $75 = ((0) + 4|0);
     $expanded31 = $75;
     $expanded30 = (($expanded31) - 1)|0;
     $expanded29 = $expanded30 ^ -1;
     $76 = $74 & $expanded29;
     $77 = $76;
     $78 = HEAP32[$77>>2]|0;
     $arglist_next6 = ((($77)) + 4|0);
     HEAP32[$argp>>2] = $arglist_next6;
     $79 = (+($78|0));
     HEAPF64[$29>>3] = $79;
     $80 = ((($30)) + 24|0);
     HEAP32[$80>>2] = 3;
     $81 = HEAP32[$4>>2]|0;
     $82 = HEAP32[$0>>2]|0;
     $83 = (($81) - ($82))|0;
     $84 = ($83|0)<(17);
     $85 = $82;
     if ($84) {
      _luaD_growstack($L,1);
      $$pre22 = HEAP32[$0>>2]|0;
      $87 = $$pre22;
     } else {
      $87 = $85;
     }
     $86 = ((($87)) + 16|0);
     HEAP32[$0>>2] = $86;
     $167 = $86;
     break;
    }
    case 102:  {
     $arglist_current8 = HEAP32[$argp>>2]|0;
     $88 = $arglist_current8;
     $89 = ((0) + 8|0);
     $expanded34 = $89;
     $expanded33 = (($expanded34) - 1)|0;
     $90 = (($88) + ($expanded33))|0;
     $91 = ((0) + 8|0);
     $expanded38 = $91;
     $expanded37 = (($expanded38) - 1)|0;
     $expanded36 = $expanded37 ^ -1;
     $92 = $90 & $expanded36;
     $93 = $92;
     $94 = +HEAPF64[$93>>3];
     $arglist_next9 = ((($93)) + 8|0);
     HEAP32[$argp>>2] = $arglist_next9;
     HEAPF64[$29>>3] = $94;
     $95 = ((($30)) + 24|0);
     HEAP32[$95>>2] = 3;
     $96 = HEAP32[$4>>2]|0;
     $97 = HEAP32[$0>>2]|0;
     $98 = (($96) - ($97))|0;
     $99 = ($98|0)<(17);
     $100 = $97;
     if ($99) {
      _luaD_growstack($L,1);
      $$pre21 = HEAP32[$0>>2]|0;
      $102 = $$pre21;
     } else {
      $102 = $100;
     }
     $101 = ((($102)) + 16|0);
     HEAP32[$0>>2] = $101;
     $167 = $101;
     break;
    }
    case 112:  {
     $arglist_current11 = HEAP32[$argp>>2]|0;
     $103 = $arglist_current11;
     $104 = ((0) + 4|0);
     $expanded41 = $104;
     $expanded40 = (($expanded41) - 1)|0;
     $105 = (($103) + ($expanded40))|0;
     $106 = ((0) + 4|0);
     $expanded45 = $106;
     $expanded44 = (($expanded45) - 1)|0;
     $expanded43 = $expanded44 ^ -1;
     $107 = $105 & $expanded43;
     $108 = $107;
     $109 = HEAP32[$108>>2]|0;
     $arglist_next12 = ((($108)) + 4|0);
     HEAP32[$argp>>2] = $arglist_next12;
     HEAP32[$vararg_buffer>>2] = $109;
     (_sprintf($buff3,620521,$vararg_buffer)|0);
     $110 = HEAP32[$0>>2]|0;
     $111 = (_strlen($buff3)|0);
     $112 = (_luaS_newlstr($L,$buff3,$111)|0);
     HEAP32[$110>>2] = $112;
     $113 = ((($110)) + 8|0);
     HEAP32[$113>>2] = 4;
     $114 = HEAP32[$4>>2]|0;
     $115 = HEAP32[$0>>2]|0;
     $116 = (($114) - ($115))|0;
     $117 = ($116|0)<(17);
     $118 = $115;
     if ($117) {
      _luaD_growstack($L,1);
      $$pre$i6 = HEAP32[$0>>2]|0;
      $120 = $$pre$i6;
     } else {
      $120 = $118;
     }
     $119 = ((($120)) + 16|0);
     HEAP32[$0>>2] = $119;
     $167 = $119;
     break;
    }
    case 37:  {
     $121 = (_luaS_newlstr($L,530575,1)|0);
     HEAP32[$29>>2] = $121;
     $122 = ((($30)) + 24|0);
     HEAP32[$122>>2] = 4;
     $123 = HEAP32[$4>>2]|0;
     $124 = HEAP32[$0>>2]|0;
     $125 = (($123) - ($124))|0;
     $126 = ($125|0)<(17);
     $127 = $124;
     if ($126) {
      _luaD_growstack($L,1);
      $$pre$i8 = HEAP32[$0>>2]|0;
      $129 = $$pre$i8;
     } else {
      $129 = $127;
     }
     $128 = ((($129)) + 16|0);
     HEAP32[$0>>2] = $128;
     $167 = $128;
     break;
    }
    default: {
     HEAP8[$buff4>>0] = 37;
     $130 = HEAP8[$31>>0]|0;
     HEAP8[$14>>0] = $130;
     HEAP8[$15>>0] = 0;
     $131 = (_strlen($buff4)|0);
     $132 = (_luaS_newlstr($L,$buff4,$131)|0);
     HEAP32[$29>>2] = $132;
     $133 = ((($30)) + 24|0);
     HEAP32[$133>>2] = 4;
     $134 = HEAP32[$4>>2]|0;
     $135 = HEAP32[$0>>2]|0;
     $136 = (($134) - ($135))|0;
     $137 = ($136|0)<(17);
     $138 = $135;
     if ($137) {
      _luaD_growstack($L,1);
      $$pre$i10 = HEAP32[$0>>2]|0;
      $140 = $$pre$i10;
     } else {
      $140 = $138;
     }
     $139 = ((($140)) + 16|0);
     HEAP32[$0>>2] = $139;
     $167 = $139;
    }
    }
    $141 = (($n$014) + 2)|0;
    $142 = ((($18)) + 2|0);
    $143 = (_strchr($142,37)|0);
    $144 = ($143|0)==(0|0);
    if ($144) {
     $$0$lcssa = $142;$$lcssa = $167;$n$0$lcssa = $141;
     break L4;
    } else {
     $$015 = $142;$18 = $143;$22 = $167;$n$014 = $141;
    }
   }
  }
 } while(0);
 $145 = (_strlen($$0$lcssa)|0);
 $146 = (_luaS_newlstr($L,$$0$lcssa,$145)|0);
 HEAP32[$$lcssa>>2] = $146;
 $147 = ((($$lcssa)) + 8|0);
 HEAP32[$147>>2] = 4;
 $148 = HEAP32[$4>>2]|0;
 $149 = HEAP32[$0>>2]|0;
 $150 = (($148) - ($149))|0;
 $151 = ($150|0)<(17);
 $152 = $149;
 if (!($151)) {
  $154 = $152;
  $153 = ((($154)) + 16|0);
  HEAP32[$0>>2] = $153;
  $155 = (($n$0$lcssa) + 1)|0;
  $$cast = $153;
  $156 = ((($L)) + 12|0);
  $157 = HEAP32[$156>>2]|0;
  $158 = (($$cast) - ($157))|0;
  $159 = $158 >> 4;
  $160 = (($159) + -1)|0;
  _luaV_concat($L,$155,$160);
  $161 = HEAP32[$0>>2]|0;
  $162 = (0 - ($n$0$lcssa))|0;
  $163 = (($161) + ($162<<4)|0);
  HEAP32[$0>>2] = $163;
  $164 = ((($163)) + -16|0);
  $165 = HEAP32[$164>>2]|0;
  $166 = ((($165)) + 16|0);
  STACKTOP = sp;return ($166|0);
 }
 _luaD_growstack($L,1);
 $$pre$i12 = HEAP32[$0>>2]|0;
 $154 = $$pre$i12;
 $153 = ((($154)) + 16|0);
 HEAP32[$0>>2] = $153;
 $155 = (($n$0$lcssa) + 1)|0;
 $$cast = $153;
 $156 = ((($L)) + 12|0);
 $157 = HEAP32[$156>>2]|0;
 $158 = (($$cast) - ($157))|0;
 $159 = $158 >> 4;
 $160 = (($159) + -1)|0;
 _luaV_concat($L,$155,$160);
 $161 = HEAP32[$0>>2]|0;
 $162 = (0 - ($n$0$lcssa))|0;
 $163 = (($161) + ($162<<4)|0);
 HEAP32[$0>>2] = $163;
 $164 = ((($163)) + -16|0);
 $165 = HEAP32[$164>>2]|0;
 $166 = ((($165)) + 16|0);
 STACKTOP = sp;return ($166|0);
}
function _luaO_pushfstring($L,$fmt,$varargs) {
 $L = $L|0;
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $0 = 0, $argp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $argp = sp;
 HEAP32[$argp>>2] = $varargs;
 $0 = (_luaO_pushvfstring($L,$fmt,$argp)|0);
 STACKTOP = sp;return ($0|0);
}
function _luaO_chunkid($out,$source,$bufflen) {
 $out = $out|0;
 $source = $source|0;
 $bufflen = $bufflen|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $endptr = 0, $endptr2 = 0;
 var $endptr4 = 0, $strlen = 0, $strlen1 = 0, $strlen3 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = HEAP8[$source>>0]|0;
 switch ($0<<24>>24) {
 case 61:  {
  $1 = ((($source)) + 1|0);
  (_strncpy($out,$1,$bufflen)|0);
  $2 = (($bufflen) + -1)|0;
  $3 = (($out) + ($2)|0);
  HEAP8[$3>>0] = 0;
  return;
  break;
 }
 case 64:  {
  $4 = ((($source)) + 1|0);
  $5 = (($bufflen) + -8)|0;
  $6 = (_strlen($4)|0);
  HEAP8[$out>>0] = 0;
  $7 = ($6>>>0)>($5>>>0);
  if ($7) {
   $8 = (($6) - ($5))|0;
   $9 = (($4) + ($8)|0);
   $strlen3 = (_strlen($out)|0);
   $endptr4 = (($out) + ($strlen3)|0);
   HEAP8[$endptr4>>0]=3026478&255;HEAP8[$endptr4+1>>0]=(3026478>>8)&255;HEAP8[$endptr4+2>>0]=(3026478>>16)&255;HEAP8[$endptr4+3>>0]=3026478>>24;
   $$0 = $9;
  } else {
   $$0 = $4;
  }
  (_strcat($out,$$0)|0);
  return;
  break;
 }
 default: {
  $10 = (_strcspn($source,518231)|0);
  $11 = (($bufflen) + -17)|0;
  $12 = ($10>>>0)>($11>>>0);
  $$ = $12 ? $11 : $10;
  dest=$out; src=518234; stop=dest+10|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
  $13 = (($source) + ($$)|0);
  $14 = HEAP8[$13>>0]|0;
  $15 = ($14<<24>>24)==(0);
  if ($15) {
   (_strcat($out,$source)|0);
  } else {
   (_strncat($out,$source,$$)|0);
   $strlen1 = (_strlen($out)|0);
   $endptr2 = (($out) + ($strlen1)|0);
   HEAP8[$endptr2>>0]=3026478&255;HEAP8[$endptr2+1>>0]=(3026478>>8)&255;HEAP8[$endptr2+2>>0]=(3026478>>16)&255;HEAP8[$endptr2+3>>0]=3026478>>24;
  }
  $strlen = (_strlen($out)|0);
  $endptr = (($out) + ($strlen)|0);
  ;HEAP8[$endptr>>0]=HEAP8[518244>>0]|0;HEAP8[$endptr+1>>0]=HEAP8[518244+1>>0]|0;HEAP8[$endptr+2>>0]=HEAP8[518244+2>>0]|0;
  return;
 }
 }
}
function _luaY_parser($L,$z,$buff,$name) {
 $L = $L|0;
 $z = $z|0;
 $buff = $buff|0;
 $name = $name|0;
 var $$pre = 0, $$pre$i = 0, $$pre1$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $7 = 0, $8 = 0, $9 = 0, $funcstate = 0, $lexstate = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 656|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp + 72|0;
 $lexstate = sp;
 $funcstate = sp + 76|0;
 $0 = ((($lexstate)) + 60|0);
 HEAP32[$0>>2] = $buff;
 $1 = (_strlen($name)|0);
 $2 = (_luaS_newlstr($L,$name,$1)|0);
 _luaX_setinput($L,$lexstate,$z,$2);
 $3 = ((($lexstate)) + 52|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (_luaF_newproto($4)|0);
 HEAP32[$funcstate>>2] = $5;
 $6 = ((($lexstate)) + 48|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($funcstate)) + 8|0);
 HEAP32[$8>>2] = $7;
 $9 = ((($funcstate)) + 12|0);
 HEAP32[$9>>2] = $lexstate;
 $10 = ((($funcstate)) + 16|0);
 HEAP32[$10>>2] = $4;
 HEAP32[$6>>2] = $funcstate;
 $11 = ((($funcstate)) + 24|0);
 HEAP32[$11>>2] = 0;
 $12 = ((($funcstate)) + 28|0);
 HEAP32[$12>>2] = -1;
 $13 = ((($funcstate)) + 32|0);
 HEAP32[$13>>2] = -1;
 $14 = ((($funcstate)) + 36|0);
 $15 = ((($funcstate)) + 20|0);
 HEAP32[$15>>2] = 0;
 $16 = ((($lexstate)) + 64|0);
 ;HEAP32[$14>>2]=0|0;HEAP32[$14+4>>2]=0|0;HEAP32[$14+8>>2]=0|0;HEAP16[$14+12>>1]=0|0;HEAP8[$14+14>>0]=0|0;
 $17 = HEAP32[$16>>2]|0;
 $18 = ((($5)) + 32|0);
 HEAP32[$18>>2] = $17;
 $19 = ((($5)) + 75|0);
 HEAP8[$19>>0] = 2;
 $20 = (_luaH_new($4,0,0)|0);
 $21 = ((($funcstate)) + 4|0);
 HEAP32[$21>>2] = $20;
 $22 = ((($4)) + 8|0);
 $23 = HEAP32[$22>>2]|0;
 HEAP32[$23>>2] = $20;
 $24 = ((($23)) + 8|0);
 HEAP32[$24>>2] = 5;
 $25 = ((($4)) + 28|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = HEAP32[$22>>2]|0;
 $28 = (($26) - ($27))|0;
 $29 = ($28|0)<(17);
 $30 = $27;
 if ($29) {
  _luaD_growstack($4,1);
  $$pre$i = HEAP32[$22>>2]|0;
  $32 = $$pre$i;
 } else {
  $32 = $30;
 }
 $31 = ((($32)) + 16|0);
 HEAP32[$22>>2] = $31;
 HEAP32[$31>>2] = $5;
 $33 = ((($32)) + 24|0);
 HEAP32[$33>>2] = 9;
 $34 = HEAP32[$25>>2]|0;
 $35 = HEAP32[$22>>2]|0;
 $36 = (($34) - ($35))|0;
 $37 = ($36|0)<(17);
 $38 = $35;
 if ($37) {
  _luaD_growstack($4,1);
  $$pre1$i = HEAP32[$22>>2]|0;
  $40 = $$pre1$i;
 } else {
  $40 = $38;
 }
 $39 = ((($40)) + 16|0);
 HEAP32[$22>>2] = $39;
 $41 = HEAP32[$funcstate>>2]|0;
 $42 = ((($41)) + 74|0);
 HEAP8[$42>>0] = 2;
 _luaX_next($lexstate);
 $43 = HEAP32[$3>>2]|0;
 $44 = ((($43)) + 52|0);
 $45 = HEAP16[$44>>1]|0;
 $46 = (($45) + 1)<<16>>16;
 HEAP16[$44>>1] = $46;
 $47 = ($46&65535)>(200);
 if ($47) {
  _luaX_lexerror($lexstate,518285,0);
 }
 $48 = ((($lexstate)) + 16|0);
 L10: while(1) {
  $49 = HEAP32[$48>>2]|0;
  switch ($49|0) {
  case 287: case 276: case 262: case 261: case 260:  {
   $64 = $49;
   break L10;
   break;
  }
  default: {
  }
  }
  $50 = (_statement($lexstate)|0);
  $51 = HEAP32[$48>>2]|0;
  $52 = ($51|0)==(59);
  if ($52) {
   _luaX_next($lexstate);
  }
  $53 = HEAP32[$6>>2]|0;
  $54 = ((($53)) + 50|0);
  $55 = HEAP8[$54>>0]|0;
  $56 = $55&255;
  $57 = ((($53)) + 36|0);
  HEAP32[$57>>2] = $56;
  $58 = ($50|0)==(0);
  if (!($58)) {
   label = 12;
   break;
  }
 }
 if ((label|0) == 12) {
  $$pre = HEAP32[$48>>2]|0;
  $64 = $$pre;
 }
 $59 = HEAP32[$3>>2]|0;
 $60 = ((($59)) + 52|0);
 $61 = HEAP16[$60>>1]|0;
 $62 = (($61) + -1)<<16>>16;
 HEAP16[$60>>1] = $62;
 $63 = ($64|0)==(287);
 if ($63) {
  _close_func($lexstate);
  $67 = HEAP32[$funcstate>>2]|0;
  STACKTOP = sp;return ($67|0);
 }
 $65 = (_luaX_token2str($lexstate,287)|0);
 HEAP32[$vararg_buffer>>2] = $65;
 $66 = (_luaO_pushfstring($59,518361,$vararg_buffer)|0);
 _luaX_syntaxerror($lexstate,$66);
 _close_func($lexstate);
 $67 = HEAP32[$funcstate>>2]|0;
 STACKTOP = sp;return ($67|0);
}
function _statement($ls) {
 $ls = $ls|0;
 var $$$i$i = 0, $$$i$i$i = 0, $$0 = 0, $$01$i$i32 = 0, $$idx$i$i = 0, $$idx$i$i14 = 0, $$idx$i$i2 = 0, $$idx$i1$i = 0, $$idx$i10$i = 0, $$idx$val$i$i = 0, $$idx$val$i$i15 = 0, $$idx$val$i$i3 = 0, $$idx$val$i11$i = 0, $$idx$val$i2$i = 0, $$lcssa113 = 0, $$lcssa118 = 0, $$lcssa119 = 0, $$mask$i$i = 0, $$mask$i$i21 = 0, $$mask$i$i9 = 0;
 var $$mask$i17$i = 0, $$mask$i8$i = 0, $$off$i = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i16 = 0, $$pre$i$i$i4 = 0, $$pre$i$i12$i = 0, $$pre$i$i3$i = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0;
 var $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0;
 var $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0;
 var $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0;
 var $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0;
 var $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0;
 var $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0;
 var $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0;
 var $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0;
 var $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0;
 var $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0;
 var $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0;
 var $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0;
 var $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0;
 var $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0;
 var $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0;
 var $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0;
 var $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $b$i = 0, $b$i24 = 0, $bl$0$i = 0, $bl$0$i$i = 0, $bl$05$i = 0, $bl$05$i$i = 0, $bl$06$i = 0;
 var $bl$06$i$i = 0, $bl$08$i = 0, $bl$08$i$i = 0, $bl$i = 0, $bl$i1 = 0, $bl1$i = 0, $bl2$i = 0, $e$i = 0, $e$i$i = 0, $e$i33 = 0, $e$i42 = 0, $e$i43 = 0, $e$i44 = 0, $escapelist$i = 0, $first$0$i = 0, $flist$0$i = 0, $flist$0$i$lcssa = 0, $flist$0$i$lcssa112 = 0, $indvars$iv$i$i = 0, $indvars$iv$i$i18 = 0;
 var $indvars$iv$i$i6 = 0, $indvars$iv$i14$i = 0, $indvars$iv$i5$i = 0, $indvars$iv$next$i$i = 0, $indvars$iv$next$i$i19 = 0, $indvars$iv$next$i$i7 = 0, $indvars$iv$next$i15$i = 0, $indvars$iv$next$i6$i = 0, $ls$idx$val$i = 0, $ls$idx$val$i$i = 0, $ls$idx$val$i30 = 0, $ls$idx1$i = 0, $ls$idx1$val$i = 0, $n$0$i$i35$lcssa = 0, $n$0$i$i3560 = 0, $n$0$i$i56 = 0, $n$0$i$lcssa = 0, $n$0$i52 = 0, $needself$0$i$i = 0, $nexps$0$i = 0;
 var $nret$0$i = 0, $nvars$0$i = 0, $nvars$0$i$i$lcssa = 0, $nvars$0$i$i53 = 0, $nvars$0$i$i53$lcssa = 0, $or$cond$i = 0, $phitmp = 0, $switch$i = 0, $upval$0$lcssa$i = 0, $upval$0$lcssa$i$i = 0, $upval$03$i = 0, $upval$03$i$i = 0, $upval$07$i = 0, $upval$07$i$i = 0, $v$i = 0, $v$i$i = 0, $v$i$i10 = 0, $v$i23 = 0, $v$i39 = 0, $vararg_buffer = 0;
 var $vararg_buffer1 = 0, $vararg_buffer10 = 0, $vararg_buffer13 = 0, $vararg_buffer16 = 0, $vararg_buffer19 = 0, $vararg_buffer22 = 0, $vararg_buffer4 = 0, $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 448|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer22 = sp + 384|0;
 $vararg_buffer19 = sp + 376|0;
 $vararg_buffer16 = sp + 368|0;
 $vararg_buffer13 = sp + 360|0;
 $vararg_buffer10 = sp + 352|0;
 $vararg_buffer7 = sp + 344|0;
 $vararg_buffer4 = sp + 336|0;
 $vararg_buffer1 = sp + 328|0;
 $vararg_buffer = sp + 320|0;
 $e$i44 = sp + 296|0;
 $e$i43 = sp + 272|0;
 $e$i42 = sp + 248|0;
 $v$i39 = sp + 216|0;
 $e$i33 = sp + 192|0;
 $e$i = sp + 168|0;
 $v$i23 = sp + 144|0;
 $b$i24 = sp + 120|0;
 $v$i = sp + 96|0;
 $b$i = sp + 72|0;
 $v$i$i10 = sp + 48|0;
 $bl1$i = sp + 428|0;
 $bl2$i = sp + 416|0;
 $e$i$i = sp + 24|0;
 $bl$i1 = sp + 404|0;
 $v$i$i = sp;
 $bl$i = sp + 392|0;
 $escapelist$i = sp + 388|0;
 $0 = ((($ls)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($ls)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 do {
  switch ($3|0) {
  case 266:  {
   $4 = ((($ls)) + 48|0);
   $5 = HEAP32[$4>>2]|0;
   HEAP32[$escapelist$i>>2] = -1;
   $6 = (_test_then_block($ls)|0);
   $flist$0$i = $6;
   L9: while(1) {
    $7 = HEAP32[$2>>2]|0;
    switch ($7|0) {
    case 260:  {
     $flist$0$i$lcssa112 = $flist$0$i;
     label = 5;
     break L9;
     break;
    }
    case 261:  {
     break;
    }
    default: {
     $flist$0$i$lcssa = $flist$0$i;
     label = 6;
     break L9;
    }
    }
    $8 = (_luaK_jump($5)|0);
    _luaK_concat($5,$escapelist$i,$8);
    _luaK_patchtohere($5,$flist$0$i);
    $9 = (_test_then_block($ls)|0);
    $flist$0$i = $9;
   }
   if ((label|0) == 5) {
    $10 = (_luaK_jump($5)|0);
    _luaK_concat($5,$escapelist$i,$10);
    _luaK_patchtohere($5,$flist$0$i$lcssa112);
    _luaX_next($ls);
    _block($ls);
   }
   else if ((label|0) == 6) {
    _luaK_concat($5,$escapelist$i,$flist$0$i$lcssa);
   }
   $11 = HEAP32[$escapelist$i>>2]|0;
   _luaK_patchtohere($5,$11);
   _check_match($ls,262,266,$1);
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 277:  {
   $12 = ((($ls)) + 48|0);
   $13 = HEAP32[$12>>2]|0;
   _luaX_next($ls);
   $14 = (_luaK_getlabel($13)|0);
   (_subexpr($ls,$v$i$i,0)|0);
   $15 = HEAP32[$v$i$i>>2]|0;
   $16 = ($15|0)==(1);
   if ($16) {
    HEAP32[$v$i$i>>2] = 3;
   }
   $17 = HEAP32[$12>>2]|0;
   _luaK_goiftrue($17,$v$i$i);
   $18 = ((($v$i$i)) + 20|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ((($bl$i)) + 4|0);
   HEAP32[$20>>2] = -1;
   $21 = ((($bl$i)) + 10|0);
   HEAP8[$21>>0] = 1;
   $22 = ((($13)) + 50|0);
   $23 = HEAP8[$22>>0]|0;
   $24 = ((($bl$i)) + 8|0);
   HEAP8[$24>>0] = $23;
   $25 = ((($bl$i)) + 9|0);
   HEAP8[$25>>0] = 0;
   $26 = ((($13)) + 20|0);
   $27 = HEAP32[$26>>2]|0;
   HEAP32[$bl$i>>2] = $27;
   HEAP32[$26>>2] = $bl$i;
   $28 = HEAP32[$2>>2]|0;
   $29 = ($28|0)==(259);
   if (!($29)) {
    $30 = ((($ls)) + 52|0);
    $31 = HEAP32[$30>>2]|0;
    $32 = (_luaX_token2str($ls,259)|0);
    HEAP32[$vararg_buffer>>2] = $32;
    $33 = (_luaO_pushfstring($31,518361,$vararg_buffer)|0);
    _luaX_syntaxerror($ls,$33);
   }
   _luaX_next($ls);
   _block($ls);
   $34 = (_luaK_jump($13)|0);
   _luaK_patchlist($13,$34,$14);
   _check_match($ls,262,277,$1);
   $35 = HEAP32[$26>>2]|0;
   $36 = HEAP32[$35>>2]|0;
   HEAP32[$26>>2] = $36;
   $37 = ((($13)) + 12|0);
   $38 = HEAP32[$37>>2]|0;
   $39 = ((($35)) + 8|0);
   $40 = HEAP16[$39>>1]|0;
   $41 = $40&255;
   $$idx$i$i = ((($38)) + 48|0);
   $$idx$val$i$i = HEAP32[$$idx$i$i>>2]|0;
   $42 = ((($$idx$val$i$i)) + 50|0);
   $43 = HEAP8[$42>>0]|0;
   $44 = ($43&255)>($41&255);
   if ($44) {
    $45 = ((($$idx$val$i$i)) + 24|0);
    $46 = HEAP32[$$idx$val$i$i>>2]|0;
    $47 = ((($46)) + 24|0);
    $48 = HEAP32[$47>>2]|0;
    $$pre$i$i$i = HEAP32[$45>>2]|0;
    $49 = $43&255;
    $indvars$iv$i$i = $49;
    while(1) {
     $50 = $indvars$iv$i$i&255;
     $51 = (($50) + -1)<<24>>24;
     HEAP8[$42>>0] = $51;
     $52 = $51&255;
     $53 = (((($$idx$val$i$i)) + 172|0) + ($52<<1)|0);
     $54 = HEAP16[$53>>1]|0;
     $55 = $54&65535;
     $56 = (((($48) + (($55*12)|0)|0)) + 8|0);
     HEAP32[$56>>2] = $$pre$i$i$i;
     $57 = ($51&255)>($41&255);
     $indvars$iv$next$i$i = (($indvars$iv$i$i) + -1)|0;
     if ($57) {
      $indvars$iv$i$i = $indvars$iv$next$i$i;
     } else {
      break;
     }
    }
   }
   $58 = ($40&65535)<(256);
   if (!($58)) {
    $$mask$i$i = $40 & 255;
    $59 = $$mask$i$i&65535;
    (_luaK_codeABC($13,35,$59,0,0)|0);
   }
   $60 = HEAP8[$22>>0]|0;
   $61 = $60&255;
   $62 = ((($13)) + 36|0);
   HEAP32[$62>>2] = $61;
   $63 = ((($35)) + 4|0);
   $64 = HEAP32[$63>>2]|0;
   _luaK_patchtohere($13,$64);
   _luaK_patchtohere($13,$19);
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 259:  {
   _luaX_next($ls);
   _block($ls);
   _check_match($ls,262,259,$1);
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 264:  {
   $65 = ((($ls)) + 48|0);
   $66 = HEAP32[$65>>2]|0;
   $67 = ((($bl$i1)) + 4|0);
   HEAP32[$67>>2] = -1;
   $68 = ((($bl$i1)) + 10|0);
   HEAP8[$68>>0] = 1;
   $69 = ((($66)) + 50|0);
   $70 = HEAP8[$69>>0]|0;
   $71 = ((($bl$i1)) + 8|0);
   HEAP8[$71>>0] = $70;
   $72 = ((($bl$i1)) + 9|0);
   HEAP8[$72>>0] = 0;
   $73 = ((($66)) + 20|0);
   $74 = HEAP32[$73>>2]|0;
   HEAP32[$bl$i1>>2] = $74;
   HEAP32[$73>>2] = $bl$i1;
   _luaX_next($ls);
   $75 = HEAP32[$2>>2]|0;
   $76 = ($75|0)==(285);
   if (!($76)) {
    $77 = ((($ls)) + 52|0);
    $78 = HEAP32[$77>>2]|0;
    $79 = (_luaX_token2str($ls,285)|0);
    HEAP32[$vararg_buffer1>>2] = $79;
    $80 = (_luaO_pushfstring($78,518361,$vararg_buffer1)|0);
    _luaX_syntaxerror($ls,$80);
   }
   $81 = ((($ls)) + 24|0);
   $82 = HEAP32[$81>>2]|0;
   _luaX_next($ls);
   $83 = HEAP32[$2>>2]|0;
   switch ($83|0) {
   case 61:  {
    $84 = HEAP32[$65>>2]|0;
    $85 = ((($84)) + 36|0);
    $86 = HEAP32[$85>>2]|0;
    $87 = (_luaX_newstring($ls,518722,11)|0);
    _new_localvar($ls,$87,0);
    $88 = (_luaX_newstring($ls,518734,11)|0);
    _new_localvar($ls,$88,1);
    $89 = (_luaX_newstring($ls,518746,10)|0);
    _new_localvar($ls,$89,2);
    _new_localvar($ls,$82,3);
    $90 = HEAP32[$2>>2]|0;
    $91 = ($90|0)==(61);
    if (!($91)) {
     $92 = ((($ls)) + 52|0);
     $93 = HEAP32[$92>>2]|0;
     $94 = (_luaX_token2str($ls,61)|0);
     HEAP32[$vararg_buffer4>>2] = $94;
     $95 = (_luaO_pushfstring($93,518361,$vararg_buffer4)|0);
     _luaX_syntaxerror($ls,$95);
    }
    _luaX_next($ls);
    (_subexpr($ls,$e$i42,0)|0);
    $96 = HEAP32[$65>>2]|0;
    _luaK_exp2nextreg($96,$e$i42);
    $97 = HEAP32[$2>>2]|0;
    $98 = ($97|0)==(44);
    if (!($98)) {
     $99 = ((($ls)) + 52|0);
     $100 = HEAP32[$99>>2]|0;
     $101 = (_luaX_token2str($ls,44)|0);
     HEAP32[$vararg_buffer7>>2] = $101;
     $102 = (_luaO_pushfstring($100,518361,$vararg_buffer7)|0);
     _luaX_syntaxerror($ls,$102);
    }
    _luaX_next($ls);
    (_subexpr($ls,$e$i43,0)|0);
    $103 = HEAP32[$65>>2]|0;
    _luaK_exp2nextreg($103,$e$i43);
    $104 = HEAP32[$2>>2]|0;
    $105 = ($104|0)==(44);
    if ($105) {
     _luaX_next($ls);
     (_subexpr($ls,$e$i44,0)|0);
     $106 = HEAP32[$65>>2]|0;
     _luaK_exp2nextreg($106,$e$i44);
    } else {
     $107 = HEAP32[$85>>2]|0;
     $108 = (_luaK_numberK($84,1.0)|0);
     (_luaK_codeABx($84,1,$107,$108)|0);
     _luaK_reserveregs($84,1);
    }
    _forbody($ls,$86,$1,1,1);
    break;
   }
   case 267: case 44:  {
    $109 = HEAP32[$65>>2]|0;
    $110 = ((($109)) + 36|0);
    $111 = HEAP32[$110>>2]|0;
    $112 = (_luaX_newstring($ls,518757,15)|0);
    _new_localvar($ls,$112,0);
    $113 = (_luaX_newstring($ls,518773,11)|0);
    _new_localvar($ls,$113,1);
    $114 = (_luaX_newstring($ls,518785,13)|0);
    _new_localvar($ls,$114,2);
    _new_localvar($ls,$82,3);
    $115 = HEAP32[$2>>2]|0;
    $116 = ($115|0)==(44);
    if ($116) {
     $117 = ((($ls)) + 52|0);
     $nvars$0$i$i53 = 4;
     while(1) {
      _luaX_next($ls);
      $118 = HEAP32[$2>>2]|0;
      $119 = ($118|0)==(285);
      if (!($119)) {
       $120 = HEAP32[$117>>2]|0;
       $121 = (_luaX_token2str($ls,285)|0);
       HEAP32[$vararg_buffer10>>2] = $121;
       $122 = (_luaO_pushfstring($120,518361,$vararg_buffer10)|0);
       _luaX_syntaxerror($ls,$122);
      }
      $123 = HEAP32[$81>>2]|0;
      _luaX_next($ls);
      $124 = (($nvars$0$i$i53) + 1)|0;
      _new_localvar($ls,$123,$nvars$0$i$i53);
      $125 = HEAP32[$2>>2]|0;
      $126 = ($125|0)==(44);
      if ($126) {
       $nvars$0$i$i53 = $124;
      } else {
       $$lcssa113 = $125;$nvars$0$i$i53$lcssa = $nvars$0$i$i53;
       break;
      }
     }
     $phitmp = (($nvars$0$i$i53$lcssa) + -2)|0;
     $128 = $$lcssa113;$nvars$0$i$i$lcssa = $phitmp;
    } else {
     $128 = $115;$nvars$0$i$i$lcssa = 1;
    }
    $127 = ($128|0)==(267);
    if (!($127)) {
     $129 = ((($ls)) + 52|0);
     $130 = HEAP32[$129>>2]|0;
     $131 = (_luaX_token2str($ls,267)|0);
     HEAP32[$vararg_buffer13>>2] = $131;
     $132 = (_luaO_pushfstring($130,518361,$vararg_buffer13)|0);
     _luaX_syntaxerror($ls,$132);
    }
    _luaX_next($ls);
    $133 = HEAP32[$0>>2]|0;
    (_subexpr($ls,$e$i$i,0)|0);
    $134 = HEAP32[$2>>2]|0;
    $135 = ($134|0)==(44);
    if ($135) {
     $n$0$i52 = 1;
     while(1) {
      _luaX_next($ls);
      $136 = HEAP32[$65>>2]|0;
      _luaK_exp2nextreg($136,$e$i$i);
      (_subexpr($ls,$e$i$i,0)|0);
      $137 = (($n$0$i52) + 1)|0;
      $138 = HEAP32[$2>>2]|0;
      $139 = ($138|0)==(44);
      if ($139) {
       $n$0$i52 = $137;
      } else {
       $n$0$i$lcssa = $137;
       break;
      }
     }
    } else {
     $n$0$i$lcssa = 1;
    }
    $ls$idx$val$i$i = HEAP32[$65>>2]|0;
    $140 = (3 - ($n$0$i$lcssa))|0;
    $141 = HEAP32[$e$i$i>>2]|0;
    switch ($141|0) {
    case 14: case 13:  {
     $142 = (($140) + 1)|0;
     $143 = ($142|0)<(0);
     $$$i$i$i = $143 ? 0 : $142;
     _luaK_setreturns($ls$idx$val$i$i,$e$i$i,$$$i$i$i);
     $144 = ($$$i$i$i|0)>(1);
     if ($144) {
      $145 = (($$$i$i$i) + -1)|0;
      _luaK_reserveregs($ls$idx$val$i$i,$145);
     }
     break;
    }
    case 0:  {
     label = 44;
     break;
    }
    default: {
     _luaK_exp2nextreg($ls$idx$val$i$i,$e$i$i);
     label = 44;
    }
    }
    if ((label|0) == 44) {
     $146 = ($140|0)>(0);
     if ($146) {
      $147 = ((($ls$idx$val$i$i)) + 36|0);
      $148 = HEAP32[$147>>2]|0;
      _luaK_reserveregs($ls$idx$val$i$i,$140);
      _luaK_nil($ls$idx$val$i$i,$148,$140);
     }
    }
    _luaK_checkstack($109,3);
    _forbody($ls,$111,$133,$nvars$0$i$i$lcssa,0);
    break;
   }
   default: {
    _luaX_syntaxerror($ls,518799);
   }
   }
   _check_match($ls,262,264,$1);
   $149 = HEAP32[$73>>2]|0;
   $150 = HEAP32[$149>>2]|0;
   HEAP32[$73>>2] = $150;
   $151 = ((($66)) + 12|0);
   $152 = HEAP32[$151>>2]|0;
   $153 = ((($149)) + 8|0);
   $154 = HEAP16[$153>>1]|0;
   $155 = $154&255;
   $$idx$i$i2 = ((($152)) + 48|0);
   $$idx$val$i$i3 = HEAP32[$$idx$i$i2>>2]|0;
   $156 = ((($$idx$val$i$i3)) + 50|0);
   $157 = HEAP8[$156>>0]|0;
   $158 = ($157&255)>($155&255);
   if ($158) {
    $159 = ((($$idx$val$i$i3)) + 24|0);
    $160 = HEAP32[$$idx$val$i$i3>>2]|0;
    $161 = ((($160)) + 24|0);
    $162 = HEAP32[$161>>2]|0;
    $$pre$i$i$i4 = HEAP32[$159>>2]|0;
    $163 = $157&255;
    $indvars$iv$i$i6 = $163;
    while(1) {
     $164 = $indvars$iv$i$i6&255;
     $165 = (($164) + -1)<<24>>24;
     HEAP8[$156>>0] = $165;
     $166 = $165&255;
     $167 = (((($$idx$val$i$i3)) + 172|0) + ($166<<1)|0);
     $168 = HEAP16[$167>>1]|0;
     $169 = $168&65535;
     $170 = (((($162) + (($169*12)|0)|0)) + 8|0);
     HEAP32[$170>>2] = $$pre$i$i$i4;
     $171 = ($165&255)>($155&255);
     $indvars$iv$next$i$i7 = (($indvars$iv$i$i6) + -1)|0;
     if ($171) {
      $indvars$iv$i$i6 = $indvars$iv$next$i$i7;
     } else {
      break;
     }
    }
   }
   $172 = ($154&65535)<(256);
   if (!($172)) {
    $$mask$i$i9 = $154 & 255;
    $173 = $$mask$i$i9&65535;
    (_luaK_codeABC($66,35,$173,0,0)|0);
   }
   $174 = HEAP8[$69>>0]|0;
   $175 = $174&255;
   $176 = ((($66)) + 36|0);
   HEAP32[$176>>2] = $175;
   $177 = ((($149)) + 4|0);
   $178 = HEAP32[$177>>2]|0;
   _luaK_patchtohere($66,$178);
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 272:  {
   $179 = ((($ls)) + 48|0);
   $180 = HEAP32[$179>>2]|0;
   $181 = (_luaK_getlabel($180)|0);
   $182 = ((($bl1$i)) + 4|0);
   HEAP32[$182>>2] = -1;
   $183 = ((($bl1$i)) + 10|0);
   HEAP8[$183>>0] = 1;
   $184 = ((($180)) + 50|0);
   $185 = HEAP8[$184>>0]|0;
   $186 = ((($bl1$i)) + 8|0);
   HEAP8[$186>>0] = $185;
   $187 = ((($bl1$i)) + 9|0);
   HEAP8[$187>>0] = 0;
   $188 = ((($180)) + 20|0);
   $189 = HEAP32[$188>>2]|0;
   HEAP32[$bl1$i>>2] = $189;
   $190 = ((($bl2$i)) + 4|0);
   HEAP32[$190>>2] = -1;
   $191 = ((($bl2$i)) + 10|0);
   HEAP8[$191>>0] = 0;
   $192 = ((($bl2$i)) + 8|0);
   HEAP8[$192>>0] = $185;
   $193 = ((($bl2$i)) + 9|0);
   HEAP8[$193>>0] = 0;
   $194 = $bl1$i;
   HEAP32[$bl2$i>>2] = $194;
   HEAP32[$188>>2] = $bl2$i;
   _luaX_next($ls);
   $195 = ((($ls)) + 52|0);
   $196 = HEAP32[$195>>2]|0;
   $197 = ((($196)) + 52|0);
   $198 = HEAP16[$197>>1]|0;
   $199 = (($198) + 1)<<16>>16;
   HEAP16[$197>>1] = $199;
   $200 = ($199&65535)>(200);
   if ($200) {
    _luaX_lexerror($ls,518285,0);
   }
   L90: while(1) {
    $201 = HEAP32[$2>>2]|0;
    switch ($201|0) {
    case 287: case 276: case 262: case 261: case 260:  {
     break L90;
     break;
    }
    default: {
    }
    }
    $202 = (_statement($ls)|0);
    $203 = HEAP32[$2>>2]|0;
    $204 = ($203|0)==(59);
    if ($204) {
     _luaX_next($ls);
    }
    $205 = HEAP32[$179>>2]|0;
    $206 = ((($205)) + 50|0);
    $207 = HEAP8[$206>>0]|0;
    $208 = $207&255;
    $209 = ((($205)) + 36|0);
    HEAP32[$209>>2] = $208;
    $210 = ($202|0)==(0);
    if (!($210)) {
     break;
    }
   }
   $211 = HEAP32[$195>>2]|0;
   $212 = ((($211)) + 52|0);
   $213 = HEAP16[$212>>1]|0;
   $214 = (($213) + -1)<<16>>16;
   HEAP16[$212>>1] = $214;
   _check_match($ls,276,272,$1);
   (_subexpr($ls,$v$i$i10,0)|0);
   $215 = HEAP32[$v$i$i10>>2]|0;
   $216 = ($215|0)==(1);
   if ($216) {
    HEAP32[$v$i$i10>>2] = 3;
   }
   $217 = HEAP32[$179>>2]|0;
   _luaK_goiftrue($217,$v$i$i10);
   $218 = ((($v$i$i10)) + 20|0);
   $219 = HEAP32[$218>>2]|0;
   $220 = HEAP8[$193>>0]|0;
   $221 = ($220<<24>>24)==(0);
   if ($221) {
    $222 = HEAP32[$188>>2]|0;
    $223 = HEAP32[$222>>2]|0;
    HEAP32[$188>>2] = $223;
    $224 = ((($180)) + 12|0);
    $225 = HEAP32[$224>>2]|0;
    $226 = ((($222)) + 8|0);
    $227 = HEAP16[$226>>1]|0;
    $228 = $227&255;
    $$idx$i$i14 = ((($225)) + 48|0);
    $$idx$val$i$i15 = HEAP32[$$idx$i$i14>>2]|0;
    $229 = ((($$idx$val$i$i15)) + 50|0);
    $230 = HEAP8[$229>>0]|0;
    $231 = ($230&255)>($228&255);
    if ($231) {
     $232 = ((($$idx$val$i$i15)) + 24|0);
     $233 = HEAP32[$$idx$val$i$i15>>2]|0;
     $234 = ((($233)) + 24|0);
     $235 = HEAP32[$234>>2]|0;
     $$pre$i$i$i16 = HEAP32[$232>>2]|0;
     $236 = $230&255;
     $indvars$iv$i$i18 = $236;
     while(1) {
      $237 = $indvars$iv$i$i18&255;
      $238 = (($237) + -1)<<24>>24;
      HEAP8[$229>>0] = $238;
      $239 = $238&255;
      $240 = (((($$idx$val$i$i15)) + 172|0) + ($239<<1)|0);
      $241 = HEAP16[$240>>1]|0;
      $242 = $241&65535;
      $243 = (((($235) + (($242*12)|0)|0)) + 8|0);
      HEAP32[$243>>2] = $$pre$i$i$i16;
      $244 = ($238&255)>($228&255);
      $indvars$iv$next$i$i19 = (($indvars$iv$i$i18) + -1)|0;
      if ($244) {
       $indvars$iv$i$i18 = $indvars$iv$next$i$i19;
      } else {
       break;
      }
     }
    }
    $245 = ($227&65535)<(256);
    if (!($245)) {
     $$mask$i$i21 = $227 & 255;
     $246 = $$mask$i$i21&65535;
     (_luaK_codeABC($180,35,$246,0,0)|0);
    }
    $247 = HEAP8[$184>>0]|0;
    $248 = $247&255;
    $249 = ((($180)) + 36|0);
    HEAP32[$249>>2] = $248;
    $250 = ((($222)) + 4|0);
    $251 = HEAP32[$250>>2]|0;
    _luaK_patchtohere($180,$251);
    $252 = HEAP32[$179>>2]|0;
    _luaK_patchlist($252,$219,$181);
    $$pre$phiZ2D = $224;
   } else {
    $253 = HEAP32[$179>>2]|0;
    $254 = ((($253)) + 20|0);
    $bl$06$i$i = HEAP32[$254>>2]|0;
    $255 = ($bl$06$i$i|0)==(0|0);
    L111: do {
     if ($255) {
      $upval$0$lcssa$i$i = 0;
      label = 72;
     } else {
      $bl$08$i$i = $bl$06$i$i;$upval$07$i$i = 0;
      while(1) {
       $256 = ((($bl$08$i$i)) + 10|0);
       $257 = HEAP8[$256>>0]|0;
       $258 = ($257<<24>>24)==(0);
       if (!($258)) {
        $bl$05$i$i = $bl$08$i$i;$upval$03$i$i = $upval$07$i$i;
        break L111;
       }
       $259 = ((($bl$08$i$i)) + 9|0);
       $260 = HEAP8[$259>>0]|0;
       $261 = $260&255;
       $262 = $261 | $upval$07$i$i;
       $bl$0$i$i = HEAP32[$bl$08$i$i>>2]|0;
       $263 = ($bl$0$i$i|0)==(0|0);
       if ($263) {
        $upval$0$lcssa$i$i = $262;
        label = 72;
        break;
       } else {
        $bl$08$i$i = $bl$0$i$i;$upval$07$i$i = $262;
       }
      }
     }
    } while(0);
    if ((label|0) == 72) {
     _luaX_syntaxerror($ls,518820);
     $bl$05$i$i = 0;$upval$03$i$i = $upval$0$lcssa$i$i;
    }
    $264 = ($upval$03$i$i|0)==(0);
    if (!($264)) {
     $265 = ((($bl$05$i$i)) + 8|0);
     $266 = HEAP8[$265>>0]|0;
     $267 = $266&255;
     (_luaK_codeABC($253,35,$267,0,0)|0);
    }
    $268 = ((($bl$05$i$i)) + 4|0);
    $269 = (_luaK_jump($253)|0);
    _luaK_concat($253,$268,$269);
    $270 = HEAP32[$179>>2]|0;
    _luaK_patchtohere($270,$219);
    $271 = HEAP32[$188>>2]|0;
    $272 = HEAP32[$271>>2]|0;
    HEAP32[$188>>2] = $272;
    $273 = ((($180)) + 12|0);
    $274 = HEAP32[$273>>2]|0;
    $275 = ((($271)) + 8|0);
    $276 = HEAP16[$275>>1]|0;
    $277 = $276&255;
    $$idx$i1$i = ((($274)) + 48|0);
    $$idx$val$i2$i = HEAP32[$$idx$i1$i>>2]|0;
    $278 = ((($$idx$val$i2$i)) + 50|0);
    $279 = HEAP8[$278>>0]|0;
    $280 = ($279&255)>($277&255);
    if ($280) {
     $281 = ((($$idx$val$i2$i)) + 24|0);
     $282 = HEAP32[$$idx$val$i2$i>>2]|0;
     $283 = ((($282)) + 24|0);
     $284 = HEAP32[$283>>2]|0;
     $$pre$i$i3$i = HEAP32[$281>>2]|0;
     $285 = $279&255;
     $indvars$iv$i5$i = $285;
     while(1) {
      $286 = $indvars$iv$i5$i&255;
      $287 = (($286) + -1)<<24>>24;
      HEAP8[$278>>0] = $287;
      $288 = $287&255;
      $289 = (((($$idx$val$i2$i)) + 172|0) + ($288<<1)|0);
      $290 = HEAP16[$289>>1]|0;
      $291 = $290&65535;
      $292 = (((($284) + (($291*12)|0)|0)) + 8|0);
      HEAP32[$292>>2] = $$pre$i$i3$i;
      $293 = ($287&255)>($277&255);
      $indvars$iv$next$i6$i = (($indvars$iv$i5$i) + -1)|0;
      if ($293) {
       $indvars$iv$i5$i = $indvars$iv$next$i6$i;
      } else {
       break;
      }
     }
    }
    $294 = ($276&65535)<(256);
    if (!($294)) {
     $$mask$i8$i = $276 & 255;
     $295 = $$mask$i8$i&65535;
     (_luaK_codeABC($180,35,$295,0,0)|0);
    }
    $296 = HEAP8[$184>>0]|0;
    $297 = $296&255;
    $298 = ((($180)) + 36|0);
    HEAP32[$298>>2] = $297;
    $299 = ((($271)) + 4|0);
    $300 = HEAP32[$299>>2]|0;
    _luaK_patchtohere($180,$300);
    $301 = HEAP32[$179>>2]|0;
    $302 = (_luaK_jump($180)|0);
    _luaK_patchlist($301,$302,$181);
    $$pre$phiZ2D = $273;
   }
   $303 = HEAP32[$188>>2]|0;
   $304 = HEAP32[$303>>2]|0;
   HEAP32[$188>>2] = $304;
   $305 = HEAP32[$$pre$phiZ2D>>2]|0;
   $306 = ((($303)) + 8|0);
   $307 = HEAP16[$306>>1]|0;
   $308 = $307&255;
   $$idx$i10$i = ((($305)) + 48|0);
   $$idx$val$i11$i = HEAP32[$$idx$i10$i>>2]|0;
   $309 = ((($$idx$val$i11$i)) + 50|0);
   $310 = HEAP8[$309>>0]|0;
   $311 = ($310&255)>($308&255);
   if ($311) {
    $312 = ((($$idx$val$i11$i)) + 24|0);
    $313 = HEAP32[$$idx$val$i11$i>>2]|0;
    $314 = ((($313)) + 24|0);
    $315 = HEAP32[$314>>2]|0;
    $$pre$i$i12$i = HEAP32[$312>>2]|0;
    $316 = $310&255;
    $indvars$iv$i14$i = $316;
    while(1) {
     $317 = $indvars$iv$i14$i&255;
     $318 = (($317) + -1)<<24>>24;
     HEAP8[$309>>0] = $318;
     $319 = $318&255;
     $320 = (((($$idx$val$i11$i)) + 172|0) + ($319<<1)|0);
     $321 = HEAP16[$320>>1]|0;
     $322 = $321&65535;
     $323 = (((($315) + (($322*12)|0)|0)) + 8|0);
     HEAP32[$323>>2] = $$pre$i$i12$i;
     $324 = ($318&255)>($308&255);
     $indvars$iv$next$i15$i = (($indvars$iv$i14$i) + -1)|0;
     if ($324) {
      $indvars$iv$i14$i = $indvars$iv$next$i15$i;
     } else {
      break;
     }
    }
   }
   $325 = ($307&65535)<(256);
   if (!($325)) {
    $$mask$i17$i = $307 & 255;
    $326 = $$mask$i17$i&65535;
    (_luaK_codeABC($180,35,$326,0,0)|0);
   }
   $327 = HEAP8[$184>>0]|0;
   $328 = $327&255;
   $329 = ((($180)) + 36|0);
   HEAP32[$329>>2] = $328;
   $330 = ((($303)) + 4|0);
   $331 = HEAP32[$330>>2]|0;
   _luaK_patchtohere($180,$331);
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 265:  {
   _luaX_next($ls);
   $332 = HEAP32[$2>>2]|0;
   $333 = ($332|0)==(285);
   if (!($333)) {
    $334 = ((($ls)) + 52|0);
    $335 = HEAP32[$334>>2]|0;
    $336 = (_luaX_token2str($ls,285)|0);
    HEAP32[$vararg_buffer16>>2] = $336;
    $337 = (_luaO_pushfstring($335,518361,$vararg_buffer16)|0);
    _luaX_syntaxerror($ls,$337);
   }
   $338 = ((($ls)) + 24|0);
   $339 = HEAP32[$338>>2]|0;
   _luaX_next($ls);
   $340 = ((($ls)) + 48|0);
   $341 = HEAP32[$340>>2]|0;
   $342 = (_singlevaraux($341,$339,$v$i,1)|0);
   $343 = ($342|0)==(8);
   if ($343) {
    $344 = (_luaK_stringK($341,$339)|0);
    $345 = ((($v$i)) + 8|0);
    HEAP32[$345>>2] = $344;
   }
   L145: while(1) {
    $346 = HEAP32[$2>>2]|0;
    switch ($346|0) {
    case 58:  {
     label = 93;
     break L145;
     break;
    }
    case 46:  {
     break;
    }
    default: {
     $needself$0$i$i = 0;
     break L145;
    }
    }
    _field($ls,$v$i);
   }
   if ((label|0) == 93) {
    _field($ls,$v$i);
    $needself$0$i$i = 1;
   }
   _body($ls,$b$i,$needself$0$i$i,$1);
   $347 = HEAP32[$340>>2]|0;
   _luaK_storevar($347,$v$i,$b$i);
   $348 = HEAP32[$340>>2]|0;
   _luaK_fixline($348,$1);
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 268:  {
   _luaX_next($ls);
   $349 = HEAP32[$2>>2]|0;
   $350 = ($349|0)==(265);
   if ($350) {
    _luaX_next($ls);
    $351 = ((($ls)) + 48|0);
    $352 = HEAP32[$351>>2]|0;
    $353 = HEAP32[$2>>2]|0;
    $354 = ($353|0)==(285);
    if (!($354)) {
     $355 = ((($ls)) + 52|0);
     $356 = HEAP32[$355>>2]|0;
     $357 = (_luaX_token2str($ls,285)|0);
     HEAP32[$vararg_buffer19>>2] = $357;
     $358 = (_luaO_pushfstring($356,518361,$vararg_buffer19)|0);
     _luaX_syntaxerror($ls,$358);
    }
    $359 = ((($ls)) + 24|0);
    $360 = HEAP32[$359>>2]|0;
    _luaX_next($ls);
    _new_localvar($ls,$360,0);
    $361 = ((($352)) + 36|0);
    $362 = HEAP32[$361>>2]|0;
    $363 = ((($v$i23)) + 16|0);
    HEAP32[$363>>2] = -1;
    $364 = ((($v$i23)) + 20|0);
    HEAP32[$364>>2] = -1;
    HEAP32[$v$i23>>2] = 6;
    $365 = ((($v$i23)) + 8|0);
    HEAP32[$365>>2] = $362;
    _luaK_reserveregs($352,1);
    $ls$idx$val$i = HEAP32[$351>>2]|0;
    $366 = ((($ls$idx$val$i)) + 50|0);
    $367 = HEAP8[$366>>0]|0;
    $368 = $367&255;
    $369 = (($368) + 1)|0;
    $370 = $369&255;
    HEAP8[$366>>0] = $370;
    $371 = ((($ls$idx$val$i)) + 24|0);
    $372 = HEAP32[$371>>2]|0;
    $373 = $369 & 255;
    $374 = HEAP32[$ls$idx$val$i>>2]|0;
    $375 = ((($374)) + 24|0);
    $376 = HEAP32[$375>>2]|0;
    $377 = (($373) + -1)|0;
    $378 = (((($ls$idx$val$i)) + 172|0) + ($377<<1)|0);
    $379 = HEAP16[$378>>1]|0;
    $380 = $379&65535;
    $381 = (((($376) + (($380*12)|0)|0)) + 4|0);
    HEAP32[$381>>2] = $372;
    $382 = HEAP32[$0>>2]|0;
    _body($ls,$b$i24,0,$382);
    _luaK_storevar($352,$v$i23,$b$i24);
    $383 = ((($352)) + 24|0);
    $384 = HEAP32[$383>>2]|0;
    $385 = ((($352)) + 50|0);
    $386 = HEAP8[$385>>0]|0;
    $387 = $386&255;
    $388 = (($387) + -1)|0;
    $389 = (((($352)) + 172|0) + ($388<<1)|0);
    $390 = HEAP16[$389>>1]|0;
    $391 = $390&65535;
    $392 = HEAP32[$352>>2]|0;
    $393 = ((($392)) + 24|0);
    $394 = HEAP32[$393>>2]|0;
    $395 = (((($394) + (($391*12)|0)|0)) + 4|0);
    HEAP32[$395>>2] = $384;
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   $396 = ((($ls)) + 24|0);
   $397 = ((($ls)) + 52|0);
   $399 = $349;$nvars$0$i = 0;
   L160: while(1) {
    $398 = ($399|0)==(285);
    if (!($398)) {
     $400 = HEAP32[$397>>2]|0;
     $401 = (_luaX_token2str($ls,285)|0);
     HEAP32[$vararg_buffer22>>2] = $401;
     $402 = (_luaO_pushfstring($400,518361,$vararg_buffer22)|0);
     _luaX_syntaxerror($ls,$402);
    }
    $403 = HEAP32[$396>>2]|0;
    _luaX_next($ls);
    $404 = (($nvars$0$i) + 1)|0;
    _new_localvar($ls,$403,$nvars$0$i);
    $405 = HEAP32[$2>>2]|0;
    switch ($405|0) {
    case 61:  {
     $$lcssa119 = $404;
     label = 104;
     break L160;
     break;
    }
    case 44:  {
     break;
    }
    default: {
     $$lcssa118 = $404;
     label = 107;
     break L160;
    }
    }
    _luaX_next($ls);
    $$pre = HEAP32[$2>>2]|0;
    $399 = $$pre;$nvars$0$i = $404;
   }
   if ((label|0) == 104) {
    _luaX_next($ls);
    (_subexpr($ls,$e$i,0)|0);
    $406 = HEAP32[$2>>2]|0;
    $407 = ($406|0)==(44);
    if ($407) {
     $408 = ((($ls)) + 48|0);
     $n$0$i$i56 = 1;
     while(1) {
      _luaX_next($ls);
      $409 = HEAP32[$408>>2]|0;
      _luaK_exp2nextreg($409,$e$i);
      (_subexpr($ls,$e$i,0)|0);
      $410 = (($n$0$i$i56) + 1)|0;
      $411 = HEAP32[$2>>2]|0;
      $412 = ($411|0)==(44);
      if ($412) {
       $n$0$i$i56 = $410;
      } else {
       $414 = $$lcssa119;$nexps$0$i = $410;
       break;
      }
     }
    } else {
     $414 = $$lcssa119;$nexps$0$i = 1;
    }
   }
   else if ((label|0) == 107) {
    HEAP32[$e$i>>2] = 0;
    $414 = $$lcssa118;$nexps$0$i = 0;
   }
   $ls$idx1$i = ((($ls)) + 48|0);
   $ls$idx1$val$i = HEAP32[$ls$idx1$i>>2]|0;
   $413 = (($414) - ($nexps$0$i))|0;
   $415 = HEAP32[$e$i>>2]|0;
   switch ($415|0) {
   case 14: case 13:  {
    $416 = (($413) + 1)|0;
    $417 = ($416|0)<(0);
    $$$i$i = $417 ? 0 : $416;
    _luaK_setreturns($ls$idx1$val$i,$e$i,$$$i$i);
    $418 = ($$$i$i|0)>(1);
    if ($418) {
     $419 = (($$$i$i) + -1)|0;
     _luaK_reserveregs($ls$idx1$val$i,$419);
    }
    break;
   }
   case 0:  {
    label = 112;
    break;
   }
   default: {
    _luaK_exp2nextreg($ls$idx1$val$i,$e$i);
    label = 112;
   }
   }
   if ((label|0) == 112) {
    $420 = ($413|0)>(0);
    if ($420) {
     $421 = ((($ls$idx1$val$i)) + 36|0);
     $422 = HEAP32[$421>>2]|0;
     _luaK_reserveregs($ls$idx1$val$i,$413);
     _luaK_nil($ls$idx1$val$i,$422,$413);
    }
   }
   $ls$idx$val$i30 = HEAP32[$ls$idx1$i>>2]|0;
   $423 = ((($ls$idx$val$i30)) + 50|0);
   $424 = HEAP8[$423>>0]|0;
   $425 = $424&255;
   $426 = (($425) + ($414))|0;
   $427 = $426&255;
   HEAP8[$423>>0] = $427;
   $428 = ((($ls$idx$val$i30)) + 24|0);
   $429 = HEAP32[$428>>2]|0;
   $430 = $426 & 255;
   $431 = HEAP32[$ls$idx$val$i30>>2]|0;
   $432 = ((($431)) + 24|0);
   $433 = HEAP32[$432>>2]|0;
   $$01$i$i32 = $414;
   while(1) {
    $434 = (($430) - ($$01$i$i32))|0;
    $435 = (((($ls$idx$val$i30)) + 172|0) + ($434<<1)|0);
    $436 = HEAP16[$435>>1]|0;
    $437 = $436&65535;
    $438 = (((($433) + (($437*12)|0)|0)) + 4|0);
    HEAP32[$438>>2] = $429;
    $439 = (($$01$i$i32) + -1)|0;
    $440 = ($439|0)==(0);
    if ($440) {
     break;
    } else {
     $$01$i$i32 = $439;
    }
   }
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 273:  {
   $441 = ((($ls)) + 48|0);
   $442 = HEAP32[$441>>2]|0;
   _luaX_next($ls);
   $443 = HEAP32[$2>>2]|0;
   L186: do {
    switch ($443|0) {
    case 59: case 287: case 276: case 262: case 261: case 260:  {
     $first$0$i = 0;$nret$0$i = 0;
     break;
    }
    default: {
     (_subexpr($ls,$e$i33,0)|0);
     $444 = HEAP32[$2>>2]|0;
     $445 = ($444|0)==(44);
     if ($445) {
      $n$0$i$i3560 = 1;
      while(1) {
       _luaX_next($ls);
       $446 = HEAP32[$441>>2]|0;
       _luaK_exp2nextreg($446,$e$i33);
       (_subexpr($ls,$e$i33,0)|0);
       $447 = (($n$0$i$i3560) + 1)|0;
       $448 = HEAP32[$2>>2]|0;
       $449 = ($448|0)==(44);
       if ($449) {
        $n$0$i$i3560 = $447;
       } else {
        $n$0$i$i35$lcssa = $447;
        break;
       }
      }
     } else {
      $n$0$i$i35$lcssa = 1;
     }
     $450 = HEAP32[$e$i33>>2]|0;
     $$off$i = (($450) + -13)|0;
     $switch$i = ($$off$i>>>0)<(2);
     if ($switch$i) {
      _luaK_setreturns($442,$e$i33,-1);
      $451 = HEAP32[$e$i33>>2]|0;
      $452 = ($451|0)==(13);
      $453 = ($n$0$i$i35$lcssa|0)==(1);
      $or$cond$i = $453 & $452;
      if ($or$cond$i) {
       $454 = ((($e$i33)) + 8|0);
       $455 = HEAP32[$454>>2]|0;
       $456 = HEAP32[$442>>2]|0;
       $457 = ((($456)) + 12|0);
       $458 = HEAP32[$457>>2]|0;
       $459 = (($458) + ($455<<2)|0);
       $460 = HEAP32[$459>>2]|0;
       $461 = $460 & -64;
       $462 = $461 | 29;
       HEAP32[$459>>2] = $462;
      }
      $463 = ((($442)) + 50|0);
      $464 = HEAP8[$463>>0]|0;
      $465 = $464&255;
      $first$0$i = $465;$nret$0$i = -1;
      break L186;
     } else {
      $466 = ($n$0$i$i35$lcssa|0)==(1);
      if ($466) {
       $467 = (_luaK_exp2anyreg($442,$e$i33)|0);
       $first$0$i = $467;$nret$0$i = 1;
       break L186;
      } else {
       _luaK_exp2nextreg($442,$e$i33);
       $468 = ((($442)) + 50|0);
       $469 = HEAP8[$468>>0]|0;
       $470 = $469&255;
       $first$0$i = $470;$nret$0$i = $n$0$i$i35$lcssa;
       break L186;
      }
     }
    }
    }
   } while(0);
   _luaK_ret($442,$first$0$i,$nret$0$i);
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 258:  {
   _luaX_next($ls);
   $471 = ((($ls)) + 48|0);
   $472 = HEAP32[$471>>2]|0;
   $473 = ((($472)) + 20|0);
   $bl$06$i = HEAP32[$473>>2]|0;
   $474 = ($bl$06$i|0)==(0|0);
   L204: do {
    if ($474) {
     $upval$0$lcssa$i = 0;
     label = 131;
    } else {
     $bl$08$i = $bl$06$i;$upval$07$i = 0;
     while(1) {
      $475 = ((($bl$08$i)) + 10|0);
      $476 = HEAP8[$475>>0]|0;
      $477 = ($476<<24>>24)==(0);
      if (!($477)) {
       $bl$05$i = $bl$08$i;$upval$03$i = $upval$07$i;
       break L204;
      }
      $478 = ((($bl$08$i)) + 9|0);
      $479 = HEAP8[$478>>0]|0;
      $480 = $479&255;
      $481 = $480 | $upval$07$i;
      $bl$0$i = HEAP32[$bl$08$i>>2]|0;
      $482 = ($bl$0$i|0)==(0|0);
      if ($482) {
       $upval$0$lcssa$i = $481;
       label = 131;
       break;
      } else {
       $bl$08$i = $bl$0$i;$upval$07$i = $481;
      }
     }
    }
   } while(0);
   if ((label|0) == 131) {
    _luaX_syntaxerror($ls,518820);
    $bl$05$i = 0;$upval$03$i = $upval$0$lcssa$i;
   }
   $483 = ($upval$03$i|0)==(0);
   if (!($483)) {
    $484 = ((($bl$05$i)) + 8|0);
    $485 = HEAP8[$484>>0]|0;
    $486 = $485&255;
    (_luaK_codeABC($472,35,$486,0,0)|0);
   }
   $487 = ((($bl$05$i)) + 4|0);
   $488 = (_luaK_jump($472)|0);
   _luaK_concat($472,$487,$488);
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  default: {
   $489 = ((($ls)) + 48|0);
   $490 = HEAP32[$489>>2]|0;
   $491 = ((($v$i39)) + 8|0);
   _primaryexp($ls,$491);
   $492 = HEAP32[$491>>2]|0;
   $493 = ($492|0)==(13);
   if ($493) {
    $494 = ((($v$i39)) + 16|0);
    $495 = HEAP32[$494>>2]|0;
    $496 = HEAP32[$490>>2]|0;
    $497 = ((($496)) + 12|0);
    $498 = HEAP32[$497>>2]|0;
    $499 = (($498) + ($495<<2)|0);
    $500 = HEAP32[$499>>2]|0;
    $501 = $500 & -8372225;
    $502 = $501 | 16384;
    HEAP32[$499>>2] = $502;
   } else {
    HEAP32[$v$i39>>2] = 0;
    _assignment($ls,$v$i39,1);
   }
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
  }
 } while(0);
 return (0)|0;
}
function _test_then_block($ls) {
 $ls = $ls|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $v$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp + 24|0;
 $v$i = sp;
 _luaX_next($ls);
 (_subexpr($ls,$v$i,0)|0);
 $0 = HEAP32[$v$i>>2]|0;
 $1 = ($0|0)==(1);
 if ($1) {
  HEAP32[$v$i>>2] = 3;
 }
 $2 = ((($ls)) + 48|0);
 $3 = HEAP32[$2>>2]|0;
 _luaK_goiftrue($3,$v$i);
 $4 = ((($v$i)) + 20|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($ls)) + 16|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(274);
 if ($8) {
  _luaX_next($ls);
  _block($ls);
  STACKTOP = sp;return ($5|0);
 }
 $9 = ((($ls)) + 52|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (_luaX_token2str($ls,274)|0);
 HEAP32[$vararg_buffer>>2] = $11;
 $12 = (_luaO_pushfstring($10,518361,$vararg_buffer)|0);
 _luaX_syntaxerror($ls,$12);
 _luaX_next($ls);
 _block($ls);
 STACKTOP = sp;return ($5|0);
}
function _subexpr($ls,$v,$limit) {
 $ls = $ls|0;
 $v = $v|0;
 $limit = $limit|0;
 var $$0$i1$ph = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $7 = 0, $8 = 0, $9 = 0, $ls$idx$i = 0, $ls$idx$val$i = 0, $op$0$lcssa = 0, $op$0$ph$ph = 0, $op$03 = 0, $v2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v2 = sp;
 $0 = ((($ls)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 52|0);
 $3 = HEAP16[$2>>1]|0;
 $4 = (($3) + 1)<<16>>16;
 HEAP16[$2>>1] = $4;
 $5 = ($4&65535)>(200);
 if ($5) {
  _luaX_lexerror($ls,518285,0);
 }
 $6 = ((($ls)) + 16|0);
 $7 = HEAP32[$6>>2]|0;
 do {
  switch ($7|0) {
  case 270:  {
   $$0$i1$ph = 1;
   label = 6;
   break;
  }
  case 45:  {
   $$0$i1$ph = 0;
   label = 6;
   break;
  }
  case 35:  {
   $$0$i1$ph = 2;
   label = 6;
   break;
  }
  case 284:  {
   $10 = ((($v)) + 16|0);
   HEAP32[$10>>2] = -1;
   $11 = ((($v)) + 20|0);
   HEAP32[$11>>2] = -1;
   HEAP32[$v>>2] = 5;
   $12 = ((($v)) + 8|0);
   HEAP32[$12>>2] = 0;
   $13 = ((($ls)) + 24|0);
   $14 = +HEAPF64[$13>>3];
   $15 = ((($v)) + 8|0);
   HEAPF64[$15>>3] = $14;
   label = 18;
   break;
  }
  case 286:  {
   $16 = ((($ls)) + 24|0);
   $17 = HEAP32[$16>>2]|0;
   $ls$idx$i = ((($ls)) + 48|0);
   $ls$idx$val$i = HEAP32[$ls$idx$i>>2]|0;
   $18 = (_luaK_stringK($ls$idx$val$i,$17)|0);
   $19 = ((($v)) + 16|0);
   HEAP32[$19>>2] = -1;
   $20 = ((($v)) + 20|0);
   HEAP32[$20>>2] = -1;
   HEAP32[$v>>2] = 4;
   $21 = ((($v)) + 8|0);
   HEAP32[$21>>2] = $18;
   label = 18;
   break;
  }
  case 269:  {
   $22 = ((($v)) + 16|0);
   HEAP32[$22>>2] = -1;
   $23 = ((($v)) + 20|0);
   HEAP32[$23>>2] = -1;
   HEAP32[$v>>2] = 1;
   $24 = ((($v)) + 8|0);
   HEAP32[$24>>2] = 0;
   label = 18;
   break;
  }
  case 275:  {
   $25 = ((($v)) + 16|0);
   HEAP32[$25>>2] = -1;
   $26 = ((($v)) + 20|0);
   HEAP32[$26>>2] = -1;
   HEAP32[$v>>2] = 2;
   $27 = ((($v)) + 8|0);
   HEAP32[$27>>2] = 0;
   label = 18;
   break;
  }
  case 263:  {
   $28 = ((($v)) + 16|0);
   HEAP32[$28>>2] = -1;
   $29 = ((($v)) + 20|0);
   HEAP32[$29>>2] = -1;
   HEAP32[$v>>2] = 3;
   $30 = ((($v)) + 8|0);
   HEAP32[$30>>2] = 0;
   label = 18;
   break;
  }
  case 279:  {
   $31 = ((($ls)) + 48|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = HEAP32[$32>>2]|0;
   $34 = ((($33)) + 74|0);
   $35 = HEAP8[$34>>0]|0;
   $36 = ($35<<24>>24)==(0);
   if ($36) {
    _luaX_syntaxerror($ls,518318);
    $$pre = HEAP32[$32>>2]|0;
    $$phi$trans$insert = ((($$pre)) + 74|0);
    $$pre5 = HEAP8[$$phi$trans$insert>>0]|0;
    $38 = $$pre;$40 = $$pre5;
   } else {
    $38 = $33;$40 = $35;
   }
   $37 = ((($38)) + 74|0);
   $39 = $40&255;
   $41 = $39 & 251;
   $42 = $41&255;
   HEAP8[$37>>0] = $42;
   $43 = (_luaK_codeABC($32,37,0,1,0)|0);
   $44 = ((($v)) + 16|0);
   HEAP32[$44>>2] = -1;
   $45 = ((($v)) + 20|0);
   HEAP32[$45>>2] = -1;
   HEAP32[$v>>2] = 14;
   $46 = ((($v)) + 8|0);
   HEAP32[$46>>2] = $43;
   label = 18;
   break;
  }
  case 123:  {
   _constructor($ls,$v);
   break;
  }
  case 265:  {
   _luaX_next($ls);
   $47 = ((($ls)) + 4|0);
   $48 = HEAP32[$47>>2]|0;
   _body($ls,$v,0,$48);
   break;
  }
  default: {
   _primaryexp($ls,$v);
  }
  }
 } while(0);
 if ((label|0) == 6) {
  _luaX_next($ls);
  (_subexpr($ls,$v,8)|0);
  $8 = ((($ls)) + 48|0);
  $9 = HEAP32[$8>>2]|0;
  _luaK_prefix($9,$$0$i1$ph,$v);
 }
 else if ((label|0) == 18) {
  _luaX_next($ls);
 }
 $49 = HEAP32[$6>>2]|0;
 do {
  switch ($49|0) {
  case 43:  {
   $op$0$ph$ph = 0;
   break;
  }
  case 45:  {
   $op$0$ph$ph = 1;
   break;
  }
  case 42:  {
   $op$0$ph$ph = 2;
   break;
  }
  case 47:  {
   $op$0$ph$ph = 3;
   break;
  }
  case 37:  {
   $op$0$ph$ph = 4;
   break;
  }
  case 94:  {
   $op$0$ph$ph = 5;
   break;
  }
  case 278:  {
   $op$0$ph$ph = 6;
   break;
  }
  case 283:  {
   $op$0$ph$ph = 7;
   break;
  }
  case 280:  {
   $op$0$ph$ph = 8;
   break;
  }
  case 60:  {
   $op$0$ph$ph = 9;
   break;
  }
  case 282:  {
   $op$0$ph$ph = 10;
   break;
  }
  case 62:  {
   $op$0$ph$ph = 11;
   break;
  }
  case 281:  {
   $op$0$ph$ph = 12;
   break;
  }
  case 257:  {
   $op$0$ph$ph = 13;
   break;
  }
  case 271:  {
   $op$0$ph$ph = 14;
   break;
  }
  default: {
   $op$0$lcssa = 15;
   $62 = HEAP32[$0>>2]|0;
   $63 = ((($62)) + 52|0);
   $64 = HEAP16[$63>>1]|0;
   $65 = (($64) + -1)<<16>>16;
   HEAP16[$63>>1] = $65;
   STACKTOP = sp;return ($op$0$lcssa|0);
  }
  }
 } while(0);
 $50 = ((($ls)) + 48|0);
 $op$03 = $op$0$ph$ph;
 while(1) {
  $51 = (518692 + ($op$03<<1)|0);
  $52 = HEAP8[$51>>0]|0;
  $53 = $52&255;
  $54 = ($53>>>0)>($limit>>>0);
  if (!($54)) {
   $op$0$lcssa = $op$03;
   label = 37;
   break;
  }
  _luaX_next($ls);
  $55 = HEAP32[$50>>2]|0;
  _luaK_infix($55,$op$03,$v);
  $56 = (((518692 + ($op$03<<1)|0)) + 1|0);
  $57 = HEAP8[$56>>0]|0;
  $58 = $57&255;
  $59 = (_subexpr($ls,$v2,$58)|0);
  $60 = HEAP32[$50>>2]|0;
  _luaK_posfix($60,$op$03,$v,$v2);
  $61 = ($59|0)==(15);
  if ($61) {
   $op$0$lcssa = 15;
   label = 37;
   break;
  } else {
   $op$03 = $59;
  }
 }
 if ((label|0) == 37) {
  $62 = HEAP32[$0>>2]|0;
  $63 = ((($62)) + 52|0);
  $64 = HEAP16[$63>>1]|0;
  $65 = (($64) + -1)<<16>>16;
  HEAP16[$63>>1] = $65;
  STACKTOP = sp;return ($op$0$lcssa|0);
 }
 return (0)|0;
}
function _constructor($ls,$t) {
 $ls = $ls|0;
 $t = $t|0;
 var $$pre = 0, $$pre$i = 0, $$pre5 = 0, $$pre6 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cc = 0, $vararg_buffer = 0;
 var $vararg_buffer1 = 0, $vararg_buffer10 = 0, $vararg_buffer14 = 0, $vararg_buffer5 = 0, $vararg_ptr13 = 0, $vararg_ptr17 = 0, $vararg_ptr18 = 0, $vararg_ptr4 = 0, $vararg_ptr8 = 0, $vararg_ptr9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer14 = sp + 80|0;
 $vararg_buffer10 = sp + 72|0;
 $vararg_buffer5 = sp + 56|0;
 $vararg_buffer1 = sp + 48|0;
 $vararg_buffer = sp + 40|0;
 $cc = sp;
 $0 = ((($ls)) + 48|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($ls)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_luaK_codeABC($1,10,0,0,0)|0);
 $5 = ((($cc)) + 36|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($cc)) + 28|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($cc)) + 32|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($cc)) + 24|0);
 HEAP32[$8>>2] = $t;
 $9 = ((($t)) + 16|0);
 HEAP32[$9>>2] = -1;
 $10 = ((($t)) + 20|0);
 HEAP32[$10>>2] = -1;
 HEAP32[$t>>2] = 11;
 $11 = ((($t)) + 8|0);
 HEAP32[$11>>2] = $4;
 $12 = ((($cc)) + 16|0);
 HEAP32[$12>>2] = -1;
 $13 = ((($cc)) + 20|0);
 HEAP32[$13>>2] = -1;
 HEAP32[$cc>>2] = 0;
 $14 = ((($cc)) + 8|0);
 HEAP32[$14>>2] = 0;
 $15 = HEAP32[$0>>2]|0;
 _luaK_exp2nextreg($15,$t);
 $16 = ((($ls)) + 16|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ($17|0)==(123);
 if (!($18)) {
  $19 = ((($ls)) + 52|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = (_luaX_token2str($ls,123)|0);
  HEAP32[$vararg_buffer>>2] = $21;
  $22 = (_luaO_pushfstring($20,518361,$vararg_buffer)|0);
  _luaX_syntaxerror($ls,$22);
 }
 _luaX_next($ls);
 $23 = HEAP32[$16>>2]|0;
 $24 = ($23|0)==(125);
 L4: do {
  if (!($24)) {
   $25 = ((($ls)) + 32|0);
   $27 = 0;
   while(1) {
    $26 = ($27|0)==(0);
    if (!($26)) {
     _luaK_exp2nextreg($1,$cc);
     HEAP32[$cc>>2] = 0;
     $28 = HEAP32[$5>>2]|0;
     $29 = ($28|0)==(50);
     if ($29) {
      $30 = HEAP32[$8>>2]|0;
      $31 = ((($30)) + 8|0);
      $32 = HEAP32[$31>>2]|0;
      $33 = HEAP32[$7>>2]|0;
      _luaK_setlist($1,$32,$33,50);
      HEAP32[$5>>2] = 0;
     }
    }
    $34 = HEAP32[$16>>2]|0;
    L12: do {
     switch ($34|0) {
     case 285:  {
      _luaX_lookahead($ls);
      $35 = HEAP32[$25>>2]|0;
      $36 = ($35|0)==(61);
      if ($36) {
       _recfield($ls,$cc);
       break L12;
      }
      (_subexpr($ls,$cc,0)|0);
      $37 = HEAP32[$7>>2]|0;
      $38 = ($37|0)>(2147483645);
      if ($38) {
       $39 = HEAP32[$0>>2]|0;
       $40 = HEAP32[$39>>2]|0;
       $41 = ((($40)) + 60|0);
       $42 = HEAP32[$41>>2]|0;
       $43 = ($42|0)==(0);
       $44 = ((($39)) + 16|0);
       $45 = HEAP32[$44>>2]|0;
       if ($43) {
        HEAP32[$vararg_buffer1>>2] = 2147483645;
        $vararg_ptr4 = ((($vararg_buffer1)) + 4|0);
        HEAP32[$vararg_ptr4>>2] = 518409;
        $46 = (_luaO_pushfstring($45,518375,$vararg_buffer1)|0);
        $50 = $46;
       } else {
        HEAP32[$vararg_buffer5>>2] = $42;
        $vararg_ptr8 = ((($vararg_buffer5)) + 4|0);
        HEAP32[$vararg_ptr8>>2] = 2147483645;
        $vararg_ptr9 = ((($vararg_buffer5)) + 8|0);
        HEAP32[$vararg_ptr9>>2] = 518409;
        $47 = (_luaO_pushfstring($45,518432,$vararg_buffer5)|0);
        $50 = $47;
       }
       $48 = ((($39)) + 12|0);
       $49 = HEAP32[$48>>2]|0;
       _luaX_lexerror($49,$50,0);
       $$pre5 = HEAP32[$7>>2]|0;
       $52 = $$pre5;
      } else {
       $52 = $37;
      }
      $51 = (($52) + 1)|0;
      HEAP32[$7>>2] = $51;
      $53 = HEAP32[$5>>2]|0;
      $54 = (($53) + 1)|0;
      HEAP32[$5>>2] = $54;
      break;
     }
     case 91:  {
      _recfield($ls,$cc);
      break;
     }
     default: {
      (_subexpr($ls,$cc,0)|0);
      $55 = HEAP32[$7>>2]|0;
      $56 = ($55|0)>(2147483645);
      if ($56) {
       $57 = HEAP32[$0>>2]|0;
       $58 = HEAP32[$57>>2]|0;
       $59 = ((($58)) + 60|0);
       $60 = HEAP32[$59>>2]|0;
       $61 = ($60|0)==(0);
       $62 = ((($57)) + 16|0);
       $63 = HEAP32[$62>>2]|0;
       if ($61) {
        HEAP32[$vararg_buffer10>>2] = 2147483645;
        $vararg_ptr13 = ((($vararg_buffer10)) + 4|0);
        HEAP32[$vararg_ptr13>>2] = 518409;
        $64 = (_luaO_pushfstring($63,518375,$vararg_buffer10)|0);
        $68 = $64;
       } else {
        HEAP32[$vararg_buffer14>>2] = $60;
        $vararg_ptr17 = ((($vararg_buffer14)) + 4|0);
        HEAP32[$vararg_ptr17>>2] = 2147483645;
        $vararg_ptr18 = ((($vararg_buffer14)) + 8|0);
        HEAP32[$vararg_ptr18>>2] = 518409;
        $65 = (_luaO_pushfstring($63,518432,$vararg_buffer14)|0);
        $68 = $65;
       }
       $66 = ((($57)) + 12|0);
       $67 = HEAP32[$66>>2]|0;
       _luaX_lexerror($67,$68,0);
       $$pre6 = HEAP32[$7>>2]|0;
       $70 = $$pre6;
      } else {
       $70 = $55;
      }
      $69 = (($70) + 1)|0;
      HEAP32[$7>>2] = $69;
      $71 = HEAP32[$5>>2]|0;
      $72 = (($71) + 1)|0;
      HEAP32[$5>>2] = $72;
     }
     }
    } while(0);
    $73 = HEAP32[$16>>2]|0;
    switch ($73|0) {
    case 44:  {
     _luaX_next($ls);
     break;
    }
    case 59:  {
     _luaX_next($ls);
     break;
    }
    default: {
     break L4;
    }
    }
    $74 = HEAP32[$16>>2]|0;
    $75 = ($74|0)==(125);
    if ($75) {
     break L4;
    }
    $$pre = HEAP32[$cc>>2]|0;
    $27 = $$pre;
   }
  }
 } while(0);
 _check_match($ls,125,123,$3);
 $76 = HEAP32[$5>>2]|0;
 $77 = ($76|0)==(0);
 L40: do {
  if (!($77)) {
   $78 = HEAP32[$cc>>2]|0;
   switch ($78|0) {
   case 14: case 13:  {
    _luaK_setreturns($1,$cc,-1);
    $79 = HEAP32[$8>>2]|0;
    $80 = ((($79)) + 8|0);
    $81 = HEAP32[$80>>2]|0;
    $82 = HEAP32[$7>>2]|0;
    _luaK_setlist($1,$81,$82,-1);
    $83 = HEAP32[$7>>2]|0;
    $84 = (($83) + -1)|0;
    HEAP32[$7>>2] = $84;
    break L40;
    break;
   }
   case 0:  {
    $89 = $76;
    break;
   }
   default: {
    _luaK_exp2nextreg($1,$cc);
    $$pre$i = HEAP32[$5>>2]|0;
    $89 = $$pre$i;
   }
   }
   $85 = HEAP32[$8>>2]|0;
   $86 = ((($85)) + 8|0);
   $87 = HEAP32[$86>>2]|0;
   $88 = HEAP32[$7>>2]|0;
   _luaK_setlist($1,$87,$88,$89);
  }
 } while(0);
 $90 = HEAP32[$1>>2]|0;
 $91 = ((($90)) + 12|0);
 $92 = HEAP32[$91>>2]|0;
 $93 = (($92) + ($4<<2)|0);
 $94 = HEAP32[$93>>2]|0;
 $95 = $94 & 8388607;
 $96 = HEAP32[$7>>2]|0;
 $97 = (_luaO_int2fb($96)|0);
 $98 = $97 << 23;
 $99 = $98 | $95;
 $100 = HEAP32[$1>>2]|0;
 $101 = ((($100)) + 12|0);
 $102 = HEAP32[$101>>2]|0;
 $103 = (($102) + ($4<<2)|0);
 HEAP32[$103>>2] = $99;
 $104 = $99 & -8372225;
 $105 = HEAP32[$6>>2]|0;
 $106 = (_luaO_int2fb($105)|0);
 $107 = $106 << 14;
 $108 = $107 & 8372224;
 $109 = $108 | $104;
 $110 = HEAP32[$1>>2]|0;
 $111 = ((($110)) + 12|0);
 $112 = HEAP32[$111>>2]|0;
 $113 = (($112) + ($4<<2)|0);
 HEAP32[$113>>2] = $109;
 STACKTOP = sp;return;
}
function _recfield($ls,$cc) {
 $ls = $ls|0;
 $cc = $cc|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $key = 0, $ls$idx$val$i = 0, $phitmp = 0, $val = 0, $vararg_buffer = 0, $vararg_buffer10 = 0, $vararg_buffer2 = 0, $vararg_buffer7 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, $vararg_ptr6 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer10 = sp + 80|0;
 $vararg_buffer7 = sp + 72|0;
 $vararg_buffer2 = sp + 56|0;
 $vararg_buffer = sp + 48|0;
 $key = sp + 24|0;
 $val = sp;
 $0 = ((($ls)) + 48|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 36|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($ls)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(285);
 if ($6) {
  $7 = ((($cc)) + 28|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8|0)>(2147483645);
  if ($9) {
   $10 = HEAP32[$1>>2]|0;
   $11 = ((($10)) + 60|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($12|0)==(0);
   $14 = ((($1)) + 16|0);
   $15 = HEAP32[$14>>2]|0;
   if ($13) {
    HEAP32[$vararg_buffer>>2] = 2147483645;
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
    HEAP32[$vararg_ptr1>>2] = 518409;
    $16 = (_luaO_pushfstring($15,518375,$vararg_buffer)|0);
    $20 = $16;
   } else {
    HEAP32[$vararg_buffer2>>2] = $12;
    $vararg_ptr5 = ((($vararg_buffer2)) + 4|0);
    HEAP32[$vararg_ptr5>>2] = 2147483645;
    $vararg_ptr6 = ((($vararg_buffer2)) + 8|0);
    HEAP32[$vararg_ptr6>>2] = 518409;
    $17 = (_luaO_pushfstring($15,518432,$vararg_buffer2)|0);
    $20 = $17;
   }
   $18 = ((($1)) + 12|0);
   $19 = HEAP32[$18>>2]|0;
   _luaX_lexerror($19,$20,0);
   $$pre = HEAP32[$4>>2]|0;
   $phitmp = ($$pre|0)==(285);
   if (!($phitmp)) {
    $21 = ((($ls)) + 52|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = (_luaX_token2str($ls,285)|0);
    HEAP32[$vararg_buffer7>>2] = $23;
    $24 = (_luaO_pushfstring($22,518361,$vararg_buffer7)|0);
    _luaX_syntaxerror($ls,$24);
   }
  }
  $25 = ((($ls)) + 24|0);
  $26 = HEAP32[$25>>2]|0;
  _luaX_next($ls);
  $ls$idx$val$i = HEAP32[$0>>2]|0;
  $27 = (_luaK_stringK($ls$idx$val$i,$26)|0);
  $28 = ((($key)) + 16|0);
  HEAP32[$28>>2] = -1;
  $29 = ((($key)) + 20|0);
  HEAP32[$29>>2] = -1;
  HEAP32[$key>>2] = 4;
  $30 = ((($key)) + 8|0);
  HEAP32[$30>>2] = $27;
  $$pre$phiZ2D = $7;
 } else {
  _yindex($ls,$key);
  $$pre1 = ((($cc)) + 28|0);
  $$pre$phiZ2D = $$pre1;
 }
 $31 = HEAP32[$$pre$phiZ2D>>2]|0;
 $32 = (($31) + 1)|0;
 HEAP32[$$pre$phiZ2D>>2] = $32;
 $33 = HEAP32[$4>>2]|0;
 $34 = ($33|0)==(61);
 if ($34) {
  _luaX_next($ls);
  $39 = (_luaK_exp2RK($1,$key)|0);
  (_subexpr($ls,$val,0)|0);
  $40 = ((($cc)) + 24|0);
  $41 = HEAP32[$40>>2]|0;
  $42 = ((($41)) + 8|0);
  $43 = HEAP32[$42>>2]|0;
  $44 = (_luaK_exp2RK($1,$val)|0);
  (_luaK_codeABC($1,9,$43,$39,$44)|0);
  HEAP32[$2>>2] = $3;
  STACKTOP = sp;return;
 }
 $35 = ((($ls)) + 52|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = (_luaX_token2str($ls,61)|0);
 HEAP32[$vararg_buffer10>>2] = $37;
 $38 = (_luaO_pushfstring($36,518361,$vararg_buffer10)|0);
 _luaX_syntaxerror($ls,$38);
 _luaX_next($ls);
 $39 = (_luaK_exp2RK($1,$key)|0);
 (_subexpr($ls,$val,0)|0);
 $40 = ((($cc)) + 24|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = ((($41)) + 8|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = (_luaK_exp2RK($1,$val)|0);
 (_luaK_codeABC($1,9,$43,$39,$44)|0);
 HEAP32[$2>>2] = $3;
 STACKTOP = sp;return;
}
function _yindex($ls,$v) {
 $ls = $ls|0;
 $v = $v|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _luaX_next($ls);
 (_subexpr($ls,$v,0)|0);
 $0 = ((($ls)) + 48|0);
 $1 = HEAP32[$0>>2]|0;
 _luaK_exp2val($1,$v);
 $2 = ((($ls)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(93);
 if ($4) {
  _luaX_next($ls);
  STACKTOP = sp;return;
 }
 $5 = ((($ls)) + 52|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (_luaX_token2str($ls,93)|0);
 HEAP32[$vararg_buffer>>2] = $7;
 $8 = (_luaO_pushfstring($6,518361,$vararg_buffer)|0);
 _luaX_syntaxerror($ls,$8);
 _luaX_next($ls);
 STACKTOP = sp;return;
}
function _check_match($ls,$what,$who,$where) {
 $ls = $ls|0;
 $what = $what|0;
 $who = $who|0;
 $where = $where|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_ptr4 = 0, $vararg_ptr5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($ls)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==($what|0);
 if ($2) {
  _luaX_next($ls);
  STACKTOP = sp;return;
 }
 $3 = ((($ls)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==($where|0);
 $6 = ((($ls)) + 52|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (_luaX_token2str($ls,$what)|0);
 if ($5) {
  HEAP32[$vararg_buffer>>2] = $8;
  $9 = (_luaO_pushfstring($7,518361,$vararg_buffer)|0);
  _luaX_syntaxerror($ls,$9);
  STACKTOP = sp;return;
 } else {
  $10 = (_luaX_token2str($ls,$who)|0);
  HEAP32[$vararg_buffer1>>2] = $8;
  $vararg_ptr4 = ((($vararg_buffer1)) + 4|0);
  HEAP32[$vararg_ptr4>>2] = $10;
  $vararg_ptr5 = ((($vararg_buffer1)) + 8|0);
  HEAP32[$vararg_ptr5>>2] = $where;
  $11 = (_luaO_pushfstring($7,518472,$vararg_buffer1)|0);
  _luaX_syntaxerror($ls,$11);
  STACKTOP = sp;return;
 }
}
function _body($ls,$e,$needself,$line) {
 $ls = $ls|0;
 $e = $e|0;
 $needself = $needself|0;
 $line = $line|0;
 var $$01$i$i = 0, $$phi$trans$insert = 0, $$pr$i = 0, $$pr2$i = 0, $$pre = 0, $$pre$i = 0, $$pre$phi13Z2D = 0, $$pre$phiZ2D = 0, $$pre10 = 0, $$pre12 = 0, $$pre5$i = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $i$01$i = 0, $ls$idx = 0, $ls$idx$val = 0, $ls$idx$val$i = 0, $new_fs = 0, $nparams$0$i = 0, $nparams$0$i$lcssa = 0, $nparams$1$ph$i = 0, $nparams$2$i = 0, $oldsize$02$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 592|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $new_fs = sp + 12|0;
 _open_func($ls,$new_fs);
 $0 = HEAP32[$new_fs>>2]|0;
 $1 = ((($0)) + 60|0);
 HEAP32[$1>>2] = $line;
 $2 = ((($ls)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(40);
 if (!($4)) {
  $5 = ((($ls)) + 52|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (_luaX_token2str($ls,40)|0);
  HEAP32[$vararg_buffer>>2] = $7;
  $8 = (_luaO_pushfstring($6,518361,$vararg_buffer)|0);
  _luaX_syntaxerror($ls,$8);
 }
 _luaX_next($ls);
 $9 = ($needself|0)==(0);
 if ($9) {
  $$phi$trans$insert = ((($ls)) + 48|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $$pre9 = HEAP32[$$pre>>2]|0;
  $$pre$phiZ2D = $$phi$trans$insert;$28 = $$pre9;$60 = $$pre;
 } else {
  $10 = (_luaX_newstring($ls,518513,4)|0);
  _new_localvar($ls,$10,0);
  $ls$idx = ((($ls)) + 48|0);
  $ls$idx$val = HEAP32[$ls$idx>>2]|0;
  $11 = ((($ls$idx$val)) + 50|0);
  $12 = HEAP8[$11>>0]|0;
  $13 = $12&255;
  $14 = (($13) + 1)|0;
  $15 = $14&255;
  HEAP8[$11>>0] = $15;
  $16 = ((($ls$idx$val)) + 24|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = $14 & 255;
  $19 = HEAP32[$ls$idx$val>>2]|0;
  $20 = ((($19)) + 24|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = (($18) + -1)|0;
  $23 = (((($ls$idx$val)) + 172|0) + ($22<<1)|0);
  $24 = HEAP16[$23>>1]|0;
  $25 = $24&65535;
  $26 = (((($21) + (($25*12)|0)|0)) + 4|0);
  HEAP32[$26>>2] = $17;
  $$pre$phiZ2D = $ls$idx;$28 = $19;$60 = $ls$idx$val;
 }
 $27 = ((($28)) + 74|0);
 HEAP8[$27>>0] = 0;
 $29 = HEAP32[$2>>2]|0;
 $30 = ($29|0)==(41);
 L8: do {
  if ($30) {
   $64 = 0;$nparams$2$i = 0;
  } else {
   $31 = ((($ls)) + 24|0);
   $32 = $29;$nparams$0$i = 0;
   L10: while(1) {
    switch ($32|0) {
    case 279:  {
     $nparams$0$i$lcssa = $nparams$0$i;
     break L10;
     break;
    }
    case 285:  {
     $33 = HEAP32[$31>>2]|0;
     _luaX_next($ls);
     $34 = (($nparams$0$i) + 1)|0;
     _new_localvar($ls,$33,$nparams$0$i);
     $nparams$1$ph$i = $34;
     break;
    }
    default: {
     _luaX_syntaxerror($ls,518563);
     $nparams$1$ph$i = $nparams$0$i;
    }
    }
    $$pr2$i = HEAP8[$27>>0]|0;
    $37 = ($$pr2$i<<24>>24)==(0);
    if (!($37)) {
     $64 = $$pr2$i;$nparams$2$i = $nparams$1$ph$i;
     break L8;
    }
    $38 = HEAP32[$2>>2]|0;
    $39 = ($38|0)==(44);
    if (!($39)) {
     $64 = 0;$nparams$2$i = $nparams$1$ph$i;
     break L8;
    }
    _luaX_next($ls);
    $$pr$i = HEAP32[$2>>2]|0;
    $32 = $$pr$i;$nparams$0$i = $nparams$1$ph$i;
   }
   _luaX_next($ls);
   $35 = (_luaX_newstring($ls,518559,3)|0);
   $36 = (($nparams$0$i$lcssa) + 1)|0;
   _new_localvar($ls,$35,$nparams$0$i$lcssa);
   HEAP8[$27>>0] = 7;
   $64 = 7;$nparams$2$i = $36;
  }
 } while(0);
 $ls$idx$val$i = HEAP32[$$pre$phiZ2D>>2]|0;
 $40 = ((($ls$idx$val$i)) + 50|0);
 $41 = HEAP8[$40>>0]|0;
 $42 = $41&255;
 $43 = (($42) + ($nparams$2$i))|0;
 $44 = $43&255;
 HEAP8[$40>>0] = $44;
 $45 = ($nparams$2$i|0)==(0);
 if (!($45)) {
  $46 = ((($ls$idx$val$i)) + 24|0);
  $47 = HEAP32[$46>>2]|0;
  $48 = $43 & 255;
  $49 = HEAP32[$ls$idx$val$i>>2]|0;
  $50 = ((($49)) + 24|0);
  $51 = HEAP32[$50>>2]|0;
  $$01$i$i = $nparams$2$i;
  while(1) {
   $52 = (($48) - ($$01$i$i))|0;
   $53 = (((($ls$idx$val$i)) + 172|0) + ($52<<1)|0);
   $54 = HEAP16[$53>>1]|0;
   $55 = $54&65535;
   $56 = (((($51) + (($55*12)|0)|0)) + 4|0);
   HEAP32[$56>>2] = $47;
   $57 = (($$01$i$i) + -1)|0;
   $58 = ($57|0)==(0);
   if ($58) {
    break;
   } else {
    $$01$i$i = $57;
   }
  }
 }
 $59 = ((($60)) + 50|0);
 $61 = HEAP8[$59>>0]|0;
 $62 = $61&255;
 $63 = $64&255;
 $65 = $63 & 1;
 $66 = (($62) - ($65))|0;
 $67 = $66&255;
 $68 = ((($28)) + 73|0);
 HEAP8[$68>>0] = $67;
 _luaK_reserveregs($60,$62);
 $69 = HEAP32[$2>>2]|0;
 $70 = ($69|0)==(41);
 $$pre10 = ((($ls)) + 52|0);
 if (!($70)) {
  $71 = HEAP32[$$pre10>>2]|0;
  $72 = (_luaX_token2str($ls,41)|0);
  HEAP32[$vararg_buffer1>>2] = $72;
  $73 = (_luaO_pushfstring($71,518361,$vararg_buffer1)|0);
  _luaX_syntaxerror($ls,$73);
 }
 _luaX_next($ls);
 $74 = HEAP32[$$pre10>>2]|0;
 $75 = ((($74)) + 52|0);
 $76 = HEAP16[$75>>1]|0;
 $77 = (($76) + 1)<<16>>16;
 HEAP16[$75>>1] = $77;
 $78 = ($77&65535)>(200);
 if ($78) {
  _luaX_lexerror($ls,518285,0);
 }
 L30: while(1) {
  $79 = HEAP32[$2>>2]|0;
  switch ($79|0) {
  case 287: case 276: case 262: case 261: case 260:  {
   break L30;
   break;
  }
  default: {
  }
  }
  $80 = (_statement($ls)|0);
  $81 = HEAP32[$2>>2]|0;
  $82 = ($81|0)==(59);
  if ($82) {
   _luaX_next($ls);
  }
  $83 = HEAP32[$$pre$phiZ2D>>2]|0;
  $84 = ((($83)) + 50|0);
  $85 = HEAP8[$84>>0]|0;
  $86 = $85&255;
  $87 = ((($83)) + 36|0);
  HEAP32[$87>>2] = $86;
  $88 = ($80|0)==(0);
  if (!($88)) {
   break;
  }
 }
 $89 = HEAP32[$$pre10>>2]|0;
 $90 = ((($89)) + 52|0);
 $91 = HEAP16[$90>>1]|0;
 $92 = (($91) + -1)<<16>>16;
 HEAP16[$90>>1] = $92;
 $93 = ((($ls)) + 4|0);
 $94 = HEAP32[$93>>2]|0;
 $95 = HEAP32[$new_fs>>2]|0;
 $96 = ((($95)) + 64|0);
 HEAP32[$96>>2] = $94;
 _check_match($ls,262,265,$line);
 _close_func($ls);
 $97 = HEAP32[$$pre$phiZ2D>>2]|0;
 $98 = HEAP32[$97>>2]|0;
 $99 = ((($98)) + 52|0);
 $100 = HEAP32[$99>>2]|0;
 $101 = ((($97)) + 44|0);
 $102 = HEAP32[$101>>2]|0;
 $103 = ($102|0)<($100|0);
 if ($103) {
  $$pre12 = ((($98)) + 16|0);
  $$pre$phi13Z2D = $$pre12;$109 = $100;
 } else {
  $104 = HEAP32[$$pre10>>2]|0;
  $105 = ((($98)) + 16|0);
  $106 = HEAP32[$105>>2]|0;
  $107 = (_luaM_growaux_($104,$106,$99,4,262143,519733)|0);
  HEAP32[$105>>2] = $107;
  $$pre$i = HEAP32[$99>>2]|0;
  $$pre$phi13Z2D = $105;$109 = $$pre$i;
 }
 $108 = ($100|0)<($109|0);
 if ($108) {
  $oldsize$02$i = $100;
  while(1) {
   $110 = (($oldsize$02$i) + 1)|0;
   $111 = HEAP32[$$pre$phi13Z2D>>2]|0;
   $112 = (($111) + ($oldsize$02$i<<2)|0);
   HEAP32[$112>>2] = 0;
   $exitcond = ($110|0)==($109|0);
   if ($exitcond) {
    break;
   } else {
    $oldsize$02$i = $110;
   }
  }
 }
 $113 = HEAP32[$new_fs>>2]|0;
 $114 = HEAP32[$101>>2]|0;
 $115 = (($114) + 1)|0;
 HEAP32[$101>>2] = $115;
 $116 = HEAP32[$$pre$phi13Z2D>>2]|0;
 $117 = (($116) + ($114<<2)|0);
 HEAP32[$117>>2] = $113;
 $118 = HEAP32[$new_fs>>2]|0;
 $119 = ((($118)) + 5|0);
 $120 = HEAP8[$119>>0]|0;
 $121 = $120 & 3;
 $122 = ($121<<24>>24)==(0);
 if ($122) {
  $129 = $115;
 } else {
  $123 = ((($98)) + 5|0);
  $124 = HEAP8[$123>>0]|0;
  $125 = $124 & 4;
  $126 = ($125<<24>>24)==(0);
  if ($126) {
   $129 = $115;
  } else {
   $127 = HEAP32[$$pre10>>2]|0;
   _luaC_barrierf($127,$98,$118);
   $$pre5$i = HEAP32[$101>>2]|0;
   $129 = $$pre5$i;
  }
 }
 $128 = (($129) + -1)|0;
 $130 = (_luaK_codeABx($97,36,0,$128)|0);
 $131 = ((($e)) + 16|0);
 HEAP32[$131>>2] = -1;
 $132 = ((($e)) + 20|0);
 HEAP32[$132>>2] = -1;
 HEAP32[$e>>2] = 11;
 $133 = ((($e)) + 8|0);
 HEAP32[$133>>2] = $130;
 $134 = HEAP32[$new_fs>>2]|0;
 $135 = ((($134)) + 72|0);
 $136 = HEAP8[$135>>0]|0;
 $137 = ($136<<24>>24)==(0);
 if ($137) {
  STACKTOP = sp;return;
 } else {
  $i$01$i = 0;
 }
 while(1) {
  $138 = (((($new_fs)) + 51|0) + ($i$01$i<<1)|0);
  $139 = HEAP8[$138>>0]|0;
  $140 = ($139<<24>>24)==(6);
  $141 = $140 ? 0 : 4;
  $142 = (((((($new_fs)) + 51|0) + ($i$01$i<<1)|0)) + 1|0);
  $143 = HEAP8[$142>>0]|0;
  $144 = $143&255;
  (_luaK_codeABC($97,$141,0,$144,0)|0);
  $145 = (($i$01$i) + 1)|0;
  $146 = HEAP32[$new_fs>>2]|0;
  $147 = ((($146)) + 72|0);
  $148 = HEAP8[$147>>0]|0;
  $149 = $148&255;
  $150 = ($145|0)<($149|0);
  if ($150) {
   $i$01$i = $145;
  } else {
   break;
  }
 }
 STACKTOP = sp;return;
}
function _open_func($ls,$fs) {
 $ls = $ls|0;
 $fs = $fs|0;
 var $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($ls)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_luaF_newproto($1)|0);
 HEAP32[$fs>>2] = $2;
 $3 = ((($ls)) + 48|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($fs)) + 8|0);
 HEAP32[$5>>2] = $4;
 $6 = ((($fs)) + 12|0);
 HEAP32[$6>>2] = $ls;
 $7 = ((($fs)) + 16|0);
 HEAP32[$7>>2] = $1;
 HEAP32[$3>>2] = $fs;
 $8 = ((($fs)) + 24|0);
 HEAP32[$8>>2] = 0;
 $9 = ((($fs)) + 28|0);
 HEAP32[$9>>2] = -1;
 $10 = ((($fs)) + 32|0);
 HEAP32[$10>>2] = -1;
 $11 = ((($fs)) + 36|0);
 $12 = ((($fs)) + 20|0);
 HEAP32[$12>>2] = 0;
 $13 = ((($ls)) + 64|0);
 ;HEAP32[$11>>2]=0|0;HEAP32[$11+4>>2]=0|0;HEAP32[$11+8>>2]=0|0;HEAP16[$11+12>>1]=0|0;HEAP8[$11+14>>0]=0|0;
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($2)) + 32|0);
 HEAP32[$15>>2] = $14;
 $16 = ((($2)) + 75|0);
 HEAP8[$16>>0] = 2;
 $17 = (_luaH_new($1,0,0)|0);
 $18 = ((($fs)) + 4|0);
 HEAP32[$18>>2] = $17;
 $19 = ((($1)) + 8|0);
 $20 = HEAP32[$19>>2]|0;
 HEAP32[$20>>2] = $17;
 $21 = ((($20)) + 8|0);
 HEAP32[$21>>2] = 5;
 $22 = ((($1)) + 28|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = HEAP32[$19>>2]|0;
 $25 = (($23) - ($24))|0;
 $26 = ($25|0)<(17);
 $27 = $24;
 if ($26) {
  _luaD_growstack($1,1);
  $$pre = HEAP32[$19>>2]|0;
  $29 = $$pre;
 } else {
  $29 = $27;
 }
 $28 = ((($29)) + 16|0);
 HEAP32[$19>>2] = $28;
 HEAP32[$28>>2] = $2;
 $30 = ((($29)) + 24|0);
 HEAP32[$30>>2] = 9;
 $31 = HEAP32[$22>>2]|0;
 $32 = HEAP32[$19>>2]|0;
 $33 = (($31) - ($32))|0;
 $34 = ($33|0)<(17);
 $35 = $32;
 if (!($34)) {
  $37 = $35;
  $36 = ((($37)) + 16|0);
  HEAP32[$19>>2] = $36;
  return;
 }
 _luaD_growstack($1,1);
 $$pre1 = HEAP32[$19>>2]|0;
 $37 = $$pre1;
 $36 = ((($37)) + 16|0);
 HEAP32[$19>>2] = $36;
 return;
}
function _new_localvar($ls,$name,$n) {
 $ls = $ls|0;
 $name = $name|0;
 $n = $n|0;
 var $$phi$trans$insert = 0, $$pre = 0, $$pre$i = 0, $$pre1 = 0, $$pre3$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $exitcond = 0, $oldsize$01$i = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, $vararg_ptr6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($ls)) + 48|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 50|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = (($4) + ($n))|0;
 $6 = ($5|0)>(199);
 if ($6) {
  $7 = HEAP32[$1>>2]|0;
  $8 = ((($7)) + 60|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)==(0);
  $11 = ((($1)) + 16|0);
  $12 = HEAP32[$11>>2]|0;
  if ($10) {
   HEAP32[$vararg_buffer>>2] = 200;
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
   HEAP32[$vararg_ptr1>>2] = 518518;
   $13 = (_luaO_pushfstring($12,518375,$vararg_buffer)|0);
   $17 = $13;
  } else {
   HEAP32[$vararg_buffer2>>2] = $9;
   $vararg_ptr5 = ((($vararg_buffer2)) + 4|0);
   HEAP32[$vararg_ptr5>>2] = 200;
   $vararg_ptr6 = ((($vararg_buffer2)) + 8|0);
   HEAP32[$vararg_ptr6>>2] = 518518;
   $14 = (_luaO_pushfstring($12,518432,$vararg_buffer2)|0);
   $17 = $14;
  }
  $15 = ((($1)) + 12|0);
  $16 = HEAP32[$15>>2]|0;
  _luaX_lexerror($16,$17,0);
  $$pre = HEAP32[$0>>2]|0;
  $19 = $$pre;
 } else {
  $19 = $1;
 }
 $18 = HEAP32[$19>>2]|0;
 $20 = ((($18)) + 56|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ((($19)) + 48|0);
 $23 = HEAP16[$22>>1]|0;
 $24 = $23 << 16 >> 16;
 $25 = ($24|0)<($21|0);
 if ($25) {
  $$phi$trans$insert = ((($18)) + 24|0);
  $$pre1 = HEAP32[$$phi$trans$insert>>2]|0;
  $32 = $21;$35 = $$pre1;
 } else {
  $26 = ((($ls)) + 52|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = ((($18)) + 24|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = (_luaM_growaux_($27,$29,$20,12,32767,518534)|0);
  HEAP32[$28>>2] = $30;
  $$pre$i = HEAP32[$20>>2]|0;
  $32 = $$pre$i;$35 = $30;
 }
 $31 = ($21|0)<($32|0);
 if ($31) {
  $oldsize$01$i = $21;
  while(1) {
   $33 = (($oldsize$01$i) + 1)|0;
   $34 = (($35) + (($oldsize$01$i*12)|0)|0);
   HEAP32[$34>>2] = 0;
   $exitcond = ($33|0)==($32|0);
   if ($exitcond) {
    break;
   } else {
    $oldsize$01$i = $33;
   }
  }
 }
 $36 = HEAP16[$22>>1]|0;
 $37 = $36 << 16 >> 16;
 $38 = (($35) + (($37*12)|0)|0);
 HEAP32[$38>>2] = $name;
 $39 = ((($name)) + 5|0);
 $40 = HEAP8[$39>>0]|0;
 $41 = $40 & 3;
 $42 = ($41<<24>>24)==(0);
 if ($42) {
  $50 = $36;
  $49 = (($50) + 1)<<16>>16;
  HEAP16[$22>>1] = $49;
  $51 = HEAP8[$2>>0]|0;
  $52 = $51&255;
  $53 = (($52) + ($n))|0;
  $54 = (((($1)) + 172|0) + ($53<<1)|0);
  HEAP16[$54>>1] = $50;
  STACKTOP = sp;return;
 }
 $43 = ((($18)) + 5|0);
 $44 = HEAP8[$43>>0]|0;
 $45 = $44 & 4;
 $46 = ($45<<24>>24)==(0);
 if ($46) {
  $50 = $36;
  $49 = (($50) + 1)<<16>>16;
  HEAP16[$22>>1] = $49;
  $51 = HEAP8[$2>>0]|0;
  $52 = $51&255;
  $53 = (($52) + ($n))|0;
  $54 = (((($1)) + 172|0) + ($53<<1)|0);
  HEAP16[$54>>1] = $50;
  STACKTOP = sp;return;
 }
 $47 = ((($ls)) + 52|0);
 $48 = HEAP32[$47>>2]|0;
 _luaC_barrierf($48,$18,$name);
 $$pre3$i = HEAP16[$22>>1]|0;
 $50 = $$pre3$i;
 $49 = (($50) + 1)<<16>>16;
 HEAP16[$22>>1] = $49;
 $51 = HEAP8[$2>>0]|0;
 $52 = $51&255;
 $53 = (($52) + ($n))|0;
 $54 = (((($1)) + 172|0) + ($53<<1)|0);
 HEAP16[$54>>1] = $50;
 STACKTOP = sp;return;
}
function _close_func($ls) {
 $ls = $ls|0;
 var $$off$i = 0, $$pre = 0, $$pre$i = 0, $$pre$phi10Z2D = 0, $$pre$phi14Z2D = 0, $$pre$phi16Z2D = 0, $$pre$phi20Z2D = 0, $$pre$phi22Z2D = 0, $$pre$phi26Z2D = 0, $$pre$phi28Z2D = 0, $$pre$phi2Z2D = 0, $$pre$phi4Z2D = 0, $$pre$phi8Z2D = 0, $$pre11 = 0, $$pre15 = 0, $$pre17 = 0, $$pre21 = 0, $$pre23 = 0, $$pre27 = 0, $$pre29 = 0;
 var $$pre3 = 0, $$pre5 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $12 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $indvars$iv = 0, $indvars$iv$next = 0, $switch$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($ls)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($ls)) + 48|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($3)) + 50|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = ($6<<24>>24)==(0);
 $$pre29 = ((($3)) + 24|0);
 if (!($7)) {
  $8 = ((($4)) + 24|0);
  $9 = HEAP32[$8>>2]|0;
  $$pre$i = HEAP32[$$pre29>>2]|0;
  $10 = $6&255;
  $indvars$iv = $10;
  while(1) {
   $11 = $indvars$iv&255;
   $12 = (($11) + -1)<<24>>24;
   HEAP8[$5>>0] = $12;
   $13 = $12&255;
   $14 = (((($3)) + 172|0) + ($13<<1)|0);
   $15 = HEAP16[$14>>1]|0;
   $16 = $15&65535;
   $17 = (((($9) + (($16*12)|0)|0)) + 8|0);
   HEAP32[$17>>2] = $$pre$i;
   $18 = ($12<<24>>24)==(0);
   $indvars$iv$next = (($indvars$iv) + -1)|0;
   if ($18) {
    break;
   } else {
    $indvars$iv = $indvars$iv$next;
   }
  }
 }
 _luaK_ret($3,0,0);
 $19 = HEAP32[$$pre29>>2]|0;
 $20 = (($19) + 1)|0;
 $21 = ($20>>>0)<(1073741824);
 if ($21) {
  $22 = ((($4)) + 12|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = ((($4)) + 44|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = $25 << 2;
  $27 = $19 << 2;
  $28 = (_luaM_realloc_($1,$23,$26,$27)|0);
  $$pre$phi2Z2D = $22;$$pre$phi4Z2D = $24;$30 = $28;
 } else {
  $29 = (_luaM_toobig($1)|0);
  $$pre = ((($4)) + 12|0);
  $$pre3 = ((($4)) + 44|0);
  $$pre$phi2Z2D = $$pre;$$pre$phi4Z2D = $$pre3;$30 = $29;
 }
 HEAP32[$$pre$phi2Z2D>>2] = $30;
 $31 = HEAP32[$$pre29>>2]|0;
 HEAP32[$$pre$phi4Z2D>>2] = $31;
 $32 = (($31) + 1)|0;
 $33 = ($32>>>0)<(1073741824);
 if ($33) {
  $34 = ((($4)) + 20|0);
  $35 = HEAP32[$34>>2]|0;
  $36 = ((($4)) + 48|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = $37 << 2;
  $39 = $31 << 2;
  $40 = (_luaM_realloc_($1,$35,$38,$39)|0);
  $$pre$phi10Z2D = $36;$$pre$phi8Z2D = $34;$42 = $40;
 } else {
  $41 = (_luaM_toobig($1)|0);
  $$pre5 = ((($4)) + 20|0);
  $$pre9 = ((($4)) + 48|0);
  $$pre$phi10Z2D = $$pre9;$$pre$phi8Z2D = $$pre5;$42 = $41;
 }
 HEAP32[$$pre$phi8Z2D>>2] = $42;
 $43 = HEAP32[$$pre29>>2]|0;
 HEAP32[$$pre$phi10Z2D>>2] = $43;
 $44 = ((($3)) + 40|0);
 $45 = HEAP32[$44>>2]|0;
 $46 = (($45) + 1)|0;
 $47 = ($46>>>0)<(268435456);
 if ($47) {
  $48 = ((($4)) + 8|0);
  $49 = HEAP32[$48>>2]|0;
  $50 = ((($4)) + 40|0);
  $51 = HEAP32[$50>>2]|0;
  $52 = $51 << 4;
  $53 = $45 << 4;
  $54 = (_luaM_realloc_($1,$49,$52,$53)|0);
  $$pre$phi14Z2D = $48;$$pre$phi16Z2D = $50;$56 = $54;
 } else {
  $55 = (_luaM_toobig($1)|0);
  $$pre11 = ((($4)) + 8|0);
  $$pre15 = ((($4)) + 40|0);
  $$pre$phi14Z2D = $$pre11;$$pre$phi16Z2D = $$pre15;$56 = $55;
 }
 HEAP32[$$pre$phi14Z2D>>2] = $56;
 $57 = HEAP32[$44>>2]|0;
 HEAP32[$$pre$phi16Z2D>>2] = $57;
 $58 = ((($3)) + 44|0);
 $59 = HEAP32[$58>>2]|0;
 $60 = (($59) + 1)|0;
 $61 = ($60>>>0)<(1073741824);
 if ($61) {
  $62 = ((($4)) + 16|0);
  $63 = HEAP32[$62>>2]|0;
  $64 = ((($4)) + 52|0);
  $65 = HEAP32[$64>>2]|0;
  $66 = $65 << 2;
  $67 = $59 << 2;
  $68 = (_luaM_realloc_($1,$63,$66,$67)|0);
  $$pre$phi20Z2D = $62;$$pre$phi22Z2D = $64;$70 = $68;
 } else {
  $69 = (_luaM_toobig($1)|0);
  $$pre17 = ((($4)) + 16|0);
  $$pre21 = ((($4)) + 52|0);
  $$pre$phi20Z2D = $$pre17;$$pre$phi22Z2D = $$pre21;$70 = $69;
 }
 HEAP32[$$pre$phi20Z2D>>2] = $70;
 $71 = HEAP32[$58>>2]|0;
 HEAP32[$$pre$phi22Z2D>>2] = $71;
 $72 = ((($3)) + 48|0);
 $73 = HEAP16[$72>>1]|0;
 $74 = $73 << 16 >> 16;
 $75 = (($74) + 1)|0;
 $76 = ($75>>>0)<(357913942);
 if ($76) {
  $77 = ((($4)) + 24|0);
  $78 = HEAP32[$77>>2]|0;
  $79 = ((($4)) + 56|0);
  $80 = HEAP32[$79>>2]|0;
  $81 = ($80*12)|0;
  $82 = ($74*12)|0;
  $83 = (_luaM_realloc_($1,$78,$81,$82)|0);
  $$pre$phi26Z2D = $77;$$pre$phi28Z2D = $79;$85 = $83;
 } else {
  $84 = (_luaM_toobig($1)|0);
  $$pre23 = ((($4)) + 24|0);
  $$pre27 = ((($4)) + 56|0);
  $$pre$phi26Z2D = $$pre23;$$pre$phi28Z2D = $$pre27;$85 = $84;
 }
 HEAP32[$$pre$phi26Z2D>>2] = $85;
 $86 = HEAP16[$72>>1]|0;
 $87 = $86 << 16 >> 16;
 HEAP32[$$pre$phi28Z2D>>2] = $87;
 $88 = ((($4)) + 72|0);
 $89 = HEAP8[$88>>0]|0;
 $90 = $89&255;
 $91 = ((($4)) + 28|0);
 $92 = HEAP32[$91>>2]|0;
 $93 = ((($4)) + 36|0);
 $94 = HEAP32[$93>>2]|0;
 $95 = $94 << 2;
 $96 = $90 << 2;
 $97 = (_luaM_realloc_($1,$92,$95,$96)|0);
 HEAP32[$91>>2] = $97;
 $98 = HEAP8[$88>>0]|0;
 $99 = $98&255;
 HEAP32[$93>>2] = $99;
 $100 = ((($3)) + 8|0);
 $101 = HEAP32[$100>>2]|0;
 HEAP32[$2>>2] = $101;
 $102 = ((($ls)) + 16|0);
 $103 = HEAP32[$102>>2]|0;
 $$off$i = (($103) + -285)|0;
 $switch$i = ($$off$i>>>0)<(2);
 if (!($switch$i)) {
  $109 = ((($1)) + 8|0);
  $110 = HEAP32[$109>>2]|0;
  $111 = ((($110)) + -32|0);
  HEAP32[$109>>2] = $111;
  return;
 }
 $104 = ((($ls)) + 24|0);
 $105 = HEAP32[$104>>2]|0;
 $106 = ((($105)) + 16|0);
 $107 = ((($105)) + 12|0);
 $108 = HEAP32[$107>>2]|0;
 (_luaX_newstring($ls,$106,$108)|0);
 $109 = ((($1)) + 8|0);
 $110 = HEAP32[$109>>2]|0;
 $111 = ((($110)) + -32|0);
 HEAP32[$109>>2] = $111;
 return;
}
function _primaryexp($ls,$v) {
 $ls = $ls|0;
 $v = $v|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $key = 0, $key1 = 0, $ls$idx$val$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp + 48|0;
 $key = sp + 24|0;
 $key1 = sp;
 $0 = ((($ls)) + 48|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($ls)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 switch ($3|0) {
 case 40:  {
  $4 = ((($ls)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  _luaX_next($ls);
  (_subexpr($ls,$v,0)|0);
  _check_match($ls,41,40,$5);
  $6 = HEAP32[$0>>2]|0;
  _luaK_dischargevars($6,$v);
  break;
 }
 case 285:  {
  $7 = ((($ls)) + 24|0);
  $8 = HEAP32[$7>>2]|0;
  _luaX_next($ls);
  $9 = HEAP32[$0>>2]|0;
  $10 = (_singlevaraux($9,$8,$v,1)|0);
  $11 = ($10|0)==(8);
  if ($11) {
   $12 = (_luaK_stringK($9,$8)|0);
   $13 = ((($v)) + 8|0);
   HEAP32[$13>>2] = $12;
  }
  break;
 }
 default: {
  _luaX_syntaxerror($ls,518597);
 }
 }
 $14 = ((($ls)) + 24|0);
 $15 = ((($key1)) + 16|0);
 $16 = ((($key1)) + 20|0);
 $17 = ((($key1)) + 8|0);
 $18 = ((($ls)) + 52|0);
 L7: while(1) {
  $19 = HEAP32[$2>>2]|0;
  switch ($19|0) {
  case 46:  {
   _field($ls,$v);
   continue L7;
   break;
  }
  case 91:  {
   (_luaK_exp2anyreg($1,$v)|0);
   _yindex($ls,$key);
   _luaK_indexed($1,$v,$key);
   continue L7;
   break;
  }
  case 58:  {
   _luaX_next($ls);
   $20 = HEAP32[$2>>2]|0;
   $21 = ($20|0)==(285);
   if (!($21)) {
    $22 = HEAP32[$18>>2]|0;
    $23 = (_luaX_token2str($ls,285)|0);
    HEAP32[$vararg_buffer>>2] = $23;
    $24 = (_luaO_pushfstring($22,518361,$vararg_buffer)|0);
    _luaX_syntaxerror($ls,$24);
   }
   $25 = HEAP32[$14>>2]|0;
   _luaX_next($ls);
   $ls$idx$val$i = HEAP32[$0>>2]|0;
   $26 = (_luaK_stringK($ls$idx$val$i,$25)|0);
   HEAP32[$15>>2] = -1;
   HEAP32[$16>>2] = -1;
   HEAP32[$key1>>2] = 4;
   HEAP32[$17>>2] = $26;
   _luaK_self($1,$v,$key1);
   _funcargs($ls,$v);
   continue L7;
   break;
  }
  case 123: case 286: case 40:  {
   _luaK_exp2nextreg($1,$v);
   _funcargs($ls,$v);
   continue L7;
   break;
  }
  default: {
   break L7;
  }
  }
 }
 STACKTOP = sp;return;
}
function _singlevaraux($fs,$n,$var,$base) {
 $fs = $fs|0;
 $n = $n|0;
 $var = $var|0;
 $base = $base|0;
 var $$0$i1 = 0, $$1 = 0, $$pre = 0, $$pre$i = 0, $$pre$phi11Z2D = 0, $$pre$phi12Z2D = 0, $$pre$phi13Z2D = 0, $$pre12$i = 0, $$pre13$i = 0, $$pre14$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bl$0$i = 0, $bl$03$i = 0, $bl$04$i = 0, $bl$04$i$lcssa = 0, $exitcond = 0, $i$0$i = 0, $i$0$i$lcssa29 = 0;
 var $i$0$in$i = 0, $i$0$in$i$lcssa28 = 0, $i$06$i = 0, $oldsize$05$i = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, $vararg_ptr6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($fs|0)==(0|0);
 if ($0) {
  $1 = ((($var)) + 16|0);
  HEAP32[$1>>2] = -1;
  $2 = ((($var)) + 20|0);
  HEAP32[$2>>2] = -1;
  HEAP32[$var>>2] = 8;
  $3 = ((($var)) + 8|0);
  HEAP32[$3>>2] = 255;
  $$1 = 8;
  STACKTOP = sp;return ($$1|0);
 }
 $4 = ((($fs)) + 50|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $i$0$in$i = $6;
 while(1) {
  $i$0$i = (($i$0$in$i) + -1)|0;
  $7 = ($i$0$in$i|0)>(0);
  if (!($7)) {
   break;
  }
  $8 = (((($fs)) + 172|0) + ($i$0$i<<1)|0);
  $9 = HEAP16[$8>>1]|0;
  $10 = $9&65535;
  $11 = HEAP32[$fs>>2]|0;
  $12 = ((($11)) + 24|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = (($13) + (($10*12)|0)|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ($15|0)==($n|0);
  if ($16) {
   $i$0$i$lcssa29 = $i$0$i;$i$0$in$i$lcssa28 = $i$0$in$i;
   label = 6;
   break;
  } else {
   $i$0$in$i = $i$0$i;
  }
 }
 if ((label|0) == 6) {
  $17 = ((($var)) + 16|0);
  HEAP32[$17>>2] = -1;
  $18 = ((($var)) + 20|0);
  HEAP32[$18>>2] = -1;
  HEAP32[$var>>2] = 6;
  $19 = ((($var)) + 8|0);
  HEAP32[$19>>2] = $i$0$i$lcssa29;
  $20 = ($base|0)==(0);
  if (!($20)) {
   $$1 = 6;
   STACKTOP = sp;return ($$1|0);
  }
  $21 = ((($fs)) + 20|0);
  $bl$03$i = HEAP32[$21>>2]|0;
  $22 = ($bl$03$i|0)==(0|0);
  if ($22) {
   $$1 = 6;
   STACKTOP = sp;return ($$1|0);
  } else {
   $bl$04$i = $bl$03$i;
  }
  while(1) {
   $23 = ((($bl$04$i)) + 8|0);
   $24 = HEAP8[$23>>0]|0;
   $25 = $24&255;
   $26 = ($25|0)<($i$0$in$i$lcssa28|0);
   if ($26) {
    $bl$04$i$lcssa = $bl$04$i;
    break;
   }
   $bl$0$i = HEAP32[$bl$04$i>>2]|0;
   $27 = ($bl$0$i|0)==(0|0);
   if ($27) {
    $$1 = 6;
    label = 32;
    break;
   } else {
    $bl$04$i = $bl$0$i;
   }
  }
  if ((label|0) == 32) {
   STACKTOP = sp;return ($$1|0);
  }
  $28 = ((($bl$04$i$lcssa)) + 9|0);
  HEAP8[$28>>0] = 1;
  $$1 = 6;
  STACKTOP = sp;return ($$1|0);
 }
 $29 = ((($fs)) + 8|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = (_singlevaraux($30,$n,$var,0)|0);
 $32 = ($31|0)==(8);
 if ($32) {
  $$1 = 8;
  STACKTOP = sp;return ($$1|0);
 }
 $33 = HEAP32[$fs>>2]|0;
 $34 = ((($33)) + 36|0);
 $35 = HEAP32[$34>>2]|0;
 $36 = ((($33)) + 72|0);
 $37 = HEAP8[$36>>0]|0;
 $38 = $37&255;
 $39 = ($37<<24>>24)==(0);
 L26: do {
  if ($39) {
   label = 17;
  } else {
   $40 = HEAP32[$var>>2]|0;
   $41 = ((($var)) + 8|0);
   $i$06$i = 0;
   while(1) {
    $42 = (((($fs)) + 51|0) + ($i$06$i<<1)|0);
    $43 = HEAP8[$42>>0]|0;
    $44 = $43&255;
    $45 = ($44|0)==($40|0);
    if ($45) {
     $46 = (((((($fs)) + 51|0) + ($i$06$i<<1)|0)) + 1|0);
     $47 = HEAP8[$46>>0]|0;
     $48 = $47&255;
     $49 = HEAP32[$41>>2]|0;
     $50 = ($48|0)==($49|0);
     if ($50) {
      $$0$i1 = $i$06$i;$$pre$phi11Z2D = $41;$$pre$phi12Z2D = $var;
      break L26;
     }
    }
    $51 = (($i$06$i) + 1)|0;
    $52 = ($51|0)<($38|0);
    if ($52) {
     $i$06$i = $51;
    } else {
     label = 17;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 17) {
  $53 = (($38) + 1)|0;
  $54 = ($53>>>0)>(60);
  if ($54) {
   $55 = ((($33)) + 60|0);
   $56 = HEAP32[$55>>2]|0;
   $57 = ($56|0)==(0);
   $58 = ((($fs)) + 16|0);
   $59 = HEAP32[$58>>2]|0;
   if ($57) {
    HEAP32[$vararg_buffer>>2] = 60;
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
    HEAP32[$vararg_ptr1>>2] = 518588;
    $60 = (_luaO_pushfstring($59,518375,$vararg_buffer)|0);
    $64 = $60;
   } else {
    HEAP32[$vararg_buffer2>>2] = $56;
    $vararg_ptr5 = ((($vararg_buffer2)) + 4|0);
    HEAP32[$vararg_ptr5>>2] = 60;
    $vararg_ptr6 = ((($vararg_buffer2)) + 8|0);
    HEAP32[$vararg_ptr6>>2] = 518588;
    $61 = (_luaO_pushfstring($59,518432,$vararg_buffer2)|0);
    $64 = $61;
   }
   $62 = ((($fs)) + 12|0);
   $63 = HEAP32[$62>>2]|0;
   _luaX_lexerror($63,$64,0);
   $$pre$i = HEAP8[$36>>0]|0;
   $$pre12$i = HEAP32[$34>>2]|0;
   $66 = $$pre$i;$68 = $$pre12$i;
  } else {
   $66 = $37;$68 = $35;
  }
  $65 = $66&255;
  $67 = ($65|0)<($68|0);
  if ($67) {
   $$pre = ((($33)) + 28|0);
   $$pre$phi13Z2D = $$pre;$75 = $68;
  } else {
   $69 = ((($fs)) + 16|0);
   $70 = HEAP32[$69>>2]|0;
   $71 = ((($33)) + 28|0);
   $72 = HEAP32[$71>>2]|0;
   $73 = (_luaM_growaux_($70,$72,$34,4,2147483645,1282466)|0);
   HEAP32[$71>>2] = $73;
   $$pre13$i = HEAP32[$34>>2]|0;
   $$pre$phi13Z2D = $71;$75 = $$pre13$i;
  }
  $74 = ($35|0)<($75|0);
  if ($74) {
   $oldsize$05$i = $35;
   while(1) {
    $76 = (($oldsize$05$i) + 1)|0;
    $77 = HEAP32[$$pre$phi13Z2D>>2]|0;
    $78 = (($77) + ($oldsize$05$i<<2)|0);
    HEAP32[$78>>2] = 0;
    $exitcond = ($76|0)==($75|0);
    if ($exitcond) {
     break;
    } else {
     $oldsize$05$i = $76;
    }
   }
  }
  $79 = HEAP8[$36>>0]|0;
  $80 = $79&255;
  $81 = HEAP32[$$pre$phi13Z2D>>2]|0;
  $82 = (($81) + ($80<<2)|0);
  HEAP32[$82>>2] = $n;
  $83 = ((($n)) + 5|0);
  $84 = HEAP8[$83>>0]|0;
  $85 = $84 & 3;
  $86 = ($85<<24>>24)==(0);
  if ($86) {
   $96 = $79;
  } else {
   $87 = ((($33)) + 5|0);
   $88 = HEAP8[$87>>0]|0;
   $89 = $88 & 4;
   $90 = ($89<<24>>24)==(0);
   if ($90) {
    $96 = $79;
   } else {
    $91 = ((($fs)) + 16|0);
    $92 = HEAP32[$91>>2]|0;
    _luaC_barrierf($92,$33,$n);
    $$pre14$i = HEAP8[$36>>0]|0;
    $96 = $$pre14$i;
   }
  }
  $93 = HEAP32[$var>>2]|0;
  $94 = $93&255;
  $95 = $96&255;
  $97 = (((($fs)) + 51|0) + ($95<<1)|0);
  HEAP8[$97>>0] = $94;
  $98 = ((($var)) + 8|0);
  $99 = HEAP32[$98>>2]|0;
  $100 = $99&255;
  $101 = (((((($fs)) + 51|0) + ($95<<1)|0)) + 1|0);
  HEAP8[$101>>0] = $100;
  $102 = (($96) + 1)<<24>>24;
  HEAP8[$36>>0] = $102;
  $$0$i1 = $95;$$pre$phi11Z2D = $98;$$pre$phi12Z2D = $var;
 }
 HEAP32[$$pre$phi11Z2D>>2] = $$0$i1;
 HEAP32[$$pre$phi12Z2D>>2] = 7;
 $$1 = 7;
 STACKTOP = sp;return ($$1|0);
}
function _field($ls,$v) {
 $ls = $ls|0;
 $v = $v|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $key = 0, $ls$idx$val$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp + 24|0;
 $key = sp;
 $0 = ((($ls)) + 48|0);
 $1 = HEAP32[$0>>2]|0;
 (_luaK_exp2anyreg($1,$v)|0);
 _luaX_next($ls);
 $2 = ((($ls)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(285);
 if (!($4)) {
  $5 = ((($ls)) + 52|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (_luaX_token2str($ls,285)|0);
  HEAP32[$vararg_buffer>>2] = $7;
  $8 = (_luaO_pushfstring($6,518361,$vararg_buffer)|0);
  _luaX_syntaxerror($ls,$8);
 }
 $9 = ((($ls)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 _luaX_next($ls);
 $ls$idx$val$i = HEAP32[$0>>2]|0;
 $11 = (_luaK_stringK($ls$idx$val$i,$10)|0);
 $12 = ((($key)) + 16|0);
 HEAP32[$12>>2] = -1;
 $13 = ((($key)) + 20|0);
 HEAP32[$13>>2] = -1;
 HEAP32[$key>>2] = 4;
 $14 = ((($key)) + 8|0);
 HEAP32[$14>>2] = $11;
 _luaK_indexed($1,$v,$key);
 STACKTOP = sp;return;
}
function _funcargs($ls,$f) {
 $ls = $ls|0;
 $f = $f|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $args = 0, $nparams$0 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $args = sp;
 $0 = ((($ls)) + 48|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($ls)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($ls)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 switch ($5|0) {
 case 40:  {
  $6 = ((($ls)) + 8|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($3|0)==($7|0);
  if (!($8)) {
   _luaX_syntaxerror($ls,518615);
  }
  _luaX_next($ls);
  $9 = HEAP32[$4>>2]|0;
  $10 = ($9|0)==(41);
  if ($10) {
   HEAP32[$args>>2] = 0;
  } else {
   (_subexpr($ls,$args,0)|0);
   $11 = HEAP32[$4>>2]|0;
   $12 = ($11|0)==(44);
   if ($12) {
    while(1) {
     _luaX_next($ls);
     $13 = HEAP32[$0>>2]|0;
     _luaK_exp2nextreg($13,$args);
     (_subexpr($ls,$args,0)|0);
     $14 = HEAP32[$4>>2]|0;
     $15 = ($14|0)==(44);
     if (!($15)) {
      break;
     }
    }
   }
   _luaK_setreturns($1,$args,-1);
  }
  _check_match($ls,41,40,$3);
  break;
 }
 case 123:  {
  _constructor($ls,$args);
  break;
 }
 case 286:  {
  $16 = ((($ls)) + 24|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = (_luaK_stringK($1,$17)|0);
  $19 = ((($args)) + 16|0);
  HEAP32[$19>>2] = -1;
  $20 = ((($args)) + 20|0);
  HEAP32[$20>>2] = -1;
  HEAP32[$args>>2] = 4;
  $21 = ((($args)) + 8|0);
  HEAP32[$21>>2] = $18;
  _luaX_next($ls);
  break;
 }
 default: {
  _luaX_syntaxerror($ls,518664);
  STACKTOP = sp;return;
 }
 }
 $22 = ((($f)) + 8|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = HEAP32[$args>>2]|0;
 switch ($24|0) {
 case 14: case 13:  {
  $nparams$0 = 0;
  break;
 }
 case 0:  {
  label = 15;
  break;
 }
 default: {
  _luaK_exp2nextreg($1,$args);
  label = 15;
 }
 }
 if ((label|0) == 15) {
  $25 = ((($1)) + 36|0);
  $26 = HEAP32[$25>>2]|0;
  $phitmp = (($26) - ($23))|0;
  $nparams$0 = $phitmp;
 }
 $27 = (_luaK_codeABC($1,28,$23,$nparams$0,2)|0);
 $28 = ((($f)) + 16|0);
 HEAP32[$28>>2] = -1;
 $29 = ((($f)) + 20|0);
 HEAP32[$29>>2] = -1;
 HEAP32[$f>>2] = 13;
 HEAP32[$22>>2] = $27;
 _luaK_fixline($1,$3);
 $30 = (($23) + 1)|0;
 $31 = ((($1)) + 36|0);
 HEAP32[$31>>2] = $30;
 STACKTOP = sp;return;
}
function _block($ls) {
 $ls = $ls|0;
 var $$idx$i = 0, $$idx$val$i = 0, $$mask$i = 0, $$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $7 = 0, $8 = 0, $9 = 0, $bl = 0, $indvars$iv$i = 0, $indvars$iv$next$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $bl = sp;
 $0 = ((($ls)) + 48|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($bl)) + 4|0);
 HEAP32[$2>>2] = -1;
 $3 = ((($bl)) + 10|0);
 HEAP8[$3>>0] = 0;
 $4 = ((($1)) + 50|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ((($bl)) + 8|0);
 HEAP8[$6>>0] = $5;
 $7 = ((($bl)) + 9|0);
 HEAP8[$7>>0] = 0;
 $8 = ((($1)) + 20|0);
 $9 = HEAP32[$8>>2]|0;
 HEAP32[$bl>>2] = $9;
 HEAP32[$8>>2] = $bl;
 $10 = ((($ls)) + 52|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($11)) + 52|0);
 $13 = HEAP16[$12>>1]|0;
 $14 = (($13) + 1)<<16>>16;
 HEAP16[$12>>1] = $14;
 $15 = ($14&65535)>(200);
 if ($15) {
  _luaX_lexerror($ls,518285,0);
 }
 $16 = ((($ls)) + 16|0);
 L4: while(1) {
  $17 = HEAP32[$16>>2]|0;
  switch ($17|0) {
  case 287: case 276: case 262: case 261: case 260:  {
   break L4;
   break;
  }
  default: {
  }
  }
  $18 = (_statement($ls)|0);
  $19 = HEAP32[$16>>2]|0;
  $20 = ($19|0)==(59);
  if ($20) {
   _luaX_next($ls);
  }
  $21 = HEAP32[$0>>2]|0;
  $22 = ((($21)) + 50|0);
  $23 = HEAP8[$22>>0]|0;
  $24 = $23&255;
  $25 = ((($21)) + 36|0);
  HEAP32[$25>>2] = $24;
  $26 = ($18|0)==(0);
  if (!($26)) {
   break;
  }
 }
 $27 = HEAP32[$10>>2]|0;
 $28 = ((($27)) + 52|0);
 $29 = HEAP16[$28>>1]|0;
 $30 = (($29) + -1)<<16>>16;
 HEAP16[$28>>1] = $30;
 $31 = HEAP32[$8>>2]|0;
 $32 = HEAP32[$31>>2]|0;
 HEAP32[$8>>2] = $32;
 $33 = ((($1)) + 12|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = ((($31)) + 8|0);
 $36 = HEAP16[$35>>1]|0;
 $37 = $36&255;
 $$idx$i = ((($34)) + 48|0);
 $$idx$val$i = HEAP32[$$idx$i>>2]|0;
 $38 = ((($$idx$val$i)) + 50|0);
 $39 = HEAP8[$38>>0]|0;
 $40 = ($39&255)>($37&255);
 if ($40) {
  $41 = ((($$idx$val$i)) + 24|0);
  $42 = HEAP32[$$idx$val$i>>2]|0;
  $43 = ((($42)) + 24|0);
  $44 = HEAP32[$43>>2]|0;
  $$pre$i$i = HEAP32[$41>>2]|0;
  $45 = $39&255;
  $indvars$iv$i = $45;
  while(1) {
   $46 = $indvars$iv$i&255;
   $47 = (($46) + -1)<<24>>24;
   HEAP8[$38>>0] = $47;
   $48 = $47&255;
   $49 = (((($$idx$val$i)) + 172|0) + ($48<<1)|0);
   $50 = HEAP16[$49>>1]|0;
   $51 = $50&65535;
   $52 = (((($44) + (($51*12)|0)|0)) + 8|0);
   HEAP32[$52>>2] = $$pre$i$i;
   $53 = ($47&255)>($37&255);
   $indvars$iv$next$i = (($indvars$iv$i) + -1)|0;
   if ($53) {
    $indvars$iv$i = $indvars$iv$next$i;
   } else {
    break;
   }
  }
 }
 $54 = ($36&65535)<(256);
 if ($54) {
  $56 = HEAP8[$4>>0]|0;
  $57 = $56&255;
  $58 = ((($1)) + 36|0);
  HEAP32[$58>>2] = $57;
  $59 = ((($31)) + 4|0);
  $60 = HEAP32[$59>>2]|0;
  _luaK_patchtohere($1,$60);
  STACKTOP = sp;return;
 }
 $$mask$i = $36 & 255;
 $55 = $$mask$i&65535;
 (_luaK_codeABC($1,35,$55,0,0)|0);
 $56 = HEAP8[$4>>0]|0;
 $57 = $56&255;
 $58 = ((($1)) + 36|0);
 HEAP32[$58>>2] = $57;
 $59 = ((($31)) + 4|0);
 $60 = HEAP32[$59>>2]|0;
 _luaK_patchtohere($1,$60);
 STACKTOP = sp;return;
}
function _forbody($ls,$base,$line,$nvars,$isnum) {
 $ls = $ls|0;
 $base = $base|0;
 $line = $line|0;
 $nvars = $nvars|0;
 $isnum = $isnum|0;
 var $$01$i2 = 0, $$idx$i = 0, $$idx$val$i = 0, $$mask$i = 0, $$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $bl = 0, $indvars$iv$i = 0, $indvars$iv$next$i = 0, $ls$idx$val = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $bl = sp + 4|0;
 $0 = ((($ls)) + 48|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 50|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = (($4) + 3)|0;
 $6 = $5&255;
 HEAP8[$2>>0] = $6;
 $7 = ((($1)) + 24|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $5 & 255;
 $10 = HEAP32[$1>>2]|0;
 $11 = ((($10)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($9) + -3)|0;
 $14 = (((($1)) + 172|0) + ($13<<1)|0);
 $15 = HEAP16[$14>>1]|0;
 $16 = $15&65535;
 $17 = (((($12) + (($16*12)|0)|0)) + 4|0);
 HEAP32[$17>>2] = $8;
 $18 = (($9) + -2)|0;
 $19 = (((($1)) + 172|0) + ($18<<1)|0);
 $20 = HEAP16[$19>>1]|0;
 $21 = $20&65535;
 $22 = (((($12) + (($21*12)|0)|0)) + 4|0);
 HEAP32[$22>>2] = $8;
 $23 = (($9) + -1)|0;
 $24 = (((($1)) + 172|0) + ($23<<1)|0);
 $25 = HEAP16[$24>>1]|0;
 $26 = $25&65535;
 $27 = (((($12) + (($26*12)|0)|0)) + 4|0);
 HEAP32[$27>>2] = $8;
 $28 = ((($ls)) + 16|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = ($29|0)==(259);
 if (!($30)) {
  $31 = ((($ls)) + 52|0);
  $32 = HEAP32[$31>>2]|0;
  $33 = (_luaX_token2str($ls,259)|0);
  HEAP32[$vararg_buffer>>2] = $33;
  $34 = (_luaO_pushfstring($32,518361,$vararg_buffer)|0);
  _luaX_syntaxerror($ls,$34);
 }
 _luaX_next($ls);
 $35 = ($isnum|0)!=(0);
 if ($35) {
  $36 = (_luaK_codeABx($1,32,$base,131070)|0);
  $94 = $36;
 } else {
  $37 = (_luaK_jump($1)|0);
  $94 = $37;
 }
 $38 = ((($bl)) + 4|0);
 HEAP32[$38>>2] = -1;
 $39 = ((($bl)) + 10|0);
 HEAP8[$39>>0] = 0;
 $40 = HEAP8[$2>>0]|0;
 $41 = ((($bl)) + 8|0);
 HEAP8[$41>>0] = $40;
 $42 = ((($bl)) + 9|0);
 HEAP8[$42>>0] = 0;
 $43 = ((($1)) + 20|0);
 $44 = HEAP32[$43>>2]|0;
 HEAP32[$bl>>2] = $44;
 HEAP32[$43>>2] = $bl;
 $ls$idx$val = HEAP32[$0>>2]|0;
 $45 = ((($ls$idx$val)) + 50|0);
 $46 = HEAP8[$45>>0]|0;
 $47 = $46&255;
 $48 = (($47) + ($nvars))|0;
 $49 = $48&255;
 HEAP8[$45>>0] = $49;
 $50 = ($nvars|0)==(0);
 if (!($50)) {
  $51 = ((($ls$idx$val)) + 24|0);
  $52 = HEAP32[$51>>2]|0;
  $53 = $48 & 255;
  $54 = HEAP32[$ls$idx$val>>2]|0;
  $55 = ((($54)) + 24|0);
  $56 = HEAP32[$55>>2]|0;
  $$01$i2 = $nvars;
  while(1) {
   $57 = (($53) - ($$01$i2))|0;
   $58 = (((($ls$idx$val)) + 172|0) + ($57<<1)|0);
   $59 = HEAP16[$58>>1]|0;
   $60 = $59&65535;
   $61 = (((($56) + (($60*12)|0)|0)) + 4|0);
   HEAP32[$61>>2] = $52;
   $62 = (($$01$i2) + -1)|0;
   $63 = ($62|0)==(0);
   if ($63) {
    break;
   } else {
    $$01$i2 = $62;
   }
  }
 }
 _luaK_reserveregs($1,$nvars);
 _block($ls);
 $64 = HEAP32[$43>>2]|0;
 $65 = HEAP32[$64>>2]|0;
 HEAP32[$43>>2] = $65;
 $66 = ((($1)) + 12|0);
 $67 = HEAP32[$66>>2]|0;
 $68 = ((($64)) + 8|0);
 $69 = HEAP16[$68>>1]|0;
 $70 = $69&255;
 $$idx$i = ((($67)) + 48|0);
 $$idx$val$i = HEAP32[$$idx$i>>2]|0;
 $71 = ((($$idx$val$i)) + 50|0);
 $72 = HEAP8[$71>>0]|0;
 $73 = ($72&255)>($70&255);
 if ($73) {
  $74 = ((($$idx$val$i)) + 24|0);
  $75 = HEAP32[$$idx$val$i>>2]|0;
  $76 = ((($75)) + 24|0);
  $77 = HEAP32[$76>>2]|0;
  $$pre$i$i = HEAP32[$74>>2]|0;
  $78 = $72&255;
  $indvars$iv$i = $78;
  while(1) {
   $79 = $indvars$iv$i&255;
   $80 = (($79) + -1)<<24>>24;
   HEAP8[$71>>0] = $80;
   $81 = $80&255;
   $82 = (((($$idx$val$i)) + 172|0) + ($81<<1)|0);
   $83 = HEAP16[$82>>1]|0;
   $84 = $83&65535;
   $85 = (((($77) + (($84*12)|0)|0)) + 8|0);
   HEAP32[$85>>2] = $$pre$i$i;
   $86 = ($80&255)>($70&255);
   $indvars$iv$next$i = (($indvars$iv$i) + -1)|0;
   if ($86) {
    $indvars$iv$i = $indvars$iv$next$i;
   } else {
    break;
   }
  }
 }
 $87 = ($69&65535)<(256);
 if (!($87)) {
  $$mask$i = $69 & 255;
  $88 = $$mask$i&65535;
  (_luaK_codeABC($1,35,$88,0,0)|0);
 }
 $89 = HEAP8[$2>>0]|0;
 $90 = $89&255;
 $91 = ((($1)) + 36|0);
 HEAP32[$91>>2] = $90;
 $92 = ((($64)) + 4|0);
 $93 = HEAP32[$92>>2]|0;
 _luaK_patchtohere($1,$93);
 _luaK_patchtohere($1,$94);
 if ($35) {
  $95 = (_luaK_codeABx($1,31,$base,131070)|0);
  _luaK_fixline($1,$line);
  $98 = $95;
  $97 = (($94) + 1)|0;
  _luaK_patchlist($1,$98,$97);
  STACKTOP = sp;return;
 } else {
  (_luaK_codeABC($1,33,$base,0,$nvars)|0);
  _luaK_fixline($1,$line);
  $96 = (_luaK_jump($1)|0);
  $98 = $96;
  $97 = (($94) + 1)|0;
  _luaK_patchlist($1,$98,$97);
  STACKTOP = sp;return;
 }
}
function _assignment($ls,$lh,$nvars) {
 $ls = $ls|0;
 $lh = $lh|0;
 $nvars = $nvars|0;
 var $$$i = 0, $$01$i = 0, $$off = 0, $$pre$i = 0, $$pre$phi4Z2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $9 = 0, $conflict$02$i = 0, $conflict$1$i = 0, $conflict$2$i = 0, $conflict$2$i$lcssa = 0, $e = 0, $ls$idx1 = 0, $ls$idx1$val = 0, $n$0$i$lcssa = 0, $n$0$i2 = 0, $nv = 0, $vararg_buffer = 0;
 var $vararg_buffer2 = 0, $vararg_buffer7 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, $vararg_ptr6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 80|0;
 $vararg_buffer2 = sp + 64|0;
 $vararg_buffer = sp + 56|0;
 $e = sp + 32|0;
 $nv = sp;
 $0 = ((($lh)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $$off = (($1) + -6)|0;
 $2 = ($$off>>>0)<(4);
 if (!($2)) {
  _luaX_syntaxerror($ls,518837);
 }
 $3 = ((($ls)) + 16|0);
 $4 = HEAP32[$3>>2]|0;
 switch ($4|0) {
 case 44:  {
  _luaX_next($ls);
  HEAP32[$nv>>2] = $lh;
  $5 = ((($nv)) + 8|0);
  _primaryexp($ls,$5);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6|0)==(6);
  if ($7) {
   $ls$idx1 = ((($ls)) + 48|0);
   $ls$idx1$val = HEAP32[$ls$idx1>>2]|0;
   $8 = ((($ls$idx1$val)) + 36|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = ($lh|0)==(0|0);
   if (!($10)) {
    $11 = ((($nv)) + 16|0);
    $$01$i = $lh;$conflict$02$i = 0;
    while(1) {
     $12 = ((($$01$i)) + 8|0);
     $13 = HEAP32[$12>>2]|0;
     $14 = ($13|0)==(9);
     if ($14) {
      $15 = ((($$01$i)) + 16|0);
      $16 = HEAP32[$15>>2]|0;
      $17 = HEAP32[$11>>2]|0;
      $18 = ($16|0)==($17|0);
      if ($18) {
       HEAP32[$15>>2] = $9;
       $$pre$i = HEAP32[$11>>2]|0;
       $22 = $$pre$i;$conflict$1$i = 1;
      } else {
       $22 = $17;$conflict$1$i = $conflict$02$i;
      }
      $19 = ((($15)) + 4|0);
      $20 = HEAP32[$19>>2]|0;
      $21 = ($20|0)==($22|0);
      if ($21) {
       HEAP32[$19>>2] = $9;
       $conflict$2$i = 1;
      } else {
       $conflict$2$i = $conflict$1$i;
      }
     } else {
      $conflict$2$i = $conflict$02$i;
     }
     $23 = HEAP32[$$01$i>>2]|0;
     $24 = ($23|0)==(0|0);
     if ($24) {
      $conflict$2$i$lcssa = $conflict$2$i;
      break;
     } else {
      $$01$i = $23;$conflict$02$i = $conflict$2$i;
     }
    }
    $25 = ($conflict$2$i$lcssa|0)==(0);
    if (!($25)) {
     $26 = HEAP32[$11>>2]|0;
     (_luaK_codeABC($ls$idx1$val,0,$9,$26,0)|0);
     _luaK_reserveregs($ls$idx1$val,1);
    }
   }
  }
  $27 = ((($ls)) + 52|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = ((($28)) + 52|0);
  $30 = HEAP16[$29>>1]|0;
  $31 = $30&65535;
  $32 = (200 - ($31))|0;
  $33 = ($32|0)<($nvars|0);
  $34 = ((($ls)) + 48|0);
  if ($33) {
   $35 = HEAP32[$34>>2]|0;
   $36 = HEAP32[$35>>2]|0;
   $37 = ((($36)) + 60|0);
   $38 = HEAP32[$37>>2]|0;
   $39 = ($38|0)==(0);
   $40 = ((($35)) + 16|0);
   $41 = HEAP32[$40>>2]|0;
   if ($39) {
    HEAP32[$vararg_buffer>>2] = $32;
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
    HEAP32[$vararg_ptr1>>2] = 518850;
    $42 = (_luaO_pushfstring($41,518375,$vararg_buffer)|0);
    $46 = $42;
   } else {
    HEAP32[$vararg_buffer2>>2] = $38;
    $vararg_ptr5 = ((($vararg_buffer2)) + 4|0);
    HEAP32[$vararg_ptr5>>2] = $32;
    $vararg_ptr6 = ((($vararg_buffer2)) + 8|0);
    HEAP32[$vararg_ptr6>>2] = 518850;
    $43 = (_luaO_pushfstring($41,518432,$vararg_buffer2)|0);
    $46 = $43;
   }
   $44 = ((($35)) + 12|0);
   $45 = HEAP32[$44>>2]|0;
   _luaX_lexerror($45,$46,0);
  }
  $47 = (($nvars) + 1)|0;
  _assignment($ls,$nv,$47);
  $$pre$phi4Z2D = $e;$$pre$phiZ2D = $34;
  break;
 }
 case 61:  {
  label = 22;
  break;
 }
 default: {
  $48 = ((($ls)) + 52|0);
  $49 = HEAP32[$48>>2]|0;
  $50 = (_luaX_token2str($ls,61)|0);
  HEAP32[$vararg_buffer7>>2] = $50;
  $51 = (_luaO_pushfstring($49,518361,$vararg_buffer7)|0);
  _luaX_syntaxerror($ls,$51);
  label = 22;
 }
 }
 if ((label|0) == 22) {
  _luaX_next($ls);
  (_subexpr($ls,$e,0)|0);
  $52 = HEAP32[$3>>2]|0;
  $53 = ($52|0)==(44);
  $54 = ((($ls)) + 48|0);
  if ($53) {
   $n$0$i2 = 1;
   while(1) {
    _luaX_next($ls);
    $55 = HEAP32[$54>>2]|0;
    _luaK_exp2nextreg($55,$e);
    (_subexpr($ls,$e,0)|0);
    $56 = (($n$0$i2) + 1)|0;
    $57 = HEAP32[$3>>2]|0;
    $58 = ($57|0)==(44);
    if ($58) {
     $n$0$i2 = $56;
    } else {
     $n$0$i$lcssa = $56;
     break;
    }
   }
  } else {
   $n$0$i$lcssa = 1;
  }
  $59 = ($n$0$i$lcssa|0)==($nvars|0);
  $60 = HEAP32[$54>>2]|0;
  if ($59) {
   _luaK_setoneret($60,$e);
   $75 = HEAP32[$54>>2]|0;
   _luaK_storevar($75,$0,$e);
   STACKTOP = sp;return;
  }
  $61 = (($nvars) - ($n$0$i$lcssa))|0;
  $62 = HEAP32[$e>>2]|0;
  switch ($62|0) {
  case 14: case 13:  {
   $63 = (($61) + 1)|0;
   $64 = ($63|0)<(0);
   $$$i = $64 ? 0 : $63;
   _luaK_setreturns($60,$e,$$$i);
   $65 = ($$$i|0)>(1);
   if ($65) {
    $66 = (($$$i) + -1)|0;
    _luaK_reserveregs($60,$66);
   }
   break;
  }
  case 0:  {
   label = 29;
   break;
  }
  default: {
   _luaK_exp2nextreg($60,$e);
   label = 29;
  }
  }
  if ((label|0) == 29) {
   $67 = ($61|0)>(0);
   if ($67) {
    $68 = ((($60)) + 36|0);
    $69 = HEAP32[$68>>2]|0;
    _luaK_reserveregs($60,$61);
    _luaK_nil($60,$69,$61);
   }
  }
  $70 = ($n$0$i$lcssa|0)>($nvars|0);
  if ($70) {
   $71 = HEAP32[$54>>2]|0;
   $72 = ((($71)) + 36|0);
   $73 = HEAP32[$72>>2]|0;
   $74 = (($61) + ($73))|0;
   HEAP32[$72>>2] = $74;
   $$pre$phi4Z2D = $e;$$pre$phiZ2D = $54;
  } else {
   $$pre$phi4Z2D = $e;$$pre$phiZ2D = $54;
  }
 }
 $76 = HEAP32[$$pre$phiZ2D>>2]|0;
 $77 = ((($76)) + 36|0);
 $78 = HEAP32[$77>>2]|0;
 $79 = (($78) + -1)|0;
 $80 = ((($e)) + 16|0);
 HEAP32[$80>>2] = -1;
 $81 = ((($e)) + 20|0);
 HEAP32[$81>>2] = -1;
 HEAP32[$$pre$phi4Z2D>>2] = 12;
 $82 = ((($e)) + 8|0);
 HEAP32[$82>>2] = $79;
 _luaK_storevar($76,$0,$e);
 STACKTOP = sp;return;
}
function _luaE_newthread($L) {
 $L = $L|0;
 var $$cast$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaM_realloc_($L,0,0,120)|0);
 _luaC_link($L,$0,8);
 $1 = ((($L)) + 16|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($0)) + 16|0);
 HEAP32[$3>>2] = $2;
 $4 = ((($0)) + 32|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($0)) + 44|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($0)) + 112|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($0)) + 68|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($0)) + 56|0);
 HEAP8[$8>>0] = 0;
 $9 = ((($0)) + 60|0);
 HEAP32[$9>>2] = 0;
 $10 = ((($0)) + 57|0);
 HEAP8[$10>>0] = 1;
 $11 = ((($0)) + 64|0);
 HEAP32[$11>>2] = 0;
 $12 = ((($0)) + 104|0);
 HEAP32[$12>>2] = 0;
 $13 = ((($0)) + 48|0);
 HEAP32[$13>>2] = 0;
 $14 = ((($0)) + 54|0);
 HEAP16[$14>>1] = 0;
 $15 = ((($0)) + 52|0);
 HEAP16[$15>>1] = 0;
 $16 = ((($0)) + 6|0);
 HEAP8[$16>>0] = 0;
 $17 = ((($0)) + 20|0);
 HEAP32[$17>>2] = 0;
 $18 = ((($0)) + 40|0);
 HEAP32[$18>>2] = 0;
 $19 = ((($0)) + 24|0);
 HEAP32[$19>>2] = 0;
 $20 = ((($0)) + 116|0);
 HEAP32[$20>>2] = 0;
 $21 = ((($0)) + 80|0);
 HEAP32[$21>>2] = 0;
 $22 = (_luaM_realloc_($L,0,0,192)|0);
 HEAP32[$18>>2] = $22;
 HEAP32[$17>>2] = $22;
 HEAP32[$13>>2] = 8;
 $23 = ((($22)) + 168|0);
 $24 = ((($0)) + 36|0);
 HEAP32[$24>>2] = $23;
 $25 = (_luaM_realloc_($L,0,0,720)|0);
 HEAP32[$4>>2] = $25;
 HEAP32[$5>>2] = 45;
 $$cast$i = $25;
 $26 = ((($0)) + 8|0);
 $27 = ((($25)) + 624|0);
 $28 = ((($0)) + 28|0);
 HEAP32[$28>>2] = $27;
 $29 = HEAP32[$17>>2]|0;
 $30 = ((($29)) + 4|0);
 HEAP32[$30>>2] = $$cast$i;
 $31 = ((($25)) + 16|0);
 HEAP32[$26>>2] = $31;
 $32 = ((($25)) + 8|0);
 HEAP32[$32>>2] = 0;
 $33 = $31;
 HEAP32[$29>>2] = $33;
 $34 = ((($0)) + 12|0);
 HEAP32[$34>>2] = $33;
 $35 = ((($25)) + 336|0);
 $36 = ((($29)) + 8|0);
 HEAP32[$36>>2] = $35;
 $37 = ((($L)) + 72|0);
 $38 = ((($0)) + 72|0);
 $39 = $37;
 $40 = $39;
 $41 = HEAP32[$40>>2]|0;
 $42 = (($39) + 4)|0;
 $43 = $42;
 $44 = HEAP32[$43>>2]|0;
 $45 = $38;
 $46 = $45;
 HEAP32[$46>>2] = $41;
 $47 = (($45) + 4)|0;
 $48 = $47;
 HEAP32[$48>>2] = $44;
 $49 = ((($L)) + 80|0);
 $50 = HEAP32[$49>>2]|0;
 HEAP32[$21>>2] = $50;
 $51 = ((($L)) + 56|0);
 $52 = HEAP8[$51>>0]|0;
 HEAP8[$8>>0] = $52;
 $53 = ((($L)) + 60|0);
 $54 = HEAP32[$53>>2]|0;
 HEAP32[$9>>2] = $54;
 $55 = ((($L)) + 68|0);
 $56 = HEAP32[$55>>2]|0;
 HEAP32[$7>>2] = $56;
 HEAP32[$11>>2] = $54;
 return ($0|0);
}
function _luaE_freethread($L,$L1) {
 $L = $L|0;
 $L1 = $L1|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L1)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 _luaF_close($L1,$1);
 $2 = ((($L1)) + 40|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($L1)) + 48|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5*24)|0;
 (_luaM_realloc_($L,$3,$6,0)|0);
 $7 = HEAP32[$0>>2]|0;
 $8 = ((($L1)) + 44|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 4;
 (_luaM_realloc_($L,$7,$10,0)|0);
 (_luaM_realloc_($L,$L1,120,0)|0);
 return;
}
function _lua_newstate($f,$ud) {
 $f = $f|0;
 $ud = $ud|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $8 = 0, $9 = 0, $scevgep = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = (FUNCTION_TABLE_iiiii[$f & 8191]($ud,0,0,376)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  $$0 = 0;
  return ($$0|0);
 }
 $2 = ((($0)) + 120|0);
 HEAP32[$0>>2] = 0;
 $3 = ((($0)) + 4|0);
 HEAP8[$3>>0] = 8;
 $4 = ((($0)) + 140|0);
 HEAP8[$4>>0] = 33;
 $5 = ((($0)) + 5|0);
 HEAP8[$5>>0] = 97;
 $6 = ((($0)) + 16|0);
 HEAP32[$6>>2] = $2;
 $7 = ((($0)) + 32|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($0)) + 44|0);
 HEAP32[$8>>2] = 0;
 $9 = ((($0)) + 112|0);
 HEAP32[$9>>2] = 0;
 $10 = ((($0)) + 68|0);
 HEAP32[$10>>2] = 0;
 $11 = ((($0)) + 56|0);
 HEAP8[$11>>0] = 0;
 $12 = ((($0)) + 60|0);
 HEAP32[$12>>2] = 0;
 $13 = ((($0)) + 57|0);
 HEAP8[$13>>0] = 1;
 $14 = ((($0)) + 64|0);
 HEAP32[$14>>2] = 0;
 $15 = ((($0)) + 104|0);
 HEAP32[$15>>2] = 0;
 $16 = ((($0)) + 48|0);
 HEAP32[$16>>2] = 0;
 $17 = ((($0)) + 54|0);
 HEAP16[$17>>1] = 0;
 $18 = ((($0)) + 52|0);
 HEAP16[$18>>1] = 0;
 $19 = ((($0)) + 6|0);
 HEAP8[$19>>0] = 0;
 $20 = ((($0)) + 20|0);
 HEAP32[$20>>2] = 0;
 $21 = ((($0)) + 40|0);
 HEAP32[$21>>2] = 0;
 $22 = ((($0)) + 24|0);
 HEAP32[$22>>2] = 0;
 $23 = ((($0)) + 116|0);
 HEAP32[$23>>2] = 0;
 $24 = ((($0)) + 80|0);
 HEAP32[$24>>2] = 0;
 $25 = ((($0)) + 132|0);
 HEAP32[$25>>2] = $f;
 $26 = ((($0)) + 136|0);
 HEAP32[$26>>2] = $ud;
 $27 = ((($0)) + 232|0);
 HEAP32[$27>>2] = $0;
 $28 = ((($0)) + 240|0);
 $29 = ((($0)) + 256|0);
 HEAP32[$29>>2] = $28;
 $30 = ((($0)) + 260|0);
 HEAP32[$30>>2] = $28;
 $31 = ((($0)) + 184|0);
 HEAP32[$31>>2] = 0;
 $32 = ((($0)) + 128|0);
 HEAP32[$32>>2] = 0;
 $33 = ((($0)) + 124|0);
 HEAP32[$33>>2] = 0;
 HEAP32[$2>>2] = 0;
 $34 = ((($0)) + 224|0);
 HEAP32[$34>>2] = 0;
 $35 = ((($0)) + 172|0);
 HEAP32[$35>>2] = 0;
 $36 = ((($0)) + 180|0);
 HEAP32[$36>>2] = 0;
 $37 = ((($0)) + 208|0);
 HEAP32[$37>>2] = 0;
 $38 = ((($0)) + 141|0);
 HEAP8[$38>>0] = 0;
 $39 = ((($0)) + 148|0);
 HEAP32[$39>>2] = $0;
 $40 = ((($0)) + 144|0);
 HEAP32[$40>>2] = 0;
 $41 = ((($0)) + 152|0);
 HEAP32[$41>>2] = $39;
 $42 = ((($0)) + 156|0);
 $43 = ((($0)) + 188|0);
 ;HEAP32[$42>>2]=0|0;HEAP32[$42+4>>2]=0|0;HEAP32[$42+8>>2]=0|0;HEAP32[$42+12>>2]=0|0;
 HEAP32[$43>>2] = 376;
 $44 = ((($0)) + 200|0);
 HEAP32[$44>>2] = 200;
 $45 = ((($0)) + 204|0);
 HEAP32[$45>>2] = 200;
 $46 = ((($0)) + 196|0);
 HEAP32[$46>>2] = 0;
 $scevgep = ((($0)) + 272|0);
 dest=$scevgep; stop=dest+36|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $47 = (_luaD_rawrunprotected($0,4294,0)|0);
 $48 = ($47|0)==(0);
 if ($48) {
  $$0 = $0;
  return ($$0|0);
 }
 $49 = HEAP32[$6>>2]|0;
 $50 = HEAP32[$7>>2]|0;
 _luaF_close($0,$50);
 _luaC_freeall($0);
 $51 = HEAP32[$6>>2]|0;
 $52 = HEAP32[$51>>2]|0;
 $53 = ((($51)) + 8|0);
 $54 = HEAP32[$53>>2]|0;
 $55 = $54 << 2;
 (_luaM_realloc_($0,$52,$55,0)|0);
 $56 = ((($49)) + 52|0);
 $57 = HEAP32[$56>>2]|0;
 $58 = ((($49)) + 60|0);
 $59 = HEAP32[$58>>2]|0;
 $60 = (_luaM_realloc_($0,$57,$59,0)|0);
 HEAP32[$56>>2] = $60;
 HEAP32[$58>>2] = 0;
 $61 = HEAP32[$21>>2]|0;
 $62 = HEAP32[$16>>2]|0;
 $63 = ($62*24)|0;
 (_luaM_realloc_($0,$61,$63,0)|0);
 $64 = HEAP32[$7>>2]|0;
 $65 = HEAP32[$8>>2]|0;
 $66 = $65 << 4;
 (_luaM_realloc_($0,$64,$66,0)|0);
 $67 = ((($49)) + 12|0);
 $68 = HEAP32[$67>>2]|0;
 $69 = ((($49)) + 16|0);
 $70 = HEAP32[$69>>2]|0;
 (FUNCTION_TABLE_iiiii[$68 & 8191]($70,$0,376,0)|0);
 $$0 = 0;
 return ($$0|0);
}
function _f_luaopen($L,$ud) {
 $L = $L|0;
 $ud = $ud|0;
 var $$cast$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_luaM_realloc_($L,0,0,192)|0);
 $3 = ((($L)) + 40|0);
 HEAP32[$3>>2] = $2;
 $4 = ((($L)) + 20|0);
 HEAP32[$4>>2] = $2;
 $5 = ((($L)) + 48|0);
 HEAP32[$5>>2] = 8;
 $6 = ((($2)) + 168|0);
 $7 = ((($L)) + 36|0);
 HEAP32[$7>>2] = $6;
 $8 = (_luaM_realloc_($L,0,0,720)|0);
 $9 = ((($L)) + 32|0);
 HEAP32[$9>>2] = $8;
 $10 = ((($L)) + 44|0);
 HEAP32[$10>>2] = 45;
 $$cast$i = $8;
 $11 = ((($L)) + 8|0);
 $12 = ((($8)) + 624|0);
 $13 = ((($L)) + 28|0);
 HEAP32[$13>>2] = $12;
 $14 = HEAP32[$4>>2]|0;
 $15 = ((($14)) + 4|0);
 HEAP32[$15>>2] = $$cast$i;
 $16 = ((($8)) + 16|0);
 HEAP32[$11>>2] = $16;
 $17 = ((($8)) + 8|0);
 HEAP32[$17>>2] = 0;
 $18 = $16;
 HEAP32[$14>>2] = $18;
 $19 = ((($L)) + 12|0);
 HEAP32[$19>>2] = $18;
 $20 = ((($8)) + 336|0);
 $21 = ((($14)) + 8|0);
 HEAP32[$21>>2] = $20;
 $22 = ((($L)) + 72|0);
 $23 = (_luaH_new($L,0,2)|0);
 HEAP32[$22>>2] = $23;
 $24 = ((($L)) + 80|0);
 HEAP32[$24>>2] = 5;
 $25 = HEAP32[$0>>2]|0;
 $26 = ((($25)) + 96|0);
 $27 = (_luaH_new($L,0,2)|0);
 HEAP32[$26>>2] = $27;
 $28 = ((($25)) + 104|0);
 HEAP32[$28>>2] = 5;
 _luaS_resize($L,32);
 _luaT_init($L);
 _luaX_init($L);
 $29 = (_luaS_newlstr($L,518874,17)|0);
 $30 = ((($29)) + 5|0);
 $31 = HEAP8[$30>>0]|0;
 $32 = $31&255;
 $33 = $32 | 32;
 $34 = $33&255;
 HEAP8[$30>>0] = $34;
 $35 = ((($1)) + 68|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = $36 << 2;
 $38 = ((($1)) + 64|0);
 HEAP32[$38>>2] = $37;
 return;
}
function _lua_close($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 112|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 32|0);
 $5 = HEAP32[$4>>2]|0;
 _luaF_close($3,$5);
 (_luaC_separateudata($3,1)|0);
 $6 = ((($3)) + 116|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($3)) + 40|0);
 $8 = ((($3)) + 20|0);
 $9 = ((($3)) + 8|0);
 $10 = ((($3)) + 12|0);
 $11 = ((($3)) + 54|0);
 $12 = ((($3)) + 52|0);
 while(1) {
  $13 = HEAP32[$7>>2]|0;
  HEAP32[$8>>2] = $13;
  $14 = HEAP32[$13>>2]|0;
  HEAP32[$9>>2] = $14;
  HEAP32[$10>>2] = $14;
  HEAP16[$11>>1] = 0;
  HEAP16[$12>>1] = 0;
  $15 = (_luaD_rawrunprotected($3,4295,0)|0);
  $16 = ($15|0)==(0);
  if ($16) {
   break;
  }
 }
 $17 = ((($3)) + 16|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = HEAP32[$4>>2]|0;
 _luaF_close($3,$19);
 _luaC_freeall($3);
 $20 = HEAP32[$17>>2]|0;
 $21 = HEAP32[$20>>2]|0;
 $22 = ((($20)) + 8|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = $23 << 2;
 (_luaM_realloc_($3,$21,$24,0)|0);
 $25 = ((($18)) + 52|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ((($18)) + 60|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = (_luaM_realloc_($3,$26,$28,0)|0);
 HEAP32[$25>>2] = $29;
 HEAP32[$27>>2] = 0;
 $30 = HEAP32[$7>>2]|0;
 $31 = ((($3)) + 48|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ($32*24)|0;
 (_luaM_realloc_($3,$30,$33,0)|0);
 $34 = HEAP32[$4>>2]|0;
 $35 = ((($3)) + 44|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = $36 << 4;
 (_luaM_realloc_($3,$34,$37,0)|0);
 $38 = ((($18)) + 12|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ((($18)) + 16|0);
 $41 = HEAP32[$40>>2]|0;
 (FUNCTION_TABLE_iiiii[$39 & 8191]($41,$3,376,0)|0);
 return;
}
function _callallgcTM($L,$ud) {
 $L = $L|0;
 $ud = $ud|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _luaC_callGCTM($L);
 return;
}
function _luaS_resize($L,$newsize) {
 $L = $L|0;
 $newsize = $newsize|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$13 = 0, $p$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 21|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = ($3<<24>>24)==(2);
 if ($4) {
  return;
 }
 $5 = (($newsize) + 1)|0;
 $6 = ($5>>>0)<(1073741824);
 if ($6) {
  $7 = $newsize << 2;
  $8 = (_luaM_realloc_($L,0,0,$7)|0);
  $13 = $8;
 } else {
  $9 = (_luaM_toobig($L)|0);
  $13 = $9;
 }
 $10 = HEAP32[$0>>2]|0;
 $11 = ($newsize|0)>(0);
 if ($11) {
  $12 = $newsize << 2;
  _memset(($13|0),0,($12|0))|0;
 }
 $14 = ((($10)) + 8|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)>(0);
 if ($16) {
  $17 = (($newsize) + -1)|0;
  $i$13 = 0;
  while(1) {
   $18 = HEAP32[$10>>2]|0;
   $19 = (($18) + ($i$13<<2)|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = ($20|0)==(0|0);
   if (!($21)) {
    $p$02 = $20;
    while(1) {
     $22 = HEAP32[$p$02>>2]|0;
     $23 = ((($p$02)) + 8|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = $24 & $17;
     $26 = (($13) + ($25<<2)|0);
     $27 = HEAP32[$26>>2]|0;
     HEAP32[$p$02>>2] = $27;
     HEAP32[$26>>2] = $p$02;
     $28 = ($22|0)==(0|0);
     if ($28) {
      break;
     } else {
      $p$02 = $22;
     }
    }
   }
   $29 = (($i$13) + 1)|0;
   $30 = ($29|0)<($15|0);
   if ($30) {
    $i$13 = $29;
   } else {
    break;
   }
  }
 }
 $31 = HEAP32[$10>>2]|0;
 $32 = $15 << 2;
 (_luaM_realloc_($L,$31,$32,0)|0);
 HEAP32[$14>>2] = $newsize;
 HEAP32[$10>>2] = $13;
 return;
}
function _luaS_newlstr($L,$str,$l) {
 $L = $L|0;
 $str = $str|0;
 $l = $l|0;
 var $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0, $h$0$lcssa = 0, $h$06 = 0, $l1$07 = 0, $o$0 = 0, $o$04 = 0, $o$05 = 0;
 var $o$05$lcssa = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $l >>> 5;
 $1 = (($0) + 1)|0;
 $2 = ($1>>>0)>($l>>>0);
 if ($2) {
  $h$0$lcssa = $l;
 } else {
  $h$06 = $l;$l1$07 = $l;
  while(1) {
   $3 = $h$06 << 5;
   $4 = $h$06 >>> 2;
   $5 = (($3) + ($4))|0;
   $6 = (($l1$07) + -1)|0;
   $7 = (($str) + ($6)|0);
   $8 = HEAP8[$7>>0]|0;
   $9 = $8&255;
   $10 = (($5) + ($9))|0;
   $11 = $10 ^ $h$06;
   $12 = (($l1$07) - ($1))|0;
   $13 = ($12>>>0)<($1>>>0);
   if ($13) {
    $h$0$lcssa = $11;
    break;
   } else {
    $h$06 = $11;$l1$07 = $12;
   }
  }
 }
 $14 = ((($L)) + 16|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($15)) + 8|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = (($17) + -1)|0;
 $19 = $18 & $h$0$lcssa;
 $20 = HEAP32[$15>>2]|0;
 $21 = (($20) + ($19<<2)|0);
 $o$04 = HEAP32[$21>>2]|0;
 $22 = ($o$04|0)==(0|0);
 L5: do {
  if (!($22)) {
   $o$05 = $o$04;
   while(1) {
    $23 = ((($o$05)) + 12|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = ($24|0)==($l|0);
    if ($25) {
     $26 = ((($o$05)) + 16|0);
     $27 = (_memcmp($str,$26,$l)|0);
     $28 = ($27|0)==(0);
     if ($28) {
      $o$05$lcssa = $o$05;
      break;
     }
    }
    $o$0 = HEAP32[$o$05>>2]|0;
    $41 = ($o$0|0)==(0|0);
    if ($41) {
     break L5;
    } else {
     $o$05 = $o$0;
    }
   }
   $29 = ((($o$05$lcssa)) + 5|0);
   $30 = HEAP8[$29>>0]|0;
   $31 = $30&255;
   $32 = ((($15)) + 20|0);
   $33 = HEAP8[$32>>0]|0;
   $34 = $33&255;
   $35 = $34 ^ 3;
   $36 = $31 & 3;
   $37 = $36 & $35;
   $38 = ($37|0)==(0);
   if ($38) {
    $$2 = $o$05$lcssa;
    return ($$2|0);
   }
   $39 = $31 ^ 3;
   $40 = $39&255;
   HEAP8[$29>>0] = $40;
   $$2 = $o$05$lcssa;
   return ($$2|0);
  }
 } while(0);
 $42 = (($l) + 1)|0;
 $43 = ($42>>>0)>(4294967277);
 if ($43) {
  (_luaM_toobig($L)|0);
 }
 $44 = (($l) + 17)|0;
 $45 = (_luaM_realloc_($L,0,0,$44)|0);
 $46 = ((($45)) + 12|0);
 HEAP32[$46>>2] = $l;
 $47 = ((($45)) + 8|0);
 HEAP32[$47>>2] = $h$0$lcssa;
 $48 = HEAP32[$14>>2]|0;
 $49 = ((($48)) + 20|0);
 $50 = HEAP8[$49>>0]|0;
 $51 = $50&255;
 $52 = $51 & 3;
 $53 = $52&255;
 $54 = ((($45)) + 5|0);
 HEAP8[$54>>0] = $53;
 $55 = ((($45)) + 4|0);
 HEAP8[$55>>0] = 4;
 $56 = ((($45)) + 6|0);
 HEAP8[$56>>0] = 0;
 $57 = ((($45)) + 16|0);
 _memcpy(($57|0),($str|0),($l|0))|0;
 $58 = (($57) + ($l)|0);
 HEAP8[$58>>0] = 0;
 $59 = HEAP32[$14>>2]|0;
 $60 = ((($59)) + 8|0);
 $61 = HEAP32[$60>>2]|0;
 $62 = (($61) + -1)|0;
 $63 = $62 & $h$0$lcssa;
 $64 = HEAP32[$59>>2]|0;
 $65 = (($64) + ($63<<2)|0);
 $66 = HEAP32[$65>>2]|0;
 HEAP32[$45>>2] = $66;
 HEAP32[$65>>2] = $45;
 $67 = ((($59)) + 4|0);
 $68 = HEAP32[$67>>2]|0;
 $69 = (($68) + 1)|0;
 HEAP32[$67>>2] = $69;
 $70 = ($69>>>0)>($61>>>0);
 $71 = ($61|0)<(1073741823);
 $or$cond$i = $71 & $70;
 if (!($or$cond$i)) {
  $$2 = $45;
  return ($$2|0);
 }
 $72 = $61 << 1;
 _luaS_resize($L,$72);
 $$2 = $45;
 return ($$2|0);
}
function _luaS_newudata($L,$s,$e) {
 $L = $L|0;
 $s = $s|0;
 $e = $e|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($s>>>0)>(4294967269);
 if ($0) {
  (_luaM_toobig($L)|0);
 }
 $1 = (($s) + 24)|0;
 $2 = (_luaM_realloc_($L,0,0,$1)|0);
 $3 = ((($L)) + 16|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 20|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = $6&255;
 $8 = $7 & 3;
 $9 = $8&255;
 $10 = ((($2)) + 5|0);
 HEAP8[$10>>0] = $9;
 $11 = ((($2)) + 4|0);
 HEAP8[$11>>0] = 7;
 $12 = ((($2)) + 16|0);
 HEAP32[$12>>2] = $s;
 $13 = ((($2)) + 8|0);
 HEAP32[$13>>2] = 0;
 $14 = ((($2)) + 12|0);
 HEAP32[$14>>2] = $e;
 $15 = ((($4)) + 112|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = HEAP32[$16>>2]|0;
 HEAP32[$2>>2] = $17;
 HEAP32[$16>>2] = $2;
 return ($2|0);
}
function _luaH_next($L,$t,$key) {
 $L = $L|0;
 $t = $t|0;
 $key = $key|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa32 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0.0, $i$0 = 0;
 var $i$0$in = 0, $i$0$in$lcssa30 = 0, $i$0$in$ph = 0, $i$0$lcssa = 0, $i$0$lcssa31 = 0, $i$19 = 0, $i$19$lcssa = 0, $n$0$i = 0, $n$0$i$lcssa = 0, $n$0$ph$i = 0, $not$$i$i = 0, $or$cond$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($key)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 L1: do {
  switch ($1|0) {
  case 0:  {
   $i$0$in$ph = -1;
   break;
  }
  case 3:  {
   $2 = +HEAPF64[$key>>3];
   $3 = (~~(($2)));
   $4 = (+($3|0));
   $not$$i$i = $4 == $2;
   $5 = ($3|0)>(0);
   $or$cond$i = $5 & $not$$i$i;
   if ($or$cond$i) {
    $6 = ((($t)) + 28|0);
    $7 = HEAP32[$6>>2]|0;
    $8 = ($3|0)>($7|0);
    if (!($8)) {
     $9 = (($3) + -1)|0;
     $i$0$in$ph = $9;
     break L1;
    }
   }
   $10 = $2 == 0.0;
   $11 = ((($t)) + 16|0);
   $12 = HEAP32[$11>>2]|0;
   if ($10) {
    $n$0$ph$i = $12;
    label = 11;
   } else {
    HEAPF64[tempDoublePtr>>3] = $2;$13 = HEAP32[tempDoublePtr>>2]|0;
    $14 = HEAP32[tempDoublePtr+4>>2]|0;
    $15 = (($14) + ($13))|0;
    $16 = ((($t)) + 7|0);
    $17 = HEAP8[$16>>0]|0;
    $18 = $17&255;
    $19 = 1 << $18;
    $20 = (($19) + -1)|0;
    $21 = $20 | 1;
    $22 = (($15>>>0) % ($21>>>0))&-1;
    $23 = (($12) + ($22<<5)|0);
    $n$0$ph$i = $23;
    label = 11;
   }
   break;
  }
  case 2:  {
   $46 = HEAP32[$key>>2]|0;
   $47 = ((($t)) + 7|0);
   $48 = HEAP8[$47>>0]|0;
   $49 = $48&255;
   $50 = 1 << $49;
   $51 = (($50) + -1)|0;
   $52 = $51 | 1;
   $53 = (($46>>>0) % ($52>>>0))&-1;
   $54 = ((($t)) + 16|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = (($55) + ($53<<5)|0);
   $n$0$ph$i = $56;
   label = 11;
   break;
  }
  case 4:  {
   $24 = HEAP32[$key>>2]|0;
   $25 = ((($24)) + 8|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = ((($t)) + 7|0);
   $28 = HEAP8[$27>>0]|0;
   $29 = $28&255;
   $30 = 1 << $29;
   $31 = (($30) + -1)|0;
   $32 = $31 & $26;
   $33 = ((($t)) + 16|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = (($34) + ($32<<5)|0);
   $n$0$ph$i = $35;
   label = 11;
   break;
  }
  case 1:  {
   $36 = HEAP32[$key>>2]|0;
   $37 = ((($t)) + 7|0);
   $38 = HEAP8[$37>>0]|0;
   $39 = $38&255;
   $40 = 1 << $39;
   $41 = (($40) + -1)|0;
   $42 = $41 & $36;
   $43 = ((($t)) + 16|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = (($44) + ($42<<5)|0);
   $n$0$ph$i = $45;
   label = 11;
   break;
  }
  default: {
   $57 = HEAP32[$key>>2]|0;
   $58 = ((($t)) + 7|0);
   $59 = HEAP8[$58>>0]|0;
   $60 = $59&255;
   $61 = 1 << $60;
   $62 = (($61) + -1)|0;
   $63 = $62 | 1;
   $64 = (($57>>>0) % ($63>>>0))&-1;
   $65 = ((($t)) + 16|0);
   $66 = HEAP32[$65>>2]|0;
   $67 = (($66) + ($64<<5)|0);
   $n$0$ph$i = $67;
   label = 11;
  }
  }
 } while(0);
 do {
  if ((label|0) == 11) {
   $n$0$i = $n$0$ph$i;
   while(1) {
    $68 = ((($n$0$i)) + 16|0);
    $69 = (_luaO_rawequalObj($68,$key)|0);
    $70 = ($69|0)==(0);
    if (!($70)) {
     $n$0$i$lcssa = $n$0$i;
     label = 16;
     break;
    }
    $71 = ((($n$0$i)) + 24|0);
    $72 = HEAP32[$71>>2]|0;
    $73 = ($72|0)==(11);
    if ($73) {
     $74 = HEAP32[$0>>2]|0;
     $75 = ($74|0)>(3);
     if ($75) {
      $76 = HEAP32[$68>>2]|0;
      $77 = HEAP32[$key>>2]|0;
      $78 = ($76|0)==($77|0);
      if ($78) {
       $n$0$i$lcssa = $n$0$i;
       label = 16;
       break;
      }
     }
    }
    $87 = ((($n$0$i)) + 28|0);
    $88 = HEAP32[$87>>2]|0;
    $89 = ($88|0)==(0|0);
    if ($89) {
     label = 18;
     break;
    } else {
     $n$0$i = $88;
    }
   }
   if ((label|0) == 16) {
    $79 = ((($t)) + 16|0);
    $80 = HEAP32[$79>>2]|0;
    $81 = $n$0$i$lcssa;
    $82 = (($81) - ($80))|0;
    $83 = $82 >> 5;
    $84 = ((($t)) + 28|0);
    $85 = HEAP32[$84>>2]|0;
    $86 = (($83) + ($85))|0;
    $i$0$in$ph = $86;
    break;
   }
   else if ((label|0) == 18) {
    _luaG_runerror($L,518892,$vararg_buffer);
    $i$0$in$ph = 0;
    break;
   }
  }
 } while(0);
 $90 = ((($t)) + 28|0);
 $91 = HEAP32[$90>>2]|0;
 $92 = ((($t)) + 12|0);
 $i$0$in = $i$0$in$ph;
 while(1) {
  $i$0 = (($i$0$in) + 1)|0;
  $93 = ($i$0|0)<($91|0);
  if (!($93)) {
   $i$0$lcssa = $i$0;
   break;
  }
  $94 = HEAP32[$92>>2]|0;
  $95 = (((($94) + ($i$0<<4)|0)) + 8|0);
  $96 = HEAP32[$95>>2]|0;
  $97 = ($96|0)==(0);
  if ($97) {
   $i$0$in = $i$0;
  } else {
   $$lcssa = $94;$$lcssa32 = $95;$i$0$in$lcssa30 = $i$0$in;$i$0$lcssa31 = $i$0;
   label = 22;
   break;
  }
 }
 if ((label|0) == 22) {
  $98 = (($i$0$in$lcssa30) + 2)|0;
  $99 = (+($98|0));
  HEAPF64[$key>>3] = $99;
  HEAP32[$0>>2] = 3;
  $100 = (($$lcssa) + ($i$0$lcssa31<<4)|0);
  $101 = ((($key)) + 16|0);
  $102 = $100;
  $103 = $102;
  $104 = HEAP32[$103>>2]|0;
  $105 = (($102) + 4)|0;
  $106 = $105;
  $107 = HEAP32[$106>>2]|0;
  $108 = $101;
  $109 = $108;
  HEAP32[$109>>2] = $104;
  $110 = (($108) + 4)|0;
  $111 = $110;
  HEAP32[$111>>2] = $107;
  $112 = HEAP32[$$lcssa32>>2]|0;
  $113 = ((($key)) + 24|0);
  HEAP32[$113>>2] = $112;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 $114 = (($i$0$lcssa) - ($91))|0;
 $115 = ((($t)) + 7|0);
 $116 = HEAP8[$115>>0]|0;
 $117 = $116&255;
 $118 = 1 << $117;
 $119 = ($114|0)<($118|0);
 if (!($119)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $120 = ((($t)) + 16|0);
 $121 = HEAP32[$120>>2]|0;
 $i$19 = $114;
 while(1) {
  $122 = (((($121) + ($i$19<<5)|0)) + 8|0);
  $123 = HEAP32[$122>>2]|0;
  $124 = ($123|0)==(0);
  if (!($124)) {
   $i$19$lcssa = $i$19;
   break;
  }
  $154 = (($i$19) + 1)|0;
  $155 = ($154|0)<($118|0);
  if ($155) {
   $i$19 = $154;
  } else {
   $$0 = 0;
   label = 28;
   break;
  }
 }
 if ((label|0) == 28) {
  STACKTOP = sp;return ($$0|0);
 }
 $125 = (((($121) + ($i$19$lcssa<<5)|0)) + 16|0);
 $126 = $125;
 $127 = $126;
 $128 = HEAP32[$127>>2]|0;
 $129 = (($126) + 4)|0;
 $130 = $129;
 $131 = HEAP32[$130>>2]|0;
 $132 = $key;
 $133 = $132;
 HEAP32[$133>>2] = $128;
 $134 = (($132) + 4)|0;
 $135 = $134;
 HEAP32[$135>>2] = $131;
 $136 = (((($121) + ($i$19$lcssa<<5)|0)) + 24|0);
 $137 = HEAP32[$136>>2]|0;
 HEAP32[$0>>2] = $137;
 $138 = HEAP32[$120>>2]|0;
 $139 = (($138) + ($i$19$lcssa<<5)|0);
 $140 = ((($key)) + 16|0);
 $141 = $139;
 $142 = $141;
 $143 = HEAP32[$142>>2]|0;
 $144 = (($141) + 4)|0;
 $145 = $144;
 $146 = HEAP32[$145>>2]|0;
 $147 = $140;
 $148 = $147;
 HEAP32[$148>>2] = $143;
 $149 = (($147) + 4)|0;
 $150 = $149;
 HEAP32[$150>>2] = $146;
 $151 = (((($138) + ($i$19$lcssa<<5)|0)) + 8|0);
 $152 = HEAP32[$151>>2]|0;
 $153 = ((($key)) + 24|0);
 HEAP32[$153>>2] = $152;
 $$0 = 1;
 STACKTOP = sp;return ($$0|0);
}
function _luaH_resizearray($L,$t,$nasize) {
 $L = $L|0;
 $t = $t|0;
 $nasize = $nasize|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($t)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(624112|0);
 if ($2) {
  $7 = 0;
 } else {
  $3 = ((($t)) + 7|0);
  $4 = HEAP8[$3>>0]|0;
  $5 = $4&255;
  $6 = 1 << $5;
  $7 = $6;
 }
 _resize($L,$t,$nasize,$7);
 return;
}
function _resize($L,$t,$nasize,$nhsize) {
 $L = $L|0;
 $t = $t|0;
 $nasize = $nasize|0;
 $nhsize = $nhsize|0;
 var $$0$i = 0, $$0$i1 = 0, $$1$i$i = 0, $$pre = 0, $$pre$i = 0, $$pre$phi12Z2D = 0, $$pre$phi17Z2D = 0.0, $$pre$phi3$iZ2D = 0, $$pre14 = 0, $$pre16 = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0.0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0.0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $exitcond$i = 0, $i$01$i = 0;
 var $i$05 = 0, $i$14 = 0, $i$14$in = 0, $k$i = 0, $n$0$i$i = 0, $n$0$i$i$lcssa = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 24|0;
 $vararg_buffer = sp + 16|0;
 $k$i = sp;
 $0 = ((($t)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($t)) + 7|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = ((($t)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($1|0)<($nasize|0);
 if ($7) {
  $8 = (($nasize) + 1)|0;
  $9 = ($8>>>0)<(268435456);
  if ($9) {
   $10 = ((($t)) + 12|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = $1 << 4;
   $13 = $nasize << 4;
   $14 = (_luaM_realloc_($L,$11,$12,$13)|0);
   $$pre$phi3$iZ2D = $10;$16 = $14;
  } else {
   $15 = (_luaM_toobig($L)|0);
   $$pre$i = ((($t)) + 12|0);
   $$pre$phi3$iZ2D = $$pre$i;$16 = $15;
  }
  HEAP32[$$pre$phi3$iZ2D>>2] = $16;
  $17 = HEAP32[$0>>2]|0;
  $18 = ($17|0)<($nasize|0);
  if ($18) {
   $i$01$i = $17;
   while(1) {
    $19 = (((($16) + ($i$01$i<<4)|0)) + 8|0);
    HEAP32[$19>>2] = 0;
    $20 = (($i$01$i) + 1)|0;
    $exitcond$i = ($20|0)==($nasize|0);
    if ($exitcond$i) {
     break;
    } else {
     $i$01$i = $20;
    }
   }
  }
  HEAP32[$0>>2] = $nasize;
 }
 _setnodevector($L,$t,$nhsize);
 $21 = ($1|0)>($nasize|0);
 if ($21) {
  HEAP32[$0>>2] = $nasize;
  $22 = ((($t)) + 12|0);
  $23 = ((($k$i)) + 8|0);
  $i$05 = $nasize;
  while(1) {
   $24 = HEAP32[$22>>2]|0;
   $25 = (((($24) + ($i$05<<4)|0)) + 8|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = ($26|0)==(0);
   $$pre14 = (($i$05) + 1)|0;
   if (!($27)) {
    $28 = HEAP32[$0>>2]|0;
    $29 = ($i$05>>>0)<($28>>>0);
    L18: do {
     if ($29) {
      $30 = (($24) + ($i$05<<4)|0);
      $$1$i$i = $30;
      label = 19;
     } else {
      $31 = (+($$pre14|0));
      $32 = ($$pre14|0)==(0);
      $33 = HEAP32[$5>>2]|0;
      if ($32) {
       $n$0$i$i = $33;
      } else {
       HEAPF64[tempDoublePtr>>3] = $31;$34 = HEAP32[tempDoublePtr>>2]|0;
       $35 = HEAP32[tempDoublePtr+4>>2]|0;
       $36 = (($35) + ($34))|0;
       $37 = HEAP8[$2>>0]|0;
       $38 = $37&255;
       $39 = 1 << $38;
       $40 = (($39) + -1)|0;
       $41 = $40 | 1;
       $42 = (($36>>>0) % ($41>>>0))&-1;
       $43 = (($33) + ($42<<5)|0);
       $n$0$i$i = $43;
      }
      while(1) {
       $44 = ((($n$0$i$i)) + 24|0);
       $45 = HEAP32[$44>>2]|0;
       $46 = ($45|0)==(3);
       if ($46) {
        $47 = ((($n$0$i$i)) + 16|0);
        $48 = +HEAPF64[$47>>3];
        $49 = $48 == $31;
        if ($49) {
         $n$0$i$i$lcssa = $n$0$i$i;
         break;
        }
       }
       $50 = ((($n$0$i$i)) + 28|0);
       $51 = HEAP32[$50>>2]|0;
       $52 = ($51|0)==(0|0);
       if ($52) {
        $$pre$phi17Z2D = $31;
        label = 21;
        break L18;
       } else {
        $n$0$i$i = $51;
       }
      }
      $$1$i$i = $n$0$i$i$lcssa;
      label = 19;
     }
    } while(0);
    if ((label|0) == 19) {
     label = 0;
     $53 = ($$1$i$i|0)==(624096|0);
     if ($53) {
      $$pre16 = (+($$pre14|0));
      $$pre$phi17Z2D = $$pre16;
      label = 21;
     } else {
      $$0$i = $$1$i$i;
     }
    }
    if ((label|0) == 21) {
     label = 0;
     HEAPF64[$k$i>>3] = $$pre$phi17Z2D;
     HEAP32[$23>>2] = 3;
     $54 = (_newkey($L,$t,$k$i)|0);
     $$0$i = $54;
    }
    $55 = (($24) + ($i$05<<4)|0);
    $56 = $55;
    $57 = $56;
    $58 = HEAP32[$57>>2]|0;
    $59 = (($56) + 4)|0;
    $60 = $59;
    $61 = HEAP32[$60>>2]|0;
    $62 = $$0$i;
    $63 = $62;
    HEAP32[$63>>2] = $58;
    $64 = (($62) + 4)|0;
    $65 = $64;
    HEAP32[$65>>2] = $61;
    $66 = HEAP32[$25>>2]|0;
    $67 = ((($$0$i)) + 8|0);
    HEAP32[$67>>2] = $66;
   }
   $exitcond = ($$pre14|0)==($1|0);
   if ($exitcond) {
    break;
   } else {
    $i$05 = $$pre14;
   }
  }
  $68 = (($nasize) + 1)|0;
  $69 = ($68>>>0)<(268435456);
  if ($69) {
   $70 = ((($t)) + 12|0);
   $71 = HEAP32[$70>>2]|0;
   $72 = $1 << 4;
   $73 = $nasize << 4;
   $74 = (_luaM_realloc_($L,$71,$72,$73)|0);
   $$pre$phi12Z2D = $70;$76 = $74;
  } else {
   $75 = (_luaM_toobig($L)|0);
   $$pre = ((($t)) + 12|0);
   $$pre$phi12Z2D = $$pre;$76 = $75;
  }
  HEAP32[$$pre$phi12Z2D>>2] = $76;
 }
 $77 = 1 << $4;
 $78 = ($3<<24>>24)==(31);
 if (!($78)) {
  $79 = ((($t)) + 6|0);
  $i$14$in = $77;
  while(1) {
   $i$14 = (($i$14$in) + -1)|0;
   $80 = (((($6) + ($i$14<<5)|0)) + 8|0);
   $81 = HEAP32[$80>>2]|0;
   $82 = ($81|0)==(0);
   if (!($82)) {
    $83 = (($6) + ($i$14<<5)|0);
    $84 = (((($6) + ($i$14<<5)|0)) + 16|0);
    $85 = (_luaH_get($t,$84)|0);
    HEAP8[$79>>0] = 0;
    $86 = ($85|0)==(624096|0);
    if ($86) {
     $87 = (((($6) + ($i$14<<5)|0)) + 24|0);
     $88 = HEAP32[$87>>2]|0;
     switch ($88|0) {
     case 0:  {
      _luaG_runerror($L,518929,$vararg_buffer);
      break;
     }
     case 3:  {
      $89 = +HEAPF64[$84>>3];
      $90 = ($89 == $89) & (0.0 == 0.0);
      if (!($90)) {
       _luaG_runerror($L,518948,$vararg_buffer1);
      }
      break;
     }
     default: {
     }
     }
     $91 = (_newkey($L,$t,$84)|0);
     $$0$i1 = $91;
    } else {
     $$0$i1 = $85;
    }
    $92 = $83;
    $93 = $92;
    $94 = HEAP32[$93>>2]|0;
    $95 = (($92) + 4)|0;
    $96 = $95;
    $97 = HEAP32[$96>>2]|0;
    $98 = $$0$i1;
    $99 = $98;
    HEAP32[$99>>2] = $94;
    $100 = (($98) + 4)|0;
    $101 = $100;
    HEAP32[$101>>2] = $97;
    $102 = HEAP32[$80>>2]|0;
    $103 = ((($$0$i1)) + 8|0);
    HEAP32[$103>>2] = $102;
   }
   $104 = ($i$14$in|0)>(1);
   if ($104) {
    $i$14$in = $i$14;
   } else {
    break;
   }
  }
 }
 $105 = ($6|0)==(624112|0);
 if ($105) {
  STACKTOP = sp;return;
 }
 $106 = $77 << 5;
 (_luaM_realloc_($L,$6,$106,0)|0);
 STACKTOP = sp;return;
}
function _setnodevector($L,$t,$size) {
 $L = $L|0;
 $t = $t|0;
 $size = $size|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $lsize$0 = 0, $phitmp = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($size|0)==(0);
 if ($0) {
  $1 = ((($t)) + 16|0);
  HEAP32[$1>>2] = 624112;
  $$0 = 0;$22 = 624112;$lsize$0 = 0;
  $20 = ((($t)) + 7|0);
  HEAP8[$20>>0] = $lsize$0;
  $21 = (($22) + ($$0<<5)|0);
  $23 = ((($t)) + 20|0);
  HEAP32[$23>>2] = $21;
  STACKTOP = sp;return;
 }
 $2 = (($size) + -1)|0;
 $3 = (_luaO_log2($2)|0);
 $4 = (($3) + 1)|0;
 $5 = ($3|0)>(25);
 if ($5) {
  _luaG_runerror($L,518914,$vararg_buffer);
 }
 $6 = 1 << $4;
 $7 = (($6) + 1)|0;
 $8 = ($7>>>0)<(134217728);
 if ($8) {
  $9 = $6 << 5;
  $10 = (_luaM_realloc_($L,0,0,$9)|0);
  $13 = $10;
 } else {
  $11 = (_luaM_toobig($L)|0);
  $13 = $11;
 }
 $12 = ((($t)) + 16|0);
 HEAP32[$12>>2] = $13;
 $14 = ($4|0)==(31);
 if (!($14)) {
  $i$01 = 0;
  while(1) {
   $15 = (((($13) + ($i$01<<5)|0)) + 28|0);
   HEAP32[$15>>2] = 0;
   $16 = (((($13) + ($i$01<<5)|0)) + 24|0);
   HEAP32[$16>>2] = 0;
   $17 = (((($13) + ($i$01<<5)|0)) + 8|0);
   HEAP32[$17>>2] = 0;
   $18 = (($i$01) + 1)|0;
   $19 = ($18|0)<($6|0);
   if ($19) {
    $i$01 = $18;
   } else {
    break;
   }
  }
 }
 $phitmp = $4&255;
 $$0 = $6;$22 = $13;$lsize$0 = $phitmp;
 $20 = ((($t)) + 7|0);
 HEAP8[$20>>0] = $lsize$0;
 $21 = (($22) + ($$0<<5)|0);
 $23 = ((($t)) + 20|0);
 HEAP32[$23>>2] = $21;
 STACKTOP = sp;return;
}
function _newkey($L,$t,$key) {
 $L = $L|0;
 $t = $t|0;
 $key = $key|0;
 var $$0$i = 0, $$0$i$i = 0, $$0$i$i$i = 0, $$0$i4 = 0, $$1 = 0, $$lcssa = 0, $$lcssa54 = 0, $$lcssa55 = 0, $$lcssa58 = 0, $$lcssa59 = 0, $$lcssa60 = 0, $$na$0$i$i = 0, $$off$i$i = 0, $$off$i$i$i = 0, $$promoted$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0.0, $109 = 0, $11 = 0, $110 = 0.0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0.0, $123 = 0, $124 = 0.0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $138$phi = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0.0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0.0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0.0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $a$03$i$i = 0, $a$1$i$i = 0, $ause$0$lcssa$i$i = 0, $ause$0$ph$i$i = 0, $ause$0$ph$i$i$lcssa = 0, $ause$08$i$i = 0, $i$0$i$i = 0, $i$0$ph$i$i = 0;
 var $i$01$i$i = 0, $i$09$i$i = 0, $i$1$lcssa$i$i = 0, $i$13$i$i = 0, $lc$0$$i$i = 0, $lc$0$lcssa$i$i = 0, $lc$04$i$i = 0, $lg$06$i$i = 0, $lim$0$ph$i$i = 0, $mp$2 = 0, $n$05$i$i = 0, $n$1$i$i = 0, $n$2$i$i = 0, $na$04$i$i = 0, $na$1$i$i = 0, $na$2$i$i = 0, $not$$i$i = 0, $not$$i$i$i = 0, $not$$i$i$i$i = 0, $nums$i = 0;
 var $or$cond = 0, $or$cond$i$i = 0, $or$cond$i$i$i = 0, $othern$0 = 0, $totaluse$0$ph$i$i = 0, $totaluse$0$ph$i$i$lcssa = 0, $ttlg$07$i$i = 0, $twotoi$0$n$0$i$i = 0, $twotoi$02$i$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $nums$i = sp + 12|0;
 $0 = ((($key)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 switch ($1|0) {
 case 3:  {
  $2 = +HEAPF64[$key>>3];
  $3 = $2 == 0.0;
  $4 = ((($t)) + 16|0);
  $5 = HEAP32[$4>>2]|0;
  if ($3) {
   $$0$i = $5;$70 = $5;
  } else {
   HEAPF64[tempDoublePtr>>3] = $2;$6 = HEAP32[tempDoublePtr>>2]|0;
   $7 = HEAP32[tempDoublePtr+4>>2]|0;
   $8 = (($7) + ($6))|0;
   $9 = ((($t)) + 7|0);
   $10 = HEAP8[$9>>0]|0;
   $11 = $10&255;
   $12 = 1 << $11;
   $13 = (($12) + -1)|0;
   $14 = $13 | 1;
   $15 = (($8>>>0) % ($14>>>0))&-1;
   $16 = (($5) + ($15<<5)|0);
   $$0$i = $16;$70 = $5;
  }
  break;
 }
 case 4:  {
  $17 = HEAP32[$key>>2]|0;
  $18 = ((($17)) + 8|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = ((($t)) + 7|0);
  $21 = HEAP8[$20>>0]|0;
  $22 = $21&255;
  $23 = 1 << $22;
  $24 = (($23) + -1)|0;
  $25 = $24 & $19;
  $26 = ((($t)) + 16|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = (($27) + ($25<<5)|0);
  $$0$i = $28;$70 = $27;
  break;
 }
 case 1:  {
  $29 = HEAP32[$key>>2]|0;
  $30 = ((($t)) + 7|0);
  $31 = HEAP8[$30>>0]|0;
  $32 = $31&255;
  $33 = 1 << $32;
  $34 = (($33) + -1)|0;
  $35 = $34 & $29;
  $36 = ((($t)) + 16|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = (($37) + ($35<<5)|0);
  $$0$i = $38;$70 = $37;
  break;
 }
 case 2:  {
  $39 = HEAP32[$key>>2]|0;
  $40 = ((($t)) + 7|0);
  $41 = HEAP8[$40>>0]|0;
  $42 = $41&255;
  $43 = 1 << $42;
  $44 = (($43) + -1)|0;
  $45 = $44 | 1;
  $46 = (($39>>>0) % ($45>>>0))&-1;
  $47 = ((($t)) + 16|0);
  $48 = HEAP32[$47>>2]|0;
  $49 = (($48) + ($46<<5)|0);
  $$0$i = $49;$70 = $48;
  break;
 }
 default: {
  $50 = HEAP32[$key>>2]|0;
  $51 = ((($t)) + 7|0);
  $52 = HEAP8[$51>>0]|0;
  $53 = $52&255;
  $54 = 1 << $53;
  $55 = (($54) + -1)|0;
  $56 = $55 | 1;
  $57 = (($50>>>0) % ($56>>>0))&-1;
  $58 = ((($t)) + 16|0);
  $59 = HEAP32[$58>>2]|0;
  $60 = (($59) + ($57<<5)|0);
  $$0$i = $60;$70 = $59;
 }
 }
 $61 = ((($$0$i)) + 8|0);
 $62 = HEAP32[$61>>2]|0;
 $63 = ($62|0)!=(0);
 $64 = ($$0$i|0)==(624112|0);
 $or$cond = $64 | $63;
 do {
  if ($or$cond) {
   $65 = ((($t)) + 20|0);
   $66 = ((($t)) + 16|0);
   $$promoted$i = HEAP32[$65>>2]|0;
   $68 = $$promoted$i;
   while(1) {
    $67 = ((($68)) + -32|0);
    $69 = ($68>>>0)>($70>>>0);
    if (!($69)) {
     $$lcssa59 = $67;
     break;
    }
    $71 = ((($68)) + -8|0);
    $72 = HEAP32[$71>>2]|0;
    $73 = ($72|0)==(0);
    if ($73) {
     $$lcssa58 = $68;$$lcssa60 = $67;
     label = 41;
     break;
    } else {
     $68 = $67;
    }
   }
   if ((label|0) == 41) {
    HEAP32[$65>>2] = $$lcssa60;
    $153 = ((($$0$i)) + 16|0);
    $154 = ((($$0$i)) + 24|0);
    $155 = HEAP32[$154>>2]|0;
    switch ($155|0) {
    case 3:  {
     $156 = +HEAPF64[$153>>3];
     $157 = $156 == 0.0;
     if ($157) {
      $$0$i4 = $70;
     } else {
      HEAPF64[tempDoublePtr>>3] = $156;$158 = HEAP32[tempDoublePtr>>2]|0;
      $159 = HEAP32[tempDoublePtr+4>>2]|0;
      $160 = (($159) + ($158))|0;
      $161 = ((($t)) + 7|0);
      $162 = HEAP8[$161>>0]|0;
      $163 = $162&255;
      $164 = 1 << $163;
      $165 = (($164) + -1)|0;
      $166 = $165 | 1;
      $167 = (($160>>>0) % ($166>>>0))&-1;
      $168 = (($70) + ($167<<5)|0);
      $$0$i4 = $168;
     }
     break;
    }
    case 4:  {
     $169 = HEAP32[$153>>2]|0;
     $170 = ((($169)) + 8|0);
     $171 = HEAP32[$170>>2]|0;
     $172 = ((($t)) + 7|0);
     $173 = HEAP8[$172>>0]|0;
     $174 = $173&255;
     $175 = 1 << $174;
     $176 = (($175) + -1)|0;
     $177 = $176 & $171;
     $178 = (($70) + ($177<<5)|0);
     $$0$i4 = $178;
     break;
    }
    case 1:  {
     $179 = HEAP32[$153>>2]|0;
     $180 = ((($t)) + 7|0);
     $181 = HEAP8[$180>>0]|0;
     $182 = $181&255;
     $183 = 1 << $182;
     $184 = (($183) + -1)|0;
     $185 = $184 & $179;
     $186 = (($70) + ($185<<5)|0);
     $$0$i4 = $186;
     break;
    }
    case 2:  {
     $187 = HEAP32[$153>>2]|0;
     $188 = ((($t)) + 7|0);
     $189 = HEAP8[$188>>0]|0;
     $190 = $189&255;
     $191 = 1 << $190;
     $192 = (($191) + -1)|0;
     $193 = $192 | 1;
     $194 = (($187>>>0) % ($193>>>0))&-1;
     $195 = (($70) + ($194<<5)|0);
     $$0$i4 = $195;
     break;
    }
    default: {
     $196 = HEAP32[$153>>2]|0;
     $197 = ((($t)) + 7|0);
     $198 = HEAP8[$197>>0]|0;
     $199 = $198&255;
     $200 = 1 << $199;
     $201 = (($200) + -1)|0;
     $202 = $201 | 1;
     $203 = (($196>>>0) % ($202>>>0))&-1;
     $204 = (($70) + ($203<<5)|0);
     $$0$i4 = $204;
    }
    }
    $205 = ($$0$i4|0)==($$0$i|0);
    if ($205) {
     $210 = ((($$0$i)) + 28|0);
     $211 = HEAP32[$210>>2]|0;
     $212 = ((($$lcssa58)) + -4|0);
     HEAP32[$212>>2] = $211;
     HEAP32[$210>>2] = $$lcssa60;
     $mp$2 = $$lcssa60;
     break;
    } else {
     $othern$0 = $$0$i4;
    }
    while(1) {
     $206 = ((($othern$0)) + 28|0);
     $207 = HEAP32[$206>>2]|0;
     $208 = ($207|0)==($$0$i|0);
     if ($208) {
      $$lcssa = $206;
      break;
     } else {
      $othern$0 = $207;
     }
    }
    HEAP32[$$lcssa>>2] = $$lcssa60;
    ;HEAP32[$$lcssa60>>2]=HEAP32[$$0$i>>2]|0;HEAP32[$$lcssa60+4>>2]=HEAP32[$$0$i+4>>2]|0;HEAP32[$$lcssa60+8>>2]=HEAP32[$$0$i+8>>2]|0;HEAP32[$$lcssa60+12>>2]=HEAP32[$$0$i+12>>2]|0;HEAP32[$$lcssa60+16>>2]=HEAP32[$$0$i+16>>2]|0;HEAP32[$$lcssa60+20>>2]=HEAP32[$$0$i+20>>2]|0;HEAP32[$$lcssa60+24>>2]=HEAP32[$$0$i+24>>2]|0;HEAP32[$$lcssa60+28>>2]=HEAP32[$$0$i+28>>2]|0;
    $209 = ((($$0$i)) + 28|0);
    HEAP32[$209>>2] = 0;
    HEAP32[$61>>2] = 0;
    $mp$2 = $$0$i;
    break;
   }
   HEAP32[$65>>2] = $$lcssa59;
   dest=$nums$i; stop=dest+108|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
   $74 = ((($t)) + 28|0);
   $75 = ((($t)) + 12|0);
   $76 = HEAP32[$74>>2]|0;
   $ause$08$i$i = 0;$i$09$i$i = 1;$lg$06$i$i = 0;$ttlg$07$i$i = 1;
   while(1) {
    $77 = ($ttlg$07$i$i|0)>($76|0);
    if ($77) {
     $78 = ($i$09$i$i|0)>($76|0);
     if ($78) {
      $ause$0$lcssa$i$i = $ause$08$i$i;
      break;
     } else {
      $lim$0$ph$i$i = $76;
     }
    } else {
     $lim$0$ph$i$i = $ttlg$07$i$i;
    }
    $79 = ($i$09$i$i|0)>($lim$0$ph$i$i|0);
    if ($79) {
     $i$1$lcssa$i$i = $i$09$i$i;$lc$0$lcssa$i$i = 0;
    } else {
     $80 = HEAP32[$75>>2]|0;
     $i$13$i$i = $i$09$i$i;$lc$04$i$i = 0;
     while(1) {
      $81 = (($i$13$i$i) + -1)|0;
      $82 = (((($80) + ($81<<4)|0)) + 8|0);
      $83 = HEAP32[$82>>2]|0;
      $not$$i$i = ($83|0)!=(0);
      $84 = $not$$i$i&1;
      $lc$0$$i$i = (($84) + ($lc$04$i$i))|0;
      $85 = (($i$13$i$i) + 1)|0;
      $86 = ($i$13$i$i|0)<($lim$0$ph$i$i|0);
      if ($86) {
       $i$13$i$i = $85;$lc$04$i$i = $lc$0$$i$i;
      } else {
       $i$1$lcssa$i$i = $85;$lc$0$lcssa$i$i = $lc$0$$i$i;
       break;
      }
     }
    }
    $87 = (($nums$i) + ($lg$06$i$i<<2)|0);
    $88 = HEAP32[$87>>2]|0;
    $89 = (($88) + ($lc$0$lcssa$i$i))|0;
    HEAP32[$87>>2] = $89;
    $90 = (($lc$0$lcssa$i$i) + ($ause$08$i$i))|0;
    $91 = (($lg$06$i$i) + 1)|0;
    $92 = $ttlg$07$i$i << 1;
    $93 = ($91|0)<(27);
    if ($93) {
     $ause$08$i$i = $90;$i$09$i$i = $i$1$lcssa$i$i;$lg$06$i$i = $91;$ttlg$07$i$i = $92;
    } else {
     $ause$0$lcssa$i$i = $90;
     break;
    }
   }
   $94 = ((($t)) + 7|0);
   $95 = HEAP8[$94>>0]|0;
   $96 = $95&255;
   $97 = 1 << $96;
   $ause$0$ph$i$i = 0;$i$0$ph$i$i = $97;$totaluse$0$ph$i$i = 0;
   L41: while(1) {
    $i$0$i$i = $i$0$ph$i$i;
    while(1) {
     $98 = (($i$0$i$i) + -1)|0;
     $99 = ($i$0$i$i|0)==(0);
     if ($99) {
      $ause$0$ph$i$i$lcssa = $ause$0$ph$i$i;$totaluse$0$ph$i$i$lcssa = $totaluse$0$ph$i$i;
      break L41;
     }
     $100 = HEAP32[$66>>2]|0;
     $101 = (((($100) + ($98<<5)|0)) + 8|0);
     $102 = HEAP32[$101>>2]|0;
     $103 = ($102|0)==(0);
     if ($103) {
      $i$0$i$i = $98;
     } else {
      $$lcssa54 = $98;$$lcssa55 = $100;
      break;
     }
    }
    $104 = (((($$lcssa55) + ($$lcssa54<<5)|0)) + 24|0);
    $105 = HEAP32[$104>>2]|0;
    $106 = ($105|0)==(3);
    if ($106) {
     $107 = (((($$lcssa55) + ($$lcssa54<<5)|0)) + 16|0);
     $108 = +HEAPF64[$107>>3];
     $109 = (~~(($108)));
     $110 = (+($109|0));
     $not$$i$i$i$i = $110 == $108;
     $$off$i$i$i = (($109) + -1)|0;
     $111 = ($$off$i$i$i>>>0)<(67108864);
     $or$cond$i$i$i = $not$$i$i$i$i & $111;
     if ($or$cond$i$i$i) {
      $112 = (_luaO_log2($$off$i$i$i)|0);
      $113 = (($112) + 1)|0;
      $114 = (($nums$i) + ($113<<2)|0);
      $115 = HEAP32[$114>>2]|0;
      $116 = (($115) + 1)|0;
      HEAP32[$114>>2] = $116;
      $$0$i$i$i = 1;
     } else {
      $$0$i$i$i = 0;
     }
    } else {
     $$0$i$i$i = 0;
    }
    $117 = (($$0$i$i$i) + ($ause$0$ph$i$i))|0;
    $118 = (($totaluse$0$ph$i$i) + 1)|0;
    $ause$0$ph$i$i = $117;$i$0$ph$i$i = $$lcssa54;$totaluse$0$ph$i$i = $118;
   }
   $119 = (($ause$0$ph$i$i$lcssa) + ($ause$0$lcssa$i$i))|0;
   $120 = HEAP32[$0>>2]|0;
   $121 = ($120|0)==(3);
   if ($121) {
    $122 = +HEAPF64[$key>>3];
    $123 = (~~(($122)));
    $124 = (+($123|0));
    $not$$i$i$i = $124 == $122;
    $$off$i$i = (($123) + -1)|0;
    $125 = ($$off$i$i>>>0)<(67108864);
    $or$cond$i$i = $not$$i$i$i & $125;
    if ($or$cond$i$i) {
     $126 = (_luaO_log2($$off$i$i)|0);
     $127 = (($126) + 1)|0;
     $128 = (($nums$i) + ($127<<2)|0);
     $129 = HEAP32[$128>>2]|0;
     $130 = (($129) + 1)|0;
     HEAP32[$128>>2] = $130;
     $$0$i$i = 1;
    } else {
     $$0$i$i = 0;
    }
   } else {
    $$0$i$i = 0;
   }
   $131 = (($119) + ($$0$i$i))|0;
   $132 = ($131|0)>(0);
   L56: do {
    if ($132) {
     $138 = 0;$a$03$i$i = 0;$i$01$i$i = 0;$n$05$i$i = 0;$na$04$i$i = 0;$twotoi$02$i$i = 1;
     while(1) {
      $133 = (($nums$i) + ($i$01$i$i<<2)|0);
      $134 = HEAP32[$133>>2]|0;
      $135 = ($134|0)>(0);
      if ($135) {
       $136 = (($134) + ($a$03$i$i))|0;
       $137 = ($136|0)>($138|0);
       $$na$0$i$i = $137 ? $136 : $na$04$i$i;
       $twotoi$0$n$0$i$i = $137 ? $twotoi$02$i$i : $n$05$i$i;
       $a$1$i$i = $136;$n$1$i$i = $twotoi$0$n$0$i$i;$na$1$i$i = $$na$0$i$i;
      } else {
       $a$1$i$i = $a$03$i$i;$n$1$i$i = $n$05$i$i;$na$1$i$i = $na$04$i$i;
      }
      $139 = ($a$1$i$i|0)==($131|0);
      if ($139) {
       $n$2$i$i = $n$1$i$i;$na$2$i$i = $na$1$i$i;
       break L56;
      }
      $140 = (($i$01$i$i) + 1)|0;
      $141 = $twotoi$02$i$i << 1;
      $142 = ($twotoi$02$i$i|0)<($131|0);
      if ($142) {
       $138$phi = $twotoi$02$i$i;$a$03$i$i = $a$1$i$i;$i$01$i$i = $140;$n$05$i$i = $n$1$i$i;$na$04$i$i = $na$1$i$i;$twotoi$02$i$i = $141;$138 = $138$phi;
      } else {
       $n$2$i$i = $n$1$i$i;$na$2$i$i = $na$1$i$i;
       break;
      }
     }
    } else {
     $n$2$i$i = 0;$na$2$i$i = 0;
    }
   } while(0);
   $143 = (($ause$0$lcssa$i$i) + 1)|0;
   $144 = (($143) + ($totaluse$0$ph$i$i$lcssa))|0;
   $145 = (($144) - ($na$2$i$i))|0;
   _resize($L,$t,$n$2$i$i,$145);
   $146 = (_luaH_get($t,$key)|0);
   $147 = ((($t)) + 6|0);
   HEAP8[$147>>0] = 0;
   $148 = ($146|0)==(624096|0);
   if (!($148)) {
    $$1 = $146;
    STACKTOP = sp;return ($$1|0);
   }
   $149 = HEAP32[$0>>2]|0;
   switch ($149|0) {
   case 0:  {
    _luaG_runerror($L,518929,$vararg_buffer);
    break;
   }
   case 3:  {
    $150 = +HEAPF64[$key>>3];
    $151 = ($150 == $150) & (0.0 == 0.0);
    if (!($151)) {
     _luaG_runerror($L,518948,$vararg_buffer1);
    }
    break;
   }
   default: {
   }
   }
   $152 = (_newkey($L,$t,$key)|0);
   $$1 = $152;
   STACKTOP = sp;return ($$1|0);
  } else {
   $mp$2 = $$0$i;
  }
 } while(0);
 $213 = ((($mp$2)) + 16|0);
 $214 = $key;
 $215 = $214;
 $216 = HEAP32[$215>>2]|0;
 $217 = (($214) + 4)|0;
 $218 = $217;
 $219 = HEAP32[$218>>2]|0;
 $220 = $213;
 $221 = $220;
 HEAP32[$221>>2] = $216;
 $222 = (($220) + 4)|0;
 $223 = $222;
 HEAP32[$223>>2] = $219;
 $224 = HEAP32[$0>>2]|0;
 $225 = ((($mp$2)) + 24|0);
 HEAP32[$225>>2] = $224;
 $226 = ($224|0)>(3);
 if ($226) {
  $227 = HEAP32[$key>>2]|0;
  $228 = ((($227)) + 5|0);
  $229 = HEAP8[$228>>0]|0;
  $230 = $229 & 3;
  $231 = ($230<<24>>24)==(0);
  if (!($231)) {
   $232 = ((($t)) + 5|0);
   $233 = HEAP8[$232>>0]|0;
   $234 = $233 & 4;
   $235 = ($234<<24>>24)==(0);
   if (!($235)) {
    _luaC_barrierback($L,$t);
   }
  }
 }
 $$1 = $mp$2;
 STACKTOP = sp;return ($$1|0);
}
function _luaH_get($t,$key) {
 $t = $t|0;
 $key = $key|0;
 var $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0.0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0.0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $n$0$i = 0, $n$0$i$lcssa = 0, $n$0$i4 = 0, $n$0$i4$lcssa = 0, $n1$0 = 0, $n1$0$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($key)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 L1: do {
  switch ($1|0) {
  case 4:  {
   $2 = HEAP32[$key>>2]|0;
   $3 = ((($2)) + 8|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ((($t)) + 7|0);
   $6 = HEAP8[$5>>0]|0;
   $7 = $6&255;
   $8 = 1 << $7;
   $9 = (($8) + -1)|0;
   $10 = $9 & $4;
   $11 = ((($t)) + 16|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = (($12) + ($10<<5)|0);
   $n$0$i = $13;
   while(1) {
    $14 = ((($n$0$i)) + 24|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = ($15|0)==(4);
    if ($16) {
     $17 = ((($n$0$i)) + 16|0);
     $18 = HEAP32[$17>>2]|0;
     $19 = ($18|0)==($2|0);
     if ($19) {
      $n$0$i$lcssa = $n$0$i;
      break;
     }
    }
    $20 = ((($n$0$i)) + 28|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = ($21|0)==(0|0);
    if ($22) {
     $$2 = 624096;
     label = 24;
     break;
    } else {
     $n$0$i = $21;
    }
   }
   if ((label|0) == 24) {
    return ($$2|0);
   }
   $$2 = $n$0$i$lcssa;
   return ($$2|0);
   break;
  }
  case 3:  {
   $23 = +HEAPF64[$key>>3];
   $24 = (~~(($23)));
   $25 = (+($24|0));
   $26 = $25 == $23;
   if (!($26)) {
    $57 = +HEAPF64[$key>>3];
    $58 = $57 == 0.0;
    $59 = ((($t)) + 16|0);
    $60 = HEAP32[$59>>2]|0;
    if ($58) {
     $n1$0 = $60;
     break L1;
    }
    HEAPF64[tempDoublePtr>>3] = $57;$61 = HEAP32[tempDoublePtr>>2]|0;
    $62 = HEAP32[tempDoublePtr+4>>2]|0;
    $63 = (($62) + ($61))|0;
    $64 = ((($t)) + 7|0);
    $65 = HEAP8[$64>>0]|0;
    $66 = $65&255;
    $67 = 1 << $66;
    $68 = (($67) + -1)|0;
    $69 = $68 | 1;
    $70 = (($63>>>0) % ($69>>>0))&-1;
    $71 = (($60) + ($70<<5)|0);
    $n1$0 = $71;
    break L1;
   }
   $27 = (($24) + -1)|0;
   $28 = ((($t)) + 28|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = ($27>>>0)<($29>>>0);
   if ($30) {
    $31 = ((($t)) + 12|0);
    $32 = HEAP32[$31>>2]|0;
    $33 = (($32) + ($27<<4)|0);
    $$2 = $33;
    return ($$2|0);
   }
   $34 = ($24|0)==(0);
   $35 = ((($t)) + 16|0);
   $36 = HEAP32[$35>>2]|0;
   if ($34) {
    $n$0$i4 = $36;
   } else {
    HEAPF64[tempDoublePtr>>3] = $25;$37 = HEAP32[tempDoublePtr>>2]|0;
    $38 = HEAP32[tempDoublePtr+4>>2]|0;
    $39 = (($38) + ($37))|0;
    $40 = ((($t)) + 7|0);
    $41 = HEAP8[$40>>0]|0;
    $42 = $41&255;
    $43 = 1 << $42;
    $44 = (($43) + -1)|0;
    $45 = $44 | 1;
    $46 = (($39>>>0) % ($45>>>0))&-1;
    $47 = (($36) + ($46<<5)|0);
    $n$0$i4 = $47;
   }
   while(1) {
    $48 = ((($n$0$i4)) + 24|0);
    $49 = HEAP32[$48>>2]|0;
    $50 = ($49|0)==(3);
    if ($50) {
     $51 = ((($n$0$i4)) + 16|0);
     $52 = +HEAPF64[$51>>3];
     $53 = $52 == $25;
     if ($53) {
      $n$0$i4$lcssa = $n$0$i4;
      break;
     }
    }
    $54 = ((($n$0$i4)) + 28|0);
    $55 = HEAP32[$54>>2]|0;
    $56 = ($55|0)==(0|0);
    if ($56) {
     $$2 = 624096;
     label = 24;
     break;
    } else {
     $n$0$i4 = $55;
    }
   }
   if ((label|0) == 24) {
    return ($$2|0);
   }
   $$2 = $n$0$i4$lcssa;
   return ($$2|0);
   break;
  }
  case 1:  {
   $72 = HEAP32[$key>>2]|0;
   $73 = ((($t)) + 7|0);
   $74 = HEAP8[$73>>0]|0;
   $75 = $74&255;
   $76 = 1 << $75;
   $77 = (($76) + -1)|0;
   $78 = $77 & $72;
   $79 = ((($t)) + 16|0);
   $80 = HEAP32[$79>>2]|0;
   $81 = (($80) + ($78<<5)|0);
   $n1$0 = $81;
   break;
  }
  case 2:  {
   $82 = HEAP32[$key>>2]|0;
   $83 = ((($t)) + 7|0);
   $84 = HEAP8[$83>>0]|0;
   $85 = $84&255;
   $86 = 1 << $85;
   $87 = (($86) + -1)|0;
   $88 = $87 | 1;
   $89 = (($82>>>0) % ($88>>>0))&-1;
   $90 = ((($t)) + 16|0);
   $91 = HEAP32[$90>>2]|0;
   $92 = (($91) + ($89<<5)|0);
   $n1$0 = $92;
   break;
  }
  case 0:  {
   $$2 = 624096;
   return ($$2|0);
   break;
  }
  default: {
   $93 = HEAP32[$key>>2]|0;
   $94 = ((($t)) + 7|0);
   $95 = HEAP8[$94>>0]|0;
   $96 = $95&255;
   $97 = 1 << $96;
   $98 = (($97) + -1)|0;
   $99 = $98 | 1;
   $100 = (($93>>>0) % ($99>>>0))&-1;
   $101 = ((($t)) + 16|0);
   $102 = HEAP32[$101>>2]|0;
   $103 = (($102) + ($100<<5)|0);
   $n1$0 = $103;
  }
  }
 } while(0);
 while(1) {
  $104 = ((($n1$0)) + 16|0);
  $105 = (_luaO_rawequalObj($104,$key)|0);
  $106 = ($105|0)==(0);
  if (!($106)) {
   $n1$0$lcssa = $n1$0;
   break;
  }
  $107 = ((($n1$0)) + 28|0);
  $108 = HEAP32[$107>>2]|0;
  $109 = ($108|0)==(0|0);
  if ($109) {
   $$2 = 624096;
   label = 24;
   break;
  } else {
   $n1$0 = $108;
  }
 }
 if ((label|0) == 24) {
  return ($$2|0);
 }
 $$2 = $n1$0$lcssa;
 return ($$2|0);
}
function _luaH_new($L,$narray,$nhash) {
 $L = $L|0;
 $narray = $narray|0;
 $nhash = $nhash|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i = 0, $i$01$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = (_luaM_realloc_($L,0,0,32)|0);
 _luaC_link($L,$0,5);
 $1 = ((($0)) + 8|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($0)) + 6|0);
 HEAP8[$2>>0] = -1;
 $3 = ((($0)) + 12|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($0)) + 28|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($0)) + 7|0);
 HEAP8[$5>>0] = 0;
 $6 = ((($0)) + 16|0);
 HEAP32[$6>>2] = 624112;
 $7 = (($narray) + 1)|0;
 $8 = ($7>>>0)<(268435456);
 if ($8) {
  $9 = $narray << 4;
  $10 = (_luaM_realloc_($L,0,0,$9)|0);
  $12 = $10;
 } else {
  $11 = (_luaM_toobig($L)|0);
  $12 = $11;
 }
 HEAP32[$3>>2] = $12;
 $13 = HEAP32[$4>>2]|0;
 $14 = ($13|0)<($narray|0);
 if ($14) {
  $i$01$i = $13;
 } else {
  HEAP32[$4>>2] = $narray;
  _setnodevector($L,$0,$nhash);
  return ($0|0);
 }
 while(1) {
  $15 = (((($12) + ($i$01$i<<4)|0)) + 8|0);
  HEAP32[$15>>2] = 0;
  $16 = (($i$01$i) + 1)|0;
  $exitcond$i = ($16|0)==($narray|0);
  if ($exitcond$i) {
   break;
  } else {
   $i$01$i = $16;
  }
 }
 HEAP32[$4>>2] = $narray;
 _setnodevector($L,$0,$nhash);
 return ($0|0);
}
function _luaH_free($L,$t) {
 $L = $L|0;
 $t = $t|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($t)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(624112|0);
 if (!($2)) {
  $3 = ((($t)) + 7|0);
  $4 = HEAP8[$3>>0]|0;
  $5 = $4&255;
  $6 = 32 << $5;
  (_luaM_realloc_($L,$1,$6,0)|0);
 }
 $7 = ((($t)) + 12|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($t)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 4;
 (_luaM_realloc_($L,$8,$11,0)|0);
 (_luaM_realloc_($L,$t,32,0)|0);
 return;
}
function _luaH_getnum($t,$key) {
 $t = $t|0;
 $key = $key|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0.0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0, $n$0 = 0, $n$0$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($key) + -1)|0;
 $1 = ((($t)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($0>>>0)<($2>>>0);
 if ($3) {
  $4 = ((($t)) + 12|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = (($5) + ($0<<4)|0);
  $$1 = $6;
  return ($$1|0);
 }
 $7 = (+($key|0));
 $8 = ($key|0)==(0);
 $9 = ((($t)) + 16|0);
 $10 = HEAP32[$9>>2]|0;
 if ($8) {
  $n$0 = $10;
 } else {
  HEAPF64[tempDoublePtr>>3] = $7;$11 = HEAP32[tempDoublePtr>>2]|0;
  $12 = HEAP32[tempDoublePtr+4>>2]|0;
  $13 = (($12) + ($11))|0;
  $14 = ((($t)) + 7|0);
  $15 = HEAP8[$14>>0]|0;
  $16 = $15&255;
  $17 = 1 << $16;
  $18 = (($17) + -1)|0;
  $19 = $18 | 1;
  $20 = (($13>>>0) % ($19>>>0))&-1;
  $21 = (($10) + ($20<<5)|0);
  $n$0 = $21;
 }
 while(1) {
  $22 = ((($n$0)) + 24|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = ($23|0)==(3);
  if ($24) {
   $25 = ((($n$0)) + 16|0);
   $26 = +HEAPF64[$25>>3];
   $27 = $26 == $7;
   if ($27) {
    $n$0$lcssa = $n$0;
    break;
   }
  }
  $28 = ((($n$0)) + 28|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = ($29|0)==(0|0);
  if ($30) {
   $$1 = 624096;
   label = 9;
   break;
  } else {
   $n$0 = $29;
  }
 }
 if ((label|0) == 9) {
  return ($$1|0);
 }
 $$1 = $n$0$lcssa;
 return ($$1|0);
}
function _luaH_getstr($t,$key) {
 $t = $t|0;
 $key = $key|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $n$0 = 0, $n$0$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($key)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($t)) + 7|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = 1 << $4;
 $6 = (($5) + -1)|0;
 $7 = $6 & $1;
 $8 = ((($t)) + 16|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (($9) + ($7<<5)|0);
 $n$0 = $10;
 while(1) {
  $11 = ((($n$0)) + 24|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ($12|0)==(4);
  if ($13) {
   $14 = ((($n$0)) + 16|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)==($key|0);
   if ($16) {
    $n$0$lcssa = $n$0;
    break;
   }
  }
  $17 = ((($n$0)) + 28|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = ($18|0)==(0|0);
  if ($19) {
   $$0 = 624096;
   label = 6;
   break;
  } else {
   $n$0 = $18;
  }
 }
 if ((label|0) == 6) {
  return ($$0|0);
 }
 $$0 = $n$0$lcssa;
 return ($$0|0);
}
function _luaH_set($L,$t,$key) {
 $L = $L|0;
 $t = $t|0;
 $key = $key|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_luaH_get($t,$key)|0);
 $1 = ((($t)) + 6|0);
 HEAP8[$1>>0] = 0;
 $2 = ($0|0)==(624096|0);
 if (!($2)) {
  $$0 = $0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ((($key)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 switch ($4|0) {
 case 0:  {
  _luaG_runerror($L,518929,$vararg_buffer);
  break;
 }
 case 3:  {
  $5 = +HEAPF64[$key>>3];
  $6 = ($5 == $5) & (0.0 == 0.0);
  if (!($6)) {
   _luaG_runerror($L,518948,$vararg_buffer1);
  }
  break;
 }
 default: {
 }
 }
 $7 = (_newkey($L,$t,$key)|0);
 $$0 = $7;
 STACKTOP = sp;return ($$0|0);
}
function _luaH_setnum($L,$t,$key) {
 $L = $L|0;
 $t = $t|0;
 $key = $key|0;
 var $$0 = 0, $$1$i = 0, $$pre = 0.0, $$pre$phiZ2D = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0.0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0, $9 = 0, $k = 0, $n$0$i = 0;
 var $n$0$i$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $k = sp;
 $0 = (($key) + -1)|0;
 $1 = ((($t)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($0>>>0)<($2>>>0);
 L1: do {
  if ($3) {
   $4 = ((($t)) + 12|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = (($5) + ($0<<4)|0);
   $$1$i = $6;
   label = 9;
  } else {
   $7 = (+($key|0));
   $8 = ($key|0)==(0);
   $9 = ((($t)) + 16|0);
   $10 = HEAP32[$9>>2]|0;
   if ($8) {
    $n$0$i = $10;
   } else {
    HEAPF64[tempDoublePtr>>3] = $7;$11 = HEAP32[tempDoublePtr>>2]|0;
    $12 = HEAP32[tempDoublePtr+4>>2]|0;
    $13 = (($12) + ($11))|0;
    $14 = ((($t)) + 7|0);
    $15 = HEAP8[$14>>0]|0;
    $16 = $15&255;
    $17 = 1 << $16;
    $18 = (($17) + -1)|0;
    $19 = $18 | 1;
    $20 = (($13>>>0) % ($19>>>0))&-1;
    $21 = (($10) + ($20<<5)|0);
    $n$0$i = $21;
   }
   while(1) {
    $22 = ((($n$0$i)) + 24|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = ($23|0)==(3);
    if ($24) {
     $25 = ((($n$0$i)) + 16|0);
     $26 = +HEAPF64[$25>>3];
     $27 = $26 == $7;
     if ($27) {
      $n$0$i$lcssa = $n$0$i;
      break;
     }
    }
    $28 = ((($n$0$i)) + 28|0);
    $29 = HEAP32[$28>>2]|0;
    $30 = ($29|0)==(0|0);
    if ($30) {
     $$pre$phiZ2D = $7;
     break L1;
    } else {
     $n$0$i = $29;
    }
   }
   $$1$i = $n$0$i$lcssa;
   label = 9;
  }
 } while(0);
 do {
  if ((label|0) == 9) {
   $31 = ($$1$i|0)==(624096|0);
   if ($31) {
    $$pre = (+($key|0));
    $$pre$phiZ2D = $$pre;
    break;
   } else {
    $$0 = $$1$i;
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 HEAPF64[$k>>3] = $$pre$phiZ2D;
 $32 = ((($k)) + 8|0);
 HEAP32[$32>>2] = 3;
 $33 = (_newkey($L,$t,$k)|0);
 $$0 = $33;
 STACKTOP = sp;return ($$0|0);
}
function _luaH_setstr($L,$t,$key) {
 $L = $L|0;
 $t = $t|0;
 $key = $key|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $k = 0, $n$0$i = 0, $n$0$i$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $k = sp;
 $0 = ((($key)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($t)) + 7|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = 1 << $4;
 $6 = (($5) + -1)|0;
 $7 = $6 & $1;
 $8 = ((($t)) + 16|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (($9) + ($7<<5)|0);
 $n$0$i = $10;
 while(1) {
  $11 = ((($n$0$i)) + 24|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ($12|0)==(4);
  if ($13) {
   $14 = ((($n$0$i)) + 16|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)==($key|0);
   if ($16) {
    $n$0$i$lcssa = $n$0$i;
    label = 5;
    break;
   }
  }
  $17 = ((($n$0$i)) + 28|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = ($18|0)==(0|0);
  if ($19) {
   break;
  } else {
   $n$0$i = $18;
  }
 }
 if ((label|0) == 5) {
  $20 = ($n$0$i$lcssa|0)==(624096|0);
  if (!($20)) {
   $$0 = $n$0$i$lcssa;
   STACKTOP = sp;return ($$0|0);
  }
 }
 HEAP32[$k>>2] = $key;
 $21 = ((($k)) + 8|0);
 HEAP32[$21>>2] = 4;
 $22 = (_newkey($L,$t,$k)|0);
 $$0 = $22;
 STACKTOP = sp;return ($$0|0);
}
function _luaH_getn($t) {
 $t = $t|0;
 var $$$1$i = 0, $$0 = 0, $$01$i = 0, $$01$i$lcssa = 0, $$1$i$i = 0, $$1$i10$i = 0, $$1$i5$i = 0, $$119$i = 0, $$j$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0.0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0.0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0.0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0.0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$0$ = 0, $i$0$i = 0, $i$0$i$lcssa = 0, $i$0$i$phi = 0, $i$09 = 0, $i$1$i = 0, $i$2$$i = 0, $i$220$i = 0, $j$08 = 0, $n$0$i$i = 0, $n$0$i$i$lcssa = 0, $n$0$i3$i = 0, $n$0$i3$i$lcssa = 0, $n$0$i8$i = 0;
 var $n$0$i8$i$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($t)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  $3 = (($1) + -1)|0;
  $4 = ((($t)) + 12|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = (((($5) + ($3<<4)|0)) + 8|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)==(0);
  if ($8) {
   $9 = ($1>>>0)>(1);
   if ($9) {
    $i$09 = 0;$j$08 = $1;
   } else {
    $$0 = 0;
    return ($$0|0);
   }
   while(1) {
    $10 = (($i$09) + ($j$08))|0;
    $11 = $10 >>> 1;
    $12 = (($11) + -1)|0;
    $13 = (((($5) + ($12<<4)|0)) + 8|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = ($14|0)==(0);
    $$j$0 = $15 ? $11 : $j$08;
    $i$0$ = $15 ? $i$09 : $11;
    $16 = (($$j$0) - ($i$0$))|0;
    $17 = ($16>>>0)>(1);
    if ($17) {
     $i$09 = $i$0$;$j$08 = $$j$0;
    } else {
     $$0 = $i$0$;
     break;
    }
   }
   return ($$0|0);
  }
 }
 $18 = ((($t)) + 16|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($19|0)==(624112|0);
 if ($20) {
  $$0 = $1;
  return ($$0|0);
 }
 $21 = (($1) + 1)|0;
 $22 = ((($t)) + 12|0);
 $23 = ((($t)) + 7|0);
 $$01$i = $21;$i$0$i = $1;
 while(1) {
  $24 = (($$01$i) + -1)|0;
  $25 = ($24>>>0)<($1>>>0);
  L15: do {
   if ($25) {
    $26 = HEAP32[$22>>2]|0;
    $27 = (($26) + ($24<<4)|0);
    $$1$i$i = $27;
   } else {
    $28 = (+($$01$i|0));
    $29 = ($$01$i|0)==(0);
    if ($29) {
     $n$0$i$i = $19;
    } else {
     HEAPF64[tempDoublePtr>>3] = $28;$30 = HEAP32[tempDoublePtr>>2]|0;
     $31 = HEAP32[tempDoublePtr+4>>2]|0;
     $32 = (($31) + ($30))|0;
     $33 = HEAP8[$23>>0]|0;
     $34 = $33&255;
     $35 = 1 << $34;
     $36 = (($35) + -1)|0;
     $37 = $36 | 1;
     $38 = (($32>>>0) % ($37>>>0))&-1;
     $39 = (($19) + ($38<<5)|0);
     $n$0$i$i = $39;
    }
    while(1) {
     $40 = ((($n$0$i$i)) + 24|0);
     $41 = HEAP32[$40>>2]|0;
     $42 = ($41|0)==(3);
     if ($42) {
      $43 = ((($n$0$i$i)) + 16|0);
      $44 = +HEAPF64[$43>>3];
      $45 = $44 == $28;
      if ($45) {
       $n$0$i$i$lcssa = $n$0$i$i;
       break;
      }
     }
     $46 = ((($n$0$i$i)) + 28|0);
     $47 = HEAP32[$46>>2]|0;
     $48 = ($47|0)==(0|0);
     if ($48) {
      $$1$i$i = 624096;
      break L15;
     } else {
      $n$0$i$i = $47;
     }
    }
    $$1$i$i = $n$0$i$i$lcssa;
   }
  } while(0);
  $49 = ((($$1$i$i)) + 8|0);
  $50 = HEAP32[$49>>2]|0;
  $51 = ($50|0)==(0);
  if ($51) {
   $$01$i$lcssa = $$01$i;$i$0$i$lcssa = $i$0$i;
   break;
  }
  $54 = $$01$i << 1;
  $55 = ($54>>>0)>(2147483645);
  if ($55) {
   $i$1$i = 1;
   label = 18;
   break;
  } else {
   $i$0$i$phi = $$01$i;$$01$i = $54;$i$0$i = $i$0$i$phi;
  }
 }
 if ((label|0) == 18) {
  while(1) {
   label = 0;
   $56 = (($i$1$i) + -1)|0;
   $57 = ($56>>>0)<($1>>>0);
   L31: do {
    if ($57) {
     $58 = HEAP32[$22>>2]|0;
     $59 = (($58) + ($56<<4)|0);
     $$1$i10$i = $59;
    } else {
     $60 = (+($i$1$i|0));
     $61 = ($i$1$i|0)==(0);
     if ($61) {
      $n$0$i8$i = $19;
     } else {
      HEAPF64[tempDoublePtr>>3] = $60;$62 = HEAP32[tempDoublePtr>>2]|0;
      $63 = HEAP32[tempDoublePtr+4>>2]|0;
      $64 = (($63) + ($62))|0;
      $65 = HEAP8[$23>>0]|0;
      $66 = $65&255;
      $67 = 1 << $66;
      $68 = (($67) + -1)|0;
      $69 = $68 | 1;
      $70 = (($64>>>0) % ($69>>>0))&-1;
      $71 = (($19) + ($70<<5)|0);
      $n$0$i8$i = $71;
     }
     while(1) {
      $72 = ((($n$0$i8$i)) + 24|0);
      $73 = HEAP32[$72>>2]|0;
      $74 = ($73|0)==(3);
      if ($74) {
       $75 = ((($n$0$i8$i)) + 16|0);
       $76 = +HEAPF64[$75>>3];
       $77 = $76 == $60;
       if ($77) {
        $n$0$i8$i$lcssa = $n$0$i8$i;
        break;
       }
      }
      $78 = ((($n$0$i8$i)) + 28|0);
      $79 = HEAP32[$78>>2]|0;
      $80 = ($79|0)==(0|0);
      if ($80) {
       $$1$i10$i = 624096;
       break L31;
      } else {
       $n$0$i8$i = $79;
      }
     }
     $$1$i10$i = $n$0$i8$i$lcssa;
    }
   } while(0);
   $81 = ((($$1$i10$i)) + 8|0);
   $82 = HEAP32[$81>>2]|0;
   $83 = ($82|0)==(0);
   $84 = (($i$1$i) + 1)|0;
   if ($83) {
    $$0 = $56;
    break;
   } else {
    $i$1$i = $84;
    label = 18;
   }
  }
  return ($$0|0);
 }
 $52 = (($$01$i$lcssa) - ($i$0$i$lcssa))|0;
 $53 = ($52>>>0)>(1);
 if ($53) {
  $$119$i = $$01$i$lcssa;$i$220$i = $i$0$i$lcssa;
 } else {
  $$0 = $i$0$i$lcssa;
  return ($$0|0);
 }
 while(1) {
  $85 = (($$119$i) + ($i$220$i))|0;
  $86 = $85 >>> 1;
  $87 = (($86) + -1)|0;
  $88 = ($87>>>0)<($1>>>0);
  L49: do {
   if ($88) {
    $89 = HEAP32[$22>>2]|0;
    $90 = (($89) + ($87<<4)|0);
    $$1$i5$i = $90;
   } else {
    $91 = (+($86|0));
    $92 = ($86|0)==(0);
    if ($92) {
     $n$0$i3$i = $19;
    } else {
     HEAPF64[tempDoublePtr>>3] = $91;$93 = HEAP32[tempDoublePtr>>2]|0;
     $94 = HEAP32[tempDoublePtr+4>>2]|0;
     $95 = (($94) + ($93))|0;
     $96 = HEAP8[$23>>0]|0;
     $97 = $96&255;
     $98 = 1 << $97;
     $99 = (($98) + -1)|0;
     $100 = $99 | 1;
     $101 = (($95>>>0) % ($100>>>0))&-1;
     $102 = (($19) + ($101<<5)|0);
     $n$0$i3$i = $102;
    }
    while(1) {
     $103 = ((($n$0$i3$i)) + 24|0);
     $104 = HEAP32[$103>>2]|0;
     $105 = ($104|0)==(3);
     if ($105) {
      $106 = ((($n$0$i3$i)) + 16|0);
      $107 = +HEAPF64[$106>>3];
      $108 = $107 == $91;
      if ($108) {
       $n$0$i3$i$lcssa = $n$0$i3$i;
       break;
      }
     }
     $109 = ((($n$0$i3$i)) + 28|0);
     $110 = HEAP32[$109>>2]|0;
     $111 = ($110|0)==(0|0);
     if ($111) {
      $$1$i5$i = 624096;
      break L49;
     } else {
      $n$0$i3$i = $110;
     }
    }
    $$1$i5$i = $n$0$i3$i$lcssa;
   }
  } while(0);
  $112 = ((($$1$i5$i)) + 8|0);
  $113 = HEAP32[$112>>2]|0;
  $114 = ($113|0)==(0);
  $$$1$i = $114 ? $86 : $$119$i;
  $i$2$$i = $114 ? $i$220$i : $86;
  $115 = (($$$1$i) - ($i$2$$i))|0;
  $116 = ($115>>>0)>(1);
  if ($116) {
   $$119$i = $$$1$i;$i$220$i = $i$2$$i;
  } else {
   $$0 = $i$2$$i;
   break;
  }
 }
 return ($$0|0);
}
function _luaT_init($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($L)) + 16|0);
 $i$01 = 0;
 while(1) {
  $1 = (38888 + ($i$01<<2)|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = (_strlen($2)|0);
  $4 = (_luaS_newlstr($L,$2,$3)|0);
  $5 = HEAP32[$0>>2]|0;
  $6 = (((($5)) + 188|0) + ($i$01<<2)|0);
  HEAP32[$6>>2] = $4;
  $7 = HEAP32[$0>>2]|0;
  $8 = (((($7)) + 188|0) + ($i$01<<2)|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($9)) + 5|0);
  $11 = HEAP8[$10>>0]|0;
  $12 = $11&255;
  $13 = $12 | 32;
  $14 = $13&255;
  HEAP8[$10>>0] = $14;
  $15 = (($i$01) + 1)|0;
  $exitcond = ($15|0)==(17);
  if ($exitcond) {
   break;
  } else {
   $i$01 = $15;
  }
 }
 return;
}
function _luaT_gettm($events,$event,$ename) {
 $events = $events|0;
 $event = $event|0;
 $ename = $ename|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaH_getstr($events,$ename)|0);
 $1 = ((($0)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0);
 if (!($3)) {
  $$0 = $0;
  return ($$0|0);
 }
 $4 = 1 << $event;
 $5 = ((($events)) + 6|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = $6&255;
 $8 = $7 | $4;
 $9 = $8&255;
 HEAP8[$5>>0] = $9;
 $$0 = 0;
 return ($$0|0);
}
function _luaT_gettmbyobj($L,$o,$event) {
 $L = $L|0;
 $o = $o|0;
 $event = $event|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $mt$0 = 0, $mt$0$in = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($o)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 switch ($1|0) {
 case 5:  {
  $2 = HEAP32[$o>>2]|0;
  $3 = ((($2)) + 8|0);
  $mt$0$in = $3;
  break;
 }
 case 7:  {
  $4 = HEAP32[$o>>2]|0;
  $5 = ((($4)) + 8|0);
  $mt$0$in = $5;
  break;
 }
 default: {
  $6 = ((($L)) + 16|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (((($7)) + 152|0) + ($1<<2)|0);
  $mt$0$in = $8;
 }
 }
 $mt$0 = HEAP32[$mt$0$in>>2]|0;
 $9 = ($mt$0|0)==(0|0);
 if ($9) {
  $15 = 624096;
  return ($15|0);
 }
 $10 = ((($L)) + 16|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (((($11)) + 188|0) + ($event<<2)|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (_luaH_getstr($mt$0,$13)|0);
 $15 = $14;
 return ($15|0);
}
function _luaU_undump($L,$Z,$buff,$name) {
 $L = $L|0;
 $Z = $Z|0;
 $buff = $buff|0;
 $name = $name|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $S = 0, $h$i = 0, $s$i = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $h$i = sp + 32|0;
 $s$i = sp + 44|0;
 $S = sp + 16|0;
 $0 = HEAP8[$name>>0]|0;
 switch ($0<<24>>24) {
 case 61: case 64:  {
  $1 = ((($name)) + 1|0);
  $2 = ((($S)) + 12|0);
  HEAP32[$2>>2] = $1;
  $17 = $1;
  break;
 }
 case 27:  {
  $3 = ((($S)) + 12|0);
  HEAP32[$3>>2] = 519100;
  $17 = 519100;
  break;
 }
 default: {
  $4 = ((($S)) + 12|0);
  HEAP32[$4>>2] = $name;
  $17 = $name;
 }
 }
 HEAP32[$S>>2] = $L;
 $5 = ((($S)) + 4|0);
 HEAP32[$5>>2] = $Z;
 $6 = ((($S)) + 8|0);
 HEAP32[$6>>2] = $buff;
 HEAP32[$h$i>>2] = 1635077147;
 $7 = ((($h$i)) + 4|0);
 $8 = ((($h$i)) + 5|0);
 HEAP8[$7>>0] = 81;
 $9 = ((($h$i)) + 6|0);
 HEAP8[$8>>0] = 0;
 $10 = ((($h$i)) + 7|0);
 HEAP8[$9>>0] = 1;
 $11 = ((($h$i)) + 8|0);
 HEAP8[$10>>0] = 4;
 $12 = ((($h$i)) + 9|0);
 HEAP8[$11>>0] = 4;
 $13 = ((($h$i)) + 10|0);
 HEAP8[$12>>0] = 4;
 $14 = ((($h$i)) + 11|0);
 HEAP8[$13>>0] = 8;
 HEAP8[$14>>0] = 0;
 $15 = (_luaZ_read($Z,$s$i,12)|0);
 $16 = ($15|0)==(0);
 if (!($16)) {
  HEAP32[$vararg_buffer>>2] = $17;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = 519142;
  (_luaO_pushfstring($L,519114,$vararg_buffer)|0);
  $18 = HEAP32[$S>>2]|0;
  _luaD_throw($18,3);
 }
 $19 = (_memcmp($h$i,$s$i,12)|0);
 $20 = ($19|0)==(0);
 if ($20) {
  $25 = (_luaS_newlstr($L,519168,2)|0);
  $26 = (_LoadFunction($S,$25)|0);
  STACKTOP = sp;return ($26|0);
 }
 $21 = HEAP32[$S>>2]|0;
 $22 = ((($S)) + 12|0);
 $23 = HEAP32[$22>>2]|0;
 HEAP32[$vararg_buffer2>>2] = $23;
 $vararg_ptr5 = ((($vararg_buffer2)) + 4|0);
 HEAP32[$vararg_ptr5>>2] = 519157;
 (_luaO_pushfstring($21,519114,$vararg_buffer2)|0);
 $24 = HEAP32[$S>>2]|0;
 _luaD_throw($24,3);
 $25 = (_luaS_newlstr($L,519168,2)|0);
 $26 = (_LoadFunction($S,$25)|0);
 STACKTOP = sp;return ($26|0);
}
function _LoadFunction($S,$p) {
 $S = $S|0;
 $p = $p|0;
 var $$0$i = 0, $$0$i$i = 0, $$0$i$i31 = 0, $$0$i18$i = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i20 = 0, $$pre$i$i29 = 0, $$pre$i10$i = 0, $$pre$i11$i = 0, $$pre$i15$i = 0, $$pre$i16 = 0, $$pre$i24 = 0, $$pre$i3 = 0, $$pre$i3$i = 0, $$pre$i7$i = 0, $$pre$phi$i22Z2D = 0, $$pre$phi$iZ2D = 0;
 var $$pre38$i = 0, $$pre50 = 0, $$pre51 = 0, $$pre52 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0.0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0;
 var $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0;
 var $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0;
 var $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0;
 var $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0;
 var $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0;
 var $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0;
 var $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0;
 var $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0;
 var $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0;
 var $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $exitcond$i = 0, $exitcond35$i = 0, $exitcond35$i37 = 0, $exitcond36$i = 0, $exitcond37$i = 0, $exitcond47 = 0, $exitcond4753 = 0, $exitcond48 = 0, $exitcond49 = 0, $i$0$i44 = 0, $i$032$i = 0;
 var $i$1$i41 = 0, $i$129$i = 0, $i$3$i38 = 0, $i$325$i = 0, $p$ = 0, $size$i = 0, $size$i$i = 0, $size$i$i25 = 0, $size$i17$i = 0, $vararg_buffer = 0, $vararg_buffer10 = 0, $vararg_buffer102 = 0, $vararg_buffer106 = 0, $vararg_buffer110 = 0, $vararg_buffer114 = 0, $vararg_buffer118 = 0, $vararg_buffer122 = 0, $vararg_buffer126 = 0, $vararg_buffer130 = 0, $vararg_buffer134 = 0;
 var $vararg_buffer138 = 0, $vararg_buffer14 = 0, $vararg_buffer142 = 0, $vararg_buffer146 = 0, $vararg_buffer150 = 0, $vararg_buffer154 = 0, $vararg_buffer18 = 0, $vararg_buffer2 = 0, $vararg_buffer22 = 0, $vararg_buffer26 = 0, $vararg_buffer30 = 0, $vararg_buffer34 = 0, $vararg_buffer38 = 0, $vararg_buffer42 = 0, $vararg_buffer46 = 0, $vararg_buffer50 = 0, $vararg_buffer54 = 0, $vararg_buffer58 = 0, $vararg_buffer6 = 0, $vararg_buffer62 = 0;
 var $vararg_buffer66 = 0, $vararg_buffer70 = 0, $vararg_buffer74 = 0, $vararg_buffer78 = 0, $vararg_buffer82 = 0, $vararg_buffer86 = 0, $vararg_buffer90 = 0, $vararg_buffer94 = 0, $vararg_buffer98 = 0, $vararg_ptr1 = 0, $vararg_ptr101 = 0, $vararg_ptr105 = 0, $vararg_ptr109 = 0, $vararg_ptr113 = 0, $vararg_ptr117 = 0, $vararg_ptr121 = 0, $vararg_ptr125 = 0, $vararg_ptr129 = 0, $vararg_ptr13 = 0, $vararg_ptr133 = 0;
 var $vararg_ptr137 = 0, $vararg_ptr141 = 0, $vararg_ptr145 = 0, $vararg_ptr149 = 0, $vararg_ptr153 = 0, $vararg_ptr157 = 0, $vararg_ptr17 = 0, $vararg_ptr21 = 0, $vararg_ptr25 = 0, $vararg_ptr29 = 0, $vararg_ptr33 = 0, $vararg_ptr37 = 0, $vararg_ptr41 = 0, $vararg_ptr45 = 0, $vararg_ptr49 = 0, $vararg_ptr5 = 0, $vararg_ptr53 = 0, $vararg_ptr57 = 0, $vararg_ptr61 = 0, $vararg_ptr65 = 0;
 var $vararg_ptr69 = 0, $vararg_ptr73 = 0, $vararg_ptr77 = 0, $vararg_ptr81 = 0, $vararg_ptr85 = 0, $vararg_ptr89 = 0, $vararg_ptr9 = 0, $vararg_ptr93 = 0, $vararg_ptr97 = 0, $x$i = 0, $x$i$i = 0, $x$i$i18 = 0, $x$i$i27 = 0, $x$i1 = 0, $x$i1$i = 0, $x$i1$i26 = 0, $x$i10 = 0, $x$i13 = 0, $x$i13$i = 0, $x$i3$i = 0;
 var $x$i5 = 0, $x$i5$i = 0, $x$i6$i = 0, $x$i7 = 0, $x$i8$i = 0, $x$i9$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 400|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer154 = sp + 320|0;
 $vararg_buffer150 = sp + 312|0;
 $vararg_buffer146 = sp + 304|0;
 $vararg_buffer142 = sp + 296|0;
 $vararg_buffer138 = sp + 288|0;
 $vararg_buffer134 = sp + 280|0;
 $vararg_buffer130 = sp + 272|0;
 $vararg_buffer126 = sp + 264|0;
 $vararg_buffer122 = sp + 256|0;
 $vararg_buffer118 = sp + 248|0;
 $vararg_buffer114 = sp + 240|0;
 $vararg_buffer110 = sp + 232|0;
 $vararg_buffer106 = sp + 224|0;
 $vararg_buffer102 = sp + 216|0;
 $vararg_buffer98 = sp + 208|0;
 $vararg_buffer94 = sp + 200|0;
 $vararg_buffer90 = sp + 192|0;
 $vararg_buffer86 = sp + 184|0;
 $vararg_buffer82 = sp + 176|0;
 $vararg_buffer78 = sp + 168|0;
 $vararg_buffer74 = sp + 160|0;
 $vararg_buffer70 = sp + 152|0;
 $vararg_buffer66 = sp + 144|0;
 $vararg_buffer62 = sp + 136|0;
 $vararg_buffer58 = sp + 128|0;
 $vararg_buffer54 = sp + 120|0;
 $vararg_buffer50 = sp + 112|0;
 $vararg_buffer46 = sp + 104|0;
 $vararg_buffer42 = sp + 96|0;
 $vararg_buffer38 = sp + 88|0;
 $vararg_buffer34 = sp + 80|0;
 $vararg_buffer30 = sp + 72|0;
 $vararg_buffer26 = sp + 64|0;
 $vararg_buffer22 = sp + 56|0;
 $vararg_buffer18 = sp + 48|0;
 $vararg_buffer14 = sp + 40|0;
 $vararg_buffer10 = sp + 32|0;
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer2 = sp + 16|0;
 $vararg_buffer = sp + 8|0;
 $x$i8$i = sp + 380|0;
 $size$i$i25 = sp + 376|0;
 $x$i6$i = sp;
 $x$i3$i = sp + 389|0;
 $x$i1$i26 = sp + 388|0;
 $x$i$i27 = sp + 372|0;
 $size$i17$i = sp + 368|0;
 $x$i13$i = sp + 364|0;
 $x$i9$i = sp + 360|0;
 $x$i5$i = sp + 356|0;
 $size$i$i = sp + 352|0;
 $x$i1$i = sp + 348|0;
 $x$i$i18 = sp + 344|0;
 $x$i$i = sp + 340|0;
 $x$i13 = sp + 387|0;
 $x$i10 = sp + 386|0;
 $x$i7 = sp + 385|0;
 $x$i5 = sp + 384|0;
 $x$i1 = sp + 336|0;
 $x$i = sp + 332|0;
 $size$i = sp + 328|0;
 $0 = HEAP32[$S>>2]|0;
 $1 = ((($0)) + 52|0);
 $2 = HEAP16[$1>>1]|0;
 $3 = (($2) + 1)<<16>>16;
 HEAP16[$1>>1] = $3;
 $4 = ($3&65535)>(200);
 if ($4) {
  $5 = ((($S)) + 12|0);
  $6 = HEAP32[$5>>2]|0;
  HEAP32[$vararg_buffer>>2] = $6;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = 519171;
  (_luaO_pushfstring($0,519114,$vararg_buffer)|0);
  $7 = HEAP32[$S>>2]|0;
  _luaD_throw($7,3);
  $$pre = HEAP32[$S>>2]|0;
  $8 = $$pre;
 } else {
  $8 = $0;
 }
 $9 = (_luaF_newproto($8)|0);
 $10 = HEAP32[$S>>2]|0;
 $11 = ((($10)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 HEAP32[$12>>2] = $9;
 $13 = ((($12)) + 8|0);
 HEAP32[$13>>2] = 9;
 $14 = HEAP32[$S>>2]|0;
 $15 = ((($14)) + 28|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($14)) + 8|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = (($16) - ($18))|0;
 $20 = ($19|0)<(17);
 $21 = $18;
 if ($20) {
  _luaD_growstack($14,1);
  $$pre50 = HEAP32[$S>>2]|0;
  $$phi$trans$insert = ((($$pre50)) + 8|0);
  $$pre51 = HEAP32[$$phi$trans$insert>>2]|0;
  $23 = $$pre50;$25 = $$pre51;
 } else {
  $23 = $14;$25 = $21;
 }
 $22 = ((($23)) + 8|0);
 $24 = ((($25)) + 16|0);
 HEAP32[$22>>2] = $24;
 $26 = ((($S)) + 4|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = (_luaZ_read($27,$size$i,4)|0);
 $29 = ($28|0)==(0);
 if (!($29)) {
  $30 = HEAP32[$S>>2]|0;
  $31 = ((($S)) + 12|0);
  $32 = HEAP32[$31>>2]|0;
  HEAP32[$vararg_buffer2>>2] = $32;
  $vararg_ptr5 = ((($vararg_buffer2)) + 4|0);
  HEAP32[$vararg_ptr5>>2] = 519142;
  (_luaO_pushfstring($30,519114,$vararg_buffer2)|0);
  $33 = HEAP32[$S>>2]|0;
  _luaD_throw($33,3);
 }
 $34 = HEAP32[$size$i>>2]|0;
 $35 = ($34|0)==(0);
 if ($35) {
  $$0$i = 0;
 } else {
  $36 = HEAP32[$S>>2]|0;
  $37 = ((($S)) + 8|0);
  $38 = HEAP32[$37>>2]|0;
  $39 = (_luaZ_openspace($36,$38,$34)|0);
  $40 = HEAP32[$size$i>>2]|0;
  $41 = HEAP32[$26>>2]|0;
  $42 = (_luaZ_read($41,$39,$40)|0);
  $43 = ($42|0)==(0);
  if (!($43)) {
   $44 = HEAP32[$S>>2]|0;
   $45 = ((($S)) + 12|0);
   $46 = HEAP32[$45>>2]|0;
   HEAP32[$vararg_buffer6>>2] = $46;
   $vararg_ptr9 = ((($vararg_buffer6)) + 4|0);
   HEAP32[$vararg_ptr9>>2] = 519142;
   (_luaO_pushfstring($44,519114,$vararg_buffer6)|0);
   $47 = HEAP32[$S>>2]|0;
   _luaD_throw($47,3);
  }
  $48 = HEAP32[$S>>2]|0;
  $49 = HEAP32[$size$i>>2]|0;
  $50 = (($49) + -1)|0;
  $51 = (_luaS_newlstr($48,$39,$50)|0);
  $$0$i = $51;
 }
 $52 = ((($9)) + 32|0);
 $53 = ($$0$i|0)==(0|0);
 $p$ = $53 ? $p : $$0$i;
 HEAP32[$52>>2] = $p$;
 $54 = HEAP32[$26>>2]|0;
 $55 = (_luaZ_read($54,$x$i,4)|0);
 $56 = ($55|0)==(0);
 if (!($56)) {
  $57 = HEAP32[$S>>2]|0;
  $58 = ((($S)) + 12|0);
  $59 = HEAP32[$58>>2]|0;
  HEAP32[$vararg_buffer10>>2] = $59;
  $vararg_ptr13 = ((($vararg_buffer10)) + 4|0);
  HEAP32[$vararg_ptr13>>2] = 519142;
  (_luaO_pushfstring($57,519114,$vararg_buffer10)|0);
  $60 = HEAP32[$S>>2]|0;
  _luaD_throw($60,3);
 }
 $61 = HEAP32[$x$i>>2]|0;
 $62 = ($61|0)<(0);
 if ($62) {
  $63 = HEAP32[$S>>2]|0;
  $64 = ((($S)) + 12|0);
  $65 = HEAP32[$64>>2]|0;
  HEAP32[$vararg_buffer14>>2] = $65;
  $vararg_ptr17 = ((($vararg_buffer14)) + 4|0);
  HEAP32[$vararg_ptr17>>2] = 519185;
  (_luaO_pushfstring($63,519114,$vararg_buffer14)|0);
  $66 = HEAP32[$S>>2]|0;
  _luaD_throw($66,3);
  $$pre$i = HEAP32[$x$i>>2]|0;
  $68 = $$pre$i;
 } else {
  $68 = $61;
 }
 $67 = ((($9)) + 60|0);
 HEAP32[$67>>2] = $68;
 $69 = HEAP32[$26>>2]|0;
 $70 = (_luaZ_read($69,$x$i1,4)|0);
 $71 = ($70|0)==(0);
 if (!($71)) {
  $72 = HEAP32[$S>>2]|0;
  $73 = ((($S)) + 12|0);
  $74 = HEAP32[$73>>2]|0;
  HEAP32[$vararg_buffer18>>2] = $74;
  $vararg_ptr21 = ((($vararg_buffer18)) + 4|0);
  HEAP32[$vararg_ptr21>>2] = 519142;
  (_luaO_pushfstring($72,519114,$vararg_buffer18)|0);
  $75 = HEAP32[$S>>2]|0;
  _luaD_throw($75,3);
 }
 $76 = HEAP32[$x$i1>>2]|0;
 $77 = ($76|0)<(0);
 if ($77) {
  $78 = HEAP32[$S>>2]|0;
  $79 = ((($S)) + 12|0);
  $80 = HEAP32[$79>>2]|0;
  HEAP32[$vararg_buffer22>>2] = $80;
  $vararg_ptr25 = ((($vararg_buffer22)) + 4|0);
  HEAP32[$vararg_ptr25>>2] = 519185;
  (_luaO_pushfstring($78,519114,$vararg_buffer22)|0);
  $81 = HEAP32[$S>>2]|0;
  _luaD_throw($81,3);
  $$pre$i3 = HEAP32[$x$i1>>2]|0;
  $83 = $$pre$i3;
 } else {
  $83 = $76;
 }
 $82 = ((($9)) + 64|0);
 HEAP32[$82>>2] = $83;
 $84 = HEAP32[$26>>2]|0;
 $85 = (_luaZ_read($84,$x$i5,1)|0);
 $86 = ($85|0)==(0);
 if (!($86)) {
  $87 = HEAP32[$S>>2]|0;
  $88 = ((($S)) + 12|0);
  $89 = HEAP32[$88>>2]|0;
  HEAP32[$vararg_buffer26>>2] = $89;
  $vararg_ptr29 = ((($vararg_buffer26)) + 4|0);
  HEAP32[$vararg_ptr29>>2] = 519142;
  (_luaO_pushfstring($87,519114,$vararg_buffer26)|0);
  $90 = HEAP32[$S>>2]|0;
  _luaD_throw($90,3);
 }
 $91 = HEAP8[$x$i5>>0]|0;
 $92 = ((($9)) + 72|0);
 HEAP8[$92>>0] = $91;
 $93 = HEAP32[$26>>2]|0;
 $94 = (_luaZ_read($93,$x$i7,1)|0);
 $95 = ($94|0)==(0);
 if (!($95)) {
  $96 = HEAP32[$S>>2]|0;
  $97 = ((($S)) + 12|0);
  $98 = HEAP32[$97>>2]|0;
  HEAP32[$vararg_buffer30>>2] = $98;
  $vararg_ptr33 = ((($vararg_buffer30)) + 4|0);
  HEAP32[$vararg_ptr33>>2] = 519142;
  (_luaO_pushfstring($96,519114,$vararg_buffer30)|0);
  $99 = HEAP32[$S>>2]|0;
  _luaD_throw($99,3);
 }
 $100 = HEAP8[$x$i7>>0]|0;
 $101 = ((($9)) + 73|0);
 HEAP8[$101>>0] = $100;
 $102 = HEAP32[$26>>2]|0;
 $103 = (_luaZ_read($102,$x$i10,1)|0);
 $104 = ($103|0)==(0);
 if (!($104)) {
  $105 = HEAP32[$S>>2]|0;
  $106 = ((($S)) + 12|0);
  $107 = HEAP32[$106>>2]|0;
  HEAP32[$vararg_buffer34>>2] = $107;
  $vararg_ptr37 = ((($vararg_buffer34)) + 4|0);
  HEAP32[$vararg_ptr37>>2] = 519142;
  (_luaO_pushfstring($105,519114,$vararg_buffer34)|0);
  $108 = HEAP32[$S>>2]|0;
  _luaD_throw($108,3);
 }
 $109 = HEAP8[$x$i10>>0]|0;
 $110 = ((($9)) + 74|0);
 HEAP8[$110>>0] = $109;
 $111 = HEAP32[$26>>2]|0;
 $112 = (_luaZ_read($111,$x$i13,1)|0);
 $113 = ($112|0)==(0);
 if (!($113)) {
  $114 = HEAP32[$S>>2]|0;
  $115 = ((($S)) + 12|0);
  $116 = HEAP32[$115>>2]|0;
  HEAP32[$vararg_buffer38>>2] = $116;
  $vararg_ptr41 = ((($vararg_buffer38)) + 4|0);
  HEAP32[$vararg_ptr41>>2] = 519142;
  (_luaO_pushfstring($114,519114,$vararg_buffer38)|0);
  $117 = HEAP32[$S>>2]|0;
  _luaD_throw($117,3);
 }
 $118 = HEAP8[$x$i13>>0]|0;
 $119 = ((($9)) + 75|0);
 HEAP8[$119>>0] = $118;
 $120 = HEAP32[$26>>2]|0;
 $121 = (_luaZ_read($120,$x$i$i,4)|0);
 $122 = ($121|0)==(0);
 if (!($122)) {
  $123 = HEAP32[$S>>2]|0;
  $124 = ((($S)) + 12|0);
  $125 = HEAP32[$124>>2]|0;
  HEAP32[$vararg_buffer42>>2] = $125;
  $vararg_ptr45 = ((($vararg_buffer42)) + 4|0);
  HEAP32[$vararg_ptr45>>2] = 519142;
  (_luaO_pushfstring($123,519114,$vararg_buffer42)|0);
  $126 = HEAP32[$S>>2]|0;
  _luaD_throw($126,3);
 }
 $127 = HEAP32[$x$i$i>>2]|0;
 $128 = ($127|0)<(0);
 if ($128) {
  $129 = HEAP32[$S>>2]|0;
  $130 = ((($S)) + 12|0);
  $131 = HEAP32[$130>>2]|0;
  HEAP32[$vararg_buffer46>>2] = $131;
  $vararg_ptr49 = ((($vararg_buffer46)) + 4|0);
  HEAP32[$vararg_ptr49>>2] = 519185;
  (_luaO_pushfstring($129,519114,$vararg_buffer46)|0);
  $132 = HEAP32[$S>>2]|0;
  _luaD_throw($132,3);
  $$pre$i$i = HEAP32[$x$i$i>>2]|0;
  $134 = $$pre$i$i;
 } else {
  $134 = $127;
 }
 $133 = (($134) + 1)|0;
 $135 = ($133>>>0)<(1073741824);
 $136 = HEAP32[$S>>2]|0;
 if ($135) {
  $137 = $134 << 2;
  $138 = (_luaM_realloc_($136,0,0,$137)|0);
  $$pre$phi$iZ2D = $137;$141 = $138;
 } else {
  $139 = (_luaM_toobig($136)|0);
  $$pre$i16 = $134 << 2;
  $$pre$phi$iZ2D = $$pre$i16;$141 = $139;
 }
 $140 = ((($9)) + 12|0);
 HEAP32[$140>>2] = $141;
 $142 = ((($9)) + 44|0);
 HEAP32[$142>>2] = $134;
 $143 = HEAP32[$26>>2]|0;
 $144 = (_luaZ_read($143,$141,$$pre$phi$iZ2D)|0);
 $145 = ($144|0)==(0);
 if (!($145)) {
  $146 = HEAP32[$S>>2]|0;
  $147 = ((($S)) + 12|0);
  $148 = HEAP32[$147>>2]|0;
  HEAP32[$vararg_buffer50>>2] = $148;
  $vararg_ptr53 = ((($vararg_buffer50)) + 4|0);
  HEAP32[$vararg_ptr53>>2] = 519142;
  (_luaO_pushfstring($146,519114,$vararg_buffer50)|0);
  $149 = HEAP32[$S>>2]|0;
  _luaD_throw($149,3);
 }
 $150 = HEAP32[$26>>2]|0;
 $151 = (_luaZ_read($150,$x$i$i27,4)|0);
 $152 = ($151|0)==(0);
 if (!($152)) {
  $153 = HEAP32[$S>>2]|0;
  $154 = ((($S)) + 12|0);
  $155 = HEAP32[$154>>2]|0;
  HEAP32[$vararg_buffer54>>2] = $155;
  $vararg_ptr57 = ((($vararg_buffer54)) + 4|0);
  HEAP32[$vararg_ptr57>>2] = 519142;
  (_luaO_pushfstring($153,519114,$vararg_buffer54)|0);
  $156 = HEAP32[$S>>2]|0;
  _luaD_throw($156,3);
 }
 $157 = HEAP32[$x$i$i27>>2]|0;
 $158 = ($157|0)<(0);
 if ($158) {
  $159 = HEAP32[$S>>2]|0;
  $160 = ((($S)) + 12|0);
  $161 = HEAP32[$160>>2]|0;
  HEAP32[$vararg_buffer58>>2] = $161;
  $vararg_ptr61 = ((($vararg_buffer58)) + 4|0);
  HEAP32[$vararg_ptr61>>2] = 519185;
  (_luaO_pushfstring($159,519114,$vararg_buffer58)|0);
  $162 = HEAP32[$S>>2]|0;
  _luaD_throw($162,3);
  $$pre$i$i29 = HEAP32[$x$i$i27>>2]|0;
  $164 = $$pre$i$i29;
 } else {
  $164 = $157;
 }
 $163 = (($164) + 1)|0;
 $165 = ($163>>>0)<(268435456);
 $166 = HEAP32[$S>>2]|0;
 if ($165) {
  $167 = $164 << 4;
  $168 = (_luaM_realloc_($166,0,0,$167)|0);
  $171 = $168;
 } else {
  $169 = (_luaM_toobig($166)|0);
  $171 = $169;
 }
 $170 = ((($9)) + 8|0);
 HEAP32[$170>>2] = $171;
 $172 = ((($9)) + 40|0);
 HEAP32[$172>>2] = $164;
 $173 = ($164|0)>(0);
 L63: do {
  if ($173) {
   $i$0$i44 = 0;
   while(1) {
    $176 = (((($171) + ($i$0$i44<<4)|0)) + 8|0);
    HEAP32[$176>>2] = 0;
    $177 = (($i$0$i44) + 1)|0;
    $exitcond49 = ($177|0)==($164|0);
    if ($exitcond49) {
     break;
    } else {
     $i$0$i44 = $177;
    }
   }
   if (!($173)) {
    break;
   }
   $174 = ((($S)) + 12|0);
   $175 = ((($S)) + 8|0);
   $i$1$i41 = 0;
   while(1) {
    $178 = HEAP32[$170>>2]|0;
    $179 = (($178) + ($i$1$i41<<4)|0);
    $180 = HEAP32[$26>>2]|0;
    $181 = (_luaZ_read($180,$x$i1$i26,1)|0);
    $182 = ($181|0)==(0);
    if (!($182)) {
     $183 = HEAP32[$S>>2]|0;
     $184 = HEAP32[$174>>2]|0;
     HEAP32[$vararg_buffer62>>2] = $184;
     $vararg_ptr65 = ((($vararg_buffer62)) + 4|0);
     HEAP32[$vararg_ptr65>>2] = 519142;
     (_luaO_pushfstring($183,519114,$vararg_buffer62)|0);
     $185 = HEAP32[$S>>2]|0;
     _luaD_throw($185,3);
    }
    $186 = HEAP8[$x$i1$i26>>0]|0;
    $187 = $186 << 24 >> 24;
    switch ($187|0) {
    case 0:  {
     $188 = (((($178) + ($i$1$i41<<4)|0)) + 8|0);
     HEAP32[$188>>2] = 0;
     break;
    }
    case 1:  {
     $189 = HEAP32[$26>>2]|0;
     $190 = (_luaZ_read($189,$x$i3$i,1)|0);
     $191 = ($190|0)==(0);
     if (!($191)) {
      $192 = HEAP32[$S>>2]|0;
      $193 = HEAP32[$174>>2]|0;
      HEAP32[$vararg_buffer66>>2] = $193;
      $vararg_ptr69 = ((($vararg_buffer66)) + 4|0);
      HEAP32[$vararg_ptr69>>2] = 519142;
      (_luaO_pushfstring($192,519114,$vararg_buffer66)|0);
      $194 = HEAP32[$S>>2]|0;
      _luaD_throw($194,3);
     }
     $195 = HEAP8[$x$i3$i>>0]|0;
     $196 = ($195<<24>>24)!=(0);
     $197 = $196&1;
     HEAP32[$179>>2] = $197;
     $198 = (((($178) + ($i$1$i41<<4)|0)) + 8|0);
     HEAP32[$198>>2] = 1;
     break;
    }
    case 3:  {
     $199 = HEAP32[$26>>2]|0;
     $200 = (_luaZ_read($199,$x$i6$i,8)|0);
     $201 = ($200|0)==(0);
     if (!($201)) {
      $202 = HEAP32[$S>>2]|0;
      $203 = HEAP32[$174>>2]|0;
      HEAP32[$vararg_buffer70>>2] = $203;
      $vararg_ptr73 = ((($vararg_buffer70)) + 4|0);
      HEAP32[$vararg_ptr73>>2] = 519142;
      (_luaO_pushfstring($202,519114,$vararg_buffer70)|0);
      $204 = HEAP32[$S>>2]|0;
      _luaD_throw($204,3);
     }
     $205 = +HEAPF64[$x$i6$i>>3];
     HEAPF64[$179>>3] = $205;
     $206 = (((($178) + ($i$1$i41<<4)|0)) + 8|0);
     HEAP32[$206>>2] = 3;
     break;
    }
    case 4:  {
     $207 = HEAP32[$26>>2]|0;
     $208 = (_luaZ_read($207,$size$i$i25,4)|0);
     $209 = ($208|0)==(0);
     if (!($209)) {
      $210 = HEAP32[$S>>2]|0;
      $211 = HEAP32[$174>>2]|0;
      HEAP32[$vararg_buffer74>>2] = $211;
      $vararg_ptr77 = ((($vararg_buffer74)) + 4|0);
      HEAP32[$vararg_ptr77>>2] = 519142;
      (_luaO_pushfstring($210,519114,$vararg_buffer74)|0);
      $212 = HEAP32[$S>>2]|0;
      _luaD_throw($212,3);
     }
     $213 = HEAP32[$size$i$i25>>2]|0;
     $214 = ($213|0)==(0);
     if ($214) {
      $$0$i$i31 = 0;
     } else {
      $215 = HEAP32[$S>>2]|0;
      $216 = HEAP32[$175>>2]|0;
      $217 = (_luaZ_openspace($215,$216,$213)|0);
      $218 = HEAP32[$size$i$i25>>2]|0;
      $219 = HEAP32[$26>>2]|0;
      $220 = (_luaZ_read($219,$217,$218)|0);
      $221 = ($220|0)==(0);
      if (!($221)) {
       $222 = HEAP32[$S>>2]|0;
       $223 = HEAP32[$174>>2]|0;
       HEAP32[$vararg_buffer78>>2] = $223;
       $vararg_ptr81 = ((($vararg_buffer78)) + 4|0);
       HEAP32[$vararg_ptr81>>2] = 519142;
       (_luaO_pushfstring($222,519114,$vararg_buffer78)|0);
       $224 = HEAP32[$S>>2]|0;
       _luaD_throw($224,3);
      }
      $225 = HEAP32[$S>>2]|0;
      $226 = HEAP32[$size$i$i25>>2]|0;
      $227 = (($226) + -1)|0;
      $228 = (_luaS_newlstr($225,$217,$227)|0);
      $$0$i$i31 = $228;
     }
     HEAP32[$179>>2] = $$0$i$i31;
     $229 = (((($178) + ($i$1$i41<<4)|0)) + 8|0);
     HEAP32[$229>>2] = 4;
     break;
    }
    default: {
     $230 = HEAP32[$S>>2]|0;
     $231 = HEAP32[$174>>2]|0;
     HEAP32[$vararg_buffer82>>2] = $231;
     $vararg_ptr85 = ((($vararg_buffer82)) + 4|0);
     HEAP32[$vararg_ptr85>>2] = 519197;
     (_luaO_pushfstring($230,519114,$vararg_buffer82)|0);
     $232 = HEAP32[$S>>2]|0;
     _luaD_throw($232,3);
    }
    }
    $233 = (($i$1$i41) + 1)|0;
    $exitcond48 = ($233|0)==($164|0);
    if ($exitcond48) {
     break L63;
    } else {
     $i$1$i41 = $233;
    }
   }
  }
 } while(0);
 $234 = HEAP32[$26>>2]|0;
 $235 = (_luaZ_read($234,$x$i8$i,4)|0);
 $236 = ($235|0)==(0);
 if (!($236)) {
  $237 = HEAP32[$S>>2]|0;
  $238 = ((($S)) + 12|0);
  $239 = HEAP32[$238>>2]|0;
  HEAP32[$vararg_buffer86>>2] = $239;
  $vararg_ptr89 = ((($vararg_buffer86)) + 4|0);
  HEAP32[$vararg_ptr89>>2] = 519142;
  (_luaO_pushfstring($237,519114,$vararg_buffer86)|0);
  $240 = HEAP32[$S>>2]|0;
  _luaD_throw($240,3);
 }
 $241 = HEAP32[$x$i8$i>>2]|0;
 $242 = ($241|0)<(0);
 if ($242) {
  $243 = HEAP32[$S>>2]|0;
  $244 = ((($S)) + 12|0);
  $245 = HEAP32[$244>>2]|0;
  HEAP32[$vararg_buffer90>>2] = $245;
  $vararg_ptr93 = ((($vararg_buffer90)) + 4|0);
  HEAP32[$vararg_ptr93>>2] = 519185;
  (_luaO_pushfstring($243,519114,$vararg_buffer90)|0);
  $246 = HEAP32[$S>>2]|0;
  _luaD_throw($246,3);
  $$pre$i10$i = HEAP32[$x$i8$i>>2]|0;
  $248 = $$pre$i10$i;
 } else {
  $248 = $241;
 }
 $247 = (($248) + 1)|0;
 $249 = ($247>>>0)<(1073741824);
 $250 = HEAP32[$S>>2]|0;
 if ($249) {
  $251 = $248 << 2;
  $252 = (_luaM_realloc_($250,0,0,$251)|0);
  $255 = $252;
 } else {
  $253 = (_luaM_toobig($250)|0);
  $255 = $253;
 }
 $254 = ((($9)) + 16|0);
 HEAP32[$254>>2] = $255;
 $256 = ((($9)) + 52|0);
 HEAP32[$256>>2] = $248;
 $257 = ($248|0)>(0);
 do {
  if ($257) {
   HEAP32[$255>>2] = 0;
   $exitcond4753 = ($248|0)==(1);
   if (!($exitcond4753)) {
    $259 = 1;
    while(1) {
     $$pre52 = HEAP32[$254>>2]|0;
     $258 = (($$pre52) + ($259<<2)|0);
     HEAP32[$258>>2] = 0;
     $260 = (($259) + 1)|0;
     $exitcond47 = ($260|0)==($248|0);
     if ($exitcond47) {
      break;
     } else {
      $259 = $260;
     }
    }
   }
   if ($257) {
    $i$3$i38 = 0;
   } else {
    break;
   }
   while(1) {
    $261 = HEAP32[$52>>2]|0;
    $262 = (_LoadFunction($S,$261)|0);
    $263 = HEAP32[$254>>2]|0;
    $264 = (($263) + ($i$3$i38<<2)|0);
    HEAP32[$264>>2] = $262;
    $265 = (($i$3$i38) + 1)|0;
    $exitcond = ($265|0)==($248|0);
    if ($exitcond) {
     break;
    } else {
     $i$3$i38 = $265;
    }
   }
  }
 } while(0);
 $266 = HEAP32[$26>>2]|0;
 $267 = (_luaZ_read($266,$x$i$i18,4)|0);
 $268 = ($267|0)==(0);
 if (!($268)) {
  $269 = HEAP32[$S>>2]|0;
  $270 = ((($S)) + 12|0);
  $271 = HEAP32[$270>>2]|0;
  HEAP32[$vararg_buffer94>>2] = $271;
  $vararg_ptr97 = ((($vararg_buffer94)) + 4|0);
  HEAP32[$vararg_ptr97>>2] = 519142;
  (_luaO_pushfstring($269,519114,$vararg_buffer94)|0);
  $272 = HEAP32[$S>>2]|0;
  _luaD_throw($272,3);
 }
 $273 = HEAP32[$x$i$i18>>2]|0;
 $274 = ($273|0)<(0);
 if ($274) {
  $275 = HEAP32[$S>>2]|0;
  $276 = ((($S)) + 12|0);
  $277 = HEAP32[$276>>2]|0;
  HEAP32[$vararg_buffer98>>2] = $277;
  $vararg_ptr101 = ((($vararg_buffer98)) + 4|0);
  HEAP32[$vararg_ptr101>>2] = 519185;
  (_luaO_pushfstring($275,519114,$vararg_buffer98)|0);
  $278 = HEAP32[$S>>2]|0;
  _luaD_throw($278,3);
  $$pre$i$i20 = HEAP32[$x$i$i18>>2]|0;
  $280 = $$pre$i$i20;
 } else {
  $280 = $273;
 }
 $279 = (($280) + 1)|0;
 $281 = ($279>>>0)<(1073741824);
 $282 = HEAP32[$S>>2]|0;
 if ($281) {
  $283 = $280 << 2;
  $284 = (_luaM_realloc_($282,0,0,$283)|0);
  $$pre$phi$i22Z2D = $283;$287 = $284;
 } else {
  $285 = (_luaM_toobig($282)|0);
  $$pre38$i = $280 << 2;
  $$pre$phi$i22Z2D = $$pre38$i;$287 = $285;
 }
 $286 = ((($9)) + 20|0);
 HEAP32[$286>>2] = $287;
 $288 = ((($9)) + 48|0);
 HEAP32[$288>>2] = $280;
 $289 = HEAP32[$26>>2]|0;
 $290 = (_luaZ_read($289,$287,$$pre$phi$i22Z2D)|0);
 $291 = ($290|0)==(0);
 if (!($291)) {
  $292 = HEAP32[$S>>2]|0;
  $293 = ((($S)) + 12|0);
  $294 = HEAP32[$293>>2]|0;
  HEAP32[$vararg_buffer102>>2] = $294;
  $vararg_ptr105 = ((($vararg_buffer102)) + 4|0);
  HEAP32[$vararg_ptr105>>2] = 519142;
  (_luaO_pushfstring($292,519114,$vararg_buffer102)|0);
  $295 = HEAP32[$S>>2]|0;
  _luaD_throw($295,3);
 }
 $296 = HEAP32[$26>>2]|0;
 $297 = (_luaZ_read($296,$x$i1$i,4)|0);
 $298 = ($297|0)==(0);
 if (!($298)) {
  $299 = HEAP32[$S>>2]|0;
  $300 = ((($S)) + 12|0);
  $301 = HEAP32[$300>>2]|0;
  HEAP32[$vararg_buffer106>>2] = $301;
  $vararg_ptr109 = ((($vararg_buffer106)) + 4|0);
  HEAP32[$vararg_ptr109>>2] = 519142;
  (_luaO_pushfstring($299,519114,$vararg_buffer106)|0);
  $302 = HEAP32[$S>>2]|0;
  _luaD_throw($302,3);
 }
 $303 = HEAP32[$x$i1$i>>2]|0;
 $304 = ($303|0)<(0);
 if ($304) {
  $305 = HEAP32[$S>>2]|0;
  $306 = ((($S)) + 12|0);
  $307 = HEAP32[$306>>2]|0;
  HEAP32[$vararg_buffer110>>2] = $307;
  $vararg_ptr113 = ((($vararg_buffer110)) + 4|0);
  HEAP32[$vararg_ptr113>>2] = 519185;
  (_luaO_pushfstring($305,519114,$vararg_buffer110)|0);
  $308 = HEAP32[$S>>2]|0;
  _luaD_throw($308,3);
  $$pre$i3$i = HEAP32[$x$i1$i>>2]|0;
  $310 = $$pre$i3$i;
 } else {
  $310 = $303;
 }
 $309 = (($310) + 1)|0;
 $311 = ($309>>>0)<(357913942);
 $312 = HEAP32[$S>>2]|0;
 if ($311) {
  $313 = ($310*12)|0;
  $314 = (_luaM_realloc_($312,0,0,$313)|0);
  $317 = $314;
 } else {
  $315 = (_luaM_toobig($312)|0);
  $317 = $315;
 }
 $316 = ((($9)) + 24|0);
 HEAP32[$316>>2] = $317;
 $318 = ((($9)) + 56|0);
 HEAP32[$318>>2] = $310;
 $319 = ($310|0)>(0);
 if ($319) {
  $i$032$i = 0;
  while(1) {
   $322 = (($317) + (($i$032$i*12)|0)|0);
   HEAP32[$322>>2] = 0;
   $323 = (($i$032$i) + 1)|0;
   $exitcond37$i = ($323|0)==($310|0);
   if ($exitcond37$i) {
    break;
   } else {
    $i$032$i = $323;
   }
  }
  $320 = ((($S)) + 12|0);
  $321 = ((($S)) + 8|0);
  $i$129$i = 0;
  while(1) {
   $324 = HEAP32[$26>>2]|0;
   $325 = (_luaZ_read($324,$size$i$i,4)|0);
   $326 = ($325|0)==(0);
   if (!($326)) {
    $327 = HEAP32[$S>>2]|0;
    $328 = HEAP32[$320>>2]|0;
    HEAP32[$vararg_buffer114>>2] = $328;
    $vararg_ptr117 = ((($vararg_buffer114)) + 4|0);
    HEAP32[$vararg_ptr117>>2] = 519142;
    (_luaO_pushfstring($327,519114,$vararg_buffer114)|0);
    $329 = HEAP32[$S>>2]|0;
    _luaD_throw($329,3);
   }
   $330 = HEAP32[$size$i$i>>2]|0;
   $331 = ($330|0)==(0);
   if ($331) {
    $$0$i$i = 0;
   } else {
    $332 = HEAP32[$S>>2]|0;
    $333 = HEAP32[$321>>2]|0;
    $334 = (_luaZ_openspace($332,$333,$330)|0);
    $335 = HEAP32[$size$i$i>>2]|0;
    $336 = HEAP32[$26>>2]|0;
    $337 = (_luaZ_read($336,$334,$335)|0);
    $338 = ($337|0)==(0);
    if (!($338)) {
     $339 = HEAP32[$S>>2]|0;
     $340 = HEAP32[$320>>2]|0;
     HEAP32[$vararg_buffer118>>2] = $340;
     $vararg_ptr121 = ((($vararg_buffer118)) + 4|0);
     HEAP32[$vararg_ptr121>>2] = 519142;
     (_luaO_pushfstring($339,519114,$vararg_buffer118)|0);
     $341 = HEAP32[$S>>2]|0;
     _luaD_throw($341,3);
    }
    $342 = HEAP32[$S>>2]|0;
    $343 = HEAP32[$size$i$i>>2]|0;
    $344 = (($343) + -1)|0;
    $345 = (_luaS_newlstr($342,$334,$344)|0);
    $$0$i$i = $345;
   }
   $346 = HEAP32[$316>>2]|0;
   $347 = (($346) + (($i$129$i*12)|0)|0);
   HEAP32[$347>>2] = $$0$i$i;
   $348 = HEAP32[$26>>2]|0;
   $349 = (_luaZ_read($348,$x$i5$i,4)|0);
   $350 = ($349|0)==(0);
   if (!($350)) {
    $351 = HEAP32[$S>>2]|0;
    $352 = HEAP32[$320>>2]|0;
    HEAP32[$vararg_buffer122>>2] = $352;
    $vararg_ptr125 = ((($vararg_buffer122)) + 4|0);
    HEAP32[$vararg_ptr125>>2] = 519142;
    (_luaO_pushfstring($351,519114,$vararg_buffer122)|0);
    $353 = HEAP32[$S>>2]|0;
    _luaD_throw($353,3);
   }
   $354 = HEAP32[$x$i5$i>>2]|0;
   $355 = ($354|0)<(0);
   if ($355) {
    $356 = HEAP32[$S>>2]|0;
    $357 = HEAP32[$320>>2]|0;
    HEAP32[$vararg_buffer126>>2] = $357;
    $vararg_ptr129 = ((($vararg_buffer126)) + 4|0);
    HEAP32[$vararg_ptr129>>2] = 519185;
    (_luaO_pushfstring($356,519114,$vararg_buffer126)|0);
    $358 = HEAP32[$S>>2]|0;
    _luaD_throw($358,3);
    $$pre$i7$i = HEAP32[$x$i5$i>>2]|0;
    $361 = $$pre$i7$i;
   } else {
    $361 = $354;
   }
   $359 = HEAP32[$316>>2]|0;
   $360 = (((($359) + (($i$129$i*12)|0)|0)) + 4|0);
   HEAP32[$360>>2] = $361;
   $362 = HEAP32[$26>>2]|0;
   $363 = (_luaZ_read($362,$x$i9$i,4)|0);
   $364 = ($363|0)==(0);
   if (!($364)) {
    $365 = HEAP32[$S>>2]|0;
    $366 = HEAP32[$320>>2]|0;
    HEAP32[$vararg_buffer130>>2] = $366;
    $vararg_ptr133 = ((($vararg_buffer130)) + 4|0);
    HEAP32[$vararg_ptr133>>2] = 519142;
    (_luaO_pushfstring($365,519114,$vararg_buffer130)|0);
    $367 = HEAP32[$S>>2]|0;
    _luaD_throw($367,3);
   }
   $368 = HEAP32[$x$i9$i>>2]|0;
   $369 = ($368|0)<(0);
   if ($369) {
    $370 = HEAP32[$S>>2]|0;
    $371 = HEAP32[$320>>2]|0;
    HEAP32[$vararg_buffer134>>2] = $371;
    $vararg_ptr137 = ((($vararg_buffer134)) + 4|0);
    HEAP32[$vararg_ptr137>>2] = 519185;
    (_luaO_pushfstring($370,519114,$vararg_buffer134)|0);
    $372 = HEAP32[$S>>2]|0;
    _luaD_throw($372,3);
    $$pre$i11$i = HEAP32[$x$i9$i>>2]|0;
    $375 = $$pre$i11$i;
   } else {
    $375 = $368;
   }
   $373 = HEAP32[$316>>2]|0;
   $374 = (((($373) + (($i$129$i*12)|0)|0)) + 8|0);
   HEAP32[$374>>2] = $375;
   $376 = (($i$129$i) + 1)|0;
   $exitcond36$i = ($376|0)==($310|0);
   if ($exitcond36$i) {
    break;
   } else {
    $i$129$i = $376;
   }
  }
 }
 $377 = HEAP32[$26>>2]|0;
 $378 = (_luaZ_read($377,$x$i13$i,4)|0);
 $379 = ($378|0)==(0);
 if (!($379)) {
  $380 = HEAP32[$S>>2]|0;
  $381 = ((($S)) + 12|0);
  $382 = HEAP32[$381>>2]|0;
  HEAP32[$vararg_buffer138>>2] = $382;
  $vararg_ptr141 = ((($vararg_buffer138)) + 4|0);
  HEAP32[$vararg_ptr141>>2] = 519142;
  (_luaO_pushfstring($380,519114,$vararg_buffer138)|0);
  $383 = HEAP32[$S>>2]|0;
  _luaD_throw($383,3);
 }
 $384 = HEAP32[$x$i13$i>>2]|0;
 $385 = ($384|0)<(0);
 if ($385) {
  $386 = HEAP32[$S>>2]|0;
  $387 = ((($S)) + 12|0);
  $388 = HEAP32[$387>>2]|0;
  HEAP32[$vararg_buffer142>>2] = $388;
  $vararg_ptr145 = ((($vararg_buffer142)) + 4|0);
  HEAP32[$vararg_ptr145>>2] = 519185;
  (_luaO_pushfstring($386,519114,$vararg_buffer142)|0);
  $389 = HEAP32[$S>>2]|0;
  _luaD_throw($389,3);
  $$pre$i15$i = HEAP32[$x$i13$i>>2]|0;
  $391 = $$pre$i15$i;
 } else {
  $391 = $384;
 }
 $390 = (($391) + 1)|0;
 $392 = ($390>>>0)<(1073741824);
 $393 = HEAP32[$S>>2]|0;
 if ($392) {
  $394 = $391 << 2;
  $395 = (_luaM_realloc_($393,0,0,$394)|0);
  $398 = $395;
 } else {
  $396 = (_luaM_toobig($393)|0);
  $398 = $396;
 }
 $397 = ((($9)) + 28|0);
 HEAP32[$397>>2] = $398;
 $399 = ((($9)) + 36|0);
 HEAP32[$399>>2] = $391;
 $400 = ($391|0)>(0);
 if ($400) {
  HEAP32[$398>>2] = 0;
  $exitcond35$i37 = ($391|0)==(1);
  if (!($exitcond35$i37)) {
   $404 = 1;
   while(1) {
    $$pre$i24 = HEAP32[$397>>2]|0;
    $403 = (($$pre$i24) + ($404<<2)|0);
    HEAP32[$403>>2] = 0;
    $405 = (($404) + 1)|0;
    $exitcond35$i = ($405|0)==($391|0);
    if ($exitcond35$i) {
     break;
    } else {
     $404 = $405;
    }
   }
  }
  $401 = ((($S)) + 8|0);
  $402 = ((($S)) + 12|0);
  $i$325$i = 0;
  while(1) {
   $406 = HEAP32[$26>>2]|0;
   $407 = (_luaZ_read($406,$size$i17$i,4)|0);
   $408 = ($407|0)==(0);
   if (!($408)) {
    $409 = HEAP32[$S>>2]|0;
    $410 = HEAP32[$402>>2]|0;
    HEAP32[$vararg_buffer146>>2] = $410;
    $vararg_ptr149 = ((($vararg_buffer146)) + 4|0);
    HEAP32[$vararg_ptr149>>2] = 519142;
    (_luaO_pushfstring($409,519114,$vararg_buffer146)|0);
    $411 = HEAP32[$S>>2]|0;
    _luaD_throw($411,3);
   }
   $412 = HEAP32[$size$i17$i>>2]|0;
   $413 = ($412|0)==(0);
   if ($413) {
    $$0$i18$i = 0;
   } else {
    $414 = HEAP32[$S>>2]|0;
    $415 = HEAP32[$401>>2]|0;
    $416 = (_luaZ_openspace($414,$415,$412)|0);
    $417 = HEAP32[$size$i17$i>>2]|0;
    $418 = HEAP32[$26>>2]|0;
    $419 = (_luaZ_read($418,$416,$417)|0);
    $420 = ($419|0)==(0);
    if (!($420)) {
     $421 = HEAP32[$S>>2]|0;
     $422 = HEAP32[$402>>2]|0;
     HEAP32[$vararg_buffer150>>2] = $422;
     $vararg_ptr153 = ((($vararg_buffer150)) + 4|0);
     HEAP32[$vararg_ptr153>>2] = 519142;
     (_luaO_pushfstring($421,519114,$vararg_buffer150)|0);
     $423 = HEAP32[$S>>2]|0;
     _luaD_throw($423,3);
    }
    $424 = HEAP32[$S>>2]|0;
    $425 = HEAP32[$size$i17$i>>2]|0;
    $426 = (($425) + -1)|0;
    $427 = (_luaS_newlstr($424,$416,$426)|0);
    $$0$i18$i = $427;
   }
   $428 = HEAP32[$397>>2]|0;
   $429 = (($428) + ($i$325$i<<2)|0);
   HEAP32[$429>>2] = $$0$i18$i;
   $430 = (($i$325$i) + 1)|0;
   $exitcond$i = ($430|0)==($391|0);
   if ($exitcond$i) {
    break;
   } else {
    $i$325$i = $430;
   }
  }
 }
 $431 = (_luaG_checkcode($9)|0);
 $432 = ($431|0)==(0);
 if (!($432)) {
  $437 = HEAP32[$S>>2]|0;
  $438 = ((($437)) + 8|0);
  $439 = HEAP32[$438>>2]|0;
  $440 = ((($439)) + -16|0);
  HEAP32[$438>>2] = $440;
  $441 = ((($437)) + 52|0);
  $442 = HEAP16[$441>>1]|0;
  $443 = (($442) + -1)<<16>>16;
  HEAP16[$441>>1] = $443;
  STACKTOP = sp;return ($9|0);
 }
 $433 = HEAP32[$S>>2]|0;
 $434 = ((($S)) + 12|0);
 $435 = HEAP32[$434>>2]|0;
 HEAP32[$vararg_buffer154>>2] = $435;
 $vararg_ptr157 = ((($vararg_buffer154)) + 4|0);
 HEAP32[$vararg_ptr157>>2] = 519210;
 (_luaO_pushfstring($433,519114,$vararg_buffer154)|0);
 $436 = HEAP32[$S>>2]|0;
 _luaD_throw($436,3);
 $437 = HEAP32[$S>>2]|0;
 $438 = ((($437)) + 8|0);
 $439 = HEAP32[$438>>2]|0;
 $440 = ((($439)) + -16|0);
 HEAP32[$438>>2] = $440;
 $441 = ((($437)) + 52|0);
 $442 = HEAP16[$441>>1]|0;
 $443 = (($442) + -1)<<16>>16;
 HEAP16[$441>>1] = $443;
 STACKTOP = sp;return ($9|0);
}
function _luaU_header($h) {
 $h = $h|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$h>>0]=1635077147&255;HEAP8[$h+1>>0]=(1635077147>>8)&255;HEAP8[$h+2>>0]=(1635077147>>16)&255;HEAP8[$h+3>>0]=1635077147>>24;
 $0 = ((($h)) + 4|0);
 $1 = ((($h)) + 5|0);
 HEAP8[$0>>0] = 81;
 $2 = ((($h)) + 6|0);
 HEAP8[$1>>0] = 0;
 $3 = ((($h)) + 7|0);
 HEAP8[$2>>0] = 1;
 $4 = ((($h)) + 8|0);
 HEAP8[$3>>0] = 4;
 $5 = ((($h)) + 9|0);
 HEAP8[$4>>0] = 4;
 $6 = ((($h)) + 10|0);
 HEAP8[$5>>0] = 4;
 $7 = ((($h)) + 11|0);
 HEAP8[$6>>0] = 8;
 HEAP8[$7>>0] = 0;
 return;
}
function _luaV_tonumber($obj,$n) {
 $obj = $obj|0;
 $n = $n|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0, $num = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $num = sp;
 $0 = ((($obj)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 switch ($1|0) {
 case 3:  {
  $$0 = $obj;
  break;
 }
 case 4:  {
  $2 = HEAP32[$obj>>2]|0;
  $3 = ((($2)) + 16|0);
  $4 = (_luaO_str2d($3,$num)|0);
  $5 = ($4|0)==(0);
  if ($5) {
   label = 4;
  } else {
   $6 = +HEAPF64[$num>>3];
   HEAPF64[$n>>3] = $6;
   $7 = ((($n)) + 8|0);
   HEAP32[$7>>2] = 3;
   $$0 = $n;
  }
  break;
 }
 default: {
  label = 4;
 }
 }
 if ((label|0) == 4) {
  $$0 = 0;
 }
 STACKTOP = sp;return ($$0|0);
}
function _luaV_tostring($L,$obj) {
 $L = $L|0;
 $obj = $obj|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0.0, $4 = 0, $5 = 0, $s = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $s = sp + 8|0;
 $0 = ((($obj)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(3);
 if (!($2)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = +HEAPF64[$obj>>3];
 HEAPF64[$vararg_buffer>>3] = $3;
 (_sprintf($s,521660,$vararg_buffer)|0);
 $4 = (_strlen($s)|0);
 $5 = (_luaS_newlstr($L,$s,$4)|0);
 HEAP32[$obj>>2] = $5;
 HEAP32[$0>>2] = 4;
 $$0 = 1;
 STACKTOP = sp;return ($$0|0);
}
function _luaV_gettable($L,$t,$key,$val) {
 $L = $L|0;
 $t = $t|0;
 $key = $key|0;
 $val = $val|0;
 var $$0215 = 0, $$0215$lcssa44 = 0, $$lcssa = 0, $$lcssa46 = 0, $$lcssa48 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $loop$014 = 0, $tm$3 = 0, $tm$3$lcssa = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($L)) + 16|0);
 $$phi$trans$insert = ((($t)) + 8|0);
 $$pre = HEAP32[$$phi$trans$insert>>2]|0;
 $$0215 = $t;$4 = $$pre;$loop$014 = 0;
 while(1) {
  $3 = ($4|0)==(5);
  if ($3) {
   $5 = HEAP32[$$0215>>2]|0;
   $6 = (_luaH_get($5,$key)|0);
   $7 = ((($6)) + 8|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ($8|0)==(0);
   if (!($9)) {
    $$lcssa = $6;$$lcssa46 = $7;
    label = 8;
    break;
   }
   $10 = ((($5)) + 8|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = ($11|0)==(0|0);
   if ($12) {
    $$lcssa = $6;$$lcssa46 = $7;
    label = 8;
    break;
   }
   $13 = ((($11)) + 6|0);
   $14 = HEAP8[$13>>0]|0;
   $15 = $14 & 1;
   $16 = ($15<<24>>24)==(0);
   if (!($16)) {
    $$lcssa = $6;$$lcssa46 = $7;
    label = 8;
    break;
   }
   $17 = HEAP32[$0>>2]|0;
   $18 = ((($17)) + 188|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = (_luaT_gettm($11,0,$19)|0);
   $21 = ($20|0)==(0|0);
   if ($21) {
    $$lcssa = $6;$$lcssa46 = $7;
    label = 8;
    break;
   } else {
    $tm$3 = $20;
   }
  } else {
   $34 = (_luaT_gettmbyobj($L,$$0215,0)|0);
   $35 = ((($34)) + 8|0);
   $36 = HEAP32[$35>>2]|0;
   $37 = ($36|0)==(0);
   if ($37) {
    _luaG_typeerror($L,$$0215,519219);
    $tm$3 = $34;
   } else {
    $tm$3 = $34;
   }
  }
  $38 = ((($tm$3)) + 8|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = ($39|0)==(6);
  $2 = (($loop$014) + 1)|0;
  if ($40) {
   $$0215$lcssa44 = $$0215;$$lcssa48 = $38;$tm$3$lcssa = $tm$3;
   label = 12;
   break;
  }
  $1 = ($2|0)<(100);
  if ($1) {
   $$0215 = $tm$3;$4 = $39;$loop$014 = $2;
  } else {
   label = 15;
   break;
  }
 }
 if ((label|0) == 8) {
  $22 = $$lcssa;
  $23 = $22;
  $24 = HEAP32[$23>>2]|0;
  $25 = (($22) + 4)|0;
  $26 = $25;
  $27 = HEAP32[$26>>2]|0;
  $28 = $val;
  $29 = $28;
  HEAP32[$29>>2] = $24;
  $30 = (($28) + 4)|0;
  $31 = $30;
  HEAP32[$31>>2] = $27;
  $32 = HEAP32[$$lcssa46>>2]|0;
  $33 = ((($val)) + 8|0);
  HEAP32[$33>>2] = $32;
  STACKTOP = sp;return;
 }
 else if ((label|0) == 12) {
  $41 = ((($$0215$lcssa44)) + 8|0);
  $42 = ((($L)) + 32|0);
  $43 = HEAP32[$42>>2]|0;
  $44 = $val;
  $45 = (($44) - ($43))|0;
  $46 = ((($L)) + 8|0);
  $47 = HEAP32[$46>>2]|0;
  $48 = $tm$3$lcssa;
  $49 = $48;
  $50 = HEAP32[$49>>2]|0;
  $51 = (($48) + 4)|0;
  $52 = $51;
  $53 = HEAP32[$52>>2]|0;
  $54 = $47;
  $55 = $54;
  HEAP32[$55>>2] = $50;
  $56 = (($54) + 4)|0;
  $57 = $56;
  HEAP32[$57>>2] = $53;
  $58 = HEAP32[$$lcssa48>>2]|0;
  $59 = ((($47)) + 8|0);
  HEAP32[$59>>2] = $58;
  $60 = HEAP32[$46>>2]|0;
  $61 = ((($60)) + 16|0);
  $62 = $$0215$lcssa44;
  $63 = $62;
  $64 = HEAP32[$63>>2]|0;
  $65 = (($62) + 4)|0;
  $66 = $65;
  $67 = HEAP32[$66>>2]|0;
  $68 = $61;
  $69 = $68;
  HEAP32[$69>>2] = $64;
  $70 = (($68) + 4)|0;
  $71 = $70;
  HEAP32[$71>>2] = $67;
  $72 = HEAP32[$41>>2]|0;
  $73 = ((($60)) + 24|0);
  HEAP32[$73>>2] = $72;
  $74 = HEAP32[$46>>2]|0;
  $75 = ((($74)) + 32|0);
  $76 = $key;
  $77 = $76;
  $78 = HEAP32[$77>>2]|0;
  $79 = (($76) + 4)|0;
  $80 = $79;
  $81 = HEAP32[$80>>2]|0;
  $82 = $75;
  $83 = $82;
  HEAP32[$83>>2] = $78;
  $84 = (($82) + 4)|0;
  $85 = $84;
  HEAP32[$85>>2] = $81;
  $86 = ((($key)) + 8|0);
  $87 = HEAP32[$86>>2]|0;
  $88 = ((($74)) + 40|0);
  HEAP32[$88>>2] = $87;
  $89 = ((($L)) + 28|0);
  $90 = HEAP32[$89>>2]|0;
  $91 = HEAP32[$46>>2]|0;
  $92 = (($90) - ($91))|0;
  $93 = ($92|0)<(49);
  $94 = $91;
  if ($93) {
   _luaD_growstack($L,3);
   $$pre$i = HEAP32[$46>>2]|0;
   $96 = $$pre$i;
  } else {
   $96 = $94;
  }
  $95 = ((($96)) + 48|0);
  HEAP32[$46>>2] = $95;
  _luaD_call($L,$96,1);
  $97 = HEAP32[$42>>2]|0;
  $98 = (($97) + ($45)|0);
  $99 = HEAP32[$46>>2]|0;
  $100 = ((($99)) + -16|0);
  HEAP32[$46>>2] = $100;
  $101 = $100;
  $102 = $101;
  $103 = HEAP32[$102>>2]|0;
  $104 = (($101) + 4)|0;
  $105 = $104;
  $106 = HEAP32[$105>>2]|0;
  $107 = $98;
  $108 = $107;
  HEAP32[$108>>2] = $103;
  $109 = (($107) + 4)|0;
  $110 = $109;
  HEAP32[$110>>2] = $106;
  $111 = ((($99)) + -8|0);
  $112 = HEAP32[$111>>2]|0;
  $113 = ((($98)) + 8|0);
  HEAP32[$113>>2] = $112;
  STACKTOP = sp;return;
 }
 else if ((label|0) == 15) {
  _luaG_runerror($L,519225,$vararg_buffer);
  STACKTOP = sp;return;
 }
}
function _luaV_settable($L,$t,$key,$val) {
 $L = $L|0;
 $t = $t|0;
 $key = $key|0;
 $val = $val|0;
 var $$017 = 0, $$017$lcssa50 = 0, $$lcssa = 0, $$lcssa52 = 0, $$lcssa53 = 0, $$lcssa55 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $loop$016 = 0, $temp = 0, $tm$3 = 0, $tm$3$lcssa = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp + 16|0;
 $temp = sp;
 $0 = ((($L)) + 16|0);
 $1 = ((($temp)) + 8|0);
 $$phi$trans$insert = ((($t)) + 8|0);
 $$pre = HEAP32[$$phi$trans$insert>>2]|0;
 $$017 = $t;$3 = $$pre;$loop$016 = 0;
 while(1) {
  $2 = ($3|0)==(5);
  if ($2) {
   $4 = HEAP32[$$017>>2]|0;
   $5 = (_luaH_set($L,$4,$key)|0);
   $6 = ((($5)) + 8|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0);
   if (!($8)) {
    $$lcssa = $4;$$lcssa52 = $5;$$lcssa53 = $6;
    label = 7;
    break;
   }
   $9 = ((($4)) + 8|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($10|0)==(0|0);
   if ($11) {
    $$lcssa = $4;$$lcssa52 = $5;$$lcssa53 = $6;
    label = 7;
    break;
   }
   $12 = ((($10)) + 6|0);
   $13 = HEAP8[$12>>0]|0;
   $14 = $13 & 2;
   $15 = ($14<<24>>24)==(0);
   if (!($15)) {
    $$lcssa = $4;$$lcssa52 = $5;$$lcssa53 = $6;
    label = 7;
    break;
   }
   $16 = HEAP32[$0>>2]|0;
   $17 = ((($16)) + 192|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = (_luaT_gettm($10,1,$18)|0);
   $20 = ($19|0)==(0|0);
   if ($20) {
    $$lcssa = $4;$$lcssa52 = $5;$$lcssa53 = $6;
    label = 7;
    break;
   } else {
    $tm$3 = $19;
   }
  } else {
   $45 = (_luaT_gettmbyobj($L,$$017,1)|0);
   $46 = ((($45)) + 8|0);
   $47 = HEAP32[$46>>2]|0;
   $48 = ($47|0)==(0);
   if ($48) {
    _luaG_typeerror($L,$$017,519219);
    $tm$3 = $45;
   } else {
    $tm$3 = $45;
   }
  }
  $49 = ((($tm$3)) + 8|0);
  $50 = HEAP32[$49>>2]|0;
  $51 = ($50|0)==(6);
  if ($51) {
   $$017$lcssa50 = $$017;$$lcssa55 = $49;$tm$3$lcssa = $tm$3;
   label = 14;
   break;
  }
  $119 = $tm$3;
  $120 = $119;
  $121 = HEAP32[$120>>2]|0;
  $122 = (($119) + 4)|0;
  $123 = $122;
  $124 = HEAP32[$123>>2]|0;
  $125 = $temp;
  $126 = $125;
  HEAP32[$126>>2] = $121;
  $127 = (($125) + 4)|0;
  $128 = $127;
  HEAP32[$128>>2] = $124;
  $129 = HEAP32[$49>>2]|0;
  HEAP32[$1>>2] = $129;
  $130 = (($loop$016) + 1)|0;
  $131 = ($130|0)<(100);
  if ($131) {
   $$017 = $temp;$3 = $129;$loop$016 = $130;
  } else {
   label = 18;
   break;
  }
 }
 if ((label|0) == 7) {
  $21 = $val;
  $22 = $21;
  $23 = HEAP32[$22>>2]|0;
  $24 = (($21) + 4)|0;
  $25 = $24;
  $26 = HEAP32[$25>>2]|0;
  $27 = $$lcssa52;
  $28 = $27;
  HEAP32[$28>>2] = $23;
  $29 = (($27) + 4)|0;
  $30 = $29;
  HEAP32[$30>>2] = $26;
  $31 = ((($val)) + 8|0);
  $32 = HEAP32[$31>>2]|0;
  HEAP32[$$lcssa53>>2] = $32;
  $33 = ((($$lcssa)) + 6|0);
  HEAP8[$33>>0] = 0;
  $34 = HEAP32[$31>>2]|0;
  $35 = ($34|0)>(3);
  if (!($35)) {
   STACKTOP = sp;return;
  }
  $36 = HEAP32[$val>>2]|0;
  $37 = ((($36)) + 5|0);
  $38 = HEAP8[$37>>0]|0;
  $39 = $38 & 3;
  $40 = ($39<<24>>24)==(0);
  if ($40) {
   STACKTOP = sp;return;
  }
  $41 = ((($$lcssa)) + 5|0);
  $42 = HEAP8[$41>>0]|0;
  $43 = $42 & 4;
  $44 = ($43<<24>>24)==(0);
  if ($44) {
   STACKTOP = sp;return;
  }
  _luaC_barrierback($L,$$lcssa);
  STACKTOP = sp;return;
 }
 else if ((label|0) == 14) {
  $52 = ((($$017$lcssa50)) + 8|0);
  $53 = ((($L)) + 8|0);
  $54 = HEAP32[$53>>2]|0;
  $55 = $tm$3$lcssa;
  $56 = $55;
  $57 = HEAP32[$56>>2]|0;
  $58 = (($55) + 4)|0;
  $59 = $58;
  $60 = HEAP32[$59>>2]|0;
  $61 = $54;
  $62 = $61;
  HEAP32[$62>>2] = $57;
  $63 = (($61) + 4)|0;
  $64 = $63;
  HEAP32[$64>>2] = $60;
  $65 = HEAP32[$$lcssa55>>2]|0;
  $66 = ((($54)) + 8|0);
  HEAP32[$66>>2] = $65;
  $67 = HEAP32[$53>>2]|0;
  $68 = ((($67)) + 16|0);
  $69 = $$017$lcssa50;
  $70 = $69;
  $71 = HEAP32[$70>>2]|0;
  $72 = (($69) + 4)|0;
  $73 = $72;
  $74 = HEAP32[$73>>2]|0;
  $75 = $68;
  $76 = $75;
  HEAP32[$76>>2] = $71;
  $77 = (($75) + 4)|0;
  $78 = $77;
  HEAP32[$78>>2] = $74;
  $79 = HEAP32[$52>>2]|0;
  $80 = ((($67)) + 24|0);
  HEAP32[$80>>2] = $79;
  $81 = HEAP32[$53>>2]|0;
  $82 = ((($81)) + 32|0);
  $83 = $key;
  $84 = $83;
  $85 = HEAP32[$84>>2]|0;
  $86 = (($83) + 4)|0;
  $87 = $86;
  $88 = HEAP32[$87>>2]|0;
  $89 = $82;
  $90 = $89;
  HEAP32[$90>>2] = $85;
  $91 = (($89) + 4)|0;
  $92 = $91;
  HEAP32[$92>>2] = $88;
  $93 = ((($key)) + 8|0);
  $94 = HEAP32[$93>>2]|0;
  $95 = ((($81)) + 40|0);
  HEAP32[$95>>2] = $94;
  $96 = HEAP32[$53>>2]|0;
  $97 = ((($96)) + 48|0);
  $98 = $val;
  $99 = $98;
  $100 = HEAP32[$99>>2]|0;
  $101 = (($98) + 4)|0;
  $102 = $101;
  $103 = HEAP32[$102>>2]|0;
  $104 = $97;
  $105 = $104;
  HEAP32[$105>>2] = $100;
  $106 = (($104) + 4)|0;
  $107 = $106;
  HEAP32[$107>>2] = $103;
  $108 = ((($val)) + 8|0);
  $109 = HEAP32[$108>>2]|0;
  $110 = ((($96)) + 56|0);
  HEAP32[$110>>2] = $109;
  $111 = ((($L)) + 28|0);
  $112 = HEAP32[$111>>2]|0;
  $113 = HEAP32[$53>>2]|0;
  $114 = (($112) - ($113))|0;
  $115 = ($114|0)<(65);
  $116 = $113;
  if ($115) {
   _luaD_growstack($L,4);
   $$pre$i = HEAP32[$53>>2]|0;
   $118 = $$pre$i;
  } else {
   $118 = $116;
  }
  $117 = ((($118)) + 64|0);
  HEAP32[$53>>2] = $117;
  _luaD_call($L,$118,0);
  STACKTOP = sp;return;
 }
 else if ((label|0) == 18) {
  _luaG_runerror($L,519242,$vararg_buffer);
  STACKTOP = sp;return;
 }
}
function _luaV_lessthan($L,$l,$r) {
 $L = $L|0;
 $l = $l|0;
 $r = $r|0;
 var $$0 = 0, $$25$i = 0, $$lcssa = 0, $$lobit = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0.0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0.0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $l$09$i = 0, $ll$010$i = 0, $lr$012$i = 0, $phitmp$i = 0, $r$011$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($l)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($r)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==($3|0);
 if (!($4)) {
  $5 = (_luaG_ordererror($L,$l,$r)|0);
  $$0 = $5;
  return ($$0|0);
 }
 switch ($1|0) {
 case 3:  {
  $6 = +HEAPF64[$l>>3];
  $7 = +HEAPF64[$r>>3];
  $8 = $6 < $7;
  $9 = $8&1;
  $$0 = $9;
  return ($$0|0);
  break;
 }
 case 4:  {
  $10 = HEAP32[$l>>2]|0;
  $11 = HEAP32[$r>>2]|0;
  $12 = ((($10)) + 16|0);
  $13 = ((($11)) + 16|0);
  $14 = (_strcoll($12,$13)|0);
  $15 = ($14|0)==(0);
  L23: do {
   if ($15) {
    $16 = ((($11)) + 12|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = ((($10)) + 12|0);
    $19 = HEAP32[$18>>2]|0;
    $l$09$i = $12;$ll$010$i = $19;$lr$012$i = $17;$r$011$i = $13;
    while(1) {
     $20 = (_strlen($l$09$i)|0);
     $21 = ($20|0)==($lr$012$i|0);
     $22 = ($20|0)==($ll$010$i|0);
     if ($21) {
      $$lcssa = $22;
      break;
     }
     if ($22) {
      $$25$i = -1;
      break L23;
     }
     $25 = (($20) + 1)|0;
     $26 = (($l$09$i) + ($25)|0);
     $27 = (($ll$010$i) - ($25))|0;
     $28 = (($r$011$i) + ($25)|0);
     $29 = (($lr$012$i) - ($25))|0;
     $30 = (_strcoll($26,$28)|0);
     $31 = ($30|0)==(0);
     if ($31) {
      $l$09$i = $26;$ll$010$i = $27;$lr$012$i = $29;$r$011$i = $28;
     } else {
      $$25$i = $30;
      break L23;
     }
    }
    $23 = $$lcssa&1;
    $24 = $23 ^ 1;
    $$25$i = $24;
   } else {
    $$25$i = $14;
   }
  } while(0);
  $$lobit = $$25$i >>> 31;
  $$0 = $$lobit;
  return ($$0|0);
  break;
 }
 default: {
  $32 = (_luaT_gettmbyobj($L,$l,13)|0);
  $33 = ((($32)) + 8|0);
  $34 = HEAP32[$33>>2]|0;
  $35 = ($34|0)==(0);
  if (!($35)) {
   $36 = (_luaT_gettmbyobj($L,$r,13)|0);
   $37 = (_luaO_rawequalObj($32,$36)|0);
   $38 = ($37|0)==(0);
   if (!($38)) {
    $39 = ((($L)) + 8|0);
    $40 = HEAP32[$39>>2]|0;
    $41 = ((($L)) + 32|0);
    $42 = HEAP32[$41>>2]|0;
    $43 = $40;
    $44 = (($43) - ($42))|0;
    $45 = $32;
    $46 = $45;
    $47 = HEAP32[$46>>2]|0;
    $48 = (($45) + 4)|0;
    $49 = $48;
    $50 = HEAP32[$49>>2]|0;
    $51 = $40;
    $52 = $51;
    HEAP32[$52>>2] = $47;
    $53 = (($51) + 4)|0;
    $54 = $53;
    HEAP32[$54>>2] = $50;
    $55 = HEAP32[$33>>2]|0;
    $56 = ((($40)) + 8|0);
    HEAP32[$56>>2] = $55;
    $57 = HEAP32[$39>>2]|0;
    $58 = ((($57)) + 16|0);
    $59 = $l;
    $60 = $59;
    $61 = HEAP32[$60>>2]|0;
    $62 = (($59) + 4)|0;
    $63 = $62;
    $64 = HEAP32[$63>>2]|0;
    $65 = $58;
    $66 = $65;
    HEAP32[$66>>2] = $61;
    $67 = (($65) + 4)|0;
    $68 = $67;
    HEAP32[$68>>2] = $64;
    $69 = HEAP32[$0>>2]|0;
    $70 = ((($57)) + 24|0);
    HEAP32[$70>>2] = $69;
    $71 = HEAP32[$39>>2]|0;
    $72 = ((($71)) + 32|0);
    $73 = $r;
    $74 = $73;
    $75 = HEAP32[$74>>2]|0;
    $76 = (($73) + 4)|0;
    $77 = $76;
    $78 = HEAP32[$77>>2]|0;
    $79 = $72;
    $80 = $79;
    HEAP32[$80>>2] = $75;
    $81 = (($79) + 4)|0;
    $82 = $81;
    HEAP32[$82>>2] = $78;
    $83 = HEAP32[$2>>2]|0;
    $84 = ((($71)) + 40|0);
    HEAP32[$84>>2] = $83;
    $85 = ((($L)) + 28|0);
    $86 = HEAP32[$85>>2]|0;
    $87 = HEAP32[$39>>2]|0;
    $88 = (($86) - ($87))|0;
    $89 = ($88|0)<(49);
    $90 = $87;
    if ($89) {
     _luaD_growstack($L,3);
     $$pre$i = HEAP32[$39>>2]|0;
     $92 = $$pre$i;
    } else {
     $92 = $90;
    }
    $91 = ((($92)) + 48|0);
    HEAP32[$39>>2] = $91;
    _luaD_call($L,$92,1);
    $93 = HEAP32[$41>>2]|0;
    $94 = (($93) + ($44)|0);
    $95 = HEAP32[$39>>2]|0;
    $96 = ((($95)) + -16|0);
    HEAP32[$39>>2] = $96;
    $97 = $96;
    $98 = $97;
    $99 = HEAP32[$98>>2]|0;
    $100 = (($97) + 4)|0;
    $101 = $100;
    $102 = HEAP32[$101>>2]|0;
    $103 = $94;
    $104 = $103;
    HEAP32[$104>>2] = $99;
    $105 = (($103) + 4)|0;
    $106 = $105;
    HEAP32[$106>>2] = $102;
    $107 = ((($95)) + -8|0);
    $108 = HEAP32[$107>>2]|0;
    $109 = ((($94)) + 8|0);
    HEAP32[$109>>2] = $108;
    $110 = HEAP32[$39>>2]|0;
    $111 = ((($110)) + 8|0);
    $112 = HEAP32[$111>>2]|0;
    $113 = ($112|0)==(0);
    if ($113) {
     $117 = 0;
    } else {
     $114 = ($112|0)==(1);
     if ($114) {
      $115 = HEAP32[$110>>2]|0;
      $phitmp$i = ($115|0)!=(0);
      $117 = $phitmp$i;
     } else {
      $117 = 1;
     }
    }
    $116 = $117&1;
    $$0 = $116;
    return ($$0|0);
   }
  }
  $118 = (_luaG_ordererror($L,$l,$r)|0);
  $$0 = $118;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function _luaV_equalval($L,$t1,$t2) {
 $L = $L|0;
 $t1 = $t1|0;
 $t2 = $t2|0;
 var $$0 = 0, $$0$shrunk = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $phitmp = 0, $tm$0 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($t1)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 switch ($1|0) {
 case 0:  {
  $$0$shrunk = 1;
  break;
 }
 case 3:  {
  $2 = +HEAPF64[$t1>>3];
  $3 = +HEAPF64[$t2>>3];
  $4 = $2 == $3;
  $$0$shrunk = $4;
  break;
 }
 case 1:  {
  $5 = HEAP32[$t1>>2]|0;
  $6 = HEAP32[$t2>>2]|0;
  $7 = ($5|0)==($6|0);
  $$0$shrunk = $7;
  break;
 }
 case 2:  {
  $8 = HEAP32[$t1>>2]|0;
  $9 = HEAP32[$t2>>2]|0;
  $10 = ($8|0)==($9|0);
  $$0$shrunk = $10;
  break;
 }
 case 7:  {
  $11 = HEAP32[$t1>>2]|0;
  $12 = HEAP32[$t2>>2]|0;
  $13 = ($11|0)==($12|0);
  if ($13) {
   $$0$shrunk = 1;
  } else {
   $14 = ((($11)) + 8|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ((($12)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ($15|0)==(0|0);
   if ($18) {
    $$0$shrunk = 0;
   } else {
    $19 = ((($15)) + 6|0);
    $20 = HEAP8[$19>>0]|0;
    $21 = $20 & 16;
    $22 = ($21<<24>>24)==(0);
    if ($22) {
     $23 = ((($L)) + 16|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ((($24)) + 204|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = (_luaT_gettm($15,4,$26)|0);
     $28 = ($27|0)==(0|0);
     if ($28) {
      $$0$shrunk = 0;
     } else {
      $29 = ($15|0)==($17|0);
      if ($29) {
       $tm$0 = $27;
       label = 24;
      } else {
       $30 = ($17|0)==(0|0);
       if ($30) {
        $$0$shrunk = 0;
       } else {
        $31 = ((($17)) + 6|0);
        $32 = HEAP8[$31>>0]|0;
        $33 = $32 & 16;
        $34 = ($33<<24>>24)==(0);
        if ($34) {
         $35 = HEAP32[$23>>2]|0;
         $36 = ((($35)) + 204|0);
         $37 = HEAP32[$36>>2]|0;
         $38 = (_luaT_gettm($17,4,$37)|0);
         $39 = ($38|0)==(0|0);
         if ($39) {
          $$0$shrunk = 0;
         } else {
          $40 = (_luaO_rawequalObj($27,$38)|0);
          $41 = ($40|0)==(0);
          if ($41) {
           $$0$shrunk = 0;
          } else {
           $tm$0 = $27;
           label = 24;
          }
         }
        } else {
         $$0$shrunk = 0;
        }
       }
      }
     }
    } else {
     $$0$shrunk = 0;
    }
   }
  }
  break;
 }
 case 5:  {
  $42 = HEAP32[$t1>>2]|0;
  $43 = HEAP32[$t2>>2]|0;
  $44 = ($42|0)==($43|0);
  if ($44) {
   $$0$shrunk = 1;
  } else {
   $45 = ((($42)) + 8|0);
   $46 = HEAP32[$45>>2]|0;
   $47 = ((($43)) + 8|0);
   $48 = HEAP32[$47>>2]|0;
   $49 = ($46|0)==(0|0);
   if ($49) {
    $$0$shrunk = 0;
   } else {
    $50 = ((($46)) + 6|0);
    $51 = HEAP8[$50>>0]|0;
    $52 = $51 & 16;
    $53 = ($52<<24>>24)==(0);
    if ($53) {
     $54 = ((($L)) + 16|0);
     $55 = HEAP32[$54>>2]|0;
     $56 = ((($55)) + 204|0);
     $57 = HEAP32[$56>>2]|0;
     $58 = (_luaT_gettm($46,4,$57)|0);
     $59 = ($58|0)==(0|0);
     if ($59) {
      $$0$shrunk = 0;
     } else {
      $60 = ($46|0)==($48|0);
      if ($60) {
       $tm$0 = $58;
       label = 24;
      } else {
       $61 = ($48|0)==(0|0);
       if ($61) {
        $$0$shrunk = 0;
       } else {
        $62 = ((($48)) + 6|0);
        $63 = HEAP8[$62>>0]|0;
        $64 = $63 & 16;
        $65 = ($64<<24>>24)==(0);
        if ($65) {
         $66 = HEAP32[$54>>2]|0;
         $67 = ((($66)) + 204|0);
         $68 = HEAP32[$67>>2]|0;
         $69 = (_luaT_gettm($48,4,$68)|0);
         $70 = ($69|0)==(0|0);
         if ($70) {
          $$0$shrunk = 0;
         } else {
          $71 = (_luaO_rawequalObj($58,$69)|0);
          $72 = ($71|0)==(0);
          if ($72) {
           $$0$shrunk = 0;
          } else {
           $tm$0 = $58;
           label = 24;
          }
         }
        } else {
         $$0$shrunk = 0;
        }
       }
      }
     }
    } else {
     $$0$shrunk = 0;
    }
   }
  }
  break;
 }
 default: {
  $73 = HEAP32[$t1>>2]|0;
  $74 = HEAP32[$t2>>2]|0;
  $75 = ($73|0)==($74|0);
  $$0$shrunk = $75;
 }
 }
 if ((label|0) == 24) {
  $76 = ((($L)) + 8|0);
  $77 = HEAP32[$76>>2]|0;
  $78 = ((($L)) + 32|0);
  $79 = HEAP32[$78>>2]|0;
  $80 = $77;
  $81 = (($80) - ($79))|0;
  $82 = $tm$0;
  $83 = $82;
  $84 = HEAP32[$83>>2]|0;
  $85 = (($82) + 4)|0;
  $86 = $85;
  $87 = HEAP32[$86>>2]|0;
  $88 = $77;
  $89 = $88;
  HEAP32[$89>>2] = $84;
  $90 = (($88) + 4)|0;
  $91 = $90;
  HEAP32[$91>>2] = $87;
  $92 = ((($tm$0)) + 8|0);
  $93 = HEAP32[$92>>2]|0;
  $94 = ((($77)) + 8|0);
  HEAP32[$94>>2] = $93;
  $95 = HEAP32[$76>>2]|0;
  $96 = ((($95)) + 16|0);
  $97 = $t1;
  $98 = $97;
  $99 = HEAP32[$98>>2]|0;
  $100 = (($97) + 4)|0;
  $101 = $100;
  $102 = HEAP32[$101>>2]|0;
  $103 = $96;
  $104 = $103;
  HEAP32[$104>>2] = $99;
  $105 = (($103) + 4)|0;
  $106 = $105;
  HEAP32[$106>>2] = $102;
  $107 = HEAP32[$0>>2]|0;
  $108 = ((($95)) + 24|0);
  HEAP32[$108>>2] = $107;
  $109 = HEAP32[$76>>2]|0;
  $110 = ((($109)) + 32|0);
  $111 = $t2;
  $112 = $111;
  $113 = HEAP32[$112>>2]|0;
  $114 = (($111) + 4)|0;
  $115 = $114;
  $116 = HEAP32[$115>>2]|0;
  $117 = $110;
  $118 = $117;
  HEAP32[$118>>2] = $113;
  $119 = (($117) + 4)|0;
  $120 = $119;
  HEAP32[$120>>2] = $116;
  $121 = ((($t2)) + 8|0);
  $122 = HEAP32[$121>>2]|0;
  $123 = ((($109)) + 40|0);
  HEAP32[$123>>2] = $122;
  $124 = ((($L)) + 28|0);
  $125 = HEAP32[$124>>2]|0;
  $126 = HEAP32[$76>>2]|0;
  $127 = (($125) - ($126))|0;
  $128 = ($127|0)<(49);
  $129 = $126;
  if ($128) {
   _luaD_growstack($L,3);
   $$pre$i = HEAP32[$76>>2]|0;
   $131 = $$pre$i;
  } else {
   $131 = $129;
  }
  $130 = ((($131)) + 48|0);
  HEAP32[$76>>2] = $130;
  _luaD_call($L,$131,1);
  $132 = HEAP32[$78>>2]|0;
  $133 = (($132) + ($81)|0);
  $134 = HEAP32[$76>>2]|0;
  $135 = ((($134)) + -16|0);
  HEAP32[$76>>2] = $135;
  $136 = $135;
  $137 = $136;
  $138 = HEAP32[$137>>2]|0;
  $139 = (($136) + 4)|0;
  $140 = $139;
  $141 = HEAP32[$140>>2]|0;
  $142 = $133;
  $143 = $142;
  HEAP32[$143>>2] = $138;
  $144 = (($142) + 4)|0;
  $145 = $144;
  HEAP32[$145>>2] = $141;
  $146 = ((($134)) + -8|0);
  $147 = HEAP32[$146>>2]|0;
  $148 = ((($133)) + 8|0);
  HEAP32[$148>>2] = $147;
  $149 = HEAP32[$76>>2]|0;
  $150 = ((($149)) + 8|0);
  $151 = HEAP32[$150>>2]|0;
  $152 = ($151|0)==(0);
  if ($152) {
   $$0$shrunk = 0;
  } else {
   $153 = ($151|0)==(1);
   if ($153) {
    $154 = HEAP32[$149>>2]|0;
    $phitmp = ($154|0)!=(0);
    $$0$shrunk = $phitmp;
   } else {
    $$0$shrunk = 1;
   }
  }
 }
 $$0 = $$0$shrunk&1;
 return ($$0|0);
}
function _luaV_concat($L,$total,$last) {
 $L = $L|0;
 $total = $total|0;
 $last = $last|0;
 var $$0 = 0, $$02 = 0, $$off = 0, $$phi$trans$insert$i = 0, $$pre$i = 0, $$pre$i4 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0.0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0.0, $140 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cond = 0, $i$018 = 0, $n$0$lcssa = 0, $n$012 = 0, $n$1 = 0, $s$i = 0, $switch = 0, $tl$0$lcssa = 0, $tl$011 = 0, $tl$1$lcssa = 0, $tl$117 = 0, $tm$01$i = 0, $vararg_buffer = 0;
 var $vararg_buffer1 = 0, $vararg_buffer4 = 0, $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $s$i = sp + 32|0;
 $0 = ((($L)) + 12|0);
 $1 = ((($L)) + 16|0);
 $2 = ((($L)) + 32|0);
 $3 = ((($L)) + 8|0);
 $4 = ((($L)) + 28|0);
 $$0 = $total;$$02 = $last;
 while(1) {
  $5 = HEAP32[$0>>2]|0;
  $6 = (($5) + ($$02<<4)|0);
  $7 = ((($6)) + 16|0);
  $8 = ((($7)) + -32|0);
  $9 = ((($8)) + 8|0);
  $10 = HEAP32[$9>>2]|0;
  $$off = (($10) + -3)|0;
  $switch = ($$off>>>0)<(2);
  $11 = ((($7)) + -16|0);
  L3: do {
   if ($switch) {
    $12 = ((($11)) + 8|0);
    $13 = HEAP32[$12>>2]|0;
    switch ($13|0) {
    case 4:  {
     break;
    }
    case 3:  {
     $14 = +HEAPF64[$11>>3];
     HEAPF64[$vararg_buffer>>3] = $14;
     (_sprintf($s$i,521660,$vararg_buffer)|0);
     $15 = (_strlen($s$i)|0);
     $16 = (_luaS_newlstr($L,$s$i,$15)|0);
     HEAP32[$11>>2] = $16;
     HEAP32[$12>>2] = 4;
     break;
    }
    default: {
     label = 5;
     break L3;
    }
    }
    $94 = HEAP32[$11>>2]|0;
    $95 = ((($94)) + 12|0);
    $96 = HEAP32[$95>>2]|0;
    $97 = ($96|0)==(0);
    if ($97) {
     $99 = HEAP32[$9>>2]|0;
     $cond = ($99|0)==(3);
     if (!($cond)) {
      $n$1 = 2;
      break;
     }
     $100 = +HEAPF64[$8>>3];
     HEAPF64[$vararg_buffer1>>3] = $100;
     (_sprintf($s$i,521660,$vararg_buffer1)|0);
     $101 = (_strlen($s$i)|0);
     $102 = (_luaS_newlstr($L,$s$i,$101)|0);
     HEAP32[$8>>2] = $102;
     HEAP32[$9>>2] = 4;
     $n$1 = 2;
     break;
    }
    $98 = ($$0|0)>(1);
    L12: do {
     if ($98) {
      $n$012 = 1;$tl$011 = $96;
      while(1) {
       $103 = (0 - ($n$012))|0;
       $104 = (($7) + ($103<<4)|0);
       $105 = ((($104)) + -16|0);
       $106 = ((($105)) + 8|0);
       $107 = HEAP32[$106>>2]|0;
       switch ($107|0) {
       case 4:  {
        break;
       }
       case 3:  {
        $108 = +HEAPF64[$105>>3];
        HEAPF64[$vararg_buffer4>>3] = $108;
        (_sprintf($s$i,521660,$vararg_buffer4)|0);
        $109 = (_strlen($s$i)|0);
        $110 = (_luaS_newlstr($L,$s$i,$109)|0);
        HEAP32[$105>>2] = $110;
        HEAP32[$106>>2] = 4;
        break;
       }
       default: {
        $n$0$lcssa = $n$012;$tl$0$lcssa = $tl$011;
        break L12;
       }
       }
       $111 = HEAP32[$105>>2]|0;
       $112 = ((($111)) + 12|0);
       $113 = HEAP32[$112>>2]|0;
       $114 = (-3 - ($tl$011))|0;
       $115 = ($113>>>0)<($114>>>0);
       if (!($115)) {
        _luaG_runerror($L,519259,$vararg_buffer7);
       }
       $116 = (($113) + ($tl$011))|0;
       $117 = (($n$012) + 1)|0;
       $118 = ($117|0)<($$0|0);
       if ($118) {
        $n$012 = $117;$tl$011 = $116;
       } else {
        $n$0$lcssa = $117;$tl$0$lcssa = $116;
        break L12;
       }
      }
     } else {
      $n$0$lcssa = 1;$tl$0$lcssa = $96;
     }
    } while(0);
    $119 = HEAP32[$1>>2]|0;
    $120 = ((($119)) + 52|0);
    $121 = (_luaZ_openspace($L,$120,$tl$0$lcssa)|0);
    $122 = ($n$0$lcssa|0)>(0);
    if ($122) {
     $i$018 = $n$0$lcssa;$tl$117 = 0;
     while(1) {
      $123 = (0 - ($i$018))|0;
      $124 = (($7) + ($123<<4)|0);
      $125 = HEAP32[$124>>2]|0;
      $126 = ((($125)) + 12|0);
      $127 = HEAP32[$126>>2]|0;
      $128 = (($121) + ($tl$117)|0);
      $129 = ((($125)) + 16|0);
      _memcpy(($128|0),($129|0),($127|0))|0;
      $130 = (($127) + ($tl$117))|0;
      $131 = (($i$018) + -1)|0;
      $132 = ($i$018|0)>(1);
      if ($132) {
       $i$018 = $131;$tl$117 = $130;
      } else {
       $tl$1$lcssa = $130;
       break;
      }
     }
    } else {
     $tl$1$lcssa = 0;
    }
    $133 = (0 - ($n$0$lcssa))|0;
    $134 = (($7) + ($133<<4)|0);
    $135 = (_luaS_newlstr($L,$121,$tl$1$lcssa)|0);
    HEAP32[$134>>2] = $135;
    $136 = ((($134)) + 8|0);
    HEAP32[$136>>2] = 4;
    $n$1 = $n$0$lcssa;
   } else {
    label = 5;
   }
  } while(0);
  do {
   if ((label|0) == 5) {
    label = 0;
    $17 = (_luaT_gettmbyobj($L,$8,15)|0);
    $18 = ((($17)) + 8|0);
    $19 = HEAP32[$18>>2]|0;
    $20 = ($19|0)==(0);
    if ($20) {
     $21 = (_luaT_gettmbyobj($L,$11,15)|0);
     $$phi$trans$insert$i = ((($21)) + 8|0);
     $$pre$i = HEAP32[$$phi$trans$insert$i>>2]|0;
     $22 = ($$pre$i|0)==(0);
     if ($22) {
      _luaG_concaterror($L,$8,$11);
      $n$1 = 2;
      break;
     } else {
      $tm$01$i = $21;
     }
    } else {
     $tm$01$i = $17;
    }
    $23 = ((($7)) + -16|0);
    $24 = HEAP32[$2>>2]|0;
    $25 = $8;
    $26 = (($25) - ($24))|0;
    $27 = HEAP32[$3>>2]|0;
    $28 = $tm$01$i;
    $29 = $28;
    $30 = HEAP32[$29>>2]|0;
    $31 = (($28) + 4)|0;
    $32 = $31;
    $33 = HEAP32[$32>>2]|0;
    $34 = $27;
    $35 = $34;
    HEAP32[$35>>2] = $30;
    $36 = (($34) + 4)|0;
    $37 = $36;
    HEAP32[$37>>2] = $33;
    $38 = ((($tm$01$i)) + 8|0);
    $39 = HEAP32[$38>>2]|0;
    $40 = ((($27)) + 8|0);
    HEAP32[$40>>2] = $39;
    $41 = HEAP32[$3>>2]|0;
    $42 = ((($41)) + 16|0);
    $43 = $8;
    $44 = $43;
    $45 = HEAP32[$44>>2]|0;
    $46 = (($43) + 4)|0;
    $47 = $46;
    $48 = HEAP32[$47>>2]|0;
    $49 = $42;
    $50 = $49;
    HEAP32[$50>>2] = $45;
    $51 = (($49) + 4)|0;
    $52 = $51;
    HEAP32[$52>>2] = $48;
    $53 = HEAP32[$9>>2]|0;
    $54 = ((($41)) + 24|0);
    HEAP32[$54>>2] = $53;
    $55 = HEAP32[$3>>2]|0;
    $56 = ((($55)) + 32|0);
    $57 = $11;
    $58 = $57;
    $59 = HEAP32[$58>>2]|0;
    $60 = (($57) + 4)|0;
    $61 = $60;
    $62 = HEAP32[$61>>2]|0;
    $63 = $56;
    $64 = $63;
    HEAP32[$64>>2] = $59;
    $65 = (($63) + 4)|0;
    $66 = $65;
    HEAP32[$66>>2] = $62;
    $67 = ((($23)) + 8|0);
    $68 = HEAP32[$67>>2]|0;
    $69 = ((($55)) + 40|0);
    HEAP32[$69>>2] = $68;
    $70 = HEAP32[$4>>2]|0;
    $71 = HEAP32[$3>>2]|0;
    $72 = (($70) - ($71))|0;
    $73 = ($72|0)<(49);
    $74 = $71;
    if ($73) {
     _luaD_growstack($L,3);
     $$pre$i4 = HEAP32[$3>>2]|0;
     $76 = $$pre$i4;
    } else {
     $76 = $74;
    }
    $75 = ((($76)) + 48|0);
    HEAP32[$3>>2] = $75;
    _luaD_call($L,$76,1);
    $77 = HEAP32[$2>>2]|0;
    $78 = (($77) + ($26)|0);
    $79 = HEAP32[$3>>2]|0;
    $80 = ((($79)) + -16|0);
    HEAP32[$3>>2] = $80;
    $81 = $80;
    $82 = $81;
    $83 = HEAP32[$82>>2]|0;
    $84 = (($81) + 4)|0;
    $85 = $84;
    $86 = HEAP32[$85>>2]|0;
    $87 = $78;
    $88 = $87;
    HEAP32[$88>>2] = $83;
    $89 = (($87) + 4)|0;
    $90 = $89;
    HEAP32[$90>>2] = $86;
    $91 = ((($79)) + -8|0);
    $92 = HEAP32[$91>>2]|0;
    $93 = ((($78)) + 8|0);
    HEAP32[$93>>2] = $92;
    $n$1 = 2;
   }
  } while(0);
  $137 = (($n$1) + -1)|0;
  $138 = (($$0) - ($137))|0;
  $139 = (($$02) - ($137))|0;
  $140 = ($138|0)>(1);
  if ($140) {
   $$0 = $138;$$02 = $139;
  } else {
   break;
  }
 }
 STACKTOP = sp;return;
}
function _luaV_execute($L,$nexeccalls) {
 $L = $L|0;
 $nexeccalls = $nexeccalls|0;
 var $$ = 0, $$0$i103$ph = 0, $$0$i106$ph = 0, $$0$i24 = 0, $$0$i33$ph = 0, $$0$i36$ph = 0, $$0$i47$ph = 0, $$0$i50$ph = 0, $$0$i61$ph = 0, $$0$i64$ph = 0, $$0$i75$ph = 0, $$0$i78$ph = 0, $$0$i89$ph = 0, $$0$i92$ph = 0, $$0$ph = 0, $$25$i$i = 0, $$lcssa = 0, $$lcssa181 = 0, $$lcssa359 = 0, $$lcssa365 = 0;
 var $$lcssa374 = 0, $$not = 0, $$not289 = 0, $$phi$trans$insert$i = 0, $$phi$trans$insert$i108 = 0, $$phi$trans$insert$i122 = 0, $$phi$trans$insert$i38 = 0, $$phi$trans$insert$i52 = 0, $$phi$trans$insert$i66 = 0, $$phi$trans$insert$i80 = 0, $$phi$trans$insert$i94 = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i100 = 0, $$pre$i109 = 0, $$pre$i114 = 0, $$pre$i123 = 0, $$pre$i128 = 0, $$pre$i23 = 0;
 var $$pre$i39 = 0, $$pre$i4$i = 0, $$pre$i44 = 0, $$pre$i53 = 0, $$pre$i58 = 0, $$pre$i67 = 0, $$pre$i72 = 0, $$pre$i81 = 0, $$pre$i86 = 0, $$pre$i95 = 0, $$pre239 = 0, $$pre240 = 0, $$pre242 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0;
 var $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0;
 var $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0;
 var $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0;
 var $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0;
 var $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0.0, $1086 = 0, $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0.0, $1091 = 0, $1092 = 0;
 var $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0;
 var $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0;
 var $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0;
 var $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0;
 var $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0;
 var $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0;
 var $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0;
 var $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0;
 var $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0;
 var $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0.0, $1267 = 0.0, $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0;
 var $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0, $1285 = 0, $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0;
 var $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0, $1302 = 0, $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0;
 var $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0, $1320 = 0, $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0;
 var $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0, $1339 = 0, $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0;
 var $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0, $1357 = 0, $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0;
 var $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0, $1375 = 0, $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0;
 var $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0, $1393 = 0, $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0;
 var $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0, $1410 = 0, $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0;
 var $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0, $1429 = 0, $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0;
 var $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0, $1447 = 0, $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0;
 var $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0, $1465 = 0, $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0;
 var $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0, $1483 = 0, $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0;
 var $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0, $1500 = 0, $1501 = 0, $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0;
 var $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0, $1511 = 0, $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0, $1518 = 0, $1519 = 0, $152 = 0, $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0;
 var $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0, $153 = 0, $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0, $1536 = 0, $1537 = 0, $1538 = 0, $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0;
 var $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0, $1548 = 0, $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0, $1554 = 0, $1555 = 0, $1556 = 0, $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0;
 var $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $1565 = 0, $1566 = 0, $1567 = 0, $1568 = 0, $1569 = 0, $157 = 0, $1570 = 0, $1571 = 0, $1572 = 0, $1573 = 0, $1574 = 0, $1575 = 0, $1576 = 0, $1577 = 0, $1578 = 0, $1579 = 0;
 var $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0, $1583 = 0, $1584 = 0, $1585 = 0, $1586 = 0, $1587 = 0, $1588 = 0, $1589 = 0, $159 = 0, $1590 = 0, $1591 = 0.0, $1592 = 0.0, $1593 = 0.0, $1594 = 0, $1595 = 0.0, $1596 = 0, $1597 = 0;
 var $1598 = 0, $1599 = 0, $16 = 0, $160 = 0, $1600 = 0, $1601 = 0, $1602 = 0, $1603 = 0, $1604 = 0, $1605 = 0, $1606 = 0, $1607 = 0, $1608 = 0, $1609 = 0, $161 = 0, $1610 = 0, $1611 = 0, $1612 = 0, $1613 = 0, $1614 = 0.0;
 var $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0, $1619 = 0, $162 = 0, $1620 = 0, $1621 = 0, $1622 = 0.0, $1623 = 0, $1624 = 0, $1625 = 0, $1626 = 0, $1627 = 0, $1628 = 0, $1629 = 0, $163 = 0, $1630 = 0.0, $1631 = 0.0, $1632 = 0.0;
 var $1633 = 0.0, $1634 = 0, $1635 = 0, $1636 = 0, $1637 = 0, $1638 = 0, $1639 = 0, $164 = 0, $1640 = 0, $1641 = 0, $1642 = 0, $1643 = 0, $1644 = 0, $1645 = 0, $1646 = 0, $1647 = 0, $1648 = 0, $1649 = 0, $165 = 0, $1650 = 0;
 var $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0, $1655 = 0, $1656 = 0, $1657 = 0, $1658 = 0, $1659 = 0, $166 = 0, $1660 = 0, $1661 = 0, $1662 = 0, $1663 = 0, $1664 = 0, $1665 = 0, $1666 = 0, $1667 = 0, $1668 = 0, $1669 = 0;
 var $167 = 0, $1670 = 0, $1671 = 0, $1672 = 0, $1673 = 0, $1674 = 0, $1675 = 0, $1676 = 0, $1677 = 0, $1678 = 0, $1679 = 0, $168 = 0, $1680 = 0, $1681 = 0, $1682 = 0, $1683 = 0, $1684 = 0, $1685 = 0, $1686 = 0, $1687 = 0;
 var $1688 = 0, $1689 = 0, $169 = 0, $1690 = 0, $1691 = 0, $1692 = 0, $1693 = 0, $1694 = 0, $1695 = 0, $1696 = 0, $1697 = 0, $1698 = 0, $1699 = 0, $17 = 0, $170 = 0, $1700 = 0, $1701 = 0, $1702 = 0, $1703 = 0, $1704 = 0;
 var $1705 = 0, $1706 = 0, $1707 = 0, $1708 = 0, $1709 = 0, $171 = 0, $1710 = 0, $1711 = 0, $1712 = 0, $1713 = 0, $1714 = 0, $1715 = 0, $1716 = 0, $1717 = 0, $1718 = 0, $1719 = 0, $172 = 0, $1720 = 0, $1721 = 0, $1722 = 0;
 var $1723 = 0, $1724 = 0, $1725 = 0, $1726 = 0, $1727 = 0, $1728 = 0, $1729 = 0, $173 = 0, $1730 = 0, $1731 = 0, $1732 = 0, $1733 = 0, $1734 = 0, $1735 = 0, $1736 = 0, $1737 = 0, $1738 = 0, $1739 = 0, $174 = 0, $1740 = 0;
 var $1741 = 0, $1742 = 0, $1743 = 0, $1744 = 0, $1745 = 0, $1746 = 0, $1747 = 0, $1748 = 0, $1749 = 0, $175 = 0, $1750 = 0, $1751 = 0, $1752 = 0, $1753 = 0, $1754 = 0, $1755 = 0, $1756 = 0, $1757 = 0, $1758 = 0, $1759 = 0;
 var $176 = 0, $1760 = 0, $1761 = 0, $1762 = 0, $1763 = 0, $1764 = 0, $1765 = 0, $1766 = 0, $1767 = 0, $1768 = 0, $1769 = 0, $177 = 0, $1770 = 0, $1771 = 0, $1772 = 0, $1773 = 0, $1774 = 0, $1775 = 0, $1776 = 0, $1777 = 0;
 var $1778 = 0, $1779 = 0, $178 = 0, $1780 = 0, $1781 = 0, $1782 = 0, $1783 = 0, $1784 = 0, $1785 = 0, $1786 = 0, $1787 = 0, $1788 = 0, $1789 = 0, $179 = 0, $1790 = 0, $1791 = 0, $1792 = 0, $1793 = 0, $1794 = 0, $1795 = 0;
 var $1796 = 0, $1797 = 0, $1798 = 0, $1799 = 0, $18 = 0, $180 = 0, $1800 = 0, $1801 = 0, $1802 = 0, $1803 = 0, $1804 = 0, $1805 = 0, $1806 = 0, $1807 = 0, $1808 = 0, $1809 = 0, $181 = 0, $1810 = 0, $1811 = 0, $1812 = 0;
 var $1813 = 0, $1814 = 0, $1815 = 0, $1816 = 0, $1817 = 0, $1818 = 0, $1819 = 0, $182 = 0, $1820 = 0, $1821 = 0, $1822 = 0, $1823 = 0, $1824 = 0, $1825 = 0, $1826 = 0, $1827 = 0, $1828 = 0, $1829 = 0, $183 = 0, $1830 = 0;
 var $1831 = 0, $1832 = 0, $1833 = 0, $1834 = 0, $1835 = 0, $1836 = 0, $1837 = 0, $1838 = 0, $1839 = 0, $184 = 0, $1840 = 0, $1841 = 0, $1842 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0;
 var $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0;
 var $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0;
 var $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0;
 var $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0;
 var $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0.0, $279 = 0.0, $28 = 0, $280 = 0.0;
 var $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0.0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0.0, $294 = 0.0, $295 = 0.0, $296 = 0.0, $297 = 0, $298 = 0, $299 = 0;
 var $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0;
 var $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0;
 var $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0;
 var $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0;
 var $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0;
 var $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0.0, $397 = 0.0, $398 = 0.0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0.0, $405 = 0, $406 = 0;
 var $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0.0, $412 = 0.0, $413 = 0.0, $414 = 0.0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0;
 var $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0;
 var $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0;
 var $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0;
 var $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0;
 var $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0.0;
 var $515 = 0.0, $516 = 0.0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0.0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0.0, $53 = 0, $530 = 0.0, $531 = 0.0, $532 = 0.0;
 var $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0;
 var $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0;
 var $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0;
 var $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0;
 var $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0;
 var $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0.0, $633 = 0.0, $634 = 0.0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0.0;
 var $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0.0, $648 = 0.0, $649 = 0.0, $65 = 0, $650 = 0.0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0;
 var $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0;
 var $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0;
 var $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0;
 var $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0;
 var $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0;
 var $75 = 0, $750 = 0.0, $751 = 0.0, $752 = 0.0, $753 = 0.0, $754 = 0.0, $755 = 0.0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0.0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0;
 var $768 = 0.0, $769 = 0.0, $77 = 0, $770 = 0.0, $771 = 0.0, $772 = 0.0, $773 = 0.0, $774 = 0.0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0;
 var $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0;
 var $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0;
 var $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0;
 var $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0;
 var $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0.0, $875 = 0.0;
 var $876 = 0.0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0.0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0.0, $89 = 0, $890 = 0.0, $891 = 0.0, $892 = 0.0, $893 = 0;
 var $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0;
 var $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0;
 var $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0;
 var $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0;
 var $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0.0, $977 = 0.0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0.0;
 var $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0.0, $99 = 0, $990 = 0.0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $aux$0199 = 0, $b79$0 = 0, $base$0 = 0;
 var $base$1 = 0, $base$1$lcssa = 0, $base$12 = 0, $brmerge = 0, $c75$0 = 0, $cond = 0, $cond248 = 0, $cond249 = 0, $cond250 = 0, $cond251 = 0, $cond252 = 0, $cond253 = 0, $cond254 = 0, $cond255 = 0, $cond256 = 0, $exitcond = 0, $exitcond238 = 0, $g = 0, $g7 = 0, $j$0193 = 0;
 var $j80$0192 = 0, $l$09$i$i = 0, $last$0197 = 0, $ll$010$i$i = 0, $lr$012$i$i = 0, $n$0 = 0, $n$1196 = 0, $num$i = 0, $num$i102 = 0, $num$i105 = 0, $num$i116 = 0, $num$i119 = 0, $num$i26 = 0, $num$i29 = 0, $num$i32 = 0, $num$i35 = 0, $num$i46 = 0, $num$i49 = 0, $num$i60 = 0, $num$i63 = 0;
 var $num$i74 = 0, $num$i77 = 0, $num$i88 = 0, $num$i91 = 0, $or$cond = 0, $or$cond$i = 0, $pc$0 = 0, $pc$0$lcssa = 0, $pc$10$lcssa = 0, $pc$10194 = 0, $pc$2 = 0, $pc$3 = 0, $pc$4 = 0, $pc$5 = 0, $pc$6 = 0, $pc$8 = 0, $pc$9 = 0, $phitmp$i$i = 0, $phitmp$i1$i = 0, $r$011$i$i = 0;
 var $ra$0 = 0, $rb$0 = 0, $scevgep237 = 0, $tempb$i = 0, $tempb$i11 = 0, $tempb$i14 = 0, $tempb$i17 = 0, $tempb$i20 = 0, $tempb$i5 = 0, $tempb$i8 = 0, $tempc$i = 0, $tempc$i12 = 0, $tempc$i15 = 0, $tempc$i18 = 0, $tempc$i6 = 0, $tempc$i9 = 0, $tm$01$i = 0, $tm$01$i110 = 0, $tm$01$i124 = 0, $tm$01$i40 = 0;
 var $tm$01$i54 = 0, $tm$01$i68 = 0, $tm$01$i82 = 0, $tm$01$i96 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 400|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 392|0;
 $vararg_buffer1 = sp + 384|0;
 $vararg_buffer = sp + 376|0;
 $num$i119 = sp + 368|0;
 $num$i116 = sp + 360|0;
 $num$i105 = sp + 352|0;
 $num$i102 = sp + 344|0;
 $num$i91 = sp + 336|0;
 $num$i88 = sp + 328|0;
 $num$i77 = sp + 320|0;
 $num$i74 = sp + 312|0;
 $num$i63 = sp + 304|0;
 $num$i60 = sp + 296|0;
 $num$i49 = sp + 288|0;
 $num$i46 = sp + 280|0;
 $num$i35 = sp + 272|0;
 $num$i32 = sp + 264|0;
 $num$i29 = sp + 256|0;
 $num$i26 = sp + 248|0;
 $num$i = sp + 240|0;
 $tempb$i20 = sp + 224|0;
 $tempb$i17 = sp + 208|0;
 $tempc$i18 = sp + 192|0;
 $tempb$i14 = sp + 176|0;
 $tempc$i15 = sp + 160|0;
 $tempb$i11 = sp + 144|0;
 $tempc$i12 = sp + 128|0;
 $tempb$i8 = sp + 112|0;
 $tempc$i9 = sp + 96|0;
 $tempb$i5 = sp + 80|0;
 $tempc$i6 = sp + 64|0;
 $tempb$i = sp + 48|0;
 $tempc$i = sp + 32|0;
 $g = sp + 16|0;
 $g7 = sp;
 $0 = ((($L)) + 24|0);
 $1 = ((($L)) + 20|0);
 $2 = ((($L)) + 12|0);
 $3 = ((($L)) + 56|0);
 $4 = ((($g)) + 8|0);
 $5 = ((($g7)) + 8|0);
 $6 = ((($L)) + 16|0);
 $7 = ((($L)) + 32|0);
 $8 = ((($L)) + 8|0);
 $9 = ((($L)) + 28|0);
 $10 = ((($tempc$i)) + 8|0);
 $11 = ((($tempb$i)) + 8|0);
 $12 = ((($tempc$i6)) + 8|0);
 $13 = ((($tempb$i5)) + 8|0);
 $14 = ((($tempc$i9)) + 8|0);
 $15 = ((($tempb$i8)) + 8|0);
 $16 = ((($tempc$i12)) + 8|0);
 $17 = ((($tempb$i11)) + 8|0);
 $18 = ((($tempc$i15)) + 8|0);
 $19 = ((($tempb$i14)) + 8|0);
 $20 = ((($tempc$i18)) + 8|0);
 $21 = ((($tempb$i17)) + 8|0);
 $22 = ((($tempb$i20)) + 8|0);
 $23 = 624096;
 $24 = $23;
 $25 = HEAP32[$24>>2]|0;
 $26 = (($23) + 4)|0;
 $27 = $26;
 $28 = HEAP32[$27>>2]|0;
 $29 = HEAP32[(624104)>>2]|0;
 $30 = ((($L)) + 64|0);
 $31 = ((($L)) + 6|0);
 $32 = ((($L)) + 60|0);
 $33 = ((($L)) + 104|0);
 $$0$ph = $nexeccalls;
 L1: while(1) {
  $$pre = HEAP32[$0>>2]|0;
  $$pre239 = HEAP32[$1>>2]|0;
  $1842 = $$pre;$35 = $$pre239;
  L3: while(1) {
   $34 = ((($35)) + 4|0);
   $36 = HEAP32[$34>>2]|0;
   $37 = HEAP32[$36>>2]|0;
   $38 = HEAP32[$2>>2]|0;
   $39 = ((($37)) + 16|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = ((($40)) + 8|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = ((($37)) + 12|0);
   $base$0 = $38;$pc$0 = $1842;
   L5: while(1) {
    $44 = ((($pc$0)) + 4|0);
    $45 = HEAP32[$pc$0>>2]|0;
    $46 = HEAP8[$3>>0]|0;
    $47 = $46 & 12;
    $48 = ($47<<24>>24)==(0);
    if ($48) {
     $base$1 = $base$0;
    } else {
     $49 = HEAP32[$30>>2]|0;
     $50 = (($49) + -1)|0;
     HEAP32[$30>>2] = $50;
     $51 = ($50|0)==(0);
     $$not = $51 ^ 1;
     $52 = $46 & 4;
     $53 = ($52<<24>>24)==(0);
     $or$cond = $53 & $$not;
     if ($or$cond) {
      $base$1 = $base$0;
     } else {
      $54 = HEAP32[$0>>2]|0;
      HEAP32[$0>>2] = $44;
      $55 = $46&255;
      $56 = $55 & 8;
      $57 = ($56|0)==(0);
      $$not289 = $51 ^ 1;
      $brmerge = $57 | $$not289;
      if (!($brmerge)) {
       $58 = HEAP32[$32>>2]|0;
       HEAP32[$30>>2] = $58;
       _luaD_callhook($L,3,-1);
      }
      $59 = $55 & 4;
      $60 = ($59|0)==(0);
      do {
       if (!($60)) {
        $61 = HEAP32[$1>>2]|0;
        $62 = ((($61)) + 4|0);
        $63 = HEAP32[$62>>2]|0;
        $64 = HEAP32[$63>>2]|0;
        $65 = ((($64)) + 16|0);
        $66 = HEAP32[$65>>2]|0;
        $67 = ((($66)) + 12|0);
        $68 = HEAP32[$67>>2]|0;
        $69 = $44;
        $70 = (($69) - ($68))|0;
        $71 = $70 >> 2;
        $72 = (($71) + -1)|0;
        $73 = ((($66)) + 20|0);
        $74 = HEAP32[$73>>2]|0;
        $75 = ($74|0)==(0|0);
        if ($75) {
         $87 = 0;
        } else {
         $76 = (($74) + ($72<<2)|0);
         $77 = HEAP32[$76>>2]|0;
         $87 = $77;
        }
        $78 = ($72|0)!=(0);
        $79 = ($54>>>0)<($44>>>0);
        $or$cond$i = $79 & $78;
        if ($or$cond$i) {
         if ($75) {
          $88 = 0;
         } else {
          $80 = $54;
          $81 = (($80) - ($68))|0;
          $82 = $81 >> 2;
          $83 = (($82) + -1)|0;
          $84 = (($74) + ($83<<2)|0);
          $85 = HEAP32[$84>>2]|0;
          $88 = $85;
         }
         $86 = ($87|0)==($88|0);
         if ($86) {
          break;
         }
        }
        _luaD_callhook($L,2,$87);
       }
      } while(0);
      $89 = HEAP8[$31>>0]|0;
      $90 = ($89<<24>>24)==(1);
      if ($90) {
       $pc$0$lcssa = $pc$0;
       label = 17;
       break L1;
      }
      $91 = HEAP32[$2>>2]|0;
      $base$1 = $91;
     }
    }
    $92 = $45 >>> 6;
    $93 = $92 & 255;
    $94 = (($base$1) + ($93<<4)|0);
    $95 = $45 & 63;
    do {
     switch ($95|0) {
     case 30:  {
      $$lcssa359 = $44;$$lcssa365 = $45;$$lcssa374 = $94;$base$1$lcssa = $base$1;
      break L3;
      break;
     }
     case 0:  {
      $96 = $45 >>> 23;
      $97 = (($base$1) + ($96<<4)|0);
      $98 = $97;
      $99 = $98;
      $100 = HEAP32[$99>>2]|0;
      $101 = (($98) + 4)|0;
      $102 = $101;
      $103 = HEAP32[$102>>2]|0;
      $104 = $94;
      $105 = $104;
      HEAP32[$105>>2] = $100;
      $106 = (($104) + 4)|0;
      $107 = $106;
      HEAP32[$107>>2] = $103;
      $108 = (((($base$1) + ($96<<4)|0)) + 8|0);
      $109 = HEAP32[$108>>2]|0;
      $110 = (((($base$1) + ($93<<4)|0)) + 8|0);
      HEAP32[$110>>2] = $109;
      $base$0 = $base$1;$pc$0 = $44;
      continue L5;
      break;
     }
     case 1:  {
      $111 = $45 >>> 14;
      $112 = (($42) + ($111<<4)|0);
      $113 = $112;
      $114 = $113;
      $115 = HEAP32[$114>>2]|0;
      $116 = (($113) + 4)|0;
      $117 = $116;
      $118 = HEAP32[$117>>2]|0;
      $119 = $94;
      $120 = $119;
      HEAP32[$120>>2] = $115;
      $121 = (($119) + 4)|0;
      $122 = $121;
      HEAP32[$122>>2] = $118;
      $123 = (((($42) + ($111<<4)|0)) + 8|0);
      $124 = HEAP32[$123>>2]|0;
      $125 = (((($base$1) + ($93<<4)|0)) + 8|0);
      HEAP32[$125>>2] = $124;
      $base$0 = $base$1;$pc$0 = $44;
      continue L5;
      break;
     }
     case 2:  {
      $126 = $45 >>> 23;
      HEAP32[$94>>2] = $126;
      $127 = (((($base$1) + ($93<<4)|0)) + 8|0);
      HEAP32[$127>>2] = 1;
      $128 = $45 & 8372224;
      $129 = ($128|0)==(0);
      $130 = ((($pc$0)) + 8|0);
      $$ = $129 ? $44 : $130;
      $base$0 = $base$1;$pc$0 = $$;
      continue L5;
      break;
     }
     case 3:  {
      $131 = $45 >>> 23;
      $132 = (($base$1) + ($131<<4)|0);
      $rb$0 = $132;
      while(1) {
       $133 = ((($rb$0)) + -16|0);
       $134 = ((($rb$0)) + 8|0);
       HEAP32[$134>>2] = 0;
       $135 = ($133>>>0)<($94>>>0);
       if ($135) {
        $base$0 = $base$1;$pc$0 = $44;
        continue L5;
       } else {
        $rb$0 = $133;
       }
      }
      break;
     }
     case 4:  {
      $136 = $45 >>> 23;
      $137 = (((($37)) + 20|0) + ($136<<2)|0);
      $138 = HEAP32[$137>>2]|0;
      $139 = ((($138)) + 8|0);
      $140 = HEAP32[$139>>2]|0;
      $141 = $140;
      $142 = $141;
      $143 = HEAP32[$142>>2]|0;
      $144 = (($141) + 4)|0;
      $145 = $144;
      $146 = HEAP32[$145>>2]|0;
      $147 = $94;
      $148 = $147;
      HEAP32[$148>>2] = $143;
      $149 = (($147) + 4)|0;
      $150 = $149;
      HEAP32[$150>>2] = $146;
      $151 = ((($140)) + 8|0);
      $152 = HEAP32[$151>>2]|0;
      $153 = (((($base$1) + ($93<<4)|0)) + 8|0);
      HEAP32[$153>>2] = $152;
      $base$0 = $base$1;$pc$0 = $44;
      continue L5;
      break;
     }
     case 5:  {
      $154 = $45 >>> 14;
      $155 = (($42) + ($154<<4)|0);
      $156 = HEAP32[$43>>2]|0;
      HEAP32[$g>>2] = $156;
      HEAP32[$4>>2] = 5;
      HEAP32[$0>>2] = $44;
      _luaV_gettable($L,$g,$155,$94);
      $157 = HEAP32[$2>>2]|0;
      $base$0 = $157;$pc$0 = $44;
      continue L5;
      break;
     }
     case 6:  {
      HEAP32[$0>>2] = $44;
      $158 = $45 >>> 23;
      $159 = (($base$1) + ($158<<4)|0);
      $160 = $45 >>> 14;
      $161 = $160 & 256;
      $162 = ($161|0)==(0);
      $163 = $160 & 255;
      $164 = (($42) + ($163<<4)|0);
      $165 = $160 & 511;
      $166 = (($base$1) + ($165<<4)|0);
      $167 = $162 ? $166 : $164;
      _luaV_gettable($L,$159,$167,$94);
      $168 = HEAP32[$2>>2]|0;
      $base$0 = $168;$pc$0 = $44;
      continue L5;
      break;
     }
     case 7:  {
      $169 = HEAP32[$43>>2]|0;
      HEAP32[$g7>>2] = $169;
      HEAP32[$5>>2] = 5;
      HEAP32[$0>>2] = $44;
      $170 = $45 >>> 14;
      $171 = (($42) + ($170<<4)|0);
      _luaV_settable($L,$g7,$171,$94);
      $172 = HEAP32[$2>>2]|0;
      $base$0 = $172;$pc$0 = $44;
      continue L5;
      break;
     }
     case 8:  {
      $173 = $45 >>> 23;
      $174 = (((($37)) + 20|0) + ($173<<2)|0);
      $175 = HEAP32[$174>>2]|0;
      $176 = ((($175)) + 8|0);
      $177 = HEAP32[$176>>2]|0;
      $178 = $94;
      $179 = $178;
      $180 = HEAP32[$179>>2]|0;
      $181 = (($178) + 4)|0;
      $182 = $181;
      $183 = HEAP32[$182>>2]|0;
      $184 = $177;
      $185 = $184;
      HEAP32[$185>>2] = $180;
      $186 = (($184) + 4)|0;
      $187 = $186;
      HEAP32[$187>>2] = $183;
      $188 = (((($base$1) + ($93<<4)|0)) + 8|0);
      $189 = HEAP32[$188>>2]|0;
      $190 = ((($177)) + 8|0);
      HEAP32[$190>>2] = $189;
      $191 = HEAP32[$188>>2]|0;
      $192 = ($191|0)>(3);
      if (!($192)) {
       $base$0 = $base$1;$pc$0 = $44;
       continue L5;
      }
      $193 = HEAP32[$94>>2]|0;
      $194 = ((($193)) + 5|0);
      $195 = HEAP8[$194>>0]|0;
      $196 = $195 & 3;
      $197 = ($196<<24>>24)==(0);
      if ($197) {
       $base$0 = $base$1;$pc$0 = $44;
       continue L5;
      }
      $198 = ((($175)) + 5|0);
      $199 = HEAP8[$198>>0]|0;
      $200 = $199 & 4;
      $201 = ($200<<24>>24)==(0);
      if ($201) {
       $base$0 = $base$1;$pc$0 = $44;
       continue L5;
      }
      _luaC_barrierf($L,$175,$193);
      $base$0 = $base$1;$pc$0 = $44;
      continue L5;
      break;
     }
     case 9:  {
      HEAP32[$0>>2] = $44;
      $202 = $45 >>> 23;
      $203 = $202 & 256;
      $204 = ($203|0)==(0);
      $205 = $202 & 255;
      $206 = (($42) + ($205<<4)|0);
      $207 = (($base$1) + ($202<<4)|0);
      $208 = $204 ? $207 : $206;
      $209 = $45 >>> 14;
      $210 = $209 & 256;
      $211 = ($210|0)==(0);
      $212 = $209 & 255;
      $213 = (($42) + ($212<<4)|0);
      $214 = $209 & 511;
      $215 = (($base$1) + ($214<<4)|0);
      $216 = $211 ? $215 : $213;
      _luaV_settable($L,$94,$208,$216);
      $217 = HEAP32[$2>>2]|0;
      $base$0 = $217;$pc$0 = $44;
      continue L5;
      break;
     }
     case 10:  {
      $218 = $45 >>> 23;
      $219 = $45 >>> 14;
      $220 = $219 & 511;
      $221 = (_luaO_fb2int($218)|0);
      $222 = (_luaO_fb2int($220)|0);
      $223 = (_luaH_new($L,$221,$222)|0);
      HEAP32[$94>>2] = $223;
      $224 = (((($base$1) + ($93<<4)|0)) + 8|0);
      HEAP32[$224>>2] = 5;
      HEAP32[$0>>2] = $44;
      $225 = HEAP32[$6>>2]|0;
      $226 = ((($225)) + 68|0);
      $227 = HEAP32[$226>>2]|0;
      $228 = ((($225)) + 64|0);
      $229 = HEAP32[$228>>2]|0;
      $230 = ($227>>>0)<($229>>>0);
      if (!($230)) {
       _luaC_step($L);
      }
      $231 = HEAP32[$2>>2]|0;
      $base$0 = $231;$pc$0 = $44;
      continue L5;
      break;
     }
     case 11:  {
      $232 = $45 >>> 23;
      $233 = (($base$1) + ($232<<4)|0);
      $234 = ((($94)) + 16|0);
      $235 = $233;
      $236 = $235;
      $237 = HEAP32[$236>>2]|0;
      $238 = (($235) + 4)|0;
      $239 = $238;
      $240 = HEAP32[$239>>2]|0;
      $241 = $234;
      $242 = $241;
      HEAP32[$242>>2] = $237;
      $243 = (($241) + 4)|0;
      $244 = $243;
      HEAP32[$244>>2] = $240;
      $245 = (((($base$1) + ($232<<4)|0)) + 8|0);
      $246 = HEAP32[$245>>2]|0;
      $247 = ((($234)) + 8|0);
      HEAP32[$247>>2] = $246;
      HEAP32[$0>>2] = $44;
      $248 = $45 >>> 14;
      $249 = $248 & 256;
      $250 = ($249|0)==(0);
      $251 = $248 & 255;
      $252 = (($42) + ($251<<4)|0);
      $253 = $248 & 511;
      $254 = (($base$1) + ($253<<4)|0);
      $255 = $250 ? $254 : $252;
      _luaV_gettable($L,$233,$255,$94);
      $256 = HEAP32[$2>>2]|0;
      $base$0 = $256;$pc$0 = $44;
      continue L5;
      break;
     }
     case 12:  {
      $257 = $45 >>> 23;
      $258 = $257 & 256;
      $259 = ($258|0)==(0);
      $260 = $257 & 255;
      $261 = (($42) + ($260<<4)|0);
      $262 = (($base$1) + ($257<<4)|0);
      $263 = $259 ? $262 : $261;
      $264 = $45 >>> 14;
      $265 = $264 & 256;
      $266 = ($265|0)==(0);
      $267 = $264 & 255;
      $268 = (($42) + ($267<<4)|0);
      $269 = $264 & 511;
      $270 = (($base$1) + ($269<<4)|0);
      $271 = $266 ? $270 : $268;
      $272 = ((($263)) + 8|0);
      $273 = HEAP32[$272>>2]|0;
      $274 = ($273|0)==(3);
      do {
       if ($274) {
        $275 = ((($271)) + 8|0);
        $276 = HEAP32[$275>>2]|0;
        $277 = ($276|0)==(3);
        if ($277) {
         $278 = +HEAPF64[$263>>3];
         $279 = +HEAPF64[$271>>3];
         $280 = $278 + $279;
         HEAPF64[$94>>3] = $280;
         $281 = (((($base$1) + ($93<<4)|0)) + 8|0);
         HEAP32[$281>>2] = 3;
         $base$0 = $base$1;$pc$0 = $44;
         continue L5;
        } else {
         HEAP32[$0>>2] = $44;
         $$0$i33$ph = $263;
         label = 46;
         break;
        }
       } else {
        HEAP32[$0>>2] = $44;
        $cond = ($273|0)==(4);
        if ($cond) {
         $282 = HEAP32[$263>>2]|0;
         $283 = ((($282)) + 16|0);
         $284 = (_luaO_str2d($283,$num$i32)|0);
         $285 = ($284|0)==(0);
         if (!($285)) {
          $286 = +HEAPF64[$num$i32>>3];
          HEAPF64[$tempb$i>>3] = $286;
          HEAP32[$11>>2] = 3;
          $$0$i33$ph = $tempb$i;
          label = 46;
          break;
         }
        }
        label = 51;
       }
      } while(0);
      do {
       if ((label|0) == 46) {
        label = 0;
        $287 = ((($271)) + 8|0);
        $288 = HEAP32[$287>>2]|0;
        switch ($288|0) {
        case 3:  {
         $$0$i36$ph = $271;
         label = 50;
         break;
        }
        case 4:  {
         $289 = HEAP32[$271>>2]|0;
         $290 = ((($289)) + 16|0);
         $291 = (_luaO_str2d($290,$num$i35)|0);
         $292 = ($291|0)==(0);
         if ($292) {
          label = 49;
         } else {
          $293 = +HEAPF64[$num$i35>>3];
          HEAPF64[$tempc$i>>3] = $293;
          HEAP32[$10>>2] = 3;
          $$0$i36$ph = $tempc$i;
          label = 50;
         }
         break;
        }
        default: {
         label = 49;
        }
        }
        if ((label|0) == 49) {
         label = 0;
         label = 51;
         break;
        }
        else if ((label|0) == 50) {
         label = 0;
         $294 = +HEAPF64[$$0$i33$ph>>3];
         $295 = +HEAPF64[$$0$i36$ph>>3];
         $296 = $294 + $295;
         HEAPF64[$94>>3] = $296;
         $297 = (((($base$1) + ($93<<4)|0)) + 8|0);
         HEAP32[$297>>2] = 3;
         break;
        }
       }
      } while(0);
      do {
       if ((label|0) == 51) {
        label = 0;
        $298 = (_luaT_gettmbyobj($L,$263,5)|0);
        $299 = ((($298)) + 8|0);
        $300 = HEAP32[$299>>2]|0;
        $301 = ($300|0)==(0);
        if ($301) {
         $302 = (_luaT_gettmbyobj($L,$271,5)|0);
         $$phi$trans$insert$i38 = ((($302)) + 8|0);
         $$pre$i39 = HEAP32[$$phi$trans$insert$i38>>2]|0;
         $303 = ($$pre$i39|0)==(0);
         if ($303) {
          _luaG_aritherror($L,$263,$271);
          break;
         } else {
          $tm$01$i40 = $302;
         }
        } else {
         $tm$01$i40 = $298;
        }
        $304 = HEAP32[$7>>2]|0;
        $305 = $94;
        $306 = (($305) - ($304))|0;
        $307 = HEAP32[$8>>2]|0;
        $308 = $tm$01$i40;
        $309 = $308;
        $310 = HEAP32[$309>>2]|0;
        $311 = (($308) + 4)|0;
        $312 = $311;
        $313 = HEAP32[$312>>2]|0;
        $314 = $307;
        $315 = $314;
        HEAP32[$315>>2] = $310;
        $316 = (($314) + 4)|0;
        $317 = $316;
        HEAP32[$317>>2] = $313;
        $318 = ((($tm$01$i40)) + 8|0);
        $319 = HEAP32[$318>>2]|0;
        $320 = ((($307)) + 8|0);
        HEAP32[$320>>2] = $319;
        $321 = HEAP32[$8>>2]|0;
        $322 = ((($321)) + 16|0);
        $323 = $263;
        $324 = $323;
        $325 = HEAP32[$324>>2]|0;
        $326 = (($323) + 4)|0;
        $327 = $326;
        $328 = HEAP32[$327>>2]|0;
        $329 = $322;
        $330 = $329;
        HEAP32[$330>>2] = $325;
        $331 = (($329) + 4)|0;
        $332 = $331;
        HEAP32[$332>>2] = $328;
        $333 = HEAP32[$272>>2]|0;
        $334 = ((($321)) + 24|0);
        HEAP32[$334>>2] = $333;
        $335 = HEAP32[$8>>2]|0;
        $336 = ((($335)) + 32|0);
        $337 = $271;
        $338 = $337;
        $339 = HEAP32[$338>>2]|0;
        $340 = (($337) + 4)|0;
        $341 = $340;
        $342 = HEAP32[$341>>2]|0;
        $343 = $336;
        $344 = $343;
        HEAP32[$344>>2] = $339;
        $345 = (($343) + 4)|0;
        $346 = $345;
        HEAP32[$346>>2] = $342;
        $347 = ((($271)) + 8|0);
        $348 = HEAP32[$347>>2]|0;
        $349 = ((($335)) + 40|0);
        HEAP32[$349>>2] = $348;
        $350 = HEAP32[$9>>2]|0;
        $351 = HEAP32[$8>>2]|0;
        $352 = (($350) - ($351))|0;
        $353 = ($352|0)<(49);
        $354 = $351;
        if ($353) {
         _luaD_growstack($L,3);
         $$pre$i44 = HEAP32[$8>>2]|0;
         $356 = $$pre$i44;
        } else {
         $356 = $354;
        }
        $355 = ((($356)) + 48|0);
        HEAP32[$8>>2] = $355;
        _luaD_call($L,$356,1);
        $357 = HEAP32[$7>>2]|0;
        $358 = (($357) + ($306)|0);
        $359 = HEAP32[$8>>2]|0;
        $360 = ((($359)) + -16|0);
        HEAP32[$8>>2] = $360;
        $361 = $360;
        $362 = $361;
        $363 = HEAP32[$362>>2]|0;
        $364 = (($361) + 4)|0;
        $365 = $364;
        $366 = HEAP32[$365>>2]|0;
        $367 = $358;
        $368 = $367;
        HEAP32[$368>>2] = $363;
        $369 = (($367) + 4)|0;
        $370 = $369;
        HEAP32[$370>>2] = $366;
        $371 = ((($359)) + -8|0);
        $372 = HEAP32[$371>>2]|0;
        $373 = ((($358)) + 8|0);
        HEAP32[$373>>2] = $372;
       }
      } while(0);
      $374 = HEAP32[$2>>2]|0;
      $base$0 = $374;$pc$0 = $44;
      continue L5;
      break;
     }
     case 13:  {
      $375 = $45 >>> 23;
      $376 = $375 & 256;
      $377 = ($376|0)==(0);
      $378 = $375 & 255;
      $379 = (($42) + ($378<<4)|0);
      $380 = (($base$1) + ($375<<4)|0);
      $381 = $377 ? $380 : $379;
      $382 = $45 >>> 14;
      $383 = $382 & 256;
      $384 = ($383|0)==(0);
      $385 = $382 & 255;
      $386 = (($42) + ($385<<4)|0);
      $387 = $382 & 511;
      $388 = (($base$1) + ($387<<4)|0);
      $389 = $384 ? $388 : $386;
      $390 = ((($381)) + 8|0);
      $391 = HEAP32[$390>>2]|0;
      $392 = ($391|0)==(3);
      do {
       if ($392) {
        $393 = ((($389)) + 8|0);
        $394 = HEAP32[$393>>2]|0;
        $395 = ($394|0)==(3);
        if ($395) {
         $396 = +HEAPF64[$381>>3];
         $397 = +HEAPF64[$389>>3];
         $398 = $396 - $397;
         HEAPF64[$94>>3] = $398;
         $399 = (((($base$1) + ($93<<4)|0)) + 8|0);
         HEAP32[$399>>2] = 3;
         $base$0 = $base$1;$pc$0 = $44;
         continue L5;
        } else {
         HEAP32[$0>>2] = $44;
         $$0$i47$ph = $381;
         label = 66;
         break;
        }
       } else {
        HEAP32[$0>>2] = $44;
        $cond248 = ($391|0)==(4);
        if ($cond248) {
         $400 = HEAP32[$381>>2]|0;
         $401 = ((($400)) + 16|0);
         $402 = (_luaO_str2d($401,$num$i46)|0);
         $403 = ($402|0)==(0);
         if (!($403)) {
          $404 = +HEAPF64[$num$i46>>3];
          HEAPF64[$tempb$i5>>3] = $404;
          HEAP32[$13>>2] = 3;
          $$0$i47$ph = $tempb$i5;
          label = 66;
          break;
         }
        }
        label = 71;
       }
      } while(0);
      do {
       if ((label|0) == 66) {
        label = 0;
        $405 = ((($389)) + 8|0);
        $406 = HEAP32[$405>>2]|0;
        switch ($406|0) {
        case 3:  {
         $$0$i50$ph = $389;
         label = 70;
         break;
        }
        case 4:  {
         $407 = HEAP32[$389>>2]|0;
         $408 = ((($407)) + 16|0);
         $409 = (_luaO_str2d($408,$num$i49)|0);
         $410 = ($409|0)==(0);
         if ($410) {
          label = 69;
         } else {
          $411 = +HEAPF64[$num$i49>>3];
          HEAPF64[$tempc$i6>>3] = $411;
          HEAP32[$12>>2] = 3;
          $$0$i50$ph = $tempc$i6;
          label = 70;
         }
         break;
        }
        default: {
         label = 69;
        }
        }
        if ((label|0) == 69) {
         label = 0;
         label = 71;
         break;
        }
        else if ((label|0) == 70) {
         label = 0;
         $412 = +HEAPF64[$$0$i47$ph>>3];
         $413 = +HEAPF64[$$0$i50$ph>>3];
         $414 = $412 - $413;
         HEAPF64[$94>>3] = $414;
         $415 = (((($base$1) + ($93<<4)|0)) + 8|0);
         HEAP32[$415>>2] = 3;
         break;
        }
       }
      } while(0);
      do {
       if ((label|0) == 71) {
        label = 0;
        $416 = (_luaT_gettmbyobj($L,$381,6)|0);
        $417 = ((($416)) + 8|0);
        $418 = HEAP32[$417>>2]|0;
        $419 = ($418|0)==(0);
        if ($419) {
         $420 = (_luaT_gettmbyobj($L,$389,6)|0);
         $$phi$trans$insert$i52 = ((($420)) + 8|0);
         $$pre$i53 = HEAP32[$$phi$trans$insert$i52>>2]|0;
         $421 = ($$pre$i53|0)==(0);
         if ($421) {
          _luaG_aritherror($L,$381,$389);
          break;
         } else {
          $tm$01$i54 = $420;
         }
        } else {
         $tm$01$i54 = $416;
        }
        $422 = HEAP32[$7>>2]|0;
        $423 = $94;
        $424 = (($423) - ($422))|0;
        $425 = HEAP32[$8>>2]|0;
        $426 = $tm$01$i54;
        $427 = $426;
        $428 = HEAP32[$427>>2]|0;
        $429 = (($426) + 4)|0;
        $430 = $429;
        $431 = HEAP32[$430>>2]|0;
        $432 = $425;
        $433 = $432;
        HEAP32[$433>>2] = $428;
        $434 = (($432) + 4)|0;
        $435 = $434;
        HEAP32[$435>>2] = $431;
        $436 = ((($tm$01$i54)) + 8|0);
        $437 = HEAP32[$436>>2]|0;
        $438 = ((($425)) + 8|0);
        HEAP32[$438>>2] = $437;
        $439 = HEAP32[$8>>2]|0;
        $440 = ((($439)) + 16|0);
        $441 = $381;
        $442 = $441;
        $443 = HEAP32[$442>>2]|0;
        $444 = (($441) + 4)|0;
        $445 = $444;
        $446 = HEAP32[$445>>2]|0;
        $447 = $440;
        $448 = $447;
        HEAP32[$448>>2] = $443;
        $449 = (($447) + 4)|0;
        $450 = $449;
        HEAP32[$450>>2] = $446;
        $451 = HEAP32[$390>>2]|0;
        $452 = ((($439)) + 24|0);
        HEAP32[$452>>2] = $451;
        $453 = HEAP32[$8>>2]|0;
        $454 = ((($453)) + 32|0);
        $455 = $389;
        $456 = $455;
        $457 = HEAP32[$456>>2]|0;
        $458 = (($455) + 4)|0;
        $459 = $458;
        $460 = HEAP32[$459>>2]|0;
        $461 = $454;
        $462 = $461;
        HEAP32[$462>>2] = $457;
        $463 = (($461) + 4)|0;
        $464 = $463;
        HEAP32[$464>>2] = $460;
        $465 = ((($389)) + 8|0);
        $466 = HEAP32[$465>>2]|0;
        $467 = ((($453)) + 40|0);
        HEAP32[$467>>2] = $466;
        $468 = HEAP32[$9>>2]|0;
        $469 = HEAP32[$8>>2]|0;
        $470 = (($468) - ($469))|0;
        $471 = ($470|0)<(49);
        $472 = $469;
        if ($471) {
         _luaD_growstack($L,3);
         $$pre$i58 = HEAP32[$8>>2]|0;
         $474 = $$pre$i58;
        } else {
         $474 = $472;
        }
        $473 = ((($474)) + 48|0);
        HEAP32[$8>>2] = $473;
        _luaD_call($L,$474,1);
        $475 = HEAP32[$7>>2]|0;
        $476 = (($475) + ($424)|0);
        $477 = HEAP32[$8>>2]|0;
        $478 = ((($477)) + -16|0);
        HEAP32[$8>>2] = $478;
        $479 = $478;
        $480 = $479;
        $481 = HEAP32[$480>>2]|0;
        $482 = (($479) + 4)|0;
        $483 = $482;
        $484 = HEAP32[$483>>2]|0;
        $485 = $476;
        $486 = $485;
        HEAP32[$486>>2] = $481;
        $487 = (($485) + 4)|0;
        $488 = $487;
        HEAP32[$488>>2] = $484;
        $489 = ((($477)) + -8|0);
        $490 = HEAP32[$489>>2]|0;
        $491 = ((($476)) + 8|0);
        HEAP32[$491>>2] = $490;
       }
      } while(0);
      $492 = HEAP32[$2>>2]|0;
      $base$0 = $492;$pc$0 = $44;
      continue L5;
      break;
     }
     case 14:  {
      $493 = $45 >>> 23;
      $494 = $493 & 256;
      $495 = ($494|0)==(0);
      $496 = $493 & 255;
      $497 = (($42) + ($496<<4)|0);
      $498 = (($base$1) + ($493<<4)|0);
      $499 = $495 ? $498 : $497;
      $500 = $45 >>> 14;
      $501 = $500 & 256;
      $502 = ($501|0)==(0);
      $503 = $500 & 255;
      $504 = (($42) + ($503<<4)|0);
      $505 = $500 & 511;
      $506 = (($base$1) + ($505<<4)|0);
      $507 = $502 ? $506 : $504;
      $508 = ((($499)) + 8|0);
      $509 = HEAP32[$508>>2]|0;
      $510 = ($509|0)==(3);
      do {
       if ($510) {
        $511 = ((($507)) + 8|0);
        $512 = HEAP32[$511>>2]|0;
        $513 = ($512|0)==(3);
        if ($513) {
         $514 = +HEAPF64[$499>>3];
         $515 = +HEAPF64[$507>>3];
         $516 = $514 * $515;
         HEAPF64[$94>>3] = $516;
         $517 = (((($base$1) + ($93<<4)|0)) + 8|0);
         HEAP32[$517>>2] = 3;
         $base$0 = $base$1;$pc$0 = $44;
         continue L5;
        } else {
         HEAP32[$0>>2] = $44;
         $$0$i61$ph = $499;
         label = 86;
         break;
        }
       } else {
        HEAP32[$0>>2] = $44;
        $cond249 = ($509|0)==(4);
        if ($cond249) {
         $518 = HEAP32[$499>>2]|0;
         $519 = ((($518)) + 16|0);
         $520 = (_luaO_str2d($519,$num$i60)|0);
         $521 = ($520|0)==(0);
         if (!($521)) {
          $522 = +HEAPF64[$num$i60>>3];
          HEAPF64[$tempb$i8>>3] = $522;
          HEAP32[$15>>2] = 3;
          $$0$i61$ph = $tempb$i8;
          label = 86;
          break;
         }
        }
        label = 91;
       }
      } while(0);
      do {
       if ((label|0) == 86) {
        label = 0;
        $523 = ((($507)) + 8|0);
        $524 = HEAP32[$523>>2]|0;
        switch ($524|0) {
        case 3:  {
         $$0$i64$ph = $507;
         label = 90;
         break;
        }
        case 4:  {
         $525 = HEAP32[$507>>2]|0;
         $526 = ((($525)) + 16|0);
         $527 = (_luaO_str2d($526,$num$i63)|0);
         $528 = ($527|0)==(0);
         if ($528) {
          label = 89;
         } else {
          $529 = +HEAPF64[$num$i63>>3];
          HEAPF64[$tempc$i9>>3] = $529;
          HEAP32[$14>>2] = 3;
          $$0$i64$ph = $tempc$i9;
          label = 90;
         }
         break;
        }
        default: {
         label = 89;
        }
        }
        if ((label|0) == 89) {
         label = 0;
         label = 91;
         break;
        }
        else if ((label|0) == 90) {
         label = 0;
         $530 = +HEAPF64[$$0$i61$ph>>3];
         $531 = +HEAPF64[$$0$i64$ph>>3];
         $532 = $530 * $531;
         HEAPF64[$94>>3] = $532;
         $533 = (((($base$1) + ($93<<4)|0)) + 8|0);
         HEAP32[$533>>2] = 3;
         break;
        }
       }
      } while(0);
      do {
       if ((label|0) == 91) {
        label = 0;
        $534 = (_luaT_gettmbyobj($L,$499,7)|0);
        $535 = ((($534)) + 8|0);
        $536 = HEAP32[$535>>2]|0;
        $537 = ($536|0)==(0);
        if ($537) {
         $538 = (_luaT_gettmbyobj($L,$507,7)|0);
         $$phi$trans$insert$i66 = ((($538)) + 8|0);
         $$pre$i67 = HEAP32[$$phi$trans$insert$i66>>2]|0;
         $539 = ($$pre$i67|0)==(0);
         if ($539) {
          _luaG_aritherror($L,$499,$507);
          break;
         } else {
          $tm$01$i68 = $538;
         }
        } else {
         $tm$01$i68 = $534;
        }
        $540 = HEAP32[$7>>2]|0;
        $541 = $94;
        $542 = (($541) - ($540))|0;
        $543 = HEAP32[$8>>2]|0;
        $544 = $tm$01$i68;
        $545 = $544;
        $546 = HEAP32[$545>>2]|0;
        $547 = (($544) + 4)|0;
        $548 = $547;
        $549 = HEAP32[$548>>2]|0;
        $550 = $543;
        $551 = $550;
        HEAP32[$551>>2] = $546;
        $552 = (($550) + 4)|0;
        $553 = $552;
        HEAP32[$553>>2] = $549;
        $554 = ((($tm$01$i68)) + 8|0);
        $555 = HEAP32[$554>>2]|0;
        $556 = ((($543)) + 8|0);
        HEAP32[$556>>2] = $555;
        $557 = HEAP32[$8>>2]|0;
        $558 = ((($557)) + 16|0);
        $559 = $499;
        $560 = $559;
        $561 = HEAP32[$560>>2]|0;
        $562 = (($559) + 4)|0;
        $563 = $562;
        $564 = HEAP32[$563>>2]|0;
        $565 = $558;
        $566 = $565;
        HEAP32[$566>>2] = $561;
        $567 = (($565) + 4)|0;
        $568 = $567;
        HEAP32[$568>>2] = $564;
        $569 = HEAP32[$508>>2]|0;
        $570 = ((($557)) + 24|0);
        HEAP32[$570>>2] = $569;
        $571 = HEAP32[$8>>2]|0;
        $572 = ((($571)) + 32|0);
        $573 = $507;
        $574 = $573;
        $575 = HEAP32[$574>>2]|0;
        $576 = (($573) + 4)|0;
        $577 = $576;
        $578 = HEAP32[$577>>2]|0;
        $579 = $572;
        $580 = $579;
        HEAP32[$580>>2] = $575;
        $581 = (($579) + 4)|0;
        $582 = $581;
        HEAP32[$582>>2] = $578;
        $583 = ((($507)) + 8|0);
        $584 = HEAP32[$583>>2]|0;
        $585 = ((($571)) + 40|0);
        HEAP32[$585>>2] = $584;
        $586 = HEAP32[$9>>2]|0;
        $587 = HEAP32[$8>>2]|0;
        $588 = (($586) - ($587))|0;
        $589 = ($588|0)<(49);
        $590 = $587;
        if ($589) {
         _luaD_growstack($L,3);
         $$pre$i72 = HEAP32[$8>>2]|0;
         $592 = $$pre$i72;
        } else {
         $592 = $590;
        }
        $591 = ((($592)) + 48|0);
        HEAP32[$8>>2] = $591;
        _luaD_call($L,$592,1);
        $593 = HEAP32[$7>>2]|0;
        $594 = (($593) + ($542)|0);
        $595 = HEAP32[$8>>2]|0;
        $596 = ((($595)) + -16|0);
        HEAP32[$8>>2] = $596;
        $597 = $596;
        $598 = $597;
        $599 = HEAP32[$598>>2]|0;
        $600 = (($597) + 4)|0;
        $601 = $600;
        $602 = HEAP32[$601>>2]|0;
        $603 = $594;
        $604 = $603;
        HEAP32[$604>>2] = $599;
        $605 = (($603) + 4)|0;
        $606 = $605;
        HEAP32[$606>>2] = $602;
        $607 = ((($595)) + -8|0);
        $608 = HEAP32[$607>>2]|0;
        $609 = ((($594)) + 8|0);
        HEAP32[$609>>2] = $608;
       }
      } while(0);
      $610 = HEAP32[$2>>2]|0;
      $base$0 = $610;$pc$0 = $44;
      continue L5;
      break;
     }
     case 15:  {
      $611 = $45 >>> 23;
      $612 = $611 & 256;
      $613 = ($612|0)==(0);
      $614 = $611 & 255;
      $615 = (($42) + ($614<<4)|0);
      $616 = (($base$1) + ($611<<4)|0);
      $617 = $613 ? $616 : $615;
      $618 = $45 >>> 14;
      $619 = $618 & 256;
      $620 = ($619|0)==(0);
      $621 = $618 & 255;
      $622 = (($42) + ($621<<4)|0);
      $623 = $618 & 511;
      $624 = (($base$1) + ($623<<4)|0);
      $625 = $620 ? $624 : $622;
      $626 = ((($617)) + 8|0);
      $627 = HEAP32[$626>>2]|0;
      $628 = ($627|0)==(3);
      do {
       if ($628) {
        $629 = ((($625)) + 8|0);
        $630 = HEAP32[$629>>2]|0;
        $631 = ($630|0)==(3);
        if ($631) {
         $632 = +HEAPF64[$617>>3];
         $633 = +HEAPF64[$625>>3];
         $634 = $632 / $633;
         HEAPF64[$94>>3] = $634;
         $635 = (((($base$1) + ($93<<4)|0)) + 8|0);
         HEAP32[$635>>2] = 3;
         $base$0 = $base$1;$pc$0 = $44;
         continue L5;
        } else {
         HEAP32[$0>>2] = $44;
         $$0$i75$ph = $617;
         label = 106;
         break;
        }
       } else {
        HEAP32[$0>>2] = $44;
        $cond250 = ($627|0)==(4);
        if ($cond250) {
         $636 = HEAP32[$617>>2]|0;
         $637 = ((($636)) + 16|0);
         $638 = (_luaO_str2d($637,$num$i74)|0);
         $639 = ($638|0)==(0);
         if (!($639)) {
          $640 = +HEAPF64[$num$i74>>3];
          HEAPF64[$tempb$i11>>3] = $640;
          HEAP32[$17>>2] = 3;
          $$0$i75$ph = $tempb$i11;
          label = 106;
          break;
         }
        }
        label = 111;
       }
      } while(0);
      do {
       if ((label|0) == 106) {
        label = 0;
        $641 = ((($625)) + 8|0);
        $642 = HEAP32[$641>>2]|0;
        switch ($642|0) {
        case 3:  {
         $$0$i78$ph = $625;
         label = 110;
         break;
        }
        case 4:  {
         $643 = HEAP32[$625>>2]|0;
         $644 = ((($643)) + 16|0);
         $645 = (_luaO_str2d($644,$num$i77)|0);
         $646 = ($645|0)==(0);
         if ($646) {
          label = 109;
         } else {
          $647 = +HEAPF64[$num$i77>>3];
          HEAPF64[$tempc$i12>>3] = $647;
          HEAP32[$16>>2] = 3;
          $$0$i78$ph = $tempc$i12;
          label = 110;
         }
         break;
        }
        default: {
         label = 109;
        }
        }
        if ((label|0) == 109) {
         label = 0;
         label = 111;
         break;
        }
        else if ((label|0) == 110) {
         label = 0;
         $648 = +HEAPF64[$$0$i75$ph>>3];
         $649 = +HEAPF64[$$0$i78$ph>>3];
         $650 = $648 / $649;
         HEAPF64[$94>>3] = $650;
         $651 = (((($base$1) + ($93<<4)|0)) + 8|0);
         HEAP32[$651>>2] = 3;
         break;
        }
       }
      } while(0);
      do {
       if ((label|0) == 111) {
        label = 0;
        $652 = (_luaT_gettmbyobj($L,$617,8)|0);
        $653 = ((($652)) + 8|0);
        $654 = HEAP32[$653>>2]|0;
        $655 = ($654|0)==(0);
        if ($655) {
         $656 = (_luaT_gettmbyobj($L,$625,8)|0);
         $$phi$trans$insert$i80 = ((($656)) + 8|0);
         $$pre$i81 = HEAP32[$$phi$trans$insert$i80>>2]|0;
         $657 = ($$pre$i81|0)==(0);
         if ($657) {
          _luaG_aritherror($L,$617,$625);
          break;
         } else {
          $tm$01$i82 = $656;
         }
        } else {
         $tm$01$i82 = $652;
        }
        $658 = HEAP32[$7>>2]|0;
        $659 = $94;
        $660 = (($659) - ($658))|0;
        $661 = HEAP32[$8>>2]|0;
        $662 = $tm$01$i82;
        $663 = $662;
        $664 = HEAP32[$663>>2]|0;
        $665 = (($662) + 4)|0;
        $666 = $665;
        $667 = HEAP32[$666>>2]|0;
        $668 = $661;
        $669 = $668;
        HEAP32[$669>>2] = $664;
        $670 = (($668) + 4)|0;
        $671 = $670;
        HEAP32[$671>>2] = $667;
        $672 = ((($tm$01$i82)) + 8|0);
        $673 = HEAP32[$672>>2]|0;
        $674 = ((($661)) + 8|0);
        HEAP32[$674>>2] = $673;
        $675 = HEAP32[$8>>2]|0;
        $676 = ((($675)) + 16|0);
        $677 = $617;
        $678 = $677;
        $679 = HEAP32[$678>>2]|0;
        $680 = (($677) + 4)|0;
        $681 = $680;
        $682 = HEAP32[$681>>2]|0;
        $683 = $676;
        $684 = $683;
        HEAP32[$684>>2] = $679;
        $685 = (($683) + 4)|0;
        $686 = $685;
        HEAP32[$686>>2] = $682;
        $687 = HEAP32[$626>>2]|0;
        $688 = ((($675)) + 24|0);
        HEAP32[$688>>2] = $687;
        $689 = HEAP32[$8>>2]|0;
        $690 = ((($689)) + 32|0);
        $691 = $625;
        $692 = $691;
        $693 = HEAP32[$692>>2]|0;
        $694 = (($691) + 4)|0;
        $695 = $694;
        $696 = HEAP32[$695>>2]|0;
        $697 = $690;
        $698 = $697;
        HEAP32[$698>>2] = $693;
        $699 = (($697) + 4)|0;
        $700 = $699;
        HEAP32[$700>>2] = $696;
        $701 = ((($625)) + 8|0);
        $702 = HEAP32[$701>>2]|0;
        $703 = ((($689)) + 40|0);
        HEAP32[$703>>2] = $702;
        $704 = HEAP32[$9>>2]|0;
        $705 = HEAP32[$8>>2]|0;
        $706 = (($704) - ($705))|0;
        $707 = ($706|0)<(49);
        $708 = $705;
        if ($707) {
         _luaD_growstack($L,3);
         $$pre$i86 = HEAP32[$8>>2]|0;
         $710 = $$pre$i86;
        } else {
         $710 = $708;
        }
        $709 = ((($710)) + 48|0);
        HEAP32[$8>>2] = $709;
        _luaD_call($L,$710,1);
        $711 = HEAP32[$7>>2]|0;
        $712 = (($711) + ($660)|0);
        $713 = HEAP32[$8>>2]|0;
        $714 = ((($713)) + -16|0);
        HEAP32[$8>>2] = $714;
        $715 = $714;
        $716 = $715;
        $717 = HEAP32[$716>>2]|0;
        $718 = (($715) + 4)|0;
        $719 = $718;
        $720 = HEAP32[$719>>2]|0;
        $721 = $712;
        $722 = $721;
        HEAP32[$722>>2] = $717;
        $723 = (($721) + 4)|0;
        $724 = $723;
        HEAP32[$724>>2] = $720;
        $725 = ((($713)) + -8|0);
        $726 = HEAP32[$725>>2]|0;
        $727 = ((($712)) + 8|0);
        HEAP32[$727>>2] = $726;
       }
      } while(0);
      $728 = HEAP32[$2>>2]|0;
      $base$0 = $728;$pc$0 = $44;
      continue L5;
      break;
     }
     case 16:  {
      $729 = $45 >>> 23;
      $730 = $729 & 256;
      $731 = ($730|0)==(0);
      $732 = $729 & 255;
      $733 = (($42) + ($732<<4)|0);
      $734 = (($base$1) + ($729<<4)|0);
      $735 = $731 ? $734 : $733;
      $736 = $45 >>> 14;
      $737 = $736 & 256;
      $738 = ($737|0)==(0);
      $739 = $736 & 255;
      $740 = (($42) + ($739<<4)|0);
      $741 = $736 & 511;
      $742 = (($base$1) + ($741<<4)|0);
      $743 = $738 ? $742 : $740;
      $744 = ((($735)) + 8|0);
      $745 = HEAP32[$744>>2]|0;
      $746 = ($745|0)==(3);
      do {
       if ($746) {
        $747 = ((($743)) + 8|0);
        $748 = HEAP32[$747>>2]|0;
        $749 = ($748|0)==(3);
        if ($749) {
         $750 = +HEAPF64[$735>>3];
         $751 = +HEAPF64[$743>>3];
         $752 = $750 / $751;
         $753 = (+Math_floor((+$752)));
         $754 = $751 * $753;
         $755 = $750 - $754;
         HEAPF64[$94>>3] = $755;
         $756 = (((($base$1) + ($93<<4)|0)) + 8|0);
         HEAP32[$756>>2] = 3;
         $base$0 = $base$1;$pc$0 = $44;
         continue L5;
        } else {
         HEAP32[$0>>2] = $44;
         $$0$i89$ph = $735;
         label = 126;
         break;
        }
       } else {
        HEAP32[$0>>2] = $44;
        $cond251 = ($745|0)==(4);
        if ($cond251) {
         $757 = HEAP32[$735>>2]|0;
         $758 = ((($757)) + 16|0);
         $759 = (_luaO_str2d($758,$num$i88)|0);
         $760 = ($759|0)==(0);
         if (!($760)) {
          $761 = +HEAPF64[$num$i88>>3];
          HEAPF64[$tempb$i14>>3] = $761;
          HEAP32[$19>>2] = 3;
          $$0$i89$ph = $tempb$i14;
          label = 126;
          break;
         }
        }
        label = 131;
       }
      } while(0);
      do {
       if ((label|0) == 126) {
        label = 0;
        $762 = ((($743)) + 8|0);
        $763 = HEAP32[$762>>2]|0;
        switch ($763|0) {
        case 3:  {
         $$0$i92$ph = $743;
         label = 130;
         break;
        }
        case 4:  {
         $764 = HEAP32[$743>>2]|0;
         $765 = ((($764)) + 16|0);
         $766 = (_luaO_str2d($765,$num$i91)|0);
         $767 = ($766|0)==(0);
         if ($767) {
          label = 129;
         } else {
          $768 = +HEAPF64[$num$i91>>3];
          HEAPF64[$tempc$i15>>3] = $768;
          HEAP32[$18>>2] = 3;
          $$0$i92$ph = $tempc$i15;
          label = 130;
         }
         break;
        }
        default: {
         label = 129;
        }
        }
        if ((label|0) == 129) {
         label = 0;
         label = 131;
         break;
        }
        else if ((label|0) == 130) {
         label = 0;
         $769 = +HEAPF64[$$0$i89$ph>>3];
         $770 = +HEAPF64[$$0$i92$ph>>3];
         $771 = $769 / $770;
         $772 = (+Math_floor((+$771)));
         $773 = $770 * $772;
         $774 = $769 - $773;
         HEAPF64[$94>>3] = $774;
         $775 = (((($base$1) + ($93<<4)|0)) + 8|0);
         HEAP32[$775>>2] = 3;
         break;
        }
       }
      } while(0);
      do {
       if ((label|0) == 131) {
        label = 0;
        $776 = (_luaT_gettmbyobj($L,$735,9)|0);
        $777 = ((($776)) + 8|0);
        $778 = HEAP32[$777>>2]|0;
        $779 = ($778|0)==(0);
        if ($779) {
         $780 = (_luaT_gettmbyobj($L,$743,9)|0);
         $$phi$trans$insert$i94 = ((($780)) + 8|0);
         $$pre$i95 = HEAP32[$$phi$trans$insert$i94>>2]|0;
         $781 = ($$pre$i95|0)==(0);
         if ($781) {
          _luaG_aritherror($L,$735,$743);
          break;
         } else {
          $tm$01$i96 = $780;
         }
        } else {
         $tm$01$i96 = $776;
        }
        $782 = HEAP32[$7>>2]|0;
        $783 = $94;
        $784 = (($783) - ($782))|0;
        $785 = HEAP32[$8>>2]|0;
        $786 = $tm$01$i96;
        $787 = $786;
        $788 = HEAP32[$787>>2]|0;
        $789 = (($786) + 4)|0;
        $790 = $789;
        $791 = HEAP32[$790>>2]|0;
        $792 = $785;
        $793 = $792;
        HEAP32[$793>>2] = $788;
        $794 = (($792) + 4)|0;
        $795 = $794;
        HEAP32[$795>>2] = $791;
        $796 = ((($tm$01$i96)) + 8|0);
        $797 = HEAP32[$796>>2]|0;
        $798 = ((($785)) + 8|0);
        HEAP32[$798>>2] = $797;
        $799 = HEAP32[$8>>2]|0;
        $800 = ((($799)) + 16|0);
        $801 = $735;
        $802 = $801;
        $803 = HEAP32[$802>>2]|0;
        $804 = (($801) + 4)|0;
        $805 = $804;
        $806 = HEAP32[$805>>2]|0;
        $807 = $800;
        $808 = $807;
        HEAP32[$808>>2] = $803;
        $809 = (($807) + 4)|0;
        $810 = $809;
        HEAP32[$810>>2] = $806;
        $811 = HEAP32[$744>>2]|0;
        $812 = ((($799)) + 24|0);
        HEAP32[$812>>2] = $811;
        $813 = HEAP32[$8>>2]|0;
        $814 = ((($813)) + 32|0);
        $815 = $743;
        $816 = $815;
        $817 = HEAP32[$816>>2]|0;
        $818 = (($815) + 4)|0;
        $819 = $818;
        $820 = HEAP32[$819>>2]|0;
        $821 = $814;
        $822 = $821;
        HEAP32[$822>>2] = $817;
        $823 = (($821) + 4)|0;
        $824 = $823;
        HEAP32[$824>>2] = $820;
        $825 = ((($743)) + 8|0);
        $826 = HEAP32[$825>>2]|0;
        $827 = ((($813)) + 40|0);
        HEAP32[$827>>2] = $826;
        $828 = HEAP32[$9>>2]|0;
        $829 = HEAP32[$8>>2]|0;
        $830 = (($828) - ($829))|0;
        $831 = ($830|0)<(49);
        $832 = $829;
        if ($831) {
         _luaD_growstack($L,3);
         $$pre$i100 = HEAP32[$8>>2]|0;
         $834 = $$pre$i100;
        } else {
         $834 = $832;
        }
        $833 = ((($834)) + 48|0);
        HEAP32[$8>>2] = $833;
        _luaD_call($L,$834,1);
        $835 = HEAP32[$7>>2]|0;
        $836 = (($835) + ($784)|0);
        $837 = HEAP32[$8>>2]|0;
        $838 = ((($837)) + -16|0);
        HEAP32[$8>>2] = $838;
        $839 = $838;
        $840 = $839;
        $841 = HEAP32[$840>>2]|0;
        $842 = (($839) + 4)|0;
        $843 = $842;
        $844 = HEAP32[$843>>2]|0;
        $845 = $836;
        $846 = $845;
        HEAP32[$846>>2] = $841;
        $847 = (($845) + 4)|0;
        $848 = $847;
        HEAP32[$848>>2] = $844;
        $849 = ((($837)) + -8|0);
        $850 = HEAP32[$849>>2]|0;
        $851 = ((($836)) + 8|0);
        HEAP32[$851>>2] = $850;
       }
      } while(0);
      $852 = HEAP32[$2>>2]|0;
      $base$0 = $852;$pc$0 = $44;
      continue L5;
      break;
     }
     case 17:  {
      $853 = $45 >>> 23;
      $854 = $853 & 256;
      $855 = ($854|0)==(0);
      $856 = $853 & 255;
      $857 = (($42) + ($856<<4)|0);
      $858 = (($base$1) + ($853<<4)|0);
      $859 = $855 ? $858 : $857;
      $860 = $45 >>> 14;
      $861 = $860 & 256;
      $862 = ($861|0)==(0);
      $863 = $860 & 255;
      $864 = (($42) + ($863<<4)|0);
      $865 = $860 & 511;
      $866 = (($base$1) + ($865<<4)|0);
      $867 = $862 ? $866 : $864;
      $868 = ((($859)) + 8|0);
      $869 = HEAP32[$868>>2]|0;
      $870 = ($869|0)==(3);
      do {
       if ($870) {
        $871 = ((($867)) + 8|0);
        $872 = HEAP32[$871>>2]|0;
        $873 = ($872|0)==(3);
        if ($873) {
         $874 = +HEAPF64[$859>>3];
         $875 = +HEAPF64[$867>>3];
         $876 = (+Math_pow((+$874),(+$875)));
         HEAPF64[$94>>3] = $876;
         $877 = (((($base$1) + ($93<<4)|0)) + 8|0);
         HEAP32[$877>>2] = 3;
         $base$0 = $base$1;$pc$0 = $44;
         continue L5;
        } else {
         HEAP32[$0>>2] = $44;
         $$0$i103$ph = $859;
         label = 146;
         break;
        }
       } else {
        HEAP32[$0>>2] = $44;
        $cond252 = ($869|0)==(4);
        if ($cond252) {
         $878 = HEAP32[$859>>2]|0;
         $879 = ((($878)) + 16|0);
         $880 = (_luaO_str2d($879,$num$i102)|0);
         $881 = ($880|0)==(0);
         if (!($881)) {
          $882 = +HEAPF64[$num$i102>>3];
          HEAPF64[$tempb$i17>>3] = $882;
          HEAP32[$21>>2] = 3;
          $$0$i103$ph = $tempb$i17;
          label = 146;
          break;
         }
        }
        label = 151;
       }
      } while(0);
      do {
       if ((label|0) == 146) {
        label = 0;
        $883 = ((($867)) + 8|0);
        $884 = HEAP32[$883>>2]|0;
        switch ($884|0) {
        case 3:  {
         $$0$i106$ph = $867;
         label = 150;
         break;
        }
        case 4:  {
         $885 = HEAP32[$867>>2]|0;
         $886 = ((($885)) + 16|0);
         $887 = (_luaO_str2d($886,$num$i105)|0);
         $888 = ($887|0)==(0);
         if ($888) {
          label = 149;
         } else {
          $889 = +HEAPF64[$num$i105>>3];
          HEAPF64[$tempc$i18>>3] = $889;
          HEAP32[$20>>2] = 3;
          $$0$i106$ph = $tempc$i18;
          label = 150;
         }
         break;
        }
        default: {
         label = 149;
        }
        }
        if ((label|0) == 149) {
         label = 0;
         label = 151;
         break;
        }
        else if ((label|0) == 150) {
         label = 0;
         $890 = +HEAPF64[$$0$i103$ph>>3];
         $891 = +HEAPF64[$$0$i106$ph>>3];
         $892 = (+Math_pow((+$890),(+$891)));
         HEAPF64[$94>>3] = $892;
         $893 = (((($base$1) + ($93<<4)|0)) + 8|0);
         HEAP32[$893>>2] = 3;
         break;
        }
       }
      } while(0);
      do {
       if ((label|0) == 151) {
        label = 0;
        $894 = (_luaT_gettmbyobj($L,$859,10)|0);
        $895 = ((($894)) + 8|0);
        $896 = HEAP32[$895>>2]|0;
        $897 = ($896|0)==(0);
        if ($897) {
         $898 = (_luaT_gettmbyobj($L,$867,10)|0);
         $$phi$trans$insert$i108 = ((($898)) + 8|0);
         $$pre$i109 = HEAP32[$$phi$trans$insert$i108>>2]|0;
         $899 = ($$pre$i109|0)==(0);
         if ($899) {
          _luaG_aritherror($L,$859,$867);
          break;
         } else {
          $tm$01$i110 = $898;
         }
        } else {
         $tm$01$i110 = $894;
        }
        $900 = HEAP32[$7>>2]|0;
        $901 = $94;
        $902 = (($901) - ($900))|0;
        $903 = HEAP32[$8>>2]|0;
        $904 = $tm$01$i110;
        $905 = $904;
        $906 = HEAP32[$905>>2]|0;
        $907 = (($904) + 4)|0;
        $908 = $907;
        $909 = HEAP32[$908>>2]|0;
        $910 = $903;
        $911 = $910;
        HEAP32[$911>>2] = $906;
        $912 = (($910) + 4)|0;
        $913 = $912;
        HEAP32[$913>>2] = $909;
        $914 = ((($tm$01$i110)) + 8|0);
        $915 = HEAP32[$914>>2]|0;
        $916 = ((($903)) + 8|0);
        HEAP32[$916>>2] = $915;
        $917 = HEAP32[$8>>2]|0;
        $918 = ((($917)) + 16|0);
        $919 = $859;
        $920 = $919;
        $921 = HEAP32[$920>>2]|0;
        $922 = (($919) + 4)|0;
        $923 = $922;
        $924 = HEAP32[$923>>2]|0;
        $925 = $918;
        $926 = $925;
        HEAP32[$926>>2] = $921;
        $927 = (($925) + 4)|0;
        $928 = $927;
        HEAP32[$928>>2] = $924;
        $929 = HEAP32[$868>>2]|0;
        $930 = ((($917)) + 24|0);
        HEAP32[$930>>2] = $929;
        $931 = HEAP32[$8>>2]|0;
        $932 = ((($931)) + 32|0);
        $933 = $867;
        $934 = $933;
        $935 = HEAP32[$934>>2]|0;
        $936 = (($933) + 4)|0;
        $937 = $936;
        $938 = HEAP32[$937>>2]|0;
        $939 = $932;
        $940 = $939;
        HEAP32[$940>>2] = $935;
        $941 = (($939) + 4)|0;
        $942 = $941;
        HEAP32[$942>>2] = $938;
        $943 = ((($867)) + 8|0);
        $944 = HEAP32[$943>>2]|0;
        $945 = ((($931)) + 40|0);
        HEAP32[$945>>2] = $944;
        $946 = HEAP32[$9>>2]|0;
        $947 = HEAP32[$8>>2]|0;
        $948 = (($946) - ($947))|0;
        $949 = ($948|0)<(49);
        $950 = $947;
        if ($949) {
         _luaD_growstack($L,3);
         $$pre$i114 = HEAP32[$8>>2]|0;
         $952 = $$pre$i114;
        } else {
         $952 = $950;
        }
        $951 = ((($952)) + 48|0);
        HEAP32[$8>>2] = $951;
        _luaD_call($L,$952,1);
        $953 = HEAP32[$7>>2]|0;
        $954 = (($953) + ($902)|0);
        $955 = HEAP32[$8>>2]|0;
        $956 = ((($955)) + -16|0);
        HEAP32[$8>>2] = $956;
        $957 = $956;
        $958 = $957;
        $959 = HEAP32[$958>>2]|0;
        $960 = (($957) + 4)|0;
        $961 = $960;
        $962 = HEAP32[$961>>2]|0;
        $963 = $954;
        $964 = $963;
        HEAP32[$964>>2] = $959;
        $965 = (($963) + 4)|0;
        $966 = $965;
        HEAP32[$966>>2] = $962;
        $967 = ((($955)) + -8|0);
        $968 = HEAP32[$967>>2]|0;
        $969 = ((($954)) + 8|0);
        HEAP32[$969>>2] = $968;
       }
      } while(0);
      $970 = HEAP32[$2>>2]|0;
      $base$0 = $970;$pc$0 = $44;
      continue L5;
      break;
     }
     case 18:  {
      $971 = $45 >>> 23;
      $972 = (($base$1) + ($971<<4)|0);
      $973 = (((($base$1) + ($971<<4)|0)) + 8|0);
      $974 = HEAP32[$973>>2]|0;
      $975 = ($974|0)==(3);
      if ($975) {
       $976 = +HEAPF64[$972>>3];
       $977 = -$976;
       HEAPF64[$94>>3] = $977;
       $978 = (((($base$1) + ($93<<4)|0)) + 8|0);
       HEAP32[$978>>2] = 3;
       $base$0 = $base$1;$pc$0 = $44;
       continue L5;
      }
      HEAP32[$0>>2] = $44;
      $cond253 = ($974|0)==(4);
      do {
       if ($cond253) {
        $979 = HEAP32[$972>>2]|0;
        $980 = ((($979)) + 16|0);
        $981 = (_luaO_str2d($980,$num$i116)|0);
        $982 = ($981|0)==(0);
        if ($982) {
         label = 163;
        } else {
         $983 = +HEAPF64[$num$i116>>3];
         HEAPF64[$tempb$i20>>3] = $983;
         HEAP32[$22>>2] = 3;
         $984 = HEAP32[$973>>2]|0;
         switch ($984|0) {
         case 3:  {
          label = 166;
          break;
         }
         case 4:  {
          $985 = HEAP32[$972>>2]|0;
          $986 = ((($985)) + 16|0);
          $987 = (_luaO_str2d($986,$num$i119)|0);
          $988 = ($987|0)==(0);
          if ($988) {
           label = 165;
          } else {
           label = 166;
          }
          break;
         }
         default: {
          label = 165;
         }
         }
         if ((label|0) == 165) {
          label = 0;
          label = 167;
          break;
         }
         else if ((label|0) == 166) {
          label = 0;
          $989 = +HEAPF64[$tempb$i20>>3];
          $990 = -$989;
          HEAPF64[$94>>3] = $990;
          $991 = (((($base$1) + ($93<<4)|0)) + 8|0);
          HEAP32[$991>>2] = 3;
          break;
         }
        }
       } else {
        label = 163;
       }
      } while(0);
      if ((label|0) == 163) {
       label = 0;
       label = 167;
      }
      do {
       if ((label|0) == 167) {
        label = 0;
        $992 = (_luaT_gettmbyobj($L,$972,11)|0);
        $993 = ((($992)) + 8|0);
        $994 = HEAP32[$993>>2]|0;
        $995 = ($994|0)==(0);
        if ($995) {
         $996 = (_luaT_gettmbyobj($L,$972,11)|0);
         $$phi$trans$insert$i122 = ((($996)) + 8|0);
         $$pre$i123 = HEAP32[$$phi$trans$insert$i122>>2]|0;
         $997 = ($$pre$i123|0)==(0);
         if ($997) {
          _luaG_aritherror($L,$972,$972);
          break;
         } else {
          $tm$01$i124 = $996;
         }
        } else {
         $tm$01$i124 = $992;
        }
        $998 = HEAP32[$7>>2]|0;
        $999 = $94;
        $1000 = (($999) - ($998))|0;
        $1001 = HEAP32[$8>>2]|0;
        $1002 = $tm$01$i124;
        $1003 = $1002;
        $1004 = HEAP32[$1003>>2]|0;
        $1005 = (($1002) + 4)|0;
        $1006 = $1005;
        $1007 = HEAP32[$1006>>2]|0;
        $1008 = $1001;
        $1009 = $1008;
        HEAP32[$1009>>2] = $1004;
        $1010 = (($1008) + 4)|0;
        $1011 = $1010;
        HEAP32[$1011>>2] = $1007;
        $1012 = ((($tm$01$i124)) + 8|0);
        $1013 = HEAP32[$1012>>2]|0;
        $1014 = ((($1001)) + 8|0);
        HEAP32[$1014>>2] = $1013;
        $1015 = HEAP32[$8>>2]|0;
        $1016 = ((($1015)) + 16|0);
        $1017 = $972;
        $1018 = $1017;
        $1019 = HEAP32[$1018>>2]|0;
        $1020 = (($1017) + 4)|0;
        $1021 = $1020;
        $1022 = HEAP32[$1021>>2]|0;
        $1023 = $1016;
        $1024 = $1023;
        HEAP32[$1024>>2] = $1019;
        $1025 = (($1023) + 4)|0;
        $1026 = $1025;
        HEAP32[$1026>>2] = $1022;
        $1027 = HEAP32[$973>>2]|0;
        $1028 = ((($1015)) + 24|0);
        HEAP32[$1028>>2] = $1027;
        $1029 = HEAP32[$8>>2]|0;
        $1030 = ((($1029)) + 32|0);
        $1031 = $972;
        $1032 = $1031;
        $1033 = HEAP32[$1032>>2]|0;
        $1034 = (($1031) + 4)|0;
        $1035 = $1034;
        $1036 = HEAP32[$1035>>2]|0;
        $1037 = $1030;
        $1038 = $1037;
        HEAP32[$1038>>2] = $1033;
        $1039 = (($1037) + 4)|0;
        $1040 = $1039;
        HEAP32[$1040>>2] = $1036;
        $1041 = HEAP32[$973>>2]|0;
        $1042 = ((($1029)) + 40|0);
        HEAP32[$1042>>2] = $1041;
        $1043 = HEAP32[$9>>2]|0;
        $1044 = HEAP32[$8>>2]|0;
        $1045 = (($1043) - ($1044))|0;
        $1046 = ($1045|0)<(49);
        $1047 = $1044;
        if ($1046) {
         _luaD_growstack($L,3);
         $$pre$i128 = HEAP32[$8>>2]|0;
         $1049 = $$pre$i128;
        } else {
         $1049 = $1047;
        }
        $1048 = ((($1049)) + 48|0);
        HEAP32[$8>>2] = $1048;
        _luaD_call($L,$1049,1);
        $1050 = HEAP32[$7>>2]|0;
        $1051 = (($1050) + ($1000)|0);
        $1052 = HEAP32[$8>>2]|0;
        $1053 = ((($1052)) + -16|0);
        HEAP32[$8>>2] = $1053;
        $1054 = $1053;
        $1055 = $1054;
        $1056 = HEAP32[$1055>>2]|0;
        $1057 = (($1054) + 4)|0;
        $1058 = $1057;
        $1059 = HEAP32[$1058>>2]|0;
        $1060 = $1051;
        $1061 = $1060;
        HEAP32[$1061>>2] = $1056;
        $1062 = (($1060) + 4)|0;
        $1063 = $1062;
        HEAP32[$1063>>2] = $1059;
        $1064 = ((($1052)) + -8|0);
        $1065 = HEAP32[$1064>>2]|0;
        $1066 = ((($1051)) + 8|0);
        HEAP32[$1066>>2] = $1065;
       }
      } while(0);
      $1067 = HEAP32[$2>>2]|0;
      $base$0 = $1067;$pc$0 = $44;
      continue L5;
      break;
     }
     case 19:  {
      $1068 = $45 >>> 23;
      $1069 = (((($base$1) + ($1068<<4)|0)) + 8|0);
      $1070 = HEAP32[$1069>>2]|0;
      $1071 = ($1070|0)==(0);
      if ($1071) {
       $1077 = 1;
      } else {
       $1072 = ($1070|0)==(1);
       if ($1072) {
        $1073 = (($base$1) + ($1068<<4)|0);
        $1074 = HEAP32[$1073>>2]|0;
        $1075 = ($1074|0)==(0);
        $1077 = $1075;
       } else {
        $1077 = 0;
       }
      }
      $1076 = $1077&1;
      HEAP32[$94>>2] = $1076;
      $1078 = (((($base$1) + ($93<<4)|0)) + 8|0);
      HEAP32[$1078>>2] = 1;
      $base$0 = $base$1;$pc$0 = $44;
      continue L5;
      break;
     }
     case 20:  {
      $1079 = $45 >>> 23;
      $1080 = (($base$1) + ($1079<<4)|0);
      $1081 = (((($base$1) + ($1079<<4)|0)) + 8|0);
      $1082 = HEAP32[$1081>>2]|0;
      switch ($1082|0) {
      case 5:  {
       $1083 = HEAP32[$1080>>2]|0;
       $1084 = (_luaH_getn($1083)|0);
       $1085 = (+($1084|0));
       HEAPF64[$94>>3] = $1085;
       $1086 = (((($base$1) + ($93<<4)|0)) + 8|0);
       HEAP32[$1086>>2] = 3;
       $base$0 = $base$1;$pc$0 = $44;
       continue L5;
       break;
      }
      case 4:  {
       $1087 = HEAP32[$1080>>2]|0;
       $1088 = ((($1087)) + 12|0);
       $1089 = HEAP32[$1088>>2]|0;
       $1090 = (+($1089>>>0));
       HEAPF64[$94>>3] = $1090;
       $1091 = (((($base$1) + ($93<<4)|0)) + 8|0);
       HEAP32[$1091>>2] = 3;
       $base$0 = $base$1;$pc$0 = $44;
       continue L5;
       break;
      }
      default: {
       HEAP32[$0>>2] = $44;
       $1092 = (_luaT_gettmbyobj($L,$1080,12)|0);
       $1093 = ((($1092)) + 8|0);
       $1094 = HEAP32[$1093>>2]|0;
       $1095 = ($1094|0)==(0);
       if ($1095) {
        $1096 = (_luaT_gettmbyobj($L,624096,12)|0);
        $$phi$trans$insert$i = ((($1096)) + 8|0);
        $$pre$i = HEAP32[$$phi$trans$insert$i>>2]|0;
        $1097 = ($$pre$i|0)==(0);
        if ($1097) {
         _luaG_typeerror($L,$1080,519282);
        } else {
         $tm$01$i = $1096;
         label = 183;
        }
       } else {
        $tm$01$i = $1092;
        label = 183;
       }
       if ((label|0) == 183) {
        label = 0;
        $1098 = HEAP32[$7>>2]|0;
        $1099 = $94;
        $1100 = (($1099) - ($1098))|0;
        $1101 = HEAP32[$8>>2]|0;
        $1102 = $tm$01$i;
        $1103 = $1102;
        $1104 = HEAP32[$1103>>2]|0;
        $1105 = (($1102) + 4)|0;
        $1106 = $1105;
        $1107 = HEAP32[$1106>>2]|0;
        $1108 = $1101;
        $1109 = $1108;
        HEAP32[$1109>>2] = $1104;
        $1110 = (($1108) + 4)|0;
        $1111 = $1110;
        HEAP32[$1111>>2] = $1107;
        $1112 = ((($tm$01$i)) + 8|0);
        $1113 = HEAP32[$1112>>2]|0;
        $1114 = ((($1101)) + 8|0);
        HEAP32[$1114>>2] = $1113;
        $1115 = HEAP32[$8>>2]|0;
        $1116 = ((($1115)) + 16|0);
        $1117 = $1080;
        $1118 = $1117;
        $1119 = HEAP32[$1118>>2]|0;
        $1120 = (($1117) + 4)|0;
        $1121 = $1120;
        $1122 = HEAP32[$1121>>2]|0;
        $1123 = $1116;
        $1124 = $1123;
        HEAP32[$1124>>2] = $1119;
        $1125 = (($1123) + 4)|0;
        $1126 = $1125;
        HEAP32[$1126>>2] = $1122;
        $1127 = HEAP32[$1081>>2]|0;
        $1128 = ((($1115)) + 24|0);
        HEAP32[$1128>>2] = $1127;
        $1129 = HEAP32[$8>>2]|0;
        $1130 = ((($1129)) + 32|0);
        $1131 = $1130;
        $1132 = $1131;
        HEAP32[$1132>>2] = $25;
        $1133 = (($1131) + 4)|0;
        $1134 = $1133;
        HEAP32[$1134>>2] = $28;
        $1135 = ((($1129)) + 40|0);
        HEAP32[$1135>>2] = $29;
        $1136 = HEAP32[$9>>2]|0;
        $1137 = HEAP32[$8>>2]|0;
        $1138 = (($1136) - ($1137))|0;
        $1139 = ($1138|0)<(49);
        $1140 = $1137;
        if ($1139) {
         _luaD_growstack($L,3);
         $$pre$i23 = HEAP32[$8>>2]|0;
         $1142 = $$pre$i23;
        } else {
         $1142 = $1140;
        }
        $1141 = ((($1142)) + 48|0);
        HEAP32[$8>>2] = $1141;
        _luaD_call($L,$1142,1);
        $1143 = HEAP32[$7>>2]|0;
        $1144 = (($1143) + ($1100)|0);
        $1145 = HEAP32[$8>>2]|0;
        $1146 = ((($1145)) + -16|0);
        HEAP32[$8>>2] = $1146;
        $1147 = $1146;
        $1148 = $1147;
        $1149 = HEAP32[$1148>>2]|0;
        $1150 = (($1147) + 4)|0;
        $1151 = $1150;
        $1152 = HEAP32[$1151>>2]|0;
        $1153 = $1144;
        $1154 = $1153;
        HEAP32[$1154>>2] = $1149;
        $1155 = (($1153) + 4)|0;
        $1156 = $1155;
        HEAP32[$1156>>2] = $1152;
        $1157 = ((($1145)) + -8|0);
        $1158 = HEAP32[$1157>>2]|0;
        $1159 = ((($1144)) + 8|0);
        HEAP32[$1159>>2] = $1158;
       }
       $1160 = HEAP32[$2>>2]|0;
       $base$0 = $1160;$pc$0 = $44;
       continue L5;
      }
      }
      break;
     }
     case 21:  {
      $1161 = $45 >>> 23;
      $1162 = $45 >>> 14;
      $1163 = $1162 & 511;
      HEAP32[$0>>2] = $44;
      $1164 = (1 - ($1161))|0;
      $1165 = (($1164) + ($1163))|0;
      _luaV_concat($L,$1165,$1163);
      $1166 = HEAP32[$6>>2]|0;
      $1167 = ((($1166)) + 68|0);
      $1168 = HEAP32[$1167>>2]|0;
      $1169 = ((($1166)) + 64|0);
      $1170 = HEAP32[$1169>>2]|0;
      $1171 = ($1168>>>0)<($1170>>>0);
      if (!($1171)) {
       _luaC_step($L);
      }
      $1172 = HEAP32[$2>>2]|0;
      $1173 = (($1172) + ($1161<<4)|0);
      $1174 = (($1172) + ($93<<4)|0);
      $1175 = $1173;
      $1176 = $1175;
      $1177 = HEAP32[$1176>>2]|0;
      $1178 = (($1175) + 4)|0;
      $1179 = $1178;
      $1180 = HEAP32[$1179>>2]|0;
      $1181 = $1174;
      $1182 = $1181;
      HEAP32[$1182>>2] = $1177;
      $1183 = (($1181) + 4)|0;
      $1184 = $1183;
      HEAP32[$1184>>2] = $1180;
      $1185 = (((($1172) + ($1161<<4)|0)) + 8|0);
      $1186 = HEAP32[$1185>>2]|0;
      $1187 = (((($1172) + ($93<<4)|0)) + 8|0);
      HEAP32[$1187>>2] = $1186;
      $base$0 = $1172;$pc$0 = $44;
      continue L5;
      break;
     }
     case 22:  {
      $1188 = $45 >>> 14;
      $1189 = (($1188) + -131071)|0;
      $1190 = (($44) + ($1189<<2)|0);
      $base$0 = $base$1;$pc$0 = $1190;
      continue L5;
      break;
     }
     case 23:  {
      $1191 = $45 >>> 23;
      $1192 = $1191 & 256;
      $1193 = ($1192|0)==(0);
      $1194 = $1191 & 255;
      $1195 = (($42) + ($1194<<4)|0);
      $1196 = (($base$1) + ($1191<<4)|0);
      $1197 = $1193 ? $1196 : $1195;
      $1198 = $45 >>> 14;
      $1199 = $1198 & 256;
      $1200 = ($1199|0)==(0);
      $1201 = $1198 & 255;
      $1202 = (($42) + ($1201<<4)|0);
      $1203 = $1198 & 511;
      $1204 = (($base$1) + ($1203<<4)|0);
      $1205 = $1200 ? $1204 : $1202;
      HEAP32[$0>>2] = $44;
      $1206 = ((($1197)) + 8|0);
      $1207 = HEAP32[$1206>>2]|0;
      $1208 = ((($1205)) + 8|0);
      $1209 = HEAP32[$1208>>2]|0;
      $1210 = ($1207|0)==($1209|0);
      if ($1210) {
       $1211 = (_luaV_equalval($L,$1197,$1205)|0);
       $1212 = ($1211|0)!=(0);
       $1214 = $1212;
      } else {
       $1214 = 0;
      }
      $1213 = $1214&1;
      $1215 = ($1213|0)==($93|0);
      if ($1215) {
       $1216 = HEAP32[$44>>2]|0;
       $1217 = $1216 >>> 14;
       $1218 = (($1217) + -131071)|0;
       $1219 = (($44) + ($1218<<2)|0);
       $pc$2 = $1219;
      } else {
       $pc$2 = $44;
      }
      $1220 = HEAP32[$2>>2]|0;
      $1221 = ((($pc$2)) + 4|0);
      $base$0 = $1220;$pc$0 = $1221;
      continue L5;
      break;
     }
     case 24:  {
      HEAP32[$0>>2] = $44;
      $1222 = $45 >>> 23;
      $1223 = $1222 & 256;
      $1224 = ($1223|0)==(0);
      $1225 = $1222 & 255;
      $1226 = (($42) + ($1225<<4)|0);
      $1227 = (($base$1) + ($1222<<4)|0);
      $1228 = $1224 ? $1227 : $1226;
      $1229 = $45 >>> 14;
      $1230 = $1229 & 256;
      $1231 = ($1230|0)==(0);
      $1232 = $1229 & 255;
      $1233 = (($42) + ($1232<<4)|0);
      $1234 = $1229 & 511;
      $1235 = (($base$1) + ($1234<<4)|0);
      $1236 = $1231 ? $1235 : $1233;
      $1237 = (_luaV_lessthan($L,$1228,$1236)|0);
      $1238 = ($1237|0)==($93|0);
      if ($1238) {
       $1239 = HEAP32[$44>>2]|0;
       $1240 = $1239 >>> 14;
       $1241 = (($1240) + -131071)|0;
       $1242 = (($44) + ($1241<<2)|0);
       $pc$3 = $1242;
      } else {
       $pc$3 = $44;
      }
      $1243 = HEAP32[$2>>2]|0;
      $1244 = ((($pc$3)) + 4|0);
      $base$0 = $1243;$pc$0 = $1244;
      continue L5;
      break;
     }
     case 25:  {
      HEAP32[$0>>2] = $44;
      $1245 = $45 >>> 23;
      $1246 = $1245 & 256;
      $1247 = ($1246|0)==(0);
      $1248 = $1245 & 255;
      $1249 = (($42) + ($1248<<4)|0);
      $1250 = (($base$1) + ($1245<<4)|0);
      $1251 = $1247 ? $1250 : $1249;
      $1252 = $45 >>> 14;
      $1253 = $1252 & 256;
      $1254 = ($1253|0)==(0);
      $1255 = $1252 & 255;
      $1256 = (($42) + ($1255<<4)|0);
      $1257 = $1252 & 511;
      $1258 = (($base$1) + ($1257<<4)|0);
      $1259 = $1254 ? $1258 : $1256;
      $1260 = ((($1251)) + 8|0);
      $1261 = HEAP32[$1260>>2]|0;
      $1262 = ((($1259)) + 8|0);
      $1263 = HEAP32[$1262>>2]|0;
      $1264 = ($1261|0)==($1263|0);
      L282: do {
       if ($1264) {
        switch ($1261|0) {
        case 3:  {
         $1266 = +HEAPF64[$1251>>3];
         $1267 = +HEAPF64[$1259>>3];
         $1268 = $1266 <= $1267;
         $1269 = $1268&1;
         $$0$i24 = $1269;
         break L282;
         break;
        }
        case 4:  {
         $1270 = HEAP32[$1251>>2]|0;
         $1271 = HEAP32[$1259>>2]|0;
         $1272 = ((($1270)) + 16|0);
         $1273 = ((($1271)) + 16|0);
         $1274 = (_strcoll($1272,$1273)|0);
         $1275 = ($1274|0)==(0);
         L310: do {
          if ($1275) {
           $1276 = ((($1271)) + 12|0);
           $1277 = HEAP32[$1276>>2]|0;
           $1278 = ((($1270)) + 12|0);
           $1279 = HEAP32[$1278>>2]|0;
           $l$09$i$i = $1272;$ll$010$i$i = $1279;$lr$012$i$i = $1277;$r$011$i$i = $1273;
           while(1) {
            $1280 = (_strlen($l$09$i$i)|0);
            $1281 = ($1280|0)==($lr$012$i$i|0);
            $1282 = ($1280|0)==($ll$010$i$i|0);
            if ($1281) {
             $$lcssa = $1282;
             break;
            }
            if ($1282) {
             $$25$i$i = -1;
             break L310;
            }
            $1285 = (($1280) + 1)|0;
            $1286 = (($l$09$i$i) + ($1285)|0);
            $1287 = (($ll$010$i$i) - ($1285))|0;
            $1288 = (($r$011$i$i) + ($1285)|0);
            $1289 = (($lr$012$i$i) - ($1285))|0;
            $1290 = (_strcoll($1286,$1288)|0);
            $1291 = ($1290|0)==(0);
            if ($1291) {
             $l$09$i$i = $1286;$ll$010$i$i = $1287;$lr$012$i$i = $1289;$r$011$i$i = $1288;
            } else {
             $$25$i$i = $1290;
             break L310;
            }
           }
           $1283 = $$lcssa&1;
           $1284 = $1283 ^ 1;
           $$25$i$i = $1284;
          } else {
           $$25$i$i = $1274;
          }
         } while(0);
         $1292 = ($$25$i$i|0)<(1);
         $1293 = $1292&1;
         $$0$i24 = $1293;
         break L282;
         break;
        }
        default: {
         $1294 = (_luaT_gettmbyobj($L,$1251,14)|0);
         $1295 = ((($1294)) + 8|0);
         $1296 = HEAP32[$1295>>2]|0;
         $1297 = ($1296|0)==(0);
         if (!($1297)) {
          $1298 = (_luaT_gettmbyobj($L,$1259,14)|0);
          $1299 = (_luaO_rawequalObj($1294,$1298)|0);
          $1300 = ($1299|0)==(0);
          if (!($1300)) {
           $1301 = HEAP32[$8>>2]|0;
           $1302 = HEAP32[$7>>2]|0;
           $1303 = $1301;
           $1304 = (($1303) - ($1302))|0;
           $1305 = $1294;
           $1306 = $1305;
           $1307 = HEAP32[$1306>>2]|0;
           $1308 = (($1305) + 4)|0;
           $1309 = $1308;
           $1310 = HEAP32[$1309>>2]|0;
           $1311 = $1301;
           $1312 = $1311;
           HEAP32[$1312>>2] = $1307;
           $1313 = (($1311) + 4)|0;
           $1314 = $1313;
           HEAP32[$1314>>2] = $1310;
           $1315 = HEAP32[$1295>>2]|0;
           $1316 = ((($1301)) + 8|0);
           HEAP32[$1316>>2] = $1315;
           $1317 = HEAP32[$8>>2]|0;
           $1318 = ((($1317)) + 16|0);
           $1319 = $1251;
           $1320 = $1319;
           $1321 = HEAP32[$1320>>2]|0;
           $1322 = (($1319) + 4)|0;
           $1323 = $1322;
           $1324 = HEAP32[$1323>>2]|0;
           $1325 = $1318;
           $1326 = $1325;
           HEAP32[$1326>>2] = $1321;
           $1327 = (($1325) + 4)|0;
           $1328 = $1327;
           HEAP32[$1328>>2] = $1324;
           $1329 = HEAP32[$1260>>2]|0;
           $1330 = ((($1317)) + 24|0);
           HEAP32[$1330>>2] = $1329;
           $1331 = HEAP32[$8>>2]|0;
           $1332 = ((($1331)) + 32|0);
           $1333 = $1259;
           $1334 = $1333;
           $1335 = HEAP32[$1334>>2]|0;
           $1336 = (($1333) + 4)|0;
           $1337 = $1336;
           $1338 = HEAP32[$1337>>2]|0;
           $1339 = $1332;
           $1340 = $1339;
           HEAP32[$1340>>2] = $1335;
           $1341 = (($1339) + 4)|0;
           $1342 = $1341;
           HEAP32[$1342>>2] = $1338;
           $1343 = HEAP32[$1262>>2]|0;
           $1344 = ((($1331)) + 40|0);
           HEAP32[$1344>>2] = $1343;
           $1345 = HEAP32[$9>>2]|0;
           $1346 = HEAP32[$8>>2]|0;
           $1347 = (($1345) - ($1346))|0;
           $1348 = ($1347|0)<(49);
           $1349 = $1346;
           if ($1348) {
            _luaD_growstack($L,3);
            $$pre$i$i = HEAP32[$8>>2]|0;
            $1351 = $$pre$i$i;
           } else {
            $1351 = $1349;
           }
           $1350 = ((($1351)) + 48|0);
           HEAP32[$8>>2] = $1350;
           _luaD_call($L,$1351,1);
           $1352 = HEAP32[$7>>2]|0;
           $1353 = (($1352) + ($1304)|0);
           $1354 = HEAP32[$8>>2]|0;
           $1355 = ((($1354)) + -16|0);
           HEAP32[$8>>2] = $1355;
           $1356 = $1355;
           $1357 = $1356;
           $1358 = HEAP32[$1357>>2]|0;
           $1359 = (($1356) + 4)|0;
           $1360 = $1359;
           $1361 = HEAP32[$1360>>2]|0;
           $1362 = $1353;
           $1363 = $1362;
           HEAP32[$1363>>2] = $1358;
           $1364 = (($1362) + 4)|0;
           $1365 = $1364;
           HEAP32[$1365>>2] = $1361;
           $1366 = ((($1354)) + -8|0);
           $1367 = HEAP32[$1366>>2]|0;
           $1368 = ((($1353)) + 8|0);
           HEAP32[$1368>>2] = $1367;
           $1369 = HEAP32[$8>>2]|0;
           $1370 = ((($1369)) + 8|0);
           $1371 = HEAP32[$1370>>2]|0;
           $1372 = ($1371|0)==(0);
           if ($1372) {
            $1376 = 0;
           } else {
            $1373 = ($1371|0)==(1);
            if ($1373) {
             $1374 = HEAP32[$1369>>2]|0;
             $phitmp$i$i = ($1374|0)!=(0);
             $1376 = $phitmp$i$i;
            } else {
             $1376 = 1;
            }
           }
           $1375 = $1376&1;
           $$0$i24 = $1375;
           break L282;
          }
         }
         $1377 = (_luaT_gettmbyobj($L,$1259,13)|0);
         $1378 = ((($1377)) + 8|0);
         $1379 = HEAP32[$1378>>2]|0;
         $1380 = ($1379|0)==(0);
         if (!($1380)) {
          $1381 = (_luaT_gettmbyobj($L,$1251,13)|0);
          $1382 = (_luaO_rawequalObj($1377,$1381)|0);
          $1383 = ($1382|0)==(0);
          if (!($1383)) {
           $1384 = HEAP32[$8>>2]|0;
           $1385 = HEAP32[$7>>2]|0;
           $1386 = $1384;
           $1387 = (($1386) - ($1385))|0;
           $1388 = $1377;
           $1389 = $1388;
           $1390 = HEAP32[$1389>>2]|0;
           $1391 = (($1388) + 4)|0;
           $1392 = $1391;
           $1393 = HEAP32[$1392>>2]|0;
           $1394 = $1384;
           $1395 = $1394;
           HEAP32[$1395>>2] = $1390;
           $1396 = (($1394) + 4)|0;
           $1397 = $1396;
           HEAP32[$1397>>2] = $1393;
           $1398 = HEAP32[$1378>>2]|0;
           $1399 = ((($1384)) + 8|0);
           HEAP32[$1399>>2] = $1398;
           $1400 = HEAP32[$8>>2]|0;
           $1401 = ((($1400)) + 16|0);
           $1402 = $1259;
           $1403 = $1402;
           $1404 = HEAP32[$1403>>2]|0;
           $1405 = (($1402) + 4)|0;
           $1406 = $1405;
           $1407 = HEAP32[$1406>>2]|0;
           $1408 = $1401;
           $1409 = $1408;
           HEAP32[$1409>>2] = $1404;
           $1410 = (($1408) + 4)|0;
           $1411 = $1410;
           HEAP32[$1411>>2] = $1407;
           $1412 = HEAP32[$1262>>2]|0;
           $1413 = ((($1400)) + 24|0);
           HEAP32[$1413>>2] = $1412;
           $1414 = HEAP32[$8>>2]|0;
           $1415 = ((($1414)) + 32|0);
           $1416 = $1251;
           $1417 = $1416;
           $1418 = HEAP32[$1417>>2]|0;
           $1419 = (($1416) + 4)|0;
           $1420 = $1419;
           $1421 = HEAP32[$1420>>2]|0;
           $1422 = $1415;
           $1423 = $1422;
           HEAP32[$1423>>2] = $1418;
           $1424 = (($1422) + 4)|0;
           $1425 = $1424;
           HEAP32[$1425>>2] = $1421;
           $1426 = HEAP32[$1260>>2]|0;
           $1427 = ((($1414)) + 40|0);
           HEAP32[$1427>>2] = $1426;
           $1428 = HEAP32[$9>>2]|0;
           $1429 = HEAP32[$8>>2]|0;
           $1430 = (($1428) - ($1429))|0;
           $1431 = ($1430|0)<(49);
           $1432 = $1429;
           if ($1431) {
            _luaD_growstack($L,3);
            $$pre$i4$i = HEAP32[$8>>2]|0;
            $1434 = $$pre$i4$i;
           } else {
            $1434 = $1432;
           }
           $1433 = ((($1434)) + 48|0);
           HEAP32[$8>>2] = $1433;
           _luaD_call($L,$1434,1);
           $1435 = HEAP32[$7>>2]|0;
           $1436 = (($1435) + ($1387)|0);
           $1437 = HEAP32[$8>>2]|0;
           $1438 = ((($1437)) + -16|0);
           HEAP32[$8>>2] = $1438;
           $1439 = $1438;
           $1440 = $1439;
           $1441 = HEAP32[$1440>>2]|0;
           $1442 = (($1439) + 4)|0;
           $1443 = $1442;
           $1444 = HEAP32[$1443>>2]|0;
           $1445 = $1436;
           $1446 = $1445;
           HEAP32[$1446>>2] = $1441;
           $1447 = (($1445) + 4)|0;
           $1448 = $1447;
           HEAP32[$1448>>2] = $1444;
           $1449 = ((($1437)) + -8|0);
           $1450 = HEAP32[$1449>>2]|0;
           $1451 = ((($1436)) + 8|0);
           HEAP32[$1451>>2] = $1450;
           $1452 = HEAP32[$8>>2]|0;
           $1453 = ((($1452)) + 8|0);
           $1454 = HEAP32[$1453>>2]|0;
           $1455 = ($1454|0)==(0);
           if ($1455) {
            $1459 = 0;
           } else {
            $1456 = ($1454|0)==(1);
            if ($1456) {
             $1457 = HEAP32[$1452>>2]|0;
             $phitmp$i1$i = ($1457|0)!=(0);
             $1459 = $phitmp$i1$i;
            } else {
             $1459 = 1;
            }
           }
           $1458 = $1459&1;
           $1460 = $1458 ^ 1;
           $$0$i24 = $1460;
           break L282;
          }
         }
         $1461 = (_luaG_ordererror($L,$1251,$1259)|0);
         $$0$i24 = $1461;
         break L282;
        }
        }
       } else {
        $1265 = (_luaG_ordererror($L,$1251,$1259)|0);
        $$0$i24 = $1265;
       }
      } while(0);
      $1462 = ($$0$i24|0)==($93|0);
      if ($1462) {
       $1463 = HEAP32[$44>>2]|0;
       $1464 = $1463 >>> 14;
       $1465 = (($1464) + -131071)|0;
       $1466 = (($44) + ($1465<<2)|0);
       $pc$4 = $1466;
      } else {
       $pc$4 = $44;
      }
      $1467 = HEAP32[$2>>2]|0;
      $1468 = ((($pc$4)) + 4|0);
      $base$0 = $1467;$pc$0 = $1468;
      continue L5;
      break;
     }
     case 26:  {
      $1469 = (((($base$1) + ($93<<4)|0)) + 8|0);
      $1470 = HEAP32[$1469>>2]|0;
      $1471 = ($1470|0)==(0);
      if ($1471) {
       $1476 = 1;
      } else {
       $1472 = ($1470|0)==(1);
       if ($1472) {
        $1473 = HEAP32[$94>>2]|0;
        $1474 = ($1473|0)==(0);
        $1476 = $1474;
       } else {
        $1476 = 0;
       }
      }
      $1475 = $1476&1;
      $1477 = $45 >>> 14;
      $1478 = $1477 & 511;
      $1479 = ($1475|0)==($1478|0);
      if ($1479) {
       $pc$5 = $44;
      } else {
       $1480 = HEAP32[$44>>2]|0;
       $1481 = $1480 >>> 14;
       $1482 = (($1481) + -131071)|0;
       $1483 = (($44) + ($1482<<2)|0);
       $pc$5 = $1483;
      }
      $1484 = ((($pc$5)) + 4|0);
      $base$0 = $base$1;$pc$0 = $1484;
      continue L5;
      break;
     }
     case 27:  {
      $1485 = $45 >>> 23;
      $1486 = (($base$1) + ($1485<<4)|0);
      $1487 = (((($base$1) + ($1485<<4)|0)) + 8|0);
      $1488 = HEAP32[$1487>>2]|0;
      $1489 = ($1488|0)==(0);
      if ($1489) {
       $1494 = 1;
      } else {
       $1490 = ($1488|0)==(1);
       if ($1490) {
        $1491 = HEAP32[$1486>>2]|0;
        $1492 = ($1491|0)==(0);
        $1494 = $1492;
       } else {
        $1494 = 0;
       }
      }
      $1493 = $1494&1;
      $1495 = $45 >>> 14;
      $1496 = $1495 & 511;
      $1497 = ($1493|0)==($1496|0);
      if ($1497) {
       $pc$6 = $44;
      } else {
       $1498 = $1486;
       $1499 = $1498;
       $1500 = HEAP32[$1499>>2]|0;
       $1501 = (($1498) + 4)|0;
       $1502 = $1501;
       $1503 = HEAP32[$1502>>2]|0;
       $1504 = $94;
       $1505 = $1504;
       HEAP32[$1505>>2] = $1500;
       $1506 = (($1504) + 4)|0;
       $1507 = $1506;
       HEAP32[$1507>>2] = $1503;
       $1508 = (((($base$1) + ($93<<4)|0)) + 8|0);
       HEAP32[$1508>>2] = $1488;
       $1509 = HEAP32[$44>>2]|0;
       $1510 = $1509 >>> 14;
       $1511 = (($1510) + -131071)|0;
       $1512 = (($44) + ($1511<<2)|0);
       $pc$6 = $1512;
      }
      $1513 = ((($pc$6)) + 4|0);
      $base$0 = $base$1;$pc$0 = $1513;
      continue L5;
      break;
     }
     case 28:  {
      $1514 = $45 >>> 23;
      $1515 = $45 >>> 14;
      $1516 = $1515 & 511;
      $1517 = (($1516) + -1)|0;
      $1518 = ($1514|0)==(0);
      if (!($1518)) {
       $1519 = (($94) + ($1514<<4)|0);
       HEAP32[$8>>2] = $1519;
      }
      HEAP32[$0>>2] = $44;
      $1520 = (_luaD_precall($L,$94,$1517)|0);
      switch ($1520|0) {
      case 0:  {
       label = 246;
       break L3;
       break;
      }
      case 1:  {
       break;
      }
      default: {
       label = 322;
       break L1;
      }
      }
      $1522 = ($1516|0)==(0);
      if (!($1522)) {
       $1523 = HEAP32[$1>>2]|0;
       $1524 = ((($1523)) + 8|0);
       $1525 = HEAP32[$1524>>2]|0;
       HEAP32[$8>>2] = $1525;
      }
      $1526 = HEAP32[$2>>2]|0;
      $base$0 = $1526;$pc$0 = $44;
      continue L5;
      break;
     }
     case 29:  {
      $1527 = $45 >>> 23;
      $1528 = ($1527|0)==(0);
      if (!($1528)) {
       $1529 = (($94) + ($1527<<4)|0);
       HEAP32[$8>>2] = $1529;
      }
      HEAP32[$0>>2] = $44;
      $1530 = (_luaD_precall($L,$94,-1)|0);
      switch ($1530|0) {
      case 0:  {
       break L5;
       break;
      }
      case 1:  {
       break;
      }
      default: {
       label = 322;
       break L1;
      }
      }
      $1576 = HEAP32[$2>>2]|0;
      $base$0 = $1576;$pc$0 = $44;
      continue L5;
      break;
     }
     case 31:  {
      $1590 = ((($94)) + 32|0);
      $1591 = +HEAPF64[$1590>>3];
      $1592 = +HEAPF64[$94>>3];
      $1593 = $1591 + $1592;
      $1594 = ((($94)) + 16|0);
      $1595 = +HEAPF64[$1594>>3];
      $1596 = $1591 > 0.0;
      if ($1596) {
       $1597 = !($1593 <= $1595);
       if ($1597) {
        $base$0 = $base$1;$pc$0 = $44;
        continue L5;
       }
      } else {
       $1598 = !($1595 <= $1593);
       if ($1598) {
        $base$0 = $base$1;$pc$0 = $44;
        continue L5;
       }
      }
      $1599 = $45 >>> 14;
      $1600 = (($1599) + -131071)|0;
      $1601 = (($44) + ($1600<<2)|0);
      HEAPF64[$94>>3] = $1593;
      $1602 = (((($base$1) + ($93<<4)|0)) + 8|0);
      HEAP32[$1602>>2] = 3;
      $1603 = ((($94)) + 48|0);
      HEAPF64[$1603>>3] = $1593;
      $1604 = ((($1603)) + 8|0);
      HEAP32[$1604>>2] = 3;
      $base$0 = $base$1;$pc$0 = $1601;
      continue L5;
      break;
     }
     case 32:  {
      $1605 = ((($94)) + 16|0);
      $1606 = ((($94)) + 32|0);
      HEAP32[$0>>2] = $44;
      $1607 = (((($base$1) + ($93<<4)|0)) + 8|0);
      $1608 = HEAP32[$1607>>2]|0;
      $1609 = ($1608|0)==(3);
      do {
       if ($1609) {
        label = 275;
       } else {
        $cond254 = ($1608|0)==(4);
        if ($cond254) {
         $1610 = HEAP32[$94>>2]|0;
         $1611 = ((($1610)) + 16|0);
         $1612 = (_luaO_str2d($1611,$num$i)|0);
         $1613 = ($1612|0)==(0);
         if (!($1613)) {
          $1614 = +HEAPF64[$num$i>>3];
          HEAPF64[$94>>3] = $1614;
          HEAP32[$1607>>2] = 3;
          label = 275;
          break;
         }
        }
        _luaG_runerror($L,519296,$vararg_buffer);
       }
      } while(0);
      L364: do {
       if ((label|0) == 275) {
        label = 0;
        $1615 = ((($1605)) + 8|0);
        $1616 = HEAP32[$1615>>2]|0;
        $1617 = ($1616|0)==(3);
        do {
         if (!($1617)) {
          $cond255 = ($1616|0)==(4);
          if ($cond255) {
           $1618 = HEAP32[$1605>>2]|0;
           $1619 = ((($1618)) + 16|0);
           $1620 = (_luaO_str2d($1619,$num$i26)|0);
           $1621 = ($1620|0)==(0);
           if (!($1621)) {
            $1622 = +HEAPF64[$num$i26>>3];
            HEAPF64[$1605>>3] = $1622;
            HEAP32[$1615>>2] = 3;
            break;
           }
          }
          _luaG_runerror($L,519333,$vararg_buffer1);
          break L364;
         }
        } while(0);
        $1623 = ((($1606)) + 8|0);
        $1624 = HEAP32[$1623>>2]|0;
        $1625 = ($1624|0)==(3);
        if (!($1625)) {
         $cond256 = ($1624|0)==(4);
         if (!($cond256)) {
          label = 284;
          break L1;
         }
         $1626 = HEAP32[$1606>>2]|0;
         $1627 = ((($1626)) + 16|0);
         $1628 = (_luaO_str2d($1627,$num$i29)|0);
         $1629 = ($1628|0)==(0);
         if ($1629) {
          label = 284;
          break L1;
         }
         $1630 = +HEAPF64[$num$i29>>3];
         HEAPF64[$1606>>3] = $1630;
         HEAP32[$1623>>2] = 3;
        }
       }
      } while(0);
      $1631 = +HEAPF64[$94>>3];
      $1632 = +HEAPF64[$1606>>3];
      $1633 = $1631 - $1632;
      HEAPF64[$94>>3] = $1633;
      HEAP32[$1607>>2] = 3;
      $1634 = $45 >>> 14;
      $1635 = (($1634) + -131071)|0;
      $1636 = (($44) + ($1635<<2)|0);
      $base$0 = $base$1;$pc$0 = $1636;
      continue L5;
      break;
     }
     case 33:  {
      $1637 = ((($94)) + 48|0);
      $1638 = ((($94)) + 32|0);
      $1639 = ((($1637)) + 32|0);
      $1640 = $1638;
      $1641 = $1640;
      $1642 = HEAP32[$1641>>2]|0;
      $1643 = (($1640) + 4)|0;
      $1644 = $1643;
      $1645 = HEAP32[$1644>>2]|0;
      $1646 = $1639;
      $1647 = $1646;
      HEAP32[$1647>>2] = $1642;
      $1648 = (($1646) + 4)|0;
      $1649 = $1648;
      HEAP32[$1649>>2] = $1645;
      $1650 = ((($1638)) + 8|0);
      $1651 = HEAP32[$1650>>2]|0;
      $1652 = ((($1639)) + 8|0);
      HEAP32[$1652>>2] = $1651;
      $1653 = ((($94)) + 16|0);
      $1654 = ((($1637)) + 16|0);
      $1655 = $1653;
      $1656 = $1655;
      $1657 = HEAP32[$1656>>2]|0;
      $1658 = (($1655) + 4)|0;
      $1659 = $1658;
      $1660 = HEAP32[$1659>>2]|0;
      $1661 = $1654;
      $1662 = $1661;
      HEAP32[$1662>>2] = $1657;
      $1663 = (($1661) + 4)|0;
      $1664 = $1663;
      HEAP32[$1664>>2] = $1660;
      $1665 = ((($1653)) + 8|0);
      $1666 = HEAP32[$1665>>2]|0;
      $1667 = ((($1654)) + 8|0);
      HEAP32[$1667>>2] = $1666;
      $1668 = $94;
      $1669 = $1668;
      $1670 = HEAP32[$1669>>2]|0;
      $1671 = (($1668) + 4)|0;
      $1672 = $1671;
      $1673 = HEAP32[$1672>>2]|0;
      $1674 = $1637;
      $1675 = $1674;
      HEAP32[$1675>>2] = $1670;
      $1676 = (($1674) + 4)|0;
      $1677 = $1676;
      HEAP32[$1677>>2] = $1673;
      $1678 = (((($base$1) + ($93<<4)|0)) + 8|0);
      $1679 = HEAP32[$1678>>2]|0;
      $1680 = ((($1637)) + 8|0);
      HEAP32[$1680>>2] = $1679;
      $1681 = ((($1637)) + 48|0);
      HEAP32[$8>>2] = $1681;
      HEAP32[$0>>2] = $44;
      $1682 = $45 >>> 14;
      $1683 = $1682 & 511;
      _luaD_call($L,$1637,$1683);
      $1684 = HEAP32[$2>>2]|0;
      $1685 = HEAP32[$1>>2]|0;
      $1686 = ((($1685)) + 8|0);
      $1687 = HEAP32[$1686>>2]|0;
      HEAP32[$8>>2] = $1687;
      $1688 = (($1684) + ($93<<4)|0);
      $1689 = ((($1688)) + 48|0);
      $1690 = ((($1689)) + 8|0);
      $1691 = HEAP32[$1690>>2]|0;
      $1692 = ($1691|0)==(0);
      if ($1692) {
       $pc$8 = $44;
      } else {
       $1693 = ((($1689)) + -16|0);
       $1694 = $1689;
       $1695 = $1694;
       $1696 = HEAP32[$1695>>2]|0;
       $1697 = (($1694) + 4)|0;
       $1698 = $1697;
       $1699 = HEAP32[$1698>>2]|0;
       $1700 = $1693;
       $1701 = $1700;
       HEAP32[$1701>>2] = $1696;
       $1702 = (($1700) + 4)|0;
       $1703 = $1702;
       HEAP32[$1703>>2] = $1699;
       $1704 = ((($1693)) + 8|0);
       HEAP32[$1704>>2] = $1691;
       $1705 = HEAP32[$44>>2]|0;
       $1706 = $1705 >>> 14;
       $1707 = (($1706) + -131071)|0;
       $1708 = (($44) + ($1707<<2)|0);
       $pc$8 = $1708;
      }
      $1709 = ((($pc$8)) + 4|0);
      $base$0 = $1684;$pc$0 = $1709;
      continue L5;
      break;
     }
     case 34:  {
      $1710 = $45 >>> 23;
      $1711 = $45 >>> 14;
      $1712 = $1711 & 511;
      $1713 = ($1710|0)==(0);
      if ($1713) {
       $1714 = HEAP32[$8>>2]|0;
       $1715 = $94;
       $1716 = (($1714) - ($1715))|0;
       $1717 = $1716 >> 4;
       $1718 = (($1717) + -1)|0;
       $1719 = HEAP32[$1>>2]|0;
       $1720 = ((($1719)) + 8|0);
       $1721 = HEAP32[$1720>>2]|0;
       HEAP32[$8>>2] = $1721;
       $n$0 = $1718;
      } else {
       $n$0 = $1710;
      }
      $1722 = ($1712|0)==(0);
      if ($1722) {
       $1723 = ((($pc$0)) + 8|0);
       $1724 = HEAP32[$44>>2]|0;
       $c75$0 = $1724;$pc$9 = $1723;
      } else {
       $c75$0 = $1712;$pc$9 = $44;
      }
      $1725 = (((($base$1) + ($93<<4)|0)) + 8|0);
      $1726 = HEAP32[$1725>>2]|0;
      $1727 = ($1726|0)==(5);
      if (!($1727)) {
       $base$0 = $base$1;$pc$0 = $pc$9;
       continue L5;
      }
      $1728 = HEAP32[$94>>2]|0;
      $1729 = ($c75$0*50)|0;
      $1730 = (($n$0) + -50)|0;
      $1731 = (($1730) + ($1729))|0;
      $1732 = ((($1728)) + 28|0);
      $1733 = HEAP32[$1732>>2]|0;
      $1734 = ($1731|0)>($1733|0);
      if ($1734) {
       _luaH_resizearray($L,$1728,$1731);
      }
      $1735 = ($n$0|0)>(0);
      if (!($1735)) {
       $base$0 = $base$1;$pc$0 = $pc$9;
       continue L5;
      }
      $1736 = ((($1728)) + 5|0);
      $last$0197 = $1731;$n$1196 = $n$0;
      while(1) {
       $1737 = (($94) + ($n$1196<<4)|0);
       $1738 = (($last$0197) + -1)|0;
       $1739 = (_luaH_setnum($L,$1728,$last$0197)|0);
       $1740 = $1737;
       $1741 = $1740;
       $1742 = HEAP32[$1741>>2]|0;
       $1743 = (($1740) + 4)|0;
       $1744 = $1743;
       $1745 = HEAP32[$1744>>2]|0;
       $1746 = $1739;
       $1747 = $1746;
       HEAP32[$1747>>2] = $1742;
       $1748 = (($1746) + 4)|0;
       $1749 = $1748;
       HEAP32[$1749>>2] = $1745;
       $1750 = ((($1737)) + 8|0);
       $1751 = HEAP32[$1750>>2]|0;
       $1752 = ((($1739)) + 8|0);
       HEAP32[$1752>>2] = $1751;
       $1753 = HEAP32[$1750>>2]|0;
       $1754 = ($1753|0)>(3);
       if ($1754) {
        $1755 = HEAP32[$1737>>2]|0;
        $1756 = ((($1755)) + 5|0);
        $1757 = HEAP8[$1756>>0]|0;
        $1758 = $1757 & 3;
        $1759 = ($1758<<24>>24)==(0);
        if (!($1759)) {
         $1760 = HEAP8[$1736>>0]|0;
         $1761 = $1760 & 4;
         $1762 = ($1761<<24>>24)==(0);
         if (!($1762)) {
          _luaC_barrierback($L,$1728);
         }
        }
       }
       $1763 = (($n$1196) + -1)|0;
       $1764 = ($n$1196|0)>(1);
       if ($1764) {
        $last$0197 = $1738;$n$1196 = $1763;
       } else {
        $base$0 = $base$1;$pc$0 = $pc$9;
        continue L5;
       }
      }
      break;
     }
     case 35:  {
      _luaF_close($L,$94);
      $base$0 = $base$1;$pc$0 = $44;
      continue L5;
      break;
     }
     case 36:  {
      $1765 = $45 >>> 14;
      $1766 = HEAP32[$39>>2]|0;
      $1767 = ((($1766)) + 16|0);
      $1768 = HEAP32[$1767>>2]|0;
      $1769 = (($1768) + ($1765<<2)|0);
      $1770 = HEAP32[$1769>>2]|0;
      $1771 = ((($1770)) + 72|0);
      $1772 = HEAP8[$1771>>0]|0;
      $1773 = $1772&255;
      $1774 = HEAP32[$43>>2]|0;
      $1775 = (_luaF_newLclosure($L,$1773,$1774)|0);
      $1776 = ((($1775)) + 16|0);
      HEAP32[$1776>>2] = $1770;
      $1777 = ($1772<<24>>24)==(0);
      if ($1777) {
       $pc$10$lcssa = $44;
      } else {
       $j$0193 = 0;$pc$10194 = $44;
       while(1) {
        $1778 = HEAP32[$pc$10194>>2]|0;
        $1779 = $1778 & 63;
        $1780 = ($1779|0)==(4);
        $1781 = $1778 >>> 23;
        if ($1780) {
         $1782 = (((($37)) + 20|0) + ($1781<<2)|0);
         $1783 = HEAP32[$1782>>2]|0;
         $1784 = (((($1775)) + 20|0) + ($j$0193<<2)|0);
         HEAP32[$1784>>2] = $1783;
        } else {
         $1785 = (($base$1) + ($1781<<4)|0);
         $1786 = (_luaF_findupval($L,$1785)|0);
         $1787 = (((($1775)) + 20|0) + ($j$0193<<2)|0);
         HEAP32[$1787>>2] = $1786;
        }
        $1788 = (($j$0193) + 1)|0;
        $1789 = ((($pc$10194)) + 4|0);
        $exitcond238 = ($1788|0)==($1773|0);
        if ($exitcond238) {
         break;
        } else {
         $j$0193 = $1788;$pc$10194 = $1789;
        }
       }
       $scevgep237 = (($44) + ($1773<<2)|0);
       $pc$10$lcssa = $scevgep237;
      }
      HEAP32[$94>>2] = $1775;
      $1790 = (((($base$1) + ($93<<4)|0)) + 8|0);
      HEAP32[$1790>>2] = 6;
      HEAP32[$0>>2] = $pc$10$lcssa;
      $1791 = HEAP32[$6>>2]|0;
      $1792 = ((($1791)) + 68|0);
      $1793 = HEAP32[$1792>>2]|0;
      $1794 = ((($1791)) + 64|0);
      $1795 = HEAP32[$1794>>2]|0;
      $1796 = ($1793>>>0)<($1795>>>0);
      if (!($1796)) {
       _luaC_step($L);
      }
      $1797 = HEAP32[$2>>2]|0;
      $base$0 = $1797;$pc$0 = $pc$10$lcssa;
      continue L5;
      break;
     }
     case 37:  {
      $1798 = $45 >>> 23;
      $1799 = (($1798) + -1)|0;
      $1800 = HEAP32[$1>>2]|0;
      $1801 = HEAP32[$1800>>2]|0;
      $1802 = ((($1800)) + 4|0);
      $1803 = HEAP32[$1802>>2]|0;
      $1804 = (($1801) - ($1803))|0;
      $1805 = $1804 >> 4;
      $1806 = HEAP32[$39>>2]|0;
      $1807 = ((($1806)) + 73|0);
      $1808 = HEAP8[$1807>>0]|0;
      $1809 = $1808&255;
      $1810 = (($1805) - ($1809))|0;
      $1811 = (($1810) + -1)|0;
      $1812 = ($1798|0)==(0);
      if ($1812) {
       HEAP32[$0>>2] = $44;
       $1813 = HEAP32[$9>>2]|0;
       $1814 = HEAP32[$8>>2]|0;
       $1815 = (($1813) - ($1814))|0;
       $1816 = $1811 << 4;
       $1817 = ($1815|0)>($1816|0);
       if (!($1817)) {
        _luaD_growstack($L,$1811);
       }
       $1818 = HEAP32[$2>>2]|0;
       $1819 = (($1818) + ($93<<4)|0);
       $1820 = (($1819) + ($1811<<4)|0);
       HEAP32[$8>>2] = $1820;
       $b79$0 = $1811;$base$12 = $1818;$ra$0 = $1819;
      } else {
       $b79$0 = $1799;$base$12 = $base$1;$ra$0 = $94;
      }
      $1821 = ($b79$0|0)>(0);
      if (!($1821)) {
       $base$0 = $base$12;$pc$0 = $44;
       continue L5;
      }
      $1822 = (1 - ($1810))|0;
      $j80$0192 = 0;
      while(1) {
       $1823 = ($j80$0192|0)<($1811|0);
       if ($1823) {
        $1824 = HEAP32[$1800>>2]|0;
        $1825 = (($1824) + ($1822<<4)|0);
        $1826 = (($1825) + ($j80$0192<<4)|0);
        $1827 = (($ra$0) + ($j80$0192<<4)|0);
        $1828 = $1826;
        $1829 = $1828;
        $1830 = HEAP32[$1829>>2]|0;
        $1831 = (($1828) + 4)|0;
        $1832 = $1831;
        $1833 = HEAP32[$1832>>2]|0;
        $1834 = $1827;
        $1835 = $1834;
        HEAP32[$1835>>2] = $1830;
        $1836 = (($1834) + 4)|0;
        $1837 = $1836;
        HEAP32[$1837>>2] = $1833;
        $1838 = ((($1826)) + 8|0);
        $1839 = HEAP32[$1838>>2]|0;
        $$sink = $1839;
       } else {
        $$sink = 0;
       }
       $1840 = (((($ra$0) + ($j80$0192<<4)|0)) + 8|0);
       HEAP32[$1840>>2] = $$sink;
       $1841 = (($j80$0192) + 1)|0;
       $exitcond = ($1841|0)==($b79$0|0);
       if ($exitcond) {
        $base$0 = $base$12;$pc$0 = $44;
        continue L5;
       } else {
        $j80$0192 = $1841;
       }
      }
      break;
     }
     default: {
      $base$0 = $base$1;$pc$0 = $44;
      continue L5;
     }
     }
    } while(0);
   }
   $1531 = HEAP32[$1>>2]|0;
   $1532 = ((($1531)) + -20|0);
   $1533 = HEAP32[$1532>>2]|0;
   $1534 = ((($1531)) + 4|0);
   $1535 = HEAP32[$1534>>2]|0;
   $1536 = HEAP32[$33>>2]|0;
   $1537 = ($1536|0)==(0|0);
   $$pre242 = ((($1531)) + -24|0);
   if ($1537) {
    $1544 = $1533;
   } else {
    $1538 = HEAP32[$$pre242>>2]|0;
    _luaF_close($L,$1538);
    $$pre240 = HEAP32[$1532>>2]|0;
    $1544 = $$pre240;
   }
   $1539 = HEAP32[$1531>>2]|0;
   $1540 = $1535;
   $1541 = (($1539) - ($1540))|0;
   $1542 = $1541 >> 4;
   $1543 = (($1544) + ($1542<<4)|0);
   HEAP32[$$pre242>>2] = $1543;
   HEAP32[$2>>2] = $1543;
   $1545 = HEAP32[$8>>2]|0;
   $1546 = ($1535>>>0)<($1545>>>0);
   if ($1546) {
    $1548 = $1535;$1555 = $1533;$aux$0199 = 0;
    while(1) {
     $1547 = $1548;
     $1549 = $1547;
     $1550 = HEAP32[$1549>>2]|0;
     $1551 = (($1547) + 4)|0;
     $1552 = $1551;
     $1553 = HEAP32[$1552>>2]|0;
     $1554 = $1555;
     $1556 = $1554;
     HEAP32[$1556>>2] = $1550;
     $1557 = (($1554) + 4)|0;
     $1558 = $1557;
     HEAP32[$1558>>2] = $1553;
     $1559 = (((($1535) + ($aux$0199<<4)|0)) + 8|0);
     $1560 = HEAP32[$1559>>2]|0;
     $1561 = (((($1533) + ($aux$0199<<4)|0)) + 8|0);
     HEAP32[$1561>>2] = $1560;
     $1562 = (($aux$0199) + 1)|0;
     $1563 = (($1535) + ($1562<<4)|0);
     $1564 = HEAP32[$8>>2]|0;
     $1565 = ($1563>>>0)<($1564>>>0);
     $1566 = (($1533) + ($1562<<4)|0);
     if ($1565) {
      $1548 = $1563;$1555 = $1566;$aux$0199 = $1562;
     } else {
      $$lcssa181 = $1566;
      break;
     }
    }
   } else {
    $$lcssa181 = $1533;
   }
   HEAP32[$8>>2] = $$lcssa181;
   $1567 = ((($1531)) + -16|0);
   HEAP32[$1567>>2] = $$lcssa181;
   $1568 = HEAP32[$0>>2]|0;
   $1569 = ((($1531)) + -12|0);
   HEAP32[$1569>>2] = $1568;
   $1570 = ((($1531)) + -4|0);
   $1571 = HEAP32[$1570>>2]|0;
   $1572 = (($1571) + 1)|0;
   HEAP32[$1570>>2] = $1572;
   $1573 = HEAP32[$1>>2]|0;
   $1574 = ((($1573)) + -24|0);
   HEAP32[$1>>2] = $1574;
   $1575 = $1568;
   $1842 = $1575;$35 = $1574;
  }
  if ((label|0) == 246) {
   label = 0;
   $1521 = (($$0$ph) + 1)|0;
   $$0$ph = $1521;
   continue;
  }
  $1577 = $$lcssa365 >>> 23;
  $1578 = ($1577|0)==(0);
  if (!($1578)) {
   $1579 = (($$lcssa374) + ($1577<<4)|0);
   $1580 = ((($1579)) + -16|0);
   HEAP32[$8>>2] = $1580;
  }
  $1581 = HEAP32[$33>>2]|0;
  $1582 = ($1581|0)==(0|0);
  if (!($1582)) {
   _luaF_close($L,$base$1$lcssa);
  }
  HEAP32[$0>>2] = $$lcssa359;
  $1583 = (_luaD_poscall($L,$$lcssa374)|0);
  $1584 = (($$0$ph) + -1)|0;
  $1585 = ($1584|0)==(0);
  if ($1585) {
   label = 322;
   break;
  }
  $1586 = ($1583|0)==(0);
  if ($1586) {
   $$0$ph = $1584;
   continue;
  }
  $1587 = HEAP32[$1>>2]|0;
  $1588 = ((($1587)) + 8|0);
  $1589 = HEAP32[$1588>>2]|0;
  HEAP32[$8>>2] = $1589;
  $$0$ph = $1584;
 }
 if ((label|0) == 17) {
  HEAP32[$0>>2] = $pc$0$lcssa;
  STACKTOP = sp;return;
 }
 else if ((label|0) == 284) {
  _luaG_runerror($L,519362,$vararg_buffer3);
  // unreachable;
 }
 else if ((label|0) == 322) {
  STACKTOP = sp;return;
 }
}
function _luaZ_fill($z) {
 $z = $z|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $size = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $size = sp;
 $0 = ((($z)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($z)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($z)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiii[$3 & 8191]($1,$5,$size)|0);
 $7 = ($6|0)==(0|0);
 $8 = HEAP32[$size>>2]|0;
 $9 = ($8|0)==(0);
 $or$cond = $7 | $9;
 if ($or$cond) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $10 = (($8) + -1)|0;
 HEAP32[$z>>2] = $10;
 $11 = ((($z)) + 4|0);
 $12 = ((($6)) + 1|0);
 HEAP32[$11>>2] = $12;
 $13 = HEAP8[$6>>0]|0;
 $14 = $13&255;
 $$0 = $14;
 STACKTOP = sp;return ($$0|0);
}
function _luaZ_lookahead($z) {
 $z = $z|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond$i = 0;
 var $size$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $size$i = sp;
 $0 = HEAP32[$z>>2]|0;
 $1 = ($0|0)==(0);
 do {
  if ($1) {
   $2 = ((($z)) + 16|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = ((($z)) + 8|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = ((($z)) + 12|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = (FUNCTION_TABLE_iiii[$5 & 8191]($3,$7,$size$i)|0);
   $9 = ($8|0)==(0|0);
   $10 = HEAP32[$size$i>>2]|0;
   $11 = ($10|0)==(0);
   $or$cond$i = $9 | $11;
   if (!($or$cond$i)) {
    $12 = ((($z)) + 4|0);
    HEAP32[$z>>2] = $10;
    HEAP32[$12>>2] = $8;
    $14 = $8;
    break;
   }
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  } else {
   $$phi$trans$insert = ((($z)) + 4|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0;
   $14 = $$pre;
  }
 } while(0);
 $13 = HEAP8[$14>>0]|0;
 $15 = $13&255;
 $$0 = $15;
 STACKTOP = sp;return ($$0|0);
}
function _luaZ_init($L,$z,$reader,$data) {
 $L = $L|0;
 $z = $z|0;
 $reader = $reader|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($z)) + 16|0);
 HEAP32[$0>>2] = $L;
 $1 = ((($z)) + 8|0);
 HEAP32[$1>>2] = $reader;
 $2 = ((($z)) + 12|0);
 HEAP32[$2>>2] = $data;
 HEAP32[$z>>2] = 0;
 $3 = ((($z)) + 4|0);
 HEAP32[$3>>2] = 0;
 return;
}
function _luaZ_read($z,$b,$n) {
 $z = $z|0;
 $b = $b|0;
 $n = $n|0;
 var $$$02 = 0, $$028 = 0, $$047 = 0, $$2 = 0, $$pre = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond$i$i = 0, $size$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $size$i$i = sp;
 $0 = ($n|0)==(0);
 if ($0) {
  $$2 = 0;
  STACKTOP = sp;return ($$2|0);
 }
 $1 = ((($z)) + 16|0);
 $2 = ((($z)) + 8|0);
 $3 = ((($z)) + 12|0);
 $4 = ((($z)) + 4|0);
 $$pre = HEAP32[$z>>2]|0;
 $$028 = $n;$$047 = $b;$6 = $$pre;
 while(1) {
  $5 = ($6|0)==(0);
  if ($5) {
   $7 = HEAP32[$1>>2]|0;
   $8 = HEAP32[$2>>2]|0;
   $9 = HEAP32[$3>>2]|0;
   $10 = (FUNCTION_TABLE_iiii[$8 & 8191]($7,$9,$size$i$i)|0);
   $11 = ($10|0)==(0|0);
   $12 = HEAP32[$size$i$i>>2]|0;
   $13 = ($12|0)==(0);
   $or$cond$i$i = $11 | $13;
   if ($or$cond$i$i) {
    $$2 = $$028;
    label = 8;
    break;
   }
   HEAP32[$z>>2] = $12;
   HEAP32[$4>>2] = $10;
   $15 = $12;$16 = $10;
  } else {
   $$pre10 = HEAP32[$4>>2]|0;
   $15 = $6;$16 = $$pre10;
  }
  $14 = ($$028>>>0)>($15>>>0);
  $$$02 = $14 ? $15 : $$028;
  _memcpy(($$047|0),($16|0),($$$02|0))|0;
  $17 = HEAP32[$z>>2]|0;
  $18 = (($17) - ($$$02))|0;
  HEAP32[$z>>2] = $18;
  $19 = HEAP32[$4>>2]|0;
  $20 = (($19) + ($$$02)|0);
  HEAP32[$4>>2] = $20;
  $21 = (($$047) + ($$$02)|0);
  $22 = (($$028) - ($$$02))|0;
  $23 = ($$028|0)==($$$02|0);
  if ($23) {
   $$2 = 0;
   label = 8;
   break;
  } else {
   $$028 = $22;$$047 = $21;$6 = $18;
  }
 }
 if ((label|0) == 8) {
  STACKTOP = sp;return ($$2|0);
 }
 return (0)|0;
}
function _luaZ_openspace($L,$buff,$n) {
 $L = $L|0;
 $buff = $buff|0;
 $n = $n|0;
 var $$n = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($buff)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1>>>0)<($n>>>0);
 if (!($2)) {
  $$pre = HEAP32[$buff>>2]|0;
  $10 = $$pre;
  return ($10|0);
 }
 $3 = ($n>>>0)<(32);
 $$n = $3 ? 32 : $n;
 $4 = (($$n) + 1)|0;
 $5 = ($4>>>0)<(4294967294);
 if ($5) {
  $6 = HEAP32[$buff>>2]|0;
  $7 = (_luaM_realloc_($L,$6,$1,$$n)|0);
  $$pre$phiZ2D = $buff;$9 = $7;
 } else {
  $8 = (_luaM_toobig($L)|0);
  $$pre$phiZ2D = $buff;$9 = $8;
 }
 HEAP32[$$pre$phiZ2D>>2] = $9;
 HEAP32[$0>>2] = $$n;
 $10 = $9;
 return ($10|0);
}
function _luaL_argerror($L,$narg,$extramsg) {
 $L = $L|0;
 $narg = $narg|0;
 $extramsg = $extramsg|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ar = 0;
 var $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_buffer6 = 0, $vararg_ptr1 = 0, $vararg_ptr10 = 0, $vararg_ptr5 = 0, $vararg_ptr9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer6 = sp + 16|0;
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $ar = sp + 28|0;
 $0 = (_lua_getstack($L,0,$ar)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  HEAP32[$vararg_buffer>>2] = $narg;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $extramsg;
  $2 = (_luaL_error($L,519390,$vararg_buffer)|0);
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 (_lua_getinfo($L,519423,$ar)|0);
 $3 = ((($ar)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (_strcmp($4,519425)|0);
 $6 = ($5|0)==(0);
 if ($6) {
  $7 = (($narg) + -1)|0;
  $8 = ($7|0)==(0);
  if ($8) {
   $9 = ((($ar)) + 4|0);
   $10 = HEAP32[$9>>2]|0;
   HEAP32[$vararg_buffer2>>2] = $10;
   $vararg_ptr5 = ((($vararg_buffer2)) + 4|0);
   HEAP32[$vararg_ptr5>>2] = $extramsg;
   $11 = (_luaL_error($L,519432,$vararg_buffer2)|0);
   $$0 = $11;
   STACKTOP = sp;return ($$0|0);
  } else {
   $$01 = $7;
  }
 } else {
  $$01 = $narg;
 }
 $12 = ((($ar)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13|0)==(0|0);
 if ($14) {
  HEAP32[$12>>2] = 523335;
  $15 = 523335;
 } else {
  $15 = $13;
 }
 HEAP32[$vararg_buffer6>>2] = $$01;
 $vararg_ptr9 = ((($vararg_buffer6)) + 4|0);
 HEAP32[$vararg_ptr9>>2] = $15;
 $vararg_ptr10 = ((($vararg_buffer6)) + 8|0);
 HEAP32[$vararg_ptr10>>2] = $extramsg;
 $16 = (_luaL_error($L,519462,$vararg_buffer6)|0);
 $$0 = $16;
 STACKTOP = sp;return ($$0|0);
}
function _luaL_error($L,$fmt,$varargs) {
 $L = $L|0;
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $ar$i = 0, $argp = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $ar$i = sp + 24|0;
 $argp = sp + 8|0;
 HEAP32[$argp>>2] = $varargs;
 $0 = (_lua_getstack($L,1,$ar$i)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  (_lua_getinfo($L,519412,$ar$i)|0);
  $2 = ((($ar$i)) + 20|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ($3|0)>(0);
  if ($4) {
   $5 = ((($ar$i)) + 36|0);
   HEAP32[$vararg_buffer>>2] = $5;
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
   HEAP32[$vararg_ptr1>>2] = $3;
   (_lua_pushfstring($L,519415,$vararg_buffer)|0);
   (_lua_pushvfstring($L,$fmt,$argp)|0);
   _lua_concat($L,2);
   $6 = (_lua_error($L)|0);
   STACKTOP = sp;return ($6|0);
  }
 }
 _lua_pushlstring($L,1282466,0);
 (_lua_pushvfstring($L,$fmt,$argp)|0);
 _lua_concat($L,2);
 $6 = (_lua_error($L)|0);
 STACKTOP = sp;return ($6|0);
}
function _luaL_typerror($L,$narg,$tname) {
 $L = $L|0;
 $narg = $narg|0;
 $tname = $tname|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_lua_type($L,$narg)|0);
 $1 = (_lua_typename($L,$0)|0);
 HEAP32[$vararg_buffer>>2] = $tname;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $1;
 $2 = (_lua_pushfstring($L,519492,$vararg_buffer)|0);
 $3 = (_luaL_argerror($L,$narg,$2)|0);
 STACKTOP = sp;return ($3|0);
}
function _luaL_where($L,$level) {
 $L = $L|0;
 $level = $level|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $ar = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $ar = sp + 8|0;
 $0 = (_lua_getstack($L,$level,$ar)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  (_lua_getinfo($L,519412,$ar)|0);
  $2 = ((($ar)) + 20|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ($3|0)>(0);
  if ($4) {
   $5 = ((($ar)) + 36|0);
   HEAP32[$vararg_buffer>>2] = $5;
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
   HEAP32[$vararg_ptr1>>2] = $3;
   (_lua_pushfstring($L,519415,$vararg_buffer)|0);
   STACKTOP = sp;return;
  }
 }
 _lua_pushlstring($L,1282466,0);
 STACKTOP = sp;return;
}
function _luaL_checkoption($L,$narg,$def,$lst) {
 $L = $L|0;
 $narg = $narg|0;
 $def = $def|0;
 $lst = $lst|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $i$02 = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($def|0)==(0|0);
 if ($0) {
  $2 = (_lua_tolstring($L,$narg,0)|0);
  $3 = ($2|0)==(0|0);
  if ($3) {
   $4 = (_lua_typename($L,4)|0);
   $5 = (_lua_type($L,$narg)|0);
   $6 = (_lua_typename($L,$5)|0);
   HEAP32[$vararg_buffer>>2] = $4;
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
   HEAP32[$vararg_ptr1>>2] = $6;
   $7 = (_lua_pushfstring($L,519492,$vararg_buffer)|0);
   (_luaL_argerror($L,$narg,$7)|0);
   $11 = 0;
  } else {
   $11 = $2;
  }
 } else {
  $1 = (_luaL_optlstring($L,$narg,$def,0)|0);
  $11 = $1;
 }
 $8 = HEAP32[$lst>>2]|0;
 $9 = ($8|0)==(0|0);
 L6: do {
  if (!($9)) {
   $10 = $8;$i$02 = 0;
   while(1) {
    $12 = (_strcmp($10,$11)|0);
    $13 = ($12|0)==(0);
    if ($13) {
     $$0 = $i$02;
     break;
    }
    $14 = (($i$02) + 1)|0;
    $15 = (($lst) + ($14<<2)|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ($16|0)==(0|0);
    if ($17) {
     break L6;
    } else {
     $10 = $16;$i$02 = $14;
    }
   }
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 HEAP32[$vararg_buffer2>>2] = $11;
 $18 = (_lua_pushfstring($L,519512,$vararg_buffer2)|0);
 $19 = (_luaL_argerror($L,$narg,$18)|0);
 $$0 = $19;
 STACKTOP = sp;return ($$0|0);
}
function _luaL_optlstring($L,$narg,$def,$len) {
 $L = $L|0;
 $narg = $narg|0;
 $def = $def|0;
 $len = $len|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_lua_type($L,$narg)|0);
 $1 = ($0|0)<(1);
 if (!($1)) {
  $6 = (_lua_tolstring($L,$narg,$len)|0);
  $7 = ($6|0)==(0|0);
  if (!($7)) {
   $$0 = $6;
   STACKTOP = sp;return ($$0|0);
  }
  $8 = (_lua_typename($L,4)|0);
  $9 = (_lua_type($L,$narg)|0);
  $10 = (_lua_typename($L,$9)|0);
  HEAP32[$vararg_buffer>>2] = $8;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $10;
  $11 = (_lua_pushfstring($L,519492,$vararg_buffer)|0);
  (_luaL_argerror($L,$narg,$11)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($len|0)==(0|0);
 if ($2) {
  $$0 = $def;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ($def|0)==(0|0);
 if ($3) {
  $5 = 0;
 } else {
  $4 = (_strlen($def)|0);
  $5 = $4;
 }
 HEAP32[$len>>2] = $5;
 $$0 = $def;
 STACKTOP = sp;return ($$0|0);
}
function _luaL_checklstring($L,$narg,$len) {
 $L = $L|0;
 $narg = $narg|0;
 $len = $len|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_lua_tolstring($L,$narg,$len)|0);
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  STACKTOP = sp;return ($0|0);
 }
 $2 = (_lua_typename($L,4)|0);
 $3 = (_lua_type($L,$narg)|0);
 $4 = (_lua_typename($L,$3)|0);
 HEAP32[$vararg_buffer>>2] = $2;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $4;
 $5 = (_lua_pushfstring($L,519492,$vararg_buffer)|0);
 (_luaL_argerror($L,$narg,$5)|0);
 STACKTOP = sp;return ($0|0);
}
function _luaL_newmetatable($L,$tname) {
 $L = $L|0;
 $tname = $tname|0;
 var $$0 = 0, $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _lua_getfield($L,-10000,$tname);
 $0 = (_lua_type($L,-1)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  $$0 = 0;
  return ($$0|0);
 }
 _lua_settop($L,-2);
 _lua_createtable($L,0,0);
 _lua_pushvalue($L,-1);
 _lua_setfield($L,-10000,$tname);
 $$0 = 1;
 return ($$0|0);
}
function _luaL_checkudata($L,$ud,$tname) {
 $L = $L|0;
 $ud = $ud|0;
 $tname = $tname|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_lua_touserdata($L,$ud)|0);
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  $2 = (_lua_getmetatable($L,$ud)|0);
  $3 = ($2|0)==(0);
  if (!($3)) {
   _lua_getfield($L,-10000,$tname);
   $4 = (_lua_rawequal($L,-1,-2)|0);
   $5 = ($4|0)==(0);
   if (!($5)) {
    _lua_settop($L,-3);
    $$0 = $0;
    STACKTOP = sp;return ($$0|0);
   }
  }
 }
 $6 = (_lua_type($L,$ud)|0);
 $7 = (_lua_typename($L,$6)|0);
 HEAP32[$vararg_buffer>>2] = $tname;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $7;
 $8 = (_lua_pushfstring($L,519492,$vararg_buffer)|0);
 (_luaL_argerror($L,$ud,$8)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _luaL_checkstack($L,$space,$mes) {
 $L = $L|0;
 $space = $space|0;
 $mes = $mes|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_lua_checkstack($L,$space)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  STACKTOP = sp;return;
 }
 HEAP32[$vararg_buffer>>2] = $mes;
 (_luaL_error($L,519532,$vararg_buffer)|0);
 STACKTOP = sp;return;
}
function _luaL_checktype($L,$narg,$t) {
 $L = $L|0;
 $narg = $narg|0;
 $t = $t|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_lua_type($L,$narg)|0);
 $1 = ($0|0)==($t|0);
 if ($1) {
  STACKTOP = sp;return;
 }
 $2 = (_lua_typename($L,$t)|0);
 $3 = (_lua_type($L,$narg)|0);
 $4 = (_lua_typename($L,$3)|0);
 HEAP32[$vararg_buffer>>2] = $2;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $4;
 $5 = (_lua_pushfstring($L,519492,$vararg_buffer)|0);
 (_luaL_argerror($L,$narg,$5)|0);
 STACKTOP = sp;return;
}
function _luaL_checkany($L,$narg) {
 $L = $L|0;
 $narg = $narg|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_type($L,$narg)|0);
 $1 = ($0|0)==(-1);
 if (!($1)) {
  return;
 }
 (_luaL_argerror($L,$narg,519552)|0);
 return;
}
function _luaL_checknumber($L,$narg) {
 $L = $L|0;
 $narg = $narg|0;
 var $0 = 0.0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (+_lua_tonumber($L,$narg));
 $1 = $0 == 0.0;
 if (!($1)) {
  STACKTOP = sp;return (+$0);
 }
 $2 = (_lua_isnumber($L,$narg)|0);
 $3 = ($2|0)==(0);
 if (!($3)) {
  STACKTOP = sp;return (+$0);
 }
 $4 = (_lua_typename($L,3)|0);
 $5 = (_lua_type($L,$narg)|0);
 $6 = (_lua_typename($L,$5)|0);
 HEAP32[$vararg_buffer>>2] = $4;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $6;
 $7 = (_lua_pushfstring($L,519492,$vararg_buffer)|0);
 (_luaL_argerror($L,$narg,$7)|0);
 STACKTOP = sp;return (+$0);
}
function _luaL_optnumber($L,$narg,$def) {
 $L = $L|0;
 $narg = $narg|0;
 $def = +$def;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_type($L,$narg)|0);
 $1 = ($0|0)<(1);
 if ($1) {
  $3 = $def;
  return (+$3);
 }
 $2 = (+_luaL_checknumber($L,$narg));
 $3 = $2;
 return (+$3);
}
function _luaL_checkinteger($L,$narg) {
 $L = $L|0;
 $narg = $narg|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_lua_tointeger($L,$narg)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  STACKTOP = sp;return ($0|0);
 }
 $2 = (_lua_isnumber($L,$narg)|0);
 $3 = ($2|0)==(0);
 if (!($3)) {
  STACKTOP = sp;return ($0|0);
 }
 $4 = (_lua_typename($L,3)|0);
 $5 = (_lua_type($L,$narg)|0);
 $6 = (_lua_typename($L,$5)|0);
 HEAP32[$vararg_buffer>>2] = $4;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $6;
 $7 = (_lua_pushfstring($L,519492,$vararg_buffer)|0);
 (_luaL_argerror($L,$narg,$7)|0);
 STACKTOP = sp;return ($0|0);
}
function _luaL_optinteger($L,$narg,$def) {
 $L = $L|0;
 $narg = $narg|0;
 $def = $def|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_type($L,$narg)|0);
 $1 = ($0|0)<(1);
 if ($1) {
  $3 = $def;
  return ($3|0);
 }
 $2 = (_luaL_checkinteger($L,$narg)|0);
 $3 = $2;
 return ($3|0);
}
function _luaL_getmetafield($L,$obj,$event) {
 $L = $L|0;
 $obj = $obj|0;
 $event = $event|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_getmetatable($L,$obj)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  $$0 = 0;
  return ($$0|0);
 }
 _lua_pushstring($L,$event);
 _lua_rawget($L,-2);
 $2 = (_lua_type($L,-1)|0);
 $3 = ($2|0)==(0);
 if ($3) {
  _lua_settop($L,-3);
  $$0 = 0;
  return ($$0|0);
 } else {
  _lua_remove($L,-2);
  $$0 = 1;
  return ($$0|0);
 }
 return (0)|0;
}
function _luaL_callmeta($L,$obj,$event) {
 $L = $L|0;
 $obj = $obj|0;
 $event = $event|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $obj$off = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $obj$off = (($obj) + 9999)|0;
 $0 = ($obj$off>>>0)>(9999);
 if ($0) {
  $4 = $obj;
 } else {
  $1 = (_lua_gettop($L)|0);
  $2 = (($obj) + 1)|0;
  $3 = (($2) + ($1))|0;
  $4 = $3;
 }
 $5 = (_lua_getmetatable($L,$4)|0);
 $6 = ($5|0)==(0);
 if ($6) {
  $$0 = 0;
  return ($$0|0);
 }
 _lua_pushstring($L,$event);
 _lua_rawget($L,-2);
 $7 = (_lua_type($L,-1)|0);
 $8 = ($7|0)==(0);
 if ($8) {
  _lua_settop($L,-3);
  $$0 = 0;
  return ($$0|0);
 } else {
  _lua_remove($L,-2);
  _lua_pushvalue($L,$4);
  _lua_call($L,1,1);
  $$0 = 1;
  return ($$0|0);
 }
 return (0)|0;
}
function _luaL_register($L,$libname,$l) {
 $L = $L|0;
 $libname = $libname|0;
 $l = $l|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_openlib($L,$libname,$l,0);
 return;
}
function _luaL_openlib($L,$libname,$l,$nup) {
 $L = $L|0;
 $libname = $libname|0;
 $l = $l|0;
 $nup = $nup|0;
 var $$01$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$02$us = 0, $size$0$lcssa$i = 0, $size$02$i = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($libname|0)==(0|0);
 if (!($0)) {
  $1 = HEAP32[$l>>2]|0;
  $2 = ($1|0)==(0|0);
  if ($2) {
   $size$0$lcssa$i = 0;
  } else {
   $$01$i = $l;$size$02$i = 0;
   while(1) {
    $3 = (($size$02$i) + 1)|0;
    $4 = ((($$01$i)) + 8|0);
    $5 = HEAP32[$4>>2]|0;
    $6 = ($5|0)==(0|0);
    if ($6) {
     $size$0$lcssa$i = $3;
     break;
    } else {
     $$01$i = $4;$size$02$i = $3;
    }
   }
  }
  (_luaL_findtable($L,-10000,523705,1)|0);
  _lua_getfield($L,-1,$libname);
  $7 = (_lua_type($L,-1)|0);
  $8 = ($7|0)==(5);
  if (!($8)) {
   _lua_settop($L,-2);
   $9 = (_luaL_findtable($L,-10002,$libname,$size$0$lcssa$i)|0);
   $10 = ($9|0)==(0|0);
   if (!($10)) {
    HEAP32[$vararg_buffer>>2] = $libname;
    (_luaL_error($L,523727,$vararg_buffer)|0);
   }
   _lua_pushvalue($L,-1);
   _lua_setfield($L,-3,$libname);
  }
  _lua_remove($L,-2);
  $11 = $nup ^ -1;
  _lua_insert($L,$11);
 }
 $12 = HEAP32[$l>>2]|0;
 $13 = ($12|0)==(0|0);
 if ($13) {
  $32 = $nup ^ -1;
  _lua_settop($L,$32);
  STACKTOP = sp;return;
 }
 $14 = ($nup|0)>(0);
 $15 = (-2 - ($nup))|0;
 $16 = (0 - ($nup))|0;
 if ($14) {
  $19 = $l;
 } else {
  $26 = $l;
  while(1) {
   $25 = ((($26)) + 4|0);
   $27 = HEAP32[$25>>2]|0;
   _lua_pushcclosure($L,$27,$nup);
   $28 = HEAP32[$26>>2]|0;
   _lua_setfield($L,$15,$28);
   $29 = ((($26)) + 8|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = ($30|0)==(0|0);
   if ($31) {
    break;
   } else {
    $26 = $29;
   }
  }
  $32 = $nup ^ -1;
  _lua_settop($L,$32);
  STACKTOP = sp;return;
 }
 while(1) {
  $i$02$us = 0;
  while(1) {
   _lua_pushvalue($L,$16);
   $17 = (($i$02$us) + 1)|0;
   $exitcond = ($17|0)==($nup|0);
   if ($exitcond) {
    break;
   } else {
    $i$02$us = $17;
   }
  }
  $18 = ((($19)) + 4|0);
  $20 = HEAP32[$18>>2]|0;
  _lua_pushcclosure($L,$20,$nup);
  $21 = HEAP32[$19>>2]|0;
  _lua_setfield($L,$15,$21);
  $22 = ((($19)) + 8|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = ($23|0)==(0|0);
  if ($24) {
   break;
  } else {
   $19 = $22;
  }
 }
 $32 = $nup ^ -1;
 _lua_settop($L,$32);
 STACKTOP = sp;return;
}
function _luaL_findtable($L,$idx,$fname,$szhint) {
 $L = $L|0;
 $idx = $idx|0;
 $fname = $fname|0;
 $szhint = $szhint|0;
 var $$0 = 0, $$01 = 0, $$01$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $e$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _lua_pushvalue($L,$idx);
 $$01 = $fname;
 while(1) {
  $0 = (_strchr($$01,46)|0);
  $1 = ($0|0)==(0|0);
  if ($1) {
   $2 = (_strlen($$01)|0);
   $3 = (($$01) + ($2)|0);
   $e$0 = $3;
  } else {
   $e$0 = $0;
  }
  $4 = $e$0;
  $5 = $$01;
  $6 = (($4) - ($5))|0;
  _lua_pushlstring($L,$$01,$6);
  _lua_rawget($L,-2);
  $7 = (_lua_type($L,-1)|0);
  $8 = ($7|0)==(0);
  if ($8) {
   _lua_settop($L,-2);
   $9 = HEAP8[$e$0>>0]|0;
   $10 = ($9<<24>>24)==(46);
   $11 = $10 ? 1 : $szhint;
   _lua_createtable($L,0,$11);
   _lua_pushlstring($L,$$01,$6);
   _lua_pushvalue($L,-2);
   _lua_settable($L,-4);
  } else {
   $12 = (_lua_type($L,-1)|0);
   $13 = ($12|0)==(5);
   if (!($13)) {
    $$01$lcssa = $$01;
    break;
   }
  }
  _lua_remove($L,-2);
  $14 = ((($e$0)) + 1|0);
  $15 = HEAP8[$e$0>>0]|0;
  $16 = ($15<<24>>24)==(46);
  if ($16) {
   $$01 = $14;
  } else {
   $$0 = 0;
   label = 9;
   break;
  }
 }
 if ((label|0) == 9) {
  return ($$0|0);
 }
 _lua_settop($L,-3);
 $$0 = $$01$lcssa;
 return ($$0|0);
}
function _luaL_gsub($L,$s,$p,$r) {
 $L = $L|0;
 $s = $s|0;
 $p = $p|0;
 $r = $r|0;
 var $$0$lcssa = 0, $$012$i = 0, $$012$i$i = 0, $$012$i$i4 = 0, $$07 = 0, $$in$i = 0, $$in$i$i = 0, $$in$i$i3 = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i5 = 0, $$pre$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1040|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $b = sp;
 $0 = (_strlen($p)|0);
 $1 = ((($b)) + 8|0);
 HEAP32[$1>>2] = $L;
 $2 = ((($b)) + 12|0);
 HEAP32[$b>>2] = $2;
 $3 = ((($b)) + 4|0);
 HEAP32[$3>>2] = 0;
 $4 = (_strstr($s,$p)|0);
 $5 = ($4|0)==(0|0);
 if ($5) {
  $$0$lcssa = $s;
 } else {
  $6 = ((($b)) + 1036|0);
  $$07 = $s;$8 = $4;
  while(1) {
   $7 = ($8|0)==($$07|0);
   if (!($7)) {
    $9 = $8;
    $10 = $$07;
    $11 = (($9) - ($10))|0;
    $$012$i = $$07;$$in$i = $11;
    while(1) {
     $12 = (($$in$i) + -1)|0;
     $13 = HEAP32[$b>>2]|0;
     $14 = ($13>>>0)<($6>>>0);
     if ($14) {
      $18 = $13;
     } else {
      (_luaL_prepbuffer($b)|0);
      $$pre$i1 = HEAP32[$b>>2]|0;
      $18 = $$pre$i1;
     }
     $15 = ((($$012$i)) + 1|0);
     $16 = HEAP8[$$012$i>>0]|0;
     $17 = ((($18)) + 1|0);
     HEAP32[$b>>2] = $17;
     HEAP8[$18>>0] = $16;
     $19 = ($12|0)==(0);
     if ($19) {
      break;
     } else {
      $$012$i = $15;$$in$i = $12;
     }
    }
   }
   $20 = (_strlen($r)|0);
   $21 = ($20|0)==(0);
   if (!($21)) {
    $$012$i$i = $r;$$in$i$i = $20;
    while(1) {
     $22 = (($$in$i$i) + -1)|0;
     $23 = HEAP32[$b>>2]|0;
     $24 = ($23>>>0)<($6>>>0);
     if ($24) {
      $28 = $23;
     } else {
      (_luaL_prepbuffer($b)|0);
      $$pre$i$i = HEAP32[$b>>2]|0;
      $28 = $$pre$i$i;
     }
     $25 = ((($$012$i$i)) + 1|0);
     $26 = HEAP8[$$012$i$i>>0]|0;
     $27 = ((($28)) + 1|0);
     HEAP32[$b>>2] = $27;
     HEAP8[$28>>0] = $26;
     $29 = ($22|0)==(0);
     if ($29) {
      break;
     } else {
      $$012$i$i = $25;$$in$i$i = $22;
     }
    }
   }
   $30 = (($8) + ($0)|0);
   $31 = (_strstr($30,$p)|0);
   $32 = ($31|0)==(0|0);
   if ($32) {
    $$0$lcssa = $30;
    break;
   } else {
    $$07 = $30;$8 = $31;
   }
  }
 }
 $33 = (_strlen($$0$lcssa)|0);
 $34 = ($33|0)==(0);
 if (!($34)) {
  $35 = ((($b)) + 1036|0);
  $$012$i$i4 = $$0$lcssa;$$in$i$i3 = $33;
  while(1) {
   $36 = (($$in$i$i3) + -1)|0;
   $37 = HEAP32[$b>>2]|0;
   $38 = ($37>>>0)<($35>>>0);
   if ($38) {
    $42 = $37;
   } else {
    (_luaL_prepbuffer($b)|0);
    $$pre$i$i5 = HEAP32[$b>>2]|0;
    $42 = $$pre$i$i5;
   }
   $39 = ((($$012$i$i4)) + 1|0);
   $40 = HEAP8[$$012$i$i4>>0]|0;
   $41 = ((($42)) + 1|0);
   HEAP32[$b>>2] = $41;
   HEAP8[$42>>0] = $40;
   $43 = ($36|0)==(0);
   if ($43) {
    break;
   } else {
    $$012$i$i4 = $39;$$in$i$i3 = $36;
   }
  }
 }
 $44 = HEAP32[$b>>2]|0;
 $45 = $2;
 $46 = ($44|0)==($45|0);
 if ($46) {
  $$pre$i = HEAP32[$3>>2]|0;
  $52 = $$pre$i;
  $51 = HEAP32[$1>>2]|0;
  _lua_concat($51,$52);
  HEAP32[$3>>2] = 1;
  $53 = (_lua_tolstring($L,-1,0)|0);
  STACKTOP = sp;return ($53|0);
 } else {
  $47 = (($44) - ($45))|0;
  $48 = HEAP32[$1>>2]|0;
  _lua_pushlstring($48,$2,$47);
  HEAP32[$b>>2] = $2;
  $49 = HEAP32[$3>>2]|0;
  $50 = (($49) + 1)|0;
  HEAP32[$3>>2] = $50;
  $52 = $50;
  $51 = HEAP32[$1>>2]|0;
  _lua_concat($51,$52);
  HEAP32[$3>>2] = 1;
  $53 = (_lua_tolstring($L,-1,0)|0);
  STACKTOP = sp;return ($53|0);
 }
 return (0)|0;
}
function _luaL_prepbuffer($B) {
 $B = $B|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond$i = 0, $or$cond1$i = 0, $toget$0$i = 0, $toget$1$i = 0, $toget$1$i$lcssa = 0, $toplen$0$i = 0, $toplen$1$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$B>>2]|0;
 $1 = ((($B)) + 12|0);
 $2 = $1;
 $3 = ($0|0)==($2|0);
 if ($3) {
  return ($1|0);
 }
 $4 = (($0) - ($2))|0;
 $5 = ((($B)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 _lua_pushlstring($6,$1,$4);
 HEAP32[$B>>2] = $1;
 $7 = ((($B)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (($8) + 1)|0;
 HEAP32[$7>>2] = $9;
 $10 = ($8|0)>(0);
 if (!($10)) {
  return ($1|0);
 }
 $11 = HEAP32[$5>>2]|0;
 $12 = (_lua_objlen($11,-1)|0);
 $toget$0$i = 1;$toplen$0$i = $12;
 while(1) {
  $13 = $toget$0$i ^ -1;
  $14 = (_lua_objlen($11,$13)|0);
  $15 = HEAP32[$7>>2]|0;
  $16 = (($15) - ($toget$0$i))|0;
  $17 = ($16|0)>(8);
  $18 = ($toplen$0$i>>>0)>($14>>>0);
  $or$cond$i = $18 | $17;
  $19 = $or$cond$i ? $14 : 0;
  $toplen$1$i = (($19) + ($toplen$0$i))|0;
  $20 = $or$cond$i&1;
  $toget$1$i = (($20) + ($toget$0$i))|0;
  $21 = ($toget$1$i|0)<($15|0);
  $or$cond1$i = $or$cond$i & $21;
  if ($or$cond1$i) {
   $toget$0$i = $toget$1$i;$toplen$0$i = $toplen$1$i;
  } else {
   $toget$1$i$lcssa = $toget$1$i;
   break;
  }
 }
 _lua_concat($11,$toget$1$i$lcssa);
 $22 = HEAP32[$7>>2]|0;
 $23 = (1 - ($toget$1$i$lcssa))|0;
 $24 = (($23) + ($22))|0;
 HEAP32[$7>>2] = $24;
 return ($1|0);
}
function _luaL_buffinit($L,$B) {
 $L = $L|0;
 $B = $B|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($B)) + 8|0);
 HEAP32[$0>>2] = $L;
 $1 = ((($B)) + 12|0);
 HEAP32[$B>>2] = $1;
 $2 = ((($B)) + 4|0);
 HEAP32[$2>>2] = 0;
 return;
}
function _luaL_addlstring($B,$s,$l) {
 $B = $B|0;
 $s = $s|0;
 $l = $l|0;
 var $$012 = 0, $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($l|0)==(0);
 if ($0) {
  return;
 }
 $1 = ((($B)) + 1036|0);
 $$012 = $s;$$in = $l;
 while(1) {
  $2 = (($$in) + -1)|0;
  $3 = HEAP32[$B>>2]|0;
  $4 = ($3>>>0)<($1>>>0);
  if ($4) {
   $8 = $3;
  } else {
   (_luaL_prepbuffer($B)|0);
   $$pre = HEAP32[$B>>2]|0;
   $8 = $$pre;
  }
  $5 = ((($$012)) + 1|0);
  $6 = HEAP8[$$012>>0]|0;
  $7 = ((($8)) + 1|0);
  HEAP32[$B>>2] = $7;
  HEAP8[$8>>0] = $6;
  $9 = ($2|0)==(0);
  if ($9) {
   break;
  } else {
   $$012 = $5;$$in = $2;
  }
 }
 return;
}
function _luaL_addstring($B,$s) {
 $B = $B|0;
 $s = $s|0;
 var $$012$i = 0, $$in$i = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strlen($s)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  return;
 }
 $2 = ((($B)) + 1036|0);
 $$012$i = $s;$$in$i = $0;
 while(1) {
  $3 = (($$in$i) + -1)|0;
  $4 = HEAP32[$B>>2]|0;
  $5 = ($4>>>0)<($2>>>0);
  if ($5) {
   $9 = $4;
  } else {
   (_luaL_prepbuffer($B)|0);
   $$pre$i = HEAP32[$B>>2]|0;
   $9 = $$pre$i;
  }
  $6 = ((($$012$i)) + 1|0);
  $7 = HEAP8[$$012$i>>0]|0;
  $8 = ((($9)) + 1|0);
  HEAP32[$B>>2] = $8;
  HEAP8[$9>>0] = $7;
  $10 = ($3|0)==(0);
  if ($10) {
   break;
  } else {
   $$012$i = $6;$$in$i = $3;
  }
 }
 return;
}
function _luaL_pushresult($B) {
 $B = $B|0;
 var $$phi$trans$insert = 0, $$pre = 0, $$pre$phi2Z2D = 0, $$pre$phiZ2D = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$B>>2]|0;
 $1 = ((($B)) + 12|0);
 $2 = $1;
 $3 = ($0|0)==($2|0);
 if ($3) {
  $$phi$trans$insert = ((($B)) + 4|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $$pre1 = ((($B)) + 8|0);
  $$pre$phi2Z2D = $$phi$trans$insert;$$pre$phiZ2D = $$pre1;$11 = $$pre;
  $10 = HEAP32[$$pre$phiZ2D>>2]|0;
  _lua_concat($10,$11);
  HEAP32[$$pre$phi2Z2D>>2] = 1;
  return;
 } else {
  $4 = (($0) - ($2))|0;
  $5 = ((($B)) + 8|0);
  $6 = HEAP32[$5>>2]|0;
  _lua_pushlstring($6,$1,$4);
  HEAP32[$B>>2] = $1;
  $7 = ((($B)) + 4|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = (($8) + 1)|0;
  HEAP32[$7>>2] = $9;
  $$pre$phi2Z2D = $7;$$pre$phiZ2D = $5;$11 = $9;
  $10 = HEAP32[$$pre$phiZ2D>>2]|0;
  _lua_concat($10,$11);
  HEAP32[$$pre$phi2Z2D>>2] = 1;
  return;
 }
}
function _luaL_addvalue($B) {
 $B = $B|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond$i = 0;
 var $or$cond1$i = 0, $toget$0$i = 0, $toget$1$i = 0, $toget$1$i$lcssa = 0, $toplen$0$i = 0, $toplen$1$i = 0, $vl = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vl = sp;
 $0 = ((($B)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_lua_tolstring($1,-1,$vl)|0);
 $3 = HEAP32[$vl>>2]|0;
 $4 = HEAP32[$B>>2]|0;
 $5 = ((($B)) + 12|0);
 $6 = $4;
 $7 = $5;
 $8 = (($7) + 1024)|0;
 $9 = (($8) - ($6))|0;
 $10 = ($3>>>0)>($9>>>0);
 if (!($10)) {
  _memcpy(($4|0),($2|0),($3|0))|0;
  $11 = HEAP32[$vl>>2]|0;
  $12 = HEAP32[$B>>2]|0;
  $13 = (($12) + ($11)|0);
  HEAP32[$B>>2] = $13;
  _lua_settop($1,-2);
  STACKTOP = sp;return;
 }
 $14 = ($4|0)==($5|0);
 if ($14) {
  $$pre = ((($B)) + 4|0);
  $$pre$phiZ2D = $$pre;
 } else {
  $15 = (($6) - ($7))|0;
  $16 = HEAP32[$0>>2]|0;
  _lua_pushlstring($16,$5,$15);
  HEAP32[$B>>2] = $5;
  $17 = ((($B)) + 4|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = (($18) + 1)|0;
  HEAP32[$17>>2] = $19;
  _lua_insert($1,-2);
  $$pre$phiZ2D = $17;
 }
 $20 = HEAP32[$$pre$phiZ2D>>2]|0;
 $21 = (($20) + 1)|0;
 HEAP32[$$pre$phiZ2D>>2] = $21;
 $22 = ($20|0)>(0);
 if (!($22)) {
  STACKTOP = sp;return;
 }
 $23 = HEAP32[$0>>2]|0;
 $24 = (_lua_objlen($23,-1)|0);
 $toget$0$i = 1;$toplen$0$i = $24;
 while(1) {
  $25 = $toget$0$i ^ -1;
  $26 = (_lua_objlen($23,$25)|0);
  $27 = HEAP32[$$pre$phiZ2D>>2]|0;
  $28 = (($27) - ($toget$0$i))|0;
  $29 = ($28|0)>(8);
  $30 = ($toplen$0$i>>>0)>($26>>>0);
  $or$cond$i = $30 | $29;
  $31 = $or$cond$i ? $26 : 0;
  $toplen$1$i = (($31) + ($toplen$0$i))|0;
  $32 = $or$cond$i&1;
  $toget$1$i = (($32) + ($toget$0$i))|0;
  $33 = ($toget$1$i|0)<($27|0);
  $or$cond1$i = $or$cond$i & $33;
  if ($or$cond1$i) {
   $toget$0$i = $toget$1$i;$toplen$0$i = $toplen$1$i;
  } else {
   $toget$1$i$lcssa = $toget$1$i;
   break;
  }
 }
 _lua_concat($23,$toget$1$i$lcssa);
 $34 = HEAP32[$$pre$phiZ2D>>2]|0;
 $35 = (1 - ($toget$1$i$lcssa))|0;
 $36 = (($35) + ($34))|0;
 HEAP32[$$pre$phiZ2D>>2] = $36;
 STACKTOP = sp;return;
}
function _luaL_ref($L,$t) {
 $L = $L|0;
 $t = $t|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ref$0 = 0, $t$off = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $t$off = (($t) + 9999)|0;
 $0 = ($t$off>>>0)>(9999);
 if ($0) {
  $6 = $t;
 } else {
  $1 = (_lua_gettop($L)|0);
  $2 = (($t) + 1)|0;
  $3 = (($2) + ($1))|0;
  $6 = $3;
 }
 $4 = (_lua_type($L,-1)|0);
 $5 = ($4|0)==(0);
 if ($5) {
  _lua_settop($L,-2);
  $$0 = -1;
  return ($$0|0);
 }
 _lua_rawgeti($L,$6,0);
 $7 = (_lua_tointeger($L,-1)|0);
 _lua_settop($L,-2);
 $8 = ($7|0)==(0);
 if ($8) {
  $9 = (_lua_objlen($L,$6)|0);
  $10 = (($9) + 1)|0;
  $ref$0 = $10;
 } else {
  _lua_rawgeti($L,$6,$7);
  _lua_rawseti($L,$6,0);
  $ref$0 = $7;
 }
 _lua_rawseti($L,$6,$ref$0);
 $$0 = $ref$0;
 return ($$0|0);
}
function _luaL_unref($L,$t,$ref) {
 $L = $L|0;
 $t = $t|0;
 $ref = $ref|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $t$off = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($ref|0)>(-1);
 if (!($0)) {
  return;
 }
 $t$off = (($t) + 9999)|0;
 $1 = ($t$off>>>0)>(9999);
 if ($1) {
  $5 = $t;
 } else {
  $2 = (_lua_gettop($L)|0);
  $3 = (($t) + 1)|0;
  $4 = (($3) + ($2))|0;
  $5 = $4;
 }
 _lua_rawgeti($L,$5,0);
 _lua_rawseti($L,$5,$ref);
 _lua_pushinteger($L,$ref);
 _lua_rawseti($L,$5,0);
 return;
}
function _luaL_loadfile($L,$filename) {
 $L = $L|0;
 $filename = $filename|0;
 var $$0 = 0, $$lcssa = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $c$0 = 0, $c$1 = 0, $lf = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer11 = 0, $vararg_buffer6 = 0, $vararg_ptr10 = 0, $vararg_ptr14 = 0, $vararg_ptr15 = 0, $vararg_ptr4 = 0, $vararg_ptr5 = 0, $vararg_ptr9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1088|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer11 = sp + 40|0;
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $lf = sp + 56|0;
 $0 = (_lua_gettop($L)|0);
 $1 = (($0) + 1)|0;
 HEAP32[$lf>>2] = 0;
 $2 = ($filename|0)==(0|0);
 if ($2) {
  _lua_pushlstring($L,519567,6);
  $3 = HEAP32[37573]|0;
  $4 = ((($lf)) + 4|0);
  HEAP32[$4>>2] = $3;
  $5 = $3;
  $$pre$phiZ2D = $4;$14 = $5;
 } else {
  HEAP32[$vararg_buffer>>2] = $filename;
  (_lua_pushfstring($L,519574,$vararg_buffer)|0);
  $6 = (_fopen($filename,527143)|0);
  $7 = ((($lf)) + 4|0);
  HEAP32[$7>>2] = $6;
  $8 = ($6|0)==(0|0);
  if ($8) {
   $9 = (___errno_location()|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = (_strerror($10)|0);
   $12 = (_lua_tolstring($L,$1,0)|0);
   $13 = ((($12)) + 1|0);
   HEAP32[$vararg_buffer1>>2] = 521744;
   $vararg_ptr4 = ((($vararg_buffer1)) + 4|0);
   HEAP32[$vararg_ptr4>>2] = $13;
   $vararg_ptr5 = ((($vararg_buffer1)) + 8|0);
   HEAP32[$vararg_ptr5>>2] = $11;
   (_lua_pushfstring($L,519578,$vararg_buffer1)|0);
   _lua_remove($L,$1);
   $$0 = 6;
   STACKTOP = sp;return ($$0|0);
  } else {
   $$pre$phiZ2D = $7;$14 = $6;
  }
 }
 $15 = (_getc($14)|0);
 $16 = ($15|0)==(35);
 do {
  if ($16) {
   HEAP32[$lf>>2] = 1;
   L9: while(1) {
    $17 = (_getc($14)|0);
    switch ($17|0) {
    case 10:  {
     label = 9;
     break L9;
     break;
    }
    case -1:  {
     label = 8;
     break L9;
     break;
    }
    default: {
    }
    }
   }
   if ((label|0) == 8) {
    $18 = ($filename|0)!=(0|0);
    $30 = $14;$41 = $18;$c$1 = -1;
    break;
   }
   else if ((label|0) == 9) {
    $19 = (_getc($14)|0);
    $c$0 = $19;
    label = 10;
    break;
   }
  } else {
   $c$0 = $15;
   label = 10;
  }
 } while(0);
 do {
  if ((label|0) == 10) {
   $20 = ($c$0|0)==(27);
   $21 = ($filename|0)!=(0|0);
   $or$cond = $21 & $20;
   if ($or$cond) {
    $22 = (_freopen($filename,530591,$14)|0);
    HEAP32[$$pre$phiZ2D>>2] = $22;
    $23 = ($22|0)==(0|0);
    if ($23) {
     $24 = (___errno_location()|0);
     $25 = HEAP32[$24>>2]|0;
     $26 = (_strerror($25)|0);
     $27 = (_lua_tolstring($L,$1,0)|0);
     $28 = ((($27)) + 1|0);
     HEAP32[$vararg_buffer6>>2] = 519595;
     $vararg_ptr9 = ((($vararg_buffer6)) + 4|0);
     HEAP32[$vararg_ptr9>>2] = $28;
     $vararg_ptr10 = ((($vararg_buffer6)) + 8|0);
     HEAP32[$vararg_ptr10>>2] = $26;
     (_lua_pushfstring($L,519578,$vararg_buffer6)|0);
     _lua_remove($L,$1);
     $$0 = 6;
     STACKTOP = sp;return ($$0|0);
    } else {
     L20: while(1) {
      $29 = (_getc($22)|0);
      switch ($29|0) {
      case 27: case -1:  {
       $$lcssa = $29;
       break L20;
       break;
      }
      default: {
      }
      }
     }
     HEAP32[$lf>>2] = 0;
     $30 = $22;$41 = 1;$c$1 = $$lcssa;
     break;
    }
   } else {
    $30 = $14;$41 = $21;$c$1 = $c$0;
   }
  }
 } while(0);
 (_ungetc($c$1,$30)|0);
 $31 = (_lua_tolstring($L,-1,0)|0);
 $32 = (_lua_load($L,4296,$lf,$31)|0);
 $33 = HEAP32[$$pre$phiZ2D>>2]|0;
 $34 = (_ferror($33)|0);
 if ($41) {
  (_fclose($33)|0);
 }
 $35 = ($34|0)==(0);
 if ($35) {
  _lua_remove($L,$1);
  $$0 = $32;
  STACKTOP = sp;return ($$0|0);
 } else {
  _lua_settop($L,$1);
  $36 = (___errno_location()|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = (_strerror($37)|0);
  $39 = (_lua_tolstring($L,$1,0)|0);
  $40 = ((($39)) + 1|0);
  HEAP32[$vararg_buffer11>>2] = 521561;
  $vararg_ptr14 = ((($vararg_buffer11)) + 4|0);
  HEAP32[$vararg_ptr14>>2] = $40;
  $vararg_ptr15 = ((($vararg_buffer11)) + 8|0);
  HEAP32[$vararg_ptr15>>2] = $38;
  (_lua_pushfstring($L,519578,$vararg_buffer11)|0);
  _lua_remove($L,$1);
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _getF($L,$ud,$size) {
 $L = $L|0;
 $ud = $ud|0;
 $size = $size|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$ud>>2]|0;
 $1 = ($0|0)==(0);
 if (!($1)) {
  HEAP32[$ud>>2] = 0;
  HEAP32[$size>>2] = 1;
  $$0 = 521380;
  return ($$0|0);
 }
 $2 = ((($ud)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_feof($3)|0);
 $5 = ($4|0)==(0);
 if (!($5)) {
  $$0 = 0;
  return ($$0|0);
 }
 $6 = ((($ud)) + 8|0);
 $7 = HEAP32[$2>>2]|0;
 $8 = (_fread($6,1,1024,$7)|0);
 HEAP32[$size>>2] = $8;
 $9 = ($8|0)!=(0);
 $10 = $9 ? $6 : 0;
 $$0 = $10;
 return ($$0|0);
}
function _luaL_loadbuffer($L,$buff,$size,$name) {
 $L = $L|0;
 $buff = $buff|0;
 $size = $size|0;
 $name = $name|0;
 var $0 = 0, $1 = 0, $ls = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ls = sp;
 HEAP32[$ls>>2] = $buff;
 $0 = ((($ls)) + 4|0);
 HEAP32[$0>>2] = $size;
 $1 = (_lua_load($L,4297,$ls,$name)|0);
 STACKTOP = sp;return ($1|0);
}
function _getS($L,$ud,$size) {
 $L = $L|0;
 $ud = $ud|0;
 $size = $size|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($ud)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  $$0 = 0;
  return ($$0|0);
 }
 HEAP32[$size>>2] = $1;
 HEAP32[$0>>2] = 0;
 $3 = HEAP32[$ud>>2]|0;
 $$0 = $3;
 return ($$0|0);
}
function _luaL_newstate() {
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_newstate(4298,0)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  return ($0|0);
 }
 (_lua_atpanic($0,4299)|0);
 return ($0|0);
}
function _l_alloc($ud,$ptr,$osize,$nsize) {
 $ud = $ud|0;
 $ptr = $ptr|0;
 $osize = $osize|0;
 $nsize = $nsize|0;
 var $$0 = 0, $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($nsize|0)==(0);
 if ($0) {
  _free($ptr);
  $$0 = 0;
  return ($$0|0);
 } else {
  $1 = (_realloc($ptr,$nsize)|0);
  $$0 = $1;
  return ($$0|0);
 }
 return (0)|0;
}
function _panic($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP32[37544]|0;
 $1 = (_lua_tolstring($L,-1,0)|0);
 HEAP32[$vararg_buffer>>2] = $1;
 (_fprintf($0,519602,$vararg_buffer)|0);
 STACKTOP = sp;return 0;
}
function _luaL_openlibs($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _lua_pushcclosure($L,4300,0);
 _lua_pushstring($L,1282466);
 _lua_call($L,1,0);
 _lua_pushcclosure($L,4301,0);
 _lua_pushstring($L,523111);
 _lua_call($L,1,0);
 _lua_pushcclosure($L,4302,0);
 _lua_pushstring($L,522280);
 _lua_call($L,1,0);
 _lua_pushcclosure($L,4303,0);
 _lua_pushstring($L,521701);
 _lua_call($L,1,0);
 _lua_pushcclosure($L,4304,0);
 _lua_pushstring($L,522053);
 _lua_call($L,1,0);
 _lua_pushcclosure($L,4305,0);
 _lua_pushstring($L,522489);
 _lua_call($L,1,0);
 _lua_pushcclosure($L,4306,0);
 _lua_pushstring($L,521854);
 _lua_call($L,1,0);
 _lua_pushcclosure($L,4307,0);
 _lua_pushstring($L,520956);
 _lua_call($L,1,0);
 return;
}
function _luaK_nil($fs,$from,$n) {
 $fs = $fs|0;
 $from = $from|0;
 $n = $n|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fs)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fs)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)>($3|0);
 do {
  if ($4) {
   $5 = ($1|0)==(0);
   if ($5) {
    $6 = ((($fs)) + 50|0);
    $7 = HEAP8[$6>>0]|0;
    $8 = $7&255;
    $9 = ($8|0)>($from|0);
    if ($9) {
     break;
    }
    return;
   }
   $10 = (($1) + -1)|0;
   $11 = HEAP32[$fs>>2]|0;
   $12 = ((($11)) + 12|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = (($13) + ($10<<2)|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = $15 & 63;
   $17 = ($16|0)==(3);
   if ($17) {
    $18 = $15 >>> 6;
    $19 = $18 & 255;
    $20 = $15 >>> 23;
    $21 = ($19|0)>($from|0);
    $22 = (($20) + 1)|0;
    $23 = ($22|0)<($from|0);
    $or$cond = $21 | $23;
    if (!($or$cond)) {
     $24 = (($from) + -1)|0;
     $25 = (($24) + ($n))|0;
     $26 = ($25|0)>($20|0);
     if (!($26)) {
      return;
     }
     $27 = $15 & 8388607;
     $28 = $25 << 23;
     $29 = $27 | $28;
     HEAP32[$14>>2] = $29;
     return;
    }
   }
  }
 } while(0);
 $30 = (($n) + ($from))|0;
 $31 = $from << 6;
 $32 = $30 << 23;
 $33 = (($32) + -8388608)|0;
 $34 = $31 | $33;
 $35 = $34 | 3;
 $36 = ((($fs)) + 12|0);
 $37 = HEAP32[$36>>2]|0;
 $38 = ((($37)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 (_luaK_code($fs,$35,$39)|0);
 return;
}
function _luaK_code($fs,$i,$line) {
 $fs = $fs|0;
 $i = $i|0;
 $line = $line|0;
 var $$0$i$i$i = 0, $$0$i$i1$i = 0, $$02$i$i = 0, $$neg$i$i = 0, $$neg$i$i$i = 0, $$phi$trans$insert = 0, $$phi$trans$insert5 = 0, $$pre = 0, $$pre$i = 0, $$pre2 = 0, $$pre3 = 0, $$pre4 = 0, $$pre6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $8 = 0, $9 = 0, $fs$idx$val$i$i = 0, $fs$idx$val$i$i$pre = 0, $fs$idx$val$idx$i$i = 0, $fs$idx$val$idx$val$i$i = 0, $ispos$i$i = 0, $ispos$i$i$i = 0, $neg$i$i = 0, $neg$i$i$i = 0, $storemerge$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$fs>>2]|0;
 $1 = ((($fs)) + 32|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($fs)) + 24|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2|0)==(-1);
 if ($5) {
  $51 = $4;
 } else {
  $6 = ((($fs)) + 12|0);
  $$02$i$i = $2;$fs$idx$val$i$i = $0;
  while(1) {
   $fs$idx$val$idx$i$i = ((($fs$idx$val$i$i)) + 12|0);
   $fs$idx$val$idx$val$i$i = HEAP32[$fs$idx$val$idx$i$i>>2]|0;
   $7 = (($fs$idx$val$idx$val$i$i) + ($$02$i$i<<2)|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = $8 >>> 14;
   $10 = (($9) + -131071)|0;
   $11 = ($10|0)==(-1);
   $12 = (($$02$i$i) + 1)|0;
   $13 = (($12) + ($10))|0;
   $$0$i$i$i = $11 ? -1 : $13;
   $14 = ($$02$i$i|0)>(0);
   if ($14) {
    $15 = ((($7)) + -4|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = $16 & 63;
    $18 = (518247 + ($17)|0);
    $19 = HEAP8[$18>>0]|0;
    $20 = ($19<<24>>24)<(0);
    if ($20) {
     $$0$i$i1$i = $15;$22 = $16;
    } else {
     label = 5;
    }
   } else {
    label = 5;
   }
   if ((label|0) == 5) {
    label = 0;
    $$0$i$i1$i = $7;$22 = $8;
   }
   $21 = $22 & 63;
   $23 = ($21|0)==(27);
   if ($23) {
    $24 = $22 >>> 23;
    $25 = $24 << 6;
    $26 = $22 & 8372224;
    $27 = $26 | $25;
    $28 = $27 | 26;
    HEAP32[$$0$i$i1$i>>2] = $28;
    $$neg$i$i = $$02$i$i ^ -1;
    $29 = (($4) + ($$neg$i$i))|0;
    $ispos$i$i = ($29|0)>(-1);
    $neg$i$i = (0 - ($29))|0;
    $30 = $ispos$i$i ? $29 : $neg$i$i;
    $31 = ($30|0)>(131071);
    if ($31) {
     $32 = HEAP32[$6>>2]|0;
     _luaX_syntaxerror($32,519652);
    }
    $33 = HEAP32[$7>>2]|0;
    $34 = $33 & 16383;
    $35 = $29 << 14;
    $36 = (($35) + 2147467264)|0;
    $37 = $34 | $36;
    $storemerge$i = $37;
   } else {
    $$neg$i$i$i = $$02$i$i ^ -1;
    $38 = (($4) + ($$neg$i$i$i))|0;
    $ispos$i$i$i = ($38|0)>(-1);
    $neg$i$i$i = (0 - ($38))|0;
    $39 = $ispos$i$i$i ? $38 : $neg$i$i$i;
    $40 = ($39|0)>(131071);
    if ($40) {
     $41 = HEAP32[$6>>2]|0;
     _luaX_syntaxerror($41,519652);
     $$pre$i = HEAP32[$7>>2]|0;
     $43 = $$pre$i;
    } else {
     $43 = $8;
    }
    $42 = $43 & 16383;
    $44 = $38 << 14;
    $45 = (($44) + 2147467264)|0;
    $46 = $42 | $45;
    $storemerge$i = $46;
   }
   HEAP32[$7>>2] = $storemerge$i;
   $47 = ($$0$i$i$i|0)==(-1);
   if ($47) {
    break;
   }
   $fs$idx$val$i$i$pre = HEAP32[$fs>>2]|0;
   $$02$i$i = $$0$i$i$i;$fs$idx$val$i$i = $fs$idx$val$i$i$pre;
  }
  $$pre = HEAP32[$3>>2]|0;
  $51 = $$pre;
 }
 HEAP32[$1>>2] = -1;
 $48 = ((($0)) + 44|0);
 $49 = HEAP32[$48>>2]|0;
 $50 = ($51|0)<($49|0);
 if ($50) {
  $$phi$trans$insert = ((($0)) + 12|0);
  $$pre3 = HEAP32[$$phi$trans$insert>>2]|0;
  $58 = $$pre3;$59 = $51;
 } else {
  $52 = ((($fs)) + 16|0);
  $53 = HEAP32[$52>>2]|0;
  $54 = ((($0)) + 12|0);
  $55 = HEAP32[$54>>2]|0;
  $56 = (_luaM_growaux_($53,$55,$48,4,2147483645,519679)|0);
  HEAP32[$54>>2] = $56;
  $$pre2 = HEAP32[$3>>2]|0;
  $58 = $56;$59 = $$pre2;
 }
 $57 = (($58) + ($59<<2)|0);
 HEAP32[$57>>2] = $i;
 $60 = HEAP32[$3>>2]|0;
 $61 = ((($0)) + 48|0);
 $62 = HEAP32[$61>>2]|0;
 $63 = ($60|0)<($62|0);
 if ($63) {
  $$phi$trans$insert5 = ((($0)) + 20|0);
  $$pre6 = HEAP32[$$phi$trans$insert5>>2]|0;
  $70 = $$pre6;$71 = $60;
  $69 = (($70) + ($71<<2)|0);
  HEAP32[$69>>2] = $line;
  $72 = HEAP32[$3>>2]|0;
  $73 = (($72) + 1)|0;
  HEAP32[$3>>2] = $73;
  return ($72|0);
 } else {
  $64 = ((($fs)) + 16|0);
  $65 = HEAP32[$64>>2]|0;
  $66 = ((($0)) + 20|0);
  $67 = HEAP32[$66>>2]|0;
  $68 = (_luaM_growaux_($65,$67,$61,4,2147483645,519679)|0);
  HEAP32[$66>>2] = $68;
  $$pre4 = HEAP32[$3>>2]|0;
  $70 = $68;$71 = $$pre4;
  $69 = (($70) + ($71<<2)|0);
  HEAP32[$69>>2] = $line;
  $72 = HEAP32[$3>>2]|0;
  $73 = (($72) + 1)|0;
  HEAP32[$3>>2] = $73;
  return ($72|0);
 }
 return (0)|0;
}
function _luaK_codeABC($fs,$o,$a,$b,$c) {
 $fs = $fs|0;
 $o = $o|0;
 $a = $a|0;
 $b = $b|0;
 $c = $c|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $a << 6;
 $1 = $0 | $o;
 $2 = $b << 23;
 $3 = $1 | $2;
 $4 = $c << 14;
 $5 = $3 | $4;
 $6 = ((($fs)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($7)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (_luaK_code($fs,$5,$9)|0);
 return ($10|0);
}
function _luaK_jump($fs) {
 $fs = $fs|0;
 var $$0$i$i = 0, $$lcssa = 0, $$lcssa6 = 0, $$neg$i$i = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fs$idx$val$i = 0, $fs$idx$val$idx$i = 0, $fs$idx$val$idx$val$i = 0, $ispos$i$i = 0, $j$0 = 0, $list$0$i = 0, $list$0$i$lcssa = 0, $neg$i$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($fs)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$0>>2] = -1;
 $2 = ((($fs)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (_luaK_code($fs,2147450902,$5)|0);
 $7 = ($1|0)==(-1);
 if ($7) {
  $j$0 = $6;
  return ($j$0|0);
 }
 $8 = ($6|0)==(-1);
 if ($8) {
  $j$0 = $1;
  return ($j$0|0);
 }
 $fs$idx$val$i = HEAP32[$fs>>2]|0;
 $fs$idx$val$idx$i = ((($fs$idx$val$i)) + 12|0);
 $fs$idx$val$idx$val$i = HEAP32[$fs$idx$val$idx$i>>2]|0;
 $list$0$i = $6;
 while(1) {
  $9 = (($fs$idx$val$idx$val$i) + ($list$0$i<<2)|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = $10 >>> 14;
  $12 = (($11) + -131071)|0;
  $13 = ($12|0)==(-1);
  $14 = (($list$0$i) + 1)|0;
  $15 = (($14) + ($12))|0;
  $$0$i$i = $13 ? -1 : $15;
  $16 = ($$0$i$i|0)==(-1);
  if ($16) {
   $$lcssa = $9;$$lcssa6 = $10;$list$0$i$lcssa = $list$0$i;
   break;
  } else {
   $list$0$i = $$0$i$i;
  }
 }
 $$neg$i$i = $list$0$i$lcssa ^ -1;
 $17 = (($1) + ($$neg$i$i))|0;
 $ispos$i$i = ($17|0)>(-1);
 $neg$i$i = (0 - ($17))|0;
 $18 = $ispos$i$i ? $17 : $neg$i$i;
 $19 = ($18|0)>(131071);
 if ($19) {
  $20 = HEAP32[$2>>2]|0;
  _luaX_syntaxerror($20,519652);
  $$pre$i = HEAP32[$$lcssa>>2]|0;
  $22 = $$pre$i;
 } else {
  $22 = $$lcssa6;
 }
 $21 = $22 & 16383;
 $23 = $17 << 14;
 $24 = (($23) + 2147467264)|0;
 $25 = $21 | $24;
 HEAP32[$$lcssa>>2] = $25;
 $j$0 = $6;
 return ($j$0|0);
}
function _luaK_codeABx($fs,$o,$a,$bc) {
 $fs = $fs|0;
 $o = $o|0;
 $a = $a|0;
 $bc = $bc|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $a << 6;
 $1 = $0 | $o;
 $2 = $bc << 14;
 $3 = $1 | $2;
 $4 = ((($fs)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (_luaK_code($fs,$3,$7)|0);
 return ($8|0);
}
function _luaK_concat($fs,$l1,$l2) {
 $fs = $fs|0;
 $l1 = $l1|0;
 $l2 = $l2|0;
 var $$0$i = 0, $$lcssa = 0, $$lcssa4 = 0, $$neg$i = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fs$idx$val = 0, $fs$idx$val$idx = 0, $fs$idx$val$idx$val = 0, $ispos$i = 0, $list$0 = 0, $list$0$lcssa = 0, $neg$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($l2|0)==(-1);
 if ($0) {
  return;
 }
 $1 = HEAP32[$l1>>2]|0;
 $2 = ($1|0)==(-1);
 if ($2) {
  HEAP32[$l1>>2] = $l2;
  return;
 }
 $fs$idx$val = HEAP32[$fs>>2]|0;
 $fs$idx$val$idx = ((($fs$idx$val)) + 12|0);
 $fs$idx$val$idx$val = HEAP32[$fs$idx$val$idx>>2]|0;
 $list$0 = $1;
 while(1) {
  $3 = (($fs$idx$val$idx$val) + ($list$0<<2)|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = $4 >>> 14;
  $6 = (($5) + -131071)|0;
  $7 = ($6|0)==(-1);
  $8 = (($list$0) + 1)|0;
  $9 = (($8) + ($6))|0;
  $$0$i = $7 ? -1 : $9;
  $10 = ($$0$i|0)==(-1);
  if ($10) {
   $$lcssa = $3;$$lcssa4 = $4;$list$0$lcssa = $list$0;
   break;
  } else {
   $list$0 = $$0$i;
  }
 }
 $$neg$i = $list$0$lcssa ^ -1;
 $11 = (($$neg$i) + ($l2))|0;
 $ispos$i = ($11|0)>(-1);
 $neg$i = (0 - ($11))|0;
 $12 = $ispos$i ? $11 : $neg$i;
 $13 = ($12|0)>(131071);
 if ($13) {
  $14 = ((($fs)) + 12|0);
  $15 = HEAP32[$14>>2]|0;
  _luaX_syntaxerror($15,519652);
  $$pre = HEAP32[$$lcssa>>2]|0;
  $17 = $$pre;
 } else {
  $17 = $$lcssa4;
 }
 $16 = $17 & 16383;
 $18 = $11 << 14;
 $19 = (($18) + 2147467264)|0;
 $20 = $16 | $19;
 HEAP32[$$lcssa>>2] = $20;
 return;
}
function _luaK_ret($fs,$first,$nret) {
 $fs = $fs|0;
 $first = $first|0;
 $nret = $nret|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $first << 6;
 $1 = $nret << 23;
 $2 = (($1) + 8388608)|0;
 $3 = $0 | $2;
 $4 = $3 | 30;
 $5 = ((($fs)) + 12|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($6)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 (_luaK_code($fs,$4,$8)|0);
 return;
}
function _luaK_getlabel($fs) {
 $fs = $fs|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fs)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fs)) + 28|0);
 HEAP32[$2>>2] = $1;
 return ($1|0);
}
function _luaK_patchlist($fs,$list,$target) {
 $fs = $fs|0;
 $list = $list|0;
 $target = $target|0;
 var $$0$i$i = 0, $$0$i$i$i = 0, $$0$i$i1 = 0, $$02$i = 0, $$lcssa = 0, $$lcssa7 = 0, $$neg$i = 0, $$neg$i$i = 0, $$neg$i$i$i = 0, $$pre = 0, $$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $fs$idx$val$i = 0, $fs$idx$val$i$i = 0, $fs$idx$val$idx$i = 0, $fs$idx$val$idx$i$i = 0, $fs$idx$val$idx$val$i = 0, $fs$idx$val$idx$val$i$i = 0, $ispos$i = 0, $ispos$i$i = 0, $ispos$i$i$i = 0, $list$0$i$i = 0, $list$0$i$i$lcssa = 0, $neg$i = 0, $neg$i$i = 0, $neg$i$i$i = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fs)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==($target|0);
 if ($2) {
  $3 = ((($fs)) + 28|0);
  HEAP32[$3>>2] = $target;
  $4 = ((($fs)) + 32|0);
  $5 = ($list|0)==(-1);
  if ($5) {
   return;
  }
  $6 = HEAP32[$4>>2]|0;
  $7 = ($6|0)==(-1);
  if ($7) {
   HEAP32[$4>>2] = $list;
   return;
  }
  $fs$idx$val$i$i = HEAP32[$fs>>2]|0;
  $fs$idx$val$idx$i$i = ((($fs$idx$val$i$i)) + 12|0);
  $fs$idx$val$idx$val$i$i = HEAP32[$fs$idx$val$idx$i$i>>2]|0;
  $list$0$i$i = $6;
  while(1) {
   $8 = (($fs$idx$val$idx$val$i$i) + ($list$0$i$i<<2)|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = $9 >>> 14;
   $11 = (($10) + -131071)|0;
   $12 = ($11|0)==(-1);
   $13 = (($list$0$i$i) + 1)|0;
   $14 = (($13) + ($11))|0;
   $$0$i$i$i = $12 ? -1 : $14;
   $15 = ($$0$i$i$i|0)==(-1);
   if ($15) {
    $$lcssa = $8;$$lcssa7 = $9;$list$0$i$i$lcssa = $list$0$i$i;
    break;
   } else {
    $list$0$i$i = $$0$i$i$i;
   }
  }
  $$neg$i$i$i = $list$0$i$i$lcssa ^ -1;
  $16 = (($$neg$i$i$i) + ($list))|0;
  $ispos$i$i$i = ($16|0)>(-1);
  $neg$i$i$i = (0 - ($16))|0;
  $17 = $ispos$i$i$i ? $16 : $neg$i$i$i;
  $18 = ($17|0)>(131071);
  if ($18) {
   $19 = ((($fs)) + 12|0);
   $20 = HEAP32[$19>>2]|0;
   _luaX_syntaxerror($20,519652);
   $$pre$i$i = HEAP32[$$lcssa>>2]|0;
   $22 = $$pre$i$i;
  } else {
   $22 = $$lcssa7;
  }
  $21 = $22 & 16383;
  $23 = $16 << 14;
  $24 = (($23) + 2147467264)|0;
  $25 = $21 | $24;
  HEAP32[$$lcssa>>2] = $25;
  return;
 }
 $26 = ($list|0)==(-1);
 if ($26) {
  return;
 }
 $27 = ((($fs)) + 12|0);
 $$02$i = $list;
 while(1) {
  $fs$idx$val$i = HEAP32[$fs>>2]|0;
  $fs$idx$val$idx$i = ((($fs$idx$val$i)) + 12|0);
  $fs$idx$val$idx$val$i = HEAP32[$fs$idx$val$idx$i>>2]|0;
  $28 = (($fs$idx$val$idx$val$i) + ($$02$i<<2)|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = $29 >>> 14;
  $31 = (($30) + -131071)|0;
  $32 = ($31|0)==(-1);
  $33 = (($$02$i) + 1)|0;
  $34 = (($33) + ($31))|0;
  $$0$i$i = $32 ? -1 : $34;
  $35 = ($$02$i|0)>(0);
  if ($35) {
   $36 = ((($28)) + -4|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = $37 & 63;
   $39 = (518247 + ($38)|0);
   $40 = HEAP8[$39>>0]|0;
   $41 = ($40<<24>>24)<(0);
   if ($41) {
    $$0$i$i1 = $36;$43 = $37;
   } else {
    label = 14;
   }
  } else {
   label = 14;
  }
  if ((label|0) == 14) {
   label = 0;
   $$0$i$i1 = $28;$43 = $29;
  }
  $42 = $43 & 63;
  $44 = ($42|0)==(27);
  if ($44) {
   $45 = $43 >>> 23;
   $46 = $45 << 6;
   $47 = $43 & 8372224;
   $48 = $47 | $46;
   $49 = $48 | 26;
   HEAP32[$$0$i$i1>>2] = $49;
   $$neg$i = $$02$i ^ -1;
   $50 = (($$neg$i) + ($target))|0;
   $ispos$i = ($50|0)>(-1);
   $neg$i = (0 - ($50))|0;
   $51 = $ispos$i ? $50 : $neg$i;
   $52 = ($51|0)>(131071);
   if ($52) {
    $53 = HEAP32[$27>>2]|0;
    _luaX_syntaxerror($53,519652);
   }
   $54 = HEAP32[$28>>2]|0;
   $55 = $54 & 16383;
   $56 = $50 << 14;
   $57 = (($56) + 2147467264)|0;
   $58 = $55 | $57;
   $storemerge = $58;
  } else {
   $$neg$i$i = $$02$i ^ -1;
   $59 = (($$neg$i$i) + ($target))|0;
   $ispos$i$i = ($59|0)>(-1);
   $neg$i$i = (0 - ($59))|0;
   $60 = $ispos$i$i ? $59 : $neg$i$i;
   $61 = ($60|0)>(131071);
   if ($61) {
    $62 = HEAP32[$27>>2]|0;
    _luaX_syntaxerror($62,519652);
    $$pre = HEAP32[$28>>2]|0;
    $64 = $$pre;
   } else {
    $64 = $29;
   }
   $63 = $64 & 16383;
   $65 = $59 << 14;
   $66 = (($65) + 2147467264)|0;
   $67 = $63 | $66;
   $storemerge = $67;
  }
  HEAP32[$28>>2] = $storemerge;
  $68 = ($$0$i$i|0)==(-1);
  if ($68) {
   break;
  } else {
   $$02$i = $$0$i$i;
  }
 }
 return;
}
function _luaK_patchtohere($fs,$list) {
 $fs = $fs|0;
 $list = $list|0;
 var $$0$i$i = 0, $$lcssa = 0, $$lcssa6 = 0, $$neg$i$i = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fs$idx$val$i = 0, $fs$idx$val$idx$i = 0, $fs$idx$val$idx$val$i = 0, $ispos$i$i = 0, $list$0$i = 0, $list$0$i$lcssa = 0, $neg$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fs)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fs)) + 28|0);
 HEAP32[$2>>2] = $1;
 $3 = ((($fs)) + 32|0);
 $4 = ($list|0)==(-1);
 if ($4) {
  return;
 }
 $5 = HEAP32[$3>>2]|0;
 $6 = ($5|0)==(-1);
 if ($6) {
  HEAP32[$3>>2] = $list;
  return;
 }
 $fs$idx$val$i = HEAP32[$fs>>2]|0;
 $fs$idx$val$idx$i = ((($fs$idx$val$i)) + 12|0);
 $fs$idx$val$idx$val$i = HEAP32[$fs$idx$val$idx$i>>2]|0;
 $list$0$i = $5;
 while(1) {
  $7 = (($fs$idx$val$idx$val$i) + ($list$0$i<<2)|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = $8 >>> 14;
  $10 = (($9) + -131071)|0;
  $11 = ($10|0)==(-1);
  $12 = (($list$0$i) + 1)|0;
  $13 = (($12) + ($10))|0;
  $$0$i$i = $11 ? -1 : $13;
  $14 = ($$0$i$i|0)==(-1);
  if ($14) {
   $$lcssa = $7;$$lcssa6 = $8;$list$0$i$lcssa = $list$0$i;
   break;
  } else {
   $list$0$i = $$0$i$i;
  }
 }
 $$neg$i$i = $list$0$i$lcssa ^ -1;
 $15 = (($$neg$i$i) + ($list))|0;
 $ispos$i$i = ($15|0)>(-1);
 $neg$i$i = (0 - ($15))|0;
 $16 = $ispos$i$i ? $15 : $neg$i$i;
 $17 = ($16|0)>(131071);
 if ($17) {
  $18 = ((($fs)) + 12|0);
  $19 = HEAP32[$18>>2]|0;
  _luaX_syntaxerror($19,519652);
  $$pre$i = HEAP32[$$lcssa>>2]|0;
  $21 = $$pre$i;
 } else {
  $21 = $$lcssa6;
 }
 $20 = $21 & 16383;
 $22 = $15 << 14;
 $23 = (($22) + 2147467264)|0;
 $24 = $20 | $23;
 HEAP32[$$lcssa>>2] = $24;
 return;
}
function _luaK_checkstack($fs,$n) {
 $fs = $fs|0;
 $n = $n|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fs)) + 36|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + ($n))|0;
 $3 = HEAP32[$fs>>2]|0;
 $4 = ((($3)) + 75|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = ($2|0)>($6|0);
 if (!($7)) {
  return;
 }
 $8 = ($2|0)>(249);
 if ($8) {
  $9 = ((($fs)) + 12|0);
  $10 = HEAP32[$9>>2]|0;
  _luaX_syntaxerror($10,519698);
  $$pre = HEAP32[$fs>>2]|0;
  $13 = $$pre;
 } else {
  $13 = $3;
 }
 $11 = $2&255;
 $12 = ((($13)) + 75|0);
 HEAP8[$12>>0] = $11;
 return;
}
function _luaK_reserveregs($fs,$n) {
 $fs = $fs|0;
 $n = $n|0;
 var $$pre = 0, $$pre$i = 0, $$pre$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($fs)) + 36|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + ($n))|0;
 $3 = HEAP32[$fs>>2]|0;
 $4 = ((($3)) + 75|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = ($2|0)>($6|0);
 if (!($7)) {
  $15 = $1;
  $14 = (($15) + ($n))|0;
  HEAP32[$0>>2] = $14;
  return;
 }
 $8 = ($2|0)>(249);
 if ($8) {
  $9 = ((($fs)) + 12|0);
  $10 = HEAP32[$9>>2]|0;
  _luaX_syntaxerror($10,519698);
  $$pre$i = HEAP32[$fs>>2]|0;
  $$pre$pre = HEAP32[$0>>2]|0;
  $$pre = $$pre$pre;$13 = $$pre$i;
 } else {
  $$pre = $1;$13 = $3;
 }
 $11 = $2&255;
 $12 = ((($13)) + 75|0);
 HEAP8[$12>>0] = $11;
 $15 = $$pre;
 $14 = (($15) + ($n))|0;
 HEAP32[$0>>2] = $14;
 return;
}
function _luaK_stringK($fs,$s) {
 $fs = $fs|0;
 $s = $s|0;
 var $0 = 0, $1 = 0, $o = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $o = sp;
 HEAP32[$o>>2] = $s;
 $0 = ((($o)) + 8|0);
 HEAP32[$0>>2] = 4;
 $1 = (_addk($fs,$o,$o)|0);
 STACKTOP = sp;return ($1|0);
}
function _addk($fs,$k,$v) {
 $fs = $fs|0;
 $k = $k|0;
 $v = $v|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $oldsize$01 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($fs)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fs)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_luaH_set($1,$3,$k)|0);
 $5 = HEAP32[$fs>>2]|0;
 $6 = ((($5)) + 40|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($4)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)==(3);
 if ($10) {
  $11 = +HEAPF64[$4>>3];
  $12 = (~~(($11)));
  $$0 = $12;
  return ($$0|0);
 }
 $13 = ((($fs)) + 40|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (+($14|0));
 HEAPF64[$4>>3] = $15;
 HEAP32[$8>>2] = 3;
 $16 = ($14|0)<($7|0);
 if ($16) {
  $21 = $7;
 } else {
  $17 = ((($5)) + 8|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = (_luaM_growaux_($1,$18,$6,16,262143,519733)|0);
  HEAP32[$17>>2] = $19;
  $$pre = HEAP32[$6>>2]|0;
  $21 = $$pre;
 }
 $20 = ($7|0)<($21|0);
 $22 = ((($5)) + 8|0);
 $23 = HEAP32[$22>>2]|0;
 if ($20) {
  $24 = HEAP32[$6>>2]|0;
  $oldsize$01 = $7;
  while(1) {
   $25 = (($oldsize$01) + 1)|0;
   $26 = (((($23) + ($oldsize$01<<4)|0)) + 8|0);
   HEAP32[$26>>2] = 0;
   $27 = ($25|0)<($24|0);
   if ($27) {
    $oldsize$01 = $25;
   } else {
    break;
   }
  }
 }
 $28 = HEAP32[$13>>2]|0;
 $29 = (($23) + ($28<<4)|0);
 $30 = $v;
 $31 = $30;
 $32 = HEAP32[$31>>2]|0;
 $33 = (($30) + 4)|0;
 $34 = $33;
 $35 = HEAP32[$34>>2]|0;
 $36 = $29;
 $37 = $36;
 HEAP32[$37>>2] = $32;
 $38 = (($36) + 4)|0;
 $39 = $38;
 HEAP32[$39>>2] = $35;
 $40 = ((($v)) + 8|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = (((($23) + ($28<<4)|0)) + 8|0);
 HEAP32[$42>>2] = $41;
 $43 = HEAP32[$40>>2]|0;
 $44 = ($43|0)>(3);
 if ($44) {
  $45 = HEAP32[$v>>2]|0;
  $46 = ((($45)) + 5|0);
  $47 = HEAP8[$46>>0]|0;
  $48 = $47 & 3;
  $49 = ($48<<24>>24)==(0);
  if (!($49)) {
   $50 = ((($5)) + 5|0);
   $51 = HEAP8[$50>>0]|0;
   $52 = $51 & 4;
   $53 = ($52<<24>>24)==(0);
   if (!($53)) {
    _luaC_barrierf($1,$5,$45);
   }
  }
 }
 $54 = HEAP32[$13>>2]|0;
 $55 = (($54) + 1)|0;
 HEAP32[$13>>2] = $55;
 $$0 = $54;
 return ($$0|0);
}
function _luaK_numberK($fs,$r) {
 $fs = $fs|0;
 $r = +$r;
 var $0 = 0, $1 = 0, $o = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $o = sp;
 HEAPF64[$o>>3] = $r;
 $0 = ((($o)) + 8|0);
 HEAP32[$0>>2] = 3;
 $1 = (_addk($fs,$o,$o)|0);
 STACKTOP = sp;return ($1|0);
}
function _luaK_setreturns($fs,$e,$nresults) {
 $fs = $fs|0;
 $e = $e|0;
 $nresults = $nresults|0;
 var $$pre$i = 0, $$pre$i$i = 0, $$pre$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$e>>2]|0;
 switch ($0|0) {
 case 13:  {
  $1 = ((($e)) + 8|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = HEAP32[$fs>>2]|0;
  $4 = ((($3)) + 12|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = (($5) + ($2<<2)|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = $7 & -8372225;
  $9 = $nresults << 14;
  $10 = (($9) + 16384)|0;
  $11 = $10 & 8372224;
  $12 = $8 | $11;
  HEAP32[$6>>2] = $12;
  return;
  break;
 }
 case 14:  {
  $13 = ((($e)) + 8|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = HEAP32[$fs>>2]|0;
  $16 = ((($15)) + 12|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = (($17) + ($14<<2)|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = $19 & 8388607;
  $21 = $nresults << 23;
  $22 = (($21) + 8388608)|0;
  $23 = $20 | $22;
  HEAP32[$18>>2] = $23;
  $24 = HEAP32[$13>>2]|0;
  $25 = (($17) + ($24<<2)|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = $26 & -16321;
  $28 = ((($fs)) + 36|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = $29 << 6;
  $31 = $30 & 16320;
  $32 = $31 | $27;
  HEAP32[$25>>2] = $32;
  $33 = HEAP32[$28>>2]|0;
  $34 = (($33) + 1)|0;
  $35 = ((($15)) + 75|0);
  $36 = HEAP8[$35>>0]|0;
  $37 = $36&255;
  $38 = ($33|0)<($37|0);
  if ($38) {
   $46 = $33;
  } else {
   $39 = ($33|0)>(248);
   if ($39) {
    $40 = ((($fs)) + 12|0);
    $41 = HEAP32[$40>>2]|0;
    _luaX_syntaxerror($41,519698);
    $$pre$i$i = HEAP32[$fs>>2]|0;
    $$pre$pre$i = HEAP32[$28>>2]|0;
    $$pre$i = $$pre$pre$i;$44 = $$pre$i$i;
   } else {
    $$pre$i = $33;$44 = $15;
   }
   $42 = $34&255;
   $43 = ((($44)) + 75|0);
   HEAP8[$43>>0] = $42;
   $46 = $$pre$i;
  }
  $45 = (($46) + 1)|0;
  HEAP32[$28>>2] = $45;
  return;
  break;
 }
 default: {
  return;
 }
 }
}
function _luaK_setoneret($fs,$e) {
 $fs = $fs|0;
 $e = $e|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$e>>2]|0;
 switch ($0|0) {
 case 13:  {
  HEAP32[$e>>2] = 12;
  $1 = ((($e)) + 8|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = HEAP32[$fs>>2]|0;
  $4 = ((($3)) + 12|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = (($5) + ($2<<2)|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = $7 >>> 6;
  $9 = $8 & 255;
  HEAP32[$1>>2] = $9;
  return;
  break;
 }
 case 14:  {
  $10 = ((($e)) + 8|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = HEAP32[$fs>>2]|0;
  $13 = ((($12)) + 12|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = (($14) + ($11<<2)|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = $16 & 8388607;
  $18 = $17 | 16777216;
  HEAP32[$15>>2] = $18;
  HEAP32[$e>>2] = 11;
  return;
  break;
 }
 default: {
  return;
 }
 }
}
function _luaK_dischargevars($fs,$e) {
 $fs = $fs|0;
 $e = $e|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$e>>2]|0;
 switch ($0|0) {
 case 6:  {
  HEAP32[$e>>2] = 12;
  return;
  break;
 }
 case 7:  {
  $1 = ((($e)) + 8|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = $2 << 23;
  $4 = $3 | 4;
  $5 = ((($fs)) + 12|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ((($6)) + 8|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = (_luaK_code($fs,$4,$8)|0);
  HEAP32[$1>>2] = $9;
  HEAP32[$e>>2] = 11;
  return;
  break;
 }
 case 8:  {
  $10 = ((($e)) + 8|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = $11 << 14;
  $13 = $12 | 5;
  $14 = ((($fs)) + 12|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ((($15)) + 8|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = (_luaK_code($fs,$13,$17)|0);
  HEAP32[$10>>2] = $18;
  HEAP32[$e>>2] = 11;
  return;
  break;
 }
 case 9:  {
  $19 = ((($e)) + 8|0);
  $20 = ((($19)) + 4|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = $21 & 256;
  $23 = ($22|0)==(0);
  if ($23) {
   $24 = ((($fs)) + 50|0);
   $25 = HEAP8[$24>>0]|0;
   $26 = $25&255;
   $27 = ($26|0)>($21|0);
   if (!($27)) {
    $28 = ((($fs)) + 36|0);
    $29 = HEAP32[$28>>2]|0;
    $30 = (($29) + -1)|0;
    HEAP32[$28>>2] = $30;
   }
  }
  $31 = HEAP32[$19>>2]|0;
  $32 = $31 & 256;
  $33 = ($32|0)==(0);
  if ($33) {
   $34 = ((($fs)) + 50|0);
   $35 = HEAP8[$34>>0]|0;
   $36 = $35&255;
   $37 = ($36|0)>($31|0);
   if (!($37)) {
    $38 = ((($fs)) + 36|0);
    $39 = HEAP32[$38>>2]|0;
    $40 = (($39) + -1)|0;
    HEAP32[$38>>2] = $40;
   }
  }
  $41 = $31 << 23;
  $42 = $21 << 14;
  $43 = $41 | $42;
  $44 = $43 | 6;
  $45 = ((($fs)) + 12|0);
  $46 = HEAP32[$45>>2]|0;
  $47 = ((($46)) + 8|0);
  $48 = HEAP32[$47>>2]|0;
  $49 = (_luaK_code($fs,$44,$48)|0);
  HEAP32[$19>>2] = $49;
  HEAP32[$e>>2] = 11;
  return;
  break;
 }
 case 13:  {
  HEAP32[$e>>2] = 12;
  $50 = ((($e)) + 8|0);
  $51 = HEAP32[$50>>2]|0;
  $52 = HEAP32[$fs>>2]|0;
  $53 = ((($52)) + 12|0);
  $54 = HEAP32[$53>>2]|0;
  $55 = (($54) + ($51<<2)|0);
  $56 = HEAP32[$55>>2]|0;
  $57 = $56 >>> 6;
  $58 = $57 & 255;
  HEAP32[$50>>2] = $58;
  return;
  break;
 }
 case 14:  {
  $59 = ((($e)) + 8|0);
  $60 = HEAP32[$59>>2]|0;
  $61 = HEAP32[$fs>>2]|0;
  $62 = ((($61)) + 12|0);
  $63 = HEAP32[$62>>2]|0;
  $64 = (($63) + ($60<<2)|0);
  $65 = HEAP32[$64>>2]|0;
  $66 = $65 & 8388607;
  $67 = $66 | 16777216;
  HEAP32[$64>>2] = $67;
  HEAP32[$e>>2] = 11;
  return;
  break;
 }
 default: {
  return;
 }
 }
}
function _luaK_exp2nextreg($fs,$e) {
 $fs = $fs|0;
 $e = $e|0;
 var $$pre$i = 0, $$pre$i$i = 0, $$pre$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaK_dischargevars($fs,$e);
 $0 = HEAP32[$e>>2]|0;
 $1 = ($0|0)==(12);
 if ($1) {
  $2 = ((($e)) + 8|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = $3 & 256;
  $5 = ($4|0)==(0);
  if ($5) {
   $6 = ((($fs)) + 50|0);
   $7 = HEAP8[$6>>0]|0;
   $8 = $7&255;
   $9 = ($8|0)>($3|0);
   if (!($9)) {
    $10 = ((($fs)) + 36|0);
    $11 = HEAP32[$10>>2]|0;
    $12 = (($11) + -1)|0;
    HEAP32[$10>>2] = $12;
   }
  }
 }
 $13 = ((($fs)) + 36|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($14) + 1)|0;
 $16 = HEAP32[$fs>>2]|0;
 $17 = ((($16)) + 75|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = $18&255;
 $20 = ($14|0)<($19|0);
 if ($20) {
  $28 = $14;
  $27 = (($28) + 1)|0;
  HEAP32[$13>>2] = $27;
  _exp2reg($fs,$e,$28);
  return;
 }
 $21 = ($14|0)>(248);
 if ($21) {
  $22 = ((($fs)) + 12|0);
  $23 = HEAP32[$22>>2]|0;
  _luaX_syntaxerror($23,519698);
  $$pre$i$i = HEAP32[$fs>>2]|0;
  $$pre$pre$i = HEAP32[$13>>2]|0;
  $$pre$i = $$pre$pre$i;$26 = $$pre$i$i;
 } else {
  $$pre$i = $14;$26 = $16;
 }
 $24 = $15&255;
 $25 = ((($26)) + 75|0);
 HEAP8[$25>>0] = $24;
 $28 = $$pre$i;
 $27 = (($28) + 1)|0;
 HEAP32[$13>>2] = $27;
 _exp2reg($fs,$e,$28);
 return;
}
function _exp2reg($fs,$e,$reg) {
 $fs = $fs|0;
 $e = $e|0;
 $reg = $reg|0;
 var $$0$i$i = 0, $$0$i$i$i = 0, $$0$i$i$i33 = 0, $$0$i$i25 = 0, $$0$i$i25$us = 0, $$0$i$i45 = 0, $$0$i$i45$us = 0, $$0$i$i57 = 0, $$0$i$i57$us = 0, $$0$i$i7 = 0, $$0$i$i7$us = 0, $$02$i = 0, $$02$i$us = 0, $$02$i41 = 0, $$02$i41$us = 0, $$026$i = 0, $$026$i17 = 0, $$5$i = 0, $$5$i20 = 0, $$lcssa = 0;
 var $$lcssa112 = 0, $$lcssa113 = 0, $$lcssa114 = 0, $$lcssa115 = 0, $$lcssa116 = 0, $$neg$i = 0, $$neg$i$i = 0, $$neg$i$i$i = 0, $$neg$i$i$i34 = 0, $$neg$i$i46 = 0, $$neg$i$i46$us = 0, $$neg$i$i8 = 0, $$neg$i$i8$us = 0, $$neg$i$us = 0, $$neg$i52 = 0, $$neg$i52$us = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i37 = 0;
 var $$pre$i1 = 0, $$pre$i18 = 0, $$pre$phi82Z2D = 0, $$pre$phi84Z2D = 0, $$pre77 = 0, $$pre78 = 0, $$pre79 = 0, $$pre81 = 0, $$pre83 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0;
 var $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0;
 var $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0;
 var $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0;
 var $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0;
 var $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0;
 var $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0;
 var $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fs$idx$val$i = 0, $fs$idx$val$i$i = 0, $fs$idx$val$i$i28 = 0, $fs$idx$val$i4 = 0, $fs$idx$val$i4$us = 0, $fs$idx$val$i42 = 0, $fs$idx$val$i42$us = 0, $fs$idx$val$idx$i = 0, $fs$idx$val$idx$i$i = 0, $fs$idx$val$idx$i$i29 = 0, $fs$idx$val$idx$i43 = 0, $fs$idx$val$idx$i43$us = 0, $fs$idx$val$idx$i5 = 0, $fs$idx$val$idx$i5$us = 0;
 var $fs$idx$val$idx$val$i = 0, $fs$idx$val$idx$val$i$i = 0, $fs$idx$val$idx$val$i$i30 = 0, $fs$idx$val$idx$val$i44 = 0, $fs$idx$val$idx$val$i44$us = 0, $fs$idx$val$idx$val$i6 = 0, $fs$idx$val$idx$val$i6$us = 0, $fs$idx3$val$i = 0, $fs$idx3$val$i13 = 0, $fs$idx3$val$idx$i = 0, $fs$idx3$val$idx$i14 = 0, $fs$idx3$val$idx$val$i = 0, $fs$idx3$val$idx$val$i15 = 0, $ispos$i = 0, $ispos$i$i = 0, $ispos$i$i$i = 0, $ispos$i$i$i35 = 0, $ispos$i$i47 = 0, $ispos$i$i47$us = 0, $ispos$i$i9 = 0;
 var $ispos$i$i9$us = 0, $ispos$i$us = 0, $ispos$i53 = 0, $ispos$i53$us = 0, $list$0$i = 0, $list$0$i$i = 0, $list$0$i$i$lcssa = 0, $list$0$i$i32 = 0, $list$0$i$i32$lcssa = 0, $list$0$i$lcssa = 0, $neg$i = 0, $neg$i$i = 0, $neg$i$i$i = 0, $neg$i$i$i36 = 0, $neg$i$i10 = 0, $neg$i$i10$us = 0, $neg$i$i48 = 0, $neg$i$i48$us = 0, $neg$i$us = 0, $neg$i54 = 0;
 var $neg$i54$us = 0, $p_f$0 = 0, $p_t$0 = 0, $storemerge = 0, $storemerge$i = 0, $storemerge$i60 = 0, $storemerge85 = 0, $storemerge86 = 0, $storemerge87 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _discharge2reg($fs,$e,$reg);
 $0 = HEAP32[$e>>2]|0;
 $1 = ($0|0)==(10);
 $2 = ((($e)) + 16|0);
 do {
  if ($1) {
   $3 = ((($e)) + 8|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==(-1);
   if (!($5)) {
    $6 = HEAP32[$2>>2]|0;
    $7 = ($6|0)==(-1);
    if ($7) {
     HEAP32[$2>>2] = $4;
     break;
    }
    $fs$idx$val$i = HEAP32[$fs>>2]|0;
    $fs$idx$val$idx$i = ((($fs$idx$val$i)) + 12|0);
    $fs$idx$val$idx$val$i = HEAP32[$fs$idx$val$idx$i>>2]|0;
    $list$0$i = $6;
    while(1) {
     $8 = (($fs$idx$val$idx$val$i) + ($list$0$i<<2)|0);
     $9 = HEAP32[$8>>2]|0;
     $10 = $9 >>> 14;
     $11 = (($10) + -131071)|0;
     $12 = ($11|0)==(-1);
     $13 = (($list$0$i) + 1)|0;
     $14 = (($13) + ($11))|0;
     $$0$i$i = $12 ? -1 : $14;
     $15 = ($$0$i$i|0)==(-1);
     if ($15) {
      $$lcssa115 = $8;$$lcssa116 = $9;$list$0$i$lcssa = $list$0$i;
      break;
     } else {
      $list$0$i = $$0$i$i;
     }
    }
    $$neg$i$i = $list$0$i$lcssa ^ -1;
    $16 = (($4) + ($$neg$i$i))|0;
    $ispos$i$i = ($16|0)>(-1);
    $neg$i$i = (0 - ($16))|0;
    $17 = $ispos$i$i ? $16 : $neg$i$i;
    $18 = ($17|0)>(131071);
    if ($18) {
     $19 = ((($fs)) + 12|0);
     $20 = HEAP32[$19>>2]|0;
     _luaX_syntaxerror($20,519652);
     $$pre$i = HEAP32[$$lcssa115>>2]|0;
     $22 = $$pre$i;
    } else {
     $22 = $$lcssa116;
    }
    $21 = $22 & 16383;
    $23 = $16 << 14;
    $24 = (($23) + 2147467264)|0;
    $25 = $21 | $24;
    HEAP32[$$lcssa115>>2] = $25;
   }
  }
 } while(0);
 $26 = HEAP32[$2>>2]|0;
 $27 = ((($e)) + 20|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ($26|0)==($28|0);
 if ($29) {
  HEAP32[$2>>2] = -1;
  HEAP32[$27>>2] = -1;
  $316 = ((($e)) + 8|0);
  HEAP32[$316>>2] = $reg;
  HEAP32[$e>>2] = 12;
  return;
 }
 $30 = ($26|0)==(-1);
 L17: do {
  if ($30) {
   label = 18;
  } else {
   $fs$idx3$val$i = HEAP32[$fs>>2]|0;
   $fs$idx3$val$idx$i = ((($fs$idx3$val$i)) + 12|0);
   $fs$idx3$val$idx$val$i = HEAP32[$fs$idx3$val$idx$i>>2]|0;
   $$026$i = $26;
   while(1) {
    $31 = (($fs$idx3$val$idx$val$i) + ($$026$i<<2)|0);
    $32 = ($$026$i|0)>(0);
    if ($32) {
     $33 = ((($31)) + -4|0);
     $34 = HEAP32[$33>>2]|0;
     $35 = $34 & 63;
     $36 = (518247 + ($35)|0);
     $37 = HEAP8[$36>>0]|0;
     $38 = ($37<<24>>24)<(0);
     if ($38) {
      $40 = $34;
     } else {
      label = 15;
     }
    } else {
     label = 15;
    }
    if ((label|0) == 15) {
     label = 0;
     $$pre$i1 = HEAP32[$31>>2]|0;
     $40 = $$pre$i1;
    }
    $39 = $40 & 63;
    $41 = ($39|0)==(27);
    if (!($41)) {
     label = 25;
     break L17;
    }
    $42 = HEAP32[$31>>2]|0;
    $43 = $42 >>> 14;
    $44 = (($43) + -131071)|0;
    $45 = ($44|0)==(-1);
    $46 = (($$026$i) + 1)|0;
    $47 = (($46) + ($44))|0;
    $$5$i = $45 ? -1 : $47;
    $48 = ($$5$i|0)==(-1);
    if ($48) {
     label = 18;
     break;
    } else {
     $$026$i = $$5$i;
    }
   }
  }
 } while(0);
 L27: do {
  if ((label|0) == 18) {
   $49 = ($28|0)==(-1);
   if ($49) {
    $p_f$0 = -1;$p_t$0 = -1;
   } else {
    $fs$idx3$val$i13 = HEAP32[$fs>>2]|0;
    $fs$idx3$val$idx$i14 = ((($fs$idx3$val$i13)) + 12|0);
    $fs$idx3$val$idx$val$i15 = HEAP32[$fs$idx3$val$idx$i14>>2]|0;
    $$026$i17 = $28;
    while(1) {
     $50 = (($fs$idx3$val$idx$val$i15) + ($$026$i17<<2)|0);
     $51 = ($$026$i17|0)>(0);
     if ($51) {
      $52 = ((($50)) + -4|0);
      $53 = HEAP32[$52>>2]|0;
      $54 = $53 & 63;
      $55 = (518247 + ($54)|0);
      $56 = HEAP8[$55>>0]|0;
      $57 = ($56<<24>>24)<(0);
      if ($57) {
       $59 = $53;
      } else {
       label = 22;
      }
     } else {
      label = 22;
     }
     if ((label|0) == 22) {
      label = 0;
      $$pre$i18 = HEAP32[$50>>2]|0;
      $59 = $$pre$i18;
     }
     $58 = $59 & 63;
     $60 = ($58|0)==(27);
     if (!($60)) {
      label = 25;
      break L27;
     }
     $61 = HEAP32[$50>>2]|0;
     $62 = $61 >>> 14;
     $63 = (($62) + -131071)|0;
     $64 = ($63|0)==(-1);
     $65 = (($$026$i17) + 1)|0;
     $66 = (($65) + ($63))|0;
     $$5$i20 = $64 ? -1 : $66;
     $67 = ($$5$i20|0)==(-1);
     if ($67) {
      $p_f$0 = -1;$p_t$0 = -1;
      break;
     } else {
      $$026$i17 = $$5$i20;
     }
    }
   }
  }
 } while(0);
 do {
  if ((label|0) == 25) {
   $68 = HEAP32[$e>>2]|0;
   $69 = ($68|0)==(10);
   if ($69) {
    $$pre81 = ((($fs)) + 12|0);
    $$pre83 = ((($fs)) + 32|0);
    $$pre$phi82Z2D = $$pre81;$$pre$phi84Z2D = $$pre83;$113 = -1;
   } else {
    $70 = ((($fs)) + 32|0);
    $71 = HEAP32[$70>>2]|0;
    HEAP32[$70>>2] = -1;
    $72 = ((($fs)) + 12|0);
    $73 = HEAP32[$72>>2]|0;
    $74 = ((($73)) + 8|0);
    $75 = HEAP32[$74>>2]|0;
    $76 = (_luaK_code($fs,2147450902,$75)|0);
    $77 = ($71|0)==(-1);
    if ($77) {
     $$pre$phi82Z2D = $72;$$pre$phi84Z2D = $70;$113 = $76;
    } else {
     $78 = ($76|0)==(-1);
     if ($78) {
      $$pre$phi82Z2D = $72;$$pre$phi84Z2D = $70;$113 = $71;
     } else {
      $fs$idx$val$i$i = HEAP32[$fs>>2]|0;
      $fs$idx$val$idx$i$i = ((($fs$idx$val$i$i)) + 12|0);
      $fs$idx$val$idx$val$i$i = HEAP32[$fs$idx$val$idx$i$i>>2]|0;
      $list$0$i$i = $76;
      while(1) {
       $79 = (($fs$idx$val$idx$val$i$i) + ($list$0$i$i<<2)|0);
       $80 = HEAP32[$79>>2]|0;
       $81 = $80 >>> 14;
       $82 = (($81) + -131071)|0;
       $83 = ($82|0)==(-1);
       $84 = (($list$0$i$i) + 1)|0;
       $85 = (($84) + ($82))|0;
       $$0$i$i$i = $83 ? -1 : $85;
       $86 = ($$0$i$i$i|0)==(-1);
       if ($86) {
        $$lcssa113 = $79;$$lcssa114 = $80;$list$0$i$i$lcssa = $list$0$i$i;
        break;
       } else {
        $list$0$i$i = $$0$i$i$i;
       }
      }
      $$neg$i$i$i = $list$0$i$i$lcssa ^ -1;
      $87 = (($71) + ($$neg$i$i$i))|0;
      $ispos$i$i$i = ($87|0)>(-1);
      $neg$i$i$i = (0 - ($87))|0;
      $88 = $ispos$i$i$i ? $87 : $neg$i$i$i;
      $89 = ($88|0)>(131071);
      if ($89) {
       $90 = HEAP32[$72>>2]|0;
       _luaX_syntaxerror($90,519652);
       $$pre$i$i = HEAP32[$$lcssa113>>2]|0;
       $92 = $$pre$i$i;
      } else {
       $92 = $$lcssa114;
      }
      $91 = $92 & 16383;
      $93 = $87 << 14;
      $94 = (($93) + 2147467264)|0;
      $95 = $91 | $94;
      HEAP32[$$lcssa113>>2] = $95;
      $$pre$phi82Z2D = $72;$$pre$phi84Z2D = $70;$113 = $76;
     }
    }
   }
   $96 = ((($fs)) + 24|0);
   $97 = HEAP32[$96>>2]|0;
   $98 = ((($fs)) + 28|0);
   HEAP32[$98>>2] = $97;
   $99 = $reg << 6;
   $100 = $99 | 16386;
   $101 = HEAP32[$$pre$phi82Z2D>>2]|0;
   $102 = ((($101)) + 8|0);
   $103 = HEAP32[$102>>2]|0;
   $104 = (_luaK_code($fs,$100,$103)|0);
   $105 = HEAP32[$96>>2]|0;
   HEAP32[$98>>2] = $105;
   $106 = $99 | 8388610;
   $107 = HEAP32[$$pre$phi82Z2D>>2]|0;
   $108 = ((($107)) + 8|0);
   $109 = HEAP32[$108>>2]|0;
   $110 = (_luaK_code($fs,$106,$109)|0);
   $111 = HEAP32[$96>>2]|0;
   HEAP32[$98>>2] = $111;
   $112 = ($113|0)==(-1);
   if ($112) {
    $p_f$0 = $104;$p_t$0 = $110;
   } else {
    $114 = HEAP32[$$pre$phi84Z2D>>2]|0;
    $115 = ($114|0)==(-1);
    if ($115) {
     HEAP32[$$pre$phi84Z2D>>2] = $113;
     $p_f$0 = $104;$p_t$0 = $110;
     break;
    }
    $fs$idx$val$i$i28 = HEAP32[$fs>>2]|0;
    $fs$idx$val$idx$i$i29 = ((($fs$idx$val$i$i28)) + 12|0);
    $fs$idx$val$idx$val$i$i30 = HEAP32[$fs$idx$val$idx$i$i29>>2]|0;
    $list$0$i$i32 = $114;
    while(1) {
     $116 = (($fs$idx$val$idx$val$i$i30) + ($list$0$i$i32<<2)|0);
     $117 = HEAP32[$116>>2]|0;
     $118 = $117 >>> 14;
     $119 = (($118) + -131071)|0;
     $120 = ($119|0)==(-1);
     $121 = (($list$0$i$i32) + 1)|0;
     $122 = (($121) + ($119))|0;
     $$0$i$i$i33 = $120 ? -1 : $122;
     $123 = ($$0$i$i$i33|0)==(-1);
     if ($123) {
      $$lcssa = $116;$$lcssa112 = $117;$list$0$i$i32$lcssa = $list$0$i$i32;
      break;
     } else {
      $list$0$i$i32 = $$0$i$i$i33;
     }
    }
    $$neg$i$i$i34 = $list$0$i$i32$lcssa ^ -1;
    $124 = (($113) + ($$neg$i$i$i34))|0;
    $ispos$i$i$i35 = ($124|0)>(-1);
    $neg$i$i$i36 = (0 - ($124))|0;
    $125 = $ispos$i$i$i35 ? $124 : $neg$i$i$i36;
    $126 = ($125|0)>(131071);
    if ($126) {
     $127 = HEAP32[$$pre$phi82Z2D>>2]|0;
     _luaX_syntaxerror($127,519652);
     $$pre$i$i37 = HEAP32[$$lcssa>>2]|0;
     $129 = $$pre$i$i37;
    } else {
     $129 = $$lcssa112;
    }
    $128 = $129 & 16383;
    $130 = $124 << 14;
    $131 = (($130) + 2147467264)|0;
    $132 = $128 | $131;
    HEAP32[$$lcssa>>2] = $132;
    $p_f$0 = $104;$p_t$0 = $110;
   }
  }
 } while(0);
 $133 = ((($fs)) + 24|0);
 $134 = HEAP32[$133>>2]|0;
 $135 = ((($fs)) + 28|0);
 HEAP32[$135>>2] = $134;
 $136 = HEAP32[$27>>2]|0;
 $137 = ($136|0)==(-1);
 L63: do {
  if (!($137)) {
   $138 = ((($fs)) + 12|0);
   $139 = ($reg|0)==(255);
   $140 = $reg << 6;
   $141 = $140 & 16320;
   if ($139) {
    $$02$i41$us = $136;
    while(1) {
     $fs$idx$val$i42$us = HEAP32[$fs>>2]|0;
     $fs$idx$val$idx$i43$us = ((($fs$idx$val$i42$us)) + 12|0);
     $fs$idx$val$idx$val$i44$us = HEAP32[$fs$idx$val$idx$i43$us>>2]|0;
     $142 = (($fs$idx$val$idx$val$i44$us) + ($$02$i41$us<<2)|0);
     $143 = HEAP32[$142>>2]|0;
     $144 = $143 >>> 14;
     $145 = (($144) + -131071)|0;
     $146 = ($145|0)==(-1);
     $147 = (($$02$i41$us) + 1)|0;
     $148 = (($147) + ($145))|0;
     $$0$i$i45$us = $146 ? -1 : $148;
     $149 = ($$02$i41$us|0)>(0);
     if ($149) {
      $150 = ((($142)) + -4|0);
      $151 = HEAP32[$150>>2]|0;
      $152 = $151 & 63;
      $153 = (518247 + ($152)|0);
      $154 = HEAP8[$153>>0]|0;
      $155 = ($154<<24>>24)<(0);
      if ($155) {
       $$0$i$i57$us = $150;$157 = $151;
      } else {
       label = 46;
      }
     } else {
      label = 46;
     }
     if ((label|0) == 46) {
      label = 0;
      $$0$i$i57$us = $142;$157 = $143;
     }
     $156 = $157 & 63;
     $158 = ($156|0)==(27);
     if ($158) {
      $168 = $157 >>> 23;
      $169 = $168 << 6;
      $170 = $157 & 8372224;
      $171 = $170 | $169;
      $172 = $171 | 26;
      HEAP32[$$0$i$i57$us>>2] = $172;
      $$neg$i52$us = $$02$i41$us ^ -1;
      $173 = (($134) + ($$neg$i52$us))|0;
      $ispos$i53$us = ($173|0)>(-1);
      $neg$i54$us = (0 - ($173))|0;
      $174 = $ispos$i53$us ? $173 : $neg$i54$us;
      $175 = ($174|0)>(131071);
      if ($175) {
       $176 = HEAP32[$138>>2]|0;
       _luaX_syntaxerror($176,519652);
      }
      $177 = HEAP32[$142>>2]|0;
      $178 = $177 & 16383;
      $179 = $173 << 14;
      $180 = (($179) + 2147467264)|0;
      $181 = $178 | $180;
      $storemerge87 = $181;
     } else {
      $$neg$i$i46$us = $$02$i41$us ^ -1;
      $159 = (($p_f$0) + ($$neg$i$i46$us))|0;
      $ispos$i$i47$us = ($159|0)>(-1);
      $neg$i$i48$us = (0 - ($159))|0;
      $160 = $ispos$i$i47$us ? $159 : $neg$i$i48$us;
      $161 = ($160|0)>(131071);
      if ($161) {
       $162 = HEAP32[$138>>2]|0;
       _luaX_syntaxerror($162,519652);
       $$pre77 = HEAP32[$142>>2]|0;
       $164 = $$pre77;
      } else {
       $164 = $143;
      }
      $163 = $164 & 16383;
      $165 = $159 << 14;
      $166 = (($165) + 2147467264)|0;
      $167 = $163 | $166;
      $storemerge87 = $167;
     }
     HEAP32[$142>>2] = $storemerge87;
     $182 = ($$0$i$i45$us|0)==(-1);
     if ($182) {
      break L63;
     } else {
      $$02$i41$us = $$0$i$i45$us;
     }
    }
   } else {
    $$02$i41 = $136;
   }
   while(1) {
    $fs$idx$val$i42 = HEAP32[$fs>>2]|0;
    $fs$idx$val$idx$i43 = ((($fs$idx$val$i42)) + 12|0);
    $fs$idx$val$idx$val$i44 = HEAP32[$fs$idx$val$idx$i43>>2]|0;
    $183 = (($fs$idx$val$idx$val$i44) + ($$02$i41<<2)|0);
    $184 = HEAP32[$183>>2]|0;
    $185 = $184 >>> 14;
    $186 = (($185) + -131071)|0;
    $187 = ($186|0)==(-1);
    $188 = (($$02$i41) + 1)|0;
    $189 = (($188) + ($186))|0;
    $$0$i$i45 = $187 ? -1 : $189;
    $190 = ($$02$i41|0)>(0);
    if ($190) {
     $191 = ((($183)) + -4|0);
     $192 = HEAP32[$191>>2]|0;
     $193 = $192 & 63;
     $194 = (518247 + ($193)|0);
     $195 = HEAP8[$194>>0]|0;
     $196 = ($195<<24>>24)<(0);
     if ($196) {
      $$0$i$i57 = $191;$198 = $192;
     } else {
      label = 57;
     }
    } else {
     label = 57;
    }
    if ((label|0) == 57) {
     label = 0;
     $$0$i$i57 = $183;$198 = $184;
    }
    $197 = $198 & 63;
    $199 = ($197|0)==(27);
    if ($199) {
     $200 = $198 >>> 23;
     $201 = ($200|0)==($reg|0);
     if ($201) {
      $204 = $198 & 8372224;
      $205 = $204 | $140;
      $206 = $205 | 26;
      $storemerge$i60 = $206;
     } else {
      $202 = $198 & -16321;
      $203 = $202 | $141;
      $storemerge$i60 = $203;
     }
     HEAP32[$$0$i$i57>>2] = $storemerge$i60;
     $$neg$i52 = $$02$i41 ^ -1;
     $207 = (($134) + ($$neg$i52))|0;
     $ispos$i53 = ($207|0)>(-1);
     $neg$i54 = (0 - ($207))|0;
     $208 = $ispos$i53 ? $207 : $neg$i54;
     $209 = ($208|0)>(131071);
     if ($209) {
      $210 = HEAP32[$138>>2]|0;
      _luaX_syntaxerror($210,519652);
     }
     $211 = HEAP32[$183>>2]|0;
     $212 = $211 & 16383;
     $213 = $207 << 14;
     $214 = (($213) + 2147467264)|0;
     $215 = $212 | $214;
     $storemerge = $215;
    } else {
     $$neg$i$i46 = $$02$i41 ^ -1;
     $216 = (($p_f$0) + ($$neg$i$i46))|0;
     $ispos$i$i47 = ($216|0)>(-1);
     $neg$i$i48 = (0 - ($216))|0;
     $217 = $ispos$i$i47 ? $216 : $neg$i$i48;
     $218 = ($217|0)>(131071);
     if ($218) {
      $219 = HEAP32[$138>>2]|0;
      _luaX_syntaxerror($219,519652);
      $$pre = HEAP32[$183>>2]|0;
      $221 = $$pre;
     } else {
      $221 = $184;
     }
     $220 = $221 & 16383;
     $222 = $216 << 14;
     $223 = (($222) + 2147467264)|0;
     $224 = $220 | $223;
     $storemerge = $224;
    }
    HEAP32[$183>>2] = $storemerge;
    $225 = ($$0$i$i45|0)==(-1);
    if ($225) {
     break;
    } else {
     $$02$i41 = $$0$i$i45;
    }
   }
  }
 } while(0);
 $226 = HEAP32[$2>>2]|0;
 $227 = ($226|0)==(-1);
 if ($227) {
  HEAP32[$2>>2] = -1;
  HEAP32[$27>>2] = -1;
  $316 = ((($e)) + 8|0);
  HEAP32[$316>>2] = $reg;
  HEAP32[$e>>2] = 12;
  return;
 }
 $228 = ((($fs)) + 12|0);
 $229 = ($reg|0)==(255);
 $230 = $reg << 6;
 $231 = $230 & 16320;
 if ($229) {
  $$02$i$us = $226;
  while(1) {
   $fs$idx$val$i4$us = HEAP32[$fs>>2]|0;
   $fs$idx$val$idx$i5$us = ((($fs$idx$val$i4$us)) + 12|0);
   $fs$idx$val$idx$val$i6$us = HEAP32[$fs$idx$val$idx$i5$us>>2]|0;
   $232 = (($fs$idx$val$idx$val$i6$us) + ($$02$i$us<<2)|0);
   $233 = HEAP32[$232>>2]|0;
   $234 = $233 >>> 14;
   $235 = (($234) + -131071)|0;
   $236 = ($235|0)==(-1);
   $237 = (($$02$i$us) + 1)|0;
   $238 = (($237) + ($235))|0;
   $$0$i$i7$us = $236 ? -1 : $238;
   $239 = ($$02$i$us|0)>(0);
   if ($239) {
    $240 = ((($232)) + -4|0);
    $241 = HEAP32[$240>>2]|0;
    $242 = $241 & 63;
    $243 = (518247 + ($242)|0);
    $244 = HEAP8[$243>>0]|0;
    $245 = ($244<<24>>24)<(0);
    if ($245) {
     $$0$i$i25$us = $240;$247 = $241;
    } else {
     label = 73;
    }
   } else {
    label = 73;
   }
   if ((label|0) == 73) {
    label = 0;
    $$0$i$i25$us = $232;$247 = $233;
   }
   $246 = $247 & 63;
   $248 = ($246|0)==(27);
   if ($248) {
    $258 = $247 >>> 23;
    $259 = $258 << 6;
    $260 = $247 & 8372224;
    $261 = $260 | $259;
    $262 = $261 | 26;
    HEAP32[$$0$i$i25$us>>2] = $262;
    $$neg$i$us = $$02$i$us ^ -1;
    $263 = (($134) + ($$neg$i$us))|0;
    $ispos$i$us = ($263|0)>(-1);
    $neg$i$us = (0 - ($263))|0;
    $264 = $ispos$i$us ? $263 : $neg$i$us;
    $265 = ($264|0)>(131071);
    if ($265) {
     $266 = HEAP32[$228>>2]|0;
     _luaX_syntaxerror($266,519652);
    }
    $267 = HEAP32[$232>>2]|0;
    $268 = $267 & 16383;
    $269 = $263 << 14;
    $270 = (($269) + 2147467264)|0;
    $271 = $268 | $270;
    $storemerge86 = $271;
   } else {
    $$neg$i$i8$us = $$02$i$us ^ -1;
    $249 = (($p_t$0) + ($$neg$i$i8$us))|0;
    $ispos$i$i9$us = ($249|0)>(-1);
    $neg$i$i10$us = (0 - ($249))|0;
    $250 = $ispos$i$i9$us ? $249 : $neg$i$i10$us;
    $251 = ($250|0)>(131071);
    if ($251) {
     $252 = HEAP32[$228>>2]|0;
     _luaX_syntaxerror($252,519652);
     $$pre79 = HEAP32[$232>>2]|0;
     $254 = $$pre79;
    } else {
     $254 = $233;
    }
    $253 = $254 & 16383;
    $255 = $249 << 14;
    $256 = (($255) + 2147467264)|0;
    $257 = $253 | $256;
    $storemerge86 = $257;
   }
   HEAP32[$232>>2] = $storemerge86;
   $272 = ($$0$i$i7$us|0)==(-1);
   if ($272) {
    break;
   } else {
    $$02$i$us = $$0$i$i7$us;
   }
  }
  HEAP32[$2>>2] = -1;
  HEAP32[$27>>2] = -1;
  $316 = ((($e)) + 8|0);
  HEAP32[$316>>2] = $reg;
  HEAP32[$e>>2] = 12;
  return;
 } else {
  $$02$i = $226;
 }
 while(1) {
  $fs$idx$val$i4 = HEAP32[$fs>>2]|0;
  $fs$idx$val$idx$i5 = ((($fs$idx$val$i4)) + 12|0);
  $fs$idx$val$idx$val$i6 = HEAP32[$fs$idx$val$idx$i5>>2]|0;
  $273 = (($fs$idx$val$idx$val$i6) + ($$02$i<<2)|0);
  $274 = HEAP32[$273>>2]|0;
  $275 = $274 >>> 14;
  $276 = (($275) + -131071)|0;
  $277 = ($276|0)==(-1);
  $278 = (($$02$i) + 1)|0;
  $279 = (($278) + ($276))|0;
  $$0$i$i7 = $277 ? -1 : $279;
  $280 = ($$02$i|0)>(0);
  if ($280) {
   $281 = ((($273)) + -4|0);
   $282 = HEAP32[$281>>2]|0;
   $283 = $282 & 63;
   $284 = (518247 + ($283)|0);
   $285 = HEAP8[$284>>0]|0;
   $286 = ($285<<24>>24)<(0);
   if ($286) {
    $$0$i$i25 = $281;$288 = $282;
   } else {
    label = 84;
   }
  } else {
   label = 84;
  }
  if ((label|0) == 84) {
   label = 0;
   $$0$i$i25 = $273;$288 = $274;
  }
  $287 = $288 & 63;
  $289 = ($287|0)==(27);
  if ($289) {
   $290 = $288 >>> 23;
   $291 = ($290|0)==($reg|0);
   if ($291) {
    $294 = $288 & 8372224;
    $295 = $294 | $230;
    $296 = $295 | 26;
    $storemerge$i = $296;
   } else {
    $292 = $288 & -16321;
    $293 = $292 | $231;
    $storemerge$i = $293;
   }
   HEAP32[$$0$i$i25>>2] = $storemerge$i;
   $$neg$i = $$02$i ^ -1;
   $297 = (($134) + ($$neg$i))|0;
   $ispos$i = ($297|0)>(-1);
   $neg$i = (0 - ($297))|0;
   $298 = $ispos$i ? $297 : $neg$i;
   $299 = ($298|0)>(131071);
   if ($299) {
    $300 = HEAP32[$228>>2]|0;
    _luaX_syntaxerror($300,519652);
   }
   $301 = HEAP32[$273>>2]|0;
   $302 = $301 & 16383;
   $303 = $297 << 14;
   $304 = (($303) + 2147467264)|0;
   $305 = $302 | $304;
   $storemerge85 = $305;
  } else {
   $$neg$i$i8 = $$02$i ^ -1;
   $306 = (($p_t$0) + ($$neg$i$i8))|0;
   $ispos$i$i9 = ($306|0)>(-1);
   $neg$i$i10 = (0 - ($306))|0;
   $307 = $ispos$i$i9 ? $306 : $neg$i$i10;
   $308 = ($307|0)>(131071);
   if ($308) {
    $309 = HEAP32[$228>>2]|0;
    _luaX_syntaxerror($309,519652);
    $$pre78 = HEAP32[$273>>2]|0;
    $311 = $$pre78;
   } else {
    $311 = $274;
   }
   $310 = $311 & 16383;
   $312 = $306 << 14;
   $313 = (($312) + 2147467264)|0;
   $314 = $310 | $313;
   $storemerge85 = $314;
  }
  HEAP32[$273>>2] = $storemerge85;
  $315 = ($$0$i$i7|0)==(-1);
  if ($315) {
   break;
  } else {
   $$02$i = $$0$i$i7;
  }
 }
 HEAP32[$2>>2] = -1;
 HEAP32[$27>>2] = -1;
 $316 = ((($e)) + 8|0);
 HEAP32[$316>>2] = $reg;
 HEAP32[$e>>2] = 12;
 return;
}
function _discharge2reg($fs,$e,$reg) {
 $fs = $fs|0;
 $e = $e|0;
 $reg = $reg|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $o$i = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $o$i = sp;
 _luaK_dischargevars($fs,$e);
 $0 = HEAP32[$e>>2]|0;
 L1: do {
  switch ($0|0) {
  case 1:  {
   $1 = ((($fs)) + 24|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = ((($fs)) + 28|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($2|0)>($4|0);
   do {
    if ($5) {
     $6 = ($2|0)==(0);
     if ($6) {
      $7 = ((($fs)) + 50|0);
      $8 = HEAP8[$7>>0]|0;
      $9 = $8&255;
      $10 = ($9|0)>($reg|0);
      if ($10) {
       break;
      } else {
       break L1;
      }
     }
     $11 = (($2) + -1)|0;
     $12 = HEAP32[$fs>>2]|0;
     $13 = ((($12)) + 12|0);
     $14 = HEAP32[$13>>2]|0;
     $15 = (($14) + ($11<<2)|0);
     $16 = HEAP32[$15>>2]|0;
     $17 = $16 & 63;
     $18 = ($17|0)==(3);
     if ($18) {
      $19 = $16 >>> 6;
      $20 = $19 & 255;
      $21 = $16 >>> 23;
      $22 = ($20|0)>($reg|0);
      $23 = (($21) + 1)|0;
      $24 = ($23|0)<($reg|0);
      $or$cond$i = $22 | $24;
      if (!($or$cond$i)) {
       $25 = ($21|0)<($reg|0);
       if (!($25)) {
        break L1;
       }
       $26 = $16 & 8388607;
       $27 = $reg << 23;
       $28 = $26 | $27;
       HEAP32[$15>>2] = $28;
       break L1;
      }
     }
    }
   } while(0);
   $29 = $reg << 6;
   $30 = $reg << 23;
   $31 = $30 | $29;
   $32 = $31 | 3;
   $33 = ((($fs)) + 12|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = ((($34)) + 8|0);
   $36 = HEAP32[$35>>2]|0;
   (_luaK_code($fs,$32,$36)|0);
   break;
  }
  case 2: case 3:  {
   $37 = ($0|0)==(2);
   $38 = $37&1;
   $39 = $reg << 6;
   $40 = $38 << 23;
   $41 = $39 | $40;
   $42 = $41 | 2;
   $43 = ((($fs)) + 12|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = ((($44)) + 8|0);
   $46 = HEAP32[$45>>2]|0;
   (_luaK_code($fs,$42,$46)|0);
   break;
  }
  case 4:  {
   $47 = ((($e)) + 8|0);
   $48 = HEAP32[$47>>2]|0;
   $49 = $reg << 6;
   $50 = $48 << 14;
   $51 = $49 | $50;
   $52 = $51 | 1;
   $53 = ((($fs)) + 12|0);
   $54 = HEAP32[$53>>2]|0;
   $55 = ((($54)) + 8|0);
   $56 = HEAP32[$55>>2]|0;
   (_luaK_code($fs,$52,$56)|0);
   break;
  }
  case 5:  {
   $57 = ((($e)) + 8|0);
   $58 = +HEAPF64[$57>>3];
   HEAPF64[$o$i>>3] = $58;
   $59 = ((($o$i)) + 8|0);
   HEAP32[$59>>2] = 3;
   $60 = (_addk($fs,$o$i,$o$i)|0);
   $61 = $reg << 6;
   $62 = $60 << 14;
   $63 = $61 | $62;
   $64 = $63 | 1;
   $65 = ((($fs)) + 12|0);
   $66 = HEAP32[$65>>2]|0;
   $67 = ((($66)) + 8|0);
   $68 = HEAP32[$67>>2]|0;
   (_luaK_code($fs,$64,$68)|0);
   break;
  }
  case 11:  {
   $69 = ((($e)) + 8|0);
   $70 = HEAP32[$69>>2]|0;
   $71 = HEAP32[$fs>>2]|0;
   $72 = ((($71)) + 12|0);
   $73 = HEAP32[$72>>2]|0;
   $74 = (($73) + ($70<<2)|0);
   $75 = HEAP32[$74>>2]|0;
   $76 = $75 & -16321;
   $77 = $reg << 6;
   $78 = $77 & 16320;
   $79 = $76 | $78;
   HEAP32[$74>>2] = $79;
   break;
  }
  case 12:  {
   $80 = ((($e)) + 8|0);
   $81 = HEAP32[$80>>2]|0;
   $82 = ($81|0)==($reg|0);
   if (!($82)) {
    $83 = $reg << 6;
    $84 = $81 << 23;
    $85 = $84 | $83;
    $86 = ((($fs)) + 12|0);
    $87 = HEAP32[$86>>2]|0;
    $88 = ((($87)) + 8|0);
    $89 = HEAP32[$88>>2]|0;
    (_luaK_code($fs,$85,$89)|0);
   }
   break;
  }
  default: {
   STACKTOP = sp;return;
  }
  }
 } while(0);
 $90 = ((($e)) + 8|0);
 HEAP32[$90>>2] = $reg;
 HEAP32[$e>>2] = 12;
 STACKTOP = sp;return;
}
function _luaK_exp2anyreg($fs,$e) {
 $fs = $fs|0;
 $e = $e|0;
 var $$0 = 0, $$pre$i$i = 0, $$pre$i$i$i = 0, $$pre$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaK_dischargevars($fs,$e);
 $0 = HEAP32[$e>>2]|0;
 $1 = ($0|0)==(12);
 if ($1) {
  $2 = ((($e)) + 16|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($e)) + 20|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ($3|0)==($5|0);
  $7 = ((($e)) + 8|0);
  $8 = HEAP32[$7>>2]|0;
  if ($6) {
   $$0 = $8;
   return ($$0|0);
  }
  $9 = ((($fs)) + 50|0);
  $10 = HEAP8[$9>>0]|0;
  $11 = $10&255;
  $12 = ($8|0)<($11|0);
  if (!($12)) {
   _exp2reg($fs,$e,$8);
   $13 = HEAP32[$7>>2]|0;
   $$0 = $13;
   return ($$0|0);
  }
 }
 _luaK_dischargevars($fs,$e);
 $14 = HEAP32[$e>>2]|0;
 $15 = ($14|0)==(12);
 if ($15) {
  $16 = ((($e)) + 8|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = $17 & 256;
  $19 = ($18|0)==(0);
  if ($19) {
   $20 = ((($fs)) + 50|0);
   $21 = HEAP8[$20>>0]|0;
   $22 = $21&255;
   $23 = ($22|0)>($17|0);
   if (!($23)) {
    $24 = ((($fs)) + 36|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = (($25) + -1)|0;
    HEAP32[$24>>2] = $26;
   }
  }
 }
 $27 = ((($fs)) + 36|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = (($28) + 1)|0;
 $30 = HEAP32[$fs>>2]|0;
 $31 = ((($30)) + 75|0);
 $32 = HEAP8[$31>>0]|0;
 $33 = $32&255;
 $34 = ($28|0)<($33|0);
 if ($34) {
  $42 = $28;
 } else {
  $35 = ($28|0)>(248);
  if ($35) {
   $36 = ((($fs)) + 12|0);
   $37 = HEAP32[$36>>2]|0;
   _luaX_syntaxerror($37,519698);
   $$pre$i$i$i = HEAP32[$fs>>2]|0;
   $$pre$pre$i$i = HEAP32[$27>>2]|0;
   $$pre$i$i = $$pre$pre$i$i;$40 = $$pre$i$i$i;
  } else {
   $$pre$i$i = $28;$40 = $30;
  }
  $38 = $29&255;
  $39 = ((($40)) + 75|0);
  HEAP8[$39>>0] = $38;
  $42 = $$pre$i$i;
 }
 $41 = (($42) + 1)|0;
 HEAP32[$27>>2] = $41;
 _exp2reg($fs,$e,$42);
 $43 = ((($e)) + 8|0);
 $44 = HEAP32[$43>>2]|0;
 $$0 = $44;
 return ($$0|0);
}
function _luaK_exp2val($fs,$e) {
 $fs = $fs|0;
 $e = $e|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($e)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($e)) + 20|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==($3|0);
 if ($4) {
  _luaK_dischargevars($fs,$e);
  return;
 } else {
  (_luaK_exp2anyreg($fs,$e)|0);
  return;
 }
}
function _luaK_exp2RK($fs,$e) {
 $fs = $fs|0;
 $e = $e|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $k$i = 0, $o$i = 0, $o$i1 = 0, $v$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $o$i1 = sp + 48|0;
 $o$i = sp + 32|0;
 $k$i = sp + 16|0;
 $v$i = sp;
 $0 = ((($e)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($e)) + 20|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==($3|0);
 if ($4) {
  _luaK_dischargevars($fs,$e);
 } else {
  (_luaK_exp2anyreg($fs,$e)|0);
 }
 $5 = HEAP32[$e>>2]|0;
 switch ($5|0) {
 case 1: case 3: case 2: case 5:  {
  $6 = ((($fs)) + 40|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)<(256);
  if ($8) {
   switch ($5|0) {
   case 1:  {
    $9 = ((($v$i)) + 8|0);
    HEAP32[$9>>2] = 0;
    $10 = ((($fs)) + 4|0);
    $11 = HEAP32[$10>>2]|0;
    HEAP32[$k$i>>2] = $11;
    $12 = ((($k$i)) + 8|0);
    HEAP32[$12>>2] = 5;
    $13 = (_addk($fs,$k$i,$v$i)|0);
    $23 = $13;
    break;
   }
   case 5:  {
    $14 = ((($e)) + 8|0);
    $15 = +HEAPF64[$14>>3];
    HEAPF64[$o$i>>3] = $15;
    $16 = ((($o$i)) + 8|0);
    HEAP32[$16>>2] = 3;
    $17 = (_addk($fs,$o$i,$o$i)|0);
    $23 = $17;
    break;
   }
   default: {
    $18 = ($5|0)==(2);
    $19 = $18&1;
    HEAP32[$o$i1>>2] = $19;
    $20 = ((($o$i1)) + 8|0);
    HEAP32[$20>>2] = 1;
    $21 = (_addk($fs,$o$i1,$o$i1)|0);
    $23 = $21;
   }
   }
   $22 = ((($e)) + 8|0);
   HEAP32[$22>>2] = $23;
   HEAP32[$e>>2] = 4;
   $24 = $23 | 256;
   $$0 = $24;
   STACKTOP = sp;return ($$0|0);
  }
  break;
 }
 case 4:  {
  $25 = ((($e)) + 8|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = ($26|0)<(256);
  if ($27) {
   $28 = $26 | 256;
   $$0 = $28;
   STACKTOP = sp;return ($$0|0);
  }
  break;
 }
 default: {
 }
 }
 $29 = (_luaK_exp2anyreg($fs,$e)|0);
 $$0 = $29;
 STACKTOP = sp;return ($$0|0);
}
function _luaK_storevar($fs,$var,$ex) {
 $fs = $fs|0;
 $var = $var|0;
 $ex = $ex|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$var>>2]|0;
 switch ($0|0) {
 case 6:  {
  $1 = HEAP32[$ex>>2]|0;
  $2 = ($1|0)==(12);
  if ($2) {
   $3 = ((($ex)) + 8|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = $4 & 256;
   $6 = ($5|0)==(0);
   if ($6) {
    $7 = ((($fs)) + 50|0);
    $8 = HEAP8[$7>>0]|0;
    $9 = $8&255;
    $10 = ($9|0)>($4|0);
    if (!($10)) {
     $11 = ((($fs)) + 36|0);
     $12 = HEAP32[$11>>2]|0;
     $13 = (($12) + -1)|0;
     HEAP32[$11>>2] = $13;
    }
   }
  }
  $14 = ((($var)) + 8|0);
  $15 = HEAP32[$14>>2]|0;
  _exp2reg($fs,$ex,$15);
  return;
  break;
 }
 case 7:  {
  $16 = (_luaK_exp2anyreg($fs,$ex)|0);
  $17 = ((($var)) + 8|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = $16 << 6;
  $20 = $18 << 23;
  $21 = $19 | $20;
  $22 = $21 | 8;
  $23 = ((($fs)) + 12|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = ((($24)) + 8|0);
  $26 = HEAP32[$25>>2]|0;
  (_luaK_code($fs,$22,$26)|0);
  break;
 }
 case 8:  {
  $27 = (_luaK_exp2anyreg($fs,$ex)|0);
  $28 = ((($var)) + 8|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = $27 << 6;
  $31 = $29 << 14;
  $32 = $30 | $31;
  $33 = $32 | 7;
  $34 = ((($fs)) + 12|0);
  $35 = HEAP32[$34>>2]|0;
  $36 = ((($35)) + 8|0);
  $37 = HEAP32[$36>>2]|0;
  (_luaK_code($fs,$33,$37)|0);
  break;
 }
 case 9:  {
  $38 = (_luaK_exp2RK($fs,$ex)|0);
  $39 = ((($var)) + 8|0);
  $40 = HEAP32[$39>>2]|0;
  $41 = ((($39)) + 4|0);
  $42 = HEAP32[$41>>2]|0;
  $43 = $40 << 6;
  $44 = $42 << 23;
  $45 = $38 << 14;
  $46 = $45 | $43;
  $47 = $46 | $44;
  $48 = $47 | 9;
  $49 = ((($fs)) + 12|0);
  $50 = HEAP32[$49>>2]|0;
  $51 = ((($50)) + 8|0);
  $52 = HEAP32[$51>>2]|0;
  (_luaK_code($fs,$48,$52)|0);
  break;
 }
 default: {
 }
 }
 $53 = HEAP32[$ex>>2]|0;
 $54 = ($53|0)==(12);
 if (!($54)) {
  return;
 }
 $55 = ((($ex)) + 8|0);
 $56 = HEAP32[$55>>2]|0;
 $57 = $56 & 256;
 $58 = ($57|0)==(0);
 if (!($58)) {
  return;
 }
 $59 = ((($fs)) + 50|0);
 $60 = HEAP8[$59>>0]|0;
 $61 = $60&255;
 $62 = ($61|0)>($56|0);
 if ($62) {
  return;
 }
 $63 = ((($fs)) + 36|0);
 $64 = HEAP32[$63>>2]|0;
 $65 = (($64) + -1)|0;
 HEAP32[$63>>2] = $65;
 return;
}
function _luaK_self($fs,$e,$key) {
 $fs = $fs|0;
 $e = $e|0;
 $key = $key|0;
 var $$pre$i = 0, $$pre$i$i = 0, $$pre$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 (_luaK_exp2anyreg($fs,$e)|0);
 $0 = HEAP32[$e>>2]|0;
 $1 = ($0|0)==(12);
 if ($1) {
  $2 = ((($e)) + 8|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = $3 & 256;
  $5 = ($4|0)==(0);
  if ($5) {
   $6 = ((($fs)) + 50|0);
   $7 = HEAP8[$6>>0]|0;
   $8 = $7&255;
   $9 = ($8|0)>($3|0);
   if (!($9)) {
    $10 = ((($fs)) + 36|0);
    $11 = HEAP32[$10>>2]|0;
    $12 = (($11) + -1)|0;
    HEAP32[$10>>2] = $12;
   }
  }
 }
 $13 = ((($fs)) + 36|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($14) + 2)|0;
 $16 = HEAP32[$fs>>2]|0;
 $17 = ((($16)) + 75|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = $18&255;
 $20 = ($15|0)>($19|0);
 if ($20) {
  $21 = ($15|0)>(249);
  if ($21) {
   $22 = ((($fs)) + 12|0);
   $23 = HEAP32[$22>>2]|0;
   _luaX_syntaxerror($23,519698);
   $$pre$i$i = HEAP32[$fs>>2]|0;
   $$pre$pre$i = HEAP32[$13>>2]|0;
   $$pre$i = $$pre$pre$i;$26 = $$pre$i$i;
  } else {
   $$pre$i = $14;$26 = $16;
  }
  $24 = $15&255;
  $25 = ((($26)) + 75|0);
  HEAP8[$25>>0] = $24;
  $28 = $$pre$i;
 } else {
  $28 = $14;
 }
 $27 = (($28) + 2)|0;
 HEAP32[$13>>2] = $27;
 $29 = ((($e)) + 8|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = (_luaK_exp2RK($fs,$key)|0);
 $32 = $14 << 6;
 $33 = $30 << 23;
 $34 = $31 << 14;
 $35 = $32 | $33;
 $36 = $35 | $34;
 $37 = $36 | 11;
 $38 = ((($fs)) + 12|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ((($39)) + 8|0);
 $41 = HEAP32[$40>>2]|0;
 (_luaK_code($fs,$37,$41)|0);
 $42 = HEAP32[$key>>2]|0;
 $43 = ($42|0)==(12);
 if (!($43)) {
  HEAP32[$29>>2] = $14;
  HEAP32[$e>>2] = 12;
  return;
 }
 $44 = ((($key)) + 8|0);
 $45 = HEAP32[$44>>2]|0;
 $46 = $45 & 256;
 $47 = ($46|0)==(0);
 if (!($47)) {
  HEAP32[$29>>2] = $14;
  HEAP32[$e>>2] = 12;
  return;
 }
 $48 = ((($fs)) + 50|0);
 $49 = HEAP8[$48>>0]|0;
 $50 = $49&255;
 $51 = ($50|0)>($45|0);
 if ($51) {
  HEAP32[$29>>2] = $14;
  HEAP32[$e>>2] = 12;
  return;
 }
 $52 = HEAP32[$13>>2]|0;
 $53 = (($52) + -1)|0;
 HEAP32[$13>>2] = $53;
 HEAP32[$29>>2] = $14;
 HEAP32[$e>>2] = 12;
 return;
}
function _luaK_goiftrue($fs,$e) {
 $fs = $fs|0;
 $e = $e|0;
 var $$0$i$i = 0, $$0$i$i$i = 0, $$0$i$i1 = 0, $$lcssa = 0, $$lcssa19 = 0, $$lcssa20 = 0, $$lcssa21 = 0, $$neg$i$i = 0, $$neg$i$i$i = 0, $$pr = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i$i = 0, $$pre$i$i3 = 0, $$pre$i2 = 0, $$pre$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0;
 var $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fs$idx$val = 0, $fs$idx$val$i = 0, $fs$idx$val$i$i = 0, $fs$idx$val$idx = 0, $fs$idx$val$idx$i = 0, $fs$idx$val$idx$i$i = 0, $fs$idx$val$idx$val = 0, $fs$idx$val$idx$val$i = 0, $fs$idx$val$idx$val$i$i = 0, $ispos$i$i = 0, $ispos$i$i$i = 0, $list$0$i = 0;
 var $list$0$i$i = 0, $list$0$i$i$lcssa = 0, $list$0$i$lcssa = 0, $neg$i$i = 0, $neg$i$i$i = 0, $pc$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaK_dischargevars($fs,$e);
 $0 = HEAP32[$e>>2]|0;
 L1: do {
  switch ($0|0) {
  case 2: case 5: case 4:  {
   break;
  }
  case 10:  {
   $fs$idx$val = HEAP32[$fs>>2]|0;
   $fs$idx$val$idx = ((($fs$idx$val)) + 12|0);
   $fs$idx$val$idx$val = HEAP32[$fs$idx$val$idx>>2]|0;
   $1 = ((($e)) + 8|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (($fs$idx$val$idx$val) + ($2<<2)|0);
   $4 = ($2|0)>(0);
   if ($4) {
    $5 = ((($3)) + -4|0);
    $6 = HEAP32[$5>>2]|0;
    $7 = $6 & 63;
    $8 = (518247 + ($7)|0);
    $9 = HEAP8[$8>>0]|0;
    $10 = ($9<<24>>24)<(0);
    if ($10) {
     $$0$i$i = $5;$12 = $6;
    } else {
     label = 4;
    }
   } else {
    label = 4;
   }
   if ((label|0) == 4) {
    $$pre$i = HEAP32[$3>>2]|0;
    $$0$i$i = $3;$12 = $$pre$i;
   }
   $11 = $12 & -16321;
   $13 = $12 & 16320;
   $14 = ($13|0)==(0);
   $15 = $14&1;
   $16 = $15 << 6;
   $17 = $16 | $11;
   HEAP32[$$0$i$i>>2] = $17;
   $18 = HEAP32[$1>>2]|0;
   $pc$0 = $18;
   label = 19;
   break;
  }
  default: {
   $19 = ((($e)) + 8|0);
   switch ($0|0) {
   case 11:  {
    $20 = HEAP32[$19>>2]|0;
    $21 = HEAP32[$fs>>2]|0;
    $22 = ((($21)) + 12|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = (($23) + ($20<<2)|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = $25 & 63;
    $27 = ($26|0)==(19);
    if ($27) {
     $28 = ((($fs)) + 24|0);
     $29 = HEAP32[$28>>2]|0;
     $30 = (($29) + -1)|0;
     HEAP32[$28>>2] = $30;
     $31 = $25 >>> 23;
     $32 = (_condjump($fs,26,$31,0,1)|0);
     $pc$0 = $32;
     label = 19;
     break L1;
    } else {
     label = 9;
    }
    break;
   }
   case 12:  {
    label = 15;
    break;
   }
   default: {
    label = 9;
   }
   }
   if ((label|0) == 9) {
    $33 = ((($fs)) + 36|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = (($34) + 1)|0;
    $36 = HEAP32[$fs>>2]|0;
    $37 = ((($36)) + 75|0);
    $38 = HEAP8[$37>>0]|0;
    $39 = $38&255;
    $40 = ($34|0)<($39|0);
    if ($40) {
     $48 = $34;
    } else {
     $41 = ($34|0)>(248);
     if ($41) {
      $42 = ((($fs)) + 12|0);
      $43 = HEAP32[$42>>2]|0;
      _luaX_syntaxerror($43,519698);
      $$pre$i$i$i = HEAP32[$fs>>2]|0;
      $$pre$pre$i$i = HEAP32[$33>>2]|0;
      $$pre$i$i3 = $$pre$pre$i$i;$46 = $$pre$i$i$i;
     } else {
      $$pre$i$i3 = $34;$46 = $36;
     }
     $44 = $35&255;
     $45 = ((($46)) + 75|0);
     HEAP8[$45>>0] = $44;
     $48 = $$pre$i$i3;
    }
    $47 = (($48) + 1)|0;
    HEAP32[$33>>2] = $47;
    _discharge2reg($fs,$e,$48);
    $$pr = HEAP32[$e>>2]|0;
    $49 = ($$pr|0)==(12);
    if ($49) {
     label = 15;
    } else {
     $$pre = HEAP32[$19>>2]|0;
     $60 = $$pre;
    }
   }
   if ((label|0) == 15) {
    $50 = HEAP32[$19>>2]|0;
    $51 = $50 & 256;
    $52 = ($51|0)==(0);
    if ($52) {
     $53 = ((($fs)) + 50|0);
     $54 = HEAP8[$53>>0]|0;
     $55 = $54&255;
     $56 = ($55|0)>($50|0);
     if ($56) {
      $60 = $50;
     } else {
      $57 = ((($fs)) + 36|0);
      $58 = HEAP32[$57>>2]|0;
      $59 = (($58) + -1)|0;
      HEAP32[$57>>2] = $59;
      $60 = $50;
     }
    } else {
     $60 = $50;
    }
   }
   $61 = (_condjump($fs,27,255,$60,0)|0);
   $pc$0 = $61;
   label = 19;
  }
  }
 } while(0);
 do {
  if ((label|0) == 19) {
   $62 = ((($e)) + 20|0);
   $63 = ($pc$0|0)==(-1);
   if (!($63)) {
    $64 = HEAP32[$62>>2]|0;
    $65 = ($64|0)==(-1);
    if ($65) {
     HEAP32[$62>>2] = $pc$0;
     break;
    }
    $fs$idx$val$i = HEAP32[$fs>>2]|0;
    $fs$idx$val$idx$i = ((($fs$idx$val$i)) + 12|0);
    $fs$idx$val$idx$val$i = HEAP32[$fs$idx$val$idx$i>>2]|0;
    $list$0$i = $64;
    while(1) {
     $66 = (($fs$idx$val$idx$val$i) + ($list$0$i<<2)|0);
     $67 = HEAP32[$66>>2]|0;
     $68 = $67 >>> 14;
     $69 = (($68) + -131071)|0;
     $70 = ($69|0)==(-1);
     $71 = (($list$0$i) + 1)|0;
     $72 = (($71) + ($69))|0;
     $$0$i$i1 = $70 ? -1 : $72;
     $73 = ($$0$i$i1|0)==(-1);
     if ($73) {
      $$lcssa20 = $66;$$lcssa21 = $67;$list$0$i$lcssa = $list$0$i;
      break;
     } else {
      $list$0$i = $$0$i$i1;
     }
    }
    $$neg$i$i = $list$0$i$lcssa ^ -1;
    $74 = (($pc$0) + ($$neg$i$i))|0;
    $ispos$i$i = ($74|0)>(-1);
    $neg$i$i = (0 - ($74))|0;
    $75 = $ispos$i$i ? $74 : $neg$i$i;
    $76 = ($75|0)>(131071);
    if ($76) {
     $77 = ((($fs)) + 12|0);
     $78 = HEAP32[$77>>2]|0;
     _luaX_syntaxerror($78,519652);
     $$pre$i2 = HEAP32[$$lcssa20>>2]|0;
     $80 = $$pre$i2;
    } else {
     $80 = $$lcssa21;
    }
    $79 = $80 & 16383;
    $81 = $74 << 14;
    $82 = (($81) + 2147467264)|0;
    $83 = $79 | $82;
    HEAP32[$$lcssa20>>2] = $83;
   }
  }
 } while(0);
 $84 = ((($e)) + 16|0);
 $85 = HEAP32[$84>>2]|0;
 $86 = ((($fs)) + 24|0);
 $87 = HEAP32[$86>>2]|0;
 $88 = ((($fs)) + 28|0);
 HEAP32[$88>>2] = $87;
 $89 = ((($fs)) + 32|0);
 $90 = ($85|0)==(-1);
 if ($90) {
  HEAP32[$84>>2] = -1;
  return;
 }
 $91 = HEAP32[$89>>2]|0;
 $92 = ($91|0)==(-1);
 if ($92) {
  HEAP32[$89>>2] = $85;
  HEAP32[$84>>2] = -1;
  return;
 }
 $fs$idx$val$i$i = HEAP32[$fs>>2]|0;
 $fs$idx$val$idx$i$i = ((($fs$idx$val$i$i)) + 12|0);
 $fs$idx$val$idx$val$i$i = HEAP32[$fs$idx$val$idx$i$i>>2]|0;
 $list$0$i$i = $91;
 while(1) {
  $93 = (($fs$idx$val$idx$val$i$i) + ($list$0$i$i<<2)|0);
  $94 = HEAP32[$93>>2]|0;
  $95 = $94 >>> 14;
  $96 = (($95) + -131071)|0;
  $97 = ($96|0)==(-1);
  $98 = (($list$0$i$i) + 1)|0;
  $99 = (($98) + ($96))|0;
  $$0$i$i$i = $97 ? -1 : $99;
  $100 = ($$0$i$i$i|0)==(-1);
  if ($100) {
   $$lcssa = $93;$$lcssa19 = $94;$list$0$i$i$lcssa = $list$0$i$i;
   break;
  } else {
   $list$0$i$i = $$0$i$i$i;
  }
 }
 $$neg$i$i$i = $list$0$i$i$lcssa ^ -1;
 $101 = (($85) + ($$neg$i$i$i))|0;
 $ispos$i$i$i = ($101|0)>(-1);
 $neg$i$i$i = (0 - ($101))|0;
 $102 = $ispos$i$i$i ? $101 : $neg$i$i$i;
 $103 = ($102|0)>(131071);
 if ($103) {
  $104 = ((($fs)) + 12|0);
  $105 = HEAP32[$104>>2]|0;
  _luaX_syntaxerror($105,519652);
  $$pre$i$i = HEAP32[$$lcssa>>2]|0;
  $107 = $$pre$i$i;
 } else {
  $107 = $$lcssa19;
 }
 $106 = $107 & 16383;
 $108 = $101 << 14;
 $109 = (($108) + 2147467264)|0;
 $110 = $106 | $109;
 HEAP32[$$lcssa>>2] = $110;
 HEAP32[$84>>2] = -1;
 return;
}
function _condjump($fs,$op,$A,$B,$C) {
 $fs = $fs|0;
 $op = $op|0;
 $A = $A|0;
 $B = $B|0;
 $C = $C|0;
 var $$0$i$i$i = 0, $$lcssa = 0, $$lcssa6 = 0, $$neg$i$i$i = 0, $$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $fs$idx$val$i$i = 0, $fs$idx$val$idx$i$i = 0, $fs$idx$val$idx$val$i$i = 0, $ispos$i$i$i = 0, $j$0$i = 0, $list$0$i$i = 0, $list$0$i$i$lcssa = 0, $neg$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $A << 6;
 $1 = $0 | $op;
 $2 = $B << 23;
 $3 = $1 | $2;
 $4 = $C << 14;
 $5 = $3 | $4;
 $6 = ((($fs)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($7)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 (_luaK_code($fs,$5,$9)|0);
 $10 = ((($fs)) + 32|0);
 $11 = HEAP32[$10>>2]|0;
 HEAP32[$10>>2] = -1;
 $12 = HEAP32[$6>>2]|0;
 $13 = ((($12)) + 8|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (_luaK_code($fs,2147450902,$14)|0);
 $16 = ($11|0)==(-1);
 if ($16) {
  $j$0$i = $15;
  return ($j$0$i|0);
 }
 $17 = ($15|0)==(-1);
 if ($17) {
  $j$0$i = $11;
  return ($j$0$i|0);
 }
 $fs$idx$val$i$i = HEAP32[$fs>>2]|0;
 $fs$idx$val$idx$i$i = ((($fs$idx$val$i$i)) + 12|0);
 $fs$idx$val$idx$val$i$i = HEAP32[$fs$idx$val$idx$i$i>>2]|0;
 $list$0$i$i = $15;
 while(1) {
  $18 = (($fs$idx$val$idx$val$i$i) + ($list$0$i$i<<2)|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = $19 >>> 14;
  $21 = (($20) + -131071)|0;
  $22 = ($21|0)==(-1);
  $23 = (($list$0$i$i) + 1)|0;
  $24 = (($23) + ($21))|0;
  $$0$i$i$i = $22 ? -1 : $24;
  $25 = ($$0$i$i$i|0)==(-1);
  if ($25) {
   $$lcssa = $18;$$lcssa6 = $19;$list$0$i$i$lcssa = $list$0$i$i;
   break;
  } else {
   $list$0$i$i = $$0$i$i$i;
  }
 }
 $$neg$i$i$i = $list$0$i$i$lcssa ^ -1;
 $26 = (($11) + ($$neg$i$i$i))|0;
 $ispos$i$i$i = ($26|0)>(-1);
 $neg$i$i$i = (0 - ($26))|0;
 $27 = $ispos$i$i$i ? $26 : $neg$i$i$i;
 $28 = ($27|0)>(131071);
 if ($28) {
  $29 = HEAP32[$6>>2]|0;
  _luaX_syntaxerror($29,519652);
  $$pre$i$i = HEAP32[$$lcssa>>2]|0;
  $31 = $$pre$i$i;
 } else {
  $31 = $$lcssa6;
 }
 $30 = $31 & 16383;
 $32 = $26 << 14;
 $33 = (($32) + 2147467264)|0;
 $34 = $30 | $33;
 HEAP32[$$lcssa>>2] = $34;
 $j$0$i = $15;
 return ($j$0$i|0);
}
function _luaK_indexed($fs,$t,$k) {
 $fs = $fs|0;
 $t = $t|0;
 $k = $k|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaK_exp2RK($fs,$k)|0);
 $1 = ((($t)) + 8|0);
 $2 = ((($1)) + 4|0);
 HEAP32[$2>>2] = $0;
 HEAP32[$t>>2] = 9;
 return;
}
function _luaK_prefix($fs,$op,$e) {
 $fs = $fs|0;
 $op = $op|0;
 $e = $e|0;
 var $$$i$i = 0, $$$i10$i = 0, $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i8$i = 0, $$03$i$i = 0, $$03$i6$i = 0, $$phi$trans$insert$i = 0, $$pr$i = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i$i = 0, $$pre$i$i13$i = 0, $$pre$i$i7$i = 0, $$pre$phi16$iZ2D = 0, $$pre$pre$i$i$i = 0, $$pre15$i = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $12 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $e2 = 0, $fs$idx$val$i = 0, $fs$idx$val$idx$i = 0, $fs$idx$val$idx$val$i = 0, $fs$idx1$val$idx$phi$trans$insert$i$i = 0;
 var $fs$idx1$val$idx$phi$trans$insert$i3$i = 0, $fs$idx1$val$idx$val$pre$i$i = 0, $fs$idx1$val$idx$val$pre$i4$i = 0, $fs$idx1$val$pre$i$i = 0, $fs$idx1$val$pre$i2$i = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $e2 = sp;
 $0 = ((($e2)) + 20|0);
 HEAP32[$0>>2] = -1;
 $1 = ((($e2)) + 16|0);
 HEAP32[$1>>2] = -1;
 HEAP32[$e2>>2] = 5;
 $2 = ((($e2)) + 8|0);
 HEAPF64[$2>>3] = 0.0;
 switch ($op|0) {
 case 0:  {
  $3 = HEAP32[$e>>2]|0;
  $4 = ($3|0)==(5);
  if ($4) {
   $5 = ((($e)) + 16|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($6|0)==(-1);
   if ($7) {
    $8 = ((($e)) + 20|0);
    $9 = HEAP32[$8>>2]|0;
    $10 = ($9|0)==(-1);
    if (!($10)) {
     label = 5;
    }
   } else {
    label = 5;
   }
  } else {
   label = 5;
  }
  if ((label|0) == 5) {
   (_luaK_exp2anyreg($fs,$e)|0);
  }
  _codearith($fs,18,$e,$e2);
  STACKTOP = sp;return;
  break;
 }
 case 1:  {
  _luaK_dischargevars($fs,$e);
  $11 = HEAP32[$e>>2]|0;
  switch ($11|0) {
  case 3: case 1:  {
   HEAP32[$e>>2] = 2;
   break;
  }
  case 2: case 5: case 4:  {
   HEAP32[$e>>2] = 3;
   break;
  }
  case 10:  {
   $fs$idx$val$i = HEAP32[$fs>>2]|0;
   $fs$idx$val$idx$i = ((($fs$idx$val$i)) + 12|0);
   $fs$idx$val$idx$val$i = HEAP32[$fs$idx$val$idx$i>>2]|0;
   $12 = ((($e)) + 8|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = (($fs$idx$val$idx$val$i) + ($13<<2)|0);
   $15 = ($13|0)>(0);
   if ($15) {
    $16 = ((($14)) + -4|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = $17 & 63;
    $19 = (518247 + ($18)|0);
    $20 = HEAP8[$19>>0]|0;
    $21 = ($20<<24>>24)<(0);
    if ($21) {
     $$0$i$i$i = $16;$23 = $17;
    } else {
     label = 12;
    }
   } else {
    label = 12;
   }
   if ((label|0) == 12) {
    $$pre$i$i = HEAP32[$14>>2]|0;
    $$0$i$i$i = $14;$23 = $$pre$i$i;
   }
   $22 = $23 & -16321;
   $24 = $23 & 16320;
   $25 = ($24|0)==(0);
   $26 = $25&1;
   $27 = $26 << 6;
   $28 = $27 | $22;
   HEAP32[$$0$i$i$i>>2] = $28;
   break;
  }
  case 12:  {
   label = 20;
   break;
  }
  case 11:  {
   $29 = ((($fs)) + 36|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = (($30) + 1)|0;
   $32 = HEAP32[$fs>>2]|0;
   $33 = ((($32)) + 75|0);
   $34 = HEAP8[$33>>0]|0;
   $35 = $34&255;
   $36 = ($30|0)<($35|0);
   if ($36) {
    $44 = $30;
   } else {
    $37 = ($30|0)>(248);
    if ($37) {
     $38 = ((($fs)) + 12|0);
     $39 = HEAP32[$38>>2]|0;
     _luaX_syntaxerror($39,519698);
     $$pre$i$i$i$i = HEAP32[$fs>>2]|0;
     $$pre$pre$i$i$i = HEAP32[$29>>2]|0;
     $$pre$i$i13$i = $$pre$pre$i$i$i;$42 = $$pre$i$i$i$i;
    } else {
     $$pre$i$i13$i = $30;$42 = $32;
    }
    $40 = $31&255;
    $41 = ((($42)) + 75|0);
    HEAP8[$41>>0] = $40;
    $44 = $$pre$i$i13$i;
   }
   $43 = (($44) + 1)|0;
   HEAP32[$29>>2] = $43;
   _discharge2reg($fs,$e,$44);
   $$pre$i = HEAP32[$e>>2]|0;
   $phitmp$i = ($$pre$i|0)==(12);
   if ($phitmp$i) {
    label = 20;
   } else {
    $$phi$trans$insert$i = ((($e)) + 8|0);
    $$pre15$i = HEAP32[$$phi$trans$insert$i>>2]|0;
    $$pre$phi16$iZ2D = $$phi$trans$insert$i;$57 = $$pre15$i;
    label = 23;
   }
   break;
  }
  default: {
  }
  }
  if ((label|0) == 20) {
   $45 = ((($e)) + 8|0);
   $46 = HEAP32[$45>>2]|0;
   $47 = $46 & 256;
   $48 = ($47|0)==(0);
   if ($48) {
    $49 = ((($fs)) + 50|0);
    $50 = HEAP8[$49>>0]|0;
    $51 = $50&255;
    $52 = ($51|0)>($46|0);
    if ($52) {
     $$pre$phi16$iZ2D = $45;$57 = $46;
     label = 23;
    } else {
     $53 = ((($fs)) + 36|0);
     $54 = HEAP32[$53>>2]|0;
     $55 = (($54) + -1)|0;
     HEAP32[$53>>2] = $55;
     $$pre$phi16$iZ2D = $45;$57 = $46;
     label = 23;
    }
   } else {
    $$pre$phi16$iZ2D = $45;$57 = $46;
    label = 23;
   }
  }
  if ((label|0) == 23) {
   $56 = $57 << 23;
   $58 = $56 | 19;
   $59 = ((($fs)) + 12|0);
   $60 = HEAP32[$59>>2]|0;
   $61 = ((($60)) + 8|0);
   $62 = HEAP32[$61>>2]|0;
   $63 = (_luaK_code($fs,$58,$62)|0);
   HEAP32[$$pre$phi16$iZ2D>>2] = $63;
   HEAP32[$e>>2] = 11;
  }
  $64 = ((($e)) + 20|0);
  $65 = HEAP32[$64>>2]|0;
  $66 = ((($e)) + 16|0);
  $67 = HEAP32[$66>>2]|0;
  HEAP32[$64>>2] = $67;
  HEAP32[$66>>2] = $65;
  $68 = ($67|0)==(-1);
  if ($68) {
   $93 = $65;
  } else {
   $fs$idx1$val$pre$i2$i = HEAP32[$fs>>2]|0;
   $fs$idx1$val$idx$phi$trans$insert$i3$i = ((($fs$idx1$val$pre$i2$i)) + 12|0);
   $fs$idx1$val$idx$val$pre$i4$i = HEAP32[$fs$idx1$val$idx$phi$trans$insert$i3$i>>2]|0;
   $$03$i6$i = $67;
   while(1) {
    $69 = (($fs$idx1$val$idx$val$pre$i4$i) + ($$03$i6$i<<2)|0);
    $70 = ($$03$i6$i|0)>(0);
    if ($70) {
     $71 = ((($69)) + -4|0);
     $72 = HEAP32[$71>>2]|0;
     $73 = $72 & 63;
     $74 = (518247 + ($73)|0);
     $75 = HEAP8[$74>>0]|0;
     $76 = ($75<<24>>24)<(0);
     if ($76) {
      $$0$i$i$i8$i = $71;$78 = $72;
     } else {
      label = 28;
     }
    } else {
     label = 28;
    }
    if ((label|0) == 28) {
     label = 0;
     $$pre$i$i7$i = HEAP32[$69>>2]|0;
     $$0$i$i$i8$i = $69;$78 = $$pre$i$i7$i;
    }
    $77 = $78 & 63;
    $79 = ($77|0)==(27);
    if ($79) {
     $80 = $78 >>> 23;
     $81 = $80 << 6;
     $82 = $78 & 8372224;
     $83 = $82 | $81;
     $84 = $83 | 26;
     HEAP32[$$0$i$i$i8$i>>2] = $84;
    }
    $85 = HEAP32[$69>>2]|0;
    $86 = $85 >>> 14;
    $87 = (($86) + -131071)|0;
    $88 = ($87|0)==(-1);
    $89 = (($$03$i6$i) + 1)|0;
    $90 = (($89) + ($87))|0;
    $$$i10$i = $88 ? -1 : $90;
    $91 = ($$$i10$i|0)==(-1);
    if ($91) {
     break;
    } else {
     $$03$i6$i = $$$i10$i;
    }
   }
   $$pr$i = HEAP32[$66>>2]|0;
   $93 = $$pr$i;
  }
  $92 = ($93|0)==(-1);
  if ($92) {
   STACKTOP = sp;return;
  }
  $fs$idx1$val$pre$i$i = HEAP32[$fs>>2]|0;
  $fs$idx1$val$idx$phi$trans$insert$i$i = ((($fs$idx1$val$pre$i$i)) + 12|0);
  $fs$idx1$val$idx$val$pre$i$i = HEAP32[$fs$idx1$val$idx$phi$trans$insert$i$i>>2]|0;
  $$03$i$i = $93;
  while(1) {
   $94 = (($fs$idx1$val$idx$val$pre$i$i) + ($$03$i$i<<2)|0);
   $95 = ($$03$i$i|0)>(0);
   if ($95) {
    $96 = ((($94)) + -4|0);
    $97 = HEAP32[$96>>2]|0;
    $98 = $97 & 63;
    $99 = (518247 + ($98)|0);
    $100 = HEAP8[$99>>0]|0;
    $101 = ($100<<24>>24)<(0);
    if ($101) {
     $$0$i$i$i$i = $96;$103 = $97;
    } else {
     label = 37;
    }
   } else {
    label = 37;
   }
   if ((label|0) == 37) {
    label = 0;
    $$pre$i$i$i = HEAP32[$94>>2]|0;
    $$0$i$i$i$i = $94;$103 = $$pre$i$i$i;
   }
   $102 = $103 & 63;
   $104 = ($102|0)==(27);
   if ($104) {
    $105 = $103 >>> 23;
    $106 = $105 << 6;
    $107 = $103 & 8372224;
    $108 = $107 | $106;
    $109 = $108 | 26;
    HEAP32[$$0$i$i$i$i>>2] = $109;
   }
   $110 = HEAP32[$94>>2]|0;
   $111 = $110 >>> 14;
   $112 = (($111) + -131071)|0;
   $113 = ($112|0)==(-1);
   $114 = (($$03$i$i) + 1)|0;
   $115 = (($114) + ($112))|0;
   $$$i$i = $113 ? -1 : $115;
   $116 = ($$$i$i|0)==(-1);
   if ($116) {
    break;
   } else {
    $$03$i$i = $$$i$i;
   }
  }
  STACKTOP = sp;return;
  break;
 }
 case 2:  {
  (_luaK_exp2anyreg($fs,$e)|0);
  _codearith($fs,20,$e,$e2);
  STACKTOP = sp;return;
  break;
 }
 default: {
  STACKTOP = sp;return;
 }
 }
}
function _codearith($fs,$op,$e1,$e2) {
 $fs = $fs|0;
 $op = $op|0;
 $e1 = $e1|0;
 $e2 = $e2|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0, $26 = 0.0;
 var $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $r$0$i = 0.0;
 var $r$02$i = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$e1>>2]|0;
 $1 = ($0|0)==(5);
 L1: do {
  if ($1) {
   $2 = ((($e1)) + 16|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = ($3|0)==(-1);
   if ($4) {
    $5 = ((($e1)) + 20|0);
    $6 = HEAP32[$5>>2]|0;
    $7 = ($6|0)==(-1);
    if ($7) {
     $8 = HEAP32[$e2>>2]|0;
     $9 = ($8|0)==(5);
     if ($9) {
      $10 = ((($e2)) + 16|0);
      $11 = HEAP32[$10>>2]|0;
      $12 = ($11|0)==(-1);
      if ($12) {
       $13 = ((($e2)) + 20|0);
       $14 = HEAP32[$13>>2]|0;
       $15 = ($14|0)==(-1);
       if ($15) {
        $16 = ((($e1)) + 8|0);
        $17 = +HEAPF64[$16>>3];
        $18 = ((($e2)) + 8|0);
        $19 = +HEAPF64[$18>>3];
        switch ($op|0) {
        case 20:  {
         $36 = 0;
         break L1;
         break;
        }
        case 12:  {
         $20 = $17 + $19;
         $r$0$i = $20;
         label = 17;
         break;
        }
        case 13:  {
         $21 = $17 - $19;
         $r$0$i = $21;
         label = 17;
         break;
        }
        case 14:  {
         $22 = $17 * $19;
         $r$0$i = $22;
         label = 17;
         break;
        }
        case 15:  {
         $23 = $19 == 0.0;
         if ($23) {
          label = 20;
          break L1;
         }
         $24 = $17 / $19;
         $r$0$i = $24;
         label = 17;
         break;
        }
        case 16:  {
         $25 = $19 == 0.0;
         if ($25) {
          label = 20;
          break L1;
         }
         $26 = $17 / $19;
         $27 = (+Math_floor((+$26)));
         $28 = $19 * $27;
         $29 = $17 - $28;
         $r$0$i = $29;
         label = 17;
         break;
        }
        case 17:  {
         $30 = (+Math_pow((+$17),(+$19)));
         $r$0$i = $30;
         label = 17;
         break;
        }
        case 18:  {
         $31 = -$17;
         $r$0$i = $31;
         label = 17;
         break;
        }
        default: {
         $r$02$i = 0.0;
        }
        }
        if ((label|0) == 17) {
         $32 = ($r$0$i == $r$0$i) & (0.0 == 0.0);
         if ($32) {
          $r$02$i = $r$0$i;
         } else {
          label = 19;
          break;
         }
        }
        HEAPF64[$16>>3] = $r$02$i;
        return;
       } else {
        label = 19;
       }
      } else {
       label = 19;
      }
     } else {
      label = 19;
     }
    } else {
     label = 19;
    }
   } else {
    label = 19;
   }
  } else {
   label = 19;
  }
 } while(0);
 if ((label|0) == 19) {
  switch ($op|0) {
  case 18: case 20:  {
   $36 = 0;
   break;
  }
  default: {
   label = 20;
  }
  }
 }
 if ((label|0) == 20) {
  $33 = (_luaK_exp2RK($fs,$e2)|0);
  $36 = $33;
 }
 $34 = (_luaK_exp2RK($fs,$e1)|0);
 $35 = ($34|0)>($36|0);
 if ($35) {
  $37 = HEAP32[$e1>>2]|0;
  $38 = ($37|0)==(12);
  if ($38) {
   $39 = ((($e1)) + 8|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = $40 & 256;
   $42 = ($41|0)==(0);
   if ($42) {
    $43 = ((($fs)) + 50|0);
    $44 = HEAP8[$43>>0]|0;
    $45 = $44&255;
    $46 = ($45|0)>($40|0);
    if (!($46)) {
     $47 = ((($fs)) + 36|0);
     $48 = HEAP32[$47>>2]|0;
     $49 = (($48) + -1)|0;
     HEAP32[$47>>2] = $49;
    }
   }
  }
  $50 = HEAP32[$e2>>2]|0;
  $51 = ($50|0)==(12);
  if ($51) {
   $52 = ((($e2)) + 8|0);
   $53 = HEAP32[$52>>2]|0;
   $54 = $53 & 256;
   $55 = ($54|0)==(0);
   if ($55) {
    $56 = ((($fs)) + 50|0);
    $57 = HEAP8[$56>>0]|0;
    $58 = $57&255;
    $59 = ($58|0)>($53|0);
    if (!($59)) {
     $60 = ((($fs)) + 36|0);
     $61 = HEAP32[$60>>2]|0;
     $62 = (($61) + -1)|0;
     HEAP32[$60>>2] = $62;
    }
   }
  }
 } else {
  $63 = HEAP32[$e2>>2]|0;
  $64 = ($63|0)==(12);
  if ($64) {
   $65 = ((($e2)) + 8|0);
   $66 = HEAP32[$65>>2]|0;
   $67 = $66 & 256;
   $68 = ($67|0)==(0);
   if ($68) {
    $69 = ((($fs)) + 50|0);
    $70 = HEAP8[$69>>0]|0;
    $71 = $70&255;
    $72 = ($71|0)>($66|0);
    if (!($72)) {
     $73 = ((($fs)) + 36|0);
     $74 = HEAP32[$73>>2]|0;
     $75 = (($74) + -1)|0;
     HEAP32[$73>>2] = $75;
    }
   }
  }
  $76 = HEAP32[$e1>>2]|0;
  $77 = ($76|0)==(12);
  if ($77) {
   $78 = ((($e1)) + 8|0);
   $79 = HEAP32[$78>>2]|0;
   $80 = $79 & 256;
   $81 = ($80|0)==(0);
   if ($81) {
    $82 = ((($fs)) + 50|0);
    $83 = HEAP8[$82>>0]|0;
    $84 = $83&255;
    $85 = ($84|0)>($79|0);
    if (!($85)) {
     $86 = ((($fs)) + 36|0);
     $87 = HEAP32[$86>>2]|0;
     $88 = (($87) + -1)|0;
     HEAP32[$86>>2] = $88;
    }
   }
  }
 }
 $89 = $34 << 23;
 $90 = $36 << 14;
 $91 = $90 | $op;
 $92 = $91 | $89;
 $93 = ((($fs)) + 12|0);
 $94 = HEAP32[$93>>2]|0;
 $95 = ((($94)) + 8|0);
 $96 = HEAP32[$95>>2]|0;
 $97 = (_luaK_code($fs,$92,$96)|0);
 $98 = ((($e1)) + 8|0);
 HEAP32[$98>>2] = $97;
 HEAP32[$e1>>2] = 11;
 return;
}
function _luaK_infix($fs,$op,$v) {
 $fs = $fs|0;
 $op = $op|0;
 $v = $v|0;
 var $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$lcssa = 0, $$lcssa20 = 0, $$lcssa21 = 0, $$lcssa22 = 0, $$neg$i$i$i = 0, $$neg$i$i$i$i = 0, $$pr$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i$i = 0, $$pre$i$i$i2 = 0, $$pre$i$i1$i = 0, $$pre$i$i3 = 0, $$pre$phiZ2D = 0, $$pre$pre$i$i = 0, $$pre$pre$i$i$i = 0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fs$idx$val$i$i = 0, $fs$idx$val$i$i$i = 0, $fs$idx$val$idx$i$i = 0, $fs$idx$val$idx$i$i$i = 0, $fs$idx$val$idx$val$i$i = 0, $fs$idx$val$idx$val$i$i$i = 0, $ispos$i$i$i = 0, $ispos$i$i$i$i = 0;
 var $list$0$i$i = 0, $list$0$i$i$i = 0, $list$0$i$i$i$lcssa = 0, $list$0$i$i$lcssa = 0, $neg$i$i$i = 0, $neg$i$i$i$i = 0, $pc$0$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 switch ($op|0) {
 case 13:  {
  _luaK_goiftrue($fs,$v);
  return;
  break;
 }
 case 14:  {
  _luaK_dischargevars($fs,$v);
  $0 = HEAP32[$v>>2]|0;
  L7: do {
   switch ($0|0) {
   case 3: case 1:  {
    break;
   }
   case 10:  {
    $1 = ((($v)) + 8|0);
    $2 = HEAP32[$1>>2]|0;
    $pc$0$i = $2;
    label = 19;
    break;
   }
   default: {
    $3 = ((($v)) + 8|0);
    switch ($0|0) {
    case 11:  {
     $4 = HEAP32[$3>>2]|0;
     $5 = HEAP32[$fs>>2]|0;
     $6 = ((($5)) + 12|0);
     $7 = HEAP32[$6>>2]|0;
     $8 = (($7) + ($4<<2)|0);
     $9 = HEAP32[$8>>2]|0;
     $10 = $9 & 63;
     $11 = ($10|0)==(19);
     if ($11) {
      $12 = ((($fs)) + 24|0);
      $13 = HEAP32[$12>>2]|0;
      $14 = (($13) + -1)|0;
      HEAP32[$12>>2] = $14;
      $15 = $9 >>> 23;
      $16 = (_condjump($fs,26,$15,0,0)|0);
      $pc$0$i = $16;
      label = 19;
      break L7;
     } else {
      $$pre$phiZ2D = $fs;$21 = $5;
      label = 9;
     }
     break;
    }
    case 12:  {
     label = 15;
     break;
    }
    default: {
     $$pre = HEAP32[$fs>>2]|0;
     $$pre$phiZ2D = $fs;$21 = $$pre;
     label = 9;
    }
    }
    if ((label|0) == 9) {
     $17 = ((($fs)) + 36|0);
     $18 = HEAP32[$17>>2]|0;
     $19 = (($18) + 1)|0;
     $20 = ((($21)) + 75|0);
     $22 = HEAP8[$20>>0]|0;
     $23 = $22&255;
     $24 = ($18|0)<($23|0);
     if ($24) {
      $32 = $18;
     } else {
      $25 = ($18|0)>(248);
      if ($25) {
       $26 = ((($fs)) + 12|0);
       $27 = HEAP32[$26>>2]|0;
       _luaX_syntaxerror($27,519698);
       $$pre$i$i$i$i = HEAP32[$$pre$phiZ2D>>2]|0;
       $$pre$pre$i$i$i = HEAP32[$17>>2]|0;
       $$pre$i$i1$i = $$pre$pre$i$i$i;$30 = $$pre$i$i$i$i;
      } else {
       $$pre$i$i1$i = $18;$30 = $21;
      }
      $28 = $19&255;
      $29 = ((($30)) + 75|0);
      HEAP8[$29>>0] = $28;
      $32 = $$pre$i$i1$i;
     }
     $31 = (($32) + 1)|0;
     HEAP32[$17>>2] = $31;
     _discharge2reg($fs,$v,$32);
     $$pr$i = HEAP32[$v>>2]|0;
     $33 = ($$pr$i|0)==(12);
     if ($33) {
      label = 15;
     } else {
      $$pre$i = HEAP32[$3>>2]|0;
      $44 = $$pre$i;
     }
    }
    if ((label|0) == 15) {
     $34 = HEAP32[$3>>2]|0;
     $35 = $34 & 256;
     $36 = ($35|0)==(0);
     if ($36) {
      $37 = ((($fs)) + 50|0);
      $38 = HEAP8[$37>>0]|0;
      $39 = $38&255;
      $40 = ($39|0)>($34|0);
      if ($40) {
       $44 = $34;
      } else {
       $41 = ((($fs)) + 36|0);
       $42 = HEAP32[$41>>2]|0;
       $43 = (($42) + -1)|0;
       HEAP32[$41>>2] = $43;
       $44 = $34;
      }
     } else {
      $44 = $34;
     }
    }
    $45 = (_condjump($fs,27,255,$44,1)|0);
    $pc$0$i = $45;
    label = 19;
   }
   }
  } while(0);
  do {
   if ((label|0) == 19) {
    $46 = ((($v)) + 16|0);
    $47 = ($pc$0$i|0)==(-1);
    if (!($47)) {
     $48 = HEAP32[$46>>2]|0;
     $49 = ($48|0)==(-1);
     if ($49) {
      HEAP32[$46>>2] = $pc$0$i;
      break;
     }
     $fs$idx$val$i$i = HEAP32[$fs>>2]|0;
     $fs$idx$val$idx$i$i = ((($fs$idx$val$i$i)) + 12|0);
     $fs$idx$val$idx$val$i$i = HEAP32[$fs$idx$val$idx$i$i>>2]|0;
     $list$0$i$i = $48;
     while(1) {
      $50 = (($fs$idx$val$idx$val$i$i) + ($list$0$i$i<<2)|0);
      $51 = HEAP32[$50>>2]|0;
      $52 = $51 >>> 14;
      $53 = (($52) + -131071)|0;
      $54 = ($53|0)==(-1);
      $55 = (($list$0$i$i) + 1)|0;
      $56 = (($55) + ($53))|0;
      $$0$i$i$i = $54 ? -1 : $56;
      $57 = ($$0$i$i$i|0)==(-1);
      if ($57) {
       $$lcssa21 = $50;$$lcssa22 = $51;$list$0$i$i$lcssa = $list$0$i$i;
       break;
      } else {
       $list$0$i$i = $$0$i$i$i;
      }
     }
     $$neg$i$i$i = $list$0$i$i$lcssa ^ -1;
     $58 = (($pc$0$i) + ($$neg$i$i$i))|0;
     $ispos$i$i$i = ($58|0)>(-1);
     $neg$i$i$i = (0 - ($58))|0;
     $59 = $ispos$i$i$i ? $58 : $neg$i$i$i;
     $60 = ($59|0)>(131071);
     if ($60) {
      $61 = ((($fs)) + 12|0);
      $62 = HEAP32[$61>>2]|0;
      _luaX_syntaxerror($62,519652);
      $$pre$i$i = HEAP32[$$lcssa21>>2]|0;
      $64 = $$pre$i$i;
     } else {
      $64 = $$lcssa22;
     }
     $63 = $64 & 16383;
     $65 = $58 << 14;
     $66 = (($65) + 2147467264)|0;
     $67 = $63 | $66;
     HEAP32[$$lcssa21>>2] = $67;
    }
   }
  } while(0);
  $68 = ((($v)) + 20|0);
  $69 = HEAP32[$68>>2]|0;
  $70 = ((($fs)) + 24|0);
  $71 = HEAP32[$70>>2]|0;
  $72 = ((($fs)) + 28|0);
  HEAP32[$72>>2] = $71;
  $73 = ((($fs)) + 32|0);
  $74 = ($69|0)==(-1);
  do {
   if (!($74)) {
    $75 = HEAP32[$73>>2]|0;
    $76 = ($75|0)==(-1);
    if ($76) {
     HEAP32[$73>>2] = $69;
     break;
    }
    $fs$idx$val$i$i$i = HEAP32[$fs>>2]|0;
    $fs$idx$val$idx$i$i$i = ((($fs$idx$val$i$i$i)) + 12|0);
    $fs$idx$val$idx$val$i$i$i = HEAP32[$fs$idx$val$idx$i$i$i>>2]|0;
    $list$0$i$i$i = $75;
    while(1) {
     $77 = (($fs$idx$val$idx$val$i$i$i) + ($list$0$i$i$i<<2)|0);
     $78 = HEAP32[$77>>2]|0;
     $79 = $78 >>> 14;
     $80 = (($79) + -131071)|0;
     $81 = ($80|0)==(-1);
     $82 = (($list$0$i$i$i) + 1)|0;
     $83 = (($82) + ($80))|0;
     $$0$i$i$i$i = $81 ? -1 : $83;
     $84 = ($$0$i$i$i$i|0)==(-1);
     if ($84) {
      $$lcssa = $77;$$lcssa20 = $78;$list$0$i$i$i$lcssa = $list$0$i$i$i;
      break;
     } else {
      $list$0$i$i$i = $$0$i$i$i$i;
     }
    }
    $$neg$i$i$i$i = $list$0$i$i$i$lcssa ^ -1;
    $85 = (($69) + ($$neg$i$i$i$i))|0;
    $ispos$i$i$i$i = ($85|0)>(-1);
    $neg$i$i$i$i = (0 - ($85))|0;
    $86 = $ispos$i$i$i$i ? $85 : $neg$i$i$i$i;
    $87 = ($86|0)>(131071);
    if ($87) {
     $88 = ((($fs)) + 12|0);
     $89 = HEAP32[$88>>2]|0;
     _luaX_syntaxerror($89,519652);
     $$pre$i$i$i = HEAP32[$$lcssa>>2]|0;
     $91 = $$pre$i$i$i;
    } else {
     $91 = $$lcssa20;
    }
    $90 = $91 & 16383;
    $92 = $85 << 14;
    $93 = (($92) + 2147467264)|0;
    $94 = $90 | $93;
    HEAP32[$$lcssa>>2] = $94;
   }
  } while(0);
  HEAP32[$68>>2] = -1;
  return;
  break;
 }
 case 6:  {
  _luaK_dischargevars($fs,$v);
  $95 = HEAP32[$v>>2]|0;
  $96 = ($95|0)==(12);
  if ($96) {
   $97 = ((($v)) + 8|0);
   $98 = HEAP32[$97>>2]|0;
   $99 = $98 & 256;
   $100 = ($99|0)==(0);
   if ($100) {
    $101 = ((($fs)) + 50|0);
    $102 = HEAP8[$101>>0]|0;
    $103 = $102&255;
    $104 = ($103|0)>($98|0);
    if (!($104)) {
     $105 = ((($fs)) + 36|0);
     $106 = HEAP32[$105>>2]|0;
     $107 = (($106) + -1)|0;
     HEAP32[$105>>2] = $107;
    }
   }
  }
  $108 = ((($fs)) + 36|0);
  $109 = HEAP32[$108>>2]|0;
  $110 = (($109) + 1)|0;
  $111 = HEAP32[$fs>>2]|0;
  $112 = ((($111)) + 75|0);
  $113 = HEAP8[$112>>0]|0;
  $114 = $113&255;
  $115 = ($109|0)<($114|0);
  if ($115) {
   $123 = $109;
  } else {
   $116 = ($109|0)>(248);
   if ($116) {
    $117 = ((($fs)) + 12|0);
    $118 = HEAP32[$117>>2]|0;
    _luaX_syntaxerror($118,519698);
    $$pre$i$i$i2 = HEAP32[$fs>>2]|0;
    $$pre$pre$i$i = HEAP32[$108>>2]|0;
    $$pre$i$i3 = $$pre$pre$i$i;$121 = $$pre$i$i$i2;
   } else {
    $$pre$i$i3 = $109;$121 = $111;
   }
   $119 = $110&255;
   $120 = ((($121)) + 75|0);
   HEAP8[$120>>0] = $119;
   $123 = $$pre$i$i3;
  }
  $122 = (($123) + 1)|0;
  HEAP32[$108>>2] = $122;
  _exp2reg($fs,$v,$123);
  return;
  break;
 }
 case 5: case 4: case 3: case 2: case 1: case 0:  {
  $124 = HEAP32[$v>>2]|0;
  $125 = ($124|0)==(5);
  if ($125) {
   $126 = ((($v)) + 16|0);
   $127 = HEAP32[$126>>2]|0;
   $128 = ($127|0)==(-1);
   if ($128) {
    $129 = ((($v)) + 20|0);
    $130 = HEAP32[$129>>2]|0;
    $131 = ($130|0)==(-1);
    if ($131) {
     return;
    }
   }
  }
  (_luaK_exp2RK($fs,$v)|0);
  return;
  break;
 }
 default: {
  (_luaK_exp2RK($fs,$v)|0);
  return;
 }
 }
}
function _luaK_posfix($fs,$op,$e1,$e2) {
 $fs = $fs|0;
 $op = $op|0;
 $e1 = $e1|0;
 $e2 = $e2|0;
 var $$0$i$i = 0, $$0$i$i7 = 0, $$lcssa = 0, $$lcssa74 = 0, $$lcssa75 = 0, $$lcssa76 = 0, $$neg$i$i = 0, $$neg$i$i8 = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i$i = 0, $$pre$i11 = 0, $$pre$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0;
 var $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0;
 var $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0;
 var $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0;
 var $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0;
 var $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0;
 var $284 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fs$idx$val$i = 0;
 var $fs$idx$val$i2 = 0, $fs$idx$val$idx$i = 0, $fs$idx$val$idx$i3 = 0, $fs$idx$val$idx$val$i = 0, $fs$idx$val$idx$val$i4 = 0, $ispos$i$i = 0, $ispos$i$i9 = 0, $list$0$i = 0, $list$0$i$lcssa = 0, $list$0$i6 = 0, $list$0$i6$lcssa = 0, $neg$i$i = 0, $neg$i$i10 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 do {
  switch ($op|0) {
  case 13:  {
   _luaK_dischargevars($fs,$e2);
   $0 = ((($e2)) + 20|0);
   $1 = ((($e1)) + 20|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = ($2|0)==(-1);
   do {
    if (!($3)) {
     $4 = HEAP32[$0>>2]|0;
     $5 = ($4|0)==(-1);
     if ($5) {
      HEAP32[$0>>2] = $2;
      break;
     }
     $fs$idx$val$i = HEAP32[$fs>>2]|0;
     $fs$idx$val$idx$i = ((($fs$idx$val$i)) + 12|0);
     $fs$idx$val$idx$val$i = HEAP32[$fs$idx$val$idx$i>>2]|0;
     $list$0$i = $4;
     while(1) {
      $6 = (($fs$idx$val$idx$val$i) + ($list$0$i<<2)|0);
      $7 = HEAP32[$6>>2]|0;
      $8 = $7 >>> 14;
      $9 = (($8) + -131071)|0;
      $10 = ($9|0)==(-1);
      $11 = (($list$0$i) + 1)|0;
      $12 = (($11) + ($9))|0;
      $$0$i$i = $10 ? -1 : $12;
      $13 = ($$0$i$i|0)==(-1);
      if ($13) {
       $$lcssa = $6;$$lcssa74 = $7;$list$0$i$lcssa = $list$0$i;
       break;
      } else {
       $list$0$i = $$0$i$i;
      }
     }
     $$neg$i$i = $list$0$i$lcssa ^ -1;
     $14 = (($2) + ($$neg$i$i))|0;
     $ispos$i$i = ($14|0)>(-1);
     $neg$i$i = (0 - ($14))|0;
     $15 = $ispos$i$i ? $14 : $neg$i$i;
     $16 = ($15|0)>(131071);
     if ($16) {
      $17 = ((($fs)) + 12|0);
      $18 = HEAP32[$17>>2]|0;
      _luaX_syntaxerror($18,519652);
      $$pre$i = HEAP32[$$lcssa>>2]|0;
      $20 = $$pre$i;
     } else {
      $20 = $$lcssa74;
     }
     $19 = $20 & 16383;
     $21 = $14 << 14;
     $22 = (($21) + 2147467264)|0;
     $23 = $19 | $22;
     HEAP32[$$lcssa>>2] = $23;
    }
   } while(0);
   ;HEAP32[$e1>>2]=HEAP32[$e2>>2]|0;HEAP32[$e1+4>>2]=HEAP32[$e2+4>>2]|0;HEAP32[$e1+8>>2]=HEAP32[$e2+8>>2]|0;HEAP32[$e1+12>>2]=HEAP32[$e2+12>>2]|0;HEAP32[$e1+16>>2]=HEAP32[$e2+16>>2]|0;HEAP32[$e1+20>>2]=HEAP32[$e2+20>>2]|0;
   return;
   break;
  }
  case 14:  {
   _luaK_dischargevars($fs,$e2);
   $24 = ((($e2)) + 16|0);
   $25 = ((($e1)) + 16|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = ($26|0)==(-1);
   do {
    if (!($27)) {
     $28 = HEAP32[$24>>2]|0;
     $29 = ($28|0)==(-1);
     if ($29) {
      HEAP32[$24>>2] = $26;
      break;
     }
     $fs$idx$val$i2 = HEAP32[$fs>>2]|0;
     $fs$idx$val$idx$i3 = ((($fs$idx$val$i2)) + 12|0);
     $fs$idx$val$idx$val$i4 = HEAP32[$fs$idx$val$idx$i3>>2]|0;
     $list$0$i6 = $28;
     while(1) {
      $30 = (($fs$idx$val$idx$val$i4) + ($list$0$i6<<2)|0);
      $31 = HEAP32[$30>>2]|0;
      $32 = $31 >>> 14;
      $33 = (($32) + -131071)|0;
      $34 = ($33|0)==(-1);
      $35 = (($list$0$i6) + 1)|0;
      $36 = (($35) + ($33))|0;
      $$0$i$i7 = $34 ? -1 : $36;
      $37 = ($$0$i$i7|0)==(-1);
      if ($37) {
       $$lcssa75 = $30;$$lcssa76 = $31;$list$0$i6$lcssa = $list$0$i6;
       break;
      } else {
       $list$0$i6 = $$0$i$i7;
      }
     }
     $$neg$i$i8 = $list$0$i6$lcssa ^ -1;
     $38 = (($26) + ($$neg$i$i8))|0;
     $ispos$i$i9 = ($38|0)>(-1);
     $neg$i$i10 = (0 - ($38))|0;
     $39 = $ispos$i$i9 ? $38 : $neg$i$i10;
     $40 = ($39|0)>(131071);
     if ($40) {
      $41 = ((($fs)) + 12|0);
      $42 = HEAP32[$41>>2]|0;
      _luaX_syntaxerror($42,519652);
      $$pre$i11 = HEAP32[$$lcssa75>>2]|0;
      $44 = $$pre$i11;
     } else {
      $44 = $$lcssa76;
     }
     $43 = $44 & 16383;
     $45 = $38 << 14;
     $46 = (($45) + 2147467264)|0;
     $47 = $43 | $46;
     HEAP32[$$lcssa75>>2] = $47;
    }
   } while(0);
   ;HEAP32[$e1>>2]=HEAP32[$e2>>2]|0;HEAP32[$e1+4>>2]=HEAP32[$e2+4>>2]|0;HEAP32[$e1+8>>2]=HEAP32[$e2+8>>2]|0;HEAP32[$e1+12>>2]=HEAP32[$e2+12>>2]|0;HEAP32[$e1+16>>2]=HEAP32[$e2+16>>2]|0;HEAP32[$e1+20>>2]=HEAP32[$e2+20>>2]|0;
   return;
   break;
  }
  case 6:  {
   $48 = ((($e2)) + 16|0);
   $49 = HEAP32[$48>>2]|0;
   $50 = ((($e2)) + 20|0);
   $51 = HEAP32[$50>>2]|0;
   $52 = ($49|0)==($51|0);
   if ($52) {
    _luaK_dischargevars($fs,$e2);
   } else {
    (_luaK_exp2anyreg($fs,$e2)|0);
   }
   $53 = HEAP32[$e2>>2]|0;
   $54 = ($53|0)==(11);
   if ($54) {
    $55 = ((($e2)) + 8|0);
    $56 = HEAP32[$55>>2]|0;
    $57 = HEAP32[$fs>>2]|0;
    $58 = ((($57)) + 12|0);
    $59 = HEAP32[$58>>2]|0;
    $60 = (($59) + ($56<<2)|0);
    $61 = HEAP32[$60>>2]|0;
    $62 = $61 & 63;
    $63 = ($62|0)==(21);
    if ($63) {
     $64 = HEAP32[$e1>>2]|0;
     $65 = ($64|0)==(12);
     $66 = ((($e1)) + 8|0);
     $67 = HEAP32[$66>>2]|0;
     $68 = $67 & 256;
     $69 = ($68|0)==(0);
     $or$cond = $65 & $69;
     if ($or$cond) {
      $70 = ((($fs)) + 50|0);
      $71 = HEAP8[$70>>0]|0;
      $72 = $71&255;
      $73 = ($72|0)>($67|0);
      if ($73) {
       $78 = $61;
      } else {
       $74 = ((($fs)) + 36|0);
       $75 = HEAP32[$74>>2]|0;
       $76 = (($75) + -1)|0;
       HEAP32[$74>>2] = $76;
       $$pre = HEAP32[$60>>2]|0;
       $78 = $$pre;
      }
     } else {
      $78 = $61;
     }
     $77 = $78 & 8388607;
     $79 = $67 << 23;
     $80 = $79 | $77;
     HEAP32[$60>>2] = $80;
     HEAP32[$e1>>2] = 11;
     $81 = HEAP32[$55>>2]|0;
     HEAP32[$66>>2] = $81;
     return;
    }
   }
   _luaK_dischargevars($fs,$e2);
   $82 = HEAP32[$e2>>2]|0;
   $83 = ($82|0)==(12);
   if ($83) {
    $84 = ((($e2)) + 8|0);
    $85 = HEAP32[$84>>2]|0;
    $86 = $85 & 256;
    $87 = ($86|0)==(0);
    if ($87) {
     $88 = ((($fs)) + 50|0);
     $89 = HEAP8[$88>>0]|0;
     $90 = $89&255;
     $91 = ($90|0)>($85|0);
     if (!($91)) {
      $92 = ((($fs)) + 36|0);
      $93 = HEAP32[$92>>2]|0;
      $94 = (($93) + -1)|0;
      HEAP32[$92>>2] = $94;
     }
    }
   }
   $95 = ((($fs)) + 36|0);
   $96 = HEAP32[$95>>2]|0;
   $97 = (($96) + 1)|0;
   $98 = HEAP32[$fs>>2]|0;
   $99 = ((($98)) + 75|0);
   $100 = HEAP8[$99>>0]|0;
   $101 = $100&255;
   $102 = ($96|0)<($101|0);
   if ($102) {
    $110 = $96;
   } else {
    $103 = ($96|0)>(248);
    if ($103) {
     $104 = ((($fs)) + 12|0);
     $105 = HEAP32[$104>>2]|0;
     _luaX_syntaxerror($105,519698);
     $$pre$i$i$i = HEAP32[$fs>>2]|0;
     $$pre$pre$i$i = HEAP32[$95>>2]|0;
     $$pre$i$i = $$pre$pre$i$i;$108 = $$pre$i$i$i;
    } else {
     $$pre$i$i = $96;$108 = $98;
    }
    $106 = $97&255;
    $107 = ((($108)) + 75|0);
    HEAP8[$107>>0] = $106;
    $110 = $$pre$i$i;
   }
   $109 = (($110) + 1)|0;
   HEAP32[$95>>2] = $109;
   _exp2reg($fs,$e2,$110);
   _codearith($fs,21,$e1,$e2);
   return;
   break;
  }
  case 0:  {
   _codearith($fs,12,$e1,$e2);
   return;
   break;
  }
  case 1:  {
   _codearith($fs,13,$e1,$e2);
   return;
   break;
  }
  case 2:  {
   _codearith($fs,14,$e1,$e2);
   return;
   break;
  }
  case 3:  {
   _codearith($fs,15,$e1,$e2);
   return;
   break;
  }
  case 4:  {
   _codearith($fs,16,$e1,$e2);
   return;
   break;
  }
  case 5:  {
   _codearith($fs,17,$e1,$e2);
   return;
   break;
  }
  case 8:  {
   $111 = (_luaK_exp2RK($fs,$e1)|0);
   $112 = (_luaK_exp2RK($fs,$e2)|0);
   $113 = HEAP32[$e2>>2]|0;
   $114 = ($113|0)==(12);
   if ($114) {
    $115 = ((($e2)) + 8|0);
    $116 = HEAP32[$115>>2]|0;
    $117 = $116 & 256;
    $118 = ($117|0)==(0);
    if ($118) {
     $119 = ((($fs)) + 50|0);
     $120 = HEAP8[$119>>0]|0;
     $121 = $120&255;
     $122 = ($121|0)>($116|0);
     if (!($122)) {
      $123 = ((($fs)) + 36|0);
      $124 = HEAP32[$123>>2]|0;
      $125 = (($124) + -1)|0;
      HEAP32[$123>>2] = $125;
     }
    }
   }
   $126 = HEAP32[$e1>>2]|0;
   $127 = ($126|0)==(12);
   $128 = ((($e1)) + 8|0);
   if ($127) {
    $129 = HEAP32[$128>>2]|0;
    $130 = $129 & 256;
    $131 = ($130|0)==(0);
    if ($131) {
     $132 = ((($fs)) + 50|0);
     $133 = HEAP8[$132>>0]|0;
     $134 = $133&255;
     $135 = ($134|0)>($129|0);
     if (!($135)) {
      $136 = ((($fs)) + 36|0);
      $137 = HEAP32[$136>>2]|0;
      $138 = (($137) + -1)|0;
      HEAP32[$136>>2] = $138;
     }
    }
   }
   $139 = (_condjump($fs,23,1,$111,$112)|0);
   HEAP32[$128>>2] = $139;
   HEAP32[$e1>>2] = 10;
   return;
   break;
  }
  case 7:  {
   $140 = (_luaK_exp2RK($fs,$e1)|0);
   $141 = (_luaK_exp2RK($fs,$e2)|0);
   $142 = HEAP32[$e2>>2]|0;
   $143 = ($142|0)==(12);
   if ($143) {
    $144 = ((($e2)) + 8|0);
    $145 = HEAP32[$144>>2]|0;
    $146 = $145 & 256;
    $147 = ($146|0)==(0);
    if ($147) {
     $148 = ((($fs)) + 50|0);
     $149 = HEAP8[$148>>0]|0;
     $150 = $149&255;
     $151 = ($150|0)>($145|0);
     if (!($151)) {
      $152 = ((($fs)) + 36|0);
      $153 = HEAP32[$152>>2]|0;
      $154 = (($153) + -1)|0;
      HEAP32[$152>>2] = $154;
     }
    }
   }
   $155 = HEAP32[$e1>>2]|0;
   $156 = ($155|0)==(12);
   $157 = ((($e1)) + 8|0);
   if ($156) {
    $158 = HEAP32[$157>>2]|0;
    $159 = $158 & 256;
    $160 = ($159|0)==(0);
    if ($160) {
     $161 = ((($fs)) + 50|0);
     $162 = HEAP8[$161>>0]|0;
     $163 = $162&255;
     $164 = ($163|0)>($158|0);
     if (!($164)) {
      $165 = ((($fs)) + 36|0);
      $166 = HEAP32[$165>>2]|0;
      $167 = (($166) + -1)|0;
      HEAP32[$165>>2] = $167;
     }
    }
   }
   $168 = (_condjump($fs,23,0,$140,$141)|0);
   HEAP32[$157>>2] = $168;
   HEAP32[$e1>>2] = 10;
   return;
   break;
  }
  case 9:  {
   $169 = (_luaK_exp2RK($fs,$e1)|0);
   $170 = (_luaK_exp2RK($fs,$e2)|0);
   $171 = HEAP32[$e2>>2]|0;
   $172 = ($171|0)==(12);
   if ($172) {
    $173 = ((($e2)) + 8|0);
    $174 = HEAP32[$173>>2]|0;
    $175 = $174 & 256;
    $176 = ($175|0)==(0);
    if ($176) {
     $177 = ((($fs)) + 50|0);
     $178 = HEAP8[$177>>0]|0;
     $179 = $178&255;
     $180 = ($179|0)>($174|0);
     if (!($180)) {
      $181 = ((($fs)) + 36|0);
      $182 = HEAP32[$181>>2]|0;
      $183 = (($182) + -1)|0;
      HEAP32[$181>>2] = $183;
     }
    }
   }
   $184 = HEAP32[$e1>>2]|0;
   $185 = ($184|0)==(12);
   $186 = ((($e1)) + 8|0);
   if ($185) {
    $187 = HEAP32[$186>>2]|0;
    $188 = $187 & 256;
    $189 = ($188|0)==(0);
    if ($189) {
     $190 = ((($fs)) + 50|0);
     $191 = HEAP8[$190>>0]|0;
     $192 = $191&255;
     $193 = ($192|0)>($187|0);
     if (!($193)) {
      $194 = ((($fs)) + 36|0);
      $195 = HEAP32[$194>>2]|0;
      $196 = (($195) + -1)|0;
      HEAP32[$194>>2] = $196;
     }
    }
   }
   $197 = (_condjump($fs,24,1,$169,$170)|0);
   HEAP32[$186>>2] = $197;
   HEAP32[$e1>>2] = 10;
   return;
   break;
  }
  case 10:  {
   $198 = (_luaK_exp2RK($fs,$e1)|0);
   $199 = (_luaK_exp2RK($fs,$e2)|0);
   $200 = HEAP32[$e2>>2]|0;
   $201 = ($200|0)==(12);
   if ($201) {
    $202 = ((($e2)) + 8|0);
    $203 = HEAP32[$202>>2]|0;
    $204 = $203 & 256;
    $205 = ($204|0)==(0);
    if ($205) {
     $206 = ((($fs)) + 50|0);
     $207 = HEAP8[$206>>0]|0;
     $208 = $207&255;
     $209 = ($208|0)>($203|0);
     if (!($209)) {
      $210 = ((($fs)) + 36|0);
      $211 = HEAP32[$210>>2]|0;
      $212 = (($211) + -1)|0;
      HEAP32[$210>>2] = $212;
     }
    }
   }
   $213 = HEAP32[$e1>>2]|0;
   $214 = ($213|0)==(12);
   $215 = ((($e1)) + 8|0);
   if ($214) {
    $216 = HEAP32[$215>>2]|0;
    $217 = $216 & 256;
    $218 = ($217|0)==(0);
    if ($218) {
     $219 = ((($fs)) + 50|0);
     $220 = HEAP8[$219>>0]|0;
     $221 = $220&255;
     $222 = ($221|0)>($216|0);
     if (!($222)) {
      $223 = ((($fs)) + 36|0);
      $224 = HEAP32[$223>>2]|0;
      $225 = (($224) + -1)|0;
      HEAP32[$223>>2] = $225;
     }
    }
   }
   $226 = (_condjump($fs,25,1,$198,$199)|0);
   HEAP32[$215>>2] = $226;
   HEAP32[$e1>>2] = 10;
   return;
   break;
  }
  case 11:  {
   $227 = (_luaK_exp2RK($fs,$e1)|0);
   $228 = (_luaK_exp2RK($fs,$e2)|0);
   $229 = HEAP32[$e2>>2]|0;
   $230 = ($229|0)==(12);
   if ($230) {
    $231 = ((($e2)) + 8|0);
    $232 = HEAP32[$231>>2]|0;
    $233 = $232 & 256;
    $234 = ($233|0)==(0);
    if ($234) {
     $235 = ((($fs)) + 50|0);
     $236 = HEAP8[$235>>0]|0;
     $237 = $236&255;
     $238 = ($237|0)>($232|0);
     if (!($238)) {
      $239 = ((($fs)) + 36|0);
      $240 = HEAP32[$239>>2]|0;
      $241 = (($240) + -1)|0;
      HEAP32[$239>>2] = $241;
     }
    }
   }
   $242 = HEAP32[$e1>>2]|0;
   $243 = ($242|0)==(12);
   $244 = ((($e1)) + 8|0);
   if ($243) {
    $245 = HEAP32[$244>>2]|0;
    $246 = $245 & 256;
    $247 = ($246|0)==(0);
    if ($247) {
     $248 = ((($fs)) + 50|0);
     $249 = HEAP8[$248>>0]|0;
     $250 = $249&255;
     $251 = ($250|0)>($245|0);
     if (!($251)) {
      $252 = ((($fs)) + 36|0);
      $253 = HEAP32[$252>>2]|0;
      $254 = (($253) + -1)|0;
      HEAP32[$252>>2] = $254;
     }
    }
   }
   $255 = (_condjump($fs,24,1,$228,$227)|0);
   HEAP32[$244>>2] = $255;
   HEAP32[$e1>>2] = 10;
   return;
   break;
  }
  case 12:  {
   $256 = (_luaK_exp2RK($fs,$e1)|0);
   $257 = (_luaK_exp2RK($fs,$e2)|0);
   $258 = HEAP32[$e2>>2]|0;
   $259 = ($258|0)==(12);
   if ($259) {
    $260 = ((($e2)) + 8|0);
    $261 = HEAP32[$260>>2]|0;
    $262 = $261 & 256;
    $263 = ($262|0)==(0);
    if ($263) {
     $264 = ((($fs)) + 50|0);
     $265 = HEAP8[$264>>0]|0;
     $266 = $265&255;
     $267 = ($266|0)>($261|0);
     if (!($267)) {
      $268 = ((($fs)) + 36|0);
      $269 = HEAP32[$268>>2]|0;
      $270 = (($269) + -1)|0;
      HEAP32[$268>>2] = $270;
     }
    }
   }
   $271 = HEAP32[$e1>>2]|0;
   $272 = ($271|0)==(12);
   $273 = ((($e1)) + 8|0);
   if ($272) {
    $274 = HEAP32[$273>>2]|0;
    $275 = $274 & 256;
    $276 = ($275|0)==(0);
    if ($276) {
     $277 = ((($fs)) + 50|0);
     $278 = HEAP8[$277>>0]|0;
     $279 = $278&255;
     $280 = ($279|0)>($274|0);
     if (!($280)) {
      $281 = ((($fs)) + 36|0);
      $282 = HEAP32[$281>>2]|0;
      $283 = (($282) + -1)|0;
      HEAP32[$281>>2] = $283;
     }
    }
   }
   $284 = (_condjump($fs,25,1,$257,$256)|0);
   HEAP32[$273>>2] = $284;
   HEAP32[$e1>>2] = 10;
   return;
   break;
  }
  default: {
   return;
  }
  }
 } while(0);
}
function _luaK_fixline($fs,$line) {
 $fs = $fs|0;
 $line = $line|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($fs)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + -1)|0;
 $3 = HEAP32[$fs>>2]|0;
 $4 = ((($3)) + 20|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($5) + ($2<<2)|0);
 HEAP32[$6>>2] = $line;
 return;
}
function _luaK_setlist($fs,$base,$nelems,$tostore) {
 $fs = $fs|0;
 $base = $base|0;
 $nelems = $nelems|0;
 $tostore = $tostore|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $tostore$op = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($nelems) + -1)|0;
 $1 = (($0|0) / 50)&-1;
 $2 = (($1) + 1)|0;
 $3 = ($tostore|0)==(-1);
 $4 = ($2|0)<(512);
 $5 = $base << 6;
 $tostore$op = $tostore << 23;
 $6 = $3 ? 0 : $tostore$op;
 $7 = $5 | $6;
 $8 = $7 | 34;
 if ($4) {
  $9 = $2 << 14;
  $10 = $8 | $9;
  $11 = ((($fs)) + 12|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ((($12)) + 8|0);
  $14 = HEAP32[$13>>2]|0;
  (_luaK_code($fs,$10,$14)|0);
  $22 = (($base) + 1)|0;
  $23 = ((($fs)) + 36|0);
  HEAP32[$23>>2] = $22;
  return;
 } else {
  $15 = ((($fs)) + 12|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ((($16)) + 8|0);
  $18 = HEAP32[$17>>2]|0;
  (_luaK_code($fs,$8,$18)|0);
  $19 = HEAP32[$15>>2]|0;
  $20 = ((($19)) + 8|0);
  $21 = HEAP32[$20>>2]|0;
  (_luaK_code($fs,$2,$21)|0);
  $22 = (($base) + 1)|0;
  $23 = ((($fs)) + 36|0);
  HEAP32[$23>>2] = $22;
  return;
 }
}
function _luaX_init($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $i$01 = 0;
 while(1) {
  $0 = (38956 + ($i$01<<2)|0);
  $1 = HEAP32[$0>>2]|0;
  $2 = (_strlen($1)|0);
  $3 = (_luaS_newlstr($L,$1,$2)|0);
  $4 = ((($3)) + 5|0);
  $5 = HEAP8[$4>>0]|0;
  $6 = $5&255;
  $7 = $6 | 32;
  $8 = $7&255;
  HEAP8[$4>>0] = $8;
  $9 = (($i$01) + 1)|0;
  $10 = $9&255;
  $11 = ((($3)) + 6|0);
  HEAP8[$11>>0] = $10;
  $exitcond = ($9|0)==(21);
  if ($exitcond) {
   break;
  } else {
   $i$01 = $9;
  }
 }
 return;
}
function _luaX_token2str($ls,$token) {
 $ls = $ls|0;
 $token = $token|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($token|0)<(257);
 if (!($0)) {
  $7 = (($token) + -257)|0;
  $8 = (38956 + ($7<<2)|0);
  $9 = HEAP32[$8>>2]|0;
  $$0 = $9;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = (_iscntrl($token)|0);
 $2 = ($1|0)==(0);
 $3 = ((($ls)) + 52|0);
 $4 = HEAP32[$3>>2]|0;
 if ($2) {
  HEAP32[$vararg_buffer1>>2] = $token;
  $6 = (_luaO_pushfstring($4,522067,$vararg_buffer1)|0);
  $$0 = $6;
  STACKTOP = sp;return ($$0|0);
 } else {
  HEAP32[$vararg_buffer>>2] = $token;
  $5 = (_luaO_pushfstring($4,519888,$vararg_buffer)|0);
  $$0 = $5;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _luaX_lexerror($ls,$msg,$token) {
 $ls = $ls|0;
 $msg = $msg|0;
 $token = $token|0;
 var $$0$i = 0, $$off$i = 0, $$pre$phiZ2D = 0, $$pre2 = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buff = 0, $buff$i = 0, $switch$i = 0, $vararg_buffer = 0, $vararg_buffer11 = 0, $vararg_buffer14 = 0, $vararg_buffer3 = 0, $vararg_buffer8 = 0, $vararg_ptr1 = 0;
 var $vararg_ptr17 = 0, $vararg_ptr2 = 0, $vararg_ptr6 = 0, $vararg_ptr7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer14 = sp + 48|0;
 $vararg_buffer11 = sp + 40|0;
 $vararg_buffer8 = sp + 32|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $buff$i = sp + 136|0;
 $buff = sp + 56|0;
 $0 = ((($ls)) + 64|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 16|0);
 _luaO_chunkid($buff,$2,80);
 $3 = ((($ls)) + 52|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($ls)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 HEAP32[$vararg_buffer>>2] = $buff;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $6;
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
 HEAP32[$vararg_ptr2>>2] = $msg;
 $7 = (_luaO_pushfstring($4,519897,$vararg_buffer)|0);
 $8 = ($token|0)==(0);
 if ($8) {
  $45 = HEAP32[$3>>2]|0;
  _luaD_throw($45,3);
  // unreachable;
 }
 $9 = HEAP32[$3>>2]|0;
 $$off$i = (($token) + -284)|0;
 $switch$i = ($$off$i>>>0)<(3);
 do {
  if ($switch$i) {
   $10 = ((($ls)) + 60|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = ((($11)) + 4|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = (($13) + 1)|0;
   $15 = ((($11)) + 8|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = ($14>>>0)>($16>>>0);
   if ($17) {
    $18 = ($16>>>0)>(2147483645);
    if ($18) {
     $19 = HEAP32[$0>>2]|0;
     $20 = ((($19)) + 16|0);
     _luaO_chunkid($buff$i,$20,80);
     $21 = HEAP32[$3>>2]|0;
     $22 = HEAP32[$5>>2]|0;
     HEAP32[$vararg_buffer3>>2] = $buff$i;
     $vararg_ptr6 = ((($vararg_buffer3)) + 4|0);
     HEAP32[$vararg_ptr6>>2] = $22;
     $vararg_ptr7 = ((($vararg_buffer3)) + 8|0);
     HEAP32[$vararg_ptr7>>2] = 519907;
     (_luaO_pushfstring($21,519897,$vararg_buffer3)|0);
     $23 = HEAP32[$3>>2]|0;
     _luaD_throw($23,3);
     // unreachable;
    }
    $24 = $16 << 1;
    $25 = ($24|0)==(-2);
    if ($25) {
     $28 = (_luaM_toobig($9)|0);
     $$pre$phiZ2D = $11;$29 = $28;
    } else {
     $26 = HEAP32[$11>>2]|0;
     $27 = (_luaM_realloc_($9,$26,$16,$24)|0);
     $$pre$phiZ2D = $11;$29 = $27;
    }
    HEAP32[$$pre$phiZ2D>>2] = $29;
    HEAP32[$15>>2] = $24;
    $$pre2 = HEAP32[$12>>2]|0;
    $31 = $$pre2;$33 = $29;
   } else {
    $$pre3 = HEAP32[$11>>2]|0;
    $31 = $13;$33 = $$pre3;
   }
   $30 = (($31) + 1)|0;
   HEAP32[$12>>2] = $30;
   $32 = (($33) + ($31)|0);
   HEAP8[$32>>0] = 0;
   $34 = HEAP32[$10>>2]|0;
   $35 = HEAP32[$34>>2]|0;
   $$0$i = $35;
  } else {
   $36 = ($token|0)<(257);
   if (!($36)) {
    $42 = (($token) + -257)|0;
    $43 = (38956 + ($42<<2)|0);
    $44 = HEAP32[$43>>2]|0;
    $$0$i = $44;
    break;
   }
   $37 = (_iscntrl($token)|0);
   $38 = ($37|0)==(0);
   $39 = HEAP32[$3>>2]|0;
   if ($38) {
    HEAP32[$vararg_buffer11>>2] = $token;
    $41 = (_luaO_pushfstring($39,522067,$vararg_buffer11)|0);
    $$0$i = $41;
    break;
   } else {
    HEAP32[$vararg_buffer8>>2] = $token;
    $40 = (_luaO_pushfstring($39,519888,$vararg_buffer8)|0);
    $$0$i = $40;
    break;
   }
  }
 } while(0);
 HEAP32[$vararg_buffer14>>2] = $7;
 $vararg_ptr17 = ((($vararg_buffer14)) + 4|0);
 HEAP32[$vararg_ptr17>>2] = $$0$i;
 (_luaO_pushfstring($9,519932,$vararg_buffer14)|0);
 $45 = HEAP32[$3>>2]|0;
 _luaD_throw($45,3);
 // unreachable;
}
function _luaX_syntaxerror($ls,$msg) {
 $ls = $ls|0;
 $msg = $msg|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($ls)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 _luaX_lexerror($ls,$msg,$1);
 return;
}
function _luaX_newstring($ls,$str,$l) {
 $ls = $ls|0;
 $str = $str|0;
 $l = $l|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($ls)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_luaS_newlstr($1,$str,$l)|0);
 $3 = ((($ls)) + 48|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (_luaH_setstr($1,$6,$2)|0);
 $8 = ((($7)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)==(0);
 if (!($10)) {
  return ($2|0);
 }
 HEAP32[$7>>2] = 1;
 HEAP32[$8>>2] = 1;
 $11 = ((($1)) + 16|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($12)) + 68|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($12)) + 64|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ($14>>>0)<($16>>>0);
 if ($17) {
  return ($2|0);
 }
 _luaC_step($1);
 return ($2|0);
}
function _luaX_setinput($L,$ls,$z,$source) {
 $L = $L|0;
 $ls = $ls|0;
 $z = $z|0;
 $source = $source|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($ls)) + 68|0);
 HEAP8[$0>>0] = 46;
 $1 = ((($ls)) + 52|0);
 HEAP32[$1>>2] = $L;
 $2 = ((($ls)) + 32|0);
 HEAP32[$2>>2] = 287;
 $3 = ((($ls)) + 56|0);
 HEAP32[$3>>2] = $z;
 $4 = ((($ls)) + 48|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($ls)) + 4|0);
 HEAP32[$5>>2] = 1;
 $6 = ((($ls)) + 8|0);
 HEAP32[$6>>2] = 1;
 $7 = ((($ls)) + 64|0);
 HEAP32[$7>>2] = $source;
 $8 = ((($ls)) + 60|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($9)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (_luaM_realloc_($L,$10,$12,32)|0);
 $14 = HEAP32[$8>>2]|0;
 HEAP32[$14>>2] = $13;
 $15 = ((($14)) + 8|0);
 HEAP32[$15>>2] = 32;
 $16 = HEAP32[$3>>2]|0;
 $17 = HEAP32[$16>>2]|0;
 $18 = (($17) + -1)|0;
 HEAP32[$16>>2] = $18;
 $19 = ($17|0)==(0);
 if ($19) {
  $25 = (_luaZ_fill($16)|0);
  $26 = $25;
  HEAP32[$ls>>2] = $26;
  return;
 } else {
  $20 = ((($16)) + 4|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = ((($21)) + 1|0);
  HEAP32[$20>>2] = $22;
  $23 = HEAP8[$21>>0]|0;
  $24 = $23&255;
  $26 = $24;
  HEAP32[$ls>>2] = $26;
  return;
 }
}
function _luaX_next($ls) {
 $ls = $ls|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($ls)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($ls)) + 8|0);
 HEAP32[$2>>2] = $1;
 $3 = ((($ls)) + 32|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(287);
 if ($5) {
  $7 = ((($ls)) + 24|0);
  $8 = (_llex($ls,$7)|0);
  $9 = ((($ls)) + 16|0);
  HEAP32[$9>>2] = $8;
  return;
 } else {
  $6 = ((($ls)) + 16|0);
  ;HEAP32[$6>>2]=HEAP32[$3>>2]|0;HEAP32[$6+4>>2]=HEAP32[$3+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$3+8>>2]|0;HEAP32[$6+12>>2]=HEAP32[$3+12>>2]|0;
  HEAP32[$3>>2] = 287;
  return;
 }
}
function _llex($ls,$seminfo) {
 $ls = $ls|0;
 $seminfo = $seminfo|0;
 var $$2 = 0, $$be$i = 0, $$lcssa = 0, $$lcssa115 = 0, $$lcssa122 = 0, $$pr = 0, $$pr$i = 0, $$pr78 = 0, $$pre = 0, $$pre$i = 0, $$pre$phi$i$iZ2D = 0, $$pre$phi$i10Z2D = 0, $$pre$phi$i16$iZ2D = 0, $$pre$phi$i18Z2D = 0, $$pre$phi$i24$iZ2D = 0, $$pre$phi$i27Z2D = 0, $$pre$phi$i32$iZ2D = 0, $$pre$phi$i35Z2D = 0, $$pre$phi$i40$iZ2D = 0, $$pre$phi$i48$iZ2D = 0;
 var $$pre$phi$i8$iZ2D = 0, $$pre1$i = 0, $$pre1$i$i = 0, $$pre1$i17$i = 0, $$pre1$i19 = 0, $$pre1$i25$i = 0, $$pre1$i28 = 0, $$pre1$i33$i = 0, $$pre1$i36 = 0, $$pre1$i41$i = 0, $$pre1$i49$i = 0, $$pre1$i9$i = 0, $$pre2$i = 0, $$pre2$i$i = 0, $$pre2$i12$i = 0, $$pre2$i14 = 0, $$pre2$i20$i = 0, $$pre2$i23 = 0, $$pre2$i28$i = 0, $$pre2$i31 = 0;
 var $$pre2$i36$i = 0, $$pre2$i4$i = 0, $$pre2$i44$i = 0, $$pre61$i = 0, $$pre74 = 0, $$pre76 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0;
 var $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0;
 var $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0;
 var $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0;
 var $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0;
 var $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0;
 var $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0;
 var $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0;
 var $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0;
 var $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0;
 var $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0;
 var $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0;
 var $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0;
 var $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0;
 var $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0;
 var $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0;
 var $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0;
 var $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0;
 var $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0;
 var $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0;
 var $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0;
 var $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0;
 var $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0;
 var $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0;
 var $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0;
 var $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0;
 var $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0;
 var $705 = 0, $706 = 0, $707 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $buff$i = 0, $buff$i$i = 0, $buff$i39 = 0, $buff$i41 = 0, $buff$i43 = 0, $buff$i51$i = 0;
 var $buff$i52$i = 0, $buff$i53$i = 0, $buff$i54$i = 0, $buff$i55$i = 0, $buff$i56$i = 0, $buff$i57$i = 0, $c$0$i = 0, $c$1$i = 0, $i$0$i = 0, $isdigit = 0, $isdigit$i = 0, $isdigit2$i = 0, $isdigit7 = 0, $isdigittmp = 0, $isdigittmp$i = 0, $isdigittmp1$i = 0, $isdigittmp6 = 0, $memchr = 0, $memchr44 = 0, $or$cond$i = 0;
 var $vararg_buffer = 0, $vararg_buffer12 = 0, $vararg_buffer17 = 0, $vararg_buffer22 = 0, $vararg_buffer27 = 0, $vararg_buffer3 = 0, $vararg_buffer32 = 0, $vararg_buffer37 = 0, $vararg_buffer42 = 0, $vararg_buffer47 = 0, $vararg_buffer52 = 0, $vararg_buffer57 = 0, $vararg_buffer8 = 0, $vararg_ptr1 = 0, $vararg_ptr11 = 0, $vararg_ptr15 = 0, $vararg_ptr16 = 0, $vararg_ptr2 = 0, $vararg_ptr20 = 0, $vararg_ptr21 = 0;
 var $vararg_ptr25 = 0, $vararg_ptr26 = 0, $vararg_ptr30 = 0, $vararg_ptr31 = 0, $vararg_ptr35 = 0, $vararg_ptr36 = 0, $vararg_ptr40 = 0, $vararg_ptr41 = 0, $vararg_ptr45 = 0, $vararg_ptr46 = 0, $vararg_ptr50 = 0, $vararg_ptr51 = 0, $vararg_ptr55 = 0, $vararg_ptr56 = 0, $vararg_ptr6 = 0, $vararg_ptr60 = 0, $vararg_ptr61 = 0, $vararg_ptr7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1168|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer57 = sp + 184|0;
 $vararg_buffer52 = sp + 168|0;
 $vararg_buffer47 = sp + 152|0;
 $vararg_buffer42 = sp + 136|0;
 $vararg_buffer37 = sp + 120|0;
 $vararg_buffer32 = sp + 104|0;
 $vararg_buffer27 = sp + 88|0;
 $vararg_buffer22 = sp + 72|0;
 $vararg_buffer17 = sp + 56|0;
 $vararg_buffer12 = sp + 40|0;
 $vararg_buffer8 = sp + 32|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $buff$i43 = sp + 1080|0;
 $buff$i41 = sp + 1000|0;
 $buff$i39 = sp + 920|0;
 $buff$i = sp + 840|0;
 $buff$i57$i = sp + 760|0;
 $buff$i56$i = sp + 680|0;
 $buff$i55$i = sp + 600|0;
 $buff$i54$i = sp + 520|0;
 $buff$i53$i = sp + 440|0;
 $buff$i52$i = sp + 360|0;
 $buff$i51$i = sp + 280|0;
 $buff$i$i = sp + 200|0;
 $0 = ((($ls)) + 60|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 4|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($ls)) + 56|0);
 $4 = ((($ls)) + 4|0);
 $5 = ((($ls)) + 16|0);
 L1: while(1) {
  $$pr78 = HEAP32[$ls>>2]|0;
  $6 = $$pr78;
  L3: while(1) {
   switch ($6|0) {
   case 91:  {
    label = 31;
    break L1;
    break;
   }
   case 61:  {
    break L1;
    break;
   }
   case 60:  {
    label = 43;
    break L1;
    break;
   }
   case 62:  {
    label = 51;
    break L1;
    break;
   }
   case 126:  {
    label = 59;
    break L1;
    break;
   }
   case 39: case 34:  {
    $$lcssa122 = $6;
    label = 67;
    break L1;
    break;
   }
   case 46:  {
    label = 181;
    break L1;
    break;
   }
   case -1:  {
    $$2 = 287;
    label = 252;
    break L1;
    break;
   }
   case 13: case 10:  {
    $$lcssa115 = $6;
    label = 4;
    break L3;
    break;
   }
   case 45:  {
    break L3;
    break;
   }
   default: {
   }
   }
   $610 = (_isspace($6)|0);
   $611 = ($610|0)==(0);
   if ($611) {
    label = 224;
    break L1;
   }
   $612 = HEAP32[$3>>2]|0;
   $613 = HEAP32[$612>>2]|0;
   $614 = (($613) + -1)|0;
   HEAP32[$612>>2] = $614;
   $615 = ($613|0)==(0);
   if ($615) {
    $621 = (_luaZ_fill($612)|0);
    $622 = $621;
   } else {
    $616 = ((($612)) + 4|0);
    $617 = HEAP32[$616>>2]|0;
    $618 = ((($617)) + 1|0);
    HEAP32[$616>>2] = $618;
    $619 = HEAP8[$617>>0]|0;
    $620 = $619&255;
    $622 = $620;
   }
   HEAP32[$ls>>2] = $622;
   $6 = $622;
  }
  if ((label|0) == 4) {
   label = 0;
   $7 = HEAP32[$3>>2]|0;
   $8 = HEAP32[$7>>2]|0;
   $9 = (($8) + -1)|0;
   HEAP32[$7>>2] = $9;
   $10 = ($8|0)==(0);
   if ($10) {
    $16 = (_luaZ_fill($7)|0);
    $17 = $16;
   } else {
    $11 = ((($7)) + 4|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = ((($12)) + 1|0);
    HEAP32[$11>>2] = $13;
    $14 = HEAP8[$12>>0]|0;
    $15 = $14&255;
    $17 = $15;
   }
   HEAP32[$ls>>2] = $17;
   switch ($17|0) {
   case 13: case 10:  {
    $18 = ($17|0)==($$lcssa115|0);
    if (!($18)) {
     $19 = HEAP32[$3>>2]|0;
     $20 = HEAP32[$19>>2]|0;
     $21 = (($20) + -1)|0;
     HEAP32[$19>>2] = $21;
     $22 = ($20|0)==(0);
     if ($22) {
      $28 = (_luaZ_fill($19)|0);
      $29 = $28;
     } else {
      $23 = ((($19)) + 4|0);
      $24 = HEAP32[$23>>2]|0;
      $25 = ((($24)) + 1|0);
      HEAP32[$23>>2] = $25;
      $26 = HEAP8[$24>>0]|0;
      $27 = $26&255;
      $29 = $27;
     }
     HEAP32[$ls>>2] = $29;
    }
    break;
   }
   default: {
   }
   }
   $30 = HEAP32[$4>>2]|0;
   $31 = (($30) + 1)|0;
   HEAP32[$4>>2] = $31;
   $32 = ($30|0)>(2147483643);
   if (!($32)) {
    continue;
   }
   $33 = HEAP32[$5>>2]|0;
   _luaX_lexerror($ls,519945,$33);
   continue;
  }
  $34 = HEAP32[$3>>2]|0;
  $35 = HEAP32[$34>>2]|0;
  $36 = (($35) + -1)|0;
  HEAP32[$34>>2] = $36;
  $37 = ($35|0)==(0);
  if ($37) {
   $43 = (_luaZ_fill($34)|0);
   $44 = $43;
  } else {
   $38 = ((($34)) + 4|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = ((($39)) + 1|0);
   HEAP32[$38>>2] = $40;
   $41 = HEAP8[$39>>0]|0;
   $42 = $41&255;
   $44 = $42;
  }
  HEAP32[$ls>>2] = $44;
  $45 = ($44|0)==(45);
  if (!($45)) {
   $$2 = 45;
   label = 252;
   break;
  }
  $46 = HEAP32[$3>>2]|0;
  $47 = HEAP32[$46>>2]|0;
  $48 = (($47) + -1)|0;
  HEAP32[$46>>2] = $48;
  $49 = ($47|0)==(0);
  if ($49) {
   $55 = (_luaZ_fill($46)|0);
   $56 = $55;
  } else {
   $50 = ((($46)) + 4|0);
   $51 = HEAP32[$50>>2]|0;
   $52 = ((($51)) + 1|0);
   HEAP32[$50>>2] = $52;
   $53 = HEAP8[$51>>0]|0;
   $54 = $53&255;
   $56 = $54;
  }
  HEAP32[$ls>>2] = $56;
  $57 = ($56|0)==(91);
  do {
   if ($57) {
    $58 = (_skip_sep($ls)|0);
    $59 = HEAP32[$0>>2]|0;
    $60 = ((($59)) + 4|0);
    HEAP32[$60>>2] = 0;
    $61 = ($58|0)>(-1);
    if ($61) {
     _read_long_string($ls,0,$58);
     $62 = HEAP32[$0>>2]|0;
     $63 = ((($62)) + 4|0);
     HEAP32[$63>>2] = 0;
     continue L1;
    } else {
     $$pr = HEAP32[$ls>>2]|0;
     $64 = $$pr;
     break;
    }
   } else {
    $64 = $56;
   }
  } while(0);
  while(1) {
   switch ($64|0) {
   case -1: case 13: case 10:  {
    continue L1;
    break;
   }
   default: {
   }
   }
   $65 = HEAP32[$3>>2]|0;
   $66 = HEAP32[$65>>2]|0;
   $67 = (($66) + -1)|0;
   HEAP32[$65>>2] = $67;
   $68 = ($66|0)==(0);
   if ($68) {
    $74 = (_luaZ_fill($65)|0);
    $75 = $74;
   } else {
    $69 = ((($65)) + 4|0);
    $70 = HEAP32[$69>>2]|0;
    $71 = ((($70)) + 1|0);
    HEAP32[$69>>2] = $71;
    $72 = HEAP8[$70>>0]|0;
    $73 = $72&255;
    $75 = $73;
   }
   HEAP32[$ls>>2] = $75;
   $64 = $75;
  }
 }
 do {
  if ((label|0) == 31) {
   $76 = (_skip_sep($ls)|0);
   $77 = ($76|0)>(-1);
   if ($77) {
    _read_long_string($ls,$seminfo,$76);
    $$2 = 286;
    STACKTOP = sp;return ($$2|0);
   }
   $78 = ($76|0)==(-1);
   if ($78) {
    $$2 = 91;
    STACKTOP = sp;return ($$2|0);
   } else {
    _luaX_lexerror($ls,520050,286);
    break;
   }
  }
  else if ((label|0) == 43) {
   $102 = HEAP32[$3>>2]|0;
   $103 = HEAP32[$102>>2]|0;
   $104 = (($103) + -1)|0;
   HEAP32[$102>>2] = $104;
   $105 = ($103|0)==(0);
   if ($105) {
    $111 = (_luaZ_fill($102)|0);
    $112 = $111;
   } else {
    $106 = ((($102)) + 4|0);
    $107 = HEAP32[$106>>2]|0;
    $108 = ((($107)) + 1|0);
    HEAP32[$106>>2] = $108;
    $109 = HEAP8[$107>>0]|0;
    $110 = $109&255;
    $112 = $110;
   }
   HEAP32[$ls>>2] = $112;
   $113 = ($112|0)==(61);
   if (!($113)) {
    $$2 = 60;
    STACKTOP = sp;return ($$2|0);
   }
   $114 = HEAP32[$3>>2]|0;
   $115 = HEAP32[$114>>2]|0;
   $116 = (($115) + -1)|0;
   HEAP32[$114>>2] = $116;
   $117 = ($115|0)==(0);
   if ($117) {
    $123 = (_luaZ_fill($114)|0);
    $124 = $123;
   } else {
    $118 = ((($114)) + 4|0);
    $119 = HEAP32[$118>>2]|0;
    $120 = ((($119)) + 1|0);
    HEAP32[$118>>2] = $120;
    $121 = HEAP8[$119>>0]|0;
    $122 = $121&255;
    $124 = $122;
   }
   HEAP32[$ls>>2] = $124;
   $$2 = 282;
   STACKTOP = sp;return ($$2|0);
  }
  else if ((label|0) == 51) {
   $125 = HEAP32[$3>>2]|0;
   $126 = HEAP32[$125>>2]|0;
   $127 = (($126) + -1)|0;
   HEAP32[$125>>2] = $127;
   $128 = ($126|0)==(0);
   if ($128) {
    $134 = (_luaZ_fill($125)|0);
    $135 = $134;
   } else {
    $129 = ((($125)) + 4|0);
    $130 = HEAP32[$129>>2]|0;
    $131 = ((($130)) + 1|0);
    HEAP32[$129>>2] = $131;
    $132 = HEAP8[$130>>0]|0;
    $133 = $132&255;
    $135 = $133;
   }
   HEAP32[$ls>>2] = $135;
   $136 = ($135|0)==(61);
   if (!($136)) {
    $$2 = 62;
    STACKTOP = sp;return ($$2|0);
   }
   $137 = HEAP32[$3>>2]|0;
   $138 = HEAP32[$137>>2]|0;
   $139 = (($138) + -1)|0;
   HEAP32[$137>>2] = $139;
   $140 = ($138|0)==(0);
   if ($140) {
    $146 = (_luaZ_fill($137)|0);
    $147 = $146;
   } else {
    $141 = ((($137)) + 4|0);
    $142 = HEAP32[$141>>2]|0;
    $143 = ((($142)) + 1|0);
    HEAP32[$141>>2] = $143;
    $144 = HEAP8[$142>>0]|0;
    $145 = $144&255;
    $147 = $145;
   }
   HEAP32[$ls>>2] = $147;
   $$2 = 281;
   STACKTOP = sp;return ($$2|0);
  }
  else if ((label|0) == 59) {
   $148 = HEAP32[$3>>2]|0;
   $149 = HEAP32[$148>>2]|0;
   $150 = (($149) + -1)|0;
   HEAP32[$148>>2] = $150;
   $151 = ($149|0)==(0);
   if ($151) {
    $157 = (_luaZ_fill($148)|0);
    $158 = $157;
   } else {
    $152 = ((($148)) + 4|0);
    $153 = HEAP32[$152>>2]|0;
    $154 = ((($153)) + 1|0);
    HEAP32[$152>>2] = $154;
    $155 = HEAP8[$153>>0]|0;
    $156 = $155&255;
    $158 = $156;
   }
   HEAP32[$ls>>2] = $158;
   $159 = ($158|0)==(61);
   if (!($159)) {
    $$2 = 126;
    STACKTOP = sp;return ($$2|0);
   }
   $160 = HEAP32[$3>>2]|0;
   $161 = HEAP32[$160>>2]|0;
   $162 = (($161) + -1)|0;
   HEAP32[$160>>2] = $162;
   $163 = ($161|0)==(0);
   if ($163) {
    $169 = (_luaZ_fill($160)|0);
    $170 = $169;
   } else {
    $164 = ((($160)) + 4|0);
    $165 = HEAP32[$164>>2]|0;
    $166 = ((($165)) + 1|0);
    HEAP32[$164>>2] = $166;
    $167 = HEAP8[$165>>0]|0;
    $168 = $167&255;
    $170 = $168;
   }
   HEAP32[$ls>>2] = $170;
   $$2 = 283;
   STACKTOP = sp;return ($$2|0);
  }
  else if ((label|0) == 67) {
   $171 = HEAP32[$0>>2]|0;
   $172 = ((($171)) + 4|0);
   $173 = HEAP32[$172>>2]|0;
   $174 = (($173) + 1)|0;
   $175 = ((($171)) + 8|0);
   $176 = HEAP32[$175>>2]|0;
   $177 = ($174>>>0)>($176>>>0);
   if ($177) {
    $178 = ($176>>>0)>(2147483645);
    if ($178) {
     $182 = ((($ls)) + 64|0);
     $183 = HEAP32[$182>>2]|0;
     $184 = ((($183)) + 16|0);
     _luaO_chunkid($buff$i51$i,$184,80);
     $185 = ((($ls)) + 52|0);
     $186 = HEAP32[$185>>2]|0;
     $187 = HEAP32[$4>>2]|0;
     HEAP32[$vararg_buffer>>2] = $buff$i51$i;
     $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
     HEAP32[$vararg_ptr1>>2] = $187;
     $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
     HEAP32[$vararg_ptr2>>2] = 519907;
     (_luaO_pushfstring($186,519897,$vararg_buffer)|0);
     $188 = HEAP32[$185>>2]|0;
     _luaD_throw($188,3);
     // unreachable;
    }
    $$pre$i = ((($ls)) + 52|0);
    $179 = $176 << 1;
    $180 = ($179|0)==(-2);
    $181 = HEAP32[$$pre$i>>2]|0;
    if ($180) {
     $191 = (_luaM_toobig($181)|0);
     $$pre$phi$i$iZ2D = $171;$192 = $191;
    } else {
     $189 = HEAP32[$171>>2]|0;
     $190 = (_luaM_realloc_($181,$189,$176,$179)|0);
     $$pre$phi$i$iZ2D = $171;$192 = $190;
    }
    HEAP32[$$pre$phi$i$iZ2D>>2] = $192;
    HEAP32[$175>>2] = $179;
    $$pre1$i$i = HEAP32[$172>>2]|0;
    $195 = $$pre1$i$i;$197 = $192;
   } else {
    $$pre2$i$i = HEAP32[$171>>2]|0;
    $195 = $173;$197 = $$pre2$i$i;
   }
   $193 = $$lcssa122&255;
   $194 = (($195) + 1)|0;
   HEAP32[$172>>2] = $194;
   $196 = (($197) + ($195)|0);
   HEAP8[$196>>0] = $193;
   $198 = HEAP32[$3>>2]|0;
   $199 = HEAP32[$198>>2]|0;
   $200 = (($199) + -1)|0;
   HEAP32[$198>>2] = $200;
   $201 = ($199|0)==(0);
   if ($201) {
    $207 = (_luaZ_fill($198)|0);
    $208 = $207;
   } else {
    $202 = ((($198)) + 4|0);
    $203 = HEAP32[$202>>2]|0;
    $204 = ((($203)) + 1|0);
    HEAP32[$202>>2] = $204;
    $205 = HEAP8[$203>>0]|0;
    $206 = $205&255;
    $208 = $206;
   }
   HEAP32[$ls>>2] = $208;
   $209 = ($208|0)==($$lcssa122|0);
   L112: do {
    if (!($209)) {
     $210 = ((($ls)) + 64|0);
     $211 = ((($ls)) + 52|0);
     $212 = $208;
     L114: while(1) {
      L116: do {
       switch ($212|0) {
       case -1:  {
        label = 82;
        break L114;
        break;
       }
       case 13: case 10:  {
        _luaX_lexerror($ls,520080,286);
        label = 80;
        break;
       }
       case 92:  {
        $220 = HEAP32[$3>>2]|0;
        $221 = HEAP32[$220>>2]|0;
        $222 = (($221) + -1)|0;
        HEAP32[$220>>2] = $222;
        $223 = ($221|0)==(0);
        if ($223) {
         $229 = (_luaZ_fill($220)|0);
         $230 = $229;
        } else {
         $224 = ((($220)) + 4|0);
         $225 = HEAP32[$224>>2]|0;
         $226 = ((($225)) + 1|0);
         HEAP32[$224>>2] = $226;
         $227 = HEAP8[$225>>0]|0;
         $228 = $227&255;
         $230 = $228;
        }
        HEAP32[$ls>>2] = $230;
        switch ($230|0) {
        case -1:  {
         $$be$i = -1;
         break L116;
         break;
        }
        case 97:  {
         $c$1$i = 7;
         break;
        }
        case 98:  {
         $c$1$i = 8;
         break;
        }
        case 102:  {
         $c$1$i = 12;
         break;
        }
        case 110:  {
         $c$1$i = 10;
         break;
        }
        case 114:  {
         $c$1$i = 13;
         break;
        }
        case 116:  {
         $c$1$i = 9;
         break;
        }
        case 118:  {
         $c$1$i = 11;
         break;
        }
        case 13: case 10:  {
         $231 = HEAP32[$0>>2]|0;
         $232 = ((($231)) + 4|0);
         $233 = HEAP32[$232>>2]|0;
         $234 = (($233) + 1)|0;
         $235 = ((($231)) + 8|0);
         $236 = HEAP32[$235>>2]|0;
         $237 = ($234>>>0)>($236>>>0);
         if ($237) {
          $238 = ($236>>>0)>(2147483645);
          if ($238) {
           label = 97;
           break L114;
          }
          $244 = $236 << 1;
          $245 = ($244|0)==(-2);
          $246 = HEAP32[$211>>2]|0;
          if ($245) {
           $249 = (_luaM_toobig($246)|0);
           $$pre$phi$i8$iZ2D = $231;$250 = $249;
          } else {
           $247 = HEAP32[$231>>2]|0;
           $248 = (_luaM_realloc_($246,$247,$236,$244)|0);
           $$pre$phi$i8$iZ2D = $231;$250 = $248;
          }
          HEAP32[$$pre$phi$i8$iZ2D>>2] = $250;
          HEAP32[$235>>2] = $244;
          $$pre1$i9$i = HEAP32[$232>>2]|0;
          $252 = $$pre1$i9$i;$254 = $250;
         } else {
          $$pre2$i4$i = HEAP32[$231>>2]|0;
          $252 = $233;$254 = $$pre2$i4$i;
         }
         $251 = (($252) + 1)|0;
         HEAP32[$232>>2] = $251;
         $253 = (($254) + ($252)|0);
         HEAP8[$253>>0] = 10;
         $255 = HEAP32[$ls>>2]|0;
         $256 = HEAP32[$3>>2]|0;
         $257 = HEAP32[$256>>2]|0;
         $258 = (($257) + -1)|0;
         HEAP32[$256>>2] = $258;
         $259 = ($257|0)==(0);
         if ($259) {
          $265 = (_luaZ_fill($256)|0);
          $266 = $265;
         } else {
          $260 = ((($256)) + 4|0);
          $261 = HEAP32[$260>>2]|0;
          $262 = ((($261)) + 1|0);
          HEAP32[$260>>2] = $262;
          $263 = HEAP8[$261>>0]|0;
          $264 = $263&255;
          $266 = $264;
         }
         HEAP32[$ls>>2] = $266;
         switch ($266|0) {
         case 13: case 10:  {
          $267 = ($266|0)==($255|0);
          if (!($267)) {
           $268 = HEAP32[$3>>2]|0;
           $269 = HEAP32[$268>>2]|0;
           $270 = (($269) + -1)|0;
           HEAP32[$268>>2] = $270;
           $271 = ($269|0)==(0);
           if ($271) {
            $277 = (_luaZ_fill($268)|0);
            $278 = $277;
           } else {
            $272 = ((($268)) + 4|0);
            $273 = HEAP32[$272>>2]|0;
            $274 = ((($273)) + 1|0);
            HEAP32[$272>>2] = $274;
            $275 = HEAP8[$273>>0]|0;
            $276 = $275&255;
            $278 = $276;
           }
           HEAP32[$ls>>2] = $278;
          }
          break;
         }
         default: {
         }
         }
         $279 = HEAP32[$4>>2]|0;
         $280 = (($279) + 1)|0;
         HEAP32[$4>>2] = $280;
         $281 = ($279|0)>(2147483643);
         if (!($281)) {
          label = 80;
          break L116;
         }
         $282 = HEAP32[$5>>2]|0;
         _luaX_lexerror($ls,519945,$282);
         label = 80;
         break L116;
         break;
        }
        default: {
         $isdigittmp$i = (($230) + -48)|0;
         $isdigit$i = ($isdigittmp$i>>>0)<(10);
         if ($isdigit$i) {
          $321 = $230;$c$0$i = 0;$i$0$i = 0;
         } else {
          $283 = HEAP32[$0>>2]|0;
          $284 = ((($283)) + 4|0);
          $285 = HEAP32[$284>>2]|0;
          $286 = (($285) + 1)|0;
          $287 = ((($283)) + 8|0);
          $288 = HEAP32[$287>>2]|0;
          $289 = ($286>>>0)>($288>>>0);
          if ($289) {
           $290 = ($288>>>0)>(2147483645);
           if ($290) {
            label = 117;
            break L114;
           }
           $296 = $288 << 1;
           $297 = ($296|0)==(-2);
           $298 = HEAP32[$211>>2]|0;
           if ($297) {
            $301 = (_luaM_toobig($298)|0);
            $$pre$phi$i16$iZ2D = $283;$302 = $301;
           } else {
            $299 = HEAP32[$283>>2]|0;
            $300 = (_luaM_realloc_($298,$299,$288,$296)|0);
            $$pre$phi$i16$iZ2D = $283;$302 = $300;
           }
           HEAP32[$$pre$phi$i16$iZ2D>>2] = $302;
           HEAP32[$287>>2] = $296;
           $$pre1$i17$i = HEAP32[$284>>2]|0;
           $305 = $$pre1$i17$i;$307 = $302;
          } else {
           $$pre2$i12$i = HEAP32[$283>>2]|0;
           $305 = $285;$307 = $$pre2$i12$i;
          }
          $303 = $230&255;
          $304 = (($305) + 1)|0;
          HEAP32[$284>>2] = $304;
          $306 = (($307) + ($305)|0);
          HEAP8[$306>>0] = $303;
          $308 = HEAP32[$3>>2]|0;
          $309 = HEAP32[$308>>2]|0;
          $310 = (($309) + -1)|0;
          HEAP32[$308>>2] = $310;
          $311 = ($309|0)==(0);
          if ($311) {
           $317 = (_luaZ_fill($308)|0);
           $318 = $317;
          } else {
           $312 = ((($308)) + 4|0);
           $313 = HEAP32[$312>>2]|0;
           $314 = ((($313)) + 1|0);
           HEAP32[$312>>2] = $314;
           $315 = HEAP8[$313>>0]|0;
           $316 = $315&255;
           $318 = $316;
          }
          HEAP32[$ls>>2] = $318;
          $$be$i = $318;
          break L116;
         }
         while(1) {
          $319 = ($c$0$i*10)|0;
          $320 = (($321) + -48)|0;
          $322 = (($320) + ($319))|0;
          $323 = HEAP32[$3>>2]|0;
          $324 = HEAP32[$323>>2]|0;
          $325 = (($324) + -1)|0;
          HEAP32[$323>>2] = $325;
          $326 = ($324|0)==(0);
          if ($326) {
           $332 = (_luaZ_fill($323)|0);
           $333 = $332;
          } else {
           $327 = ((($323)) + 4|0);
           $328 = HEAP32[$327>>2]|0;
           $329 = ((($328)) + 1|0);
           HEAP32[$327>>2] = $329;
           $330 = HEAP8[$328>>0]|0;
           $331 = $330&255;
           $333 = $331;
          }
          HEAP32[$ls>>2] = $333;
          $334 = (($i$0$i) + 1)|0;
          $335 = ($334|0)<(3);
          $isdigittmp1$i = (($333) + -48)|0;
          $isdigit2$i = ($isdigittmp1$i>>>0)<(10);
          $or$cond$i = $335 & $isdigit2$i;
          if ($or$cond$i) {
           $321 = $333;$c$0$i = $322;$i$0$i = $334;
          } else {
           $$lcssa = $322;
           break;
          }
         }
         $336 = ($$lcssa|0)>(255);
         if ($336) {
          _luaX_lexerror($ls,520098,286);
         }
         $337 = HEAP32[$0>>2]|0;
         $338 = ((($337)) + 4|0);
         $339 = HEAP32[$338>>2]|0;
         $340 = (($339) + 1)|0;
         $341 = ((($337)) + 8|0);
         $342 = HEAP32[$341>>2]|0;
         $343 = ($340>>>0)>($342>>>0);
         if ($343) {
          $344 = ($342>>>0)>(2147483645);
          if ($344) {
           label = 135;
           break L114;
          }
          $350 = $342 << 1;
          $351 = ($350|0)==(-2);
          $352 = HEAP32[$211>>2]|0;
          if ($351) {
           $355 = (_luaM_toobig($352)|0);
           $$pre$phi$i24$iZ2D = $337;$356 = $355;
          } else {
           $353 = HEAP32[$337>>2]|0;
           $354 = (_luaM_realloc_($352,$353,$342,$350)|0);
           $$pre$phi$i24$iZ2D = $337;$356 = $354;
          }
          HEAP32[$$pre$phi$i24$iZ2D>>2] = $356;
          HEAP32[$341>>2] = $350;
          $$pre1$i25$i = HEAP32[$338>>2]|0;
          $359 = $$pre1$i25$i;$361 = $356;
         } else {
          $$pre2$i20$i = HEAP32[$337>>2]|0;
          $359 = $339;$361 = $$pre2$i20$i;
         }
         $357 = $$lcssa&255;
         $358 = (($359) + 1)|0;
         HEAP32[$338>>2] = $358;
         $360 = (($361) + ($359)|0);
         HEAP8[$360>>0] = $357;
         label = 80;
         break L116;
        }
        }
        $362 = HEAP32[$0>>2]|0;
        $363 = ((($362)) + 4|0);
        $364 = HEAP32[$363>>2]|0;
        $365 = (($364) + 1)|0;
        $366 = ((($362)) + 8|0);
        $367 = HEAP32[$366>>2]|0;
        $368 = ($365>>>0)>($367>>>0);
        if ($368) {
         $369 = ($367>>>0)>(2147483645);
         if ($369) {
          label = 144;
          break L114;
         }
         $375 = $367 << 1;
         $376 = ($375|0)==(-2);
         $377 = HEAP32[$211>>2]|0;
         if ($376) {
          $380 = (_luaM_toobig($377)|0);
          $$pre$phi$i32$iZ2D = $362;$381 = $380;
         } else {
          $378 = HEAP32[$362>>2]|0;
          $379 = (_luaM_realloc_($377,$378,$367,$375)|0);
          $$pre$phi$i32$iZ2D = $362;$381 = $379;
         }
         HEAP32[$$pre$phi$i32$iZ2D>>2] = $381;
         HEAP32[$366>>2] = $375;
         $$pre1$i33$i = HEAP32[$363>>2]|0;
         $383 = $$pre1$i33$i;$385 = $381;
        } else {
         $$pre2$i28$i = HEAP32[$362>>2]|0;
         $383 = $364;$385 = $$pre2$i28$i;
        }
        $382 = (($383) + 1)|0;
        HEAP32[$363>>2] = $382;
        $384 = (($385) + ($383)|0);
        HEAP8[$384>>0] = $c$1$i;
        $386 = HEAP32[$3>>2]|0;
        $387 = HEAP32[$386>>2]|0;
        $388 = (($387) + -1)|0;
        HEAP32[$386>>2] = $388;
        $389 = ($387|0)==(0);
        if ($389) {
         $395 = (_luaZ_fill($386)|0);
         $396 = $395;
        } else {
         $390 = ((($386)) + 4|0);
         $391 = HEAP32[$390>>2]|0;
         $392 = ((($391)) + 1|0);
         HEAP32[$390>>2] = $392;
         $393 = HEAP8[$391>>0]|0;
         $394 = $393&255;
         $396 = $394;
        }
        HEAP32[$ls>>2] = $396;
        $$be$i = $396;
        break;
       }
       default: {
        $397 = HEAP32[$0>>2]|0;
        $398 = ((($397)) + 4|0);
        $399 = HEAP32[$398>>2]|0;
        $400 = (($399) + 1)|0;
        $401 = ((($397)) + 8|0);
        $402 = HEAP32[$401>>2]|0;
        $403 = ($400>>>0)>($402>>>0);
        if ($403) {
         $404 = ($402>>>0)>(2147483645);
         if ($404) {
          label = 156;
          break L114;
         }
         $410 = $402 << 1;
         $411 = ($410|0)==(-2);
         $412 = HEAP32[$211>>2]|0;
         if ($411) {
          $415 = (_luaM_toobig($412)|0);
          $$pre$phi$i40$iZ2D = $397;$416 = $415;
         } else {
          $413 = HEAP32[$397>>2]|0;
          $414 = (_luaM_realloc_($412,$413,$402,$410)|0);
          $$pre$phi$i40$iZ2D = $397;$416 = $414;
         }
         HEAP32[$$pre$phi$i40$iZ2D>>2] = $416;
         HEAP32[$401>>2] = $410;
         $$pre1$i41$i = HEAP32[$398>>2]|0;
         $419 = $$pre1$i41$i;$421 = $416;
        } else {
         $$pre2$i36$i = HEAP32[$397>>2]|0;
         $419 = $399;$421 = $$pre2$i36$i;
        }
        $417 = $212&255;
        $418 = (($419) + 1)|0;
        HEAP32[$398>>2] = $418;
        $420 = (($421) + ($419)|0);
        HEAP8[$420>>0] = $417;
        $422 = HEAP32[$3>>2]|0;
        $423 = HEAP32[$422>>2]|0;
        $424 = (($423) + -1)|0;
        HEAP32[$422>>2] = $424;
        $425 = ($423|0)==(0);
        if ($425) {
         $431 = (_luaZ_fill($422)|0);
         $432 = $431;
        } else {
         $426 = ((($422)) + 4|0);
         $427 = HEAP32[$426>>2]|0;
         $428 = ((($427)) + 1|0);
         HEAP32[$426>>2] = $428;
         $429 = HEAP8[$427>>0]|0;
         $430 = $429&255;
         $432 = $430;
        }
        HEAP32[$ls>>2] = $432;
        $$be$i = $432;
       }
       }
      } while(0);
      if ((label|0) == 80) {
       label = 0;
       $$pr$i = HEAP32[$ls>>2]|0;
       $$be$i = $$pr$i;
      }
      $433 = ($$be$i|0)==($$lcssa122|0);
      if ($433) {
       break L112;
      } else {
       $212 = $$be$i;
      }
     }
     if ((label|0) == 82) {
      $213 = HEAP32[$210>>2]|0;
      $214 = ((($213)) + 16|0);
      _luaO_chunkid($buff$i$i,$214,80);
      $215 = HEAP32[$211>>2]|0;
      $216 = HEAP32[$4>>2]|0;
      HEAP32[$vararg_buffer3>>2] = $buff$i$i;
      $vararg_ptr6 = ((($vararg_buffer3)) + 4|0);
      HEAP32[$vararg_ptr6>>2] = $216;
      $vararg_ptr7 = ((($vararg_buffer3)) + 8|0);
      HEAP32[$vararg_ptr7>>2] = 520080;
      $217 = (_luaO_pushfstring($215,519897,$vararg_buffer3)|0);
      $218 = HEAP32[$211>>2]|0;
      HEAP32[$vararg_buffer8>>2] = $217;
      $vararg_ptr11 = ((($vararg_buffer8)) + 4|0);
      HEAP32[$vararg_ptr11>>2] = 519882;
      (_luaO_pushfstring($218,519932,$vararg_buffer8)|0);
      $219 = HEAP32[$211>>2]|0;
      _luaD_throw($219,3);
      // unreachable;
     }
     else if ((label|0) == 97) {
      $239 = HEAP32[$210>>2]|0;
      $240 = ((($239)) + 16|0);
      _luaO_chunkid($buff$i52$i,$240,80);
      $241 = HEAP32[$211>>2]|0;
      $242 = HEAP32[$4>>2]|0;
      HEAP32[$vararg_buffer12>>2] = $buff$i52$i;
      $vararg_ptr15 = ((($vararg_buffer12)) + 4|0);
      HEAP32[$vararg_ptr15>>2] = $242;
      $vararg_ptr16 = ((($vararg_buffer12)) + 8|0);
      HEAP32[$vararg_ptr16>>2] = 519907;
      (_luaO_pushfstring($241,519897,$vararg_buffer12)|0);
      $243 = HEAP32[$211>>2]|0;
      _luaD_throw($243,3);
      // unreachable;
     }
     else if ((label|0) == 117) {
      $291 = HEAP32[$210>>2]|0;
      $292 = ((($291)) + 16|0);
      _luaO_chunkid($buff$i53$i,$292,80);
      $293 = HEAP32[$211>>2]|0;
      $294 = HEAP32[$4>>2]|0;
      HEAP32[$vararg_buffer17>>2] = $buff$i53$i;
      $vararg_ptr20 = ((($vararg_buffer17)) + 4|0);
      HEAP32[$vararg_ptr20>>2] = $294;
      $vararg_ptr21 = ((($vararg_buffer17)) + 8|0);
      HEAP32[$vararg_ptr21>>2] = 519907;
      (_luaO_pushfstring($293,519897,$vararg_buffer17)|0);
      $295 = HEAP32[$211>>2]|0;
      _luaD_throw($295,3);
      // unreachable;
     }
     else if ((label|0) == 135) {
      $345 = HEAP32[$210>>2]|0;
      $346 = ((($345)) + 16|0);
      _luaO_chunkid($buff$i54$i,$346,80);
      $347 = HEAP32[$211>>2]|0;
      $348 = HEAP32[$4>>2]|0;
      HEAP32[$vararg_buffer22>>2] = $buff$i54$i;
      $vararg_ptr25 = ((($vararg_buffer22)) + 4|0);
      HEAP32[$vararg_ptr25>>2] = $348;
      $vararg_ptr26 = ((($vararg_buffer22)) + 8|0);
      HEAP32[$vararg_ptr26>>2] = 519907;
      (_luaO_pushfstring($347,519897,$vararg_buffer22)|0);
      $349 = HEAP32[$211>>2]|0;
      _luaD_throw($349,3);
      // unreachable;
     }
     else if ((label|0) == 144) {
      $370 = HEAP32[$210>>2]|0;
      $371 = ((($370)) + 16|0);
      _luaO_chunkid($buff$i55$i,$371,80);
      $372 = HEAP32[$211>>2]|0;
      $373 = HEAP32[$4>>2]|0;
      HEAP32[$vararg_buffer27>>2] = $buff$i55$i;
      $vararg_ptr30 = ((($vararg_buffer27)) + 4|0);
      HEAP32[$vararg_ptr30>>2] = $373;
      $vararg_ptr31 = ((($vararg_buffer27)) + 8|0);
      HEAP32[$vararg_ptr31>>2] = 519907;
      (_luaO_pushfstring($372,519897,$vararg_buffer27)|0);
      $374 = HEAP32[$211>>2]|0;
      _luaD_throw($374,3);
      // unreachable;
     }
     else if ((label|0) == 156) {
      $405 = HEAP32[$210>>2]|0;
      $406 = ((($405)) + 16|0);
      _luaO_chunkid($buff$i56$i,$406,80);
      $407 = HEAP32[$211>>2]|0;
      $408 = HEAP32[$4>>2]|0;
      HEAP32[$vararg_buffer32>>2] = $buff$i56$i;
      $vararg_ptr35 = ((($vararg_buffer32)) + 4|0);
      HEAP32[$vararg_ptr35>>2] = $408;
      $vararg_ptr36 = ((($vararg_buffer32)) + 8|0);
      HEAP32[$vararg_ptr36>>2] = 519907;
      (_luaO_pushfstring($407,519897,$vararg_buffer32)|0);
      $409 = HEAP32[$211>>2]|0;
      _luaD_throw($409,3);
      // unreachable;
     }
    }
   } while(0);
   $434 = HEAP32[$0>>2]|0;
   $435 = ((($434)) + 4|0);
   $436 = HEAP32[$435>>2]|0;
   $437 = (($436) + 1)|0;
   $438 = ((($434)) + 8|0);
   $439 = HEAP32[$438>>2]|0;
   $440 = ($437>>>0)>($439>>>0);
   if ($440) {
    $441 = ($439>>>0)>(2147483645);
    if ($441) {
     $445 = ((($ls)) + 64|0);
     $446 = HEAP32[$445>>2]|0;
     $447 = ((($446)) + 16|0);
     _luaO_chunkid($buff$i57$i,$447,80);
     $448 = ((($ls)) + 52|0);
     $449 = HEAP32[$448>>2]|0;
     $450 = HEAP32[$4>>2]|0;
     HEAP32[$vararg_buffer37>>2] = $buff$i57$i;
     $vararg_ptr40 = ((($vararg_buffer37)) + 4|0);
     HEAP32[$vararg_ptr40>>2] = $450;
     $vararg_ptr41 = ((($vararg_buffer37)) + 8|0);
     HEAP32[$vararg_ptr41>>2] = 519907;
     (_luaO_pushfstring($449,519897,$vararg_buffer37)|0);
     $451 = HEAP32[$448>>2]|0;
     _luaD_throw($451,3);
     // unreachable;
    }
    $$pre61$i = ((($ls)) + 52|0);
    $442 = $439 << 1;
    $443 = ($442|0)==(-2);
    $444 = HEAP32[$$pre61$i>>2]|0;
    if ($443) {
     $454 = (_luaM_toobig($444)|0);
     $$pre$phi$i48$iZ2D = $434;$455 = $454;
    } else {
     $452 = HEAP32[$434>>2]|0;
     $453 = (_luaM_realloc_($444,$452,$439,$442)|0);
     $$pre$phi$i48$iZ2D = $434;$455 = $453;
    }
    HEAP32[$$pre$phi$i48$iZ2D>>2] = $455;
    HEAP32[$438>>2] = $442;
    $$pre1$i49$i = HEAP32[$435>>2]|0;
    $457 = $$pre1$i49$i;$459 = $455;
   } else {
    $$pre2$i44$i = HEAP32[$434>>2]|0;
    $457 = $436;$459 = $$pre2$i44$i;
   }
   $456 = (($457) + 1)|0;
   HEAP32[$435>>2] = $456;
   $458 = (($459) + ($457)|0);
   HEAP8[$458>>0] = $193;
   $460 = HEAP32[$3>>2]|0;
   $461 = HEAP32[$460>>2]|0;
   $462 = (($461) + -1)|0;
   HEAP32[$460>>2] = $462;
   $463 = ($461|0)==(0);
   if ($463) {
    $469 = (_luaZ_fill($460)|0);
    $470 = $469;
   } else {
    $464 = ((($460)) + 4|0);
    $465 = HEAP32[$464>>2]|0;
    $466 = ((($465)) + 1|0);
    HEAP32[$464>>2] = $466;
    $467 = HEAP8[$465>>0]|0;
    $468 = $467&255;
    $470 = $468;
   }
   HEAP32[$ls>>2] = $470;
   $471 = HEAP32[$0>>2]|0;
   $472 = HEAP32[$471>>2]|0;
   $473 = ((($472)) + 1|0);
   $474 = ((($471)) + 4|0);
   $475 = HEAP32[$474>>2]|0;
   $476 = (($475) + -2)|0;
   $477 = ((($ls)) + 52|0);
   $478 = HEAP32[$477>>2]|0;
   $479 = (_luaS_newlstr($478,$473,$476)|0);
   $480 = ((($ls)) + 48|0);
   $481 = HEAP32[$480>>2]|0;
   $482 = ((($481)) + 4|0);
   $483 = HEAP32[$482>>2]|0;
   $484 = (_luaH_setstr($478,$483,$479)|0);
   $485 = ((($484)) + 8|0);
   $486 = HEAP32[$485>>2]|0;
   $487 = ($486|0)==(0);
   if ($487) {
    HEAP32[$484>>2] = 1;
    HEAP32[$485>>2] = 1;
    $488 = ((($478)) + 16|0);
    $489 = HEAP32[$488>>2]|0;
    $490 = ((($489)) + 68|0);
    $491 = HEAP32[$490>>2]|0;
    $492 = ((($489)) + 64|0);
    $493 = HEAP32[$492>>2]|0;
    $494 = ($491>>>0)<($493>>>0);
    if (!($494)) {
     _luaC_step($478);
    }
   }
   HEAP32[$seminfo>>2] = $479;
   $$2 = 286;
   STACKTOP = sp;return ($$2|0);
  }
  else if ((label|0) == 181) {
   $495 = HEAP32[$0>>2]|0;
   $496 = ((($495)) + 4|0);
   $497 = HEAP32[$496>>2]|0;
   $498 = (($497) + 1)|0;
   $499 = ((($495)) + 8|0);
   $500 = HEAP32[$499>>2]|0;
   $501 = ($498>>>0)>($500>>>0);
   if ($501) {
    $502 = ($500>>>0)>(2147483645);
    if ($502) {
     $506 = ((($ls)) + 64|0);
     $507 = HEAP32[$506>>2]|0;
     $508 = ((($507)) + 16|0);
     _luaO_chunkid($buff$i,$508,80);
     $509 = ((($ls)) + 52|0);
     $510 = HEAP32[$509>>2]|0;
     $511 = HEAP32[$4>>2]|0;
     HEAP32[$vararg_buffer42>>2] = $buff$i;
     $vararg_ptr45 = ((($vararg_buffer42)) + 4|0);
     HEAP32[$vararg_ptr45>>2] = $511;
     $vararg_ptr46 = ((($vararg_buffer42)) + 8|0);
     HEAP32[$vararg_ptr46>>2] = 519907;
     (_luaO_pushfstring($510,519897,$vararg_buffer42)|0);
     $512 = HEAP32[$509>>2]|0;
     _luaD_throw($512,3);
     // unreachable;
    }
    $$pre = ((($ls)) + 52|0);
    $503 = $500 << 1;
    $504 = ($503|0)==(-2);
    $505 = HEAP32[$$pre>>2]|0;
    if ($504) {
     $515 = (_luaM_toobig($505)|0);
     $$pre$phi$i10Z2D = $495;$516 = $515;
    } else {
     $513 = HEAP32[$495>>2]|0;
     $514 = (_luaM_realloc_($505,$513,$500,$503)|0);
     $$pre$phi$i10Z2D = $495;$516 = $514;
    }
    HEAP32[$$pre$phi$i10Z2D>>2] = $516;
    HEAP32[$499>>2] = $503;
    $$pre1$i = HEAP32[$496>>2]|0;
    $518 = $$pre1$i;$520 = $516;
   } else {
    $$pre2$i = HEAP32[$495>>2]|0;
    $518 = $497;$520 = $$pre2$i;
   }
   $517 = (($518) + 1)|0;
   HEAP32[$496>>2] = $517;
   $519 = (($520) + ($518)|0);
   HEAP8[$519>>0] = 46;
   $521 = HEAP32[$3>>2]|0;
   $522 = HEAP32[$521>>2]|0;
   $523 = (($522) + -1)|0;
   HEAP32[$521>>2] = $523;
   $524 = ($522|0)==(0);
   if ($524) {
    $530 = (_luaZ_fill($521)|0);
    $531 = $530;
   } else {
    $525 = ((($521)) + 4|0);
    $526 = HEAP32[$525>>2]|0;
    $527 = ((($526)) + 1|0);
    HEAP32[$525>>2] = $527;
    $528 = HEAP8[$526>>0]|0;
    $529 = $528&255;
    $531 = $529;
   }
   HEAP32[$ls>>2] = $531;
   $memchr = (_memchr(619449,$531,2)|0);
   $532 = ($memchr|0)==(0|0);
   if ($532) {
    $isdigittmp = (($531) + -48)|0;
    $isdigit = ($isdigittmp>>>0)<(10);
    if (!($isdigit)) {
     $$2 = 46;
     STACKTOP = sp;return ($$2|0);
    }
    _read_numeral($ls,$seminfo);
    $$2 = 284;
    STACKTOP = sp;return ($$2|0);
   }
   $533 = HEAP32[$0>>2]|0;
   $534 = ((($533)) + 4|0);
   $535 = HEAP32[$534>>2]|0;
   $536 = (($535) + 1)|0;
   $537 = ((($533)) + 8|0);
   $538 = HEAP32[$537>>2]|0;
   $539 = ($536>>>0)>($538>>>0);
   if ($539) {
    $540 = ($538>>>0)>(2147483645);
    if ($540) {
     $544 = ((($ls)) + 64|0);
     $545 = HEAP32[$544>>2]|0;
     $546 = ((($545)) + 16|0);
     _luaO_chunkid($buff$i41,$546,80);
     $547 = ((($ls)) + 52|0);
     $548 = HEAP32[$547>>2]|0;
     $549 = HEAP32[$4>>2]|0;
     HEAP32[$vararg_buffer47>>2] = $buff$i41;
     $vararg_ptr50 = ((($vararg_buffer47)) + 4|0);
     HEAP32[$vararg_ptr50>>2] = $549;
     $vararg_ptr51 = ((($vararg_buffer47)) + 8|0);
     HEAP32[$vararg_ptr51>>2] = 519907;
     (_luaO_pushfstring($548,519897,$vararg_buffer47)|0);
     $550 = HEAP32[$547>>2]|0;
     _luaD_throw($550,3);
     // unreachable;
    }
    $$pre74 = ((($ls)) + 52|0);
    $541 = $538 << 1;
    $542 = ($541|0)==(-2);
    $543 = HEAP32[$$pre74>>2]|0;
    if ($542) {
     $553 = (_luaM_toobig($543)|0);
     $$pre$phi$i27Z2D = $533;$554 = $553;
    } else {
     $551 = HEAP32[$533>>2]|0;
     $552 = (_luaM_realloc_($543,$551,$538,$541)|0);
     $$pre$phi$i27Z2D = $533;$554 = $552;
    }
    HEAP32[$$pre$phi$i27Z2D>>2] = $554;
    HEAP32[$537>>2] = $541;
    $$pre1$i28 = HEAP32[$534>>2]|0;
    $557 = $$pre1$i28;$559 = $554;
   } else {
    $$pre2$i23 = HEAP32[$533>>2]|0;
    $557 = $535;$559 = $$pre2$i23;
   }
   $555 = $531&255;
   $556 = (($557) + 1)|0;
   HEAP32[$534>>2] = $556;
   $558 = (($559) + ($557)|0);
   HEAP8[$558>>0] = $555;
   $560 = HEAP32[$3>>2]|0;
   $561 = HEAP32[$560>>2]|0;
   $562 = (($561) + -1)|0;
   HEAP32[$560>>2] = $562;
   $563 = ($561|0)==(0);
   if ($563) {
    $569 = (_luaZ_fill($560)|0);
    $570 = $569;
   } else {
    $564 = ((($560)) + 4|0);
    $565 = HEAP32[$564>>2]|0;
    $566 = ((($565)) + 1|0);
    HEAP32[$564>>2] = $566;
    $567 = HEAP8[$565>>0]|0;
    $568 = $567&255;
    $570 = $568;
   }
   HEAP32[$ls>>2] = $570;
   $memchr44 = (_memchr(619449,$570,2)|0);
   $571 = ($memchr44|0)==(0|0);
   if ($571) {
    $$2 = 278;
    STACKTOP = sp;return ($$2|0);
   }
   $572 = HEAP32[$0>>2]|0;
   $573 = ((($572)) + 4|0);
   $574 = HEAP32[$573>>2]|0;
   $575 = (($574) + 1)|0;
   $576 = ((($572)) + 8|0);
   $577 = HEAP32[$576>>2]|0;
   $578 = ($575>>>0)>($577>>>0);
   if ($578) {
    $579 = ($577>>>0)>(2147483645);
    if ($579) {
     $583 = ((($ls)) + 64|0);
     $584 = HEAP32[$583>>2]|0;
     $585 = ((($584)) + 16|0);
     _luaO_chunkid($buff$i43,$585,80);
     $586 = ((($ls)) + 52|0);
     $587 = HEAP32[$586>>2]|0;
     $588 = HEAP32[$4>>2]|0;
     HEAP32[$vararg_buffer52>>2] = $buff$i43;
     $vararg_ptr55 = ((($vararg_buffer52)) + 4|0);
     HEAP32[$vararg_ptr55>>2] = $588;
     $vararg_ptr56 = ((($vararg_buffer52)) + 8|0);
     HEAP32[$vararg_ptr56>>2] = 519907;
     (_luaO_pushfstring($587,519897,$vararg_buffer52)|0);
     $589 = HEAP32[$586>>2]|0;
     _luaD_throw($589,3);
     // unreachable;
    }
    $$pre76 = ((($ls)) + 52|0);
    $580 = $577 << 1;
    $581 = ($580|0)==(-2);
    $582 = HEAP32[$$pre76>>2]|0;
    if ($581) {
     $592 = (_luaM_toobig($582)|0);
     $$pre$phi$i35Z2D = $572;$593 = $592;
    } else {
     $590 = HEAP32[$572>>2]|0;
     $591 = (_luaM_realloc_($582,$590,$577,$580)|0);
     $$pre$phi$i35Z2D = $572;$593 = $591;
    }
    HEAP32[$$pre$phi$i35Z2D>>2] = $593;
    HEAP32[$576>>2] = $580;
    $$pre1$i36 = HEAP32[$573>>2]|0;
    $596 = $$pre1$i36;$598 = $593;
   } else {
    $$pre2$i31 = HEAP32[$572>>2]|0;
    $596 = $574;$598 = $$pre2$i31;
   }
   $594 = $570&255;
   $595 = (($596) + 1)|0;
   HEAP32[$573>>2] = $595;
   $597 = (($598) + ($596)|0);
   HEAP8[$597>>0] = $594;
   $599 = HEAP32[$3>>2]|0;
   $600 = HEAP32[$599>>2]|0;
   $601 = (($600) + -1)|0;
   HEAP32[$599>>2] = $601;
   $602 = ($600|0)==(0);
   if ($602) {
    $608 = (_luaZ_fill($599)|0);
    $609 = $608;
   } else {
    $603 = ((($599)) + 4|0);
    $604 = HEAP32[$603>>2]|0;
    $605 = ((($604)) + 1|0);
    HEAP32[$603>>2] = $605;
    $606 = HEAP8[$604>>0]|0;
    $607 = $606&255;
    $609 = $607;
   }
   HEAP32[$ls>>2] = $609;
   $$2 = 279;
   STACKTOP = sp;return ($$2|0);
  }
  else if ((label|0) == 224) {
   $623 = HEAP32[$ls>>2]|0;
   $isdigittmp6 = (($623) + -48)|0;
   $isdigit7 = ($isdigittmp6>>>0)<(10);
   if ($isdigit7) {
    _read_numeral($ls,$seminfo);
    $$2 = 284;
    STACKTOP = sp;return ($$2|0);
   }
   $624 = (_isalpha($623)|0);
   $625 = ($624|0)==(0);
   if ($625) {
    $626 = HEAP32[$ls>>2]|0;
    $627 = ($626|0)==(95);
    if (!($627)) {
     $697 = HEAP32[$3>>2]|0;
     $698 = HEAP32[$697>>2]|0;
     $699 = (($698) + -1)|0;
     HEAP32[$697>>2] = $699;
     $700 = ($698|0)==(0);
     if ($700) {
      $706 = (_luaZ_fill($697)|0);
      $707 = $706;
     } else {
      $701 = ((($697)) + 4|0);
      $702 = HEAP32[$701>>2]|0;
      $703 = ((($702)) + 1|0);
      HEAP32[$701>>2] = $703;
      $704 = HEAP8[$702>>0]|0;
      $705 = $704&255;
      $707 = $705;
     }
     HEAP32[$ls>>2] = $707;
     $$2 = $626;
     STACKTOP = sp;return ($$2|0);
    }
   }
   $628 = ((($ls)) + 64|0);
   $629 = ((($ls)) + 52|0);
   while(1) {
    $630 = HEAP32[$ls>>2]|0;
    $631 = HEAP32[$0>>2]|0;
    $632 = ((($631)) + 4|0);
    $633 = HEAP32[$632>>2]|0;
    $634 = (($633) + 1)|0;
    $635 = ((($631)) + 8|0);
    $636 = HEAP32[$635>>2]|0;
    $637 = ($634>>>0)>($636>>>0);
    if ($637) {
     $638 = ($636>>>0)>(2147483645);
     if ($638) {
      label = 232;
      break;
     }
     $644 = $636 << 1;
     $645 = ($644|0)==(-2);
     $646 = HEAP32[$629>>2]|0;
     if ($645) {
      $649 = (_luaM_toobig($646)|0);
      $$pre$phi$i18Z2D = $631;$650 = $649;
     } else {
      $647 = HEAP32[$631>>2]|0;
      $648 = (_luaM_realloc_($646,$647,$636,$644)|0);
      $$pre$phi$i18Z2D = $631;$650 = $648;
     }
     HEAP32[$$pre$phi$i18Z2D>>2] = $650;
     HEAP32[$635>>2] = $644;
     $$pre1$i19 = HEAP32[$632>>2]|0;
     $653 = $$pre1$i19;$655 = $650;
    } else {
     $$pre2$i14 = HEAP32[$631>>2]|0;
     $653 = $633;$655 = $$pre2$i14;
    }
    $651 = $630&255;
    $652 = (($653) + 1)|0;
    HEAP32[$632>>2] = $652;
    $654 = (($655) + ($653)|0);
    HEAP8[$654>>0] = $651;
    $656 = HEAP32[$3>>2]|0;
    $657 = HEAP32[$656>>2]|0;
    $658 = (($657) + -1)|0;
    HEAP32[$656>>2] = $658;
    $659 = ($657|0)==(0);
    if ($659) {
     $665 = (_luaZ_fill($656)|0);
     $666 = $665;
    } else {
     $660 = ((($656)) + 4|0);
     $661 = HEAP32[$660>>2]|0;
     $662 = ((($661)) + 1|0);
     HEAP32[$660>>2] = $662;
     $663 = HEAP8[$661>>0]|0;
     $664 = $663&255;
     $666 = $664;
    }
    HEAP32[$ls>>2] = $666;
    $667 = (_isalnum($666)|0);
    $668 = ($667|0)==(0);
    if (!($668)) {
     continue;
    }
    $669 = HEAP32[$ls>>2]|0;
    $670 = ($669|0)==(95);
    if (!($670)) {
     break;
    }
   }
   if ((label|0) == 232) {
    $639 = HEAP32[$628>>2]|0;
    $640 = ((($639)) + 16|0);
    _luaO_chunkid($buff$i39,$640,80);
    $641 = HEAP32[$629>>2]|0;
    $642 = HEAP32[$4>>2]|0;
    HEAP32[$vararg_buffer57>>2] = $buff$i39;
    $vararg_ptr60 = ((($vararg_buffer57)) + 4|0);
    HEAP32[$vararg_ptr60>>2] = $642;
    $vararg_ptr61 = ((($vararg_buffer57)) + 8|0);
    HEAP32[$vararg_ptr61>>2] = 519907;
    (_luaO_pushfstring($641,519897,$vararg_buffer57)|0);
    $643 = HEAP32[$629>>2]|0;
    _luaD_throw($643,3);
    // unreachable;
   }
   $671 = HEAP32[$0>>2]|0;
   $672 = HEAP32[$671>>2]|0;
   $673 = ((($671)) + 4|0);
   $674 = HEAP32[$673>>2]|0;
   $675 = HEAP32[$629>>2]|0;
   $676 = (_luaS_newlstr($675,$672,$674)|0);
   $677 = ((($ls)) + 48|0);
   $678 = HEAP32[$677>>2]|0;
   $679 = ((($678)) + 4|0);
   $680 = HEAP32[$679>>2]|0;
   $681 = (_luaH_setstr($675,$680,$676)|0);
   $682 = ((($681)) + 8|0);
   $683 = HEAP32[$682>>2]|0;
   $684 = ($683|0)==(0);
   if ($684) {
    HEAP32[$681>>2] = 1;
    HEAP32[$682>>2] = 1;
    $685 = ((($675)) + 16|0);
    $686 = HEAP32[$685>>2]|0;
    $687 = ((($686)) + 68|0);
    $688 = HEAP32[$687>>2]|0;
    $689 = ((($686)) + 64|0);
    $690 = HEAP32[$689>>2]|0;
    $691 = ($688>>>0)<($690>>>0);
    if (!($691)) {
     _luaC_step($675);
    }
   }
   $692 = ((($676)) + 6|0);
   $693 = HEAP8[$692>>0]|0;
   $694 = ($693<<24>>24)==(0);
   if ($694) {
    HEAP32[$seminfo>>2] = $676;
    $$2 = 285;
    STACKTOP = sp;return ($$2|0);
   } else {
    $695 = $693&255;
    $696 = $695 | 256;
    $$2 = $696;
    STACKTOP = sp;return ($$2|0);
   }
  }
  else if ((label|0) == 252) {
   STACKTOP = sp;return ($$2|0);
  }
 } while(0);
 $79 = HEAP32[$3>>2]|0;
 $80 = HEAP32[$79>>2]|0;
 $81 = (($80) + -1)|0;
 HEAP32[$79>>2] = $81;
 $82 = ($80|0)==(0);
 if ($82) {
  $88 = (_luaZ_fill($79)|0);
  $89 = $88;
 } else {
  $83 = ((($79)) + 4|0);
  $84 = HEAP32[$83>>2]|0;
  $85 = ((($84)) + 1|0);
  HEAP32[$83>>2] = $85;
  $86 = HEAP8[$84>>0]|0;
  $87 = $86&255;
  $89 = $87;
 }
 HEAP32[$ls>>2] = $89;
 $90 = ($89|0)==(61);
 if (!($90)) {
  $$2 = 61;
  STACKTOP = sp;return ($$2|0);
 }
 $91 = HEAP32[$3>>2]|0;
 $92 = HEAP32[$91>>2]|0;
 $93 = (($92) + -1)|0;
 HEAP32[$91>>2] = $93;
 $94 = ($92|0)==(0);
 if ($94) {
  $100 = (_luaZ_fill($91)|0);
  $101 = $100;
 } else {
  $95 = ((($91)) + 4|0);
  $96 = HEAP32[$95>>2]|0;
  $97 = ((($96)) + 1|0);
  HEAP32[$95>>2] = $97;
  $98 = HEAP8[$96>>0]|0;
  $99 = $98&255;
  $101 = $99;
 }
 HEAP32[$ls>>2] = $101;
 $$2 = 280;
 STACKTOP = sp;return ($$2|0);
}
function _skip_sep($ls) {
 $ls = $ls|0;
 var $$lcssa = 0, $$pre = 0, $$pre$phi$i6Z2D = 0, $$pre$phi$iZ2D = 0, $$pre1$i = 0, $$pre1$i7 = 0, $$pre2$i = 0, $$pre2$i2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $9 = 0, $buff$i = 0, $buff$i9 = 0, $count$0$lcssa = 0, $count$010 = 0, $not$ = 0, $vararg_buffer = 0, $vararg_buffer3 = 0;
 var $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr6 = 0, $vararg_ptr7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $buff$i9 = sp + 112|0;
 $buff$i = sp + 32|0;
 $0 = HEAP32[$ls>>2]|0;
 $1 = ((($ls)) + 60|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + 1)|0;
 $6 = ((($2)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($5>>>0)>($7>>>0);
 if ($8) {
  $9 = ($7>>>0)>(2147483645);
  if ($9) {
   $13 = ((($ls)) + 64|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ((($14)) + 16|0);
   _luaO_chunkid($buff$i,$15,80);
   $16 = ((($ls)) + 52|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($ls)) + 4|0);
   $19 = HEAP32[$18>>2]|0;
   HEAP32[$vararg_buffer>>2] = $buff$i;
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
   HEAP32[$vararg_ptr1>>2] = $19;
   $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
   HEAP32[$vararg_ptr2>>2] = 519907;
   (_luaO_pushfstring($17,519897,$vararg_buffer)|0);
   $20 = HEAP32[$16>>2]|0;
   _luaD_throw($20,3);
   // unreachable;
  }
  $$pre = ((($ls)) + 52|0);
  $10 = $7 << 1;
  $11 = ($10|0)==(-2);
  $12 = HEAP32[$$pre>>2]|0;
  if ($11) {
   $23 = (_luaM_toobig($12)|0);
   $$pre$phi$iZ2D = $2;$24 = $23;
  } else {
   $21 = HEAP32[$2>>2]|0;
   $22 = (_luaM_realloc_($12,$21,$7,$10)|0);
   $$pre$phi$iZ2D = $2;$24 = $22;
  }
  HEAP32[$$pre$phi$iZ2D>>2] = $24;
  HEAP32[$6>>2] = $10;
  $$pre1$i = HEAP32[$3>>2]|0;
  $27 = $$pre1$i;$29 = $24;
 } else {
  $$pre2$i = HEAP32[$2>>2]|0;
  $27 = $4;$29 = $$pre2$i;
 }
 $25 = $0&255;
 $26 = (($27) + 1)|0;
 HEAP32[$3>>2] = $26;
 $28 = (($29) + ($27)|0);
 HEAP8[$28>>0] = $25;
 $30 = ((($ls)) + 56|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = HEAP32[$31>>2]|0;
 $33 = (($32) + -1)|0;
 HEAP32[$31>>2] = $33;
 $34 = ($32|0)==(0);
 if ($34) {
  $40 = (_luaZ_fill($31)|0);
  $41 = $40;
 } else {
  $35 = ((($31)) + 4|0);
  $36 = HEAP32[$35>>2]|0;
  $37 = ((($36)) + 1|0);
  HEAP32[$35>>2] = $37;
  $38 = HEAP8[$36>>0]|0;
  $39 = $38&255;
  $41 = $39;
 }
 HEAP32[$ls>>2] = $41;
 $42 = ($41|0)==(61);
 if (!($42)) {
  $$lcssa = $41;$count$0$lcssa = 0;
  $not$ = ($$lcssa|0)!=($0|0);
  $83 = $not$ << 31 >> 31;
  $84 = $83 ^ $count$0$lcssa;
  STACKTOP = sp;return ($84|0);
 }
 $43 = ((($ls)) + 64|0);
 $44 = ((($ls)) + 52|0);
 $45 = ((($ls)) + 4|0);
 $count$010 = 0;
 while(1) {
  $46 = HEAP32[$1>>2]|0;
  $47 = ((($46)) + 4|0);
  $48 = HEAP32[$47>>2]|0;
  $49 = (($48) + 1)|0;
  $50 = ((($46)) + 8|0);
  $51 = HEAP32[$50>>2]|0;
  $52 = ($49>>>0)>($51>>>0);
  if ($52) {
   $53 = ($51>>>0)>(2147483645);
   if ($53) {
    label = 17;
    break;
   }
   $59 = $51 << 1;
   $60 = ($59|0)==(-2);
   $61 = HEAP32[$44>>2]|0;
   if ($60) {
    $64 = (_luaM_toobig($61)|0);
    $$pre$phi$i6Z2D = $46;$65 = $64;
   } else {
    $62 = HEAP32[$46>>2]|0;
    $63 = (_luaM_realloc_($61,$62,$51,$59)|0);
    $$pre$phi$i6Z2D = $46;$65 = $63;
   }
   HEAP32[$$pre$phi$i6Z2D>>2] = $65;
   HEAP32[$50>>2] = $59;
   $$pre1$i7 = HEAP32[$47>>2]|0;
   $67 = $$pre1$i7;$69 = $65;
  } else {
   $$pre2$i2 = HEAP32[$46>>2]|0;
   $67 = $48;$69 = $$pre2$i2;
  }
  $66 = (($67) + 1)|0;
  HEAP32[$47>>2] = $66;
  $68 = (($69) + ($67)|0);
  HEAP8[$68>>0] = 61;
  $70 = HEAP32[$30>>2]|0;
  $71 = HEAP32[$70>>2]|0;
  $72 = (($71) + -1)|0;
  HEAP32[$70>>2] = $72;
  $73 = ($71|0)==(0);
  if ($73) {
   $79 = (_luaZ_fill($70)|0);
   $80 = $79;
  } else {
   $74 = ((($70)) + 4|0);
   $75 = HEAP32[$74>>2]|0;
   $76 = ((($75)) + 1|0);
   HEAP32[$74>>2] = $76;
   $77 = HEAP8[$75>>0]|0;
   $78 = $77&255;
   $80 = $78;
  }
  HEAP32[$ls>>2] = $80;
  $81 = (($count$010) + 1)|0;
  $82 = ($80|0)==(61);
  if ($82) {
   $count$010 = $81;
  } else {
   $$lcssa = $80;$count$0$lcssa = $81;
   label = 26;
   break;
  }
 }
 if ((label|0) == 17) {
  $54 = HEAP32[$43>>2]|0;
  $55 = ((($54)) + 16|0);
  _luaO_chunkid($buff$i9,$55,80);
  $56 = HEAP32[$44>>2]|0;
  $57 = HEAP32[$45>>2]|0;
  HEAP32[$vararg_buffer3>>2] = $buff$i9;
  $vararg_ptr6 = ((($vararg_buffer3)) + 4|0);
  HEAP32[$vararg_ptr6>>2] = $57;
  $vararg_ptr7 = ((($vararg_buffer3)) + 8|0);
  HEAP32[$vararg_ptr7>>2] = 519907;
  (_luaO_pushfstring($56,519897,$vararg_buffer3)|0);
  $58 = HEAP32[$44>>2]|0;
  _luaD_throw($58,3);
  // unreachable;
 }
 else if ((label|0) == 26) {
  $not$ = ($$lcssa|0)!=($0|0);
  $83 = $not$ << 31 >> 31;
  $84 = $83 ^ $count$0$lcssa;
  STACKTOP = sp;return ($84|0);
 }
 return (0)|0;
}
function _read_long_string($ls,$seminfo,$sep) {
 $ls = $ls|0;
 $seminfo = $seminfo|0;
 $sep = $sep|0;
 var $$ph = 0, $$ph40 = 0, $$pr = 0, $$pre = 0, $$pre$phi$i14Z2D = 0, $$pre$phi$i22Z2D = 0, $$pre$phi$i31Z2D = 0, $$pre$phi$i6Z2D = 0, $$pre$phi$iZ2D = 0, $$pre1$i = 0, $$pre1$i15 = 0, $$pre1$i23 = 0, $$pre1$i32 = 0, $$pre1$i7 = 0, $$pre2$i = 0, $$pre2$i10 = 0, $$pre2$i18 = 0, $$pre2$i2 = 0, $$pre2$i27 = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0;
 var $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0;
 var $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $buff$i = 0, $buff$i34 = 0, $buff$i35 = 0, $buff$i36 = 0, $buff$i37 = 0, $buff$i38 = 0, $vararg_buffer = 0, $vararg_buffer12 = 0, $vararg_buffer17 = 0;
 var $vararg_buffer22 = 0, $vararg_buffer27 = 0, $vararg_buffer3 = 0, $vararg_buffer8 = 0, $vararg_ptr1 = 0, $vararg_ptr11 = 0, $vararg_ptr15 = 0, $vararg_ptr16 = 0, $vararg_ptr2 = 0, $vararg_ptr20 = 0, $vararg_ptr21 = 0, $vararg_ptr25 = 0, $vararg_ptr26 = 0, $vararg_ptr30 = 0, $vararg_ptr31 = 0, $vararg_ptr6 = 0, $vararg_ptr7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 592|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer27 = sp + 88|0;
 $vararg_buffer22 = sp + 72|0;
 $vararg_buffer17 = sp + 56|0;
 $vararg_buffer12 = sp + 40|0;
 $vararg_buffer8 = sp + 32|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $buff$i38 = sp + 504|0;
 $buff$i37 = sp + 424|0;
 $buff$i36 = sp + 344|0;
 $buff$i35 = sp + 264|0;
 $buff$i34 = sp + 184|0;
 $buff$i = sp + 104|0;
 $0 = HEAP32[$ls>>2]|0;
 $1 = ((($ls)) + 60|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + 1)|0;
 $6 = ((($2)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($5>>>0)>($7>>>0);
 if ($8) {
  $9 = ($7>>>0)>(2147483645);
  if ($9) {
   $13 = ((($ls)) + 64|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ((($14)) + 16|0);
   _luaO_chunkid($buff$i34,$15,80);
   $16 = ((($ls)) + 52|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($ls)) + 4|0);
   $19 = HEAP32[$18>>2]|0;
   HEAP32[$vararg_buffer>>2] = $buff$i34;
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
   HEAP32[$vararg_ptr1>>2] = $19;
   $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
   HEAP32[$vararg_ptr2>>2] = 519907;
   (_luaO_pushfstring($17,519897,$vararg_buffer)|0);
   $20 = HEAP32[$16>>2]|0;
   _luaD_throw($20,3);
   // unreachable;
  }
  $$pre = ((($ls)) + 52|0);
  $10 = $7 << 1;
  $11 = ($10|0)==(-2);
  $12 = HEAP32[$$pre>>2]|0;
  if ($11) {
   $23 = (_luaM_toobig($12)|0);
   $$pre$phi$iZ2D = $2;$24 = $23;
  } else {
   $21 = HEAP32[$2>>2]|0;
   $22 = (_luaM_realloc_($12,$21,$7,$10)|0);
   $$pre$phi$iZ2D = $2;$24 = $22;
  }
  HEAP32[$$pre$phi$iZ2D>>2] = $24;
  HEAP32[$6>>2] = $10;
  $$pre1$i = HEAP32[$3>>2]|0;
  $27 = $$pre1$i;$29 = $24;
 } else {
  $$pre2$i = HEAP32[$2>>2]|0;
  $27 = $4;$29 = $$pre2$i;
 }
 $25 = $0&255;
 $26 = (($27) + 1)|0;
 HEAP32[$3>>2] = $26;
 $28 = (($29) + ($27)|0);
 HEAP8[$28>>0] = $25;
 $30 = ((($ls)) + 56|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = HEAP32[$31>>2]|0;
 $33 = (($32) + -1)|0;
 HEAP32[$31>>2] = $33;
 $34 = ($32|0)==(0);
 if ($34) {
  $40 = (_luaZ_fill($31)|0);
  $41 = $40;
 } else {
  $35 = ((($31)) + 4|0);
  $36 = HEAP32[$35>>2]|0;
  $37 = ((($36)) + 1|0);
  HEAP32[$35>>2] = $37;
  $38 = HEAP8[$36>>0]|0;
  $39 = $38&255;
  $41 = $39;
 }
 HEAP32[$ls>>2] = $41;
 switch ($41|0) {
 case 13: case 10:  {
  $42 = HEAP32[$30>>2]|0;
  $43 = HEAP32[$42>>2]|0;
  $44 = (($43) + -1)|0;
  HEAP32[$42>>2] = $44;
  $45 = ($43|0)==(0);
  if ($45) {
   $51 = (_luaZ_fill($42)|0);
   $52 = $51;
  } else {
   $46 = ((($42)) + 4|0);
   $47 = HEAP32[$46>>2]|0;
   $48 = ((($47)) + 1|0);
   HEAP32[$46>>2] = $48;
   $49 = HEAP8[$47>>0]|0;
   $50 = $49&255;
   $52 = $50;
  }
  HEAP32[$ls>>2] = $52;
  switch ($52|0) {
  case 13: case 10:  {
   $53 = ($52|0)==($41|0);
   if (!($53)) {
    $54 = HEAP32[$30>>2]|0;
    $55 = HEAP32[$54>>2]|0;
    $56 = (($55) + -1)|0;
    HEAP32[$54>>2] = $56;
    $57 = ($55|0)==(0);
    if ($57) {
     $63 = (_luaZ_fill($54)|0);
     $64 = $63;
    } else {
     $58 = ((($54)) + 4|0);
     $59 = HEAP32[$58>>2]|0;
     $60 = ((($59)) + 1|0);
     HEAP32[$58>>2] = $60;
     $61 = HEAP8[$59>>0]|0;
     $62 = $61&255;
     $64 = $62;
    }
    HEAP32[$ls>>2] = $64;
   }
   break;
  }
  default: {
  }
  }
  $65 = ((($ls)) + 4|0);
  $66 = HEAP32[$65>>2]|0;
  $67 = (($66) + 1)|0;
  HEAP32[$65>>2] = $67;
  $68 = ($66|0)>(2147483643);
  if ($68) {
   $69 = ((($ls)) + 16|0);
   $70 = HEAP32[$69>>2]|0;
   _luaX_lexerror($ls,519945,$70);
   label = 24;
  } else {
   label = 24;
  }
  break;
 }
 default: {
  $$ph = $41;
 }
 }
 L31: while(1) {
  if ((label|0) == 24) {
   label = 0;
   $$pr = HEAP32[$ls>>2]|0;
   $$ph = $$pr;
  }
  $71 = ($seminfo|0)==(0|0);
  $72 = ((($ls)) + 64|0);
  $73 = ((($ls)) + 52|0);
  $74 = ((($ls)) + 4|0);
  $75 = ($sep|0)==(0);
  $$ph40 = $$ph;
  L35: while(1) {
   L37: do {
    if ($71) {
     $76 = $$ph40;
     while(1) {
      switch ($76|0) {
      case -1:  {
       label = 33;
       break L31;
       break;
      }
      case 93:  {
       label = 48;
       break L35;
       break;
      }
      case 13: case 10:  {
       label = 61;
       break L35;
       break;
      }
      case 91:  {
       break L37;
       break;
      }
      default: {
      }
      }
      $77 = HEAP32[$30>>2]|0;
      $78 = HEAP32[$77>>2]|0;
      $79 = (($78) + -1)|0;
      HEAP32[$77>>2] = $79;
      $80 = ($78|0)==(0);
      if ($80) {
       $86 = (_luaZ_fill($77)|0);
       $87 = $86;
      } else {
       $81 = ((($77)) + 4|0);
       $82 = HEAP32[$81>>2]|0;
       $83 = ((($82)) + 1|0);
       HEAP32[$81>>2] = $83;
       $84 = HEAP8[$82>>0]|0;
       $85 = $84&255;
       $87 = $85;
      }
      HEAP32[$ls>>2] = $87;
      $76 = $87;
     }
    } else {
     $88 = $$ph40;
     while(1) {
      switch ($88|0) {
      case -1:  {
       label = 33;
       break L31;
       break;
      }
      case 93:  {
       label = 48;
       break L35;
       break;
      }
      case 13: case 10:  {
       label = 61;
       break L35;
       break;
      }
      case 91:  {
       break L37;
       break;
      }
      default: {
      }
      }
      $231 = HEAP32[$1>>2]|0;
      $232 = ((($231)) + 4|0);
      $233 = HEAP32[$232>>2]|0;
      $234 = (($233) + 1)|0;
      $235 = ((($231)) + 8|0);
      $236 = HEAP32[$235>>2]|0;
      $237 = ($234>>>0)>($236>>>0);
      if ($237) {
       $238 = ($236>>>0)>(2147483645);
       if ($238) {
        label = 85;
        break L31;
       }
       $244 = $236 << 1;
       $245 = ($244|0)==(-2);
       $246 = HEAP32[$73>>2]|0;
       if ($245) {
        $249 = (_luaM_toobig($246)|0);
        $$pre$phi$i31Z2D = $231;$250 = $249;
       } else {
        $247 = HEAP32[$231>>2]|0;
        $248 = (_luaM_realloc_($246,$247,$236,$244)|0);
        $$pre$phi$i31Z2D = $231;$250 = $248;
       }
       HEAP32[$$pre$phi$i31Z2D>>2] = $250;
       HEAP32[$235>>2] = $244;
       $$pre1$i32 = HEAP32[$232>>2]|0;
       $253 = $$pre1$i32;$255 = $250;
      } else {
       $$pre2$i27 = HEAP32[$231>>2]|0;
       $253 = $233;$255 = $$pre2$i27;
      }
      $251 = $88&255;
      $252 = (($253) + 1)|0;
      HEAP32[$232>>2] = $252;
      $254 = (($255) + ($253)|0);
      HEAP8[$254>>0] = $251;
      $256 = HEAP32[$30>>2]|0;
      $257 = HEAP32[$256>>2]|0;
      $258 = (($257) + -1)|0;
      HEAP32[$256>>2] = $258;
      $259 = ($257|0)==(0);
      if ($259) {
       $265 = (_luaZ_fill($256)|0);
       $266 = $265;
      } else {
       $260 = ((($256)) + 4|0);
       $261 = HEAP32[$260>>2]|0;
       $262 = ((($261)) + 1|0);
       HEAP32[$260>>2] = $262;
       $263 = HEAP8[$261>>0]|0;
       $264 = $263&255;
       $266 = $264;
      }
      HEAP32[$ls>>2] = $266;
      $88 = $266;
     }
    }
   } while(0);
   $98 = (_skip_sep($ls)|0);
   $99 = ($98|0)==($sep|0);
   if (!($99)) {
    label = 24;
    continue L31;
   }
   $100 = HEAP32[$ls>>2]|0;
   $101 = HEAP32[$1>>2]|0;
   $102 = ((($101)) + 4|0);
   $103 = HEAP32[$102>>2]|0;
   $104 = (($103) + 1)|0;
   $105 = ((($101)) + 8|0);
   $106 = HEAP32[$105>>2]|0;
   $107 = ($104>>>0)>($106>>>0);
   if ($107) {
    $108 = ($106>>>0)>(2147483645);
    if ($108) {
     label = 38;
     break L31;
    }
    $114 = $106 << 1;
    $115 = ($114|0)==(-2);
    $116 = HEAP32[$73>>2]|0;
    if ($115) {
     $119 = (_luaM_toobig($116)|0);
     $$pre$phi$i6Z2D = $101;$120 = $119;
    } else {
     $117 = HEAP32[$101>>2]|0;
     $118 = (_luaM_realloc_($116,$117,$106,$114)|0);
     $$pre$phi$i6Z2D = $101;$120 = $118;
    }
    HEAP32[$$pre$phi$i6Z2D>>2] = $120;
    HEAP32[$105>>2] = $114;
    $$pre1$i7 = HEAP32[$102>>2]|0;
    $123 = $$pre1$i7;$125 = $120;
   } else {
    $$pre2$i2 = HEAP32[$101>>2]|0;
    $123 = $103;$125 = $$pre2$i2;
   }
   $121 = $100&255;
   $122 = (($123) + 1)|0;
   HEAP32[$102>>2] = $122;
   $124 = (($125) + ($123)|0);
   HEAP8[$124>>0] = $121;
   $126 = HEAP32[$30>>2]|0;
   $127 = HEAP32[$126>>2]|0;
   $128 = (($127) + -1)|0;
   HEAP32[$126>>2] = $128;
   $129 = ($127|0)==(0);
   if ($129) {
    $135 = (_luaZ_fill($126)|0);
    $136 = $135;
   } else {
    $130 = ((($126)) + 4|0);
    $131 = HEAP32[$130>>2]|0;
    $132 = ((($131)) + 1|0);
    HEAP32[$130>>2] = $132;
    $133 = HEAP8[$131>>0]|0;
    $134 = $133&255;
    $136 = $134;
   }
   HEAP32[$ls>>2] = $136;
   if ($75) {
    label = 47;
    break;
   } else {
    $$ph40 = $136;
   }
  }
  if ((label|0) == 47) {
   label = 0;
   _luaX_lexerror($ls,520017,91);
   label = 24;
   continue;
  }
  else if ((label|0) == 48) {
   label = 0;
   $137 = (_skip_sep($ls)|0);
   $138 = ($137|0)==($sep|0);
   if ($138) {
    label = 49;
    break;
   } else {
    label = 24;
    continue;
   }
  }
  else if ((label|0) == 61) {
   label = 0;
   $176 = HEAP32[$1>>2]|0;
   $177 = ((($176)) + 4|0);
   $178 = HEAP32[$177>>2]|0;
   $179 = (($178) + 1)|0;
   $180 = ((($176)) + 8|0);
   $181 = HEAP32[$180>>2]|0;
   $182 = ($179>>>0)>($181>>>0);
   if ($182) {
    $183 = ($181>>>0)>(2147483645);
    if ($183) {
     label = 64;
     break;
    }
    $189 = $181 << 1;
    $190 = ($189|0)==(-2);
    $191 = HEAP32[$73>>2]|0;
    if ($190) {
     $194 = (_luaM_toobig($191)|0);
     $$pre$phi$i22Z2D = $176;$195 = $194;
    } else {
     $192 = HEAP32[$176>>2]|0;
     $193 = (_luaM_realloc_($191,$192,$181,$189)|0);
     $$pre$phi$i22Z2D = $176;$195 = $193;
    }
    HEAP32[$$pre$phi$i22Z2D>>2] = $195;
    HEAP32[$180>>2] = $189;
    $$pre1$i23 = HEAP32[$177>>2]|0;
    $197 = $$pre1$i23;$199 = $195;
   } else {
    $$pre2$i18 = HEAP32[$176>>2]|0;
    $197 = $178;$199 = $$pre2$i18;
   }
   $196 = (($197) + 1)|0;
   HEAP32[$177>>2] = $196;
   $198 = (($199) + ($197)|0);
   HEAP8[$198>>0] = 10;
   $200 = HEAP32[$ls>>2]|0;
   $201 = HEAP32[$30>>2]|0;
   $202 = HEAP32[$201>>2]|0;
   $203 = (($202) + -1)|0;
   HEAP32[$201>>2] = $203;
   $204 = ($202|0)==(0);
   if ($204) {
    $210 = (_luaZ_fill($201)|0);
    $211 = $210;
   } else {
    $205 = ((($201)) + 4|0);
    $206 = HEAP32[$205>>2]|0;
    $207 = ((($206)) + 1|0);
    HEAP32[$205>>2] = $207;
    $208 = HEAP8[$206>>0]|0;
    $209 = $208&255;
    $211 = $209;
   }
   HEAP32[$ls>>2] = $211;
   switch ($211|0) {
   case 13: case 10:  {
    $212 = ($211|0)==($200|0);
    if (!($212)) {
     $213 = HEAP32[$30>>2]|0;
     $214 = HEAP32[$213>>2]|0;
     $215 = (($214) + -1)|0;
     HEAP32[$213>>2] = $215;
     $216 = ($214|0)==(0);
     if ($216) {
      $222 = (_luaZ_fill($213)|0);
      $223 = $222;
     } else {
      $217 = ((($213)) + 4|0);
      $218 = HEAP32[$217>>2]|0;
      $219 = ((($218)) + 1|0);
      HEAP32[$217>>2] = $219;
      $220 = HEAP8[$218>>0]|0;
      $221 = $220&255;
      $223 = $221;
     }
     HEAP32[$ls>>2] = $223;
    }
    break;
   }
   default: {
   }
   }
   $224 = HEAP32[$74>>2]|0;
   $225 = (($224) + 1)|0;
   HEAP32[$74>>2] = $225;
   $226 = ($224|0)>(2147483643);
   if ($226) {
    $227 = ((($ls)) + 16|0);
    $228 = HEAP32[$227>>2]|0;
    _luaX_lexerror($ls,519945,$228);
   }
   if (!($71)) {
    label = 24;
    continue;
   }
   $229 = HEAP32[$1>>2]|0;
   $230 = ((($229)) + 4|0);
   HEAP32[$230>>2] = 0;
   label = 24;
   continue;
  }
 }
 if ((label|0) == 33) {
  $89 = ($seminfo|0)!=(0|0);
  $90 = $89 ? 519970 : 519993;
  $91 = HEAP32[$72>>2]|0;
  $92 = ((($91)) + 16|0);
  _luaO_chunkid($buff$i,$92,80);
  $93 = HEAP32[$73>>2]|0;
  $94 = HEAP32[$74>>2]|0;
  HEAP32[$vararg_buffer3>>2] = $buff$i;
  $vararg_ptr6 = ((($vararg_buffer3)) + 4|0);
  HEAP32[$vararg_ptr6>>2] = $94;
  $vararg_ptr7 = ((($vararg_buffer3)) + 8|0);
  HEAP32[$vararg_ptr7>>2] = $90;
  $95 = (_luaO_pushfstring($93,519897,$vararg_buffer3)|0);
  $96 = HEAP32[$73>>2]|0;
  HEAP32[$vararg_buffer8>>2] = $95;
  $vararg_ptr11 = ((($vararg_buffer8)) + 4|0);
  HEAP32[$vararg_ptr11>>2] = 519882;
  (_luaO_pushfstring($96,519932,$vararg_buffer8)|0);
  $97 = HEAP32[$73>>2]|0;
  _luaD_throw($97,3);
  // unreachable;
 }
 else if ((label|0) == 38) {
  $109 = HEAP32[$72>>2]|0;
  $110 = ((($109)) + 16|0);
  _luaO_chunkid($buff$i35,$110,80);
  $111 = HEAP32[$73>>2]|0;
  $112 = HEAP32[$74>>2]|0;
  HEAP32[$vararg_buffer12>>2] = $buff$i35;
  $vararg_ptr15 = ((($vararg_buffer12)) + 4|0);
  HEAP32[$vararg_ptr15>>2] = $112;
  $vararg_ptr16 = ((($vararg_buffer12)) + 8|0);
  HEAP32[$vararg_ptr16>>2] = 519907;
  (_luaO_pushfstring($111,519897,$vararg_buffer12)|0);
  $113 = HEAP32[$73>>2]|0;
  _luaD_throw($113,3);
  // unreachable;
 }
 else if ((label|0) == 49) {
  $139 = HEAP32[$ls>>2]|0;
  $140 = HEAP32[$1>>2]|0;
  $141 = ((($140)) + 4|0);
  $142 = HEAP32[$141>>2]|0;
  $143 = (($142) + 1)|0;
  $144 = ((($140)) + 8|0);
  $145 = HEAP32[$144>>2]|0;
  $146 = ($143>>>0)>($145>>>0);
  if ($146) {
   $147 = ($145>>>0)>(2147483645);
   if ($147) {
    $148 = HEAP32[$72>>2]|0;
    $149 = ((($148)) + 16|0);
    _luaO_chunkid($buff$i36,$149,80);
    $150 = HEAP32[$73>>2]|0;
    $151 = HEAP32[$74>>2]|0;
    HEAP32[$vararg_buffer17>>2] = $buff$i36;
    $vararg_ptr20 = ((($vararg_buffer17)) + 4|0);
    HEAP32[$vararg_ptr20>>2] = $151;
    $vararg_ptr21 = ((($vararg_buffer17)) + 8|0);
    HEAP32[$vararg_ptr21>>2] = 519907;
    (_luaO_pushfstring($150,519897,$vararg_buffer17)|0);
    $152 = HEAP32[$73>>2]|0;
    _luaD_throw($152,3);
    // unreachable;
   }
   $153 = $145 << 1;
   $154 = ($153|0)==(-2);
   $155 = HEAP32[$73>>2]|0;
   if ($154) {
    $158 = (_luaM_toobig($155)|0);
    $$pre$phi$i14Z2D = $140;$159 = $158;
   } else {
    $156 = HEAP32[$140>>2]|0;
    $157 = (_luaM_realloc_($155,$156,$145,$153)|0);
    $$pre$phi$i14Z2D = $140;$159 = $157;
   }
   HEAP32[$$pre$phi$i14Z2D>>2] = $159;
   HEAP32[$144>>2] = $153;
   $$pre1$i15 = HEAP32[$141>>2]|0;
   $162 = $$pre1$i15;$164 = $159;
  } else {
   $$pre2$i10 = HEAP32[$140>>2]|0;
   $162 = $142;$164 = $$pre2$i10;
  }
  $160 = $139&255;
  $161 = (($162) + 1)|0;
  HEAP32[$141>>2] = $161;
  $163 = (($164) + ($162)|0);
  HEAP8[$163>>0] = $160;
  $165 = HEAP32[$30>>2]|0;
  $166 = HEAP32[$165>>2]|0;
  $167 = (($166) + -1)|0;
  HEAP32[$165>>2] = $167;
  $168 = ($166|0)==(0);
  if ($168) {
   $174 = (_luaZ_fill($165)|0);
   $175 = $174;
  } else {
   $169 = ((($165)) + 4|0);
   $170 = HEAP32[$169>>2]|0;
   $171 = ((($170)) + 1|0);
   HEAP32[$169>>2] = $171;
   $172 = HEAP8[$170>>0]|0;
   $173 = $172&255;
   $175 = $173;
  }
  HEAP32[$ls>>2] = $175;
  if ($71) {
   STACKTOP = sp;return;
  }
  $267 = HEAP32[$1>>2]|0;
  $268 = HEAP32[$267>>2]|0;
  $269 = (($sep) + 2)|0;
  $270 = (($268) + ($269)|0);
  $271 = ((($267)) + 4|0);
  $272 = HEAP32[$271>>2]|0;
  $273 = $269 << 1;
  $274 = (($272) - ($273))|0;
  $275 = HEAP32[$73>>2]|0;
  $276 = (_luaS_newlstr($275,$270,$274)|0);
  $277 = ((($ls)) + 48|0);
  $278 = HEAP32[$277>>2]|0;
  $279 = ((($278)) + 4|0);
  $280 = HEAP32[$279>>2]|0;
  $281 = (_luaH_setstr($275,$280,$276)|0);
  $282 = ((($281)) + 8|0);
  $283 = HEAP32[$282>>2]|0;
  $284 = ($283|0)==(0);
  if ($284) {
   HEAP32[$281>>2] = 1;
   HEAP32[$282>>2] = 1;
   $285 = ((($275)) + 16|0);
   $286 = HEAP32[$285>>2]|0;
   $287 = ((($286)) + 68|0);
   $288 = HEAP32[$287>>2]|0;
   $289 = ((($286)) + 64|0);
   $290 = HEAP32[$289>>2]|0;
   $291 = ($288>>>0)<($290>>>0);
   if (!($291)) {
    _luaC_step($275);
   }
  }
  HEAP32[$seminfo>>2] = $276;
  STACKTOP = sp;return;
 }
 else if ((label|0) == 64) {
  $184 = HEAP32[$72>>2]|0;
  $185 = ((($184)) + 16|0);
  _luaO_chunkid($buff$i37,$185,80);
  $186 = HEAP32[$73>>2]|0;
  $187 = HEAP32[$74>>2]|0;
  HEAP32[$vararg_buffer22>>2] = $buff$i37;
  $vararg_ptr25 = ((($vararg_buffer22)) + 4|0);
  HEAP32[$vararg_ptr25>>2] = $187;
  $vararg_ptr26 = ((($vararg_buffer22)) + 8|0);
  HEAP32[$vararg_ptr26>>2] = 519907;
  (_luaO_pushfstring($186,519897,$vararg_buffer22)|0);
  $188 = HEAP32[$73>>2]|0;
  _luaD_throw($188,3);
  // unreachable;
 }
 else if ((label|0) == 85) {
  $239 = HEAP32[$72>>2]|0;
  $240 = ((($239)) + 16|0);
  _luaO_chunkid($buff$i38,$240,80);
  $241 = HEAP32[$73>>2]|0;
  $242 = HEAP32[$74>>2]|0;
  HEAP32[$vararg_buffer27>>2] = $buff$i38;
  $vararg_ptr30 = ((($vararg_buffer27)) + 4|0);
  HEAP32[$vararg_ptr30>>2] = $242;
  $vararg_ptr31 = ((($vararg_buffer27)) + 8|0);
  HEAP32[$vararg_ptr31>>2] = 519907;
  (_luaO_pushfstring($241,519897,$vararg_buffer27)|0);
  $243 = HEAP32[$73>>2]|0;
  _luaD_throw($243,3);
  // unreachable;
 }
}
function _read_numeral($ls,$seminfo) {
 $ls = $ls|0;
 $seminfo = $seminfo|0;
 var $$in$i = 0, $$in$i$i = 0, $$in$i4$i = 0, $$lcssa = 0, $$pre = 0, $$pre$i22 = 0, $$pre$phi$i10Z2D = 0, $$pre$phi$i18Z2D = 0, $$pre$phi$i29Z2D = 0, $$pre$phi$i37Z2D = 0, $$pre$phi$iZ2D = 0, $$pre1$i = 0, $$pre1$i11 = 0, $$pre1$i19 = 0, $$pre1$i30 = 0, $$pre1$i38 = 0, $$pre2$i = 0, $$pre2$i14 = 0, $$pre2$i25 = 0, $$pre2$i33 = 0;
 var $$pre2$i6 = 0, $$pre48 = 0, $$pre49 = 0, $$pre8$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $buff$i = 0, $buff$i40 = 0, $buff$i41 = 0, $buff$i42 = 0, $buff$i43 = 0, $isdigit = 0, $isdigittmp = 0, $ls$idx$val = 0, $ls$idx$val$i = 0, $ls$idx$val$idx$val = 0, $ls$idx$val$idx$val$i = 0, $ls$idx$val$idx2 = 0, $ls$idx$val$idx2$val = 0, $ls$idx$val$idx3$i = 0, $ls$idx$val$idx3$val$i = 0, $ls$idx1$val$i = 0, $ls$idx1$val$idx$val$i = 0, $ls$idx1$val$idx2$i = 0, $ls$idx1$val$idx2$val$i = 0;
 var $memchr = 0, $memchr44 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer13 = 0, $vararg_buffer18 = 0, $vararg_buffer3 = 0, $vararg_buffer8 = 0, $vararg_ptr1 = 0, $vararg_ptr11 = 0, $vararg_ptr12 = 0, $vararg_ptr16 = 0, $vararg_ptr17 = 0, $vararg_ptr2 = 0, $vararg_ptr21 = 0, $vararg_ptr22 = 0, $vararg_ptr6 = 0, $vararg_ptr7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 480|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer18 = sp + 64|0;
 $vararg_buffer13 = sp + 48|0;
 $vararg_buffer8 = sp + 32|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $buff$i43 = sp + 400|0;
 $buff$i42 = sp + 320|0;
 $buff$i41 = sp + 240|0;
 $buff$i40 = sp + 160|0;
 $buff$i = sp + 80|0;
 $0 = ((($ls)) + 60|0);
 $1 = ((($ls)) + 64|0);
 $2 = ((($ls)) + 52|0);
 $3 = ((($ls)) + 4|0);
 $4 = ((($ls)) + 56|0);
 $$pre = HEAP32[$ls>>2]|0;
 $26 = $$pre;
 while(1) {
  $5 = HEAP32[$0>>2]|0;
  $6 = ((($5)) + 4|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (($7) + 1)|0;
  $9 = ((($5)) + 8|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ($8>>>0)>($10>>>0);
  if ($11) {
   $12 = ($10>>>0)>(2147483645);
   if ($12) {
    label = 5;
    break;
   }
   $18 = $10 << 1;
   $19 = ($18|0)==(-2);
   $20 = HEAP32[$2>>2]|0;
   if ($19) {
    $23 = (_luaM_toobig($20)|0);
    $$pre$phi$iZ2D = $5;$24 = $23;
   } else {
    $21 = HEAP32[$5>>2]|0;
    $22 = (_luaM_realloc_($20,$21,$10,$18)|0);
    $$pre$phi$iZ2D = $5;$24 = $22;
   }
   HEAP32[$$pre$phi$iZ2D>>2] = $24;
   HEAP32[$9>>2] = $18;
   $$pre1$i = HEAP32[$6>>2]|0;
   $28 = $$pre1$i;$30 = $24;
  } else {
   $$pre2$i = HEAP32[$5>>2]|0;
   $28 = $7;$30 = $$pre2$i;
  }
  $25 = $26&255;
  $27 = (($28) + 1)|0;
  HEAP32[$6>>2] = $27;
  $29 = (($30) + ($28)|0);
  HEAP8[$29>>0] = $25;
  $31 = HEAP32[$4>>2]|0;
  $32 = HEAP32[$31>>2]|0;
  $33 = (($32) + -1)|0;
  HEAP32[$31>>2] = $33;
  $34 = ($32|0)==(0);
  if ($34) {
   $40 = (_luaZ_fill($31)|0);
   $41 = $40;
  } else {
   $35 = ((($31)) + 4|0);
   $36 = HEAP32[$35>>2]|0;
   $37 = ((($36)) + 1|0);
   HEAP32[$35>>2] = $37;
   $38 = HEAP8[$36>>0]|0;
   $39 = $38&255;
   $41 = $39;
  }
  HEAP32[$ls>>2] = $41;
  $isdigittmp = (($41) + -48)|0;
  $isdigit = ($isdigittmp>>>0)<(10);
  $42 = ($41|0)==(46);
  $or$cond = $42 | $isdigit;
  if ($or$cond) {
   $26 = $41;
  } else {
   $$lcssa = $41;
   break;
  }
 }
 if ((label|0) == 5) {
  $13 = HEAP32[$1>>2]|0;
  $14 = ((($13)) + 16|0);
  _luaO_chunkid($buff$i,$14,80);
  $15 = HEAP32[$2>>2]|0;
  $16 = HEAP32[$3>>2]|0;
  HEAP32[$vararg_buffer>>2] = $buff$i;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $16;
  $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
  HEAP32[$vararg_ptr2>>2] = 519907;
  (_luaO_pushfstring($15,519897,$vararg_buffer)|0);
  $17 = HEAP32[$2>>2]|0;
  _luaD_throw($17,3);
  // unreachable;
 }
 $memchr = (_memchr(520124,$$lcssa,3)|0);
 $43 = ($memchr|0)==(0|0);
 if ($43) {
  $216 = $$lcssa;
 } else {
  $44 = HEAP32[$0>>2]|0;
  $45 = ((($44)) + 4|0);
  $46 = HEAP32[$45>>2]|0;
  $47 = (($46) + 1)|0;
  $48 = ((($44)) + 8|0);
  $49 = HEAP32[$48>>2]|0;
  $50 = ($47>>>0)>($49>>>0);
  if ($50) {
   $51 = ($49>>>0)>(2147483645);
   if ($51) {
    $52 = HEAP32[$1>>2]|0;
    $53 = ((($52)) + 16|0);
    _luaO_chunkid($buff$i42,$53,80);
    $54 = HEAP32[$2>>2]|0;
    $55 = HEAP32[$3>>2]|0;
    HEAP32[$vararg_buffer3>>2] = $buff$i42;
    $vararg_ptr6 = ((($vararg_buffer3)) + 4|0);
    HEAP32[$vararg_ptr6>>2] = $55;
    $vararg_ptr7 = ((($vararg_buffer3)) + 8|0);
    HEAP32[$vararg_ptr7>>2] = 519907;
    (_luaO_pushfstring($54,519897,$vararg_buffer3)|0);
    $56 = HEAP32[$2>>2]|0;
    _luaD_throw($56,3);
    // unreachable;
   }
   $57 = $49 << 1;
   $58 = ($57|0)==(-2);
   $59 = HEAP32[$2>>2]|0;
   if ($58) {
    $62 = (_luaM_toobig($59)|0);
    $$pre$phi$i29Z2D = $44;$63 = $62;
   } else {
    $60 = HEAP32[$44>>2]|0;
    $61 = (_luaM_realloc_($59,$60,$49,$57)|0);
    $$pre$phi$i29Z2D = $44;$63 = $61;
   }
   HEAP32[$$pre$phi$i29Z2D>>2] = $63;
   HEAP32[$48>>2] = $57;
   $$pre1$i30 = HEAP32[$45>>2]|0;
   $66 = $$pre1$i30;$68 = $63;
  } else {
   $$pre2$i25 = HEAP32[$44>>2]|0;
   $66 = $46;$68 = $$pre2$i25;
  }
  $64 = $$lcssa&255;
  $65 = (($66) + 1)|0;
  HEAP32[$45>>2] = $65;
  $67 = (($68) + ($66)|0);
  HEAP8[$67>>0] = $64;
  $69 = HEAP32[$4>>2]|0;
  $70 = HEAP32[$69>>2]|0;
  $71 = (($70) + -1)|0;
  HEAP32[$69>>2] = $71;
  $72 = ($70|0)==(0);
  if ($72) {
   $78 = (_luaZ_fill($69)|0);
   $79 = $78;
  } else {
   $73 = ((($69)) + 4|0);
   $74 = HEAP32[$73>>2]|0;
   $75 = ((($74)) + 1|0);
   HEAP32[$73>>2] = $75;
   $76 = HEAP8[$74>>0]|0;
   $77 = $76&255;
   $79 = $77;
  }
  HEAP32[$ls>>2] = $79;
  $memchr44 = (_memchr(525092,$79,3)|0);
  $80 = ($memchr44|0)==(0|0);
  if ($80) {
   $216 = $79;
  } else {
   $81 = HEAP32[$0>>2]|0;
   $82 = ((($81)) + 4|0);
   $83 = HEAP32[$82>>2]|0;
   $84 = (($83) + 1)|0;
   $85 = ((($81)) + 8|0);
   $86 = HEAP32[$85>>2]|0;
   $87 = ($84>>>0)>($86>>>0);
   if ($87) {
    $88 = ($86>>>0)>(2147483645);
    if ($88) {
     $89 = HEAP32[$1>>2]|0;
     $90 = ((($89)) + 16|0);
     _luaO_chunkid($buff$i43,$90,80);
     $91 = HEAP32[$2>>2]|0;
     $92 = HEAP32[$3>>2]|0;
     HEAP32[$vararg_buffer8>>2] = $buff$i43;
     $vararg_ptr11 = ((($vararg_buffer8)) + 4|0);
     HEAP32[$vararg_ptr11>>2] = $92;
     $vararg_ptr12 = ((($vararg_buffer8)) + 8|0);
     HEAP32[$vararg_ptr12>>2] = 519907;
     (_luaO_pushfstring($91,519897,$vararg_buffer8)|0);
     $93 = HEAP32[$2>>2]|0;
     _luaD_throw($93,3);
     // unreachable;
    }
    $94 = $86 << 1;
    $95 = ($94|0)==(-2);
    $96 = HEAP32[$2>>2]|0;
    if ($95) {
     $99 = (_luaM_toobig($96)|0);
     $$pre$phi$i37Z2D = $81;$100 = $99;
    } else {
     $97 = HEAP32[$81>>2]|0;
     $98 = (_luaM_realloc_($96,$97,$86,$94)|0);
     $$pre$phi$i37Z2D = $81;$100 = $98;
    }
    HEAP32[$$pre$phi$i37Z2D>>2] = $100;
    HEAP32[$85>>2] = $94;
    $$pre1$i38 = HEAP32[$82>>2]|0;
    $103 = $$pre1$i38;$105 = $100;
   } else {
    $$pre2$i33 = HEAP32[$81>>2]|0;
    $103 = $83;$105 = $$pre2$i33;
   }
   $101 = $79&255;
   $102 = (($103) + 1)|0;
   HEAP32[$82>>2] = $102;
   $104 = (($105) + ($103)|0);
   HEAP8[$104>>0] = $101;
   $106 = HEAP32[$4>>2]|0;
   $107 = HEAP32[$106>>2]|0;
   $108 = (($107) + -1)|0;
   HEAP32[$106>>2] = $108;
   $109 = ($107|0)==(0);
   if ($109) {
    $115 = (_luaZ_fill($106)|0);
    $116 = $115;
   } else {
    $110 = ((($106)) + 4|0);
    $111 = HEAP32[$110>>2]|0;
    $112 = ((($111)) + 1|0);
    HEAP32[$110>>2] = $112;
    $113 = HEAP8[$111>>0]|0;
    $114 = $113&255;
    $116 = $114;
   }
   HEAP32[$ls>>2] = $116;
   $216 = $116;
  }
 }
 $117 = $216;
 while(1) {
  $118 = (_isalnum($117)|0);
  $119 = ($118|0)==(0);
  $120 = HEAP32[$ls>>2]|0;
  if ($119) {
   $121 = ($120|0)==(95);
   if ($121) {
    $143 = 95;
   } else {
    break;
   }
  } else {
   $143 = $120;
  }
  $122 = HEAP32[$0>>2]|0;
  $123 = ((($122)) + 4|0);
  $124 = HEAP32[$123>>2]|0;
  $125 = (($124) + 1)|0;
  $126 = ((($122)) + 8|0);
  $127 = HEAP32[$126>>2]|0;
  $128 = ($125>>>0)>($127>>>0);
  if ($128) {
   $129 = ($127>>>0)>(2147483645);
   if ($129) {
    label = 45;
    break;
   }
   $135 = $127 << 1;
   $136 = ($135|0)==(-2);
   $137 = HEAP32[$2>>2]|0;
   if ($136) {
    $140 = (_luaM_toobig($137)|0);
    $$pre$phi$i10Z2D = $122;$141 = $140;
   } else {
    $138 = HEAP32[$122>>2]|0;
    $139 = (_luaM_realloc_($137,$138,$127,$135)|0);
    $$pre$phi$i10Z2D = $122;$141 = $139;
   }
   HEAP32[$$pre$phi$i10Z2D>>2] = $141;
   HEAP32[$126>>2] = $135;
   $$pre1$i11 = HEAP32[$123>>2]|0;
   $145 = $$pre1$i11;$147 = $141;
  } else {
   $$pre2$i6 = HEAP32[$122>>2]|0;
   $145 = $124;$147 = $$pre2$i6;
  }
  $142 = $143&255;
  $144 = (($145) + 1)|0;
  HEAP32[$123>>2] = $144;
  $146 = (($147) + ($145)|0);
  HEAP8[$146>>0] = $142;
  $148 = HEAP32[$4>>2]|0;
  $149 = HEAP32[$148>>2]|0;
  $150 = (($149) + -1)|0;
  HEAP32[$148>>2] = $150;
  $151 = ($149|0)==(0);
  if ($151) {
   $157 = (_luaZ_fill($148)|0);
   $158 = $157;
  } else {
   $152 = ((($148)) + 4|0);
   $153 = HEAP32[$152>>2]|0;
   $154 = ((($153)) + 1|0);
   HEAP32[$152>>2] = $154;
   $155 = HEAP8[$153>>0]|0;
   $156 = $155&255;
   $158 = $156;
  }
  HEAP32[$ls>>2] = $158;
  $117 = $158;
 }
 if ((label|0) == 45) {
  $130 = HEAP32[$1>>2]|0;
  $131 = ((($130)) + 16|0);
  _luaO_chunkid($buff$i40,$131,80);
  $132 = HEAP32[$2>>2]|0;
  $133 = HEAP32[$3>>2]|0;
  HEAP32[$vararg_buffer13>>2] = $buff$i40;
  $vararg_ptr16 = ((($vararg_buffer13)) + 4|0);
  HEAP32[$vararg_ptr16>>2] = $133;
  $vararg_ptr17 = ((($vararg_buffer13)) + 8|0);
  HEAP32[$vararg_ptr17>>2] = 519907;
  (_luaO_pushfstring($132,519897,$vararg_buffer13)|0);
  $134 = HEAP32[$2>>2]|0;
  _luaD_throw($134,3);
  // unreachable;
 }
 $159 = HEAP32[$0>>2]|0;
 $160 = ((($159)) + 4|0);
 $161 = HEAP32[$160>>2]|0;
 $162 = (($161) + 1)|0;
 $163 = ((($159)) + 8|0);
 $164 = HEAP32[$163>>2]|0;
 $165 = ($162>>>0)>($164>>>0);
 if ($165) {
  $166 = ($164>>>0)>(2147483645);
  if ($166) {
   $167 = HEAP32[$1>>2]|0;
   $168 = ((($167)) + 16|0);
   _luaO_chunkid($buff$i41,$168,80);
   $169 = HEAP32[$2>>2]|0;
   $170 = HEAP32[$3>>2]|0;
   HEAP32[$vararg_buffer18>>2] = $buff$i41;
   $vararg_ptr21 = ((($vararg_buffer18)) + 4|0);
   HEAP32[$vararg_ptr21>>2] = $170;
   $vararg_ptr22 = ((($vararg_buffer18)) + 8|0);
   HEAP32[$vararg_ptr22>>2] = 519907;
   (_luaO_pushfstring($169,519897,$vararg_buffer18)|0);
   $171 = HEAP32[$2>>2]|0;
   _luaD_throw($171,3);
   // unreachable;
  }
  $172 = $164 << 1;
  $173 = ($172|0)==(-2);
  $174 = HEAP32[$2>>2]|0;
  if ($173) {
   $177 = (_luaM_toobig($174)|0);
   $$pre$phi$i18Z2D = $159;$178 = $177;
  } else {
   $175 = HEAP32[$159>>2]|0;
   $176 = (_luaM_realloc_($174,$175,$164,$172)|0);
   $$pre$phi$i18Z2D = $159;$178 = $176;
  }
  HEAP32[$$pre$phi$i18Z2D>>2] = $178;
  HEAP32[$163>>2] = $172;
  $$pre1$i19 = HEAP32[$160>>2]|0;
  $180 = $$pre1$i19;$182 = $178;
 } else {
  $$pre2$i14 = HEAP32[$159>>2]|0;
  $180 = $161;$182 = $$pre2$i14;
 }
 $179 = (($180) + 1)|0;
 HEAP32[$160>>2] = $179;
 $181 = (($182) + ($180)|0);
 HEAP8[$181>>0] = 0;
 $183 = ((($ls)) + 68|0);
 $184 = HEAP8[$183>>0]|0;
 $ls$idx$val = HEAP32[$0>>2]|0;
 $ls$idx$val$idx$val = HEAP32[$ls$idx$val>>2]|0;
 $ls$idx$val$idx2 = ((($ls$idx$val)) + 4|0);
 $ls$idx$val$idx2$val = HEAP32[$ls$idx$val$idx2>>2]|0;
 $185 = ($ls$idx$val$idx2$val|0)==(0);
 if ($185) {
  $191 = $ls$idx$val$idx$val;
 } else {
  $$in$i = $ls$idx$val$idx2$val;
  while(1) {
   $186 = (($$in$i) + -1)|0;
   $187 = (($ls$idx$val$idx$val) + ($186)|0);
   $188 = HEAP8[$187>>0]|0;
   $189 = ($188<<24>>24)==(46);
   if ($189) {
    HEAP8[$187>>0] = $184;
   }
   $190 = ($186|0)==(0);
   if ($190) {
    break;
   } else {
    $$in$i = $186;
   }
  }
  $$pre48 = HEAP32[$0>>2]|0;
  $$pre49 = HEAP32[$$pre48>>2]|0;
  $191 = $$pre49;
 }
 $192 = (_luaO_str2d($191,$seminfo)|0);
 $193 = ($192|0)==(0);
 if (!($193)) {
  STACKTOP = sp;return;
 }
 $194 = (_localeconv()|0);
 $195 = HEAP8[$183>>0]|0;
 $196 = ($194|0)==(0|0);
 if ($196) {
  $199 = 46;
 } else {
  $197 = HEAP32[$194>>2]|0;
  $198 = HEAP8[$197>>0]|0;
  $199 = $198;
 }
 HEAP8[$183>>0] = $199;
 $ls$idx1$val$i = HEAP32[$0>>2]|0;
 $ls$idx1$val$idx$val$i = HEAP32[$ls$idx1$val$i>>2]|0;
 $ls$idx1$val$idx2$i = ((($ls$idx1$val$i)) + 4|0);
 $ls$idx1$val$idx2$val$i = HEAP32[$ls$idx1$val$idx2$i>>2]|0;
 $200 = ($ls$idx1$val$idx2$val$i|0)==(0);
 if ($200) {
  $206 = $ls$idx1$val$idx$val$i;
 } else {
  $$in$i$i = $ls$idx1$val$idx2$val$i;
  while(1) {
   $201 = (($$in$i$i) + -1)|0;
   $202 = (($ls$idx1$val$idx$val$i) + ($201)|0);
   $203 = HEAP8[$202>>0]|0;
   $204 = ($203<<24>>24)==($195<<24>>24);
   if ($204) {
    HEAP8[$202>>0] = $199;
   }
   $205 = ($201|0)==(0);
   if ($205) {
    break;
   } else {
    $$in$i$i = $201;
   }
  }
  $$pre$i22 = HEAP32[$0>>2]|0;
  $$pre8$i = HEAP32[$$pre$i22>>2]|0;
  $206 = $$pre8$i;
 }
 $207 = (_luaO_str2d($206,$seminfo)|0);
 $208 = ($207|0)==(0);
 if (!($208)) {
  STACKTOP = sp;return;
 }
 $209 = HEAP8[$183>>0]|0;
 $ls$idx$val$i = HEAP32[$0>>2]|0;
 $ls$idx$val$idx$val$i = HEAP32[$ls$idx$val$i>>2]|0;
 $ls$idx$val$idx3$i = ((($ls$idx$val$i)) + 4|0);
 $ls$idx$val$idx3$val$i = HEAP32[$ls$idx$val$idx3$i>>2]|0;
 $210 = ($ls$idx$val$idx3$val$i|0)==(0);
 if (!($210)) {
  $$in$i4$i = $ls$idx$val$idx3$val$i;
  while(1) {
   $211 = (($$in$i4$i) + -1)|0;
   $212 = (($ls$idx$val$idx$val$i) + ($211)|0);
   $213 = HEAP8[$212>>0]|0;
   $214 = ($213<<24>>24)==($209<<24>>24);
   if ($214) {
    HEAP8[$212>>0] = 46;
   }
   $215 = ($211|0)==(0);
   if ($215) {
    break;
   } else {
    $$in$i4$i = $211;
   }
  }
 }
 _luaX_lexerror($ls,520127,284);
 STACKTOP = sp;return;
}
function _luaX_lookahead($ls) {
 $ls = $ls|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($ls)) + 40|0);
 $1 = (_llex($ls,$0)|0);
 $2 = ((($ls)) + 32|0);
 HEAP32[$2>>2] = $1;
 return;
}
function _luaopen_base($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _lua_pushvalue($L,-10002);
 _lua_setfield($L,-10002,520144);
 _luaL_register($L,520144,39084);
 _lua_pushlstring($L,520685,7);
 _lua_setfield($L,-10002,520693);
 _lua_pushcclosure($L,4308,0);
 _lua_pushcclosure($L,4309,1);
 _lua_setfield($L,-2,520702);
 _lua_pushcclosure($L,2212,0);
 _lua_pushcclosure($L,4310,1);
 _lua_setfield($L,-2,520709);
 _lua_createtable($L,0,1);
 _lua_pushvalue($L,-1);
 (_lua_setmetatable($L,-2)|0);
 _lua_pushlstring($L,520715,2);
 _lua_setfield($L,-2,520718);
 _lua_pushcclosure($L,4311,1);
 _lua_setfield($L,-10002,520751);
 _luaL_register($L,520760,39344);
 return 2;
}
function _luaB_assert($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _luaL_checkany($L,1);
 $0 = (_lua_toboolean($L,1)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  $2 = (_luaL_optlstring($L,2,520154,0)|0);
  HEAP32[$vararg_buffer>>2] = $2;
  $3 = (_luaL_error($L,595867,$vararg_buffer)|0);
  $$0 = $3;
  STACKTOP = sp;return ($$0|0);
 } else {
  $4 = (_lua_gettop($L)|0);
  $$0 = $4;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _luaB_collectgarbage($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkoption($L,1,520187,39284)|0);
 $1 = (_luaL_optinteger($L,2,0)|0);
 $2 = (39316 + ($0<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_lua_gc($L,$3,$1)|0);
 switch ($3|0) {
 case 3:  {
  $5 = (_lua_gc($L,4,0)|0);
  $6 = (+($4|0));
  $7 = (+($5|0));
  $8 = $7 * 9.765625E-4;
  $9 = $6 + $8;
  _lua_pushnumber($L,$9);
  return 1;
  break;
 }
 case 5:  {
  _lua_pushboolean($L,$4);
  return 1;
  break;
 }
 default: {
  $10 = (+($4|0));
  _lua_pushnumber($L,$10);
  return 1;
 }
 }
 return (0)|0;
}
function _luaB_dofile($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_optlstring($L,1,0,0)|0);
 $1 = (_lua_gettop($L)|0);
 $2 = (_luaL_loadfile($L,$0)|0);
 $3 = ($2|0)==(0);
 if (!($3)) {
  (_lua_error($L)|0);
 }
 _lua_call($L,0,-1);
 $4 = (_lua_gettop($L)|0);
 $5 = (($4) - ($1))|0;
 return ($5|0);
}
function _luaB_error($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_optinteger($L,2,1)|0);
 _lua_settop($L,1);
 $1 = (_lua_isstring($L,1)|0);
 $2 = ($1|0)!=(0);
 $3 = ($0|0)>(0);
 $or$cond = $3 & $2;
 if (!($or$cond)) {
  $4 = (_lua_error($L)|0);
  return ($4|0);
 }
 _luaL_where($L,$0);
 _lua_pushvalue($L,1);
 _lua_concat($L,2);
 $4 = (_lua_error($L)|0);
 return ($4|0);
}
function _luaB_gcinfo($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_gc($L,3,0)|0);
 _lua_pushinteger($L,$0);
 return 1;
}
function _luaB_getfenv($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _getfunc($L,1);
 $0 = (_lua_iscfunction($L,-1)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  _lua_getfenv($L,-1);
  return 1;
 } else {
  _lua_pushvalue($L,-10002);
  return 1;
 }
 return (0)|0;
}
function _getfunc($L,$opt) {
 $L = $L|0;
 $opt = $opt|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ar = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $ar = sp + 4|0;
 $0 = (_lua_type($L,1)|0);
 $1 = ($0|0)==(6);
 if ($1) {
  _lua_pushvalue($L,1);
  STACKTOP = sp;return;
 }
 $2 = ($opt|0)==(0);
 if ($2) {
  $4 = (_luaL_checkinteger($L,1)|0);
  $6 = $4;
 } else {
  $3 = (_luaL_optinteger($L,1,1)|0);
  $6 = $3;
 }
 $5 = ($6|0)>(-1);
 if (!($5)) {
  (_luaL_argerror($L,1,520253)|0);
 }
 $7 = (_lua_getstack($L,$6,$ar)|0);
 $8 = ($7|0)==(0);
 if ($8) {
  (_luaL_argerror($L,1,520280)|0);
 }
 (_lua_getinfo($L,523775,$ar)|0);
 $9 = (_lua_type($L,-1)|0);
 $10 = ($9|0)==(0);
 if ($10) {
  HEAP32[$vararg_buffer>>2] = $6;
  (_luaL_error($L,520294,$vararg_buffer)|0);
 }
 STACKTOP = sp;return;
}
function _luaB_getmetatable($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checkany($L,1);
 $0 = (_lua_getmetatable($L,1)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  _lua_pushnil($L);
  return 1;
 } else {
  (_luaL_getmetafield($L,1,520344)|0);
  return 1;
 }
 return (0)|0;
}
function _luaB_loadfile($L) {
 $L = $L|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_optlstring($L,1,0,0)|0);
 $1 = (_luaL_loadfile($L,$0)|0);
 $2 = ($1|0)==(0);
 if ($2) {
  $$0$i = 1;
  return ($$0$i|0);
 }
 _lua_pushnil($L);
 _lua_insert($L,-2);
 $$0$i = 2;
 return ($$0$i|0);
}
function _luaB_load($L) {
 $L = $L|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_optlstring($L,2,520370,0)|0);
 _luaL_checktype($L,1,6);
 _lua_settop($L,3);
 $1 = (_lua_load($L,4312,0,$0)|0);
 $2 = ($1|0)==(0);
 if ($2) {
  $$0$i = 1;
  return ($$0$i|0);
 }
 _lua_pushnil($L);
 _lua_insert($L,-2);
 $$0$i = 2;
 return ($$0$i|0);
}
function _generic_reader($L,$ud,$size) {
 $L = $L|0;
 $ud = $ud|0;
 $size = $size|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _luaL_checkstack($L,2,520378);
 _lua_pushvalue($L,1);
 _lua_call($L,0,1);
 $0 = (_lua_type($L,-1)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  HEAP32[$size>>2] = 0;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = (_lua_isstring($L,-1)|0);
 $3 = ($2|0)==(0);
 if ($3) {
  (_luaL_error($L,520404,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  _lua_replace($L,3);
  $4 = (_lua_tolstring($L,3,$size)|0);
  $$0 = $4;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _luaB_loadstring($L) {
 $L = $L|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = (_luaL_checklstring($L,1,$l)|0);
 $1 = (_luaL_optlstring($L,2,$0,0)|0);
 $2 = HEAP32[$l>>2]|0;
 $3 = (_luaL_loadbuffer($L,$0,$2,$1)|0);
 $4 = ($3|0)==(0);
 if ($4) {
  $$0$i = 1;
  STACKTOP = sp;return ($$0$i|0);
 }
 _lua_pushnil($L);
 _lua_insert($L,-2);
 $$0$i = 2;
 STACKTOP = sp;return ($$0$i|0);
}
function _luaB_next($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checktype($L,1,5);
 _lua_settop($L,2);
 $0 = (_lua_next($L,1)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  $$0 = 2;
  return ($$0|0);
 }
 _lua_pushnil($L);
 $$0 = 1;
 return ($$0|0);
}
function _luaB_pcall($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checkany($L,1);
 $0 = (_lua_gettop($L)|0);
 $1 = (($0) + -1)|0;
 $2 = (_lua_pcall($L,$1,-1,0)|0);
 $3 = ($2|0)==(0);
 $4 = $3&1;
 _lua_pushboolean($L,$4);
 _lua_insert($L,1);
 $5 = (_lua_gettop($L)|0);
 return ($5|0);
}
function _luaB_print($L) {
 $L = $L|0;
 var $$2 = 0, $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $i$06 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_lua_gettop($L)|0);
 _lua_getfield($L,-10002,520469);
 $1 = ($0|0)<(1);
 $$pre = HEAP32[37602]|0;
 L1: do {
  if (!($1)) {
   $i$06 = 1;
   while(1) {
    _lua_pushvalue($L,-1);
    _lua_pushvalue($L,$i$06);
    _lua_call($L,1,1);
    $2 = (_lua_tolstring($L,-1,0)|0);
    $3 = ($2|0)==(0|0);
    if ($3) {
     break;
    }
    $4 = ($i$06|0)>(1);
    if ($4) {
     (_fputc(9,$$pre)|0);
    }
    (_fputs($2,$$pre)|0);
    _lua_settop($L,-2);
    $6 = (($i$06) + 1)|0;
    $7 = ($i$06|0)<($0|0);
    if ($7) {
     $i$06 = $6;
    } else {
     break L1;
    }
   }
   $5 = (_luaL_error($L,520478,$vararg_buffer)|0);
   $$2 = $5;
   STACKTOP = sp;return ($$2|0);
  }
 } while(0);
 (_fputc(10,$$pre)|0);
 $$2 = 0;
 STACKTOP = sp;return ($$2|0);
}
function _luaB_rawequal($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checkany($L,1);
 _luaL_checkany($L,2);
 $0 = (_lua_rawequal($L,1,2)|0);
 _lua_pushboolean($L,$0);
 return 1;
}
function _luaB_rawget($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checktype($L,1,5);
 _luaL_checkany($L,2);
 _lua_settop($L,2);
 _lua_rawget($L,1);
 return 1;
}
function _luaB_rawset($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checktype($L,1,5);
 _luaL_checkany($L,2);
 _luaL_checkany($L,3);
 _lua_settop($L,3);
 _lua_rawset($L,1);
 return 1;
}
function _luaB_select($L) {
 $L = $L|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_gettop($L)|0);
 $1 = (_lua_type($L,1)|0);
 $2 = ($1|0)==(4);
 if ($2) {
  $3 = (_lua_tolstring($L,1,0)|0);
  $4 = HEAP8[$3>>0]|0;
  $5 = ($4<<24>>24)==(35);
  if ($5) {
   $6 = (($0) + -1)|0;
   _lua_pushinteger($L,$6);
   $$0 = 1;
   return ($$0|0);
  }
 }
 $7 = (_luaL_checkinteger($L,1)|0);
 $8 = ($7|0)<(0);
 $9 = (($7) + ($0))|0;
 $10 = ($7|0)>($0|0);
 $$ = $10 ? $0 : $7;
 $i$0 = $8 ? $9 : $$;
 $11 = ($i$0|0)>(0);
 if (!($11)) {
  (_luaL_argerror($L,1,520551)|0);
 }
 $12 = (($0) - ($i$0))|0;
 $$0 = $12;
 return ($$0|0);
}
function _luaB_setfenv($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0.0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _luaL_checktype($L,2,5);
 _getfunc($L,0);
 _lua_pushvalue($L,2);
 $0 = (_lua_isnumber($L,1)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  $2 = (+_lua_tonumber($L,1));
  $3 = $2 == 0.0;
  if ($3) {
   (_lua_pushthread($L)|0);
   _lua_insert($L,-2);
   (_lua_setfenv($L,-2)|0);
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $4 = (_lua_iscfunction($L,-2)|0);
 $5 = ($4|0)==(0);
 if ($5) {
  $6 = (_lua_setfenv($L,-2)|0);
  $7 = ($6|0)==(0);
  if (!($7)) {
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_luaL_error($L,521255,$vararg_buffer)|0);
 $$0 = 1;
 STACKTOP = sp;return ($$0|0);
}
function _luaB_setmetatable($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_lua_type($L,2)|0);
 _luaL_checktype($L,1,5);
 switch ($0|0) {
 case 0: case 5:  {
  break;
 }
 default: {
  (_luaL_argerror($L,2,521337)|0);
 }
 }
 $1 = (_luaL_getmetafield($L,1,520344)|0);
 $2 = ($1|0)==(0);
 if (!($2)) {
  (_luaL_error($L,520570,$vararg_buffer)|0);
 }
 _lua_settop($L,2);
 (_lua_setmetatable($L,1)|0);
 STACKTOP = sp;return 1;
}
function _luaB_tonumber($L) {
 $L = $L|0;
 var $$lcssa = 0, $$lcssa8 = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0.0, $3 = 0;
 var $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $s2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $s2 = sp;
 $0 = (_luaL_optinteger($L,2,10)|0);
 $1 = ($0|0)==(10);
 if ($1) {
  _luaL_checkany($L,1);
  $2 = (_lua_isnumber($L,1)|0);
  $3 = ($2|0)==(0);
  if (!($3)) {
   $4 = (+_lua_tonumber($L,1));
   _lua_pushnumber($L,$4);
   STACKTOP = sp;return 1;
  }
 } else {
  $5 = (_luaL_checklstring($L,1,0)|0);
  $$off = (($0) + -2)|0;
  $6 = ($$off>>>0)<(35);
  if (!($6)) {
   (_luaL_argerror($L,2,520615)|0);
  }
  $7 = (_strtoul($5,$s2,$0)|0);
  $8 = HEAP32[$s2>>2]|0;
  $9 = ($5|0)==($8|0);
  if (!($9)) {
   $10 = HEAP8[$8>>0]|0;
   $11 = $10&255;
   $12 = (_isspace($11)|0);
   $13 = ($12|0)==(0);
   if ($13) {
    $$lcssa = $8;
   } else {
    $15 = $8;
    while(1) {
     $14 = ((($15)) + 1|0);
     $16 = HEAP8[$14>>0]|0;
     $17 = $16&255;
     $18 = (_isspace($17)|0);
     $19 = ($18|0)==(0);
     if ($19) {
      $$lcssa8 = $14;
      break;
     } else {
      $15 = $14;
     }
    }
    HEAP32[$s2>>2] = $$lcssa8;
    $$lcssa = $$lcssa8;
   }
   $20 = HEAP8[$$lcssa>>0]|0;
   $21 = ($20<<24>>24)==(0);
   if ($21) {
    $22 = (+($7>>>0));
    _lua_pushnumber($L,$22);
    STACKTOP = sp;return 1;
   }
  }
 }
 _lua_pushnil($L);
 STACKTOP = sp;return 1;
}
function _luaB_tostring($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _luaL_checkany($L,1);
 $0 = (_luaL_callmeta($L,1,521666)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  STACKTOP = sp;return 1;
 }
 $2 = (_lua_type($L,1)|0);
 switch ($2|0) {
 case 3:  {
  $3 = (_lua_tolstring($L,1,0)|0);
  _lua_pushstring($L,$3);
  STACKTOP = sp;return 1;
  break;
 }
 case 4:  {
  _lua_pushvalue($L,1);
  STACKTOP = sp;return 1;
  break;
 }
 case 1:  {
  $4 = (_lua_toboolean($L,1)|0);
  $5 = ($4|0)!=(0);
  $6 = $5 ? 622520 : 622525;
  _lua_pushstring($L,$6);
  STACKTOP = sp;return 1;
  break;
 }
 case 0:  {
  _lua_pushlstring($L,520633,3);
  STACKTOP = sp;return 1;
  break;
 }
 default: {
  $7 = (_lua_type($L,1)|0);
  $8 = (_lua_typename($L,$7)|0);
  $9 = (_lua_topointer($L,1)|0);
  HEAP32[$vararg_buffer>>2] = $8;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $9;
  (_lua_pushfstring($L,520637,$vararg_buffer)|0);
  STACKTOP = sp;return 1;
 }
 }
 return (0)|0;
}
function _luaB_type($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checkany($L,1);
 $0 = (_lua_type($L,1)|0);
 $1 = (_lua_typename($L,$0)|0);
 _lua_pushstring($L,$1);
 return 1;
}
function _luaB_unpack($L) {
 $L = $L|0;
 var $$0 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _luaL_checktype($L,1,5);
 $0 = (_luaL_optinteger($L,2,1)|0);
 $1 = (_lua_type($L,3)|0);
 $2 = ($1|0)<(1);
 if ($2) {
  $3 = (_lua_objlen($L,1)|0);
  $6 = $3;
 } else {
  $4 = (_luaL_checkinteger($L,3)|0);
  $6 = $4;
 }
 $5 = ($6|0)<($0|0);
 if ($5) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $7 = (($6) - ($0))|0;
 $8 = (($7) + 1)|0;
 $9 = ($7|0)<(0);
 if (!($9)) {
  $10 = (_lua_checkstack($L,$8)|0);
  $11 = ($10|0)==(0);
  if (!($11)) {
   _lua_rawgeti($L,1,$0);
   $13 = ($6|0)>($0|0);
   if ($13) {
    $$in = $0;
   } else {
    $$0 = $8;
    STACKTOP = sp;return ($$0|0);
   }
   while(1) {
    $14 = (($$in) + 1)|0;
    _lua_rawgeti($L,1,$14);
    $exitcond = ($14|0)==($6|0);
    if ($exitcond) {
     $$0 = $8;
     break;
    } else {
     $$in = $14;
    }
   }
   STACKTOP = sp;return ($$0|0);
  }
 }
 $12 = (_luaL_error($L,520651,$vararg_buffer)|0);
 $$0 = $12;
 STACKTOP = sp;return ($$0|0);
}
function _luaB_xpcall($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checkany($L,2);
 _lua_settop($L,2);
 _lua_insert($L,1);
 $0 = (_lua_pcall($L,0,-1,1)|0);
 $1 = ($0|0)==(0);
 $2 = $1&1;
 _lua_pushboolean($L,$2);
 _lua_replace($L,1);
 $3 = (_lua_gettop($L)|0);
 return ($3|0);
}
function _ipairsaux($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkinteger($L,2)|0);
 _luaL_checktype($L,1,5);
 $1 = (($0) + 1)|0;
 _lua_pushinteger($L,$1);
 _lua_rawgeti($L,1,$1);
 $2 = (_lua_type($L,-1)|0);
 $3 = ($2|0)==(0);
 $4 = $3 ? 0 : 2;
 return ($4|0);
}
function _luaB_ipairs($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checktype($L,1,5);
 _lua_pushvalue($L,-10003);
 _lua_pushvalue($L,1);
 _lua_pushinteger($L,0);
 return 3;
}
function _luaB_pairs($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checktype($L,1,5);
 _lua_pushvalue($L,-10003);
 _lua_pushvalue($L,1);
 _lua_pushnil($L);
 return 3;
}
function _luaB_newproxy($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _lua_settop($L,1);
 (_lua_newuserdata($L,0)|0);
 $0 = (_lua_toboolean($L,1)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  return 1;
 }
 $2 = (_lua_type($L,1)|0);
 $3 = ($2|0)==(1);
 if ($3) {
  _lua_createtable($L,0,0);
  _lua_pushvalue($L,-1);
  _lua_pushboolean($L,1);
  _lua_rawset($L,-10003);
 } else {
  $4 = (_lua_getmetatable($L,1)|0);
  $5 = ($4|0)==(0);
  if ($5) {
   label = 6;
  } else {
   _lua_rawget($L,-10003);
   $6 = (_lua_toboolean($L,-1)|0);
   _lua_settop($L,-2);
   $phitmp = ($6|0)==(0);
   if ($phitmp) {
    label = 6;
   }
  }
  if ((label|0) == 6) {
   (_luaL_argerror($L,1,520725)|0);
  }
  (_lua_getmetatable($L,1)|0);
 }
 (_lua_setmetatable($L,2)|0);
 return 1;
}
function _luaB_cocreate($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_newthread($L)|0);
 $1 = (_lua_type($L,1)|0);
 $2 = ($1|0)==(6);
 if ($2) {
  $3 = (_lua_iscfunction($L,1)|0);
  $4 = ($3|0)==(0);
  if (!($4)) {
   label = 3;
  }
 } else {
  label = 3;
 }
 if ((label|0) == 3) {
  (_luaL_argerror($L,1,520777)|0);
 }
 _lua_pushvalue($L,1);
 _lua_xmove($L,$0,1);
 return 1;
}
function _luaB_coresume($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_tothread($L,1)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_luaL_argerror($L,1,520806)|0);
 }
 $2 = (_lua_gettop($L)|0);
 $3 = (($2) + -1)|0;
 $4 = (_auxresume($L,$0,$3)|0);
 $5 = ($4|0)<(0);
 if ($5) {
  _lua_pushboolean($L,0);
  _lua_insert($L,-2);
  $$0 = 2;
  return ($$0|0);
 } else {
  _lua_pushboolean($L,1);
  $6 = (($4) + 1)|0;
  $7 = $4 ^ -1;
  _lua_insert($L,$7);
  $$0 = $6;
  return ($$0|0);
 }
 return (0)|0;
}
function _auxresume($L,$co,$narg) {
 $L = $L|0;
 $co = $co|0;
 $narg = $narg|0;
 var $$$i = 0, $$0 = 0, $$0$i = 0, $$1$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $ar$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $ar$i = sp + 20|0;
 $0 = ($L|0)==($co|0);
 L1: do {
  if ($0) {
   $$1$i = 0;
  } else {
   $1 = (_lua_status($co)|0);
   switch ($1|0) {
   case 1:  {
    $$1$i = 1;
    break L1;
    break;
   }
   case 0:  {
    break;
   }
   default: {
    $$1$i = 3;
    break L1;
   }
   }
   $2 = (_lua_getstack($co,0,$ar$i)|0);
   $3 = ($2|0)>(0);
   if ($3) {
    $$0$i = 2;
   } else {
    $4 = (_lua_gettop($co)|0);
    $5 = ($4|0)==(0);
    $$$i = $5 ? 3 : 1;
    $$0$i = $$$i;
   }
   $$1$i = $$0$i;
  }
 } while(0);
 $6 = (_lua_checkstack($co,$narg)|0);
 $7 = ($6|0)==(0);
 if ($7) {
  (_luaL_error($L,520825,$vararg_buffer)|0);
 }
 $8 = ($$1$i|0)==(1);
 if (!($8)) {
  $9 = (39400 + ($$1$i<<2)|0);
  $10 = HEAP32[$9>>2]|0;
  HEAP32[$vararg_buffer1>>2] = $10;
  (_lua_pushfstring($L,520884,$vararg_buffer1)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 _lua_xmove($L,$co,$narg);
 _lua_setlevel($L,$co);
 $11 = (_lua_resume($co,$narg)|0);
 $12 = $11 | 1;
 $13 = ($12|0)==(1);
 if (!($13)) {
  _lua_xmove($co,$L,1);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $14 = (_lua_gettop($co)|0);
 $15 = (($14) + 1)|0;
 $16 = (_lua_checkstack($L,$15)|0);
 $17 = ($16|0)==(0);
 if ($17) {
  (_luaL_error($L,520911,$vararg_buffer3)|0);
 }
 _lua_xmove($co,$L,$14);
 $$0 = $14;
 STACKTOP = sp;return ($$0|0);
}
function _luaB_corunning($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_pushthread($L)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  return 1;
 }
 _lua_pushnil($L);
 return 1;
}
function _luaB_costatus($L) {
 $L = $L|0;
 var $$$i = 0, $$0$i = 0, $$1$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ar$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ar$i = sp;
 $0 = (_lua_tothread($L,1)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_luaL_argerror($L,1,520806)|0);
 }
 $2 = ($0|0)==($L|0);
 L4: do {
  if ($2) {
   $$1$i = 0;
  } else {
   $3 = (_lua_status($0)|0);
   switch ($3|0) {
   case 1:  {
    $$1$i = 1;
    break L4;
    break;
   }
   case 0:  {
    break;
   }
   default: {
    $$1$i = 3;
    break L4;
   }
   }
   $4 = (_lua_getstack($0,0,$ar$i)|0);
   $5 = ($4|0)>(0);
   if ($5) {
    $$0$i = 2;
   } else {
    $6 = (_lua_gettop($0)|0);
    $7 = ($6|0)==(0);
    $$$i = $7 ? 3 : 1;
    $$0$i = $$$i;
   }
   $$1$i = $$0$i;
  }
 } while(0);
 $8 = (39400 + ($$1$i<<2)|0);
 $9 = HEAP32[$8>>2]|0;
 _lua_pushstring($L,$9);
 STACKTOP = sp;return 1;
}
function _luaB_cowrap($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_newthread($L)|0);
 $1 = (_lua_type($L,1)|0);
 $2 = ($1|0)==(6);
 if ($2) {
  $3 = (_lua_iscfunction($L,1)|0);
  $4 = ($3|0)==(0);
  if (!($4)) {
   label = 3;
  }
 } else {
  label = 3;
 }
 if ((label|0) == 3) {
  (_luaL_argerror($L,1,520777)|0);
 }
 _lua_pushvalue($L,1);
 _lua_xmove($L,$0,1);
 _lua_pushcclosure($L,4313,1);
 return 1;
}
function _luaB_auxwrap($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_tothread($L,-10003)|0);
 $1 = (_lua_gettop($L)|0);
 $2 = (_auxresume($L,$0,$1)|0);
 $3 = ($2|0)<(0);
 if (!($3)) {
  return ($2|0);
 }
 $4 = (_lua_isstring($L,-1)|0);
 $5 = ($4|0)==(0);
 if (!($5)) {
  _luaL_where($L,1);
  _lua_insert($L,-2);
  _lua_concat($L,2);
 }
 (_lua_error($L)|0);
 return ($2|0);
}
function _luaB_yield($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_gettop($L)|0);
 $1 = (_lua_yield($L,$0)|0);
 return ($1|0);
}
function _luaopen_debug($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_register($L,520956,39416);
 return 1;
}
function _db_debug($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $buffer = sp;
 $0 = HEAP32[37544]|0;
 (_fwrite(520962,11,1,$0)|0);
 $1 = HEAP32[37573]|0;
 $2 = (_fgets($buffer,250,$1)|0);
 $3 = ($2|0)==(0|0);
 if ($3) {
  STACKTOP = sp;return 0;
 }
 while(1) {
  $4 = (_strcmp($buffer,520974)|0);
  $5 = ($4|0)==(0);
  if ($5) {
   label = 7;
   break;
  }
  $6 = (_strlen($buffer)|0);
  $7 = (_luaL_loadbuffer($L,$buffer,$6,520980)|0);
  $8 = ($7|0)==(0);
  if ($8) {
   $9 = (_lua_pcall($L,0,0,0)|0);
   $10 = ($9|0)==(0);
   if (!($10)) {
    label = 5;
   }
  } else {
   label = 5;
  }
  if ((label|0) == 5) {
   label = 0;
   $11 = (_lua_tolstring($L,-1,0)|0);
   (_fputs($11,$0)|0);
   (_fputc(10,$0)|0);
  }
  _lua_settop($L,0);
  (_fwrite(520962,11,1,$0)|0);
  $12 = (_fgets($buffer,250,$1)|0);
  $13 = ($12|0)==(0|0);
  if ($13) {
   label = 7;
   break;
  }
 }
 if ((label|0) == 7) {
  STACKTOP = sp;return 0;
 }
 return (0)|0;
}
function _db_getfenv($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checkany($L,1);
 _lua_getfenv($L,1);
 return 1;
}
function _db_gethook($L) {
 $L = $L|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $buff = 0, $i$0$i = 0, $i$1$i = 0, $i$2$i = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $buff = sp;
 $0 = (_lua_type($L,1)|0);
 $1 = ($0|0)==(8);
 if ($1) {
  $2 = (_lua_tothread($L,1)|0);
  $$0$i = $2;
 } else {
  $$0$i = $L;
 }
 $3 = (_lua_gethookmask($$0$i)|0);
 $4 = (_lua_gethook($$0$i)|0);
 $5 = ($4|0)!=(0|0);
 $6 = ($4|0)!=(4314|0);
 $or$cond = $5 & $6;
 if ($or$cond) {
  _lua_pushlstring($L,521038,13);
 } else {
  _gethooktable($L);
  _lua_pushlightuserdata($L,$$0$i);
  _lua_rawget($L,-2);
  _lua_remove($L,-2);
 }
 $7 = $3 & 1;
 $8 = ($7|0)==(0);
 if ($8) {
  $i$0$i = 0;
 } else {
  HEAP8[$buff>>0] = 99;
  $i$0$i = 1;
 }
 $9 = $3 & 2;
 $10 = ($9|0)==(0);
 if ($10) {
  $i$1$i = $i$0$i;
 } else {
  $11 = (($i$0$i) + 1)|0;
  $12 = (($buff) + ($i$0$i)|0);
  HEAP8[$12>>0] = 114;
  $i$1$i = $11;
 }
 $13 = $3 & 4;
 $14 = ($13|0)==(0);
 if ($14) {
  $i$2$i = $i$1$i;
 } else {
  $15 = (($i$1$i) + 1)|0;
  $16 = (($buff) + ($i$1$i)|0);
  HEAP8[$16>>0] = 108;
  $i$2$i = $15;
 }
 $17 = (($buff) + ($i$2$i)|0);
 HEAP8[$17>>0] = 0;
 _lua_pushstring($L,$buff);
 $18 = (_lua_gethookcount($$0$i)|0);
 _lua_pushinteger($L,$18);
 STACKTOP = sp;return 3;
}
function _hookf($L,$ar) {
 $L = $L|0;
 $ar = $ar|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _lua_pushlightuserdata($L,521013);
 _lua_rawget($L,-10000);
 _lua_pushlightuserdata($L,$L);
 _lua_rawget($L,-2);
 $0 = (_lua_type($L,-1)|0);
 $1 = ($0|0)==(6);
 if (!($1)) {
  return;
 }
 $2 = HEAP32[$ar>>2]|0;
 $3 = (39536 + ($2<<2)|0);
 $4 = HEAP32[$3>>2]|0;
 _lua_pushstring($L,$4);
 $5 = ((($ar)) + 20|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)>(-1);
 if ($7) {
  _lua_pushinteger($L,$6);
 } else {
  _lua_pushnil($L);
 }
 _lua_call($L,2,0);
 return;
}
function _gethooktable($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _lua_pushlightuserdata($L,521013);
 _lua_rawget($L,-10000);
 $0 = (_lua_type($L,-1)|0);
 $1 = ($0|0)==(5);
 if ($1) {
  return;
 }
 _lua_settop($L,-2);
 _lua_createtable($L,0,1);
 _lua_pushlightuserdata($L,521013);
 _lua_pushvalue($L,-2);
 _lua_rawset($L,-10000);
 return;
}
function _db_getinfo($L) {
 $L = $L|0;
 var $$0 = 0, $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ar = 0, $arg$0 = 0, $options$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $ar = sp + 4|0;
 $0 = (_lua_type($L,1)|0);
 $1 = ($0|0)==(8);
 if ($1) {
  $2 = (_lua_tothread($L,1)|0);
  $$0$i = $2;$arg$0 = 1;
 } else {
  $$0$i = $L;$arg$0 = 0;
 }
 $3 = $arg$0 | 2;
 $4 = (_luaL_optlstring($L,$3,521060,0)|0);
 $5 = (($arg$0) + 1)|0;
 $6 = (_lua_isnumber($L,$5)|0);
 $7 = ($6|0)==(0);
 do {
  if ($7) {
   $11 = (_lua_type($L,$5)|0);
   $12 = ($11|0)==(6);
   if ($12) {
    HEAP32[$vararg_buffer>>2] = $4;
    (_lua_pushfstring($L,521066,$vararg_buffer)|0);
    $13 = (_lua_tolstring($L,-1,0)|0);
    _lua_pushvalue($L,$5);
    _lua_xmove($L,$$0$i,1);
    $options$0 = $13;
    break;
   }
   $14 = (_luaL_argerror($L,$5,521070)|0);
   $$0 = $14;
   STACKTOP = sp;return ($$0|0);
  } else {
   $8 = (_lua_tointeger($L,$5)|0);
   $9 = (_lua_getstack($$0$i,$8,$ar)|0);
   $10 = ($9|0)==(0);
   if ($10) {
    _lua_pushnil($L);
    $$0 = 1;
    STACKTOP = sp;return ($$0|0);
   } else {
    $options$0 = $4;
   }
  }
 } while(0);
 $15 = (_lua_getinfo($$0$i,$options$0,$ar)|0);
 $16 = ($15|0)==(0);
 if ($16) {
  $17 = (_luaL_argerror($L,$3,521585)|0);
  $$0 = $17;
  STACKTOP = sp;return ($$0|0);
 }
 _lua_createtable($L,0,2);
 $18 = (_strchr($options$0,83)|0);
 $19 = ($18|0)==(0|0);
 if (!($19)) {
  $20 = ((($ar)) + 16|0);
  $21 = HEAP32[$20>>2]|0;
  _lua_pushstring($L,$21);
  _lua_setfield($L,-2,524524);
  $22 = ((($ar)) + 36|0);
  _lua_pushstring($L,$22);
  _lua_setfield($L,-2,521097);
  $23 = ((($ar)) + 28|0);
  $24 = HEAP32[$23>>2]|0;
  _lua_pushinteger($L,$24);
  _lua_setfield($L,-2,521107);
  $25 = ((($ar)) + 32|0);
  $26 = HEAP32[$25>>2]|0;
  _lua_pushinteger($L,$26);
  _lua_setfield($L,-2,521119);
  $27 = ((($ar)) + 12|0);
  $28 = HEAP32[$27>>2]|0;
  _lua_pushstring($L,$28);
  _lua_setfield($L,-2,521135);
 }
 $29 = (_strchr($options$0,108)|0);
 $30 = ($29|0)==(0|0);
 if (!($30)) {
  $31 = ((($ar)) + 20|0);
  $32 = HEAP32[$31>>2]|0;
  _lua_pushinteger($L,$32);
  _lua_setfield($L,-2,521140);
 }
 $33 = (_strchr($options$0,117)|0);
 $34 = ($33|0)==(0|0);
 if (!($34)) {
  $35 = ((($ar)) + 24|0);
  $36 = HEAP32[$35>>2]|0;
  _lua_pushinteger($L,$36);
  _lua_setfield($L,-2,521152);
 }
 $37 = (_strchr($options$0,110)|0);
 $38 = ($37|0)==(0|0);
 if (!($38)) {
  $39 = ((($ar)) + 4|0);
  $40 = HEAP32[$39>>2]|0;
  _lua_pushstring($L,$40);
  _lua_setfield($L,-2,610375);
  $41 = ((($ar)) + 8|0);
  $42 = HEAP32[$41>>2]|0;
  _lua_pushstring($L,$42);
  _lua_setfield($L,-2,521157);
 }
 $43 = (_strchr($options$0,76)|0);
 $44 = ($43|0)==(0|0);
 if (!($44)) {
  $45 = ($$0$i|0)==($L|0);
  if ($45) {
   _lua_pushvalue($L,-2);
   _lua_remove($L,-3);
  } else {
   _lua_xmove($$0$i,$L,1);
  }
  _lua_setfield($L,-2,521166);
 }
 $46 = (_strchr($options$0,102)|0);
 $47 = ($46|0)==(0|0);
 if ($47) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 $48 = ($$0$i|0)==($L|0);
 if ($48) {
  _lua_pushvalue($L,-2);
  _lua_remove($L,-3);
 } else {
  _lua_xmove($$0$i,$L,1);
 }
 _lua_setfield($L,-2,521178);
 $$0 = 1;
 STACKTOP = sp;return ($$0|0);
}
function _db_getlocal($L) {
 $L = $L|0;
 var $$0 = 0, $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ar = 0, $arg$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ar = sp;
 $0 = (_lua_type($L,1)|0);
 $1 = ($0|0)==(8);
 if ($1) {
  $2 = (_lua_tothread($L,1)|0);
  $$0$i = $2;$arg$0 = 1;
 } else {
  $$0$i = $L;$arg$0 = 0;
 }
 $3 = (($arg$0) + 1)|0;
 $4 = (_luaL_checkinteger($L,$3)|0);
 $5 = (_lua_getstack($$0$i,$4,$ar)|0);
 $6 = ($5|0)==(0);
 if ($6) {
  $7 = (_luaL_argerror($L,$3,521192)|0);
  $$0 = $7;
  STACKTOP = sp;return ($$0|0);
 }
 $8 = $arg$0 | 2;
 $9 = (_luaL_checkinteger($L,$8)|0);
 $10 = (_lua_getlocal($$0$i,$ar,$9)|0);
 $11 = ($10|0)==(0|0);
 if ($11) {
  _lua_pushnil($L);
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 } else {
  _lua_xmove($$0$i,$L,1);
  _lua_pushstring($L,$10);
  _lua_pushvalue($L,-2);
  $$0 = 2;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _db_getregistry($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _lua_pushvalue($L,-10000);
 return 1;
}
function _db_getmetatable($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checkany($L,1);
 $0 = (_lua_getmetatable($L,1)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  return 1;
 }
 _lua_pushnil($L);
 return 1;
}
function _db_getupvalue($L) {
 $L = $L|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkinteger($L,2)|0);
 _luaL_checktype($L,1,6);
 $1 = (_lua_iscfunction($L,1)|0);
 $2 = ($1|0)==(0);
 if (!($2)) {
  $$0$i = 0;
  return ($$0$i|0);
 }
 $3 = (_lua_getupvalue($L,1,$0)|0);
 $4 = ($3|0)==(0|0);
 if ($4) {
  $$0$i = 0;
  return ($$0$i|0);
 }
 _lua_pushstring($L,$3);
 _lua_insert($L,-2);
 $$0$i = 2;
 return ($$0$i|0);
}
function _db_setfenv($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _luaL_checktype($L,2,5);
 _lua_settop($L,2);
 $0 = (_lua_setfenv($L,1)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  STACKTOP = sp;return 1;
 }
 (_luaL_error($L,521255,$vararg_buffer)|0);
 STACKTOP = sp;return 1;
}
function _db_sethook($L) {
 $L = $L|0;
 var $$$i = 0, $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $arg$0 = 0, $count$0 = 0, $func$0 = 0, $mask$0 = 0, $mask$1$$i = 0, $mask$1$i = 0, $mask$3$i = 0, $not$$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_type($L,1)|0);
 $1 = ($0|0)==(8);
 if ($1) {
  $2 = (_lua_tothread($L,1)|0);
  $$0$i = $2;$arg$0 = 1;
 } else {
  $$0$i = $L;$arg$0 = 0;
 }
 $3 = (($arg$0) + 1)|0;
 $4 = (_lua_type($L,$3)|0);
 $5 = ($4|0)<(1);
 if ($5) {
  _lua_settop($L,$3);
  $count$0 = 0;$func$0 = 0;$mask$0 = 0;
 } else {
  $6 = $arg$0 | 2;
  $7 = (_luaL_checklstring($L,$6,0)|0);
  _luaL_checktype($L,$3,6);
  $8 = (($arg$0) + 3)|0;
  $9 = (_luaL_optinteger($L,$8,0)|0);
  $10 = (_strchr($7,99)|0);
  $not$$i = ($10|0)!=(0|0);
  $$$i = $not$$i&1;
  $11 = (_strchr($7,114)|0);
  $12 = ($11|0)==(0|0);
  $13 = $$$i | 2;
  $mask$1$i = $12 ? $$$i : $13;
  $14 = (_strchr($7,108)|0);
  $15 = ($14|0)==(0|0);
  $16 = $mask$1$i | 4;
  $mask$1$$i = $15 ? $mask$1$i : $16;
  $17 = ($9|0)>(0);
  $18 = $mask$1$$i | 8;
  $mask$3$i = $17 ? $18 : $mask$1$$i;
  $count$0 = $9;$func$0 = 4314;$mask$0 = $mask$3$i;
 }
 _gethooktable($L);
 _lua_pushlightuserdata($L,$$0$i);
 _lua_pushvalue($L,$3);
 _lua_rawset($L,-3);
 _lua_settop($L,-2);
 (_lua_sethook($$0$i,$func$0,$mask$0,$count$0)|0);
 return 0;
}
function _db_setlocal($L) {
 $L = $L|0;
 var $$0 = 0, $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ar = 0, $arg$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ar = sp;
 $0 = (_lua_type($L,1)|0);
 $1 = ($0|0)==(8);
 if ($1) {
  $2 = (_lua_tothread($L,1)|0);
  $$0$i = $2;$arg$0 = 1;
 } else {
  $$0$i = $L;$arg$0 = 0;
 }
 $3 = (($arg$0) + 1)|0;
 $4 = (_luaL_checkinteger($L,$3)|0);
 $5 = (_lua_getstack($$0$i,$4,$ar)|0);
 $6 = ($5|0)==(0);
 if ($6) {
  $7 = (_luaL_argerror($L,$3,521192)|0);
  $$0 = $7;
  STACKTOP = sp;return ($$0|0);
 } else {
  $8 = (($arg$0) + 3)|0;
  _luaL_checkany($L,$8);
  _lua_settop($L,$8);
  _lua_xmove($L,$$0$i,1);
  $9 = $arg$0 | 2;
  $10 = (_luaL_checkinteger($L,$9)|0);
  $11 = (_lua_setlocal($$0$i,$ar,$10)|0);
  _lua_pushstring($L,$11);
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _db_setmetatable($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_type($L,2)|0);
 switch ($0|0) {
 case 0: case 5:  {
  break;
 }
 default: {
  (_luaL_argerror($L,2,521337)|0);
 }
 }
 _lua_settop($L,2);
 $1 = (_lua_setmetatable($L,1)|0);
 _lua_pushboolean($L,$1);
 return 1;
}
function _db_setupvalue($L) {
 $L = $L|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checkany($L,3);
 $0 = (_luaL_checkinteger($L,2)|0);
 _luaL_checktype($L,1,6);
 $1 = (_lua_iscfunction($L,1)|0);
 $2 = ($1|0)==(0);
 if (!($2)) {
  $$0$i = 0;
  return ($$0$i|0);
 }
 $3 = (_lua_setupvalue($L,1,$0)|0);
 $4 = ($3|0)==(0|0);
 if ($4) {
  $$0$i = 0;
  return ($$0$i|0);
 }
 _lua_pushstring($L,$3);
 _lua_insert($L,-1);
 $$0$i = 1;
 return ($$0$i|0);
}
function _db_errorfb($L) {
 $L = $L|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ar = 0, $arg$0 = 0, $firstpart$0$lcssa2 = 0, $firstpart$0$ph = 0, $firstpart$0$us = 0, $level$0 = 0, $level$1$ph = 0, $level$2 = 0, $level$2$in = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, $vararg_buffer7 = 0;
 var $vararg_buffer9 = 0, $vararg_ptr12 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer9 = sp + 32|0;
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $ar = sp + 40|0;
 $0 = (_lua_type($L,1)|0);
 $1 = ($0|0)==(8);
 if ($1) {
  $2 = (_lua_tothread($L,1)|0);
  $$0$i = $2;$arg$0 = 1;
 } else {
  $$0$i = $L;$arg$0 = 0;
 }
 $3 = $arg$0 | 2;
 $4 = (_lua_isnumber($L,$3)|0);
 $5 = ($4|0)==(0);
 if ($5) {
  $7 = ($$0$i|0)==($L|0);
  $8 = $7&1;
  $level$0 = $8;
 } else {
  $6 = (_lua_tointeger($L,$3)|0);
  _lua_settop($L,-2);
  $level$0 = $6;
 }
 $9 = (_lua_gettop($L)|0);
 $10 = ($9|0)==($arg$0|0);
 do {
  if ($10) {
   _lua_pushlstring($L,1282466,0);
  } else {
   $11 = (($arg$0) + 1)|0;
   $12 = (_lua_isstring($L,$11)|0);
   $13 = ($12|0)==(0);
   if ($13) {
    STACKTOP = sp;return 1;
   } else {
    _lua_pushlstring($L,521380,1);
    break;
   }
  }
 } while(0);
 _lua_pushlstring($L,521382,16);
 $14 = ((($ar)) + 36|0);
 $15 = ((($ar)) + 20|0);
 $16 = ((($ar)) + 8|0);
 $17 = ((($ar)) + 12|0);
 $18 = ((($ar)) + 28|0);
 $19 = ((($ar)) + 4|0);
 $firstpart$0$ph = 1;$level$1$ph = $level$0;
 L15: while(1) {
  $20 = ($level$1$ph|0)>(11);
  $21 = (($level$1$ph) + 11)|0;
  L17: do {
   if ($20) {
    $firstpart$0$us = $firstpart$0$ph;
    while(1) {
     $22 = (_lua_getstack($$0$i,$level$1$ph,$ar)|0);
     $23 = ($22|0)==(0);
     if ($23) {
      break L15;
     }
     $24 = ($firstpart$0$us|0)==(0);
     if ($24) {
      $firstpart$0$lcssa2 = 0;
      break L17;
     }
     $25 = (_lua_getstack($$0$i,$21,$ar)|0);
     $26 = ($25|0)==(0);
     if ($26) {
      $firstpart$0$us = 0;
     } else {
      break;
     }
    }
    _lua_pushlstring($L,521399,5);
    $level$2$in = $level$1$ph;
    while(1) {
     $level$2 = (($level$2$in) + 1)|0;
     $29 = (($level$2$in) + 11)|0;
     $30 = (_lua_getstack($$0$i,$29,$ar)|0);
     $31 = ($30|0)==(0);
     if ($31) {
      $firstpart$0$ph = 0;$level$1$ph = $level$2;
      continue L15;
     } else {
      $level$2$in = $level$2;
     }
    }
   } else {
    $27 = (_lua_getstack($$0$i,$level$1$ph,$ar)|0);
    $28 = ($27|0)==(0);
    if ($28) {
     break L15;
    } else {
     $firstpart$0$lcssa2 = $firstpart$0$ph;
    }
   }
  } while(0);
  $32 = (($level$1$ph) + 1)|0;
  _lua_pushlstring($L,521405,2);
  (_lua_getinfo($$0$i,521408,$ar)|0);
  HEAP32[$vararg_buffer>>2] = $14;
  (_lua_pushfstring($L,521412,$vararg_buffer)|0);
  $33 = HEAP32[$15>>2]|0;
  $34 = ($33|0)>(0);
  if ($34) {
   HEAP32[$vararg_buffer1>>2] = $33;
   (_lua_pushfstring($L,521416,$vararg_buffer1)|0);
  }
  $35 = HEAP32[$16>>2]|0;
  $36 = HEAP8[$35>>0]|0;
  $37 = ($36<<24>>24)==(0);
  L30: do {
   if ($37) {
    $39 = HEAP32[$17>>2]|0;
    $40 = HEAP8[$39>>0]|0;
    switch ($40<<24>>24) {
    case 109:  {
     (_lua_pushfstring($L,521438,$vararg_buffer7)|0);
     break L30;
     break;
    }
    case 116: case 67:  {
     _lua_pushlstring($L,521453,2);
     break L30;
     break;
    }
    default: {
     $41 = HEAP32[$18>>2]|0;
     HEAP32[$vararg_buffer9>>2] = $14;
     $vararg_ptr12 = ((($vararg_buffer9)) + 4|0);
     HEAP32[$vararg_ptr12>>2] = $41;
     (_lua_pushfstring($L,521456,$vararg_buffer9)|0);
     break L30;
    }
    }
   } else {
    $38 = HEAP32[$19>>2]|0;
    HEAP32[$vararg_buffer4>>2] = $38;
    (_lua_pushfstring($L,521420,$vararg_buffer4)|0);
   }
  } while(0);
  $42 = (_lua_gettop($L)|0);
  $43 = (($42) - ($arg$0))|0;
  _lua_concat($L,$43);
  $firstpart$0$ph = $firstpart$0$lcssa2;$level$1$ph = $32;
 }
 $44 = (_lua_gettop($L)|0);
 $45 = (($44) - ($arg$0))|0;
 _lua_concat($L,$45);
 STACKTOP = sp;return 1;
}
function _luaopen_io($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 (_luaL_newmetatable($L,521477)|0);
 _lua_pushvalue($L,-1);
 _lua_setfield($L,-2,523215);
 _luaL_register($L,0,39556);
 _lua_createtable($L,0,1);
 _lua_pushcclosure($L,4315,0);
 _lua_setfield($L,-2,521518);
 _lua_replace($L,-10001);
 _luaL_register($L,521701,39692);
 _lua_createtable($L,0,1);
 _lua_pushcclosure($L,4316,0);
 _lua_setfield($L,-2,521518);
 $0 = HEAP32[37573]|0;
 $1 = (_lua_newuserdata($L,4)|0);
 HEAP32[$1>>2] = 0;
 _lua_getfield($L,-10000,521477);
 (_lua_setmetatable($L,-2)|0);
 HEAP32[$1>>2] = $0;
 _lua_pushvalue($L,-1);
 _lua_rawseti($L,-10001,1);
 _lua_pushvalue($L,-2);
 (_lua_setfenv($L,-2)|0);
 _lua_setfield($L,-3,521834);
 $2 = HEAP32[37602]|0;
 $3 = (_lua_newuserdata($L,4)|0);
 HEAP32[$3>>2] = 0;
 _lua_getfield($L,-10000,521477);
 (_lua_setmetatable($L,-2)|0);
 HEAP32[$3>>2] = $2;
 _lua_pushvalue($L,-1);
 _lua_rawseti($L,-10001,2);
 _lua_pushvalue($L,-2);
 (_lua_setfenv($L,-2)|0);
 _lua_setfield($L,-3,521840);
 $4 = HEAP32[37544]|0;
 $5 = (_lua_newuserdata($L,4)|0);
 HEAP32[$5>>2] = 0;
 _lua_getfield($L,-10000,521477);
 (_lua_setmetatable($L,-2)|0);
 HEAP32[$5>>2] = $4;
 _lua_pushvalue($L,-2);
 (_lua_setfenv($L,-2)|0);
 _lua_setfield($L,-3,521847);
 _lua_settop($L,-2);
 _lua_getfield($L,-1,521749);
 _lua_createtable($L,0,1);
 _lua_pushcclosure($L,4317,0);
 _lua_setfield($L,-2,521518);
 (_lua_setfenv($L,-2)|0);
 _lua_settop($L,-2);
 return 1;
}
function _io_close($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_lua_type($L,1)|0);
 $1 = ($0|0)==(-1);
 if ($1) {
  _lua_rawgeti($L,-10001,2);
 }
 $2 = (_luaL_checkudata($L,1,521477)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  (_luaL_error($L,521489,$vararg_buffer)|0);
 }
 _lua_getfenv($L,1);
 _lua_getfield($L,-1,521518);
 $5 = (_lua_tocfunction($L,-1)|0);
 $6 = (FUNCTION_TABLE_ii[$5 & 8191]($L)|0);
 STACKTOP = sp;return ($6|0);
}
function _f_flush($L) {
 $L = $L|0;
 var $$0$i = 0, $$pre$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_luaL_checkudata($L,1,521477)|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  (_luaL_error($L,521489,$vararg_buffer)|0);
  $$pre$i = HEAP32[$0>>2]|0;
  $3 = $$pre$i;
 } else {
  $3 = $1;
 }
 $4 = (_fflush($3)|0);
 $5 = ($4|0)==(0);
 $6 = (___errno_location()|0);
 $7 = HEAP32[$6>>2]|0;
 if ($5) {
  _lua_pushboolean($L,1);
  $$0$i = 1;
  STACKTOP = sp;return ($$0$i|0);
 } else {
  _lua_pushnil($L);
  $8 = (_strerror($7)|0);
  HEAP32[$vararg_buffer1>>2] = $8;
  (_lua_pushfstring($L,595867,$vararg_buffer1)|0);
  _lua_pushinteger($L,$7);
  $$0$i = 3;
  STACKTOP = sp;return ($$0$i|0);
 }
 return (0)|0;
}
function _f_lines($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_luaL_checkudata($L,1,521477)|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  (_luaL_error($L,521489,$vararg_buffer)|0);
 }
 _lua_pushvalue($L,1);
 _lua_pushboolean($L,0);
 _lua_pushcclosure($L,4318,2);
 STACKTOP = sp;return 1;
}
function _io_readline($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_lua_touserdata($L,-10003)|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  (_luaL_error($L,521538,$vararg_buffer)|0);
 }
 $3 = (_read_line($L,$1)|0);
 $4 = (_ferror($1)|0);
 $5 = ($4|0)==(0);
 if (!($5)) {
  $6 = (___errno_location()|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (_strerror($7)|0);
  HEAP32[$vararg_buffer1>>2] = $8;
  $9 = (_luaL_error($L,595867,$vararg_buffer1)|0);
  $$0 = $9;
  STACKTOP = sp;return ($$0|0);
 }
 $10 = ($3|0)==(0);
 if (!($10)) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 $11 = (_lua_toboolean($L,-10004)|0);
 $12 = ($11|0)==(0);
 if ($12) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 _lua_settop($L,0);
 _lua_pushvalue($L,-10003);
 _lua_getfenv($L,1);
 _lua_getfield($L,-1,521518);
 $13 = (_lua_tocfunction($L,-1)|0);
 (FUNCTION_TABLE_ii[$13 & 8191]($L)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _read_line($L,$f) {
 $L = $L|0;
 $f = $f|0;
 var $$1$ph = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $b = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1040|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $b = sp;
 _luaL_buffinit($L,$b);
 $0 = (_luaL_prepbuffer($b)|0);
 $1 = (_fgets($0,1024,$f)|0);
 $2 = ($1|0)==(0|0);
 L1: do {
  if (!($2)) {
   $6 = $0;
   while(1) {
    $7 = (_strlen($6)|0);
    $8 = ($7|0)==(0);
    if (!($8)) {
     $9 = (($7) + -1)|0;
     $10 = (($6) + ($9)|0);
     $11 = HEAP8[$10>>0]|0;
     $12 = ($11<<24>>24)==(10);
     if ($12) {
      $$lcssa = $9;
      break;
     }
    }
    $15 = HEAP32[$b>>2]|0;
    $16 = (($15) + ($7)|0);
    HEAP32[$b>>2] = $16;
    $17 = (_luaL_prepbuffer($b)|0);
    $18 = (_fgets($17,1024,$f)|0);
    $19 = ($18|0)==(0|0);
    if ($19) {
     break L1;
    } else {
     $6 = $17;
    }
   }
   $13 = HEAP32[$b>>2]|0;
   $14 = (($13) + ($$lcssa)|0);
   HEAP32[$b>>2] = $14;
   _luaL_pushresult($b);
   $$1$ph = 1;
   STACKTOP = sp;return ($$1$ph|0);
  }
 } while(0);
 _luaL_pushresult($b);
 $3 = (_lua_objlen($L,-1)|0);
 $4 = ($3|0)!=(0);
 $5 = $4&1;
 $$1$ph = $5;
 STACKTOP = sp;return ($$1$ph|0);
}
function _f_read($L) {
 $L = $L|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_luaL_checkudata($L,1,521477)|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  (_luaL_error($L,521489,$vararg_buffer)|0);
  $$pre$i = HEAP32[$0>>2]|0;
  $3 = $$pre$i;
 } else {
  $3 = $1;
 }
 $4 = (_g_read($L,$3,2)|0);
 STACKTOP = sp;return ($4|0);
}
function _g_read($L,$f,$first) {
 $L = $L|0;
 $f = $f|0;
 $first = $first|0;
 var $$0$i = 0, $$3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $d$i = 0, $n$05 = 0, $n$05$lcssa = 0, $n$1 = 0, $success$3 = 0, $success$4 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 16|0;
 $vararg_buffer = sp + 8|0;
 $d$i = sp;
 $0 = (_lua_gettop($L)|0);
 _clearerr($f);
 $1 = ($0|0)==(1);
 L1: do {
  if ($1) {
   $2 = (_read_line($L,$f)|0);
   $3 = (($first) + 1)|0;
   $n$1 = $3;$success$4 = $2;
  } else {
   $4 = (($0) + 19)|0;
   _luaL_checkstack($L,$4,521566);
   $5 = (($0) + -2)|0;
   $28 = $5;$n$05 = $first;
   L4: while(1) {
    $6 = (_lua_type($L,$n$05)|0);
    $7 = ($6|0)==(3);
    L6: do {
     if ($7) {
      $8 = (_lua_tointeger($L,$n$05)|0);
      $9 = ($8|0)==(0);
      if ($9) {
       $10 = (_getc($f)|0);
       (_ungetc($10,$f)|0);
       _lua_pushlstring($L,0,0);
       $11 = ($10|0)!=(-1);
       $12 = $11&1;
       $success$3 = $12;
       break;
      } else {
       $13 = (_read_chars($L,$f,$8)|0);
       $success$3 = $13;
       break;
      }
     } else {
      $14 = (_lua_tolstring($L,$n$05,0)|0);
      $15 = ($14|0)==(0|0);
      if ($15) {
       label = 10;
      } else {
       $16 = HEAP8[$14>>0]|0;
       $17 = ($16<<24>>24)==(42);
       if (!($17)) {
        label = 10;
       }
      }
      if ((label|0) == 10) {
       label = 0;
       (_luaL_argerror($L,$n$05,521585)|0);
      }
      $18 = ((($14)) + 1|0);
      $19 = HEAP8[$18>>0]|0;
      $20 = $19 << 24 >> 24;
      switch ($20|0) {
      case 110:  {
       HEAP32[$vararg_buffer>>2] = $d$i;
       $21 = (_fscanf($f,521600,$vararg_buffer)|0);
       $22 = ($21|0)==(1);
       if ($22) {
        $23 = +HEAPF64[$d$i>>3];
        _lua_pushnumber($L,$23);
        $$0$i = 1;
       } else {
        _lua_pushnil($L);
        $$0$i = 0;
       }
       $success$3 = $$0$i;
       break L6;
       break;
      }
      case 108:  {
       $24 = (_read_line($L,$f)|0);
       $success$3 = $24;
       break L6;
       break;
      }
      case 97:  {
       (_read_chars($L,$f,-1)|0);
       $success$3 = 1;
       break L6;
       break;
      }
      default: {
       $n$05$lcssa = $n$05;
       break L4;
      }
      }
     }
    } while(0);
    $26 = (($n$05) + 1)|0;
    $27 = (($28) + -1)|0;
    $29 = ($28|0)!=(0);
    $30 = ($success$3|0)!=(0);
    $31 = $30 & $29;
    if ($31) {
     $28 = $27;$n$05 = $26;
    } else {
     $n$1 = $26;$success$4 = $success$3;
     break L1;
    }
   }
   $25 = (_luaL_argerror($L,$n$05$lcssa,521604)|0);
   $$3 = $25;
   STACKTOP = sp;return ($$3|0);
  }
 } while(0);
 $32 = (_ferror($f)|0);
 $33 = ($32|0)==(0);
 if (!($33)) {
  $34 = (___errno_location()|0);
  $35 = HEAP32[$34>>2]|0;
  _lua_pushnil($L);
  $36 = (_strerror($35)|0);
  HEAP32[$vararg_buffer1>>2] = $36;
  (_lua_pushfstring($L,595867,$vararg_buffer1)|0);
  _lua_pushinteger($L,$35);
  $$3 = 3;
  STACKTOP = sp;return ($$3|0);
 }
 $37 = ($success$4|0)==(0);
 if ($37) {
  _lua_settop($L,-2);
  _lua_pushnil($L);
 }
 $38 = (($n$1) - ($first))|0;
 $$3 = $38;
 STACKTOP = sp;return ($$3|0);
}
function _read_chars($L,$f,$n) {
 $L = $L|0;
 $f = $f|0;
 $n = $n|0;
 var $$0 = 0, $$0$rlen$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b = 0, $rlen$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1040|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $b = sp;
 _luaL_buffinit($L,$b);
 $$0 = $n;$rlen$0 = 1024;
 while(1) {
  $0 = (_luaL_prepbuffer($b)|0);
  $1 = ($rlen$0>>>0)>($$0>>>0);
  $$0$rlen$0 = $1 ? $$0 : $rlen$0;
  $2 = (_fread($0,1,$$0$rlen$0,$f)|0);
  $3 = HEAP32[$b>>2]|0;
  $4 = (($3) + ($2)|0);
  HEAP32[$b>>2] = $4;
  $5 = ($$0|0)==($2|0);
  if ($5) {
   label = 3;
   break;
  }
  $6 = (($$0) - ($2))|0;
  $7 = ($2|0)==($$0$rlen$0|0);
  if ($7) {
   $$0 = $6;$rlen$0 = $$0$rlen$0;
  } else {
   label = 5;
   break;
  }
 }
 if ((label|0) == 3) {
  _luaL_pushresult($b);
  $11 = 1;
  $10 = $11&1;
  STACKTOP = sp;return ($10|0);
 }
 else if ((label|0) == 5) {
  _luaL_pushresult($b);
  $8 = (_lua_objlen($L,-1)|0);
  $9 = ($8|0)!=(0);
  $11 = $9;
  $10 = $11&1;
  STACKTOP = sp;return ($10|0);
 }
 return (0)|0;
}
function _f_seek($L) {
 $L = $L|0;
 var $$0 = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_luaL_checkudata($L,1,521477)|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  (_luaL_error($L,521489,$vararg_buffer)|0);
  $$pre$i = HEAP32[$0>>2]|0;
  $7 = $$pre$i;
 } else {
  $7 = $1;
 }
 $3 = (_luaL_checkoption($L,2,521624,39636)|0);
 $4 = (_luaL_optinteger($L,3,0)|0);
 $5 = (39652 + ($3<<2)|0);
 $6 = HEAP32[$5>>2]|0;
 $8 = (_fseek($7,$4,$6)|0);
 $9 = ($8|0)==(0);
 if ($9) {
  $13 = (_ftell($7)|0);
  _lua_pushinteger($L,$13);
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 } else {
  $10 = (___errno_location()|0);
  $11 = HEAP32[$10>>2]|0;
  _lua_pushnil($L);
  $12 = (_strerror($11)|0);
  HEAP32[$vararg_buffer1>>2] = $12;
  (_lua_pushfstring($L,595867,$vararg_buffer1)|0);
  _lua_pushinteger($L,$11);
  $$0 = 3;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _f_setvbuf($L) {
 $L = $L|0;
 var $$0$i = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_luaL_checkudata($L,1,521477)|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  (_luaL_error($L,521489,$vararg_buffer)|0);
  $$pre$i = HEAP32[$0>>2]|0;
  $7 = $$pre$i;
 } else {
  $7 = $1;
 }
 $3 = (_luaL_checkoption($L,2,0,39664)|0);
 $4 = (_luaL_optinteger($L,3,1024)|0);
 $5 = (39680 + ($3<<2)|0);
 $6 = HEAP32[$5>>2]|0;
 $8 = (_setvbuf($7,0,$6,$4)|0);
 $9 = ($8|0)==(0);
 $10 = (___errno_location()|0);
 $11 = HEAP32[$10>>2]|0;
 if ($9) {
  _lua_pushboolean($L,1);
  $$0$i = 1;
  STACKTOP = sp;return ($$0$i|0);
 } else {
  _lua_pushnil($L);
  $12 = (_strerror($11)|0);
  HEAP32[$vararg_buffer1>>2] = $12;
  (_lua_pushfstring($L,595867,$vararg_buffer1)|0);
  _lua_pushinteger($L,$11);
  $$0$i = 3;
  STACKTOP = sp;return ($$0$i|0);
 }
 return (0)|0;
}
function _f_write($L) {
 $L = $L|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_luaL_checkudata($L,1,521477)|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  (_luaL_error($L,521489,$vararg_buffer)|0);
  $$pre$i = HEAP32[$0>>2]|0;
  $3 = $$pre$i;
 } else {
  $3 = $1;
 }
 $4 = (_g_write($L,$3,2)|0);
 STACKTOP = sp;return ($4|0);
}
function _g_write($L,$f,$arg) {
 $L = $L|0;
 $f = $f|0;
 $arg = $arg|0;
 var $$0$i = 0, $$03 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0;
 var $4 = 0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $l = 0, $nargs$0 = 0, $nargs$01 = 0, $nargs$04 = 0, $status$02 = 0, $status$1 = 0, $status$1$lcssa = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $l = sp + 12|0;
 $0 = (_lua_gettop($L)|0);
 $nargs$01 = (($0) + -1)|0;
 $1 = ($nargs$01|0)==(0);
 if (!($1)) {
  $$03 = $arg;$nargs$04 = $nargs$01;$status$02 = 1;
  while(1) {
   $2 = (_lua_type($L,$$03)|0);
   $3 = ($2|0)==(3);
   if ($3) {
    $4 = ($status$02|0)==(0);
    if ($4) {
     $9 = 0;
    } else {
     $5 = (+_lua_tonumber($L,$$03));
     HEAPF64[$vararg_buffer>>3] = $5;
     $6 = (_fprintf($f,521660,$vararg_buffer)|0);
     $7 = ($6|0)>(0);
     $9 = $7;
    }
    $8 = $9&1;
    $status$1 = $8;
   } else {
    $10 = (_luaL_checklstring($L,$$03,$l)|0);
    $11 = ($status$02|0)==(0);
    if ($11) {
     $17 = 0;
    } else {
     $12 = HEAP32[$l>>2]|0;
     $13 = (_fwrite($10,1,$12,$f)|0);
     $14 = HEAP32[$l>>2]|0;
     $15 = ($13|0)==($14|0);
     $17 = $15;
    }
    $16 = $17&1;
    $status$1 = $16;
   }
   $18 = (($$03) + 1)|0;
   $nargs$0 = (($nargs$04) + -1)|0;
   $19 = ($nargs$0|0)==(0);
   if ($19) {
    $status$1$lcssa = $status$1;
    break;
   } else {
    $$03 = $18;$nargs$04 = $nargs$0;$status$02 = $status$1;
   }
  }
  $20 = (___errno_location()|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = ($status$1$lcssa|0)==(0);
  if ($22) {
   _lua_pushnil($L);
   $23 = (_strerror($21)|0);
   HEAP32[$vararg_buffer1>>2] = $23;
   (_lua_pushfstring($L,595867,$vararg_buffer1)|0);
   _lua_pushinteger($L,$21);
   $$0$i = 3;
   STACKTOP = sp;return ($$0$i|0);
  }
 }
 _lua_pushboolean($L,1);
 $$0$i = 1;
 STACKTOP = sp;return ($$0$i|0);
}
function _io_gc($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkudata($L,1,521477)|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return 0;
 }
 _lua_getfenv($L,1);
 _lua_getfield($L,-1,521518);
 $3 = (_lua_tocfunction($L,-1)|0);
 (FUNCTION_TABLE_ii[$3 & 8191]($L)|0);
 return 0;
}
function _io_tostring($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_luaL_checkudata($L,1,521477)|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  _lua_pushlstring($L,521677,13);
  STACKTOP = sp;return 1;
 } else {
  HEAP32[$vararg_buffer>>2] = $1;
  (_lua_pushfstring($L,521691,$vararg_buffer)|0);
  STACKTOP = sp;return 1;
 }
 return (0)|0;
}
function _io_fclose($L) {
 $L = $L|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_luaL_checkudata($L,1,521477)|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_fclose($1)|0);
 $3 = ($2|0)==(0);
 HEAP32[$0>>2] = 0;
 $4 = (___errno_location()|0);
 $5 = HEAP32[$4>>2]|0;
 if ($3) {
  _lua_pushboolean($L,1);
  $$0$i = 1;
  STACKTOP = sp;return ($$0$i|0);
 } else {
  _lua_pushnil($L);
  $6 = (_strerror($5)|0);
  HEAP32[$vararg_buffer>>2] = $6;
  (_lua_pushfstring($L,595867,$vararg_buffer)|0);
  _lua_pushinteger($L,$5);
  $$0$i = 3;
  STACKTOP = sp;return ($$0$i|0);
 }
 return (0)|0;
}
function _io_flush($L) {
 $L = $L|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 _lua_rawgeti($L,-10001,2);
 $0 = (_lua_touserdata($L,-1)|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  HEAP32[$vararg_buffer>>2] = 521731;
  (_luaL_error($L,521704,$vararg_buffer)|0);
 }
 $3 = (_fflush($1)|0);
 $4 = ($3|0)==(0);
 $5 = (___errno_location()|0);
 $6 = HEAP32[$5>>2]|0;
 if ($4) {
  _lua_pushboolean($L,1);
  $$0$i = 1;
  STACKTOP = sp;return ($$0$i|0);
 } else {
  _lua_pushnil($L);
  $7 = (_strerror($6)|0);
  HEAP32[$vararg_buffer1>>2] = $7;
  (_lua_pushfstring($L,595867,$vararg_buffer1)|0);
  _lua_pushinteger($L,$6);
  $$0$i = 3;
  STACKTOP = sp;return ($$0$i|0);
 }
 return (0)|0;
}
function _io_input($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _g_iofile($L,1,527143);
 return 1;
}
function _g_iofile($L,$f,$mode) {
 $L = $L|0;
 $f = $f|0;
 $mode = $mode|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_lua_type($L,1)|0);
 $1 = ($0|0)<(1);
 if ($1) {
  _lua_rawgeti($L,-10001,$f);
  STACKTOP = sp;return;
 }
 $2 = (_lua_tolstring($L,1,0)|0);
 $3 = ($2|0)==(0|0);
 if ($3) {
  $11 = (_luaL_checkudata($L,1,521477)|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ($12|0)==(0|0);
  if ($13) {
   (_luaL_error($L,521489,$vararg_buffer2)|0);
  }
  _lua_pushvalue($L,1);
 } else {
  $4 = (_lua_newuserdata($L,4)|0);
  HEAP32[$4>>2] = 0;
  _lua_getfield($L,-10000,521477);
  (_lua_setmetatable($L,-2)|0);
  $5 = (_fopen($2,$mode)|0);
  HEAP32[$4>>2] = $5;
  $6 = ($5|0)==(0|0);
  if ($6) {
   $7 = (___errno_location()|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = (_strerror($8)|0);
   HEAP32[$vararg_buffer>>2] = $2;
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
   HEAP32[$vararg_ptr1>>2] = $9;
   (_lua_pushfstring($L,522142,$vararg_buffer)|0);
   $10 = (_lua_tolstring($L,-1,0)|0);
   (_luaL_argerror($L,1,$10)|0);
  }
 }
 _lua_rawseti($L,-10001,$f);
 _lua_rawgeti($L,-10001,$f);
 STACKTOP = sp;return;
}
function _io_lines($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_ptr3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_lua_type($L,1)|0);
 $1 = ($0|0)<(1);
 if ($1) {
  _lua_rawgeti($L,-10001,1);
  $2 = (_luaL_checkudata($L,1,521477)|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ($3|0)==(0|0);
  if ($4) {
   (_luaL_error($L,521489,$vararg_buffer)|0);
  }
  _lua_pushvalue($L,1);
  _lua_pushboolean($L,0);
  _lua_pushcclosure($L,4318,2);
  STACKTOP = sp;return 1;
 } else {
  $5 = (_luaL_checklstring($L,1,0)|0);
  $6 = (_lua_newuserdata($L,4)|0);
  HEAP32[$6>>2] = 0;
  _lua_getfield($L,-10000,521477);
  (_lua_setmetatable($L,-2)|0);
  $7 = (_fopen($5,527143)|0);
  HEAP32[$6>>2] = $7;
  $8 = ($7|0)==(0|0);
  if ($8) {
   $9 = (___errno_location()|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = (_strerror($10)|0);
   HEAP32[$vararg_buffer1>>2] = $5;
   $vararg_ptr3 = ((($vararg_buffer1)) + 4|0);
   HEAP32[$vararg_ptr3>>2] = $11;
   (_lua_pushfstring($L,522142,$vararg_buffer1)|0);
   $12 = (_lua_tolstring($L,-1,0)|0);
   (_luaL_argerror($L,1,$12)|0);
  }
  $13 = (_lua_gettop($L)|0);
  _lua_pushvalue($L,$13);
  _lua_pushboolean($L,1);
  _lua_pushcclosure($L,4318,2);
  STACKTOP = sp;return 1;
 }
 return (0)|0;
}
function _io_open($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_luaL_checklstring($L,1,0)|0);
 $1 = (_luaL_optlstring($L,2,527143,0)|0);
 $2 = (_lua_newuserdata($L,4)|0);
 HEAP32[$2>>2] = 0;
 _lua_getfield($L,-10000,521477);
 (_lua_setmetatable($L,-2)|0);
 $3 = (_fopen($0,$1)|0);
 HEAP32[$2>>2] = $3;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  $9 = 1;
  STACKTOP = sp;return ($9|0);
 }
 $5 = (___errno_location()|0);
 $6 = HEAP32[$5>>2]|0;
 _lua_pushnil($L);
 $7 = ($0|0)==(0|0);
 $8 = (_strerror($6)|0);
 if ($7) {
  HEAP32[$vararg_buffer2>>2] = $8;
  (_lua_pushfstring($L,595867,$vararg_buffer2)|0);
 } else {
  HEAP32[$vararg_buffer>>2] = $0;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $8;
  (_lua_pushfstring($L,522142,$vararg_buffer)|0);
 }
 _lua_pushinteger($L,$6);
 $9 = 3;
 STACKTOP = sp;return ($9|0);
}
function _io_output($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _g_iofile($L,2,612357);
 return 1;
}
function _io_popen($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, $vararg_ptr3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_luaL_checklstring($L,1,0)|0);
 (_luaL_optlstring($L,2,527143,0)|0);
 $1 = (_lua_newuserdata($L,4)|0);
 HEAP32[$1>>2] = 0;
 _lua_getfield($L,-10000,521477);
 (_lua_setmetatable($L,-2)|0);
 (_luaL_error($L,521755,$vararg_buffer)|0);
 HEAP32[$1>>2] = 0;
 $2 = (___errno_location()|0);
 $3 = HEAP32[$2>>2]|0;
 _lua_pushnil($L);
 $4 = ($0|0)==(0|0);
 $5 = (_strerror($3)|0);
 if ($4) {
  HEAP32[$vararg_buffer4>>2] = $5;
  (_lua_pushfstring($L,595867,$vararg_buffer4)|0);
  _lua_pushinteger($L,$3);
  STACKTOP = sp;return 3;
 } else {
  HEAP32[$vararg_buffer1>>2] = $0;
  $vararg_ptr3 = ((($vararg_buffer1)) + 4|0);
  HEAP32[$vararg_ptr3>>2] = $5;
  (_lua_pushfstring($L,522142,$vararg_buffer1)|0);
  _lua_pushinteger($L,$3);
  STACKTOP = sp;return 3;
 }
 return (0)|0;
}
function _io_read($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _lua_rawgeti($L,-10001,1);
 $0 = (_lua_touserdata($L,-1)|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  HEAP32[$vararg_buffer>>2] = 521738;
  (_luaL_error($L,521704,$vararg_buffer)|0);
 }
 $3 = (_g_read($L,$1,1)|0);
 STACKTOP = sp;return ($3|0);
}
function _io_tmpfile($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_lua_newuserdata($L,4)|0);
 HEAP32[$0>>2] = 0;
 _lua_getfield($L,-10000,521477);
 (_lua_setmetatable($L,-2)|0);
 $1 = (_tmpfile()|0);
 HEAP32[$0>>2] = $1;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $6 = 1;
  STACKTOP = sp;return ($6|0);
 }
 $3 = (___errno_location()|0);
 $4 = HEAP32[$3>>2]|0;
 _lua_pushnil($L);
 $5 = (_strerror($4)|0);
 HEAP32[$vararg_buffer>>2] = $5;
 (_lua_pushfstring($L,595867,$vararg_buffer)|0);
 _lua_pushinteger($L,$4);
 $6 = 3;
 STACKTOP = sp;return ($6|0);
}
function _io_type($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checkany($L,1);
 $0 = (_lua_touserdata($L,1)|0);
 _lua_getfield($L,-10000,521477);
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  $2 = (_lua_getmetatable($L,1)|0);
  $3 = ($2|0)==(0);
  if (!($3)) {
   $4 = (_lua_rawequal($L,-2,-1)|0);
   $5 = ($4|0)==(0);
   if (!($5)) {
    $6 = HEAP32[$0>>2]|0;
    $7 = ($6|0)==(0|0);
    if ($7) {
     _lua_pushlstring($L,521790,11);
     return 1;
    } else {
     _lua_pushlstring($L,521802,4);
     return 1;
    }
   }
  }
 }
 _lua_pushnil($L);
 return 1;
}
function _io_write($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _lua_rawgeti($L,-10001,2);
 $0 = (_lua_touserdata($L,-1)|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  HEAP32[$vararg_buffer>>2] = 521731;
  (_luaL_error($L,521704,$vararg_buffer)|0);
 }
 $3 = (_g_write($L,$1,1)|0);
 STACKTOP = sp;return ($3|0);
}
function _io_noclose($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _lua_pushnil($L);
 _lua_pushlstring($L,521807,26);
 return 2;
}
function _io_pclose($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_luaL_checkudata($L,1,521477)|0);
 HEAP32[$0>>2] = 0;
 $1 = (___errno_location()|0);
 $2 = HEAP32[$1>>2]|0;
 _lua_pushnil($L);
 $3 = (_strerror($2)|0);
 HEAP32[$vararg_buffer>>2] = $3;
 (_lua_pushfstring($L,595867,$vararg_buffer)|0);
 _lua_pushinteger($L,$2);
 STACKTOP = sp;return 3;
}
function _luaopen_math($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_register($L,521854,39788);
 _lua_pushnumber($L,3.1415926535897931);
 _lua_setfield($L,-2,522041);
 _lua_pushnumber($L,inf);
 _lua_setfield($L,-2,522044);
 _lua_getfield($L,-1,521912);
 _lua_setfield($L,-2,522049);
 return 1;
}
function _math_abs($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+Math_abs((+$0)));
 _lua_pushnumber($L,$1);
 return 1;
}
function _math_acos($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+Math_acos((+$0)));
 _lua_pushnumber($L,$1);
 return 1;
}
function _math_asin($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+Math_asin((+$0)));
 _lua_pushnumber($L,$1);
 return 1;
}
function _math_atan2($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+_luaL_checknumber($L,2));
 $2 = (+Math_atan2((+$0),(+$1)));
 _lua_pushnumber($L,$2);
 return 1;
}
function _math_atan($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+Math_atan((+$0)));
 _lua_pushnumber($L,$1);
 return 1;
}
function _math_ceil($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+Math_ceil((+$0)));
 _lua_pushnumber($L,$1);
 return 1;
}
function _math_cosh($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+_cosh($0));
 _lua_pushnumber($L,$1);
 return 1;
}
function _math_cos($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+Math_cos((+$0)));
 _lua_pushnumber($L,$1);
 return 1;
}
function _math_deg($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = $0 / 0.017453292519943295;
 _lua_pushnumber($L,$1);
 return 1;
}
function _math_exp($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+Math_exp((+$0)));
 _lua_pushnumber($L,$1);
 return 1;
}
function _math_floor($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+Math_floor((+$0)));
 _lua_pushnumber($L,$1);
 return 1;
}
function _math_fmod($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+_luaL_checknumber($L,2));
 $2 = (+_fmod($0,$1));
 _lua_pushnumber($L,$2);
 return 1;
}
function _math_frexp($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, $2 = 0, $e = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $e = sp;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+_frexp($0,$e));
 _lua_pushnumber($L,$1);
 $2 = HEAP32[$e>>2]|0;
 _lua_pushinteger($L,$2);
 STACKTOP = sp;return 2;
}
function _math_ldexp($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (_luaL_checkinteger($L,2)|0);
 $2 = (+_ldexp($0,$1));
 _lua_pushnumber($L,$2);
 return 1;
}
function _math_log10($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+_log10($0));
 _lua_pushnumber($L,$1);
 return 1;
}
function _math_log($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+Math_log((+$0)));
 _lua_pushnumber($L,$1);
 return 1;
}
function _math_max($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0.0, $4 = 0, $5 = 0, $dmax$0$lcssa = 0.0, $dmax$02 = 0.0, $dmax$1 = 0.0, $exitcond = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_gettop($L)|0);
 $1 = (+_luaL_checknumber($L,1));
 $2 = ($0|0)<(2);
 if ($2) {
  $dmax$0$lcssa = $1;
  _lua_pushnumber($L,$dmax$0$lcssa);
  return 1;
 } else {
  $dmax$02 = $1;$i$01 = 2;
 }
 while(1) {
  $3 = (+_luaL_checknumber($L,$i$01));
  $4 = $3 > $dmax$02;
  $dmax$1 = $4 ? $3 : $dmax$02;
  $5 = (($i$01) + 1)|0;
  $exitcond = ($i$01|0)==($0|0);
  if ($exitcond) {
   $dmax$0$lcssa = $dmax$1;
   break;
  } else {
   $dmax$02 = $dmax$1;$i$01 = $5;
  }
 }
 _lua_pushnumber($L,$dmax$0$lcssa);
 return 1;
}
function _math_min($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0.0, $4 = 0, $5 = 0, $dmin$0$lcssa = 0.0, $dmin$02 = 0.0, $dmin$1 = 0.0, $exitcond = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_gettop($L)|0);
 $1 = (+_luaL_checknumber($L,1));
 $2 = ($0|0)<(2);
 if ($2) {
  $dmin$0$lcssa = $1;
  _lua_pushnumber($L,$dmin$0$lcssa);
  return 1;
 } else {
  $dmin$02 = $1;$i$01 = 2;
 }
 while(1) {
  $3 = (+_luaL_checknumber($L,$i$01));
  $4 = $3 < $dmin$02;
  $dmin$1 = $4 ? $3 : $dmin$02;
  $5 = (($i$01) + 1)|0;
  $exitcond = ($i$01|0)==($0|0);
  if ($exitcond) {
   $dmin$0$lcssa = $dmin$1;
   break;
  } else {
   $dmin$02 = $dmin$1;$i$01 = $5;
  }
 }
 _lua_pushnumber($L,$dmin$0$lcssa);
 return 1;
}
function _math_modf($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, $2 = 0.0, $ip = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ip = sp;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+_modf($0,$ip));
 $2 = +HEAPF64[$ip>>3];
 _lua_pushnumber($L,$2);
 _lua_pushnumber($L,$1);
 STACKTOP = sp;return 2;
}
function _math_pow($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+_luaL_checknumber($L,2));
 $2 = (+Math_pow((+$0),(+$1)));
 _lua_pushnumber($L,$2);
 return 1;
}
function _math_rad($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = $0 * 0.017453292519943295;
 _lua_pushnumber($L,$1);
 return 1;
}
function _math_random($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0.0, $21 = 0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0.0, $8 = 0.0, $9 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_rand()|0);
 $1 = (($0|0) % 2147483647)&-1;
 $2 = (+($1|0));
 $3 = $2 / 2147483647.0;
 $4 = (_lua_gettop($L)|0);
 switch ($4|0) {
 case 0:  {
  _lua_pushnumber($L,$3);
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 case 1:  {
  $5 = (_luaL_checkinteger($L,1)|0);
  $6 = ($5|0)>(0);
  if (!($6)) {
   (_luaL_argerror($L,1,521963)|0);
  }
  $7 = (+($5|0));
  $8 = $3 * $7;
  $9 = (+Math_floor((+$8)));
  $10 = $9 + 1.0;
  _lua_pushnumber($L,$10);
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 case 2:  {
  $11 = (_luaL_checkinteger($L,1)|0);
  $12 = (_luaL_checkinteger($L,2)|0);
  $13 = ($12|0)<($11|0);
  if ($13) {
   (_luaL_argerror($L,2,521963)|0);
  }
  $14 = (1 - ($11))|0;
  $15 = (($14) + ($12))|0;
  $16 = (+($15|0));
  $17 = $3 * $16;
  $18 = (+Math_floor((+$17)));
  $19 = (+($11|0));
  $20 = $19 + $18;
  _lua_pushnumber($L,$20);
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 default: {
  $21 = (_luaL_error($L,521981,$vararg_buffer)|0);
  $$0 = $21;
  STACKTOP = sp;return ($$0|0);
 }
 }
 return (0)|0;
}
function _math_randomseed($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkinteger($L,1)|0);
 _srand($0);
 return 0;
}
function _math_sinh($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+_sinh($0));
 _lua_pushnumber($L,$1);
 return 1;
}
function _math_sin($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+Math_sin((+$0)));
 _lua_pushnumber($L,$1);
 return 1;
}
function _math_sqrt($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+Math_sqrt((+$0)));
 _lua_pushnumber($L,$1);
 return 1;
}
function _math_tanh($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+_tanh($0));
 _lua_pushnumber($L,$1);
 return 1;
}
function _math_tan($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (+Math_tan((+$0)));
 _lua_pushnumber($L,$1);
 return 1;
}
function _luaopen_os($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_register($L,522053,40020);
 return 1;
}
function _os_clock($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_clock()|0);
 $1 = (+($0|0));
 $2 = $1 / 1.0E+6;
 _lua_pushnumber($L,$2);
 return 1;
}
function _os_date($L) {
 $L = $L|0;
 var $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b = 0, $buff = 0, $cc = 0, $s$0 = 0, $s$1 = 0, $s$2 = 0, $stm$0 = 0;
 var $t = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1248|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $t = sp + 1036|0;
 $cc = sp + 1240|0;
 $b = sp;
 $buff = sp + 1040|0;
 $0 = (_luaL_optlstring($L,1,522067,0)|0);
 $1 = (_lua_type($L,2)|0);
 $2 = ($1|0)<(1);
 if ($2) {
  $3 = (_time((0|0))|0);
  $6 = $3;
 } else {
  $4 = (+_luaL_checknumber($L,2));
  $5 = (~~(($4)));
  $6 = $5;
 }
 HEAP32[$t>>2] = $6;
 $7 = HEAP8[$0>>0]|0;
 $8 = ($7<<24>>24)==(33);
 if ($8) {
  $9 = (_gmtime(($t|0))|0);
  $10 = ((($0)) + 1|0);
  $s$0 = $10;$stm$0 = $9;
 } else {
  $11 = (_localtime(($t|0))|0);
  $s$0 = $0;$stm$0 = $11;
 }
 $12 = ($stm$0|0)==(0|0);
 if ($12) {
  _lua_pushnil($L);
  STACKTOP = sp;return 1;
 }
 $13 = (_strcmp($s$0,522070)|0);
 $14 = ($13|0)==(0);
 if ($14) {
  _lua_createtable($L,0,9);
  $15 = HEAP32[$stm$0>>2]|0;
  _lua_pushinteger($L,$15);
  _lua_setfield($L,-2,522073);
  $16 = ((($stm$0)) + 4|0);
  $17 = HEAP32[$16>>2]|0;
  _lua_pushinteger($L,$17);
  _lua_setfield($L,-2,525433);
  $18 = ((($stm$0)) + 8|0);
  $19 = HEAP32[$18>>2]|0;
  _lua_pushinteger($L,$19);
  _lua_setfield($L,-2,522077);
  $20 = ((($stm$0)) + 12|0);
  $21 = HEAP32[$20>>2]|0;
  _lua_pushinteger($L,$21);
  _lua_setfield($L,-2,522082);
  $22 = ((($stm$0)) + 16|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = (($23) + 1)|0;
  _lua_pushinteger($L,$24);
  _lua_setfield($L,-2,522086);
  $25 = ((($stm$0)) + 20|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = (($26) + 1900)|0;
  _lua_pushinteger($L,$27);
  _lua_setfield($L,-2,522092);
  $28 = ((($stm$0)) + 24|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = (($29) + 1)|0;
  _lua_pushinteger($L,$30);
  _lua_setfield($L,-2,522097);
  $31 = ((($stm$0)) + 28|0);
  $32 = HEAP32[$31>>2]|0;
  $33 = (($32) + 1)|0;
  _lua_pushinteger($L,$33);
  _lua_setfield($L,-2,522102);
  $34 = ((($stm$0)) + 32|0);
  $35 = HEAP32[$34>>2]|0;
  $36 = ($35|0)<(0);
  if ($36) {
   STACKTOP = sp;return 1;
  }
  _lua_pushboolean($L,$35);
  _lua_setfield($L,-2,522107);
  STACKTOP = sp;return 1;
 }
 HEAP8[$cc>>0] = 37;
 $37 = ((($cc)) + 2|0);
 HEAP8[$37>>0] = 0;
 _luaL_buffinit($L,$b);
 $38 = ((($b)) + 1036|0);
 $39 = ((($cc)) + 1|0);
 $s$1 = $s$0;
 L20: while(1) {
  $40 = HEAP8[$s$1>>0]|0;
  switch ($40<<24>>24) {
  case 0:  {
   break L20;
   break;
  }
  case 37:  {
   $41 = ((($s$1)) + 1|0);
   $42 = HEAP8[$41>>0]|0;
   $43 = ($42<<24>>24)==(0);
   if ($43) {
    label = 15;
   } else {
    HEAP8[$39>>0] = $42;
    $49 = (_strftime(($buff|0),200,($cc|0),($stm$0|0))|0);
    _luaL_addlstring($b,$buff,$49);
    $s$2 = $41;
   }
   break;
  }
  default: {
   label = 15;
  }
  }
  if ((label|0) == 15) {
   label = 0;
   $44 = HEAP32[$b>>2]|0;
   $45 = ($44>>>0)<($38>>>0);
   if ($45) {
    $47 = $44;$48 = $40;
   } else {
    (_luaL_prepbuffer($b)|0);
    $$pre = HEAP8[$s$1>>0]|0;
    $$pre1 = HEAP32[$b>>2]|0;
    $47 = $$pre1;$48 = $$pre;
   }
   $46 = ((($47)) + 1|0);
   HEAP32[$b>>2] = $46;
   HEAP8[$47>>0] = $48;
   $s$2 = $s$1;
  }
  $50 = ((($s$2)) + 1|0);
  $s$1 = $50;
 }
 _luaL_pushresult($b);
 STACKTOP = sp;return 1;
}
function _os_difftime($L) {
 $L = $L|0;
 var $0 = 0.0, $1 = 0, $2 = 0.0, $3 = 0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_luaL_checknumber($L,1));
 $1 = (~~(($0)));
 $2 = (+_luaL_optnumber($L,2,0.0));
 $3 = (~~(($2)));
 $4 = (+_difftime(($1|0),($3|0)));
 _lua_pushnumber($L,$4);
 return 1;
}
function _os_execute($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_optlstring($L,1,0,0)|0);
 $1 = (_system(($0|0))|0);
 _lua_pushinteger($L,$1);
 return 1;
}
function _os_exit($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_optinteger($L,1,0)|0);
 _exit(($0|0));
 // unreachable;
 return (0)|0;
}
function _os_getenv($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checklstring($L,1,0)|0);
 $1 = (_getenv(($0|0))|0);
 _lua_pushstring($L,$1);
 return 1;
}
function _os_remove($L) {
 $L = $L|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_luaL_checklstring($L,1,0)|0);
 $1 = (_remove($0)|0);
 $2 = ($1|0)==(0);
 $3 = (___errno_location()|0);
 $4 = HEAP32[$3>>2]|0;
 if ($2) {
  _lua_pushboolean($L,1);
  $$0$i = 1;
  STACKTOP = sp;return ($$0$i|0);
 } else {
  _lua_pushnil($L);
  $5 = (_strerror($4)|0);
  HEAP32[$vararg_buffer>>2] = $0;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $5;
  (_lua_pushfstring($L,522142,$vararg_buffer)|0);
  _lua_pushinteger($L,$4);
  $$0$i = 3;
  STACKTOP = sp;return ($$0$i|0);
 }
 return (0)|0;
}
function _os_rename($L) {
 $L = $L|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_luaL_checklstring($L,1,0)|0);
 $1 = (_luaL_checklstring($L,2,0)|0);
 $2 = (_rename($0,$1)|0);
 $3 = ($2|0)==(0);
 $4 = (___errno_location()|0);
 $5 = HEAP32[$4>>2]|0;
 if ($3) {
  _lua_pushboolean($L,1);
  $$0$i = 1;
  STACKTOP = sp;return ($$0$i|0);
 } else {
  _lua_pushnil($L);
  $6 = (_strerror($5)|0);
  HEAP32[$vararg_buffer>>2] = $0;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $6;
  (_lua_pushfstring($L,522142,$vararg_buffer)|0);
  _lua_pushinteger($L,$5);
  $$0$i = 3;
  STACKTOP = sp;return ($$0$i|0);
 }
 return (0)|0;
}
function _os_setlocale($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_optlstring($L,1,0,0)|0);
 $1 = (_luaL_checkoption($L,2,527870,40116)|0);
 $2 = (40144 + ($1<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_setlocale($3,$0)|0);
 _lua_pushstring($L,$4);
 return 1;
}
function _os_time($L) {
 $L = $L|0;
 var $$0$i = 0, $$0$i11 = 0, $$0$i8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $res$0$i = 0, $res$0$i1 = 0, $res$0$i3 = 0, $t$0 = 0, $ts = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $ts = sp + 20|0;
 $0 = (_lua_type($L,1)|0);
 $1 = ($0|0)<(1);
 if ($1) {
  $2 = (_time((0|0))|0);
  $t$0 = $2;
 } else {
  _luaL_checktype($L,1,5);
  _lua_settop($L,1);
  _lua_getfield($L,-1,522073);
  $3 = (_lua_isnumber($L,-1)|0);
  $4 = ($3|0)==(0);
  if ($4) {
   $res$0$i = 0;
  } else {
   $5 = (_lua_tointeger($L,-1)|0);
   $res$0$i = $5;
  }
  _lua_settop($L,-2);
  HEAP32[$ts>>2] = $res$0$i;
  _lua_getfield($L,-1,525433);
  $6 = (_lua_isnumber($L,-1)|0);
  $7 = ($6|0)==(0);
  if ($7) {
   $res$0$i1 = 0;
  } else {
   $8 = (_lua_tointeger($L,-1)|0);
   $res$0$i1 = $8;
  }
  _lua_settop($L,-2);
  $9 = ((($ts)) + 4|0);
  HEAP32[$9>>2] = $res$0$i1;
  _lua_getfield($L,-1,522077);
  $10 = (_lua_isnumber($L,-1)|0);
  $11 = ($10|0)==(0);
  if ($11) {
   $res$0$i3 = 12;
  } else {
   $12 = (_lua_tointeger($L,-1)|0);
   $res$0$i3 = $12;
  }
  _lua_settop($L,-2);
  $13 = ((($ts)) + 8|0);
  HEAP32[$13>>2] = $res$0$i3;
  _lua_getfield($L,-1,522082);
  $14 = (_lua_isnumber($L,-1)|0);
  $15 = ($14|0)==(0);
  if ($15) {
   HEAP32[$vararg_buffer>>2] = 522082;
   $17 = (_luaL_error($L,522202,$vararg_buffer)|0);
   $$0$i = $17;
  } else {
   $16 = (_lua_tointeger($L,-1)|0);
   _lua_settop($L,-2);
   $$0$i = $16;
  }
  $18 = ((($ts)) + 12|0);
  HEAP32[$18>>2] = $$0$i;
  _lua_getfield($L,-1,522086);
  $19 = (_lua_isnumber($L,-1)|0);
  $20 = ($19|0)==(0);
  if ($20) {
   HEAP32[$vararg_buffer1>>2] = 522086;
   $22 = (_luaL_error($L,522202,$vararg_buffer1)|0);
   $$0$i8 = $22;
  } else {
   $21 = (_lua_tointeger($L,-1)|0);
   _lua_settop($L,-2);
   $$0$i8 = $21;
  }
  $23 = (($$0$i8) + -1)|0;
  $24 = ((($ts)) + 16|0);
  HEAP32[$24>>2] = $23;
  _lua_getfield($L,-1,522092);
  $25 = (_lua_isnumber($L,-1)|0);
  $26 = ($25|0)==(0);
  if ($26) {
   HEAP32[$vararg_buffer4>>2] = 522092;
   $28 = (_luaL_error($L,522202,$vararg_buffer4)|0);
   $$0$i11 = $28;
  } else {
   $27 = (_lua_tointeger($L,-1)|0);
   _lua_settop($L,-2);
   $$0$i11 = $27;
  }
  $29 = (($$0$i11) + -1900)|0;
  $30 = ((($ts)) + 20|0);
  HEAP32[$30>>2] = $29;
  _lua_getfield($L,-1,522107);
  $31 = (_lua_type($L,-1)|0);
  $32 = ($31|0)==(0);
  if ($32) {
   $35 = -1;
  } else {
   $33 = (_lua_toboolean($L,-1)|0);
   $35 = $33;
  }
  _lua_settop($L,-2);
  $34 = ((($ts)) + 32|0);
  HEAP32[$34>>2] = $35;
  $36 = (_mktime(($ts|0))|0);
  $t$0 = $36;
 }
 $37 = ($t$0|0)==(-1);
 if ($37) {
  _lua_pushnil($L);
  STACKTOP = sp;return 1;
 } else {
  $38 = (+($t$0|0));
  _lua_pushnumber($L,$38);
  STACKTOP = sp;return 1;
 }
 return (0)|0;
}
function _os_tmpname($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $buff = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $buff = sp + 4|0;
 $0 = (_tmpnam($buff)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  $2 = (_luaL_error($L,522243,$vararg_buffer)|0);
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 } else {
  _lua_pushstring($L,$buff);
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _luaopen_table($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_register($L,522280,40168);
 return 1;
}
function _tconcat($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b = 0;
 var $exitcond = 0, $i$02 = 0, $lsep = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1056|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $b = sp + 20|0;
 $lsep = sp + 16|0;
 $0 = (_luaL_optlstring($L,2,1282466,$lsep)|0);
 _luaL_checktype($L,1,5);
 $1 = (_luaL_optinteger($L,3,1)|0);
 $2 = (_lua_type($L,4)|0);
 $3 = ($2|0)<(1);
 if ($3) {
  $4 = (_lua_objlen($L,1)|0);
  $7 = $4;
 } else {
  $5 = (_luaL_checkinteger($L,4)|0);
  $7 = $5;
 }
 _luaL_buffinit($L,$b);
 $6 = ($1|0)<($7|0);
 if ($6) {
  $i$02 = $1;
  while(1) {
   _lua_rawgeti($L,1,$i$02);
   $8 = (_lua_isstring($L,-1)|0);
   $9 = ($8|0)==(0);
   if ($9) {
    $10 = (_lua_type($L,-1)|0);
    $11 = (_lua_typename($L,$10)|0);
    HEAP32[$vararg_buffer>>2] = $11;
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
    HEAP32[$vararg_ptr1>>2] = $i$02;
    (_luaL_error($L,522293,$vararg_buffer)|0);
   }
   _luaL_addvalue($b);
   $12 = HEAP32[$lsep>>2]|0;
   _luaL_addlstring($b,$0,$12);
   $13 = (($i$02) + 1)|0;
   $exitcond = ($13|0)==($7|0);
   if ($exitcond) {
    break;
   } else {
    $i$02 = $13;
   }
  }
 } else {
  $14 = ($1|0)==($7|0);
  if (!($14)) {
   _luaL_pushresult($b);
   STACKTOP = sp;return 1;
  }
 }
 _lua_rawgeti($L,1,$7);
 $15 = (_lua_isstring($L,-1)|0);
 $16 = ($15|0)==(0);
 if ($16) {
  $17 = (_lua_type($L,-1)|0);
  $18 = (_lua_typename($L,$17)|0);
  HEAP32[$vararg_buffer2>>2] = $18;
  $vararg_ptr5 = ((($vararg_buffer2)) + 4|0);
  HEAP32[$vararg_ptr5>>2] = $7;
  (_luaL_error($L,522293,$vararg_buffer2)|0);
 }
 _luaL_addvalue($b);
 _luaL_pushresult($b);
 STACKTOP = sp;return 1;
}
function _foreach($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checktype($L,1,5);
 _luaL_checktype($L,2,6);
 _lua_pushnil($L);
 $0 = (_lua_next($L,1)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  $$0 = 0;
  return ($$0|0);
 }
 while(1) {
  _lua_pushvalue($L,2);
  _lua_pushvalue($L,-3);
  _lua_pushvalue($L,-3);
  _lua_call($L,2,1);
  $2 = (_lua_type($L,-1)|0);
  $3 = ($2|0)==(0);
  if (!($3)) {
   $$0 = 1;
   label = 4;
   break;
  }
  _lua_settop($L,-3);
  $4 = (_lua_next($L,1)|0);
  $5 = ($4|0)==(0);
  if ($5) {
   $$0 = 0;
   label = 4;
   break;
  }
 }
 if ((label|0) == 4) {
  return ($$0|0);
 }
 return (0)|0;
}
function _foreachi($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checktype($L,1,5);
 $0 = (_lua_objlen($L,1)|0);
 _luaL_checktype($L,2,6);
 $1 = ($0|0)<(1);
 if ($1) {
  $$0 = 0;
  return ($$0|0);
 } else {
  $i$01 = 1;
 }
 while(1) {
  _lua_pushvalue($L,2);
  _lua_pushinteger($L,$i$01);
  _lua_rawgeti($L,1,$i$01);
  _lua_call($L,2,1);
  $2 = (_lua_type($L,-1)|0);
  $3 = ($2|0)==(0);
  if (!($3)) {
   $$0 = 1;
   label = 4;
   break;
  }
  _lua_settop($L,-2);
  $4 = (($i$01) + 1)|0;
  $5 = ($i$01|0)<($0|0);
  if ($5) {
   $i$01 = $4;
  } else {
   $$0 = 0;
   label = 4;
   break;
  }
 }
 if ((label|0) == 4) {
  return ($$0|0);
 }
 return (0)|0;
}
function _getn($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checktype($L,1,5);
 $0 = (_lua_objlen($L,1)|0);
 _lua_pushinteger($L,$0);
 return 1;
}
function _maxn($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $max$0$ph = 0.0, $max$0$ph$lcssa = 0.0, $max$1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checktype($L,1,5);
 _lua_pushnil($L);
 $max$0$ph = 0.0;
 L1: while(1) {
  while(1) {
   $0 = (_lua_next($L,1)|0);
   $1 = ($0|0)==(0);
   if ($1) {
    $max$0$ph$lcssa = $max$0$ph;
    break L1;
   }
   _lua_settop($L,-2);
   $2 = (_lua_type($L,-1)|0);
   $3 = ($2|0)==(3);
   if ($3) {
    break;
   }
  }
  $4 = (+_lua_tonumber($L,-1));
  $5 = $4 > $max$0$ph;
  $max$1 = $5 ? $4 : $max$0$ph;
  $max$0$ph = $max$1;
 }
 _lua_pushnumber($L,$max$0$ph$lcssa);
 return 1;
}
function _tinsert($L) {
 $L = $L|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $i$01 = 0, $pos$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _luaL_checktype($L,1,5);
 $0 = (_lua_objlen($L,1)|0);
 $1 = (($0) + 1)|0;
 $2 = (_lua_gettop($L)|0);
 switch ($2|0) {
 case 2:  {
  $pos$0 = $1;
  break;
 }
 case 3:  {
  $3 = (_luaL_checkinteger($L,2)|0);
  $4 = ($3|0)>($1|0);
  $$ = $4 ? $3 : $1;
  $5 = ($$|0)>($3|0);
  if ($5) {
   $i$01 = $$;
   while(1) {
    $6 = (($i$01) + -1)|0;
    _lua_rawgeti($L,1,$6);
    _lua_rawseti($L,1,$i$01);
    $7 = ($6|0)>($3|0);
    if ($7) {
     $i$01 = $6;
    } else {
     $pos$0 = $3;
     break;
    }
   }
  } else {
   $pos$0 = $3;
  }
  break;
 }
 default: {
  $8 = (_luaL_error($L,522380,$vararg_buffer)|0);
  $$0 = $8;
  STACKTOP = sp;return ($$0|0);
 }
 }
 _lua_rawseti($L,1,$pos$0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _tremove($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $exitcond = 0, $or$cond = 0, $pos$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checktype($L,1,5);
 $0 = (_lua_objlen($L,1)|0);
 $1 = (_luaL_optinteger($L,2,$0)|0);
 $2 = ($1|0)<(1);
 $3 = ($1|0)>($0|0);
 $or$cond = $2 | $3;
 if ($or$cond) {
  $$0 = 0;
  return ($$0|0);
 }
 _lua_rawgeti($L,1,$1);
 $4 = ($1|0)<($0|0);
 if ($4) {
  $pos$01 = $1;
  while(1) {
   $5 = (($pos$01) + 1)|0;
   _lua_rawgeti($L,1,$5);
   _lua_rawseti($L,1,$pos$01);
   $exitcond = ($5|0)==($0|0);
   if ($exitcond) {
    break;
   } else {
    $pos$01 = $5;
   }
  }
 }
 _lua_pushnil($L);
 _lua_rawseti($L,1,$0);
 $$0 = 1;
 return ($$0|0);
}
function _setn($L) {
 $L = $L|0;
 var $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _luaL_checktype($L,1,5);
 (_luaL_error($L,522430,$vararg_buffer)|0);
 _lua_pushvalue($L,1);
 STACKTOP = sp;return 1;
}
function _sort($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checktype($L,1,5);
 $0 = (_lua_objlen($L,1)|0);
 _luaL_checkstack($L,40,1282466);
 $1 = (_lua_type($L,2)|0);
 $2 = ($1|0)<(1);
 if (!($2)) {
  _luaL_checktype($L,2,6);
 }
 _lua_settop($L,2);
 _auxsort($L,1,$0);
 return 0;
}
function _auxsort($L,$l,$u) {
 $L = $L|0;
 $l = $l|0;
 $u = $u|0;
 var $$$02 = 0, $$01$i$0 = 0, $$0113 = 0, $$02$ = 0, $$0212 = 0, $$lcssa = 0, $$lcssa$lcssa = 0, $$lcssa5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $i$0$$01 = 0, $i$0$lcssa = 0, $i$0$lcssa$lcssa = 0, $i$0$ph = 0, $i$06 = 0, $i$06$phi = 0, $j$0$ph = 0, $j$1$lcssa = 0, $j$18 = 0, $j$18$phi = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($u|0)>($l|0);
 if ($0) {
  $$0113 = $u;$$0212 = $l;
 } else {
  STACKTOP = sp;return;
 }
 while(1) {
  _lua_rawgeti($L,1,$$0212);
  _lua_rawgeti($L,1,$$0113);
  $1 = (_sort_comp($L,-1,-2)|0);
  $2 = ($1|0)==(0);
  if ($2) {
   _lua_settop($L,-3);
  } else {
   _lua_rawseti($L,1,$$0212);
   _lua_rawseti($L,1,$$0113);
  }
  $3 = (($$0113) - ($$0212))|0;
  $4 = ($3|0)==(1);
  if ($4) {
   label = 24;
   break;
  }
  $5 = (($$0113) + ($$0212))|0;
  $6 = (($5|0) / 2)&-1;
  _lua_rawgeti($L,1,$6);
  _lua_rawgeti($L,1,$$0212);
  $7 = (_sort_comp($L,-2,-1)|0);
  $8 = ($7|0)==(0);
  do {
   if ($8) {
    _lua_settop($L,-2);
    _lua_rawgeti($L,1,$$0113);
    $9 = (_sort_comp($L,-1,-2)|0);
    $10 = ($9|0)==(0);
    if ($10) {
     _lua_settop($L,-3);
     break;
    } else {
     _lua_rawseti($L,1,$6);
     _lua_rawseti($L,1,$$0113);
     break;
    }
   } else {
    _lua_rawseti($L,1,$6);
    _lua_rawseti($L,1,$$0212);
   }
  } while(0);
  $11 = ($3|0)==(2);
  if ($11) {
   label = 24;
   break;
  }
  _lua_rawgeti($L,1,$6);
  _lua_pushvalue($L,-1);
  $12 = (($$0113) + -1)|0;
  _lua_rawgeti($L,1,$12);
  _lua_rawseti($L,1,$6);
  _lua_rawseti($L,1,$12);
  $i$0$ph = $$0212;$j$0$ph = $12;
  while(1) {
   $13 = (($i$0$ph) + 1)|0;
   _lua_rawgeti($L,1,$13);
   $14 = (_sort_comp($L,-1,-2)|0);
   $15 = ($14|0)==(0);
   if ($15) {
    $$lcssa = $13;$i$0$lcssa = $i$0$ph;
   } else {
    $21 = $13;$i$06 = $i$0$ph;
    while(1) {
     $19 = ($$0113|0)>($i$06|0);
     if (!($19)) {
      (_luaL_error($L,522454,$vararg_buffer)|0);
     }
     _lua_settop($L,-2);
     $20 = (($21) + 1)|0;
     _lua_rawgeti($L,1,$20);
     $22 = (_sort_comp($L,-1,-2)|0);
     $23 = ($22|0)==(0);
     if ($23) {
      $$lcssa = $20;$i$0$lcssa = $21;
      break;
     } else {
      $i$06$phi = $21;$21 = $20;$i$06 = $i$06$phi;
     }
    }
   }
   $16 = (($j$0$ph) + -1)|0;
   _lua_rawgeti($L,1,$16);
   $17 = (_sort_comp($L,-3,-1)|0);
   $18 = ($17|0)==(0);
   if ($18) {
    $$lcssa5 = $16;$j$1$lcssa = $j$0$ph;
   } else {
    $26 = $16;$j$18 = $j$0$ph;
    while(1) {
     $24 = ($j$18|0)>($$0212|0);
     if (!($24)) {
      (_luaL_error($L,522454,$vararg_buffer1)|0);
     }
     _lua_settop($L,-2);
     $25 = (($26) + -1)|0;
     _lua_rawgeti($L,1,$25);
     $27 = (_sort_comp($L,-3,-1)|0);
     $28 = ($27|0)==(0);
     if ($28) {
      $$lcssa5 = $25;$j$1$lcssa = $26;
      break;
     } else {
      $j$18$phi = $26;$26 = $25;$j$18 = $j$18$phi;
     }
    }
   }
   $29 = ($j$1$lcssa|0)>($$lcssa|0);
   if (!($29)) {
    $$lcssa$lcssa = $$lcssa;$i$0$lcssa$lcssa = $i$0$lcssa;
    break;
   }
   _lua_rawseti($L,1,$$lcssa);
   _lua_rawseti($L,1,$$lcssa5);
   $i$0$ph = $$lcssa;$j$0$ph = $$lcssa5;
  }
  _lua_settop($L,-4);
  _lua_rawgeti($L,1,$12);
  _lua_rawgeti($L,1,$$lcssa$lcssa);
  _lua_rawseti($L,1,$12);
  _lua_rawseti($L,1,$$lcssa$lcssa);
  $30 = (($$lcssa$lcssa) - ($$0212))|0;
  $31 = (($$0113) - ($$lcssa$lcssa))|0;
  $32 = ($30|0)<($31|0);
  $33 = (($i$0$lcssa$lcssa) + 2)|0;
  $$$02 = $32 ? $33 : $$0212;
  $$01$i$0 = $32 ? $$0113 : $i$0$lcssa$lcssa;
  $i$0$$01 = $32 ? $i$0$lcssa$lcssa : $$0113;
  $$02$ = $32 ? $$0212 : $33;
  _auxsort($L,$$02$,$i$0$$01);
  $34 = ($$01$i$0|0)>($$$02|0);
  if ($34) {
   $$0113 = $$01$i$0;$$0212 = $$$02;
  } else {
   label = 24;
   break;
  }
 }
 if ((label|0) == 24) {
  STACKTOP = sp;return;
 }
}
function _sort_comp($L,$a,$b) {
 $L = $L|0;
 $a = $a|0;
 $b = $b|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_lua_type($L,2)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  $5 = (_lua_lessthan($L,$a,$b)|0);
  $$0 = $5;
  return ($$0|0);
 } else {
  _lua_pushvalue($L,2);
  $2 = (($a) + -1)|0;
  _lua_pushvalue($L,$2);
  $3 = (($b) + -2)|0;
  _lua_pushvalue($L,$3);
  _lua_call($L,2,1);
  $4 = (_lua_toboolean($L,-1)|0);
  _lua_settop($L,-2);
  $$0 = $4;
  return ($$0|0);
 }
 return (0)|0;
}
function _luaopen_string($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_register($L,522489,40248);
 _lua_getfield($L,-1,522983);
 _lua_setfield($L,-2,522931);
 _lua_createtable($L,0,1);
 _lua_pushlstring($L,1282466,0);
 _lua_pushvalue($L,-2);
 (_lua_setmetatable($L,-2)|0);
 _lua_settop($L,-2);
 _lua_pushvalue($L,-2);
 _lua_setfield($L,-2,523215);
 _lua_settop($L,-2);
 return 1;
}
function _str_byte($L) {
 $L = $L|0;
 var $$ = 0, $$0 = 0, $$0$i = 0, $$0$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$02 = 0, $l = 0, $pose$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $l = sp + 4|0;
 $0 = (_luaL_checklstring($L,1,$l)|0);
 $1 = (_luaL_optinteger($L,2,1)|0);
 $2 = HEAP32[$l>>2]|0;
 $3 = ($1|0)<(0);
 $4 = (($2) + 1)|0;
 $5 = $3 ? $4 : 0;
 $$0$i = (($5) + ($1))|0;
 $6 = ($$0$i|0)<(0);
 $7 = $6 ? 0 : $$0$i;
 $8 = (_luaL_optinteger($L,3,$7)|0);
 $9 = HEAP32[$l>>2]|0;
 $10 = ($8|0)<(0);
 $11 = (($9) + 1)|0;
 $12 = $10 ? $11 : 0;
 $$0$i1 = (($12) + ($8))|0;
 $13 = ($$0$i1|0)<(0);
 $14 = $13 ? 0 : $$0$i1;
 $15 = ($7|0)<(1);
 $$ = $15 ? 1 : $$0$i;
 $16 = ($14>>>0)>($9>>>0);
 $pose$0 = $16 ? $9 : $14;
 $17 = ($pose$0|0)<($$|0);
 if ($17) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $18 = (($pose$0) - ($$))|0;
 $19 = (($18) + 1)|0;
 $20 = ($pose$0|0)==(2147483647);
 if ($20) {
  (_luaL_error($L,522501,$vararg_buffer)|0);
 }
 _luaL_checkstack($L,$19,522501);
 $21 = ($18|0)<(0);
 if ($21) {
  $$0 = $19;
  STACKTOP = sp;return ($$0|0);
 }
 $22 = (($$) + -1)|0;
 $i$02 = 0;
 while(1) {
  $23 = (($22) + ($i$02))|0;
  $24 = (($0) + ($23)|0);
  $25 = HEAP8[$24>>0]|0;
  $26 = $25&255;
  _lua_pushinteger($L,$26);
  $27 = (($i$02) + 1)|0;
  $28 = ($i$02|0)<($18|0);
  if ($28) {
   $i$02 = $27;
  } else {
   $$0 = $19;
   break;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _str_char($L) {
 $L = $L|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b = 0, $exitcond = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1040|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $b = sp;
 $0 = (_lua_gettop($L)|0);
 _luaL_buffinit($L,$b);
 $1 = ($0|0)<(1);
 if ($1) {
  _luaL_pushresult($b);
  STACKTOP = sp;return 1;
 }
 $2 = ((($b)) + 1036|0);
 $i$01 = 1;
 while(1) {
  $3 = (_luaL_checkinteger($L,$i$01)|0);
  $4 = $3&255;
  $5 = $3 & 255;
  $6 = ($5|0)==($3|0);
  if (!($6)) {
   (_luaL_argerror($L,$i$01,522528)|0);
  }
  $7 = HEAP32[$b>>2]|0;
  $8 = ($7>>>0)<($2>>>0);
  if ($8) {
   $10 = $7;
  } else {
   (_luaL_prepbuffer($b)|0);
   $$pre = HEAP32[$b>>2]|0;
   $10 = $$pre;
  }
  $9 = ((($10)) + 1|0);
  HEAP32[$b>>2] = $9;
  HEAP8[$10>>0] = $4;
  $11 = (($i$01) + 1)|0;
  $exitcond = ($i$01|0)==($0|0);
  if ($exitcond) {
   break;
  } else {
   $i$01 = $11;
  }
 }
 _luaL_pushresult($b);
 STACKTOP = sp;return 1;
}
function _str_dump($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $b = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1040|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $b = sp + 4|0;
 _luaL_checktype($L,1,6);
 _lua_settop($L,1);
 _luaL_buffinit($L,$b);
 $0 = (_lua_dump($L,4319,$b)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  _luaL_pushresult($b);
  STACKTOP = sp;return 1;
 }
 (_luaL_error($L,522547,$vararg_buffer)|0);
 _luaL_pushresult($b);
 STACKTOP = sp;return 1;
}
function _writer($L,$b,$size,$B) {
 $L = $L|0;
 $b = $b|0;
 $size = $size|0;
 $B = $B|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_addlstring($B,$b,$size);
 return 0;
}
function _str_find($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_str_find_aux($L,1)|0);
 return ($0|0);
}
function _str_find_aux($L,$find) {
 $L = $L|0;
 $find = $find|0;
 var $$ = 0, $$$i = 0, $$0$i = 0, $$014$i = 0, $$023$i = 0, $$1$i15 = 0, $$2$ph = 0, $$3 = 0, $$7 = 0, $$lcssa22 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $exitcond$i = 0, $exitcond$i11 = 0, $i$02$i = 0, $i$02$i10 = 0, $init$0 = 0, $l1 = 0, $l2 = 0, $ms = 0, $or$cond$i = 0, $s1$0$lcssa = 0;
 var $s1$036 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 288|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l1 = sp + 276|0;
 $l2 = sp + 272|0;
 $ms = sp;
 $0 = (_luaL_checklstring($L,1,$l1)|0);
 $1 = (_luaL_checklstring($L,2,$l2)|0);
 $2 = (_luaL_optinteger($L,3,1)|0);
 $3 = HEAP32[$l1>>2]|0;
 $4 = ($2|0)<(0);
 $5 = (($3) + 1)|0;
 $6 = $4 ? $5 : 0;
 $$0$i = (($6) + ($2))|0;
 $7 = ($$0$i|0)<(0);
 $8 = $7 ? 0 : $$0$i;
 $9 = (($8) + -1)|0;
 $10 = ($8|0)<(1);
 $11 = ($9>>>0)>($3>>>0);
 $$ = $11 ? $3 : $9;
 $init$0 = $10 ? 0 : $$;
 $12 = ($find|0)!=(0);
 L1: do {
  if ($12) {
   $13 = (_lua_toboolean($L,4)|0);
   $14 = ($13|0)==(0);
   if ($14) {
    $15 = (_strpbrk($1,522582)|0);
    $16 = ($15|0)==(0|0);
    if (!($16)) {
     $$pre = HEAP32[$l1>>2]|0;
     $52 = $$pre;
     label = 14;
     break;
    }
   }
   $17 = (($0) + ($init$0)|0);
   $18 = HEAP32[$l1>>2]|0;
   $19 = (($18) - ($init$0))|0;
   $20 = HEAP32[$l2>>2]|0;
   $21 = ($20|0)==(0);
   L7: do {
    if ($21) {
     $39 = ($17|0)==(0|0);
     if ($39) {
      break L1;
     } else {
      $$1$i15 = $17;
     }
    } else {
     $22 = ($20>>>0)>($19>>>0);
     if ($22) {
      break L1;
     }
     $23 = (($20) + -1)|0;
     $24 = ($19|0)==($23|0);
     if ($24) {
      break L1;
     }
     $25 = (($19) - ($23))|0;
     $26 = HEAP8[$1>>0]|0;
     $27 = $26 << 24 >> 24;
     $28 = ((($1)) + 1|0);
     $$014$i = $25;$$023$i = $17;
     while(1) {
      $29 = (_memchr($$023$i,$27,$$014$i)|0);
      $30 = ($29|0)==(0|0);
      if ($30) {
       break L1;
      }
      $31 = ((($29)) + 1|0);
      $32 = (_memcmp($31,$28,$23)|0);
      $33 = ($32|0)==(0);
      if ($33) {
       $$1$i15 = $29;
       break L7;
      }
      $34 = $31;
      $35 = $$023$i;
      $36 = (($35) + ($$014$i))|0;
      $37 = (($36) - ($34))|0;
      $38 = ($36|0)==($34|0);
      if ($38) {
       break L1;
      } else {
       $$014$i = $37;$$023$i = $31;
      }
     }
    }
   } while(0);
   $40 = $$1$i15;
   $41 = $0;
   $42 = (($40) - ($41))|0;
   $43 = (($42) + 1)|0;
   _lua_pushinteger($L,$43);
   $44 = HEAP32[$l2>>2]|0;
   $45 = (($42) + ($44))|0;
   _lua_pushinteger($L,$45);
   $$3 = 2;
   STACKTOP = sp;return ($$3|0);
  } else {
   $52 = $3;
   label = 14;
  }
 } while(0);
 do {
  if ((label|0) == 14) {
   $46 = HEAP8[$1>>0]|0;
   $47 = ($46<<24>>24)==(94);
   $48 = ((($1)) + 1|0);
   $$7 = $47 ? $48 : $1;
   $49 = (($0) + ($init$0)|0);
   $50 = ((($ms)) + 8|0);
   HEAP32[$50>>2] = $L;
   HEAP32[$ms>>2] = $0;
   $51 = (($0) + ($52)|0);
   $53 = ((($ms)) + 4|0);
   HEAP32[$53>>2] = $51;
   $54 = ((($ms)) + 12|0);
   HEAP32[$54>>2] = 0;
   $55 = (_match($ms,$49,$$7)|0);
   $56 = ($55|0)==(0|0);
   L20: do {
    if ($47) {
     if ($56) {
      label = 25;
     } else {
      $$lcssa22 = $55;$s1$0$lcssa = $49;
     }
    } else {
     if ($56) {
      $s1$036 = $49;
      while(1) {
       $74 = HEAP32[$53>>2]|0;
       $75 = ($s1$036>>>0)<($74>>>0);
       if (!($75)) {
        label = 25;
        break L20;
       }
       $76 = ((($s1$036)) + 1|0);
       HEAP32[$54>>2] = 0;
       $77 = (_match($ms,$76,$$7)|0);
       $78 = ($77|0)==(0|0);
       if ($78) {
        $s1$036 = $76;
       } else {
        $$lcssa22 = $77;$s1$0$lcssa = $76;
        break;
       }
      }
     } else {
      $$lcssa22 = $55;$s1$0$lcssa = $49;
     }
    }
   } while(0);
   if ((label|0) == 25) {
    break;
   }
   if ($12) {
    $57 = $s1$0$lcssa;
    $58 = $0;
    $59 = (1 - ($58))|0;
    $60 = (($59) + ($57))|0;
    _lua_pushinteger($L,$60);
    $61 = $$lcssa22;
    $62 = (($61) - ($58))|0;
    _lua_pushinteger($L,$62);
    $63 = HEAP32[$54>>2]|0;
    $64 = HEAP32[$50>>2]|0;
    _luaL_checkstack($64,$63,522593);
    $65 = ($63|0)>(0);
    if ($65) {
     $i$02$i10 = 0;
     while(1) {
      _push_onecapture($ms,$i$02$i10,0,0);
      $66 = (($i$02$i10) + 1)|0;
      $exitcond$i11 = ($66|0)==($63|0);
      if ($exitcond$i11) {
       break;
      } else {
       $i$02$i10 = $66;
      }
     }
    }
    $67 = (($63) + 2)|0;
    $$2$ph = $67;
   } else {
    $68 = HEAP32[$54>>2]|0;
    $69 = ($68|0)==(0);
    $70 = ($s1$0$lcssa|0)!=(0|0);
    $or$cond$i = $70 & $69;
    $$$i = $or$cond$i ? 1 : $68;
    $71 = HEAP32[$50>>2]|0;
    _luaL_checkstack($71,$$$i,522593);
    $72 = ($$$i|0)>(0);
    if ($72) {
     $i$02$i = 0;
     while(1) {
      _push_onecapture($ms,$i$02$i,$s1$0$lcssa,$$lcssa22);
      $73 = (($i$02$i) + 1)|0;
      $exitcond$i = ($73|0)==($$$i|0);
      if ($exitcond$i) {
       $$2$ph = $$$i;
       break;
      } else {
       $i$02$i = $73;
      }
     }
    } else {
     $$2$ph = $$$i;
    }
   }
   $$3 = $$2$ph;
   STACKTOP = sp;return ($$3|0);
  }
 } while(0);
 _lua_pushnil($L);
 $$3 = 1;
 STACKTOP = sp;return ($$3|0);
}
function _match($ms,$s,$p) {
 $ms = $ms|0;
 $s = $s|0;
 $p = $p|0;
 var $$$i = 0, $$$i$i = 0, $$$i$i$i = 0, $$$i16 = 0, $$$i19 = 0, $$$i26 = 0, $$0$i = 0, $$0$i$i = 0, $$0$i$i37 = 0, $$0$i$i40 = 0, $$0$i13 = 0, $$0$i17 = 0, $$0$i24 = 0, $$0$i28 = 0, $$01$be$i = 0, $$01$be$i$i = 0, $$01$be$i$i$i = 0, $$01$be$i22 = 0, $$01$i = 0, $$01$i$lcssa394 = 0;
 var $$01$i$lcssa396 = 0, $$01$i32 = 0, $$012$i = 0, $$012$i$i = 0, $$012$i$i$i = 0, $$012$i20 = 0, $$02$i$ph = 0, $$02$i125 = 0, $$02$i125$lcssa369 = 0, $$02$i125$lcssa370 = 0, $$02$i125$lcssa371 = 0, $$02$i125372 = 0, $$02$ph = 0, $$02$ph$lcssa = 0, $$02$ph$lcssa402 = 0, $$02$ph$lcssa404 = 0, $$02$ph$lcssa407 = 0, $$02$ph$lcssa409 = 0, $$024$i = 0, $$024$i$lcssa = 0;
 var $$024$i$phi = 0, $$04 = 0, $$04$lcssa = 0, $$04$lcssa379 = 0, $$04$lcssa382 = 0, $$04$lcssa383 = 0, $$04$lcssa385 = 0, $$04$lcssa387 = 0, $$04$ph = 0, $$1$i = 0, $$1$i$lcssa = 0, $$1$i31 = 0, $$1$i31$lcssa = 0, $$2$i = 0, $$2$i30 = 0, $$4 = 0, $$cont$0$i = 0, $$lcssa = 0, $$lcssa373 = 0, $$lcssa374 = 0;
 var $$lcssa375 = 0, $$lcssa389 = 0, $$lcssa390 = 0, $$p$i = 0, $$p$i$i = 0, $$p$i$i$i = 0, $$p$i18 = 0, $$phi$trans$insert$i = 0, $$pr$i = 0, $$pr$i29 = 0, $$pre = 0, $$pre$i = 0, $$pre$i14 = 0, $$pre$i27 = 0, $$pre$i34 = 0, $$pre$phi198$lcssa393Z2D = 0, $$pre$phi198Z2D = 0, $$pre195 = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0;
 var $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0;
 var $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0;
 var $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0;
 var $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0;
 var $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cond = 0, $cont$0$be$i = 0, $cont$05$i = 0, $isdigit = 0, $isdigittmp = 0, $level$0$i$i = 0, $level$0$in$i$i = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer11 = 0;
 var $vararg_buffer13 = 0, $vararg_buffer15 = 0, $vararg_buffer17 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, $vararg_buffer9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer17 = sp + 72|0;
 $vararg_buffer15 = sp + 64|0;
 $vararg_buffer13 = sp + 56|0;
 $vararg_buffer11 = sp + 48|0;
 $vararg_buffer9 = sp + 40|0;
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($ms)) + 4|0);
 $1 = ((($ms)) + 8|0);
 $2 = ((($ms)) + 12|0);
 $$02$ph = $s;$$04$ph = $p;
 L1: while(1) {
  $3 = ((($$02$ph)) + 1|0);
  $4 = ((($$02$ph)) + -1|0);
  $$04 = $$04$ph;
  L3: while(1) {
   $5 = HEAP8[$$04>>0]|0;
   $6 = $5 << 24 >> 24;
   L5: do {
    switch ($6|0) {
    case 40:  {
     $$02$ph$lcssa = $$02$ph;$$04$lcssa = $$04;
     label = 4;
     break L1;
     break;
    }
    case 41:  {
     $$02$ph$lcssa402 = $$02$ph;$$04$lcssa379 = $$04;
     label = 13;
     break L1;
     break;
    }
    case 0:  {
     $$4 = $$02$ph;
     label = 125;
     break L1;
     break;
    }
    case 37:  {
     $46 = ((($$04)) + 1|0);
     $47 = HEAP8[$46>>0]|0;
     $48 = $47 << 24 >> 24;
     switch ($48|0) {
     case 98:  {
      $$04$lcssa382 = $$04;
      label = 20;
      break L3;
      break;
     }
     case 102:  {
      break;
     }
     default: {
      $158 = $47&255;
      $isdigittmp = (($158) + -48)|0;
      $isdigit = ($isdigittmp>>>0)<(10);
      if ($isdigit) {
       $$04$lcssa383 = $$04;$$lcssa389 = $47;$$lcssa390 = $158;
       label = 67;
       break L3;
      }
      $186 = HEAP8[$46>>0]|0;
      $187 = ($186<<24>>24)==(0);
      if ($187) {
       $188 = HEAP32[$1>>2]|0;
       (_luaL_error($188,522689,$vararg_buffer15)|0);
      }
      $189 = ((($$04)) + 2|0);
      $$01$i = $189;$$pre$phi198Z2D = $46;
      break L5;
     }
     }
     $74 = ((($$04)) + 2|0);
     $75 = HEAP8[$74>>0]|0;
     $76 = ($75<<24>>24)==(91);
     if ($76) {
      $80 = 91;
     } else {
      $77 = HEAP32[$1>>2]|0;
      (_luaL_error($77,522654,$vararg_buffer7)|0);
      $$pre = HEAP8[$74>>0]|0;
      $80 = $$pre;
     }
     $78 = ((($$04)) + 3|0);
     $79 = $80 << 24 >> 24;
     switch ($79|0) {
     case 37:  {
      $81 = HEAP8[$78>>0]|0;
      $82 = ($81<<24>>24)==(0);
      if ($82) {
       $83 = HEAP32[$1>>2]|0;
       (_luaL_error($83,522689,$vararg_buffer9)|0);
      }
      $84 = ((($$04)) + 4|0);
      $$01$i32 = $84;
      break;
     }
     case 91:  {
      $85 = HEAP8[$78>>0]|0;
      $86 = ($85<<24>>24)==(94);
      $87 = ((($$04)) + 4|0);
      $$$i26 = $86 ? $87 : $78;
      $$pre$i27 = HEAP8[$$$i26>>0]|0;
      $$0$i28 = $$$i26;$89 = $$pre$i27;
      while(1) {
       $88 = ($89<<24>>24)==(0);
       if ($88) {
        $90 = HEAP32[$1>>2]|0;
        (_luaL_error($90,522724,$vararg_buffer11)|0);
        $$pr$i29 = HEAP8[$$0$i28>>0]|0;
        $93 = $$pr$i29;
       } else {
        $93 = $89;
       }
       $91 = ((($$0$i28)) + 1|0);
       $92 = ($93<<24>>24)==(37);
       if ($92) {
        $94 = HEAP8[$91>>0]|0;
        $95 = ($94<<24>>24)==(0);
        $96 = ((($$0$i28)) + 2|0);
        $$2$i30 = $95 ? $91 : $96;
        $$1$i31 = $$2$i30;
       } else {
        $$1$i31 = $91;
       }
       $97 = HEAP8[$$1$i31>>0]|0;
       $98 = ($97<<24>>24)==(93);
       if ($98) {
        $$1$i31$lcssa = $$1$i31;
        break;
       } else {
        $$0$i28 = $$1$i31;$89 = $97;
       }
      }
      $99 = ((($$1$i31$lcssa)) + 1|0);
      $$01$i32 = $99;
      break;
     }
     default: {
      $$01$i32 = $78;
     }
     }
     $100 = HEAP32[$ms>>2]|0;
     $101 = ($$02$ph|0)==($100|0);
     if ($101) {
      $105 = 0;
     } else {
      $102 = HEAP8[$4>>0]|0;
      $103 = $102 << 24 >> 24;
      $105 = $103;
     }
     $104 = $105 & 255;
     $106 = ((($$01$i32)) + -1|0);
     $107 = HEAP8[$78>>0]|0;
     $108 = ($107<<24>>24)==(94);
     $$p$i18 = $108 ? $78 : $74;
     $109 = $108&1;
     $$$i19 = $109 ^ 1;
     $110 = ((($$p$i18)) + 1|0);
     $111 = ($110>>>0)<($106>>>0);
     L47: do {
      if ($111) {
       $112 = $105&255;
       $$012$i20 = $$p$i18;$114 = $110;
       while(1) {
        $113 = HEAP8[$114>>0]|0;
        $115 = ($113<<24>>24)==(37);
        $116 = ((($$012$i20)) + 2|0);
        $117 = HEAP8[$116>>0]|0;
        do {
         if ($115) {
          $118 = $117&255;
          $119 = (_match_class($104,$118)|0);
          $120 = ($119|0)==(0);
          if ($120) {
           $$01$be$i22 = $116;
          } else {
           $$0$i24 = $$$i19;
           break L47;
          }
         } else {
          $121 = ($117<<24>>24)==(45);
          if ($121) {
           $122 = ((($$012$i20)) + 3|0);
           $123 = ($122>>>0)<($106>>>0);
           if ($123) {
            $124 = $113&255;
            $125 = ($124>>>0)>($104>>>0);
            if ($125) {
             $$01$be$i22 = $122;
             break;
            }
            $126 = HEAP8[$122>>0]|0;
            $127 = $126&255;
            $128 = ($127>>>0)<($104>>>0);
            if ($128) {
             $$01$be$i22 = $122;
             break;
            } else {
             $$0$i24 = $$$i19;
             break L47;
            }
           }
          }
          $129 = ($113<<24>>24)==($112<<24>>24);
          if ($129) {
           $$0$i24 = $$$i19;
           break L47;
          } else {
           $$01$be$i22 = $114;
          }
         }
        } while(0);
        $130 = ((($$01$be$i22)) + 1|0);
        $131 = ($130>>>0)<($106>>>0);
        if ($131) {
         $$012$i20 = $$01$be$i22;$114 = $130;
        } else {
         $$0$i24 = $109;
         break;
        }
       }
      } else {
       $$0$i24 = $109;
      }
     } while(0);
     $132 = ($$0$i24|0)==(0);
     if (!($132)) {
      $$4 = 0;
      label = 125;
      break L1;
     }
     $133 = HEAP8[$$02$ph>>0]|0;
     $134 = $133&255;
     $135 = HEAP8[$78>>0]|0;
     $136 = ($135<<24>>24)==(94);
     $$p$i = $136 ? $78 : $74;
     $137 = $136&1;
     $$$i16 = $137 ^ 1;
     $138 = ((($$p$i)) + 1|0);
     $139 = ($138>>>0)<($106>>>0);
     L62: do {
      if ($139) {
       $$012$i = $$p$i;$141 = $138;
       while(1) {
        $140 = HEAP8[$141>>0]|0;
        $142 = ($140<<24>>24)==(37);
        $143 = ((($$012$i)) + 2|0);
        $144 = HEAP8[$143>>0]|0;
        do {
         if ($142) {
          $145 = $144&255;
          $146 = (_match_class($134,$145)|0);
          $147 = ($146|0)==(0);
          if ($147) {
           $$01$be$i = $143;
          } else {
           $$0$i17 = $$$i16;
           break L62;
          }
         } else {
          $148 = ($144<<24>>24)==(45);
          if ($148) {
           $149 = ((($$012$i)) + 3|0);
           $150 = ($149>>>0)<($106>>>0);
           if ($150) {
            $151 = ($140&255)>($133&255);
            if ($151) {
             $$01$be$i = $149;
             break;
            }
            $152 = HEAP8[$149>>0]|0;
            $153 = ($152&255)<($133&255);
            if ($153) {
             $$01$be$i = $149;
             break;
            } else {
             $$0$i17 = $$$i16;
             break L62;
            }
           }
          }
          $154 = ($140<<24>>24)==($133<<24>>24);
          if ($154) {
           $$0$i17 = $$$i16;
           break L62;
          } else {
           $$01$be$i = $141;
          }
         }
        } while(0);
        $155 = ((($$01$be$i)) + 1|0);
        $156 = ($155>>>0)<($106>>>0);
        if ($156) {
         $$012$i = $$01$be$i;$141 = $155;
        } else {
         $$0$i17 = $137;
         break;
        }
       }
      } else {
       $$0$i17 = $137;
      }
     } while(0);
     $157 = ($$0$i17|0)==(0);
     if ($157) {
      $$4 = 0;
      label = 125;
      break L1;
     } else {
      $$04 = $$01$i32;
      continue L3;
     }
     break;
    }
    case 36:  {
     $180 = ((($$04)) + 1|0);
     $181 = HEAP8[$180>>0]|0;
     $182 = ($181<<24>>24)==(0);
     if ($182) {
      $$02$ph$lcssa404 = $$02$ph;
      label = 75;
      break L1;
     } else {
      $$01$i = $180;$$pre$phi198Z2D = $180;
     }
     break;
    }
    default: {
     $$pre195 = ((($$04)) + 1|0);
     $cond = ($5<<24>>24)==(91);
     if ($cond) {
      $190 = HEAP8[$$pre195>>0]|0;
      $191 = ($190<<24>>24)==(94);
      $192 = ((($$04)) + 2|0);
      $$$i = $191 ? $192 : $$pre195;
      $$pre$i = HEAP8[$$$i>>0]|0;
      $$0$i13 = $$$i;$194 = $$pre$i;
      while(1) {
       $193 = ($194<<24>>24)==(0);
       if ($193) {
        $195 = HEAP32[$1>>2]|0;
        (_luaL_error($195,522724,$vararg_buffer17)|0);
        $$pr$i = HEAP8[$$0$i13>>0]|0;
        $198 = $$pr$i;
       } else {
        $198 = $194;
       }
       $196 = ((($$0$i13)) + 1|0);
       $197 = ($198<<24>>24)==(37);
       if ($197) {
        $199 = HEAP8[$196>>0]|0;
        $200 = ($199<<24>>24)==(0);
        $201 = ((($$0$i13)) + 2|0);
        $$2$i = $200 ? $196 : $201;
        $$1$i = $$2$i;
       } else {
        $$1$i = $196;
       }
       $202 = HEAP8[$$1$i>>0]|0;
       $203 = ($202<<24>>24)==(93);
       if ($203) {
        $$1$i$lcssa = $$1$i;
        break;
       } else {
        $$0$i13 = $$1$i;$194 = $202;
       }
      }
      $204 = ((($$1$i$lcssa)) + 1|0);
      $$01$i = $204;$$pre$phi198Z2D = $$pre195;
     } else {
      $$01$i = $$pre195;$$pre$phi198Z2D = $$pre195;
     }
    }
    }
   } while(0);
   $205 = HEAP32[$0>>2]|0;
   $206 = ($$02$ph>>>0)<($205>>>0);
   if ($206) {
    $207 = HEAP8[$$02$ph>>0]|0;
    $208 = $207&255;
    $209 = HEAP8[$$04>>0]|0;
    $210 = $209 << 24 >> 24;
    L79: do {
     switch ($210|0) {
     case 46:  {
      $$0$i = 1;
      break;
     }
     case 37:  {
      $211 = HEAP8[$$pre$phi198Z2D>>0]|0;
      $212 = $211&255;
      $213 = (_match_class($208,$212)|0);
      $$0$i = $213;
      break;
     }
     case 91:  {
      $214 = ((($$01$i)) + -1|0);
      $215 = HEAP8[$$pre$phi198Z2D>>0]|0;
      $216 = ($215<<24>>24)==(94);
      $$p$i$i = $216 ? $$pre$phi198Z2D : $$04;
      $217 = $216&1;
      $$$i$i = $217 ^ 1;
      $218 = ((($$p$i$i)) + 1|0);
      $219 = ($218>>>0)<($214>>>0);
      if ($219) {
       $$012$i$i = $$p$i$i;$221 = $218;
       while(1) {
        $220 = HEAP8[$221>>0]|0;
        $222 = ($220<<24>>24)==(37);
        $223 = ((($$012$i$i)) + 2|0);
        $224 = HEAP8[$223>>0]|0;
        do {
         if ($222) {
          $225 = $224&255;
          $226 = (_match_class($208,$225)|0);
          $227 = ($226|0)==(0);
          if ($227) {
           $$01$be$i$i = $223;
          } else {
           $$0$i = $$$i$i;
           break L79;
          }
         } else {
          $228 = ($224<<24>>24)==(45);
          if ($228) {
           $229 = ((($$012$i$i)) + 3|0);
           $230 = ($229>>>0)<($214>>>0);
           if ($230) {
            $231 = ($220&255)>($207&255);
            if ($231) {
             $$01$be$i$i = $229;
             break;
            }
            $232 = HEAP8[$229>>0]|0;
            $233 = ($232&255)<($207&255);
            if ($233) {
             $$01$be$i$i = $229;
             break;
            } else {
             $$0$i = $$$i$i;
             break L79;
            }
           }
          }
          $234 = ($220<<24>>24)==($207<<24>>24);
          if ($234) {
           $$0$i = $$$i$i;
           break L79;
          } else {
           $$01$be$i$i = $221;
          }
         }
        } while(0);
        $235 = ((($$01$be$i$i)) + 1|0);
        $236 = ($235>>>0)<($214>>>0);
        if ($236) {
         $$012$i$i = $$01$be$i$i;$221 = $235;
        } else {
         $$0$i = $217;
         break;
        }
       }
      } else {
       $$0$i = $217;
      }
      break;
     }
     default: {
      $237 = ($209<<24>>24)==($207<<24>>24);
      $238 = $237&1;
      $$0$i = $238;
     }
     }
    } while(0);
    $239 = ($$0$i|0)!=(0);
    $290 = $239;
   } else {
    $290 = 0;
   }
   $240 = HEAP8[$$01$i>>0]|0;
   $241 = $240 << 24 >> 24;
   switch ($241|0) {
   case 42:  {
    $$01$i$lcssa394 = $$01$i;$$02$ph$lcssa407 = $$02$ph;$$04$lcssa385 = $$04;
    label = 105;
    break L1;
    break;
   }
   case 43:  {
    label = 106;
    break L1;
    break;
   }
   case 45:  {
    $$01$i$lcssa396 = $$01$i;$$02$ph$lcssa409 = $$02$ph;$$04$lcssa387 = $$04;$$pre$phi198$lcssa393Z2D = $$pre$phi198Z2D;
    label = 102;
    break L1;
    break;
   }
   case 63:  {
    break;
   }
   default: {
    label = 124;
    break L3;
   }
   }
   $244 = ((($$01$i)) + 1|0);
   if (!($290)) {
    $$04 = $244;
    continue;
   }
   $245 = (_match($ms,$3,$244)|0);
   $246 = ($245|0)==(0|0);
   if ($246) {
    $$04 = $244;
   } else {
    $$4 = $245;
    label = 125;
    break L1;
   }
  }
  if ((label|0) == 20) {
   label = 0;
   $49 = ((($$04$lcssa382)) + 2|0);
   $50 = HEAP8[$49>>0]|0;
   $51 = ($50<<24>>24)==(0);
   if ($51) {
    label = 22;
   } else {
    $52 = ((($$04$lcssa382)) + 3|0);
    $53 = HEAP8[$52>>0]|0;
    $54 = ($53<<24>>24)==(0);
    if ($54) {
     label = 22;
    } else {
     $58 = $50;
    }
   }
   if ((label|0) == 22) {
    label = 0;
    $55 = HEAP32[$1>>2]|0;
    (_luaL_error($55,522635,$vararg_buffer5)|0);
    $$pre$i34 = HEAP8[$49>>0]|0;
    $58 = $$pre$i34;
   }
   $56 = HEAP8[$$02$ph>>0]|0;
   $57 = ($56<<24>>24)==($58<<24>>24);
   if (!($57)) {
    $$4 = 0;
    label = 125;
    break;
   }
   $59 = ((($$04$lcssa382)) + 3|0);
   $60 = HEAP8[$59>>0]|0;
   $61 = HEAP32[$0>>2]|0;
   $62 = ($3>>>0)<($61>>>0);
   if ($62) {
    $$024$i = $$02$ph;$64 = $3;$cont$05$i = 1;
   } else {
    $$4 = 0;
    label = 125;
    break;
   }
   while(1) {
    $63 = HEAP8[$64>>0]|0;
    $65 = ($63<<24>>24)==($60<<24>>24);
    if ($65) {
     $66 = (($cont$05$i) + -1)|0;
     $67 = ($66|0)==(0);
     if ($67) {
      $$024$i$lcssa = $$024$i;
      break;
     } else {
      $cont$0$be$i = $66;
     }
    } else {
     $70 = ($63<<24>>24)==($58<<24>>24);
     $71 = $70&1;
     $$cont$0$i = (($71) + ($cont$05$i))|0;
     $cont$0$be$i = $$cont$0$i;
    }
    $68 = ((($64)) + 1|0);
    $69 = ($68>>>0)<($61>>>0);
    if ($69) {
     $$024$i$phi = $64;$64 = $68;$cont$05$i = $cont$0$be$i;$$024$i = $$024$i$phi;
    } else {
     $$4 = 0;
     label = 125;
     break L1;
    }
   }
   $72 = ((($$024$i$lcssa)) + 2|0);
   $73 = ((($$04$lcssa382)) + 4|0);
   $$02$ph = $72;$$04$ph = $73;
   continue;
  }
  else if ((label|0) == 67) {
   label = 0;
   $159 = (($$lcssa390) + -49)|0;
   $160 = ($$lcssa389&255)<(49);
   if ($160) {
    label = 70;
   } else {
    $161 = HEAP32[$2>>2]|0;
    $162 = ($159|0)<($161|0);
    if ($162) {
     $163 = (((((($ms)) + 16|0) + ($159<<3)|0)) + 4|0);
     $164 = HEAP32[$163>>2]|0;
     $165 = ($164|0)==(-1);
     if ($165) {
      label = 70;
     } else {
      $$0$i$i = $159;$172 = $164;
     }
    } else {
     label = 70;
    }
   }
   if ((label|0) == 70) {
    label = 0;
    $166 = HEAP32[$1>>2]|0;
    $167 = (_luaL_error($166,522756,$vararg_buffer13)|0);
    $$phi$trans$insert$i = (((((($ms)) + 16|0) + ($167<<3)|0)) + 4|0);
    $$pre$i14 = HEAP32[$$phi$trans$insert$i>>2]|0;
    $$0$i$i = $167;$172 = $$pre$i14;
   }
   $168 = HEAP32[$0>>2]|0;
   $169 = $$02$ph;
   $170 = (($168) - ($169))|0;
   $171 = ($170>>>0)<($172>>>0);
   if ($171) {
    $$4 = 0;
    label = 125;
    break;
   }
   $173 = (((($ms)) + 16|0) + ($$0$i$i<<3)|0);
   $174 = HEAP32[$173>>2]|0;
   $175 = (_memcmp($174,$$02$ph,$172)|0);
   $176 = ($175|0)!=(0);
   $177 = (($$02$ph) + ($172)|0);
   $178 = ($177|0)==(0|0);
   $or$cond = $178 | $176;
   if ($or$cond) {
    $$4 = 0;
    label = 125;
    break;
   }
   $179 = ((($$04$lcssa383)) + 2|0);
   $$02$ph = $177;$$04$ph = $179;
   continue;
  }
  else if ((label|0) == 124) {
   label = 0;
   if ($290) {
    $$02$ph = $3;$$04$ph = $$01$i;
    continue;
   } else {
    $$4 = 0;
    label = 125;
    break;
   }
  }
 }
 if ((label|0) == 4) {
  $7 = ((($$04$lcssa)) + 1|0);
  $8 = HEAP8[$7>>0]|0;
  $9 = ($8<<24>>24)==(41);
  if ($9) {
   $10 = ((($$04$lcssa)) + 2|0);
   $11 = HEAP32[$2>>2]|0;
   $12 = ($11|0)>(31);
   if ($12) {
    $13 = HEAP32[$1>>2]|0;
    (_luaL_error($13,522593,$vararg_buffer)|0);
   }
   $14 = (((($ms)) + 16|0) + ($11<<3)|0);
   HEAP32[$14>>2] = $$02$ph$lcssa;
   $15 = (((((($ms)) + 16|0) + ($11<<3)|0)) + 4|0);
   HEAP32[$15>>2] = -2;
   $16 = (($11) + 1)|0;
   HEAP32[$2>>2] = $16;
   $17 = (_match($ms,$$02$ph$lcssa,$10)|0);
   $18 = ($17|0)==(0|0);
   if (!($18)) {
    $$4 = $17;
    STACKTOP = sp;return ($$4|0);
   }
   $19 = HEAP32[$2>>2]|0;
   $20 = (($19) + -1)|0;
   HEAP32[$2>>2] = $20;
   $$4 = 0;
   STACKTOP = sp;return ($$4|0);
  } else {
   $21 = HEAP32[$2>>2]|0;
   $22 = ($21|0)>(31);
   if ($22) {
    $23 = HEAP32[$1>>2]|0;
    (_luaL_error($23,522593,$vararg_buffer1)|0);
   }
   $24 = (((($ms)) + 16|0) + ($21<<3)|0);
   HEAP32[$24>>2] = $$02$ph$lcssa;
   $25 = (((((($ms)) + 16|0) + ($21<<3)|0)) + 4|0);
   HEAP32[$25>>2] = -1;
   $26 = (($21) + 1)|0;
   HEAP32[$2>>2] = $26;
   $27 = (_match($ms,$$02$ph$lcssa,$7)|0);
   $28 = ($27|0)==(0|0);
   if (!($28)) {
    $$4 = $27;
    STACKTOP = sp;return ($$4|0);
   }
   $29 = HEAP32[$2>>2]|0;
   $30 = (($29) + -1)|0;
   HEAP32[$2>>2] = $30;
   $$4 = 0;
   STACKTOP = sp;return ($$4|0);
  }
 }
 else if ((label|0) == 13) {
  $31 = ((($$04$lcssa379)) + 1|0);
  $32 = HEAP32[$2>>2]|0;
  $level$0$in$i$i = $32;
  while(1) {
   $level$0$i$i = (($level$0$in$i$i) + -1)|0;
   $33 = ($level$0$in$i$i|0)>(0);
   if (!($33)) {
    label = 16;
    break;
   }
   $34 = (((((($ms)) + 16|0) + ($level$0$i$i<<3)|0)) + 4|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = ($35|0)==(-1);
   if ($36) {
    $$0$i$i37 = $level$0$i$i;
    break;
   } else {
    $level$0$in$i$i = $level$0$i$i;
   }
  }
  if ((label|0) == 16) {
   $37 = HEAP32[$1>>2]|0;
   $38 = (_luaL_error($37,522611,$vararg_buffer3)|0);
   $$0$i$i37 = $38;
  }
  $39 = (((($ms)) + 16|0) + ($$0$i$i37<<3)|0);
  $40 = HEAP32[$39>>2]|0;
  $41 = $$02$ph$lcssa402;
  $42 = (($41) - ($40))|0;
  $43 = (((((($ms)) + 16|0) + ($$0$i$i37<<3)|0)) + 4|0);
  HEAP32[$43>>2] = $42;
  $44 = (_match($ms,$$02$ph$lcssa402,$31)|0);
  $45 = ($44|0)==(0|0);
  if (!($45)) {
   $$4 = $44;
   STACKTOP = sp;return ($$4|0);
  }
  HEAP32[$43>>2] = -1;
  $$4 = 0;
  STACKTOP = sp;return ($$4|0);
 }
 else if ((label|0) == 75) {
  $183 = HEAP32[$0>>2]|0;
  $184 = ($$02$ph$lcssa404|0)==($183|0);
  $185 = $184 ? $$02$ph$lcssa404 : 0;
  $$4 = $185;
  STACKTOP = sp;return ($$4|0);
 }
 else if ((label|0) == 102) {
  $242 = ((($$01$i$lcssa396)) + 1|0);
  $243 = ((($$01$i$lcssa396)) + -1|0);
  $$02$i$ph = $$02$ph$lcssa409;
  L134: while(1) {
   $288 = (_match($ms,$$02$i$ph,$242)|0);
   $289 = ($288|0)==(0|0);
   if ($289) {
    $$02$i125 = $$02$i$ph;
   } else {
    $$4 = $288;
    label = 125;
    break;
   }
   L136: while(1) {
    $249 = HEAP32[$0>>2]|0;
    $250 = ($$02$i125>>>0)<($249>>>0);
    if (!($250)) {
     $$4 = 0;
     label = 125;
     break L134;
    }
    $251 = HEAP8[$$02$i125>>0]|0;
    $252 = HEAP8[$$04$lcssa387>>0]|0;
    $253 = $252 << 24 >> 24;
    switch ($253|0) {
    case 37:  {
     $$02$i125$lcssa370 = $$02$i125;$$lcssa373 = $251;
     label = 111;
     break L136;
     break;
    }
    case 91:  {
     $$02$i125$lcssa371 = $$02$i125;$$lcssa374 = $251;
     label = 112;
     break L136;
     break;
    }
    case 46:  {
     break;
    }
    default: {
     $$02$i125$lcssa369 = $$02$i125;$$lcssa = $251;$$lcssa375 = $252;
     label = 121;
     break L136;
    }
    }
    $254 = ((($$02$i125)) + 1|0);
    $255 = (_match($ms,$254,$242)|0);
    $256 = ($255|0)==(0|0);
    if ($256) {
     $$02$i125 = $254;
    } else {
     $$4 = $255;
     label = 125;
     break L134;
    }
   }
   L140: do {
    if ((label|0) == 111) {
     label = 0;
     $257 = $$lcssa373&255;
     $258 = HEAP8[$$pre$phi198$lcssa393Z2D>>0]|0;
     $259 = $258&255;
     $260 = (_match_class($257,$259)|0);
     $$0$i$i40 = $260;$$02$i125372 = $$02$i125$lcssa370;
    }
    else if ((label|0) == 112) {
     label = 0;
     $261 = $$lcssa374&255;
     $262 = HEAP8[$$pre$phi198$lcssa393Z2D>>0]|0;
     $263 = ($262<<24>>24)==(94);
     $$p$i$i$i = $263 ? $$pre$phi198$lcssa393Z2D : $$04$lcssa387;
     $264 = $263&1;
     $$$i$i$i = $264 ^ 1;
     $265 = ((($$p$i$i$i)) + 1|0);
     $266 = ($265>>>0)<($243>>>0);
     if ($266) {
      $$012$i$i$i = $$p$i$i$i;$268 = $265;
      while(1) {
       $267 = HEAP8[$268>>0]|0;
       $269 = ($267<<24>>24)==(37);
       $270 = ((($$012$i$i$i)) + 2|0);
       $271 = HEAP8[$270>>0]|0;
       do {
        if ($269) {
         $272 = $271&255;
         $273 = (_match_class($261,$272)|0);
         $274 = ($273|0)==(0);
         if ($274) {
          $$01$be$i$i$i = $270;
         } else {
          $$0$i$i40 = $$$i$i$i;$$02$i125372 = $$02$i125$lcssa371;
          break L140;
         }
        } else {
         $275 = ($271<<24>>24)==(45);
         if ($275) {
          $276 = ((($$012$i$i$i)) + 3|0);
          $277 = ($276>>>0)<($243>>>0);
          if ($277) {
           $278 = ($267&255)>($$lcssa374&255);
           if ($278) {
            $$01$be$i$i$i = $276;
            break;
           }
           $279 = HEAP8[$276>>0]|0;
           $280 = ($279&255)<($$lcssa374&255);
           if ($280) {
            $$01$be$i$i$i = $276;
            break;
           } else {
            $$0$i$i40 = $$$i$i$i;$$02$i125372 = $$02$i125$lcssa371;
            break L140;
           }
          }
         }
         $281 = ($267<<24>>24)==($$lcssa374<<24>>24);
         if ($281) {
          $$0$i$i40 = $$$i$i$i;$$02$i125372 = $$02$i125$lcssa371;
          break L140;
         } else {
          $$01$be$i$i$i = $268;
         }
        }
       } while(0);
       $282 = ((($$01$be$i$i$i)) + 1|0);
       $283 = ($282>>>0)<($243>>>0);
       if ($283) {
        $$012$i$i$i = $$01$be$i$i$i;$268 = $282;
       } else {
        $$0$i$i40 = $264;$$02$i125372 = $$02$i125$lcssa371;
        break;
       }
      }
     } else {
      $$0$i$i40 = $264;$$02$i125372 = $$02$i125$lcssa371;
     }
    }
    else if ((label|0) == 121) {
     label = 0;
     $284 = ($$lcssa375<<24>>24)==($$lcssa<<24>>24);
     $285 = $284&1;
     $$0$i$i40 = $285;$$02$i125372 = $$02$i125$lcssa369;
    }
   } while(0);
   $286 = ($$0$i$i40|0)==(0);
   $287 = ((($$02$i125372)) + 1|0);
   if ($286) {
    $$4 = 0;
    label = 125;
    break;
   } else {
    $$02$i$ph = $287;
   }
  }
  if ((label|0) == 125) {
   STACKTOP = sp;return ($$4|0);
  }
 }
 else if ((label|0) == 105) {
  $247 = (_max_expand($ms,$$02$ph$lcssa407,$$04$lcssa385,$$01$i$lcssa394)|0);
  $$4 = $247;
  STACKTOP = sp;return ($$4|0);
 }
 else if ((label|0) == 106) {
  if (!($290)) {
   $$4 = 0;
   STACKTOP = sp;return ($$4|0);
  }
  $248 = (_max_expand($ms,$3,$$04,$$01$i)|0);
  $$4 = $248;
  STACKTOP = sp;return ($$4|0);
 }
 else if ((label|0) == 125) {
  STACKTOP = sp;return ($$4|0);
 }
 return (0)|0;
}
function _match_class($c,$cl) {
 $c = $c|0;
 $cl = $cl|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $isdigit = 0, $isdigittmp = 0, $res$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_tolower($cl)|0);
 do {
  switch ($0|0) {
  case 97:  {
   $1 = (_isalpha($c)|0);
   $res$0 = $1;
   break;
  }
  case 99:  {
   $2 = (_iscntrl($c)|0);
   $res$0 = $2;
   break;
  }
  case 100:  {
   $isdigittmp = (($c) + -48)|0;
   $isdigit = ($isdigittmp>>>0)<(10);
   $3 = $isdigit&1;
   $res$0 = $3;
   break;
  }
  case 108:  {
   $4 = (_islower($c)|0);
   $res$0 = $4;
   break;
  }
  case 112:  {
   $5 = (_ispunct($c)|0);
   $res$0 = $5;
   break;
  }
  case 115:  {
   $6 = (_isspace($c)|0);
   $res$0 = $6;
   break;
  }
  case 117:  {
   $7 = (_isupper($c)|0);
   $res$0 = $7;
   break;
  }
  case 119:  {
   $8 = (_isalnum($c)|0);
   $res$0 = $8;
   break;
  }
  case 120:  {
   $9 = (_isxdigit($c)|0);
   $res$0 = $9;
   break;
  }
  case 122:  {
   $10 = ($c|0)==(0);
   $11 = $10&1;
   $res$0 = $11;
   break;
  }
  default: {
   $12 = ($cl|0)==($c|0);
   $13 = $12&1;
   $$0 = $13;
   return ($$0|0);
  }
  }
 } while(0);
 $14 = (_islower($cl)|0);
 $15 = ($14|0)==(0);
 $16 = ($res$0|0)==(0);
 $17 = $16&1;
 $18 = $15 ? $17 : $res$0;
 $$0 = $18;
 return ($$0|0);
}
function _max_expand($ms,$s,$p,$ep) {
 $ms = $ms|0;
 $s = $s|0;
 $p = $p|0;
 $ep = $ep|0;
 var $$$i$i = 0, $$0$i = 0, $$01$be$i$i = 0, $$012$i$i = 0, $$2 = 0, $$p$i$i = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0$lcssa = 0, $i$02 = 0, $i$1 = 0, $i$2 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($ms)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1>>>0)>($s>>>0);
 L1: do {
  if ($2) {
   $3 = ((($p)) + 1|0);
   $4 = ((($ep)) + -1|0);
   $48 = $1;$6 = $s;$i$02 = 0;
   while(1) {
    $5 = HEAP8[$6>>0]|0;
    $7 = $5&255;
    $8 = HEAP8[$p>>0]|0;
    $9 = $8 << 24 >> 24;
    L5: do {
     switch ($9|0) {
     case 46:  {
      $42 = $48;
      break;
     }
     case 37:  {
      $10 = HEAP8[$3>>0]|0;
      $11 = $10&255;
      $12 = (_match_class($7,$11)|0);
      $$0$i = $12;
      label = 15;
      break;
     }
     case 91:  {
      $13 = HEAP8[$3>>0]|0;
      $14 = ($13<<24>>24)==(94);
      $$p$i$i = $14 ? $3 : $p;
      $15 = $14&1;
      $$$i$i = $15 ^ 1;
      $16 = ((($$p$i$i)) + 1|0);
      $17 = ($16>>>0)<($4>>>0);
      if ($17) {
       $$012$i$i = $$p$i$i;$19 = $16;
       while(1) {
        $18 = HEAP8[$19>>0]|0;
        $20 = ($18<<24>>24)==(37);
        $21 = ((($$012$i$i)) + 2|0);
        $22 = HEAP8[$21>>0]|0;
        do {
         if ($20) {
          $23 = $22&255;
          $24 = (_match_class($7,$23)|0);
          $25 = ($24|0)==(0);
          if ($25) {
           $$01$be$i$i = $21;
          } else {
           $$0$i = $$$i$i;
           label = 15;
           break L5;
          }
         } else {
          $26 = ($22<<24>>24)==(45);
          if ($26) {
           $27 = ((($$012$i$i)) + 3|0);
           $28 = ($27>>>0)<($4>>>0);
           if ($28) {
            $29 = ($18&255)>($5&255);
            if ($29) {
             $$01$be$i$i = $27;
             break;
            }
            $30 = HEAP8[$27>>0]|0;
            $31 = ($30&255)<($5&255);
            if ($31) {
             $$01$be$i$i = $27;
             break;
            } else {
             $$0$i = $$$i$i;
             label = 15;
             break L5;
            }
           }
          }
          $32 = ($18<<24>>24)==($5<<24>>24);
          if ($32) {
           $$0$i = $$$i$i;
           label = 15;
           break L5;
          } else {
           $$01$be$i$i = $19;
          }
         }
        } while(0);
        $33 = ((($$01$be$i$i)) + 1|0);
        $34 = ($33>>>0)<($4>>>0);
        if ($34) {
         $$012$i$i = $$01$be$i$i;$19 = $33;
        } else {
         $$0$i = $15;
         label = 15;
         break;
        }
       }
      } else {
       $$0$i = $15;
       label = 15;
      }
      break;
     }
     default: {
      $35 = ($8<<24>>24)==($5<<24>>24);
      $36 = $35&1;
      $$0$i = $36;
      label = 15;
     }
     }
    } while(0);
    if ((label|0) == 15) {
     label = 0;
     $37 = ($$0$i|0)==(0);
     if ($37) {
      $i$0$lcssa = $i$02;
      break L1;
     }
     $$pre = HEAP32[$0>>2]|0;
     $42 = $$pre;
    }
    $39 = (($i$02) + 1)|0;
    $40 = (($s) + ($39)|0);
    $41 = ($40>>>0)<($42>>>0);
    if ($41) {
     $48 = $42;$6 = $40;$i$02 = $39;
    } else {
     $i$0$lcssa = $39;
     break L1;
    }
   }
  } else {
   $i$0$lcssa = 0;
  }
 } while(0);
 $38 = ((($ep)) + 1|0);
 $i$1 = $i$0$lcssa;
 while(1) {
  $43 = ($i$1|0)>(-1);
  if (!($43)) {
   $$2 = 0;
   label = 21;
   break;
  }
  $44 = (($s) + ($i$1)|0);
  $45 = (_match($ms,$44,$38)|0);
  $46 = ($45|0)==(0|0);
  $47 = $46 << 31 >> 31;
  $i$2 = (($47) + ($i$1))|0;
  if ($46) {
   $i$1 = $i$2;
  } else {
   $$2 = $45;
   label = 21;
   break;
  }
 }
 if ((label|0) == 21) {
  return ($$2|0);
 }
 return (0)|0;
}
function _push_onecapture($ms,$i,$s,$e) {
 $ms = $ms|0;
 $i = $i|0;
 $s = $s|0;
 $e = $e|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($ms)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>($i|0);
 if (!($2)) {
  $3 = ($i|0)==(0);
  $4 = ((($ms)) + 8|0);
  $5 = HEAP32[$4>>2]|0;
  if ($3) {
   $6 = $e;
   $7 = $s;
   $8 = (($6) - ($7))|0;
   _lua_pushlstring($5,$s,$8);
   STACKTOP = sp;return;
  } else {
   (_luaL_error($5,522756,$vararg_buffer)|0);
   STACKTOP = sp;return;
  }
 }
 $9 = (((((($ms)) + 16|0) + ($i<<3)|0)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10|0)==(-1);
 if ($11) {
  $12 = ((($ms)) + 8|0);
  $13 = HEAP32[$12>>2]|0;
  (_luaL_error($13,522778,$vararg_buffer1)|0);
  $14 = HEAP32[$12>>2]|0;
  $15 = (((($ms)) + 16|0) + ($i<<3)|0);
  $16 = HEAP32[$15>>2]|0;
  $26 = $14;$27 = $16;
 } else {
  $17 = ($10|0)==(-2);
  $18 = ((($ms)) + 8|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = (((($ms)) + 16|0) + ($i<<3)|0);
  $21 = HEAP32[$20>>2]|0;
  if ($17) {
   $22 = HEAP32[$ms>>2]|0;
   $23 = $21;
   $24 = (($23) + 1)|0;
   $25 = (($24) - ($22))|0;
   _lua_pushinteger($19,$25);
   STACKTOP = sp;return;
  } else {
   $26 = $19;$27 = $21;
  }
 }
 _lua_pushlstring($26,$27,$10);
 STACKTOP = sp;return;
}
function _str_format($L) {
 $L = $L|0;
 var $$$$i = 0, $$$i = 0, $$2 = 0, $$lcssa = 0, $$lcssa33 = 0, $$lcssa34 = 0, $$p$0$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i4 = 0, $$pre12$i = 0, $$pre17 = 0, $$pre18 = 0, $$pre19 = 0, $$pre20 = 0, $$pre3$i = 0, $$pre4$i = 0, $$pre5$i = 0, $$pre6$i = 0, $$pre7$i = 0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0.0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0.0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $arg$0$ph13 = 0, $b = 0, $buff = 0, $form = 0, $isdigit$i = 0, $isdigit2$i = 0, $isdigit4$i = 0, $isdigit6$i = 0, $isdigit8$i = 0, $isdigittmp$i = 0, $isdigittmp1$i = 0, $isdigittmp3$i = 0, $isdigittmp5$i = 0;
 var $isdigittmp7$i = 0, $l = 0, $l$i = 0, $memchr$i = 0, $or$cond = 0, $p$0$lcssa$i = 0, $p$09$i = 0, $p$2$i = 0, $p$4$i = 0, $s$01$i = 0, $sfl = 0, $strfrmt$0$be = 0, $strfrmt$0$ph12 = 0, $strfrmt$011 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer11 = 0, $vararg_buffer14 = 0, $vararg_buffer17 = 0, $vararg_buffer3 = 0;
 var $vararg_buffer5 = 0, $vararg_buffer8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1648|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer17 = sp + 56|0;
 $vararg_buffer14 = sp + 48|0;
 $vararg_buffer11 = sp + 40|0;
 $vararg_buffer8 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $l$i = sp + 1104|0;
 $sfl = sp + 1100|0;
 $b = sp + 64|0;
 $form = sp + 1624|0;
 $buff = sp + 1112|0;
 $l = sp + 60|0;
 $0 = (_lua_gettop($L)|0);
 $1 = (_luaL_checklstring($L,1,$sfl)|0);
 $2 = HEAP32[$sfl>>2]|0;
 $3 = (($1) + ($2)|0);
 _luaL_buffinit($L,$b);
 $4 = ($2|0)>(0);
 L1: do {
  if ($4) {
   $5 = ((($b)) + 1036|0);
   $6 = ((($form)) + 1|0);
   $arg$0$ph13 = 1;$strfrmt$0$ph12 = $1;
   L3: while(1) {
    $strfrmt$011 = $strfrmt$0$ph12;
    while(1) {
     $7 = HEAP8[$strfrmt$011>>0]|0;
     $8 = ($7<<24>>24)==(37);
     if ($8) {
      $15 = ((($strfrmt$011)) + 1|0);
      $16 = HEAP8[$15>>0]|0;
      $17 = ($16<<24>>24)==(37);
      if (!($17)) {
       $$lcssa = $15;$$lcssa33 = $16;
       break;
      }
      $18 = HEAP32[$b>>2]|0;
      $19 = ($18>>>0)<($5>>>0);
      if ($19) {
       $22 = $18;$23 = 37;
      } else {
       (_luaL_prepbuffer($b)|0);
       $$pre19 = HEAP8[$15>>0]|0;
       $$pre20 = HEAP32[$b>>2]|0;
       $22 = $$pre20;$23 = $$pre19;
      }
      $20 = ((($strfrmt$011)) + 2|0);
      $21 = ((($22)) + 1|0);
      HEAP32[$b>>2] = $21;
      HEAP8[$22>>0] = $23;
      $strfrmt$0$be = $20;
     } else {
      $9 = HEAP32[$b>>2]|0;
      $10 = ($9>>>0)<($5>>>0);
      if ($10) {
       $13 = $9;$14 = $7;
      } else {
       (_luaL_prepbuffer($b)|0);
       $$pre = HEAP8[$strfrmt$011>>0]|0;
       $$pre17 = HEAP32[$b>>2]|0;
       $13 = $$pre17;$14 = $$pre;
      }
      $11 = ((($strfrmt$011)) + 1|0);
      $12 = ((($13)) + 1|0);
      HEAP32[$b>>2] = $12;
      HEAP8[$13>>0] = $14;
      $strfrmt$0$be = $11;
     }
     $24 = ($strfrmt$0$be>>>0)<($3>>>0);
     if ($24) {
      $strfrmt$011 = $strfrmt$0$be;
     } else {
      break L1;
     }
    }
    $25 = (($arg$0$ph13) + 1)|0;
    $26 = ($arg$0$ph13|0)<($0|0);
    if ($26) {
     $28 = $$lcssa33;
    } else {
     (_luaL_argerror($L,$25,522797)|0);
     $$pre18 = HEAP8[$$lcssa>>0]|0;
     $28 = $$pre18;
    }
    $27 = ($28<<24>>24)==(0);
    L22: do {
     if ($27) {
      $126 = 0;$p$0$lcssa$i = $$lcssa;
     } else {
      $30 = $28;$p$09$i = $$lcssa;
      while(1) {
       $29 = $30 << 24 >> 24;
       $memchr$i = (_memchr(522806,$29,6)|0);
       $31 = ($memchr$i|0)==(0|0);
       if ($31) {
        $126 = $30;$p$0$lcssa$i = $p$09$i;
        break L22;
       }
       $32 = ((($p$09$i)) + 1|0);
       $33 = HEAP8[$32>>0]|0;
       $34 = ($33<<24>>24)==(0);
       if ($34) {
        $126 = 0;$p$0$lcssa$i = $32;
        break;
       } else {
        $30 = $33;$p$09$i = $32;
       }
      }
     }
    } while(0);
    $35 = $p$0$lcssa$i;
    $36 = $$lcssa;
    $37 = (($35) - ($36))|0;
    $38 = ($37>>>0)>(5);
    if ($38) {
     (_luaL_error($L,522812,$vararg_buffer)|0);
     $$pre$i = HEAP8[$p$0$lcssa$i>>0]|0;
     $40 = $$pre$i;
    } else {
     $40 = $126;
    }
    $39 = $40&255;
    $isdigittmp$i = (($39) + -48)|0;
    $isdigit$i = ($isdigittmp$i>>>0)<(10);
    $41 = ((($p$0$lcssa$i)) + 1|0);
    $$p$0$i = $isdigit$i ? $41 : $p$0$lcssa$i;
    $42 = HEAP8[$$p$0$i>>0]|0;
    $43 = $42&255;
    $isdigittmp1$i = (($43) + -48)|0;
    $isdigit2$i = ($isdigittmp1$i>>>0)<(10);
    $44 = ((($$p$0$i)) + 1|0);
    $p$2$i = $isdigit2$i ? $44 : $$p$0$i;
    $45 = HEAP8[$p$2$i>>0]|0;
    $46 = ($45<<24>>24)==(46);
    if ($46) {
     $47 = ((($p$2$i)) + 1|0);
     $48 = HEAP8[$47>>0]|0;
     $49 = $48&255;
     $isdigittmp5$i = (($49) + -48)|0;
     $isdigit6$i = ($isdigittmp5$i>>>0)<(10);
     $50 = ((($p$2$i)) + 2|0);
     $$$i = $isdigit6$i ? $50 : $47;
     $51 = HEAP8[$$$i>>0]|0;
     $52 = $51&255;
     $isdigittmp7$i = (($52) + -48)|0;
     $isdigit8$i = ($isdigittmp7$i>>>0)<(10);
     $53 = ((($$$i)) + 1|0);
     $$$$i = $isdigit8$i ? $53 : $$$i;
     $$pre12$i = HEAP8[$$$$i>>0]|0;
     $55 = $$pre12$i;$p$4$i = $$$$i;
    } else {
     $55 = $45;$p$4$i = $p$2$i;
    }
    $54 = $55&255;
    $isdigittmp3$i = (($54) + -48)|0;
    $isdigit4$i = ($isdigittmp3$i>>>0)<(10);
    if ($isdigit4$i) {
     (_luaL_error($L,522844,$vararg_buffer1)|0);
    }
    HEAP8[$form>>0] = 37;
    $56 = $p$4$i;
    $57 = (($56) - ($36))|0;
    $58 = (($57) + 1)|0;
    (_strncpy($6,$$lcssa,$58)|0);
    $59 = (($6) + ($58)|0);
    HEAP8[$59>>0] = 0;
    $60 = ((($p$4$i)) + 1|0);
    $61 = HEAP8[$p$4$i>>0]|0;
    $62 = $61 << 24 >> 24;
    L36: do {
     switch ($62|0) {
     case 99:  {
      $63 = (+_luaL_checknumber($L,$25));
      $64 = (~~(($63)));
      HEAP32[$vararg_buffer3>>2] = $64;
      (_sprintf($buff,$form,$vararg_buffer3)|0);
      label = 50;
      break;
     }
     case 105: case 100:  {
      $65 = (_strlen($form)|0);
      $66 = (($65) + -1)|0;
      $67 = (($form) + ($66)|0);
      $68 = HEAP8[$67>>0]|0;
      $69 = (($form) + ($65)|0);
      $70 = ((($69)) + -1|0);
      HEAP8[$70>>0]=108&255;HEAP8[$70+1>>0]=108>>8;
      HEAP8[$69>>0] = $68;
      $71 = (($65) + 1)|0;
      $72 = (($form) + ($71)|0);
      HEAP8[$72>>0] = 0;
      $73 = (+_luaL_checknumber($L,$25));
      $74 = (~~(($73)));
      HEAP32[$vararg_buffer5>>2] = $74;
      (_sprintf($buff,$form,$vararg_buffer5)|0);
      label = 50;
      break;
     }
     case 88: case 120: case 117: case 111:  {
      $75 = (_strlen($form)|0);
      $76 = (($75) + -1)|0;
      $77 = (($form) + ($76)|0);
      $78 = HEAP8[$77>>0]|0;
      $79 = (($form) + ($75)|0);
      $80 = ((($79)) + -1|0);
      HEAP8[$80>>0]=108&255;HEAP8[$80+1>>0]=108>>8;
      HEAP8[$79>>0] = $78;
      $81 = (($75) + 1)|0;
      $82 = (($form) + ($81)|0);
      HEAP8[$82>>0] = 0;
      $83 = (+_luaL_checknumber($L,$25));
      $84 = (~~(($83))>>>0);
      HEAP32[$vararg_buffer8>>2] = $84;
      (_sprintf($buff,$form,$vararg_buffer8)|0);
      label = 50;
      break;
     }
     case 71: case 103: case 102: case 69: case 101:  {
      $85 = (+_luaL_checknumber($L,$25));
      HEAPF64[$vararg_buffer11>>3] = $85;
      (_sprintf($buff,$form,$vararg_buffer11)|0);
      label = 50;
      break;
     }
     case 113:  {
      $86 = (_luaL_checklstring($L,$25,$l$i)|0);
      $87 = HEAP32[$b>>2]|0;
      $88 = ($87>>>0)<($5>>>0);
      if ($88) {
       $90 = $87;
      } else {
       (_luaL_prepbuffer($b)|0);
       $$pre$i4 = HEAP32[$b>>2]|0;
       $90 = $$pre$i4;
      }
      $89 = ((($90)) + 1|0);
      HEAP32[$b>>2] = $89;
      HEAP8[$90>>0] = 34;
      $91 = HEAP32[$l$i>>2]|0;
      $92 = (($91) + -1)|0;
      HEAP32[$l$i>>2] = $92;
      $93 = ($91|0)==(0);
      L45: do {
       if (!($93)) {
        $s$01$i = $86;
        while(1) {
         $94 = HEAP8[$s$01$i>>0]|0;
         $95 = $94 << 24 >> 24;
         switch ($95|0) {
         case 10: case 92: case 34:  {
          $96 = HEAP32[$b>>2]|0;
          $97 = ($96>>>0)<($5>>>0);
          if ($97) {
           $99 = $96;
          } else {
           (_luaL_prepbuffer($b)|0);
           $$pre3$i = HEAP32[$b>>2]|0;
           $99 = $$pre3$i;
          }
          $98 = ((($99)) + 1|0);
          HEAP32[$b>>2] = $98;
          HEAP8[$99>>0] = 92;
          $100 = HEAP32[$b>>2]|0;
          $101 = ($100>>>0)<($5>>>0);
          if ($101) {
           $104 = $100;
          } else {
           (_luaL_prepbuffer($b)|0);
           $$pre4$i = HEAP32[$b>>2]|0;
           $104 = $$pre4$i;
          }
          $102 = HEAP8[$s$01$i>>0]|0;
          $103 = ((($104)) + 1|0);
          HEAP32[$b>>2] = $103;
          HEAP8[$104>>0] = $102;
          break;
         }
         case 13:  {
          _luaL_addlstring($b,522889,2);
          break;
         }
         case 0:  {
          _luaL_addlstring($b,522892,4);
          break;
         }
         default: {
          $105 = HEAP32[$b>>2]|0;
          $106 = ($105>>>0)<($5>>>0);
          if ($106) {
           $108 = $105;$109 = $94;
          } else {
           (_luaL_prepbuffer($b)|0);
           $$pre5$i = HEAP8[$s$01$i>>0]|0;
           $$pre6$i = HEAP32[$b>>2]|0;
           $108 = $$pre6$i;$109 = $$pre5$i;
          }
          $107 = ((($108)) + 1|0);
          HEAP32[$b>>2] = $107;
          HEAP8[$108>>0] = $109;
         }
         }
         $110 = ((($s$01$i)) + 1|0);
         $111 = HEAP32[$l$i>>2]|0;
         $112 = (($111) + -1)|0;
         HEAP32[$l$i>>2] = $112;
         $113 = ($111|0)==(0);
         if ($113) {
          break L45;
         } else {
          $s$01$i = $110;
         }
        }
       }
      } while(0);
      $114 = HEAP32[$b>>2]|0;
      $115 = ($114>>>0)<($5>>>0);
      if ($115) {
       $117 = $114;
      } else {
       (_luaL_prepbuffer($b)|0);
       $$pre7$i = HEAP32[$b>>2]|0;
       $117 = $$pre7$i;
      }
      $116 = ((($117)) + 1|0);
      HEAP32[$b>>2] = $116;
      HEAP8[$117>>0] = 34;
      break;
     }
     case 115:  {
      $118 = (_luaL_checklstring($L,$25,$l)|0);
      $119 = (_strchr($form,46)|0);
      $120 = ($119|0)==(0|0);
      $121 = HEAP32[$l>>2]|0;
      $122 = ($121>>>0)>(99);
      $or$cond = $120 & $122;
      if ($or$cond) {
       _lua_pushvalue($L,$25);
       _luaL_addvalue($b);
       break L36;
      } else {
       HEAP32[$vararg_buffer14>>2] = $118;
       (_sprintf($buff,$form,$vararg_buffer14)|0);
       label = 50;
       break L36;
      }
      break;
     }
     default: {
      $$lcssa34 = $62;
      break L3;
     }
     }
    } while(0);
    if ((label|0) == 50) {
     label = 0;
     $123 = (_strlen($buff)|0);
     _luaL_addlstring($b,$buff,$123);
    }
    $124 = ($60>>>0)<($3>>>0);
    if ($124) {
     $arg$0$ph13 = $25;$strfrmt$0$ph12 = $60;
    } else {
     break L1;
    }
   }
   HEAP32[$vararg_buffer17>>2] = $$lcssa34;
   $125 = (_luaL_error($L,522897,$vararg_buffer17)|0);
   $$2 = $125;
   STACKTOP = sp;return ($$2|0);
  }
 } while(0);
 _luaL_pushresult($b);
 $$2 = 1;
 STACKTOP = sp;return ($$2|0);
}
function _gfind_nodef($L) {
 $L = $L|0;
 var $0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_luaL_error($L,522937,$vararg_buffer)|0);
 STACKTOP = sp;return ($0|0);
}
function _gmatch($L) {
 $L = $L|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 (_luaL_checklstring($L,1,0)|0);
 (_luaL_checklstring($L,2,0)|0);
 _lua_settop($L,2);
 _lua_pushinteger($L,0);
 _lua_pushcclosure($L,4320,3);
 return 1;
}
function _gmatch_aux($L) {
 $L = $L|0;
 var $$ = 0, $$$i = 0, $$2 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i = 0, $i$02$i = 0, $ls = 0, $ms = 0, $or$cond$i = 0, $src$05 = 0, $src$05$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 288|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ms = sp + 8|0;
 $ls = sp;
 $0 = (_lua_tolstring($L,-10003,$ls)|0);
 $1 = (_lua_tolstring($L,-10004,0)|0);
 $2 = ((($ms)) + 8|0);
 HEAP32[$2>>2] = $L;
 HEAP32[$ms>>2] = $0;
 $3 = HEAP32[$ls>>2]|0;
 $4 = (($0) + ($3)|0);
 $5 = ((($ms)) + 4|0);
 HEAP32[$5>>2] = $4;
 $6 = (_lua_tointeger($L,-10005)|0);
 $7 = ($6|0)>($3|0);
 if ($7) {
  $$2 = 0;
  STACKTOP = sp;return ($$2|0);
 }
 $8 = (($0) + ($6)|0);
 $9 = ((($ms)) + 12|0);
 $src$05 = $8;
 while(1) {
  HEAP32[$9>>2] = 0;
  $10 = (_match($ms,$src$05,$1)|0);
  $11 = ($10|0)==(0|0);
  if (!($11)) {
   $$lcssa = $10;$src$05$lcssa = $src$05;
   break;
  }
  $23 = ((($src$05)) + 1|0);
  $24 = HEAP32[$5>>2]|0;
  $25 = ($23>>>0)>($24>>>0);
  if ($25) {
   $$2 = 0;
   label = 7;
   break;
  } else {
   $src$05 = $23;
  }
 }
 if ((label|0) == 7) {
  STACKTOP = sp;return ($$2|0);
 }
 $12 = $$lcssa;
 $13 = $0;
 $14 = (($12) - ($13))|0;
 $15 = ($$lcssa|0)==($src$05$lcssa|0);
 $16 = $15&1;
 $$ = (($14) + ($16))|0;
 _lua_pushinteger($L,$$);
 _lua_replace($L,-10005);
 $17 = HEAP32[$9>>2]|0;
 $18 = ($17|0)==(0);
 $19 = ($src$05$lcssa|0)!=(0|0);
 $or$cond$i = $19 & $18;
 $$$i = $or$cond$i ? 1 : $17;
 $20 = HEAP32[$2>>2]|0;
 _luaL_checkstack($20,$$$i,522593);
 $21 = ($$$i|0)>(0);
 if ($21) {
  $i$02$i = 0;
 } else {
  $$2 = $$$i;
  STACKTOP = sp;return ($$2|0);
 }
 while(1) {
  _push_onecapture($ms,$i$02$i,$src$05$lcssa,$$lcssa);
  $22 = (($i$02$i) + 1)|0;
  $exitcond$i = ($22|0)==($$$i|0);
  if ($exitcond$i) {
   $$2 = $$$i;
   break;
  } else {
   $i$02$i = $22;
  }
 }
 STACKTOP = sp;return ($$2|0);
}
function _str_gsub($L) {
 $L = $L|0;
 var $$ = 0, $$$i$i = 0, $$off = 0, $$pre = 0, $$pre$i$i = 0, $$pre2$i$i = 0, $$pre3$i$i = 0, $$pre4$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $b = 0, $exitcond$i$i = 0, $i$01$i$i = 0, $i$02$i$i = 0, $i$1$i$i = 0, $isdigit$i$i = 0, $isdigittmp$i$i = 0, $l$i$i = 0, $ms = 0, $n$0 = 0, $n$110 = 0;
 var $n$111 = 0, $n$2 = 0, $or$cond$i$i = 0, $src$0 = 0, $src$1 = 0, $src$3 = 0, $srcl = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1328|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $l$i$i = sp + 1324|0;
 $srcl = sp + 1320|0;
 $ms = sp + 1048|0;
 $b = sp + 12|0;
 $0 = (_luaL_checklstring($L,1,$srcl)|0);
 $1 = (_luaL_checklstring($L,2,0)|0);
 $2 = (_lua_type($L,3)|0);
 $3 = HEAP32[$srcl>>2]|0;
 $4 = (($3) + 1)|0;
 $5 = (_luaL_optinteger($L,4,$4)|0);
 $6 = HEAP8[$1>>0]|0;
 $7 = ($6<<24>>24)==(94);
 $8 = ((($1)) + 1|0);
 $$ = $7 ? $8 : $1;
 $$off = (($2) + -3)|0;
 $9 = ($$off>>>0)<(4);
 if (!($9)) {
  (_luaL_argerror($L,3,522995)|0);
 }
 _luaL_buffinit($L,$b);
 $10 = ((($ms)) + 8|0);
 HEAP32[$10>>2] = $L;
 HEAP32[$ms>>2] = $0;
 $11 = HEAP32[$srcl>>2]|0;
 $12 = (($0) + ($11)|0);
 $13 = ((($ms)) + 4|0);
 HEAP32[$13>>2] = $12;
 $14 = ((($ms)) + 12|0);
 $15 = ((($b)) + 1036|0);
 $16 = ((($ms)) + 20|0);
 $17 = ((($ms)) + 16|0);
 $n$0 = 0;$src$0 = $0;
 while(1) {
  $18 = ($n$0|0)<($5|0);
  if (!($18)) {
   $n$2 = $n$0;$src$3 = $src$0;
   label = 44;
   break;
  }
  HEAP32[$14>>2] = 0;
  $19 = (_match($ms,$src$0,$$)|0);
  $20 = ($19|0)==(0|0);
  if ($20) {
   $n$110 = $n$0;
   label = 39;
  } else {
   $21 = (($n$0) + 1)|0;
   $22 = HEAP32[$10>>2]|0;
   $23 = (_lua_type($22,3)|0);
   switch ($23|0) {
   case 4: case 3:  {
    $24 = HEAP32[$10>>2]|0;
    $25 = (_lua_tolstring($24,3,$l$i$i)|0);
    $26 = HEAP32[$l$i$i>>2]|0;
    $27 = ($26|0)==(0);
    if (!($27)) {
     $28 = $19;
     $29 = $src$0;
     $30 = (($28) - ($29))|0;
     $i$01$i$i = 0;
     while(1) {
      $31 = (($25) + ($i$01$i$i)|0);
      $32 = HEAP8[$31>>0]|0;
      $33 = ($32<<24>>24)==(37);
      do {
       if ($33) {
        $39 = (($i$01$i$i) + 1)|0;
        $40 = (($25) + ($39)|0);
        $41 = HEAP8[$40>>0]|0;
        $42 = $41&255;
        $isdigittmp$i$i = (($42) + -48)|0;
        $isdigit$i$i = ($isdigittmp$i$i>>>0)<(10);
        if ($isdigit$i$i) {
         $48 = ($41<<24>>24)==(48);
         if ($48) {
          _luaL_addlstring($b,$src$0,$30);
          $i$1$i$i = $39;
          break;
         } else {
          $49 = $41 << 24 >> 24;
          $50 = (($49) + -49)|0;
          _push_onecapture($ms,$50,$src$0,$19);
          _luaL_addvalue($b);
          $i$1$i$i = $39;
          break;
         }
        } else {
         $43 = HEAP32[$b>>2]|0;
         $44 = ($43>>>0)<($15>>>0);
         if ($44) {
          $46 = $43;$47 = $41;
         } else {
          (_luaL_prepbuffer($b)|0);
          $$pre3$i$i = HEAP8[$40>>0]|0;
          $$pre4$i$i = HEAP32[$b>>2]|0;
          $46 = $$pre4$i$i;$47 = $$pre3$i$i;
         }
         $45 = ((($46)) + 1|0);
         HEAP32[$b>>2] = $45;
         HEAP8[$46>>0] = $47;
         $i$1$i$i = $39;
         break;
        }
       } else {
        $34 = HEAP32[$b>>2]|0;
        $35 = ($34>>>0)<($15>>>0);
        if ($35) {
         $37 = $34;$38 = $32;
        } else {
         (_luaL_prepbuffer($b)|0);
         $$pre$i$i = HEAP8[$31>>0]|0;
         $$pre2$i$i = HEAP32[$b>>2]|0;
         $37 = $$pre2$i$i;$38 = $$pre$i$i;
        }
        $36 = ((($37)) + 1|0);
        HEAP32[$b>>2] = $36;
        HEAP8[$37>>0] = $38;
        $i$1$i$i = $i$01$i$i;
       }
      } while(0);
      $51 = (($i$1$i$i) + 1)|0;
      $52 = HEAP32[$l$i$i>>2]|0;
      $53 = ($51>>>0)<($52>>>0);
      if ($53) {
       $i$01$i$i = $51;
      } else {
       break;
      }
     }
    }
    break;
   }
   case 6:  {
    _lua_pushvalue($22,3);
    $54 = HEAP32[$14>>2]|0;
    $55 = ($54|0)==(0);
    $56 = ($src$0|0)!=(0|0);
    $or$cond$i$i = $56 & $55;
    $$$i$i = $or$cond$i$i ? 1 : $54;
    $57 = HEAP32[$10>>2]|0;
    _luaL_checkstack($57,$$$i$i,522593);
    $58 = ($$$i$i|0)>(0);
    if ($58) {
     $i$02$i$i = 0;
     while(1) {
      _push_onecapture($ms,$i$02$i$i,$src$0,$19);
      $59 = (($i$02$i$i) + 1)|0;
      $exitcond$i$i = ($59|0)==($$$i$i|0);
      if ($exitcond$i$i) {
       break;
      } else {
       $i$02$i$i = $59;
      }
     }
    }
    _lua_call($22,$$$i$i,1);
    label = 33;
    break;
   }
   case 5:  {
    $60 = HEAP32[$14>>2]|0;
    $61 = ($60|0)>(0);
    do {
     if ($61) {
      $66 = HEAP32[$16>>2]|0;
      $67 = ($66|0)==(-1);
      if ($67) {
       $68 = HEAP32[$10>>2]|0;
       (_luaL_error($68,522778,$vararg_buffer)|0);
       $69 = HEAP32[$10>>2]|0;
       $70 = HEAP32[$17>>2]|0;
       $78 = $69;$79 = $70;
      } else {
       $71 = ($66|0)==(-2);
       $72 = HEAP32[$10>>2]|0;
       $73 = HEAP32[$17>>2]|0;
       if ($71) {
        $74 = HEAP32[$ms>>2]|0;
        $75 = $73;
        $76 = (($75) + 1)|0;
        $77 = (($76) - ($74))|0;
        _lua_pushinteger($72,$77);
        break;
       } else {
        $78 = $72;$79 = $73;
       }
      }
      _lua_pushlstring($78,$79,$66);
     } else {
      $62 = HEAP32[$10>>2]|0;
      $63 = $19;
      $64 = $src$0;
      $65 = (($63) - ($64))|0;
      _lua_pushlstring($62,$src$0,$65);
     }
    } while(0);
    _lua_gettable($22,3);
    label = 33;
    break;
   }
   default: {
    label = 33;
   }
   }
   if ((label|0) == 33) {
    label = 0;
    $80 = (_lua_toboolean($22,-1)|0);
    $81 = ($80|0)==(0);
    if ($81) {
     _lua_settop($22,-2);
     $82 = $19;
     $83 = $src$0;
     $84 = (($82) - ($83))|0;
     _lua_pushlstring($22,$src$0,$84);
    } else {
     $85 = (_lua_isstring($22,-1)|0);
     $86 = ($85|0)==(0);
     if ($86) {
      $87 = (_lua_type($22,-1)|0);
      $88 = (_lua_typename($22,$87)|0);
      HEAP32[$vararg_buffer1>>2] = $88;
      (_luaL_error($22,523026,$vararg_buffer1)|0);
     }
    }
    _luaL_addvalue($b);
   }
   $89 = ($19>>>0)>($src$0>>>0);
   if ($89) {
    $n$111 = $21;$src$1 = $19;
   } else {
    $n$110 = $21;
    label = 39;
   }
  }
  if ((label|0) == 39) {
   label = 0;
   $90 = HEAP32[$13>>2]|0;
   $91 = ($src$0>>>0)<($90>>>0);
   if (!($91)) {
    $n$2 = $n$110;$src$3 = $src$0;
    label = 44;
    break;
   }
   $92 = HEAP32[$b>>2]|0;
   $93 = ($92>>>0)<($15>>>0);
   if ($93) {
    $97 = $92;
   } else {
    (_luaL_prepbuffer($b)|0);
    $$pre = HEAP32[$b>>2]|0;
    $97 = $$pre;
   }
   $94 = ((($src$0)) + 1|0);
   $95 = HEAP8[$src$0>>0]|0;
   $96 = ((($97)) + 1|0);
   HEAP32[$b>>2] = $96;
   HEAP8[$97>>0] = $95;
   $n$111 = $n$110;$src$1 = $94;
  }
  if ($7) {
   $n$2 = $n$111;$src$3 = $src$1;
   label = 44;
   break;
  } else {
   $n$0 = $n$111;$src$0 = $src$1;
  }
 }
 if ((label|0) == 44) {
  $98 = HEAP32[$13>>2]|0;
  $99 = $src$3;
  $100 = (($98) - ($99))|0;
  _luaL_addlstring($b,$src$3,$100);
  _luaL_pushresult($b);
  _lua_pushinteger($L,$n$2);
  STACKTOP = sp;return 2;
 }
 return (0)|0;
}
function _str_len($L) {
 $L = $L|0;
 var $0 = 0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 (_luaL_checklstring($L,1,$l)|0);
 $0 = HEAP32[$l>>2]|0;
 _lua_pushinteger($L,$0);
 STACKTOP = sp;return 1;
}
function _str_lower($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b = 0, $i$01 = 0, $l = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1040|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp + 1036|0;
 $b = sp;
 $0 = (_luaL_checklstring($L,1,$l)|0);
 _luaL_buffinit($L,$b);
 $1 = HEAP32[$l>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  _luaL_pushresult($b);
  STACKTOP = sp;return 1;
 }
 $3 = ((($b)) + 1036|0);
 $i$01 = 0;
 while(1) {
  $4 = HEAP32[$b>>2]|0;
  $5 = ($4>>>0)<($3>>>0);
  if (!($5)) {
   (_luaL_prepbuffer($b)|0);
  }
  $6 = (($0) + ($i$01)|0);
  $7 = HEAP8[$6>>0]|0;
  $8 = $7&255;
  $9 = (_tolower($8)|0);
  $10 = $9&255;
  $11 = HEAP32[$b>>2]|0;
  $12 = ((($11)) + 1|0);
  HEAP32[$b>>2] = $12;
  HEAP8[$11>>0] = $10;
  $13 = (($i$01) + 1)|0;
  $14 = HEAP32[$l>>2]|0;
  $15 = ($13>>>0)<($14>>>0);
  if ($15) {
   $i$01 = $13;
  } else {
   break;
  }
 }
 _luaL_pushresult($b);
 STACKTOP = sp;return 1;
}
function _str_match($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_str_find_aux($L,0)|0);
 return ($0|0);
}
function _str_rep($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $b = 0, $l = 0, $n$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1040|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp + 1036|0;
 $b = sp;
 $0 = (_luaL_checklstring($L,1,$l)|0);
 $1 = (_luaL_checkinteger($L,2)|0);
 _luaL_buffinit($L,$b);
 $2 = ($1|0)>(0);
 if ($2) {
  $n$01 = $1;
 } else {
  _luaL_pushresult($b);
  STACKTOP = sp;return 1;
 }
 while(1) {
  $3 = (($n$01) + -1)|0;
  $4 = HEAP32[$l>>2]|0;
  _luaL_addlstring($b,$0,$4);
  $5 = ($n$01|0)>(1);
  if ($5) {
   $n$01 = $3;
  } else {
   break;
  }
 }
 _luaL_pushresult($b);
 STACKTOP = sp;return 1;
}
function _str_reverse($L) {
 $L = $L|0;
 var $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b = 0, $l = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1040|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp + 1036|0;
 $b = sp;
 $0 = (_luaL_checklstring($L,1,$l)|0);
 _luaL_buffinit($L,$b);
 $1 = HEAP32[$l>>2]|0;
 $2 = (($1) + -1)|0;
 HEAP32[$l>>2] = $2;
 $3 = ($1|0)==(0);
 if ($3) {
  _luaL_pushresult($b);
  STACKTOP = sp;return 1;
 }
 $4 = ((($b)) + 1036|0);
 $15 = $2;
 while(1) {
  $5 = HEAP32[$b>>2]|0;
  $6 = ($5>>>0)<($4>>>0);
  if ($6) {
   $11 = $5;$8 = $15;
  } else {
   (_luaL_prepbuffer($b)|0);
   $$pre = HEAP32[$l>>2]|0;
   $$pre1 = HEAP32[$b>>2]|0;
   $11 = $$pre1;$8 = $$pre;
  }
  $7 = (($0) + ($8)|0);
  $9 = HEAP8[$7>>0]|0;
  $10 = ((($11)) + 1|0);
  HEAP32[$b>>2] = $10;
  HEAP8[$11>>0] = $9;
  $12 = HEAP32[$l>>2]|0;
  $13 = (($12) + -1)|0;
  HEAP32[$l>>2] = $13;
  $14 = ($12|0)==(0);
  if ($14) {
   break;
  } else {
   $15 = $13;
  }
 }
 _luaL_pushresult($b);
 STACKTOP = sp;return 1;
}
function _str_sub($L) {
 $L = $L|0;
 var $$ = 0, $$0$i = 0, $$0$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $end$0 = 0, $l = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp;
 $0 = (_luaL_checklstring($L,1,$l)|0);
 $1 = (_luaL_checkinteger($L,2)|0);
 $2 = HEAP32[$l>>2]|0;
 $3 = ($1|0)<(0);
 $4 = (($2) + 1)|0;
 $5 = $3 ? $4 : 0;
 $$0$i = (($5) + ($1))|0;
 $6 = (_luaL_optinteger($L,3,-1)|0);
 $7 = HEAP32[$l>>2]|0;
 $8 = ($6|0)<(0);
 $9 = (($7) + 1)|0;
 $10 = $8 ? $9 : 0;
 $$0$i1 = (($10) + ($6))|0;
 $11 = ($$0$i1|0)<(0);
 $12 = $11 ? 0 : $$0$i1;
 $13 = ($$0$i|0)<(1);
 $$ = $13 ? 1 : $$0$i;
 $14 = ($12|0)>($7|0);
 $end$0 = $14 ? $7 : $12;
 $15 = ($end$0|0)<($$|0);
 if ($15) {
  _lua_pushlstring($L,1282466,0);
  STACKTOP = sp;return 1;
 } else {
  $16 = (($0) + ($$)|0);
  $17 = ((($16)) + -1|0);
  $18 = (1 - ($$))|0;
  $19 = (($18) + ($end$0))|0;
  _lua_pushlstring($L,$17,$19);
  STACKTOP = sp;return 1;
 }
 return (0)|0;
}
function _str_upper($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b = 0, $i$01 = 0, $l = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1040|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $l = sp + 1036|0;
 $b = sp;
 $0 = (_luaL_checklstring($L,1,$l)|0);
 _luaL_buffinit($L,$b);
 $1 = HEAP32[$l>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  _luaL_pushresult($b);
  STACKTOP = sp;return 1;
 }
 $3 = ((($b)) + 1036|0);
 $i$01 = 0;
 while(1) {
  $4 = HEAP32[$b>>2]|0;
  $5 = ($4>>>0)<($3>>>0);
  if (!($5)) {
   (_luaL_prepbuffer($b)|0);
  }
  $6 = (($0) + ($i$01)|0);
  $7 = HEAP8[$6>>0]|0;
  $8 = $7&255;
  $9 = (_toupper($8)|0);
  $10 = $9&255;
  $11 = HEAP32[$b>>2]|0;
  $12 = ((($11)) + 1|0);
  HEAP32[$b>>2] = $12;
  HEAP8[$11>>0] = $10;
  $13 = (($i$01) + 1)|0;
  $14 = HEAP32[$l>>2]|0;
  $15 = ($13>>>0)<($14>>>0);
  if ($15) {
   $i$01 = $13;
  } else {
   break;
  }
 }
 _luaL_pushresult($b);
 STACKTOP = sp;return 1;
}
function _luaopen_package($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 (_luaL_newmetatable($L,523097)|0);
 _lua_pushcclosure($L,4321,0);
 _lua_setfield($L,-2,523106);
 _luaL_register($L,523111,40376);
 _lua_pushvalue($L,-1);
 _lua_replace($L,-10001);
 _lua_createtable($L,4,0);
 _lua_pushcclosure($L,4322,0);
 _lua_rawseti($L,-2,1);
 _lua_pushcclosure($L,4323,0);
 _lua_rawseti($L,-2,2);
 _lua_pushcclosure($L,4324,0);
 _lua_rawseti($L,-2,3);
 _lua_pushcclosure($L,4325,0);
 _lua_rawseti($L,-2,4);
 _lua_setfield($L,-2,523447);
 $0 = (_getenv((523455|0))|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  _lua_pushstring($L,523464);
 } else {
  $2 = (_luaL_gsub($L,$0,523602,523605)|0);
  (_luaL_gsub($L,$2,530008,523464)|0);
  _lua_remove($L,-2);
 }
 _lua_setfield($L,-2,523298);
 $3 = (_getenv((523609|0))|0);
 $4 = ($3|0)==(0|0);
 if ($4) {
  _lua_pushstring($L,523619);
 } else {
  $5 = (_luaL_gsub($L,$3,523602,523605)|0);
  (_luaL_gsub($L,$5,530008,523619)|0);
  _lua_remove($L,-2);
 }
 _lua_setfield($L,-2,523398);
 _lua_pushlstring($L,523688,9);
 _lua_setfield($L,-2,523698);
 (_luaL_findtable($L,-10000,523705,2)|0);
 _lua_setfield($L,-2,523713);
 _lua_createtable($L,0,0);
 _lua_setfield($L,-2,523223);
 _lua_pushvalue($L,-10002);
 _luaL_register($L,0,40400);
 _lua_settop($L,-2);
 return 1;
}
function _gctm($L) {
 $L = $L|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checkudata($L,1,523097)|0);
 HEAP32[$0>>2] = 0;
 return 0;
}
function _ll_loadlib($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_luaL_checklstring($L,1,0)|0);
 (_luaL_checklstring($L,2,0)|0);
 $1 = (_ll_loadfunc($L,$0)|0);
 $2 = ($1|0)==(0);
 if ($2) {
  $$0 = 1;
  return ($$0|0);
 }
 _lua_pushnil($L);
 _lua_insert($L,-2);
 $3 = ($1|0)==(1);
 $4 = $3 ? 523196 : 523203;
 _lua_pushstring($L,$4);
 $$0 = 3;
 return ($$0|0);
}
function _ll_loadfunc($L,$path) {
 $L = $L|0;
 $path = $path|0;
 var $$1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $plib$0$i = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 HEAP32[$vararg_buffer>>2] = 523127;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $path;
 (_lua_pushfstring($L,526499,$vararg_buffer)|0);
 _lua_gettable($L,-10000);
 $0 = (_lua_type($L,-1)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  _lua_settop($L,-2);
  $3 = (_lua_newuserdata($L,4)|0);
  HEAP32[$3>>2] = 0;
  _lua_getfield($L,-10000,523097);
  (_lua_setmetatable($L,-2)|0);
  HEAP32[$vararg_buffer2>>2] = 523127;
  $vararg_ptr5 = ((($vararg_buffer2)) + 4|0);
  HEAP32[$vararg_ptr5>>2] = $path;
  (_lua_pushfstring($L,526499,$vararg_buffer2)|0);
  _lua_pushvalue($L,-2);
  _lua_settable($L,-10000);
  $plib$0$i = $3;
 } else {
  $2 = (_lua_touserdata($L,-1)|0);
  $plib$0$i = $2;
 }
 $4 = HEAP32[$plib$0$i>>2]|0;
 $5 = ($4|0)==(0|0);
 _lua_pushlstring($L,523137,58);
 if (!($5)) {
  $$1 = 2;
  STACKTOP = sp;return ($$1|0);
 }
 HEAP32[$plib$0$i>>2] = 0;
 $$1 = 1;
 STACKTOP = sp;return ($$1|0);
}
function _ll_seeall($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _luaL_checktype($L,1,5);
 $0 = (_lua_getmetatable($L,1)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  _lua_createtable($L,0,1);
  _lua_pushvalue($L,-1);
  (_lua_setmetatable($L,1)|0);
 }
 _lua_pushvalue($L,-10002);
 _lua_setfield($L,-2,523215);
 return 0;
}
function _loader_preload($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_luaL_checklstring($L,1,0)|0);
 _lua_getfield($L,-10001,523223);
 $1 = (_lua_type($L,-1)|0);
 $2 = ($1|0)==(5);
 if (!($2)) {
  (_luaL_error($L,523231,$vararg_buffer)|0);
 }
 _lua_getfield($L,-1,$0);
 $3 = (_lua_type($L,-1)|0);
 $4 = ($3|0)==(0);
 if (!($4)) {
  STACKTOP = sp;return 1;
 }
 HEAP32[$vararg_buffer1>>2] = $0;
 (_lua_pushfstring($L,523265,$vararg_buffer1)|0);
 STACKTOP = sp;return 1;
}
function _loader_Lua($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_luaL_checklstring($L,1,0)|0);
 $1 = (_findfile($L,$0,523298)|0);
 $2 = ($1|0)==(0|0);
 if ($2) {
  STACKTOP = sp;return 1;
 }
 $3 = (_luaL_loadfile($L,$1)|0);
 $4 = ($3|0)==(0);
 if ($4) {
  STACKTOP = sp;return 1;
 }
 $5 = (_lua_tolstring($L,1,0)|0);
 $6 = (_lua_tolstring($L,-1,0)|0);
 HEAP32[$vararg_buffer>>2] = $5;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $1;
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
 HEAP32[$vararg_ptr2>>2] = $6;
 (_luaL_error($L,523352,$vararg_buffer)|0);
 STACKTOP = sp;return 1;
}
function _findfile($L,$name,$pname) {
 $L = $L|0;
 $name = $name|0;
 $pname = $pname|0;
 var $$0$i = 0, $$2 = 0, $$lcssa = 0, $$lcssa8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $l$0$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_luaL_gsub($L,$name,619449,523303)|0);
 _lua_getfield($L,-10001,$pname);
 $1 = (_lua_tolstring($L,-1,0)|0);
 $2 = ($1|0)==(0|0);
 if ($2) {
  HEAP32[$vararg_buffer>>2] = $pname;
  (_luaL_error($L,523305,$vararg_buffer)|0);
 }
 _lua_pushlstring($L,1282466,0);
 $$0$i = $1;
 L4: while(1) {
  $3 = HEAP8[$$0$i>>0]|0;
  switch ($3<<24>>24) {
  case 0:  {
   $$2 = 0;
   label = 12;
   break L4;
   break;
  }
  case 59:  {
   $4 = ((($$0$i)) + 1|0);
   $$0$i = $4;
   continue L4;
   break;
  }
  default: {
  }
  }
  $5 = (_strchr($$0$i,59)|0);
  $6 = ($5|0)==(0|0);
  if ($6) {
   $7 = (_strlen($$0$i)|0);
   $8 = (($$0$i) + ($7)|0);
   $l$0$i = $8;
  } else {
   $l$0$i = $5;
  }
  $9 = $l$0$i;
  $10 = $$0$i;
  $11 = (($9) - ($10))|0;
  _lua_pushlstring($L,$$0$i,$11);
  $12 = ($l$0$i|0)==(0|0);
  if ($12) {
   $$2 = 0;
   label = 12;
   break;
  }
  $13 = (_lua_tolstring($L,-1,0)|0);
  $14 = (_luaL_gsub($L,$13,523335,$0)|0);
  _lua_remove($L,-2);
  $15 = (_fopen($14,527143)|0);
  $16 = ($15|0)==(0|0);
  if (!($16)) {
   $$lcssa = $14;$$lcssa8 = $15;
   label = 11;
   break;
  }
  HEAP32[$vararg_buffer1>>2] = $14;
  (_lua_pushfstring($L,523337,$vararg_buffer1)|0);
  _lua_remove($L,-2);
  _lua_concat($L,2);
  $$0$i = $l$0$i;
 }
 if ((label|0) == 11) {
  (_fclose($$lcssa8)|0);
  $$2 = $$lcssa;
  STACKTOP = sp;return ($$2|0);
 }
 else if ((label|0) == 12) {
  STACKTOP = sp;return ($$2|0);
 }
 return (0)|0;
}
function _loader_C($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $modname$$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_ptr4 = 0, $vararg_ptr5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_luaL_checklstring($L,1,0)|0);
 $1 = (_findfile($L,$0,523398)|0);
 $2 = ($1|0)==(0|0);
 if ($2) {
  STACKTOP = sp;return 1;
 }
 $3 = (_strchr($0,45)|0);
 $4 = ($3|0)==(0|0);
 $5 = ((($3)) + 1|0);
 $modname$$i = $4 ? $0 : $5;
 $6 = (_luaL_gsub($L,$modname$$i,619449,523404)|0);
 HEAP32[$vararg_buffer>>2] = $6;
 (_lua_pushfstring($L,523406,$vararg_buffer)|0);
 _lua_remove($L,-2);
 $7 = (_ll_loadfunc($L,$1)|0);
 $8 = ($7|0)==(0);
 if ($8) {
  STACKTOP = sp;return 1;
 }
 $9 = (_lua_tolstring($L,1,0)|0);
 $10 = (_lua_tolstring($L,-1,0)|0);
 HEAP32[$vararg_buffer1>>2] = $9;
 $vararg_ptr4 = ((($vararg_buffer1)) + 4|0);
 HEAP32[$vararg_ptr4>>2] = $1;
 $vararg_ptr5 = ((($vararg_buffer1)) + 8|0);
 HEAP32[$vararg_ptr5>>2] = $10;
 (_luaL_error($L,523352,$vararg_buffer1)|0);
 STACKTOP = sp;return 1;
}
function _loader_Croot($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $modname$$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0;
 var $vararg_buffer6 = 0, $vararg_ptr4 = 0, $vararg_ptr5 = 0, $vararg_ptr9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_luaL_checklstring($L,1,0)|0);
 $1 = (_strchr($0,46)|0);
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = $1;
 $4 = $0;
 $5 = (($3) - ($4))|0;
 _lua_pushlstring($L,$0,$5);
 $6 = (_lua_tolstring($L,-1,0)|0);
 $7 = (_findfile($L,$6,523398)|0);
 $8 = ($7|0)==(0|0);
 if ($8) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 $9 = (_strchr($0,45)|0);
 $10 = ($9|0)==(0|0);
 $11 = ((($9)) + 1|0);
 $modname$$i = $10 ? $0 : $11;
 $12 = (_luaL_gsub($L,$modname$$i,619449,523404)|0);
 HEAP32[$vararg_buffer>>2] = $12;
 (_lua_pushfstring($L,523406,$vararg_buffer)|0);
 _lua_remove($L,-2);
 $13 = (_ll_loadfunc($L,$7)|0);
 switch ($13|0) {
 case 2:  {
  break;
 }
 case 0:  {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 default: {
  $14 = (_lua_tolstring($L,1,0)|0);
  $15 = (_lua_tolstring($L,-1,0)|0);
  HEAP32[$vararg_buffer1>>2] = $14;
  $vararg_ptr4 = ((($vararg_buffer1)) + 4|0);
  HEAP32[$vararg_ptr4>>2] = $7;
  $vararg_ptr5 = ((($vararg_buffer1)) + 8|0);
  HEAP32[$vararg_ptr5>>2] = $15;
  (_luaL_error($L,523352,$vararg_buffer1)|0);
 }
 }
 HEAP32[$vararg_buffer6>>2] = $0;
 $vararg_ptr9 = ((($vararg_buffer6)) + 4|0);
 HEAP32[$vararg_ptr9>>2] = $7;
 (_lua_pushfstring($L,523417,$vararg_buffer6)|0);
 $$0 = 1;
 STACKTOP = sp;return ($$0|0);
}
function _ll_module($L) {
 $L = $L|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ar$i = 0, $dot$0$i = 0, $exitcond$i = 0, $i$01$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $ar$i = sp + 12|0;
 $0 = (_luaL_checklstring($L,1,0)|0);
 $1 = (_lua_gettop($L)|0);
 $2 = (($1) + 1)|0;
 _lua_getfield($L,-10000,523705);
 _lua_getfield($L,$2,$0);
 $3 = (_lua_type($L,-1)|0);
 $4 = ($3|0)==(5);
 do {
  if (!($4)) {
   _lua_settop($L,-2);
   $5 = (_luaL_findtable($L,-10002,$0,1)|0);
   $6 = ($5|0)==(0|0);
   if ($6) {
    _lua_pushvalue($L,-1);
    _lua_setfield($L,$2,$0);
    break;
   }
   HEAP32[$vararg_buffer>>2] = $0;
   $7 = (_luaL_error($L,523727,$vararg_buffer)|0);
   $$0 = $7;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 _lua_getfield($L,-1,523757);
 $8 = (_lua_type($L,-1)|0);
 $9 = ($8|0)==(0);
 _lua_settop($L,-2);
 if ($9) {
  _lua_pushvalue($L,-1);
  _lua_setfield($L,-2,523763);
  _lua_pushstring($L,$0);
  _lua_setfield($L,-2,523757);
  $10 = (_strrchr($0,46)|0);
  $11 = ($10|0)==(0|0);
  $12 = ((($10)) + 1|0);
  $dot$0$i = $11 ? $0 : $12;
  $13 = $dot$0$i;
  $14 = $0;
  $15 = (($13) - ($14))|0;
  _lua_pushlstring($L,$0,$15);
  _lua_setfield($L,-2,523766);
 }
 _lua_pushvalue($L,-1);
 $16 = (_lua_getstack($L,1,$ar$i)|0);
 $17 = ($16|0)==(0);
 if ($17) {
  label = 10;
 } else {
  $18 = (_lua_getinfo($L,523775,$ar$i)|0);
  $19 = ($18|0)==(0);
  if ($19) {
   label = 10;
  } else {
   $20 = (_lua_iscfunction($L,-1)|0);
   $21 = ($20|0)==(0);
   if (!($21)) {
    label = 10;
   }
  }
 }
 if ((label|0) == 10) {
  (_luaL_error($L,523777,$vararg_buffer1)|0);
 }
 _lua_pushvalue($L,-2);
 (_lua_setfenv($L,-2)|0);
 _lua_settop($L,-2);
 $22 = ($1|0)<(2);
 if ($22) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $i$01$i = 2;
 }
 while(1) {
  _lua_pushvalue($L,$i$01$i);
  _lua_pushvalue($L,-2);
  _lua_call($L,1,0);
  $23 = (($i$01$i) + 1)|0;
  $exitcond$i = ($i$01$i|0)==($1|0);
  if ($exitcond$i) {
   $$0 = 0;
   break;
  } else {
   $i$01$i = $23;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _ll_require($L) {
 $L = $L|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_ptr6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_luaL_checklstring($L,1,0)|0);
 _lua_settop($L,1);
 _lua_getfield($L,-10000,523705);
 _lua_getfield($L,2,$0);
 $1 = (_lua_toboolean($L,-1)|0);
 $2 = ($1|0)==(0);
 if (!($2)) {
  $3 = (_lua_touserdata($L,-1)|0);
  $4 = ($3|0)==(732312|0);
  if (!($4)) {
   STACKTOP = sp;return 1;
  }
  HEAP32[$vararg_buffer>>2] = $0;
  (_luaL_error($L,523825,$vararg_buffer)|0);
  STACKTOP = sp;return 1;
 }
 _lua_getfield($L,-10001,523447);
 $5 = (_lua_type($L,-1)|0);
 $6 = ($5|0)==(5);
 if (!($6)) {
  (_luaL_error($L,523868,$vararg_buffer1)|0);
 }
 _lua_pushlstring($L,1282466,0);
 $i$0 = 1;
 while(1) {
  _lua_rawgeti($L,-2,$i$0);
  $7 = (_lua_type($L,-1)|0);
  $8 = ($7|0)==(0);
  if ($8) {
   $9 = (_lua_tolstring($L,-2,0)|0);
   HEAP32[$vararg_buffer3>>2] = $0;
   $vararg_ptr6 = ((($vararg_buffer3)) + 4|0);
   HEAP32[$vararg_ptr6>>2] = $9;
   (_luaL_error($L,523902,$vararg_buffer3)|0);
  }
  _lua_pushstring($L,$0);
  _lua_call($L,1,1);
  $10 = (_lua_type($L,-1)|0);
  $11 = ($10|0)==(6);
  if ($11) {
   break;
  }
  $12 = (_lua_isstring($L,-1)|0);
  $13 = ($12|0)==(0);
  if ($13) {
   _lua_settop($L,-2);
  } else {
   _lua_concat($L,2);
  }
  $14 = (($i$0) + 1)|0;
  $i$0 = $14;
 }
 _lua_pushlightuserdata($L,732312);
 _lua_setfield($L,2,$0);
 _lua_pushstring($L,$0);
 _lua_call($L,1,1);
 $15 = (_lua_type($L,-1)|0);
 $16 = ($15|0)==(0);
 if (!($16)) {
  _lua_setfield($L,2,$0);
 }
 _lua_getfield($L,2,$0);
 $17 = (_lua_touserdata($L,-1)|0);
 $18 = ($17|0)==(732312|0);
 if (!($18)) {
  STACKTOP = sp;return 1;
 }
 _lua_pushboolean($L,1);
 _lua_pushvalue($L,-1);
 _lua_setfield($L,2,$0);
 STACKTOP = sp;return 1;
}
function __GLOBAL__sub_I_fastmix_cpp() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN13CzCUBICSPLINEC2Ev(0); //@line 142 "fastmix.cpp"
 __ZN13CzWINDOWEDFIRC2Ev(1252507); //@line 285 "fastmix.cpp"
 return;
}
function __ZN13CzCUBICSPLINEC2Ev($this) {
 $this = $this|0;
 var $$ = 0, $$_LMax$1 = 0, $0 = 0.0, $1 = 0.0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0.0;
 var $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0.0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0.0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0.0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0.0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0.0, $80 = 0, $81 = 0, $82 = 0, $9 = 0.0, $_LIi$05 = 0, $_LMax$1 = 0, $exitcond = 0, $tmp2 = 0.0, $tmp3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $_LIi$05 = 0;
 while(1) {
  $0 = (+($_LIi$05|0)); //@line 115 "fastmix.cpp"
  $1 = $0 * 9.765625E-4; //@line 115 "fastmix.cpp"
  $2 = $_LIi$05 << 2; //@line 116 "fastmix.cpp"
  $3 = $1; //@line 117 "fastmix.cpp"
  $4 = $3 * $3; //@line 117 "fastmix.cpp"
  $5 = $4 * 0.5; //@line 117 "fastmix.cpp"
  $6 = $3 * 0.5; //@line 117 "fastmix.cpp"
  $7 = $3 + -0.5; //@line 117 "fastmix.cpp"
  $8 = $7 - $5; //@line 117 "fastmix.cpp"
  $9 = $3 * 16384.0; //@line 117 "fastmix.cpp"
  $10 = $9 * $8; //@line 117 "fastmix.cpp"
  $11 = $10 + 0.5; //@line 117 "fastmix.cpp"
  $12 = (+Math_floor((+$11))); //@line 117 "fastmix.cpp"
  $13 = $12; //@line 117 "fastmix.cpp"
  $14 = $3 * 1.5; //@line 118 "fastmix.cpp"
  $15 = $14 * $4; //@line 118 "fastmix.cpp"
  $16 = $4 * 2.5; //@line 118 "fastmix.cpp"
  $17 = 1.0 - $16; //@line 118 "fastmix.cpp"
  $18 = $17 + $15; //@line 118 "fastmix.cpp"
  $19 = $18 * 16384.0; //@line 118 "fastmix.cpp"
  $20 = $19 + 0.5; //@line 118 "fastmix.cpp"
  $21 = (+Math_floor((+$20))); //@line 118 "fastmix.cpp"
  $22 = $21; //@line 118 "fastmix.cpp"
  $tmp3 = 2.0 - $14; //@line 119 "fastmix.cpp"
  $tmp2 = $tmp3 * $4;
  $23 = $tmp2 + $6; //@line 119 "fastmix.cpp"
  $24 = $23 * 16384.0; //@line 119 "fastmix.cpp"
  $25 = $24 + 0.5; //@line 119 "fastmix.cpp"
  $26 = (+Math_floor((+$25))); //@line 119 "fastmix.cpp"
  $27 = $26; //@line 119 "fastmix.cpp"
  $28 = $6 + -0.5; //@line 120 "fastmix.cpp"
  $29 = $4 * 16384.0; //@line 120 "fastmix.cpp"
  $30 = $29 * $28; //@line 120 "fastmix.cpp"
  $31 = $30 + 0.5; //@line 120 "fastmix.cpp"
  $32 = (+Math_floor((+$31))); //@line 120 "fastmix.cpp"
  $33 = $32; //@line 120 "fastmix.cpp"
  $34 = $13 < -16384.0; //@line 121 "fastmix.cpp"
  $35 = $13 > 16384.0; //@line 121 "fastmix.cpp"
  $36 = (~~(($13))); //@line 121 "fastmix.cpp"
  $37 = $35 ? 16384 : $36; //@line 121 "fastmix.cpp"
  $38 = $34 ? -16384 : $37; //@line 121 "fastmix.cpp"
  $39 = (1209506 + ($2<<1)|0); //@line 121 "fastmix.cpp"
  HEAP16[$39>>1] = $38; //@line 121 "fastmix.cpp"
  $40 = $22 < -16384.0; //@line 122 "fastmix.cpp"
  $41 = $22 > 16384.0; //@line 122 "fastmix.cpp"
  $42 = (~~(($22))); //@line 122 "fastmix.cpp"
  $43 = $41 ? 16384 : $42; //@line 122 "fastmix.cpp"
  $44 = $40 ? -16384 : $43; //@line 122 "fastmix.cpp"
  $45 = $2 | 1; //@line 122 "fastmix.cpp"
  $46 = (1209506 + ($45<<1)|0); //@line 122 "fastmix.cpp"
  HEAP16[$46>>1] = $44; //@line 122 "fastmix.cpp"
  $47 = $27 < -16384.0; //@line 123 "fastmix.cpp"
  $48 = $27 > 16384.0; //@line 123 "fastmix.cpp"
  $49 = (~~(($27))); //@line 123 "fastmix.cpp"
  $50 = $48 ? 16384 : $49; //@line 123 "fastmix.cpp"
  $51 = $47 ? -16384 : $50; //@line 123 "fastmix.cpp"
  $52 = $2 | 2; //@line 123 "fastmix.cpp"
  $53 = (1209506 + ($52<<1)|0); //@line 123 "fastmix.cpp"
  HEAP16[$53>>1] = $51; //@line 123 "fastmix.cpp"
  $54 = $33 < -16384.0; //@line 124 "fastmix.cpp"
  $55 = $33 > 16384.0; //@line 124 "fastmix.cpp"
  $56 = (~~(($33))); //@line 124 "fastmix.cpp"
  $57 = $55 ? 16384 : $56; //@line 124 "fastmix.cpp"
  $58 = $54 ? -16384 : $57; //@line 124 "fastmix.cpp"
  $59 = $2 | 3; //@line 124 "fastmix.cpp"
  $60 = (1209506 + ($59<<1)|0); //@line 124 "fastmix.cpp"
  HEAP16[$60>>1] = $58; //@line 124 "fastmix.cpp"
  $61 = $38 << 16 >> 16; //@line 126 "fastmix.cpp"
  $62 = $44 << 16 >> 16; //@line 126 "fastmix.cpp"
  $63 = (($62) + ($61))|0; //@line 126 "fastmix.cpp"
  $64 = $51 << 16 >> 16; //@line 126 "fastmix.cpp"
  $65 = (($63) + ($64))|0; //@line 126 "fastmix.cpp"
  $66 = $58 << 16 >> 16; //@line 126 "fastmix.cpp"
  $67 = (($65) + ($66))|0; //@line 126 "fastmix.cpp"
  $68 = ($67|0)==(16384); //@line 127 "fastmix.cpp"
  if (!($68)) {
   $69 = ($44<<16>>16)>($38<<16>>16); //@line 129 "fastmix.cpp"
   $$ = $69 ? $45 : $2; //@line 129 "fastmix.cpp"
   $70 = (1209506 + ($$<<1)|0); //@line 130 "fastmix.cpp"
   $71 = HEAP16[$70>>1]|0; //@line 130 "fastmix.cpp"
   $72 = ($51<<16>>16)>($71<<16>>16); //@line 130 "fastmix.cpp"
   $_LMax$1 = $72 ? $52 : $$; //@line 130 "fastmix.cpp"
   $73 = (1209506 + ($_LMax$1<<1)|0); //@line 131 "fastmix.cpp"
   $74 = HEAP16[$73>>1]|0; //@line 131 "fastmix.cpp"
   $75 = ($58<<16>>16)>($74<<16>>16); //@line 131 "fastmix.cpp"
   $$_LMax$1 = $75 ? $59 : $_LMax$1; //@line 131 "fastmix.cpp"
   $76 = (16384 - ($67))|0; //@line 132 "fastmix.cpp"
   $77 = (1209506 + ($$_LMax$1<<1)|0); //@line 132 "fastmix.cpp"
   $78 = HEAP16[$77>>1]|0; //@line 132 "fastmix.cpp"
   $79 = $78&65535; //@line 132 "fastmix.cpp"
   $80 = (($76) + ($79))|0; //@line 132 "fastmix.cpp"
   $81 = $80&65535; //@line 132 "fastmix.cpp"
   HEAP16[$77>>1] = $81; //@line 132 "fastmix.cpp"
  }
  $82 = (($_LIi$05) + 1)|0; //@line 113 "fastmix.cpp"
  $exitcond = ($82|0)==(1024); //@line 113 "fastmix.cpp"
  if ($exitcond) {
   break;
  } else {
   $_LIi$05 = $82;
  }
 }
 return; //@line 136 "fastmix.cpp"
}
function __ZN13CzWINDOWEDFIRC2Ev($this) {
 $this = $this|0;
 var $$lcssa = 0.0, $0 = 0, $1 = 0, $10 = 0.0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0.0, $110 = 0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0;
 var $115 = 0.0, $116 = 0.0, $117 = 0, $118 = 0, $119 = 0, $12 = 0.0, $120 = 0, $121 = 0, $122 = 0, $13 = 0.0, $14 = 0.0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0;
 var $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0;
 var $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0.0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0.0, $79 = 0.0, $8 = 0.0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $_LCc$01 = 0, $_LCoefs = 0, $_LGain$02 = 0.0, $_LPcl$04 = 0, $_LSi$0$i = 0.0, $_LWc$1$i = 0.0, $exitcond = 0, $exitcond6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $_LCoefs = sp;
 $0 = ((($_LCoefs)) + 4|0); //@line 275 "fastmix.cpp"
 $1 = ((($_LCoefs)) + 8|0); //@line 275 "fastmix.cpp"
 $2 = ((($_LCoefs)) + 12|0); //@line 275 "fastmix.cpp"
 $3 = ((($_LCoefs)) + 16|0); //@line 275 "fastmix.cpp"
 $4 = ((($_LCoefs)) + 20|0); //@line 275 "fastmix.cpp"
 $5 = ((($_LCoefs)) + 24|0); //@line 275 "fastmix.cpp"
 $6 = ((($_LCoefs)) + 28|0); //@line 275 "fastmix.cpp"
 $_LPcl$04 = 0;
 while(1) {
  $7 = (+($_LPcl$04|0)); //@line 268 "fastmix.cpp"
  $8 = $7 + -1024.0; //@line 268 "fastmix.cpp"
  $9 = $8 * 4.8828125E-4; //@line 268 "fastmix.cpp"
  $10 = $9; //@line 201 "fastmix.cpp"
  $_LCc$01 = 0;$_LGain$02 = 0.0;
  while(1) {
   $11 = (+($_LCc$01|0)); //@line 201 "fastmix.cpp"
   $12 = $11 - $10; //@line 201 "fastmix.cpp"
   $13 = $12 + -3.5; //@line 202 "fastmix.cpp"
   $14 = (+Math_abs((+$13))); //@line 205 "fastmix.cpp"
   $15 = $14 < 1.0E-8; //@line 205 "fastmix.cpp"
   if ($15) {
    $_LSi$0$i = 0.89999997615814208;$_LWc$1$i = 1.0;
   } else {
    $16 = $12 * 0.89759790102565518; //@line 218 "fastmix.cpp"
    $17 = (+Math_cos((+$16))); //@line 218 "fastmix.cpp"
    $18 = $17 * 0.5; //@line 218 "fastmix.cpp"
    $19 = 0.41999999999999998 - $18; //@line 218 "fastmix.cpp"
    $20 = $12 * 1.7951958020513104; //@line 219 "fastmix.cpp"
    $21 = (+Math_cos((+$20))); //@line 219 "fastmix.cpp"
    $22 = $21 * 0.080000000000000002; //@line 219 "fastmix.cpp"
    $23 = $19 + $22; //@line 218 "fastmix.cpp"
    $24 = $13 * 3.1415926535897931; //@line 248 "fastmix.cpp"
    $25 = $13 * 2.8274333133294083; //@line 249 "fastmix.cpp"
    $26 = (+Math_sin((+$25))); //@line 249 "fastmix.cpp"
    $27 = $26 / $24; //@line 249 "fastmix.cpp"
    $_LSi$0$i = $27;$_LWc$1$i = $23;
   }
   $28 = $_LSi$0$i * $_LWc$1$i; //@line 251 "fastmix.cpp"
   $29 = $28; //@line 251 "fastmix.cpp"
   $30 = (($_LCoefs) + ($_LCc$01<<2)|0); //@line 271 "fastmix.cpp"
   HEAPF32[$30>>2] = $29; //@line 271 "fastmix.cpp"
   $31 = $29 + $_LGain$02; //@line 271 "fastmix.cpp"
   $32 = (($_LCc$01) + 1)|0; //@line 270 "fastmix.cpp"
   $exitcond = ($32|0)==(8); //@line 270 "fastmix.cpp"
   if ($exitcond) {
    $$lcssa = $31;
    break;
   } else {
    $_LCc$01 = $32;$_LGain$02 = $31;
   }
  }
  $33 = $_LPcl$04 << 3; //@line 269 "fastmix.cpp"
  $34 = 32768.0 / $$lcssa; //@line 275 "fastmix.cpp"
  $35 = +HEAPF32[$_LCoefs>>2]; //@line 275 "fastmix.cpp"
  $36 = $34 * $35; //@line 275 "fastmix.cpp"
  $37 = $36; //@line 275 "fastmix.cpp"
  $38 = $37 + 0.5; //@line 275 "fastmix.cpp"
  $39 = (+Math_floor((+$38))); //@line 275 "fastmix.cpp"
  $40 = $39; //@line 275 "fastmix.cpp"
  $41 = $40 < -32768.0; //@line 276 "fastmix.cpp"
  $42 = (~~(($40))); //@line 276 "fastmix.cpp"
  $43 = $41 ? -32768 : $42; //@line 276 "fastmix.cpp"
  $44 = (1217698 + ($33<<1)|0); //@line 276 "fastmix.cpp"
  HEAP16[$44>>1] = $43; //@line 276 "fastmix.cpp"
  $45 = +HEAPF32[$0>>2]; //@line 275 "fastmix.cpp"
  $46 = $34 * $45; //@line 275 "fastmix.cpp"
  $47 = $46; //@line 275 "fastmix.cpp"
  $48 = $47 + 0.5; //@line 275 "fastmix.cpp"
  $49 = (+Math_floor((+$48))); //@line 275 "fastmix.cpp"
  $50 = $49; //@line 275 "fastmix.cpp"
  $51 = $50 < -32768.0; //@line 276 "fastmix.cpp"
  $52 = (~~(($50))); //@line 276 "fastmix.cpp"
  $53 = $51 ? -32768 : $52; //@line 276 "fastmix.cpp"
  $54 = $33 | 1; //@line 276 "fastmix.cpp"
  $55 = (1217698 + ($54<<1)|0); //@line 276 "fastmix.cpp"
  HEAP16[$55>>1] = $53; //@line 276 "fastmix.cpp"
  $56 = +HEAPF32[$1>>2]; //@line 275 "fastmix.cpp"
  $57 = $34 * $56; //@line 275 "fastmix.cpp"
  $58 = $57; //@line 275 "fastmix.cpp"
  $59 = $58 + 0.5; //@line 275 "fastmix.cpp"
  $60 = (+Math_floor((+$59))); //@line 275 "fastmix.cpp"
  $61 = $60; //@line 275 "fastmix.cpp"
  $62 = $61 < -32768.0; //@line 276 "fastmix.cpp"
  $63 = (~~(($61))); //@line 276 "fastmix.cpp"
  $64 = $62 ? -32768 : $63; //@line 276 "fastmix.cpp"
  $65 = $33 | 2; //@line 276 "fastmix.cpp"
  $66 = (1217698 + ($65<<1)|0); //@line 276 "fastmix.cpp"
  HEAP16[$66>>1] = $64; //@line 276 "fastmix.cpp"
  $67 = +HEAPF32[$2>>2]; //@line 275 "fastmix.cpp"
  $68 = $34 * $67; //@line 275 "fastmix.cpp"
  $69 = $68; //@line 275 "fastmix.cpp"
  $70 = $69 + 0.5; //@line 275 "fastmix.cpp"
  $71 = (+Math_floor((+$70))); //@line 275 "fastmix.cpp"
  $72 = $71; //@line 275 "fastmix.cpp"
  $73 = $72 < -32768.0; //@line 276 "fastmix.cpp"
  $74 = (~~(($72))); //@line 276 "fastmix.cpp"
  $75 = $73 ? -32768 : $74; //@line 276 "fastmix.cpp"
  $76 = $33 | 3; //@line 276 "fastmix.cpp"
  $77 = (1217698 + ($76<<1)|0); //@line 276 "fastmix.cpp"
  HEAP16[$77>>1] = $75; //@line 276 "fastmix.cpp"
  $78 = +HEAPF32[$3>>2]; //@line 275 "fastmix.cpp"
  $79 = $34 * $78; //@line 275 "fastmix.cpp"
  $80 = $79; //@line 275 "fastmix.cpp"
  $81 = $80 + 0.5; //@line 275 "fastmix.cpp"
  $82 = (+Math_floor((+$81))); //@line 275 "fastmix.cpp"
  $83 = $82; //@line 275 "fastmix.cpp"
  $84 = $83 < -32768.0; //@line 276 "fastmix.cpp"
  $85 = (~~(($83))); //@line 276 "fastmix.cpp"
  $86 = $84 ? -32768 : $85; //@line 276 "fastmix.cpp"
  $87 = $33 | 4; //@line 276 "fastmix.cpp"
  $88 = (1217698 + ($87<<1)|0); //@line 276 "fastmix.cpp"
  HEAP16[$88>>1] = $86; //@line 276 "fastmix.cpp"
  $89 = +HEAPF32[$4>>2]; //@line 275 "fastmix.cpp"
  $90 = $34 * $89; //@line 275 "fastmix.cpp"
  $91 = $90; //@line 275 "fastmix.cpp"
  $92 = $91 + 0.5; //@line 275 "fastmix.cpp"
  $93 = (+Math_floor((+$92))); //@line 275 "fastmix.cpp"
  $94 = $93; //@line 275 "fastmix.cpp"
  $95 = $94 < -32768.0; //@line 276 "fastmix.cpp"
  $96 = (~~(($94))); //@line 276 "fastmix.cpp"
  $97 = $95 ? -32768 : $96; //@line 276 "fastmix.cpp"
  $98 = $33 | 5; //@line 276 "fastmix.cpp"
  $99 = (1217698 + ($98<<1)|0); //@line 276 "fastmix.cpp"
  HEAP16[$99>>1] = $97; //@line 276 "fastmix.cpp"
  $100 = +HEAPF32[$5>>2]; //@line 275 "fastmix.cpp"
  $101 = $34 * $100; //@line 275 "fastmix.cpp"
  $102 = $101; //@line 275 "fastmix.cpp"
  $103 = $102 + 0.5; //@line 275 "fastmix.cpp"
  $104 = (+Math_floor((+$103))); //@line 275 "fastmix.cpp"
  $105 = $104; //@line 275 "fastmix.cpp"
  $106 = $105 < -32768.0; //@line 276 "fastmix.cpp"
  $107 = (~~(($105))); //@line 276 "fastmix.cpp"
  $108 = $106 ? -32768 : $107; //@line 276 "fastmix.cpp"
  $109 = $33 | 6; //@line 276 "fastmix.cpp"
  $110 = (1217698 + ($109<<1)|0); //@line 276 "fastmix.cpp"
  HEAP16[$110>>1] = $108; //@line 276 "fastmix.cpp"
  $111 = +HEAPF32[$6>>2]; //@line 275 "fastmix.cpp"
  $112 = $34 * $111; //@line 275 "fastmix.cpp"
  $113 = $112; //@line 275 "fastmix.cpp"
  $114 = $113 + 0.5; //@line 275 "fastmix.cpp"
  $115 = (+Math_floor((+$114))); //@line 275 "fastmix.cpp"
  $116 = $115; //@line 275 "fastmix.cpp"
  $117 = $116 < -32768.0; //@line 276 "fastmix.cpp"
  $118 = (~~(($116))); //@line 276 "fastmix.cpp"
  $119 = $117 ? -32768 : $118; //@line 276 "fastmix.cpp"
  $120 = $33 | 7; //@line 276 "fastmix.cpp"
  $121 = (1217698 + ($120<<1)|0); //@line 276 "fastmix.cpp"
  HEAP16[$121>>1] = $119; //@line 276 "fastmix.cpp"
  $122 = (($_LPcl$04) + 1)|0; //@line 265 "fastmix.cpp"
  $exitcond6 = ($122|0)==(2049); //@line 265 "fastmix.cpp"
  if ($exitcond6) {
   break;
  } else {
   $_LPcl$04 = $122;
  }
 }
 STACKTOP = sp;return; //@line 279 "fastmix.cpp"
}
function __Z7_muldivlll($a,$b,$c) {
 $a = $a|0;
 $b = $b|0;
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($a|0)<(0); //@line 122 "sndmix.cpp"
 $1 = $0 << 31 >> 31; //@line 122 "sndmix.cpp"
 $2 = ($b|0)<(0); //@line 122 "sndmix.cpp"
 $3 = $2 << 31 >> 31; //@line 122 "sndmix.cpp"
 $4 = (___muldi3(($b|0),($3|0),($a|0),($1|0))|0); //@line 122 "sndmix.cpp"
 $5 = tempRet0; //@line 122 "sndmix.cpp"
 $6 = ($c|0)<(0); //@line 122 "sndmix.cpp"
 $7 = $6 << 31 >> 31; //@line 122 "sndmix.cpp"
 $8 = (___udivdi3(($4|0),($5|0),($c|0),($7|0))|0); //@line 122 "sndmix.cpp"
 $9 = tempRet0; //@line 122 "sndmix.cpp"
 return ($8|0); //@line 122 "sndmix.cpp"
}
function __Z8_muldivrlll($a,$b,$c) {
 $a = $a|0;
 $b = $b|0;
 $c = $c|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($a|0)<(0); //@line 173 "sndmix.cpp"
 $1 = $0 << 31 >> 31; //@line 173 "sndmix.cpp"
 $2 = ($b|0)<(0); //@line 173 "sndmix.cpp"
 $3 = $2 << 31 >> 31; //@line 173 "sndmix.cpp"
 $4 = (___muldi3(($b|0),($3|0),($a|0),($1|0))|0); //@line 173 "sndmix.cpp"
 $5 = tempRet0; //@line 173 "sndmix.cpp"
 $6 = $c >> 1; //@line 173 "sndmix.cpp"
 $7 = ($6|0)<(0); //@line 173 "sndmix.cpp"
 $8 = $7 << 31 >> 31; //@line 173 "sndmix.cpp"
 $9 = (_i64Add(($6|0),($8|0),($4|0),($5|0))|0); //@line 173 "sndmix.cpp"
 $10 = tempRet0; //@line 173 "sndmix.cpp"
 $11 = ($c|0)<(0); //@line 173 "sndmix.cpp"
 $12 = $11 << 31 >> 31; //@line 173 "sndmix.cpp"
 $13 = (___udivdi3(($9|0),($10|0),($c|0),($12|0))|0); //@line 173 "sndmix.cpp"
 $14 = tempRet0; //@line 173 "sndmix.cpp"
 return ($13|0); //@line 173 "sndmix.cpp"
}
function __ZN10CSoundFile10InitPlayerEb($bReset) {
 $bReset = $bReset|0;
 var $$ = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[10107]|0; //@line 181 "sndmix.cpp"
 $1 = ($0>>>0)>(128); //@line 181 "sndmix.cpp"
 if ($1) {
  HEAP32[10107] = 128; //@line 181 "sndmix.cpp"
 }
 $2 = HEAP32[10109]|0; //@line 182 "sndmix.cpp"
 $3 = ($2>>>0)<(4000); //@line 182 "sndmix.cpp"
 if ($3) {
  HEAP32[10109] = 4000; //@line 182 "sndmix.cpp"
  $6 = 4000;
 } else {
  $4 = ($2>>>0)>(192000); //@line 183 "sndmix.cpp"
  if ($4) {
   HEAP32[10109] = 192000; //@line 183 "sndmix.cpp"
   $6 = 192000;
  } else {
   $6 = $2;
  }
 }
 $5 = ($6*146)|0; //@line 184 "sndmix.cpp"
 $7 = (($5>>>0) / 100000)&-1; //@line 184 "sndmix.cpp"
 $8 = ($7>>>0)<(8); //@line 185 "sndmix.cpp"
 $$ = $8 ? 8 : $7; //@line 185 "sndmix.cpp"
 HEAP32[10112] = $$; //@line 185 "sndmix.cpp"
 HEAP32[183086] = 0; //@line 186 "sndmix.cpp"
 HEAP32[183085] = 0; //@line 186 "sndmix.cpp"
 HEAP32[183088] = 0; //@line 187 "sndmix.cpp"
 HEAP32[183087] = 0; //@line 187 "sndmix.cpp"
 if (!($bReset)) {
  $9 = $bReset ? 3 : 1; //@line 193 "sndmix.cpp"
  HEAP32[183089] = $9; //@line 193 "sndmix.cpp"
  __ZN10CSoundFile13InitializeDSPEb($bReset); //@line 194 "sndmix.cpp"
  return 1; //@line 195 "sndmix.cpp"
 }
 HEAP32[183081] = 0; //@line 190 "sndmix.cpp"
 HEAP32[183082] = 0; //@line 191 "sndmix.cpp"
 $9 = $bReset ? 3 : 1; //@line 193 "sndmix.cpp"
 HEAP32[183089] = $9; //@line 193 "sndmix.cpp"
 __ZN10CSoundFile13InitializeDSPEb($bReset); //@line 194 "sndmix.cpp"
 return 1; //@line 195 "sndmix.cpp"
}
function __ZN10CSoundFile13InitializeDSPEb($bReset) {
 $bReset = $bReset|0;
 var $$ = 0, $$1 = 0, $$2 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $bReset$not = 0, $bReset$not9 = 0, $n$0$i = 0, $n$0$i$lcssa = 0, $or$cond = 0, $or$cond11 = 0, $or$cond8 = 0, $storemerge = 0, $storemerge13 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[10116]|0; //@line 118 "snd_dsp.cpp"
 $1 = ($0|0)==(0); //@line 118 "snd_dsp.cpp"
 if ($1) {
  HEAP32[10116] = 100; //@line 118 "snd_dsp.cpp"
  $30 = 100;
 } else {
  $30 = $0;
 }
 $2 = HEAP32[10114]|0; //@line 119 "snd_dsp.cpp"
 $3 = ($2|0)==(0); //@line 119 "snd_dsp.cpp"
 if ($3) {
  HEAP32[10114] = 14; //@line 119 "snd_dsp.cpp"
  $51 = 14;
 } else {
  $51 = $2;
 }
 $4 = HEAP32[10118]|0; //@line 120 "snd_dsp.cpp"
 $5 = ($4|0)==(0); //@line 120 "snd_dsp.cpp"
 if ($5) {
  HEAP32[10118] = 20; //@line 120 "snd_dsp.cpp"
  $14 = 20;
 } else {
  $14 = $4;
 }
 $6 = HEAP32[10113]|0; //@line 121 "snd_dsp.cpp"
 $7 = ($6>>>0)>(8); //@line 121 "snd_dsp.cpp"
 if ($7) {
  HEAP32[10113] = 8; //@line 121 "snd_dsp.cpp"
 } else {
  $8 = ($6>>>0)<(2); //@line 122 "snd_dsp.cpp"
  if ($8) {
   HEAP32[10113] = 2; //@line 122 "snd_dsp.cpp"
  }
 }
 if ($bReset) {
  HEAP32[187188] = 0; //@line 126 "snd_dsp.cpp"
  HEAP32[187189] = 0; //@line 126 "snd_dsp.cpp"
 }
 HEAP32[187190] = 0; //@line 129 "snd_dsp.cpp"
 HEAP32[187191] = 0; //@line 129 "snd_dsp.cpp"
 HEAP32[187192] = 0; //@line 130 "snd_dsp.cpp"
 HEAP32[187193] = 0; //@line 130 "snd_dsp.cpp"
 HEAP32[187194] = 0; //@line 130 "snd_dsp.cpp"
 HEAP32[187195] = 0; //@line 131 "snd_dsp.cpp"
 HEAP32[187196] = 0; //@line 131 "snd_dsp.cpp"
 $9 = HEAP32[183080]|0; //@line 132 "snd_dsp.cpp"
 $10 = $9 & 64; //@line 132 "snd_dsp.cpp"
 $11 = ($10|0)==(0); //@line 132 "snd_dsp.cpp"
 if (!($11)) {
  _memset((748788|0),0,256)|0; //@line 134 "snd_dsp.cpp"
  _memset((749044|0),0,256)|0; //@line 135 "snd_dsp.cpp"
  _memset((749300|0),0,256)|0; //@line 136 "snd_dsp.cpp"
  _memset((749556|0),0,38400)|0; //@line 137 "snd_dsp.cpp"
  $12 = HEAP32[10109]|0; //@line 138 "snd_dsp.cpp"
  $13 = Math_imul($14, $12)|0; //@line 138 "snd_dsp.cpp"
  $15 = (($13>>>0) / 1000)&-1; //@line 138 "snd_dsp.cpp"
  $16 = ($15>>>0)>(9600); //@line 139 "snd_dsp.cpp"
  $$ = $16 ? 9600 : $15; //@line 139 "snd_dsp.cpp"
  HEAP32[187190] = $$; //@line 139 "snd_dsp.cpp"
  $17 = HEAP32[10117]|0; //@line 140 "snd_dsp.cpp"
  $18 = ($17>>>0)<(8); //@line 140 "snd_dsp.cpp"
  if ($18) {
   $19 = 32 >>> $17; //@line 140 "snd_dsp.cpp"
   $20 = (($19) + 32)|0; //@line 140 "snd_dsp.cpp"
   $storemerge = $20;
  } else {
   $21 = ($17>>>0)<(16); //@line 141 "snd_dsp.cpp"
   $22 = ($17*7)|0; //@line 141 "snd_dsp.cpp"
   $23 = (($22) + -48)|0; //@line 141 "snd_dsp.cpp"
   $24 = $21 ? $23 : 64; //@line 141 "snd_dsp.cpp"
   $storemerge = $24;
  }
  $25 = $storemerge >> 2; //@line 142 "snd_dsp.cpp"
  HEAP32[196989] = $25; //@line 142 "snd_dsp.cpp"
 }
 $26 = $9 & 128; //@line 146 "snd_dsp.cpp"
 $27 = ($26|0)==(0); //@line 146 "snd_dsp.cpp"
 if ($27) {
  HEAP32[196990] = 0; //@line 172 "snd_dsp.cpp"
 } else {
  $28 = HEAP32[10109]|0; //@line 148 "snd_dsp.cpp"
  $29 = Math_imul($30, $28)|0; //@line 148 "snd_dsp.cpp"
  $31 = (($29>>>0) / 1000)&-1; //@line 148 "snd_dsp.cpp"
  $32 = HEAP32[10115]|0; //@line 149 "snd_dsp.cpp"
  $33 = (($32) + 1)|0; //@line 149 "snd_dsp.cpp"
  $34 = ($29>>>0)>(38400999); //@line 150 "snd_dsp.cpp"
  $$5 = $34 ? 38400 : $31; //@line 150 "snd_dsp.cpp"
  $bReset$not = $bReset ^ 1; //@line 151 "snd_dsp.cpp"
  $35 = HEAP32[196990]|0;
  $36 = ($$5|0)==($35|0); //@line 151 "snd_dsp.cpp"
  $or$cond = $36 & $bReset$not; //@line 151 "snd_dsp.cpp"
  $37 = HEAP32[196991]|0;
  $38 = ($33|0)==($37|0); //@line 151 "snd_dsp.cpp"
  $or$cond8 = $or$cond & $38; //@line 151 "snd_dsp.cpp"
  if (!($or$cond8)) {
   HEAP32[196991] = $33; //@line 153 "snd_dsp.cpp"
   HEAP32[196990] = $$5; //@line 154 "snd_dsp.cpp"
   HEAP32[196992] = 0; //@line 155 "snd_dsp.cpp"
   HEAP32[196993] = 0; //@line 155 "snd_dsp.cpp"
   HEAP32[196994] = 0; //@line 155 "snd_dsp.cpp"
   HEAP32[196995] = 0; //@line 155 "snd_dsp.cpp"
   HEAP32[196996] = 0; //@line 156 "snd_dsp.cpp"
   HEAP32[196997] = 0; //@line 156 "snd_dsp.cpp"
   HEAP32[196998] = 0; //@line 156 "snd_dsp.cpp"
   HEAP32[196999] = 0; //@line 157 "snd_dsp.cpp"
   HEAP32[197000] = 0; //@line 157 "snd_dsp.cpp"
   $39 = ($$5*13)|0; //@line 158 "snd_dsp.cpp"
   $40 = (($39>>>0) / 17)&-1; //@line 158 "snd_dsp.cpp"
   $41 = ($40>>>0)>(29364); //@line 159 "snd_dsp.cpp"
   $$1 = $41 ? 29364 : $40; //@line 159 "snd_dsp.cpp"
   HEAP32[197001] = $$1; //@line 159 "snd_dsp.cpp"
   $42 = ($$5*7)|0; //@line 160 "snd_dsp.cpp"
   $43 = (($42>>>0) / 13)&-1; //@line 160 "snd_dsp.cpp"
   $44 = ($43>>>0)>(20676); //@line 161 "snd_dsp.cpp"
   $$4 = $44 ? 20676 : $43; //@line 161 "snd_dsp.cpp"
   HEAP32[197002] = $$4; //@line 161 "snd_dsp.cpp"
   $45 = (($42>>>0) / 19)&-1; //@line 162 "snd_dsp.cpp"
   $46 = ($45>>>0)>(14147); //@line 163 "snd_dsp.cpp"
   $$2 = $46 ? 14147 : $45; //@line 163 "snd_dsp.cpp"
   HEAP32[197003] = $$2; //@line 163 "snd_dsp.cpp"
   _memset((788016|0),0,256)|0; //@line 164 "snd_dsp.cpp"
   _memset((788272|0),0,256)|0; //@line 165 "snd_dsp.cpp"
   _memset((788528|0),0,153600)|0; //@line 166 "snd_dsp.cpp"
   _memset((942128|0),0,117456)|0; //@line 167 "snd_dsp.cpp"
   _memset((1059584|0),0,82704)|0; //@line 168 "snd_dsp.cpp"
   _memset((1142288|0),0,56588)|0; //@line 169 "snd_dsp.cpp"
   ;HEAP32[1198876>>2]=0|0;HEAP32[1198876+4>>2]=0|0;HEAP32[1198876+8>>2]=0|0;HEAP32[1198876+12>>2]=0|0;HEAP32[1198876+16>>2]=0|0;HEAP32[1198876+20>>2]=0|0;HEAP32[1198876+24>>2]=0|0;HEAP32[1198876+28>>2]=0|0; //@line 170 "snd_dsp.cpp"
  }
 }
 $47 = $9 & 32; //@line 176 "snd_dsp.cpp"
 $48 = ($47|0)==(0); //@line 176 "snd_dsp.cpp"
 if ($48) {
  $storemerge13 = 0;
 } else {
  $49 = HEAP32[10109]|0; //@line 178 "snd_dsp.cpp"
  $50 = Math_imul($51, $49)|0; //@line 178 "snd_dsp.cpp"
  $52 = (($50>>>0) / 10000)&-1; //@line 178 "snd_dsp.cpp"
  $53 = ($52>>>0)>(64); //@line 179 "snd_dsp.cpp"
  $$3 = $53 ? 64 : $52; //@line 179 "snd_dsp.cpp"
  $n$0$i = 2;
  while(1) {
   $54 = ($n$0$i>>>0)>($$3>>>0); //@line 110 "snd_dsp.cpp"
   $55 = $n$0$i << 1; //@line 110 "snd_dsp.cpp"
   if ($54) {
    $n$0$i$lcssa = $n$0$i;
    break;
   } else {
    $n$0$i = $55;
   }
  }
  $56 = $n$0$i$lcssa >>> 1; //@line 111 "snd_dsp.cpp"
  $57 = (($56) + -1)|0; //@line 111 "snd_dsp.cpp"
  $bReset$not9 = $bReset ^ 1; //@line 181 "snd_dsp.cpp"
  $58 = HEAP32[299727]|0;
  $59 = ($57|0)==($58|0); //@line 181 "snd_dsp.cpp"
  $or$cond11 = $59 & $bReset$not9; //@line 181 "snd_dsp.cpp"
  if ($or$cond11) {
   return; //@line 197 "snd_dsp.cpp"
  } else {
   $storemerge13 = $57;
  }
 }
 HEAP32[299727] = $storemerge13; //@line 188 "snd_dsp.cpp"
 HEAP32[299728] = 0; //@line 193 "snd_dsp.cpp"
 HEAP32[299729] = 0; //@line 193 "snd_dsp.cpp"
 HEAP32[299730] = 0; //@line 193 "snd_dsp.cpp"
 _memset((1198924|0),0,256)|0; //@line 194 "snd_dsp.cpp"
 _memset((1199180|0),0,256)|0; //@line 195 "snd_dsp.cpp"
 return; //@line 197 "snd_dsp.cpp"
}
function __ZN10CSoundFile14GlobalFadeSongEj($this,$msec) {
 $this = $this|0;
 $msec = $msec|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 59660|0); //@line 228 "sndmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 228 "sndmix.cpp"
 $2 = $1 & 1024; //@line 228 "sndmix.cpp"
 $3 = ($2|0)==(0); //@line 228 "sndmix.cpp"
 if (!($3)) {
  $$0 = 0;
  return ($$0|0); //@line 233 "sndmix.cpp"
 }
 $4 = HEAP32[10109]|0; //@line 229 "sndmix.cpp"
 $5 = ($msec|0)<(0); //@line 122 "sndmix.cpp"
 $6 = $5 << 31 >> 31; //@line 122 "sndmix.cpp"
 $7 = ($4|0)<(0); //@line 122 "sndmix.cpp"
 $8 = $7 << 31 >> 31; //@line 122 "sndmix.cpp"
 $9 = (___muldi3(($4|0),($8|0),($msec|0),($6|0))|0); //@line 122 "sndmix.cpp"
 $10 = tempRet0; //@line 122 "sndmix.cpp"
 $11 = (___udivdi3(($9|0),($10|0),1000,0)|0); //@line 122 "sndmix.cpp"
 $12 = tempRet0; //@line 122 "sndmix.cpp"
 $13 = ((($this)) + 59784|0); //@line 229 "sndmix.cpp"
 HEAP32[$13>>2] = $11; //@line 229 "sndmix.cpp"
 $14 = ((($this)) + 59780|0); //@line 230 "sndmix.cpp"
 HEAP32[$14>>2] = $11; //@line 230 "sndmix.cpp"
 $15 = $1 | 1024; //@line 231 "sndmix.cpp"
 HEAP32[$0>>2] = $15; //@line 231 "sndmix.cpp"
 $$0 = 1;
 return ($$0|0); //@line 233 "sndmix.cpp"
}
function __ZN10CSoundFile4ReadEPvj($this,$lpDestBuffer,$cbBuffer) {
 $this = $this|0;
 $lpDestBuffer = $lpDestBuffer|0;
 $cbBuffer = $cbBuffer|0;
 var $$ = 0, $$$i = 0, $$0 = 0, $$2 = 0, $$5 = 0, $$old3 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $lRead$0 = 0;
 var $lRead$0$$ = 0, $lRead$1 = 0, $lSampleCount$0 = 0, $lSampleSize$0 = 0, $lTotalSampleCount$0 = 0, $lpBuffer$0 = 0, $lpBuffer$1 = 0, $nStat$0 = 0, $nStat$1 = 0, $nVUMeterMax = 0, $nVUMeterMin = 0, $noff$01$i = 0, $or$cond = 0, $pCvt$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $nVUMeterMin = sp + 4|0;
 $nVUMeterMax = sp;
 HEAP32[$nVUMeterMin>>2] = 2147483647; //@line 242 "sndmix.cpp"
 HEAP32[$nVUMeterMax>>2] = -2147483647; //@line 242 "sndmix.cpp"
 $0 = ((($this)) + 59672|0); //@line 249 "sndmix.cpp"
 HEAP32[$0>>2] = 0; //@line 249 "sndmix.cpp"
 $1 = HEAP32[10108]|0; //@line 250 "sndmix.cpp"
 $2 = HEAP32[10110]|0; //@line 251 "sndmix.cpp"
 switch ($2|0) {
 case 16:  {
  $3 = $1 << 1; //@line 251 "sndmix.cpp"
  $lSampleSize$0 = $3;$pCvt$0 = 4327;
  break;
 }
 case 24:  {
  $4 = ($1*3)|0; //@line 253 "sndmix.cpp"
  $lSampleSize$0 = $4;$pCvt$0 = 4328;
  break;
 }
 case 32:  {
  $5 = $1 << 2; //@line 254 "sndmix.cpp"
  $lSampleSize$0 = $5;$pCvt$0 = 4329;
  break;
 }
 default: {
  $lSampleSize$0 = $1;$pCvt$0 = 4326;
 }
 }
 $6 = (($cbBuffer>>>0) / ($lSampleSize$0>>>0))&-1; //@line 256 "sndmix.cpp"
 $7 = ($6|0)!=(0); //@line 257 "sndmix.cpp"
 $8 = ($lpDestBuffer|0)!=(0|0); //@line 257 "sndmix.cpp"
 $or$cond = $8 & $7; //@line 257 "sndmix.cpp"
 if (!($or$cond)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 340 "sndmix.cpp"
 }
 $9 = ((($this)) + 59664|0); //@line 257 "sndmix.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 257 "sndmix.cpp"
 $11 = ($10|0)==(0); //@line 257 "sndmix.cpp"
 if ($11) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 340 "sndmix.cpp"
 }
 $12 = ((($this)) + 59660|0); //@line 259 "sndmix.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 259 "sndmix.cpp"
 $14 = $13 & 512; //@line 259 "sndmix.cpp"
 $15 = ($14|0)==(0); //@line 259 "sndmix.cpp"
 L12: do {
  if ($15) {
   $16 = ((($this)) + 59676|0); //@line 263 "sndmix.cpp"
   $17 = ((($this)) + 59668|0); //@line 208 "sndmix.cpp"
   $$pre = HEAP32[$16>>2]|0; //@line 263 "sndmix.cpp"
   $19 = $$pre;$lRead$0 = $6;$lpBuffer$0 = $lpDestBuffer;$nStat$0 = 0;
   while(1) {
    $18 = ($19|0)==(0); //@line 263 "sndmix.cpp"
    do {
     if ($18) {
      $20 = HEAP32[$12>>2]|0; //@line 266 "sndmix.cpp"
      $21 = $20 & 256; //@line 266 "sndmix.cpp"
      $22 = ($21|0)==(0); //@line 266 "sndmix.cpp"
      if (!($22)) {
       $23 = $20 | 512; //@line 268 "sndmix.cpp"
       HEAP32[$12>>2] = $23; //@line 268 "sndmix.cpp"
       HEAP32[$16>>2] = $lRead$0; //@line 269 "sndmix.cpp"
       break;
      }
      $24 = (__ZN10CSoundFile8ReadNoteEv($this)|0); //@line 272 "sndmix.cpp"
      if (!($24)) {
       $25 = HEAP32[10109]|0; //@line 202 "sndmix.cpp"
       $26 = ($25|0)<(0); //@line 122 "sndmix.cpp"
       $27 = $26 << 31 >> 31; //@line 122 "sndmix.cpp"
       $28 = (___muldi3(($25|0),($27|0),100,0)|0); //@line 122 "sndmix.cpp"
       $29 = tempRet0; //@line 122 "sndmix.cpp"
       $30 = (___udivdi3(($28|0),($29|0),1000,0)|0); //@line 122 "sndmix.cpp"
       $31 = tempRet0; //@line 122 "sndmix.cpp"
       $32 = ($30|0)<(1); //@line 203 "sndmix.cpp"
       if ($32) {
        $65 = HEAP32[$12>>2]|0; //@line 278 "sndmix.cpp"
        $66 = $65 | 512; //@line 278 "sndmix.cpp"
        HEAP32[$12>>2] = $66; //@line 278 "sndmix.cpp"
        $67 = ($lRead$0|0)==($6|0); //@line 279 "sndmix.cpp"
        if ($67) {
         $lRead$1 = $6;$lpBuffer$1 = $lpBuffer$0;$nStat$1 = $nStat$0;
         break L12;
        }
        HEAP32[$16>>2] = $lRead$0; //@line 280 "sndmix.cpp"
        break;
       }
       $33 = ($30|0)>(1048576); //@line 204 "sndmix.cpp"
       $$$i = $33 ? 1048576 : $30; //@line 204 "sndmix.cpp"
       HEAP32[$16>>2] = $$$i; //@line 205 "sndmix.cpp"
       $34 = HEAP32[$17>>2]|0; //@line 208 "sndmix.cpp"
       $35 = ($34|0)==(0); //@line 208 "sndmix.cpp"
       if (!($35)) {
        $noff$01$i = 0;
        while(1) {
         $36 = (((($this)) + 32768|0) + ($noff$01$i<<2)|0); //@line 210 "sndmix.cpp"
         $37 = HEAP32[$36>>2]|0; //@line 210 "sndmix.cpp"
         $38 = (($this) + ($37<<8)|0); //@line 210 "sndmix.cpp"
         $39 = ($38|0)==(0|0); //@line 211 "sndmix.cpp"
         if (!($39)) {
          $40 = (((($this) + ($37<<8)|0)) + 100|0); //@line 212 "sndmix.cpp"
          HEAP32[$40>>2] = 0; //@line 212 "sndmix.cpp"
          $41 = (((($this) + ($37<<8)|0)) + 104|0); //@line 212 "sndmix.cpp"
          HEAP32[$41>>2] = 0; //@line 212 "sndmix.cpp"
          $42 = (((($this) + ($37<<8)|0)) + 16|0); //@line 213 "sndmix.cpp"
          $43 = HEAP32[$42>>2]|0; //@line 213 "sndmix.cpp"
          $44 = (0 - ($43))|0; //@line 213 "sndmix.cpp"
          $45 = $44 << 12; //@line 213 "sndmix.cpp"
          $46 = (($45|0) / ($$$i|0))&-1; //@line 213 "sndmix.cpp"
          $47 = (((($this) + ($37<<8)|0)) + 24|0); //@line 213 "sndmix.cpp"
          HEAP32[$47>>2] = $46; //@line 213 "sndmix.cpp"
          $48 = (((($this) + ($37<<8)|0)) + 20|0); //@line 214 "sndmix.cpp"
          $49 = HEAP32[$48>>2]|0; //@line 214 "sndmix.cpp"
          $50 = (0 - ($49))|0; //@line 214 "sndmix.cpp"
          $51 = $50 << 12; //@line 214 "sndmix.cpp"
          $52 = (($51|0) / ($$$i|0))&-1; //@line 214 "sndmix.cpp"
          $53 = (((($this) + ($37<<8)|0)) + 28|0); //@line 214 "sndmix.cpp"
          HEAP32[$53>>2] = $52; //@line 214 "sndmix.cpp"
          $54 = $43 << 12; //@line 215 "sndmix.cpp"
          $55 = (((($this) + ($37<<8)|0)) + 48|0); //@line 215 "sndmix.cpp"
          HEAP32[$55>>2] = $54; //@line 215 "sndmix.cpp"
          $56 = $49 << 12; //@line 216 "sndmix.cpp"
          $57 = (((($this) + ($37<<8)|0)) + 52|0); //@line 216 "sndmix.cpp"
          HEAP32[$57>>2] = $56; //@line 216 "sndmix.cpp"
          $58 = (((($this) + ($37<<8)|0)) + 92|0); //@line 217 "sndmix.cpp"
          HEAP32[$58>>2] = $$$i; //@line 217 "sndmix.cpp"
          $59 = (((($this) + ($37<<8)|0)) + 36|0); //@line 218 "sndmix.cpp"
          $60 = HEAP32[$59>>2]|0; //@line 218 "sndmix.cpp"
          $61 = $60 | 32768; //@line 218 "sndmix.cpp"
          HEAP32[$59>>2] = $61; //@line 218 "sndmix.cpp"
         }
         $62 = (($noff$01$i) + 1)|0; //@line 208 "sndmix.cpp"
         $exitcond = ($62|0)==($34|0); //@line 208 "sndmix.cpp"
         if ($exitcond) {
          break;
         } else {
          $noff$01$i = $62;
         }
        }
       }
       $63 = HEAP32[$12>>2]|0; //@line 220 "sndmix.cpp"
       $64 = $63 | 256; //@line 220 "sndmix.cpp"
       HEAP32[$12>>2] = $64; //@line 220 "sndmix.cpp"
      }
     }
    } while(0);
    $68 = HEAP32[$16>>2]|0; //@line 284 "sndmix.cpp"
    $69 = ($68>>>0)>(512); //@line 285 "sndmix.cpp"
    $$ = $69 ? 512 : $68; //@line 285 "sndmix.cpp"
    $70 = ($$>>>0)>($lRead$0>>>0); //@line 286 "sndmix.cpp"
    $lRead$0$$ = $70 ? $lRead$0 : $$; //@line 286 "sndmix.cpp"
    $71 = ($lRead$0$$|0)==(0); //@line 287 "sndmix.cpp"
    if ($71) {
     $lRead$1 = $lRead$0;$lpBuffer$1 = $lpBuffer$0;$nStat$1 = $nStat$0;
     break L12;
    }
    HEAP32[183090] = 0; //@line 290 "sndmix.cpp"
    __Z14X86_StereoFillPijS_S_(732364,$lRead$0$$,732340,732344); //@line 293 "sndmix.cpp"
    $72 = HEAP32[10108]|0; //@line 294 "sndmix.cpp"
    $73 = ($72>>>0)>(1); //@line 294 "sndmix.cpp"
    if ($73) {
     $74 = $lRead$0$$ << 1; //@line 296 "sndmix.cpp"
     $75 = (__ZN10CSoundFile15CreateStereoMixEi($this,$lRead$0$$)|0); //@line 297 "sndmix.cpp"
     $76 = HEAP32[$0>>2]|0; //@line 297 "sndmix.cpp"
     $77 = (($76) + ($75))|0; //@line 297 "sndmix.cpp"
     HEAP32[$0>>2] = $77; //@line 297 "sndmix.cpp"
     __ZN10CSoundFile16ProcessStereoDSPEi($lRead$0$$); //@line 298 "sndmix.cpp"
     $lSampleCount$0 = $74;
    } else {
     $78 = (__ZN10CSoundFile15CreateStereoMixEi($this,$lRead$0$$)|0); //@line 301 "sndmix.cpp"
     $79 = HEAP32[$0>>2]|0; //@line 301 "sndmix.cpp"
     $80 = (($79) + ($78))|0; //@line 301 "sndmix.cpp"
     HEAP32[$0>>2] = $80; //@line 301 "sndmix.cpp"
     __ZN10CSoundFile16ProcessStereoDSPEi($lRead$0$$); //@line 303 "sndmix.cpp"
     __Z18X86_MonoFromStereoPij(732364,$lRead$0$$); //@line 304 "sndmix.cpp"
     $lSampleCount$0 = $lRead$0$$;
    }
    $81 = (($nStat$0) + 1)|0; //@line 306 "sndmix.cpp"
    $82 = HEAP32[10108]|0; //@line 314 "sndmix.cpp"
    $83 = ($82>>>0)>(2); //@line 314 "sndmix.cpp"
    if ($83) {
     $84 = $lSampleCount$0 << 1; //@line 317 "sndmix.cpp"
     $lTotalSampleCount$0 = $84;
    } else {
     $lTotalSampleCount$0 = $lSampleCount$0;
    }
    $85 = HEAP32[183083]|0; //@line 320 "sndmix.cpp"
    $86 = ($85|0)==(0|0); //@line 320 "sndmix.cpp"
    if (!($86)) {
     $87 = HEAP32[10108]|0; //@line 322 "sndmix.cpp"
     FUNCTION_TABLE_viii[$85 & 8191](732364,$lTotalSampleCount$0,$87); //@line 322 "sndmix.cpp"
    }
    $88 = (FUNCTION_TABLE_iiiiii[$pCvt$0 & 8191]($lpBuffer$0,732364,$lTotalSampleCount$0,$nVUMeterMin,$nVUMeterMax)|0); //@line 326 "sndmix.cpp"
    $89 = (($lpBuffer$0) + ($88)|0); //@line 326 "sndmix.cpp"
    $90 = (($lRead$0) - ($lRead$0$$))|0; //@line 328 "sndmix.cpp"
    $91 = HEAP32[$16>>2]|0; //@line 329 "sndmix.cpp"
    $92 = (($91) - ($lRead$0$$))|0; //@line 329 "sndmix.cpp"
    HEAP32[$16>>2] = $92; //@line 329 "sndmix.cpp"
    $$old3 = ($lRead$0|0)==($lRead$0$$|0); //@line 260 "sndmix.cpp"
    if ($$old3) {
     $lRead$1 = $90;$lpBuffer$1 = $89;$nStat$1 = $81;
     break;
    } else {
     $19 = $92;$lRead$0 = $90;$lpBuffer$0 = $89;$nStat$0 = $81;
    }
   }
  } else {
   $lRead$1 = $6;$lpBuffer$1 = $lpDestBuffer;$nStat$1 = 0;
  }
 } while(0);
 $93 = ($lRead$1|0)==(0); //@line 332 "sndmix.cpp"
 if (!($93)) {
  $94 = HEAP32[10110]|0; //@line 332 "sndmix.cpp"
  $95 = ($94|0)==(8); //@line 332 "sndmix.cpp"
  $96 = $95 ? -128 : 0; //@line 332 "sndmix.cpp"
  $97 = Math_imul($lRead$1, $lSampleSize$0)|0; //@line 332 "sndmix.cpp"
  _memset(($lpBuffer$1|0),($96|0),($97|0))|0; //@line 332 "sndmix.cpp"
 }
 $98 = HEAP32[$nVUMeterMin>>2]|0; //@line 334 "sndmix.cpp"
 $99 = $98 >> 20; //@line 334 "sndmix.cpp"
 HEAP32[$nVUMeterMin>>2] = $99; //@line 334 "sndmix.cpp"
 $100 = HEAP32[$nVUMeterMax>>2]|0; //@line 335 "sndmix.cpp"
 $101 = $100 >> 20; //@line 335 "sndmix.cpp"
 $102 = ($101|0)<($99|0); //@line 336 "sndmix.cpp"
 $$5 = $102 ? $99 : $101; //@line 336 "sndmix.cpp"
 HEAP32[$nVUMeterMax>>2] = $$5; //@line 335 "sndmix.cpp"
 $103 = (($$5) - ($99))|0; //@line 337 "sndmix.cpp"
 $104 = ($103>>>0)>(255); //@line 337 "sndmix.cpp"
 $$2 = $104 ? 255 : $103; //@line 337 "sndmix.cpp"
 HEAP32[183081] = $$2; //@line 337 "sndmix.cpp"
 $105 = ($nStat$1|0)==(0); //@line 338 "sndmix.cpp"
 if (!($105)) {
  $106 = (($nStat$1) + -1)|0; //@line 338 "sndmix.cpp"
  $107 = HEAP32[$0>>2]|0; //@line 338 "sndmix.cpp"
  $108 = (($106) + ($107))|0; //@line 338 "sndmix.cpp"
  $109 = (($108>>>0) / ($nStat$1>>>0))&-1; //@line 338 "sndmix.cpp"
  HEAP32[$0>>2] = $109; //@line 338 "sndmix.cpp"
 }
 $110 = (($6) - ($lRead$1))|0; //@line 339 "sndmix.cpp"
 $$0 = $110;
 STACKTOP = sp;return ($$0|0); //@line 340 "sndmix.cpp"
}
function __Z16X86_Convert32To8PvPijS0_S0_($lp8,$pBuffer,$lSampleCount,$lpMin,$lpMax) {
 $lp8 = $lp8|0;
 $pBuffer = $pBuffer|0;
 $lSampleCount = $lSampleCount|0;
 $lpMin = $lpMin|0;
 $lpMax = $lpMax|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$03 = 0, $n$0 = 0, $n$0$vumax$0 = 0, $vumax$0$lcssa = 0;
 var $vumax$01 = 0, $vumax$1 = 0, $vumin$0$lcssa = 0, $vumin$02 = 0, $vumin$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$lpMin>>2]|0; //@line 1682 "fastmix.cpp"
 $1 = HEAP32[$lpMax>>2]|0; //@line 1682 "fastmix.cpp"
 $2 = ($lSampleCount|0)==(0); //@line 1684 "fastmix.cpp"
 if ($2) {
  $vumax$0$lcssa = $1;$vumin$0$lcssa = $0;
  HEAP32[$lpMin>>2] = $vumin$0$lcssa; //@line 1697 "fastmix.cpp"
  HEAP32[$lpMax>>2] = $vumax$0$lcssa; //@line 1698 "fastmix.cpp"
  return ($lSampleCount|0); //@line 1699 "fastmix.cpp"
 } else {
  $i$03 = 0;$vumax$01 = $1;$vumin$02 = $0;
 }
 while(1) {
  $3 = (($pBuffer) + ($i$03<<2)|0); //@line 1686 "fastmix.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 1686 "fastmix.cpp"
  $5 = ($4|0)<(-134217728); //@line 1687 "fastmix.cpp"
  $6 = ($4|0)>(134217727); //@line 1689 "fastmix.cpp"
  $$ = $6 ? 134217727 : $4; //@line 1689 "fastmix.cpp"
  $n$0 = $5 ? -134217728 : $$; //@line 1687 "fastmix.cpp"
  $7 = ($n$0|0)<($vumin$02|0); //@line 1691 "fastmix.cpp"
  $8 = ($n$0|0)>($vumax$01|0); //@line 1693 "fastmix.cpp"
  $n$0$vumax$0 = $8 ? $n$0 : $vumax$01; //@line 1693 "fastmix.cpp"
  $vumax$1 = $7 ? $vumax$01 : $n$0$vumax$0; //@line 1691 "fastmix.cpp"
  $vumin$1 = $7 ? $n$0 : $vumin$02; //@line 1691 "fastmix.cpp"
  $9 = $n$0 >>> 20; //@line 1695 "fastmix.cpp"
  $10 = $9 ^ 128; //@line 1695 "fastmix.cpp"
  $11 = $10&255; //@line 1695 "fastmix.cpp"
  $12 = (($lp8) + ($i$03)|0); //@line 1695 "fastmix.cpp"
  HEAP8[$12>>0] = $11; //@line 1695 "fastmix.cpp"
  $13 = (($i$03) + 1)|0; //@line 1684 "fastmix.cpp"
  $exitcond = ($13|0)==($lSampleCount|0); //@line 1684 "fastmix.cpp"
  if ($exitcond) {
   $vumax$0$lcssa = $vumax$1;$vumin$0$lcssa = $vumin$1;
   break;
  } else {
   $i$03 = $13;$vumax$01 = $vumax$1;$vumin$02 = $vumin$1;
  }
 }
 HEAP32[$lpMin>>2] = $vumin$0$lcssa; //@line 1697 "fastmix.cpp"
 HEAP32[$lpMax>>2] = $vumax$0$lcssa; //@line 1698 "fastmix.cpp"
 return ($lSampleCount|0); //@line 1699 "fastmix.cpp"
}
function __Z17X86_Convert32To16PvPijS0_S0_($lp16,$pBuffer,$lSampleCount,$lpMin,$lpMax) {
 $lp16 = $lp16|0;
 $pBuffer = $pBuffer|0;
 $lSampleCount = $lSampleCount|0;
 $lpMin = $lpMin|0;
 $lpMax = $lpMax|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$03 = 0, $n$0 = 0, $n$0$vumax$0 = 0, $vumax$0$lcssa = 0;
 var $vumax$01 = 0, $vumax$1 = 0, $vumin$0$lcssa = 0, $vumin$02 = 0, $vumin$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$lpMin>>2]|0; //@line 1776 "fastmix.cpp"
 $1 = HEAP32[$lpMax>>2]|0; //@line 1776 "fastmix.cpp"
 $2 = ($lSampleCount|0)==(0); //@line 1778 "fastmix.cpp"
 if ($2) {
  $vumax$0$lcssa = $1;$vumin$0$lcssa = $0;
  HEAP32[$lpMin>>2] = $vumin$0$lcssa; //@line 1791 "fastmix.cpp"
  HEAP32[$lpMax>>2] = $vumax$0$lcssa; //@line 1792 "fastmix.cpp"
  $13 = $lSampleCount << 1; //@line 1793 "fastmix.cpp"
  return ($13|0); //@line 1793 "fastmix.cpp"
 } else {
  $i$03 = 0;$vumax$01 = $1;$vumin$02 = $0;
 }
 while(1) {
  $3 = (($pBuffer) + ($i$03<<2)|0); //@line 1780 "fastmix.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 1780 "fastmix.cpp"
  $5 = ($4|0)<(-134217728); //@line 1781 "fastmix.cpp"
  $6 = ($4|0)>(134217727); //@line 1783 "fastmix.cpp"
  $$ = $6 ? 134217727 : $4; //@line 1783 "fastmix.cpp"
  $n$0 = $5 ? -134217728 : $$; //@line 1781 "fastmix.cpp"
  $7 = ($n$0|0)<($vumin$02|0); //@line 1785 "fastmix.cpp"
  $8 = ($n$0|0)>($vumax$01|0); //@line 1787 "fastmix.cpp"
  $n$0$vumax$0 = $8 ? $n$0 : $vumax$01; //@line 1787 "fastmix.cpp"
  $vumax$1 = $7 ? $vumax$01 : $n$0$vumax$0; //@line 1785 "fastmix.cpp"
  $vumin$1 = $7 ? $n$0 : $vumin$02; //@line 1785 "fastmix.cpp"
  $9 = $n$0 >>> 12; //@line 1789 "fastmix.cpp"
  $10 = $9&65535; //@line 1789 "fastmix.cpp"
  $11 = (($lp16) + ($i$03<<1)|0); //@line 1789 "fastmix.cpp"
  HEAP16[$11>>1] = $10; //@line 1789 "fastmix.cpp"
  $12 = (($i$03) + 1)|0; //@line 1778 "fastmix.cpp"
  $exitcond = ($12|0)==($lSampleCount|0); //@line 1778 "fastmix.cpp"
  if ($exitcond) {
   $vumax$0$lcssa = $vumax$1;$vumin$0$lcssa = $vumin$1;
   break;
  } else {
   $i$03 = $12;$vumax$01 = $vumax$1;$vumin$02 = $vumin$1;
  }
 }
 HEAP32[$lpMin>>2] = $vumin$0$lcssa; //@line 1791 "fastmix.cpp"
 HEAP32[$lpMax>>2] = $vumax$0$lcssa; //@line 1792 "fastmix.cpp"
 $13 = $lSampleCount << 1; //@line 1793 "fastmix.cpp"
 return ($13|0); //@line 1793 "fastmix.cpp"
}
function __Z17X86_Convert32To24PvPijS0_S0_($lp16,$pBuffer,$lSampleCount,$lpMin,$lpMax) {
 $lp16 = $lp16|0;
 $pBuffer = $pBuffer|0;
 $lSampleCount = $lSampleCount|0;
 $lpMin = $lpMin|0;
 $lpMax = $lpMax|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $exitcond = 0, $i$03 = 0, $n$0 = 0, $n$0$vumax$0 = 0, $vumax$0$lcssa = 0, $vumax$01 = 0, $vumax$1 = 0, $vumin$0$lcssa = 0, $vumin$02 = 0, $vumin$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$lpMin>>2]|0; //@line 1873 "fastmix.cpp"
 $1 = HEAP32[$lpMax>>2]|0; //@line 1873 "fastmix.cpp"
 $2 = ($lSampleCount|0)==(0); //@line 1877 "fastmix.cpp"
 if ($2) {
  $vumax$0$lcssa = $1;$vumin$0$lcssa = $0;
  HEAP32[$lpMin>>2] = $vumin$0$lcssa; //@line 1899 "fastmix.cpp"
  HEAP32[$lpMax>>2] = $vumax$0$lcssa; //@line 1900 "fastmix.cpp"
  $18 = ($lSampleCount*3)|0; //@line 1901 "fastmix.cpp"
  return ($18|0); //@line 1901 "fastmix.cpp"
 } else {
  $i$03 = 0;$vumax$01 = $1;$vumin$02 = $0;
 }
 while(1) {
  $3 = (($pBuffer) + ($i$03<<2)|0); //@line 1879 "fastmix.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 1879 "fastmix.cpp"
  $5 = ($4|0)<(-134217728); //@line 1880 "fastmix.cpp"
  $6 = ($4|0)>(134217727); //@line 1882 "fastmix.cpp"
  $$ = $6 ? 134217727 : $4; //@line 1882 "fastmix.cpp"
  $n$0 = $5 ? -134217728 : $$; //@line 1880 "fastmix.cpp"
  $7 = ($n$0|0)<($vumin$02|0); //@line 1884 "fastmix.cpp"
  $8 = ($n$0|0)>($vumax$01|0); //@line 1886 "fastmix.cpp"
  $n$0$vumax$0 = $8 ? $n$0 : $vumax$01; //@line 1886 "fastmix.cpp"
  $vumax$1 = $7 ? $vumax$01 : $n$0$vumax$0; //@line 1884 "fastmix.cpp"
  $vumin$1 = $7 ? $n$0 : $vumin$02; //@line 1884 "fastmix.cpp"
  $9 = $n$0 >>> 4; //@line 1888 "fastmix.cpp"
  $10 = $9&255; //@line 1894 "fastmix.cpp"
  $11 = ($i$03*3)|0; //@line 1894 "fastmix.cpp"
  $12 = (($lp16) + ($11)|0); //@line 1894 "fastmix.cpp"
  HEAP8[$12>>0] = $10; //@line 1894 "fastmix.cpp"
  $13 = (($11) + 1)|0; //@line 1895 "fastmix.cpp"
  $14 = (($lp16) + ($13)|0); //@line 1895 "fastmix.cpp"
  HEAP8[$14>>0] = 0; //@line 1895 "fastmix.cpp"
  $15 = (($11) + 2)|0; //@line 1896 "fastmix.cpp"
  $16 = (($lp16) + ($15)|0); //@line 1896 "fastmix.cpp"
  HEAP8[$16>>0] = 0; //@line 1896 "fastmix.cpp"
  $17 = (($i$03) + 1)|0; //@line 1877 "fastmix.cpp"
  $exitcond = ($17|0)==($lSampleCount|0); //@line 1877 "fastmix.cpp"
  if ($exitcond) {
   $vumax$0$lcssa = $vumax$1;$vumin$0$lcssa = $vumin$1;
   break;
  } else {
   $i$03 = $17;$vumax$01 = $vumax$1;$vumin$02 = $vumin$1;
  }
 }
 HEAP32[$lpMin>>2] = $vumin$0$lcssa; //@line 1899 "fastmix.cpp"
 HEAP32[$lpMax>>2] = $vumax$0$lcssa; //@line 1900 "fastmix.cpp"
 $18 = ($lSampleCount*3)|0; //@line 1901 "fastmix.cpp"
 return ($18|0); //@line 1901 "fastmix.cpp"
}
function __Z17X86_Convert32To32PvPijS0_S0_($lp16,$pBuffer,$lSampleCount,$lpMin,$lpMax) {
 $lp16 = $lp16|0;
 $pBuffer = $pBuffer|0;
 $lSampleCount = $lSampleCount|0;
 $lpMin = $lpMin|0;
 $lpMax = $lpMax|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$03 = 0, $n$0 = 0, $n$0$vumax$0 = 0, $vumax$0$lcssa = 0, $vumax$02 = 0;
 var $vumax$1 = 0, $vumin$0$lcssa = 0, $vumin$01 = 0, $vumin$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$lpMin>>2]|0; //@line 1972 "fastmix.cpp"
 $1 = HEAP32[$lpMax>>2]|0; //@line 1972 "fastmix.cpp"
 $2 = ($lSampleCount|0)==(0); //@line 1975 "fastmix.cpp"
 if ($2) {
  $vumax$0$lcssa = $1;$vumin$0$lcssa = $0;
  HEAP32[$lpMin>>2] = $vumin$0$lcssa; //@line 1988 "fastmix.cpp"
  HEAP32[$lpMax>>2] = $vumax$0$lcssa; //@line 1989 "fastmix.cpp"
  $12 = $lSampleCount << 2; //@line 1990 "fastmix.cpp"
  return ($12|0); //@line 1990 "fastmix.cpp"
 } else {
  $i$03 = 0;$vumax$02 = $1;$vumin$01 = $0;
 }
 while(1) {
  $3 = (($pBuffer) + ($i$03<<2)|0); //@line 1977 "fastmix.cpp"
  $4 = HEAP32[$3>>2]|0; //@line 1977 "fastmix.cpp"
  $5 = ($4|0)<(-134217728); //@line 1978 "fastmix.cpp"
  $6 = ($4|0)>(134217727); //@line 1980 "fastmix.cpp"
  $$ = $6 ? 134217727 : $4; //@line 1980 "fastmix.cpp"
  $n$0 = $5 ? -134217728 : $$; //@line 1978 "fastmix.cpp"
  $7 = ($n$0|0)<($vumin$01|0); //@line 1982 "fastmix.cpp"
  $8 = ($n$0|0)>($vumax$02|0); //@line 1984 "fastmix.cpp"
  $n$0$vumax$0 = $8 ? $n$0 : $vumax$02; //@line 1984 "fastmix.cpp"
  $vumin$1 = $7 ? $n$0 : $vumin$01; //@line 1982 "fastmix.cpp"
  $vumax$1 = $7 ? $vumax$02 : $n$0$vumax$0; //@line 1982 "fastmix.cpp"
  $9 = $n$0 << 4; //@line 1986 "fastmix.cpp"
  $10 = (($lp16) + ($i$03<<2)|0); //@line 1986 "fastmix.cpp"
  HEAP32[$10>>2] = $9; //@line 1986 "fastmix.cpp"
  $11 = (($i$03) + 1)|0; //@line 1975 "fastmix.cpp"
  $exitcond = ($11|0)==($lSampleCount|0); //@line 1975 "fastmix.cpp"
  if ($exitcond) {
   $vumax$0$lcssa = $vumax$1;$vumin$0$lcssa = $vumin$1;
   break;
  } else {
   $i$03 = $11;$vumax$02 = $vumax$1;$vumin$01 = $vumin$1;
  }
 }
 HEAP32[$lpMin>>2] = $vumin$0$lcssa; //@line 1988 "fastmix.cpp"
 HEAP32[$lpMax>>2] = $vumax$0$lcssa; //@line 1989 "fastmix.cpp"
 $12 = $lSampleCount << 2; //@line 1990 "fastmix.cpp"
 return ($12|0); //@line 1990 "fastmix.cpp"
}
function __ZN10CSoundFile8ReadNoteEv($this) {
 $this = $this|0;
 var $$ = 0, $$$14 = 0, $$$14$ = 0, $$$19 = 0, $$$24 = 0, $$$3 = 0, $$$5 = 0, $$$7 = 0, $$$period$1 = 0, $$0 = 0, $$14 = 0, $$19 = 0, $$2 = 0, $$20 = 0, $$24 = 0, $$25 = 0, $$26 = 0, $$29 = 0, $$3 = 0, $$30 = 0;
 var $$32 = 0, $$33 = 0, $$4 = 0, $$5 = 0, $$6 = 0, $$7 = 0, $$envpan$1 = 0, $$envpan$1$op = 0, $$envpitch$1 = 0, $$envpitch$3 = 0, $$envvol$1 = 0, $$in = 0, $$not = 0, $$not107 = 0, $$off = 0, $$old = 0, $$old116 = 0, $$old121 = 0, $$old122 = 0, $$pan$0 = 0;
 var $$period$1 = 0, $$phi$trans$insert = 0, $$phi$trans$insert85 = 0, $$pn = 0, $$pn$in = 0, $$pn23 = 0, $$pre = 0, $$pre$phi89Z2D = 0, $$pre$phi94Z2D = 0, $$pre77 = 0, $$pre79 = 0, $$pre84 = 0, $$pre86 = 0, $$pre91 = 0, $$pre93 = 0, $$vol$4 = 0, $$vol$4$op = 0, $$vol$7 = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0;
 var $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0;
 var $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0;
 var $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0;
 var $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0, $1088 = 0, $1089 = 0;
 var $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0, $1106 = 0;
 var $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0, $1124 = 0;
 var $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0;
 var $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0;
 var $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0;
 var $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0;
 var $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0;
 var $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0;
 var $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0;
 var $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0;
 var $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0;
 var $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0;
 var $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0;
 var $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0;
 var $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0;
 var $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0;
 var $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0;
 var $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0;
 var $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0;
 var $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0;
 var $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0;
 var $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0;
 var $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0;
 var $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0;
 var $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0;
 var $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0;
 var $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0;
 var $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0;
 var $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0;
 var $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0;
 var $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0;
 var $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0;
 var $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0;
 var $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0;
 var $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0;
 var $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0;
 var $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0;
 var $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0;
 var $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0;
 var $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0;
 var $_ZL11PreAmpTable$_ZL14PreAmpAGCTable = 0, $brmerge = 0, $df1$0 = 0, $df1$0$in = 0, $df2$0 = 0, $df2$0$in = 0, $envpan$0 = 0, $envpan$1 = 0, $envpitch$0 = 0, $envpitch$1 = 0, $envpitch$3 = 0, $envvol$0 = 0, $envvol$1 = 0, $envvol$3 = 0, $i$046 = 0, $i17$045 = 0, $i3$050 = 0, $i9$057 = 0, $j$043 = 0, $j$043$phi = 0;
 var $mastervol$0 = 0, $n$0 = 0, $n13$0 = 0, $nChn$064 = 0, $nPeriodFrac$0 = 0, $nPeriodFrac$1 = 0, $nRampLength$0 = 0, $nchn32$0 = 0, $ninc$0 = 0, $ninc$0$op = 0, $ninc$0$op$op = 0, $not$ = 0, $not$128 = 0, $not$40 = 0, $not$41 = 0, $notlhs = 0, $notlhs129 = 0, $notrhs = 0, $notrhs130 = 0, $ontime$0 = 0;
 var $or$cond = 0, $or$cond100 = 0, $or$cond102 = 0, $or$cond104 = 0, $or$cond109 = 0, $or$cond11 = 0, $or$cond111 = 0, $or$cond112 = 0, $or$cond113 = 0, $or$cond115 = 0, $or$cond118 = 0, $or$cond118$not = 0, $or$cond120$not = 0, $or$cond123 = 0, $or$cond125 = 0, $or$cond127 = 0, $or$cond16 = 0, $or$cond18$not = 0, $or$cond27 = 0, $or$cond28 = 0;
 var $or$cond31 = 0, $or$cond37 = 0, $or$cond9 = 0, $or$cond98 = 0, $pChn$068 = 0, $pan$0 = 0, $pan$2 = 0, $pdelta$0 = 0, $period$0 = 0, $period$1 = 0, $period$3 = 0, $period$4 = 0, $period$5 = 0, $period$6 = 0, $period$7 = 0, $period$8 = 0, $pt$0 = 0, $pt2$0 = 0, $pt8$0 = 0, $realmastervol$0 = 0;
 var $storemerge22 = 0, $val$0 = 0, $vdelta$0 = 0, $vdelta$1 = 0, $vol$0 = 0, $vol$1 = 0, $vol$1$ = 0, $vol$3 = 0, $vol$4 = 0, $vol$7 = 0, $vol$8 = 0, $vol$9 = 0, $x1$0 = 0, $x111$0 = 0, $x15$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN10CSoundFile10ProcessRowEv($this)|0); //@line 465 "sndmix.cpp"
 if (!($0)) {
  $$0 = 0;
  return ($$0|0); //@line 1235 "sndmix.cpp"
 }
 $1 = ((($this)) + 59696|0); //@line 467 "sndmix.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 467 "sndmix.cpp"
 $3 = (($2) + 1)|0; //@line 467 "sndmix.cpp"
 HEAP32[$1>>2] = $3; //@line 467 "sndmix.cpp"
 $4 = ((($this)) + 59712|0); //@line 468 "sndmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 468 "sndmix.cpp"
 $6 = ($5|0)==(0); //@line 468 "sndmix.cpp"
 if ($6) {
  $$0 = 0;
  return ($$0|0); //@line 1235 "sndmix.cpp"
 }
 $7 = HEAP32[10109]|0; //@line 469 "sndmix.cpp"
 $8 = ($7*5)|0; //@line 469 "sndmix.cpp"
 $9 = ((($this)) + 59756|0); //@line 469 "sndmix.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 469 "sndmix.cpp"
 $11 = Math_imul($8, $10)|0; //@line 469 "sndmix.cpp"
 $12 = $5 << 8; //@line 469 "sndmix.cpp"
 $13 = (($11>>>0) / ($12>>>0))&-1; //@line 469 "sndmix.cpp"
 $14 = ((($this)) + 59676|0); //@line 469 "sndmix.cpp"
 HEAP32[$14>>2] = $13; //@line 469 "sndmix.cpp"
 $15 = ((($this)) + 59664|0); //@line 473 "sndmix.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 473 "sndmix.cpp"
 $17 = ($16>>>0)>(31); //@line 473 "sndmix.cpp"
 $$24 = $17 ? 31 : $16; //@line 473 "sndmix.cpp"
 $18 = ((($this)) + 59680|0); //@line 474 "sndmix.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 474 "sndmix.cpp"
 $20 = $19 & 32; //@line 474 "sndmix.cpp"
 $21 = ($20|0)==(0); //@line 474 "sndmix.cpp"
 if ($21) {
  $nchn32$0 = $$24;
 } else {
  $22 = ((($this)) + 59688|0); //@line 474 "sndmix.cpp"
  $23 = HEAP32[$22>>2]|0; //@line 474 "sndmix.cpp"
  $24 = ($23|0)!=(0); //@line 474 "sndmix.cpp"
  $25 = ($$24|0)<(6); //@line 474 "sndmix.cpp"
  $or$cond = $25 & $24; //@line 474 "sndmix.cpp"
  $$$24 = $or$cond ? 6 : $$24; //@line 474 "sndmix.cpp"
  $nchn32$0 = $$$24;
 }
 $26 = ((($this)) + 59740|0); //@line 475 "sndmix.cpp"
 $27 = HEAP32[$26>>2]|0; //@line 475 "sndmix.cpp"
 $28 = ($27|0)>(128); //@line 476 "sndmix.cpp"
 if ($28) {
  $29 = (($27) + -128)|0; //@line 478 "sndmix.cpp"
  $30 = (($nchn32$0) + 4)|0; //@line 478 "sndmix.cpp"
  $31 = Math_imul($29, $30)|0; //@line 478 "sndmix.cpp"
  $32 = (($31|0) / 16)&-1; //@line 478 "sndmix.cpp"
  $33 = (($32) + 128)|0; //@line 478 "sndmix.cpp"
  $realmastervol$0 = $33;
 } else {
  $realmastervol$0 = $27;
 }
 $34 = HEAP32[183080]|0; //@line 480 "sndmix.cpp"
 $35 = $34 & 4; //@line 480 "sndmix.cpp"
 $36 = ($35|0)==(0); //@line 480 "sndmix.cpp"
 $37 = $nchn32$0 >> 1; //@line 480 "sndmix.cpp"
 $_ZL11PreAmpTable$_ZL14PreAmpAGCTable = $36 ? 40532 : 40596; //@line 480 "sndmix.cpp"
 $$in = (($_ZL11PreAmpTable$_ZL14PreAmpAGCTable) + ($37<<2)|0); //@line 480 "sndmix.cpp"
 $38 = HEAP32[$$in>>2]|0; //@line 480 "sndmix.cpp"
 $39 = ((($this)) + 59748|0); //@line 481 "sndmix.cpp"
 $40 = HEAP32[$39>>2]|0; //@line 481 "sndmix.cpp"
 $41 = (($40) + 16)|0; //@line 481 "sndmix.cpp"
 $42 = Math_imul($41, $realmastervol$0)|0; //@line 481 "sndmix.cpp"
 $43 = $42 >>> 6; //@line 481 "sndmix.cpp"
 $44 = ($43>>>0)>(512); //@line 482 "sndmix.cpp"
 $$ = $44 ? 512 : $43; //@line 482 "sndmix.cpp"
 $45 = ((($this)) + 59660|0); //@line 483 "sndmix.cpp"
 $46 = HEAP32[$45>>2]|0; //@line 483 "sndmix.cpp"
 $47 = $46 & 1024; //@line 483 "sndmix.cpp"
 $48 = ($47|0)==(0); //@line 483 "sndmix.cpp"
 if ($48) {
  $mastervol$0 = $$;
 } else {
  $49 = ((($this)) + 59784|0); //@line 483 "sndmix.cpp"
  $50 = HEAP32[$49>>2]|0; //@line 483 "sndmix.cpp"
  $51 = ($50|0)==(0); //@line 483 "sndmix.cpp"
  if ($51) {
   $mastervol$0 = $$;
  } else {
   $52 = ((($this)) + 59780|0); //@line 485 "sndmix.cpp"
   $53 = HEAP32[$52>>2]|0; //@line 485 "sndmix.cpp"
   $54 = ($53|0)<(0); //@line 122 "sndmix.cpp"
   $55 = $54 << 31 >> 31; //@line 122 "sndmix.cpp"
   $56 = (___muldi3(($53|0),($55|0),($$|0),0)|0); //@line 122 "sndmix.cpp"
   $57 = tempRet0; //@line 122 "sndmix.cpp"
   $58 = ($50|0)<(0); //@line 122 "sndmix.cpp"
   $59 = $58 << 31 >> 31; //@line 122 "sndmix.cpp"
   $60 = (___udivdi3(($56|0),($57|0),($50|0),($59|0))|0); //@line 122 "sndmix.cpp"
   $61 = tempRet0; //@line 122 "sndmix.cpp"
   $mastervol$0 = $60;
  }
 }
 $62 = $mastervol$0 << 7; //@line 487 "sndmix.cpp"
 $63 = (($62>>>0) / ($38>>>0))&-1; //@line 487 "sndmix.cpp"
 $64 = ($63>>>0)>(384); //@line 488 "sndmix.cpp"
 $$2 = $64 ? 384 : $63; //@line 488 "sndmix.cpp"
 $65 = ((($this)) + 59668|0); //@line 492 "sndmix.cpp"
 HEAP32[$65>>2] = 0; //@line 492 "sndmix.cpp"
 $66 = ((($this)) + 59692|0); //@line 559 "sndmix.cpp"
 $67 = ((($this)) + 59744|0); //@line 696 "sndmix.cpp"
 $68 = ((($this)) + 59764|0); //@line 698 "sndmix.cpp"
 $69 = ((($this)) + 59768|0); //@line 923 "sndmix.cpp"
 $70 = ((($this)) + 59752|0); //@line 943 "sndmix.cpp"
 $nChn$064 = 0;$pChn$068 = $this;
 L17: while(1) {
  $71 = ((($pChn$068)) + 36|0); //@line 496 "sndmix.cpp"
  $72 = HEAP32[$71>>2]|0; //@line 496 "sndmix.cpp"
  $73 = $72 & 1024; //@line 496 "sndmix.cpp"
  $74 = ($73|0)==(0); //@line 496 "sndmix.cpp"
  if (!($74)) {
   $75 = ((($pChn$068)) + 124|0); //@line 496 "sndmix.cpp"
   $76 = HEAP32[$75>>2]|0; //@line 496 "sndmix.cpp"
   $77 = ((($pChn$068)) + 16|0); //@line 496 "sndmix.cpp"
   $78 = HEAP32[$77>>2]|0; //@line 496 "sndmix.cpp"
   $79 = $78 | $76; //@line 496 "sndmix.cpp"
   $80 = ((($pChn$068)) + 20|0); //@line 496 "sndmix.cpp"
   $81 = HEAP32[$80>>2]|0; //@line 496 "sndmix.cpp"
   $82 = $79 | $81; //@line 496 "sndmix.cpp"
   $83 = ($82|0)==(0); //@line 496 "sndmix.cpp"
   if ($83) {
    $84 = ((($pChn$068)) + 32|0); //@line 498 "sndmix.cpp"
    HEAP32[$84>>2] = 0; //@line 498 "sndmix.cpp"
    $85 = ((($pChn$068)) + 88|0); //@line 499 "sndmix.cpp"
    HEAP32[$85>>2] = 0; //@line 499 "sndmix.cpp"
    $86 = ((($pChn$068)) + 84|0); //@line 499 "sndmix.cpp"
    HEAP32[$86>>2] = 0; //@line 499 "sndmix.cpp"
   }
  }
  $87 = $72 & 256; //@line 502 "sndmix.cpp"
  $88 = ($87|0)==(0); //@line 502 "sndmix.cpp"
  do {
   if ($88) {
    $89 = HEAP32[$15>>2]|0; //@line 502 "sndmix.cpp"
    $90 = ($nChn$064>>>0)<($89>>>0); //@line 502 "sndmix.cpp"
    if (!($90)) {
     $91 = ((($pChn$068)) + 32|0); //@line 502 "sndmix.cpp"
     $92 = HEAP32[$91>>2]|0; //@line 502 "sndmix.cpp"
     $93 = ($92|0)==(0); //@line 502 "sndmix.cpp"
     if ($93) {
      label = 17;
      break;
     }
    }
    $95 = ((($pChn$068)) + 12|0); //@line 511 "sndmix.cpp"
    HEAP32[$95>>2] = 0; //@line 511 "sndmix.cpp"
    $96 = ((($pChn$068)) + 108|0); //@line 512 "sndmix.cpp"
    HEAP32[$96>>2] = 0; //@line 512 "sndmix.cpp"
    $97 = ((($pChn$068)) + 120|0); //@line 513 "sndmix.cpp"
    $98 = HEAP32[$97>>2]|0; //@line 513 "sndmix.cpp"
    $99 = ((($pChn$068)) + 210|0); //@line 513 "sndmix.cpp"
    $100 = HEAP16[$99>>1]|0; //@line 513 "sndmix.cpp"
    $101 = $100 << 16 >> 16; //@line 513 "sndmix.cpp"
    $102 = (($101) + ($98))|0; //@line 513 "sndmix.cpp"
    $103 = ((($pChn$068)) + 112|0); //@line 513 "sndmix.cpp"
    $104 = ($102|0)<(0); //@line 514 "sndmix.cpp"
    $$25 = $104 ? 0 : $102; //@line 514 "sndmix.cpp"
    $105 = ($$25|0)>(256); //@line 515 "sndmix.cpp"
    $storemerge22 = $105 ? 256 : $$25; //@line 515 "sndmix.cpp"
    HEAP32[$103>>2] = $storemerge22; //@line 513 "sndmix.cpp"
    $106 = ((($pChn$068)) + 92|0); //@line 516 "sndmix.cpp"
    HEAP32[$106>>2] = 0; //@line 516 "sndmix.cpp"
    $107 = ((($pChn$068)) + 128|0); //@line 518 "sndmix.cpp"
    $108 = HEAP32[$107>>2]|0; //@line 518 "sndmix.cpp"
    $109 = ($108|0)==(0); //@line 518 "sndmix.cpp"
    if ($109) {
     $961 = 0;
    } else {
     $110 = ((($pChn$068)) + 32|0); //@line 518 "sndmix.cpp"
     $111 = HEAP32[$110>>2]|0; //@line 518 "sndmix.cpp"
     $112 = ($111|0)==(0); //@line 518 "sndmix.cpp"
     if ($112) {
      $961 = 0;
     } else {
      $113 = ((($pChn$068)) + 116|0); //@line 520 "sndmix.cpp"
      $114 = HEAP32[$113>>2]|0; //@line 520 "sndmix.cpp"
      $115 = ((($pChn$068)) + 208|0); //@line 520 "sndmix.cpp"
      $116 = HEAP16[$115>>1]|0; //@line 520 "sndmix.cpp"
      $117 = $116 << 16 >> 16; //@line 520 "sndmix.cpp"
      $118 = (($117) + ($114))|0; //@line 520 "sndmix.cpp"
      $119 = ($118|0)<(0); //@line 522 "sndmix.cpp"
      $$3 = $119 ? 0 : $118; //@line 522 "sndmix.cpp"
      $120 = ($$3|0)>(256); //@line 523 "sndmix.cpp"
      $$$3 = $120 ? 256 : $$3; //@line 523 "sndmix.cpp"
      $121 = $72 & 131072; //@line 525 "sndmix.cpp"
      $122 = ($121|0)==(0); //@line 525 "sndmix.cpp"
      do {
       if ($122) {
        $vol$1 = $$$3;
       } else {
        $123 = ((($pChn$068)) + 200|0); //@line 527 "sndmix.cpp"
        $124 = HEAP32[$123>>2]|0; //@line 527 "sndmix.cpp"
        $125 = $124 & 63; //@line 527 "sndmix.cpp"
        $126 = ($$$3|0)>(0); //@line 528 "sndmix.cpp"
        L33: do {
         if ($126) {
          $127 = HEAP32[$18>>2]|0; //@line 530 "sndmix.cpp"
          $128 = $127 >>> 2; //@line 530 "sndmix.cpp"
          $129 = $128 & 1; //@line 530 "sndmix.cpp"
          $130 = (6 - ($129))|0; //@line 530 "sndmix.cpp"
          $131 = ((($pChn$068)) + 227|0); //@line 531 "sndmix.cpp"
          $132 = HEAP8[$131>>0]|0; //@line 531 "sndmix.cpp"
          $133 = $132&255; //@line 531 "sndmix.cpp"
          $134 = $133 & 3; //@line 531 "sndmix.cpp"
          switch ($134|0) {
          case 1:  {
           $135 = (169324 + ($125<<1)|0); //@line 534 "sndmix.cpp"
           $136 = HEAP16[$135>>1]|0; //@line 534 "sndmix.cpp"
           $137 = $136 << 16 >> 16; //@line 534 "sndmix.cpp"
           $138 = ((($pChn$068)) + 229|0); //@line 534 "sndmix.cpp"
           $139 = HEAP8[$138>>0]|0; //@line 534 "sndmix.cpp"
           $140 = $139&255; //@line 534 "sndmix.cpp"
           $141 = Math_imul($140, $137)|0; //@line 534 "sndmix.cpp"
           $142 = $141 >> $130; //@line 534 "sndmix.cpp"
           $143 = (($142) + ($$$3))|0; //@line 534 "sndmix.cpp"
           $vol$0 = $143;
           break L33;
           break;
          }
          case 2:  {
           $144 = (169452 + ($125<<1)|0); //@line 537 "sndmix.cpp"
           $145 = HEAP16[$144>>1]|0; //@line 537 "sndmix.cpp"
           $146 = $145 << 16 >> 16; //@line 537 "sndmix.cpp"
           $147 = ((($pChn$068)) + 229|0); //@line 537 "sndmix.cpp"
           $148 = HEAP8[$147>>0]|0; //@line 537 "sndmix.cpp"
           $149 = $148&255; //@line 537 "sndmix.cpp"
           $150 = Math_imul($149, $146)|0; //@line 537 "sndmix.cpp"
           $151 = $150 >> $130; //@line 537 "sndmix.cpp"
           $152 = (($151) + ($$$3))|0; //@line 537 "sndmix.cpp"
           $vol$0 = $152;
           break L33;
           break;
          }
          case 3:  {
           $153 = (169580 + ($125<<1)|0); //@line 540 "sndmix.cpp"
           $154 = HEAP16[$153>>1]|0; //@line 540 "sndmix.cpp"
           $155 = $154 << 16 >> 16; //@line 540 "sndmix.cpp"
           $156 = ((($pChn$068)) + 229|0); //@line 540 "sndmix.cpp"
           $157 = HEAP8[$156>>0]|0; //@line 540 "sndmix.cpp"
           $158 = $157&255; //@line 540 "sndmix.cpp"
           $159 = Math_imul($158, $155)|0; //@line 540 "sndmix.cpp"
           $160 = $159 >> $130; //@line 540 "sndmix.cpp"
           $161 = (($160) + ($$$3))|0; //@line 540 "sndmix.cpp"
           $vol$0 = $161;
           break L33;
           break;
          }
          default: {
           $162 = (169708 + ($125<<1)|0); //@line 543 "sndmix.cpp"
           $163 = HEAP16[$162>>1]|0; //@line 543 "sndmix.cpp"
           $164 = $163 << 16 >> 16; //@line 543 "sndmix.cpp"
           $165 = ((($pChn$068)) + 229|0); //@line 543 "sndmix.cpp"
           $166 = HEAP8[$165>>0]|0; //@line 543 "sndmix.cpp"
           $167 = $166&255; //@line 543 "sndmix.cpp"
           $168 = Math_imul($167, $164)|0; //@line 543 "sndmix.cpp"
           $169 = $168 >> $130; //@line 543 "sndmix.cpp"
           $170 = (($169) + ($$$3))|0; //@line 543 "sndmix.cpp"
           $vol$0 = $170;
           break L33;
          }
          }
         } else {
          $vol$0 = $$$3;
         }
        } while(0);
        $171 = HEAP32[$66>>2]|0; //@line 546 "sndmix.cpp"
        $172 = ($171|0)==(0); //@line 546 "sndmix.cpp"
        if ($172) {
         $173 = HEAP32[$18>>2]|0; //@line 546 "sndmix.cpp"
         $174 = $173 & 290; //@line 546 "sndmix.cpp"
         $175 = ($174|0)==(0); //@line 546 "sndmix.cpp"
         if ($175) {
          $vol$1 = $vol$0;
          break;
         }
         $176 = HEAP32[$45>>2]|0; //@line 546 "sndmix.cpp"
         $177 = $176 & 4; //@line 546 "sndmix.cpp"
         $178 = ($177|0)==(0); //@line 546 "sndmix.cpp"
         if (!($178)) {
          $vol$1 = $vol$0;
          break;
         }
        }
        $179 = ((($pChn$068)) + 228|0); //@line 548 "sndmix.cpp"
        $180 = HEAP8[$179>>0]|0; //@line 548 "sndmix.cpp"
        $181 = $180&255; //@line 548 "sndmix.cpp"
        $182 = (($181) + ($124))|0; //@line 548 "sndmix.cpp"
        $183 = $182 & 63; //@line 548 "sndmix.cpp"
        HEAP32[$123>>2] = $183; //@line 548 "sndmix.cpp"
        $vol$1 = $vol$0;
       }
      } while(0);
      $184 = ((($pChn$068)) + 216|0); //@line 552 "sndmix.cpp"
      $185 = HEAP8[$184>>0]|0; //@line 552 "sndmix.cpp"
      $186 = ($185<<24>>24)==(18); //@line 552 "sndmix.cpp"
      if ($186) {
       $187 = ((($pChn$068)) + 243|0); //@line 554 "sndmix.cpp"
       $188 = HEAP8[$187>>0]|0; //@line 554 "sndmix.cpp"
       $189 = $188&255; //@line 554 "sndmix.cpp"
       $190 = $189 >>> 4; //@line 554 "sndmix.cpp"
       $191 = $189 & 15; //@line 554 "sndmix.cpp"
       $192 = (($190) + ($191))|0; //@line 554 "sndmix.cpp"
       $193 = HEAP32[$18>>2]|0; //@line 556 "sndmix.cpp"
       $194 = $193 & 32; //@line 556 "sndmix.cpp"
       $195 = ($194|0)==(0); //@line 556 "sndmix.cpp"
       if ($195) {
        label = 34;
       } else {
        $196 = HEAP32[$45>>2]|0; //@line 556 "sndmix.cpp"
        $197 = $196 & 4; //@line 556 "sndmix.cpp"
        $198 = ($197|0)==(0); //@line 556 "sndmix.cpp"
        if ($198) {
         $n$0 = $192;$ontime$0 = $190;
        } else {
         label = 34;
        }
       }
       if ((label|0) == 34) {
        label = 0;
        $199 = (($192) + 2)|0; //@line 556 "sndmix.cpp"
        $200 = (($190) + 1)|0; //@line 556 "sndmix.cpp"
        $n$0 = $199;$ontime$0 = $200;
       }
       $201 = ((($pChn$068)) + 242|0); //@line 557 "sndmix.cpp"
       $202 = HEAP8[$201>>0]|0; //@line 557 "sndmix.cpp"
       $203 = $202&255; //@line 557 "sndmix.cpp"
       $204 = ($203>>>0)>=($n$0>>>0); //@line 558 "sndmix.cpp"
       $$4 = $204 ? 0 : $203; //@line 558 "sndmix.cpp"
       $205 = HEAP32[$66>>2]|0; //@line 559 "sndmix.cpp"
       $206 = $193 & 34;
       $207 = $205 | $206; //@line 559 "sndmix.cpp"
       $208 = ($207|0)==(0); //@line 559 "sndmix.cpp"
       if ($208) {
        $vol$3 = $vol$1;
       } else {
        $209 = ($$4>>>0)<($ontime$0>>>0); //@line 561 "sndmix.cpp"
        $vol$1$ = $209 ? $vol$1 : 0; //@line 561 "sndmix.cpp"
        $210 = (($$4) + 1)|0; //@line 562 "sndmix.cpp"
        $211 = $210&255; //@line 562 "sndmix.cpp"
        HEAP8[$201>>0] = $211; //@line 562 "sndmix.cpp"
        $vol$3 = $vol$1$;
       }
       $212 = $72 | 16777216; //@line 564 "sndmix.cpp"
       HEAP32[$71>>2] = $212; //@line 564 "sndmix.cpp"
       $220 = $212;$vol$4 = $vol$3;
      } else {
       $220 = $72;$vol$4 = $vol$1;
      }
      $213 = ($vol$4|0)<(0); //@line 567 "sndmix.cpp"
      $$vol$4 = $213 ? 0 : $vol$4; //@line 567 "sndmix.cpp"
      $214 = ($$vol$4|0)>(256); //@line 568 "sndmix.cpp"
      $$vol$4$op = $$vol$4 << 6; //@line 569 "sndmix.cpp"
      $215 = $214 ? 16384 : $$vol$4$op; //@line 569 "sndmix.cpp"
      $216 = ((($pChn$068)) + 140|0); //@line 571 "sndmix.cpp"
      $217 = HEAP32[$216>>2]|0; //@line 571 "sndmix.cpp"
      $218 = ($217|0)==(0|0); //@line 571 "sndmix.cpp"
      do {
       if ($218) {
        $350 = $220 & 1024; //@line 685 "sndmix.cpp"
        $351 = ($350|0)==(0); //@line 685 "sndmix.cpp"
        if ($351) {
         $vol$9 = $215;
         label = 71;
        } else {
         $352 = ((($pChn$068)) + 124|0); //@line 687 "sndmix.cpp"
         HEAP32[$352>>2] = 0; //@line 687 "sndmix.cpp"
        }
       } else {
        $219 = $220 & 2097152; //@line 575 "sndmix.cpp"
        $221 = ($219|0)==(0); //@line 575 "sndmix.cpp"
        if ($221) {
         $vol$7 = $215;
        } else {
         $222 = ((($217)) + 556|0); //@line 575 "sndmix.cpp"
         $223 = HEAP8[$222>>0]|0; //@line 575 "sndmix.cpp"
         $224 = ($223<<24>>24)==(0); //@line 575 "sndmix.cpp"
         if ($224) {
          $vol$7 = $215;
         } else {
          $225 = ((($pChn$068)) + 148|0); //@line 577 "sndmix.cpp"
          $226 = HEAP32[$225>>2]|0; //@line 577 "sndmix.cpp"
          $227 = $223&255; //@line 578 "sndmix.cpp"
          $228 = (($227) + -1)|0; //@line 578 "sndmix.cpp"
          $229 = ($223<<24>>24)==(1); //@line 579 "sndmix.cpp"
          L64: do {
           if ($229) {
            $pt$0 = $228;
           } else {
            $i$046 = 0;
            while(1) {
             $230 = (((($217)) + 12|0) + ($i$046<<1)|0); //@line 581 "sndmix.cpp"
             $231 = HEAP16[$230>>1]|0; //@line 581 "sndmix.cpp"
             $232 = $231&65535; //@line 581 "sndmix.cpp"
             $233 = ($226|0)>($232|0); //@line 581 "sndmix.cpp"
             if (!($233)) {
              $pt$0 = $i$046;
              break L64;
             }
             $234 = (($i$046) + 1)|0; //@line 579 "sndmix.cpp"
             $235 = ($234>>>0)<($228>>>0); //@line 579 "sndmix.cpp"
             if ($235) {
              $i$046 = $234;
             } else {
              $pt$0 = $228;
              break;
             }
            }
           }
          } while(0);
          $236 = (((($217)) + 12|0) + ($pt$0<<1)|0); //@line 587 "sndmix.cpp"
          $237 = HEAP16[$236>>1]|0; //@line 587 "sndmix.cpp"
          $238 = $237&65535; //@line 587 "sndmix.cpp"
          $239 = ($226|0)<($238|0); //@line 589 "sndmix.cpp"
          do {
           if ($239) {
            $244 = ($pt$0|0)==(0); //@line 594 "sndmix.cpp"
            if ($244) {
             $envvol$0 = 0;$x1$0 = 0;
             break;
            }
            $245 = (($pt$0) + -1)|0; //@line 596 "sndmix.cpp"
            $246 = (((($217)) + 204|0) + ($245)|0); //@line 596 "sndmix.cpp"
            $247 = HEAP8[$246>>0]|0; //@line 596 "sndmix.cpp"
            $248 = $247&255; //@line 596 "sndmix.cpp"
            $249 = $248 << 2; //@line 596 "sndmix.cpp"
            $250 = (((($217)) + 12|0) + ($245<<1)|0); //@line 597 "sndmix.cpp"
            $251 = HEAP16[$250>>1]|0; //@line 597 "sndmix.cpp"
            $252 = $251&65535; //@line 597 "sndmix.cpp"
            $envvol$0 = $249;$x1$0 = $252;
           } else {
            $240 = (((($217)) + 204|0) + ($pt$0)|0); //@line 591 "sndmix.cpp"
            $241 = HEAP8[$240>>0]|0; //@line 591 "sndmix.cpp"
            $242 = $241&255; //@line 591 "sndmix.cpp"
            $243 = $242 << 2; //@line 591 "sndmix.cpp"
            $envvol$0 = $243;$x1$0 = $238;
           }
          } while(0);
          $253 = ($226|0)>($238|0); //@line 603 "sndmix.cpp"
          $$26 = $253 ? $238 : $226; //@line 603 "sndmix.cpp"
          $254 = ($238>>>0)>($x1$0>>>0); //@line 604 "sndmix.cpp"
          $255 = ($$26|0)>($x1$0|0); //@line 604 "sndmix.cpp"
          $or$cond27 = $254 & $255; //@line 604 "sndmix.cpp"
          if ($or$cond27) {
           $256 = (($$26) - ($x1$0))|0; //@line 606 "sndmix.cpp"
           $257 = (((($217)) + 204|0) + ($pt$0)|0); //@line 606 "sndmix.cpp"
           $258 = HEAP8[$257>>0]|0; //@line 606 "sndmix.cpp"
           $259 = $258&255; //@line 606 "sndmix.cpp"
           $260 = $259 << 2; //@line 606 "sndmix.cpp"
           $261 = (($260) - ($envvol$0))|0; //@line 606 "sndmix.cpp"
           $262 = Math_imul($261, $256)|0; //@line 606 "sndmix.cpp"
           $263 = (($238) - ($x1$0))|0; //@line 606 "sndmix.cpp"
           $264 = (($262|0) / ($263|0))&-1; //@line 606 "sndmix.cpp"
           $265 = (($264) + ($envvol$0))|0; //@line 606 "sndmix.cpp"
           $envvol$1 = $265;
          } else {
           $envvol$1 = $envvol$0;
          }
          $266 = ($envvol$1|0)<(0); //@line 608 "sndmix.cpp"
          $$envvol$1 = $266 ? 0 : $envvol$1; //@line 608 "sndmix.cpp"
          $267 = ($$envvol$1|0)>(256); //@line 609 "sndmix.cpp"
          $envvol$3 = $267 ? 256 : $$envvol$1; //@line 609 "sndmix.cpp"
          $268 = Math_imul($envvol$3, $215)|0; //@line 610 "sndmix.cpp"
          $269 = $268 >> 8; //@line 610 "sndmix.cpp"
          $vol$7 = $269;
         }
        }
        $270 = $220 & 4194304; //@line 613 "sndmix.cpp"
        $271 = ($270|0)==(0); //@line 613 "sndmix.cpp"
        do {
         if ($271) {
          $334 = $storemerge22;
         } else {
          $272 = ((($217)) + 557|0); //@line 613 "sndmix.cpp"
          $273 = HEAP8[$272>>0]|0; //@line 613 "sndmix.cpp"
          $274 = ($273<<24>>24)==(0); //@line 613 "sndmix.cpp"
          if ($274) {
           $334 = $storemerge22;
           break;
          }
          $275 = ((($pChn$068)) + 152|0); //@line 615 "sndmix.cpp"
          $276 = HEAP32[$275>>2]|0; //@line 615 "sndmix.cpp"
          $277 = $273&255; //@line 616 "sndmix.cpp"
          $278 = (($277) + -1)|0; //@line 616 "sndmix.cpp"
          $279 = ($273<<24>>24)==(1); //@line 617 "sndmix.cpp"
          L81: do {
           if ($279) {
            $pt2$0 = $278;
           } else {
            $i3$050 = 0;
            while(1) {
             $280 = (((($217)) + 76|0) + ($i3$050<<1)|0); //@line 619 "sndmix.cpp"
             $281 = HEAP16[$280>>1]|0; //@line 619 "sndmix.cpp"
             $282 = $281&65535; //@line 619 "sndmix.cpp"
             $283 = ($276|0)>($282|0); //@line 619 "sndmix.cpp"
             if (!($283)) {
              $pt2$0 = $i3$050;
              break L81;
             }
             $284 = (($i3$050) + 1)|0; //@line 617 "sndmix.cpp"
             $285 = ($284>>>0)<($278>>>0); //@line 617 "sndmix.cpp"
             if ($285) {
              $i3$050 = $284;
             } else {
              $pt2$0 = $278;
              break;
             }
            }
           }
          } while(0);
          $286 = (((($217)) + 76|0) + ($pt2$0<<1)|0); //@line 625 "sndmix.cpp"
          $287 = HEAP16[$286>>1]|0; //@line 625 "sndmix.cpp"
          $288 = $287&65535; //@line 625 "sndmix.cpp"
          $289 = (((($217)) + 236|0) + ($pt2$0)|0); //@line 625 "sndmix.cpp"
          $290 = HEAP8[$289>>0]|0; //@line 625 "sndmix.cpp"
          $291 = $290&255; //@line 625 "sndmix.cpp"
          $292 = ($276|0)<($288|0); //@line 627 "sndmix.cpp"
          do {
           if ($292) {
            $293 = ($pt2$0|0)==(0); //@line 632 "sndmix.cpp"
            if ($293) {
             $envpan$0 = 128;$x15$0 = 0;
             break;
            }
            $294 = (($pt2$0) + -1)|0; //@line 634 "sndmix.cpp"
            $295 = (((($217)) + 236|0) + ($294)|0); //@line 634 "sndmix.cpp"
            $296 = HEAP8[$295>>0]|0; //@line 634 "sndmix.cpp"
            $297 = $296&255; //@line 634 "sndmix.cpp"
            $298 = (((($217)) + 76|0) + ($294<<1)|0); //@line 635 "sndmix.cpp"
            $299 = HEAP16[$298>>1]|0; //@line 635 "sndmix.cpp"
            $300 = $299&65535; //@line 635 "sndmix.cpp"
            $envpan$0 = $297;$x15$0 = $300;
           } else {
            $envpan$0 = $291;$x15$0 = $288;
           }
          } while(0);
          $301 = ($288>>>0)>($x15$0>>>0); //@line 641 "sndmix.cpp"
          $302 = ($276|0)>($x15$0|0); //@line 641 "sndmix.cpp"
          $or$cond28 = $301 & $302; //@line 641 "sndmix.cpp"
          if ($or$cond28) {
           $303 = (($276) - ($x15$0))|0; //@line 643 "sndmix.cpp"
           $304 = (($291) - ($envpan$0))|0; //@line 643 "sndmix.cpp"
           $305 = Math_imul($304, $303)|0; //@line 643 "sndmix.cpp"
           $306 = (($288) - ($x15$0))|0; //@line 643 "sndmix.cpp"
           $307 = (($305|0) / ($306|0))&-1; //@line 643 "sndmix.cpp"
           $308 = (($307) + ($envpan$0))|0; //@line 643 "sndmix.cpp"
           $envpan$1 = $308;
          } else {
           $envpan$1 = $envpan$0;
          }
          $309 = ($envpan$1|0)<(0); //@line 645 "sndmix.cpp"
          $$envpan$1 = $309 ? 0 : $envpan$1; //@line 645 "sndmix.cpp"
          $310 = ($$envpan$1|0)>(64); //@line 646 "sndmix.cpp"
          $311 = ($98|0)>(127); //@line 648 "sndmix.cpp"
          $$envpan$1$op = (($$envpan$1) + -32)|0; //@line 650 "sndmix.cpp"
          $312 = $310 ? 32 : $$envpan$1$op; //@line 650 "sndmix.cpp"
          $313 = (256 - ($98))|0; //@line 650 "sndmix.cpp"
          $$pn23 = $311 ? $313 : $98; //@line 648 "sndmix.cpp"
          $$pn$in = Math_imul($312, $$pn23)|0; //@line 650 "sndmix.cpp"
          $$pn = (($$pn$in|0) / 32)&-1; //@line 650 "sndmix.cpp"
          $pan$0 = (($$pn) + ($98))|0; //@line 650 "sndmix.cpp"
          $314 = ($pan$0|0)<(0); //@line 655 "sndmix.cpp"
          $$pan$0 = $314 ? 0 : $pan$0; //@line 655 "sndmix.cpp"
          $315 = ($$pan$0|0)>(256); //@line 656 "sndmix.cpp"
          $pan$2 = $315 ? 256 : $$pan$0; //@line 656 "sndmix.cpp"
          HEAP32[$103>>2] = $pan$2; //@line 657 "sndmix.cpp"
          $334 = $pan$2;
         }
        } while(0);
        $316 = $220 & 1024; //@line 660 "sndmix.cpp"
        $317 = ($316|0)==(0); //@line 660 "sndmix.cpp"
        do {
         if ($317) {
          $vol$8 = $vol$7;
         } else {
          $318 = HEAP32[$217>>2]|0; //@line 662 "sndmix.cpp"
          $319 = ($318|0)==(0); //@line 663 "sndmix.cpp"
          if ($319) {
           $327 = ((($pChn$068)) + 124|0); //@line 669 "sndmix.cpp"
           $328 = HEAP32[$327>>2]|0; //@line 669 "sndmix.cpp"
           $329 = ($328|0)==(0); //@line 669 "sndmix.cpp"
           $$vol$7 = $329 ? 0 : $vol$7; //@line 669 "sndmix.cpp"
           $vol$8 = $$vol$7;
           break;
          } else {
           $320 = $318 << 1; //@line 665 "sndmix.cpp"
           $321 = ((($pChn$068)) + 124|0); //@line 665 "sndmix.cpp"
           $322 = HEAP32[$321>>2]|0; //@line 665 "sndmix.cpp"
           $323 = (($322) - ($320))|0; //@line 665 "sndmix.cpp"
           $324 = ($323|0)>(0); //@line 666 "sndmix.cpp"
           $$29 = $324 ? $323 : 0; //@line 666 "sndmix.cpp"
           HEAP32[$321>>2] = $$29; //@line 665 "sndmix.cpp"
           $325 = Math_imul($$29, $vol$7)|0; //@line 667 "sndmix.cpp"
           $326 = $325 >> 16; //@line 667 "sndmix.cpp"
           $vol$8 = $326;
           break;
          }
         }
        } while(0);
        $330 = ((($217)) + 583|0); //@line 675 "sndmix.cpp"
        $331 = HEAP8[$330>>0]|0; //@line 675 "sndmix.cpp"
        $332 = ($331<<24>>24)==(0); //@line 675 "sndmix.cpp"
        $333 = ($334|0)==(0); //@line 675 "sndmix.cpp"
        $or$cond98 = $332 | $333; //@line 675 "sndmix.cpp"
        if ($or$cond98) {
         $vol$9 = $vol$8;
         label = 71;
         break;
        }
        $335 = ((($pChn$068)) + 212|0); //@line 675 "sndmix.cpp"
        $336 = HEAP8[$335>>0]|0; //@line 675 "sndmix.cpp"
        $337 = ($336<<24>>24)==(0); //@line 675 "sndmix.cpp"
        if ($337) {
         $vol$9 = $vol$8;
         label = 71;
         break;
        }
        $338 = $336&255; //@line 677 "sndmix.cpp"
        $339 = ((($217)) + 584|0); //@line 677 "sndmix.cpp"
        $340 = HEAP8[$339>>0]|0; //@line 677 "sndmix.cpp"
        $341 = $340&255; //@line 677 "sndmix.cpp"
        $342 = (($338) + -1)|0; //@line 677 "sndmix.cpp"
        $343 = (($342) - ($341))|0; //@line 677 "sndmix.cpp"
        $344 = $331 << 24 >> 24; //@line 677 "sndmix.cpp"
        $345 = Math_imul($343, $344)|0; //@line 677 "sndmix.cpp"
        $346 = (($345|0) / 8)&-1; //@line 677 "sndmix.cpp"
        $347 = (($346) + ($334))|0; //@line 677 "sndmix.cpp"
        $348 = ($347|0)<(0); //@line 678 "sndmix.cpp"
        $$5 = $348 ? 0 : $347; //@line 678 "sndmix.cpp"
        $349 = ($$5|0)>(256); //@line 679 "sndmix.cpp"
        $$$5 = $349 ? 256 : $$5; //@line 679 "sndmix.cpp"
        HEAP32[$103>>2] = $$$5; //@line 680 "sndmix.cpp"
        $vol$9 = $vol$8;
        label = 71;
       }
      } while(0);
      if ((label|0) == 71) {
       label = 0;
       $353 = ($vol$9|0)==(0); //@line 692 "sndmix.cpp"
       if (!($353)) {
        $354 = HEAP32[$67>>2]|0; //@line 696 "sndmix.cpp"
        $355 = Math_imul($354, $vol$9)|0; //@line 696 "sndmix.cpp"
        $356 = ((($pChn$068)) + 168|0); //@line 696 "sndmix.cpp"
        $357 = HEAP32[$356>>2]|0; //@line 696 "sndmix.cpp"
        $358 = ((($pChn$068)) + 172|0); //@line 696 "sndmix.cpp"
        $359 = HEAP32[$358>>2]|0; //@line 696 "sndmix.cpp"
        $360 = Math_imul($359, $357)|0; //@line 696 "sndmix.cpp"
        $361 = ($355|0)<(0); //@line 122 "sndmix.cpp"
        $362 = $361 << 31 >> 31; //@line 122 "sndmix.cpp"
        $363 = ($360|0)<(0); //@line 122 "sndmix.cpp"
        $364 = $363 << 31 >> 31; //@line 122 "sndmix.cpp"
        $365 = (___muldi3(($360|0),($364|0),($355|0),($362|0))|0); //@line 122 "sndmix.cpp"
        $366 = tempRet0; //@line 122 "sndmix.cpp"
        $367 = (_bitshift64Lshr(($365|0),($366|0),20)|0); //@line 122 "sndmix.cpp"
        $368 = tempRet0; //@line 122 "sndmix.cpp"
        HEAP32[$96>>2] = $367; //@line 696 "sndmix.cpp"
       }
      }
      $369 = HEAP32[$68>>2]|0; //@line 698 "sndmix.cpp"
      $370 = ($108|0)<($369|0); //@line 698 "sndmix.cpp"
      if ($370) {
       HEAP32[$107>>2] = $369; //@line 698 "sndmix.cpp"
       $373 = $369;
      } else {
       $373 = $108;
      }
      $371 = $220 & 1572864; //@line 700 "sndmix.cpp"
      $372 = ($371|0)==(1572864); //@line 700 "sndmix.cpp"
      if ($372) {
       $374 = (__ZNK10CSoundFile17GetNoteFromPeriodEj($this,$373)|0); //@line 702 "sndmix.cpp"
       $375 = ((($pChn$068)) + 176|0); //@line 702 "sndmix.cpp"
       $376 = HEAP32[$375>>2]|0; //@line 702 "sndmix.cpp"
       $377 = ((($pChn$068)) + 132|0); //@line 702 "sndmix.cpp"
       $378 = HEAP32[$377>>2]|0; //@line 702 "sndmix.cpp"
       $379 = (__ZNK10CSoundFile17GetPeriodFromNoteEjij($this,$374,$376,$378)|0); //@line 702 "sndmix.cpp"
       $$pre = HEAP8[$184>>0]|0; //@line 706 "sndmix.cpp"
       $381 = $$pre;$period$0 = $379;
      } else {
       $381 = $185;$period$0 = $373;
      }
      $380 = ($381<<24>>24)==(1); //@line 706 "sndmix.cpp"
      L112: do {
       if ($380) {
        $382 = HEAP32[$66>>2]|0; //@line 708 "sndmix.cpp"
        $383 = (($382>>>0) % 3)&-1; //@line 708 "sndmix.cpp"
        switch ($383|0) {
        case 1:  {
         $384 = ((($pChn$068)) + 212|0); //@line 710 "sndmix.cpp"
         $385 = HEAP8[$384>>0]|0; //@line 710 "sndmix.cpp"
         $386 = $385&255; //@line 710 "sndmix.cpp"
         $387 = ((($pChn$068)) + 217|0); //@line 710 "sndmix.cpp"
         $388 = HEAP8[$387>>0]|0; //@line 710 "sndmix.cpp"
         $389 = $388&255; //@line 710 "sndmix.cpp"
         $390 = $389 >>> 4; //@line 710 "sndmix.cpp"
         $391 = (($390) + ($386))|0; //@line 710 "sndmix.cpp"
         $392 = ((($pChn$068)) + 176|0); //@line 710 "sndmix.cpp"
         $393 = HEAP32[$392>>2]|0; //@line 710 "sndmix.cpp"
         $394 = ((($pChn$068)) + 132|0); //@line 710 "sndmix.cpp"
         $395 = HEAP32[$394>>2]|0; //@line 710 "sndmix.cpp"
         $396 = (__ZNK10CSoundFile17GetPeriodFromNoteEjij($this,$391,$393,$395)|0); //@line 710 "sndmix.cpp"
         $period$1 = $396;
         break L112;
         break;
        }
        case 2:  {
         $397 = ((($pChn$068)) + 212|0); //@line 711 "sndmix.cpp"
         $398 = HEAP8[$397>>0]|0; //@line 711 "sndmix.cpp"
         $399 = $398&255; //@line 711 "sndmix.cpp"
         $400 = ((($pChn$068)) + 217|0); //@line 711 "sndmix.cpp"
         $401 = HEAP8[$400>>0]|0; //@line 711 "sndmix.cpp"
         $402 = $401&255; //@line 711 "sndmix.cpp"
         $403 = $402 & 15; //@line 711 "sndmix.cpp"
         $404 = (($403) + ($399))|0; //@line 711 "sndmix.cpp"
         $405 = ((($pChn$068)) + 176|0); //@line 711 "sndmix.cpp"
         $406 = HEAP32[$405>>2]|0; //@line 711 "sndmix.cpp"
         $407 = ((($pChn$068)) + 132|0); //@line 711 "sndmix.cpp"
         $408 = HEAP32[$407>>2]|0; //@line 711 "sndmix.cpp"
         $409 = (__ZNK10CSoundFile17GetPeriodFromNoteEjij($this,$404,$406,$408)|0); //@line 711 "sndmix.cpp"
         $period$1 = $409;
         break L112;
         break;
        }
        default: {
         $period$1 = $period$0;
         break L112;
        }
        }
       } else {
        $period$1 = $period$0;
       }
      } while(0);
      $410 = HEAP32[$45>>2]|0; //@line 715 "sndmix.cpp"
      $411 = $410 & 65536; //@line 715 "sndmix.cpp"
      $412 = ($411|0)==(0); //@line 715 "sndmix.cpp"
      if ($412) {
       $period$3 = $period$1;
      } else {
       $413 = ($period$1|0)<(452); //@line 717 "sndmix.cpp"
       $$period$1 = $413 ? 452 : $period$1; //@line 717 "sndmix.cpp"
       $414 = ($$period$1|0)>(3424); //@line 718 "sndmix.cpp"
       $$$period$1 = $414 ? 3424 : $$period$1; //@line 718 "sndmix.cpp"
       $period$3 = $$$period$1;
      }
      $415 = HEAP32[$216>>2]|0; //@line 722 "sndmix.cpp"
      $416 = ($415|0)==(0|0); //@line 722 "sndmix.cpp"
      do {
       if ($416) {
        $period$4 = $period$3;
       } else {
        $417 = HEAP32[$71>>2]|0; //@line 722 "sndmix.cpp"
        $418 = $417 & 8388608; //@line 722 "sndmix.cpp"
        $419 = ($418|0)==(0); //@line 722 "sndmix.cpp"
        if ($419) {
         $period$4 = $period$3;
         break;
        }
        $420 = ((($415)) + 558|0); //@line 722 "sndmix.cpp"
        $421 = HEAP8[$420>>0]|0; //@line 722 "sndmix.cpp"
        $422 = ($421<<24>>24)==(0); //@line 722 "sndmix.cpp"
        if ($422) {
         $period$4 = $period$3;
         break;
        }
        $423 = ((($pChn$068)) + 156|0); //@line 725 "sndmix.cpp"
        $424 = HEAP32[$423>>2]|0; //@line 725 "sndmix.cpp"
        $425 = $421&255; //@line 726 "sndmix.cpp"
        $426 = (($425) + -1)|0; //@line 726 "sndmix.cpp"
        $427 = ($421<<24>>24)==(1); //@line 727 "sndmix.cpp"
        L125: do {
         if ($427) {
          $pt8$0 = $426;
         } else {
          $i9$057 = 0;
          while(1) {
           $428 = (((($415)) + 140|0) + ($i9$057<<1)|0); //@line 729 "sndmix.cpp"
           $429 = HEAP16[$428>>1]|0; //@line 729 "sndmix.cpp"
           $430 = $429&65535; //@line 729 "sndmix.cpp"
           $431 = ($424|0)>($430|0); //@line 729 "sndmix.cpp"
           if (!($431)) {
            $pt8$0 = $i9$057;
            break L125;
           }
           $432 = (($i9$057) + 1)|0; //@line 727 "sndmix.cpp"
           $433 = ($432>>>0)<($426>>>0); //@line 727 "sndmix.cpp"
           if ($433) {
            $i9$057 = $432;
           } else {
            $pt8$0 = $426;
            break;
           }
          }
         }
        } while(0);
        $434 = (((($415)) + 140|0) + ($pt8$0<<1)|0); //@line 735 "sndmix.cpp"
        $435 = HEAP16[$434>>1]|0; //@line 735 "sndmix.cpp"
        $436 = $435&65535; //@line 735 "sndmix.cpp"
        $437 = ($424|0)<($436|0); //@line 737 "sndmix.cpp"
        do {
         if ($437) {
          $443 = ($pt8$0|0)==(0); //@line 742 "sndmix.cpp"
          if ($443) {
           $envpitch$0 = 0;$x111$0 = 0;
           break;
          }
          $444 = (($pt8$0) + -1)|0; //@line 744 "sndmix.cpp"
          $445 = (((($415)) + 268|0) + ($444)|0); //@line 744 "sndmix.cpp"
          $446 = HEAP8[$445>>0]|0; //@line 744 "sndmix.cpp"
          $447 = $446&255; //@line 744 "sndmix.cpp"
          $448 = $447 << 3; //@line 744 "sndmix.cpp"
          $449 = (($448) + -256)|0; //@line 744 "sndmix.cpp"
          $450 = (((($415)) + 140|0) + ($444<<1)|0); //@line 745 "sndmix.cpp"
          $451 = HEAP16[$450>>1]|0; //@line 745 "sndmix.cpp"
          $452 = $451&65535; //@line 745 "sndmix.cpp"
          $envpitch$0 = $449;$x111$0 = $452;
         } else {
          $438 = (((($415)) + 268|0) + ($pt8$0)|0); //@line 739 "sndmix.cpp"
          $439 = HEAP8[$438>>0]|0; //@line 739 "sndmix.cpp"
          $440 = $439&255; //@line 739 "sndmix.cpp"
          $441 = $440 << 3; //@line 739 "sndmix.cpp"
          $442 = (($441) + -256)|0; //@line 739 "sndmix.cpp"
          $envpitch$0 = $442;$x111$0 = $436;
         }
        } while(0);
        $453 = ($424|0)>($436|0); //@line 751 "sndmix.cpp"
        $$30 = $453 ? $436 : $424; //@line 751 "sndmix.cpp"
        $454 = ($436>>>0)>($x111$0>>>0); //@line 752 "sndmix.cpp"
        $455 = ($$30|0)>($x111$0|0); //@line 752 "sndmix.cpp"
        $or$cond31 = $454 & $455; //@line 752 "sndmix.cpp"
        if ($or$cond31) {
         $456 = (((($415)) + 268|0) + ($pt8$0)|0); //@line 754 "sndmix.cpp"
         $457 = HEAP8[$456>>0]|0; //@line 754 "sndmix.cpp"
         $458 = $457&255; //@line 754 "sndmix.cpp"
         $459 = $458 << 3; //@line 754 "sndmix.cpp"
         $460 = (($$30) - ($x111$0))|0; //@line 755 "sndmix.cpp"
         $461 = (-256 - ($envpitch$0))|0; //@line 754 "sndmix.cpp"
         $462 = (($461) + ($459))|0; //@line 755 "sndmix.cpp"
         $463 = Math_imul($462, $460)|0; //@line 755 "sndmix.cpp"
         $464 = (($436) - ($x111$0))|0; //@line 755 "sndmix.cpp"
         $465 = (($463|0) / ($464|0))&-1; //@line 755 "sndmix.cpp"
         $466 = (($465) + ($envpitch$0))|0; //@line 755 "sndmix.cpp"
         $envpitch$1 = $466;
        } else {
         $envpitch$1 = $envpitch$0;
        }
        $467 = ($envpitch$1|0)<(-256); //@line 757 "sndmix.cpp"
        $$envpitch$1 = $467 ? -256 : $envpitch$1; //@line 757 "sndmix.cpp"
        $468 = ($$envpitch$1|0)>(256); //@line 758 "sndmix.cpp"
        $envpitch$3 = $468 ? 256 : $$envpitch$1; //@line 758 "sndmix.cpp"
        $469 = ((($415)) + 4|0); //@line 760 "sndmix.cpp"
        $470 = HEAP32[$469>>2]|0; //@line 760 "sndmix.cpp"
        $471 = $470 & 1024; //@line 760 "sndmix.cpp"
        $472 = ($471|0)==(0); //@line 760 "sndmix.cpp"
        if (!($472)) {
         $473 = $417 & 16384; //@line 763 "sndmix.cpp"
         $not$ = ($473|0)==(0); //@line 763 "sndmix.cpp"
         __ZNK10CSoundFile18SetupChannelFilterEP11_MODCHANNELbi($this,$pChn$068,$not$,$envpitch$3); //@line 763 "sndmix.cpp"
         $period$4 = $period$3;
         break;
        }
        $474 = ($envpitch$3|0)<(0); //@line 769 "sndmix.cpp"
        if ($474) {
         $475 = (0 - ($envpitch$3))|0; //@line 771 "sndmix.cpp"
         $476 = ($475|0)>(255); //@line 772 "sndmix.cpp"
         $$6 = $476 ? 255 : $475; //@line 772 "sndmix.cpp"
         $477 = (40660 + ($$6<<2)|0); //@line 773 "sndmix.cpp"
         $478 = HEAP32[$477>>2]|0; //@line 773 "sndmix.cpp"
         $479 = ($period$3|0)<(0); //@line 122 "sndmix.cpp"
         $480 = $479 << 31 >> 31; //@line 122 "sndmix.cpp"
         $481 = ($478|0)<(0); //@line 122 "sndmix.cpp"
         $482 = $481 << 31 >> 31; //@line 122 "sndmix.cpp"
         $483 = (___muldi3(($478|0),($482|0),($period$3|0),($480|0))|0); //@line 122 "sndmix.cpp"
         $484 = tempRet0; //@line 122 "sndmix.cpp"
         $485 = (_bitshift64Lshr(($483|0),($484|0),16)|0); //@line 122 "sndmix.cpp"
         $486 = tempRet0; //@line 122 "sndmix.cpp"
         $period$4 = $485;
         break;
        } else {
         $487 = ($envpitch$3|0)>(255); //@line 776 "sndmix.cpp"
         $$envpitch$3 = $487 ? 255 : $$envpitch$1; //@line 776 "sndmix.cpp"
         $488 = (41684 + ($$envpitch$3<<2)|0); //@line 777 "sndmix.cpp"
         $489 = HEAP32[$488>>2]|0; //@line 777 "sndmix.cpp"
         $490 = ($period$3|0)<(0); //@line 122 "sndmix.cpp"
         $491 = $490 << 31 >> 31; //@line 122 "sndmix.cpp"
         $492 = ($489|0)<(0); //@line 122 "sndmix.cpp"
         $493 = $492 << 31 >> 31; //@line 122 "sndmix.cpp"
         $494 = (___muldi3(($489|0),($493|0),($period$3|0),($491|0))|0); //@line 122 "sndmix.cpp"
         $495 = tempRet0; //@line 122 "sndmix.cpp"
         $496 = (_bitshift64Lshr(($494|0),($495|0),16)|0); //@line 122 "sndmix.cpp"
         $497 = tempRet0; //@line 122 "sndmix.cpp"
         $period$4 = $496;
         break;
        }
       }
      } while(0);
      $498 = HEAP32[$71>>2]|0; //@line 783 "sndmix.cpp"
      $499 = $498 & 65536; //@line 783 "sndmix.cpp"
      $500 = ($499|0)==(0); //@line 783 "sndmix.cpp"
      do {
       if ($500) {
        $period$5 = $period$4;
       } else {
        $501 = ((($pChn$068)) + 196|0); //@line 785 "sndmix.cpp"
        $502 = HEAP32[$501>>2]|0; //@line 785 "sndmix.cpp"
        $503 = ((($pChn$068)) + 224|0); //@line 787 "sndmix.cpp"
        $504 = HEAP8[$503>>0]|0; //@line 787 "sndmix.cpp"
        $505 = $504&255; //@line 787 "sndmix.cpp"
        $506 = $505 & 3; //@line 787 "sndmix.cpp"
        switch ($506|0) {
        case 1:  {
         $507 = (169324 + ($502<<1)|0); //@line 790 "sndmix.cpp"
         $508 = HEAP16[$507>>1]|0; //@line 790 "sndmix.cpp"
         $509 = $508 << 16 >> 16; //@line 790 "sndmix.cpp"
         $vdelta$0 = $509;
         break;
        }
        case 2:  {
         $510 = (169452 + ($502<<1)|0); //@line 793 "sndmix.cpp"
         $511 = HEAP16[$510>>1]|0; //@line 793 "sndmix.cpp"
         $512 = $511 << 16 >> 16; //@line 793 "sndmix.cpp"
         $vdelta$0 = $512;
         break;
        }
        case 3:  {
         $513 = (169580 + ($502<<1)|0); //@line 796 "sndmix.cpp"
         $514 = HEAP16[$513>>1]|0; //@line 796 "sndmix.cpp"
         $515 = $514 << 16 >> 16; //@line 796 "sndmix.cpp"
         $vdelta$0 = $515;
         break;
        }
        default: {
         $516 = (169708 + ($502<<1)|0); //@line 799 "sndmix.cpp"
         $517 = HEAP16[$516>>1]|0; //@line 799 "sndmix.cpp"
         $518 = $517 << 16 >> 16; //@line 799 "sndmix.cpp"
         $vdelta$0 = $518;
        }
        }
        $519 = HEAP32[$18>>2]|0; //@line 801 "sndmix.cpp"
        $520 = HEAP32[$45>>2]|0; //@line 801 "sndmix.cpp"
        $521 = $520 & 4; //@line 801 "sndmix.cpp"
        $522 = ($521|0)!=(0); //@line 801 "sndmix.cpp"
        $not$128 = ($519|0)!=(32); //@line 801 "sndmix.cpp"
        $523 = $522 | $not$128; //@line 801 "sndmix.cpp"
        $524 = $523 ? 6 : 7; //@line 801 "sndmix.cpp"
        $525 = ((($pChn$068)) + 226|0); //@line 802 "sndmix.cpp"
        $526 = HEAP8[$525>>0]|0; //@line 802 "sndmix.cpp"
        $527 = $526&255; //@line 802 "sndmix.cpp"
        $528 = Math_imul($527, $vdelta$0)|0; //@line 802 "sndmix.cpp"
        $529 = $528 >> $524; //@line 802 "sndmix.cpp"
        $530 = $520 & 16; //@line 803 "sndmix.cpp"
        $531 = ($530|0)==(0); //@line 803 "sndmix.cpp"
        $532 = $519 & 32;
        $533 = ($532|0)==(0); //@line 803 "sndmix.cpp"
        $or$cond100 = $531 | $533; //@line 803 "sndmix.cpp"
        do {
         if ($or$cond100) {
          $vdelta$1 = $529;
         } else {
          $534 = ($529|0)<(0); //@line 806 "sndmix.cpp"
          if ($534) {
           $535 = (0 - ($529))|0; //@line 808 "sndmix.cpp"
           $536 = $535 >> 2; //@line 809 "sndmix.cpp"
           $537 = (41684 + ($536<<2)|0); //@line 809 "sndmix.cpp"
           $538 = HEAP32[$537>>2]|0; //@line 809 "sndmix.cpp"
           $539 = ($period$4|0)<(0); //@line 122 "sndmix.cpp"
           $540 = $539 << 31 >> 31; //@line 122 "sndmix.cpp"
           $541 = ($538|0)<(0); //@line 122 "sndmix.cpp"
           $542 = $541 << 31 >> 31; //@line 122 "sndmix.cpp"
           $543 = (___muldi3(($538|0),($542|0),($period$4|0),($540|0))|0); //@line 122 "sndmix.cpp"
           $544 = tempRet0; //@line 122 "sndmix.cpp"
           $545 = (_bitshift64Lshr(($543|0),($544|0),16)|0); //@line 122 "sndmix.cpp"
           $546 = tempRet0; //@line 122 "sndmix.cpp"
           $547 = (($545) - ($period$4))|0; //@line 809 "sndmix.cpp"
           $548 = $535 & 3; //@line 810 "sndmix.cpp"
           $549 = ($548|0)==(0); //@line 810 "sndmix.cpp"
           if ($549) {
            $vdelta$1 = $547;
            break;
           }
           $550 = (42708 + ($548<<2)|0); //@line 810 "sndmix.cpp"
           $551 = HEAP32[$550>>2]|0; //@line 810 "sndmix.cpp"
           $552 = ($551|0)<(0); //@line 122 "sndmix.cpp"
           $553 = $552 << 31 >> 31; //@line 122 "sndmix.cpp"
           $554 = (___muldi3(($551|0),($553|0),($period$4|0),($540|0))|0); //@line 122 "sndmix.cpp"
           $555 = tempRet0; //@line 122 "sndmix.cpp"
           $556 = (_bitshift64Lshr(($554|0),($555|0),16)|0); //@line 122 "sndmix.cpp"
           $557 = tempRet0; //@line 122 "sndmix.cpp"
           $558 = (($547) - ($period$4))|0; //@line 810 "sndmix.cpp"
           $559 = (($558) + ($556))|0; //@line 810 "sndmix.cpp"
           $vdelta$1 = $559;
           break;
          } else {
           $560 = $529 >> 2; //@line 814 "sndmix.cpp"
           $561 = (40660 + ($560<<2)|0); //@line 814 "sndmix.cpp"
           $562 = HEAP32[$561>>2]|0; //@line 814 "sndmix.cpp"
           $563 = ($period$4|0)<(0); //@line 122 "sndmix.cpp"
           $564 = $563 << 31 >> 31; //@line 122 "sndmix.cpp"
           $565 = ($562|0)<(0); //@line 122 "sndmix.cpp"
           $566 = $565 << 31 >> 31; //@line 122 "sndmix.cpp"
           $567 = (___muldi3(($562|0),($566|0),($period$4|0),($564|0))|0); //@line 122 "sndmix.cpp"
           $568 = tempRet0; //@line 122 "sndmix.cpp"
           $569 = (_bitshift64Lshr(($567|0),($568|0),16)|0); //@line 122 "sndmix.cpp"
           $570 = tempRet0; //@line 122 "sndmix.cpp"
           $571 = (($569) - ($period$4))|0; //@line 814 "sndmix.cpp"
           $572 = $529 & 3; //@line 815 "sndmix.cpp"
           $573 = ($572|0)==(0); //@line 815 "sndmix.cpp"
           if ($573) {
            $vdelta$1 = $571;
            break;
           }
           $574 = (42772 + ($572<<2)|0); //@line 815 "sndmix.cpp"
           $575 = HEAP32[$574>>2]|0; //@line 815 "sndmix.cpp"
           $576 = ($575|0)<(0); //@line 122 "sndmix.cpp"
           $577 = $576 << 31 >> 31; //@line 122 "sndmix.cpp"
           $578 = (___muldi3(($575|0),($577|0),($period$4|0),($564|0))|0); //@line 122 "sndmix.cpp"
           $579 = tempRet0; //@line 122 "sndmix.cpp"
           $580 = (_bitshift64Lshr(($578|0),($579|0),16)|0); //@line 122 "sndmix.cpp"
           $581 = tempRet0; //@line 122 "sndmix.cpp"
           $582 = (($571) - ($period$4))|0; //@line 815 "sndmix.cpp"
           $583 = (($582) + ($580))|0; //@line 815 "sndmix.cpp"
           $vdelta$1 = $583;
           break;
          }
         }
        } while(0);
        $584 = (($vdelta$1) + ($period$4))|0; //@line 819 "sndmix.cpp"
        $585 = HEAP32[$66>>2]|0; //@line 820 "sndmix.cpp"
        $586 = ($585|0)==(0); //@line 820 "sndmix.cpp"
        if ($586) {
         $587 = $519 & 32; //@line 820 "sndmix.cpp"
         $588 = ($587|0)!=(0); //@line 820 "sndmix.cpp"
         $589 = $520 & 4;
         $590 = ($589|0)==(0); //@line 820 "sndmix.cpp"
         $or$cond102 = $588 & $590; //@line 820 "sndmix.cpp"
         if (!($or$cond102)) {
          $period$5 = $584;
          break;
         }
        }
        $591 = ((($pChn$068)) + 225|0); //@line 822 "sndmix.cpp"
        $592 = HEAP8[$591>>0]|0; //@line 822 "sndmix.cpp"
        $593 = $592&255; //@line 822 "sndmix.cpp"
        $594 = (($593) + ($502))|0; //@line 822 "sndmix.cpp"
        $595 = $594 & 63; //@line 822 "sndmix.cpp"
        HEAP32[$501>>2] = $595; //@line 822 "sndmix.cpp"
        $period$5 = $584;
       }
      } while(0);
      $596 = $498 & 262144; //@line 826 "sndmix.cpp"
      $597 = ($596|0)==(0); //@line 826 "sndmix.cpp"
      if (!($597)) {
       $598 = ((($pChn$068)) + 204|0); //@line 828 "sndmix.cpp"
       $599 = HEAP32[$598>>2]|0; //@line 828 "sndmix.cpp"
       $600 = (($599) + 16)|0; //@line 828 "sndmix.cpp"
       $601 = $600 >>> 2; //@line 828 "sndmix.cpp"
       $602 = $601 & 63; //@line 828 "sndmix.cpp"
       $603 = ((($pChn$068)) + 230|0); //@line 830 "sndmix.cpp"
       $604 = HEAP8[$603>>0]|0; //@line 830 "sndmix.cpp"
       $605 = $604&255; //@line 830 "sndmix.cpp"
       $606 = $605 & 3; //@line 830 "sndmix.cpp"
       switch ($606|0) {
       case 1:  {
        $607 = (169324 + ($602<<1)|0); //@line 833 "sndmix.cpp"
        $608 = HEAP16[$607>>1]|0; //@line 833 "sndmix.cpp"
        $609 = $608 << 16 >> 16; //@line 833 "sndmix.cpp"
        $pdelta$0 = $609;
        break;
       }
       case 2:  {
        $610 = (169452 + ($602<<1)|0); //@line 836 "sndmix.cpp"
        $611 = HEAP16[$610>>1]|0; //@line 836 "sndmix.cpp"
        $612 = $611 << 16 >> 16; //@line 836 "sndmix.cpp"
        $pdelta$0 = $612;
        break;
       }
       case 3:  {
        $613 = (169580 + ($602<<1)|0); //@line 839 "sndmix.cpp"
        $614 = HEAP16[$613>>1]|0; //@line 839 "sndmix.cpp"
        $615 = $614 << 16 >> 16; //@line 839 "sndmix.cpp"
        $pdelta$0 = $615;
        break;
       }
       default: {
        $616 = (169708 + ($602<<1)|0); //@line 842 "sndmix.cpp"
        $617 = HEAP16[$616>>1]|0; //@line 842 "sndmix.cpp"
        $618 = $617 << 16 >> 16; //@line 842 "sndmix.cpp"
        $pdelta$0 = $618;
       }
       }
       $619 = ((($pChn$068)) + 231|0); //@line 844 "sndmix.cpp"
       $620 = HEAP8[$619>>0]|0; //@line 844 "sndmix.cpp"
       $621 = $620&255; //@line 844 "sndmix.cpp"
       $622 = (($599) + ($621))|0; //@line 844 "sndmix.cpp"
       HEAP32[$598>>2] = $622; //@line 844 "sndmix.cpp"
       $623 = ((($pChn$068)) + 232|0); //@line 845 "sndmix.cpp"
       $624 = HEAP8[$623>>0]|0; //@line 845 "sndmix.cpp"
       $625 = $624&255; //@line 845 "sndmix.cpp"
       $626 = Math_imul($625, $pdelta$0)|0; //@line 845 "sndmix.cpp"
       $627 = (($626) + 2)|0; //@line 845 "sndmix.cpp"
       $628 = $627 >> 3; //@line 845 "sndmix.cpp"
       $629 = HEAP32[$103>>2]|0; //@line 846 "sndmix.cpp"
       $630 = (($628) + ($629))|0; //@line 846 "sndmix.cpp"
       $631 = ($630|0)<(0); //@line 847 "sndmix.cpp"
       $$7 = $631 ? 0 : $630; //@line 847 "sndmix.cpp"
       $632 = ($$7|0)>(256); //@line 848 "sndmix.cpp"
       $$$7 = $632 ? 256 : $$7; //@line 848 "sndmix.cpp"
       HEAP32[$103>>2] = $$$7; //@line 849 "sndmix.cpp"
      }
      $633 = ((($pChn$068)) + 144|0); //@line 853 "sndmix.cpp"
      $634 = HEAP32[$633>>2]|0; //@line 853 "sndmix.cpp"
      $635 = ($634|0)==(0|0); //@line 853 "sndmix.cpp"
      do {
       if ($635) {
        $nPeriodFrac$0 = 0;$period$6 = $period$5;
       } else {
        $636 = ((($634)) + 40|0); //@line 853 "sndmix.cpp"
        $637 = HEAP8[$636>>0]|0; //@line 853 "sndmix.cpp"
        $638 = ($637<<24>>24)==(0); //@line 853 "sndmix.cpp"
        if ($638) {
         $nPeriodFrac$0 = 0;$period$6 = $period$5;
         break;
        }
        $639 = ((($634)) + 39|0); //@line 856 "sndmix.cpp"
        $640 = HEAP8[$639>>0]|0; //@line 856 "sndmix.cpp"
        $641 = $640&255; //@line 856 "sndmix.cpp"
        $642 = ($640<<24>>24)==(0); //@line 856 "sndmix.cpp"
        do {
         if ($642) {
          $643 = $637&255; //@line 858 "sndmix.cpp"
          $644 = $643 << 8; //@line 858 "sndmix.cpp"
          $645 = ((($pChn$068)) + 188|0); //@line 858 "sndmix.cpp"
          HEAP32[$645>>2] = $644; //@line 858 "sndmix.cpp"
          $695 = $644;
         } else {
          $646 = HEAP32[$18>>2]|0; //@line 861 "sndmix.cpp"
          $647 = $646 & 32; //@line 861 "sndmix.cpp"
          $648 = ($647|0)==(0); //@line 861 "sndmix.cpp"
          do {
           if ($648) {
            $653 = $498 & 512; //@line 865 "sndmix.cpp"
            $654 = ($653|0)==(0); //@line 865 "sndmix.cpp"
            if ($654) {
             $655 = $637&255; //@line 867 "sndmix.cpp"
             $656 = $655 << 8; //@line 867 "sndmix.cpp"
             $657 = (($656>>>0) / ($641>>>0))&-1; //@line 867 "sndmix.cpp"
             $658 = ((($pChn$068)) + 188|0); //@line 867 "sndmix.cpp"
             $659 = HEAP32[$658>>2]|0; //@line 867 "sndmix.cpp"
             $660 = (($659) + ($657))|0; //@line 867 "sndmix.cpp"
             HEAP32[$658>>2] = $660; //@line 867 "sndmix.cpp"
             $$pre$phi89Z2D = $658;$662 = $660;
             break;
            } else {
             $$phi$trans$insert = ((($pChn$068)) + 188|0);
             $$pre77 = HEAP32[$$phi$trans$insert>>2]|0; //@line 869 "sndmix.cpp"
             $$pre$phi89Z2D = $$phi$trans$insert;$662 = $$pre77;
             break;
            }
           } else {
            $649 = $641 << 3; //@line 863 "sndmix.cpp"
            $650 = ((($pChn$068)) + 188|0); //@line 863 "sndmix.cpp"
            $651 = HEAP32[$650>>2]|0; //@line 863 "sndmix.cpp"
            $652 = (($651) + ($649))|0; //@line 863 "sndmix.cpp"
            HEAP32[$650>>2] = $652; //@line 863 "sndmix.cpp"
            $$pre$phi89Z2D = $650;$662 = $652;
           }
          } while(0);
          $661 = $662 >> 8; //@line 869 "sndmix.cpp"
          $663 = $637&255; //@line 869 "sndmix.cpp"
          $664 = ($661|0)>($663|0); //@line 869 "sndmix.cpp"
          if (!($664)) {
           $695 = $662;
           break;
          }
          $665 = $663 << 8; //@line 870 "sndmix.cpp"
          HEAP32[$$pre$phi89Z2D>>2] = $665; //@line 870 "sndmix.cpp"
          $695 = $665;
         }
        } while(0);
        $666 = ((($634)) + 41|0); //@line 872 "sndmix.cpp"
        $667 = HEAP8[$666>>0]|0; //@line 872 "sndmix.cpp"
        $668 = $667&255; //@line 872 "sndmix.cpp"
        $669 = ((($pChn$068)) + 192|0); //@line 872 "sndmix.cpp"
        $670 = HEAP32[$669>>2]|0; //@line 872 "sndmix.cpp"
        $671 = (($670) + ($668))|0; //@line 872 "sndmix.cpp"
        HEAP32[$669>>2] = $671; //@line 872 "sndmix.cpp"
        $672 = ((($634)) + 38|0); //@line 874 "sndmix.cpp"
        $673 = HEAP8[$672>>0]|0; //@line 874 "sndmix.cpp"
        $674 = $673&255; //@line 874 "sndmix.cpp"
        switch ($674|0) {
        case 4:  {
         $675 = $671 & 63; //@line 877 "sndmix.cpp"
         $676 = (169580 + ($675<<1)|0); //@line 877 "sndmix.cpp"
         $677 = HEAP16[$676>>1]|0; //@line 877 "sndmix.cpp"
         $678 = $677 << 16 >> 16; //@line 877 "sndmix.cpp"
         $679 = (($671) + 1)|0; //@line 878 "sndmix.cpp"
         HEAP32[$669>>2] = $679; //@line 878 "sndmix.cpp"
         $val$0 = $678;
         break;
        }
        case 3:  {
         $680 = $671 >>> 1; //@line 881 "sndmix.cpp"
         $681 = (64 - ($680))|0; //@line 881 "sndmix.cpp"
         $682 = $681 & 127; //@line 881 "sndmix.cpp"
         $683 = (($682) + -64)|0; //@line 881 "sndmix.cpp"
         $val$0 = $683;
         break;
        }
        case 2:  {
         $684 = $671 >>> 1; //@line 884 "sndmix.cpp"
         $685 = (($684) + 64)|0; //@line 884 "sndmix.cpp"
         $686 = $685 & 127; //@line 884 "sndmix.cpp"
         $687 = (($686) + -64)|0; //@line 884 "sndmix.cpp"
         $val$0 = $687;
         break;
        }
        case 1:  {
         $688 = $671 & 128; //@line 887 "sndmix.cpp"
         $689 = (($688) + -64)|0; //@line 887 "sndmix.cpp"
         $val$0 = $689;
         break;
        }
        default: {
         $690 = $671 & 255; //@line 890 "sndmix.cpp"
         $691 = (523983 + ($690)|0); //@line 890 "sndmix.cpp"
         $692 = HEAP8[$691>>0]|0; //@line 890 "sndmix.cpp"
         $693 = $692 << 24 >> 24; //@line 890 "sndmix.cpp"
         $val$0 = $693;
        }
        }
        $694 = Math_imul($695, $val$0)|0; //@line 892 "sndmix.cpp"
        $696 = $694 >> 8; //@line 892 "sndmix.cpp"
        $697 = HEAP32[$18>>2]|0; //@line 893 "sndmix.cpp"
        $698 = $697 & 32; //@line 893 "sndmix.cpp"
        $699 = ($698|0)==(0); //@line 893 "sndmix.cpp"
        if ($699) {
         $726 = $694 >> 14; //@line 914 "sndmix.cpp"
         $727 = (($726) + ($period$5))|0; //@line 914 "sndmix.cpp"
         $nPeriodFrac$0 = 0;$period$6 = $727;
         break;
        }
        $700 = ($696|0)<(0); //@line 896 "sndmix.cpp"
        if ($700) {
         $701 = (0 - ($696))|0; //@line 898 "sndmix.cpp"
         $702 = $701 >> 8; //@line 899 "sndmix.cpp"
         $703 = (40660 + ($702<<2)|0); //@line 900 "sndmix.cpp"
         $704 = (($702) + 1)|0; //@line 901 "sndmix.cpp"
         $705 = (40660 + ($704<<2)|0); //@line 901 "sndmix.cpp"
         $df1$0$in = $703;$df2$0$in = $705;$n13$0 = $701;
        } else {
         $706 = $694 >> 16; //@line 904 "sndmix.cpp"
         $707 = (41684 + ($706<<2)|0); //@line 905 "sndmix.cpp"
         $708 = (($706) + 1)|0; //@line 906 "sndmix.cpp"
         $709 = (41684 + ($708<<2)|0); //@line 906 "sndmix.cpp"
         $df1$0$in = $707;$df2$0$in = $709;$n13$0 = $696;
        }
        $df2$0 = HEAP32[$df2$0$in>>2]|0; //@line 901 "sndmix.cpp"
        $df1$0 = HEAP32[$df1$0$in>>2]|0; //@line 900 "sndmix.cpp"
        $710 = $n13$0 >>> 2; //@line 908 "sndmix.cpp"
        $711 = (($df2$0) - ($df1$0))|0; //@line 909 "sndmix.cpp"
        $712 = $710 & 63; //@line 909 "sndmix.cpp"
        $713 = Math_imul($711, $712)|0; //@line 909 "sndmix.cpp"
        $714 = $713 >> 6; //@line 909 "sndmix.cpp"
        $715 = (($714) + ($df1$0))|0; //@line 909 "sndmix.cpp"
        $716 = ($period$5|0)<(0); //@line 122 "sndmix.cpp"
        $717 = $716 << 31 >> 31; //@line 122 "sndmix.cpp"
        $718 = ($715|0)<(0); //@line 122 "sndmix.cpp"
        $719 = $718 << 31 >> 31; //@line 122 "sndmix.cpp"
        $720 = (___muldi3(($715|0),($719|0),($period$5|0),($717|0))|0); //@line 122 "sndmix.cpp"
        $721 = tempRet0; //@line 122 "sndmix.cpp"
        $722 = (_bitshift64Lshr(($720|0),($721|0),8)|0); //@line 122 "sndmix.cpp"
        $723 = tempRet0; //@line 122 "sndmix.cpp"
        $724 = $722 & 255; //@line 910 "sndmix.cpp"
        $725 = $722 >> 8; //@line 911 "sndmix.cpp"
        $nPeriodFrac$0 = $724;$period$6 = $725;
       }
      } while(0);
      $728 = HEAP32[$68>>2]|0; //@line 918 "sndmix.cpp"
      $729 = ($period$6|0)>($728|0); //@line 918 "sndmix.cpp"
      do {
       if ($729) {
        $period$7 = $period$6;
       } else {
        $730 = HEAP32[$18>>2]|0; //@line 920 "sndmix.cpp"
        $731 = $730 & 2; //@line 920 "sndmix.cpp"
        $732 = ($731|0)==(0); //@line 920 "sndmix.cpp"
        if ($732) {
         $period$7 = $728;
         break;
        }
        HEAP32[$110>>2] = 0; //@line 920 "sndmix.cpp"
        $period$7 = $728;
       }
      } while(0);
      $733 = HEAP32[$69>>2]|0; //@line 923 "sndmix.cpp"
      $734 = ($period$7|0)>($733|0); //@line 923 "sndmix.cpp"
      do {
       if ($734) {
        $735 = HEAP32[$18>>2]|0; //@line 925 "sndmix.cpp"
        $736 = $735 & 32; //@line 925 "sndmix.cpp"
        $737 = ($736|0)!=(0); //@line 925 "sndmix.cpp"
        $738 = ($period$7|0)>(1048575); //@line 925 "sndmix.cpp"
        $or$cond9 = $738 | $737; //@line 925 "sndmix.cpp"
        if (!($or$cond9)) {
         $nPeriodFrac$1 = 0;$period$8 = $733;
         break;
        }
        $739 = ((($pChn$068)) + 124|0); //@line 927 "sndmix.cpp"
        HEAP32[$739>>2] = 0; //@line 927 "sndmix.cpp"
        $740 = $498 | 1024; //@line 928 "sndmix.cpp"
        HEAP32[$71>>2] = $740; //@line 928 "sndmix.cpp"
        HEAP32[$96>>2] = 0; //@line 929 "sndmix.cpp"
        $nPeriodFrac$1 = 0;$period$8 = $733;
       } else {
        $nPeriodFrac$1 = $nPeriodFrac$0;$period$8 = $period$7;
       }
      } while(0);
      $741 = ((($pChn$068)) + 132|0); //@line 934 "sndmix.cpp"
      $742 = HEAP32[$741>>2]|0; //@line 934 "sndmix.cpp"
      $743 = (__ZNK10CSoundFile17GetFreqFromPeriodEjji($this,$period$8,$742,$nPeriodFrac$1)|0); //@line 934 "sndmix.cpp"
      $744 = HEAP32[$18>>2]|0; //@line 935 "sndmix.cpp"
      $745 = $744 & 32; //@line 935 "sndmix.cpp"
      $746 = ($745|0)!=(0); //@line 935 "sndmix.cpp"
      $747 = ($743>>>0)<(256); //@line 935 "sndmix.cpp"
      $or$cond11 = $747 & $746; //@line 935 "sndmix.cpp"
      if ($or$cond11) {
       $748 = ((($pChn$068)) + 124|0); //@line 937 "sndmix.cpp"
       HEAP32[$748>>2] = 0; //@line 937 "sndmix.cpp"
       $749 = HEAP32[$71>>2]|0; //@line 938 "sndmix.cpp"
       $750 = $749 | 1024; //@line 938 "sndmix.cpp"
       HEAP32[$71>>2] = $750; //@line 938 "sndmix.cpp"
       HEAP32[$96>>2] = 0; //@line 939 "sndmix.cpp"
      }
      $751 = HEAP32[10109]|0; //@line 941 "sndmix.cpp"
      $752 = ($743|0)<(0); //@line 122 "sndmix.cpp"
      $753 = $752 << 31 >> 31; //@line 122 "sndmix.cpp"
      $754 = (_bitshift64Shl(($743|0),($753|0),16)|0); //@line 122 "sndmix.cpp"
      $755 = tempRet0; //@line 122 "sndmix.cpp"
      $756 = ($751|0)<(0); //@line 122 "sndmix.cpp"
      $757 = $756 << 31 >> 31; //@line 122 "sndmix.cpp"
      $758 = (___udivdi3(($754|0),($755|0),($751|0),($757|0))|0); //@line 122 "sndmix.cpp"
      $759 = tempRet0; //@line 122 "sndmix.cpp"
      $$off = (($758) + -65456)|0; //@line 942 "sndmix.cpp"
      $760 = ($$off>>>0)<(225); //@line 942 "sndmix.cpp"
      $$20 = $760 ? 65536 : $758; //@line 942 "sndmix.cpp"
      $761 = HEAP32[$70>>2]|0; //@line 943 "sndmix.cpp"
      $762 = ($761|0)==(128); //@line 943 "sndmix.cpp"
      $763 = Math_imul($$20, $761)|0; //@line 943 "sndmix.cpp"
      $764 = $763 >>> 7; //@line 943 "sndmix.cpp"
      $ninc$0 = $762 ? $$20 : $764; //@line 943 "sndmix.cpp"
      $765 = ($ninc$0>>>0)>(16711680); //@line 944 "sndmix.cpp"
      $ninc$0$op = (($ninc$0) + 1)|0; //@line 945 "sndmix.cpp"
      $ninc$0$op$op = $ninc$0$op & -4; //@line 945 "sndmix.cpp"
      $766 = $765 ? 16711680 : $ninc$0$op$op; //@line 945 "sndmix.cpp"
      HEAP32[$95>>2] = $766; //@line 945 "sndmix.cpp"
      $961 = $766;
     }
    }
    $767 = ((($pChn$068)) + 140|0); //@line 949 "sndmix.cpp"
    $768 = HEAP32[$767>>2]|0; //@line 949 "sndmix.cpp"
    $769 = ($768|0)==(0|0); //@line 949 "sndmix.cpp"
    $$pre79 = HEAP32[$71>>2]|0; //@line 1047 "sndmix.cpp"
    do {
     if ($769) {
      $950 = $$pre79;
     } else {
      $770 = $$pre79 & 2097152; //@line 953 "sndmix.cpp"
      $771 = ($770|0)==(0); //@line 953 "sndmix.cpp"
      do {
       if ($771) {
        $849 = $$pre79;
       } else {
        $772 = ((($pChn$068)) + 148|0); //@line 956 "sndmix.cpp"
        $773 = HEAP32[$772>>2]|0; //@line 956 "sndmix.cpp"
        $774 = (($773) + 1)|0; //@line 956 "sndmix.cpp"
        HEAP32[$772>>2] = $774; //@line 956 "sndmix.cpp"
        $775 = ((($768)) + 4|0); //@line 958 "sndmix.cpp"
        $776 = HEAP32[$775>>2]|0; //@line 958 "sndmix.cpp"
        $777 = $776 & 4; //@line 958 "sndmix.cpp"
        $778 = ($777|0)==(0); //@line 958 "sndmix.cpp"
        do {
         if ($778) {
          $810 = $$pre79;$820 = $774;
         } else {
          $779 = ((($768)) + 560|0); //@line 960 "sndmix.cpp"
          $780 = HEAP8[$779>>0]|0; //@line 960 "sndmix.cpp"
          $781 = $780&255; //@line 960 "sndmix.cpp"
          $782 = (((($768)) + 12|0) + ($781<<1)|0); //@line 960 "sndmix.cpp"
          $783 = HEAP16[$782>>1]|0; //@line 960 "sndmix.cpp"
          $784 = $783&65535; //@line 960 "sndmix.cpp"
          $785 = HEAP32[$18>>2]|0; //@line 961 "sndmix.cpp"
          $not$40 = ($785|0)!=(4); //@line 961 "sndmix.cpp"
          $786 = $not$40&1; //@line 961 "sndmix.cpp"
          $$32 = (($786) + ($784))|0; //@line 961 "sndmix.cpp"
          $787 = ($774|0)==($$32|0); //@line 962 "sndmix.cpp"
          if ($787) {
           $788 = ((($768)) + 559|0); //@line 964 "sndmix.cpp"
           $789 = HEAP8[$788>>0]|0; //@line 964 "sndmix.cpp"
           $790 = $789&255; //@line 964 "sndmix.cpp"
           $791 = (((($768)) + 12|0) + ($790<<1)|0); //@line 964 "sndmix.cpp"
           $792 = HEAP16[$791>>1]|0; //@line 964 "sndmix.cpp"
           $793 = $792&65535; //@line 964 "sndmix.cpp"
           HEAP32[$772>>2] = $793; //@line 964 "sndmix.cpp"
           $794 = ($780<<24>>24)==($789<<24>>24); //@line 965 "sndmix.cpp"
           if ($794) {
            $795 = (((($768)) + 204|0) + ($781)|0); //@line 965 "sndmix.cpp"
            $796 = HEAP8[$795>>0]|0; //@line 965 "sndmix.cpp"
            $797 = ($796<<24>>24)==(0); //@line 965 "sndmix.cpp"
            if ($797) {
             $798 = $785 & 4; //@line 966 "sndmix.cpp"
             $799 = ($798|0)==(0); //@line 966 "sndmix.cpp"
             if (!($799)) {
              $800 = (($781) + 1)|0; //@line 966 "sndmix.cpp"
              $801 = ((($768)) + 556|0); //@line 966 "sndmix.cpp"
              $802 = HEAP8[$801>>0]|0; //@line 966 "sndmix.cpp"
              $803 = $802&255; //@line 966 "sndmix.cpp"
              $804 = ($800|0)==($803|0); //@line 966 "sndmix.cpp"
              if (!($804)) {
               $810 = $$pre79;$820 = $793;
               break;
              }
             }
             $805 = $$pre79 | 1024; //@line 968 "sndmix.cpp"
             HEAP32[$71>>2] = $805; //@line 968 "sndmix.cpp"
             $806 = ((($pChn$068)) + 124|0); //@line 969 "sndmix.cpp"
             HEAP32[$806>>2] = 0; //@line 969 "sndmix.cpp"
             $810 = $805;$820 = $793;
            } else {
             $810 = $$pre79;$820 = $793;
            }
           } else {
            $810 = $$pre79;$820 = $793;
           }
          } else {
           $810 = $$pre79;$820 = $774;
          }
         }
        } while(0);
        $807 = $776 & 2; //@line 974 "sndmix.cpp"
        $808 = ($807|0)!=(0); //@line 974 "sndmix.cpp"
        $809 = $810 & 512;
        $811 = ($809|0)==(0); //@line 974 "sndmix.cpp"
        $or$cond104 = $808 & $811; //@line 974 "sndmix.cpp"
        if ($or$cond104) {
         $812 = ((($768)) + 562|0); //@line 976 "sndmix.cpp"
         $813 = HEAP8[$812>>0]|0; //@line 976 "sndmix.cpp"
         $814 = $813&255; //@line 976 "sndmix.cpp"
         $815 = (((($768)) + 12|0) + ($814<<1)|0); //@line 976 "sndmix.cpp"
         $816 = HEAP16[$815>>1]|0; //@line 976 "sndmix.cpp"
         $817 = $816&65535; //@line 976 "sndmix.cpp"
         $818 = (($817) + 1)|0; //@line 976 "sndmix.cpp"
         $819 = ($820|0)==($818|0); //@line 976 "sndmix.cpp"
         if (!($819)) {
          $849 = $810;
          break;
         }
         $821 = ((($768)) + 561|0); //@line 977 "sndmix.cpp"
         $822 = HEAP8[$821>>0]|0; //@line 977 "sndmix.cpp"
         $823 = $822&255; //@line 977 "sndmix.cpp"
         $824 = (((($768)) + 12|0) + ($823<<1)|0); //@line 977 "sndmix.cpp"
         $825 = HEAP16[$824>>1]|0; //@line 977 "sndmix.cpp"
         $826 = $825&65535; //@line 977 "sndmix.cpp"
         HEAP32[$772>>2] = $826; //@line 977 "sndmix.cpp"
         $849 = $810;
         break;
        }
        $827 = ((($768)) + 556|0); //@line 980 "sndmix.cpp"
        $828 = HEAP8[$827>>0]|0; //@line 980 "sndmix.cpp"
        $829 = $828&255; //@line 980 "sndmix.cpp"
        $830 = (($829) + -1)|0; //@line 980 "sndmix.cpp"
        $831 = (((($768)) + 12|0) + ($830<<1)|0); //@line 980 "sndmix.cpp"
        $832 = HEAP16[$831>>1]|0; //@line 980 "sndmix.cpp"
        $833 = $832&65535; //@line 980 "sndmix.cpp"
        $834 = ($820>>>0)>($833>>>0); //@line 980 "sndmix.cpp"
        if ($834) {
         $835 = HEAP32[$18>>2]|0; //@line 982 "sndmix.cpp"
         $836 = $835 & 32; //@line 982 "sndmix.cpp"
         $837 = $810 & 512;
         $838 = $836 | $837; //@line 982 "sndmix.cpp"
         $839 = ($838|0)==(0); //@line 982 "sndmix.cpp"
         if ($839) {
          $846 = $810;
         } else {
          $840 = $810 | 1024; //@line 982 "sndmix.cpp"
          HEAP32[$71>>2] = $840; //@line 982 "sndmix.cpp"
          $846 = $840;
         }
         HEAP32[$772>>2] = $833; //@line 983 "sndmix.cpp"
         $841 = (((($768)) + 204|0) + ($830)|0); //@line 984 "sndmix.cpp"
         $842 = HEAP8[$841>>0]|0; //@line 984 "sndmix.cpp"
         $843 = ($842<<24>>24)==(0); //@line 984 "sndmix.cpp"
         if ($843) {
          $844 = HEAP32[$15>>2]|0; //@line 984 "sndmix.cpp"
          $$not = ($nChn$064>>>0)>=($844>>>0); //@line 984 "sndmix.cpp"
          $$not107 = ($836|0)!=(0); //@line 984 "sndmix.cpp"
          $brmerge = $$not | $$not107; //@line 984 "sndmix.cpp"
          if (!($brmerge)) {
           $849 = $846;
           break;
          }
          $845 = $846 | 1024; //@line 986 "sndmix.cpp"
          HEAP32[$71>>2] = $845; //@line 986 "sndmix.cpp"
          $847 = ((($pChn$068)) + 124|0); //@line 987 "sndmix.cpp"
          HEAP32[$847>>2] = 0; //@line 987 "sndmix.cpp"
          HEAP32[$96>>2] = 0; //@line 989 "sndmix.cpp"
          $849 = $845;
         } else {
          $849 = $846;
         }
        } else {
         $849 = $810;
        }
       }
      } while(0);
      $848 = $849 & 4194304; //@line 994 "sndmix.cpp"
      $850 = ($848|0)==(0); //@line 994 "sndmix.cpp"
      do {
       if (!($850)) {
        $851 = ((($pChn$068)) + 152|0); //@line 996 "sndmix.cpp"
        $852 = HEAP32[$851>>2]|0; //@line 996 "sndmix.cpp"
        $853 = (($852) + 1)|0; //@line 996 "sndmix.cpp"
        HEAP32[$851>>2] = $853; //@line 996 "sndmix.cpp"
        $854 = ((($768)) + 4|0); //@line 997 "sndmix.cpp"
        $855 = HEAP32[$854>>2]|0; //@line 997 "sndmix.cpp"
        $856 = $855 & 32; //@line 997 "sndmix.cpp"
        $857 = ($856|0)==(0); //@line 997 "sndmix.cpp"
        if ($857) {
         $883 = $853;
        } else {
         $858 = ((($768)) + 564|0); //@line 999 "sndmix.cpp"
         $859 = HEAP8[$858>>0]|0; //@line 999 "sndmix.cpp"
         $860 = $859&255; //@line 999 "sndmix.cpp"
         $861 = (((($768)) + 76|0) + ($860<<1)|0); //@line 999 "sndmix.cpp"
         $862 = HEAP16[$861>>1]|0; //@line 999 "sndmix.cpp"
         $863 = $862&65535; //@line 999 "sndmix.cpp"
         $864 = HEAP32[$18>>2]|0; //@line 1000 "sndmix.cpp"
         $not$41 = ($864|0)!=(4); //@line 1000 "sndmix.cpp"
         $865 = $not$41&1; //@line 1000 "sndmix.cpp"
         $$33 = (($865) + ($863))|0; //@line 1000 "sndmix.cpp"
         $866 = ($853|0)==($$33|0); //@line 1001 "sndmix.cpp"
         if ($866) {
          $867 = ((($768)) + 563|0); //@line 1002 "sndmix.cpp"
          $868 = HEAP8[$867>>0]|0; //@line 1002 "sndmix.cpp"
          $869 = $868&255; //@line 1002 "sndmix.cpp"
          $870 = (((($768)) + 76|0) + ($869<<1)|0); //@line 1002 "sndmix.cpp"
          $871 = HEAP16[$870>>1]|0; //@line 1002 "sndmix.cpp"
          $872 = $871&65535; //@line 1002 "sndmix.cpp"
          HEAP32[$851>>2] = $872; //@line 1002 "sndmix.cpp"
          $883 = $872;
         } else {
          $883 = $853;
         }
        }
        $873 = $855 & 16; //@line 1005 "sndmix.cpp"
        $874 = ($873|0)==(0); //@line 1005 "sndmix.cpp"
        if (!($874)) {
         $875 = ((($768)) + 566|0); //@line 1005 "sndmix.cpp"
         $876 = HEAP8[$875>>0]|0; //@line 1005 "sndmix.cpp"
         $877 = $876&255; //@line 1005 "sndmix.cpp"
         $878 = (((($768)) + 76|0) + ($877<<1)|0); //@line 1005 "sndmix.cpp"
         $879 = HEAP16[$878>>1]|0; //@line 1005 "sndmix.cpp"
         $880 = $879&65535; //@line 1005 "sndmix.cpp"
         $881 = (($880) + 1)|0; //@line 1005 "sndmix.cpp"
         $882 = ($883|0)==($881|0); //@line 1005 "sndmix.cpp"
         $884 = $849 & 512;
         $885 = ($884|0)==(0); //@line 1006 "sndmix.cpp"
         $or$cond109 = $882 & $885; //@line 1006 "sndmix.cpp"
         if ($or$cond109) {
          $886 = ((($768)) + 565|0); //@line 1009 "sndmix.cpp"
          $887 = HEAP8[$886>>0]|0; //@line 1009 "sndmix.cpp"
          $888 = $887&255; //@line 1009 "sndmix.cpp"
          $889 = (((($768)) + 76|0) + ($888<<1)|0); //@line 1009 "sndmix.cpp"
          $890 = HEAP16[$889>>1]|0; //@line 1009 "sndmix.cpp"
          $891 = $890&65535; //@line 1009 "sndmix.cpp"
          HEAP32[$851>>2] = $891; //@line 1009 "sndmix.cpp"
          break;
         }
        }
        $892 = ((($768)) + 557|0); //@line 1012 "sndmix.cpp"
        $893 = HEAP8[$892>>0]|0; //@line 1012 "sndmix.cpp"
        $894 = $893&255; //@line 1012 "sndmix.cpp"
        $895 = (($894) + -1)|0; //@line 1012 "sndmix.cpp"
        $896 = (((($768)) + 76|0) + ($895<<1)|0); //@line 1012 "sndmix.cpp"
        $897 = HEAP16[$896>>1]|0; //@line 1012 "sndmix.cpp"
        $898 = $897&65535; //@line 1012 "sndmix.cpp"
        $899 = ($883>>>0)>($898>>>0); //@line 1012 "sndmix.cpp"
        if ($899) {
         HEAP32[$851>>2] = $898; //@line 1013 "sndmix.cpp"
        }
       }
      } while(0);
      $900 = $849 & 8388608; //@line 1017 "sndmix.cpp"
      $901 = ($900|0)==(0); //@line 1017 "sndmix.cpp"
      if ($901) {
       $950 = $849;
      } else {
       $902 = ((($pChn$068)) + 156|0); //@line 1020 "sndmix.cpp"
       $903 = HEAP32[$902>>2]|0; //@line 1020 "sndmix.cpp"
       $904 = (($903) + 1)|0; //@line 1020 "sndmix.cpp"
       HEAP32[$902>>2] = $904; //@line 1020 "sndmix.cpp"
       $905 = ((($768)) + 4|0); //@line 1022 "sndmix.cpp"
       $906 = HEAP32[$905>>2]|0; //@line 1022 "sndmix.cpp"
       $907 = $906 & 256; //@line 1022 "sndmix.cpp"
       $908 = ($907|0)==(0); //@line 1022 "sndmix.cpp"
       if ($908) {
        $934 = $904;
       } else {
        $909 = ((($768)) + 568|0); //@line 1024 "sndmix.cpp"
        $910 = HEAP8[$909>>0]|0; //@line 1024 "sndmix.cpp"
        $911 = $910&255; //@line 1024 "sndmix.cpp"
        $912 = (((($768)) + 140|0) + ($911<<1)|0); //@line 1024 "sndmix.cpp"
        $913 = HEAP16[$912>>1]|0; //@line 1024 "sndmix.cpp"
        $914 = $913&65535; //@line 1024 "sndmix.cpp"
        $915 = ($904>>>0)<($914>>>0); //@line 1024 "sndmix.cpp"
        if ($915) {
         $934 = $904;
        } else {
         $916 = ((($768)) + 567|0); //@line 1025 "sndmix.cpp"
         $917 = HEAP8[$916>>0]|0; //@line 1025 "sndmix.cpp"
         $918 = $917&255; //@line 1025 "sndmix.cpp"
         $919 = (((($768)) + 140|0) + ($918<<1)|0); //@line 1025 "sndmix.cpp"
         $920 = HEAP16[$919>>1]|0; //@line 1025 "sndmix.cpp"
         $921 = $920&65535; //@line 1025 "sndmix.cpp"
         HEAP32[$902>>2] = $921; //@line 1025 "sndmix.cpp"
         $934 = $921;
        }
       }
       $922 = $906 & 128; //@line 1028 "sndmix.cpp"
       $923 = ($922|0)!=(0); //@line 1028 "sndmix.cpp"
       $924 = $849 & 512;
       $925 = ($924|0)==(0); //@line 1028 "sndmix.cpp"
       $or$cond111 = $923 & $925; //@line 1028 "sndmix.cpp"
       if ($or$cond111) {
        $926 = ((($768)) + 570|0); //@line 1030 "sndmix.cpp"
        $927 = HEAP8[$926>>0]|0; //@line 1030 "sndmix.cpp"
        $928 = $927&255; //@line 1030 "sndmix.cpp"
        $929 = (((($768)) + 140|0) + ($928<<1)|0); //@line 1030 "sndmix.cpp"
        $930 = HEAP16[$929>>1]|0; //@line 1030 "sndmix.cpp"
        $931 = $930&65535; //@line 1030 "sndmix.cpp"
        $932 = (($931) + 1)|0; //@line 1030 "sndmix.cpp"
        $933 = ($934|0)==($932|0); //@line 1030 "sndmix.cpp"
        if (!($933)) {
         $950 = $849;
         break;
        }
        $935 = ((($768)) + 569|0); //@line 1031 "sndmix.cpp"
        $936 = HEAP8[$935>>0]|0; //@line 1031 "sndmix.cpp"
        $937 = $936&255; //@line 1031 "sndmix.cpp"
        $938 = (((($768)) + 140|0) + ($937<<1)|0); //@line 1031 "sndmix.cpp"
        $939 = HEAP16[$938>>1]|0; //@line 1031 "sndmix.cpp"
        $940 = $939&65535; //@line 1031 "sndmix.cpp"
        HEAP32[$902>>2] = $940; //@line 1031 "sndmix.cpp"
        $950 = $849;
        break;
       } else {
        $941 = ((($768)) + 558|0); //@line 1034 "sndmix.cpp"
        $942 = HEAP8[$941>>0]|0; //@line 1034 "sndmix.cpp"
        $943 = $942&255; //@line 1034 "sndmix.cpp"
        $944 = (($943) + -1)|0; //@line 1034 "sndmix.cpp"
        $945 = (((($768)) + 140|0) + ($944<<1)|0); //@line 1034 "sndmix.cpp"
        $946 = HEAP16[$945>>1]|0; //@line 1034 "sndmix.cpp"
        $947 = $946&65535; //@line 1034 "sndmix.cpp"
        $948 = ($934>>>0)>($947>>>0); //@line 1034 "sndmix.cpp"
        if (!($948)) {
         $950 = $849;
         break;
        }
        HEAP32[$902>>2] = $947; //@line 1035 "sndmix.cpp"
        $950 = $849;
        break;
       }
      }
     }
    } while(0);
    $949 = $950 & -32769; //@line 1047 "sndmix.cpp"
    HEAP32[$71>>2] = $949; //@line 1047 "sndmix.cpp"
    $951 = HEAP32[$96>>2]|0; //@line 1048 "sndmix.cpp"
    $952 = ($951|0)==(0); //@line 1048 "sndmix.cpp"
    if ($952) {
     $953 = ((($pChn$068)) + 20|0); //@line 1048 "sndmix.cpp"
     $954 = HEAP32[$953>>2]|0; //@line 1048 "sndmix.cpp"
     $955 = ($954|0)==(0); //@line 1048 "sndmix.cpp"
     if ($955) {
      $956 = ((($pChn$068)) + 16|0); //@line 1048 "sndmix.cpp"
      $957 = HEAP32[$956>>2]|0; //@line 1048 "sndmix.cpp"
      $958 = ($957|0)==(0); //@line 1048 "sndmix.cpp"
      if ($958) {
       $970 = $949;
      } else {
       label = 193;
      }
     } else {
      label = 193;
     }
    } else {
     label = 193;
    }
    if ((label|0) == 193) {
     label = 0;
     $959 = $950 | 32768; //@line 1049 "sndmix.cpp"
     HEAP32[$71>>2] = $959; //@line 1049 "sndmix.cpp"
     $970 = $959;
    }
    $960 = $961 >> 16; //@line 1059 "sndmix.cpp"
    $962 = (($960) + 1)|0; //@line 1059 "sndmix.cpp"
    $963 = ((($pChn$068)) + 44|0); //@line 1059 "sndmix.cpp"
    $964 = HEAP32[$963>>2]|0; //@line 1059 "sndmix.cpp"
    $965 = ((($pChn$068)) + 40|0); //@line 1059 "sndmix.cpp"
    $966 = HEAP32[$965>>2]|0; //@line 1059 "sndmix.cpp"
    $967 = (($964) - ($966))|0; //@line 1059 "sndmix.cpp"
    $968 = ($962|0)<($967|0); //@line 1059 "sndmix.cpp"
    if ($968) {
     $1021 = $970;
    } else {
     $969 = $970 & -3; //@line 1059 "sndmix.cpp"
     HEAP32[$71>>2] = $969; //@line 1059 "sndmix.cpp"
     $1021 = $969;
    }
    $971 = ((($pChn$068)) + 104|0); //@line 1060 "sndmix.cpp"
    HEAP32[$971>>2] = 0; //@line 1060 "sndmix.cpp"
    $972 = ((($pChn$068)) + 100|0); //@line 1060 "sndmix.cpp"
    HEAP32[$972>>2] = 0; //@line 1060 "sndmix.cpp"
    $973 = ((($pChn$068)) + 96|0); //@line 1061 "sndmix.cpp"
    $974 = HEAP32[$973>>2]|0; //@line 1061 "sndmix.cpp"
    $975 = ($974|0)==(0|0); //@line 1061 "sndmix.cpp"
    if (!($975)) {
     $976 = ((($pChn$068)) + 32|0); //@line 1061 "sndmix.cpp"
     $977 = HEAP32[$976>>2]|0; //@line 1061 "sndmix.cpp"
     $978 = ($977|0)==(0); //@line 1061 "sndmix.cpp"
     $979 = ($961|0)==(0); //@line 1061 "sndmix.cpp"
     $or$cond112 = $978 | $979; //@line 1061 "sndmix.cpp"
     if (!($or$cond112)) {
      HEAP32[$pChn$068>>2] = $974; //@line 1061 "sndmix.cpp"
      $980 = HEAP32[10108]|0; //@line 1090 "sndmix.cpp"
      $981 = ($980>>>0)>(1); //@line 1090 "sndmix.cpp"
      do {
       if ($981) {
        $982 = HEAP32[$103>>2]|0; //@line 1092 "sndmix.cpp"
        $983 = (($982) + -128)|0; //@line 1092 "sndmix.cpp"
        $984 = HEAP32[10106]|0; //@line 1093 "sndmix.cpp"
        $985 = Math_imul($983, $984)|0; //@line 1093 "sndmix.cpp"
        $986 = (($985|0) / 128)&-1; //@line 1094 "sndmix.cpp"
        $987 = (($986) + 128)|0; //@line 1095 "sndmix.cpp"
        $988 = ($987|0)<(0); //@line 1097 "sndmix.cpp"
        $$14 = $988 ? 0 : $987; //@line 1097 "sndmix.cpp"
        $989 = ($$14|0)>(256); //@line 1098 "sndmix.cpp"
        $$$14 = $989 ? 256 : $$14; //@line 1098 "sndmix.cpp"
        $990 = HEAP32[183080]|0; //@line 1100 "sndmix.cpp"
        $991 = $990 & 1; //@line 1100 "sndmix.cpp"
        $992 = ($991|0)==(0); //@line 1100 "sndmix.cpp"
        $993 = (256 - ($$$14))|0; //@line 1100 "sndmix.cpp"
        $$$14$ = $992 ? $$$14 : $993; //@line 1100 "sndmix.cpp"
        $994 = Math_imul($951, $$2)|0; //@line 1102 "sndmix.cpp"
        $995 = $994 >>> 7; //@line 1102 "sndmix.cpp"
        $996 = $990 & 512; //@line 1103 "sndmix.cpp"
        $997 = ($996|0)==(0); //@line 1103 "sndmix.cpp"
        if ($997) {
         $1006 = Math_imul($$$14$, $995)|0; //@line 1116 "sndmix.cpp"
         $1007 = $1006 >> 8; //@line 1116 "sndmix.cpp"
         HEAP32[$971>>2] = $1007; //@line 1116 "sndmix.cpp"
         $1008 = (256 - ($$$14$))|0; //@line 1117 "sndmix.cpp"
         $1009 = Math_imul($1008, $995)|0; //@line 1117 "sndmix.cpp"
         $1010 = $1009 >> 8; //@line 1117 "sndmix.cpp"
         HEAP32[$972>>2] = $1010; //@line 1117 "sndmix.cpp"
         $1014 = $1010;$1016 = $1007;
         break;
        }
        $998 = ($$$14$|0)<(128); //@line 1105 "sndmix.cpp"
        if ($998) {
         $999 = Math_imul($$$14$, $995)|0; //@line 1107 "sndmix.cpp"
         $1000 = $999 >> 8; //@line 1107 "sndmix.cpp"
         HEAP32[$971>>2] = $1000; //@line 1107 "sndmix.cpp"
         $1001 = $994 >> 8; //@line 1108 "sndmix.cpp"
         HEAP32[$972>>2] = $1001; //@line 1108 "sndmix.cpp"
         $1014 = $1001;$1016 = $1000;
         break;
        } else {
         $1002 = $994 >> 8; //@line 1111 "sndmix.cpp"
         HEAP32[$971>>2] = $1002; //@line 1111 "sndmix.cpp"
         $1003 = (256 - ($$$14$))|0; //@line 1112 "sndmix.cpp"
         $1004 = Math_imul($1003, $995)|0; //@line 1112 "sndmix.cpp"
         $1005 = $1004 >> 8; //@line 1112 "sndmix.cpp"
         HEAP32[$972>>2] = $1005; //@line 1112 "sndmix.cpp"
         $1014 = $1005;$1016 = $1002;
         break;
        }
       } else {
        $1011 = Math_imul($951, $$2)|0; //@line 1121 "sndmix.cpp"
        $1012 = $1011 >>> 8; //@line 1121 "sndmix.cpp"
        HEAP32[$972>>2] = $1012; //@line 1121 "sndmix.cpp"
        HEAP32[$971>>2] = $1012; //@line 1122 "sndmix.cpp"
        $1014 = $1012;$1016 = $1012;
       }
      } while(0);
      $1013 = ($1014|0)>(65535); //@line 1125 "sndmix.cpp"
      if ($1013) {
       HEAP32[$972>>2] = 65535; //@line 1125 "sndmix.cpp"
       $1031 = 65535;
      } else {
       $1031 = $1014;
      }
      $1015 = ($1016|0)>(65535); //@line 1126 "sndmix.cpp"
      if ($1015) {
       HEAP32[$971>>2] = 65535; //@line 1126 "sndmix.cpp"
       $1033 = 65535;
      } else {
       $1033 = $1016;
      }
      $1017 = HEAP32[183080]|0; //@line 1128 "sndmix.cpp"
      $1018 = $1017 & 8; //@line 1128 "sndmix.cpp"
      $1019 = ($1018|0)==(0); //@line 1128 "sndmix.cpp"
      do {
       if ($1019) {
        $1022 = $1021 & -12289; //@line 1133 "sndmix.cpp"
        HEAP32[$71>>2] = $1022; //@line 1133 "sndmix.cpp"
        $1023 = ($961|0)==(65536); //@line 1134 "sndmix.cpp"
        if ($1023) {
         $1024 = $1022 | 4096; //@line 1135 "sndmix.cpp"
         HEAP32[$71>>2] = $1024; //@line 1135 "sndmix.cpp"
         $1037 = $1024;
         break;
        }
        $1025 = HEAP32[183080]|0; //@line 1138 "sndmix.cpp"
        $1026 = $1025 & 1040; //@line 1138 "sndmix.cpp"
        $1027 = ($1026|0)==(0); //@line 1138 "sndmix.cpp"
        $1028 = ($961|0)>(65279); //@line 1139 "sndmix.cpp"
        $or$cond37 = $1028 & $1027; //@line 1138 "sndmix.cpp"
        if (!($or$cond37)) {
         $1037 = $1022;
         break;
        }
        $1029 = $1022 | 4096; //@line 1139 "sndmix.cpp"
        HEAP32[$71>>2] = $1029; //@line 1139 "sndmix.cpp"
        $1037 = $1029;
       } else {
        $1020 = $1021 | 4096; //@line 1130 "sndmix.cpp"
        HEAP32[$71>>2] = $1020; //@line 1130 "sndmix.cpp"
        $1037 = $1020;
       }
      } while(0);
      $1030 = $1031 >> 4; //@line 1143 "sndmix.cpp"
      HEAP32[$972>>2] = $1030; //@line 1143 "sndmix.cpp"
      $1032 = $1033 >> 4; //@line 1144 "sndmix.cpp"
      HEAP32[$971>>2] = $1032; //@line 1144 "sndmix.cpp"
      $1034 = ((($pChn$068)) + 28|0); //@line 1145 "sndmix.cpp"
      HEAP32[$1034>>2] = 0; //@line 1145 "sndmix.cpp"
      $1035 = ((($pChn$068)) + 24|0); //@line 1145 "sndmix.cpp"
      HEAP32[$1035>>2] = 0; //@line 1145 "sndmix.cpp"
      $1036 = $1037 & 2048; //@line 1147 "sndmix.cpp"
      $1038 = ($1036|0)!=(0); //@line 1147 "sndmix.cpp"
      $1039 = HEAP32[10108]|0;
      $1040 = ($1039>>>0)<(3); //@line 1147 "sndmix.cpp"
      $or$cond16 = $1038 & $1040; //@line 1147 "sndmix.cpp"
      if ($or$cond16) {
       $1041 = (0 - ($1032))|0; //@line 1147 "sndmix.cpp"
       HEAP32[$971>>2] = $1041; //@line 1147 "sndmix.cpp"
       $1052 = $1041;
      } else {
       $1052 = $1032;
      }
      $1042 = $1037 & 128; //@line 1149 "sndmix.cpp"
      $1043 = ($1042|0)==(0); //@line 1149 "sndmix.cpp"
      if (!($1043)) {
       $1044 = (0 - ($961))|0; //@line 1149 "sndmix.cpp"
       HEAP32[$95>>2] = $1044; //@line 1149 "sndmix.cpp"
      }
      $1045 = $1037 & 32768; //@line 1151 "sndmix.cpp"
      $1046 = ($1045|0)==(0); //@line 1151 "sndmix.cpp"
      $$pre91 = ((($pChn$068)) + 16|0); //@line 1187 "sndmix.cpp"
      do {
       if ($1046) {
        $$pre93 = ((($pChn$068)) + 20|0); //@line 1188 "sndmix.cpp"
        $$pre$phi94Z2D = $$pre93;
        label = 230;
       } else {
        $1047 = HEAP32[$$pre91>>2]|0; //@line 1152 "sndmix.cpp"
        $1048 = ($1030|0)==($1047|0); //@line 1152 "sndmix.cpp"
        $1049 = ((($pChn$068)) + 20|0); //@line 1153 "sndmix.cpp"
        $1050 = HEAP32[$1049>>2]|0; //@line 1153 "sndmix.cpp"
        $1051 = ($1052|0)==($1050|0); //@line 1153 "sndmix.cpp"
        $or$cond113 = $1048 & $1051; //@line 1153 "sndmix.cpp"
        if ($or$cond113) {
         $$pre$phi94Z2D = $1049;
         label = 230;
         break;
        }
        $1053 = HEAP32[10112]|0; //@line 1155 "sndmix.cpp"
        $1054 = (($1030) - ($1047))|0; //@line 1156 "sndmix.cpp"
        $1055 = $1054 << 12; //@line 1156 "sndmix.cpp"
        $1056 = (($1052) - ($1050))|0; //@line 1157 "sndmix.cpp"
        $1057 = $1056 << 12; //@line 1157 "sndmix.cpp"
        $1058 = HEAP32[183080]|0; //@line 1159 "sndmix.cpp"
        $1059 = $1058 & 65536; //@line 1159 "sndmix.cpp"
        $1060 = ($1059|0)==(0); //@line 1159 "sndmix.cpp"
        do {
         if ($1060) {
          $1061 = HEAP32[183079]|0; //@line 1160 "sndmix.cpp"
          $1062 = $1061 & 9; //@line 1160 "sndmix.cpp"
          $1063 = ($1062|0)==(0); //@line 1160 "sndmix.cpp"
          if ($1063) {
           $nRampLength$0 = $1053;
           break;
          }
          $1064 = $1058 & 16; //@line 1161 "sndmix.cpp"
          $1065 = HEAP32[183082]|0;
          $notlhs129 = ($1064|0)==(0); //@line 1161 "sndmix.cpp"
          $notrhs130 = ($1065>>>0)>(20); //@line 1161 "sndmix.cpp"
          $or$cond18$not = $notrhs130 | $notlhs129; //@line 1161 "sndmix.cpp"
          $1066 = $1050 | $1047;
          $1067 = ($1066|0)==(0); //@line 1163 "sndmix.cpp"
          $or$cond115 = $or$cond18$not | $1067; //@line 1161 "sndmix.cpp"
          $1068 = $1052 | $1030;
          $1069 = ($1068|0)==(0); //@line 1163 "sndmix.cpp"
          $or$cond118 = $or$cond115 | $1069; //@line 1161 "sndmix.cpp"
          $or$cond118$not = $or$cond118 ^ 1; //@line 1161 "sndmix.cpp"
          $$old121 = $1037 & 16777216;
          $$old122 = ($$old121|0)==(0); //@line 1163 "sndmix.cpp"
          $or$cond125 = $$old122 & $or$cond118$not; //@line 1161 "sndmix.cpp"
          if ($or$cond125) {
           label = 226;
          } else {
           $nRampLength$0 = $1053;
          }
         } else {
          $$old = $1050 | $1047; //@line 1163 "sndmix.cpp"
          $$old116 = $1052 | $1030;
          $notlhs = ($$old|0)!=(0); //@line 1163 "sndmix.cpp"
          $notrhs = ($$old116|0)!=(0); //@line 1163 "sndmix.cpp"
          $or$cond120$not = $notrhs & $notlhs; //@line 1163 "sndmix.cpp"
          $1070 = $1037 & 16777216;
          $1071 = ($1070|0)==(0); //@line 1163 "sndmix.cpp"
          $or$cond123 = $or$cond120$not & $1071; //@line 1163 "sndmix.cpp"
          if ($or$cond123) {
           label = 226;
          } else {
           $nRampLength$0 = $1053;
          }
         }
        } while(0);
        if ((label|0) == 226) {
         label = 0;
         $1072 = HEAP32[$14>>2]|0; //@line 1165 "sndmix.cpp"
         $1073 = ($1072|0)>(2048); //@line 1166 "sndmix.cpp"
         $$19 = $1073 ? 2048 : $1072; //@line 1166 "sndmix.cpp"
         $1074 = ($$19|0)<($1053|0); //@line 1167 "sndmix.cpp"
         $$$19 = $1074 ? $1053 : $$19; //@line 1167 "sndmix.cpp"
         $nRampLength$0 = $$$19;
        }
        $1075 = (($1055|0) / ($nRampLength$0|0))&-1; //@line 1171 "sndmix.cpp"
        HEAP32[$1035>>2] = $1075; //@line 1171 "sndmix.cpp"
        $1076 = (($1057|0) / ($nRampLength$0|0))&-1; //@line 1172 "sndmix.cpp"
        HEAP32[$1034>>2] = $1076; //@line 1172 "sndmix.cpp"
        $1077 = Math_imul($1075, $nRampLength$0)|0; //@line 1173 "sndmix.cpp"
        $1078 = $1077 >> 12; //@line 1173 "sndmix.cpp"
        $1079 = (($1030) - ($1078))|0; //@line 1173 "sndmix.cpp"
        HEAP32[$$pre91>>2] = $1079; //@line 1173 "sndmix.cpp"
        $1080 = Math_imul($1076, $nRampLength$0)|0; //@line 1174 "sndmix.cpp"
        $1081 = $1080 >> 12; //@line 1174 "sndmix.cpp"
        $1082 = (($1052) - ($1081))|0; //@line 1174 "sndmix.cpp"
        HEAP32[$1049>>2] = $1082; //@line 1174 "sndmix.cpp"
        $1083 = $1076 | $1075; //@line 1175 "sndmix.cpp"
        $1084 = ($1083|0)==(0); //@line 1175 "sndmix.cpp"
        if ($1084) {
         $1085 = $1037 & -32769; //@line 1180 "sndmix.cpp"
         HEAP32[$71>>2] = $1085; //@line 1180 "sndmix.cpp"
         HEAP32[$$pre91>>2] = $1030; //@line 1181 "sndmix.cpp"
         HEAP32[$1049>>2] = $1052; //@line 1182 "sndmix.cpp"
         $1088 = $1030;$1091 = $1052;
         break;
        } else {
         HEAP32[$106>>2] = $nRampLength$0; //@line 1177 "sndmix.cpp"
         $1088 = $1079;$1091 = $1082;
         break;
        }
       }
      } while(0);
      if ((label|0) == 230) {
       label = 0;
       $1086 = $1037 & -32769; //@line 1186 "sndmix.cpp"
       HEAP32[$71>>2] = $1086; //@line 1186 "sndmix.cpp"
       HEAP32[$$pre91>>2] = $1030; //@line 1187 "sndmix.cpp"
       HEAP32[$$pre$phi94Z2D>>2] = $1052; //@line 1188 "sndmix.cpp"
       $1088 = $1030;$1091 = $1052;
      }
      $1087 = $1088 << 12; //@line 1190 "sndmix.cpp"
      $1089 = ((($pChn$068)) + 48|0); //@line 1190 "sndmix.cpp"
      HEAP32[$1089>>2] = $1087; //@line 1190 "sndmix.cpp"
      $1090 = $1091 << 12; //@line 1191 "sndmix.cpp"
      $1092 = ((($pChn$068)) + 52|0); //@line 1191 "sndmix.cpp"
      HEAP32[$1092>>2] = $1090; //@line 1191 "sndmix.cpp"
      $1093 = HEAP32[$65>>2]|0; //@line 1193 "sndmix.cpp"
      $1094 = (($1093) + 1)|0; //@line 1193 "sndmix.cpp"
      HEAP32[$65>>2] = $1094; //@line 1193 "sndmix.cpp"
      $1095 = (((($this)) + 32768|0) + ($1093<<2)|0); //@line 1193 "sndmix.cpp"
      HEAP32[$1095>>2] = $nChn$064; //@line 1193 "sndmix.cpp"
      $1096 = HEAP32[$65>>2]|0; //@line 1194 "sndmix.cpp"
      $1097 = ($1096>>>0)>(127); //@line 1194 "sndmix.cpp"
      if ($1097) {
       $1109 = $1096;
       break L17;
      } else {
       break;
      }
     }
    }
    HEAP32[$pChn$068>>2] = 0; //@line 1061 "sndmix.cpp"
    $1098 = ((($pChn$068)) + 164|0); //@line 1202 "sndmix.cpp"
    $1099 = HEAP32[$1098>>2]|0; //@line 1202 "sndmix.cpp"
    $1100 = ($1099>>>0)>(255); //@line 1202 "sndmix.cpp"
    if ($1100) {
     HEAP32[$1098>>2] = 0; //@line 1202 "sndmix.cpp"
    }
    $1101 = ((($pChn$068)) + 16|0); //@line 1203 "sndmix.cpp"
    HEAP32[$1101>>2] = 0; //@line 1203 "sndmix.cpp"
    $1102 = ((($pChn$068)) + 20|0); //@line 1203 "sndmix.cpp"
    HEAP32[$1102>>2] = 0; //@line 1203 "sndmix.cpp"
    $1103 = ((($pChn$068)) + 32|0); //@line 1204 "sndmix.cpp"
    HEAP32[$1103>>2] = 0; //@line 1204 "sndmix.cpp"
   } else {
    label = 17;
   }
  } while(0);
  if ((label|0) == 17) {
   label = 0;
   $94 = ((($pChn$068)) + 164|0); //@line 504 "sndmix.cpp"
   HEAP32[$94>>2] = 0; //@line 504 "sndmix.cpp"
  }
  $1104 = (($nChn$064) + 1)|0; //@line 494 "sndmix.cpp"
  $1105 = ((($pChn$068)) + 256|0); //@line 494 "sndmix.cpp"
  $1106 = ($1104>>>0)<(128); //@line 494 "sndmix.cpp"
  if ($1106) {
   $nChn$064 = $1104;$pChn$068 = $1105;
  } else {
   label = 236;
   break;
  }
 }
 if ((label|0) == 236) {
  $$pre84 = HEAP32[$65>>2]|0; //@line 1208 "sndmix.cpp"
  $1109 = $$pre84;
 }
 $1107 = HEAP32[10107]|0; //@line 1208 "sndmix.cpp"
 $1108 = ($1109>>>0)<($1107>>>0); //@line 1208 "sndmix.cpp"
 if (!($1108)) {
  $1110 = HEAP32[183080]|0; //@line 1208 "sndmix.cpp"
  $1111 = $1110 & 65536; //@line 1208 "sndmix.cpp"
  $1112 = ($1111|0)!=(0); //@line 1208 "sndmix.cpp"
  $1113 = ($1109|0)==(0); //@line 1210 "sndmix.cpp"
  $or$cond127 = $1112 | $1113; //@line 1208 "sndmix.cpp"
  if (!($or$cond127)) {
   $1116 = $1109;$i17$045 = 0;
   while(1) {
    $1114 = (($i17$045) + 1)|0; //@line 1213 "sndmix.cpp"
    $1115 = ($1114>>>0)<($1116>>>0); //@line 1213 "sndmix.cpp"
    L336: do {
     if ($1115) {
      $$phi$trans$insert85 = (((($this)) + 32768|0) + ($i17$045<<2)|0);
      $$pre86 = HEAP32[$$phi$trans$insert85>>2]|0; //@line 1213 "sndmix.cpp"
      $1117 = (((($this) + ($$pre86<<8)|0)) + 108|0); //@line 1213 "sndmix.cpp"
      $1120 = $1114;$1141 = $1116;$j$043 = $i17$045;
      while(1) {
       $1118 = HEAP32[$1117>>2]|0; //@line 1213 "sndmix.cpp"
       $1119 = (((($this)) + 32768|0) + ($1120<<2)|0); //@line 1213 "sndmix.cpp"
       $1121 = HEAP32[$1119>>2]|0; //@line 1213 "sndmix.cpp"
       $1122 = (((($this) + ($1121<<8)|0)) + 108|0); //@line 1213 "sndmix.cpp"
       $1123 = HEAP32[$1122>>2]|0; //@line 1213 "sndmix.cpp"
       $1124 = ($1118|0)<($1123|0); //@line 1213 "sndmix.cpp"
       if (!($1124)) {
        $1130 = $1141;
        break L336;
       }
       $1125 = (((($this)) + 32768|0) + ($j$043<<2)|0); //@line 1213 "sndmix.cpp"
       HEAP32[$1125>>2] = $1121; //@line 1216 "sndmix.cpp"
       HEAP32[$1119>>2] = $$pre86; //@line 1217 "sndmix.cpp"
       $1126 = (($1120) + 1)|0; //@line 1213 "sndmix.cpp"
       $1127 = HEAP32[$65>>2]|0; //@line 1213 "sndmix.cpp"
       $1128 = ($1126>>>0)<($1127>>>0); //@line 1213 "sndmix.cpp"
       if ($1128) {
        $j$043$phi = $1120;$1120 = $1126;$1141 = $1127;$j$043 = $j$043$phi;
       } else {
        $1130 = $1127;
        break;
       }
      }
     } else {
      $1130 = $1116;
     }
    } while(0);
    $1129 = ($1114>>>0)<($1130>>>0); //@line 1210 "sndmix.cpp"
    if ($1129) {
     $1116 = $1130;$i17$045 = $1114;
    } else {
     break;
    }
   }
  }
 }
 $1131 = HEAP32[$45>>2]|0; //@line 1222 "sndmix.cpp"
 $1132 = $1131 & 1024; //@line 1222 "sndmix.cpp"
 $1133 = ($1132|0)==(0); //@line 1222 "sndmix.cpp"
 if ($1133) {
  $$0 = 1;
  return ($$0|0); //@line 1235 "sndmix.cpp"
 }
 $1134 = ((($this)) + 59780|0); //@line 1224 "sndmix.cpp"
 $1135 = HEAP32[$1134>>2]|0; //@line 1224 "sndmix.cpp"
 $1136 = ($1135|0)==(0); //@line 1224 "sndmix.cpp"
 if ($1136) {
  $1137 = $1131 | 512; //@line 1226 "sndmix.cpp"
  HEAP32[$45>>2] = $1137; //@line 1226 "sndmix.cpp"
  $$0 = 0;
  return ($$0|0); //@line 1235 "sndmix.cpp"
 }
 $1138 = HEAP32[$14>>2]|0; //@line 1229 "sndmix.cpp"
 $1139 = ($1135>>>0)>($1138>>>0); //@line 1229 "sndmix.cpp"
 if ($1139) {
  $1140 = (($1135) - ($1138))|0; //@line 1230 "sndmix.cpp"
  HEAP32[$1134>>2] = $1140; //@line 1230 "sndmix.cpp"
  $$0 = 1;
  return ($$0|0); //@line 1235 "sndmix.cpp"
 } else {
  HEAP32[$1134>>2] = 0; //@line 1232 "sndmix.cpp"
  $$0 = 1;
  return ($$0|0); //@line 1235 "sndmix.cpp"
 }
 return (0)|0;
}
function __ZN10CSoundFile10ProcessRowEv($this) {
 $this = $this|0;
 var $$0 = 0, $$be = 0, $$ph9 = 0, $$pre = 0, $$pre7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $m$04 = 0, $nChn$05 = 0, $pChn$03 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 59692|0); //@line 350 "sndmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 350 "sndmix.cpp"
 $2 = (($1) + 1)|0; //@line 350 "sndmix.cpp"
 HEAP32[$0>>2] = $2; //@line 350 "sndmix.cpp"
 $3 = ((($this)) + 59708|0); //@line 350 "sndmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 350 "sndmix.cpp"
 $5 = ((($this)) + 59700|0); //@line 350 "sndmix.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 350 "sndmix.cpp"
 $7 = (($6) + 1)|0; //@line 350 "sndmix.cpp"
 $8 = Math_imul($7, $4)|0; //@line 350 "sndmix.cpp"
 $9 = ((($this)) + 59704|0); //@line 350 "sndmix.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 350 "sndmix.cpp"
 $11 = (($8) + ($10))|0; //@line 350 "sndmix.cpp"
 $12 = ($2>>>0)<($11>>>0); //@line 350 "sndmix.cpp"
 if ($12) {
  $112 = $4;
 } else {
  HEAP32[$5>>2] = 0; //@line 352 "sndmix.cpp"
  HEAP32[$9>>2] = 0; //@line 353 "sndmix.cpp"
  HEAP32[$0>>2] = 0; //@line 354 "sndmix.cpp"
  $13 = ((($this)) + 59716|0); //@line 355 "sndmix.cpp"
  $14 = HEAP32[$13>>2]|0; //@line 355 "sndmix.cpp"
  $15 = ((($this)) + 59720|0); //@line 355 "sndmix.cpp"
  HEAP32[$15>>2] = $14; //@line 355 "sndmix.cpp"
  $16 = ((($this)) + 59728|0); //@line 357 "sndmix.cpp"
  $17 = HEAP32[$16>>2]|0; //@line 357 "sndmix.cpp"
  $18 = ((($this)) + 59732|0); //@line 357 "sndmix.cpp"
  $19 = HEAP32[$18>>2]|0; //@line 357 "sndmix.cpp"
  $20 = ($17|0)==($19|0); //@line 357 "sndmix.cpp"
  if ($20) {
   $26 = $17;
  } else {
   HEAP32[$16>>2] = $19; //@line 357 "sndmix.cpp"
   $26 = $19;
  }
  $21 = ((($this)) + 59660|0); //@line 359 "sndmix.cpp"
  $22 = HEAP32[$21>>2]|0; //@line 359 "sndmix.cpp"
  $23 = $22 & 32; //@line 359 "sndmix.cpp"
  $24 = ($23|0)==(0); //@line 359 "sndmix.cpp"
  if ($24) {
   $25 = ($26>>>0)<(256); //@line 361 "sndmix.cpp"
   do {
    if ($25) {
     $28 = (((($this)) + 53344|0) + ($26)|0); //@line 361 "sndmix.cpp"
     $29 = HEAP8[$28>>0]|0; //@line 361 "sndmix.cpp"
     $30 = $29&255; //@line 361 "sndmix.cpp"
     $31 = ((($this)) + 59724|0); //@line 361 "sndmix.cpp"
     HEAP32[$31>>2] = $30; //@line 361 "sndmix.cpp"
     $32 = ($29&255)<(240); //@line 362 "sndmix.cpp"
     if ($32) {
      $33 = (((($this)) + 51904|0) + ($30<<2)|0); //@line 362 "sndmix.cpp"
      $34 = HEAP32[$33>>2]|0; //@line 362 "sndmix.cpp"
      $35 = ($34|0)==(0|0); //@line 362 "sndmix.cpp"
      if ($35) {
       HEAP32[$31>>2] = 254; //@line 362 "sndmix.cpp"
       $$ph9 = $31;$38 = 254;
       label = 11;
       break;
      }
     }
     $36 = ($29&255)>(239); //@line 363 "sndmix.cpp"
     if ($36) {
      $$ph9 = $31;$38 = $30;
      label = 11;
     } else {
      $52 = $26;
     }
    } else {
     $27 = ((($this)) + 59724|0); //@line 361 "sndmix.cpp"
     HEAP32[$27>>2] = 255; //@line 361 "sndmix.cpp"
     $$ph9 = $27;$38 = 255;
     label = 11;
    }
   } while(0);
   do {
    if ((label|0) == 11) {
     $37 = ($38|0)==(255); //@line 366 "sndmix.cpp"
     if ($37) {
      $$0 = 0;
      return ($$0|0); //@line 456 "sndmix.cpp"
     }
     L19: while(1) {
      $39 = HEAP32[$16>>2]|0; //@line 366 "sndmix.cpp"
      $40 = ($39>>>0)>(255); //@line 366 "sndmix.cpp"
      if ($40) {
       $$0 = 0;
       label = 42;
       break;
      }
      $41 = (($39) + 1)|0; //@line 407 "sndmix.cpp"
      HEAP32[$16>>2] = $41; //@line 407 "sndmix.cpp"
      $42 = ($41>>>0)<(256); //@line 409 "sndmix.cpp"
      do {
       if ($42) {
        $43 = (((($this)) + 53344|0) + ($41)|0); //@line 409 "sndmix.cpp"
        $44 = HEAP8[$43>>0]|0; //@line 409 "sndmix.cpp"
        $45 = $44&255; //@line 409 "sndmix.cpp"
        HEAP32[$$ph9>>2] = $45; //@line 409 "sndmix.cpp"
        $46 = ($44&255)<(240); //@line 410 "sndmix.cpp"
        if ($46) {
         $47 = (((($this)) + 51904|0) + ($45<<2)|0); //@line 410 "sndmix.cpp"
         $48 = HEAP32[$47>>2]|0; //@line 410 "sndmix.cpp"
         $49 = ($48|0)==(0|0); //@line 410 "sndmix.cpp"
         if ($49) {
          HEAP32[$$ph9>>2] = 254; //@line 410 "sndmix.cpp"
          $$be = 254;
          break;
         }
        }
        $50 = ($44&255)>(239); //@line 363 "sndmix.cpp"
        if ($50) {
         $$be = $45;
        } else {
         label = 20;
         break L19;
        }
       } else {
        HEAP32[$$ph9>>2] = 255; //@line 409 "sndmix.cpp"
        $$be = 255;
       }
      } while(0);
      $51 = ($$be|0)==(255); //@line 366 "sndmix.cpp"
      if ($51) {
       $$0 = 0;
       label = 42;
       break;
      }
     }
     if ((label|0) == 20) {
      $$pre = HEAP32[$16>>2]|0; //@line 412 "sndmix.cpp"
      $52 = $$pre;
      break;
     }
     else if ((label|0) == 42) {
      return ($$0|0); //@line 456 "sndmix.cpp"
     }
    }
   } while(0);
   HEAP32[$18>>2] = $52; //@line 412 "sndmix.cpp"
   $73 = $52;
  } else {
   $73 = $26;
  }
  $53 = ((($this)) + 59724|0); //@line 415 "sndmix.cpp"
  $54 = HEAP32[$53>>2]|0; //@line 415 "sndmix.cpp"
  $55 = ($54>>>0)>(239); //@line 415 "sndmix.cpp"
  if ($55) {
   $$0 = 0;
   return ($$0|0); //@line 456 "sndmix.cpp"
  }
  $56 = (((($this)) + 51904|0) + ($54<<2)|0); //@line 415 "sndmix.cpp"
  $57 = HEAP32[$56>>2]|0; //@line 415 "sndmix.cpp"
  $58 = ($57|0)==(0|0); //@line 415 "sndmix.cpp"
  if ($58) {
   $$0 = 0;
   return ($$0|0); //@line 456 "sndmix.cpp"
  }
  $59 = HEAP32[$15>>2]|0; //@line 417 "sndmix.cpp"
  $60 = (((($this)) + 52864|0) + ($54<<1)|0); //@line 417 "sndmix.cpp"
  $61 = HEAP16[$60>>1]|0; //@line 417 "sndmix.cpp"
  $62 = $61&65535; //@line 417 "sndmix.cpp"
  $63 = ($59>>>0)<($62>>>0); //@line 417 "sndmix.cpp"
  if ($63) {
   $65 = $59;
  } else {
   HEAP32[$15>>2] = 0; //@line 417 "sndmix.cpp"
   $65 = 0;
  }
  $64 = (($65) + 1)|0; //@line 418 "sndmix.cpp"
  HEAP32[$13>>2] = $64; //@line 418 "sndmix.cpp"
  $66 = HEAP16[$60>>1]|0; //@line 419 "sndmix.cpp"
  $67 = $66&65535; //@line 419 "sndmix.cpp"
  $68 = ($64>>>0)<($67>>>0); //@line 419 "sndmix.cpp"
  if (!($68)) {
   $69 = HEAP32[$21>>2]|0; //@line 421 "sndmix.cpp"
   $70 = $69 & 32; //@line 421 "sndmix.cpp"
   $71 = ($70|0)==(0); //@line 421 "sndmix.cpp"
   if ($71) {
    $72 = (($73) + 1)|0; //@line 421 "sndmix.cpp"
    HEAP32[$18>>2] = $72; //@line 421 "sndmix.cpp"
   }
   HEAP32[$13>>2] = 0; //@line 422 "sndmix.cpp"
  }
  $74 = ((($this)) + 59664|0); //@line 426 "sndmix.cpp"
  $75 = HEAP32[$74>>2]|0; //@line 427 "sndmix.cpp"
  $76 = ($75|0)==(0); //@line 427 "sndmix.cpp"
  if (!($76)) {
   $77 = HEAP32[$56>>2]|0; //@line 426 "sndmix.cpp"
   $78 = Math_imul($75, $65)|0; //@line 426 "sndmix.cpp"
   $79 = (($77) + (($78*6)|0)|0); //@line 426 "sndmix.cpp"
   $m$04 = $79;$nChn$05 = 0;$pChn$03 = $this;
   while(1) {
    $80 = HEAP8[$m$04>>0]|0; //@line 429 "sndmix.cpp"
    $81 = ((($pChn$03)) + 246|0); //@line 429 "sndmix.cpp"
    HEAP8[$81>>0] = $80; //@line 429 "sndmix.cpp"
    $82 = ((($m$04)) + 1|0); //@line 430 "sndmix.cpp"
    $83 = HEAP8[$82>>0]|0; //@line 430 "sndmix.cpp"
    $84 = ((($pChn$03)) + 247|0); //@line 430 "sndmix.cpp"
    HEAP8[$84>>0] = $83; //@line 430 "sndmix.cpp"
    $85 = ((($m$04)) + 2|0); //@line 431 "sndmix.cpp"
    $86 = HEAP8[$85>>0]|0; //@line 431 "sndmix.cpp"
    $87 = ((($pChn$03)) + 248|0); //@line 431 "sndmix.cpp"
    HEAP8[$87>>0] = $86; //@line 431 "sndmix.cpp"
    $88 = ((($m$04)) + 4|0); //@line 432 "sndmix.cpp"
    $89 = HEAP8[$88>>0]|0; //@line 432 "sndmix.cpp"
    $90 = ((($pChn$03)) + 249|0); //@line 432 "sndmix.cpp"
    HEAP8[$90>>0] = $89; //@line 432 "sndmix.cpp"
    $91 = ((($m$04)) + 3|0); //@line 433 "sndmix.cpp"
    $92 = HEAP8[$91>>0]|0; //@line 433 "sndmix.cpp"
    $93 = ((($pChn$03)) + 250|0); //@line 433 "sndmix.cpp"
    HEAP8[$93>>0] = $92; //@line 433 "sndmix.cpp"
    $94 = ((($m$04)) + 5|0); //@line 434 "sndmix.cpp"
    $95 = HEAP8[$94>>0]|0; //@line 434 "sndmix.cpp"
    $96 = ((($pChn$03)) + 251|0); //@line 434 "sndmix.cpp"
    HEAP8[$96>>0] = $95; //@line 434 "sndmix.cpp"
    $97 = ((($pChn$03)) + 104|0); //@line 436 "sndmix.cpp"
    $98 = HEAP32[$97>>2]|0; //@line 436 "sndmix.cpp"
    $99 = ((($pChn$03)) + 20|0); //@line 436 "sndmix.cpp"
    HEAP32[$99>>2] = $98; //@line 436 "sndmix.cpp"
    $100 = ((($pChn$03)) + 100|0); //@line 437 "sndmix.cpp"
    $101 = HEAP32[$100>>2]|0; //@line 437 "sndmix.cpp"
    $102 = ((($pChn$03)) + 16|0); //@line 437 "sndmix.cpp"
    HEAP32[$102>>2] = $101; //@line 437 "sndmix.cpp"
    $103 = ((($pChn$03)) + 36|0); //@line 438 "sndmix.cpp"
    $104 = HEAP32[$103>>2]|0; //@line 438 "sndmix.cpp"
    $105 = $104 & -983041; //@line 438 "sndmix.cpp"
    HEAP32[$103>>2] = $105; //@line 438 "sndmix.cpp"
    $106 = ((($pChn$03)) + 216|0); //@line 439 "sndmix.cpp"
    HEAP8[$106>>0] = 0; //@line 439 "sndmix.cpp"
    $107 = ((($pChn$03)) + 256|0); //@line 427 "sndmix.cpp"
    $108 = (($nChn$05) + 1)|0; //@line 427 "sndmix.cpp"
    $109 = ((($m$04)) + 6|0); //@line 427 "sndmix.cpp"
    $110 = ($108>>>0)<($75>>>0); //@line 427 "sndmix.cpp"
    if ($110) {
     $m$04 = $109;$nChn$05 = $108;$pChn$03 = $107;
    } else {
     break;
    }
   }
  }
  $$pre7 = HEAP32[$3>>2]|0; //@line 443 "sndmix.cpp"
  $112 = $$pre7;
 }
 $111 = ($112|0)==(0); //@line 443 "sndmix.cpp"
 if ($111) {
  HEAP32[$3>>2] = 1; //@line 443 "sndmix.cpp"
  $126 = 1;
 } else {
  $126 = $112;
 }
 $113 = ((($this)) + 59660|0); //@line 444 "sndmix.cpp"
 $114 = HEAP32[$113>>2]|0; //@line 444 "sndmix.cpp"
 $115 = $114 | 4096; //@line 444 "sndmix.cpp"
 HEAP32[$113>>2] = $115; //@line 444 "sndmix.cpp"
 $116 = HEAP32[$0>>2]|0; //@line 445 "sndmix.cpp"
 $117 = ($116|0)==(0); //@line 445 "sndmix.cpp"
 if (!($117)) {
  $118 = $114 & -4097; //@line 447 "sndmix.cpp"
  HEAP32[$113>>2] = $118; //@line 447 "sndmix.cpp"
  $119 = ((($this)) + 59680|0); //@line 448 "sndmix.cpp"
  $120 = HEAP32[$119>>2]|0; //@line 448 "sndmix.cpp"
  $121 = $120 & 4; //@line 448 "sndmix.cpp"
  $122 = ($121|0)==(0); //@line 448 "sndmix.cpp"
  if ($122) {
   $123 = HEAP32[$5>>2]|0; //@line 448 "sndmix.cpp"
   $124 = (($123) + 1)|0; //@line 448 "sndmix.cpp"
   $125 = Math_imul($124, $126)|0; //@line 448 "sndmix.cpp"
   $127 = ($116>>>0)<($125>>>0); //@line 448 "sndmix.cpp"
   if ($127) {
    $128 = (($116>>>0) % ($126>>>0))&-1; //@line 450 "sndmix.cpp"
    $129 = ($128|0)==(0); //@line 450 "sndmix.cpp"
    if ($129) {
     HEAP32[$113>>2] = $115; //@line 450 "sndmix.cpp"
    }
   }
  }
 }
 $130 = (__ZN10CSoundFile14ProcessEffectsEv($this)|0); //@line 455 "sndmix.cpp"
 $$0 = $130;
 return ($$0|0); //@line 456 "sndmix.cpp"
}
function __ZN10CSoundFile14ProcessEffectsEv($this) {
 $this = $this|0;
 var $$ = 0, $$$$i = 0, $$$$i52 = 0, $$$$i65 = 0, $$$$i73 = 0, $$$$i80 = 0, $$$i = 0, $$$i42 = 0, $$$i44 = 0, $$$i51 = 0, $$$i64 = 0, $$$i69 = 0, $$$i72 = 0, $$$i79 = 0, $$0$i = 0, $$0$i40 = 0, $$0$i41 = 0, $$0$i43 = 0, $$0$i46 = 0, $$0$i48 = 0;
 var $$0$i53 = 0, $$0$i59 = 0, $$0$i70 = 0, $$0$i74 = 0, $$1$i = 0, $$instr$0 = 0, $$nBreakRow$0 = 0, $$nPatLoopRow$0 = 0, $$not$i = 0, $$not$i55 = 0, $$off = 0, $$op = 0, $$op106 = 0, $$param$i = 0, $$phi$trans$insert127 = 0, $$pre$i = 0, $$pre$phi$i134135Z2D = 0, $$pre$phi$i60137138Z2D = 0, $$pre$phi$i75140141Z2D = 0, $$pre$phi132Z2D = 0;
 var $$pre128 = 0, $$pre131 = 0, $$storemerge$i = 0, $$storemerge$i58 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0;
 var $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0;
 var $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0;
 var $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0;
 var $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0;
 var $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0;
 var $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0;
 var $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0;
 var $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0;
 var $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0;
 var $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0;
 var $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0;
 var $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0;
 var $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0;
 var $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0;
 var $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0;
 var $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0;
 var $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0;
 var $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0;
 var $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0;
 var $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0;
 var $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0;
 var $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0;
 var $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0;
 var $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0;
 var $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0;
 var $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0;
 var $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0;
 var $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0;
 var $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0;
 var $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0;
 var $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bPorta$093 = 0, $bPorta$09597 = 0, $bPorta$1$off0 = 0, $bPorta$1$off0100 = 0, $brmerge$i = 0, $brmerge$i56 = 0, $extract$t = 0, $extract$t31 = 0, $extract$t32 = 0, $i$0107 = 0, $i$02$i = 0, $instr$1 = 0, $instr$3 = 0;
 var $nBreakRow$0$lcssa = 0, $nBreakRow$0115 = 0, $nBreakRow$1 = 0, $nChn$0108 = 0, $nChnSlide$0$i = 0, $nGlbSlide$0$$i = 0, $nGlbSlide$0$i = 0, $nPanSlide$0$i = 0, $nPanSlide$0$i63 = 0, $nPanSlide$0$i78 = 0, $nPatLoopRow$0$lcssa = 0, $nPatLoopRow$0113 = 0, $nPatLoopRow$1 = 0, $nPosJump$0$lcssa = 0, $nPosJump$0114 = 0, $nPosJump$1 = 0, $nPosJump$2 = 0, $nStartTick$0 = 0, $not$ = 0, $not$$i = 0;
 var $not$105 = 0, $not$33 = 0, $note$0$off = 0, $or$cond = 0, $or$cond$i = 0, $or$cond$i47 = 0, $or$cond$i49 = 0, $or$cond$i54 = 0, $or$cond$i61 = 0, $or$cond$i71 = 0, $or$cond$i76 = 0, $or$cond143 = 0, $or$cond19 = 0, $or$cond2$i = 0, $or$cond21 = 0, $or$cond27 = 0, $or$cond29 = 0, $or$cond3$i = 0, $or$cond3$i68 = 0, $or$cond5 = 0;
 var $or$cond5$i = 0, $or$cond7 = 0, $or$cond9 = 0, $p$01$i = 0, $pChn$0116 = 0, $param$0 = 0, $param$1 = 0, $param$1$ = 0, $param$1$$op = 0, $param$2 = 0, $param$3 = 0, $param$4 = 0, $param$5 = 0, $param$6102 = 0, $param$7 = 0, $storemerge = 0, $storemerge$i = 0, $storemerge$i57 = 0, $vol$3 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 59664|0); //@line 734 "snd_fx.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 734 "snd_fx.cpp"
 $2 = ($1|0)==(0); //@line 734 "snd_fx.cpp"
 L1: do {
  if ($2) {
   $$pre131 = ((($this)) + 59692|0); //@line 1194 "snd_fx.cpp"
   $$pre$phi132Z2D = $$pre131;$813 = 0;$nBreakRow$0$lcssa = -1;$nPatLoopRow$0$lcssa = -1;$nPosJump$0$lcssa = -1;
  } else {
   $3 = ((($this)) + 59680|0); //@line 748 "snd_fx.cpp"
   $4 = ((($this)) + 59692|0); //@line 754 "snd_fx.cpp"
   $5 = ((($this)) + 59720|0); //@line 2131 "snd_fx.cpp"
   $6 = ((($this)) + 59700|0); //@line 766 "snd_fx.cpp"
   $7 = ((($this)) + 59688|0); //@line 788 "snd_fx.cpp"
   $8 = ((($this)) + 59728|0); //@line 2075 "snd_fx.cpp"
   $9 = ((($this)) + 59708|0); //@line 2083 "snd_fx.cpp"
   $10 = ((($this)) + 59712|0); //@line 2095 "snd_fx.cpp"
   $11 = ((($this)) + 59660|0); //@line 1015 "snd_fx.cpp"
   $12 = ((($this)) + 59744|0); //@line 1060 "snd_fx.cpp"
   $13 = ((($this)) + 59760|0); //@line 2141 "snd_fx.cpp"
   $62 = $1;$nBreakRow$0115 = -1;$nChn$0108 = 0;$nPatLoopRow$0113 = -1;$nPosJump$0114 = -1;$pChn$0116 = $this;
   while(1) {
    $14 = ((($pChn$0116)) + 247|0); //@line 736 "snd_fx.cpp"
    $15 = HEAP8[$14>>0]|0; //@line 736 "snd_fx.cpp"
    $16 = $15&255; //@line 736 "snd_fx.cpp"
    $17 = ((($pChn$0116)) + 248|0); //@line 737 "snd_fx.cpp"
    $18 = HEAP8[$17>>0]|0; //@line 737 "snd_fx.cpp"
    $19 = $18&255; //@line 737 "snd_fx.cpp"
    $20 = ((($pChn$0116)) + 249|0); //@line 738 "snd_fx.cpp"
    $21 = HEAP8[$20>>0]|0; //@line 738 "snd_fx.cpp"
    $22 = $21&255; //@line 738 "snd_fx.cpp"
    $23 = ((($pChn$0116)) + 250|0); //@line 739 "snd_fx.cpp"
    $24 = HEAP8[$23>>0]|0; //@line 739 "snd_fx.cpp"
    $25 = $24&255; //@line 739 "snd_fx.cpp"
    $26 = ((($pChn$0116)) + 251|0); //@line 740 "snd_fx.cpp"
    $27 = HEAP8[$26>>0]|0; //@line 740 "snd_fx.cpp"
    $28 = $27&255; //@line 740 "snd_fx.cpp"
    switch ($24<<24>>24) {
    case 4: case 6:  {
     $31 = 0;
     break;
    }
    default: {
     $29 = ($18<<24>>24)!=(11); //@line 741 "snd_fx.cpp"
     $31 = $29;
    }
    }
    $30 = $31&1; //@line 741 "snd_fx.cpp"
    $32 = $30 ^ 1; //@line 741 "snd_fx.cpp"
    $33 = ((($pChn$0116)) + 36|0); //@line 744 "snd_fx.cpp"
    $34 = HEAP32[$33>>2]|0; //@line 744 "snd_fx.cpp"
    $35 = $34 & -16777217; //@line 744 "snd_fx.cpp"
    HEAP32[$33>>2] = $35; //@line 744 "snd_fx.cpp"
    $36 = ($24<<24>>24)==(19); //@line 746 "snd_fx.cpp"
    $37 = ($24<<24>>24)==(20); //@line 746 "snd_fx.cpp"
    $$off = (($24) + -19)<<24>>24; //@line 746 "snd_fx.cpp"
    $38 = ($$off&255)<(2); //@line 746 "snd_fx.cpp"
    do {
     if ($38) {
      $39 = ($27<<24>>24)==(0); //@line 748 "snd_fx.cpp"
      if ($39) {
       $40 = HEAP32[$3>>2]|0; //@line 748 "snd_fx.cpp"
       $41 = $40 & 34; //@line 748 "snd_fx.cpp"
       $42 = ($41|0)==(0); //@line 748 "snd_fx.cpp"
       if ($42) {
        label = 10;
       } else {
        $43 = ((($pChn$0116)) + 233|0); //@line 748 "snd_fx.cpp"
        $44 = HEAP8[$43>>0]|0; //@line 748 "snd_fx.cpp"
        $45 = $44&255; //@line 748 "snd_fx.cpp"
        $param$0 = $45;
       }
      } else {
       label = 10;
      }
      if ((label|0) == 10) {
       label = 0;
       $46 = ((($pChn$0116)) + 233|0); //@line 748 "snd_fx.cpp"
       HEAP8[$46>>0] = $27; //@line 748 "snd_fx.cpp"
       $param$0 = $28;
      }
      $47 = $param$0 & 240; //@line 750 "snd_fx.cpp"
      $48 = ($47|0)==(208); //@line 750 "snd_fx.cpp"
      if ($48) {
       $49 = $param$0 & 15; //@line 752 "snd_fx.cpp"
       $nPatLoopRow$1 = $nPatLoopRow$0113;$nStartTick$0 = $49;$param$1 = $param$0;
       break;
      }
      $50 = HEAP32[$4>>2]|0; //@line 754 "snd_fx.cpp"
      $51 = ($50|0)==(0); //@line 754 "snd_fx.cpp"
      if ($51) {
       $52 = ($47|0)==(96); //@line 757 "snd_fx.cpp"
       $or$cond5 = $36 & $52; //@line 757 "snd_fx.cpp"
       $53 = ($47|0)==(176);
       $or$cond7 = $37 & $53; //@line 758 "snd_fx.cpp"
       $or$cond = $or$cond5 | $or$cond7; //@line 757 "snd_fx.cpp"
       if (!($or$cond)) {
        $78 = ($47|0)==(224); //@line 764 "snd_fx.cpp"
        if (!($78)) {
         $nPatLoopRow$1 = $nPatLoopRow$0113;$nStartTick$0 = 0;$param$1 = $param$0;
         break;
        }
        $79 = $param$0 & 15; //@line 766 "snd_fx.cpp"
        HEAP32[$6>>2] = $79; //@line 766 "snd_fx.cpp"
        $nPatLoopRow$1 = $nPatLoopRow$0113;$nStartTick$0 = 0;$param$1 = $param$0;
        break;
       }
       $54 = $param$0 & 15; //@line 760 "snd_fx.cpp"
       $55 = ($54|0)==(0); //@line 2112 "snd_fx.cpp"
       L25: do {
        if ($55) {
         $74 = HEAP32[$5>>2]|0; //@line 2131 "snd_fx.cpp"
         $75 = $74&255; //@line 2131 "snd_fx.cpp"
         $76 = ((($pChn$0116)) + 244|0); //@line 2131 "snd_fx.cpp"
         HEAP8[$76>>0] = $75; //@line 2131 "snd_fx.cpp"
         $$0$i = -1;
        } else {
         $56 = ((($pChn$0116)) + 245|0); //@line 2114 "snd_fx.cpp"
         $57 = HEAP8[$56>>0]|0; //@line 2114 "snd_fx.cpp"
         $58 = ($57<<24>>24)==(0); //@line 2114 "snd_fx.cpp"
         if ($58) {
          $61 = ($62|0)==(0); //@line 2121 "snd_fx.cpp"
          if (!($61)) {
           $i$02$i = 0;$p$01$i = $this;
           while(1) {
            $63 = ($p$01$i|0)==($pChn$0116|0); //@line 2121 "snd_fx.cpp"
            if (!($63)) {
             $64 = ((($p$01$i)) + 245|0); //@line 2124 "snd_fx.cpp"
             $65 = HEAP8[$64>>0]|0; //@line 2124 "snd_fx.cpp"
             $66 = ($65<<24>>24)==(0); //@line 2124 "snd_fx.cpp"
             if (!($66)) {
              $$0$i = -1;
              break L25;
             }
            }
            $67 = (($i$02$i) + 1)|0; //@line 2121 "snd_fx.cpp"
            $68 = ((($p$01$i)) + 256|0); //@line 2121 "snd_fx.cpp"
            $69 = ($67>>>0)<($62>>>0); //@line 2121 "snd_fx.cpp"
            if ($69) {
             $i$02$i = $67;$p$01$i = $68;
            } else {
             break;
            }
           }
          }
          $70 = $54&255; //@line 2126 "snd_fx.cpp"
          HEAP8[$56>>0] = $70; //@line 2126 "snd_fx.cpp"
         } else {
          $59 = (($57) + -1)<<24>>24; //@line 2116 "snd_fx.cpp"
          HEAP8[$56>>0] = $59; //@line 2116 "snd_fx.cpp"
          $60 = ($59<<24>>24)==(0); //@line 2117 "snd_fx.cpp"
          if ($60) {
           $$0$i = -1;
           break;
          }
         }
         $71 = ((($pChn$0116)) + 244|0); //@line 2128 "snd_fx.cpp"
         $72 = HEAP8[$71>>0]|0; //@line 2128 "snd_fx.cpp"
         $73 = $72&255; //@line 2128 "snd_fx.cpp"
         $$0$i = $73;
        }
       } while(0);
       $77 = ($$0$i|0)>(-1); //@line 761 "snd_fx.cpp"
       $$nPatLoopRow$0 = $77 ? $$0$i : $nPatLoopRow$0113; //@line 761 "snd_fx.cpp"
       $nPatLoopRow$1 = $$nPatLoopRow$0;$nStartTick$0 = 0;$param$1 = $param$0;
      } else {
       $nPatLoopRow$1 = $nPatLoopRow$0113;$nStartTick$0 = 0;$param$1 = $param$0;
      }
     } else {
      $nPatLoopRow$1 = $nPatLoopRow$0113;$nStartTick$0 = 0;$param$1 = $28;
     }
    } while(0);
    $80 = HEAP32[$4>>2]|0; //@line 772 "snd_fx.cpp"
    $81 = ($80|0)==($nStartTick$0|0); //@line 772 "snd_fx.cpp"
    $extract$t32 = $31 ^ 1;
    do {
     if ($81) {
      $82 = ((($pChn$0116)) + 246|0); //@line 774 "snd_fx.cpp"
      $83 = HEAP8[$82>>0]|0; //@line 774 "snd_fx.cpp"
      $84 = $83&255; //@line 774 "snd_fx.cpp"
      $85 = ($15<<24>>24)==(0); //@line 775 "snd_fx.cpp"
      if (!($85)) {
       $86 = ((($pChn$0116)) + 215|0); //@line 775 "snd_fx.cpp"
       HEAP8[$86>>0] = $15; //@line 775 "snd_fx.cpp"
      }
      $87 = HEAP32[$3>>2]|0; //@line 777 "snd_fx.cpp"
      $88 = $87 & 1048581; //@line 777 "snd_fx.cpp"
      $89 = ($88|0)!=(0); //@line 777 "snd_fx.cpp"
      $90 = ($83<<24>>24)==(-1); //@line 779 "snd_fx.cpp"
      $or$cond27 = $90 & $89; //@line 777 "snd_fx.cpp"
      do {
       if ($or$cond27) {
        $91 = ((($pChn$0116)) + 140|0); //@line 779 "snd_fx.cpp"
        $92 = HEAP32[$91>>2]|0; //@line 779 "snd_fx.cpp"
        $93 = ($92|0)==(0|0); //@line 779 "snd_fx.cpp"
        if (!($93)) {
         $94 = ((($92)) + 4|0); //@line 779 "snd_fx.cpp"
         $95 = HEAP32[$94>>2]|0; //@line 779 "snd_fx.cpp"
         $96 = $95 & 1; //@line 779 "snd_fx.cpp"
         $97 = ($96|0)==(0); //@line 779 "snd_fx.cpp"
         if (!($97)) {
          label = 35;
          break;
         }
        }
        $98 = $34 | 16777216; //@line 781 "snd_fx.cpp"
        HEAP32[$33>>2] = $98; //@line 781 "snd_fx.cpp"
        $99 = ((($pChn$0116)) + 116|0); //@line 782 "snd_fx.cpp"
        HEAP32[$99>>2] = 0; //@line 782 "snd_fx.cpp"
        $bPorta$093 = $32;
       } else {
        label = 35;
       }
      } while(0);
      do {
       if ((label|0) == 35) {
        label = 0;
        $100 = ($83<<24>>24)==(0); //@line 786 "snd_fx.cpp"
        $101 = ($15<<24>>24)!=(0); //@line 786 "snd_fx.cpp"
        $or$cond9 = $101 & $100; //@line 786 "snd_fx.cpp"
        if ($or$cond9) {
         $102 = HEAP32[$7>>2]|0; //@line 788 "snd_fx.cpp"
         $103 = ($102|0)==(0); //@line 788 "snd_fx.cpp"
         if ($103) {
          $121 = ($15&255)<(240); //@line 804 "snd_fx.cpp"
          if ($121) {
           $122 = (((((($this)) + 33280|0) + ($16<<6)|0)) + 30|0); //@line 804 "snd_fx.cpp"
           $123 = HEAP16[$122>>1]|0; //@line 804 "snd_fx.cpp"
           $124 = $123&65535; //@line 804 "snd_fx.cpp"
           $125 = ((($pChn$0116)) + 116|0); //@line 804 "snd_fx.cpp"
           HEAP32[$125>>2] = $124; //@line 804 "snd_fx.cpp"
          }
         } else {
          $104 = ((($pChn$0116)) + 144|0); //@line 790 "snd_fx.cpp"
          $105 = HEAP32[$104>>2]|0; //@line 790 "snd_fx.cpp"
          $106 = ($105|0)==(0|0); //@line 790 "snd_fx.cpp"
          if (!($106)) {
           $107 = ((($105)) + 30|0); //@line 790 "snd_fx.cpp"
           $108 = HEAP16[$107>>1]|0; //@line 790 "snd_fx.cpp"
           $109 = $108&65535; //@line 790 "snd_fx.cpp"
           $110 = ((($pChn$0116)) + 116|0); //@line 790 "snd_fx.cpp"
           HEAP32[$110>>2] = $109; //@line 790 "snd_fx.cpp"
          }
          $111 = $87 & 1048580; //@line 791 "snd_fx.cpp"
          $112 = ($111|0)==(0); //@line 791 "snd_fx.cpp"
          if (!($112)) {
           $113 = ((($pChn$0116)) + 148|0); //@line 794 "snd_fx.cpp"
           HEAP32[$113>>2] = 0; //@line 794 "snd_fx.cpp"
           $114 = ((($pChn$0116)) + 152|0); //@line 795 "snd_fx.cpp"
           HEAP32[$114>>2] = 0; //@line 795 "snd_fx.cpp"
           $115 = ((($pChn$0116)) + 156|0); //@line 796 "snd_fx.cpp"
           HEAP32[$115>>2] = 0; //@line 796 "snd_fx.cpp"
           $116 = ((($pChn$0116)) + 188|0); //@line 797 "snd_fx.cpp"
           HEAP32[$116>>2] = 0; //@line 797 "snd_fx.cpp"
           $117 = ((($pChn$0116)) + 192|0); //@line 798 "snd_fx.cpp"
           HEAP32[$117>>2] = 0; //@line 798 "snd_fx.cpp"
           $118 = $34 & -16778241; //@line 799 "snd_fx.cpp"
           $119 = $118 | 16777216; //@line 799 "snd_fx.cpp"
           HEAP32[$33>>2] = $119; //@line 799 "snd_fx.cpp"
           $120 = ((($pChn$0116)) + 124|0); //@line 800 "snd_fx.cpp"
           HEAP32[$120>>2] = 65536; //@line 800 "snd_fx.cpp"
          }
         }
         $126 = $87 & 32; //@line 806 "snd_fx.cpp"
         $127 = ($126|0)==(0); //@line 806 "snd_fx.cpp"
         $$instr$0 = $127 ? 0 : $16; //@line 806 "snd_fx.cpp"
         $instr$1 = $$instr$0;
        } else {
         $instr$1 = $16;
        }
        $128 = ($instr$1>>>0)>(239); //@line 809 "snd_fx.cpp"
        $129 = ($83&255)>(253); //@line 811 "snd_fx.cpp"
        $130 = $129 | $128; //@line 811 "snd_fx.cpp"
        $instr$3 = $130 ? 0 : $instr$1; //@line 811 "snd_fx.cpp"
        $note$0$off = (($84) + -1)|0; //@line 812 "snd_fx.cpp"
        $131 = ($note$0$off>>>0)<(128); //@line 812 "snd_fx.cpp"
        if ($131) {
         $132 = ((($pChn$0116)) + 214|0); //@line 812 "snd_fx.cpp"
         HEAP8[$132>>0] = $83; //@line 812 "snd_fx.cpp"
         if ($31) {
          __ZN10CSoundFile8CheckNNAEjjib($this,$nChn$0108,$instr$3,$84,0); //@line 816 "snd_fx.cpp"
         }
        }
        $133 = ($instr$3|0)==(0); //@line 819 "snd_fx.cpp"
        if ($133) {
         $819 = 0;
         label = 50;
        } else {
         $134 = ((($pChn$0116)) + 144|0); //@line 821 "snd_fx.cpp"
         $135 = HEAP32[$134>>2]|0; //@line 821 "snd_fx.cpp"
         __ZN10CSoundFile16InstrumentChangeEP11_MODCHANNELjbbb($this,$pChn$0116,$instr$3,$extract$t32,1,1); //@line 822 "snd_fx.cpp"
         $136 = ((($pChn$0116)) + 215|0); //@line 823 "snd_fx.cpp"
         HEAP8[$136>>0] = 0; //@line 823 "snd_fx.cpp"
         $137 = HEAP32[$3>>2]|0; //@line 825 "snd_fx.cpp"
         $138 = $137 & 34; //@line 825 "snd_fx.cpp"
         $139 = ($138|0)==(0); //@line 825 "snd_fx.cpp"
         if ($139) {
          $819 = 1;
          label = 50;
         } else {
          $140 = HEAP32[$134>>2]|0; //@line 826 "snd_fx.cpp"
          $141 = ($135|0)!=($140|0); //@line 826 "snd_fx.cpp"
          $142 = ($note$0$off>>>0)<(127); //@line 826 "snd_fx.cpp"
          $143 = $142 & $141; //@line 826 "snd_fx.cpp"
          $$ = $143 ? 0 : $32; //@line 826 "snd_fx.cpp"
          if ($100) {
           $bPorta$093 = $$;
           break;
          } else {
           $158 = 1;$bPorta$09597 = $$;
          }
         }
        }
        if ((label|0) == 50) {
         label = 0;
         if ($100) {
          $bPorta$093 = $32;
          break;
         }
         if ($819) {
          $158 = 1;$bPorta$09597 = $32;
         } else {
          $144 = ((($pChn$0116)) + 215|0); //@line 834 "snd_fx.cpp"
          $145 = HEAP8[$144>>0]|0; //@line 834 "snd_fx.cpp"
          $146 = ($145<<24>>24)!=(0); //@line 834 "snd_fx.cpp"
          $147 = ($83<<24>>24)>(-1); //@line 834 "snd_fx.cpp"
          $or$cond19 = $147 & $146; //@line 834 "snd_fx.cpp"
          if ($or$cond19) {
           $148 = $145&255; //@line 836 "snd_fx.cpp"
           $149 = HEAP32[$3>>2]|0; //@line 836 "snd_fx.cpp"
           $150 = $149 & 1048580; //@line 836 "snd_fx.cpp"
           $not$33 = ($150|0)==(0); //@line 836 "snd_fx.cpp"
           __ZN10CSoundFile16InstrumentChangeEP11_MODCHANNELjbbb($this,$pChn$0116,$148,$extract$t32,0,$not$33); //@line 836 "snd_fx.cpp"
           HEAP8[$144>>0] = 0; //@line 837 "snd_fx.cpp"
           $158 = 0;$bPorta$09597 = $32;
          } else {
           $158 = 0;$bPorta$09597 = $32;
          }
         }
        }
        $151 = $bPorta$09597 & 1; //@line 839 "snd_fx.cpp"
        $152 = ($151<<24>>24)!=(0); //@line 839 "snd_fx.cpp"
        $153 = HEAP32[$3>>2]|0; //@line 839 "snd_fx.cpp"
        $154 = $153 & 1048580; //@line 839 "snd_fx.cpp"
        $not$ = ($154|0)==(0); //@line 839 "snd_fx.cpp"
        __ZN10CSoundFile10NoteChangeEjibb($this,$nChn$0108,$84,$152,$not$); //@line 839 "snd_fx.cpp"
        if ($152) {
         $155 = HEAP32[$3>>2]|0; //@line 840 "snd_fx.cpp"
         $156 = $155 & 1048580; //@line 840 "snd_fx.cpp"
         $157 = ($156|0)!=(0); //@line 840 "snd_fx.cpp"
         $or$cond21 = $158 & $157; //@line 840 "snd_fx.cpp"
         if ($or$cond21) {
          $159 = HEAP32[$33>>2]|0; //@line 842 "snd_fx.cpp"
          $160 = $159 | 16777216; //@line 842 "snd_fx.cpp"
          HEAP32[$33>>2] = $160; //@line 842 "snd_fx.cpp"
          $161 = ((($pChn$0116)) + 148|0); //@line 843 "snd_fx.cpp"
          HEAP32[$161>>2] = 0; //@line 843 "snd_fx.cpp"
          $162 = ((($pChn$0116)) + 152|0); //@line 844 "snd_fx.cpp"
          HEAP32[$162>>2] = 0; //@line 844 "snd_fx.cpp"
          $163 = ((($pChn$0116)) + 156|0); //@line 845 "snd_fx.cpp"
          HEAP32[$163>>2] = 0; //@line 845 "snd_fx.cpp"
          $164 = ((($pChn$0116)) + 188|0); //@line 846 "snd_fx.cpp"
          HEAP32[$164>>2] = 0; //@line 846 "snd_fx.cpp"
          $165 = ((($pChn$0116)) + 192|0); //@line 847 "snd_fx.cpp"
          HEAP32[$165>>2] = 0; //@line 847 "snd_fx.cpp"
          $bPorta$093 = $bPorta$09597;
         } else {
          $bPorta$093 = $bPorta$09597;
         }
        } else {
         $bPorta$093 = $bPorta$09597;
        }
       }
      } while(0);
      $166 = ($18<<24>>24)==(1); //@line 851 "snd_fx.cpp"
      if ($166) {
       $167 = ($21&255)>(64); //@line 853 "snd_fx.cpp"
       $$op106 = $22 << 2; //@line 854 "snd_fx.cpp"
       $168 = $167 ? 256 : $$op106; //@line 854 "snd_fx.cpp"
       $169 = ((($pChn$0116)) + 116|0); //@line 854 "snd_fx.cpp"
       HEAP32[$169>>2] = $168; //@line 854 "snd_fx.cpp"
       $170 = HEAP32[$33>>2]|0; //@line 855 "snd_fx.cpp"
       $171 = $170 | 16777216; //@line 855 "snd_fx.cpp"
       HEAP32[$33>>2] = $171; //@line 855 "snd_fx.cpp"
       $172 = $bPorta$093 & 1;
       $extract$t = ($172<<24>>24)!=(0);
       $bPorta$1$off0100 = $extract$t;
       break;
      }
      $173 = ($18<<24>>24)==(2); //@line 857 "snd_fx.cpp"
      $174 = $bPorta$093 & 1;
      $extract$t31 = ($174<<24>>24)!=(0);
      if ($173) {
       $175 = ($21&255)>(64); //@line 859 "snd_fx.cpp"
       $$op = $22 << 2; //@line 860 "snd_fx.cpp"
       $176 = $175 ? 256 : $$op; //@line 860 "snd_fx.cpp"
       $177 = ((($pChn$0116)) + 120|0); //@line 860 "snd_fx.cpp"
       HEAP32[$177>>2] = $176; //@line 860 "snd_fx.cpp"
       $178 = HEAP32[$33>>2]|0; //@line 861 "snd_fx.cpp"
       $179 = $178 | 16777216; //@line 861 "snd_fx.cpp"
       HEAP32[$33>>2] = $179; //@line 861 "snd_fx.cpp"
       $bPorta$1$off0100 = $extract$t31;
      } else {
       $bPorta$1$off0 = $extract$t31;
       label = 61;
      }
     } else {
      $bPorta$1$off0 = $extract$t32;
      label = 61;
     }
    } while(0);
    L85: do {
     if ((label|0) == 61) {
      label = 0;
      $180 = ($18&255)>(2); //@line 866 "snd_fx.cpp"
      if ($180) {
       $181 = HEAP32[$4>>2]|0; //@line 866 "snd_fx.cpp"
       $182 = ($181>>>0)<($nStartTick$0>>>0); //@line 866 "snd_fx.cpp"
       if ($182) {
        $bPorta$1$off0100 = $bPorta$1$off0;
       } else {
        $183 = ($18<<24>>24)==(11); //@line 868 "snd_fx.cpp"
        if ($183) {
         $184 = HEAP32[$3>>2]|0; //@line 870 "snd_fx.cpp"
         $185 = $184 & 32; //@line 870 "snd_fx.cpp"
         $186 = ($185|0)==(0); //@line 870 "snd_fx.cpp"
         if ($186) {
          $191 = $22 << 4; //@line 873 "snd_fx.cpp"
          __ZN10CSoundFile14TonePortamentoEP11_MODCHANNELj($this,$pChn$0116,$191); //@line 873 "snd_fx.cpp"
          $bPorta$1$off0100 = $bPorta$1$off0;
          break;
         } else {
          $187 = $22 & 15; //@line 871 "snd_fx.cpp"
          $188 = (523935 + ($187)|0); //@line 871 "snd_fx.cpp"
          $189 = HEAP8[$188>>0]|0; //@line 871 "snd_fx.cpp"
          $190 = $189&255; //@line 871 "snd_fx.cpp"
          __ZN10CSoundFile14TonePortamentoEP11_MODCHANNELj($this,$pChn$0116,$190); //@line 871 "snd_fx.cpp"
          $bPorta$1$off0100 = $bPorta$1$off0;
          break;
         }
        }
        $192 = ($21<<24>>24)==(0); //@line 876 "snd_fx.cpp"
        $193 = ((($pChn$0116)) + 234|0); //@line 876 "snd_fx.cpp"
        if ($192) {
         $194 = HEAP8[$193>>0]|0; //@line 876 "snd_fx.cpp"
         $195 = $194&255; //@line 876 "snd_fx.cpp"
         $vol$3 = $195;
        } else {
         HEAP8[$193>>0] = $21; //@line 876 "snd_fx.cpp"
         $vol$3 = $22;
        }
        do {
         switch ($19|0) {
         case 3:  {
          $196 = $vol$3 << 4; //@line 880 "snd_fx.cpp"
          __ZN10CSoundFile11VolumeSlideEP11_MODCHANNELj($this,$pChn$0116,$196); //@line 880 "snd_fx.cpp"
          $bPorta$1$off0100 = $bPorta$1$off0;
          break L85;
          break;
         }
         case 4:  {
          __ZN10CSoundFile11VolumeSlideEP11_MODCHANNELj($this,$pChn$0116,$vol$3); //@line 884 "snd_fx.cpp"
          $bPorta$1$off0100 = $bPorta$1$off0;
          break L85;
          break;
         }
         case 5:  {
          $197 = HEAP32[$3>>2]|0; //@line 888 "snd_fx.cpp"
          $198 = $197 & 32; //@line 888 "snd_fx.cpp"
          $199 = ($198|0)==(0); //@line 888 "snd_fx.cpp"
          if (!($199)) {
           $200 = ($181|0)==($nStartTick$0|0); //@line 890 "snd_fx.cpp"
           if (!($200)) {
            $bPorta$1$off0100 = $bPorta$1$off0;
            break L85;
           }
           $201 = $vol$3 << 4; //@line 890 "snd_fx.cpp"
           $202 = $201 | 15; //@line 890 "snd_fx.cpp"
           __ZN10CSoundFile11VolumeSlideEP11_MODCHANNELj($this,$pChn$0116,$202); //@line 890 "snd_fx.cpp"
           $bPorta$1$off0100 = $bPorta$1$off0;
           break L85;
          }
          $203 = ($vol$3|0)==(0); //@line 1563 "snd_fx.cpp"
          if ($203) {
           $206 = ((($pChn$0116)) + 219|0); //@line 1563 "snd_fx.cpp"
           $207 = HEAP8[$206>>0]|0; //@line 1563 "snd_fx.cpp"
           $208 = $207&255; //@line 1563 "snd_fx.cpp"
           $$0$i40 = $208;
          } else {
           $204 = $vol$3&255; //@line 1563 "snd_fx.cpp"
           $205 = ((($pChn$0116)) + 219|0); //@line 1563 "snd_fx.cpp"
           HEAP8[$205>>0] = $204; //@line 1563 "snd_fx.cpp"
           $$0$i40 = $vol$3;
          }
          $209 = HEAP32[$11>>2]|0; //@line 1564 "snd_fx.cpp"
          $210 = $209 & 4096; //@line 1564 "snd_fx.cpp"
          $211 = ($210|0)==(0); //@line 1564 "snd_fx.cpp"
          if ($211) {
           $bPorta$1$off0100 = $bPorta$1$off0;
           break L85;
          }
          $212 = $$0$i40 << 2; //@line 1566 "snd_fx.cpp"
          $213 = ((($pChn$0116)) + 116|0); //@line 1566 "snd_fx.cpp"
          $214 = HEAP32[$213>>2]|0; //@line 1566 "snd_fx.cpp"
          $215 = (($214) + ($212))|0; //@line 1566 "snd_fx.cpp"
          $216 = ($215|0)>(256); //@line 1567 "snd_fx.cpp"
          $$$i = $216 ? 256 : $215; //@line 1567 "snd_fx.cpp"
          HEAP32[$213>>2] = $$$i; //@line 1566 "snd_fx.cpp"
          $217 = $197 & 1; //@line 1568 "snd_fx.cpp"
          $218 = ($217|0)==(0); //@line 1568 "snd_fx.cpp"
          if ($218) {
           $bPorta$1$off0100 = $bPorta$1$off0;
           break L85;
          }
          $219 = HEAP32[$33>>2]|0; //@line 1568 "snd_fx.cpp"
          $220 = $219 | 16777216; //@line 1568 "snd_fx.cpp"
          HEAP32[$33>>2] = $220; //@line 1568 "snd_fx.cpp"
          $bPorta$1$off0100 = $bPorta$1$off0;
          break L85;
          break;
         }
         case 6:  {
          $221 = HEAP32[$3>>2]|0; //@line 896 "snd_fx.cpp"
          $222 = $221 & 32; //@line 896 "snd_fx.cpp"
          $223 = ($222|0)==(0); //@line 896 "snd_fx.cpp"
          if (!($223)) {
           $224 = ($181|0)==($nStartTick$0|0); //@line 898 "snd_fx.cpp"
           if (!($224)) {
            $bPorta$1$off0100 = $bPorta$1$off0;
            break L85;
           }
           $225 = $vol$3 | 240; //@line 898 "snd_fx.cpp"
           __ZN10CSoundFile11VolumeSlideEP11_MODCHANNELj($this,$pChn$0116,$225); //@line 898 "snd_fx.cpp"
           $bPorta$1$off0100 = $bPorta$1$off0;
           break L85;
          }
          $226 = ($vol$3|0)==(0); //@line 1576 "snd_fx.cpp"
          if ($226) {
           $229 = ((($pChn$0116)) + 219|0); //@line 1576 "snd_fx.cpp"
           $230 = HEAP8[$229>>0]|0; //@line 1576 "snd_fx.cpp"
           $231 = $230&255; //@line 1576 "snd_fx.cpp"
           $$0$i41 = $231;
          } else {
           $227 = $vol$3&255; //@line 1576 "snd_fx.cpp"
           $228 = ((($pChn$0116)) + 219|0); //@line 1576 "snd_fx.cpp"
           HEAP8[$228>>0] = $227; //@line 1576 "snd_fx.cpp"
           $$0$i41 = $vol$3;
          }
          $232 = HEAP32[$11>>2]|0; //@line 1577 "snd_fx.cpp"
          $233 = $232 & 4096; //@line 1577 "snd_fx.cpp"
          $234 = ($233|0)==(0); //@line 1577 "snd_fx.cpp"
          if ($234) {
           $bPorta$1$off0100 = $bPorta$1$off0;
           break L85;
          }
          $235 = $$0$i41 << 2; //@line 1579 "snd_fx.cpp"
          $236 = ((($pChn$0116)) + 116|0); //@line 1579 "snd_fx.cpp"
          $237 = HEAP32[$236>>2]|0; //@line 1579 "snd_fx.cpp"
          $238 = (($237) - ($235))|0; //@line 1579 "snd_fx.cpp"
          $239 = ($238|0)<(0); //@line 1580 "snd_fx.cpp"
          $$$i42 = $239 ? 0 : $238; //@line 1580 "snd_fx.cpp"
          HEAP32[$236>>2] = $$$i42; //@line 1579 "snd_fx.cpp"
          $240 = $221 & 1; //@line 1581 "snd_fx.cpp"
          $241 = ($240|0)==(0); //@line 1581 "snd_fx.cpp"
          if ($241) {
           $bPorta$1$off0100 = $bPorta$1$off0;
           break L85;
          }
          $242 = HEAP32[$33>>2]|0; //@line 1581 "snd_fx.cpp"
          $243 = $242 | 16777216; //@line 1581 "snd_fx.cpp"
          HEAP32[$33>>2] = $243; //@line 1581 "snd_fx.cpp"
          $bPorta$1$off0100 = $bPorta$1$off0;
          break L85;
          break;
         }
         case 7:  {
          $244 = $vol$3 & 15; //@line 1442 "snd_fx.cpp"
          $245 = ($244|0)==(0); //@line 1442 "snd_fx.cpp"
          if (!($245)) {
           $246 = $244&255; //@line 1442 "snd_fx.cpp"
           $247 = ((($pChn$0116)) + 225|0); //@line 1442 "snd_fx.cpp"
           HEAP8[$247>>0] = $246; //@line 1442 "snd_fx.cpp"
          }
          $248 = HEAP32[$33>>2]|0; //@line 1443 "snd_fx.cpp"
          $249 = $248 | 65536; //@line 1443 "snd_fx.cpp"
          HEAP32[$33>>2] = $249; //@line 1443 "snd_fx.cpp"
          $bPorta$1$off0100 = $bPorta$1$off0;
          break L85;
          break;
         }
         case 8:  {
          $250 = $vol$3 & 15; //@line 1441 "snd_fx.cpp"
          $251 = ($250|0)==(0); //@line 1441 "snd_fx.cpp"
          if (!($251)) {
           $252 = $250 << 2; //@line 1441 "snd_fx.cpp"
           $253 = $252&255; //@line 1441 "snd_fx.cpp"
           $254 = ((($pChn$0116)) + 226|0); //@line 1441 "snd_fx.cpp"
           HEAP8[$254>>0] = $253; //@line 1441 "snd_fx.cpp"
          }
          $255 = $vol$3 & 240; //@line 1442 "snd_fx.cpp"
          $256 = ($255|0)==(0); //@line 1442 "snd_fx.cpp"
          if (!($256)) {
           $257 = $vol$3 >>> 4; //@line 1442 "snd_fx.cpp"
           $258 = $257&255; //@line 1442 "snd_fx.cpp"
           $259 = ((($pChn$0116)) + 225|0); //@line 1442 "snd_fx.cpp"
           HEAP8[$259>>0] = $258; //@line 1442 "snd_fx.cpp"
          }
          $260 = HEAP32[$33>>2]|0; //@line 1443 "snd_fx.cpp"
          $261 = $260 | 65536; //@line 1443 "snd_fx.cpp"
          HEAP32[$33>>2] = $261; //@line 1443 "snd_fx.cpp"
          $bPorta$1$off0100 = $bPorta$1$off0;
          break L85;
          break;
         }
         case 9:  {
          $262 = ($vol$3|0)==(0); //@line 1517 "snd_fx.cpp"
          if ($262) {
           $265 = ((($pChn$0116)) + 222|0); //@line 1517 "snd_fx.cpp"
           $266 = HEAP8[$265>>0]|0; //@line 1517 "snd_fx.cpp"
           $267 = $266&255; //@line 1517 "snd_fx.cpp"
           $$0$i48 = $267;
          } else {
           $263 = $vol$3&255; //@line 1517 "snd_fx.cpp"
           $264 = ((($pChn$0116)) + 222|0); //@line 1517 "snd_fx.cpp"
           HEAP8[$264>>0] = $263; //@line 1517 "snd_fx.cpp"
           $$0$i48 = $vol$3;
          }
          $268 = HEAP32[$3>>2]|0; //@line 1518 "snd_fx.cpp"
          $269 = $268 & 290; //@line 1518 "snd_fx.cpp"
          $270 = ($269|0)==(0); //@line 1518 "snd_fx.cpp"
          L140: do {
           if ($270) {
            $289 = HEAP32[$11>>2]|0; //@line 1544 "snd_fx.cpp"
            $290 = $289 & 4096; //@line 1544 "snd_fx.cpp"
            $291 = ($290|0)==(0); //@line 1544 "snd_fx.cpp"
            if (!($291)) {
             $bPorta$1$off0100 = $bPorta$1$off0;
             break L85;
            }
            $292 = $$0$i48 & 15; //@line 1546 "snd_fx.cpp"
            $293 = ($292|0)==(0); //@line 1546 "snd_fx.cpp"
            if ($293) {
             $296 = $$0$i48 >>> 2; //@line 1547 "snd_fx.cpp"
             $297 = $296 & 60; //@line 1547 "snd_fx.cpp"
             $nPanSlide$0$i = $297;
             break;
            } else {
             $294 = $292 << 2; //@line 1546 "snd_fx.cpp"
             $295 = (0 - ($294))|0; //@line 1546 "snd_fx.cpp"
             $nPanSlide$0$i = $295;
             break;
            }
           } else {
            $271 = $$0$i48 & 15; //@line 1520 "snd_fx.cpp"
            $272 = ($271|0)==(15); //@line 1520 "snd_fx.cpp"
            $273 = $$0$i48 & 240; //@line 1520 "snd_fx.cpp"
            do {
             if ($272) {
              $274 = ($273|0)==(0); //@line 1520 "snd_fx.cpp"
              $275 = HEAP32[$11>>2]|0; //@line 1536 "snd_fx.cpp"
              $276 = $275 & 4096; //@line 1536 "snd_fx.cpp"
              $277 = ($276|0)==(0); //@line 1536 "snd_fx.cpp"
              if ($274) {
               if ($277) {
                $$pre$phi$i134135Z2D = 0;$820 = 0;
                break;
               } else {
                $bPorta$1$off0100 = $bPorta$1$off0;
                break L85;
               }
              }
              if ($277) {
               $bPorta$1$off0100 = $bPorta$1$off0;
               break L85;
              }
              $278 = $273 >>> 2; //@line 1524 "snd_fx.cpp"
              $279 = (0 - ($278))|0; //@line 1525 "snd_fx.cpp"
              $nPanSlide$0$i = $279;
              break L140;
             } else {
              $280 = ($273|0)!=(240); //@line 1528 "snd_fx.cpp"
              $281 = ($271|0)==(0); //@line 1528 "snd_fx.cpp"
              $or$cond$i49 = $281 | $280; //@line 1528 "snd_fx.cpp"
              $282 = HEAP32[$11>>2]|0; //@line 1536 "snd_fx.cpp"
              $283 = $282 & 4096; //@line 1536 "snd_fx.cpp"
              $284 = ($283|0)==(0); //@line 1536 "snd_fx.cpp"
              if ($or$cond$i49) {
               if ($284) {
                $$pre$phi$i134135Z2D = $273;$820 = $281;
                break;
               } else {
                $bPorta$1$off0100 = $bPorta$1$off0;
                break L85;
               }
              } else {
               $285 = $271 << 2; //@line 1532 "snd_fx.cpp"
               if ($284) {
                $bPorta$1$off0100 = $bPorta$1$off0;
                break L85;
               } else {
                $nPanSlide$0$i = $285;
                break L140;
               }
              }
             }
            } while(0);
            if ($820) {
             $287 = $$pre$phi$i134135Z2D >>> 2; //@line 1539 "snd_fx.cpp"
             $288 = (0 - ($287))|0; //@line 1539 "snd_fx.cpp"
             $nPanSlide$0$i = $288;
             break;
            } else {
             $286 = $271 << 2; //@line 1538 "snd_fx.cpp"
             $nPanSlide$0$i = $286;
             break;
            }
           }
          } while(0);
          $298 = ($nPanSlide$0$i|0)==(0); //@line 1550 "snd_fx.cpp"
          if ($298) {
           $bPorta$1$off0100 = $bPorta$1$off0;
           break L85;
          }
          $299 = ((($pChn$0116)) + 120|0); //@line 1552 "snd_fx.cpp"
          $300 = HEAP32[$299>>2]|0; //@line 1552 "snd_fx.cpp"
          $301 = (($300) + ($nPanSlide$0$i))|0; //@line 1552 "snd_fx.cpp"
          $302 = ($301|0)<(0); //@line 1553 "snd_fx.cpp"
          $$$i51 = $302 ? 0 : $301; //@line 1553 "snd_fx.cpp"
          $303 = ($$$i51|0)>(256); //@line 1554 "snd_fx.cpp"
          $$$$i52 = $303 ? 256 : $$$i51; //@line 1554 "snd_fx.cpp"
          HEAP32[$299>>2] = $$$$i52; //@line 1555 "snd_fx.cpp"
          $bPorta$1$off0100 = $bPorta$1$off0;
          break L85;
          break;
         }
         case 10:  {
          $304 = $vol$3 << 4; //@line 916 "snd_fx.cpp"
          $305 = ($vol$3|0)==(0); //@line 1517 "snd_fx.cpp"
          if ($305) {
           $308 = ((($pChn$0116)) + 222|0); //@line 1517 "snd_fx.cpp"
           $309 = HEAP8[$308>>0]|0; //@line 1517 "snd_fx.cpp"
           $310 = $309&255; //@line 1517 "snd_fx.cpp"
           $$0$i59 = $310;
          } else {
           $306 = $304&255; //@line 1517 "snd_fx.cpp"
           $307 = ((($pChn$0116)) + 222|0); //@line 1517 "snd_fx.cpp"
           HEAP8[$307>>0] = $306; //@line 1517 "snd_fx.cpp"
           $$0$i59 = $304;
          }
          $311 = HEAP32[$3>>2]|0; //@line 1518 "snd_fx.cpp"
          $312 = $311 & 290; //@line 1518 "snd_fx.cpp"
          $313 = ($312|0)==(0); //@line 1518 "snd_fx.cpp"
          L168: do {
           if ($313) {
            $332 = HEAP32[$11>>2]|0; //@line 1544 "snd_fx.cpp"
            $333 = $332 & 4096; //@line 1544 "snd_fx.cpp"
            $334 = ($333|0)==(0); //@line 1544 "snd_fx.cpp"
            if (!($334)) {
             $bPorta$1$off0100 = $bPorta$1$off0;
             break L85;
            }
            $335 = $$0$i59 & 15; //@line 1546 "snd_fx.cpp"
            $336 = ($335|0)==(0); //@line 1546 "snd_fx.cpp"
            if ($336) {
             $339 = $$0$i59 >>> 2; //@line 1547 "snd_fx.cpp"
             $340 = $339 & 60; //@line 1547 "snd_fx.cpp"
             $nPanSlide$0$i63 = $340;
             break;
            } else {
             $337 = $335 << 2; //@line 1546 "snd_fx.cpp"
             $338 = (0 - ($337))|0; //@line 1546 "snd_fx.cpp"
             $nPanSlide$0$i63 = $338;
             break;
            }
           } else {
            $314 = $$0$i59 & 15; //@line 1520 "snd_fx.cpp"
            $315 = ($314|0)==(15); //@line 1520 "snd_fx.cpp"
            $316 = $$0$i59 & 240; //@line 1520 "snd_fx.cpp"
            do {
             if ($315) {
              $317 = ($316|0)==(0); //@line 1520 "snd_fx.cpp"
              $318 = HEAP32[$11>>2]|0; //@line 1536 "snd_fx.cpp"
              $319 = $318 & 4096; //@line 1536 "snd_fx.cpp"
              $320 = ($319|0)==(0); //@line 1536 "snd_fx.cpp"
              if ($317) {
               if ($320) {
                $$pre$phi$i60137138Z2D = 0;$821 = 0;
                break;
               } else {
                $bPorta$1$off0100 = $bPorta$1$off0;
                break L85;
               }
              }
              if ($320) {
               $bPorta$1$off0100 = $bPorta$1$off0;
               break L85;
              }
              $321 = $316 >>> 2; //@line 1524 "snd_fx.cpp"
              $322 = (0 - ($321))|0; //@line 1525 "snd_fx.cpp"
              $nPanSlide$0$i63 = $322;
              break L168;
             } else {
              $323 = ($316|0)!=(240); //@line 1528 "snd_fx.cpp"
              $324 = ($314|0)==(0); //@line 1528 "snd_fx.cpp"
              $or$cond$i61 = $324 | $323; //@line 1528 "snd_fx.cpp"
              $325 = HEAP32[$11>>2]|0; //@line 1536 "snd_fx.cpp"
              $326 = $325 & 4096; //@line 1536 "snd_fx.cpp"
              $327 = ($326|0)==(0); //@line 1536 "snd_fx.cpp"
              if ($or$cond$i61) {
               if ($327) {
                $$pre$phi$i60137138Z2D = $316;$821 = $324;
                break;
               } else {
                $bPorta$1$off0100 = $bPorta$1$off0;
                break L85;
               }
              } else {
               $328 = $314 << 2; //@line 1532 "snd_fx.cpp"
               if ($327) {
                $bPorta$1$off0100 = $bPorta$1$off0;
                break L85;
               } else {
                $nPanSlide$0$i63 = $328;
                break L168;
               }
              }
             }
            } while(0);
            if ($821) {
             $330 = $$pre$phi$i60137138Z2D >>> 2; //@line 1539 "snd_fx.cpp"
             $331 = (0 - ($330))|0; //@line 1539 "snd_fx.cpp"
             $nPanSlide$0$i63 = $331;
             break;
            } else {
             $329 = $314 << 2; //@line 1538 "snd_fx.cpp"
             $nPanSlide$0$i63 = $329;
             break;
            }
           }
          } while(0);
          $341 = ($nPanSlide$0$i63|0)==(0); //@line 1550 "snd_fx.cpp"
          if ($341) {
           $bPorta$1$off0100 = $bPorta$1$off0;
           break L85;
          }
          $342 = ((($pChn$0116)) + 120|0); //@line 1552 "snd_fx.cpp"
          $343 = HEAP32[$342>>2]|0; //@line 1552 "snd_fx.cpp"
          $344 = (($343) + ($nPanSlide$0$i63))|0; //@line 1552 "snd_fx.cpp"
          $345 = ($344|0)<(0); //@line 1553 "snd_fx.cpp"
          $$$i64 = $345 ? 0 : $344; //@line 1553 "snd_fx.cpp"
          $346 = ($$$i64|0)>(256); //@line 1554 "snd_fx.cpp"
          $$$$i65 = $346 ? 256 : $$$i64; //@line 1554 "snd_fx.cpp"
          HEAP32[$342>>2] = $$$$i65; //@line 1555 "snd_fx.cpp"
          $bPorta$1$off0100 = $bPorta$1$off0;
          break L85;
          break;
         }
         case 12:  {
          $347 = $vol$3 << 2; //@line 920 "snd_fx.cpp"
          __ZN10CSoundFile12PortamentoUpEP11_MODCHANNELj($this,$pChn$0116,$347); //@line 920 "snd_fx.cpp"
          $bPorta$1$off0100 = $bPorta$1$off0;
          break L85;
          break;
         }
         case 13:  {
          $348 = $vol$3 << 2; //@line 924 "snd_fx.cpp"
          __ZN10CSoundFile14PortamentoDownEP11_MODCHANNELj($this,$pChn$0116,$348); //@line 924 "snd_fx.cpp"
          $bPorta$1$off0100 = $bPorta$1$off0;
          break L85;
          break;
         }
         default: {
          $bPorta$1$off0100 = $bPorta$1$off0;
          break L85;
         }
         }
        } while(0);
       }
      } else {
       $bPorta$1$off0100 = $bPorta$1$off0;
      }
     }
    } while(0);
    $349 = ($24<<24>>24)==(0); //@line 931 "snd_fx.cpp"
    L194: do {
     if ($349) {
      $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
     } else {
      do {
       switch ($25|0) {
       case 13:  {
        $350 = HEAP32[$4>>2]|0; //@line 935 "snd_fx.cpp"
        $351 = ($350|0)==(0); //@line 935 "snd_fx.cpp"
        if (!($351)) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $352 = ($param$1>>>0)<(64); //@line 937 "snd_fx.cpp"
        $353 = $param$1 << 2; //@line 937 "snd_fx.cpp"
        $354 = $352 ? $353 : 256; //@line 937 "snd_fx.cpp"
        $355 = ((($pChn$0116)) + 116|0); //@line 937 "snd_fx.cpp"
        HEAP32[$355>>2] = $354; //@line 937 "snd_fx.cpp"
        $356 = HEAP32[$33>>2]|0; //@line 938 "snd_fx.cpp"
        $357 = $356 | 16777216; //@line 938 "snd_fx.cpp"
        HEAP32[$33>>2] = $357; //@line 938 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 2:  {
        $358 = ($param$1|0)==(0); //@line 944 "snd_fx.cpp"
        if ($358) {
         $359 = HEAP32[$3>>2]|0; //@line 944 "snd_fx.cpp"
         $360 = $359 & 1; //@line 944 "snd_fx.cpp"
         $361 = ($360|0)==(0); //@line 944 "snd_fx.cpp"
         if (!($361)) {
          $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
          break L194;
         }
        }
        __ZN10CSoundFile12PortamentoUpEP11_MODCHANNELj($this,$pChn$0116,$param$1); //@line 945 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 3:  {
        $362 = ($param$1|0)==(0); //@line 950 "snd_fx.cpp"
        if ($362) {
         $363 = HEAP32[$3>>2]|0; //@line 950 "snd_fx.cpp"
         $364 = $363 & 1; //@line 950 "snd_fx.cpp"
         $365 = ($364|0)==(0); //@line 950 "snd_fx.cpp"
         if (!($365)) {
          $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
          break L194;
         }
        }
        __ZN10CSoundFile14PortamentoDownEP11_MODCHANNELj($this,$pChn$0116,$param$1); //@line 951 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 11:  {
        $366 = ($param$1|0)==(0); //@line 956 "snd_fx.cpp"
        if ($366) {
         $367 = HEAP32[$3>>2]|0; //@line 956 "snd_fx.cpp"
         $368 = ($367|0)==(1); //@line 956 "snd_fx.cpp"
         if ($368) {
          $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
          break L194;
         }
        }
        __ZN10CSoundFile11VolumeSlideEP11_MODCHANNELj($this,$pChn$0116,$param$1); //@line 956 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 4:  {
        __ZN10CSoundFile14TonePortamentoEP11_MODCHANNELj($this,$pChn$0116,$param$1); //@line 961 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 6:  {
        $369 = ($param$1|0)==(0); //@line 966 "snd_fx.cpp"
        if ($369) {
         $370 = HEAP32[$3>>2]|0; //@line 966 "snd_fx.cpp"
         $371 = ($370|0)==(1); //@line 966 "snd_fx.cpp"
         if (!($371)) {
          label = 159;
         }
        } else {
         label = 159;
        }
        if ((label|0) == 159) {
         label = 0;
         __ZN10CSoundFile11VolumeSlideEP11_MODCHANNELj($this,$pChn$0116,$param$1); //@line 966 "snd_fx.cpp"
        }
        __ZN10CSoundFile14TonePortamentoEP11_MODCHANNELj($this,$pChn$0116,0); //@line 967 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 5:  {
        $372 = $param$1 & 15; //@line 1441 "snd_fx.cpp"
        $373 = ($372|0)==(0); //@line 1441 "snd_fx.cpp"
        if (!($373)) {
         $374 = $372 << 2; //@line 1441 "snd_fx.cpp"
         $375 = $374&255; //@line 1441 "snd_fx.cpp"
         $376 = ((($pChn$0116)) + 226|0); //@line 1441 "snd_fx.cpp"
         HEAP8[$376>>0] = $375; //@line 1441 "snd_fx.cpp"
        }
        $377 = $param$1 & 240; //@line 1442 "snd_fx.cpp"
        $378 = ($377|0)==(0); //@line 1442 "snd_fx.cpp"
        if (!($378)) {
         $379 = $param$1 >>> 4; //@line 1442 "snd_fx.cpp"
         $380 = $379 & 15; //@line 1442 "snd_fx.cpp"
         $381 = $380&255; //@line 1442 "snd_fx.cpp"
         $382 = ((($pChn$0116)) + 225|0); //@line 1442 "snd_fx.cpp"
         HEAP8[$382>>0] = $381; //@line 1442 "snd_fx.cpp"
        }
        $383 = HEAP32[$33>>2]|0; //@line 1443 "snd_fx.cpp"
        $384 = $383 | 65536; //@line 1443 "snd_fx.cpp"
        HEAP32[$33>>2] = $384; //@line 1443 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 7:  {
        $385 = ($param$1|0)==(0); //@line 977 "snd_fx.cpp"
        if ($385) {
         $386 = HEAP32[$3>>2]|0; //@line 977 "snd_fx.cpp"
         $387 = ($386|0)==(1); //@line 977 "snd_fx.cpp"
         if (!($387)) {
          label = 168;
         }
        } else {
         label = 168;
        }
        if ((label|0) == 168) {
         label = 0;
         __ZN10CSoundFile11VolumeSlideEP11_MODCHANNELj($this,$pChn$0116,$param$1); //@line 977 "snd_fx.cpp"
        }
        $388 = HEAP32[$33>>2]|0; //@line 1443 "snd_fx.cpp"
        $389 = $388 | 65536; //@line 1443 "snd_fx.cpp"
        HEAP32[$33>>2] = $389; //@line 1443 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 16:  {
        $390 = HEAP32[$4>>2]|0; //@line 983 "snd_fx.cpp"
        $391 = ($390|0)==(0); //@line 983 "snd_fx.cpp"
        if (!($391)) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $392 = HEAP32[$3>>2]|0; //@line 2068 "snd_fx.cpp"
        $393 = ($392|0)==(32); //@line 2068 "snd_fx.cpp"
        $394 = $393 ? 256 : 128; //@line 2068 "snd_fx.cpp"
        $395 = (($param$1) + -1)|0; //@line 2073 "snd_fx.cpp"
        $396 = ($395>>>0)>(126); //@line 2073 "snd_fx.cpp"
        if ($396) {
         label = 173;
        } else {
         $397 = $392 & 1048581; //@line 2073 "snd_fx.cpp"
         $398 = ($397|0)!=(0); //@line 2073 "snd_fx.cpp"
         $399 = ($param$1>>>0)>(29); //@line 2073 "snd_fx.cpp"
         $or$cond3$i68 = $399 & $398; //@line 2073 "snd_fx.cpp"
         if ($or$cond3$i68) {
          label = 173;
         } else {
          $406 = $392;
         }
        }
        if ((label|0) == 173) {
         label = 0;
         $400 = HEAP32[$8>>2]|0; //@line 2075 "snd_fx.cpp"
         $401 = HEAP32[$5>>2]|0; //@line 2075 "snd_fx.cpp"
         $402 = (($401) + 1)|0; //@line 2075 "snd_fx.cpp"
         $403 = (__ZNK10CSoundFile14IsSongFinishedEjj($this,$400,$402)|0); //@line 2075 "snd_fx.cpp"
         $404 = ($403|0)==(0); //@line 2075 "snd_fx.cpp"
         if ($404) {
          $406 = $392;
         } else {
          (__ZN10CSoundFile14GlobalFadeSongEj($this,1000)|0); //@line 2077 "snd_fx.cpp"
          $$pre$i = HEAP32[$3>>2]|0; //@line 2082 "snd_fx.cpp"
          $406 = $$pre$i;
         }
        }
        $405 = $406 & 2; //@line 2082 "snd_fx.cpp"
        $407 = ($405|0)!=(0); //@line 2082 "snd_fx.cpp"
        $408 = ($param$1>>>0)>(128); //@line 2082 "snd_fx.cpp"
        $or$cond5$i = $408 & $407; //@line 2082 "snd_fx.cpp"
        $409 = (($param$1) + -128)|0; //@line 2082 "snd_fx.cpp"
        $$param$i = $or$cond5$i ? $409 : $param$1; //@line 2082 "snd_fx.cpp"
        $410 = (($$param$i) + -1)|0; //@line 2083 "snd_fx.cpp"
        $411 = ($410>>>0)<($394>>>0); //@line 2083 "snd_fx.cpp"
        if (!($411)) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        HEAP32[$9>>2] = $$param$i; //@line 2083 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 17:  {
        $412 = HEAP32[$4>>2]|0; //@line 988 "snd_fx.cpp"
        $413 = ($412|0)==(0); //@line 988 "snd_fx.cpp"
        if (!($413)) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $414 = HEAP32[$3>>2]|0; //@line 990 "snd_fx.cpp"
        $415 = $414 & 34; //@line 990 "snd_fx.cpp"
        $416 = ($415|0)==(0); //@line 990 "snd_fx.cpp"
        do {
         if ($416) {
          $param$2 = $param$1;
         } else {
          $417 = ($param$1|0)==(0); //@line 992 "snd_fx.cpp"
          if ($417) {
           $420 = ((($pChn$0116)) + 235|0); //@line 992 "snd_fx.cpp"
           $421 = HEAP8[$420>>0]|0; //@line 992 "snd_fx.cpp"
           $422 = $421&255; //@line 992 "snd_fx.cpp"
           $param$2 = $422;
           break;
          } else {
           $418 = $param$1&255; //@line 992 "snd_fx.cpp"
           $419 = ((($pChn$0116)) + 235|0); //@line 992 "snd_fx.cpp"
           HEAP8[$419>>0] = $418; //@line 992 "snd_fx.cpp"
           $param$2 = $param$1;
           break;
          }
         }
        } while(0);
        $423 = ($param$2>>>0)<(32); //@line 2090 "snd_fx.cpp"
        if (!($423)) {
         HEAP32[$10>>2] = $param$2; //@line 2104 "snd_fx.cpp"
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $424 = $param$2 & 240; //@line 2093 "snd_fx.cpp"
        $425 = ($424|0)==(16); //@line 2093 "snd_fx.cpp"
        $426 = $param$2 << 1; //@line 2095 "snd_fx.cpp"
        $427 = $426 & 30; //@line 2095 "snd_fx.cpp"
        $428 = HEAP32[$10>>2]|0; //@line 2095 "snd_fx.cpp"
        if ($425) {
         $429 = (($428) + ($427))|0; //@line 2095 "snd_fx.cpp"
         $430 = ($429>>>0)>(255); //@line 2096 "snd_fx.cpp"
         $$$i69 = $430 ? 255 : $429; //@line 2096 "snd_fx.cpp"
         HEAP32[$10>>2] = $$$i69; //@line 2096 "snd_fx.cpp"
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        } else {
         $431 = (($428) - ($427))|0; //@line 2099 "snd_fx.cpp"
         $432 = ($431|0)<(32); //@line 2100 "snd_fx.cpp"
         $$1$i = $432 ? 32 : $431; //@line 2100 "snd_fx.cpp"
         HEAP32[$10>>2] = $$1$i; //@line 2100 "snd_fx.cpp"
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        break;
       }
       case 10:  {
        $433 = HEAP32[$4>>2]|0; //@line 1000 "snd_fx.cpp"
        $434 = ($433|0)==(0); //@line 1000 "snd_fx.cpp"
        if (!($434)) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $435 = ($param$1|0)==(0); //@line 1001 "snd_fx.cpp"
        if ($435) {
         $438 = ((($pChn$0116)) + 236|0); //@line 1001 "snd_fx.cpp"
         $439 = HEAP8[$438>>0]|0; //@line 1001 "snd_fx.cpp"
         $440 = $439&255; //@line 1001 "snd_fx.cpp"
         $param$3 = $440;
        } else {
         $436 = $param$1&255; //@line 1001 "snd_fx.cpp"
         $437 = ((($pChn$0116)) + 236|0); //@line 1001 "snd_fx.cpp"
         HEAP8[$437>>0] = $436; //@line 1001 "snd_fx.cpp"
         $param$3 = $param$1;
        }
        $441 = $param$3 << 8; //@line 1002 "snd_fx.cpp"
        $442 = ((($pChn$0116)) + 237|0); //@line 1003 "snd_fx.cpp"
        $443 = HEAP8[$442>>0]|0; //@line 1003 "snd_fx.cpp"
        $444 = $443&255; //@line 1003 "snd_fx.cpp"
        $445 = $444 << 16; //@line 1003 "snd_fx.cpp"
        $446 = $445 | $441; //@line 1003 "snd_fx.cpp"
        $447 = ((($pChn$0116)) + 246|0); //@line 1004 "snd_fx.cpp"
        $448 = HEAP8[$447>>0]|0; //@line 1004 "snd_fx.cpp"
        $449 = ($448<<24>>24)>(0); //@line 1004 "snd_fx.cpp"
        if (!($449)) {
         $466 = ((($pChn$0116)) + 32|0); //@line 1022 "snd_fx.cpp"
         $467 = HEAP32[$466>>2]|0; //@line 1022 "snd_fx.cpp"
         $468 = ($446>>>0)<($467>>>0); //@line 1022 "snd_fx.cpp"
         if (!($468)) {
          $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
          break L194;
         }
         $469 = HEAP32[$3>>2]|0; //@line 1022 "snd_fx.cpp"
         $470 = $469 & 131088; //@line 1022 "snd_fx.cpp"
         $471 = ($470|0)==(0); //@line 1022 "snd_fx.cpp"
         if ($471) {
          $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
          break L194;
         }
         $472 = ((($pChn$0116)) + 4|0); //@line 1024 "snd_fx.cpp"
         HEAP32[$472>>2] = $446; //@line 1024 "snd_fx.cpp"
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $450 = ((($pChn$0116)) + 4|0); //@line 1007 "snd_fx.cpp"
        if ($bPorta$1$off0100) {
         $storemerge = $446;
        } else {
         $451 = HEAP32[$450>>2]|0; //@line 1009 "snd_fx.cpp"
         $452 = (($451) + ($446))|0; //@line 1009 "snd_fx.cpp"
         $storemerge = $452;
        }
        HEAP32[$450>>2] = $storemerge; //@line 1007 "snd_fx.cpp"
        $453 = ((($pChn$0116)) + 32|0); //@line 1010 "snd_fx.cpp"
        $454 = HEAP32[$453>>2]|0; //@line 1010 "snd_fx.cpp"
        $455 = ($storemerge>>>0)<($454>>>0); //@line 1010 "snd_fx.cpp"
        if ($455) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $456 = HEAP32[$3>>2]|0; //@line 1012 "snd_fx.cpp"
        $457 = $456 & 1048580; //@line 1012 "snd_fx.cpp"
        $458 = ($457|0)==(0); //@line 1012 "snd_fx.cpp"
        if (!($458)) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $459 = ((($pChn$0116)) + 40|0); //@line 1014 "snd_fx.cpp"
        $460 = HEAP32[$459>>2]|0; //@line 1014 "snd_fx.cpp"
        HEAP32[$450>>2] = $460; //@line 1014 "snd_fx.cpp"
        $461 = HEAP32[$11>>2]|0; //@line 1015 "snd_fx.cpp"
        $462 = $461 & 4; //@line 1015 "snd_fx.cpp"
        $463 = ($462|0)!=(0); //@line 1015 "snd_fx.cpp"
        $464 = ($454>>>0)>(4); //@line 1015 "snd_fx.cpp"
        $or$cond143 = $463 & $464; //@line 1015 "snd_fx.cpp"
        if (!($or$cond143)) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $465 = (($454) + -2)|0; //@line 1017 "snd_fx.cpp"
        HEAP32[$450>>2] = $465; //@line 1017 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 1:  {
        $473 = HEAP32[$4>>2]|0; //@line 1030 "snd_fx.cpp"
        $474 = ($473|0)==(0); //@line 1030 "snd_fx.cpp"
        if (!($474)) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $475 = ((($pChn$0116)) + 128|0); //@line 1030 "snd_fx.cpp"
        $476 = HEAP32[$475>>2]|0; //@line 1030 "snd_fx.cpp"
        $477 = ($476|0)==(0); //@line 1030 "snd_fx.cpp"
        if ($477) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $478 = ((($pChn$0116)) + 212|0); //@line 1030 "snd_fx.cpp"
        $479 = HEAP8[$478>>0]|0; //@line 1030 "snd_fx.cpp"
        $480 = ($479<<24>>24)==(0); //@line 1030 "snd_fx.cpp"
        if ($480) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $481 = ($param$1|0)==(0); //@line 1031 "snd_fx.cpp"
        if (!($481)) {
         $486 = ((($pChn$0116)) + 216|0); //@line 1032 "snd_fx.cpp"
         HEAP8[$486>>0] = 1; //@line 1032 "snd_fx.cpp"
         $487 = $param$1&255; //@line 1033 "snd_fx.cpp"
         $488 = ((($pChn$0116)) + 217|0); //@line 1033 "snd_fx.cpp"
         HEAP8[$488>>0] = $487; //@line 1033 "snd_fx.cpp"
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $482 = HEAP32[$3>>2]|0; //@line 1031 "snd_fx.cpp"
        $483 = $482 & 34; //@line 1031 "snd_fx.cpp"
        $484 = ($483|0)==(0); //@line 1031 "snd_fx.cpp"
        if ($484) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $485 = ((($pChn$0116)) + 216|0); //@line 1032 "snd_fx.cpp"
        HEAP8[$485>>0] = 1; //@line 1032 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 15:  {
        $489 = HEAP32[$3>>2]|0; //@line 1038 "snd_fx.cpp"
        $490 = $489 & 1048580; //@line 1038 "snd_fx.cpp"
        $491 = ($490|0)==(0); //@line 1038 "snd_fx.cpp"
        if ($491) {
         $507 = ($param$1|0)==(0); //@line 1044 "snd_fx.cpp"
         if ($507) {
          $510 = ((($pChn$0116)) + 241|0); //@line 1044 "snd_fx.cpp"
          $511 = HEAP8[$510>>0]|0; //@line 1044 "snd_fx.cpp"
          $512 = $511&255; //@line 1044 "snd_fx.cpp"
          $param$7 = $512;
         } else {
          $param$6102 = $param$1;
          label = 215;
         }
        } else {
         $492 = $param$1 & 240; //@line 1040 "snd_fx.cpp"
         $493 = ($492|0)==(0); //@line 1040 "snd_fx.cpp"
         if ($493) {
          $494 = ((($pChn$0116)) + 241|0); //@line 1040 "snd_fx.cpp"
          $495 = HEAP8[$494>>0]|0; //@line 1040 "snd_fx.cpp"
          $496 = $495&255; //@line 1040 "snd_fx.cpp"
          $497 = $496 & 240; //@line 1040 "snd_fx.cpp"
          $498 = $497 | $param$1; //@line 1040 "snd_fx.cpp"
          $param$4 = $498;
         } else {
          $param$4 = $param$1;
         }
         $499 = $param$4 & 15; //@line 1041 "snd_fx.cpp"
         $500 = ($499|0)==(0); //@line 1041 "snd_fx.cpp"
         if ($500) {
          $501 = ((($pChn$0116)) + 241|0); //@line 1041 "snd_fx.cpp"
          $502 = HEAP8[$501>>0]|0; //@line 1041 "snd_fx.cpp"
          $503 = $502&255; //@line 1041 "snd_fx.cpp"
          $504 = $503 & 15; //@line 1041 "snd_fx.cpp"
          $505 = $504 | $param$4; //@line 1041 "snd_fx.cpp"
          $param$5 = $505;
         } else {
          $param$5 = $param$4;
         }
         $506 = $param$5 | 256; //@line 1042 "snd_fx.cpp"
         $param$6102 = $506;
         label = 215;
        }
        if ((label|0) == 215) {
         label = 0;
         $508 = $param$6102&255; //@line 1044 "snd_fx.cpp"
         $509 = ((($pChn$0116)) + 241|0); //@line 1044 "snd_fx.cpp"
         HEAP8[$509>>0] = $508; //@line 1044 "snd_fx.cpp"
         $param$7 = $param$6102;
        }
        __ZN10CSoundFile10RetrigNoteEjj($this,$nChn$0108,$param$7); //@line 1045 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 18:  {
        $513 = HEAP32[$4>>2]|0; //@line 1050 "snd_fx.cpp"
        $514 = ($513|0)==(0); //@line 1050 "snd_fx.cpp"
        if (!($514)) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $515 = ((($pChn$0116)) + 216|0); //@line 1051 "snd_fx.cpp"
        HEAP8[$515>>0] = 18; //@line 1051 "snd_fx.cpp"
        $516 = ($param$1|0)==(0); //@line 1052 "snd_fx.cpp"
        if ($516) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $517 = $param$1&255; //@line 1052 "snd_fx.cpp"
        $518 = ((($pChn$0116)) + 243|0); //@line 1052 "snd_fx.cpp"
        HEAP8[$518>>0] = $517; //@line 1052 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 23:  {
        $519 = HEAP32[$4>>2]|0; //@line 1057 "snd_fx.cpp"
        $520 = ($519|0)==(0); //@line 1057 "snd_fx.cpp"
        if (!($520)) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $521 = HEAP32[$3>>2]|0; //@line 1058 "snd_fx.cpp"
        $not$105 = ($521|0)!=(32); //@line 1058 "snd_fx.cpp"
        $522 = $not$105&1; //@line 1058 "snd_fx.cpp"
        $param$1$ = $param$1 << $522; //@line 1058 "snd_fx.cpp"
        $523 = ($param$1$>>>0)>(128); //@line 1059 "snd_fx.cpp"
        $param$1$$op = $param$1$ << 1; //@line 1060 "snd_fx.cpp"
        $524 = $523 ? 256 : $param$1$$op; //@line 1060 "snd_fx.cpp"
        HEAP32[$12>>2] = $524; //@line 1060 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 24:  {
        $525 = ($param$1|0)==(0); //@line 2141 "snd_fx.cpp"
        if ($525) {
         $526 = HEAP32[$13>>2]|0; //@line 2141 "snd_fx.cpp"
         $$0$i70 = $526;
        } else {
         HEAP32[$13>>2] = $param$1; //@line 2141 "snd_fx.cpp"
         $$0$i70 = $param$1;
        }
        $527 = $$0$i70 & 15; //@line 2142 "snd_fx.cpp"
        $528 = ($527|0)!=(15); //@line 2142 "snd_fx.cpp"
        $529 = $$0$i70 & 240; //@line 1065 "snd_fx.cpp"
        $530 = ($529|0)==(0); //@line 2142 "snd_fx.cpp"
        $or$cond$i71 = $528 | $530; //@line 2142 "snd_fx.cpp"
        do {
         if ($or$cond$i71) {
          $536 = ($529|0)!=(240); //@line 2146 "snd_fx.cpp"
          $537 = ($527|0)==(0); //@line 2146 "snd_fx.cpp"
          $or$cond2$i = $536 | $537; //@line 2146 "snd_fx.cpp"
          $538 = HEAP32[$11>>2]|0; //@line 2151 "snd_fx.cpp"
          $539 = $538 & 4096; //@line 2151 "snd_fx.cpp"
          $540 = ($539|0)==(0); //@line 2151 "snd_fx.cpp"
          if (!($or$cond2$i)) {
           if ($540) {
            $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
            break L194;
           }
           $541 = $527 << 1; //@line 2148 "snd_fx.cpp"
           $542 = (0 - ($541))|0; //@line 2148 "snd_fx.cpp"
           $nGlbSlide$0$i = $542;
           break;
          }
          if (!($540)) {
           $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
           break L194;
          }
          if ($530) {
           $544 = $527 << 1; //@line 2154 "snd_fx.cpp"
           $545 = (0 - ($544))|0; //@line 2154 "snd_fx.cpp"
           $nGlbSlide$0$i = $545;
           break;
          } else {
           $543 = $529 >>> 3; //@line 2153 "snd_fx.cpp"
           $nGlbSlide$0$i = $543;
           break;
          }
         } else {
          $531 = HEAP32[$11>>2]|0; //@line 2144 "snd_fx.cpp"
          $532 = $531 & 4096; //@line 2144 "snd_fx.cpp"
          $533 = ($532|0)==(0); //@line 2144 "snd_fx.cpp"
          if ($533) {
           $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
           break L194;
          }
          $534 = $$0$i70 >>> 4; //@line 2144 "snd_fx.cpp"
          $535 = $534 << 1; //@line 2144 "snd_fx.cpp"
          $nGlbSlide$0$i = $535;
         }
        } while(0);
        $546 = ($nGlbSlide$0$i|0)==(0); //@line 2157 "snd_fx.cpp"
        if ($546) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $547 = HEAP32[$3>>2]|0; //@line 2159 "snd_fx.cpp"
        $not$$i = ($547|0)!=(32); //@line 2159 "snd_fx.cpp"
        $548 = $not$$i&1; //@line 2159 "snd_fx.cpp"
        $nGlbSlide$0$$i = $nGlbSlide$0$i << $548; //@line 2159 "snd_fx.cpp"
        $549 = HEAP32[$12>>2]|0; //@line 2160 "snd_fx.cpp"
        $550 = (($nGlbSlide$0$$i) + ($549))|0; //@line 2160 "snd_fx.cpp"
        $551 = ($550|0)<(0); //@line 2161 "snd_fx.cpp"
        $$$i72 = $551 ? 0 : $550; //@line 2161 "snd_fx.cpp"
        $552 = ($$$i72|0)>(256); //@line 2162 "snd_fx.cpp"
        $$$$i73 = $552 ? 256 : $$$i72; //@line 2162 "snd_fx.cpp"
        HEAP32[$12>>2] = $$$$i73; //@line 2163 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 9:  {
        $553 = HEAP32[$4>>2]|0; //@line 1070 "snd_fx.cpp"
        $554 = ($553|0)==(0); //@line 1070 "snd_fx.cpp"
        if (!($554)) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $555 = HEAP32[$11>>2]|0; //@line 1071 "snd_fx.cpp"
        $556 = $555 & 16384; //@line 1071 "snd_fx.cpp"
        $557 = ($556|0)==(0); //@line 1071 "snd_fx.cpp"
        if ($557) {
         $558 = HEAP32[$33>>2]|0; //@line 1071 "snd_fx.cpp"
         $559 = $558 & -2049; //@line 1071 "snd_fx.cpp"
         HEAP32[$33>>2] = $559; //@line 1071 "snd_fx.cpp"
        }
        $560 = HEAP32[$3>>2]|0; //@line 1072 "snd_fx.cpp"
        $561 = $560 & 1048612; //@line 1072 "snd_fx.cpp"
        $562 = ($561|0)==(0); //@line 1072 "snd_fx.cpp"
        do {
         if ($562) {
          $564 = ($param$1>>>0)<(129); //@line 1076 "snd_fx.cpp"
          if ($564) {
           $565 = $param$1 << 1; //@line 1078 "snd_fx.cpp"
           $566 = ((($pChn$0116)) + 120|0); //@line 1078 "snd_fx.cpp"
           HEAP32[$566>>2] = $565; //@line 1078 "snd_fx.cpp"
           break;
          }
          $567 = ($param$1|0)==(164); //@line 1080 "snd_fx.cpp"
          if ($567) {
           $568 = HEAP32[$33>>2]|0; //@line 1082 "snd_fx.cpp"
           $569 = $568 | 2048; //@line 1082 "snd_fx.cpp"
           HEAP32[$33>>2] = $569; //@line 1082 "snd_fx.cpp"
           $570 = ((($pChn$0116)) + 120|0); //@line 1083 "snd_fx.cpp"
           HEAP32[$570>>2] = 128; //@line 1083 "snd_fx.cpp"
          }
         } else {
          $563 = ((($pChn$0116)) + 120|0); //@line 1074 "snd_fx.cpp"
          HEAP32[$563>>2] = $param$1; //@line 1074 "snd_fx.cpp"
         }
        } while(0);
        $571 = HEAP32[$33>>2]|0; //@line 1085 "snd_fx.cpp"
        $572 = $571 | 16777216; //@line 1085 "snd_fx.cpp"
        HEAP32[$33>>2] = $572; //@line 1085 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 29:  {
        $573 = ($param$1|0)==(0); //@line 1517 "snd_fx.cpp"
        if ($573) {
         $576 = ((($pChn$0116)) + 222|0); //@line 1517 "snd_fx.cpp"
         $577 = HEAP8[$576>>0]|0; //@line 1517 "snd_fx.cpp"
         $578 = $577&255; //@line 1517 "snd_fx.cpp"
         $$0$i74 = $578;
        } else {
         $574 = $param$1&255; //@line 1517 "snd_fx.cpp"
         $575 = ((($pChn$0116)) + 222|0); //@line 1517 "snd_fx.cpp"
         HEAP8[$575>>0] = $574; //@line 1517 "snd_fx.cpp"
         $$0$i74 = $param$1;
        }
        $579 = HEAP32[$3>>2]|0; //@line 1518 "snd_fx.cpp"
        $580 = $579 & 290; //@line 1518 "snd_fx.cpp"
        $581 = ($580|0)==(0); //@line 1518 "snd_fx.cpp"
        L335: do {
         if ($581) {
          $600 = HEAP32[$11>>2]|0; //@line 1544 "snd_fx.cpp"
          $601 = $600 & 4096; //@line 1544 "snd_fx.cpp"
          $602 = ($601|0)==(0); //@line 1544 "snd_fx.cpp"
          if (!($602)) {
           $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
           break L194;
          }
          $603 = $$0$i74 & 15; //@line 1546 "snd_fx.cpp"
          $604 = ($603|0)==(0); //@line 1546 "snd_fx.cpp"
          if ($604) {
           $607 = $$0$i74 >>> 2; //@line 1547 "snd_fx.cpp"
           $608 = $607 & 60; //@line 1547 "snd_fx.cpp"
           $nPanSlide$0$i78 = $608;
           break;
          } else {
           $605 = $603 << 2; //@line 1546 "snd_fx.cpp"
           $606 = (0 - ($605))|0; //@line 1546 "snd_fx.cpp"
           $nPanSlide$0$i78 = $606;
           break;
          }
         } else {
          $582 = $$0$i74 & 15; //@line 1520 "snd_fx.cpp"
          $583 = ($582|0)==(15); //@line 1520 "snd_fx.cpp"
          $584 = $$0$i74 & 240; //@line 1520 "snd_fx.cpp"
          do {
           if ($583) {
            $585 = ($584|0)==(0); //@line 1520 "snd_fx.cpp"
            $586 = HEAP32[$11>>2]|0; //@line 1536 "snd_fx.cpp"
            $587 = $586 & 4096; //@line 1536 "snd_fx.cpp"
            $588 = ($587|0)==(0); //@line 1536 "snd_fx.cpp"
            if ($585) {
             if ($588) {
              $$pre$phi$i75140141Z2D = 0;$822 = 0;
              break;
             } else {
              $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
              break L194;
             }
            }
            if ($588) {
             $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
             break L194;
            }
            $589 = $584 >>> 2; //@line 1524 "snd_fx.cpp"
            $590 = (0 - ($589))|0; //@line 1525 "snd_fx.cpp"
            $nPanSlide$0$i78 = $590;
            break L335;
           } else {
            $591 = ($584|0)!=(240); //@line 1528 "snd_fx.cpp"
            $592 = ($582|0)==(0); //@line 1528 "snd_fx.cpp"
            $or$cond$i76 = $592 | $591; //@line 1528 "snd_fx.cpp"
            $593 = HEAP32[$11>>2]|0; //@line 1536 "snd_fx.cpp"
            $594 = $593 & 4096; //@line 1536 "snd_fx.cpp"
            $595 = ($594|0)==(0); //@line 1536 "snd_fx.cpp"
            if ($or$cond$i76) {
             if ($595) {
              $$pre$phi$i75140141Z2D = $584;$822 = $592;
              break;
             } else {
              $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
              break L194;
             }
            } else {
             $596 = $582 << 2; //@line 1532 "snd_fx.cpp"
             if ($595) {
              $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
              break L194;
             } else {
              $nPanSlide$0$i78 = $596;
              break L335;
             }
            }
           }
          } while(0);
          if ($822) {
           $598 = $$pre$phi$i75140141Z2D >>> 2; //@line 1539 "snd_fx.cpp"
           $599 = (0 - ($598))|0; //@line 1539 "snd_fx.cpp"
           $nPanSlide$0$i78 = $599;
           break;
          } else {
           $597 = $582 << 2; //@line 1538 "snd_fx.cpp"
           $nPanSlide$0$i78 = $597;
           break;
          }
         }
        } while(0);
        $609 = ($nPanSlide$0$i78|0)==(0); //@line 1550 "snd_fx.cpp"
        if ($609) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $610 = ((($pChn$0116)) + 120|0); //@line 1552 "snd_fx.cpp"
        $611 = HEAP32[$610>>2]|0; //@line 1552 "snd_fx.cpp"
        $612 = (($611) + ($nPanSlide$0$i78))|0; //@line 1552 "snd_fx.cpp"
        $613 = ($612|0)<(0); //@line 1553 "snd_fx.cpp"
        $$$i79 = $613 ? 0 : $612; //@line 1553 "snd_fx.cpp"
        $614 = ($$$i79|0)>(256); //@line 1554 "snd_fx.cpp"
        $$$$i80 = $614 ? 256 : $$$i79; //@line 1554 "snd_fx.cpp"
        HEAP32[$610>>2] = $$$$i80; //@line 1555 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 8:  {
        $615 = $param$1 & 15; //@line 1589 "snd_fx.cpp"
        $616 = ($615|0)==(0); //@line 1589 "snd_fx.cpp"
        if (!($616)) {
         $617 = $615 << 2; //@line 1589 "snd_fx.cpp"
         $618 = $617&255; //@line 1589 "snd_fx.cpp"
         $619 = ((($pChn$0116)) + 229|0); //@line 1589 "snd_fx.cpp"
         HEAP8[$619>>0] = $618; //@line 1589 "snd_fx.cpp"
        }
        $620 = $param$1 & 240; //@line 1590 "snd_fx.cpp"
        $621 = ($620|0)==(0); //@line 1590 "snd_fx.cpp"
        if (!($621)) {
         $622 = $param$1 >>> 4; //@line 1590 "snd_fx.cpp"
         $623 = $622 & 15; //@line 1590 "snd_fx.cpp"
         $624 = $623&255; //@line 1590 "snd_fx.cpp"
         $625 = ((($pChn$0116)) + 228|0); //@line 1590 "snd_fx.cpp"
         HEAP8[$625>>0] = $624; //@line 1590 "snd_fx.cpp"
        }
        $626 = HEAP32[$33>>2]|0; //@line 1591 "snd_fx.cpp"
        $627 = $626 | 131072; //@line 1591 "snd_fx.cpp"
        HEAP32[$33>>2] = $627; //@line 1591 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 26:  {
        $628 = $param$1 & 15; //@line 1450 "snd_fx.cpp"
        $629 = ($628|0)==(0); //@line 1450 "snd_fx.cpp"
        if (!($629)) {
         $630 = $628&255; //@line 1450 "snd_fx.cpp"
         $631 = ((($pChn$0116)) + 226|0); //@line 1450 "snd_fx.cpp"
         HEAP8[$631>>0] = $630; //@line 1450 "snd_fx.cpp"
        }
        $632 = $param$1 & 240; //@line 1451 "snd_fx.cpp"
        $633 = ($632|0)==(0); //@line 1451 "snd_fx.cpp"
        if (!($633)) {
         $634 = $param$1 >>> 4; //@line 1451 "snd_fx.cpp"
         $635 = $634 & 15; //@line 1451 "snd_fx.cpp"
         $636 = $635&255; //@line 1451 "snd_fx.cpp"
         $637 = ((($pChn$0116)) + 225|0); //@line 1451 "snd_fx.cpp"
         HEAP8[$637>>0] = $636; //@line 1451 "snd_fx.cpp"
        }
        $638 = HEAP32[$33>>2]|0; //@line 1452 "snd_fx.cpp"
        $639 = $638 | 65536; //@line 1452 "snd_fx.cpp"
        HEAP32[$33>>2] = $639; //@line 1452 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 19:  {
        __ZN10CSoundFile19ExtendedMODCommandsEjj($this,$nChn$0108,$param$1); //@line 1105 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 20:  {
        __ZN10CSoundFile19ExtendedS3MCommandsEjj($this,$nChn$0108,$param$1); //@line 1110 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 25:  {
        $640 = HEAP32[$4>>2]|0; //@line 1115 "snd_fx.cpp"
        $641 = ($640|0)==(0); //@line 1115 "snd_fx.cpp"
        if (!($641)) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        __ZN10CSoundFile6KeyOffEj($this,$nChn$0108); //@line 1115 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 28:  {
        $642 = $param$1 & 240; //@line 1120 "snd_fx.cpp"
        switch ($642|0) {
        case 16:  {
         $643 = $param$1 & 15; //@line 1122 "snd_fx.cpp"
         $644 = HEAP32[$3>>2]|0; //@line 1355 "snd_fx.cpp"
         $645 = $644 & 1048580; //@line 1355 "snd_fx.cpp"
         $646 = ($645|0)==(0); //@line 1355 "snd_fx.cpp"
         do {
          if ($646) {
           $$0$i53 = $643;
          } else {
           $647 = ($643|0)==(0); //@line 1357 "snd_fx.cpp"
           if ($647) {
            $650 = ((($pChn$0116)) + 221|0); //@line 1357 "snd_fx.cpp"
            $651 = HEAP8[$650>>0]|0; //@line 1357 "snd_fx.cpp"
            $652 = $651&255; //@line 1357 "snd_fx.cpp"
            $$0$i53 = $652;
            break;
           } else {
            $648 = $643&255; //@line 1357 "snd_fx.cpp"
            $649 = ((($pChn$0116)) + 221|0); //@line 1357 "snd_fx.cpp"
            HEAP8[$649>>0] = $648; //@line 1357 "snd_fx.cpp"
            $$0$i53 = $643;
            break;
           }
          }
         } while(0);
         $653 = HEAP32[$11>>2]|0; //@line 1359 "snd_fx.cpp"
         $654 = $653 & 4096; //@line 1359 "snd_fx.cpp"
         $655 = ($654|0)==(0); //@line 1359 "snd_fx.cpp"
         if ($655) {
          $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
          break L194;
         }
         $656 = ((($pChn$0116)) + 128|0); //@line 1361 "snd_fx.cpp"
         $657 = HEAP32[$656>>2]|0; //@line 1361 "snd_fx.cpp"
         $658 = ($657|0)!=(0); //@line 1361 "snd_fx.cpp"
         $659 = ($$0$i53|0)!=(0); //@line 1361 "snd_fx.cpp"
         $or$cond$i54 = $659 & $658; //@line 1361 "snd_fx.cpp"
         if (!($or$cond$i54)) {
          $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
          break L194;
         }
         $660 = $653 & 16; //@line 1363 "snd_fx.cpp"
         $661 = ($660|0)==(0); //@line 1363 "snd_fx.cpp"
         $$not$i55 = $646 ^ 1; //@line 1363 "snd_fx.cpp"
         $brmerge$i56 = $661 | $$not$i55; //@line 1363 "snd_fx.cpp"
         if ($brmerge$i56) {
          $666 = (($657) - ($$0$i53))|0; //@line 1368 "snd_fx.cpp"
          $storemerge$i57 = $666;
         } else {
          $662 = $$0$i53 & 15; //@line 1365 "snd_fx.cpp"
          $663 = (42708 + ($662<<2)|0); //@line 1365 "snd_fx.cpp"
          $664 = HEAP32[$663>>2]|0; //@line 1365 "snd_fx.cpp"
          $665 = (__Z8_muldivrlll($657,$664,65536)|0); //@line 1365 "snd_fx.cpp"
          $storemerge$i57 = $665;
         }
         $667 = ($storemerge$i57|0)<(1); //@line 1370 "snd_fx.cpp"
         $$storemerge$i58 = $667 ? 1 : $storemerge$i57; //@line 1370 "snd_fx.cpp"
         HEAP32[$656>>2] = $$storemerge$i58; //@line 1370 "snd_fx.cpp"
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
         break;
        }
        case 32:  {
         $668 = $param$1 & 15; //@line 1123 "snd_fx.cpp"
         $669 = HEAP32[$3>>2]|0; //@line 1379 "snd_fx.cpp"
         $670 = $669 & 1048580; //@line 1379 "snd_fx.cpp"
         $671 = ($670|0)==(0); //@line 1379 "snd_fx.cpp"
         do {
          if ($671) {
           $$0$i46 = $668;
          } else {
           $672 = ($668|0)==(0); //@line 1381 "snd_fx.cpp"
           if ($672) {
            $675 = ((($pChn$0116)) + 221|0); //@line 1381 "snd_fx.cpp"
            $676 = HEAP8[$675>>0]|0; //@line 1381 "snd_fx.cpp"
            $677 = $676&255; //@line 1381 "snd_fx.cpp"
            $$0$i46 = $677;
            break;
           } else {
            $673 = $668&255; //@line 1381 "snd_fx.cpp"
            $674 = ((($pChn$0116)) + 221|0); //@line 1381 "snd_fx.cpp"
            HEAP8[$674>>0] = $673; //@line 1381 "snd_fx.cpp"
            $$0$i46 = $668;
            break;
           }
          }
         } while(0);
         $678 = HEAP32[$11>>2]|0; //@line 1383 "snd_fx.cpp"
         $679 = $678 & 4096; //@line 1383 "snd_fx.cpp"
         $680 = ($679|0)==(0); //@line 1383 "snd_fx.cpp"
         if ($680) {
          $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
          break L194;
         }
         $681 = ((($pChn$0116)) + 128|0); //@line 1385 "snd_fx.cpp"
         $682 = HEAP32[$681>>2]|0; //@line 1385 "snd_fx.cpp"
         $683 = ($682|0)!=(0); //@line 1385 "snd_fx.cpp"
         $684 = ($$0$i46|0)!=(0); //@line 1385 "snd_fx.cpp"
         $or$cond$i47 = $684 & $683; //@line 1385 "snd_fx.cpp"
         if (!($or$cond$i47)) {
          $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
          break L194;
         }
         $685 = $678 & 16; //@line 1387 "snd_fx.cpp"
         $686 = ($685|0)==(0); //@line 1387 "snd_fx.cpp"
         $$not$i = $671 ^ 1; //@line 1387 "snd_fx.cpp"
         $brmerge$i = $686 | $$not$i; //@line 1387 "snd_fx.cpp"
         if ($brmerge$i) {
          $691 = (($682) + ($$0$i46))|0; //@line 1392 "snd_fx.cpp"
          $storemerge$i = $691;
         } else {
          $687 = $$0$i46 & 15; //@line 1389 "snd_fx.cpp"
          $688 = (42772 + ($687<<2)|0); //@line 1389 "snd_fx.cpp"
          $689 = HEAP32[$688>>2]|0; //@line 1389 "snd_fx.cpp"
          $690 = (__Z8_muldivrlll($682,$689,65536)|0); //@line 1389 "snd_fx.cpp"
          $storemerge$i = $690;
         }
         $692 = ($storemerge$i|0)>(65535); //@line 1394 "snd_fx.cpp"
         $$storemerge$i = $692 ? 65535 : $storemerge$i; //@line 1394 "snd_fx.cpp"
         HEAP32[$681>>2] = $$storemerge$i; //@line 1394 "snd_fx.cpp"
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
         break;
        }
        case 160: case 144: case 112: case 96: case 80:  {
         __ZN10CSoundFile19ExtendedS3MCommandsEjj($this,$nChn$0108,$param$1); //@line 1129 "snd_fx.cpp"
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
         break;
        }
        default: {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        }
        break;
       }
       case 21:  {
        $693 = HEAP32[$4>>2]|0; //@line 1135 "snd_fx.cpp"
        $694 = ($693|0)==(0); //@line 1135 "snd_fx.cpp"
        $695 = ($param$1>>>0)<(65); //@line 1136 "snd_fx.cpp"
        $or$cond29 = $695 & $694; //@line 1135 "snd_fx.cpp"
        if (!($or$cond29)) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $696 = ((($pChn$0116)) + 168|0); //@line 1138 "snd_fx.cpp"
        HEAP32[$696>>2] = $param$1; //@line 1138 "snd_fx.cpp"
        $697 = HEAP32[$33>>2]|0; //@line 1139 "snd_fx.cpp"
        $698 = $697 | 16777216; //@line 1139 "snd_fx.cpp"
        HEAP32[$33>>2] = $698; //@line 1139 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 22:  {
        $699 = ($param$1|0)==(0); //@line 1599 "snd_fx.cpp"
        if ($699) {
         $702 = ((($pChn$0116)) + 223|0); //@line 1599 "snd_fx.cpp"
         $703 = HEAP8[$702>>0]|0; //@line 1599 "snd_fx.cpp"
         $704 = $703&255; //@line 1599 "snd_fx.cpp"
         $$0$i43 = $704;
        } else {
         $700 = $param$1&255; //@line 1599 "snd_fx.cpp"
         $701 = ((($pChn$0116)) + 223|0); //@line 1599 "snd_fx.cpp"
         HEAP8[$701>>0] = $700; //@line 1599 "snd_fx.cpp"
         $$0$i43 = $param$1;
        }
        $705 = $$0$i43 & 15; //@line 1600 "snd_fx.cpp"
        $706 = ($705|0)!=(15); //@line 1600 "snd_fx.cpp"
        $707 = $$0$i43 & 240; //@line 1145 "snd_fx.cpp"
        $708 = ($707|0)==(0); //@line 1600 "snd_fx.cpp"
        $or$cond$i = $706 | $708; //@line 1600 "snd_fx.cpp"
        do {
         if ($or$cond$i) {
          $713 = ($707|0)!=(240); //@line 1604 "snd_fx.cpp"
          $714 = ($705|0)==(0); //@line 1604 "snd_fx.cpp"
          $or$cond3$i = $713 | $714; //@line 1604 "snd_fx.cpp"
          $715 = HEAP32[$11>>2]|0; //@line 1609 "snd_fx.cpp"
          $716 = $715 & 4096; //@line 1609 "snd_fx.cpp"
          $717 = ($716|0)==(0); //@line 1609 "snd_fx.cpp"
          if (!($or$cond3$i)) {
           $718 = (0 - ($705))|0; //@line 1606 "snd_fx.cpp"
           if ($717) {
            $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
            break L194;
           } else {
            $nChnSlide$0$i = $718;
            break;
           }
          }
          if (!($717)) {
           $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
           break L194;
          }
          if ($714) {
           $720 = $707 >>> 4; //@line 1612 "snd_fx.cpp"
           $nChnSlide$0$i = $720;
           break;
          } else {
           $719 = (0 - ($705))|0; //@line 1611 "snd_fx.cpp"
           $nChnSlide$0$i = $719;
           break;
          }
         } else {
          $709 = HEAP32[$11>>2]|0; //@line 1602 "snd_fx.cpp"
          $710 = $709 & 4096; //@line 1602 "snd_fx.cpp"
          $711 = ($710|0)==(0); //@line 1602 "snd_fx.cpp"
          $712 = $$0$i43 >>> 4; //@line 1602 "snd_fx.cpp"
          if ($711) {
           $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
           break L194;
          } else {
           $nChnSlide$0$i = $712;
          }
         }
        } while(0);
        $721 = ($nChnSlide$0$i|0)==(0); //@line 1615 "snd_fx.cpp"
        if ($721) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $722 = ((($pChn$0116)) + 168|0); //@line 1617 "snd_fx.cpp"
        $723 = HEAP32[$722>>2]|0; //@line 1617 "snd_fx.cpp"
        $724 = (($723) + ($nChnSlide$0$i))|0; //@line 1617 "snd_fx.cpp"
        $725 = ($724|0)<(0); //@line 1618 "snd_fx.cpp"
        $$$i44 = $725 ? 0 : $724; //@line 1618 "snd_fx.cpp"
        $726 = ($$$i44|0)>(64); //@line 1619 "snd_fx.cpp"
        $$$$i = $726 ? 64 : $$$i44; //@line 1619 "snd_fx.cpp"
        HEAP32[$722>>2] = $$$$i; //@line 1620 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 27:  {
        $727 = $param$1 & 15; //@line 1459 "snd_fx.cpp"
        $728 = ($727|0)==(0); //@line 1459 "snd_fx.cpp"
        if (!($728)) {
         $729 = $727&255; //@line 1459 "snd_fx.cpp"
         $730 = ((($pChn$0116)) + 232|0); //@line 1459 "snd_fx.cpp"
         HEAP8[$730>>0] = $729; //@line 1459 "snd_fx.cpp"
        }
        $731 = $param$1 & 240; //@line 1460 "snd_fx.cpp"
        $732 = ($731|0)==(0); //@line 1460 "snd_fx.cpp"
        if (!($732)) {
         $733 = $param$1 >>> 4; //@line 1460 "snd_fx.cpp"
         $734 = $733 & 15; //@line 1460 "snd_fx.cpp"
         $735 = $734&255; //@line 1460 "snd_fx.cpp"
         $736 = ((($pChn$0116)) + 231|0); //@line 1460 "snd_fx.cpp"
         HEAP8[$736>>0] = $735; //@line 1460 "snd_fx.cpp"
        }
        $737 = HEAP32[$33>>2]|0; //@line 1461 "snd_fx.cpp"
        $738 = $737 | 262144; //@line 1461 "snd_fx.cpp"
        HEAP32[$33>>2] = $738; //@line 1461 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 30:  {
        $739 = HEAP32[$4>>2]|0; //@line 1155 "snd_fx.cpp"
        $740 = ($739|0)==(0); //@line 1155 "snd_fx.cpp"
        if (!($740)) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $741 = ((($pChn$0116)) + 148|0); //@line 1157 "snd_fx.cpp"
        HEAP32[$741>>2] = $param$1; //@line 1157 "snd_fx.cpp"
        $742 = ((($pChn$0116)) + 152|0); //@line 1158 "snd_fx.cpp"
        HEAP32[$742>>2] = $param$1; //@line 1158 "snd_fx.cpp"
        $743 = ((($pChn$0116)) + 156|0); //@line 1159 "snd_fx.cpp"
        HEAP32[$743>>2] = $param$1; //@line 1159 "snd_fx.cpp"
        $744 = ((($pChn$0116)) + 140|0); //@line 1160 "snd_fx.cpp"
        $745 = HEAP32[$744>>2]|0; //@line 1160 "snd_fx.cpp"
        $746 = ($745|0)==(0|0); //@line 1160 "snd_fx.cpp"
        if ($746) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $747 = HEAP32[$33>>2]|0; //@line 1163 "snd_fx.cpp"
        $748 = $747 & 4194304; //@line 1163 "snd_fx.cpp"
        $749 = ($748|0)==(0); //@line 1163 "snd_fx.cpp"
        if ($749) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $750 = ((($745)) + 557|0); //@line 1163 "snd_fx.cpp"
        $751 = HEAP8[$750>>0]|0; //@line 1163 "snd_fx.cpp"
        $752 = ($751<<24>>24)==(0); //@line 1163 "snd_fx.cpp"
        if ($752) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $753 = $751&255; //@line 1163 "snd_fx.cpp"
        $754 = (($753) + -1)|0; //@line 1163 "snd_fx.cpp"
        $755 = (((($745)) + 76|0) + ($754<<1)|0); //@line 1163 "snd_fx.cpp"
        $756 = HEAP16[$755>>1]|0; //@line 1163 "snd_fx.cpp"
        $757 = $756&65535; //@line 1163 "snd_fx.cpp"
        $758 = ($param$1>>>0)>($757>>>0); //@line 1163 "snd_fx.cpp"
        if (!($758)) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $759 = $747 & -4194305; //@line 1165 "snd_fx.cpp"
        HEAP32[$33>>2] = $759; //@line 1165 "snd_fx.cpp"
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 12:  {
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $param$1;
        break L194;
        break;
       }
       case 14:  {
        $nBreakRow$1 = $param$1;$nPosJump$1 = $nPosJump$0114;
        break L194;
        break;
       }
       case 31:  {
        $760 = HEAP32[$4>>2]|0; //@line 1183 "snd_fx.cpp"
        $761 = ($760|0)==(0); //@line 1183 "snd_fx.cpp"
        if (!($761)) {
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        $762 = ($param$1>>>0)<(128); //@line 1184 "snd_fx.cpp"
        if ($762) {
         $763 = ((($pChn$0116)) + 254|0); //@line 1185 "snd_fx.cpp"
         $764 = HEAP8[$763>>0]|0; //@line 1185 "snd_fx.cpp"
         $765 = $764&255; //@line 1185 "snd_fx.cpp"
         $766 = $765 << 5; //@line 1185 "snd_fx.cpp"
         $767 = (((($this)) + 53888|0) + ($766)|0); //@line 1185 "snd_fx.cpp"
         __ZN10CSoundFile16ProcessMidiMacroEjPKcj($this,$nChn$0108,$767,$param$1); //@line 1185 "snd_fx.cpp"
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        } else {
         $768 = $param$1 << 5; //@line 1187 "snd_fx.cpp"
         $769 = $768 & 4064; //@line 1187 "snd_fx.cpp"
         $770 = (((($this)) + 54400|0) + ($769)|0); //@line 1187 "snd_fx.cpp"
         __ZN10CSoundFile16ProcessMidiMacroEjPKcj($this,$nChn$0108,$770,0); //@line 1187 "snd_fx.cpp"
         $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
         break L194;
        }
        break;
       }
       default: {
        $nBreakRow$1 = $nBreakRow$0115;$nPosJump$1 = $nPosJump$0114;
        break L194;
       }
       }
      } while(0);
     }
    } while(0);
    $771 = (($nChn$0108) + 1)|0; //@line 734 "snd_fx.cpp"
    $772 = ((($pChn$0116)) + 256|0); //@line 734 "snd_fx.cpp"
    $773 = HEAP32[$0>>2]|0; //@line 734 "snd_fx.cpp"
    $774 = ($771>>>0)<($773>>>0); //@line 734 "snd_fx.cpp"
    if ($774) {
     $62 = $773;$nBreakRow$0115 = $nBreakRow$1;$nChn$0108 = $771;$nPatLoopRow$0113 = $nPatLoopRow$1;$nPosJump$0114 = $nPosJump$1;$pChn$0116 = $772;
    } else {
     $$pre$phi132Z2D = $4;$813 = $773;$nBreakRow$0$lcssa = $nBreakRow$1;$nPatLoopRow$0$lcssa = $nPatLoopRow$1;$nPosJump$0$lcssa = $nPosJump$1;
     break L1;
    }
   }
  }
 } while(0);
 $775 = HEAP32[$$pre$phi132Z2D>>2]|0; //@line 1194 "snd_fx.cpp"
 $776 = ($775|0)==(0); //@line 1194 "snd_fx.cpp"
 if (!($776)) {
  return 1; //@line 1245 "snd_fx.cpp"
 }
 $777 = ($nPatLoopRow$0$lcssa|0)>(-1); //@line 1197 "snd_fx.cpp"
 if ($777) {
  $778 = ((($this)) + 59728|0); //@line 1199 "snd_fx.cpp"
  $779 = HEAP32[$778>>2]|0; //@line 1199 "snd_fx.cpp"
  $780 = ((($this)) + 59732|0); //@line 1199 "snd_fx.cpp"
  HEAP32[$780>>2] = $779; //@line 1199 "snd_fx.cpp"
  $781 = ((($this)) + 59716|0); //@line 1200 "snd_fx.cpp"
  HEAP32[$781>>2] = $nPatLoopRow$0$lcssa; //@line 1200 "snd_fx.cpp"
  $782 = ((($this)) + 59700|0); //@line 1201 "snd_fx.cpp"
  $783 = HEAP32[$782>>2]|0; //@line 1201 "snd_fx.cpp"
  $784 = ($783|0)==(0); //@line 1201 "snd_fx.cpp"
  if ($784) {
   return 1; //@line 1245 "snd_fx.cpp"
  }
  $785 = (($nPatLoopRow$0$lcssa) + 1)|0; //@line 1201 "snd_fx.cpp"
  HEAP32[$781>>2] = $785; //@line 1201 "snd_fx.cpp"
  return 1; //@line 1245 "snd_fx.cpp"
 }
 $786 = $nBreakRow$0$lcssa & $nPosJump$0$lcssa; //@line 1204 "snd_fx.cpp"
 $787 = ($786|0)>(-1); //@line 1204 "snd_fx.cpp"
 if (!($787)) {
  return 1; //@line 1245 "snd_fx.cpp"
 }
 $788 = ($nPosJump$0$lcssa|0)<(0); //@line 1207 "snd_fx.cpp"
 $789 = ((($this)) + 59728|0); //@line 1207 "snd_fx.cpp"
 $790 = HEAP32[$789>>2]|0; //@line 1207 "snd_fx.cpp"
 $791 = (($790) + 1)|0; //@line 1207 "snd_fx.cpp"
 $nPosJump$2 = $788 ? $791 : $nPosJump$0$lcssa; //@line 1207 "snd_fx.cpp"
 $792 = ($nBreakRow$0$lcssa|0)<(0); //@line 1208 "snd_fx.cpp"
 $$nBreakRow$0 = $792 ? 0 : $nBreakRow$0$lcssa; //@line 1208 "snd_fx.cpp"
 $793 = ($nPosJump$2|0)<($790|0); //@line 1211 "snd_fx.cpp"
 if ($793) {
  $$phi$trans$insert127 = ((($this)) + 59720|0);
  $$pre128 = HEAP32[$$phi$trans$insert127>>2]|0; //@line 1214 "snd_fx.cpp"
  $798 = $$pre128;
  label = 347;
 } else {
  $794 = ($nPosJump$2|0)==($790|0); //@line 1212 "snd_fx.cpp"
  if ($794) {
   $795 = ((($this)) + 59720|0); //@line 1212 "snd_fx.cpp"
   $796 = HEAP32[$795>>2]|0; //@line 1212 "snd_fx.cpp"
   $797 = ($$nBreakRow$0|0)>($796|0); //@line 1212 "snd_fx.cpp"
   if (!($797)) {
    $798 = $796;
    label = 347;
   }
  }
 }
 do {
  if ((label|0) == 347) {
   $799 = (__ZNK10CSoundFile19IsValidBackwardJumpEjjjj($this,$790,$798,$nPosJump$2,$$nBreakRow$0)|0); //@line 1214 "snd_fx.cpp"
   if (!($799)) {
    $800 = ((($this)) + 59772|0); //@line 1216 "snd_fx.cpp"
    $801 = HEAP32[$800>>2]|0; //@line 1216 "snd_fx.cpp"
    $802 = ($801|0)==(0); //@line 1216 "snd_fx.cpp"
    if ($802) {
     $805 = ((($this)) + 59776|0); //@line 1228 "snd_fx.cpp"
     $806 = HEAP32[$805>>2]|0; //@line 1228 "snd_fx.cpp"
     HEAP32[$800>>2] = $806; //@line 1228 "snd_fx.cpp"
     return 1; //@line 1245 "snd_fx.cpp"
    } else {
     $803 = ($801|0)>(0); //@line 1218 "snd_fx.cpp"
     if (!($803)) {
      break;
     }
     $804 = (($801) + -1)|0; //@line 1218 "snd_fx.cpp"
     HEAP32[$800>>2] = $804; //@line 1218 "snd_fx.cpp"
     break;
    }
   }
  }
 } while(0);
 $807 = ($nPosJump$2|0)<(256); //@line 1233 "snd_fx.cpp"
 if (!($807)) {
  return 1; //@line 1245 "snd_fx.cpp"
 }
 $808 = ($nPosJump$2|0)==($790|0); //@line 1234 "snd_fx.cpp"
 if ($808) {
  $809 = ((($this)) + 59720|0); //@line 1234 "snd_fx.cpp"
  $810 = HEAP32[$809>>2]|0; //@line 1234 "snd_fx.cpp"
  $811 = ($$nBreakRow$0|0)==($810|0); //@line 1234 "snd_fx.cpp"
  if ($811) {
   return 1; //@line 1245 "snd_fx.cpp"
  }
 } else {
  $812 = ($813|0)==(0); //@line 1238 "snd_fx.cpp"
  if (!($812)) {
   $i$0107 = 0;
   while(1) {
    $814 = (((($this) + ($i$0107<<8)|0)) + 245|0); //@line 1238 "snd_fx.cpp"
    HEAP8[$814>>0] = 0; //@line 1238 "snd_fx.cpp"
    $815 = (($i$0107) + 1)|0; //@line 1238 "snd_fx.cpp"
    $816 = ($815>>>0)<($813>>>0); //@line 1238 "snd_fx.cpp"
    if ($816) {
     $i$0107 = $815;
    } else {
     break;
    }
   }
  }
 }
 $817 = ((($this)) + 59732|0); //@line 1240 "snd_fx.cpp"
 HEAP32[$817>>2] = $nPosJump$2; //@line 1240 "snd_fx.cpp"
 $818 = ((($this)) + 59716|0); //@line 1241 "snd_fx.cpp"
 HEAP32[$818>>2] = $$nBreakRow$0; //@line 1241 "snd_fx.cpp"
 return 1; //@line 1245 "snd_fx.cpp"
}
function __ZN10CSoundFile8CheckNNAEjjib($this,$nChn,$instr,$note,$bForceCut) {
 $this = $this|0;
 $nChn = $nChn|0;
 $instr = $instr|0;
 $note = $note|0;
 $bForceCut = $bForceCut|0;
 var $$0$i = 0, $$0$i34 = 0, $$1 = 0, $$lobit$i = 0, $$lobit$i26 = 0, $$note = 0, $$off = 0, $$phi$trans$insert$i = 0, $$phi$trans$insert$i28 = 0, $$pre$i = 0, $$pre$i29 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0;
 var $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0;
 var $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $brmerge = 0, $envpos$05$i = 0, $envpos$05$i22 = 0, $envpos$1$i = 0, $envpos$1$i33 = 0, $exitcond = 0, $exitcond50 = 0, $i$011$i = 0, $i$011$i14 = 0;
 var $i$046 = 0, $i$046$us = 0, $j$07$i = 0, $j$07$i20 = 0, $or$cond5 = 0, $or$cond5$us = 0, $or$cond6 = 0, $or$cond6$us = 0, $or$cond7 = 0, $or$cond7$us = 0, $p2$044 = 0, $p2$044$us = 0, $pHeader$0 = 0, $pSample$0 = 0, $pi$010$i = 0, $pi$010$i15 = 0, $pj$06$i = 0, $pj$06$i21 = 0, $result$03$i = 0, $result$03$i24 = 0;
 var $result$1$i = 0, $result$1$i31 = 0, $v$0$$i = 0, $v$0$$i27 = 0, $v$0$i = 0, $v$0$i25 = 0, $vol$04$i = 0, $vol$04$i23 = 0, $vol$1$i = 0, $vol$1$i32 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($this) + ($nChn<<8)|0); //@line 598 "snd_fx.cpp"
 $1 = (((($this) + ($nChn<<8)|0)) + 140|0); //@line 599 "snd_fx.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 599 "snd_fx.cpp"
 $3 = ($note|0)>(128); //@line 601 "snd_fx.cpp"
 $$note = $3 ? 0 : $note; //@line 601 "snd_fx.cpp"
 $4 = ($$note|0)<(1); //@line 602 "snd_fx.cpp"
 if ($4) {
  return; //@line 725 "snd_fx.cpp"
 }
 $5 = ((($this)) + 59680|0); //@line 604 "snd_fx.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 604 "snd_fx.cpp"
 $7 = $6 & 1048608; //@line 604 "snd_fx.cpp"
 $8 = ($7|0)==(0); //@line 604 "snd_fx.cpp"
 if (!($8)) {
  $9 = ((($this)) + 59688|0); //@line 604 "snd_fx.cpp"
  $10 = HEAP32[$9>>2]|0; //@line 604 "snd_fx.cpp"
  $11 = ($10|0)==(0); //@line 604 "snd_fx.cpp"
  $brmerge = $11 | $bForceCut; //@line 604 "snd_fx.cpp"
  if (!($brmerge)) {
   $77 = (((($this) + ($nChn<<8)|0)) + 96|0); //@line 627 "snd_fx.cpp"
   $78 = HEAP32[$77>>2]|0; //@line 627 "snd_fx.cpp"
   $79 = (($instr) + -1)|0; //@line 629 "snd_fx.cpp"
   $80 = ($79>>>0)<(239); //@line 629 "snd_fx.cpp"
   if ($80) {
    $81 = (((($this)) + 48640|0) + ($instr<<2)|0); //@line 631 "snd_fx.cpp"
    $82 = HEAP32[$81>>2]|0; //@line 631 "snd_fx.cpp"
    $83 = ($82|0)==(0|0); //@line 632 "snd_fx.cpp"
    if ($83) {
     $$1 = $$note;$pHeader$0 = 0;$pSample$0 = 0;
    } else {
     $84 = ($$note|0)<(129); //@line 635 "snd_fx.cpp"
     if ($84) {
      $85 = (($$note) + -1)|0; //@line 637 "snd_fx.cpp"
      $86 = (((($82)) + 300|0) + ($85)|0); //@line 637 "snd_fx.cpp"
      $87 = HEAP8[$86>>0]|0; //@line 637 "snd_fx.cpp"
      $88 = (((($82)) + 428|0) + ($85)|0); //@line 638 "snd_fx.cpp"
      $89 = HEAP8[$88>>0]|0; //@line 638 "snd_fx.cpp"
      $90 = $89&255; //@line 638 "snd_fx.cpp"
      $$off = (($87) + -1)<<24>>24; //@line 639 "snd_fx.cpp"
      $91 = ($$off&255)<(239); //@line 639 "snd_fx.cpp"
      if ($91) {
       $92 = $87&255; //@line 637 "snd_fx.cpp"
       $93 = (((((($this)) + 33280|0) + ($92<<6)|0)) + 20|0); //@line 639 "snd_fx.cpp"
       $94 = HEAP32[$93>>2]|0; //@line 639 "snd_fx.cpp"
       $$1 = $90;$pHeader$0 = $82;$pSample$0 = $94;
      } else {
       $$1 = $90;$pHeader$0 = $82;$pSample$0 = $78;
      }
     } else {
      $$1 = $$note;$pHeader$0 = $82;$pSample$0 = $78;
     }
    }
   } else {
    $$1 = $$note;$pHeader$0 = $2;$pSample$0 = $78;
   }
   $95 = ($2|0)==(0|0); //@line 643 "snd_fx.cpp"
   if ($95) {
    return; //@line 725 "snd_fx.cpp"
   }
   $96 = ($nChn>>>0)<(128); //@line 645 "snd_fx.cpp"
   L16: do {
    if ($96) {
     $97 = ((($this)) + 59664|0); //@line 646 "snd_fx.cpp"
     $98 = (($nChn) + 1)|0; //@line 648 "snd_fx.cpp"
     $99 = ($$1|0)==(0); //@line 656 "snd_fx.cpp"
     $100 = ($pSample$0|0)==(0|0); //@line 660 "snd_fx.cpp"
     if ($100) {
      $i$046$us = $nChn;$p2$044$us = $0;
      while(1) {
       $101 = HEAP32[$97>>2]|0; //@line 646 "snd_fx.cpp"
       $102 = ($i$046$us>>>0)>=($101>>>0); //@line 646 "snd_fx.cpp"
       $103 = ($p2$044$us|0)==($0|0); //@line 646 "snd_fx.cpp"
       $or$cond5$us = $103 | $102; //@line 646 "snd_fx.cpp"
       L21: do {
        if ($or$cond5$us) {
         $104 = ((($p2$044$us)) + 160|0); //@line 648 "snd_fx.cpp"
         $105 = HEAP32[$104>>2]|0; //@line 648 "snd_fx.cpp"
         $106 = ($105|0)==($98|0); //@line 648 "snd_fx.cpp"
         $or$cond6$us = $103 | $106; //@line 648 "snd_fx.cpp"
         if ($or$cond6$us) {
          $107 = ((($p2$044$us)) + 140|0); //@line 648 "snd_fx.cpp"
          $108 = HEAP32[$107>>2]|0; //@line 648 "snd_fx.cpp"
          $109 = ($108|0)==(0|0); //@line 648 "snd_fx.cpp"
          if (!($109)) {
           $110 = ((($108)) + 572|0); //@line 652 "snd_fx.cpp"
           $111 = HEAP8[$110>>0]|0; //@line 652 "snd_fx.cpp"
           $112 = $111&255; //@line 652 "snd_fx.cpp"
           switch ($112|0) {
           case 1:  {
            if ($99) {
             break L21;
            }
            $114 = ((($p2$044$us)) + 212|0); //@line 656 "snd_fx.cpp"
            $115 = HEAP8[$114>>0]|0; //@line 656 "snd_fx.cpp"
            $116 = $115&255; //@line 656 "snd_fx.cpp"
            $117 = ($116|0)==($$1|0); //@line 656 "snd_fx.cpp"
            $118 = ($pHeader$0|0)==($108|0); //@line 656 "snd_fx.cpp"
            $or$cond7$us = $118 & $117; //@line 656 "snd_fx.cpp"
            if (!($or$cond7$us)) {
             break L21;
            }
            break;
           }
           case 3:  {
            $113 = ($pHeader$0|0)==($108|0); //@line 664 "snd_fx.cpp"
            if (!($113)) {
             break L21;
            }
            break;
           }
           default: {
            break L21;
           }
           }
           $119 = ((($108)) + 573|0); //@line 670 "snd_fx.cpp"
           $120 = HEAP8[$119>>0]|0; //@line 670 "snd_fx.cpp"
           $121 = $120&255; //@line 670 "snd_fx.cpp"
           switch ($121|0) {
           case 0:  {
            __ZN10CSoundFile6KeyOffEj($this,$i$046$us); //@line 674 "snd_fx.cpp"
            $125 = ((($p2$044$us)) + 116|0); //@line 675 "snd_fx.cpp"
            HEAP32[$125>>2] = 0; //@line 675 "snd_fx.cpp"
            break;
           }
           case 1:  {
            __ZN10CSoundFile6KeyOffEj($this,$i$046$us); //@line 679 "snd_fx.cpp"
            break;
           }
           case 2:  {
            $122 = ((($p2$044$us)) + 36|0); //@line 683 "snd_fx.cpp"
            $123 = HEAP32[$122>>2]|0; //@line 683 "snd_fx.cpp"
            $124 = $123 | 1024; //@line 683 "snd_fx.cpp"
            HEAP32[$122>>2] = $124; //@line 683 "snd_fx.cpp"
            break;
           }
           default: {
           }
           }
           $126 = ((($p2$044$us)) + 116|0); //@line 686 "snd_fx.cpp"
           $127 = HEAP32[$126>>2]|0; //@line 686 "snd_fx.cpp"
           $128 = ($127|0)==(0); //@line 686 "snd_fx.cpp"
           if ($128) {
            $129 = ((($p2$044$us)) + 124|0); //@line 688 "snd_fx.cpp"
            HEAP32[$129>>2] = 0; //@line 688 "snd_fx.cpp"
            $130 = ((($p2$044$us)) + 36|0); //@line 689 "snd_fx.cpp"
            $131 = HEAP32[$130>>2]|0; //@line 689 "snd_fx.cpp"
            $132 = $131 | 16778240; //@line 689 "snd_fx.cpp"
            HEAP32[$130>>2] = $132; //@line 689 "snd_fx.cpp"
           }
          }
         }
        }
       } while(0);
       $133 = ((($p2$044$us)) + 256|0); //@line 645 "snd_fx.cpp"
       $134 = (($i$046$us) + 1)|0; //@line 645 "snd_fx.cpp"
       $exitcond50 = ($134|0)==(128); //@line 645 "snd_fx.cpp"
       if ($exitcond50) {
        break L16;
       } else {
        $i$046$us = $134;$p2$044$us = $133;
       }
      }
     } else {
      $i$046 = $nChn;$p2$044 = $0;
     }
     while(1) {
      $135 = HEAP32[$97>>2]|0; //@line 646 "snd_fx.cpp"
      $136 = ($i$046>>>0)>=($135>>>0); //@line 646 "snd_fx.cpp"
      $137 = ($p2$044|0)==($0|0); //@line 646 "snd_fx.cpp"
      $or$cond5 = $137 | $136; //@line 646 "snd_fx.cpp"
      L39: do {
       if ($or$cond5) {
        $138 = ((($p2$044)) + 160|0); //@line 648 "snd_fx.cpp"
        $139 = HEAP32[$138>>2]|0; //@line 648 "snd_fx.cpp"
        $140 = ($139|0)==($98|0); //@line 648 "snd_fx.cpp"
        $or$cond6 = $137 | $140; //@line 648 "snd_fx.cpp"
        if ($or$cond6) {
         $141 = ((($p2$044)) + 140|0); //@line 648 "snd_fx.cpp"
         $142 = HEAP32[$141>>2]|0; //@line 648 "snd_fx.cpp"
         $143 = ($142|0)==(0|0); //@line 648 "snd_fx.cpp"
         if (!($143)) {
          $144 = ((($142)) + 572|0); //@line 652 "snd_fx.cpp"
          $145 = HEAP8[$144>>0]|0; //@line 652 "snd_fx.cpp"
          $146 = $145&255; //@line 652 "snd_fx.cpp"
          switch ($146|0) {
          case 1:  {
           if ($99) {
            break L39;
           }
           $147 = ((($p2$044)) + 212|0); //@line 656 "snd_fx.cpp"
           $148 = HEAP8[$147>>0]|0; //@line 656 "snd_fx.cpp"
           $149 = $148&255; //@line 656 "snd_fx.cpp"
           $150 = ($149|0)==($$1|0); //@line 656 "snd_fx.cpp"
           $151 = ($pHeader$0|0)==($142|0); //@line 656 "snd_fx.cpp"
           $or$cond7 = $151 & $150; //@line 656 "snd_fx.cpp"
           if (!($or$cond7)) {
            break L39;
           }
           break;
          }
          case 2:  {
           $153 = ((($p2$044)) + 96|0); //@line 660 "snd_fx.cpp"
           $154 = HEAP32[$153>>2]|0; //@line 660 "snd_fx.cpp"
           $155 = ($pSample$0|0)==($154|0); //@line 660 "snd_fx.cpp"
           if (!($155)) {
            break L39;
           }
           break;
          }
          case 3:  {
           $152 = ($pHeader$0|0)==($142|0); //@line 664 "snd_fx.cpp"
           if (!($152)) {
            break L39;
           }
           break;
          }
          default: {
           break L39;
          }
          }
          $156 = ((($142)) + 573|0); //@line 670 "snd_fx.cpp"
          $157 = HEAP8[$156>>0]|0; //@line 670 "snd_fx.cpp"
          $158 = $157&255; //@line 670 "snd_fx.cpp"
          switch ($158|0) {
          case 0:  {
           __ZN10CSoundFile6KeyOffEj($this,$i$046); //@line 674 "snd_fx.cpp"
           $159 = ((($p2$044)) + 116|0); //@line 675 "snd_fx.cpp"
           HEAP32[$159>>2] = 0; //@line 675 "snd_fx.cpp"
           break;
          }
          case 1:  {
           __ZN10CSoundFile6KeyOffEj($this,$i$046); //@line 679 "snd_fx.cpp"
           break;
          }
          case 2:  {
           $160 = ((($p2$044)) + 36|0); //@line 683 "snd_fx.cpp"
           $161 = HEAP32[$160>>2]|0; //@line 683 "snd_fx.cpp"
           $162 = $161 | 1024; //@line 683 "snd_fx.cpp"
           HEAP32[$160>>2] = $162; //@line 683 "snd_fx.cpp"
           break;
          }
          default: {
          }
          }
          $163 = ((($p2$044)) + 116|0); //@line 686 "snd_fx.cpp"
          $164 = HEAP32[$163>>2]|0; //@line 686 "snd_fx.cpp"
          $165 = ($164|0)==(0); //@line 686 "snd_fx.cpp"
          if ($165) {
           $166 = ((($p2$044)) + 124|0); //@line 688 "snd_fx.cpp"
           HEAP32[$166>>2] = 0; //@line 688 "snd_fx.cpp"
           $167 = ((($p2$044)) + 36|0); //@line 689 "snd_fx.cpp"
           $168 = HEAP32[$167>>2]|0; //@line 689 "snd_fx.cpp"
           $169 = $168 | 16778240; //@line 689 "snd_fx.cpp"
           HEAP32[$167>>2] = $169; //@line 689 "snd_fx.cpp"
          }
         }
        }
       }
      } while(0);
      $170 = ((($p2$044)) + 256|0); //@line 645 "snd_fx.cpp"
      $171 = (($i$046) + 1)|0; //@line 645 "snd_fx.cpp"
      $exitcond = ($171|0)==(128); //@line 645 "snd_fx.cpp"
      if ($exitcond) {
       break;
      } else {
       $i$046 = $171;$p2$044 = $170;
      }
     }
    }
   } while(0);
   $172 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 694 "snd_fx.cpp"
   $173 = HEAP32[$172>>2]|0; //@line 694 "snd_fx.cpp"
   $174 = $173 & 256; //@line 694 "snd_fx.cpp"
   $175 = ($174|0)==(0); //@line 694 "snd_fx.cpp"
   if (!($175)) {
    return; //@line 725 "snd_fx.cpp"
   }
   $176 = (((($this) + ($nChn<<8)|0)) + 116|0); //@line 696 "snd_fx.cpp"
   $177 = HEAP32[$176>>2]|0; //@line 696 "snd_fx.cpp"
   $178 = ($177|0)==(0); //@line 696 "snd_fx.cpp"
   if ($178) {
    return; //@line 725 "snd_fx.cpp"
   }
   $179 = (((($this) + ($nChn<<8)|0)) + 32|0); //@line 696 "snd_fx.cpp"
   $180 = HEAP32[$179>>2]|0; //@line 696 "snd_fx.cpp"
   $181 = ($180|0)==(0); //@line 696 "snd_fx.cpp"
   if ($181) {
    return; //@line 725 "snd_fx.cpp"
   }
   $182 = ((($this)) + 59664|0); //@line 567 "snd_fx.cpp"
   $183 = HEAP32[$182>>2]|0; //@line 567 "snd_fx.cpp"
   $184 = ($183>>>0)<(128); //@line 568 "snd_fx.cpp"
   if (!($184)) {
    return; //@line 725 "snd_fx.cpp"
   }
   $185 = (($this) + ($183<<8)|0); //@line 567 "snd_fx.cpp"
   $i$011$i14 = $183;$pi$010$i15 = $185;
   while(1) {
    $186 = ((($pi$010$i15)) + 32|0); //@line 568 "snd_fx.cpp"
    $187 = HEAP32[$186>>2]|0; //@line 568 "snd_fx.cpp"
    $188 = ($187|0)==(0); //@line 568 "snd_fx.cpp"
    if ($188) {
     $$0$i34 = $i$011$i14;
     break;
    }
    $189 = (($i$011$i14) + 1)|0; //@line 568 "snd_fx.cpp"
    $190 = ((($pi$010$i15)) + 256|0); //@line 568 "snd_fx.cpp"
    $191 = ($189>>>0)<(128); //@line 568 "snd_fx.cpp"
    if ($191) {
     $i$011$i14 = $189;$pi$010$i15 = $190;
    } else {
     label = 67;
     break;
    }
   }
   L72: do {
    if ((label|0) == 67) {
     $192 = (((($this) + ($nChn<<8)|0)) + 124|0); //@line 569 "snd_fx.cpp"
     $193 = HEAP32[$192>>2]|0; //@line 569 "snd_fx.cpp"
     $194 = ($193|0)==(0); //@line 569 "snd_fx.cpp"
     if ($194) {
      return; //@line 725 "snd_fx.cpp"
     } else {
      $envpos$05$i22 = 16777215;$j$07$i20 = $183;$pj$06$i21 = $185;$result$03$i24 = 0;$vol$04$i23 = 4194304;
     }
     while(1) {
      $195 = ((($pj$06$i21)) + 124|0); //@line 577 "snd_fx.cpp"
      $196 = HEAP32[$195>>2]|0; //@line 577 "snd_fx.cpp"
      $197 = ($196|0)==(0); //@line 577 "snd_fx.cpp"
      if ($197) {
       $$0$i34 = $j$07$i20;
       break L72;
      }
      $198 = ((($pj$06$i21)) + 116|0); //@line 578 "snd_fx.cpp"
      $199 = HEAP32[$198>>2]|0; //@line 578 "snd_fx.cpp"
      $200 = ((($pj$06$i21)) + 36|0); //@line 579 "snd_fx.cpp"
      $201 = HEAP32[$200>>2]|0; //@line 579 "snd_fx.cpp"
      $202 = $201 & 1024; //@line 579 "snd_fx.cpp"
      $203 = ($202|0)==(0); //@line 579 "snd_fx.cpp"
      $204 = Math_imul($199, $196)|0; //@line 580 "snd_fx.cpp"
      $205 = $199 << 16; //@line 582 "snd_fx.cpp"
      $v$0$i25 = $203 ? $205 : $204; //@line 579 "snd_fx.cpp"
      $206 = $201 >>> 1; //@line 583 "snd_fx.cpp"
      $$lobit$i26 = $206 & 1; //@line 583 "snd_fx.cpp"
      $v$0$$i27 = $v$0$i25 >>> $$lobit$i26; //@line 583 "snd_fx.cpp"
      $207 = ($v$0$$i27>>>0)<($vol$04$i23>>>0); //@line 584 "snd_fx.cpp"
      if ($207) {
       $$phi$trans$insert$i28 = ((($pj$06$i21)) + 148|0); //@line 698 "snd_fx.cpp"
       $$pre$i29 = HEAP32[$$phi$trans$insert$i28>>2]|0; //@line 586 "snd_fx.cpp"
       $239 = $$pre$i29;
       label = 73;
      } else {
       $208 = ($v$0$$i27|0)==($vol$04$i23|0); //@line 584 "snd_fx.cpp"
       if ($208) {
        $209 = ((($pj$06$i21)) + 148|0); //@line 584 "snd_fx.cpp"
        $210 = HEAP32[$209>>2]|0; //@line 584 "snd_fx.cpp"
        $211 = ($210>>>0)>($envpos$05$i22>>>0); //@line 584 "snd_fx.cpp"
        if ($211) {
         $239 = $210;
         label = 73;
        } else {
         $envpos$1$i33 = $envpos$05$i22;$result$1$i31 = $result$03$i24;$vol$1$i32 = $vol$04$i23;
        }
       } else {
        $envpos$1$i33 = $envpos$05$i22;$result$1$i31 = $result$03$i24;$vol$1$i32 = $vol$04$i23;
       }
      }
      if ((label|0) == 73) {
       label = 0;
       $envpos$1$i33 = $239;$result$1$i31 = $j$07$i20;$vol$1$i32 = $v$0$$i27;
      }
      $212 = (($j$07$i20) + 1)|0; //@line 575 "snd_fx.cpp"
      $213 = ((($pj$06$i21)) + 256|0); //@line 575 "snd_fx.cpp"
      $214 = ($212>>>0)<(128); //@line 575 "snd_fx.cpp"
      if ($214) {
       $envpos$05$i22 = $envpos$1$i33;$j$07$i20 = $212;$pj$06$i21 = $213;$result$03$i24 = $result$1$i31;$vol$04$i23 = $vol$1$i32;
      } else {
       $$0$i34 = $result$1$i31;
       break;
      }
     }
    }
   } while(0);
   $215 = ($$0$i34|0)==(0); //@line 699 "snd_fx.cpp"
   if ($215) {
    return; //@line 725 "snd_fx.cpp"
   }
   $216 = (($this) + ($$0$i34<<8)|0); //@line 701 "snd_fx.cpp"
   _memcpy(($216|0),($0|0),256)|0; //@line 703 "snd_fx.cpp"
   $217 = (((($this) + ($$0$i34<<8)|0)) + 36|0); //@line 704 "snd_fx.cpp"
   $218 = HEAP32[$217>>2]|0; //@line 704 "snd_fx.cpp"
   $219 = $218 & -983297; //@line 704 "snd_fx.cpp"
   HEAP32[$217>>2] = $219; //@line 704 "snd_fx.cpp"
   $220 = (($nChn) + 1)|0; //@line 705 "snd_fx.cpp"
   $221 = (((($this) + ($$0$i34<<8)|0)) + 160|0); //@line 705 "snd_fx.cpp"
   HEAP32[$221>>2] = $220; //@line 705 "snd_fx.cpp"
   $222 = (((($this) + ($$0$i34<<8)|0)) + 216|0); //@line 706 "snd_fx.cpp"
   HEAP8[$222>>0] = 0; //@line 706 "snd_fx.cpp"
   $223 = (((($this) + ($nChn<<8)|0)) + 213|0); //@line 708 "snd_fx.cpp"
   $224 = HEAP8[$223>>0]|0; //@line 708 "snd_fx.cpp"
   $225 = $224&255; //@line 708 "snd_fx.cpp"
   switch ($225|0) {
   case 2:  {
    __ZN10CSoundFile6KeyOffEj($this,$$0$i34); //@line 710 "snd_fx.cpp"
    break;
   }
   case 0:  {
    $226 = (((($this) + ($$0$i34<<8)|0)) + 124|0); //@line 712 "snd_fx.cpp"
    HEAP32[$226>>2] = 0; //@line 712 "snd_fx.cpp"
    label = 79;
    break;
   }
   case 3:  {
    label = 79;
    break;
   }
   default: {
   }
   }
   if ((label|0) == 79) {
    $227 = $219 | 1024; //@line 713 "snd_fx.cpp"
    HEAP32[$217>>2] = $227; //@line 713 "snd_fx.cpp"
   }
   $228 = (((($this) + ($$0$i34<<8)|0)) + 116|0); //@line 715 "snd_fx.cpp"
   $229 = HEAP32[$228>>2]|0; //@line 715 "snd_fx.cpp"
   $230 = ($229|0)==(0); //@line 715 "snd_fx.cpp"
   if ($230) {
    $231 = (((($this) + ($$0$i34<<8)|0)) + 124|0); //@line 717 "snd_fx.cpp"
    HEAP32[$231>>2] = 0; //@line 717 "snd_fx.cpp"
    $232 = HEAP32[$217>>2]|0; //@line 718 "snd_fx.cpp"
    $233 = $232 | 16778240; //@line 718 "snd_fx.cpp"
    HEAP32[$217>>2] = $233; //@line 718 "snd_fx.cpp"
   }
   $234 = (((($this) + ($nChn<<8)|0)) + 8|0); //@line 721 "snd_fx.cpp"
   HEAP32[$234>>2] = 0; //@line 721 "snd_fx.cpp"
   $235 = (((($this) + ($nChn<<8)|0)) + 4|0); //@line 721 "snd_fx.cpp"
   HEAP32[$235>>2] = 0; //@line 721 "snd_fx.cpp"
   HEAP32[$179>>2] = 0; //@line 721 "snd_fx.cpp"
   $236 = (((($this) + ($nChn<<8)|0)) + 88|0); //@line 722 "snd_fx.cpp"
   HEAP32[$236>>2] = 0; //@line 722 "snd_fx.cpp"
   $237 = (((($this) + ($nChn<<8)|0)) + 84|0); //@line 722 "snd_fx.cpp"
   HEAP32[$237>>2] = 0; //@line 722 "snd_fx.cpp"
   return; //@line 725 "snd_fx.cpp"
  }
 }
 $12 = ((($this)) + 59660|0); //@line 606 "snd_fx.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 606 "snd_fx.cpp"
 $14 = $13 & 2048; //@line 606 "snd_fx.cpp"
 $15 = ($14|0)==(0); //@line 606 "snd_fx.cpp"
 if (!($15)) {
  return; //@line 725 "snd_fx.cpp"
 }
 $16 = (((($this) + ($nChn<<8)|0)) + 32|0); //@line 607 "snd_fx.cpp"
 $17 = HEAP32[$16>>2]|0; //@line 607 "snd_fx.cpp"
 $18 = ($17|0)==(0); //@line 607 "snd_fx.cpp"
 if ($18) {
  return; //@line 725 "snd_fx.cpp"
 }
 $19 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 607 "snd_fx.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 607 "snd_fx.cpp"
 $21 = $20 & 256; //@line 607 "snd_fx.cpp"
 $22 = ($21|0)==(0); //@line 607 "snd_fx.cpp"
 if (!($22)) {
  return; //@line 725 "snd_fx.cpp"
 }
 $23 = (((($this) + ($nChn<<8)|0)) + 20|0); //@line 608 "snd_fx.cpp"
 $24 = HEAP32[$23>>2]|0; //@line 608 "snd_fx.cpp"
 $25 = ($24|0)==(0); //@line 608 "snd_fx.cpp"
 if ($25) {
  $26 = (((($this) + ($nChn<<8)|0)) + 16|0); //@line 608 "snd_fx.cpp"
  $27 = HEAP32[$26>>2]|0; //@line 608 "snd_fx.cpp"
  $28 = ($27|0)==(0); //@line 608 "snd_fx.cpp"
  if ($28) {
   return; //@line 725 "snd_fx.cpp"
  }
 }
 $29 = ((($this)) + 59664|0); //@line 567 "snd_fx.cpp"
 $30 = HEAP32[$29>>2]|0; //@line 567 "snd_fx.cpp"
 $31 = ($30>>>0)<(128); //@line 568 "snd_fx.cpp"
 if (!($31)) {
  return; //@line 725 "snd_fx.cpp"
 }
 $32 = (($this) + ($30<<8)|0); //@line 567 "snd_fx.cpp"
 $i$011$i = $30;$pi$010$i = $32;
 while(1) {
  $33 = ((($pi$010$i)) + 32|0); //@line 568 "snd_fx.cpp"
  $34 = HEAP32[$33>>2]|0; //@line 568 "snd_fx.cpp"
  $35 = ($34|0)==(0); //@line 568 "snd_fx.cpp"
  if ($35) {
   $$0$i = $i$011$i;
   break;
  }
  $36 = (($i$011$i) + 1)|0; //@line 568 "snd_fx.cpp"
  $37 = ((($pi$010$i)) + 256|0); //@line 568 "snd_fx.cpp"
  $38 = ($36>>>0)<(128); //@line 568 "snd_fx.cpp"
  if ($38) {
   $i$011$i = $36;$pi$010$i = $37;
  } else {
   label = 13;
   break;
  }
 }
 L120: do {
  if ((label|0) == 13) {
   $39 = (((($this) + ($nChn<<8)|0)) + 124|0); //@line 569 "snd_fx.cpp"
   $40 = HEAP32[$39>>2]|0; //@line 569 "snd_fx.cpp"
   $41 = ($40|0)==(0); //@line 569 "snd_fx.cpp"
   if ($41) {
    return; //@line 725 "snd_fx.cpp"
   } else {
    $envpos$05$i = 16777215;$j$07$i = $30;$pj$06$i = $32;$result$03$i = 0;$vol$04$i = 4194304;
   }
   while(1) {
    $42 = ((($pj$06$i)) + 124|0); //@line 577 "snd_fx.cpp"
    $43 = HEAP32[$42>>2]|0; //@line 577 "snd_fx.cpp"
    $44 = ($43|0)==(0); //@line 577 "snd_fx.cpp"
    if ($44) {
     $$0$i = $j$07$i;
     break L120;
    }
    $45 = ((($pj$06$i)) + 116|0); //@line 578 "snd_fx.cpp"
    $46 = HEAP32[$45>>2]|0; //@line 578 "snd_fx.cpp"
    $47 = ((($pj$06$i)) + 36|0); //@line 579 "snd_fx.cpp"
    $48 = HEAP32[$47>>2]|0; //@line 579 "snd_fx.cpp"
    $49 = $48 & 1024; //@line 579 "snd_fx.cpp"
    $50 = ($49|0)==(0); //@line 579 "snd_fx.cpp"
    $51 = Math_imul($46, $43)|0; //@line 580 "snd_fx.cpp"
    $52 = $46 << 16; //@line 582 "snd_fx.cpp"
    $v$0$i = $50 ? $52 : $51; //@line 579 "snd_fx.cpp"
    $53 = $48 >>> 1; //@line 583 "snd_fx.cpp"
    $$lobit$i = $53 & 1; //@line 583 "snd_fx.cpp"
    $v$0$$i = $v$0$i >>> $$lobit$i; //@line 583 "snd_fx.cpp"
    $54 = ($v$0$$i>>>0)<($vol$04$i>>>0); //@line 584 "snd_fx.cpp"
    if ($54) {
     $$phi$trans$insert$i = ((($pj$06$i)) + 148|0); //@line 609 "snd_fx.cpp"
     $$pre$i = HEAP32[$$phi$trans$insert$i>>2]|0; //@line 586 "snd_fx.cpp"
     $238 = $$pre$i;
     label = 19;
    } else {
     $55 = ($v$0$$i|0)==($vol$04$i|0); //@line 584 "snd_fx.cpp"
     if ($55) {
      $56 = ((($pj$06$i)) + 148|0); //@line 584 "snd_fx.cpp"
      $57 = HEAP32[$56>>2]|0; //@line 584 "snd_fx.cpp"
      $58 = ($57>>>0)>($envpos$05$i>>>0); //@line 584 "snd_fx.cpp"
      if ($58) {
       $238 = $57;
       label = 19;
      } else {
       $envpos$1$i = $envpos$05$i;$result$1$i = $result$03$i;$vol$1$i = $vol$04$i;
      }
     } else {
      $envpos$1$i = $envpos$05$i;$result$1$i = $result$03$i;$vol$1$i = $vol$04$i;
     }
    }
    if ((label|0) == 19) {
     label = 0;
     $envpos$1$i = $238;$result$1$i = $j$07$i;$vol$1$i = $v$0$$i;
    }
    $59 = (($j$07$i) + 1)|0; //@line 575 "snd_fx.cpp"
    $60 = ((($pj$06$i)) + 256|0); //@line 575 "snd_fx.cpp"
    $61 = ($59>>>0)<(128); //@line 575 "snd_fx.cpp"
    if ($61) {
     $envpos$05$i = $envpos$1$i;$j$07$i = $59;$pj$06$i = $60;$result$03$i = $result$1$i;$vol$04$i = $vol$1$i;
    } else {
     $$0$i = $result$1$i;
     break;
    }
   }
  }
 } while(0);
 $62 = ($$0$i|0)==(0); //@line 610 "snd_fx.cpp"
 if ($62) {
  return; //@line 725 "snd_fx.cpp"
 }
 $63 = (($this) + ($$0$i<<8)|0); //@line 611 "snd_fx.cpp"
 _memcpy(($63|0),($0|0),256)|0; //@line 613 "snd_fx.cpp"
 $64 = (((($this) + ($$0$i<<8)|0)) + 36|0); //@line 614 "snd_fx.cpp"
 $65 = HEAP32[$64>>2]|0; //@line 614 "snd_fx.cpp"
 $66 = $65 & -17761537; //@line 614 "snd_fx.cpp"
 $67 = (($nChn) + 1)|0; //@line 615 "snd_fx.cpp"
 $68 = (((($this) + ($$0$i<<8)|0)) + 160|0); //@line 615 "snd_fx.cpp"
 HEAP32[$68>>2] = $67; //@line 615 "snd_fx.cpp"
 $69 = (((($this) + ($$0$i<<8)|0)) + 216|0); //@line 616 "snd_fx.cpp"
 HEAP8[$69>>0] = 0; //@line 616 "snd_fx.cpp"
 $70 = (((($this) + ($$0$i<<8)|0)) + 124|0); //@line 618 "snd_fx.cpp"
 HEAP32[$70>>2] = 0; //@line 618 "snd_fx.cpp"
 $71 = $66 | 16778240; //@line 619 "snd_fx.cpp"
 HEAP32[$64>>2] = $71; //@line 619 "snd_fx.cpp"
 $72 = (((($this) + ($nChn<<8)|0)) + 8|0); //@line 621 "snd_fx.cpp"
 HEAP32[$72>>2] = 0; //@line 621 "snd_fx.cpp"
 $73 = (((($this) + ($nChn<<8)|0)) + 4|0); //@line 621 "snd_fx.cpp"
 HEAP32[$73>>2] = 0; //@line 621 "snd_fx.cpp"
 HEAP32[$16>>2] = 0; //@line 621 "snd_fx.cpp"
 $74 = (((($this) + ($nChn<<8)|0)) + 88|0); //@line 622 "snd_fx.cpp"
 HEAP32[$74>>2] = 0; //@line 622 "snd_fx.cpp"
 $75 = (((($this) + ($nChn<<8)|0)) + 84|0); //@line 622 "snd_fx.cpp"
 HEAP32[$75>>2] = 0; //@line 622 "snd_fx.cpp"
 $76 = (((($this) + ($nChn<<8)|0)) + 16|0); //@line 623 "snd_fx.cpp"
 HEAP32[$76>>2] = 0; //@line 623 "snd_fx.cpp"
 HEAP32[$23>>2] = 0; //@line 623 "snd_fx.cpp"
 return; //@line 725 "snd_fx.cpp"
}
function __ZN10CSoundFile6KeyOffEj($this,$nChn) {
 $this = $this|0;
 $nChn = $nChn|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $brmerge = 0, $not$$not = 0, $or$cond = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 2024 "snd_fx.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 2024 "snd_fx.cpp"
 $2 = $1 & 512; //@line 2024 "snd_fx.cpp"
 $3 = $1 | 512; //@line 2025 "snd_fx.cpp"
 HEAP32[$0>>2] = $3; //@line 2025 "snd_fx.cpp"
 $4 = (((($this) + ($nChn<<8)|0)) + 140|0); //@line 2027 "snd_fx.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 2027 "snd_fx.cpp"
 $6 = ($5|0)!=(0|0); //@line 2027 "snd_fx.cpp"
 $7 = $1 & 2097152;
 $8 = ($7|0)==(0); //@line 2027 "snd_fx.cpp"
 $or$cond = $6 & $8; //@line 2027 "snd_fx.cpp"
 if ($or$cond) {
  $9 = $1 | 1536; //@line 2029 "snd_fx.cpp"
  HEAP32[$0>>2] = $9; //@line 2029 "snd_fx.cpp"
  $14 = $9;
 } else {
  $14 = $3;
 }
 $10 = (((($this) + ($nChn<<8)|0)) + 32|0); //@line 2031 "snd_fx.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 2031 "snd_fx.cpp"
 $12 = ($11|0)==(0); //@line 2031 "snd_fx.cpp"
 if ($12) {
  return; //@line 2058 "snd_fx.cpp"
 }
 $13 = $14 & 8; //@line 2032 "snd_fx.cpp"
 $15 = ($13|0)==(0); //@line 2032 "snd_fx.cpp"
 do {
  if ($15) {
   $51 = $14;
  } else {
   $16 = (((($this) + ($nChn<<8)|0)) + 144|0); //@line 2032 "snd_fx.cpp"
   $17 = HEAP32[$16>>2]|0; //@line 2032 "snd_fx.cpp"
   $18 = ($17|0)==(0|0); //@line 2032 "snd_fx.cpp"
   $not$$not = ($2|0)!=(0); //@line 2032 "snd_fx.cpp"
   $brmerge = $not$$not | $18; //@line 2032 "snd_fx.cpp"
   if ($brmerge) {
    $51 = $14;
   } else {
    $19 = ((($17)) + 34|0); //@line 2035 "snd_fx.cpp"
    $20 = HEAP16[$19>>1]|0; //@line 2035 "snd_fx.cpp"
    $21 = $20&65535; //@line 2035 "snd_fx.cpp"
    $22 = $21 & 2; //@line 2035 "snd_fx.cpp"
    $23 = ($22|0)==(0); //@line 2035 "snd_fx.cpp"
    if ($23) {
     $37 = $14 & -135; //@line 2048 "snd_fx.cpp"
     HEAP32[$0>>2] = $37; //@line 2048 "snd_fx.cpp"
     $38 = HEAP32[$17>>2]|0; //@line 2049 "snd_fx.cpp"
     HEAP32[$10>>2] = $38; //@line 2049 "snd_fx.cpp"
     $51 = $37;
     break;
    }
    $24 = $21 & 4; //@line 2037 "snd_fx.cpp"
    $25 = ($24|0)==(0); //@line 2037 "snd_fx.cpp"
    $26 = $14 & -135; //@line 2040 "snd_fx.cpp"
    $27 = $14 | 4; //@line 2038 "snd_fx.cpp"
    $storemerge = $25 ? $26 : $27; //@line 2037 "snd_fx.cpp"
    $28 = $storemerge | 2; //@line 2041 "snd_fx.cpp"
    HEAP32[$0>>2] = $28; //@line 2041 "snd_fx.cpp"
    $29 = HEAP32[$17>>2]|0; //@line 2042 "snd_fx.cpp"
    HEAP32[$10>>2] = $29; //@line 2042 "snd_fx.cpp"
    $30 = ((($17)) + 4|0); //@line 2043 "snd_fx.cpp"
    $31 = HEAP32[$30>>2]|0; //@line 2043 "snd_fx.cpp"
    $32 = (((($this) + ($nChn<<8)|0)) + 40|0); //@line 2043 "snd_fx.cpp"
    HEAP32[$32>>2] = $31; //@line 2043 "snd_fx.cpp"
    $33 = ((($17)) + 8|0); //@line 2044 "snd_fx.cpp"
    $34 = HEAP32[$33>>2]|0; //@line 2044 "snd_fx.cpp"
    $35 = (((($this) + ($nChn<<8)|0)) + 44|0); //@line 2044 "snd_fx.cpp"
    HEAP32[$35>>2] = $34; //@line 2044 "snd_fx.cpp"
    $36 = ($29>>>0)>($34>>>0); //@line 2045 "snd_fx.cpp"
    if ($36) {
     HEAP32[$10>>2] = $34; //@line 2045 "snd_fx.cpp"
     $51 = $28;
    } else {
     $51 = $28;
    }
   }
  }
 } while(0);
 $39 = ($5|0)==(0|0); //@line 2052 "snd_fx.cpp"
 if ($39) {
  return; //@line 2058 "snd_fx.cpp"
 }
 $40 = ((($5)) + 4|0); //@line 2055 "snd_fx.cpp"
 $41 = HEAP32[$40>>2]|0; //@line 2055 "snd_fx.cpp"
 $42 = $41 & 4; //@line 2055 "snd_fx.cpp"
 $43 = ($42|0)==(0); //@line 2055 "snd_fx.cpp"
 if ($43) {
  $44 = ((($this)) + 59680|0); //@line 2055 "snd_fx.cpp"
  $45 = HEAP32[$44>>2]|0; //@line 2055 "snd_fx.cpp"
  $46 = $45 & 1048580; //@line 2055 "snd_fx.cpp"
  $47 = ($46|0)==(0); //@line 2055 "snd_fx.cpp"
  if ($47) {
   return; //@line 2058 "snd_fx.cpp"
  }
 }
 $48 = HEAP32[$5>>2]|0; //@line 2055 "snd_fx.cpp"
 $49 = ($48|0)==(0); //@line 2055 "snd_fx.cpp"
 if ($49) {
  return; //@line 2058 "snd_fx.cpp"
 }
 $50 = $51 | 1024; //@line 2056 "snd_fx.cpp"
 HEAP32[$0>>2] = $50; //@line 2056 "snd_fx.cpp"
 return; //@line 2058 "snd_fx.cpp"
}
function __ZN10CSoundFile16InstrumentChangeEP11_MODCHANNELjbbb($this,$pChn,$instr,$bPorta,$bUpdVol,$bResetEnv) {
 $this = $this|0;
 $pChn = $pChn|0;
 $instr = $instr|0;
 $bPorta = $bPorta|0;
 $bUpdVol = $bUpdVol|0;
 $bResetEnv = $bResetEnv|0;
 var $$ = 0, $$$ = 0, $$off = 0, $$off10 = 0, $$pre$phi17Z2D = 0, $$pre$phi18Z2D = 0, $$pre16 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0;
 var $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0;
 var $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0;
 var $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bInstrumentChanged$0$off0 = 0, $bPorta$not = 0, $or$cond = 0, $or$cond11 = 0, $or$cond13 = 0, $or$cond20 = 0, $or$cond21 = 0, $or$cond7 = 0, $or$cond9 = 0;
 var $psmp$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($instr>>>0)>(239); //@line 285 "snd_fx.cpp"
 if ($0) {
  return; //@line 408 "snd_fx.cpp"
 }
 $1 = (((($this)) + 48640|0) + ($instr<<2)|0); //@line 286 "snd_fx.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 286 "snd_fx.cpp"
 $3 = ((($pChn)) + 214|0); //@line 288 "snd_fx.cpp"
 $4 = HEAP8[$3>>0]|0; //@line 288 "snd_fx.cpp"
 $5 = ($2|0)!=(0|0); //@line 289 "snd_fx.cpp"
 $$off = (($4) + -1)<<24>>24; //@line 289 "snd_fx.cpp"
 $6 = ($$off<<24>>24)>(-1); //@line 289 "snd_fx.cpp"
 $7 = $5 & $6; //@line 289 "snd_fx.cpp"
 do {
  if ($7) {
   $8 = $4&255; //@line 288 "snd_fx.cpp"
   $9 = (($8) + -1)|0; //@line 291 "snd_fx.cpp"
   $10 = (((($2)) + 428|0) + ($9)|0); //@line 291 "snd_fx.cpp"
   $11 = HEAP8[$10>>0]|0; //@line 291 "snd_fx.cpp"
   $12 = ($11&255)>(253); //@line 291 "snd_fx.cpp"
   if ($12) {
    return; //@line 408 "snd_fx.cpp"
   } else {
    $13 = (((($2)) + 300|0) + ($9)|0); //@line 292 "snd_fx.cpp"
    $14 = HEAP8[$13>>0]|0; //@line 292 "snd_fx.cpp"
    $15 = $14&255; //@line 292 "snd_fx.cpp"
    $$off10 = (($14) + -1)<<24>>24; //@line 293 "snd_fx.cpp"
    $16 = ($$off10&255)<(239); //@line 293 "snd_fx.cpp"
    $17 = (((($this)) + 33280|0) + ($15<<6)|0); //@line 293 "snd_fx.cpp"
    $18 = $16 ? $17 : 0; //@line 293 "snd_fx.cpp"
    $psmp$0 = $18;
    break;
   }
  } else {
   $19 = (((($this)) + 33280|0) + ($instr<<6)|0); //@line 287 "snd_fx.cpp"
   $20 = ((($this)) + 59688|0); //@line 295 "snd_fx.cpp"
   $21 = HEAP32[$20>>2]|0; //@line 295 "snd_fx.cpp"
   $22 = ($21|0)==(0); //@line 295 "snd_fx.cpp"
   if ($22) {
    $psmp$0 = $19;
   } else {
    $23 = ($4&255)>(253); //@line 297 "snd_fx.cpp"
    if ($23) {
     return; //@line 408 "snd_fx.cpp"
    } else {
     $psmp$0 = 0;
    }
   }
  }
 } while(0);
 if ($bUpdVol) {
  $24 = ($psmp$0|0)==(0|0); //@line 301 "snd_fx.cpp"
  if ($24) {
   $29 = 0;
  } else {
   $25 = ((($psmp$0)) + 30|0); //@line 301 "snd_fx.cpp"
   $26 = HEAP16[$25>>1]|0; //@line 301 "snd_fx.cpp"
   $27 = $26&65535; //@line 301 "snd_fx.cpp"
   $29 = $27;
  }
  $28 = ((($pChn)) + 116|0); //@line 301 "snd_fx.cpp"
  HEAP32[$28>>2] = $29; //@line 301 "snd_fx.cpp"
 }
 $30 = ((($pChn)) + 140|0); //@line 303 "snd_fx.cpp"
 $31 = HEAP32[$30>>2]|0; //@line 303 "snd_fx.cpp"
 $32 = ($2|0)==($31|0); //@line 303 "snd_fx.cpp"
 if ($32) {
  if ($bPorta) {
   $33 = ((($this)) + 59680|0); //@line 310 "snd_fx.cpp"
   $34 = HEAP32[$33>>2]|0; //@line 310 "snd_fx.cpp"
   $35 = $34 & 1048580; //@line 310 "snd_fx.cpp"
   $36 = ($35|0)!=(0); //@line 310 "snd_fx.cpp"
   $or$cond7 = $5 & $36; //@line 310 "snd_fx.cpp"
   if ($or$cond7) {
    $37 = ((($pChn)) + 144|0); //@line 311 "snd_fx.cpp"
    $38 = HEAP32[$37>>2]|0; //@line 311 "snd_fx.cpp"
    $39 = ($38|0)==(0|0); //@line 311 "snd_fx.cpp"
    $40 = ($psmp$0|0)==($38|0); //@line 311 "snd_fx.cpp"
    $or$cond = $39 | $40; //@line 311 "snd_fx.cpp"
    if ($or$cond) {
     $bInstrumentChanged$0$off0 = 1;
    } else {
     return; //@line 408 "snd_fx.cpp"
    }
   } else {
    $bInstrumentChanged$0$off0 = 1;
   }
  } else {
   $bInstrumentChanged$0$off0 = 1;
  }
 } else {
  HEAP32[$30>>2] = $2; //@line 306 "snd_fx.cpp"
  $bInstrumentChanged$0$off0 = 0;
 }
 $41 = ((($pChn)) + 215|0); //@line 319 "snd_fx.cpp"
 HEAP8[$41>>0] = 0; //@line 319 "snd_fx.cpp"
 $42 = ($psmp$0|0)!=(0|0); //@line 320 "snd_fx.cpp"
 if ($42) {
  $43 = ((($psmp$0)) + 32|0); //@line 324 "snd_fx.cpp"
  $44 = HEAP16[$43>>1]|0; //@line 324 "snd_fx.cpp"
  $45 = $44&65535; //@line 324 "snd_fx.cpp"
  if ($5) {
   $46 = ((($2)) + 8|0); //@line 324 "snd_fx.cpp"
   $47 = HEAP16[$46>>1]|0; //@line 324 "snd_fx.cpp"
   $48 = $47&65535; //@line 324 "snd_fx.cpp"
   $49 = Math_imul($48, $45)|0; //@line 324 "snd_fx.cpp"
   $50 = $49 >>> 6; //@line 324 "snd_fx.cpp"
   $51 = ((($pChn)) + 172|0); //@line 324 "snd_fx.cpp"
   HEAP32[$51>>2] = $50; //@line 324 "snd_fx.cpp"
   $52 = ((($2)) + 4|0); //@line 325 "snd_fx.cpp"
   $53 = HEAP32[$52>>2]|0; //@line 325 "snd_fx.cpp"
   $54 = $53 & 512; //@line 325 "snd_fx.cpp"
   $55 = ($54|0)==(0); //@line 325 "snd_fx.cpp"
   if (!($55)) {
    $56 = ((($2)) + 10|0); //@line 325 "snd_fx.cpp"
    $57 = HEAP16[$56>>1]|0; //@line 325 "snd_fx.cpp"
    $58 = $57&65535; //@line 325 "snd_fx.cpp"
    $59 = ((($pChn)) + 120|0); //@line 325 "snd_fx.cpp"
    HEAP32[$59>>2] = $58; //@line 325 "snd_fx.cpp"
   }
   $60 = ((($2)) + 571|0); //@line 326 "snd_fx.cpp"
   $61 = HEAP8[$60>>0]|0; //@line 326 "snd_fx.cpp"
   $62 = ((($pChn)) + 213|0); //@line 326 "snd_fx.cpp"
   HEAP8[$62>>0] = $61; //@line 326 "snd_fx.cpp"
  } else {
   $63 = ((($pChn)) + 172|0); //@line 329 "snd_fx.cpp"
   HEAP32[$63>>2] = $45; //@line 329 "snd_fx.cpp"
  }
  $64 = ((($psmp$0)) + 34|0); //@line 331 "snd_fx.cpp"
  $65 = HEAP16[$64>>1]|0; //@line 331 "snd_fx.cpp"
  $66 = $65 & 32; //@line 331 "snd_fx.cpp"
  $67 = ($66<<16>>16)==(0); //@line 331 "snd_fx.cpp"
  if (!($67)) {
   $68 = ((($psmp$0)) + 28|0); //@line 331 "snd_fx.cpp"
   $69 = HEAP16[$68>>1]|0; //@line 331 "snd_fx.cpp"
   $70 = $69&65535; //@line 331 "snd_fx.cpp"
   $71 = ((($pChn)) + 120|0); //@line 331 "snd_fx.cpp"
   HEAP32[$71>>2] = $70; //@line 331 "snd_fx.cpp"
  }
 }
 L37: do {
  if ($bResetEnv) {
   $72 = ((($this)) + 59680|0); //@line 336 "snd_fx.cpp"
   $73 = HEAP32[$72>>2]|0; //@line 336 "snd_fx.cpp"
   $bPorta$not = $bPorta ^ 1; //@line 336 "snd_fx.cpp"
   $74 = $73 & 32;
   $75 = ($74|0)==(0); //@line 336 "snd_fx.cpp"
   $or$cond20 = $75 | $bPorta$not; //@line 336 "snd_fx.cpp"
   do {
    if (!($or$cond20)) {
     $76 = ((($this)) + 59660|0); //@line 336 "snd_fx.cpp"
     $77 = HEAP32[$76>>2]|0; //@line 336 "snd_fx.cpp"
     $78 = $77 & 8; //@line 336 "snd_fx.cpp"
     $79 = ($78|0)==(0); //@line 336 "snd_fx.cpp"
     if ($79) {
      $80 = ((($pChn)) + 32|0); //@line 337 "snd_fx.cpp"
      $81 = HEAP32[$80>>2]|0; //@line 337 "snd_fx.cpp"
      $82 = ($81|0)==(0); //@line 337 "snd_fx.cpp"
      if (!($82)) {
       $83 = ((($pChn)) + 36|0); //@line 337 "snd_fx.cpp"
       $84 = HEAP32[$83>>2]|0; //@line 337 "snd_fx.cpp"
       $85 = $84 & 1024; //@line 337 "snd_fx.cpp"
       $86 = ($85|0)==(0); //@line 337 "snd_fx.cpp"
       if (!($86)) {
        $87 = ((($pChn)) + 124|0); //@line 337 "snd_fx.cpp"
        $88 = HEAP32[$87>>2]|0; //@line 337 "snd_fx.cpp"
        $89 = ($88|0)==(0); //@line 337 "snd_fx.cpp"
        if ($89) {
         break;
        }
       }
       if (!($5)) {
        break L37;
       }
       $113 = ((($2)) + 4|0); //@line 354 "snd_fx.cpp"
       $114 = HEAP32[$113>>2]|0; //@line 354 "snd_fx.cpp"
       $115 = $114 & 1; //@line 354 "snd_fx.cpp"
       $116 = ($115|0)==(0); //@line 354 "snd_fx.cpp"
       if (!($116)) {
        break L37;
       }
       $117 = ((($pChn)) + 148|0); //@line 356 "snd_fx.cpp"
       HEAP32[$117>>2] = 0; //@line 356 "snd_fx.cpp"
       $118 = ((($pChn)) + 188|0); //@line 357 "snd_fx.cpp"
       HEAP32[$118>>2] = 0; //@line 357 "snd_fx.cpp"
       $119 = ((($pChn)) + 192|0); //@line 358 "snd_fx.cpp"
       HEAP32[$119>>2] = 0; //@line 358 "snd_fx.cpp"
       break L37;
      }
     }
    }
   } while(0);
   $90 = ((($pChn)) + 36|0); //@line 339 "snd_fx.cpp"
   $91 = HEAP32[$90>>2]|0; //@line 339 "snd_fx.cpp"
   $92 = $91 | 16777216; //@line 339 "snd_fx.cpp"
   HEAP32[$90>>2] = $92; //@line 339 "snd_fx.cpp"
   $93 = $73 & 32; //@line 340 "snd_fx.cpp"
   $94 = ($93|0)!=(0); //@line 340 "snd_fx.cpp"
   $or$cond9 = $5 & $bInstrumentChanged$0$off0; //@line 340 "snd_fx.cpp"
   $or$cond11 = $or$cond9 & $94; //@line 340 "snd_fx.cpp"
   $95 = $91 & 1536;
   $96 = ($95|0)==(0); //@line 340 "snd_fx.cpp"
   $or$cond13 = $96 & $or$cond11; //@line 340 "snd_fx.cpp"
   if ($or$cond13) {
    $97 = ((($2)) + 4|0); //@line 342 "snd_fx.cpp"
    $98 = HEAP32[$97>>2]|0; //@line 342 "snd_fx.cpp"
    $99 = $98 & 2048; //@line 342 "snd_fx.cpp"
    $100 = ($99|0)==(0); //@line 342 "snd_fx.cpp"
    if ($100) {
     $101 = ((($pChn)) + 148|0); //@line 342 "snd_fx.cpp"
     HEAP32[$101>>2] = 0; //@line 342 "snd_fx.cpp"
    }
    $102 = $98 & 4096; //@line 343 "snd_fx.cpp"
    $103 = ($102|0)==(0); //@line 343 "snd_fx.cpp"
    if ($103) {
     $104 = ((($pChn)) + 152|0); //@line 343 "snd_fx.cpp"
     HEAP32[$104>>2] = 0; //@line 343 "snd_fx.cpp"
    }
    $105 = $98 & 8192; //@line 344 "snd_fx.cpp"
    $106 = ($105|0)==(0); //@line 344 "snd_fx.cpp"
    if ($106) {
     $107 = ((($pChn)) + 156|0); //@line 344 "snd_fx.cpp"
     HEAP32[$107>>2] = 0; //@line 344 "snd_fx.cpp"
    }
   } else {
    $108 = ((($pChn)) + 148|0); //@line 347 "snd_fx.cpp"
    HEAP32[$108>>2] = 0; //@line 347 "snd_fx.cpp"
    $109 = ((($pChn)) + 152|0); //@line 348 "snd_fx.cpp"
    HEAP32[$109>>2] = 0; //@line 348 "snd_fx.cpp"
    $110 = ((($pChn)) + 156|0); //@line 349 "snd_fx.cpp"
    HEAP32[$110>>2] = 0; //@line 349 "snd_fx.cpp"
   }
   $111 = ((($pChn)) + 188|0); //@line 351 "snd_fx.cpp"
   HEAP32[$111>>2] = 0; //@line 351 "snd_fx.cpp"
   $112 = ((($pChn)) + 192|0); //@line 352 "snd_fx.cpp"
   HEAP32[$112>>2] = 0; //@line 352 "snd_fx.cpp"
  }
 } while(0);
 if (!($42)) {
  $120 = ((($pChn)) + 144|0); //@line 364 "snd_fx.cpp"
  HEAP32[$120>>2] = 0; //@line 364 "snd_fx.cpp"
  $121 = ((($pChn)) + 172|0); //@line 365 "snd_fx.cpp"
  HEAP32[$121>>2] = 0; //@line 365 "snd_fx.cpp"
  return; //@line 408 "snd_fx.cpp"
 }
 do {
  if ($bPorta) {
   $122 = ((($pChn)) + 144|0); //@line 369 "snd_fx.cpp"
   $123 = HEAP32[$122>>2]|0; //@line 369 "snd_fx.cpp"
   $124 = ($psmp$0|0)==($123|0); //@line 369 "snd_fx.cpp"
   if ($124) {
    $125 = ((($this)) + 59680|0); //@line 371 "snd_fx.cpp"
    $126 = HEAP32[$125>>2]|0; //@line 371 "snd_fx.cpp"
    $127 = $126 & 34; //@line 371 "snd_fx.cpp"
    $128 = ($127|0)==(0); //@line 371 "snd_fx.cpp"
    if ($128) {
     $129 = ((($pChn)) + 36|0); //@line 372 "snd_fx.cpp"
     $130 = HEAP32[$129>>2]|0; //@line 372 "snd_fx.cpp"
     $131 = $130 & -1664; //@line 373 "snd_fx.cpp"
     $132 = ((($psmp$0)) + 34|0); //@line 373 "snd_fx.cpp"
     $133 = HEAP16[$132>>1]|0; //@line 373 "snd_fx.cpp"
     $134 = $133&65535; //@line 373 "snd_fx.cpp"
     $135 = $134 | $131; //@line 373 "snd_fx.cpp"
     HEAP32[$129>>2] = $135; //@line 373 "snd_fx.cpp"
     $$pre$phi17Z2D = $122;$$pre$phi18Z2D = $129;$208 = $135;
     break;
    } else {
     return; //@line 408 "snd_fx.cpp"
    }
   } else {
    label = 48;
   }
  } else {
   label = 48;
  }
 } while(0);
 if ((label|0) == 48) {
  $136 = ((($pChn)) + 36|0); //@line 376 "snd_fx.cpp"
  $137 = HEAP32[$136>>2]|0; //@line 376 "snd_fx.cpp"
  $138 = $137 & -14681856; //@line 377 "snd_fx.cpp"
  $139 = ((($psmp$0)) + 34|0); //@line 377 "snd_fx.cpp"
  $140 = HEAP16[$139>>1]|0; //@line 377 "snd_fx.cpp"
  $141 = $140&65535; //@line 377 "snd_fx.cpp"
  $142 = $141 | $138; //@line 377 "snd_fx.cpp"
  HEAP32[$136>>2] = $142; //@line 377 "snd_fx.cpp"
  if ($5) {
   $143 = ((($2)) + 4|0); //@line 380 "snd_fx.cpp"
   $144 = HEAP32[$143>>2]|0; //@line 380 "snd_fx.cpp"
   $145 = $144 & 1; //@line 380 "snd_fx.cpp"
   $146 = ($145|0)==(0); //@line 380 "snd_fx.cpp"
   $147 = $142 | 2097152; //@line 380 "snd_fx.cpp"
   $148 = $146 ? $142 : $147; //@line 380 "snd_fx.cpp"
   $149 = $144 & 8; //@line 381 "snd_fx.cpp"
   $150 = ($149|0)==(0); //@line 381 "snd_fx.cpp"
   $151 = $148 | 4194304; //@line 381 "snd_fx.cpp"
   $$ = $150 ? $148 : $151; //@line 381 "snd_fx.cpp"
   $152 = $146 ^ 1; //@line 382 "snd_fx.cpp"
   $153 = $150 ^ 1; //@line 382 "snd_fx.cpp"
   $154 = $152 | $153; //@line 382 "snd_fx.cpp"
   $155 = $144 & 64; //@line 382 "snd_fx.cpp"
   $156 = ($155|0)==(0); //@line 382 "snd_fx.cpp"
   $157 = $$ | 8388608; //@line 382 "snd_fx.cpp"
   $$$ = $156 ? $$ : $157; //@line 382 "snd_fx.cpp"
   $158 = $154 ^ 1; //@line 383 "snd_fx.cpp"
   $159 = $156 ^ 1; //@line 383 "snd_fx.cpp"
   $160 = $158 ^ 1; //@line 383 "snd_fx.cpp"
   $161 = $160 | $159; //@line 383 "snd_fx.cpp"
   if ($161) {
    HEAP32[$136>>2] = $$$; //@line 383 "snd_fx.cpp"
   }
   $162 = $144 & 1088; //@line 383 "snd_fx.cpp"
   $163 = ($162|0)==(1088); //@line 383 "snd_fx.cpp"
   if ($163) {
    $164 = ((($pChn)) + 238|0); //@line 385 "snd_fx.cpp"
    $165 = HEAP8[$164>>0]|0; //@line 385 "snd_fx.cpp"
    $166 = ($165<<24>>24)==(0); //@line 385 "snd_fx.cpp"
    if ($166) {
     HEAP8[$164>>0] = 127; //@line 385 "snd_fx.cpp"
    }
   }
   $167 = ((($2)) + 576|0); //@line 387 "snd_fx.cpp"
   $168 = HEAP8[$167>>0]|0; //@line 387 "snd_fx.cpp"
   $169 = $168&255; //@line 387 "snd_fx.cpp"
   $170 = $169 & 128; //@line 387 "snd_fx.cpp"
   $171 = ($170|0)==(0); //@line 387 "snd_fx.cpp"
   if (!($171)) {
    $172 = $169 & 127; //@line 387 "snd_fx.cpp"
    $173 = $172&255; //@line 387 "snd_fx.cpp"
    $174 = ((($pChn)) + 238|0); //@line 387 "snd_fx.cpp"
    HEAP8[$174>>0] = $173; //@line 387 "snd_fx.cpp"
   }
   $175 = ((($2)) + 577|0); //@line 388 "snd_fx.cpp"
   $176 = HEAP8[$175>>0]|0; //@line 388 "snd_fx.cpp"
   $177 = $176&255; //@line 388 "snd_fx.cpp"
   $178 = $177 & 128; //@line 388 "snd_fx.cpp"
   $179 = ($178|0)==(0); //@line 388 "snd_fx.cpp"
   if ($179) {
    $223 = $$$;
   } else {
    $180 = $177 & 127; //@line 388 "snd_fx.cpp"
    $181 = $180&255; //@line 388 "snd_fx.cpp"
    $182 = ((($pChn)) + 239|0); //@line 388 "snd_fx.cpp"
    HEAP8[$182>>0] = $181; //@line 388 "snd_fx.cpp"
    $223 = $$$;
   }
  } else {
   $223 = $142;
  }
  $183 = ((($pChn)) + 210|0); //@line 390 "snd_fx.cpp"
  HEAP16[$183>>1] = 0; //@line 390 "snd_fx.cpp"
  $184 = ((($pChn)) + 208|0); //@line 390 "snd_fx.cpp"
  HEAP16[$184>>1] = 0; //@line 390 "snd_fx.cpp"
  $$pre16 = ((($pChn)) + 144|0); //@line 392 "snd_fx.cpp"
  $$pre$phi17Z2D = $$pre16;$$pre$phi18Z2D = $136;$208 = $223;
 }
 HEAP32[$$pre$phi17Z2D>>2] = $psmp$0; //@line 392 "snd_fx.cpp"
 $185 = HEAP32[$psmp$0>>2]|0; //@line 393 "snd_fx.cpp"
 $186 = ((($pChn)) + 32|0); //@line 393 "snd_fx.cpp"
 HEAP32[$186>>2] = $185; //@line 393 "snd_fx.cpp"
 $187 = ((($psmp$0)) + 4|0); //@line 394 "snd_fx.cpp"
 $188 = HEAP32[$187>>2]|0; //@line 394 "snd_fx.cpp"
 $189 = ((($pChn)) + 40|0); //@line 394 "snd_fx.cpp"
 HEAP32[$189>>2] = $188; //@line 394 "snd_fx.cpp"
 $190 = ((($psmp$0)) + 8|0); //@line 395 "snd_fx.cpp"
 $191 = HEAP32[$190>>2]|0; //@line 395 "snd_fx.cpp"
 $192 = ((($pChn)) + 44|0); //@line 395 "snd_fx.cpp"
 HEAP32[$192>>2] = $191; //@line 395 "snd_fx.cpp"
 $193 = ((($psmp$0)) + 24|0); //@line 396 "snd_fx.cpp"
 $194 = HEAP32[$193>>2]|0; //@line 396 "snd_fx.cpp"
 $195 = ((($pChn)) + 132|0); //@line 396 "snd_fx.cpp"
 HEAP32[$195>>2] = $194; //@line 396 "snd_fx.cpp"
 $196 = ((($psmp$0)) + 20|0); //@line 397 "snd_fx.cpp"
 $197 = HEAP32[$196>>2]|0; //@line 397 "snd_fx.cpp"
 $198 = ((($pChn)) + 96|0); //@line 397 "snd_fx.cpp"
 HEAP32[$198>>2] = $197; //@line 397 "snd_fx.cpp"
 $199 = ((($psmp$0)) + 36|0); //@line 398 "snd_fx.cpp"
 $200 = HEAP8[$199>>0]|0; //@line 398 "snd_fx.cpp"
 $201 = $200 << 24 >> 24; //@line 398 "snd_fx.cpp"
 $202 = ((($pChn)) + 180|0); //@line 398 "snd_fx.cpp"
 HEAP32[$202>>2] = $201; //@line 398 "snd_fx.cpp"
 $203 = ((($psmp$0)) + 37|0); //@line 399 "snd_fx.cpp"
 $204 = HEAP8[$203>>0]|0; //@line 399 "snd_fx.cpp"
 $205 = $204 << 24 >> 24; //@line 399 "snd_fx.cpp"
 $206 = ((($pChn)) + 176|0); //@line 399 "snd_fx.cpp"
 HEAP32[$206>>2] = $205; //@line 399 "snd_fx.cpp"
 $207 = $208 & 8; //@line 400 "snd_fx.cpp"
 $209 = ($207|0)==(0); //@line 400 "snd_fx.cpp"
 if ($209) {
  $219 = $208;$222 = $191;
 } else {
  $210 = ((($psmp$0)) + 12|0); //@line 402 "snd_fx.cpp"
  $211 = HEAP32[$210>>2]|0; //@line 402 "snd_fx.cpp"
  HEAP32[$189>>2] = $211; //@line 402 "snd_fx.cpp"
  $212 = ((($psmp$0)) + 16|0); //@line 403 "snd_fx.cpp"
  $213 = HEAP32[$212>>2]|0; //@line 403 "snd_fx.cpp"
  HEAP32[$192>>2] = $213; //@line 403 "snd_fx.cpp"
  $214 = $208 | 2; //@line 404 "snd_fx.cpp"
  HEAP32[$$pre$phi18Z2D>>2] = $214; //@line 404 "snd_fx.cpp"
  $215 = $208 & 16; //@line 405 "snd_fx.cpp"
  $216 = ($215|0)==(0); //@line 405 "snd_fx.cpp"
  if ($216) {
   $219 = $214;$222 = $213;
  } else {
   $217 = $208 | 6; //@line 405 "snd_fx.cpp"
   HEAP32[$$pre$phi18Z2D>>2] = $217; //@line 405 "snd_fx.cpp"
   $219 = $217;$222 = $213;
  }
 }
 $218 = $219 & 2; //@line 407 "snd_fx.cpp"
 $220 = ($218|0)!=(0); //@line 407 "snd_fx.cpp"
 $221 = ($222>>>0)<($185>>>0); //@line 407 "snd_fx.cpp"
 $or$cond21 = $220 & $221; //@line 407 "snd_fx.cpp"
 if (!($or$cond21)) {
  return; //@line 408 "snd_fx.cpp"
 }
 HEAP32[$186>>2] = $222; //@line 407 "snd_fx.cpp"
 return; //@line 408 "snd_fx.cpp"
}
function __ZN10CSoundFile10NoteChangeEjibb($this,$nChn,$note,$bPorta,$bResetEnv) {
 $this = $this|0;
 $nChn = $nChn|0;
 $note = $note|0;
 $bPorta = $bPorta|0;
 $bResetEnv = $bResetEnv|0;
 var $$ = 0, $$$1 = 0, $$0 = 0, $$04$off011 = 0, $$1 = 0, $$3 = 0, $$lobit = 0, $$not = 0, $$off = 0, $$phi$trans$insert = 0, $$phi$trans$insert15 = 0, $$pre = 0, $$pre$phi19Z2D = 0, $$pre16 = 0, $$pre18 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0;
 var $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0;
 var $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bFlt$0 = 0, $bFlt$1$off0 = 0, $bFlt$1$off0$not = 0, $brmerge = 0;
 var $extract$t6 = 0, $or$cond = 0, $or$cond23 = 0, $or$cond25 = 0, $pins$0 = 0, $pins$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($note|0)<(1); //@line 414 "snd_fx.cpp"
 if ($0) {
  return; //@line 559 "snd_fx.cpp"
 }
 $1 = (($this) + ($nChn<<8)|0); //@line 415 "snd_fx.cpp"
 $2 = (((($this) + ($nChn<<8)|0)) + 144|0); //@line 416 "snd_fx.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 416 "snd_fx.cpp"
 $4 = (((($this) + ($nChn<<8)|0)) + 140|0); //@line 417 "snd_fx.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 417 "snd_fx.cpp"
 $6 = ($5|0)!=(0|0); //@line 418 "snd_fx.cpp"
 $7 = ($note|0)<(129); //@line 418 "snd_fx.cpp"
 $or$cond = $7 & $6; //@line 418 "snd_fx.cpp"
 if ($or$cond) {
  $8 = (($note) + -1)|0; //@line 420 "snd_fx.cpp"
  $9 = (((($5)) + 300|0) + ($8)|0); //@line 420 "snd_fx.cpp"
  $10 = HEAP8[$9>>0]|0; //@line 420 "snd_fx.cpp"
  $$off = (($10) + -1)<<24>>24; //@line 421 "snd_fx.cpp"
  $11 = ($$off&255)<(239); //@line 421 "snd_fx.cpp"
  $12 = $10&255; //@line 420 "snd_fx.cpp"
  $13 = (((($this)) + 33280|0) + ($12<<6)|0); //@line 421 "snd_fx.cpp"
  $pins$0 = $11 ? $13 : $3; //@line 421 "snd_fx.cpp"
  $14 = (((($5)) + 428|0) + ($8)|0); //@line 422 "snd_fx.cpp"
  $15 = HEAP8[$14>>0]|0; //@line 422 "snd_fx.cpp"
  $16 = $15&255; //@line 422 "snd_fx.cpp"
  $$0 = $16;$pins$1 = $pins$0;
 } else {
  $$0 = $note;$pins$1 = $3;
 }
 $17 = ($$0|0)>(127); //@line 425 "snd_fx.cpp"
 if ($17) {
  __ZN10CSoundFile6KeyOffEj($this,$nChn); //@line 428 "snd_fx.cpp"
  $18 = ($$0|0)==(254); //@line 430 "snd_fx.cpp"
  if (!($18)) {
   return; //@line 559 "snd_fx.cpp"
  }
  $19 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 432 "snd_fx.cpp"
  $20 = HEAP32[$19>>2]|0; //@line 432 "snd_fx.cpp"
  $21 = $20 | 16778240; //@line 432 "snd_fx.cpp"
  HEAP32[$19>>2] = $21; //@line 432 "snd_fx.cpp"
  $22 = ((($this)) + 59680|0); //@line 433 "snd_fx.cpp"
  $23 = HEAP32[$22>>2]|0; //@line 433 "snd_fx.cpp"
  $24 = $23 & 32; //@line 433 "snd_fx.cpp"
  $25 = ($24|0)==(0); //@line 433 "snd_fx.cpp"
  if ($25) {
   label = 8;
  } else {
   $26 = ((($this)) + 59688|0); //@line 433 "snd_fx.cpp"
   $27 = HEAP32[$26>>2]|0; //@line 433 "snd_fx.cpp"
   $28 = ($27|0)==(0); //@line 433 "snd_fx.cpp"
   if (!($28)) {
    label = 8;
   }
  }
  if ((label|0) == 8) {
   $29 = (((($this) + ($nChn<<8)|0)) + 116|0); //@line 433 "snd_fx.cpp"
   HEAP32[$29>>2] = 0; //@line 433 "snd_fx.cpp"
  }
  $30 = (((($this) + ($nChn<<8)|0)) + 124|0); //@line 434 "snd_fx.cpp"
  HEAP32[$30>>2] = 0; //@line 434 "snd_fx.cpp"
  return; //@line 559 "snd_fx.cpp"
 }
 $31 = ($pins$1|0)==(0|0); //@line 438 "snd_fx.cpp"
 if ($31) {
  return; //@line 559 "snd_fx.cpp"
 }
 $$phi$trans$insert = ((($this)) + 59680|0);
 $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 444 "snd_fx.cpp"
 $32 = $$pre & 1048588;
 $33 = ($32|0)==(0); //@line 439 "snd_fx.cpp"
 $or$cond23 = $33 | $bPorta; //@line 439 "snd_fx.cpp"
 if (!($or$cond23)) {
  $34 = ((($pins$1)) + 36|0); //@line 441 "snd_fx.cpp"
  $35 = HEAP8[$34>>0]|0; //@line 441 "snd_fx.cpp"
  $36 = $35 << 24 >> 24; //@line 441 "snd_fx.cpp"
  $37 = (((($this) + ($nChn<<8)|0)) + 180|0); //@line 441 "snd_fx.cpp"
  HEAP32[$37>>2] = $36; //@line 441 "snd_fx.cpp"
  $38 = ((($pins$1)) + 37|0); //@line 442 "snd_fx.cpp"
  $39 = HEAP8[$38>>0]|0; //@line 442 "snd_fx.cpp"
  $40 = $39 << 24 >> 24; //@line 442 "snd_fx.cpp"
  $41 = (((($this) + ($nChn<<8)|0)) + 176|0); //@line 442 "snd_fx.cpp"
  HEAP32[$41>>2] = $40; //@line 442 "snd_fx.cpp"
 }
 $42 = $$pre & 1048588; //@line 444 "snd_fx.cpp"
 $43 = ($42|0)==(0); //@line 444 "snd_fx.cpp"
 if ($43) {
  $$1 = $$0;
 } else {
  $44 = (((($this) + ($nChn<<8)|0)) + 180|0); //@line 444 "snd_fx.cpp"
  $45 = HEAP32[$44>>2]|0; //@line 444 "snd_fx.cpp"
  $46 = (($45) + ($$0))|0; //@line 444 "snd_fx.cpp"
  $$1 = $46;
 }
 $47 = ($$1|0)<(1); //@line 445 "snd_fx.cpp"
 $$$1 = $47 ? 1 : $$1; //@line 445 "snd_fx.cpp"
 $48 = ($$$1|0)>(132); //@line 446 "snd_fx.cpp"
 $$3 = $48 ? 132 : $$$1; //@line 446 "snd_fx.cpp"
 $49 = $$3&255; //@line 447 "snd_fx.cpp"
 $50 = (((($this) + ($nChn<<8)|0)) + 212|0); //@line 447 "snd_fx.cpp"
 HEAP8[$50>>0] = $49; //@line 447 "snd_fx.cpp"
 $51 = $$pre & 34;
 $52 = ($51|0)==(0); //@line 448 "snd_fx.cpp"
 $or$cond25 = $52 & $bPorta; //@line 448 "snd_fx.cpp"
 if (!($or$cond25)) {
  $53 = (((($this) + ($nChn<<8)|0)) + 215|0); //@line 448 "snd_fx.cpp"
  HEAP8[$53>>0] = 0; //@line 448 "snd_fx.cpp"
 }
 $54 = (((($this) + ($nChn<<8)|0)) + 176|0); //@line 449 "snd_fx.cpp"
 $55 = HEAP32[$54>>2]|0; //@line 449 "snd_fx.cpp"
 $56 = (((($this) + ($nChn<<8)|0)) + 132|0); //@line 449 "snd_fx.cpp"
 $57 = HEAP32[$56>>2]|0; //@line 449 "snd_fx.cpp"
 $58 = (__ZNK10CSoundFile17GetPeriodFromNoteEjij($this,$$3,$55,$57)|0); //@line 449 "snd_fx.cpp"
 $59 = ($58|0)==(0); //@line 450 "snd_fx.cpp"
 do {
  if ($59) {
   $$04$off011 = 0;
  } else {
   $60 = (((($this) + ($nChn<<8)|0)) + 128|0); //@line 452 "snd_fx.cpp"
   if ($bPorta) {
    $62 = HEAP32[$60>>2]|0; //@line 452 "snd_fx.cpp"
    $63 = ($62|0)==(0); //@line 452 "snd_fx.cpp"
    if ($63) {
     HEAP32[$60>>2] = $58; //@line 452 "snd_fx.cpp"
    }
    $64 = (((($this) + ($nChn<<8)|0)) + 136|0); //@line 453 "snd_fx.cpp"
    HEAP32[$64>>2] = $58; //@line 453 "snd_fx.cpp"
    $65 = (((($this) + ($nChn<<8)|0)) + 32|0); //@line 454 "snd_fx.cpp"
    $66 = HEAP32[$65>>2]|0; //@line 454 "snd_fx.cpp"
    $67 = ($66|0)==(0); //@line 454 "snd_fx.cpp"
    if ($67) {
     $68 = HEAP32[$$phi$trans$insert>>2]|0; //@line 454 "snd_fx.cpp"
     $69 = $68 & 2; //@line 454 "snd_fx.cpp"
     $70 = ($69|0)==(0); //@line 454 "snd_fx.cpp"
     if ($70) {
      $$pre$phi19Z2D = $65;
      label = 24;
     } else {
      $125 = 0;
     }
    } else {
     $125 = $66;
    }
   } else {
    HEAP32[$60>>2] = $58; //@line 452 "snd_fx.cpp"
    $61 = (((($this) + ($nChn<<8)|0)) + 136|0); //@line 453 "snd_fx.cpp"
    HEAP32[$61>>2] = $58; //@line 453 "snd_fx.cpp"
    $$pre18 = (((($this) + ($nChn<<8)|0)) + 32|0); //@line 458 "snd_fx.cpp"
    $$pre$phi19Z2D = $$pre18;
    label = 24;
   }
   if ((label|0) == 24) {
    HEAP32[$2>>2] = $pins$1; //@line 456 "snd_fx.cpp"
    $71 = ((($pins$1)) + 20|0); //@line 457 "snd_fx.cpp"
    $72 = HEAP32[$71>>2]|0; //@line 457 "snd_fx.cpp"
    $73 = (((($this) + ($nChn<<8)|0)) + 96|0); //@line 457 "snd_fx.cpp"
    HEAP32[$73>>2] = $72; //@line 457 "snd_fx.cpp"
    $74 = HEAP32[$pins$1>>2]|0; //@line 458 "snd_fx.cpp"
    HEAP32[$$pre$phi19Z2D>>2] = $74; //@line 458 "snd_fx.cpp"
    $75 = (((($this) + ($nChn<<8)|0)) + 44|0); //@line 459 "snd_fx.cpp"
    HEAP32[$75>>2] = $74; //@line 459 "snd_fx.cpp"
    $76 = (((($this) + ($nChn<<8)|0)) + 40|0); //@line 460 "snd_fx.cpp"
    HEAP32[$76>>2] = 0; //@line 460 "snd_fx.cpp"
    $77 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 461 "snd_fx.cpp"
    $78 = HEAP32[$77>>2]|0; //@line 461 "snd_fx.cpp"
    $79 = $78 & -256; //@line 461 "snd_fx.cpp"
    $80 = ((($pins$1)) + 34|0); //@line 461 "snd_fx.cpp"
    $81 = HEAP16[$80>>1]|0; //@line 461 "snd_fx.cpp"
    $82 = $81&65535; //@line 461 "snd_fx.cpp"
    $83 = $82 | $79; //@line 461 "snd_fx.cpp"
    HEAP32[$77>>2] = $83; //@line 461 "snd_fx.cpp"
    $84 = $82 & 8; //@line 462 "snd_fx.cpp"
    $85 = ($84|0)==(0); //@line 462 "snd_fx.cpp"
    if ($85) {
     $96 = $82 & 2; //@line 471 "snd_fx.cpp"
     $97 = ($96|0)==(0); //@line 471 "snd_fx.cpp"
     if ($97) {
      $260 = $74;
     } else {
      $98 = ((($pins$1)) + 4|0); //@line 473 "snd_fx.cpp"
      $99 = HEAP32[$98>>2]|0; //@line 473 "snd_fx.cpp"
      HEAP32[$76>>2] = $99; //@line 473 "snd_fx.cpp"
      $100 = ((($pins$1)) + 8|0); //@line 474 "snd_fx.cpp"
      $101 = HEAP32[$100>>2]|0; //@line 474 "snd_fx.cpp"
      HEAP32[$75>>2] = $101; //@line 474 "snd_fx.cpp"
      $102 = ($74>>>0)>($101>>>0); //@line 475 "snd_fx.cpp"
      if ($102) {
       HEAP32[$$pre$phi19Z2D>>2] = $101; //@line 475 "snd_fx.cpp"
       $260 = $101;
      } else {
       $260 = $74;
      }
     }
    } else {
     $86 = ((($pins$1)) + 12|0); //@line 464 "snd_fx.cpp"
     $87 = HEAP32[$86>>2]|0; //@line 464 "snd_fx.cpp"
     HEAP32[$76>>2] = $87; //@line 464 "snd_fx.cpp"
     $88 = ((($pins$1)) + 16|0); //@line 465 "snd_fx.cpp"
     $89 = HEAP32[$88>>2]|0; //@line 465 "snd_fx.cpp"
     HEAP32[$75>>2] = $89; //@line 465 "snd_fx.cpp"
     $90 = $83 & -7; //@line 466 "snd_fx.cpp"
     $91 = $90 | 2; //@line 467 "snd_fx.cpp"
     $92 = $82 & 16; //@line 468 "snd_fx.cpp"
     $93 = ($92|0)==(0); //@line 468 "snd_fx.cpp"
     $94 = $83 | 6; //@line 468 "snd_fx.cpp"
     $$ = $93 ? $91 : $94; //@line 468 "snd_fx.cpp"
     HEAP32[$77>>2] = $$; //@line 467 "snd_fx.cpp"
     $95 = ($74>>>0)>($89>>>0); //@line 469 "snd_fx.cpp"
     if ($95) {
      HEAP32[$$pre$phi19Z2D>>2] = $89; //@line 469 "snd_fx.cpp"
      $260 = $89;
     } else {
      $260 = $74;
     }
    }
    $103 = (((($this) + ($nChn<<8)|0)) + 4|0); //@line 477 "snd_fx.cpp"
    HEAP32[$103>>2] = 0; //@line 477 "snd_fx.cpp"
    $104 = (((($this) + ($nChn<<8)|0)) + 8|0); //@line 478 "snd_fx.cpp"
    HEAP32[$104>>2] = 0; //@line 478 "snd_fx.cpp"
    $105 = (((($this) + ($nChn<<8)|0)) + 224|0); //@line 479 "snd_fx.cpp"
    $106 = HEAP8[$105>>0]|0; //@line 479 "snd_fx.cpp"
    $107 = ($106&255)<(4); //@line 479 "snd_fx.cpp"
    if ($107) {
     $108 = HEAP32[$$phi$trans$insert>>2]|0; //@line 479 "snd_fx.cpp"
     $109 = $108 & 32; //@line 479 "snd_fx.cpp"
     $110 = ($109|0)==(0); //@line 479 "snd_fx.cpp"
     if ($110) {
      $116 = 0;
     } else {
      $111 = ((($this)) + 59660|0); //@line 479 "snd_fx.cpp"
      $112 = HEAP32[$111>>2]|0; //@line 479 "snd_fx.cpp"
      $113 = $112 & 4; //@line 479 "snd_fx.cpp"
      $114 = ($113|0)==(0); //@line 479 "snd_fx.cpp"
      $116 = $114;
     }
     $115 = $116 ? 16 : 0; //@line 479 "snd_fx.cpp"
     $117 = (((($this) + ($nChn<<8)|0)) + 196|0); //@line 479 "snd_fx.cpp"
     HEAP32[$117>>2] = $115; //@line 479 "snd_fx.cpp"
    }
    $118 = (((($this) + ($nChn<<8)|0)) + 227|0); //@line 480 "snd_fx.cpp"
    $119 = HEAP8[$118>>0]|0; //@line 480 "snd_fx.cpp"
    $120 = ($119&255)<(4); //@line 480 "snd_fx.cpp"
    if ($120) {
     $121 = (((($this) + ($nChn<<8)|0)) + 200|0); //@line 480 "snd_fx.cpp"
     HEAP32[$121>>2] = 0; //@line 480 "snd_fx.cpp"
     $125 = $260;
    } else {
     $125 = $260;
    }
   }
   $122 = (((($this) + ($nChn<<8)|0)) + 4|0); //@line 482 "snd_fx.cpp"
   $123 = HEAP32[$122>>2]|0; //@line 482 "snd_fx.cpp"
   $124 = ($123>>>0)<($125>>>0); //@line 482 "snd_fx.cpp"
   if ($124) {
    if (!($bPorta)) {
     $$04$off011 = 0;
     break;
    }
   } else {
    $126 = (((($this) + ($nChn<<8)|0)) + 40|0); //@line 482 "snd_fx.cpp"
    $127 = HEAP32[$126>>2]|0; //@line 482 "snd_fx.cpp"
    HEAP32[$122>>2] = $127; //@line 482 "snd_fx.cpp"
    if (!($bPorta)) {
     $$04$off011 = 0;
     break;
    }
   }
   $128 = HEAP32[$$phi$trans$insert>>2]|0; //@line 484 "snd_fx.cpp"
   $129 = $128 & 32; //@line 484 "snd_fx.cpp"
   $130 = ($129|0)==(0); //@line 484 "snd_fx.cpp"
   if ($130) {
    $$04$off011 = 1;
   } else {
    $131 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 485 "snd_fx.cpp"
    $132 = HEAP32[$131>>2]|0; //@line 485 "snd_fx.cpp"
    $133 = $132 & 1024; //@line 485 "snd_fx.cpp"
    $134 = ($133|0)==(0); //@line 485 "snd_fx.cpp"
    if (!($134)) {
     $135 = (((($this) + ($nChn<<8)|0)) + 124|0); //@line 485 "snd_fx.cpp"
     $136 = HEAP32[$135>>2]|0; //@line 485 "snd_fx.cpp"
     $137 = ($136|0)==(0); //@line 485 "snd_fx.cpp"
     if ($137) {
      $$04$off011 = 1;
      break;
     }
    }
    $138 = ((($this)) + 59660|0); //@line 486 "snd_fx.cpp"
    $139 = HEAP32[$138>>2]|0; //@line 486 "snd_fx.cpp"
    $140 = $139 & 8; //@line 486 "snd_fx.cpp"
    $141 = ($140|0)==(0); //@line 486 "snd_fx.cpp"
    if (!($141)) {
     $142 = (((($this) + ($nChn<<8)|0)) + 247|0); //@line 486 "snd_fx.cpp"
     $143 = HEAP8[$142>>0]|0; //@line 486 "snd_fx.cpp"
     $144 = ($143<<24>>24)==(0); //@line 486 "snd_fx.cpp"
     if (!($144)) {
      $$04$off011 = 1;
      break;
     }
    }
    $180 = $132 & -33554945; //@line 507 "snd_fx.cpp"
    HEAP32[$131>>2] = $180; //@line 507 "snd_fx.cpp"
    return; //@line 559 "snd_fx.cpp"
   }
  }
 } while(0);
 $145 = HEAP32[$$phi$trans$insert>>2]|0; //@line 488 "snd_fx.cpp"
 $146 = $145 & 32; //@line 488 "snd_fx.cpp"
 $147 = ($146|0)==(0); //@line 488 "snd_fx.cpp"
 if (!($147)) {
  $148 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 488 "snd_fx.cpp"
  $149 = HEAP32[$148>>2]|0; //@line 488 "snd_fx.cpp"
  $150 = $149 & 1024; //@line 488 "snd_fx.cpp"
  $151 = ($150|0)==(0); //@line 488 "snd_fx.cpp"
  if (!($151)) {
   $152 = (((($this) + ($nChn<<8)|0)) + 124|0); //@line 488 "snd_fx.cpp"
   $153 = HEAP32[$152>>2]|0; //@line 488 "snd_fx.cpp"
   $154 = ($153|0)==(0); //@line 488 "snd_fx.cpp"
   if ($154) {
    $155 = (((($this) + ($nChn<<8)|0)) + 148|0); //@line 490 "snd_fx.cpp"
    HEAP32[$155>>2] = 0; //@line 490 "snd_fx.cpp"
    $156 = (((($this) + ($nChn<<8)|0)) + 152|0); //@line 491 "snd_fx.cpp"
    HEAP32[$156>>2] = 0; //@line 491 "snd_fx.cpp"
    $157 = (((($this) + ($nChn<<8)|0)) + 156|0); //@line 492 "snd_fx.cpp"
    HEAP32[$157>>2] = 0; //@line 492 "snd_fx.cpp"
    $158 = (((($this) + ($nChn<<8)|0)) + 188|0); //@line 493 "snd_fx.cpp"
    HEAP32[$158>>2] = 0; //@line 493 "snd_fx.cpp"
    $159 = (((($this) + ($nChn<<8)|0)) + 192|0); //@line 494 "snd_fx.cpp"
    HEAP32[$159>>2] = 0; //@line 494 "snd_fx.cpp"
    $160 = $149 & -1025; //@line 495 "snd_fx.cpp"
    HEAP32[$148>>2] = $160; //@line 495 "snd_fx.cpp"
    HEAP32[$152>>2] = 65536; //@line 496 "snd_fx.cpp"
   }
  }
 }
 if ($$04$off011) {
  $161 = ((($this)) + 59660|0); //@line 498 "snd_fx.cpp"
  $162 = HEAP32[$161>>2]|0; //@line 498 "snd_fx.cpp"
  $163 = $162 & 8; //@line 498 "snd_fx.cpp"
  $164 = ($163|0)==(0); //@line 498 "snd_fx.cpp"
  if (!($164)) {
   $165 = (((($this) + ($nChn<<8)|0)) + 247|0); //@line 498 "snd_fx.cpp"
   $166 = HEAP8[$165>>0]|0; //@line 498 "snd_fx.cpp"
   $167 = ($166<<24>>24)==(0); //@line 498 "snd_fx.cpp"
   if ($167) {
    $168 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 507 "snd_fx.cpp"
    $169 = HEAP32[$168>>2]|0; //@line 507 "snd_fx.cpp"
    $170 = $169 & -33554945; //@line 507 "snd_fx.cpp"
    HEAP32[$168>>2] = $170; //@line 507 "snd_fx.cpp"
    return; //@line 559 "snd_fx.cpp"
   }
  }
 }
 $171 = $145 & 1048580; //@line 500 "snd_fx.cpp"
 $172 = ($171|0)==(0); //@line 500 "snd_fx.cpp"
 if ($172) {
  label = 54;
 } else {
  $173 = (((($this) + ($nChn<<8)|0)) + 247|0); //@line 500 "snd_fx.cpp"
  $174 = HEAP8[$173>>0]|0; //@line 500 "snd_fx.cpp"
  $175 = ($174<<24>>24)==(0); //@line 500 "snd_fx.cpp"
  if ($175) {
   $$phi$trans$insert15 = (((($this) + ($nChn<<8)|0)) + 36|0);
   $$pre16 = HEAP32[$$phi$trans$insert15>>2]|0; //@line 507 "snd_fx.cpp"
   $181 = $$pre16 & -33554945; //@line 507 "snd_fx.cpp"
   HEAP32[$$phi$trans$insert15>>2] = $181; //@line 507 "snd_fx.cpp"
   if ($$04$off011) {
    return; //@line 559 "snd_fx.cpp"
   } else {
    $186 = $$phi$trans$insert15;
   }
  } else {
   label = 54;
  }
 }
 if ((label|0) == 54) {
  $176 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 502 "snd_fx.cpp"
  $177 = HEAP32[$176>>2]|0; //@line 502 "snd_fx.cpp"
  $178 = (((($this) + ($nChn<<8)|0)) + 124|0); //@line 503 "snd_fx.cpp"
  HEAP32[$178>>2] = 65536; //@line 503 "snd_fx.cpp"
  $179 = $177 & -33555969; //@line 507 "snd_fx.cpp"
  HEAP32[$176>>2] = $179; //@line 507 "snd_fx.cpp"
  if ($$04$off011) {
   return; //@line 559 "snd_fx.cpp"
  } else {
   $186 = $176;
  }
 }
 $182 = (((($this) + ($nChn<<8)|0)) + 164|0); //@line 511 "snd_fx.cpp"
 HEAP32[$182>>2] = 256; //@line 511 "snd_fx.cpp"
 $183 = (((($this) + ($nChn<<8)|0)) + 253|0); //@line 512 "snd_fx.cpp"
 HEAP8[$183>>0] = -1; //@line 512 "snd_fx.cpp"
 $184 = (((($this) + ($nChn<<8)|0)) + 252|0); //@line 512 "snd_fx.cpp"
 HEAP8[$184>>0] = -1; //@line 512 "snd_fx.cpp"
 $185 = HEAP32[$186>>2]|0; //@line 513 "snd_fx.cpp"
 $187 = $185 & -16793601; //@line 513 "snd_fx.cpp"
 $188 = $187 | 16777216; //@line 514 "snd_fx.cpp"
 HEAP32[$186>>2] = $188; //@line 514 "snd_fx.cpp"
 $189 = (((($this) + ($nChn<<8)|0)) + 240|0); //@line 515 "snd_fx.cpp"
 HEAP8[$189>>0] = 0; //@line 515 "snd_fx.cpp"
 $190 = (((($this) + ($nChn<<8)|0)) + 242|0); //@line 516 "snd_fx.cpp"
 HEAP8[$190>>0] = 0; //@line 516 "snd_fx.cpp"
 if ($bResetEnv) {
  $191 = (((($this) + ($nChn<<8)|0)) + 210|0); //@line 519 "snd_fx.cpp"
  HEAP16[$191>>1] = 0; //@line 519 "snd_fx.cpp"
  $192 = (((($this) + ($nChn<<8)|0)) + 208|0); //@line 519 "snd_fx.cpp"
  HEAP16[$192>>1] = 0; //@line 519 "snd_fx.cpp"
  do {
   if ($6) {
    $193 = ((($5)) + 4|0); //@line 522 "snd_fx.cpp"
    $194 = HEAP32[$193>>2]|0; //@line 522 "snd_fx.cpp"
    $195 = $194 & 2048; //@line 522 "snd_fx.cpp"
    $196 = ($195|0)==(0); //@line 522 "snd_fx.cpp"
    if ($196) {
     $197 = (((($this) + ($nChn<<8)|0)) + 148|0); //@line 522 "snd_fx.cpp"
     HEAP32[$197>>2] = 0; //@line 522 "snd_fx.cpp"
    }
    $198 = $194 & 4096; //@line 523 "snd_fx.cpp"
    $199 = ($198|0)==(0); //@line 523 "snd_fx.cpp"
    if ($199) {
     $200 = (((($this) + ($nChn<<8)|0)) + 152|0); //@line 523 "snd_fx.cpp"
     HEAP32[$200>>2] = 0; //@line 523 "snd_fx.cpp"
    }
    $201 = $194 & 8192; //@line 524 "snd_fx.cpp"
    $202 = ($201|0)==(0); //@line 524 "snd_fx.cpp"
    if ($202) {
     $203 = (((($this) + ($nChn<<8)|0)) + 156|0); //@line 524 "snd_fx.cpp"
     HEAP32[$203>>2] = 0; //@line 524 "snd_fx.cpp"
    }
    $204 = HEAP32[$$phi$trans$insert>>2]|0; //@line 525 "snd_fx.cpp"
    $205 = $204 & 32; //@line 525 "snd_fx.cpp"
    $206 = ($205|0)==(0); //@line 525 "snd_fx.cpp"
    if (!($206)) {
     $207 = ((($5)) + 575|0); //@line 528 "snd_fx.cpp"
     $208 = HEAP8[$207>>0]|0; //@line 528 "snd_fx.cpp"
     $209 = ($208<<24>>24)==(0); //@line 528 "snd_fx.cpp"
     if (!($209)) {
      $210 = $208&255; //@line 530 "snd_fx.cpp"
      $211 = (_rand()|0); //@line 530 "snd_fx.cpp"
      $212 = $211 & 255; //@line 530 "snd_fx.cpp"
      $213 = (($212) + -127)|0; //@line 530 "snd_fx.cpp"
      $214 = Math_imul($213, $210)|0; //@line 530 "snd_fx.cpp"
      $215 = (($214|0) / 128)&-1; //@line 530 "snd_fx.cpp"
      $216 = (((($this) + ($nChn<<8)|0)) + 116|0); //@line 531 "snd_fx.cpp"
      $217 = HEAP32[$216>>2]|0; //@line 531 "snd_fx.cpp"
      $218 = Math_imul($217, $215)|0; //@line 531 "snd_fx.cpp"
      $219 = (($218) + 1)|0; //@line 531 "snd_fx.cpp"
      $220 = (($219|0) / 128)&-1; //@line 531 "snd_fx.cpp"
      $221 = $220&65535; //@line 531 "snd_fx.cpp"
      HEAP16[$192>>1] = $221; //@line 531 "snd_fx.cpp"
     }
     $222 = ((($5)) + 574|0); //@line 534 "snd_fx.cpp"
     $223 = HEAP8[$222>>0]|0; //@line 534 "snd_fx.cpp"
     $224 = ($223<<24>>24)==(0); //@line 534 "snd_fx.cpp"
     if ($224) {
      break;
     }
     $225 = $223&255; //@line 536 "snd_fx.cpp"
     $226 = (_rand()|0); //@line 536 "snd_fx.cpp"
     $227 = $226 & 255; //@line 536 "snd_fx.cpp"
     $228 = (($227) + -127)|0; //@line 536 "snd_fx.cpp"
     $229 = Math_imul($228, $225)|0; //@line 536 "snd_fx.cpp"
     $230 = (($229|0) / 128)&-1; //@line 536 "snd_fx.cpp"
     $231 = $230&65535; //@line 537 "snd_fx.cpp"
     HEAP16[$191>>1] = $231; //@line 537 "snd_fx.cpp"
    }
   }
  } while(0);
  $232 = (((($this) + ($nChn<<8)|0)) + 188|0); //@line 541 "snd_fx.cpp"
  HEAP32[$232>>2] = 0; //@line 541 "snd_fx.cpp"
  $233 = (((($this) + ($nChn<<8)|0)) + 192|0); //@line 542 "snd_fx.cpp"
  HEAP32[$233>>2] = 0; //@line 542 "snd_fx.cpp"
 }
 $234 = (((($this) + ($nChn<<8)|0)) + 16|0); //@line 544 "snd_fx.cpp"
 HEAP32[$234>>2] = 0; //@line 544 "snd_fx.cpp"
 $235 = (((($this) + ($nChn<<8)|0)) + 20|0); //@line 544 "snd_fx.cpp"
 HEAP32[$235>>2] = 0; //@line 544 "snd_fx.cpp"
 $236 = ((($this)) + 59660|0); //@line 545 "snd_fx.cpp"
 $237 = HEAP32[$236>>2]|0; //@line 545 "snd_fx.cpp"
 $238 = $237 >>> 13; //@line 545 "snd_fx.cpp"
 $$lobit = $238 & 1; //@line 545 "snd_fx.cpp"
 if ($6) {
  $239 = ($$lobit|0)==(0); //@line 545 "snd_fx.cpp"
  $240 = ((($5)) + 577|0); //@line 549 "snd_fx.cpp"
  $241 = HEAP8[$240>>0]|0; //@line 549 "snd_fx.cpp"
  $242 = $241&255; //@line 549 "snd_fx.cpp"
  $243 = $242 & 128; //@line 549 "snd_fx.cpp"
  $244 = ($243|0)==(0); //@line 549 "snd_fx.cpp"
  if ($244) {
   $bFlt$0 = $239;
  } else {
   $245 = $242 & 127; //@line 549 "snd_fx.cpp"
   $246 = $245&255; //@line 549 "snd_fx.cpp"
   $247 = (((($this) + ($nChn<<8)|0)) + 239|0); //@line 549 "snd_fx.cpp"
   HEAP8[$247>>0] = $246; //@line 549 "snd_fx.cpp"
   $bFlt$0 = 1;
  }
  $248 = ((($5)) + 576|0); //@line 550 "snd_fx.cpp"
  $249 = HEAP8[$248>>0]|0; //@line 550 "snd_fx.cpp"
  $250 = $249&255; //@line 550 "snd_fx.cpp"
  $251 = $250 & 128; //@line 550 "snd_fx.cpp"
  $252 = ($251|0)==(0); //@line 550 "snd_fx.cpp"
  if ($252) {
   $bFlt$1$off0 = $bFlt$0;
  } else {
   $253 = $250 & 127; //@line 550 "snd_fx.cpp"
   $254 = $253&255; //@line 550 "snd_fx.cpp"
   $255 = (((($this) + ($nChn<<8)|0)) + 238|0); //@line 550 "snd_fx.cpp"
   HEAP8[$255>>0] = $254; //@line 550 "snd_fx.cpp"
   $bFlt$1$off0 = 1;
  }
 } else {
  $256 = (((($this) + ($nChn<<8)|0)) + 210|0); //@line 553 "snd_fx.cpp"
  HEAP16[$256>>1] = 0; //@line 553 "snd_fx.cpp"
  $257 = (((($this) + ($nChn<<8)|0)) + 208|0); //@line 553 "snd_fx.cpp"
  HEAP16[$257>>1] = 0; //@line 553 "snd_fx.cpp"
  $extract$t6 = ($$lobit|0)==(0);
  $bFlt$1$off0 = $extract$t6;
 }
 $258 = (((($this) + ($nChn<<8)|0)) + 238|0); //@line 556 "snd_fx.cpp"
 $259 = HEAP8[$258>>0]|0; //@line 556 "snd_fx.cpp"
 $$not = ($259&255)>(126); //@line 556 "snd_fx.cpp"
 $bFlt$1$off0$not = $bFlt$1$off0 ^ 1; //@line 556 "snd_fx.cpp"
 $brmerge = $$not | $bFlt$1$off0$not; //@line 556 "snd_fx.cpp"
 if ($brmerge) {
  return; //@line 559 "snd_fx.cpp"
 }
 __ZNK10CSoundFile18SetupChannelFilterEP11_MODCHANNELbi($this,$1,1,256); //@line 556 "snd_fx.cpp"
 return; //@line 559 "snd_fx.cpp"
}
function __ZNK10CSoundFile17GetPeriodFromNoteEjij($this,$note,$nFineTune,$nC4Speed) {
 $this = $this|0;
 $note = $note|0;
 $nFineTune = $nFineTune|0;
 $nC4Speed = $nC4Speed|0;
 var $$ = 0, $$$2 = 0, $$$3 = 0, $$0 = 0, $$2 = 0, $$3 = 0, $$nC4Speed = 0, $$note = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $9 = 0, $finetune$0 = 0, $rfine$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($note) + -1)|0; //@line 2293 "snd_fx.cpp"
 $1 = ($0>>>0)>(239); //@line 2293 "snd_fx.cpp"
 if ($1) {
  $$0 = 0;
  return ($$0|0); //@line 2349 "snd_fx.cpp"
 }
 $2 = ((($this)) + 59680|0); //@line 2294 "snd_fx.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 2294 "snd_fx.cpp"
 $4 = $3 & 5136290; //@line 2294 "snd_fx.cpp"
 $5 = ($4|0)==(0); //@line 2294 "snd_fx.cpp"
 if (!($5)) {
  $6 = ((($this)) + 59660|0); //@line 2298 "snd_fx.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 2298 "snd_fx.cpp"
  $8 = $7 & 16; //@line 2298 "snd_fx.cpp"
  $9 = ($8|0)==(0); //@line 2298 "snd_fx.cpp"
  if ($9) {
   $17 = ($nC4Speed|0)==(0); //@line 2303 "snd_fx.cpp"
   $$nC4Speed = $17 ? 8363 : $nC4Speed; //@line 2303 "snd_fx.cpp"
   $18 = (($0>>>0) % 12)&-1; //@line 2304 "snd_fx.cpp"
   $19 = (168524 + ($18<<1)|0); //@line 2304 "snd_fx.cpp"
   $20 = HEAP16[$19>>1]|0; //@line 2304 "snd_fx.cpp"
   $21 = $20&65535; //@line 2304 "snd_fx.cpp"
   $22 = $21 << 5; //@line 2304 "snd_fx.cpp"
   $23 = (($0>>>0) / 12)&-1; //@line 2304 "snd_fx.cpp"
   $24 = $$nC4Speed << $23; //@line 2304 "snd_fx.cpp"
   $25 = (__Z7_muldivlll(8363,$22,$24)|0); //@line 2304 "snd_fx.cpp"
   $$0 = $25;
   return ($$0|0); //@line 2349 "snd_fx.cpp"
  } else {
   $10 = (($0>>>0) % 12)&-1; //@line 2300 "snd_fx.cpp"
   $11 = (168524 + ($10<<1)|0); //@line 2300 "snd_fx.cpp"
   $12 = HEAP16[$11>>1]|0; //@line 2300 "snd_fx.cpp"
   $13 = $12&65535; //@line 2300 "snd_fx.cpp"
   $14 = $13 << 5; //@line 2300 "snd_fx.cpp"
   $15 = (($0>>>0) / 12)&-1; //@line 2300 "snd_fx.cpp"
   $16 = $14 >>> $15; //@line 2300 "snd_fx.cpp"
   $$0 = $16;
   return ($$0|0); //@line 2349 "snd_fx.cpp"
  }
 }
 $26 = $3 & 1048580; //@line 2307 "snd_fx.cpp"
 $27 = ($26|0)==(0); //@line 2307 "snd_fx.cpp"
 if ($27) {
  $68 = $nFineTune >>> 4; //@line 2343 "snd_fx.cpp"
  $69 = $68 & 15; //@line 2343 "snd_fx.cpp"
  $70 = ($69|0)!=(0); //@line 2344 "snd_fx.cpp"
  $$off = (($note) + -37)|0; //@line 2344 "snd_fx.cpp"
  $71 = ($$off>>>0)>(71); //@line 2344 "snd_fx.cpp"
  $72 = $71 | $70; //@line 2344 "snd_fx.cpp"
  if ($72) {
   $73 = ($69*12)|0; //@line 2345 "snd_fx.cpp"
   $74 = (($0>>>0) % 12)&-1; //@line 2345 "snd_fx.cpp"
   $75 = (($74) + ($73))|0; //@line 2345 "snd_fx.cpp"
   $76 = (168764 + ($75<<1)|0); //@line 2345 "snd_fx.cpp"
   $77 = HEAP16[$76>>1]|0; //@line 2345 "snd_fx.cpp"
   $78 = $77&65535; //@line 2345 "snd_fx.cpp"
   $79 = $78 << 5; //@line 2345 "snd_fx.cpp"
   $80 = (($0>>>0) / 12)&-1; //@line 2345 "snd_fx.cpp"
   $81 = $79 >>> $80; //@line 2345 "snd_fx.cpp"
   $$0 = $81;
   return ($$0|0); //@line 2349 "snd_fx.cpp"
  } else {
   $82 = (169148 + ($$off<<1)|0); //@line 2347 "snd_fx.cpp"
   $83 = HEAP16[$82>>1]|0; //@line 2347 "snd_fx.cpp"
   $84 = $83&65535; //@line 2347 "snd_fx.cpp"
   $85 = $84 << 2; //@line 2347 "snd_fx.cpp"
   $$0 = $85;
   return ($$0|0); //@line 2349 "snd_fx.cpp"
  }
 } else {
  $28 = ($note>>>0)<(13); //@line 2309 "snd_fx.cpp"
  $$note = $28 ? 13 : $note; //@line 2309 "snd_fx.cpp"
  $29 = (($$note) + -13)|0; //@line 2310 "snd_fx.cpp"
  $30 = ((($this)) + 59660|0); //@line 2311 "snd_fx.cpp"
  $31 = HEAP32[$30>>2]|0; //@line 2311 "snd_fx.cpp"
  $32 = $31 & 16; //@line 2311 "snd_fx.cpp"
  $33 = ($32|0)==(0); //@line 2311 "snd_fx.cpp"
  if ($33) {
   $39 = (($29>>>0) % 12)&-1; //@line 2319 "snd_fx.cpp"
   $40 = $39 << 3; //@line 2319 "snd_fx.cpp"
   $41 = (($29>>>0) / 12)&-1; //@line 2320 "snd_fx.cpp"
   $42 = (($nFineTune|0) / 16)&-1; //@line 2321 "snd_fx.cpp"
   $43 = (($40) + 8)|0; //@line 2322 "snd_fx.cpp"
   $44 = (($43) + ($42))|0; //@line 2322 "snd_fx.cpp"
   $45 = ($44|0)<(0); //@line 2323 "snd_fx.cpp"
   $$2 = $45 ? 0 : $44; //@line 2323 "snd_fx.cpp"
   $46 = ($$2|0)>(103); //@line 2324 "snd_fx.cpp"
   $$$2 = $46 ? 103 : $$2; //@line 2324 "snd_fx.cpp"
   $47 = (168556 + ($$$2<<1)|0); //@line 2325 "snd_fx.cpp"
   $48 = HEAP16[$47>>1]|0; //@line 2325 "snd_fx.cpp"
   $49 = $48&65535; //@line 2325 "snd_fx.cpp"
   $50 = ($nFineTune|0)<(0); //@line 2326 "snd_fx.cpp"
   $51 = (0 - ($nFineTune))|0; //@line 2329 "snd_fx.cpp"
   $finetune$0 = $50 ? $51 : $nFineTune; //@line 2326 "snd_fx.cpp"
   $52 = $nFineTune >> 31; //@line 2326 "snd_fx.cpp"
   $53 = $52 | 1; //@line 2326 "snd_fx.cpp"
   $rfine$0 = (($53) + 8)|0; //@line 2326 "snd_fx.cpp"
   $54 = (($rfine$0) + ($40))|0; //@line 2331 "snd_fx.cpp"
   $55 = (($54) + ($42))|0; //@line 2331 "snd_fx.cpp"
   $56 = ($55|0)<(0); //@line 2332 "snd_fx.cpp"
   $$3 = $56 ? 0 : $55; //@line 2332 "snd_fx.cpp"
   $57 = ($$3|0)>(103); //@line 2333 "snd_fx.cpp"
   $$$3 = $57 ? 103 : $$3; //@line 2333 "snd_fx.cpp"
   $58 = (168556 + ($$$3<<1)|0); //@line 2334 "snd_fx.cpp"
   $59 = HEAP16[$58>>1]|0; //@line 2334 "snd_fx.cpp"
   $60 = $59&65535; //@line 2334 "snd_fx.cpp"
   $61 = $finetune$0 & 15; //@line 2335 "snd_fx.cpp"
   $62 = (16 - ($61))|0; //@line 2336 "snd_fx.cpp"
   $63 = Math_imul($49, $62)|0; //@line 2336 "snd_fx.cpp"
   $64 = Math_imul($60, $61)|0; //@line 2337 "snd_fx.cpp"
   $65 = (($64) + ($63))|0; //@line 2338 "snd_fx.cpp"
   $66 = $65 << 1; //@line 2338 "snd_fx.cpp"
   $67 = $66 >>> $41; //@line 2338 "snd_fx.cpp"
   $$0 = $67;
   return ($$0|0); //@line 2349 "snd_fx.cpp"
  } else {
   $34 = (133 - ($$note))|0; //@line 2313 "snd_fx.cpp"
   $35 = $34 << 6; //@line 2313 "snd_fx.cpp"
   $36 = (($nFineTune|0) / 2)&-1; //@line 2313 "snd_fx.cpp"
   $37 = (($35) - ($36))|0; //@line 2313 "snd_fx.cpp"
   $38 = ($37|0)<(1); //@line 2314 "snd_fx.cpp"
   $$ = $38 ? 1 : $37; //@line 2314 "snd_fx.cpp"
   $$0 = $$;
   return ($$0|0); //@line 2349 "snd_fx.cpp"
  }
 }
 return (0)|0;
}
function __ZNK10CSoundFile18SetupChannelFilterEP11_MODCHANNELbi($this,$pChn,$bReset,$flt_modifier) {
 $this = $this|0;
 $pChn = $pChn|0;
 $bReset = $bReset|0;
 $flt_modifier = $flt_modifier|0;
 var $$$i = 0, $$0$i = 0.0, $$sink$i = 0.0, $$sink$v$i = 0.0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0;
 var $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0;
 var $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0, $52 = 0.0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0.0, $Fc$0$i = 0.0, $Fc$0$in$i = 0.0, $d$0 = 0.0, $exp21$i = 0.0, $factor = 0.0, $phitmp = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChn)) + 238|0); //@line 74 "snd_flt.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 74 "snd_flt.cpp"
 $2 = $1&255; //@line 74 "snd_flt.cpp"
 $3 = ((($this)) + 59660|0); //@line 58 "snd_flt.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 58 "snd_flt.cpp"
 $5 = $4 & 32768; //@line 58 "snd_flt.cpp"
 $6 = ($5|0)==(0); //@line 58 "snd_flt.cpp"
 $7 = (($flt_modifier) + 256)|0; //@line 59 "snd_flt.cpp"
 $8 = Math_imul($2, $7)|0; //@line 59 "snd_flt.cpp"
 $9 = (+($8>>>0)); //@line 59 "snd_flt.cpp"
 $$sink$v$i = $6 ? 8.1380210758652538E-5 : 9.3005954113323241E-5; //@line 58 "snd_flt.cpp"
 $$sink$i = $9 * $$sink$v$i; //@line 58 "snd_flt.cpp"
 $10 = $$sink$i + 0.25; //@line 59 "snd_flt.cpp"
 $11 = $10; //@line 59 "snd_flt.cpp"
 $exp21$i = (+_exp2($11)); //@line 59 "snd_flt.cpp"
 $Fc$0$in$i = $exp21$i * 110.0; //@line 59 "snd_flt.cpp"
 $Fc$0$i = $Fc$0$in$i; //@line 59 "snd_flt.cpp"
 $12 = (~~(($Fc$0$i))); //@line 62 "snd_flt.cpp"
 $13 = ($12|0)<(120); //@line 63 "snd_flt.cpp"
 if ($13) {
  $$0$i = 120.0;
 } else {
  $14 = ($12|0)>(10000); //@line 64 "snd_flt.cpp"
  if ($14) {
   $$0$i = 1.0E+4;
  } else {
   $15 = $12 << 1; //@line 65 "snd_flt.cpp"
   $16 = HEAP32[10109]|0; //@line 65 "snd_flt.cpp"
   $17 = ($15|0)>($16|0); //@line 65 "snd_flt.cpp"
   $18 = $16 >>> 1; //@line 65 "snd_flt.cpp"
   $$$i = $17 ? $18 : $12; //@line 65 "snd_flt.cpp"
   $phitmp = (+($$$i>>>0)); //@line 66 "snd_flt.cpp"
   $$0$i = $phitmp;
  }
 }
 $19 = HEAP32[10109]|0; //@line 75 "snd_flt.cpp"
 $20 = (+($19>>>0)); //@line 75 "snd_flt.cpp"
 $21 = $20; //@line 78 "snd_flt.cpp"
 $22 = 6.2831853071600001 / $21; //@line 78 "snd_flt.cpp"
 $23 = $22; //@line 78 "snd_flt.cpp"
 $24 = $23 * $$0$i; //@line 78 "snd_flt.cpp"
 $25 = ((($pChn)) + 239|0); //@line 79 "snd_flt.cpp"
 $26 = HEAP8[$25>>0]|0; //@line 79 "snd_flt.cpp"
 $27 = (+($26&255)); //@line 79 "snd_flt.cpp"
 $28 = $27 * -0.0093750003725290298; //@line 79 "snd_flt.cpp"
 $29 = $28; //@line 79 "snd_flt.cpp"
 $30 = (+Math_pow(10.0,(+$29))); //@line 79 "snd_flt.cpp"
 $31 = $30; //@line 79 "snd_flt.cpp"
 $32 = $31 * 2.0; //@line 80 "snd_flt.cpp"
 $33 = 1.0 - $32; //@line 80 "snd_flt.cpp"
 $34 = $33 * $24; //@line 80 "snd_flt.cpp"
 $35 = $34 > 2.0; //@line 81 "snd_flt.cpp"
 $d$0 = $35 ? 2.0 : $34; //@line 81 "snd_flt.cpp"
 $36 = $32 - $d$0; //@line 82 "snd_flt.cpp"
 $37 = $36 / $24; //@line 82 "snd_flt.cpp"
 $38 = 1.0 / $24; //@line 83 "snd_flt.cpp"
 $39 = $38 * $38; //@line 83 "snd_flt.cpp"
 $40 = $37 + 1.0; //@line 85 "snd_flt.cpp"
 $41 = $40 + $39; //@line 85 "snd_flt.cpp"
 $factor = $39 * 2.0;
 $42 = $factor + $37; //@line 86 "snd_flt.cpp"
 $43 = $42 / $41; //@line 86 "snd_flt.cpp"
 $44 = -$39; //@line 87 "snd_flt.cpp"
 $45 = $44 / $41; //@line 87 "snd_flt.cpp"
 $46 = 8192.0 / $41; //@line 89 "snd_flt.cpp"
 $47 = (~~(($46))); //@line 89 "snd_flt.cpp"
 $48 = ((($pChn)) + 72|0); //@line 89 "snd_flt.cpp"
 HEAP32[$48>>2] = $47; //@line 89 "snd_flt.cpp"
 $49 = $43 * 8192.0; //@line 90 "snd_flt.cpp"
 $50 = (~~(($49))); //@line 90 "snd_flt.cpp"
 $51 = ((($pChn)) + 76|0); //@line 90 "snd_flt.cpp"
 HEAP32[$51>>2] = $50; //@line 90 "snd_flt.cpp"
 $52 = $45 * 8192.0; //@line 91 "snd_flt.cpp"
 $53 = (~~(($52))); //@line 91 "snd_flt.cpp"
 $54 = ((($pChn)) + 80|0); //@line 91 "snd_flt.cpp"
 HEAP32[$54>>2] = $53; //@line 91 "snd_flt.cpp"
 if (!($bReset)) {
  $56 = ((($pChn)) + 36|0); //@line 98 "snd_flt.cpp"
  $57 = HEAP32[$56>>2]|0; //@line 98 "snd_flt.cpp"
  $58 = $57 | 16384; //@line 98 "snd_flt.cpp"
  HEAP32[$56>>2] = $58; //@line 98 "snd_flt.cpp"
  return; //@line 99 "snd_flt.cpp"
 }
 $55 = ((($pChn)) + 56|0); //@line 95 "snd_flt.cpp"
 ;HEAP32[$55>>2]=0|0;HEAP32[$55+4>>2]=0|0;HEAP32[$55+8>>2]=0|0;HEAP32[$55+12>>2]=0|0; //@line 95 "snd_flt.cpp"
 $56 = ((($pChn)) + 36|0); //@line 98 "snd_flt.cpp"
 $57 = HEAP32[$56>>2]|0; //@line 98 "snd_flt.cpp"
 $58 = $57 | 16384; //@line 98 "snd_flt.cpp"
 HEAP32[$56>>2] = $58; //@line 98 "snd_flt.cpp"
 return; //@line 99 "snd_flt.cpp"
}
function __ZN10CSoundFile14TonePortamentoEP11_MODCHANNELj($this,$pChn,$param) {
 $this = $this|0;
 $pChn = $pChn|0;
 $param = $param|0;
 var $$ = 0, $$1 = 0, $$2 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$pre = 0, $$pre6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $delta$0 = 0, $delta1$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($param|0)==(0); //@line 1404 "snd_fx.cpp"
 if (!($0)) {
  $1 = $param << 2; //@line 1404 "snd_fx.cpp"
  $2 = ((($pChn)) + 184|0); //@line 1404 "snd_fx.cpp"
  HEAP32[$2>>2] = $1; //@line 1404 "snd_fx.cpp"
 }
 $3 = ((($pChn)) + 36|0); //@line 1405 "snd_fx.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 1405 "snd_fx.cpp"
 $5 = $4 | 524288; //@line 1405 "snd_fx.cpp"
 HEAP32[$3>>2] = $5; //@line 1405 "snd_fx.cpp"
 $6 = ((($pChn)) + 128|0); //@line 1406 "snd_fx.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1406 "snd_fx.cpp"
 $8 = ($7|0)==(0); //@line 1406 "snd_fx.cpp"
 if ($8) {
  return; //@line 1435 "snd_fx.cpp"
 }
 $9 = ((($pChn)) + 136|0); //@line 1406 "snd_fx.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 1406 "snd_fx.cpp"
 $11 = ($10|0)==(0); //@line 1406 "snd_fx.cpp"
 if ($11) {
  return; //@line 1435 "snd_fx.cpp"
 }
 $12 = ((($this)) + 59660|0); //@line 1406 "snd_fx.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 1406 "snd_fx.cpp"
 $14 = $13 & 4096; //@line 1406 "snd_fx.cpp"
 $15 = ($14|0)==(0); //@line 1406 "snd_fx.cpp"
 if (!($15)) {
  return; //@line 1435 "snd_fx.cpp"
 }
 $16 = ($7|0)<($10|0); //@line 1408 "snd_fx.cpp"
 if ($16) {
  $17 = ((($pChn)) + 184|0); //@line 1410 "snd_fx.cpp"
  $18 = HEAP32[$17>>2]|0; //@line 1410 "snd_fx.cpp"
  $19 = $13 & 16; //@line 1411 "snd_fx.cpp"
  $20 = ($19|0)==(0); //@line 1411 "snd_fx.cpp"
  if ($20) {
   $34 = $7;$36 = $10;$delta$0 = $18;
  } else {
   $21 = ((($this)) + 59680|0); //@line 1411 "snd_fx.cpp"
   $22 = HEAP32[$21>>2]|0; //@line 1411 "snd_fx.cpp"
   $23 = $22 & 1048580; //@line 1411 "snd_fx.cpp"
   $24 = ($23|0)==(0); //@line 1411 "snd_fx.cpp"
   if ($24) {
    $25 = $18 >> 2; //@line 1413 "snd_fx.cpp"
    $26 = ($25>>>0)>(255); //@line 1414 "snd_fx.cpp"
    $$ = $26 ? 255 : $25; //@line 1414 "snd_fx.cpp"
    $27 = (40660 + ($$<<2)|0); //@line 1415 "snd_fx.cpp"
    $28 = HEAP32[$27>>2]|0; //@line 1415 "snd_fx.cpp"
    $29 = (__Z8_muldivrlll($7,$28,65536)|0); //@line 1415 "snd_fx.cpp"
    $30 = HEAP32[$6>>2]|0; //@line 1415 "snd_fx.cpp"
    $31 = (($29) - ($30))|0; //@line 1415 "snd_fx.cpp"
    $32 = ($31|0)<(1); //@line 1416 "snd_fx.cpp"
    $$2 = $32 ? 1 : $31; //@line 1416 "snd_fx.cpp"
    $$pre6 = HEAP32[$9>>2]|0; //@line 1419 "snd_fx.cpp"
    $34 = $30;$36 = $$pre6;$delta$0 = $$2;
   } else {
    $34 = $7;$36 = $10;$delta$0 = $18;
   }
  }
  $33 = (($34) + ($delta$0))|0; //@line 1418 "snd_fx.cpp"
  $35 = ($33|0)>($36|0); //@line 1419 "snd_fx.cpp"
  $$4 = $35 ? $36 : $33; //@line 1419 "snd_fx.cpp"
  HEAP32[$6>>2] = $$4; //@line 1419 "snd_fx.cpp"
  return; //@line 1435 "snd_fx.cpp"
 }
 $37 = ($7|0)>($10|0); //@line 1421 "snd_fx.cpp"
 if (!($37)) {
  return; //@line 1435 "snd_fx.cpp"
 }
 $38 = ((($pChn)) + 184|0); //@line 1423 "snd_fx.cpp"
 $39 = HEAP32[$38>>2]|0; //@line 1423 "snd_fx.cpp"
 $40 = (0 - ($39))|0; //@line 1423 "snd_fx.cpp"
 $41 = $13 & 16; //@line 1424 "snd_fx.cpp"
 $42 = ($41|0)==(0); //@line 1424 "snd_fx.cpp"
 if ($42) {
  $56 = $7;$58 = $10;$delta1$0 = $40;
  $55 = (($56) + ($delta1$0))|0; //@line 1431 "snd_fx.cpp"
  $57 = ($55|0)<($58|0); //@line 1432 "snd_fx.cpp"
  $$5 = $57 ? $58 : $55; //@line 1432 "snd_fx.cpp"
  HEAP32[$6>>2] = $$5; //@line 1432 "snd_fx.cpp"
  return; //@line 1432 "snd_fx.cpp"
 }
 $43 = ((($this)) + 59680|0); //@line 1424 "snd_fx.cpp"
 $44 = HEAP32[$43>>2]|0; //@line 1424 "snd_fx.cpp"
 $45 = $44 & 1048580; //@line 1424 "snd_fx.cpp"
 $46 = ($45|0)==(0); //@line 1424 "snd_fx.cpp"
 if (!($46)) {
  $56 = $7;$58 = $10;$delta1$0 = $40;
  $55 = (($56) + ($delta1$0))|0; //@line 1431 "snd_fx.cpp"
  $57 = ($55|0)<($58|0); //@line 1432 "snd_fx.cpp"
  $$5 = $57 ? $58 : $55; //@line 1432 "snd_fx.cpp"
  HEAP32[$6>>2] = $$5; //@line 1432 "snd_fx.cpp"
  return; //@line 1432 "snd_fx.cpp"
 }
 $47 = $39 >> 2; //@line 1426 "snd_fx.cpp"
 $48 = ($47>>>0)>(255); //@line 1427 "snd_fx.cpp"
 $$1 = $48 ? 255 : $47; //@line 1427 "snd_fx.cpp"
 $49 = (41684 + ($$1<<2)|0); //@line 1428 "snd_fx.cpp"
 $50 = HEAP32[$49>>2]|0; //@line 1428 "snd_fx.cpp"
 $51 = (__Z8_muldivrlll($7,$50,65536)|0); //@line 1428 "snd_fx.cpp"
 $52 = HEAP32[$6>>2]|0; //@line 1428 "snd_fx.cpp"
 $53 = (($51) - ($52))|0; //@line 1428 "snd_fx.cpp"
 $54 = ($53|0)>(-1); //@line 1429 "snd_fx.cpp"
 $$3 = $54 ? -1 : $53; //@line 1429 "snd_fx.cpp"
 $$pre = HEAP32[$9>>2]|0; //@line 1432 "snd_fx.cpp"
 $56 = $52;$58 = $$pre;$delta1$0 = $$3;
 $55 = (($56) + ($delta1$0))|0; //@line 1431 "snd_fx.cpp"
 $57 = ($55|0)<($58|0); //@line 1432 "snd_fx.cpp"
 $$5 = $57 ? $58 : $55; //@line 1432 "snd_fx.cpp"
 HEAP32[$6>>2] = $$5; //@line 1432 "snd_fx.cpp"
 return; //@line 1432 "snd_fx.cpp"
}
function __ZN10CSoundFile11VolumeSlideEP11_MODCHANNELj($this,$pChn,$param) {
 $this = $this|0;
 $pChn = $pChn|0;
 $param = $param|0;
 var $$ = 0, $$$i = 0, $$$i8 = 0, $$0 = 0, $$0$i = 0, $$6 = 0, $$newvolume$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $newvolume$0 = 0, $newvolume$1 = 0, $newvolume$2 = 0, $newvolume$4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($param|0)==(0); //@line 1468 "snd_fx.cpp"
 if ($0) {
  $3 = ((($pChn)) + 218|0); //@line 1468 "snd_fx.cpp"
  $4 = HEAP8[$3>>0]|0; //@line 1468 "snd_fx.cpp"
  $5 = $4&255; //@line 1468 "snd_fx.cpp"
  $$0 = $5;
 } else {
  $1 = $param&255; //@line 1468 "snd_fx.cpp"
  $2 = ((($pChn)) + 218|0); //@line 1468 "snd_fx.cpp"
  HEAP8[$2>>0] = $1; //@line 1468 "snd_fx.cpp"
  $$0 = $param;
 }
 $6 = ((($pChn)) + 116|0); //@line 1469 "snd_fx.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1469 "snd_fx.cpp"
 $8 = ((($this)) + 59680|0); //@line 1470 "snd_fx.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1470 "snd_fx.cpp"
 $10 = $9 & 2338; //@line 1470 "snd_fx.cpp"
 $11 = ($10|0)==(0); //@line 1470 "snd_fx.cpp"
 do {
  if ($11) {
   $newvolume$0 = $7;
  } else {
   $12 = $$0 & 15; //@line 1472 "snd_fx.cpp"
   $13 = ($12|0)==(15); //@line 1472 "snd_fx.cpp"
   $14 = $$0 & 240; //@line 1474 "snd_fx.cpp"
   if (!($13)) {
    $40 = ($14|0)==(240); //@line 1486 "snd_fx.cpp"
    if (!($40)) {
     $newvolume$0 = $7;
     break;
    }
    $41 = ($12|0)==(0); //@line 1488 "snd_fx.cpp"
    if ($41) {
     $56 = ((($this)) + 59660|0); //@line 1494 "snd_fx.cpp"
     $57 = HEAP32[$56>>2]|0; //@line 1494 "snd_fx.cpp"
     $58 = $57 & 4098; //@line 1494 "snd_fx.cpp"
     $59 = ($58|0)==(4096); //@line 1494 "snd_fx.cpp"
     $60 = (($7) + 60)|0; //@line 1496 "snd_fx.cpp"
     $$6 = $59 ? $60 : $7; //@line 1494 "snd_fx.cpp"
     $newvolume$0 = $$6;
     break;
    }
    $42 = $12&255; //@line 1576 "snd_fx.cpp"
    $43 = ((($pChn)) + 219|0); //@line 1576 "snd_fx.cpp"
    HEAP8[$43>>0] = $42; //@line 1576 "snd_fx.cpp"
    $44 = ((($this)) + 59660|0); //@line 1577 "snd_fx.cpp"
    $45 = HEAP32[$44>>2]|0; //@line 1577 "snd_fx.cpp"
    $46 = $45 & 4096; //@line 1577 "snd_fx.cpp"
    $47 = ($46|0)==(0); //@line 1577 "snd_fx.cpp"
    if ($47) {
     return; //@line 1510 "snd_fx.cpp"
    }
    $48 = $12 << 2; //@line 1579 "snd_fx.cpp"
    $49 = (($7) - ($48))|0; //@line 1579 "snd_fx.cpp"
    $50 = ($49|0)<(0); //@line 1580 "snd_fx.cpp"
    $$$i8 = $50 ? 0 : $49; //@line 1580 "snd_fx.cpp"
    HEAP32[$6>>2] = $$$i8; //@line 1579 "snd_fx.cpp"
    $51 = $9 & 1; //@line 1581 "snd_fx.cpp"
    $52 = ($51|0)==(0); //@line 1581 "snd_fx.cpp"
    if ($52) {
     return; //@line 1510 "snd_fx.cpp"
    }
    $53 = ((($pChn)) + 36|0); //@line 1581 "snd_fx.cpp"
    $54 = HEAP32[$53>>2]|0; //@line 1581 "snd_fx.cpp"
    $55 = $54 | 16777216; //@line 1581 "snd_fx.cpp"
    HEAP32[$53>>2] = $55; //@line 1581 "snd_fx.cpp"
    return; //@line 1510 "snd_fx.cpp"
   }
   $15 = ($14|0)==(0); //@line 1474 "snd_fx.cpp"
   if ($15) {
    $35 = ((($this)) + 59660|0); //@line 1480 "snd_fx.cpp"
    $36 = HEAP32[$35>>2]|0; //@line 1480 "snd_fx.cpp"
    $37 = $36 & 4098; //@line 1480 "snd_fx.cpp"
    $38 = ($37|0)==(4096); //@line 1480 "snd_fx.cpp"
    $39 = (($7) + -60)|0; //@line 1482 "snd_fx.cpp"
    $$ = $38 ? $39 : $7; //@line 1480 "snd_fx.cpp"
    $newvolume$0 = $$;
    break;
   }
   $16 = $$0 >>> 4; //@line 1476 "snd_fx.cpp"
   $17 = ($16|0)==(0); //@line 1563 "snd_fx.cpp"
   if ($17) {
    $20 = ((($pChn)) + 219|0); //@line 1563 "snd_fx.cpp"
    $21 = HEAP8[$20>>0]|0; //@line 1563 "snd_fx.cpp"
    $22 = $21&255; //@line 1563 "snd_fx.cpp"
    $$0$i = $22;
   } else {
    $18 = $16&255; //@line 1563 "snd_fx.cpp"
    $19 = ((($pChn)) + 219|0); //@line 1563 "snd_fx.cpp"
    HEAP8[$19>>0] = $18; //@line 1563 "snd_fx.cpp"
    $$0$i = $16;
   }
   $23 = ((($this)) + 59660|0); //@line 1564 "snd_fx.cpp"
   $24 = HEAP32[$23>>2]|0; //@line 1564 "snd_fx.cpp"
   $25 = $24 & 4096; //@line 1564 "snd_fx.cpp"
   $26 = ($25|0)==(0); //@line 1564 "snd_fx.cpp"
   if ($26) {
    return; //@line 1510 "snd_fx.cpp"
   }
   $27 = $$0$i << 2; //@line 1566 "snd_fx.cpp"
   $28 = (($7) + ($27))|0; //@line 1566 "snd_fx.cpp"
   $29 = ($28|0)>(256); //@line 1567 "snd_fx.cpp"
   $$$i = $29 ? 256 : $28; //@line 1567 "snd_fx.cpp"
   HEAP32[$6>>2] = $$$i; //@line 1566 "snd_fx.cpp"
   $30 = $9 & 1; //@line 1568 "snd_fx.cpp"
   $31 = ($30|0)==(0); //@line 1568 "snd_fx.cpp"
   if ($31) {
    return; //@line 1510 "snd_fx.cpp"
   }
   $32 = ((($pChn)) + 36|0); //@line 1568 "snd_fx.cpp"
   $33 = HEAP32[$32>>2]|0; //@line 1568 "snd_fx.cpp"
   $34 = $33 | 16777216; //@line 1568 "snd_fx.cpp"
   HEAP32[$32>>2] = $34; //@line 1568 "snd_fx.cpp"
   return; //@line 1510 "snd_fx.cpp"
  }
 } while(0);
 $61 = ((($this)) + 59660|0); //@line 1501 "snd_fx.cpp"
 $62 = HEAP32[$61>>2]|0; //@line 1501 "snd_fx.cpp"
 $63 = $62 & 4098; //@line 1501 "snd_fx.cpp"
 $64 = ($63|0)==(4096); //@line 1501 "snd_fx.cpp"
 if ($64) {
  $newvolume$2 = $newvolume$0;
 } else {
  $65 = $$0 & 15; //@line 1503 "snd_fx.cpp"
  $66 = ($65|0)==(0); //@line 1503 "snd_fx.cpp"
  if ($66) {
   $69 = $$0 >>> 2; //@line 1504 "snd_fx.cpp"
   $70 = $69 & 60; //@line 1504 "snd_fx.cpp"
   $71 = (($newvolume$0) + ($70))|0; //@line 1504 "snd_fx.cpp"
   $newvolume$1 = $71;
  } else {
   $67 = $65 << 2; //@line 1503 "snd_fx.cpp"
   $68 = (($newvolume$0) - ($67))|0; //@line 1503 "snd_fx.cpp"
   $newvolume$1 = $68;
  }
  $72 = $9 & 1; //@line 1505 "snd_fx.cpp"
  $73 = ($72|0)==(0); //@line 1505 "snd_fx.cpp"
  if ($73) {
   $newvolume$2 = $newvolume$1;
  } else {
   $74 = ((($pChn)) + 36|0); //@line 1505 "snd_fx.cpp"
   $75 = HEAP32[$74>>2]|0; //@line 1505 "snd_fx.cpp"
   $76 = $75 | 16777216; //@line 1505 "snd_fx.cpp"
   HEAP32[$74>>2] = $76; //@line 1505 "snd_fx.cpp"
   $newvolume$2 = $newvolume$1;
  }
 }
 $77 = ($newvolume$2|0)<(0); //@line 1507 "snd_fx.cpp"
 $$newvolume$2 = $77 ? 0 : $newvolume$2; //@line 1507 "snd_fx.cpp"
 $78 = ($$newvolume$2|0)>(256); //@line 1508 "snd_fx.cpp"
 $newvolume$4 = $78 ? 256 : $$newvolume$2; //@line 1508 "snd_fx.cpp"
 HEAP32[$6>>2] = $newvolume$4; //@line 1509 "snd_fx.cpp"
 return; //@line 1510 "snd_fx.cpp"
}
function __ZN10CSoundFile12PortamentoUpEP11_MODCHANNELj($this,$pChn,$param) {
 $this = $this|0;
 $pChn = $pChn|0;
 $param = $param|0;
 var $$$i = 0, $$0 = 0, $$1$i = 0, $$not$i = 0, $$not$i3 = 0, $$storemerge$i = 0, $$storemerge$i6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $9 = 0, $brmerge$i = 0, $brmerge$i4 = 0, $or$cond = 0, $storemerge$i = 0, $storemerge$i5 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($param|0)==(0); //@line 1255 "snd_fx.cpp"
 if ($0) {
  $3 = ((($pChn)) + 220|0); //@line 1255 "snd_fx.cpp"
  $4 = HEAP8[$3>>0]|0; //@line 1255 "snd_fx.cpp"
  $5 = $4&255; //@line 1255 "snd_fx.cpp"
  $$0 = $5;
 } else {
  $1 = $param&255; //@line 1255 "snd_fx.cpp"
  $2 = ((($pChn)) + 220|0); //@line 1255 "snd_fx.cpp"
  HEAP8[$2>>0] = $1; //@line 1255 "snd_fx.cpp"
  $$0 = $param;
 }
 $6 = ((($this)) + 59680|0); //@line 1256 "snd_fx.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1256 "snd_fx.cpp"
 $8 = $7 & 290; //@line 1256 "snd_fx.cpp"
 $9 = ($8|0)==(0); //@line 1256 "snd_fx.cpp"
 if (!($9)) {
  $10 = $$0 & 240; //@line 1256 "snd_fx.cpp"
  $11 = ($10>>>0)>(223); //@line 1256 "snd_fx.cpp"
  if ($11) {
   $12 = $$0 & 15; //@line 1258 "snd_fx.cpp"
   $13 = ($12|0)==(0); //@line 1258 "snd_fx.cpp"
   if ($13) {
    return; //@line 1276 "snd_fx.cpp"
   }
   switch ($10|0) {
   case 240:  {
    $14 = $7 & 1048580; //@line 1307 "snd_fx.cpp"
    $15 = ($14|0)==(0); //@line 1307 "snd_fx.cpp"
    if (!($15)) {
     $16 = $12&255; //@line 1309 "snd_fx.cpp"
     $17 = ((($pChn)) + 221|0); //@line 1309 "snd_fx.cpp"
     HEAP8[$17>>0] = $16; //@line 1309 "snd_fx.cpp"
    }
    $18 = ((($this)) + 59660|0); //@line 1311 "snd_fx.cpp"
    $19 = HEAP32[$18>>2]|0; //@line 1311 "snd_fx.cpp"
    $20 = $19 & 4096; //@line 1311 "snd_fx.cpp"
    $21 = ($20|0)==(0); //@line 1311 "snd_fx.cpp"
    if ($21) {
     return; //@line 1276 "snd_fx.cpp"
    }
    $22 = ((($pChn)) + 128|0); //@line 1313 "snd_fx.cpp"
    $23 = HEAP32[$22>>2]|0; //@line 1313 "snd_fx.cpp"
    $24 = ($23|0)==(0); //@line 1313 "snd_fx.cpp"
    if ($24) {
     return; //@line 1276 "snd_fx.cpp"
    }
    $25 = $19 & 16; //@line 1315 "snd_fx.cpp"
    $26 = ($25|0)==(0); //@line 1315 "snd_fx.cpp"
    $$not$i = $15 ^ 1; //@line 1315 "snd_fx.cpp"
    $brmerge$i = $26 | $$not$i; //@line 1315 "snd_fx.cpp"
    if ($brmerge$i) {
     $30 = $12 << 2; //@line 1320 "snd_fx.cpp"
     $31 = (($23) - ($30))|0; //@line 1320 "snd_fx.cpp"
     $storemerge$i = $31;
    } else {
     $27 = (41684 + ($12<<2)|0); //@line 1317 "snd_fx.cpp"
     $28 = HEAP32[$27>>2]|0; //@line 1317 "snd_fx.cpp"
     $29 = (__Z8_muldivrlll($23,$28,65536)|0); //@line 1317 "snd_fx.cpp"
     $storemerge$i = $29;
    }
    $32 = ($storemerge$i|0)<(1); //@line 1322 "snd_fx.cpp"
    $$storemerge$i = $32 ? 1 : $storemerge$i; //@line 1322 "snd_fx.cpp"
    HEAP32[$22>>2] = $$storemerge$i; //@line 1322 "snd_fx.cpp"
    return; //@line 1276 "snd_fx.cpp"
    break;
   }
   case 224:  {
    $33 = $7 & 1048580; //@line 1355 "snd_fx.cpp"
    $34 = ($33|0)==(0); //@line 1355 "snd_fx.cpp"
    if (!($34)) {
     $35 = $12&255; //@line 1357 "snd_fx.cpp"
     $36 = ((($pChn)) + 221|0); //@line 1357 "snd_fx.cpp"
     HEAP8[$36>>0] = $35; //@line 1357 "snd_fx.cpp"
    }
    $37 = ((($this)) + 59660|0); //@line 1359 "snd_fx.cpp"
    $38 = HEAP32[$37>>2]|0; //@line 1359 "snd_fx.cpp"
    $39 = $38 & 4096; //@line 1359 "snd_fx.cpp"
    $40 = ($39|0)==(0); //@line 1359 "snd_fx.cpp"
    if ($40) {
     return; //@line 1276 "snd_fx.cpp"
    }
    $41 = ((($pChn)) + 128|0); //@line 1361 "snd_fx.cpp"
    $42 = HEAP32[$41>>2]|0; //@line 1361 "snd_fx.cpp"
    $43 = ($42|0)==(0); //@line 1361 "snd_fx.cpp"
    if ($43) {
     return; //@line 1276 "snd_fx.cpp"
    }
    $44 = $38 & 16; //@line 1363 "snd_fx.cpp"
    $45 = ($44|0)==(0); //@line 1363 "snd_fx.cpp"
    $$not$i3 = $34 ^ 1; //@line 1363 "snd_fx.cpp"
    $brmerge$i4 = $45 | $$not$i3; //@line 1363 "snd_fx.cpp"
    if ($brmerge$i4) {
     $49 = (($42) - ($12))|0; //@line 1368 "snd_fx.cpp"
     $storemerge$i5 = $49;
    } else {
     $46 = (42708 + ($12<<2)|0); //@line 1365 "snd_fx.cpp"
     $47 = HEAP32[$46>>2]|0; //@line 1365 "snd_fx.cpp"
     $48 = (__Z8_muldivrlll($42,$47,65536)|0); //@line 1365 "snd_fx.cpp"
     $storemerge$i5 = $48;
    }
    $50 = ($storemerge$i5|0)<(1); //@line 1370 "snd_fx.cpp"
    $$storemerge$i6 = $50 ? 1 : $storemerge$i5; //@line 1370 "snd_fx.cpp"
    HEAP32[$41>>2] = $$storemerge$i6; //@line 1370 "snd_fx.cpp"
    return; //@line 1276 "snd_fx.cpp"
    break;
   }
   default: {
    return; //@line 1276 "snd_fx.cpp"
   }
   }
  }
 }
 $51 = ((($this)) + 59660|0); //@line 1272 "snd_fx.cpp"
 $52 = HEAP32[$51>>2]|0; //@line 1272 "snd_fx.cpp"
 $53 = $52 & 4096; //@line 1272 "snd_fx.cpp"
 $54 = ($53|0)==(0); //@line 1272 "snd_fx.cpp"
 if (!($54)) {
  $55 = ((($this)) + 59708|0); //@line 1272 "snd_fx.cpp"
  $56 = HEAP32[$55>>2]|0; //@line 1272 "snd_fx.cpp"
  $57 = ($56|0)==(1); //@line 1272 "snd_fx.cpp"
  if (!($57)) {
   return; //@line 1276 "snd_fx.cpp"
  }
 }
 $58 = $$0 << 2; //@line 1274 "snd_fx.cpp"
 $59 = (0 - ($58))|0; //@line 1274 "snd_fx.cpp"
 $60 = ((($pChn)) + 128|0); //@line 1976 "snd_fx.cpp"
 $61 = HEAP32[$60>>2]|0; //@line 1976 "snd_fx.cpp"
 $62 = ($61|0)==(0); //@line 1976 "snd_fx.cpp"
 if ($62) {
  return; //@line 1276 "snd_fx.cpp"
 }
 $63 = $52 & 16; //@line 1977 "snd_fx.cpp"
 $64 = ($63|0)!=(0); //@line 1977 "snd_fx.cpp"
 $65 = $7 & 1048580;
 $66 = ($65|0)==(0); //@line 1977 "snd_fx.cpp"
 $or$cond = $64 & $66; //@line 1977 "snd_fx.cpp"
 do {
  if ($or$cond) {
   $67 = ($58|0)>(0); //@line 1979 "snd_fx.cpp"
   if ($67) {
    $68 = $58 >> 2; //@line 1981 "snd_fx.cpp"
    $69 = ($68>>>0)>(255); //@line 1982 "snd_fx.cpp"
    $$$i = $69 ? 255 : $68; //@line 1982 "snd_fx.cpp"
    $70 = (41684 + ($$$i<<2)|0); //@line 1983 "snd_fx.cpp"
    $71 = HEAP32[$70>>2]|0; //@line 1983 "snd_fx.cpp"
    $72 = (__Z8_muldivrlll($61,$71,65536)|0); //@line 1983 "snd_fx.cpp"
    HEAP32[$60>>2] = $72; //@line 1983 "snd_fx.cpp"
    $80 = $72;
    break;
   } else {
    $73 = $59 >> 2; //@line 1986 "snd_fx.cpp"
    $74 = ($73>>>0)>(255); //@line 1988 "snd_fx.cpp"
    $$1$i = $74 ? 255 : $73; //@line 1988 "snd_fx.cpp"
    $75 = (40660 + ($$1$i<<2)|0); //@line 1989 "snd_fx.cpp"
    $76 = HEAP32[$75>>2]|0; //@line 1989 "snd_fx.cpp"
    $77 = (__Z8_muldivrlll($61,$76,65536)|0); //@line 1989 "snd_fx.cpp"
    HEAP32[$60>>2] = $77; //@line 1989 "snd_fx.cpp"
    $80 = $77;
    break;
   }
  } else {
   $78 = (($61) - ($58))|0; //@line 1993 "snd_fx.cpp"
   HEAP32[$60>>2] = $78; //@line 1993 "snd_fx.cpp"
   $80 = $78;
  }
 } while(0);
 $79 = ($80|0)<(1); //@line 1995 "snd_fx.cpp"
 if (!($79)) {
  return; //@line 1276 "snd_fx.cpp"
 }
 HEAP32[$60>>2] = 1; //@line 1997 "snd_fx.cpp"
 $81 = HEAP32[$6>>2]|0; //@line 1998 "snd_fx.cpp"
 $82 = $81 & 32; //@line 1998 "snd_fx.cpp"
 $83 = ($82|0)==(0); //@line 1998 "snd_fx.cpp"
 if ($83) {
  return; //@line 1276 "snd_fx.cpp"
 }
 $84 = ((($pChn)) + 36|0); //@line 2000 "snd_fx.cpp"
 $85 = HEAP32[$84>>2]|0; //@line 2000 "snd_fx.cpp"
 $86 = $85 | 1024; //@line 2000 "snd_fx.cpp"
 HEAP32[$84>>2] = $86; //@line 2000 "snd_fx.cpp"
 $87 = ((($pChn)) + 124|0); //@line 2001 "snd_fx.cpp"
 HEAP32[$87>>2] = 0; //@line 2001 "snd_fx.cpp"
 return; //@line 1276 "snd_fx.cpp"
}
function __ZN10CSoundFile14PortamentoDownEP11_MODCHANNELj($this,$pChn,$param) {
 $this = $this|0;
 $pChn = $pChn|0;
 $param = $param|0;
 var $$$i = 0, $$0 = 0, $$1$i = 0, $$not$i = 0, $$not$i3 = 0, $$storemerge$i = 0, $$storemerge$i6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $9 = 0, $brmerge$i = 0, $brmerge$i4 = 0, $or$cond = 0, $storemerge$i = 0, $storemerge$i5 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($param|0)==(0); //@line 1282 "snd_fx.cpp"
 if ($0) {
  $3 = ((($pChn)) + 220|0); //@line 1282 "snd_fx.cpp"
  $4 = HEAP8[$3>>0]|0; //@line 1282 "snd_fx.cpp"
  $5 = $4&255; //@line 1282 "snd_fx.cpp"
  $$0 = $5;
 } else {
  $1 = $param&255; //@line 1282 "snd_fx.cpp"
  $2 = ((($pChn)) + 220|0); //@line 1282 "snd_fx.cpp"
  HEAP8[$2>>0] = $1; //@line 1282 "snd_fx.cpp"
  $$0 = $param;
 }
 $6 = ((($this)) + 59680|0); //@line 1283 "snd_fx.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1283 "snd_fx.cpp"
 $8 = $7 & 290; //@line 1283 "snd_fx.cpp"
 $9 = ($8|0)==(0); //@line 1283 "snd_fx.cpp"
 if (!($9)) {
  $10 = $$0 & 240; //@line 1283 "snd_fx.cpp"
  $11 = ($10>>>0)>(223); //@line 1283 "snd_fx.cpp"
  if ($11) {
   $12 = $$0 & 15; //@line 1285 "snd_fx.cpp"
   $13 = ($12|0)==(0); //@line 1285 "snd_fx.cpp"
   if ($13) {
    return; //@line 1301 "snd_fx.cpp"
   }
   switch ($10|0) {
   case 240:  {
    $14 = $7 & 1048580; //@line 1331 "snd_fx.cpp"
    $15 = ($14|0)==(0); //@line 1331 "snd_fx.cpp"
    if (!($15)) {
     $16 = $12&255; //@line 1333 "snd_fx.cpp"
     $17 = ((($pChn)) + 221|0); //@line 1333 "snd_fx.cpp"
     HEAP8[$17>>0] = $16; //@line 1333 "snd_fx.cpp"
    }
    $18 = ((($this)) + 59660|0); //@line 1335 "snd_fx.cpp"
    $19 = HEAP32[$18>>2]|0; //@line 1335 "snd_fx.cpp"
    $20 = $19 & 4096; //@line 1335 "snd_fx.cpp"
    $21 = ($20|0)==(0); //@line 1335 "snd_fx.cpp"
    if ($21) {
     return; //@line 1301 "snd_fx.cpp"
    }
    $22 = ((($pChn)) + 128|0); //@line 1337 "snd_fx.cpp"
    $23 = HEAP32[$22>>2]|0; //@line 1337 "snd_fx.cpp"
    $24 = ($23|0)==(0); //@line 1337 "snd_fx.cpp"
    if ($24) {
     return; //@line 1301 "snd_fx.cpp"
    }
    $25 = $19 & 16; //@line 1339 "snd_fx.cpp"
    $26 = ($25|0)==(0); //@line 1339 "snd_fx.cpp"
    $$not$i = $15 ^ 1; //@line 1339 "snd_fx.cpp"
    $brmerge$i = $26 | $$not$i; //@line 1339 "snd_fx.cpp"
    if ($brmerge$i) {
     $30 = $12 << 2; //@line 1344 "snd_fx.cpp"
     $31 = (($23) + ($30))|0; //@line 1344 "snd_fx.cpp"
     $storemerge$i = $31;
    } else {
     $27 = (40660 + ($12<<2)|0); //@line 1341 "snd_fx.cpp"
     $28 = HEAP32[$27>>2]|0; //@line 1341 "snd_fx.cpp"
     $29 = (__Z8_muldivrlll($23,$28,65536)|0); //@line 1341 "snd_fx.cpp"
     $storemerge$i = $29;
    }
    $32 = ($storemerge$i|0)>(65535); //@line 1346 "snd_fx.cpp"
    $$storemerge$i = $32 ? 65535 : $storemerge$i; //@line 1346 "snd_fx.cpp"
    HEAP32[$22>>2] = $$storemerge$i; //@line 1346 "snd_fx.cpp"
    return; //@line 1301 "snd_fx.cpp"
    break;
   }
   case 224:  {
    $33 = $7 & 1048580; //@line 1379 "snd_fx.cpp"
    $34 = ($33|0)==(0); //@line 1379 "snd_fx.cpp"
    if (!($34)) {
     $35 = $12&255; //@line 1381 "snd_fx.cpp"
     $36 = ((($pChn)) + 221|0); //@line 1381 "snd_fx.cpp"
     HEAP8[$36>>0] = $35; //@line 1381 "snd_fx.cpp"
    }
    $37 = ((($this)) + 59660|0); //@line 1383 "snd_fx.cpp"
    $38 = HEAP32[$37>>2]|0; //@line 1383 "snd_fx.cpp"
    $39 = $38 & 4096; //@line 1383 "snd_fx.cpp"
    $40 = ($39|0)==(0); //@line 1383 "snd_fx.cpp"
    if ($40) {
     return; //@line 1301 "snd_fx.cpp"
    }
    $41 = ((($pChn)) + 128|0); //@line 1385 "snd_fx.cpp"
    $42 = HEAP32[$41>>2]|0; //@line 1385 "snd_fx.cpp"
    $43 = ($42|0)==(0); //@line 1385 "snd_fx.cpp"
    if ($43) {
     return; //@line 1301 "snd_fx.cpp"
    }
    $44 = $38 & 16; //@line 1387 "snd_fx.cpp"
    $45 = ($44|0)==(0); //@line 1387 "snd_fx.cpp"
    $$not$i3 = $34 ^ 1; //@line 1387 "snd_fx.cpp"
    $brmerge$i4 = $45 | $$not$i3; //@line 1387 "snd_fx.cpp"
    if ($brmerge$i4) {
     $49 = (($42) + ($12))|0; //@line 1392 "snd_fx.cpp"
     $storemerge$i5 = $49;
    } else {
     $46 = (42772 + ($12<<2)|0); //@line 1389 "snd_fx.cpp"
     $47 = HEAP32[$46>>2]|0; //@line 1389 "snd_fx.cpp"
     $48 = (__Z8_muldivrlll($42,$47,65536)|0); //@line 1389 "snd_fx.cpp"
     $storemerge$i5 = $48;
    }
    $50 = ($storemerge$i5|0)>(65535); //@line 1394 "snd_fx.cpp"
    $$storemerge$i6 = $50 ? 65535 : $storemerge$i5; //@line 1394 "snd_fx.cpp"
    HEAP32[$41>>2] = $$storemerge$i6; //@line 1394 "snd_fx.cpp"
    return; //@line 1301 "snd_fx.cpp"
    break;
   }
   default: {
    return; //@line 1301 "snd_fx.cpp"
   }
   }
  }
 }
 $51 = ((($this)) + 59660|0); //@line 1298 "snd_fx.cpp"
 $52 = HEAP32[$51>>2]|0; //@line 1298 "snd_fx.cpp"
 $53 = $52 & 4096; //@line 1298 "snd_fx.cpp"
 $54 = ($53|0)==(0); //@line 1298 "snd_fx.cpp"
 if (!($54)) {
  $55 = ((($this)) + 59708|0); //@line 1298 "snd_fx.cpp"
  $56 = HEAP32[$55>>2]|0; //@line 1298 "snd_fx.cpp"
  $57 = ($56|0)==(1); //@line 1298 "snd_fx.cpp"
  if (!($57)) {
   return; //@line 1301 "snd_fx.cpp"
  }
 }
 $58 = $$0 << 2; //@line 1299 "snd_fx.cpp"
 $59 = ((($pChn)) + 128|0); //@line 1976 "snd_fx.cpp"
 $60 = HEAP32[$59>>2]|0; //@line 1976 "snd_fx.cpp"
 $61 = ($60|0)==(0); //@line 1976 "snd_fx.cpp"
 if ($61) {
  return; //@line 1301 "snd_fx.cpp"
 }
 $62 = $52 & 16; //@line 1977 "snd_fx.cpp"
 $63 = ($62|0)!=(0); //@line 1977 "snd_fx.cpp"
 $64 = $7 & 1048580;
 $65 = ($64|0)==(0); //@line 1977 "snd_fx.cpp"
 $or$cond = $63 & $65; //@line 1977 "snd_fx.cpp"
 do {
  if ($or$cond) {
   $66 = ($58|0)<(0); //@line 1979 "snd_fx.cpp"
   if ($66) {
    $67 = (0 - ($58))|0; //@line 1981 "snd_fx.cpp"
    $68 = $67 >> 2; //@line 1981 "snd_fx.cpp"
    $69 = ($68>>>0)>(255); //@line 1982 "snd_fx.cpp"
    $$$i = $69 ? 255 : $68; //@line 1982 "snd_fx.cpp"
    $70 = (41684 + ($$$i<<2)|0); //@line 1983 "snd_fx.cpp"
    $71 = HEAP32[$70>>2]|0; //@line 1983 "snd_fx.cpp"
    $72 = (__Z8_muldivrlll($60,$71,65536)|0); //@line 1983 "snd_fx.cpp"
    HEAP32[$59>>2] = $72; //@line 1983 "snd_fx.cpp"
    $80 = $72;
    break;
   } else {
    $73 = $58 >> 2; //@line 1986 "snd_fx.cpp"
    $74 = ($73>>>0)>(255); //@line 1988 "snd_fx.cpp"
    $$1$i = $74 ? 255 : $73; //@line 1988 "snd_fx.cpp"
    $75 = (40660 + ($$1$i<<2)|0); //@line 1989 "snd_fx.cpp"
    $76 = HEAP32[$75>>2]|0; //@line 1989 "snd_fx.cpp"
    $77 = (__Z8_muldivrlll($60,$76,65536)|0); //@line 1989 "snd_fx.cpp"
    HEAP32[$59>>2] = $77; //@line 1989 "snd_fx.cpp"
    $80 = $77;
    break;
   }
  } else {
   $78 = (($60) + ($58))|0; //@line 1993 "snd_fx.cpp"
   HEAP32[$59>>2] = $78; //@line 1993 "snd_fx.cpp"
   $80 = $78;
  }
 } while(0);
 $79 = ($80|0)<(1); //@line 1995 "snd_fx.cpp"
 if (!($79)) {
  return; //@line 1301 "snd_fx.cpp"
 }
 HEAP32[$59>>2] = 1; //@line 1997 "snd_fx.cpp"
 $81 = HEAP32[$6>>2]|0; //@line 1998 "snd_fx.cpp"
 $82 = $81 & 32; //@line 1998 "snd_fx.cpp"
 $83 = ($82|0)==(0); //@line 1998 "snd_fx.cpp"
 if ($83) {
  return; //@line 1301 "snd_fx.cpp"
 }
 $84 = ((($pChn)) + 36|0); //@line 2000 "snd_fx.cpp"
 $85 = HEAP32[$84>>2]|0; //@line 2000 "snd_fx.cpp"
 $86 = $85 | 1024; //@line 2000 "snd_fx.cpp"
 HEAP32[$84>>2] = $86; //@line 2000 "snd_fx.cpp"
 $87 = ((($pChn)) + 124|0); //@line 2001 "snd_fx.cpp"
 HEAP32[$87>>2] = 0; //@line 2001 "snd_fx.cpp"
 return; //@line 1301 "snd_fx.cpp"
}
function __ZNK10CSoundFile14IsSongFinishedEjj($this,$nStartOrder,$nStartRow) {
 $this = $this|0;
 $nStartOrder = $nStartOrder|0;
 $nStartRow = $nStartRow|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cmd$0 = 0, $cmd$0$cmp = 0, $nOrd$014 = 0, $or$cond5 = 0, $pos$013 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($nStartOrder>>>0)<(256); //@line 2173 "snd_fx.cpp"
 if (!($0)) {
  $$0 = 255;
  return ($$0|0); //@line 2204 "snd_fx.cpp"
 }
 $1 = ((($this)) + 59664|0); //@line 2184 "snd_fx.cpp"
 $nOrd$014 = $nStartOrder;
 L4: while(1) {
  $2 = (((($this)) + 53344|0) + ($nOrd$014)|0); //@line 2175 "snd_fx.cpp"
  $3 = HEAP8[$2>>0]|0; //@line 2175 "snd_fx.cpp"
  $4 = $3&255; //@line 2175 "snd_fx.cpp"
  $5 = ($3<<24>>24)==(-2); //@line 2176 "snd_fx.cpp"
  if (!($5)) {
   $6 = ($3&255)>(239); //@line 2180 "snd_fx.cpp"
   if ($6) {
    $$0 = $nOrd$014;
    label = 14;
    break;
   }
   $7 = (((($this)) + 51904|0) + ($4<<2)|0); //@line 2181 "snd_fx.cpp"
   $8 = HEAP32[$7>>2]|0; //@line 2181 "snd_fx.cpp"
   $9 = ($8|0)==(0|0); //@line 2182 "snd_fx.cpp"
   if (!($9)) {
    $10 = (((($this)) + 52864|0) + ($4<<1)|0); //@line 2184 "snd_fx.cpp"
    $11 = HEAP16[$10>>1]|0; //@line 2184 "snd_fx.cpp"
    $12 = $11&65535; //@line 2184 "snd_fx.cpp"
    $13 = HEAP32[$1>>2]|0; //@line 2184 "snd_fx.cpp"
    $14 = Math_imul($12, $13)|0; //@line 2184 "snd_fx.cpp"
    $15 = ($nOrd$014|0)==($nStartOrder|0); //@line 2185 "snd_fx.cpp"
    $16 = $15 ? $nStartRow : 0; //@line 2185 "snd_fx.cpp"
    $17 = Math_imul($13, $16)|0; //@line 2186 "snd_fx.cpp"
    $18 = ($17>>>0)<($14>>>0); //@line 2187 "snd_fx.cpp"
    if ($18) {
     $pos$013 = $17;
     while(1) {
      $21 = (($8) + (($pos$013*6)|0)|0); //@line 2190 "snd_fx.cpp"
      $22 = HEAP8[$21>>0]|0; //@line 2190 "snd_fx.cpp"
      $23 = ($22<<24>>24)==(0); //@line 2190 "snd_fx.cpp"
      if (!($23)) {
       $$0 = 0;
       label = 14;
       break L4;
      }
      $24 = (((($8) + (($pos$013*6)|0)|0)) + 2|0); //@line 2190 "snd_fx.cpp"
      $25 = HEAP8[$24>>0]|0; //@line 2190 "snd_fx.cpp"
      $26 = ($25<<24>>24)==(0); //@line 2190 "snd_fx.cpp"
      if (!($26)) {
       $$0 = 0;
       label = 14;
       break L4;
      }
      $27 = (((($8) + (($pos$013*6)|0)|0)) + 3|0); //@line 2191 "snd_fx.cpp"
      $28 = HEAP8[$27>>0]|0; //@line 2191 "snd_fx.cpp"
      $29 = $28&255; //@line 2191 "snd_fx.cpp"
      $30 = ($28<<24>>24)==(19); //@line 2192 "snd_fx.cpp"
      if ($30) {
       $31 = (((($8) + (($pos$013*6)|0)|0)) + 5|0); //@line 2194 "snd_fx.cpp"
       $32 = HEAP8[$31>>0]|0; //@line 2194 "snd_fx.cpp"
       $33 = $32&255; //@line 2194 "snd_fx.cpp"
       $34 = $33 & 240; //@line 2194 "snd_fx.cpp"
       $35 = ($34|0)==(0); //@line 2195 "snd_fx.cpp"
       $36 = $33 & 112; //@line 2195 "snd_fx.cpp"
       $37 = ($36|0)==(96); //@line 2195 "snd_fx.cpp"
       $38 = $37 | $35; //@line 2195 "snd_fx.cpp"
       $39 = ($34|0)==(240); //@line 2195 "snd_fx.cpp"
       $or$cond5 = $39 | $38; //@line 2195 "snd_fx.cpp"
       $$ = $or$cond5 ? 0 : $29; //@line 2195 "snd_fx.cpp"
       $cmd$0 = $$;
      } else {
       $cmd$0 = $29;
      }
      $40 = ($cmd$0|0)!=(0); //@line 2197 "snd_fx.cpp"
      $41 = $cmd$0 & 254; //@line 2197 "snd_fx.cpp"
      $cmd$0$cmp = ($41|0)!=(16); //@line 2197 "snd_fx.cpp"
      $42 = $40 & $cmd$0$cmp; //@line 2197 "snd_fx.cpp"
      $20 = (($pos$013) + 1)|0; //@line 2198 "snd_fx.cpp"
      if ($42) {
       $$0 = 0;
       label = 14;
       break L4;
      }
      $19 = ($20>>>0)<($14>>>0); //@line 2187 "snd_fx.cpp"
      if ($19) {
       $pos$013 = $20;
      } else {
       break;
      }
     }
    }
   }
  }
  $43 = (($nOrd$014) + 1)|0; //@line 2173 "snd_fx.cpp"
  $44 = ($43>>>0)<(256); //@line 2173 "snd_fx.cpp"
  if ($44) {
   $nOrd$014 = $43;
  } else {
   $$0 = 255;
   label = 14;
   break;
  }
 }
 if ((label|0) == 14) {
  return ($$0|0); //@line 2204 "snd_fx.cpp"
 }
 return (0)|0;
}
function __ZN10CSoundFile10RetrigNoteEjj($this,$nChn,$param) {
 $this = $this|0;
 $nChn = $nChn|0;
 $param = $param|0;
 var $$ = 0, $$12 = 0, $$14 = 0, $$bResetEnv$0 = 0, $$nRetrigCount$1 = 0, $$off = 0, $$pre = 0, $$realspeed$0 = 0, $$realspeed$019 = 0, $$realspeed$020 = 0, $$vol$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $bDoRetrig$1 = 0, $bDoRetrig$2$off0 = 0, $bResetEnv$0 = 0, $bResetEnv$1$off0 = 0, $brmerge24 = 0, $nRetrigCount$0 = 0, $nRetrigCount$122 = 0, $nRetrigCount$2 = 0, $or$cond3 = 0, $or$cond5 = 0, $realspeed$0$ph = 0, $realspeed$2 = 0, $vol$0 = 0;
 var $vol$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($this) + ($nChn<<8)|0); //@line 1915 "snd_fx.cpp"
 $1 = $param & 15; //@line 1916 "snd_fx.cpp"
 $2 = (((($this) + ($nChn<<8)|0)) + 240|0); //@line 1917 "snd_fx.cpp"
 $3 = HEAP8[$2>>0]|0; //@line 1917 "snd_fx.cpp"
 $4 = $3&255; //@line 1917 "snd_fx.cpp"
 $5 = ((($this)) + 59680|0); //@line 1920 "snd_fx.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 1920 "snd_fx.cpp"
 $7 = $6 & 34; //@line 1920 "snd_fx.cpp"
 $8 = ($7|0)==(0); //@line 1920 "snd_fx.cpp"
 do {
  if ($8) {
   $15 = $param & 256; //@line 1928 "snd_fx.cpp"
   $16 = ($15|0)!=(0); //@line 1928 "snd_fx.cpp"
   do {
    if ($16) {
     $17 = (((($this) + ($nChn<<8)|0)) + 248|0); //@line 1928 "snd_fx.cpp"
     $18 = HEAP8[$17>>0]|0; //@line 1928 "snd_fx.cpp"
     $19 = ($18<<24>>24)==(1); //@line 1928 "snd_fx.cpp"
     if ($19) {
      $20 = (((($this) + ($nChn<<8)|0)) + 251|0); //@line 1928 "snd_fx.cpp"
      $21 = HEAP8[$20>>0]|0; //@line 1928 "snd_fx.cpp"
      $22 = ($21&255)>(15); //@line 1928 "snd_fx.cpp"
      $23 = $22&1; //@line 1928 "snd_fx.cpp"
      $$12 = (($23) + ($1))|0; //@line 1928 "snd_fx.cpp"
      $realspeed$0$ph = $$12;
     } else {
      $realspeed$0$ph = $1;
     }
     $24 = ((($this)) + 59692|0); //@line 1929 "snd_fx.cpp"
     $25 = ($realspeed$0$ph|0)==(0); //@line 1931 "snd_fx.cpp"
     $$realspeed$019 = $25 ? 1 : $realspeed$0$ph; //@line 1931 "snd_fx.cpp"
     $$realspeed$020 = $$realspeed$019;$95 = $24;$bDoRetrig$1 = 0;
     label = 12;
    } else {
     $26 = ((($this)) + 59692|0); //@line 1929 "snd_fx.cpp"
     $27 = HEAP32[$26>>2]|0; //@line 1929 "snd_fx.cpp"
     $28 = ($27|0)==(0); //@line 1929 "snd_fx.cpp"
     if ($28) {
      $36 = $6 & 1048580; //@line 1934 "snd_fx.cpp"
      $37 = ($36|0)==(0); //@line 1934 "snd_fx.cpp"
      $$14 = $37 ? $4 : 0; //@line 1934 "snd_fx.cpp"
      $40 = $26;$bDoRetrig$2$off0 = 0;$nRetrigCount$0 = $$14;$realspeed$2 = $1;
      break;
     }
     $29 = ($1|0)==(0); //@line 1931 "snd_fx.cpp"
     $$realspeed$0 = $29 ? 1 : $1; //@line 1931 "snd_fx.cpp"
     $30 = ((($this)) + 59708|0); //@line 1932 "snd_fx.cpp"
     $31 = HEAP32[$30>>2]|0; //@line 1932 "snd_fx.cpp"
     $32 = ($31|0)==(0); //@line 1932 "snd_fx.cpp"
     if ($32) {
      $$realspeed$020 = $$realspeed$0;$95 = $26;$bDoRetrig$1 = 0;
      label = 12;
     } else {
      $33 = (($27>>>0) % ($$realspeed$0>>>0))&-1; //@line 1932 "snd_fx.cpp"
      $34 = ($33|0)==(0); //@line 1932 "snd_fx.cpp"
      $$realspeed$020 = $$realspeed$0;$95 = $26;$bDoRetrig$1 = $34;
      label = 12;
     }
    }
   } while(0);
   if ((label|0) == 12) {
    $35 = (($4) + 1)|0; //@line 1933 "snd_fx.cpp"
    $40 = $95;$bDoRetrig$2$off0 = $bDoRetrig$1;$nRetrigCount$0 = $35;$realspeed$2 = $$realspeed$020;
   }
   $38 = ($nRetrigCount$0>>>0)<($realspeed$2>>>0); //@line 1935 "snd_fx.cpp"
   if (!($38)) {
    $39 = HEAP32[$40>>2]|0; //@line 1937 "snd_fx.cpp"
    $41 = ($39|0)==(0); //@line 1937 "snd_fx.cpp"
    if (!($41)) {
     $nRetrigCount$122 = $nRetrigCount$0;
     break;
    }
    if ($16) {
     $42 = (((($this) + ($nChn<<8)|0)) + 246|0); //@line 1937 "snd_fx.cpp"
     $43 = HEAP8[$42>>0]|0; //@line 1937 "snd_fx.cpp"
     $44 = ($43<<24>>24)==(0); //@line 1937 "snd_fx.cpp"
     $brmerge24 = $bDoRetrig$2$off0 | $44; //@line 1937 "snd_fx.cpp"
     if ($brmerge24) {
      $nRetrigCount$122 = $nRetrigCount$0;
      break;
     } else {
      $nRetrigCount$2 = $nRetrigCount$0;
     }
     $94 = $nRetrigCount$2&255; //@line 1968 "snd_fx.cpp"
     HEAP8[$2>>0] = $94; //@line 1968 "snd_fx.cpp"
     return; //@line 1969 "snd_fx.cpp"
    }
   }
   if ($bDoRetrig$2$off0) {
    $nRetrigCount$122 = $nRetrigCount$0;
   } else {
    $nRetrigCount$2 = $nRetrigCount$0;
    $94 = $nRetrigCount$2&255; //@line 1968 "snd_fx.cpp"
    HEAP8[$2>>0] = $94; //@line 1968 "snd_fx.cpp"
    return; //@line 1969 "snd_fx.cpp"
   }
  } else {
   $9 = ($3<<24>>24)==(0); //@line 1923 "snd_fx.cpp"
   if ($9) {
    $10 = (($4) + 1)|0; //@line 1924 "snd_fx.cpp"
    $nRetrigCount$2 = $10;
    $94 = $nRetrigCount$2&255; //@line 1968 "snd_fx.cpp"
    HEAP8[$2>>0] = $94; //@line 1968 "snd_fx.cpp"
    return; //@line 1969 "snd_fx.cpp"
   } else {
    $11 = ($1|0)==(0); //@line 1922 "snd_fx.cpp"
    $$ = $11 ? 1 : $1; //@line 1922 "snd_fx.cpp"
    $12 = (($4>>>0) % ($$>>>0))&-1; //@line 1923 "snd_fx.cpp"
    $13 = ($12|0)==(0); //@line 1923 "snd_fx.cpp"
    $14 = (($4) + 1)|0; //@line 1924 "snd_fx.cpp"
    if ($13) {
     $nRetrigCount$122 = $14;
     break;
    } else {
     $nRetrigCount$2 = $14;
    }
    $94 = $nRetrigCount$2&255; //@line 1968 "snd_fx.cpp"
    HEAP8[$2>>0] = $94; //@line 1968 "snd_fx.cpp"
    return; //@line 1969 "snd_fx.cpp"
   }
  }
 } while(0);
 $45 = $param >>> 4; //@line 1942 "snd_fx.cpp"
 $46 = $45 & 15; //@line 1942 "snd_fx.cpp"
 $47 = ($46|0)==(0); //@line 1943 "snd_fx.cpp"
 if (!($47)) {
  $48 = (((($this) + ($nChn<<8)|0)) + 116|0); //@line 1945 "snd_fx.cpp"
  $49 = HEAP32[$48>>2]|0; //@line 1945 "snd_fx.cpp"
  $50 = 49344 >>> $46; //@line 1946 "snd_fx.cpp"
  $51 = $50 & 1; //@line 1946 "snd_fx.cpp"
  $52 = ($51|0)==(0); //@line 1946 "snd_fx.cpp"
  if ($52) {
   $58 = (523967 + ($46)|0); //@line 1949 "snd_fx.cpp"
   $59 = HEAP8[$58>>0]|0; //@line 1949 "snd_fx.cpp"
   $60 = $59 << 24 >> 24; //@line 1949 "snd_fx.cpp"
   $61 = $60 << 2; //@line 1949 "snd_fx.cpp"
   $62 = (($61) + ($49))|0; //@line 1949 "snd_fx.cpp"
   $vol$0 = $62;
  } else {
   $53 = (523951 + ($46)|0); //@line 1946 "snd_fx.cpp"
   $54 = HEAP8[$53>>0]|0; //@line 1946 "snd_fx.cpp"
   $55 = $54 << 24 >> 24; //@line 1947 "snd_fx.cpp"
   $56 = Math_imul($55, $49)|0; //@line 1947 "snd_fx.cpp"
   $57 = $56 >> 4; //@line 1947 "snd_fx.cpp"
   $vol$0 = $57;
  }
  $63 = ($vol$0|0)<(0); //@line 1950 "snd_fx.cpp"
  $$vol$0 = $63 ? 0 : $vol$0; //@line 1950 "snd_fx.cpp"
  $64 = ($$vol$0|0)>(256); //@line 1951 "snd_fx.cpp"
  $vol$2 = $64 ? 256 : $$vol$0; //@line 1951 "snd_fx.cpp"
  HEAP32[$48>>2] = $vol$2; //@line 1952 "snd_fx.cpp"
  $65 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 1953 "snd_fx.cpp"
  $66 = HEAP32[$65>>2]|0; //@line 1953 "snd_fx.cpp"
  $67 = $66 | 16777216; //@line 1953 "snd_fx.cpp"
  HEAP32[$65>>2] = $67; //@line 1953 "snd_fx.cpp"
 }
 $68 = (((($this) + ($nChn<<8)|0)) + 214|0); //@line 1955 "snd_fx.cpp"
 $69 = HEAP8[$68>>0]|0; //@line 1955 "snd_fx.cpp"
 $70 = $69&255; //@line 1955 "snd_fx.cpp"
 $71 = (((($this) + ($nChn<<8)|0)) + 128|0); //@line 1956 "snd_fx.cpp"
 $72 = HEAP32[$71>>2]|0; //@line 1956 "snd_fx.cpp"
 $$off = (($69) + -1)<<24>>24; //@line 1957 "snd_fx.cpp"
 $73 = ($$off&255)<(120); //@line 1957 "snd_fx.cpp"
 if ($73) {
  $74 = (((($this) + ($nChn<<8)|0)) + 32|0); //@line 1957 "snd_fx.cpp"
  $75 = HEAP32[$74>>2]|0; //@line 1957 "snd_fx.cpp"
  $76 = ($75|0)==(0); //@line 1957 "snd_fx.cpp"
  if ($76) {
   $78 = $6;
  } else {
   __ZN10CSoundFile8CheckNNAEjjib($this,$nChn,0,$70,1); //@line 1957 "snd_fx.cpp"
   $$pre = HEAP32[$5>>2]|0; //@line 1959 "snd_fx.cpp"
   $78 = $$pre;
  }
 } else {
  $78 = $6;
 }
 $77 = $78 & 1048580; //@line 1959 "snd_fx.cpp"
 $79 = ($77|0)==(0); //@line 1959 "snd_fx.cpp"
 if ($79) {
  $bResetEnv$1$off0 = 0;
 } else {
  $80 = (((($this) + ($nChn<<8)|0)) + 247|0); //@line 1961 "snd_fx.cpp"
  $81 = HEAP8[$80>>0]|0; //@line 1961 "snd_fx.cpp"
  $82 = ($81<<24>>24)!=(0); //@line 1961 "snd_fx.cpp"
  $83 = ($param>>>0)<(256); //@line 1961 "snd_fx.cpp"
  $or$cond3 = $83 & $82; //@line 1961 "snd_fx.cpp"
  if ($or$cond3) {
   $84 = $81&255; //@line 1961 "snd_fx.cpp"
   __ZN10CSoundFile16InstrumentChangeEP11_MODCHANNELjbbb($this,$0,$84,0,0,1); //@line 1961 "snd_fx.cpp"
   $bResetEnv$0 = 1;
  } else {
   $bResetEnv$0 = 0;
  }
  $$bResetEnv$0 = $83 | $bResetEnv$0; //@line 1962 "snd_fx.cpp"
  $bResetEnv$1$off0 = $$bResetEnv$0;
 }
 __ZN10CSoundFile10NoteChangeEjibb($this,$nChn,$70,0,$bResetEnv$1$off0); //@line 1964 "snd_fx.cpp"
 $85 = HEAP32[$5>>2]|0; //@line 1965 "snd_fx.cpp"
 $86 = $85 & 32; //@line 1965 "snd_fx.cpp"
 $87 = ($86|0)==(0); //@line 1965 "snd_fx.cpp"
 if (!($87)) {
  $88 = (((($this) + ($nChn<<8)|0)) + 246|0); //@line 1965 "snd_fx.cpp"
  $89 = HEAP8[$88>>0]|0; //@line 1965 "snd_fx.cpp"
  $90 = ($89<<24>>24)==(0); //@line 1965 "snd_fx.cpp"
  $91 = ($72|0)!=(0); //@line 1965 "snd_fx.cpp"
  $or$cond5 = $91 & $90; //@line 1965 "snd_fx.cpp"
  if ($or$cond5) {
   HEAP32[$71>>2] = $72; //@line 1965 "snd_fx.cpp"
  }
 }
 $92 = $85 & 34; //@line 1966 "snd_fx.cpp"
 $93 = ($92|0)==(0); //@line 1966 "snd_fx.cpp"
 $$nRetrigCount$1 = $93 ? 0 : $nRetrigCount$122; //@line 1966 "snd_fx.cpp"
 $nRetrigCount$2 = $$nRetrigCount$1;
 $94 = $nRetrigCount$2&255; //@line 1968 "snd_fx.cpp"
 HEAP8[$2>>0] = $94; //@line 1968 "snd_fx.cpp"
 return; //@line 1969 "snd_fx.cpp"
}
function __ZN10CSoundFile19ExtendedMODCommandsEjj($this,$nChn,$param) {
 $this = $this|0;
 $nChn = $nChn|0;
 $param = $param|0;
 var $$$i = 0, $$$i9 = 0, $$0$i = 0, $$0$i1 = 0, $$0$i7 = 0, $$0$i8 = 0, $$not$i = 0, $$not$i3 = 0, $$storemerge$i = 0, $$storemerge$i6 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $brmerge$i = 0, $brmerge$i4 = 0, $or$cond$i = 0, $or$cond$i2 = 0;
 var $storemerge$i = 0, $storemerge$i5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $param & 240; //@line 1629 "snd_fx.cpp"
 $1 = $param & 15; //@line 1630 "snd_fx.cpp"
 do {
  switch ($0|0) {
  case 16:  {
   $2 = ($1|0)==(0); //@line 1635 "snd_fx.cpp"
   $3 = ((($this)) + 59680|0); //@line 1635 "snd_fx.cpp"
   $4 = HEAP32[$3>>2]|0; //@line 1635 "snd_fx.cpp"
   $5 = $4 & 1048580;
   $6 = $1 | $5; //@line 1635 "snd_fx.cpp"
   $7 = ($6|0)==(0); //@line 1635 "snd_fx.cpp"
   if ($7) {
    return; //@line 1669 "snd_fx.cpp"
   }
   $8 = $4 & 1048580; //@line 1307 "snd_fx.cpp"
   $9 = ($8|0)==(0); //@line 1307 "snd_fx.cpp"
   do {
    if ($9) {
     $$0$i = $1;
    } else {
     if ($2) {
      $12 = (((($this) + ($nChn<<8)|0)) + 221|0); //@line 1309 "snd_fx.cpp"
      $13 = HEAP8[$12>>0]|0; //@line 1309 "snd_fx.cpp"
      $14 = $13&255; //@line 1309 "snd_fx.cpp"
      $$0$i = $14;
      break;
     } else {
      $10 = $1&255; //@line 1309 "snd_fx.cpp"
      $11 = (((($this) + ($nChn<<8)|0)) + 221|0); //@line 1309 "snd_fx.cpp"
      HEAP8[$11>>0] = $10; //@line 1309 "snd_fx.cpp"
      $$0$i = $1;
      break;
     }
    }
   } while(0);
   $15 = ((($this)) + 59660|0); //@line 1311 "snd_fx.cpp"
   $16 = HEAP32[$15>>2]|0; //@line 1311 "snd_fx.cpp"
   $17 = $16 & 4096; //@line 1311 "snd_fx.cpp"
   $18 = ($17|0)==(0); //@line 1311 "snd_fx.cpp"
   if ($18) {
    return; //@line 1669 "snd_fx.cpp"
   }
   $19 = (((($this) + ($nChn<<8)|0)) + 128|0); //@line 1313 "snd_fx.cpp"
   $20 = HEAP32[$19>>2]|0; //@line 1313 "snd_fx.cpp"
   $21 = ($20|0)!=(0); //@line 1313 "snd_fx.cpp"
   $22 = ($$0$i|0)!=(0); //@line 1313 "snd_fx.cpp"
   $or$cond$i = $22 & $21; //@line 1313 "snd_fx.cpp"
   if (!($or$cond$i)) {
    return; //@line 1669 "snd_fx.cpp"
   }
   $23 = $16 & 16; //@line 1315 "snd_fx.cpp"
   $24 = ($23|0)==(0); //@line 1315 "snd_fx.cpp"
   $$not$i = $9 ^ 1; //@line 1315 "snd_fx.cpp"
   $brmerge$i = $24 | $$not$i; //@line 1315 "snd_fx.cpp"
   if ($brmerge$i) {
    $29 = $$0$i << 2; //@line 1320 "snd_fx.cpp"
    $30 = (($20) - ($29))|0; //@line 1320 "snd_fx.cpp"
    $storemerge$i = $30;
   } else {
    $25 = $$0$i & 15; //@line 1317 "snd_fx.cpp"
    $26 = (41684 + ($25<<2)|0); //@line 1317 "snd_fx.cpp"
    $27 = HEAP32[$26>>2]|0; //@line 1317 "snd_fx.cpp"
    $28 = (__Z8_muldivrlll($20,$27,65536)|0); //@line 1317 "snd_fx.cpp"
    $storemerge$i = $28;
   }
   $31 = ($storemerge$i|0)<(1); //@line 1322 "snd_fx.cpp"
   $$storemerge$i = $31 ? 1 : $storemerge$i; //@line 1322 "snd_fx.cpp"
   HEAP32[$19>>2] = $$storemerge$i; //@line 1322 "snd_fx.cpp"
   return; //@line 1669 "snd_fx.cpp"
   break;
  }
  case 32:  {
   $32 = ($1|0)==(0); //@line 1637 "snd_fx.cpp"
   $33 = ((($this)) + 59680|0); //@line 1637 "snd_fx.cpp"
   $34 = HEAP32[$33>>2]|0; //@line 1637 "snd_fx.cpp"
   $35 = $34 & 1048580;
   $36 = $1 | $35; //@line 1637 "snd_fx.cpp"
   $37 = ($36|0)==(0); //@line 1637 "snd_fx.cpp"
   if ($37) {
    return; //@line 1669 "snd_fx.cpp"
   }
   $38 = $34 & 1048580; //@line 1331 "snd_fx.cpp"
   $39 = ($38|0)==(0); //@line 1331 "snd_fx.cpp"
   do {
    if ($39) {
     $$0$i1 = $1;
    } else {
     if ($32) {
      $42 = (((($this) + ($nChn<<8)|0)) + 221|0); //@line 1333 "snd_fx.cpp"
      $43 = HEAP8[$42>>0]|0; //@line 1333 "snd_fx.cpp"
      $44 = $43&255; //@line 1333 "snd_fx.cpp"
      $$0$i1 = $44;
      break;
     } else {
      $40 = $1&255; //@line 1333 "snd_fx.cpp"
      $41 = (((($this) + ($nChn<<8)|0)) + 221|0); //@line 1333 "snd_fx.cpp"
      HEAP8[$41>>0] = $40; //@line 1333 "snd_fx.cpp"
      $$0$i1 = $1;
      break;
     }
    }
   } while(0);
   $45 = ((($this)) + 59660|0); //@line 1335 "snd_fx.cpp"
   $46 = HEAP32[$45>>2]|0; //@line 1335 "snd_fx.cpp"
   $47 = $46 & 4096; //@line 1335 "snd_fx.cpp"
   $48 = ($47|0)==(0); //@line 1335 "snd_fx.cpp"
   if ($48) {
    return; //@line 1669 "snd_fx.cpp"
   }
   $49 = (((($this) + ($nChn<<8)|0)) + 128|0); //@line 1337 "snd_fx.cpp"
   $50 = HEAP32[$49>>2]|0; //@line 1337 "snd_fx.cpp"
   $51 = ($50|0)!=(0); //@line 1337 "snd_fx.cpp"
   $52 = ($$0$i1|0)!=(0); //@line 1337 "snd_fx.cpp"
   $or$cond$i2 = $52 & $51; //@line 1337 "snd_fx.cpp"
   if (!($or$cond$i2)) {
    return; //@line 1669 "snd_fx.cpp"
   }
   $53 = $46 & 16; //@line 1339 "snd_fx.cpp"
   $54 = ($53|0)==(0); //@line 1339 "snd_fx.cpp"
   $$not$i3 = $39 ^ 1; //@line 1339 "snd_fx.cpp"
   $brmerge$i4 = $54 | $$not$i3; //@line 1339 "snd_fx.cpp"
   if ($brmerge$i4) {
    $59 = $$0$i1 << 2; //@line 1344 "snd_fx.cpp"
    $60 = (($50) + ($59))|0; //@line 1344 "snd_fx.cpp"
    $storemerge$i5 = $60;
   } else {
    $55 = $$0$i1 & 15; //@line 1341 "snd_fx.cpp"
    $56 = (40660 + ($55<<2)|0); //@line 1341 "snd_fx.cpp"
    $57 = HEAP32[$56>>2]|0; //@line 1341 "snd_fx.cpp"
    $58 = (__Z8_muldivrlll($50,$57,65536)|0); //@line 1341 "snd_fx.cpp"
    $storemerge$i5 = $58;
   }
   $61 = ($storemerge$i5|0)>(65535); //@line 1346 "snd_fx.cpp"
   $$storemerge$i6 = $61 ? 65535 : $storemerge$i5; //@line 1346 "snd_fx.cpp"
   HEAP32[$49>>2] = $$storemerge$i6; //@line 1346 "snd_fx.cpp"
   return; //@line 1669 "snd_fx.cpp"
   break;
  }
  case 48:  {
   $62 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 1639 "snd_fx.cpp"
   $63 = HEAP32[$62>>2]|0; //@line 1639 "snd_fx.cpp"
   $64 = $63 & -1048577; //@line 1639 "snd_fx.cpp"
   HEAP32[$62>>2] = $64; //@line 1639 "snd_fx.cpp"
   $65 = ($1|0)==(0); //@line 1639 "snd_fx.cpp"
   if ($65) {
    return; //@line 1669 "snd_fx.cpp"
   }
   $66 = $63 | 1048576; //@line 1639 "snd_fx.cpp"
   HEAP32[$62>>2] = $66; //@line 1639 "snd_fx.cpp"
   return; //@line 1669 "snd_fx.cpp"
   break;
  }
  case 64:  {
   $67 = $param & 7; //@line 1641 "snd_fx.cpp"
   $68 = $67&255; //@line 1641 "snd_fx.cpp"
   $69 = (((($this) + ($nChn<<8)|0)) + 224|0); //@line 1641 "snd_fx.cpp"
   HEAP8[$69>>0] = $68; //@line 1641 "snd_fx.cpp"
   return; //@line 1669 "snd_fx.cpp"
   break;
  }
  case 80:  {
   $70 = ((($this)) + 59692|0); //@line 1643 "snd_fx.cpp"
   $71 = HEAP32[$70>>2]|0; //@line 1643 "snd_fx.cpp"
   $72 = ($71|0)==(0); //@line 1643 "snd_fx.cpp"
   if (!($72)) {
    return; //@line 1669 "snd_fx.cpp"
   }
   $73 = (169292 + ($1<<1)|0); //@line 1644 "snd_fx.cpp"
   $74 = HEAP16[$73>>1]|0; //@line 1644 "snd_fx.cpp"
   $75 = $74&65535; //@line 1644 "snd_fx.cpp"
   $76 = (((($this) + ($nChn<<8)|0)) + 132|0); //@line 1644 "snd_fx.cpp"
   HEAP32[$76>>2] = $75; //@line 1644 "snd_fx.cpp"
   $77 = ((($this)) + 59680|0); //@line 1645 "snd_fx.cpp"
   $78 = HEAP32[$77>>2]|0; //@line 1645 "snd_fx.cpp"
   $79 = $78 & 1048580; //@line 1645 "snd_fx.cpp"
   $80 = ($79|0)==(0); //@line 1645 "snd_fx.cpp"
   $81 = $param << 28; //@line 1648 "snd_fx.cpp"
   $82 = $81 >> 24; //@line 1648 "snd_fx.cpp"
   $83 = $1 << 1; //@line 1646 "snd_fx.cpp"
   $84 = $80 ? $82 : $83; //@line 1645 "snd_fx.cpp"
   $85 = (((($this) + ($nChn<<8)|0)) + 176|0); //@line 1648 "snd_fx.cpp"
   HEAP32[$85>>2] = $84; //@line 1648 "snd_fx.cpp"
   $86 = (((($this) + ($nChn<<8)|0)) + 128|0); //@line 1649 "snd_fx.cpp"
   $87 = HEAP32[$86>>2]|0; //@line 1649 "snd_fx.cpp"
   $88 = ($87|0)==(0); //@line 1649 "snd_fx.cpp"
   if ($88) {
    return; //@line 1669 "snd_fx.cpp"
   }
   $89 = (((($this) + ($nChn<<8)|0)) + 212|0); //@line 1649 "snd_fx.cpp"
   $90 = HEAP8[$89>>0]|0; //@line 1649 "snd_fx.cpp"
   $91 = $90&255; //@line 1649 "snd_fx.cpp"
   $92 = (__ZNK10CSoundFile17GetPeriodFromNoteEjij($this,$91,$84,$75)|0); //@line 1649 "snd_fx.cpp"
   HEAP32[$86>>2] = $92; //@line 1649 "snd_fx.cpp"
   return; //@line 1669 "snd_fx.cpp"
   break;
  }
  case 112:  {
   $93 = $param & 7; //@line 1653 "snd_fx.cpp"
   $94 = $93&255; //@line 1653 "snd_fx.cpp"
   $95 = (((($this) + ($nChn<<8)|0)) + 227|0); //@line 1653 "snd_fx.cpp"
   HEAP8[$95>>0] = $94; //@line 1653 "snd_fx.cpp"
   return; //@line 1669 "snd_fx.cpp"
   break;
  }
  case 128:  {
   $96 = ((($this)) + 59692|0); //@line 1655 "snd_fx.cpp"
   $97 = HEAP32[$96>>2]|0; //@line 1655 "snd_fx.cpp"
   $98 = ($97|0)==(0); //@line 1655 "snd_fx.cpp"
   if (!($98)) {
    return; //@line 1669 "snd_fx.cpp"
   }
   $99 = $1 << 4; //@line 1655 "snd_fx.cpp"
   $100 = $99 | 8; //@line 1655 "snd_fx.cpp"
   $101 = (((($this) + ($nChn<<8)|0)) + 120|0); //@line 1655 "snd_fx.cpp"
   HEAP32[$101>>2] = $100; //@line 1655 "snd_fx.cpp"
   $102 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 1655 "snd_fx.cpp"
   $103 = HEAP32[$102>>2]|0; //@line 1655 "snd_fx.cpp"
   $104 = $103 | 16777216; //@line 1655 "snd_fx.cpp"
   HEAP32[$102>>2] = $104; //@line 1655 "snd_fx.cpp"
   return; //@line 1669 "snd_fx.cpp"
   break;
  }
  case 144:  {
   __ZN10CSoundFile10RetrigNoteEjj($this,$nChn,$1); //@line 1657 "snd_fx.cpp"
   return; //@line 1669 "snd_fx.cpp"
   break;
  }
  case 160:  {
   $105 = ($1|0)==(0); //@line 1659 "snd_fx.cpp"
   do {
    if ($105) {
     $106 = ((($this)) + 59680|0); //@line 1659 "snd_fx.cpp"
     $107 = HEAP32[$106>>2]|0; //@line 1659 "snd_fx.cpp"
     $108 = $107 & 1048580; //@line 1659 "snd_fx.cpp"
     $109 = ($108|0)==(0); //@line 1659 "snd_fx.cpp"
     if ($109) {
      return; //@line 1669 "snd_fx.cpp"
     } else {
      $112 = (((($this) + ($nChn<<8)|0)) + 219|0); //@line 1563 "snd_fx.cpp"
      $113 = HEAP8[$112>>0]|0; //@line 1563 "snd_fx.cpp"
      $114 = $113&255; //@line 1563 "snd_fx.cpp"
      $$0$i7 = $114;
      break;
     }
    } else {
     $110 = $1&255; //@line 1563 "snd_fx.cpp"
     $111 = (((($this) + ($nChn<<8)|0)) + 219|0); //@line 1563 "snd_fx.cpp"
     HEAP8[$111>>0] = $110; //@line 1563 "snd_fx.cpp"
     $$0$i7 = $1;
    }
   } while(0);
   $115 = ((($this)) + 59660|0); //@line 1564 "snd_fx.cpp"
   $116 = HEAP32[$115>>2]|0; //@line 1564 "snd_fx.cpp"
   $117 = $116 & 4096; //@line 1564 "snd_fx.cpp"
   $118 = ($117|0)==(0); //@line 1564 "snd_fx.cpp"
   if ($118) {
    return; //@line 1669 "snd_fx.cpp"
   }
   $119 = $$0$i7 << 2; //@line 1566 "snd_fx.cpp"
   $120 = (((($this) + ($nChn<<8)|0)) + 116|0); //@line 1566 "snd_fx.cpp"
   $121 = HEAP32[$120>>2]|0; //@line 1566 "snd_fx.cpp"
   $122 = (($121) + ($119))|0; //@line 1566 "snd_fx.cpp"
   $123 = ($122|0)>(256); //@line 1567 "snd_fx.cpp"
   $$$i = $123 ? 256 : $122; //@line 1567 "snd_fx.cpp"
   HEAP32[$120>>2] = $$$i; //@line 1566 "snd_fx.cpp"
   $124 = ((($this)) + 59680|0); //@line 1568 "snd_fx.cpp"
   $125 = HEAP32[$124>>2]|0; //@line 1568 "snd_fx.cpp"
   $126 = $125 & 1; //@line 1568 "snd_fx.cpp"
   $127 = ($126|0)==(0); //@line 1568 "snd_fx.cpp"
   if ($127) {
    return; //@line 1669 "snd_fx.cpp"
   }
   $128 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 1568 "snd_fx.cpp"
   $129 = HEAP32[$128>>2]|0; //@line 1568 "snd_fx.cpp"
   $130 = $129 | 16777216; //@line 1568 "snd_fx.cpp"
   HEAP32[$128>>2] = $130; //@line 1568 "snd_fx.cpp"
   return; //@line 1669 "snd_fx.cpp"
   break;
  }
  case 176:  {
   $131 = ($1|0)==(0); //@line 1661 "snd_fx.cpp"
   do {
    if ($131) {
     $132 = ((($this)) + 59680|0); //@line 1661 "snd_fx.cpp"
     $133 = HEAP32[$132>>2]|0; //@line 1661 "snd_fx.cpp"
     $134 = $133 & 1048580; //@line 1661 "snd_fx.cpp"
     $135 = ($134|0)==(0); //@line 1661 "snd_fx.cpp"
     if ($135) {
      return; //@line 1669 "snd_fx.cpp"
     } else {
      $138 = (((($this) + ($nChn<<8)|0)) + 219|0); //@line 1576 "snd_fx.cpp"
      $139 = HEAP8[$138>>0]|0; //@line 1576 "snd_fx.cpp"
      $140 = $139&255; //@line 1576 "snd_fx.cpp"
      $$0$i8 = $140;
      break;
     }
    } else {
     $136 = $1&255; //@line 1576 "snd_fx.cpp"
     $137 = (((($this) + ($nChn<<8)|0)) + 219|0); //@line 1576 "snd_fx.cpp"
     HEAP8[$137>>0] = $136; //@line 1576 "snd_fx.cpp"
     $$0$i8 = $1;
    }
   } while(0);
   $141 = ((($this)) + 59660|0); //@line 1577 "snd_fx.cpp"
   $142 = HEAP32[$141>>2]|0; //@line 1577 "snd_fx.cpp"
   $143 = $142 & 4096; //@line 1577 "snd_fx.cpp"
   $144 = ($143|0)==(0); //@line 1577 "snd_fx.cpp"
   if ($144) {
    return; //@line 1669 "snd_fx.cpp"
   }
   $145 = $$0$i8 << 2; //@line 1579 "snd_fx.cpp"
   $146 = (((($this) + ($nChn<<8)|0)) + 116|0); //@line 1579 "snd_fx.cpp"
   $147 = HEAP32[$146>>2]|0; //@line 1579 "snd_fx.cpp"
   $148 = (($147) - ($145))|0; //@line 1579 "snd_fx.cpp"
   $149 = ($148|0)<(0); //@line 1580 "snd_fx.cpp"
   $$$i9 = $149 ? 0 : $148; //@line 1580 "snd_fx.cpp"
   HEAP32[$146>>2] = $$$i9; //@line 1579 "snd_fx.cpp"
   $150 = ((($this)) + 59680|0); //@line 1581 "snd_fx.cpp"
   $151 = HEAP32[$150>>2]|0; //@line 1581 "snd_fx.cpp"
   $152 = $151 & 1; //@line 1581 "snd_fx.cpp"
   $153 = ($152|0)==(0); //@line 1581 "snd_fx.cpp"
   if ($153) {
    return; //@line 1669 "snd_fx.cpp"
   }
   $154 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 1581 "snd_fx.cpp"
   $155 = HEAP32[$154>>2]|0; //@line 1581 "snd_fx.cpp"
   $156 = $155 | 16777216; //@line 1581 "snd_fx.cpp"
   HEAP32[$154>>2] = $156; //@line 1581 "snd_fx.cpp"
   return; //@line 1669 "snd_fx.cpp"
   break;
  }
  case 192:  {
   $157 = ((($this)) + 59692|0); //@line 2010 "snd_fx.cpp"
   $158 = HEAP32[$157>>2]|0; //@line 2010 "snd_fx.cpp"
   $159 = ($158|0)==($1|0); //@line 2010 "snd_fx.cpp"
   if (!($159)) {
    return; //@line 1669 "snd_fx.cpp"
   }
   $160 = (((($this) + ($nChn<<8)|0)) + 116|0); //@line 2014 "snd_fx.cpp"
   HEAP32[$160>>2] = 0; //@line 2014 "snd_fx.cpp"
   $161 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 2015 "snd_fx.cpp"
   $162 = HEAP32[$161>>2]|0; //@line 2015 "snd_fx.cpp"
   $163 = $162 | 16777216; //@line 2015 "snd_fx.cpp"
   HEAP32[$161>>2] = $163; //@line 2015 "snd_fx.cpp"
   return; //@line 1669 "snd_fx.cpp"
   break;
  }
  case 240:  {
   $164 = $1&255; //@line 1667 "snd_fx.cpp"
   $165 = (((($this) + ($nChn<<8)|0)) + 254|0); //@line 1667 "snd_fx.cpp"
   HEAP8[$165>>0] = $164; //@line 1667 "snd_fx.cpp"
   return; //@line 1669 "snd_fx.cpp"
   break;
  }
  default: {
   return; //@line 1669 "snd_fx.cpp"
  }
  }
 } while(0);
}
function __ZN10CSoundFile19ExtendedS3MCommandsEjj($this,$nChn,$param) {
 $this = $this|0;
 $nChn = $nChn|0;
 $param = $param|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bkp$01 = 0, $bkp$01$us = 0;
 var $bkp$01$us4 = 0, $exitcond = 0, $exitcond5 = 0, $exitcond6 = 0, $i$02 = 0, $i$02$us = 0, $i$02$us3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $param & 240; //@line 1676 "snd_fx.cpp"
 $1 = $param & 15; //@line 1677 "snd_fx.cpp"
 do {
  switch ($0|0) {
  case 16:  {
   $2 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 1682 "snd_fx.cpp"
   $3 = HEAP32[$2>>2]|0; //@line 1682 "snd_fx.cpp"
   $4 = $3 & -1048577; //@line 1682 "snd_fx.cpp"
   HEAP32[$2>>2] = $4; //@line 1682 "snd_fx.cpp"
   $5 = ($1|0)==(0); //@line 1682 "snd_fx.cpp"
   if ($5) {
    return; //@line 1753 "snd_fx.cpp"
   }
   $6 = $3 | 1048576; //@line 1682 "snd_fx.cpp"
   HEAP32[$2>>2] = $6; //@line 1682 "snd_fx.cpp"
   return; //@line 1753 "snd_fx.cpp"
   break;
  }
  case 32:  {
   $7 = ((($this)) + 59692|0); //@line 1684 "snd_fx.cpp"
   $8 = HEAP32[$7>>2]|0; //@line 1684 "snd_fx.cpp"
   $9 = ($8|0)==(0); //@line 1684 "snd_fx.cpp"
   if (!($9)) {
    return; //@line 1753 "snd_fx.cpp"
   }
   $10 = (169292 + ($1<<1)|0); //@line 1685 "snd_fx.cpp"
   $11 = HEAP16[$10>>1]|0; //@line 1685 "snd_fx.cpp"
   $12 = $11&65535; //@line 1685 "snd_fx.cpp"
   $13 = (((($this) + ($nChn<<8)|0)) + 132|0); //@line 1685 "snd_fx.cpp"
   HEAP32[$13>>2] = $12; //@line 1685 "snd_fx.cpp"
   $14 = $param << 28; //@line 1686 "snd_fx.cpp"
   $15 = $14 >> 24; //@line 1686 "snd_fx.cpp"
   $16 = (((($this) + ($nChn<<8)|0)) + 176|0); //@line 1686 "snd_fx.cpp"
   HEAP32[$16>>2] = $15; //@line 1686 "snd_fx.cpp"
   $17 = (((($this) + ($nChn<<8)|0)) + 128|0); //@line 1687 "snd_fx.cpp"
   $18 = HEAP32[$17>>2]|0; //@line 1687 "snd_fx.cpp"
   $19 = ($18|0)==(0); //@line 1687 "snd_fx.cpp"
   if ($19) {
    return; //@line 1753 "snd_fx.cpp"
   }
   $20 = (((($this) + ($nChn<<8)|0)) + 212|0); //@line 1687 "snd_fx.cpp"
   $21 = HEAP8[$20>>0]|0; //@line 1687 "snd_fx.cpp"
   $22 = $21&255; //@line 1687 "snd_fx.cpp"
   $23 = (__ZNK10CSoundFile17GetPeriodFromNoteEjij($this,$22,$15,$12)|0); //@line 1687 "snd_fx.cpp"
   HEAP32[$17>>2] = $23; //@line 1687 "snd_fx.cpp"
   return; //@line 1753 "snd_fx.cpp"
   break;
  }
  case 48:  {
   $24 = $param & 7; //@line 1690 "snd_fx.cpp"
   $25 = $24&255; //@line 1690 "snd_fx.cpp"
   $26 = (((($this) + ($nChn<<8)|0)) + 224|0); //@line 1690 "snd_fx.cpp"
   HEAP8[$26>>0] = $25; //@line 1690 "snd_fx.cpp"
   return; //@line 1753 "snd_fx.cpp"
   break;
  }
  case 64:  {
   $27 = $param & 7; //@line 1692 "snd_fx.cpp"
   $28 = $27&255; //@line 1692 "snd_fx.cpp"
   $29 = (((($this) + ($nChn<<8)|0)) + 227|0); //@line 1692 "snd_fx.cpp"
   HEAP8[$29>>0] = $28; //@line 1692 "snd_fx.cpp"
   return; //@line 1753 "snd_fx.cpp"
   break;
  }
  case 80:  {
   $30 = $param & 7; //@line 1694 "snd_fx.cpp"
   $31 = $30&255; //@line 1694 "snd_fx.cpp"
   $32 = (((($this) + ($nChn<<8)|0)) + 230|0); //@line 1694 "snd_fx.cpp"
   HEAP8[$32>>0] = $31; //@line 1694 "snd_fx.cpp"
   return; //@line 1753 "snd_fx.cpp"
   break;
  }
  case 96:  {
   $33 = ((($this)) + 59704|0); //@line 1696 "snd_fx.cpp"
   HEAP32[$33>>2] = $1; //@line 1696 "snd_fx.cpp"
   return; //@line 1753 "snd_fx.cpp"
   break;
  }
  case 112:  {
   $34 = ((($this)) + 59692|0); //@line 1698 "snd_fx.cpp"
   $35 = HEAP32[$34>>2]|0; //@line 1698 "snd_fx.cpp"
   $36 = ($35|0)==(0); //@line 1698 "snd_fx.cpp"
   if (!($36)) {
    return; //@line 1753 "snd_fx.cpp"
   }
   do {
    switch ($1|0) {
    case 2: case 1: case 0:  {
     $37 = ((($this)) + 59664|0); //@line 1705 "snd_fx.cpp"
     $38 = HEAP32[$37>>2]|0; //@line 1705 "snd_fx.cpp"
     $39 = (($this) + ($38<<8)|0); //@line 1705 "snd_fx.cpp"
     $40 = ($38>>>0)<(128); //@line 1706 "snd_fx.cpp"
     if (!($40)) {
      return; //@line 1753 "snd_fx.cpp"
     }
     $41 = (($nChn) + 1)|0; //@line 1708 "snd_fx.cpp"
     switch ($1|0) {
     case 1:  {
      $bkp$01$us = $39;$i$02$us = $38;
      while(1) {
       $42 = ((($bkp$01$us)) + 160|0); //@line 1708 "snd_fx.cpp"
       $43 = HEAP32[$42>>2]|0; //@line 1708 "snd_fx.cpp"
       $44 = ($43|0)==($41|0); //@line 1708 "snd_fx.cpp"
       if ($44) {
        __ZN10CSoundFile6KeyOffEj($this,$i$02$us); //@line 1710 "snd_fx.cpp"
       }
       $45 = (($i$02$us) + 1)|0; //@line 1706 "snd_fx.cpp"
       $46 = ((($bkp$01$us)) + 256|0); //@line 1706 "snd_fx.cpp"
       $exitcond5 = ($45|0)==(128); //@line 1706 "snd_fx.cpp"
       if ($exitcond5) {
        break;
       } else {
        $bkp$01$us = $46;$i$02$us = $45;
       }
      }
      return; //@line 1753 "snd_fx.cpp"
      break;
     }
     case 2:  {
      $bkp$01$us4 = $39;$i$02$us3 = $38;
      while(1) {
       $47 = ((($bkp$01$us4)) + 160|0); //@line 1708 "snd_fx.cpp"
       $48 = HEAP32[$47>>2]|0; //@line 1708 "snd_fx.cpp"
       $49 = ($48|0)==($41|0); //@line 1708 "snd_fx.cpp"
       if ($49) {
        $50 = ((($bkp$01$us4)) + 36|0); //@line 1711 "snd_fx.cpp"
        $51 = HEAP32[$50>>2]|0; //@line 1711 "snd_fx.cpp"
        $52 = $51 | 1024; //@line 1711 "snd_fx.cpp"
        HEAP32[$50>>2] = $52; //@line 1711 "snd_fx.cpp"
       }
       $53 = (($i$02$us3) + 1)|0; //@line 1706 "snd_fx.cpp"
       $54 = ((($bkp$01$us4)) + 256|0); //@line 1706 "snd_fx.cpp"
       $exitcond6 = ($53|0)==(128); //@line 1706 "snd_fx.cpp"
       if ($exitcond6) {
        break;
       } else {
        $bkp$01$us4 = $54;$i$02$us3 = $53;
       }
      }
      return; //@line 1753 "snd_fx.cpp"
      break;
     }
     default: {
      $bkp$01 = $39;$i$02 = $38;
      while(1) {
       $55 = ((($bkp$01)) + 160|0); //@line 1708 "snd_fx.cpp"
       $56 = HEAP32[$55>>2]|0; //@line 1708 "snd_fx.cpp"
       $57 = ($56|0)==($41|0); //@line 1708 "snd_fx.cpp"
       if ($57) {
        $58 = ((($bkp$01)) + 36|0); //@line 1711 "snd_fx.cpp"
        $59 = HEAP32[$58>>2]|0; //@line 1711 "snd_fx.cpp"
        $60 = $59 | 1024; //@line 1711 "snd_fx.cpp"
        HEAP32[$58>>2] = $60; //@line 1711 "snd_fx.cpp"
        $61 = ((($bkp$01)) + 124|0); //@line 1712 "snd_fx.cpp"
        HEAP32[$61>>2] = 0; //@line 1712 "snd_fx.cpp"
       }
       $62 = (($i$02) + 1)|0; //@line 1706 "snd_fx.cpp"
       $63 = ((($bkp$01)) + 256|0); //@line 1706 "snd_fx.cpp"
       $exitcond = ($62|0)==(128); //@line 1706 "snd_fx.cpp"
       if ($exitcond) {
        break;
       } else {
        $bkp$01 = $63;$i$02 = $62;
       }
      }
      return; //@line 1753 "snd_fx.cpp"
     }
     }
     break;
    }
    case 3:  {
     $64 = (((($this) + ($nChn<<8)|0)) + 213|0); //@line 1717 "snd_fx.cpp"
     HEAP8[$64>>0] = 0; //@line 1717 "snd_fx.cpp"
     return; //@line 1753 "snd_fx.cpp"
     break;
    }
    case 4:  {
     $65 = (((($this) + ($nChn<<8)|0)) + 213|0); //@line 1718 "snd_fx.cpp"
     HEAP8[$65>>0] = 1; //@line 1718 "snd_fx.cpp"
     return; //@line 1753 "snd_fx.cpp"
     break;
    }
    case 5:  {
     $66 = (((($this) + ($nChn<<8)|0)) + 213|0); //@line 1719 "snd_fx.cpp"
     HEAP8[$66>>0] = 2; //@line 1719 "snd_fx.cpp"
     return; //@line 1753 "snd_fx.cpp"
     break;
    }
    case 6:  {
     $67 = (((($this) + ($nChn<<8)|0)) + 213|0); //@line 1720 "snd_fx.cpp"
     HEAP8[$67>>0] = 3; //@line 1720 "snd_fx.cpp"
     return; //@line 1753 "snd_fx.cpp"
     break;
    }
    case 7:  {
     $68 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 1721 "snd_fx.cpp"
     $69 = HEAP32[$68>>2]|0; //@line 1721 "snd_fx.cpp"
     $70 = $69 & -2097153; //@line 1721 "snd_fx.cpp"
     HEAP32[$68>>2] = $70; //@line 1721 "snd_fx.cpp"
     return; //@line 1753 "snd_fx.cpp"
     break;
    }
    case 8:  {
     $71 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 1722 "snd_fx.cpp"
     $72 = HEAP32[$71>>2]|0; //@line 1722 "snd_fx.cpp"
     $73 = $72 | 2097152; //@line 1722 "snd_fx.cpp"
     HEAP32[$71>>2] = $73; //@line 1722 "snd_fx.cpp"
     return; //@line 1753 "snd_fx.cpp"
     break;
    }
    case 9:  {
     $74 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 1723 "snd_fx.cpp"
     $75 = HEAP32[$74>>2]|0; //@line 1723 "snd_fx.cpp"
     $76 = $75 & -4194305; //@line 1723 "snd_fx.cpp"
     HEAP32[$74>>2] = $76; //@line 1723 "snd_fx.cpp"
     return; //@line 1753 "snd_fx.cpp"
     break;
    }
    case 10:  {
     $77 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 1724 "snd_fx.cpp"
     $78 = HEAP32[$77>>2]|0; //@line 1724 "snd_fx.cpp"
     $79 = $78 | 4194304; //@line 1724 "snd_fx.cpp"
     HEAP32[$77>>2] = $79; //@line 1724 "snd_fx.cpp"
     return; //@line 1753 "snd_fx.cpp"
     break;
    }
    case 11:  {
     $80 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 1725 "snd_fx.cpp"
     $81 = HEAP32[$80>>2]|0; //@line 1725 "snd_fx.cpp"
     $82 = $81 & -8388609; //@line 1725 "snd_fx.cpp"
     HEAP32[$80>>2] = $82; //@line 1725 "snd_fx.cpp"
     return; //@line 1753 "snd_fx.cpp"
     break;
    }
    case 12:  {
     $83 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 1726 "snd_fx.cpp"
     $84 = HEAP32[$83>>2]|0; //@line 1726 "snd_fx.cpp"
     $85 = $84 | 8388608; //@line 1726 "snd_fx.cpp"
     HEAP32[$83>>2] = $85; //@line 1726 "snd_fx.cpp"
     return; //@line 1753 "snd_fx.cpp"
     break;
    }
    default: {
     return; //@line 1753 "snd_fx.cpp"
    }
    }
   } while(0);
   break;
  }
  case 128:  {
   $86 = ((($this)) + 59692|0); //@line 1730 "snd_fx.cpp"
   $87 = HEAP32[$86>>2]|0; //@line 1730 "snd_fx.cpp"
   $88 = ($87|0)==(0); //@line 1730 "snd_fx.cpp"
   if (!($88)) {
    return; //@line 1753 "snd_fx.cpp"
   }
   $89 = $1 << 4; //@line 1730 "snd_fx.cpp"
   $90 = $89 | 8; //@line 1730 "snd_fx.cpp"
   $91 = (((($this) + ($nChn<<8)|0)) + 120|0); //@line 1730 "snd_fx.cpp"
   HEAP32[$91>>2] = $90; //@line 1730 "snd_fx.cpp"
   $92 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 1730 "snd_fx.cpp"
   $93 = HEAP32[$92>>2]|0; //@line 1730 "snd_fx.cpp"
   $94 = $93 | 16777216; //@line 1730 "snd_fx.cpp"
   HEAP32[$92>>2] = $94; //@line 1730 "snd_fx.cpp"
   return; //@line 1753 "snd_fx.cpp"
   break;
  }
  case 144:  {
   $95 = (($this) + ($nChn<<8)|0); //@line 1675 "snd_fx.cpp"
   __ZN10CSoundFile21ExtendedChannelEffectEP11_MODCHANNELj($this,$95,$1); //@line 1732 "snd_fx.cpp"
   return; //@line 1753 "snd_fx.cpp"
   break;
  }
  case 160:  {
   $96 = ((($this)) + 59692|0); //@line 1734 "snd_fx.cpp"
   $97 = HEAP32[$96>>2]|0; //@line 1734 "snd_fx.cpp"
   $98 = ($97|0)==(0); //@line 1734 "snd_fx.cpp"
   if (!($98)) {
    return; //@line 1753 "snd_fx.cpp"
   }
   $99 = $1&255; //@line 1736 "snd_fx.cpp"
   $100 = (((($this) + ($nChn<<8)|0)) + 237|0); //@line 1736 "snd_fx.cpp"
   HEAP8[$100>>0] = $99; //@line 1736 "snd_fx.cpp"
   $101 = (((($this) + ($nChn<<8)|0)) + 246|0); //@line 1737 "snd_fx.cpp"
   $102 = HEAP8[$101>>0]|0; //@line 1737 "snd_fx.cpp"
   $103 = ($102<<24>>24)>(0); //@line 1737 "snd_fx.cpp"
   if (!($103)) {
    return; //@line 1753 "snd_fx.cpp"
   }
   $104 = $1 << 16; //@line 1739 "snd_fx.cpp"
   $105 = (((($this) + ($nChn<<8)|0)) + 32|0); //@line 1740 "snd_fx.cpp"
   $106 = HEAP32[$105>>2]|0; //@line 1740 "snd_fx.cpp"
   $107 = ($104>>>0)<($106>>>0); //@line 1740 "snd_fx.cpp"
   if (!($107)) {
    return; //@line 1753 "snd_fx.cpp"
   }
   $108 = (((($this) + ($nChn<<8)|0)) + 4|0); //@line 1740 "snd_fx.cpp"
   HEAP32[$108>>2] = $104; //@line 1740 "snd_fx.cpp"
   return; //@line 1753 "snd_fx.cpp"
   break;
  }
  case 192:  {
   $109 = ((($this)) + 59692|0); //@line 2010 "snd_fx.cpp"
   $110 = HEAP32[$109>>2]|0; //@line 2010 "snd_fx.cpp"
   $111 = ($110|0)==($1|0); //@line 2010 "snd_fx.cpp"
   if (!($111)) {
    return; //@line 1753 "snd_fx.cpp"
   }
   $112 = (((($this) + ($nChn<<8)|0)) + 116|0); //@line 2014 "snd_fx.cpp"
   HEAP32[$112>>2] = 0; //@line 2014 "snd_fx.cpp"
   $113 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 2015 "snd_fx.cpp"
   $114 = HEAP32[$113>>2]|0; //@line 2015 "snd_fx.cpp"
   $115 = $114 | 16777216; //@line 2015 "snd_fx.cpp"
   HEAP32[$113>>2] = $115; //@line 2015 "snd_fx.cpp"
   return; //@line 1753 "snd_fx.cpp"
   break;
  }
  case 240:  {
   $116 = $1&255; //@line 1751 "snd_fx.cpp"
   $117 = (((($this) + ($nChn<<8)|0)) + 254|0); //@line 1751 "snd_fx.cpp"
   HEAP8[$117>>0] = $116; //@line 1751 "snd_fx.cpp"
   return; //@line 1753 "snd_fx.cpp"
   break;
  }
  default: {
   return; //@line 1753 "snd_fx.cpp"
  }
  }
 } while(0);
}
function __ZN10CSoundFile21ExtendedChannelEffectEP11_MODCHANNELj($this,$pChn,$param) {
 $this = $this|0;
 $pChn = $pChn|0;
 $param = $param|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 59692|0); //@line 1760 "snd_fx.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1760 "snd_fx.cpp"
 $2 = ($1|0)==(0); //@line 1760 "snd_fx.cpp"
 if (!($2)) {
  return; //@line 1809 "snd_fx.cpp"
 }
 $3 = $param & 15; //@line 1761 "snd_fx.cpp"
 do {
  switch ($3|0) {
  case 0:  {
   $4 = ((($pChn)) + 36|0); //@line 1764 "snd_fx.cpp"
   $5 = HEAP32[$4>>2]|0; //@line 1764 "snd_fx.cpp"
   $6 = $5 & -2049; //@line 1764 "snd_fx.cpp"
   HEAP32[$4>>2] = $6; //@line 1764 "snd_fx.cpp"
   return; //@line 1809 "snd_fx.cpp"
   break;
  }
  case 1:  {
   $7 = ((($pChn)) + 36|0); //@line 1766 "snd_fx.cpp"
   $8 = HEAP32[$7>>2]|0; //@line 1766 "snd_fx.cpp"
   $9 = $8 | 2048; //@line 1766 "snd_fx.cpp"
   HEAP32[$7>>2] = $9; //@line 1766 "snd_fx.cpp"
   $10 = ((($pChn)) + 120|0); //@line 1766 "snd_fx.cpp"
   HEAP32[$10>>2] = 128; //@line 1766 "snd_fx.cpp"
   return; //@line 1809 "snd_fx.cpp"
   break;
  }
  case 8:  {
   $11 = ((($pChn)) + 36|0); //@line 1771 "snd_fx.cpp"
   $12 = HEAP32[$11>>2]|0; //@line 1771 "snd_fx.cpp"
   $13 = $12 & -201326593; //@line 1771 "snd_fx.cpp"
   $14 = $13 | 134217728; //@line 1772 "snd_fx.cpp"
   HEAP32[$11>>2] = $14; //@line 1772 "snd_fx.cpp"
   return; //@line 1809 "snd_fx.cpp"
   break;
  }
  case 9:  {
   $15 = ((($pChn)) + 36|0); //@line 1776 "snd_fx.cpp"
   $16 = HEAP32[$15>>2]|0; //@line 1776 "snd_fx.cpp"
   $17 = $16 & -201326593; //@line 1776 "snd_fx.cpp"
   $18 = $17 | 67108864; //@line 1777 "snd_fx.cpp"
   HEAP32[$15>>2] = $18; //@line 1777 "snd_fx.cpp"
   return; //@line 1809 "snd_fx.cpp"
   break;
  }
  case 10:  {
   $19 = ((($this)) + 59660|0); //@line 1781 "snd_fx.cpp"
   $20 = HEAP32[$19>>2]|0; //@line 1781 "snd_fx.cpp"
   $21 = $20 & -16385; //@line 1781 "snd_fx.cpp"
   HEAP32[$19>>2] = $21; //@line 1781 "snd_fx.cpp"
   return; //@line 1809 "snd_fx.cpp"
   break;
  }
  case 11:  {
   $22 = ((($this)) + 59660|0); //@line 1785 "snd_fx.cpp"
   $23 = HEAP32[$22>>2]|0; //@line 1785 "snd_fx.cpp"
   $24 = $23 | 16384; //@line 1785 "snd_fx.cpp"
   HEAP32[$22>>2] = $24; //@line 1785 "snd_fx.cpp"
   return; //@line 1809 "snd_fx.cpp"
   break;
  }
  case 12:  {
   $25 = ((($this)) + 59660|0); //@line 1789 "snd_fx.cpp"
   $26 = HEAP32[$25>>2]|0; //@line 1789 "snd_fx.cpp"
   $27 = $26 & -8193; //@line 1789 "snd_fx.cpp"
   HEAP32[$25>>2] = $27; //@line 1789 "snd_fx.cpp"
   return; //@line 1809 "snd_fx.cpp"
   break;
  }
  case 13:  {
   $28 = ((($this)) + 59660|0); //@line 1793 "snd_fx.cpp"
   $29 = HEAP32[$28>>2]|0; //@line 1793 "snd_fx.cpp"
   $30 = $29 | 8192; //@line 1793 "snd_fx.cpp"
   HEAP32[$28>>2] = $30; //@line 1793 "snd_fx.cpp"
   return; //@line 1809 "snd_fx.cpp"
   break;
  }
  case 14:  {
   $31 = ((($pChn)) + 36|0); //@line 1797 "snd_fx.cpp"
   $32 = HEAP32[$31>>2]|0; //@line 1797 "snd_fx.cpp"
   $33 = $32 & -129; //@line 1797 "snd_fx.cpp"
   HEAP32[$31>>2] = $33; //@line 1797 "snd_fx.cpp"
   return; //@line 1809 "snd_fx.cpp"
   break;
  }
  case 15:  {
   $34 = ((($pChn)) + 36|0); //@line 1801 "snd_fx.cpp"
   $35 = HEAP32[$34>>2]|0; //@line 1801 "snd_fx.cpp"
   $36 = $35 & 2; //@line 1801 "snd_fx.cpp"
   $37 = ($36|0)==(0); //@line 1801 "snd_fx.cpp"
   if ($37) {
    $38 = ((($pChn)) + 4|0); //@line 1801 "snd_fx.cpp"
    $39 = HEAP32[$38>>2]|0; //@line 1801 "snd_fx.cpp"
    $40 = ($39|0)==(0); //@line 1801 "snd_fx.cpp"
    if ($40) {
     $41 = ((($pChn)) + 32|0); //@line 1801 "snd_fx.cpp"
     $42 = HEAP32[$41>>2]|0; //@line 1801 "snd_fx.cpp"
     $43 = ($42|0)==(0); //@line 1801 "snd_fx.cpp"
     if (!($43)) {
      $44 = (($42) + -1)|0; //@line 1803 "snd_fx.cpp"
      HEAP32[$38>>2] = $44; //@line 1803 "snd_fx.cpp"
      $45 = ((($pChn)) + 8|0); //@line 1804 "snd_fx.cpp"
      HEAP32[$45>>2] = 65535; //@line 1804 "snd_fx.cpp"
     }
    }
   }
   $46 = $35 | 128; //@line 1806 "snd_fx.cpp"
   HEAP32[$34>>2] = $46; //@line 1806 "snd_fx.cpp"
   return; //@line 1809 "snd_fx.cpp"
   break;
  }
  default: {
   return; //@line 1809 "snd_fx.cpp"
  }
  }
 } while(0);
}
function __ZN10CSoundFile16ProcessMidiMacroEjPKcj($this,$nChn,$pszMidiMacro,$param) {
 $this = $this|0;
 $nChn = $nChn|0;
 $pszMidiMacro = $pszMidiMacro|0;
 $param = $param|0;
 var $$ = 0, $$off = 0, $$off13 = 0, $$off14 = 0, $$off15 = 0, $$off16 = 0, $$off17 = 0, $$off18 = 0, $$off19 = 0, $$phi$trans$insert = 0, $$phi$trans$insert25 = 0, $$pre = 0, $$pre26 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dwByteCode$0$be = 0, $dwByteCode$02033 = 0, $dwMidiCode$0$be = 0, $dwMidiCode$02132 = 0, $dwParam$0 = 0, $dwParam$1 = 0, $nBytes$0$be = 0, $nBytes$02231 = 0, $nNib$0$be = 0, $nNib$02330 = 0, $not$ = 0, $not$6 = 0, $or$cond28 = 0, $or$cond5 = 0, $pos$02429 = 0;
 var $pos$02429$phi = 0, $switch$split2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($this) + ($nChn<<8)|0); //@line 1815 "snd_fx.cpp"
 $1 = HEAP32[$pszMidiMacro>>2]|0; //@line 1816 "snd_fx.cpp"
 $2 = $1 & 2136964959; //@line 1816 "snd_fx.cpp"
 $switch$split2D = ($2|0)<(826683462);
 if ($switch$split2D) {
  switch ($2|0) {
  case 809906246:  {
   label = 23;
   break;
  }
  default: {
   label = 2;
  }
  }
 } else {
  switch ($2|0) {
  case 826683462:  {
   label = 23;
   break;
  }
  default: {
   label = 2;
  }
  }
 }
 if ((label|0) == 2) {
  $3 = $1&255;
  $4 = ((($this)) + 59664|0); //@line 1840 "snd_fx.cpp"
  $5 = (($nChn) + 1)|0; //@line 1840 "snd_fx.cpp"
  $6 = (((($this) + ($nChn<<8)|0)) + 160|0); //@line 1840 "snd_fx.cpp"
  $7 = $param & 127; //@line 1829 "snd_fx.cpp"
  $8 = $param & 112; //@line 1830 "snd_fx.cpp"
  $9 = $param << 3; //@line 1831 "snd_fx.cpp"
  $10 = $9 & 120; //@line 1831 "snd_fx.cpp"
  $11 = ($3<<24>>24)==(0); //@line 1825 "snd_fx.cpp"
  if ($11) {
   return; //@line 1908 "snd_fx.cpp"
  } else {
   $13 = $3;$26 = 1;$dwByteCode$02033 = 0;$dwMidiCode$02132 = 0;$nBytes$02231 = 0;$nNib$02330 = 0;$pos$02429 = 0;
  }
  while(1) {
   $12 = $13 << 24 >> 24; //@line 1826 "snd_fx.cpp"
   $$off17 = (($13) + -48)<<24>>24; //@line 1826 "snd_fx.cpp"
   $14 = ($$off17&255)<(10); //@line 1826 "snd_fx.cpp"
   L10: do {
    if ($14) {
     $15 = $dwByteCode$02033 << 4; //@line 1826 "snd_fx.cpp"
     $16 = (($12) + -48)|0; //@line 1826 "snd_fx.cpp"
     $17 = $16 | $15; //@line 1826 "snd_fx.cpp"
     $18 = (($nNib$02330) + 1)|0; //@line 1826 "snd_fx.cpp"
     $dwByteCode$0$be = $17;$dwMidiCode$0$be = $dwMidiCode$02132;$nBytes$0$be = $nBytes$02231;$nNib$0$be = $18;
    } else {
     $$off18 = (($13) + -65)<<24>>24; //@line 1827 "snd_fx.cpp"
     $19 = ($$off18&255)<(6); //@line 1827 "snd_fx.cpp"
     if ($19) {
      $20 = $dwByteCode$02033 << 4; //@line 1827 "snd_fx.cpp"
      $21 = (($12) + -55)|0; //@line 1827 "snd_fx.cpp"
      $22 = $21 | $20; //@line 1827 "snd_fx.cpp"
      $23 = (($nNib$02330) + 1)|0; //@line 1827 "snd_fx.cpp"
      $dwByteCode$0$be = $22;$dwMidiCode$0$be = $dwMidiCode$02132;$nBytes$0$be = $nBytes$02231;$nNib$0$be = $23;
      break;
     }
     $$off19 = (($13) + -97)<<24>>24; //@line 1828 "snd_fx.cpp"
     $29 = ($$off19&255)<(6); //@line 1828 "snd_fx.cpp"
     if ($29) {
      $30 = $dwByteCode$02033 << 4; //@line 1828 "snd_fx.cpp"
      $31 = (($12) + -87)|0; //@line 1828 "snd_fx.cpp"
      $32 = $31 | $30; //@line 1828 "snd_fx.cpp"
      $33 = (($nNib$02330) + 1)|0; //@line 1828 "snd_fx.cpp"
      $dwByteCode$0$be = $32;$dwMidiCode$0$be = $dwMidiCode$02132;$nBytes$0$be = $nBytes$02231;$nNib$0$be = $33;
      break;
     }
     switch ($13<<24>>24) {
     case 90: case 122:  {
      $dwByteCode$0$be = $7;$dwMidiCode$0$be = $dwMidiCode$02132;$nBytes$0$be = $nBytes$02231;$nNib$0$be = 2;
      break L10;
      break;
     }
     case 88: case 120:  {
      $dwByteCode$0$be = $8;$dwMidiCode$0$be = $dwMidiCode$02132;$nBytes$0$be = $nBytes$02231;$nNib$0$be = 2;
      break L10;
      break;
     }
     case 89: case 121:  {
      $dwByteCode$0$be = $10;$dwMidiCode$0$be = $dwMidiCode$02132;$nBytes$0$be = $nBytes$02231;$nNib$0$be = 2;
      break L10;
      break;
     }
     default: {
      $34 = ($nNib$02330>>>0)>(1); //@line 1832 "snd_fx.cpp"
      if (!($34)) {
       $dwByteCode$0$be = $dwByteCode$02033;$dwMidiCode$0$be = $dwMidiCode$02132;$nBytes$0$be = $nBytes$02231;$nNib$0$be = $nNib$02330;
       break L10;
      }
      $35 = $nBytes$02231 << 3; //@line 1835 "snd_fx.cpp"
      $36 = $dwByteCode$02033 << $35; //@line 1835 "snd_fx.cpp"
      $37 = $36 | $dwMidiCode$02132; //@line 1835 "snd_fx.cpp"
      $38 = (($nBytes$02231) + 1)|0; //@line 1837 "snd_fx.cpp"
      $39 = ($38>>>0)>(2); //@line 1838 "snd_fx.cpp"
      if (!($39)) {
       $dwByteCode$0$be = 0;$dwMidiCode$0$be = $37;$nBytes$0$be = $38;$nNib$0$be = 0;
       break L10;
      }
      $40 = HEAP32[$4>>2]|0; //@line 1840 "snd_fx.cpp"
      $41 = ($40>>>0)>($nChn>>>0); //@line 1840 "snd_fx.cpp"
      if ($41) {
       $44 = $5;
      } else {
       $42 = HEAP32[$6>>2]|0; //@line 1840 "snd_fx.cpp"
       $44 = $42;
      }
      $43 = (($44) + -1)|0; //@line 1841 "snd_fx.cpp"
      $45 = ($43>>>0)<($40>>>0); //@line 1841 "snd_fx.cpp"
      if (!($45)) {
       $dwByteCode$0$be = 0;$dwMidiCode$0$be = 0;$nBytes$0$be = 0;$nNib$0$be = 0;
       break L10;
      }
      $46 = (($44) + -1)|0; //@line 1843 "snd_fx.cpp"
      $47 = (((((($this)) + 49600|0) + (($46*36)|0)|0)) + 12|0); //@line 1843 "snd_fx.cpp"
      $48 = HEAP32[$47>>2]|0; //@line 1843 "snd_fx.cpp"
      $$off = (($48) + -1)|0; //@line 1844 "snd_fx.cpp"
      $49 = ($$off>>>0)<(8); //@line 1844 "snd_fx.cpp"
      if (!($49)) {
       $dwByteCode$0$be = 0;$dwMidiCode$0$be = 0;$nBytes$0$be = 0;$nNib$0$be = 0;
       break L10;
      }
      $50 = (((($this)) + 58496|0) + (($$off*144)|0)|0); //@line 1846 "snd_fx.cpp"
      $51 = HEAP32[$50>>2]|0; //@line 1846 "snd_fx.cpp"
      $52 = ($51|0)==(0|0); //@line 1847 "snd_fx.cpp"
      if ($52) {
       $dwByteCode$0$be = 0;$dwMidiCode$0$be = 0;$nBytes$0$be = 0;$nNib$0$be = 0;
       break L10;
      }
      $53 = (((((($this)) + 58496|0) + (($$off*144)|0)|0)) + 4|0); //@line 1847 "snd_fx.cpp"
      $54 = HEAP32[$53>>2]|0; //@line 1847 "snd_fx.cpp"
      $55 = ($54|0)==(0|0); //@line 1847 "snd_fx.cpp"
      if ($55) {
       $dwByteCode$0$be = 0;$dwMidiCode$0$be = 0;$nBytes$0$be = 0;$nNib$0$be = 0;
       break L10;
      }
      $56 = HEAP32[$51>>2]|0; //@line 1849 "snd_fx.cpp"
      $57 = ((($56)) + 32|0); //@line 1849 "snd_fx.cpp"
      $58 = HEAP32[$57>>2]|0; //@line 1849 "snd_fx.cpp"
      FUNCTION_TABLE_vii[$58 & 8191]($51,$37); //@line 1849 "snd_fx.cpp"
      $dwByteCode$0$be = 0;$dwMidiCode$0$be = 0;$nBytes$0$be = 0;$nNib$0$be = 0;
      break L10;
     }
     }
    }
   } while(0);
   $24 = (($pos$02429) + 7)|0; //@line 1822 "snd_fx.cpp"
   $25 = ($24>>>0)<(33); //@line 1822 "snd_fx.cpp"
   if (!($25)) {
    label = 42;
    break;
   }
   $$phi$trans$insert25 = (($pszMidiMacro) + ($26)|0);
   $$pre26 = HEAP8[$$phi$trans$insert25>>0]|0; //@line 1824 "snd_fx.cpp"
   $27 = (($26) + 1)|0; //@line 1824 "snd_fx.cpp"
   $28 = ($$pre26<<24>>24)==(0); //@line 1825 "snd_fx.cpp"
   if ($28) {
    label = 42;
    break;
   } else {
    $pos$02429$phi = $26;$13 = $$pre26;$26 = $27;$dwByteCode$02033 = $dwByteCode$0$be;$dwMidiCode$02132 = $dwMidiCode$0$be;$nBytes$02231 = $nBytes$0$be;$nNib$02330 = $nNib$0$be;$pos$02429 = $pos$02429$phi;
   }
  }
  if ((label|0) == 42) {
   return; //@line 1908 "snd_fx.cpp"
  }
 }
 else if ((label|0) == 23) {
  $59 = ((($pszMidiMacro)) + 4|0); //@line 1862 "snd_fx.cpp"
  $60 = HEAP8[$59>>0]|0; //@line 1864 "snd_fx.cpp"
  $61 = ($60<<24>>24)==(48); //@line 1864 "snd_fx.cpp"
  if (!($61)) {
   return; //@line 1908 "snd_fx.cpp"
  }
  $62 = ((($pszMidiMacro)) + 6|0); //@line 1866 "snd_fx.cpp"
  $63 = HEAP8[$62>>0]|0; //@line 1866 "snd_fx.cpp"
  $64 = $63 << 24 >> 24; //@line 1868 "snd_fx.cpp"
  L41: do {
   switch ($63<<24>>24) {
   case 90: case 122:  {
    $dwParam$1 = $param;
    break;
   }
   default: {
    $65 = ((($pszMidiMacro)) + 7|0); //@line 1873 "snd_fx.cpp"
    $66 = HEAP8[$65>>0]|0; //@line 1873 "snd_fx.cpp"
    $$off13 = (($63) + -48)<<24>>24; //@line 1874 "snd_fx.cpp"
    $67 = ($$off13&255)<(10); //@line 1874 "snd_fx.cpp"
    if ($67) {
     $68 = $64 << 4; //@line 1874 "snd_fx.cpp"
     $69 = (($68) + -768)|0; //@line 1874 "snd_fx.cpp"
     $dwParam$0 = $69;
    } else {
     $$off14 = (($63) + -65)<<24>>24; //@line 1875 "snd_fx.cpp"
     $70 = ($$off14&255)<(6); //@line 1875 "snd_fx.cpp"
     if ($70) {
      $71 = $64 << 4; //@line 1875 "snd_fx.cpp"
      $72 = (($71) + -880)|0; //@line 1875 "snd_fx.cpp"
      $dwParam$0 = $72;
     } else {
      $dwParam$0 = 0;
     }
    }
    $73 = $66 << 24 >> 24; //@line 1876 "snd_fx.cpp"
    $$off15 = (($66) + -48)<<24>>24; //@line 1876 "snd_fx.cpp"
    $74 = ($$off15&255)<(10); //@line 1876 "snd_fx.cpp"
    if ($74) {
     $75 = (($73) + -48)|0; //@line 1876 "snd_fx.cpp"
     $76 = (($75) + ($dwParam$0))|0; //@line 1876 "snd_fx.cpp"
     $dwParam$1 = $76;
     break L41;
    }
    $$off16 = (($66) + -65)<<24>>24; //@line 1877 "snd_fx.cpp"
    $77 = ($$off16&255)<(6); //@line 1877 "snd_fx.cpp"
    if ($77) {
     $78 = (($73) + -55)|0; //@line 1877 "snd_fx.cpp"
     $79 = (($78) + ($dwParam$0))|0; //@line 1877 "snd_fx.cpp"
     $dwParam$1 = $79;
    } else {
     $dwParam$1 = $dwParam$0;
    }
   }
   }
  } while(0);
  $80 = ((($pszMidiMacro)) + 5|0); //@line 1879 "snd_fx.cpp"
  $81 = HEAP8[$80>>0]|0; //@line 1879 "snd_fx.cpp"
  $82 = $81 << 24 >> 24; //@line 1879 "snd_fx.cpp"
  switch ($82|0) {
  case 48:  {
   $83 = (((($this) + ($nChn<<8)|0)) + 238|0); //@line 1884 "snd_fx.cpp"
   $84 = HEAP8[$83>>0]|0; //@line 1884 "snd_fx.cpp"
   $85 = $84&255; //@line 1884 "snd_fx.cpp"
   $86 = ($dwParam$1>>>0)<(128); //@line 1885 "snd_fx.cpp"
   if ($86) {
    $87 = $dwParam$1&255; //@line 1885 "snd_fx.cpp"
    HEAP8[$83>>0] = $87; //@line 1885 "snd_fx.cpp"
    $89 = $87;
   } else {
    $89 = $84;
   }
   $88 = $89&255; //@line 1887 "snd_fx.cpp"
   $90 = (($85) - ($88))|0; //@line 1887 "snd_fx.cpp"
   $91 = ($90|0)<(0); //@line 1889 "snd_fx.cpp"
   $92 = (0 - ($90))|0; //@line 1889 "snd_fx.cpp"
   $$ = $91 ? $92 : $90; //@line 1889 "snd_fx.cpp"
   $93 = (((($this) + ($nChn<<8)|0)) + 116|0); //@line 1890 "snd_fx.cpp"
   $94 = HEAP32[$93>>2]|0; //@line 1890 "snd_fx.cpp"
   $95 = ($94|0)>(0); //@line 1890 "snd_fx.cpp"
   $96 = ($$|0)<(16); //@line 1890 "snd_fx.cpp"
   $or$cond5 = $95 | $96; //@line 1890 "snd_fx.cpp"
   $$phi$trans$insert = (((($this) + ($nChn<<8)|0)) + 36|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 1892 "snd_fx.cpp"
   $97 = $$pre & 16384;
   $98 = ($97|0)==(0); //@line 1891 "snd_fx.cpp"
   $or$cond28 = $or$cond5 | $98; //@line 1890 "snd_fx.cpp"
   if (!($or$cond28)) {
    $99 = (((($this) + ($nChn<<8)|0)) + 20|0); //@line 1891 "snd_fx.cpp"
    $100 = HEAP32[$99>>2]|0; //@line 1891 "snd_fx.cpp"
    $101 = (((($this) + ($nChn<<8)|0)) + 16|0); //@line 1891 "snd_fx.cpp"
    $102 = HEAP32[$101>>2]|0; //@line 1891 "snd_fx.cpp"
    $103 = $102 | $100; //@line 1891 "snd_fx.cpp"
    $104 = ($103|0)==(0); //@line 1891 "snd_fx.cpp"
    if (!($104)) {
     return; //@line 1908 "snd_fx.cpp"
    }
   }
   $105 = $$pre & 16384; //@line 1892 "snd_fx.cpp"
   $not$6 = ($105|0)==(0); //@line 1892 "snd_fx.cpp"
   __ZNK10CSoundFile18SetupChannelFilterEP11_MODCHANNELbi($this,$0,$not$6,256); //@line 1892 "snd_fx.cpp"
   return; //@line 1908 "snd_fx.cpp"
   break;
  }
  case 49:  {
   $106 = ($dwParam$1>>>0)<(128); //@line 1899 "snd_fx.cpp"
   if ($106) {
    $107 = $dwParam$1&255; //@line 1899 "snd_fx.cpp"
    $108 = (((($this) + ($nChn<<8)|0)) + 239|0); //@line 1899 "snd_fx.cpp"
    HEAP8[$108>>0] = $107; //@line 1899 "snd_fx.cpp"
   }
   $109 = (((($this) + ($nChn<<8)|0)) + 36|0); //@line 1901 "snd_fx.cpp"
   $110 = HEAP32[$109>>2]|0; //@line 1901 "snd_fx.cpp"
   $111 = $110 & 16384; //@line 1901 "snd_fx.cpp"
   $not$ = ($111|0)==(0); //@line 1901 "snd_fx.cpp"
   __ZNK10CSoundFile18SetupChannelFilterEP11_MODCHANNELbi($this,$0,$not$,256); //@line 1901 "snd_fx.cpp"
   return; //@line 1908 "snd_fx.cpp"
   break;
  }
  default: {
   return; //@line 1908 "snd_fx.cpp"
  }
  }
 }
}
function __ZNK10CSoundFile19IsValidBackwardJumpEjjjj($this,$nStartOrder,$nStartRow,$nJumpOrder,$nJumpRow) {
 $this = $this|0;
 $nStartOrder = $nStartOrder|0;
 $nStartRow = $nStartRow|0;
 $nJumpOrder = $nJumpOrder|0;
 $nJumpRow = $nJumpRow|0;
 var $$ = 0, $$0 = 0, $$0719 = 0, $$0719$lcssa = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $breakrow$015 = 0, $breakrow$1 = 0, $breakrow$1$lcssa = 0, $i$017 = 0, $p$014 = 0, $posjump$016 = 0, $posjump$1 = 0;
 var $posjump$1$lcssa = 0, $row$0 = 0, $row$1 = 0, $row_hist = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $row_hist = sp;
 $0 = ($nJumpOrder>>>0)<(240); //@line 2210 "snd_fx.cpp"
 if ($0) {
  $$0719 = $nJumpOrder;
 } else {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 2250 "snd_fx.cpp"
 }
 while(1) {
  $1 = (((($this)) + 53344|0) + ($$0719)|0); //@line 2210 "snd_fx.cpp"
  $2 = HEAP8[$1>>0]|0; //@line 2210 "snd_fx.cpp"
  $3 = ($2<<24>>24)==(-2); //@line 2210 "snd_fx.cpp"
  if (!($3)) {
   $$0719$lcssa = $$0719;
   break;
  }
  $4 = (($$0719) + 1)|0; //@line 2210 "snd_fx.cpp"
  $5 = ($4>>>0)<(240); //@line 2210 "snd_fx.cpp"
  if ($5) {
   $$0719 = $4;
  } else {
   $$1 = 0;
   label = 24;
   break;
  }
 }
 if ((label|0) == 24) {
  STACKTOP = sp;return ($$1|0); //@line 2250 "snd_fx.cpp"
 }
 $6 = ($nStartOrder>>>0)>(239); //@line 2211 "snd_fx.cpp"
 if ($6) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 2250 "snd_fx.cpp"
 }
 $7 = ($$0719$lcssa>>>0)>($nStartOrder>>>0); //@line 2213 "snd_fx.cpp"
 if ($7) {
  $$1 = 1;
  STACKTOP = sp;return ($$1|0); //@line 2250 "snd_fx.cpp"
 }
 $8 = ($$0719$lcssa>>>0)<($nStartOrder>>>0); //@line 2214 "snd_fx.cpp"
 if ($8) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 2250 "snd_fx.cpp"
 }
 $9 = (((($this)) + 52864|0) + ($nStartOrder<<1)|0); //@line 2214 "snd_fx.cpp"
 $10 = HEAP16[$9>>1]|0; //@line 2214 "snd_fx.cpp"
 $11 = $10&65535; //@line 2214 "snd_fx.cpp"
 $12 = ($11>>>0)>($nJumpRow>>>0); //@line 2214 "snd_fx.cpp"
 if (!($12)) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 2250 "snd_fx.cpp"
 }
 $13 = (((($this)) + 51904|0) + ($nStartOrder<<2)|0); //@line 2215 "snd_fx.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 2215 "snd_fx.cpp"
 $15 = ($14|0)==(0|0); //@line 2215 "snd_fx.cpp"
 $16 = $nJumpRow | $nStartRow; //@line 2215 "snd_fx.cpp"
 $17 = ($16>>>0)>(255); //@line 2215 "snd_fx.cpp"
 $18 = $17 | $15; //@line 2215 "snd_fx.cpp"
 if ($18) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 2250 "snd_fx.cpp"
 }
 _memset(($row_hist|0),0,256)|0; //@line 2218 "snd_fx.cpp"
 $19 = ($10&65535)>(256); //@line 2220 "snd_fx.cpp"
 $$ = $19 ? 256 : $11; //@line 2220 "snd_fx.cpp"
 $20 = (($row_hist) + ($nStartRow)|0); //@line 2221 "snd_fx.cpp"
 HEAP8[$20>>0] = 1; //@line 2221 "snd_fx.cpp"
 $21 = ((($this)) + 59664|0); //@line 2226 "snd_fx.cpp"
 $row$0 = $nJumpRow;
 L24: while(1) {
  $22 = ($row$0>>>0)<(256); //@line 2222 "snd_fx.cpp"
  if (!($22)) {
   $$0 = 0;
   break;
  }
  $23 = (($row_hist) + ($row$0)|0); //@line 2222 "snd_fx.cpp"
  $24 = HEAP8[$23>>0]|0; //@line 2222 "snd_fx.cpp"
  $25 = ($24<<24>>24)==(0); //@line 2222 "snd_fx.cpp"
  if (!($25)) {
   $$0 = 0;
   break;
  }
  $26 = ($row$0>>>0)<($$>>>0); //@line 2224 "snd_fx.cpp"
  if (!($26)) {
   $$0 = 1;
   break;
  }
  HEAP8[$23>>0] = 1; //@line 2225 "snd_fx.cpp"
  $27 = (($row$0) + 1)|0; //@line 2227 "snd_fx.cpp"
  $28 = HEAP32[$21>>2]|0; //@line 2229 "snd_fx.cpp"
  $29 = ($28|0)==(0); //@line 2229 "snd_fx.cpp"
  if ($29) {
   $row$1 = $27;
  } else {
   $30 = Math_imul($28, $row$0)|0; //@line 2226 "snd_fx.cpp"
   $31 = (($14) + (($30*6)|0)|0); //@line 2226 "snd_fx.cpp"
   $breakrow$015 = -1;$i$017 = 0;$p$014 = $31;$posjump$016 = 0;
   while(1) {
    $32 = ((($p$014)) + 3|0); //@line 2231 "snd_fx.cpp"
    $33 = HEAP8[$32>>0]|0; //@line 2231 "snd_fx.cpp"
    switch ($33<<24>>24) {
    case 12:  {
     $34 = ((($p$014)) + 5|0); //@line 2233 "snd_fx.cpp"
     $35 = HEAP8[$34>>0]|0; //@line 2233 "snd_fx.cpp"
     $36 = $35&255; //@line 2233 "snd_fx.cpp"
     $37 = ($36>>>0)<($nStartOrder>>>0); //@line 2233 "snd_fx.cpp"
     if ($37) {
      $$0 = 0;
      break L24;
     }
     $38 = ($36>>>0)>($nStartOrder>>>0); //@line 2234 "snd_fx.cpp"
     if ($38) {
      $$0 = 1;
      break L24;
     } else {
      $breakrow$1 = $breakrow$015;$posjump$1 = 1;
     }
     break;
    }
    case 14:  {
     $39 = ((($p$014)) + 5|0); //@line 2239 "snd_fx.cpp"
     $40 = HEAP8[$39>>0]|0; //@line 2239 "snd_fx.cpp"
     $41 = $40&255; //@line 2239 "snd_fx.cpp"
     $breakrow$1 = $41;$posjump$1 = $posjump$016;
     break;
    }
    default: {
     $breakrow$1 = $breakrow$015;$posjump$1 = $posjump$016;
    }
    }
    $42 = (($i$017) + 1)|0; //@line 2229 "snd_fx.cpp"
    $43 = ((($p$014)) + 6|0); //@line 2229 "snd_fx.cpp"
    $44 = ($42>>>0)<($28>>>0); //@line 2229 "snd_fx.cpp"
    if ($44) {
     $breakrow$015 = $breakrow$1;$i$017 = $42;$p$014 = $43;$posjump$016 = $posjump$1;
    } else {
     $breakrow$1$lcssa = $breakrow$1;$posjump$1$lcssa = $posjump$1;
     break;
    }
   }
   $45 = ($breakrow$1$lcssa|0)>(-1); //@line 2242 "snd_fx.cpp"
   if ($45) {
    $46 = ($posjump$1$lcssa|0)==(0); //@line 2244 "snd_fx.cpp"
    if ($46) {
     $$0 = 1;
     break;
    } else {
     $row$1 = $breakrow$1$lcssa;
    }
   } else {
    $row$1 = $27;
   }
  }
  $47 = ($row$1>>>0)<($$>>>0); //@line 2247 "snd_fx.cpp"
  if ($47) {
   $row$0 = $row$1;
  } else {
   $$0 = 1;
   break;
  }
 }
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0); //@line 2250 "snd_fx.cpp"
}
function __ZNK10CSoundFile17GetNoteFromPeriodEj($this,$period) {
 $this = $this|0;
 $period = $period|0;
 var $$0 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$06 = 0, $i$06$lcssa = 0, $i1$05 = 0, $or$cond = 0, $or$cond2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($period|0)==(0); //@line 2259 "snd_fx.cpp"
 if ($0) {
  $$0 = 0;
  return ($$0|0); //@line 2286 "snd_fx.cpp"
 }
 $1 = ((($this)) + 59680|0); //@line 2260 "snd_fx.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 2260 "snd_fx.cpp"
 $3 = $2 & 2130009; //@line 2260 "snd_fx.cpp"
 $4 = ($3|0)==(0); //@line 2260 "snd_fx.cpp"
 if ($4) {
  $i1$05 = 1;
  while(1) {
   $23 = (__ZNK10CSoundFile17GetPeriodFromNoteEjij($this,$i1$05,0,0)|0); //@line 2281 "snd_fx.cpp"
   $24 = ($23|0)<(1); //@line 2282 "snd_fx.cpp"
   $25 = ($23|0)>($period|0); //@line 2282 "snd_fx.cpp"
   $or$cond2 = $24 | $25; //@line 2282 "snd_fx.cpp"
   if (!($or$cond2)) {
    $$0 = $i1$05;
    label = 12;
    break;
   }
   $26 = (($i1$05) + 1)|0; //@line 2279 "snd_fx.cpp"
   $27 = ($26>>>0)<(120); //@line 2279 "snd_fx.cpp"
   if ($27) {
    $i1$05 = $26;
   } else {
    $$0 = 120;
    label = 12;
    break;
   }
  }
  if ((label|0) == 12) {
   return ($$0|0); //@line 2286 "snd_fx.cpp"
  }
 }
 $5 = $period >>> 2; //@line 2262 "snd_fx.cpp"
 $i$06 = 0;
 while(1) {
  $6 = (169148 + ($i$06<<1)|0); //@line 2265 "snd_fx.cpp"
  $7 = HEAP16[$6>>1]|0; //@line 2265 "snd_fx.cpp"
  $8 = $7&65535; //@line 2265 "snd_fx.cpp"
  $9 = ($5>>>0)<($8>>>0); //@line 2265 "snd_fx.cpp"
  if (!($9)) {
   $$lcssa = $8;$i$06$lcssa = $i$06;
   break;
  }
  $21 = (($i$06) + 1)|0; //@line 2263 "snd_fx.cpp"
  $22 = ($21>>>0)<(72); //@line 2263 "snd_fx.cpp"
  if ($22) {
   $i$06 = $21;
  } else {
   $$0 = 108;
   label = 12;
   break;
  }
 }
 if ((label|0) == 12) {
  return ($$0|0); //@line 2286 "snd_fx.cpp"
 }
 $10 = ($5|0)!=($$lcssa|0); //@line 2267 "snd_fx.cpp"
 $11 = ($i$06$lcssa|0)!=(0); //@line 2267 "snd_fx.cpp"
 $or$cond = $11 & $10; //@line 2267 "snd_fx.cpp"
 if ($or$cond) {
  $12 = (($i$06$lcssa) + -1)|0; //@line 2269 "snd_fx.cpp"
  $13 = (169148 + ($12<<1)|0); //@line 2269 "snd_fx.cpp"
  $14 = HEAP16[$13>>1]|0; //@line 2269 "snd_fx.cpp"
  $15 = $14&65535; //@line 2269 "snd_fx.cpp"
  $16 = (($15) - ($5))|0; //@line 2271 "snd_fx.cpp"
  $17 = (($5) - ($$lcssa))|0; //@line 2271 "snd_fx.cpp"
  $18 = ($16>>>0)<($17>>>0); //@line 2271 "snd_fx.cpp"
  if ($18) {
   $19 = (($i$06$lcssa) + 36)|0; //@line 2271 "snd_fx.cpp"
   $$0 = $19;
   return ($$0|0); //@line 2286 "snd_fx.cpp"
  }
 }
 $20 = (($i$06$lcssa) + 37)|0; //@line 2273 "snd_fx.cpp"
 $$0 = $20;
 return ($$0|0); //@line 2286 "snd_fx.cpp"
}
function __ZNK10CSoundFile17GetFreqFromPeriodEjji($this,$period,$nC4Speed,$nPeriodFrac) {
 $this = $this|0;
 $period = $period|0;
 $nC4Speed = $nC4Speed|0;
 $nPeriodFrac = $nPeriodFrac|0;
 var $$0 = 0, $$nC4Speed = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($period|0)==(0); //@line 2355 "snd_fx.cpp"
 if ($0) {
  $$0 = 0;
  return ($$0|0); //@line 2377 "snd_fx.cpp"
 }
 $1 = ((($this)) + 59680|0); //@line 2356 "snd_fx.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 2356 "snd_fx.cpp"
 $3 = $2 & 2130009; //@line 2356 "snd_fx.cpp"
 $4 = ($3|0)==(0); //@line 2356 "snd_fx.cpp"
 if (!($4)) {
  $5 = (14187580 / ($period>>>0))&-1; //@line 2358 "snd_fx.cpp"
  $$0 = $5;
  return ($$0|0); //@line 2377 "snd_fx.cpp"
 }
 $6 = $2 & 1048580; //@line 2360 "snd_fx.cpp"
 $7 = ($6|0)==(0); //@line 2360 "snd_fx.cpp"
 $8 = ((($this)) + 59660|0); //@line 2362 "snd_fx.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 2362 "snd_fx.cpp"
 $10 = $9 & 16; //@line 2362 "snd_fx.cpp"
 $11 = ($10|0)!=(0); //@line 2362 "snd_fx.cpp"
 if ($7) {
  if ($11) {
   $18 = ($nC4Speed|0)==(0); //@line 2370 "snd_fx.cpp"
   $$nC4Speed = $18 ? 8363 : $nC4Speed; //@line 2370 "snd_fx.cpp"
   $19 = $period << 8; //@line 2371 "snd_fx.cpp"
   $20 = (($19) + ($nPeriodFrac))|0; //@line 2371 "snd_fx.cpp"
   $21 = (__Z7_muldivlll($$nC4Speed,438272,$20)|0); //@line 2371 "snd_fx.cpp"
   $$0 = $21;
   return ($$0|0); //@line 2377 "snd_fx.cpp"
  } else {
   $22 = $period << 8; //@line 2374 "snd_fx.cpp"
   $23 = (($22) + ($nPeriodFrac))|0; //@line 2374 "snd_fx.cpp"
   $24 = (__Z7_muldivlll(8363,438272,$23)|0); //@line 2374 "snd_fx.cpp"
   $$0 = $24;
   return ($$0|0); //@line 2377 "snd_fx.cpp"
  }
 } else {
  if ($11) {
   $12 = (($period>>>0) % 768)&-1; //@line 2363 "snd_fx.cpp"
   $13 = (42836 + ($12<<2)|0); //@line 2363 "snd_fx.cpp"
   $14 = HEAP32[$13>>2]|0; //@line 2363 "snd_fx.cpp"
   $15 = (($period>>>0) / 768)&-1; //@line 2363 "snd_fx.cpp"
   $16 = $14 >>> $15; //@line 2363 "snd_fx.cpp"
   $$0 = $16;
   return ($$0|0); //@line 2377 "snd_fx.cpp"
  } else {
   $17 = (14317456 / ($period>>>0))&-1; //@line 2365 "snd_fx.cpp"
   $$0 = $17;
   return ($$0|0); //@line 2377 "snd_fx.cpp"
  }
 }
 return (0)|0;
}
function __Z14X86_StereoFillPijS_S_($pBuffer,$nSamples,$lpROfs,$lpLOfs) {
 $pBuffer = $pBuffer|0;
 $nSamples = $nSamples|0;
 $lpROfs = $lpROfs|0;
 $lpLOfs = $lpLOfs|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$02 = 0, $lofs$0$lcssa = 0, $lofs$03 = 0, $rofs$0$lcssa = 0, $rofs$04 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$lpROfs>>2]|0; //@line 2198 "fastmix.cpp"
 $1 = HEAP32[$lpLOfs>>2]|0; //@line 2199 "fastmix.cpp"
 $2 = $1 | $0; //@line 2201 "fastmix.cpp"
 $3 = ($2|0)==(0); //@line 2201 "fastmix.cpp"
 if ($3) {
  $5 = $nSamples << 3; //@line 2203 "fastmix.cpp"
  _memset(($pBuffer|0),0,($5|0))|0; //@line 2032 "fastmix.cpp"
  return; //@line 2217 "fastmix.cpp"
 }
 $4 = ($nSamples|0)==(0); //@line 2206 "fastmix.cpp"
 if ($4) {
  $lofs$0$lcssa = $1;$rofs$0$lcssa = $0;
 } else {
  $i$02 = 0;$lofs$03 = $1;$rofs$04 = $0;
  while(1) {
   $6 = (0 - ($rofs$04))|0; //@line 2208 "fastmix.cpp"
   $7 = $6 >> 31; //@line 2208 "fastmix.cpp"
   $8 = $7 & 255; //@line 2208 "fastmix.cpp"
   $9 = (($8) + ($rofs$04))|0; //@line 2208 "fastmix.cpp"
   $10 = $9 >> 8; //@line 2208 "fastmix.cpp"
   $11 = (0 - ($lofs$03))|0; //@line 2209 "fastmix.cpp"
   $12 = $11 >> 31; //@line 2209 "fastmix.cpp"
   $13 = $12 & 255; //@line 2209 "fastmix.cpp"
   $14 = (($13) + ($lofs$03))|0; //@line 2209 "fastmix.cpp"
   $15 = $14 >> 8; //@line 2209 "fastmix.cpp"
   $16 = (($rofs$04) - ($10))|0; //@line 2210 "fastmix.cpp"
   $17 = (($lofs$03) - ($15))|0; //@line 2211 "fastmix.cpp"
   $18 = $i$02 << 1; //@line 2212 "fastmix.cpp"
   $19 = (($pBuffer) + ($18<<2)|0); //@line 2212 "fastmix.cpp"
   HEAP32[$19>>2] = $10; //@line 2212 "fastmix.cpp"
   $20 = $18 | 1; //@line 2213 "fastmix.cpp"
   $21 = (($pBuffer) + ($20<<2)|0); //@line 2213 "fastmix.cpp"
   HEAP32[$21>>2] = $15; //@line 2213 "fastmix.cpp"
   $22 = (($i$02) + 1)|0; //@line 2206 "fastmix.cpp"
   $exitcond = ($22|0)==($nSamples|0); //@line 2206 "fastmix.cpp"
   if ($exitcond) {
    $lofs$0$lcssa = $17;$rofs$0$lcssa = $16;
    break;
   } else {
    $i$02 = $22;$lofs$03 = $17;$rofs$04 = $16;
   }
  }
 }
 HEAP32[$lpROfs>>2] = $rofs$0$lcssa; //@line 2215 "fastmix.cpp"
 HEAP32[$lpLOfs>>2] = $lofs$0$lcssa; //@line 2216 "fastmix.cpp"
 return; //@line 2217 "fastmix.cpp"
}
function __ZN10CSoundFile15CreateStereoMixEi($this,$count) {
 $this = $this|0;
 $count = $count|0;
 var $$ = 0, $$$ = 0, $$$i = 0, $$$nSamples$i = 0, $$0 = 0, $$0$i7 = 0, $$2$i = 0, $$4$i = 0, $$4$nSamples$i = 0, $$in = 0, $$lcssa = 0, $$not$i = 0, $$pn = 0, $$pre = 0, $$pre40 = 0, $$sum = 0, $$sum1 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0;
 var $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0;
 var $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0;
 var $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0;
 var $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0;
 var $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0;
 var $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0;
 var $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0;
 var $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0;
 var $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0;
 var $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i = 0, $factor$i = 0, $i$02$i = 0, $lofs$0$lcssa$i = 0, $lofs$03$i = 0, $nChn$026 = 0, $nFlags$3 = 0, $nInc$0$i = 0, $nInc$1$i = 0, $nSmpCount$0$i = 0, $naddmix$0 = 0, $naddmix$0$lcssa = 0, $nchmixed$027 = 0, $nchmixed$1 = 0, $nchused$030 = 0, $nchused$1 = 0;
 var $nrampsamples$0 = 0, $nsamples$0 = 0, $nsamples$0$lcssa68 = 0, $nsamples$070 = 0, $or$cond = 0, $or$cond$i = 0, $or$cond12$i = 0, $or$cond13$i = 0, $or$cond14$i = 0, $or$cond7$i = 0, $or$cond8$i = 0, $or$cond8$not$i = 0, $pMixFuncTable$0 = 0, $pbuffer$0 = 0, $pbuffer$0$lcssa71 = 0, $pbuffer$073 = 0, $pbuffer$1 = 0, $rofs$0$lcssa$i = 0, $rofs$04$i = 0, $sum$i$neg = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($count|0)==(0); //@line 1479 "fastmix.cpp"
 if ($0) {
  $$0 = 0;
  return ($$0|0); //@line 1606 "fastmix.cpp"
 }
 $1 = HEAP32[10108]|0; //@line 1481 "fastmix.cpp"
 $2 = ($1>>>0)>(2); //@line 1481 "fastmix.cpp"
 if ($2) {
  $3 = $count << 3; //@line 1481 "fastmix.cpp"
  _memset((744652|0),0,($3|0))|0; //@line 2032 "fastmix.cpp"
 }
 $4 = ((($this)) + 59668|0); //@line 1484 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1484 "fastmix.cpp"
 $6 = ($5|0)==(0); //@line 1484 "fastmix.cpp"
 if ($6) {
  $$0 = 0;
  return ($$0|0); //@line 1606 "fastmix.cpp"
 }
 $7 = $count << 3; //@line 1529 "fastmix.cpp"
 $nChn$026 = 0;$nchmixed$027 = 0;$nchused$030 = 0;
 while(1) {
  $8 = (((($this)) + 32768|0) + ($nChn$026<<2)|0); //@line 1487 "fastmix.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 1487 "fastmix.cpp"
  $10 = (($this) + ($9<<8)|0); //@line 1487 "fastmix.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 1493 "fastmix.cpp"
  $12 = ($11|0)==(0|0); //@line 1493 "fastmix.cpp"
  do {
   if ($12) {
    $nchmixed$1 = $nchmixed$027;$nchused$1 = $nchused$030;
   } else {
    $13 = (((($this) + ($9<<8)|0)) + 36|0); //@line 1498 "fastmix.cpp"
    $14 = HEAP32[$13>>2]|0; //@line 1498 "fastmix.cpp"
    $15 = $14 & 1; //@line 1498 "fastmix.cpp"
    $16 = $14 >>> 5; //@line 1499 "fastmix.cpp"
    $17 = $16 & 2; //@line 1499 "fastmix.cpp"
    $18 = $17 | $15; //@line 1499 "fastmix.cpp"
    $19 = $14 >>> 11; //@line 1501 "fastmix.cpp"
    $20 = $19 & 8; //@line 1501 "fastmix.cpp"
    $21 = $18 | $20; //@line 1501 "fastmix.cpp"
    $22 = $14 & 4096; //@line 1503 "fastmix.cpp"
    $23 = ($22|0)==(0); //@line 1503 "fastmix.cpp"
    do {
     if ($23) {
      $24 = HEAP32[183080]|0; //@line 1506 "fastmix.cpp"
      $25 = $24 & 1040; //@line 1506 "fastmix.cpp"
      $26 = ($25|0)==(1040); //@line 1506 "fastmix.cpp"
      if ($26) {
       $27 = $21 | 48; //@line 1508 "fastmix.cpp"
       $nFlags$3 = $27;
       break;
      }
      $28 = $24 & 16; //@line 1509 "fastmix.cpp"
      $29 = ($28|0)==(0); //@line 1509 "fastmix.cpp"
      if ($29) {
       $31 = $21 | 16; //@line 1512 "fastmix.cpp"
       $nFlags$3 = $31;
       break;
      } else {
       $30 = $21 | 32; //@line 1510 "fastmix.cpp"
       $nFlags$3 = $30;
       break;
      }
     } else {
      $nFlags$3 = $21;
     }
    } while(0);
    $32 = ($nFlags$3>>>0)<(64); //@line 1514 "fastmix.cpp"
    if ($32) {
     $33 = (((($this) + ($9<<8)|0)) + 20|0); //@line 1514 "fastmix.cpp"
     $34 = HEAP32[$33>>2]|0; //@line 1514 "fastmix.cpp"
     $35 = (((($this) + ($9<<8)|0)) + 16|0); //@line 1514 "fastmix.cpp"
     $36 = HEAP32[$35>>2]|0; //@line 1514 "fastmix.cpp"
     $37 = ($34|0)==($36|0); //@line 1514 "fastmix.cpp"
     if ($37) {
      $38 = (((($this) + ($9<<8)|0)) + 92|0); //@line 1515 "fastmix.cpp"
      $39 = HEAP32[$38>>2]|0; //@line 1515 "fastmix.cpp"
      $40 = ($39|0)==(0); //@line 1515 "fastmix.cpp"
      if ($40) {
       $pMixFuncTable$0 = 46164;
      } else {
       $41 = (((($this) + ($9<<8)|0)) + 28|0); //@line 1515 "fastmix.cpp"
       $42 = HEAP32[$41>>2]|0; //@line 1515 "fastmix.cpp"
       $43 = (((($this) + ($9<<8)|0)) + 24|0); //@line 1515 "fastmix.cpp"
       $44 = HEAP32[$43>>2]|0; //@line 1515 "fastmix.cpp"
       $45 = ($42|0)==($44|0); //@line 1515 "fastmix.cpp"
       if ($45) {
        $pMixFuncTable$0 = 46164;
       } else {
        label = 17;
       }
      }
     } else {
      label = 17;
     }
    } else {
     label = 17;
    }
    if ((label|0) == 17) {
     label = 0;
     $pMixFuncTable$0 = 45908;
    }
    $46 = HEAP32[183080]|0; //@line 1524 "fastmix.cpp"
    $47 = $46 & 128; //@line 1524 "fastmix.cpp"
    $48 = ($47|0)!=(0); //@line 1524 "fastmix.cpp"
    $49 = $48 ? 740556 : 732364; //@line 1524 "fastmix.cpp"
    $50 = $14 & 134217728; //@line 1525 "fastmix.cpp"
    $51 = ($50|0)!=(0); //@line 1525 "fastmix.cpp"
    $$ = $51 ? 732364 : $49; //@line 1525 "fastmix.cpp"
    $52 = $14 & 67108864; //@line 1526 "fastmix.cpp"
    $53 = ($52|0)!=(0); //@line 1526 "fastmix.cpp"
    $$$ = $53 ? 740556 : $$; //@line 1526 "fastmix.cpp"
    $54 = ($$$|0)==(740556|0); //@line 1527 "fastmix.cpp"
    if ($54) {
     $55 = HEAP32[183090]|0; //@line 1529 "fastmix.cpp"
     $56 = ($55|0)==(0); //@line 1529 "fastmix.cpp"
     if ($56) {
      _memset((740556|0),0,($7|0))|0; //@line 1529 "fastmix.cpp"
     }
     $57 = (($55) + ($count))|0; //@line 1530 "fastmix.cpp"
     HEAP32[183090] = $57; //@line 1530 "fastmix.cpp"
    }
    $58 = (($nchused$030) + 1)|0; //@line 1535 "fastmix.cpp"
    $59 = (((($this) + ($9<<8)|0)) + 92|0); //@line 1539 "fastmix.cpp"
    $60 = (((($this) + ($9<<8)|0)) + 12|0); //@line 1366 "fastmix.cpp"
    $61 = (((($this) + ($9<<8)|0)) + 32|0); //@line 1368 "fastmix.cpp"
    $62 = (((($this) + ($9<<8)|0)) + 4|0); //@line 1370 "fastmix.cpp"
    $63 = (((($this) + ($9<<8)|0)) + 8|0); //@line 1375 "fastmix.cpp"
    $64 = (((($this) + ($9<<8)|0)) + 20|0); //@line 1561 "fastmix.cpp"
    $65 = (((($this) + ($9<<8)|0)) + 16|0); //@line 1561 "fastmix.cpp"
    $66 = $nFlags$3 | 4; //@line 1574 "fastmix.cpp"
    $67 = (((($this) + ($9<<8)|0)) + 84|0); //@line 1576 "fastmix.cpp"
    $68 = (((($this) + ($9<<8)|0)) + 88|0); //@line 1577 "fastmix.cpp"
    $69 = (((($this) + ($9<<8)|0)) + 100|0); //@line 1592 "fastmix.cpp"
    $70 = (((($this) + ($9<<8)|0)) + 104|0); //@line 1593 "fastmix.cpp"
    $71 = (((($this) + ($9<<8)|0)) + 28|0); //@line 1594 "fastmix.cpp"
    $72 = (((($this) + ($9<<8)|0)) + 24|0); //@line 1594 "fastmix.cpp"
    $73 = (((($this) + ($9<<8)|0)) + 124|0); //@line 1595 "fastmix.cpp"
    $74 = (((($this) + ($9<<8)|0)) + 40|0); //@line 1430 "fastmix.cpp"
    $$pre = HEAP32[$59>>2]|0; //@line 1539 "fastmix.cpp"
    $76 = $$pre;$nsamples$0 = $count;$pbuffer$0 = $$$;
    L36: while(1) {
     $75 = ($76|0)>(0); //@line 1539 "fastmix.cpp"
     $77 = ($nsamples$0|0)>($76|0); //@line 1541 "fastmix.cpp"
     $or$cond = $75 & $77; //@line 1539 "fastmix.cpp"
     $nrampsamples$0 = $or$cond ? $76 : $nsamples$0; //@line 1539 "fastmix.cpp"
     $78 = HEAP32[$13>>2]|0; //@line 1365 "fastmix.cpp"
     $79 = $78 & 2; //@line 1365 "fastmix.cpp"
     $80 = ($79|0)==(0); //@line 1365 "fastmix.cpp"
     if ($80) {
      $89 = 0;
     } else {
      $81 = HEAP32[$74>>2]|0; //@line 1365 "fastmix.cpp"
      $89 = $81;
     }
     $82 = HEAP32[$60>>2]|0; //@line 1366 "fastmix.cpp"
     $83 = ($nrampsamples$0|0)>(0); //@line 1368 "fastmix.cpp"
     $84 = ($82|0)!=(0); //@line 1368 "fastmix.cpp"
     $or$cond$i = $83 & $84; //@line 1368 "fastmix.cpp"
     if (!($or$cond$i)) {
      $nsamples$070 = $nsamples$0;$pbuffer$073 = $pbuffer$0;
      break;
     }
     $85 = HEAP32[$61>>2]|0; //@line 1368 "fastmix.cpp"
     $86 = ($85|0)==(0); //@line 1368 "fastmix.cpp"
     if ($86) {
      $nsamples$070 = $nsamples$0;$pbuffer$073 = $pbuffer$0;
      break;
     }
     $87 = HEAP32[$62>>2]|0; //@line 1370 "fastmix.cpp"
     $88 = ($89|0)>($87|0); //@line 1370 "fastmix.cpp"
     do {
      if ($88) {
       $90 = ($82|0)<(0); //@line 1372 "fastmix.cpp"
       if (!($90)) {
        $107 = ($87|0)<(0); //@line 1396 "fastmix.cpp"
        if (!($107)) {
         $131 = $87;$133 = $82;
         break;
        }
        HEAP32[$62>>2] = 0; //@line 1396 "fastmix.cpp"
        $131 = 0;$133 = $82;
        break;
       }
       $91 = (($89) - ($87))|0; //@line 1375 "fastmix.cpp"
       $92 = $91 << 16; //@line 1375 "fastmix.cpp"
       $93 = HEAP32[$63>>2]|0; //@line 1375 "fastmix.cpp"
       $94 = $93 & 65535; //@line 1375 "fastmix.cpp"
       $95 = (($92) - ($94))|0; //@line 1375 "fastmix.cpp"
       $96 = $95 >> 16; //@line 1376 "fastmix.cpp"
       $97 = $96 | $89; //@line 1376 "fastmix.cpp"
       HEAP32[$62>>2] = $97; //@line 1376 "fastmix.cpp"
       $98 = $95 & 65535; //@line 1377 "fastmix.cpp"
       HEAP32[$63>>2] = $98; //@line 1377 "fastmix.cpp"
       $99 = ($97|0)<($89|0); //@line 1378 "fastmix.cpp"
       if ($99) {
        label = 31;
       } else {
        $100 = (($85) + ($89))|0; //@line 1379 "fastmix.cpp"
        $101 = $100 >>> 1; //@line 1379 "fastmix.cpp"
        $102 = ($97>>>0)<($101>>>0); //@line 1379 "fastmix.cpp"
        if ($102) {
         $106 = $97;
        } else {
         label = 31;
        }
       }
       if ((label|0) == 31) {
        label = 0;
        HEAP32[$62>>2] = $89; //@line 1381 "fastmix.cpp"
        HEAP32[$63>>2] = 0; //@line 1381 "fastmix.cpp"
        $106 = $89;
       }
       $103 = (0 - ($82))|0; //@line 1383 "fastmix.cpp"
       HEAP32[$60>>2] = $103; //@line 1384 "fastmix.cpp"
       $104 = $78 & -129; //@line 1385 "fastmix.cpp"
       HEAP32[$13>>2] = $104; //@line 1385 "fastmix.cpp"
       $$not$i = $80 ^ 1; //@line 1386 "fastmix.cpp"
       $105 = ($106>>>0)<($85>>>0); //@line 1386 "fastmix.cpp"
       $or$cond12$i = $105 & $$not$i; //@line 1386 "fastmix.cpp"
       if ($or$cond12$i) {
        $131 = $106;$133 = $103;
       } else {
        $$lcssa = $85;$nsamples$0$lcssa68 = $nsamples$0;$pbuffer$0$lcssa71 = $pbuffer$0;
        label = 33;
        break L36;
       }
      } else {
       $108 = ($87>>>0)<($85>>>0); //@line 1400 "fastmix.cpp"
       if ($108) {
        $131 = $87;$133 = $82;
       } else {
        if ($80) {
         $nsamples$070 = $nsamples$0;$pbuffer$073 = $pbuffer$0;
         break L36;
        }
        $109 = $78 & 4; //@line 1403 "fastmix.cpp"
        $110 = ($109|0)==(0); //@line 1403 "fastmix.cpp"
        if ($110) {
         $124 = ($82|0)<(0); //@line 1422 "fastmix.cpp"
         if ($124) {
          $125 = (0 - ($82))|0; //@line 1424 "fastmix.cpp"
          HEAP32[$60>>2] = $125; //@line 1425 "fastmix.cpp"
          $nInc$1$i = $125;
         } else {
          $nInc$1$i = $82;
         }
         $126 = (($89) - ($85))|0; //@line 1428 "fastmix.cpp"
         $127 = (($87) + ($126))|0; //@line 1428 "fastmix.cpp"
         HEAP32[$62>>2] = $127; //@line 1428 "fastmix.cpp"
         $128 = ($127|0)<($89|0); //@line 1429 "fastmix.cpp"
         if (!($128)) {
          $131 = $127;$133 = $nInc$1$i;
          break;
         }
         $129 = HEAP32[$74>>2]|0; //@line 1430 "fastmix.cpp"
         HEAP32[$62>>2] = $129; //@line 1430 "fastmix.cpp"
         $131 = $129;$133 = $nInc$1$i;
         break;
        } else {
         $111 = ($82|0)>(0); //@line 1406 "fastmix.cpp"
         if ($111) {
          $112 = (0 - ($82))|0; //@line 1408 "fastmix.cpp"
          HEAP32[$60>>2] = $112; //@line 1409 "fastmix.cpp"
          $nInc$0$i = $112;
         } else {
          $nInc$0$i = $82;
         }
         $113 = $78 | 128; //@line 1411 "fastmix.cpp"
         HEAP32[$13>>2] = $113; //@line 1411 "fastmix.cpp"
         $114 = HEAP32[$63>>2]|0; //@line 1414 "fastmix.cpp"
         $115 = $114 & 65535; //@line 1414 "fastmix.cpp"
         $116 = (65536 - ($115))|0; //@line 1414 "fastmix.cpp"
         $117 = $116 >>> 16; //@line 1415 "fastmix.cpp"
         $factor$i = $85 << 1; //@line 1543 "fastmix.cpp"
         $sum$i$neg = (($factor$i) - ($87))|0; //@line 1415 "fastmix.cpp"
         $118 = (($sum$i$neg) - ($117))|0; //@line 1415 "fastmix.cpp"
         HEAP32[$62>>2] = $118; //@line 1415 "fastmix.cpp"
         $119 = $116 & 65535; //@line 1416 "fastmix.cpp"
         HEAP32[$63>>2] = $119; //@line 1416 "fastmix.cpp"
         $120 = HEAP32[$74>>2]|0; //@line 1417 "fastmix.cpp"
         $121 = ($118>>>0)>($120>>>0); //@line 1417 "fastmix.cpp"
         $122 = ($118>>>0)<($85>>>0); //@line 1418 "fastmix.cpp"
         $or$cond13$i = $121 & $122; //@line 1417 "fastmix.cpp"
         if ($or$cond13$i) {
          $131 = $118;$133 = $nInc$0$i;
          break;
         }
         $123 = (($85) + -1)|0; //@line 1419 "fastmix.cpp"
         HEAP32[$62>>2] = $123; //@line 1419 "fastmix.cpp"
         $131 = $123;$133 = $nInc$0$i;
         break;
        }
       }
      }
     } while(0);
     $130 = ($131|0)<($89|0); //@line 1435 "fastmix.cpp"
     $132 = $133 | $131; //@line 1543 "fastmix.cpp"
     $134 = ($132|0)<(0); //@line 1437 "fastmix.cpp"
     $or$cond7$i = $130 & $134; //@line 1435 "fastmix.cpp"
     $135 = ($131|0)<(0); //@line 1439 "fastmix.cpp"
     $or$cond8$i = $135 | $or$cond7$i; //@line 1435 "fastmix.cpp"
     $or$cond8$not$i = $or$cond8$i ^ 1; //@line 1435 "fastmix.cpp"
     $136 = ($131|0)<($85|0); //@line 1439 "fastmix.cpp"
     $or$cond14$i = $136 & $or$cond8$not$i; //@line 1435 "fastmix.cpp"
     if (!($or$cond14$i)) {
      $nsamples$070 = $nsamples$0;$pbuffer$073 = $pbuffer$0;
      break;
     }
     $137 = HEAP32[$63>>2]|0; //@line 1440 "fastmix.cpp"
     $138 = $137 & 65535; //@line 1440 "fastmix.cpp"
     $139 = ($133|0)<(0); //@line 1441 "fastmix.cpp"
     if ($139) {
      $140 = (0 - ($133))|0; //@line 1443 "fastmix.cpp"
      $141 = $140 >> 16; //@line 1444 "fastmix.cpp"
      $142 = (($141) + 1)|0; //@line 1444 "fastmix.cpp"
      $143 = (16384 / ($142|0))&-1; //@line 1444 "fastmix.cpp"
      $144 = ($143|0)<(2); //@line 1445 "fastmix.cpp"
      $$$i = $144 ? 2 : $143; //@line 1445 "fastmix.cpp"
      $145 = ($$$i|0)<($nrampsamples$0|0); //@line 1446 "fastmix.cpp"
      $$$nSamples$i = $145 ? $$$i : $nrampsamples$0; //@line 1446 "fastmix.cpp"
      $146 = (($$$nSamples$i) + -1)|0; //@line 1447 "fastmix.cpp"
      $147 = Math_imul($146, $141)|0; //@line 1447 "fastmix.cpp"
      $148 = $140 & 65535; //@line 1448 "fastmix.cpp"
      $149 = Math_imul($146, $148)|0; //@line 1448 "fastmix.cpp"
      $150 = (($131) - ($147))|0; //@line 1449 "fastmix.cpp"
      $151 = (($138) - ($149))|0; //@line 1449 "fastmix.cpp"
      $152 = $151 >> 16; //@line 1449 "fastmix.cpp"
      $153 = (($150) + ($152))|0; //@line 1449 "fastmix.cpp"
      $154 = ($153|0)<($89|0); //@line 1450 "fastmix.cpp"
      if ($154) {
       $155 = ($131|0)<(0); //@line 1452 "fastmix.cpp"
       $156 = $155 << 31 >> 31; //@line 1452 "fastmix.cpp"
       $157 = ($89|0)<(0); //@line 1452 "fastmix.cpp"
       $158 = $157 << 31 >> 31; //@line 1452 "fastmix.cpp"
       $159 = (_i64Subtract(($131|0),($156|0),($89|0),($158|0))|0); //@line 1452 "fastmix.cpp"
       $160 = tempRet0; //@line 1452 "fastmix.cpp"
       $161 = (_bitshift64Shl(($159|0),($160|0),16)|0); //@line 1452 "fastmix.cpp"
       $162 = tempRet0; //@line 1452 "fastmix.cpp"
       $163 = $138 | $161; //@line 1452 "fastmix.cpp"
       $164 = (_i64Add(($163|0),($162|0),-1,-1)|0); //@line 1452 "fastmix.cpp"
       $165 = tempRet0; //@line 1452 "fastmix.cpp"
       $166 = ($140|0)<(0); //@line 1452 "fastmix.cpp"
       $167 = $166 << 31 >> 31; //@line 1452 "fastmix.cpp"
       $168 = (___divdi3(($164|0),($165|0),($140|0),($167|0))|0); //@line 1452 "fastmix.cpp"
       $169 = tempRet0; //@line 1452 "fastmix.cpp"
       $170 = (($168) + 1)|0; //@line 1452 "fastmix.cpp"
       $$2$i = $$$nSamples$i;$nSmpCount$0$i = $170;
      } else {
       $$2$i = $$$nSamples$i;$nSmpCount$0$i = $nrampsamples$0;
      }
     } else {
      $171 = $133 >> 16; //@line 1456 "fastmix.cpp"
      $172 = (($171) + 1)|0; //@line 1456 "fastmix.cpp"
      $173 = (16384 / ($172|0))&-1; //@line 1456 "fastmix.cpp"
      $174 = ($173|0)<(2); //@line 1457 "fastmix.cpp"
      $$4$i = $174 ? 2 : $173; //@line 1457 "fastmix.cpp"
      $175 = ($$4$i|0)<($nrampsamples$0|0); //@line 1458 "fastmix.cpp"
      $$4$nSamples$i = $175 ? $$4$i : $nrampsamples$0; //@line 1458 "fastmix.cpp"
      $176 = (($$4$nSamples$i) + -1)|0; //@line 1459 "fastmix.cpp"
      $177 = Math_imul($176, $171)|0; //@line 1459 "fastmix.cpp"
      $178 = $133 & 65535; //@line 1460 "fastmix.cpp"
      $179 = Math_imul($176, $178)|0; //@line 1460 "fastmix.cpp"
      $180 = (($177) + ($131))|0; //@line 1461 "fastmix.cpp"
      $181 = (($179) + ($138))|0; //@line 1461 "fastmix.cpp"
      $182 = $181 >> 16; //@line 1461 "fastmix.cpp"
      $183 = (($180) + ($182))|0; //@line 1461 "fastmix.cpp"
      $184 = ($183|0)<($85|0); //@line 1462 "fastmix.cpp"
      if ($184) {
       $$2$i = $$4$nSamples$i;$nSmpCount$0$i = $nrampsamples$0;
      } else {
       $185 = ($131|0)<(0); //@line 1464 "fastmix.cpp"
       $186 = $185 << 31 >> 31; //@line 1464 "fastmix.cpp"
       $187 = (_i64Subtract(($85|0),0,($131|0),($186|0))|0); //@line 1464 "fastmix.cpp"
       $188 = tempRet0; //@line 1464 "fastmix.cpp"
       $189 = (_bitshift64Shl(($187|0),($188|0),16)|0); //@line 1464 "fastmix.cpp"
       $190 = tempRet0; //@line 1464 "fastmix.cpp"
       $191 = (_i64Add(($189|0),($190|0),-1,-1)|0); //@line 1464 "fastmix.cpp"
       $192 = tempRet0; //@line 1464 "fastmix.cpp"
       $193 = (_i64Subtract(($191|0),($192|0),($138|0),0)|0); //@line 1464 "fastmix.cpp"
       $194 = tempRet0; //@line 1464 "fastmix.cpp"
       $195 = ($133|0)<(0); //@line 1464 "fastmix.cpp"
       $196 = $195 << 31 >> 31; //@line 1464 "fastmix.cpp"
       $197 = (___divdi3(($193|0),($194|0),($133|0),($196|0))|0); //@line 1464 "fastmix.cpp"
       $198 = tempRet0; //@line 1464 "fastmix.cpp"
       $199 = (($197) + 1)|0; //@line 1464 "fastmix.cpp"
       $$2$i = $$4$nSamples$i;$nSmpCount$0$i = $199;
      }
     }
     $200 = ($nSmpCount$0$i|0)<(2); //@line 1467 "fastmix.cpp"
     if ($200) {
      $$0$i7 = 1;
     } else {
      $201 = ($nSmpCount$0$i|0)>($$2$i|0); //@line 1468 "fastmix.cpp"
      if ($201) {
       $202 = ($$2$i|0)<(1); //@line 1543 "fastmix.cpp"
       if ($202) {
        $nsamples$070 = $nsamples$0;$pbuffer$073 = $pbuffer$0;
        break;
       } else {
        $$0$i7 = $$2$i;
       }
      } else {
       $$0$i7 = $nSmpCount$0$i;
      }
     }
     $237 = HEAP32[10107]|0; //@line 1560 "fastmix.cpp"
     $238 = ($nchmixed$027>>>0)<($237>>>0); //@line 1560 "fastmix.cpp"
     if ($238) {
      label = 63;
     } else {
      $239 = HEAP32[183080]|0; //@line 1560 "fastmix.cpp"
      $240 = $239 & 65536; //@line 1560 "fastmix.cpp"
      $241 = ($240|0)==(0); //@line 1560 "fastmix.cpp"
      if ($241) {
       label = 65;
      } else {
       label = 63;
      }
     }
     do {
      if ((label|0) == 63) {
       label = 0;
       $242 = ($76|0)==(0); //@line 1561 "fastmix.cpp"
       if ($242) {
        $243 = HEAP32[$64>>2]|0; //@line 1561 "fastmix.cpp"
        $244 = HEAP32[$65>>2]|0; //@line 1561 "fastmix.cpp"
        $245 = $244 | $243; //@line 1561 "fastmix.cpp"
        $246 = ($245|0)==(0); //@line 1561 "fastmix.cpp"
        if ($246) {
         label = 65;
         break;
        }
       }
       $$pn = $242 ? $nFlags$3 : $66; //@line 1574 "fastmix.cpp"
       $$in = (($pMixFuncTable$0) + ($$pn<<2)|0); //@line 1574 "fastmix.cpp"
       $254 = HEAP32[$$in>>2]|0; //@line 1574 "fastmix.cpp"
       $255 = $$0$i7 << 1; //@line 1575 "fastmix.cpp"
       $256 = (($pbuffer$0) + ($255<<2)|0); //@line 1575 "fastmix.cpp"
       $$sum = (($255) + -2)|0; //@line 1576 "fastmix.cpp"
       $257 = (($pbuffer$0) + ($$sum<<2)|0); //@line 1576 "fastmix.cpp"
       $258 = HEAP32[$257>>2]|0; //@line 1576 "fastmix.cpp"
       $259 = (0 - ($258))|0; //@line 1576 "fastmix.cpp"
       HEAP32[$67>>2] = $259; //@line 1576 "fastmix.cpp"
       $$sum1 = (($255) + -1)|0; //@line 1577 "fastmix.cpp"
       $260 = (($pbuffer$0) + ($$sum1<<2)|0); //@line 1577 "fastmix.cpp"
       $261 = HEAP32[$260>>2]|0; //@line 1577 "fastmix.cpp"
       $262 = (0 - ($261))|0; //@line 1577 "fastmix.cpp"
       HEAP32[$68>>2] = $262; //@line 1577 "fastmix.cpp"
       FUNCTION_TABLE_viii[$254 & 8191]($10,$pbuffer$0,$256); //@line 1578 "fastmix.cpp"
       $263 = HEAP32[$257>>2]|0; //@line 1579 "fastmix.cpp"
       $264 = HEAP32[$67>>2]|0; //@line 1579 "fastmix.cpp"
       $265 = (($264) + ($263))|0; //@line 1579 "fastmix.cpp"
       HEAP32[$67>>2] = $265; //@line 1579 "fastmix.cpp"
       $266 = HEAP32[$260>>2]|0; //@line 1580 "fastmix.cpp"
       $267 = HEAP32[$68>>2]|0; //@line 1580 "fastmix.cpp"
       $268 = (($267) + ($266))|0; //@line 1580 "fastmix.cpp"
       HEAP32[$68>>2] = $268; //@line 1580 "fastmix.cpp"
       $$pre40 = HEAP32[$59>>2]|0; //@line 1586 "fastmix.cpp"
       $271 = $$pre40;$naddmix$0 = 1;$pbuffer$1 = $256;
      }
     } while(0);
     if ((label|0) == 65) {
      label = 0;
      $247 = Math_imul($133, $$0$i7)|0; //@line 1563 "fastmix.cpp"
      $248 = (($247) + ($137))|0; //@line 1563 "fastmix.cpp"
      $249 = $248 & 65535; //@line 1564 "fastmix.cpp"
      HEAP32[$63>>2] = $249; //@line 1564 "fastmix.cpp"
      $250 = $248 >> 16; //@line 1565 "fastmix.cpp"
      $251 = (($250) + ($131))|0; //@line 1565 "fastmix.cpp"
      HEAP32[$62>>2] = $251; //@line 1565 "fastmix.cpp"
      HEAP32[$68>>2] = 0; //@line 1566 "fastmix.cpp"
      HEAP32[$67>>2] = 0; //@line 1566 "fastmix.cpp"
      $252 = $$0$i7 << 1; //@line 1567 "fastmix.cpp"
      $253 = (($pbuffer$0) + ($252<<2)|0); //@line 1567 "fastmix.cpp"
      $271 = $76;$naddmix$0 = 0;$pbuffer$1 = $253;
     }
     $269 = (($nsamples$0) - ($$0$i7))|0; //@line 1585 "fastmix.cpp"
     $270 = ($271|0)==(0); //@line 1586 "fastmix.cpp"
     do {
      if ($270) {
       $286 = 0;
      } else {
       $272 = (($271) - ($$0$i7))|0; //@line 1588 "fastmix.cpp"
       HEAP32[$59>>2] = $272; //@line 1588 "fastmix.cpp"
       $273 = ($272|0)<(1); //@line 1589 "fastmix.cpp"
       if (!($273)) {
        $286 = $272;
        break;
       }
       HEAP32[$59>>2] = 0; //@line 1591 "fastmix.cpp"
       $274 = HEAP32[$69>>2]|0; //@line 1592 "fastmix.cpp"
       HEAP32[$65>>2] = $274; //@line 1592 "fastmix.cpp"
       $275 = HEAP32[$70>>2]|0; //@line 1593 "fastmix.cpp"
       HEAP32[$64>>2] = $275; //@line 1593 "fastmix.cpp"
       HEAP32[$71>>2] = 0; //@line 1594 "fastmix.cpp"
       HEAP32[$72>>2] = 0; //@line 1594 "fastmix.cpp"
       $276 = HEAP32[$13>>2]|0; //@line 1595 "fastmix.cpp"
       $277 = $276 & 1024; //@line 1595 "fastmix.cpp"
       $278 = ($277|0)==(0); //@line 1595 "fastmix.cpp"
       if ($278) {
        $286 = 0;
        break;
       }
       $279 = HEAP32[$73>>2]|0; //@line 1595 "fastmix.cpp"
       $280 = ($279|0)==(0); //@line 1595 "fastmix.cpp"
       if (!($280)) {
        $286 = 0;
        break;
       }
       HEAP32[$61>>2] = 0; //@line 1597 "fastmix.cpp"
       HEAP32[$10>>2] = 0; //@line 1598 "fastmix.cpp"
       $286 = 0;
      }
     } while(0);
     $281 = ($269|0)>(0); //@line 1602 "fastmix.cpp"
     if ($281) {
      $76 = $286;$nsamples$0 = $269;$pbuffer$0 = $pbuffer$1;
     } else {
      $naddmix$0$lcssa = $naddmix$0;
      label = 73;
      break;
     }
    }
    if ((label|0) == 33) {
     label = 0;
     HEAP32[$62>>2] = $$lcssa; //@line 1388 "fastmix.cpp"
     HEAP32[$63>>2] = 0; //@line 1389 "fastmix.cpp"
     $nsamples$070 = $nsamples$0$lcssa68;$pbuffer$073 = $pbuffer$0$lcssa71;
    }
    else if ((label|0) == 73) {
     label = 0;
     $282 = (($naddmix$0$lcssa) + ($nchmixed$027))|0; //@line 1603 "fastmix.cpp"
     $nchmixed$1 = $282;$nchused$1 = $58;
     break;
    }
    HEAP32[$10>>2] = 0; //@line 1546 "fastmix.cpp"
    HEAP32[$61>>2] = 0; //@line 1547 "fastmix.cpp"
    HEAP32[$62>>2] = 0; //@line 1548 "fastmix.cpp"
    HEAP32[$63>>2] = 0; //@line 1549 "fastmix.cpp"
    HEAP32[$59>>2] = 0; //@line 1550 "fastmix.cpp"
    $203 = HEAP32[$67>>2]|0; //@line 2266 "fastmix.cpp"
    $204 = HEAP32[$68>>2]|0; //@line 2267 "fastmix.cpp"
    $205 = $204 | $203; //@line 2269 "fastmix.cpp"
    $206 = ($205|0)==(0); //@line 2269 "fastmix.cpp"
    if ($206) {
     $231 = $203;
    } else {
     $207 = ($nsamples$070|0)==(0); //@line 2270 "fastmix.cpp"
     if ($207) {
      $lofs$0$lcssa$i = $204;$rofs$0$lcssa$i = $203;
     } else {
      $i$02$i = 0;$lofs$03$i = $204;$rofs$04$i = $203;
      while(1) {
       $208 = (0 - ($rofs$04$i))|0; //@line 2272 "fastmix.cpp"
       $209 = $208 >> 31; //@line 2272 "fastmix.cpp"
       $210 = $209 & 255; //@line 2272 "fastmix.cpp"
       $211 = (($210) + ($rofs$04$i))|0; //@line 2272 "fastmix.cpp"
       $212 = $211 >> 8; //@line 2272 "fastmix.cpp"
       $213 = (0 - ($lofs$03$i))|0; //@line 2273 "fastmix.cpp"
       $214 = $213 >> 31; //@line 2273 "fastmix.cpp"
       $215 = $214 & 255; //@line 2273 "fastmix.cpp"
       $216 = (($215) + ($lofs$03$i))|0; //@line 2273 "fastmix.cpp"
       $217 = $216 >> 8; //@line 2273 "fastmix.cpp"
       $218 = (($rofs$04$i) - ($212))|0; //@line 2274 "fastmix.cpp"
       $219 = (($lofs$03$i) - ($217))|0; //@line 2275 "fastmix.cpp"
       $220 = $i$02$i << 1; //@line 2276 "fastmix.cpp"
       $221 = (($pbuffer$073) + ($220<<2)|0); //@line 2276 "fastmix.cpp"
       $222 = HEAP32[$221>>2]|0; //@line 2276 "fastmix.cpp"
       $223 = (($212) + ($222))|0; //@line 2276 "fastmix.cpp"
       HEAP32[$221>>2] = $223; //@line 2276 "fastmix.cpp"
       $224 = $220 | 1; //@line 2277 "fastmix.cpp"
       $225 = (($pbuffer$073) + ($224<<2)|0); //@line 2277 "fastmix.cpp"
       $226 = HEAP32[$225>>2]|0; //@line 2277 "fastmix.cpp"
       $227 = (($217) + ($226))|0; //@line 2277 "fastmix.cpp"
       HEAP32[$225>>2] = $227; //@line 2277 "fastmix.cpp"
       $228 = (($i$02$i) + 1)|0; //@line 2270 "fastmix.cpp"
       $exitcond$i = ($228|0)==($nsamples$070|0); //@line 2270 "fastmix.cpp"
       if ($exitcond$i) {
        $lofs$0$lcssa$i = $219;$rofs$0$lcssa$i = $218;
        break;
       } else {
        $i$02$i = $228;$lofs$03$i = $219;$rofs$04$i = $218;
       }
      }
     }
     HEAP32[$67>>2] = $rofs$0$lcssa$i; //@line 2279 "fastmix.cpp"
     HEAP32[$68>>2] = $lofs$0$lcssa$i; //@line 2280 "fastmix.cpp"
     $231 = $rofs$0$lcssa$i;
    }
    $229 = HEAP32[183085]|0; //@line 1552 "fastmix.cpp"
    $230 = (($229) + ($231))|0; //@line 1552 "fastmix.cpp"
    HEAP32[183085] = $230; //@line 1552 "fastmix.cpp"
    $232 = HEAP32[$68>>2]|0; //@line 1553 "fastmix.cpp"
    $233 = HEAP32[183086]|0; //@line 1553 "fastmix.cpp"
    $234 = (($233) + ($232))|0; //@line 1553 "fastmix.cpp"
    HEAP32[183086] = $234; //@line 1553 "fastmix.cpp"
    HEAP32[$68>>2] = 0; //@line 1554 "fastmix.cpp"
    HEAP32[$67>>2] = 0; //@line 1554 "fastmix.cpp"
    $235 = HEAP32[$13>>2]|0; //@line 1555 "fastmix.cpp"
    $236 = $235 & -129; //@line 1555 "fastmix.cpp"
    HEAP32[$13>>2] = $236; //@line 1555 "fastmix.cpp"
    $nchmixed$1 = $nchmixed$027;$nchused$1 = $58;
   }
  } while(0);
  $283 = (($nChn$026) + 1)|0; //@line 1484 "fastmix.cpp"
  $284 = HEAP32[$4>>2]|0; //@line 1484 "fastmix.cpp"
  $285 = ($283>>>0)<($284>>>0); //@line 1484 "fastmix.cpp"
  if ($285) {
   $nChn$026 = $283;$nchmixed$027 = $nchmixed$1;$nchused$030 = $nchused$1;
  } else {
   $$0 = $nchused$1;
   break;
  }
 }
 return ($$0|0); //@line 1606 "fastmix.cpp"
}
function __Z11Mono8BitMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$sum = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $pvol$0 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 678 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 678 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 678 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 678 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 678 "fastmix.cpp"
 $5 = ((($pChannel)) + 36|0); //@line 678 "fastmix.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 678 "fastmix.cpp"
 $7 = $6 & 64; //@line 678 "fastmix.cpp"
 $8 = ($7|0)==(0); //@line 678 "fastmix.cpp"
 $9 = (($2) + ($4)|0); //@line 678 "fastmix.cpp"
 $$ = $8 ? $2 : $9; //@line 678 "fastmix.cpp"
 $10 = ((($pChannel)) + 16|0); //@line 680 "fastmix.cpp"
 $11 = ((($pChannel)) + 20|0); //@line 680 "fastmix.cpp"
 $12 = ((($pChannel)) + 12|0); //@line 681 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $13 = $nPos$0 >> 16; //@line 679 "fastmix.cpp"
  $$sum = (($13) + ($4))|0; //@line 679 "fastmix.cpp"
  $14 = (($$) + ($$sum)|0); //@line 679 "fastmix.cpp"
  $15 = HEAP8[$14>>0]|0; //@line 679 "fastmix.cpp"
  $16 = $15 << 24 >> 24; //@line 679 "fastmix.cpp"
  $17 = $16 << 8; //@line 679 "fastmix.cpp"
  $18 = HEAP32[$10>>2]|0; //@line 680 "fastmix.cpp"
  $19 = Math_imul($17, $18)|0; //@line 680 "fastmix.cpp"
  $20 = HEAP32[$pvol$0>>2]|0; //@line 680 "fastmix.cpp"
  $21 = (($19) + ($20))|0; //@line 680 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $21; //@line 680 "fastmix.cpp"
  $22 = HEAP32[$11>>2]|0; //@line 680 "fastmix.cpp"
  $23 = Math_imul($22, $17)|0; //@line 680 "fastmix.cpp"
  $24 = ((($pvol$0)) + 4|0); //@line 680 "fastmix.cpp"
  $25 = HEAP32[$24>>2]|0; //@line 680 "fastmix.cpp"
  $26 = (($25) + ($23))|0; //@line 680 "fastmix.cpp"
  HEAP32[$24>>2] = $26; //@line 680 "fastmix.cpp"
  $27 = ((($pvol$0)) + 8|0); //@line 680 "fastmix.cpp"
  $28 = HEAP32[$12>>2]|0; //@line 681 "fastmix.cpp"
  $29 = (($28) + ($nPos$0))|0; //@line 681 "fastmix.cpp"
  $30 = ($27>>>0)<($pbufmax>>>0); //@line 681 "fastmix.cpp"
  if ($30) {
   $nPos$0 = $29;$pvol$0 = $27;
  } else {
   $$lcssa = $29;
   break;
  }
 }
 $31 = $$lcssa >> 16; //@line 681 "fastmix.cpp"
 $32 = HEAP32[$3>>2]|0; //@line 681 "fastmix.cpp"
 $33 = (($32) + ($31))|0; //@line 681 "fastmix.cpp"
 HEAP32[$3>>2] = $33; //@line 681 "fastmix.cpp"
 $34 = $$lcssa & 65535; //@line 681 "fastmix.cpp"
 HEAP32[$0>>2] = $34; //@line 681 "fastmix.cpp"
 return; //@line 681 "fastmix.cpp"
}
function __Z12Mono16BitMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $pvol$0 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 684 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 684 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 684 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 684 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 684 "fastmix.cpp"
 $5 = $4 << 1; //@line 684 "fastmix.cpp"
 $6 = (($2) + ($5)|0); //@line 684 "fastmix.cpp"
 $7 = ((($pChannel)) + 36|0); //@line 684 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 684 "fastmix.cpp"
 $9 = $8 & 64; //@line 684 "fastmix.cpp"
 $10 = ($9|0)==(0); //@line 684 "fastmix.cpp"
 $11 = (($6) + ($4<<1)|0); //@line 684 "fastmix.cpp"
 $$ = $10 ? $6 : $11; //@line 684 "fastmix.cpp"
 $12 = ((($pChannel)) + 16|0); //@line 686 "fastmix.cpp"
 $13 = ((($pChannel)) + 20|0); //@line 686 "fastmix.cpp"
 $14 = ((($pChannel)) + 12|0); //@line 687 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $15 = $nPos$0 >> 16; //@line 685 "fastmix.cpp"
  $16 = (($$) + ($15<<1)|0); //@line 685 "fastmix.cpp"
  $17 = HEAP16[$16>>1]|0; //@line 685 "fastmix.cpp"
  $18 = $17 << 16 >> 16; //@line 685 "fastmix.cpp"
  $19 = HEAP32[$12>>2]|0; //@line 686 "fastmix.cpp"
  $20 = Math_imul($18, $19)|0; //@line 686 "fastmix.cpp"
  $21 = HEAP32[$pvol$0>>2]|0; //@line 686 "fastmix.cpp"
  $22 = (($21) + ($20))|0; //@line 686 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $22; //@line 686 "fastmix.cpp"
  $23 = HEAP32[$13>>2]|0; //@line 686 "fastmix.cpp"
  $24 = Math_imul($23, $18)|0; //@line 686 "fastmix.cpp"
  $25 = ((($pvol$0)) + 4|0); //@line 686 "fastmix.cpp"
  $26 = HEAP32[$25>>2]|0; //@line 686 "fastmix.cpp"
  $27 = (($26) + ($24))|0; //@line 686 "fastmix.cpp"
  HEAP32[$25>>2] = $27; //@line 686 "fastmix.cpp"
  $28 = ((($pvol$0)) + 8|0); //@line 686 "fastmix.cpp"
  $29 = HEAP32[$14>>2]|0; //@line 687 "fastmix.cpp"
  $30 = (($29) + ($nPos$0))|0; //@line 687 "fastmix.cpp"
  $31 = ($28>>>0)<($pbufmax>>>0); //@line 687 "fastmix.cpp"
  if ($31) {
   $nPos$0 = $30;$pvol$0 = $28;
  } else {
   $$lcssa = $30;
   break;
  }
 }
 $32 = $$lcssa >> 16; //@line 687 "fastmix.cpp"
 $33 = HEAP32[$3>>2]|0; //@line 687 "fastmix.cpp"
 $34 = (($33) + ($32))|0; //@line 687 "fastmix.cpp"
 HEAP32[$3>>2] = $34; //@line 687 "fastmix.cpp"
 $35 = $$lcssa & 65535; //@line 687 "fastmix.cpp"
 HEAP32[$0>>2] = $35; //@line 687 "fastmix.cpp"
 return; //@line 687 "fastmix.cpp"
}
function __Z13Stereo8BitMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$sum = 0, $$sum1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 882 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 882 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 882 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 882 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 882 "fastmix.cpp"
 $5 = (($2) + ($4)|0); //@line 882 "fastmix.cpp"
 $6 = ((($pChannel)) + 36|0); //@line 882 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 882 "fastmix.cpp"
 $8 = $7 & 64; //@line 882 "fastmix.cpp"
 $9 = ($8|0)==(0); //@line 882 "fastmix.cpp"
 $$ = $9 ? $2 : $5; //@line 882 "fastmix.cpp"
 $10 = ((($pChannel)) + 16|0); //@line 884 "fastmix.cpp"
 $11 = ((($pChannel)) + 20|0); //@line 884 "fastmix.cpp"
 $12 = ((($pChannel)) + 12|0); //@line 885 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $13 = $nPos$0 >> 16; //@line 883 "fastmix.cpp"
  $14 = $13 << 1; //@line 883 "fastmix.cpp"
  $$sum = (($14) + ($4))|0; //@line 883 "fastmix.cpp"
  $15 = (($$) + ($$sum)|0); //@line 883 "fastmix.cpp"
  $16 = HEAP8[$15>>0]|0; //@line 883 "fastmix.cpp"
  $17 = $16 << 24 >> 24; //@line 883 "fastmix.cpp"
  $18 = $17 << 8; //@line 883 "fastmix.cpp"
  $19 = $14 | 1; //@line 883 "fastmix.cpp"
  $$sum1 = (($19) + ($4))|0; //@line 883 "fastmix.cpp"
  $20 = (($$) + ($$sum1)|0); //@line 883 "fastmix.cpp"
  $21 = HEAP8[$20>>0]|0; //@line 883 "fastmix.cpp"
  $22 = $21 << 24 >> 24; //@line 883 "fastmix.cpp"
  $23 = $22 << 8; //@line 883 "fastmix.cpp"
  $24 = HEAP32[$10>>2]|0; //@line 884 "fastmix.cpp"
  $25 = Math_imul($18, $24)|0; //@line 884 "fastmix.cpp"
  $26 = HEAP32[$pvol$0>>2]|0; //@line 884 "fastmix.cpp"
  $27 = (($26) + ($25))|0; //@line 884 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $27; //@line 884 "fastmix.cpp"
  $28 = HEAP32[$11>>2]|0; //@line 884 "fastmix.cpp"
  $29 = Math_imul($23, $28)|0; //@line 884 "fastmix.cpp"
  $30 = ((($pvol$0)) + 4|0); //@line 884 "fastmix.cpp"
  $31 = HEAP32[$30>>2]|0; //@line 884 "fastmix.cpp"
  $32 = (($31) + ($29))|0; //@line 884 "fastmix.cpp"
  HEAP32[$30>>2] = $32; //@line 884 "fastmix.cpp"
  $33 = ((($pvol$0)) + 8|0); //@line 884 "fastmix.cpp"
  $34 = HEAP32[$12>>2]|0; //@line 885 "fastmix.cpp"
  $35 = (($34) + ($nPos$0))|0; //@line 885 "fastmix.cpp"
  $36 = ($33>>>0)<($pbufmax>>>0); //@line 885 "fastmix.cpp"
  if ($36) {
   $nPos$0 = $35;$pvol$0 = $33;
  } else {
   $$lcssa = $35;
   break;
  }
 }
 $37 = $$lcssa >> 16; //@line 885 "fastmix.cpp"
 $38 = HEAP32[$3>>2]|0; //@line 885 "fastmix.cpp"
 $39 = (($38) + ($37))|0; //@line 885 "fastmix.cpp"
 HEAP32[$3>>2] = $39; //@line 885 "fastmix.cpp"
 $40 = $$lcssa & 65535; //@line 885 "fastmix.cpp"
 HEAP32[$0>>2] = $40; //@line 885 "fastmix.cpp"
 return; //@line 885 "fastmix.cpp"
}
function __Z14Stereo16BitMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 888 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 888 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 888 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 888 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 888 "fastmix.cpp"
 $5 = $4 << 1; //@line 888 "fastmix.cpp"
 $6 = (($2) + ($5)|0); //@line 888 "fastmix.cpp"
 $7 = ((($pChannel)) + 36|0); //@line 888 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 888 "fastmix.cpp"
 $9 = $8 & 64; //@line 888 "fastmix.cpp"
 $10 = ($9|0)==(0); //@line 888 "fastmix.cpp"
 $11 = (($6) + ($4<<1)|0); //@line 888 "fastmix.cpp"
 $$ = $10 ? $6 : $11; //@line 888 "fastmix.cpp"
 $12 = ((($pChannel)) + 16|0); //@line 890 "fastmix.cpp"
 $13 = ((($pChannel)) + 20|0); //@line 890 "fastmix.cpp"
 $14 = ((($pChannel)) + 12|0); //@line 891 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $15 = $nPos$0 >> 16; //@line 889 "fastmix.cpp"
  $16 = $15 << 1; //@line 889 "fastmix.cpp"
  $17 = (($$) + ($16<<1)|0); //@line 889 "fastmix.cpp"
  $18 = HEAP16[$17>>1]|0; //@line 889 "fastmix.cpp"
  $19 = $18 << 16 >> 16; //@line 889 "fastmix.cpp"
  $20 = $16 | 1; //@line 889 "fastmix.cpp"
  $21 = (($$) + ($20<<1)|0); //@line 889 "fastmix.cpp"
  $22 = HEAP16[$21>>1]|0; //@line 889 "fastmix.cpp"
  $23 = $22 << 16 >> 16; //@line 889 "fastmix.cpp"
  $24 = HEAP32[$12>>2]|0; //@line 890 "fastmix.cpp"
  $25 = Math_imul($24, $19)|0; //@line 890 "fastmix.cpp"
  $26 = HEAP32[$pvol$0>>2]|0; //@line 890 "fastmix.cpp"
  $27 = (($26) + ($25))|0; //@line 890 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $27; //@line 890 "fastmix.cpp"
  $28 = HEAP32[$13>>2]|0; //@line 890 "fastmix.cpp"
  $29 = Math_imul($28, $23)|0; //@line 890 "fastmix.cpp"
  $30 = ((($pvol$0)) + 4|0); //@line 890 "fastmix.cpp"
  $31 = HEAP32[$30>>2]|0; //@line 890 "fastmix.cpp"
  $32 = (($31) + ($29))|0; //@line 890 "fastmix.cpp"
  HEAP32[$30>>2] = $32; //@line 890 "fastmix.cpp"
  $33 = ((($pvol$0)) + 8|0); //@line 890 "fastmix.cpp"
  $34 = HEAP32[$14>>2]|0; //@line 891 "fastmix.cpp"
  $35 = (($34) + ($nPos$0))|0; //@line 891 "fastmix.cpp"
  $36 = ($33>>>0)<($pbufmax>>>0); //@line 891 "fastmix.cpp"
  if ($36) {
   $nPos$0 = $35;$pvol$0 = $33;
  } else {
   $$lcssa = $35;
   break;
  }
 }
 $37 = $$lcssa >> 16; //@line 891 "fastmix.cpp"
 $38 = HEAP32[$3>>2]|0; //@line 891 "fastmix.cpp"
 $39 = (($38) + ($37))|0; //@line 891 "fastmix.cpp"
 HEAP32[$3>>2] = $39; //@line 891 "fastmix.cpp"
 $40 = $$lcssa & 65535; //@line 891 "fastmix.cpp"
 HEAP32[$0>>2] = $40; //@line 891 "fastmix.cpp"
 return; //@line 891 "fastmix.cpp"
}
function __Z15Mono8BitRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa22 = 0, $$lcssa23 = 0, $$lcssa24 = 0, $$lcssa25 = 0, $$sum = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 727 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 727 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 727 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 727 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 728 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 728 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 728 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 728 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 728 "fastmix.cpp"
 $9 = ((($pChannel)) + 36|0); //@line 728 "fastmix.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 728 "fastmix.cpp"
 $11 = $10 & 64; //@line 728 "fastmix.cpp"
 $12 = ($11|0)==(0); //@line 728 "fastmix.cpp"
 $13 = (($6) + ($8)|0); //@line 728 "fastmix.cpp"
 $$ = $12 ? $6 : $13; //@line 728 "fastmix.cpp"
 $14 = ((($pChannel)) + 28|0); //@line 730 "fastmix.cpp"
 $15 = ((($pChannel)) + 24|0); //@line 730 "fastmix.cpp"
 $16 = ((($pChannel)) + 12|0); //@line 731 "fastmix.cpp"
 $nPos$0 = $5;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $17 = $nPos$0 >> 16; //@line 729 "fastmix.cpp"
  $$sum = (($17) + ($8))|0; //@line 729 "fastmix.cpp"
  $18 = (($$) + ($$sum)|0); //@line 729 "fastmix.cpp"
  $19 = HEAP8[$18>>0]|0; //@line 729 "fastmix.cpp"
  $20 = $19 << 24 >> 24; //@line 729 "fastmix.cpp"
  $21 = $20 << 8; //@line 729 "fastmix.cpp"
  $22 = HEAP32[$14>>2]|0; //@line 730 "fastmix.cpp"
  $23 = (($22) + ($nRampLeftVol$0))|0; //@line 730 "fastmix.cpp"
  $24 = HEAP32[$15>>2]|0; //@line 730 "fastmix.cpp"
  $25 = (($24) + ($nRampRightVol$0))|0; //@line 730 "fastmix.cpp"
  $26 = $25 >> 12; //@line 730 "fastmix.cpp"
  $27 = Math_imul($26, $21)|0; //@line 730 "fastmix.cpp"
  $28 = HEAP32[$pvol$0>>2]|0; //@line 730 "fastmix.cpp"
  $29 = (($27) + ($28))|0; //@line 730 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $29; //@line 730 "fastmix.cpp"
  $30 = $23 >> 12; //@line 730 "fastmix.cpp"
  $31 = Math_imul($30, $21)|0; //@line 730 "fastmix.cpp"
  $32 = ((($pvol$0)) + 4|0); //@line 730 "fastmix.cpp"
  $33 = HEAP32[$32>>2]|0; //@line 730 "fastmix.cpp"
  $34 = (($33) + ($31))|0; //@line 730 "fastmix.cpp"
  HEAP32[$32>>2] = $34; //@line 730 "fastmix.cpp"
  $35 = ((($pvol$0)) + 8|0); //@line 730 "fastmix.cpp"
  $36 = HEAP32[$16>>2]|0; //@line 731 "fastmix.cpp"
  $37 = (($36) + ($nPos$0))|0; //@line 731 "fastmix.cpp"
  $38 = ($35>>>0)<($pbufmax>>>0); //@line 731 "fastmix.cpp"
  if ($38) {
   $nPos$0 = $37;$nRampLeftVol$0 = $23;$nRampRightVol$0 = $25;$pvol$0 = $35;
  } else {
   $$lcssa = $23;$$lcssa22 = $25;$$lcssa23 = $26;$$lcssa24 = $30;$$lcssa25 = $37;
   break;
  }
 }
 $39 = $$lcssa25 >> 16; //@line 731 "fastmix.cpp"
 $40 = HEAP32[$7>>2]|0; //@line 731 "fastmix.cpp"
 $41 = (($40) + ($39))|0; //@line 731 "fastmix.cpp"
 HEAP32[$7>>2] = $41; //@line 731 "fastmix.cpp"
 $42 = $$lcssa25 & 65535; //@line 731 "fastmix.cpp"
 HEAP32[$4>>2] = $42; //@line 731 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa22; //@line 731 "fastmix.cpp"
 $43 = ((($pChannel)) + 16|0); //@line 731 "fastmix.cpp"
 HEAP32[$43>>2] = $$lcssa23; //@line 731 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa; //@line 731 "fastmix.cpp"
 $44 = ((($pChannel)) + 20|0); //@line 731 "fastmix.cpp"
 HEAP32[$44>>2] = $$lcssa24; //@line 731 "fastmix.cpp"
 return; //@line 731 "fastmix.cpp"
}
function __Z16Mono16BitRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa21 = 0, $$lcssa22 = 0, $$lcssa23 = 0, $$lcssa24 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 733 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 733 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 733 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 733 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 734 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 734 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 734 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 734 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 734 "fastmix.cpp"
 $9 = $8 << 1; //@line 734 "fastmix.cpp"
 $10 = (($6) + ($9)|0); //@line 734 "fastmix.cpp"
 $11 = ((($pChannel)) + 36|0); //@line 734 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 734 "fastmix.cpp"
 $13 = $12 & 64; //@line 734 "fastmix.cpp"
 $14 = ($13|0)==(0); //@line 734 "fastmix.cpp"
 $15 = (($10) + ($8<<1)|0); //@line 734 "fastmix.cpp"
 $$ = $14 ? $10 : $15; //@line 734 "fastmix.cpp"
 $16 = ((($pChannel)) + 28|0); //@line 736 "fastmix.cpp"
 $17 = ((($pChannel)) + 24|0); //@line 736 "fastmix.cpp"
 $18 = ((($pChannel)) + 12|0); //@line 737 "fastmix.cpp"
 $nPos$0 = $5;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $19 = $nPos$0 >> 16; //@line 735 "fastmix.cpp"
  $20 = (($$) + ($19<<1)|0); //@line 735 "fastmix.cpp"
  $21 = HEAP16[$20>>1]|0; //@line 735 "fastmix.cpp"
  $22 = $21 << 16 >> 16; //@line 735 "fastmix.cpp"
  $23 = HEAP32[$16>>2]|0; //@line 736 "fastmix.cpp"
  $24 = (($23) + ($nRampLeftVol$0))|0; //@line 736 "fastmix.cpp"
  $25 = HEAP32[$17>>2]|0; //@line 736 "fastmix.cpp"
  $26 = (($25) + ($nRampRightVol$0))|0; //@line 736 "fastmix.cpp"
  $27 = $26 >> 12; //@line 736 "fastmix.cpp"
  $28 = Math_imul($27, $22)|0; //@line 736 "fastmix.cpp"
  $29 = HEAP32[$pvol$0>>2]|0; //@line 736 "fastmix.cpp"
  $30 = (($28) + ($29))|0; //@line 736 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $30; //@line 736 "fastmix.cpp"
  $31 = $24 >> 12; //@line 736 "fastmix.cpp"
  $32 = Math_imul($31, $22)|0; //@line 736 "fastmix.cpp"
  $33 = ((($pvol$0)) + 4|0); //@line 736 "fastmix.cpp"
  $34 = HEAP32[$33>>2]|0; //@line 736 "fastmix.cpp"
  $35 = (($34) + ($32))|0; //@line 736 "fastmix.cpp"
  HEAP32[$33>>2] = $35; //@line 736 "fastmix.cpp"
  $36 = ((($pvol$0)) + 8|0); //@line 736 "fastmix.cpp"
  $37 = HEAP32[$18>>2]|0; //@line 737 "fastmix.cpp"
  $38 = (($37) + ($nPos$0))|0; //@line 737 "fastmix.cpp"
  $39 = ($36>>>0)<($pbufmax>>>0); //@line 737 "fastmix.cpp"
  if ($39) {
   $nPos$0 = $38;$nRampLeftVol$0 = $24;$nRampRightVol$0 = $26;$pvol$0 = $36;
  } else {
   $$lcssa = $24;$$lcssa21 = $26;$$lcssa22 = $27;$$lcssa23 = $31;$$lcssa24 = $38;
   break;
  }
 }
 $40 = $$lcssa24 >> 16; //@line 737 "fastmix.cpp"
 $41 = HEAP32[$7>>2]|0; //@line 737 "fastmix.cpp"
 $42 = (($41) + ($40))|0; //@line 737 "fastmix.cpp"
 HEAP32[$7>>2] = $42; //@line 737 "fastmix.cpp"
 $43 = $$lcssa24 & 65535; //@line 737 "fastmix.cpp"
 HEAP32[$4>>2] = $43; //@line 737 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa21; //@line 737 "fastmix.cpp"
 $44 = ((($pChannel)) + 16|0); //@line 737 "fastmix.cpp"
 HEAP32[$44>>2] = $$lcssa22; //@line 737 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa; //@line 737 "fastmix.cpp"
 $45 = ((($pChannel)) + 20|0); //@line 737 "fastmix.cpp"
 HEAP32[$45>>2] = $$lcssa23; //@line 737 "fastmix.cpp"
 return; //@line 737 "fastmix.cpp"
}
function __Z17Stereo8BitRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa24 = 0, $$lcssa25 = 0, $$lcssa26 = 0, $$lcssa27 = 0, $$sum = 0, $$sum1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0;
 var $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 931 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 931 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 931 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 931 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 932 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 932 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 932 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 932 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 932 "fastmix.cpp"
 $9 = (($6) + ($8)|0); //@line 932 "fastmix.cpp"
 $10 = ((($pChannel)) + 36|0); //@line 932 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 932 "fastmix.cpp"
 $12 = $11 & 64; //@line 932 "fastmix.cpp"
 $13 = ($12|0)==(0); //@line 932 "fastmix.cpp"
 $$ = $13 ? $6 : $9; //@line 932 "fastmix.cpp"
 $14 = ((($pChannel)) + 28|0); //@line 934 "fastmix.cpp"
 $15 = ((($pChannel)) + 24|0); //@line 934 "fastmix.cpp"
 $16 = ((($pChannel)) + 12|0); //@line 935 "fastmix.cpp"
 $nPos$0 = $5;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $17 = $nPos$0 >> 16; //@line 933 "fastmix.cpp"
  $18 = $17 << 1; //@line 933 "fastmix.cpp"
  $$sum = (($18) + ($8))|0; //@line 933 "fastmix.cpp"
  $19 = (($$) + ($$sum)|0); //@line 933 "fastmix.cpp"
  $20 = HEAP8[$19>>0]|0; //@line 933 "fastmix.cpp"
  $21 = $20 << 24 >> 24; //@line 933 "fastmix.cpp"
  $22 = $21 << 8; //@line 933 "fastmix.cpp"
  $23 = $18 | 1; //@line 933 "fastmix.cpp"
  $$sum1 = (($23) + ($8))|0; //@line 933 "fastmix.cpp"
  $24 = (($$) + ($$sum1)|0); //@line 933 "fastmix.cpp"
  $25 = HEAP8[$24>>0]|0; //@line 933 "fastmix.cpp"
  $26 = $25 << 24 >> 24; //@line 933 "fastmix.cpp"
  $27 = $26 << 8; //@line 933 "fastmix.cpp"
  $28 = HEAP32[$14>>2]|0; //@line 934 "fastmix.cpp"
  $29 = (($28) + ($nRampLeftVol$0))|0; //@line 934 "fastmix.cpp"
  $30 = HEAP32[$15>>2]|0; //@line 934 "fastmix.cpp"
  $31 = (($30) + ($nRampRightVol$0))|0; //@line 934 "fastmix.cpp"
  $32 = $31 >> 12; //@line 934 "fastmix.cpp"
  $33 = Math_imul($22, $32)|0; //@line 934 "fastmix.cpp"
  $34 = HEAP32[$pvol$0>>2]|0; //@line 934 "fastmix.cpp"
  $35 = (($33) + ($34))|0; //@line 934 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $35; //@line 934 "fastmix.cpp"
  $36 = $29 >> 12; //@line 934 "fastmix.cpp"
  $37 = Math_imul($27, $36)|0; //@line 934 "fastmix.cpp"
  $38 = ((($pvol$0)) + 4|0); //@line 934 "fastmix.cpp"
  $39 = HEAP32[$38>>2]|0; //@line 934 "fastmix.cpp"
  $40 = (($39) + ($37))|0; //@line 934 "fastmix.cpp"
  HEAP32[$38>>2] = $40; //@line 934 "fastmix.cpp"
  $41 = ((($pvol$0)) + 8|0); //@line 934 "fastmix.cpp"
  $42 = HEAP32[$16>>2]|0; //@line 935 "fastmix.cpp"
  $43 = (($42) + ($nPos$0))|0; //@line 935 "fastmix.cpp"
  $44 = ($41>>>0)<($pbufmax>>>0); //@line 935 "fastmix.cpp"
  if ($44) {
   $nPos$0 = $43;$nRampLeftVol$0 = $29;$nRampRightVol$0 = $31;$pvol$0 = $41;
  } else {
   $$lcssa = $29;$$lcssa24 = $31;$$lcssa25 = $32;$$lcssa26 = $36;$$lcssa27 = $43;
   break;
  }
 }
 $45 = $$lcssa27 >> 16; //@line 935 "fastmix.cpp"
 $46 = HEAP32[$7>>2]|0; //@line 935 "fastmix.cpp"
 $47 = (($46) + ($45))|0; //@line 935 "fastmix.cpp"
 HEAP32[$7>>2] = $47; //@line 935 "fastmix.cpp"
 $48 = $$lcssa27 & 65535; //@line 935 "fastmix.cpp"
 HEAP32[$4>>2] = $48; //@line 935 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa24; //@line 935 "fastmix.cpp"
 $49 = ((($pChannel)) + 16|0); //@line 935 "fastmix.cpp"
 HEAP32[$49>>2] = $$lcssa25; //@line 935 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa; //@line 935 "fastmix.cpp"
 $50 = ((($pChannel)) + 20|0); //@line 935 "fastmix.cpp"
 HEAP32[$50>>2] = $$lcssa26; //@line 935 "fastmix.cpp"
 return; //@line 935 "fastmix.cpp"
}
function __Z18Stereo16BitRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa21 = 0, $$lcssa22 = 0, $$lcssa23 = 0, $$lcssa24 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0;
 var $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 937 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 937 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 937 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 937 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 938 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 938 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 938 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 938 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 938 "fastmix.cpp"
 $9 = $8 << 1; //@line 938 "fastmix.cpp"
 $10 = (($6) + ($9)|0); //@line 938 "fastmix.cpp"
 $11 = ((($pChannel)) + 36|0); //@line 938 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 938 "fastmix.cpp"
 $13 = $12 & 64; //@line 938 "fastmix.cpp"
 $14 = ($13|0)==(0); //@line 938 "fastmix.cpp"
 $15 = (($10) + ($8<<1)|0); //@line 938 "fastmix.cpp"
 $$ = $14 ? $10 : $15; //@line 938 "fastmix.cpp"
 $16 = ((($pChannel)) + 28|0); //@line 940 "fastmix.cpp"
 $17 = ((($pChannel)) + 24|0); //@line 940 "fastmix.cpp"
 $18 = ((($pChannel)) + 12|0); //@line 941 "fastmix.cpp"
 $nPos$0 = $5;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $19 = $nPos$0 >> 16; //@line 939 "fastmix.cpp"
  $20 = $19 << 1; //@line 939 "fastmix.cpp"
  $21 = (($$) + ($20<<1)|0); //@line 939 "fastmix.cpp"
  $22 = HEAP16[$21>>1]|0; //@line 939 "fastmix.cpp"
  $23 = $22 << 16 >> 16; //@line 939 "fastmix.cpp"
  $24 = $20 | 1; //@line 939 "fastmix.cpp"
  $25 = (($$) + ($24<<1)|0); //@line 939 "fastmix.cpp"
  $26 = HEAP16[$25>>1]|0; //@line 939 "fastmix.cpp"
  $27 = $26 << 16 >> 16; //@line 939 "fastmix.cpp"
  $28 = HEAP32[$16>>2]|0; //@line 940 "fastmix.cpp"
  $29 = (($28) + ($nRampLeftVol$0))|0; //@line 940 "fastmix.cpp"
  $30 = HEAP32[$17>>2]|0; //@line 940 "fastmix.cpp"
  $31 = (($30) + ($nRampRightVol$0))|0; //@line 940 "fastmix.cpp"
  $32 = $31 >> 12; //@line 940 "fastmix.cpp"
  $33 = Math_imul($32, $23)|0; //@line 940 "fastmix.cpp"
  $34 = HEAP32[$pvol$0>>2]|0; //@line 940 "fastmix.cpp"
  $35 = (($33) + ($34))|0; //@line 940 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $35; //@line 940 "fastmix.cpp"
  $36 = $29 >> 12; //@line 940 "fastmix.cpp"
  $37 = Math_imul($36, $27)|0; //@line 940 "fastmix.cpp"
  $38 = ((($pvol$0)) + 4|0); //@line 940 "fastmix.cpp"
  $39 = HEAP32[$38>>2]|0; //@line 940 "fastmix.cpp"
  $40 = (($39) + ($37))|0; //@line 940 "fastmix.cpp"
  HEAP32[$38>>2] = $40; //@line 940 "fastmix.cpp"
  $41 = ((($pvol$0)) + 8|0); //@line 940 "fastmix.cpp"
  $42 = HEAP32[$18>>2]|0; //@line 941 "fastmix.cpp"
  $43 = (($42) + ($nPos$0))|0; //@line 941 "fastmix.cpp"
  $44 = ($41>>>0)<($pbufmax>>>0); //@line 941 "fastmix.cpp"
  if ($44) {
   $nPos$0 = $43;$nRampLeftVol$0 = $29;$nRampRightVol$0 = $31;$pvol$0 = $41;
  } else {
   $$lcssa = $29;$$lcssa21 = $31;$$lcssa22 = $32;$$lcssa23 = $36;$$lcssa24 = $43;
   break;
  }
 }
 $45 = $$lcssa24 >> 16; //@line 941 "fastmix.cpp"
 $46 = HEAP32[$7>>2]|0; //@line 941 "fastmix.cpp"
 $47 = (($46) + ($45))|0; //@line 941 "fastmix.cpp"
 HEAP32[$7>>2] = $47; //@line 941 "fastmix.cpp"
 $48 = $$lcssa24 & 65535; //@line 941 "fastmix.cpp"
 HEAP32[$4>>2] = $48; //@line 941 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa21; //@line 941 "fastmix.cpp"
 $49 = ((($pChannel)) + 16|0); //@line 941 "fastmix.cpp"
 HEAP32[$49>>2] = $$lcssa22; //@line 941 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa; //@line 941 "fastmix.cpp"
 $50 = ((($pChannel)) + 20|0); //@line 941 "fastmix.cpp"
 HEAP32[$50>>2] = $$lcssa23; //@line 941 "fastmix.cpp"
 return; //@line 941 "fastmix.cpp"
}
function __Z17FilterMono8BitMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa7 = 0, $$sum = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0;
 var $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 56|0); //@line 987 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 987 "fastmix.cpp"
 $2 = ((($pChannel)) + 60|0); //@line 987 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 987 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 988 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 988 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 988 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 988 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 988 "fastmix.cpp"
 $9 = ((($pChannel)) + 36|0); //@line 988 "fastmix.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 988 "fastmix.cpp"
 $11 = $10 & 64; //@line 988 "fastmix.cpp"
 $12 = ($11|0)==(0); //@line 988 "fastmix.cpp"
 $13 = (($6) + ($8)|0); //@line 988 "fastmix.cpp"
 $$ = $12 ? $6 : $13; //@line 988 "fastmix.cpp"
 $14 = ((($pChannel)) + 72|0); //@line 990 "fastmix.cpp"
 $15 = ((($pChannel)) + 76|0); //@line 990 "fastmix.cpp"
 $16 = ((($pChannel)) + 80|0); //@line 990 "fastmix.cpp"
 $17 = ((($pChannel)) + 16|0); //@line 991 "fastmix.cpp"
 $18 = ((($pChannel)) + 20|0); //@line 991 "fastmix.cpp"
 $19 = ((($pChannel)) + 12|0); //@line 992 "fastmix.cpp"
 $fy1$0 = $1;$fy2$0 = $3;$nPos$0 = $5;$pvol$0 = $pbuffer;
 while(1) {
  $20 = $nPos$0 >> 16; //@line 989 "fastmix.cpp"
  $$sum = (($20) + ($8))|0; //@line 989 "fastmix.cpp"
  $21 = (($$) + ($$sum)|0); //@line 989 "fastmix.cpp"
  $22 = HEAP8[$21>>0]|0; //@line 989 "fastmix.cpp"
  $23 = $22 << 24 >> 24; //@line 989 "fastmix.cpp"
  $24 = $23 << 8; //@line 989 "fastmix.cpp"
  $25 = HEAP32[$14>>2]|0; //@line 990 "fastmix.cpp"
  $26 = Math_imul($24, $25)|0; //@line 990 "fastmix.cpp"
  $27 = HEAP32[$15>>2]|0; //@line 990 "fastmix.cpp"
  $28 = Math_imul($27, $fy1$0)|0; //@line 990 "fastmix.cpp"
  $29 = HEAP32[$16>>2]|0; //@line 990 "fastmix.cpp"
  $30 = Math_imul($29, $fy2$0)|0; //@line 990 "fastmix.cpp"
  $31 = (($28) + 4096)|0; //@line 990 "fastmix.cpp"
  $32 = (($31) + ($26))|0; //@line 990 "fastmix.cpp"
  $33 = (($32) + ($30))|0; //@line 990 "fastmix.cpp"
  $34 = $33 >> 13; //@line 990 "fastmix.cpp"
  $35 = HEAP32[$17>>2]|0; //@line 991 "fastmix.cpp"
  $36 = Math_imul($34, $35)|0; //@line 991 "fastmix.cpp"
  $37 = HEAP32[$pvol$0>>2]|0; //@line 991 "fastmix.cpp"
  $38 = (($36) + ($37))|0; //@line 991 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $38; //@line 991 "fastmix.cpp"
  $39 = HEAP32[$18>>2]|0; //@line 991 "fastmix.cpp"
  $40 = Math_imul($34, $39)|0; //@line 991 "fastmix.cpp"
  $41 = ((($pvol$0)) + 4|0); //@line 991 "fastmix.cpp"
  $42 = HEAP32[$41>>2]|0; //@line 991 "fastmix.cpp"
  $43 = (($40) + ($42))|0; //@line 991 "fastmix.cpp"
  HEAP32[$41>>2] = $43; //@line 991 "fastmix.cpp"
  $44 = ((($pvol$0)) + 8|0); //@line 991 "fastmix.cpp"
  $45 = HEAP32[$19>>2]|0; //@line 992 "fastmix.cpp"
  $46 = (($45) + ($nPos$0))|0; //@line 992 "fastmix.cpp"
  $47 = ($44>>>0)<($pbufmax>>>0); //@line 992 "fastmix.cpp"
  if ($47) {
   $fy2$0$phi = $fy1$0;$fy1$0 = $34;$nPos$0 = $46;$pvol$0 = $44;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $34;$$lcssa7 = $46;$fy1$0$lcssa = $fy1$0;
   break;
  }
 }
 $48 = $$lcssa7 >> 16; //@line 992 "fastmix.cpp"
 $49 = HEAP32[$7>>2]|0; //@line 992 "fastmix.cpp"
 $50 = (($49) + ($48))|0; //@line 992 "fastmix.cpp"
 HEAP32[$7>>2] = $50; //@line 992 "fastmix.cpp"
 $51 = $$lcssa7 & 65535; //@line 992 "fastmix.cpp"
 HEAP32[$4>>2] = $51; //@line 992 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 992 "fastmix.cpp"
 HEAP32[$2>>2] = $fy1$0$lcssa; //@line 992 "fastmix.cpp"
 return; //@line 992 "fastmix.cpp"
}
function __Z18FilterMono16BitMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0;
 var $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 56|0); //@line 994 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 994 "fastmix.cpp"
 $2 = ((($pChannel)) + 60|0); //@line 994 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 994 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 995 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 995 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 995 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 995 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 995 "fastmix.cpp"
 $9 = $8 << 1; //@line 995 "fastmix.cpp"
 $10 = (($6) + ($9)|0); //@line 995 "fastmix.cpp"
 $11 = ((($pChannel)) + 36|0); //@line 995 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 995 "fastmix.cpp"
 $13 = $12 & 64; //@line 995 "fastmix.cpp"
 $14 = ($13|0)==(0); //@line 995 "fastmix.cpp"
 $15 = (($10) + ($8<<1)|0); //@line 995 "fastmix.cpp"
 $$ = $14 ? $10 : $15; //@line 995 "fastmix.cpp"
 $16 = ((($pChannel)) + 72|0); //@line 997 "fastmix.cpp"
 $17 = ((($pChannel)) + 76|0); //@line 997 "fastmix.cpp"
 $18 = ((($pChannel)) + 80|0); //@line 997 "fastmix.cpp"
 $19 = ((($pChannel)) + 16|0); //@line 998 "fastmix.cpp"
 $20 = ((($pChannel)) + 20|0); //@line 998 "fastmix.cpp"
 $21 = ((($pChannel)) + 12|0); //@line 999 "fastmix.cpp"
 $fy1$0 = $1;$fy2$0 = $3;$nPos$0 = $5;$pvol$0 = $pbuffer;
 while(1) {
  $22 = $nPos$0 >> 16; //@line 996 "fastmix.cpp"
  $23 = (($$) + ($22<<1)|0); //@line 996 "fastmix.cpp"
  $24 = HEAP16[$23>>1]|0; //@line 996 "fastmix.cpp"
  $25 = $24 << 16 >> 16; //@line 996 "fastmix.cpp"
  $26 = HEAP32[$16>>2]|0; //@line 997 "fastmix.cpp"
  $27 = Math_imul($25, $26)|0; //@line 997 "fastmix.cpp"
  $28 = HEAP32[$17>>2]|0; //@line 997 "fastmix.cpp"
  $29 = Math_imul($28, $fy1$0)|0; //@line 997 "fastmix.cpp"
  $30 = HEAP32[$18>>2]|0; //@line 997 "fastmix.cpp"
  $31 = Math_imul($30, $fy2$0)|0; //@line 997 "fastmix.cpp"
  $32 = (($27) + 4096)|0; //@line 997 "fastmix.cpp"
  $33 = (($32) + ($29))|0; //@line 997 "fastmix.cpp"
  $34 = (($33) + ($31))|0; //@line 997 "fastmix.cpp"
  $35 = $34 >> 13; //@line 997 "fastmix.cpp"
  $36 = HEAP32[$19>>2]|0; //@line 998 "fastmix.cpp"
  $37 = Math_imul($35, $36)|0; //@line 998 "fastmix.cpp"
  $38 = HEAP32[$pvol$0>>2]|0; //@line 998 "fastmix.cpp"
  $39 = (($37) + ($38))|0; //@line 998 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $39; //@line 998 "fastmix.cpp"
  $40 = HEAP32[$20>>2]|0; //@line 998 "fastmix.cpp"
  $41 = Math_imul($35, $40)|0; //@line 998 "fastmix.cpp"
  $42 = ((($pvol$0)) + 4|0); //@line 998 "fastmix.cpp"
  $43 = HEAP32[$42>>2]|0; //@line 998 "fastmix.cpp"
  $44 = (($41) + ($43))|0; //@line 998 "fastmix.cpp"
  HEAP32[$42>>2] = $44; //@line 998 "fastmix.cpp"
  $45 = ((($pvol$0)) + 8|0); //@line 998 "fastmix.cpp"
  $46 = HEAP32[$21>>2]|0; //@line 999 "fastmix.cpp"
  $47 = (($46) + ($nPos$0))|0; //@line 999 "fastmix.cpp"
  $48 = ($45>>>0)<($pbufmax>>>0); //@line 999 "fastmix.cpp"
  if ($48) {
   $fy2$0$phi = $fy1$0;$fy1$0 = $35;$nPos$0 = $47;$pvol$0 = $45;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $35;$$lcssa6 = $47;$fy1$0$lcssa = $fy1$0;
   break;
  }
 }
 $49 = $$lcssa6 >> 16; //@line 999 "fastmix.cpp"
 $50 = HEAP32[$7>>2]|0; //@line 999 "fastmix.cpp"
 $51 = (($50) + ($49))|0; //@line 999 "fastmix.cpp"
 HEAP32[$7>>2] = $51; //@line 999 "fastmix.cpp"
 $52 = $$lcssa6 & 65535; //@line 999 "fastmix.cpp"
 HEAP32[$4>>2] = $52; //@line 999 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 999 "fastmix.cpp"
 HEAP32[$2>>2] = $fy1$0$lcssa; //@line 999 "fastmix.cpp"
 return; //@line 999 "fastmix.cpp"
}
function __Z19FilterStereo8BitMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa14 = 0, $$lcssa15 = 0, $$sum = 0, $$sum1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $fy3$0 = 0;
 var $fy3$0$lcssa = 0, $fy4$0 = 0, $fy4$0$phi = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 56|0); //@line 1102 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1102 "fastmix.cpp"
 $2 = ((($pChannel)) + 60|0); //@line 1102 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1102 "fastmix.cpp"
 $4 = ((($pChannel)) + 64|0); //@line 1102 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1102 "fastmix.cpp"
 $6 = ((($pChannel)) + 68|0); //@line 1102 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1102 "fastmix.cpp"
 $8 = ((($pChannel)) + 8|0); //@line 1103 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1103 "fastmix.cpp"
 $10 = HEAP32[$pChannel>>2]|0; //@line 1103 "fastmix.cpp"
 $11 = ((($pChannel)) + 4|0); //@line 1103 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1103 "fastmix.cpp"
 $13 = (($10) + ($12)|0); //@line 1103 "fastmix.cpp"
 $14 = ((($pChannel)) + 36|0); //@line 1103 "fastmix.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 1103 "fastmix.cpp"
 $16 = $15 & 64; //@line 1103 "fastmix.cpp"
 $17 = ($16|0)==(0); //@line 1103 "fastmix.cpp"
 $$ = $17 ? $10 : $13; //@line 1103 "fastmix.cpp"
 $18 = ((($pChannel)) + 72|0); //@line 1105 "fastmix.cpp"
 $19 = ((($pChannel)) + 76|0); //@line 1105 "fastmix.cpp"
 $20 = ((($pChannel)) + 80|0); //@line 1105 "fastmix.cpp"
 $21 = ((($pChannel)) + 16|0); //@line 1106 "fastmix.cpp"
 $22 = ((($pChannel)) + 20|0); //@line 1106 "fastmix.cpp"
 $23 = ((($pChannel)) + 12|0); //@line 1107 "fastmix.cpp"
 $fy1$0 = $1;$fy2$0 = $3;$fy3$0 = $5;$fy4$0 = $7;$nPos$0 = $9;$pvol$0 = $pbuffer;
 while(1) {
  $24 = $nPos$0 >> 16; //@line 1104 "fastmix.cpp"
  $25 = $24 << 1; //@line 1104 "fastmix.cpp"
  $$sum = (($25) + ($12))|0; //@line 1104 "fastmix.cpp"
  $26 = (($$) + ($$sum)|0); //@line 1104 "fastmix.cpp"
  $27 = HEAP8[$26>>0]|0; //@line 1104 "fastmix.cpp"
  $28 = $27 << 24 >> 24; //@line 1104 "fastmix.cpp"
  $29 = $28 << 8; //@line 1104 "fastmix.cpp"
  $30 = $25 | 1; //@line 1104 "fastmix.cpp"
  $$sum1 = (($30) + ($12))|0; //@line 1104 "fastmix.cpp"
  $31 = (($$) + ($$sum1)|0); //@line 1104 "fastmix.cpp"
  $32 = HEAP8[$31>>0]|0; //@line 1104 "fastmix.cpp"
  $33 = $32 << 24 >> 24; //@line 1104 "fastmix.cpp"
  $34 = $33 << 8; //@line 1104 "fastmix.cpp"
  $35 = HEAP32[$18>>2]|0; //@line 1105 "fastmix.cpp"
  $36 = Math_imul($29, $35)|0; //@line 1105 "fastmix.cpp"
  $37 = HEAP32[$19>>2]|0; //@line 1105 "fastmix.cpp"
  $38 = Math_imul($37, $fy1$0)|0; //@line 1105 "fastmix.cpp"
  $39 = HEAP32[$20>>2]|0; //@line 1105 "fastmix.cpp"
  $40 = Math_imul($39, $fy2$0)|0; //@line 1105 "fastmix.cpp"
  $41 = (($36) + 4096)|0; //@line 1105 "fastmix.cpp"
  $42 = (($41) + ($38))|0; //@line 1105 "fastmix.cpp"
  $43 = (($42) + ($40))|0; //@line 1105 "fastmix.cpp"
  $44 = $43 >> 13; //@line 1105 "fastmix.cpp"
  $45 = Math_imul($34, $35)|0; //@line 1105 "fastmix.cpp"
  $46 = Math_imul($37, $fy3$0)|0; //@line 1105 "fastmix.cpp"
  $47 = Math_imul($39, $fy4$0)|0; //@line 1105 "fastmix.cpp"
  $48 = (($46) + 4096)|0; //@line 1105 "fastmix.cpp"
  $49 = (($48) + ($45))|0; //@line 1105 "fastmix.cpp"
  $50 = (($49) + ($47))|0; //@line 1105 "fastmix.cpp"
  $51 = $50 >> 13; //@line 1105 "fastmix.cpp"
  $52 = HEAP32[$21>>2]|0; //@line 1106 "fastmix.cpp"
  $53 = Math_imul($44, $52)|0; //@line 1106 "fastmix.cpp"
  $54 = HEAP32[$pvol$0>>2]|0; //@line 1106 "fastmix.cpp"
  $55 = (($53) + ($54))|0; //@line 1106 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $55; //@line 1106 "fastmix.cpp"
  $56 = HEAP32[$22>>2]|0; //@line 1106 "fastmix.cpp"
  $57 = Math_imul($51, $56)|0; //@line 1106 "fastmix.cpp"
  $58 = ((($pvol$0)) + 4|0); //@line 1106 "fastmix.cpp"
  $59 = HEAP32[$58>>2]|0; //@line 1106 "fastmix.cpp"
  $60 = (($57) + ($59))|0; //@line 1106 "fastmix.cpp"
  HEAP32[$58>>2] = $60; //@line 1106 "fastmix.cpp"
  $61 = ((($pvol$0)) + 8|0); //@line 1106 "fastmix.cpp"
  $62 = HEAP32[$23>>2]|0; //@line 1107 "fastmix.cpp"
  $63 = (($62) + ($nPos$0))|0; //@line 1107 "fastmix.cpp"
  $64 = ($61>>>0)<($pbufmax>>>0); //@line 1107 "fastmix.cpp"
  if ($64) {
   $fy4$0$phi = $fy3$0;$fy2$0$phi = $fy1$0;$fy1$0 = $44;$fy3$0 = $51;$nPos$0 = $63;$pvol$0 = $61;$fy4$0 = $fy4$0$phi;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $44;$$lcssa14 = $51;$$lcssa15 = $63;$fy1$0$lcssa = $fy1$0;$fy3$0$lcssa = $fy3$0;
   break;
  }
 }
 $65 = $$lcssa15 >> 16; //@line 1107 "fastmix.cpp"
 $66 = HEAP32[$11>>2]|0; //@line 1107 "fastmix.cpp"
 $67 = (($66) + ($65))|0; //@line 1107 "fastmix.cpp"
 HEAP32[$11>>2] = $67; //@line 1107 "fastmix.cpp"
 $68 = $$lcssa15 & 65535; //@line 1107 "fastmix.cpp"
 HEAP32[$8>>2] = $68; //@line 1107 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 1107 "fastmix.cpp"
 HEAP32[$2>>2] = $fy1$0$lcssa; //@line 1107 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa14; //@line 1107 "fastmix.cpp"
 HEAP32[$6>>2] = $fy3$0$lcssa; //@line 1107 "fastmix.cpp"
 return; //@line 1107 "fastmix.cpp"
}
function __Z20FilterStereo16BitMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa11 = 0, $$lcssa12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $fy3$0 = 0, $fy3$0$lcssa = 0, $fy4$0 = 0;
 var $fy4$0$phi = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 56|0); //@line 1109 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1109 "fastmix.cpp"
 $2 = ((($pChannel)) + 60|0); //@line 1109 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1109 "fastmix.cpp"
 $4 = ((($pChannel)) + 64|0); //@line 1109 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1109 "fastmix.cpp"
 $6 = ((($pChannel)) + 68|0); //@line 1109 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1109 "fastmix.cpp"
 $8 = ((($pChannel)) + 8|0); //@line 1110 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1110 "fastmix.cpp"
 $10 = HEAP32[$pChannel>>2]|0; //@line 1110 "fastmix.cpp"
 $11 = ((($pChannel)) + 4|0); //@line 1110 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1110 "fastmix.cpp"
 $13 = $12 << 1; //@line 1110 "fastmix.cpp"
 $14 = (($10) + ($13)|0); //@line 1110 "fastmix.cpp"
 $15 = ((($pChannel)) + 36|0); //@line 1110 "fastmix.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1110 "fastmix.cpp"
 $17 = $16 & 64; //@line 1110 "fastmix.cpp"
 $18 = ($17|0)==(0); //@line 1110 "fastmix.cpp"
 $19 = (($14) + ($12<<1)|0); //@line 1110 "fastmix.cpp"
 $$ = $18 ? $14 : $19; //@line 1110 "fastmix.cpp"
 $20 = ((($pChannel)) + 72|0); //@line 1112 "fastmix.cpp"
 $21 = ((($pChannel)) + 76|0); //@line 1112 "fastmix.cpp"
 $22 = ((($pChannel)) + 80|0); //@line 1112 "fastmix.cpp"
 $23 = ((($pChannel)) + 16|0); //@line 1113 "fastmix.cpp"
 $24 = ((($pChannel)) + 20|0); //@line 1113 "fastmix.cpp"
 $25 = ((($pChannel)) + 12|0); //@line 1114 "fastmix.cpp"
 $fy1$0 = $1;$fy2$0 = $3;$fy3$0 = $5;$fy4$0 = $7;$nPos$0 = $9;$pvol$0 = $pbuffer;
 while(1) {
  $26 = $nPos$0 >> 16; //@line 1111 "fastmix.cpp"
  $27 = $26 << 1; //@line 1111 "fastmix.cpp"
  $28 = (($$) + ($27<<1)|0); //@line 1111 "fastmix.cpp"
  $29 = HEAP16[$28>>1]|0; //@line 1111 "fastmix.cpp"
  $30 = $29 << 16 >> 16; //@line 1111 "fastmix.cpp"
  $31 = $27 | 1; //@line 1111 "fastmix.cpp"
  $32 = (($$) + ($31<<1)|0); //@line 1111 "fastmix.cpp"
  $33 = HEAP16[$32>>1]|0; //@line 1111 "fastmix.cpp"
  $34 = $33 << 16 >> 16; //@line 1111 "fastmix.cpp"
  $35 = HEAP32[$20>>2]|0; //@line 1112 "fastmix.cpp"
  $36 = Math_imul($35, $30)|0; //@line 1112 "fastmix.cpp"
  $37 = HEAP32[$21>>2]|0; //@line 1112 "fastmix.cpp"
  $38 = Math_imul($37, $fy1$0)|0; //@line 1112 "fastmix.cpp"
  $39 = HEAP32[$22>>2]|0; //@line 1112 "fastmix.cpp"
  $40 = Math_imul($39, $fy2$0)|0; //@line 1112 "fastmix.cpp"
  $41 = (($36) + 4096)|0; //@line 1112 "fastmix.cpp"
  $42 = (($41) + ($38))|0; //@line 1112 "fastmix.cpp"
  $43 = (($42) + ($40))|0; //@line 1112 "fastmix.cpp"
  $44 = $43 >> 13; //@line 1112 "fastmix.cpp"
  $45 = Math_imul($34, $35)|0; //@line 1112 "fastmix.cpp"
  $46 = Math_imul($37, $fy3$0)|0; //@line 1112 "fastmix.cpp"
  $47 = Math_imul($39, $fy4$0)|0; //@line 1112 "fastmix.cpp"
  $48 = (($45) + 4096)|0; //@line 1112 "fastmix.cpp"
  $49 = (($48) + ($46))|0; //@line 1112 "fastmix.cpp"
  $50 = (($49) + ($47))|0; //@line 1112 "fastmix.cpp"
  $51 = $50 >> 13; //@line 1112 "fastmix.cpp"
  $52 = HEAP32[$23>>2]|0; //@line 1113 "fastmix.cpp"
  $53 = Math_imul($44, $52)|0; //@line 1113 "fastmix.cpp"
  $54 = HEAP32[$pvol$0>>2]|0; //@line 1113 "fastmix.cpp"
  $55 = (($53) + ($54))|0; //@line 1113 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $55; //@line 1113 "fastmix.cpp"
  $56 = HEAP32[$24>>2]|0; //@line 1113 "fastmix.cpp"
  $57 = Math_imul($51, $56)|0; //@line 1113 "fastmix.cpp"
  $58 = ((($pvol$0)) + 4|0); //@line 1113 "fastmix.cpp"
  $59 = HEAP32[$58>>2]|0; //@line 1113 "fastmix.cpp"
  $60 = (($57) + ($59))|0; //@line 1113 "fastmix.cpp"
  HEAP32[$58>>2] = $60; //@line 1113 "fastmix.cpp"
  $61 = ((($pvol$0)) + 8|0); //@line 1113 "fastmix.cpp"
  $62 = HEAP32[$25>>2]|0; //@line 1114 "fastmix.cpp"
  $63 = (($62) + ($nPos$0))|0; //@line 1114 "fastmix.cpp"
  $64 = ($61>>>0)<($pbufmax>>>0); //@line 1114 "fastmix.cpp"
  if ($64) {
   $fy4$0$phi = $fy3$0;$fy2$0$phi = $fy1$0;$fy1$0 = $44;$fy3$0 = $51;$nPos$0 = $63;$pvol$0 = $61;$fy4$0 = $fy4$0$phi;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $44;$$lcssa11 = $51;$$lcssa12 = $63;$fy1$0$lcssa = $fy1$0;$fy3$0$lcssa = $fy3$0;
   break;
  }
 }
 $65 = $$lcssa12 >> 16; //@line 1114 "fastmix.cpp"
 $66 = HEAP32[$11>>2]|0; //@line 1114 "fastmix.cpp"
 $67 = (($66) + ($65))|0; //@line 1114 "fastmix.cpp"
 HEAP32[$11>>2] = $67; //@line 1114 "fastmix.cpp"
 $68 = $$lcssa12 & 65535; //@line 1114 "fastmix.cpp"
 HEAP32[$8>>2] = $68; //@line 1114 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 1114 "fastmix.cpp"
 HEAP32[$2>>2] = $fy1$0$lcssa; //@line 1114 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa11; //@line 1114 "fastmix.cpp"
 HEAP32[$6>>2] = $fy3$0$lcssa; //@line 1114 "fastmix.cpp"
 return; //@line 1114 "fastmix.cpp"
}
function __Z21FilterMono8BitRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa27 = 0, $$lcssa28 = 0, $$lcssa29 = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$sum = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $7 = 0, $8 = 0, $9 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 1044 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1044 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 1044 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1044 "fastmix.cpp"
 $4 = ((($pChannel)) + 56|0); //@line 1044 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1044 "fastmix.cpp"
 $6 = ((($pChannel)) + 60|0); //@line 1044 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1044 "fastmix.cpp"
 $8 = ((($pChannel)) + 8|0); //@line 1045 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1045 "fastmix.cpp"
 $10 = HEAP32[$pChannel>>2]|0; //@line 1045 "fastmix.cpp"
 $11 = ((($pChannel)) + 4|0); //@line 1045 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1045 "fastmix.cpp"
 $13 = ((($pChannel)) + 36|0); //@line 1045 "fastmix.cpp"
 $14 = HEAP32[$13>>2]|0; //@line 1045 "fastmix.cpp"
 $15 = $14 & 64; //@line 1045 "fastmix.cpp"
 $16 = ($15|0)==(0); //@line 1045 "fastmix.cpp"
 $17 = (($10) + ($12)|0); //@line 1045 "fastmix.cpp"
 $$ = $16 ? $10 : $17; //@line 1045 "fastmix.cpp"
 $18 = ((($pChannel)) + 72|0); //@line 1047 "fastmix.cpp"
 $19 = ((($pChannel)) + 76|0); //@line 1047 "fastmix.cpp"
 $20 = ((($pChannel)) + 80|0); //@line 1047 "fastmix.cpp"
 $21 = ((($pChannel)) + 28|0); //@line 1048 "fastmix.cpp"
 $22 = ((($pChannel)) + 24|0); //@line 1048 "fastmix.cpp"
 $23 = ((($pChannel)) + 12|0); //@line 1049 "fastmix.cpp"
 $fy1$0 = $5;$fy2$0 = $7;$nPos$0 = $9;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $24 = $nPos$0 >> 16; //@line 1046 "fastmix.cpp"
  $$sum = (($24) + ($12))|0; //@line 1046 "fastmix.cpp"
  $25 = (($$) + ($$sum)|0); //@line 1046 "fastmix.cpp"
  $26 = HEAP8[$25>>0]|0; //@line 1046 "fastmix.cpp"
  $27 = $26 << 24 >> 24; //@line 1046 "fastmix.cpp"
  $28 = $27 << 8; //@line 1046 "fastmix.cpp"
  $29 = HEAP32[$18>>2]|0; //@line 1047 "fastmix.cpp"
  $30 = Math_imul($28, $29)|0; //@line 1047 "fastmix.cpp"
  $31 = HEAP32[$19>>2]|0; //@line 1047 "fastmix.cpp"
  $32 = Math_imul($31, $fy1$0)|0; //@line 1047 "fastmix.cpp"
  $33 = HEAP32[$20>>2]|0; //@line 1047 "fastmix.cpp"
  $34 = Math_imul($33, $fy2$0)|0; //@line 1047 "fastmix.cpp"
  $35 = (($32) + 4096)|0; //@line 1047 "fastmix.cpp"
  $36 = (($35) + ($30))|0; //@line 1047 "fastmix.cpp"
  $37 = (($36) + ($34))|0; //@line 1047 "fastmix.cpp"
  $38 = $37 >> 13; //@line 1047 "fastmix.cpp"
  $39 = HEAP32[$21>>2]|0; //@line 1048 "fastmix.cpp"
  $40 = (($39) + ($nRampLeftVol$0))|0; //@line 1048 "fastmix.cpp"
  $41 = HEAP32[$22>>2]|0; //@line 1048 "fastmix.cpp"
  $42 = (($41) + ($nRampRightVol$0))|0; //@line 1048 "fastmix.cpp"
  $43 = $42 >> 12; //@line 1048 "fastmix.cpp"
  $44 = Math_imul($38, $43)|0; //@line 1048 "fastmix.cpp"
  $45 = HEAP32[$pvol$0>>2]|0; //@line 1048 "fastmix.cpp"
  $46 = (($44) + ($45))|0; //@line 1048 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $46; //@line 1048 "fastmix.cpp"
  $47 = $40 >> 12; //@line 1048 "fastmix.cpp"
  $48 = Math_imul($38, $47)|0; //@line 1048 "fastmix.cpp"
  $49 = ((($pvol$0)) + 4|0); //@line 1048 "fastmix.cpp"
  $50 = HEAP32[$49>>2]|0; //@line 1048 "fastmix.cpp"
  $51 = (($48) + ($50))|0; //@line 1048 "fastmix.cpp"
  HEAP32[$49>>2] = $51; //@line 1048 "fastmix.cpp"
  $52 = ((($pvol$0)) + 8|0); //@line 1048 "fastmix.cpp"
  $53 = HEAP32[$23>>2]|0; //@line 1049 "fastmix.cpp"
  $54 = (($53) + ($nPos$0))|0; //@line 1049 "fastmix.cpp"
  $55 = ($52>>>0)<($pbufmax>>>0); //@line 1049 "fastmix.cpp"
  if ($55) {
   $fy2$0$phi = $fy1$0;$fy1$0 = $38;$nPos$0 = $54;$nRampLeftVol$0 = $40;$nRampRightVol$0 = $42;$pvol$0 = $52;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $38;$$lcssa27 = $40;$$lcssa28 = $42;$$lcssa29 = $43;$$lcssa30 = $47;$$lcssa31 = $54;$fy1$0$lcssa = $fy1$0;
   break;
  }
 }
 $56 = $$lcssa31 >> 16; //@line 1049 "fastmix.cpp"
 $57 = HEAP32[$11>>2]|0; //@line 1049 "fastmix.cpp"
 $58 = (($57) + ($56))|0; //@line 1049 "fastmix.cpp"
 HEAP32[$11>>2] = $58; //@line 1049 "fastmix.cpp"
 $59 = $$lcssa31 & 65535; //@line 1049 "fastmix.cpp"
 HEAP32[$8>>2] = $59; //@line 1049 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa; //@line 1049 "fastmix.cpp"
 HEAP32[$6>>2] = $fy1$0$lcssa; //@line 1049 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa28; //@line 1049 "fastmix.cpp"
 $60 = ((($pChannel)) + 16|0); //@line 1049 "fastmix.cpp"
 HEAP32[$60>>2] = $$lcssa29; //@line 1049 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa27; //@line 1049 "fastmix.cpp"
 $61 = ((($pChannel)) + 20|0); //@line 1049 "fastmix.cpp"
 HEAP32[$61>>2] = $$lcssa30; //@line 1049 "fastmix.cpp"
 return; //@line 1049 "fastmix.cpp"
}
function __Z22FilterMono16BitRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa26 = 0, $$lcssa27 = 0, $$lcssa28 = 0, $$lcssa29 = 0, $$lcssa30 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $7 = 0, $8 = 0, $9 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 1051 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1051 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 1051 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1051 "fastmix.cpp"
 $4 = ((($pChannel)) + 56|0); //@line 1051 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1051 "fastmix.cpp"
 $6 = ((($pChannel)) + 60|0); //@line 1051 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1051 "fastmix.cpp"
 $8 = ((($pChannel)) + 8|0); //@line 1052 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1052 "fastmix.cpp"
 $10 = HEAP32[$pChannel>>2]|0; //@line 1052 "fastmix.cpp"
 $11 = ((($pChannel)) + 4|0); //@line 1052 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1052 "fastmix.cpp"
 $13 = $12 << 1; //@line 1052 "fastmix.cpp"
 $14 = (($10) + ($13)|0); //@line 1052 "fastmix.cpp"
 $15 = ((($pChannel)) + 36|0); //@line 1052 "fastmix.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1052 "fastmix.cpp"
 $17 = $16 & 64; //@line 1052 "fastmix.cpp"
 $18 = ($17|0)==(0); //@line 1052 "fastmix.cpp"
 $19 = (($14) + ($12<<1)|0); //@line 1052 "fastmix.cpp"
 $$ = $18 ? $14 : $19; //@line 1052 "fastmix.cpp"
 $20 = ((($pChannel)) + 72|0); //@line 1054 "fastmix.cpp"
 $21 = ((($pChannel)) + 76|0); //@line 1054 "fastmix.cpp"
 $22 = ((($pChannel)) + 80|0); //@line 1054 "fastmix.cpp"
 $23 = ((($pChannel)) + 28|0); //@line 1055 "fastmix.cpp"
 $24 = ((($pChannel)) + 24|0); //@line 1055 "fastmix.cpp"
 $25 = ((($pChannel)) + 12|0); //@line 1056 "fastmix.cpp"
 $fy1$0 = $5;$fy2$0 = $7;$nPos$0 = $9;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $26 = $nPos$0 >> 16; //@line 1053 "fastmix.cpp"
  $27 = (($$) + ($26<<1)|0); //@line 1053 "fastmix.cpp"
  $28 = HEAP16[$27>>1]|0; //@line 1053 "fastmix.cpp"
  $29 = $28 << 16 >> 16; //@line 1053 "fastmix.cpp"
  $30 = HEAP32[$20>>2]|0; //@line 1054 "fastmix.cpp"
  $31 = Math_imul($29, $30)|0; //@line 1054 "fastmix.cpp"
  $32 = HEAP32[$21>>2]|0; //@line 1054 "fastmix.cpp"
  $33 = Math_imul($32, $fy1$0)|0; //@line 1054 "fastmix.cpp"
  $34 = HEAP32[$22>>2]|0; //@line 1054 "fastmix.cpp"
  $35 = Math_imul($34, $fy2$0)|0; //@line 1054 "fastmix.cpp"
  $36 = (($31) + 4096)|0; //@line 1054 "fastmix.cpp"
  $37 = (($36) + ($33))|0; //@line 1054 "fastmix.cpp"
  $38 = (($37) + ($35))|0; //@line 1054 "fastmix.cpp"
  $39 = $38 >> 13; //@line 1054 "fastmix.cpp"
  $40 = HEAP32[$23>>2]|0; //@line 1055 "fastmix.cpp"
  $41 = (($40) + ($nRampLeftVol$0))|0; //@line 1055 "fastmix.cpp"
  $42 = HEAP32[$24>>2]|0; //@line 1055 "fastmix.cpp"
  $43 = (($42) + ($nRampRightVol$0))|0; //@line 1055 "fastmix.cpp"
  $44 = $43 >> 12; //@line 1055 "fastmix.cpp"
  $45 = Math_imul($39, $44)|0; //@line 1055 "fastmix.cpp"
  $46 = HEAP32[$pvol$0>>2]|0; //@line 1055 "fastmix.cpp"
  $47 = (($45) + ($46))|0; //@line 1055 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $47; //@line 1055 "fastmix.cpp"
  $48 = $41 >> 12; //@line 1055 "fastmix.cpp"
  $49 = Math_imul($39, $48)|0; //@line 1055 "fastmix.cpp"
  $50 = ((($pvol$0)) + 4|0); //@line 1055 "fastmix.cpp"
  $51 = HEAP32[$50>>2]|0; //@line 1055 "fastmix.cpp"
  $52 = (($49) + ($51))|0; //@line 1055 "fastmix.cpp"
  HEAP32[$50>>2] = $52; //@line 1055 "fastmix.cpp"
  $53 = ((($pvol$0)) + 8|0); //@line 1055 "fastmix.cpp"
  $54 = HEAP32[$25>>2]|0; //@line 1056 "fastmix.cpp"
  $55 = (($54) + ($nPos$0))|0; //@line 1056 "fastmix.cpp"
  $56 = ($53>>>0)<($pbufmax>>>0); //@line 1056 "fastmix.cpp"
  if ($56) {
   $fy2$0$phi = $fy1$0;$fy1$0 = $39;$nPos$0 = $55;$nRampLeftVol$0 = $41;$nRampRightVol$0 = $43;$pvol$0 = $53;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $39;$$lcssa26 = $41;$$lcssa27 = $43;$$lcssa28 = $44;$$lcssa29 = $48;$$lcssa30 = $55;$fy1$0$lcssa = $fy1$0;
   break;
  }
 }
 $57 = $$lcssa30 >> 16; //@line 1056 "fastmix.cpp"
 $58 = HEAP32[$11>>2]|0; //@line 1056 "fastmix.cpp"
 $59 = (($58) + ($57))|0; //@line 1056 "fastmix.cpp"
 HEAP32[$11>>2] = $59; //@line 1056 "fastmix.cpp"
 $60 = $$lcssa30 & 65535; //@line 1056 "fastmix.cpp"
 HEAP32[$8>>2] = $60; //@line 1056 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa; //@line 1056 "fastmix.cpp"
 HEAP32[$6>>2] = $fy1$0$lcssa; //@line 1056 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa27; //@line 1056 "fastmix.cpp"
 $61 = ((($pChannel)) + 16|0); //@line 1056 "fastmix.cpp"
 HEAP32[$61>>2] = $$lcssa28; //@line 1056 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa26; //@line 1056 "fastmix.cpp"
 $62 = ((($pChannel)) + 20|0); //@line 1056 "fastmix.cpp"
 HEAP32[$62>>2] = $$lcssa29; //@line 1056 "fastmix.cpp"
 return; //@line 1056 "fastmix.cpp"
}
function __Z23FilterStereo8BitRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa34 = 0, $$lcssa35 = 0, $$lcssa36 = 0, $$lcssa37 = 0, $$lcssa38 = 0, $$lcssa39 = 0, $$sum = 0, $$sum1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $fy3$0 = 0, $fy3$0$lcssa = 0, $fy4$0 = 0, $fy4$0$phi = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0;
 var $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 1159 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1159 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 1159 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1159 "fastmix.cpp"
 $4 = ((($pChannel)) + 56|0); //@line 1159 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1159 "fastmix.cpp"
 $6 = ((($pChannel)) + 60|0); //@line 1159 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1159 "fastmix.cpp"
 $8 = ((($pChannel)) + 64|0); //@line 1159 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1159 "fastmix.cpp"
 $10 = ((($pChannel)) + 68|0); //@line 1159 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 1159 "fastmix.cpp"
 $12 = ((($pChannel)) + 8|0); //@line 1160 "fastmix.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 1160 "fastmix.cpp"
 $14 = HEAP32[$pChannel>>2]|0; //@line 1160 "fastmix.cpp"
 $15 = ((($pChannel)) + 4|0); //@line 1160 "fastmix.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1160 "fastmix.cpp"
 $17 = (($14) + ($16)|0); //@line 1160 "fastmix.cpp"
 $18 = ((($pChannel)) + 36|0); //@line 1160 "fastmix.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 1160 "fastmix.cpp"
 $20 = $19 & 64; //@line 1160 "fastmix.cpp"
 $21 = ($20|0)==(0); //@line 1160 "fastmix.cpp"
 $$ = $21 ? $14 : $17; //@line 1160 "fastmix.cpp"
 $22 = ((($pChannel)) + 72|0); //@line 1162 "fastmix.cpp"
 $23 = ((($pChannel)) + 76|0); //@line 1162 "fastmix.cpp"
 $24 = ((($pChannel)) + 80|0); //@line 1162 "fastmix.cpp"
 $25 = ((($pChannel)) + 28|0); //@line 1163 "fastmix.cpp"
 $26 = ((($pChannel)) + 24|0); //@line 1163 "fastmix.cpp"
 $27 = ((($pChannel)) + 12|0); //@line 1164 "fastmix.cpp"
 $fy1$0 = $5;$fy2$0 = $7;$fy3$0 = $9;$fy4$0 = $11;$nPos$0 = $13;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $28 = $nPos$0 >> 16; //@line 1161 "fastmix.cpp"
  $29 = $28 << 1; //@line 1161 "fastmix.cpp"
  $$sum = (($29) + ($16))|0; //@line 1161 "fastmix.cpp"
  $30 = (($$) + ($$sum)|0); //@line 1161 "fastmix.cpp"
  $31 = HEAP8[$30>>0]|0; //@line 1161 "fastmix.cpp"
  $32 = $31 << 24 >> 24; //@line 1161 "fastmix.cpp"
  $33 = $32 << 8; //@line 1161 "fastmix.cpp"
  $34 = $29 | 1; //@line 1161 "fastmix.cpp"
  $$sum1 = (($34) + ($16))|0; //@line 1161 "fastmix.cpp"
  $35 = (($$) + ($$sum1)|0); //@line 1161 "fastmix.cpp"
  $36 = HEAP8[$35>>0]|0; //@line 1161 "fastmix.cpp"
  $37 = $36 << 24 >> 24; //@line 1161 "fastmix.cpp"
  $38 = $37 << 8; //@line 1161 "fastmix.cpp"
  $39 = HEAP32[$22>>2]|0; //@line 1162 "fastmix.cpp"
  $40 = Math_imul($33, $39)|0; //@line 1162 "fastmix.cpp"
  $41 = HEAP32[$23>>2]|0; //@line 1162 "fastmix.cpp"
  $42 = Math_imul($41, $fy1$0)|0; //@line 1162 "fastmix.cpp"
  $43 = HEAP32[$24>>2]|0; //@line 1162 "fastmix.cpp"
  $44 = Math_imul($43, $fy2$0)|0; //@line 1162 "fastmix.cpp"
  $45 = (($40) + 4096)|0; //@line 1162 "fastmix.cpp"
  $46 = (($45) + ($42))|0; //@line 1162 "fastmix.cpp"
  $47 = (($46) + ($44))|0; //@line 1162 "fastmix.cpp"
  $48 = $47 >> 13; //@line 1162 "fastmix.cpp"
  $49 = Math_imul($38, $39)|0; //@line 1162 "fastmix.cpp"
  $50 = Math_imul($41, $fy3$0)|0; //@line 1162 "fastmix.cpp"
  $51 = Math_imul($43, $fy4$0)|0; //@line 1162 "fastmix.cpp"
  $52 = (($50) + 4096)|0; //@line 1162 "fastmix.cpp"
  $53 = (($52) + ($49))|0; //@line 1162 "fastmix.cpp"
  $54 = (($53) + ($51))|0; //@line 1162 "fastmix.cpp"
  $55 = $54 >> 13; //@line 1162 "fastmix.cpp"
  $56 = HEAP32[$25>>2]|0; //@line 1163 "fastmix.cpp"
  $57 = (($56) + ($nRampLeftVol$0))|0; //@line 1163 "fastmix.cpp"
  $58 = HEAP32[$26>>2]|0; //@line 1163 "fastmix.cpp"
  $59 = (($58) + ($nRampRightVol$0))|0; //@line 1163 "fastmix.cpp"
  $60 = $59 >> 12; //@line 1163 "fastmix.cpp"
  $61 = Math_imul($48, $60)|0; //@line 1163 "fastmix.cpp"
  $62 = HEAP32[$pvol$0>>2]|0; //@line 1163 "fastmix.cpp"
  $63 = (($61) + ($62))|0; //@line 1163 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $63; //@line 1163 "fastmix.cpp"
  $64 = $57 >> 12; //@line 1163 "fastmix.cpp"
  $65 = Math_imul($55, $64)|0; //@line 1163 "fastmix.cpp"
  $66 = ((($pvol$0)) + 4|0); //@line 1163 "fastmix.cpp"
  $67 = HEAP32[$66>>2]|0; //@line 1163 "fastmix.cpp"
  $68 = (($65) + ($67))|0; //@line 1163 "fastmix.cpp"
  HEAP32[$66>>2] = $68; //@line 1163 "fastmix.cpp"
  $69 = ((($pvol$0)) + 8|0); //@line 1163 "fastmix.cpp"
  $70 = HEAP32[$27>>2]|0; //@line 1164 "fastmix.cpp"
  $71 = (($70) + ($nPos$0))|0; //@line 1164 "fastmix.cpp"
  $72 = ($69>>>0)<($pbufmax>>>0); //@line 1164 "fastmix.cpp"
  if ($72) {
   $fy4$0$phi = $fy3$0;$fy2$0$phi = $fy1$0;$fy1$0 = $48;$fy3$0 = $55;$nPos$0 = $71;$nRampLeftVol$0 = $57;$nRampRightVol$0 = $59;$pvol$0 = $69;$fy4$0 = $fy4$0$phi;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $48;$$lcssa34 = $55;$$lcssa35 = $57;$$lcssa36 = $59;$$lcssa37 = $60;$$lcssa38 = $64;$$lcssa39 = $71;$fy1$0$lcssa = $fy1$0;$fy3$0$lcssa = $fy3$0;
   break;
  }
 }
 $73 = $$lcssa39 >> 16; //@line 1164 "fastmix.cpp"
 $74 = HEAP32[$15>>2]|0; //@line 1164 "fastmix.cpp"
 $75 = (($74) + ($73))|0; //@line 1164 "fastmix.cpp"
 HEAP32[$15>>2] = $75; //@line 1164 "fastmix.cpp"
 $76 = $$lcssa39 & 65535; //@line 1164 "fastmix.cpp"
 HEAP32[$12>>2] = $76; //@line 1164 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa; //@line 1164 "fastmix.cpp"
 HEAP32[$6>>2] = $fy1$0$lcssa; //@line 1164 "fastmix.cpp"
 HEAP32[$8>>2] = $$lcssa34; //@line 1164 "fastmix.cpp"
 HEAP32[$10>>2] = $fy3$0$lcssa; //@line 1164 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa36; //@line 1164 "fastmix.cpp"
 $77 = ((($pChannel)) + 16|0); //@line 1164 "fastmix.cpp"
 HEAP32[$77>>2] = $$lcssa37; //@line 1164 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa35; //@line 1164 "fastmix.cpp"
 $78 = ((($pChannel)) + 20|0); //@line 1164 "fastmix.cpp"
 HEAP32[$78>>2] = $$lcssa38; //@line 1164 "fastmix.cpp"
 return; //@line 1164 "fastmix.cpp"
}
function __Z24FilterStereo16BitRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$lcssa33 = 0, $$lcssa34 = 0, $$lcssa35 = 0, $$lcssa36 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $fy3$0 = 0, $fy3$0$lcssa = 0, $fy4$0 = 0, $fy4$0$phi = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 1166 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1166 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 1166 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1166 "fastmix.cpp"
 $4 = ((($pChannel)) + 56|0); //@line 1166 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1166 "fastmix.cpp"
 $6 = ((($pChannel)) + 60|0); //@line 1166 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1166 "fastmix.cpp"
 $8 = ((($pChannel)) + 64|0); //@line 1166 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1166 "fastmix.cpp"
 $10 = ((($pChannel)) + 68|0); //@line 1166 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 1166 "fastmix.cpp"
 $12 = ((($pChannel)) + 8|0); //@line 1167 "fastmix.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 1167 "fastmix.cpp"
 $14 = HEAP32[$pChannel>>2]|0; //@line 1167 "fastmix.cpp"
 $15 = ((($pChannel)) + 4|0); //@line 1167 "fastmix.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1167 "fastmix.cpp"
 $17 = $16 << 1; //@line 1167 "fastmix.cpp"
 $18 = (($14) + ($17)|0); //@line 1167 "fastmix.cpp"
 $19 = ((($pChannel)) + 36|0); //@line 1167 "fastmix.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 1167 "fastmix.cpp"
 $21 = $20 & 64; //@line 1167 "fastmix.cpp"
 $22 = ($21|0)==(0); //@line 1167 "fastmix.cpp"
 $23 = (($18) + ($16<<1)|0); //@line 1167 "fastmix.cpp"
 $$ = $22 ? $18 : $23; //@line 1167 "fastmix.cpp"
 $24 = ((($pChannel)) + 72|0); //@line 1169 "fastmix.cpp"
 $25 = ((($pChannel)) + 76|0); //@line 1169 "fastmix.cpp"
 $26 = ((($pChannel)) + 80|0); //@line 1169 "fastmix.cpp"
 $27 = ((($pChannel)) + 28|0); //@line 1170 "fastmix.cpp"
 $28 = ((($pChannel)) + 24|0); //@line 1170 "fastmix.cpp"
 $29 = ((($pChannel)) + 12|0); //@line 1171 "fastmix.cpp"
 $fy1$0 = $5;$fy2$0 = $7;$fy3$0 = $9;$fy4$0 = $11;$nPos$0 = $13;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $30 = $nPos$0 >> 16; //@line 1168 "fastmix.cpp"
  $31 = $30 << 1; //@line 1168 "fastmix.cpp"
  $32 = (($$) + ($31<<1)|0); //@line 1168 "fastmix.cpp"
  $33 = HEAP16[$32>>1]|0; //@line 1168 "fastmix.cpp"
  $34 = $33 << 16 >> 16; //@line 1168 "fastmix.cpp"
  $35 = $31 | 1; //@line 1168 "fastmix.cpp"
  $36 = (($$) + ($35<<1)|0); //@line 1168 "fastmix.cpp"
  $37 = HEAP16[$36>>1]|0; //@line 1168 "fastmix.cpp"
  $38 = $37 << 16 >> 16; //@line 1168 "fastmix.cpp"
  $39 = HEAP32[$24>>2]|0; //@line 1169 "fastmix.cpp"
  $40 = Math_imul($39, $34)|0; //@line 1169 "fastmix.cpp"
  $41 = HEAP32[$25>>2]|0; //@line 1169 "fastmix.cpp"
  $42 = Math_imul($41, $fy1$0)|0; //@line 1169 "fastmix.cpp"
  $43 = HEAP32[$26>>2]|0; //@line 1169 "fastmix.cpp"
  $44 = Math_imul($43, $fy2$0)|0; //@line 1169 "fastmix.cpp"
  $45 = (($40) + 4096)|0; //@line 1169 "fastmix.cpp"
  $46 = (($45) + ($42))|0; //@line 1169 "fastmix.cpp"
  $47 = (($46) + ($44))|0; //@line 1169 "fastmix.cpp"
  $48 = $47 >> 13; //@line 1169 "fastmix.cpp"
  $49 = Math_imul($38, $39)|0; //@line 1169 "fastmix.cpp"
  $50 = Math_imul($41, $fy3$0)|0; //@line 1169 "fastmix.cpp"
  $51 = Math_imul($43, $fy4$0)|0; //@line 1169 "fastmix.cpp"
  $52 = (($49) + 4096)|0; //@line 1169 "fastmix.cpp"
  $53 = (($52) + ($50))|0; //@line 1169 "fastmix.cpp"
  $54 = (($53) + ($51))|0; //@line 1169 "fastmix.cpp"
  $55 = $54 >> 13; //@line 1169 "fastmix.cpp"
  $56 = HEAP32[$27>>2]|0; //@line 1170 "fastmix.cpp"
  $57 = (($56) + ($nRampLeftVol$0))|0; //@line 1170 "fastmix.cpp"
  $58 = HEAP32[$28>>2]|0; //@line 1170 "fastmix.cpp"
  $59 = (($58) + ($nRampRightVol$0))|0; //@line 1170 "fastmix.cpp"
  $60 = $59 >> 12; //@line 1170 "fastmix.cpp"
  $61 = Math_imul($48, $60)|0; //@line 1170 "fastmix.cpp"
  $62 = HEAP32[$pvol$0>>2]|0; //@line 1170 "fastmix.cpp"
  $63 = (($61) + ($62))|0; //@line 1170 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $63; //@line 1170 "fastmix.cpp"
  $64 = $57 >> 12; //@line 1170 "fastmix.cpp"
  $65 = Math_imul($55, $64)|0; //@line 1170 "fastmix.cpp"
  $66 = ((($pvol$0)) + 4|0); //@line 1170 "fastmix.cpp"
  $67 = HEAP32[$66>>2]|0; //@line 1170 "fastmix.cpp"
  $68 = (($65) + ($67))|0; //@line 1170 "fastmix.cpp"
  HEAP32[$66>>2] = $68; //@line 1170 "fastmix.cpp"
  $69 = ((($pvol$0)) + 8|0); //@line 1170 "fastmix.cpp"
  $70 = HEAP32[$29>>2]|0; //@line 1171 "fastmix.cpp"
  $71 = (($70) + ($nPos$0))|0; //@line 1171 "fastmix.cpp"
  $72 = ($69>>>0)<($pbufmax>>>0); //@line 1171 "fastmix.cpp"
  if ($72) {
   $fy4$0$phi = $fy3$0;$fy2$0$phi = $fy1$0;$fy1$0 = $48;$fy3$0 = $55;$nPos$0 = $71;$nRampLeftVol$0 = $57;$nRampRightVol$0 = $59;$pvol$0 = $69;$fy4$0 = $fy4$0$phi;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $48;$$lcssa31 = $55;$$lcssa32 = $57;$$lcssa33 = $59;$$lcssa34 = $60;$$lcssa35 = $64;$$lcssa36 = $71;$fy1$0$lcssa = $fy1$0;$fy3$0$lcssa = $fy3$0;
   break;
  }
 }
 $73 = $$lcssa36 >> 16; //@line 1171 "fastmix.cpp"
 $74 = HEAP32[$15>>2]|0; //@line 1171 "fastmix.cpp"
 $75 = (($74) + ($73))|0; //@line 1171 "fastmix.cpp"
 HEAP32[$15>>2] = $75; //@line 1171 "fastmix.cpp"
 $76 = $$lcssa36 & 65535; //@line 1171 "fastmix.cpp"
 HEAP32[$12>>2] = $76; //@line 1171 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa; //@line 1171 "fastmix.cpp"
 HEAP32[$6>>2] = $fy1$0$lcssa; //@line 1171 "fastmix.cpp"
 HEAP32[$8>>2] = $$lcssa31; //@line 1171 "fastmix.cpp"
 HEAP32[$10>>2] = $fy3$0$lcssa; //@line 1171 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa33; //@line 1171 "fastmix.cpp"
 $77 = ((($pChannel)) + 16|0); //@line 1171 "fastmix.cpp"
 HEAP32[$77>>2] = $$lcssa34; //@line 1171 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa32; //@line 1171 "fastmix.cpp"
 $78 = ((($pChannel)) + 20|0); //@line 1171 "fastmix.cpp"
 HEAP32[$78>>2] = $$lcssa35; //@line 1171 "fastmix.cpp"
 return; //@line 1171 "fastmix.cpp"
}
function __Z17Mono8BitLinearMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$sum = 0, $$sum1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 690 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 690 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 690 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 690 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 690 "fastmix.cpp"
 $5 = (($2) + ($4)|0); //@line 690 "fastmix.cpp"
 $6 = ((($pChannel)) + 36|0); //@line 690 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 690 "fastmix.cpp"
 $8 = $7 & 64; //@line 690 "fastmix.cpp"
 $9 = ($8|0)==(0); //@line 690 "fastmix.cpp"
 $$ = $9 ? $2 : $5; //@line 690 "fastmix.cpp"
 $10 = (($4) + 1)|0; //@line 691 "fastmix.cpp"
 $11 = ((($pChannel)) + 16|0); //@line 692 "fastmix.cpp"
 $12 = ((($pChannel)) + 20|0); //@line 692 "fastmix.cpp"
 $13 = ((($pChannel)) + 12|0); //@line 693 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $14 = $nPos$0 >> 16; //@line 691 "fastmix.cpp"
  $15 = $nPos$0 >>> 8; //@line 691 "fastmix.cpp"
  $16 = $15 & 255; //@line 691 "fastmix.cpp"
  $$sum = (($14) + ($4))|0; //@line 691 "fastmix.cpp"
  $17 = (($$) + ($$sum)|0); //@line 691 "fastmix.cpp"
  $18 = HEAP8[$17>>0]|0; //@line 691 "fastmix.cpp"
  $19 = $18 << 24 >> 24; //@line 691 "fastmix.cpp"
  $$sum1 = (($10) + ($14))|0; //@line 691 "fastmix.cpp"
  $20 = (($$) + ($$sum1)|0); //@line 691 "fastmix.cpp"
  $21 = HEAP8[$20>>0]|0; //@line 691 "fastmix.cpp"
  $22 = $21 << 24 >> 24; //@line 691 "fastmix.cpp"
  $23 = $19 << 8; //@line 691 "fastmix.cpp"
  $24 = (($22) - ($19))|0; //@line 691 "fastmix.cpp"
  $25 = Math_imul($24, $16)|0; //@line 691 "fastmix.cpp"
  $26 = (($25) + ($23))|0; //@line 691 "fastmix.cpp"
  $27 = HEAP32[$11>>2]|0; //@line 692 "fastmix.cpp"
  $28 = Math_imul($26, $27)|0; //@line 692 "fastmix.cpp"
  $29 = HEAP32[$pvol$0>>2]|0; //@line 692 "fastmix.cpp"
  $30 = (($28) + ($29))|0; //@line 692 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $30; //@line 692 "fastmix.cpp"
  $31 = HEAP32[$12>>2]|0; //@line 692 "fastmix.cpp"
  $32 = Math_imul($26, $31)|0; //@line 692 "fastmix.cpp"
  $33 = ((($pvol$0)) + 4|0); //@line 692 "fastmix.cpp"
  $34 = HEAP32[$33>>2]|0; //@line 692 "fastmix.cpp"
  $35 = (($32) + ($34))|0; //@line 692 "fastmix.cpp"
  HEAP32[$33>>2] = $35; //@line 692 "fastmix.cpp"
  $36 = ((($pvol$0)) + 8|0); //@line 692 "fastmix.cpp"
  $37 = HEAP32[$13>>2]|0; //@line 693 "fastmix.cpp"
  $38 = (($37) + ($nPos$0))|0; //@line 693 "fastmix.cpp"
  $39 = ($36>>>0)<($pbufmax>>>0); //@line 693 "fastmix.cpp"
  if ($39) {
   $nPos$0 = $38;$pvol$0 = $36;
  } else {
   $$lcssa = $38;
   break;
  }
 }
 $40 = $$lcssa >> 16; //@line 693 "fastmix.cpp"
 $41 = HEAP32[$3>>2]|0; //@line 693 "fastmix.cpp"
 $42 = (($41) + ($40))|0; //@line 693 "fastmix.cpp"
 HEAP32[$3>>2] = $42; //@line 693 "fastmix.cpp"
 $43 = $$lcssa & 65535; //@line 693 "fastmix.cpp"
 HEAP32[$0>>2] = $43; //@line 693 "fastmix.cpp"
 return; //@line 693 "fastmix.cpp"
}
function __Z18Mono16BitLinearMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 696 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 696 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 696 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 696 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 696 "fastmix.cpp"
 $5 = $4 << 1; //@line 696 "fastmix.cpp"
 $6 = (($2) + ($5)|0); //@line 696 "fastmix.cpp"
 $7 = ((($pChannel)) + 36|0); //@line 696 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 696 "fastmix.cpp"
 $9 = $8 & 64; //@line 696 "fastmix.cpp"
 $10 = ($9|0)==(0); //@line 696 "fastmix.cpp"
 $11 = (($6) + ($4<<1)|0); //@line 696 "fastmix.cpp"
 $$ = $10 ? $6 : $11; //@line 696 "fastmix.cpp"
 $12 = ((($pChannel)) + 16|0); //@line 698 "fastmix.cpp"
 $13 = ((($pChannel)) + 20|0); //@line 698 "fastmix.cpp"
 $14 = ((($pChannel)) + 12|0); //@line 699 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $15 = $nPos$0 >> 16; //@line 697 "fastmix.cpp"
  $16 = $nPos$0 >>> 8; //@line 697 "fastmix.cpp"
  $17 = $16 & 255; //@line 697 "fastmix.cpp"
  $18 = (($$) + ($15<<1)|0); //@line 697 "fastmix.cpp"
  $19 = HEAP16[$18>>1]|0; //@line 697 "fastmix.cpp"
  $20 = $19 << 16 >> 16; //@line 697 "fastmix.cpp"
  $21 = (($15) + 1)|0; //@line 697 "fastmix.cpp"
  $22 = (($$) + ($21<<1)|0); //@line 697 "fastmix.cpp"
  $23 = HEAP16[$22>>1]|0; //@line 697 "fastmix.cpp"
  $24 = $23 << 16 >> 16; //@line 697 "fastmix.cpp"
  $25 = (($24) - ($20))|0; //@line 697 "fastmix.cpp"
  $26 = Math_imul($25, $17)|0; //@line 697 "fastmix.cpp"
  $27 = $26 >> 8; //@line 697 "fastmix.cpp"
  $28 = (($27) + ($20))|0; //@line 697 "fastmix.cpp"
  $29 = HEAP32[$12>>2]|0; //@line 698 "fastmix.cpp"
  $30 = Math_imul($28, $29)|0; //@line 698 "fastmix.cpp"
  $31 = HEAP32[$pvol$0>>2]|0; //@line 698 "fastmix.cpp"
  $32 = (($30) + ($31))|0; //@line 698 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $32; //@line 698 "fastmix.cpp"
  $33 = HEAP32[$13>>2]|0; //@line 698 "fastmix.cpp"
  $34 = Math_imul($28, $33)|0; //@line 698 "fastmix.cpp"
  $35 = ((($pvol$0)) + 4|0); //@line 698 "fastmix.cpp"
  $36 = HEAP32[$35>>2]|0; //@line 698 "fastmix.cpp"
  $37 = (($34) + ($36))|0; //@line 698 "fastmix.cpp"
  HEAP32[$35>>2] = $37; //@line 698 "fastmix.cpp"
  $38 = ((($pvol$0)) + 8|0); //@line 698 "fastmix.cpp"
  $39 = HEAP32[$14>>2]|0; //@line 699 "fastmix.cpp"
  $40 = (($39) + ($nPos$0))|0; //@line 699 "fastmix.cpp"
  $41 = ($38>>>0)<($pbufmax>>>0); //@line 699 "fastmix.cpp"
  if ($41) {
   $nPos$0 = $40;$pvol$0 = $38;
  } else {
   $$lcssa = $40;
   break;
  }
 }
 $42 = $$lcssa >> 16; //@line 699 "fastmix.cpp"
 $43 = HEAP32[$3>>2]|0; //@line 699 "fastmix.cpp"
 $44 = (($43) + ($42))|0; //@line 699 "fastmix.cpp"
 HEAP32[$3>>2] = $44; //@line 699 "fastmix.cpp"
 $45 = $$lcssa & 65535; //@line 699 "fastmix.cpp"
 HEAP32[$0>>2] = $45; //@line 699 "fastmix.cpp"
 return; //@line 699 "fastmix.cpp"
}
function __Z19Stereo8BitLinearMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$sum = 0, $$sum1 = 0, $$sum2 = 0, $$sum3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 894 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 894 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 894 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 894 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 894 "fastmix.cpp"
 $5 = (($2) + ($4)|0); //@line 894 "fastmix.cpp"
 $6 = ((($pChannel)) + 36|0); //@line 894 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 894 "fastmix.cpp"
 $8 = $7 & 64; //@line 894 "fastmix.cpp"
 $9 = ($8|0)==(0); //@line 894 "fastmix.cpp"
 $$ = $9 ? $2 : $5; //@line 894 "fastmix.cpp"
 $10 = (($4) + 2)|0; //@line 895 "fastmix.cpp"
 $11 = (($4) + 3)|0; //@line 895 "fastmix.cpp"
 $12 = ((($pChannel)) + 16|0); //@line 896 "fastmix.cpp"
 $13 = ((($pChannel)) + 20|0); //@line 896 "fastmix.cpp"
 $14 = ((($pChannel)) + 12|0); //@line 897 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $15 = $nPos$0 >> 16; //@line 895 "fastmix.cpp"
  $16 = $nPos$0 >>> 8; //@line 895 "fastmix.cpp"
  $17 = $16 & 255; //@line 895 "fastmix.cpp"
  $18 = $15 << 1; //@line 895 "fastmix.cpp"
  $$sum = (($18) + ($4))|0; //@line 895 "fastmix.cpp"
  $19 = (($$) + ($$sum)|0); //@line 895 "fastmix.cpp"
  $20 = HEAP8[$19>>0]|0; //@line 895 "fastmix.cpp"
  $21 = $20 << 24 >> 24; //@line 895 "fastmix.cpp"
  $22 = $21 << 8; //@line 895 "fastmix.cpp"
  $$sum1 = (($10) + ($18))|0; //@line 895 "fastmix.cpp"
  $23 = (($$) + ($$sum1)|0); //@line 895 "fastmix.cpp"
  $24 = HEAP8[$23>>0]|0; //@line 895 "fastmix.cpp"
  $25 = $24 << 24 >> 24; //@line 895 "fastmix.cpp"
  $26 = (($25) - ($21))|0; //@line 895 "fastmix.cpp"
  $27 = Math_imul($26, $17)|0; //@line 895 "fastmix.cpp"
  $28 = (($27) + ($22))|0; //@line 895 "fastmix.cpp"
  $29 = $18 | 1; //@line 895 "fastmix.cpp"
  $$sum2 = (($29) + ($4))|0; //@line 895 "fastmix.cpp"
  $30 = (($$) + ($$sum2)|0); //@line 895 "fastmix.cpp"
  $31 = HEAP8[$30>>0]|0; //@line 895 "fastmix.cpp"
  $32 = $31 << 24 >> 24; //@line 895 "fastmix.cpp"
  $33 = $32 << 8; //@line 895 "fastmix.cpp"
  $$sum3 = (($11) + ($18))|0; //@line 895 "fastmix.cpp"
  $34 = (($$) + ($$sum3)|0); //@line 895 "fastmix.cpp"
  $35 = HEAP8[$34>>0]|0; //@line 895 "fastmix.cpp"
  $36 = $35 << 24 >> 24; //@line 895 "fastmix.cpp"
  $37 = (($36) - ($32))|0; //@line 895 "fastmix.cpp"
  $38 = Math_imul($37, $17)|0; //@line 895 "fastmix.cpp"
  $39 = (($38) + ($33))|0; //@line 895 "fastmix.cpp"
  $40 = HEAP32[$12>>2]|0; //@line 896 "fastmix.cpp"
  $41 = Math_imul($28, $40)|0; //@line 896 "fastmix.cpp"
  $42 = HEAP32[$pvol$0>>2]|0; //@line 896 "fastmix.cpp"
  $43 = (($41) + ($42))|0; //@line 896 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $43; //@line 896 "fastmix.cpp"
  $44 = HEAP32[$13>>2]|0; //@line 896 "fastmix.cpp"
  $45 = Math_imul($39, $44)|0; //@line 896 "fastmix.cpp"
  $46 = ((($pvol$0)) + 4|0); //@line 896 "fastmix.cpp"
  $47 = HEAP32[$46>>2]|0; //@line 896 "fastmix.cpp"
  $48 = (($45) + ($47))|0; //@line 896 "fastmix.cpp"
  HEAP32[$46>>2] = $48; //@line 896 "fastmix.cpp"
  $49 = ((($pvol$0)) + 8|0); //@line 896 "fastmix.cpp"
  $50 = HEAP32[$14>>2]|0; //@line 897 "fastmix.cpp"
  $51 = (($50) + ($nPos$0))|0; //@line 897 "fastmix.cpp"
  $52 = ($49>>>0)<($pbufmax>>>0); //@line 897 "fastmix.cpp"
  if ($52) {
   $nPos$0 = $51;$pvol$0 = $49;
  } else {
   $$lcssa = $51;
   break;
  }
 }
 $53 = $$lcssa >> 16; //@line 897 "fastmix.cpp"
 $54 = HEAP32[$3>>2]|0; //@line 897 "fastmix.cpp"
 $55 = (($54) + ($53))|0; //@line 897 "fastmix.cpp"
 HEAP32[$3>>2] = $55; //@line 897 "fastmix.cpp"
 $56 = $$lcssa & 65535; //@line 897 "fastmix.cpp"
 HEAP32[$0>>2] = $56; //@line 897 "fastmix.cpp"
 return; //@line 897 "fastmix.cpp"
}
function __Z20Stereo16BitLinearMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 900 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 900 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 900 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 900 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 900 "fastmix.cpp"
 $5 = $4 << 1; //@line 900 "fastmix.cpp"
 $6 = (($2) + ($5)|0); //@line 900 "fastmix.cpp"
 $7 = ((($pChannel)) + 36|0); //@line 900 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 900 "fastmix.cpp"
 $9 = $8 & 64; //@line 900 "fastmix.cpp"
 $10 = ($9|0)==(0); //@line 900 "fastmix.cpp"
 $11 = (($6) + ($4<<1)|0); //@line 900 "fastmix.cpp"
 $$ = $10 ? $6 : $11; //@line 900 "fastmix.cpp"
 $12 = ((($pChannel)) + 16|0); //@line 902 "fastmix.cpp"
 $13 = ((($pChannel)) + 20|0); //@line 902 "fastmix.cpp"
 $14 = ((($pChannel)) + 12|0); //@line 903 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $15 = $nPos$0 >> 16; //@line 901 "fastmix.cpp"
  $16 = $nPos$0 >>> 8; //@line 901 "fastmix.cpp"
  $17 = $16 & 255; //@line 901 "fastmix.cpp"
  $18 = $15 << 1; //@line 901 "fastmix.cpp"
  $19 = (($$) + ($18<<1)|0); //@line 901 "fastmix.cpp"
  $20 = HEAP16[$19>>1]|0; //@line 901 "fastmix.cpp"
  $21 = $20 << 16 >> 16; //@line 901 "fastmix.cpp"
  $22 = (($18) + 2)|0; //@line 901 "fastmix.cpp"
  $23 = (($$) + ($22<<1)|0); //@line 901 "fastmix.cpp"
  $24 = HEAP16[$23>>1]|0; //@line 901 "fastmix.cpp"
  $25 = $24 << 16 >> 16; //@line 901 "fastmix.cpp"
  $26 = (($25) - ($21))|0; //@line 901 "fastmix.cpp"
  $27 = Math_imul($26, $17)|0; //@line 901 "fastmix.cpp"
  $28 = $27 >> 8; //@line 901 "fastmix.cpp"
  $29 = (($28) + ($21))|0; //@line 901 "fastmix.cpp"
  $30 = $18 | 1; //@line 901 "fastmix.cpp"
  $31 = (($$) + ($30<<1)|0); //@line 901 "fastmix.cpp"
  $32 = HEAP16[$31>>1]|0; //@line 901 "fastmix.cpp"
  $33 = $32 << 16 >> 16; //@line 901 "fastmix.cpp"
  $34 = (($18) + 3)|0; //@line 901 "fastmix.cpp"
  $35 = (($$) + ($34<<1)|0); //@line 901 "fastmix.cpp"
  $36 = HEAP16[$35>>1]|0; //@line 901 "fastmix.cpp"
  $37 = $36 << 16 >> 16; //@line 901 "fastmix.cpp"
  $38 = (($37) - ($33))|0; //@line 901 "fastmix.cpp"
  $39 = Math_imul($38, $17)|0; //@line 901 "fastmix.cpp"
  $40 = $39 >> 8; //@line 901 "fastmix.cpp"
  $41 = (($40) + ($33))|0; //@line 901 "fastmix.cpp"
  $42 = HEAP32[$12>>2]|0; //@line 902 "fastmix.cpp"
  $43 = Math_imul($29, $42)|0; //@line 902 "fastmix.cpp"
  $44 = HEAP32[$pvol$0>>2]|0; //@line 902 "fastmix.cpp"
  $45 = (($43) + ($44))|0; //@line 902 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $45; //@line 902 "fastmix.cpp"
  $46 = HEAP32[$13>>2]|0; //@line 902 "fastmix.cpp"
  $47 = Math_imul($41, $46)|0; //@line 902 "fastmix.cpp"
  $48 = ((($pvol$0)) + 4|0); //@line 902 "fastmix.cpp"
  $49 = HEAP32[$48>>2]|0; //@line 902 "fastmix.cpp"
  $50 = (($47) + ($49))|0; //@line 902 "fastmix.cpp"
  HEAP32[$48>>2] = $50; //@line 902 "fastmix.cpp"
  $51 = ((($pvol$0)) + 8|0); //@line 902 "fastmix.cpp"
  $52 = HEAP32[$14>>2]|0; //@line 903 "fastmix.cpp"
  $53 = (($52) + ($nPos$0))|0; //@line 903 "fastmix.cpp"
  $54 = ($51>>>0)<($pbufmax>>>0); //@line 903 "fastmix.cpp"
  if ($54) {
   $nPos$0 = $53;$pvol$0 = $51;
  } else {
   $$lcssa = $53;
   break;
  }
 }
 $55 = $$lcssa >> 16; //@line 903 "fastmix.cpp"
 $56 = HEAP32[$3>>2]|0; //@line 903 "fastmix.cpp"
 $57 = (($56) + ($55))|0; //@line 903 "fastmix.cpp"
 HEAP32[$3>>2] = $57; //@line 903 "fastmix.cpp"
 $58 = $$lcssa & 65535; //@line 903 "fastmix.cpp"
 HEAP32[$0>>2] = $58; //@line 903 "fastmix.cpp"
 return; //@line 903 "fastmix.cpp"
}
function __Z21Mono8BitLinearRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa24 = 0, $$lcssa25 = 0, $$lcssa26 = 0, $$lcssa27 = 0, $$sum = 0, $$sum1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 739 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 739 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 739 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 739 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 740 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 740 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 740 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 740 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 740 "fastmix.cpp"
 $9 = (($6) + ($8)|0); //@line 740 "fastmix.cpp"
 $10 = ((($pChannel)) + 36|0); //@line 740 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 740 "fastmix.cpp"
 $12 = $11 & 64; //@line 740 "fastmix.cpp"
 $13 = ($12|0)==(0); //@line 740 "fastmix.cpp"
 $$ = $13 ? $6 : $9; //@line 740 "fastmix.cpp"
 $14 = (($8) + 1)|0; //@line 741 "fastmix.cpp"
 $15 = ((($pChannel)) + 28|0); //@line 742 "fastmix.cpp"
 $16 = ((($pChannel)) + 24|0); //@line 742 "fastmix.cpp"
 $17 = ((($pChannel)) + 12|0); //@line 743 "fastmix.cpp"
 $nPos$0 = $5;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $18 = $nPos$0 >> 16; //@line 741 "fastmix.cpp"
  $19 = $nPos$0 >>> 8; //@line 741 "fastmix.cpp"
  $20 = $19 & 255; //@line 741 "fastmix.cpp"
  $$sum = (($18) + ($8))|0; //@line 741 "fastmix.cpp"
  $21 = (($$) + ($$sum)|0); //@line 741 "fastmix.cpp"
  $22 = HEAP8[$21>>0]|0; //@line 741 "fastmix.cpp"
  $23 = $22 << 24 >> 24; //@line 741 "fastmix.cpp"
  $$sum1 = (($14) + ($18))|0; //@line 741 "fastmix.cpp"
  $24 = (($$) + ($$sum1)|0); //@line 741 "fastmix.cpp"
  $25 = HEAP8[$24>>0]|0; //@line 741 "fastmix.cpp"
  $26 = $25 << 24 >> 24; //@line 741 "fastmix.cpp"
  $27 = $23 << 8; //@line 741 "fastmix.cpp"
  $28 = (($26) - ($23))|0; //@line 741 "fastmix.cpp"
  $29 = Math_imul($28, $20)|0; //@line 741 "fastmix.cpp"
  $30 = (($29) + ($27))|0; //@line 741 "fastmix.cpp"
  $31 = HEAP32[$15>>2]|0; //@line 742 "fastmix.cpp"
  $32 = (($31) + ($nRampLeftVol$0))|0; //@line 742 "fastmix.cpp"
  $33 = HEAP32[$16>>2]|0; //@line 742 "fastmix.cpp"
  $34 = (($33) + ($nRampRightVol$0))|0; //@line 742 "fastmix.cpp"
  $35 = $34 >> 12; //@line 742 "fastmix.cpp"
  $36 = Math_imul($30, $35)|0; //@line 742 "fastmix.cpp"
  $37 = HEAP32[$pvol$0>>2]|0; //@line 742 "fastmix.cpp"
  $38 = (($36) + ($37))|0; //@line 742 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $38; //@line 742 "fastmix.cpp"
  $39 = $32 >> 12; //@line 742 "fastmix.cpp"
  $40 = Math_imul($30, $39)|0; //@line 742 "fastmix.cpp"
  $41 = ((($pvol$0)) + 4|0); //@line 742 "fastmix.cpp"
  $42 = HEAP32[$41>>2]|0; //@line 742 "fastmix.cpp"
  $43 = (($40) + ($42))|0; //@line 742 "fastmix.cpp"
  HEAP32[$41>>2] = $43; //@line 742 "fastmix.cpp"
  $44 = ((($pvol$0)) + 8|0); //@line 742 "fastmix.cpp"
  $45 = HEAP32[$17>>2]|0; //@line 743 "fastmix.cpp"
  $46 = (($45) + ($nPos$0))|0; //@line 743 "fastmix.cpp"
  $47 = ($44>>>0)<($pbufmax>>>0); //@line 743 "fastmix.cpp"
  if ($47) {
   $nPos$0 = $46;$nRampLeftVol$0 = $32;$nRampRightVol$0 = $34;$pvol$0 = $44;
  } else {
   $$lcssa = $32;$$lcssa24 = $34;$$lcssa25 = $35;$$lcssa26 = $39;$$lcssa27 = $46;
   break;
  }
 }
 $48 = $$lcssa27 >> 16; //@line 743 "fastmix.cpp"
 $49 = HEAP32[$7>>2]|0; //@line 743 "fastmix.cpp"
 $50 = (($49) + ($48))|0; //@line 743 "fastmix.cpp"
 HEAP32[$7>>2] = $50; //@line 743 "fastmix.cpp"
 $51 = $$lcssa27 & 65535; //@line 743 "fastmix.cpp"
 HEAP32[$4>>2] = $51; //@line 743 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa24; //@line 743 "fastmix.cpp"
 $52 = ((($pChannel)) + 16|0); //@line 743 "fastmix.cpp"
 HEAP32[$52>>2] = $$lcssa25; //@line 743 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa; //@line 743 "fastmix.cpp"
 $53 = ((($pChannel)) + 20|0); //@line 743 "fastmix.cpp"
 HEAP32[$53>>2] = $$lcssa26; //@line 743 "fastmix.cpp"
 return; //@line 743 "fastmix.cpp"
}
function __Z22Mono16BitLinearRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa21 = 0, $$lcssa22 = 0, $$lcssa23 = 0, $$lcssa24 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 745 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 745 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 745 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 745 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 746 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 746 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 746 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 746 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 746 "fastmix.cpp"
 $9 = $8 << 1; //@line 746 "fastmix.cpp"
 $10 = (($6) + ($9)|0); //@line 746 "fastmix.cpp"
 $11 = ((($pChannel)) + 36|0); //@line 746 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 746 "fastmix.cpp"
 $13 = $12 & 64; //@line 746 "fastmix.cpp"
 $14 = ($13|0)==(0); //@line 746 "fastmix.cpp"
 $15 = (($10) + ($8<<1)|0); //@line 746 "fastmix.cpp"
 $$ = $14 ? $10 : $15; //@line 746 "fastmix.cpp"
 $16 = ((($pChannel)) + 28|0); //@line 748 "fastmix.cpp"
 $17 = ((($pChannel)) + 24|0); //@line 748 "fastmix.cpp"
 $18 = ((($pChannel)) + 12|0); //@line 749 "fastmix.cpp"
 $nPos$0 = $5;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $19 = $nPos$0 >> 16; //@line 747 "fastmix.cpp"
  $20 = $nPos$0 >>> 8; //@line 747 "fastmix.cpp"
  $21 = $20 & 255; //@line 747 "fastmix.cpp"
  $22 = (($$) + ($19<<1)|0); //@line 747 "fastmix.cpp"
  $23 = HEAP16[$22>>1]|0; //@line 747 "fastmix.cpp"
  $24 = $23 << 16 >> 16; //@line 747 "fastmix.cpp"
  $25 = (($19) + 1)|0; //@line 747 "fastmix.cpp"
  $26 = (($$) + ($25<<1)|0); //@line 747 "fastmix.cpp"
  $27 = HEAP16[$26>>1]|0; //@line 747 "fastmix.cpp"
  $28 = $27 << 16 >> 16; //@line 747 "fastmix.cpp"
  $29 = (($28) - ($24))|0; //@line 747 "fastmix.cpp"
  $30 = Math_imul($29, $21)|0; //@line 747 "fastmix.cpp"
  $31 = $30 >> 8; //@line 747 "fastmix.cpp"
  $32 = (($31) + ($24))|0; //@line 747 "fastmix.cpp"
  $33 = HEAP32[$16>>2]|0; //@line 748 "fastmix.cpp"
  $34 = (($33) + ($nRampLeftVol$0))|0; //@line 748 "fastmix.cpp"
  $35 = HEAP32[$17>>2]|0; //@line 748 "fastmix.cpp"
  $36 = (($35) + ($nRampRightVol$0))|0; //@line 748 "fastmix.cpp"
  $37 = $36 >> 12; //@line 748 "fastmix.cpp"
  $38 = Math_imul($32, $37)|0; //@line 748 "fastmix.cpp"
  $39 = HEAP32[$pvol$0>>2]|0; //@line 748 "fastmix.cpp"
  $40 = (($38) + ($39))|0; //@line 748 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $40; //@line 748 "fastmix.cpp"
  $41 = $34 >> 12; //@line 748 "fastmix.cpp"
  $42 = Math_imul($32, $41)|0; //@line 748 "fastmix.cpp"
  $43 = ((($pvol$0)) + 4|0); //@line 748 "fastmix.cpp"
  $44 = HEAP32[$43>>2]|0; //@line 748 "fastmix.cpp"
  $45 = (($42) + ($44))|0; //@line 748 "fastmix.cpp"
  HEAP32[$43>>2] = $45; //@line 748 "fastmix.cpp"
  $46 = ((($pvol$0)) + 8|0); //@line 748 "fastmix.cpp"
  $47 = HEAP32[$18>>2]|0; //@line 749 "fastmix.cpp"
  $48 = (($47) + ($nPos$0))|0; //@line 749 "fastmix.cpp"
  $49 = ($46>>>0)<($pbufmax>>>0); //@line 749 "fastmix.cpp"
  if ($49) {
   $nPos$0 = $48;$nRampLeftVol$0 = $34;$nRampRightVol$0 = $36;$pvol$0 = $46;
  } else {
   $$lcssa = $34;$$lcssa21 = $36;$$lcssa22 = $37;$$lcssa23 = $41;$$lcssa24 = $48;
   break;
  }
 }
 $50 = $$lcssa24 >> 16; //@line 749 "fastmix.cpp"
 $51 = HEAP32[$7>>2]|0; //@line 749 "fastmix.cpp"
 $52 = (($51) + ($50))|0; //@line 749 "fastmix.cpp"
 HEAP32[$7>>2] = $52; //@line 749 "fastmix.cpp"
 $53 = $$lcssa24 & 65535; //@line 749 "fastmix.cpp"
 HEAP32[$4>>2] = $53; //@line 749 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa21; //@line 749 "fastmix.cpp"
 $54 = ((($pChannel)) + 16|0); //@line 749 "fastmix.cpp"
 HEAP32[$54>>2] = $$lcssa22; //@line 749 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa; //@line 749 "fastmix.cpp"
 $55 = ((($pChannel)) + 20|0); //@line 749 "fastmix.cpp"
 HEAP32[$55>>2] = $$lcssa23; //@line 749 "fastmix.cpp"
 return; //@line 749 "fastmix.cpp"
}
function __Z23Stereo8BitLinearRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa28 = 0, $$lcssa29 = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$sum = 0, $$sum1 = 0, $$sum2 = 0, $$sum3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0;
 var $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 943 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 943 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 943 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 943 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 944 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 944 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 944 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 944 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 944 "fastmix.cpp"
 $9 = (($6) + ($8)|0); //@line 944 "fastmix.cpp"
 $10 = ((($pChannel)) + 36|0); //@line 944 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 944 "fastmix.cpp"
 $12 = $11 & 64; //@line 944 "fastmix.cpp"
 $13 = ($12|0)==(0); //@line 944 "fastmix.cpp"
 $$ = $13 ? $6 : $9; //@line 944 "fastmix.cpp"
 $14 = (($8) + 2)|0; //@line 945 "fastmix.cpp"
 $15 = (($8) + 3)|0; //@line 945 "fastmix.cpp"
 $16 = ((($pChannel)) + 28|0); //@line 946 "fastmix.cpp"
 $17 = ((($pChannel)) + 24|0); //@line 946 "fastmix.cpp"
 $18 = ((($pChannel)) + 12|0); //@line 947 "fastmix.cpp"
 $nPos$0 = $5;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $19 = $nPos$0 >> 16; //@line 945 "fastmix.cpp"
  $20 = $nPos$0 >>> 8; //@line 945 "fastmix.cpp"
  $21 = $20 & 255; //@line 945 "fastmix.cpp"
  $22 = $19 << 1; //@line 945 "fastmix.cpp"
  $$sum = (($22) + ($8))|0; //@line 945 "fastmix.cpp"
  $23 = (($$) + ($$sum)|0); //@line 945 "fastmix.cpp"
  $24 = HEAP8[$23>>0]|0; //@line 945 "fastmix.cpp"
  $25 = $24 << 24 >> 24; //@line 945 "fastmix.cpp"
  $26 = $25 << 8; //@line 945 "fastmix.cpp"
  $$sum1 = (($14) + ($22))|0; //@line 945 "fastmix.cpp"
  $27 = (($$) + ($$sum1)|0); //@line 945 "fastmix.cpp"
  $28 = HEAP8[$27>>0]|0; //@line 945 "fastmix.cpp"
  $29 = $28 << 24 >> 24; //@line 945 "fastmix.cpp"
  $30 = (($29) - ($25))|0; //@line 945 "fastmix.cpp"
  $31 = Math_imul($30, $21)|0; //@line 945 "fastmix.cpp"
  $32 = (($31) + ($26))|0; //@line 945 "fastmix.cpp"
  $33 = $22 | 1; //@line 945 "fastmix.cpp"
  $$sum2 = (($33) + ($8))|0; //@line 945 "fastmix.cpp"
  $34 = (($$) + ($$sum2)|0); //@line 945 "fastmix.cpp"
  $35 = HEAP8[$34>>0]|0; //@line 945 "fastmix.cpp"
  $36 = $35 << 24 >> 24; //@line 945 "fastmix.cpp"
  $37 = $36 << 8; //@line 945 "fastmix.cpp"
  $$sum3 = (($15) + ($22))|0; //@line 945 "fastmix.cpp"
  $38 = (($$) + ($$sum3)|0); //@line 945 "fastmix.cpp"
  $39 = HEAP8[$38>>0]|0; //@line 945 "fastmix.cpp"
  $40 = $39 << 24 >> 24; //@line 945 "fastmix.cpp"
  $41 = (($40) - ($36))|0; //@line 945 "fastmix.cpp"
  $42 = Math_imul($41, $21)|0; //@line 945 "fastmix.cpp"
  $43 = (($42) + ($37))|0; //@line 945 "fastmix.cpp"
  $44 = HEAP32[$16>>2]|0; //@line 946 "fastmix.cpp"
  $45 = (($44) + ($nRampLeftVol$0))|0; //@line 946 "fastmix.cpp"
  $46 = HEAP32[$17>>2]|0; //@line 946 "fastmix.cpp"
  $47 = (($46) + ($nRampRightVol$0))|0; //@line 946 "fastmix.cpp"
  $48 = $47 >> 12; //@line 946 "fastmix.cpp"
  $49 = Math_imul($48, $32)|0; //@line 946 "fastmix.cpp"
  $50 = HEAP32[$pvol$0>>2]|0; //@line 946 "fastmix.cpp"
  $51 = (($49) + ($50))|0; //@line 946 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $51; //@line 946 "fastmix.cpp"
  $52 = $45 >> 12; //@line 946 "fastmix.cpp"
  $53 = Math_imul($43, $52)|0; //@line 946 "fastmix.cpp"
  $54 = ((($pvol$0)) + 4|0); //@line 946 "fastmix.cpp"
  $55 = HEAP32[$54>>2]|0; //@line 946 "fastmix.cpp"
  $56 = (($53) + ($55))|0; //@line 946 "fastmix.cpp"
  HEAP32[$54>>2] = $56; //@line 946 "fastmix.cpp"
  $57 = ((($pvol$0)) + 8|0); //@line 946 "fastmix.cpp"
  $58 = HEAP32[$18>>2]|0; //@line 947 "fastmix.cpp"
  $59 = (($58) + ($nPos$0))|0; //@line 947 "fastmix.cpp"
  $60 = ($57>>>0)<($pbufmax>>>0); //@line 947 "fastmix.cpp"
  if ($60) {
   $nPos$0 = $59;$nRampLeftVol$0 = $45;$nRampRightVol$0 = $47;$pvol$0 = $57;
  } else {
   $$lcssa = $45;$$lcssa28 = $47;$$lcssa29 = $48;$$lcssa30 = $52;$$lcssa31 = $59;
   break;
  }
 }
 $61 = $$lcssa31 >> 16; //@line 947 "fastmix.cpp"
 $62 = HEAP32[$7>>2]|0; //@line 947 "fastmix.cpp"
 $63 = (($62) + ($61))|0; //@line 947 "fastmix.cpp"
 HEAP32[$7>>2] = $63; //@line 947 "fastmix.cpp"
 $64 = $$lcssa31 & 65535; //@line 947 "fastmix.cpp"
 HEAP32[$4>>2] = $64; //@line 947 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa28; //@line 947 "fastmix.cpp"
 $65 = ((($pChannel)) + 16|0); //@line 947 "fastmix.cpp"
 HEAP32[$65>>2] = $$lcssa29; //@line 947 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa; //@line 947 "fastmix.cpp"
 $66 = ((($pChannel)) + 20|0); //@line 947 "fastmix.cpp"
 HEAP32[$66>>2] = $$lcssa30; //@line 947 "fastmix.cpp"
 return; //@line 947 "fastmix.cpp"
}
function __Z24Stereo16BitLinearRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa21 = 0, $$lcssa22 = 0, $$lcssa23 = 0, $$lcssa24 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 949 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 949 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 949 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 949 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 950 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 950 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 950 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 950 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 950 "fastmix.cpp"
 $9 = $8 << 1; //@line 950 "fastmix.cpp"
 $10 = (($6) + ($9)|0); //@line 950 "fastmix.cpp"
 $11 = ((($pChannel)) + 36|0); //@line 950 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 950 "fastmix.cpp"
 $13 = $12 & 64; //@line 950 "fastmix.cpp"
 $14 = ($13|0)==(0); //@line 950 "fastmix.cpp"
 $15 = (($10) + ($8<<1)|0); //@line 950 "fastmix.cpp"
 $$ = $14 ? $10 : $15; //@line 950 "fastmix.cpp"
 $16 = ((($pChannel)) + 28|0); //@line 952 "fastmix.cpp"
 $17 = ((($pChannel)) + 24|0); //@line 952 "fastmix.cpp"
 $18 = ((($pChannel)) + 12|0); //@line 953 "fastmix.cpp"
 $nPos$0 = $5;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $19 = $nPos$0 >> 16; //@line 951 "fastmix.cpp"
  $20 = $nPos$0 >>> 8; //@line 951 "fastmix.cpp"
  $21 = $20 & 255; //@line 951 "fastmix.cpp"
  $22 = $19 << 1; //@line 951 "fastmix.cpp"
  $23 = (($$) + ($22<<1)|0); //@line 951 "fastmix.cpp"
  $24 = HEAP16[$23>>1]|0; //@line 951 "fastmix.cpp"
  $25 = $24 << 16 >> 16; //@line 951 "fastmix.cpp"
  $26 = (($22) + 2)|0; //@line 951 "fastmix.cpp"
  $27 = (($$) + ($26<<1)|0); //@line 951 "fastmix.cpp"
  $28 = HEAP16[$27>>1]|0; //@line 951 "fastmix.cpp"
  $29 = $28 << 16 >> 16; //@line 951 "fastmix.cpp"
  $30 = (($29) - ($25))|0; //@line 951 "fastmix.cpp"
  $31 = Math_imul($30, $21)|0; //@line 951 "fastmix.cpp"
  $32 = $31 >> 8; //@line 951 "fastmix.cpp"
  $33 = (($32) + ($25))|0; //@line 951 "fastmix.cpp"
  $34 = $22 | 1; //@line 951 "fastmix.cpp"
  $35 = (($$) + ($34<<1)|0); //@line 951 "fastmix.cpp"
  $36 = HEAP16[$35>>1]|0; //@line 951 "fastmix.cpp"
  $37 = $36 << 16 >> 16; //@line 951 "fastmix.cpp"
  $38 = (($22) + 3)|0; //@line 951 "fastmix.cpp"
  $39 = (($$) + ($38<<1)|0); //@line 951 "fastmix.cpp"
  $40 = HEAP16[$39>>1]|0; //@line 951 "fastmix.cpp"
  $41 = $40 << 16 >> 16; //@line 951 "fastmix.cpp"
  $42 = (($41) - ($37))|0; //@line 951 "fastmix.cpp"
  $43 = Math_imul($42, $21)|0; //@line 951 "fastmix.cpp"
  $44 = $43 >> 8; //@line 951 "fastmix.cpp"
  $45 = (($44) + ($37))|0; //@line 951 "fastmix.cpp"
  $46 = HEAP32[$16>>2]|0; //@line 952 "fastmix.cpp"
  $47 = (($46) + ($nRampLeftVol$0))|0; //@line 952 "fastmix.cpp"
  $48 = HEAP32[$17>>2]|0; //@line 952 "fastmix.cpp"
  $49 = (($48) + ($nRampRightVol$0))|0; //@line 952 "fastmix.cpp"
  $50 = $49 >> 12; //@line 952 "fastmix.cpp"
  $51 = Math_imul($50, $33)|0; //@line 952 "fastmix.cpp"
  $52 = HEAP32[$pvol$0>>2]|0; //@line 952 "fastmix.cpp"
  $53 = (($51) + ($52))|0; //@line 952 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $53; //@line 952 "fastmix.cpp"
  $54 = $47 >> 12; //@line 952 "fastmix.cpp"
  $55 = Math_imul($45, $54)|0; //@line 952 "fastmix.cpp"
  $56 = ((($pvol$0)) + 4|0); //@line 952 "fastmix.cpp"
  $57 = HEAP32[$56>>2]|0; //@line 952 "fastmix.cpp"
  $58 = (($55) + ($57))|0; //@line 952 "fastmix.cpp"
  HEAP32[$56>>2] = $58; //@line 952 "fastmix.cpp"
  $59 = ((($pvol$0)) + 8|0); //@line 952 "fastmix.cpp"
  $60 = HEAP32[$18>>2]|0; //@line 953 "fastmix.cpp"
  $61 = (($60) + ($nPos$0))|0; //@line 953 "fastmix.cpp"
  $62 = ($59>>>0)<($pbufmax>>>0); //@line 953 "fastmix.cpp"
  if ($62) {
   $nPos$0 = $61;$nRampLeftVol$0 = $47;$nRampRightVol$0 = $49;$pvol$0 = $59;
  } else {
   $$lcssa = $47;$$lcssa21 = $49;$$lcssa22 = $50;$$lcssa23 = $54;$$lcssa24 = $61;
   break;
  }
 }
 $63 = $$lcssa24 >> 16; //@line 953 "fastmix.cpp"
 $64 = HEAP32[$7>>2]|0; //@line 953 "fastmix.cpp"
 $65 = (($64) + ($63))|0; //@line 953 "fastmix.cpp"
 HEAP32[$7>>2] = $65; //@line 953 "fastmix.cpp"
 $66 = $$lcssa24 & 65535; //@line 953 "fastmix.cpp"
 HEAP32[$4>>2] = $66; //@line 953 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa21; //@line 953 "fastmix.cpp"
 $67 = ((($pChannel)) + 16|0); //@line 953 "fastmix.cpp"
 HEAP32[$67>>2] = $$lcssa22; //@line 953 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa; //@line 953 "fastmix.cpp"
 $68 = ((($pChannel)) + 20|0); //@line 953 "fastmix.cpp"
 HEAP32[$68>>2] = $$lcssa23; //@line 953 "fastmix.cpp"
 return; //@line 953 "fastmix.cpp"
}
function __Z23FilterMono8BitLinearMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa9 = 0, $$sum = 0, $$sum1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0, $9 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 56|0); //@line 1001 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1001 "fastmix.cpp"
 $2 = ((($pChannel)) + 60|0); //@line 1001 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1001 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 1002 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1002 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 1002 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 1002 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 1002 "fastmix.cpp"
 $9 = (($6) + ($8)|0); //@line 1002 "fastmix.cpp"
 $10 = ((($pChannel)) + 36|0); //@line 1002 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 1002 "fastmix.cpp"
 $12 = $11 & 64; //@line 1002 "fastmix.cpp"
 $13 = ($12|0)==(0); //@line 1002 "fastmix.cpp"
 $$ = $13 ? $6 : $9; //@line 1002 "fastmix.cpp"
 $14 = (($8) + 1)|0; //@line 1003 "fastmix.cpp"
 $15 = ((($pChannel)) + 72|0); //@line 1004 "fastmix.cpp"
 $16 = ((($pChannel)) + 76|0); //@line 1004 "fastmix.cpp"
 $17 = ((($pChannel)) + 80|0); //@line 1004 "fastmix.cpp"
 $18 = ((($pChannel)) + 16|0); //@line 1005 "fastmix.cpp"
 $19 = ((($pChannel)) + 20|0); //@line 1005 "fastmix.cpp"
 $20 = ((($pChannel)) + 12|0); //@line 1006 "fastmix.cpp"
 $fy1$0 = $1;$fy2$0 = $3;$nPos$0 = $5;$pvol$0 = $pbuffer;
 while(1) {
  $21 = $nPos$0 >> 16; //@line 1003 "fastmix.cpp"
  $22 = $nPos$0 >>> 8; //@line 1003 "fastmix.cpp"
  $23 = $22 & 255; //@line 1003 "fastmix.cpp"
  $$sum = (($21) + ($8))|0; //@line 1003 "fastmix.cpp"
  $24 = (($$) + ($$sum)|0); //@line 1003 "fastmix.cpp"
  $25 = HEAP8[$24>>0]|0; //@line 1003 "fastmix.cpp"
  $26 = $25 << 24 >> 24; //@line 1003 "fastmix.cpp"
  $$sum1 = (($14) + ($21))|0; //@line 1003 "fastmix.cpp"
  $27 = (($$) + ($$sum1)|0); //@line 1003 "fastmix.cpp"
  $28 = HEAP8[$27>>0]|0; //@line 1003 "fastmix.cpp"
  $29 = $28 << 24 >> 24; //@line 1003 "fastmix.cpp"
  $30 = $26 << 8; //@line 1003 "fastmix.cpp"
  $31 = (($29) - ($26))|0; //@line 1003 "fastmix.cpp"
  $32 = Math_imul($31, $23)|0; //@line 1003 "fastmix.cpp"
  $33 = (($32) + ($30))|0; //@line 1003 "fastmix.cpp"
  $34 = HEAP32[$15>>2]|0; //@line 1004 "fastmix.cpp"
  $35 = Math_imul($33, $34)|0; //@line 1004 "fastmix.cpp"
  $36 = HEAP32[$16>>2]|0; //@line 1004 "fastmix.cpp"
  $37 = Math_imul($36, $fy1$0)|0; //@line 1004 "fastmix.cpp"
  $38 = HEAP32[$17>>2]|0; //@line 1004 "fastmix.cpp"
  $39 = Math_imul($38, $fy2$0)|0; //@line 1004 "fastmix.cpp"
  $40 = (($37) + 4096)|0; //@line 1004 "fastmix.cpp"
  $41 = (($40) + ($39))|0; //@line 1004 "fastmix.cpp"
  $42 = (($41) + ($35))|0; //@line 1004 "fastmix.cpp"
  $43 = $42 >> 13; //@line 1004 "fastmix.cpp"
  $44 = HEAP32[$18>>2]|0; //@line 1005 "fastmix.cpp"
  $45 = Math_imul($43, $44)|0; //@line 1005 "fastmix.cpp"
  $46 = HEAP32[$pvol$0>>2]|0; //@line 1005 "fastmix.cpp"
  $47 = (($45) + ($46))|0; //@line 1005 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $47; //@line 1005 "fastmix.cpp"
  $48 = HEAP32[$19>>2]|0; //@line 1005 "fastmix.cpp"
  $49 = Math_imul($43, $48)|0; //@line 1005 "fastmix.cpp"
  $50 = ((($pvol$0)) + 4|0); //@line 1005 "fastmix.cpp"
  $51 = HEAP32[$50>>2]|0; //@line 1005 "fastmix.cpp"
  $52 = (($49) + ($51))|0; //@line 1005 "fastmix.cpp"
  HEAP32[$50>>2] = $52; //@line 1005 "fastmix.cpp"
  $53 = ((($pvol$0)) + 8|0); //@line 1005 "fastmix.cpp"
  $54 = HEAP32[$20>>2]|0; //@line 1006 "fastmix.cpp"
  $55 = (($54) + ($nPos$0))|0; //@line 1006 "fastmix.cpp"
  $56 = ($53>>>0)<($pbufmax>>>0); //@line 1006 "fastmix.cpp"
  if ($56) {
   $fy2$0$phi = $fy1$0;$fy1$0 = $43;$nPos$0 = $55;$pvol$0 = $53;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $43;$$lcssa9 = $55;$fy1$0$lcssa = $fy1$0;
   break;
  }
 }
 $57 = $$lcssa9 >> 16; //@line 1006 "fastmix.cpp"
 $58 = HEAP32[$7>>2]|0; //@line 1006 "fastmix.cpp"
 $59 = (($58) + ($57))|0; //@line 1006 "fastmix.cpp"
 HEAP32[$7>>2] = $59; //@line 1006 "fastmix.cpp"
 $60 = $$lcssa9 & 65535; //@line 1006 "fastmix.cpp"
 HEAP32[$4>>2] = $60; //@line 1006 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 1006 "fastmix.cpp"
 HEAP32[$2>>2] = $fy1$0$lcssa; //@line 1006 "fastmix.cpp"
 return; //@line 1006 "fastmix.cpp"
}
function __Z24FilterMono16BitLinearMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $7 = 0, $8 = 0, $9 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 56|0); //@line 1008 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1008 "fastmix.cpp"
 $2 = ((($pChannel)) + 60|0); //@line 1008 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1008 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 1009 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1009 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 1009 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 1009 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 1009 "fastmix.cpp"
 $9 = $8 << 1; //@line 1009 "fastmix.cpp"
 $10 = (($6) + ($9)|0); //@line 1009 "fastmix.cpp"
 $11 = ((($pChannel)) + 36|0); //@line 1009 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1009 "fastmix.cpp"
 $13 = $12 & 64; //@line 1009 "fastmix.cpp"
 $14 = ($13|0)==(0); //@line 1009 "fastmix.cpp"
 $15 = (($10) + ($8<<1)|0); //@line 1009 "fastmix.cpp"
 $$ = $14 ? $10 : $15; //@line 1009 "fastmix.cpp"
 $16 = ((($pChannel)) + 72|0); //@line 1011 "fastmix.cpp"
 $17 = ((($pChannel)) + 76|0); //@line 1011 "fastmix.cpp"
 $18 = ((($pChannel)) + 80|0); //@line 1011 "fastmix.cpp"
 $19 = ((($pChannel)) + 16|0); //@line 1012 "fastmix.cpp"
 $20 = ((($pChannel)) + 20|0); //@line 1012 "fastmix.cpp"
 $21 = ((($pChannel)) + 12|0); //@line 1013 "fastmix.cpp"
 $fy1$0 = $1;$fy2$0 = $3;$nPos$0 = $5;$pvol$0 = $pbuffer;
 while(1) {
  $22 = $nPos$0 >> 16; //@line 1010 "fastmix.cpp"
  $23 = $nPos$0 >>> 8; //@line 1010 "fastmix.cpp"
  $24 = $23 & 255; //@line 1010 "fastmix.cpp"
  $25 = (($$) + ($22<<1)|0); //@line 1010 "fastmix.cpp"
  $26 = HEAP16[$25>>1]|0; //@line 1010 "fastmix.cpp"
  $27 = $26 << 16 >> 16; //@line 1010 "fastmix.cpp"
  $28 = (($22) + 1)|0; //@line 1010 "fastmix.cpp"
  $29 = (($$) + ($28<<1)|0); //@line 1010 "fastmix.cpp"
  $30 = HEAP16[$29>>1]|0; //@line 1010 "fastmix.cpp"
  $31 = $30 << 16 >> 16; //@line 1010 "fastmix.cpp"
  $32 = (($31) - ($27))|0; //@line 1010 "fastmix.cpp"
  $33 = Math_imul($32, $24)|0; //@line 1010 "fastmix.cpp"
  $34 = $33 >> 8; //@line 1010 "fastmix.cpp"
  $35 = (($34) + ($27))|0; //@line 1010 "fastmix.cpp"
  $36 = HEAP32[$16>>2]|0; //@line 1011 "fastmix.cpp"
  $37 = Math_imul($35, $36)|0; //@line 1011 "fastmix.cpp"
  $38 = HEAP32[$17>>2]|0; //@line 1011 "fastmix.cpp"
  $39 = Math_imul($38, $fy1$0)|0; //@line 1011 "fastmix.cpp"
  $40 = HEAP32[$18>>2]|0; //@line 1011 "fastmix.cpp"
  $41 = Math_imul($40, $fy2$0)|0; //@line 1011 "fastmix.cpp"
  $42 = (($39) + 4096)|0; //@line 1011 "fastmix.cpp"
  $43 = (($42) + ($41))|0; //@line 1011 "fastmix.cpp"
  $44 = (($43) + ($37))|0; //@line 1011 "fastmix.cpp"
  $45 = $44 >> 13; //@line 1011 "fastmix.cpp"
  $46 = HEAP32[$19>>2]|0; //@line 1012 "fastmix.cpp"
  $47 = Math_imul($45, $46)|0; //@line 1012 "fastmix.cpp"
  $48 = HEAP32[$pvol$0>>2]|0; //@line 1012 "fastmix.cpp"
  $49 = (($47) + ($48))|0; //@line 1012 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $49; //@line 1012 "fastmix.cpp"
  $50 = HEAP32[$20>>2]|0; //@line 1012 "fastmix.cpp"
  $51 = Math_imul($45, $50)|0; //@line 1012 "fastmix.cpp"
  $52 = ((($pvol$0)) + 4|0); //@line 1012 "fastmix.cpp"
  $53 = HEAP32[$52>>2]|0; //@line 1012 "fastmix.cpp"
  $54 = (($51) + ($53))|0; //@line 1012 "fastmix.cpp"
  HEAP32[$52>>2] = $54; //@line 1012 "fastmix.cpp"
  $55 = ((($pvol$0)) + 8|0); //@line 1012 "fastmix.cpp"
  $56 = HEAP32[$21>>2]|0; //@line 1013 "fastmix.cpp"
  $57 = (($56) + ($nPos$0))|0; //@line 1013 "fastmix.cpp"
  $58 = ($55>>>0)<($pbufmax>>>0); //@line 1013 "fastmix.cpp"
  if ($58) {
   $fy2$0$phi = $fy1$0;$fy1$0 = $45;$nPos$0 = $57;$pvol$0 = $55;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $45;$$lcssa6 = $57;$fy1$0$lcssa = $fy1$0;
   break;
  }
 }
 $59 = $$lcssa6 >> 16; //@line 1013 "fastmix.cpp"
 $60 = HEAP32[$7>>2]|0; //@line 1013 "fastmix.cpp"
 $61 = (($60) + ($59))|0; //@line 1013 "fastmix.cpp"
 HEAP32[$7>>2] = $61; //@line 1013 "fastmix.cpp"
 $62 = $$lcssa6 & 65535; //@line 1013 "fastmix.cpp"
 HEAP32[$4>>2] = $62; //@line 1013 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 1013 "fastmix.cpp"
 HEAP32[$2>>2] = $fy1$0$lcssa; //@line 1013 "fastmix.cpp"
 return; //@line 1013 "fastmix.cpp"
}
function __Z25FilterStereo8BitLinearMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa18 = 0, $$lcssa19 = 0, $$sum = 0, $$sum1 = 0, $$sum2 = 0, $$sum3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $9 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $fy3$0 = 0, $fy3$0$lcssa = 0, $fy4$0 = 0;
 var $fy4$0$phi = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 56|0); //@line 1116 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1116 "fastmix.cpp"
 $2 = ((($pChannel)) + 60|0); //@line 1116 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1116 "fastmix.cpp"
 $4 = ((($pChannel)) + 64|0); //@line 1116 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1116 "fastmix.cpp"
 $6 = ((($pChannel)) + 68|0); //@line 1116 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1116 "fastmix.cpp"
 $8 = ((($pChannel)) + 8|0); //@line 1117 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1117 "fastmix.cpp"
 $10 = HEAP32[$pChannel>>2]|0; //@line 1117 "fastmix.cpp"
 $11 = ((($pChannel)) + 4|0); //@line 1117 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1117 "fastmix.cpp"
 $13 = (($10) + ($12)|0); //@line 1117 "fastmix.cpp"
 $14 = ((($pChannel)) + 36|0); //@line 1117 "fastmix.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 1117 "fastmix.cpp"
 $16 = $15 & 64; //@line 1117 "fastmix.cpp"
 $17 = ($16|0)==(0); //@line 1117 "fastmix.cpp"
 $$ = $17 ? $10 : $13; //@line 1117 "fastmix.cpp"
 $18 = (($12) + 2)|0; //@line 1118 "fastmix.cpp"
 $19 = (($12) + 3)|0; //@line 1118 "fastmix.cpp"
 $20 = ((($pChannel)) + 72|0); //@line 1119 "fastmix.cpp"
 $21 = ((($pChannel)) + 76|0); //@line 1119 "fastmix.cpp"
 $22 = ((($pChannel)) + 80|0); //@line 1119 "fastmix.cpp"
 $23 = ((($pChannel)) + 16|0); //@line 1120 "fastmix.cpp"
 $24 = ((($pChannel)) + 20|0); //@line 1120 "fastmix.cpp"
 $25 = ((($pChannel)) + 12|0); //@line 1121 "fastmix.cpp"
 $fy1$0 = $1;$fy2$0 = $3;$fy3$0 = $5;$fy4$0 = $7;$nPos$0 = $9;$pvol$0 = $pbuffer;
 while(1) {
  $26 = $nPos$0 >> 16; //@line 1118 "fastmix.cpp"
  $27 = $nPos$0 >>> 8; //@line 1118 "fastmix.cpp"
  $28 = $27 & 255; //@line 1118 "fastmix.cpp"
  $29 = $26 << 1; //@line 1118 "fastmix.cpp"
  $$sum = (($29) + ($12))|0; //@line 1118 "fastmix.cpp"
  $30 = (($$) + ($$sum)|0); //@line 1118 "fastmix.cpp"
  $31 = HEAP8[$30>>0]|0; //@line 1118 "fastmix.cpp"
  $32 = $31 << 24 >> 24; //@line 1118 "fastmix.cpp"
  $33 = $32 << 8; //@line 1118 "fastmix.cpp"
  $$sum1 = (($18) + ($29))|0; //@line 1118 "fastmix.cpp"
  $34 = (($$) + ($$sum1)|0); //@line 1118 "fastmix.cpp"
  $35 = HEAP8[$34>>0]|0; //@line 1118 "fastmix.cpp"
  $36 = $35 << 24 >> 24; //@line 1118 "fastmix.cpp"
  $37 = (($36) - ($32))|0; //@line 1118 "fastmix.cpp"
  $38 = Math_imul($37, $28)|0; //@line 1118 "fastmix.cpp"
  $39 = (($38) + ($33))|0; //@line 1118 "fastmix.cpp"
  $40 = $29 | 1; //@line 1118 "fastmix.cpp"
  $$sum2 = (($40) + ($12))|0; //@line 1118 "fastmix.cpp"
  $41 = (($$) + ($$sum2)|0); //@line 1118 "fastmix.cpp"
  $42 = HEAP8[$41>>0]|0; //@line 1118 "fastmix.cpp"
  $43 = $42 << 24 >> 24; //@line 1118 "fastmix.cpp"
  $44 = $43 << 8; //@line 1118 "fastmix.cpp"
  $$sum3 = (($19) + ($29))|0; //@line 1118 "fastmix.cpp"
  $45 = (($$) + ($$sum3)|0); //@line 1118 "fastmix.cpp"
  $46 = HEAP8[$45>>0]|0; //@line 1118 "fastmix.cpp"
  $47 = $46 << 24 >> 24; //@line 1118 "fastmix.cpp"
  $48 = (($47) - ($43))|0; //@line 1118 "fastmix.cpp"
  $49 = Math_imul($48, $28)|0; //@line 1118 "fastmix.cpp"
  $50 = (($49) + ($44))|0; //@line 1118 "fastmix.cpp"
  $51 = HEAP32[$20>>2]|0; //@line 1119 "fastmix.cpp"
  $52 = Math_imul($39, $51)|0; //@line 1119 "fastmix.cpp"
  $53 = HEAP32[$21>>2]|0; //@line 1119 "fastmix.cpp"
  $54 = Math_imul($53, $fy1$0)|0; //@line 1119 "fastmix.cpp"
  $55 = HEAP32[$22>>2]|0; //@line 1119 "fastmix.cpp"
  $56 = Math_imul($55, $fy2$0)|0; //@line 1119 "fastmix.cpp"
  $57 = (($54) + 4096)|0; //@line 1119 "fastmix.cpp"
  $58 = (($57) + ($52))|0; //@line 1119 "fastmix.cpp"
  $59 = (($58) + ($56))|0; //@line 1119 "fastmix.cpp"
  $60 = $59 >> 13; //@line 1119 "fastmix.cpp"
  $61 = Math_imul($50, $51)|0; //@line 1119 "fastmix.cpp"
  $62 = Math_imul($53, $fy3$0)|0; //@line 1119 "fastmix.cpp"
  $63 = Math_imul($55, $fy4$0)|0; //@line 1119 "fastmix.cpp"
  $64 = (($62) + 4096)|0; //@line 1119 "fastmix.cpp"
  $65 = (($64) + ($63))|0; //@line 1119 "fastmix.cpp"
  $66 = (($65) + ($61))|0; //@line 1119 "fastmix.cpp"
  $67 = $66 >> 13; //@line 1119 "fastmix.cpp"
  $68 = HEAP32[$23>>2]|0; //@line 1120 "fastmix.cpp"
  $69 = Math_imul($60, $68)|0; //@line 1120 "fastmix.cpp"
  $70 = HEAP32[$pvol$0>>2]|0; //@line 1120 "fastmix.cpp"
  $71 = (($69) + ($70))|0; //@line 1120 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $71; //@line 1120 "fastmix.cpp"
  $72 = HEAP32[$24>>2]|0; //@line 1120 "fastmix.cpp"
  $73 = Math_imul($67, $72)|0; //@line 1120 "fastmix.cpp"
  $74 = ((($pvol$0)) + 4|0); //@line 1120 "fastmix.cpp"
  $75 = HEAP32[$74>>2]|0; //@line 1120 "fastmix.cpp"
  $76 = (($73) + ($75))|0; //@line 1120 "fastmix.cpp"
  HEAP32[$74>>2] = $76; //@line 1120 "fastmix.cpp"
  $77 = ((($pvol$0)) + 8|0); //@line 1120 "fastmix.cpp"
  $78 = HEAP32[$25>>2]|0; //@line 1121 "fastmix.cpp"
  $79 = (($78) + ($nPos$0))|0; //@line 1121 "fastmix.cpp"
  $80 = ($77>>>0)<($pbufmax>>>0); //@line 1121 "fastmix.cpp"
  if ($80) {
   $fy4$0$phi = $fy3$0;$fy2$0$phi = $fy1$0;$fy1$0 = $60;$fy3$0 = $67;$nPos$0 = $79;$pvol$0 = $77;$fy4$0 = $fy4$0$phi;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $60;$$lcssa18 = $67;$$lcssa19 = $79;$fy1$0$lcssa = $fy1$0;$fy3$0$lcssa = $fy3$0;
   break;
  }
 }
 $81 = $$lcssa19 >> 16; //@line 1121 "fastmix.cpp"
 $82 = HEAP32[$11>>2]|0; //@line 1121 "fastmix.cpp"
 $83 = (($82) + ($81))|0; //@line 1121 "fastmix.cpp"
 HEAP32[$11>>2] = $83; //@line 1121 "fastmix.cpp"
 $84 = $$lcssa19 & 65535; //@line 1121 "fastmix.cpp"
 HEAP32[$8>>2] = $84; //@line 1121 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 1121 "fastmix.cpp"
 HEAP32[$2>>2] = $fy1$0$lcssa; //@line 1121 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa18; //@line 1121 "fastmix.cpp"
 HEAP32[$6>>2] = $fy3$0$lcssa; //@line 1121 "fastmix.cpp"
 return; //@line 1121 "fastmix.cpp"
}
function __Z26FilterStereo16BitLinearMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa11 = 0, $$lcssa12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $fy3$0 = 0, $fy3$0$lcssa = 0, $fy4$0 = 0, $fy4$0$phi = 0, $nPos$0 = 0;
 var $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 56|0); //@line 1123 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1123 "fastmix.cpp"
 $2 = ((($pChannel)) + 60|0); //@line 1123 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1123 "fastmix.cpp"
 $4 = ((($pChannel)) + 64|0); //@line 1123 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1123 "fastmix.cpp"
 $6 = ((($pChannel)) + 68|0); //@line 1123 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1123 "fastmix.cpp"
 $8 = ((($pChannel)) + 8|0); //@line 1124 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1124 "fastmix.cpp"
 $10 = HEAP32[$pChannel>>2]|0; //@line 1124 "fastmix.cpp"
 $11 = ((($pChannel)) + 4|0); //@line 1124 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1124 "fastmix.cpp"
 $13 = $12 << 1; //@line 1124 "fastmix.cpp"
 $14 = (($10) + ($13)|0); //@line 1124 "fastmix.cpp"
 $15 = ((($pChannel)) + 36|0); //@line 1124 "fastmix.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1124 "fastmix.cpp"
 $17 = $16 & 64; //@line 1124 "fastmix.cpp"
 $18 = ($17|0)==(0); //@line 1124 "fastmix.cpp"
 $19 = (($14) + ($12<<1)|0); //@line 1124 "fastmix.cpp"
 $$ = $18 ? $14 : $19; //@line 1124 "fastmix.cpp"
 $20 = ((($pChannel)) + 72|0); //@line 1126 "fastmix.cpp"
 $21 = ((($pChannel)) + 76|0); //@line 1126 "fastmix.cpp"
 $22 = ((($pChannel)) + 80|0); //@line 1126 "fastmix.cpp"
 $23 = ((($pChannel)) + 16|0); //@line 1127 "fastmix.cpp"
 $24 = ((($pChannel)) + 20|0); //@line 1127 "fastmix.cpp"
 $25 = ((($pChannel)) + 12|0); //@line 1128 "fastmix.cpp"
 $fy1$0 = $1;$fy2$0 = $3;$fy3$0 = $5;$fy4$0 = $7;$nPos$0 = $9;$pvol$0 = $pbuffer;
 while(1) {
  $26 = $nPos$0 >> 16; //@line 1125 "fastmix.cpp"
  $27 = $nPos$0 >>> 8; //@line 1125 "fastmix.cpp"
  $28 = $27 & 255; //@line 1125 "fastmix.cpp"
  $29 = $26 << 1; //@line 1125 "fastmix.cpp"
  $30 = (($$) + ($29<<1)|0); //@line 1125 "fastmix.cpp"
  $31 = HEAP16[$30>>1]|0; //@line 1125 "fastmix.cpp"
  $32 = $31 << 16 >> 16; //@line 1125 "fastmix.cpp"
  $33 = (($29) + 2)|0; //@line 1125 "fastmix.cpp"
  $34 = (($$) + ($33<<1)|0); //@line 1125 "fastmix.cpp"
  $35 = HEAP16[$34>>1]|0; //@line 1125 "fastmix.cpp"
  $36 = $35 << 16 >> 16; //@line 1125 "fastmix.cpp"
  $37 = (($36) - ($32))|0; //@line 1125 "fastmix.cpp"
  $38 = Math_imul($37, $28)|0; //@line 1125 "fastmix.cpp"
  $39 = $38 >> 8; //@line 1125 "fastmix.cpp"
  $40 = (($39) + ($32))|0; //@line 1125 "fastmix.cpp"
  $41 = $29 | 1; //@line 1125 "fastmix.cpp"
  $42 = (($$) + ($41<<1)|0); //@line 1125 "fastmix.cpp"
  $43 = HEAP16[$42>>1]|0; //@line 1125 "fastmix.cpp"
  $44 = $43 << 16 >> 16; //@line 1125 "fastmix.cpp"
  $45 = (($29) + 3)|0; //@line 1125 "fastmix.cpp"
  $46 = (($$) + ($45<<1)|0); //@line 1125 "fastmix.cpp"
  $47 = HEAP16[$46>>1]|0; //@line 1125 "fastmix.cpp"
  $48 = $47 << 16 >> 16; //@line 1125 "fastmix.cpp"
  $49 = (($48) - ($44))|0; //@line 1125 "fastmix.cpp"
  $50 = Math_imul($49, $28)|0; //@line 1125 "fastmix.cpp"
  $51 = $50 >> 8; //@line 1125 "fastmix.cpp"
  $52 = (($51) + ($44))|0; //@line 1125 "fastmix.cpp"
  $53 = HEAP32[$20>>2]|0; //@line 1126 "fastmix.cpp"
  $54 = Math_imul($40, $53)|0; //@line 1126 "fastmix.cpp"
  $55 = HEAP32[$21>>2]|0; //@line 1126 "fastmix.cpp"
  $56 = Math_imul($55, $fy1$0)|0; //@line 1126 "fastmix.cpp"
  $57 = HEAP32[$22>>2]|0; //@line 1126 "fastmix.cpp"
  $58 = Math_imul($57, $fy2$0)|0; //@line 1126 "fastmix.cpp"
  $59 = (($56) + 4096)|0; //@line 1126 "fastmix.cpp"
  $60 = (($59) + ($54))|0; //@line 1126 "fastmix.cpp"
  $61 = (($60) + ($58))|0; //@line 1126 "fastmix.cpp"
  $62 = $61 >> 13; //@line 1126 "fastmix.cpp"
  $63 = Math_imul($52, $53)|0; //@line 1126 "fastmix.cpp"
  $64 = Math_imul($55, $fy3$0)|0; //@line 1126 "fastmix.cpp"
  $65 = Math_imul($57, $fy4$0)|0; //@line 1126 "fastmix.cpp"
  $66 = (($64) + 4096)|0; //@line 1126 "fastmix.cpp"
  $67 = (($66) + ($65))|0; //@line 1126 "fastmix.cpp"
  $68 = (($67) + ($63))|0; //@line 1126 "fastmix.cpp"
  $69 = $68 >> 13; //@line 1126 "fastmix.cpp"
  $70 = HEAP32[$23>>2]|0; //@line 1127 "fastmix.cpp"
  $71 = Math_imul($62, $70)|0; //@line 1127 "fastmix.cpp"
  $72 = HEAP32[$pvol$0>>2]|0; //@line 1127 "fastmix.cpp"
  $73 = (($71) + ($72))|0; //@line 1127 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $73; //@line 1127 "fastmix.cpp"
  $74 = HEAP32[$24>>2]|0; //@line 1127 "fastmix.cpp"
  $75 = Math_imul($69, $74)|0; //@line 1127 "fastmix.cpp"
  $76 = ((($pvol$0)) + 4|0); //@line 1127 "fastmix.cpp"
  $77 = HEAP32[$76>>2]|0; //@line 1127 "fastmix.cpp"
  $78 = (($75) + ($77))|0; //@line 1127 "fastmix.cpp"
  HEAP32[$76>>2] = $78; //@line 1127 "fastmix.cpp"
  $79 = ((($pvol$0)) + 8|0); //@line 1127 "fastmix.cpp"
  $80 = HEAP32[$25>>2]|0; //@line 1128 "fastmix.cpp"
  $81 = (($80) + ($nPos$0))|0; //@line 1128 "fastmix.cpp"
  $82 = ($79>>>0)<($pbufmax>>>0); //@line 1128 "fastmix.cpp"
  if ($82) {
   $fy4$0$phi = $fy3$0;$fy2$0$phi = $fy1$0;$fy1$0 = $62;$fy3$0 = $69;$nPos$0 = $81;$pvol$0 = $79;$fy4$0 = $fy4$0$phi;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $62;$$lcssa11 = $69;$$lcssa12 = $81;$fy1$0$lcssa = $fy1$0;$fy3$0$lcssa = $fy3$0;
   break;
  }
 }
 $83 = $$lcssa12 >> 16; //@line 1128 "fastmix.cpp"
 $84 = HEAP32[$11>>2]|0; //@line 1128 "fastmix.cpp"
 $85 = (($84) + ($83))|0; //@line 1128 "fastmix.cpp"
 HEAP32[$11>>2] = $85; //@line 1128 "fastmix.cpp"
 $86 = $$lcssa12 & 65535; //@line 1128 "fastmix.cpp"
 HEAP32[$8>>2] = $86; //@line 1128 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 1128 "fastmix.cpp"
 HEAP32[$2>>2] = $fy1$0$lcssa; //@line 1128 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa11; //@line 1128 "fastmix.cpp"
 HEAP32[$6>>2] = $fy3$0$lcssa; //@line 1128 "fastmix.cpp"
 return; //@line 1128 "fastmix.cpp"
}
function __Z27FilterMono8BitLinearRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa29 = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$lcssa33 = 0, $$sum = 0, $$sum1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $8 = 0, $9 = 0;
 var $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 1058 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1058 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 1058 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1058 "fastmix.cpp"
 $4 = ((($pChannel)) + 56|0); //@line 1058 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1058 "fastmix.cpp"
 $6 = ((($pChannel)) + 60|0); //@line 1058 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1058 "fastmix.cpp"
 $8 = ((($pChannel)) + 8|0); //@line 1059 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1059 "fastmix.cpp"
 $10 = HEAP32[$pChannel>>2]|0; //@line 1059 "fastmix.cpp"
 $11 = ((($pChannel)) + 4|0); //@line 1059 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1059 "fastmix.cpp"
 $13 = (($10) + ($12)|0); //@line 1059 "fastmix.cpp"
 $14 = ((($pChannel)) + 36|0); //@line 1059 "fastmix.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 1059 "fastmix.cpp"
 $16 = $15 & 64; //@line 1059 "fastmix.cpp"
 $17 = ($16|0)==(0); //@line 1059 "fastmix.cpp"
 $$ = $17 ? $10 : $13; //@line 1059 "fastmix.cpp"
 $18 = (($12) + 1)|0; //@line 1060 "fastmix.cpp"
 $19 = ((($pChannel)) + 72|0); //@line 1061 "fastmix.cpp"
 $20 = ((($pChannel)) + 76|0); //@line 1061 "fastmix.cpp"
 $21 = ((($pChannel)) + 80|0); //@line 1061 "fastmix.cpp"
 $22 = ((($pChannel)) + 28|0); //@line 1062 "fastmix.cpp"
 $23 = ((($pChannel)) + 24|0); //@line 1062 "fastmix.cpp"
 $24 = ((($pChannel)) + 12|0); //@line 1063 "fastmix.cpp"
 $fy1$0 = $5;$fy2$0 = $7;$nPos$0 = $9;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $25 = $nPos$0 >> 16; //@line 1060 "fastmix.cpp"
  $26 = $nPos$0 >>> 8; //@line 1060 "fastmix.cpp"
  $27 = $26 & 255; //@line 1060 "fastmix.cpp"
  $$sum = (($25) + ($12))|0; //@line 1060 "fastmix.cpp"
  $28 = (($$) + ($$sum)|0); //@line 1060 "fastmix.cpp"
  $29 = HEAP8[$28>>0]|0; //@line 1060 "fastmix.cpp"
  $30 = $29 << 24 >> 24; //@line 1060 "fastmix.cpp"
  $$sum1 = (($18) + ($25))|0; //@line 1060 "fastmix.cpp"
  $31 = (($$) + ($$sum1)|0); //@line 1060 "fastmix.cpp"
  $32 = HEAP8[$31>>0]|0; //@line 1060 "fastmix.cpp"
  $33 = $32 << 24 >> 24; //@line 1060 "fastmix.cpp"
  $34 = $30 << 8; //@line 1060 "fastmix.cpp"
  $35 = (($33) - ($30))|0; //@line 1060 "fastmix.cpp"
  $36 = Math_imul($35, $27)|0; //@line 1060 "fastmix.cpp"
  $37 = (($36) + ($34))|0; //@line 1060 "fastmix.cpp"
  $38 = HEAP32[$19>>2]|0; //@line 1061 "fastmix.cpp"
  $39 = Math_imul($37, $38)|0; //@line 1061 "fastmix.cpp"
  $40 = HEAP32[$20>>2]|0; //@line 1061 "fastmix.cpp"
  $41 = Math_imul($40, $fy1$0)|0; //@line 1061 "fastmix.cpp"
  $42 = HEAP32[$21>>2]|0; //@line 1061 "fastmix.cpp"
  $43 = Math_imul($42, $fy2$0)|0; //@line 1061 "fastmix.cpp"
  $44 = (($41) + 4096)|0; //@line 1061 "fastmix.cpp"
  $45 = (($44) + ($43))|0; //@line 1061 "fastmix.cpp"
  $46 = (($45) + ($39))|0; //@line 1061 "fastmix.cpp"
  $47 = $46 >> 13; //@line 1061 "fastmix.cpp"
  $48 = HEAP32[$22>>2]|0; //@line 1062 "fastmix.cpp"
  $49 = (($48) + ($nRampLeftVol$0))|0; //@line 1062 "fastmix.cpp"
  $50 = HEAP32[$23>>2]|0; //@line 1062 "fastmix.cpp"
  $51 = (($50) + ($nRampRightVol$0))|0; //@line 1062 "fastmix.cpp"
  $52 = $51 >> 12; //@line 1062 "fastmix.cpp"
  $53 = Math_imul($47, $52)|0; //@line 1062 "fastmix.cpp"
  $54 = HEAP32[$pvol$0>>2]|0; //@line 1062 "fastmix.cpp"
  $55 = (($53) + ($54))|0; //@line 1062 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $55; //@line 1062 "fastmix.cpp"
  $56 = $49 >> 12; //@line 1062 "fastmix.cpp"
  $57 = Math_imul($47, $56)|0; //@line 1062 "fastmix.cpp"
  $58 = ((($pvol$0)) + 4|0); //@line 1062 "fastmix.cpp"
  $59 = HEAP32[$58>>2]|0; //@line 1062 "fastmix.cpp"
  $60 = (($57) + ($59))|0; //@line 1062 "fastmix.cpp"
  HEAP32[$58>>2] = $60; //@line 1062 "fastmix.cpp"
  $61 = ((($pvol$0)) + 8|0); //@line 1062 "fastmix.cpp"
  $62 = HEAP32[$24>>2]|0; //@line 1063 "fastmix.cpp"
  $63 = (($62) + ($nPos$0))|0; //@line 1063 "fastmix.cpp"
  $64 = ($61>>>0)<($pbufmax>>>0); //@line 1063 "fastmix.cpp"
  if ($64) {
   $fy2$0$phi = $fy1$0;$fy1$0 = $47;$nPos$0 = $63;$nRampLeftVol$0 = $49;$nRampRightVol$0 = $51;$pvol$0 = $61;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $47;$$lcssa29 = $49;$$lcssa30 = $51;$$lcssa31 = $52;$$lcssa32 = $56;$$lcssa33 = $63;$fy1$0$lcssa = $fy1$0;
   break;
  }
 }
 $65 = $$lcssa33 >> 16; //@line 1063 "fastmix.cpp"
 $66 = HEAP32[$11>>2]|0; //@line 1063 "fastmix.cpp"
 $67 = (($66) + ($65))|0; //@line 1063 "fastmix.cpp"
 HEAP32[$11>>2] = $67; //@line 1063 "fastmix.cpp"
 $68 = $$lcssa33 & 65535; //@line 1063 "fastmix.cpp"
 HEAP32[$8>>2] = $68; //@line 1063 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa; //@line 1063 "fastmix.cpp"
 HEAP32[$6>>2] = $fy1$0$lcssa; //@line 1063 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa30; //@line 1063 "fastmix.cpp"
 $69 = ((($pChannel)) + 16|0); //@line 1063 "fastmix.cpp"
 HEAP32[$69>>2] = $$lcssa31; //@line 1063 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa29; //@line 1063 "fastmix.cpp"
 $70 = ((($pChannel)) + 20|0); //@line 1063 "fastmix.cpp"
 HEAP32[$70>>2] = $$lcssa32; //@line 1063 "fastmix.cpp"
 return; //@line 1063 "fastmix.cpp"
}
function __Z28FilterMono16BitLinearRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa26 = 0, $$lcssa27 = 0, $$lcssa28 = 0, $$lcssa29 = 0, $$lcssa30 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0;
 var $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 1065 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1065 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 1065 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1065 "fastmix.cpp"
 $4 = ((($pChannel)) + 56|0); //@line 1065 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1065 "fastmix.cpp"
 $6 = ((($pChannel)) + 60|0); //@line 1065 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1065 "fastmix.cpp"
 $8 = ((($pChannel)) + 8|0); //@line 1066 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1066 "fastmix.cpp"
 $10 = HEAP32[$pChannel>>2]|0; //@line 1066 "fastmix.cpp"
 $11 = ((($pChannel)) + 4|0); //@line 1066 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1066 "fastmix.cpp"
 $13 = $12 << 1; //@line 1066 "fastmix.cpp"
 $14 = (($10) + ($13)|0); //@line 1066 "fastmix.cpp"
 $15 = ((($pChannel)) + 36|0); //@line 1066 "fastmix.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1066 "fastmix.cpp"
 $17 = $16 & 64; //@line 1066 "fastmix.cpp"
 $18 = ($17|0)==(0); //@line 1066 "fastmix.cpp"
 $19 = (($14) + ($12<<1)|0); //@line 1066 "fastmix.cpp"
 $$ = $18 ? $14 : $19; //@line 1066 "fastmix.cpp"
 $20 = ((($pChannel)) + 72|0); //@line 1068 "fastmix.cpp"
 $21 = ((($pChannel)) + 76|0); //@line 1068 "fastmix.cpp"
 $22 = ((($pChannel)) + 80|0); //@line 1068 "fastmix.cpp"
 $23 = ((($pChannel)) + 28|0); //@line 1069 "fastmix.cpp"
 $24 = ((($pChannel)) + 24|0); //@line 1069 "fastmix.cpp"
 $25 = ((($pChannel)) + 12|0); //@line 1070 "fastmix.cpp"
 $fy1$0 = $5;$fy2$0 = $7;$nPos$0 = $9;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $26 = $nPos$0 >> 16; //@line 1067 "fastmix.cpp"
  $27 = $nPos$0 >>> 8; //@line 1067 "fastmix.cpp"
  $28 = $27 & 255; //@line 1067 "fastmix.cpp"
  $29 = (($$) + ($26<<1)|0); //@line 1067 "fastmix.cpp"
  $30 = HEAP16[$29>>1]|0; //@line 1067 "fastmix.cpp"
  $31 = $30 << 16 >> 16; //@line 1067 "fastmix.cpp"
  $32 = (($26) + 1)|0; //@line 1067 "fastmix.cpp"
  $33 = (($$) + ($32<<1)|0); //@line 1067 "fastmix.cpp"
  $34 = HEAP16[$33>>1]|0; //@line 1067 "fastmix.cpp"
  $35 = $34 << 16 >> 16; //@line 1067 "fastmix.cpp"
  $36 = (($35) - ($31))|0; //@line 1067 "fastmix.cpp"
  $37 = Math_imul($36, $28)|0; //@line 1067 "fastmix.cpp"
  $38 = $37 >> 8; //@line 1067 "fastmix.cpp"
  $39 = (($38) + ($31))|0; //@line 1067 "fastmix.cpp"
  $40 = HEAP32[$20>>2]|0; //@line 1068 "fastmix.cpp"
  $41 = Math_imul($39, $40)|0; //@line 1068 "fastmix.cpp"
  $42 = HEAP32[$21>>2]|0; //@line 1068 "fastmix.cpp"
  $43 = Math_imul($42, $fy1$0)|0; //@line 1068 "fastmix.cpp"
  $44 = HEAP32[$22>>2]|0; //@line 1068 "fastmix.cpp"
  $45 = Math_imul($44, $fy2$0)|0; //@line 1068 "fastmix.cpp"
  $46 = (($43) + 4096)|0; //@line 1068 "fastmix.cpp"
  $47 = (($46) + ($45))|0; //@line 1068 "fastmix.cpp"
  $48 = (($47) + ($41))|0; //@line 1068 "fastmix.cpp"
  $49 = $48 >> 13; //@line 1068 "fastmix.cpp"
  $50 = HEAP32[$23>>2]|0; //@line 1069 "fastmix.cpp"
  $51 = (($50) + ($nRampLeftVol$0))|0; //@line 1069 "fastmix.cpp"
  $52 = HEAP32[$24>>2]|0; //@line 1069 "fastmix.cpp"
  $53 = (($52) + ($nRampRightVol$0))|0; //@line 1069 "fastmix.cpp"
  $54 = $53 >> 12; //@line 1069 "fastmix.cpp"
  $55 = Math_imul($49, $54)|0; //@line 1069 "fastmix.cpp"
  $56 = HEAP32[$pvol$0>>2]|0; //@line 1069 "fastmix.cpp"
  $57 = (($55) + ($56))|0; //@line 1069 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $57; //@line 1069 "fastmix.cpp"
  $58 = $51 >> 12; //@line 1069 "fastmix.cpp"
  $59 = Math_imul($49, $58)|0; //@line 1069 "fastmix.cpp"
  $60 = ((($pvol$0)) + 4|0); //@line 1069 "fastmix.cpp"
  $61 = HEAP32[$60>>2]|0; //@line 1069 "fastmix.cpp"
  $62 = (($59) + ($61))|0; //@line 1069 "fastmix.cpp"
  HEAP32[$60>>2] = $62; //@line 1069 "fastmix.cpp"
  $63 = ((($pvol$0)) + 8|0); //@line 1069 "fastmix.cpp"
  $64 = HEAP32[$25>>2]|0; //@line 1070 "fastmix.cpp"
  $65 = (($64) + ($nPos$0))|0; //@line 1070 "fastmix.cpp"
  $66 = ($63>>>0)<($pbufmax>>>0); //@line 1070 "fastmix.cpp"
  if ($66) {
   $fy2$0$phi = $fy1$0;$fy1$0 = $49;$nPos$0 = $65;$nRampLeftVol$0 = $51;$nRampRightVol$0 = $53;$pvol$0 = $63;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $49;$$lcssa26 = $51;$$lcssa27 = $53;$$lcssa28 = $54;$$lcssa29 = $58;$$lcssa30 = $65;$fy1$0$lcssa = $fy1$0;
   break;
  }
 }
 $67 = $$lcssa30 >> 16; //@line 1070 "fastmix.cpp"
 $68 = HEAP32[$11>>2]|0; //@line 1070 "fastmix.cpp"
 $69 = (($68) + ($67))|0; //@line 1070 "fastmix.cpp"
 HEAP32[$11>>2] = $69; //@line 1070 "fastmix.cpp"
 $70 = $$lcssa30 & 65535; //@line 1070 "fastmix.cpp"
 HEAP32[$8>>2] = $70; //@line 1070 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa; //@line 1070 "fastmix.cpp"
 HEAP32[$6>>2] = $fy1$0$lcssa; //@line 1070 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa27; //@line 1070 "fastmix.cpp"
 $71 = ((($pChannel)) + 16|0); //@line 1070 "fastmix.cpp"
 HEAP32[$71>>2] = $$lcssa28; //@line 1070 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa26; //@line 1070 "fastmix.cpp"
 $72 = ((($pChannel)) + 20|0); //@line 1070 "fastmix.cpp"
 HEAP32[$72>>2] = $$lcssa29; //@line 1070 "fastmix.cpp"
 return; //@line 1070 "fastmix.cpp"
}
function __Z29FilterStereo8BitLinearRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa38 = 0, $$lcssa39 = 0, $$lcssa40 = 0, $$lcssa41 = 0, $$lcssa42 = 0, $$lcssa43 = 0, $$sum = 0, $$sum1 = 0, $$sum2 = 0, $$sum3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $fy3$0 = 0, $fy3$0$lcssa = 0, $fy4$0 = 0, $fy4$0$phi = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 1173 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1173 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 1173 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1173 "fastmix.cpp"
 $4 = ((($pChannel)) + 56|0); //@line 1173 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1173 "fastmix.cpp"
 $6 = ((($pChannel)) + 60|0); //@line 1173 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1173 "fastmix.cpp"
 $8 = ((($pChannel)) + 64|0); //@line 1173 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1173 "fastmix.cpp"
 $10 = ((($pChannel)) + 68|0); //@line 1173 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 1173 "fastmix.cpp"
 $12 = ((($pChannel)) + 8|0); //@line 1174 "fastmix.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 1174 "fastmix.cpp"
 $14 = HEAP32[$pChannel>>2]|0; //@line 1174 "fastmix.cpp"
 $15 = ((($pChannel)) + 4|0); //@line 1174 "fastmix.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1174 "fastmix.cpp"
 $17 = (($14) + ($16)|0); //@line 1174 "fastmix.cpp"
 $18 = ((($pChannel)) + 36|0); //@line 1174 "fastmix.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 1174 "fastmix.cpp"
 $20 = $19 & 64; //@line 1174 "fastmix.cpp"
 $21 = ($20|0)==(0); //@line 1174 "fastmix.cpp"
 $$ = $21 ? $14 : $17; //@line 1174 "fastmix.cpp"
 $22 = (($16) + 2)|0; //@line 1175 "fastmix.cpp"
 $23 = (($16) + 3)|0; //@line 1175 "fastmix.cpp"
 $24 = ((($pChannel)) + 72|0); //@line 1176 "fastmix.cpp"
 $25 = ((($pChannel)) + 76|0); //@line 1176 "fastmix.cpp"
 $26 = ((($pChannel)) + 80|0); //@line 1176 "fastmix.cpp"
 $27 = ((($pChannel)) + 28|0); //@line 1177 "fastmix.cpp"
 $28 = ((($pChannel)) + 24|0); //@line 1177 "fastmix.cpp"
 $29 = ((($pChannel)) + 12|0); //@line 1178 "fastmix.cpp"
 $fy1$0 = $5;$fy2$0 = $7;$fy3$0 = $9;$fy4$0 = $11;$nPos$0 = $13;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $30 = $nPos$0 >> 16; //@line 1175 "fastmix.cpp"
  $31 = $nPos$0 >>> 8; //@line 1175 "fastmix.cpp"
  $32 = $31 & 255; //@line 1175 "fastmix.cpp"
  $33 = $30 << 1; //@line 1175 "fastmix.cpp"
  $$sum = (($33) + ($16))|0; //@line 1175 "fastmix.cpp"
  $34 = (($$) + ($$sum)|0); //@line 1175 "fastmix.cpp"
  $35 = HEAP8[$34>>0]|0; //@line 1175 "fastmix.cpp"
  $36 = $35 << 24 >> 24; //@line 1175 "fastmix.cpp"
  $37 = $36 << 8; //@line 1175 "fastmix.cpp"
  $$sum1 = (($22) + ($33))|0; //@line 1175 "fastmix.cpp"
  $38 = (($$) + ($$sum1)|0); //@line 1175 "fastmix.cpp"
  $39 = HEAP8[$38>>0]|0; //@line 1175 "fastmix.cpp"
  $40 = $39 << 24 >> 24; //@line 1175 "fastmix.cpp"
  $41 = (($40) - ($36))|0; //@line 1175 "fastmix.cpp"
  $42 = Math_imul($41, $32)|0; //@line 1175 "fastmix.cpp"
  $43 = (($42) + ($37))|0; //@line 1175 "fastmix.cpp"
  $44 = $33 | 1; //@line 1175 "fastmix.cpp"
  $$sum2 = (($44) + ($16))|0; //@line 1175 "fastmix.cpp"
  $45 = (($$) + ($$sum2)|0); //@line 1175 "fastmix.cpp"
  $46 = HEAP8[$45>>0]|0; //@line 1175 "fastmix.cpp"
  $47 = $46 << 24 >> 24; //@line 1175 "fastmix.cpp"
  $48 = $47 << 8; //@line 1175 "fastmix.cpp"
  $$sum3 = (($23) + ($33))|0; //@line 1175 "fastmix.cpp"
  $49 = (($$) + ($$sum3)|0); //@line 1175 "fastmix.cpp"
  $50 = HEAP8[$49>>0]|0; //@line 1175 "fastmix.cpp"
  $51 = $50 << 24 >> 24; //@line 1175 "fastmix.cpp"
  $52 = (($51) - ($47))|0; //@line 1175 "fastmix.cpp"
  $53 = Math_imul($52, $32)|0; //@line 1175 "fastmix.cpp"
  $54 = (($53) + ($48))|0; //@line 1175 "fastmix.cpp"
  $55 = HEAP32[$24>>2]|0; //@line 1176 "fastmix.cpp"
  $56 = Math_imul($43, $55)|0; //@line 1176 "fastmix.cpp"
  $57 = HEAP32[$25>>2]|0; //@line 1176 "fastmix.cpp"
  $58 = Math_imul($57, $fy1$0)|0; //@line 1176 "fastmix.cpp"
  $59 = HEAP32[$26>>2]|0; //@line 1176 "fastmix.cpp"
  $60 = Math_imul($59, $fy2$0)|0; //@line 1176 "fastmix.cpp"
  $61 = (($58) + 4096)|0; //@line 1176 "fastmix.cpp"
  $62 = (($61) + ($56))|0; //@line 1176 "fastmix.cpp"
  $63 = (($62) + ($60))|0; //@line 1176 "fastmix.cpp"
  $64 = $63 >> 13; //@line 1176 "fastmix.cpp"
  $65 = Math_imul($54, $55)|0; //@line 1176 "fastmix.cpp"
  $66 = Math_imul($57, $fy3$0)|0; //@line 1176 "fastmix.cpp"
  $67 = Math_imul($59, $fy4$0)|0; //@line 1176 "fastmix.cpp"
  $68 = (($66) + 4096)|0; //@line 1176 "fastmix.cpp"
  $69 = (($68) + ($67))|0; //@line 1176 "fastmix.cpp"
  $70 = (($69) + ($65))|0; //@line 1176 "fastmix.cpp"
  $71 = $70 >> 13; //@line 1176 "fastmix.cpp"
  $72 = HEAP32[$27>>2]|0; //@line 1177 "fastmix.cpp"
  $73 = (($72) + ($nRampLeftVol$0))|0; //@line 1177 "fastmix.cpp"
  $74 = HEAP32[$28>>2]|0; //@line 1177 "fastmix.cpp"
  $75 = (($74) + ($nRampRightVol$0))|0; //@line 1177 "fastmix.cpp"
  $76 = $75 >> 12; //@line 1177 "fastmix.cpp"
  $77 = Math_imul($64, $76)|0; //@line 1177 "fastmix.cpp"
  $78 = HEAP32[$pvol$0>>2]|0; //@line 1177 "fastmix.cpp"
  $79 = (($77) + ($78))|0; //@line 1177 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $79; //@line 1177 "fastmix.cpp"
  $80 = $73 >> 12; //@line 1177 "fastmix.cpp"
  $81 = Math_imul($71, $80)|0; //@line 1177 "fastmix.cpp"
  $82 = ((($pvol$0)) + 4|0); //@line 1177 "fastmix.cpp"
  $83 = HEAP32[$82>>2]|0; //@line 1177 "fastmix.cpp"
  $84 = (($81) + ($83))|0; //@line 1177 "fastmix.cpp"
  HEAP32[$82>>2] = $84; //@line 1177 "fastmix.cpp"
  $85 = ((($pvol$0)) + 8|0); //@line 1177 "fastmix.cpp"
  $86 = HEAP32[$29>>2]|0; //@line 1178 "fastmix.cpp"
  $87 = (($86) + ($nPos$0))|0; //@line 1178 "fastmix.cpp"
  $88 = ($85>>>0)<($pbufmax>>>0); //@line 1178 "fastmix.cpp"
  if ($88) {
   $fy4$0$phi = $fy3$0;$fy2$0$phi = $fy1$0;$fy1$0 = $64;$fy3$0 = $71;$nPos$0 = $87;$nRampLeftVol$0 = $73;$nRampRightVol$0 = $75;$pvol$0 = $85;$fy4$0 = $fy4$0$phi;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $64;$$lcssa38 = $71;$$lcssa39 = $73;$$lcssa40 = $75;$$lcssa41 = $76;$$lcssa42 = $80;$$lcssa43 = $87;$fy1$0$lcssa = $fy1$0;$fy3$0$lcssa = $fy3$0;
   break;
  }
 }
 $89 = $$lcssa43 >> 16; //@line 1178 "fastmix.cpp"
 $90 = HEAP32[$15>>2]|0; //@line 1178 "fastmix.cpp"
 $91 = (($90) + ($89))|0; //@line 1178 "fastmix.cpp"
 HEAP32[$15>>2] = $91; //@line 1178 "fastmix.cpp"
 $92 = $$lcssa43 & 65535; //@line 1178 "fastmix.cpp"
 HEAP32[$12>>2] = $92; //@line 1178 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa; //@line 1178 "fastmix.cpp"
 HEAP32[$6>>2] = $fy1$0$lcssa; //@line 1178 "fastmix.cpp"
 HEAP32[$8>>2] = $$lcssa38; //@line 1178 "fastmix.cpp"
 HEAP32[$10>>2] = $fy3$0$lcssa; //@line 1178 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa40; //@line 1178 "fastmix.cpp"
 $93 = ((($pChannel)) + 16|0); //@line 1178 "fastmix.cpp"
 HEAP32[$93>>2] = $$lcssa41; //@line 1178 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa39; //@line 1178 "fastmix.cpp"
 $94 = ((($pChannel)) + 20|0); //@line 1178 "fastmix.cpp"
 HEAP32[$94>>2] = $$lcssa42; //@line 1178 "fastmix.cpp"
 return; //@line 1178 "fastmix.cpp"
}
function __Z30FilterStereo16BitLinearRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$lcssa33 = 0, $$lcssa34 = 0, $$lcssa35 = 0, $$lcssa36 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $fy3$0 = 0, $fy3$0$lcssa = 0, $fy4$0 = 0, $fy4$0$phi = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 1180 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1180 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 1180 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1180 "fastmix.cpp"
 $4 = ((($pChannel)) + 56|0); //@line 1180 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1180 "fastmix.cpp"
 $6 = ((($pChannel)) + 60|0); //@line 1180 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1180 "fastmix.cpp"
 $8 = ((($pChannel)) + 64|0); //@line 1180 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1180 "fastmix.cpp"
 $10 = ((($pChannel)) + 68|0); //@line 1180 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 1180 "fastmix.cpp"
 $12 = ((($pChannel)) + 8|0); //@line 1181 "fastmix.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 1181 "fastmix.cpp"
 $14 = HEAP32[$pChannel>>2]|0; //@line 1181 "fastmix.cpp"
 $15 = ((($pChannel)) + 4|0); //@line 1181 "fastmix.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1181 "fastmix.cpp"
 $17 = $16 << 1; //@line 1181 "fastmix.cpp"
 $18 = (($14) + ($17)|0); //@line 1181 "fastmix.cpp"
 $19 = ((($pChannel)) + 36|0); //@line 1181 "fastmix.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 1181 "fastmix.cpp"
 $21 = $20 & 64; //@line 1181 "fastmix.cpp"
 $22 = ($21|0)==(0); //@line 1181 "fastmix.cpp"
 $23 = (($18) + ($16<<1)|0); //@line 1181 "fastmix.cpp"
 $$ = $22 ? $18 : $23; //@line 1181 "fastmix.cpp"
 $24 = ((($pChannel)) + 72|0); //@line 1183 "fastmix.cpp"
 $25 = ((($pChannel)) + 76|0); //@line 1183 "fastmix.cpp"
 $26 = ((($pChannel)) + 80|0); //@line 1183 "fastmix.cpp"
 $27 = ((($pChannel)) + 28|0); //@line 1184 "fastmix.cpp"
 $28 = ((($pChannel)) + 24|0); //@line 1184 "fastmix.cpp"
 $29 = ((($pChannel)) + 12|0); //@line 1185 "fastmix.cpp"
 $fy1$0 = $5;$fy2$0 = $7;$fy3$0 = $9;$fy4$0 = $11;$nPos$0 = $13;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $30 = $nPos$0 >> 16; //@line 1182 "fastmix.cpp"
  $31 = $nPos$0 >>> 8; //@line 1182 "fastmix.cpp"
  $32 = $31 & 255; //@line 1182 "fastmix.cpp"
  $33 = $30 << 1; //@line 1182 "fastmix.cpp"
  $34 = (($$) + ($33<<1)|0); //@line 1182 "fastmix.cpp"
  $35 = HEAP16[$34>>1]|0; //@line 1182 "fastmix.cpp"
  $36 = $35 << 16 >> 16; //@line 1182 "fastmix.cpp"
  $37 = (($33) + 2)|0; //@line 1182 "fastmix.cpp"
  $38 = (($$) + ($37<<1)|0); //@line 1182 "fastmix.cpp"
  $39 = HEAP16[$38>>1]|0; //@line 1182 "fastmix.cpp"
  $40 = $39 << 16 >> 16; //@line 1182 "fastmix.cpp"
  $41 = (($40) - ($36))|0; //@line 1182 "fastmix.cpp"
  $42 = Math_imul($41, $32)|0; //@line 1182 "fastmix.cpp"
  $43 = $42 >> 8; //@line 1182 "fastmix.cpp"
  $44 = (($43) + ($36))|0; //@line 1182 "fastmix.cpp"
  $45 = $33 | 1; //@line 1182 "fastmix.cpp"
  $46 = (($$) + ($45<<1)|0); //@line 1182 "fastmix.cpp"
  $47 = HEAP16[$46>>1]|0; //@line 1182 "fastmix.cpp"
  $48 = $47 << 16 >> 16; //@line 1182 "fastmix.cpp"
  $49 = (($33) + 3)|0; //@line 1182 "fastmix.cpp"
  $50 = (($$) + ($49<<1)|0); //@line 1182 "fastmix.cpp"
  $51 = HEAP16[$50>>1]|0; //@line 1182 "fastmix.cpp"
  $52 = $51 << 16 >> 16; //@line 1182 "fastmix.cpp"
  $53 = (($52) - ($48))|0; //@line 1182 "fastmix.cpp"
  $54 = Math_imul($53, $32)|0; //@line 1182 "fastmix.cpp"
  $55 = $54 >> 8; //@line 1182 "fastmix.cpp"
  $56 = (($55) + ($48))|0; //@line 1182 "fastmix.cpp"
  $57 = HEAP32[$24>>2]|0; //@line 1183 "fastmix.cpp"
  $58 = Math_imul($44, $57)|0; //@line 1183 "fastmix.cpp"
  $59 = HEAP32[$25>>2]|0; //@line 1183 "fastmix.cpp"
  $60 = Math_imul($59, $fy1$0)|0; //@line 1183 "fastmix.cpp"
  $61 = HEAP32[$26>>2]|0; //@line 1183 "fastmix.cpp"
  $62 = Math_imul($61, $fy2$0)|0; //@line 1183 "fastmix.cpp"
  $63 = (($60) + 4096)|0; //@line 1183 "fastmix.cpp"
  $64 = (($63) + ($58))|0; //@line 1183 "fastmix.cpp"
  $65 = (($64) + ($62))|0; //@line 1183 "fastmix.cpp"
  $66 = $65 >> 13; //@line 1183 "fastmix.cpp"
  $67 = Math_imul($56, $57)|0; //@line 1183 "fastmix.cpp"
  $68 = Math_imul($59, $fy3$0)|0; //@line 1183 "fastmix.cpp"
  $69 = Math_imul($61, $fy4$0)|0; //@line 1183 "fastmix.cpp"
  $70 = (($68) + 4096)|0; //@line 1183 "fastmix.cpp"
  $71 = (($70) + ($69))|0; //@line 1183 "fastmix.cpp"
  $72 = (($71) + ($67))|0; //@line 1183 "fastmix.cpp"
  $73 = $72 >> 13; //@line 1183 "fastmix.cpp"
  $74 = HEAP32[$27>>2]|0; //@line 1184 "fastmix.cpp"
  $75 = (($74) + ($nRampLeftVol$0))|0; //@line 1184 "fastmix.cpp"
  $76 = HEAP32[$28>>2]|0; //@line 1184 "fastmix.cpp"
  $77 = (($76) + ($nRampRightVol$0))|0; //@line 1184 "fastmix.cpp"
  $78 = $77 >> 12; //@line 1184 "fastmix.cpp"
  $79 = Math_imul($66, $78)|0; //@line 1184 "fastmix.cpp"
  $80 = HEAP32[$pvol$0>>2]|0; //@line 1184 "fastmix.cpp"
  $81 = (($79) + ($80))|0; //@line 1184 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $81; //@line 1184 "fastmix.cpp"
  $82 = $75 >> 12; //@line 1184 "fastmix.cpp"
  $83 = Math_imul($73, $82)|0; //@line 1184 "fastmix.cpp"
  $84 = ((($pvol$0)) + 4|0); //@line 1184 "fastmix.cpp"
  $85 = HEAP32[$84>>2]|0; //@line 1184 "fastmix.cpp"
  $86 = (($83) + ($85))|0; //@line 1184 "fastmix.cpp"
  HEAP32[$84>>2] = $86; //@line 1184 "fastmix.cpp"
  $87 = ((($pvol$0)) + 8|0); //@line 1184 "fastmix.cpp"
  $88 = HEAP32[$29>>2]|0; //@line 1185 "fastmix.cpp"
  $89 = (($88) + ($nPos$0))|0; //@line 1185 "fastmix.cpp"
  $90 = ($87>>>0)<($pbufmax>>>0); //@line 1185 "fastmix.cpp"
  if ($90) {
   $fy4$0$phi = $fy3$0;$fy2$0$phi = $fy1$0;$fy1$0 = $66;$fy3$0 = $73;$nPos$0 = $89;$nRampLeftVol$0 = $75;$nRampRightVol$0 = $77;$pvol$0 = $87;$fy4$0 = $fy4$0$phi;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $66;$$lcssa31 = $73;$$lcssa32 = $75;$$lcssa33 = $77;$$lcssa34 = $78;$$lcssa35 = $82;$$lcssa36 = $89;$fy1$0$lcssa = $fy1$0;$fy3$0$lcssa = $fy3$0;
   break;
  }
 }
 $91 = $$lcssa36 >> 16; //@line 1185 "fastmix.cpp"
 $92 = HEAP32[$15>>2]|0; //@line 1185 "fastmix.cpp"
 $93 = (($92) + ($91))|0; //@line 1185 "fastmix.cpp"
 HEAP32[$15>>2] = $93; //@line 1185 "fastmix.cpp"
 $94 = $$lcssa36 & 65535; //@line 1185 "fastmix.cpp"
 HEAP32[$12>>2] = $94; //@line 1185 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa; //@line 1185 "fastmix.cpp"
 HEAP32[$6>>2] = $fy1$0$lcssa; //@line 1185 "fastmix.cpp"
 HEAP32[$8>>2] = $$lcssa31; //@line 1185 "fastmix.cpp"
 HEAP32[$10>>2] = $fy3$0$lcssa; //@line 1185 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa33; //@line 1185 "fastmix.cpp"
 $95 = ((($pChannel)) + 16|0); //@line 1185 "fastmix.cpp"
 HEAP32[$95>>2] = $$lcssa34; //@line 1185 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa32; //@line 1185 "fastmix.cpp"
 $96 = ((($pChannel)) + 20|0); //@line 1185 "fastmix.cpp"
 HEAP32[$96>>2] = $$lcssa35; //@line 1185 "fastmix.cpp"
 return; //@line 1185 "fastmix.cpp"
}
function __Z17Mono8BitSplineMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$sum = 0, $$sum1 = 0, $$sum2 = 0, $$sum3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 702 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 702 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 702 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 702 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 702 "fastmix.cpp"
 $5 = (($2) + ($4)|0); //@line 702 "fastmix.cpp"
 $6 = ((($pChannel)) + 36|0); //@line 702 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 702 "fastmix.cpp"
 $8 = $7 & 64; //@line 702 "fastmix.cpp"
 $9 = ($8|0)==(0); //@line 702 "fastmix.cpp"
 $$ = $9 ? $2 : $5; //@line 702 "fastmix.cpp"
 $10 = (($4) + -1)|0; //@line 703 "fastmix.cpp"
 $11 = (($4) + 2)|0; //@line 703 "fastmix.cpp"
 $12 = (($4) + 1)|0; //@line 703 "fastmix.cpp"
 $13 = ((($pChannel)) + 16|0); //@line 704 "fastmix.cpp"
 $14 = ((($pChannel)) + 20|0); //@line 704 "fastmix.cpp"
 $15 = ((($pChannel)) + 12|0); //@line 705 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $16 = $nPos$0 >> 16; //@line 703 "fastmix.cpp"
  $17 = $nPos$0 >>> 4; //@line 703 "fastmix.cpp"
  $18 = $17 & 4092; //@line 703 "fastmix.cpp"
  $19 = (1209506 + ($18<<1)|0); //@line 703 "fastmix.cpp"
  $20 = HEAP16[$19>>1]|0; //@line 703 "fastmix.cpp"
  $21 = $20 << 16 >> 16; //@line 703 "fastmix.cpp"
  $$sum = (($10) + ($16))|0; //@line 703 "fastmix.cpp"
  $22 = (($$) + ($$sum)|0); //@line 703 "fastmix.cpp"
  $23 = HEAP8[$22>>0]|0; //@line 703 "fastmix.cpp"
  $24 = $23 << 24 >> 24; //@line 703 "fastmix.cpp"
  $25 = Math_imul($24, $21)|0; //@line 703 "fastmix.cpp"
  $26 = $18 | 1; //@line 703 "fastmix.cpp"
  $27 = (1209506 + ($26<<1)|0); //@line 703 "fastmix.cpp"
  $28 = HEAP16[$27>>1]|0; //@line 703 "fastmix.cpp"
  $29 = $28 << 16 >> 16; //@line 703 "fastmix.cpp"
  $$sum1 = (($16) + ($4))|0; //@line 703 "fastmix.cpp"
  $30 = (($$) + ($$sum1)|0); //@line 703 "fastmix.cpp"
  $31 = HEAP8[$30>>0]|0; //@line 703 "fastmix.cpp"
  $32 = $31 << 24 >> 24; //@line 703 "fastmix.cpp"
  $33 = Math_imul($32, $29)|0; //@line 703 "fastmix.cpp"
  $34 = (($33) + ($25))|0; //@line 703 "fastmix.cpp"
  $35 = $18 | 3; //@line 703 "fastmix.cpp"
  $36 = (1209506 + ($35<<1)|0); //@line 703 "fastmix.cpp"
  $37 = HEAP16[$36>>1]|0; //@line 703 "fastmix.cpp"
  $38 = $37 << 16 >> 16; //@line 703 "fastmix.cpp"
  $$sum2 = (($11) + ($16))|0; //@line 703 "fastmix.cpp"
  $39 = (($$) + ($$sum2)|0); //@line 703 "fastmix.cpp"
  $40 = HEAP8[$39>>0]|0; //@line 703 "fastmix.cpp"
  $41 = $40 << 24 >> 24; //@line 703 "fastmix.cpp"
  $42 = Math_imul($41, $38)|0; //@line 703 "fastmix.cpp"
  $43 = (($34) + ($42))|0; //@line 703 "fastmix.cpp"
  $44 = $18 | 2; //@line 703 "fastmix.cpp"
  $45 = (1209506 + ($44<<1)|0); //@line 703 "fastmix.cpp"
  $46 = HEAP16[$45>>1]|0; //@line 703 "fastmix.cpp"
  $47 = $46 << 16 >> 16; //@line 703 "fastmix.cpp"
  $$sum3 = (($12) + ($16))|0; //@line 703 "fastmix.cpp"
  $48 = (($$) + ($$sum3)|0); //@line 703 "fastmix.cpp"
  $49 = HEAP8[$48>>0]|0; //@line 703 "fastmix.cpp"
  $50 = $49 << 24 >> 24; //@line 703 "fastmix.cpp"
  $51 = Math_imul($50, $47)|0; //@line 703 "fastmix.cpp"
  $52 = (($43) + ($51))|0; //@line 703 "fastmix.cpp"
  $53 = $52 >> 6; //@line 703 "fastmix.cpp"
  $54 = HEAP32[$13>>2]|0; //@line 704 "fastmix.cpp"
  $55 = Math_imul($53, $54)|0; //@line 704 "fastmix.cpp"
  $56 = HEAP32[$pvol$0>>2]|0; //@line 704 "fastmix.cpp"
  $57 = (($55) + ($56))|0; //@line 704 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $57; //@line 704 "fastmix.cpp"
  $58 = HEAP32[$14>>2]|0; //@line 704 "fastmix.cpp"
  $59 = Math_imul($53, $58)|0; //@line 704 "fastmix.cpp"
  $60 = ((($pvol$0)) + 4|0); //@line 704 "fastmix.cpp"
  $61 = HEAP32[$60>>2]|0; //@line 704 "fastmix.cpp"
  $62 = (($59) + ($61))|0; //@line 704 "fastmix.cpp"
  HEAP32[$60>>2] = $62; //@line 704 "fastmix.cpp"
  $63 = ((($pvol$0)) + 8|0); //@line 704 "fastmix.cpp"
  $64 = HEAP32[$15>>2]|0; //@line 705 "fastmix.cpp"
  $65 = (($64) + ($nPos$0))|0; //@line 705 "fastmix.cpp"
  $66 = ($63>>>0)<($pbufmax>>>0); //@line 705 "fastmix.cpp"
  if ($66) {
   $nPos$0 = $65;$pvol$0 = $63;
  } else {
   $$lcssa = $65;
   break;
  }
 }
 $67 = $$lcssa >> 16; //@line 705 "fastmix.cpp"
 $68 = HEAP32[$3>>2]|0; //@line 705 "fastmix.cpp"
 $69 = (($68) + ($67))|0; //@line 705 "fastmix.cpp"
 HEAP32[$3>>2] = $69; //@line 705 "fastmix.cpp"
 $70 = $$lcssa & 65535; //@line 705 "fastmix.cpp"
 HEAP32[$0>>2] = $70; //@line 705 "fastmix.cpp"
 return; //@line 705 "fastmix.cpp"
}
function __Z18Mono16BitSplineMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 708 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 708 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 708 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 708 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 708 "fastmix.cpp"
 $5 = $4 << 1; //@line 708 "fastmix.cpp"
 $6 = (($2) + ($5)|0); //@line 708 "fastmix.cpp"
 $7 = ((($pChannel)) + 36|0); //@line 708 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 708 "fastmix.cpp"
 $9 = $8 & 64; //@line 708 "fastmix.cpp"
 $10 = ($9|0)==(0); //@line 708 "fastmix.cpp"
 $11 = (($6) + ($4<<1)|0); //@line 708 "fastmix.cpp"
 $$ = $10 ? $6 : $11; //@line 708 "fastmix.cpp"
 $12 = ((($pChannel)) + 16|0); //@line 710 "fastmix.cpp"
 $13 = ((($pChannel)) + 20|0); //@line 710 "fastmix.cpp"
 $14 = ((($pChannel)) + 12|0); //@line 711 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $15 = $nPos$0 >> 16; //@line 709 "fastmix.cpp"
  $16 = $nPos$0 >>> 4; //@line 709 "fastmix.cpp"
  $17 = $16 & 4092; //@line 709 "fastmix.cpp"
  $18 = (1209506 + ($17<<1)|0); //@line 709 "fastmix.cpp"
  $19 = HEAP16[$18>>1]|0; //@line 709 "fastmix.cpp"
  $20 = $19 << 16 >> 16; //@line 709 "fastmix.cpp"
  $21 = (($15) + -1)|0; //@line 709 "fastmix.cpp"
  $22 = (($$) + ($21<<1)|0); //@line 709 "fastmix.cpp"
  $23 = HEAP16[$22>>1]|0; //@line 709 "fastmix.cpp"
  $24 = $23 << 16 >> 16; //@line 709 "fastmix.cpp"
  $25 = Math_imul($24, $20)|0; //@line 709 "fastmix.cpp"
  $26 = $17 | 1; //@line 709 "fastmix.cpp"
  $27 = (1209506 + ($26<<1)|0); //@line 709 "fastmix.cpp"
  $28 = HEAP16[$27>>1]|0; //@line 709 "fastmix.cpp"
  $29 = $28 << 16 >> 16; //@line 709 "fastmix.cpp"
  $30 = (($$) + ($15<<1)|0); //@line 709 "fastmix.cpp"
  $31 = HEAP16[$30>>1]|0; //@line 709 "fastmix.cpp"
  $32 = $31 << 16 >> 16; //@line 709 "fastmix.cpp"
  $33 = Math_imul($32, $29)|0; //@line 709 "fastmix.cpp"
  $34 = (($33) + ($25))|0; //@line 709 "fastmix.cpp"
  $35 = $17 | 3; //@line 709 "fastmix.cpp"
  $36 = (1209506 + ($35<<1)|0); //@line 709 "fastmix.cpp"
  $37 = HEAP16[$36>>1]|0; //@line 709 "fastmix.cpp"
  $38 = $37 << 16 >> 16; //@line 709 "fastmix.cpp"
  $39 = (($15) + 2)|0; //@line 709 "fastmix.cpp"
  $40 = (($$) + ($39<<1)|0); //@line 709 "fastmix.cpp"
  $41 = HEAP16[$40>>1]|0; //@line 709 "fastmix.cpp"
  $42 = $41 << 16 >> 16; //@line 709 "fastmix.cpp"
  $43 = Math_imul($42, $38)|0; //@line 709 "fastmix.cpp"
  $44 = (($34) + ($43))|0; //@line 709 "fastmix.cpp"
  $45 = $17 | 2; //@line 709 "fastmix.cpp"
  $46 = (1209506 + ($45<<1)|0); //@line 709 "fastmix.cpp"
  $47 = HEAP16[$46>>1]|0; //@line 709 "fastmix.cpp"
  $48 = $47 << 16 >> 16; //@line 709 "fastmix.cpp"
  $49 = (($15) + 1)|0; //@line 709 "fastmix.cpp"
  $50 = (($$) + ($49<<1)|0); //@line 709 "fastmix.cpp"
  $51 = HEAP16[$50>>1]|0; //@line 709 "fastmix.cpp"
  $52 = $51 << 16 >> 16; //@line 709 "fastmix.cpp"
  $53 = Math_imul($52, $48)|0; //@line 709 "fastmix.cpp"
  $54 = (($44) + ($53))|0; //@line 709 "fastmix.cpp"
  $55 = $54 >> 14; //@line 709 "fastmix.cpp"
  $56 = HEAP32[$12>>2]|0; //@line 710 "fastmix.cpp"
  $57 = Math_imul($55, $56)|0; //@line 710 "fastmix.cpp"
  $58 = HEAP32[$pvol$0>>2]|0; //@line 710 "fastmix.cpp"
  $59 = (($57) + ($58))|0; //@line 710 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $59; //@line 710 "fastmix.cpp"
  $60 = HEAP32[$13>>2]|0; //@line 710 "fastmix.cpp"
  $61 = Math_imul($55, $60)|0; //@line 710 "fastmix.cpp"
  $62 = ((($pvol$0)) + 4|0); //@line 710 "fastmix.cpp"
  $63 = HEAP32[$62>>2]|0; //@line 710 "fastmix.cpp"
  $64 = (($61) + ($63))|0; //@line 710 "fastmix.cpp"
  HEAP32[$62>>2] = $64; //@line 710 "fastmix.cpp"
  $65 = ((($pvol$0)) + 8|0); //@line 710 "fastmix.cpp"
  $66 = HEAP32[$14>>2]|0; //@line 711 "fastmix.cpp"
  $67 = (($66) + ($nPos$0))|0; //@line 711 "fastmix.cpp"
  $68 = ($65>>>0)<($pbufmax>>>0); //@line 711 "fastmix.cpp"
  if ($68) {
   $nPos$0 = $67;$pvol$0 = $65;
  } else {
   $$lcssa = $67;
   break;
  }
 }
 $69 = $$lcssa >> 16; //@line 711 "fastmix.cpp"
 $70 = HEAP32[$3>>2]|0; //@line 711 "fastmix.cpp"
 $71 = (($70) + ($69))|0; //@line 711 "fastmix.cpp"
 HEAP32[$3>>2] = $71; //@line 711 "fastmix.cpp"
 $72 = $$lcssa & 65535; //@line 711 "fastmix.cpp"
 HEAP32[$0>>2] = $72; //@line 711 "fastmix.cpp"
 return; //@line 711 "fastmix.cpp"
}
function __Z19Stereo8BitSplineMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$sum = 0, $$sum1 = 0, $$sum2 = 0, $$sum3 = 0, $$sum4 = 0, $$sum5 = 0, $$sum6 = 0, $$sum7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 906 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 906 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 906 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 906 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 906 "fastmix.cpp"
 $5 = (($2) + ($4)|0); //@line 906 "fastmix.cpp"
 $6 = ((($pChannel)) + 36|0); //@line 906 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 906 "fastmix.cpp"
 $8 = $7 & 64; //@line 906 "fastmix.cpp"
 $9 = ($8|0)==(0); //@line 906 "fastmix.cpp"
 $$ = $9 ? $2 : $5; //@line 906 "fastmix.cpp"
 $10 = (($4) + -2)|0; //@line 907 "fastmix.cpp"
 $11 = (($4) + 2)|0; //@line 907 "fastmix.cpp"
 $12 = (($4) + 4)|0; //@line 907 "fastmix.cpp"
 $13 = (($4) + -1)|0; //@line 907 "fastmix.cpp"
 $14 = (($4) + 3)|0; //@line 907 "fastmix.cpp"
 $15 = (($4) + 5)|0; //@line 907 "fastmix.cpp"
 $16 = ((($pChannel)) + 16|0); //@line 908 "fastmix.cpp"
 $17 = ((($pChannel)) + 20|0); //@line 908 "fastmix.cpp"
 $18 = ((($pChannel)) + 12|0); //@line 909 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $19 = $nPos$0 >> 16; //@line 907 "fastmix.cpp"
  $20 = $nPos$0 >>> 4; //@line 907 "fastmix.cpp"
  $21 = $20 & 4092; //@line 907 "fastmix.cpp"
  $22 = (1209506 + ($21<<1)|0); //@line 907 "fastmix.cpp"
  $23 = HEAP16[$22>>1]|0; //@line 907 "fastmix.cpp"
  $24 = $23 << 16 >> 16; //@line 907 "fastmix.cpp"
  $25 = $19 << 1; //@line 907 "fastmix.cpp"
  $$sum = (($10) + ($25))|0; //@line 907 "fastmix.cpp"
  $26 = (($$) + ($$sum)|0); //@line 907 "fastmix.cpp"
  $27 = HEAP8[$26>>0]|0; //@line 907 "fastmix.cpp"
  $28 = $27 << 24 >> 24; //@line 907 "fastmix.cpp"
  $29 = Math_imul($28, $24)|0; //@line 907 "fastmix.cpp"
  $30 = $21 | 1; //@line 907 "fastmix.cpp"
  $31 = (1209506 + ($30<<1)|0); //@line 907 "fastmix.cpp"
  $32 = HEAP16[$31>>1]|0; //@line 907 "fastmix.cpp"
  $33 = $32 << 16 >> 16; //@line 907 "fastmix.cpp"
  $$sum1 = (($25) + ($4))|0; //@line 907 "fastmix.cpp"
  $34 = (($$) + ($$sum1)|0); //@line 907 "fastmix.cpp"
  $35 = HEAP8[$34>>0]|0; //@line 907 "fastmix.cpp"
  $36 = $35 << 24 >> 24; //@line 907 "fastmix.cpp"
  $37 = Math_imul($36, $33)|0; //@line 907 "fastmix.cpp"
  $38 = (($37) + ($29))|0; //@line 907 "fastmix.cpp"
  $39 = $21 | 2; //@line 907 "fastmix.cpp"
  $40 = (1209506 + ($39<<1)|0); //@line 907 "fastmix.cpp"
  $41 = HEAP16[$40>>1]|0; //@line 907 "fastmix.cpp"
  $42 = $41 << 16 >> 16; //@line 907 "fastmix.cpp"
  $$sum2 = (($11) + ($25))|0; //@line 907 "fastmix.cpp"
  $43 = (($$) + ($$sum2)|0); //@line 907 "fastmix.cpp"
  $44 = HEAP8[$43>>0]|0; //@line 907 "fastmix.cpp"
  $45 = $44 << 24 >> 24; //@line 907 "fastmix.cpp"
  $46 = Math_imul($45, $42)|0; //@line 907 "fastmix.cpp"
  $47 = (($38) + ($46))|0; //@line 907 "fastmix.cpp"
  $48 = $21 | 3; //@line 907 "fastmix.cpp"
  $49 = (1209506 + ($48<<1)|0); //@line 907 "fastmix.cpp"
  $50 = HEAP16[$49>>1]|0; //@line 907 "fastmix.cpp"
  $51 = $50 << 16 >> 16; //@line 907 "fastmix.cpp"
  $$sum3 = (($12) + ($25))|0; //@line 907 "fastmix.cpp"
  $52 = (($$) + ($$sum3)|0); //@line 907 "fastmix.cpp"
  $53 = HEAP8[$52>>0]|0; //@line 907 "fastmix.cpp"
  $54 = $53 << 24 >> 24; //@line 907 "fastmix.cpp"
  $55 = Math_imul($54, $51)|0; //@line 907 "fastmix.cpp"
  $56 = (($47) + ($55))|0; //@line 907 "fastmix.cpp"
  $57 = $56 >> 6; //@line 907 "fastmix.cpp"
  $$sum4 = (($13) + ($25))|0; //@line 907 "fastmix.cpp"
  $58 = (($$) + ($$sum4)|0); //@line 907 "fastmix.cpp"
  $59 = HEAP8[$58>>0]|0; //@line 907 "fastmix.cpp"
  $60 = $59 << 24 >> 24; //@line 907 "fastmix.cpp"
  $61 = Math_imul($60, $24)|0; //@line 907 "fastmix.cpp"
  $62 = $25 | 1; //@line 907 "fastmix.cpp"
  $$sum5 = (($62) + ($4))|0; //@line 907 "fastmix.cpp"
  $63 = (($$) + ($$sum5)|0); //@line 907 "fastmix.cpp"
  $64 = HEAP8[$63>>0]|0; //@line 907 "fastmix.cpp"
  $65 = $64 << 24 >> 24; //@line 907 "fastmix.cpp"
  $66 = Math_imul($65, $33)|0; //@line 907 "fastmix.cpp"
  $67 = (($66) + ($61))|0; //@line 907 "fastmix.cpp"
  $$sum6 = (($14) + ($25))|0; //@line 907 "fastmix.cpp"
  $68 = (($$) + ($$sum6)|0); //@line 907 "fastmix.cpp"
  $69 = HEAP8[$68>>0]|0; //@line 907 "fastmix.cpp"
  $70 = $69 << 24 >> 24; //@line 907 "fastmix.cpp"
  $71 = Math_imul($70, $42)|0; //@line 907 "fastmix.cpp"
  $72 = (($67) + ($71))|0; //@line 907 "fastmix.cpp"
  $$sum7 = (($15) + ($25))|0; //@line 907 "fastmix.cpp"
  $73 = (($$) + ($$sum7)|0); //@line 907 "fastmix.cpp"
  $74 = HEAP8[$73>>0]|0; //@line 907 "fastmix.cpp"
  $75 = $74 << 24 >> 24; //@line 907 "fastmix.cpp"
  $76 = Math_imul($75, $51)|0; //@line 907 "fastmix.cpp"
  $77 = (($72) + ($76))|0; //@line 907 "fastmix.cpp"
  $78 = $77 >> 6; //@line 907 "fastmix.cpp"
  $79 = HEAP32[$16>>2]|0; //@line 908 "fastmix.cpp"
  $80 = Math_imul($79, $57)|0; //@line 908 "fastmix.cpp"
  $81 = HEAP32[$pvol$0>>2]|0; //@line 908 "fastmix.cpp"
  $82 = (($81) + ($80))|0; //@line 908 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $82; //@line 908 "fastmix.cpp"
  $83 = HEAP32[$17>>2]|0; //@line 908 "fastmix.cpp"
  $84 = Math_imul($78, $83)|0; //@line 908 "fastmix.cpp"
  $85 = ((($pvol$0)) + 4|0); //@line 908 "fastmix.cpp"
  $86 = HEAP32[$85>>2]|0; //@line 908 "fastmix.cpp"
  $87 = (($84) + ($86))|0; //@line 908 "fastmix.cpp"
  HEAP32[$85>>2] = $87; //@line 908 "fastmix.cpp"
  $88 = ((($pvol$0)) + 8|0); //@line 908 "fastmix.cpp"
  $89 = HEAP32[$18>>2]|0; //@line 909 "fastmix.cpp"
  $90 = (($89) + ($nPos$0))|0; //@line 909 "fastmix.cpp"
  $91 = ($88>>>0)<($pbufmax>>>0); //@line 909 "fastmix.cpp"
  if ($91) {
   $nPos$0 = $90;$pvol$0 = $88;
  } else {
   $$lcssa = $90;
   break;
  }
 }
 $92 = $$lcssa >> 16; //@line 909 "fastmix.cpp"
 $93 = HEAP32[$3>>2]|0; //@line 909 "fastmix.cpp"
 $94 = (($93) + ($92))|0; //@line 909 "fastmix.cpp"
 HEAP32[$3>>2] = $94; //@line 909 "fastmix.cpp"
 $95 = $$lcssa & 65535; //@line 909 "fastmix.cpp"
 HEAP32[$0>>2] = $95; //@line 909 "fastmix.cpp"
 return; //@line 909 "fastmix.cpp"
}
function __Z20Stereo16BitSplineMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 912 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 912 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 912 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 912 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 912 "fastmix.cpp"
 $5 = $4 << 1; //@line 912 "fastmix.cpp"
 $6 = (($2) + ($5)|0); //@line 912 "fastmix.cpp"
 $7 = ((($pChannel)) + 36|0); //@line 912 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 912 "fastmix.cpp"
 $9 = $8 & 64; //@line 912 "fastmix.cpp"
 $10 = ($9|0)==(0); //@line 912 "fastmix.cpp"
 $11 = (($6) + ($4<<1)|0); //@line 912 "fastmix.cpp"
 $$ = $10 ? $6 : $11; //@line 912 "fastmix.cpp"
 $12 = ((($pChannel)) + 16|0); //@line 914 "fastmix.cpp"
 $13 = ((($pChannel)) + 20|0); //@line 914 "fastmix.cpp"
 $14 = ((($pChannel)) + 12|0); //@line 915 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $15 = $nPos$0 >> 16; //@line 913 "fastmix.cpp"
  $16 = $nPos$0 >>> 4; //@line 913 "fastmix.cpp"
  $17 = $16 & 4092; //@line 913 "fastmix.cpp"
  $18 = (1209506 + ($17<<1)|0); //@line 913 "fastmix.cpp"
  $19 = HEAP16[$18>>1]|0; //@line 913 "fastmix.cpp"
  $20 = $19 << 16 >> 16; //@line 913 "fastmix.cpp"
  $21 = $15 << 1; //@line 913 "fastmix.cpp"
  $22 = (($21) + -2)|0; //@line 913 "fastmix.cpp"
  $23 = (($$) + ($22<<1)|0); //@line 913 "fastmix.cpp"
  $24 = HEAP16[$23>>1]|0; //@line 913 "fastmix.cpp"
  $25 = $24 << 16 >> 16; //@line 913 "fastmix.cpp"
  $26 = Math_imul($25, $20)|0; //@line 913 "fastmix.cpp"
  $27 = $17 | 1; //@line 913 "fastmix.cpp"
  $28 = (1209506 + ($27<<1)|0); //@line 913 "fastmix.cpp"
  $29 = HEAP16[$28>>1]|0; //@line 913 "fastmix.cpp"
  $30 = $29 << 16 >> 16; //@line 913 "fastmix.cpp"
  $31 = (($$) + ($21<<1)|0); //@line 913 "fastmix.cpp"
  $32 = HEAP16[$31>>1]|0; //@line 913 "fastmix.cpp"
  $33 = $32 << 16 >> 16; //@line 913 "fastmix.cpp"
  $34 = Math_imul($33, $30)|0; //@line 913 "fastmix.cpp"
  $35 = (($34) + ($26))|0; //@line 913 "fastmix.cpp"
  $36 = $17 | 2; //@line 913 "fastmix.cpp"
  $37 = (1209506 + ($36<<1)|0); //@line 913 "fastmix.cpp"
  $38 = HEAP16[$37>>1]|0; //@line 913 "fastmix.cpp"
  $39 = $38 << 16 >> 16; //@line 913 "fastmix.cpp"
  $40 = (($21) + 2)|0; //@line 913 "fastmix.cpp"
  $41 = (($$) + ($40<<1)|0); //@line 913 "fastmix.cpp"
  $42 = HEAP16[$41>>1]|0; //@line 913 "fastmix.cpp"
  $43 = $42 << 16 >> 16; //@line 913 "fastmix.cpp"
  $44 = Math_imul($43, $39)|0; //@line 913 "fastmix.cpp"
  $45 = (($35) + ($44))|0; //@line 913 "fastmix.cpp"
  $46 = $17 | 3; //@line 913 "fastmix.cpp"
  $47 = (1209506 + ($46<<1)|0); //@line 913 "fastmix.cpp"
  $48 = HEAP16[$47>>1]|0; //@line 913 "fastmix.cpp"
  $49 = $48 << 16 >> 16; //@line 913 "fastmix.cpp"
  $50 = (($21) + 4)|0; //@line 913 "fastmix.cpp"
  $51 = (($$) + ($50<<1)|0); //@line 913 "fastmix.cpp"
  $52 = HEAP16[$51>>1]|0; //@line 913 "fastmix.cpp"
  $53 = $52 << 16 >> 16; //@line 913 "fastmix.cpp"
  $54 = Math_imul($53, $49)|0; //@line 913 "fastmix.cpp"
  $55 = (($45) + ($54))|0; //@line 913 "fastmix.cpp"
  $56 = $55 >> 14; //@line 913 "fastmix.cpp"
  $57 = (($21) + -1)|0; //@line 913 "fastmix.cpp"
  $58 = (($$) + ($57<<1)|0); //@line 913 "fastmix.cpp"
  $59 = HEAP16[$58>>1]|0; //@line 913 "fastmix.cpp"
  $60 = $59 << 16 >> 16; //@line 913 "fastmix.cpp"
  $61 = Math_imul($60, $20)|0; //@line 913 "fastmix.cpp"
  $62 = $21 | 1; //@line 913 "fastmix.cpp"
  $63 = (($$) + ($62<<1)|0); //@line 913 "fastmix.cpp"
  $64 = HEAP16[$63>>1]|0; //@line 913 "fastmix.cpp"
  $65 = $64 << 16 >> 16; //@line 913 "fastmix.cpp"
  $66 = Math_imul($65, $30)|0; //@line 913 "fastmix.cpp"
  $67 = (($66) + ($61))|0; //@line 913 "fastmix.cpp"
  $68 = (($21) + 3)|0; //@line 913 "fastmix.cpp"
  $69 = (($$) + ($68<<1)|0); //@line 913 "fastmix.cpp"
  $70 = HEAP16[$69>>1]|0; //@line 913 "fastmix.cpp"
  $71 = $70 << 16 >> 16; //@line 913 "fastmix.cpp"
  $72 = Math_imul($71, $39)|0; //@line 913 "fastmix.cpp"
  $73 = (($67) + ($72))|0; //@line 913 "fastmix.cpp"
  $74 = (($21) + 5)|0; //@line 913 "fastmix.cpp"
  $75 = (($$) + ($74<<1)|0); //@line 913 "fastmix.cpp"
  $76 = HEAP16[$75>>1]|0; //@line 913 "fastmix.cpp"
  $77 = $76 << 16 >> 16; //@line 913 "fastmix.cpp"
  $78 = Math_imul($77, $49)|0; //@line 913 "fastmix.cpp"
  $79 = (($73) + ($78))|0; //@line 913 "fastmix.cpp"
  $80 = $79 >> 14; //@line 913 "fastmix.cpp"
  $81 = HEAP32[$12>>2]|0; //@line 914 "fastmix.cpp"
  $82 = Math_imul($81, $56)|0; //@line 914 "fastmix.cpp"
  $83 = HEAP32[$pvol$0>>2]|0; //@line 914 "fastmix.cpp"
  $84 = (($83) + ($82))|0; //@line 914 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $84; //@line 914 "fastmix.cpp"
  $85 = HEAP32[$13>>2]|0; //@line 914 "fastmix.cpp"
  $86 = Math_imul($80, $85)|0; //@line 914 "fastmix.cpp"
  $87 = ((($pvol$0)) + 4|0); //@line 914 "fastmix.cpp"
  $88 = HEAP32[$87>>2]|0; //@line 914 "fastmix.cpp"
  $89 = (($86) + ($88))|0; //@line 914 "fastmix.cpp"
  HEAP32[$87>>2] = $89; //@line 914 "fastmix.cpp"
  $90 = ((($pvol$0)) + 8|0); //@line 914 "fastmix.cpp"
  $91 = HEAP32[$14>>2]|0; //@line 915 "fastmix.cpp"
  $92 = (($91) + ($nPos$0))|0; //@line 915 "fastmix.cpp"
  $93 = ($90>>>0)<($pbufmax>>>0); //@line 915 "fastmix.cpp"
  if ($93) {
   $nPos$0 = $92;$pvol$0 = $90;
  } else {
   $$lcssa = $92;
   break;
  }
 }
 $94 = $$lcssa >> 16; //@line 915 "fastmix.cpp"
 $95 = HEAP32[$3>>2]|0; //@line 915 "fastmix.cpp"
 $96 = (($95) + ($94))|0; //@line 915 "fastmix.cpp"
 HEAP32[$3>>2] = $96; //@line 915 "fastmix.cpp"
 $97 = $$lcssa & 65535; //@line 915 "fastmix.cpp"
 HEAP32[$0>>2] = $97; //@line 915 "fastmix.cpp"
 return; //@line 915 "fastmix.cpp"
}
function __Z21Mono8BitSplineRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa28 = 0, $$lcssa29 = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$sum = 0, $$sum1 = 0, $$sum2 = 0, $$sum3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $9 = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 751 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 751 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 751 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 751 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 752 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 752 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 752 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 752 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 752 "fastmix.cpp"
 $9 = (($6) + ($8)|0); //@line 752 "fastmix.cpp"
 $10 = ((($pChannel)) + 36|0); //@line 752 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 752 "fastmix.cpp"
 $12 = $11 & 64; //@line 752 "fastmix.cpp"
 $13 = ($12|0)==(0); //@line 752 "fastmix.cpp"
 $$ = $13 ? $6 : $9; //@line 752 "fastmix.cpp"
 $14 = (($8) + -1)|0; //@line 753 "fastmix.cpp"
 $15 = (($8) + 2)|0; //@line 753 "fastmix.cpp"
 $16 = (($8) + 1)|0; //@line 753 "fastmix.cpp"
 $17 = ((($pChannel)) + 28|0); //@line 754 "fastmix.cpp"
 $18 = ((($pChannel)) + 24|0); //@line 754 "fastmix.cpp"
 $19 = ((($pChannel)) + 12|0); //@line 755 "fastmix.cpp"
 $nPos$0 = $5;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $20 = $nPos$0 >> 16; //@line 753 "fastmix.cpp"
  $21 = $nPos$0 >>> 4; //@line 753 "fastmix.cpp"
  $22 = $21 & 4092; //@line 753 "fastmix.cpp"
  $23 = (1209506 + ($22<<1)|0); //@line 753 "fastmix.cpp"
  $24 = HEAP16[$23>>1]|0; //@line 753 "fastmix.cpp"
  $25 = $24 << 16 >> 16; //@line 753 "fastmix.cpp"
  $$sum = (($14) + ($20))|0; //@line 753 "fastmix.cpp"
  $26 = (($$) + ($$sum)|0); //@line 753 "fastmix.cpp"
  $27 = HEAP8[$26>>0]|0; //@line 753 "fastmix.cpp"
  $28 = $27 << 24 >> 24; //@line 753 "fastmix.cpp"
  $29 = Math_imul($28, $25)|0; //@line 753 "fastmix.cpp"
  $30 = $22 | 1; //@line 753 "fastmix.cpp"
  $31 = (1209506 + ($30<<1)|0); //@line 753 "fastmix.cpp"
  $32 = HEAP16[$31>>1]|0; //@line 753 "fastmix.cpp"
  $33 = $32 << 16 >> 16; //@line 753 "fastmix.cpp"
  $$sum1 = (($20) + ($8))|0; //@line 753 "fastmix.cpp"
  $34 = (($$) + ($$sum1)|0); //@line 753 "fastmix.cpp"
  $35 = HEAP8[$34>>0]|0; //@line 753 "fastmix.cpp"
  $36 = $35 << 24 >> 24; //@line 753 "fastmix.cpp"
  $37 = Math_imul($36, $33)|0; //@line 753 "fastmix.cpp"
  $38 = (($37) + ($29))|0; //@line 753 "fastmix.cpp"
  $39 = $22 | 3; //@line 753 "fastmix.cpp"
  $40 = (1209506 + ($39<<1)|0); //@line 753 "fastmix.cpp"
  $41 = HEAP16[$40>>1]|0; //@line 753 "fastmix.cpp"
  $42 = $41 << 16 >> 16; //@line 753 "fastmix.cpp"
  $$sum2 = (($15) + ($20))|0; //@line 753 "fastmix.cpp"
  $43 = (($$) + ($$sum2)|0); //@line 753 "fastmix.cpp"
  $44 = HEAP8[$43>>0]|0; //@line 753 "fastmix.cpp"
  $45 = $44 << 24 >> 24; //@line 753 "fastmix.cpp"
  $46 = Math_imul($45, $42)|0; //@line 753 "fastmix.cpp"
  $47 = (($38) + ($46))|0; //@line 753 "fastmix.cpp"
  $48 = $22 | 2; //@line 753 "fastmix.cpp"
  $49 = (1209506 + ($48<<1)|0); //@line 753 "fastmix.cpp"
  $50 = HEAP16[$49>>1]|0; //@line 753 "fastmix.cpp"
  $51 = $50 << 16 >> 16; //@line 753 "fastmix.cpp"
  $$sum3 = (($16) + ($20))|0; //@line 753 "fastmix.cpp"
  $52 = (($$) + ($$sum3)|0); //@line 753 "fastmix.cpp"
  $53 = HEAP8[$52>>0]|0; //@line 753 "fastmix.cpp"
  $54 = $53 << 24 >> 24; //@line 753 "fastmix.cpp"
  $55 = Math_imul($54, $51)|0; //@line 753 "fastmix.cpp"
  $56 = (($47) + ($55))|0; //@line 753 "fastmix.cpp"
  $57 = $56 >> 6; //@line 753 "fastmix.cpp"
  $58 = HEAP32[$17>>2]|0; //@line 754 "fastmix.cpp"
  $59 = (($58) + ($nRampLeftVol$0))|0; //@line 754 "fastmix.cpp"
  $60 = HEAP32[$18>>2]|0; //@line 754 "fastmix.cpp"
  $61 = (($60) + ($nRampRightVol$0))|0; //@line 754 "fastmix.cpp"
  $62 = $61 >> 12; //@line 754 "fastmix.cpp"
  $63 = Math_imul($57, $62)|0; //@line 754 "fastmix.cpp"
  $64 = HEAP32[$pvol$0>>2]|0; //@line 754 "fastmix.cpp"
  $65 = (($63) + ($64))|0; //@line 754 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $65; //@line 754 "fastmix.cpp"
  $66 = $59 >> 12; //@line 754 "fastmix.cpp"
  $67 = Math_imul($57, $66)|0; //@line 754 "fastmix.cpp"
  $68 = ((($pvol$0)) + 4|0); //@line 754 "fastmix.cpp"
  $69 = HEAP32[$68>>2]|0; //@line 754 "fastmix.cpp"
  $70 = (($67) + ($69))|0; //@line 754 "fastmix.cpp"
  HEAP32[$68>>2] = $70; //@line 754 "fastmix.cpp"
  $71 = ((($pvol$0)) + 8|0); //@line 754 "fastmix.cpp"
  $72 = HEAP32[$19>>2]|0; //@line 755 "fastmix.cpp"
  $73 = (($72) + ($nPos$0))|0; //@line 755 "fastmix.cpp"
  $74 = ($71>>>0)<($pbufmax>>>0); //@line 755 "fastmix.cpp"
  if ($74) {
   $nPos$0 = $73;$nRampLeftVol$0 = $59;$nRampRightVol$0 = $61;$pvol$0 = $71;
  } else {
   $$lcssa = $59;$$lcssa28 = $61;$$lcssa29 = $62;$$lcssa30 = $66;$$lcssa31 = $73;
   break;
  }
 }
 $75 = $$lcssa31 >> 16; //@line 755 "fastmix.cpp"
 $76 = HEAP32[$7>>2]|0; //@line 755 "fastmix.cpp"
 $77 = (($76) + ($75))|0; //@line 755 "fastmix.cpp"
 HEAP32[$7>>2] = $77; //@line 755 "fastmix.cpp"
 $78 = $$lcssa31 & 65535; //@line 755 "fastmix.cpp"
 HEAP32[$4>>2] = $78; //@line 755 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa28; //@line 755 "fastmix.cpp"
 $79 = ((($pChannel)) + 16|0); //@line 755 "fastmix.cpp"
 HEAP32[$79>>2] = $$lcssa29; //@line 755 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa; //@line 755 "fastmix.cpp"
 $80 = ((($pChannel)) + 20|0); //@line 755 "fastmix.cpp"
 HEAP32[$80>>2] = $$lcssa30; //@line 755 "fastmix.cpp"
 return; //@line 755 "fastmix.cpp"
}
function __Z22Mono16BitSplineRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa21 = 0, $$lcssa22 = 0, $$lcssa23 = 0, $$lcssa24 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $9 = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 757 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 757 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 757 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 757 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 758 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 758 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 758 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 758 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 758 "fastmix.cpp"
 $9 = $8 << 1; //@line 758 "fastmix.cpp"
 $10 = (($6) + ($9)|0); //@line 758 "fastmix.cpp"
 $11 = ((($pChannel)) + 36|0); //@line 758 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 758 "fastmix.cpp"
 $13 = $12 & 64; //@line 758 "fastmix.cpp"
 $14 = ($13|0)==(0); //@line 758 "fastmix.cpp"
 $15 = (($10) + ($8<<1)|0); //@line 758 "fastmix.cpp"
 $$ = $14 ? $10 : $15; //@line 758 "fastmix.cpp"
 $16 = ((($pChannel)) + 28|0); //@line 760 "fastmix.cpp"
 $17 = ((($pChannel)) + 24|0); //@line 760 "fastmix.cpp"
 $18 = ((($pChannel)) + 12|0); //@line 761 "fastmix.cpp"
 $nPos$0 = $5;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $19 = $nPos$0 >> 16; //@line 759 "fastmix.cpp"
  $20 = $nPos$0 >>> 4; //@line 759 "fastmix.cpp"
  $21 = $20 & 4092; //@line 759 "fastmix.cpp"
  $22 = (1209506 + ($21<<1)|0); //@line 759 "fastmix.cpp"
  $23 = HEAP16[$22>>1]|0; //@line 759 "fastmix.cpp"
  $24 = $23 << 16 >> 16; //@line 759 "fastmix.cpp"
  $25 = (($19) + -1)|0; //@line 759 "fastmix.cpp"
  $26 = (($$) + ($25<<1)|0); //@line 759 "fastmix.cpp"
  $27 = HEAP16[$26>>1]|0; //@line 759 "fastmix.cpp"
  $28 = $27 << 16 >> 16; //@line 759 "fastmix.cpp"
  $29 = Math_imul($28, $24)|0; //@line 759 "fastmix.cpp"
  $30 = $21 | 1; //@line 759 "fastmix.cpp"
  $31 = (1209506 + ($30<<1)|0); //@line 759 "fastmix.cpp"
  $32 = HEAP16[$31>>1]|0; //@line 759 "fastmix.cpp"
  $33 = $32 << 16 >> 16; //@line 759 "fastmix.cpp"
  $34 = (($$) + ($19<<1)|0); //@line 759 "fastmix.cpp"
  $35 = HEAP16[$34>>1]|0; //@line 759 "fastmix.cpp"
  $36 = $35 << 16 >> 16; //@line 759 "fastmix.cpp"
  $37 = Math_imul($36, $33)|0; //@line 759 "fastmix.cpp"
  $38 = (($37) + ($29))|0; //@line 759 "fastmix.cpp"
  $39 = $21 | 3; //@line 759 "fastmix.cpp"
  $40 = (1209506 + ($39<<1)|0); //@line 759 "fastmix.cpp"
  $41 = HEAP16[$40>>1]|0; //@line 759 "fastmix.cpp"
  $42 = $41 << 16 >> 16; //@line 759 "fastmix.cpp"
  $43 = (($19) + 2)|0; //@line 759 "fastmix.cpp"
  $44 = (($$) + ($43<<1)|0); //@line 759 "fastmix.cpp"
  $45 = HEAP16[$44>>1]|0; //@line 759 "fastmix.cpp"
  $46 = $45 << 16 >> 16; //@line 759 "fastmix.cpp"
  $47 = Math_imul($46, $42)|0; //@line 759 "fastmix.cpp"
  $48 = (($38) + ($47))|0; //@line 759 "fastmix.cpp"
  $49 = $21 | 2; //@line 759 "fastmix.cpp"
  $50 = (1209506 + ($49<<1)|0); //@line 759 "fastmix.cpp"
  $51 = HEAP16[$50>>1]|0; //@line 759 "fastmix.cpp"
  $52 = $51 << 16 >> 16; //@line 759 "fastmix.cpp"
  $53 = (($19) + 1)|0; //@line 759 "fastmix.cpp"
  $54 = (($$) + ($53<<1)|0); //@line 759 "fastmix.cpp"
  $55 = HEAP16[$54>>1]|0; //@line 759 "fastmix.cpp"
  $56 = $55 << 16 >> 16; //@line 759 "fastmix.cpp"
  $57 = Math_imul($56, $52)|0; //@line 759 "fastmix.cpp"
  $58 = (($48) + ($57))|0; //@line 759 "fastmix.cpp"
  $59 = $58 >> 14; //@line 759 "fastmix.cpp"
  $60 = HEAP32[$16>>2]|0; //@line 760 "fastmix.cpp"
  $61 = (($60) + ($nRampLeftVol$0))|0; //@line 760 "fastmix.cpp"
  $62 = HEAP32[$17>>2]|0; //@line 760 "fastmix.cpp"
  $63 = (($62) + ($nRampRightVol$0))|0; //@line 760 "fastmix.cpp"
  $64 = $63 >> 12; //@line 760 "fastmix.cpp"
  $65 = Math_imul($59, $64)|0; //@line 760 "fastmix.cpp"
  $66 = HEAP32[$pvol$0>>2]|0; //@line 760 "fastmix.cpp"
  $67 = (($65) + ($66))|0; //@line 760 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $67; //@line 760 "fastmix.cpp"
  $68 = $61 >> 12; //@line 760 "fastmix.cpp"
  $69 = Math_imul($59, $68)|0; //@line 760 "fastmix.cpp"
  $70 = ((($pvol$0)) + 4|0); //@line 760 "fastmix.cpp"
  $71 = HEAP32[$70>>2]|0; //@line 760 "fastmix.cpp"
  $72 = (($69) + ($71))|0; //@line 760 "fastmix.cpp"
  HEAP32[$70>>2] = $72; //@line 760 "fastmix.cpp"
  $73 = ((($pvol$0)) + 8|0); //@line 760 "fastmix.cpp"
  $74 = HEAP32[$18>>2]|0; //@line 761 "fastmix.cpp"
  $75 = (($74) + ($nPos$0))|0; //@line 761 "fastmix.cpp"
  $76 = ($73>>>0)<($pbufmax>>>0); //@line 761 "fastmix.cpp"
  if ($76) {
   $nPos$0 = $75;$nRampLeftVol$0 = $61;$nRampRightVol$0 = $63;$pvol$0 = $73;
  } else {
   $$lcssa = $61;$$lcssa21 = $63;$$lcssa22 = $64;$$lcssa23 = $68;$$lcssa24 = $75;
   break;
  }
 }
 $77 = $$lcssa24 >> 16; //@line 761 "fastmix.cpp"
 $78 = HEAP32[$7>>2]|0; //@line 761 "fastmix.cpp"
 $79 = (($78) + ($77))|0; //@line 761 "fastmix.cpp"
 HEAP32[$7>>2] = $79; //@line 761 "fastmix.cpp"
 $80 = $$lcssa24 & 65535; //@line 761 "fastmix.cpp"
 HEAP32[$4>>2] = $80; //@line 761 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa21; //@line 761 "fastmix.cpp"
 $81 = ((($pChannel)) + 16|0); //@line 761 "fastmix.cpp"
 HEAP32[$81>>2] = $$lcssa22; //@line 761 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa; //@line 761 "fastmix.cpp"
 $82 = ((($pChannel)) + 20|0); //@line 761 "fastmix.cpp"
 HEAP32[$82>>2] = $$lcssa23; //@line 761 "fastmix.cpp"
 return; //@line 761 "fastmix.cpp"
}
function __Z23Stereo8BitSplineRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa36 = 0, $$lcssa37 = 0, $$lcssa38 = 0, $$lcssa39 = 0, $$sum = 0, $$sum1 = 0, $$sum2 = 0, $$sum3 = 0, $$sum4 = 0, $$sum5 = 0, $$sum6 = 0, $$sum7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 955 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 955 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 955 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 955 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 956 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 956 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 956 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 956 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 956 "fastmix.cpp"
 $9 = (($6) + ($8)|0); //@line 956 "fastmix.cpp"
 $10 = ((($pChannel)) + 36|0); //@line 956 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 956 "fastmix.cpp"
 $12 = $11 & 64; //@line 956 "fastmix.cpp"
 $13 = ($12|0)==(0); //@line 956 "fastmix.cpp"
 $$ = $13 ? $6 : $9; //@line 956 "fastmix.cpp"
 $14 = (($8) + -2)|0; //@line 957 "fastmix.cpp"
 $15 = (($8) + 2)|0; //@line 957 "fastmix.cpp"
 $16 = (($8) + 4)|0; //@line 957 "fastmix.cpp"
 $17 = (($8) + -1)|0; //@line 957 "fastmix.cpp"
 $18 = (($8) + 3)|0; //@line 957 "fastmix.cpp"
 $19 = (($8) + 5)|0; //@line 957 "fastmix.cpp"
 $20 = ((($pChannel)) + 28|0); //@line 958 "fastmix.cpp"
 $21 = ((($pChannel)) + 24|0); //@line 958 "fastmix.cpp"
 $22 = ((($pChannel)) + 12|0); //@line 959 "fastmix.cpp"
 $nPos$0 = $5;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $23 = $nPos$0 >> 16; //@line 957 "fastmix.cpp"
  $24 = $nPos$0 >>> 4; //@line 957 "fastmix.cpp"
  $25 = $24 & 4092; //@line 957 "fastmix.cpp"
  $26 = (1209506 + ($25<<1)|0); //@line 957 "fastmix.cpp"
  $27 = HEAP16[$26>>1]|0; //@line 957 "fastmix.cpp"
  $28 = $27 << 16 >> 16; //@line 957 "fastmix.cpp"
  $29 = $23 << 1; //@line 957 "fastmix.cpp"
  $$sum = (($14) + ($29))|0; //@line 957 "fastmix.cpp"
  $30 = (($$) + ($$sum)|0); //@line 957 "fastmix.cpp"
  $31 = HEAP8[$30>>0]|0; //@line 957 "fastmix.cpp"
  $32 = $31 << 24 >> 24; //@line 957 "fastmix.cpp"
  $33 = Math_imul($32, $28)|0; //@line 957 "fastmix.cpp"
  $34 = $25 | 1; //@line 957 "fastmix.cpp"
  $35 = (1209506 + ($34<<1)|0); //@line 957 "fastmix.cpp"
  $36 = HEAP16[$35>>1]|0; //@line 957 "fastmix.cpp"
  $37 = $36 << 16 >> 16; //@line 957 "fastmix.cpp"
  $$sum1 = (($29) + ($8))|0; //@line 957 "fastmix.cpp"
  $38 = (($$) + ($$sum1)|0); //@line 957 "fastmix.cpp"
  $39 = HEAP8[$38>>0]|0; //@line 957 "fastmix.cpp"
  $40 = $39 << 24 >> 24; //@line 957 "fastmix.cpp"
  $41 = Math_imul($40, $37)|0; //@line 957 "fastmix.cpp"
  $42 = (($41) + ($33))|0; //@line 957 "fastmix.cpp"
  $43 = $25 | 2; //@line 957 "fastmix.cpp"
  $44 = (1209506 + ($43<<1)|0); //@line 957 "fastmix.cpp"
  $45 = HEAP16[$44>>1]|0; //@line 957 "fastmix.cpp"
  $46 = $45 << 16 >> 16; //@line 957 "fastmix.cpp"
  $$sum2 = (($15) + ($29))|0; //@line 957 "fastmix.cpp"
  $47 = (($$) + ($$sum2)|0); //@line 957 "fastmix.cpp"
  $48 = HEAP8[$47>>0]|0; //@line 957 "fastmix.cpp"
  $49 = $48 << 24 >> 24; //@line 957 "fastmix.cpp"
  $50 = Math_imul($49, $46)|0; //@line 957 "fastmix.cpp"
  $51 = (($42) + ($50))|0; //@line 957 "fastmix.cpp"
  $52 = $25 | 3; //@line 957 "fastmix.cpp"
  $53 = (1209506 + ($52<<1)|0); //@line 957 "fastmix.cpp"
  $54 = HEAP16[$53>>1]|0; //@line 957 "fastmix.cpp"
  $55 = $54 << 16 >> 16; //@line 957 "fastmix.cpp"
  $$sum3 = (($16) + ($29))|0; //@line 957 "fastmix.cpp"
  $56 = (($$) + ($$sum3)|0); //@line 957 "fastmix.cpp"
  $57 = HEAP8[$56>>0]|0; //@line 957 "fastmix.cpp"
  $58 = $57 << 24 >> 24; //@line 957 "fastmix.cpp"
  $59 = Math_imul($58, $55)|0; //@line 957 "fastmix.cpp"
  $60 = (($51) + ($59))|0; //@line 957 "fastmix.cpp"
  $61 = $60 >> 6; //@line 957 "fastmix.cpp"
  $$sum4 = (($17) + ($29))|0; //@line 957 "fastmix.cpp"
  $62 = (($$) + ($$sum4)|0); //@line 957 "fastmix.cpp"
  $63 = HEAP8[$62>>0]|0; //@line 957 "fastmix.cpp"
  $64 = $63 << 24 >> 24; //@line 957 "fastmix.cpp"
  $65 = Math_imul($64, $28)|0; //@line 957 "fastmix.cpp"
  $66 = $29 | 1; //@line 957 "fastmix.cpp"
  $$sum5 = (($66) + ($8))|0; //@line 957 "fastmix.cpp"
  $67 = (($$) + ($$sum5)|0); //@line 957 "fastmix.cpp"
  $68 = HEAP8[$67>>0]|0; //@line 957 "fastmix.cpp"
  $69 = $68 << 24 >> 24; //@line 957 "fastmix.cpp"
  $70 = Math_imul($69, $37)|0; //@line 957 "fastmix.cpp"
  $71 = (($70) + ($65))|0; //@line 957 "fastmix.cpp"
  $$sum6 = (($18) + ($29))|0; //@line 957 "fastmix.cpp"
  $72 = (($$) + ($$sum6)|0); //@line 957 "fastmix.cpp"
  $73 = HEAP8[$72>>0]|0; //@line 957 "fastmix.cpp"
  $74 = $73 << 24 >> 24; //@line 957 "fastmix.cpp"
  $75 = Math_imul($74, $46)|0; //@line 957 "fastmix.cpp"
  $76 = (($71) + ($75))|0; //@line 957 "fastmix.cpp"
  $$sum7 = (($19) + ($29))|0; //@line 957 "fastmix.cpp"
  $77 = (($$) + ($$sum7)|0); //@line 957 "fastmix.cpp"
  $78 = HEAP8[$77>>0]|0; //@line 957 "fastmix.cpp"
  $79 = $78 << 24 >> 24; //@line 957 "fastmix.cpp"
  $80 = Math_imul($79, $55)|0; //@line 957 "fastmix.cpp"
  $81 = (($76) + ($80))|0; //@line 957 "fastmix.cpp"
  $82 = $81 >> 6; //@line 957 "fastmix.cpp"
  $83 = HEAP32[$20>>2]|0; //@line 958 "fastmix.cpp"
  $84 = (($83) + ($nRampLeftVol$0))|0; //@line 958 "fastmix.cpp"
  $85 = HEAP32[$21>>2]|0; //@line 958 "fastmix.cpp"
  $86 = (($85) + ($nRampRightVol$0))|0; //@line 958 "fastmix.cpp"
  $87 = $86 >> 12; //@line 958 "fastmix.cpp"
  $88 = Math_imul($87, $61)|0; //@line 958 "fastmix.cpp"
  $89 = HEAP32[$pvol$0>>2]|0; //@line 958 "fastmix.cpp"
  $90 = (($88) + ($89))|0; //@line 958 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $90; //@line 958 "fastmix.cpp"
  $91 = $84 >> 12; //@line 958 "fastmix.cpp"
  $92 = Math_imul($82, $91)|0; //@line 958 "fastmix.cpp"
  $93 = ((($pvol$0)) + 4|0); //@line 958 "fastmix.cpp"
  $94 = HEAP32[$93>>2]|0; //@line 958 "fastmix.cpp"
  $95 = (($92) + ($94))|0; //@line 958 "fastmix.cpp"
  HEAP32[$93>>2] = $95; //@line 958 "fastmix.cpp"
  $96 = ((($pvol$0)) + 8|0); //@line 958 "fastmix.cpp"
  $97 = HEAP32[$22>>2]|0; //@line 959 "fastmix.cpp"
  $98 = (($97) + ($nPos$0))|0; //@line 959 "fastmix.cpp"
  $99 = ($96>>>0)<($pbufmax>>>0); //@line 959 "fastmix.cpp"
  if ($99) {
   $nPos$0 = $98;$nRampLeftVol$0 = $84;$nRampRightVol$0 = $86;$pvol$0 = $96;
  } else {
   $$lcssa = $84;$$lcssa36 = $86;$$lcssa37 = $87;$$lcssa38 = $91;$$lcssa39 = $98;
   break;
  }
 }
 $100 = $$lcssa39 >> 16; //@line 959 "fastmix.cpp"
 $101 = HEAP32[$7>>2]|0; //@line 959 "fastmix.cpp"
 $102 = (($101) + ($100))|0; //@line 959 "fastmix.cpp"
 HEAP32[$7>>2] = $102; //@line 959 "fastmix.cpp"
 $103 = $$lcssa39 & 65535; //@line 959 "fastmix.cpp"
 HEAP32[$4>>2] = $103; //@line 959 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa36; //@line 959 "fastmix.cpp"
 $104 = ((($pChannel)) + 16|0); //@line 959 "fastmix.cpp"
 HEAP32[$104>>2] = $$lcssa37; //@line 959 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa; //@line 959 "fastmix.cpp"
 $105 = ((($pChannel)) + 20|0); //@line 959 "fastmix.cpp"
 HEAP32[$105>>2] = $$lcssa38; //@line 959 "fastmix.cpp"
 return; //@line 959 "fastmix.cpp"
}
function __Z24Stereo16BitSplineRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa21 = 0, $$lcssa22 = 0, $$lcssa23 = 0, $$lcssa24 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $11 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 961 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 961 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 961 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 961 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 962 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 962 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 962 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 962 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 962 "fastmix.cpp"
 $9 = $8 << 1; //@line 962 "fastmix.cpp"
 $10 = (($6) + ($9)|0); //@line 962 "fastmix.cpp"
 $11 = ((($pChannel)) + 36|0); //@line 962 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 962 "fastmix.cpp"
 $13 = $12 & 64; //@line 962 "fastmix.cpp"
 $14 = ($13|0)==(0); //@line 962 "fastmix.cpp"
 $15 = (($10) + ($8<<1)|0); //@line 962 "fastmix.cpp"
 $$ = $14 ? $10 : $15; //@line 962 "fastmix.cpp"
 $16 = ((($pChannel)) + 28|0); //@line 964 "fastmix.cpp"
 $17 = ((($pChannel)) + 24|0); //@line 964 "fastmix.cpp"
 $18 = ((($pChannel)) + 12|0); //@line 965 "fastmix.cpp"
 $nPos$0 = $5;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $19 = $nPos$0 >> 16; //@line 963 "fastmix.cpp"
  $20 = $nPos$0 >>> 4; //@line 963 "fastmix.cpp"
  $21 = $20 & 4092; //@line 963 "fastmix.cpp"
  $22 = (1209506 + ($21<<1)|0); //@line 963 "fastmix.cpp"
  $23 = HEAP16[$22>>1]|0; //@line 963 "fastmix.cpp"
  $24 = $23 << 16 >> 16; //@line 963 "fastmix.cpp"
  $25 = $19 << 1; //@line 963 "fastmix.cpp"
  $26 = (($25) + -2)|0; //@line 963 "fastmix.cpp"
  $27 = (($$) + ($26<<1)|0); //@line 963 "fastmix.cpp"
  $28 = HEAP16[$27>>1]|0; //@line 963 "fastmix.cpp"
  $29 = $28 << 16 >> 16; //@line 963 "fastmix.cpp"
  $30 = Math_imul($29, $24)|0; //@line 963 "fastmix.cpp"
  $31 = $21 | 1; //@line 963 "fastmix.cpp"
  $32 = (1209506 + ($31<<1)|0); //@line 963 "fastmix.cpp"
  $33 = HEAP16[$32>>1]|0; //@line 963 "fastmix.cpp"
  $34 = $33 << 16 >> 16; //@line 963 "fastmix.cpp"
  $35 = (($$) + ($25<<1)|0); //@line 963 "fastmix.cpp"
  $36 = HEAP16[$35>>1]|0; //@line 963 "fastmix.cpp"
  $37 = $36 << 16 >> 16; //@line 963 "fastmix.cpp"
  $38 = Math_imul($37, $34)|0; //@line 963 "fastmix.cpp"
  $39 = (($38) + ($30))|0; //@line 963 "fastmix.cpp"
  $40 = $21 | 2; //@line 963 "fastmix.cpp"
  $41 = (1209506 + ($40<<1)|0); //@line 963 "fastmix.cpp"
  $42 = HEAP16[$41>>1]|0; //@line 963 "fastmix.cpp"
  $43 = $42 << 16 >> 16; //@line 963 "fastmix.cpp"
  $44 = (($25) + 2)|0; //@line 963 "fastmix.cpp"
  $45 = (($$) + ($44<<1)|0); //@line 963 "fastmix.cpp"
  $46 = HEAP16[$45>>1]|0; //@line 963 "fastmix.cpp"
  $47 = $46 << 16 >> 16; //@line 963 "fastmix.cpp"
  $48 = Math_imul($47, $43)|0; //@line 963 "fastmix.cpp"
  $49 = (($39) + ($48))|0; //@line 963 "fastmix.cpp"
  $50 = $21 | 3; //@line 963 "fastmix.cpp"
  $51 = (1209506 + ($50<<1)|0); //@line 963 "fastmix.cpp"
  $52 = HEAP16[$51>>1]|0; //@line 963 "fastmix.cpp"
  $53 = $52 << 16 >> 16; //@line 963 "fastmix.cpp"
  $54 = (($25) + 4)|0; //@line 963 "fastmix.cpp"
  $55 = (($$) + ($54<<1)|0); //@line 963 "fastmix.cpp"
  $56 = HEAP16[$55>>1]|0; //@line 963 "fastmix.cpp"
  $57 = $56 << 16 >> 16; //@line 963 "fastmix.cpp"
  $58 = Math_imul($57, $53)|0; //@line 963 "fastmix.cpp"
  $59 = (($49) + ($58))|0; //@line 963 "fastmix.cpp"
  $60 = $59 >> 14; //@line 963 "fastmix.cpp"
  $61 = (($25) + -1)|0; //@line 963 "fastmix.cpp"
  $62 = (($$) + ($61<<1)|0); //@line 963 "fastmix.cpp"
  $63 = HEAP16[$62>>1]|0; //@line 963 "fastmix.cpp"
  $64 = $63 << 16 >> 16; //@line 963 "fastmix.cpp"
  $65 = Math_imul($64, $24)|0; //@line 963 "fastmix.cpp"
  $66 = $25 | 1; //@line 963 "fastmix.cpp"
  $67 = (($$) + ($66<<1)|0); //@line 963 "fastmix.cpp"
  $68 = HEAP16[$67>>1]|0; //@line 963 "fastmix.cpp"
  $69 = $68 << 16 >> 16; //@line 963 "fastmix.cpp"
  $70 = Math_imul($69, $34)|0; //@line 963 "fastmix.cpp"
  $71 = (($70) + ($65))|0; //@line 963 "fastmix.cpp"
  $72 = (($25) + 3)|0; //@line 963 "fastmix.cpp"
  $73 = (($$) + ($72<<1)|0); //@line 963 "fastmix.cpp"
  $74 = HEAP16[$73>>1]|0; //@line 963 "fastmix.cpp"
  $75 = $74 << 16 >> 16; //@line 963 "fastmix.cpp"
  $76 = Math_imul($75, $43)|0; //@line 963 "fastmix.cpp"
  $77 = (($71) + ($76))|0; //@line 963 "fastmix.cpp"
  $78 = (($25) + 5)|0; //@line 963 "fastmix.cpp"
  $79 = (($$) + ($78<<1)|0); //@line 963 "fastmix.cpp"
  $80 = HEAP16[$79>>1]|0; //@line 963 "fastmix.cpp"
  $81 = $80 << 16 >> 16; //@line 963 "fastmix.cpp"
  $82 = Math_imul($81, $53)|0; //@line 963 "fastmix.cpp"
  $83 = (($77) + ($82))|0; //@line 963 "fastmix.cpp"
  $84 = $83 >> 14; //@line 963 "fastmix.cpp"
  $85 = HEAP32[$16>>2]|0; //@line 964 "fastmix.cpp"
  $86 = (($85) + ($nRampLeftVol$0))|0; //@line 964 "fastmix.cpp"
  $87 = HEAP32[$17>>2]|0; //@line 964 "fastmix.cpp"
  $88 = (($87) + ($nRampRightVol$0))|0; //@line 964 "fastmix.cpp"
  $89 = $88 >> 12; //@line 964 "fastmix.cpp"
  $90 = Math_imul($89, $60)|0; //@line 964 "fastmix.cpp"
  $91 = HEAP32[$pvol$0>>2]|0; //@line 964 "fastmix.cpp"
  $92 = (($90) + ($91))|0; //@line 964 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $92; //@line 964 "fastmix.cpp"
  $93 = $86 >> 12; //@line 964 "fastmix.cpp"
  $94 = Math_imul($84, $93)|0; //@line 964 "fastmix.cpp"
  $95 = ((($pvol$0)) + 4|0); //@line 964 "fastmix.cpp"
  $96 = HEAP32[$95>>2]|0; //@line 964 "fastmix.cpp"
  $97 = (($94) + ($96))|0; //@line 964 "fastmix.cpp"
  HEAP32[$95>>2] = $97; //@line 964 "fastmix.cpp"
  $98 = ((($pvol$0)) + 8|0); //@line 964 "fastmix.cpp"
  $99 = HEAP32[$18>>2]|0; //@line 965 "fastmix.cpp"
  $100 = (($99) + ($nPos$0))|0; //@line 965 "fastmix.cpp"
  $101 = ($98>>>0)<($pbufmax>>>0); //@line 965 "fastmix.cpp"
  if ($101) {
   $nPos$0 = $100;$nRampLeftVol$0 = $86;$nRampRightVol$0 = $88;$pvol$0 = $98;
  } else {
   $$lcssa = $86;$$lcssa21 = $88;$$lcssa22 = $89;$$lcssa23 = $93;$$lcssa24 = $100;
   break;
  }
 }
 $102 = $$lcssa24 >> 16; //@line 965 "fastmix.cpp"
 $103 = HEAP32[$7>>2]|0; //@line 965 "fastmix.cpp"
 $104 = (($103) + ($102))|0; //@line 965 "fastmix.cpp"
 HEAP32[$7>>2] = $104; //@line 965 "fastmix.cpp"
 $105 = $$lcssa24 & 65535; //@line 965 "fastmix.cpp"
 HEAP32[$4>>2] = $105; //@line 965 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa21; //@line 965 "fastmix.cpp"
 $106 = ((($pChannel)) + 16|0); //@line 965 "fastmix.cpp"
 HEAP32[$106>>2] = $$lcssa22; //@line 965 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa; //@line 965 "fastmix.cpp"
 $107 = ((($pChannel)) + 20|0); //@line 965 "fastmix.cpp"
 HEAP32[$107>>2] = $$lcssa23; //@line 965 "fastmix.cpp"
 return; //@line 965 "fastmix.cpp"
}
function __Z23FilterMono8BitSplineMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa13 = 0, $$sum = 0, $$sum1 = 0, $$sum2 = 0, $$sum3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $9 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $nPos$0 = 0;
 var $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 56|0); //@line 1015 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1015 "fastmix.cpp"
 $2 = ((($pChannel)) + 60|0); //@line 1015 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1015 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 1016 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1016 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 1016 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 1016 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 1016 "fastmix.cpp"
 $9 = (($6) + ($8)|0); //@line 1016 "fastmix.cpp"
 $10 = ((($pChannel)) + 36|0); //@line 1016 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 1016 "fastmix.cpp"
 $12 = $11 & 64; //@line 1016 "fastmix.cpp"
 $13 = ($12|0)==(0); //@line 1016 "fastmix.cpp"
 $$ = $13 ? $6 : $9; //@line 1016 "fastmix.cpp"
 $14 = (($8) + -1)|0; //@line 1017 "fastmix.cpp"
 $15 = (($8) + 2)|0; //@line 1017 "fastmix.cpp"
 $16 = (($8) + 1)|0; //@line 1017 "fastmix.cpp"
 $17 = ((($pChannel)) + 72|0); //@line 1018 "fastmix.cpp"
 $18 = ((($pChannel)) + 76|0); //@line 1018 "fastmix.cpp"
 $19 = ((($pChannel)) + 80|0); //@line 1018 "fastmix.cpp"
 $20 = ((($pChannel)) + 16|0); //@line 1019 "fastmix.cpp"
 $21 = ((($pChannel)) + 20|0); //@line 1019 "fastmix.cpp"
 $22 = ((($pChannel)) + 12|0); //@line 1020 "fastmix.cpp"
 $fy1$0 = $1;$fy2$0 = $3;$nPos$0 = $5;$pvol$0 = $pbuffer;
 while(1) {
  $23 = $nPos$0 >> 16; //@line 1017 "fastmix.cpp"
  $24 = $nPos$0 >>> 4; //@line 1017 "fastmix.cpp"
  $25 = $24 & 4092; //@line 1017 "fastmix.cpp"
  $26 = (1209506 + ($25<<1)|0); //@line 1017 "fastmix.cpp"
  $27 = HEAP16[$26>>1]|0; //@line 1017 "fastmix.cpp"
  $28 = $27 << 16 >> 16; //@line 1017 "fastmix.cpp"
  $$sum = (($14) + ($23))|0; //@line 1017 "fastmix.cpp"
  $29 = (($$) + ($$sum)|0); //@line 1017 "fastmix.cpp"
  $30 = HEAP8[$29>>0]|0; //@line 1017 "fastmix.cpp"
  $31 = $30 << 24 >> 24; //@line 1017 "fastmix.cpp"
  $32 = Math_imul($31, $28)|0; //@line 1017 "fastmix.cpp"
  $33 = $25 | 1; //@line 1017 "fastmix.cpp"
  $34 = (1209506 + ($33<<1)|0); //@line 1017 "fastmix.cpp"
  $35 = HEAP16[$34>>1]|0; //@line 1017 "fastmix.cpp"
  $36 = $35 << 16 >> 16; //@line 1017 "fastmix.cpp"
  $$sum1 = (($23) + ($8))|0; //@line 1017 "fastmix.cpp"
  $37 = (($$) + ($$sum1)|0); //@line 1017 "fastmix.cpp"
  $38 = HEAP8[$37>>0]|0; //@line 1017 "fastmix.cpp"
  $39 = $38 << 24 >> 24; //@line 1017 "fastmix.cpp"
  $40 = Math_imul($39, $36)|0; //@line 1017 "fastmix.cpp"
  $41 = (($40) + ($32))|0; //@line 1017 "fastmix.cpp"
  $42 = $25 | 3; //@line 1017 "fastmix.cpp"
  $43 = (1209506 + ($42<<1)|0); //@line 1017 "fastmix.cpp"
  $44 = HEAP16[$43>>1]|0; //@line 1017 "fastmix.cpp"
  $45 = $44 << 16 >> 16; //@line 1017 "fastmix.cpp"
  $$sum2 = (($15) + ($23))|0; //@line 1017 "fastmix.cpp"
  $46 = (($$) + ($$sum2)|0); //@line 1017 "fastmix.cpp"
  $47 = HEAP8[$46>>0]|0; //@line 1017 "fastmix.cpp"
  $48 = $47 << 24 >> 24; //@line 1017 "fastmix.cpp"
  $49 = Math_imul($48, $45)|0; //@line 1017 "fastmix.cpp"
  $50 = (($41) + ($49))|0; //@line 1017 "fastmix.cpp"
  $51 = $25 | 2; //@line 1017 "fastmix.cpp"
  $52 = (1209506 + ($51<<1)|0); //@line 1017 "fastmix.cpp"
  $53 = HEAP16[$52>>1]|0; //@line 1017 "fastmix.cpp"
  $54 = $53 << 16 >> 16; //@line 1017 "fastmix.cpp"
  $$sum3 = (($16) + ($23))|0; //@line 1017 "fastmix.cpp"
  $55 = (($$) + ($$sum3)|0); //@line 1017 "fastmix.cpp"
  $56 = HEAP8[$55>>0]|0; //@line 1017 "fastmix.cpp"
  $57 = $56 << 24 >> 24; //@line 1017 "fastmix.cpp"
  $58 = Math_imul($57, $54)|0; //@line 1017 "fastmix.cpp"
  $59 = (($50) + ($58))|0; //@line 1017 "fastmix.cpp"
  $60 = $59 >> 6; //@line 1017 "fastmix.cpp"
  $61 = HEAP32[$17>>2]|0; //@line 1018 "fastmix.cpp"
  $62 = Math_imul($60, $61)|0; //@line 1018 "fastmix.cpp"
  $63 = HEAP32[$18>>2]|0; //@line 1018 "fastmix.cpp"
  $64 = Math_imul($63, $fy1$0)|0; //@line 1018 "fastmix.cpp"
  $65 = HEAP32[$19>>2]|0; //@line 1018 "fastmix.cpp"
  $66 = Math_imul($65, $fy2$0)|0; //@line 1018 "fastmix.cpp"
  $67 = (($64) + 4096)|0; //@line 1018 "fastmix.cpp"
  $68 = (($67) + ($66))|0; //@line 1018 "fastmix.cpp"
  $69 = (($68) + ($62))|0; //@line 1018 "fastmix.cpp"
  $70 = $69 >> 13; //@line 1018 "fastmix.cpp"
  $71 = HEAP32[$20>>2]|0; //@line 1019 "fastmix.cpp"
  $72 = Math_imul($70, $71)|0; //@line 1019 "fastmix.cpp"
  $73 = HEAP32[$pvol$0>>2]|0; //@line 1019 "fastmix.cpp"
  $74 = (($72) + ($73))|0; //@line 1019 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $74; //@line 1019 "fastmix.cpp"
  $75 = HEAP32[$21>>2]|0; //@line 1019 "fastmix.cpp"
  $76 = Math_imul($70, $75)|0; //@line 1019 "fastmix.cpp"
  $77 = ((($pvol$0)) + 4|0); //@line 1019 "fastmix.cpp"
  $78 = HEAP32[$77>>2]|0; //@line 1019 "fastmix.cpp"
  $79 = (($76) + ($78))|0; //@line 1019 "fastmix.cpp"
  HEAP32[$77>>2] = $79; //@line 1019 "fastmix.cpp"
  $80 = ((($pvol$0)) + 8|0); //@line 1019 "fastmix.cpp"
  $81 = HEAP32[$22>>2]|0; //@line 1020 "fastmix.cpp"
  $82 = (($81) + ($nPos$0))|0; //@line 1020 "fastmix.cpp"
  $83 = ($80>>>0)<($pbufmax>>>0); //@line 1020 "fastmix.cpp"
  if ($83) {
   $fy2$0$phi = $fy1$0;$fy1$0 = $70;$nPos$0 = $82;$pvol$0 = $80;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $70;$$lcssa13 = $82;$fy1$0$lcssa = $fy1$0;
   break;
  }
 }
 $84 = $$lcssa13 >> 16; //@line 1020 "fastmix.cpp"
 $85 = HEAP32[$7>>2]|0; //@line 1020 "fastmix.cpp"
 $86 = (($85) + ($84))|0; //@line 1020 "fastmix.cpp"
 HEAP32[$7>>2] = $86; //@line 1020 "fastmix.cpp"
 $87 = $$lcssa13 & 65535; //@line 1020 "fastmix.cpp"
 HEAP32[$4>>2] = $87; //@line 1020 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 1020 "fastmix.cpp"
 HEAP32[$2>>2] = $fy1$0$lcssa; //@line 1020 "fastmix.cpp"
 return; //@line 1020 "fastmix.cpp"
}
function __Z24FilterMono16BitSplineMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 56|0); //@line 1022 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1022 "fastmix.cpp"
 $2 = ((($pChannel)) + 60|0); //@line 1022 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1022 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 1023 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1023 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 1023 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 1023 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 1023 "fastmix.cpp"
 $9 = $8 << 1; //@line 1023 "fastmix.cpp"
 $10 = (($6) + ($9)|0); //@line 1023 "fastmix.cpp"
 $11 = ((($pChannel)) + 36|0); //@line 1023 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1023 "fastmix.cpp"
 $13 = $12 & 64; //@line 1023 "fastmix.cpp"
 $14 = ($13|0)==(0); //@line 1023 "fastmix.cpp"
 $15 = (($10) + ($8<<1)|0); //@line 1023 "fastmix.cpp"
 $$ = $14 ? $10 : $15; //@line 1023 "fastmix.cpp"
 $16 = ((($pChannel)) + 72|0); //@line 1025 "fastmix.cpp"
 $17 = ((($pChannel)) + 76|0); //@line 1025 "fastmix.cpp"
 $18 = ((($pChannel)) + 80|0); //@line 1025 "fastmix.cpp"
 $19 = ((($pChannel)) + 16|0); //@line 1026 "fastmix.cpp"
 $20 = ((($pChannel)) + 20|0); //@line 1026 "fastmix.cpp"
 $21 = ((($pChannel)) + 12|0); //@line 1027 "fastmix.cpp"
 $fy1$0 = $1;$fy2$0 = $3;$nPos$0 = $5;$pvol$0 = $pbuffer;
 while(1) {
  $22 = $nPos$0 >> 16; //@line 1024 "fastmix.cpp"
  $23 = $nPos$0 >>> 4; //@line 1024 "fastmix.cpp"
  $24 = $23 & 4092; //@line 1024 "fastmix.cpp"
  $25 = (1209506 + ($24<<1)|0); //@line 1024 "fastmix.cpp"
  $26 = HEAP16[$25>>1]|0; //@line 1024 "fastmix.cpp"
  $27 = $26 << 16 >> 16; //@line 1024 "fastmix.cpp"
  $28 = (($22) + -1)|0; //@line 1024 "fastmix.cpp"
  $29 = (($$) + ($28<<1)|0); //@line 1024 "fastmix.cpp"
  $30 = HEAP16[$29>>1]|0; //@line 1024 "fastmix.cpp"
  $31 = $30 << 16 >> 16; //@line 1024 "fastmix.cpp"
  $32 = Math_imul($31, $27)|0; //@line 1024 "fastmix.cpp"
  $33 = $24 | 1; //@line 1024 "fastmix.cpp"
  $34 = (1209506 + ($33<<1)|0); //@line 1024 "fastmix.cpp"
  $35 = HEAP16[$34>>1]|0; //@line 1024 "fastmix.cpp"
  $36 = $35 << 16 >> 16; //@line 1024 "fastmix.cpp"
  $37 = (($$) + ($22<<1)|0); //@line 1024 "fastmix.cpp"
  $38 = HEAP16[$37>>1]|0; //@line 1024 "fastmix.cpp"
  $39 = $38 << 16 >> 16; //@line 1024 "fastmix.cpp"
  $40 = Math_imul($39, $36)|0; //@line 1024 "fastmix.cpp"
  $41 = (($40) + ($32))|0; //@line 1024 "fastmix.cpp"
  $42 = $24 | 3; //@line 1024 "fastmix.cpp"
  $43 = (1209506 + ($42<<1)|0); //@line 1024 "fastmix.cpp"
  $44 = HEAP16[$43>>1]|0; //@line 1024 "fastmix.cpp"
  $45 = $44 << 16 >> 16; //@line 1024 "fastmix.cpp"
  $46 = (($22) + 2)|0; //@line 1024 "fastmix.cpp"
  $47 = (($$) + ($46<<1)|0); //@line 1024 "fastmix.cpp"
  $48 = HEAP16[$47>>1]|0; //@line 1024 "fastmix.cpp"
  $49 = $48 << 16 >> 16; //@line 1024 "fastmix.cpp"
  $50 = Math_imul($49, $45)|0; //@line 1024 "fastmix.cpp"
  $51 = (($41) + ($50))|0; //@line 1024 "fastmix.cpp"
  $52 = $24 | 2; //@line 1024 "fastmix.cpp"
  $53 = (1209506 + ($52<<1)|0); //@line 1024 "fastmix.cpp"
  $54 = HEAP16[$53>>1]|0; //@line 1024 "fastmix.cpp"
  $55 = $54 << 16 >> 16; //@line 1024 "fastmix.cpp"
  $56 = (($22) + 1)|0; //@line 1024 "fastmix.cpp"
  $57 = (($$) + ($56<<1)|0); //@line 1024 "fastmix.cpp"
  $58 = HEAP16[$57>>1]|0; //@line 1024 "fastmix.cpp"
  $59 = $58 << 16 >> 16; //@line 1024 "fastmix.cpp"
  $60 = Math_imul($59, $55)|0; //@line 1024 "fastmix.cpp"
  $61 = (($51) + ($60))|0; //@line 1024 "fastmix.cpp"
  $62 = $61 >> 14; //@line 1024 "fastmix.cpp"
  $63 = HEAP32[$16>>2]|0; //@line 1025 "fastmix.cpp"
  $64 = Math_imul($62, $63)|0; //@line 1025 "fastmix.cpp"
  $65 = HEAP32[$17>>2]|0; //@line 1025 "fastmix.cpp"
  $66 = Math_imul($65, $fy1$0)|0; //@line 1025 "fastmix.cpp"
  $67 = HEAP32[$18>>2]|0; //@line 1025 "fastmix.cpp"
  $68 = Math_imul($67, $fy2$0)|0; //@line 1025 "fastmix.cpp"
  $69 = (($66) + 4096)|0; //@line 1025 "fastmix.cpp"
  $70 = (($69) + ($68))|0; //@line 1025 "fastmix.cpp"
  $71 = (($70) + ($64))|0; //@line 1025 "fastmix.cpp"
  $72 = $71 >> 13; //@line 1025 "fastmix.cpp"
  $73 = HEAP32[$19>>2]|0; //@line 1026 "fastmix.cpp"
  $74 = Math_imul($72, $73)|0; //@line 1026 "fastmix.cpp"
  $75 = HEAP32[$pvol$0>>2]|0; //@line 1026 "fastmix.cpp"
  $76 = (($74) + ($75))|0; //@line 1026 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $76; //@line 1026 "fastmix.cpp"
  $77 = HEAP32[$20>>2]|0; //@line 1026 "fastmix.cpp"
  $78 = Math_imul($72, $77)|0; //@line 1026 "fastmix.cpp"
  $79 = ((($pvol$0)) + 4|0); //@line 1026 "fastmix.cpp"
  $80 = HEAP32[$79>>2]|0; //@line 1026 "fastmix.cpp"
  $81 = (($78) + ($80))|0; //@line 1026 "fastmix.cpp"
  HEAP32[$79>>2] = $81; //@line 1026 "fastmix.cpp"
  $82 = ((($pvol$0)) + 8|0); //@line 1026 "fastmix.cpp"
  $83 = HEAP32[$21>>2]|0; //@line 1027 "fastmix.cpp"
  $84 = (($83) + ($nPos$0))|0; //@line 1027 "fastmix.cpp"
  $85 = ($82>>>0)<($pbufmax>>>0); //@line 1027 "fastmix.cpp"
  if ($85) {
   $fy2$0$phi = $fy1$0;$fy1$0 = $72;$nPos$0 = $84;$pvol$0 = $82;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $72;$$lcssa6 = $84;$fy1$0$lcssa = $fy1$0;
   break;
  }
 }
 $86 = $$lcssa6 >> 16; //@line 1027 "fastmix.cpp"
 $87 = HEAP32[$7>>2]|0; //@line 1027 "fastmix.cpp"
 $88 = (($87) + ($86))|0; //@line 1027 "fastmix.cpp"
 HEAP32[$7>>2] = $88; //@line 1027 "fastmix.cpp"
 $89 = $$lcssa6 & 65535; //@line 1027 "fastmix.cpp"
 HEAP32[$4>>2] = $89; //@line 1027 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 1027 "fastmix.cpp"
 HEAP32[$2>>2] = $fy1$0$lcssa; //@line 1027 "fastmix.cpp"
 return; //@line 1027 "fastmix.cpp"
}
function __Z25FilterStereo8BitSplineMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa26 = 0, $$lcssa27 = 0, $$sum = 0, $$sum1 = 0, $$sum2 = 0, $$sum3 = 0, $$sum4 = 0, $$sum5 = 0, $$sum6 = 0, $$sum7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0;
 var $fy3$0 = 0, $fy3$0$lcssa = 0, $fy4$0 = 0, $fy4$0$phi = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 56|0); //@line 1130 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1130 "fastmix.cpp"
 $2 = ((($pChannel)) + 60|0); //@line 1130 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1130 "fastmix.cpp"
 $4 = ((($pChannel)) + 64|0); //@line 1130 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1130 "fastmix.cpp"
 $6 = ((($pChannel)) + 68|0); //@line 1130 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1130 "fastmix.cpp"
 $8 = ((($pChannel)) + 8|0); //@line 1131 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1131 "fastmix.cpp"
 $10 = HEAP32[$pChannel>>2]|0; //@line 1131 "fastmix.cpp"
 $11 = ((($pChannel)) + 4|0); //@line 1131 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1131 "fastmix.cpp"
 $13 = (($10) + ($12)|0); //@line 1131 "fastmix.cpp"
 $14 = ((($pChannel)) + 36|0); //@line 1131 "fastmix.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 1131 "fastmix.cpp"
 $16 = $15 & 64; //@line 1131 "fastmix.cpp"
 $17 = ($16|0)==(0); //@line 1131 "fastmix.cpp"
 $$ = $17 ? $10 : $13; //@line 1131 "fastmix.cpp"
 $18 = (($12) + -2)|0; //@line 1132 "fastmix.cpp"
 $19 = (($12) + 2)|0; //@line 1132 "fastmix.cpp"
 $20 = (($12) + 4)|0; //@line 1132 "fastmix.cpp"
 $21 = (($12) + -1)|0; //@line 1132 "fastmix.cpp"
 $22 = (($12) + 3)|0; //@line 1132 "fastmix.cpp"
 $23 = (($12) + 5)|0; //@line 1132 "fastmix.cpp"
 $24 = ((($pChannel)) + 72|0); //@line 1133 "fastmix.cpp"
 $25 = ((($pChannel)) + 76|0); //@line 1133 "fastmix.cpp"
 $26 = ((($pChannel)) + 80|0); //@line 1133 "fastmix.cpp"
 $27 = ((($pChannel)) + 16|0); //@line 1134 "fastmix.cpp"
 $28 = ((($pChannel)) + 20|0); //@line 1134 "fastmix.cpp"
 $29 = ((($pChannel)) + 12|0); //@line 1135 "fastmix.cpp"
 $fy1$0 = $1;$fy2$0 = $3;$fy3$0 = $5;$fy4$0 = $7;$nPos$0 = $9;$pvol$0 = $pbuffer;
 while(1) {
  $30 = $nPos$0 >> 16; //@line 1132 "fastmix.cpp"
  $31 = $nPos$0 >>> 4; //@line 1132 "fastmix.cpp"
  $32 = $31 & 4092; //@line 1132 "fastmix.cpp"
  $33 = (1209506 + ($32<<1)|0); //@line 1132 "fastmix.cpp"
  $34 = HEAP16[$33>>1]|0; //@line 1132 "fastmix.cpp"
  $35 = $34 << 16 >> 16; //@line 1132 "fastmix.cpp"
  $36 = $30 << 1; //@line 1132 "fastmix.cpp"
  $$sum = (($18) + ($36))|0; //@line 1132 "fastmix.cpp"
  $37 = (($$) + ($$sum)|0); //@line 1132 "fastmix.cpp"
  $38 = HEAP8[$37>>0]|0; //@line 1132 "fastmix.cpp"
  $39 = $38 << 24 >> 24; //@line 1132 "fastmix.cpp"
  $40 = Math_imul($39, $35)|0; //@line 1132 "fastmix.cpp"
  $41 = $32 | 1; //@line 1132 "fastmix.cpp"
  $42 = (1209506 + ($41<<1)|0); //@line 1132 "fastmix.cpp"
  $43 = HEAP16[$42>>1]|0; //@line 1132 "fastmix.cpp"
  $44 = $43 << 16 >> 16; //@line 1132 "fastmix.cpp"
  $$sum1 = (($36) + ($12))|0; //@line 1132 "fastmix.cpp"
  $45 = (($$) + ($$sum1)|0); //@line 1132 "fastmix.cpp"
  $46 = HEAP8[$45>>0]|0; //@line 1132 "fastmix.cpp"
  $47 = $46 << 24 >> 24; //@line 1132 "fastmix.cpp"
  $48 = Math_imul($47, $44)|0; //@line 1132 "fastmix.cpp"
  $49 = (($48) + ($40))|0; //@line 1132 "fastmix.cpp"
  $50 = $32 | 2; //@line 1132 "fastmix.cpp"
  $51 = (1209506 + ($50<<1)|0); //@line 1132 "fastmix.cpp"
  $52 = HEAP16[$51>>1]|0; //@line 1132 "fastmix.cpp"
  $53 = $52 << 16 >> 16; //@line 1132 "fastmix.cpp"
  $$sum2 = (($19) + ($36))|0; //@line 1132 "fastmix.cpp"
  $54 = (($$) + ($$sum2)|0); //@line 1132 "fastmix.cpp"
  $55 = HEAP8[$54>>0]|0; //@line 1132 "fastmix.cpp"
  $56 = $55 << 24 >> 24; //@line 1132 "fastmix.cpp"
  $57 = Math_imul($56, $53)|0; //@line 1132 "fastmix.cpp"
  $58 = (($49) + ($57))|0; //@line 1132 "fastmix.cpp"
  $59 = $32 | 3; //@line 1132 "fastmix.cpp"
  $60 = (1209506 + ($59<<1)|0); //@line 1132 "fastmix.cpp"
  $61 = HEAP16[$60>>1]|0; //@line 1132 "fastmix.cpp"
  $62 = $61 << 16 >> 16; //@line 1132 "fastmix.cpp"
  $$sum3 = (($20) + ($36))|0; //@line 1132 "fastmix.cpp"
  $63 = (($$) + ($$sum3)|0); //@line 1132 "fastmix.cpp"
  $64 = HEAP8[$63>>0]|0; //@line 1132 "fastmix.cpp"
  $65 = $64 << 24 >> 24; //@line 1132 "fastmix.cpp"
  $66 = Math_imul($65, $62)|0; //@line 1132 "fastmix.cpp"
  $67 = (($58) + ($66))|0; //@line 1132 "fastmix.cpp"
  $68 = $67 >> 6; //@line 1132 "fastmix.cpp"
  $$sum4 = (($21) + ($36))|0; //@line 1132 "fastmix.cpp"
  $69 = (($$) + ($$sum4)|0); //@line 1132 "fastmix.cpp"
  $70 = HEAP8[$69>>0]|0; //@line 1132 "fastmix.cpp"
  $71 = $70 << 24 >> 24; //@line 1132 "fastmix.cpp"
  $72 = Math_imul($71, $35)|0; //@line 1132 "fastmix.cpp"
  $73 = $36 | 1; //@line 1132 "fastmix.cpp"
  $$sum5 = (($73) + ($12))|0; //@line 1132 "fastmix.cpp"
  $74 = (($$) + ($$sum5)|0); //@line 1132 "fastmix.cpp"
  $75 = HEAP8[$74>>0]|0; //@line 1132 "fastmix.cpp"
  $76 = $75 << 24 >> 24; //@line 1132 "fastmix.cpp"
  $77 = Math_imul($76, $44)|0; //@line 1132 "fastmix.cpp"
  $78 = (($77) + ($72))|0; //@line 1132 "fastmix.cpp"
  $$sum6 = (($22) + ($36))|0; //@line 1132 "fastmix.cpp"
  $79 = (($$) + ($$sum6)|0); //@line 1132 "fastmix.cpp"
  $80 = HEAP8[$79>>0]|0; //@line 1132 "fastmix.cpp"
  $81 = $80 << 24 >> 24; //@line 1132 "fastmix.cpp"
  $82 = Math_imul($81, $53)|0; //@line 1132 "fastmix.cpp"
  $83 = (($78) + ($82))|0; //@line 1132 "fastmix.cpp"
  $$sum7 = (($23) + ($36))|0; //@line 1132 "fastmix.cpp"
  $84 = (($$) + ($$sum7)|0); //@line 1132 "fastmix.cpp"
  $85 = HEAP8[$84>>0]|0; //@line 1132 "fastmix.cpp"
  $86 = $85 << 24 >> 24; //@line 1132 "fastmix.cpp"
  $87 = Math_imul($86, $62)|0; //@line 1132 "fastmix.cpp"
  $88 = (($83) + ($87))|0; //@line 1132 "fastmix.cpp"
  $89 = $88 >> 6; //@line 1132 "fastmix.cpp"
  $90 = HEAP32[$24>>2]|0; //@line 1133 "fastmix.cpp"
  $91 = Math_imul($90, $68)|0; //@line 1133 "fastmix.cpp"
  $92 = HEAP32[$25>>2]|0; //@line 1133 "fastmix.cpp"
  $93 = Math_imul($92, $fy1$0)|0; //@line 1133 "fastmix.cpp"
  $94 = HEAP32[$26>>2]|0; //@line 1133 "fastmix.cpp"
  $95 = Math_imul($94, $fy2$0)|0; //@line 1133 "fastmix.cpp"
  $96 = (($91) + 4096)|0; //@line 1133 "fastmix.cpp"
  $97 = (($96) + ($93))|0; //@line 1133 "fastmix.cpp"
  $98 = (($97) + ($95))|0; //@line 1133 "fastmix.cpp"
  $99 = $98 >> 13; //@line 1133 "fastmix.cpp"
  $100 = Math_imul($89, $90)|0; //@line 1133 "fastmix.cpp"
  $101 = Math_imul($92, $fy3$0)|0; //@line 1133 "fastmix.cpp"
  $102 = Math_imul($94, $fy4$0)|0; //@line 1133 "fastmix.cpp"
  $103 = (($101) + 4096)|0; //@line 1133 "fastmix.cpp"
  $104 = (($103) + ($102))|0; //@line 1133 "fastmix.cpp"
  $105 = (($104) + ($100))|0; //@line 1133 "fastmix.cpp"
  $106 = $105 >> 13; //@line 1133 "fastmix.cpp"
  $107 = HEAP32[$27>>2]|0; //@line 1134 "fastmix.cpp"
  $108 = Math_imul($99, $107)|0; //@line 1134 "fastmix.cpp"
  $109 = HEAP32[$pvol$0>>2]|0; //@line 1134 "fastmix.cpp"
  $110 = (($108) + ($109))|0; //@line 1134 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $110; //@line 1134 "fastmix.cpp"
  $111 = HEAP32[$28>>2]|0; //@line 1134 "fastmix.cpp"
  $112 = Math_imul($106, $111)|0; //@line 1134 "fastmix.cpp"
  $113 = ((($pvol$0)) + 4|0); //@line 1134 "fastmix.cpp"
  $114 = HEAP32[$113>>2]|0; //@line 1134 "fastmix.cpp"
  $115 = (($112) + ($114))|0; //@line 1134 "fastmix.cpp"
  HEAP32[$113>>2] = $115; //@line 1134 "fastmix.cpp"
  $116 = ((($pvol$0)) + 8|0); //@line 1134 "fastmix.cpp"
  $117 = HEAP32[$29>>2]|0; //@line 1135 "fastmix.cpp"
  $118 = (($117) + ($nPos$0))|0; //@line 1135 "fastmix.cpp"
  $119 = ($116>>>0)<($pbufmax>>>0); //@line 1135 "fastmix.cpp"
  if ($119) {
   $fy4$0$phi = $fy3$0;$fy2$0$phi = $fy1$0;$fy1$0 = $99;$fy3$0 = $106;$nPos$0 = $118;$pvol$0 = $116;$fy4$0 = $fy4$0$phi;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $99;$$lcssa26 = $106;$$lcssa27 = $118;$fy1$0$lcssa = $fy1$0;$fy3$0$lcssa = $fy3$0;
   break;
  }
 }
 $120 = $$lcssa27 >> 16; //@line 1135 "fastmix.cpp"
 $121 = HEAP32[$11>>2]|0; //@line 1135 "fastmix.cpp"
 $122 = (($121) + ($120))|0; //@line 1135 "fastmix.cpp"
 HEAP32[$11>>2] = $122; //@line 1135 "fastmix.cpp"
 $123 = $$lcssa27 & 65535; //@line 1135 "fastmix.cpp"
 HEAP32[$8>>2] = $123; //@line 1135 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 1135 "fastmix.cpp"
 HEAP32[$2>>2] = $fy1$0$lcssa; //@line 1135 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa26; //@line 1135 "fastmix.cpp"
 HEAP32[$6>>2] = $fy3$0$lcssa; //@line 1135 "fastmix.cpp"
 return; //@line 1135 "fastmix.cpp"
}
function __Z26FilterStereo16BitSplineMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa11 = 0, $$lcssa12 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $fy3$0 = 0, $fy3$0$lcssa = 0, $fy4$0 = 0, $fy4$0$phi = 0, $nPos$0 = 0, $pvol$0 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 56|0); //@line 1137 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1137 "fastmix.cpp"
 $2 = ((($pChannel)) + 60|0); //@line 1137 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1137 "fastmix.cpp"
 $4 = ((($pChannel)) + 64|0); //@line 1137 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1137 "fastmix.cpp"
 $6 = ((($pChannel)) + 68|0); //@line 1137 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1137 "fastmix.cpp"
 $8 = ((($pChannel)) + 8|0); //@line 1138 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1138 "fastmix.cpp"
 $10 = HEAP32[$pChannel>>2]|0; //@line 1138 "fastmix.cpp"
 $11 = ((($pChannel)) + 4|0); //@line 1138 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1138 "fastmix.cpp"
 $13 = $12 << 1; //@line 1138 "fastmix.cpp"
 $14 = (($10) + ($13)|0); //@line 1138 "fastmix.cpp"
 $15 = ((($pChannel)) + 36|0); //@line 1138 "fastmix.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1138 "fastmix.cpp"
 $17 = $16 & 64; //@line 1138 "fastmix.cpp"
 $18 = ($17|0)==(0); //@line 1138 "fastmix.cpp"
 $19 = (($14) + ($12<<1)|0); //@line 1138 "fastmix.cpp"
 $$ = $18 ? $14 : $19; //@line 1138 "fastmix.cpp"
 $20 = ((($pChannel)) + 72|0); //@line 1140 "fastmix.cpp"
 $21 = ((($pChannel)) + 76|0); //@line 1140 "fastmix.cpp"
 $22 = ((($pChannel)) + 80|0); //@line 1140 "fastmix.cpp"
 $23 = ((($pChannel)) + 16|0); //@line 1141 "fastmix.cpp"
 $24 = ((($pChannel)) + 20|0); //@line 1141 "fastmix.cpp"
 $25 = ((($pChannel)) + 12|0); //@line 1142 "fastmix.cpp"
 $fy1$0 = $1;$fy2$0 = $3;$fy3$0 = $5;$fy4$0 = $7;$nPos$0 = $9;$pvol$0 = $pbuffer;
 while(1) {
  $26 = $nPos$0 >> 16; //@line 1139 "fastmix.cpp"
  $27 = $nPos$0 >>> 4; //@line 1139 "fastmix.cpp"
  $28 = $27 & 4092; //@line 1139 "fastmix.cpp"
  $29 = (1209506 + ($28<<1)|0); //@line 1139 "fastmix.cpp"
  $30 = HEAP16[$29>>1]|0; //@line 1139 "fastmix.cpp"
  $31 = $30 << 16 >> 16; //@line 1139 "fastmix.cpp"
  $32 = $26 << 1; //@line 1139 "fastmix.cpp"
  $33 = (($32) + -2)|0; //@line 1139 "fastmix.cpp"
  $34 = (($$) + ($33<<1)|0); //@line 1139 "fastmix.cpp"
  $35 = HEAP16[$34>>1]|0; //@line 1139 "fastmix.cpp"
  $36 = $35 << 16 >> 16; //@line 1139 "fastmix.cpp"
  $37 = Math_imul($36, $31)|0; //@line 1139 "fastmix.cpp"
  $38 = $28 | 1; //@line 1139 "fastmix.cpp"
  $39 = (1209506 + ($38<<1)|0); //@line 1139 "fastmix.cpp"
  $40 = HEAP16[$39>>1]|0; //@line 1139 "fastmix.cpp"
  $41 = $40 << 16 >> 16; //@line 1139 "fastmix.cpp"
  $42 = (($$) + ($32<<1)|0); //@line 1139 "fastmix.cpp"
  $43 = HEAP16[$42>>1]|0; //@line 1139 "fastmix.cpp"
  $44 = $43 << 16 >> 16; //@line 1139 "fastmix.cpp"
  $45 = Math_imul($44, $41)|0; //@line 1139 "fastmix.cpp"
  $46 = (($45) + ($37))|0; //@line 1139 "fastmix.cpp"
  $47 = $28 | 2; //@line 1139 "fastmix.cpp"
  $48 = (1209506 + ($47<<1)|0); //@line 1139 "fastmix.cpp"
  $49 = HEAP16[$48>>1]|0; //@line 1139 "fastmix.cpp"
  $50 = $49 << 16 >> 16; //@line 1139 "fastmix.cpp"
  $51 = (($32) + 2)|0; //@line 1139 "fastmix.cpp"
  $52 = (($$) + ($51<<1)|0); //@line 1139 "fastmix.cpp"
  $53 = HEAP16[$52>>1]|0; //@line 1139 "fastmix.cpp"
  $54 = $53 << 16 >> 16; //@line 1139 "fastmix.cpp"
  $55 = Math_imul($54, $50)|0; //@line 1139 "fastmix.cpp"
  $56 = (($46) + ($55))|0; //@line 1139 "fastmix.cpp"
  $57 = $28 | 3; //@line 1139 "fastmix.cpp"
  $58 = (1209506 + ($57<<1)|0); //@line 1139 "fastmix.cpp"
  $59 = HEAP16[$58>>1]|0; //@line 1139 "fastmix.cpp"
  $60 = $59 << 16 >> 16; //@line 1139 "fastmix.cpp"
  $61 = (($32) + 4)|0; //@line 1139 "fastmix.cpp"
  $62 = (($$) + ($61<<1)|0); //@line 1139 "fastmix.cpp"
  $63 = HEAP16[$62>>1]|0; //@line 1139 "fastmix.cpp"
  $64 = $63 << 16 >> 16; //@line 1139 "fastmix.cpp"
  $65 = Math_imul($64, $60)|0; //@line 1139 "fastmix.cpp"
  $66 = (($56) + ($65))|0; //@line 1139 "fastmix.cpp"
  $67 = $66 >> 14; //@line 1139 "fastmix.cpp"
  $68 = (($32) + -1)|0; //@line 1139 "fastmix.cpp"
  $69 = (($$) + ($68<<1)|0); //@line 1139 "fastmix.cpp"
  $70 = HEAP16[$69>>1]|0; //@line 1139 "fastmix.cpp"
  $71 = $70 << 16 >> 16; //@line 1139 "fastmix.cpp"
  $72 = Math_imul($71, $31)|0; //@line 1139 "fastmix.cpp"
  $73 = $32 | 1; //@line 1139 "fastmix.cpp"
  $74 = (($$) + ($73<<1)|0); //@line 1139 "fastmix.cpp"
  $75 = HEAP16[$74>>1]|0; //@line 1139 "fastmix.cpp"
  $76 = $75 << 16 >> 16; //@line 1139 "fastmix.cpp"
  $77 = Math_imul($76, $41)|0; //@line 1139 "fastmix.cpp"
  $78 = (($77) + ($72))|0; //@line 1139 "fastmix.cpp"
  $79 = (($32) + 3)|0; //@line 1139 "fastmix.cpp"
  $80 = (($$) + ($79<<1)|0); //@line 1139 "fastmix.cpp"
  $81 = HEAP16[$80>>1]|0; //@line 1139 "fastmix.cpp"
  $82 = $81 << 16 >> 16; //@line 1139 "fastmix.cpp"
  $83 = Math_imul($82, $50)|0; //@line 1139 "fastmix.cpp"
  $84 = (($78) + ($83))|0; //@line 1139 "fastmix.cpp"
  $85 = (($32) + 5)|0; //@line 1139 "fastmix.cpp"
  $86 = (($$) + ($85<<1)|0); //@line 1139 "fastmix.cpp"
  $87 = HEAP16[$86>>1]|0; //@line 1139 "fastmix.cpp"
  $88 = $87 << 16 >> 16; //@line 1139 "fastmix.cpp"
  $89 = Math_imul($88, $60)|0; //@line 1139 "fastmix.cpp"
  $90 = (($84) + ($89))|0; //@line 1139 "fastmix.cpp"
  $91 = $90 >> 14; //@line 1139 "fastmix.cpp"
  $92 = HEAP32[$20>>2]|0; //@line 1140 "fastmix.cpp"
  $93 = Math_imul($92, $67)|0; //@line 1140 "fastmix.cpp"
  $94 = HEAP32[$21>>2]|0; //@line 1140 "fastmix.cpp"
  $95 = Math_imul($94, $fy1$0)|0; //@line 1140 "fastmix.cpp"
  $96 = HEAP32[$22>>2]|0; //@line 1140 "fastmix.cpp"
  $97 = Math_imul($96, $fy2$0)|0; //@line 1140 "fastmix.cpp"
  $98 = (($93) + 4096)|0; //@line 1140 "fastmix.cpp"
  $99 = (($98) + ($95))|0; //@line 1140 "fastmix.cpp"
  $100 = (($99) + ($97))|0; //@line 1140 "fastmix.cpp"
  $101 = $100 >> 13; //@line 1140 "fastmix.cpp"
  $102 = Math_imul($91, $92)|0; //@line 1140 "fastmix.cpp"
  $103 = Math_imul($94, $fy3$0)|0; //@line 1140 "fastmix.cpp"
  $104 = Math_imul($96, $fy4$0)|0; //@line 1140 "fastmix.cpp"
  $105 = (($103) + 4096)|0; //@line 1140 "fastmix.cpp"
  $106 = (($105) + ($104))|0; //@line 1140 "fastmix.cpp"
  $107 = (($106) + ($102))|0; //@line 1140 "fastmix.cpp"
  $108 = $107 >> 13; //@line 1140 "fastmix.cpp"
  $109 = HEAP32[$23>>2]|0; //@line 1141 "fastmix.cpp"
  $110 = Math_imul($101, $109)|0; //@line 1141 "fastmix.cpp"
  $111 = HEAP32[$pvol$0>>2]|0; //@line 1141 "fastmix.cpp"
  $112 = (($110) + ($111))|0; //@line 1141 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $112; //@line 1141 "fastmix.cpp"
  $113 = HEAP32[$24>>2]|0; //@line 1141 "fastmix.cpp"
  $114 = Math_imul($108, $113)|0; //@line 1141 "fastmix.cpp"
  $115 = ((($pvol$0)) + 4|0); //@line 1141 "fastmix.cpp"
  $116 = HEAP32[$115>>2]|0; //@line 1141 "fastmix.cpp"
  $117 = (($114) + ($116))|0; //@line 1141 "fastmix.cpp"
  HEAP32[$115>>2] = $117; //@line 1141 "fastmix.cpp"
  $118 = ((($pvol$0)) + 8|0); //@line 1141 "fastmix.cpp"
  $119 = HEAP32[$25>>2]|0; //@line 1142 "fastmix.cpp"
  $120 = (($119) + ($nPos$0))|0; //@line 1142 "fastmix.cpp"
  $121 = ($118>>>0)<($pbufmax>>>0); //@line 1142 "fastmix.cpp"
  if ($121) {
   $fy4$0$phi = $fy3$0;$fy2$0$phi = $fy1$0;$fy1$0 = $101;$fy3$0 = $108;$nPos$0 = $120;$pvol$0 = $118;$fy4$0 = $fy4$0$phi;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $101;$$lcssa11 = $108;$$lcssa12 = $120;$fy1$0$lcssa = $fy1$0;$fy3$0$lcssa = $fy3$0;
   break;
  }
 }
 $122 = $$lcssa12 >> 16; //@line 1142 "fastmix.cpp"
 $123 = HEAP32[$11>>2]|0; //@line 1142 "fastmix.cpp"
 $124 = (($123) + ($122))|0; //@line 1142 "fastmix.cpp"
 HEAP32[$11>>2] = $124; //@line 1142 "fastmix.cpp"
 $125 = $$lcssa12 & 65535; //@line 1142 "fastmix.cpp"
 HEAP32[$8>>2] = $125; //@line 1142 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 1142 "fastmix.cpp"
 HEAP32[$2>>2] = $fy1$0$lcssa; //@line 1142 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa11; //@line 1142 "fastmix.cpp"
 HEAP32[$6>>2] = $fy3$0$lcssa; //@line 1142 "fastmix.cpp"
 return; //@line 1142 "fastmix.cpp"
}
function __Z27FilterMono8BitSplineRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa33 = 0, $$lcssa34 = 0, $$lcssa35 = 0, $$lcssa36 = 0, $$lcssa37 = 0, $$sum = 0, $$sum1 = 0, $$sum2 = 0, $$sum3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 1072 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1072 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 1072 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1072 "fastmix.cpp"
 $4 = ((($pChannel)) + 56|0); //@line 1072 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1072 "fastmix.cpp"
 $6 = ((($pChannel)) + 60|0); //@line 1072 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1072 "fastmix.cpp"
 $8 = ((($pChannel)) + 8|0); //@line 1073 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1073 "fastmix.cpp"
 $10 = HEAP32[$pChannel>>2]|0; //@line 1073 "fastmix.cpp"
 $11 = ((($pChannel)) + 4|0); //@line 1073 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1073 "fastmix.cpp"
 $13 = (($10) + ($12)|0); //@line 1073 "fastmix.cpp"
 $14 = ((($pChannel)) + 36|0); //@line 1073 "fastmix.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 1073 "fastmix.cpp"
 $16 = $15 & 64; //@line 1073 "fastmix.cpp"
 $17 = ($16|0)==(0); //@line 1073 "fastmix.cpp"
 $$ = $17 ? $10 : $13; //@line 1073 "fastmix.cpp"
 $18 = (($12) + -1)|0; //@line 1074 "fastmix.cpp"
 $19 = (($12) + 2)|0; //@line 1074 "fastmix.cpp"
 $20 = (($12) + 1)|0; //@line 1074 "fastmix.cpp"
 $21 = ((($pChannel)) + 72|0); //@line 1075 "fastmix.cpp"
 $22 = ((($pChannel)) + 76|0); //@line 1075 "fastmix.cpp"
 $23 = ((($pChannel)) + 80|0); //@line 1075 "fastmix.cpp"
 $24 = ((($pChannel)) + 28|0); //@line 1076 "fastmix.cpp"
 $25 = ((($pChannel)) + 24|0); //@line 1076 "fastmix.cpp"
 $26 = ((($pChannel)) + 12|0); //@line 1077 "fastmix.cpp"
 $fy1$0 = $5;$fy2$0 = $7;$nPos$0 = $9;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $27 = $nPos$0 >> 16; //@line 1074 "fastmix.cpp"
  $28 = $nPos$0 >>> 4; //@line 1074 "fastmix.cpp"
  $29 = $28 & 4092; //@line 1074 "fastmix.cpp"
  $30 = (1209506 + ($29<<1)|0); //@line 1074 "fastmix.cpp"
  $31 = HEAP16[$30>>1]|0; //@line 1074 "fastmix.cpp"
  $32 = $31 << 16 >> 16; //@line 1074 "fastmix.cpp"
  $$sum = (($18) + ($27))|0; //@line 1074 "fastmix.cpp"
  $33 = (($$) + ($$sum)|0); //@line 1074 "fastmix.cpp"
  $34 = HEAP8[$33>>0]|0; //@line 1074 "fastmix.cpp"
  $35 = $34 << 24 >> 24; //@line 1074 "fastmix.cpp"
  $36 = Math_imul($35, $32)|0; //@line 1074 "fastmix.cpp"
  $37 = $29 | 1; //@line 1074 "fastmix.cpp"
  $38 = (1209506 + ($37<<1)|0); //@line 1074 "fastmix.cpp"
  $39 = HEAP16[$38>>1]|0; //@line 1074 "fastmix.cpp"
  $40 = $39 << 16 >> 16; //@line 1074 "fastmix.cpp"
  $$sum1 = (($27) + ($12))|0; //@line 1074 "fastmix.cpp"
  $41 = (($$) + ($$sum1)|0); //@line 1074 "fastmix.cpp"
  $42 = HEAP8[$41>>0]|0; //@line 1074 "fastmix.cpp"
  $43 = $42 << 24 >> 24; //@line 1074 "fastmix.cpp"
  $44 = Math_imul($43, $40)|0; //@line 1074 "fastmix.cpp"
  $45 = (($44) + ($36))|0; //@line 1074 "fastmix.cpp"
  $46 = $29 | 3; //@line 1074 "fastmix.cpp"
  $47 = (1209506 + ($46<<1)|0); //@line 1074 "fastmix.cpp"
  $48 = HEAP16[$47>>1]|0; //@line 1074 "fastmix.cpp"
  $49 = $48 << 16 >> 16; //@line 1074 "fastmix.cpp"
  $$sum2 = (($19) + ($27))|0; //@line 1074 "fastmix.cpp"
  $50 = (($$) + ($$sum2)|0); //@line 1074 "fastmix.cpp"
  $51 = HEAP8[$50>>0]|0; //@line 1074 "fastmix.cpp"
  $52 = $51 << 24 >> 24; //@line 1074 "fastmix.cpp"
  $53 = Math_imul($52, $49)|0; //@line 1074 "fastmix.cpp"
  $54 = (($45) + ($53))|0; //@line 1074 "fastmix.cpp"
  $55 = $29 | 2; //@line 1074 "fastmix.cpp"
  $56 = (1209506 + ($55<<1)|0); //@line 1074 "fastmix.cpp"
  $57 = HEAP16[$56>>1]|0; //@line 1074 "fastmix.cpp"
  $58 = $57 << 16 >> 16; //@line 1074 "fastmix.cpp"
  $$sum3 = (($20) + ($27))|0; //@line 1074 "fastmix.cpp"
  $59 = (($$) + ($$sum3)|0); //@line 1074 "fastmix.cpp"
  $60 = HEAP8[$59>>0]|0; //@line 1074 "fastmix.cpp"
  $61 = $60 << 24 >> 24; //@line 1074 "fastmix.cpp"
  $62 = Math_imul($61, $58)|0; //@line 1074 "fastmix.cpp"
  $63 = (($54) + ($62))|0; //@line 1074 "fastmix.cpp"
  $64 = $63 >> 6; //@line 1074 "fastmix.cpp"
  $65 = HEAP32[$21>>2]|0; //@line 1075 "fastmix.cpp"
  $66 = Math_imul($64, $65)|0; //@line 1075 "fastmix.cpp"
  $67 = HEAP32[$22>>2]|0; //@line 1075 "fastmix.cpp"
  $68 = Math_imul($67, $fy1$0)|0; //@line 1075 "fastmix.cpp"
  $69 = HEAP32[$23>>2]|0; //@line 1075 "fastmix.cpp"
  $70 = Math_imul($69, $fy2$0)|0; //@line 1075 "fastmix.cpp"
  $71 = (($68) + 4096)|0; //@line 1075 "fastmix.cpp"
  $72 = (($71) + ($70))|0; //@line 1075 "fastmix.cpp"
  $73 = (($72) + ($66))|0; //@line 1075 "fastmix.cpp"
  $74 = $73 >> 13; //@line 1075 "fastmix.cpp"
  $75 = HEAP32[$24>>2]|0; //@line 1076 "fastmix.cpp"
  $76 = (($75) + ($nRampLeftVol$0))|0; //@line 1076 "fastmix.cpp"
  $77 = HEAP32[$25>>2]|0; //@line 1076 "fastmix.cpp"
  $78 = (($77) + ($nRampRightVol$0))|0; //@line 1076 "fastmix.cpp"
  $79 = $78 >> 12; //@line 1076 "fastmix.cpp"
  $80 = Math_imul($74, $79)|0; //@line 1076 "fastmix.cpp"
  $81 = HEAP32[$pvol$0>>2]|0; //@line 1076 "fastmix.cpp"
  $82 = (($80) + ($81))|0; //@line 1076 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $82; //@line 1076 "fastmix.cpp"
  $83 = $76 >> 12; //@line 1076 "fastmix.cpp"
  $84 = Math_imul($74, $83)|0; //@line 1076 "fastmix.cpp"
  $85 = ((($pvol$0)) + 4|0); //@line 1076 "fastmix.cpp"
  $86 = HEAP32[$85>>2]|0; //@line 1076 "fastmix.cpp"
  $87 = (($84) + ($86))|0; //@line 1076 "fastmix.cpp"
  HEAP32[$85>>2] = $87; //@line 1076 "fastmix.cpp"
  $88 = ((($pvol$0)) + 8|0); //@line 1076 "fastmix.cpp"
  $89 = HEAP32[$26>>2]|0; //@line 1077 "fastmix.cpp"
  $90 = (($89) + ($nPos$0))|0; //@line 1077 "fastmix.cpp"
  $91 = ($88>>>0)<($pbufmax>>>0); //@line 1077 "fastmix.cpp"
  if ($91) {
   $fy2$0$phi = $fy1$0;$fy1$0 = $74;$nPos$0 = $90;$nRampLeftVol$0 = $76;$nRampRightVol$0 = $78;$pvol$0 = $88;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $74;$$lcssa33 = $76;$$lcssa34 = $78;$$lcssa35 = $79;$$lcssa36 = $83;$$lcssa37 = $90;$fy1$0$lcssa = $fy1$0;
   break;
  }
 }
 $92 = $$lcssa37 >> 16; //@line 1077 "fastmix.cpp"
 $93 = HEAP32[$11>>2]|0; //@line 1077 "fastmix.cpp"
 $94 = (($93) + ($92))|0; //@line 1077 "fastmix.cpp"
 HEAP32[$11>>2] = $94; //@line 1077 "fastmix.cpp"
 $95 = $$lcssa37 & 65535; //@line 1077 "fastmix.cpp"
 HEAP32[$8>>2] = $95; //@line 1077 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa; //@line 1077 "fastmix.cpp"
 HEAP32[$6>>2] = $fy1$0$lcssa; //@line 1077 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa34; //@line 1077 "fastmix.cpp"
 $96 = ((($pChannel)) + 16|0); //@line 1077 "fastmix.cpp"
 HEAP32[$96>>2] = $$lcssa35; //@line 1077 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa33; //@line 1077 "fastmix.cpp"
 $97 = ((($pChannel)) + 20|0); //@line 1077 "fastmix.cpp"
 HEAP32[$97>>2] = $$lcssa36; //@line 1077 "fastmix.cpp"
 return; //@line 1077 "fastmix.cpp"
}
function __Z28FilterMono16BitSplineRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa26 = 0, $$lcssa27 = 0, $$lcssa28 = 0, $$lcssa29 = 0, $$lcssa30 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 1079 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1079 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 1079 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1079 "fastmix.cpp"
 $4 = ((($pChannel)) + 56|0); //@line 1079 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1079 "fastmix.cpp"
 $6 = ((($pChannel)) + 60|0); //@line 1079 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1079 "fastmix.cpp"
 $8 = ((($pChannel)) + 8|0); //@line 1080 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1080 "fastmix.cpp"
 $10 = HEAP32[$pChannel>>2]|0; //@line 1080 "fastmix.cpp"
 $11 = ((($pChannel)) + 4|0); //@line 1080 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1080 "fastmix.cpp"
 $13 = $12 << 1; //@line 1080 "fastmix.cpp"
 $14 = (($10) + ($13)|0); //@line 1080 "fastmix.cpp"
 $15 = ((($pChannel)) + 36|0); //@line 1080 "fastmix.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1080 "fastmix.cpp"
 $17 = $16 & 64; //@line 1080 "fastmix.cpp"
 $18 = ($17|0)==(0); //@line 1080 "fastmix.cpp"
 $19 = (($14) + ($12<<1)|0); //@line 1080 "fastmix.cpp"
 $$ = $18 ? $14 : $19; //@line 1080 "fastmix.cpp"
 $20 = ((($pChannel)) + 72|0); //@line 1082 "fastmix.cpp"
 $21 = ((($pChannel)) + 76|0); //@line 1082 "fastmix.cpp"
 $22 = ((($pChannel)) + 80|0); //@line 1082 "fastmix.cpp"
 $23 = ((($pChannel)) + 28|0); //@line 1083 "fastmix.cpp"
 $24 = ((($pChannel)) + 24|0); //@line 1083 "fastmix.cpp"
 $25 = ((($pChannel)) + 12|0); //@line 1084 "fastmix.cpp"
 $fy1$0 = $5;$fy2$0 = $7;$nPos$0 = $9;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $26 = $nPos$0 >> 16; //@line 1081 "fastmix.cpp"
  $27 = $nPos$0 >>> 4; //@line 1081 "fastmix.cpp"
  $28 = $27 & 4092; //@line 1081 "fastmix.cpp"
  $29 = (1209506 + ($28<<1)|0); //@line 1081 "fastmix.cpp"
  $30 = HEAP16[$29>>1]|0; //@line 1081 "fastmix.cpp"
  $31 = $30 << 16 >> 16; //@line 1081 "fastmix.cpp"
  $32 = (($26) + -1)|0; //@line 1081 "fastmix.cpp"
  $33 = (($$) + ($32<<1)|0); //@line 1081 "fastmix.cpp"
  $34 = HEAP16[$33>>1]|0; //@line 1081 "fastmix.cpp"
  $35 = $34 << 16 >> 16; //@line 1081 "fastmix.cpp"
  $36 = Math_imul($35, $31)|0; //@line 1081 "fastmix.cpp"
  $37 = $28 | 1; //@line 1081 "fastmix.cpp"
  $38 = (1209506 + ($37<<1)|0); //@line 1081 "fastmix.cpp"
  $39 = HEAP16[$38>>1]|0; //@line 1081 "fastmix.cpp"
  $40 = $39 << 16 >> 16; //@line 1081 "fastmix.cpp"
  $41 = (($$) + ($26<<1)|0); //@line 1081 "fastmix.cpp"
  $42 = HEAP16[$41>>1]|0; //@line 1081 "fastmix.cpp"
  $43 = $42 << 16 >> 16; //@line 1081 "fastmix.cpp"
  $44 = Math_imul($43, $40)|0; //@line 1081 "fastmix.cpp"
  $45 = (($44) + ($36))|0; //@line 1081 "fastmix.cpp"
  $46 = $28 | 3; //@line 1081 "fastmix.cpp"
  $47 = (1209506 + ($46<<1)|0); //@line 1081 "fastmix.cpp"
  $48 = HEAP16[$47>>1]|0; //@line 1081 "fastmix.cpp"
  $49 = $48 << 16 >> 16; //@line 1081 "fastmix.cpp"
  $50 = (($26) + 2)|0; //@line 1081 "fastmix.cpp"
  $51 = (($$) + ($50<<1)|0); //@line 1081 "fastmix.cpp"
  $52 = HEAP16[$51>>1]|0; //@line 1081 "fastmix.cpp"
  $53 = $52 << 16 >> 16; //@line 1081 "fastmix.cpp"
  $54 = Math_imul($53, $49)|0; //@line 1081 "fastmix.cpp"
  $55 = (($45) + ($54))|0; //@line 1081 "fastmix.cpp"
  $56 = $28 | 2; //@line 1081 "fastmix.cpp"
  $57 = (1209506 + ($56<<1)|0); //@line 1081 "fastmix.cpp"
  $58 = HEAP16[$57>>1]|0; //@line 1081 "fastmix.cpp"
  $59 = $58 << 16 >> 16; //@line 1081 "fastmix.cpp"
  $60 = (($26) + 1)|0; //@line 1081 "fastmix.cpp"
  $61 = (($$) + ($60<<1)|0); //@line 1081 "fastmix.cpp"
  $62 = HEAP16[$61>>1]|0; //@line 1081 "fastmix.cpp"
  $63 = $62 << 16 >> 16; //@line 1081 "fastmix.cpp"
  $64 = Math_imul($63, $59)|0; //@line 1081 "fastmix.cpp"
  $65 = (($55) + ($64))|0; //@line 1081 "fastmix.cpp"
  $66 = $65 >> 14; //@line 1081 "fastmix.cpp"
  $67 = HEAP32[$20>>2]|0; //@line 1082 "fastmix.cpp"
  $68 = Math_imul($66, $67)|0; //@line 1082 "fastmix.cpp"
  $69 = HEAP32[$21>>2]|0; //@line 1082 "fastmix.cpp"
  $70 = Math_imul($69, $fy1$0)|0; //@line 1082 "fastmix.cpp"
  $71 = HEAP32[$22>>2]|0; //@line 1082 "fastmix.cpp"
  $72 = Math_imul($71, $fy2$0)|0; //@line 1082 "fastmix.cpp"
  $73 = (($70) + 4096)|0; //@line 1082 "fastmix.cpp"
  $74 = (($73) + ($72))|0; //@line 1082 "fastmix.cpp"
  $75 = (($74) + ($68))|0; //@line 1082 "fastmix.cpp"
  $76 = $75 >> 13; //@line 1082 "fastmix.cpp"
  $77 = HEAP32[$23>>2]|0; //@line 1083 "fastmix.cpp"
  $78 = (($77) + ($nRampLeftVol$0))|0; //@line 1083 "fastmix.cpp"
  $79 = HEAP32[$24>>2]|0; //@line 1083 "fastmix.cpp"
  $80 = (($79) + ($nRampRightVol$0))|0; //@line 1083 "fastmix.cpp"
  $81 = $80 >> 12; //@line 1083 "fastmix.cpp"
  $82 = Math_imul($76, $81)|0; //@line 1083 "fastmix.cpp"
  $83 = HEAP32[$pvol$0>>2]|0; //@line 1083 "fastmix.cpp"
  $84 = (($82) + ($83))|0; //@line 1083 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $84; //@line 1083 "fastmix.cpp"
  $85 = $78 >> 12; //@line 1083 "fastmix.cpp"
  $86 = Math_imul($76, $85)|0; //@line 1083 "fastmix.cpp"
  $87 = ((($pvol$0)) + 4|0); //@line 1083 "fastmix.cpp"
  $88 = HEAP32[$87>>2]|0; //@line 1083 "fastmix.cpp"
  $89 = (($86) + ($88))|0; //@line 1083 "fastmix.cpp"
  HEAP32[$87>>2] = $89; //@line 1083 "fastmix.cpp"
  $90 = ((($pvol$0)) + 8|0); //@line 1083 "fastmix.cpp"
  $91 = HEAP32[$25>>2]|0; //@line 1084 "fastmix.cpp"
  $92 = (($91) + ($nPos$0))|0; //@line 1084 "fastmix.cpp"
  $93 = ($90>>>0)<($pbufmax>>>0); //@line 1084 "fastmix.cpp"
  if ($93) {
   $fy2$0$phi = $fy1$0;$fy1$0 = $76;$nPos$0 = $92;$nRampLeftVol$0 = $78;$nRampRightVol$0 = $80;$pvol$0 = $90;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $76;$$lcssa26 = $78;$$lcssa27 = $80;$$lcssa28 = $81;$$lcssa29 = $85;$$lcssa30 = $92;$fy1$0$lcssa = $fy1$0;
   break;
  }
 }
 $94 = $$lcssa30 >> 16; //@line 1084 "fastmix.cpp"
 $95 = HEAP32[$11>>2]|0; //@line 1084 "fastmix.cpp"
 $96 = (($95) + ($94))|0; //@line 1084 "fastmix.cpp"
 HEAP32[$11>>2] = $96; //@line 1084 "fastmix.cpp"
 $97 = $$lcssa30 & 65535; //@line 1084 "fastmix.cpp"
 HEAP32[$8>>2] = $97; //@line 1084 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa; //@line 1084 "fastmix.cpp"
 HEAP32[$6>>2] = $fy1$0$lcssa; //@line 1084 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa27; //@line 1084 "fastmix.cpp"
 $98 = ((($pChannel)) + 16|0); //@line 1084 "fastmix.cpp"
 HEAP32[$98>>2] = $$lcssa28; //@line 1084 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa26; //@line 1084 "fastmix.cpp"
 $99 = ((($pChannel)) + 20|0); //@line 1084 "fastmix.cpp"
 HEAP32[$99>>2] = $$lcssa29; //@line 1084 "fastmix.cpp"
 return; //@line 1084 "fastmix.cpp"
}
function __Z29FilterStereo8BitSplineRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa46 = 0, $$lcssa47 = 0, $$lcssa48 = 0, $$lcssa49 = 0, $$lcssa50 = 0, $$lcssa51 = 0, $$sum = 0, $$sum1 = 0, $$sum2 = 0, $$sum3 = 0, $$sum4 = 0, $$sum5 = 0, $$sum6 = 0, $$sum7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $fy3$0 = 0, $fy3$0$lcssa = 0, $fy4$0 = 0, $fy4$0$phi = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0;
 var $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 1187 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1187 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 1187 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1187 "fastmix.cpp"
 $4 = ((($pChannel)) + 56|0); //@line 1187 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1187 "fastmix.cpp"
 $6 = ((($pChannel)) + 60|0); //@line 1187 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1187 "fastmix.cpp"
 $8 = ((($pChannel)) + 64|0); //@line 1187 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1187 "fastmix.cpp"
 $10 = ((($pChannel)) + 68|0); //@line 1187 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 1187 "fastmix.cpp"
 $12 = ((($pChannel)) + 8|0); //@line 1188 "fastmix.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 1188 "fastmix.cpp"
 $14 = HEAP32[$pChannel>>2]|0; //@line 1188 "fastmix.cpp"
 $15 = ((($pChannel)) + 4|0); //@line 1188 "fastmix.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1188 "fastmix.cpp"
 $17 = (($14) + ($16)|0); //@line 1188 "fastmix.cpp"
 $18 = ((($pChannel)) + 36|0); //@line 1188 "fastmix.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 1188 "fastmix.cpp"
 $20 = $19 & 64; //@line 1188 "fastmix.cpp"
 $21 = ($20|0)==(0); //@line 1188 "fastmix.cpp"
 $$ = $21 ? $14 : $17; //@line 1188 "fastmix.cpp"
 $22 = (($16) + -2)|0; //@line 1189 "fastmix.cpp"
 $23 = (($16) + 2)|0; //@line 1189 "fastmix.cpp"
 $24 = (($16) + 4)|0; //@line 1189 "fastmix.cpp"
 $25 = (($16) + -1)|0; //@line 1189 "fastmix.cpp"
 $26 = (($16) + 3)|0; //@line 1189 "fastmix.cpp"
 $27 = (($16) + 5)|0; //@line 1189 "fastmix.cpp"
 $28 = ((($pChannel)) + 72|0); //@line 1190 "fastmix.cpp"
 $29 = ((($pChannel)) + 76|0); //@line 1190 "fastmix.cpp"
 $30 = ((($pChannel)) + 80|0); //@line 1190 "fastmix.cpp"
 $31 = ((($pChannel)) + 28|0); //@line 1191 "fastmix.cpp"
 $32 = ((($pChannel)) + 24|0); //@line 1191 "fastmix.cpp"
 $33 = ((($pChannel)) + 12|0); //@line 1192 "fastmix.cpp"
 $fy1$0 = $5;$fy2$0 = $7;$fy3$0 = $9;$fy4$0 = $11;$nPos$0 = $13;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $34 = $nPos$0 >> 16; //@line 1189 "fastmix.cpp"
  $35 = $nPos$0 >>> 4; //@line 1189 "fastmix.cpp"
  $36 = $35 & 4092; //@line 1189 "fastmix.cpp"
  $37 = (1209506 + ($36<<1)|0); //@line 1189 "fastmix.cpp"
  $38 = HEAP16[$37>>1]|0; //@line 1189 "fastmix.cpp"
  $39 = $38 << 16 >> 16; //@line 1189 "fastmix.cpp"
  $40 = $34 << 1; //@line 1189 "fastmix.cpp"
  $$sum = (($22) + ($40))|0; //@line 1189 "fastmix.cpp"
  $41 = (($$) + ($$sum)|0); //@line 1189 "fastmix.cpp"
  $42 = HEAP8[$41>>0]|0; //@line 1189 "fastmix.cpp"
  $43 = $42 << 24 >> 24; //@line 1189 "fastmix.cpp"
  $44 = Math_imul($43, $39)|0; //@line 1189 "fastmix.cpp"
  $45 = $36 | 1; //@line 1189 "fastmix.cpp"
  $46 = (1209506 + ($45<<1)|0); //@line 1189 "fastmix.cpp"
  $47 = HEAP16[$46>>1]|0; //@line 1189 "fastmix.cpp"
  $48 = $47 << 16 >> 16; //@line 1189 "fastmix.cpp"
  $$sum1 = (($40) + ($16))|0; //@line 1189 "fastmix.cpp"
  $49 = (($$) + ($$sum1)|0); //@line 1189 "fastmix.cpp"
  $50 = HEAP8[$49>>0]|0; //@line 1189 "fastmix.cpp"
  $51 = $50 << 24 >> 24; //@line 1189 "fastmix.cpp"
  $52 = Math_imul($51, $48)|0; //@line 1189 "fastmix.cpp"
  $53 = (($52) + ($44))|0; //@line 1189 "fastmix.cpp"
  $54 = $36 | 2; //@line 1189 "fastmix.cpp"
  $55 = (1209506 + ($54<<1)|0); //@line 1189 "fastmix.cpp"
  $56 = HEAP16[$55>>1]|0; //@line 1189 "fastmix.cpp"
  $57 = $56 << 16 >> 16; //@line 1189 "fastmix.cpp"
  $$sum2 = (($23) + ($40))|0; //@line 1189 "fastmix.cpp"
  $58 = (($$) + ($$sum2)|0); //@line 1189 "fastmix.cpp"
  $59 = HEAP8[$58>>0]|0; //@line 1189 "fastmix.cpp"
  $60 = $59 << 24 >> 24; //@line 1189 "fastmix.cpp"
  $61 = Math_imul($60, $57)|0; //@line 1189 "fastmix.cpp"
  $62 = (($53) + ($61))|0; //@line 1189 "fastmix.cpp"
  $63 = $36 | 3; //@line 1189 "fastmix.cpp"
  $64 = (1209506 + ($63<<1)|0); //@line 1189 "fastmix.cpp"
  $65 = HEAP16[$64>>1]|0; //@line 1189 "fastmix.cpp"
  $66 = $65 << 16 >> 16; //@line 1189 "fastmix.cpp"
  $$sum3 = (($24) + ($40))|0; //@line 1189 "fastmix.cpp"
  $67 = (($$) + ($$sum3)|0); //@line 1189 "fastmix.cpp"
  $68 = HEAP8[$67>>0]|0; //@line 1189 "fastmix.cpp"
  $69 = $68 << 24 >> 24; //@line 1189 "fastmix.cpp"
  $70 = Math_imul($69, $66)|0; //@line 1189 "fastmix.cpp"
  $71 = (($62) + ($70))|0; //@line 1189 "fastmix.cpp"
  $72 = $71 >> 6; //@line 1189 "fastmix.cpp"
  $$sum4 = (($25) + ($40))|0; //@line 1189 "fastmix.cpp"
  $73 = (($$) + ($$sum4)|0); //@line 1189 "fastmix.cpp"
  $74 = HEAP8[$73>>0]|0; //@line 1189 "fastmix.cpp"
  $75 = $74 << 24 >> 24; //@line 1189 "fastmix.cpp"
  $76 = Math_imul($75, $39)|0; //@line 1189 "fastmix.cpp"
  $77 = $40 | 1; //@line 1189 "fastmix.cpp"
  $$sum5 = (($77) + ($16))|0; //@line 1189 "fastmix.cpp"
  $78 = (($$) + ($$sum5)|0); //@line 1189 "fastmix.cpp"
  $79 = HEAP8[$78>>0]|0; //@line 1189 "fastmix.cpp"
  $80 = $79 << 24 >> 24; //@line 1189 "fastmix.cpp"
  $81 = Math_imul($80, $48)|0; //@line 1189 "fastmix.cpp"
  $82 = (($81) + ($76))|0; //@line 1189 "fastmix.cpp"
  $$sum6 = (($26) + ($40))|0; //@line 1189 "fastmix.cpp"
  $83 = (($$) + ($$sum6)|0); //@line 1189 "fastmix.cpp"
  $84 = HEAP8[$83>>0]|0; //@line 1189 "fastmix.cpp"
  $85 = $84 << 24 >> 24; //@line 1189 "fastmix.cpp"
  $86 = Math_imul($85, $57)|0; //@line 1189 "fastmix.cpp"
  $87 = (($82) + ($86))|0; //@line 1189 "fastmix.cpp"
  $$sum7 = (($27) + ($40))|0; //@line 1189 "fastmix.cpp"
  $88 = (($$) + ($$sum7)|0); //@line 1189 "fastmix.cpp"
  $89 = HEAP8[$88>>0]|0; //@line 1189 "fastmix.cpp"
  $90 = $89 << 24 >> 24; //@line 1189 "fastmix.cpp"
  $91 = Math_imul($90, $66)|0; //@line 1189 "fastmix.cpp"
  $92 = (($87) + ($91))|0; //@line 1189 "fastmix.cpp"
  $93 = $92 >> 6; //@line 1189 "fastmix.cpp"
  $94 = HEAP32[$28>>2]|0; //@line 1190 "fastmix.cpp"
  $95 = Math_imul($94, $72)|0; //@line 1190 "fastmix.cpp"
  $96 = HEAP32[$29>>2]|0; //@line 1190 "fastmix.cpp"
  $97 = Math_imul($96, $fy1$0)|0; //@line 1190 "fastmix.cpp"
  $98 = HEAP32[$30>>2]|0; //@line 1190 "fastmix.cpp"
  $99 = Math_imul($98, $fy2$0)|0; //@line 1190 "fastmix.cpp"
  $100 = (($95) + 4096)|0; //@line 1190 "fastmix.cpp"
  $101 = (($100) + ($97))|0; //@line 1190 "fastmix.cpp"
  $102 = (($101) + ($99))|0; //@line 1190 "fastmix.cpp"
  $103 = $102 >> 13; //@line 1190 "fastmix.cpp"
  $104 = Math_imul($93, $94)|0; //@line 1190 "fastmix.cpp"
  $105 = Math_imul($96, $fy3$0)|0; //@line 1190 "fastmix.cpp"
  $106 = Math_imul($98, $fy4$0)|0; //@line 1190 "fastmix.cpp"
  $107 = (($105) + 4096)|0; //@line 1190 "fastmix.cpp"
  $108 = (($107) + ($106))|0; //@line 1190 "fastmix.cpp"
  $109 = (($108) + ($104))|0; //@line 1190 "fastmix.cpp"
  $110 = $109 >> 13; //@line 1190 "fastmix.cpp"
  $111 = HEAP32[$31>>2]|0; //@line 1191 "fastmix.cpp"
  $112 = (($111) + ($nRampLeftVol$0))|0; //@line 1191 "fastmix.cpp"
  $113 = HEAP32[$32>>2]|0; //@line 1191 "fastmix.cpp"
  $114 = (($113) + ($nRampRightVol$0))|0; //@line 1191 "fastmix.cpp"
  $115 = $114 >> 12; //@line 1191 "fastmix.cpp"
  $116 = Math_imul($103, $115)|0; //@line 1191 "fastmix.cpp"
  $117 = HEAP32[$pvol$0>>2]|0; //@line 1191 "fastmix.cpp"
  $118 = (($116) + ($117))|0; //@line 1191 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $118; //@line 1191 "fastmix.cpp"
  $119 = $112 >> 12; //@line 1191 "fastmix.cpp"
  $120 = Math_imul($110, $119)|0; //@line 1191 "fastmix.cpp"
  $121 = ((($pvol$0)) + 4|0); //@line 1191 "fastmix.cpp"
  $122 = HEAP32[$121>>2]|0; //@line 1191 "fastmix.cpp"
  $123 = (($120) + ($122))|0; //@line 1191 "fastmix.cpp"
  HEAP32[$121>>2] = $123; //@line 1191 "fastmix.cpp"
  $124 = ((($pvol$0)) + 8|0); //@line 1191 "fastmix.cpp"
  $125 = HEAP32[$33>>2]|0; //@line 1192 "fastmix.cpp"
  $126 = (($125) + ($nPos$0))|0; //@line 1192 "fastmix.cpp"
  $127 = ($124>>>0)<($pbufmax>>>0); //@line 1192 "fastmix.cpp"
  if ($127) {
   $fy4$0$phi = $fy3$0;$fy2$0$phi = $fy1$0;$fy1$0 = $103;$fy3$0 = $110;$nPos$0 = $126;$nRampLeftVol$0 = $112;$nRampRightVol$0 = $114;$pvol$0 = $124;$fy4$0 = $fy4$0$phi;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $103;$$lcssa46 = $110;$$lcssa47 = $112;$$lcssa48 = $114;$$lcssa49 = $115;$$lcssa50 = $119;$$lcssa51 = $126;$fy1$0$lcssa = $fy1$0;$fy3$0$lcssa = $fy3$0;
   break;
  }
 }
 $128 = $$lcssa51 >> 16; //@line 1192 "fastmix.cpp"
 $129 = HEAP32[$15>>2]|0; //@line 1192 "fastmix.cpp"
 $130 = (($129) + ($128))|0; //@line 1192 "fastmix.cpp"
 HEAP32[$15>>2] = $130; //@line 1192 "fastmix.cpp"
 $131 = $$lcssa51 & 65535; //@line 1192 "fastmix.cpp"
 HEAP32[$12>>2] = $131; //@line 1192 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa; //@line 1192 "fastmix.cpp"
 HEAP32[$6>>2] = $fy1$0$lcssa; //@line 1192 "fastmix.cpp"
 HEAP32[$8>>2] = $$lcssa46; //@line 1192 "fastmix.cpp"
 HEAP32[$10>>2] = $fy3$0$lcssa; //@line 1192 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa48; //@line 1192 "fastmix.cpp"
 $132 = ((($pChannel)) + 16|0); //@line 1192 "fastmix.cpp"
 HEAP32[$132>>2] = $$lcssa49; //@line 1192 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa47; //@line 1192 "fastmix.cpp"
 $133 = ((($pChannel)) + 20|0); //@line 1192 "fastmix.cpp"
 HEAP32[$133>>2] = $$lcssa50; //@line 1192 "fastmix.cpp"
 return; //@line 1192 "fastmix.cpp"
}
function __Z30FilterStereo16BitSplineRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$lcssa33 = 0, $$lcssa34 = 0, $$lcssa35 = 0, $$lcssa36 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $fy3$0 = 0, $fy3$0$lcssa = 0, $fy4$0 = 0, $fy4$0$phi = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 1194 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1194 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 1194 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1194 "fastmix.cpp"
 $4 = ((($pChannel)) + 56|0); //@line 1194 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1194 "fastmix.cpp"
 $6 = ((($pChannel)) + 60|0); //@line 1194 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1194 "fastmix.cpp"
 $8 = ((($pChannel)) + 64|0); //@line 1194 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1194 "fastmix.cpp"
 $10 = ((($pChannel)) + 68|0); //@line 1194 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 1194 "fastmix.cpp"
 $12 = ((($pChannel)) + 8|0); //@line 1195 "fastmix.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 1195 "fastmix.cpp"
 $14 = HEAP32[$pChannel>>2]|0; //@line 1195 "fastmix.cpp"
 $15 = ((($pChannel)) + 4|0); //@line 1195 "fastmix.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1195 "fastmix.cpp"
 $17 = $16 << 1; //@line 1195 "fastmix.cpp"
 $18 = (($14) + ($17)|0); //@line 1195 "fastmix.cpp"
 $19 = ((($pChannel)) + 36|0); //@line 1195 "fastmix.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 1195 "fastmix.cpp"
 $21 = $20 & 64; //@line 1195 "fastmix.cpp"
 $22 = ($21|0)==(0); //@line 1195 "fastmix.cpp"
 $23 = (($18) + ($16<<1)|0); //@line 1195 "fastmix.cpp"
 $$ = $22 ? $18 : $23; //@line 1195 "fastmix.cpp"
 $24 = ((($pChannel)) + 72|0); //@line 1197 "fastmix.cpp"
 $25 = ((($pChannel)) + 76|0); //@line 1197 "fastmix.cpp"
 $26 = ((($pChannel)) + 80|0); //@line 1197 "fastmix.cpp"
 $27 = ((($pChannel)) + 28|0); //@line 1198 "fastmix.cpp"
 $28 = ((($pChannel)) + 24|0); //@line 1198 "fastmix.cpp"
 $29 = ((($pChannel)) + 12|0); //@line 1199 "fastmix.cpp"
 $fy1$0 = $5;$fy2$0 = $7;$fy3$0 = $9;$fy4$0 = $11;$nPos$0 = $13;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $30 = $nPos$0 >> 16; //@line 1196 "fastmix.cpp"
  $31 = $nPos$0 >>> 4; //@line 1196 "fastmix.cpp"
  $32 = $31 & 4092; //@line 1196 "fastmix.cpp"
  $33 = (1209506 + ($32<<1)|0); //@line 1196 "fastmix.cpp"
  $34 = HEAP16[$33>>1]|0; //@line 1196 "fastmix.cpp"
  $35 = $34 << 16 >> 16; //@line 1196 "fastmix.cpp"
  $36 = $30 << 1; //@line 1196 "fastmix.cpp"
  $37 = (($36) + -2)|0; //@line 1196 "fastmix.cpp"
  $38 = (($$) + ($37<<1)|0); //@line 1196 "fastmix.cpp"
  $39 = HEAP16[$38>>1]|0; //@line 1196 "fastmix.cpp"
  $40 = $39 << 16 >> 16; //@line 1196 "fastmix.cpp"
  $41 = Math_imul($40, $35)|0; //@line 1196 "fastmix.cpp"
  $42 = $32 | 1; //@line 1196 "fastmix.cpp"
  $43 = (1209506 + ($42<<1)|0); //@line 1196 "fastmix.cpp"
  $44 = HEAP16[$43>>1]|0; //@line 1196 "fastmix.cpp"
  $45 = $44 << 16 >> 16; //@line 1196 "fastmix.cpp"
  $46 = (($$) + ($36<<1)|0); //@line 1196 "fastmix.cpp"
  $47 = HEAP16[$46>>1]|0; //@line 1196 "fastmix.cpp"
  $48 = $47 << 16 >> 16; //@line 1196 "fastmix.cpp"
  $49 = Math_imul($48, $45)|0; //@line 1196 "fastmix.cpp"
  $50 = (($49) + ($41))|0; //@line 1196 "fastmix.cpp"
  $51 = $32 | 2; //@line 1196 "fastmix.cpp"
  $52 = (1209506 + ($51<<1)|0); //@line 1196 "fastmix.cpp"
  $53 = HEAP16[$52>>1]|0; //@line 1196 "fastmix.cpp"
  $54 = $53 << 16 >> 16; //@line 1196 "fastmix.cpp"
  $55 = (($36) + 2)|0; //@line 1196 "fastmix.cpp"
  $56 = (($$) + ($55<<1)|0); //@line 1196 "fastmix.cpp"
  $57 = HEAP16[$56>>1]|0; //@line 1196 "fastmix.cpp"
  $58 = $57 << 16 >> 16; //@line 1196 "fastmix.cpp"
  $59 = Math_imul($58, $54)|0; //@line 1196 "fastmix.cpp"
  $60 = (($50) + ($59))|0; //@line 1196 "fastmix.cpp"
  $61 = $32 | 3; //@line 1196 "fastmix.cpp"
  $62 = (1209506 + ($61<<1)|0); //@line 1196 "fastmix.cpp"
  $63 = HEAP16[$62>>1]|0; //@line 1196 "fastmix.cpp"
  $64 = $63 << 16 >> 16; //@line 1196 "fastmix.cpp"
  $65 = (($36) + 4)|0; //@line 1196 "fastmix.cpp"
  $66 = (($$) + ($65<<1)|0); //@line 1196 "fastmix.cpp"
  $67 = HEAP16[$66>>1]|0; //@line 1196 "fastmix.cpp"
  $68 = $67 << 16 >> 16; //@line 1196 "fastmix.cpp"
  $69 = Math_imul($68, $64)|0; //@line 1196 "fastmix.cpp"
  $70 = (($60) + ($69))|0; //@line 1196 "fastmix.cpp"
  $71 = $70 >> 14; //@line 1196 "fastmix.cpp"
  $72 = (($36) + -1)|0; //@line 1196 "fastmix.cpp"
  $73 = (($$) + ($72<<1)|0); //@line 1196 "fastmix.cpp"
  $74 = HEAP16[$73>>1]|0; //@line 1196 "fastmix.cpp"
  $75 = $74 << 16 >> 16; //@line 1196 "fastmix.cpp"
  $76 = Math_imul($75, $35)|0; //@line 1196 "fastmix.cpp"
  $77 = $36 | 1; //@line 1196 "fastmix.cpp"
  $78 = (($$) + ($77<<1)|0); //@line 1196 "fastmix.cpp"
  $79 = HEAP16[$78>>1]|0; //@line 1196 "fastmix.cpp"
  $80 = $79 << 16 >> 16; //@line 1196 "fastmix.cpp"
  $81 = Math_imul($80, $45)|0; //@line 1196 "fastmix.cpp"
  $82 = (($81) + ($76))|0; //@line 1196 "fastmix.cpp"
  $83 = (($36) + 3)|0; //@line 1196 "fastmix.cpp"
  $84 = (($$) + ($83<<1)|0); //@line 1196 "fastmix.cpp"
  $85 = HEAP16[$84>>1]|0; //@line 1196 "fastmix.cpp"
  $86 = $85 << 16 >> 16; //@line 1196 "fastmix.cpp"
  $87 = Math_imul($86, $54)|0; //@line 1196 "fastmix.cpp"
  $88 = (($82) + ($87))|0; //@line 1196 "fastmix.cpp"
  $89 = (($36) + 5)|0; //@line 1196 "fastmix.cpp"
  $90 = (($$) + ($89<<1)|0); //@line 1196 "fastmix.cpp"
  $91 = HEAP16[$90>>1]|0; //@line 1196 "fastmix.cpp"
  $92 = $91 << 16 >> 16; //@line 1196 "fastmix.cpp"
  $93 = Math_imul($92, $64)|0; //@line 1196 "fastmix.cpp"
  $94 = (($88) + ($93))|0; //@line 1196 "fastmix.cpp"
  $95 = $94 >> 14; //@line 1196 "fastmix.cpp"
  $96 = HEAP32[$24>>2]|0; //@line 1197 "fastmix.cpp"
  $97 = Math_imul($96, $71)|0; //@line 1197 "fastmix.cpp"
  $98 = HEAP32[$25>>2]|0; //@line 1197 "fastmix.cpp"
  $99 = Math_imul($98, $fy1$0)|0; //@line 1197 "fastmix.cpp"
  $100 = HEAP32[$26>>2]|0; //@line 1197 "fastmix.cpp"
  $101 = Math_imul($100, $fy2$0)|0; //@line 1197 "fastmix.cpp"
  $102 = (($97) + 4096)|0; //@line 1197 "fastmix.cpp"
  $103 = (($102) + ($99))|0; //@line 1197 "fastmix.cpp"
  $104 = (($103) + ($101))|0; //@line 1197 "fastmix.cpp"
  $105 = $104 >> 13; //@line 1197 "fastmix.cpp"
  $106 = Math_imul($95, $96)|0; //@line 1197 "fastmix.cpp"
  $107 = Math_imul($98, $fy3$0)|0; //@line 1197 "fastmix.cpp"
  $108 = Math_imul($100, $fy4$0)|0; //@line 1197 "fastmix.cpp"
  $109 = (($107) + 4096)|0; //@line 1197 "fastmix.cpp"
  $110 = (($109) + ($108))|0; //@line 1197 "fastmix.cpp"
  $111 = (($110) + ($106))|0; //@line 1197 "fastmix.cpp"
  $112 = $111 >> 13; //@line 1197 "fastmix.cpp"
  $113 = HEAP32[$27>>2]|0; //@line 1198 "fastmix.cpp"
  $114 = (($113) + ($nRampLeftVol$0))|0; //@line 1198 "fastmix.cpp"
  $115 = HEAP32[$28>>2]|0; //@line 1198 "fastmix.cpp"
  $116 = (($115) + ($nRampRightVol$0))|0; //@line 1198 "fastmix.cpp"
  $117 = $116 >> 12; //@line 1198 "fastmix.cpp"
  $118 = Math_imul($105, $117)|0; //@line 1198 "fastmix.cpp"
  $119 = HEAP32[$pvol$0>>2]|0; //@line 1198 "fastmix.cpp"
  $120 = (($118) + ($119))|0; //@line 1198 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $120; //@line 1198 "fastmix.cpp"
  $121 = $114 >> 12; //@line 1198 "fastmix.cpp"
  $122 = Math_imul($112, $121)|0; //@line 1198 "fastmix.cpp"
  $123 = ((($pvol$0)) + 4|0); //@line 1198 "fastmix.cpp"
  $124 = HEAP32[$123>>2]|0; //@line 1198 "fastmix.cpp"
  $125 = (($122) + ($124))|0; //@line 1198 "fastmix.cpp"
  HEAP32[$123>>2] = $125; //@line 1198 "fastmix.cpp"
  $126 = ((($pvol$0)) + 8|0); //@line 1198 "fastmix.cpp"
  $127 = HEAP32[$29>>2]|0; //@line 1199 "fastmix.cpp"
  $128 = (($127) + ($nPos$0))|0; //@line 1199 "fastmix.cpp"
  $129 = ($126>>>0)<($pbufmax>>>0); //@line 1199 "fastmix.cpp"
  if ($129) {
   $fy4$0$phi = $fy3$0;$fy2$0$phi = $fy1$0;$fy1$0 = $105;$fy3$0 = $112;$nPos$0 = $128;$nRampLeftVol$0 = $114;$nRampRightVol$0 = $116;$pvol$0 = $126;$fy4$0 = $fy4$0$phi;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $105;$$lcssa31 = $112;$$lcssa32 = $114;$$lcssa33 = $116;$$lcssa34 = $117;$$lcssa35 = $121;$$lcssa36 = $128;$fy1$0$lcssa = $fy1$0;$fy3$0$lcssa = $fy3$0;
   break;
  }
 }
 $130 = $$lcssa36 >> 16; //@line 1199 "fastmix.cpp"
 $131 = HEAP32[$15>>2]|0; //@line 1199 "fastmix.cpp"
 $132 = (($131) + ($130))|0; //@line 1199 "fastmix.cpp"
 HEAP32[$15>>2] = $132; //@line 1199 "fastmix.cpp"
 $133 = $$lcssa36 & 65535; //@line 1199 "fastmix.cpp"
 HEAP32[$12>>2] = $133; //@line 1199 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa; //@line 1199 "fastmix.cpp"
 HEAP32[$6>>2] = $fy1$0$lcssa; //@line 1199 "fastmix.cpp"
 HEAP32[$8>>2] = $$lcssa31; //@line 1199 "fastmix.cpp"
 HEAP32[$10>>2] = $fy3$0$lcssa; //@line 1199 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa33; //@line 1199 "fastmix.cpp"
 $134 = ((($pChannel)) + 16|0); //@line 1199 "fastmix.cpp"
 HEAP32[$134>>2] = $$lcssa34; //@line 1199 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa32; //@line 1199 "fastmix.cpp"
 $135 = ((($pChannel)) + 20|0); //@line 1199 "fastmix.cpp"
 HEAP32[$135>>2] = $$lcssa35; //@line 1199 "fastmix.cpp"
 return; //@line 1199 "fastmix.cpp"
}
function __Z20Mono8BitFirFilterMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$sum = 0, $$sum1 = 0, $$sum2 = 0, $$sum3 = 0, $$sum4 = 0, $$sum5 = 0, $$sum6 = 0, $$sum7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 714 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 714 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 714 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 714 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 714 "fastmix.cpp"
 $5 = (($2) + ($4)|0); //@line 714 "fastmix.cpp"
 $6 = ((($pChannel)) + 36|0); //@line 714 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 714 "fastmix.cpp"
 $8 = $7 & 64; //@line 714 "fastmix.cpp"
 $9 = ($8|0)==(0); //@line 714 "fastmix.cpp"
 $$ = $9 ? $2 : $5; //@line 714 "fastmix.cpp"
 $10 = (($4) + -3)|0; //@line 715 "fastmix.cpp"
 $11 = (($4) + -2)|0; //@line 715 "fastmix.cpp"
 $12 = (($4) + -1)|0; //@line 715 "fastmix.cpp"
 $13 = (($4) + 1)|0; //@line 715 "fastmix.cpp"
 $14 = (($4) + 2)|0; //@line 715 "fastmix.cpp"
 $15 = (($4) + 3)|0; //@line 715 "fastmix.cpp"
 $16 = (($4) + 4)|0; //@line 715 "fastmix.cpp"
 $17 = ((($pChannel)) + 16|0); //@line 716 "fastmix.cpp"
 $18 = ((($pChannel)) + 20|0); //@line 716 "fastmix.cpp"
 $19 = ((($pChannel)) + 12|0); //@line 717 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $20 = $nPos$0 >> 16; //@line 715 "fastmix.cpp"
  $21 = $nPos$0 & 65535; //@line 715 "fastmix.cpp"
  $22 = (($21) + 16)|0; //@line 715 "fastmix.cpp"
  $23 = $22 >>> 2; //@line 715 "fastmix.cpp"
  $24 = $23 & 32760; //@line 715 "fastmix.cpp"
  $25 = (1217698 + ($24<<1)|0); //@line 715 "fastmix.cpp"
  $26 = HEAP16[$25>>1]|0; //@line 715 "fastmix.cpp"
  $27 = $26 << 16 >> 16; //@line 715 "fastmix.cpp"
  $$sum = (($10) + ($20))|0; //@line 715 "fastmix.cpp"
  $28 = (($$) + ($$sum)|0); //@line 715 "fastmix.cpp"
  $29 = HEAP8[$28>>0]|0; //@line 715 "fastmix.cpp"
  $30 = $29 << 24 >> 24; //@line 715 "fastmix.cpp"
  $31 = Math_imul($30, $27)|0; //@line 715 "fastmix.cpp"
  $32 = $24 | 1; //@line 715 "fastmix.cpp"
  $33 = (1217698 + ($32<<1)|0); //@line 715 "fastmix.cpp"
  $34 = HEAP16[$33>>1]|0; //@line 715 "fastmix.cpp"
  $35 = $34 << 16 >> 16; //@line 715 "fastmix.cpp"
  $$sum1 = (($11) + ($20))|0; //@line 715 "fastmix.cpp"
  $36 = (($$) + ($$sum1)|0); //@line 715 "fastmix.cpp"
  $37 = HEAP8[$36>>0]|0; //@line 715 "fastmix.cpp"
  $38 = $37 << 24 >> 24; //@line 715 "fastmix.cpp"
  $39 = Math_imul($38, $35)|0; //@line 715 "fastmix.cpp"
  $40 = (($39) + ($31))|0; //@line 715 "fastmix.cpp"
  $41 = $24 | 2; //@line 715 "fastmix.cpp"
  $42 = (1217698 + ($41<<1)|0); //@line 715 "fastmix.cpp"
  $43 = HEAP16[$42>>1]|0; //@line 715 "fastmix.cpp"
  $44 = $43 << 16 >> 16; //@line 715 "fastmix.cpp"
  $$sum2 = (($12) + ($20))|0; //@line 715 "fastmix.cpp"
  $45 = (($$) + ($$sum2)|0); //@line 715 "fastmix.cpp"
  $46 = HEAP8[$45>>0]|0; //@line 715 "fastmix.cpp"
  $47 = $46 << 24 >> 24; //@line 715 "fastmix.cpp"
  $48 = Math_imul($47, $44)|0; //@line 715 "fastmix.cpp"
  $49 = (($40) + ($48))|0; //@line 715 "fastmix.cpp"
  $50 = $24 | 3; //@line 715 "fastmix.cpp"
  $51 = (1217698 + ($50<<1)|0); //@line 715 "fastmix.cpp"
  $52 = HEAP16[$51>>1]|0; //@line 715 "fastmix.cpp"
  $53 = $52 << 16 >> 16; //@line 715 "fastmix.cpp"
  $$sum3 = (($20) + ($4))|0; //@line 715 "fastmix.cpp"
  $54 = (($$) + ($$sum3)|0); //@line 715 "fastmix.cpp"
  $55 = HEAP8[$54>>0]|0; //@line 715 "fastmix.cpp"
  $56 = $55 << 24 >> 24; //@line 715 "fastmix.cpp"
  $57 = Math_imul($56, $53)|0; //@line 715 "fastmix.cpp"
  $58 = (($49) + ($57))|0; //@line 715 "fastmix.cpp"
  $59 = $24 | 4; //@line 715 "fastmix.cpp"
  $60 = (1217698 + ($59<<1)|0); //@line 715 "fastmix.cpp"
  $61 = HEAP16[$60>>1]|0; //@line 715 "fastmix.cpp"
  $62 = $61 << 16 >> 16; //@line 715 "fastmix.cpp"
  $$sum4 = (($13) + ($20))|0; //@line 715 "fastmix.cpp"
  $63 = (($$) + ($$sum4)|0); //@line 715 "fastmix.cpp"
  $64 = HEAP8[$63>>0]|0; //@line 715 "fastmix.cpp"
  $65 = $64 << 24 >> 24; //@line 715 "fastmix.cpp"
  $66 = Math_imul($65, $62)|0; //@line 715 "fastmix.cpp"
  $67 = (($58) + ($66))|0; //@line 715 "fastmix.cpp"
  $68 = $24 | 5; //@line 715 "fastmix.cpp"
  $69 = (1217698 + ($68<<1)|0); //@line 715 "fastmix.cpp"
  $70 = HEAP16[$69>>1]|0; //@line 715 "fastmix.cpp"
  $71 = $70 << 16 >> 16; //@line 715 "fastmix.cpp"
  $$sum5 = (($14) + ($20))|0; //@line 715 "fastmix.cpp"
  $72 = (($$) + ($$sum5)|0); //@line 715 "fastmix.cpp"
  $73 = HEAP8[$72>>0]|0; //@line 715 "fastmix.cpp"
  $74 = $73 << 24 >> 24; //@line 715 "fastmix.cpp"
  $75 = Math_imul($74, $71)|0; //@line 715 "fastmix.cpp"
  $76 = (($67) + ($75))|0; //@line 715 "fastmix.cpp"
  $77 = $24 | 6; //@line 715 "fastmix.cpp"
  $78 = (1217698 + ($77<<1)|0); //@line 715 "fastmix.cpp"
  $79 = HEAP16[$78>>1]|0; //@line 715 "fastmix.cpp"
  $80 = $79 << 16 >> 16; //@line 715 "fastmix.cpp"
  $$sum6 = (($15) + ($20))|0; //@line 715 "fastmix.cpp"
  $81 = (($$) + ($$sum6)|0); //@line 715 "fastmix.cpp"
  $82 = HEAP8[$81>>0]|0; //@line 715 "fastmix.cpp"
  $83 = $82 << 24 >> 24; //@line 715 "fastmix.cpp"
  $84 = Math_imul($83, $80)|0; //@line 715 "fastmix.cpp"
  $85 = (($76) + ($84))|0; //@line 715 "fastmix.cpp"
  $86 = $23 | 7; //@line 715 "fastmix.cpp"
  $87 = (1217698 + ($86<<1)|0); //@line 715 "fastmix.cpp"
  $88 = HEAP16[$87>>1]|0; //@line 715 "fastmix.cpp"
  $89 = $88 << 16 >> 16; //@line 715 "fastmix.cpp"
  $$sum7 = (($16) + ($20))|0; //@line 715 "fastmix.cpp"
  $90 = (($$) + ($$sum7)|0); //@line 715 "fastmix.cpp"
  $91 = HEAP8[$90>>0]|0; //@line 715 "fastmix.cpp"
  $92 = $91 << 24 >> 24; //@line 715 "fastmix.cpp"
  $93 = Math_imul($92, $89)|0; //@line 715 "fastmix.cpp"
  $94 = (($85) + ($93))|0; //@line 715 "fastmix.cpp"
  $95 = $94 >> 7; //@line 715 "fastmix.cpp"
  $96 = HEAP32[$17>>2]|0; //@line 716 "fastmix.cpp"
  $97 = Math_imul($95, $96)|0; //@line 716 "fastmix.cpp"
  $98 = HEAP32[$pvol$0>>2]|0; //@line 716 "fastmix.cpp"
  $99 = (($97) + ($98))|0; //@line 716 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $99; //@line 716 "fastmix.cpp"
  $100 = HEAP32[$18>>2]|0; //@line 716 "fastmix.cpp"
  $101 = Math_imul($95, $100)|0; //@line 716 "fastmix.cpp"
  $102 = ((($pvol$0)) + 4|0); //@line 716 "fastmix.cpp"
  $103 = HEAP32[$102>>2]|0; //@line 716 "fastmix.cpp"
  $104 = (($101) + ($103))|0; //@line 716 "fastmix.cpp"
  HEAP32[$102>>2] = $104; //@line 716 "fastmix.cpp"
  $105 = ((($pvol$0)) + 8|0); //@line 716 "fastmix.cpp"
  $106 = HEAP32[$19>>2]|0; //@line 717 "fastmix.cpp"
  $107 = (($106) + ($nPos$0))|0; //@line 717 "fastmix.cpp"
  $108 = ($105>>>0)<($pbufmax>>>0); //@line 717 "fastmix.cpp"
  if ($108) {
   $nPos$0 = $107;$pvol$0 = $105;
  } else {
   $$lcssa = $107;
   break;
  }
 }
 $109 = $$lcssa >> 16; //@line 717 "fastmix.cpp"
 $110 = HEAP32[$3>>2]|0; //@line 717 "fastmix.cpp"
 $111 = (($110) + ($109))|0; //@line 717 "fastmix.cpp"
 HEAP32[$3>>2] = $111; //@line 717 "fastmix.cpp"
 $112 = $$lcssa & 65535; //@line 717 "fastmix.cpp"
 HEAP32[$0>>2] = $112; //@line 717 "fastmix.cpp"
 return; //@line 717 "fastmix.cpp"
}
function __Z21Mono16BitFirFilterMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $nPos$0 = 0;
 var $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 720 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 720 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 720 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 720 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 720 "fastmix.cpp"
 $5 = $4 << 1; //@line 720 "fastmix.cpp"
 $6 = (($2) + ($5)|0); //@line 720 "fastmix.cpp"
 $7 = ((($pChannel)) + 36|0); //@line 720 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 720 "fastmix.cpp"
 $9 = $8 & 64; //@line 720 "fastmix.cpp"
 $10 = ($9|0)==(0); //@line 720 "fastmix.cpp"
 $11 = (($6) + ($4<<1)|0); //@line 720 "fastmix.cpp"
 $$ = $10 ? $6 : $11; //@line 720 "fastmix.cpp"
 $12 = ((($pChannel)) + 16|0); //@line 722 "fastmix.cpp"
 $13 = ((($pChannel)) + 20|0); //@line 722 "fastmix.cpp"
 $14 = ((($pChannel)) + 12|0); //@line 723 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $15 = $nPos$0 >> 16; //@line 721 "fastmix.cpp"
  $16 = $nPos$0 & 65535; //@line 721 "fastmix.cpp"
  $17 = (($16) + 16)|0; //@line 721 "fastmix.cpp"
  $18 = $17 >>> 2; //@line 721 "fastmix.cpp"
  $19 = $18 & 32760; //@line 721 "fastmix.cpp"
  $20 = (1217698 + ($19<<1)|0); //@line 721 "fastmix.cpp"
  $21 = HEAP16[$20>>1]|0; //@line 721 "fastmix.cpp"
  $22 = $21 << 16 >> 16; //@line 721 "fastmix.cpp"
  $23 = (($15) + -3)|0; //@line 721 "fastmix.cpp"
  $24 = (($$) + ($23<<1)|0); //@line 721 "fastmix.cpp"
  $25 = HEAP16[$24>>1]|0; //@line 721 "fastmix.cpp"
  $26 = $25 << 16 >> 16; //@line 721 "fastmix.cpp"
  $27 = Math_imul($26, $22)|0; //@line 721 "fastmix.cpp"
  $28 = $19 | 1; //@line 721 "fastmix.cpp"
  $29 = (1217698 + ($28<<1)|0); //@line 721 "fastmix.cpp"
  $30 = HEAP16[$29>>1]|0; //@line 721 "fastmix.cpp"
  $31 = $30 << 16 >> 16; //@line 721 "fastmix.cpp"
  $32 = (($15) + -2)|0; //@line 721 "fastmix.cpp"
  $33 = (($$) + ($32<<1)|0); //@line 721 "fastmix.cpp"
  $34 = HEAP16[$33>>1]|0; //@line 721 "fastmix.cpp"
  $35 = $34 << 16 >> 16; //@line 721 "fastmix.cpp"
  $36 = Math_imul($35, $31)|0; //@line 721 "fastmix.cpp"
  $37 = (($36) + ($27))|0; //@line 721 "fastmix.cpp"
  $38 = $19 | 2; //@line 721 "fastmix.cpp"
  $39 = (1217698 + ($38<<1)|0); //@line 721 "fastmix.cpp"
  $40 = HEAP16[$39>>1]|0; //@line 721 "fastmix.cpp"
  $41 = $40 << 16 >> 16; //@line 721 "fastmix.cpp"
  $42 = (($15) + -1)|0; //@line 721 "fastmix.cpp"
  $43 = (($$) + ($42<<1)|0); //@line 721 "fastmix.cpp"
  $44 = HEAP16[$43>>1]|0; //@line 721 "fastmix.cpp"
  $45 = $44 << 16 >> 16; //@line 721 "fastmix.cpp"
  $46 = Math_imul($45, $41)|0; //@line 721 "fastmix.cpp"
  $47 = (($37) + ($46))|0; //@line 721 "fastmix.cpp"
  $48 = $19 | 3; //@line 721 "fastmix.cpp"
  $49 = (1217698 + ($48<<1)|0); //@line 721 "fastmix.cpp"
  $50 = HEAP16[$49>>1]|0; //@line 721 "fastmix.cpp"
  $51 = $50 << 16 >> 16; //@line 721 "fastmix.cpp"
  $52 = (($$) + ($15<<1)|0); //@line 721 "fastmix.cpp"
  $53 = HEAP16[$52>>1]|0; //@line 721 "fastmix.cpp"
  $54 = $53 << 16 >> 16; //@line 721 "fastmix.cpp"
  $55 = Math_imul($54, $51)|0; //@line 721 "fastmix.cpp"
  $56 = (($47) + ($55))|0; //@line 721 "fastmix.cpp"
  $57 = $19 | 4; //@line 721 "fastmix.cpp"
  $58 = (1217698 + ($57<<1)|0); //@line 721 "fastmix.cpp"
  $59 = HEAP16[$58>>1]|0; //@line 721 "fastmix.cpp"
  $60 = $59 << 16 >> 16; //@line 721 "fastmix.cpp"
  $61 = (($15) + 1)|0; //@line 721 "fastmix.cpp"
  $62 = (($$) + ($61<<1)|0); //@line 721 "fastmix.cpp"
  $63 = HEAP16[$62>>1]|0; //@line 721 "fastmix.cpp"
  $64 = $63 << 16 >> 16; //@line 721 "fastmix.cpp"
  $65 = Math_imul($64, $60)|0; //@line 721 "fastmix.cpp"
  $66 = $19 | 5; //@line 721 "fastmix.cpp"
  $67 = (1217698 + ($66<<1)|0); //@line 721 "fastmix.cpp"
  $68 = HEAP16[$67>>1]|0; //@line 721 "fastmix.cpp"
  $69 = $68 << 16 >> 16; //@line 721 "fastmix.cpp"
  $70 = (($15) + 2)|0; //@line 721 "fastmix.cpp"
  $71 = (($$) + ($70<<1)|0); //@line 721 "fastmix.cpp"
  $72 = HEAP16[$71>>1]|0; //@line 721 "fastmix.cpp"
  $73 = $72 << 16 >> 16; //@line 721 "fastmix.cpp"
  $74 = Math_imul($73, $69)|0; //@line 721 "fastmix.cpp"
  $75 = (($74) + ($65))|0; //@line 721 "fastmix.cpp"
  $76 = $19 | 6; //@line 721 "fastmix.cpp"
  $77 = (1217698 + ($76<<1)|0); //@line 721 "fastmix.cpp"
  $78 = HEAP16[$77>>1]|0; //@line 721 "fastmix.cpp"
  $79 = $78 << 16 >> 16; //@line 721 "fastmix.cpp"
  $80 = (($15) + 3)|0; //@line 721 "fastmix.cpp"
  $81 = (($$) + ($80<<1)|0); //@line 721 "fastmix.cpp"
  $82 = HEAP16[$81>>1]|0; //@line 721 "fastmix.cpp"
  $83 = $82 << 16 >> 16; //@line 721 "fastmix.cpp"
  $84 = Math_imul($83, $79)|0; //@line 721 "fastmix.cpp"
  $85 = (($75) + ($84))|0; //@line 721 "fastmix.cpp"
  $86 = $18 | 7; //@line 721 "fastmix.cpp"
  $87 = (1217698 + ($86<<1)|0); //@line 721 "fastmix.cpp"
  $88 = HEAP16[$87>>1]|0; //@line 721 "fastmix.cpp"
  $89 = $88 << 16 >> 16; //@line 721 "fastmix.cpp"
  $90 = (($15) + 4)|0; //@line 721 "fastmix.cpp"
  $91 = (($$) + ($90<<1)|0); //@line 721 "fastmix.cpp"
  $92 = HEAP16[$91>>1]|0; //@line 721 "fastmix.cpp"
  $93 = $92 << 16 >> 16; //@line 721 "fastmix.cpp"
  $94 = Math_imul($93, $89)|0; //@line 721 "fastmix.cpp"
  $95 = (($85) + ($94))|0; //@line 721 "fastmix.cpp"
  $96 = $56 >> 1; //@line 721 "fastmix.cpp"
  $97 = $95 >> 1; //@line 721 "fastmix.cpp"
  $98 = (($97) + ($96))|0; //@line 721 "fastmix.cpp"
  $99 = $98 >> 14; //@line 721 "fastmix.cpp"
  $100 = HEAP32[$12>>2]|0; //@line 722 "fastmix.cpp"
  $101 = Math_imul($99, $100)|0; //@line 722 "fastmix.cpp"
  $102 = HEAP32[$pvol$0>>2]|0; //@line 722 "fastmix.cpp"
  $103 = (($101) + ($102))|0; //@line 722 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $103; //@line 722 "fastmix.cpp"
  $104 = HEAP32[$13>>2]|0; //@line 722 "fastmix.cpp"
  $105 = Math_imul($99, $104)|0; //@line 722 "fastmix.cpp"
  $106 = ((($pvol$0)) + 4|0); //@line 722 "fastmix.cpp"
  $107 = HEAP32[$106>>2]|0; //@line 722 "fastmix.cpp"
  $108 = (($105) + ($107))|0; //@line 722 "fastmix.cpp"
  HEAP32[$106>>2] = $108; //@line 722 "fastmix.cpp"
  $109 = ((($pvol$0)) + 8|0); //@line 722 "fastmix.cpp"
  $110 = HEAP32[$14>>2]|0; //@line 723 "fastmix.cpp"
  $111 = (($110) + ($nPos$0))|0; //@line 723 "fastmix.cpp"
  $112 = ($109>>>0)<($pbufmax>>>0); //@line 723 "fastmix.cpp"
  if ($112) {
   $nPos$0 = $111;$pvol$0 = $109;
  } else {
   $$lcssa = $111;
   break;
  }
 }
 $113 = $$lcssa >> 16; //@line 723 "fastmix.cpp"
 $114 = HEAP32[$3>>2]|0; //@line 723 "fastmix.cpp"
 $115 = (($114) + ($113))|0; //@line 723 "fastmix.cpp"
 HEAP32[$3>>2] = $115; //@line 723 "fastmix.cpp"
 $116 = $$lcssa & 65535; //@line 723 "fastmix.cpp"
 HEAP32[$0>>2] = $116; //@line 723 "fastmix.cpp"
 return; //@line 723 "fastmix.cpp"
}
function __Z22Stereo8BitFirFilterMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$sum = 0, $$sum1 = 0, $$sum10 = 0, $$sum11 = 0, $$sum12 = 0, $$sum13 = 0, $$sum14 = 0, $$sum15 = 0, $$sum2 = 0, $$sum3 = 0, $$sum4 = 0, $$sum5 = 0, $$sum6 = 0, $$sum7 = 0, $$sum8 = 0, $$sum9 = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 918 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 918 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 918 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 918 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 918 "fastmix.cpp"
 $5 = (($2) + ($4)|0); //@line 918 "fastmix.cpp"
 $6 = ((($pChannel)) + 36|0); //@line 918 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 918 "fastmix.cpp"
 $8 = $7 & 64; //@line 918 "fastmix.cpp"
 $9 = ($8|0)==(0); //@line 918 "fastmix.cpp"
 $$ = $9 ? $2 : $5; //@line 918 "fastmix.cpp"
 $10 = (($4) + -6)|0; //@line 919 "fastmix.cpp"
 $11 = (($4) + -4)|0; //@line 919 "fastmix.cpp"
 $12 = (($4) + -2)|0; //@line 919 "fastmix.cpp"
 $13 = (($4) + 2)|0; //@line 919 "fastmix.cpp"
 $14 = (($4) + 4)|0; //@line 919 "fastmix.cpp"
 $15 = (($4) + 6)|0; //@line 919 "fastmix.cpp"
 $16 = (($4) + 8)|0; //@line 919 "fastmix.cpp"
 $17 = (($4) + -5)|0; //@line 919 "fastmix.cpp"
 $18 = (($4) + -3)|0; //@line 919 "fastmix.cpp"
 $19 = (($4) + -1)|0; //@line 919 "fastmix.cpp"
 $20 = (($4) + 3)|0; //@line 919 "fastmix.cpp"
 $21 = (($4) + 5)|0; //@line 919 "fastmix.cpp"
 $22 = (($4) + 7)|0; //@line 919 "fastmix.cpp"
 $23 = (($4) + 9)|0; //@line 919 "fastmix.cpp"
 $24 = ((($pChannel)) + 16|0); //@line 920 "fastmix.cpp"
 $25 = ((($pChannel)) + 20|0); //@line 920 "fastmix.cpp"
 $26 = ((($pChannel)) + 12|0); //@line 921 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $27 = $nPos$0 >> 16; //@line 919 "fastmix.cpp"
  $28 = $nPos$0 & 65535; //@line 919 "fastmix.cpp"
  $29 = (($28) + 16)|0; //@line 919 "fastmix.cpp"
  $30 = $29 >>> 2; //@line 919 "fastmix.cpp"
  $31 = $30 & 32760; //@line 919 "fastmix.cpp"
  $32 = (1217698 + ($31<<1)|0); //@line 919 "fastmix.cpp"
  $33 = HEAP16[$32>>1]|0; //@line 919 "fastmix.cpp"
  $34 = $33 << 16 >> 16; //@line 919 "fastmix.cpp"
  $35 = $27 << 1; //@line 919 "fastmix.cpp"
  $$sum = (($10) + ($35))|0; //@line 919 "fastmix.cpp"
  $36 = (($$) + ($$sum)|0); //@line 919 "fastmix.cpp"
  $37 = HEAP8[$36>>0]|0; //@line 919 "fastmix.cpp"
  $38 = $37 << 24 >> 24; //@line 919 "fastmix.cpp"
  $39 = Math_imul($38, $34)|0; //@line 919 "fastmix.cpp"
  $40 = $31 | 1; //@line 919 "fastmix.cpp"
  $41 = (1217698 + ($40<<1)|0); //@line 919 "fastmix.cpp"
  $42 = HEAP16[$41>>1]|0; //@line 919 "fastmix.cpp"
  $43 = $42 << 16 >> 16; //@line 919 "fastmix.cpp"
  $$sum1 = (($11) + ($35))|0; //@line 919 "fastmix.cpp"
  $44 = (($$) + ($$sum1)|0); //@line 919 "fastmix.cpp"
  $45 = HEAP8[$44>>0]|0; //@line 919 "fastmix.cpp"
  $46 = $45 << 24 >> 24; //@line 919 "fastmix.cpp"
  $47 = Math_imul($46, $43)|0; //@line 919 "fastmix.cpp"
  $48 = (($47) + ($39))|0; //@line 919 "fastmix.cpp"
  $49 = $31 | 2; //@line 919 "fastmix.cpp"
  $50 = (1217698 + ($49<<1)|0); //@line 919 "fastmix.cpp"
  $51 = HEAP16[$50>>1]|0; //@line 919 "fastmix.cpp"
  $52 = $51 << 16 >> 16; //@line 919 "fastmix.cpp"
  $$sum2 = (($12) + ($35))|0; //@line 919 "fastmix.cpp"
  $53 = (($$) + ($$sum2)|0); //@line 919 "fastmix.cpp"
  $54 = HEAP8[$53>>0]|0; //@line 919 "fastmix.cpp"
  $55 = $54 << 24 >> 24; //@line 919 "fastmix.cpp"
  $56 = Math_imul($55, $52)|0; //@line 919 "fastmix.cpp"
  $57 = (($48) + ($56))|0; //@line 919 "fastmix.cpp"
  $58 = $31 | 3; //@line 919 "fastmix.cpp"
  $59 = (1217698 + ($58<<1)|0); //@line 919 "fastmix.cpp"
  $60 = HEAP16[$59>>1]|0; //@line 919 "fastmix.cpp"
  $61 = $60 << 16 >> 16; //@line 919 "fastmix.cpp"
  $$sum3 = (($35) + ($4))|0; //@line 919 "fastmix.cpp"
  $62 = (($$) + ($$sum3)|0); //@line 919 "fastmix.cpp"
  $63 = HEAP8[$62>>0]|0; //@line 919 "fastmix.cpp"
  $64 = $63 << 24 >> 24; //@line 919 "fastmix.cpp"
  $65 = Math_imul($64, $61)|0; //@line 919 "fastmix.cpp"
  $66 = (($57) + ($65))|0; //@line 919 "fastmix.cpp"
  $67 = $31 | 4; //@line 919 "fastmix.cpp"
  $68 = (1217698 + ($67<<1)|0); //@line 919 "fastmix.cpp"
  $69 = HEAP16[$68>>1]|0; //@line 919 "fastmix.cpp"
  $70 = $69 << 16 >> 16; //@line 919 "fastmix.cpp"
  $$sum4 = (($13) + ($35))|0; //@line 919 "fastmix.cpp"
  $71 = (($$) + ($$sum4)|0); //@line 919 "fastmix.cpp"
  $72 = HEAP8[$71>>0]|0; //@line 919 "fastmix.cpp"
  $73 = $72 << 24 >> 24; //@line 919 "fastmix.cpp"
  $74 = Math_imul($73, $70)|0; //@line 919 "fastmix.cpp"
  $75 = (($66) + ($74))|0; //@line 919 "fastmix.cpp"
  $76 = $31 | 5; //@line 919 "fastmix.cpp"
  $77 = (1217698 + ($76<<1)|0); //@line 919 "fastmix.cpp"
  $78 = HEAP16[$77>>1]|0; //@line 919 "fastmix.cpp"
  $79 = $78 << 16 >> 16; //@line 919 "fastmix.cpp"
  $$sum5 = (($14) + ($35))|0; //@line 919 "fastmix.cpp"
  $80 = (($$) + ($$sum5)|0); //@line 919 "fastmix.cpp"
  $81 = HEAP8[$80>>0]|0; //@line 919 "fastmix.cpp"
  $82 = $81 << 24 >> 24; //@line 919 "fastmix.cpp"
  $83 = Math_imul($82, $79)|0; //@line 919 "fastmix.cpp"
  $84 = (($75) + ($83))|0; //@line 919 "fastmix.cpp"
  $85 = $31 | 6; //@line 919 "fastmix.cpp"
  $86 = (1217698 + ($85<<1)|0); //@line 919 "fastmix.cpp"
  $87 = HEAP16[$86>>1]|0; //@line 919 "fastmix.cpp"
  $88 = $87 << 16 >> 16; //@line 919 "fastmix.cpp"
  $$sum6 = (($15) + ($35))|0; //@line 919 "fastmix.cpp"
  $89 = (($$) + ($$sum6)|0); //@line 919 "fastmix.cpp"
  $90 = HEAP8[$89>>0]|0; //@line 919 "fastmix.cpp"
  $91 = $90 << 24 >> 24; //@line 919 "fastmix.cpp"
  $92 = Math_imul($91, $88)|0; //@line 919 "fastmix.cpp"
  $93 = (($84) + ($92))|0; //@line 919 "fastmix.cpp"
  $94 = $30 | 7; //@line 919 "fastmix.cpp"
  $95 = (1217698 + ($94<<1)|0); //@line 919 "fastmix.cpp"
  $96 = HEAP16[$95>>1]|0; //@line 919 "fastmix.cpp"
  $97 = $96 << 16 >> 16; //@line 919 "fastmix.cpp"
  $$sum7 = (($16) + ($35))|0; //@line 919 "fastmix.cpp"
  $98 = (($$) + ($$sum7)|0); //@line 919 "fastmix.cpp"
  $99 = HEAP8[$98>>0]|0; //@line 919 "fastmix.cpp"
  $100 = $99 << 24 >> 24; //@line 919 "fastmix.cpp"
  $101 = Math_imul($100, $97)|0; //@line 919 "fastmix.cpp"
  $102 = (($93) + ($101))|0; //@line 919 "fastmix.cpp"
  $103 = $102 >> 7; //@line 919 "fastmix.cpp"
  $$sum8 = (($17) + ($35))|0; //@line 919 "fastmix.cpp"
  $104 = (($$) + ($$sum8)|0); //@line 919 "fastmix.cpp"
  $105 = HEAP8[$104>>0]|0; //@line 919 "fastmix.cpp"
  $106 = $105 << 24 >> 24; //@line 919 "fastmix.cpp"
  $107 = Math_imul($106, $34)|0; //@line 919 "fastmix.cpp"
  $$sum9 = (($18) + ($35))|0; //@line 919 "fastmix.cpp"
  $108 = (($$) + ($$sum9)|0); //@line 919 "fastmix.cpp"
  $109 = HEAP8[$108>>0]|0; //@line 919 "fastmix.cpp"
  $110 = $109 << 24 >> 24; //@line 919 "fastmix.cpp"
  $111 = Math_imul($110, $43)|0; //@line 919 "fastmix.cpp"
  $112 = (($111) + ($107))|0; //@line 919 "fastmix.cpp"
  $$sum10 = (($19) + ($35))|0; //@line 919 "fastmix.cpp"
  $113 = (($$) + ($$sum10)|0); //@line 919 "fastmix.cpp"
  $114 = HEAP8[$113>>0]|0; //@line 919 "fastmix.cpp"
  $115 = $114 << 24 >> 24; //@line 919 "fastmix.cpp"
  $116 = Math_imul($115, $52)|0; //@line 919 "fastmix.cpp"
  $117 = (($112) + ($116))|0; //@line 919 "fastmix.cpp"
  $118 = $35 | 1; //@line 919 "fastmix.cpp"
  $$sum11 = (($118) + ($4))|0; //@line 919 "fastmix.cpp"
  $119 = (($$) + ($$sum11)|0); //@line 919 "fastmix.cpp"
  $120 = HEAP8[$119>>0]|0; //@line 919 "fastmix.cpp"
  $121 = $120 << 24 >> 24; //@line 919 "fastmix.cpp"
  $122 = Math_imul($121, $61)|0; //@line 919 "fastmix.cpp"
  $123 = (($117) + ($122))|0; //@line 919 "fastmix.cpp"
  $$sum12 = (($20) + ($35))|0; //@line 919 "fastmix.cpp"
  $124 = (($$) + ($$sum12)|0); //@line 919 "fastmix.cpp"
  $125 = HEAP8[$124>>0]|0; //@line 919 "fastmix.cpp"
  $126 = $125 << 24 >> 24; //@line 919 "fastmix.cpp"
  $127 = Math_imul($126, $70)|0; //@line 919 "fastmix.cpp"
  $128 = (($123) + ($127))|0; //@line 919 "fastmix.cpp"
  $$sum13 = (($21) + ($35))|0; //@line 919 "fastmix.cpp"
  $129 = (($$) + ($$sum13)|0); //@line 919 "fastmix.cpp"
  $130 = HEAP8[$129>>0]|0; //@line 919 "fastmix.cpp"
  $131 = $130 << 24 >> 24; //@line 919 "fastmix.cpp"
  $132 = Math_imul($131, $79)|0; //@line 919 "fastmix.cpp"
  $133 = (($128) + ($132))|0; //@line 919 "fastmix.cpp"
  $$sum14 = (($22) + ($35))|0; //@line 919 "fastmix.cpp"
  $134 = (($$) + ($$sum14)|0); //@line 919 "fastmix.cpp"
  $135 = HEAP8[$134>>0]|0; //@line 919 "fastmix.cpp"
  $136 = $135 << 24 >> 24; //@line 919 "fastmix.cpp"
  $137 = Math_imul($136, $88)|0; //@line 919 "fastmix.cpp"
  $138 = (($133) + ($137))|0; //@line 919 "fastmix.cpp"
  $$sum15 = (($23) + ($35))|0; //@line 919 "fastmix.cpp"
  $139 = (($$) + ($$sum15)|0); //@line 919 "fastmix.cpp"
  $140 = HEAP8[$139>>0]|0; //@line 919 "fastmix.cpp"
  $141 = $140 << 24 >> 24; //@line 919 "fastmix.cpp"
  $142 = Math_imul($141, $97)|0; //@line 919 "fastmix.cpp"
  $143 = (($138) + ($142))|0; //@line 919 "fastmix.cpp"
  $144 = $143 >> 7; //@line 919 "fastmix.cpp"
  $145 = HEAP32[$24>>2]|0; //@line 920 "fastmix.cpp"
  $146 = Math_imul($145, $103)|0; //@line 920 "fastmix.cpp"
  $147 = HEAP32[$pvol$0>>2]|0; //@line 920 "fastmix.cpp"
  $148 = (($147) + ($146))|0; //@line 920 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $148; //@line 920 "fastmix.cpp"
  $149 = HEAP32[$25>>2]|0; //@line 920 "fastmix.cpp"
  $150 = Math_imul($144, $149)|0; //@line 920 "fastmix.cpp"
  $151 = ((($pvol$0)) + 4|0); //@line 920 "fastmix.cpp"
  $152 = HEAP32[$151>>2]|0; //@line 920 "fastmix.cpp"
  $153 = (($150) + ($152))|0; //@line 920 "fastmix.cpp"
  HEAP32[$151>>2] = $153; //@line 920 "fastmix.cpp"
  $154 = ((($pvol$0)) + 8|0); //@line 920 "fastmix.cpp"
  $155 = HEAP32[$26>>2]|0; //@line 921 "fastmix.cpp"
  $156 = (($155) + ($nPos$0))|0; //@line 921 "fastmix.cpp"
  $157 = ($154>>>0)<($pbufmax>>>0); //@line 921 "fastmix.cpp"
  if ($157) {
   $nPos$0 = $156;$pvol$0 = $154;
  } else {
   $$lcssa = $156;
   break;
  }
 }
 $158 = $$lcssa >> 16; //@line 921 "fastmix.cpp"
 $159 = HEAP32[$3>>2]|0; //@line 921 "fastmix.cpp"
 $160 = (($159) + ($158))|0; //@line 921 "fastmix.cpp"
 HEAP32[$3>>2] = $160; //@line 921 "fastmix.cpp"
 $161 = $$lcssa & 65535; //@line 921 "fastmix.cpp"
 HEAP32[$0>>2] = $161; //@line 921 "fastmix.cpp"
 return; //@line 921 "fastmix.cpp"
}
function __Z23Stereo16BitFirFilterMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 924 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 924 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 924 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 924 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 924 "fastmix.cpp"
 $5 = $4 << 1; //@line 924 "fastmix.cpp"
 $6 = (($2) + ($5)|0); //@line 924 "fastmix.cpp"
 $7 = ((($pChannel)) + 36|0); //@line 924 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 924 "fastmix.cpp"
 $9 = $8 & 64; //@line 924 "fastmix.cpp"
 $10 = ($9|0)==(0); //@line 924 "fastmix.cpp"
 $11 = (($6) + ($4<<1)|0); //@line 924 "fastmix.cpp"
 $$ = $10 ? $6 : $11; //@line 924 "fastmix.cpp"
 $12 = ((($pChannel)) + 16|0); //@line 926 "fastmix.cpp"
 $13 = ((($pChannel)) + 20|0); //@line 926 "fastmix.cpp"
 $14 = ((($pChannel)) + 12|0); //@line 927 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $15 = $nPos$0 >> 16; //@line 925 "fastmix.cpp"
  $16 = $nPos$0 & 65535; //@line 925 "fastmix.cpp"
  $17 = (($16) + 16)|0; //@line 925 "fastmix.cpp"
  $18 = $17 >>> 2; //@line 925 "fastmix.cpp"
  $19 = $18 & 32760; //@line 925 "fastmix.cpp"
  $20 = (1217698 + ($19<<1)|0); //@line 925 "fastmix.cpp"
  $21 = HEAP16[$20>>1]|0; //@line 925 "fastmix.cpp"
  $22 = $21 << 16 >> 16; //@line 925 "fastmix.cpp"
  $23 = $15 << 1; //@line 925 "fastmix.cpp"
  $24 = (($23) + -6)|0; //@line 925 "fastmix.cpp"
  $25 = (($$) + ($24<<1)|0); //@line 925 "fastmix.cpp"
  $26 = HEAP16[$25>>1]|0; //@line 925 "fastmix.cpp"
  $27 = $26 << 16 >> 16; //@line 925 "fastmix.cpp"
  $28 = Math_imul($27, $22)|0; //@line 925 "fastmix.cpp"
  $29 = $19 | 1; //@line 925 "fastmix.cpp"
  $30 = (1217698 + ($29<<1)|0); //@line 925 "fastmix.cpp"
  $31 = HEAP16[$30>>1]|0; //@line 925 "fastmix.cpp"
  $32 = $31 << 16 >> 16; //@line 925 "fastmix.cpp"
  $33 = (($23) + -4)|0; //@line 925 "fastmix.cpp"
  $34 = (($$) + ($33<<1)|0); //@line 925 "fastmix.cpp"
  $35 = HEAP16[$34>>1]|0; //@line 925 "fastmix.cpp"
  $36 = $35 << 16 >> 16; //@line 925 "fastmix.cpp"
  $37 = Math_imul($36, $32)|0; //@line 925 "fastmix.cpp"
  $38 = (($37) + ($28))|0; //@line 925 "fastmix.cpp"
  $39 = $19 | 2; //@line 925 "fastmix.cpp"
  $40 = (1217698 + ($39<<1)|0); //@line 925 "fastmix.cpp"
  $41 = HEAP16[$40>>1]|0; //@line 925 "fastmix.cpp"
  $42 = $41 << 16 >> 16; //@line 925 "fastmix.cpp"
  $43 = (($23) + -2)|0; //@line 925 "fastmix.cpp"
  $44 = (($$) + ($43<<1)|0); //@line 925 "fastmix.cpp"
  $45 = HEAP16[$44>>1]|0; //@line 925 "fastmix.cpp"
  $46 = $45 << 16 >> 16; //@line 925 "fastmix.cpp"
  $47 = Math_imul($46, $42)|0; //@line 925 "fastmix.cpp"
  $48 = (($38) + ($47))|0; //@line 925 "fastmix.cpp"
  $49 = $19 | 3; //@line 925 "fastmix.cpp"
  $50 = (1217698 + ($49<<1)|0); //@line 925 "fastmix.cpp"
  $51 = HEAP16[$50>>1]|0; //@line 925 "fastmix.cpp"
  $52 = $51 << 16 >> 16; //@line 925 "fastmix.cpp"
  $53 = (($$) + ($23<<1)|0); //@line 925 "fastmix.cpp"
  $54 = HEAP16[$53>>1]|0; //@line 925 "fastmix.cpp"
  $55 = $54 << 16 >> 16; //@line 925 "fastmix.cpp"
  $56 = Math_imul($55, $52)|0; //@line 925 "fastmix.cpp"
  $57 = (($48) + ($56))|0; //@line 925 "fastmix.cpp"
  $58 = $19 | 4; //@line 925 "fastmix.cpp"
  $59 = (1217698 + ($58<<1)|0); //@line 925 "fastmix.cpp"
  $60 = HEAP16[$59>>1]|0; //@line 925 "fastmix.cpp"
  $61 = $60 << 16 >> 16; //@line 925 "fastmix.cpp"
  $62 = (($23) + 2)|0; //@line 925 "fastmix.cpp"
  $63 = (($$) + ($62<<1)|0); //@line 925 "fastmix.cpp"
  $64 = HEAP16[$63>>1]|0; //@line 925 "fastmix.cpp"
  $65 = $64 << 16 >> 16; //@line 925 "fastmix.cpp"
  $66 = Math_imul($65, $61)|0; //@line 925 "fastmix.cpp"
  $67 = $19 | 5; //@line 925 "fastmix.cpp"
  $68 = (1217698 + ($67<<1)|0); //@line 925 "fastmix.cpp"
  $69 = HEAP16[$68>>1]|0; //@line 925 "fastmix.cpp"
  $70 = $69 << 16 >> 16; //@line 925 "fastmix.cpp"
  $71 = (($23) + 4)|0; //@line 925 "fastmix.cpp"
  $72 = (($$) + ($71<<1)|0); //@line 925 "fastmix.cpp"
  $73 = HEAP16[$72>>1]|0; //@line 925 "fastmix.cpp"
  $74 = $73 << 16 >> 16; //@line 925 "fastmix.cpp"
  $75 = Math_imul($74, $70)|0; //@line 925 "fastmix.cpp"
  $76 = (($75) + ($66))|0; //@line 925 "fastmix.cpp"
  $77 = $19 | 6; //@line 925 "fastmix.cpp"
  $78 = (1217698 + ($77<<1)|0); //@line 925 "fastmix.cpp"
  $79 = HEAP16[$78>>1]|0; //@line 925 "fastmix.cpp"
  $80 = $79 << 16 >> 16; //@line 925 "fastmix.cpp"
  $81 = (($23) + 6)|0; //@line 925 "fastmix.cpp"
  $82 = (($$) + ($81<<1)|0); //@line 925 "fastmix.cpp"
  $83 = HEAP16[$82>>1]|0; //@line 925 "fastmix.cpp"
  $84 = $83 << 16 >> 16; //@line 925 "fastmix.cpp"
  $85 = Math_imul($84, $80)|0; //@line 925 "fastmix.cpp"
  $86 = (($76) + ($85))|0; //@line 925 "fastmix.cpp"
  $87 = $18 | 7; //@line 925 "fastmix.cpp"
  $88 = (1217698 + ($87<<1)|0); //@line 925 "fastmix.cpp"
  $89 = HEAP16[$88>>1]|0; //@line 925 "fastmix.cpp"
  $90 = $89 << 16 >> 16; //@line 925 "fastmix.cpp"
  $91 = (($23) + 8)|0; //@line 925 "fastmix.cpp"
  $92 = (($$) + ($91<<1)|0); //@line 925 "fastmix.cpp"
  $93 = HEAP16[$92>>1]|0; //@line 925 "fastmix.cpp"
  $94 = $93 << 16 >> 16; //@line 925 "fastmix.cpp"
  $95 = Math_imul($94, $90)|0; //@line 925 "fastmix.cpp"
  $96 = (($86) + ($95))|0; //@line 925 "fastmix.cpp"
  $97 = $57 >> 1; //@line 925 "fastmix.cpp"
  $98 = $96 >> 1; //@line 925 "fastmix.cpp"
  $99 = (($98) + ($97))|0; //@line 925 "fastmix.cpp"
  $100 = $99 >> 14; //@line 925 "fastmix.cpp"
  $101 = (($23) + -5)|0; //@line 925 "fastmix.cpp"
  $102 = (($$) + ($101<<1)|0); //@line 925 "fastmix.cpp"
  $103 = HEAP16[$102>>1]|0; //@line 925 "fastmix.cpp"
  $104 = $103 << 16 >> 16; //@line 925 "fastmix.cpp"
  $105 = Math_imul($104, $22)|0; //@line 925 "fastmix.cpp"
  $106 = (($23) + -3)|0; //@line 925 "fastmix.cpp"
  $107 = (($$) + ($106<<1)|0); //@line 925 "fastmix.cpp"
  $108 = HEAP16[$107>>1]|0; //@line 925 "fastmix.cpp"
  $109 = $108 << 16 >> 16; //@line 925 "fastmix.cpp"
  $110 = Math_imul($109, $32)|0; //@line 925 "fastmix.cpp"
  $111 = (($110) + ($105))|0; //@line 925 "fastmix.cpp"
  $112 = (($23) + -1)|0; //@line 925 "fastmix.cpp"
  $113 = (($$) + ($112<<1)|0); //@line 925 "fastmix.cpp"
  $114 = HEAP16[$113>>1]|0; //@line 925 "fastmix.cpp"
  $115 = $114 << 16 >> 16; //@line 925 "fastmix.cpp"
  $116 = Math_imul($115, $42)|0; //@line 925 "fastmix.cpp"
  $117 = (($111) + ($116))|0; //@line 925 "fastmix.cpp"
  $118 = $23 | 1; //@line 925 "fastmix.cpp"
  $119 = (($$) + ($118<<1)|0); //@line 925 "fastmix.cpp"
  $120 = HEAP16[$119>>1]|0; //@line 925 "fastmix.cpp"
  $121 = $120 << 16 >> 16; //@line 925 "fastmix.cpp"
  $122 = Math_imul($121, $52)|0; //@line 925 "fastmix.cpp"
  $123 = (($117) + ($122))|0; //@line 925 "fastmix.cpp"
  $124 = (($23) + 3)|0; //@line 925 "fastmix.cpp"
  $125 = (($$) + ($124<<1)|0); //@line 925 "fastmix.cpp"
  $126 = HEAP16[$125>>1]|0; //@line 925 "fastmix.cpp"
  $127 = $126 << 16 >> 16; //@line 925 "fastmix.cpp"
  $128 = Math_imul($127, $61)|0; //@line 925 "fastmix.cpp"
  $129 = (($23) + 5)|0; //@line 925 "fastmix.cpp"
  $130 = (($$) + ($129<<1)|0); //@line 925 "fastmix.cpp"
  $131 = HEAP16[$130>>1]|0; //@line 925 "fastmix.cpp"
  $132 = $131 << 16 >> 16; //@line 925 "fastmix.cpp"
  $133 = Math_imul($132, $70)|0; //@line 925 "fastmix.cpp"
  $134 = (($133) + ($128))|0; //@line 925 "fastmix.cpp"
  $135 = (($23) + 7)|0; //@line 925 "fastmix.cpp"
  $136 = (($$) + ($135<<1)|0); //@line 925 "fastmix.cpp"
  $137 = HEAP16[$136>>1]|0; //@line 925 "fastmix.cpp"
  $138 = $137 << 16 >> 16; //@line 925 "fastmix.cpp"
  $139 = Math_imul($138, $80)|0; //@line 925 "fastmix.cpp"
  $140 = (($134) + ($139))|0; //@line 925 "fastmix.cpp"
  $141 = (($23) + 9)|0; //@line 925 "fastmix.cpp"
  $142 = (($$) + ($141<<1)|0); //@line 925 "fastmix.cpp"
  $143 = HEAP16[$142>>1]|0; //@line 925 "fastmix.cpp"
  $144 = $143 << 16 >> 16; //@line 925 "fastmix.cpp"
  $145 = Math_imul($144, $90)|0; //@line 925 "fastmix.cpp"
  $146 = (($140) + ($145))|0; //@line 925 "fastmix.cpp"
  $147 = $123 >> 1; //@line 925 "fastmix.cpp"
  $148 = $146 >> 1; //@line 925 "fastmix.cpp"
  $149 = (($148) + ($147))|0; //@line 925 "fastmix.cpp"
  $150 = $149 >> 14; //@line 925 "fastmix.cpp"
  $151 = HEAP32[$12>>2]|0; //@line 926 "fastmix.cpp"
  $152 = Math_imul($151, $100)|0; //@line 926 "fastmix.cpp"
  $153 = HEAP32[$pvol$0>>2]|0; //@line 926 "fastmix.cpp"
  $154 = (($153) + ($152))|0; //@line 926 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $154; //@line 926 "fastmix.cpp"
  $155 = HEAP32[$13>>2]|0; //@line 926 "fastmix.cpp"
  $156 = Math_imul($150, $155)|0; //@line 926 "fastmix.cpp"
  $157 = ((($pvol$0)) + 4|0); //@line 926 "fastmix.cpp"
  $158 = HEAP32[$157>>2]|0; //@line 926 "fastmix.cpp"
  $159 = (($156) + ($158))|0; //@line 926 "fastmix.cpp"
  HEAP32[$157>>2] = $159; //@line 926 "fastmix.cpp"
  $160 = ((($pvol$0)) + 8|0); //@line 926 "fastmix.cpp"
  $161 = HEAP32[$14>>2]|0; //@line 927 "fastmix.cpp"
  $162 = (($161) + ($nPos$0))|0; //@line 927 "fastmix.cpp"
  $163 = ($160>>>0)<($pbufmax>>>0); //@line 927 "fastmix.cpp"
  if ($163) {
   $nPos$0 = $162;$pvol$0 = $160;
  } else {
   $$lcssa = $162;
   break;
  }
 }
 $164 = $$lcssa >> 16; //@line 927 "fastmix.cpp"
 $165 = HEAP32[$3>>2]|0; //@line 927 "fastmix.cpp"
 $166 = (($165) + ($164))|0; //@line 927 "fastmix.cpp"
 HEAP32[$3>>2] = $166; //@line 927 "fastmix.cpp"
 $167 = $$lcssa & 65535; //@line 927 "fastmix.cpp"
 HEAP32[$0>>2] = $167; //@line 927 "fastmix.cpp"
 return; //@line 927 "fastmix.cpp"
}
function __Z24Mono8BitFirFilterRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa36 = 0, $$lcssa37 = 0, $$lcssa38 = 0, $$lcssa39 = 0, $$sum = 0, $$sum1 = 0, $$sum2 = 0, $$sum3 = 0, $$sum4 = 0, $$sum5 = 0, $$sum6 = 0, $$sum7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0;
 var $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 763 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 763 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 763 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 763 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 764 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 764 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 764 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 764 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 764 "fastmix.cpp"
 $9 = (($6) + ($8)|0); //@line 764 "fastmix.cpp"
 $10 = ((($pChannel)) + 36|0); //@line 764 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 764 "fastmix.cpp"
 $12 = $11 & 64; //@line 764 "fastmix.cpp"
 $13 = ($12|0)==(0); //@line 764 "fastmix.cpp"
 $$ = $13 ? $6 : $9; //@line 764 "fastmix.cpp"
 $14 = (($8) + -3)|0; //@line 765 "fastmix.cpp"
 $15 = (($8) + -2)|0; //@line 765 "fastmix.cpp"
 $16 = (($8) + -1)|0; //@line 765 "fastmix.cpp"
 $17 = (($8) + 1)|0; //@line 765 "fastmix.cpp"
 $18 = (($8) + 2)|0; //@line 765 "fastmix.cpp"
 $19 = (($8) + 3)|0; //@line 765 "fastmix.cpp"
 $20 = (($8) + 4)|0; //@line 765 "fastmix.cpp"
 $21 = ((($pChannel)) + 28|0); //@line 766 "fastmix.cpp"
 $22 = ((($pChannel)) + 24|0); //@line 766 "fastmix.cpp"
 $23 = ((($pChannel)) + 12|0); //@line 767 "fastmix.cpp"
 $nPos$0 = $5;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $24 = $nPos$0 >> 16; //@line 765 "fastmix.cpp"
  $25 = $nPos$0 & 65535; //@line 765 "fastmix.cpp"
  $26 = (($25) + 16)|0; //@line 765 "fastmix.cpp"
  $27 = $26 >>> 2; //@line 765 "fastmix.cpp"
  $28 = $27 & 32760; //@line 765 "fastmix.cpp"
  $29 = (1217698 + ($28<<1)|0); //@line 765 "fastmix.cpp"
  $30 = HEAP16[$29>>1]|0; //@line 765 "fastmix.cpp"
  $31 = $30 << 16 >> 16; //@line 765 "fastmix.cpp"
  $$sum = (($14) + ($24))|0; //@line 765 "fastmix.cpp"
  $32 = (($$) + ($$sum)|0); //@line 765 "fastmix.cpp"
  $33 = HEAP8[$32>>0]|0; //@line 765 "fastmix.cpp"
  $34 = $33 << 24 >> 24; //@line 765 "fastmix.cpp"
  $35 = Math_imul($34, $31)|0; //@line 765 "fastmix.cpp"
  $36 = $28 | 1; //@line 765 "fastmix.cpp"
  $37 = (1217698 + ($36<<1)|0); //@line 765 "fastmix.cpp"
  $38 = HEAP16[$37>>1]|0; //@line 765 "fastmix.cpp"
  $39 = $38 << 16 >> 16; //@line 765 "fastmix.cpp"
  $$sum1 = (($15) + ($24))|0; //@line 765 "fastmix.cpp"
  $40 = (($$) + ($$sum1)|0); //@line 765 "fastmix.cpp"
  $41 = HEAP8[$40>>0]|0; //@line 765 "fastmix.cpp"
  $42 = $41 << 24 >> 24; //@line 765 "fastmix.cpp"
  $43 = Math_imul($42, $39)|0; //@line 765 "fastmix.cpp"
  $44 = (($43) + ($35))|0; //@line 765 "fastmix.cpp"
  $45 = $28 | 2; //@line 765 "fastmix.cpp"
  $46 = (1217698 + ($45<<1)|0); //@line 765 "fastmix.cpp"
  $47 = HEAP16[$46>>1]|0; //@line 765 "fastmix.cpp"
  $48 = $47 << 16 >> 16; //@line 765 "fastmix.cpp"
  $$sum2 = (($16) + ($24))|0; //@line 765 "fastmix.cpp"
  $49 = (($$) + ($$sum2)|0); //@line 765 "fastmix.cpp"
  $50 = HEAP8[$49>>0]|0; //@line 765 "fastmix.cpp"
  $51 = $50 << 24 >> 24; //@line 765 "fastmix.cpp"
  $52 = Math_imul($51, $48)|0; //@line 765 "fastmix.cpp"
  $53 = (($44) + ($52))|0; //@line 765 "fastmix.cpp"
  $54 = $28 | 3; //@line 765 "fastmix.cpp"
  $55 = (1217698 + ($54<<1)|0); //@line 765 "fastmix.cpp"
  $56 = HEAP16[$55>>1]|0; //@line 765 "fastmix.cpp"
  $57 = $56 << 16 >> 16; //@line 765 "fastmix.cpp"
  $$sum3 = (($24) + ($8))|0; //@line 765 "fastmix.cpp"
  $58 = (($$) + ($$sum3)|0); //@line 765 "fastmix.cpp"
  $59 = HEAP8[$58>>0]|0; //@line 765 "fastmix.cpp"
  $60 = $59 << 24 >> 24; //@line 765 "fastmix.cpp"
  $61 = Math_imul($60, $57)|0; //@line 765 "fastmix.cpp"
  $62 = (($53) + ($61))|0; //@line 765 "fastmix.cpp"
  $63 = $28 | 4; //@line 765 "fastmix.cpp"
  $64 = (1217698 + ($63<<1)|0); //@line 765 "fastmix.cpp"
  $65 = HEAP16[$64>>1]|0; //@line 765 "fastmix.cpp"
  $66 = $65 << 16 >> 16; //@line 765 "fastmix.cpp"
  $$sum4 = (($17) + ($24))|0; //@line 765 "fastmix.cpp"
  $67 = (($$) + ($$sum4)|0); //@line 765 "fastmix.cpp"
  $68 = HEAP8[$67>>0]|0; //@line 765 "fastmix.cpp"
  $69 = $68 << 24 >> 24; //@line 765 "fastmix.cpp"
  $70 = Math_imul($69, $66)|0; //@line 765 "fastmix.cpp"
  $71 = (($62) + ($70))|0; //@line 765 "fastmix.cpp"
  $72 = $28 | 5; //@line 765 "fastmix.cpp"
  $73 = (1217698 + ($72<<1)|0); //@line 765 "fastmix.cpp"
  $74 = HEAP16[$73>>1]|0; //@line 765 "fastmix.cpp"
  $75 = $74 << 16 >> 16; //@line 765 "fastmix.cpp"
  $$sum5 = (($18) + ($24))|0; //@line 765 "fastmix.cpp"
  $76 = (($$) + ($$sum5)|0); //@line 765 "fastmix.cpp"
  $77 = HEAP8[$76>>0]|0; //@line 765 "fastmix.cpp"
  $78 = $77 << 24 >> 24; //@line 765 "fastmix.cpp"
  $79 = Math_imul($78, $75)|0; //@line 765 "fastmix.cpp"
  $80 = (($71) + ($79))|0; //@line 765 "fastmix.cpp"
  $81 = $28 | 6; //@line 765 "fastmix.cpp"
  $82 = (1217698 + ($81<<1)|0); //@line 765 "fastmix.cpp"
  $83 = HEAP16[$82>>1]|0; //@line 765 "fastmix.cpp"
  $84 = $83 << 16 >> 16; //@line 765 "fastmix.cpp"
  $$sum6 = (($19) + ($24))|0; //@line 765 "fastmix.cpp"
  $85 = (($$) + ($$sum6)|0); //@line 765 "fastmix.cpp"
  $86 = HEAP8[$85>>0]|0; //@line 765 "fastmix.cpp"
  $87 = $86 << 24 >> 24; //@line 765 "fastmix.cpp"
  $88 = Math_imul($87, $84)|0; //@line 765 "fastmix.cpp"
  $89 = (($80) + ($88))|0; //@line 765 "fastmix.cpp"
  $90 = $27 | 7; //@line 765 "fastmix.cpp"
  $91 = (1217698 + ($90<<1)|0); //@line 765 "fastmix.cpp"
  $92 = HEAP16[$91>>1]|0; //@line 765 "fastmix.cpp"
  $93 = $92 << 16 >> 16; //@line 765 "fastmix.cpp"
  $$sum7 = (($20) + ($24))|0; //@line 765 "fastmix.cpp"
  $94 = (($$) + ($$sum7)|0); //@line 765 "fastmix.cpp"
  $95 = HEAP8[$94>>0]|0; //@line 765 "fastmix.cpp"
  $96 = $95 << 24 >> 24; //@line 765 "fastmix.cpp"
  $97 = Math_imul($96, $93)|0; //@line 765 "fastmix.cpp"
  $98 = (($89) + ($97))|0; //@line 765 "fastmix.cpp"
  $99 = $98 >> 7; //@line 765 "fastmix.cpp"
  $100 = HEAP32[$21>>2]|0; //@line 766 "fastmix.cpp"
  $101 = (($100) + ($nRampLeftVol$0))|0; //@line 766 "fastmix.cpp"
  $102 = HEAP32[$22>>2]|0; //@line 766 "fastmix.cpp"
  $103 = (($102) + ($nRampRightVol$0))|0; //@line 766 "fastmix.cpp"
  $104 = $103 >> 12; //@line 766 "fastmix.cpp"
  $105 = Math_imul($99, $104)|0; //@line 766 "fastmix.cpp"
  $106 = HEAP32[$pvol$0>>2]|0; //@line 766 "fastmix.cpp"
  $107 = (($105) + ($106))|0; //@line 766 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $107; //@line 766 "fastmix.cpp"
  $108 = $101 >> 12; //@line 766 "fastmix.cpp"
  $109 = Math_imul($99, $108)|0; //@line 766 "fastmix.cpp"
  $110 = ((($pvol$0)) + 4|0); //@line 766 "fastmix.cpp"
  $111 = HEAP32[$110>>2]|0; //@line 766 "fastmix.cpp"
  $112 = (($109) + ($111))|0; //@line 766 "fastmix.cpp"
  HEAP32[$110>>2] = $112; //@line 766 "fastmix.cpp"
  $113 = ((($pvol$0)) + 8|0); //@line 766 "fastmix.cpp"
  $114 = HEAP32[$23>>2]|0; //@line 767 "fastmix.cpp"
  $115 = (($114) + ($nPos$0))|0; //@line 767 "fastmix.cpp"
  $116 = ($113>>>0)<($pbufmax>>>0); //@line 767 "fastmix.cpp"
  if ($116) {
   $nPos$0 = $115;$nRampLeftVol$0 = $101;$nRampRightVol$0 = $103;$pvol$0 = $113;
  } else {
   $$lcssa = $101;$$lcssa36 = $103;$$lcssa37 = $104;$$lcssa38 = $108;$$lcssa39 = $115;
   break;
  }
 }
 $117 = $$lcssa39 >> 16; //@line 767 "fastmix.cpp"
 $118 = HEAP32[$7>>2]|0; //@line 767 "fastmix.cpp"
 $119 = (($118) + ($117))|0; //@line 767 "fastmix.cpp"
 HEAP32[$7>>2] = $119; //@line 767 "fastmix.cpp"
 $120 = $$lcssa39 & 65535; //@line 767 "fastmix.cpp"
 HEAP32[$4>>2] = $120; //@line 767 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa36; //@line 767 "fastmix.cpp"
 $121 = ((($pChannel)) + 16|0); //@line 767 "fastmix.cpp"
 HEAP32[$121>>2] = $$lcssa37; //@line 767 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa; //@line 767 "fastmix.cpp"
 $122 = ((($pChannel)) + 20|0); //@line 767 "fastmix.cpp"
 HEAP32[$122>>2] = $$lcssa38; //@line 767 "fastmix.cpp"
 return; //@line 767 "fastmix.cpp"
}
function __Z25Mono16BitFirFilterRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa21 = 0, $$lcssa22 = 0, $$lcssa23 = 0, $$lcssa24 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 769 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 769 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 769 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 769 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 770 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 770 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 770 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 770 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 770 "fastmix.cpp"
 $9 = $8 << 1; //@line 770 "fastmix.cpp"
 $10 = (($6) + ($9)|0); //@line 770 "fastmix.cpp"
 $11 = ((($pChannel)) + 36|0); //@line 770 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 770 "fastmix.cpp"
 $13 = $12 & 64; //@line 770 "fastmix.cpp"
 $14 = ($13|0)==(0); //@line 770 "fastmix.cpp"
 $15 = (($10) + ($8<<1)|0); //@line 770 "fastmix.cpp"
 $$ = $14 ? $10 : $15; //@line 770 "fastmix.cpp"
 $16 = ((($pChannel)) + 28|0); //@line 772 "fastmix.cpp"
 $17 = ((($pChannel)) + 24|0); //@line 772 "fastmix.cpp"
 $18 = ((($pChannel)) + 12|0); //@line 773 "fastmix.cpp"
 $nPos$0 = $5;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $19 = $nPos$0 >> 16; //@line 771 "fastmix.cpp"
  $20 = $nPos$0 & 65535; //@line 771 "fastmix.cpp"
  $21 = (($20) + 16)|0; //@line 771 "fastmix.cpp"
  $22 = $21 >>> 2; //@line 771 "fastmix.cpp"
  $23 = $22 & 32760; //@line 771 "fastmix.cpp"
  $24 = (1217698 + ($23<<1)|0); //@line 771 "fastmix.cpp"
  $25 = HEAP16[$24>>1]|0; //@line 771 "fastmix.cpp"
  $26 = $25 << 16 >> 16; //@line 771 "fastmix.cpp"
  $27 = (($19) + -3)|0; //@line 771 "fastmix.cpp"
  $28 = (($$) + ($27<<1)|0); //@line 771 "fastmix.cpp"
  $29 = HEAP16[$28>>1]|0; //@line 771 "fastmix.cpp"
  $30 = $29 << 16 >> 16; //@line 771 "fastmix.cpp"
  $31 = Math_imul($30, $26)|0; //@line 771 "fastmix.cpp"
  $32 = $23 | 1; //@line 771 "fastmix.cpp"
  $33 = (1217698 + ($32<<1)|0); //@line 771 "fastmix.cpp"
  $34 = HEAP16[$33>>1]|0; //@line 771 "fastmix.cpp"
  $35 = $34 << 16 >> 16; //@line 771 "fastmix.cpp"
  $36 = (($19) + -2)|0; //@line 771 "fastmix.cpp"
  $37 = (($$) + ($36<<1)|0); //@line 771 "fastmix.cpp"
  $38 = HEAP16[$37>>1]|0; //@line 771 "fastmix.cpp"
  $39 = $38 << 16 >> 16; //@line 771 "fastmix.cpp"
  $40 = Math_imul($39, $35)|0; //@line 771 "fastmix.cpp"
  $41 = (($40) + ($31))|0; //@line 771 "fastmix.cpp"
  $42 = $23 | 2; //@line 771 "fastmix.cpp"
  $43 = (1217698 + ($42<<1)|0); //@line 771 "fastmix.cpp"
  $44 = HEAP16[$43>>1]|0; //@line 771 "fastmix.cpp"
  $45 = $44 << 16 >> 16; //@line 771 "fastmix.cpp"
  $46 = (($19) + -1)|0; //@line 771 "fastmix.cpp"
  $47 = (($$) + ($46<<1)|0); //@line 771 "fastmix.cpp"
  $48 = HEAP16[$47>>1]|0; //@line 771 "fastmix.cpp"
  $49 = $48 << 16 >> 16; //@line 771 "fastmix.cpp"
  $50 = Math_imul($49, $45)|0; //@line 771 "fastmix.cpp"
  $51 = (($41) + ($50))|0; //@line 771 "fastmix.cpp"
  $52 = $23 | 3; //@line 771 "fastmix.cpp"
  $53 = (1217698 + ($52<<1)|0); //@line 771 "fastmix.cpp"
  $54 = HEAP16[$53>>1]|0; //@line 771 "fastmix.cpp"
  $55 = $54 << 16 >> 16; //@line 771 "fastmix.cpp"
  $56 = (($$) + ($19<<1)|0); //@line 771 "fastmix.cpp"
  $57 = HEAP16[$56>>1]|0; //@line 771 "fastmix.cpp"
  $58 = $57 << 16 >> 16; //@line 771 "fastmix.cpp"
  $59 = Math_imul($58, $55)|0; //@line 771 "fastmix.cpp"
  $60 = (($51) + ($59))|0; //@line 771 "fastmix.cpp"
  $61 = $23 | 4; //@line 771 "fastmix.cpp"
  $62 = (1217698 + ($61<<1)|0); //@line 771 "fastmix.cpp"
  $63 = HEAP16[$62>>1]|0; //@line 771 "fastmix.cpp"
  $64 = $63 << 16 >> 16; //@line 771 "fastmix.cpp"
  $65 = (($19) + 1)|0; //@line 771 "fastmix.cpp"
  $66 = (($$) + ($65<<1)|0); //@line 771 "fastmix.cpp"
  $67 = HEAP16[$66>>1]|0; //@line 771 "fastmix.cpp"
  $68 = $67 << 16 >> 16; //@line 771 "fastmix.cpp"
  $69 = Math_imul($68, $64)|0; //@line 771 "fastmix.cpp"
  $70 = $23 | 5; //@line 771 "fastmix.cpp"
  $71 = (1217698 + ($70<<1)|0); //@line 771 "fastmix.cpp"
  $72 = HEAP16[$71>>1]|0; //@line 771 "fastmix.cpp"
  $73 = $72 << 16 >> 16; //@line 771 "fastmix.cpp"
  $74 = (($19) + 2)|0; //@line 771 "fastmix.cpp"
  $75 = (($$) + ($74<<1)|0); //@line 771 "fastmix.cpp"
  $76 = HEAP16[$75>>1]|0; //@line 771 "fastmix.cpp"
  $77 = $76 << 16 >> 16; //@line 771 "fastmix.cpp"
  $78 = Math_imul($77, $73)|0; //@line 771 "fastmix.cpp"
  $79 = (($78) + ($69))|0; //@line 771 "fastmix.cpp"
  $80 = $23 | 6; //@line 771 "fastmix.cpp"
  $81 = (1217698 + ($80<<1)|0); //@line 771 "fastmix.cpp"
  $82 = HEAP16[$81>>1]|0; //@line 771 "fastmix.cpp"
  $83 = $82 << 16 >> 16; //@line 771 "fastmix.cpp"
  $84 = (($19) + 3)|0; //@line 771 "fastmix.cpp"
  $85 = (($$) + ($84<<1)|0); //@line 771 "fastmix.cpp"
  $86 = HEAP16[$85>>1]|0; //@line 771 "fastmix.cpp"
  $87 = $86 << 16 >> 16; //@line 771 "fastmix.cpp"
  $88 = Math_imul($87, $83)|0; //@line 771 "fastmix.cpp"
  $89 = (($79) + ($88))|0; //@line 771 "fastmix.cpp"
  $90 = $22 | 7; //@line 771 "fastmix.cpp"
  $91 = (1217698 + ($90<<1)|0); //@line 771 "fastmix.cpp"
  $92 = HEAP16[$91>>1]|0; //@line 771 "fastmix.cpp"
  $93 = $92 << 16 >> 16; //@line 771 "fastmix.cpp"
  $94 = (($19) + 4)|0; //@line 771 "fastmix.cpp"
  $95 = (($$) + ($94<<1)|0); //@line 771 "fastmix.cpp"
  $96 = HEAP16[$95>>1]|0; //@line 771 "fastmix.cpp"
  $97 = $96 << 16 >> 16; //@line 771 "fastmix.cpp"
  $98 = Math_imul($97, $93)|0; //@line 771 "fastmix.cpp"
  $99 = (($89) + ($98))|0; //@line 771 "fastmix.cpp"
  $100 = $60 >> 1; //@line 771 "fastmix.cpp"
  $101 = $99 >> 1; //@line 771 "fastmix.cpp"
  $102 = (($101) + ($100))|0; //@line 771 "fastmix.cpp"
  $103 = $102 >> 14; //@line 771 "fastmix.cpp"
  $104 = HEAP32[$16>>2]|0; //@line 772 "fastmix.cpp"
  $105 = (($104) + ($nRampLeftVol$0))|0; //@line 772 "fastmix.cpp"
  $106 = HEAP32[$17>>2]|0; //@line 772 "fastmix.cpp"
  $107 = (($106) + ($nRampRightVol$0))|0; //@line 772 "fastmix.cpp"
  $108 = $107 >> 12; //@line 772 "fastmix.cpp"
  $109 = Math_imul($103, $108)|0; //@line 772 "fastmix.cpp"
  $110 = HEAP32[$pvol$0>>2]|0; //@line 772 "fastmix.cpp"
  $111 = (($109) + ($110))|0; //@line 772 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $111; //@line 772 "fastmix.cpp"
  $112 = $105 >> 12; //@line 772 "fastmix.cpp"
  $113 = Math_imul($103, $112)|0; //@line 772 "fastmix.cpp"
  $114 = ((($pvol$0)) + 4|0); //@line 772 "fastmix.cpp"
  $115 = HEAP32[$114>>2]|0; //@line 772 "fastmix.cpp"
  $116 = (($113) + ($115))|0; //@line 772 "fastmix.cpp"
  HEAP32[$114>>2] = $116; //@line 772 "fastmix.cpp"
  $117 = ((($pvol$0)) + 8|0); //@line 772 "fastmix.cpp"
  $118 = HEAP32[$18>>2]|0; //@line 773 "fastmix.cpp"
  $119 = (($118) + ($nPos$0))|0; //@line 773 "fastmix.cpp"
  $120 = ($117>>>0)<($pbufmax>>>0); //@line 773 "fastmix.cpp"
  if ($120) {
   $nPos$0 = $119;$nRampLeftVol$0 = $105;$nRampRightVol$0 = $107;$pvol$0 = $117;
  } else {
   $$lcssa = $105;$$lcssa21 = $107;$$lcssa22 = $108;$$lcssa23 = $112;$$lcssa24 = $119;
   break;
  }
 }
 $121 = $$lcssa24 >> 16; //@line 773 "fastmix.cpp"
 $122 = HEAP32[$7>>2]|0; //@line 773 "fastmix.cpp"
 $123 = (($122) + ($121))|0; //@line 773 "fastmix.cpp"
 HEAP32[$7>>2] = $123; //@line 773 "fastmix.cpp"
 $124 = $$lcssa24 & 65535; //@line 773 "fastmix.cpp"
 HEAP32[$4>>2] = $124; //@line 773 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa21; //@line 773 "fastmix.cpp"
 $125 = ((($pChannel)) + 16|0); //@line 773 "fastmix.cpp"
 HEAP32[$125>>2] = $$lcssa22; //@line 773 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa; //@line 773 "fastmix.cpp"
 $126 = ((($pChannel)) + 20|0); //@line 773 "fastmix.cpp"
 HEAP32[$126>>2] = $$lcssa23; //@line 773 "fastmix.cpp"
 return; //@line 773 "fastmix.cpp"
}
function __Z26Stereo8BitFirFilterRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa52 = 0, $$lcssa53 = 0, $$lcssa54 = 0, $$lcssa55 = 0, $$sum = 0, $$sum1 = 0, $$sum10 = 0, $$sum11 = 0, $$sum12 = 0, $$sum13 = 0, $$sum14 = 0, $$sum15 = 0, $$sum2 = 0, $$sum3 = 0, $$sum4 = 0, $$sum5 = 0, $$sum6 = 0, $$sum7 = 0;
 var $$sum8 = 0, $$sum9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 967 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 967 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 967 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 967 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 968 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 968 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 968 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 968 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 968 "fastmix.cpp"
 $9 = (($6) + ($8)|0); //@line 968 "fastmix.cpp"
 $10 = ((($pChannel)) + 36|0); //@line 968 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 968 "fastmix.cpp"
 $12 = $11 & 64; //@line 968 "fastmix.cpp"
 $13 = ($12|0)==(0); //@line 968 "fastmix.cpp"
 $$ = $13 ? $6 : $9; //@line 968 "fastmix.cpp"
 $14 = (($8) + -6)|0; //@line 969 "fastmix.cpp"
 $15 = (($8) + -4)|0; //@line 969 "fastmix.cpp"
 $16 = (($8) + -2)|0; //@line 969 "fastmix.cpp"
 $17 = (($8) + 2)|0; //@line 969 "fastmix.cpp"
 $18 = (($8) + 4)|0; //@line 969 "fastmix.cpp"
 $19 = (($8) + 6)|0; //@line 969 "fastmix.cpp"
 $20 = (($8) + 8)|0; //@line 969 "fastmix.cpp"
 $21 = (($8) + -5)|0; //@line 969 "fastmix.cpp"
 $22 = (($8) + -3)|0; //@line 969 "fastmix.cpp"
 $23 = (($8) + -1)|0; //@line 969 "fastmix.cpp"
 $24 = (($8) + 3)|0; //@line 969 "fastmix.cpp"
 $25 = (($8) + 5)|0; //@line 969 "fastmix.cpp"
 $26 = (($8) + 7)|0; //@line 969 "fastmix.cpp"
 $27 = (($8) + 9)|0; //@line 969 "fastmix.cpp"
 $28 = ((($pChannel)) + 28|0); //@line 970 "fastmix.cpp"
 $29 = ((($pChannel)) + 24|0); //@line 970 "fastmix.cpp"
 $30 = ((($pChannel)) + 12|0); //@line 971 "fastmix.cpp"
 $nPos$0 = $5;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $31 = $nPos$0 >> 16; //@line 969 "fastmix.cpp"
  $32 = $nPos$0 & 65535; //@line 969 "fastmix.cpp"
  $33 = (($32) + 16)|0; //@line 969 "fastmix.cpp"
  $34 = $33 >>> 2; //@line 969 "fastmix.cpp"
  $35 = $34 & 32760; //@line 969 "fastmix.cpp"
  $36 = (1217698 + ($35<<1)|0); //@line 969 "fastmix.cpp"
  $37 = HEAP16[$36>>1]|0; //@line 969 "fastmix.cpp"
  $38 = $37 << 16 >> 16; //@line 969 "fastmix.cpp"
  $39 = $31 << 1; //@line 969 "fastmix.cpp"
  $$sum = (($14) + ($39))|0; //@line 969 "fastmix.cpp"
  $40 = (($$) + ($$sum)|0); //@line 969 "fastmix.cpp"
  $41 = HEAP8[$40>>0]|0; //@line 969 "fastmix.cpp"
  $42 = $41 << 24 >> 24; //@line 969 "fastmix.cpp"
  $43 = Math_imul($42, $38)|0; //@line 969 "fastmix.cpp"
  $44 = $35 | 1; //@line 969 "fastmix.cpp"
  $45 = (1217698 + ($44<<1)|0); //@line 969 "fastmix.cpp"
  $46 = HEAP16[$45>>1]|0; //@line 969 "fastmix.cpp"
  $47 = $46 << 16 >> 16; //@line 969 "fastmix.cpp"
  $$sum1 = (($15) + ($39))|0; //@line 969 "fastmix.cpp"
  $48 = (($$) + ($$sum1)|0); //@line 969 "fastmix.cpp"
  $49 = HEAP8[$48>>0]|0; //@line 969 "fastmix.cpp"
  $50 = $49 << 24 >> 24; //@line 969 "fastmix.cpp"
  $51 = Math_imul($50, $47)|0; //@line 969 "fastmix.cpp"
  $52 = (($51) + ($43))|0; //@line 969 "fastmix.cpp"
  $53 = $35 | 2; //@line 969 "fastmix.cpp"
  $54 = (1217698 + ($53<<1)|0); //@line 969 "fastmix.cpp"
  $55 = HEAP16[$54>>1]|0; //@line 969 "fastmix.cpp"
  $56 = $55 << 16 >> 16; //@line 969 "fastmix.cpp"
  $$sum2 = (($16) + ($39))|0; //@line 969 "fastmix.cpp"
  $57 = (($$) + ($$sum2)|0); //@line 969 "fastmix.cpp"
  $58 = HEAP8[$57>>0]|0; //@line 969 "fastmix.cpp"
  $59 = $58 << 24 >> 24; //@line 969 "fastmix.cpp"
  $60 = Math_imul($59, $56)|0; //@line 969 "fastmix.cpp"
  $61 = (($52) + ($60))|0; //@line 969 "fastmix.cpp"
  $62 = $35 | 3; //@line 969 "fastmix.cpp"
  $63 = (1217698 + ($62<<1)|0); //@line 969 "fastmix.cpp"
  $64 = HEAP16[$63>>1]|0; //@line 969 "fastmix.cpp"
  $65 = $64 << 16 >> 16; //@line 969 "fastmix.cpp"
  $$sum3 = (($39) + ($8))|0; //@line 969 "fastmix.cpp"
  $66 = (($$) + ($$sum3)|0); //@line 969 "fastmix.cpp"
  $67 = HEAP8[$66>>0]|0; //@line 969 "fastmix.cpp"
  $68 = $67 << 24 >> 24; //@line 969 "fastmix.cpp"
  $69 = Math_imul($68, $65)|0; //@line 969 "fastmix.cpp"
  $70 = (($61) + ($69))|0; //@line 969 "fastmix.cpp"
  $71 = $35 | 4; //@line 969 "fastmix.cpp"
  $72 = (1217698 + ($71<<1)|0); //@line 969 "fastmix.cpp"
  $73 = HEAP16[$72>>1]|0; //@line 969 "fastmix.cpp"
  $74 = $73 << 16 >> 16; //@line 969 "fastmix.cpp"
  $$sum4 = (($17) + ($39))|0; //@line 969 "fastmix.cpp"
  $75 = (($$) + ($$sum4)|0); //@line 969 "fastmix.cpp"
  $76 = HEAP8[$75>>0]|0; //@line 969 "fastmix.cpp"
  $77 = $76 << 24 >> 24; //@line 969 "fastmix.cpp"
  $78 = Math_imul($77, $74)|0; //@line 969 "fastmix.cpp"
  $79 = (($70) + ($78))|0; //@line 969 "fastmix.cpp"
  $80 = $35 | 5; //@line 969 "fastmix.cpp"
  $81 = (1217698 + ($80<<1)|0); //@line 969 "fastmix.cpp"
  $82 = HEAP16[$81>>1]|0; //@line 969 "fastmix.cpp"
  $83 = $82 << 16 >> 16; //@line 969 "fastmix.cpp"
  $$sum5 = (($18) + ($39))|0; //@line 969 "fastmix.cpp"
  $84 = (($$) + ($$sum5)|0); //@line 969 "fastmix.cpp"
  $85 = HEAP8[$84>>0]|0; //@line 969 "fastmix.cpp"
  $86 = $85 << 24 >> 24; //@line 969 "fastmix.cpp"
  $87 = Math_imul($86, $83)|0; //@line 969 "fastmix.cpp"
  $88 = (($79) + ($87))|0; //@line 969 "fastmix.cpp"
  $89 = $35 | 6; //@line 969 "fastmix.cpp"
  $90 = (1217698 + ($89<<1)|0); //@line 969 "fastmix.cpp"
  $91 = HEAP16[$90>>1]|0; //@line 969 "fastmix.cpp"
  $92 = $91 << 16 >> 16; //@line 969 "fastmix.cpp"
  $$sum6 = (($19) + ($39))|0; //@line 969 "fastmix.cpp"
  $93 = (($$) + ($$sum6)|0); //@line 969 "fastmix.cpp"
  $94 = HEAP8[$93>>0]|0; //@line 969 "fastmix.cpp"
  $95 = $94 << 24 >> 24; //@line 969 "fastmix.cpp"
  $96 = Math_imul($95, $92)|0; //@line 969 "fastmix.cpp"
  $97 = (($88) + ($96))|0; //@line 969 "fastmix.cpp"
  $98 = $34 | 7; //@line 969 "fastmix.cpp"
  $99 = (1217698 + ($98<<1)|0); //@line 969 "fastmix.cpp"
  $100 = HEAP16[$99>>1]|0; //@line 969 "fastmix.cpp"
  $101 = $100 << 16 >> 16; //@line 969 "fastmix.cpp"
  $$sum7 = (($20) + ($39))|0; //@line 969 "fastmix.cpp"
  $102 = (($$) + ($$sum7)|0); //@line 969 "fastmix.cpp"
  $103 = HEAP8[$102>>0]|0; //@line 969 "fastmix.cpp"
  $104 = $103 << 24 >> 24; //@line 969 "fastmix.cpp"
  $105 = Math_imul($104, $101)|0; //@line 969 "fastmix.cpp"
  $106 = (($97) + ($105))|0; //@line 969 "fastmix.cpp"
  $107 = $106 >> 7; //@line 969 "fastmix.cpp"
  $$sum8 = (($21) + ($39))|0; //@line 969 "fastmix.cpp"
  $108 = (($$) + ($$sum8)|0); //@line 969 "fastmix.cpp"
  $109 = HEAP8[$108>>0]|0; //@line 969 "fastmix.cpp"
  $110 = $109 << 24 >> 24; //@line 969 "fastmix.cpp"
  $111 = Math_imul($110, $38)|0; //@line 969 "fastmix.cpp"
  $$sum9 = (($22) + ($39))|0; //@line 969 "fastmix.cpp"
  $112 = (($$) + ($$sum9)|0); //@line 969 "fastmix.cpp"
  $113 = HEAP8[$112>>0]|0; //@line 969 "fastmix.cpp"
  $114 = $113 << 24 >> 24; //@line 969 "fastmix.cpp"
  $115 = Math_imul($114, $47)|0; //@line 969 "fastmix.cpp"
  $116 = (($115) + ($111))|0; //@line 969 "fastmix.cpp"
  $$sum10 = (($23) + ($39))|0; //@line 969 "fastmix.cpp"
  $117 = (($$) + ($$sum10)|0); //@line 969 "fastmix.cpp"
  $118 = HEAP8[$117>>0]|0; //@line 969 "fastmix.cpp"
  $119 = $118 << 24 >> 24; //@line 969 "fastmix.cpp"
  $120 = Math_imul($119, $56)|0; //@line 969 "fastmix.cpp"
  $121 = (($116) + ($120))|0; //@line 969 "fastmix.cpp"
  $122 = $39 | 1; //@line 969 "fastmix.cpp"
  $$sum11 = (($122) + ($8))|0; //@line 969 "fastmix.cpp"
  $123 = (($$) + ($$sum11)|0); //@line 969 "fastmix.cpp"
  $124 = HEAP8[$123>>0]|0; //@line 969 "fastmix.cpp"
  $125 = $124 << 24 >> 24; //@line 969 "fastmix.cpp"
  $126 = Math_imul($125, $65)|0; //@line 969 "fastmix.cpp"
  $127 = (($121) + ($126))|0; //@line 969 "fastmix.cpp"
  $$sum12 = (($24) + ($39))|0; //@line 969 "fastmix.cpp"
  $128 = (($$) + ($$sum12)|0); //@line 969 "fastmix.cpp"
  $129 = HEAP8[$128>>0]|0; //@line 969 "fastmix.cpp"
  $130 = $129 << 24 >> 24; //@line 969 "fastmix.cpp"
  $131 = Math_imul($130, $74)|0; //@line 969 "fastmix.cpp"
  $132 = (($127) + ($131))|0; //@line 969 "fastmix.cpp"
  $$sum13 = (($25) + ($39))|0; //@line 969 "fastmix.cpp"
  $133 = (($$) + ($$sum13)|0); //@line 969 "fastmix.cpp"
  $134 = HEAP8[$133>>0]|0; //@line 969 "fastmix.cpp"
  $135 = $134 << 24 >> 24; //@line 969 "fastmix.cpp"
  $136 = Math_imul($135, $83)|0; //@line 969 "fastmix.cpp"
  $137 = (($132) + ($136))|0; //@line 969 "fastmix.cpp"
  $$sum14 = (($26) + ($39))|0; //@line 969 "fastmix.cpp"
  $138 = (($$) + ($$sum14)|0); //@line 969 "fastmix.cpp"
  $139 = HEAP8[$138>>0]|0; //@line 969 "fastmix.cpp"
  $140 = $139 << 24 >> 24; //@line 969 "fastmix.cpp"
  $141 = Math_imul($140, $92)|0; //@line 969 "fastmix.cpp"
  $142 = (($137) + ($141))|0; //@line 969 "fastmix.cpp"
  $$sum15 = (($27) + ($39))|0; //@line 969 "fastmix.cpp"
  $143 = (($$) + ($$sum15)|0); //@line 969 "fastmix.cpp"
  $144 = HEAP8[$143>>0]|0; //@line 969 "fastmix.cpp"
  $145 = $144 << 24 >> 24; //@line 969 "fastmix.cpp"
  $146 = Math_imul($145, $101)|0; //@line 969 "fastmix.cpp"
  $147 = (($142) + ($146))|0; //@line 969 "fastmix.cpp"
  $148 = $147 >> 7; //@line 969 "fastmix.cpp"
  $149 = HEAP32[$28>>2]|0; //@line 970 "fastmix.cpp"
  $150 = (($149) + ($nRampLeftVol$0))|0; //@line 970 "fastmix.cpp"
  $151 = HEAP32[$29>>2]|0; //@line 970 "fastmix.cpp"
  $152 = (($151) + ($nRampRightVol$0))|0; //@line 970 "fastmix.cpp"
  $153 = $152 >> 12; //@line 970 "fastmix.cpp"
  $154 = Math_imul($153, $107)|0; //@line 970 "fastmix.cpp"
  $155 = HEAP32[$pvol$0>>2]|0; //@line 970 "fastmix.cpp"
  $156 = (($154) + ($155))|0; //@line 970 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $156; //@line 970 "fastmix.cpp"
  $157 = $150 >> 12; //@line 970 "fastmix.cpp"
  $158 = Math_imul($148, $157)|0; //@line 970 "fastmix.cpp"
  $159 = ((($pvol$0)) + 4|0); //@line 970 "fastmix.cpp"
  $160 = HEAP32[$159>>2]|0; //@line 970 "fastmix.cpp"
  $161 = (($158) + ($160))|0; //@line 970 "fastmix.cpp"
  HEAP32[$159>>2] = $161; //@line 970 "fastmix.cpp"
  $162 = ((($pvol$0)) + 8|0); //@line 970 "fastmix.cpp"
  $163 = HEAP32[$30>>2]|0; //@line 971 "fastmix.cpp"
  $164 = (($163) + ($nPos$0))|0; //@line 971 "fastmix.cpp"
  $165 = ($162>>>0)<($pbufmax>>>0); //@line 971 "fastmix.cpp"
  if ($165) {
   $nPos$0 = $164;$nRampLeftVol$0 = $150;$nRampRightVol$0 = $152;$pvol$0 = $162;
  } else {
   $$lcssa = $150;$$lcssa52 = $152;$$lcssa53 = $153;$$lcssa54 = $157;$$lcssa55 = $164;
   break;
  }
 }
 $166 = $$lcssa55 >> 16; //@line 971 "fastmix.cpp"
 $167 = HEAP32[$7>>2]|0; //@line 971 "fastmix.cpp"
 $168 = (($167) + ($166))|0; //@line 971 "fastmix.cpp"
 HEAP32[$7>>2] = $168; //@line 971 "fastmix.cpp"
 $169 = $$lcssa55 & 65535; //@line 971 "fastmix.cpp"
 HEAP32[$4>>2] = $169; //@line 971 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa52; //@line 971 "fastmix.cpp"
 $170 = ((($pChannel)) + 16|0); //@line 971 "fastmix.cpp"
 HEAP32[$170>>2] = $$lcssa53; //@line 971 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa; //@line 971 "fastmix.cpp"
 $171 = ((($pChannel)) + 20|0); //@line 971 "fastmix.cpp"
 HEAP32[$171>>2] = $$lcssa54; //@line 971 "fastmix.cpp"
 return; //@line 971 "fastmix.cpp"
}
function __Z27Stereo16BitFirFilterRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa21 = 0, $$lcssa22 = 0, $$lcssa23 = 0, $$lcssa24 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 973 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 973 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 973 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 973 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 974 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 974 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 974 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 974 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 974 "fastmix.cpp"
 $9 = $8 << 1; //@line 974 "fastmix.cpp"
 $10 = (($6) + ($9)|0); //@line 974 "fastmix.cpp"
 $11 = ((($pChannel)) + 36|0); //@line 974 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 974 "fastmix.cpp"
 $13 = $12 & 64; //@line 974 "fastmix.cpp"
 $14 = ($13|0)==(0); //@line 974 "fastmix.cpp"
 $15 = (($10) + ($8<<1)|0); //@line 974 "fastmix.cpp"
 $$ = $14 ? $10 : $15; //@line 974 "fastmix.cpp"
 $16 = ((($pChannel)) + 28|0); //@line 976 "fastmix.cpp"
 $17 = ((($pChannel)) + 24|0); //@line 976 "fastmix.cpp"
 $18 = ((($pChannel)) + 12|0); //@line 977 "fastmix.cpp"
 $nPos$0 = $5;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $19 = $nPos$0 >> 16; //@line 975 "fastmix.cpp"
  $20 = $nPos$0 & 65535; //@line 975 "fastmix.cpp"
  $21 = (($20) + 16)|0; //@line 975 "fastmix.cpp"
  $22 = $21 >>> 2; //@line 975 "fastmix.cpp"
  $23 = $22 & 32760; //@line 975 "fastmix.cpp"
  $24 = (1217698 + ($23<<1)|0); //@line 975 "fastmix.cpp"
  $25 = HEAP16[$24>>1]|0; //@line 975 "fastmix.cpp"
  $26 = $25 << 16 >> 16; //@line 975 "fastmix.cpp"
  $27 = $19 << 1; //@line 975 "fastmix.cpp"
  $28 = (($27) + -6)|0; //@line 975 "fastmix.cpp"
  $29 = (($$) + ($28<<1)|0); //@line 975 "fastmix.cpp"
  $30 = HEAP16[$29>>1]|0; //@line 975 "fastmix.cpp"
  $31 = $30 << 16 >> 16; //@line 975 "fastmix.cpp"
  $32 = Math_imul($31, $26)|0; //@line 975 "fastmix.cpp"
  $33 = $23 | 1; //@line 975 "fastmix.cpp"
  $34 = (1217698 + ($33<<1)|0); //@line 975 "fastmix.cpp"
  $35 = HEAP16[$34>>1]|0; //@line 975 "fastmix.cpp"
  $36 = $35 << 16 >> 16; //@line 975 "fastmix.cpp"
  $37 = (($27) + -4)|0; //@line 975 "fastmix.cpp"
  $38 = (($$) + ($37<<1)|0); //@line 975 "fastmix.cpp"
  $39 = HEAP16[$38>>1]|0; //@line 975 "fastmix.cpp"
  $40 = $39 << 16 >> 16; //@line 975 "fastmix.cpp"
  $41 = Math_imul($40, $36)|0; //@line 975 "fastmix.cpp"
  $42 = (($41) + ($32))|0; //@line 975 "fastmix.cpp"
  $43 = $23 | 2; //@line 975 "fastmix.cpp"
  $44 = (1217698 + ($43<<1)|0); //@line 975 "fastmix.cpp"
  $45 = HEAP16[$44>>1]|0; //@line 975 "fastmix.cpp"
  $46 = $45 << 16 >> 16; //@line 975 "fastmix.cpp"
  $47 = (($27) + -2)|0; //@line 975 "fastmix.cpp"
  $48 = (($$) + ($47<<1)|0); //@line 975 "fastmix.cpp"
  $49 = HEAP16[$48>>1]|0; //@line 975 "fastmix.cpp"
  $50 = $49 << 16 >> 16; //@line 975 "fastmix.cpp"
  $51 = Math_imul($50, $46)|0; //@line 975 "fastmix.cpp"
  $52 = (($42) + ($51))|0; //@line 975 "fastmix.cpp"
  $53 = $23 | 3; //@line 975 "fastmix.cpp"
  $54 = (1217698 + ($53<<1)|0); //@line 975 "fastmix.cpp"
  $55 = HEAP16[$54>>1]|0; //@line 975 "fastmix.cpp"
  $56 = $55 << 16 >> 16; //@line 975 "fastmix.cpp"
  $57 = (($$) + ($27<<1)|0); //@line 975 "fastmix.cpp"
  $58 = HEAP16[$57>>1]|0; //@line 975 "fastmix.cpp"
  $59 = $58 << 16 >> 16; //@line 975 "fastmix.cpp"
  $60 = Math_imul($59, $56)|0; //@line 975 "fastmix.cpp"
  $61 = (($52) + ($60))|0; //@line 975 "fastmix.cpp"
  $62 = $23 | 4; //@line 975 "fastmix.cpp"
  $63 = (1217698 + ($62<<1)|0); //@line 975 "fastmix.cpp"
  $64 = HEAP16[$63>>1]|0; //@line 975 "fastmix.cpp"
  $65 = $64 << 16 >> 16; //@line 975 "fastmix.cpp"
  $66 = (($27) + 2)|0; //@line 975 "fastmix.cpp"
  $67 = (($$) + ($66<<1)|0); //@line 975 "fastmix.cpp"
  $68 = HEAP16[$67>>1]|0; //@line 975 "fastmix.cpp"
  $69 = $68 << 16 >> 16; //@line 975 "fastmix.cpp"
  $70 = Math_imul($69, $65)|0; //@line 975 "fastmix.cpp"
  $71 = $23 | 5; //@line 975 "fastmix.cpp"
  $72 = (1217698 + ($71<<1)|0); //@line 975 "fastmix.cpp"
  $73 = HEAP16[$72>>1]|0; //@line 975 "fastmix.cpp"
  $74 = $73 << 16 >> 16; //@line 975 "fastmix.cpp"
  $75 = (($27) + 4)|0; //@line 975 "fastmix.cpp"
  $76 = (($$) + ($75<<1)|0); //@line 975 "fastmix.cpp"
  $77 = HEAP16[$76>>1]|0; //@line 975 "fastmix.cpp"
  $78 = $77 << 16 >> 16; //@line 975 "fastmix.cpp"
  $79 = Math_imul($78, $74)|0; //@line 975 "fastmix.cpp"
  $80 = (($79) + ($70))|0; //@line 975 "fastmix.cpp"
  $81 = $23 | 6; //@line 975 "fastmix.cpp"
  $82 = (1217698 + ($81<<1)|0); //@line 975 "fastmix.cpp"
  $83 = HEAP16[$82>>1]|0; //@line 975 "fastmix.cpp"
  $84 = $83 << 16 >> 16; //@line 975 "fastmix.cpp"
  $85 = (($27) + 6)|0; //@line 975 "fastmix.cpp"
  $86 = (($$) + ($85<<1)|0); //@line 975 "fastmix.cpp"
  $87 = HEAP16[$86>>1]|0; //@line 975 "fastmix.cpp"
  $88 = $87 << 16 >> 16; //@line 975 "fastmix.cpp"
  $89 = Math_imul($88, $84)|0; //@line 975 "fastmix.cpp"
  $90 = (($80) + ($89))|0; //@line 975 "fastmix.cpp"
  $91 = $22 | 7; //@line 975 "fastmix.cpp"
  $92 = (1217698 + ($91<<1)|0); //@line 975 "fastmix.cpp"
  $93 = HEAP16[$92>>1]|0; //@line 975 "fastmix.cpp"
  $94 = $93 << 16 >> 16; //@line 975 "fastmix.cpp"
  $95 = (($27) + 8)|0; //@line 975 "fastmix.cpp"
  $96 = (($$) + ($95<<1)|0); //@line 975 "fastmix.cpp"
  $97 = HEAP16[$96>>1]|0; //@line 975 "fastmix.cpp"
  $98 = $97 << 16 >> 16; //@line 975 "fastmix.cpp"
  $99 = Math_imul($98, $94)|0; //@line 975 "fastmix.cpp"
  $100 = (($90) + ($99))|0; //@line 975 "fastmix.cpp"
  $101 = $61 >> 1; //@line 975 "fastmix.cpp"
  $102 = $100 >> 1; //@line 975 "fastmix.cpp"
  $103 = (($102) + ($101))|0; //@line 975 "fastmix.cpp"
  $104 = $103 >> 14; //@line 975 "fastmix.cpp"
  $105 = (($27) + -5)|0; //@line 975 "fastmix.cpp"
  $106 = (($$) + ($105<<1)|0); //@line 975 "fastmix.cpp"
  $107 = HEAP16[$106>>1]|0; //@line 975 "fastmix.cpp"
  $108 = $107 << 16 >> 16; //@line 975 "fastmix.cpp"
  $109 = Math_imul($108, $26)|0; //@line 975 "fastmix.cpp"
  $110 = (($27) + -3)|0; //@line 975 "fastmix.cpp"
  $111 = (($$) + ($110<<1)|0); //@line 975 "fastmix.cpp"
  $112 = HEAP16[$111>>1]|0; //@line 975 "fastmix.cpp"
  $113 = $112 << 16 >> 16; //@line 975 "fastmix.cpp"
  $114 = Math_imul($113, $36)|0; //@line 975 "fastmix.cpp"
  $115 = (($114) + ($109))|0; //@line 975 "fastmix.cpp"
  $116 = (($27) + -1)|0; //@line 975 "fastmix.cpp"
  $117 = (($$) + ($116<<1)|0); //@line 975 "fastmix.cpp"
  $118 = HEAP16[$117>>1]|0; //@line 975 "fastmix.cpp"
  $119 = $118 << 16 >> 16; //@line 975 "fastmix.cpp"
  $120 = Math_imul($119, $46)|0; //@line 975 "fastmix.cpp"
  $121 = (($115) + ($120))|0; //@line 975 "fastmix.cpp"
  $122 = $27 | 1; //@line 975 "fastmix.cpp"
  $123 = (($$) + ($122<<1)|0); //@line 975 "fastmix.cpp"
  $124 = HEAP16[$123>>1]|0; //@line 975 "fastmix.cpp"
  $125 = $124 << 16 >> 16; //@line 975 "fastmix.cpp"
  $126 = Math_imul($125, $56)|0; //@line 975 "fastmix.cpp"
  $127 = (($121) + ($126))|0; //@line 975 "fastmix.cpp"
  $128 = (($27) + 3)|0; //@line 975 "fastmix.cpp"
  $129 = (($$) + ($128<<1)|0); //@line 975 "fastmix.cpp"
  $130 = HEAP16[$129>>1]|0; //@line 975 "fastmix.cpp"
  $131 = $130 << 16 >> 16; //@line 975 "fastmix.cpp"
  $132 = Math_imul($131, $65)|0; //@line 975 "fastmix.cpp"
  $133 = (($27) + 5)|0; //@line 975 "fastmix.cpp"
  $134 = (($$) + ($133<<1)|0); //@line 975 "fastmix.cpp"
  $135 = HEAP16[$134>>1]|0; //@line 975 "fastmix.cpp"
  $136 = $135 << 16 >> 16; //@line 975 "fastmix.cpp"
  $137 = Math_imul($136, $74)|0; //@line 975 "fastmix.cpp"
  $138 = (($137) + ($132))|0; //@line 975 "fastmix.cpp"
  $139 = (($27) + 7)|0; //@line 975 "fastmix.cpp"
  $140 = (($$) + ($139<<1)|0); //@line 975 "fastmix.cpp"
  $141 = HEAP16[$140>>1]|0; //@line 975 "fastmix.cpp"
  $142 = $141 << 16 >> 16; //@line 975 "fastmix.cpp"
  $143 = Math_imul($142, $84)|0; //@line 975 "fastmix.cpp"
  $144 = (($138) + ($143))|0; //@line 975 "fastmix.cpp"
  $145 = (($27) + 9)|0; //@line 975 "fastmix.cpp"
  $146 = (($$) + ($145<<1)|0); //@line 975 "fastmix.cpp"
  $147 = HEAP16[$146>>1]|0; //@line 975 "fastmix.cpp"
  $148 = $147 << 16 >> 16; //@line 975 "fastmix.cpp"
  $149 = Math_imul($148, $94)|0; //@line 975 "fastmix.cpp"
  $150 = (($144) + ($149))|0; //@line 975 "fastmix.cpp"
  $151 = $127 >> 1; //@line 975 "fastmix.cpp"
  $152 = $150 >> 1; //@line 975 "fastmix.cpp"
  $153 = (($152) + ($151))|0; //@line 975 "fastmix.cpp"
  $154 = $153 >> 14; //@line 975 "fastmix.cpp"
  $155 = HEAP32[$16>>2]|0; //@line 976 "fastmix.cpp"
  $156 = (($155) + ($nRampLeftVol$0))|0; //@line 976 "fastmix.cpp"
  $157 = HEAP32[$17>>2]|0; //@line 976 "fastmix.cpp"
  $158 = (($157) + ($nRampRightVol$0))|0; //@line 976 "fastmix.cpp"
  $159 = $158 >> 12; //@line 976 "fastmix.cpp"
  $160 = Math_imul($159, $104)|0; //@line 976 "fastmix.cpp"
  $161 = HEAP32[$pvol$0>>2]|0; //@line 976 "fastmix.cpp"
  $162 = (($160) + ($161))|0; //@line 976 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $162; //@line 976 "fastmix.cpp"
  $163 = $156 >> 12; //@line 976 "fastmix.cpp"
  $164 = Math_imul($154, $163)|0; //@line 976 "fastmix.cpp"
  $165 = ((($pvol$0)) + 4|0); //@line 976 "fastmix.cpp"
  $166 = HEAP32[$165>>2]|0; //@line 976 "fastmix.cpp"
  $167 = (($164) + ($166))|0; //@line 976 "fastmix.cpp"
  HEAP32[$165>>2] = $167; //@line 976 "fastmix.cpp"
  $168 = ((($pvol$0)) + 8|0); //@line 976 "fastmix.cpp"
  $169 = HEAP32[$18>>2]|0; //@line 977 "fastmix.cpp"
  $170 = (($169) + ($nPos$0))|0; //@line 977 "fastmix.cpp"
  $171 = ($168>>>0)<($pbufmax>>>0); //@line 977 "fastmix.cpp"
  if ($171) {
   $nPos$0 = $170;$nRampLeftVol$0 = $156;$nRampRightVol$0 = $158;$pvol$0 = $168;
  } else {
   $$lcssa = $156;$$lcssa21 = $158;$$lcssa22 = $159;$$lcssa23 = $163;$$lcssa24 = $170;
   break;
  }
 }
 $172 = $$lcssa24 >> 16; //@line 977 "fastmix.cpp"
 $173 = HEAP32[$7>>2]|0; //@line 977 "fastmix.cpp"
 $174 = (($173) + ($172))|0; //@line 977 "fastmix.cpp"
 HEAP32[$7>>2] = $174; //@line 977 "fastmix.cpp"
 $175 = $$lcssa24 & 65535; //@line 977 "fastmix.cpp"
 HEAP32[$4>>2] = $175; //@line 977 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa21; //@line 977 "fastmix.cpp"
 $176 = ((($pChannel)) + 16|0); //@line 977 "fastmix.cpp"
 HEAP32[$176>>2] = $$lcssa22; //@line 977 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa; //@line 977 "fastmix.cpp"
 $177 = ((($pChannel)) + 20|0); //@line 977 "fastmix.cpp"
 HEAP32[$177>>2] = $$lcssa23; //@line 977 "fastmix.cpp"
 return; //@line 977 "fastmix.cpp"
}
function __Z26FilterMono8BitFirFilterMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa21 = 0, $$sum = 0, $$sum1 = 0, $$sum2 = 0, $$sum3 = 0, $$sum4 = 0, $$sum5 = 0, $$sum6 = 0, $$sum7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 56|0); //@line 1029 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1029 "fastmix.cpp"
 $2 = ((($pChannel)) + 60|0); //@line 1029 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1029 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 1030 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1030 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 1030 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 1030 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 1030 "fastmix.cpp"
 $9 = (($6) + ($8)|0); //@line 1030 "fastmix.cpp"
 $10 = ((($pChannel)) + 36|0); //@line 1030 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 1030 "fastmix.cpp"
 $12 = $11 & 64; //@line 1030 "fastmix.cpp"
 $13 = ($12|0)==(0); //@line 1030 "fastmix.cpp"
 $$ = $13 ? $6 : $9; //@line 1030 "fastmix.cpp"
 $14 = (($8) + -3)|0; //@line 1031 "fastmix.cpp"
 $15 = (($8) + -2)|0; //@line 1031 "fastmix.cpp"
 $16 = (($8) + -1)|0; //@line 1031 "fastmix.cpp"
 $17 = (($8) + 1)|0; //@line 1031 "fastmix.cpp"
 $18 = (($8) + 2)|0; //@line 1031 "fastmix.cpp"
 $19 = (($8) + 3)|0; //@line 1031 "fastmix.cpp"
 $20 = (($8) + 4)|0; //@line 1031 "fastmix.cpp"
 $21 = ((($pChannel)) + 72|0); //@line 1032 "fastmix.cpp"
 $22 = ((($pChannel)) + 76|0); //@line 1032 "fastmix.cpp"
 $23 = ((($pChannel)) + 80|0); //@line 1032 "fastmix.cpp"
 $24 = ((($pChannel)) + 16|0); //@line 1033 "fastmix.cpp"
 $25 = ((($pChannel)) + 20|0); //@line 1033 "fastmix.cpp"
 $26 = ((($pChannel)) + 12|0); //@line 1034 "fastmix.cpp"
 $fy1$0 = $1;$fy2$0 = $3;$nPos$0 = $5;$pvol$0 = $pbuffer;
 while(1) {
  $27 = $nPos$0 >> 16; //@line 1031 "fastmix.cpp"
  $28 = $nPos$0 & 65535; //@line 1031 "fastmix.cpp"
  $29 = (($28) + 16)|0; //@line 1031 "fastmix.cpp"
  $30 = $29 >>> 2; //@line 1031 "fastmix.cpp"
  $31 = $30 & 32760; //@line 1031 "fastmix.cpp"
  $32 = (1217698 + ($31<<1)|0); //@line 1031 "fastmix.cpp"
  $33 = HEAP16[$32>>1]|0; //@line 1031 "fastmix.cpp"
  $34 = $33 << 16 >> 16; //@line 1031 "fastmix.cpp"
  $$sum = (($14) + ($27))|0; //@line 1031 "fastmix.cpp"
  $35 = (($$) + ($$sum)|0); //@line 1031 "fastmix.cpp"
  $36 = HEAP8[$35>>0]|0; //@line 1031 "fastmix.cpp"
  $37 = $36 << 24 >> 24; //@line 1031 "fastmix.cpp"
  $38 = Math_imul($37, $34)|0; //@line 1031 "fastmix.cpp"
  $39 = $31 | 1; //@line 1031 "fastmix.cpp"
  $40 = (1217698 + ($39<<1)|0); //@line 1031 "fastmix.cpp"
  $41 = HEAP16[$40>>1]|0; //@line 1031 "fastmix.cpp"
  $42 = $41 << 16 >> 16; //@line 1031 "fastmix.cpp"
  $$sum1 = (($15) + ($27))|0; //@line 1031 "fastmix.cpp"
  $43 = (($$) + ($$sum1)|0); //@line 1031 "fastmix.cpp"
  $44 = HEAP8[$43>>0]|0; //@line 1031 "fastmix.cpp"
  $45 = $44 << 24 >> 24; //@line 1031 "fastmix.cpp"
  $46 = Math_imul($45, $42)|0; //@line 1031 "fastmix.cpp"
  $47 = (($46) + ($38))|0; //@line 1031 "fastmix.cpp"
  $48 = $31 | 2; //@line 1031 "fastmix.cpp"
  $49 = (1217698 + ($48<<1)|0); //@line 1031 "fastmix.cpp"
  $50 = HEAP16[$49>>1]|0; //@line 1031 "fastmix.cpp"
  $51 = $50 << 16 >> 16; //@line 1031 "fastmix.cpp"
  $$sum2 = (($16) + ($27))|0; //@line 1031 "fastmix.cpp"
  $52 = (($$) + ($$sum2)|0); //@line 1031 "fastmix.cpp"
  $53 = HEAP8[$52>>0]|0; //@line 1031 "fastmix.cpp"
  $54 = $53 << 24 >> 24; //@line 1031 "fastmix.cpp"
  $55 = Math_imul($54, $51)|0; //@line 1031 "fastmix.cpp"
  $56 = (($47) + ($55))|0; //@line 1031 "fastmix.cpp"
  $57 = $31 | 3; //@line 1031 "fastmix.cpp"
  $58 = (1217698 + ($57<<1)|0); //@line 1031 "fastmix.cpp"
  $59 = HEAP16[$58>>1]|0; //@line 1031 "fastmix.cpp"
  $60 = $59 << 16 >> 16; //@line 1031 "fastmix.cpp"
  $$sum3 = (($27) + ($8))|0; //@line 1031 "fastmix.cpp"
  $61 = (($$) + ($$sum3)|0); //@line 1031 "fastmix.cpp"
  $62 = HEAP8[$61>>0]|0; //@line 1031 "fastmix.cpp"
  $63 = $62 << 24 >> 24; //@line 1031 "fastmix.cpp"
  $64 = Math_imul($63, $60)|0; //@line 1031 "fastmix.cpp"
  $65 = (($56) + ($64))|0; //@line 1031 "fastmix.cpp"
  $66 = $31 | 4; //@line 1031 "fastmix.cpp"
  $67 = (1217698 + ($66<<1)|0); //@line 1031 "fastmix.cpp"
  $68 = HEAP16[$67>>1]|0; //@line 1031 "fastmix.cpp"
  $69 = $68 << 16 >> 16; //@line 1031 "fastmix.cpp"
  $$sum4 = (($17) + ($27))|0; //@line 1031 "fastmix.cpp"
  $70 = (($$) + ($$sum4)|0); //@line 1031 "fastmix.cpp"
  $71 = HEAP8[$70>>0]|0; //@line 1031 "fastmix.cpp"
  $72 = $71 << 24 >> 24; //@line 1031 "fastmix.cpp"
  $73 = Math_imul($72, $69)|0; //@line 1031 "fastmix.cpp"
  $74 = (($65) + ($73))|0; //@line 1031 "fastmix.cpp"
  $75 = $31 | 5; //@line 1031 "fastmix.cpp"
  $76 = (1217698 + ($75<<1)|0); //@line 1031 "fastmix.cpp"
  $77 = HEAP16[$76>>1]|0; //@line 1031 "fastmix.cpp"
  $78 = $77 << 16 >> 16; //@line 1031 "fastmix.cpp"
  $$sum5 = (($18) + ($27))|0; //@line 1031 "fastmix.cpp"
  $79 = (($$) + ($$sum5)|0); //@line 1031 "fastmix.cpp"
  $80 = HEAP8[$79>>0]|0; //@line 1031 "fastmix.cpp"
  $81 = $80 << 24 >> 24; //@line 1031 "fastmix.cpp"
  $82 = Math_imul($81, $78)|0; //@line 1031 "fastmix.cpp"
  $83 = (($74) + ($82))|0; //@line 1031 "fastmix.cpp"
  $84 = $31 | 6; //@line 1031 "fastmix.cpp"
  $85 = (1217698 + ($84<<1)|0); //@line 1031 "fastmix.cpp"
  $86 = HEAP16[$85>>1]|0; //@line 1031 "fastmix.cpp"
  $87 = $86 << 16 >> 16; //@line 1031 "fastmix.cpp"
  $$sum6 = (($19) + ($27))|0; //@line 1031 "fastmix.cpp"
  $88 = (($$) + ($$sum6)|0); //@line 1031 "fastmix.cpp"
  $89 = HEAP8[$88>>0]|0; //@line 1031 "fastmix.cpp"
  $90 = $89 << 24 >> 24; //@line 1031 "fastmix.cpp"
  $91 = Math_imul($90, $87)|0; //@line 1031 "fastmix.cpp"
  $92 = (($83) + ($91))|0; //@line 1031 "fastmix.cpp"
  $93 = $30 | 7; //@line 1031 "fastmix.cpp"
  $94 = (1217698 + ($93<<1)|0); //@line 1031 "fastmix.cpp"
  $95 = HEAP16[$94>>1]|0; //@line 1031 "fastmix.cpp"
  $96 = $95 << 16 >> 16; //@line 1031 "fastmix.cpp"
  $$sum7 = (($20) + ($27))|0; //@line 1031 "fastmix.cpp"
  $97 = (($$) + ($$sum7)|0); //@line 1031 "fastmix.cpp"
  $98 = HEAP8[$97>>0]|0; //@line 1031 "fastmix.cpp"
  $99 = $98 << 24 >> 24; //@line 1031 "fastmix.cpp"
  $100 = Math_imul($99, $96)|0; //@line 1031 "fastmix.cpp"
  $101 = (($92) + ($100))|0; //@line 1031 "fastmix.cpp"
  $102 = $101 >> 7; //@line 1031 "fastmix.cpp"
  $103 = HEAP32[$21>>2]|0; //@line 1032 "fastmix.cpp"
  $104 = Math_imul($102, $103)|0; //@line 1032 "fastmix.cpp"
  $105 = HEAP32[$22>>2]|0; //@line 1032 "fastmix.cpp"
  $106 = Math_imul($105, $fy1$0)|0; //@line 1032 "fastmix.cpp"
  $107 = HEAP32[$23>>2]|0; //@line 1032 "fastmix.cpp"
  $108 = Math_imul($107, $fy2$0)|0; //@line 1032 "fastmix.cpp"
  $109 = (($106) + 4096)|0; //@line 1032 "fastmix.cpp"
  $110 = (($109) + ($108))|0; //@line 1032 "fastmix.cpp"
  $111 = (($110) + ($104))|0; //@line 1032 "fastmix.cpp"
  $112 = $111 >> 13; //@line 1032 "fastmix.cpp"
  $113 = HEAP32[$24>>2]|0; //@line 1033 "fastmix.cpp"
  $114 = Math_imul($112, $113)|0; //@line 1033 "fastmix.cpp"
  $115 = HEAP32[$pvol$0>>2]|0; //@line 1033 "fastmix.cpp"
  $116 = (($114) + ($115))|0; //@line 1033 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $116; //@line 1033 "fastmix.cpp"
  $117 = HEAP32[$25>>2]|0; //@line 1033 "fastmix.cpp"
  $118 = Math_imul($112, $117)|0; //@line 1033 "fastmix.cpp"
  $119 = ((($pvol$0)) + 4|0); //@line 1033 "fastmix.cpp"
  $120 = HEAP32[$119>>2]|0; //@line 1033 "fastmix.cpp"
  $121 = (($118) + ($120))|0; //@line 1033 "fastmix.cpp"
  HEAP32[$119>>2] = $121; //@line 1033 "fastmix.cpp"
  $122 = ((($pvol$0)) + 8|0); //@line 1033 "fastmix.cpp"
  $123 = HEAP32[$26>>2]|0; //@line 1034 "fastmix.cpp"
  $124 = (($123) + ($nPos$0))|0; //@line 1034 "fastmix.cpp"
  $125 = ($122>>>0)<($pbufmax>>>0); //@line 1034 "fastmix.cpp"
  if ($125) {
   $fy2$0$phi = $fy1$0;$fy1$0 = $112;$nPos$0 = $124;$pvol$0 = $122;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $112;$$lcssa21 = $124;$fy1$0$lcssa = $fy1$0;
   break;
  }
 }
 $126 = $$lcssa21 >> 16; //@line 1034 "fastmix.cpp"
 $127 = HEAP32[$7>>2]|0; //@line 1034 "fastmix.cpp"
 $128 = (($127) + ($126))|0; //@line 1034 "fastmix.cpp"
 HEAP32[$7>>2] = $128; //@line 1034 "fastmix.cpp"
 $129 = $$lcssa21 & 65535; //@line 1034 "fastmix.cpp"
 HEAP32[$4>>2] = $129; //@line 1034 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 1034 "fastmix.cpp"
 HEAP32[$2>>2] = $fy1$0$lcssa; //@line 1034 "fastmix.cpp"
 return; //@line 1034 "fastmix.cpp"
}
function __Z27FilterMono16BitFirFilterMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa6 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0;
 var $fy2$0$phi = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 56|0); //@line 1036 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1036 "fastmix.cpp"
 $2 = ((($pChannel)) + 60|0); //@line 1036 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1036 "fastmix.cpp"
 $4 = ((($pChannel)) + 8|0); //@line 1037 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1037 "fastmix.cpp"
 $6 = HEAP32[$pChannel>>2]|0; //@line 1037 "fastmix.cpp"
 $7 = ((($pChannel)) + 4|0); //@line 1037 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 1037 "fastmix.cpp"
 $9 = $8 << 1; //@line 1037 "fastmix.cpp"
 $10 = (($6) + ($9)|0); //@line 1037 "fastmix.cpp"
 $11 = ((($pChannel)) + 36|0); //@line 1037 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1037 "fastmix.cpp"
 $13 = $12 & 64; //@line 1037 "fastmix.cpp"
 $14 = ($13|0)==(0); //@line 1037 "fastmix.cpp"
 $15 = (($10) + ($8<<1)|0); //@line 1037 "fastmix.cpp"
 $$ = $14 ? $10 : $15; //@line 1037 "fastmix.cpp"
 $16 = ((($pChannel)) + 72|0); //@line 1039 "fastmix.cpp"
 $17 = ((($pChannel)) + 76|0); //@line 1039 "fastmix.cpp"
 $18 = ((($pChannel)) + 80|0); //@line 1039 "fastmix.cpp"
 $19 = ((($pChannel)) + 16|0); //@line 1040 "fastmix.cpp"
 $20 = ((($pChannel)) + 20|0); //@line 1040 "fastmix.cpp"
 $21 = ((($pChannel)) + 12|0); //@line 1041 "fastmix.cpp"
 $fy1$0 = $1;$fy2$0 = $3;$nPos$0 = $5;$pvol$0 = $pbuffer;
 while(1) {
  $22 = $nPos$0 >> 16; //@line 1038 "fastmix.cpp"
  $23 = $nPos$0 & 65535; //@line 1038 "fastmix.cpp"
  $24 = (($23) + 16)|0; //@line 1038 "fastmix.cpp"
  $25 = $24 >>> 2; //@line 1038 "fastmix.cpp"
  $26 = $25 & 32760; //@line 1038 "fastmix.cpp"
  $27 = (1217698 + ($26<<1)|0); //@line 1038 "fastmix.cpp"
  $28 = HEAP16[$27>>1]|0; //@line 1038 "fastmix.cpp"
  $29 = $28 << 16 >> 16; //@line 1038 "fastmix.cpp"
  $30 = (($22) + -3)|0; //@line 1038 "fastmix.cpp"
  $31 = (($$) + ($30<<1)|0); //@line 1038 "fastmix.cpp"
  $32 = HEAP16[$31>>1]|0; //@line 1038 "fastmix.cpp"
  $33 = $32 << 16 >> 16; //@line 1038 "fastmix.cpp"
  $34 = Math_imul($33, $29)|0; //@line 1038 "fastmix.cpp"
  $35 = $26 | 1; //@line 1038 "fastmix.cpp"
  $36 = (1217698 + ($35<<1)|0); //@line 1038 "fastmix.cpp"
  $37 = HEAP16[$36>>1]|0; //@line 1038 "fastmix.cpp"
  $38 = $37 << 16 >> 16; //@line 1038 "fastmix.cpp"
  $39 = (($22) + -2)|0; //@line 1038 "fastmix.cpp"
  $40 = (($$) + ($39<<1)|0); //@line 1038 "fastmix.cpp"
  $41 = HEAP16[$40>>1]|0; //@line 1038 "fastmix.cpp"
  $42 = $41 << 16 >> 16; //@line 1038 "fastmix.cpp"
  $43 = Math_imul($42, $38)|0; //@line 1038 "fastmix.cpp"
  $44 = (($43) + ($34))|0; //@line 1038 "fastmix.cpp"
  $45 = $26 | 2; //@line 1038 "fastmix.cpp"
  $46 = (1217698 + ($45<<1)|0); //@line 1038 "fastmix.cpp"
  $47 = HEAP16[$46>>1]|0; //@line 1038 "fastmix.cpp"
  $48 = $47 << 16 >> 16; //@line 1038 "fastmix.cpp"
  $49 = (($22) + -1)|0; //@line 1038 "fastmix.cpp"
  $50 = (($$) + ($49<<1)|0); //@line 1038 "fastmix.cpp"
  $51 = HEAP16[$50>>1]|0; //@line 1038 "fastmix.cpp"
  $52 = $51 << 16 >> 16; //@line 1038 "fastmix.cpp"
  $53 = Math_imul($52, $48)|0; //@line 1038 "fastmix.cpp"
  $54 = (($44) + ($53))|0; //@line 1038 "fastmix.cpp"
  $55 = $26 | 3; //@line 1038 "fastmix.cpp"
  $56 = (1217698 + ($55<<1)|0); //@line 1038 "fastmix.cpp"
  $57 = HEAP16[$56>>1]|0; //@line 1038 "fastmix.cpp"
  $58 = $57 << 16 >> 16; //@line 1038 "fastmix.cpp"
  $59 = (($$) + ($22<<1)|0); //@line 1038 "fastmix.cpp"
  $60 = HEAP16[$59>>1]|0; //@line 1038 "fastmix.cpp"
  $61 = $60 << 16 >> 16; //@line 1038 "fastmix.cpp"
  $62 = Math_imul($61, $58)|0; //@line 1038 "fastmix.cpp"
  $63 = (($54) + ($62))|0; //@line 1038 "fastmix.cpp"
  $64 = $26 | 4; //@line 1038 "fastmix.cpp"
  $65 = (1217698 + ($64<<1)|0); //@line 1038 "fastmix.cpp"
  $66 = HEAP16[$65>>1]|0; //@line 1038 "fastmix.cpp"
  $67 = $66 << 16 >> 16; //@line 1038 "fastmix.cpp"
  $68 = (($22) + 1)|0; //@line 1038 "fastmix.cpp"
  $69 = (($$) + ($68<<1)|0); //@line 1038 "fastmix.cpp"
  $70 = HEAP16[$69>>1]|0; //@line 1038 "fastmix.cpp"
  $71 = $70 << 16 >> 16; //@line 1038 "fastmix.cpp"
  $72 = Math_imul($71, $67)|0; //@line 1038 "fastmix.cpp"
  $73 = $26 | 5; //@line 1038 "fastmix.cpp"
  $74 = (1217698 + ($73<<1)|0); //@line 1038 "fastmix.cpp"
  $75 = HEAP16[$74>>1]|0; //@line 1038 "fastmix.cpp"
  $76 = $75 << 16 >> 16; //@line 1038 "fastmix.cpp"
  $77 = (($22) + 2)|0; //@line 1038 "fastmix.cpp"
  $78 = (($$) + ($77<<1)|0); //@line 1038 "fastmix.cpp"
  $79 = HEAP16[$78>>1]|0; //@line 1038 "fastmix.cpp"
  $80 = $79 << 16 >> 16; //@line 1038 "fastmix.cpp"
  $81 = Math_imul($80, $76)|0; //@line 1038 "fastmix.cpp"
  $82 = (($81) + ($72))|0; //@line 1038 "fastmix.cpp"
  $83 = $26 | 6; //@line 1038 "fastmix.cpp"
  $84 = (1217698 + ($83<<1)|0); //@line 1038 "fastmix.cpp"
  $85 = HEAP16[$84>>1]|0; //@line 1038 "fastmix.cpp"
  $86 = $85 << 16 >> 16; //@line 1038 "fastmix.cpp"
  $87 = (($22) + 3)|0; //@line 1038 "fastmix.cpp"
  $88 = (($$) + ($87<<1)|0); //@line 1038 "fastmix.cpp"
  $89 = HEAP16[$88>>1]|0; //@line 1038 "fastmix.cpp"
  $90 = $89 << 16 >> 16; //@line 1038 "fastmix.cpp"
  $91 = Math_imul($90, $86)|0; //@line 1038 "fastmix.cpp"
  $92 = (($82) + ($91))|0; //@line 1038 "fastmix.cpp"
  $93 = $25 | 7; //@line 1038 "fastmix.cpp"
  $94 = (1217698 + ($93<<1)|0); //@line 1038 "fastmix.cpp"
  $95 = HEAP16[$94>>1]|0; //@line 1038 "fastmix.cpp"
  $96 = $95 << 16 >> 16; //@line 1038 "fastmix.cpp"
  $97 = (($22) + 4)|0; //@line 1038 "fastmix.cpp"
  $98 = (($$) + ($97<<1)|0); //@line 1038 "fastmix.cpp"
  $99 = HEAP16[$98>>1]|0; //@line 1038 "fastmix.cpp"
  $100 = $99 << 16 >> 16; //@line 1038 "fastmix.cpp"
  $101 = Math_imul($100, $96)|0; //@line 1038 "fastmix.cpp"
  $102 = (($92) + ($101))|0; //@line 1038 "fastmix.cpp"
  $103 = $63 >> 1; //@line 1038 "fastmix.cpp"
  $104 = $102 >> 1; //@line 1038 "fastmix.cpp"
  $105 = (($104) + ($103))|0; //@line 1038 "fastmix.cpp"
  $106 = $105 >> 14; //@line 1038 "fastmix.cpp"
  $107 = HEAP32[$16>>2]|0; //@line 1039 "fastmix.cpp"
  $108 = Math_imul($106, $107)|0; //@line 1039 "fastmix.cpp"
  $109 = HEAP32[$17>>2]|0; //@line 1039 "fastmix.cpp"
  $110 = Math_imul($109, $fy1$0)|0; //@line 1039 "fastmix.cpp"
  $111 = HEAP32[$18>>2]|0; //@line 1039 "fastmix.cpp"
  $112 = Math_imul($111, $fy2$0)|0; //@line 1039 "fastmix.cpp"
  $113 = (($110) + 4096)|0; //@line 1039 "fastmix.cpp"
  $114 = (($113) + ($112))|0; //@line 1039 "fastmix.cpp"
  $115 = (($114) + ($108))|0; //@line 1039 "fastmix.cpp"
  $116 = $115 >> 13; //@line 1039 "fastmix.cpp"
  $117 = HEAP32[$19>>2]|0; //@line 1040 "fastmix.cpp"
  $118 = Math_imul($116, $117)|0; //@line 1040 "fastmix.cpp"
  $119 = HEAP32[$pvol$0>>2]|0; //@line 1040 "fastmix.cpp"
  $120 = (($118) + ($119))|0; //@line 1040 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $120; //@line 1040 "fastmix.cpp"
  $121 = HEAP32[$20>>2]|0; //@line 1040 "fastmix.cpp"
  $122 = Math_imul($116, $121)|0; //@line 1040 "fastmix.cpp"
  $123 = ((($pvol$0)) + 4|0); //@line 1040 "fastmix.cpp"
  $124 = HEAP32[$123>>2]|0; //@line 1040 "fastmix.cpp"
  $125 = (($122) + ($124))|0; //@line 1040 "fastmix.cpp"
  HEAP32[$123>>2] = $125; //@line 1040 "fastmix.cpp"
  $126 = ((($pvol$0)) + 8|0); //@line 1040 "fastmix.cpp"
  $127 = HEAP32[$21>>2]|0; //@line 1041 "fastmix.cpp"
  $128 = (($127) + ($nPos$0))|0; //@line 1041 "fastmix.cpp"
  $129 = ($126>>>0)<($pbufmax>>>0); //@line 1041 "fastmix.cpp"
  if ($129) {
   $fy2$0$phi = $fy1$0;$fy1$0 = $116;$nPos$0 = $128;$pvol$0 = $126;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $116;$$lcssa6 = $128;$fy1$0$lcssa = $fy1$0;
   break;
  }
 }
 $130 = $$lcssa6 >> 16; //@line 1041 "fastmix.cpp"
 $131 = HEAP32[$7>>2]|0; //@line 1041 "fastmix.cpp"
 $132 = (($131) + ($130))|0; //@line 1041 "fastmix.cpp"
 HEAP32[$7>>2] = $132; //@line 1041 "fastmix.cpp"
 $133 = $$lcssa6 & 65535; //@line 1041 "fastmix.cpp"
 HEAP32[$4>>2] = $133; //@line 1041 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 1041 "fastmix.cpp"
 HEAP32[$2>>2] = $fy1$0$lcssa; //@line 1041 "fastmix.cpp"
 return; //@line 1041 "fastmix.cpp"
}
function __Z28FilterStereo8BitFirFilterMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa42 = 0, $$lcssa43 = 0, $$sum = 0, $$sum1 = 0, $$sum10 = 0, $$sum11 = 0, $$sum12 = 0, $$sum13 = 0, $$sum14 = 0, $$sum15 = 0, $$sum2 = 0, $$sum3 = 0, $$sum4 = 0, $$sum5 = 0, $$sum6 = 0, $$sum7 = 0, $$sum8 = 0, $$sum9 = 0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $fy3$0 = 0, $fy3$0$lcssa = 0, $fy4$0 = 0, $fy4$0$phi = 0, $nPos$0 = 0, $pvol$0 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 56|0); //@line 1144 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1144 "fastmix.cpp"
 $2 = ((($pChannel)) + 60|0); //@line 1144 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1144 "fastmix.cpp"
 $4 = ((($pChannel)) + 64|0); //@line 1144 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1144 "fastmix.cpp"
 $6 = ((($pChannel)) + 68|0); //@line 1144 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1144 "fastmix.cpp"
 $8 = ((($pChannel)) + 8|0); //@line 1145 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1145 "fastmix.cpp"
 $10 = HEAP32[$pChannel>>2]|0; //@line 1145 "fastmix.cpp"
 $11 = ((($pChannel)) + 4|0); //@line 1145 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1145 "fastmix.cpp"
 $13 = (($10) + ($12)|0); //@line 1145 "fastmix.cpp"
 $14 = ((($pChannel)) + 36|0); //@line 1145 "fastmix.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 1145 "fastmix.cpp"
 $16 = $15 & 64; //@line 1145 "fastmix.cpp"
 $17 = ($16|0)==(0); //@line 1145 "fastmix.cpp"
 $$ = $17 ? $10 : $13; //@line 1145 "fastmix.cpp"
 $18 = (($12) + -6)|0; //@line 1146 "fastmix.cpp"
 $19 = (($12) + -4)|0; //@line 1146 "fastmix.cpp"
 $20 = (($12) + -2)|0; //@line 1146 "fastmix.cpp"
 $21 = (($12) + 2)|0; //@line 1146 "fastmix.cpp"
 $22 = (($12) + 4)|0; //@line 1146 "fastmix.cpp"
 $23 = (($12) + 6)|0; //@line 1146 "fastmix.cpp"
 $24 = (($12) + 8)|0; //@line 1146 "fastmix.cpp"
 $25 = (($12) + -5)|0; //@line 1146 "fastmix.cpp"
 $26 = (($12) + -3)|0; //@line 1146 "fastmix.cpp"
 $27 = (($12) + -1)|0; //@line 1146 "fastmix.cpp"
 $28 = (($12) + 3)|0; //@line 1146 "fastmix.cpp"
 $29 = (($12) + 5)|0; //@line 1146 "fastmix.cpp"
 $30 = (($12) + 7)|0; //@line 1146 "fastmix.cpp"
 $31 = (($12) + 9)|0; //@line 1146 "fastmix.cpp"
 $32 = ((($pChannel)) + 72|0); //@line 1147 "fastmix.cpp"
 $33 = ((($pChannel)) + 76|0); //@line 1147 "fastmix.cpp"
 $34 = ((($pChannel)) + 80|0); //@line 1147 "fastmix.cpp"
 $35 = ((($pChannel)) + 16|0); //@line 1148 "fastmix.cpp"
 $36 = ((($pChannel)) + 20|0); //@line 1148 "fastmix.cpp"
 $37 = ((($pChannel)) + 12|0); //@line 1149 "fastmix.cpp"
 $fy1$0 = $1;$fy2$0 = $3;$fy3$0 = $5;$fy4$0 = $7;$nPos$0 = $9;$pvol$0 = $pbuffer;
 while(1) {
  $38 = $nPos$0 >> 16; //@line 1146 "fastmix.cpp"
  $39 = $nPos$0 & 65535; //@line 1146 "fastmix.cpp"
  $40 = (($39) + 16)|0; //@line 1146 "fastmix.cpp"
  $41 = $40 >>> 2; //@line 1146 "fastmix.cpp"
  $42 = $41 & 32760; //@line 1146 "fastmix.cpp"
  $43 = (1217698 + ($42<<1)|0); //@line 1146 "fastmix.cpp"
  $44 = HEAP16[$43>>1]|0; //@line 1146 "fastmix.cpp"
  $45 = $44 << 16 >> 16; //@line 1146 "fastmix.cpp"
  $46 = $38 << 1; //@line 1146 "fastmix.cpp"
  $$sum = (($18) + ($46))|0; //@line 1146 "fastmix.cpp"
  $47 = (($$) + ($$sum)|0); //@line 1146 "fastmix.cpp"
  $48 = HEAP8[$47>>0]|0; //@line 1146 "fastmix.cpp"
  $49 = $48 << 24 >> 24; //@line 1146 "fastmix.cpp"
  $50 = Math_imul($49, $45)|0; //@line 1146 "fastmix.cpp"
  $51 = $42 | 1; //@line 1146 "fastmix.cpp"
  $52 = (1217698 + ($51<<1)|0); //@line 1146 "fastmix.cpp"
  $53 = HEAP16[$52>>1]|0; //@line 1146 "fastmix.cpp"
  $54 = $53 << 16 >> 16; //@line 1146 "fastmix.cpp"
  $$sum1 = (($19) + ($46))|0; //@line 1146 "fastmix.cpp"
  $55 = (($$) + ($$sum1)|0); //@line 1146 "fastmix.cpp"
  $56 = HEAP8[$55>>0]|0; //@line 1146 "fastmix.cpp"
  $57 = $56 << 24 >> 24; //@line 1146 "fastmix.cpp"
  $58 = Math_imul($57, $54)|0; //@line 1146 "fastmix.cpp"
  $59 = (($58) + ($50))|0; //@line 1146 "fastmix.cpp"
  $60 = $42 | 2; //@line 1146 "fastmix.cpp"
  $61 = (1217698 + ($60<<1)|0); //@line 1146 "fastmix.cpp"
  $62 = HEAP16[$61>>1]|0; //@line 1146 "fastmix.cpp"
  $63 = $62 << 16 >> 16; //@line 1146 "fastmix.cpp"
  $$sum2 = (($20) + ($46))|0; //@line 1146 "fastmix.cpp"
  $64 = (($$) + ($$sum2)|0); //@line 1146 "fastmix.cpp"
  $65 = HEAP8[$64>>0]|0; //@line 1146 "fastmix.cpp"
  $66 = $65 << 24 >> 24; //@line 1146 "fastmix.cpp"
  $67 = Math_imul($66, $63)|0; //@line 1146 "fastmix.cpp"
  $68 = (($59) + ($67))|0; //@line 1146 "fastmix.cpp"
  $69 = $42 | 3; //@line 1146 "fastmix.cpp"
  $70 = (1217698 + ($69<<1)|0); //@line 1146 "fastmix.cpp"
  $71 = HEAP16[$70>>1]|0; //@line 1146 "fastmix.cpp"
  $72 = $71 << 16 >> 16; //@line 1146 "fastmix.cpp"
  $$sum3 = (($46) + ($12))|0; //@line 1146 "fastmix.cpp"
  $73 = (($$) + ($$sum3)|0); //@line 1146 "fastmix.cpp"
  $74 = HEAP8[$73>>0]|0; //@line 1146 "fastmix.cpp"
  $75 = $74 << 24 >> 24; //@line 1146 "fastmix.cpp"
  $76 = Math_imul($75, $72)|0; //@line 1146 "fastmix.cpp"
  $77 = (($68) + ($76))|0; //@line 1146 "fastmix.cpp"
  $78 = $42 | 4; //@line 1146 "fastmix.cpp"
  $79 = (1217698 + ($78<<1)|0); //@line 1146 "fastmix.cpp"
  $80 = HEAP16[$79>>1]|0; //@line 1146 "fastmix.cpp"
  $81 = $80 << 16 >> 16; //@line 1146 "fastmix.cpp"
  $$sum4 = (($21) + ($46))|0; //@line 1146 "fastmix.cpp"
  $82 = (($$) + ($$sum4)|0); //@line 1146 "fastmix.cpp"
  $83 = HEAP8[$82>>0]|0; //@line 1146 "fastmix.cpp"
  $84 = $83 << 24 >> 24; //@line 1146 "fastmix.cpp"
  $85 = Math_imul($84, $81)|0; //@line 1146 "fastmix.cpp"
  $86 = (($77) + ($85))|0; //@line 1146 "fastmix.cpp"
  $87 = $42 | 5; //@line 1146 "fastmix.cpp"
  $88 = (1217698 + ($87<<1)|0); //@line 1146 "fastmix.cpp"
  $89 = HEAP16[$88>>1]|0; //@line 1146 "fastmix.cpp"
  $90 = $89 << 16 >> 16; //@line 1146 "fastmix.cpp"
  $$sum5 = (($22) + ($46))|0; //@line 1146 "fastmix.cpp"
  $91 = (($$) + ($$sum5)|0); //@line 1146 "fastmix.cpp"
  $92 = HEAP8[$91>>0]|0; //@line 1146 "fastmix.cpp"
  $93 = $92 << 24 >> 24; //@line 1146 "fastmix.cpp"
  $94 = Math_imul($93, $90)|0; //@line 1146 "fastmix.cpp"
  $95 = (($86) + ($94))|0; //@line 1146 "fastmix.cpp"
  $96 = $42 | 6; //@line 1146 "fastmix.cpp"
  $97 = (1217698 + ($96<<1)|0); //@line 1146 "fastmix.cpp"
  $98 = HEAP16[$97>>1]|0; //@line 1146 "fastmix.cpp"
  $99 = $98 << 16 >> 16; //@line 1146 "fastmix.cpp"
  $$sum6 = (($23) + ($46))|0; //@line 1146 "fastmix.cpp"
  $100 = (($$) + ($$sum6)|0); //@line 1146 "fastmix.cpp"
  $101 = HEAP8[$100>>0]|0; //@line 1146 "fastmix.cpp"
  $102 = $101 << 24 >> 24; //@line 1146 "fastmix.cpp"
  $103 = Math_imul($102, $99)|0; //@line 1146 "fastmix.cpp"
  $104 = (($95) + ($103))|0; //@line 1146 "fastmix.cpp"
  $105 = $41 | 7; //@line 1146 "fastmix.cpp"
  $106 = (1217698 + ($105<<1)|0); //@line 1146 "fastmix.cpp"
  $107 = HEAP16[$106>>1]|0; //@line 1146 "fastmix.cpp"
  $108 = $107 << 16 >> 16; //@line 1146 "fastmix.cpp"
  $$sum7 = (($24) + ($46))|0; //@line 1146 "fastmix.cpp"
  $109 = (($$) + ($$sum7)|0); //@line 1146 "fastmix.cpp"
  $110 = HEAP8[$109>>0]|0; //@line 1146 "fastmix.cpp"
  $111 = $110 << 24 >> 24; //@line 1146 "fastmix.cpp"
  $112 = Math_imul($111, $108)|0; //@line 1146 "fastmix.cpp"
  $113 = (($104) + ($112))|0; //@line 1146 "fastmix.cpp"
  $114 = $113 >> 7; //@line 1146 "fastmix.cpp"
  $$sum8 = (($25) + ($46))|0; //@line 1146 "fastmix.cpp"
  $115 = (($$) + ($$sum8)|0); //@line 1146 "fastmix.cpp"
  $116 = HEAP8[$115>>0]|0; //@line 1146 "fastmix.cpp"
  $117 = $116 << 24 >> 24; //@line 1146 "fastmix.cpp"
  $118 = Math_imul($117, $45)|0; //@line 1146 "fastmix.cpp"
  $$sum9 = (($26) + ($46))|0; //@line 1146 "fastmix.cpp"
  $119 = (($$) + ($$sum9)|0); //@line 1146 "fastmix.cpp"
  $120 = HEAP8[$119>>0]|0; //@line 1146 "fastmix.cpp"
  $121 = $120 << 24 >> 24; //@line 1146 "fastmix.cpp"
  $122 = Math_imul($121, $54)|0; //@line 1146 "fastmix.cpp"
  $123 = (($122) + ($118))|0; //@line 1146 "fastmix.cpp"
  $$sum10 = (($27) + ($46))|0; //@line 1146 "fastmix.cpp"
  $124 = (($$) + ($$sum10)|0); //@line 1146 "fastmix.cpp"
  $125 = HEAP8[$124>>0]|0; //@line 1146 "fastmix.cpp"
  $126 = $125 << 24 >> 24; //@line 1146 "fastmix.cpp"
  $127 = Math_imul($126, $63)|0; //@line 1146 "fastmix.cpp"
  $128 = (($123) + ($127))|0; //@line 1146 "fastmix.cpp"
  $129 = $46 | 1; //@line 1146 "fastmix.cpp"
  $$sum11 = (($129) + ($12))|0; //@line 1146 "fastmix.cpp"
  $130 = (($$) + ($$sum11)|0); //@line 1146 "fastmix.cpp"
  $131 = HEAP8[$130>>0]|0; //@line 1146 "fastmix.cpp"
  $132 = $131 << 24 >> 24; //@line 1146 "fastmix.cpp"
  $133 = Math_imul($132, $72)|0; //@line 1146 "fastmix.cpp"
  $134 = (($128) + ($133))|0; //@line 1146 "fastmix.cpp"
  $$sum12 = (($28) + ($46))|0; //@line 1146 "fastmix.cpp"
  $135 = (($$) + ($$sum12)|0); //@line 1146 "fastmix.cpp"
  $136 = HEAP8[$135>>0]|0; //@line 1146 "fastmix.cpp"
  $137 = $136 << 24 >> 24; //@line 1146 "fastmix.cpp"
  $138 = Math_imul($137, $81)|0; //@line 1146 "fastmix.cpp"
  $139 = (($134) + ($138))|0; //@line 1146 "fastmix.cpp"
  $$sum13 = (($29) + ($46))|0; //@line 1146 "fastmix.cpp"
  $140 = (($$) + ($$sum13)|0); //@line 1146 "fastmix.cpp"
  $141 = HEAP8[$140>>0]|0; //@line 1146 "fastmix.cpp"
  $142 = $141 << 24 >> 24; //@line 1146 "fastmix.cpp"
  $143 = Math_imul($142, $90)|0; //@line 1146 "fastmix.cpp"
  $144 = (($139) + ($143))|0; //@line 1146 "fastmix.cpp"
  $$sum14 = (($30) + ($46))|0; //@line 1146 "fastmix.cpp"
  $145 = (($$) + ($$sum14)|0); //@line 1146 "fastmix.cpp"
  $146 = HEAP8[$145>>0]|0; //@line 1146 "fastmix.cpp"
  $147 = $146 << 24 >> 24; //@line 1146 "fastmix.cpp"
  $148 = Math_imul($147, $99)|0; //@line 1146 "fastmix.cpp"
  $149 = (($144) + ($148))|0; //@line 1146 "fastmix.cpp"
  $$sum15 = (($31) + ($46))|0; //@line 1146 "fastmix.cpp"
  $150 = (($$) + ($$sum15)|0); //@line 1146 "fastmix.cpp"
  $151 = HEAP8[$150>>0]|0; //@line 1146 "fastmix.cpp"
  $152 = $151 << 24 >> 24; //@line 1146 "fastmix.cpp"
  $153 = Math_imul($152, $108)|0; //@line 1146 "fastmix.cpp"
  $154 = (($149) + ($153))|0; //@line 1146 "fastmix.cpp"
  $155 = $154 >> 7; //@line 1146 "fastmix.cpp"
  $156 = HEAP32[$32>>2]|0; //@line 1147 "fastmix.cpp"
  $157 = Math_imul($156, $114)|0; //@line 1147 "fastmix.cpp"
  $158 = HEAP32[$33>>2]|0; //@line 1147 "fastmix.cpp"
  $159 = Math_imul($158, $fy1$0)|0; //@line 1147 "fastmix.cpp"
  $160 = HEAP32[$34>>2]|0; //@line 1147 "fastmix.cpp"
  $161 = Math_imul($160, $fy2$0)|0; //@line 1147 "fastmix.cpp"
  $162 = (($157) + 4096)|0; //@line 1147 "fastmix.cpp"
  $163 = (($162) + ($159))|0; //@line 1147 "fastmix.cpp"
  $164 = (($163) + ($161))|0; //@line 1147 "fastmix.cpp"
  $165 = $164 >> 13; //@line 1147 "fastmix.cpp"
  $166 = Math_imul($155, $156)|0; //@line 1147 "fastmix.cpp"
  $167 = Math_imul($158, $fy3$0)|0; //@line 1147 "fastmix.cpp"
  $168 = Math_imul($160, $fy4$0)|0; //@line 1147 "fastmix.cpp"
  $169 = (($167) + 4096)|0; //@line 1147 "fastmix.cpp"
  $170 = (($169) + ($168))|0; //@line 1147 "fastmix.cpp"
  $171 = (($170) + ($166))|0; //@line 1147 "fastmix.cpp"
  $172 = $171 >> 13; //@line 1147 "fastmix.cpp"
  $173 = HEAP32[$35>>2]|0; //@line 1148 "fastmix.cpp"
  $174 = Math_imul($165, $173)|0; //@line 1148 "fastmix.cpp"
  $175 = HEAP32[$pvol$0>>2]|0; //@line 1148 "fastmix.cpp"
  $176 = (($174) + ($175))|0; //@line 1148 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $176; //@line 1148 "fastmix.cpp"
  $177 = HEAP32[$36>>2]|0; //@line 1148 "fastmix.cpp"
  $178 = Math_imul($172, $177)|0; //@line 1148 "fastmix.cpp"
  $179 = ((($pvol$0)) + 4|0); //@line 1148 "fastmix.cpp"
  $180 = HEAP32[$179>>2]|0; //@line 1148 "fastmix.cpp"
  $181 = (($178) + ($180))|0; //@line 1148 "fastmix.cpp"
  HEAP32[$179>>2] = $181; //@line 1148 "fastmix.cpp"
  $182 = ((($pvol$0)) + 8|0); //@line 1148 "fastmix.cpp"
  $183 = HEAP32[$37>>2]|0; //@line 1149 "fastmix.cpp"
  $184 = (($183) + ($nPos$0))|0; //@line 1149 "fastmix.cpp"
  $185 = ($182>>>0)<($pbufmax>>>0); //@line 1149 "fastmix.cpp"
  if ($185) {
   $fy4$0$phi = $fy3$0;$fy2$0$phi = $fy1$0;$fy1$0 = $165;$fy3$0 = $172;$nPos$0 = $184;$pvol$0 = $182;$fy4$0 = $fy4$0$phi;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $165;$$lcssa42 = $172;$$lcssa43 = $184;$fy1$0$lcssa = $fy1$0;$fy3$0$lcssa = $fy3$0;
   break;
  }
 }
 $186 = $$lcssa43 >> 16; //@line 1149 "fastmix.cpp"
 $187 = HEAP32[$11>>2]|0; //@line 1149 "fastmix.cpp"
 $188 = (($187) + ($186))|0; //@line 1149 "fastmix.cpp"
 HEAP32[$11>>2] = $188; //@line 1149 "fastmix.cpp"
 $189 = $$lcssa43 & 65535; //@line 1149 "fastmix.cpp"
 HEAP32[$8>>2] = $189; //@line 1149 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 1149 "fastmix.cpp"
 HEAP32[$2>>2] = $fy1$0$lcssa; //@line 1149 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa42; //@line 1149 "fastmix.cpp"
 HEAP32[$6>>2] = $fy3$0$lcssa; //@line 1149 "fastmix.cpp"
 return; //@line 1149 "fastmix.cpp"
}
function __Z29FilterStereo16BitFirFilterMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa11 = 0, $$lcssa12 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $fy3$0 = 0, $fy3$0$lcssa = 0, $fy4$0 = 0, $fy4$0$phi = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 56|0); //@line 1151 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1151 "fastmix.cpp"
 $2 = ((($pChannel)) + 60|0); //@line 1151 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1151 "fastmix.cpp"
 $4 = ((($pChannel)) + 64|0); //@line 1151 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1151 "fastmix.cpp"
 $6 = ((($pChannel)) + 68|0); //@line 1151 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1151 "fastmix.cpp"
 $8 = ((($pChannel)) + 8|0); //@line 1152 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1152 "fastmix.cpp"
 $10 = HEAP32[$pChannel>>2]|0; //@line 1152 "fastmix.cpp"
 $11 = ((($pChannel)) + 4|0); //@line 1152 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1152 "fastmix.cpp"
 $13 = $12 << 1; //@line 1152 "fastmix.cpp"
 $14 = (($10) + ($13)|0); //@line 1152 "fastmix.cpp"
 $15 = ((($pChannel)) + 36|0); //@line 1152 "fastmix.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1152 "fastmix.cpp"
 $17 = $16 & 64; //@line 1152 "fastmix.cpp"
 $18 = ($17|0)==(0); //@line 1152 "fastmix.cpp"
 $19 = (($14) + ($12<<1)|0); //@line 1152 "fastmix.cpp"
 $$ = $18 ? $14 : $19; //@line 1152 "fastmix.cpp"
 $20 = ((($pChannel)) + 72|0); //@line 1154 "fastmix.cpp"
 $21 = ((($pChannel)) + 76|0); //@line 1154 "fastmix.cpp"
 $22 = ((($pChannel)) + 80|0); //@line 1154 "fastmix.cpp"
 $23 = ((($pChannel)) + 16|0); //@line 1155 "fastmix.cpp"
 $24 = ((($pChannel)) + 20|0); //@line 1155 "fastmix.cpp"
 $25 = ((($pChannel)) + 12|0); //@line 1156 "fastmix.cpp"
 $fy1$0 = $1;$fy2$0 = $3;$fy3$0 = $5;$fy4$0 = $7;$nPos$0 = $9;$pvol$0 = $pbuffer;
 while(1) {
  $26 = $nPos$0 >> 16; //@line 1153 "fastmix.cpp"
  $27 = $nPos$0 & 65535; //@line 1153 "fastmix.cpp"
  $28 = (($27) + 16)|0; //@line 1153 "fastmix.cpp"
  $29 = $28 >>> 2; //@line 1153 "fastmix.cpp"
  $30 = $29 & 32760; //@line 1153 "fastmix.cpp"
  $31 = (1217698 + ($30<<1)|0); //@line 1153 "fastmix.cpp"
  $32 = HEAP16[$31>>1]|0; //@line 1153 "fastmix.cpp"
  $33 = $32 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $34 = $26 << 1; //@line 1153 "fastmix.cpp"
  $35 = (($34) + -6)|0; //@line 1153 "fastmix.cpp"
  $36 = (($$) + ($35<<1)|0); //@line 1153 "fastmix.cpp"
  $37 = HEAP16[$36>>1]|0; //@line 1153 "fastmix.cpp"
  $38 = $37 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $39 = Math_imul($38, $33)|0; //@line 1153 "fastmix.cpp"
  $40 = $30 | 1; //@line 1153 "fastmix.cpp"
  $41 = (1217698 + ($40<<1)|0); //@line 1153 "fastmix.cpp"
  $42 = HEAP16[$41>>1]|0; //@line 1153 "fastmix.cpp"
  $43 = $42 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $44 = (($34) + -4)|0; //@line 1153 "fastmix.cpp"
  $45 = (($$) + ($44<<1)|0); //@line 1153 "fastmix.cpp"
  $46 = HEAP16[$45>>1]|0; //@line 1153 "fastmix.cpp"
  $47 = $46 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $48 = Math_imul($47, $43)|0; //@line 1153 "fastmix.cpp"
  $49 = (($48) + ($39))|0; //@line 1153 "fastmix.cpp"
  $50 = $30 | 2; //@line 1153 "fastmix.cpp"
  $51 = (1217698 + ($50<<1)|0); //@line 1153 "fastmix.cpp"
  $52 = HEAP16[$51>>1]|0; //@line 1153 "fastmix.cpp"
  $53 = $52 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $54 = (($34) + -2)|0; //@line 1153 "fastmix.cpp"
  $55 = (($$) + ($54<<1)|0); //@line 1153 "fastmix.cpp"
  $56 = HEAP16[$55>>1]|0; //@line 1153 "fastmix.cpp"
  $57 = $56 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $58 = Math_imul($57, $53)|0; //@line 1153 "fastmix.cpp"
  $59 = (($49) + ($58))|0; //@line 1153 "fastmix.cpp"
  $60 = $30 | 3; //@line 1153 "fastmix.cpp"
  $61 = (1217698 + ($60<<1)|0); //@line 1153 "fastmix.cpp"
  $62 = HEAP16[$61>>1]|0; //@line 1153 "fastmix.cpp"
  $63 = $62 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $64 = (($$) + ($34<<1)|0); //@line 1153 "fastmix.cpp"
  $65 = HEAP16[$64>>1]|0; //@line 1153 "fastmix.cpp"
  $66 = $65 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $67 = Math_imul($66, $63)|0; //@line 1153 "fastmix.cpp"
  $68 = (($59) + ($67))|0; //@line 1153 "fastmix.cpp"
  $69 = $30 | 4; //@line 1153 "fastmix.cpp"
  $70 = (1217698 + ($69<<1)|0); //@line 1153 "fastmix.cpp"
  $71 = HEAP16[$70>>1]|0; //@line 1153 "fastmix.cpp"
  $72 = $71 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $73 = (($34) + 2)|0; //@line 1153 "fastmix.cpp"
  $74 = (($$) + ($73<<1)|0); //@line 1153 "fastmix.cpp"
  $75 = HEAP16[$74>>1]|0; //@line 1153 "fastmix.cpp"
  $76 = $75 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $77 = Math_imul($76, $72)|0; //@line 1153 "fastmix.cpp"
  $78 = $30 | 5; //@line 1153 "fastmix.cpp"
  $79 = (1217698 + ($78<<1)|0); //@line 1153 "fastmix.cpp"
  $80 = HEAP16[$79>>1]|0; //@line 1153 "fastmix.cpp"
  $81 = $80 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $82 = (($34) + 4)|0; //@line 1153 "fastmix.cpp"
  $83 = (($$) + ($82<<1)|0); //@line 1153 "fastmix.cpp"
  $84 = HEAP16[$83>>1]|0; //@line 1153 "fastmix.cpp"
  $85 = $84 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $86 = Math_imul($85, $81)|0; //@line 1153 "fastmix.cpp"
  $87 = (($86) + ($77))|0; //@line 1153 "fastmix.cpp"
  $88 = $30 | 6; //@line 1153 "fastmix.cpp"
  $89 = (1217698 + ($88<<1)|0); //@line 1153 "fastmix.cpp"
  $90 = HEAP16[$89>>1]|0; //@line 1153 "fastmix.cpp"
  $91 = $90 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $92 = (($34) + 6)|0; //@line 1153 "fastmix.cpp"
  $93 = (($$) + ($92<<1)|0); //@line 1153 "fastmix.cpp"
  $94 = HEAP16[$93>>1]|0; //@line 1153 "fastmix.cpp"
  $95 = $94 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $96 = Math_imul($95, $91)|0; //@line 1153 "fastmix.cpp"
  $97 = (($87) + ($96))|0; //@line 1153 "fastmix.cpp"
  $98 = $29 | 7; //@line 1153 "fastmix.cpp"
  $99 = (1217698 + ($98<<1)|0); //@line 1153 "fastmix.cpp"
  $100 = HEAP16[$99>>1]|0; //@line 1153 "fastmix.cpp"
  $101 = $100 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $102 = (($34) + 8)|0; //@line 1153 "fastmix.cpp"
  $103 = (($$) + ($102<<1)|0); //@line 1153 "fastmix.cpp"
  $104 = HEAP16[$103>>1]|0; //@line 1153 "fastmix.cpp"
  $105 = $104 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $106 = Math_imul($105, $101)|0; //@line 1153 "fastmix.cpp"
  $107 = (($97) + ($106))|0; //@line 1153 "fastmix.cpp"
  $108 = $68 >> 1; //@line 1153 "fastmix.cpp"
  $109 = $107 >> 1; //@line 1153 "fastmix.cpp"
  $110 = (($109) + ($108))|0; //@line 1153 "fastmix.cpp"
  $111 = $110 >> 14; //@line 1153 "fastmix.cpp"
  $112 = (($34) + -5)|0; //@line 1153 "fastmix.cpp"
  $113 = (($$) + ($112<<1)|0); //@line 1153 "fastmix.cpp"
  $114 = HEAP16[$113>>1]|0; //@line 1153 "fastmix.cpp"
  $115 = $114 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $116 = Math_imul($115, $33)|0; //@line 1153 "fastmix.cpp"
  $117 = (($34) + -3)|0; //@line 1153 "fastmix.cpp"
  $118 = (($$) + ($117<<1)|0); //@line 1153 "fastmix.cpp"
  $119 = HEAP16[$118>>1]|0; //@line 1153 "fastmix.cpp"
  $120 = $119 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $121 = Math_imul($120, $43)|0; //@line 1153 "fastmix.cpp"
  $122 = (($121) + ($116))|0; //@line 1153 "fastmix.cpp"
  $123 = (($34) + -1)|0; //@line 1153 "fastmix.cpp"
  $124 = (($$) + ($123<<1)|0); //@line 1153 "fastmix.cpp"
  $125 = HEAP16[$124>>1]|0; //@line 1153 "fastmix.cpp"
  $126 = $125 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $127 = Math_imul($126, $53)|0; //@line 1153 "fastmix.cpp"
  $128 = (($122) + ($127))|0; //@line 1153 "fastmix.cpp"
  $129 = $34 | 1; //@line 1153 "fastmix.cpp"
  $130 = (($$) + ($129<<1)|0); //@line 1153 "fastmix.cpp"
  $131 = HEAP16[$130>>1]|0; //@line 1153 "fastmix.cpp"
  $132 = $131 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $133 = Math_imul($132, $63)|0; //@line 1153 "fastmix.cpp"
  $134 = (($128) + ($133))|0; //@line 1153 "fastmix.cpp"
  $135 = (($34) + 3)|0; //@line 1153 "fastmix.cpp"
  $136 = (($$) + ($135<<1)|0); //@line 1153 "fastmix.cpp"
  $137 = HEAP16[$136>>1]|0; //@line 1153 "fastmix.cpp"
  $138 = $137 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $139 = Math_imul($138, $72)|0; //@line 1153 "fastmix.cpp"
  $140 = (($34) + 5)|0; //@line 1153 "fastmix.cpp"
  $141 = (($$) + ($140<<1)|0); //@line 1153 "fastmix.cpp"
  $142 = HEAP16[$141>>1]|0; //@line 1153 "fastmix.cpp"
  $143 = $142 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $144 = Math_imul($143, $81)|0; //@line 1153 "fastmix.cpp"
  $145 = (($144) + ($139))|0; //@line 1153 "fastmix.cpp"
  $146 = (($34) + 7)|0; //@line 1153 "fastmix.cpp"
  $147 = (($$) + ($146<<1)|0); //@line 1153 "fastmix.cpp"
  $148 = HEAP16[$147>>1]|0; //@line 1153 "fastmix.cpp"
  $149 = $148 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $150 = Math_imul($149, $91)|0; //@line 1153 "fastmix.cpp"
  $151 = (($145) + ($150))|0; //@line 1153 "fastmix.cpp"
  $152 = (($34) + 9)|0; //@line 1153 "fastmix.cpp"
  $153 = (($$) + ($152<<1)|0); //@line 1153 "fastmix.cpp"
  $154 = HEAP16[$153>>1]|0; //@line 1153 "fastmix.cpp"
  $155 = $154 << 16 >> 16; //@line 1153 "fastmix.cpp"
  $156 = Math_imul($155, $101)|0; //@line 1153 "fastmix.cpp"
  $157 = (($151) + ($156))|0; //@line 1153 "fastmix.cpp"
  $158 = $134 >> 1; //@line 1153 "fastmix.cpp"
  $159 = $157 >> 1; //@line 1153 "fastmix.cpp"
  $160 = (($159) + ($158))|0; //@line 1153 "fastmix.cpp"
  $161 = $160 >> 14; //@line 1153 "fastmix.cpp"
  $162 = HEAP32[$20>>2]|0; //@line 1154 "fastmix.cpp"
  $163 = Math_imul($162, $111)|0; //@line 1154 "fastmix.cpp"
  $164 = HEAP32[$21>>2]|0; //@line 1154 "fastmix.cpp"
  $165 = Math_imul($164, $fy1$0)|0; //@line 1154 "fastmix.cpp"
  $166 = HEAP32[$22>>2]|0; //@line 1154 "fastmix.cpp"
  $167 = Math_imul($166, $fy2$0)|0; //@line 1154 "fastmix.cpp"
  $168 = (($163) + 4096)|0; //@line 1154 "fastmix.cpp"
  $169 = (($168) + ($165))|0; //@line 1154 "fastmix.cpp"
  $170 = (($169) + ($167))|0; //@line 1154 "fastmix.cpp"
  $171 = $170 >> 13; //@line 1154 "fastmix.cpp"
  $172 = Math_imul($161, $162)|0; //@line 1154 "fastmix.cpp"
  $173 = Math_imul($164, $fy3$0)|0; //@line 1154 "fastmix.cpp"
  $174 = Math_imul($166, $fy4$0)|0; //@line 1154 "fastmix.cpp"
  $175 = (($173) + 4096)|0; //@line 1154 "fastmix.cpp"
  $176 = (($175) + ($174))|0; //@line 1154 "fastmix.cpp"
  $177 = (($176) + ($172))|0; //@line 1154 "fastmix.cpp"
  $178 = $177 >> 13; //@line 1154 "fastmix.cpp"
  $179 = HEAP32[$23>>2]|0; //@line 1155 "fastmix.cpp"
  $180 = Math_imul($171, $179)|0; //@line 1155 "fastmix.cpp"
  $181 = HEAP32[$pvol$0>>2]|0; //@line 1155 "fastmix.cpp"
  $182 = (($180) + ($181))|0; //@line 1155 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $182; //@line 1155 "fastmix.cpp"
  $183 = HEAP32[$24>>2]|0; //@line 1155 "fastmix.cpp"
  $184 = Math_imul($178, $183)|0; //@line 1155 "fastmix.cpp"
  $185 = ((($pvol$0)) + 4|0); //@line 1155 "fastmix.cpp"
  $186 = HEAP32[$185>>2]|0; //@line 1155 "fastmix.cpp"
  $187 = (($184) + ($186))|0; //@line 1155 "fastmix.cpp"
  HEAP32[$185>>2] = $187; //@line 1155 "fastmix.cpp"
  $188 = ((($pvol$0)) + 8|0); //@line 1155 "fastmix.cpp"
  $189 = HEAP32[$25>>2]|0; //@line 1156 "fastmix.cpp"
  $190 = (($189) + ($nPos$0))|0; //@line 1156 "fastmix.cpp"
  $191 = ($188>>>0)<($pbufmax>>>0); //@line 1156 "fastmix.cpp"
  if ($191) {
   $fy4$0$phi = $fy3$0;$fy2$0$phi = $fy1$0;$fy1$0 = $171;$fy3$0 = $178;$nPos$0 = $190;$pvol$0 = $188;$fy4$0 = $fy4$0$phi;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $171;$$lcssa11 = $178;$$lcssa12 = $190;$fy1$0$lcssa = $fy1$0;$fy3$0$lcssa = $fy3$0;
   break;
  }
 }
 $192 = $$lcssa12 >> 16; //@line 1156 "fastmix.cpp"
 $193 = HEAP32[$11>>2]|0; //@line 1156 "fastmix.cpp"
 $194 = (($193) + ($192))|0; //@line 1156 "fastmix.cpp"
 HEAP32[$11>>2] = $194; //@line 1156 "fastmix.cpp"
 $195 = $$lcssa12 & 65535; //@line 1156 "fastmix.cpp"
 HEAP32[$8>>2] = $195; //@line 1156 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 1156 "fastmix.cpp"
 HEAP32[$2>>2] = $fy1$0$lcssa; //@line 1156 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa11; //@line 1156 "fastmix.cpp"
 HEAP32[$6>>2] = $fy3$0$lcssa; //@line 1156 "fastmix.cpp"
 return; //@line 1156 "fastmix.cpp"
}
function __Z30FilterMono8BitFirFilterRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa41 = 0, $$lcssa42 = 0, $$lcssa43 = 0, $$lcssa44 = 0, $$lcssa45 = 0, $$sum = 0, $$sum1 = 0, $$sum2 = 0, $$sum3 = 0, $$sum4 = 0, $$sum5 = 0, $$sum6 = 0, $$sum7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $nPos$0 = 0;
 var $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 1086 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1086 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 1086 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1086 "fastmix.cpp"
 $4 = ((($pChannel)) + 56|0); //@line 1086 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1086 "fastmix.cpp"
 $6 = ((($pChannel)) + 60|0); //@line 1086 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1086 "fastmix.cpp"
 $8 = ((($pChannel)) + 8|0); //@line 1087 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1087 "fastmix.cpp"
 $10 = HEAP32[$pChannel>>2]|0; //@line 1087 "fastmix.cpp"
 $11 = ((($pChannel)) + 4|0); //@line 1087 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1087 "fastmix.cpp"
 $13 = (($10) + ($12)|0); //@line 1087 "fastmix.cpp"
 $14 = ((($pChannel)) + 36|0); //@line 1087 "fastmix.cpp"
 $15 = HEAP32[$14>>2]|0; //@line 1087 "fastmix.cpp"
 $16 = $15 & 64; //@line 1087 "fastmix.cpp"
 $17 = ($16|0)==(0); //@line 1087 "fastmix.cpp"
 $$ = $17 ? $10 : $13; //@line 1087 "fastmix.cpp"
 $18 = (($12) + -3)|0; //@line 1088 "fastmix.cpp"
 $19 = (($12) + -2)|0; //@line 1088 "fastmix.cpp"
 $20 = (($12) + -1)|0; //@line 1088 "fastmix.cpp"
 $21 = (($12) + 1)|0; //@line 1088 "fastmix.cpp"
 $22 = (($12) + 2)|0; //@line 1088 "fastmix.cpp"
 $23 = (($12) + 3)|0; //@line 1088 "fastmix.cpp"
 $24 = (($12) + 4)|0; //@line 1088 "fastmix.cpp"
 $25 = ((($pChannel)) + 72|0); //@line 1089 "fastmix.cpp"
 $26 = ((($pChannel)) + 76|0); //@line 1089 "fastmix.cpp"
 $27 = ((($pChannel)) + 80|0); //@line 1089 "fastmix.cpp"
 $28 = ((($pChannel)) + 28|0); //@line 1090 "fastmix.cpp"
 $29 = ((($pChannel)) + 24|0); //@line 1090 "fastmix.cpp"
 $30 = ((($pChannel)) + 12|0); //@line 1091 "fastmix.cpp"
 $fy1$0 = $5;$fy2$0 = $7;$nPos$0 = $9;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $31 = $nPos$0 >> 16; //@line 1088 "fastmix.cpp"
  $32 = $nPos$0 & 65535; //@line 1088 "fastmix.cpp"
  $33 = (($32) + 16)|0; //@line 1088 "fastmix.cpp"
  $34 = $33 >>> 2; //@line 1088 "fastmix.cpp"
  $35 = $34 & 32760; //@line 1088 "fastmix.cpp"
  $36 = (1217698 + ($35<<1)|0); //@line 1088 "fastmix.cpp"
  $37 = HEAP16[$36>>1]|0; //@line 1088 "fastmix.cpp"
  $38 = $37 << 16 >> 16; //@line 1088 "fastmix.cpp"
  $$sum = (($18) + ($31))|0; //@line 1088 "fastmix.cpp"
  $39 = (($$) + ($$sum)|0); //@line 1088 "fastmix.cpp"
  $40 = HEAP8[$39>>0]|0; //@line 1088 "fastmix.cpp"
  $41 = $40 << 24 >> 24; //@line 1088 "fastmix.cpp"
  $42 = Math_imul($41, $38)|0; //@line 1088 "fastmix.cpp"
  $43 = $35 | 1; //@line 1088 "fastmix.cpp"
  $44 = (1217698 + ($43<<1)|0); //@line 1088 "fastmix.cpp"
  $45 = HEAP16[$44>>1]|0; //@line 1088 "fastmix.cpp"
  $46 = $45 << 16 >> 16; //@line 1088 "fastmix.cpp"
  $$sum1 = (($19) + ($31))|0; //@line 1088 "fastmix.cpp"
  $47 = (($$) + ($$sum1)|0); //@line 1088 "fastmix.cpp"
  $48 = HEAP8[$47>>0]|0; //@line 1088 "fastmix.cpp"
  $49 = $48 << 24 >> 24; //@line 1088 "fastmix.cpp"
  $50 = Math_imul($49, $46)|0; //@line 1088 "fastmix.cpp"
  $51 = (($50) + ($42))|0; //@line 1088 "fastmix.cpp"
  $52 = $35 | 2; //@line 1088 "fastmix.cpp"
  $53 = (1217698 + ($52<<1)|0); //@line 1088 "fastmix.cpp"
  $54 = HEAP16[$53>>1]|0; //@line 1088 "fastmix.cpp"
  $55 = $54 << 16 >> 16; //@line 1088 "fastmix.cpp"
  $$sum2 = (($20) + ($31))|0; //@line 1088 "fastmix.cpp"
  $56 = (($$) + ($$sum2)|0); //@line 1088 "fastmix.cpp"
  $57 = HEAP8[$56>>0]|0; //@line 1088 "fastmix.cpp"
  $58 = $57 << 24 >> 24; //@line 1088 "fastmix.cpp"
  $59 = Math_imul($58, $55)|0; //@line 1088 "fastmix.cpp"
  $60 = (($51) + ($59))|0; //@line 1088 "fastmix.cpp"
  $61 = $35 | 3; //@line 1088 "fastmix.cpp"
  $62 = (1217698 + ($61<<1)|0); //@line 1088 "fastmix.cpp"
  $63 = HEAP16[$62>>1]|0; //@line 1088 "fastmix.cpp"
  $64 = $63 << 16 >> 16; //@line 1088 "fastmix.cpp"
  $$sum3 = (($31) + ($12))|0; //@line 1088 "fastmix.cpp"
  $65 = (($$) + ($$sum3)|0); //@line 1088 "fastmix.cpp"
  $66 = HEAP8[$65>>0]|0; //@line 1088 "fastmix.cpp"
  $67 = $66 << 24 >> 24; //@line 1088 "fastmix.cpp"
  $68 = Math_imul($67, $64)|0; //@line 1088 "fastmix.cpp"
  $69 = (($60) + ($68))|0; //@line 1088 "fastmix.cpp"
  $70 = $35 | 4; //@line 1088 "fastmix.cpp"
  $71 = (1217698 + ($70<<1)|0); //@line 1088 "fastmix.cpp"
  $72 = HEAP16[$71>>1]|0; //@line 1088 "fastmix.cpp"
  $73 = $72 << 16 >> 16; //@line 1088 "fastmix.cpp"
  $$sum4 = (($21) + ($31))|0; //@line 1088 "fastmix.cpp"
  $74 = (($$) + ($$sum4)|0); //@line 1088 "fastmix.cpp"
  $75 = HEAP8[$74>>0]|0; //@line 1088 "fastmix.cpp"
  $76 = $75 << 24 >> 24; //@line 1088 "fastmix.cpp"
  $77 = Math_imul($76, $73)|0; //@line 1088 "fastmix.cpp"
  $78 = (($69) + ($77))|0; //@line 1088 "fastmix.cpp"
  $79 = $35 | 5; //@line 1088 "fastmix.cpp"
  $80 = (1217698 + ($79<<1)|0); //@line 1088 "fastmix.cpp"
  $81 = HEAP16[$80>>1]|0; //@line 1088 "fastmix.cpp"
  $82 = $81 << 16 >> 16; //@line 1088 "fastmix.cpp"
  $$sum5 = (($22) + ($31))|0; //@line 1088 "fastmix.cpp"
  $83 = (($$) + ($$sum5)|0); //@line 1088 "fastmix.cpp"
  $84 = HEAP8[$83>>0]|0; //@line 1088 "fastmix.cpp"
  $85 = $84 << 24 >> 24; //@line 1088 "fastmix.cpp"
  $86 = Math_imul($85, $82)|0; //@line 1088 "fastmix.cpp"
  $87 = (($78) + ($86))|0; //@line 1088 "fastmix.cpp"
  $88 = $35 | 6; //@line 1088 "fastmix.cpp"
  $89 = (1217698 + ($88<<1)|0); //@line 1088 "fastmix.cpp"
  $90 = HEAP16[$89>>1]|0; //@line 1088 "fastmix.cpp"
  $91 = $90 << 16 >> 16; //@line 1088 "fastmix.cpp"
  $$sum6 = (($23) + ($31))|0; //@line 1088 "fastmix.cpp"
  $92 = (($$) + ($$sum6)|0); //@line 1088 "fastmix.cpp"
  $93 = HEAP8[$92>>0]|0; //@line 1088 "fastmix.cpp"
  $94 = $93 << 24 >> 24; //@line 1088 "fastmix.cpp"
  $95 = Math_imul($94, $91)|0; //@line 1088 "fastmix.cpp"
  $96 = (($87) + ($95))|0; //@line 1088 "fastmix.cpp"
  $97 = $34 | 7; //@line 1088 "fastmix.cpp"
  $98 = (1217698 + ($97<<1)|0); //@line 1088 "fastmix.cpp"
  $99 = HEAP16[$98>>1]|0; //@line 1088 "fastmix.cpp"
  $100 = $99 << 16 >> 16; //@line 1088 "fastmix.cpp"
  $$sum7 = (($24) + ($31))|0; //@line 1088 "fastmix.cpp"
  $101 = (($$) + ($$sum7)|0); //@line 1088 "fastmix.cpp"
  $102 = HEAP8[$101>>0]|0; //@line 1088 "fastmix.cpp"
  $103 = $102 << 24 >> 24; //@line 1088 "fastmix.cpp"
  $104 = Math_imul($103, $100)|0; //@line 1088 "fastmix.cpp"
  $105 = (($96) + ($104))|0; //@line 1088 "fastmix.cpp"
  $106 = $105 >> 7; //@line 1088 "fastmix.cpp"
  $107 = HEAP32[$25>>2]|0; //@line 1089 "fastmix.cpp"
  $108 = Math_imul($106, $107)|0; //@line 1089 "fastmix.cpp"
  $109 = HEAP32[$26>>2]|0; //@line 1089 "fastmix.cpp"
  $110 = Math_imul($109, $fy1$0)|0; //@line 1089 "fastmix.cpp"
  $111 = HEAP32[$27>>2]|0; //@line 1089 "fastmix.cpp"
  $112 = Math_imul($111, $fy2$0)|0; //@line 1089 "fastmix.cpp"
  $113 = (($110) + 4096)|0; //@line 1089 "fastmix.cpp"
  $114 = (($113) + ($112))|0; //@line 1089 "fastmix.cpp"
  $115 = (($114) + ($108))|0; //@line 1089 "fastmix.cpp"
  $116 = $115 >> 13; //@line 1089 "fastmix.cpp"
  $117 = HEAP32[$28>>2]|0; //@line 1090 "fastmix.cpp"
  $118 = (($117) + ($nRampLeftVol$0))|0; //@line 1090 "fastmix.cpp"
  $119 = HEAP32[$29>>2]|0; //@line 1090 "fastmix.cpp"
  $120 = (($119) + ($nRampRightVol$0))|0; //@line 1090 "fastmix.cpp"
  $121 = $120 >> 12; //@line 1090 "fastmix.cpp"
  $122 = Math_imul($116, $121)|0; //@line 1090 "fastmix.cpp"
  $123 = HEAP32[$pvol$0>>2]|0; //@line 1090 "fastmix.cpp"
  $124 = (($122) + ($123))|0; //@line 1090 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $124; //@line 1090 "fastmix.cpp"
  $125 = $118 >> 12; //@line 1090 "fastmix.cpp"
  $126 = Math_imul($116, $125)|0; //@line 1090 "fastmix.cpp"
  $127 = ((($pvol$0)) + 4|0); //@line 1090 "fastmix.cpp"
  $128 = HEAP32[$127>>2]|0; //@line 1090 "fastmix.cpp"
  $129 = (($126) + ($128))|0; //@line 1090 "fastmix.cpp"
  HEAP32[$127>>2] = $129; //@line 1090 "fastmix.cpp"
  $130 = ((($pvol$0)) + 8|0); //@line 1090 "fastmix.cpp"
  $131 = HEAP32[$30>>2]|0; //@line 1091 "fastmix.cpp"
  $132 = (($131) + ($nPos$0))|0; //@line 1091 "fastmix.cpp"
  $133 = ($130>>>0)<($pbufmax>>>0); //@line 1091 "fastmix.cpp"
  if ($133) {
   $fy2$0$phi = $fy1$0;$fy1$0 = $116;$nPos$0 = $132;$nRampLeftVol$0 = $118;$nRampRightVol$0 = $120;$pvol$0 = $130;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $116;$$lcssa41 = $118;$$lcssa42 = $120;$$lcssa43 = $121;$$lcssa44 = $125;$$lcssa45 = $132;$fy1$0$lcssa = $fy1$0;
   break;
  }
 }
 $134 = $$lcssa45 >> 16; //@line 1091 "fastmix.cpp"
 $135 = HEAP32[$11>>2]|0; //@line 1091 "fastmix.cpp"
 $136 = (($135) + ($134))|0; //@line 1091 "fastmix.cpp"
 HEAP32[$11>>2] = $136; //@line 1091 "fastmix.cpp"
 $137 = $$lcssa45 & 65535; //@line 1091 "fastmix.cpp"
 HEAP32[$8>>2] = $137; //@line 1091 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa; //@line 1091 "fastmix.cpp"
 HEAP32[$6>>2] = $fy1$0$lcssa; //@line 1091 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa42; //@line 1091 "fastmix.cpp"
 $138 = ((($pChannel)) + 16|0); //@line 1091 "fastmix.cpp"
 HEAP32[$138>>2] = $$lcssa43; //@line 1091 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa41; //@line 1091 "fastmix.cpp"
 $139 = ((($pChannel)) + 20|0); //@line 1091 "fastmix.cpp"
 HEAP32[$139>>2] = $$lcssa44; //@line 1091 "fastmix.cpp"
 return; //@line 1091 "fastmix.cpp"
}
function __Z31FilterMono16BitFirFilterRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa26 = 0, $$lcssa27 = 0, $$lcssa28 = 0, $$lcssa29 = 0, $$lcssa30 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 1093 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1093 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 1093 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1093 "fastmix.cpp"
 $4 = ((($pChannel)) + 56|0); //@line 1093 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1093 "fastmix.cpp"
 $6 = ((($pChannel)) + 60|0); //@line 1093 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1093 "fastmix.cpp"
 $8 = ((($pChannel)) + 8|0); //@line 1094 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1094 "fastmix.cpp"
 $10 = HEAP32[$pChannel>>2]|0; //@line 1094 "fastmix.cpp"
 $11 = ((($pChannel)) + 4|0); //@line 1094 "fastmix.cpp"
 $12 = HEAP32[$11>>2]|0; //@line 1094 "fastmix.cpp"
 $13 = $12 << 1; //@line 1094 "fastmix.cpp"
 $14 = (($10) + ($13)|0); //@line 1094 "fastmix.cpp"
 $15 = ((($pChannel)) + 36|0); //@line 1094 "fastmix.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1094 "fastmix.cpp"
 $17 = $16 & 64; //@line 1094 "fastmix.cpp"
 $18 = ($17|0)==(0); //@line 1094 "fastmix.cpp"
 $19 = (($14) + ($12<<1)|0); //@line 1094 "fastmix.cpp"
 $$ = $18 ? $14 : $19; //@line 1094 "fastmix.cpp"
 $20 = ((($pChannel)) + 72|0); //@line 1096 "fastmix.cpp"
 $21 = ((($pChannel)) + 76|0); //@line 1096 "fastmix.cpp"
 $22 = ((($pChannel)) + 80|0); //@line 1096 "fastmix.cpp"
 $23 = ((($pChannel)) + 28|0); //@line 1097 "fastmix.cpp"
 $24 = ((($pChannel)) + 24|0); //@line 1097 "fastmix.cpp"
 $25 = ((($pChannel)) + 12|0); //@line 1098 "fastmix.cpp"
 $fy1$0 = $5;$fy2$0 = $7;$nPos$0 = $9;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $26 = $nPos$0 >> 16; //@line 1095 "fastmix.cpp"
  $27 = $nPos$0 & 65535; //@line 1095 "fastmix.cpp"
  $28 = (($27) + 16)|0; //@line 1095 "fastmix.cpp"
  $29 = $28 >>> 2; //@line 1095 "fastmix.cpp"
  $30 = $29 & 32760; //@line 1095 "fastmix.cpp"
  $31 = (1217698 + ($30<<1)|0); //@line 1095 "fastmix.cpp"
  $32 = HEAP16[$31>>1]|0; //@line 1095 "fastmix.cpp"
  $33 = $32 << 16 >> 16; //@line 1095 "fastmix.cpp"
  $34 = (($26) + -3)|0; //@line 1095 "fastmix.cpp"
  $35 = (($$) + ($34<<1)|0); //@line 1095 "fastmix.cpp"
  $36 = HEAP16[$35>>1]|0; //@line 1095 "fastmix.cpp"
  $37 = $36 << 16 >> 16; //@line 1095 "fastmix.cpp"
  $38 = Math_imul($37, $33)|0; //@line 1095 "fastmix.cpp"
  $39 = $30 | 1; //@line 1095 "fastmix.cpp"
  $40 = (1217698 + ($39<<1)|0); //@line 1095 "fastmix.cpp"
  $41 = HEAP16[$40>>1]|0; //@line 1095 "fastmix.cpp"
  $42 = $41 << 16 >> 16; //@line 1095 "fastmix.cpp"
  $43 = (($26) + -2)|0; //@line 1095 "fastmix.cpp"
  $44 = (($$) + ($43<<1)|0); //@line 1095 "fastmix.cpp"
  $45 = HEAP16[$44>>1]|0; //@line 1095 "fastmix.cpp"
  $46 = $45 << 16 >> 16; //@line 1095 "fastmix.cpp"
  $47 = Math_imul($46, $42)|0; //@line 1095 "fastmix.cpp"
  $48 = (($47) + ($38))|0; //@line 1095 "fastmix.cpp"
  $49 = $30 | 2; //@line 1095 "fastmix.cpp"
  $50 = (1217698 + ($49<<1)|0); //@line 1095 "fastmix.cpp"
  $51 = HEAP16[$50>>1]|0; //@line 1095 "fastmix.cpp"
  $52 = $51 << 16 >> 16; //@line 1095 "fastmix.cpp"
  $53 = (($26) + -1)|0; //@line 1095 "fastmix.cpp"
  $54 = (($$) + ($53<<1)|0); //@line 1095 "fastmix.cpp"
  $55 = HEAP16[$54>>1]|0; //@line 1095 "fastmix.cpp"
  $56 = $55 << 16 >> 16; //@line 1095 "fastmix.cpp"
  $57 = Math_imul($56, $52)|0; //@line 1095 "fastmix.cpp"
  $58 = (($48) + ($57))|0; //@line 1095 "fastmix.cpp"
  $59 = $30 | 3; //@line 1095 "fastmix.cpp"
  $60 = (1217698 + ($59<<1)|0); //@line 1095 "fastmix.cpp"
  $61 = HEAP16[$60>>1]|0; //@line 1095 "fastmix.cpp"
  $62 = $61 << 16 >> 16; //@line 1095 "fastmix.cpp"
  $63 = (($$) + ($26<<1)|0); //@line 1095 "fastmix.cpp"
  $64 = HEAP16[$63>>1]|0; //@line 1095 "fastmix.cpp"
  $65 = $64 << 16 >> 16; //@line 1095 "fastmix.cpp"
  $66 = Math_imul($65, $62)|0; //@line 1095 "fastmix.cpp"
  $67 = (($58) + ($66))|0; //@line 1095 "fastmix.cpp"
  $68 = $30 | 4; //@line 1095 "fastmix.cpp"
  $69 = (1217698 + ($68<<1)|0); //@line 1095 "fastmix.cpp"
  $70 = HEAP16[$69>>1]|0; //@line 1095 "fastmix.cpp"
  $71 = $70 << 16 >> 16; //@line 1095 "fastmix.cpp"
  $72 = (($26) + 1)|0; //@line 1095 "fastmix.cpp"
  $73 = (($$) + ($72<<1)|0); //@line 1095 "fastmix.cpp"
  $74 = HEAP16[$73>>1]|0; //@line 1095 "fastmix.cpp"
  $75 = $74 << 16 >> 16; //@line 1095 "fastmix.cpp"
  $76 = Math_imul($75, $71)|0; //@line 1095 "fastmix.cpp"
  $77 = $30 | 5; //@line 1095 "fastmix.cpp"
  $78 = (1217698 + ($77<<1)|0); //@line 1095 "fastmix.cpp"
  $79 = HEAP16[$78>>1]|0; //@line 1095 "fastmix.cpp"
  $80 = $79 << 16 >> 16; //@line 1095 "fastmix.cpp"
  $81 = (($26) + 2)|0; //@line 1095 "fastmix.cpp"
  $82 = (($$) + ($81<<1)|0); //@line 1095 "fastmix.cpp"
  $83 = HEAP16[$82>>1]|0; //@line 1095 "fastmix.cpp"
  $84 = $83 << 16 >> 16; //@line 1095 "fastmix.cpp"
  $85 = Math_imul($84, $80)|0; //@line 1095 "fastmix.cpp"
  $86 = (($85) + ($76))|0; //@line 1095 "fastmix.cpp"
  $87 = $30 | 6; //@line 1095 "fastmix.cpp"
  $88 = (1217698 + ($87<<1)|0); //@line 1095 "fastmix.cpp"
  $89 = HEAP16[$88>>1]|0; //@line 1095 "fastmix.cpp"
  $90 = $89 << 16 >> 16; //@line 1095 "fastmix.cpp"
  $91 = (($26) + 3)|0; //@line 1095 "fastmix.cpp"
  $92 = (($$) + ($91<<1)|0); //@line 1095 "fastmix.cpp"
  $93 = HEAP16[$92>>1]|0; //@line 1095 "fastmix.cpp"
  $94 = $93 << 16 >> 16; //@line 1095 "fastmix.cpp"
  $95 = Math_imul($94, $90)|0; //@line 1095 "fastmix.cpp"
  $96 = (($86) + ($95))|0; //@line 1095 "fastmix.cpp"
  $97 = $29 | 7; //@line 1095 "fastmix.cpp"
  $98 = (1217698 + ($97<<1)|0); //@line 1095 "fastmix.cpp"
  $99 = HEAP16[$98>>1]|0; //@line 1095 "fastmix.cpp"
  $100 = $99 << 16 >> 16; //@line 1095 "fastmix.cpp"
  $101 = (($26) + 4)|0; //@line 1095 "fastmix.cpp"
  $102 = (($$) + ($101<<1)|0); //@line 1095 "fastmix.cpp"
  $103 = HEAP16[$102>>1]|0; //@line 1095 "fastmix.cpp"
  $104 = $103 << 16 >> 16; //@line 1095 "fastmix.cpp"
  $105 = Math_imul($104, $100)|0; //@line 1095 "fastmix.cpp"
  $106 = (($96) + ($105))|0; //@line 1095 "fastmix.cpp"
  $107 = $67 >> 1; //@line 1095 "fastmix.cpp"
  $108 = $106 >> 1; //@line 1095 "fastmix.cpp"
  $109 = (($108) + ($107))|0; //@line 1095 "fastmix.cpp"
  $110 = $109 >> 14; //@line 1095 "fastmix.cpp"
  $111 = HEAP32[$20>>2]|0; //@line 1096 "fastmix.cpp"
  $112 = Math_imul($110, $111)|0; //@line 1096 "fastmix.cpp"
  $113 = HEAP32[$21>>2]|0; //@line 1096 "fastmix.cpp"
  $114 = Math_imul($113, $fy1$0)|0; //@line 1096 "fastmix.cpp"
  $115 = HEAP32[$22>>2]|0; //@line 1096 "fastmix.cpp"
  $116 = Math_imul($115, $fy2$0)|0; //@line 1096 "fastmix.cpp"
  $117 = (($114) + 4096)|0; //@line 1096 "fastmix.cpp"
  $118 = (($117) + ($116))|0; //@line 1096 "fastmix.cpp"
  $119 = (($118) + ($112))|0; //@line 1096 "fastmix.cpp"
  $120 = $119 >> 13; //@line 1096 "fastmix.cpp"
  $121 = HEAP32[$23>>2]|0; //@line 1097 "fastmix.cpp"
  $122 = (($121) + ($nRampLeftVol$0))|0; //@line 1097 "fastmix.cpp"
  $123 = HEAP32[$24>>2]|0; //@line 1097 "fastmix.cpp"
  $124 = (($123) + ($nRampRightVol$0))|0; //@line 1097 "fastmix.cpp"
  $125 = $124 >> 12; //@line 1097 "fastmix.cpp"
  $126 = Math_imul($120, $125)|0; //@line 1097 "fastmix.cpp"
  $127 = HEAP32[$pvol$0>>2]|0; //@line 1097 "fastmix.cpp"
  $128 = (($126) + ($127))|0; //@line 1097 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $128; //@line 1097 "fastmix.cpp"
  $129 = $122 >> 12; //@line 1097 "fastmix.cpp"
  $130 = Math_imul($120, $129)|0; //@line 1097 "fastmix.cpp"
  $131 = ((($pvol$0)) + 4|0); //@line 1097 "fastmix.cpp"
  $132 = HEAP32[$131>>2]|0; //@line 1097 "fastmix.cpp"
  $133 = (($130) + ($132))|0; //@line 1097 "fastmix.cpp"
  HEAP32[$131>>2] = $133; //@line 1097 "fastmix.cpp"
  $134 = ((($pvol$0)) + 8|0); //@line 1097 "fastmix.cpp"
  $135 = HEAP32[$25>>2]|0; //@line 1098 "fastmix.cpp"
  $136 = (($135) + ($nPos$0))|0; //@line 1098 "fastmix.cpp"
  $137 = ($134>>>0)<($pbufmax>>>0); //@line 1098 "fastmix.cpp"
  if ($137) {
   $fy2$0$phi = $fy1$0;$fy1$0 = $120;$nPos$0 = $136;$nRampLeftVol$0 = $122;$nRampRightVol$0 = $124;$pvol$0 = $134;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $120;$$lcssa26 = $122;$$lcssa27 = $124;$$lcssa28 = $125;$$lcssa29 = $129;$$lcssa30 = $136;$fy1$0$lcssa = $fy1$0;
   break;
  }
 }
 $138 = $$lcssa30 >> 16; //@line 1098 "fastmix.cpp"
 $139 = HEAP32[$11>>2]|0; //@line 1098 "fastmix.cpp"
 $140 = (($139) + ($138))|0; //@line 1098 "fastmix.cpp"
 HEAP32[$11>>2] = $140; //@line 1098 "fastmix.cpp"
 $141 = $$lcssa30 & 65535; //@line 1098 "fastmix.cpp"
 HEAP32[$8>>2] = $141; //@line 1098 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa; //@line 1098 "fastmix.cpp"
 HEAP32[$6>>2] = $fy1$0$lcssa; //@line 1098 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa27; //@line 1098 "fastmix.cpp"
 $142 = ((($pChannel)) + 16|0); //@line 1098 "fastmix.cpp"
 HEAP32[$142>>2] = $$lcssa28; //@line 1098 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa26; //@line 1098 "fastmix.cpp"
 $143 = ((($pChannel)) + 20|0); //@line 1098 "fastmix.cpp"
 HEAP32[$143>>2] = $$lcssa29; //@line 1098 "fastmix.cpp"
 return; //@line 1098 "fastmix.cpp"
}
function __Z32FilterStereo8BitFirFilterRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa62 = 0, $$lcssa63 = 0, $$lcssa64 = 0, $$lcssa65 = 0, $$lcssa66 = 0, $$lcssa67 = 0, $$sum = 0, $$sum1 = 0, $$sum10 = 0, $$sum11 = 0, $$sum12 = 0, $$sum13 = 0, $$sum14 = 0, $$sum15 = 0, $$sum2 = 0, $$sum3 = 0, $$sum4 = 0, $$sum5 = 0;
 var $$sum6 = 0, $$sum7 = 0, $$sum8 = 0, $$sum9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $fy3$0 = 0, $fy3$0$lcssa = 0, $fy4$0 = 0, $fy4$0$phi = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 1201 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1201 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 1201 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1201 "fastmix.cpp"
 $4 = ((($pChannel)) + 56|0); //@line 1201 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1201 "fastmix.cpp"
 $6 = ((($pChannel)) + 60|0); //@line 1201 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1201 "fastmix.cpp"
 $8 = ((($pChannel)) + 64|0); //@line 1201 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1201 "fastmix.cpp"
 $10 = ((($pChannel)) + 68|0); //@line 1201 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 1201 "fastmix.cpp"
 $12 = ((($pChannel)) + 8|0); //@line 1202 "fastmix.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 1202 "fastmix.cpp"
 $14 = HEAP32[$pChannel>>2]|0; //@line 1202 "fastmix.cpp"
 $15 = ((($pChannel)) + 4|0); //@line 1202 "fastmix.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1202 "fastmix.cpp"
 $17 = (($14) + ($16)|0); //@line 1202 "fastmix.cpp"
 $18 = ((($pChannel)) + 36|0); //@line 1202 "fastmix.cpp"
 $19 = HEAP32[$18>>2]|0; //@line 1202 "fastmix.cpp"
 $20 = $19 & 64; //@line 1202 "fastmix.cpp"
 $21 = ($20|0)==(0); //@line 1202 "fastmix.cpp"
 $$ = $21 ? $14 : $17; //@line 1202 "fastmix.cpp"
 $22 = (($16) + -6)|0; //@line 1203 "fastmix.cpp"
 $23 = (($16) + -4)|0; //@line 1203 "fastmix.cpp"
 $24 = (($16) + -2)|0; //@line 1203 "fastmix.cpp"
 $25 = (($16) + 2)|0; //@line 1203 "fastmix.cpp"
 $26 = (($16) + 4)|0; //@line 1203 "fastmix.cpp"
 $27 = (($16) + 6)|0; //@line 1203 "fastmix.cpp"
 $28 = (($16) + 8)|0; //@line 1203 "fastmix.cpp"
 $29 = (($16) + -5)|0; //@line 1203 "fastmix.cpp"
 $30 = (($16) + -3)|0; //@line 1203 "fastmix.cpp"
 $31 = (($16) + -1)|0; //@line 1203 "fastmix.cpp"
 $32 = (($16) + 3)|0; //@line 1203 "fastmix.cpp"
 $33 = (($16) + 5)|0; //@line 1203 "fastmix.cpp"
 $34 = (($16) + 7)|0; //@line 1203 "fastmix.cpp"
 $35 = (($16) + 9)|0; //@line 1203 "fastmix.cpp"
 $36 = ((($pChannel)) + 72|0); //@line 1204 "fastmix.cpp"
 $37 = ((($pChannel)) + 76|0); //@line 1204 "fastmix.cpp"
 $38 = ((($pChannel)) + 80|0); //@line 1204 "fastmix.cpp"
 $39 = ((($pChannel)) + 28|0); //@line 1205 "fastmix.cpp"
 $40 = ((($pChannel)) + 24|0); //@line 1205 "fastmix.cpp"
 $41 = ((($pChannel)) + 12|0); //@line 1206 "fastmix.cpp"
 $fy1$0 = $5;$fy2$0 = $7;$fy3$0 = $9;$fy4$0 = $11;$nPos$0 = $13;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $42 = $nPos$0 >> 16; //@line 1203 "fastmix.cpp"
  $43 = $nPos$0 & 65535; //@line 1203 "fastmix.cpp"
  $44 = (($43) + 16)|0; //@line 1203 "fastmix.cpp"
  $45 = $44 >>> 2; //@line 1203 "fastmix.cpp"
  $46 = $45 & 32760; //@line 1203 "fastmix.cpp"
  $47 = (1217698 + ($46<<1)|0); //@line 1203 "fastmix.cpp"
  $48 = HEAP16[$47>>1]|0; //@line 1203 "fastmix.cpp"
  $49 = $48 << 16 >> 16; //@line 1203 "fastmix.cpp"
  $50 = $42 << 1; //@line 1203 "fastmix.cpp"
  $$sum = (($22) + ($50))|0; //@line 1203 "fastmix.cpp"
  $51 = (($$) + ($$sum)|0); //@line 1203 "fastmix.cpp"
  $52 = HEAP8[$51>>0]|0; //@line 1203 "fastmix.cpp"
  $53 = $52 << 24 >> 24; //@line 1203 "fastmix.cpp"
  $54 = Math_imul($53, $49)|0; //@line 1203 "fastmix.cpp"
  $55 = $46 | 1; //@line 1203 "fastmix.cpp"
  $56 = (1217698 + ($55<<1)|0); //@line 1203 "fastmix.cpp"
  $57 = HEAP16[$56>>1]|0; //@line 1203 "fastmix.cpp"
  $58 = $57 << 16 >> 16; //@line 1203 "fastmix.cpp"
  $$sum1 = (($23) + ($50))|0; //@line 1203 "fastmix.cpp"
  $59 = (($$) + ($$sum1)|0); //@line 1203 "fastmix.cpp"
  $60 = HEAP8[$59>>0]|0; //@line 1203 "fastmix.cpp"
  $61 = $60 << 24 >> 24; //@line 1203 "fastmix.cpp"
  $62 = Math_imul($61, $58)|0; //@line 1203 "fastmix.cpp"
  $63 = (($62) + ($54))|0; //@line 1203 "fastmix.cpp"
  $64 = $46 | 2; //@line 1203 "fastmix.cpp"
  $65 = (1217698 + ($64<<1)|0); //@line 1203 "fastmix.cpp"
  $66 = HEAP16[$65>>1]|0; //@line 1203 "fastmix.cpp"
  $67 = $66 << 16 >> 16; //@line 1203 "fastmix.cpp"
  $$sum2 = (($24) + ($50))|0; //@line 1203 "fastmix.cpp"
  $68 = (($$) + ($$sum2)|0); //@line 1203 "fastmix.cpp"
  $69 = HEAP8[$68>>0]|0; //@line 1203 "fastmix.cpp"
  $70 = $69 << 24 >> 24; //@line 1203 "fastmix.cpp"
  $71 = Math_imul($70, $67)|0; //@line 1203 "fastmix.cpp"
  $72 = (($63) + ($71))|0; //@line 1203 "fastmix.cpp"
  $73 = $46 | 3; //@line 1203 "fastmix.cpp"
  $74 = (1217698 + ($73<<1)|0); //@line 1203 "fastmix.cpp"
  $75 = HEAP16[$74>>1]|0; //@line 1203 "fastmix.cpp"
  $76 = $75 << 16 >> 16; //@line 1203 "fastmix.cpp"
  $$sum3 = (($50) + ($16))|0; //@line 1203 "fastmix.cpp"
  $77 = (($$) + ($$sum3)|0); //@line 1203 "fastmix.cpp"
  $78 = HEAP8[$77>>0]|0; //@line 1203 "fastmix.cpp"
  $79 = $78 << 24 >> 24; //@line 1203 "fastmix.cpp"
  $80 = Math_imul($79, $76)|0; //@line 1203 "fastmix.cpp"
  $81 = (($72) + ($80))|0; //@line 1203 "fastmix.cpp"
  $82 = $46 | 4; //@line 1203 "fastmix.cpp"
  $83 = (1217698 + ($82<<1)|0); //@line 1203 "fastmix.cpp"
  $84 = HEAP16[$83>>1]|0; //@line 1203 "fastmix.cpp"
  $85 = $84 << 16 >> 16; //@line 1203 "fastmix.cpp"
  $$sum4 = (($25) + ($50))|0; //@line 1203 "fastmix.cpp"
  $86 = (($$) + ($$sum4)|0); //@line 1203 "fastmix.cpp"
  $87 = HEAP8[$86>>0]|0; //@line 1203 "fastmix.cpp"
  $88 = $87 << 24 >> 24; //@line 1203 "fastmix.cpp"
  $89 = Math_imul($88, $85)|0; //@line 1203 "fastmix.cpp"
  $90 = (($81) + ($89))|0; //@line 1203 "fastmix.cpp"
  $91 = $46 | 5; //@line 1203 "fastmix.cpp"
  $92 = (1217698 + ($91<<1)|0); //@line 1203 "fastmix.cpp"
  $93 = HEAP16[$92>>1]|0; //@line 1203 "fastmix.cpp"
  $94 = $93 << 16 >> 16; //@line 1203 "fastmix.cpp"
  $$sum5 = (($26) + ($50))|0; //@line 1203 "fastmix.cpp"
  $95 = (($$) + ($$sum5)|0); //@line 1203 "fastmix.cpp"
  $96 = HEAP8[$95>>0]|0; //@line 1203 "fastmix.cpp"
  $97 = $96 << 24 >> 24; //@line 1203 "fastmix.cpp"
  $98 = Math_imul($97, $94)|0; //@line 1203 "fastmix.cpp"
  $99 = (($90) + ($98))|0; //@line 1203 "fastmix.cpp"
  $100 = $46 | 6; //@line 1203 "fastmix.cpp"
  $101 = (1217698 + ($100<<1)|0); //@line 1203 "fastmix.cpp"
  $102 = HEAP16[$101>>1]|0; //@line 1203 "fastmix.cpp"
  $103 = $102 << 16 >> 16; //@line 1203 "fastmix.cpp"
  $$sum6 = (($27) + ($50))|0; //@line 1203 "fastmix.cpp"
  $104 = (($$) + ($$sum6)|0); //@line 1203 "fastmix.cpp"
  $105 = HEAP8[$104>>0]|0; //@line 1203 "fastmix.cpp"
  $106 = $105 << 24 >> 24; //@line 1203 "fastmix.cpp"
  $107 = Math_imul($106, $103)|0; //@line 1203 "fastmix.cpp"
  $108 = (($99) + ($107))|0; //@line 1203 "fastmix.cpp"
  $109 = $45 | 7; //@line 1203 "fastmix.cpp"
  $110 = (1217698 + ($109<<1)|0); //@line 1203 "fastmix.cpp"
  $111 = HEAP16[$110>>1]|0; //@line 1203 "fastmix.cpp"
  $112 = $111 << 16 >> 16; //@line 1203 "fastmix.cpp"
  $$sum7 = (($28) + ($50))|0; //@line 1203 "fastmix.cpp"
  $113 = (($$) + ($$sum7)|0); //@line 1203 "fastmix.cpp"
  $114 = HEAP8[$113>>0]|0; //@line 1203 "fastmix.cpp"
  $115 = $114 << 24 >> 24; //@line 1203 "fastmix.cpp"
  $116 = Math_imul($115, $112)|0; //@line 1203 "fastmix.cpp"
  $117 = (($108) + ($116))|0; //@line 1203 "fastmix.cpp"
  $118 = $117 >> 7; //@line 1203 "fastmix.cpp"
  $$sum8 = (($29) + ($50))|0; //@line 1203 "fastmix.cpp"
  $119 = (($$) + ($$sum8)|0); //@line 1203 "fastmix.cpp"
  $120 = HEAP8[$119>>0]|0; //@line 1203 "fastmix.cpp"
  $121 = $120 << 24 >> 24; //@line 1203 "fastmix.cpp"
  $122 = Math_imul($121, $49)|0; //@line 1203 "fastmix.cpp"
  $$sum9 = (($30) + ($50))|0; //@line 1203 "fastmix.cpp"
  $123 = (($$) + ($$sum9)|0); //@line 1203 "fastmix.cpp"
  $124 = HEAP8[$123>>0]|0; //@line 1203 "fastmix.cpp"
  $125 = $124 << 24 >> 24; //@line 1203 "fastmix.cpp"
  $126 = Math_imul($125, $58)|0; //@line 1203 "fastmix.cpp"
  $127 = (($126) + ($122))|0; //@line 1203 "fastmix.cpp"
  $$sum10 = (($31) + ($50))|0; //@line 1203 "fastmix.cpp"
  $128 = (($$) + ($$sum10)|0); //@line 1203 "fastmix.cpp"
  $129 = HEAP8[$128>>0]|0; //@line 1203 "fastmix.cpp"
  $130 = $129 << 24 >> 24; //@line 1203 "fastmix.cpp"
  $131 = Math_imul($130, $67)|0; //@line 1203 "fastmix.cpp"
  $132 = (($127) + ($131))|0; //@line 1203 "fastmix.cpp"
  $133 = $50 | 1; //@line 1203 "fastmix.cpp"
  $$sum11 = (($133) + ($16))|0; //@line 1203 "fastmix.cpp"
  $134 = (($$) + ($$sum11)|0); //@line 1203 "fastmix.cpp"
  $135 = HEAP8[$134>>0]|0; //@line 1203 "fastmix.cpp"
  $136 = $135 << 24 >> 24; //@line 1203 "fastmix.cpp"
  $137 = Math_imul($136, $76)|0; //@line 1203 "fastmix.cpp"
  $138 = (($132) + ($137))|0; //@line 1203 "fastmix.cpp"
  $$sum12 = (($32) + ($50))|0; //@line 1203 "fastmix.cpp"
  $139 = (($$) + ($$sum12)|0); //@line 1203 "fastmix.cpp"
  $140 = HEAP8[$139>>0]|0; //@line 1203 "fastmix.cpp"
  $141 = $140 << 24 >> 24; //@line 1203 "fastmix.cpp"
  $142 = Math_imul($141, $85)|0; //@line 1203 "fastmix.cpp"
  $143 = (($138) + ($142))|0; //@line 1203 "fastmix.cpp"
  $$sum13 = (($33) + ($50))|0; //@line 1203 "fastmix.cpp"
  $144 = (($$) + ($$sum13)|0); //@line 1203 "fastmix.cpp"
  $145 = HEAP8[$144>>0]|0; //@line 1203 "fastmix.cpp"
  $146 = $145 << 24 >> 24; //@line 1203 "fastmix.cpp"
  $147 = Math_imul($146, $94)|0; //@line 1203 "fastmix.cpp"
  $148 = (($143) + ($147))|0; //@line 1203 "fastmix.cpp"
  $$sum14 = (($34) + ($50))|0; //@line 1203 "fastmix.cpp"
  $149 = (($$) + ($$sum14)|0); //@line 1203 "fastmix.cpp"
  $150 = HEAP8[$149>>0]|0; //@line 1203 "fastmix.cpp"
  $151 = $150 << 24 >> 24; //@line 1203 "fastmix.cpp"
  $152 = Math_imul($151, $103)|0; //@line 1203 "fastmix.cpp"
  $153 = (($148) + ($152))|0; //@line 1203 "fastmix.cpp"
  $$sum15 = (($35) + ($50))|0; //@line 1203 "fastmix.cpp"
  $154 = (($$) + ($$sum15)|0); //@line 1203 "fastmix.cpp"
  $155 = HEAP8[$154>>0]|0; //@line 1203 "fastmix.cpp"
  $156 = $155 << 24 >> 24; //@line 1203 "fastmix.cpp"
  $157 = Math_imul($156, $112)|0; //@line 1203 "fastmix.cpp"
  $158 = (($153) + ($157))|0; //@line 1203 "fastmix.cpp"
  $159 = $158 >> 7; //@line 1203 "fastmix.cpp"
  $160 = HEAP32[$36>>2]|0; //@line 1204 "fastmix.cpp"
  $161 = Math_imul($160, $118)|0; //@line 1204 "fastmix.cpp"
  $162 = HEAP32[$37>>2]|0; //@line 1204 "fastmix.cpp"
  $163 = Math_imul($162, $fy1$0)|0; //@line 1204 "fastmix.cpp"
  $164 = HEAP32[$38>>2]|0; //@line 1204 "fastmix.cpp"
  $165 = Math_imul($164, $fy2$0)|0; //@line 1204 "fastmix.cpp"
  $166 = (($161) + 4096)|0; //@line 1204 "fastmix.cpp"
  $167 = (($166) + ($163))|0; //@line 1204 "fastmix.cpp"
  $168 = (($167) + ($165))|0; //@line 1204 "fastmix.cpp"
  $169 = $168 >> 13; //@line 1204 "fastmix.cpp"
  $170 = Math_imul($159, $160)|0; //@line 1204 "fastmix.cpp"
  $171 = Math_imul($162, $fy3$0)|0; //@line 1204 "fastmix.cpp"
  $172 = Math_imul($164, $fy4$0)|0; //@line 1204 "fastmix.cpp"
  $173 = (($171) + 4096)|0; //@line 1204 "fastmix.cpp"
  $174 = (($173) + ($172))|0; //@line 1204 "fastmix.cpp"
  $175 = (($174) + ($170))|0; //@line 1204 "fastmix.cpp"
  $176 = $175 >> 13; //@line 1204 "fastmix.cpp"
  $177 = HEAP32[$39>>2]|0; //@line 1205 "fastmix.cpp"
  $178 = (($177) + ($nRampLeftVol$0))|0; //@line 1205 "fastmix.cpp"
  $179 = HEAP32[$40>>2]|0; //@line 1205 "fastmix.cpp"
  $180 = (($179) + ($nRampRightVol$0))|0; //@line 1205 "fastmix.cpp"
  $181 = $180 >> 12; //@line 1205 "fastmix.cpp"
  $182 = Math_imul($169, $181)|0; //@line 1205 "fastmix.cpp"
  $183 = HEAP32[$pvol$0>>2]|0; //@line 1205 "fastmix.cpp"
  $184 = (($182) + ($183))|0; //@line 1205 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $184; //@line 1205 "fastmix.cpp"
  $185 = $178 >> 12; //@line 1205 "fastmix.cpp"
  $186 = Math_imul($176, $185)|0; //@line 1205 "fastmix.cpp"
  $187 = ((($pvol$0)) + 4|0); //@line 1205 "fastmix.cpp"
  $188 = HEAP32[$187>>2]|0; //@line 1205 "fastmix.cpp"
  $189 = (($186) + ($188))|0; //@line 1205 "fastmix.cpp"
  HEAP32[$187>>2] = $189; //@line 1205 "fastmix.cpp"
  $190 = ((($pvol$0)) + 8|0); //@line 1205 "fastmix.cpp"
  $191 = HEAP32[$41>>2]|0; //@line 1206 "fastmix.cpp"
  $192 = (($191) + ($nPos$0))|0; //@line 1206 "fastmix.cpp"
  $193 = ($190>>>0)<($pbufmax>>>0); //@line 1206 "fastmix.cpp"
  if ($193) {
   $fy4$0$phi = $fy3$0;$fy2$0$phi = $fy1$0;$fy1$0 = $169;$fy3$0 = $176;$nPos$0 = $192;$nRampLeftVol$0 = $178;$nRampRightVol$0 = $180;$pvol$0 = $190;$fy4$0 = $fy4$0$phi;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $169;$$lcssa62 = $176;$$lcssa63 = $178;$$lcssa64 = $180;$$lcssa65 = $181;$$lcssa66 = $185;$$lcssa67 = $192;$fy1$0$lcssa = $fy1$0;$fy3$0$lcssa = $fy3$0;
   break;
  }
 }
 $194 = $$lcssa67 >> 16; //@line 1206 "fastmix.cpp"
 $195 = HEAP32[$15>>2]|0; //@line 1206 "fastmix.cpp"
 $196 = (($195) + ($194))|0; //@line 1206 "fastmix.cpp"
 HEAP32[$15>>2] = $196; //@line 1206 "fastmix.cpp"
 $197 = $$lcssa67 & 65535; //@line 1206 "fastmix.cpp"
 HEAP32[$12>>2] = $197; //@line 1206 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa; //@line 1206 "fastmix.cpp"
 HEAP32[$6>>2] = $fy1$0$lcssa; //@line 1206 "fastmix.cpp"
 HEAP32[$8>>2] = $$lcssa62; //@line 1206 "fastmix.cpp"
 HEAP32[$10>>2] = $fy3$0$lcssa; //@line 1206 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa64; //@line 1206 "fastmix.cpp"
 $198 = ((($pChannel)) + 16|0); //@line 1206 "fastmix.cpp"
 HEAP32[$198>>2] = $$lcssa65; //@line 1206 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa63; //@line 1206 "fastmix.cpp"
 $199 = ((($pChannel)) + 20|0); //@line 1206 "fastmix.cpp"
 HEAP32[$199>>2] = $$lcssa66; //@line 1206 "fastmix.cpp"
 return; //@line 1206 "fastmix.cpp"
}
function __Z33FilterStereo16BitFirFilterRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$lcssa33 = 0, $$lcssa34 = 0, $$lcssa35 = 0, $$lcssa36 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0;
 var $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fy1$0 = 0, $fy1$0$lcssa = 0, $fy2$0 = 0, $fy2$0$phi = 0, $fy3$0 = 0, $fy3$0$lcssa = 0;
 var $fy4$0 = 0, $fy4$0$phi = 0, $nPos$0 = 0, $nRampLeftVol$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 1208 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1208 "fastmix.cpp"
 $2 = ((($pChannel)) + 52|0); //@line 1208 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1208 "fastmix.cpp"
 $4 = ((($pChannel)) + 56|0); //@line 1208 "fastmix.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 1208 "fastmix.cpp"
 $6 = ((($pChannel)) + 60|0); //@line 1208 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 1208 "fastmix.cpp"
 $8 = ((($pChannel)) + 64|0); //@line 1208 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 1208 "fastmix.cpp"
 $10 = ((($pChannel)) + 68|0); //@line 1208 "fastmix.cpp"
 $11 = HEAP32[$10>>2]|0; //@line 1208 "fastmix.cpp"
 $12 = ((($pChannel)) + 8|0); //@line 1209 "fastmix.cpp"
 $13 = HEAP32[$12>>2]|0; //@line 1209 "fastmix.cpp"
 $14 = HEAP32[$pChannel>>2]|0; //@line 1209 "fastmix.cpp"
 $15 = ((($pChannel)) + 4|0); //@line 1209 "fastmix.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 1209 "fastmix.cpp"
 $17 = $16 << 1; //@line 1209 "fastmix.cpp"
 $18 = (($14) + ($17)|0); //@line 1209 "fastmix.cpp"
 $19 = ((($pChannel)) + 36|0); //@line 1209 "fastmix.cpp"
 $20 = HEAP32[$19>>2]|0; //@line 1209 "fastmix.cpp"
 $21 = $20 & 64; //@line 1209 "fastmix.cpp"
 $22 = ($21|0)==(0); //@line 1209 "fastmix.cpp"
 $23 = (($18) + ($16<<1)|0); //@line 1209 "fastmix.cpp"
 $$ = $22 ? $18 : $23; //@line 1209 "fastmix.cpp"
 $24 = ((($pChannel)) + 72|0); //@line 1211 "fastmix.cpp"
 $25 = ((($pChannel)) + 76|0); //@line 1211 "fastmix.cpp"
 $26 = ((($pChannel)) + 80|0); //@line 1211 "fastmix.cpp"
 $27 = ((($pChannel)) + 28|0); //@line 1212 "fastmix.cpp"
 $28 = ((($pChannel)) + 24|0); //@line 1212 "fastmix.cpp"
 $29 = ((($pChannel)) + 12|0); //@line 1213 "fastmix.cpp"
 $fy1$0 = $5;$fy2$0 = $7;$fy3$0 = $9;$fy4$0 = $11;$nPos$0 = $13;$nRampLeftVol$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $30 = $nPos$0 >> 16; //@line 1210 "fastmix.cpp"
  $31 = $nPos$0 & 65535; //@line 1210 "fastmix.cpp"
  $32 = (($31) + 16)|0; //@line 1210 "fastmix.cpp"
  $33 = $32 >>> 2; //@line 1210 "fastmix.cpp"
  $34 = $33 & 32760; //@line 1210 "fastmix.cpp"
  $35 = (1217698 + ($34<<1)|0); //@line 1210 "fastmix.cpp"
  $36 = HEAP16[$35>>1]|0; //@line 1210 "fastmix.cpp"
  $37 = $36 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $38 = $30 << 1; //@line 1210 "fastmix.cpp"
  $39 = (($38) + -6)|0; //@line 1210 "fastmix.cpp"
  $40 = (($$) + ($39<<1)|0); //@line 1210 "fastmix.cpp"
  $41 = HEAP16[$40>>1]|0; //@line 1210 "fastmix.cpp"
  $42 = $41 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $43 = Math_imul($42, $37)|0; //@line 1210 "fastmix.cpp"
  $44 = $34 | 1; //@line 1210 "fastmix.cpp"
  $45 = (1217698 + ($44<<1)|0); //@line 1210 "fastmix.cpp"
  $46 = HEAP16[$45>>1]|0; //@line 1210 "fastmix.cpp"
  $47 = $46 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $48 = (($38) + -4)|0; //@line 1210 "fastmix.cpp"
  $49 = (($$) + ($48<<1)|0); //@line 1210 "fastmix.cpp"
  $50 = HEAP16[$49>>1]|0; //@line 1210 "fastmix.cpp"
  $51 = $50 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $52 = Math_imul($51, $47)|0; //@line 1210 "fastmix.cpp"
  $53 = (($52) + ($43))|0; //@line 1210 "fastmix.cpp"
  $54 = $34 | 2; //@line 1210 "fastmix.cpp"
  $55 = (1217698 + ($54<<1)|0); //@line 1210 "fastmix.cpp"
  $56 = HEAP16[$55>>1]|0; //@line 1210 "fastmix.cpp"
  $57 = $56 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $58 = (($38) + -2)|0; //@line 1210 "fastmix.cpp"
  $59 = (($$) + ($58<<1)|0); //@line 1210 "fastmix.cpp"
  $60 = HEAP16[$59>>1]|0; //@line 1210 "fastmix.cpp"
  $61 = $60 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $62 = Math_imul($61, $57)|0; //@line 1210 "fastmix.cpp"
  $63 = (($53) + ($62))|0; //@line 1210 "fastmix.cpp"
  $64 = $34 | 3; //@line 1210 "fastmix.cpp"
  $65 = (1217698 + ($64<<1)|0); //@line 1210 "fastmix.cpp"
  $66 = HEAP16[$65>>1]|0; //@line 1210 "fastmix.cpp"
  $67 = $66 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $68 = (($$) + ($38<<1)|0); //@line 1210 "fastmix.cpp"
  $69 = HEAP16[$68>>1]|0; //@line 1210 "fastmix.cpp"
  $70 = $69 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $71 = Math_imul($70, $67)|0; //@line 1210 "fastmix.cpp"
  $72 = (($63) + ($71))|0; //@line 1210 "fastmix.cpp"
  $73 = $34 | 4; //@line 1210 "fastmix.cpp"
  $74 = (1217698 + ($73<<1)|0); //@line 1210 "fastmix.cpp"
  $75 = HEAP16[$74>>1]|0; //@line 1210 "fastmix.cpp"
  $76 = $75 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $77 = (($38) + 2)|0; //@line 1210 "fastmix.cpp"
  $78 = (($$) + ($77<<1)|0); //@line 1210 "fastmix.cpp"
  $79 = HEAP16[$78>>1]|0; //@line 1210 "fastmix.cpp"
  $80 = $79 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $81 = Math_imul($80, $76)|0; //@line 1210 "fastmix.cpp"
  $82 = $34 | 5; //@line 1210 "fastmix.cpp"
  $83 = (1217698 + ($82<<1)|0); //@line 1210 "fastmix.cpp"
  $84 = HEAP16[$83>>1]|0; //@line 1210 "fastmix.cpp"
  $85 = $84 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $86 = (($38) + 4)|0; //@line 1210 "fastmix.cpp"
  $87 = (($$) + ($86<<1)|0); //@line 1210 "fastmix.cpp"
  $88 = HEAP16[$87>>1]|0; //@line 1210 "fastmix.cpp"
  $89 = $88 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $90 = Math_imul($89, $85)|0; //@line 1210 "fastmix.cpp"
  $91 = (($90) + ($81))|0; //@line 1210 "fastmix.cpp"
  $92 = $34 | 6; //@line 1210 "fastmix.cpp"
  $93 = (1217698 + ($92<<1)|0); //@line 1210 "fastmix.cpp"
  $94 = HEAP16[$93>>1]|0; //@line 1210 "fastmix.cpp"
  $95 = $94 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $96 = (($38) + 6)|0; //@line 1210 "fastmix.cpp"
  $97 = (($$) + ($96<<1)|0); //@line 1210 "fastmix.cpp"
  $98 = HEAP16[$97>>1]|0; //@line 1210 "fastmix.cpp"
  $99 = $98 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $100 = Math_imul($99, $95)|0; //@line 1210 "fastmix.cpp"
  $101 = (($91) + ($100))|0; //@line 1210 "fastmix.cpp"
  $102 = $33 | 7; //@line 1210 "fastmix.cpp"
  $103 = (1217698 + ($102<<1)|0); //@line 1210 "fastmix.cpp"
  $104 = HEAP16[$103>>1]|0; //@line 1210 "fastmix.cpp"
  $105 = $104 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $106 = (($38) + 8)|0; //@line 1210 "fastmix.cpp"
  $107 = (($$) + ($106<<1)|0); //@line 1210 "fastmix.cpp"
  $108 = HEAP16[$107>>1]|0; //@line 1210 "fastmix.cpp"
  $109 = $108 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $110 = Math_imul($109, $105)|0; //@line 1210 "fastmix.cpp"
  $111 = (($101) + ($110))|0; //@line 1210 "fastmix.cpp"
  $112 = $72 >> 1; //@line 1210 "fastmix.cpp"
  $113 = $111 >> 1; //@line 1210 "fastmix.cpp"
  $114 = (($113) + ($112))|0; //@line 1210 "fastmix.cpp"
  $115 = $114 >> 14; //@line 1210 "fastmix.cpp"
  $116 = (($38) + -5)|0; //@line 1210 "fastmix.cpp"
  $117 = (($$) + ($116<<1)|0); //@line 1210 "fastmix.cpp"
  $118 = HEAP16[$117>>1]|0; //@line 1210 "fastmix.cpp"
  $119 = $118 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $120 = Math_imul($119, $37)|0; //@line 1210 "fastmix.cpp"
  $121 = (($38) + -3)|0; //@line 1210 "fastmix.cpp"
  $122 = (($$) + ($121<<1)|0); //@line 1210 "fastmix.cpp"
  $123 = HEAP16[$122>>1]|0; //@line 1210 "fastmix.cpp"
  $124 = $123 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $125 = Math_imul($124, $47)|0; //@line 1210 "fastmix.cpp"
  $126 = (($125) + ($120))|0; //@line 1210 "fastmix.cpp"
  $127 = (($38) + -1)|0; //@line 1210 "fastmix.cpp"
  $128 = (($$) + ($127<<1)|0); //@line 1210 "fastmix.cpp"
  $129 = HEAP16[$128>>1]|0; //@line 1210 "fastmix.cpp"
  $130 = $129 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $131 = Math_imul($130, $57)|0; //@line 1210 "fastmix.cpp"
  $132 = (($126) + ($131))|0; //@line 1210 "fastmix.cpp"
  $133 = $38 | 1; //@line 1210 "fastmix.cpp"
  $134 = (($$) + ($133<<1)|0); //@line 1210 "fastmix.cpp"
  $135 = HEAP16[$134>>1]|0; //@line 1210 "fastmix.cpp"
  $136 = $135 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $137 = Math_imul($136, $67)|0; //@line 1210 "fastmix.cpp"
  $138 = (($132) + ($137))|0; //@line 1210 "fastmix.cpp"
  $139 = (($38) + 3)|0; //@line 1210 "fastmix.cpp"
  $140 = (($$) + ($139<<1)|0); //@line 1210 "fastmix.cpp"
  $141 = HEAP16[$140>>1]|0; //@line 1210 "fastmix.cpp"
  $142 = $141 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $143 = Math_imul($142, $76)|0; //@line 1210 "fastmix.cpp"
  $144 = (($38) + 5)|0; //@line 1210 "fastmix.cpp"
  $145 = (($$) + ($144<<1)|0); //@line 1210 "fastmix.cpp"
  $146 = HEAP16[$145>>1]|0; //@line 1210 "fastmix.cpp"
  $147 = $146 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $148 = Math_imul($147, $85)|0; //@line 1210 "fastmix.cpp"
  $149 = (($148) + ($143))|0; //@line 1210 "fastmix.cpp"
  $150 = (($38) + 7)|0; //@line 1210 "fastmix.cpp"
  $151 = (($$) + ($150<<1)|0); //@line 1210 "fastmix.cpp"
  $152 = HEAP16[$151>>1]|0; //@line 1210 "fastmix.cpp"
  $153 = $152 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $154 = Math_imul($153, $95)|0; //@line 1210 "fastmix.cpp"
  $155 = (($149) + ($154))|0; //@line 1210 "fastmix.cpp"
  $156 = (($38) + 9)|0; //@line 1210 "fastmix.cpp"
  $157 = (($$) + ($156<<1)|0); //@line 1210 "fastmix.cpp"
  $158 = HEAP16[$157>>1]|0; //@line 1210 "fastmix.cpp"
  $159 = $158 << 16 >> 16; //@line 1210 "fastmix.cpp"
  $160 = Math_imul($159, $105)|0; //@line 1210 "fastmix.cpp"
  $161 = (($155) + ($160))|0; //@line 1210 "fastmix.cpp"
  $162 = $138 >> 1; //@line 1210 "fastmix.cpp"
  $163 = $161 >> 1; //@line 1210 "fastmix.cpp"
  $164 = (($163) + ($162))|0; //@line 1210 "fastmix.cpp"
  $165 = $164 >> 14; //@line 1210 "fastmix.cpp"
  $166 = HEAP32[$24>>2]|0; //@line 1211 "fastmix.cpp"
  $167 = Math_imul($166, $115)|0; //@line 1211 "fastmix.cpp"
  $168 = HEAP32[$25>>2]|0; //@line 1211 "fastmix.cpp"
  $169 = Math_imul($168, $fy1$0)|0; //@line 1211 "fastmix.cpp"
  $170 = HEAP32[$26>>2]|0; //@line 1211 "fastmix.cpp"
  $171 = Math_imul($170, $fy2$0)|0; //@line 1211 "fastmix.cpp"
  $172 = (($167) + 4096)|0; //@line 1211 "fastmix.cpp"
  $173 = (($172) + ($169))|0; //@line 1211 "fastmix.cpp"
  $174 = (($173) + ($171))|0; //@line 1211 "fastmix.cpp"
  $175 = $174 >> 13; //@line 1211 "fastmix.cpp"
  $176 = Math_imul($165, $166)|0; //@line 1211 "fastmix.cpp"
  $177 = Math_imul($168, $fy3$0)|0; //@line 1211 "fastmix.cpp"
  $178 = Math_imul($170, $fy4$0)|0; //@line 1211 "fastmix.cpp"
  $179 = (($177) + 4096)|0; //@line 1211 "fastmix.cpp"
  $180 = (($179) + ($178))|0; //@line 1211 "fastmix.cpp"
  $181 = (($180) + ($176))|0; //@line 1211 "fastmix.cpp"
  $182 = $181 >> 13; //@line 1211 "fastmix.cpp"
  $183 = HEAP32[$27>>2]|0; //@line 1212 "fastmix.cpp"
  $184 = (($183) + ($nRampLeftVol$0))|0; //@line 1212 "fastmix.cpp"
  $185 = HEAP32[$28>>2]|0; //@line 1212 "fastmix.cpp"
  $186 = (($185) + ($nRampRightVol$0))|0; //@line 1212 "fastmix.cpp"
  $187 = $186 >> 12; //@line 1212 "fastmix.cpp"
  $188 = Math_imul($175, $187)|0; //@line 1212 "fastmix.cpp"
  $189 = HEAP32[$pvol$0>>2]|0; //@line 1212 "fastmix.cpp"
  $190 = (($188) + ($189))|0; //@line 1212 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $190; //@line 1212 "fastmix.cpp"
  $191 = $184 >> 12; //@line 1212 "fastmix.cpp"
  $192 = Math_imul($182, $191)|0; //@line 1212 "fastmix.cpp"
  $193 = ((($pvol$0)) + 4|0); //@line 1212 "fastmix.cpp"
  $194 = HEAP32[$193>>2]|0; //@line 1212 "fastmix.cpp"
  $195 = (($192) + ($194))|0; //@line 1212 "fastmix.cpp"
  HEAP32[$193>>2] = $195; //@line 1212 "fastmix.cpp"
  $196 = ((($pvol$0)) + 8|0); //@line 1212 "fastmix.cpp"
  $197 = HEAP32[$29>>2]|0; //@line 1213 "fastmix.cpp"
  $198 = (($197) + ($nPos$0))|0; //@line 1213 "fastmix.cpp"
  $199 = ($196>>>0)<($pbufmax>>>0); //@line 1213 "fastmix.cpp"
  if ($199) {
   $fy4$0$phi = $fy3$0;$fy2$0$phi = $fy1$0;$fy1$0 = $175;$fy3$0 = $182;$nPos$0 = $198;$nRampLeftVol$0 = $184;$nRampRightVol$0 = $186;$pvol$0 = $196;$fy4$0 = $fy4$0$phi;$fy2$0 = $fy2$0$phi;
  } else {
   $$lcssa = $175;$$lcssa31 = $182;$$lcssa32 = $184;$$lcssa33 = $186;$$lcssa34 = $187;$$lcssa35 = $191;$$lcssa36 = $198;$fy1$0$lcssa = $fy1$0;$fy3$0$lcssa = $fy3$0;
   break;
  }
 }
 $200 = $$lcssa36 >> 16; //@line 1213 "fastmix.cpp"
 $201 = HEAP32[$15>>2]|0; //@line 1213 "fastmix.cpp"
 $202 = (($201) + ($200))|0; //@line 1213 "fastmix.cpp"
 HEAP32[$15>>2] = $202; //@line 1213 "fastmix.cpp"
 $203 = $$lcssa36 & 65535; //@line 1213 "fastmix.cpp"
 HEAP32[$12>>2] = $203; //@line 1213 "fastmix.cpp"
 HEAP32[$4>>2] = $$lcssa; //@line 1213 "fastmix.cpp"
 HEAP32[$6>>2] = $fy1$0$lcssa; //@line 1213 "fastmix.cpp"
 HEAP32[$8>>2] = $$lcssa31; //@line 1213 "fastmix.cpp"
 HEAP32[$10>>2] = $fy3$0$lcssa; //@line 1213 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa33; //@line 1213 "fastmix.cpp"
 $204 = ((($pChannel)) + 16|0); //@line 1213 "fastmix.cpp"
 HEAP32[$204>>2] = $$lcssa34; //@line 1213 "fastmix.cpp"
 HEAP32[$2>>2] = $$lcssa32; //@line 1213 "fastmix.cpp"
 $205 = ((($pChannel)) + 20|0); //@line 1213 "fastmix.cpp"
 HEAP32[$205>>2] = $$lcssa35; //@line 1213 "fastmix.cpp"
 return; //@line 1213 "fastmix.cpp"
}
function __Z15FastMono8BitMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$sum = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 780 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 780 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 780 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 780 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 780 "fastmix.cpp"
 $5 = ((($pChannel)) + 36|0); //@line 780 "fastmix.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 780 "fastmix.cpp"
 $7 = $6 & 64; //@line 780 "fastmix.cpp"
 $8 = ($7|0)==(0); //@line 780 "fastmix.cpp"
 $9 = (($2) + ($4)|0); //@line 780 "fastmix.cpp"
 $$ = $8 ? $2 : $9; //@line 780 "fastmix.cpp"
 $10 = ((($pChannel)) + 16|0); //@line 782 "fastmix.cpp"
 $11 = ((($pChannel)) + 12|0); //@line 783 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $12 = $nPos$0 >> 16; //@line 781 "fastmix.cpp"
  $$sum = (($12) + ($4))|0; //@line 781 "fastmix.cpp"
  $13 = (($$) + ($$sum)|0); //@line 781 "fastmix.cpp"
  $14 = HEAP8[$13>>0]|0; //@line 781 "fastmix.cpp"
  $15 = $14 << 24 >> 24; //@line 781 "fastmix.cpp"
  $16 = $15 << 8; //@line 781 "fastmix.cpp"
  $17 = HEAP32[$10>>2]|0; //@line 782 "fastmix.cpp"
  $18 = Math_imul($16, $17)|0; //@line 782 "fastmix.cpp"
  $19 = HEAP32[$pvol$0>>2]|0; //@line 782 "fastmix.cpp"
  $20 = (($18) + ($19))|0; //@line 782 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $20; //@line 782 "fastmix.cpp"
  $21 = ((($pvol$0)) + 4|0); //@line 782 "fastmix.cpp"
  $22 = HEAP32[$21>>2]|0; //@line 782 "fastmix.cpp"
  $23 = (($22) + ($18))|0; //@line 782 "fastmix.cpp"
  HEAP32[$21>>2] = $23; //@line 782 "fastmix.cpp"
  $24 = ((($pvol$0)) + 8|0); //@line 782 "fastmix.cpp"
  $25 = HEAP32[$11>>2]|0; //@line 783 "fastmix.cpp"
  $26 = (($25) + ($nPos$0))|0; //@line 783 "fastmix.cpp"
  $27 = ($24>>>0)<($pbufmax>>>0); //@line 783 "fastmix.cpp"
  if ($27) {
   $nPos$0 = $26;$pvol$0 = $24;
  } else {
   $$lcssa = $26;
   break;
  }
 }
 $28 = $$lcssa >> 16; //@line 783 "fastmix.cpp"
 $29 = HEAP32[$3>>2]|0; //@line 783 "fastmix.cpp"
 $30 = (($29) + ($28))|0; //@line 783 "fastmix.cpp"
 HEAP32[$3>>2] = $30; //@line 783 "fastmix.cpp"
 $31 = $$lcssa & 65535; //@line 783 "fastmix.cpp"
 HEAP32[$0>>2] = $31; //@line 783 "fastmix.cpp"
 return; //@line 783 "fastmix.cpp"
}
function __Z16FastMono16BitMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 786 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 786 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 786 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 786 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 786 "fastmix.cpp"
 $5 = $4 << 1; //@line 786 "fastmix.cpp"
 $6 = (($2) + ($5)|0); //@line 786 "fastmix.cpp"
 $7 = ((($pChannel)) + 36|0); //@line 786 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 786 "fastmix.cpp"
 $9 = $8 & 64; //@line 786 "fastmix.cpp"
 $10 = ($9|0)==(0); //@line 786 "fastmix.cpp"
 $11 = (($6) + ($4<<1)|0); //@line 786 "fastmix.cpp"
 $$ = $10 ? $6 : $11; //@line 786 "fastmix.cpp"
 $12 = ((($pChannel)) + 16|0); //@line 788 "fastmix.cpp"
 $13 = ((($pChannel)) + 12|0); //@line 789 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $14 = $nPos$0 >> 16; //@line 787 "fastmix.cpp"
  $15 = (($$) + ($14<<1)|0); //@line 787 "fastmix.cpp"
  $16 = HEAP16[$15>>1]|0; //@line 787 "fastmix.cpp"
  $17 = $16 << 16 >> 16; //@line 787 "fastmix.cpp"
  $18 = HEAP32[$12>>2]|0; //@line 788 "fastmix.cpp"
  $19 = Math_imul($17, $18)|0; //@line 788 "fastmix.cpp"
  $20 = HEAP32[$pvol$0>>2]|0; //@line 788 "fastmix.cpp"
  $21 = (($20) + ($19))|0; //@line 788 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $21; //@line 788 "fastmix.cpp"
  $22 = ((($pvol$0)) + 4|0); //@line 788 "fastmix.cpp"
  $23 = HEAP32[$22>>2]|0; //@line 788 "fastmix.cpp"
  $24 = (($23) + ($19))|0; //@line 788 "fastmix.cpp"
  HEAP32[$22>>2] = $24; //@line 788 "fastmix.cpp"
  $25 = ((($pvol$0)) + 8|0); //@line 788 "fastmix.cpp"
  $26 = HEAP32[$13>>2]|0; //@line 789 "fastmix.cpp"
  $27 = (($26) + ($nPos$0))|0; //@line 789 "fastmix.cpp"
  $28 = ($25>>>0)<($pbufmax>>>0); //@line 789 "fastmix.cpp"
  if ($28) {
   $nPos$0 = $27;$pvol$0 = $25;
  } else {
   $$lcssa = $27;
   break;
  }
 }
 $29 = $$lcssa >> 16; //@line 789 "fastmix.cpp"
 $30 = HEAP32[$3>>2]|0; //@line 789 "fastmix.cpp"
 $31 = (($30) + ($29))|0; //@line 789 "fastmix.cpp"
 HEAP32[$3>>2] = $31; //@line 789 "fastmix.cpp"
 $32 = $$lcssa & 65535; //@line 789 "fastmix.cpp"
 HEAP32[$0>>2] = $32; //@line 789 "fastmix.cpp"
 return; //@line 789 "fastmix.cpp"
}
function __Z19FastMono8BitRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa12 = 0, $$lcssa13 = 0, $$sum = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 829 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 829 "fastmix.cpp"
 $2 = ((($pChannel)) + 8|0); //@line 830 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 830 "fastmix.cpp"
 $4 = HEAP32[$pChannel>>2]|0; //@line 830 "fastmix.cpp"
 $5 = ((($pChannel)) + 4|0); //@line 830 "fastmix.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 830 "fastmix.cpp"
 $7 = ((($pChannel)) + 36|0); //@line 830 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 830 "fastmix.cpp"
 $9 = $8 & 64; //@line 830 "fastmix.cpp"
 $10 = ($9|0)==(0); //@line 830 "fastmix.cpp"
 $11 = (($4) + ($6)|0); //@line 830 "fastmix.cpp"
 $$ = $10 ? $4 : $11; //@line 830 "fastmix.cpp"
 $12 = ((($pChannel)) + 24|0); //@line 832 "fastmix.cpp"
 $13 = ((($pChannel)) + 12|0); //@line 833 "fastmix.cpp"
 $nPos$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $14 = $nPos$0 >> 16; //@line 831 "fastmix.cpp"
  $$sum = (($14) + ($6))|0; //@line 831 "fastmix.cpp"
  $15 = (($$) + ($$sum)|0); //@line 831 "fastmix.cpp"
  $16 = HEAP8[$15>>0]|0; //@line 831 "fastmix.cpp"
  $17 = $16 << 24 >> 24; //@line 831 "fastmix.cpp"
  $18 = $17 << 8; //@line 831 "fastmix.cpp"
  $19 = HEAP32[$12>>2]|0; //@line 832 "fastmix.cpp"
  $20 = (($19) + ($nRampRightVol$0))|0; //@line 832 "fastmix.cpp"
  $21 = $20 >> 12; //@line 832 "fastmix.cpp"
  $22 = Math_imul($21, $18)|0; //@line 832 "fastmix.cpp"
  $23 = HEAP32[$pvol$0>>2]|0; //@line 832 "fastmix.cpp"
  $24 = (($22) + ($23))|0; //@line 832 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $24; //@line 832 "fastmix.cpp"
  $25 = ((($pvol$0)) + 4|0); //@line 832 "fastmix.cpp"
  $26 = HEAP32[$25>>2]|0; //@line 832 "fastmix.cpp"
  $27 = (($22) + ($26))|0; //@line 832 "fastmix.cpp"
  HEAP32[$25>>2] = $27; //@line 832 "fastmix.cpp"
  $28 = ((($pvol$0)) + 8|0); //@line 832 "fastmix.cpp"
  $29 = HEAP32[$13>>2]|0; //@line 833 "fastmix.cpp"
  $30 = (($29) + ($nPos$0))|0; //@line 833 "fastmix.cpp"
  $31 = ($28>>>0)<($pbufmax>>>0); //@line 833 "fastmix.cpp"
  if ($31) {
   $nPos$0 = $30;$nRampRightVol$0 = $20;$pvol$0 = $28;
  } else {
   $$lcssa = $20;$$lcssa12 = $21;$$lcssa13 = $30;
   break;
  }
 }
 $32 = $$lcssa13 >> 16; //@line 833 "fastmix.cpp"
 $33 = HEAP32[$5>>2]|0; //@line 833 "fastmix.cpp"
 $34 = (($33) + ($32))|0; //@line 833 "fastmix.cpp"
 HEAP32[$5>>2] = $34; //@line 833 "fastmix.cpp"
 $35 = $$lcssa13 & 65535; //@line 833 "fastmix.cpp"
 HEAP32[$2>>2] = $35; //@line 833 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 833 "fastmix.cpp"
 $36 = ((($pChannel)) + 52|0); //@line 833 "fastmix.cpp"
 HEAP32[$36>>2] = $$lcssa; //@line 833 "fastmix.cpp"
 $37 = ((($pChannel)) + 16|0); //@line 833 "fastmix.cpp"
 HEAP32[$37>>2] = $$lcssa12; //@line 833 "fastmix.cpp"
 $38 = ((($pChannel)) + 20|0); //@line 833 "fastmix.cpp"
 HEAP32[$38>>2] = $$lcssa12; //@line 833 "fastmix.cpp"
 return; //@line 833 "fastmix.cpp"
}
function __Z20FastMono16BitRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa11 = 0, $$lcssa12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 835 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 835 "fastmix.cpp"
 $2 = ((($pChannel)) + 8|0); //@line 836 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 836 "fastmix.cpp"
 $4 = HEAP32[$pChannel>>2]|0; //@line 836 "fastmix.cpp"
 $5 = ((($pChannel)) + 4|0); //@line 836 "fastmix.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 836 "fastmix.cpp"
 $7 = $6 << 1; //@line 836 "fastmix.cpp"
 $8 = (($4) + ($7)|0); //@line 836 "fastmix.cpp"
 $9 = ((($pChannel)) + 36|0); //@line 836 "fastmix.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 836 "fastmix.cpp"
 $11 = $10 & 64; //@line 836 "fastmix.cpp"
 $12 = ($11|0)==(0); //@line 836 "fastmix.cpp"
 $13 = (($8) + ($6<<1)|0); //@line 836 "fastmix.cpp"
 $$ = $12 ? $8 : $13; //@line 836 "fastmix.cpp"
 $14 = ((($pChannel)) + 24|0); //@line 838 "fastmix.cpp"
 $15 = ((($pChannel)) + 12|0); //@line 839 "fastmix.cpp"
 $nPos$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $16 = $nPos$0 >> 16; //@line 837 "fastmix.cpp"
  $17 = (($$) + ($16<<1)|0); //@line 837 "fastmix.cpp"
  $18 = HEAP16[$17>>1]|0; //@line 837 "fastmix.cpp"
  $19 = $18 << 16 >> 16; //@line 837 "fastmix.cpp"
  $20 = HEAP32[$14>>2]|0; //@line 838 "fastmix.cpp"
  $21 = (($20) + ($nRampRightVol$0))|0; //@line 838 "fastmix.cpp"
  $22 = $21 >> 12; //@line 838 "fastmix.cpp"
  $23 = Math_imul($22, $19)|0; //@line 838 "fastmix.cpp"
  $24 = HEAP32[$pvol$0>>2]|0; //@line 838 "fastmix.cpp"
  $25 = (($23) + ($24))|0; //@line 838 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $25; //@line 838 "fastmix.cpp"
  $26 = ((($pvol$0)) + 4|0); //@line 838 "fastmix.cpp"
  $27 = HEAP32[$26>>2]|0; //@line 838 "fastmix.cpp"
  $28 = (($23) + ($27))|0; //@line 838 "fastmix.cpp"
  HEAP32[$26>>2] = $28; //@line 838 "fastmix.cpp"
  $29 = ((($pvol$0)) + 8|0); //@line 838 "fastmix.cpp"
  $30 = HEAP32[$15>>2]|0; //@line 839 "fastmix.cpp"
  $31 = (($30) + ($nPos$0))|0; //@line 839 "fastmix.cpp"
  $32 = ($29>>>0)<($pbufmax>>>0); //@line 839 "fastmix.cpp"
  if ($32) {
   $nPos$0 = $31;$nRampRightVol$0 = $21;$pvol$0 = $29;
  } else {
   $$lcssa = $21;$$lcssa11 = $22;$$lcssa12 = $31;
   break;
  }
 }
 $33 = $$lcssa12 >> 16; //@line 839 "fastmix.cpp"
 $34 = HEAP32[$5>>2]|0; //@line 839 "fastmix.cpp"
 $35 = (($34) + ($33))|0; //@line 839 "fastmix.cpp"
 HEAP32[$5>>2] = $35; //@line 839 "fastmix.cpp"
 $36 = $$lcssa12 & 65535; //@line 839 "fastmix.cpp"
 HEAP32[$2>>2] = $36; //@line 839 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 839 "fastmix.cpp"
 $37 = ((($pChannel)) + 52|0); //@line 839 "fastmix.cpp"
 HEAP32[$37>>2] = $$lcssa; //@line 839 "fastmix.cpp"
 $38 = ((($pChannel)) + 16|0); //@line 839 "fastmix.cpp"
 HEAP32[$38>>2] = $$lcssa11; //@line 839 "fastmix.cpp"
 $39 = ((($pChannel)) + 20|0); //@line 839 "fastmix.cpp"
 HEAP32[$39>>2] = $$lcssa11; //@line 839 "fastmix.cpp"
 return; //@line 839 "fastmix.cpp"
}
function __Z21FastMono8BitLinearMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$sum = 0, $$sum1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 792 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 792 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 792 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 792 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 792 "fastmix.cpp"
 $5 = (($2) + ($4)|0); //@line 792 "fastmix.cpp"
 $6 = ((($pChannel)) + 36|0); //@line 792 "fastmix.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 792 "fastmix.cpp"
 $8 = $7 & 64; //@line 792 "fastmix.cpp"
 $9 = ($8|0)==(0); //@line 792 "fastmix.cpp"
 $$ = $9 ? $2 : $5; //@line 792 "fastmix.cpp"
 $10 = (($4) + 1)|0; //@line 793 "fastmix.cpp"
 $11 = ((($pChannel)) + 16|0); //@line 794 "fastmix.cpp"
 $12 = ((($pChannel)) + 12|0); //@line 795 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $13 = $nPos$0 >> 16; //@line 793 "fastmix.cpp"
  $14 = $nPos$0 >>> 8; //@line 793 "fastmix.cpp"
  $15 = $14 & 255; //@line 793 "fastmix.cpp"
  $$sum = (($13) + ($4))|0; //@line 793 "fastmix.cpp"
  $16 = (($$) + ($$sum)|0); //@line 793 "fastmix.cpp"
  $17 = HEAP8[$16>>0]|0; //@line 793 "fastmix.cpp"
  $18 = $17 << 24 >> 24; //@line 793 "fastmix.cpp"
  $$sum1 = (($10) + ($13))|0; //@line 793 "fastmix.cpp"
  $19 = (($$) + ($$sum1)|0); //@line 793 "fastmix.cpp"
  $20 = HEAP8[$19>>0]|0; //@line 793 "fastmix.cpp"
  $21 = $20 << 24 >> 24; //@line 793 "fastmix.cpp"
  $22 = $18 << 8; //@line 793 "fastmix.cpp"
  $23 = (($21) - ($18))|0; //@line 793 "fastmix.cpp"
  $24 = Math_imul($23, $15)|0; //@line 793 "fastmix.cpp"
  $25 = (($24) + ($22))|0; //@line 793 "fastmix.cpp"
  $26 = HEAP32[$11>>2]|0; //@line 794 "fastmix.cpp"
  $27 = Math_imul($25, $26)|0; //@line 794 "fastmix.cpp"
  $28 = HEAP32[$pvol$0>>2]|0; //@line 794 "fastmix.cpp"
  $29 = (($27) + ($28))|0; //@line 794 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $29; //@line 794 "fastmix.cpp"
  $30 = ((($pvol$0)) + 4|0); //@line 794 "fastmix.cpp"
  $31 = HEAP32[$30>>2]|0; //@line 794 "fastmix.cpp"
  $32 = (($27) + ($31))|0; //@line 794 "fastmix.cpp"
  HEAP32[$30>>2] = $32; //@line 794 "fastmix.cpp"
  $33 = ((($pvol$0)) + 8|0); //@line 794 "fastmix.cpp"
  $34 = HEAP32[$12>>2]|0; //@line 795 "fastmix.cpp"
  $35 = (($34) + ($nPos$0))|0; //@line 795 "fastmix.cpp"
  $36 = ($33>>>0)<($pbufmax>>>0); //@line 795 "fastmix.cpp"
  if ($36) {
   $nPos$0 = $35;$pvol$0 = $33;
  } else {
   $$lcssa = $35;
   break;
  }
 }
 $37 = $$lcssa >> 16; //@line 795 "fastmix.cpp"
 $38 = HEAP32[$3>>2]|0; //@line 795 "fastmix.cpp"
 $39 = (($38) + ($37))|0; //@line 795 "fastmix.cpp"
 HEAP32[$3>>2] = $39; //@line 795 "fastmix.cpp"
 $40 = $$lcssa & 65535; //@line 795 "fastmix.cpp"
 HEAP32[$0>>2] = $40; //@line 795 "fastmix.cpp"
 return; //@line 795 "fastmix.cpp"
}
function __Z22FastMono16BitLinearMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 8|0); //@line 798 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 798 "fastmix.cpp"
 $2 = HEAP32[$pChannel>>2]|0; //@line 798 "fastmix.cpp"
 $3 = ((($pChannel)) + 4|0); //@line 798 "fastmix.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 798 "fastmix.cpp"
 $5 = $4 << 1; //@line 798 "fastmix.cpp"
 $6 = (($2) + ($5)|0); //@line 798 "fastmix.cpp"
 $7 = ((($pChannel)) + 36|0); //@line 798 "fastmix.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 798 "fastmix.cpp"
 $9 = $8 & 64; //@line 798 "fastmix.cpp"
 $10 = ($9|0)==(0); //@line 798 "fastmix.cpp"
 $11 = (($6) + ($4<<1)|0); //@line 798 "fastmix.cpp"
 $$ = $10 ? $6 : $11; //@line 798 "fastmix.cpp"
 $12 = ((($pChannel)) + 16|0); //@line 800 "fastmix.cpp"
 $13 = ((($pChannel)) + 12|0); //@line 801 "fastmix.cpp"
 $nPos$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $14 = $nPos$0 >> 16; //@line 799 "fastmix.cpp"
  $15 = $nPos$0 >>> 8; //@line 799 "fastmix.cpp"
  $16 = $15 & 255; //@line 799 "fastmix.cpp"
  $17 = (($$) + ($14<<1)|0); //@line 799 "fastmix.cpp"
  $18 = HEAP16[$17>>1]|0; //@line 799 "fastmix.cpp"
  $19 = $18 << 16 >> 16; //@line 799 "fastmix.cpp"
  $20 = (($14) + 1)|0; //@line 799 "fastmix.cpp"
  $21 = (($$) + ($20<<1)|0); //@line 799 "fastmix.cpp"
  $22 = HEAP16[$21>>1]|0; //@line 799 "fastmix.cpp"
  $23 = $22 << 16 >> 16; //@line 799 "fastmix.cpp"
  $24 = (($23) - ($19))|0; //@line 799 "fastmix.cpp"
  $25 = Math_imul($24, $16)|0; //@line 799 "fastmix.cpp"
  $26 = $25 >> 8; //@line 799 "fastmix.cpp"
  $27 = (($26) + ($19))|0; //@line 799 "fastmix.cpp"
  $28 = HEAP32[$12>>2]|0; //@line 800 "fastmix.cpp"
  $29 = Math_imul($27, $28)|0; //@line 800 "fastmix.cpp"
  $30 = HEAP32[$pvol$0>>2]|0; //@line 800 "fastmix.cpp"
  $31 = (($29) + ($30))|0; //@line 800 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $31; //@line 800 "fastmix.cpp"
  $32 = ((($pvol$0)) + 4|0); //@line 800 "fastmix.cpp"
  $33 = HEAP32[$32>>2]|0; //@line 800 "fastmix.cpp"
  $34 = (($29) + ($33))|0; //@line 800 "fastmix.cpp"
  HEAP32[$32>>2] = $34; //@line 800 "fastmix.cpp"
  $35 = ((($pvol$0)) + 8|0); //@line 800 "fastmix.cpp"
  $36 = HEAP32[$13>>2]|0; //@line 801 "fastmix.cpp"
  $37 = (($36) + ($nPos$0))|0; //@line 801 "fastmix.cpp"
  $38 = ($35>>>0)<($pbufmax>>>0); //@line 801 "fastmix.cpp"
  if ($38) {
   $nPos$0 = $37;$pvol$0 = $35;
  } else {
   $$lcssa = $37;
   break;
  }
 }
 $39 = $$lcssa >> 16; //@line 801 "fastmix.cpp"
 $40 = HEAP32[$3>>2]|0; //@line 801 "fastmix.cpp"
 $41 = (($40) + ($39))|0; //@line 801 "fastmix.cpp"
 HEAP32[$3>>2] = $41; //@line 801 "fastmix.cpp"
 $42 = $$lcssa & 65535; //@line 801 "fastmix.cpp"
 HEAP32[$0>>2] = $42; //@line 801 "fastmix.cpp"
 return; //@line 801 "fastmix.cpp"
}
function __Z25FastMono8BitLinearRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa14 = 0, $$lcssa15 = 0, $$sum = 0, $$sum1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 841 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 841 "fastmix.cpp"
 $2 = ((($pChannel)) + 8|0); //@line 842 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 842 "fastmix.cpp"
 $4 = HEAP32[$pChannel>>2]|0; //@line 842 "fastmix.cpp"
 $5 = ((($pChannel)) + 4|0); //@line 842 "fastmix.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 842 "fastmix.cpp"
 $7 = (($4) + ($6)|0); //@line 842 "fastmix.cpp"
 $8 = ((($pChannel)) + 36|0); //@line 842 "fastmix.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 842 "fastmix.cpp"
 $10 = $9 & 64; //@line 842 "fastmix.cpp"
 $11 = ($10|0)==(0); //@line 842 "fastmix.cpp"
 $$ = $11 ? $4 : $7; //@line 842 "fastmix.cpp"
 $12 = (($6) + 1)|0; //@line 843 "fastmix.cpp"
 $13 = ((($pChannel)) + 24|0); //@line 844 "fastmix.cpp"
 $14 = ((($pChannel)) + 12|0); //@line 845 "fastmix.cpp"
 $nPos$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $15 = $nPos$0 >> 16; //@line 843 "fastmix.cpp"
  $16 = $nPos$0 >>> 8; //@line 843 "fastmix.cpp"
  $17 = $16 & 255; //@line 843 "fastmix.cpp"
  $$sum = (($15) + ($6))|0; //@line 843 "fastmix.cpp"
  $18 = (($$) + ($$sum)|0); //@line 843 "fastmix.cpp"
  $19 = HEAP8[$18>>0]|0; //@line 843 "fastmix.cpp"
  $20 = $19 << 24 >> 24; //@line 843 "fastmix.cpp"
  $$sum1 = (($12) + ($15))|0; //@line 843 "fastmix.cpp"
  $21 = (($$) + ($$sum1)|0); //@line 843 "fastmix.cpp"
  $22 = HEAP8[$21>>0]|0; //@line 843 "fastmix.cpp"
  $23 = $22 << 24 >> 24; //@line 843 "fastmix.cpp"
  $24 = $20 << 8; //@line 843 "fastmix.cpp"
  $25 = (($23) - ($20))|0; //@line 843 "fastmix.cpp"
  $26 = Math_imul($25, $17)|0; //@line 843 "fastmix.cpp"
  $27 = (($26) + ($24))|0; //@line 843 "fastmix.cpp"
  $28 = HEAP32[$13>>2]|0; //@line 844 "fastmix.cpp"
  $29 = (($28) + ($nRampRightVol$0))|0; //@line 844 "fastmix.cpp"
  $30 = $29 >> 12; //@line 844 "fastmix.cpp"
  $31 = Math_imul($27, $30)|0; //@line 844 "fastmix.cpp"
  $32 = HEAP32[$pvol$0>>2]|0; //@line 844 "fastmix.cpp"
  $33 = (($31) + ($32))|0; //@line 844 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $33; //@line 844 "fastmix.cpp"
  $34 = ((($pvol$0)) + 4|0); //@line 844 "fastmix.cpp"
  $35 = HEAP32[$34>>2]|0; //@line 844 "fastmix.cpp"
  $36 = (($31) + ($35))|0; //@line 844 "fastmix.cpp"
  HEAP32[$34>>2] = $36; //@line 844 "fastmix.cpp"
  $37 = ((($pvol$0)) + 8|0); //@line 844 "fastmix.cpp"
  $38 = HEAP32[$14>>2]|0; //@line 845 "fastmix.cpp"
  $39 = (($38) + ($nPos$0))|0; //@line 845 "fastmix.cpp"
  $40 = ($37>>>0)<($pbufmax>>>0); //@line 845 "fastmix.cpp"
  if ($40) {
   $nPos$0 = $39;$nRampRightVol$0 = $29;$pvol$0 = $37;
  } else {
   $$lcssa = $29;$$lcssa14 = $30;$$lcssa15 = $39;
   break;
  }
 }
 $41 = $$lcssa15 >> 16; //@line 845 "fastmix.cpp"
 $42 = HEAP32[$5>>2]|0; //@line 845 "fastmix.cpp"
 $43 = (($42) + ($41))|0; //@line 845 "fastmix.cpp"
 HEAP32[$5>>2] = $43; //@line 845 "fastmix.cpp"
 $44 = $$lcssa15 & 65535; //@line 845 "fastmix.cpp"
 HEAP32[$2>>2] = $44; //@line 845 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 845 "fastmix.cpp"
 $45 = ((($pChannel)) + 52|0); //@line 845 "fastmix.cpp"
 HEAP32[$45>>2] = $$lcssa; //@line 845 "fastmix.cpp"
 $46 = ((($pChannel)) + 16|0); //@line 845 "fastmix.cpp"
 HEAP32[$46>>2] = $$lcssa14; //@line 845 "fastmix.cpp"
 $47 = ((($pChannel)) + 20|0); //@line 845 "fastmix.cpp"
 HEAP32[$47>>2] = $$lcssa14; //@line 845 "fastmix.cpp"
 return; //@line 845 "fastmix.cpp"
}
function __Z26FastMono16BitLinearRampMixP11_MODCHANNELPiS1_($pChannel,$pbuffer,$pbufmax) {
 $pChannel = $pChannel|0;
 $pbuffer = $pbuffer|0;
 $pbufmax = $pbufmax|0;
 var $$ = 0, $$lcssa = 0, $$lcssa11 = 0, $$lcssa12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nPos$0 = 0, $nRampRightVol$0 = 0, $pvol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pChannel)) + 48|0); //@line 847 "fastmix.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 847 "fastmix.cpp"
 $2 = ((($pChannel)) + 8|0); //@line 848 "fastmix.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 848 "fastmix.cpp"
 $4 = HEAP32[$pChannel>>2]|0; //@line 848 "fastmix.cpp"
 $5 = ((($pChannel)) + 4|0); //@line 848 "fastmix.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 848 "fastmix.cpp"
 $7 = $6 << 1; //@line 848 "fastmix.cpp"
 $8 = (($4) + ($7)|0); //@line 848 "fastmix.cpp"
 $9 = ((($pChannel)) + 36|0); //@line 848 "fastmix.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 848 "fastmix.cpp"
 $11 = $10 & 64; //@line 848 "fastmix.cpp"
 $12 = ($11|0)==(0); //@line 848 "fastmix.cpp"
 $13 = (($8) + ($6<<1)|0); //@line 848 "fastmix.cpp"
 $$ = $12 ? $8 : $13; //@line 848 "fastmix.cpp"
 $14 = ((($pChannel)) + 24|0); //@line 850 "fastmix.cpp"
 $15 = ((($pChannel)) + 12|0); //@line 851 "fastmix.cpp"
 $nPos$0 = $3;$nRampRightVol$0 = $1;$pvol$0 = $pbuffer;
 while(1) {
  $16 = $nPos$0 >> 16; //@line 849 "fastmix.cpp"
  $17 = $nPos$0 >>> 8; //@line 849 "fastmix.cpp"
  $18 = $17 & 255; //@line 849 "fastmix.cpp"
  $19 = (($$) + ($16<<1)|0); //@line 849 "fastmix.cpp"
  $20 = HEAP16[$19>>1]|0; //@line 849 "fastmix.cpp"
  $21 = $20 << 16 >> 16; //@line 849 "fastmix.cpp"
  $22 = (($16) + 1)|0; //@line 849 "fastmix.cpp"
  $23 = (($$) + ($22<<1)|0); //@line 849 "fastmix.cpp"
  $24 = HEAP16[$23>>1]|0; //@line 849 "fastmix.cpp"
  $25 = $24 << 16 >> 16; //@line 849 "fastmix.cpp"
  $26 = (($25) - ($21))|0; //@line 849 "fastmix.cpp"
  $27 = Math_imul($26, $18)|0; //@line 849 "fastmix.cpp"
  $28 = $27 >> 8; //@line 849 "fastmix.cpp"
  $29 = (($28) + ($21))|0; //@line 849 "fastmix.cpp"
  $30 = HEAP32[$14>>2]|0; //@line 850 "fastmix.cpp"
  $31 = (($30) + ($nRampRightVol$0))|0; //@line 850 "fastmix.cpp"
  $32 = $31 >> 12; //@line 850 "fastmix.cpp"
  $33 = Math_imul($29, $32)|0; //@line 850 "fastmix.cpp"
  $34 = HEAP32[$pvol$0>>2]|0; //@line 850 "fastmix.cpp"
  $35 = (($33) + ($34))|0; //@line 850 "fastmix.cpp"
  HEAP32[$pvol$0>>2] = $35; //@line 850 "fastmix.cpp"
  $36 = ((($pvol$0)) + 4|0); //@line 850 "fastmix.cpp"
  $37 = HEAP32[$36>>2]|0; //@line 850 "fastmix.cpp"
  $38 = (($33) + ($37))|0; //@line 850 "fastmix.cpp"
  HEAP32[$36>>2] = $38; //@line 850 "fastmix.cpp"
  $39 = ((($pvol$0)) + 8|0); //@line 850 "fastmix.cpp"
  $40 = HEAP32[$15>>2]|0; //@line 851 "fastmix.cpp"
  $41 = (($40) + ($nPos$0))|0; //@line 851 "fastmix.cpp"
  $42 = ($39>>>0)<($pbufmax>>>0); //@line 851 "fastmix.cpp"
  if ($42) {
   $nPos$0 = $41;$nRampRightVol$0 = $31;$pvol$0 = $39;
  } else {
   $$lcssa = $31;$$lcssa11 = $32;$$lcssa12 = $41;
   break;
  }
 }
 $43 = $$lcssa12 >> 16; //@line 851 "fastmix.cpp"
 $44 = HEAP32[$5>>2]|0; //@line 851 "fastmix.cpp"
 $45 = (($44) + ($43))|0; //@line 851 "fastmix.cpp"
 HEAP32[$5>>2] = $45; //@line 851 "fastmix.cpp"
 $46 = $$lcssa12 & 65535; //@line 851 "fastmix.cpp"
 HEAP32[$2>>2] = $46; //@line 851 "fastmix.cpp"
 HEAP32[$0>>2] = $$lcssa; //@line 851 "fastmix.cpp"
 $47 = ((($pChannel)) + 52|0); //@line 851 "fastmix.cpp"
 HEAP32[$47>>2] = $$lcssa; //@line 851 "fastmix.cpp"
 $48 = ((($pChannel)) + 16|0); //@line 851 "fastmix.cpp"
 HEAP32[$48>>2] = $$lcssa11; //@line 851 "fastmix.cpp"
 $49 = ((($pChannel)) + 20|0); //@line 851 "fastmix.cpp"
 HEAP32[$49>>2] = $$lcssa11; //@line 851 "fastmix.cpp"
 return; //@line 851 "fastmix.cpp"
}
function __ZN10CSoundFile16ProcessStereoDSPEi($count) {
 $count = $count|0;
 var $$ = 0, $$1 = 0, $$2 = 0, $$3 = 0, $$4 = 0, $$lcssa47 = 0, $$lcssa48 = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $$pre29 = 0, $$pre30 = 0, $$pre31 = 0, $$pre32 = 0, $$pre32$pre = 0, $$pre33 = 0, $$pre34 = 0, $$pre34$pre = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $n1$0$lcssa = 0, $n1$05 = 0, $n2$0$lcssa = 0, $n2$018 = 0, $n28$0$lcssa = 0, $n28$06 = 0, $n5$0$lcssa = 0, $n5$011 = 0, $nr$08 = 0;
 var $pin$0 = 0, $pnr$07 = 0, $pr$0 = 0, $pr1$019 = 0, $px$010 = 0, $r$017 = 0, $rvbcount$0 = 0, $x$012 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[183080]|0; //@line 205 "snd_dsp.cpp"
 $1 = $0 & 128; //@line 205 "snd_dsp.cpp"
 $2 = ($1|0)==(0); //@line 205 "snd_dsp.cpp"
 if ($2) {
  $86 = $0;
 } else {
  $$pre = HEAP32[196995]|0; //@line 210 "snd_dsp.cpp"
  $$pre27 = HEAP32[196994]|0; //@line 210 "snd_dsp.cpp"
  $$pre28 = HEAP32[196993]|0; //@line 211 "snd_dsp.cpp"
  $$pre29 = HEAP32[196992]|0; //@line 211 "snd_dsp.cpp"
  $11 = $$pre28;$15 = $$pre29;$4 = $$pre;$7 = $$pre27;$pin$0 = 740556;$pr$0 = 732364;$rvbcount$0 = $count;
  while(1) {
   $3 = (788528 + ($4<<2)|0); //@line 210 "snd_dsp.cpp"
   $5 = HEAP32[$3>>2]|0; //@line 210 "snd_dsp.cpp"
   $6 = (942128 + ($7<<2)|0); //@line 210 "snd_dsp.cpp"
   $8 = HEAP32[$6>>2]|0; //@line 210 "snd_dsp.cpp"
   $9 = (($8) + ($5))|0; //@line 210 "snd_dsp.cpp"
   $10 = (1059584 + ($11<<2)|0); //@line 211 "snd_dsp.cpp"
   $12 = HEAP32[$10>>2]|0; //@line 211 "snd_dsp.cpp"
   $13 = (($9) + ($12))|0; //@line 211 "snd_dsp.cpp"
   $14 = (1142288 + ($15<<2)|0); //@line 211 "snd_dsp.cpp"
   $16 = HEAP32[$14>>2]|0; //@line 211 "snd_dsp.cpp"
   $17 = (($13) + ($16))|0; //@line 211 "snd_dsp.cpp"
   $18 = HEAP32[196996]|0; //@line 213 "snd_dsp.cpp"
   $19 = (788272 + ($18<<2)|0); //@line 213 "snd_dsp.cpp"
   $20 = HEAP32[$19>>2]|0; //@line 213 "snd_dsp.cpp"
   $21 = $17 >> 1; //@line 214 "snd_dsp.cpp"
   HEAP32[$19>>2] = $21; //@line 214 "snd_dsp.cpp"
   $22 = (($18) + 1)|0; //@line 215 "snd_dsp.cpp"
   $23 = $22 & 31; //@line 216 "snd_dsp.cpp"
   HEAP32[196996] = $23; //@line 216 "snd_dsp.cpp"
   $24 = HEAP32[196997]|0; //@line 217 "snd_dsp.cpp"
   $25 = (788016 + ($24<<2)|0); //@line 218 "snd_dsp.cpp"
   $26 = HEAP32[$25>>2]|0; //@line 218 "snd_dsp.cpp"
   $27 = HEAP32[196998]|0; //@line 218 "snd_dsp.cpp"
   $28 = (($27) - ($26))|0; //@line 218 "snd_dsp.cpp"
   $29 = (($17|0) / 128)&-1; //@line 219 "snd_dsp.cpp"
   HEAP32[$25>>2] = $29; //@line 220 "snd_dsp.cpp"
   $30 = (($28) + ($29))|0; //@line 221 "snd_dsp.cpp"
   HEAP32[196998] = $30; //@line 221 "snd_dsp.cpp"
   $31 = (($20) - ($30))|0; //@line 222 "snd_dsp.cpp"
   $32 = (($24) + 1)|0; //@line 223 "snd_dsp.cpp"
   $33 = $32 & 63; //@line 223 "snd_dsp.cpp"
   HEAP32[196997] = $33; //@line 223 "snd_dsp.cpp"
   $34 = HEAP32[$pin$0>>2]|0; //@line 225 "snd_dsp.cpp"
   $35 = ((($pin$0)) + 4|0); //@line 225 "snd_dsp.cpp"
   $36 = HEAP32[$35>>2]|0; //@line 225 "snd_dsp.cpp"
   $37 = (($36) + ($34))|0; //@line 225 "snd_dsp.cpp"
   $38 = HEAP32[196991]|0; //@line 225 "snd_dsp.cpp"
   $39 = $37 >> $38; //@line 225 "snd_dsp.cpp"
   $40 = (($31) + ($34))|0; //@line 226 "snd_dsp.cpp"
   $41 = HEAP32[$pr$0>>2]|0; //@line 226 "snd_dsp.cpp"
   $42 = (($40) + ($41))|0; //@line 226 "snd_dsp.cpp"
   HEAP32[$pr$0>>2] = $42; //@line 226 "snd_dsp.cpp"
   $43 = (($36) + ($31))|0; //@line 227 "snd_dsp.cpp"
   $44 = ((($pr$0)) + 4|0); //@line 227 "snd_dsp.cpp"
   $45 = HEAP32[$44>>2]|0; //@line 227 "snd_dsp.cpp"
   $46 = (($43) + ($45))|0; //@line 227 "snd_dsp.cpp"
   HEAP32[$44>>2] = $46; //@line 227 "snd_dsp.cpp"
   $47 = $31 >> 2; //@line 228 "snd_dsp.cpp"
   $48 = (($39) + ($47))|0; //@line 228 "snd_dsp.cpp"
   $49 = HEAP32[196993]|0; //@line 229 "snd_dsp.cpp"
   $50 = (1059584 + ($49<<2)|0); //@line 229 "snd_dsp.cpp"
   HEAP32[$50>>2] = $48; //@line 229 "snd_dsp.cpp"
   $51 = HEAP32[196992]|0; //@line 230 "snd_dsp.cpp"
   $52 = (1142288 + ($51<<2)|0); //@line 230 "snd_dsp.cpp"
   HEAP32[$52>>2] = $48; //@line 230 "snd_dsp.cpp"
   $53 = $31 >> 4; //@line 231 "snd_dsp.cpp"
   $54 = (($48) + ($53))|0; //@line 231 "snd_dsp.cpp"
   $55 = $54 >> 1; //@line 232 "snd_dsp.cpp"
   $56 = HEAP32[196999]|0; //@line 233 "snd_dsp.cpp"
   $57 = (1198876 + ($56<<2)|0); //@line 233 "snd_dsp.cpp"
   $58 = HEAP32[$57>>2]|0; //@line 233 "snd_dsp.cpp"
   $59 = HEAP32[197000]|0; //@line 233 "snd_dsp.cpp"
   $60 = (($55) - ($58))|0; //@line 233 "snd_dsp.cpp"
   $61 = (($60) + ($59))|0; //@line 234 "snd_dsp.cpp"
   HEAP32[197000] = $61; //@line 234 "snd_dsp.cpp"
   HEAP32[$57>>2] = $55; //@line 235 "snd_dsp.cpp"
   $62 = (($56) + 1)|0; //@line 236 "snd_dsp.cpp"
   $63 = $62 & 7; //@line 237 "snd_dsp.cpp"
   HEAP32[196999] = $63; //@line 237 "snd_dsp.cpp"
   $64 = $61 >> 2; //@line 238 "snd_dsp.cpp"
   $65 = HEAP32[196995]|0; //@line 239 "snd_dsp.cpp"
   $66 = (788528 + ($65<<2)|0); //@line 239 "snd_dsp.cpp"
   HEAP32[$66>>2] = $64; //@line 239 "snd_dsp.cpp"
   $67 = HEAP32[196994]|0; //@line 240 "snd_dsp.cpp"
   $68 = (942128 + ($67<<2)|0); //@line 240 "snd_dsp.cpp"
   HEAP32[$68>>2] = $64; //@line 240 "snd_dsp.cpp"
   $69 = (($65) + 1)|0; //@line 241 "snd_dsp.cpp"
   $70 = HEAP32[196990]|0; //@line 241 "snd_dsp.cpp"
   $71 = ($69|0)>=($70|0); //@line 241 "snd_dsp.cpp"
   $$ = $71 ? 0 : $69; //@line 241 "snd_dsp.cpp"
   HEAP32[196995] = $$; //@line 241 "snd_dsp.cpp"
   $72 = (($67) + 1)|0; //@line 242 "snd_dsp.cpp"
   $73 = HEAP32[197001]|0; //@line 242 "snd_dsp.cpp"
   $74 = ($72|0)>=($73|0); //@line 242 "snd_dsp.cpp"
   $$3 = $74 ? 0 : $72; //@line 242 "snd_dsp.cpp"
   HEAP32[196994] = $$3; //@line 242 "snd_dsp.cpp"
   $75 = (($49) + 1)|0; //@line 243 "snd_dsp.cpp"
   $76 = HEAP32[197002]|0; //@line 243 "snd_dsp.cpp"
   $77 = ($75|0)>=($76|0); //@line 243 "snd_dsp.cpp"
   $$1 = $77 ? 0 : $75; //@line 243 "snd_dsp.cpp"
   HEAP32[196993] = $$1; //@line 243 "snd_dsp.cpp"
   $78 = (($51) + 1)|0; //@line 244 "snd_dsp.cpp"
   $79 = HEAP32[197003]|0; //@line 244 "snd_dsp.cpp"
   $80 = ($78|0)>=($79|0); //@line 244 "snd_dsp.cpp"
   $$4 = $80 ? 0 : $78; //@line 244 "snd_dsp.cpp"
   HEAP32[196992] = $$4; //@line 244 "snd_dsp.cpp"
   $81 = ((($pr$0)) + 8|0); //@line 245 "snd_dsp.cpp"
   $82 = ((($pin$0)) + 8|0); //@line 246 "snd_dsp.cpp"
   $83 = (($rvbcount$0) + -1)|0; //@line 247 "snd_dsp.cpp"
   $84 = ($83|0)==(0); //@line 247 "snd_dsp.cpp"
   if ($84) {
    break;
   } else {
    $11 = $$1;$15 = $$4;$4 = $$;$7 = $$3;$pin$0 = $82;$pr$0 = $81;$rvbcount$0 = $83;
   }
  }
  $$pre30 = HEAP32[183080]|0; //@line 251 "snd_dsp.cpp"
  $86 = $$pre30;
 }
 $85 = $86 & 64; //@line 251 "snd_dsp.cpp"
 $87 = ($85|0)==(0); //@line 251 "snd_dsp.cpp"
 if ($87) {
  $135 = $86;
 } else {
  $88 = HEAP32[187194]|0; //@line 253 "snd_dsp.cpp"
  $89 = ($count|0)==(0); //@line 254 "snd_dsp.cpp"
  if ($89) {
   $$pre32 = $86;$n2$0$lcssa = $88;
  } else {
   $$pre31 = HEAP32[187191]|0; //@line 268 "snd_dsp.cpp"
   $107 = $$pre31;$n2$018 = $88;$pr1$019 = 732364;$r$017 = $count;
   while(1) {
    $90 = HEAP32[$pr1$019>>2]|0; //@line 256 "snd_dsp.cpp"
    $91 = ((($pr1$019)) + 4|0); //@line 256 "snd_dsp.cpp"
    $92 = HEAP32[$91>>2]|0; //@line 256 "snd_dsp.cpp"
    $93 = (($90) + 31)|0; //@line 256 "snd_dsp.cpp"
    $94 = (($93) + ($92))|0; //@line 256 "snd_dsp.cpp"
    $95 = $94 >> 7; //@line 256 "snd_dsp.cpp"
    $96 = HEAP32[196989]|0; //@line 258 "snd_dsp.cpp"
    $97 = Math_imul($95, $96)|0; //@line 258 "snd_dsp.cpp"
    $98 = HEAP32[187196]|0; //@line 261 "snd_dsp.cpp"
    $99 = (749044 + ($98<<2)|0); //@line 261 "snd_dsp.cpp"
    $100 = HEAP32[$99>>2]|0; //@line 261 "snd_dsp.cpp"
    $101 = HEAP32[187195]|0; //@line 261 "snd_dsp.cpp"
    $102 = (($101) - ($100))|0; //@line 261 "snd_dsp.cpp"
    HEAP32[$99>>2] = $97; //@line 262 "snd_dsp.cpp"
    $103 = (($102) + ($97))|0; //@line 263 "snd_dsp.cpp"
    HEAP32[187195] = $103; //@line 263 "snd_dsp.cpp"
    $104 = (($98) + 1)|0; //@line 265 "snd_dsp.cpp"
    $105 = $104 & 3; //@line 266 "snd_dsp.cpp"
    HEAP32[187196] = $105; //@line 266 "snd_dsp.cpp"
    $106 = (749556 + ($107<<2)|0); //@line 268 "snd_dsp.cpp"
    $108 = HEAP32[$106>>2]|0; //@line 268 "snd_dsp.cpp"
    HEAP32[$106>>2] = $103; //@line 269 "snd_dsp.cpp"
    $109 = HEAP32[187192]|0; //@line 271 "snd_dsp.cpp"
    $110 = (749300 + ($109<<2)|0); //@line 271 "snd_dsp.cpp"
    $111 = HEAP32[$110>>2]|0; //@line 271 "snd_dsp.cpp"
    HEAP32[$110>>2] = $108; //@line 272 "snd_dsp.cpp"
    $112 = (($109) + 1)|0; //@line 273 "snd_dsp.cpp"
    $113 = $112 & 31; //@line 274 "snd_dsp.cpp"
    HEAP32[187192] = $113; //@line 274 "snd_dsp.cpp"
    $114 = (748788 + ($n2$018<<2)|0); //@line 275 "snd_dsp.cpp"
    $115 = HEAP32[$114>>2]|0; //@line 275 "snd_dsp.cpp"
    $116 = HEAP32[187193]|0; //@line 275 "snd_dsp.cpp"
    $117 = (($116) - ($115))|0; //@line 275 "snd_dsp.cpp"
    $118 = (($108|0) / 64)&-1; //@line 276 "snd_dsp.cpp"
    HEAP32[$114>>2] = $118; //@line 277 "snd_dsp.cpp"
    $119 = (($117) + ($118))|0; //@line 278 "snd_dsp.cpp"
    HEAP32[187193] = $119; //@line 278 "snd_dsp.cpp"
    $120 = (($111) - ($119))|0; //@line 279 "snd_dsp.cpp"
    $121 = (($n2$018) + 1)|0; //@line 280 "snd_dsp.cpp"
    $122 = $121 & 63; //@line 281 "snd_dsp.cpp"
    $123 = HEAP32[$pr1$019>>2]|0; //@line 283 "snd_dsp.cpp"
    $124 = (($120) + ($123))|0; //@line 283 "snd_dsp.cpp"
    HEAP32[$pr1$019>>2] = $124; //@line 283 "snd_dsp.cpp"
    $125 = HEAP32[$91>>2]|0; //@line 284 "snd_dsp.cpp"
    $126 = (($125) - ($120))|0; //@line 284 "snd_dsp.cpp"
    HEAP32[$91>>2] = $126; //@line 284 "snd_dsp.cpp"
    $127 = HEAP32[187191]|0; //@line 285 "snd_dsp.cpp"
    $128 = (($127) + 1)|0; //@line 285 "snd_dsp.cpp"
    $129 = HEAP32[187190]|0; //@line 285 "snd_dsp.cpp"
    $130 = ($128|0)>=($129|0); //@line 285 "snd_dsp.cpp"
    $$2 = $130 ? 0 : $128; //@line 285 "snd_dsp.cpp"
    HEAP32[187191] = $$2; //@line 285 "snd_dsp.cpp"
    $131 = ((($pr1$019)) + 8|0); //@line 286 "snd_dsp.cpp"
    $132 = (($r$017) + -1)|0; //@line 254 "snd_dsp.cpp"
    $133 = ($132|0)==(0); //@line 254 "snd_dsp.cpp"
    if ($133) {
     $$lcssa48 = $122;
     break;
    } else {
     $107 = $$2;$n2$018 = $122;$pr1$019 = $131;$r$017 = $132;
    }
   }
   $$pre32$pre = HEAP32[183080]|0; //@line 291 "snd_dsp.cpp"
   $$pre32 = $$pre32$pre;$n2$0$lcssa = $$lcssa48;
  }
  HEAP32[187194] = $n2$0$lcssa; //@line 288 "snd_dsp.cpp"
  $135 = $$pre32;
 }
 $134 = $135 & 32; //@line 291 "snd_dsp.cpp"
 $136 = ($134|0)==(0); //@line 291 "snd_dsp.cpp"
 if ($136) {
  $178 = $135;
 } else {
  $137 = HEAP32[10113]|0; //@line 294 "snd_dsp.cpp"
  $138 = (($137) + 1)|0; //@line 294 "snd_dsp.cpp"
  $139 = 1 << $138; //@line 294 "snd_dsp.cpp"
  $140 = (($139) + -1)|0; //@line 294 "snd_dsp.cpp"
  $141 = HEAP32[299729]|0; //@line 295 "snd_dsp.cpp"
  $142 = ($count|0)==(0); //@line 296 "snd_dsp.cpp"
  if ($142) {
   $$pre34 = $135;$n5$0$lcssa = $141;
  } else {
   $$pre33 = HEAP32[299728]|0; //@line 303 "snd_dsp.cpp"
   $157 = $$pre33;$n5$011 = $141;$px$010 = 732364;$x$012 = $count;
   while(1) {
    $143 = (1198924 + ($n5$011<<2)|0); //@line 298 "snd_dsp.cpp"
    $144 = HEAP32[$143>>2]|0; //@line 298 "snd_dsp.cpp"
    $145 = HEAP32[299730]|0; //@line 298 "snd_dsp.cpp"
    $146 = (($145) - ($144))|0; //@line 298 "snd_dsp.cpp"
    HEAP32[299730] = $146; //@line 298 "snd_dsp.cpp"
    $147 = HEAP32[$px$010>>2]|0; //@line 299 "snd_dsp.cpp"
    $148 = ((($px$010)) + 4|0); //@line 299 "snd_dsp.cpp"
    $149 = HEAP32[$148>>2]|0; //@line 299 "snd_dsp.cpp"
    $150 = (($149) + ($147))|0; //@line 299 "snd_dsp.cpp"
    $151 = $150 >> 31; //@line 300 "snd_dsp.cpp"
    $152 = $151 & $140; //@line 300 "snd_dsp.cpp"
    $153 = (($152) + ($150))|0; //@line 300 "snd_dsp.cpp"
    $154 = $153 >> $138; //@line 300 "snd_dsp.cpp"
    HEAP32[$143>>2] = $154; //@line 301 "snd_dsp.cpp"
    $155 = (($154) + ($146))|0; //@line 302 "snd_dsp.cpp"
    HEAP32[299730] = $155; //@line 302 "snd_dsp.cpp"
    $156 = (1199180 + ($157<<2)|0); //@line 303 "snd_dsp.cpp"
    $158 = HEAP32[$156>>2]|0; //@line 303 "snd_dsp.cpp"
    $159 = HEAP32[$px$010>>2]|0; //@line 304 "snd_dsp.cpp"
    HEAP32[$156>>2] = $159; //@line 304 "snd_dsp.cpp"
    $160 = (($155) + ($158))|0; //@line 305 "snd_dsp.cpp"
    HEAP32[$px$010>>2] = $160; //@line 305 "snd_dsp.cpp"
    $161 = HEAP32[299728]|0; //@line 306 "snd_dsp.cpp"
    $162 = (($161) + 1)|0; //@line 306 "snd_dsp.cpp"
    $163 = (1199180 + ($162<<2)|0); //@line 306 "snd_dsp.cpp"
    $164 = HEAP32[$163>>2]|0; //@line 306 "snd_dsp.cpp"
    $165 = HEAP32[$148>>2]|0; //@line 307 "snd_dsp.cpp"
    HEAP32[$163>>2] = $165; //@line 307 "snd_dsp.cpp"
    $166 = HEAP32[299730]|0; //@line 308 "snd_dsp.cpp"
    $167 = (($166) + ($164))|0; //@line 308 "snd_dsp.cpp"
    HEAP32[$148>>2] = $167; //@line 308 "snd_dsp.cpp"
    $168 = HEAP32[299728]|0; //@line 309 "snd_dsp.cpp"
    $169 = (($168) + 2)|0; //@line 309 "snd_dsp.cpp"
    $170 = HEAP32[299727]|0; //@line 309 "snd_dsp.cpp"
    $171 = $169 & $170; //@line 309 "snd_dsp.cpp"
    HEAP32[299728] = $171; //@line 309 "snd_dsp.cpp"
    $172 = ((($px$010)) + 8|0); //@line 310 "snd_dsp.cpp"
    $173 = (($n5$011) + 1)|0; //@line 311 "snd_dsp.cpp"
    $174 = $170 & $173; //@line 312 "snd_dsp.cpp"
    $175 = (($x$012) + -1)|0; //@line 296 "snd_dsp.cpp"
    $176 = ($175|0)==(0); //@line 296 "snd_dsp.cpp"
    if ($176) {
     $$lcssa47 = $174;
     break;
    } else {
     $157 = $171;$n5$011 = $174;$px$010 = $172;$x$012 = $175;
    }
   }
   $$pre34$pre = HEAP32[183080]|0; //@line 317 "snd_dsp.cpp"
   $$pre34 = $$pre34$pre;$n5$0$lcssa = $$lcssa47;
  }
  HEAP32[299729] = $n5$0$lcssa; //@line 314 "snd_dsp.cpp"
  $178 = $$pre34;
 }
 $177 = $178 & 2; //@line 317 "snd_dsp.cpp"
 $179 = ($177|0)==(0); //@line 317 "snd_dsp.cpp"
 if ($179) {
  return; //@line 334 "snd_dsp.cpp"
 }
 $180 = HEAP32[187189]|0; //@line 319 "snd_dsp.cpp"
 $181 = HEAP32[187188]|0; //@line 319 "snd_dsp.cpp"
 $182 = ($count|0)==(0); //@line 321 "snd_dsp.cpp"
 if ($182) {
  $n1$0$lcssa = $180;$n28$0$lcssa = $181;
 } else {
  $n1$05 = $180;$n28$06 = $181;$nr$08 = $count;$pnr$07 = 732364;
  while(1) {
   $183 = HEAP32[$pnr$07>>2]|0; //@line 323 "snd_dsp.cpp"
   $184 = $183 >> 1; //@line 323 "snd_dsp.cpp"
   $185 = (($184) + ($n1$05))|0; //@line 324 "snd_dsp.cpp"
   HEAP32[$pnr$07>>2] = $185; //@line 324 "snd_dsp.cpp"
   $186 = ((($pnr$07)) + 4|0); //@line 326 "snd_dsp.cpp"
   $187 = HEAP32[$186>>2]|0; //@line 326 "snd_dsp.cpp"
   $188 = $187 >> 1; //@line 326 "snd_dsp.cpp"
   $189 = (($188) + ($n28$06))|0; //@line 327 "snd_dsp.cpp"
   HEAP32[$186>>2] = $189; //@line 327 "snd_dsp.cpp"
   $190 = ((($pnr$07)) + 8|0); //@line 329 "snd_dsp.cpp"
   $191 = (($nr$08) + -1)|0; //@line 321 "snd_dsp.cpp"
   $192 = ($191|0)==(0); //@line 321 "snd_dsp.cpp"
   if ($192) {
    $n1$0$lcssa = $184;$n28$0$lcssa = $188;
    break;
   } else {
    $n1$05 = $184;$n28$06 = $188;$nr$08 = $191;$pnr$07 = $190;
   }
  }
 }
 HEAP32[187189] = $n1$0$lcssa; //@line 331 "snd_dsp.cpp"
 HEAP32[187188] = $n28$0$lcssa; //@line 332 "snd_dsp.cpp"
 return; //@line 334 "snd_dsp.cpp"
}
function __Z18X86_MonoFromStereoPij($pMixBuf,$nSamples) {
 $pMixBuf = $pMixBuf|0;
 $nSamples = $nSamples|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($nSamples|0)==(0); //@line 2107 "fastmix.cpp"
 if ($0) {
  return; //@line 2112 "fastmix.cpp"
 } else {
  $i$01 = 0;
 }
 while(1) {
  $1 = $i$01 << 1; //@line 2109 "fastmix.cpp"
  $2 = (($pMixBuf) + ($1<<2)|0); //@line 2110 "fastmix.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 2110 "fastmix.cpp"
  $4 = $1 | 1; //@line 2110 "fastmix.cpp"
  $5 = (($pMixBuf) + ($4<<2)|0); //@line 2110 "fastmix.cpp"
  $6 = HEAP32[$5>>2]|0; //@line 2110 "fastmix.cpp"
  $7 = (($6) + ($3))|0; //@line 2110 "fastmix.cpp"
  $8 = $7 >> 1; //@line 2110 "fastmix.cpp"
  $9 = (($pMixBuf) + ($i$01<<2)|0); //@line 2110 "fastmix.cpp"
  HEAP32[$9>>2] = $8; //@line 2110 "fastmix.cpp"
  $10 = (($i$01) + 1)|0; //@line 2107 "fastmix.cpp"
  $exitcond = ($10|0)==($nSamples|0); //@line 2107 "fastmix.cpp"
  if ($exitcond) {
   break;
  } else {
   $i$01 = $10;
  }
 }
 return; //@line 2112 "fastmix.cpp"
}
function __Z23X86_InterleaveFrontRearPiS_j($pFrontBuf,$pRearBuf,$nSamples) {
 $pFrontBuf = $pFrontBuf|0;
 $pRearBuf = $pRearBuf|0;
 $nSamples = $nSamples|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return; //@line 2078 "fastmix.cpp"
}
function __ZN10CSoundFileC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 59680|0); //@line 51 "sndfile.cpp"
 HEAP32[$0>>2] = 0; //@line 51 "sndfile.cpp"
 $1 = ((($this)) + 59660|0); //@line 52 "sndfile.cpp"
 HEAP32[$1>>2] = 0; //@line 52 "sndfile.cpp"
 $2 = ((($this)) + 59664|0); //@line 53 "sndfile.cpp"
 HEAP32[$2>>2] = 0; //@line 53 "sndfile.cpp"
 $3 = ((($this)) + 59668|0); //@line 54 "sndfile.cpp"
 HEAP32[$3>>2] = 0; //@line 54 "sndfile.cpp"
 $4 = ((($this)) + 59684|0); //@line 55 "sndfile.cpp"
 HEAP32[$4>>2] = 0; //@line 55 "sndfile.cpp"
 $5 = ((($this)) + 59688|0); //@line 56 "sndfile.cpp"
 HEAP32[$5>>2] = 0; //@line 56 "sndfile.cpp"
 $6 = ((($this)) + 59792|0); //@line 57 "sndfile.cpp"
 HEAP32[$6>>2] = 0; //@line 57 "sndfile.cpp"
 $7 = ((($this)) + 59800|0); //@line 58 "sndfile.cpp"
 HEAP32[$7>>2] = 0; //@line 58 "sndfile.cpp"
 $8 = ((($this)) + 59796|0); //@line 59 "sndfile.cpp"
 HEAP32[$8>>2] = 0; //@line 59 "sndfile.cpp"
 $9 = ((($this)) + 59756|0); //@line 60 "sndfile.cpp"
 HEAP32[$9>>2] = 128; //@line 60 "sndfile.cpp"
 $10 = ((($this)) + 59752|0); //@line 60 "sndfile.cpp"
 HEAP32[$10>>2] = 128; //@line 60 "sndfile.cpp"
 $11 = ((($this)) + 59740|0); //@line 61 "sndfile.cpp"
 HEAP32[$11>>2] = 128; //@line 61 "sndfile.cpp"
 $12 = ((($this)) + 59764|0); //@line 62 "sndfile.cpp"
 HEAP32[$12>>2] = 32; //@line 62 "sndfile.cpp"
 $13 = ((($this)) + 59768|0); //@line 63 "sndfile.cpp"
 HEAP32[$13>>2] = 32767; //@line 63 "sndfile.cpp"
 $14 = ((($this)) + 59772|0); //@line 64 "sndfile.cpp"
 HEAP32[$14>>2] = 0; //@line 64 "sndfile.cpp"
 $15 = ((($this)) + 53344|0); //@line 70 "sndfile.cpp"
 _memset(($this|0),0,51904)|0; //@line 65 "sndfile.cpp"
 _memset(($15|0),-1,256)|0; //@line 70 "sndfile.cpp"
 $16 = ((($this)) + 51904|0); //@line 71 "sndfile.cpp"
 _memset(($16|0),0,960)|0; //@line 71 "sndfile.cpp"
 $17 = ((($this)) + 59804|0); //@line 72 "sndfile.cpp"
 _memset(($17|0),0,7680)|0; //@line 72 "sndfile.cpp"
 $18 = ((($this)) + 58496|0); //@line 73 "sndfile.cpp"
 _memset(($18|0),0,1152)|0; //@line 73 "sndfile.cpp"
 return; //@line 74 "sndfile.cpp"
}
function __ZN10CSoundFileD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 (__ZN10CSoundFile7DestroyEv($this)|0); //@line 80 "sndfile.cpp"
 return; //@line 81 "sndfile.cpp"
}
function __ZN10CSoundFile7DestroyEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $exitcond7 = 0;
 var $exitcond8 = 0, $exitcond9 = 0, $i$06 = 0, $i$15 = 0, $i$24 = 0, $i$33 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $i$06 = 0;
 while(1) {
  $0 = (((($this)) + 51904|0) + ($i$06<<2)|0); //@line 280 "sndfile.cpp"
  $1 = HEAP32[$0>>2]|0; //@line 280 "sndfile.cpp"
  $2 = ($1|0)==(0|0); //@line 280 "sndfile.cpp"
  if (!($2)) {
   __ZdaPv($1); //@line 349 "sndfile.cpp"
   HEAP32[$0>>2] = 0; //@line 283 "sndfile.cpp"
  }
  $3 = (($i$06) + 1)|0; //@line 280 "sndfile.cpp"
  $exitcond9 = ($3|0)==(240); //@line 280 "sndfile.cpp"
  if ($exitcond9) {
   break;
  } else {
   $i$06 = $3;
  }
 }
 $4 = ((($this)) + 59792|0); //@line 285 "sndfile.cpp"
 HEAP32[$4>>2] = 0; //@line 285 "sndfile.cpp"
 $5 = ((($this)) + 59800|0); //@line 286 "sndfile.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 286 "sndfile.cpp"
 $7 = ($6|0)==(0|0); //@line 286 "sndfile.cpp"
 if (!($7)) {
  __ZdaPv($6); //@line 288 "sndfile.cpp"
  HEAP32[$5>>2] = 0; //@line 289 "sndfile.cpp"
 }
 $8 = ((($this)) + 59796|0); //@line 291 "sndfile.cpp"
 $9 = HEAP32[$8>>2]|0; //@line 291 "sndfile.cpp"
 $10 = ($9|0)==(0|0); //@line 291 "sndfile.cpp"
 if ($10) {
  $i$15 = 1;
 } else {
  __ZdaPv($9); //@line 293 "sndfile.cpp"
  HEAP32[$8>>2] = 0; //@line 294 "sndfile.cpp"
  $i$15 = 1;
 }
 while(1) {
  $11 = (((((($this)) + 33280|0) + ($i$15<<6)|0)) + 20|0); //@line 299 "sndfile.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 299 "sndfile.cpp"
  $13 = ($12|0)==(0|0); //@line 299 "sndfile.cpp"
  if (!($13)) {
   $14 = ((($12)) + -16|0); //@line 367 "sndfile.cpp"
   _free($14); //@line 367 "sndfile.cpp"
   HEAP32[$11>>2] = 0; //@line 302 "sndfile.cpp"
  }
  $15 = (($i$15) + 1)|0; //@line 296 "sndfile.cpp"
  $exitcond8 = ($15|0)==(240); //@line 296 "sndfile.cpp"
  if ($exitcond8) {
   $i$24 = 0;
   break;
  } else {
   $i$15 = $15;
  }
 }
 while(1) {
  $16 = (((($this)) + 48640|0) + ($i$24<<2)|0); //@line 307 "sndfile.cpp"
  $17 = HEAP32[$16>>2]|0; //@line 307 "sndfile.cpp"
  $18 = ($17|0)==(0|0); //@line 307 "sndfile.cpp"
  if (!($18)) {
   __ZdlPv($17); //@line 309 "sndfile.cpp"
   HEAP32[$16>>2] = 0; //@line 310 "sndfile.cpp"
  }
  $19 = (($i$24) + 1)|0; //@line 305 "sndfile.cpp"
  $exitcond7 = ($19|0)==(240); //@line 305 "sndfile.cpp"
  if ($exitcond7) {
   $i$33 = 0;
   break;
  } else {
   $i$24 = $19;
  }
 }
 while(1) {
  $20 = (((((($this)) + 58496|0) + (($i$33*144)|0)|0)) + 8|0); //@line 315 "sndfile.cpp"
  $21 = HEAP32[$20>>2]|0; //@line 315 "sndfile.cpp"
  $22 = ($21|0)==(0); //@line 315 "sndfile.cpp"
  if (!($22)) {
   $23 = (((((($this)) + 58496|0) + (($i$33*144)|0)|0)) + 12|0); //@line 315 "sndfile.cpp"
   $24 = HEAP32[$23>>2]|0; //@line 315 "sndfile.cpp"
   $25 = ($24|0)==(0|0); //@line 315 "sndfile.cpp"
   if (!($25)) {
    HEAP32[$20>>2] = 0; //@line 317 "sndfile.cpp"
    __ZdaPv($24); //@line 318 "sndfile.cpp"
    HEAP32[$23>>2] = 0; //@line 319 "sndfile.cpp"
   }
  }
  $26 = (((((($this)) + 58496|0) + (($i$33*144)|0)|0)) + 4|0); //@line 321 "sndfile.cpp"
  HEAP32[$26>>2] = 0; //@line 321 "sndfile.cpp"
  $27 = (((($this)) + 58496|0) + (($i$33*144)|0)|0); //@line 322 "sndfile.cpp"
  $28 = HEAP32[$27>>2]|0; //@line 322 "sndfile.cpp"
  $29 = ($28|0)==(0|0); //@line 322 "sndfile.cpp"
  if (!($29)) {
   $30 = HEAP32[$28>>2]|0; //@line 324 "sndfile.cpp"
   $31 = ((($30)) + 12|0); //@line 324 "sndfile.cpp"
   $32 = HEAP32[$31>>2]|0; //@line 324 "sndfile.cpp"
   (FUNCTION_TABLE_ii[$32 & 8191]($28)|0); //@line 324 "sndfile.cpp"
   HEAP32[$27>>2] = 0; //@line 325 "sndfile.cpp"
  }
  $33 = (($i$33) + 1)|0; //@line 313 "sndfile.cpp"
  $exitcond = ($33|0)==(8); //@line 313 "sndfile.cpp"
  if ($exitcond) {
   break;
  } else {
   $i$33 = $33;
  }
 }
 $34 = ((($this)) + 59680|0); //@line 328 "sndfile.cpp"
 HEAP32[$34>>2] = 0; //@line 328 "sndfile.cpp"
 $35 = ((($this)) + 59688|0); //@line 329 "sndfile.cpp"
 HEAP32[$35>>2] = 0; //@line 329 "sndfile.cpp"
 $36 = ((($this)) + 59684|0); //@line 329 "sndfile.cpp"
 HEAP32[$36>>2] = 0; //@line 329 "sndfile.cpp"
 $37 = ((($this)) + 59664|0); //@line 329 "sndfile.cpp"
 HEAP32[$37>>2] = 0; //@line 329 "sndfile.cpp"
 return 1; //@line 330 "sndfile.cpp"
}
function __ZN10CSoundFile6CreateEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$0 = 0, $$lcssa = 0, $$phi$trans$insert = 0, $$pr = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $exitcond = 0, $exitcond17 = 0, $exitcond18 = 0, $exitcond19 = 0, $exitcond20 = 0, $exitcond21 = 0, $i$013 = 0, $i$19 = 0, $i$27 = 0, $iPlug$05 = 0, $j$01022 = 0, $j$01022$lcssa = 0, $j$111 = 0, $nch$015 = 0, $npt$016 = 0, $pins$08 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 4|0;
 $1 = sp;
 HEAP32[$0>>2] = $lpStream;
 HEAP32[$1>>2] = $dwMemLength;
 $2 = ((($this)) + 59680|0); //@line 89 "sndfile.cpp"
 HEAP32[$2>>2] = 0; //@line 89 "sndfile.cpp"
 $3 = ((($this)) + 59660|0); //@line 90 "sndfile.cpp"
 HEAP32[$3>>2] = 0; //@line 90 "sndfile.cpp"
 $4 = ((($this)) + 59664|0); //@line 91 "sndfile.cpp"
 HEAP32[$4>>2] = 0; //@line 91 "sndfile.cpp"
 $5 = ((($this)) + 59668|0); //@line 92 "sndfile.cpp"
 HEAP32[$5>>2] = 0; //@line 92 "sndfile.cpp"
 $6 = ((($this)) + 59684|0); //@line 93 "sndfile.cpp"
 HEAP32[$6>>2] = 0; //@line 93 "sndfile.cpp"
 $7 = ((($this)) + 59688|0); //@line 94 "sndfile.cpp"
 HEAP32[$7>>2] = 0; //@line 94 "sndfile.cpp"
 $8 = ((($this)) + 59756|0); //@line 95 "sndfile.cpp"
 HEAP32[$8>>2] = 128; //@line 95 "sndfile.cpp"
 $9 = ((($this)) + 59752|0); //@line 95 "sndfile.cpp"
 HEAP32[$9>>2] = 128; //@line 95 "sndfile.cpp"
 $10 = ((($this)) + 59740|0); //@line 96 "sndfile.cpp"
 HEAP32[$10>>2] = 128; //@line 96 "sndfile.cpp"
 $11 = ((($this)) + 59656|0); //@line 97 "sndfile.cpp"
 HEAP32[$11>>2] = 256; //@line 97 "sndfile.cpp"
 $12 = ((($this)) + 59744|0); //@line 98 "sndfile.cpp"
 HEAP32[$12>>2] = 256; //@line 98 "sndfile.cpp"
 $13 = ((($this)) + 59760|0); //@line 99 "sndfile.cpp"
 HEAP32[$13>>2] = 0; //@line 99 "sndfile.cpp"
 $14 = ((($this)) + 59648|0); //@line 100 "sndfile.cpp"
 HEAP32[$14>>2] = 6; //@line 100 "sndfile.cpp"
 $15 = ((($this)) + 59652|0); //@line 101 "sndfile.cpp"
 HEAP32[$15>>2] = 125; //@line 101 "sndfile.cpp"
 $16 = ((($this)) + 59700|0); //@line 102 "sndfile.cpp"
 HEAP32[$16>>2] = 0; //@line 102 "sndfile.cpp"
 $17 = ((($this)) + 59704|0); //@line 103 "sndfile.cpp"
 HEAP32[$17>>2] = 0; //@line 103 "sndfile.cpp"
 $18 = ((($this)) + 59716|0); //@line 104 "sndfile.cpp"
 $19 = ((($this)) + 59720|0); //@line 105 "sndfile.cpp"
 $20 = ((($this)) + 59724|0); //@line 106 "sndfile.cpp"
 $21 = ((($this)) + 59728|0); //@line 107 "sndfile.cpp"
 $22 = ((($this)) + 59732|0); //@line 108 "sndfile.cpp"
 $23 = ((($this)) + 59736|0); //@line 109 "sndfile.cpp"
 $24 = ((($this)) + 59764|0); //@line 110 "sndfile.cpp"
 ;HEAP32[$18>>2]=0|0;HEAP32[$18+4>>2]=0|0;HEAP32[$18+8>>2]=0|0;HEAP32[$18+12>>2]=0|0;HEAP32[$18+16>>2]=0|0;HEAP32[$18+20>>2]=0|0; //@line 104 "sndfile.cpp"
 HEAP32[$24>>2] = 16; //@line 110 "sndfile.cpp"
 $25 = ((($this)) + 59768|0); //@line 111 "sndfile.cpp"
 HEAP32[$25>>2] = 32767; //@line 111 "sndfile.cpp"
 $26 = ((($this)) + 59748|0); //@line 112 "sndfile.cpp"
 HEAP32[$26>>2] = 48; //@line 112 "sndfile.cpp"
 $27 = ((($this)) + 59788|0); //@line 114 "sndfile.cpp"
 $28 = ((($this)) + 53344|0); //@line 121 "sndfile.cpp"
 _memset(($this|0),0,49600)|0; //@line 117 "sndfile.cpp"
 ;HEAP32[$27>>2]=0|0;HEAP32[$27+4>>2]=0|0;HEAP32[$27+8>>2]=0|0;HEAP32[$27+12>>2]=0|0; //@line 114 "sndfile.cpp"
 _memset(($28|0),-1,256)|0; //@line 121 "sndfile.cpp"
 $29 = ((($this)) + 51904|0); //@line 122 "sndfile.cpp"
 _memset(($29|0),0,960)|0; //@line 122 "sndfile.cpp"
 $30 = ((($this)) + 59804|0); //@line 123 "sndfile.cpp"
 _memset(($30|0),0,7680)|0; //@line 123 "sndfile.cpp"
 $31 = ((($this)) + 58496|0); //@line 124 "sndfile.cpp"
 _memset(($31|0),0,1152)|0; //@line 124 "sndfile.cpp"
 __ZN10CSoundFile12ResetMidiCfgEv($this); //@line 125 "sndfile.cpp"
 $npt$016 = 0;
 while(1) {
  $32 = (((($this)) + 52864|0) + ($npt$016<<1)|0); //@line 126 "sndfile.cpp"
  HEAP16[$32>>1] = 64; //@line 126 "sndfile.cpp"
  $33 = (($npt$016) + 1)|0; //@line 126 "sndfile.cpp"
  $exitcond21 = ($33|0)==(240); //@line 126 "sndfile.cpp"
  if ($exitcond21) {
   $nch$015 = 0;
   break;
  } else {
   $npt$016 = $33;
  }
 }
 while(1) {
  $34 = (((($this)) + 49600|0) + (($nch$015*36)|0)|0); //@line 129 "sndfile.cpp"
  HEAP32[$34>>2] = 128; //@line 129 "sndfile.cpp"
  $35 = (((((($this)) + 49600|0) + (($nch$015*36)|0)|0)) + 4|0); //@line 130 "sndfile.cpp"
  HEAP32[$35>>2] = 64; //@line 130 "sndfile.cpp"
  $36 = (((((($this)) + 49600|0) + (($nch$015*36)|0)|0)) + 8|0); //@line 131 "sndfile.cpp"
  HEAP32[$36>>2] = 0; //@line 131 "sndfile.cpp"
  $37 = (((((($this)) + 49600|0) + (($nch$015*36)|0)|0)) + 16|0); //@line 132 "sndfile.cpp"
  HEAP8[$37>>0] = 0; //@line 132 "sndfile.cpp"
  $38 = (($nch$015) + 1)|0; //@line 127 "sndfile.cpp"
  $exitcond20 = ($38|0)==(64); //@line 127 "sndfile.cpp"
  if ($exitcond20) {
   break;
  } else {
   $nch$015 = $38;
  }
 }
 $39 = ($lpStream|0)==(0|0); //@line 134 "sndfile.cpp"
 if ($39) {
  $i$013 = 0;
 } else {
  $40 = (__Z12MMCMP_UnpackPPKhPj($0,$1)|0); //@line 137 "sndfile.cpp"
  $41 = HEAP32[$0>>2]|0; //@line 139 "sndfile.cpp"
  $42 = HEAP32[$1>>2]|0; //@line 139 "sndfile.cpp"
  $43 = (__ZN10CSoundFile6ReadXMEPKhj($this,$41,$42)|0); //@line 139 "sndfile.cpp"
  do {
   if (!($43)) {
    $44 = HEAP32[$0>>2]|0; //@line 140 "sndfile.cpp"
    $45 = HEAP32[$1>>2]|0; //@line 140 "sndfile.cpp"
    $46 = (__ZN10CSoundFile7ReadS3MEPKhj($this,$44,$45)|0); //@line 140 "sndfile.cpp"
    if (!($46)) {
     $47 = HEAP32[$0>>2]|0; //@line 141 "sndfile.cpp"
     $48 = HEAP32[$1>>2]|0; //@line 141 "sndfile.cpp"
     $49 = (__ZN10CSoundFile6ReadITEPKhj($this,$47,$48)|0); //@line 141 "sndfile.cpp"
     if (!($49)) {
      $50 = HEAP32[$0>>2]|0; //@line 142 "sndfile.cpp"
      $51 = HEAP32[$1>>2]|0; //@line 142 "sndfile.cpp"
      $52 = (__ZN10CSoundFile7ReadWavEPKhj($this,$50,$51)|0); //@line 142 "sndfile.cpp"
      if (!($52)) {
       $53 = HEAP32[$0>>2]|0; //@line 145 "sndfile.cpp"
       $54 = HEAP32[$1>>2]|0; //@line 145 "sndfile.cpp"
       $55 = (__ZN10CSoundFile7ReadABCEPKhj($this,$53,$54)|0); //@line 145 "sndfile.cpp"
       if (!($55)) {
        $56 = HEAP32[$0>>2]|0; //@line 146 "sndfile.cpp"
        $57 = HEAP32[$1>>2]|0; //@line 146 "sndfile.cpp"
        $58 = (__ZN10CSoundFile7ReadMIDEPKhj($this,$56,$57)|0); //@line 146 "sndfile.cpp"
        if (!($58)) {
         $59 = HEAP32[$0>>2]|0; //@line 147 "sndfile.cpp"
         $60 = HEAP32[$1>>2]|0; //@line 147 "sndfile.cpp"
         $61 = (__ZN10CSoundFile7ReadPATEPKhj($this,$59,$60)|0); //@line 147 "sndfile.cpp"
         if (!($61)) {
          $62 = HEAP32[$0>>2]|0; //@line 148 "sndfile.cpp"
          $63 = HEAP32[$1>>2]|0; //@line 148 "sndfile.cpp"
          $64 = (__ZN10CSoundFile7ReadSTMEPKhj($this,$62,$63)|0); //@line 148 "sndfile.cpp"
          if (!($64)) {
           $65 = HEAP32[$0>>2]|0; //@line 149 "sndfile.cpp"
           $66 = HEAP32[$1>>2]|0; //@line 149 "sndfile.cpp"
           $67 = (__ZN10CSoundFile7ReadMedEPKhj($this,$65,$66)|0); //@line 149 "sndfile.cpp"
           if (!($67)) {
            $68 = HEAP32[$0>>2]|0; //@line 150 "sndfile.cpp"
            $69 = HEAP32[$1>>2]|0; //@line 150 "sndfile.cpp"
            $70 = (__ZN10CSoundFile7ReadMTMEPKhj($this,$68,$69)|0); //@line 150 "sndfile.cpp"
            if (!($70)) {
             $71 = HEAP32[$0>>2]|0; //@line 151 "sndfile.cpp"
             $72 = HEAP32[$1>>2]|0; //@line 151 "sndfile.cpp"
             $73 = (__ZN10CSoundFile7ReadMDLEPKhj($this,$71,$72)|0); //@line 151 "sndfile.cpp"
             if (!($73)) {
              $74 = HEAP32[$0>>2]|0; //@line 152 "sndfile.cpp"
              $75 = HEAP32[$1>>2]|0; //@line 152 "sndfile.cpp"
              $76 = (__ZN10CSoundFile7ReadDBMEPKhj($this,$74,$75)|0); //@line 152 "sndfile.cpp"
              if (!($76)) {
               $77 = HEAP32[$0>>2]|0; //@line 153 "sndfile.cpp"
               $78 = HEAP32[$1>>2]|0; //@line 153 "sndfile.cpp"
               $79 = (__ZN10CSoundFile7Read669EPKhj($this,$77,$78)|0); //@line 153 "sndfile.cpp"
               if (!($79)) {
                $80 = HEAP32[$0>>2]|0; //@line 154 "sndfile.cpp"
                $81 = HEAP32[$1>>2]|0; //@line 154 "sndfile.cpp"
                $82 = (__ZN10CSoundFile7ReadFAREPKhj($this,$80,$81)|0); //@line 154 "sndfile.cpp"
                if (!($82)) {
                 $83 = HEAP32[$0>>2]|0; //@line 155 "sndfile.cpp"
                 $84 = HEAP32[$1>>2]|0; //@line 155 "sndfile.cpp"
                 $85 = (__ZN10CSoundFile7ReadAMSEPKhj($this,$83,$84)|0); //@line 155 "sndfile.cpp"
                 if (!($85)) {
                  $86 = HEAP32[$0>>2]|0; //@line 156 "sndfile.cpp"
                  $87 = HEAP32[$1>>2]|0; //@line 156 "sndfile.cpp"
                  $88 = (__ZN10CSoundFile7ReadOKTEPKhj($this,$86,$87)|0); //@line 156 "sndfile.cpp"
                  if (!($88)) {
                   $89 = HEAP32[$0>>2]|0; //@line 157 "sndfile.cpp"
                   $90 = HEAP32[$1>>2]|0; //@line 157 "sndfile.cpp"
                   $91 = (__ZN10CSoundFile7ReadPTMEPKhj($this,$89,$90)|0); //@line 157 "sndfile.cpp"
                   if (!($91)) {
                    $92 = HEAP32[$0>>2]|0; //@line 158 "sndfile.cpp"
                    $93 = HEAP32[$1>>2]|0; //@line 158 "sndfile.cpp"
                    $94 = (__ZN10CSoundFile7ReadUltEPKhj($this,$92,$93)|0); //@line 158 "sndfile.cpp"
                    if ($94) {
                     break;
                    }
                    $95 = HEAP32[$0>>2]|0; //@line 159 "sndfile.cpp"
                    $96 = HEAP32[$1>>2]|0; //@line 159 "sndfile.cpp"
                    $97 = (__ZN10CSoundFile7ReadDMFEPKhj($this,$95,$96)|0); //@line 159 "sndfile.cpp"
                    if ($97) {
                     break;
                    }
                    $98 = HEAP32[$0>>2]|0; //@line 160 "sndfile.cpp"
                    $99 = HEAP32[$1>>2]|0; //@line 160 "sndfile.cpp"
                    $100 = (__ZN10CSoundFile7ReadDSMEPKhj($this,$98,$99)|0); //@line 160 "sndfile.cpp"
                    if ($100) {
                     break;
                    }
                    $101 = HEAP32[$0>>2]|0; //@line 161 "sndfile.cpp"
                    $102 = HEAP32[$1>>2]|0; //@line 161 "sndfile.cpp"
                    $103 = (__ZN10CSoundFile7ReadUMXEPKhj($this,$101,$102)|0); //@line 161 "sndfile.cpp"
                    if ($103) {
                     break;
                    }
                    $104 = HEAP32[$0>>2]|0; //@line 162 "sndfile.cpp"
                    $105 = HEAP32[$1>>2]|0; //@line 162 "sndfile.cpp"
                    $106 = (__ZN10CSoundFile7ReadAMFEPKhj($this,$104,$105)|0); //@line 162 "sndfile.cpp"
                    if ($106) {
                     break;
                    }
                    $107 = HEAP32[$0>>2]|0; //@line 163 "sndfile.cpp"
                    $108 = HEAP32[$1>>2]|0; //@line 163 "sndfile.cpp"
                    $109 = (__ZN10CSoundFile7ReadPSMEPKhj($this,$107,$108)|0); //@line 163 "sndfile.cpp"
                    if ($109) {
                     break;
                    }
                    $110 = HEAP32[$0>>2]|0; //@line 164 "sndfile.cpp"
                    $111 = HEAP32[$1>>2]|0; //@line 164 "sndfile.cpp"
                    $112 = (__ZN10CSoundFile7ReadMT2EPKhj($this,$110,$111)|0); //@line 164 "sndfile.cpp"
                    if ($112) {
                     break;
                    }
                    $113 = HEAP32[$0>>2]|0; //@line 166 "sndfile.cpp"
                    $114 = HEAP32[$1>>2]|0; //@line 166 "sndfile.cpp"
                    $115 = (__ZN10CSoundFile7ReadModEPKhj($this,$113,$114)|0); //@line 166 "sndfile.cpp"
                    if ($115) {
                     break;
                    }
                    HEAP32[$2>>2] = 0; //@line 166 "sndfile.cpp"
                   }
                  }
                 }
                }
               }
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    }
   }
  } while(0);
  if ($40) {
   $116 = HEAP32[$0>>2]|0; //@line 170 "sndfile.cpp"
   _free($116); //@line 170 "sndfile.cpp"
   HEAP32[$0>>2] = 0; //@line 171 "sndfile.cpp"
   $i$013 = 0;
  } else {
   $i$013 = 0;
  }
 }
 while(1) {
  $117 = (((((($this)) + 59804|0) + ($i$013<<5)|0)) + 31|0); //@line 180 "sndfile.cpp"
  HEAP8[$117>>0] = 0; //@line 180 "sndfile.cpp"
  $j$01022 = 31;
  while(1) {
   $119 = ((((($this)) + 59804|0) + ($i$013<<5)|0) + ($j$01022)|0); //@line 181 "sndfile.cpp"
   $120 = (($j$01022) + -1)|0; //@line 181 "sndfile.cpp"
   HEAP8[$119>>0] = 0; //@line 181 "sndfile.cpp"
   $121 = ($j$01022|0)>(0); //@line 181 "sndfile.cpp"
   if (!($121)) {
    $$lcssa = $120;$j$01022$lcssa = $j$01022;
    break;
   }
   $$phi$trans$insert = ((((($this)) + 59804|0) + ($i$013<<5)|0) + ($120)|0);
   $$pre = HEAP8[$$phi$trans$insert>>0]|0; //@line 181 "sndfile.cpp"
   $122 = ($$pre<<24>>24)<(33); //@line 181 "sndfile.cpp"
   if ($122) {
    $j$01022 = $120;
   } else {
    $$lcssa = $120;$j$01022$lcssa = $j$01022;
    break;
   }
  }
  $118 = ($j$01022$lcssa|0)>(0); //@line 182 "sndfile.cpp"
  if ($118) {
   $j$111 = $$lcssa;
   while(1) {
    $123 = ((((($this)) + 59804|0) + ($i$013<<5)|0) + ($j$111)|0); //@line 184 "sndfile.cpp"
    $124 = HEAP8[$123>>0]|0; //@line 184 "sndfile.cpp"
    $125 = ($124&255)<(32); //@line 184 "sndfile.cpp"
    if ($125) {
     HEAP8[$123>>0] = 32; //@line 184 "sndfile.cpp"
    }
    $126 = (($j$111) + -1)|0; //@line 185 "sndfile.cpp"
    $127 = ($j$111|0)>(0); //@line 182 "sndfile.cpp"
    if ($127) {
     $j$111 = $126;
    } else {
     break;
    }
   }
  }
  $128 = (($i$013) + 1)|0; //@line 176 "sndfile.cpp"
  $exitcond19 = ($128|0)==(240); //@line 176 "sndfile.cpp"
  if ($exitcond19) {
   $i$19 = 0;
   break;
  } else {
   $i$013 = $128;
  }
 }
 while(1) {
  $129 = (((((($this)) + 49600|0) + (($i$19*36)|0)|0)) + 4|0); //@line 191 "sndfile.cpp"
  $130 = HEAP32[$129>>2]|0; //@line 191 "sndfile.cpp"
  $131 = ($130>>>0)>(64); //@line 191 "sndfile.cpp"
  if ($131) {
   HEAP32[$129>>2] = 64; //@line 191 "sndfile.cpp"
  }
  $132 = (((($this)) + 49600|0) + (($i$19*36)|0)|0); //@line 192 "sndfile.cpp"
  $133 = HEAP32[$132>>2]|0; //@line 192 "sndfile.cpp"
  $134 = ($133>>>0)>(256); //@line 192 "sndfile.cpp"
  if ($134) {
   HEAP32[$132>>2] = 128; //@line 192 "sndfile.cpp"
   $136 = 128;
  } else {
   $136 = $133;
  }
  $135 = (((($this) + ($i$19<<8)|0)) + 120|0); //@line 193 "sndfile.cpp"
  HEAP32[$135>>2] = $136; //@line 193 "sndfile.cpp"
  $137 = HEAP32[$129>>2]|0; //@line 194 "sndfile.cpp"
  $138 = (((($this) + ($i$19<<8)|0)) + 168|0); //@line 194 "sndfile.cpp"
  HEAP32[$138>>2] = $137; //@line 194 "sndfile.cpp"
  $139 = (((((($this)) + 49600|0) + (($i$19*36)|0)|0)) + 8|0); //@line 195 "sndfile.cpp"
  $140 = HEAP32[$139>>2]|0; //@line 195 "sndfile.cpp"
  $141 = (((($this) + ($i$19<<8)|0)) + 36|0); //@line 195 "sndfile.cpp"
  HEAP32[$141>>2] = $140; //@line 195 "sndfile.cpp"
  $142 = (((($this) + ($i$19<<8)|0)) + 116|0); //@line 196 "sndfile.cpp"
  HEAP32[$142>>2] = 256; //@line 196 "sndfile.cpp"
  $143 = (((($this) + ($i$19<<8)|0)) + 238|0); //@line 197 "sndfile.cpp"
  HEAP8[$143>>0] = 127; //@line 197 "sndfile.cpp"
  $144 = (($i$19) + 1)|0; //@line 189 "sndfile.cpp"
  $exitcond18 = ($144|0)==(64); //@line 189 "sndfile.cpp"
  if ($exitcond18) {
   break;
  } else {
   $i$19 = $144;
  }
 }
 $145 = ((($this)) + 33280|0); //@line 200 "sndfile.cpp"
 $i$27 = 0;$pins$08 = $145;
 while(1) {
  $146 = ((($pins$08)) + 20|0); //@line 204 "sndfile.cpp"
  $147 = HEAP32[$146>>2]|0; //@line 204 "sndfile.cpp"
  $148 = ($147|0)==(0|0); //@line 204 "sndfile.cpp"
  if ($148) {
   ;HEAP32[$pins$08>>2]=0|0;HEAP32[$pins$08+4>>2]=0|0;HEAP32[$pins$08+8>>2]=0|0;HEAP32[$pins$08+12>>2]=0|0;HEAP32[$pins$08+16>>2]=0|0; //@line 221 "sndfile.cpp"
   $235 = 0;
   label = 58;
  } else {
   $149 = ((($pins$08)) + 8|0); //@line 206 "sndfile.cpp"
   $150 = HEAP32[$149>>2]|0; //@line 206 "sndfile.cpp"
   $151 = HEAP32[$pins$08>>2]|0; //@line 206 "sndfile.cpp"
   $152 = ($150>>>0)>($151>>>0); //@line 206 "sndfile.cpp"
   if ($152) {
    HEAP32[$149>>2] = $151; //@line 206 "sndfile.cpp"
    $157 = $151;
   } else {
    $157 = $150;
   }
   $153 = ((($pins$08)) + 4|0); //@line 207 "sndfile.cpp"
   $154 = HEAP32[$153>>2]|0; //@line 207 "sndfile.cpp"
   $155 = (($154) + 3)|0; //@line 207 "sndfile.cpp"
   $156 = ($155>>>0)<($157>>>0); //@line 207 "sndfile.cpp"
   if ($156) {
    $167 = $157;
   } else {
    HEAP32[$153>>2] = 0; //@line 209 "sndfile.cpp"
    HEAP32[$149>>2] = 0; //@line 210 "sndfile.cpp"
    $167 = 0;
   }
   $158 = ((($pins$08)) + 16|0); //@line 212 "sndfile.cpp"
   $159 = HEAP32[$158>>2]|0; //@line 212 "sndfile.cpp"
   $160 = ($159>>>0)>($151>>>0); //@line 212 "sndfile.cpp"
   if ($160) {
    HEAP32[$158>>2] = $151; //@line 212 "sndfile.cpp"
    $165 = $151;
   } else {
    $165 = $159;
   }
   $161 = ((($pins$08)) + 12|0); //@line 213 "sndfile.cpp"
   $162 = HEAP32[$161>>2]|0; //@line 213 "sndfile.cpp"
   $163 = (($162) + 3)|0; //@line 213 "sndfile.cpp"
   $164 = ($163>>>0)<($165>>>0); //@line 213 "sndfile.cpp"
   if ($164) {
    $234 = $165;
   } else {
    HEAP32[$161>>2] = 0; //@line 215 "sndfile.cpp"
    HEAP32[$158>>2] = 0; //@line 216 "sndfile.cpp"
    $234 = 0;
   }
   $166 = ($167|0)==(0); //@line 226 "sndfile.cpp"
   if ($166) {
    $235 = $234;
    label = 58;
   } else {
    $174 = $234;
   }
  }
  if ((label|0) == 58) {
   label = 0;
   $168 = ((($pins$08)) + 34|0); //@line 226 "sndfile.cpp"
   $169 = HEAP16[$168>>1]|0; //@line 226 "sndfile.cpp"
   $170 = $169&65535; //@line 226 "sndfile.cpp"
   $171 = $170 & 65533; //@line 226 "sndfile.cpp"
   $172 = $171&65535; //@line 226 "sndfile.cpp"
   HEAP16[$168>>1] = $172; //@line 226 "sndfile.cpp"
   $174 = $235;
  }
  $173 = ($174|0)==(0); //@line 227 "sndfile.cpp"
  if ($173) {
   $175 = ((($pins$08)) + 34|0); //@line 227 "sndfile.cpp"
   $176 = HEAP16[$175>>1]|0; //@line 227 "sndfile.cpp"
   $177 = $176&65535; //@line 227 "sndfile.cpp"
   $178 = $177 & 65527; //@line 227 "sndfile.cpp"
   $179 = $178&65535; //@line 227 "sndfile.cpp"
   HEAP16[$175>>1] = $179; //@line 227 "sndfile.cpp"
  }
  $180 = ((($pins$08)) + 32|0); //@line 228 "sndfile.cpp"
  $181 = HEAP16[$180>>1]|0; //@line 228 "sndfile.cpp"
  $182 = ($181&65535)>(64); //@line 228 "sndfile.cpp"
  if ($182) {
   HEAP16[$180>>1] = 64; //@line 228 "sndfile.cpp"
  }
  $183 = (($i$27) + 1)|0; //@line 202 "sndfile.cpp"
  $184 = ((($pins$08)) + 64|0); //@line 202 "sndfile.cpp"
  $exitcond17 = ($183|0)==(240); //@line 202 "sndfile.cpp"
  if ($exitcond17) {
   break;
  } else {
   $i$27 = $183;$pins$08 = $184;
  }
 }
 $$pr = HEAP32[$7>>2]|0; //@line 231 "sndfile.cpp"
 $185 = ($$pr|0)==(0); //@line 231 "sndfile.cpp"
 L85: do {
  if (!($185)) {
   $187 = $$pr;
   while(1) {
    $186 = (((($this)) + 48640|0) + ($187<<2)|0); //@line 231 "sndfile.cpp"
    $188 = HEAP32[$186>>2]|0; //@line 231 "sndfile.cpp"
    $189 = ($188|0)==(0|0); //@line 231 "sndfile.cpp"
    if (!($189)) {
     break L85;
    }
    $190 = (($187) + -1)|0; //@line 232 "sndfile.cpp"
    HEAP32[$7>>2] = $190; //@line 232 "sndfile.cpp"
    $191 = ($190|0)==(0); //@line 231 "sndfile.cpp"
    if ($191) {
     break;
    } else {
     $187 = $190;
    }
   }
  }
 } while(0);
 $192 = HEAP32[$26>>2]|0; //@line 234 "sndfile.cpp"
 $193 = ($192>>>0)<(32); //@line 234 "sndfile.cpp"
 if ($193) {
  HEAP32[$26>>2] = 32; //@line 234 "sndfile.cpp"
 }
 $194 = HEAP32[$15>>2]|0; //@line 235 "sndfile.cpp"
 $195 = ($194>>>0)<(32); //@line 235 "sndfile.cpp"
 if ($195) {
  HEAP32[$15>>2] = 125; //@line 235 "sndfile.cpp"
  $201 = 125;
 } else {
  $201 = $194;
 }
 $196 = HEAP32[$14>>2]|0; //@line 236 "sndfile.cpp"
 $197 = ($196|0)==(0); //@line 236 "sndfile.cpp"
 if ($197) {
  HEAP32[$14>>2] = 6; //@line 236 "sndfile.cpp"
  $199 = 6;
 } else {
  $199 = $196;
 }
 $198 = ((($this)) + 59708|0); //@line 237 "sndfile.cpp"
 HEAP32[$198>>2] = $199; //@line 237 "sndfile.cpp"
 $200 = ((($this)) + 59712|0); //@line 238 "sndfile.cpp"
 HEAP32[$200>>2] = $201; //@line 238 "sndfile.cpp"
 $202 = HEAP32[$11>>2]|0; //@line 239 "sndfile.cpp"
 HEAP32[$12>>2] = $202; //@line 239 "sndfile.cpp"
 HEAP32[$22>>2] = 0; //@line 240 "sndfile.cpp"
 HEAP32[$21>>2] = 0; //@line 241 "sndfile.cpp"
 HEAP32[$20>>2] = 0; //@line 242 "sndfile.cpp"
 $203 = ((($this)) + 59676|0); //@line 243 "sndfile.cpp"
 HEAP32[$203>>2] = 0; //@line 243 "sndfile.cpp"
 $204 = ((($this)) + 59692|0); //@line 244 "sndfile.cpp"
 HEAP32[$204>>2] = $199; //@line 244 "sndfile.cpp"
 HEAP32[$18>>2] = 0; //@line 245 "sndfile.cpp"
 HEAP32[$19>>2] = 0; //@line 246 "sndfile.cpp"
 $205 = HEAP32[$23>>2]|0; //@line 247 "sndfile.cpp"
 $206 = ($205>>>0)>(255); //@line 247 "sndfile.cpp"
 if ($206) {
  label = 75;
 } else {
  $207 = (((($this)) + 53344|0) + ($205)|0); //@line 247 "sndfile.cpp"
  $208 = HEAP8[$207>>0]|0; //@line 247 "sndfile.cpp"
  $209 = ($208&255)>(239); //@line 247 "sndfile.cpp"
  if ($209) {
   label = 75;
  }
 }
 if ((label|0) == 75) {
  HEAP32[$23>>2] = 0; //@line 247 "sndfile.cpp"
 }
 $210 = HEAP32[183084]|0; //@line 249 "sndfile.cpp"
 $211 = ($210|0)==(0|0); //@line 249 "sndfile.cpp"
 if (!($211)) {
  $iPlug$05 = 0;
  while(1) {
   $212 = (((($this)) + 58496|0) + (($iPlug$05*144)|0)|0); //@line 253 "sndfile.cpp"
   $213 = (((((($this)) + 58496|0) + (($iPlug$05*144)|0)|0)) + 16|0); //@line 253 "sndfile.cpp"
   $214 = HEAP32[$213>>2]|0; //@line 253 "sndfile.cpp"
   $215 = ($214|0)==(0); //@line 253 "sndfile.cpp"
   if ($215) {
    $216 = (((((($this)) + 58496|0) + (($iPlug$05*144)|0)|0)) + 20|0); //@line 254 "sndfile.cpp"
    $217 = HEAP32[$216>>2]|0; //@line 254 "sndfile.cpp"
    $218 = ($217|0)==(0); //@line 254 "sndfile.cpp"
    if (!($218)) {
     label = 79;
    }
   } else {
    label = 79;
   }
   if ((label|0) == 79) {
    label = 0;
    $219 = HEAP32[183084]|0; //@line 256 "sndfile.cpp"
    (FUNCTION_TABLE_ii[$219 & 8191]($212)|0); //@line 256 "sndfile.cpp"
    $220 = HEAP32[$212>>2]|0; //@line 257 "sndfile.cpp"
    $221 = ($220|0)==(0|0); //@line 257 "sndfile.cpp"
    if (!($221)) {
     $222 = HEAP32[$220>>2]|0; //@line 259 "sndfile.cpp"
     $223 = ((($222)) + 20|0); //@line 259 "sndfile.cpp"
     $224 = HEAP32[$223>>2]|0; //@line 259 "sndfile.cpp"
     FUNCTION_TABLE_vi[$224 & 8191]($220); //@line 259 "sndfile.cpp"
    }
   }
   $225 = (($iPlug$05) + 1)|0; //@line 251 "sndfile.cpp"
   $exitcond = ($225|0)==(8); //@line 251 "sndfile.cpp"
   if ($exitcond) {
    break;
   } else {
    $iPlug$05 = $225;
   }
  }
 }
 $226 = HEAP32[$2>>2]|0; //@line 264 "sndfile.cpp"
 $227 = ($226|0)==(0); //@line 264 "sndfile.cpp"
 if ($227) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 272 "sndfile.cpp"
 }
 $228 = HEAP32[$4>>2]|0; //@line 266 "sndfile.cpp"
 $229 = $228 << 3; //@line 266 "sndfile.cpp"
 $230 = (($229) + 16)|0; //@line 266 "sndfile.cpp"
 $231 = ($230>>>0)>(100); //@line 267 "sndfile.cpp"
 $$ = $231 ? 100 : $230; //@line 267 "sndfile.cpp"
 $232 = HEAP32[$26>>2]|0; //@line 268 "sndfile.cpp"
 $233 = ($232>>>0)>($$>>>0); //@line 268 "sndfile.cpp"
 if (!($233)) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 272 "sndfile.cpp"
 }
 HEAP32[$26>>2] = $$; //@line 268 "sndfile.cpp"
 $$0 = 1;
 STACKTOP = sp;return ($$0|0); //@line 272 "sndfile.cpp"
}
function __ZN10CSoundFile12ResetMidiCfgEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer10 = 0, $vararg_buffer13 = 0, $vararg_buffer16 = 0, $vararg_buffer19 = 0, $vararg_buffer22 = 0, $vararg_buffer25 = 0, $vararg_buffer28 = 0, $vararg_buffer31 = 0, $vararg_buffer34 = 0, $vararg_buffer37 = 0, $vararg_buffer4 = 0, $vararg_buffer40 = 0;
 var $vararg_buffer43 = 0, $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer43 = sp + 120|0;
 $vararg_buffer40 = sp + 112|0;
 $vararg_buffer37 = sp + 104|0;
 $vararg_buffer34 = sp + 96|0;
 $vararg_buffer31 = sp + 88|0;
 $vararg_buffer28 = sp + 80|0;
 $vararg_buffer25 = sp + 72|0;
 $vararg_buffer22 = sp + 64|0;
 $vararg_buffer19 = sp + 56|0;
 $vararg_buffer16 = sp + 48|0;
 $vararg_buffer13 = sp + 40|0;
 $vararg_buffer10 = sp + 32|0;
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($this)) + 53600|0); //@line 378 "sndfile.cpp"
 _memset(($0|0),0,4896)|0; //@line 378 "sndfile.cpp"
 ;HEAP8[$0>>0]=HEAP8[524239>>0]|0;HEAP8[$0+1>>0]=HEAP8[524239+1>>0]|0;HEAP8[$0+2>>0]=HEAP8[524239+2>>0]|0; //@line 379 "sndfile.cpp"
 $1 = ((($this)) + 53632|0); //@line 380 "sndfile.cpp"
 ;HEAP8[$1>>0]=HEAP8[524242>>0]|0;HEAP8[$1+1>>0]=HEAP8[524242+1>>0]|0;HEAP8[$1+2>>0]=HEAP8[524242+2>>0]|0; //@line 380 "sndfile.cpp"
 $2 = ((($this)) + 53696|0); //@line 381 "sndfile.cpp"
 ;HEAP8[$2>>0]=HEAP8[524245>>0]|0;HEAP8[$2+1>>0]=HEAP8[524245+1>>0]|0;HEAP8[$2+2>>0]=HEAP8[524245+2>>0]|0;HEAP8[$2+3>>0]=HEAP8[524245+3>>0]|0;HEAP8[$2+4>>0]=HEAP8[524245+4>>0]|0;HEAP8[$2+5>>0]=HEAP8[524245+5>>0]|0;HEAP8[$2+6>>0]=HEAP8[524245+6>>0]|0; //@line 381 "sndfile.cpp"
 $3 = ((($this)) + 53728|0); //@line 382 "sndfile.cpp"
 ;HEAP8[$3>>0]=HEAP8[524252>>0]|0;HEAP8[$3+1>>0]=HEAP8[524252+1>>0]|0;HEAP8[$3+2>>0]=HEAP8[524252+2>>0]|0;HEAP8[$3+3>>0]=HEAP8[524252+3>>0]|0;HEAP8[$3+4>>0]=HEAP8[524252+4>>0]|0;HEAP8[$3+5>>0]=HEAP8[524252+5>>0]|0;HEAP8[$3+6>>0]=HEAP8[524252+6>>0]|0; //@line 382 "sndfile.cpp"
 $4 = ((($this)) + 53856|0); //@line 383 "sndfile.cpp"
 ;HEAP8[$4>>0]=HEAP8[524259>>0]|0;HEAP8[$4+1>>0]=HEAP8[524259+1>>0]|0;HEAP8[$4+2>>0]=HEAP8[524259+2>>0]|0;HEAP8[$4+3>>0]=HEAP8[524259+3>>0]|0;HEAP8[$4+4>>0]=HEAP8[524259+4>>0]|0; //@line 383 "sndfile.cpp"
 $5 = ((($this)) + 53888|0); //@line 384 "sndfile.cpp"
 $6 = $5; //@line 384 "sndfile.cpp"
 $7 = $6; //@line 384 "sndfile.cpp"
 HEAP8[$7>>0]=809906246&255;HEAP8[$7+1>>0]=(809906246>>8)&255;HEAP8[$7+2>>0]=(809906246>>16)&255;HEAP8[$7+3>>0]=809906246>>24; //@line 384 "sndfile.cpp"
 $8 = (($6) + 4)|0; //@line 384 "sndfile.cpp"
 $9 = $8; //@line 384 "sndfile.cpp"
 HEAP8[$9>>0]=8007728&255;HEAP8[$9+1>>0]=(8007728>>8)&255;HEAP8[$9+2>>0]=(8007728>>16)&255;HEAP8[$9+3>>0]=8007728>>24; //@line 384 "sndfile.cpp"
 $10 = ((($this)) + 54400|0); //@line 385 "sndfile.cpp"
 HEAP32[$vararg_buffer>>2] = 0; //@line 385 "sndfile.cpp"
 (_sprintf($10,524264,$vararg_buffer)|0); //@line 385 "sndfile.cpp"
 $11 = ((($this)) + 54432|0); //@line 385 "sndfile.cpp"
 HEAP32[$vararg_buffer1>>2] = 8; //@line 385 "sndfile.cpp"
 (_sprintf($11,524264,$vararg_buffer1)|0); //@line 385 "sndfile.cpp"
 $12 = ((($this)) + 54464|0); //@line 385 "sndfile.cpp"
 HEAP32[$vararg_buffer4>>2] = 16; //@line 385 "sndfile.cpp"
 (_sprintf($12,524264,$vararg_buffer4)|0); //@line 385 "sndfile.cpp"
 $13 = ((($this)) + 54496|0); //@line 385 "sndfile.cpp"
 HEAP32[$vararg_buffer7>>2] = 24; //@line 385 "sndfile.cpp"
 (_sprintf($13,524264,$vararg_buffer7)|0); //@line 385 "sndfile.cpp"
 $14 = ((($this)) + 54528|0); //@line 385 "sndfile.cpp"
 HEAP32[$vararg_buffer10>>2] = 32; //@line 385 "sndfile.cpp"
 (_sprintf($14,524264,$vararg_buffer10)|0); //@line 385 "sndfile.cpp"
 $15 = ((($this)) + 54560|0); //@line 385 "sndfile.cpp"
 HEAP32[$vararg_buffer13>>2] = 40; //@line 385 "sndfile.cpp"
 (_sprintf($15,524264,$vararg_buffer13)|0); //@line 385 "sndfile.cpp"
 $16 = ((($this)) + 54592|0); //@line 385 "sndfile.cpp"
 HEAP32[$vararg_buffer16>>2] = 48; //@line 385 "sndfile.cpp"
 (_sprintf($16,524264,$vararg_buffer16)|0); //@line 385 "sndfile.cpp"
 $17 = ((($this)) + 54624|0); //@line 385 "sndfile.cpp"
 HEAP32[$vararg_buffer19>>2] = 56; //@line 385 "sndfile.cpp"
 (_sprintf($17,524264,$vararg_buffer19)|0); //@line 385 "sndfile.cpp"
 $18 = ((($this)) + 54656|0); //@line 385 "sndfile.cpp"
 HEAP32[$vararg_buffer22>>2] = 64; //@line 385 "sndfile.cpp"
 (_sprintf($18,524264,$vararg_buffer22)|0); //@line 385 "sndfile.cpp"
 $19 = ((($this)) + 54688|0); //@line 385 "sndfile.cpp"
 HEAP32[$vararg_buffer25>>2] = 72; //@line 385 "sndfile.cpp"
 (_sprintf($19,524264,$vararg_buffer25)|0); //@line 385 "sndfile.cpp"
 $20 = ((($this)) + 54720|0); //@line 385 "sndfile.cpp"
 HEAP32[$vararg_buffer28>>2] = 80; //@line 385 "sndfile.cpp"
 (_sprintf($20,524264,$vararg_buffer28)|0); //@line 385 "sndfile.cpp"
 $21 = ((($this)) + 54752|0); //@line 385 "sndfile.cpp"
 HEAP32[$vararg_buffer31>>2] = 88; //@line 385 "sndfile.cpp"
 (_sprintf($21,524264,$vararg_buffer31)|0); //@line 385 "sndfile.cpp"
 $22 = ((($this)) + 54784|0); //@line 385 "sndfile.cpp"
 HEAP32[$vararg_buffer34>>2] = 96; //@line 385 "sndfile.cpp"
 (_sprintf($22,524264,$vararg_buffer34)|0); //@line 385 "sndfile.cpp"
 $23 = ((($this)) + 54816|0); //@line 385 "sndfile.cpp"
 HEAP32[$vararg_buffer37>>2] = 104; //@line 385 "sndfile.cpp"
 (_sprintf($23,524264,$vararg_buffer37)|0); //@line 385 "sndfile.cpp"
 $24 = ((($this)) + 54848|0); //@line 385 "sndfile.cpp"
 HEAP32[$vararg_buffer40>>2] = 112; //@line 385 "sndfile.cpp"
 (_sprintf($24,524264,$vararg_buffer40)|0); //@line 385 "sndfile.cpp"
 $25 = ((($this)) + 54880|0); //@line 385 "sndfile.cpp"
 HEAP32[$vararg_buffer43>>2] = 120; //@line 385 "sndfile.cpp"
 (_sprintf($25,524264,$vararg_buffer43)|0); //@line 385 "sndfile.cpp"
 STACKTOP = sp;return; //@line 386 "sndfile.cpp"
}
function __Z12MMCMP_UnpackPPKhPj($ppMemFile,$pdwMemLength) {
 $ppMemFile = $ppMemFile|0;
 $pdwMemLength = $pdwMemLength|0;
 var $$ = 0, $$0 = 0, $$0$i19 = 0, $$0$i29 = 0, $$0$i62 = 0, $$0$i73 = 0, $$18 = 0, $$lcssa = 0, $$lcssa255 = 0, $$lcssa256 = 0, $$lcssa257 = 0, $$lcssa258 = 0, $$lcssa259 = 0, $$lcssa260 = 0, $$lcssa261 = 0, $$lcssa262 = 0, $$lcssa263 = 0, $$lcssa264 = 0, $$lcssa265 = 0, $$lcssa266 = 0;
 var $$lcssa267 = 0, $$lcssa268 = 0, $$lcssa269 = 0, $$lcssa270 = 0, $$lcssa271 = 0, $$lcssa272 = 0, $$lcssa273 = 0, $$lcssa274 = 0, $$lcssa275 = 0, $$off = 0, $$off$i = 0, $$ph = 0, $$ph111 = 0, $$ph112 = 0, $$ph115 = 0, $$ph116 = 0, $$ph117 = 0, $$pn = 0, $$sum = 0, $$sum10 = 0;
 var $$sum156 = 0, $$sum2 = 0, $$sum3 = 0, $$sum4 = 0, $$sum5 = 0, $$sum6 = 0, $$sum7 = 0, $$sum8 = 0, $$sum9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0;
 var $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0;
 var $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0;
 var $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0;
 var $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0;
 var $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0;
 var $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0;
 var $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0;
 var $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0;
 var $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0;
 var $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0;
 var $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0;
 var $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0;
 var $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0;
 var $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dwMemPos$0149 = 0, $dwPos$0 = 0, $dwPos$1 = 0, $dwPos4$0 = 0, $dwPos4$1 = 0;
 var $dwSize$0$ph = 0, $dwSize$0$ph$in = 0, $dwSize$0$ph$in$in = 0, $dwSize3$0$ph = 0, $dwSize3$0$ph$in = 0, $i$0147 = 0, $nBlock$0152 = 0, $newval$0 = 0, $newval$099102 = 0, $newval$1 = 0, $newval8$0 = 0, $newval8$0108 = 0, $numbits$0 = 0, $numbits$0$ph = 0, $numbits$197 = 0, $numbits$197$lcssa = 0, $numbits5$0 = 0, $numbits5$0$ph = 0, $numbits5$1106 = 0, $numbits5$1106$lcssa = 0;
 var $oldval$0 = 0, $oldval$0$ph = 0, $oldval$1 = 0, $oldval$2 = 0, $oldval$2$lcssa = 0, $oldval7$0 = 0, $oldval7$0$ = 0, $oldval7$0$ph = 0, $oldval7$2 = 0, $oldval7$2$lcssa = 0, $or$cond = 0, $or$cond$i = 0, $or$cond5$i = 0, $pDest$0$ph$in = 0, $psubblk$0148 = 0, $subblk$0$ph = 0, $subblk6$0$ph = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$pdwMemLength>>2]|0; //@line 112 "mmcmp.cpp"
 $1 = HEAP32[$ppMemFile>>2]|0; //@line 113 "mmcmp.cpp"
 $2 = ($1|0)==(0|0); //@line 392 "mmcmp.cpp"
 $3 = ($0>>>0)<(256); //@line 392 "mmcmp.cpp"
 $or$cond$i = $3 | $2; //@line 392 "mmcmp.cpp"
 if (!($or$cond$i)) {
  $4 = HEAP32[$1>>2]|0; //@line 392 "mmcmp.cpp"
  $5 = ($4|0)==(808603728); //@line 392 "mmcmp.cpp"
  if ($5) {
   $6 = (($0) + -4)|0; //@line 393 "mmcmp.cpp"
   $7 = (($1) + ($6)|0); //@line 393 "mmcmp.cpp"
   $8 = HEAP8[$7>>0]|0; //@line 393 "mmcmp.cpp"
   $9 = $8&255; //@line 393 "mmcmp.cpp"
   $10 = $9 << 16; //@line 393 "mmcmp.cpp"
   $11 = (($0) + -3)|0; //@line 393 "mmcmp.cpp"
   $12 = (($1) + ($11)|0); //@line 393 "mmcmp.cpp"
   $13 = HEAP8[$12>>0]|0; //@line 393 "mmcmp.cpp"
   $14 = $13&255; //@line 393 "mmcmp.cpp"
   $15 = $14 << 8; //@line 393 "mmcmp.cpp"
   $16 = $15 | $10; //@line 393 "mmcmp.cpp"
   $17 = (($0) + -2)|0; //@line 393 "mmcmp.cpp"
   $18 = (($1) + ($17)|0); //@line 393 "mmcmp.cpp"
   $19 = HEAP8[$18>>0]|0; //@line 393 "mmcmp.cpp"
   $20 = $19&255; //@line 393 "mmcmp.cpp"
   $21 = $16 | $20; //@line 393 "mmcmp.cpp"
   $$off$i = (($21) + -512)|0; //@line 395 "mmcmp.cpp"
   $22 = ($$off$i>>>0)>(4193792); //@line 395 "mmcmp.cpp"
   $23 = $0 << 4; //@line 120 "mmcmp.cpp"
   $24 = ($21>>>0)>($23>>>0); //@line 395 "mmcmp.cpp"
   $or$cond5$i = $24 | $22; //@line 395 "mmcmp.cpp"
   if (!($or$cond5$i)) {
    $25 = (($21) + 31)|0; //@line 396 "mmcmp.cpp"
    $26 = $25 & 33554416; //@line 396 "mmcmp.cpp"
    $27 = (_malloc($26)|0); //@line 101 "../src/libmodplug/stdafx.h"
    $28 = ($27|0)==(0|0); //@line 103 "../src/libmodplug/stdafx.h"
    if (!($28)) {
     _memset(($27|0),0,($26|0))|0; //@line 103 "../src/libmodplug/stdafx.h"
     $29 = ((($1)) + 4|0); //@line 397 "mmcmp.cpp"
     __Z13PP20_DoUnpackPKhjPhj($29,$6,$27,$21); //@line 397 "mmcmp.cpp"
     HEAP32[$ppMemFile>>2] = $27; //@line 398 "mmcmp.cpp"
     HEAP32[$pdwMemLength>>2] = $21; //@line 399 "mmcmp.cpp"
     $$0 = 1;
     return ($$0|0); //@line 291 "mmcmp.cpp"
    }
   }
  }
 }
 $30 = ($0>>>0)>(255); //@line 124 "mmcmp.cpp"
 $31 = ($1|0)!=(0|0); //@line 124 "mmcmp.cpp"
 $or$cond = $30 & $31; //@line 124 "mmcmp.cpp"
 if (!($or$cond)) {
  $$0 = 0;
  return ($$0|0); //@line 291 "mmcmp.cpp"
 }
 $32 = HEAP32[$1>>2]|0; //@line 124 "mmcmp.cpp"
 $33 = ($32|0)==(1129474426); //@line 124 "mmcmp.cpp"
 if (!($33)) {
  $$0 = 0;
  return ($$0|0); //@line 291 "mmcmp.cpp"
 }
 $34 = ((($1)) + 4|0); //@line 124 "mmcmp.cpp"
 $35 = HEAP32[$34>>2]|0; //@line 124 "mmcmp.cpp"
 $36 = ($35|0)==(1634291279); //@line 124 "mmcmp.cpp"
 if (!($36)) {
  $$0 = 0;
  return ($$0|0); //@line 291 "mmcmp.cpp"
 }
 $37 = ((($1)) + 8|0); //@line 124 "mmcmp.cpp"
 $38 = HEAP16[$37>>1]|0; //@line 124 "mmcmp.cpp"
 $39 = ($38&65535)<(14); //@line 124 "mmcmp.cpp"
 if ($39) {
  $$0 = 0;
  return ($$0|0); //@line 291 "mmcmp.cpp"
 }
 $40 = ((($1)) + 12|0); //@line 125 "mmcmp.cpp"
 $41 = HEAP16[$40>>1]|0; //@line 125 "mmcmp.cpp"
 $42 = ($41<<16>>16)==(0); //@line 125 "mmcmp.cpp"
 if ($42) {
  $$0 = 0;
  return ($$0|0); //@line 291 "mmcmp.cpp"
 }
 $43 = ((($1)) + 14|0); //@line 125 "mmcmp.cpp"
 $44 = HEAP32[$43>>2]|0; //@line 125 "mmcmp.cpp"
 $$off = (($44) + -16)|0; //@line 125 "mmcmp.cpp"
 $45 = ($$off>>>0)>(134217712); //@line 125 "mmcmp.cpp"
 if ($45) {
  $$0 = 0;
  return ($$0|0); //@line 291 "mmcmp.cpp"
 }
 $46 = ((($1)) + 18|0); //@line 126 "mmcmp.cpp"
 $47 = HEAP32[$46>>2]|0; //@line 126 "mmcmp.cpp"
 $48 = ($47>>>0)<($0>>>0); //@line 126 "mmcmp.cpp"
 if (!($48)) {
  $$0 = 0;
  return ($$0|0); //@line 291 "mmcmp.cpp"
 }
 $49 = $41&65535; //@line 126 "mmcmp.cpp"
 $50 = $49 << 2; //@line 126 "mmcmp.cpp"
 $51 = (($47) + ($50))|0; //@line 126 "mmcmp.cpp"
 $52 = ($51>>>0)>($0>>>0); //@line 126 "mmcmp.cpp"
 if ($52) {
  $$0 = 0;
  return ($$0|0); //@line 291 "mmcmp.cpp"
 }
 $53 = (($44) + 31)|0; //@line 128 "mmcmp.cpp"
 $54 = $53 & -16; //@line 128 "mmcmp.cpp"
 $55 = (_malloc($54)|0); //@line 101 "../src/libmodplug/stdafx.h"
 $56 = ($55|0)==(0|0); //@line 103 "../src/libmodplug/stdafx.h"
 if ($56) {
  $$0 = 0;
  return ($$0|0); //@line 291 "mmcmp.cpp"
 }
 _memset(($55|0),0,($54|0))|0; //@line 103 "../src/libmodplug/stdafx.h"
 $57 = HEAP32[$46>>2]|0; //@line 129 "mmcmp.cpp"
 $58 = (($1) + ($57)|0); //@line 129 "mmcmp.cpp"
 $59 = HEAP16[$40>>1]|0; //@line 130 "mmcmp.cpp"
 $60 = ($59<<16>>16)==(0); //@line 130 "mmcmp.cpp"
 L35: do {
  if (!($60)) {
   $nBlock$0152 = 0;
   while(1) {
    $61 = (($58) + ($nBlock$0152<<2)|0); //@line 132 "mmcmp.cpp"
    $62 = HEAP32[$61>>2]|0; //@line 132 "mmcmp.cpp"
    $$sum = (($62) + 20)|0; //@line 134 "mmcmp.cpp"
    $63 = (($1) + ($$sum)|0); //@line 134 "mmcmp.cpp"
    $64 = ($$sum>>>0)<($0>>>0); //@line 136 "mmcmp.cpp"
    if (!($64)) {
     break L35;
    }
    $$sum2 = (($62) + 12)|0; //@line 136 "mmcmp.cpp"
    $65 = (($1) + ($$sum2)|0); //@line 136 "mmcmp.cpp"
    $66 = HEAP16[$65>>1]|0; //@line 136 "mmcmp.cpp"
    $67 = $66&65535; //@line 136 "mmcmp.cpp"
    $68 = $67 << 3; //@line 136 "mmcmp.cpp"
    $69 = (($68) + ($$sum))|0; //@line 136 "mmcmp.cpp"
    $70 = ($69>>>0)<($0>>>0); //@line 136 "mmcmp.cpp"
    if (!($70)) {
     break L35;
    }
    $$sum3 = (($62) + 14)|0; //@line 144 "mmcmp.cpp"
    $71 = (($1) + ($$sum3)|0); //@line 144 "mmcmp.cpp"
    $72 = HEAP16[$71>>1]|0; //@line 144 "mmcmp.cpp"
    $73 = $72&65535; //@line 144 "mmcmp.cpp"
    $74 = $73 & 1; //@line 144 "mmcmp.cpp"
    $75 = ($74|0)==(0); //@line 144 "mmcmp.cpp"
    L40: do {
     if ($75) {
      $76 = ($66<<16>>16)==(0); //@line 146 "mmcmp.cpp"
      if (!($76)) {
       $dwMemPos$0149 = $69;$i$0147 = 0;$psubblk$0148 = $63;
       while(1) {
        $77 = HEAP32[$psubblk$0148>>2]|0; //@line 148 "mmcmp.cpp"
        $78 = ($77>>>0)>($44>>>0); //@line 148 "mmcmp.cpp"
        if ($78) {
         break L40;
        }
        $79 = ((($psubblk$0148)) + 4|0); //@line 148 "mmcmp.cpp"
        $80 = HEAP32[$79>>2]|0; //@line 148 "mmcmp.cpp"
        $81 = (($80) + ($77))|0; //@line 148 "mmcmp.cpp"
        $82 = ($81>>>0)>($44>>>0); //@line 148 "mmcmp.cpp"
        if ($82) {
         break L40;
        }
        $83 = (($55) + ($77)|0); //@line 152 "mmcmp.cpp"
        $84 = (($1) + ($dwMemPos$0149)|0); //@line 152 "mmcmp.cpp"
        _memcpy(($83|0),($84|0),($80|0))|0; //@line 152 "mmcmp.cpp"
        $85 = HEAP32[$79>>2]|0; //@line 153 "mmcmp.cpp"
        $86 = (($85) + ($dwMemPos$0149))|0; //@line 153 "mmcmp.cpp"
        $87 = ((($psubblk$0148)) + 8|0); //@line 154 "mmcmp.cpp"
        $88 = (($i$0147) + 1)|0; //@line 146 "mmcmp.cpp"
        $89 = HEAP16[$65>>1]|0; //@line 146 "mmcmp.cpp"
        $90 = $89&65535; //@line 146 "mmcmp.cpp"
        $91 = ($88>>>0)<($90>>>0); //@line 146 "mmcmp.cpp"
        if ($91) {
         $dwMemPos$0149 = $86;$i$0147 = $88;$psubblk$0148 = $87;
        } else {
         break;
        }
       }
      }
     } else {
      $92 = $73 & 4; //@line 158 "mmcmp.cpp"
      $93 = ($92|0)==(0); //@line 158 "mmcmp.cpp"
      $94 = HEAP32[$63>>2]|0; //@line 161 "mmcmp.cpp"
      $$sum4 = (($62) + 24)|0; //@line 231 "mmcmp.cpp"
      $95 = (($1) + ($$sum4)|0); //@line 231 "mmcmp.cpp"
      $$sum5 = (($62) + 18)|0; //@line 233 "mmcmp.cpp"
      $96 = (($1) + ($$sum5)|0); //@line 233 "mmcmp.cpp"
      $97 = HEAP16[$96>>1]|0; //@line 233 "mmcmp.cpp"
      $98 = $97&65535; //@line 233 "mmcmp.cpp"
      $$sum6 = (($62) + 16)|0; //@line 239 "mmcmp.cpp"
      $99 = (($1) + ($$sum6)|0); //@line 239 "mmcmp.cpp"
      $100 = HEAP16[$99>>1]|0; //@line 239 "mmcmp.cpp"
      $101 = $100&65535; //@line 239 "mmcmp.cpp"
      $$sum7 = (($101) + ($69))|0; //@line 239 "mmcmp.cpp"
      $102 = (($1) + ($$sum7)|0); //@line 239 "mmcmp.cpp"
      $$sum8 = (($62) + 4)|0; //@line 240 "mmcmp.cpp"
      $103 = (($1) + ($$sum8)|0); //@line 240 "mmcmp.cpp"
      $104 = HEAP32[$103>>2]|0; //@line 240 "mmcmp.cpp"
      $$sum9 = (($104) + ($69))|0; //@line 240 "mmcmp.cpp"
      $105 = (($1) + ($$sum9)|0); //@line 240 "mmcmp.cpp"
      if ($93) {
       $$ph = 0;$$ph111 = $102;$$ph112 = 0;$370 = $94;$dwSize3$0$ph$in = $95;$numbits5$0$ph = $98;$oldval7$0$ph = 0;$subblk6$0$ph = 0;
       while(1) {
        $dwSize3$0$ph = HEAP32[$dwSize3$0$ph$in>>2]|0; //@line 282 "mmcmp.cpp"
        $251 = $$ph112;$415 = $$ph;$416 = $$ph111;$dwPos4$0 = 0;$numbits5$0 = $numbits5$0$ph;$oldval7$0 = $oldval7$0$ph;
        while(1) {
         $245 = HEAP16[$65>>1]|0; //@line 241 "mmcmp.cpp"
         $246 = $245&65535; //@line 241 "mmcmp.cpp"
         $247 = ($subblk6$0$ph>>>0)<($246>>>0); //@line 241 "mmcmp.cpp"
         if (!($247)) {
          break L40;
         }
         $248 = (($numbits5$0) + 1)|0; //@line 244 "mmcmp.cpp"
         $249 = ($248|0)==(0); //@line 68 "mmcmp.cpp"
         if ($249) {
          $$0$i73 = 0;$283 = $251;$417 = $415;$418 = $416;
         } else {
          $250 = ($251>>>0)<(24); //@line 69 "mmcmp.cpp"
          if ($250) {
           $252 = (23 - ($251))|0; //@line 71 "mmcmp.cpp"
           $253 = $252 & -8; //@line 71 "mmcmp.cpp"
           $255 = $416;$261 = $251;$263 = $415;$420 = $416;
           while(1) {
            $254 = ($255>>>0)<($105>>>0); //@line 71 "mmcmp.cpp"
            if ($254) {
             $256 = ((($255)) + 1|0); //@line 71 "mmcmp.cpp"
             $257 = HEAP8[$255>>0]|0; //@line 71 "mmcmp.cpp"
             $258 = $257&255; //@line 71 "mmcmp.cpp"
             $260 = $258;$421 = $256;$422 = $256;
            } else {
             $260 = 0;$421 = $420;$422 = $255;
            }
            $259 = $260 << $261; //@line 71 "mmcmp.cpp"
            $262 = $259 | $263; //@line 71 "mmcmp.cpp"
            $264 = (($261) + 8)|0; //@line 72 "mmcmp.cpp"
            $265 = ($264>>>0)<(24); //@line 69 "mmcmp.cpp"
            if ($265) {
             $255 = $422;$261 = $264;$263 = $262;$420 = $421;
            } else {
             $$lcssa265 = $421;$$lcssa266 = $262;
             break;
            }
           }
           $266 = (($251) + 8)|0; //@line 71 "mmcmp.cpp"
           $267 = (($266) + ($253))|0; //@line 71 "mmcmp.cpp"
           $271 = $$lcssa266;$274 = $267;$419 = $$lcssa265;
          } else {
           $271 = $415;$274 = $251;$419 = $416;
          }
          $268 = 1 << $248; //@line 74 "mmcmp.cpp"
          $269 = (($268) + -1)|0; //@line 74 "mmcmp.cpp"
          $270 = $271 & $269; //@line 74 "mmcmp.cpp"
          $272 = $271 >>> $248; //@line 75 "mmcmp.cpp"
          $273 = (($274) - ($248))|0; //@line 76 "mmcmp.cpp"
          $$0$i73 = $270;$283 = $273;$417 = $272;$418 = $419;
         }
         $275 = (46548 + ($numbits5$0<<2)|0); //@line 246 "mmcmp.cpp"
         $276 = HEAP32[$275>>2]|0; //@line 246 "mmcmp.cpp"
         $277 = ($$0$i73>>>0)<($276>>>0); //@line 246 "mmcmp.cpp"
         do {
          if ($277) {
           $423 = $417;$424 = $418;$425 = $283;$newval8$0 = $$0$i73;
           label = 106;
          } else {
           $278 = (46580 + ($numbits5$0<<2)|0); //@line 248 "mmcmp.cpp"
           $279 = HEAP32[$278>>2]|0; //@line 248 "mmcmp.cpp"
           $280 = $numbits5$0 & -2; //@line 68 "mmcmp.cpp"
           $281 = ($280|0)==(6); //@line 68 "mmcmp.cpp"
           if ($281) {
            $$0$i62 = 0;$313 = $283;$426 = $417;$427 = $418;
           } else {
            $282 = ($283>>>0)<(24); //@line 69 "mmcmp.cpp"
            if ($282) {
             $284 = (23 - ($283))|0; //@line 71 "mmcmp.cpp"
             $285 = $284 & -8; //@line 71 "mmcmp.cpp"
             $287 = $418;$293 = $283;$295 = $417;$429 = $418;
             while(1) {
              $286 = ($287>>>0)<($105>>>0); //@line 71 "mmcmp.cpp"
              if ($286) {
               $288 = ((($287)) + 1|0); //@line 71 "mmcmp.cpp"
               $289 = HEAP8[$287>>0]|0; //@line 71 "mmcmp.cpp"
               $290 = $289&255; //@line 71 "mmcmp.cpp"
               $292 = $290;$430 = $288;$431 = $288;
              } else {
               $292 = 0;$430 = $429;$431 = $287;
              }
              $291 = $292 << $293; //@line 71 "mmcmp.cpp"
              $294 = $291 | $295; //@line 71 "mmcmp.cpp"
              $296 = (($293) + 8)|0; //@line 72 "mmcmp.cpp"
              $297 = ($296>>>0)<(24); //@line 69 "mmcmp.cpp"
              if ($297) {
               $287 = $431;$293 = $296;$295 = $294;$429 = $430;
              } else {
               $$lcssa267 = $430;$$lcssa268 = $294;
               break;
              }
             }
             $298 = (($283) + 8)|0; //@line 71 "mmcmp.cpp"
             $299 = (($298) + ($285))|0; //@line 71 "mmcmp.cpp"
             $303 = $$lcssa268;$306 = $299;$428 = $$lcssa267;
            } else {
             $303 = $417;$306 = $283;$428 = $418;
            }
            $300 = 1 << $279; //@line 74 "mmcmp.cpp"
            $301 = (($300) + -1)|0; //@line 74 "mmcmp.cpp"
            $302 = $303 & $301; //@line 74 "mmcmp.cpp"
            $304 = $303 >>> $279; //@line 75 "mmcmp.cpp"
            $305 = (($306) - ($279))|0; //@line 76 "mmcmp.cpp"
            $$0$i62 = $302;$313 = $305;$426 = $304;$427 = $428;
           }
           $307 = (($$0$i73) - ($276))|0; //@line 249 "mmcmp.cpp"
           $308 = $307 << $279; //@line 249 "mmcmp.cpp"
           $309 = (($308) + ($$0$i62))|0; //@line 249 "mmcmp.cpp"
           $310 = ($309|0)==($numbits5$0|0); //@line 250 "mmcmp.cpp"
           if (!($310)) {
            $311 = $309 & 7; //@line 252 "mmcmp.cpp"
            $432 = $313;$433 = $427;$434 = $426;$dwPos4$1 = $dwPos4$0;$numbits5$1106 = $311;$oldval7$2 = $oldval7$0;
            break;
           }
           $312 = ($313>>>0)<(24); //@line 69 "mmcmp.cpp"
           if ($312) {
            $314 = (23 - ($313))|0; //@line 71 "mmcmp.cpp"
            $315 = $314 & -8; //@line 71 "mmcmp.cpp"
            $317 = $427;$323 = $313;$325 = $426;$436 = $427;
            while(1) {
             $316 = ($317>>>0)<($105>>>0); //@line 71 "mmcmp.cpp"
             if ($316) {
              $318 = ((($317)) + 1|0); //@line 71 "mmcmp.cpp"
              $319 = HEAP8[$317>>0]|0; //@line 71 "mmcmp.cpp"
              $320 = $319&255; //@line 71 "mmcmp.cpp"
              $322 = $320;$437 = $318;$438 = $318;
             } else {
              $322 = 0;$437 = $436;$438 = $317;
             }
             $321 = $322 << $323; //@line 71 "mmcmp.cpp"
             $324 = $321 | $325; //@line 71 "mmcmp.cpp"
             $326 = (($323) + 8)|0; //@line 72 "mmcmp.cpp"
             $327 = ($326>>>0)<(24); //@line 69 "mmcmp.cpp"
             if ($327) {
              $317 = $438;$323 = $326;$325 = $324;$436 = $437;
             } else {
              $$lcssa269 = $437;$$lcssa270 = $324;
              break;
             }
            }
            $328 = (($313) + 8)|0; //@line 71 "mmcmp.cpp"
            $329 = (($328) + ($315))|0; //@line 71 "mmcmp.cpp"
            $331 = $$lcssa270;$334 = $329;$435 = $$lcssa269;
           } else {
            $331 = $426;$334 = $313;$435 = $427;
           }
           $330 = $331 & 7; //@line 74 "mmcmp.cpp"
           $332 = $331 >>> 3; //@line 75 "mmcmp.cpp"
           $333 = (($334) + -3)|0; //@line 76 "mmcmp.cpp"
           $335 = ($330|0)==(7); //@line 255 "mmcmp.cpp"
           if (!($335)) {
            $359 = $330 | 248; //@line 261 "mmcmp.cpp"
            $423 = $332;$424 = $435;$425 = $333;$newval8$0 = $359;
            label = 106;
            break;
           }
           $336 = ($333>>>0)<(24); //@line 69 "mmcmp.cpp"
           if ($336) {
            $337 = (26 - ($334))|0; //@line 71 "mmcmp.cpp"
            $338 = $337 & -8; //@line 71 "mmcmp.cpp"
            $340 = $435;$346 = $333;$348 = $332;$440 = $435;
            while(1) {
             $339 = ($340>>>0)<($105>>>0); //@line 71 "mmcmp.cpp"
             if ($339) {
              $341 = ((($340)) + 1|0); //@line 71 "mmcmp.cpp"
              $342 = HEAP8[$340>>0]|0; //@line 71 "mmcmp.cpp"
              $343 = $342&255; //@line 71 "mmcmp.cpp"
              $345 = $343;$441 = $341;$442 = $341;
             } else {
              $345 = 0;$441 = $440;$442 = $340;
             }
             $344 = $345 << $346; //@line 71 "mmcmp.cpp"
             $347 = $344 | $348; //@line 71 "mmcmp.cpp"
             $349 = (($346) + 8)|0; //@line 72 "mmcmp.cpp"
             $350 = ($349>>>0)<(24); //@line 69 "mmcmp.cpp"
             if ($350) {
              $340 = $442;$346 = $349;$348 = $347;$440 = $441;
             } else {
              $$lcssa271 = $441;$$lcssa272 = $347;
              break;
             }
            }
            $351 = (($334) + 5)|0; //@line 71 "mmcmp.cpp"
            $352 = (($351) + ($338))|0; //@line 71 "mmcmp.cpp"
            $354 = $$lcssa272;$357 = $352;$439 = $$lcssa271;
           } else {
            $354 = $332;$357 = $333;$439 = $435;
           }
           $353 = $354 & 1; //@line 74 "mmcmp.cpp"
           $355 = $354 >>> 1; //@line 75 "mmcmp.cpp"
           $356 = (($357) + -1)|0; //@line 76 "mmcmp.cpp"
           $358 = ($353|0)==(0); //@line 257 "mmcmp.cpp"
           if ($358) {
            $443 = $356;$444 = $439;$445 = $355;$newval8$0108 = 255;
            label = 107;
           } else {
            break L40;
           }
          }
         } while(0);
         if ((label|0) == 106) {
          label = 0;
          $360 = ($newval8$0>>>0)<(256); //@line 268 "mmcmp.cpp"
          if ($360) {
           $443 = $425;$444 = $424;$445 = $423;$newval8$0108 = $newval8$0;
           label = 107;
          } else {
           $432 = $425;$433 = $424;$434 = $423;$dwPos4$1 = $dwPos4$0;$numbits5$1106 = $numbits5$0;$oldval7$2 = $oldval7$0;
          }
         }
         if ((label|0) == 107) {
          label = 0;
          $$sum10 = (($newval8$0108) + ($69))|0; //@line 270 "mmcmp.cpp"
          $361 = (($1) + ($$sum10)|0); //@line 270 "mmcmp.cpp"
          $362 = HEAP8[$361>>0]|0; //@line 270 "mmcmp.cpp"
          $363 = $362&255; //@line 270 "mmcmp.cpp"
          $364 = HEAP16[$71>>1]|0; //@line 271 "mmcmp.cpp"
          $365 = $364 & 2; //@line 271 "mmcmp.cpp"
          $366 = ($365<<16>>16)==(0); //@line 271 "mmcmp.cpp"
          $367 = (($363) + ($oldval7$0))|0; //@line 273 "mmcmp.cpp"
          $oldval7$0$ = $366 ? $oldval7$0 : $367; //@line 271 "mmcmp.cpp"
          $$18 = $366 ? $363 : $367; //@line 271 "mmcmp.cpp"
          $368 = $$18&255; //@line 276 "mmcmp.cpp"
          $369 = (($dwPos4$0) + 1)|0; //@line 276 "mmcmp.cpp"
          $$sum156 = (($370) + ($dwPos4$0))|0; //@line 276 "mmcmp.cpp"
          $371 = (($55) + ($$sum156)|0); //@line 276 "mmcmp.cpp"
          HEAP8[$371>>0] = $368; //@line 276 "mmcmp.cpp"
          $432 = $443;$433 = $444;$434 = $445;$dwPos4$1 = $369;$numbits5$1106 = $numbits5$0;$oldval7$2 = $oldval7$0$;
         }
         $372 = ($dwPos4$1>>>0)<($dwSize3$0$ph>>>0); //@line 278 "mmcmp.cpp"
         if ($372) {
          $251 = $432;$415 = $434;$416 = $433;$dwPos4$0 = $dwPos4$1;$numbits5$0 = $numbits5$1106;$oldval7$0 = $oldval7$2;
         } else {
          $$lcssa273 = $432;$$lcssa274 = $433;$$lcssa275 = $434;$numbits5$1106$lcssa = $numbits5$1106;$oldval7$2$lcssa = $oldval7$2;
          break;
         }
        }
        $373 = (($subblk6$0$ph) + 1)|0; //@line 280 "mmcmp.cpp"
        $374 = (((($63) + ($373<<3)|0)) + 4|0); //@line 282 "mmcmp.cpp"
        $375 = (($63) + ($373<<3)|0); //@line 283 "mmcmp.cpp"
        $376 = HEAP32[$375>>2]|0; //@line 283 "mmcmp.cpp"
        $$ph = $$lcssa275;$$ph111 = $$lcssa274;$$ph112 = $$lcssa273;$370 = $376;$dwSize3$0$ph$in = $374;$numbits5$0$ph = $numbits5$1106$lcssa;$oldval7$0$ph = $oldval7$2$lcssa;$subblk6$0$ph = $373;
       }
      } else {
       $$ph115 = 0;$$ph116 = $102;$$ph117 = 0;$$pn = $94;$dwSize$0$ph$in$in = $95;$numbits$0$ph = $98;$oldval$0$ph = 0;$subblk$0$ph = 0;
      }
      while(1) {
       $dwSize$0$ph$in = HEAP32[$dwSize$0$ph$in$in>>2]|0; //@line 222 "mmcmp.cpp"
       $pDest$0$ph$in = (($55) + ($$pn)|0); //@line 223 "mmcmp.cpp"
       $dwSize$0$ph = $dwSize$0$ph$in >>> 1; //@line 222 "mmcmp.cpp"
       $112 = $$ph117;$381 = $$ph115;$382 = $$ph116;$dwPos$0 = 0;$numbits$0 = $numbits$0$ph;$oldval$0 = $oldval$0$ph;
       while(1) {
        $106 = HEAP16[$65>>1]|0; //@line 177 "mmcmp.cpp"
        $107 = $106&65535; //@line 177 "mmcmp.cpp"
        $108 = ($subblk$0$ph>>>0)<($107>>>0); //@line 177 "mmcmp.cpp"
        if (!($108)) {
         break L40;
        }
        $109 = (($numbits$0) + 1)|0; //@line 180 "mmcmp.cpp"
        $110 = ($109|0)==(0); //@line 68 "mmcmp.cpp"
        if ($110) {
         $$0$i19 = 0;$144 = $112;$383 = $381;$384 = $382;
        } else {
         $111 = ($112>>>0)<(24); //@line 69 "mmcmp.cpp"
         if ($111) {
          $113 = (23 - ($112))|0; //@line 71 "mmcmp.cpp"
          $114 = $113 & -8; //@line 71 "mmcmp.cpp"
          $116 = $382;$122 = $112;$124 = $381;$386 = $382;
          while(1) {
           $115 = ($116>>>0)<($105>>>0); //@line 71 "mmcmp.cpp"
           if ($115) {
            $117 = ((($116)) + 1|0); //@line 71 "mmcmp.cpp"
            $118 = HEAP8[$116>>0]|0; //@line 71 "mmcmp.cpp"
            $119 = $118&255; //@line 71 "mmcmp.cpp"
            $121 = $119;$387 = $117;$388 = $117;
           } else {
            $121 = 0;$387 = $386;$388 = $116;
           }
           $120 = $121 << $122; //@line 71 "mmcmp.cpp"
           $123 = $120 | $124; //@line 71 "mmcmp.cpp"
           $125 = (($122) + 8)|0; //@line 72 "mmcmp.cpp"
           $126 = ($125>>>0)<(24); //@line 69 "mmcmp.cpp"
           if ($126) {
            $116 = $388;$122 = $125;$124 = $123;$386 = $387;
           } else {
            $$lcssa = $387;$$lcssa255 = $123;
            break;
           }
          }
          $127 = (($112) + 8)|0; //@line 71 "mmcmp.cpp"
          $128 = (($127) + ($114))|0; //@line 71 "mmcmp.cpp"
          $132 = $$lcssa255;$135 = $128;$385 = $$lcssa;
         } else {
          $132 = $381;$135 = $112;$385 = $382;
         }
         $129 = 1 << $109; //@line 74 "mmcmp.cpp"
         $130 = (($129) + -1)|0; //@line 74 "mmcmp.cpp"
         $131 = $132 & $130; //@line 74 "mmcmp.cpp"
         $133 = $132 >>> $109; //@line 75 "mmcmp.cpp"
         $134 = (($135) - ($109))|0; //@line 76 "mmcmp.cpp"
         $$0$i19 = $131;$144 = $134;$383 = $133;$384 = $385;
        }
        $136 = (46420 + ($numbits$0<<2)|0); //@line 182 "mmcmp.cpp"
        $137 = HEAP32[$136>>2]|0; //@line 182 "mmcmp.cpp"
        $138 = ($$0$i19>>>0)<($137>>>0); //@line 182 "mmcmp.cpp"
        do {
         if ($138) {
          $389 = $383;$390 = $384;$391 = $144;$newval$0 = $$0$i19;
          label = 60;
         } else {
          $139 = (46484 + ($numbits$0<<2)|0); //@line 184 "mmcmp.cpp"
          $140 = HEAP32[$139>>2]|0; //@line 184 "mmcmp.cpp"
          $141 = (($numbits$0) + -7)|0; //@line 68 "mmcmp.cpp"
          $142 = ($141>>>0)<(9); //@line 68 "mmcmp.cpp"
          if ($142) {
           $$0$i29 = 0;$174 = $144;$392 = $383;$393 = $384;
          } else {
           $143 = ($144>>>0)<(24); //@line 69 "mmcmp.cpp"
           if ($143) {
            $145 = (23 - ($144))|0; //@line 71 "mmcmp.cpp"
            $146 = $145 & -8; //@line 71 "mmcmp.cpp"
            $148 = $384;$154 = $144;$156 = $383;$395 = $384;
            while(1) {
             $147 = ($148>>>0)<($105>>>0); //@line 71 "mmcmp.cpp"
             if ($147) {
              $149 = ((($148)) + 1|0); //@line 71 "mmcmp.cpp"
              $150 = HEAP8[$148>>0]|0; //@line 71 "mmcmp.cpp"
              $151 = $150&255; //@line 71 "mmcmp.cpp"
              $153 = $151;$396 = $149;$397 = $149;
             } else {
              $153 = 0;$396 = $395;$397 = $148;
             }
             $152 = $153 << $154; //@line 71 "mmcmp.cpp"
             $155 = $152 | $156; //@line 71 "mmcmp.cpp"
             $157 = (($154) + 8)|0; //@line 72 "mmcmp.cpp"
             $158 = ($157>>>0)<(24); //@line 69 "mmcmp.cpp"
             if ($158) {
              $148 = $397;$154 = $157;$156 = $155;$395 = $396;
             } else {
              $$lcssa256 = $396;$$lcssa257 = $155;
              break;
             }
            }
            $159 = (($144) + 8)|0; //@line 71 "mmcmp.cpp"
            $160 = (($159) + ($146))|0; //@line 71 "mmcmp.cpp"
            $164 = $$lcssa257;$167 = $160;$394 = $$lcssa256;
           } else {
            $164 = $383;$167 = $144;$394 = $384;
           }
           $161 = 1 << $140; //@line 74 "mmcmp.cpp"
           $162 = (($161) + -1)|0; //@line 74 "mmcmp.cpp"
           $163 = $164 & $162; //@line 74 "mmcmp.cpp"
           $165 = $164 >>> $140; //@line 75 "mmcmp.cpp"
           $166 = (($167) - ($140))|0; //@line 76 "mmcmp.cpp"
           $$0$i29 = $163;$174 = $166;$392 = $165;$393 = $394;
          }
          $168 = (($$0$i19) - ($137))|0; //@line 185 "mmcmp.cpp"
          $169 = $168 << $140; //@line 185 "mmcmp.cpp"
          $170 = (($169) + ($$0$i29))|0; //@line 185 "mmcmp.cpp"
          $171 = ($170|0)==($numbits$0|0); //@line 186 "mmcmp.cpp"
          if (!($171)) {
           $172 = $170 & 15; //@line 188 "mmcmp.cpp"
           $398 = $174;$399 = $393;$400 = $392;$dwPos$1 = $dwPos$0;$numbits$197 = $172;$oldval$2 = $oldval$0;
           break;
          }
          $173 = ($174>>>0)<(24); //@line 69 "mmcmp.cpp"
          if ($173) {
           $175 = (23 - ($174))|0; //@line 71 "mmcmp.cpp"
           $176 = $175 & -8; //@line 71 "mmcmp.cpp"
           $178 = $393;$184 = $174;$186 = $392;$402 = $393;
           while(1) {
            $177 = ($178>>>0)<($105>>>0); //@line 71 "mmcmp.cpp"
            if ($177) {
             $179 = ((($178)) + 1|0); //@line 71 "mmcmp.cpp"
             $180 = HEAP8[$178>>0]|0; //@line 71 "mmcmp.cpp"
             $181 = $180&255; //@line 71 "mmcmp.cpp"
             $183 = $181;$403 = $179;$404 = $179;
            } else {
             $183 = 0;$403 = $402;$404 = $178;
            }
            $182 = $183 << $184; //@line 71 "mmcmp.cpp"
            $185 = $182 | $186; //@line 71 "mmcmp.cpp"
            $187 = (($184) + 8)|0; //@line 72 "mmcmp.cpp"
            $188 = ($187>>>0)<(24); //@line 69 "mmcmp.cpp"
            if ($188) {
             $178 = $404;$184 = $187;$186 = $185;$402 = $403;
            } else {
             $$lcssa258 = $403;$$lcssa259 = $185;
             break;
            }
           }
           $189 = (($174) + 8)|0; //@line 71 "mmcmp.cpp"
           $190 = (($189) + ($176))|0; //@line 71 "mmcmp.cpp"
           $192 = $$lcssa259;$195 = $190;$401 = $$lcssa258;
          } else {
           $192 = $392;$195 = $174;$401 = $393;
          }
          $191 = $192 & 15; //@line 74 "mmcmp.cpp"
          $193 = $192 >>> 4; //@line 75 "mmcmp.cpp"
          $194 = (($195) + -4)|0; //@line 76 "mmcmp.cpp"
          $196 = ($191|0)==(15); //@line 191 "mmcmp.cpp"
          if (!($196)) {
           $220 = $191 | 65520; //@line 197 "mmcmp.cpp"
           $389 = $193;$390 = $401;$391 = $194;$newval$0 = $220;
           label = 60;
           break;
          }
          $197 = ($194>>>0)<(24); //@line 69 "mmcmp.cpp"
          if ($197) {
           $198 = (27 - ($195))|0; //@line 71 "mmcmp.cpp"
           $199 = $198 & -8; //@line 71 "mmcmp.cpp"
           $201 = $401;$207 = $194;$209 = $193;$406 = $401;
           while(1) {
            $200 = ($201>>>0)<($105>>>0); //@line 71 "mmcmp.cpp"
            if ($200) {
             $202 = ((($201)) + 1|0); //@line 71 "mmcmp.cpp"
             $203 = HEAP8[$201>>0]|0; //@line 71 "mmcmp.cpp"
             $204 = $203&255; //@line 71 "mmcmp.cpp"
             $206 = $204;$407 = $202;$408 = $202;
            } else {
             $206 = 0;$407 = $406;$408 = $201;
            }
            $205 = $206 << $207; //@line 71 "mmcmp.cpp"
            $208 = $205 | $209; //@line 71 "mmcmp.cpp"
            $210 = (($207) + 8)|0; //@line 72 "mmcmp.cpp"
            $211 = ($210>>>0)<(24); //@line 69 "mmcmp.cpp"
            if ($211) {
             $201 = $408;$207 = $210;$209 = $208;$406 = $407;
            } else {
             $$lcssa260 = $407;$$lcssa261 = $208;
             break;
            }
           }
           $212 = (($195) + 4)|0; //@line 71 "mmcmp.cpp"
           $213 = (($212) + ($199))|0; //@line 71 "mmcmp.cpp"
           $215 = $$lcssa261;$218 = $213;$405 = $$lcssa260;
          } else {
           $215 = $193;$218 = $194;$405 = $401;
          }
          $214 = $215 & 1; //@line 74 "mmcmp.cpp"
          $216 = $215 >>> 1; //@line 75 "mmcmp.cpp"
          $217 = (($218) + -1)|0; //@line 76 "mmcmp.cpp"
          $219 = ($214|0)==(0); //@line 193 "mmcmp.cpp"
          if ($219) {
           $409 = $216;$410 = $405;$411 = $217;$newval$099102 = 65535;
           label = 62;
          } else {
           break L40;
          }
         }
        } while(0);
        do {
         if ((label|0) == 60) {
          label = 0;
          $221 = ($newval$0>>>0)<(65536); //@line 204 "mmcmp.cpp"
          if (!($221)) {
           $398 = $391;$399 = $390;$400 = $389;$dwPos$1 = $dwPos$0;$numbits$197 = $numbits$0;$oldval$2 = $oldval$0;
           break;
          }
          $222 = $newval$0 & 1; //@line 206 "mmcmp.cpp"
          $223 = ($222|0)==(0); //@line 206 "mmcmp.cpp"
          if (!($223)) {
           $409 = $389;$410 = $390;$411 = $391;$newval$099102 = $newval$0;
           label = 62;
           break;
          }
          $227 = $newval$0 >>> 1; //@line 206 "mmcmp.cpp"
          $233 = $227;$412 = $389;$413 = $390;$414 = $391;
          label = 64;
         }
        } while(0);
        if ((label|0) == 62) {
         label = 0;
         $224 = (($newval$099102) + 1)|0; //@line 206 "mmcmp.cpp"
         $225 = $224 >>> 1; //@line 206 "mmcmp.cpp"
         $226 = (0 - ($225))|0; //@line 206 "mmcmp.cpp"
         $233 = $226;$412 = $409;$413 = $410;$414 = $411;
         label = 64;
        }
        if ((label|0) == 64) {
         label = 0;
         $228 = HEAP16[$71>>1]|0; //@line 207 "mmcmp.cpp"
         $229 = $228&65535; //@line 207 "mmcmp.cpp"
         $230 = $229 & 2; //@line 207 "mmcmp.cpp"
         $231 = ($230|0)==(0); //@line 207 "mmcmp.cpp"
         if ($231) {
          $234 = $229 & 512; //@line 212 "mmcmp.cpp"
          $235 = ($234|0)==(0); //@line 212 "mmcmp.cpp"
          $236 = $233 ^ 32768; //@line 214 "mmcmp.cpp"
          $$ = $235 ? $236 : $233; //@line 212 "mmcmp.cpp"
          $newval$1 = $$;$oldval$1 = $oldval$0;
         } else {
          $232 = (($233) + ($oldval$0))|0; //@line 209 "mmcmp.cpp"
          $newval$1 = $232;$oldval$1 = $232;
         }
         $237 = $newval$1&65535; //@line 216 "mmcmp.cpp"
         $238 = (($dwPos$0) + 1)|0; //@line 216 "mmcmp.cpp"
         $239 = (($pDest$0$ph$in) + ($dwPos$0<<1)|0); //@line 216 "mmcmp.cpp"
         HEAP16[$239>>1] = $237; //@line 216 "mmcmp.cpp"
         $398 = $414;$399 = $413;$400 = $412;$dwPos$1 = $238;$numbits$197 = $numbits$0;$oldval$2 = $oldval$1;
        }
        $240 = ($dwPos$1>>>0)<($dwSize$0$ph>>>0); //@line 218 "mmcmp.cpp"
        if ($240) {
         $112 = $398;$381 = $400;$382 = $399;$dwPos$0 = $dwPos$1;$numbits$0 = $numbits$197;$oldval$0 = $oldval$2;
        } else {
         $$lcssa262 = $398;$$lcssa263 = $399;$$lcssa264 = $400;$numbits$197$lcssa = $numbits$197;$oldval$2$lcssa = $oldval$2;
         break;
        }
       }
       $241 = (($subblk$0$ph) + 1)|0; //@line 220 "mmcmp.cpp"
       $242 = (((($63) + ($241<<3)|0)) + 4|0); //@line 222 "mmcmp.cpp"
       $243 = (($63) + ($241<<3)|0); //@line 223 "mmcmp.cpp"
       $244 = HEAP32[$243>>2]|0; //@line 223 "mmcmp.cpp"
       $$ph115 = $$lcssa264;$$ph116 = $$lcssa263;$$ph117 = $$lcssa262;$$pn = $244;$dwSize$0$ph$in$in = $242;$numbits$0$ph = $numbits$197$lcssa;$oldval$0$ph = $oldval$2$lcssa;$subblk$0$ph = $241;
      }
     }
    } while(0);
    $377 = (($nBlock$0152) + 1)|0; //@line 130 "mmcmp.cpp"
    $378 = HEAP16[$40>>1]|0; //@line 130 "mmcmp.cpp"
    $379 = $378&65535; //@line 130 "mmcmp.cpp"
    $380 = ($377>>>0)<($379>>>0); //@line 130 "mmcmp.cpp"
    if ($380) {
     $nBlock$0152 = $377;
    } else {
     break;
    }
   }
  }
 } while(0);
 HEAP32[$ppMemFile>>2] = $55; //@line 288 "mmcmp.cpp"
 HEAP32[$pdwMemLength>>2] = $44; //@line 289 "mmcmp.cpp"
 $$0 = 1;
 return ($$0|0); //@line 291 "mmcmp.cpp"
}
function __Z13PP20_DoUnpackPKhjPhj($pSrc,$nSrcLen,$pDst,$nDstLen) {
 $pSrc = $pSrc|0;
 $nSrcLen = $nSrcLen|0;
 $pDst = $pDst|0;
 $nDstLen = $nDstLen|0;
 var $$ = 0, $$74 = 0, $$76 = 0, $$76$1 = 0, $$78 = 0, $$78$1 = 0, $$78$2 = 0, $$78$3 = 0, $$78$4 = 0, $$78$5 = 0, $$78$6 = 0, $$78$7 = 0, $$80 = 0, $$80$1 = 0, $$82 = 0, $$84 = 0, $$86 = 0, $$86$1 = 0, $$86$2 = 0, $$88 = 0;
 var $$lcssa = 0, $$lcssa112 = 0, $$lcssa113 = 0, $$lcssa114 = 0, $$lcssa250 = 0, $$lcssa251 = 0, $$ph = 0, $$ph100 = 0, $$ph183 = 0, $$ph184 = 0, $$ph185 = 0, $$ph186 = 0, $$ph187 = 0, $$ph188 = 0, $$ph189 = 0, $$ph98 = 0, $$ph99 = 0, $$sum = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0;
 var $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0;
 var $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0;
 var $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0;
 var $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0;
 var $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0;
 var $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0;
 var $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $exitcond$i = 0, $exitcond$i25 = 0, $exitcond$i7 = 0, $i$0138 = 0, $i$02$i = 0, $i$02$i21 = 0, $i$02$i3 = 0, $i3$0144 = 0, $n$0 = 0, $n$1 = 0, $n1$0 = 0;
 var $n1$1 = 0, $nBytesLeft$0145 = 0, $nBytesLeft$1$lcssa = 0, $nBytesLeft$1139 = 0, $nBytesLeft$2 = 0, $nBytesLeft$3143 = 0, $nofs$0 = 0, $result$0$lcssa$i26 = 0, $result$01$i22 = 0, $result$01$i4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$sum = (($nSrcLen) + -4)|0; //@line 336 "mmcmp.cpp"
 $0 = (($pSrc) + ($$sum)|0); //@line 336 "mmcmp.cpp"
 $1 = (($nSrcLen) + -1)|0; //@line 339 "mmcmp.cpp"
 $2 = (($pSrc) + ($1)|0); //@line 339 "mmcmp.cpp"
 $3 = HEAP8[$2>>0]|0; //@line 339 "mmcmp.cpp"
 $4 = $3&255; //@line 339 "mmcmp.cpp"
 $5 = ($3<<24>>24)==(0); //@line 314 "mmcmp.cpp"
 if ($5) {
  $$ph100 = 0;$$ph98 = $0;$$ph99 = 0;
 } else {
  $309 = 0;$7 = 0;$9 = $0;$i$02$i = 0;
  while(1) {
   $6 = ($7|0)==(0); //@line 316 "mmcmp.cpp"
   if ($6) {
    $8 = ($9|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
    $10 = ((($9)) + -1|0); //@line 319 "mmcmp.cpp"
    $$ = $8 ? $9 : $10; //@line 319 "mmcmp.cpp"
    $11 = HEAP8[$$>>0]|0; //@line 320 "mmcmp.cpp"
    $12 = $11&255; //@line 320 "mmcmp.cpp"
    $14 = $12;$16 = 8;$310 = $$;
   } else {
    $14 = $309;$16 = $7;$310 = $9;
   }
   $13 = $14 >>> 1; //@line 323 "mmcmp.cpp"
   $15 = (($16) + -1)|0; //@line 324 "mmcmp.cpp"
   $17 = (($i$02$i) + 1)|0; //@line 314 "mmcmp.cpp"
   $exitcond$i = ($17|0)==($4|0); //@line 314 "mmcmp.cpp"
   if ($exitcond$i) {
    $$ph100 = $15;$$ph98 = $310;$$ph99 = $13;
    break;
   } else {
    $309 = $13;$7 = $15;$9 = $310;$i$02$i = $17;
   }
  }
 }
 $18 = ($nDstLen|0)==(0); //@line 341 "mmcmp.cpp"
 if ($18) {
  return; //@line 382 "mmcmp.cpp"
 } else {
  $20 = $$ph100;$22 = $$ph98;$311 = $$ph99;$nBytesLeft$0145 = $nDstLen;
 }
 L10: while(1) {
  $19 = ($20|0)==(0); //@line 316 "mmcmp.cpp"
  if ($19) {
   $21 = ($22|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
   $23 = ((($22)) + -1|0); //@line 319 "mmcmp.cpp"
   $$74 = $21 ? $22 : $23; //@line 319 "mmcmp.cpp"
   $24 = HEAP8[$$74>>0]|0; //@line 320 "mmcmp.cpp"
   $25 = $24&255; //@line 320 "mmcmp.cpp"
   $27 = $25;$30 = 8;$312 = $$74;
  } else {
   $27 = $311;$30 = $20;$312 = $22;
  }
  $26 = $27 & 1; //@line 322 "mmcmp.cpp"
  $28 = $27 >>> 1; //@line 323 "mmcmp.cpp"
  $29 = (($30) + -1)|0; //@line 324 "mmcmp.cpp"
  $31 = ($26|0)==(0); //@line 343 "mmcmp.cpp"
  if ($31) {
   $34 = $29;$36 = $312;$42 = $28;$n$0 = 1;
   while(1) {
    $32 = ($n$0>>>0)<($nBytesLeft$0145>>>0); //@line 346 "mmcmp.cpp"
    if (!($32)) {
     $313 = $36;$314 = $42;$315 = $34;$n$1 = $n$0;
     break;
    }
    $33 = ($34|0)==(0); //@line 316 "mmcmp.cpp"
    if ($33) {
     $35 = ($36|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
     $37 = ((($36)) + -1|0); //@line 319 "mmcmp.cpp"
     $$76 = $35 ? $36 : $37; //@line 319 "mmcmp.cpp"
     $38 = HEAP8[$$76>>0]|0; //@line 320 "mmcmp.cpp"
     $39 = $38&255; //@line 320 "mmcmp.cpp"
     $40 = $39 >>> 1; //@line 323 "mmcmp.cpp"
     $150 = $39;$153 = $40;$157 = 7;$316 = $$76;
    } else {
     $41 = $42 >>> 1; //@line 323 "mmcmp.cpp"
     $43 = (($34) + -1)|0; //@line 324 "mmcmp.cpp"
     $44 = ($43|0)==(0); //@line 316 "mmcmp.cpp"
     if ($44) {
      $145 = ($36|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
      $146 = ((($36)) + -1|0); //@line 319 "mmcmp.cpp"
      $$76$1 = $145 ? $36 : $146; //@line 319 "mmcmp.cpp"
      $147 = HEAP8[$$76$1>>0]|0; //@line 320 "mmcmp.cpp"
      $148 = $147&255; //@line 320 "mmcmp.cpp"
      $150 = $42;$153 = $148;$157 = 8;$316 = $$76$1;
     } else {
      $150 = $42;$153 = $41;$157 = $43;$316 = $36;
     }
    }
    $149 = $150 << 1; //@line 322 "mmcmp.cpp"
    $151 = $149 & 2; //@line 322 "mmcmp.cpp"
    $152 = $153 & 1; //@line 322 "mmcmp.cpp"
    $154 = $152 | $151; //@line 322 "mmcmp.cpp"
    $155 = $153 >>> 1; //@line 323 "mmcmp.cpp"
    $156 = (($157) + -1)|0; //@line 324 "mmcmp.cpp"
    $158 = (($154) + ($n$0))|0; //@line 349 "mmcmp.cpp"
    $159 = ($154|0)==(3); //@line 350 "mmcmp.cpp"
    if ($159) {
     $34 = $156;$36 = $316;$42 = $155;$n$0 = $158;
    } else {
     $313 = $316;$314 = $155;$315 = $156;$n$1 = $158;
     break;
    }
   }
   $45 = ($n$1|0)==(0); //@line 352 "mmcmp.cpp"
   if ($45) {
    $$lcssa112 = $313;$$lcssa113 = $314;$$lcssa114 = $315;$nBytesLeft$1$lcssa = $nBytesLeft$0145;
   } else {
    $47 = $315;$49 = $313;$55 = $314;$i$0138 = 0;$nBytesLeft$1139 = $nBytesLeft$0145;
    while(1) {
     $46 = ($47|0)==(0); //@line 316 "mmcmp.cpp"
     if ($46) {
      $48 = ($49|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
      $50 = ((($49)) + -1|0); //@line 319 "mmcmp.cpp"
      $$78 = $48 ? $49 : $50; //@line 319 "mmcmp.cpp"
      $51 = HEAP8[$$78>>0]|0; //@line 320 "mmcmp.cpp"
      $52 = $51&255; //@line 320 "mmcmp.cpp"
      $53 = $52 >>> 1; //@line 323 "mmcmp.cpp"
      $165 = $52;$168 = $53;$171 = 7;$173 = $$78;
     } else {
      $54 = $55 >>> 1; //@line 323 "mmcmp.cpp"
      $56 = (($47) + -1)|0; //@line 324 "mmcmp.cpp"
      $57 = ($56|0)==(0); //@line 316 "mmcmp.cpp"
      if ($57) {
       $160 = ($49|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
       $161 = ((($49)) + -1|0); //@line 319 "mmcmp.cpp"
       $$78$1 = $160 ? $49 : $161; //@line 319 "mmcmp.cpp"
       $162 = HEAP8[$$78$1>>0]|0; //@line 320 "mmcmp.cpp"
       $163 = $162&255; //@line 320 "mmcmp.cpp"
       $165 = $55;$168 = $163;$171 = 8;$173 = $$78$1;
      } else {
       $165 = $55;$168 = $54;$171 = $56;$173 = $49;
      }
     }
     $164 = $165 << 1; //@line 322 "mmcmp.cpp"
     $166 = $164 & 2; //@line 322 "mmcmp.cpp"
     $167 = $168 & 1; //@line 322 "mmcmp.cpp"
     $169 = $167 | $166; //@line 322 "mmcmp.cpp"
     $170 = ($171|0)==(1); //@line 316 "mmcmp.cpp"
     do {
      if ($170) {
       $172 = ($173|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
       $174 = ((($173)) + -1|0); //@line 319 "mmcmp.cpp"
       $$78$2 = $172 ? $173 : $174; //@line 319 "mmcmp.cpp"
       $175 = HEAP8[$$78$2>>0]|0; //@line 320 "mmcmp.cpp"
       $176 = $175&255; //@line 320 "mmcmp.cpp"
       $177 = $169 << 1; //@line 322 "mmcmp.cpp"
       $178 = $176 & 1; //@line 322 "mmcmp.cpp"
       $179 = $178 | $177; //@line 322 "mmcmp.cpp"
       $180 = $176 >>> 1; //@line 323 "mmcmp.cpp"
       $$ph = $179;$$ph183 = $$78$2;$$ph184 = 6;$$ph185 = $180;
       label = 50;
      } else {
       $181 = $168 >>> 1; //@line 323 "mmcmp.cpp"
       $182 = $169 << 1; //@line 322 "mmcmp.cpp"
       $183 = $181 & 1; //@line 322 "mmcmp.cpp"
       $184 = $183 | $182; //@line 322 "mmcmp.cpp"
       $185 = ($171|0)==(2); //@line 316 "mmcmp.cpp"
       if ($185) {
        $186 = ($173|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
        $187 = ((($173)) + -1|0); //@line 319 "mmcmp.cpp"
        $$78$3 = $186 ? $173 : $187; //@line 319 "mmcmp.cpp"
        $188 = HEAP8[$$78$3>>0]|0; //@line 320 "mmcmp.cpp"
        $189 = $188&255; //@line 320 "mmcmp.cpp"
        $$ph = $184;$$ph183 = $$78$3;$$ph184 = 7;$$ph185 = $189;
        label = 50;
        break;
       }
       $194 = $168 >>> 2; //@line 323 "mmcmp.cpp"
       $195 = $184 << 1; //@line 322 "mmcmp.cpp"
       $196 = $194 & 1; //@line 322 "mmcmp.cpp"
       $197 = $196 | $195; //@line 322 "mmcmp.cpp"
       $198 = $168 >>> 3; //@line 323 "mmcmp.cpp"
       $199 = (($171) + -3)|0; //@line 324 "mmcmp.cpp"
       $200 = ($199|0)==(0); //@line 316 "mmcmp.cpp"
       if ($200) {
        $201 = ($173|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
        $202 = ((($173)) + -1|0); //@line 319 "mmcmp.cpp"
        $$78$4 = $201 ? $173 : $202; //@line 319 "mmcmp.cpp"
        $203 = HEAP8[$$78$4>>0]|0; //@line 320 "mmcmp.cpp"
        $204 = $203&255; //@line 320 "mmcmp.cpp"
        $205 = $197 << 1; //@line 322 "mmcmp.cpp"
        $206 = $204 & 1; //@line 322 "mmcmp.cpp"
        $207 = $206 | $205; //@line 322 "mmcmp.cpp"
        $208 = $204 >>> 1; //@line 323 "mmcmp.cpp"
        $$ph186 = $207;$$ph187 = $$78$4;$$ph188 = 6;$$ph189 = $208;
        label = 55;
       } else {
        $210 = $197;$212 = $198;$215 = $199;$217 = $173;
        label = 53;
       }
      }
     } while(0);
     if ((label|0) == 50) {
      label = 0;
      $190 = $$ph << 1; //@line 322 "mmcmp.cpp"
      $191 = $$ph185 & 1; //@line 322 "mmcmp.cpp"
      $192 = $191 | $190; //@line 322 "mmcmp.cpp"
      $193 = $$ph185 >>> 1; //@line 323 "mmcmp.cpp"
      $210 = $192;$212 = $193;$215 = $$ph184;$217 = $$ph183;
      label = 53;
     }
     do {
      if ((label|0) == 53) {
       label = 0;
       $209 = $210 << 1; //@line 322 "mmcmp.cpp"
       $211 = $212 & 1; //@line 322 "mmcmp.cpp"
       $213 = $211 | $209; //@line 322 "mmcmp.cpp"
       $214 = ($215|0)==(1); //@line 316 "mmcmp.cpp"
       if ($214) {
        $216 = ($217|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
        $218 = ((($217)) + -1|0); //@line 319 "mmcmp.cpp"
        $$78$5 = $216 ? $217 : $218; //@line 319 "mmcmp.cpp"
        $219 = HEAP8[$$78$5>>0]|0; //@line 320 "mmcmp.cpp"
        $220 = $219&255; //@line 320 "mmcmp.cpp"
        $$ph186 = $213;$$ph187 = $$78$5;$$ph188 = 7;$$ph189 = $220;
        label = 55;
        break;
       }
       $225 = $212 >>> 1; //@line 323 "mmcmp.cpp"
       $226 = $213 << 1; //@line 322 "mmcmp.cpp"
       $227 = $225 & 1; //@line 322 "mmcmp.cpp"
       $228 = $227 | $226; //@line 322 "mmcmp.cpp"
       $229 = $212 >>> 2; //@line 323 "mmcmp.cpp"
       $230 = (($215) + -2)|0; //@line 324 "mmcmp.cpp"
       $231 = ($230|0)==(0); //@line 316 "mmcmp.cpp"
       if ($231) {
        $232 = ($217|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
        $233 = ((($217)) + -1|0); //@line 319 "mmcmp.cpp"
        $$78$6 = $232 ? $217 : $233; //@line 319 "mmcmp.cpp"
        $234 = HEAP8[$$78$6>>0]|0; //@line 320 "mmcmp.cpp"
        $235 = $234&255; //@line 320 "mmcmp.cpp"
        $236 = $228 << 1; //@line 322 "mmcmp.cpp"
        $237 = $235 & 1; //@line 322 "mmcmp.cpp"
        $238 = $237 | $236; //@line 322 "mmcmp.cpp"
        $239 = $235 >>> 1; //@line 323 "mmcmp.cpp"
        $255 = $238;$257 = $239;$261 = 7;$328 = $$78$6;
       } else {
        $241 = $228;$243 = $229;$247 = $230;$250 = $217;
        label = 58;
       }
      }
     } while(0);
     if ((label|0) == 55) {
      label = 0;
      $221 = $$ph186 << 1; //@line 322 "mmcmp.cpp"
      $222 = $$ph189 & 1; //@line 322 "mmcmp.cpp"
      $223 = $222 | $221; //@line 322 "mmcmp.cpp"
      $224 = $$ph189 >>> 1; //@line 323 "mmcmp.cpp"
      $241 = $223;$243 = $224;$247 = $$ph188;$250 = $$ph187;
      label = 58;
     }
     if ((label|0) == 58) {
      label = 0;
      $240 = $241 << 1; //@line 322 "mmcmp.cpp"
      $242 = $243 & 1; //@line 322 "mmcmp.cpp"
      $244 = $242 | $240; //@line 322 "mmcmp.cpp"
      $245 = $243 >>> 1; //@line 323 "mmcmp.cpp"
      $246 = (($247) + -1)|0; //@line 324 "mmcmp.cpp"
      $248 = ($246|0)==(0); //@line 316 "mmcmp.cpp"
      if ($248) {
       $249 = ($250|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
       $251 = ((($250)) + -1|0); //@line 319 "mmcmp.cpp"
       $$78$7 = $249 ? $250 : $251; //@line 319 "mmcmp.cpp"
       $252 = HEAP8[$$78$7>>0]|0; //@line 320 "mmcmp.cpp"
       $253 = $252&255; //@line 320 "mmcmp.cpp"
       $255 = $244;$257 = $253;$261 = 8;$328 = $$78$7;
      } else {
       $255 = $244;$257 = $245;$261 = $246;$328 = $250;
      }
     }
     $254 = $255 << 1; //@line 322 "mmcmp.cpp"
     $256 = $257 & 1; //@line 322 "mmcmp.cpp"
     $258 = $256 | $254; //@line 322 "mmcmp.cpp"
     $259 = $257 >>> 1; //@line 323 "mmcmp.cpp"
     $260 = (($261) + -1)|0; //@line 324 "mmcmp.cpp"
     $262 = $258&255; //@line 354 "mmcmp.cpp"
     $263 = (($nBytesLeft$1139) + -1)|0; //@line 354 "mmcmp.cpp"
     $264 = (($pDst) + ($263)|0); //@line 354 "mmcmp.cpp"
     HEAP8[$264>>0] = $262; //@line 354 "mmcmp.cpp"
     $265 = (($i$0138) + 1)|0; //@line 352 "mmcmp.cpp"
     $exitcond = ($265|0)==($n$1|0); //@line 352 "mmcmp.cpp"
     if ($exitcond) {
      $$lcssa = $328;$$lcssa250 = $259;$$lcssa251 = $260;
      break;
     } else {
      $47 = $260;$49 = $328;$55 = $259;$i$0138 = $265;$nBytesLeft$1139 = $263;
     }
    }
    $58 = (($nBytesLeft$0145) - ($n$1))|0; //@line 352 "mmcmp.cpp"
    $$lcssa112 = $$lcssa;$$lcssa113 = $$lcssa250;$$lcssa114 = $$lcssa251;$nBytesLeft$1$lcssa = $58;
   }
   $59 = ($nBytesLeft$1$lcssa|0)==(0); //@line 356 "mmcmp.cpp"
   if ($59) {
    label = 42;
    break;
   } else {
    $61 = $$lcssa114;$63 = $$lcssa112;$69 = $$lcssa113;$nBytesLeft$2 = $nBytesLeft$1$lcssa;
   }
  } else {
   $61 = $29;$63 = $312;$69 = $28;$nBytesLeft$2 = $nBytesLeft$0145;
  }
  $60 = ($61|0)==(0); //@line 316 "mmcmp.cpp"
  if ($60) {
   $62 = ($63|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
   $64 = ((($63)) + -1|0); //@line 319 "mmcmp.cpp"
   $$80 = $62 ? $63 : $64; //@line 319 "mmcmp.cpp"
   $65 = HEAP8[$$80>>0]|0; //@line 320 "mmcmp.cpp"
   $66 = $65&255; //@line 320 "mmcmp.cpp"
   $67 = $66 >>> 1; //@line 323 "mmcmp.cpp"
   $271 = $66;$274 = $67;$277 = 7;$75 = $$80;
  } else {
   $68 = $69 >>> 1; //@line 323 "mmcmp.cpp"
   $70 = (($61) + -1)|0; //@line 324 "mmcmp.cpp"
   $71 = ($70|0)==(0); //@line 316 "mmcmp.cpp"
   if ($71) {
    $266 = ($63|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
    $267 = ((($63)) + -1|0); //@line 319 "mmcmp.cpp"
    $$80$1 = $266 ? $63 : $267; //@line 319 "mmcmp.cpp"
    $268 = HEAP8[$$80$1>>0]|0; //@line 320 "mmcmp.cpp"
    $269 = $268&255; //@line 320 "mmcmp.cpp"
    $271 = $69;$274 = $269;$277 = 8;$75 = $$80$1;
   } else {
    $271 = $69;$274 = $68;$277 = $70;$75 = $63;
   }
  }
  $270 = $271 << 1; //@line 322 "mmcmp.cpp"
  $272 = $270 & 2; //@line 322 "mmcmp.cpp"
  $273 = $274 & 1; //@line 322 "mmcmp.cpp"
  $275 = $273 | $272; //@line 322 "mmcmp.cpp"
  $276 = $274 >>> 1; //@line 323 "mmcmp.cpp"
  $73 = (($277) + -1)|0; //@line 324 "mmcmp.cpp"
  $278 = (($275) + 1)|0; //@line 359 "mmcmp.cpp"
  $279 = (($pSrc) + ($275)|0); //@line 360 "mmcmp.cpp"
  $117 = HEAP8[$279>>0]|0; //@line 360 "mmcmp.cpp"
  $86 = $117&255; //@line 360 "mmcmp.cpp"
  $280 = ($278|0)==(4); //@line 362 "mmcmp.cpp"
  L62: do {
   if ($280) {
    $72 = ($73|0)==(0); //@line 316 "mmcmp.cpp"
    if ($72) {
     $74 = ($75|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
     $76 = ((($75)) + -1|0); //@line 319 "mmcmp.cpp"
     $$82 = $74 ? $75 : $76; //@line 319 "mmcmp.cpp"
     $77 = HEAP8[$$82>>0]|0; //@line 320 "mmcmp.cpp"
     $78 = $77&255; //@line 320 "mmcmp.cpp"
     $317 = $$82;$80 = $78;$83 = 8;
    } else {
     $317 = $75;$80 = $276;$83 = $73;
    }
    $79 = $80 & 1; //@line 322 "mmcmp.cpp"
    $81 = $80 >>> 1; //@line 323 "mmcmp.cpp"
    $82 = (($83) + -1)|0; //@line 324 "mmcmp.cpp"
    $84 = ($79|0)!=(0); //@line 364 "mmcmp.cpp"
    $85 = $84 ? $86 : 7; //@line 364 "mmcmp.cpp"
    $87 = ($85|0)==(0); //@line 314 "mmcmp.cpp"
    if ($87) {
     $318 = $317;$319 = $81;$320 = $82;$result$0$lcssa$i26 = 0;
    } else {
     $321 = $81;$89 = $82;$91 = $317;$i$02$i21 = 0;$result$01$i22 = 0;
     while(1) {
      $88 = ($89|0)==(0); //@line 316 "mmcmp.cpp"
      if ($88) {
       $90 = ($91|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
       $92 = ((($91)) + -1|0); //@line 319 "mmcmp.cpp"
       $$84 = $90 ? $91 : $92; //@line 319 "mmcmp.cpp"
       $93 = HEAP8[$$84>>0]|0; //@line 320 "mmcmp.cpp"
       $94 = $93&255; //@line 320 "mmcmp.cpp"
       $101 = 8;$322 = $$84;$97 = $94;
      } else {
       $101 = $89;$322 = $91;$97 = $321;
      }
      $95 = $result$01$i22 << 1; //@line 322 "mmcmp.cpp"
      $96 = $97 & 1; //@line 322 "mmcmp.cpp"
      $98 = $96 | $95; //@line 322 "mmcmp.cpp"
      $99 = $97 >>> 1; //@line 323 "mmcmp.cpp"
      $100 = (($101) + -1)|0; //@line 324 "mmcmp.cpp"
      $102 = (($i$02$i21) + 1)|0; //@line 314 "mmcmp.cpp"
      $exitcond$i25 = ($102|0)==($85|0); //@line 314 "mmcmp.cpp"
      if ($exitcond$i25) {
       $318 = $322;$319 = $99;$320 = $100;$result$0$lcssa$i26 = $98;
       break;
      } else {
       $321 = $99;$89 = $100;$91 = $322;$i$02$i21 = $102;$result$01$i22 = $98;
      }
     }
    }
    $105 = $320;$107 = $318;$113 = $319;$n1$0 = 4;
    while(1) {
     $103 = ($n1$0>>>0)<($nBytesLeft$2>>>0); //@line 365 "mmcmp.cpp"
     if (!($103)) {
      $323 = $107;$324 = $113;$325 = $105;$n1$1 = $n1$0;$nofs$0 = $result$0$lcssa$i26;
      break L62;
     }
     $104 = ($105|0)==(0); //@line 316 "mmcmp.cpp"
     if ($104) {
      $106 = ($107|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
      $108 = ((($107)) + -1|0); //@line 319 "mmcmp.cpp"
      $$86 = $106 ? $107 : $108; //@line 319 "mmcmp.cpp"
      $109 = HEAP8[$$86>>0]|0; //@line 320 "mmcmp.cpp"
      $110 = $109&255; //@line 320 "mmcmp.cpp"
      $111 = $110 >>> 1; //@line 323 "mmcmp.cpp"
      $286 = $110;$289 = $111;$293 = 7;$296 = $$86;
     } else {
      $112 = $113 >>> 1; //@line 323 "mmcmp.cpp"
      $114 = (($105) + -1)|0; //@line 324 "mmcmp.cpp"
      $115 = ($114|0)==(0); //@line 316 "mmcmp.cpp"
      if ($115) {
       $281 = ($107|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
       $282 = ((($107)) + -1|0); //@line 319 "mmcmp.cpp"
       $$86$1 = $281 ? $107 : $282; //@line 319 "mmcmp.cpp"
       $283 = HEAP8[$$86$1>>0]|0; //@line 320 "mmcmp.cpp"
       $284 = $283&255; //@line 320 "mmcmp.cpp"
       $286 = $113;$289 = $284;$293 = 8;$296 = $$86$1;
      } else {
       $286 = $113;$289 = $112;$293 = $114;$296 = $107;
      }
     }
     $285 = $286 << 1; //@line 322 "mmcmp.cpp"
     $287 = $285 & 2; //@line 322 "mmcmp.cpp"
     $288 = $289 & 1; //@line 322 "mmcmp.cpp"
     $290 = $288 | $287; //@line 322 "mmcmp.cpp"
     $291 = $289 >>> 1; //@line 323 "mmcmp.cpp"
     $292 = (($293) + -1)|0; //@line 324 "mmcmp.cpp"
     $294 = ($292|0)==(0); //@line 316 "mmcmp.cpp"
     if ($294) {
      $295 = ($296|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
      $297 = ((($296)) + -1|0); //@line 319 "mmcmp.cpp"
      $$86$2 = $295 ? $296 : $297; //@line 319 "mmcmp.cpp"
      $298 = HEAP8[$$86$2>>0]|0; //@line 320 "mmcmp.cpp"
      $299 = $298&255; //@line 320 "mmcmp.cpp"
      $302 = $299;$306 = 8;$329 = $$86$2;
     } else {
      $302 = $291;$306 = $292;$329 = $296;
     }
     $300 = $290 << 1; //@line 322 "mmcmp.cpp"
     $301 = $302 & 1; //@line 322 "mmcmp.cpp"
     $303 = $301 | $300; //@line 322 "mmcmp.cpp"
     $304 = $302 >>> 1; //@line 323 "mmcmp.cpp"
     $305 = (($306) + -1)|0; //@line 324 "mmcmp.cpp"
     $307 = (($303) + ($n1$0))|0; //@line 368 "mmcmp.cpp"
     $308 = ($303|0)==(7); //@line 369 "mmcmp.cpp"
     if ($308) {
      $105 = $305;$107 = $329;$113 = $304;$n1$0 = $307;
     } else {
      $323 = $329;$324 = $304;$325 = $305;$n1$1 = $307;$nofs$0 = $result$0$lcssa$i26;
      break;
     }
    }
   } else {
    $116 = ($117<<24>>24)==(0); //@line 314 "mmcmp.cpp"
    if ($116) {
     $323 = $75;$324 = $276;$325 = $73;$n1$1 = $278;$nofs$0 = 0;
    } else {
     $119 = $73;$121 = $75;$326 = $276;$i$02$i3 = 0;$result$01$i4 = 0;
     while(1) {
      $118 = ($119|0)==(0); //@line 316 "mmcmp.cpp"
      if ($118) {
       $120 = ($121|0)==($pSrc|0); //@line 319 "mmcmp.cpp"
       $122 = ((($121)) + -1|0); //@line 319 "mmcmp.cpp"
       $$88 = $120 ? $121 : $122; //@line 319 "mmcmp.cpp"
       $123 = HEAP8[$$88>>0]|0; //@line 320 "mmcmp.cpp"
       $124 = $123&255; //@line 320 "mmcmp.cpp"
       $127 = $124;$131 = 8;$327 = $$88;
      } else {
       $127 = $326;$131 = $119;$327 = $121;
      }
      $125 = $result$01$i4 << 1; //@line 322 "mmcmp.cpp"
      $126 = $127 & 1; //@line 322 "mmcmp.cpp"
      $128 = $126 | $125; //@line 322 "mmcmp.cpp"
      $129 = $127 >>> 1; //@line 323 "mmcmp.cpp"
      $130 = (($131) + -1)|0; //@line 324 "mmcmp.cpp"
      $132 = (($i$02$i3) + 1)|0; //@line 314 "mmcmp.cpp"
      $exitcond$i7 = ($132|0)==($86|0); //@line 314 "mmcmp.cpp"
      if ($exitcond$i7) {
       $323 = $327;$324 = $129;$325 = $130;$n1$1 = $278;$nofs$0 = $128;
       break;
      } else {
       $119 = $130;$121 = $327;$326 = $129;$i$02$i3 = $132;$result$01$i4 = $128;
      }
     }
    }
   }
  } while(0);
  $i3$0144 = 0;$nBytesLeft$3143 = $nBytesLeft$2;
  while(1) {
   $135 = (($nBytesLeft$3143) + ($nofs$0))|0; //@line 377 "mmcmp.cpp"
   $136 = ($135>>>0)<($nDstLen>>>0); //@line 377 "mmcmp.cpp"
   if ($136) {
    $137 = (($pDst) + ($135)|0); //@line 377 "mmcmp.cpp"
    $138 = HEAP8[$137>>0]|0; //@line 377 "mmcmp.cpp"
    $139 = $138&255; //@line 377 "mmcmp.cpp"
    $141 = $139;
   } else {
    $141 = 0;
   }
   $140 = $141&255; //@line 377 "mmcmp.cpp"
   $142 = (($nBytesLeft$3143) + -1)|0; //@line 377 "mmcmp.cpp"
   $143 = (($pDst) + ($142)|0); //@line 377 "mmcmp.cpp"
   HEAP8[$143>>0] = $140; //@line 377 "mmcmp.cpp"
   $144 = ($142|0)==(0); //@line 378 "mmcmp.cpp"
   $134 = (($i3$0144) + 1)|0; //@line 375 "mmcmp.cpp"
   if ($144) {
    label = 42;
    break L10;
   }
   $133 = ($134>>>0)>($n1$1>>>0); //@line 375 "mmcmp.cpp"
   if ($133) {
    $20 = $325;$22 = $323;$311 = $324;$nBytesLeft$0145 = $142;
    continue L10;
   } else {
    $i3$0144 = $134;$nBytesLeft$3143 = $142;
   }
  }
 }
 if ((label|0) == 42) {
  return; //@line 382 "mmcmp.cpp"
 }
}
function __ZN10CSoundFile6ReadXMEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$2 = 0, $$4 = 0, $$58 = 0, $$59 = 0, $$60 = 0, $$lcssa = 0, $$lcssa302 = 0, $$lobit = 0, $$lobit73 = 0, $$lobit74 = 0, $$lobit75 = 0, $$off = 0, $$off71 = 0, $$off72 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre252 = 0, $$pre254 = 0, $$sum = 0;
 var $$sum17 = 0, $$sum18 = 0, $$sum19 = 0, $$sum20 = 0, $$sum21 = 0, $$sum22 = 0, $$sum24 = 0, $$sum25 = 0, $$sum26 = 0, $$sum27 = 0, $$sum28 = 0, $$sum29 = 0, $$sum30 = 0, $$sum31 = 0, $$sum32 = 0, $$sum33 = 0, $$sum37 = 0, $$sum38 = 0, $$sum39 = 0, $$sum40 = 0;
 var $$sum41 = 0, $$sum42 = 0, $$sum43 = 0, $$sum44 = 0, $$sum45 = 0, $$sum46 = 0, $$sum47 = 0, $$sum48 = 0, $$sum49 = 0, $$xmss$sroa$0$1 = 0, $$xmss$sroa$0$1$ = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0;
 var $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0;
 var $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0;
 var $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0;
 var $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0;
 var $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0;
 var $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0;
 var $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0;
 var $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0;
 var $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0;
 var $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0;
 var $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0;
 var $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0;
 var $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0;
 var $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0;
 var $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0;
 var $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0;
 var $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0;
 var $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0;
 var $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0;
 var $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0;
 var $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0;
 var $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0;
 var $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0;
 var $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0;
 var $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0;
 var $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0;
 var $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $InstUsed = 0, $chn$0187 = 0, $clrs$0131 = 0, $clrs5$0140 = 0, $cond = 0, $cond$us = 0, $cond124 = 0, $dwMemPos$0$lcssa = 0, $dwMemPos$0$pn = 0, $dwMemPos$0$pn$lcssa = 0, $dwMemPos$0199 = 0, $dwMemPos$10 = 0, $dwMemPos$11 = 0, $dwMemPos$12 = 0, $dwMemPos$13 = 0, $dwMemPos$2 = 0;
 var $dwMemPos$3$lcssa = 0, $dwMemPos$3179 = 0, $dwMemPos$4$lcssa = 0, $dwMemPos$4174 = 0, $dwMemPos$5 = 0, $dwMemPos$6161 = 0, $dwMemPos$7164 = 0, $dwMemPos$9$ph67 = 0, $dwSize$0 = 0, $dwSize$0$in$in = 0, $exitcond = 0, $exitcond233 = 0, $exitcond236 = 0, $exitcond237 = 0, $exitcond240 = 0, $exitcond241 = 0, $exitcond242 = 0, $exitcond243 = 0, $exitcond244 = 0, $exitcond245 = 0;
 var $exitcond246 = 0, $exitcond247 = 0, $exitcond248 = 0, $exitcond249 = 0, $flags = 0, $i$0209 = 0, $i$1208 = 0, $i$2206 = 0, $i$3203 = 0, $i1$0211 = 0, $i13$0118 = 0, $ienv$0158 = 0, $iext$0133 = 0, $iext$0133$lcssa = 0, $indvars$iv$next239 = 0, $indvars$iv234 = 0, $indvars$iv238 = 0, $ins$0160 = 0, $ipat$0198 = 0, $ismpd$0165 = 0;
 var $j$0207 = 0, $j$1 = 0, $j$2205 = 0, $j$3 = 0, $j2$0196 = 0, $j2$1$lcssa = 0, $j2$10 = 0, $j2$10$ = 0, $j2$11 = 0, $j2$1186 = 0, $j2$2 = 0, $j2$3 = 0, $j2$4 = 0, $j2$5 = 0, $j2$6 = 0, $j2$8 = 0, $j2$8$ = 0, $j8$0159 = 0, $ks$0130 = 0, $ks7$0139 = 0;
 var $mapchk$0136 = 0, $n$0$lcssa105 = 0, $n$0125 = 0, $n$0125$us = 0, $n$1 = 0, $n$1$newsamples$0 = 0, $n$1$newsamples$0$lcssa = 0, $newsamples$0147 = 0, $or$cond = 0, $or$cond14 = 0, $or$cond261 = 0, $or$cond53 = 0, $or$cond55 = 0, $or$cond57 = 0, $or$cond62 = 0, $or$cond63 = 0, $or$cond77 = 0, $or$cond8 = 0, $p$1$ph = 0, $p$1195 = 0;
 var $p$2$lcssa = 0, $p$2185 = 0, $p$3 = 0, $pattern_map = 0, $row$0197 = 0, $rows$0 = 0, $samplemap = 0, $samples_used = 0, $samplesize = 0, $storemerge34 = 0, $storemerge34$in = 0, $switch$split2D = 0, $unused_samples$0170 = 0, $unused_samples$1143 = 0, $unused_samples$2 = 0, $unused_samples$3 = 0, $unused_samples$3$lcssa = 0, $unused_samples$4$ph66 = 0, $vol$0 = 0, $vol$1 = 0;
 var $vol$1$off = 0, $xmapchk$0119$us = 0, $xmsh = 0, $xmss$sroa$0$0 = 0, $xmss$sroa$0$1 = 0, $xmss$sroa$17$0 = 0, $xmss$sroa$17$1 = 0, $xmss$sroa$26$0 = 0, $xmss$sroa$26$1 = 0, $xmss$sroa$40$0 = 0, $xmss$sroa$40$1 = 0, $xmss$sroa$55 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1312|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $xmsh = sp + 256|0;
 $xmss$sroa$55 = sp + 1280|0;
 $InstUsed = sp + 1024|0;
 $pattern_map = sp + 768|0;
 $samples_used = sp + 528|0;
 $flags = sp + 496|0;
 $samplesize = sp + 128|0;
 $samplemap = sp;
 $0 = ((($this)) + 59664|0); //@line 90 "load_xm.cpp"
 HEAP32[$0>>2] = 0; //@line 90 "load_xm.cpp"
 $1 = ($lpStream|0)==(0|0); //@line 91 "load_xm.cpp"
 $2 = ($dwMemLength>>>0)<(512); //@line 91 "load_xm.cpp"
 $or$cond = $1 | $2; //@line 91 "load_xm.cpp"
 if ($or$cond) {
  $$2 = 0;
  STACKTOP = sp;return ($$2|0); //@line 581 "load_xm.cpp"
 }
 $3 = (_strncasecmp($lpStream,529876,15)|0); //@line 92 "load_xm.cpp"
 $4 = ($3|0)==(0); //@line 92 "load_xm.cpp"
 if (!($4)) {
  $$2 = 0;
  STACKTOP = sp;return ($$2|0); //@line 581 "load_xm.cpp"
 }
 $5 = ((($this)) + 59804|0); //@line 94 "load_xm.cpp"
 $6 = ((($lpStream)) + 17|0); //@line 94 "load_xm.cpp"
 dest=$5; src=$6; stop=dest+20|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 94 "load_xm.cpp"
 $7 = ((($lpStream)) + 60|0); //@line 95 "load_xm.cpp"
 $8 = HEAPU8[$7>>0]|(HEAPU8[$7+1>>0]<<8)|(HEAPU8[$7+2>>0]<<16)|(HEAPU8[$7+3>>0]<<24); //@line 95 "load_xm.cpp"
 $9 = ((($lpStream)) + 64|0); //@line 95 "load_xm.cpp"
 $10 = HEAPU8[$9>>0]|(HEAPU8[$9+1>>0]<<8); //@line 95 "load_xm.cpp"
 $11 = ((($lpStream)) + 66|0); //@line 95 "load_xm.cpp"
 $12 = HEAPU8[$11>>0]|(HEAPU8[$11+1>>0]<<8); //@line 95 "load_xm.cpp"
 $13 = ((($lpStream)) + 68|0); //@line 95 "load_xm.cpp"
 $14 = HEAPU8[$13>>0]|(HEAPU8[$13+1>>0]<<8); //@line 95 "load_xm.cpp"
 $15 = ((($lpStream)) + 70|0); //@line 95 "load_xm.cpp"
 $16 = HEAPU8[$15>>0]|(HEAPU8[$15+1>>0]<<8); //@line 95 "load_xm.cpp"
 $17 = ((($lpStream)) + 72|0); //@line 95 "load_xm.cpp"
 $18 = HEAPU8[$17>>0]|(HEAPU8[$17+1>>0]<<8); //@line 95 "load_xm.cpp"
 $19 = ((($lpStream)) + 74|0); //@line 95 "load_xm.cpp"
 $20 = HEAPU8[$19>>0]|(HEAPU8[$19+1>>0]<<8); //@line 95 "load_xm.cpp"
 $21 = ((($lpStream)) + 76|0); //@line 95 "load_xm.cpp"
 $22 = HEAPU8[$21>>0]|(HEAPU8[$21+1>>0]<<8); //@line 95 "load_xm.cpp"
 $23 = ((($lpStream)) + 78|0); //@line 95 "load_xm.cpp"
 $24 = HEAPU8[$23>>0]|(HEAPU8[$23+1>>0]<<8); //@line 95 "load_xm.cpp"
 $25 = ($10<<16>>16)==(0); //@line 98 "load_xm.cpp"
 if ($25) {
  $$2 = 0;
  STACKTOP = sp;return ($$2|0); //@line 581 "load_xm.cpp"
 }
 $26 = $10&65535; //@line 98 "load_xm.cpp"
 $27 = ($10&65535)>(256); //@line 98 "load_xm.cpp"
 $28 = (($14) + -1)<<16>>16;
 $29 = ($28&65535)>(63); //@line 101 "load_xm.cpp"
 $or$cond77 = $27 | $29; //@line 98 "load_xm.cpp"
 if ($or$cond77) {
  $$2 = 0;
  STACKTOP = sp;return ($$2|0); //@line 581 "load_xm.cpp"
 }
 $30 = $14&65535; //@line 101 "load_xm.cpp"
 $31 = ((($this)) + 59680|0); //@line 102 "load_xm.cpp"
 HEAP32[$31>>2] = 4; //@line 102 "load_xm.cpp"
 $32 = ((($this)) + 59764|0); //@line 103 "load_xm.cpp"
 HEAP32[$32>>2] = 27; //@line 103 "load_xm.cpp"
 $33 = ((($this)) + 59768|0); //@line 104 "load_xm.cpp"
 HEAP32[$33>>2] = 54784; //@line 104 "load_xm.cpp"
 HEAP32[$0>>2] = $30; //@line 105 "load_xm.cpp"
 $34 = ($12&65535)<($10&65535); //@line 106 "load_xm.cpp"
 if ($34) {
  $35 = $12&65535; //@line 106 "load_xm.cpp"
  $36 = ((($this)) + 59736|0); //@line 106 "load_xm.cpp"
  HEAP32[$36>>2] = $35; //@line 106 "load_xm.cpp"
 }
 $37 = ($16&65535)>(256); //@line 108 "load_xm.cpp"
 $38 = ($18&65535)>(239); //@line 110 "load_xm.cpp"
 $39 = $18&65535; //@line 111 "load_xm.cpp"
 $40 = $38 ? 239 : $39; //@line 111 "load_xm.cpp"
 $41 = ((($this)) + 59688|0); //@line 111 "load_xm.cpp"
 HEAP32[$41>>2] = $40; //@line 111 "load_xm.cpp"
 $42 = ((($this)) + 59684|0); //@line 112 "load_xm.cpp"
 HEAP32[$42>>2] = 0; //@line 112 "load_xm.cpp"
 $43 = $20&65535; //@line 114 "load_xm.cpp"
 $44 = $43 & 1; //@line 114 "load_xm.cpp"
 $45 = ($44|0)==(0); //@line 114 "load_xm.cpp"
 if (!($45)) {
  $46 = ((($this)) + 59660|0); //@line 114 "load_xm.cpp"
  $47 = HEAP32[$46>>2]|0; //@line 114 "load_xm.cpp"
  $48 = $47 | 16; //@line 114 "load_xm.cpp"
  HEAP32[$46>>2] = $48; //@line 114 "load_xm.cpp"
 }
 $49 = $43 & 4096; //@line 115 "load_xm.cpp"
 $50 = ($49|0)==(0); //@line 115 "load_xm.cpp"
 if (!($50)) {
  $51 = ((($this)) + 59660|0); //@line 115 "load_xm.cpp"
  $52 = HEAP32[$51>>2]|0; //@line 115 "load_xm.cpp"
  $53 = $52 | 32768; //@line 115 "load_xm.cpp"
  HEAP32[$51>>2] = $53; //@line 115 "load_xm.cpp"
 }
 $$off71 = (($24) + -32)<<16>>16; //@line 118 "load_xm.cpp"
 $54 = ($$off71&65535)<(224); //@line 118 "load_xm.cpp"
 if ($54) {
  $55 = $24&65535; //@line 118 "load_xm.cpp"
  $56 = ((($this)) + 59652|0); //@line 118 "load_xm.cpp"
  HEAP32[$56>>2] = $55; //@line 118 "load_xm.cpp"
 }
 $$off72 = (($22) + -1)<<16>>16; //@line 119 "load_xm.cpp"
 $57 = ($$off72&65535)<(39); //@line 119 "load_xm.cpp"
 if ($57) {
  $58 = $22&65535; //@line 119 "load_xm.cpp"
  $59 = ((($this)) + 59648|0); //@line 119 "load_xm.cpp"
  HEAP32[$59>>2] = $58; //@line 119 "load_xm.cpp"
 }
 $60 = ((($this)) + 53344|0); //@line 120 "load_xm.cpp"
 $61 = ((($lpStream)) + 80|0); //@line 120 "load_xm.cpp"
 _memcpy(($60|0),($61|0),($26|0))|0; //@line 120 "load_xm.cpp"
 _memset(($InstUsed|0),0,256)|0; //@line 121 "load_xm.cpp"
 $62 = $16&65535; //@line 122 "load_xm.cpp"
 $63 = $37 ? 256 : $62; //@line 122 "load_xm.cpp"
 $64 = ($63>>>0)>(240); //@line 122 "load_xm.cpp"
 if ($64) {
  $i$0209 = 0;
  while(1) {
   $65 = (((($this)) + 53344|0) + ($i$0209)|0); //@line 127 "load_xm.cpp"
   $66 = HEAP8[$65>>0]|0; //@line 127 "load_xm.cpp"
   $67 = $66&255; //@line 127 "load_xm.cpp"
   $68 = ($67>>>0)<($63>>>0); //@line 127 "load_xm.cpp"
   if ($68) {
    $69 = (($InstUsed) + ($67)|0); //@line 127 "load_xm.cpp"
    HEAP8[$69>>0] = 1; //@line 127 "load_xm.cpp"
   }
   $70 = (($i$0209) + 1)|0; //@line 125 "load_xm.cpp"
   $exitcond248 = ($70|0)==($26|0); //@line 125 "load_xm.cpp"
   if ($exitcond248) {
    $i$1208 = 0;$j$0207 = 0;
    break;
   } else {
    $i$0209 = $70;
   }
  }
  while(1) {
   $71 = (($InstUsed) + ($i$1208)|0); //@line 132 "load_xm.cpp"
   $72 = HEAP8[$71>>0]|0; //@line 132 "load_xm.cpp"
   $73 = ($72<<24>>24)==(0); //@line 132 "load_xm.cpp"
   if ($73) {
    $j$1 = $j$0207;
   } else {
    $74 = (($j$0207) + 1)|0; //@line 132 "load_xm.cpp"
    $75 = $j$0207&255; //@line 132 "load_xm.cpp"
    $76 = (($pattern_map) + ($i$1208)|0); //@line 132 "load_xm.cpp"
    HEAP8[$76>>0] = $75; //@line 132 "load_xm.cpp"
    $j$1 = $74;
   }
   $77 = (($i$1208) + 1)|0; //@line 130 "load_xm.cpp"
   $exitcond247 = ($77|0)==(256); //@line 130 "load_xm.cpp"
   if ($exitcond247) {
    $i$2206 = 0;$j$2205 = $j$1;
    break;
   } else {
    $i$1208 = $77;$j$0207 = $j$1;
   }
  }
  while(1) {
   $78 = (($InstUsed) + ($i$2206)|0); //@line 136 "load_xm.cpp"
   $79 = HEAP8[$78>>0]|0; //@line 136 "load_xm.cpp"
   $80 = ($79<<24>>24)==(0); //@line 136 "load_xm.cpp"
   if ($80) {
    $81 = ($j$2205>>>0)<(240); //@line 138 "load_xm.cpp"
    $82 = $j$2205&255; //@line 138 "load_xm.cpp"
    $83 = $81 ? $82 : -2; //@line 138 "load_xm.cpp"
    $84 = (($pattern_map) + ($i$2206)|0); //@line 138 "load_xm.cpp"
    HEAP8[$84>>0] = $83; //@line 138 "load_xm.cpp"
    $85 = (($j$2205) + 1)|0; //@line 139 "load_xm.cpp"
    $j$3 = $85;
   } else {
    $j$3 = $j$2205;
   }
   $86 = (($i$2206) + 1)|0; //@line 134 "load_xm.cpp"
   $exitcond246 = ($86|0)==(256); //@line 134 "load_xm.cpp"
   if ($exitcond246) {
    $i$3203 = 0;
    break;
   } else {
    $i$2206 = $86;$j$2205 = $j$3;
   }
  }
  while(1) {
   $87 = (((($this)) + 53344|0) + ($i$3203)|0); //@line 144 "load_xm.cpp"
   $88 = HEAP8[$87>>0]|0; //@line 144 "load_xm.cpp"
   $89 = $88&255; //@line 144 "load_xm.cpp"
   $90 = (($pattern_map) + ($89)|0); //@line 144 "load_xm.cpp"
   $91 = HEAP8[$90>>0]|0; //@line 144 "load_xm.cpp"
   HEAP8[$87>>0] = $91; //@line 144 "load_xm.cpp"
   $92 = (($i$3203) + 1)|0; //@line 142 "load_xm.cpp"
   $exitcond245 = ($92|0)==($26|0); //@line 142 "load_xm.cpp"
   if ($exitcond245) {
    break;
   } else {
    $i$3203 = $92;
   }
  }
 } else {
  $i1$0211 = 0;
  while(1) {
   $93 = $i1$0211&255; //@line 148 "load_xm.cpp"
   $94 = (($pattern_map) + ($i1$0211)|0); //@line 148 "load_xm.cpp"
   HEAP8[$94>>0] = $93; //@line 148 "load_xm.cpp"
   $95 = (($i1$0211) + 1)|0; //@line 148 "load_xm.cpp"
   $exitcond249 = ($95|0)==(256); //@line 148 "load_xm.cpp"
   if ($exitcond249) {
    break;
   } else {
    $i1$0211 = $95;
   }
  }
 }
 _memset(($InstUsed|0),0,256)|0; //@line 150 "load_xm.cpp"
 $96 = (($8) + 68)|0; //@line 152 "load_xm.cpp"
 $97 = ($96>>>0)<($dwMemLength>>>0); //@line 152 "load_xm.cpp"
 if (!($97)) {
  $$2 = 1;
  STACKTOP = sp;return ($$2|0); //@line 581 "load_xm.cpp"
 }
 $98 = (($8) + 60)|0; //@line 151 "load_xm.cpp"
 $99 = ($63|0)==(0); //@line 155 "load_xm.cpp"
 L52: do {
  if ($99) {
   $dwMemPos$0$lcssa = $98;
  } else {
   $dwMemPos$0199 = $98;$ipat$0198 = 0;
   while(1) {
    $102 = (($pattern_map) + ($ipat$0198)|0); //@line 157 "load_xm.cpp"
    $103 = HEAP8[$102>>0]|0; //@line 157 "load_xm.cpp"
    $dwMemPos$0$pn = $dwMemPos$0199;
    while(1) {
     $dwSize$0$in$in = (($lpStream) + ($dwMemPos$0$pn)|0); //@line 160 "load_xm.cpp"
     $dwSize$0 = HEAP32[$dwSize$0$in$in>>2]|0; //@line 160 "load_xm.cpp"
     $104 = (($dwSize$0) + ($dwMemPos$0$pn))|0; //@line 161 "load_xm.cpp"
     $105 = ($104>>>0)>=($dwMemLength>>>0); //@line 161 "load_xm.cpp"
     $106 = ($dwSize$0>>>0)>(255); //@line 161 "load_xm.cpp"
     $or$cond53 = $106 | $105; //@line 161 "load_xm.cpp"
     $107 = (($dwMemPos$0$pn) + 4)|0;
     $108 = ($107>>>0)<($dwMemLength>>>0); //@line 163 "load_xm.cpp"
     $or$cond55 = $108 & $or$cond53; //@line 161 "load_xm.cpp"
     $109 = (($dwMemPos$0$pn) + 1)|0; //@line 164 "load_xm.cpp"
     if ($or$cond55) {
      $dwMemPos$0$pn = $109;
     } else {
      $$lcssa302 = $104;$dwMemPos$0$pn$lcssa = $dwMemPos$0$pn;
      break;
     }
    }
    $110 = $103&255; //@line 157 "load_xm.cpp"
    $$sum37 = (($dwMemPos$0$pn$lcssa) + 5)|0; //@line 167 "load_xm.cpp"
    $111 = (($lpStream) + ($$sum37)|0); //@line 167 "load_xm.cpp"
    $112 = HEAP16[$111>>1]|0; //@line 167 "load_xm.cpp"
    $113 = (($112) + -1)<<16>>16; //@line 168 "load_xm.cpp"
    $114 = ($113&65535)>(255); //@line 168 "load_xm.cpp"
    $rows$0 = $114 ? 64 : $112; //@line 168 "load_xm.cpp"
    $$sum38 = (($dwMemPos$0$pn$lcssa) + 7)|0; //@line 169 "load_xm.cpp"
    $115 = (($lpStream) + ($$sum38)|0); //@line 169 "load_xm.cpp"
    $116 = HEAP16[$115>>1]|0; //@line 169 "load_xm.cpp"
    $117 = (($$lcssa302) + 4)|0; //@line 170 "load_xm.cpp"
    $118 = ($117>>>0)>($dwMemLength>>>0); //@line 170 "load_xm.cpp"
    if ($118) {
     $$2 = 1;
     label = 214;
     break;
    }
    $119 = $116&65535; //@line 172 "load_xm.cpp"
    $120 = (($119) + ($$lcssa302))|0; //@line 172 "load_xm.cpp"
    $121 = (($120) + 4)|0; //@line 172 "load_xm.cpp"
    $122 = ($121>>>0)>($dwMemLength>>>0); //@line 172 "load_xm.cpp"
    if ($122) {
     $$2 = 1;
     label = 214;
     break;
    }
    $123 = ($103&255)<(240); //@line 174 "load_xm.cpp"
    if ($123) {
     $124 = (((($this)) + 52864|0) + ($110<<1)|0); //@line 176 "load_xm.cpp"
     HEAP16[$124>>1] = $rows$0; //@line 176 "load_xm.cpp"
     $125 = $rows$0&65535; //@line 177 "load_xm.cpp"
     $126 = HEAP32[$0>>2]|0; //@line 177 "load_xm.cpp"
     $127 = (__ZN10CSoundFile15AllocatePatternEjj($125,$126)|0); //@line 177 "load_xm.cpp"
     $128 = (((($this)) + 51904|0) + ($110<<2)|0); //@line 177 "load_xm.cpp"
     HEAP32[$128>>2] = $127; //@line 177 "load_xm.cpp"
     $129 = ($127|0)==(0|0); //@line 177 "load_xm.cpp"
     if ($129) {
      $$2 = 1;
      label = 214;
      break;
     }
     $130 = ($116<<16>>16)==(0); //@line 178 "load_xm.cpp"
     if ($130) {
      $dwMemPos$2 = $$lcssa302;
     } else {
      $p$1$ph = $127;
      label = 37;
     }
    } else {
     $p$1$ph = 0;
     label = 37;
    }
    if ((label|0) == 37) {
     label = 0;
     $131 = ($rows$0<<16>>16)==(0); //@line 183 "load_xm.cpp"
     if ($131) {
      $dwMemPos$2 = $120;
     } else {
      $132 = (($$lcssa302) + 2)|0; //@line 201 "load_xm.cpp"
      $133 = (($$lcssa302) + 3)|0; //@line 202 "load_xm.cpp"
      $134 = $rows$0&65535;
      $j2$0196 = 0;$p$1195 = $p$1$ph;$row$0197 = 0;
      while(1) {
       $135 = HEAP32[$0>>2]|0; //@line 185 "load_xm.cpp"
       $136 = ($135|0)==(0); //@line 185 "load_xm.cpp"
       L68: do {
        if ($136) {
         $j2$1$lcssa = $j2$0196;$p$2$lcssa = $p$1195;
        } else {
         $692 = $135;$chn$0187 = 0;$j2$1186 = $j2$0196;$p$2185 = $p$1195;
         while(1) {
          $137 = ($p$2185|0)!=(0|0); //@line 187 "load_xm.cpp"
          $138 = ($j2$1186>>>0)<($119>>>0); //@line 187 "load_xm.cpp"
          $or$cond57 = $137 & $138; //@line 187 "load_xm.cpp"
          do {
           if ($or$cond57) {
            $139 = (($j2$1186) + 1)|0; //@line 189 "load_xm.cpp"
            $$sum40 = (($j2$1186) + ($$lcssa302))|0; //@line 189 "load_xm.cpp"
            $140 = (($lpStream) + ($$sum40)|0); //@line 189 "load_xm.cpp"
            $141 = HEAP8[$140>>0]|0; //@line 189 "load_xm.cpp"
            $142 = $141&255; //@line 191 "load_xm.cpp"
            $143 = $142 & 128; //@line 191 "load_xm.cpp"
            $144 = ($143|0)==(0); //@line 191 "load_xm.cpp"
            do {
             if ($144) {
              HEAP8[$p$2185>>0] = $141; //@line 200 "load_xm.cpp"
              $$sum41 = (($139) + ($$lcssa302))|0; //@line 201 "load_xm.cpp"
              $174 = (($lpStream) + ($$sum41)|0); //@line 201 "load_xm.cpp"
              $175 = HEAP8[$174>>0]|0; //@line 201 "load_xm.cpp"
              $176 = ((($p$2185)) + 1|0); //@line 201 "load_xm.cpp"
              HEAP8[$176>>0] = $175; //@line 201 "load_xm.cpp"
              $$sum42 = (($132) + ($j2$1186))|0; //@line 202 "load_xm.cpp"
              $177 = (($lpStream) + ($$sum42)|0); //@line 202 "load_xm.cpp"
              $178 = HEAP8[$177>>0]|0; //@line 202 "load_xm.cpp"
              $179 = $178&255; //@line 202 "load_xm.cpp"
              $$sum43 = (($133) + ($j2$1186))|0; //@line 203 "load_xm.cpp"
              $180 = (($lpStream) + ($$sum43)|0); //@line 203 "load_xm.cpp"
              $181 = HEAP8[$180>>0]|0; //@line 203 "load_xm.cpp"
              $182 = ((($p$2185)) + 3|0); //@line 203 "load_xm.cpp"
              HEAP8[$182>>0] = $181; //@line 203 "load_xm.cpp"
              $183 = (($j2$1186) + 5)|0; //@line 204 "load_xm.cpp"
              $$sum44 = (($117) + ($j2$1186))|0; //@line 204 "load_xm.cpp"
              $184 = (($lpStream) + ($$sum44)|0); //@line 204 "load_xm.cpp"
              $185 = HEAP8[$184>>0]|0; //@line 204 "load_xm.cpp"
              $186 = ((($p$2185)) + 5|0); //@line 204 "load_xm.cpp"
              HEAP8[$186>>0] = $185; //@line 204 "load_xm.cpp"
              $j2$6 = $183;$vol$1 = $179;
             } else {
              $145 = $142 & 1; //@line 193 "load_xm.cpp"
              $146 = ($145|0)==(0); //@line 193 "load_xm.cpp"
              if ($146) {
               $j2$2 = $139;
              } else {
               $147 = (($j2$1186) + 2)|0; //@line 193 "load_xm.cpp"
               $$sum49 = (($139) + ($$lcssa302))|0; //@line 193 "load_xm.cpp"
               $148 = (($lpStream) + ($$sum49)|0); //@line 193 "load_xm.cpp"
               $149 = HEAP8[$148>>0]|0; //@line 193 "load_xm.cpp"
               HEAP8[$p$2185>>0] = $149; //@line 193 "load_xm.cpp"
               $j2$2 = $147;
              }
              $150 = $142 & 2; //@line 194 "load_xm.cpp"
              $151 = ($150|0)==(0); //@line 194 "load_xm.cpp"
              if ($151) {
               $j2$3 = $j2$2;
              } else {
               $152 = (($j2$2) + 1)|0; //@line 194 "load_xm.cpp"
               $$sum48 = (($j2$2) + ($$lcssa302))|0; //@line 194 "load_xm.cpp"
               $153 = (($lpStream) + ($$sum48)|0); //@line 194 "load_xm.cpp"
               $154 = HEAP8[$153>>0]|0; //@line 194 "load_xm.cpp"
               $155 = ((($p$2185)) + 1|0); //@line 194 "load_xm.cpp"
               HEAP8[$155>>0] = $154; //@line 194 "load_xm.cpp"
               $j2$3 = $152;
              }
              $156 = $142 & 4; //@line 195 "load_xm.cpp"
              $157 = ($156|0)==(0); //@line 195 "load_xm.cpp"
              if ($157) {
               $j2$4 = $j2$3;$vol$0 = 0;
              } else {
               $158 = (($j2$3) + 1)|0; //@line 195 "load_xm.cpp"
               $$sum47 = (($j2$3) + ($$lcssa302))|0; //@line 195 "load_xm.cpp"
               $159 = (($lpStream) + ($$sum47)|0); //@line 195 "load_xm.cpp"
               $160 = HEAP8[$159>>0]|0; //@line 195 "load_xm.cpp"
               $161 = $160&255; //@line 195 "load_xm.cpp"
               $j2$4 = $158;$vol$0 = $161;
              }
              $162 = $142 & 8; //@line 196 "load_xm.cpp"
              $163 = ($162|0)==(0); //@line 196 "load_xm.cpp"
              if ($163) {
               $j2$5 = $j2$4;
              } else {
               $164 = (($j2$4) + 1)|0; //@line 196 "load_xm.cpp"
               $$sum46 = (($j2$4) + ($$lcssa302))|0; //@line 196 "load_xm.cpp"
               $165 = (($lpStream) + ($$sum46)|0); //@line 196 "load_xm.cpp"
               $166 = HEAP8[$165>>0]|0; //@line 196 "load_xm.cpp"
               $167 = ((($p$2185)) + 3|0); //@line 196 "load_xm.cpp"
               HEAP8[$167>>0] = $166; //@line 196 "load_xm.cpp"
               $j2$5 = $164;
              }
              $168 = $142 & 16; //@line 197 "load_xm.cpp"
              $169 = ($168|0)==(0); //@line 197 "load_xm.cpp"
              if ($169) {
               $j2$6 = $j2$5;$vol$1 = $vol$0;
               break;
              }
              $170 = (($j2$5) + 1)|0; //@line 197 "load_xm.cpp"
              $$sum45 = (($j2$5) + ($$lcssa302))|0; //@line 197 "load_xm.cpp"
              $171 = (($lpStream) + ($$sum45)|0); //@line 197 "load_xm.cpp"
              $172 = HEAP8[$171>>0]|0; //@line 197 "load_xm.cpp"
              $173 = ((($p$2185)) + 5|0); //@line 197 "load_xm.cpp"
              HEAP8[$173>>0] = $172; //@line 197 "load_xm.cpp"
              $j2$6 = $170;$vol$1 = $vol$0;
             }
            } while(0);
            $187 = HEAP8[$p$2185>>0]|0; //@line 206 "load_xm.cpp"
            $188 = $187&255; //@line 206 "load_xm.cpp"
            L90: do {
             switch ($187<<24>>24) {
             case 97:  {
              HEAP8[$p$2185>>0] = -1; //@line 206 "load_xm.cpp"
              break;
             }
             case 0:  {
              break;
             }
             default: {
              $189 = ($187&255)<(97); //@line 207 "load_xm.cpp"
              if (!($189)) {
               break L90;
              }
              $190 = (($188) + 12)|0; //@line 207 "load_xm.cpp"
              $191 = $190&255; //@line 207 "load_xm.cpp"
              HEAP8[$p$2185>>0] = $191; //@line 207 "load_xm.cpp"
             }
             }
            } while(0);
            $192 = ((($p$2185)) + 3|0); //@line 209 "load_xm.cpp"
            $193 = HEAP8[$192>>0]|0; //@line 209 "load_xm.cpp"
            $194 = ((($p$2185)) + 5|0); //@line 209 "load_xm.cpp"
            $195 = HEAP8[$194>>0]|0; //@line 209 "load_xm.cpp"
            $196 = $195 | $193; //@line 209 "load_xm.cpp"
            $197 = ($196<<24>>24)==(0); //@line 209 "load_xm.cpp"
            if (!($197)) {
             __ZNK10CSoundFile17ConvertModCommandEP11_MODCOMMAND($this,$p$2185); //@line 209 "load_xm.cpp"
            }
            $198 = ((($p$2185)) + 1|0); //@line 210 "load_xm.cpp"
            $199 = HEAP8[$198>>0]|0; //@line 210 "load_xm.cpp"
            switch ($199<<24>>24) {
            case -1:  {
             HEAP8[$198>>0] = 0; //@line 210 "load_xm.cpp"
             break;
            }
            case 0:  {
             break;
            }
            default: {
             $200 = $199&255; //@line 211 "load_xm.cpp"
             $201 = (($InstUsed) + ($200)|0); //@line 211 "load_xm.cpp"
             HEAP8[$201>>0] = 1; //@line 211 "load_xm.cpp"
            }
            }
            $vol$1$off = (($vol$1) + -16)|0; //@line 212 "load_xm.cpp"
            $202 = ($vol$1$off>>>0)<(65); //@line 212 "load_xm.cpp"
            L102: do {
             if ($202) {
              $203 = ((($p$2185)) + 2|0); //@line 214 "load_xm.cpp"
              HEAP8[$203>>0] = 1; //@line 214 "load_xm.cpp"
              $204 = (($vol$1) + 240)|0; //@line 215 "load_xm.cpp"
              $205 = $204&255; //@line 215 "load_xm.cpp"
              $206 = ((($p$2185)) + 4|0); //@line 215 "load_xm.cpp"
              HEAP8[$206>>0] = $205; //@line 215 "load_xm.cpp"
             } else {
              $207 = ($vol$1>>>0)>(95); //@line 217 "load_xm.cpp"
              if (!($207)) {
               break;
              }
              $208 = $vol$1 & 240; //@line 219 "load_xm.cpp"
              $209 = $vol$1 & 15; //@line 220 "load_xm.cpp"
              $210 = $209&255; //@line 221 "load_xm.cpp"
              $211 = ((($p$2185)) + 4|0); //@line 221 "load_xm.cpp"
              HEAP8[$211>>0] = $210; //@line 221 "load_xm.cpp"
              do {
               switch ($208|0) {
               case 96:  {
                $212 = ((($p$2185)) + 2|0); //@line 225 "load_xm.cpp"
                HEAP8[$212>>0] = 4; //@line 225 "load_xm.cpp"
                break L102;
                break;
               }
               case 112:  {
                $213 = ((($p$2185)) + 2|0); //@line 227 "load_xm.cpp"
                HEAP8[$213>>0] = 3; //@line 227 "load_xm.cpp"
                break L102;
                break;
               }
               case 128:  {
                $214 = ((($p$2185)) + 2|0); //@line 229 "load_xm.cpp"
                HEAP8[$214>>0] = 6; //@line 229 "load_xm.cpp"
                break L102;
                break;
               }
               case 144:  {
                $215 = ((($p$2185)) + 2|0); //@line 231 "load_xm.cpp"
                HEAP8[$215>>0] = 5; //@line 231 "load_xm.cpp"
                break L102;
                break;
               }
               case 160:  {
                $216 = ((($p$2185)) + 2|0); //@line 233 "load_xm.cpp"
                HEAP8[$216>>0] = 7; //@line 233 "load_xm.cpp"
                break L102;
                break;
               }
               case 176:  {
                $217 = ((($p$2185)) + 2|0); //@line 235 "load_xm.cpp"
                HEAP8[$217>>0] = 8; //@line 235 "load_xm.cpp"
                break L102;
                break;
               }
               case 192:  {
                $218 = ((($p$2185)) + 2|0); //@line 237 "load_xm.cpp"
                HEAP8[$218>>0] = 2; //@line 237 "load_xm.cpp"
                $219 = $209 << 2; //@line 237 "load_xm.cpp"
                $220 = $219 | 2; //@line 237 "load_xm.cpp"
                $221 = $220&255; //@line 237 "load_xm.cpp"
                HEAP8[$211>>0] = $221; //@line 237 "load_xm.cpp"
                break L102;
                break;
               }
               case 208:  {
                $222 = ((($p$2185)) + 2|0); //@line 239 "load_xm.cpp"
                HEAP8[$222>>0] = 9; //@line 239 "load_xm.cpp"
                break L102;
                break;
               }
               case 224:  {
                $223 = ((($p$2185)) + 2|0); //@line 241 "load_xm.cpp"
                HEAP8[$223>>0] = 10; //@line 241 "load_xm.cpp"
                break L102;
                break;
               }
               case 240:  {
                $224 = ((($p$2185)) + 2|0); //@line 243 "load_xm.cpp"
                HEAP8[$224>>0] = 11; //@line 243 "load_xm.cpp"
                break L102;
                break;
               }
               default: {
                break L102;
               }
               }
              } while(0);
             }
            } while(0);
            $225 = ((($p$2185)) + 6|0); //@line 246 "load_xm.cpp"
            $$pre = HEAP32[$0>>2]|0; //@line 185 "load_xm.cpp"
            $240 = $$pre;$j2$11 = $j2$6;$p$3 = $225;
           } else {
            if (!($138)) {
             $j2$1$lcssa = $j2$1186;$p$2$lcssa = $p$2185;
             break L68;
            }
            $$sum39 = (($j2$1186) + ($$lcssa302))|0; //@line 250 "load_xm.cpp"
            $226 = (($lpStream) + ($$sum39)|0); //@line 250 "load_xm.cpp"
            $227 = HEAP8[$226>>0]|0; //@line 250 "load_xm.cpp"
            $228 = $227&255; //@line 251 "load_xm.cpp"
            $229 = $228 & 128; //@line 251 "load_xm.cpp"
            $230 = ($229|0)==(0); //@line 251 "load_xm.cpp"
            if ($230) {
             $237 = (($j2$1186) + 5)|0; //@line 258 "load_xm.cpp"
             $240 = $692;$j2$11 = $237;$p$3 = $p$2185;
             break;
            } else {
             $231 = $228 & 1; //@line 253 "load_xm.cpp"
             $232 = $228 >>> 1; //@line 254 "load_xm.cpp"
             $$lobit = $232 & 1; //@line 254 "load_xm.cpp"
             $233 = $228 >>> 2; //@line 255 "load_xm.cpp"
             $$lobit73 = $233 & 1; //@line 255 "load_xm.cpp"
             $234 = $228 >>> 3; //@line 256 "load_xm.cpp"
             $$lobit74 = $234 & 1; //@line 256 "load_xm.cpp"
             $235 = $228 >>> 4; //@line 257 "load_xm.cpp"
             $$lobit75 = $235 & 1; //@line 257 "load_xm.cpp"
             $236 = (($j2$1186) + 1)|0; //@line 253 "load_xm.cpp"
             $$ = (($236) + ($231))|0; //@line 253 "load_xm.cpp"
             $j2$8 = (($$) + ($$lobit))|0; //@line 254 "load_xm.cpp"
             $j2$8$ = (($j2$8) + ($$lobit73))|0; //@line 255 "load_xm.cpp"
             $j2$10 = (($j2$8$) + ($$lobit74))|0; //@line 256 "load_xm.cpp"
             $j2$10$ = (($j2$10) + ($$lobit75))|0; //@line 257 "load_xm.cpp"
             $240 = $692;$j2$11 = $j2$10$;$p$3 = $p$2185;
             break;
            }
           }
          } while(0);
          $238 = (($chn$0187) + 1)|0; //@line 185 "load_xm.cpp"
          $239 = ($238>>>0)<($240>>>0); //@line 185 "load_xm.cpp"
          if ($239) {
           $692 = $240;$chn$0187 = $238;$j2$1186 = $j2$11;$p$2185 = $p$3;
          } else {
           $j2$1$lcssa = $j2$11;$p$2$lcssa = $p$3;
           break;
          }
         }
        }
       } while(0);
       $241 = (($row$0197) + 1)|0; //@line 183 "load_xm.cpp"
       $exitcond244 = ($241|0)==($134|0); //@line 183 "load_xm.cpp"
       if ($exitcond244) {
        $dwMemPos$2 = $120;
        break;
       } else {
        $j2$0196 = $j2$1$lcssa;$p$1195 = $p$2$lcssa;$row$0197 = $241;
       }
      }
     }
    }
    $242 = (($ipat$0198) + 1)|0; //@line 155 "load_xm.cpp"
    $243 = ($242>>>0)<($63>>>0); //@line 155 "load_xm.cpp"
    if ($243) {
     $dwMemPos$0199 = $dwMemPos$2;$ipat$0198 = $242;
    } else {
     $dwMemPos$0$lcssa = $dwMemPos$2;
     break L52;
    }
   }
   if ((label|0) == 214) {
    STACKTOP = sp;return ($$2|0); //@line 581 "load_xm.cpp"
   }
  }
 } while(0);
 $100 = (($dwMemPos$0$lcssa) + 4)|0; //@line 265 "load_xm.cpp"
 $101 = ($100>>>0)<($dwMemLength>>>0); //@line 265 "load_xm.cpp"
 L131: do {
  if ($101) {
   $dwMemPos$3179 = $dwMemPos$0$lcssa;
   while(1) {
    $244 = (($lpStream) + ($dwMemPos$3179)|0); //@line 267 "load_xm.cpp"
    $245 = HEAP32[$244>>2]|0; //@line 267 "load_xm.cpp"
    $246 = ($245>>>0)<(768); //@line 268 "load_xm.cpp"
    if ($246) {
     $dwMemPos$3$lcssa = $dwMemPos$3179;
     break L131;
    }
    $247 = (($dwMemPos$3179) + 1)|0; //@line 269 "load_xm.cpp"
    $248 = (($dwMemPos$3179) + 5)|0; //@line 265 "load_xm.cpp"
    $249 = ($248>>>0)<($dwMemLength>>>0); //@line 265 "load_xm.cpp"
    if ($249) {
     $dwMemPos$3179 = $247;
    } else {
     $dwMemPos$3$lcssa = $247;
     break;
    }
   }
  } else {
   $dwMemPos$3$lcssa = $dwMemPos$0$lcssa;
  }
 } while(0);
 _memset(($samples_used|0),0,240)|0; //@line 271 "load_xm.cpp"
 $250 = ($40|0)==(0); //@line 274 "load_xm.cpp"
 L136: do {
  if ($250) {
   $dwMemPos$4$lcssa = $dwMemPos$3$lcssa;
  } else {
   $251 = ((($xmsh)) + 210|0); //@line 297 "load_xm.cpp"
   $252 = ((($xmsh)) + 204|0); //@line 382 "load_xm.cpp"
   $253 = ((($xmsh)) + 196|0); //@line 388 "load_xm.cpp"
   $254 = ((($xmsh)) + 197|0); //@line 389 "load_xm.cpp"
   $255 = ((($xmsh)) + 200|0); //@line 397 "load_xm.cpp"
   $256 = ((($xmsh)) + 206|0); //@line 492 "load_xm.cpp"
   $257 = ((($xmsh)) + 208|0); //@line 494 "load_xm.cpp"
   $dwMemPos$4174 = $dwMemPos$3$lcssa;$indvars$iv238 = 1;$unused_samples$0170 = 0;
   L138: while(1) {
    $258 = (($dwMemPos$4174) + 29)|0; //@line 282 "load_xm.cpp"
    $259 = ($258>>>0)<($dwMemLength>>>0); //@line 282 "load_xm.cpp"
    if (!($259)) {
     break;
    }
    $260 = (($lpStream) + ($dwMemPos$4174)|0); //@line 283 "load_xm.cpp"
    $261 = HEAPU8[$260>>0]|(HEAPU8[$260+1>>0]<<8)|(HEAPU8[$260+2>>0]<<16)|(HEAPU8[$260+3>>0]<<24); //@line 284 "load_xm.cpp"
    $262 = (($261) + ($dwMemPos$4174))|0; //@line 284 "load_xm.cpp"
    $263 = ($262>>>0)>($dwMemLength>>>0); //@line 284 "load_xm.cpp"
    if ($263) {
     break;
    }
    $264 = (__Znwj(632)|0); //@line 285 "load_xm.cpp"
    $265 = (((($this)) + 48640|0) + ($indvars$iv238<<2)|0); //@line 285 "load_xm.cpp"
    HEAP32[$265>>2] = $264; //@line 285 "load_xm.cpp"
    _memset(($264|0),0,632)|0; //@line 286 "load_xm.cpp"
    $266 = ((($264)) + 585|0); //@line 287 "load_xm.cpp"
    $$sum21 = (($dwMemPos$4174) + 4)|0; //@line 287 "load_xm.cpp"
    $267 = (($lpStream) + ($$sum21)|0); //@line 287 "load_xm.cpp"
    dest=$266; src=$267; stop=dest+22|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 287 "load_xm.cpp"
    $$sum22 = (($dwMemPos$4174) + 27)|0; //@line 288 "load_xm.cpp"
    $268 = (($lpStream) + ($$sum22)|0); //@line 288 "load_xm.cpp"
    $269 = HEAP8[$268>>0]|0; //@line 288 "load_xm.cpp"
    $270 = $269&255; //@line 288 "load_xm.cpp"
    $271 = ($269<<24>>24)==(0); //@line 288 "load_xm.cpp"
    if ($271) {
     $276 = ($261|0)==(0); //@line 302 "load_xm.cpp"
     $dwMemPos$5 = $276 ? $258 : $262; //@line 302 "load_xm.cpp"
     $dwMemPos$9$ph67 = $dwMemPos$5;$unused_samples$4$ph66 = $unused_samples$0170;
    } else {
     $272 = (($dwMemPos$4174) + 234)|0; //@line 290 "load_xm.cpp"
     $273 = ($272>>>0)>($dwMemLength>>>0); //@line 290 "load_xm.cpp"
     if ($273) {
      break;
     }
     $274 = (($lpStream) + ($258)|0); //@line 291 "load_xm.cpp"
     _memcpy(($xmsh|0),($274|0),234)|0; //@line 291 "load_xm.cpp"
     dest=$samplemap; stop=dest+128|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 306 "load_xm.cpp"
     $275 = ($269&255)>(32); //@line 307 "load_xm.cpp"
     if ($275) {
      break;
     }
     $277 = HEAP32[$42>>2]|0; //@line 308 "load_xm.cpp"
     $278 = ($indvars$iv238>>>0)>(1); //@line 357 "load_xm.cpp"
     $281 = $277;$indvars$iv234 = 0;$newsamples$0147 = $277;$unused_samples$1143 = $unused_samples$0170;
     while(1) {
      $279 = (($indvars$iv234) + 1)|0; //@line 311 "load_xm.cpp"
      $280 = (($279) + ($281))|0; //@line 311 "load_xm.cpp"
      $282 = ($280>>>0)>(239); //@line 312 "load_xm.cpp"
      L149: do {
       if ($282) {
        $cond124 = ($281|0)==(0); //@line 315 "load_xm.cpp"
        L151: do {
         if (!($cond124)) {
          $283 = ($indvars$iv234|0)==(0); //@line 319 "load_xm.cpp"
          L153: do {
           if ($283) {
            $n$0125 = $281;
            while(1) {
             $293 = (((((($this)) + 33280|0) + ($n$0125<<6)|0)) + 20|0); //@line 317 "load_xm.cpp"
             $294 = HEAP32[$293>>2]|0; //@line 317 "load_xm.cpp"
             $295 = ($294|0)==(0|0); //@line 317 "load_xm.cpp"
             if ($295) {
              $n$0$lcssa105 = $n$0125;
              break L153;
             }
             $305 = (($n$0125) + -1)|0; //@line 334 "load_xm.cpp"
             $cond = ($305|0)==(0); //@line 315 "load_xm.cpp"
             if ($cond) {
              break L151;
             } else {
              $n$0125 = $305;
             }
            }
           } else {
            $n$0125$us = $281;
            while(1) {
             $284 = (((((($this)) + 33280|0) + ($n$0125$us<<6)|0)) + 20|0); //@line 317 "load_xm.cpp"
             $285 = HEAP32[$284>>2]|0; //@line 317 "load_xm.cpp"
             $286 = ($285|0)==(0|0); //@line 317 "load_xm.cpp"
             L159: do {
              if ($286) {
               $xmapchk$0119$us = 0;
               while(1) {
                $289 = (($samplemap) + ($xmapchk$0119$us<<2)|0); //@line 321 "load_xm.cpp"
                $290 = HEAP32[$289>>2]|0; //@line 321 "load_xm.cpp"
                $291 = ($290|0)==($n$0125$us|0); //@line 321 "load_xm.cpp"
                $288 = (($xmapchk$0119$us) + 1)|0; //@line 319 "load_xm.cpp"
                if ($291) {
                 break L159;
                }
                $287 = ($288>>>0)<($indvars$iv234>>>0); //@line 319 "load_xm.cpp"
                if ($287) {
                 $xmapchk$0119$us = $288;
                } else {
                 $n$0$lcssa105 = $n$0125$us;
                 break L153;
                }
               }
              }
             } while(0);
             $292 = (($n$0125$us) + -1)|0; //@line 334 "load_xm.cpp"
             $cond$us = ($292|0)==(0); //@line 315 "load_xm.cpp"
             if ($cond$us) {
              break L151;
             } else {
              $n$0125$us = $292;
             }
            }
           }
          } while(0);
          if ($278) {
           $clrs$0131 = 1;
          } else {
           $n$1 = $n$0$lcssa105;$unused_samples$3 = $unused_samples$1143;
           break L149;
          }
          while(1) {
           $296 = (((($this)) + 48640|0) + ($clrs$0131<<2)|0); //@line 323 "load_xm.cpp"
           $297 = HEAP32[$296>>2]|0; //@line 323 "load_xm.cpp"
           $298 = ($297|0)==(0|0); //@line 323 "load_xm.cpp"
           if (!($298)) {
            $ks$0130 = 0;
            while(1) {
             $299 = (((($297)) + 300|0) + ($ks$0130)|0); //@line 328 "load_xm.cpp"
             $300 = HEAP8[$299>>0]|0; //@line 328 "load_xm.cpp"
             $301 = $300&255; //@line 328 "load_xm.cpp"
             $302 = ($301|0)==($n$0$lcssa105|0); //@line 328 "load_xm.cpp"
             if ($302) {
              HEAP8[$299>>0] = 0; //@line 328 "load_xm.cpp"
             }
             $303 = (($ks$0130) + 1)|0; //@line 326 "load_xm.cpp"
             $exitcond = ($303|0)==(128); //@line 326 "load_xm.cpp"
             if ($exitcond) {
              break;
             } else {
              $ks$0130 = $303;
             }
            }
           }
           $304 = (($clrs$0131) + 1)|0; //@line 323 "load_xm.cpp"
           $exitcond233 = ($304|0)==($indvars$iv238|0); //@line 323 "load_xm.cpp"
           if ($exitcond233) {
            $n$1 = $n$0$lcssa105;$unused_samples$3 = $unused_samples$1143;
            break L149;
           } else {
            $clrs$0131 = $304;
           }
          }
         }
        } while(0);
        $306 = ($unused_samples$1143|0)==(0); //@line 340 "load_xm.cpp"
        if ($306) {
         $307 = (__ZN10CSoundFile19DetectUnusedSamplesEPb($this,$samples_used)|0); //@line 342 "load_xm.cpp"
         $308 = ($307|0)!=(0); //@line 343 "load_xm.cpp"
         $$4 = $308 ? $307 : 65535; //@line 343 "load_xm.cpp"
         $unused_samples$2 = $$4;
        } else {
         $unused_samples$2 = $unused_samples$1143;
        }
        $switch$split2D = ($unused_samples$2|0)<(65535);
        if ($switch$split2D) {
         switch ($unused_samples$2|0) {
         case 0:  {
          $n$1 = 0;$unused_samples$3 = $unused_samples$2;
          break L149;
          break;
         }
         default: {
         }
         }
        } else {
         switch ($unused_samples$2|0) {
         case 65535:  {
          $n$1 = 0;$unused_samples$3 = $unused_samples$2;
          break L149;
          break;
         }
         default: {
         }
         }
        }
        $309 = HEAP32[$42>>2]|0; //@line 347 "load_xm.cpp"
        $310 = ($309|0)==(0); //@line 347 "load_xm.cpp"
        if ($310) {
         $n$1 = 0;$unused_samples$3 = $unused_samples$2;
         break;
        } else {
         $iext$0133 = $309;
        }
        while(1) {
         $311 = (($samples_used) + ($iext$0133)|0); //@line 347 "load_xm.cpp"
         $312 = HEAP8[$311>>0]|0; //@line 347 "load_xm.cpp"
         $313 = ($312<<24>>24)==(0); //@line 347 "load_xm.cpp"
         if ($313) {
          $$lcssa = $311;$iext$0133$lcssa = $iext$0133;
          break;
         }
         $330 = (($iext$0133) + -1)|0; //@line 347 "load_xm.cpp"
         $331 = ($330|0)==(0); //@line 347 "load_xm.cpp"
         if ($331) {
          $n$1 = 0;$unused_samples$3 = $unused_samples$2;
          break L149;
         } else {
          $iext$0133 = $330;
         }
        }
        $314 = (($unused_samples$2) + -1)|0; //@line 349 "load_xm.cpp"
        HEAP8[$$lcssa>>0] = 1; //@line 350 "load_xm.cpp"
        (__ZN10CSoundFile13DestroySampleEj($this,$iext$0133$lcssa)|0); //@line 351 "load_xm.cpp"
        $315 = ($indvars$iv234|0)==(0); //@line 353 "load_xm.cpp"
        if (!($315)) {
         $mapchk$0136 = 0;
         while(1) {
          $316 = (($samplemap) + ($mapchk$0136<<2)|0); //@line 355 "load_xm.cpp"
          $317 = HEAP32[$316>>2]|0; //@line 355 "load_xm.cpp"
          $318 = ($317|0)==($iext$0133$lcssa|0); //@line 355 "load_xm.cpp"
          if ($318) {
           HEAP32[$316>>2] = 0; //@line 355 "load_xm.cpp"
          }
          $319 = (($mapchk$0136) + 1)|0; //@line 353 "load_xm.cpp"
          $exitcond236 = ($319|0)==($indvars$iv234|0); //@line 353 "load_xm.cpp"
          if ($exitcond236) {
           break;
          } else {
           $mapchk$0136 = $319;
          }
         }
        }
        if ($278) {
         $clrs5$0140 = 1;
         while(1) {
          $320 = (((($this)) + 48640|0) + ($clrs5$0140<<2)|0); //@line 357 "load_xm.cpp"
          $321 = HEAP32[$320>>2]|0; //@line 357 "load_xm.cpp"
          $322 = ($321|0)==(0|0); //@line 357 "load_xm.cpp"
          if (!($322)) {
           $ks7$0139 = 0;
           while(1) {
            $323 = (((($321)) + 300|0) + ($ks7$0139)|0); //@line 362 "load_xm.cpp"
            $324 = HEAP8[$323>>0]|0; //@line 362 "load_xm.cpp"
            $325 = $324&255; //@line 362 "load_xm.cpp"
            $326 = ($325|0)==($iext$0133$lcssa|0); //@line 362 "load_xm.cpp"
            if ($326) {
             HEAP8[$323>>0] = 0; //@line 362 "load_xm.cpp"
            }
            $327 = (($ks7$0139) + 1)|0; //@line 360 "load_xm.cpp"
            $exitcond237 = ($327|0)==(128); //@line 360 "load_xm.cpp"
            if ($exitcond237) {
             break;
            } else {
             $ks7$0139 = $327;
            }
           }
          }
          $328 = (($clrs5$0140) + 1)|0; //@line 357 "load_xm.cpp"
          $exitcond240 = ($328|0)==($indvars$iv238|0); //@line 357 "load_xm.cpp"
          if ($exitcond240) {
           break;
          } else {
           $clrs5$0140 = $328;
          }
         }
        }
        $329 = (((($this)) + 33280|0) + ($iext$0133$lcssa<<6)|0); //@line 365 "load_xm.cpp"
        dest=$329; stop=dest+64|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0)); //@line 365 "load_xm.cpp"
        $n$1 = $iext$0133$lcssa;$unused_samples$3 = $314;
       } else {
        $n$1 = $280;$unused_samples$3 = $unused_samples$1143;
       }
      } while(0);
      $332 = ($newsamples$0147>>>0)<($n$1>>>0); //@line 372 "load_xm.cpp"
      $n$1$newsamples$0 = $332 ? $n$1 : $newsamples$0147; //@line 372 "load_xm.cpp"
      $333 = (($samplemap) + ($indvars$iv234<<2)|0); //@line 373 "load_xm.cpp"
      HEAP32[$333>>2] = $n$1; //@line 373 "load_xm.cpp"
      $exitcond241 = ($279|0)==($270|0); //@line 309 "load_xm.cpp"
      if ($exitcond241) {
       $n$1$newsamples$0$lcssa = $n$1$newsamples$0;$unused_samples$3$lcssa = $unused_samples$3;
       break;
      }
      $$pre252 = HEAP32[$42>>2]|0; //@line 311 "load_xm.cpp"
      $281 = $$pre252;$indvars$iv234 = $279;$newsamples$0147 = $n$1$newsamples$0;$unused_samples$1143 = $unused_samples$3;
     }
     HEAP32[$42>>2] = $n$1$newsamples$0$lcssa; //@line 375 "load_xm.cpp"
     $334 = HEAP32[$265>>2]|0; //@line 377 "load_xm.cpp"
     $$sum24 = (($dwMemPos$4174) + 26)|0; //@line 378 "load_xm.cpp"
     $335 = (($lpStream) + ($$sum24)|0); //@line 378 "load_xm.cpp"
     $336 = HEAP8[$335>>0]|0; //@line 378 "load_xm.cpp"
     $337 = ((($334)) + 580|0); //@line 378 "load_xm.cpp"
     HEAP8[$337>>0] = $336; //@line 378 "load_xm.cpp"
     $338 = HEAP16[$251>>1]|0; //@line 379 "load_xm.cpp"
     $339 = $338&65535; //@line 379 "load_xm.cpp"
     HEAP32[$334>>2] = $339; //@line 379 "load_xm.cpp"
     $340 = ((($334)) + 10|0); //@line 380 "load_xm.cpp"
     HEAP16[$340>>1] = 128; //@line 380 "load_xm.cpp"
     $341 = ((($334)) + 584|0); //@line 381 "load_xm.cpp"
     HEAP8[$341>>0] = 60; //@line 381 "load_xm.cpp"
     $342 = HEAP16[$252>>1]|0; //@line 382 "load_xm.cpp"
     $343 = $342&255; //@line 382 "load_xm.cpp"
     $344 = $343 & 1; //@line 382 "load_xm.cpp"
     $345 = ($344<<24>>24)==(0); //@line 382 "load_xm.cpp"
     $346 = ($342&65535) >>> 8;
     $347 = $346&255;
     if (!($345)) {
      $348 = ((($334)) + 4|0); //@line 382 "load_xm.cpp"
      $349 = HEAP32[$348>>2]|0; //@line 382 "load_xm.cpp"
      $350 = $349 | 1; //@line 382 "load_xm.cpp"
      HEAP32[$348>>2] = $350; //@line 382 "load_xm.cpp"
     }
     $351 = $343 & 2; //@line 383 "load_xm.cpp"
     $352 = ($351<<24>>24)==(0); //@line 383 "load_xm.cpp"
     if (!($352)) {
      $353 = ((($334)) + 4|0); //@line 383 "load_xm.cpp"
      $354 = HEAP32[$353>>2]|0; //@line 383 "load_xm.cpp"
      $355 = $354 | 2; //@line 383 "load_xm.cpp"
      HEAP32[$353>>2] = $355; //@line 383 "load_xm.cpp"
     }
     $356 = $343 & 4; //@line 384 "load_xm.cpp"
     $357 = ($356<<24>>24)==(0); //@line 384 "load_xm.cpp"
     if (!($357)) {
      $358 = ((($334)) + 4|0); //@line 384 "load_xm.cpp"
      $359 = HEAP32[$358>>2]|0; //@line 384 "load_xm.cpp"
      $360 = $359 | 4; //@line 384 "load_xm.cpp"
      HEAP32[$358>>2] = $360; //@line 384 "load_xm.cpp"
     }
     $361 = $347 & 1; //@line 385 "load_xm.cpp"
     $362 = ($361<<24>>24)==(0); //@line 385 "load_xm.cpp"
     if (!($362)) {
      $363 = ((($334)) + 4|0); //@line 385 "load_xm.cpp"
      $364 = HEAP32[$363>>2]|0; //@line 385 "load_xm.cpp"
      $365 = $364 | 8; //@line 385 "load_xm.cpp"
      HEAP32[$363>>2] = $365; //@line 385 "load_xm.cpp"
     }
     $366 = $347 & 2; //@line 386 "load_xm.cpp"
     $367 = ($366<<24>>24)==(0); //@line 386 "load_xm.cpp"
     if (!($367)) {
      $368 = ((($334)) + 4|0); //@line 386 "load_xm.cpp"
      $369 = HEAP32[$368>>2]|0; //@line 386 "load_xm.cpp"
      $370 = $369 | 16; //@line 386 "load_xm.cpp"
      HEAP32[$368>>2] = $370; //@line 386 "load_xm.cpp"
     }
     $371 = $347 & 4; //@line 387 "load_xm.cpp"
     $372 = ($371<<24>>24)==(0); //@line 387 "load_xm.cpp"
     if (!($372)) {
      $373 = ((($334)) + 4|0); //@line 387 "load_xm.cpp"
      $374 = HEAP32[$373>>2]|0; //@line 387 "load_xm.cpp"
      $375 = $374 | 32; //@line 387 "load_xm.cpp"
      HEAP32[$373>>2] = $375; //@line 387 "load_xm.cpp"
     }
     $376 = HEAP16[$253>>1]|0; //@line 388 "load_xm.cpp"
     $377 = HEAP32[$253>>2]|0; //@line 388 "load_xm.cpp"
     $378 = $377&65535; //@line 388 "load_xm.cpp"
     $379 = $377&255; //@line 388 "load_xm.cpp"
     $380 = ($379&255)>(12); //@line 388 "load_xm.cpp"
     $381 = ($378&65535) >>> 8;
     $382 = $381&255;
     $383 = ($376&65535) >>> 8;
     $384 = $383&255;
     $385 = $377 >>> 16;
     $386 = $385&255;
     $387 = $377 >>> 24;
     $388 = $387&255;
     if ($380) {
      HEAP8[$253>>0] = 12; //@line 388 "load_xm.cpp"
      $391 = 12;
     } else {
      $391 = $379;
     }
     $389 = ($382&255)>(12); //@line 389 "load_xm.cpp"
     if ($389) {
      HEAP8[$254>>0] = 12; //@line 389 "load_xm.cpp"
      $397 = 12;
     } else {
      $397 = $384;
     }
     $390 = ((($334)) + 556|0); //@line 390 "load_xm.cpp"
     HEAP8[$390>>0] = $391; //@line 390 "load_xm.cpp"
     $392 = ($391<<24>>24)==(0); //@line 391 "load_xm.cpp"
     if ($392) {
      $393 = ((($334)) + 4|0); //@line 391 "load_xm.cpp"
      $394 = HEAP32[$393>>2]|0; //@line 391 "load_xm.cpp"
      $395 = $394 & -2; //@line 391 "load_xm.cpp"
      HEAP32[$393>>2] = $395; //@line 391 "load_xm.cpp"
     }
     $396 = ($397<<24>>24)==(0); //@line 392 "load_xm.cpp"
     if ($396) {
      $398 = ((($334)) + 4|0); //@line 392 "load_xm.cpp"
      $399 = HEAP32[$398>>2]|0; //@line 392 "load_xm.cpp"
      $400 = $399 & -9; //@line 392 "load_xm.cpp"
      HEAP32[$398>>2] = $400; //@line 392 "load_xm.cpp"
     }
     $401 = ((($334)) + 557|0); //@line 393 "load_xm.cpp"
     HEAP8[$401>>0] = $397; //@line 393 "load_xm.cpp"
     $402 = ((($334)) + 562|0); //@line 394 "load_xm.cpp"
     HEAP8[$402>>0] = $386; //@line 394 "load_xm.cpp"
     $403 = ((($334)) + 561|0); //@line 394 "load_xm.cpp"
     HEAP8[$403>>0] = $386; //@line 394 "load_xm.cpp"
     $404 = ($386&255)>(11); //@line 395 "load_xm.cpp"
     if ($404) {
      $405 = ((($334)) + 4|0); //@line 395 "load_xm.cpp"
      $406 = HEAP32[$405>>2]|0; //@line 395 "load_xm.cpp"
      $407 = $406 & -3; //@line 395 "load_xm.cpp"
      HEAP32[$405>>2] = $407; //@line 395 "load_xm.cpp"
     }
     $408 = ((($334)) + 559|0); //@line 396 "load_xm.cpp"
     HEAP8[$408>>0] = $388; //@line 396 "load_xm.cpp"
     $409 = HEAP16[$255>>1]|0; //@line 397 "load_xm.cpp"
     $410 = HEAP32[$255>>2]|0; //@line 397 "load_xm.cpp"
     $411 = $410&65535; //@line 397 "load_xm.cpp"
     $412 = $410&255; //@line 397 "load_xm.cpp"
     $413 = ((($334)) + 560|0); //@line 397 "load_xm.cpp"
     $414 = ($412&255)>(11); //@line 398 "load_xm.cpp"
     $$58 = $414 ? 0 : $412; //@line 398 "load_xm.cpp"
     HEAP8[$413>>0] = $$58; //@line 397 "load_xm.cpp"
     $415 = ($388&255)<($$58&255); //@line 399 "load_xm.cpp"
     $416 = ($411&65535) >>> 8;
     $417 = $416&255;
     $418 = ($409&65535) >>> 8;
     $419 = $418&255;
     $420 = $410 >>> 16;
     $421 = $420&255;
     $422 = $410 >>> 24;
     $423 = $422&255;
     if (!($415)) {
      $424 = ((($334)) + 4|0); //@line 399 "load_xm.cpp"
      $425 = HEAP32[$424>>2]|0; //@line 399 "load_xm.cpp"
      $426 = $425 & -5; //@line 399 "load_xm.cpp"
      HEAP32[$424>>2] = $426; //@line 399 "load_xm.cpp"
     }
     $427 = ((($334)) + 566|0); //@line 400 "load_xm.cpp"
     HEAP8[$427>>0] = $417; //@line 400 "load_xm.cpp"
     $428 = ((($334)) + 565|0); //@line 400 "load_xm.cpp"
     HEAP8[$428>>0] = $417; //@line 400 "load_xm.cpp"
     $429 = ($419&255)>(11); //@line 401 "load_xm.cpp"
     if ($429) {
      $430 = ((($334)) + 4|0); //@line 401 "load_xm.cpp"
      $431 = HEAP32[$430>>2]|0; //@line 401 "load_xm.cpp"
      $432 = $431 & -17; //@line 401 "load_xm.cpp"
      HEAP32[$430>>2] = $432; //@line 401 "load_xm.cpp"
     }
     $433 = ((($334)) + 563|0); //@line 402 "load_xm.cpp"
     HEAP8[$433>>0] = $421; //@line 402 "load_xm.cpp"
     $434 = ((($334)) + 564|0); //@line 403 "load_xm.cpp"
     $435 = ($423&255)>(11); //@line 404 "load_xm.cpp"
     $$59 = $435 ? 0 : $423; //@line 404 "load_xm.cpp"
     HEAP8[$434>>0] = $$59; //@line 403 "load_xm.cpp"
     $436 = ($421&255)<($$59&255); //@line 405 "load_xm.cpp"
     if (!($436)) {
      $437 = ((($334)) + 4|0); //@line 405 "load_xm.cpp"
      $438 = HEAP32[$437>>2]|0; //@line 405 "load_xm.cpp"
      $439 = $438 & -33; //@line 405 "load_xm.cpp"
      HEAP32[$437>>2] = $439; //@line 405 "load_xm.cpp"
     }
     $440 = ((($334)) + 8|0); //@line 406 "load_xm.cpp"
     HEAP16[$440>>1] = 64; //@line 406 "load_xm.cpp"
     $ienv$0158 = 0;
     while(1) {
      $441 = $ienv$0158 << 1; //@line 409 "load_xm.cpp"
      $442 = (((($xmsh)) + 100|0) + ($441<<1)|0); //@line 409 "load_xm.cpp"
      $443 = HEAP16[$442>>1]|0; //@line 409 "load_xm.cpp"
      $444 = (((($334)) + 12|0) + ($ienv$0158<<1)|0); //@line 409 "load_xm.cpp"
      HEAP16[$444>>1] = $443; //@line 409 "load_xm.cpp"
      $445 = $441 | 1; //@line 410 "load_xm.cpp"
      $446 = (((($xmsh)) + 100|0) + ($445<<1)|0); //@line 410 "load_xm.cpp"
      $447 = HEAP16[$446>>1]|0; //@line 410 "load_xm.cpp"
      $448 = $447&255; //@line 410 "load_xm.cpp"
      $449 = (((($334)) + 204|0) + ($ienv$0158)|0); //@line 410 "load_xm.cpp"
      HEAP8[$449>>0] = $448; //@line 410 "load_xm.cpp"
      $450 = (((($xmsh)) + 148|0) + ($441<<1)|0); //@line 411 "load_xm.cpp"
      $451 = HEAP16[$450>>1]|0; //@line 411 "load_xm.cpp"
      $452 = (((($334)) + 76|0) + ($ienv$0158<<1)|0); //@line 411 "load_xm.cpp"
      HEAP16[$452>>1] = $451; //@line 411 "load_xm.cpp"
      $453 = (((($xmsh)) + 148|0) + ($445<<1)|0); //@line 412 "load_xm.cpp"
      $454 = HEAP16[$453>>1]|0; //@line 412 "load_xm.cpp"
      $455 = $454&255; //@line 412 "load_xm.cpp"
      $456 = (((($334)) + 236|0) + ($ienv$0158)|0); //@line 412 "load_xm.cpp"
      HEAP8[$456>>0] = $455; //@line 412 "load_xm.cpp"
      $457 = ($ienv$0158|0)==(0); //@line 413 "load_xm.cpp"
      if ($457) {
       $ienv$0158 = 1;
       continue;
      }
      $458 = HEAP16[$444>>1]|0; //@line 415 "load_xm.cpp"
      $459 = (($ienv$0158) + -1)|0; //@line 415 "load_xm.cpp"
      $460 = (((($334)) + 12|0) + ($459<<1)|0); //@line 415 "load_xm.cpp"
      $461 = HEAP16[$460>>1]|0; //@line 415 "load_xm.cpp"
      $462 = ($458&65535)<($461&65535); //@line 415 "load_xm.cpp"
      do {
       if ($462) {
        $463 = $458&65535; //@line 415 "load_xm.cpp"
        $464 = $463 & 255; //@line 417 "load_xm.cpp"
        $465 = $461&65535; //@line 418 "load_xm.cpp"
        $466 = $465 & 65280; //@line 418 "load_xm.cpp"
        $467 = $466 | $464; //@line 418 "load_xm.cpp"
        $468 = $467&65535; //@line 418 "load_xm.cpp"
        HEAP16[$444>>1] = $468; //@line 418 "load_xm.cpp"
        $469 = ($467>>>0)<($465>>>0); //@line 419 "load_xm.cpp"
        if (!($469)) {
         break;
        }
        $470 = (($467) + 256)|0; //@line 419 "load_xm.cpp"
        $471 = $470&65535; //@line 419 "load_xm.cpp"
        HEAP16[$444>>1] = $471; //@line 419 "load_xm.cpp"
       }
      } while(0);
      $472 = HEAP16[$452>>1]|0; //@line 421 "load_xm.cpp"
      $473 = (((($334)) + 76|0) + ($459<<1)|0); //@line 421 "load_xm.cpp"
      $474 = HEAP16[$473>>1]|0; //@line 421 "load_xm.cpp"
      $475 = ($472&65535)<($474&65535); //@line 421 "load_xm.cpp"
      do {
       if ($475) {
        $476 = $472&65535; //@line 421 "load_xm.cpp"
        $477 = $476 & 255; //@line 423 "load_xm.cpp"
        $478 = $474&65535; //@line 424 "load_xm.cpp"
        $479 = $478 & 65280; //@line 424 "load_xm.cpp"
        $480 = $479 | $477; //@line 424 "load_xm.cpp"
        $481 = $480&65535; //@line 424 "load_xm.cpp"
        HEAP16[$452>>1] = $481; //@line 424 "load_xm.cpp"
        $482 = ($480>>>0)<($478>>>0); //@line 425 "load_xm.cpp"
        if (!($482)) {
         break;
        }
        $483 = (($480) + 256)|0; //@line 425 "load_xm.cpp"
        $484 = $483&65535; //@line 425 "load_xm.cpp"
        HEAP16[$452>>1] = $484; //@line 425 "load_xm.cpp"
       }
      } while(0);
      $485 = (($ienv$0158) + 1)|0; //@line 407 "load_xm.cpp"
      $exitcond242 = ($485|0)==(12); //@line 407 "load_xm.cpp"
      if ($exitcond242) {
       $j8$0159 = 0;
       break;
      } else {
       $ienv$0158 = $485;
      }
     }
     while(1) {
      $495 = (($j8$0159) + 1)|0; //@line 431 "load_xm.cpp"
      $496 = (($j8$0159) + 13)|0; //@line 431 "load_xm.cpp"
      $497 = $496&255; //@line 431 "load_xm.cpp"
      $498 = (($j8$0159) + 12)|0; //@line 431 "load_xm.cpp"
      $499 = (((($334)) + 428|0) + ($498)|0); //@line 431 "load_xm.cpp"
      HEAP8[$499>>0] = $497; //@line 431 "load_xm.cpp"
      $500 = (((($xmsh)) + 4|0) + ($j8$0159)|0); //@line 432 "load_xm.cpp"
      $501 = HEAP8[$500>>0]|0; //@line 432 "load_xm.cpp"
      $502 = ($501&255)<($269&255); //@line 432 "load_xm.cpp"
      if ($502) {
       $503 = $501&255; //@line 432 "load_xm.cpp"
       $504 = (($samplemap) + ($503<<2)|0); //@line 433 "load_xm.cpp"
       $505 = HEAP32[$504>>2]|0; //@line 433 "load_xm.cpp"
       $506 = $505&255; //@line 433 "load_xm.cpp"
       $507 = (((($334)) + 300|0) + ($498)|0); //@line 433 "load_xm.cpp"
       HEAP8[$507>>0] = $506; //@line 433 "load_xm.cpp"
      }
      $exitcond243 = ($495|0)==(96); //@line 429 "load_xm.cpp"
      if ($exitcond243) {
       break;
      } else {
       $j8$0159 = $495;
      }
     }
     $486 = HEAP32[$xmsh>>2]|0; //@line 439 "load_xm.cpp"
     $487 = HEAP16[$256>>1]|0; //@line 492 "load_xm.cpp"
     $488 = $487&255; //@line 492 "load_xm.cpp"
     $489 = ($487&65535) >>> 8;
     $490 = $489&255;
     $491 = HEAP16[$257>>1]|0; //@line 494 "load_xm.cpp"
     $492 = $491&255; //@line 494 "load_xm.cpp"
     $493 = ($491&65535) >>> 8;
     $494 = $493&255;
     $dwMemPos$6161 = $262;$ins$0160 = 0;
     while(1) {
      $508 = (($dwMemPos$6161) + 40)|0; //@line 438 "load_xm.cpp"
      $509 = ($508>>>0)>($dwMemLength>>>0); //@line 438 "load_xm.cpp"
      if ($509) {
       break L138;
      }
      $510 = (($486) + ($dwMemPos$6161))|0; //@line 439 "load_xm.cpp"
      $511 = ($510>>>0)>($dwMemLength>>>0); //@line 439 "load_xm.cpp"
      if ($511) {
       break L138;
      }
      $512 = (($lpStream) + ($dwMemPos$6161)|0); //@line 440 "load_xm.cpp"
      $513 = HEAPU8[$512>>0]|(HEAPU8[$512+1>>0]<<8)|(HEAPU8[$512+2>>0]<<16)|(HEAPU8[$512+3>>0]<<24); //@line 440 "load_xm.cpp"
      $$sum25 = (($dwMemPos$6161) + 4)|0; //@line 440 "load_xm.cpp"
      $514 = (($lpStream) + ($$sum25)|0); //@line 440 "load_xm.cpp"
      $515 = HEAPU8[$514>>0]|(HEAPU8[$514+1>>0]<<8)|(HEAPU8[$514+2>>0]<<16)|(HEAPU8[$514+3>>0]<<24); //@line 440 "load_xm.cpp"
      $$sum26 = (($dwMemPos$6161) + 8)|0; //@line 440 "load_xm.cpp"
      $516 = (($lpStream) + ($$sum26)|0); //@line 440 "load_xm.cpp"
      $517 = HEAPU8[$516>>0]|(HEAPU8[$516+1>>0]<<8)|(HEAPU8[$516+2>>0]<<16)|(HEAPU8[$516+3>>0]<<24); //@line 440 "load_xm.cpp"
      $$sum27 = (($dwMemPos$6161) + 12)|0; //@line 440 "load_xm.cpp"
      $518 = (($lpStream) + ($$sum27)|0); //@line 440 "load_xm.cpp"
      $519 = HEAP8[$518>>0]|0; //@line 440 "load_xm.cpp"
      $$sum28 = (($dwMemPos$6161) + 13)|0; //@line 440 "load_xm.cpp"
      $520 = (($lpStream) + ($$sum28)|0); //@line 440 "load_xm.cpp"
      $521 = HEAP8[$520>>0]|0; //@line 440 "load_xm.cpp"
      $$sum29 = (($dwMemPos$6161) + 14)|0; //@line 440 "load_xm.cpp"
      $522 = (($lpStream) + ($$sum29)|0); //@line 440 "load_xm.cpp"
      $523 = HEAP8[$522>>0]|0; //@line 440 "load_xm.cpp"
      $$sum30 = (($dwMemPos$6161) + 15)|0; //@line 440 "load_xm.cpp"
      $524 = (($lpStream) + ($$sum30)|0); //@line 440 "load_xm.cpp"
      $525 = HEAP8[$524>>0]|0; //@line 440 "load_xm.cpp"
      $$sum31 = (($dwMemPos$6161) + 16)|0; //@line 440 "load_xm.cpp"
      $526 = (($lpStream) + ($$sum31)|0); //@line 440 "load_xm.cpp"
      $527 = HEAP8[$526>>0]|0; //@line 440 "load_xm.cpp"
      $$sum32 = (($dwMemPos$6161) + 17)|0; //@line 440 "load_xm.cpp"
      $528 = (($lpStream) + ($$sum32)|0); //@line 440 "load_xm.cpp"
      $529 = HEAP8[$528>>0]|0; //@line 440 "load_xm.cpp"
      $$sum33 = (($dwMemPos$6161) + 18)|0; //@line 440 "load_xm.cpp"
      $530 = (($lpStream) + ($$sum33)|0); //@line 440 "load_xm.cpp"
      dest=$xmss$sroa$55; src=$530; stop=dest+22|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 440 "load_xm.cpp"
      $531 = $523&255; //@line 445 "load_xm.cpp"
      $532 = $531 & 16; //@line 445 "load_xm.cpp"
      $533 = $532 >>> 3; //@line 445 "load_xm.cpp"
      $534 = (($flags) + ($ins$0160)|0); //@line 445 "load_xm.cpp"
      $535 = $531 & 32; //@line 446 "load_xm.cpp"
      $536 = ($535|0)==(0); //@line 446 "load_xm.cpp"
      $537 = $535 >>> 2; //@line 446 "load_xm.cpp"
      $538 = $537 | 2; //@line 446 "load_xm.cpp"
      $storemerge34$in = (($538) + ($533))|0; //@line 446 "load_xm.cpp"
      $storemerge34 = $storemerge34$in&255; //@line 446 "load_xm.cpp"
      HEAP8[$534>>0] = $storemerge34; //@line 445 "load_xm.cpp"
      $539 = (($samplesize) + ($ins$0160<<2)|0); //@line 447 "load_xm.cpp"
      HEAP32[$539>>2] = $513; //@line 447 "load_xm.cpp"
      $540 = (($samplemap) + ($ins$0160<<2)|0); //@line 448 "load_xm.cpp"
      $541 = HEAP32[$540>>2]|0; //@line 448 "load_xm.cpp"
      $542 = ($541|0)==(0); //@line 448 "load_xm.cpp"
      if (!($542)) {
       $543 = ($532|0)==(0); //@line 445 "load_xm.cpp"
       if ($543) {
        $xmss$sroa$0$0 = $513;$xmss$sroa$17$0 = $515;$xmss$sroa$26$0 = $517;
       } else {
        $544 = $517 >>> 1; //@line 451 "load_xm.cpp"
        $545 = $515 >>> 1; //@line 452 "load_xm.cpp"
        $546 = $513 >>> 1; //@line 453 "load_xm.cpp"
        $xmss$sroa$0$0 = $546;$xmss$sroa$17$0 = $545;$xmss$sroa$26$0 = $544;
       }
       if ($536) {
        $xmss$sroa$0$1 = $xmss$sroa$0$0;$xmss$sroa$17$1 = $xmss$sroa$17$0;$xmss$sroa$26$1 = $xmss$sroa$26$0;
       } else {
        $547 = $xmss$sroa$26$0 >>> 1; //@line 457 "load_xm.cpp"
        $548 = $xmss$sroa$17$0 >>> 1; //@line 458 "load_xm.cpp"
        $549 = $xmss$sroa$0$0 >>> 1; //@line 459 "load_xm.cpp"
        $xmss$sroa$0$1 = $549;$xmss$sroa$17$1 = $548;$xmss$sroa$26$1 = $547;
       }
       $550 = ($xmss$sroa$0$1>>>0)>(16000000); //@line 461 "load_xm.cpp"
       $$xmss$sroa$0$1 = $550 ? 16000000 : $xmss$sroa$0$1; //@line 461 "load_xm.cpp"
       $551 = ($xmss$sroa$17$1>>>0)<($$xmss$sroa$0$1>>>0); //@line 462 "load_xm.cpp"
       $552 = $531 & 252; //@line 462 "load_xm.cpp"
       $553 = $552&255; //@line 462 "load_xm.cpp"
       $xmss$sroa$40$0 = $551 ? $523 : $553; //@line 462 "load_xm.cpp"
       $554 = (($xmss$sroa$26$1) + ($xmss$sroa$17$1))|0; //@line 463 "load_xm.cpp"
       $555 = ($554>>>0)>($$xmss$sroa$0$1>>>0); //@line 464 "load_xm.cpp"
       $$xmss$sroa$0$1$ = $555 ? $$xmss$sroa$0$1 : $554; //@line 464 "load_xm.cpp"
       $556 = ($$xmss$sroa$0$1$|0)==(0); //@line 465 "load_xm.cpp"
       if ($556) {
        $557 = $xmss$sroa$40$0&255; //@line 465 "load_xm.cpp"
        $558 = $557 & 252; //@line 465 "load_xm.cpp"
        $559 = $558&255; //@line 465 "load_xm.cpp"
        $xmss$sroa$40$1 = $559;
       } else {
        $xmss$sroa$40$1 = $xmss$sroa$40$0;
       }
       $560 = (((($this)) + 59804|0) + ($541<<5)|0); //@line 467 "load_xm.cpp"
       dest=$560; src=$xmss$sroa$55; stop=dest+22|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 467 "load_xm.cpp"
       $561 = (((((($this)) + 59804|0) + ($541<<5)|0)) + 22|0); //@line 468 "load_xm.cpp"
       HEAP8[$561>>0] = 0; //@line 468 "load_xm.cpp"
       $562 = (((($this)) + 33280|0) + ($541<<6)|0); //@line 470 "load_xm.cpp"
       HEAP32[$562>>2] = $$xmss$sroa$0$1; //@line 470 "load_xm.cpp"
       $563 = (((((($this)) + 33280|0) + ($541<<6)|0)) + 4|0); //@line 471 "load_xm.cpp"
       HEAP32[$563>>2] = $xmss$sroa$17$1; //@line 471 "load_xm.cpp"
       $564 = (((((($this)) + 33280|0) + ($541<<6)|0)) + 8|0); //@line 472 "load_xm.cpp"
       HEAP32[$564>>2] = $$xmss$sroa$0$1$; //@line 473 "load_xm.cpp"
       $565 = ($xmss$sroa$17$1>>>0)<($$xmss$sroa$0$1$>>>0); //@line 474 "load_xm.cpp"
       if (!($565)) {
        HEAP32[$564>>2] = 0; //@line 476 "load_xm.cpp"
        HEAP32[$563>>2] = 0; //@line 476 "load_xm.cpp"
       }
       $566 = $xmss$sroa$40$1&255; //@line 478 "load_xm.cpp"
       $567 = $566 & 3; //@line 478 "load_xm.cpp"
       $568 = ($567|0)==(0); //@line 478 "load_xm.cpp"
       if (!($568)) {
        $569 = (((((($this)) + 33280|0) + ($541<<6)|0)) + 34|0); //@line 478 "load_xm.cpp"
        $570 = HEAP16[$569>>1]|0; //@line 478 "load_xm.cpp"
        $571 = $570&65535; //@line 478 "load_xm.cpp"
        $572 = $571 | 2; //@line 478 "load_xm.cpp"
        $573 = $572&65535; //@line 478 "load_xm.cpp"
        HEAP16[$569>>1] = $573; //@line 478 "load_xm.cpp"
       }
       $574 = $566 & 2; //@line 479 "load_xm.cpp"
       $575 = ($574|0)==(0); //@line 479 "load_xm.cpp"
       if (!($575)) {
        $576 = (((((($this)) + 33280|0) + ($541<<6)|0)) + 34|0); //@line 479 "load_xm.cpp"
        $577 = HEAP16[$576>>1]|0; //@line 479 "load_xm.cpp"
        $578 = $577&65535; //@line 479 "load_xm.cpp"
        $579 = $578 | 4; //@line 479 "load_xm.cpp"
        $580 = $579&65535; //@line 479 "load_xm.cpp"
        HEAP16[$576>>1] = $580; //@line 479 "load_xm.cpp"
       }
       $581 = $519&255; //@line 480 "load_xm.cpp"
       $582 = $581 << 2; //@line 480 "load_xm.cpp"
       $583 = $582&65535; //@line 480 "load_xm.cpp"
       $584 = (((((($this)) + 33280|0) + ($541<<6)|0)) + 30|0); //@line 480 "load_xm.cpp"
       $585 = ($582>>>0)>(256); //@line 481 "load_xm.cpp"
       $$60 = $585 ? 256 : $583; //@line 481 "load_xm.cpp"
       HEAP16[$584>>1] = $$60; //@line 480 "load_xm.cpp"
       $586 = (((((($this)) + 33280|0) + ($541<<6)|0)) + 32|0); //@line 482 "load_xm.cpp"
       HEAP16[$586>>1] = 64; //@line 482 "load_xm.cpp"
       $587 = ($529<<24>>24)==(-83); //@line 483 "load_xm.cpp"
       $588 = $566 & 48;
       $589 = ($588|0)==(0); //@line 483 "load_xm.cpp"
       $or$cond62 = $587 & $589; //@line 483 "load_xm.cpp"
       if ($or$cond62) {
        HEAP8[$534>>0] = 3; //@line 485 "load_xm.cpp"
        $590 = (($513) + 1)|0; //@line 486 "load_xm.cpp"
        $591 = $590 >>> 1; //@line 486 "load_xm.cpp"
        $592 = (($591) + 16)|0; //@line 486 "load_xm.cpp"
        HEAP32[$539>>2] = $592; //@line 486 "load_xm.cpp"
       }
       $593 = (((((($this)) + 33280|0) + ($541<<6)|0)) + 37|0); //@line 488 "load_xm.cpp"
       HEAP8[$593>>0] = $521; //@line 488 "load_xm.cpp"
       $594 = (((((($this)) + 33280|0) + ($541<<6)|0)) + 36|0); //@line 489 "load_xm.cpp"
       HEAP8[$594>>0] = $527; //@line 489 "load_xm.cpp"
       $595 = $525&255; //@line 490 "load_xm.cpp"
       $596 = (((((($this)) + 33280|0) + ($541<<6)|0)) + 28|0); //@line 490 "load_xm.cpp"
       HEAP16[$596>>1] = $595; //@line 490 "load_xm.cpp"
       $597 = (((((($this)) + 33280|0) + ($541<<6)|0)) + 34|0); //@line 491 "load_xm.cpp"
       $598 = HEAP16[$597>>1]|0; //@line 491 "load_xm.cpp"
       $599 = $598&65535; //@line 491 "load_xm.cpp"
       $600 = $599 | 32; //@line 491 "load_xm.cpp"
       $601 = $600&65535; //@line 491 "load_xm.cpp"
       HEAP16[$597>>1] = $601; //@line 491 "load_xm.cpp"
       $602 = (((((($this)) + 33280|0) + ($541<<6)|0)) + 38|0); //@line 492 "load_xm.cpp"
       HEAP8[$602>>0] = $488; //@line 492 "load_xm.cpp"
       $603 = (((((($this)) + 33280|0) + ($541<<6)|0)) + 39|0); //@line 493 "load_xm.cpp"
       HEAP8[$603>>0] = $490; //@line 493 "load_xm.cpp"
       $604 = (((((($this)) + 33280|0) + ($541<<6)|0)) + 40|0); //@line 494 "load_xm.cpp"
       HEAP8[$604>>0] = $492; //@line 494 "load_xm.cpp"
       $605 = (((((($this)) + 33280|0) + ($541<<6)|0)) + 41|0); //@line 495 "load_xm.cpp"
       HEAP8[$605>>0] = $494; //@line 495 "load_xm.cpp"
       $606 = (((((($this)) + 33280|0) + ($541<<6)|0)) + 42|0); //@line 496 "load_xm.cpp"
       dest=$606; src=$xmss$sroa$55; stop=dest+22|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 496 "load_xm.cpp"
       $607 = (((((($this)) + 33280|0) + ($541<<6)|0)) + 63|0); //@line 497 "load_xm.cpp"
       HEAP8[$607>>0] = 0; //@line 497 "load_xm.cpp"
      }
      $608 = (($ins$0160) + 1)|0; //@line 436 "load_xm.cpp"
      $609 = ($608>>>0)<($270>>>0); //@line 436 "load_xm.cpp"
      if ($609) {
       $dwMemPos$6161 = $510;$ins$0160 = $608;
      } else {
       $dwMemPos$7164 = $510;$ismpd$0165 = 0;
       break;
      }
     }
     while(1) {
      $610 = (($samplemap) + ($ismpd$0165<<2)|0); //@line 507 "load_xm.cpp"
      $611 = HEAP32[$610>>2]|0; //@line 507 "load_xm.cpp"
      $612 = ($611|0)==(0); //@line 507 "load_xm.cpp"
      $$phi$trans$insert = (($samplesize) + ($ismpd$0165<<2)|0);
      $$pre254 = HEAP32[$$phi$trans$insert>>2]|0; //@line 511 "load_xm.cpp"
      do {
       if (!($612)) {
        $613 = ($$pre254|0)!=(0); //@line 507 "load_xm.cpp"
        $614 = ($dwMemPos$7164>>>0)<($dwMemLength>>>0); //@line 507 "load_xm.cpp"
        $or$cond63 = $614 & $613; //@line 507 "load_xm.cpp"
        if (!($or$cond63)) {
         break;
        }
        $615 = (((($this)) + 33280|0) + ($611<<6)|0); //@line 509 "load_xm.cpp"
        $616 = (($flags) + ($ismpd$0165)|0); //@line 509 "load_xm.cpp"
        $617 = HEAP8[$616>>0]|0; //@line 509 "load_xm.cpp"
        $618 = $617&255; //@line 509 "load_xm.cpp"
        $619 = (($lpStream) + ($dwMemPos$7164)|0); //@line 509 "load_xm.cpp"
        $620 = (($dwMemLength) - ($dwMemPos$7164))|0; //@line 509 "load_xm.cpp"
        (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$615,$618,$619,$620)|0); //@line 509 "load_xm.cpp"
       }
      } while(0);
      $621 = (($$pre254) + ($dwMemPos$7164))|0; //@line 511 "load_xm.cpp"
      $622 = ($621>>>0)<($dwMemLength>>>0); //@line 512 "load_xm.cpp"
      $623 = (($ismpd$0165) + 1)|0; //@line 505 "load_xm.cpp"
      $624 = ($623>>>0)<($270>>>0); //@line 505 "load_xm.cpp"
      $or$cond261 = $622 & $624; //@line 512 "load_xm.cpp"
      if ($or$cond261) {
       $dwMemPos$7164 = $621;$ismpd$0165 = $623;
      } else {
       $dwMemPos$9$ph67 = $621;$unused_samples$4$ph66 = $unused_samples$3$lcssa;
       break;
      }
     }
    }
    $indvars$iv$next239 = (($indvars$iv238) + 1)|0; //@line 274 "load_xm.cpp"
    $625 = ($indvars$iv$next239>>>0)>($40>>>0); //@line 274 "load_xm.cpp"
    if ($625) {
     $dwMemPos$4$lcssa = $dwMemPos$9$ph67;
     break L136;
    } else {
     $dwMemPos$4174 = $dwMemPos$9$ph67;$indvars$iv238 = $indvars$iv$next239;$unused_samples$0170 = $unused_samples$4$ph66;
    }
   }
   $$2 = 1;
   STACKTOP = sp;return ($$2|0); //@line 581 "load_xm.cpp"
  }
 } while(0);
 $626 = (($dwMemPos$4$lcssa) + 8)|0; //@line 516 "load_xm.cpp"
 $627 = ($626>>>0)<($dwMemLength>>>0); //@line 516 "load_xm.cpp"
 if ($627) {
  $628 = (($lpStream) + ($dwMemPos$4$lcssa)|0); //@line 516 "load_xm.cpp"
  $629 = HEAP32[$628>>2]|0; //@line 516 "load_xm.cpp"
  $630 = ($629|0)==(1954047348); //@line 516 "load_xm.cpp"
  if ($630) {
   $$sum20 = (($dwMemPos$4$lcssa) + 4)|0; //@line 518 "load_xm.cpp"
   $631 = (($lpStream) + ($$sum20)|0); //@line 518 "load_xm.cpp"
   $632 = HEAP32[$631>>2]|0; //@line 518 "load_xm.cpp"
   $633 = (($632) + ($626))|0; //@line 520 "load_xm.cpp"
   $634 = ($633>>>0)<=($dwMemLength>>>0); //@line 520 "load_xm.cpp"
   $635 = ($632>>>0)<(16384); //@line 520 "load_xm.cpp"
   $or$cond8 = $635 & $634; //@line 520 "load_xm.cpp"
   if ($or$cond8) {
    $636 = (($632) + 1)|0; //@line 522 "load_xm.cpp"
    $637 = (__Znaj($636)|0); //@line 522 "load_xm.cpp"
    $638 = ((($this)) + 59796|0); //@line 522 "load_xm.cpp"
    HEAP32[$638>>2] = $637; //@line 522 "load_xm.cpp"
    $639 = (($lpStream) + ($626)|0); //@line 525 "load_xm.cpp"
    _memcpy(($637|0),($639|0),($632|0))|0; //@line 525 "load_xm.cpp"
    $640 = (($637) + ($632)|0); //@line 526 "load_xm.cpp"
    HEAP8[$640>>0] = 0; //@line 526 "load_xm.cpp"
    $dwMemPos$10 = $633;
   } else {
    $dwMemPos$10 = $626;
   }
  } else {
   $dwMemPos$10 = $dwMemPos$4$lcssa;
  }
 } else {
  $dwMemPos$10 = $dwMemPos$4$lcssa;
 }
 $641 = (($dwMemPos$10) + 8)|0; //@line 532 "load_xm.cpp"
 $642 = ($641>>>0)<($dwMemLength>>>0); //@line 532 "load_xm.cpp"
 if ($642) {
  $643 = (($lpStream) + ($dwMemPos$10)|0); //@line 532 "load_xm.cpp"
  $644 = HEAP32[$643>>2]|0; //@line 532 "load_xm.cpp"
  $645 = ($644|0)==(1229211981); //@line 532 "load_xm.cpp"
  if ($645) {
   $$sum19 = (($dwMemPos$10) + 4)|0; //@line 534 "load_xm.cpp"
   $646 = (($lpStream) + ($$sum19)|0); //@line 534 "load_xm.cpp"
   $647 = HEAP32[$646>>2]|0; //@line 534 "load_xm.cpp"
   $648 = ($647|0)==(4896); //@line 536 "load_xm.cpp"
   if ($648) {
    $649 = ((($this)) + 53600|0); //@line 538 "load_xm.cpp"
    $650 = (($lpStream) + ($641)|0); //@line 538 "load_xm.cpp"
    _memcpy(($649|0),($650|0),4896)|0; //@line 538 "load_xm.cpp"
    $651 = ((($this)) + 59660|0); //@line 539 "load_xm.cpp"
    $652 = HEAP32[$651>>2]|0; //@line 539 "load_xm.cpp"
    $653 = $652 | 1; //@line 539 "load_xm.cpp"
    HEAP32[$651>>2] = $653; //@line 539 "load_xm.cpp"
    $dwMemPos$11 = $641;
   } else {
    $dwMemPos$11 = $641;
   }
  } else {
   $dwMemPos$11 = $dwMemPos$10;
  }
 } else {
  $dwMemPos$11 = $dwMemPos$10;
 }
 $654 = (($dwMemPos$11) + 8)|0; //@line 543 "load_xm.cpp"
 $655 = ($654>>>0)<($dwMemLength>>>0); //@line 543 "load_xm.cpp"
 if ($655) {
  $656 = (($lpStream) + ($dwMemPos$11)|0); //@line 543 "load_xm.cpp"
  $657 = HEAP32[$656>>2]|0; //@line 543 "load_xm.cpp"
  $658 = ($657|0)==(1296125520); //@line 543 "load_xm.cpp"
  if ($658) {
   $$sum18 = (($dwMemPos$11) + 4)|0; //@line 545 "load_xm.cpp"
   $659 = (($lpStream) + ($$sum18)|0); //@line 545 "load_xm.cpp"
   $660 = HEAP32[$659>>2]|0; //@line 545 "load_xm.cpp"
   $661 = (($660) + ($654))|0; //@line 547 "load_xm.cpp"
   $662 = ($661>>>0)<=($dwMemLength>>>0); //@line 547 "load_xm.cpp"
   $$off = (($660) + -32)|0; //@line 547 "load_xm.cpp"
   $663 = ($$off>>>0)<(7649); //@line 547 "load_xm.cpp"
   $664 = $663 & $662; //@line 547 "load_xm.cpp"
   if ($664) {
    $665 = (__Znaj($660)|0); //@line 549 "load_xm.cpp"
    $666 = ((($this)) + 59800|0); //@line 549 "load_xm.cpp"
    HEAP32[$666>>2] = $665; //@line 549 "load_xm.cpp"
    $667 = $660 >>> 5; //@line 553 "load_xm.cpp"
    $668 = ((($this)) + 59792|0); //@line 553 "load_xm.cpp"
    HEAP32[$668>>2] = $667; //@line 553 "load_xm.cpp"
    $669 = (($lpStream) + ($654)|0); //@line 554 "load_xm.cpp"
    _memcpy(($665|0),($669|0),($660|0))|0; //@line 554 "load_xm.cpp"
    $dwMemPos$12 = $661;
   } else {
    $dwMemPos$12 = $654;
   }
  } else {
   $dwMemPos$12 = $dwMemPos$11;
  }
 } else {
  $dwMemPos$12 = $dwMemPos$11;
 }
 $670 = (($dwMemPos$12) + 8)|0; //@line 560 "load_xm.cpp"
 $671 = ($670>>>0)<($dwMemLength>>>0); //@line 560 "load_xm.cpp"
 do {
  if ($671) {
   $672 = (($lpStream) + ($dwMemPos$12)|0); //@line 560 "load_xm.cpp"
   $673 = HEAP32[$672>>2]|0; //@line 560 "load_xm.cpp"
   $674 = ($673|0)==(1296125507); //@line 560 "load_xm.cpp"
   if ($674) {
    $$sum = (($dwMemPos$12) + 4)|0; //@line 562 "load_xm.cpp"
    $675 = (($lpStream) + ($$sum)|0); //@line 562 "load_xm.cpp"
    $676 = HEAP32[$675>>2]|0; //@line 562 "load_xm.cpp"
    $677 = (($676) + ($670))|0; //@line 564 "load_xm.cpp"
    $678 = ($677>>>0)<=($dwMemLength>>>0); //@line 564 "load_xm.cpp"
    $679 = ($676>>>0)<(1281); //@line 564 "load_xm.cpp"
    $or$cond14 = $679 & $678; //@line 564 "load_xm.cpp"
    if (!($or$cond14)) {
     $dwMemPos$13 = $670;
     break;
    }
    $680 = (($676>>>0) / 20)&-1; //@line 566 "load_xm.cpp"
    $681 = ($676>>>0)>(19); //@line 567 "load_xm.cpp"
    if ($681) {
     $i13$0118 = 0;
    } else {
     $dwMemPos$13 = $677;
     break;
    }
    while(1) {
     $682 = (((((($this)) + 49600|0) + (($i13$0118*36)|0)|0)) + 16|0); //@line 569 "load_xm.cpp"
     $683 = ($i13$0118*20)|0; //@line 569 "load_xm.cpp"
     $$sum17 = (($683) + ($670))|0; //@line 569 "load_xm.cpp"
     $684 = (($lpStream) + ($$sum17)|0); //@line 569 "load_xm.cpp"
     dest=$682; src=$684; stop=dest+20|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 569 "load_xm.cpp"
     $685 = (((((($this)) + 49600|0) + (($i13$0118*36)|0)|0)) + 35|0); //@line 570 "load_xm.cpp"
     HEAP8[$685>>0] = 0; //@line 570 "load_xm.cpp"
     $686 = (($i13$0118) + 1)|0; //@line 567 "load_xm.cpp"
     $687 = ($686>>>0)<($680>>>0); //@line 567 "load_xm.cpp"
     if ($687) {
      $i13$0118 = $686;
     } else {
      $dwMemPos$13 = $677;
      break;
     }
    }
   } else {
    $dwMemPos$13 = $dwMemPos$12;
   }
  } else {
   $dwMemPos$13 = $dwMemPos$12;
  }
 } while(0);
 $688 = (($dwMemPos$13) + 8)|0; //@line 576 "load_xm.cpp"
 $689 = ($688>>>0)<($dwMemLength>>>0); //@line 576 "load_xm.cpp"
 if (!($689)) {
  $$2 = 1;
  STACKTOP = sp;return ($$2|0); //@line 581 "load_xm.cpp"
 }
 $690 = (($lpStream) + ($dwMemPos$13)|0); //@line 578 "load_xm.cpp"
 $691 = (($dwMemLength) - ($dwMemPos$13))|0; //@line 578 "load_xm.cpp"
 (__ZN10CSoundFile14LoadMixPluginsEPKvj($this,$690,$691)|0); //@line 578 "load_xm.cpp"
 $$2 = 1;
 STACKTOP = sp;return ($$2|0); //@line 581 "load_xm.cpp"
}
function __ZN10CSoundFile15AllocatePatternEjj($rows,$nchns) {
 $rows = $rows|0;
 $nchns = $nchns|0;
 var $$arith = 0, $$overflow = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = Math_imul($nchns, $rows)|0; //@line 340 "sndfile.cpp"
 $$arith = ($0*6)|0; //@line 340 "sndfile.cpp"
 $$overflow = ($0>>>0)>(715827882); //@line 340 "sndfile.cpp"
 $1 = $$overflow ? -1 : $$arith; //@line 340 "sndfile.cpp"
 $2 = (__Znaj($1)|0); //@line 340 "sndfile.cpp"
 _memset(($2|0),0,($$arith|0))|0; //@line 341 "sndfile.cpp"
 return ($2|0); //@line 342 "sndfile.cpp"
}
function __ZNK10CSoundFile17ConvertModCommandEP11_MODCOMMAND($this,$m) {
 $this = $this|0;
 $m = $m|0;
 var $$ = 0, $$1 = 0, $$2 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $command$0 = 0, $not$ = 0, $param$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($m)) + 3|0); //@line 22 "load_mod.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 22 "load_mod.cpp"
 $2 = $1&255; //@line 22 "load_mod.cpp"
 $3 = ((($m)) + 5|0); //@line 22 "load_mod.cpp"
 $4 = HEAP8[$3>>0]|0; //@line 22 "load_mod.cpp"
 $5 = $4&255; //@line 22 "load_mod.cpp"
 do {
  switch ($2|0) {
  case 0:  {
   $not$ = ($4<<24>>24)!=(0); //@line 26 "load_mod.cpp"
   $$ = $not$&1; //@line 26 "load_mod.cpp"
   $command$0 = $$;$param$0 = $5;
   break;
  }
  case 1:  {
   $command$0 = 2;$param$0 = $5;
   break;
  }
  case 2:  {
   $command$0 = 3;$param$0 = $5;
   break;
  }
  case 3:  {
   $command$0 = 4;$param$0 = $5;
   break;
  }
  case 4:  {
   $command$0 = 5;$param$0 = $5;
   break;
  }
  case 5:  {
   $6 = $5 & 240; //@line 31 "load_mod.cpp"
   $7 = ($6|0)==(0); //@line 31 "load_mod.cpp"
   $$1 = $7 ? $5 : $6; //@line 31 "load_mod.cpp"
   $command$0 = 6;$param$0 = $$1;
   break;
  }
  case 6:  {
   $8 = $5 & 240; //@line 32 "load_mod.cpp"
   $9 = ($8|0)==(0); //@line 32 "load_mod.cpp"
   $$2 = $9 ? $5 : $8; //@line 32 "load_mod.cpp"
   $command$0 = 7;$param$0 = $$2;
   break;
  }
  case 7:  {
   $command$0 = 8;$param$0 = $5;
   break;
  }
  case 8:  {
   $command$0 = 9;$param$0 = $5;
   break;
  }
  case 9:  {
   $command$0 = 10;$param$0 = $5;
   break;
  }
  case 10:  {
   $10 = $5 & 240; //@line 36 "load_mod.cpp"
   $11 = ($10|0)==(0); //@line 36 "load_mod.cpp"
   $$3 = $11 ? $5 : $10; //@line 36 "load_mod.cpp"
   $command$0 = 11;$param$0 = $$3;
   break;
  }
  case 11:  {
   $command$0 = 12;$param$0 = $5;
   break;
  }
  case 12:  {
   $command$0 = 13;$param$0 = $5;
   break;
  }
  case 13:  {
   $12 = $5 >>> 4; //@line 39 "load_mod.cpp"
   $13 = ($12*10)|0; //@line 39 "load_mod.cpp"
   $14 = $5 & 15; //@line 39 "load_mod.cpp"
   $15 = (($13) + ($14))|0; //@line 39 "load_mod.cpp"
   $command$0 = 14;$param$0 = $15;
   break;
  }
  case 14:  {
   $command$0 = 19;$param$0 = $5;
   break;
  }
  case 15:  {
   $16 = ((($this)) + 59680|0); //@line 41 "load_mod.cpp"
   $17 = HEAP32[$16>>2]|0; //@line 41 "load_mod.cpp"
   $18 = $17 & 1048580; //@line 41 "load_mod.cpp"
   $19 = ($18|0)!=(0); //@line 41 "load_mod.cpp"
   $20 = $19 ? 31 : 32; //@line 41 "load_mod.cpp"
   $21 = ($5>>>0)<=($20>>>0); //@line 41 "load_mod.cpp"
   $22 = $21 ? 16 : 17; //@line 41 "load_mod.cpp"
   $23 = ($4<<24>>24)==(-1); //@line 42 "load_mod.cpp"
   if ($23) {
    $24 = ((($this)) + 59684|0); //@line 42 "load_mod.cpp"
    $25 = HEAP32[$24>>2]|0; //@line 42 "load_mod.cpp"
    $26 = ($25|0)==(15); //@line 42 "load_mod.cpp"
    $$4 = $26 ? 0 : $22; //@line 42 "load_mod.cpp"
    $command$0 = $$4;$param$0 = $5;
   } else {
    $command$0 = $22;$param$0 = $5;
   }
   break;
  }
  case 16:  {
   $command$0 = 23;$param$0 = $5;
   break;
  }
  case 17:  {
   $27 = $5 & 240; //@line 45 "load_mod.cpp"
   $28 = ($27|0)==(0); //@line 45 "load_mod.cpp"
   $$5 = $28 ? $5 : $27; //@line 45 "load_mod.cpp"
   $command$0 = 24;$param$0 = $$5;
   break;
  }
  case 20:  {
   $command$0 = 25;$param$0 = $5;
   break;
  }
  case 21:  {
   $command$0 = 30;$param$0 = $5;
   break;
  }
  case 22:  {
   $command$0 = 21;$param$0 = $5;
   break;
  }
  case 23:  {
   $command$0 = 22;$param$0 = $5;
   break;
  }
  case 25:  {
   $29 = $5 & 240; //@line 50 "load_mod.cpp"
   $30 = ($29|0)==(0); //@line 50 "load_mod.cpp"
   $$6 = $30 ? $5 : $29; //@line 50 "load_mod.cpp"
   $command$0 = 29;$param$0 = $$6;
   break;
  }
  case 27:  {
   $command$0 = 15;$param$0 = $5;
   break;
  }
  case 29:  {
   $command$0 = 18;$param$0 = $5;
   break;
  }
  case 33:  {
   $command$0 = 28;$param$0 = $5;
   break;
  }
  case 34:  {
   $command$0 = 27;$param$0 = $5;
   break;
  }
  case 35:  {
   $command$0 = 31;$param$0 = $5;
   break;
  }
  default: {
   $command$0 = 0;$param$0 = $5;
  }
  }
 } while(0);
 $31 = $command$0&255; //@line 58 "load_mod.cpp"
 HEAP8[$0>>0] = $31; //@line 58 "load_mod.cpp"
 $32 = $param$0&255; //@line 59 "load_mod.cpp"
 HEAP8[$3>>0] = $32; //@line 59 "load_mod.cpp"
 return; //@line 60 "load_mod.cpp"
}
function __ZN10CSoundFile19DetectUnusedSamplesEPb($this,$pbIns) {
 $this = $this|0;
 $pbIns = $pbIns|0;
 var $$0 = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $exitcond = 0, $exitcond12 = 0, $ichk$05 = 0, $ipat$011 = 0, $j$08 = 0, $k$06 = 0, $nExt$0$ = 0, $nExt$04 = 0, $nExt$1 = 0, $not$ = 0, $p$09 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($pbIns|0)==(0|0); //@line 1813 "sndfile.cpp"
 if ($0) {
  $$0 = 0;
  return ($$0|0); //@line 1857 "sndfile.cpp"
 }
 $1 = ((($this)) + 59688|0); //@line 1814 "sndfile.cpp"
 $2 = HEAP32[$1>>2]|0; //@line 1814 "sndfile.cpp"
 $3 = ($2|0)==(0); //@line 1814 "sndfile.cpp"
 if ($3) {
  $$0 = 0;
  return ($$0|0); //@line 1857 "sndfile.cpp"
 }
 _memset(($pbIns|0),0,240)|0; //@line 1816 "sndfile.cpp"
 $4 = ((($this)) + 59664|0); //@line 1822 "sndfile.cpp"
 $ipat$011 = 0;
 while(1) {
  $8 = (((($this)) + 51904|0) + ($ipat$011<<2)|0); //@line 1819 "sndfile.cpp"
  $9 = HEAP32[$8>>2]|0; //@line 1819 "sndfile.cpp"
  $10 = ($9|0)==(0|0); //@line 1820 "sndfile.cpp"
  if (!($10)) {
   $11 = (((($this)) + 52864|0) + ($ipat$011<<1)|0); //@line 1822 "sndfile.cpp"
   $12 = HEAP16[$11>>1]|0; //@line 1822 "sndfile.cpp"
   $13 = $12&65535; //@line 1822 "sndfile.cpp"
   $14 = HEAP32[$4>>2]|0; //@line 1822 "sndfile.cpp"
   $15 = Math_imul($13, $14)|0; //@line 1822 "sndfile.cpp"
   $16 = ($15|0)==(0); //@line 1823 "sndfile.cpp"
   if (!($16)) {
    $j$08 = 0;$p$09 = $9;
    while(1) {
     $17 = HEAP8[$p$09>>0]|0; //@line 1825 "sndfile.cpp"
     $18 = ($17<<24>>24)==(0); //@line 1825 "sndfile.cpp"
     do {
      if (!($18)) {
       $19 = $17&255; //@line 1825 "sndfile.cpp"
       $20 = ($17&255)<(121); //@line 1825 "sndfile.cpp"
       if ($20) {
        $21 = ((($p$09)) + 1|0); //@line 1827 "sndfile.cpp"
        $22 = HEAP8[$21>>0]|0; //@line 1827 "sndfile.cpp"
        $$off = (($22) + -1)<<24>>24; //@line 1827 "sndfile.cpp"
        $23 = ($$off&255)<(239); //@line 1827 "sndfile.cpp"
        if ($23) {
         $27 = $22&255; //@line 1827 "sndfile.cpp"
         $28 = (((($this)) + 48640|0) + ($27<<2)|0); //@line 1829 "sndfile.cpp"
         $29 = HEAP32[$28>>2]|0; //@line 1829 "sndfile.cpp"
         $30 = ($29|0)==(0|0); //@line 1830 "sndfile.cpp"
         if ($30) {
          break;
         }
         $31 = (($19) + -1)|0; //@line 1832 "sndfile.cpp"
         $32 = (((($29)) + 300|0) + ($31)|0); //@line 1832 "sndfile.cpp"
         $33 = HEAP8[$32>>0]|0; //@line 1832 "sndfile.cpp"
         $34 = ($33&255)<(240); //@line 1833 "sndfile.cpp"
         if (!($34)) {
          break;
         }
         $35 = $33&255; //@line 1832 "sndfile.cpp"
         $36 = (($pbIns) + ($35)|0); //@line 1833 "sndfile.cpp"
         HEAP8[$36>>0] = 1; //@line 1833 "sndfile.cpp"
         break;
        }
        $24 = HEAP32[$1>>2]|0; //@line 1837 "sndfile.cpp"
        $25 = ($24|0)==(0); //@line 1837 "sndfile.cpp"
        if (!($25)) {
         $26 = (($19) + -1)|0; //@line 1842 "sndfile.cpp"
         $k$06 = 1;
         while(1) {
          $37 = (((($this)) + 48640|0) + ($k$06<<2)|0); //@line 1839 "sndfile.cpp"
          $38 = HEAP32[$37>>2]|0; //@line 1839 "sndfile.cpp"
          $39 = ($38|0)==(0|0); //@line 1840 "sndfile.cpp"
          if (!($39)) {
           $40 = (((($38)) + 300|0) + ($26)|0); //@line 1842 "sndfile.cpp"
           $41 = HEAP8[$40>>0]|0; //@line 1842 "sndfile.cpp"
           $42 = ($41&255)<(240); //@line 1843 "sndfile.cpp"
           if ($42) {
            $43 = $41&255; //@line 1842 "sndfile.cpp"
            $44 = (($pbIns) + ($43)|0); //@line 1843 "sndfile.cpp"
            HEAP8[$44>>0] = 1; //@line 1843 "sndfile.cpp"
           }
          }
          $45 = (($k$06) + 1)|0; //@line 1837 "sndfile.cpp"
          $46 = ($45>>>0)>($24>>>0); //@line 1837 "sndfile.cpp"
          if ($46) {
           break;
          } else {
           $k$06 = $45;
          }
         }
        }
       }
      }
     } while(0);
     $47 = (($j$08) + 1)|0; //@line 1823 "sndfile.cpp"
     $48 = ((($p$09)) + 6|0); //@line 1823 "sndfile.cpp"
     $exitcond = ($47|0)==($15|0); //@line 1823 "sndfile.cpp"
     if ($exitcond) {
      break;
     } else {
      $j$08 = $47;$p$09 = $48;
     }
    }
   }
  }
  $49 = (($ipat$011) + 1)|0; //@line 1817 "sndfile.cpp"
  $exitcond12 = ($49|0)==(240); //@line 1817 "sndfile.cpp"
  if ($exitcond12) {
   break;
  } else {
   $ipat$011 = $49;
  }
 }
 $5 = ((($this)) + 59684|0); //@line 1851 "sndfile.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 1851 "sndfile.cpp"
 $7 = ($6|0)==(0); //@line 1851 "sndfile.cpp"
 if ($7) {
  $$0 = 0;
  return ($$0|0); //@line 1857 "sndfile.cpp"
 } else {
  $ichk$05 = 1;$nExt$04 = 0;
 }
 while(1) {
  $50 = (($pbIns) + ($ichk$05)|0); //@line 1853 "sndfile.cpp"
  $51 = HEAP8[$50>>0]|0; //@line 1853 "sndfile.cpp"
  $52 = ($51<<24>>24)==(0); //@line 1853 "sndfile.cpp"
  if ($52) {
   $53 = (((((($this)) + 33280|0) + ($ichk$05<<6)|0)) + 20|0); //@line 1853 "sndfile.cpp"
   $54 = HEAP32[$53>>2]|0; //@line 1853 "sndfile.cpp"
   $not$ = ($54|0)!=(0|0); //@line 1853 "sndfile.cpp"
   $55 = $not$&1; //@line 1853 "sndfile.cpp"
   $nExt$0$ = (($55) + ($nExt$04))|0; //@line 1853 "sndfile.cpp"
   $nExt$1 = $nExt$0$;
  } else {
   $nExt$1 = $nExt$04;
  }
  $56 = (($ichk$05) + 1)|0; //@line 1851 "sndfile.cpp"
  $57 = ($56>>>0)>($6>>>0); //@line 1851 "sndfile.cpp"
  if ($57) {
   $$0 = $nExt$1;
   break;
  } else {
   $ichk$05 = $56;$nExt$04 = $nExt$1;
  }
 }
 return ($$0|0); //@line 1857 "sndfile.cpp"
}
function __ZN10CSoundFile13DestroySampleEj($this,$nSample) {
 $this = $this|0;
 $nSample = $nSample|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $exitcond = 0, $i$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($nSample) + -1)|0; //@line 1879 "sndfile.cpp"
 $1 = ($0>>>0)>(238); //@line 1879 "sndfile.cpp"
 if ($1) {
  $$0 = 0;
  return ($$0|0); //@line 1896 "sndfile.cpp"
 }
 $2 = (((((($this)) + 33280|0) + ($nSample<<6)|0)) + 20|0); //@line 1880 "sndfile.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 1880 "sndfile.cpp"
 $4 = ($3|0)==(0|0); //@line 1880 "sndfile.cpp"
 if ($4) {
  $$0 = 1;
  return ($$0|0); //@line 1896 "sndfile.cpp"
 }
 HEAP32[$2>>2] = 0; //@line 1883 "sndfile.cpp"
 $5 = (((($this)) + 33280|0) + ($nSample<<6)|0); //@line 1884 "sndfile.cpp"
 HEAP32[$5>>2] = 0; //@line 1884 "sndfile.cpp"
 $6 = (((((($this)) + 33280|0) + ($nSample<<6)|0)) + 34|0); //@line 1885 "sndfile.cpp"
 $7 = HEAP16[$6>>1]|0; //@line 1885 "sndfile.cpp"
 $8 = $7&65535; //@line 1885 "sndfile.cpp"
 $9 = $8 & 65534; //@line 1885 "sndfile.cpp"
 $10 = $9&65535; //@line 1885 "sndfile.cpp"
 HEAP16[$6>>1] = $10; //@line 1885 "sndfile.cpp"
 $i$02 = 0;
 while(1) {
  $11 = (((($this) + ($i$02<<8)|0)) + 96|0); //@line 1888 "sndfile.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 1888 "sndfile.cpp"
  $13 = ($12|0)==($3|0); //@line 1888 "sndfile.cpp"
  if ($13) {
   $14 = (((($this) + ($i$02<<8)|0)) + 32|0); //@line 1890 "sndfile.cpp"
   HEAP32[$14>>2] = 0; //@line 1890 "sndfile.cpp"
   $15 = (((($this) + ($i$02<<8)|0)) + 4|0); //@line 1890 "sndfile.cpp"
   HEAP32[$15>>2] = 0; //@line 1890 "sndfile.cpp"
   $16 = (($this) + ($i$02<<8)|0); //@line 1891 "sndfile.cpp"
   HEAP32[$16>>2] = 0; //@line 1891 "sndfile.cpp"
   HEAP32[$11>>2] = 0; //@line 1891 "sndfile.cpp"
  }
  $17 = (($i$02) + 1)|0; //@line 1886 "sndfile.cpp"
  $exitcond = ($17|0)==(128); //@line 1886 "sndfile.cpp"
  if ($exitcond) {
   break;
  } else {
   $i$02 = $17;
  }
 }
 $18 = ((($3)) + -16|0); //@line 367 "sndfile.cpp"
 _free($18); //@line 367 "sndfile.cpp"
 $$0 = 1;
 return ($$0|0); //@line 1896 "sndfile.cpp"
}
function __ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$pIns,$nFlags,$lpMemFile,$dwMemLength) {
 $this = $this|0;
 $pIns = $pIns|0;
 $nFlags = $nFlags|0;
 $lpMemFile = $lpMemFile|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$0 = 0, $$10 = 0, $$11 = 0, $$12 = 0, $$13 = 0, $$4 = 0, $$lobit90 = 0, $$sum = 0, $$sum100 = 0, $$sum101 = 0, $$sum102 = 0, $$sum103 = 0, $$sum104 = 0, $$sum105 = 0, $$sum107 = 0, $$sum17 = 0, $$sum18 = 0, $$sum19 = 0, $$sum78 = 0;
 var $$sum9 = 0, $$sum99 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0;
 var $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0;
 var $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bitbuf = 0, $bitnum = 0, $delta$038 = 0, $delta16$046 = 0, $delta4$042 = 0, $delta8$081 = 0, $dlt$060 = 0, $dmax$0 = 0, $dwMemLength$ = 0, $exitcond = 0, $exitcond91 = 0, $exitcond92 = 0, $exitcond93 = 0;
 var $exitcond94 = 0, $exitcond95 = 0, $exitcond96 = 0, $exitcond97 = 0, $hibyte$0 = 0, $hibyte$0$lcssa = 0, $hibyte$1 = 0, $hibyte$2 = 0, $iadd_l$0 = 0, $iadd_l$165 = 0, $iadd_l$2 = 0, $iadd_l23$0 = 0, $iadd_l23$184 = 0, $iadd_l23$2 = 0, $iadd_r$166 = 0, $iadd_r$2 = 0, $iadd_r24$185 = 0, $iadd_r24$2 = 0, $ibuf = 0, $j$036 = 0;
 var $j12$051 = 0, $j15$055 = 0, $j18$058 = 0, $j2$037 = 0, $j22$068 = 0, $j27$087 = 0, $j30$071 = 0, $j34$076 = 0, $j37$082 = 0, $j41$062 = 0, $j5$041 = 0, $j8$045 = 0, $len$0 = 0, $len$5 = 0, $len$6 = 0, $len$615 = 0, $lowbyte$061 = 0, $lowbyte$1 = 0, $mem$0 = 0, $mem$1 = 0;
 var $or$cond = 0, $p7$047 = 0, $pSample1$039 = 0, $pSample13$053 = 0, $pSample3$043 = 0, $pSample36$080 = 0, $pSample6$048 = 0, $pSrc14$054 = 0, $psrc$067 = 0, $psrc25$086 = 0, $psrc28$070 = 0, $psrc32$075 = 0, $sext = 0, $sext106 = 0, $sext5 = 0, $sext6 = 0, $sext98 = 0, $storemerge16 = 0, dest = 0, label = 0;
 var sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ibuf = sp + 8|0;
 $bitbuf = sp + 4|0;
 $bitnum = sp;
 $0 = HEAP32[$pIns>>2]|0; //@line 1097 "sndfile.cpp"
 $1 = (($0) + 6)|0; //@line 1097 "sndfile.cpp"
 $2 = ($pIns|0)==(0|0); //@line 1100 "sndfile.cpp"
 if ($2) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 1546 "sndfile.cpp"
 }
 $3 = ($0|0)>(3); //@line 1100 "sndfile.cpp"
 $4 = ($lpMemFile|0)!=(0|0); //@line 1100 "sndfile.cpp"
 $or$cond = $4 & $3; //@line 1100 "sndfile.cpp"
 if (!($or$cond)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 1546 "sndfile.cpp"
 }
 $5 = ($0>>>0)>(16000000); //@line 1101 "sndfile.cpp"
 if ($5) {
  HEAP32[$pIns>>2] = 16000000; //@line 1101 "sndfile.cpp"
  $24 = 16000000;
 } else {
  $24 = $0;
 }
 $6 = ((($pIns)) + 34|0); //@line 1102 "sndfile.cpp"
 $7 = HEAP16[$6>>1]|0; //@line 1102 "sndfile.cpp"
 $8 = $7&65535; //@line 1102 "sndfile.cpp"
 $9 = $8 & 65470; //@line 1102 "sndfile.cpp"
 $10 = $nFlags >>> 2; //@line 1103 "sndfile.cpp"
 $11 = $10 & 1; //@line 1103 "sndfile.cpp"
 $12 = $9 | $11; //@line 1103 "sndfile.cpp"
 $mem$0 = $1 << $11; //@line 1103 "sndfile.cpp"
 $13 = $nFlags & 8; //@line 1108 "sndfile.cpp"
 $14 = $13 << 3; //@line 1108 "sndfile.cpp"
 $15 = $14 | $12; //@line 1108 "sndfile.cpp"
 $$lobit90 = $13 >>> 3; //@line 1108 "sndfile.cpp"
 $mem$1 = $mem$0 << $$lobit90; //@line 1108 "sndfile.cpp"
 $storemerge16 = $15&65535; //@line 1111 "sndfile.cpp"
 HEAP16[$6>>1] = $storemerge16; //@line 1102 "sndfile.cpp"
 $16 = (($mem$1) + 39)|0; //@line 356 "sndfile.cpp"
 $17 = $16 & -8; //@line 356 "sndfile.cpp"
 $18 = (_malloc($17)|0); //@line 101 "./libmodplug/stdafx.h"
 $19 = ($18|0)==(0|0); //@line 103 "./libmodplug/stdafx.h"
 if ($19) {
  $20 = ((($pIns)) + 20|0); //@line 1113 "sndfile.cpp"
  HEAP32[$20>>2] = 0; //@line 1113 "sndfile.cpp"
  HEAP32[$pIns>>2] = 0; //@line 1115 "sndfile.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 1546 "sndfile.cpp"
 }
 _memset(($18|0),0,($17|0))|0; //@line 103 "./libmodplug/stdafx.h"
 $21 = ((($18)) + 16|0); //@line 357 "sndfile.cpp"
 $22 = ((($pIns)) + 20|0); //@line 1113 "sndfile.cpp"
 HEAP32[$22>>2] = $21; //@line 1113 "sndfile.cpp"
 L14: do {
  switch ($nFlags|0) {
  case 1:  {
   $23 = ($24>>>0)>($dwMemLength>>>0); //@line 1124 "sndfile.cpp"
   if ($23) {
    HEAP32[$pIns>>2] = $dwMemLength; //@line 1124 "sndfile.cpp"
    $len$0 = $dwMemLength;
   } else {
    $len$0 = $24;
   }
   $25 = ($len$0|0)==(0); //@line 1126 "sndfile.cpp"
   if ($25) {
    $len$6 = 0;
    label = 86;
   } else {
    $j$036 = 0;
    while(1) {
     $26 = (($lpMemFile) + ($j$036)|0); //@line 1126 "sndfile.cpp"
     $27 = HEAP8[$26>>0]|0; //@line 1126 "sndfile.cpp"
     $28 = $27&255; //@line 1126 "sndfile.cpp"
     $29 = (($28) + 128)|0; //@line 1126 "sndfile.cpp"
     $30 = $29&255; //@line 1126 "sndfile.cpp"
     $$sum107 = (($j$036) + 16)|0; //@line 1126 "sndfile.cpp"
     $31 = (($18) + ($$sum107)|0); //@line 1126 "sndfile.cpp"
     HEAP8[$31>>0] = $30; //@line 1126 "sndfile.cpp"
     $32 = (($j$036) + 1)|0; //@line 1126 "sndfile.cpp"
     $exitcond = ($32|0)==($len$0|0); //@line 1126 "sndfile.cpp"
     if ($exitcond) {
      $len$6 = $len$0;
      label = 86;
      break;
     } else {
      $j$036 = $32;
     }
    }
   }
   break;
  }
  case 2:  {
   $33 = ($24>>>0)>($dwMemLength>>>0); //@line 1134 "sndfile.cpp"
   if ($33) {
    $len$6 = $24;
    label = 86;
   } else {
    $34 = ($24|0)==(0); //@line 1139 "sndfile.cpp"
    if ($34) {
     $len$6 = 0;
     label = 86;
    } else {
     $delta$038 = 0;$j2$037 = 0;$pSample1$039 = $21;
     while(1) {
      $35 = (($lpMemFile) + ($j2$037)|0); //@line 1141 "sndfile.cpp"
      $36 = HEAP8[$35>>0]|0; //@line 1141 "sndfile.cpp"
      $37 = $36 << 24 >> 24; //@line 1141 "sndfile.cpp"
      $38 = (($37) + ($delta$038))|0; //@line 1141 "sndfile.cpp"
      $39 = $38&255; //@line 1142 "sndfile.cpp"
      $40 = ((($pSample1$039)) + 1|0); //@line 1142 "sndfile.cpp"
      HEAP8[$pSample1$039>>0] = $39; //@line 1142 "sndfile.cpp"
      $41 = (($j2$037) + 1)|0; //@line 1139 "sndfile.cpp"
      $exitcond91 = ($41|0)==($24|0); //@line 1139 "sndfile.cpp"
      if ($exitcond91) {
       $len$6 = $24;
       label = 86;
       break;
      } else {
       $delta$038 = $38;$j2$037 = $41;$pSample1$039 = $40;
      }
     }
    }
   }
   break;
  }
  case 3:  {
   $42 = (($24) + 1)|0; //@line 1150 "sndfile.cpp"
   $43 = $42 >>> 1; //@line 1150 "sndfile.cpp"
   $44 = (($dwMemLength) + -16)|0; //@line 1151 "sndfile.cpp"
   $45 = ($43>>>0)>($44>>>0); //@line 1151 "sndfile.cpp"
   if ($45) {
    $len$6 = $43;
    label = 86;
   } else {
    $46 = ((($this)) + 67484|0); //@line 1152 "sndfile.cpp"
    dest=$46; src=$lpMemFile; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 1152 "sndfile.cpp"
    $47 = ($43|0)==(0); //@line 1156 "sndfile.cpp"
    if (!($47)) {
     $48 = HEAP32[$22>>2]|0; //@line 1154 "sndfile.cpp"
     $delta4$042 = 0;$j5$041 = 0;$pSample3$043 = $48;
     while(1) {
      $$sum9 = (($j5$041) + 16)|0; //@line 1158 "sndfile.cpp"
      $49 = (($lpMemFile) + ($$sum9)|0); //@line 1158 "sndfile.cpp"
      $50 = HEAP8[$49>>0]|0; //@line 1158 "sndfile.cpp"
      $51 = $50&255; //@line 1159 "sndfile.cpp"
      $52 = $51 >>> 4; //@line 1159 "sndfile.cpp"
      $53 = $51 & 15; //@line 750 "./libmodplug/sndfile.h"
      $54 = (((($this)) + 67484|0) + ($53)|0); //@line 750 "./libmodplug/sndfile.h"
      $55 = HEAP8[$54>>0]|0; //@line 750 "./libmodplug/sndfile.h"
      $56 = $55&255; //@line 750 "./libmodplug/sndfile.h"
      $57 = (($56) + ($delta4$042))|0; //@line 750 "./libmodplug/sndfile.h"
      $58 = $57&255; //@line 750 "./libmodplug/sndfile.h"
      HEAP8[$pSample3$043>>0] = $58; //@line 1161 "sndfile.cpp"
      $59 = $57 & 255; //@line 750 "./libmodplug/sndfile.h"
      $60 = (((($this)) + 67484|0) + ($52)|0); //@line 750 "./libmodplug/sndfile.h"
      $61 = HEAP8[$60>>0]|0; //@line 750 "./libmodplug/sndfile.h"
      $62 = $61&255; //@line 750 "./libmodplug/sndfile.h"
      $63 = (($59) + ($62))|0; //@line 750 "./libmodplug/sndfile.h"
      $64 = $63&255; //@line 750 "./libmodplug/sndfile.h"
      $65 = ((($pSample3$043)) + 1|0); //@line 1163 "sndfile.cpp"
      HEAP8[$65>>0] = $64; //@line 1163 "sndfile.cpp"
      $66 = ((($pSample3$043)) + 2|0); //@line 1164 "sndfile.cpp"
      $67 = (($j5$041) + 1)|0; //@line 1156 "sndfile.cpp"
      $exitcond92 = ($67|0)==($43|0); //@line 1156 "sndfile.cpp"
      if ($exitcond92) {
       break;
      } else {
       $delta4$042 = $63;$j5$041 = $67;$pSample3$043 = $66;
      }
     }
    }
    $68 = (($43) + 16)|0; //@line 1166 "sndfile.cpp"
    $len$6 = $68;
    label = 86;
   }
   break;
  }
  case 4:  {
   $69 = $24 << 1; //@line 1173 "sndfile.cpp"
   $70 = ($69>>>0)>($dwMemLength>>>0); //@line 1174 "sndfile.cpp"
   if ($70) {
    $len$6 = $69;
    label = 86;
   } else {
    $71 = ($69|0)==(0); //@line 1178 "sndfile.cpp"
    if ($71) {
     $len$6 = 0;
     label = 86;
    } else {
     $delta16$046 = 0;$j8$045 = 0;$p7$047 = $lpMemFile;$pSample6$048 = $21;
     while(1) {
      $72 = ((($p7$047)) + 2|0); //@line 1180 "sndfile.cpp"
      $73 = HEAP16[$p7$047>>1]|0; //@line 1180 "sndfile.cpp"
      $74 = $73 << 16 >> 16; //@line 1180 "sndfile.cpp"
      $75 = (($74) + ($delta16$046))|0; //@line 1180 "sndfile.cpp"
      $76 = $75&65535; //@line 1181 "sndfile.cpp"
      $77 = ((($pSample6$048)) + 2|0); //@line 1181 "sndfile.cpp"
      HEAP16[$pSample6$048>>1] = $76; //@line 1181 "sndfile.cpp"
      $78 = (($j8$045) + 2)|0; //@line 1178 "sndfile.cpp"
      $79 = ($78>>>0)<($69>>>0); //@line 1178 "sndfile.cpp"
      if ($79) {
       $delta16$046 = $75;$j8$045 = $78;$p7$047 = $72;$pSample6$048 = $77;
      } else {
       $len$6 = $69;
       label = 86;
       break;
      }
     }
    }
   }
   break;
  }
  case 5:  {
   $80 = $24 << 1; //@line 1189 "sndfile.cpp"
   $81 = ($80>>>0)>($dwMemLength>>>0); //@line 1190 "sndfile.cpp"
   if ($81) {
    $len$6 = $80;
    label = 86;
   } else {
    _memcpy(($21|0),($lpMemFile|0),($80|0))|0; //@line 1190 "sndfile.cpp"
    $len$6 = $80;
    label = 86;
   }
   break;
  }
  case 7:  {
   $82 = $24 << 1; //@line 1202 "sndfile.cpp"
   $83 = ($82>>>0)>($dwMemLength>>>0); //@line 1203 "sndfile.cpp"
   $84 = $dwMemLength & -2; //@line 1203 "sndfile.cpp"
   $$10 = $83 ? $84 : $82; //@line 1203 "sndfile.cpp"
   $85 = ($$10>>>0)>(1); //@line 1204 "sndfile.cpp"
   if ($85) {
    $j12$051 = 0;
    while(1) {
     $86 = (($lpMemFile) + ($j12$051)|0); //@line 1212 "sndfile.cpp"
     $87 = HEAP16[$86>>1]|0; //@line 1212 "sndfile.cpp"
     $88 = $87&65535; //@line 1212 "sndfile.cpp"
     $89 = $88 >>> 8; //@line 1212 "sndfile.cpp"
     $90 = $88 << 8; //@line 1212 "sndfile.cpp"
     $91 = $89 | $90; //@line 1212 "sndfile.cpp"
     $92 = $91&65535; //@line 1212 "sndfile.cpp"
     $$sum19 = (($j12$051) + 16)|0; //@line 1212 "sndfile.cpp"
     $93 = (($18) + ($$sum19)|0); //@line 1212 "sndfile.cpp"
     HEAP16[$93>>1] = $92; //@line 1212 "sndfile.cpp"
     $94 = (($j12$051) + 2)|0; //@line 1208 "sndfile.cpp"
     $95 = ($94>>>0)<($$10>>>0); //@line 1208 "sndfile.cpp"
     if ($95) {
      $j12$051 = $94;
     } else {
      $len$615 = $$10;
      break;
     }
    }
   } else {
    $len$615 = $$10;
   }
   break;
  }
  case 6:  {
   $96 = $24 << 1; //@line 1220 "sndfile.cpp"
   $97 = ($96>>>0)>($dwMemLength>>>0); //@line 1221 "sndfile.cpp"
   if ($97) {
    $len$6 = $96;
    label = 86;
   } else {
    $98 = ($96|0)==(0); //@line 1224 "sndfile.cpp"
    if ($98) {
     $len$6 = 0;
     label = 86;
    } else {
     $j15$055 = 0;$pSample13$053 = $21;$pSrc14$054 = $lpMemFile;
     while(1) {
      $99 = ((($pSrc14$054)) + 2|0); //@line 1224 "sndfile.cpp"
      $100 = HEAP16[$pSrc14$054>>1]|0; //@line 1224 "sndfile.cpp"
      $101 = $100&65535; //@line 1224 "sndfile.cpp"
      $102 = (($101) + 32768)|0; //@line 1224 "sndfile.cpp"
      $103 = $102&65535; //@line 1224 "sndfile.cpp"
      $104 = ((($pSample13$053)) + 2|0); //@line 1224 "sndfile.cpp"
      HEAP16[$pSample13$053>>1] = $103; //@line 1224 "sndfile.cpp"
      $105 = (($j15$055) + 2)|0; //@line 1224 "sndfile.cpp"
      $106 = ($105>>>0)<($96>>>0); //@line 1224 "sndfile.cpp"
      if ($106) {
       $j15$055 = $105;$pSample13$053 = $104;$pSrc14$054 = $99;
      } else {
       $len$6 = $96;
       label = 86;
       break;
      }
     }
    }
   }
   break;
  }
  case 15:  {
   $107 = $24 << 1; //@line 1230 "sndfile.cpp"
   $108 = $24 << 2; //@line 1231 "sndfile.cpp"
   $109 = ($108>>>0)>($dwMemLength>>>0); //@line 1231 "sndfile.cpp"
   if ($109) {
    $len$6 = $107;
    label = 86;
   } else {
    $110 = ($107|0)==(0); //@line 1235 "sndfile.cpp"
    if ($110) {
     $len$6 = $108;
     label = 86;
    } else {
     $j18$058 = 0;
     while(1) {
      $111 = (($lpMemFile) + ($j18$058)|0); //@line 1241 "sndfile.cpp"
      $112 = HEAP16[$111>>1]|0; //@line 1241 "sndfile.cpp"
      $113 = $112&65535; //@line 1241 "sndfile.cpp"
      $114 = $113 >>> 8; //@line 1241 "sndfile.cpp"
      $115 = $113 << 8; //@line 1241 "sndfile.cpp"
      $116 = $114 | $115; //@line 1241 "sndfile.cpp"
      $117 = $116&65535; //@line 1241 "sndfile.cpp"
      $118 = $j18$058 << 1; //@line 1241 "sndfile.cpp"
      $$sum17 = (($118) + 16)|0; //@line 1241 "sndfile.cpp"
      $119 = (($18) + ($$sum17)|0); //@line 1241 "sndfile.cpp"
      HEAP16[$119>>1] = $117; //@line 1241 "sndfile.cpp"
      $$sum = (($j18$058) + ($107))|0; //@line 1242 "sndfile.cpp"
      $120 = (($lpMemFile) + ($$sum)|0); //@line 1242 "sndfile.cpp"
      $121 = HEAP16[$120>>1]|0; //@line 1242 "sndfile.cpp"
      $122 = $121&65535; //@line 1242 "sndfile.cpp"
      $123 = $122 >>> 8; //@line 1242 "sndfile.cpp"
      $124 = $122 << 8; //@line 1242 "sndfile.cpp"
      $125 = $123 | $124; //@line 1242 "sndfile.cpp"
      $126 = $125&65535; //@line 1242 "sndfile.cpp"
      $$sum78 = $118 | 2; //@line 1242 "sndfile.cpp"
      $$sum18 = (($$sum78) + 16)|0; //@line 1242 "sndfile.cpp"
      $127 = (($18) + ($$sum18)|0); //@line 1242 "sndfile.cpp"
      HEAP16[$127>>1] = $126; //@line 1242 "sndfile.cpp"
      $128 = (($j18$058) + 2)|0; //@line 1235 "sndfile.cpp"
      $129 = ($128>>>0)<($107>>>0); //@line 1235 "sndfile.cpp"
      if ($129) {
       $j18$058 = $128;
      } else {
       $len$6 = $108;
       label = 86;
       break;
      }
     }
    }
   }
   break;
  }
  case 19: case 23:  {
   $277 = ($dwMemLength>>>0)>(3); //@line 1435 "sndfile.cpp"
   if ($277) {
    $$13 = $24 << $11; //@line 1438 "sndfile.cpp"
    $278 = (($lpMemFile) + ($dwMemLength)|0); //@line 1439 "sndfile.cpp"
    $279 = (__Z9DMFUnpackPhS_S_j($21,$lpMemFile,$278,$$13)|0); //@line 1440 "sndfile.cpp"
    $len$6 = $279;
    label = 86;
   } else {
    $len$615 = $dwMemLength;
   }
   break;
  }
  case 32: case 36:  {
   $240 = ($dwMemLength>>>0)>(3); //@line 1396 "sndfile.cpp"
   if ($240) {
    $241 = HEAP32[$lpMemFile>>2]|0; //@line 1400 "sndfile.cpp"
    HEAP32[$bitbuf>>2] = $241; //@line 1400 "sndfile.cpp"
    HEAP32[$bitnum>>2] = 32; //@line 1401 "sndfile.cpp"
    $242 = ((($lpMemFile)) + 4|0); //@line 1403 "sndfile.cpp"
    HEAP32[$ibuf>>2] = $242; //@line 1403 "sndfile.cpp"
    $243 = ($24|0)==(0); //@line 1404 "sndfile.cpp"
    if ($243) {
     $len$615 = $dwMemLength;
    } else {
     $244 = ($nFlags|0)==(36); //@line 1408 "sndfile.cpp"
     $dlt$060 = 0;$j41$062 = 0;$lowbyte$061 = 0;
     while(1) {
      if ($244) {
       $245 = (__Z11MDLReadBitsRjS_RPha($bitbuf,$bitnum,$ibuf,8)|0); //@line 1408 "sndfile.cpp"
       $246 = $245&255; //@line 1408 "sndfile.cpp"
       $lowbyte$1 = $246;
      } else {
       $lowbyte$1 = $lowbyte$061;
      }
      $247 = (__Z11MDLReadBitsRjS_RPha($bitbuf,$bitnum,$ibuf,1)|0); //@line 1409 "sndfile.cpp"
      $248 = $247&255; //@line 1409 "sndfile.cpp"
      $249 = (__Z11MDLReadBitsRjS_RPha($bitbuf,$bitnum,$ibuf,1)|0); //@line 1410 "sndfile.cpp"
      $250 = ($249<<16>>16)==(0); //@line 1410 "sndfile.cpp"
      if ($250) {
       $hibyte$0 = 8;
       while(1) {
        $253 = (__Z11MDLReadBitsRjS_RPha($bitbuf,$bitnum,$ibuf,1)|0); //@line 1416 "sndfile.cpp"
        $254 = ($253<<16>>16)==(0); //@line 1416 "sndfile.cpp"
        $255 = $hibyte$0 & 255; //@line 1416 "sndfile.cpp"
        $256 = (($255) + 16)|0; //@line 1416 "sndfile.cpp"
        if ($254) {
         $hibyte$0 = $256;
        } else {
         $hibyte$0$lcssa = $hibyte$0;
         break;
        }
       }
       $257 = (__Z11MDLReadBitsRjS_RPha($bitbuf,$bitnum,$ibuf,4)|0); //@line 1417 "sndfile.cpp"
       $258 = $257&65535; //@line 1417 "sndfile.cpp"
       $259 = (($258) + ($hibyte$0$lcssa))|0; //@line 1417 "sndfile.cpp"
       $260 = $259&255; //@line 1417 "sndfile.cpp"
       $hibyte$1 = $260;
      } else {
       $251 = (__Z11MDLReadBitsRjS_RPha($bitbuf,$bitnum,$ibuf,3)|0); //@line 1412 "sndfile.cpp"
       $252 = $251&255; //@line 1412 "sndfile.cpp"
       $hibyte$1 = $252;
      }
      $261 = ($248<<24>>24)==(0); //@line 1419 "sndfile.cpp"
      if ($261) {
       $hibyte$2 = $hibyte$1;
      } else {
       $262 = $hibyte$1&255; //@line 1419 "sndfile.cpp"
       $263 = $262 ^ 255; //@line 1419 "sndfile.cpp"
       $264 = $263&255; //@line 1419 "sndfile.cpp"
       $hibyte$2 = $264;
      }
      $265 = $hibyte$2&255; //@line 1420 "sndfile.cpp"
      $266 = $dlt$060 & 255; //@line 1420 "sndfile.cpp"
      $267 = (($265) + ($266))|0; //@line 1420 "sndfile.cpp"
      $268 = $267&255; //@line 1420 "sndfile.cpp"
      if ($244) {
       $270 = $j41$062 << 1; //@line 1425 "sndfile.cpp"
       $$sum100 = (($270) + 16)|0; //@line 1425 "sndfile.cpp"
       $271 = (($18) + ($$sum100)|0); //@line 1425 "sndfile.cpp"
       HEAP8[$271>>0] = $lowbyte$1; //@line 1425 "sndfile.cpp"
       $272 = $270 | 1; //@line 1426 "sndfile.cpp"
       $$sum101 = (($272) + 16)|0; //@line 1426 "sndfile.cpp"
       $273 = (($18) + ($$sum101)|0); //@line 1426 "sndfile.cpp"
       HEAP8[$273>>0] = $268; //@line 1426 "sndfile.cpp"
      } else {
       $$sum99 = (($j41$062) + 16)|0; //@line 1422 "sndfile.cpp"
       $269 = (($18) + ($$sum99)|0); //@line 1422 "sndfile.cpp"
       HEAP8[$269>>0] = $268; //@line 1422 "sndfile.cpp"
      }
      $274 = (($j41$062) + 1)|0; //@line 1404 "sndfile.cpp"
      $275 = HEAP32[$pIns>>2]|0; //@line 1404 "sndfile.cpp"
      $276 = ($274>>>0)<($275>>>0); //@line 1404 "sndfile.cpp"
      if ($276) {
       $dlt$060 = $267;$j41$062 = $274;$lowbyte$061 = $lowbyte$1;
      } else {
       $len$615 = $dwMemLength;
       break;
      }
     }
    }
   } else {
    $len$615 = $dwMemLength;
   }
   break;
  }
  case 8: case 10:  {
   $iadd_l$0 = 0;
   label = 36;
   break;
  }
  case 9:  {
   $iadd_l$0 = -128;
   label = 36;
   break;
  }
  case 22: case 18: case 20: case 16:  {
   $172 = ($dwMemLength>>>0)<(4); //@line 1306 "sndfile.cpp"
   if ($172) {
    $len$615 = $dwMemLength;
   } else {
    $173 = $nFlags & -3; //@line 1307 "sndfile.cpp"
    $174 = ($173|0)==(16); //@line 1307 "sndfile.cpp"
    if ($174) {
     $175 = ($nFlags|0)==(18); //@line 1307 "sndfile.cpp"
     __Z12ITUnpack8BitPajPhjb($21,$24,$lpMemFile,$dwMemLength,$175); //@line 1308 "sndfile.cpp"
     $len$615 = $dwMemLength;
     break L14;
    } else {
     $176 = ($nFlags|0)==(22); //@line 1310 "sndfile.cpp"
     __Z13ITUnpack16BitPajPhjb($21,$24,$lpMemFile,$dwMemLength,$176); //@line 1310 "sndfile.cpp"
     $len$615 = $dwMemLength;
     break L14;
    }
   }
   break;
  }
  case 73: case 72:  {
   $177 = ($nFlags|0)==(73); //@line 1320 "sndfile.cpp"
   $$ = $177 ? -128 : 0; //@line 1320 "sndfile.cpp"
   $178 = $24 << 1; //@line 1322 "sndfile.cpp"
   $179 = ($178>>>0)>($dwMemLength>>>0); //@line 1322 "sndfile.cpp"
   $180 = $dwMemLength >>> 1; //@line 1322 "sndfile.cpp"
   $$11 = $179 ? $180 : $24; //@line 1322 "sndfile.cpp"
   $181 = ($$11|0)==(0); //@line 1325 "sndfile.cpp"
   if (!($181)) {
    $j30$071 = 0;$psrc28$070 = $lpMemFile;
    while(1) {
     $182 = HEAP8[$psrc28$070>>0]|0; //@line 1327 "sndfile.cpp"
     $183 = $182&255; //@line 1327 "sndfile.cpp"
     $184 = (($183) + ($$))|0; //@line 1327 "sndfile.cpp"
     $185 = $184&255; //@line 1327 "sndfile.cpp"
     $186 = $j30$071 << 1; //@line 1327 "sndfile.cpp"
     $$sum102 = (($186) + 16)|0; //@line 1327 "sndfile.cpp"
     $187 = (($18) + ($$sum102)|0); //@line 1327 "sndfile.cpp"
     HEAP8[$187>>0] = $185; //@line 1327 "sndfile.cpp"
     $188 = ((($psrc28$070)) + 1|0); //@line 1328 "sndfile.cpp"
     $189 = HEAP8[$188>>0]|0; //@line 1328 "sndfile.cpp"
     $190 = $189&255; //@line 1328 "sndfile.cpp"
     $191 = (($190) + ($$))|0; //@line 1328 "sndfile.cpp"
     $192 = $191&255; //@line 1328 "sndfile.cpp"
     $193 = $186 | 1; //@line 1328 "sndfile.cpp"
     $$sum103 = (($193) + 16)|0; //@line 1328 "sndfile.cpp"
     $194 = (($18) + ($$sum103)|0); //@line 1328 "sndfile.cpp"
     HEAP8[$194>>0] = $192; //@line 1328 "sndfile.cpp"
     $195 = ((($psrc28$070)) + 2|0); //@line 1329 "sndfile.cpp"
     $196 = (($j30$071) + 1)|0; //@line 1325 "sndfile.cpp"
     $exitcond94 = ($196|0)==($$11|0); //@line 1325 "sndfile.cpp"
     if ($exitcond94) {
      break;
     } else {
      $j30$071 = $196;$psrc28$070 = $195;
     }
    }
   }
   $197 = $$11 << 1; //@line 1331 "sndfile.cpp"
   $len$6 = $197;
   label = 86;
   break;
  }
  case 78: case 77:  {
   $198 = ($nFlags|0)==(78); //@line 1340 "sndfile.cpp"
   $$4 = $198 ? -32768 : 0; //@line 1340 "sndfile.cpp"
   $199 = $24 << 2; //@line 1342 "sndfile.cpp"
   $200 = ($199>>>0)>($dwMemLength>>>0); //@line 1342 "sndfile.cpp"
   $201 = $dwMemLength >>> 2; //@line 1342 "sndfile.cpp"
   $$12 = $200 ? $201 : $24; //@line 1342 "sndfile.cpp"
   $202 = ($$12|0)==(0); //@line 1345 "sndfile.cpp"
   if (!($202)) {
    $j34$076 = 0;$psrc32$075 = $lpMemFile;
    while(1) {
     $203 = HEAP16[$psrc32$075>>1]|0; //@line 1347 "sndfile.cpp"
     $204 = $203&65535; //@line 1347 "sndfile.cpp"
     $205 = (($204) + ($$4))|0; //@line 1347 "sndfile.cpp"
     $206 = $205&65535; //@line 1347 "sndfile.cpp"
     $207 = $j34$076 << 1; //@line 1347 "sndfile.cpp"
     $208 = (($21) + ($207<<1)|0); //@line 1347 "sndfile.cpp"
     HEAP16[$208>>1] = $206; //@line 1347 "sndfile.cpp"
     $209 = ((($psrc32$075)) + 2|0); //@line 1348 "sndfile.cpp"
     $210 = HEAP16[$209>>1]|0; //@line 1348 "sndfile.cpp"
     $211 = $210&65535; //@line 1348 "sndfile.cpp"
     $212 = (($211) + ($$4))|0; //@line 1348 "sndfile.cpp"
     $213 = $212&65535; //@line 1348 "sndfile.cpp"
     $214 = $207 | 1; //@line 1348 "sndfile.cpp"
     $215 = (($21) + ($214<<1)|0); //@line 1348 "sndfile.cpp"
     HEAP16[$215>>1] = $213; //@line 1348 "sndfile.cpp"
     $216 = ((($psrc32$075)) + 4|0); //@line 1349 "sndfile.cpp"
     $217 = (($j34$076) + 1)|0; //@line 1345 "sndfile.cpp"
     $exitcond95 = ($217|0)==($$12|0); //@line 1345 "sndfile.cpp"
     if ($exitcond95) {
      break;
     } else {
      $j34$076 = $217;$psrc32$075 = $216;
     }
    }
   }
   $218 = $$12 << 2; //@line 1351 "sndfile.cpp"
   $len$6 = $218;
   label = 86;
   break;
  }
  case 21: case 17:  {
   $219 = ($dwMemLength>>>0)>(9); //@line 1359 "sndfile.cpp"
   if ($219) {
    $220 = ((($lpMemFile)) + 8|0); //@line 1362 "sndfile.cpp"
    $221 = HEAP8[$220>>0]|0; //@line 1362 "sndfile.cpp"
    $222 = ((($lpMemFile)) + 4|0); //@line 1363 "sndfile.cpp"
    $223 = HEAP32[$222>>2]|0; //@line 1363 "sndfile.cpp"
    $224 = (($223) + 9)|0; //@line 1363 "sndfile.cpp"
    $225 = ($224>>>0)>($dwMemLength>>>0); //@line 1364 "sndfile.cpp"
    $dwMemLength$ = $225 ? $dwMemLength : $224; //@line 1364 "sndfile.cpp"
    $dmax$0 = $24 << $11; //@line 1366 "sndfile.cpp"
    $226 = ((($lpMemFile)) + 9|0); //@line 1367 "sndfile.cpp"
    $227 = (($dwMemLength$) + -9)|0; //@line 1367 "sndfile.cpp"
    __Z9AMSUnpackPKcjPcjc($226,$227,$21,$dmax$0,$221); //@line 1367 "sndfile.cpp"
    $len$615 = $dwMemLength$;
   } else {
    $len$6 = 9;
    label = 86;
   }
   break;
  }
  case 37:  {
   $228 = $24 << 1; //@line 1374 "sndfile.cpp"
   $229 = ($228>>>0)>($dwMemLength>>>0); //@line 1375 "sndfile.cpp"
   if ($229) {
    $len$6 = $228;
    label = 86;
   } else {
    $230 = ($228|0)==(0); //@line 1378 "sndfile.cpp"
    if ($230) {
     $len$6 = 0;
     label = 86;
    } else {
     $231 = $24 << 1; //@line 1378 "sndfile.cpp"
     $delta8$081 = 0;$j37$082 = 0;$pSample36$080 = $21;
     while(1) {
      $232 = (($lpMemFile) + ($j37$082)|0); //@line 1380 "sndfile.cpp"
      $233 = HEAP8[$232>>0]|0; //@line 1380 "sndfile.cpp"
      $234 = $233 << 24 >> 24; //@line 1380 "sndfile.cpp"
      $sext = $delta8$081 << 24; //@line 1380 "sndfile.cpp"
      $235 = $sext >> 24; //@line 1380 "sndfile.cpp"
      $236 = (($234) + ($235))|0; //@line 1380 "sndfile.cpp"
      $237 = $236&255; //@line 1380 "sndfile.cpp"
      $238 = ((($pSample36$080)) + 1|0); //@line 1381 "sndfile.cpp"
      HEAP8[$pSample36$080>>0] = $237; //@line 1381 "sndfile.cpp"
      $239 = (($j37$082) + 1)|0; //@line 1378 "sndfile.cpp"
      $exitcond96 = ($239|0)==($231|0); //@line 1378 "sndfile.cpp"
      if ($exitcond96) {
       $len$6 = $228;
       label = 86;
       break;
      } else {
       $delta8$081 = $236;$j37$082 = $239;$pSample36$080 = $238;
      }
     }
    }
   }
   break;
  }
  case 14:  {
   $iadd_l23$0 = -32768;
   label = 43;
   break;
  }
  case 13: case 12:  {
   $iadd_l23$0 = 0;
   label = 43;
   break;
  }
  default: {
   $280 = ($24>>>0)>($dwMemLength>>>0); //@line 1531 "sndfile.cpp"
   if ($280) {
    HEAP32[$pIns>>2] = $dwMemLength; //@line 1531 "sndfile.cpp"
    $len$5 = $dwMemLength;
   } else {
    $len$5 = $24;
   }
   _memcpy(($21|0),($lpMemFile|0),($len$5|0))|0; //@line 1532 "sndfile.cpp"
   $len$6 = $len$5;
   label = 86;
  }
  }
 } while(0);
 if ((label|0) == 36) {
  $130 = $24 << 1; //@line 1258 "sndfile.cpp"
  $131 = ($130>>>0)>($dwMemLength>>>0); //@line 1258 "sndfile.cpp"
  if ($131) {
   $len$6 = $24;
   label = 86;
  } else {
   $132 = ($24|0)==(0); //@line 1259 "sndfile.cpp"
   if ($132) {
    $len$6 = $130;
    label = 86;
   } else {
    $133 = ($nFlags|0)==(10); //@line 1264 "sndfile.cpp"
    $iadd_l$165 = $iadd_l$0;$iadd_r$166 = $iadd_l$0;$j22$068 = 0;$psrc$067 = $lpMemFile;
    while(1) {
     $134 = HEAP8[$psrc$067>>0]|0; //@line 1261 "sndfile.cpp"
     $135 = $134&255; //@line 1261 "sndfile.cpp"
     $136 = (($135) + ($iadd_l$165))|0; //@line 1261 "sndfile.cpp"
     $137 = $136&255; //@line 1261 "sndfile.cpp"
     $138 = $j22$068 << 1; //@line 1261 "sndfile.cpp"
     $$sum104 = (($138) + 16)|0; //@line 1261 "sndfile.cpp"
     $139 = (($18) + ($$sum104)|0); //@line 1261 "sndfile.cpp"
     HEAP8[$139>>0] = $137; //@line 1261 "sndfile.cpp"
     $140 = (($psrc$067) + ($24)|0); //@line 1262 "sndfile.cpp"
     $141 = HEAP8[$140>>0]|0; //@line 1262 "sndfile.cpp"
     $142 = $141&255; //@line 1262 "sndfile.cpp"
     $143 = (($142) + ($iadd_r$166))|0; //@line 1262 "sndfile.cpp"
     $144 = $143&255; //@line 1262 "sndfile.cpp"
     $145 = $138 | 1; //@line 1262 "sndfile.cpp"
     $$sum105 = (($145) + 16)|0; //@line 1262 "sndfile.cpp"
     $146 = (($18) + ($$sum105)|0); //@line 1262 "sndfile.cpp"
     HEAP8[$146>>0] = $144; //@line 1262 "sndfile.cpp"
     $147 = ((($psrc$067)) + 1|0); //@line 1263 "sndfile.cpp"
     if ($133) {
      $sext106 = $136 << 24; //@line 1266 "sndfile.cpp"
      $148 = $sext106 >> 24; //@line 1266 "sndfile.cpp"
      $sext6 = $143 << 24; //@line 1267 "sndfile.cpp"
      $149 = $sext6 >> 24; //@line 1267 "sndfile.cpp"
      $iadd_l$2 = $148;$iadd_r$2 = $149;
     } else {
      $iadd_l$2 = $iadd_l$165;$iadd_r$2 = $iadd_r$166;
     }
     $150 = (($j22$068) + 1)|0; //@line 1259 "sndfile.cpp"
     $exitcond93 = ($150|0)==($24|0); //@line 1259 "sndfile.cpp"
     if ($exitcond93) {
      $len$6 = $130;
      label = 86;
      break;
     } else {
      $iadd_l$165 = $iadd_l$2;$iadd_r$166 = $iadd_r$2;$j22$068 = $150;$psrc$067 = $147;
     }
    }
   }
  }
 }
 else if ((label|0) == 43) {
  $151 = $24 << 2; //@line 1284 "sndfile.cpp"
  $152 = ($151>>>0)>($dwMemLength>>>0); //@line 1284 "sndfile.cpp"
  if ($152) {
   $len$6 = $24;
   label = 86;
  } else {
   $153 = ($24|0)==(0); //@line 1285 "sndfile.cpp"
   if ($153) {
    $len$6 = $151;
    label = 86;
   } else {
    $154 = ($nFlags|0)==(12); //@line 1290 "sndfile.cpp"
    $iadd_l23$184 = $iadd_l23$0;$iadd_r24$185 = $iadd_l23$0;$j27$087 = 0;$psrc25$086 = $lpMemFile;
    while(1) {
     $155 = HEAP16[$psrc25$086>>1]|0; //@line 1287 "sndfile.cpp"
     $156 = $155&65535; //@line 1287 "sndfile.cpp"
     $157 = (($156) + ($iadd_l23$184))|0; //@line 1287 "sndfile.cpp"
     $158 = $157&65535; //@line 1287 "sndfile.cpp"
     $159 = $j27$087 << 1; //@line 1287 "sndfile.cpp"
     $160 = (($21) + ($159<<1)|0); //@line 1287 "sndfile.cpp"
     HEAP16[$160>>1] = $158; //@line 1287 "sndfile.cpp"
     $161 = (($psrc25$086) + ($24<<1)|0); //@line 1288 "sndfile.cpp"
     $162 = HEAP16[$161>>1]|0; //@line 1288 "sndfile.cpp"
     $163 = $162&65535; //@line 1288 "sndfile.cpp"
     $164 = (($163) + ($iadd_r24$185))|0; //@line 1288 "sndfile.cpp"
     $165 = $164&65535; //@line 1288 "sndfile.cpp"
     $166 = $159 | 1; //@line 1288 "sndfile.cpp"
     $167 = (($21) + ($166<<1)|0); //@line 1288 "sndfile.cpp"
     HEAP16[$167>>1] = $165; //@line 1288 "sndfile.cpp"
     $168 = ((($psrc25$086)) + 2|0); //@line 1289 "sndfile.cpp"
     if ($154) {
      $sext98 = $157 << 16; //@line 1292 "sndfile.cpp"
      $169 = $sext98 >> 16; //@line 1292 "sndfile.cpp"
      $sext5 = $164 << 16; //@line 1293 "sndfile.cpp"
      $170 = $sext5 >> 16; //@line 1293 "sndfile.cpp"
      $iadd_l23$2 = $169;$iadd_r24$2 = $170;
     } else {
      $iadd_l23$2 = $iadd_l23$184;$iadd_r24$2 = $iadd_r24$185;
     }
     $171 = (($j27$087) + 1)|0; //@line 1285 "sndfile.cpp"
     $exitcond97 = ($171|0)==($24|0); //@line 1285 "sndfile.cpp"
     if ($exitcond97) {
      $len$6 = $151;
      label = 86;
      break;
     } else {
      $iadd_l23$184 = $iadd_l23$2;$iadd_r24$185 = $iadd_r24$2;$j27$087 = $171;$psrc25$086 = $168;
     }
    }
   }
  }
 }
 if ((label|0) == 86) {
  $281 = ($len$6>>>0)>($dwMemLength>>>0); //@line 1534 "sndfile.cpp"
  if ($281) {
   $282 = HEAP32[$22>>2]|0; //@line 1536 "sndfile.cpp"
   $283 = ($282|0)==(0|0); //@line 1536 "sndfile.cpp"
   if ($283) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0); //@line 1546 "sndfile.cpp"
   }
   HEAP32[$pIns>>2] = 0; //@line 1538 "sndfile.cpp"
   $284 = ((($282)) + -16|0); //@line 367 "sndfile.cpp"
   _free($284); //@line 367 "sndfile.cpp"
   HEAP32[$22>>2] = 0; //@line 1540 "sndfile.cpp"
   $$0 = 0;
   STACKTOP = sp;return ($$0|0); //@line 1546 "sndfile.cpp"
  } else {
   $len$615 = $len$6;
  }
 }
 __ZN10CSoundFile16AdjustSampleLoopEP14_MODINSTRUMENT($this,$pIns); //@line 1544 "sndfile.cpp"
 $$0 = $len$615;
 STACKTOP = sp;return ($$0|0); //@line 1546 "sndfile.cpp"
}
function __Z12ITUnpack8BitPajPhjb($pSample,$dwLen,$lpMemFile,$dwMemLength,$b215) {
 $pSample = $pSample|0;
 $dwLen = $dwLen|0;
 $lpMemFile = $lpMemFile|0;
 $dwMemLength = $dwMemLength|0;
 $b215 = $b215|0;
 var $$ = 0, $$0$wCount$1 = 0, $$038 = 0, $$17 = 0, $$bLeft$035 = 0, $$bTemp$0 = 0, $$bTemp2$0 = 0, $$in = 0, $$lcssa92 = 0, $$lcssa94 = 0, $$lpMemFile16 = 0, $$pre = 0, $$pre$phiZ2D = 0, $$sink1 = 0, $$sum = 0, $$wCount$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bLeft$035 = 0, $bLeft$2 = 0, $bLeft$3 = 0, $bLeft$3$lcssa97 = 0, $bTemp$034 = 0, $bTemp$2 = 0, $bTemp$3 = 0, $bTemp$4 = 0, $bTemp$4$lcssa96 = 0, $bTemp2$033 = 0;
 var $bTemp2$2 = 0, $bTemp2$3 = 0, $bTemp2$4 = 0, $bTemp2$4$lcssa95 = 0, $dwPos$0 = 0, $dwPos$1 = 0, $i$0$i = 0, $lpMemFile11 = 0, $lpMemFile12 = 0, $lpMemFile13 = 0, $lpMemFile14 = 0, $lpMemFile1632 = 0, $lpMemFile8 = 0, $lpMemFile8$lcssa90 = 0, $lpMemFile9$1 = 0, $lpMemFile9$2 = 0, $not$ = 0, $pDst$037 = 0, $retval$0$i = 0, $retval$1$i = 0;
 var $sext = 0, $wBits$0 = 0, $wCount$036 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($dwLen|0)==(0); //@line 1222 "load_it.cpp"
 if ($0) {
  return; //@line 1290 "load_it.cpp"
 }
 $$sum = (($dwMemLength) + 1)|0; //@line 1283 "load_it.cpp"
 $1 = (($lpMemFile) + ($$sum)|0); //@line 1283 "load_it.cpp"
 $$038 = $dwLen;$4 = 0;$5 = 0;$bLeft$035 = 0;$bTemp$034 = 0;$bTemp2$033 = 0;$lpMemFile1632 = $lpMemFile;$pDst$037 = $pSample;$wCount$036 = 0;
 L4: while(1) {
  $2 = ($wCount$036|0)==(0); //@line 1224 "load_it.cpp"
  $3 = ((($lpMemFile1632)) + 2|0); //@line 1228 "load_it.cpp"
  $$lpMemFile16 = $2 ? $3 : $lpMemFile1632; //@line 1224 "load_it.cpp"
  $$ = $2 ? 0 : $4; //@line 1224 "load_it.cpp"
  $$17 = $2 ? 0 : $5; //@line 1224 "load_it.cpp"
  $$bTemp2$0 = $2 ? 0 : $bTemp2$033; //@line 1224 "load_it.cpp"
  $$bTemp$0 = $2 ? 0 : $bTemp$034; //@line 1224 "load_it.cpp"
  $$wCount$0 = $2 ? 32768 : $wCount$036; //@line 1224 "load_it.cpp"
  $6 = ($$wCount$0>>>0)>($$038>>>0); //@line 1234 "load_it.cpp"
  $$0$wCount$1 = $6 ? $$038 : $$wCount$0; //@line 1234 "load_it.cpp"
  $$bLeft$035 = $2 ? 9 : $bLeft$035;
  $109 = $$;$110 = $$17;$bLeft$2 = $$bLeft$035;$bTemp$2 = $$bTemp$0;$bTemp2$2 = $$bTemp2$0;$dwPos$0 = 0;$lpMemFile14 = $$lpMemFile16;
  while(1) {
   $7 = $bLeft$2 << 24 >> 24; //@line 1188 "load_it.cpp"
   $8 = ($bLeft$2<<24>>24)>(0); //@line 1190 "load_it.cpp"
   if ($8) {
    $10 = $110;$111 = $109;$i$0$i = $7;$lpMemFile13 = $lpMemFile14;$retval$0$i = 0;
    while(1) {
     $9 = ($10|0)==(0); //@line 1194 "load_it.cpp"
     if ($9) {
      $11 = ((($lpMemFile13)) + 1|0); //@line 1196 "load_it.cpp"
      $12 = HEAP8[$lpMemFile13>>0]|0; //@line 1196 "load_it.cpp"
      $13 = $12&255; //@line 1196 "load_it.cpp"
      $16 = $13;$20 = 8;$lpMemFile12 = $11;
     } else {
      $16 = $111;$20 = $10;$lpMemFile12 = $lpMemFile13;
     }
     $14 = $retval$0$i >>> 1; //@line 1199 "load_it.cpp"
     $15 = $16 << 31; //@line 1200 "load_it.cpp"
     $17 = $15 | $14; //@line 1200 "load_it.cpp"
     $18 = $16 >>> 1; //@line 1201 "load_it.cpp"
     $19 = (($20) + -1)|0; //@line 1202 "load_it.cpp"
     $21 = (($i$0$i) + -1)|0; //@line 1203 "load_it.cpp"
     $22 = ($21|0)==(0); //@line 1204 "load_it.cpp"
     if ($22) {
      $32 = $19;$38 = $18;$lpMemFile11 = $lpMemFile12;$retval$1$i = $17;
      break;
     } else {
      $10 = $19;$111 = $18;$i$0$i = $21;$lpMemFile13 = $lpMemFile12;$retval$0$i = $17;
     }
    }
   } else {
    $32 = $110;$38 = $109;$lpMemFile11 = $lpMemFile14;$retval$1$i = 0;
   }
   $23 = (32 - ($7))|0; //@line 1207 "load_it.cpp"
   $24 = $retval$1$i >>> $23; //@line 1207 "load_it.cpp"
   $25 = $bLeft$2&255; //@line 1240 "load_it.cpp"
   $26 = ($bLeft$2&255)<(7); //@line 1240 "load_it.cpp"
   do {
    if ($26) {
     $27 = (($25) + -1)|0; //@line 1242 "load_it.cpp"
     $28 = 1 << $27; //@line 1242 "load_it.cpp"
     $29 = $24 & 65535; //@line 1243 "load_it.cpp"
     $30 = ($28|0)==($29|0); //@line 1244 "load_it.cpp"
     if ($30) {
      $31 = ($32|0)==(0); //@line 1194 "load_it.cpp"
      if ($31) {
       $33 = ((($lpMemFile11)) + 1|0); //@line 1196 "load_it.cpp"
       $34 = HEAP8[$lpMemFile11>>0]|0; //@line 1196 "load_it.cpp"
       $35 = $34&255; //@line 1196 "load_it.cpp"
       $36 = $35 >>> 1; //@line 1201 "load_it.cpp"
       $85 = $35;$88 = $36;$92 = 7;$lpMemFile9$1 = $33;
      } else {
       $37 = $38 >>> 1; //@line 1201 "load_it.cpp"
       $39 = (($32) + -1)|0; //@line 1202 "load_it.cpp"
       $40 = ($39|0)==(0); //@line 1194 "load_it.cpp"
       if ($40) {
        $81 = ((($lpMemFile11)) + 1|0); //@line 1196 "load_it.cpp"
        $82 = HEAP8[$lpMemFile11>>0]|0; //@line 1196 "load_it.cpp"
        $83 = $82&255; //@line 1196 "load_it.cpp"
        $85 = $38;$88 = $83;$92 = 8;$lpMemFile9$1 = $81;
       } else {
        $85 = $38;$88 = $37;$92 = $39;$lpMemFile9$1 = $lpMemFile11;
       }
      }
      $84 = $85 << 30; //@line 1199 "load_it.cpp"
      $86 = $84 & 1073741824; //@line 1199 "load_it.cpp"
      $87 = $88 << 31; //@line 1200 "load_it.cpp"
      $89 = $87 | $86; //@line 1200 "load_it.cpp"
      $90 = $88 >>> 1; //@line 1201 "load_it.cpp"
      $91 = (($92) + -1)|0; //@line 1202 "load_it.cpp"
      $93 = ($91|0)==(0); //@line 1194 "load_it.cpp"
      if ($93) {
       $94 = ((($lpMemFile9$1)) + 1|0); //@line 1196 "load_it.cpp"
       $95 = HEAP8[$lpMemFile9$1>>0]|0; //@line 1196 "load_it.cpp"
       $96 = $95&255; //@line 1196 "load_it.cpp"
       $103 = 8;$99 = $96;$lpMemFile9$2 = $94;
      } else {
       $103 = $91;$99 = $90;$lpMemFile9$2 = $lpMemFile9$1;
      }
      $97 = $89 >>> 1; //@line 1199 "load_it.cpp"
      $98 = $99 << 31; //@line 1200 "load_it.cpp"
      $100 = $98 | $97; //@line 1200 "load_it.cpp"
      $101 = $99 >>> 1; //@line 1201 "load_it.cpp"
      $102 = (($103) + -1)|0; //@line 1202 "load_it.cpp"
      $104 = $100 >>> 29; //@line 1207 "load_it.cpp"
      $105 = (($104) + 1)|0; //@line 1245 "load_it.cpp"
      $106 = ($105>>>0)<($25>>>0); //@line 1246 "load_it.cpp"
      $107 = (($104) + 2)|0; //@line 1246 "load_it.cpp"
      $$in = $106 ? $105 : $107; //@line 1246 "load_it.cpp"
      $108 = $$in&255; //@line 1246 "load_it.cpp"
      $112 = $101;$113 = $102;$bLeft$3 = $108;$bTemp$4 = $bTemp$2;$bTemp2$4 = $bTemp2$2;$dwPos$1 = $dwPos$0;$lpMemFile8 = $lpMemFile9$2;
     } else {
      $$pre$phiZ2D = $29;
      label = 22;
     }
    } else {
     $41 = ($bLeft$2&255)<(9); //@line 1249 "load_it.cpp"
     if ($41) {
      $42 = (9 - ($25))|0; //@line 1251 "load_it.cpp"
      $43 = 255 >>> $42; //@line 1251 "load_it.cpp"
      $44 = (($43) + 65532)|0; //@line 1252 "load_it.cpp"
      $45 = $24 & 65535; //@line 1253 "load_it.cpp"
      $46 = $44 & 65535; //@line 1253 "load_it.cpp"
      $47 = ($45>>>0)>($46>>>0); //@line 1253 "load_it.cpp"
      if ($47) {
       $48 = (($43) + 4)|0; //@line 1251 "load_it.cpp"
       $49 = $48 & 65535; //@line 1252 "load_it.cpp"
       $50 = ($45>>>0)>($49>>>0); //@line 1253 "load_it.cpp"
       if (!($50)) {
        $51 = (($24) - ($44))|0; //@line 1254 "load_it.cpp"
        $52 = $51 & 255; //@line 1255 "load_it.cpp"
        $not$ = ($52>>>0)>=($25>>>0); //@line 1255 "load_it.cpp"
        $53 = $not$&1; //@line 1255 "load_it.cpp"
        $$sink1 = (($53) + ($51))|0; //@line 1255 "load_it.cpp"
        $54 = $$sink1&255; //@line 1255 "load_it.cpp"
        $112 = $38;$113 = $32;$bLeft$3 = $54;$bTemp$4 = $bTemp$2;$bTemp2$4 = $bTemp2$2;$dwPos$1 = $dwPos$0;$lpMemFile8 = $lpMemFile11;
        break;
       }
      }
     } else {
      $55 = ($bLeft$2&255)>(9); //@line 1258 "load_it.cpp"
      if ($55) {
       $bTemp$3 = $bTemp$2;$bTemp2$3 = $bTemp2$2;
       label = 24;
       break;
      }
      $56 = $24 & 65280; //@line 1259 "load_it.cpp"
      $57 = ($56>>>0)>(255); //@line 1259 "load_it.cpp"
      if ($57) {
       $58 = (($24) + 1)|0; //@line 1261 "load_it.cpp"
       $59 = $58&255; //@line 1261 "load_it.cpp"
       $112 = $38;$113 = $32;$bLeft$3 = $59;$bTemp$4 = $bTemp$2;$bTemp2$4 = $bTemp2$2;$dwPos$1 = $dwPos$0;$lpMemFile8 = $lpMemFile11;
       break;
      }
     }
     $60 = ($bLeft$2&255)<(8); //@line 1265 "load_it.cpp"
     if ($60) {
      $$pre = $24 & 65535; //@line 1268 "load_it.cpp"
      $$pre$phiZ2D = $$pre;
      label = 22;
     } else {
      $wBits$0 = $24;
      label = 23;
     }
    }
   } while(0);
   if ((label|0) == 22) {
    label = 0;
    $61 = (8 - ($25))|0; //@line 1267 "load_it.cpp"
    $62 = $61 & 255; //@line 1268 "load_it.cpp"
    $63 = $$pre$phiZ2D << $62; //@line 1268 "load_it.cpp"
    $sext = $63 << 24; //@line 1269 "load_it.cpp"
    $64 = $sext >> 24; //@line 1269 "load_it.cpp"
    $65 = $64 >> $62; //@line 1269 "load_it.cpp"
    $wBits$0 = $65;
    label = 23;
   }
   if ((label|0) == 23) {
    label = 0;
    $66 = $bTemp$2&255; //@line 1272 "load_it.cpp"
    $67 = (($wBits$0) + ($66))|0; //@line 1272 "load_it.cpp"
    $68 = $67&255; //@line 1273 "load_it.cpp"
    $69 = $bTemp2$2&255; //@line 1274 "load_it.cpp"
    $70 = (($67) + ($69))|0; //@line 1274 "load_it.cpp"
    $71 = $70&255; //@line 1274 "load_it.cpp"
    $72 = $b215 ? $71 : $68; //@line 1276 "load_it.cpp"
    $73 = (($pDst$037) + ($dwPos$0)|0); //@line 1276 "load_it.cpp"
    HEAP8[$73>>0] = $72; //@line 1276 "load_it.cpp"
    $bTemp$3 = $68;$bTemp2$3 = $71;
    label = 24;
   }
   if ((label|0) == 24) {
    label = 0;
    $74 = (($dwPos$0) + 1)|0; //@line 1281 "load_it.cpp"
    $112 = $38;$113 = $32;$bLeft$3 = $bLeft$2;$bTemp$4 = $bTemp$3;$bTemp2$4 = $bTemp2$3;$dwPos$1 = $74;$lpMemFile8 = $lpMemFile11;
   }
   $75 = ($lpMemFile8>>>0)<($1>>>0); //@line 1283 "load_it.cpp"
   if (!($75)) {
    label = 28;
    break L4;
   }
   $76 = ($dwPos$1>>>0)<($$0$wCount$1>>>0); //@line 1284 "load_it.cpp"
   if ($76) {
    $109 = $112;$110 = $113;$bLeft$2 = $bLeft$3;$bTemp$2 = $bTemp$4;$bTemp2$2 = $bTemp2$4;$dwPos$0 = $dwPos$1;$lpMemFile14 = $lpMemFile8;
   } else {
    $$lcssa92 = $112;$$lcssa94 = $113;$bLeft$3$lcssa97 = $bLeft$3;$bTemp$4$lcssa96 = $bTemp$4;$bTemp2$4$lcssa95 = $bTemp2$4;$lpMemFile8$lcssa90 = $lpMemFile8;
    break;
   }
  }
  $77 = (($$wCount$0) - ($$0$wCount$1))|0; //@line 1286 "load_it.cpp"
  $78 = (($$038) - ($$0$wCount$1))|0; //@line 1287 "load_it.cpp"
  $79 = (($pDst$037) + ($$0$wCount$1)|0); //@line 1288 "load_it.cpp"
  $80 = ($$038|0)==($$0$wCount$1|0); //@line 1222 "load_it.cpp"
  if ($80) {
   label = 28;
   break;
  } else {
   $$038 = $78;$4 = $$lcssa92;$5 = $$lcssa94;$bLeft$035 = $bLeft$3$lcssa97;$bTemp$034 = $bTemp$4$lcssa96;$bTemp2$033 = $bTemp2$4$lcssa95;$lpMemFile1632 = $lpMemFile8$lcssa90;$pDst$037 = $79;$wCount$036 = $77;
  }
 }
 if ((label|0) == 28) {
  return; //@line 1290 "load_it.cpp"
 }
}
function __Z13ITUnpack16BitPajPhjb($pSample,$dwLen,$lpMemFile,$dwMemLength,$b215) {
 $pSample = $pSample|0;
 $dwLen = $dwLen|0;
 $lpMemFile = $lpMemFile|0;
 $dwMemLength = $dwMemLength|0;
 $b215 = $b215|0;
 var $$ = 0, $$0 = 0, $$0$wCount$1 = 0, $$20 = 0, $$bLeft$0 = 0, $$lcssa92 = 0, $$lcssa94 = 0, $$lpMemFile19 = 0, $$old3 = 0, $$sink1 = 0, $$sink2 = 0, $$sum = 0, $$wCount$0 = 0, $$wTemp$0 = 0, $$wTemp2$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $bLeft$0 = 0, $bLeft$2 = 0, $bLeft$3 = 0, $bLeft$3$lcssa97 = 0, $dwBits$0 = 0, $dwPos$0 = 0, $dwPos$1 = 0, $i$0$i = 0, $lpMemFile11 = 0, $lpMemFile11$lcssa90 = 0, $lpMemFile12$1 = 0, $lpMemFile12$3 = 0, $lpMemFile14 = 0, $lpMemFile15 = 0, $lpMemFile16 = 0, $lpMemFile17 = 0, $lpMemFile19 = 0, $not$ = 0, $or$cond = 0, $pDst$0 = 0;
 var $retval$0$i = 0, $retval$1$i = 0, $sext4 = 0, $wCount$0 = 0, $wTemp$0 = 0, $wTemp$2 = 0, $wTemp$3 = 0, $wTemp$4 = 0, $wTemp$4$lcssa96 = 0, $wTemp2$0 = 0, $wTemp2$2 = 0, $wTemp2$3 = 0, $wTemp2$4 = 0, $wTemp2$4$lcssa95 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$old3 = ($dwLen|0)==(0); //@line 1305 "load_it.cpp"
 if ($$old3) {
  return; //@line 1374 "load_it.cpp"
 }
 $$sum = (($dwMemLength) + 1)|0; //@line 1366 "load_it.cpp"
 $0 = (($lpMemFile) + ($$sum)|0); //@line 1366 "load_it.cpp"
 $1 = (($lpMemFile) + ($dwMemLength)|0); //@line 1366 "load_it.cpp"
 $$0 = $dwLen;$4 = 0;$5 = 0;$bLeft$0 = 0;$lpMemFile19 = $lpMemFile;$pDst$0 = $pSample;$wCount$0 = 0;$wTemp$0 = 0;$wTemp2$0 = 0;
 L4: while(1) {
  $2 = ($wCount$0|0)==(0); //@line 1307 "load_it.cpp"
  $3 = ((($lpMemFile19)) + 2|0); //@line 1311 "load_it.cpp"
  $$lpMemFile19 = $2 ? $3 : $lpMemFile19; //@line 1307 "load_it.cpp"
  $$ = $2 ? 0 : $4; //@line 1307 "load_it.cpp"
  $$20 = $2 ? 0 : $5; //@line 1307 "load_it.cpp"
  $$wTemp2$0 = $2 ? 0 : $wTemp2$0; //@line 1307 "load_it.cpp"
  $$wTemp$0 = $2 ? 0 : $wTemp$0; //@line 1307 "load_it.cpp"
  $$wCount$0 = $2 ? 16384 : $wCount$0; //@line 1307 "load_it.cpp"
  $6 = ($$wCount$0>>>0)>($$0>>>0); //@line 1317 "load_it.cpp"
  $$0$wCount$1 = $6 ? $$0 : $$wCount$0; //@line 1317 "load_it.cpp"
  $$bLeft$0 = $2 ? 17 : $bLeft$0;
  $120 = $$;$121 = $$20;$bLeft$2 = $$bLeft$0;$dwPos$0 = 0;$lpMemFile17 = $$lpMemFile19;$wTemp$2 = $$wTemp$0;$wTemp2$2 = $$wTemp2$0;
  while(1) {
   $7 = $bLeft$2 << 24 >> 24; //@line 1188 "load_it.cpp"
   $8 = ($bLeft$2<<24>>24)>(0); //@line 1190 "load_it.cpp"
   if ($8) {
    $10 = $121;$122 = $120;$i$0$i = $7;$lpMemFile16 = $lpMemFile17;$retval$0$i = 0;
    while(1) {
     $9 = ($10|0)==(0); //@line 1194 "load_it.cpp"
     if ($9) {
      $11 = ((($lpMemFile16)) + 1|0); //@line 1196 "load_it.cpp"
      $12 = HEAP8[$lpMemFile16>>0]|0; //@line 1196 "load_it.cpp"
      $13 = $12&255; //@line 1196 "load_it.cpp"
      $16 = $13;$20 = 8;$lpMemFile15 = $11;
     } else {
      $16 = $122;$20 = $10;$lpMemFile15 = $lpMemFile16;
     }
     $14 = $retval$0$i >>> 1; //@line 1199 "load_it.cpp"
     $15 = $16 << 31; //@line 1200 "load_it.cpp"
     $17 = $15 | $14; //@line 1200 "load_it.cpp"
     $18 = $16 >>> 1; //@line 1201 "load_it.cpp"
     $19 = (($20) + -1)|0; //@line 1202 "load_it.cpp"
     $21 = (($i$0$i) + -1)|0; //@line 1203 "load_it.cpp"
     $22 = ($21|0)==(0); //@line 1204 "load_it.cpp"
     if ($22) {
      $31 = $19;$37 = $18;$lpMemFile14 = $lpMemFile15;$retval$1$i = $17;
      break;
     } else {
      $10 = $19;$122 = $18;$i$0$i = $21;$lpMemFile16 = $lpMemFile15;$retval$0$i = $17;
     }
    }
   } else {
    $31 = $121;$37 = $120;$lpMemFile14 = $lpMemFile17;$retval$1$i = 0;
   }
   $23 = (32 - ($7))|0; //@line 1207 "load_it.cpp"
   $24 = $retval$1$i >>> $23; //@line 1207 "load_it.cpp"
   $25 = $bLeft$2&255; //@line 1323 "load_it.cpp"
   $26 = ($bLeft$2&255)<(7); //@line 1323 "load_it.cpp"
   do {
    if ($26) {
     $27 = (($25) + -1)|0; //@line 1325 "load_it.cpp"
     $28 = 1 << $27; //@line 1325 "load_it.cpp"
     $29 = ($28|0)==($24|0); //@line 1327 "load_it.cpp"
     if ($29) {
      $30 = ($31|0)==(0); //@line 1194 "load_it.cpp"
      if ($30) {
       $32 = ((($lpMemFile14)) + 1|0); //@line 1196 "load_it.cpp"
       $33 = HEAP8[$lpMemFile14>>0]|0; //@line 1196 "load_it.cpp"
       $34 = $33&255; //@line 1196 "load_it.cpp"
       $35 = $34 >>> 1; //@line 1201 "load_it.cpp"
       $83 = $34;$86 = $35;$89 = 7;$lpMemFile12$1 = $32;
      } else {
       $36 = $37 >>> 1; //@line 1201 "load_it.cpp"
       $38 = (($31) + -1)|0; //@line 1202 "load_it.cpp"
       $39 = ($38|0)==(0); //@line 1194 "load_it.cpp"
       if ($39) {
        $79 = ((($lpMemFile14)) + 1|0); //@line 1196 "load_it.cpp"
        $80 = HEAP8[$lpMemFile14>>0]|0; //@line 1196 "load_it.cpp"
        $81 = $80&255; //@line 1196 "load_it.cpp"
        $83 = $37;$86 = $81;$89 = 8;$lpMemFile12$1 = $79;
       } else {
        $83 = $37;$86 = $36;$89 = $38;$lpMemFile12$1 = $lpMemFile14;
       }
      }
      $82 = $83 << 30; //@line 1199 "load_it.cpp"
      $84 = $82 & 1073741824; //@line 1199 "load_it.cpp"
      $85 = $86 << 31; //@line 1200 "load_it.cpp"
      $87 = $85 | $84; //@line 1200 "load_it.cpp"
      $88 = ($89|0)==(1); //@line 1194 "load_it.cpp"
      if ($88) {
       $90 = ((($lpMemFile12$1)) + 1|0); //@line 1196 "load_it.cpp"
       $91 = HEAP8[$lpMemFile12$1>>0]|0; //@line 1196 "load_it.cpp"
       $92 = $91&255; //@line 1196 "load_it.cpp"
       $93 = $87 >>> 1; //@line 1199 "load_it.cpp"
       $94 = $92 << 31; //@line 1200 "load_it.cpp"
       $95 = $94 | $93; //@line 1200 "load_it.cpp"
       $96 = $92 >>> 1; //@line 1201 "load_it.cpp"
       $108 = $95;$110 = $96;$114 = 7;$lpMemFile12$3 = $90;
      } else {
       $97 = $86 >>> 1; //@line 1201 "load_it.cpp"
       $98 = $87 >>> 1; //@line 1199 "load_it.cpp"
       $99 = $97 << 31; //@line 1200 "load_it.cpp"
       $100 = $99 | $98; //@line 1200 "load_it.cpp"
       $101 = $86 >>> 2; //@line 1201 "load_it.cpp"
       $102 = (($89) + -2)|0; //@line 1202 "load_it.cpp"
       $103 = ($102|0)==(0); //@line 1194 "load_it.cpp"
       if ($103) {
        $104 = ((($lpMemFile12$1)) + 1|0); //@line 1196 "load_it.cpp"
        $105 = HEAP8[$lpMemFile12$1>>0]|0; //@line 1196 "load_it.cpp"
        $106 = $105&255; //@line 1196 "load_it.cpp"
        $108 = $100;$110 = $106;$114 = 8;$lpMemFile12$3 = $104;
       } else {
        $108 = $100;$110 = $101;$114 = $102;$lpMemFile12$3 = $lpMemFile12$1;
       }
      }
      $107 = $108 >>> 1; //@line 1199 "load_it.cpp"
      $109 = $110 << 31; //@line 1200 "load_it.cpp"
      $111 = $109 | $107; //@line 1200 "load_it.cpp"
      $112 = $110 >>> 1; //@line 1201 "load_it.cpp"
      $113 = (($114) + -1)|0; //@line 1202 "load_it.cpp"
      $115 = $111 >>> 28; //@line 1207 "load_it.cpp"
      $116 = (($115) + 1)|0; //@line 1328 "load_it.cpp"
      $117 = ($116>>>0)<($25>>>0); //@line 1329 "load_it.cpp"
      $118 = (($115) + 2)|0; //@line 1329 "load_it.cpp"
      $$sink1 = $117 ? $116 : $118; //@line 1329 "load_it.cpp"
      $119 = $$sink1&255; //@line 1329 "load_it.cpp"
      $123 = $112;$124 = $113;$bLeft$3 = $119;$dwPos$1 = $dwPos$0;$lpMemFile11 = $lpMemFile12$3;$wTemp$4 = $wTemp$2;$wTemp2$4 = $wTemp2$2;
     } else {
      label = 21;
     }
    } else {
     $40 = ($bLeft$2&255)<(17); //@line 1332 "load_it.cpp"
     if ($40) {
      $41 = (17 - ($25))|0; //@line 1334 "load_it.cpp"
      $42 = 65535 >>> $41; //@line 1334 "load_it.cpp"
      $43 = (($42) + 65528)|0; //@line 1335 "load_it.cpp"
      $44 = $43 & 65535; //@line 1335 "load_it.cpp"
      $45 = ($24>>>0)>($44>>>0); //@line 1336 "load_it.cpp"
      if ($45) {
       $46 = (($42) + 8)|0; //@line 1334 "load_it.cpp"
       $47 = $46 & 65535; //@line 1336 "load_it.cpp"
       $48 = ($24>>>0)>($47>>>0); //@line 1336 "load_it.cpp"
       if (!($48)) {
        $49 = (($24) - ($44))|0; //@line 1337 "load_it.cpp"
        $50 = $49 & 255; //@line 1338 "load_it.cpp"
        $not$ = ($50>>>0)>=($25>>>0); //@line 1338 "load_it.cpp"
        $51 = $not$&1; //@line 1338 "load_it.cpp"
        $$sink2 = (($51) + ($49))|0; //@line 1338 "load_it.cpp"
        $52 = $$sink2&255; //@line 1338 "load_it.cpp"
        $123 = $37;$124 = $31;$bLeft$3 = $52;$dwPos$1 = $dwPos$0;$lpMemFile11 = $lpMemFile14;$wTemp$4 = $wTemp$2;$wTemp2$4 = $wTemp2$2;
        break;
       }
      }
     } else {
      $53 = ($bLeft$2&255)>(17); //@line 1341 "load_it.cpp"
      if ($53) {
       $wTemp$3 = $wTemp$2;$wTemp2$3 = $wTemp2$2;
       label = 23;
       break;
      }
      $54 = ($24>>>0)>(65535); //@line 1342 "load_it.cpp"
      if ($54) {
       $55 = (($24) + 1)|0; //@line 1344 "load_it.cpp"
       $56 = $55&255; //@line 1344 "load_it.cpp"
       $123 = $37;$124 = $31;$bLeft$3 = $56;$dwPos$1 = $dwPos$0;$lpMemFile11 = $lpMemFile14;$wTemp$4 = $wTemp$2;$wTemp2$4 = $wTemp2$2;
       break;
      }
     }
     $57 = ($bLeft$2&255)<(16); //@line 1348 "load_it.cpp"
     if ($57) {
      label = 21;
     } else {
      $dwBits$0 = $24;
      label = 22;
     }
    }
   } while(0);
   if ((label|0) == 21) {
    label = 0;
    $58 = (16 - ($25))|0; //@line 1350 "load_it.cpp"
    $59 = $58 & 255; //@line 1351 "load_it.cpp"
    $60 = $24 << $59; //@line 1351 "load_it.cpp"
    $sext4 = $60 << 16; //@line 1352 "load_it.cpp"
    $61 = $sext4 >> 16; //@line 1352 "load_it.cpp"
    $62 = $61 >> $59; //@line 1352 "load_it.cpp"
    $dwBits$0 = $62;
    label = 22;
   }
   if ((label|0) == 22) {
    label = 0;
    $63 = $wTemp$2&65535; //@line 1355 "load_it.cpp"
    $64 = (($dwBits$0) + ($63))|0; //@line 1355 "load_it.cpp"
    $65 = $64&65535; //@line 1356 "load_it.cpp"
    $66 = $wTemp2$2&65535; //@line 1357 "load_it.cpp"
    $67 = (($64) + ($66))|0; //@line 1357 "load_it.cpp"
    $68 = $67&65535; //@line 1357 "load_it.cpp"
    $69 = $b215 ? $68 : $65; //@line 1359 "load_it.cpp"
    $70 = (($pDst$0) + ($dwPos$0<<1)|0); //@line 1359 "load_it.cpp"
    HEAP16[$70>>1] = $69; //@line 1359 "load_it.cpp"
    $wTemp$3 = $65;$wTemp2$3 = $68;
    label = 23;
   }
   if ((label|0) == 23) {
    label = 0;
    $71 = (($dwPos$0) + 1)|0; //@line 1364 "load_it.cpp"
    $123 = $37;$124 = $31;$bLeft$3 = $bLeft$2;$dwPos$1 = $71;$lpMemFile11 = $lpMemFile14;$wTemp$4 = $wTemp$3;$wTemp2$4 = $wTemp2$3;
   }
   $72 = ($lpMemFile11>>>0)<($0>>>0); //@line 1366 "load_it.cpp"
   if (!($72)) {
    label = 27;
    break L4;
   }
   $73 = ($dwPos$1>>>0)<($$0$wCount$1>>>0); //@line 1367 "load_it.cpp"
   if ($73) {
    $120 = $123;$121 = $124;$bLeft$2 = $bLeft$3;$dwPos$0 = $dwPos$1;$lpMemFile17 = $lpMemFile11;$wTemp$2 = $wTemp$4;$wTemp2$2 = $wTemp2$4;
   } else {
    $$lcssa92 = $123;$$lcssa94 = $124;$bLeft$3$lcssa97 = $bLeft$3;$lpMemFile11$lcssa90 = $lpMemFile11;$wTemp$4$lcssa96 = $wTemp$4;$wTemp2$4$lcssa95 = $wTemp2$4;
    break;
   }
  }
  $74 = (($$wCount$0) - ($$0$wCount$1))|0; //@line 1369 "load_it.cpp"
  $75 = (($$0) - ($$0$wCount$1))|0; //@line 1370 "load_it.cpp"
  $76 = (($pDst$0) + ($$0$wCount$1<<1)|0); //@line 1371 "load_it.cpp"
  $77 = ($lpMemFile11$lcssa90>>>0)<($1>>>0); //@line 1372 "load_it.cpp"
  $78 = ($$0|0)!=($$0$wCount$1|0); //@line 1305 "load_it.cpp"
  $or$cond = $78 & $77; //@line 1372 "load_it.cpp"
  if ($or$cond) {
   $$0 = $75;$4 = $$lcssa92;$5 = $$lcssa94;$bLeft$0 = $bLeft$3$lcssa97;$lpMemFile19 = $lpMemFile11$lcssa90;$pDst$0 = $76;$wCount$0 = $74;$wTemp$0 = $wTemp$4$lcssa96;$wTemp2$0 = $wTemp2$4$lcssa95;
  } else {
   label = 27;
   break;
  }
 }
 if ((label|0) == 27) {
  return; //@line 1374 "load_it.cpp"
 }
}
function __Z9AMSUnpackPKcjPcjc($psrc,$inputlen,$pdest,$dmax,$packcharacter) {
 $psrc = $psrc|0;
 $inputlen = $inputlen|0;
 $pdest = $pdest|0;
 $dmax = $dmax|0;
 $packcharacter = $packcharacter|0;
 var $$ = 0, $$$lcssa = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, $bitcount$011 = 0, $bitcount$16 = 0, $ch2$0 = 0, $count$09 = 0, $dh$0$ = 0, $dh$0$$lcssa = 0, $dh$07 = 0, $exitcond = 0, $exitcond20 = 0, $exitcond21 = 0;
 var $i$0$be = 0, $i$017 = 0, $i2$013 = 0, $i3$05 = 0, $j$0$be = 0, $j$016 = 0, $j$1 = 0, $k$012 = 0, $k$18 = 0, $old$04 = 0, $or$cond = 0, $or$cond15 = 0, $or$cond3 = 0, $p1$010 = 0, $pos$0 = 0, $sext = 0, $sext1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znaj($dmax)|0); //@line 566 "load_ams.cpp"
 $1 = ($inputlen|0)!=(0); //@line 573 "load_ams.cpp"
 $2 = ($dmax|0)!=(0); //@line 573 "load_ams.cpp"
 $or$cond15 = $2 & $1; //@line 573 "load_ams.cpp"
 if ($or$cond15) {
  $i$017 = 0;$j$016 = 0;
  while(1) {
   $4 = (($i$017) + 1)|0; //@line 575 "load_ams.cpp"
   $5 = (($psrc) + ($i$017)|0); //@line 575 "load_ams.cpp"
   $6 = HEAP8[$5>>0]|0; //@line 575 "load_ams.cpp"
   $7 = ($6<<24>>24)==($packcharacter<<24>>24); //@line 576 "load_ams.cpp"
   L4: do {
    if ($7) {
     $8 = (($i$017) + 2)|0; //@line 578 "load_ams.cpp"
     $9 = (($psrc) + ($4)|0); //@line 578 "load_ams.cpp"
     $10 = HEAP8[$9>>0]|0; //@line 578 "load_ams.cpp"
     $11 = ($10<<24>>24)==(0); //@line 579 "load_ams.cpp"
     if ($11) {
      $20 = (($j$016) + 1)|0; //@line 587 "load_ams.cpp"
      $21 = (($0) + ($j$016)|0); //@line 587 "load_ams.cpp"
      HEAP8[$21>>0] = $packcharacter; //@line 587 "load_ams.cpp"
      $i$0$be = $8;$j$0$be = $20;
      break;
     }
     $12 = (($i$017) + 3)|0; //@line 581 "load_ams.cpp"
     $13 = (($psrc) + ($8)|0); //@line 581 "load_ams.cpp"
     $14 = HEAP8[$13>>0]|0; //@line 581 "load_ams.cpp"
     $ch2$0 = $10;$j$1 = $j$016;
     while(1) {
      $15 = ($ch2$0<<24>>24)==(0); //@line 582 "load_ams.cpp"
      if ($15) {
       $i$0$be = $12;$j$0$be = $j$1;
       break L4;
      }
      $16 = (($ch2$0) + -1)<<24>>24; //@line 582 "load_ams.cpp"
      $17 = (($j$1) + 1)|0; //@line 584 "load_ams.cpp"
      $18 = (($0) + ($j$1)|0); //@line 584 "load_ams.cpp"
      HEAP8[$18>>0] = $14; //@line 584 "load_ams.cpp"
      $19 = ($17>>>0)<($dmax>>>0); //@line 585 "load_ams.cpp"
      if ($19) {
       $ch2$0 = $16;$j$1 = $17;
      } else {
       $i$0$be = $12;$j$0$be = $17;
       break;
      }
     }
    } else {
     $24 = (($j$016) + 1)|0; //@line 588 "load_ams.cpp"
     $25 = (($0) + ($j$016)|0); //@line 588 "load_ams.cpp"
     HEAP8[$25>>0] = $6; //@line 588 "load_ams.cpp"
     $i$0$be = $4;$j$0$be = $24;
    }
   } while(0);
   $22 = ($i$0$be>>>0)<($inputlen>>>0); //@line 573 "load_ams.cpp"
   $23 = ($j$0$be>>>0)<($dmax>>>0); //@line 573 "load_ams.cpp"
   $or$cond = $23 & $22; //@line 573 "load_ams.cpp"
   if ($or$cond) {
    $i$017 = $i$0$be;$j$016 = $j$0$be;
   } else {
    break;
   }
  }
 }
 $3 = ($dmax|0)==(0); //@line 596 "load_ams.cpp"
 if ($3) {
  __ZdaPv($0); //@line 626 "load_ams.cpp"
  return; //@line 627 "load_ams.cpp"
 } else {
  $bitcount$011 = 128;$i2$013 = 0;$k$012 = 0;$p1$010 = $0;
 }
 while(1) {
  $26 = HEAP8[$p1$010>>0]|0; //@line 598 "load_ams.cpp"
  $27 = $26&255; //@line 602 "load_ams.cpp"
  $bitcount$16 = $bitcount$011;$count$09 = 0;$dh$07 = 0;$k$18 = $k$012;
  while(1) {
   $28 = $bitcount$16 & $27; //@line 602 "load_ams.cpp"
   $29 = $28 << 8; //@line 603 "load_ams.cpp"
   $30 = $29 | $28; //@line 603 "load_ams.cpp"
   $31 = (($dh$07) - ($count$09))|0; //@line 603 "load_ams.cpp"
   $32 = $31 & 7; //@line 603 "load_ams.cpp"
   $33 = $30 >>> $32; //@line 603 "load_ams.cpp"
   $34 = $bitcount$16 << 8; //@line 604 "load_ams.cpp"
   $35 = $34 | $bitcount$16; //@line 604 "load_ams.cpp"
   $36 = $35 >>> 1; //@line 604 "load_ams.cpp"
   $37 = $36 & 255; //@line 604 "load_ams.cpp"
   $38 = (($k$18) + 1)|0; //@line 605 "load_ams.cpp"
   $39 = (($pdest) + ($k$18)|0); //@line 605 "load_ams.cpp"
   $40 = HEAP8[$39>>0]|0; //@line 605 "load_ams.cpp"
   $41 = $40&255; //@line 605 "load_ams.cpp"
   $42 = $41 | $33; //@line 605 "load_ams.cpp"
   $43 = $42&255; //@line 605 "load_ams.cpp"
   HEAP8[$39>>0] = $43; //@line 605 "load_ams.cpp"
   $44 = ($38>>>0)<($dmax>>>0); //@line 606 "load_ams.cpp"
   $45 = $44&1; //@line 606 "load_ams.cpp"
   $46 = $45 ^ 1; //@line 606 "load_ams.cpp"
   $dh$0$ = (($46) + ($dh$07))|0; //@line 606 "load_ams.cpp"
   $$ = $44 ? $38 : 0; //@line 606 "load_ams.cpp"
   $47 = (($count$09) + 1)|0; //@line 600 "load_ams.cpp"
   $exitcond20 = ($47|0)==(8); //@line 600 "load_ams.cpp"
   if ($exitcond20) {
    $$$lcssa = $$;$$lcssa = $37;$dh$0$$lcssa = $dh$0$;
    break;
   } else {
    $bitcount$16 = $37;$count$09 = $47;$dh$07 = $dh$0$;$k$18 = $$;
   }
  }
  $48 = ((($p1$010)) + 1|0); //@line 598 "load_ams.cpp"
  $49 = $$lcssa << 8; //@line 612 "load_ams.cpp"
  $50 = $49 | $$lcssa; //@line 612 "load_ams.cpp"
  $51 = $50 >>> $dh$0$$lcssa; //@line 612 "load_ams.cpp"
  $52 = $51 & 255; //@line 612 "load_ams.cpp"
  $53 = (($i2$013) + 1)|0; //@line 596 "load_ams.cpp"
  $exitcond21 = ($53|0)==($dmax|0); //@line 596 "load_ams.cpp"
  if ($exitcond21) {
   break;
  } else {
   $bitcount$011 = $52;$i2$013 = $53;$k$012 = $$$lcssa;$p1$010 = $48;
  }
 }
 if ($3) {
  __ZdaPv($0); //@line 626 "load_ams.cpp"
  return; //@line 627 "load_ams.cpp"
 } else {
  $i3$05 = 0;$old$04 = 0;
 }
 while(1) {
  $54 = (($pdest) + ($i3$05)|0); //@line 620 "load_ams.cpp"
  $55 = HEAP8[$54>>0]|0; //@line 620 "load_ams.cpp"
  $56 = $55&255; //@line 620 "load_ams.cpp"
  $57 = ($55<<24>>24)==(-128); //@line 621 "load_ams.cpp"
  $58 = $56 & 128;
  $59 = ($58|0)==(0); //@line 621 "load_ams.cpp"
  $or$cond3 = $57 | $59; //@line 621 "load_ams.cpp"
  $60 = $56 & 127; //@line 621 "load_ams.cpp"
  $61 = (0 - ($60))|0; //@line 621 "load_ams.cpp"
  $pos$0 = $or$cond3 ? $56 : $61; //@line 621 "load_ams.cpp"
  $sext = $pos$0 << 24; //@line 622 "load_ams.cpp"
  $62 = $sext >> 24; //@line 622 "load_ams.cpp"
  $sext1 = $old$04 << 24; //@line 622 "load_ams.cpp"
  $63 = $sext1 >> 24; //@line 622 "load_ams.cpp"
  $64 = (($63) - ($62))|0; //@line 622 "load_ams.cpp"
  $65 = $64&255; //@line 622 "load_ams.cpp"
  HEAP8[$54>>0] = $65; //@line 623 "load_ams.cpp"
  $66 = (($i3$05) + 1)|0; //@line 618 "load_ams.cpp"
  $exitcond = ($66|0)==($dmax|0); //@line 618 "load_ams.cpp"
  if ($exitcond) {
   break;
  } else {
   $i3$05 = $66;$old$04 = $64;
  }
 }
 __ZdaPv($0); //@line 626 "load_ams.cpp"
 return; //@line 627 "load_ams.cpp"
}
function __Z11MDLReadBitsRjS_RPha($bitbuf,$bitnum,$ibuf,$n) {
 $bitbuf = $bitbuf|0;
 $bitnum = $bitnum|0;
 $ibuf = $ibuf|0;
 $n = $n|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$bitbuf>>2]|0; //@line 492 "load_mdl.cpp"
 $1 = $n << 24 >> 24; //@line 492 "load_mdl.cpp"
 $2 = 1 << $1; //@line 492 "load_mdl.cpp"
 $3 = (($2) + 65535)|0; //@line 492 "load_mdl.cpp"
 $4 = $0 & $3; //@line 492 "load_mdl.cpp"
 $5 = $4&65535; //@line 492 "load_mdl.cpp"
 $6 = $0 >>> $1; //@line 493 "load_mdl.cpp"
 HEAP32[$bitbuf>>2] = $6; //@line 493 "load_mdl.cpp"
 $7 = HEAP32[$bitnum>>2]|0; //@line 494 "load_mdl.cpp"
 $8 = (($7) - ($1))|0; //@line 494 "load_mdl.cpp"
 HEAP32[$bitnum>>2] = $8; //@line 494 "load_mdl.cpp"
 $9 = ($8>>>0)<(25); //@line 495 "load_mdl.cpp"
 if (!($9)) {
  return ($5|0); //@line 500 "load_mdl.cpp"
 }
 $10 = HEAP32[$ibuf>>2]|0; //@line 497 "load_mdl.cpp"
 $11 = ((($10)) + 1|0); //@line 497 "load_mdl.cpp"
 HEAP32[$ibuf>>2] = $11; //@line 497 "load_mdl.cpp"
 $12 = HEAP8[$10>>0]|0; //@line 497 "load_mdl.cpp"
 $13 = $12&255; //@line 497 "load_mdl.cpp"
 $14 = $13 << $8; //@line 497 "load_mdl.cpp"
 $15 = HEAP32[$bitbuf>>2]|0; //@line 497 "load_mdl.cpp"
 $16 = $15 | $14; //@line 497 "load_mdl.cpp"
 HEAP32[$bitbuf>>2] = $16; //@line 497 "load_mdl.cpp"
 $17 = HEAP32[$bitnum>>2]|0; //@line 498 "load_mdl.cpp"
 $18 = (($17) + 8)|0; //@line 498 "load_mdl.cpp"
 HEAP32[$bitnum>>2] = $18; //@line 498 "load_mdl.cpp"
 return ($5|0); //@line 500 "load_mdl.cpp"
}
function __Z9DMFUnpackPhS_S_j($psample,$ibuf,$ibufmax,$maxlen) {
 $psample = $psample|0;
 $ibuf = $ibuf|0;
 $ibufmax = $ibufmax|0;
 $maxlen = $maxlen|0;
 var $$pre$i = 0, $$pre$i$pre = 0, $$pre$i1 = 0, $$pre$i16 = 0, $$pre3$i = 0, $$pre3$i5 = 0, $$sink = 0, $$sink$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $7 = 0, $8 = 0, $9 = 0, $actnode$0 = 0, $delta$011 = 0, $delta$1 = 0, $delta$2 = 0, $delta$3 = 0, $exitcond = 0, $fold = 0, $i$012 = 0, $or$cond = 0, $storemerge$i = 0;
 var $tree = 0, $value$010 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1312|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tree = sp;
 _memset(($tree|0),0,1304)|0; //@line 577 "load_dmf.cpp"
 HEAP32[$tree>>2] = $ibuf; //@line 578 "load_dmf.cpp"
 $0 = ((($tree)) + 4|0); //@line 579 "load_dmf.cpp"
 HEAP32[$0>>2] = $ibufmax; //@line 579 "load_dmf.cpp"
 __Z10DMFNewNodeP9DMF_HTREE($tree); //@line 580 "load_dmf.cpp"
 $1 = ($maxlen|0)==(0); //@line 582 "load_dmf.cpp"
 if ($1) {
  $58 = HEAP32[$tree>>2]|0; //@line 603 "load_dmf.cpp"
  $59 = $58; //@line 603 "load_dmf.cpp"
  $60 = $ibuf; //@line 603 "load_dmf.cpp"
  $61 = (($59) - ($60))|0; //@line 603 "load_dmf.cpp"
  STACKTOP = sp;return ($61|0); //@line 604 "load_dmf.cpp"
 }
 $2 = ((($tree)) + 12|0); //@line 516 "load_dmf.cpp"
 $3 = ((($tree)) + 8|0); //@line 521 "load_dmf.cpp"
 $$pre$i$pre = HEAP32[$2>>2]|0; //@line 516 "load_dmf.cpp"
 $$pre$i = $$pre$i$pre;$delta$011 = 0;$i$012 = 0;$value$010 = 0;
 while(1) {
  $4 = ($$pre$i|0)==(0); //@line 516 "load_dmf.cpp"
  if ($4) {
   $6 = HEAP32[$tree>>2]|0; //@line 521 "load_dmf.cpp"
   $7 = HEAP32[$0>>2]|0; //@line 521 "load_dmf.cpp"
   $8 = ($6>>>0)<($7>>>0); //@line 521 "load_dmf.cpp"
   if ($8) {
    $9 = ((($6)) + 1|0); //@line 521 "load_dmf.cpp"
    HEAP32[$tree>>2] = $9; //@line 521 "load_dmf.cpp"
    $10 = HEAP8[$6>>0]|0; //@line 521 "load_dmf.cpp"
    $11 = $10&255; //@line 521 "load_dmf.cpp"
    $12 = $11;
   } else {
    $12 = 0;
   }
   HEAP32[$3>>2] = $12; //@line 521 "load_dmf.cpp"
   $14 = $12;$storemerge$i = 7;
  } else {
   $5 = (($$pre$i) + -1)|0; //@line 518 "load_dmf.cpp"
   $$pre3$i = HEAP32[$3>>2]|0; //@line 524 "load_dmf.cpp"
   $14 = $$pre3$i;$storemerge$i = $5;
  }
  HEAP32[$2>>2] = $storemerge$i; //@line 518 "load_dmf.cpp"
  $13 = $14 & 1; //@line 524 "load_dmf.cpp"
  $15 = ($13|0)==(0); //@line 524 "load_dmf.cpp"
  $16 = $14 >>> 1; //@line 526 "load_dmf.cpp"
  HEAP32[$3>>2] = $16; //@line 526 "load_dmf.cpp"
  $$pre$i1 = $storemerge$i;$$pre3$i5 = $16;$actnode$0 = 0;$delta$1 = $delta$011;
  while(1) {
   $17 = ($$pre$i1|0)==(0); //@line 516 "load_dmf.cpp"
   if ($17) {
    $19 = HEAP32[$tree>>2]|0; //@line 521 "load_dmf.cpp"
    $20 = HEAP32[$0>>2]|0; //@line 521 "load_dmf.cpp"
    $21 = ($19>>>0)<($20>>>0); //@line 521 "load_dmf.cpp"
    if ($21) {
     $22 = ((($19)) + 1|0); //@line 521 "load_dmf.cpp"
     HEAP32[$tree>>2] = $22; //@line 521 "load_dmf.cpp"
     $23 = HEAP8[$19>>0]|0; //@line 521 "load_dmf.cpp"
     $24 = $23&255; //@line 521 "load_dmf.cpp"
     $25 = $24;
    } else {
     $25 = 0;
    }
    HEAP32[$3>>2] = $25; //@line 521 "load_dmf.cpp"
    $26 = 7;$28 = $25;
   } else {
    $18 = (($$pre$i1) + -1)|0; //@line 518 "load_dmf.cpp"
    $26 = $18;$28 = $$pre3$i5;
   }
   HEAP32[$2>>2] = $26; //@line 518 "load_dmf.cpp"
   $27 = $28 & 1; //@line 524 "load_dmf.cpp"
   $29 = ($27|0)==(0); //@line 524 "load_dmf.cpp"
   $30 = $28 >>> 1; //@line 526 "load_dmf.cpp"
   HEAP32[$3>>2] = $30; //@line 526 "load_dmf.cpp"
   $31 = (((((($tree)) + 24|0) + (($actnode$0*5)|0)|0)) + 2|0); //@line 589 "load_dmf.cpp"
   $32 = (((($tree)) + 24|0) + (($actnode$0*5)|0)|0); //@line 591 "load_dmf.cpp"
   $$sink$in = $29 ? $32 : $31; //@line 588 "load_dmf.cpp"
   $$sink = HEAPU8[$$sink$in>>0]|(HEAPU8[$$sink$in+1>>0]<<8); //@line 589 "load_dmf.cpp"
   $33 = $$sink << 16 >> 16; //@line 589 "load_dmf.cpp"
   $34 = ($$sink&65535)>(255); //@line 592 "load_dmf.cpp"
   if ($34) {
    $$pre$i16 = $26;$delta$2 = $delta$1;
    break;
   }
   $35 = (((((($tree)) + 24|0) + (($33*5)|0)|0)) + 4|0); //@line 593 "load_dmf.cpp"
   $36 = HEAP8[$35>>0]|0; //@line 593 "load_dmf.cpp"
   $37 = HEAP32[$tree>>2]|0; //@line 594 "load_dmf.cpp"
   $38 = HEAP32[$0>>2]|0; //@line 594 "load_dmf.cpp"
   $39 = ($37>>>0)>=($38>>>0); //@line 594 "load_dmf.cpp"
   $40 = ($26|0)==(0); //@line 594 "load_dmf.cpp"
   $or$cond = $39 & $40; //@line 594 "load_dmf.cpp"
   if ($or$cond) {
    $$pre$i16 = 0;$delta$2 = $36;
    break;
   }
   $41 = (((($tree)) + 24|0) + (($33*5)|0)|0); //@line 595 "load_dmf.cpp"
   $42 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8); //@line 595 "load_dmf.cpp"
   $43 = ($42<<16>>16)>(-1); //@line 595 "load_dmf.cpp"
   if (!($43)) {
    $$pre$i16 = $26;$delta$2 = $36;
    break;
   }
   $44 = (((((($tree)) + 24|0) + (($33*5)|0)|0)) + 2|0); //@line 595 "load_dmf.cpp"
   $45 = HEAPU8[$44>>0]|(HEAPU8[$44+1>>0]<<8); //@line 595 "load_dmf.cpp"
   $46 = ($45<<16>>16)>(-1); //@line 595 "load_dmf.cpp"
   if ($46) {
    $$pre$i1 = $26;$$pre3$i5 = $30;$actnode$0 = $33;$delta$1 = $36;
   } else {
    $$pre$i16 = $26;$delta$2 = $36;
    break;
   }
  }
  if ($15) {
   $delta$3 = $delta$2;
  } else {
   $47 = $delta$2&255; //@line 596 "load_dmf.cpp"
   $48 = $47 ^ 255; //@line 596 "load_dmf.cpp"
   $49 = $48&255; //@line 596 "load_dmf.cpp"
   $delta$3 = $49;
  }
  $50 = $delta$3&255; //@line 597 "load_dmf.cpp"
  $51 = $value$010 & 255; //@line 597 "load_dmf.cpp"
  $52 = (($50) + ($51))|0; //@line 597 "load_dmf.cpp"
  $53 = ($i$012|0)!=(0); //@line 598 "load_dmf.cpp"
  $fold = (($50) + ($value$010))|0; //@line 598 "load_dmf.cpp"
  $54 = $fold&255; //@line 598 "load_dmf.cpp"
  $55 = $53 ? $54 : 0; //@line 598 "load_dmf.cpp"
  $56 = (($psample) + ($i$012)|0); //@line 598 "load_dmf.cpp"
  HEAP8[$56>>0] = $55; //@line 598 "load_dmf.cpp"
  $57 = (($i$012) + 1)|0; //@line 582 "load_dmf.cpp"
  $exitcond = ($57|0)==($maxlen|0); //@line 582 "load_dmf.cpp"
  if ($exitcond) {
   break;
  } else {
   $$pre$i = $$pre$i16;$delta$011 = $delta$3;$i$012 = $57;$value$010 = $52;
  }
 }
 $58 = HEAP32[$tree>>2]|0; //@line 603 "load_dmf.cpp"
 $59 = $58; //@line 603 "load_dmf.cpp"
 $60 = $ibuf; //@line 603 "load_dmf.cpp"
 $61 = (($59) - ($60))|0; //@line 603 "load_dmf.cpp"
 STACKTOP = sp;return ($61|0); //@line 604 "load_dmf.cpp"
}
function __Z10DMFNewNodeP9DMF_HTREE($tree) {
 $tree = $tree|0;
 var $$in$i = 0, $$lcssa25 = 0, $$pre$i = 0, $$pre$i1 = 0, $$pre3$i = 0, $$pre3$i5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0, $bitv$02$i = 0;
 var $storemerge$i = 0, $storemerge$i14 = 0, $storemerge$i621 = 0, $x$0$$i = 0, $x$0$$i$lcssa = 0, $x$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($tree)) + 20|0); //@line 541 "load_dmf.cpp"
 $1 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 541 "load_dmf.cpp"
 $2 = ($1>>>0)>(255); //@line 542 "load_dmf.cpp"
 if ($2) {
  return; //@line 567 "load_dmf.cpp"
 }
 $3 = ((($tree)) + 12|0); //@line 516 "load_dmf.cpp"
 $4 = ((($tree)) + 4|0); //@line 521 "load_dmf.cpp"
 $5 = ((($tree)) + 8|0); //@line 521 "load_dmf.cpp"
 $6 = ((($tree)) + 16|0); //@line 546 "load_dmf.cpp"
 $28 = $1;
 while(1) {
  $$pre$i = HEAPU8[$3>>0]|(HEAPU8[$3+1>>0]<<8)|(HEAPU8[$3+2>>0]<<16)|(HEAPU8[$3+3>>0]<<24); //@line 516 "load_dmf.cpp"
  $$in$i = 7;$9 = $$pre$i;$bitv$02$i = 1;$x$01$i = 0;
  while(1) {
   $7 = (($$in$i) + -1)|0; //@line 514 "load_dmf.cpp"
   $8 = ($9|0)==(0); //@line 516 "load_dmf.cpp"
   if ($8) {
    $11 = HEAPU8[$tree>>0]|(HEAPU8[$tree+1>>0]<<8)|(HEAPU8[$tree+2>>0]<<16)|(HEAPU8[$tree+3>>0]<<24); //@line 521 "load_dmf.cpp"
    $12 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8)|(HEAPU8[$4+2>>0]<<16)|(HEAPU8[$4+3>>0]<<24); //@line 521 "load_dmf.cpp"
    $13 = ($11>>>0)<($12>>>0); //@line 521 "load_dmf.cpp"
    if ($13) {
     $14 = ((($11)) + 1|0); //@line 521 "load_dmf.cpp"
     HEAP8[$tree>>0]=$14&255;HEAP8[$tree+1>>0]=($14>>8)&255;HEAP8[$tree+2>>0]=($14>>16)&255;HEAP8[$tree+3>>0]=$14>>24; //@line 521 "load_dmf.cpp"
     $15 = HEAP8[$11>>0]|0; //@line 521 "load_dmf.cpp"
     $16 = $15&255; //@line 521 "load_dmf.cpp"
     $17 = $16;
    } else {
     $17 = 0;
    }
    HEAP8[$5>>0]=$17&255;HEAP8[$5+1>>0]=($17>>8)&255;HEAP8[$5+2>>0]=($17>>16)&255;HEAP8[$5+3>>0]=$17>>24; //@line 521 "load_dmf.cpp"
    $19 = $17;$storemerge$i = 7;
   } else {
    $10 = (($9) + -1)|0; //@line 518 "load_dmf.cpp"
    $$pre3$i = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8)|(HEAPU8[$5+2>>0]<<16)|(HEAPU8[$5+3>>0]<<24); //@line 524 "load_dmf.cpp"
    $19 = $$pre3$i;$storemerge$i = $10;
   }
   HEAP8[$3>>0]=$storemerge$i&255;HEAP8[$3+1>>0]=($storemerge$i>>8)&255;HEAP8[$3+2>>0]=($storemerge$i>>16)&255;HEAP8[$3+3>>0]=$storemerge$i>>24; //@line 518 "load_dmf.cpp"
   $18 = $19 & 1; //@line 524 "load_dmf.cpp"
   $20 = ($18|0)==(0); //@line 524 "load_dmf.cpp"
   $21 = $20 ? 0 : $bitv$02$i; //@line 524 "load_dmf.cpp"
   $x$0$$i = $21 | $x$01$i; //@line 524 "load_dmf.cpp"
   $22 = $bitv$02$i&255; //@line 525 "load_dmf.cpp"
   $23 = $22 << 1; //@line 525 "load_dmf.cpp"
   $24 = $23&255; //@line 525 "load_dmf.cpp"
   $25 = $19 >>> 1; //@line 526 "load_dmf.cpp"
   HEAP8[$5>>0]=$25&255;HEAP8[$5+1>>0]=($25>>8)&255;HEAP8[$5+2>>0]=($25>>16)&255;HEAP8[$5+3>>0]=$25>>24; //@line 526 "load_dmf.cpp"
   $26 = ($7|0)==(0); //@line 514 "load_dmf.cpp"
   if ($26) {
    $x$0$$i$lcssa = $x$0$$i;
    break;
   } else {
    $$in$i = $7;$9 = $storemerge$i;$bitv$02$i = $24;$x$01$i = $x$0$$i;
   }
  }
  $27 = (((((($tree)) + 24|0) + (($28*5)|0)|0)) + 4|0); //@line 543 "load_dmf.cpp"
  HEAP8[$27>>0] = $x$0$$i$lcssa; //@line 543 "load_dmf.cpp"
  $$pre$i1 = HEAPU8[$3>>0]|(HEAPU8[$3+1>>0]<<8)|(HEAPU8[$3+2>>0]<<16)|(HEAPU8[$3+3>>0]<<24); //@line 516 "load_dmf.cpp"
  $29 = ($$pre$i1|0)==(0); //@line 516 "load_dmf.cpp"
  if ($29) {
   $30 = HEAPU8[$tree>>0]|(HEAPU8[$tree+1>>0]<<8)|(HEAPU8[$tree+2>>0]<<16)|(HEAPU8[$tree+3>>0]<<24); //@line 521 "load_dmf.cpp"
   $31 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8)|(HEAPU8[$4+2>>0]<<16)|(HEAPU8[$4+3>>0]<<24); //@line 521 "load_dmf.cpp"
   $32 = ($30>>>0)<($31>>>0); //@line 521 "load_dmf.cpp"
   if ($32) {
    $33 = ((($30)) + 1|0); //@line 521 "load_dmf.cpp"
    HEAP8[$tree>>0]=$33&255;HEAP8[$tree+1>>0]=($33>>8)&255;HEAP8[$tree+2>>0]=($33>>16)&255;HEAP8[$tree+3>>0]=$33>>24; //@line 521 "load_dmf.cpp"
    $34 = HEAP8[$30>>0]|0; //@line 521 "load_dmf.cpp"
    $35 = $34&255; //@line 521 "load_dmf.cpp"
    $37 = $35;
   } else {
    $37 = 0;
   }
   HEAP8[$3>>0]=7&255;HEAP8[$3+1>>0]=(7>>8)&255;HEAP8[$3+2>>0]=(7>>16)&255;HEAP8[$3+3>>0]=7>>24; //@line 518 "load_dmf.cpp"
   $36 = $37 & 1; //@line 524 "load_dmf.cpp"
   $38 = ($36|0)==(0); //@line 524 "load_dmf.cpp"
   $39 = $37 >>> 1; //@line 526 "load_dmf.cpp"
   HEAP8[$5>>0]=$39&255;HEAP8[$5+1>>0]=($39>>8)&255;HEAP8[$5+2>>0]=($39>>16)&255;HEAP8[$5+3>>0]=$39>>24; //@line 526 "load_dmf.cpp"
   $70 = $39;$71 = $38;$storemerge$i621 = 7;
   label = 15;
  } else {
   $40 = (($$pre$i1) + -1)|0; //@line 518 "load_dmf.cpp"
   $$pre3$i5 = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8)|(HEAPU8[$5+2>>0]<<16)|(HEAPU8[$5+3>>0]<<24); //@line 524 "load_dmf.cpp"
   HEAP8[$3>>0]=$40&255;HEAP8[$3+1>>0]=($40>>8)&255;HEAP8[$3+2>>0]=($40>>16)&255;HEAP8[$3+3>>0]=$40>>24; //@line 518 "load_dmf.cpp"
   $41 = $$pre3$i5 & 1; //@line 524 "load_dmf.cpp"
   $42 = ($41|0)==(0); //@line 524 "load_dmf.cpp"
   $43 = $$pre3$i5 >>> 1; //@line 526 "load_dmf.cpp"
   HEAP8[$5>>0]=$43&255;HEAP8[$5+1>>0]=($43>>8)&255;HEAP8[$5+2>>0]=($43>>16)&255;HEAP8[$5+3>>0]=$43>>24; //@line 526 "load_dmf.cpp"
   $44 = ($40|0)==(0); //@line 516 "load_dmf.cpp"
   if ($44) {
    $46 = HEAPU8[$tree>>0]|(HEAPU8[$tree+1>>0]<<8)|(HEAPU8[$tree+2>>0]<<16)|(HEAPU8[$tree+3>>0]<<24); //@line 521 "load_dmf.cpp"
    $47 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8)|(HEAPU8[$4+2>>0]<<16)|(HEAPU8[$4+3>>0]<<24); //@line 521 "load_dmf.cpp"
    $48 = ($46>>>0)<($47>>>0); //@line 521 "load_dmf.cpp"
    if ($48) {
     $49 = ((($46)) + 1|0); //@line 521 "load_dmf.cpp"
     HEAP8[$tree>>0]=$49&255;HEAP8[$tree+1>>0]=($49>>8)&255;HEAP8[$tree+2>>0]=($49>>16)&255;HEAP8[$tree+3>>0]=$49>>24; //@line 521 "load_dmf.cpp"
     $50 = HEAP8[$46>>0]|0; //@line 521 "load_dmf.cpp"
     $51 = $50&255; //@line 521 "load_dmf.cpp"
     $52 = $51;
    } else {
     $52 = 0;
    }
    HEAP8[$5>>0]=$52&255;HEAP8[$5+1>>0]=($52>>8)&255;HEAP8[$5+2>>0]=($52>>16)&255;HEAP8[$5+3>>0]=$52>>24; //@line 521 "load_dmf.cpp"
    $54 = $52;$72 = $42;$storemerge$i14 = 7;
   } else {
    $70 = $43;$71 = $42;$storemerge$i621 = $40;
    label = 15;
   }
  }
  if ((label|0) == 15) {
   label = 0;
   $45 = (($storemerge$i621) + -1)|0; //@line 518 "load_dmf.cpp"
   $54 = $70;$72 = $71;$storemerge$i14 = $45;
  }
  HEAP8[$3>>0]=$storemerge$i14&255;HEAP8[$3+1>>0]=($storemerge$i14>>8)&255;HEAP8[$3+2>>0]=($storemerge$i14>>16)&255;HEAP8[$3+3>>0]=$storemerge$i14>>24; //@line 518 "load_dmf.cpp"
  $53 = $54 & 1; //@line 524 "load_dmf.cpp"
  $55 = ($53|0)==(0); //@line 524 "load_dmf.cpp"
  $56 = $54 >>> 1; //@line 526 "load_dmf.cpp"
  HEAP8[$5>>0]=$56&255;HEAP8[$5+1>>0]=($56>>8)&255;HEAP8[$5+2>>0]=($56>>16)&255;HEAP8[$5+3>>0]=$56>>24; //@line 526 "load_dmf.cpp"
  $57 = HEAPU8[$6>>0]|(HEAPU8[$6+1>>0]<<8)|(HEAPU8[$6+2>>0]<<16)|(HEAPU8[$6+3>>0]<<24); //@line 546 "load_dmf.cpp"
  $58 = ($57>>>0)>(255); //@line 547 "load_dmf.cpp"
  if ($58) {
   label = 26;
   break;
  }
  $59 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 548 "load_dmf.cpp"
  $60 = (($59) + 1)|0; //@line 548 "load_dmf.cpp"
  HEAP8[$0>>0]=$60&255;HEAP8[$0+1>>0]=($60>>8)&255;HEAP8[$0+2>>0]=($60>>16)&255;HEAP8[$0+3>>0]=$60>>24; //@line 548 "load_dmf.cpp"
  HEAP8[$6>>0]=$60&255;HEAP8[$6+1>>0]=($60>>8)&255;HEAP8[$6+2>>0]=($60>>16)&255;HEAP8[$6+3>>0]=$60>>24; //@line 549 "load_dmf.cpp"
  if ($72) {
   $63 = (((($tree)) + 24|0) + (($57*5)|0)|0); //@line 556 "load_dmf.cpp"
   HEAP8[$63>>0]=-1&255;HEAP8[$63+1>>0]=-1>>8; //@line 556 "load_dmf.cpp"
  } else {
   $61 = $60&65535; //@line 552 "load_dmf.cpp"
   $62 = (((($tree)) + 24|0) + (($57*5)|0)|0); //@line 552 "load_dmf.cpp"
   HEAP8[$62>>0]=$61&255;HEAP8[$62+1>>0]=$61>>8; //@line 552 "load_dmf.cpp"
   __Z10DMFNewNodeP9DMF_HTREE($tree); //@line 553 "load_dmf.cpp"
  }
  $64 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 558 "load_dmf.cpp"
  HEAP8[$6>>0]=$64&255;HEAP8[$6+1>>0]=($64>>8)&255;HEAP8[$6+2>>0]=($64>>16)&255;HEAP8[$6+3>>0]=$64>>24; //@line 558 "load_dmf.cpp"
  if ($55) {
   $$lcssa25 = $57;
   label = 25;
   break;
  }
  $65 = $64&65535; //@line 561 "load_dmf.cpp"
  $66 = (((((($tree)) + 24|0) + (($57*5)|0)|0)) + 2|0); //@line 561 "load_dmf.cpp"
  HEAP8[$66>>0]=$65&255;HEAP8[$66+1>>0]=$65>>8; //@line 561 "load_dmf.cpp"
  $67 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 541 "load_dmf.cpp"
  $68 = ($67>>>0)>(255); //@line 542 "load_dmf.cpp"
  if ($68) {
   label = 26;
   break;
  } else {
   $28 = $67;
  }
 }
 if ((label|0) == 25) {
  $69 = (((((($tree)) + 24|0) + (($$lcssa25*5)|0)|0)) + 2|0); //@line 565 "load_dmf.cpp"
  HEAP8[$69>>0]=-1&255;HEAP8[$69+1>>0]=-1>>8; //@line 565 "load_dmf.cpp"
  return; //@line 567 "load_dmf.cpp"
 }
 else if ((label|0) == 26) {
  return; //@line 567 "load_dmf.cpp"
 }
}
function __ZN10CSoundFile16AdjustSampleLoopEP14_MODINSTRUMENT($this,$pIns) {
 $this = $this|0;
 $pIns = $pIns|0;
 var $$$smpfix$2 = 0, $$off = 0, $$old = 0, $$old3 = 0, $$old9 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phiZ2D = 0, $$pre7$pre = 0, $$smpfix$2 = 0, $$smpfix$2$lobit = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $kscan$0 = 0, $kscan$0$lcssa = 0;
 var $kscan$15 = 0, $or$cond = 0, $or$cond10 = 0, $or$cond11 = 0, $or$cond4 = 0, $smpfix$0 = 0, $smpfix$1 = 0, $smpfix$26 = 0, $smpfix$4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pIns)) + 20|0); //@line 1552 "sndfile.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1552 "sndfile.cpp"
 $2 = ($1|0)==(0|0); //@line 1552 "sndfile.cpp"
 if ($2) {
  return; //@line 1639 "sndfile.cpp"
 }
 $3 = HEAP32[$pIns>>2]|0; //@line 1553 "sndfile.cpp"
 $4 = ($3>>>0)>(16000000); //@line 1553 "sndfile.cpp"
 if ($4) {
  HEAP32[$pIns>>2] = 16000000; //@line 1553 "sndfile.cpp"
  $8 = 16000000;
 } else {
  $8 = $3;
 }
 $5 = ((($pIns)) + 8|0); //@line 1554 "sndfile.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 1554 "sndfile.cpp"
 $7 = ($6>>>0)>($8>>>0); //@line 1554 "sndfile.cpp"
 if ($7) {
  HEAP32[$5>>2] = $8; //@line 1554 "sndfile.cpp"
  $16 = $8;
 } else {
  $16 = $6;
 }
 $9 = ((($pIns)) + 4|0); //@line 1555 "sndfile.cpp"
 $10 = HEAP32[$9>>2]|0; //@line 1555 "sndfile.cpp"
 $11 = (($8) + 2)|0; //@line 1555 "sndfile.cpp"
 $12 = ($10>>>0)>($11>>>0); //@line 1555 "sndfile.cpp"
 if ($12) {
  HEAP32[$9>>2] = $11; //@line 1555 "sndfile.cpp"
  $14 = $11;
 } else {
  $14 = $10;
 }
 $13 = (($14) + 2)|0; //@line 1556 "sndfile.cpp"
 $15 = ($13>>>0)<($16>>>0); //@line 1556 "sndfile.cpp"
 if ($15) {
  $$phi$trans$insert = ((($pIns)) + 34|0);
  $$pre = HEAP16[$$phi$trans$insert>>1]|0; //@line 1562 "sndfile.cpp"
  $$pre$phiZ2D = $$phi$trans$insert;$23 = $$pre;$56 = $16;$63 = $14;
 } else {
  HEAP32[$5>>2] = 0; //@line 1558 "sndfile.cpp"
  HEAP32[$9>>2] = 0; //@line 1558 "sndfile.cpp"
  $17 = ((($pIns)) + 34|0); //@line 1559 "sndfile.cpp"
  $18 = HEAP16[$17>>1]|0; //@line 1559 "sndfile.cpp"
  $19 = $18&65535; //@line 1559 "sndfile.cpp"
  $20 = $19 & 65533; //@line 1559 "sndfile.cpp"
  $21 = $20&65535; //@line 1559 "sndfile.cpp"
  HEAP16[$17>>1] = $21; //@line 1559 "sndfile.cpp"
  $$pre$phiZ2D = $17;$23 = $21;$56 = 0;$63 = 0;
 }
 $22 = $23&65535; //@line 1562 "sndfile.cpp"
 $24 = $22 & 1; //@line 1562 "sndfile.cpp"
 $25 = ($24|0)==(0); //@line 1562 "sndfile.cpp"
 if (!($25)) {
  $26 = $22 & 64; //@line 1566 "sndfile.cpp"
  $27 = ($26|0)==(0); //@line 1566 "sndfile.cpp"
  if ($27) {
   $44 = (($1) + ($8<<1)|0); //@line 1572 "sndfile.cpp"
   HEAP16[$44>>1] = 0; //@line 1572 "sndfile.cpp"
   $45 = (($8) + 1)|0; //@line 1572 "sndfile.cpp"
   $46 = (($1) + ($45<<1)|0); //@line 1572 "sndfile.cpp"
   HEAP16[$46>>1] = 0; //@line 1572 "sndfile.cpp"
   $47 = (($1) + ($11<<1)|0); //@line 1572 "sndfile.cpp"
   HEAP16[$47>>1] = 0; //@line 1572 "sndfile.cpp"
   $48 = (($8) + 3)|0; //@line 1572 "sndfile.cpp"
   $49 = (($1) + ($48<<1)|0); //@line 1572 "sndfile.cpp"
   HEAP16[$49>>1] = 0; //@line 1572 "sndfile.cpp"
   $50 = (($8) + 4)|0; //@line 1572 "sndfile.cpp"
   $51 = (($1) + ($50<<1)|0); //@line 1572 "sndfile.cpp"
   HEAP16[$51>>1] = 0; //@line 1572 "sndfile.cpp"
  } else {
   $28 = $8 << 1; //@line 1568 "sndfile.cpp"
   $29 = (($1) + ($28<<1)|0); //@line 1568 "sndfile.cpp"
   HEAP16[$29>>1] = 0; //@line 1568 "sndfile.cpp"
   $30 = (($28) + 2)|0; //@line 1568 "sndfile.cpp"
   $31 = (($1) + ($30<<1)|0); //@line 1568 "sndfile.cpp"
   HEAP16[$31>>1] = 0; //@line 1568 "sndfile.cpp"
   $32 = (($28) + 4)|0; //@line 1568 "sndfile.cpp"
   $33 = (($1) + ($32<<1)|0); //@line 1568 "sndfile.cpp"
   HEAP16[$33>>1] = 0; //@line 1568 "sndfile.cpp"
   $34 = (($28) + 6)|0; //@line 1568 "sndfile.cpp"
   $35 = (($1) + ($34<<1)|0); //@line 1568 "sndfile.cpp"
   HEAP16[$35>>1] = 0; //@line 1568 "sndfile.cpp"
   $36 = $28 | 1; //@line 1569 "sndfile.cpp"
   $37 = (($1) + ($36<<1)|0); //@line 1569 "sndfile.cpp"
   HEAP16[$37>>1] = 0; //@line 1569 "sndfile.cpp"
   $38 = (($28) + 3)|0; //@line 1569 "sndfile.cpp"
   $39 = (($1) + ($38<<1)|0); //@line 1569 "sndfile.cpp"
   HEAP16[$39>>1] = 0; //@line 1569 "sndfile.cpp"
   $40 = (($28) + 5)|0; //@line 1569 "sndfile.cpp"
   $41 = (($1) + ($40<<1)|0); //@line 1569 "sndfile.cpp"
   HEAP16[$41>>1] = 0; //@line 1569 "sndfile.cpp"
   $42 = (($28) + 7)|0; //@line 1569 "sndfile.cpp"
   $43 = (($1) + ($42<<1)|0); //@line 1569 "sndfile.cpp"
   HEAP16[$43>>1] = 0; //@line 1569 "sndfile.cpp"
  }
  $52 = HEAP16[$$pre$phiZ2D>>1]|0; //@line 1574 "sndfile.cpp"
  $53 = $52 & 70; //@line 1574 "sndfile.cpp"
  $54 = ($53<<16>>16)==(2); //@line 1574 "sndfile.cpp"
  if (!($54)) {
   return; //@line 1639 "sndfile.cpp"
  }
  $55 = (($56) + 3)|0; //@line 1577 "sndfile.cpp"
  $57 = ($55>>>0)<($8>>>0); //@line 1577 "sndfile.cpp"
  if ($57) {
   $58 = ((($this)) + 59680|0); //@line 1577 "sndfile.cpp"
   $59 = HEAP32[$58>>2]|0; //@line 1577 "sndfile.cpp"
   $60 = $59 & 2; //@line 1577 "sndfile.cpp"
   $61 = ($60|0)==(0); //@line 1577 "sndfile.cpp"
   if ($61) {
    return; //@line 1639 "sndfile.cpp"
   }
  }
  $62 = (($1) + ($63<<1)|0); //@line 1579 "sndfile.cpp"
  $64 = HEAP16[$62>>1]|0; //@line 1579 "sndfile.cpp"
  $65 = (($1) + ($56<<1)|0); //@line 1579 "sndfile.cpp"
  HEAP16[$65>>1] = $64; //@line 1579 "sndfile.cpp"
  $66 = (($63) + 1)|0; //@line 1580 "sndfile.cpp"
  $67 = (($1) + ($66<<1)|0); //@line 1580 "sndfile.cpp"
  $68 = HEAP16[$67>>1]|0; //@line 1580 "sndfile.cpp"
  $69 = (($56) + 1)|0; //@line 1580 "sndfile.cpp"
  $70 = (($1) + ($69<<1)|0); //@line 1580 "sndfile.cpp"
  HEAP16[$70>>1] = $68; //@line 1580 "sndfile.cpp"
  $71 = (($63) + 2)|0; //@line 1581 "sndfile.cpp"
  $72 = (($1) + ($71<<1)|0); //@line 1581 "sndfile.cpp"
  $73 = HEAP16[$72>>1]|0; //@line 1581 "sndfile.cpp"
  $74 = (($56) + 2)|0; //@line 1581 "sndfile.cpp"
  $75 = (($1) + ($74<<1)|0); //@line 1581 "sndfile.cpp"
  HEAP16[$75>>1] = $73; //@line 1581 "sndfile.cpp"
  $76 = (($63) + 3)|0; //@line 1582 "sndfile.cpp"
  $77 = (($1) + ($76<<1)|0); //@line 1582 "sndfile.cpp"
  $78 = HEAP16[$77>>1]|0; //@line 1582 "sndfile.cpp"
  $79 = (($1) + ($55<<1)|0); //@line 1582 "sndfile.cpp"
  HEAP16[$79>>1] = $78; //@line 1582 "sndfile.cpp"
  $80 = (($63) + 4)|0; //@line 1583 "sndfile.cpp"
  $81 = (($1) + ($80<<1)|0); //@line 1583 "sndfile.cpp"
  $82 = HEAP16[$81>>1]|0; //@line 1583 "sndfile.cpp"
  $83 = (($56) + 4)|0; //@line 1583 "sndfile.cpp"
  $84 = (($1) + ($83<<1)|0); //@line 1583 "sndfile.cpp"
  HEAP16[$84>>1] = $82; //@line 1583 "sndfile.cpp"
  return; //@line 1639 "sndfile.cpp"
 }
 $85 = ($8>>>0)>(256); //@line 1591 "sndfile.cpp"
 do {
  if ($85) {
   $86 = ((($this)) + 59680|0); //@line 1591 "sndfile.cpp"
   $87 = HEAP32[$86>>2]|0; //@line 1591 "sndfile.cpp"
   $88 = $87 & 3; //@line 1591 "sndfile.cpp"
   $89 = ($88|0)!=(0); //@line 1591 "sndfile.cpp"
   $90 = $22 & 64;
   $91 = ($90|0)==(0); //@line 1592 "sndfile.cpp"
   $or$cond = $91 & $89; //@line 1592 "sndfile.cpp"
   if ($or$cond) {
    $92 = (($8) + -1)|0; //@line 1594 "sndfile.cpp"
    $93 = (($1) + ($92)|0); //@line 1594 "sndfile.cpp"
    $94 = HEAP8[$93>>0]|0; //@line 1594 "sndfile.cpp"
    $95 = $94 << 24 >> 24; //@line 1594 "sndfile.cpp"
    $kscan$0 = $92;$smpfix$0 = 0;
    while(1) {
     $96 = ($kscan$0|0)>(0); //@line 1595 "sndfile.cpp"
     if (!($96)) {
      $kscan$0$lcssa = $kscan$0;$smpfix$1 = $smpfix$0;
      break;
     }
     $97 = (($kscan$0) + -1)|0; //@line 1597 "sndfile.cpp"
     $98 = (($1) + ($97)|0); //@line 1597 "sndfile.cpp"
     $99 = HEAP8[$98>>0]|0; //@line 1597 "sndfile.cpp"
     $100 = $99 << 24 >> 24; //@line 1597 "sndfile.cpp"
     $101 = ($99<<24>>24)==($94<<24>>24); //@line 1598 "sndfile.cpp"
     if ($101) {
      $kscan$0 = $97;$smpfix$0 = $100;
     } else {
      $kscan$0$lcssa = $kscan$0;$smpfix$1 = $100;
      break;
     }
    }
    $102 = (($smpfix$1) - ($95))|0; //@line 1600 "sndfile.cpp"
    $103 = $23 & 2; //@line 1601 "sndfile.cpp"
    $104 = ($103<<16>>16)==(0); //@line 1601 "sndfile.cpp"
    if ($104) {
     $$old = (($102) + 8)|0; //@line 1602 "sndfile.cpp"
     $$old3 = ($$old>>>0)>(16); //@line 1602 "sndfile.cpp"
     $107 = ($kscan$0$lcssa|0)<($8|0); //@line 1604 "sndfile.cpp"
     $or$cond10 = $$old3 & $107; //@line 1602 "sndfile.cpp"
     if ($or$cond10) {
      $kscan$15 = $kscan$0$lcssa;$smpfix$26 = $smpfix$1;
     } else {
      $118 = $23;
      break;
     }
    } else {
     $105 = ($kscan$0$lcssa|0)>($56|0); //@line 1601 "sndfile.cpp"
     $$off = (($102) + 8)|0;
     $106 = ($$off>>>0)>(16); //@line 1602 "sndfile.cpp"
     $or$cond4 = $106 & $105; //@line 1602 "sndfile.cpp"
     $$old9 = ($kscan$0$lcssa|0)<($8|0); //@line 1604 "sndfile.cpp"
     $or$cond11 = $or$cond4 & $$old9; //@line 1602 "sndfile.cpp"
     if ($or$cond11) {
      $kscan$15 = $kscan$0$lcssa;$smpfix$26 = $smpfix$1;
     } else {
      $118 = $23;
      break;
     }
    }
    while(1) {
     $108 = $kscan$15 & 7; //@line 1606 "sndfile.cpp"
     $109 = ($108|0)==(0); //@line 1606 "sndfile.cpp"
     if ($109) {
      $110 = ($smpfix$26|0)>(0); //@line 1608 "sndfile.cpp"
      $111 = $110 << 31 >> 31; //@line 1608 "sndfile.cpp"
      $$smpfix$2 = (($111) + ($smpfix$26))|0; //@line 1608 "sndfile.cpp"
      $$smpfix$2$lobit = $$smpfix$2 >>> 31; //@line 1609 "sndfile.cpp"
      $$$smpfix$2 = (($$smpfix$2$lobit) + ($$smpfix$2))|0; //@line 1609 "sndfile.cpp"
      $smpfix$4 = $$$smpfix$2;
     } else {
      $smpfix$4 = $smpfix$26;
     }
     $112 = $smpfix$4&255; //@line 1611 "sndfile.cpp"
     $113 = (($1) + ($kscan$15)|0); //@line 1611 "sndfile.cpp"
     HEAP8[$113>>0] = $112; //@line 1611 "sndfile.cpp"
     $114 = (($kscan$15) + 1)|0; //@line 1612 "sndfile.cpp"
     $115 = HEAP32[$pIns>>2]|0; //@line 1604 "sndfile.cpp"
     $116 = ($114|0)<($115|0); //@line 1604 "sndfile.cpp"
     if ($116) {
      $kscan$15 = $114;$smpfix$26 = $smpfix$4;
     } else {
      break;
     }
    }
    $$pre7$pre = HEAP16[$$pre$phiZ2D>>1]|0; //@line 1618 "sndfile.cpp"
    $118 = $$pre7$pre;
   } else {
    $118 = $23;
   }
  } else {
   $118 = $23;
  }
 } while(0);
 $117 = $118 & 64; //@line 1618 "sndfile.cpp"
 $119 = ($117<<16>>16)==(0); //@line 1618 "sndfile.cpp"
 if ($119) {
  $136 = (($1) + ($8)|0); //@line 1625 "sndfile.cpp"
  HEAP8[$136>>0] = 0; //@line 1625 "sndfile.cpp"
  $137 = (($8) + 1)|0; //@line 1625 "sndfile.cpp"
  $138 = (($1) + ($137)|0); //@line 1625 "sndfile.cpp"
  HEAP8[$138>>0] = 0; //@line 1625 "sndfile.cpp"
  $139 = (($1) + ($11)|0); //@line 1625 "sndfile.cpp"
  HEAP8[$139>>0] = 0; //@line 1625 "sndfile.cpp"
  $140 = (($8) + 3)|0; //@line 1625 "sndfile.cpp"
  $141 = (($1) + ($140)|0); //@line 1625 "sndfile.cpp"
  HEAP8[$141>>0] = 0; //@line 1625 "sndfile.cpp"
  $142 = (($8) + 4)|0; //@line 1625 "sndfile.cpp"
  $143 = (($1) + ($142)|0); //@line 1625 "sndfile.cpp"
  HEAP8[$143>>0] = 0; //@line 1625 "sndfile.cpp"
 } else {
  $120 = $8 << 1; //@line 1620 "sndfile.cpp"
  $121 = (($1) + ($120)|0); //@line 1620 "sndfile.cpp"
  HEAP8[$121>>0] = 0; //@line 1620 "sndfile.cpp"
  $122 = (($120) + 2)|0; //@line 1620 "sndfile.cpp"
  $123 = (($1) + ($122)|0); //@line 1620 "sndfile.cpp"
  HEAP8[$123>>0] = 0; //@line 1620 "sndfile.cpp"
  $124 = (($120) + 4)|0; //@line 1620 "sndfile.cpp"
  $125 = (($1) + ($124)|0); //@line 1620 "sndfile.cpp"
  HEAP8[$125>>0] = 0; //@line 1620 "sndfile.cpp"
  $126 = (($120) + 6)|0; //@line 1620 "sndfile.cpp"
  $127 = (($1) + ($126)|0); //@line 1620 "sndfile.cpp"
  HEAP8[$127>>0] = 0; //@line 1620 "sndfile.cpp"
  $128 = $120 | 1; //@line 1621 "sndfile.cpp"
  $129 = (($1) + ($128)|0); //@line 1621 "sndfile.cpp"
  HEAP8[$129>>0] = 0; //@line 1621 "sndfile.cpp"
  $130 = (($120) + 3)|0; //@line 1621 "sndfile.cpp"
  $131 = (($1) + ($130)|0); //@line 1621 "sndfile.cpp"
  HEAP8[$131>>0] = 0; //@line 1621 "sndfile.cpp"
  $132 = (($120) + 5)|0; //@line 1621 "sndfile.cpp"
  $133 = (($1) + ($132)|0); //@line 1621 "sndfile.cpp"
  HEAP8[$133>>0] = 0; //@line 1621 "sndfile.cpp"
  $134 = (($120) + 7)|0; //@line 1621 "sndfile.cpp"
  $135 = (($1) + ($134)|0); //@line 1621 "sndfile.cpp"
  HEAP8[$135>>0] = 0; //@line 1621 "sndfile.cpp"
 }
 $144 = HEAP16[$$pre$phiZ2D>>1]|0; //@line 1627 "sndfile.cpp"
 $145 = $144 & 70; //@line 1627 "sndfile.cpp"
 $146 = ($145<<16>>16)==(2); //@line 1627 "sndfile.cpp"
 if (!($146)) {
  return; //@line 1639 "sndfile.cpp"
 }
 $147 = HEAP32[$5>>2]|0; //@line 1629 "sndfile.cpp"
 $148 = (($147) + 3)|0; //@line 1629 "sndfile.cpp"
 $149 = HEAP32[$pIns>>2]|0; //@line 1629 "sndfile.cpp"
 $150 = ($148>>>0)<($149>>>0); //@line 1629 "sndfile.cpp"
 if ($150) {
  $151 = ((($this)) + 59680|0); //@line 1629 "sndfile.cpp"
  $152 = HEAP32[$151>>2]|0; //@line 1629 "sndfile.cpp"
  $153 = $152 & 3; //@line 1629 "sndfile.cpp"
  $154 = ($153|0)==(0); //@line 1629 "sndfile.cpp"
  if ($154) {
   return; //@line 1639 "sndfile.cpp"
  }
 }
 $155 = HEAP32[$9>>2]|0; //@line 1631 "sndfile.cpp"
 $156 = (($1) + ($155)|0); //@line 1631 "sndfile.cpp"
 $157 = HEAP8[$156>>0]|0; //@line 1631 "sndfile.cpp"
 $158 = (($1) + ($147)|0); //@line 1631 "sndfile.cpp"
 HEAP8[$158>>0] = $157; //@line 1631 "sndfile.cpp"
 $159 = HEAP32[$9>>2]|0; //@line 1632 "sndfile.cpp"
 $160 = (($159) + 1)|0; //@line 1632 "sndfile.cpp"
 $161 = (($1) + ($160)|0); //@line 1632 "sndfile.cpp"
 $162 = HEAP8[$161>>0]|0; //@line 1632 "sndfile.cpp"
 $163 = HEAP32[$5>>2]|0; //@line 1632 "sndfile.cpp"
 $164 = (($163) + 1)|0; //@line 1632 "sndfile.cpp"
 $165 = (($1) + ($164)|0); //@line 1632 "sndfile.cpp"
 HEAP8[$165>>0] = $162; //@line 1632 "sndfile.cpp"
 $166 = HEAP32[$9>>2]|0; //@line 1633 "sndfile.cpp"
 $167 = (($166) + 2)|0; //@line 1633 "sndfile.cpp"
 $168 = (($1) + ($167)|0); //@line 1633 "sndfile.cpp"
 $169 = HEAP8[$168>>0]|0; //@line 1633 "sndfile.cpp"
 $170 = HEAP32[$5>>2]|0; //@line 1633 "sndfile.cpp"
 $171 = (($170) + 2)|0; //@line 1633 "sndfile.cpp"
 $172 = (($1) + ($171)|0); //@line 1633 "sndfile.cpp"
 HEAP8[$172>>0] = $169; //@line 1633 "sndfile.cpp"
 $173 = HEAP32[$9>>2]|0; //@line 1634 "sndfile.cpp"
 $174 = (($173) + 3)|0; //@line 1634 "sndfile.cpp"
 $175 = (($1) + ($174)|0); //@line 1634 "sndfile.cpp"
 $176 = HEAP8[$175>>0]|0; //@line 1634 "sndfile.cpp"
 $177 = HEAP32[$5>>2]|0; //@line 1634 "sndfile.cpp"
 $178 = (($177) + 3)|0; //@line 1634 "sndfile.cpp"
 $179 = (($1) + ($178)|0); //@line 1634 "sndfile.cpp"
 HEAP8[$179>>0] = $176; //@line 1634 "sndfile.cpp"
 $180 = HEAP32[$9>>2]|0; //@line 1635 "sndfile.cpp"
 $181 = (($180) + 4)|0; //@line 1635 "sndfile.cpp"
 $182 = (($1) + ($181)|0); //@line 1635 "sndfile.cpp"
 $183 = HEAP8[$182>>0]|0; //@line 1635 "sndfile.cpp"
 $184 = HEAP32[$5>>2]|0; //@line 1635 "sndfile.cpp"
 $185 = (($184) + 4)|0; //@line 1635 "sndfile.cpp"
 $186 = (($1) + ($185)|0); //@line 1635 "sndfile.cpp"
 HEAP8[$186>>0] = $183; //@line 1635 "sndfile.cpp"
 return; //@line 1639 "sndfile.cpp"
}
function __ZN10CSoundFile14LoadMixPluginsEPKvj($this,$pData,$nLen) {
 $this = $this|0;
 $pData = $pData|0;
 $nLen = $nLen|0;
 var $$sum = 0, $$sum3 = 0, $$sum4 = 0, $$sum6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ch$09 = 0, $exitcond = 0, $nPos$0$lcssa = 0, $nPos$010 = 0;
 var $or$cond = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ($nLen>>>0)>(8); //@line 1467 "load_it.cpp"
 if (!($0)) {
  $nPos$0$lcssa = 0;
  return ($nPos$0$lcssa|0); //@line 1514 "load_it.cpp"
 }
 $1 = (($nLen) + -8)|0; //@line 1473 "load_it.cpp"
 $11 = 8;$nPos$010 = 0;
 while(1) {
  $$sum = (($nPos$010) + 4)|0; //@line 1472 "load_it.cpp"
  $2 = (($pData) + ($$sum)|0); //@line 1472 "load_it.cpp"
  $3 = HEAP32[$2>>2]|0; //@line 1472 "load_it.cpp"
  $4 = (($1) - ($nPos$010))|0; //@line 1473 "load_it.cpp"
  $5 = ($3>>>0)>($4>>>0); //@line 1473 "load_it.cpp"
  if ($5) {
   $nPos$0$lcssa = $nPos$010;
   label = 16;
   break;
  }
  $6 = (($pData) + ($nPos$010)|0); //@line 1472 "load_it.cpp"
  $7 = HEAP32[$6>>2]|0; //@line 1474 "load_it.cpp"
  $8 = ($7|0)==(1481001027); //@line 1474 "load_it.cpp"
  if ($8) {
   $ch$09 = 0;
   while(1) {
    $9 = $ch$09 << 2; //@line 1476 "load_it.cpp"
    $10 = ($9>>>0)<($3>>>0); //@line 1476 "load_it.cpp"
    if ($10) {
     $$sum6 = (($9) + ($11))|0; //@line 1478 "load_it.cpp"
     $12 = (($pData) + ($$sum6)|0); //@line 1478 "load_it.cpp"
     $13 = HEAP32[$12>>2]|0; //@line 1478 "load_it.cpp"
     $14 = (((((($this)) + 49600|0) + (($ch$09*36)|0)|0)) + 12|0); //@line 1478 "load_it.cpp"
     HEAP32[$14>>2] = $13; //@line 1478 "load_it.cpp"
    }
    $15 = (($ch$09) + 1)|0; //@line 1476 "load_it.cpp"
    $exitcond = ($15|0)==(64); //@line 1476 "load_it.cpp"
    if ($exitcond) {
     break;
    } else {
     $ch$09 = $15;
    }
   }
  } else {
   $16 = $7&255;
   $17 = ($16<<24>>24)==(70); //@line 1482 "load_it.cpp"
   if (!($17)) {
    $nPos$0$lcssa = $nPos$010;
    label = 16;
    break;
   }
   $18 = (($nPos$010) + 1)|0; //@line 1482 "load_it.cpp"
   $19 = (($pData) + ($18)|0); //@line 1482 "load_it.cpp"
   $20 = HEAP8[$19>>0]|0; //@line 1482 "load_it.cpp"
   $21 = ($20<<24>>24)==(88); //@line 1482 "load_it.cpp"
   if (!($21)) {
    $nPos$0$lcssa = $nPos$010;
    label = 16;
    break;
   }
   $22 = (($nPos$010) + 2)|0; //@line 1483 "load_it.cpp"
   $23 = (($pData) + ($22)|0); //@line 1483 "load_it.cpp"
   $24 = HEAP8[$23>>0]|0; //@line 1483 "load_it.cpp"
   $25 = $24&255; //@line 1483 "load_it.cpp"
   $26 = ($24&255)<(48); //@line 1483 "load_it.cpp"
   if ($26) {
    $nPos$0$lcssa = $nPos$010;
    label = 16;
    break;
   }
   $27 = (($nPos$010) + 3)|0; //@line 1483 "load_it.cpp"
   $28 = (($pData) + ($27)|0); //@line 1483 "load_it.cpp"
   $29 = HEAP8[$28>>0]|0; //@line 1483 "load_it.cpp"
   $30 = ($29&255)<(48); //@line 1483 "load_it.cpp"
   if ($30) {
    $nPos$0$lcssa = $nPos$010;
    label = 16;
    break;
   }
   $31 = $29&255; //@line 1483 "load_it.cpp"
   $32 = ($25*10)|0; //@line 1487 "load_it.cpp"
   $33 = (($32) + -528)|0; //@line 1487 "load_it.cpp"
   $34 = (($33) + ($31))|0; //@line 1487 "load_it.cpp"
   $35 = ($34>>>0)<(8); //@line 1488 "load_it.cpp"
   $36 = ($3>>>0)>(131); //@line 1488 "load_it.cpp"
   $or$cond = $36 & $35; //@line 1488 "load_it.cpp"
   if ($or$cond) {
    $37 = (($pData) + ($11)|0); //@line 1490 "load_it.cpp"
    $$sum3 = (($nPos$010) + 136)|0; //@line 1490 "load_it.cpp"
    $38 = (($pData) + ($$sum3)|0); //@line 1490 "load_it.cpp"
    $39 = HEAP32[$38>>2]|0; //@line 1490 "load_it.cpp"
    $40 = (((((($this)) + 58496|0) + (($34*144)|0)|0)) + 16|0); //@line 1491 "load_it.cpp"
    dest=$40; src=$37; stop=dest+128|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1491 "load_it.cpp"
    $41 = (($3) + -132)|0;
    $42 = (($39) + -1)|0; //@line 1500 "load_it.cpp"
    $43 = ($42>>>0)<($41>>>0); //@line 1500 "load_it.cpp"
    if ($43) {
     $44 = (((((($this)) + 58496|0) + (($34*144)|0)|0)) + 8|0); //@line 1502 "load_it.cpp"
     $45 = (__Znaj($39)|0); //@line 1503 "load_it.cpp"
     $46 = (((((($this)) + 58496|0) + (($34*144)|0)|0)) + 12|0); //@line 1503 "load_it.cpp"
     HEAP32[$46>>2] = $45; //@line 1503 "load_it.cpp"
     HEAP32[$44>>2] = $39; //@line 1506 "load_it.cpp"
     $$sum4 = (($nPos$010) + 140)|0; //@line 1507 "load_it.cpp"
     $47 = (($pData) + ($$sum4)|0); //@line 1507 "load_it.cpp"
     _memcpy(($45|0),($47|0),($39|0))|0; //@line 1507 "load_it.cpp"
    }
   }
  }
  $48 = (($nPos$010) + 8)|0; //@line 1512 "load_it.cpp"
  $49 = (($48) + ($3))|0; //@line 1512 "load_it.cpp"
  $50 = (($49) + 8)|0; //@line 1467 "load_it.cpp"
  $51 = ($50>>>0)<($nLen>>>0); //@line 1467 "load_it.cpp"
  if ($51) {
   $11 = $50;$nPos$010 = $49;
  } else {
   $nPos$0$lcssa = $49;
   label = 16;
   break;
  }
 }
 if ((label|0) == 16) {
  return ($nPos$0$lcssa|0); //@line 1514 "load_it.cpp"
 }
 return (0)|0;
}
function __ZN10CSoundFile7ReadS3MEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$$ = 0, $$$3 = 0, $$0 = 0, $$1 = 0, $$2 = 0, $$3 = 0, $$34 = 0, $$35 = 0, $$43 = 0, $$44 = 0, $$45 = 0, $$46 = 0, $$dwMemPos$0 = 0, $$flags$1 = 0, $$input$i = 0, $$input$i48 = 0, $$input$i50 = 0, $$lcssa = 0, $$lobit = 0;
 var $$op = 0, $$pre = 0, $$sink = 0, $$sum = 0, $$sum10 = 0, $$sum11 = 0, $$sum12 = 0, $$sum13 = 0, $$sum1415 = 0, $$sum1617 = 0, $$sum1819 = 0, $$sum20 = 0, $$sum21 = 0, $$sum22 = 0, $$sum23 = 0, $$sum24 = 0, $$sum25 = 0, $$sum26 = 0, $$sum27 = 0, $$sum28 = 0;
 var $$sum29 = 0, $$sum30 = 0, $$sum9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0;
 var $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0;
 var $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0;
 var $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0;
 var $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $command$0$i = 0, $dwMemPos$1 = 0, $exitcond = 0, $exitcond73 = 0, $exitcond75 = 0, $i$071 = 0, $iPat$062 = 0, $iRaw$061 = 0, $iSmp$066 = 0, $ich$072 = 0, $insfile = 0, $insflags = 0, $inspack = 0, $j3$0 = 0, $j3$0$ph = 0;
 var $j3$1 = 0, $j3$2 = 0, $j3$4 = 0, $j3$4$ = 0, $j3$5 = 0, $not$ = 0, $or$cond = 0, $or$cond36 = 0, $or$cond38 = 0, $or$cond39 = 0, $or$cond40 = 0, $or$cond42 = 0, $or$cond5 = 0, $pSmp$sroa$15 = 0, $param$0$i = 0, $psfh = 0, $ptr = 0, $row$0$ph = 0, $storemerge32 = 0, $storemerge33 = 0;
 var $vol$1 = 0, $z$064 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 2096|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $insfile = sp + 40|0;
 $ptr = sp + 1000|0;
 $insflags = sp + 1848|0;
 $inspack = sp + 1608|0;
 $psfh = sp + 1512|0;
 $pSmp$sroa$15 = sp;
 $0 = ($lpStream|0)==(0|0); //@line 203 "load_s3m.cpp"
 $1 = ($dwMemLength>>>0)<(241); //@line 203 "load_s3m.cpp"
 $or$cond = $0 | $1; //@line 203 "load_s3m.cpp"
 if ($or$cond) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 407 "load_s3m.cpp"
 }
 dest=$psfh; src=$lpStream; stop=dest+96|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 204 "load_s3m.cpp"
 $2 = ((($psfh)) + 32|0); //@line 207 "load_s3m.cpp"
 $3 = ((($psfh)) + 34|0); //@line 208 "load_s3m.cpp"
 $4 = ((($psfh)) + 36|0); //@line 209 "load_s3m.cpp"
 $5 = ((($psfh)) + 38|0); //@line 210 "load_s3m.cpp"
 $6 = ((($psfh)) + 40|0); //@line 211 "load_s3m.cpp"
 $7 = ((($psfh)) + 42|0); //@line 212 "load_s3m.cpp"
 $8 = ((($psfh)) + 44|0); //@line 213 "load_s3m.cpp"
 $9 = HEAPU8[$8>>0]|(HEAPU8[$8+1>>0]<<8)|(HEAPU8[$8+2>>0]<<16)|(HEAPU8[$8+3>>0]<<24); //@line 216 "load_s3m.cpp"
 $10 = ($9|0)==(1297236819); //@line 216 "load_s3m.cpp"
 if ($10) {
  $11 = ((($this)) + 59680|0); //@line 218 "load_s3m.cpp"
  HEAP32[$11>>2] = 2; //@line 218 "load_s3m.cpp"
  $12 = ((($this)) + 59804|0); //@line 219 "load_s3m.cpp"
  _memset(($12|0),0,7680)|0; //@line 219 "load_s3m.cpp"
  dest=$12; src=$psfh; stop=dest+28|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 220 "load_s3m.cpp"
  $13 = ((($psfh)) + 49|0); //@line 222 "load_s3m.cpp"
  $14 = HEAP8[$13>>0]|0; //@line 222 "load_s3m.cpp"
  $15 = $14&255; //@line 222 "load_s3m.cpp"
  $16 = ((($this)) + 59648|0); //@line 222 "load_s3m.cpp"
  $17 = ($14<<24>>24)==(0); //@line 223 "load_s3m.cpp"
  $$34 = $17 ? 6 : $15; //@line 223 "load_s3m.cpp"
  $18 = ($$34>>>0)>(31); //@line 224 "load_s3m.cpp"
  $storemerge33 = $18 ? 31 : $$34; //@line 224 "load_s3m.cpp"
  HEAP32[$16>>2] = $storemerge33; //@line 222 "load_s3m.cpp"
  $19 = ((($psfh)) + 50|0); //@line 226 "load_s3m.cpp"
  $20 = HEAP8[$19>>0]|0; //@line 226 "load_s3m.cpp"
  $21 = $20&255; //@line 226 "load_s3m.cpp"
  $22 = ((($this)) + 59652|0); //@line 226 "load_s3m.cpp"
  $23 = ($20&255)<(40); //@line 227 "load_s3m.cpp"
  $$35 = $23 ? 40 : $21; //@line 227 "load_s3m.cpp"
  $24 = ($$35>>>0)>(240); //@line 228 "load_s3m.cpp"
  $storemerge32 = $24 ? 240 : $$35; //@line 228 "load_s3m.cpp"
  HEAP32[$22>>2] = $storemerge32; //@line 226 "load_s3m.cpp"
  $25 = ((($psfh)) + 48|0); //@line 230 "load_s3m.cpp"
  $26 = HEAP8[$25>>0]|0; //@line 230 "load_s3m.cpp"
  $27 = $26&255; //@line 230 "load_s3m.cpp"
  $28 = $27 << 2; //@line 230 "load_s3m.cpp"
  $29 = ((($this)) + 59656|0); //@line 230 "load_s3m.cpp"
  $30 = ($26<<24>>24)==(0); //@line 231 "load_s3m.cpp"
  $31 = ($28>>>0)>(256); //@line 231 "load_s3m.cpp"
  $or$cond36 = $30 | $31; //@line 231 "load_s3m.cpp"
  $$46 = $or$cond36 ? 256 : $28; //@line 231 "load_s3m.cpp"
  HEAP32[$29>>2] = $$46; //@line 231 "load_s3m.cpp"
  $32 = ((($psfh)) + 51|0); //@line 232 "load_s3m.cpp"
  $33 = HEAP8[$32>>0]|0; //@line 232 "load_s3m.cpp"
  $34 = $33&255; //@line 232 "load_s3m.cpp"
  $35 = $34 & 127; //@line 232 "load_s3m.cpp"
  $36 = ((($this)) + 59748|0); //@line 232 "load_s3m.cpp"
  HEAP32[$36>>2] = $35; //@line 232 "load_s3m.cpp"
  $37 = ((($this)) + 59664|0); //@line 234 "load_s3m.cpp"
  HEAP32[$37>>2] = 4; //@line 234 "load_s3m.cpp"
  $ich$072 = 0;
  while(1) {
   $38 = (((($this)) + 49600|0) + (($ich$072*36)|0)|0); //@line 237 "load_s3m.cpp"
   HEAP32[$38>>2] = 128; //@line 237 "load_s3m.cpp"
   $39 = (((((($this)) + 49600|0) + (($ich$072*36)|0)|0)) + 4|0); //@line 238 "load_s3m.cpp"
   HEAP32[$39>>2] = 64; //@line 238 "load_s3m.cpp"
   $40 = (((((($this)) + 49600|0) + (($ich$072*36)|0)|0)) + 8|0); //@line 240 "load_s3m.cpp"
   HEAP32[$40>>2] = 256; //@line 240 "load_s3m.cpp"
   $41 = (((($psfh)) + 64|0) + ($ich$072)|0); //@line 241 "load_s3m.cpp"
   $42 = HEAP8[$41>>0]|0; //@line 241 "load_s3m.cpp"
   $43 = ($42<<24>>24)==(-1); //@line 241 "load_s3m.cpp"
   $$pre = (($ich$072) + 1)|0; //@line 235 "load_s3m.cpp"
   if (!($43)) {
    HEAP32[$37>>2] = $$pre; //@line 243 "load_s3m.cpp"
    $44 = $42&255; //@line 244 "load_s3m.cpp"
    $45 = $44 << 4; //@line 245 "load_s3m.cpp"
    $46 = $45 & 128; //@line 245 "load_s3m.cpp"
    $47 = $46 | 64; //@line 245 "load_s3m.cpp"
    HEAP32[$38>>2] = $47; //@line 245 "load_s3m.cpp"
    HEAP32[$40>>2] = 0; //@line 246 "load_s3m.cpp"
   }
   $exitcond75 = ($$pre|0)==(32); //@line 235 "load_s3m.cpp"
   if ($exitcond75) {
    break;
   } else {
    $ich$072 = $$pre;
   }
  }
  $48 = HEAP32[$37>>2]|0; //@line 249 "load_s3m.cpp"
  $49 = ($48>>>0)<(4); //@line 249 "load_s3m.cpp"
  if ($49) {
   HEAP32[$37>>2] = 4; //@line 249 "load_s3m.cpp"
  }
  $50 = HEAPU8[$6>>0]|(HEAPU8[$6+1>>0]<<8); //@line 250 "load_s3m.cpp"
  $51 = ($50&65535)<(4896); //@line 250 "load_s3m.cpp"
  if ($51) {
   label = 11;
  } else {
   $52 = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8); //@line 250 "load_s3m.cpp"
   $53 = $52 & 64; //@line 250 "load_s3m.cpp"
   $54 = ($53<<16>>16)==(0); //@line 250 "load_s3m.cpp"
   if (!($54)) {
    label = 11;
   }
  }
  if ((label|0) == 11) {
   $55 = ((($this)) + 59660|0); //@line 250 "load_s3m.cpp"
   $56 = HEAP32[$55>>2]|0; //@line 250 "load_s3m.cpp"
   $57 = $56 | 2; //@line 250 "load_s3m.cpp"
   HEAP32[$55>>2] = $57; //@line 250 "load_s3m.cpp"
  }
  $58 = HEAPU8[$2>>0]|(HEAPU8[$2+1>>0]<<8); //@line 252 "load_s3m.cpp"
  $59 = $58&65535; //@line 252 "load_s3m.cpp"
  $60 = ($58<<16>>16)==(0); //@line 253 "load_s3m.cpp"
  $$ = $60 ? 1 : $59; //@line 253 "load_s3m.cpp"
  $61 = ($$>>>0)>(256); //@line 254 "load_s3m.cpp"
  $$$ = $61 ? 256 : $$; //@line 254 "load_s3m.cpp"
  $62 = ($$$|0)==(0); //@line 255 "load_s3m.cpp"
  if ($62) {
   $dwMemPos$1 = 96;
  } else {
   $63 = ((($this)) + 53344|0); //@line 257 "load_s3m.cpp"
   $64 = ((($lpStream)) + 96|0); //@line 257 "load_s3m.cpp"
   _memcpy(($63|0),($64|0),($$$|0))|0; //@line 257 "load_s3m.cpp"
   $65 = (($$$) + 96)|0; //@line 258 "load_s3m.cpp"
   $66 = $$$ & 1; //@line 260 "load_s3m.cpp"
   $67 = ($66|0)==(0); //@line 260 "load_s3m.cpp"
   if ($67) {
    $dwMemPos$1 = $65;
   } else {
    $68 = (($lpStream) + ($65)|0); //@line 260 "load_s3m.cpp"
    $69 = HEAP8[$68>>0]|0; //@line 260 "load_s3m.cpp"
    $70 = ($69<<24>>24)==(-1); //@line 260 "load_s3m.cpp"
    $71 = (($$$) + 97)|0; //@line 260 "load_s3m.cpp"
    $$dwMemPos$0 = $70 ? $71 : $65; //@line 260 "load_s3m.cpp"
    $dwMemPos$1 = $$dwMemPos$0;
   }
  }
  $72 = HEAPU8[$3>>0]|(HEAPU8[$3+1>>0]<<8); //@line 262 "load_s3m.cpp"
  $73 = $72&65535; //@line 262 "load_s3m.cpp"
  $74 = ($73>>>0)>(239); //@line 263 "load_s3m.cpp"
  $$2 = $74 ? 239 : $73; //@line 263 "load_s3m.cpp"
  $75 = ((($this)) + 59684|0); //@line 264 "load_s3m.cpp"
  HEAP32[$75>>2] = $$2; //@line 264 "load_s3m.cpp"
  $76 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8); //@line 265 "load_s3m.cpp"
  $77 = $76&65535; //@line 265 "load_s3m.cpp"
  _memset(($ptr|0),0,512)|0; //@line 267 "load_s3m.cpp"
  $78 = (($77) + ($73))|0; //@line 270 "load_s3m.cpp"
  $79 = ($78>>>0)>(256); //@line 270 "load_s3m.cpp"
  if ($79) {
   $$0 = 0;
  } else {
   $80 = ($78|0)==(0); //@line 272 "load_s3m.cpp"
   if (!($80)) {
    $81 = (($lpStream) + ($dwMemPos$1)|0); //@line 274 "load_s3m.cpp"
    $82 = $78 << 1; //@line 274 "load_s3m.cpp"
    _memcpy(($ptr|0),($81|0),($82|0))|0; //@line 274 "load_s3m.cpp"
    $83 = (($82) + ($dwMemPos$1))|0; //@line 275 "load_s3m.cpp"
    $84 = ((($psfh)) + 53|0); //@line 279 "load_s3m.cpp"
    $85 = HEAP8[$84>>0]|0; //@line 279 "load_s3m.cpp"
    $86 = ($85<<24>>24)==(-4); //@line 279 "load_s3m.cpp"
    if ($86) {
     $i$071 = 0;
     while(1) {
      $$sum30 = (($83) + ($i$071))|0; //@line 282 "load_s3m.cpp"
      $87 = (($lpStream) + ($$sum30)|0); //@line 282 "load_s3m.cpp"
      $88 = HEAP8[$87>>0]|0; //@line 282 "load_s3m.cpp"
      $89 = $88&255; //@line 282 "load_s3m.cpp"
      $90 = $89 & 32; //@line 282 "load_s3m.cpp"
      $91 = ($90|0)==(0); //@line 282 "load_s3m.cpp"
      if (!($91)) {
       $92 = $89 << 4; //@line 284 "load_s3m.cpp"
       $93 = $92 & 240; //@line 284 "load_s3m.cpp"
       $94 = $93 | 8; //@line 284 "load_s3m.cpp"
       $95 = (((($this)) + 49600|0) + (($i$071*36)|0)|0); //@line 284 "load_s3m.cpp"
       HEAP32[$95>>2] = $94; //@line 284 "load_s3m.cpp"
      }
      $96 = (($i$071) + 1)|0; //@line 282 "load_s3m.cpp"
      $exitcond73 = ($96|0)==(32); //@line 282 "load_s3m.cpp"
      if ($exitcond73) {
       break;
      } else {
       $i$071 = $96;
      }
     }
    }
   }
   $97 = HEAP32[$37>>2]|0; //@line 288 "load_s3m.cpp"
   $98 = ($97|0)==(0); //@line 288 "load_s3m.cpp"
   if ($98) {
    $$0 = 1;
   } else {
    _memset(($insfile|0),0,960)|0; //@line 290 "load_s3m.cpp"
    $99 = ($$2|0)==(0); //@line 291 "load_s3m.cpp"
    if (!($99)) {
     $100 = ((($pSmp$sroa$15)) + 12|0); //@line 300 "load_s3m.cpp"
     $iSmp$066 = 1;
     while(1) {
      $105 = (($iSmp$066) + -1)|0; //@line 293 "load_s3m.cpp"
      $106 = (($ptr) + ($105<<1)|0); //@line 293 "load_s3m.cpp"
      $107 = HEAP16[$106>>1]|0; //@line 293 "load_s3m.cpp"
      $108 = $107&65535; //@line 293 "load_s3m.cpp"
      $109 = $108 << 4; //@line 293 "load_s3m.cpp"
      $110 = ($107<<16>>16)==(0); //@line 294 "load_s3m.cpp"
      $111 = (($109) + 80)|0;
      $112 = ($111>>>0)>($dwMemLength>>>0); //@line 294 "load_s3m.cpp"
      $or$cond38 = $110 | $112; //@line 294 "load_s3m.cpp"
      if (!($or$cond38)) {
       $113 = (($lpStream) + ($109)|0); //@line 296 "load_s3m.cpp"
       $114 = HEAP8[$113>>0]|0; //@line 296 "load_s3m.cpp"
       $$sum1415 = $109 | 1; //@line 296 "load_s3m.cpp"
       $115 = (($lpStream) + ($$sum1415)|0); //@line 296 "load_s3m.cpp"
       $$sum1617 = $109 | 13; //@line 296 "load_s3m.cpp"
       $116 = (($lpStream) + ($$sum1617)|0); //@line 296 "load_s3m.cpp"
       $117 = HEAP8[$116>>0]|0; //@line 296 "load_s3m.cpp"
       $$sum1819 = $109 | 14; //@line 296 "load_s3m.cpp"
       $118 = (($lpStream) + ($$sum1819)|0); //@line 296 "load_s3m.cpp"
       $119 = HEAPU8[$118>>0]|(HEAPU8[$118+1>>0]<<8); //@line 296 "load_s3m.cpp"
       $$sum20 = (($109) + 16)|0; //@line 296 "load_s3m.cpp"
       $120 = (($lpStream) + ($$sum20)|0); //@line 296 "load_s3m.cpp"
       $121 = HEAPU8[$120>>0]|(HEAPU8[$120+1>>0]<<8)|(HEAPU8[$120+2>>0]<<16)|(HEAPU8[$120+3>>0]<<24); //@line 296 "load_s3m.cpp"
       $$sum21 = (($109) + 20)|0; //@line 296 "load_s3m.cpp"
       $122 = (($lpStream) + ($$sum21)|0); //@line 296 "load_s3m.cpp"
       $123 = HEAPU8[$122>>0]|(HEAPU8[$122+1>>0]<<8)|(HEAPU8[$122+2>>0]<<16)|(HEAPU8[$122+3>>0]<<24); //@line 296 "load_s3m.cpp"
       $$sum22 = (($109) + 24)|0; //@line 296 "load_s3m.cpp"
       $124 = (($lpStream) + ($$sum22)|0); //@line 296 "load_s3m.cpp"
       $125 = HEAPU8[$124>>0]|(HEAPU8[$124+1>>0]<<8)|(HEAPU8[$124+2>>0]<<16)|(HEAPU8[$124+3>>0]<<24); //@line 296 "load_s3m.cpp"
       $$sum23 = (($109) + 28)|0; //@line 296 "load_s3m.cpp"
       $126 = (($lpStream) + ($$sum23)|0); //@line 296 "load_s3m.cpp"
       $127 = HEAP8[$126>>0]|0; //@line 296 "load_s3m.cpp"
       $$sum24 = (($109) + 30)|0; //@line 296 "load_s3m.cpp"
       $128 = (($lpStream) + ($$sum24)|0); //@line 296 "load_s3m.cpp"
       $129 = HEAP8[$128>>0]|0; //@line 296 "load_s3m.cpp"
       $$sum25 = (($109) + 31)|0; //@line 296 "load_s3m.cpp"
       $130 = (($lpStream) + ($$sum25)|0); //@line 296 "load_s3m.cpp"
       $131 = HEAP8[$130>>0]|0; //@line 296 "load_s3m.cpp"
       $$sum26 = (($109) + 32)|0; //@line 296 "load_s3m.cpp"
       $132 = (($lpStream) + ($$sum26)|0); //@line 296 "load_s3m.cpp"
       $133 = HEAPU8[$132>>0]|(HEAPU8[$132+1>>0]<<8)|(HEAPU8[$132+2>>0]<<16)|(HEAPU8[$132+3>>0]<<24); //@line 296 "load_s3m.cpp"
       $$sum27 = (($109) + 36)|0; //@line 296 "load_s3m.cpp"
       $134 = (($lpStream) + ($$sum27)|0); //@line 296 "load_s3m.cpp"
       dest=$pSmp$sroa$15; src=$134; stop=dest+40|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 296 "load_s3m.cpp"
       $$sum28 = (($109) + 78)|0; //@line 296 "load_s3m.cpp"
       $135 = (($lpStream) + ($$sum28)|0); //@line 296 "load_s3m.cpp"
       $136 = HEAP8[$135>>0]|0; //@line 296 "load_s3m.cpp"
       $$sum29 = (($109) + 79)|0; //@line 296 "load_s3m.cpp"
       $137 = (($lpStream) + ($$sum29)|0); //@line 296 "load_s3m.cpp"
       $138 = HEAP8[$137>>0]|0; //@line 296 "load_s3m.cpp"
       $139 = (((((($this)) + 33280|0) + ($iSmp$066<<6)|0)) + 42|0); //@line 297 "load_s3m.cpp"
       _memmove(($139|0),($115|0),12)|0; //@line 297 "load_s3m.cpp"
       $140 = (($insflags) + ($105)|0); //@line 298 "load_s3m.cpp"
       HEAP8[$140>>0] = $131; //@line 298 "load_s3m.cpp"
       $141 = (($inspack) + ($105)|0); //@line 299 "load_s3m.cpp"
       HEAP8[$141>>0] = $129; //@line 299 "load_s3m.cpp"
       $142 = (((($this)) + 59804|0) + ($iSmp$066<<5)|0); //@line 300 "load_s3m.cpp"
       dest=$142; src=$100; stop=dest+28|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 300 "load_s3m.cpp"
       $143 = (((((($this)) + 59804|0) + ($iSmp$066<<5)|0)) + 28|0); //@line 301 "load_s3m.cpp"
       HEAP8[$143>>0] = 0; //@line 301 "load_s3m.cpp"
       $144 = ($114<<24>>24)==(1); //@line 302 "load_s3m.cpp"
       $145 = ($136<<24>>24)==(82); //@line 302 "load_s3m.cpp"
       $or$cond39 = $144 & $145; //@line 302 "load_s3m.cpp"
       $146 = ($138<<24>>24)==(83); //@line 302 "load_s3m.cpp"
       $or$cond40 = $or$cond39 & $146; //@line 302 "load_s3m.cpp"
       if ($or$cond40) {
        $147 = ($121>>>0)>(16000000); //@line 188 "load_s3m.cpp"
        $148 = ($121>>>0)<(4); //@line 189 "load_s3m.cpp"
        $$input$i = $148 ? 0 : $121; //@line 189 "load_s3m.cpp"
        $149 = $147 ? 16000000 : $$input$i; //@line 188 "load_s3m.cpp"
        $150 = (((($this)) + 33280|0) + ($iSmp$066<<6)|0); //@line 304 "load_s3m.cpp"
        HEAP32[$150>>2] = $149; //@line 304 "load_s3m.cpp"
        $151 = (($149) + -1)|0; //@line 305 "load_s3m.cpp"
        $152 = ($123>>>0)>($151>>>0); //@line 188 "load_s3m.cpp"
        $153 = ($123>>>0)<(4); //@line 189 "load_s3m.cpp"
        $$input$i50 = $153 ? 0 : $123; //@line 189 "load_s3m.cpp"
        $154 = $152 ? $151 : $$input$i50; //@line 188 "load_s3m.cpp"
        $155 = (((((($this)) + 33280|0) + ($iSmp$066<<6)|0)) + 4|0); //@line 305 "load_s3m.cpp"
        HEAP32[$155>>2] = $154; //@line 305 "load_s3m.cpp"
        $156 = ($125>>>0)>($149>>>0); //@line 188 "load_s3m.cpp"
        $157 = ($125>>>0)<(4); //@line 189 "load_s3m.cpp"
        $$input$i48 = $157 ? 0 : $125; //@line 189 "load_s3m.cpp"
        $158 = $156 ? $149 : $$input$i48; //@line 188 "load_s3m.cpp"
        $159 = (((((($this)) + 33280|0) + ($iSmp$066<<6)|0)) + 8|0); //@line 306 "load_s3m.cpp"
        HEAP32[$159>>2] = $158; //@line 306 "load_s3m.cpp"
        $160 = $127&255; //@line 307 "load_s3m.cpp"
        $161 = ($127&255)>(64); //@line 188 "load_s3m.cpp"
        $$op = $160 << 2; //@line 307 "load_s3m.cpp"
        $162 = $$op&65535; //@line 307 "load_s3m.cpp"
        $163 = $161 ? 256 : $162; //@line 307 "load_s3m.cpp"
        $164 = (((((($this)) + 33280|0) + ($iSmp$066<<6)|0)) + 30|0); //@line 307 "load_s3m.cpp"
        HEAP16[$164>>1] = $163; //@line 307 "load_s3m.cpp"
        $165 = (((((($this)) + 33280|0) + ($iSmp$066<<6)|0)) + 32|0); //@line 308 "load_s3m.cpp"
        HEAP16[$165>>1] = 64; //@line 308 "load_s3m.cpp"
        $166 = $131 & 1; //@line 309 "load_s3m.cpp"
        $167 = ($166<<24>>24)==(0); //@line 309 "load_s3m.cpp"
        if (!($167)) {
         $168 = (((((($this)) + 33280|0) + ($iSmp$066<<6)|0)) + 34|0); //@line 309 "load_s3m.cpp"
         $169 = HEAP16[$168>>1]|0; //@line 309 "load_s3m.cpp"
         $170 = $169&65535; //@line 309 "load_s3m.cpp"
         $171 = $170 | 2; //@line 309 "load_s3m.cpp"
         $172 = $171&65535; //@line 309 "load_s3m.cpp"
         HEAP16[$168>>1] = $172; //@line 309 "load_s3m.cpp"
        }
        $173 = ($133|0)!=(0); //@line 311 "load_s3m.cpp"
        $$3 = $173 ? $133 : 8363; //@line 311 "load_s3m.cpp"
        $174 = ($$3>>>0)<(1024); //@line 312 "load_s3m.cpp"
        $$$3 = $174 ? 1024 : $$3; //@line 312 "load_s3m.cpp"
        $175 = (((((($this)) + 33280|0) + ($iSmp$066<<6)|0)) + 24|0); //@line 313 "load_s3m.cpp"
        HEAP32[$175>>2] = $$$3; //@line 313 "load_s3m.cpp"
        $176 = $117&255; //@line 314 "load_s3m.cpp"
        $177 = $176 << 20; //@line 314 "load_s3m.cpp"
        $178 = $119&65535; //@line 314 "load_s3m.cpp"
        $179 = $178 << 4; //@line 314 "load_s3m.cpp"
        $180 = $179 | $177; //@line 314 "load_s3m.cpp"
        $181 = (($insfile) + ($iSmp$066<<2)|0); //@line 314 "load_s3m.cpp"
        HEAP32[$181>>2] = $180; //@line 314 "load_s3m.cpp"
        $182 = ($180>>>0)>($dwMemLength>>>0); //@line 316 "load_s3m.cpp"
        if ($182) {
         HEAP32[$181>>2] = 0; //@line 316 "load_s3m.cpp"
        } else {
         $183 = ($180|0)==(0); //@line 317 "load_s3m.cpp"
         if (!($183)) {
          $187 = $180;$z$064 = $105;
          while(1) {
           $184 = (($insfile) + ($z$064<<2)|0); //@line 320 "load_s3m.cpp"
           $185 = HEAP32[$184>>2]|0; //@line 320 "load_s3m.cpp"
           $186 = ($187|0)==($185|0); //@line 320 "load_s3m.cpp"
           if ($186) {
            HEAP32[$181>>2] = 0; //@line 321 "load_s3m.cpp"
            $321 = 0;
           } else {
            $321 = $187;
           }
           $188 = (($z$064) + -1)|0; //@line 319 "load_s3m.cpp"
           $189 = ($z$064|0)>(0); //@line 319 "load_s3m.cpp"
           if ($189) {
            $187 = $321;$z$064 = $188;
           } else {
            break;
           }
          }
         }
        }
        $190 = ($158>>>0)<=($154>>>0); //@line 323 "load_s3m.cpp"
        $191 = (($158) - ($154))|0;
        $192 = ($191>>>0)<(8); //@line 323 "load_s3m.cpp"
        $or$cond42 = $190 | $192; //@line 323 "load_s3m.cpp"
        if ($or$cond42) {
         HEAP32[$159>>2] = 0; //@line 324 "load_s3m.cpp"
         HEAP32[$155>>2] = 0; //@line 324 "load_s3m.cpp"
        }
        $193 = (((((($this)) + 33280|0) + ($iSmp$066<<6)|0)) + 28|0); //@line 325 "load_s3m.cpp"
        HEAP16[$193>>1] = 128; //@line 325 "load_s3m.cpp"
       }
      }
      $194 = (($iSmp$066) + 1)|0; //@line 291 "load_s3m.cpp"
      $195 = ($194>>>0)>($$2>>>0); //@line 291 "load_s3m.cpp"
      if ($195) {
       break;
      } else {
       $iSmp$066 = $194;
      }
     }
    }
    $101 = ($76<<16>>16)==(0); //@line 329 "load_s3m.cpp"
    if (!($101)) {
     $102 = (($dwMemLength) + -6)|0; //@line 336 "load_s3m.cpp"
     $103 = ($76&65535)<(240);
     $104 = $103 ? $77 : 240; //@line 329 "load_s3m.cpp"
     $iPat$062 = 0;
     while(1) {
      $199 = (($iPat$062) + ($73))|0; //@line 331 "load_s3m.cpp"
      $200 = (($ptr) + ($199<<1)|0); //@line 331 "load_s3m.cpp"
      $201 = HEAP16[$200>>1]|0; //@line 331 "load_s3m.cpp"
      $202 = $201&65535; //@line 331 "load_s3m.cpp"
      $203 = $202 << 4; //@line 331 "load_s3m.cpp"
      $204 = (($203) + 64)|0; //@line 332 "load_s3m.cpp"
      $205 = ($204>>>0)>($dwMemLength>>>0); //@line 332 "load_s3m.cpp"
      L64: do {
       if (!($205)) {
        $206 = (($lpStream) + ($203)|0); //@line 333 "load_s3m.cpp"
        $207 = HEAP16[$206>>1]|0; //@line 333 "load_s3m.cpp"
        $208 = $203 | 2; //@line 334 "load_s3m.cpp"
        $209 = (((($this)) + 52864|0) + ($iPat$062<<1)|0); //@line 335 "load_s3m.cpp"
        HEAP16[$209>>1] = 64; //@line 335 "load_s3m.cpp"
        $210 = ($207<<16>>16)==(0); //@line 336 "load_s3m.cpp"
        if (!($210)) {
         $211 = $207&65535; //@line 336 "load_s3m.cpp"
         $212 = (($211) + ($208))|0; //@line 336 "load_s3m.cpp"
         $213 = ($212>>>0)>($102>>>0); //@line 336 "load_s3m.cpp"
         if (!($213)) {
          $214 = HEAP32[$37>>2]|0; //@line 337 "load_s3m.cpp"
          $215 = (__ZN10CSoundFile15AllocatePatternEjj(64,$214)|0); //@line 337 "load_s3m.cpp"
          $216 = (((($this)) + 51904|0) + ($iPat$062<<2)|0); //@line 337 "load_s3m.cpp"
          HEAP32[$216>>2] = $215; //@line 337 "load_s3m.cpp"
          $217 = ($215|0)==(0|0); //@line 337 "load_s3m.cpp"
          if (!($217)) {
           $218 = $203 | 3; //@line 378 "load_s3m.cpp"
           $219 = (($208) + 2)|0; //@line 357 "load_s3m.cpp"
           $j3$0$ph = 0;$row$0$ph = 0;
           while(1) {
            $j3$0 = $j3$0$ph;
            while(1) {
             $220 = ($j3$0>>>0)<($211>>>0); //@line 343 "load_s3m.cpp"
             if (!($220)) {
              break L64;
             }
             $221 = (($j3$0) + 1)|0; //@line 345 "load_s3m.cpp"
             $$sum = (($j3$0) + ($208))|0; //@line 345 "load_s3m.cpp"
             $222 = (($lpStream) + ($$sum)|0); //@line 345 "load_s3m.cpp"
             $223 = HEAP8[$222>>0]|0; //@line 345 "load_s3m.cpp"
             $224 = ($223<<24>>24)==(0); //@line 346 "load_s3m.cpp"
             if ($224) {
              $$lcssa = $221;
              break;
             }
             $227 = $223&255; //@line 351 "load_s3m.cpp"
             $228 = $227 & 31; //@line 351 "load_s3m.cpp"
             $229 = HEAP32[$37>>2]|0; //@line 352 "load_s3m.cpp"
             $230 = ($228>>>0)<($229>>>0); //@line 352 "load_s3m.cpp"
             do {
              if ($230) {
               $231 = Math_imul($229, $row$0$ph)|0; //@line 354 "load_s3m.cpp"
               $232 = (($231) + ($228))|0; //@line 354 "load_s3m.cpp"
               $233 = $227 & 32; //@line 355 "load_s3m.cpp"
               $234 = ($233|0)==(0); //@line 355 "load_s3m.cpp"
               if ($234) {
                $j3$1 = $221;
               } else {
                $$sum12 = (($221) + ($208))|0; //@line 357 "load_s3m.cpp"
                $235 = (($lpStream) + ($$sum12)|0); //@line 357 "load_s3m.cpp"
                $236 = HEAP8[$235>>0]|0; //@line 357 "load_s3m.cpp"
                $237 = (($215) + (($232*6)|0)|0); //@line 357 "load_s3m.cpp"
                HEAP8[$237>>0] = $236; //@line 357 "load_s3m.cpp"
                $238 = $236&255; //@line 358 "load_s3m.cpp"
                $239 = ($236&255)<(240); //@line 358 "load_s3m.cpp"
                do {
                 if ($239) {
                  $240 = $238 & 15; //@line 358 "load_s3m.cpp"
                  $241 = $238 >>> 4; //@line 358 "load_s3m.cpp"
                  $242 = ($241*12)|0; //@line 358 "load_s3m.cpp"
                  $243 = (($240) + 13)|0; //@line 358 "load_s3m.cpp"
                  $244 = (($243) + ($242))|0; //@line 358 "load_s3m.cpp"
                  $245 = $244&255; //@line 358 "load_s3m.cpp"
                  HEAP8[$237>>0] = $245; //@line 358 "load_s3m.cpp"
                 } else {
                  $246 = ($236<<24>>24)==(-1); //@line 359 "load_s3m.cpp"
                  if (!($246)) {
                   break;
                  }
                  HEAP8[$237>>0] = 0; //@line 359 "load_s3m.cpp"
                 }
                } while(0);
                $247 = (($j3$0) + 3)|0; //@line 360 "load_s3m.cpp"
                $$sum13 = (($219) + ($j3$0))|0; //@line 360 "load_s3m.cpp"
                $248 = (($lpStream) + ($$sum13)|0); //@line 360 "load_s3m.cpp"
                $249 = HEAP8[$248>>0]|0; //@line 360 "load_s3m.cpp"
                $250 = (((($215) + (($232*6)|0)|0)) + 1|0); //@line 360 "load_s3m.cpp"
                HEAP8[$250>>0] = $249; //@line 360 "load_s3m.cpp"
                $j3$1 = $247;
               }
               $251 = $227 & 64; //@line 362 "load_s3m.cpp"
               $252 = ($251|0)==(0); //@line 362 "load_s3m.cpp"
               if ($252) {
                $j3$2 = $j3$1;
               } else {
                $253 = (($j3$1) + 1)|0; //@line 364 "load_s3m.cpp"
                $$sum11 = (($j3$1) + ($208))|0; //@line 364 "load_s3m.cpp"
                $254 = (($lpStream) + ($$sum11)|0); //@line 364 "load_s3m.cpp"
                $255 = HEAP8[$254>>0]|0; //@line 364 "load_s3m.cpp"
                $256 = $255&255; //@line 364 "load_s3m.cpp"
                $257 = ($255<<24>>24)<(0); //@line 365 "load_s3m.cpp"
                $258 = ($255&255)<(193); //@line 365 "load_s3m.cpp"
                $or$cond5 = $257 & $258; //@line 365 "load_s3m.cpp"
                $259 = ($255&255)>(64); //@line 371 "load_s3m.cpp"
                $$43 = $259 ? 64 : $256; //@line 371 "load_s3m.cpp"
                $260 = (($256) + 128)|0; //@line 367 "load_s3m.cpp"
                $$sink = $or$cond5 ? 2 : 1; //@line 365 "load_s3m.cpp"
                $vol$1 = $or$cond5 ? $260 : $$43; //@line 365 "load_s3m.cpp"
                $261 = (((($215) + (($232*6)|0)|0)) + 2|0); //@line 372 "load_s3m.cpp"
                HEAP8[$261>>0] = $$sink; //@line 372 "load_s3m.cpp"
                $262 = $vol$1&255; //@line 374 "load_s3m.cpp"
                $263 = (((($215) + (($232*6)|0)|0)) + 4|0); //@line 374 "load_s3m.cpp"
                HEAP8[$263>>0] = $262; //@line 374 "load_s3m.cpp"
                $j3$2 = $253;
               }
               $264 = $227 & 128; //@line 376 "load_s3m.cpp"
               $265 = ($264|0)==(0); //@line 376 "load_s3m.cpp"
               if ($265) {
                $j3$5 = $j3$2;
                break;
               }
               $$sum9 = (($j3$2) + ($208))|0; //@line 378 "load_s3m.cpp"
               $266 = (($lpStream) + ($$sum9)|0); //@line 378 "load_s3m.cpp"
               $267 = HEAP8[$266>>0]|0; //@line 378 "load_s3m.cpp"
               $268 = (((($215) + (($232*6)|0)|0)) + 3|0); //@line 378 "load_s3m.cpp"
               HEAP8[$268>>0] = $267; //@line 378 "load_s3m.cpp"
               $269 = (($j3$2) + 2)|0; //@line 379 "load_s3m.cpp"
               $$sum10 = (($218) + ($j3$2))|0; //@line 379 "load_s3m.cpp"
               $270 = (($lpStream) + ($$sum10)|0); //@line 379 "load_s3m.cpp"
               $271 = HEAP8[$270>>0]|0; //@line 379 "load_s3m.cpp"
               $272 = (((($215) + (($232*6)|0)|0)) + 5|0); //@line 379 "load_s3m.cpp"
               HEAP8[$272>>0] = $271; //@line 379 "load_s3m.cpp"
               $273 = ($267<<24>>24)==(0); //@line 380 "load_s3m.cpp"
               if ($273) {
                $j3$5 = $269;
                break;
               }
               $274 = $267&255; //@line 71 "load_s3m.cpp"
               $275 = $271&255; //@line 72 "load_s3m.cpp"
               do {
                switch ($274|0) {
                case 1:  {
                 $command$0$i = 16;$param$0$i = $275;
                 break;
                }
                case 2:  {
                 $command$0$i = 12;$param$0$i = $275;
                 break;
                }
                case 3:  {
                 $276 = $275 >>> 4; //@line 77 "load_s3m.cpp"
                 $277 = ($276*10)|0; //@line 77 "load_s3m.cpp"
                 $278 = $275 & 15; //@line 77 "load_s3m.cpp"
                 $279 = (($277) + ($278))|0; //@line 77 "load_s3m.cpp"
                 $command$0$i = 14;$param$0$i = $279;
                 break;
                }
                case 4:  {
                 $command$0$i = 11;$param$0$i = $275;
                 break;
                }
                case 5:  {
                 $command$0$i = 3;$param$0$i = $275;
                 break;
                }
                case 6:  {
                 $command$0$i = 2;$param$0$i = $275;
                 break;
                }
                case 7:  {
                 $command$0$i = 4;$param$0$i = $275;
                 break;
                }
                case 8:  {
                 $command$0$i = 5;$param$0$i = $275;
                 break;
                }
                case 9:  {
                 $command$0$i = 18;$param$0$i = $275;
                 break;
                }
                case 10:  {
                 $command$0$i = 1;$param$0$i = $275;
                 break;
                }
                case 11:  {
                 $command$0$i = 7;$param$0$i = $275;
                 break;
                }
                case 12:  {
                 $command$0$i = 6;$param$0$i = $275;
                 break;
                }
                case 13:  {
                 $command$0$i = 21;$param$0$i = $275;
                 break;
                }
                case 14:  {
                 $command$0$i = 22;$param$0$i = $275;
                 break;
                }
                case 15:  {
                 $command$0$i = 10;$param$0$i = $275;
                 break;
                }
                case 16:  {
                 $command$0$i = 29;$param$0$i = $275;
                 break;
                }
                case 17:  {
                 $command$0$i = 15;$param$0$i = $275;
                 break;
                }
                case 18:  {
                 $command$0$i = 8;$param$0$i = $275;
                 break;
                }
                case 19:  {
                 $command$0$i = 20;$param$0$i = $275;
                 break;
                }
                case 20:  {
                 $command$0$i = 17;$param$0$i = $275;
                 break;
                }
                case 21:  {
                 $command$0$i = 26;$param$0$i = $275;
                 break;
                }
                case 22:  {
                 $command$0$i = 23;$param$0$i = $275;
                 break;
                }
                case 23:  {
                 $command$0$i = 24;$param$0$i = $275;
                 break;
                }
                case 24:  {
                 $command$0$i = 9;$param$0$i = $275;
                 break;
                }
                case 25:  {
                 $command$0$i = 27;$param$0$i = $275;
                 break;
                }
                case 26:  {
                 $command$0$i = 31;$param$0$i = $275;
                 break;
                }
                default: {
                 $command$0$i = 0;$param$0$i = $275;
                }
                }
               } while(0);
               HEAP8[$268>>0] = $command$0$i; //@line 103 "load_s3m.cpp"
               $280 = $param$0$i&255; //@line 104 "load_s3m.cpp"
               HEAP8[$272>>0] = $280; //@line 104 "load_s3m.cpp"
               $j3$5 = $269;
              } else {
               $281 = $227 & 32; //@line 384 "load_s3m.cpp"
               $282 = ($281|0)==(0); //@line 384 "load_s3m.cpp"
               $283 = (($j3$0) + 3)|0; //@line 384 "load_s3m.cpp"
               $$44 = $282 ? $221 : $283; //@line 384 "load_s3m.cpp"
               $284 = $227 >>> 6; //@line 385 "load_s3m.cpp"
               $$lobit = $284 & 1; //@line 385 "load_s3m.cpp"
               $j3$4 = (($$44) + ($$lobit))|0; //@line 385 "load_s3m.cpp"
               $285 = $227 & 128; //@line 386 "load_s3m.cpp"
               $286 = ($285|0)==(0); //@line 386 "load_s3m.cpp"
               $287 = (($j3$4) + 2)|0; //@line 386 "load_s3m.cpp"
               $j3$4$ = $286 ? $j3$4 : $287; //@line 386 "load_s3m.cpp"
               $j3$5 = $j3$4$;
              }
             } while(0);
             $288 = ($j3$5>>>0)<($211>>>0); //@line 388 "load_s3m.cpp"
             if ($288) {
              $j3$0 = $j3$5;
             } else {
              break L64;
             }
            }
            $225 = (($row$0$ph) + 1)|0; //@line 348 "load_s3m.cpp"
            $226 = ($225>>>0)>(63); //@line 348 "load_s3m.cpp"
            if ($226) {
             break;
            } else {
             $j3$0$ph = $$lcssa;$row$0$ph = $225;
            }
           }
          }
         }
        }
       }
      } while(0);
      $289 = (($iPat$062) + 1)|0; //@line 329 "load_s3m.cpp"
      $exitcond = ($289|0)==($104|0); //@line 329 "load_s3m.cpp"
      if ($exitcond) {
       break;
      } else {
       $iPat$062 = $289;
      }
     }
    }
    if (!($99)) {
     $196 = HEAPU8[$7>>0]|(HEAPU8[$7+1>>0]<<8); //@line 395 "load_s3m.cpp"
     $not$ = ($196<<16>>16)!=(1); //@line 395 "load_s3m.cpp"
     $197 = $not$&1; //@line 395 "load_s3m.cpp"
     $198 = $not$ ? 6 : 5; //@line 396 "load_s3m.cpp"
     $iRaw$061 = 1;
     while(1) {
      $290 = (((($this)) + 33280|0) + ($iRaw$061<<6)|0); //@line 393 "load_s3m.cpp"
      $291 = HEAP32[$290>>2]|0; //@line 393 "load_s3m.cpp"
      $292 = ($291|0)==(0); //@line 393 "load_s3m.cpp"
      if (!($292)) {
       $293 = (($insfile) + ($iRaw$061<<2)|0); //@line 393 "load_s3m.cpp"
       $294 = HEAP32[$293>>2]|0; //@line 393 "load_s3m.cpp"
       $295 = ($294|0)==(0); //@line 393 "load_s3m.cpp"
       if (!($295)) {
        $296 = (($iRaw$061) + -1)|0; //@line 396 "load_s3m.cpp"
        $297 = (($insflags) + ($296)|0); //@line 396 "load_s3m.cpp"
        $298 = HEAP8[$297>>0]|0; //@line 396 "load_s3m.cpp"
        $299 = ($294>>>0)<($dwMemLength>>>0); //@line 400 "load_s3m.cpp"
        if ($299) {
         $300 = (($inspack) + ($296)|0); //@line 398 "load_s3m.cpp"
         $301 = HEAP8[$300>>0]|0; //@line 398 "load_s3m.cpp"
         $302 = ($301<<24>>24)==(4); //@line 398 "load_s3m.cpp"
         $303 = $298 & 2; //@line 397 "load_s3m.cpp"
         $304 = $303&255; //@line 397 "load_s3m.cpp"
         $305 = $304 << 2; //@line 397 "load_s3m.cpp"
         $306 = $298 & 4; //@line 396 "load_s3m.cpp"
         $307 = ($306<<24>>24)==(0); //@line 396 "load_s3m.cpp"
         $$45 = $307 ? $197 : $198; //@line 396 "load_s3m.cpp"
         $308 = $305 | $$45; //@line 397 "load_s3m.cpp"
         $$flags$1 = $302 ? 3 : $308; //@line 398 "load_s3m.cpp"
         $309 = (($lpStream) + ($294)|0); //@line 401 "load_s3m.cpp"
         $310 = (($dwMemLength) - ($294))|0; //@line 401 "load_s3m.cpp"
         (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$290,$$flags$1,$309,$310)|0); //@line 401 "load_s3m.cpp"
        }
       }
      }
      $311 = (($iRaw$061) + 1)|0; //@line 393 "load_s3m.cpp"
      $312 = ($311>>>0)>($$2>>>0); //@line 393 "load_s3m.cpp"
      if ($312) {
       break;
      } else {
       $iRaw$061 = $311;
      }
     }
    }
    $313 = ((($this)) + 59764|0); //@line 403 "load_s3m.cpp"
    HEAP32[$313>>2] = 64; //@line 403 "load_s3m.cpp"
    $314 = ((($this)) + 59768|0); //@line 404 "load_s3m.cpp"
    HEAP32[$314>>2] = 32767; //@line 404 "load_s3m.cpp"
    $315 = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8); //@line 405 "load_s3m.cpp"
    $316 = $315 & 16; //@line 405 "load_s3m.cpp"
    $317 = ($316<<16>>16)==(0); //@line 405 "load_s3m.cpp"
    if ($317) {
     $$0 = 1;
    } else {
     $318 = ((($this)) + 59660|0); //@line 405 "load_s3m.cpp"
     $319 = HEAP32[$318>>2]|0; //@line 405 "load_s3m.cpp"
     $320 = $319 | 65536; //@line 405 "load_s3m.cpp"
     HEAP32[$318>>2] = $320; //@line 405 "load_s3m.cpp"
     $$0 = 1;
    }
   }
  }
 } else {
  $$0 = 0;
 }
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0); //@line 407 "load_s3m.cpp"
}
function __ZN10CSoundFile6ReadITEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$0 = 0, $$1 = 0, $$100 = 0, $$73 = 0, $$84 = 0, $$85 = 0, $$86 = 0, $$89 = 0, $$97 = 0, $$98 = 0, $$lcssa = 0, $$lcssa195 = 0, $$lobit = 0, $$lobit101$us = 0, $$lobit102$us = 0, $$not = 0, $$off = 0, $$off34 = 0, $$off35 = 0;
 var $$off64 = 0, $$off71 = 0, $$op = 0, $$op27 = 0, $$op28 = 0, $$phi$trans$insert = 0, $$phi$trans$insert152 = 0, $$phi$trans$insert163 = 0, $$pr$pre = 0, $$pre = 0, $$pre$phi173Z2D = 0, $$pre$phi175Z2D = 0, $$pre$phi177Z2D = 0, $$pre$phiZ2D = 0, $$pre150 = 0, $$pre153 = 0, $$pre154 = 0, $$pre155 = 0, $$pre157 = 0, $$pre158 = 0;
 var $$pre159 = 0, $$pre162 = 0, $$pre164 = 0, $$pre165 = 0, $$pre172 = 0, $$pre174 = 0, $$pre176 = 0, $$sum = 0, $$sum29 = 0, $$sum30 = 0, $$sum31 = 0, $$sum32 = 0, $$sum33 = 0, $$sum36 = 0, $$sum37 = 0, $$sum38 = 0, $$sum39 = 0, $$sum40 = 0, $$sum41 = 0, $$sum42 = 0;
 var $$sum43 = 0, $$sum44 = 0, $$sum45 = 0, $$sum46 = 0, $$sum47 = 0, $$sum48 = 0, $$sum49 = 0, $$sum50 = 0, $$sum51 = 0, $$sum52 = 0, $$sum53 = 0, $$sum54 = 0, $$sum55 = 0, $$sum56 = 0, $$sum63 = 0, $$sum65 = 0, $$sum66$us = 0, $$sum67$us = 0, $$sum68 = 0, $$sum69 = 0;
 var $$sum70 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0;
 var $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0;
 var $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0;
 var $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0;
 var $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0;
 var $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0;
 var $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0;
 var $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0;
 var $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0;
 var $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0;
 var $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0;
 var $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0;
 var $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0;
 var $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0;
 var $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0;
 var $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0;
 var $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0;
 var $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0;
 var $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0;
 var $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0;
 var $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0;
 var $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0;
 var $62 = 0, $620 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $ch$0$us = 0, $ch13$0 = 0, $chnmask = 0, $dwMemPos$0 = 0, $dwMemPos$1 = 0, $dwMemPos$2 = 0, $dwMemPos$3 = 0, $exitcond = 0, $exitcond142 = 0, $exitcond144 = 0, $exitcond148 = 0, $flags$2 = 0, $i$0131 = 0, $i6$0$ph189 = 0, $i6$0$us = 0, $i6$1$$us = 0, $i6$1$us = 0, $i6$3$$us = 0;
 var $i6$3$us = 0, $i6$5$us = 0, $i9$0$be = 0, $i9$0$ph121 = 0, $i9$0113 = 0, $i9$1 = 0, $i9$2 = 0, $i9$3 = 0, $i9$4 = 0, $inspos = 0, $ipan$0133 = 0, $lastvalue = 0, $m$0$ph120 = 0, $ncu$0111 = 0, $nins$0127 = 0, $not$ = 0, $notlhs = 0, $notrhs = 0, $npat$0123 = 0, $nrow$0$ph188 = 0;
 var $nrow11$0$ph122 = 0, $nsmp$0125 = 0, $or$cond = 0, $or$cond18 = 0, $or$cond25 = 0, $or$cond75 = 0, $or$cond77 = 0, $or$cond80 = 0, $or$cond87$not = 0, $or$cond88 = 0, $or$cond9 = 0, $or$cond93 = 0, $or$cond95 = 0, $or$cond96112 = 0, $or$cond96112119 = 0, $patchk$0129 = 0, $patpos = 0, $pifh = 0, $pis$sroa$18 = 0, $pis$sroa$6 = 0;
 var $smppos = 0, $storemerge59 = 0, $storemerge61 = 0, $tmp = 0, $tmp103 = 0, $tmp104 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 3568|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $inspos = sp + 2112|0;
 $smppos = sp + 1152|0;
 $patpos = sp + 192|0;
 $chnmask = sp + 3496|0;
 $lastvalue = sp + 3112|0;
 $pifh = sp;
 $pis$sroa$6 = sp + 3100|0;
 $pis$sroa$18 = sp + 3072|0;
 $0 = ($lpStream|0)==(0|0); //@line 169 "load_it.cpp"
 $1 = ($dwMemLength>>>0)<(192); //@line 169 "load_it.cpp"
 $or$cond = $0 | $1; //@line 169 "load_it.cpp"
 if ($or$cond) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 591 "load_it.cpp"
 }
 _memcpy(($pifh|0),($lpStream|0),192)|0; //@line 170 "load_it.cpp"
 $2 = ((($pifh)) + 32|0); //@line 174 "load_it.cpp"
 $3 = ((($pifh)) + 36|0); //@line 176 "load_it.cpp"
 $4 = ((($pifh)) + 42|0); //@line 179 "load_it.cpp"
 $5 = ((($pifh)) + 44|0); //@line 180 "load_it.cpp"
 $6 = ((($pifh)) + 54|0); //@line 182 "load_it.cpp"
 $7 = ((($pifh)) + 56|0); //@line 183 "load_it.cpp"
 $8 = HEAP32[$pifh>>2]|0; //@line 186 "load_it.cpp"
 $9 = ($8|0)==(1297108297); //@line 186 "load_it.cpp"
 if ($9) {
  $10 = ((($pifh)) + 34|0); //@line 175 "load_it.cpp"
  $11 = HEAP16[$10>>1]|0; //@line 186 "load_it.cpp"
  $12 = $11&65535; //@line 186 "load_it.cpp"
  $13 = ($11&65535)>(239); //@line 186 "load_it.cpp"
  if ($13) {
   $$0 = 0;
  } else {
   $14 = HEAP32[$3>>2]|0; //@line 187 "load_it.cpp"
   $15 = $14&65535; //@line 187 "load_it.cpp"
   $16 = ($15<<16>>16)==(0); //@line 187 "load_it.cpp"
   $17 = $14 >>> 16;
   $18 = $17&65535;
   if ($16) {
    $$0 = 0;
   } else {
    $19 = $14 & 65535; //@line 187 "load_it.cpp"
    $20 = ($15&65535)>(239); //@line 187 "load_it.cpp"
    if ($20) {
     $$0 = 0;
    } else {
     $21 = HEAP32[$2>>2]|0; //@line 187 "load_it.cpp"
     $22 = $21&65535; //@line 187 "load_it.cpp"
     $23 = ($22<<16>>16)==(0); //@line 187 "load_it.cpp"
     $24 = $21 >>> 16;
     $25 = $24&65535;
     if ($23) {
      $$0 = 0;
     } else {
      $26 = $21 & 65535; //@line 188 "load_it.cpp"
      $tmp = (($19) + ($12))|0;
      $tmp103 = (($tmp) + ($17))|0;
      $tmp104 = $tmp103 << 2;
      $27 = (($26) + 192)|0; //@line 189 "load_it.cpp"
      $28 = (($27) + ($tmp104))|0; //@line 189 "load_it.cpp"
      $29 = ($28>>>0)>($dwMemLength>>>0); //@line 189 "load_it.cpp"
      if ($29) {
       $$0 = 0;
      } else {
       $30 = ((($this)) + 59680|0); //@line 190 "load_it.cpp"
       HEAP32[$30>>2] = 32; //@line 190 "load_it.cpp"
       $31 = HEAP16[$5>>1]|0; //@line 191 "load_it.cpp"
       $32 = HEAP32[$5>>2]|0; //@line 191 "load_it.cpp"
       $33 = $32&65535; //@line 191 "load_it.cpp"
       $34 = $33 & 8; //@line 191 "load_it.cpp"
       $35 = ($34<<16>>16)==(0); //@line 191 "load_it.cpp"
       if (!($35)) {
        $36 = ((($this)) + 59660|0); //@line 191 "load_it.cpp"
        $37 = HEAP32[$36>>2]|0; //@line 191 "load_it.cpp"
        $38 = $37 | 16; //@line 191 "load_it.cpp"
        HEAP32[$36>>2] = $38; //@line 191 "load_it.cpp"
       }
       $39 = $33 & 16; //@line 192 "load_it.cpp"
       $40 = ($39<<16>>16)==(0); //@line 192 "load_it.cpp"
       if (!($40)) {
        $41 = ((($this)) + 59660|0); //@line 192 "load_it.cpp"
        $42 = HEAP32[$41>>2]|0; //@line 192 "load_it.cpp"
        $43 = $42 | 4; //@line 192 "load_it.cpp"
        HEAP32[$41>>2] = $43; //@line 192 "load_it.cpp"
       }
       $44 = $33 & 32; //@line 193 "load_it.cpp"
       $45 = ($44<<16>>16)==(0); //@line 193 "load_it.cpp"
       if (!($45)) {
        $46 = ((($this)) + 59660|0); //@line 193 "load_it.cpp"
        $47 = HEAP32[$46>>2]|0; //@line 193 "load_it.cpp"
        $48 = $47 | 8; //@line 193 "load_it.cpp"
        HEAP32[$46>>2] = $48; //@line 193 "load_it.cpp"
       }
       $49 = $33 & 128; //@line 194 "load_it.cpp"
       $50 = ($49<<16>>16)==(0); //@line 194 "load_it.cpp"
       if (!($50)) {
        $51 = ((($this)) + 59660|0); //@line 194 "load_it.cpp"
        $52 = HEAP32[$51>>2]|0; //@line 194 "load_it.cpp"
        $53 = $52 | 1; //@line 194 "load_it.cpp"
        HEAP32[$51>>2] = $53; //@line 194 "load_it.cpp"
       }
       $54 = $33 & 4096; //@line 195 "load_it.cpp"
       $55 = ($54<<16>>16)==(0); //@line 195 "load_it.cpp"
       if (!($55)) {
        $56 = ((($this)) + 59660|0); //@line 195 "load_it.cpp"
        $57 = HEAP32[$56>>2]|0; //@line 195 "load_it.cpp"
        $58 = $57 | 32768; //@line 195 "load_it.cpp"
        HEAP32[$56>>2] = $58; //@line 195 "load_it.cpp"
       }
       $59 = ((($this)) + 59804|0); //@line 196 "load_it.cpp"
       $60 = ((($pifh)) + 4|0); //@line 196 "load_it.cpp"
       dest=$59; src=$60; stop=dest+26|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 196 "load_it.cpp"
       $61 = ((($this)) + 59830|0); //@line 197 "load_it.cpp"
       HEAP8[$61>>0] = 0; //@line 197 "load_it.cpp"
       $62 = ((($pifh)) + 48|0); //@line 199 "load_it.cpp"
       $63 = HEAP16[$62>>1]|0; //@line 199 "load_it.cpp"
       $64 = HEAP32[$62>>2]|0; //@line 199 "load_it.cpp"
       $65 = $64&255; //@line 199 "load_it.cpp"
       $66 = ($65<<24>>24)==(0); //@line 199 "load_it.cpp"
       $67 = $64 >>> 16;
       $68 = $67&255;
       $69 = $64 >>> 24;
       $70 = ($63&65535) >>> 8;
       $71 = $70&65535;
       if (!($66)) {
        $72 = $64 << 1; //@line 201 "load_it.cpp"
        $73 = $72 & 510; //@line 201 "load_it.cpp"
        $74 = ((($this)) + 59656|0); //@line 201 "load_it.cpp"
        $75 = ($73>>>0)>(256); //@line 203 "load_it.cpp"
        $$98 = $75 ? 256 : $73; //@line 203 "load_it.cpp"
        HEAP32[$74>>2] = $$98; //@line 203 "load_it.cpp"
       }
       $76 = ($68<<24>>24)==(0); //@line 205 "load_it.cpp"
       if (!($76)) {
        $77 = $67 & 255; //@line 205 "load_it.cpp"
        $78 = ((($this)) + 59648|0); //@line 205 "load_it.cpp"
        HEAP32[$78>>2] = $77; //@line 205 "load_it.cpp"
       }
       $79 = ($69|0)==(0); //@line 206 "load_it.cpp"
       if (!($79)) {
        $80 = ((($this)) + 59652|0); //@line 206 "load_it.cpp"
        HEAP32[$80>>2] = $69; //@line 206 "load_it.cpp"
       }
       $81 = $71 & 127; //@line 207 "load_it.cpp"
       $82 = ((($this)) + 59748|0); //@line 207 "load_it.cpp"
       HEAP32[$82>>2] = $81; //@line 207 "load_it.cpp"
       $ipan$0133 = 0;
       while(1) {
        $83 = (((($pifh)) + 64|0) + ($ipan$0133)|0); //@line 209 "load_it.cpp"
        $84 = HEAP8[$83>>0]|0; //@line 209 "load_it.cpp"
        $85 = ($84<<24>>24)==(-1); //@line 209 "load_it.cpp"
        do {
         if (!($85)) {
          $86 = (((($pifh)) + 128|0) + ($ipan$0133)|0); //@line 211 "load_it.cpp"
          $87 = HEAP8[$86>>0]|0; //@line 211 "load_it.cpp"
          $88 = $87&255; //@line 211 "load_it.cpp"
          $89 = (((((($this)) + 49600|0) + (($ipan$0133*36)|0)|0)) + 4|0); //@line 211 "load_it.cpp"
          HEAP32[$89>>2] = $88; //@line 211 "load_it.cpp"
          $90 = (((($this)) + 49600|0) + (($ipan$0133*36)|0)|0); //@line 212 "load_it.cpp"
          HEAP32[$90>>2] = 128; //@line 212 "load_it.cpp"
          $91 = ($84<<24>>24)<(0); //@line 213 "load_it.cpp"
          if ($91) {
           $92 = (((((($this)) + 49600|0) + (($ipan$0133*36)|0)|0)) + 8|0); //@line 213 "load_it.cpp"
           $93 = HEAP32[$92>>2]|0; //@line 213 "load_it.cpp"
           $94 = $93 | 256; //@line 213 "load_it.cpp"
           HEAP32[$92>>2] = $94; //@line 213 "load_it.cpp"
          }
          $95 = $84&255; //@line 214 "load_it.cpp"
          $96 = $95 & 127; //@line 214 "load_it.cpp"
          $97 = ($96>>>0)<(65); //@line 215 "load_it.cpp"
          if ($97) {
           $98 = $96 << 2; //@line 215 "load_it.cpp"
           HEAP32[$90>>2] = $98; //@line 215 "load_it.cpp"
           break;
          }
          $99 = ($96|0)==(100); //@line 216 "load_it.cpp"
          if (!($99)) {
           break;
          }
          $100 = (((((($this)) + 49600|0) + (($ipan$0133*36)|0)|0)) + 8|0); //@line 216 "load_it.cpp"
          $101 = HEAP32[$100>>2]|0; //@line 216 "load_it.cpp"
          $102 = $101 | 2048; //@line 216 "load_it.cpp"
          HEAP32[$100>>2] = $102; //@line 216 "load_it.cpp"
         }
        } while(0);
        $103 = (($ipan$0133) + 1)|0; //@line 209 "load_it.cpp"
        $exitcond148 = ($103|0)==(64); //@line 209 "load_it.cpp"
        if ($exitcond148) {
         break;
        } else {
         $ipan$0133 = $103;
        }
       }
       $104 = ((($this)) + 59664|0); //@line 218 "load_it.cpp"
       $105 = HEAP32[$104>>2]|0; //@line 218 "load_it.cpp"
       $106 = ($105>>>0)<(4); //@line 218 "load_it.cpp"
       if ($106) {
        HEAP32[$104>>2] = 4; //@line 218 "load_it.cpp"
       }
       $107 = $32 & 65536; //@line 220 "load_it.cpp"
       $108 = ($107|0)==(0); //@line 220 "load_it.cpp"
       do {
        if (!($108)) {
         $109 = HEAP16[$6>>1]|0; //@line 220 "load_it.cpp"
         $110 = ($109<<16>>16)==(0); //@line 220 "load_it.cpp"
         if (!($110)) {
          $111 = $109&65535; //@line 220 "load_it.cpp"
          $112 = ($111>>>0)>($dwMemLength>>>0); //@line 220 "load_it.cpp"
          if ($112) {
           break;
          }
          $113 = HEAP32[$7>>2]|0; //@line 220 "load_it.cpp"
          $114 = (($dwMemLength) - ($111))|0; //@line 220 "load_it.cpp"
          $115 = ($113>>>0)<($114>>>0); //@line 220 "load_it.cpp"
          if (!($115)) {
           break;
          }
          $116 = (($111) + 1)|0; //@line 222 "load_it.cpp"
          $117 = (__Znaj($116)|0); //@line 222 "load_it.cpp"
          $118 = ((($this)) + 59796|0); //@line 222 "load_it.cpp"
          HEAP32[$118>>2] = $117; //@line 222 "load_it.cpp"
          $119 = (($lpStream) + ($113)|0); //@line 225 "load_it.cpp"
          _memcpy(($117|0),($119|0),($111|0))|0; //@line 225 "load_it.cpp"
          $120 = (($117) + ($111)|0); //@line 226 "load_it.cpp"
          HEAP8[$120>>0] = 0; //@line 226 "load_it.cpp"
         }
        }
       } while(0);
       $121 = ($26>>>0)>(256); //@line 231 "load_it.cpp"
       $$ = $121 ? 256 : $26; //@line 231 "load_it.cpp"
       $122 = ((($this)) + 53344|0); //@line 232 "load_it.cpp"
       $123 = ((($lpStream)) + 192|0); //@line 232 "load_it.cpp"
       _memcpy(($122|0),($123|0),($$|0))|0; //@line 232 "load_it.cpp"
       _memset(($inspos|0),0,960)|0; //@line 235 "load_it.cpp"
       $124 = ($25&65535)>(240); //@line 237 "load_it.cpp"
       $$op = $24 << 2; //@line 238 "load_it.cpp"
       $125 = $124 ? 960 : $$op; //@line 238 "load_it.cpp"
       $126 = (($lpStream) + ($27)|0); //@line 239 "load_it.cpp"
       _memcpy(($inspos|0),($126|0),($125|0))|0; //@line 239 "load_it.cpp"
       $127 = (($$op) + ($27))|0; //@line 244 "load_it.cpp"
       _memset(($smppos|0),0,960)|0; //@line 246 "load_it.cpp"
       $128 = ($15&65535)>(240); //@line 248 "load_it.cpp"
       $$op27 = $19 << 2; //@line 249 "load_it.cpp"
       $129 = $128 ? 960 : $$op27; //@line 249 "load_it.cpp"
       $130 = (($lpStream) + ($127)|0); //@line 250 "load_it.cpp"
       _memcpy(($smppos|0),($130|0),($129|0))|0; //@line 250 "load_it.cpp"
       $131 = (($$op27) + ($127))|0; //@line 255 "load_it.cpp"
       _memset(($patpos|0),0,960)|0; //@line 257 "load_it.cpp"
       $132 = ($18&65535)>(240); //@line 259 "load_it.cpp"
       $$op28 = $17 << 2; //@line 260 "load_it.cpp"
       $133 = $132 ? 960 : $$op28; //@line 260 "load_it.cpp"
       $134 = (($lpStream) + ($131)|0); //@line 261 "load_it.cpp"
       _memcpy(($patpos|0),($134|0),($133|0))|0; //@line 261 "load_it.cpp"
       $135 = (($$op28) + ($131))|0; //@line 266 "load_it.cpp"
       $136 = (($135) + 2)|0; //@line 268 "load_it.cpp"
       $137 = ($136>>>0)<($dwMemLength>>>0); //@line 268 "load_it.cpp"
       if ($137) {
        $138 = (($lpStream) + ($135)|0); //@line 270 "load_it.cpp"
        $139 = HEAP16[$138>>1]|0; //@line 270 "load_it.cpp"
        $140 = $139&65535; //@line 270 "load_it.cpp"
        $141 = $140 << 3; //@line 272 "load_it.cpp"
        $142 = (($141) + ($136))|0; //@line 272 "load_it.cpp"
        $143 = ($142>>>0)<($dwMemLength>>>0); //@line 272 "load_it.cpp"
        $$73 = $143 ? $142 : $136; //@line 272 "load_it.cpp"
        $dwMemPos$0 = $$73;
       } else {
        $dwMemPos$0 = $135;
       }
       $144 = ((($this)) + 59660|0); //@line 275 "load_it.cpp"
       $145 = HEAP32[$144>>2]|0; //@line 275 "load_it.cpp"
       $146 = $145 & 1; //@line 275 "load_it.cpp"
       $147 = ($146|0)==(0); //@line 275 "load_it.cpp"
       do {
        if ($147) {
         $dwMemPos$1 = $dwMemPos$0;
        } else {
         $148 = (($dwMemPos$0) + 4896)|0; //@line 277 "load_it.cpp"
         $149 = ($148>>>0)<($dwMemLength>>>0); //@line 277 "load_it.cpp"
         if (!($149)) {
          $dwMemPos$1 = $dwMemPos$0;
          break;
         }
         $150 = ((($this)) + 53600|0); //@line 279 "load_it.cpp"
         $151 = (($lpStream) + ($dwMemPos$0)|0); //@line 279 "load_it.cpp"
         _memcpy(($150|0),($151|0),4896)|0; //@line 279 "load_it.cpp"
         $dwMemPos$1 = $148;
        }
       } while(0);
       $152 = (($dwMemPos$1) + 8)|0; //@line 284 "load_it.cpp"
       $153 = ($152>>>0)<($dwMemLength>>>0); //@line 284 "load_it.cpp"
       do {
        if ($153) {
         $154 = (($lpStream) + ($dwMemPos$1)|0); //@line 284 "load_it.cpp"
         $155 = HEAP32[$154>>2]|0; //@line 284 "load_it.cpp"
         $156 = ($155|0)==(1296125520); //@line 284 "load_it.cpp"
         if (!($156)) {
          $dwMemPos$2 = $dwMemPos$1;
          break;
         }
         $$sum70 = (($dwMemPos$1) + 4)|0; //@line 286 "load_it.cpp"
         $157 = (($lpStream) + ($$sum70)|0); //@line 286 "load_it.cpp"
         $158 = HEAP32[$157>>2]|0; //@line 286 "load_it.cpp"
         $159 = (($158) + ($152))|0; //@line 288 "load_it.cpp"
         $160 = ($159>>>0)<=($dwMemLength>>>0); //@line 288 "load_it.cpp"
         $$off71 = (($158) + -32)|0; //@line 288 "load_it.cpp"
         $161 = ($$off71>>>0)<(7649); //@line 288 "load_it.cpp"
         $162 = $161 & $160; //@line 288 "load_it.cpp"
         if (!($162)) {
          $dwMemPos$2 = $152;
          break;
         }
         $163 = (__Znaj($158)|0); //@line 290 "load_it.cpp"
         $164 = ((($this)) + 59800|0); //@line 290 "load_it.cpp"
         HEAP32[$164>>2] = $163; //@line 290 "load_it.cpp"
         $165 = $158 >>> 5; //@line 293 "load_it.cpp"
         $166 = ((($this)) + 59792|0); //@line 293 "load_it.cpp"
         HEAP32[$166>>2] = $165; //@line 293 "load_it.cpp"
         $167 = (($lpStream) + ($152)|0); //@line 294 "load_it.cpp"
         _memcpy(($163|0),($167|0),($158|0))|0; //@line 294 "load_it.cpp"
         $dwMemPos$2 = $159;
        } else {
         $dwMemPos$2 = $dwMemPos$1;
        }
       } while(0);
       HEAP32[$104>>2] = 4; //@line 300 "load_it.cpp"
       $168 = (($dwMemPos$2) + 8)|0; //@line 302 "load_it.cpp"
       $169 = ($168>>>0)<($dwMemLength>>>0); //@line 302 "load_it.cpp"
       do {
        if ($169) {
         $170 = (($lpStream) + ($dwMemPos$2)|0); //@line 302 "load_it.cpp"
         $171 = HEAP32[$170>>2]|0; //@line 302 "load_it.cpp"
         $172 = ($171|0)==(1296125507); //@line 302 "load_it.cpp"
         if (!($172)) {
          $dwMemPos$3 = $dwMemPos$2;
          break;
         }
         $$sum68 = (($dwMemPos$2) + 4)|0; //@line 304 "load_it.cpp"
         $173 = (($lpStream) + ($$sum68)|0); //@line 304 "load_it.cpp"
         $174 = HEAP32[$173>>2]|0; //@line 304 "load_it.cpp"
         $175 = (($174) + ($168))|0; //@line 306 "load_it.cpp"
         $176 = ($175>>>0)<=($dwMemLength>>>0); //@line 306 "load_it.cpp"
         $177 = ($174>>>0)<(1281); //@line 306 "load_it.cpp"
         $or$cond9 = $177 & $176; //@line 306 "load_it.cpp"
         if (!($or$cond9)) {
          $dwMemPos$3 = $168;
          break;
         }
         $178 = (($174>>>0) / 20)&-1; //@line 308 "load_it.cpp"
         $179 = ($174>>>0)>(99); //@line 309 "load_it.cpp"
         if ($179) {
          HEAP32[$104>>2] = $178; //@line 309 "load_it.cpp"
          $i$0131 = 0;
         } else {
          $180 = ($174>>>0)>(19); //@line 310 "load_it.cpp"
          if ($180) {
           $i$0131 = 0;
          } else {
           $dwMemPos$3 = $175;
           break;
          }
         }
         while(1) {
          $181 = (((((($this)) + 49600|0) + (($i$0131*36)|0)|0)) + 16|0); //@line 312 "load_it.cpp"
          $182 = ($i$0131*20)|0; //@line 312 "load_it.cpp"
          $$sum69 = (($182) + ($168))|0; //@line 312 "load_it.cpp"
          $183 = (($lpStream) + ($$sum69)|0); //@line 312 "load_it.cpp"
          dest=$181; src=$183; stop=dest+20|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 312 "load_it.cpp"
          $184 = (((((($this)) + 49600|0) + (($i$0131*36)|0)|0)) + 35|0); //@line 313 "load_it.cpp"
          HEAP8[$184>>0] = 0; //@line 313 "load_it.cpp"
          $185 = (($i$0131) + 1)|0; //@line 310 "load_it.cpp"
          $186 = ($185>>>0)<($178>>>0); //@line 310 "load_it.cpp"
          if ($186) {
           $i$0131 = $185;
          } else {
           $dwMemPos$3 = $175;
           break;
          }
         }
        } else {
         $dwMemPos$3 = $dwMemPos$2;
        }
       } while(0);
       $187 = (($dwMemPos$3) + 8)|0; //@line 319 "load_it.cpp"
       $188 = ($187>>>0)<($dwMemLength>>>0); //@line 319 "load_it.cpp"
       if ($188) {
        $189 = (($lpStream) + ($dwMemPos$3)|0); //@line 321 "load_it.cpp"
        $190 = (($dwMemLength) - ($dwMemPos$3))|0; //@line 321 "load_it.cpp"
        (__ZN10CSoundFile14LoadMixPluginsEPKvj($this,$189,$190)|0); //@line 321 "load_it.cpp"
       }
       $191 = ($17|0)==(0); //@line 326 "load_it.cpp"
       $not$ = ($14>>>0)<(15794176); //@line 326 "load_it.cpp"
       $192 = $191 & $not$; //@line 326 "load_it.cpp"
       if (!($192)) {
        $193 = (($dwMemLength) + -4)|0;
        $194 = ($14>>>0)<(15728640);
        $195 = $194 ? $17 : 240; //@line 326 "load_it.cpp"
        $patchk$0129 = 0;
        while(1) {
         dest=$chnmask; stop=dest+64|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 328 "load_it.cpp"
         $196 = (($patpos) + ($patchk$0129<<2)|0); //@line 329 "load_it.cpp"
         $197 = HEAP32[$196>>2]|0; //@line 329 "load_it.cpp"
         $198 = ($197|0)!=(0); //@line 329 "load_it.cpp"
         $199 = ($197>>>0)<($193>>>0); //@line 329 "load_it.cpp"
         $or$cond75 = $198 & $199; //@line 329 "load_it.cpp"
         L86: do {
          if ($or$cond75) {
           $200 = (($lpStream) + ($197)|0); //@line 330 "load_it.cpp"
           $201 = HEAP16[$200>>1]|0; //@line 330 "load_it.cpp"
           $202 = $201&65535; //@line 330 "load_it.cpp"
           $$sum63 = (($197) + 2)|0; //@line 331 "load_it.cpp"
           $203 = (($lpStream) + ($$sum63)|0); //@line 331 "load_it.cpp"
           $204 = HEAP16[$203>>1]|0; //@line 331 "load_it.cpp"
           $205 = $204&65535; //@line 331 "load_it.cpp"
           $$off64 = (($204) + -4)<<16>>16; //@line 332 "load_it.cpp"
           $206 = ($$off64&65535)>(252); //@line 332 "load_it.cpp"
           if ($206) {
            break;
           }
           $207 = (($202) + 8)|0; //@line 333 "load_it.cpp"
           $208 = ($207>>>0)>($dwMemLength>>>0); //@line 333 "load_it.cpp"
           $209 = (($dwMemLength) - ($207))|0;
           $210 = ($197>>>0)>($209>>>0); //@line 333 "load_it.cpp"
           $or$cond77 = $208 | $210; //@line 333 "load_it.cpp"
           if ($or$cond77) {
            break;
           }
           $$sum65 = (($197) + 8)|0; //@line 335 "load_it.cpp"
           $211 = ($204<<16>>16)==(0); //@line 337 "load_it.cpp"
           if ($211) {
            break;
           } else {
            $i6$0$ph189 = 0;$nrow$0$ph188 = 0;
           }
           while(1) {
            $i6$0$us = $i6$0$ph189;
            while(1) {
             $212 = ($i6$0$us>>>0)<($202>>>0); //@line 339 "load_it.cpp"
             if (!($212)) {
              break L86;
             }
             $213 = (($i6$0$us) + 1)|0; //@line 340 "load_it.cpp"
             $$sum66$us = (($i6$0$us) + ($$sum65))|0; //@line 340 "load_it.cpp"
             $214 = (($lpStream) + ($$sum66$us)|0); //@line 340 "load_it.cpp"
             $215 = HEAP8[$214>>0]|0; //@line 340 "load_it.cpp"
             $216 = ($215<<24>>24)==(0); //@line 341 "load_it.cpp"
             if ($216) {
              $$lcssa195 = $213;
              break;
             }
             $217 = $215&255; //@line 346 "load_it.cpp"
             $218 = $217 & 127; //@line 346 "load_it.cpp"
             $219 = ($218|0)==(0); //@line 347 "load_it.cpp"
             $220 = (($217) + 63)|0; //@line 347 "load_it.cpp"
             $221 = $220 & 63; //@line 347 "load_it.cpp"
             $ch$0$us = $219 ? 0 : $221; //@line 347 "load_it.cpp"
             $222 = $217 & 128; //@line 348 "load_it.cpp"
             $223 = ($222|0)==(0); //@line 348 "load_it.cpp"
             if ($223) {
              $$phi$trans$insert = (($chnmask) + ($ch$0$us)|0);
              $$pre = HEAP8[$$phi$trans$insert>>0]|0; //@line 354 "load_it.cpp"
              $230 = $$pre;$i6$1$us = $213;
             } else {
              $224 = ($213>>>0)<($202>>>0); //@line 350 "load_it.cpp"
              if (!($224)) {
               break L86;
              }
              $225 = (($i6$0$us) + 2)|0; //@line 351 "load_it.cpp"
              $$sum67$us = (($213) + ($$sum65))|0; //@line 351 "load_it.cpp"
              $226 = (($lpStream) + ($$sum67$us)|0); //@line 351 "load_it.cpp"
              $227 = HEAP8[$226>>0]|0; //@line 351 "load_it.cpp"
              $228 = (($chnmask) + ($ch$0$us)|0); //@line 351 "load_it.cpp"
              HEAP8[$228>>0] = $227; //@line 351 "load_it.cpp"
              $230 = $227;$i6$1$us = $225;
             }
             $229 = $230 & 15; //@line 354 "load_it.cpp"
             $231 = ($229<<24>>24)==(0); //@line 354 "load_it.cpp"
             do {
              if (!($231)) {
               $232 = HEAP32[$104>>2]|0; //@line 356 "load_it.cpp"
               $233 = ($ch$0$us>>>0)<($232>>>0); //@line 356 "load_it.cpp"
               if ($233) {
                break;
               }
               $234 = (($ch$0$us) + 1)|0; //@line 356 "load_it.cpp"
               HEAP32[$104>>2] = $234; //@line 356 "load_it.cpp"
              }
             } while(0);
             $235 = $230 & 1; //@line 359 "load_it.cpp"
             $236 = $235&255; //@line 359 "load_it.cpp"
             $i6$1$$us = (($236) + ($i6$1$us))|0; //@line 359 "load_it.cpp"
             $237 = ($230&255) >>> 1; //@line 361 "load_it.cpp"
             $$lobit101$us = $237 & 1; //@line 361 "load_it.cpp"
             $238 = $$lobit101$us&255; //@line 361 "load_it.cpp"
             $i6$3$us = (($i6$1$$us) + ($238))|0; //@line 361 "load_it.cpp"
             $239 = ($230&255) >>> 2; //@line 363 "load_it.cpp"
             $$lobit102$us = $239 & 1; //@line 363 "load_it.cpp"
             $240 = $$lobit102$us&255; //@line 363 "load_it.cpp"
             $i6$3$$us = (($i6$3$us) + ($240))|0; //@line 363 "load_it.cpp"
             $241 = $230 & 8; //@line 365 "load_it.cpp"
             $242 = ($241<<24>>24)==(0); //@line 365 "load_it.cpp"
             $243 = (($i6$3$$us) + 2)|0; //@line 365 "load_it.cpp"
             $i6$5$us = $242 ? $i6$3$$us : $243; //@line 365 "load_it.cpp"
             $244 = ($i6$5$us>>>0)<($202>>>0); //@line 366 "load_it.cpp"
             if ($244) {
              $i6$0$us = $i6$5$us;
             } else {
              break L86;
             }
            }
            $245 = (($nrow$0$ph188) + 1)|0; //@line 343 "load_it.cpp"
            $246 = ($245>>>0)<($205>>>0); //@line 337 "load_it.cpp"
            if ($246) {
             $i6$0$ph189 = $$lcssa195;$nrow$0$ph188 = $245;
            } else {
             break;
            }
           }
          }
         } while(0);
         $247 = (($patchk$0129) + 1)|0; //@line 326 "load_it.cpp"
         $exitcond144 = ($247|0)==($195|0); //@line 326 "load_it.cpp"
         if ($exitcond144) {
          break;
         } else {
          $patchk$0129 = $247;
         }
        }
       }
       $248 = ((($this)) + 59688|0); //@line 370 "load_it.cpp"
       $249 = $31 & 4; //@line 371 "load_it.cpp"
       $250 = ($249<<16>>16)==(0); //@line 371 "load_it.cpp"
       do {
        if ($250) {
         HEAP32[$248>>2] = 0; //@line 370 "load_it.cpp"
        } else {
         $251 = ($24>>>0)>(239); //@line 372 "load_it.cpp"
         $$100 = $251 ? 239 : $24; //@line 372 "load_it.cpp"
         HEAP32[$248>>2] = $$100; //@line 370 "load_it.cpp"
         $252 = ($$100|0)==(0); //@line 373 "load_it.cpp"
         if ($252) {
          break;
         }
         $253 = (($dwMemLength) + -554)|0;
         $254 = HEAP16[$4>>1]|0; //@line 381 "load_it.cpp"
         $255 = $254&65535; //@line 381 "load_it.cpp"
         $620 = $$100;$nins$0127 = 0;
         while(1) {
          $256 = (($inspos) + ($nins$0127<<2)|0); //@line 375 "load_it.cpp"
          $257 = HEAP32[$256>>2]|0; //@line 375 "load_it.cpp"
          $258 = ($257|0)!=(0); //@line 375 "load_it.cpp"
          $259 = ($257>>>0)<($253>>>0); //@line 375 "load_it.cpp"
          $or$cond80 = $258 & $259; //@line 375 "load_it.cpp"
          if ($or$cond80) {
           $260 = (__Znwj(632)|0); //@line 377 "load_it.cpp"
           $261 = (($nins$0127) + 1)|0; //@line 379 "load_it.cpp"
           $262 = (((($this)) + 48640|0) + ($261<<2)|0); //@line 379 "load_it.cpp"
           HEAP32[$262>>2] = $260; //@line 379 "load_it.cpp"
           _memset(($260|0),0,632)|0; //@line 380 "load_it.cpp"
           $263 = (($lpStream) + ($257)|0); //@line 381 "load_it.cpp"
           (__ZN10CSoundFile12ITInstrToMPTEPKvP17_INSTRUMENTHEADERj(0,$263,$260,$255)|0); //@line 381 "load_it.cpp"
           $$pre150 = HEAP32[$248>>2]|0; //@line 373 "load_it.cpp"
           $$pre$phi177Z2D = $261;$265 = $$pre150;
          } else {
           $$pre176 = (($nins$0127) + 1)|0; //@line 373 "load_it.cpp"
           $$pre$phi177Z2D = $$pre176;$265 = $620;
          }
          $264 = ($$pre$phi177Z2D>>>0)<($265>>>0); //@line 373 "load_it.cpp"
          if ($264) {
           $620 = $265;$nins$0127 = $$pre$phi177Z2D;
          } else {
           break;
          }
         }
        }
       } while(0);
       $266 = ((($this)) + 59684|0); //@line 385 "load_it.cpp"
       HEAP32[$266>>2] = $19; //@line 385 "load_it.cpp"
       $267 = (($dwMemLength) + -80)|0;
       $268 = HEAP16[$4>>1]|0; //@line 444 "load_it.cpp"
       $269 = ($268&65535)>(532); //@line 444 "load_it.cpp"
       $nsmp$0125 = 0;
       while(1) {
        $273 = (($smppos) + ($nsmp$0125<<2)|0); //@line 387 "load_it.cpp"
        $274 = HEAP32[$273>>2]|0; //@line 387 "load_it.cpp"
        $275 = (($274) + -1)|0; //@line 387 "load_it.cpp"
        $276 = ($275>>>0)<($267>>>0); //@line 387 "load_it.cpp"
        if ($276) {
         $277 = (($lpStream) + ($274)|0); //@line 389 "load_it.cpp"
         $278 = HEAPU8[$277>>0]|(HEAPU8[$277+1>>0]<<8)|(HEAPU8[$277+2>>0]<<16)|(HEAPU8[$277+3>>0]<<24); //@line 389 "load_it.cpp"
         $$sum39 = (($274) + 4)|0; //@line 389 "load_it.cpp"
         $279 = (($lpStream) + ($$sum39)|0); //@line 389 "load_it.cpp"
         dest=$pis$sroa$6; src=$279; stop=dest+12|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 389 "load_it.cpp"
         $$sum40 = (($274) + 17)|0; //@line 389 "load_it.cpp"
         $280 = (($lpStream) + ($$sum40)|0); //@line 389 "load_it.cpp"
         $281 = HEAP8[$280>>0]|0; //@line 389 "load_it.cpp"
         $$sum41 = (($274) + 18)|0; //@line 389 "load_it.cpp"
         $282 = (($lpStream) + ($$sum41)|0); //@line 389 "load_it.cpp"
         $283 = HEAP8[$282>>0]|0; //@line 389 "load_it.cpp"
         $$sum42 = (($274) + 19)|0; //@line 389 "load_it.cpp"
         $284 = (($lpStream) + ($$sum42)|0); //@line 389 "load_it.cpp"
         $285 = HEAP8[$284>>0]|0; //@line 389 "load_it.cpp"
         $$sum43 = (($274) + 20)|0; //@line 389 "load_it.cpp"
         $286 = (($lpStream) + ($$sum43)|0); //@line 389 "load_it.cpp"
         dest=$pis$sroa$18; src=$286; stop=dest+26|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 389 "load_it.cpp"
         $$sum44 = (($274) + 46)|0; //@line 389 "load_it.cpp"
         $287 = (($lpStream) + ($$sum44)|0); //@line 389 "load_it.cpp"
         $288 = HEAP8[$287>>0]|0; //@line 389 "load_it.cpp"
         $$sum45 = (($274) + 47)|0; //@line 389 "load_it.cpp"
         $289 = (($lpStream) + ($$sum45)|0); //@line 389 "load_it.cpp"
         $290 = HEAP8[$289>>0]|0; //@line 389 "load_it.cpp"
         $$sum46 = (($274) + 48)|0; //@line 389 "load_it.cpp"
         $291 = (($lpStream) + ($$sum46)|0); //@line 389 "load_it.cpp"
         $292 = HEAPU8[$291>>0]|(HEAPU8[$291+1>>0]<<8)|(HEAPU8[$291+2>>0]<<16)|(HEAPU8[$291+3>>0]<<24); //@line 389 "load_it.cpp"
         $$sum49 = (($274) + 60)|0; //@line 389 "load_it.cpp"
         $293 = (($lpStream) + ($$sum49)|0); //@line 389 "load_it.cpp"
         $294 = HEAPU8[$293>>0]|(HEAPU8[$293+1>>0]<<8)|(HEAPU8[$293+2>>0]<<16)|(HEAPU8[$293+3>>0]<<24); //@line 389 "load_it.cpp"
         $$sum52 = (($274) + 72)|0; //@line 389 "load_it.cpp"
         $295 = (($lpStream) + ($$sum52)|0); //@line 389 "load_it.cpp"
         $296 = HEAPU8[$295>>0]|(HEAPU8[$295+1>>0]<<8)|(HEAPU8[$295+2>>0]<<16)|(HEAPU8[$295+3>>0]<<24); //@line 389 "load_it.cpp"
         $$sum53 = (($274) + 76)|0; //@line 389 "load_it.cpp"
         $297 = (($lpStream) + ($$sum53)|0); //@line 389 "load_it.cpp"
         $298 = HEAP8[$297>>0]|0; //@line 389 "load_it.cpp"
         $$sum54 = (($274) + 77)|0; //@line 389 "load_it.cpp"
         $299 = (($lpStream) + ($$sum54)|0); //@line 389 "load_it.cpp"
         $300 = HEAP8[$299>>0]|0; //@line 389 "load_it.cpp"
         $$sum55 = (($274) + 78)|0; //@line 389 "load_it.cpp"
         $301 = (($lpStream) + ($$sum55)|0); //@line 389 "load_it.cpp"
         $302 = HEAP8[$301>>0]|0; //@line 389 "load_it.cpp"
         $$sum56 = (($274) + 79)|0; //@line 389 "load_it.cpp"
         $303 = (($lpStream) + ($$sum56)|0); //@line 389 "load_it.cpp"
         $304 = HEAP8[$303>>0]|0; //@line 389 "load_it.cpp"
         $305 = ($278|0)==(1397771593); //@line 399 "load_it.cpp"
         do {
          if ($305) {
           $$sum51 = (($274) + 68)|0; //@line 389 "load_it.cpp"
           $306 = (($lpStream) + ($$sum51)|0); //@line 389 "load_it.cpp"
           $307 = HEAPU8[$306>>0]|(HEAPU8[$306+1>>0]<<8)|(HEAPU8[$306+2>>0]<<16)|(HEAPU8[$306+3>>0]<<24); //@line 389 "load_it.cpp"
           $$sum50 = (($274) + 64)|0; //@line 389 "load_it.cpp"
           $308 = (($lpStream) + ($$sum50)|0); //@line 389 "load_it.cpp"
           $309 = HEAPU8[$308>>0]|(HEAPU8[$308+1>>0]<<8)|(HEAPU8[$308+2>>0]<<16)|(HEAPU8[$308+3>>0]<<24); //@line 389 "load_it.cpp"
           $$sum48 = (($274) + 56)|0; //@line 389 "load_it.cpp"
           $310 = (($lpStream) + ($$sum48)|0); //@line 389 "load_it.cpp"
           $311 = HEAPU8[$310>>0]|(HEAPU8[$310+1>>0]<<8)|(HEAPU8[$310+2>>0]<<16)|(HEAPU8[$310+3>>0]<<24); //@line 389 "load_it.cpp"
           $$sum47 = (($274) + 52)|0; //@line 389 "load_it.cpp"
           $312 = (($lpStream) + ($$sum47)|0); //@line 389 "load_it.cpp"
           $313 = HEAPU8[$312>>0]|(HEAPU8[$312+1>>0]<<8)|(HEAPU8[$312+2>>0]<<16)|(HEAPU8[$312+3>>0]<<24); //@line 389 "load_it.cpp"
           $314 = (($nsmp$0125) + 1)|0; //@line 401 "load_it.cpp"
           $315 = (((($this)) + 33280|0) + ($314<<6)|0); //@line 401 "load_it.cpp"
           $316 = (((((($this)) + 33280|0) + ($314<<6)|0)) + 42|0); //@line 402 "load_it.cpp"
           dest=$316; src=$pis$sroa$6; stop=dest+12|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 402 "load_it.cpp"
           $317 = (((((($this)) + 33280|0) + ($314<<6)|0)) + 34|0); //@line 403 "load_it.cpp"
           HEAP16[$317>>1] = 0; //@line 403 "load_it.cpp"
           HEAP32[$315>>2] = 0; //@line 404 "load_it.cpp"
           $318 = (((((($this)) + 33280|0) + ($314<<6)|0)) + 4|0); //@line 405 "load_it.cpp"
           HEAP32[$318>>2] = $313; //@line 405 "load_it.cpp"
           $319 = (((((($this)) + 33280|0) + ($314<<6)|0)) + 8|0); //@line 406 "load_it.cpp"
           HEAP32[$319>>2] = $311; //@line 406 "load_it.cpp"
           $320 = (((((($this)) + 33280|0) + ($314<<6)|0)) + 12|0); //@line 407 "load_it.cpp"
           HEAP32[$320>>2] = $309; //@line 407 "load_it.cpp"
           $321 = (((((($this)) + 33280|0) + ($314<<6)|0)) + 16|0); //@line 408 "load_it.cpp"
           HEAP32[$321>>2] = $307; //@line 408 "load_it.cpp"
           $322 = (((((($this)) + 33280|0) + ($314<<6)|0)) + 24|0); //@line 409 "load_it.cpp"
           $323 = ($294|0)==(0); //@line 410 "load_it.cpp"
           $$84 = $323 ? 8363 : $294; //@line 410 "load_it.cpp"
           $324 = ($294>>>0)<(256); //@line 411 "load_it.cpp"
           $storemerge61 = $324 ? 256 : $$84; //@line 411 "load_it.cpp"
           HEAP32[$322>>2] = $storemerge61; //@line 409 "load_it.cpp"
           $325 = $285&255; //@line 412 "load_it.cpp"
           $326 = $325 << 2; //@line 412 "load_it.cpp"
           $327 = $326&65535; //@line 412 "load_it.cpp"
           $328 = (((((($this)) + 33280|0) + ($314<<6)|0)) + 30|0); //@line 412 "load_it.cpp"
           $329 = ($326>>>0)>(256); //@line 413 "load_it.cpp"
           $$85 = $329 ? 256 : $327; //@line 413 "load_it.cpp"
           HEAP16[$328>>1] = $$85; //@line 412 "load_it.cpp"
           $330 = $281&255; //@line 414 "load_it.cpp"
           $331 = (((((($this)) + 33280|0) + ($314<<6)|0)) + 32|0); //@line 414 "load_it.cpp"
           $332 = ($281&255)>(64); //@line 415 "load_it.cpp"
           $storemerge59 = $332 ? 64 : $330; //@line 415 "load_it.cpp"
           HEAP16[$331>>1] = $storemerge59; //@line 414 "load_it.cpp"
           $333 = $283&255; //@line 416 "load_it.cpp"
           $334 = $333 & 16; //@line 416 "load_it.cpp"
           $335 = ($334|0)==(0); //@line 416 "load_it.cpp"
           if ($335) {
            $339 = 0;
           } else {
            HEAP16[$317>>1] = 2; //@line 416 "load_it.cpp"
            $339 = 2;
           }
           $336 = $333 & 32; //@line 417 "load_it.cpp"
           $337 = ($336|0)==(0); //@line 417 "load_it.cpp"
           if ($337) {
            $345 = $339;
           } else {
            $338 = $339&65535; //@line 417 "load_it.cpp"
            $340 = $338 | 8; //@line 417 "load_it.cpp"
            $341 = $340&65535; //@line 417 "load_it.cpp"
            HEAP16[$317>>1] = $341; //@line 417 "load_it.cpp"
            $345 = $341;
           }
           $342 = $333 & 64; //@line 418 "load_it.cpp"
           $343 = ($342|0)==(0); //@line 418 "load_it.cpp"
           if ($343) {
            $351 = $345;
           } else {
            $344 = $345&65535; //@line 418 "load_it.cpp"
            $346 = $344 | 4; //@line 418 "load_it.cpp"
            $347 = $346&65535; //@line 418 "load_it.cpp"
            HEAP16[$317>>1] = $347; //@line 418 "load_it.cpp"
            $351 = $347;
           }
           $348 = $333 & 128; //@line 419 "load_it.cpp"
           $349 = ($348|0)==(0); //@line 419 "load_it.cpp"
           if ($349) {
            $363 = $351;
           } else {
            $350 = $351&65535; //@line 419 "load_it.cpp"
            $352 = $350 | 16; //@line 419 "load_it.cpp"
            $353 = $352&65535; //@line 419 "load_it.cpp"
            HEAP16[$317>>1] = $353; //@line 419 "load_it.cpp"
            $363 = $353;
           }
           $354 = $290&255; //@line 420 "load_it.cpp"
           $355 = $354 << 2; //@line 420 "load_it.cpp"
           $356 = $355 & 508; //@line 420 "load_it.cpp"
           $357 = $356&65535; //@line 420 "load_it.cpp"
           $358 = (((((($this)) + 33280|0) + ($314<<6)|0)) + 28|0); //@line 420 "load_it.cpp"
           $359 = ($356>>>0)>(256); //@line 421 "load_it.cpp"
           $$86 = $359 ? 256 : $357; //@line 421 "load_it.cpp"
           HEAP16[$358>>1] = $$86; //@line 420 "load_it.cpp"
           $360 = $354 & 128; //@line 422 "load_it.cpp"
           $361 = ($360|0)==(0); //@line 422 "load_it.cpp"
           if ($361) {
            $393 = $363;
           } else {
            $362 = $363&65535; //@line 422 "load_it.cpp"
            $364 = $362 | 32; //@line 422 "load_it.cpp"
            $365 = $364&65535; //@line 422 "load_it.cpp"
            HEAP16[$317>>1] = $365; //@line 422 "load_it.cpp"
            $393 = $365;
           }
           $366 = $304&255; //@line 423 "load_it.cpp"
           $367 = $366 & 7; //@line 423 "load_it.cpp"
           $368 = (523927 + ($367)|0); //@line 423 "load_it.cpp"
           $369 = HEAP8[$368>>0]|0; //@line 423 "load_it.cpp"
           $370 = (((((($this)) + 33280|0) + ($314<<6)|0)) + 38|0); //@line 423 "load_it.cpp"
           HEAP8[$370>>0] = $369; //@line 423 "load_it.cpp"
           $371 = (((((($this)) + 33280|0) + ($314<<6)|0)) + 41|0); //@line 424 "load_it.cpp"
           HEAP8[$371>>0] = $298; //@line 424 "load_it.cpp"
           $372 = $300&255; //@line 425 "load_it.cpp"
           $373 = $372 & 127; //@line 425 "load_it.cpp"
           $374 = $373&255; //@line 425 "load_it.cpp"
           $375 = (((((($this)) + 33280|0) + ($314<<6)|0)) + 40|0); //@line 425 "load_it.cpp"
           HEAP8[$375>>0] = $374; //@line 425 "load_it.cpp"
           $376 = $302&255; //@line 426 "load_it.cpp"
           $377 = (($376) + 3)|0; //@line 426 "load_it.cpp"
           $378 = $377 >>> 2; //@line 426 "load_it.cpp"
           $379 = $378&255; //@line 426 "load_it.cpp"
           $380 = (((((($this)) + 33280|0) + ($314<<6)|0)) + 39|0); //@line 426 "load_it.cpp"
           HEAP8[$380>>0] = $379; //@line 426 "load_it.cpp"
           $notlhs = ($296|0)==(0); //@line 427 "load_it.cpp"
           $notrhs = ($296>>>0)>=($dwMemLength>>>0); //@line 427 "load_it.cpp"
           $or$cond87$not = $notrhs | $notlhs; //@line 427 "load_it.cpp"
           $381 = ($292|0)==(0); //@line 427 "load_it.cpp"
           $or$cond88 = $381 | $or$cond87$not; //@line 427 "load_it.cpp"
           if ($or$cond88) {
            $$pre$phi173Z2D = $314;
            break;
           }
           $382 = ($292>>>0)>(16000000); //@line 430 "load_it.cpp"
           $$89 = $382 ? 16000000 : $292; //@line 430 "load_it.cpp"
           HEAP32[$315>>2] = $$89; //@line 429 "load_it.cpp"
           $383 = $288&255; //@line 431 "load_it.cpp"
           $384 = $383 & 1; //@line 431 "load_it.cpp"
           $385 = $384 ^ 1; //@line 431 "load_it.cpp"
           $386 = $333 & 2; //@line 432 "load_it.cpp"
           $387 = ($386|0)==(0); //@line 432 "load_it.cpp"
           do {
            if ($387) {
             $402 = ($288<<24>>24)==(-1); //@line 442 "load_it.cpp"
             if ($402) {
              $flags$2 = 3;
              break;
             }
             $403 = $333 << 1; //@line 441 "load_it.cpp"
             $404 = $403 & 8; //@line 441 "load_it.cpp"
             $405 = $385 | $404; //@line 441 "load_it.cpp"
             $406 = $333 & 8; //@line 444 "load_it.cpp"
             $407 = ($406|0)==(0); //@line 444 "load_it.cpp"
             if ($407) {
              $flags$2 = $405;
              break;
             }
             $408 = $383 & 4; //@line 444 "load_it.cpp"
             $409 = ($408|0)!=(0); //@line 444 "load_it.cpp"
             $410 = $409 & $269; //@line 444 "load_it.cpp"
             $411 = $410 ? 18 : 16; //@line 444 "load_it.cpp"
             $flags$2 = $411;
            } else {
             $388 = (($385) + 5)|0; //@line 434 "load_it.cpp"
             $389 = $333 << 1; //@line 435 "load_it.cpp"
             $390 = $389 & 8; //@line 435 "load_it.cpp"
             $391 = $388 | $390; //@line 435 "load_it.cpp"
             $392 = $393&65535; //@line 436 "load_it.cpp"
             $394 = $392 | 1; //@line 436 "load_it.cpp"
             $395 = $394&65535; //@line 436 "load_it.cpp"
             HEAP16[$317>>1] = $395; //@line 436 "load_it.cpp"
             $396 = $333 & 8; //@line 438 "load_it.cpp"
             $397 = ($396|0)==(0); //@line 438 "load_it.cpp"
             if ($397) {
              $flags$2 = $391;
              break;
             }
             $398 = $383 & 4; //@line 438 "load_it.cpp"
             $399 = ($398|0)!=(0); //@line 438 "load_it.cpp"
             $400 = $399 & $269; //@line 438 "load_it.cpp"
             $401 = $400 ? 22 : 20; //@line 438 "load_it.cpp"
             $flags$2 = $401;
            }
           } while(0);
           $412 = (($lpStream) + ($296)|0); //@line 446 "load_it.cpp"
           $413 = (($dwMemLength) - ($296))|0; //@line 446 "load_it.cpp"
           (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$315,$flags$2,$412,$413)|0); //@line 446 "load_it.cpp"
           $$pre$phi173Z2D = $314;
          } else {
           $$pre172 = (($nsmp$0125) + 1)|0; //@line 449 "load_it.cpp"
           $$pre$phi173Z2D = $$pre172;
          }
         } while(0);
         $414 = (((($this)) + 59804|0) + ($$pre$phi173Z2D<<5)|0); //@line 449 "load_it.cpp"
         dest=$414; src=$pis$sroa$18; stop=dest+26|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 449 "load_it.cpp"
         $$pre$phi175Z2D = $$pre$phi173Z2D;
        } else {
         $$pre174 = (($nsmp$0125) + 1)|0; //@line 387 "load_it.cpp"
         $$pre$phi175Z2D = $$pre174;
        }
        $415 = ($$pre$phi175Z2D>>>0)<($19>>>0); //@line 387 "load_it.cpp"
        if ($415) {
         $nsmp$0125 = $$pre$phi175Z2D;
        } else {
         break;
        }
       }
       if ($192) {
        $ncu$0111 = 0;
       } else {
        $270 = (($dwMemLength) + -4)|0;
        $271 = ($14>>>0)<(15728640);
        $272 = $271 ? $17 : 240; //@line 452 "load_it.cpp"
        $npat$0123 = 0;
        while(1) {
         $416 = (($patpos) + ($npat$0123<<2)|0); //@line 454 "load_it.cpp"
         $417 = HEAP32[$416>>2]|0; //@line 454 "load_it.cpp"
         $418 = ($417|0)!=(0); //@line 454 "load_it.cpp"
         $419 = ($417>>>0)<($270>>>0); //@line 454 "load_it.cpp"
         $or$cond93 = $418 & $419; //@line 454 "load_it.cpp"
         L157: do {
          if ($or$cond93) {
           $424 = (($lpStream) + ($417)|0); //@line 461 "load_it.cpp"
           $425 = HEAP16[$424>>1]|0; //@line 461 "load_it.cpp"
           $426 = $425&65535; //@line 461 "load_it.cpp"
           $$sum = (($417) + 2)|0; //@line 462 "load_it.cpp"
           $427 = (($lpStream) + ($$sum)|0); //@line 462 "load_it.cpp"
           $428 = HEAP16[$427>>1]|0; //@line 462 "load_it.cpp"
           $429 = $428&65535; //@line 462 "load_it.cpp"
           $$off = (($428) + -4)<<16>>16; //@line 463 "load_it.cpp"
           $430 = ($$off&65535)>(252); //@line 463 "load_it.cpp"
           if ($430) {
            break;
           }
           $431 = (($426) + 8)|0; //@line 464 "load_it.cpp"
           $432 = ($431>>>0)>($dwMemLength>>>0); //@line 464 "load_it.cpp"
           $433 = (($dwMemLength) - ($431))|0;
           $434 = ($417>>>0)>($433>>>0); //@line 464 "load_it.cpp"
           $or$cond95 = $432 | $434; //@line 464 "load_it.cpp"
           if ($or$cond95) {
            break;
           }
           $435 = (((($this)) + 52864|0) + ($npat$0123<<1)|0); //@line 465 "load_it.cpp"
           HEAP16[$435>>1] = $428; //@line 465 "load_it.cpp"
           $436 = HEAP32[$104>>2]|0; //@line 466 "load_it.cpp"
           $437 = (__ZN10CSoundFile15AllocatePatternEjj($429,$436)|0); //@line 466 "load_it.cpp"
           $438 = (((($this)) + 51904|0) + ($npat$0123<<2)|0); //@line 466 "load_it.cpp"
           HEAP32[$438>>2] = $437; //@line 466 "load_it.cpp"
           $439 = ($437|0)==(0|0); //@line 466 "load_it.cpp"
           if ($439) {
            break;
           }
           _memset(($lastvalue|0),0,384)|0; //@line 467 "load_it.cpp"
           dest=$chnmask; stop=dest+64|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 468 "load_it.cpp"
           $$sum29 = (($417) + 8)|0; //@line 471 "load_it.cpp"
           $440 = (($426) + -2)|0; //@line 563 "load_it.cpp"
           $441 = ($428<<16>>16)!=(0); //@line 473 "load_it.cpp"
           $442 = ($425<<16>>16)!=(0); //@line 475 "load_it.cpp"
           $or$cond96112119 = $442 & $441; //@line 473 "load_it.cpp"
           if (!($or$cond96112119)) {
            break;
           }
           $443 = (($417) + 9)|0; //@line 564 "load_it.cpp"
           $i9$0$ph121 = 0;$m$0$ph120 = $437;$nrow11$0$ph122 = 0;
           while(1) {
            $i9$0113 = $i9$0$ph121;
            while(1) {
             $444 = (($i9$0113) + 1)|0; //@line 476 "load_it.cpp"
             $$sum30 = (($i9$0113) + ($$sum29))|0; //@line 476 "load_it.cpp"
             $445 = (($lpStream) + ($$sum30)|0); //@line 476 "load_it.cpp"
             $446 = HEAP8[$445>>0]|0; //@line 476 "load_it.cpp"
             $447 = ($446<<24>>24)==(0); //@line 477 "load_it.cpp"
             if ($447) {
              $$lcssa = $444;
              break;
             }
             $453 = $446&255; //@line 483 "load_it.cpp"
             $454 = $453 & 127; //@line 483 "load_it.cpp"
             $455 = ($454|0)==(0); //@line 484 "load_it.cpp"
             $456 = (($453) + 63)|0; //@line 484 "load_it.cpp"
             $457 = $456 & 63; //@line 484 "load_it.cpp"
             $ch13$0 = $455 ? 0 : $457; //@line 484 "load_it.cpp"
             $458 = $453 & 128; //@line 485 "load_it.cpp"
             $459 = ($458|0)==(0); //@line 485 "load_it.cpp"
             if ($459) {
              $$phi$trans$insert152 = (($chnmask) + ($ch13$0)|0);
              $$pre153 = HEAP8[$$phi$trans$insert152>>0]|0; //@line 490 "load_it.cpp"
              $$pre$phiZ2D = $$phi$trans$insert152;$466 = $$pre153;$i9$1 = $444;
             } else {
              $460 = ($444>>>0)<($426>>>0); //@line 487 "load_it.cpp"
              if (!($460)) {
               break L157;
              }
              $461 = (($i9$0113) + 2)|0; //@line 488 "load_it.cpp"
              $$sum38 = (($444) + ($$sum29))|0; //@line 488 "load_it.cpp"
              $462 = (($lpStream) + ($$sum38)|0); //@line 488 "load_it.cpp"
              $463 = HEAP8[$462>>0]|0; //@line 488 "load_it.cpp"
              $464 = (($chnmask) + ($ch13$0)|0); //@line 488 "load_it.cpp"
              HEAP8[$464>>0] = $463; //@line 488 "load_it.cpp"
              $$pre$phiZ2D = $464;$466 = $463;$i9$1 = $461;
             }
             $465 = $466 & 16; //@line 490 "load_it.cpp"
             $467 = ($465<<24>>24)==(0); //@line 490 "load_it.cpp"
             do {
              if ($467) {
               $474 = $466;
              } else {
               $468 = HEAP32[$104>>2]|0; //@line 490 "load_it.cpp"
               $469 = ($ch13$0>>>0)<($468>>>0); //@line 490 "load_it.cpp"
               if (!($469)) {
                $474 = $466;
                break;
               }
               $470 = (($lastvalue) + (($ch13$0*6)|0)|0); //@line 492 "load_it.cpp"
               $471 = HEAP8[$470>>0]|0; //@line 492 "load_it.cpp"
               $472 = (($m$0$ph120) + (($ch13$0*6)|0)|0); //@line 492 "load_it.cpp"
               HEAP8[$472>>0] = $471; //@line 492 "load_it.cpp"
               $$pre154 = HEAP8[$$pre$phiZ2D>>0]|0; //@line 494 "load_it.cpp"
               $474 = $$pre154;
              }
             } while(0);
             $473 = $474 & 32; //@line 494 "load_it.cpp"
             $475 = ($473<<24>>24)==(0); //@line 494 "load_it.cpp"
             do {
              if ($475) {
               $482 = $474;
              } else {
               $476 = HEAP32[$104>>2]|0; //@line 494 "load_it.cpp"
               $477 = ($ch13$0>>>0)<($476>>>0); //@line 494 "load_it.cpp"
               if (!($477)) {
                $482 = $474;
                break;
               }
               $478 = (((($lastvalue) + (($ch13$0*6)|0)|0)) + 1|0); //@line 496 "load_it.cpp"
               $479 = HEAP8[$478>>0]|0; //@line 496 "load_it.cpp"
               $480 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 1|0); //@line 496 "load_it.cpp"
               HEAP8[$480>>0] = $479; //@line 496 "load_it.cpp"
               $$pre155 = HEAP8[$$pre$phiZ2D>>0]|0; //@line 498 "load_it.cpp"
               $482 = $$pre155;
              }
             } while(0);
             $481 = $482 & 64; //@line 498 "load_it.cpp"
             $483 = ($481<<24>>24)==(0); //@line 498 "load_it.cpp"
             do {
              if ($483) {
               $493 = $482;
              } else {
               $484 = HEAP32[$104>>2]|0; //@line 498 "load_it.cpp"
               $485 = ($ch13$0>>>0)<($484>>>0); //@line 498 "load_it.cpp"
               if (!($485)) {
                $493 = $482;
                break;
               }
               $486 = (((($lastvalue) + (($ch13$0*6)|0)|0)) + 2|0); //@line 500 "load_it.cpp"
               $487 = HEAP8[$486>>0]|0; //@line 500 "load_it.cpp"
               $488 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 2|0); //@line 500 "load_it.cpp"
               HEAP8[$488>>0] = $487; //@line 500 "load_it.cpp"
               $489 = (((($lastvalue) + (($ch13$0*6)|0)|0)) + 4|0); //@line 501 "load_it.cpp"
               $490 = HEAP8[$489>>0]|0; //@line 501 "load_it.cpp"
               $491 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 4|0); //@line 501 "load_it.cpp"
               HEAP8[$491>>0] = $490; //@line 501 "load_it.cpp"
               $$pr$pre = HEAP8[$$pre$phiZ2D>>0]|0; //@line 503 "load_it.cpp"
               $493 = $$pr$pre;
              }
             } while(0);
             $492 = ($493<<24>>24)<(0); //@line 503 "load_it.cpp"
             do {
              if ($492) {
               $494 = HEAP32[$104>>2]|0; //@line 503 "load_it.cpp"
               $495 = ($ch13$0>>>0)<($494>>>0); //@line 503 "load_it.cpp"
               if (!($495)) {
                $503 = $493;
                break;
               }
               $496 = (((($lastvalue) + (($ch13$0*6)|0)|0)) + 3|0); //@line 505 "load_it.cpp"
               $497 = HEAP8[$496>>0]|0; //@line 505 "load_it.cpp"
               $498 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 3|0); //@line 505 "load_it.cpp"
               HEAP8[$498>>0] = $497; //@line 505 "load_it.cpp"
               $499 = (((($lastvalue) + (($ch13$0*6)|0)|0)) + 5|0); //@line 506 "load_it.cpp"
               $500 = HEAP8[$499>>0]|0; //@line 506 "load_it.cpp"
               $501 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 5|0); //@line 506 "load_it.cpp"
               HEAP8[$501>>0] = $500; //@line 506 "load_it.cpp"
               $$pre157 = HEAP8[$$pre$phiZ2D>>0]|0; //@line 508 "load_it.cpp"
               $503 = $$pre157;
              } else {
               $503 = $493;
              }
             } while(0);
             $502 = $503 & 1; //@line 508 "load_it.cpp"
             $504 = ($502<<24>>24)==(0); //@line 508 "load_it.cpp"
             do {
              if ($504) {
               $517 = $503;$i9$2 = $i9$1;
              } else {
               $505 = ($i9$1>>>0)<($426>>>0); //@line 510 "load_it.cpp"
               if (!($505)) {
                break L157;
               }
               $506 = (($i9$1) + 1)|0; //@line 511 "load_it.cpp"
               $$sum37 = (($i9$1) + ($$sum29))|0; //@line 511 "load_it.cpp"
               $507 = (($lpStream) + ($$sum37)|0); //@line 511 "load_it.cpp"
               $508 = HEAP8[$507>>0]|0; //@line 511 "load_it.cpp"
               $509 = HEAP32[$104>>2]|0; //@line 512 "load_it.cpp"
               $510 = ($ch13$0>>>0)<($509>>>0); //@line 512 "load_it.cpp"
               if (!($510)) {
                $517 = $503;$i9$2 = $506;
                break;
               }
               $511 = $508&255; //@line 511 "load_it.cpp"
               $$lobit = ($508&255) >>> 7; //@line 514 "load_it.cpp"
               $512 = $$lobit&255; //@line 514 "load_it.cpp"
               $$not = $512 ^ 1; //@line 514 "load_it.cpp"
               $$97 = (($$not) + ($511))|0; //@line 514 "load_it.cpp"
               $513 = $$97&255; //@line 515 "load_it.cpp"
               $514 = (($m$0$ph120) + (($ch13$0*6)|0)|0); //@line 515 "load_it.cpp"
               HEAP8[$514>>0] = $513; //@line 515 "load_it.cpp"
               $515 = (($lastvalue) + (($ch13$0*6)|0)|0); //@line 516 "load_it.cpp"
               HEAP8[$515>>0] = $513; //@line 516 "load_it.cpp"
               $$pre158 = HEAP8[$$pre$phiZ2D>>0]|0; //@line 520 "load_it.cpp"
               $517 = $$pre158;$i9$2 = $506;
              }
             } while(0);
             $516 = $517 & 2; //@line 520 "load_it.cpp"
             $518 = ($516<<24>>24)==(0); //@line 520 "load_it.cpp"
             do {
              if ($518) {
               $528 = $517;$i9$3 = $i9$2;
              } else {
               $519 = ($i9$2>>>0)<($426>>>0); //@line 522 "load_it.cpp"
               if (!($519)) {
                break L157;
               }
               $520 = (($i9$2) + 1)|0; //@line 523 "load_it.cpp"
               $521 = HEAP32[$104>>2]|0; //@line 524 "load_it.cpp"
               $522 = ($ch13$0>>>0)<($521>>>0); //@line 524 "load_it.cpp"
               if (!($522)) {
                $528 = $517;$i9$3 = $520;
                break;
               }
               $$sum36 = (($i9$2) + ($$sum29))|0; //@line 523 "load_it.cpp"
               $523 = (($lpStream) + ($$sum36)|0); //@line 523 "load_it.cpp"
               $524 = HEAP8[$523>>0]|0; //@line 523 "load_it.cpp"
               $525 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 1|0); //@line 526 "load_it.cpp"
               HEAP8[$525>>0] = $524; //@line 526 "load_it.cpp"
               $526 = (((($lastvalue) + (($ch13$0*6)|0)|0)) + 1|0); //@line 527 "load_it.cpp"
               HEAP8[$526>>0] = $524; //@line 527 "load_it.cpp"
               $$pre159 = HEAP8[$$pre$phiZ2D>>0]|0; //@line 530 "load_it.cpp"
               $528 = $$pre159;$i9$3 = $520;
              }
             } while(0);
             $527 = $528 & 4; //@line 530 "load_it.cpp"
             $529 = ($527<<24>>24)==(0); //@line 530 "load_it.cpp"
             do {
              if ($529) {
               $591 = $528;$i9$4 = $i9$3;
              } else {
               $530 = ($i9$3>>>0)<($426>>>0); //@line 532 "load_it.cpp"
               if (!($530)) {
                break L157;
               }
               $531 = (($i9$3) + 1)|0; //@line 533 "load_it.cpp"
               $$sum33 = (($i9$3) + ($$sum29))|0; //@line 533 "load_it.cpp"
               $532 = (($lpStream) + ($$sum33)|0); //@line 533 "load_it.cpp"
               $533 = HEAP8[$532>>0]|0; //@line 533 "load_it.cpp"
               $534 = $533&255; //@line 533 "load_it.cpp"
               $535 = HEAP32[$104>>2]|0; //@line 534 "load_it.cpp"
               $536 = ($ch13$0>>>0)<($535>>>0); //@line 534 "load_it.cpp"
               if (!($536)) {
                $591 = $528;$i9$4 = $531;
                break;
               }
               $537 = ($533&255)<(65); //@line 537 "load_it.cpp"
               do {
                if ($537) {
                 $538 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 2|0); //@line 537 "load_it.cpp"
                 HEAP8[$538>>0] = 1; //@line 537 "load_it.cpp"
                 $539 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 4|0); //@line 537 "load_it.cpp"
                 HEAP8[$539>>0] = $533; //@line 537 "load_it.cpp"
                 $587 = 1;$589 = $533;
                } else {
                 $540 = ($533<<24>>24)<(0); //@line 539 "load_it.cpp"
                 $541 = ($533&255)<(193); //@line 539 "load_it.cpp"
                 $or$cond18 = $540 & $541; //@line 539 "load_it.cpp"
                 if ($or$cond18) {
                  $542 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 2|0); //@line 539 "load_it.cpp"
                  HEAP8[$542>>0] = 2; //@line 539 "load_it.cpp"
                  $543 = (($534) + 128)|0; //@line 539 "load_it.cpp"
                  $544 = $543&255; //@line 539 "load_it.cpp"
                  $545 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 4|0); //@line 539 "load_it.cpp"
                  HEAP8[$545>>0] = $544; //@line 539 "load_it.cpp"
                  $587 = 2;$589 = $544;
                  break;
                 }
                 $546 = ($533&255)<(75); //@line 541 "load_it.cpp"
                 if ($546) {
                  $547 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 2|0); //@line 541 "load_it.cpp"
                  HEAP8[$547>>0] = 5; //@line 541 "load_it.cpp"
                  $548 = (($534) + 191)|0; //@line 541 "load_it.cpp"
                  $549 = $548&255; //@line 541 "load_it.cpp"
                  $550 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 4|0); //@line 541 "load_it.cpp"
                  HEAP8[$550>>0] = $549; //@line 541 "load_it.cpp"
                  $587 = 5;$589 = $549;
                  break;
                 }
                 $551 = ($533&255)<(85); //@line 543 "load_it.cpp"
                 if ($551) {
                  $552 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 2|0); //@line 543 "load_it.cpp"
                  HEAP8[$552>>0] = 6; //@line 543 "load_it.cpp"
                  $553 = (($534) + 181)|0; //@line 543 "load_it.cpp"
                  $554 = $553&255; //@line 543 "load_it.cpp"
                  $555 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 4|0); //@line 543 "load_it.cpp"
                  HEAP8[$555>>0] = $554; //@line 543 "load_it.cpp"
                  $587 = 6;$589 = $554;
                  break;
                 }
                 $556 = ($533&255)<(95); //@line 545 "load_it.cpp"
                 if ($556) {
                  $557 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 2|0); //@line 545 "load_it.cpp"
                  HEAP8[$557>>0] = 3; //@line 545 "load_it.cpp"
                  $558 = (($534) + 171)|0; //@line 545 "load_it.cpp"
                  $559 = $558&255; //@line 545 "load_it.cpp"
                  $560 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 4|0); //@line 545 "load_it.cpp"
                  HEAP8[$560>>0] = $559; //@line 545 "load_it.cpp"
                  $587 = 3;$589 = $559;
                  break;
                 }
                 $561 = ($533&255)<(105); //@line 547 "load_it.cpp"
                 if ($561) {
                  $562 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 2|0); //@line 547 "load_it.cpp"
                  HEAP8[$562>>0] = 4; //@line 547 "load_it.cpp"
                  $563 = (($534) + 161)|0; //@line 547 "load_it.cpp"
                  $564 = $563&255; //@line 547 "load_it.cpp"
                  $565 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 4|0); //@line 547 "load_it.cpp"
                  HEAP8[$565>>0] = $564; //@line 547 "load_it.cpp"
                  $587 = 4;$589 = $564;
                  break;
                 }
                 $566 = ($533&255)<(115); //@line 549 "load_it.cpp"
                 if ($566) {
                  $567 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 2|0); //@line 549 "load_it.cpp"
                  HEAP8[$567>>0] = 13; //@line 549 "load_it.cpp"
                  $568 = (($534) + 151)|0; //@line 549 "load_it.cpp"
                  $569 = $568&255; //@line 549 "load_it.cpp"
                  $570 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 4|0); //@line 549 "load_it.cpp"
                  HEAP8[$570>>0] = $569; //@line 549 "load_it.cpp"
                  $587 = 13;$589 = $569;
                  break;
                 }
                 $571 = ($533&255)<(125); //@line 551 "load_it.cpp"
                 if ($571) {
                  $572 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 2|0); //@line 551 "load_it.cpp"
                  HEAP8[$572>>0] = 12; //@line 551 "load_it.cpp"
                  $573 = (($534) + 141)|0; //@line 551 "load_it.cpp"
                  $574 = $573&255; //@line 551 "load_it.cpp"
                  $575 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 4|0); //@line 551 "load_it.cpp"
                  HEAP8[$575>>0] = $574; //@line 551 "load_it.cpp"
                  $587 = 12;$589 = $574;
                  break;
                 }
                 $$off34 = (($533) + 63)<<24>>24; //@line 553 "load_it.cpp"
                 $576 = ($$off34&255)<(10); //@line 553 "load_it.cpp"
                 if ($576) {
                  $577 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 2|0); //@line 553 "load_it.cpp"
                  HEAP8[$577>>0] = 11; //@line 553 "load_it.cpp"
                  $578 = (($534) + 63)|0; //@line 553 "load_it.cpp"
                  $579 = $578&255; //@line 553 "load_it.cpp"
                  $580 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 4|0); //@line 553 "load_it.cpp"
                  HEAP8[$580>>0] = $579; //@line 553 "load_it.cpp"
                  $587 = 11;$589 = $579;
                  break;
                 }
                 $$off35 = (($533) + 53)<<24>>24; //@line 555 "load_it.cpp"
                 $581 = ($$off35&255)<(10); //@line 555 "load_it.cpp"
                 $582 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 2|0); //@line 555 "load_it.cpp"
                 if ($581) {
                  HEAP8[$582>>0] = 7; //@line 555 "load_it.cpp"
                  $583 = (($534) + 53)|0; //@line 555 "load_it.cpp"
                  $584 = $583&255; //@line 555 "load_it.cpp"
                  $585 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 4|0); //@line 555 "load_it.cpp"
                  HEAP8[$585>>0] = $584; //@line 555 "load_it.cpp"
                  $587 = 7;$589 = $584;
                  break;
                 } else {
                  $$pre162 = HEAP8[$582>>0]|0; //@line 556 "load_it.cpp"
                  $$phi$trans$insert163 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 4|0);
                  $$pre164 = HEAP8[$$phi$trans$insert163>>0]|0; //@line 557 "load_it.cpp"
                  $587 = $$pre162;$589 = $$pre164;
                  break;
                 }
                }
               } while(0);
               $586 = (((($lastvalue) + (($ch13$0*6)|0)|0)) + 2|0); //@line 556 "load_it.cpp"
               HEAP8[$586>>0] = $587; //@line 556 "load_it.cpp"
               $588 = (((($lastvalue) + (($ch13$0*6)|0)|0)) + 4|0); //@line 557 "load_it.cpp"
               HEAP8[$588>>0] = $589; //@line 557 "load_it.cpp"
               $$pre165 = HEAP8[$$pre$phiZ2D>>0]|0; //@line 561 "load_it.cpp"
               $591 = $$pre165;$i9$4 = $531;
              }
             } while(0);
             $590 = $591 & 8; //@line 561 "load_it.cpp"
             $592 = ($590<<24>>24)==(0); //@line 561 "load_it.cpp"
             do {
              if ($592) {
               $i9$0$be = $i9$4;
              } else {
               $593 = ($i9$4>>>0)>($440>>>0); //@line 563 "load_it.cpp"
               if ($593) {
                break L157;
               }
               $$sum31 = (($i9$4) + ($$sum29))|0; //@line 564 "load_it.cpp"
               $594 = (($lpStream) + ($$sum31)|0); //@line 564 "load_it.cpp"
               $595 = HEAP8[$594>>0]|0; //@line 564 "load_it.cpp"
               $596 = (($i9$4) + 2)|0; //@line 565 "load_it.cpp"
               $597 = HEAP32[$104>>2]|0; //@line 566 "load_it.cpp"
               $598 = ($ch13$0>>>0)<($597>>>0); //@line 566 "load_it.cpp"
               $599 = ($595<<24>>24)!=(0); //@line 568 "load_it.cpp"
               $or$cond25 = $599 & $598; //@line 566 "load_it.cpp"
               if (!($or$cond25)) {
                $i9$0$be = $596;
                break;
               }
               $$sum32 = (($443) + ($i9$4))|0; //@line 565 "load_it.cpp"
               $601 = (($lpStream) + ($$sum32)|0); //@line 565 "load_it.cpp"
               $602 = HEAP8[$601>>0]|0; //@line 565 "load_it.cpp"
               $603 = (($m$0$ph120) + (($ch13$0*6)|0)|0); //@line 570 "load_it.cpp"
               $604 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 3|0); //@line 570 "load_it.cpp"
               HEAP8[$604>>0] = $595; //@line 570 "load_it.cpp"
               $605 = (((($m$0$ph120) + (($ch13$0*6)|0)|0)) + 5|0); //@line 571 "load_it.cpp"
               HEAP8[$605>>0] = $602; //@line 571 "load_it.cpp"
               __ZNK10CSoundFile10S3MConvertEP11_MODCOMMANDb($this,$603,1); //@line 572 "load_it.cpp"
               $606 = HEAP8[$604>>0]|0; //@line 573 "load_it.cpp"
               $607 = (((($lastvalue) + (($ch13$0*6)|0)|0)) + 3|0); //@line 573 "load_it.cpp"
               HEAP8[$607>>0] = $606; //@line 573 "load_it.cpp"
               $608 = HEAP8[$605>>0]|0; //@line 574 "load_it.cpp"
               $609 = (((($lastvalue) + (($ch13$0*6)|0)|0)) + 5|0); //@line 574 "load_it.cpp"
               HEAP8[$609>>0] = $608; //@line 574 "load_it.cpp"
               $i9$0$be = $596;
              }
             } while(0);
             $600 = ($i9$0$be>>>0)<($426>>>0); //@line 475 "load_it.cpp"
             if ($600) {
              $i9$0113 = $i9$0$be;
             } else {
              break L157;
             }
            }
            $448 = (($nrow11$0$ph122) + 1)|0; //@line 479 "load_it.cpp"
            $449 = HEAP32[$104>>2]|0; //@line 480 "load_it.cpp"
            $450 = (($m$0$ph120) + (($449*6)|0)|0); //@line 480 "load_it.cpp"
            $451 = ($448>>>0)<($429>>>0); //@line 473 "load_it.cpp"
            $452 = ($$lcssa>>>0)<($426>>>0); //@line 475 "load_it.cpp"
            $or$cond96112 = $452 & $451; //@line 473 "load_it.cpp"
            if ($or$cond96112) {
             $i9$0$ph121 = $$lcssa;$m$0$ph120 = $450;$nrow11$0$ph122 = $448;
            } else {
             break;
            }
           }
          } else {
           $420 = (((($this)) + 52864|0) + ($npat$0123<<1)|0); //@line 456 "load_it.cpp"
           HEAP16[$420>>1] = 64; //@line 456 "load_it.cpp"
           $421 = HEAP32[$104>>2]|0; //@line 457 "load_it.cpp"
           $422 = (__ZN10CSoundFile15AllocatePatternEjj(64,$421)|0); //@line 457 "load_it.cpp"
           $423 = (((($this)) + 51904|0) + ($npat$0123<<2)|0); //@line 457 "load_it.cpp"
           HEAP32[$423>>2] = $422; //@line 457 "load_it.cpp"
          }
         } while(0);
         $610 = (($npat$0123) + 1)|0; //@line 452 "load_it.cpp"
         $exitcond142 = ($610|0)==($272|0); //@line 452 "load_it.cpp"
         if ($exitcond142) {
          $ncu$0111 = 0;
          break;
         } else {
          $npat$0123 = $610;
         }
        }
       }
       while(1) {
        $611 = HEAP32[$104>>2]|0; //@line 582 "load_it.cpp"
        $612 = ($ncu$0111>>>0)<($611>>>0); //@line 582 "load_it.cpp"
        if (!($612)) {
         $613 = (((((($this)) + 49600|0) + (($ncu$0111*36)|0)|0)) + 4|0); //@line 584 "load_it.cpp"
         HEAP32[$613>>2] = 64; //@line 584 "load_it.cpp"
         $614 = (((((($this)) + 49600|0) + (($ncu$0111*36)|0)|0)) + 8|0); //@line 585 "load_it.cpp"
         $615 = HEAP32[$614>>2]|0; //@line 585 "load_it.cpp"
         $616 = $615 & -257; //@line 585 "load_it.cpp"
         HEAP32[$614>>2] = $616; //@line 585 "load_it.cpp"
        }
        $617 = (($ncu$0111) + 1)|0; //@line 580 "load_it.cpp"
        $exitcond = ($617|0)==(64); //@line 580 "load_it.cpp"
        if ($exitcond) {
         break;
        } else {
         $ncu$0111 = $617;
        }
       }
       $618 = ((($this)) + 59764|0); //@line 588 "load_it.cpp"
       HEAP32[$618>>2] = 8; //@line 588 "load_it.cpp"
       $619 = ((($this)) + 59768|0); //@line 589 "load_it.cpp"
       HEAP32[$619>>2] = 61440; //@line 589 "load_it.cpp"
       $$0 = 1;
      }
     }
    }
   }
  }
 } else {
  $$0 = 0;
 }
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0); //@line 591 "load_it.cpp"
}
function __ZN10CSoundFile12ITInstrToMPTEPKvP17_INSTRUMENTHEADERj($this,$p,$penv,$trkvers) {
 $this = $this|0;
 $p = $p|0;
 $penv = $penv|0;
 $trkvers = $trkvers|0;
 var $$ = 0, $$19 = 0, $$20 = 0, $$21 = 0, $$22 = 0, $$pre$phiZ2D = 0, $$sum = 0, $$sum1 = 0, $$sum12 = 0, $$sum13 = 0, $$sum15 = 0, $$sum16 = 0, $$sum17 = 0, $$sum18 = 0, $$sum2 = 0, $$sum3 = 0, $$sum4 = 0, $$sum5 = 0, $$sum6 = 0, $$sum7 = 0;
 var $$sum8 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0;
 var $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0;
 var $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0;
 var $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0;
 var $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0;
 var $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0;
 var $331 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $ev$024 = 0, $ev$024$lcssa = 0, $ev5$026 = 0, $exitcond = 0, $exitcond29 = 0, $exitcond30 = 0, $j$025 = 0;
 var $j2$027 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ($trkvers>>>0)<(512); //@line 41 "load_it.cpp"
 $1 = ((($penv)) + 585|0); //@line 44 "load_it.cpp"
 $2 = ((($p)) + 32|0); //@line 44 "load_it.cpp"
 dest=$1; src=$2; stop=dest+26|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 44 "load_it.cpp"
 $3 = ((($penv)) + 617|0); //@line 45 "load_it.cpp"
 $4 = ((($p)) + 4|0); //@line 45 "load_it.cpp"
 dest=$3; src=$4; stop=dest+12|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 45 "load_it.cpp"
 if ($0) {
  $5 = ((($p)) + 24|0); //@line 46 "load_it.cpp"
  $6 = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8); //@line 46 "load_it.cpp"
  $7 = $6&65535; //@line 46 "load_it.cpp"
  $8 = $7 << 6; //@line 46 "load_it.cpp"
  HEAP32[$penv>>2] = $8; //@line 46 "load_it.cpp"
  $9 = ((($penv)) + 8|0); //@line 47 "load_it.cpp"
  HEAP16[$9>>1] = 64; //@line 47 "load_it.cpp"
  $j$025 = 0;
  while(1) {
   $10 = $j$025 << 1; //@line 50 "load_it.cpp"
   $$sum17 = (($10) + 64)|0; //@line 50 "load_it.cpp"
   $11 = (($p) + ($$sum17)|0); //@line 50 "load_it.cpp"
   $12 = HEAP8[$11>>0]|0; //@line 50 "load_it.cpp"
   $13 = $12&255; //@line 50 "load_it.cpp"
   $14 = $10 | 1; //@line 51 "load_it.cpp"
   $$sum18 = (($14) + 64)|0; //@line 51 "load_it.cpp"
   $15 = (($p) + ($$sum18)|0); //@line 51 "load_it.cpp"
   $16 = HEAP8[$15>>0]|0; //@line 51 "load_it.cpp"
   $17 = ($16&255)<(240); //@line 52 "load_it.cpp"
   if ($17) {
    $18 = (((($penv)) + 300|0) + ($j$025)|0); //@line 52 "load_it.cpp"
    HEAP8[$18>>0] = $16; //@line 52 "load_it.cpp"
   }
   $19 = ($12<<24>>24)>(-1); //@line 53 "load_it.cpp"
   if ($19) {
    $20 = (($13) + 1)|0; //@line 53 "load_it.cpp"
    $21 = $20&255; //@line 53 "load_it.cpp"
    $22 = (((($penv)) + 428|0) + ($j$025)|0); //@line 53 "load_it.cpp"
    HEAP8[$22>>0] = $21; //@line 53 "load_it.cpp"
   } else {
    $23 = ($12&255)>(253); //@line 54 "load_it.cpp"
    if ($23) {
     $24 = (((($penv)) + 428|0) + ($j$025)|0); //@line 54 "load_it.cpp"
     HEAP8[$24>>0] = $12; //@line 54 "load_it.cpp"
    }
   }
   $25 = (($j$025) + 1)|0; //@line 48 "load_it.cpp"
   $exitcond = ($25|0)==(120); //@line 48 "load_it.cpp"
   if ($exitcond) {
    break;
   } else {
    $j$025 = $25;
   }
  }
  $26 = ((($p)) + 17|0); //@line 56 "load_it.cpp"
  $27 = HEAP8[$26>>0]|0; //@line 56 "load_it.cpp"
  $28 = $27 & 1; //@line 56 "load_it.cpp"
  $29 = ($28<<24>>24)==(0); //@line 56 "load_it.cpp"
  if (!($29)) {
   $30 = ((($penv)) + 4|0); //@line 56 "load_it.cpp"
   $31 = HEAP32[$30>>2]|0; //@line 56 "load_it.cpp"
   $32 = $31 | 1; //@line 56 "load_it.cpp"
   HEAP32[$30>>2] = $32; //@line 56 "load_it.cpp"
  }
  $33 = $27 & 2; //@line 57 "load_it.cpp"
  $34 = ($33<<24>>24)==(0); //@line 57 "load_it.cpp"
  if (!($34)) {
   $35 = ((($penv)) + 4|0); //@line 57 "load_it.cpp"
   $36 = HEAP32[$35>>2]|0; //@line 57 "load_it.cpp"
   $37 = $36 | 4; //@line 57 "load_it.cpp"
   HEAP32[$35>>2] = $37; //@line 57 "load_it.cpp"
  }
  $38 = $27 & 4; //@line 58 "load_it.cpp"
  $39 = ($38<<24>>24)==(0); //@line 58 "load_it.cpp"
  if (!($39)) {
   $40 = ((($penv)) + 4|0); //@line 58 "load_it.cpp"
   $41 = HEAP32[$40>>2]|0; //@line 58 "load_it.cpp"
   $42 = $41 | 2; //@line 58 "load_it.cpp"
   HEAP32[$40>>2] = $42; //@line 58 "load_it.cpp"
  }
  $43 = ((($p)) + 18|0); //@line 59 "load_it.cpp"
  $44 = HEAP8[$43>>0]|0; //@line 59 "load_it.cpp"
  $45 = ((($penv)) + 559|0); //@line 59 "load_it.cpp"
  HEAP8[$45>>0] = $44; //@line 59 "load_it.cpp"
  $46 = ((($p)) + 19|0); //@line 60 "load_it.cpp"
  $47 = HEAP8[$46>>0]|0; //@line 60 "load_it.cpp"
  $48 = ((($penv)) + 560|0); //@line 60 "load_it.cpp"
  HEAP8[$48>>0] = $47; //@line 60 "load_it.cpp"
  $49 = ((($p)) + 20|0); //@line 61 "load_it.cpp"
  $50 = HEAP8[$49>>0]|0; //@line 61 "load_it.cpp"
  $51 = ((($penv)) + 561|0); //@line 61 "load_it.cpp"
  HEAP8[$51>>0] = $50; //@line 61 "load_it.cpp"
  $52 = ((($p)) + 21|0); //@line 62 "load_it.cpp"
  $53 = HEAP8[$52>>0]|0; //@line 62 "load_it.cpp"
  $54 = ((($penv)) + 562|0); //@line 62 "load_it.cpp"
  HEAP8[$54>>0] = $53; //@line 62 "load_it.cpp"
  $55 = ((($penv)) + 556|0); //@line 63 "load_it.cpp"
  HEAP8[$55>>0] = 25; //@line 63 "load_it.cpp"
  $ev$024 = 0;
  while(1) {
   $56 = $ev$024 << 1; //@line 66 "load_it.cpp"
   $$sum15 = (($56) + 504)|0; //@line 66 "load_it.cpp"
   $57 = (($p) + ($$sum15)|0); //@line 66 "load_it.cpp"
   $58 = HEAP8[$57>>0]|0; //@line 66 "load_it.cpp"
   $59 = $58&255; //@line 66 "load_it.cpp"
   $60 = (((($penv)) + 12|0) + ($ev$024<<1)|0); //@line 66 "load_it.cpp"
   HEAP16[$60>>1] = $59; //@line 66 "load_it.cpp"
   $61 = ($58<<24>>24)==(-1); //@line 66 "load_it.cpp"
   if ($61) {
    $ev$024$lcssa = $ev$024;
    label = 18;
    break;
   }
   $63 = $56 | 1; //@line 71 "load_it.cpp"
   $$sum16 = (($63) + 504)|0; //@line 71 "load_it.cpp"
   $64 = (($p) + ($$sum16)|0); //@line 71 "load_it.cpp"
   $65 = HEAP8[$64>>0]|0; //@line 71 "load_it.cpp"
   $66 = (((($penv)) + 204|0) + ($ev$024)|0); //@line 71 "load_it.cpp"
   HEAP8[$66>>0] = $65; //@line 71 "load_it.cpp"
   $67 = (($ev$024) + 1)|0; //@line 64 "load_it.cpp"
   $68 = ($67>>>0)<(25); //@line 64 "load_it.cpp"
   if ($68) {
    $ev$024 = $67;
   } else {
    break;
   }
  }
  if ((label|0) == 18) {
   $62 = $ev$024$lcssa&255; //@line 68 "load_it.cpp"
   HEAP8[$55>>0] = $62; //@line 68 "load_it.cpp"
  }
  $69 = ((($p)) + 26|0); //@line 73 "load_it.cpp"
  $70 = HEAP8[$69>>0]|0; //@line 73 "load_it.cpp"
  $71 = ((($penv)) + 571|0); //@line 73 "load_it.cpp"
  HEAP8[$71>>0] = $70; //@line 73 "load_it.cpp"
  $72 = ((($p)) + 27|0); //@line 74 "load_it.cpp"
  $73 = HEAP8[$72>>0]|0; //@line 74 "load_it.cpp"
  $74 = ((($penv)) + 572|0); //@line 74 "load_it.cpp"
  HEAP8[$74>>0] = $73; //@line 74 "load_it.cpp"
  $75 = ((($penv)) + 10|0); //@line 75 "load_it.cpp"
  HEAP16[$75>>1] = 128; //@line 75 "load_it.cpp"
  $$pre$phiZ2D = $45;
 } else {
  $76 = ((($p)) + 61|0); //@line 81 "load_it.cpp"
  $77 = HEAP8[$76>>0]|0; //@line 81 "load_it.cpp"
  $78 = ((($penv)) + 580|0); //@line 81 "load_it.cpp"
  HEAP8[$78>>0] = $77; //@line 81 "load_it.cpp"
  $79 = ((($p)) + 60|0); //@line 82 "load_it.cpp"
  $80 = HEAP8[$79>>0]|0; //@line 82 "load_it.cpp"
  $81 = ((($penv)) + 581|0); //@line 82 "load_it.cpp"
  HEAP8[$81>>0] = $80; //@line 82 "load_it.cpp"
  $82 = ((($p)) + 62|0); //@line 83 "load_it.cpp"
  $83 = HEAPU8[$82>>0]|(HEAPU8[$82+1>>0]<<8); //@line 83 "load_it.cpp"
  $84 = ((($penv)) + 578|0); //@line 83 "load_it.cpp"
  HEAP16[$84>>1] = $83; //@line 83 "load_it.cpp"
  $85 = ((($p)) + 20|0); //@line 84 "load_it.cpp"
  $86 = HEAPU8[$85>>0]|(HEAPU8[$85+1>>0]<<8); //@line 84 "load_it.cpp"
  $87 = $86&65535; //@line 84 "load_it.cpp"
  $88 = $87 << 5; //@line 84 "load_it.cpp"
  HEAP32[$penv>>2] = $88; //@line 84 "load_it.cpp"
  $89 = ((($p)) + 24|0); //@line 85 "load_it.cpp"
  $90 = HEAP8[$89>>0]|0; //@line 85 "load_it.cpp"
  $91 = ($90&255) >>> 1; //@line 85 "load_it.cpp"
  $92 = $91&255; //@line 85 "load_it.cpp"
  $93 = ((($penv)) + 8|0); //@line 85 "load_it.cpp"
  $94 = ($90&255)>(129); //@line 86 "load_it.cpp"
  $$ = $94 ? 64 : $92; //@line 86 "load_it.cpp"
  HEAP16[$93>>1] = $$; //@line 85 "load_it.cpp"
  $j2$027 = 0;
  while(1) {
   $95 = $j2$027 << 1; //@line 89 "load_it.cpp"
   $$sum12 = (($95) + 64)|0; //@line 89 "load_it.cpp"
   $96 = (($p) + ($$sum12)|0); //@line 89 "load_it.cpp"
   $97 = HEAP8[$96>>0]|0; //@line 89 "load_it.cpp"
   $98 = $97&255; //@line 89 "load_it.cpp"
   $99 = $95 | 1; //@line 90 "load_it.cpp"
   $$sum13 = (($99) + 64)|0; //@line 90 "load_it.cpp"
   $100 = (($p) + ($$sum13)|0); //@line 90 "load_it.cpp"
   $101 = HEAP8[$100>>0]|0; //@line 90 "load_it.cpp"
   $102 = ($101&255)<(240); //@line 91 "load_it.cpp"
   if ($102) {
    $103 = (((($penv)) + 300|0) + ($j2$027)|0); //@line 91 "load_it.cpp"
    HEAP8[$103>>0] = $101; //@line 91 "load_it.cpp"
   }
   $104 = ($97<<24>>24)>(-1); //@line 92 "load_it.cpp"
   if ($104) {
    $105 = (($98) + 1)|0; //@line 92 "load_it.cpp"
    $106 = $105&255; //@line 92 "load_it.cpp"
    $107 = (((($penv)) + 428|0) + ($j2$027)|0); //@line 92 "load_it.cpp"
    HEAP8[$107>>0] = $106; //@line 92 "load_it.cpp"
   } else {
    $108 = ($97&255)>(253); //@line 93 "load_it.cpp"
    if ($108) {
     $109 = (((($penv)) + 428|0) + ($j2$027)|0); //@line 93 "load_it.cpp"
     HEAP8[$109>>0] = $97; //@line 93 "load_it.cpp"
    }
   }
   $110 = (($j2$027) + 1)|0; //@line 87 "load_it.cpp"
   $exitcond30 = ($110|0)==(120); //@line 87 "load_it.cpp"
   if ($exitcond30) {
    break;
   } else {
    $j2$027 = $110;
   }
  }
  $111 = ((($p)) + 304|0); //@line 96 "load_it.cpp"
  $112 = HEAP8[$111>>0]|0; //@line 96 "load_it.cpp"
  $113 = $112 & 1; //@line 96 "load_it.cpp"
  $114 = ($113<<24>>24)==(0); //@line 96 "load_it.cpp"
  if (!($114)) {
   $115 = ((($penv)) + 4|0); //@line 96 "load_it.cpp"
   $116 = HEAP32[$115>>2]|0; //@line 96 "load_it.cpp"
   $117 = $116 | 1; //@line 96 "load_it.cpp"
   HEAP32[$115>>2] = $117; //@line 96 "load_it.cpp"
  }
  $118 = $112 & 2; //@line 97 "load_it.cpp"
  $119 = ($118<<24>>24)==(0); //@line 97 "load_it.cpp"
  if (!($119)) {
   $120 = ((($penv)) + 4|0); //@line 97 "load_it.cpp"
   $121 = HEAP32[$120>>2]|0; //@line 97 "load_it.cpp"
   $122 = $121 | 4; //@line 97 "load_it.cpp"
   HEAP32[$120>>2] = $122; //@line 97 "load_it.cpp"
  }
  $123 = $112 & 4; //@line 98 "load_it.cpp"
  $124 = ($123<<24>>24)==(0); //@line 98 "load_it.cpp"
  if (!($124)) {
   $125 = ((($penv)) + 4|0); //@line 98 "load_it.cpp"
   $126 = HEAP32[$125>>2]|0; //@line 98 "load_it.cpp"
   $127 = $126 | 2; //@line 98 "load_it.cpp"
   HEAP32[$125>>2] = $127; //@line 98 "load_it.cpp"
  }
  $128 = $112 & 8; //@line 99 "load_it.cpp"
  $129 = ($128<<24>>24)==(0); //@line 99 "load_it.cpp"
  if (!($129)) {
   $130 = ((($penv)) + 4|0); //@line 99 "load_it.cpp"
   $131 = HEAP32[$130>>2]|0; //@line 99 "load_it.cpp"
   $132 = $131 | 2048; //@line 99 "load_it.cpp"
   HEAP32[$130>>2] = $132; //@line 99 "load_it.cpp"
  }
  $133 = ((($p)) + 305|0); //@line 100 "load_it.cpp"
  $134 = HEAP8[$133>>0]|0; //@line 100 "load_it.cpp"
  $135 = ((($penv)) + 556|0); //@line 100 "load_it.cpp"
  $136 = ($134&255)>(25); //@line 101 "load_it.cpp"
  $$19 = $136 ? 25 : $134; //@line 101 "load_it.cpp"
  HEAP8[$135>>0] = $$19; //@line 100 "load_it.cpp"
  $137 = ((($p)) + 306|0); //@line 103 "load_it.cpp"
  $138 = HEAP8[$137>>0]|0; //@line 103 "load_it.cpp"
  $139 = ((($penv)) + 559|0); //@line 103 "load_it.cpp"
  HEAP8[$139>>0] = $138; //@line 103 "load_it.cpp"
  $140 = ((($p)) + 307|0); //@line 104 "load_it.cpp"
  $141 = HEAP8[$140>>0]|0; //@line 104 "load_it.cpp"
  $142 = ((($penv)) + 560|0); //@line 104 "load_it.cpp"
  HEAP8[$142>>0] = $141; //@line 104 "load_it.cpp"
  $143 = ((($p)) + 308|0); //@line 105 "load_it.cpp"
  $144 = HEAP8[$143>>0]|0; //@line 105 "load_it.cpp"
  $145 = ((($penv)) + 561|0); //@line 105 "load_it.cpp"
  HEAP8[$145>>0] = $144; //@line 105 "load_it.cpp"
  $146 = ((($p)) + 309|0); //@line 106 "load_it.cpp"
  $147 = HEAP8[$146>>0]|0; //@line 106 "load_it.cpp"
  $148 = ((($penv)) + 562|0); //@line 106 "load_it.cpp"
  HEAP8[$148>>0] = $147; //@line 106 "load_it.cpp"
  $149 = ((($p)) + 386|0); //@line 108 "load_it.cpp"
  $150 = HEAP8[$149>>0]|0; //@line 108 "load_it.cpp"
  $151 = $150 & 1; //@line 108 "load_it.cpp"
  $152 = ($151<<24>>24)==(0); //@line 108 "load_it.cpp"
  if (!($152)) {
   $153 = ((($penv)) + 4|0); //@line 108 "load_it.cpp"
   $154 = HEAP32[$153>>2]|0; //@line 108 "load_it.cpp"
   $155 = $154 | 8; //@line 108 "load_it.cpp"
   HEAP32[$153>>2] = $155; //@line 108 "load_it.cpp"
  }
  $156 = $150 & 2; //@line 109 "load_it.cpp"
  $157 = ($156<<24>>24)==(0); //@line 109 "load_it.cpp"
  if (!($157)) {
   $158 = ((($penv)) + 4|0); //@line 109 "load_it.cpp"
   $159 = HEAP32[$158>>2]|0; //@line 109 "load_it.cpp"
   $160 = $159 | 32; //@line 109 "load_it.cpp"
   HEAP32[$158>>2] = $160; //@line 109 "load_it.cpp"
  }
  $161 = $150 & 4; //@line 110 "load_it.cpp"
  $162 = ($161<<24>>24)==(0); //@line 110 "load_it.cpp"
  if (!($162)) {
   $163 = ((($penv)) + 4|0); //@line 110 "load_it.cpp"
   $164 = HEAP32[$163>>2]|0; //@line 110 "load_it.cpp"
   $165 = $164 | 16; //@line 110 "load_it.cpp"
   HEAP32[$163>>2] = $165; //@line 110 "load_it.cpp"
  }
  $166 = $150 & 8; //@line 111 "load_it.cpp"
  $167 = ($166<<24>>24)==(0); //@line 111 "load_it.cpp"
  if (!($167)) {
   $168 = ((($penv)) + 4|0); //@line 111 "load_it.cpp"
   $169 = HEAP32[$168>>2]|0; //@line 111 "load_it.cpp"
   $170 = $169 | 4096; //@line 111 "load_it.cpp"
   HEAP32[$168>>2] = $170; //@line 111 "load_it.cpp"
  }
  $171 = ((($p)) + 387|0); //@line 112 "load_it.cpp"
  $172 = HEAP8[$171>>0]|0; //@line 112 "load_it.cpp"
  $173 = ((($penv)) + 557|0); //@line 112 "load_it.cpp"
  $174 = ($172&255)>(25); //@line 113 "load_it.cpp"
  $$20 = $174 ? 25 : $172; //@line 113 "load_it.cpp"
  HEAP8[$173>>0] = $$20; //@line 112 "load_it.cpp"
  $175 = ((($p)) + 388|0); //@line 114 "load_it.cpp"
  $176 = HEAP8[$175>>0]|0; //@line 114 "load_it.cpp"
  $177 = ((($penv)) + 563|0); //@line 114 "load_it.cpp"
  HEAP8[$177>>0] = $176; //@line 114 "load_it.cpp"
  $178 = ((($p)) + 389|0); //@line 115 "load_it.cpp"
  $179 = HEAP8[$178>>0]|0; //@line 115 "load_it.cpp"
  $180 = ((($penv)) + 564|0); //@line 115 "load_it.cpp"
  HEAP8[$180>>0] = $179; //@line 115 "load_it.cpp"
  $181 = ((($p)) + 390|0); //@line 116 "load_it.cpp"
  $182 = HEAP8[$181>>0]|0; //@line 116 "load_it.cpp"
  $183 = ((($penv)) + 565|0); //@line 116 "load_it.cpp"
  HEAP8[$183>>0] = $182; //@line 116 "load_it.cpp"
  $184 = ((($p)) + 391|0); //@line 117 "load_it.cpp"
  $185 = HEAP8[$184>>0]|0; //@line 117 "load_it.cpp"
  $186 = ((($penv)) + 566|0); //@line 117 "load_it.cpp"
  HEAP8[$186>>0] = $185; //@line 117 "load_it.cpp"
  $187 = ((($p)) + 468|0); //@line 119 "load_it.cpp"
  $188 = HEAP8[$187>>0]|0; //@line 119 "load_it.cpp"
  $189 = $188 & 1; //@line 119 "load_it.cpp"
  $190 = ($189<<24>>24)==(0); //@line 119 "load_it.cpp"
  if (!($190)) {
   $191 = ((($penv)) + 4|0); //@line 119 "load_it.cpp"
   $192 = HEAP32[$191>>2]|0; //@line 119 "load_it.cpp"
   $193 = $192 | 64; //@line 119 "load_it.cpp"
   HEAP32[$191>>2] = $193; //@line 119 "load_it.cpp"
  }
  $194 = $188 & 2; //@line 120 "load_it.cpp"
  $195 = ($194<<24>>24)==(0); //@line 120 "load_it.cpp"
  if (!($195)) {
   $196 = ((($penv)) + 4|0); //@line 120 "load_it.cpp"
   $197 = HEAP32[$196>>2]|0; //@line 120 "load_it.cpp"
   $198 = $197 | 256; //@line 120 "load_it.cpp"
   HEAP32[$196>>2] = $198; //@line 120 "load_it.cpp"
  }
  $199 = $188 & 4; //@line 121 "load_it.cpp"
  $200 = ($199<<24>>24)==(0); //@line 121 "load_it.cpp"
  if (!($200)) {
   $201 = ((($penv)) + 4|0); //@line 121 "load_it.cpp"
   $202 = HEAP32[$201>>2]|0; //@line 121 "load_it.cpp"
   $203 = $202 | 128; //@line 121 "load_it.cpp"
   HEAP32[$201>>2] = $203; //@line 121 "load_it.cpp"
  }
  $204 = $188 & 8; //@line 122 "load_it.cpp"
  $205 = ($204<<24>>24)==(0); //@line 122 "load_it.cpp"
  if (!($205)) {
   $206 = ((($penv)) + 4|0); //@line 122 "load_it.cpp"
   $207 = HEAP32[$206>>2]|0; //@line 122 "load_it.cpp"
   $208 = $207 | 8192; //@line 122 "load_it.cpp"
   HEAP32[$206>>2] = $208; //@line 122 "load_it.cpp"
  }
  $209 = ($188<<24>>24)<(0); //@line 123 "load_it.cpp"
  if ($209) {
   $210 = ((($penv)) + 4|0); //@line 123 "load_it.cpp"
   $211 = HEAP32[$210>>2]|0; //@line 123 "load_it.cpp"
   $212 = $211 | 1024; //@line 123 "load_it.cpp"
   HEAP32[$210>>2] = $212; //@line 123 "load_it.cpp"
  }
  $213 = ((($p)) + 469|0); //@line 124 "load_it.cpp"
  $214 = HEAP8[$213>>0]|0; //@line 124 "load_it.cpp"
  $215 = ((($penv)) + 558|0); //@line 124 "load_it.cpp"
  $216 = ($214&255)>(25); //@line 125 "load_it.cpp"
  $$21 = $216 ? 25 : $214; //@line 125 "load_it.cpp"
  HEAP8[$215>>0] = $$21; //@line 124 "load_it.cpp"
  $217 = ((($p)) + 470|0); //@line 126 "load_it.cpp"
  $218 = HEAP8[$217>>0]|0; //@line 126 "load_it.cpp"
  $219 = ((($penv)) + 567|0); //@line 126 "load_it.cpp"
  HEAP8[$219>>0] = $218; //@line 126 "load_it.cpp"
  $220 = ((($p)) + 471|0); //@line 127 "load_it.cpp"
  $221 = HEAP8[$220>>0]|0; //@line 127 "load_it.cpp"
  $222 = ((($penv)) + 568|0); //@line 127 "load_it.cpp"
  HEAP8[$222>>0] = $221; //@line 127 "load_it.cpp"
  $223 = ((($p)) + 472|0); //@line 128 "load_it.cpp"
  $224 = HEAP8[$223>>0]|0; //@line 128 "load_it.cpp"
  $225 = ((($penv)) + 569|0); //@line 128 "load_it.cpp"
  HEAP8[$225>>0] = $224; //@line 128 "load_it.cpp"
  $226 = ((($p)) + 473|0); //@line 129 "load_it.cpp"
  $227 = HEAP8[$226>>0]|0; //@line 129 "load_it.cpp"
  $228 = ((($penv)) + 570|0); //@line 129 "load_it.cpp"
  HEAP8[$228>>0] = $227; //@line 129 "load_it.cpp"
  $ev5$026 = 0;
  while(1) {
   $229 = ($ev5$026*3)|0; //@line 133 "load_it.cpp"
   $$sum = (($229) + 310)|0; //@line 133 "load_it.cpp"
   $230 = (($p) + ($$sum)|0); //@line 133 "load_it.cpp"
   $231 = HEAP8[$230>>0]|0; //@line 133 "load_it.cpp"
   $232 = (((($penv)) + 204|0) + ($ev5$026)|0); //@line 133 "load_it.cpp"
   HEAP8[$232>>0] = $231; //@line 133 "load_it.cpp"
   $$sum1 = (($229) + 312)|0; //@line 134 "load_it.cpp"
   $233 = (($p) + ($$sum1)|0); //@line 134 "load_it.cpp"
   $234 = HEAP8[$233>>0]|0; //@line 134 "load_it.cpp"
   $235 = $234&255; //@line 134 "load_it.cpp"
   $236 = $235 << 8; //@line 134 "load_it.cpp"
   $$sum2 = (($229) + 311)|0; //@line 134 "load_it.cpp"
   $237 = (($p) + ($$sum2)|0); //@line 134 "load_it.cpp"
   $238 = HEAP8[$237>>0]|0; //@line 134 "load_it.cpp"
   $239 = $238&255; //@line 134 "load_it.cpp"
   $240 = $236 | $239; //@line 134 "load_it.cpp"
   $241 = $240&65535; //@line 134 "load_it.cpp"
   $242 = (((($penv)) + 12|0) + ($ev5$026<<1)|0); //@line 134 "load_it.cpp"
   HEAP16[$242>>1] = $241; //@line 134 "load_it.cpp"
   $$sum3 = (($229) + 392)|0; //@line 135 "load_it.cpp"
   $243 = (($p) + ($$sum3)|0); //@line 135 "load_it.cpp"
   $244 = HEAP8[$243>>0]|0; //@line 135 "load_it.cpp"
   $245 = $244&255; //@line 135 "load_it.cpp"
   $246 = (($245) + 32)|0; //@line 135 "load_it.cpp"
   $247 = $246&255; //@line 135 "load_it.cpp"
   $248 = (((($penv)) + 236|0) + ($ev5$026)|0); //@line 135 "load_it.cpp"
   HEAP8[$248>>0] = $247; //@line 135 "load_it.cpp"
   $$sum4 = (($229) + 394)|0; //@line 136 "load_it.cpp"
   $249 = (($p) + ($$sum4)|0); //@line 136 "load_it.cpp"
   $250 = HEAP8[$249>>0]|0; //@line 136 "load_it.cpp"
   $251 = $250&255; //@line 136 "load_it.cpp"
   $252 = $251 << 8; //@line 136 "load_it.cpp"
   $$sum5 = (($229) + 393)|0; //@line 136 "load_it.cpp"
   $253 = (($p) + ($$sum5)|0); //@line 136 "load_it.cpp"
   $254 = HEAP8[$253>>0]|0; //@line 136 "load_it.cpp"
   $255 = $254&255; //@line 136 "load_it.cpp"
   $256 = $252 | $255; //@line 136 "load_it.cpp"
   $257 = $256&65535; //@line 136 "load_it.cpp"
   $258 = (((($penv)) + 76|0) + ($ev5$026<<1)|0); //@line 136 "load_it.cpp"
   HEAP16[$258>>1] = $257; //@line 136 "load_it.cpp"
   $$sum6 = (($229) + 474)|0; //@line 137 "load_it.cpp"
   $259 = (($p) + ($$sum6)|0); //@line 137 "load_it.cpp"
   $260 = HEAP8[$259>>0]|0; //@line 137 "load_it.cpp"
   $261 = $260&255; //@line 137 "load_it.cpp"
   $262 = (($261) + 32)|0; //@line 137 "load_it.cpp"
   $263 = $262&255; //@line 137 "load_it.cpp"
   $264 = (((($penv)) + 268|0) + ($ev5$026)|0); //@line 137 "load_it.cpp"
   HEAP8[$264>>0] = $263; //@line 137 "load_it.cpp"
   $$sum7 = (($229) + 476)|0; //@line 138 "load_it.cpp"
   $265 = (($p) + ($$sum7)|0); //@line 138 "load_it.cpp"
   $266 = HEAP8[$265>>0]|0; //@line 138 "load_it.cpp"
   $267 = $266&255; //@line 138 "load_it.cpp"
   $268 = $267 << 8; //@line 138 "load_it.cpp"
   $$sum8 = (($229) + 475)|0; //@line 138 "load_it.cpp"
   $269 = (($p) + ($$sum8)|0); //@line 138 "load_it.cpp"
   $270 = HEAP8[$269>>0]|0; //@line 138 "load_it.cpp"
   $271 = $270&255; //@line 138 "load_it.cpp"
   $272 = $268 | $271; //@line 138 "load_it.cpp"
   $273 = $272&65535; //@line 138 "load_it.cpp"
   $274 = (((($penv)) + 140|0) + ($ev5$026<<1)|0); //@line 138 "load_it.cpp"
   HEAP16[$274>>1] = $273; //@line 138 "load_it.cpp"
   $275 = (($ev5$026) + 1)|0; //@line 131 "load_it.cpp"
   $exitcond29 = ($275|0)==(25); //@line 131 "load_it.cpp"
   if ($exitcond29) {
    break;
   } else {
    $ev5$026 = $275;
   }
  }
  $276 = ((($p)) + 17|0); //@line 140 "load_it.cpp"
  $277 = HEAP8[$276>>0]|0; //@line 140 "load_it.cpp"
  $278 = ((($penv)) + 571|0); //@line 140 "load_it.cpp"
  HEAP8[$278>>0] = $277; //@line 140 "load_it.cpp"
  $279 = ((($p)) + 18|0); //@line 141 "load_it.cpp"
  $280 = HEAP8[$279>>0]|0; //@line 141 "load_it.cpp"
  $281 = ((($penv)) + 572|0); //@line 141 "load_it.cpp"
  HEAP8[$281>>0] = $280; //@line 141 "load_it.cpp"
  $282 = ((($p)) + 19|0); //@line 142 "load_it.cpp"
  $283 = HEAP8[$282>>0]|0; //@line 142 "load_it.cpp"
  $284 = ((($penv)) + 573|0); //@line 142 "load_it.cpp"
  HEAP8[$284>>0] = $283; //@line 142 "load_it.cpp"
  $285 = ((($p)) + 22|0); //@line 143 "load_it.cpp"
  $286 = HEAP8[$285>>0]|0; //@line 143 "load_it.cpp"
  $287 = ((($penv)) + 583|0); //@line 143 "load_it.cpp"
  HEAP8[$287>>0] = $286; //@line 143 "load_it.cpp"
  $288 = ((($p)) + 23|0); //@line 144 "load_it.cpp"
  $289 = HEAP8[$288>>0]|0; //@line 144 "load_it.cpp"
  $290 = ((($penv)) + 584|0); //@line 144 "load_it.cpp"
  HEAP8[$290>>0] = $289; //@line 144 "load_it.cpp"
  $291 = ((($p)) + 58|0); //@line 145 "load_it.cpp"
  $292 = HEAP8[$291>>0]|0; //@line 145 "load_it.cpp"
  $293 = ((($penv)) + 576|0); //@line 145 "load_it.cpp"
  HEAP8[$293>>0] = $292; //@line 145 "load_it.cpp"
  $294 = ((($p)) + 59|0); //@line 146 "load_it.cpp"
  $295 = HEAP8[$294>>0]|0; //@line 146 "load_it.cpp"
  $296 = ((($penv)) + 577|0); //@line 146 "load_it.cpp"
  HEAP8[$296>>0] = $295; //@line 146 "load_it.cpp"
  $297 = ((($p)) + 26|0); //@line 147 "load_it.cpp"
  $298 = HEAP8[$297>>0]|0; //@line 147 "load_it.cpp"
  $299 = ((($penv)) + 575|0); //@line 147 "load_it.cpp"
  HEAP8[$299>>0] = $298; //@line 147 "load_it.cpp"
  $300 = ((($p)) + 27|0); //@line 148 "load_it.cpp"
  $301 = HEAP8[$300>>0]|0; //@line 148 "load_it.cpp"
  $302 = ((($penv)) + 574|0); //@line 148 "load_it.cpp"
  HEAP8[$302>>0] = $301; //@line 148 "load_it.cpp"
  $303 = ((($p)) + 25|0); //@line 149 "load_it.cpp"
  $304 = HEAP8[$303>>0]|0; //@line 149 "load_it.cpp"
  $305 = $304&255; //@line 149 "load_it.cpp"
  $306 = $305 << 2; //@line 149 "load_it.cpp"
  $307 = $306 & 508; //@line 149 "load_it.cpp"
  $308 = $307&65535; //@line 149 "load_it.cpp"
  $309 = ((($penv)) + 10|0); //@line 149 "load_it.cpp"
  $310 = ($307>>>0)>(256); //@line 150 "load_it.cpp"
  $$22 = $310 ? 128 : $308; //@line 150 "load_it.cpp"
  HEAP16[$309>>1] = $$22; //@line 149 "load_it.cpp"
  $311 = ($304<<24>>24)>(-1); //@line 151 "load_it.cpp"
  if ($311) {
   $312 = ((($penv)) + 4|0); //@line 151 "load_it.cpp"
   $313 = HEAP32[$312>>2]|0; //@line 151 "load_it.cpp"
   $314 = $313 | 512; //@line 151 "load_it.cpp"
   HEAP32[$312>>2] = $314; //@line 151 "load_it.cpp"
   $$pre$phiZ2D = $139;
  } else {
   $$pre$phiZ2D = $139;
  }
 }
 $315 = HEAP8[$$pre$phiZ2D>>0]|0; //@line 153 "load_it.cpp"
 $316 = ($315&255)>(24); //@line 153 "load_it.cpp"
 if ($316) {
  label = 61;
 } else {
  $317 = ((($penv)) + 560|0); //@line 153 "load_it.cpp"
  $318 = HEAP8[$317>>0]|0; //@line 153 "load_it.cpp"
  $319 = ($318&255)>(24); //@line 153 "load_it.cpp"
  if ($319) {
   label = 61;
  }
 }
 if ((label|0) == 61) {
  $320 = ((($penv)) + 4|0); //@line 153 "load_it.cpp"
  $321 = HEAP32[$320>>2]|0; //@line 153 "load_it.cpp"
  $322 = $321 & -5; //@line 153 "load_it.cpp"
  HEAP32[$320>>2] = $322; //@line 153 "load_it.cpp"
 }
 $323 = ((($penv)) + 561|0); //@line 154 "load_it.cpp"
 $324 = HEAP8[$323>>0]|0; //@line 154 "load_it.cpp"
 $325 = ($324&255)>(24); //@line 154 "load_it.cpp"
 if (!($325)) {
  $326 = ((($penv)) + 562|0); //@line 154 "load_it.cpp"
  $327 = HEAP8[$326>>0]|0; //@line 154 "load_it.cpp"
  $328 = ($327&255)>(24); //@line 154 "load_it.cpp"
  if (!($328)) {
   return 1; //@line 155 "load_it.cpp"
  }
 }
 $329 = ((($penv)) + 4|0); //@line 154 "load_it.cpp"
 $330 = HEAP32[$329>>2]|0; //@line 154 "load_it.cpp"
 $331 = $330 & -3; //@line 154 "load_it.cpp"
 HEAP32[$329>>2] = $331; //@line 154 "load_it.cpp"
 return 1; //@line 155 "load_it.cpp"
}
function __ZNK10CSoundFile10S3MConvertEP11_MODCOMMANDb($this,$m,$bIT) {
 $this = $this|0;
 $m = $m|0;
 $bIT = $bIT|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $command$0 = 0, $param$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($m)) + 3|0); //@line 71 "load_s3m.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 71 "load_s3m.cpp"
 $2 = $1&255; //@line 71 "load_s3m.cpp"
 $3 = ((($m)) + 5|0); //@line 72 "load_s3m.cpp"
 $4 = HEAP8[$3>>0]|0; //@line 72 "load_s3m.cpp"
 $5 = $4&255; //@line 72 "load_s3m.cpp"
 do {
  switch ($2|0) {
  case 1:  {
   $command$0 = 16;$param$0 = $5;
   break;
  }
  case 2:  {
   $command$0 = 12;$param$0 = $5;
   break;
  }
  case 3:  {
   if ($bIT) {
    $command$0 = 14;$param$0 = $5;
   } else {
    $6 = $5 >>> 4; //@line 77 "load_s3m.cpp"
    $7 = ($6*10)|0; //@line 77 "load_s3m.cpp"
    $8 = $5 & 15; //@line 77 "load_s3m.cpp"
    $9 = (($7) + ($8))|0; //@line 77 "load_s3m.cpp"
    $command$0 = 14;$param$0 = $9;
   }
   break;
  }
  case 4:  {
   $command$0 = 11;$param$0 = $5;
   break;
  }
  case 5:  {
   $command$0 = 3;$param$0 = $5;
   break;
  }
  case 6:  {
   $command$0 = 2;$param$0 = $5;
   break;
  }
  case 7:  {
   $command$0 = 4;$param$0 = $5;
   break;
  }
  case 8:  {
   $command$0 = 5;$param$0 = $5;
   break;
  }
  case 9:  {
   $command$0 = 18;$param$0 = $5;
   break;
  }
  case 10:  {
   $command$0 = 1;$param$0 = $5;
   break;
  }
  case 11:  {
   $command$0 = 7;$param$0 = $5;
   break;
  }
  case 12:  {
   $command$0 = 6;$param$0 = $5;
   break;
  }
  case 13:  {
   $command$0 = 21;$param$0 = $5;
   break;
  }
  case 14:  {
   $command$0 = 22;$param$0 = $5;
   break;
  }
  case 15:  {
   $command$0 = 10;$param$0 = $5;
   break;
  }
  case 16:  {
   $command$0 = 29;$param$0 = $5;
   break;
  }
  case 17:  {
   $command$0 = 15;$param$0 = $5;
   break;
  }
  case 18:  {
   $command$0 = 8;$param$0 = $5;
   break;
  }
  case 19:  {
   $command$0 = 20;$param$0 = $5;
   break;
  }
  case 20:  {
   $command$0 = 17;$param$0 = $5;
   break;
  }
  case 21:  {
   $command$0 = 26;$param$0 = $5;
   break;
  }
  case 22:  {
   $command$0 = 23;$param$0 = $5;
   break;
  }
  case 23:  {
   $command$0 = 24;$param$0 = $5;
   break;
  }
  case 24:  {
   $command$0 = 9;$param$0 = $5;
   break;
  }
  case 25:  {
   $command$0 = 27;$param$0 = $5;
   break;
  }
  case 26:  {
   $command$0 = 31;$param$0 = $5;
   break;
  }
  default: {
   $command$0 = 0;$param$0 = $5;
  }
  }
 } while(0);
 HEAP8[$0>>0] = $command$0; //@line 103 "load_s3m.cpp"
 $10 = $param$0&255; //@line 104 "load_s3m.cpp"
 HEAP8[$3>>0] = $10; //@line 104 "load_s3m.cpp"
 return; //@line 105 "load_s3m.cpp"
}
function __ZN10CSoundFile7ReadWavEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$0 = 0, $$14 = 0, $$lcssa = 0, $$off = 0, $$old2 = 0, $$pre = 0, $$pre42 = 0, $$sum = 0, $$sum10$us = 0, $$sum4 = 0, $$sum5 = 0, $$sum6$us = 0, $$sum8 = 0, $$sum9$us = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dwMemPos$0 = 0, $dwMemPos$0$lcssa = 0, $exitcond38 = 0, $exitcond41 = 0, $framesperrow$0 = 0, $i$023$us = 0, $i3$021$us = 0, $len$0 = 0;
 var $len$1 = 0, $nChn$027$us = 0, $norders$0 = 0, $or$cond = 0, $or$cond3 = 0, $psrc$022$us = 0, $psrc2$020$us = 0, $umax36 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($lpStream)) + 12|0); //@line 22 "load_wav.cpp"
 $1 = ($lpStream|0)==(0|0); //@line 23 "load_wav.cpp"
 $2 = ($dwMemLength>>>0)<(12); //@line 23 "load_wav.cpp"
 $or$cond = $1 | $2; //@line 23 "load_wav.cpp"
 if ($or$cond) {
  $$0 = 0;
  return ($$0|0); //@line 140 "load_wav.cpp"
 }
 $3 = HEAPU8[$lpStream>>0]|(HEAPU8[$lpStream+1>>0]<<8)|(HEAPU8[$lpStream+2>>0]<<16)|(HEAPU8[$lpStream+3>>0]<<24); //@line 24 "load_wav.cpp"
 $4 = ($3|0)==(1179011410); //@line 24 "load_wav.cpp"
 if (!($4)) {
  $$0 = 0;
  return ($$0|0); //@line 140 "load_wav.cpp"
 }
 $5 = ((($lpStream)) + 8|0); //@line 24 "load_wav.cpp"
 $6 = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8)|(HEAPU8[$5+2>>0]<<16)|(HEAPU8[$5+3>>0]<<24); //@line 24 "load_wav.cpp"
 $7 = ($6|0)==(1163280727); //@line 24 "load_wav.cpp"
 if (!($7)) {
  $$0 = 0;
  return ($$0|0); //@line 140 "load_wav.cpp"
 }
 $8 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 25 "load_wav.cpp"
 $9 = ($8|0)==(544501094); //@line 25 "load_wav.cpp"
 if (!($9)) {
  $$0 = 0;
  return ($$0|0); //@line 140 "load_wav.cpp"
 }
 $10 = ((($lpStream)) + 16|0); //@line 26 "load_wav.cpp"
 $11 = HEAPU8[$10>>0]|(HEAPU8[$10+1>>0]<<8)|(HEAPU8[$10+2>>0]<<16)|(HEAPU8[$10+3>>0]<<24); //@line 26 "load_wav.cpp"
 $12 = (($11) + 20)|0; //@line 26 "load_wav.cpp"
 $13 = (($11) + 28)|0; //@line 27 "load_wav.cpp"
 $14 = ($13>>>0)<($dwMemLength>>>0); //@line 27 "load_wav.cpp"
 if (!($14)) {
  $$0 = 0;
  return ($$0|0); //@line 140 "load_wav.cpp"
 }
 $15 = ((($lpStream)) + 20|0); //@line 28 "load_wav.cpp"
 $16 = HEAPU8[$15>>0]|(HEAPU8[$15+1>>0]<<8); //@line 28 "load_wav.cpp"
 switch ($16<<16>>16) {
 case -2: case 1:  {
  break;
 }
 default: {
  $$0 = 0;
  return ($$0|0); //@line 140 "load_wav.cpp"
 }
 }
 $17 = ((($lpStream)) + 22|0); //@line 29 "load_wav.cpp"
 $18 = HEAPU8[$17>>0]|(HEAPU8[$17+1>>0]<<8); //@line 29 "load_wav.cpp"
 $19 = (($18) + -1)<<16>>16; //@line 30 "load_wav.cpp"
 $20 = ($19&65535)>(3); //@line 30 "load_wav.cpp"
 if ($20) {
  $$0 = 0;
  return ($$0|0); //@line 140 "load_wav.cpp"
 }
 $21 = ((($lpStream)) + 24|0); //@line 31 "load_wav.cpp"
 $22 = HEAPU8[$21>>0]|(HEAPU8[$21+1>>0]<<8)|(HEAPU8[$21+2>>0]<<16)|(HEAPU8[$21+3>>0]<<24); //@line 31 "load_wav.cpp"
 $23 = ($22|0)==(0); //@line 31 "load_wav.cpp"
 if ($23) {
  $$0 = 0;
  return ($$0|0); //@line 140 "load_wav.cpp"
 }
 $24 = ((($lpStream)) + 34|0); //@line 32 "load_wav.cpp"
 $25 = HEAPU8[$24>>0]|(HEAPU8[$24+1>>0]<<8); //@line 32 "load_wav.cpp"
 $26 = $25 & 7; //@line 32 "load_wav.cpp"
 $27 = ($26<<16>>16)!=(0); //@line 32 "load_wav.cpp"
 $$off = (($25) + -8)<<16>>16; //@line 33 "load_wav.cpp"
 $28 = ($$off&65535)>(24); //@line 33 "load_wav.cpp"
 $29 = $28 | $27; //@line 33 "load_wav.cpp"
 if ($29) {
  $$0 = 0;
  return ($$0|0); //@line 140 "load_wav.cpp"
 } else {
  $dwMemPos$0 = $12;
 }
 while(1) {
  $30 = (($lpStream) + ($dwMemPos$0)|0); //@line 38 "load_wav.cpp"
  $31 = HEAPU8[$30>>0]|(HEAPU8[$30+1>>0]<<8)|(HEAPU8[$30+2>>0]<<16)|(HEAPU8[$30+3>>0]<<24); //@line 39 "load_wav.cpp"
  $32 = ($31|0)==(1635017060); //@line 39 "load_wav.cpp"
  if ($32) {
   $dwMemPos$0$lcssa = $dwMemPos$0;
   break;
  }
  $$sum = (($dwMemPos$0) + 4)|0; //@line 40 "load_wav.cpp"
  $33 = (($lpStream) + ($$sum)|0); //@line 40 "load_wav.cpp"
  $34 = HEAPU8[$33>>0]|(HEAPU8[$33+1>>0]<<8)|(HEAPU8[$33+2>>0]<<16)|(HEAPU8[$33+3>>0]<<24); //@line 40 "load_wav.cpp"
  $35 = (($dwMemPos$0) + 8)|0; //@line 40 "load_wav.cpp"
  $36 = (($35) + ($34))|0; //@line 40 "load_wav.cpp"
  $37 = (($36) + 8)|0; //@line 41 "load_wav.cpp"
  $38 = ($37>>>0)<($dwMemLength>>>0); //@line 41 "load_wav.cpp"
  if ($38) {
   $dwMemPos$0 = $36;
  } else {
   $$0 = 0;
   label = 37;
   break;
  }
 }
 if ((label|0) == 37) {
  return ($$0|0); //@line 140 "load_wav.cpp"
 }
 $39 = ((($this)) + 59680|0); //@line 43 "load_wav.cpp"
 HEAP32[$39>>2] = 1024; //@line 43 "load_wav.cpp"
 $40 = ((($this)) + 59684|0); //@line 44 "load_wav.cpp"
 HEAP32[$40>>2] = 0; //@line 44 "load_wav.cpp"
 $41 = ((($this)) + 59688|0); //@line 45 "load_wav.cpp"
 HEAP32[$41>>2] = 0; //@line 45 "load_wav.cpp"
 $42 = ((($this)) + 59664|0); //@line 46 "load_wav.cpp"
 HEAP32[$42>>2] = 4; //@line 46 "load_wav.cpp"
 $43 = ((($this)) + 59648|0); //@line 47 "load_wav.cpp"
 HEAP32[$43>>2] = 8; //@line 47 "load_wav.cpp"
 $44 = ((($this)) + 59652|0); //@line 48 "load_wav.cpp"
 HEAP32[$44>>2] = 125; //@line 48 "load_wav.cpp"
 $45 = ((($this)) + 59660|0); //@line 49 "load_wav.cpp"
 $46 = HEAP32[$45>>2]|0; //@line 49 "load_wav.cpp"
 $47 = $46 | 16; //@line 49 "load_wav.cpp"
 HEAP32[$45>>2] = $47; //@line 49 "load_wav.cpp"
 $48 = ((($this)) + 53344|0); //@line 50 "load_wav.cpp"
 HEAP8[$48>>0] = 0; //@line 50 "load_wav.cpp"
 $49 = ((($this)) + 53345|0); //@line 51 "load_wav.cpp"
 HEAP8[$49>>0] = -1; //@line 51 "load_wav.cpp"
 $50 = ((($this)) + 52866|0); //@line 52 "load_wav.cpp"
 HEAP16[$50>>1] = 64; //@line 52 "load_wav.cpp"
 $51 = ((($this)) + 52864|0); //@line 52 "load_wav.cpp"
 HEAP16[$51>>1] = 64; //@line 52 "load_wav.cpp"
 $52 = (__ZN10CSoundFile15AllocatePatternEjj(64,4)|0); //@line 53 "load_wav.cpp"
 $53 = ((($this)) + 51904|0); //@line 53 "load_wav.cpp"
 HEAP32[$53>>2] = $52; //@line 53 "load_wav.cpp"
 $54 = ($52|0)==(0|0); //@line 53 "load_wav.cpp"
 if ($54) {
  $$0 = 1;
  return ($$0|0); //@line 140 "load_wav.cpp"
 }
 $55 = (__ZN10CSoundFile15AllocatePatternEjj(64,4)|0); //@line 54 "load_wav.cpp"
 $56 = ((($this)) + 51908|0); //@line 54 "load_wav.cpp"
 HEAP32[$56>>2] = $55; //@line 54 "load_wav.cpp"
 $57 = ($55|0)==(0|0); //@line 54 "load_wav.cpp"
 if ($57) {
  $$0 = 1;
  return ($$0|0); //@line 140 "load_wav.cpp"
 }
 $58 = HEAPU8[$17>>0]|(HEAPU8[$17+1>>0]<<8); //@line 55 "load_wav.cpp"
 $59 = $58&65535; //@line 55 "load_wav.cpp"
 $60 = HEAPU8[$24>>0]|(HEAPU8[$24+1>>0]<<8); //@line 55 "load_wav.cpp"
 $61 = $60&65535; //@line 55 "load_wav.cpp"
 $62 = Math_imul($61, $59)|0; //@line 55 "load_wav.cpp"
 $63 = $62 >>> 3; //@line 55 "load_wav.cpp"
 $$sum4 = (($dwMemPos$0$lcssa) + 4)|0; //@line 56 "load_wav.cpp"
 $64 = (($lpStream) + ($$sum4)|0); //@line 56 "load_wav.cpp"
 $65 = HEAPU8[$64>>0]|(HEAPU8[$64+1>>0]<<8)|(HEAPU8[$64+2>>0]<<16)|(HEAPU8[$64+3>>0]<<24); //@line 56 "load_wav.cpp"
 $66 = (($dwMemLength) + -8)|0; //@line 57 "load_wav.cpp"
 $67 = (($66) - ($dwMemPos$0$lcssa))|0; //@line 57 "load_wav.cpp"
 $68 = ($65>>>0)>($67>>>0); //@line 57 "load_wav.cpp"
 $len$0 = $68 ? $67 : $65; //@line 57 "load_wav.cpp"
 $69 = (($len$0>>>0) / ($63>>>0))&-1; //@line 58 "load_wav.cpp"
 $70 = ($60&65535)>(15); //@line 60 "load_wav.cpp"
 $71 = $70&1; //@line 60 "load_wav.cpp"
 $$14 = $69 << $71; //@line 60 "load_wav.cpp"
 $72 = ($69>>>0)>(16000000); //@line 61 "load_wav.cpp"
 $len$1 = $72 ? 16000000 : $69; //@line 61 "load_wav.cpp"
 $73 = ($len$1|0)==(0); //@line 62 "load_wav.cpp"
 if ($73) {
  $$0 = 1;
  return ($$0|0); //@line 140 "load_wav.cpp"
 }
 $74 = ($len$1*50)|0; //@line 64 "load_wav.cpp"
 $75 = HEAPU8[$21>>0]|(HEAPU8[$21+1>>0]<<8)|(HEAPU8[$21+2>>0]<<16)|(HEAPU8[$21+3>>0]<<24); //@line 64 "load_wav.cpp"
 $76 = (($74>>>0) / ($75>>>0))&-1; //@line 64 "load_wav.cpp"
 $77 = (($76) + 64)|0; //@line 65 "load_wav.cpp"
 $78 = (($77>>>0) / 63)&-1; //@line 65 "load_wav.cpp"
 $79 = ($78>>>0)<(4); //@line 66 "load_wav.cpp"
 $$ = $79 ? 4 : $78; //@line 66 "load_wav.cpp"
 $$old2 = ($$>>>0)>(31); //@line 68 "load_wav.cpp"
 if ($$old2) {
  $norders$0 = 1;
  while(1) {
   $80 = (($norders$0) + 1)|0; //@line 70 "load_wav.cpp"
   $81 = (((($this)) + 53344|0) + ($norders$0)|0); //@line 70 "load_wav.cpp"
   HEAP8[$81>>0] = 1; //@line 70 "load_wav.cpp"
   $82 = (((($this)) + 53344|0) + ($80)|0); //@line 71 "load_wav.cpp"
   HEAP8[$82>>0] = -1; //@line 71 "load_wav.cpp"
   $83 = $80 << 6; //@line 72 "load_wav.cpp"
   $84 = (($83) + ($76))|0; //@line 72 "load_wav.cpp"
   $85 = (($84>>>0) / ($83>>>0))&-1; //@line 72 "load_wav.cpp"
   $86 = ($80>>>0)<(255); //@line 73 "load_wav.cpp"
   $87 = ($85>>>0)>(31); //@line 68 "load_wav.cpp"
   $or$cond3 = $86 & $87; //@line 73 "load_wav.cpp"
   if ($or$cond3) {
    $norders$0 = $80;
   } else {
    $$lcssa = $85;
    break;
   }
  }
  $$pre = HEAPU8[$17>>0]|(HEAPU8[$17+1>>0]<<8); //@line 90 "load_wav.cpp"
  $108 = $$pre;$framesperrow$0 = $$lcssa;
 } else {
  $108 = $58;$framesperrow$0 = $$;
 }
 HEAP32[$43>>2] = $framesperrow$0; //@line 75 "load_wav.cpp"
 $88 = ((($this)) + 49600|0); //@line 78 "load_wav.cpp"
 HEAP32[$88>>2] = 0; //@line 78 "load_wav.cpp"
 $89 = ((($this)) + 49604|0); //@line 79 "load_wav.cpp"
 HEAP32[$89>>2] = 64; //@line 79 "load_wav.cpp"
 $90 = ((($this)) + 49608|0); //@line 80 "load_wav.cpp"
 HEAP32[$90>>2] = 0; //@line 80 "load_wav.cpp"
 $91 = ((($this)) + 49636|0); //@line 78 "load_wav.cpp"
 HEAP32[$91>>2] = 256; //@line 78 "load_wav.cpp"
 $92 = ((($this)) + 49640|0); //@line 79 "load_wav.cpp"
 HEAP32[$92>>2] = 64; //@line 79 "load_wav.cpp"
 $93 = ((($this)) + 49644|0); //@line 80 "load_wav.cpp"
 HEAP32[$93>>2] = 0; //@line 80 "load_wav.cpp"
 $94 = ((($this)) + 49672|0); //@line 78 "load_wav.cpp"
 HEAP32[$94>>2] = 0; //@line 78 "load_wav.cpp"
 $95 = ((($this)) + 49676|0); //@line 79 "load_wav.cpp"
 HEAP32[$95>>2] = 64; //@line 79 "load_wav.cpp"
 $96 = ((($this)) + 49680|0); //@line 80 "load_wav.cpp"
 HEAP32[$96>>2] = 0; //@line 80 "load_wav.cpp"
 $97 = ((($this)) + 49708|0); //@line 78 "load_wav.cpp"
 HEAP32[$97>>2] = 256; //@line 78 "load_wav.cpp"
 $98 = ((($this)) + 49712|0); //@line 79 "load_wav.cpp"
 HEAP32[$98>>2] = 64; //@line 79 "load_wav.cpp"
 $99 = ((($this)) + 49716|0); //@line 80 "load_wav.cpp"
 HEAP32[$99>>2] = 0; //@line 80 "load_wav.cpp"
 $100 = HEAP32[$53>>2]|0; //@line 83 "load_wav.cpp"
 $101 = ((($100)) + 3|0); //@line 84 "load_wav.cpp"
 HEAP8[$101>>0] = 16; //@line 84 "load_wav.cpp"
 $102 = $framesperrow$0&255; //@line 85 "load_wav.cpp"
 $103 = ((($100)) + 5|0); //@line 85 "load_wav.cpp"
 HEAP8[$103>>0] = $102; //@line 85 "load_wav.cpp"
 HEAP8[$100>>0] = 61; //@line 86 "load_wav.cpp"
 $104 = ((($100)) + 1|0); //@line 87 "load_wav.cpp"
 HEAP8[$104>>0] = 1; //@line 87 "load_wav.cpp"
 $105 = ((($100)) + 6|0); //@line 88 "load_wav.cpp"
 HEAP8[$105>>0] = 61; //@line 88 "load_wav.cpp"
 $106 = ((($100)) + 7|0); //@line 89 "load_wav.cpp"
 HEAP8[$106>>0] = 1; //@line 89 "load_wav.cpp"
 $107 = $108&65535; //@line 90 "load_wav.cpp"
 HEAP32[$40>>2] = $107; //@line 90 "load_wav.cpp"
 $109 = ($108<<16>>16)==(0); //@line 92 "load_wav.cpp"
 if ($109) {
  $$0 = 1;
  return ($$0|0); //@line 140 "load_wav.cpp"
 }
 $110 = (($$14) + 8)|0; //@line 115 "load_wav.cpp"
 $$sum8 = (($dwMemPos$0$lcssa) + 6)|0; //@line 120 "load_wav.cpp"
 $111 = (($len$1) + -1)|0; //@line 126 "load_wav.cpp"
 $112 = (($len$1) + 1)|0; //@line 126 "load_wav.cpp"
 $$sum5 = (($dwMemPos$0$lcssa) + 8)|0; //@line 130 "load_wav.cpp"
 $113 = (($dwMemPos$0$lcssa) + 7)|0; //@line 92 "load_wav.cpp"
 $114 = (($113) - ($dwMemLength))|0; //@line 92 "load_wav.cpp"
 $115 = $65 ^ -1; //@line 92 "load_wav.cpp"
 $116 = ($114>>>0)>($115>>>0);
 $umax36 = $116 ? $114 : $115;
 $117 = $umax36 ^ -1; //@line 92 "load_wav.cpp"
 $118 = (($117>>>0) / ($63>>>0))&-1; //@line 92 "load_wav.cpp"
 $119 = ($118>>>0)<(16000000);
 $120 = $119 ? $118 : 16000000; //@line 92 "load_wav.cpp"
 $121 = ((($100)) + 15|0); //@line 110 "load_wav.cpp"
 $122 = ((($100)) + 17|0); //@line 110 "load_wav.cpp"
 $123 = ((($100)) + 21|0); //@line 111 "load_wav.cpp"
 $124 = ((($100)) + 23|0); //@line 111 "load_wav.cpp"
 $127 = 61;$143 = $107;$nChn$027$us = 0;
 while(1) {
  $125 = (($nChn$027$us) + 1)|0; //@line 94 "load_wav.cpp"
  $126 = (($100) + (($nChn$027$us*6)|0)|0); //@line 95 "load_wav.cpp"
  HEAP8[$126>>0] = $127; //@line 95 "load_wav.cpp"
  $128 = $125&255; //@line 96 "load_wav.cpp"
  $129 = (((($100) + (($nChn$027$us*6)|0)|0)) + 1|0); //@line 96 "load_wav.cpp"
  HEAP8[$129>>0] = $128; //@line 96 "load_wav.cpp"
  $130 = (((($this)) + 33280|0) + ($125<<6)|0); //@line 97 "load_wav.cpp"
  HEAP32[$130>>2] = $len$1; //@line 97 "load_wav.cpp"
  $131 = HEAPU8[$21>>0]|(HEAPU8[$21+1>>0]<<8)|(HEAPU8[$21+2>>0]<<16)|(HEAPU8[$21+3>>0]<<24); //@line 98 "load_wav.cpp"
  $132 = (((((($this)) + 33280|0) + ($125<<6)|0)) + 24|0); //@line 98 "load_wav.cpp"
  HEAP32[$132>>2] = $131; //@line 98 "load_wav.cpp"
  $133 = (((((($this)) + 33280|0) + ($125<<6)|0)) + 30|0); //@line 99 "load_wav.cpp"
  HEAP16[$133>>1] = 256; //@line 99 "load_wav.cpp"
  $134 = (((((($this)) + 33280|0) + ($125<<6)|0)) + 28|0); //@line 100 "load_wav.cpp"
  HEAP16[$134>>1] = 128; //@line 100 "load_wav.cpp"
  $135 = (((((($this)) + 33280|0) + ($125<<6)|0)) + 32|0); //@line 101 "load_wav.cpp"
  HEAP16[$135>>1] = 64; //@line 101 "load_wav.cpp"
  $136 = HEAPU8[$24>>0]|(HEAPU8[$24+1>>0]<<8); //@line 102 "load_wav.cpp"
  $137 = ($136&65535)>(15); //@line 102 "load_wav.cpp"
  $138 = (((((($this)) + 33280|0) + ($125<<6)|0)) + 34|0); //@line 102 "load_wav.cpp"
  $139 = $137&1; //@line 103 "load_wav.cpp"
  $140 = $139 | 32; //@line 103 "load_wav.cpp"
  $141 = $140&65535; //@line 103 "load_wav.cpp"
  HEAP16[$138>>1] = $141; //@line 103 "load_wav.cpp"
  $142 = ($143>>>0)>(1); //@line 104 "load_wav.cpp"
  L52: do {
   if ($142) {
    switch ($nChn$027$us|0) {
    case 0:  {
     HEAP16[$134>>1] = 0; //@line 108 "load_wav.cpp"
     break L52;
     break;
    }
    case 1:  {
     HEAP16[$134>>1] = 256; //@line 109 "load_wav.cpp"
     break L52;
     break;
    }
    case 2:  {
     $144 = ($143|0)==(3); //@line 110 "load_wav.cpp"
     $145 = $144 ? 128 : 64; //@line 110 "load_wav.cpp"
     HEAP16[$134>>1] = $145; //@line 110 "load_wav.cpp"
     HEAP8[$121>>0] = 20; //@line 110 "load_wav.cpp"
     HEAP8[$122>>0] = -111; //@line 110 "load_wav.cpp"
     break L52;
     break;
    }
    case 3:  {
     HEAP16[$134>>1] = 192; //@line 111 "load_wav.cpp"
     HEAP8[$123>>0] = 20; //@line 111 "load_wav.cpp"
     HEAP8[$124>>0] = -111; //@line 111 "load_wav.cpp"
     break L52;
     break;
    }
    default: {
     HEAP16[$134>>1] = 128; //@line 112 "load_wav.cpp"
     break L52;
    }
    }
   }
  } while(0);
  $146 = (__ZN10CSoundFile14AllocateSampleEj($110)|0); //@line 115 "load_wav.cpp"
  $147 = (((((($this)) + 33280|0) + ($125<<6)|0)) + 20|0); //@line 115 "load_wav.cpp"
  HEAP32[$147>>2] = $146; //@line 115 "load_wav.cpp"
  $148 = ($146|0)==(0|0); //@line 115 "load_wav.cpp"
  if ($148) {
   $$0 = 1;
   label = 37;
   break;
  }
  $149 = HEAPU8[$24>>0]|(HEAPU8[$24+1>>0]<<8); //@line 116 "load_wav.cpp"
  $150 = ($149&65535)>(15); //@line 116 "load_wav.cpp"
  if ($150) {
   $173 = $149&65535; //@line 116 "load_wav.cpp"
   $174 = $173 >>> 3; //@line 118 "load_wav.cpp"
   $$sum9$us = (($$sum8) + ($174))|0; //@line 120 "load_wav.cpp"
   $175 = Math_imul($174, $nChn$027$us)|0; //@line 120 "load_wav.cpp"
   $$sum10$us = (($$sum9$us) + ($175))|0; //@line 120 "load_wav.cpp"
   $176 = (($lpStream) + ($$sum10$us)|0); //@line 120 "load_wav.cpp"
   $i$023$us = 0;$psrc$022$us = $176;
   while(1) {
    $166 = HEAP16[$psrc$022$us>>1]|0; //@line 123 "load_wav.cpp"
    $167 = (($146) + ($i$023$us<<1)|0); //@line 123 "load_wav.cpp"
    HEAP16[$167>>1] = $166; //@line 123 "load_wav.cpp"
    $168 = (($psrc$022$us) + ($63)|0); //@line 124 "load_wav.cpp"
    $169 = (($i$023$us) + 1)|0; //@line 121 "load_wav.cpp"
    $exitcond41 = ($169|0)==($120|0); //@line 121 "load_wav.cpp"
    if ($exitcond41) {
     break;
    } else {
     $i$023$us = $169;$psrc$022$us = $168;
    }
   }
   $162 = (($146) + ($111<<1)|0); //@line 126 "load_wav.cpp"
   $163 = HEAP16[$162>>1]|0; //@line 126 "load_wav.cpp"
   $164 = (($146) + ($len$1<<1)|0); //@line 126 "load_wav.cpp"
   HEAP16[$164>>1] = $163; //@line 126 "load_wav.cpp"
   $165 = (($146) + ($112<<1)|0); //@line 126 "load_wav.cpp"
   HEAP16[$165>>1] = $163; //@line 126 "load_wav.cpp"
  } else {
   $$sum6$us = (($$sum5) + ($nChn$027$us))|0; //@line 130 "load_wav.cpp"
   $172 = (($lpStream) + ($$sum6$us)|0); //@line 130 "load_wav.cpp"
   $i3$021$us = 0;$psrc2$020$us = $172;
   while(1) {
    $155 = HEAP8[$psrc2$020$us>>0]|0; //@line 133 "load_wav.cpp"
    $156 = $155&255; //@line 133 "load_wav.cpp"
    $157 = (($156) + 128)|0; //@line 133 "load_wav.cpp"
    $158 = $157&255; //@line 133 "load_wav.cpp"
    $159 = (($146) + ($i3$021$us)|0); //@line 133 "load_wav.cpp"
    HEAP8[$159>>0] = $158; //@line 133 "load_wav.cpp"
    $160 = (($psrc2$020$us) + ($63)|0); //@line 134 "load_wav.cpp"
    $161 = (($i3$021$us) + 1)|0; //@line 131 "load_wav.cpp"
    $exitcond38 = ($161|0)==($120|0); //@line 131 "load_wav.cpp"
    if ($exitcond38) {
     break;
    } else {
     $i3$021$us = $161;$psrc2$020$us = $160;
    }
   }
   $151 = (($146) + ($111)|0); //@line 136 "load_wav.cpp"
   $152 = HEAP8[$151>>0]|0; //@line 136 "load_wav.cpp"
   $153 = (($146) + ($len$1)|0); //@line 136 "load_wav.cpp"
   HEAP8[$153>>0] = $152; //@line 136 "load_wav.cpp"
   $154 = (($146) + ($112)|0); //@line 136 "load_wav.cpp"
   HEAP8[$154>>0] = $152; //@line 136 "load_wav.cpp"
  }
  $170 = HEAP32[$40>>2]|0; //@line 92 "load_wav.cpp"
  $171 = ($125>>>0)<($170>>>0); //@line 92 "load_wav.cpp"
  if (!($171)) {
   $$0 = 1;
   label = 37;
   break;
  }
  $$pre42 = HEAP8[$100>>0]|0; //@line 95 "load_wav.cpp"
  $127 = $$pre42;$143 = $170;$nChn$027$us = $125;
 }
 if ((label|0) == 37) {
  return ($$0|0); //@line 140 "load_wav.cpp"
 }
 return (0)|0;
}
function __ZN10CSoundFile14AllocateSampleEj($nbytes) {
 $nbytes = $nbytes|0;
 var $$ = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($nbytes) + 39)|0; //@line 356 "sndfile.cpp"
 $1 = $0 & -8; //@line 356 "sndfile.cpp"
 $2 = (_malloc($1)|0); //@line 101 "./libmodplug/stdafx.h"
 $3 = ($2|0)==(0|0); //@line 103 "./libmodplug/stdafx.h"
 if (!($3)) {
  _memset(($2|0),0,($1|0))|0; //@line 103 "./libmodplug/stdafx.h"
 }
 $4 = ((($2)) + 16|0); //@line 357 "sndfile.cpp"
 $$ = $3 ? 0 : $4; //@line 357 "sndfile.cpp"
 return ($$|0); //@line 358 "sndfile.cpp"
}
function __ZN10CSoundFile7ReadABCEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$$$i$i = 0, $$$h$0$i$i = 0, $$$h$0$i$i$1 = 0, $$$h$0$i$i$2 = 0, $$$h$0$i$i$3 = 0, $$$h$0$i$i121 = 0, $$$h$0$i$i121$1 = 0, $$$h$0$i$i121$2 = 0, $$$h$0$i$i121$3 = 0, $$$i = 0, $$$i$i = 0, $$$i739 = 0, $$0 = 0, $$0$i$i195 = 0, $$0$i$i224 = 0, $$0$i107 = 0, $$0$i107$lcssa = 0, $$0$i126 = 0, $$0$i126$lcssa = 0;
 var $$0$i133 = 0, $$0$i133$lcssa = 0, $$0$i136 = 0, $$0$i154 = 0, $$0$i156 = 0, $$0$i156$lcssa = 0, $$0$i164 = 0, $$0$i170 = 0, $$0$i206 = 0, $$0$i206$lcssa = 0, $$0$i214 = 0, $$0$i226 = 0, $$0$i230 = 0, $$0$i284 = 0, $$0$i286 = 0, $$0$i287 = 0, $$0$i288 = 0, $$0$i290 = 0, $$0$i358 = 0, $$0$i366 = 0;
 var $$0$i401 = 0, $$0$i404 = 0, $$0$i442 = 0, $$0$i444 = 0, $$0$i446 = 0, $$0$i448 = 0, $$0$i464 = 0, $$0$i466 = 0, $$0$i468 = 0, $$0$i477 = 0, $$0$i483 = 0, $$0$i492 = 0, $$0$i501 = 0, $$0$i510 = 0, $$0$i519 = 0, $$0$i528 = 0, $$0$i537 = 0, $$0$i539 = 0, $$0$i541 = 0, $$0$i543 = 0;
 var $$0$i545 = 0, $$0$i547 = 0, $$0$i549 = 0, $$0$i551 = 0, $$0$i553 = 0, $$0$i555 = 0, $$0$i563 = 0, $$0$i598 = 0, $$0$i600 = 0, $$0$i603 = 0, $$0$i607 = 0, $$0$i645 = 0, $$0$i647 = 0, $$0$i650 = 0, $$0$i654 = 0, $$0$i693 = 0, $$0$i699 = 0, $$0$i723 = 0, $$01$i = 0, $$01$i601 = 0;
 var $$01$i648 = 0, $$1$i = 0, $$1$i$1 = 0, $$1$i$1$lcssa = 0, $$1$i$2 = 0, $$1$i$2$lcssa = 0, $$1$i$3 = 0, $$1$i$3$lcssa = 0, $$1$i$lcssa = 0, $$1$i111 = 0, $$1$i111$1 = 0, $$1$i111$1$lcssa = 0, $$1$i111$2 = 0, $$1$i111$2$lcssa = 0, $$1$i111$3 = 0, $$1$i111$3$lcssa = 0, $$1$i111$lcssa = 0, $$1$i450 = 0, $$1$i742 = 0, $$12$i = 0;
 var $$12$i158 = 0, $$12$i208 = 0, $$2$i135 = 0, $$2$i135$lcssa = 0, $$2$i160 = 0, $$2$i160$lcssa = 0, $$2$i210 = 0, $$2$i210$lcssa = 0, $$39$i = 0, $$786 = 0, $$788 = 0, $$789 = 0, $$790 = 0, $$791 = 0, $$795 = 0, $$796 = 0, $$797 = 0, $$799 = 0, $$80 = 0, $$800 = 0;
 var $$801 = 0, $$81 = 0, $$82 = 0, $$85 = 0, $$89 = 0, $$93 = 0, $$95 = 0, $$97 = 0, $$98 = 0, $$abcnoslurs$4 = 0, $$abcnoslurs$6 = 0, $$abcnoslurs$9 = 0, $$abcstate$3 = 0, $$abctempo$0 = 0, $$abcto$4 = 0, $$abcxcount$0 = 0, $$abcxcount$0$lcssa = 0, $$be = 0, $$be865 = 0, $$d$1$i$i = 0;
 var $$h$0$i = 0, $$h$0$i$i = 0, $$h$0$i$i$1 = 0, $$h$0$i$i$2 = 0, $$h$0$i$i$3 = 0, $$h$0$i$i119 = 0, $$h$0$i$i119$1 = 0, $$h$0$i$i119$2 = 0, $$h$0$i$i119$3 = 0, $$h$0$i$i239 = 0, $$h$0$i$i262 = 0, $$h$0$i$i576 = 0, $$h$0$i$i623 = 0, $$h$0$i10$i = 0, $$h$0$i10$i279 = 0, $$h$0$i10$i396 = 0, $$h$0$i10$i593 = 0, $$h$0$i10$i640 = 0, $$h$0$i10$i688 = 0, $$h$0$i307 = 0;
 var $$h$0$i319 = 0, $$h$0$i331 = 0, $$h$0$i343 = 0, $$h$0$i355 = 0, $$h$0$i415 = 0, $$h$0$i427 = 0, $$h$0$i439 = 0, $$idx = 0, $$idx$val = 0, $$idx101 = 0, $$idx101$val = 0, $$idx102 = 0, $$idx102$val = 0, $$idx103 = 0, $$idx103$val = 0, $$idx104 = 0, $$idx104$val = 0, $$idx105 = 0, $$idx105$val = 0, $$idx106 = 0;
 var $$idx106$val = 0, $$lcssa13$i = 0, $$lcssa13$i267 = 0, $$lcssa13$i384 = 0, $$lcssa13$i581 = 0, $$lcssa13$i628 = 0, $$lcssa13$i676 = 0, $$lcssa2123 = 0, $$lcssa2124 = 0, $$lcssa2125 = 0, $$lcssa2127 = 0, $$lcssa2129 = 0, $$lcssa2130 = 0, $$lcssa2131 = 0, $$lcssa2132 = 0, $$lcssa2133 = 0, $$lcssa2134 = 0, $$lcssa2135 = 0, $$lcssa2136 = 0, $$lcssa2137 = 0;
 var $$lcssa2138 = 0, $$lcssa2139 = 0, $$lcssa2140 = 0, $$lcssa2141 = 0, $$lcssa2142 = 0, $$lcssa2143 = 0, $$lcssa2148 = 0, $$lcssa2153 = 0, $$lcssa2157 = 0, $$lcssa2158 = 0, $$lcssa2159 = 0, $$lcssa2160 = 0, $$lcssa2179 = 0, $$lcssa2188 = 0, $$lcssa2202 = 0, $$lcssa2210 = 0, $$lcssa2211 = 0, $$lcssa2212 = 0, $$lcssa2213 = 0, $$lcssa2214 = 0;
 var $$lcssa2215 = 0, $$lcssa2216 = 0, $$lcssa2217 = 0, $$lcssa2218 = 0, $$lcssa2219 = 0, $$lcssa2220 = 0, $$lcssa2221 = 0, $$lcssa2222 = 0, $$lcssa2223 = 0, $$lcssa2230 = 0, $$lcssa2231 = 0, $$lcssa2232 = 0, $$lcssa2239 = 0, $$lcssa2240 = 0, $$lcssa2245 = 0, $$lcssa2248 = 0, $$lcssa2249 = 0, $$lcssa2250 = 0, $$lcssa2256 = 0, $$lcssa2257 = 0;
 var $$lcssa2264 = 0, $$lcssa2265 = 0, $$lcssa2272 = 0, $$lcssa2273 = 0, $$lcssa2275 = 0, $$lcssa2276 = 0, $$lcssa2277 = 0, $$lcssa2278 = 0, $$lcssa2288 = 0, $$lcssa2289 = 0, $$lcssa2290 = 0, $$lcssa2291 = 0, $$lcssa2292 = 0, $$lcssa2293 = 0, $$lcssa2294 = 0, $$lcssa2295 = 0, $$lcssa2296 = 0, $$lcssa2297 = 0, $$lcssa2298 = 0, $$lcssa928 = 0;
 var $$lcssa948 = 0, $$lcssa959 = 0, $$lcssa960 = 0, $$lcssa961 = 0, $$nd0$0 = 0, $$neg840 = 0, $$nl0$0 = 0, $$not = 0, $$not1187 = 0, $$op = 0, $$op1547 = 0, $$p$i = 0, $$ph$i = 0, $$ph1257 = 0, $$ph892$lcssa = 0, $$ph8921258 = 0, $$ph893$lcssa = 0, $$ph8931259 = 0, $$ph894$lcssa = 0, $$ph8941260 = 0;
 var $$phi$trans$insert = 0, $$phi$trans$insert$i = 0, $$phi$trans$insert$i$i = 0, $$pp$1 = 0, $$pr = 0, $$pr$i = 0, $$pr$pre = 0, $$pr759 = 0, $$pr760 = 0, $$pr761 = 0, $$pr766 = 0, $$pr768 = 0, $$pr774 = 0, $$pr774$pre = 0, $$pr777 = 0, $$pr777$pre = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i457 = 0;
 var $$pre15$i$i = 0, $$pre1509 = 0, $$pre1510 = 0, $$pre1511 = 0, $$pre1512 = 0, $$pre1514 = 0, $$pre1515 = 0, $$pre1516 = 0, $$pre1518 = 0, $$pre1519 = 0, $$pre1521 = 0, $$pre1522 = 0, $$pre1523 = 0, $$pre1525 = 0, $$pre1526 = 0, $$pre1527 = 0, $$pre1531 = 0, $$pre1532 = 0, $$pre1533 = 0, $$pre1535 = 0;
 var $$pre1537 = 0, $$pre1538 = 0, $$pre1539 = 0, $$pre1540 = 0, $$pre1542 = 0, $$pre1543 = 0, $$promoted = 0, $$promoted1212 = 0, $$sum$i = 0, $$sum$i$lcssa = 0, $$sum$i275 = 0, $$sum$i275$lcssa = 0, $$sum$i392 = 0, $$sum$i392$lcssa = 0, $$sum$i453 = 0, $$sum$i589 = 0, $$sum$i589$lcssa = 0, $$sum$i636 = 0, $$sum$i636$lcssa = 0, $$sum$i684 = 0;
 var $$sum$i684$lcssa = 0, $$sum1$i = 0, $$sum43$pn = 0, $$sum43$pn$in = 0, $$sum44 = 0, $$sum45 = 0, $$sum46 = 0, $$sum49 = 0, $$sum54 = 0, $$sum59$pn = 0, $$sum59$pn$in = 0, $$sum61 = 0, $$sum62 = 0, $$sum63 = 0, $$sum64 = 0, $$sum65 = 0, $$sum69 = 0, $$sum70 = 0, $$sum78 = 0, $$sum79 = 0;
 var $$sum803 = 0, $$sum804 = 0, $$sum805 = 0, $$sum806 = 0, $$sum807 = 0, $$sum808 = 0, $$sum809 = 0, $$sum810 = 0, $$sum810$lcssa = 0, $$sum813 = 0, $$sum814 = 0, $$sum815 = 0, $$sum816 = 0, $$sum817 = 0, $$sum818 = 0, $$sum819 = 0, $$sum820 = 0, $$sum821 = 0, $$sum822 = 0, $$sum823 = 0;
 var $$sum825 = 0, $$sum826 = 0, $$sum827 = 0, $$sum830 = 0, $$sum833 = 0, $$tracktime$0$i = 0, $$tracktime$0$i185 = 0, $$tracktime$0$i204 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0;
 var $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0;
 var $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0;
 var $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0;
 var $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0;
 var $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0;
 var $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0;
 var $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0;
 var $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0;
 var $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0;
 var $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0;
 var $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0;
 var $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0;
 var $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0;
 var $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0;
 var $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0, $1267 = 0, $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0;
 var $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0, $1285 = 0, $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0;
 var $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0, $1302 = 0, $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0;
 var $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0, $1320 = 0, $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0;
 var $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0, $1339 = 0, $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0;
 var $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0, $1357 = 0, $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0;
 var $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0, $1375 = 0, $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0;
 var $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0, $1393 = 0, $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0;
 var $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0, $1410 = 0, $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0;
 var $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0, $1429 = 0, $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0;
 var $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0, $1447 = 0, $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0;
 var $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0, $1465 = 0, $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0;
 var $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0, $1483 = 0, $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0;
 var $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0, $1500 = 0, $1501 = 0, $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0, $1511 = 0;
 var $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0, $1518 = 0, $1519 = 0, $152 = 0, $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0, $153 = 0;
 var $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0, $1536 = 0, $1537 = 0, $1538 = 0, $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0, $1548 = 0;
 var $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0, $1554 = 0, $1555 = 0, $1556 = 0, $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $1565 = 0, $1566 = 0;
 var $1567 = 0, $1568 = 0, $1569 = 0, $157 = 0, $1570 = 0, $1571 = 0, $1572 = 0, $1573 = 0, $1574 = 0, $1575 = 0, $1576 = 0, $1577 = 0, $1578 = 0, $1579 = 0, $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0, $1583 = 0, $1584 = 0;
 var $1585 = 0, $1586 = 0, $1587 = 0, $1588 = 0, $1589 = 0, $159 = 0, $1590 = 0, $1591 = 0, $1592 = 0, $1593 = 0, $1594 = 0, $1595 = 0, $1596 = 0, $1597 = 0, $1598 = 0, $1599 = 0, $16 = 0, $160 = 0, $1600 = 0, $1601 = 0;
 var $1602 = 0, $1603 = 0, $1604 = 0, $1605 = 0, $1606 = 0, $1607 = 0, $1608 = 0, $1609 = 0, $161 = 0, $1610 = 0, $1611 = 0, $1612 = 0, $1613 = 0, $1614 = 0, $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0, $1619 = 0, $162 = 0;
 var $1620 = 0, $1621 = 0, $1622 = 0, $1623 = 0, $1624 = 0, $1625 = 0, $1626 = 0, $1627 = 0, $1628 = 0, $1629 = 0, $163 = 0, $1630 = 0, $1631 = 0, $1632 = 0, $1633 = 0, $1634 = 0, $1635 = 0, $1636 = 0, $1637 = 0, $1638 = 0;
 var $1639 = 0, $164 = 0, $1640 = 0, $1641 = 0, $1642 = 0, $1643 = 0, $1644 = 0, $1645 = 0, $1646 = 0, $1647 = 0, $1648 = 0, $1649 = 0, $165 = 0, $1650 = 0, $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0, $1655 = 0, $1656 = 0;
 var $1657 = 0, $1658 = 0, $1659 = 0, $166 = 0, $1660 = 0, $1661 = 0, $1662 = 0, $1663 = 0, $1664 = 0, $1665 = 0, $1666 = 0, $1667 = 0, $1668 = 0, $1669 = 0, $167 = 0, $1670 = 0, $1671 = 0, $1672 = 0, $1673 = 0, $1674 = 0;
 var $1675 = 0, $1676 = 0, $1677 = 0, $1678 = 0, $1679 = 0, $168 = 0, $1680 = 0, $1681 = 0, $1682 = 0, $1683 = 0, $1684 = 0, $1685 = 0, $1686 = 0, $1687 = 0, $1688 = 0, $1689 = 0, $169 = 0, $1690 = 0, $1691 = 0, $1692 = 0;
 var $1693 = 0, $1694 = 0, $1695 = 0, $1696 = 0, $1697 = 0, $1698 = 0, $1699 = 0, $17 = 0, $170 = 0, $1700 = 0, $1701 = 0, $1702 = 0, $1703 = 0, $1704 = 0, $1705 = 0, $1706 = 0, $1707 = 0, $1708 = 0, $1709 = 0, $171 = 0;
 var $1710 = 0, $1711 = 0, $1712 = 0, $1713 = 0, $1714 = 0, $1715 = 0, $1716 = 0, $1717 = 0, $1718 = 0, $1719 = 0, $172 = 0, $1720 = 0, $1721 = 0, $1722 = 0, $1723 = 0, $1724 = 0, $1725 = 0, $1726 = 0, $1727 = 0, $1728 = 0;
 var $1729 = 0, $173 = 0, $1730 = 0, $1731 = 0, $1732 = 0, $1733 = 0, $1734 = 0, $1735 = 0, $1736 = 0, $1737 = 0, $1738 = 0, $1739 = 0, $174 = 0, $1740 = 0, $1741 = 0, $1742 = 0, $1743 = 0, $1744 = 0, $1745 = 0, $1746 = 0;
 var $1747 = 0, $1748 = 0, $1749 = 0, $175 = 0, $1750 = 0, $1751 = 0, $1752 = 0, $1753 = 0, $1754 = 0, $1755 = 0, $1756 = 0, $1757 = 0, $1758 = 0, $1759 = 0, $176 = 0, $1760 = 0, $1761 = 0, $1762 = 0, $1763 = 0, $1764 = 0;
 var $1765 = 0, $1766 = 0, $1767 = 0, $1768 = 0, $1769 = 0, $177 = 0, $1770 = 0, $1771 = 0, $1772 = 0, $1773 = 0, $1774 = 0, $1775 = 0, $1776 = 0, $1777 = 0, $1778 = 0, $1779 = 0, $178 = 0, $1780 = 0, $1781 = 0, $1782 = 0;
 var $1783 = 0, $1784 = 0, $1785 = 0, $1786 = 0, $1787 = 0, $1788 = 0, $1789 = 0, $179 = 0, $1790 = 0, $1791 = 0, $1792 = 0, $1793 = 0, $1794 = 0, $1795 = 0, $1796 = 0, $1797 = 0, $1798 = 0, $1799 = 0, $18 = 0, $180 = 0;
 var $1800 = 0, $1801 = 0, $1802 = 0, $1803 = 0, $1804 = 0, $1805 = 0, $1806 = 0, $1807 = 0, $1808 = 0, $1809 = 0, $181 = 0, $1810 = 0, $1811 = 0, $1812 = 0, $1813 = 0, $1814 = 0, $1815 = 0, $1816 = 0, $1817 = 0, $1818 = 0;
 var $1819 = 0, $182 = 0, $1820 = 0, $1821 = 0, $1822 = 0, $1823 = 0, $1824 = 0, $1825 = 0, $1826 = 0, $1827 = 0, $1828 = 0, $1829 = 0, $183 = 0, $1830 = 0, $1831 = 0, $1832 = 0, $1833 = 0, $1834 = 0, $1835 = 0, $1836 = 0;
 var $1837 = 0, $1838 = 0, $1839 = 0, $184 = 0, $1840 = 0, $1841 = 0, $1842 = 0, $1843 = 0, $1844 = 0, $1845 = 0, $1846 = 0, $1847 = 0, $1848 = 0, $1849 = 0, $185 = 0, $1850 = 0, $1851 = 0, $1852 = 0, $1853 = 0, $1854 = 0;
 var $1855 = 0, $1856 = 0, $1857 = 0, $1858 = 0, $1859 = 0, $186 = 0, $1860 = 0, $1861 = 0, $1862 = 0, $1863 = 0, $1864 = 0, $1865 = 0, $1866 = 0, $1867 = 0, $1868 = 0, $1869 = 0, $187 = 0, $1870 = 0, $1871 = 0, $1872 = 0;
 var $1873 = 0, $1874 = 0, $1875 = 0, $1876 = 0, $1877 = 0, $1878 = 0, $1879 = 0, $188 = 0, $1880 = 0, $1881 = 0, $1882 = 0, $1883 = 0, $1884 = 0, $1885 = 0, $1886 = 0, $1887 = 0, $1888 = 0, $1889 = 0, $189 = 0, $1890 = 0;
 var $1891 = 0, $1892 = 0, $1893 = 0, $1894 = 0, $1895 = 0, $1896 = 0, $1897 = 0, $1898 = 0, $1899 = 0, $19 = 0, $190 = 0, $1900 = 0, $1901 = 0, $1902 = 0, $1903 = 0, $1904 = 0, $1905 = 0, $1906 = 0, $1907 = 0, $1908 = 0;
 var $1909 = 0, $191 = 0, $1910 = 0, $1911 = 0, $1912 = 0, $1913 = 0, $1914 = 0, $1915 = 0, $1916 = 0, $1917 = 0, $1918 = 0, $1919 = 0, $192 = 0, $1920 = 0, $1921 = 0, $1922 = 0, $1923 = 0, $1924 = 0, $1925 = 0, $1926 = 0;
 var $1927 = 0, $1928 = 0, $1929 = 0, $193 = 0, $1930 = 0, $1931 = 0, $1932 = 0, $1933 = 0, $1934 = 0, $1935 = 0, $1936 = 0, $1937 = 0, $1938 = 0, $1939 = 0, $194 = 0, $1940 = 0, $1941 = 0, $1942 = 0, $1943 = 0, $1944 = 0;
 var $1945 = 0, $1946 = 0, $1947 = 0, $1948 = 0, $1949 = 0, $195 = 0, $1950 = 0, $1951 = 0, $1952 = 0, $1953 = 0, $1954 = 0, $1955 = 0, $1956 = 0, $1957 = 0, $1958 = 0, $1959 = 0, $196 = 0, $1960 = 0, $1961 = 0, $1962 = 0;
 var $1963 = 0, $1964 = 0, $1965 = 0, $1966 = 0, $1967 = 0, $1968 = 0, $1969 = 0, $197 = 0, $1970 = 0, $1971 = 0, $1972 = 0, $1973 = 0, $1974 = 0, $1975 = 0, $1976 = 0, $1977 = 0, $1978 = 0, $1979 = 0, $198 = 0, $1980 = 0;
 var $1981 = 0, $1982 = 0, $1983 = 0, $1984 = 0, $1985 = 0, $1986 = 0, $1987 = 0, $1988 = 0, $1989 = 0, $199 = 0, $1990 = 0, $1991 = 0, $1992 = 0, $1993 = 0, $1994 = 0, $1995 = 0, $1996 = 0, $1997 = 0, $1998 = 0, $1999 = 0;
 var $2 = 0, $20 = 0.0, $200 = 0, $2000 = 0, $2001 = 0, $2002 = 0, $2003 = 0, $2004 = 0, $2005 = 0, $2006 = 0, $2007 = 0, $2008 = 0, $2009 = 0, $201 = 0, $2010 = 0, $2011 = 0, $2012 = 0, $2013 = 0, $2014 = 0, $2015 = 0;
 var $2016 = 0, $2017 = 0, $2018 = 0, $2019 = 0, $202 = 0, $2020 = 0, $2021 = 0, $2022 = 0, $2023 = 0, $2024 = 0, $2025 = 0, $2026 = 0, $2027 = 0, $2028 = 0, $2029 = 0, $203 = 0, $2030 = 0, $2031 = 0, $2032 = 0, $2033 = 0;
 var $2034 = 0, $2035 = 0, $2036 = 0, $2037 = 0, $2038 = 0, $2039 = 0, $204 = 0, $2040 = 0, $2041 = 0, $2042 = 0, $2043 = 0, $2044 = 0, $2045 = 0, $2046 = 0, $2047 = 0, $2048 = 0, $2049 = 0, $205 = 0, $2050 = 0, $2051 = 0;
 var $2052 = 0, $2053 = 0, $2054 = 0, $2055 = 0, $2056 = 0, $2057 = 0, $2058 = 0, $2059 = 0, $206 = 0, $2060 = 0, $2061 = 0, $2062 = 0, $2063 = 0, $2064 = 0, $2065 = 0, $2066 = 0, $2067 = 0, $2068 = 0, $2069 = 0, $207 = 0;
 var $2070 = 0, $2071 = 0, $2072 = 0, $2073 = 0, $2074 = 0, $2075 = 0, $2076 = 0, $2077 = 0, $2078 = 0, $2079 = 0, $208 = 0, $2080 = 0, $2081 = 0, $2082 = 0, $2083 = 0, $2084 = 0, $2085 = 0, $2086 = 0, $2087 = 0, $2088 = 0;
 var $2089 = 0, $209 = 0, $2090 = 0, $2091 = 0, $2092 = 0, $2093 = 0, $2094 = 0, $2095 = 0, $2096 = 0, $2097 = 0, $2098 = 0, $2099 = 0, $21 = 0.0, $210 = 0, $2100 = 0, $2101 = 0, $2102 = 0, $2103 = 0, $2104 = 0, $2105 = 0;
 var $2106 = 0, $2107 = 0, $2108 = 0, $2109 = 0, $211 = 0, $2110 = 0, $2111 = 0, $2112 = 0, $2113 = 0, $2114 = 0, $2115 = 0, $2116 = 0, $2117 = 0, $2118 = 0, $2119 = 0, $212 = 0, $2120 = 0, $2121 = 0, $2122 = 0, $2123 = 0;
 var $2124 = 0, $2125 = 0, $2126 = 0, $2127 = 0, $2128 = 0, $2129 = 0, $213 = 0, $2130 = 0, $2131 = 0, $2132 = 0, $2133 = 0, $2134 = 0, $2135 = 0, $2136 = 0, $2137 = 0, $2138 = 0, $2139 = 0, $214 = 0, $2140 = 0, $2141 = 0;
 var $2142 = 0, $2143 = 0, $2144 = 0, $2145 = 0, $2146 = 0, $2147 = 0, $2148 = 0, $2149 = 0, $215 = 0, $2150 = 0, $2151 = 0, $2152 = 0, $2153 = 0, $2154 = 0, $2155 = 0, $2156 = 0, $2157 = 0, $2158 = 0, $2159 = 0, $216 = 0;
 var $2160 = 0, $2161 = 0, $2162 = 0, $2163 = 0, $2164 = 0, $2165 = 0, $2166 = 0, $2167 = 0, $2168 = 0, $2169 = 0, $217 = 0, $2170 = 0, $2171 = 0, $2172 = 0, $2173 = 0, $2174 = 0, $2175 = 0, $2176 = 0, $2177 = 0, $2178 = 0;
 var $2179 = 0, $218 = 0, $2180 = 0, $2181 = 0, $2182 = 0, $2183 = 0, $2184 = 0, $2185 = 0, $2186 = 0, $2187 = 0, $2188 = 0, $2189 = 0, $219 = 0, $2190 = 0, $2191 = 0, $2192 = 0, $2193 = 0, $2194 = 0, $2195 = 0, $2196 = 0;
 var $2197 = 0, $2198 = 0, $2199 = 0, $22 = 0, $220 = 0, $2200 = 0, $2201 = 0, $2202 = 0, $2203 = 0, $2204 = 0, $2205 = 0, $2206 = 0, $2207 = 0, $2208 = 0, $2209 = 0, $221 = 0, $2210 = 0, $2211 = 0, $2212 = 0, $2213 = 0;
 var $2214 = 0, $2215 = 0, $2216 = 0, $2217 = 0, $2218 = 0, $2219 = 0, $222 = 0, $2220 = 0, $2221 = 0, $2222 = 0, $2223 = 0, $2224 = 0, $2225 = 0, $2226 = 0, $2227 = 0, $2228 = 0, $2229 = 0, $223 = 0, $2230 = 0, $2231 = 0;
 var $2232 = 0, $2233 = 0, $2234 = 0, $2235 = 0, $2236 = 0, $2237 = 0, $2238 = 0, $2239 = 0, $224 = 0, $2240 = 0, $2241 = 0, $2242 = 0, $2243 = 0, $2244 = 0, $2245 = 0, $2246 = 0, $2247 = 0, $2248 = 0, $2249 = 0, $225 = 0;
 var $2250 = 0, $2251 = 0, $2252 = 0, $2253 = 0, $2254 = 0, $2255 = 0, $2256 = 0, $2257 = 0, $2258 = 0, $2259 = 0, $226 = 0, $2260 = 0, $2261 = 0, $2262 = 0, $2263 = 0, $2264 = 0, $2265 = 0, $2266 = 0, $2267 = 0, $2268 = 0;
 var $2269 = 0, $227 = 0, $2270 = 0, $2271 = 0, $2272 = 0, $2273 = 0, $2274 = 0, $2275 = 0, $2276 = 0, $2277 = 0, $2278 = 0, $2279 = 0, $228 = 0, $2280 = 0, $2281 = 0, $2282 = 0, $2283 = 0, $2284 = 0, $2285 = 0, $2286 = 0;
 var $2287 = 0, $2288 = 0, $2289 = 0, $229 = 0, $2290 = 0, $2291 = 0, $2292 = 0, $2293 = 0, $2294 = 0, $2295 = 0, $2296 = 0, $2297 = 0, $2298 = 0, $2299 = 0, $23 = 0, $230 = 0, $2300 = 0, $2301 = 0, $2302 = 0, $2303 = 0;
 var $2304 = 0, $2305 = 0, $2306 = 0, $2307 = 0, $2308 = 0, $2309 = 0, $231 = 0, $2310 = 0, $2311 = 0, $2312 = 0, $2313 = 0, $2314 = 0, $2315 = 0, $2316 = 0, $2317 = 0, $2318 = 0, $2319 = 0, $232 = 0, $2320 = 0, $2321 = 0;
 var $2322 = 0, $2323 = 0, $2324 = 0, $2325 = 0, $2326 = 0, $2327 = 0, $2328 = 0, $2329 = 0, $233 = 0, $2330 = 0, $2331 = 0, $2332 = 0, $2333 = 0, $2334 = 0, $2335 = 0, $2336 = 0, $2337 = 0, $2338 = 0, $2339 = 0, $234 = 0;
 var $2340 = 0, $2341 = 0, $2342 = 0, $2343 = 0, $2344 = 0, $2345 = 0, $2346 = 0, $2347 = 0, $2348 = 0, $2349 = 0, $235 = 0, $2350 = 0, $2351 = 0, $2352 = 0, $2353 = 0, $2354 = 0, $2355 = 0, $2356 = 0, $2357 = 0, $2358 = 0;
 var $2359 = 0, $236 = 0, $2360 = 0, $2361 = 0, $2362 = 0, $2363 = 0, $2364 = 0, $2365 = 0, $2366 = 0, $2367 = 0, $2368 = 0, $2369 = 0, $237 = 0, $2370 = 0, $2371 = 0, $2372 = 0, $2373 = 0, $2374 = 0, $2375 = 0, $2376 = 0;
 var $2377 = 0, $2378 = 0, $2379 = 0, $238 = 0, $2380 = 0, $2381 = 0, $2382 = 0, $2383 = 0, $2384 = 0, $2385 = 0, $2386 = 0, $2387 = 0, $2388 = 0, $2389 = 0, $239 = 0, $2390 = 0, $2391 = 0, $2392 = 0, $2393 = 0, $2394 = 0;
 var $2395 = 0, $2396 = 0, $2397 = 0, $2398 = 0, $2399 = 0, $24 = 0, $240 = 0, $2400 = 0, $2401 = 0, $2402 = 0, $2403 = 0, $2404 = 0, $2405 = 0, $2406 = 0, $2407 = 0, $2408 = 0, $2409 = 0, $241 = 0, $2410 = 0, $2411 = 0;
 var $2412 = 0, $2413 = 0, $2414 = 0, $2415 = 0, $2416 = 0, $2417 = 0, $2418 = 0, $2419 = 0, $242 = 0, $2420 = 0, $2421 = 0, $2422 = 0, $2423 = 0, $2424 = 0, $2425 = 0, $2426 = 0, $2427 = 0, $2428 = 0, $2429 = 0, $243 = 0;
 var $2430 = 0, $2431 = 0, $2432 = 0, $2433 = 0, $2434 = 0, $2435 = 0, $2436 = 0, $2437 = 0, $2438 = 0, $2439 = 0, $244 = 0, $2440 = 0, $2441 = 0, $2442 = 0, $2443 = 0, $2444 = 0, $2445 = 0, $2446 = 0, $2447 = 0, $2448 = 0;
 var $2449 = 0, $245 = 0, $2450 = 0, $2451 = 0, $2452 = 0, $2453 = 0, $2454 = 0, $2455 = 0, $2456 = 0, $2457 = 0, $2458 = 0, $2459 = 0, $246 = 0, $2460 = 0, $2461 = 0, $2462 = 0, $2463 = 0, $2464 = 0, $2465 = 0, $2466 = 0;
 var $2467 = 0, $2468 = 0, $2469 = 0, $247 = 0, $2470 = 0, $2471 = 0, $2472 = 0, $2473 = 0, $2474 = 0, $2475 = 0, $2476 = 0, $2477 = 0, $2478 = 0, $2479 = 0, $248 = 0, $2480 = 0, $2481 = 0, $2482 = 0, $2483 = 0, $2484 = 0;
 var $2485 = 0, $2486 = 0, $2487 = 0, $2488 = 0, $2489 = 0, $249 = 0, $2490 = 0, $2491 = 0, $2492 = 0, $2493 = 0, $2494 = 0, $2495 = 0, $2496 = 0, $2497 = 0, $2498 = 0, $2499 = 0, $25 = 0, $250 = 0, $2500 = 0, $2501 = 0;
 var $2502 = 0, $2503 = 0, $2504 = 0, $2505 = 0, $2506 = 0, $2507 = 0, $2508 = 0, $2509 = 0, $251 = 0, $2510 = 0, $2511 = 0, $2512 = 0, $2513 = 0, $2514 = 0, $2515 = 0, $2516 = 0, $2517 = 0, $2518 = 0, $2519 = 0, $252 = 0;
 var $2520 = 0, $2521 = 0, $2522 = 0, $2523 = 0, $2524 = 0, $2525 = 0, $2526 = 0, $2527 = 0, $2528 = 0, $2529 = 0, $253 = 0, $2530 = 0, $2531 = 0, $2532 = 0, $2533 = 0, $2534 = 0, $2535 = 0, $2536 = 0, $2537 = 0, $2538 = 0;
 var $2539 = 0, $254 = 0, $2540 = 0, $2541 = 0, $2542 = 0, $2543 = 0, $2544 = 0, $2545 = 0, $2546 = 0, $2547 = 0, $2548 = 0, $2549 = 0, $255 = 0, $2550 = 0, $2551 = 0, $2552 = 0, $2553 = 0, $2554 = 0, $2555 = 0, $2556 = 0;
 var $2557 = 0, $2558 = 0, $2559 = 0, $256 = 0, $2560 = 0, $2561 = 0, $2562 = 0, $2563 = 0, $2564 = 0, $2565 = 0, $2566 = 0, $2567 = 0, $2568 = 0, $2569 = 0, $257 = 0, $2570 = 0, $2571 = 0, $2572 = 0, $2573 = 0, $2574 = 0;
 var $2575 = 0, $2576 = 0, $2577 = 0, $2578 = 0, $2579 = 0, $258 = 0, $2580 = 0, $2581 = 0, $2582 = 0, $2583 = 0, $2584 = 0, $2585 = 0, $2586 = 0, $2587 = 0, $2588 = 0, $2589 = 0, $259 = 0, $2590 = 0, $2591 = 0, $2592 = 0;
 var $2593 = 0, $2594 = 0, $2595 = 0, $2596 = 0, $2597 = 0, $2598 = 0, $2599 = 0, $26 = 0, $260 = 0, $2600 = 0, $2601 = 0, $2602 = 0, $2603 = 0, $2604 = 0, $2605 = 0, $2606 = 0, $2607 = 0, $2608 = 0, $2609 = 0, $261 = 0;
 var $2610 = 0, $2611 = 0, $2612 = 0, $2613 = 0, $2614 = 0, $2615 = 0, $2616 = 0, $2617 = 0, $2618 = 0, $2619 = 0, $262 = 0, $2620 = 0, $2621 = 0, $2622 = 0, $2623 = 0, $2624 = 0, $2625 = 0, $2626 = 0, $2627 = 0, $2628 = 0;
 var $2629 = 0, $263 = 0, $2630 = 0, $2631 = 0, $2632 = 0, $2633 = 0, $2634 = 0, $2635 = 0, $2636 = 0, $2637 = 0, $2638 = 0, $2639 = 0, $264 = 0, $2640 = 0, $2641 = 0, $2642 = 0, $2643 = 0, $2644 = 0, $2645 = 0, $2646 = 0;
 var $2647 = 0, $2648 = 0, $2649 = 0, $265 = 0, $2650 = 0, $2651 = 0, $2652 = 0, $2653 = 0, $2654 = 0, $2655 = 0, $2656 = 0, $2657 = 0, $2658 = 0, $2659 = 0, $266 = 0, $2660 = 0, $2661 = 0, $2662 = 0, $2663 = 0, $2664 = 0;
 var $2665 = 0, $2666 = 0, $2667 = 0, $2668 = 0, $2669 = 0, $267 = 0, $2670 = 0, $2671 = 0, $2672 = 0, $2673 = 0, $2674 = 0, $2675 = 0, $2676 = 0, $2677 = 0, $2678 = 0, $2679 = 0, $268 = 0, $2680 = 0, $2681 = 0, $2682 = 0;
 var $2683 = 0, $2684 = 0, $2685 = 0, $2686 = 0, $2687 = 0, $2688 = 0, $2689 = 0, $269 = 0, $2690 = 0, $2691 = 0, $2692 = 0, $2693 = 0, $2694 = 0, $2695 = 0, $2696 = 0, $2697 = 0, $2698 = 0, $2699 = 0, $27 = 0, $270 = 0;
 var $2700 = 0, $2701 = 0, $2702 = 0, $2703 = 0, $2704 = 0, $2705 = 0, $2706 = 0, $2707 = 0, $2708 = 0, $2709 = 0, $271 = 0, $2710 = 0, $2711 = 0, $2712 = 0, $2713 = 0, $2714 = 0, $2715 = 0, $2716 = 0, $2717 = 0, $2718 = 0;
 var $2719 = 0, $272 = 0, $2720 = 0, $2721 = 0, $2722 = 0, $2723 = 0, $2724 = 0, $2725 = 0, $2726 = 0, $2727 = 0, $2728 = 0, $2729 = 0, $273 = 0, $2730 = 0, $2731 = 0, $2732 = 0, $2733 = 0, $2734 = 0, $2735 = 0, $2736 = 0;
 var $2737 = 0, $2738 = 0, $2739 = 0, $274 = 0, $2740 = 0, $2741 = 0, $2742 = 0, $2743 = 0, $2744 = 0, $2745 = 0, $2746 = 0, $2747 = 0, $2748 = 0, $2749 = 0, $275 = 0, $2750 = 0, $2751 = 0, $2752 = 0, $2753 = 0, $2754 = 0;
 var $2755 = 0, $2756 = 0, $2757 = 0, $2758 = 0, $2759 = 0, $276 = 0, $2760 = 0, $2761 = 0, $2762 = 0, $2763 = 0, $2764 = 0, $2765 = 0, $2766 = 0, $2767 = 0, $2768 = 0, $2769 = 0, $277 = 0, $2770 = 0, $2771 = 0, $2772 = 0;
 var $2773 = 0, $2774 = 0, $2775 = 0, $2776 = 0, $2777 = 0, $2778 = 0, $2779 = 0, $278 = 0, $2780 = 0, $2781 = 0, $2782 = 0, $2783 = 0, $2784 = 0, $2785 = 0, $2786 = 0, $2787 = 0, $2788 = 0, $2789 = 0, $279 = 0, $2790 = 0;
 var $2791 = 0, $2792 = 0, $2793 = 0, $2794 = 0, $2795 = 0, $2796 = 0, $2797 = 0, $2798 = 0, $2799 = 0, $28 = 0, $280 = 0, $2800 = 0, $2801 = 0, $2802 = 0, $2803 = 0, $2804 = 0, $2805 = 0, $2806 = 0, $2807 = 0, $2808 = 0;
 var $2809 = 0, $281 = 0, $2810 = 0, $2811 = 0, $2812 = 0, $2813 = 0, $2814 = 0, $2815 = 0, $2816 = 0, $2817 = 0, $2818 = 0, $2819 = 0, $282 = 0, $2820 = 0, $2821 = 0, $2822 = 0, $2823 = 0, $2824 = 0, $2825 = 0, $2826 = 0;
 var $2827 = 0, $2828 = 0, $2829 = 0, $283 = 0, $2830 = 0, $2831 = 0, $2832 = 0, $2833 = 0, $2834 = 0, $2835 = 0, $2836 = 0, $2837 = 0, $2838 = 0, $2839 = 0, $284 = 0, $2840 = 0, $2841 = 0, $2842 = 0, $2843 = 0, $2844 = 0;
 var $2845 = 0, $2846 = 0, $2847 = 0, $2848 = 0, $2849 = 0, $285 = 0, $2850 = 0, $2851 = 0, $2852 = 0, $2853 = 0, $2854 = 0, $2855 = 0, $2856 = 0, $2857 = 0, $2858 = 0, $2859 = 0, $286 = 0, $2860 = 0, $2861 = 0, $2862 = 0;
 var $2863 = 0, $2864 = 0, $2865 = 0, $2866 = 0, $2867 = 0, $2868 = 0, $2869 = 0, $287 = 0, $2870 = 0, $2871 = 0, $2872 = 0, $2873 = 0, $2874 = 0, $2875 = 0, $2876 = 0, $2877 = 0, $2878 = 0, $2879 = 0, $288 = 0, $2880 = 0;
 var $2881 = 0, $2882 = 0, $2883 = 0, $2884 = 0, $2885 = 0, $2886 = 0, $2887 = 0, $2888 = 0, $2889 = 0, $289 = 0, $2890 = 0, $2891 = 0, $2892 = 0, $2893 = 0, $2894 = 0, $2895 = 0, $2896 = 0, $2897 = 0, $2898 = 0, $2899 = 0;
 var $29 = 0, $290 = 0, $2900 = 0, $2901 = 0, $2902 = 0, $2903 = 0, $2904 = 0, $2905 = 0, $2906 = 0, $2907 = 0, $2908 = 0, $2909 = 0, $291 = 0, $2910 = 0, $2911 = 0, $2912 = 0, $2913 = 0, $2914 = 0, $2915 = 0, $2916 = 0;
 var $2917 = 0, $2918 = 0, $2919 = 0, $292 = 0, $2920 = 0, $2921 = 0, $2922 = 0, $2923 = 0, $2924 = 0, $2925 = 0, $2926 = 0, $2927 = 0, $2928 = 0, $2929 = 0, $293 = 0, $2930 = 0, $2931 = 0, $2932 = 0, $2933 = 0, $2934 = 0;
 var $2935 = 0, $2936 = 0, $2937 = 0, $2938 = 0, $2939 = 0, $294 = 0, $2940 = 0, $2941 = 0, $2942 = 0, $2943 = 0, $2944 = 0, $2945 = 0, $2946 = 0, $2947 = 0, $2948 = 0, $2949 = 0, $295 = 0, $2950 = 0, $2951 = 0, $2952 = 0;
 var $2953 = 0, $2954 = 0, $2955 = 0, $2956 = 0, $2957 = 0, $2958 = 0, $2959 = 0, $296 = 0, $2960 = 0, $2961 = 0, $2962 = 0, $2963 = 0, $2964 = 0, $2965 = 0, $2966 = 0, $2967 = 0, $2968 = 0, $2969 = 0, $297 = 0, $2970 = 0;
 var $2971 = 0, $2972 = 0, $2973 = 0, $2974 = 0, $2975 = 0, $2976 = 0, $2977 = 0, $2978 = 0, $2979 = 0, $298 = 0, $2980 = 0, $2981 = 0, $2982 = 0, $2983 = 0, $2984 = 0, $2985 = 0, $2986 = 0, $2987 = 0, $2988 = 0, $2989 = 0;
 var $299 = 0, $2990 = 0, $2991 = 0, $2992 = 0, $2993 = 0, $2994 = 0, $2995 = 0, $2996 = 0, $2997 = 0, $2998 = 0, $2999 = 0, $3 = 0, $30 = 0, $300 = 0, $3000 = 0, $3001 = 0, $3002 = 0, $3003 = 0, $3004 = 0, $3005 = 0;
 var $3006 = 0, $3007 = 0, $3008 = 0, $3009 = 0, $301 = 0, $3010 = 0, $3011 = 0, $3012 = 0, $3013 = 0, $3014 = 0, $3015 = 0, $3016 = 0, $3017 = 0, $3018 = 0, $3019 = 0, $302 = 0, $3020 = 0, $3021 = 0, $3022 = 0, $3023 = 0;
 var $3024 = 0, $3025 = 0, $3026 = 0, $3027 = 0, $3028 = 0, $3029 = 0, $303 = 0, $3030 = 0, $3031 = 0, $3032 = 0, $3033 = 0, $3034 = 0, $3035 = 0, $3036 = 0, $3037 = 0, $3038 = 0, $3039 = 0, $304 = 0, $3040 = 0, $3041 = 0;
 var $3042 = 0, $3043 = 0, $3044 = 0, $3045 = 0, $3046 = 0, $3047 = 0, $3048 = 0, $3049 = 0, $305 = 0, $3050 = 0, $3051 = 0, $3052 = 0, $3053 = 0, $3054 = 0, $3055 = 0, $3056 = 0, $3057 = 0, $3058 = 0, $3059 = 0, $306 = 0;
 var $3060 = 0, $3061 = 0, $3062 = 0, $3063 = 0, $3064 = 0, $3065 = 0, $3066 = 0, $3067 = 0, $3068 = 0, $3069 = 0, $307 = 0, $3070 = 0, $3071 = 0, $3072 = 0, $3073 = 0, $3074 = 0, $3075 = 0, $3076 = 0, $3077 = 0, $3078 = 0;
 var $3079 = 0, $308 = 0, $3080 = 0, $3081 = 0, $3082 = 0, $3083 = 0, $3084 = 0, $3085 = 0, $3086 = 0, $3087 = 0, $3088 = 0, $3089 = 0, $309 = 0, $3090 = 0, $3091 = 0, $3092 = 0, $3093 = 0, $3094 = 0, $3095 = 0, $3096 = 0;
 var $3097 = 0, $3098 = 0, $3099 = 0, $31 = 0, $310 = 0, $3100 = 0, $3101 = 0, $3102 = 0, $3103 = 0, $3104 = 0, $3105 = 0, $3106 = 0, $3107 = 0, $3108 = 0, $3109 = 0, $311 = 0, $3110 = 0, $3111 = 0, $3112 = 0, $3113 = 0;
 var $3114 = 0, $3115 = 0, $3116 = 0, $3117 = 0, $3118 = 0, $3119 = 0, $312 = 0, $3120 = 0, $3121 = 0, $3122 = 0, $3123 = 0, $3124 = 0, $3125 = 0, $3126 = 0, $3127 = 0, $3128 = 0, $3129 = 0, $313 = 0, $3130 = 0, $3131 = 0;
 var $3132 = 0, $3133 = 0, $3134 = 0, $3135 = 0, $3136 = 0, $3137 = 0, $3138 = 0, $3139 = 0, $314 = 0, $3140 = 0, $3141 = 0, $3142 = 0, $3143 = 0, $3144 = 0, $3145 = 0, $3146 = 0, $3147 = 0, $3148 = 0, $3149 = 0, $315 = 0;
 var $3150 = 0, $3151 = 0, $3152 = 0, $3153 = 0, $3154 = 0, $3155 = 0, $3156 = 0, $3157 = 0, $3158 = 0, $3159 = 0, $316 = 0, $3160 = 0, $3161 = 0, $3162 = 0, $3163 = 0, $3164 = 0, $3165 = 0, $3166 = 0, $3167 = 0, $3168 = 0;
 var $3169 = 0, $317 = 0, $3170 = 0, $3171 = 0, $3172 = 0, $3173 = 0, $3174 = 0, $3175 = 0, $3176 = 0, $3177 = 0, $3178 = 0, $3179 = 0, $318 = 0, $3180 = 0, $3181 = 0, $3182 = 0, $3183 = 0, $3184 = 0, $3185 = 0, $3186 = 0;
 var $3187 = 0, $3188 = 0, $3189 = 0, $319 = 0, $3190 = 0, $3191 = 0, $3192 = 0, $3193 = 0, $3194 = 0, $3195 = 0, $3196 = 0, $3197 = 0, $3198 = 0, $3199 = 0, $32 = 0, $320 = 0, $3200 = 0, $3201 = 0, $3202 = 0, $3203 = 0;
 var $3204 = 0, $3205 = 0, $3206 = 0, $3207 = 0, $3208 = 0, $3209 = 0, $321 = 0, $3210 = 0, $3211 = 0, $3212 = 0, $3213 = 0, $3214 = 0, $3215 = 0, $3216 = 0, $3217 = 0, $3218 = 0, $3219 = 0, $322 = 0, $3220 = 0, $3221 = 0;
 var $3222 = 0, $3223 = 0, $3224 = 0, $3225 = 0, $3226 = 0, $3227 = 0, $3228 = 0, $3229 = 0, $323 = 0, $3230 = 0, $3231 = 0, $3232 = 0, $3233 = 0, $3234 = 0, $3235 = 0, $3236 = 0, $3237 = 0, $3238 = 0, $3239 = 0, $324 = 0;
 var $3240 = 0, $3241 = 0, $3242 = 0, $3243 = 0, $3244 = 0, $3245 = 0, $3246 = 0, $3247 = 0, $3248 = 0, $3249 = 0, $325 = 0, $3250 = 0, $3251 = 0, $3252 = 0, $3253 = 0, $3254 = 0, $3255 = 0, $3256 = 0, $3257 = 0, $3258 = 0;
 var $3259 = 0, $326 = 0, $3260 = 0, $3261 = 0, $3262 = 0, $3263 = 0, $3264 = 0, $3265 = 0, $3266 = 0, $3267 = 0, $3268 = 0, $3269 = 0, $327 = 0, $3270 = 0, $3271 = 0, $3272 = 0, $3273 = 0, $3274 = 0, $3275 = 0, $3276 = 0;
 var $3277 = 0, $3278 = 0, $3279 = 0, $328 = 0, $3280 = 0, $3281 = 0, $3282 = 0, $3283 = 0, $3284 = 0, $3285 = 0, $3286 = 0, $3287 = 0, $3288 = 0, $3289 = 0, $329 = 0, $3290 = 0, $3291 = 0, $3292 = 0, $3293 = 0, $3294 = 0;
 var $3295 = 0, $3296 = 0, $3297 = 0, $3298 = 0, $3299 = 0, $33 = 0, $330 = 0, $3300 = 0, $3301 = 0, $3302 = 0, $3303 = 0, $3304 = 0, $3305 = 0, $3306 = 0, $3307 = 0, $3308 = 0, $3309 = 0, $331 = 0, $3310 = 0, $3311 = 0;
 var $3312 = 0, $3313 = 0, $3314 = 0, $3315 = 0, $3316 = 0, $3317 = 0, $3318 = 0, $3319 = 0, $332 = 0, $3320 = 0, $3321 = 0, $3322 = 0, $3323 = 0, $3324 = 0, $3325 = 0, $3326 = 0, $3327 = 0, $3328 = 0, $3329 = 0, $333 = 0;
 var $3330 = 0, $3331 = 0, $3332 = 0, $3333 = 0, $3334 = 0, $3335 = 0, $3336 = 0, $3337 = 0, $3338 = 0, $3339 = 0, $334 = 0, $3340 = 0, $3341 = 0, $3342 = 0, $3343 = 0, $3344 = 0, $3345 = 0, $3346 = 0, $3347 = 0, $3348 = 0;
 var $3349 = 0, $335 = 0, $3350 = 0, $3351 = 0, $3352 = 0, $3353 = 0, $3354 = 0, $3355 = 0, $3356 = 0, $3357 = 0, $3358 = 0, $3359 = 0, $336 = 0, $3360 = 0, $3361 = 0, $3362 = 0, $3363 = 0, $3364 = 0, $3365 = 0, $3366 = 0;
 var $3367 = 0, $3368 = 0, $3369 = 0, $337 = 0, $3370 = 0, $3371 = 0, $3372 = 0, $3373 = 0, $3374 = 0, $3375 = 0, $3376 = 0, $3377 = 0, $3378 = 0, $3379 = 0, $338 = 0, $3380 = 0, $3381 = 0, $3382 = 0, $3383 = 0, $3384 = 0;
 var $3385 = 0, $3386 = 0, $3387 = 0, $3388 = 0, $3389 = 0, $339 = 0, $3390 = 0, $3391 = 0, $3392 = 0, $3393 = 0, $3394 = 0, $3395 = 0, $3396 = 0, $3397 = 0, $3398 = 0, $3399 = 0, $34 = 0, $340 = 0, $3400 = 0, $3401 = 0;
 var $3402 = 0, $3403 = 0, $3404 = 0, $3405 = 0, $3406 = 0, $3407 = 0, $3408 = 0, $3409 = 0, $341 = 0, $3410 = 0, $3411 = 0, $3412 = 0, $3413 = 0, $3414 = 0, $3415 = 0, $3416 = 0, $3417 = 0, $3418 = 0, $3419 = 0, $342 = 0;
 var $3420 = 0, $3421 = 0, $3422 = 0, $3423 = 0, $3424 = 0, $3425 = 0, $3426 = 0, $3427 = 0, $3428 = 0, $3429 = 0, $343 = 0, $3430 = 0, $3431 = 0, $3432 = 0, $3433 = 0, $3434 = 0, $3435 = 0, $3436 = 0, $3437 = 0, $3438 = 0;
 var $3439 = 0, $344 = 0, $3440 = 0, $3441 = 0, $3442 = 0, $3443 = 0, $3444 = 0, $3445 = 0, $3446 = 0, $3447 = 0, $3448 = 0, $3449 = 0, $345 = 0, $3450 = 0, $3451 = 0, $3452 = 0, $3453 = 0, $3454 = 0, $3455 = 0, $3456 = 0;
 var $3457 = 0, $3458 = 0, $3459 = 0, $346 = 0, $3460 = 0, $3461 = 0, $3462 = 0, $3463 = 0, $3464 = 0, $3465 = 0, $3466 = 0, $3467 = 0, $3468 = 0, $3469 = 0, $347 = 0, $3470 = 0, $3471 = 0, $3472 = 0, $3473 = 0, $3474 = 0;
 var $3475 = 0, $3476 = 0, $3477 = 0, $3478 = 0, $3479 = 0, $348 = 0, $3480 = 0, $3481 = 0, $3482 = 0, $3483 = 0, $3484 = 0, $3485 = 0, $3486 = 0, $3487 = 0, $3488 = 0, $3489 = 0, $349 = 0, $3490 = 0, $3491 = 0, $3492 = 0;
 var $3493 = 0, $3494 = 0, $3495 = 0, $3496 = 0, $3497 = 0, $3498 = 0, $3499 = 0, $35 = 0, $350 = 0, $3500 = 0, $3501 = 0, $3502 = 0, $3503 = 0, $3504 = 0, $3505 = 0, $3506 = 0, $3507 = 0, $3508 = 0, $3509 = 0, $351 = 0;
 var $3510 = 0, $3511 = 0, $3512 = 0, $3513 = 0, $3514 = 0, $3515 = 0, $3516 = 0, $3517 = 0, $3518 = 0, $3519 = 0, $352 = 0, $3520 = 0, $3521 = 0, $3522 = 0, $3523 = 0, $3524 = 0, $3525 = 0, $3526 = 0, $3527 = 0, $3528 = 0;
 var $3529 = 0, $353 = 0, $3530 = 0, $3531 = 0, $3532 = 0, $3533 = 0, $3534 = 0, $3535 = 0, $3536 = 0, $3537 = 0, $3538 = 0, $3539 = 0, $354 = 0, $3540 = 0, $3541 = 0, $3542 = 0, $3543 = 0, $3544 = 0, $3545 = 0, $3546 = 0;
 var $3547 = 0, $3548 = 0, $3549 = 0, $355 = 0, $3550 = 0, $3551 = 0, $3552 = 0, $3553 = 0, $3554 = 0, $3555 = 0, $3556 = 0, $3557 = 0, $3558 = 0, $3559 = 0, $356 = 0, $3560 = 0, $3561 = 0, $3562 = 0, $3563 = 0, $3564 = 0;
 var $3565 = 0, $3566 = 0, $3567 = 0, $3568 = 0, $3569 = 0, $357 = 0, $3570 = 0, $3571 = 0, $3572 = 0, $3573 = 0, $3574 = 0, $3575 = 0, $3576 = 0, $3577 = 0, $3578 = 0, $3579 = 0, $358 = 0, $3580 = 0, $3581 = 0, $3582 = 0;
 var $3583 = 0, $3584 = 0, $3585 = 0, $3586 = 0, $3587 = 0, $3588 = 0, $3589 = 0, $359 = 0, $3590 = 0, $3591 = 0, $3592 = 0, $3593 = 0, $3594 = 0, $3595 = 0, $3596 = 0, $3597 = 0, $3598 = 0, $3599 = 0, $36 = 0, $360 = 0;
 var $3600 = 0, $3601 = 0, $3602 = 0, $3603 = 0, $3604 = 0, $3605 = 0, $3606 = 0, $3607 = 0, $3608 = 0, $3609 = 0, $361 = 0, $3610 = 0, $3611 = 0, $3612 = 0, $3613 = 0, $3614 = 0, $3615 = 0, $3616 = 0, $3617 = 0, $3618 = 0;
 var $3619 = 0, $362 = 0, $3620 = 0, $3621 = 0, $3622 = 0, $3623 = 0, $3624 = 0, $3625 = 0, $3626 = 0, $3627 = 0, $3628 = 0, $3629 = 0, $363 = 0, $3630 = 0, $3631 = 0, $3632 = 0, $3633 = 0, $3634 = 0, $3635 = 0, $3636 = 0;
 var $3637 = 0, $3638 = 0, $3639 = 0, $364 = 0, $3640 = 0, $3641 = 0, $3642 = 0, $3643 = 0, $3644 = 0, $3645 = 0, $3646 = 0, $3647 = 0, $3648 = 0, $3649 = 0, $365 = 0, $3650 = 0, $3651 = 0, $3652 = 0, $3653 = 0, $3654 = 0;
 var $3655 = 0, $3656 = 0, $3657 = 0, $3658 = 0, $3659 = 0, $366 = 0, $3660 = 0, $3661 = 0, $3662 = 0, $3663 = 0, $3664 = 0, $3665 = 0, $3666 = 0, $3667 = 0, $3668 = 0, $3669 = 0, $367 = 0, $3670 = 0, $3671 = 0, $3672 = 0;
 var $3673 = 0, $3674 = 0, $3675 = 0, $3676 = 0, $3677 = 0, $3678 = 0, $3679 = 0, $368 = 0, $3680 = 0, $3681 = 0, $3682 = 0, $3683 = 0, $3684 = 0, $3685 = 0, $3686 = 0, $3687 = 0, $3688 = 0, $3689 = 0, $369 = 0, $3690 = 0;
 var $3691 = 0, $3692 = 0, $3693 = 0, $3694 = 0, $3695 = 0, $3696 = 0, $3697 = 0, $3698 = 0, $3699 = 0, $37 = 0, $370 = 0, $3700 = 0, $3701 = 0, $3702 = 0, $3703 = 0, $3704 = 0, $3705 = 0, $3706 = 0, $3707 = 0, $3708 = 0;
 var $3709 = 0, $371 = 0, $3710 = 0, $3711 = 0, $3712 = 0, $3713 = 0, $3714 = 0, $3715 = 0, $3716 = 0, $3717 = 0, $3718 = 0, $3719 = 0, $372 = 0, $3720 = 0, $3721 = 0, $3722 = 0, $3723 = 0, $3724 = 0, $3725 = 0, $3726 = 0;
 var $3727 = 0, $3728 = 0, $3729 = 0, $373 = 0, $3730 = 0, $3731 = 0, $3732 = 0, $3733 = 0, $3734 = 0, $3735 = 0, $3736 = 0, $3737 = 0, $3738 = 0, $3739 = 0, $374 = 0, $3740 = 0, $3741 = 0, $3742 = 0, $3743 = 0, $3744 = 0;
 var $3745 = 0, $3746 = 0, $3747 = 0, $3748 = 0, $3749 = 0, $375 = 0, $3750 = 0, $3751 = 0, $3752 = 0, $3753 = 0, $3754 = 0, $3755 = 0, $3756 = 0, $3757 = 0, $3758 = 0, $3759 = 0, $376 = 0, $3760 = 0, $3761 = 0, $3762 = 0;
 var $3763 = 0, $3764 = 0, $3765 = 0, $3766 = 0, $3767 = 0, $3768 = 0, $3769 = 0, $377 = 0, $3770 = 0, $3771 = 0, $3772 = 0, $3773 = 0, $3774 = 0, $3775 = 0, $3776 = 0, $3777 = 0, $3778 = 0, $3779 = 0, $378 = 0, $3780 = 0;
 var $3781 = 0, $3782 = 0, $3783 = 0, $3784 = 0, $3785 = 0, $3786 = 0, $3787 = 0, $3788 = 0, $3789 = 0, $379 = 0, $3790 = 0, $3791 = 0, $3792 = 0, $3793 = 0, $3794 = 0, $3795 = 0, $3796 = 0, $3797 = 0, $3798 = 0, $3799 = 0;
 var $38 = 0, $380 = 0, $3800 = 0, $3801 = 0, $3802 = 0, $3803 = 0, $3804 = 0, $3805 = 0, $3806 = 0, $3807 = 0, $3808 = 0, $3809 = 0, $381 = 0, $3810 = 0, $3811 = 0, $3812 = 0, $3813 = 0, $3814 = 0, $3815 = 0, $3816 = 0;
 var $3817 = 0, $3818 = 0, $3819 = 0, $382 = 0, $3820 = 0, $3821 = 0, $3822 = 0, $3823 = 0, $3824 = 0, $3825 = 0, $3826 = 0, $3827 = 0, $3828 = 0, $3829 = 0, $383 = 0, $3830 = 0, $3831 = 0, $3832 = 0, $3833 = 0, $3834 = 0;
 var $3835 = 0, $3836 = 0, $3837 = 0, $3838 = 0, $3839 = 0, $384 = 0, $3840 = 0, $3841 = 0, $3842 = 0, $3843 = 0, $3844 = 0, $3845 = 0, $3846 = 0, $3847 = 0, $3848 = 0, $3849 = 0, $385 = 0, $3850 = 0, $3851 = 0, $3852 = 0;
 var $3853 = 0, $3854 = 0, $3855 = 0, $3856 = 0, $3857 = 0, $3858 = 0, $3859 = 0, $386 = 0, $3860 = 0, $3861 = 0, $3862 = 0, $3863 = 0, $3864 = 0, $3865 = 0, $3866 = 0, $3867 = 0, $3868 = 0, $3869 = 0, $387 = 0, $3870 = 0;
 var $3871 = 0, $3872 = 0, $3873 = 0, $3874 = 0, $3875 = 0, $3876 = 0, $3877 = 0, $3878 = 0, $3879 = 0, $388 = 0, $3880 = 0, $3881 = 0, $3882 = 0, $3883 = 0, $3884 = 0, $3885 = 0, $3886 = 0, $3887 = 0, $3888 = 0, $3889 = 0;
 var $389 = 0, $3890 = 0, $3891 = 0, $3892 = 0, $3893 = 0, $3894 = 0, $3895 = 0, $3896 = 0, $3897 = 0, $3898 = 0, $3899 = 0, $39 = 0, $390 = 0, $3900 = 0, $3901 = 0, $3902 = 0, $3903 = 0, $3904 = 0, $3905 = 0, $3906 = 0;
 var $3907 = 0, $3908 = 0, $3909 = 0, $391 = 0, $3910 = 0, $3911 = 0, $3912 = 0, $3913 = 0, $3914 = 0, $3915 = 0, $3916 = 0, $3917 = 0, $3918 = 0, $3919 = 0, $392 = 0, $3920 = 0, $3921 = 0, $3922 = 0, $3923 = 0, $3924 = 0;
 var $3925 = 0, $3926 = 0, $3927 = 0, $3928 = 0, $3929 = 0, $393 = 0, $3930 = 0, $3931 = 0, $3932 = 0, $3933 = 0, $3934 = 0, $3935 = 0, $3936 = 0, $3937 = 0, $3938 = 0, $3939 = 0, $394 = 0, $3940 = 0, $3941 = 0, $3942 = 0;
 var $3943 = 0, $3944 = 0, $3945 = 0, $3946 = 0, $3947 = 0, $3948 = 0, $3949 = 0, $395 = 0, $3950 = 0, $3951 = 0, $3952 = 0, $3953 = 0, $3954 = 0, $3955 = 0, $3956 = 0, $3957 = 0, $3958 = 0, $3959 = 0, $396 = 0, $3960 = 0;
 var $3961 = 0, $3962 = 0, $3963 = 0, $3964 = 0, $3965 = 0, $3966 = 0, $3967 = 0, $3968 = 0, $3969 = 0, $397 = 0, $3970 = 0, $3971 = 0, $3972 = 0, $3973 = 0, $3974 = 0, $3975 = 0, $3976 = 0, $3977 = 0, $3978 = 0, $3979 = 0;
 var $398 = 0, $3980 = 0, $3981 = 0, $3982 = 0, $3983 = 0, $3984 = 0, $3985 = 0, $3986 = 0, $3987 = 0, $3988 = 0, $3989 = 0, $399 = 0, $3990 = 0, $3991 = 0, $3992 = 0, $3993 = 0, $3994 = 0, $3995 = 0, $3996 = 0, $3997 = 0;
 var $3998 = 0, $3999 = 0, $4 = 0, $40 = 0, $400 = 0, $4000 = 0, $4001 = 0, $4002 = 0, $4003 = 0, $4004 = 0, $4005 = 0, $4006 = 0, $4007 = 0, $4008 = 0, $4009 = 0, $401 = 0, $4010 = 0, $4011 = 0, $4012 = 0, $4013 = 0;
 var $4014 = 0, $4015 = 0, $4016 = 0, $4017 = 0, $4018 = 0, $4019 = 0, $402 = 0, $4020 = 0, $4021 = 0, $4022 = 0, $4023 = 0, $4024 = 0, $4025 = 0, $4026 = 0, $4027 = 0, $4028 = 0, $4029 = 0, $403 = 0, $4030 = 0, $4031 = 0;
 var $4032 = 0, $4033 = 0, $4034 = 0, $4035 = 0, $4036 = 0, $4037 = 0, $4038 = 0, $4039 = 0, $404 = 0, $4040 = 0, $4041 = 0, $4042 = 0, $4043 = 0, $4044 = 0, $4045 = 0, $4046 = 0, $4047 = 0, $4048 = 0, $4049 = 0, $405 = 0;
 var $4050 = 0, $4051 = 0, $4052 = 0, $4053 = 0, $4054 = 0, $4055 = 0, $4056 = 0, $4057 = 0, $4058 = 0, $4059 = 0, $406 = 0, $4060 = 0, $4061 = 0, $4062 = 0, $4063 = 0, $4064 = 0, $4065 = 0, $4066 = 0, $4067 = 0, $4068 = 0;
 var $4069 = 0, $407 = 0, $4070 = 0, $4071 = 0, $4072 = 0, $4073 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0;
 var $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0;
 var $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0;
 var $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0;
 var $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0;
 var $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0;
 var $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0;
 var $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0;
 var $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0;
 var $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0;
 var $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0;
 var $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0;
 var $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0;
 var $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0;
 var $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0;
 var $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0;
 var $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0;
 var $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0;
 var $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0;
 var $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0;
 var $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0;
 var $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0;
 var $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0;
 var $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0;
 var $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0;
 var $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0;
 var $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0;
 var $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0;
 var $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0;
 var $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0;
 var $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0;
 var $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0;
 var $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0;
 var $998 = 0, $999 = 0, $abcarpeggio$01324 = 0, $abcarpeggio$1$ph$lcssa = 0, $abcarpeggio$1$ph1273 = 0, $abcarpeggio$21229 = 0, $abcarpeggio$3 = 0, $abcarpeggio$3$lcssa = 0, $abcchord$0 = 0, $abcchord$0$lcssa = 0, $abcchord$11206 = 0, $abceffect$01315 = 0, $abceffect$1$lcssa = 0, $abceffect$1$ph1264 = 0, $abceffect$11093 = 0, $abceffect$2 = 0, $abceffect$2$lcssa = 0, $abceffect$3$ = 0, $abceffect$3$91 = 0, $abceffect$31222 = 0;
 var $abceffect$5 = 0, $abceffect$5$lcssa = 0, $abceffoper$01316 = 0, $abceffoper$1$lcssa = 0, $abceffoper$1$ph1265 = 0, $abceffoper$11095 = 0, $abceffoper$2 = 0, $abceffoper$2$lcssa = 0, $abceffoper$31223 = 0, $abceffoper$4 = 0, $abceffoper$4$lcssa = 0, $abcfermata$01323 = 0, $abcfermata$1$ph$lcssa = 0, $abcfermata$1$ph1272 = 0, $abcfermata$21228 = 0, $abcfermata$7 = 0, $abcfermata$7$lcssa = 0, $abcgrace$01331 = 0, $abcgrace$1$lcssa = 0, $abcgrace$1$ph1280 = 0;
 var $abcgrace$10 = 0, $abcgrace$10$lcssa = 0, $abcgrace$11117 = 0, $abcgrace$2 = 0, $abcgrace$3 = 0, $abcgrace$3$lcssa = 0, $abcgrace$41234 = 0, $abcgrace$5 = 0, $abcgrace$6 = 0, $abcgrace$6$lcssa = 0, $abcgrace$7 = 0, $abcgrace$8 = 0, $abcgrace$9 = 0, $abchornpipe$01329 = 0, $abchornpipe$1$ = 0, $abchornpipe$1$lcssa = 0, $abchornpipe$1$ph1278 = 0, $abchornpipe$11113 = 0, $abchornpipe$3 = 0, $abchornpipe$3$lcssa = 0;
 var $abckey$01319 = 0, $abckey$1$lcssa = 0, $abckey$1$ph1268 = 0, $abckey$11101 = 0, $abckey$2 = 0, $abckey$3 = 0, $abckey$3$lcssa = 0, $abckey$41224 = 0, $abckey$5 = 0, $abckey$5$lcssa = 0, $abcnolegato$01322 = 0, $abcnolegato$1$lcssa = 0, $abcnolegato$1$ph1271 = 0, $abcnolegato$101227 = 0, $abcnolegato$11 = 0, $abcnolegato$11107 = 0, $abcnolegato$12$ph = 0, $abcnolegato$13 = 0, $abcnolegato$14 = 0, $abcnolegato$14$lcssa = 0;
 var $abcnolegato$2 = 0, $abcnolegato$3 = 0, $abcnolegato$4 = 0, $abcnolegato$5 = 0, $abcnolegato$6 = 0, $abcnolegato$7 = 0, $abcnolegato$8 = 0, $abcnolegato$9 = 0, $abcnolegato$9$lcssa = 0, $abcnoslurs$01321 = 0, $abcnoslurs$1$ = 0, $abcnoslurs$1$83 = 0, $abcnoslurs$1$lcssa = 0, $abcnoslurs$1$ph1270 = 0, $abcnoslurs$10$ph = 0, $abcnoslurs$11 = 0, $abcnoslurs$11106 = 0, $abcnoslurs$12 = 0, $abcnoslurs$12$lcssa = 0, $abcnoslurs$3 = 0;
 var $abcnoslurs$4 = 0, $abcnoslurs$5 = 0, $abcnoslurs$6 = 0, $abcnoslurs$7 = 0, $abcnoslurs$8 = 0, $abcnoslurs$8$lcssa = 0, $abcnoslurs$9$ = 0, $abcnoslurs$91226 = 0, $abcparts = 0, $abcstate$0 = 0, $abcstate$1$ph = 0, $abcstate$11314 = 0, $abcstate$2$lcssa = 0, $abcstate$2$lcssa$lcssa = 0, $abcstate$2$ph1263 = 0, $abcstate$21092 = 0, $abcstate$5 = 0, $abcstate$6 = 0, $abcstate$7$ = 0, $abcstate$71221 = 0;
 var $abcstate$8 = 0, $abcstate$8$lcssa = 0, $abctempo$01326 = 0, $abctempo$1$lcssa = 0, $abctempo$1$lcssa$lcssa = 0, $abctempo$1$ph1275 = 0, $abctempo$11111 = 0, $abctempo$2 = 0, $abctempo$3 = 0, $abctempo$4 = 0, $abctempo$4$lcssa = 0, $abctempo$51231 = 0, $abctempo$6 = 0, $abctempo$6$lcssa = 0, $abcto$01325 = 0, $abcto$1$lcssa = 0, $abcto$1$ph1274 = 0, $abcto$11109 = 0, $abcto$2 = 0, $abcto$2$lcssa = 0;
 var $abcto$31230 = 0, $abcto$4$be = 0, $abcto$4$lcssa = 0, $abcto$41190 = 0, $abcto$5 = 0, $abcto$6 = 0, $abcto$6$lcssa = 0, $abcvol$01320 = 0, $abcvol$1$lcssa = 0, $abcvol$1$ph1269 = 0, $abcvol$11104 = 0, $abcvol$2 = 0, $abcvol$2$lcssa = 0, $abcvol$31225 = 0, $abcvol$4 = 0, $abcvol$4$lcssa = 0, $abcxcount$01335 = 0, $abcxcount$11317 = 0, $abcxcount$2$lcssa = 0, $abcxcount$2$ph1266 = 0;
 var $abcxcount$21097 = 0, $abcxcount$3 = 0, $abcxcount$4 = 0, $abcxcount$4$lcssa = 0, $abcxnumber$01318 = 0, $abcxnumber$1$lcssa = 0, $abcxnumber$1$lcssa$lcssa = 0, $abcxnumber$1$ph1267 = 0, $abcxnumber$11099 = 0, $abcxnumber$2 = 0, $abcxnumber$2$lcssa = 0, $abcxwanted$0 = 0, $abcxwanted$1$ph = 0, $barsig = 0, $bartime$01332 = 0, $bartime$1$lcssa = 0, $bartime$1$ph1281 = 0, $bartime$11119 = 0, $bartime$2 = 0, $bartime$3 = 0;
 var $bartime$3$lcssa = 0, $bartime$41235 = 0, $bartime$5 = 0, $bartime$5$lcssa = 0, $buf = 0, $buf$i = 0, $buf1 = 0, $buf2 = 0, $ch$0 = 0, $ch$1 = 0, $ch$1$lcssa = 0, $ch0$01312 = 0, $ch0$1$ph$lcssa = 0, $ch0$1$ph1261 = 0, $ch0$21220 = 0, $cnd = 0, $cnl = 0, $cnotediv$01328 = 0, $cnotediv$1$ph$lcssa = 0, $cnotediv$1$ph1277 = 0;
 var $cnotediv$21233 = 0, $cnotediv$3 = 0, $cnotediv$3$lcssa = 0, $cnotediv$4 = 0, $cnotediv$5 = 0, $cnotediv$6 = 0, $cnotediv$6$lcssa = 0, $cnotelen$01327 = 0, $cnotelen$1$ph$lcssa = 0, $cnotelen$1$ph1276 = 0, $cnotelen$21232 = 0, $cnotelen$3 = 0, $cnotelen$3$lcssa = 0, $cnotelen$4 = 0, $cnotelen$5 = 0, $cnotelen$6 = 0, $cnotelen$6$lcssa = 0, $ct$03$i$i = 0, $d$04$i$i = 0, $d$1$i$i = 0;
 var $dt$0$lcssa$i = 0, $dt$0$lcssa$i$i = 0, $dt$0$lcssa$i$i146 = 0, $dt$0$lcssa$i$i191 = 0, $dt$0$lcssa$i$i220 = 0, $dt$0$lcssa$i$i711 = 0, $dt$0$lcssa$i721 = 0, $dt$03$i = 0, $dt$03$i$i = 0, $dt$03$i$i142 = 0, $dt$03$i$i187 = 0, $dt$03$i$i216 = 0, $dt$03$i$i707 = 0, $dt$03$i717 = 0, $dt$1$i = 0, $dt$1$i$i = 0, $dt$1$i$i145 = 0, $dt$1$i$i190 = 0, $dt$1$i$i219 = 0, $dt$1$i$i710 = 0;
 var $dt$1$i720 = 0, $e$0$i = 0, $e$0$i743 = 0, $e$02$i = 0, $e$02$i$i = 0, $e$02$i$i143 = 0, $e$02$i$i188 = 0, $e$02$i$i217 = 0, $e$02$i$i708 = 0, $e$02$i718 = 0, $e$03$i = 0, $e$05$i = 0, $e$05$i741 = 0, $e$06$i = 0, $el$02$i$i = 0, $el$1$i$i = 0, $em$0$i = 0, $em$011$i = 0, $em$012$i = 0, $endptr = 0;
 var $endptr74 = 0, $ep$0$i$i = 0, $ep$0$i4$i = 0, $ep$01$i$i = 0, $ep$01$i$i$i = 0, $ep$01$i2$i = 0, $ep$03$i$i = 0, $ep$05$i$i = 0, $ep$1$i$i = 0, $ep$1$in$i$i = 0, $ep$14$pre$i$i = 0, $ep$16$i$i = 0, $es$0$lcssa$ph$i = 0, $es$0$lcssa22$i = 0, $es$07$i = 0, $es$07$i$phi = 0, $et$08$i = 0, $exitcond = 0, $exitcond$i132 = 0, $exitcond1491 = 0;
 var $exitcond1492 = 0, $exitcond1493 = 0, $exitcond1494 = 0, $exitcond1495 = 0, $exitcond1496 = 0, $exitcond1500 = 0, $exitcond1501 = 0, $exitcond1502 = 0, $h$0$lcssa$i = 0, $h$0$lcssa$i$i = 0, $h$0$lcssa$i$i260 = 0, $h$0$lcssa$i$i377 = 0, $h$0$lcssa$i$i377$op = 0, $h$0$lcssa$i$i574 = 0, $h$0$lcssa$i$i621 = 0, $h$0$lcssa$i$i669 = 0, $h$0$lcssa$i$i669$op = 0, $h$0$lcssa$i268 = 0, $h$0$lcssa$i295 = 0, $h$0$lcssa$i305 = 0;
 var $h$0$lcssa$i317 = 0, $h$0$lcssa$i385 = 0, $h$0$lcssa$i413 = 0, $h$0$lcssa$i425 = 0, $h$0$lcssa$i437 = 0, $h$0$lcssa$i582 = 0, $h$0$lcssa$i629 = 0, $h$0$lcssa$i677 = 0, $h$017$i = 0, $h$017$i264 = 0, $h$017$i381 = 0, $h$017$i578 = 0, $h$017$i625 = 0, $h$017$i673 = 0, $h$04$i = 0, $h$04$i$i = 0, $h$04$i$i$1 = 0, $h$04$i$i$2 = 0, $h$04$i$i$3 = 0, $h$04$i$i114 = 0;
 var $h$04$i$i114$1 = 0, $h$04$i$i114$2 = 0, $h$04$i$i114$3 = 0, $h$04$i$i234 = 0, $h$04$i$i255 = 0, $h$04$i$i372 = 0, $h$04$i$i569 = 0, $h$04$i$i616 = 0, $h$04$i$i664 = 0, $h$04$i3$i = 0, $h$04$i3$i273 = 0, $h$04$i3$i390 = 0, $h$04$i3$i587 = 0, $h$04$i3$i634 = 0, $h$04$i3$i682 = 0, $h$04$i300 = 0, $h$04$i312 = 0, $h$04$i324 = 0, $h$04$i336 = 0, $h$04$i348 = 0;
 var $h$04$i408 = 0, $h$04$i420 = 0, $h$04$i432 = 0, $h$1$i = 0, $h$1$i282 = 0, $h$1$i399 = 0, $h$1$i596 = 0, $h$1$i643 = 0, $h$1$i691 = 0, $i$0$lcssa$i = 0, $i$0$lcssa$i$i = 0, $i$0$lcssa$i$i261 = 0, $i$0$lcssa$i$i378 = 0, $i$0$lcssa$i$i575 = 0, $i$0$lcssa$i$i622 = 0, $i$0$lcssa$i$i670 = 0, $i$0$lcssa$i269 = 0, $i$0$lcssa$i296 = 0, $i$0$lcssa$i306 = 0, $i$0$lcssa$i318 = 0;
 var $i$0$lcssa$i386 = 0, $i$0$lcssa$i414 = 0, $i$0$lcssa$i426 = 0, $i$0$lcssa$i438 = 0, $i$0$lcssa$i583 = 0, $i$0$lcssa$i630 = 0, $i$0$lcssa$i678 = 0, $i$01$i = 0, $i$016$i = 0, $i$016$i265 = 0, $i$016$i382 = 0, $i$016$i579 = 0, $i$016$i626 = 0, $i$016$i674 = 0, $i$03$i = 0, $i$03$i$i = 0, $i$03$i$i$1 = 0, $i$03$i$i$2 = 0, $i$03$i$i$3 = 0, $i$03$i$i115 = 0;
 var $i$03$i$i115$1 = 0, $i$03$i$i115$2 = 0, $i$03$i$i115$3 = 0, $i$03$i$i235 = 0, $i$03$i$i256 = 0, $i$03$i$i373 = 0, $i$03$i$i570 = 0, $i$03$i$i617 = 0, $i$03$i$i665 = 0, $i$03$i301 = 0, $i$03$i313 = 0, $i$03$i325 = 0, $i$03$i325$lcssa = 0, $i$03$i337 = 0, $i$03$i349 = 0, $i$03$i349$lcssa = 0, $i$03$i4$i = 0, $i$03$i4$i274 = 0, $i$03$i4$i391 = 0, $i$03$i4$i588 = 0;
 var $i$03$i4$i635 = 0, $i$03$i4$i683 = 0, $i$03$i409 = 0, $i$03$i421 = 0, $i$03$i433 = 0, $i$1$i = 0, $i$1$i281 = 0, $i$1$i398 = 0, $i$1$i595 = 0, $i$1$i642 = 0, $i$1$i690 = 0, $i$1$lcssa$i = 0, $i$310$i = 0, $i$310$i$lcssa = 0, $i$37$i = 0, $isdigit = 0, $isdigit$i = 0, $isdigit$i$i = 0, $isdigit$i$i$1 = 0, $isdigit$i$i$2 = 0;
 var $isdigit$i$i$3 = 0, $isdigit$i$i117 = 0, $isdigit$i$i117$1 = 0, $isdigit$i$i117$2 = 0, $isdigit$i$i117$3 = 0, $isdigit$i$i237 = 0, $isdigit$i$i258 = 0, $isdigit$i$i375 = 0, $isdigit$i$i572 = 0, $isdigit$i$i619 = 0, $isdigit$i$i667 = 0, $isdigit$i243 = 0, $isdigit$i271 = 0, $isdigit$i293 = 0, $isdigit$i303 = 0, $isdigit$i315 = 0, $isdigit$i327 = 0, $isdigit$i339 = 0, $isdigit$i351 = 0, $isdigit$i388 = 0;
 var $isdigit$i411 = 0, $isdigit$i423 = 0, $isdigit$i435 = 0, $isdigit$i585 = 0, $isdigit$i6$i = 0, $isdigit$i6$i277 = 0, $isdigit$i6$i394 = 0, $isdigit$i6$i591 = 0, $isdigit$i6$i638 = 0, $isdigit$i6$i686 = 0, $isdigit$i632 = 0, $isdigit$i680 = 0, $isdigit2$i = 0, $isdigit2$i$i = 0, $isdigit2$i$i$1 = 0, $isdigit2$i$i$2 = 0, $isdigit2$i$i$3 = 0, $isdigit2$i$i113 = 0, $isdigit2$i$i113$1 = 0, $isdigit2$i$i113$2 = 0;
 var $isdigit2$i$i113$3 = 0, $isdigit2$i$i233 = 0, $isdigit2$i$i254 = 0, $isdigit2$i$i371 = 0, $isdigit2$i$i568 = 0, $isdigit2$i$i615 = 0, $isdigit2$i$i663 = 0, $isdigit2$i299 = 0, $isdigit2$i407 = 0, $isdigit2$i419 = 0, $isdigit42 = 0, $isdigit48 = 0, $isdigit51 = 0, $isdigit53 = 0, $isdigit57 = 0, $isdigittmp = 0, $isdigittmp$i = 0, $isdigittmp$i$i = 0, $isdigittmp$i$i$1 = 0, $isdigittmp$i$i$2 = 0;
 var $isdigittmp$i$i$3 = 0, $isdigittmp$i$i116 = 0, $isdigittmp$i$i116$1 = 0, $isdigittmp$i$i116$2 = 0, $isdigittmp$i$i116$3 = 0, $isdigittmp$i$i236 = 0, $isdigittmp$i$i257 = 0, $isdigittmp$i$i374 = 0, $isdigittmp$i$i571 = 0, $isdigittmp$i$i618 = 0, $isdigittmp$i$i666 = 0, $isdigittmp$i242 = 0, $isdigittmp$i270 = 0, $isdigittmp$i292 = 0, $isdigittmp$i302 = 0, $isdigittmp$i314 = 0, $isdigittmp$i326 = 0, $isdigittmp$i338 = 0, $isdigittmp$i350 = 0, $isdigittmp$i387 = 0;
 var $isdigittmp$i410 = 0, $isdigittmp$i422 = 0, $isdigittmp$i434 = 0, $isdigittmp$i5$i = 0, $isdigittmp$i5$i276 = 0, $isdigittmp$i5$i393 = 0, $isdigittmp$i5$i590 = 0, $isdigittmp$i5$i637 = 0, $isdigittmp$i5$i685 = 0, $isdigittmp$i584 = 0, $isdigittmp$i631 = 0, $isdigittmp$i679 = 0, $isdigittmp1$i = 0, $isdigittmp1$i$i = 0, $isdigittmp1$i$i$1 = 0, $isdigittmp1$i$i$2 = 0, $isdigittmp1$i$i$3 = 0, $isdigittmp1$i$i112 = 0, $isdigittmp1$i$i112$1 = 0, $isdigittmp1$i$i112$2 = 0;
 var $isdigittmp1$i$i112$3 = 0, $isdigittmp1$i$i232 = 0, $isdigittmp1$i$i253 = 0, $isdigittmp1$i$i370 = 0, $isdigittmp1$i$i567 = 0, $isdigittmp1$i$i614 = 0, $isdigittmp1$i$i662 = 0, $isdigittmp1$i298 = 0, $isdigittmp1$i406 = 0, $isdigittmp1$i418 = 0, $isdigittmp41 = 0, $isdigittmp47 = 0, $isdigittmp50 = 0, $isdigittmp52 = 0, $isdigittmp56 = 0, $ktempo$01330 = 0, $ktempo$1$lcssa = 0, $ktempo$1$ph1279 = 0, $ktempo$11115 = 0, $ktempo$2 = 0;
 var $ktempo$2$lcssa = 0, $lhsc = 0, $memchr = 0, $memchr39 = 0, $memchr60 = 0, $memchr66 = 0, $memchr67 = 0, $memchr68 = 0, $mm = 0, $mmsp$01334 = 0, $mmsp$1$lcssa = 0, $mmsp$1$ph1283 = 0, $mmsp$11123 = 0, $mmsp$2 = 0, $mmsp$3 = 0, $mmsp$3$lcssa = 0, $mmstack = 0, $mnotediv = 0, $mnotelen = 0, $mp$0 = 0;
 var $mp$01181 = 0, $mp$01182 = 0, $mp$01182$lcssa = 0, $n$0$lcssa$i = 0, $n$02$i = 0, $n$1$$i = 0, $n$14$i = 0, $n$2$i = 0, $nd0$0 = 0, $nd0$0$lcssa = 0, $nd0$1 = 0, $nl0$0 = 0, $nl0$0$lcssa = 0, $nl0$1 = 0, $nl0$21204 = 0, $nn$i = 0, $not$ = 0, $not$$i = 0, $not$2$i = 0, $not$812 = 0;
 var $not$824 = 0, $not$828 = 0, $not$829 = 0, $not$831 = 0, $not$832 = 0, $or$cond = 0, $or$cond12 = 0, $or$cond1743 = 0, $or$cond23 = 0, $or$cond3 = 0, $or$cond5 = 0, $or$cond7 = 0, $or$cond787 = 0, $or$cond792 = 0, $or$cond793 = 0, $or$cond794 = 0, $or$cond798 = 0, $or$cond88 = 0, $or$cond90 = 0, $or$cond901188 = 0;
 var $orderlen$01313 = 0, $orderlen$1$lcssa = 0, $orderlen$1$lcssa$lcssa = 0, $orderlen$1$ph1262 = 0, $orderlen$11090 = 0, $orderlen$2 = 0, $orderlen$3 = 0, $orderlen$3$lcssa = 0, $orderlen$4783 = 0, $orderlist = 0, $p$0 = 0, $p$0$lcssa = 0, $p$1 = 0, $p$1$lcssa = 0, $p$10 = 0, $p$11 = 0, $p$12 = 0, $p$13 = 0, $p$14 = 0, $p$14$lcssa = 0;
 var $p$15 = 0, $p$16 = 0, $p$16$lcssa = 0, $p$16$pn = 0, $p$17$be = 0, $p$18 = 0, $p$18$pn = 0, $p$19 = 0, $p$2 = 0, $p$2$lcssa = 0, $p$20 = 0, $p$20$lcssa = 0, $p$21 = 0, $p$22 = 0, $p$22$lcssa = 0, $p$23 = 0, $p$23$lcssa = 0, $p$24 = 0, $p$25 = 0, $p$25$lcssa = 0;
 var $p$26 = 0, $p$26$lcssa = 0, $p$27 = 0, $p$27$lcssa = 0, $p$27$pn = 0, $p$28 = 0, $p$29 = 0, $p$3 = 0, $p$3$lcssa = 0, $p$30 = 0, $p$30$ph = 0, $p$31 = 0, $p$32 = 0, $p$33 = 0, $p$33$lcssa = 0, $p$34 = 0, $p$35 = 0, $p$36 = 0, $p$36$lcssa = 0, $p$37$be = 0;
 var $p$371215 = 0, $p$38 = 0, $p$39 = 0, $p$4 = 0, $p$40 = 0, $p$41$be = 0, $p$411200 = 0, $p$421189 = 0, $p$43 = 0, $p$44 = 0, $p$45 = 0, $p$46 = 0, $p$47 = 0, $p$47$lcssa = 0, $p$48 = 0, $p$48$lcssa = 0, $p$50 = 0, $p$51550 = 0, $p$6 = 0, $p$7 = 0;
 var $p$8$lcssa = 0, $p$81079 = 0, $p$9$pn = 0, $p$91085 = 0, $p$91085$lcssa = 0, $p2$026$i = 0, $p2$1$lcssa$i = 0, $p2$117$i = 0, $p2$2$pn$i = 0, $p2$222$i = 0, $p2$222$i$lcssa = 0, $p2$3$i = 0, $p2$4$i = 0, $partpat = 0, $phitmp$i = 0, $phitmp$i147 = 0, $phitmp$i192 = 0, $phitmp$i221 = 0, $phitmp$i712 = 0, $phitmp1$i = 0;
 var $phitmp1$i148 = 0, $phitmp1$i193 = 0, $phitmp1$i222 = 0, $phitmp1$i713 = 0, $pm$025$i = 0, $pm$2$lcssa$i = 0, $pm$2$lcssa$i$lcssa = 0, $pm$214$i = 0, $pp$0$be = 0, $pp$01219 = 0, $pp$1 = 0, $pp$1$p$50 = 0, $ptp$07$i$i = 0, $s$i = 0, $snotediv = 0, $snotelen = 0, $storemerge$i = 0, $storemerge$i128 = 0, $strlen = 0, $strlen73 = 0;
 var $strlenfirst$i = 0, $strlenfirst$i127 = 0, $strlenfirst$pre = 0, $strlenfirst38 = 0, $strlenfirst72 = 0, $strlenfirst75 = 0, $strlenfirst77 = 0, $switch = 0, $switch100 = 0, $switch1001199 = 0, $switch1214 = 0, $t$0 = 0, $t$0$i = 0, $t$0$i746 = 0, $t$01$i = 0, $t$01068 = 0, $t$01069 = 0, $t$03$i = 0, $t$08$i = 0, $t$09$i = 0;
 var $t$1 = 0, $t$1$lcssa = 0, $t$10$ph = 0, $t$11 = 0, $t$121185 = 0, $t$121185$lcssa2190 = 0, $t$1211852192 = 0, $t$141056 = 0, $t$151055 = 0, $t$161054 = 0, $t$171053 = 0, $t$181049 = 0, $t$191045 = 0, $t$2 = 0, $t$201042 = 0, $t$211041 = 0, $t$221040 = 0, $t$41063 = 0, $t$5 = 0, $t$6 = 0;
 var $t$61064 = 0, $t$61065 = 0, $t$7 = 0, $t$7$lcssa = 0, $t$8 = 0, $t$91089 = 0, $thistime$2 = 0, $thistime$4 = 0, $thistime$6 = 0, $thistime$8 = 0, $tp$0$i = 0, $tp$0$i$i = 0, $tp$0$i249 = 0, $tp$0$i363 = 0, $tp$0$i459 = 0, $tp$0$i473 = 0, $tp$0$i488 = 0, $tp$0$i506 = 0, $tp$0$i524 = 0, $tp$0$i560 = 0;
 var $tp$0$i6$i = 0, $tp$0$i612 = 0, $tp$0$i659 = 0, $tp$0$i729 = 0, $tp$0$us$i = 0, $tp$0$us$i497 = 0, $tp$0$us$i515 = 0, $tp$0$us$i533 = 0, $tp$0$us15$i = 0, $tp$01$i = 0, $tp$01$i$i = 0, $tp$01$i245 = 0, $tp$01$i360 = 0, $tp$01$i557 = 0, $tp$01$i609 = 0, $tp$01$i656 = 0, $tp$015$i = 0, $tp$016$i = 0, $tp$02$i = 0, $tp$02$i247 = 0;
 var $tp$02$i361 = 0, $tp$02$i558 = 0, $tp$02$i610 = 0, $tp$02$i657 = 0, $tp$08$i = 0, $tp$08$i$i = 0, $tp$08$i1$i = 0, $tp$08$i1$pre$i = 0, $tp$08$i471 = 0, $tp$08$i486 = 0, $tp$08$i504 = 0, $tp$08$i522 = 0, $tp$08$us$i = 0, $tp$08$us$i495 = 0, $tp$08$us$i513 = 0, $tp$08$us$i531 = 0, $tp$08$us11$i = 0, $tp$09$i$i = 0, $tp$1$i$i = 0, $tpd$01333 = 0;
 var $tpd$1$be = 0, $tpd$1$lcssa = 0, $tpd$1$ph1282 = 0, $tpd$11121 = 0, $tpd$2 = 0, $tpd$3 = 0, $tpd$4 = 0, $tpd$4$lcssa = 0, $tracktime$0$i = 0, $tracktime$0$i184 = 0, $tracktime$0$i203 = 0, $tt$02$i$i = 0, $tt$15$i$i = 0, $tt$2$i$i = 0, $ttp$0 = 0, $ttp$0$i = 0, $ttp$0$i704 = 0, $ttp$0$i736 = 0, $ttp$01$i733$pre = 0, $ttp$01072 = 0;
 var $ttp$01073 = 0, $ttp$02$i = 0, $ttp$02$i702 = 0, $ttp$02$i734 = 0, $ttp$1 = 0, $ttp$11061 = 0, $ttp$2 = 0, $ttp$21057 = 0, $ttp$21058 = 0, $txt$i = 0, $value$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer101 = 0, $vararg_buffer104 = 0, $vararg_buffer107 = 0, $vararg_buffer11 = 0, $vararg_buffer111 = 0, $vararg_buffer114 = 0, $vararg_buffer117 = 0;
 var $vararg_buffer121 = 0, $vararg_buffer124 = 0, $vararg_buffer127 = 0, $vararg_buffer130 = 0, $vararg_buffer133 = 0, $vararg_buffer136 = 0, $vararg_buffer139 = 0, $vararg_buffer142 = 0, $vararg_buffer144 = 0, $vararg_buffer147 = 0, $vararg_buffer15 = 0, $vararg_buffer150 = 0, $vararg_buffer153 = 0, $vararg_buffer156 = 0, $vararg_buffer162 = 0, $vararg_buffer172 = 0, $vararg_buffer177 = 0, $vararg_buffer18 = 0, $vararg_buffer180 = 0, $vararg_buffer185 = 0;
 var $vararg_buffer195 = 0, $vararg_buffer21 = 0, $vararg_buffer25 = 0, $vararg_buffer29 = 0, $vararg_buffer33 = 0, $vararg_buffer36 = 0, $vararg_buffer39 = 0, $vararg_buffer4 = 0, $vararg_buffer42 = 0, $vararg_buffer45 = 0, $vararg_buffer48 = 0, $vararg_buffer51 = 0, $vararg_buffer54 = 0, $vararg_buffer57 = 0, $vararg_buffer60 = 0, $vararg_buffer63 = 0, $vararg_buffer66 = 0, $vararg_buffer69 = 0, $vararg_buffer72 = 0, $vararg_buffer75 = 0;
 var $vararg_buffer78 = 0, $vararg_buffer8 = 0, $vararg_buffer81 = 0, $vararg_buffer84 = 0, $vararg_buffer88 = 0, $vararg_buffer92 = 0, $vararg_buffer95 = 0, $vararg_buffer98 = 0, $vararg_ptr110 = 0, $vararg_ptr120 = 0, $vararg_ptr14 = 0, $vararg_ptr159 = 0, $vararg_ptr160 = 0, $vararg_ptr161 = 0, $vararg_ptr165 = 0, $vararg_ptr166 = 0, $vararg_ptr167 = 0, $vararg_ptr168 = 0, $vararg_ptr169 = 0, $vararg_ptr170 = 0;
 var $vararg_ptr171 = 0, $vararg_ptr175 = 0, $vararg_ptr176 = 0, $vararg_ptr183 = 0, $vararg_ptr184 = 0, $vararg_ptr188 = 0, $vararg_ptr189 = 0, $vararg_ptr190 = 0, $vararg_ptr191 = 0, $vararg_ptr192 = 0, $vararg_ptr193 = 0, $vararg_ptr194 = 0, $vararg_ptr198 = 0, $vararg_ptr199 = 0, $vararg_ptr200 = 0, $vararg_ptr24 = 0, $vararg_ptr28 = 0, $vararg_ptr32 = 0, $vararg_ptr7 = 0, $vararg_ptr87 = 0;
 var $vararg_ptr91 = 0, $vnl$0 = 0, $vnl$0$lcssa = 0, $vnl$1 = 0, $vol$0$i = 0, $vol$0$i368 = 0, $vol$0$i565 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1728|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer195 = sp + 520|0;
 $vararg_buffer185 = sp + 488|0;
 $vararg_buffer180 = sp + 472|0;
 $vararg_buffer177 = sp + 464|0;
 $vararg_buffer172 = sp + 448|0;
 $vararg_buffer162 = sp + 416|0;
 $vararg_buffer156 = sp + 400|0;
 $vararg_buffer153 = sp + 392|0;
 $vararg_buffer150 = sp + 384|0;
 $vararg_buffer147 = sp + 376|0;
 $vararg_buffer144 = sp + 368|0;
 $vararg_buffer142 = sp + 360|0;
 $vararg_buffer139 = sp + 352|0;
 $vararg_buffer136 = sp + 344|0;
 $vararg_buffer133 = sp + 336|0;
 $vararg_buffer130 = sp + 328|0;
 $vararg_buffer127 = sp + 320|0;
 $vararg_buffer124 = sp + 312|0;
 $vararg_buffer121 = sp + 304|0;
 $vararg_buffer117 = sp + 296|0;
 $vararg_buffer114 = sp + 288|0;
 $vararg_buffer111 = sp + 280|0;
 $vararg_buffer107 = sp + 272|0;
 $vararg_buffer104 = sp + 264|0;
 $vararg_buffer101 = sp + 256|0;
 $vararg_buffer98 = sp + 248|0;
 $vararg_buffer95 = sp + 240|0;
 $vararg_buffer92 = sp + 232|0;
 $vararg_buffer88 = sp + 224|0;
 $vararg_buffer84 = sp + 216|0;
 $vararg_buffer81 = sp + 208|0;
 $vararg_buffer78 = sp + 200|0;
 $vararg_buffer75 = sp + 192|0;
 $vararg_buffer72 = sp + 184|0;
 $vararg_buffer69 = sp + 176|0;
 $vararg_buffer66 = sp + 168|0;
 $vararg_buffer63 = sp + 160|0;
 $vararg_buffer60 = sp + 152|0;
 $vararg_buffer57 = sp + 144|0;
 $vararg_buffer54 = sp + 136|0;
 $vararg_buffer51 = sp + 128|0;
 $vararg_buffer48 = sp + 120|0;
 $vararg_buffer45 = sp + 112|0;
 $vararg_buffer42 = sp + 104|0;
 $vararg_buffer39 = sp + 96|0;
 $vararg_buffer36 = sp + 88|0;
 $vararg_buffer33 = sp + 80|0;
 $vararg_buffer29 = sp + 72|0;
 $vararg_buffer25 = sp + 64|0;
 $vararg_buffer21 = sp + 56|0;
 $vararg_buffer18 = sp + 48|0;
 $vararg_buffer15 = sp + 40|0;
 $vararg_buffer11 = sp + 32|0;
 $vararg_buffer8 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $nn$i = sp + 1720|0;
 $s$i = sp + 1712|0;
 $value$i = sp + 1456|0;
 $txt$i = sp + 1200|0;
 $buf$i = sp + 1188|0;
 $mm = sp + 656|0;
 $barsig = sp + 1136|0;
 $abcparts = sp + 652|0;
 $partpat = sp + 1080|0;
 $orderlist = sp + 648|0;
 $snotelen = sp + 644|0;
 $snotediv = sp + 640|0;
 $mnotelen = sp + 636|0;
 $mnotediv = sp + 632|0;
 $mmstack = sp + 600|0;
 $buf = sp + 880|0;
 $buf1 = sp + 680|0;
 $cnl = sp + 568|0;
 $cnd = sp + 536|0;
 $buf2 = sp + 668|0;
 $0 = (__ZN10CSoundFile7TestABCEPKhj(0,$lpStream,$dwMemLength)|0); //@line 3811 "load_abc.cpp"
 if (!($0)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 5158 "load_abc.cpp"
 }
 $1 = (_calloc(1,428)|0); //@line 2383 "load_abc.cpp"
 $2 = ($1|0)==(0|0); //@line 2384 "load_abc.cpp"
 if ($2) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 5158 "load_abc.cpp"
 }
 $3 = ((($1)) + 28|0); //@line 2389 "load_abc.cpp"
 HEAP8[$3>>0]=0&255;HEAP8[$3+1>>0]=(0>>8)&255;HEAP8[$3+2>>0]=(0>>16)&255;HEAP8[$3+3>>0]=0>>24; //@line 2389 "load_abc.cpp"
 $4 = ((($1)) + 12|0); //@line 2390 "load_abc.cpp"
 $5 = ((($1)) + 24|0); //@line 2392 "load_abc.cpp"
 HEAP8[$5>>0]=0&255;HEAP8[$5+1>>0]=(0>>8)&255;HEAP8[$5+2>>0]=(0>>16)&255;HEAP8[$5+3>>0]=0>>24; //@line 2392 "load_abc.cpp"
 $6 = ((($1)) + 36|0); //@line 2393 "load_abc.cpp"
 HEAP8[$6>>0] = 0; //@line 2393 "load_abc.cpp"
 $7 = ((($1)) + 356|0); //@line 2394 "load_abc.cpp"
 HEAP8[$7>>0]=0&255;HEAP8[$7+1>>0]=(0>>8)&255;HEAP8[$7+2>>0]=(0>>16)&255;HEAP8[$7+3>>0]=0>>24; //@line 2394 "load_abc.cpp"
 dest=$1; stop=dest+20|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 2387 "load_abc.cpp"
 $8 = (_getenv((524275|0))|0); //@line 2395 "load_abc.cpp"
 $9 = ($8|0)==(0|0); //@line 2396 "load_abc.cpp"
 if ($9) {
  $18 = (_time((0|0))|0); //@line 2411 "load_abc.cpp"
  _srandom($18); //@line 2411 "load_abc.cpp"
  $19 = (_random()|0); //@line 2412 "load_abc.cpp"
  $20 = (+($19|0)); //@line 2412 "load_abc.cpp"
  $21 = $20 * 4.6566128730773926E-6; //@line 2412 "load_abc.cpp"
  $22 = (~~(($21))); //@line 2412 "load_abc.cpp"
  $23 = (($22) + 1)|0; //@line 2412 "load_abc.cpp"
  HEAP8[$4>>0]=$23&255;HEAP8[$4+1>>0]=($23>>8)&255;HEAP8[$4+2>>0]=($23>>16)&255;HEAP8[$4+3>>0]=$23>>24; //@line 2412 "load_abc.cpp"
  HEAP32[$vararg_buffer1>>2] = $23; //@line 2417 "load_abc.cpp"
  (_sprintf($buf$i,524296,$vararg_buffer1)|0); //@line 2417 "load_abc.cpp"
  (_setenv((524275|0),($buf$i|0),1)|0); //@line 2419 "load_abc.cpp"
 } else {
  $10 = HEAP8[$8>>0]|0; //@line 2397 "load_abc.cpp"
  $11 = $10 << 24 >> 24; //@line 2397 "load_abc.cpp"
  $isdigittmp$i = (($11) + -48)|0; //@line 2397 "load_abc.cpp"
  $isdigit$i = ($isdigittmp$i>>>0)<(10); //@line 2397 "load_abc.cpp"
  if ($isdigit$i) {
   $12 = (_atoi($8)|0); //@line 2398 "load_abc.cpp"
   HEAP8[$4>>0]=$12&255;HEAP8[$4+1>>0]=($12>>8)&255;HEAP8[$4+2>>0]=($12>>16)&255;HEAP8[$4+3>>0]=$12>>24; //@line 2398 "load_abc.cpp"
  }
  $13 = ($10<<24>>24)==(45); //@line 2399 "load_abc.cpp"
  if ($13) {
   $14 = ((($8)) + 1|0); //@line 2404 "load_abc.cpp"
   $15 = (_atoi($14)|0); //@line 2404 "load_abc.cpp"
   $16 = (($15) + -1)|0; //@line 2404 "load_abc.cpp"
   HEAP8[$4>>0]=$16&255;HEAP8[$4+1>>0]=($16>>8)&255;HEAP8[$4+2>>0]=($16>>16)&255;HEAP8[$4+3>>0]=$16>>24; //@line 2404 "load_abc.cpp"
   $17 = (($15) + 1)|0; //@line 2405 "load_abc.cpp"
   HEAP32[$vararg_buffer>>2] = $17; //@line 2405 "load_abc.cpp"
   (_sprintf($buf$i,524296,$vararg_buffer)|0); //@line 2405 "load_abc.cpp"
   (_setenv((524275|0),($buf$i|0),1)|0); //@line 2407 "load_abc.cpp"
  }
 }
 HEAP32[$mm>>2] = $lpStream; //@line 3815 "load_abc.cpp"
 $24 = ((($mm)) + 4|0); //@line 3816 "load_abc.cpp"
 HEAP32[$24>>2] = $dwMemLength; //@line 3816 "load_abc.cpp"
 $25 = ((($mm)) + 8|0); //@line 3817 "load_abc.cpp"
 HEAP32[$25>>2] = 0; //@line 3817 "load_abc.cpp"
 $26 = HEAP32[299859]|0; //@line 3819 "load_abc.cpp"
 $27 = ($26|0)==(0); //@line 3819 "load_abc.cpp"
 if (!($27)) {
  while(1) {
   (_sleep(1)|0); //@line 3819 "load_abc.cpp"
   $28 = HEAP32[299859]|0; //@line 3819 "load_abc.cpp"
   $29 = ($28|0)==(0); //@line 3819 "load_abc.cpp"
   if ($29) {
    break;
   }
  }
 }
 HEAP32[299859] = 1; //@line 3820 "load_abc.cpp"
 _pat_resetsmp(); //@line 3821 "load_abc.cpp"
 _pat_init_patnames(); //@line 3822 "load_abc.cpp"
 $30 = ((($this)) + 59652|0); //@line 3823 "load_abc.cpp"
 HEAP32[$30>>2] = 0; //@line 3823 "load_abc.cpp"
 HEAP32[299860] = 0; //@line 3824 "load_abc.cpp"
 $31 = ((($1)) + 424|0); //@line 3826 "load_abc.cpp"
 HEAP32[299861] = 0; //@line 3827 "load_abc.cpp"
 $32 = ((($1)) + 20|0); //@line 3828 "load_abc.cpp"
 HEAP8[$32>>0]=6&255;HEAP8[$32+1>>0]=(6>>8)&255;HEAP8[$32+2>>0]=(6>>16)&255;HEAP8[$32+3>>0]=6>>24; //@line 3828 "load_abc.cpp"
 HEAP32[299862] = 2; //@line 3830 "load_abc.cpp"
 HEAP32[299863] = 1; //@line 3831 "load_abc.cpp"
 $33 = ((($1)) + 360|0); //@line 3837 "load_abc.cpp"
 HEAP8[$33>>0]=120&255;HEAP8[$33+1>>0]=(120>>8)&255;HEAP8[$33+2>>0]=(120>>16)&255;HEAP8[$33+3>>0]=120>>24; //@line 3837 "load_abc.cpp"
 $34 = ((($1)) + 368|0); //@line 3838 "load_abc.cpp"
 HEAP8[$34>>0]=120&255;HEAP8[$34+1>>0]=(120>>8)&255;HEAP8[$34+2>>0]=(120>>16)&255;HEAP8[$34+3>>0]=120>>24; //@line 3838 "load_abc.cpp"
 $35 = ((($1)) + 364|0); //@line 3839 "load_abc.cpp"
 HEAP8[$35>>0]=25&255;HEAP8[$35+1>>0]=(25>>8)&255;HEAP8[$35+2>>0]=(25>>16)&255;HEAP8[$35+3>>0]=25>>24; //@line 3839 "load_abc.cpp"
 $36 = ((($1)) + 372|0); //@line 3840 "load_abc.cpp"
 HEAP8[$36>>0]=33&255;HEAP8[$36+1>>0]=(33>>8)&255;HEAP8[$36+2>>0]=(33>>16)&255;HEAP8[$36+3>>0]=33>>24; //@line 3840 "load_abc.cpp"
 HEAP32[$abcparts>>2] = 0; //@line 3841 "load_abc.cpp"
 HEAP32[$snotelen>>2] = 0; //@line 3847 "load_abc.cpp"
 HEAP32[$snotediv>>2] = 0; //@line 3848 "load_abc.cpp"
 HEAP32[$mnotelen>>2] = 1; //@line 3849 "load_abc.cpp"
 HEAP32[$mnotediv>>2] = 1; //@line 3850 "load_abc.cpp"
 $37 = ((($1)) + 376|0); //@line 3856 "load_abc.cpp"
 HEAP8[$37>>0]=0&255;HEAP8[$37+1>>0]=(0>>8)&255;HEAP8[$37+2>>0]=(0>>16)&255;HEAP8[$37+3>>0]=0>>24; //@line 3856 "load_abc.cpp"
 $38 = ((($1)) + 380|0); //@line 3857 "load_abc.cpp"
 HEAP8[$38>>0]=0&255;HEAP8[$38+1>>0]=(0>>8)&255;HEAP8[$38+2>>0]=(0>>16)&255;HEAP8[$38+3>>0]=0>>24; //@line 3857 "load_abc.cpp"
 $39 = ((($1)) + 384|0); //@line 3858 "load_abc.cpp"
 HEAP8[$39>>0]=1&255;HEAP8[$39+1>>0]=(1>>8)&255;HEAP8[$39+2>>0]=(1>>16)&255;HEAP8[$39+3>>0]=1>>24; //@line 3858 "load_abc.cpp"
 $40 = ((($1)) + 388|0); //@line 3859 "load_abc.cpp"
 HEAP8[$40>>0]=0&255;HEAP8[$40+1>>0]=(0>>8)&255;HEAP8[$40+2>>0]=(0>>16)&255;HEAP8[$40+3>>0]=0>>24; //@line 3859 "load_abc.cpp"
 $41 = ((($1)) + 412|0); //@line 3862 "load_abc.cpp"
 $42 = ((($1)) + 416|0); //@line 3863 "load_abc.cpp"
 $43 = ((($1)) + 420|0); //@line 3864 "load_abc.cpp"
 $44 = ((($1)) + 392|0); //@line 3866 "load_abc.cpp"
 dest=$41; stop=dest+16|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 3862 "load_abc.cpp"
 HEAP8[$44>>0]=71&255;HEAP8[$44+1>>0]=(71>>8)&255;HEAP8[$44+2>>0]=(71>>16)&255;HEAP8[$44+3>>0]=71>>24; //@line 3866 "load_abc.cpp"
 $45 = ((($1)) + 396|0); //@line 3867 "load_abc.cpp"
 HEAP8[$45>>0]=45&255;HEAP8[$45+1>>0]=(45>>8)&255;HEAP8[$45+2>>0]=(45>>16)&255;HEAP8[$45+3>>0]=45>>24; //@line 3867 "load_abc.cpp"
 $46 = ((($1)) + 400|0); //@line 3868 "load_abc.cpp"
 HEAP8[$46>>0]=33&255;HEAP8[$46+1>>0]=(33>>8)&255;HEAP8[$46+2>>0]=(33>>16)&255;HEAP8[$46+3>>0]=33>>24; //@line 3868 "load_abc.cpp"
 $47 = ((($1)) + 404|0); //@line 3869 "load_abc.cpp"
 HEAP8[$47>>0]=80&255;HEAP8[$47+1>>0]=(80>>8)&255;HEAP8[$47+2>>0]=(80>>16)&255;HEAP8[$47+3>>0]=80>>24; //@line 3869 "load_abc.cpp"
 $48 = ((($1)) + 408|0); //@line 3870 "load_abc.cpp"
 HEAP8[$48>>0]=80&255;HEAP8[$48+1>>0]=(80>>8)&255;HEAP8[$48+2>>0]=(80>>16)&255;HEAP8[$48+3>>0]=80>>24; //@line 3870 "load_abc.cpp"
 __ZL14abc_new_umacroP10_ABCHANDLEPKc($1,524531); //@line 3871 "load_abc.cpp"
 __ZL14abc_new_umacroP10_ABCHANDLEPKc($1,524571); //@line 3872 "load_abc.cpp"
 __ZL14abc_new_umacroP10_ABCHANDLEPKc($1,524583); //@line 3873 "load_abc.cpp"
 __ZL14abc_new_umacroP10_ABCHANDLEPKc($1,524594); //@line 3874 "load_abc.cpp"
 __ZL14abc_new_umacroP10_ABCHANDLEPKc($1,524606); //@line 3875 "load_abc.cpp"
 __ZL14abc_new_umacroP10_ABCHANDLEPKc($1,524625); //@line 3876 "load_abc.cpp"
 __ZL14abc_new_umacroP10_ABCHANDLEPKc($1,524644); //@line 3877 "load_abc.cpp"
 __ZL14abc_new_umacroP10_ABCHANDLEPKc($1,524659); //@line 3878 "load_abc.cpp"
 __ZL14abc_new_umacroP10_ABCHANDLEPKc($1,524673); //@line 3879 "load_abc.cpp"
 __ZL14abc_new_umacroP10_ABCHANDLEPKc($1,524685); //@line 3880 "load_abc.cpp"
 HEAP32[299864] = 0; //@line 2994 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(1282466,3,46612); //@line 2995 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524863,3,46624); //@line 2996 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524865,4,46636); //@line 2997 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524867,4,46652); //@line 2998 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524870,4,46668); //@line 2999 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524875,4,46684); //@line 3000 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524878,4,46700); //@line 3001 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524880,4,46716); //@line 3002 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524883,3,46732); //@line 3003 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524887,3,46744); //@line 3004 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524889,4,46756); //@line 3005 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524894,4,46756); //@line 3006 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524897,3,46772); //@line 3007 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524901,4,46784); //@line 3008 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524906,5,46800); //@line 3009 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524908,5,46820); //@line 3010 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524911,5,46840); //@line 3011 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524916,5,46860); //@line 3012 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524919,6,46880); //@line 3013 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524922,5,46904); //@line 3014 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524927,3,46924); //@line 3015 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524931,3,46936); //@line 3016 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524936,4,46948); //@line 3017 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524941,4,46964); //@line 3018 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524947,4,46980); //@line 3019 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524953,5,46996); //@line 3020 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524959,2,47016); //@line 3021 "load_abc.cpp"
 __ZL16abc_addchordnamePKciPKi(524961,6,47024); //@line 3022 "load_abc.cpp"
 HEAP8[$partpat>>0] = -1; //@line 3583 "load_abc.cpp"
 $49 = ((($partpat)) + 1|0); //@line 3584 "load_abc.cpp"
 HEAP8[$49>>0] = 0; //@line 3584 "load_abc.cpp"
 $50 = ((($partpat)) + 2|0); //@line 3583 "load_abc.cpp"
 HEAP8[$50>>0] = -1; //@line 3583 "load_abc.cpp"
 $51 = ((($partpat)) + 3|0); //@line 3584 "load_abc.cpp"
 HEAP8[$51>>0] = 0; //@line 3584 "load_abc.cpp"
 $52 = ((($partpat)) + 4|0); //@line 3583 "load_abc.cpp"
 HEAP8[$52>>0] = -1; //@line 3583 "load_abc.cpp"
 $53 = ((($partpat)) + 5|0); //@line 3584 "load_abc.cpp"
 HEAP8[$53>>0] = 0; //@line 3584 "load_abc.cpp"
 $54 = ((($partpat)) + 6|0); //@line 3583 "load_abc.cpp"
 HEAP8[$54>>0] = -1; //@line 3583 "load_abc.cpp"
 $55 = ((($partpat)) + 7|0); //@line 3584 "load_abc.cpp"
 HEAP8[$55>>0] = 0; //@line 3584 "load_abc.cpp"
 $56 = ((($partpat)) + 8|0); //@line 3583 "load_abc.cpp"
 HEAP8[$56>>0] = -1; //@line 3583 "load_abc.cpp"
 $57 = ((($partpat)) + 9|0); //@line 3584 "load_abc.cpp"
 HEAP8[$57>>0] = 0; //@line 3584 "load_abc.cpp"
 $58 = ((($partpat)) + 10|0); //@line 3583 "load_abc.cpp"
 HEAP8[$58>>0] = -1; //@line 3583 "load_abc.cpp"
 $59 = ((($partpat)) + 11|0); //@line 3584 "load_abc.cpp"
 HEAP8[$59>>0] = 0; //@line 3584 "load_abc.cpp"
 $60 = ((($partpat)) + 12|0); //@line 3583 "load_abc.cpp"
 HEAP8[$60>>0] = -1; //@line 3583 "load_abc.cpp"
 $61 = ((($partpat)) + 13|0); //@line 3584 "load_abc.cpp"
 HEAP8[$61>>0] = 0; //@line 3584 "load_abc.cpp"
 $62 = ((($partpat)) + 14|0); //@line 3583 "load_abc.cpp"
 HEAP8[$62>>0] = -1; //@line 3583 "load_abc.cpp"
 $63 = ((($partpat)) + 15|0); //@line 3584 "load_abc.cpp"
 HEAP8[$63>>0] = 0; //@line 3584 "load_abc.cpp"
 $64 = ((($partpat)) + 16|0); //@line 3583 "load_abc.cpp"
 HEAP8[$64>>0] = -1; //@line 3583 "load_abc.cpp"
 $65 = ((($partpat)) + 17|0); //@line 3584 "load_abc.cpp"
 HEAP8[$65>>0] = 0; //@line 3584 "load_abc.cpp"
 $66 = ((($partpat)) + 18|0); //@line 3583 "load_abc.cpp"
 HEAP8[$66>>0] = -1; //@line 3583 "load_abc.cpp"
 $67 = ((($partpat)) + 19|0); //@line 3584 "load_abc.cpp"
 HEAP8[$67>>0] = 0; //@line 3584 "load_abc.cpp"
 $68 = ((($partpat)) + 20|0); //@line 3583 "load_abc.cpp"
 HEAP8[$68>>0] = -1; //@line 3583 "load_abc.cpp"
 $69 = ((($partpat)) + 21|0); //@line 3584 "load_abc.cpp"
 HEAP8[$69>>0] = 0; //@line 3584 "load_abc.cpp"
 $70 = ((($partpat)) + 22|0); //@line 3583 "load_abc.cpp"
 HEAP8[$70>>0] = -1; //@line 3583 "load_abc.cpp"
 $71 = ((($partpat)) + 23|0); //@line 3584 "load_abc.cpp"
 HEAP8[$71>>0] = 0; //@line 3584 "load_abc.cpp"
 $72 = ((($partpat)) + 24|0); //@line 3583 "load_abc.cpp"
 HEAP8[$72>>0] = -1; //@line 3583 "load_abc.cpp"
 $73 = ((($partpat)) + 25|0); //@line 3584 "load_abc.cpp"
 HEAP8[$73>>0] = 0; //@line 3584 "load_abc.cpp"
 $74 = ((($partpat)) + 26|0); //@line 3583 "load_abc.cpp"
 HEAP8[$74>>0] = -1; //@line 3583 "load_abc.cpp"
 $75 = ((($partpat)) + 27|0); //@line 3584 "load_abc.cpp"
 HEAP8[$75>>0] = 0; //@line 3584 "load_abc.cpp"
 $76 = ((($partpat)) + 28|0); //@line 3583 "load_abc.cpp"
 HEAP8[$76>>0] = -1; //@line 3583 "load_abc.cpp"
 $77 = ((($partpat)) + 29|0); //@line 3584 "load_abc.cpp"
 HEAP8[$77>>0] = 0; //@line 3584 "load_abc.cpp"
 $78 = ((($partpat)) + 30|0); //@line 3583 "load_abc.cpp"
 HEAP8[$78>>0] = -1; //@line 3583 "load_abc.cpp"
 $79 = ((($partpat)) + 31|0); //@line 3584 "load_abc.cpp"
 HEAP8[$79>>0] = 0; //@line 3584 "load_abc.cpp"
 $80 = ((($partpat)) + 32|0); //@line 3583 "load_abc.cpp"
 HEAP8[$80>>0] = -1; //@line 3583 "load_abc.cpp"
 $81 = ((($partpat)) + 33|0); //@line 3584 "load_abc.cpp"
 HEAP8[$81>>0] = 0; //@line 3584 "load_abc.cpp"
 $82 = ((($partpat)) + 34|0); //@line 3583 "load_abc.cpp"
 HEAP8[$82>>0] = -1; //@line 3583 "load_abc.cpp"
 $83 = ((($partpat)) + 35|0); //@line 3584 "load_abc.cpp"
 HEAP8[$83>>0] = 0; //@line 3584 "load_abc.cpp"
 $84 = ((($partpat)) + 36|0); //@line 3583 "load_abc.cpp"
 HEAP8[$84>>0] = -1; //@line 3583 "load_abc.cpp"
 $85 = ((($partpat)) + 37|0); //@line 3584 "load_abc.cpp"
 HEAP8[$85>>0] = 0; //@line 3584 "load_abc.cpp"
 $86 = ((($partpat)) + 38|0); //@line 3583 "load_abc.cpp"
 HEAP8[$86>>0] = -1; //@line 3583 "load_abc.cpp"
 $87 = ((($partpat)) + 39|0); //@line 3584 "load_abc.cpp"
 HEAP8[$87>>0] = 0; //@line 3584 "load_abc.cpp"
 $88 = ((($partpat)) + 40|0); //@line 3583 "load_abc.cpp"
 HEAP8[$88>>0] = -1; //@line 3583 "load_abc.cpp"
 $89 = ((($partpat)) + 41|0); //@line 3584 "load_abc.cpp"
 HEAP8[$89>>0] = 0; //@line 3584 "load_abc.cpp"
 $90 = ((($partpat)) + 42|0); //@line 3583 "load_abc.cpp"
 HEAP8[$90>>0] = -1; //@line 3583 "load_abc.cpp"
 $91 = ((($partpat)) + 43|0); //@line 3584 "load_abc.cpp"
 HEAP8[$91>>0] = 0; //@line 3584 "load_abc.cpp"
 $92 = ((($partpat)) + 44|0); //@line 3583 "load_abc.cpp"
 HEAP8[$92>>0] = -1; //@line 3583 "load_abc.cpp"
 $93 = ((($partpat)) + 45|0); //@line 3584 "load_abc.cpp"
 HEAP8[$93>>0] = 0; //@line 3584 "load_abc.cpp"
 $94 = ((($partpat)) + 46|0); //@line 3583 "load_abc.cpp"
 HEAP8[$94>>0] = -1; //@line 3583 "load_abc.cpp"
 $95 = ((($partpat)) + 47|0); //@line 3584 "load_abc.cpp"
 HEAP8[$95>>0] = 0; //@line 3584 "load_abc.cpp"
 $96 = ((($partpat)) + 48|0); //@line 3583 "load_abc.cpp"
 HEAP8[$96>>0] = -1; //@line 3583 "load_abc.cpp"
 $97 = ((($partpat)) + 49|0); //@line 3584 "load_abc.cpp"
 HEAP8[$97>>0] = 0; //@line 3584 "load_abc.cpp"
 $98 = ((($partpat)) + 50|0); //@line 3583 "load_abc.cpp"
 HEAP8[$98>>0] = -1; //@line 3583 "load_abc.cpp"
 $99 = ((($partpat)) + 51|0); //@line 3584 "load_abc.cpp"
 HEAP8[$99>>0] = 0; //@line 3584 "load_abc.cpp"
 $100 = ((($partpat)) + 52|0); //@line 3583 "load_abc.cpp"
 HEAP8[$100>>0] = -1; //@line 3583 "load_abc.cpp"
 $101 = ((($partpat)) + 53|0); //@line 3584 "load_abc.cpp"
 HEAP8[$101>>0] = 0; //@line 3584 "load_abc.cpp"
 $102 = ((($1)) + 32|0); //@line 3507 "load_abc.cpp"
 HEAP8[$102>>0] = 127; //@line 3507 "load_abc.cpp"
 $103 = ((($1)) + 33|0); //@line 3508 "load_abc.cpp"
 HEAP8[$103>>0] = 125; //@line 3508 "load_abc.cpp"
 $104 = ((($1)) + 34|0); //@line 3509 "load_abc.cpp"
 HEAP8[$104>>0] = 110; //@line 3509 "load_abc.cpp"
 $105 = ((($1)) + 35|0); //@line 3510 "load_abc.cpp"
 HEAP8[$105>>0] = 1; //@line 3510 "load_abc.cpp"
 $$1$i = 1282466;
 while(1) {
  $106 = HEAP8[$$1$i>>0]|0; //@line 3512 "load_abc.cpp"
  $107 = $106 << 24 >> 24; //@line 3512 "load_abc.cpp"
  $108 = (_isspace($107)|0); //@line 3512 "load_abc.cpp"
  $109 = ($108|0)==(0); //@line 3512 "load_abc.cpp"
  $110 = ((($$1$i)) + 1|0); //@line 3512 "load_abc.cpp"
  if ($109) {
   $$1$i$lcssa = $$1$i;
   break;
  } else {
   $$1$i = $110;
  }
 }
 $111 = HEAP8[$$1$i$lcssa>>0]|0; //@line 3513 "load_abc.cpp"
 $112 = ($111<<24>>24)==(0); //@line 3513 "load_abc.cpp"
 if ($112) {
  $$1$i$1 = $$1$i$lcssa;
 } else {
  $113 = $111 << 24 >> 24; //@line 1625 "load_abc.cpp"
  $isdigittmp1$i$i = (($113) + -48)|0; //@line 1625 "load_abc.cpp"
  $isdigit2$i$i = ($isdigittmp1$i$i>>>0)<(10); //@line 1625 "load_abc.cpp"
  if ($isdigit2$i$i) {
   $116 = $113;$h$04$i$i = 0;$i$03$i$i = 0;
   while(1) {
    $114 = ($h$04$i$i*10)|0; //@line 1626 "load_abc.cpp"
    $115 = (($116) + -48)|0; //@line 1626 "load_abc.cpp"
    $117 = (($115) + ($114))|0; //@line 1626 "load_abc.cpp"
    $118 = (($i$03$i$i) + 1)|0; //@line 1627 "load_abc.cpp"
    $119 = (($$1$i$lcssa) + ($118)|0); //@line 1625 "load_abc.cpp"
    $120 = HEAP8[$119>>0]|0; //@line 1625 "load_abc.cpp"
    $121 = $120 << 24 >> 24; //@line 1625 "load_abc.cpp"
    $isdigittmp$i$i = (($121) + -48)|0; //@line 1625 "load_abc.cpp"
    $isdigit$i$i = ($isdigittmp$i$i>>>0)<(10); //@line 1625 "load_abc.cpp"
    if ($isdigit$i$i) {
     $116 = $121;$h$04$i$i = $117;$i$03$i$i = $118;
    } else {
     $$lcssa2296 = $117;$$lcssa2297 = $118;$$lcssa2298 = $119;
     break;
    }
   }
   $122 = ($$lcssa2297|0)==(0); //@line 1629 "load_abc.cpp"
   $$h$0$i$i = $122 ? 1 : $$lcssa2296; //@line 1630 "load_abc.cpp"
   $123 = ($$h$0$i$i|0)<(0); //@line 3515 "load_abc.cpp"
   if ($123) {
    $126 = 0;$4042 = $$lcssa2298;
   } else {
    $124 = ($$h$0$i$i|0)>(127); //@line 3516 "load_abc.cpp"
    $$$h$0$i$i = $124 ? 127 : $$h$0$i$i; //@line 3516 "load_abc.cpp"
    $126 = $$$h$0$i$i;$4042 = $$lcssa2298;
   }
  } else {
   $126 = 1;$4042 = $$1$i$lcssa;
  }
  $125 = $126&255; //@line 3517 "load_abc.cpp"
  $127 = ((($1)) + 32|0); //@line 3517 "load_abc.cpp"
  HEAP8[$127>>0] = $125; //@line 3517 "load_abc.cpp"
  $$1$i$1 = $4042;
 }
 while(1) {
  $3974 = HEAP8[$$1$i$1>>0]|0; //@line 3512 "load_abc.cpp"
  $3975 = $3974 << 24 >> 24; //@line 3512 "load_abc.cpp"
  $3976 = (_isspace($3975)|0); //@line 3512 "load_abc.cpp"
  $3977 = ($3976|0)==(0); //@line 3512 "load_abc.cpp"
  $3978 = ((($$1$i$1)) + 1|0); //@line 3512 "load_abc.cpp"
  if ($3977) {
   $$1$i$1$lcssa = $$1$i$1;
   break;
  } else {
   $$1$i$1 = $3978;
  }
 }
 $3979 = HEAP8[$$1$i$1$lcssa>>0]|0; //@line 3513 "load_abc.cpp"
 $3980 = ($3979<<24>>24)==(0); //@line 3513 "load_abc.cpp"
 if ($3980) {
  $$1$i$2 = $$1$i$1$lcssa;
 } else {
  $3981 = $3979 << 24 >> 24; //@line 1625 "load_abc.cpp"
  $isdigittmp1$i$i$1 = (($3981) + -48)|0; //@line 1625 "load_abc.cpp"
  $isdigit2$i$i$1 = ($isdigittmp1$i$i$1>>>0)<(10); //@line 1625 "load_abc.cpp"
  if ($isdigit2$i$i$1) {
   $3984 = $3981;$h$04$i$i$1 = 0;$i$03$i$i$1 = 0;
   while(1) {
    $3982 = ($h$04$i$i$1*10)|0; //@line 1626 "load_abc.cpp"
    $3983 = (($3984) + -48)|0; //@line 1626 "load_abc.cpp"
    $3985 = (($3983) + ($3982))|0; //@line 1626 "load_abc.cpp"
    $3986 = (($i$03$i$i$1) + 1)|0; //@line 1627 "load_abc.cpp"
    $3987 = (($$1$i$1$lcssa) + ($3986)|0); //@line 1625 "load_abc.cpp"
    $3988 = HEAP8[$3987>>0]|0; //@line 1625 "load_abc.cpp"
    $3989 = $3988 << 24 >> 24; //@line 1625 "load_abc.cpp"
    $isdigittmp$i$i$1 = (($3989) + -48)|0; //@line 1625 "load_abc.cpp"
    $isdigit$i$i$1 = ($isdigittmp$i$i$1>>>0)<(10); //@line 1625 "load_abc.cpp"
    if ($isdigit$i$i$1) {
     $3984 = $3989;$h$04$i$i$1 = $3985;$i$03$i$i$1 = $3986;
    } else {
     $$lcssa2293 = $3985;$$lcssa2294 = $3986;$$lcssa2295 = $3987;
     break;
    }
   }
   $3990 = ($$lcssa2294|0)==(0); //@line 1629 "load_abc.cpp"
   $$h$0$i$i$1 = $3990 ? 1 : $$lcssa2293; //@line 1630 "load_abc.cpp"
   $3991 = ($$h$0$i$i$1|0)<(0); //@line 3515 "load_abc.cpp"
   if ($3991) {
    $3994 = 0;$4072 = $$lcssa2295;
   } else {
    $3992 = ($$h$0$i$i$1|0)>(127); //@line 3516 "load_abc.cpp"
    $$$h$0$i$i$1 = $3992 ? 127 : $$h$0$i$i$1; //@line 3516 "load_abc.cpp"
    $3994 = $$$h$0$i$i$1;$4072 = $$lcssa2295;
   }
  } else {
   $3994 = 1;$4072 = $$1$i$1$lcssa;
  }
  $3993 = $3994&255; //@line 3517 "load_abc.cpp"
  $3995 = ((($1)) + 33|0); //@line 3517 "load_abc.cpp"
  HEAP8[$3995>>0] = $3993; //@line 3517 "load_abc.cpp"
  $$1$i$2 = $4072;
 }
 while(1) {
  $3996 = HEAP8[$$1$i$2>>0]|0; //@line 3512 "load_abc.cpp"
  $3997 = $3996 << 24 >> 24; //@line 3512 "load_abc.cpp"
  $3998 = (_isspace($3997)|0); //@line 3512 "load_abc.cpp"
  $3999 = ($3998|0)==(0); //@line 3512 "load_abc.cpp"
  $4000 = ((($$1$i$2)) + 1|0); //@line 3512 "load_abc.cpp"
  if ($3999) {
   $$1$i$2$lcssa = $$1$i$2;
   break;
  } else {
   $$1$i$2 = $4000;
  }
 }
 $4001 = HEAP8[$$1$i$2$lcssa>>0]|0; //@line 3513 "load_abc.cpp"
 $4002 = ($4001<<24>>24)==(0); //@line 3513 "load_abc.cpp"
 if ($4002) {
  $$1$i$3 = $$1$i$2$lcssa;
 } else {
  $4003 = $4001 << 24 >> 24; //@line 1625 "load_abc.cpp"
  $isdigittmp1$i$i$2 = (($4003) + -48)|0; //@line 1625 "load_abc.cpp"
  $isdigit2$i$i$2 = ($isdigittmp1$i$i$2>>>0)<(10); //@line 1625 "load_abc.cpp"
  if ($isdigit2$i$i$2) {
   $4006 = $4003;$h$04$i$i$2 = 0;$i$03$i$i$2 = 0;
   while(1) {
    $4004 = ($h$04$i$i$2*10)|0; //@line 1626 "load_abc.cpp"
    $4005 = (($4006) + -48)|0; //@line 1626 "load_abc.cpp"
    $4007 = (($4005) + ($4004))|0; //@line 1626 "load_abc.cpp"
    $4008 = (($i$03$i$i$2) + 1)|0; //@line 1627 "load_abc.cpp"
    $4009 = (($$1$i$2$lcssa) + ($4008)|0); //@line 1625 "load_abc.cpp"
    $4010 = HEAP8[$4009>>0]|0; //@line 1625 "load_abc.cpp"
    $4011 = $4010 << 24 >> 24; //@line 1625 "load_abc.cpp"
    $isdigittmp$i$i$2 = (($4011) + -48)|0; //@line 1625 "load_abc.cpp"
    $isdigit$i$i$2 = ($isdigittmp$i$i$2>>>0)<(10); //@line 1625 "load_abc.cpp"
    if ($isdigit$i$i$2) {
     $4006 = $4011;$h$04$i$i$2 = $4007;$i$03$i$i$2 = $4008;
    } else {
     $$lcssa2290 = $4007;$$lcssa2291 = $4008;$$lcssa2292 = $4009;
     break;
    }
   }
   $4012 = ($$lcssa2291|0)==(0); //@line 1629 "load_abc.cpp"
   $$h$0$i$i$2 = $4012 ? 1 : $$lcssa2290; //@line 1630 "load_abc.cpp"
   $4013 = ($$h$0$i$i$2|0)<(0); //@line 3515 "load_abc.cpp"
   if ($4013) {
    $4016 = 0;$4073 = $$lcssa2292;
   } else {
    $4014 = ($$h$0$i$i$2|0)>(127); //@line 3516 "load_abc.cpp"
    $$$h$0$i$i$2 = $4014 ? 127 : $$h$0$i$i$2; //@line 3516 "load_abc.cpp"
    $4016 = $$$h$0$i$i$2;$4073 = $$lcssa2292;
   }
  } else {
   $4016 = 1;$4073 = $$1$i$2$lcssa;
  }
  $4015 = $4016&255; //@line 3517 "load_abc.cpp"
  $4017 = ((($1)) + 34|0); //@line 3517 "load_abc.cpp"
  HEAP8[$4017>>0] = $4015; //@line 3517 "load_abc.cpp"
  $$1$i$3 = $4073;
 }
 while(1) {
  $4018 = HEAP8[$$1$i$3>>0]|0; //@line 3512 "load_abc.cpp"
  $4019 = $4018 << 24 >> 24; //@line 3512 "load_abc.cpp"
  $4020 = (_isspace($4019)|0); //@line 3512 "load_abc.cpp"
  $4021 = ($4020|0)==(0); //@line 3512 "load_abc.cpp"
  $4022 = ((($$1$i$3)) + 1|0); //@line 3512 "load_abc.cpp"
  if ($4021) {
   $$1$i$3$lcssa = $$1$i$3;
   break;
  } else {
   $$1$i$3 = $4022;
  }
 }
 $4023 = HEAP8[$$1$i$3$lcssa>>0]|0; //@line 3513 "load_abc.cpp"
 $4024 = ($4023<<24>>24)==(0); //@line 3513 "load_abc.cpp"
 if ($4024) {
  $$pre = HEAP8[$105>>0]|0; //@line 3520 "load_abc.cpp"
  $4041 = $$pre;
 } else {
  $4025 = $4023 << 24 >> 24; //@line 1625 "load_abc.cpp"
  $isdigittmp1$i$i$3 = (($4025) + -48)|0; //@line 1625 "load_abc.cpp"
  $isdigit2$i$i$3 = ($isdigittmp1$i$i$3>>>0)<(10); //@line 1625 "load_abc.cpp"
  if ($isdigit2$i$i$3) {
   $4028 = $4025;$h$04$i$i$3 = 0;$i$03$i$i$3 = 0;
   while(1) {
    $4026 = ($h$04$i$i$3*10)|0; //@line 1626 "load_abc.cpp"
    $4027 = (($4028) + -48)|0; //@line 1626 "load_abc.cpp"
    $4029 = (($4027) + ($4026))|0; //@line 1626 "load_abc.cpp"
    $4030 = (($i$03$i$i$3) + 1)|0; //@line 1627 "load_abc.cpp"
    $4031 = (($$1$i$3$lcssa) + ($4030)|0); //@line 1625 "load_abc.cpp"
    $4032 = HEAP8[$4031>>0]|0; //@line 1625 "load_abc.cpp"
    $4033 = $4032 << 24 >> 24; //@line 1625 "load_abc.cpp"
    $isdigittmp$i$i$3 = (($4033) + -48)|0; //@line 1625 "load_abc.cpp"
    $isdigit$i$i$3 = ($isdigittmp$i$i$3>>>0)<(10); //@line 1625 "load_abc.cpp"
    if ($isdigit$i$i$3) {
     $4028 = $4033;$h$04$i$i$3 = $4029;$i$03$i$i$3 = $4030;
    } else {
     $$lcssa2288 = $4029;$$lcssa2289 = $4030;
     break;
    }
   }
   $4034 = ($$lcssa2289|0)==(0); //@line 1629 "load_abc.cpp"
   $$h$0$i$i$3 = $4034 ? 1 : $$lcssa2288; //@line 1630 "load_abc.cpp"
   $4035 = ($$h$0$i$i$3|0)<(0); //@line 3515 "load_abc.cpp"
   if ($4035) {
    $4038 = 0;
   } else {
    $4036 = ($$h$0$i$i$3|0)>(127); //@line 3516 "load_abc.cpp"
    $$$h$0$i$i$3 = $4036 ? 127 : $$h$0$i$i$3; //@line 3516 "load_abc.cpp"
    $4038 = $$$h$0$i$i$3;
   }
  } else {
   $4038 = 1;
  }
  $4037 = $4038&255; //@line 3517 "load_abc.cpp"
  $4039 = ((($1)) + 35|0); //@line 3517 "load_abc.cpp"
  HEAP8[$4039>>0] = $4037; //@line 3517 "load_abc.cpp"
  $4041 = $4037;
 }
 $4040 = ($4041<<24>>24)==(0); //@line 3520 "load_abc.cpp"
 if ($4040) {
  HEAP8[$105>>0] = 1; //@line 3520 "load_abc.cpp"
  $$0$i107 = 1282466;
 } else {
  $$0$i107 = 1282466;
 }
 while(1) {
  $128 = HEAP8[$$0$i107>>0]|0; //@line 3535 "load_abc.cpp"
  $129 = $128 << 24 >> 24; //@line 3535 "load_abc.cpp"
  $130 = (_isspace($129)|0); //@line 3535 "load_abc.cpp"
  $131 = ($130|0)==(0); //@line 3535 "load_abc.cpp"
  $132 = ((($$0$i107)) + 1|0); //@line 3535 "load_abc.cpp"
  if ($131) {
   $$0$i107$lcssa = $$0$i107;
   break;
  } else {
   $$0$i107 = $132;
  }
 }
 $133 = HEAPU8[$3>>0]|(HEAPU8[$3+1>>0]<<8)|(HEAPU8[$3+2>>0]<<16)|(HEAPU8[$3+3>>0]<<24); //@line 3536 "load_abc.cpp"
 $134 = ($133|0)==(0|0); //@line 3536 "load_abc.cpp"
 if (!($134)) {
  _free($133); //@line 3536 "load_abc.cpp"
 }
 $strlenfirst$i = HEAP8[$$0$i107$lcssa>>0]|0; //@line 3537 "load_abc.cpp"
 $135 = ($strlenfirst$i<<24>>24)==(0); //@line 3537 "load_abc.cpp"
 if ($135) {
  $storemerge$i = 0;
 } else {
  $136 = (___strdup($$0$i107$lcssa)|0); //@line 3538 "load_abc.cpp"
  $storemerge$i = $136;
 }
 HEAP8[$3>>0]=$storemerge$i&255;HEAP8[$3+1>>0]=($storemerge$i>>8)&255;HEAP8[$3+2>>0]=($storemerge$i>>16)&255;HEAP8[$3+3>>0]=$storemerge$i>>24; //@line 3538 "load_abc.cpp"
 HEAP32[$orderlist>>2] = 0; //@line 3885 "load_abc.cpp"
 HEAP32[$mmstack>>2] = $mm; //@line 3888 "load_abc.cpp"
 HEAP32[$25>>2] = 0; //@line 520 "load_abc.cpp"
 $137 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8)|(HEAPU8[$4+2>>0]<<16)|(HEAPU8[$4+3>>0]<<24); //@line 3891 "load_abc.cpp"
 $138 = ($137|0)==(0); //@line 3891 "load_abc.cpp"
 if ($138) {
  $abcstate$1$ph = 0;$abcxwanted$1$ph = 0;
 } else {
  HEAP32[$25>>2] = 0; //@line 520 "load_abc.cpp"
  $139 = (__ZL8abc_getsP10_ABCHANDLEP6MMFILE($1,$mm)|0); //@line 3895 "load_abc.cpp"
  $140 = ($139|0)==(0|0); //@line 3895 "load_abc.cpp"
  if ($140) {
   $abcstate$0 = 0;$abcxwanted$0 = 0;
  } else {
   $4043 = $139;$abcxcount$01335 = 0;
   while(1) {
    $p$0 = $4043;
    while(1) {
     $141 = HEAP8[$p$0>>0]|0; //@line 3896 "load_abc.cpp"
     $142 = $141 << 24 >> 24; //@line 3896 "load_abc.cpp"
     $143 = (_isspace($142)|0); //@line 3896 "load_abc.cpp"
     $144 = ($143|0)==(0); //@line 3896 "load_abc.cpp"
     $145 = ((($p$0)) + 1|0); //@line 3896 "load_abc.cpp"
     if ($144) {
      $p$0$lcssa = $p$0;
      break;
     } else {
      $p$0 = $145;
     }
    }
    $146 = (_strncmp($p$0$lcssa,525008,2)|0); //@line 3897 "load_abc.cpp"
    $147 = ($146|0)==(0); //@line 3897 "load_abc.cpp"
    $148 = $147&1; //@line 3897 "load_abc.cpp"
    $$abcxcount$0 = (($148) + ($abcxcount$01335))|0; //@line 3897 "load_abc.cpp"
    $149 = (__ZL8abc_getsP10_ABCHANDLEP6MMFILE($1,$mm)|0); //@line 3895 "load_abc.cpp"
    $150 = ($149|0)==(0|0); //@line 3895 "load_abc.cpp"
    if ($150) {
     $$abcxcount$0$lcssa = $$abcxcount$0;
     break;
    } else {
     $4043 = $149;$abcxcount$01335 = $$abcxcount$0;
    }
   }
   $151 = ($$abcxcount$0$lcssa|0)==(0); //@line 3899 "load_abc.cpp"
   if ($151) {
    $abcstate$0 = 0;$abcxwanted$0 = 0;
   } else {
    $152 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8)|(HEAPU8[$4+2>>0]<<16)|(HEAPU8[$4+3>>0]<<24); //@line 3902 "load_abc.cpp"
    $153 = (($152) + -1)|0; //@line 3902 "load_abc.cpp"
    $154 = (($153|0) % ($$abcxcount$0$lcssa|0))&-1; //@line 3902 "load_abc.cpp"
    $abcstate$0 = 4;$abcxwanted$0 = $154;
   }
  }
  HEAP32[$25>>2] = 0; //@line 520 "load_abc.cpp"
  $abcstate$1$ph = $abcstate$0;$abcxwanted$1$ph = $abcxwanted$0;
 }
 $155 = HEAP32[37544]|0; //@line 328 "load_abc.cpp"
 $156 = ((($1)) + 8|0); //@line 4244 "load_abc.cpp"
 $157 = ((($this)) + 59804|0); //@line 3923 "load_abc.cpp"
 $158 = ((($buf)) + 31|0); //@line 4011 "load_abc.cpp"
 $159 = ((($buf1)) + 31|0); //@line 4169 "load_abc.cpp"
 $160 = ((($1)) + 4|0); //@line 4306 "load_abc.cpp"
 $161 = ((($s$i)) + 1|0); //@line 1519 "load_abc.cpp"
 $162 = ((($s$i)) + 2|0); //@line 1519 "load_abc.cpp"
 $163 = ((($s$i)) + 3|0); //@line 1519 "load_abc.cpp"
 $164 = ((($s$i)) + 4|0); //@line 1519 "load_abc.cpp"
 $165 = ((($s$i)) + 5|0); //@line 1519 "load_abc.cpp"
 $166 = ((($s$i)) + 6|0); //@line 1519 "load_abc.cpp"
 $167 = ((($1)) + 32|0); //@line 3517 "load_abc.cpp"
 $168 = ((($1)) + 33|0); //@line 3517 "load_abc.cpp"
 $169 = ((($1)) + 34|0); //@line 3517 "load_abc.cpp"
 $170 = ((($1)) + 35|0); //@line 3517 "load_abc.cpp"
 $4044 = 0;$4045 = 0;$4046 = 0;$4047 = 0;$abcarpeggio$01324 = 0;$abceffect$01315 = 0;$abceffoper$01316 = 0;$abcfermata$01323 = 0;$abcgrace$01331 = 0;$abchornpipe$01329 = 0;$abckey$01319 = 0;$abcnolegato$01322 = 1;$abcnoslurs$01321 = 1;$abcstate$11314 = $abcstate$1$ph;$abctempo$01326 = 0;$abcto$01325 = 0;$abcvol$01320 = 120;$abcxcount$11317 = 0;$abcxnumber$01318 = 1;$bartime$01332 = 0;$ch0$01312 = 0;$cnotediv$01328 = 0;$cnotelen$01327 = 0;$ktempo$01330 = 0;$mmsp$01334 = 1;$orderlen$01313 = 0;$tpd$01333 = 0;
 while(1) {
  $171 = (($mmsp$01334) + -1)|0; //@line 3907 "load_abc.cpp"
  $172 = (($mmstack) + ($171<<2)|0); //@line 3908 "load_abc.cpp"
  $173 = HEAP32[$172>>2]|0; //@line 3908 "load_abc.cpp"
  $174 = (__ZL8abc_getsP10_ABCHANDLEP6MMFILE($1,$173)|0); //@line 3908 "load_abc.cpp"
  $175 = ($174|0)==(0|0); //@line 3908 "load_abc.cpp"
  L91: do {
   if ($175) {
    $$lcssa928 = $4047;$$lcssa948 = $172;$$ph892$lcssa = $4046;$$ph893$lcssa = $4045;$$ph894$lcssa = $4044;$abcarpeggio$1$ph$lcssa = $abcarpeggio$01324;$abceffect$1$lcssa = $abceffect$01315;$abceffoper$1$lcssa = $abceffoper$01316;$abcfermata$1$ph$lcssa = $abcfermata$01323;$abcgrace$1$lcssa = $abcgrace$01331;$abchornpipe$1$lcssa = $abchornpipe$01329;$abckey$1$lcssa = $abckey$01319;$abcnolegato$1$lcssa = $abcnolegato$01322;$abcnoslurs$1$lcssa = $abcnoslurs$01321;$abcstate$2$lcssa = $abcstate$11314;$abctempo$1$lcssa = $abctempo$01326;$abcto$1$lcssa = $abcto$01325;$abcvol$1$lcssa = $abcvol$01320;$abcxcount$2$lcssa = $abcxcount$11317;$abcxnumber$1$lcssa = $abcxnumber$01318;$bartime$1$lcssa = $bartime$01332;$ch0$1$ph$lcssa = $ch0$01312;$cnotediv$1$ph$lcssa = $cnotediv$01328;$cnotelen$1$ph$lcssa = $cnotelen$01327;$ktempo$1$lcssa = $ktempo$01330;$mmsp$1$lcssa = $171;$orderlen$1$lcssa = $orderlen$01313;$tpd$1$lcssa = $tpd$01333;
   } else {
    $$ph1257 = $4047;$$ph8921258 = $4046;$$ph8931259 = $4045;$$ph8941260 = $4044;$4048 = $174;$4049 = $172;$abcarpeggio$1$ph1273 = $abcarpeggio$01324;$abceffect$1$ph1264 = $abceffect$01315;$abceffoper$1$ph1265 = $abceffoper$01316;$abcfermata$1$ph1272 = $abcfermata$01323;$abcgrace$1$ph1280 = $abcgrace$01331;$abchornpipe$1$ph1278 = $abchornpipe$01329;$abckey$1$ph1268 = $abckey$01319;$abcnolegato$1$ph1271 = $abcnolegato$01322;$abcnoslurs$1$ph1270 = $abcnoslurs$01321;$abcstate$2$ph1263 = $abcstate$11314;$abctempo$1$ph1275 = $abctempo$01326;$abcto$1$ph1274 = $abcto$01325;$abcvol$1$ph1269 = $abcvol$01320;$abcxcount$2$ph1266 = $abcxcount$11317;$abcxnumber$1$ph1267 = $abcxnumber$01318;$bartime$1$ph1281 = $bartime$01332;$ch0$1$ph1261 = $ch0$01312;$cnotediv$1$ph1277 = $cnotediv$01328;$cnotelen$1$ph1276 = $cnotelen$01327;$ktempo$1$ph1279 = $ktempo$01330;$mmsp$1$ph1283 = $171;$orderlen$1$ph1262 = $orderlen$01313;$tpd$1$ph1282 = $tpd$01333;
    while(1) {
     $180 = ($ch0$1$ph1261<<24>>24)==(92); //@line 4141 "load_abc.cpp"
     $181 = $180&1; //@line 4141 "load_abc.cpp"
     $4050 = $4048;$4051 = $$ph1257;$776 = $4049;$abceffect$11093 = $abceffect$1$ph1264;$abceffoper$11095 = $abceffoper$1$ph1265;$abcgrace$11117 = $abcgrace$1$ph1280;$abchornpipe$11113 = $abchornpipe$1$ph1278;$abckey$11101 = $abckey$1$ph1268;$abcnolegato$11107 = $abcnolegato$1$ph1271;$abcnoslurs$11106 = $abcnoslurs$1$ph1270;$abcstate$21092 = $abcstate$2$ph1263;$abctempo$11111 = $abctempo$1$ph1275;$abcto$11109 = $abcto$1$ph1274;$abcvol$11104 = $abcvol$1$ph1269;$abcxcount$21097 = $abcxcount$2$ph1266;$abcxnumber$11099 = $abcxnumber$1$ph1267;$bartime$11119 = $bartime$1$ph1281;$ktempo$11115 = $ktempo$1$ph1279;$mmsp$11123 = $mmsp$1$ph1283;$orderlen$11090 = $orderlen$1$ph1262;$tpd$11121 = $tpd$1$ph1282;
     L94: while(1) {
      $p$1 = $4050;
      while(1) {
       $182 = HEAP8[$p$1>>0]|0; //@line 3909 "load_abc.cpp"
       $183 = $182 << 24 >> 24; //@line 3909 "load_abc.cpp"
       $184 = (_isspace($183)|0); //@line 3909 "load_abc.cpp"
       $185 = ($184|0)==(0); //@line 3909 "load_abc.cpp"
       $186 = ((($p$1)) + 1|0); //@line 3909 "load_abc.cpp"
       if ($185) {
        $$lcssa2125 = $186;$p$1$lcssa = $p$1;
        break;
       } else {
        $p$1 = $186;
       }
      }
      switch ($abcstate$21092|0) {
      case 4:  {
       $187 = (_strncmp($p$1$lcssa,525008,2)|0); //@line 3912 "load_abc.cpp"
       $188 = ($187|0)==(0); //@line 3912 "load_abc.cpp"
       if ($188) {
        $189 = (($abcxcount$21097) + 1)|0; //@line 3913 "load_abc.cpp"
        $190 = ($abcxcount$21097|0)==($abcxwanted$1$ph|0); //@line 3913 "load_abc.cpp"
        if ($190) {
         $abcxcount$3 = $189;
         label = 43;
        } else {
         $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = 4;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $189;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$1$lcssa;$tpd$2 = $tpd$11121;
        }
       } else {
        $abcxcount$3 = $abcxcount$21097;
        label = 43;
       }
       break;
      }
      case 1:  {
       $abcxcount$3 = $abcxcount$21097;
       label = 43;
       break;
      }
      case 0:  {
       $237 = HEAP8[$p$1$lcssa>>0]|0; //@line 3956 "load_abc.cpp"
       $238 = ($237<<24>>24)==(0); //@line 3956 "load_abc.cpp"
       if ($238) {
        $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = 0;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$1$lcssa;$tpd$2 = $tpd$11121;
       } else {
        $239 = HEAP8[$$lcssa2125>>0]|0; //@line 3956 "load_abc.cpp"
        $240 = ($239<<24>>24)==(58); //@line 3956 "load_abc.cpp"
        if ($240) {
         $241 = HEAP32[$abcparts>>2]|0; //@line 1820 "load_abc.cpp"
         $242 = ($241|0)==(0|0); //@line 1820 "load_abc.cpp"
         if (!($242)) {
          _free($241); //@line 1820 "load_abc.cpp"
         }
         HEAP32[$abcparts>>2] = 0; //@line 1822 "load_abc.cpp"
         HEAP8[$6>>0] = 0; //@line 3959 "load_abc.cpp"
         $243 = HEAPU8[$38>>0]|(HEAPU8[$38+1>>0]<<8)|(HEAPU8[$38+2>>0]<<16)|(HEAPU8[$38+3>>0]<<24); //@line 3960 "load_abc.cpp"
         $244 = ($243|0)==(0); //@line 3960 "load_abc.cpp"
         L108: do {
          if (!($244)) {
           $245 = HEAPU8[$43>>0]|(HEAPU8[$43+1>>0]<<8)|(HEAPU8[$43+2>>0]<<16)|(HEAPU8[$43+3>>0]<<24); //@line 3960 "load_abc.cpp"
           $246 = ($245|0)==(0|0); //@line 3960 "load_abc.cpp"
           if ($246) {
            break;
           }
           $247 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 3960 "load_abc.cpp"
           $248 = ((($245)) + 44|0); //@line 1616 "load_abc.cpp"
           $249 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,$248,8)|0); //@line 1616 "load_abc.cpp"
           $250 = ((($249)) + 8|0); //@line 1590 "load_abc.cpp"
           $251 = HEAPU8[$250>>0]|(HEAPU8[$250+1>>0]<<8)|(HEAPU8[$250+2>>0]<<16)|(HEAPU8[$250+3>>0]<<24); //@line 1590 "load_abc.cpp"
           $252 = ($251|0)==(0|0); //@line 1591 "load_abc.cpp"
           do {
            if (!($252)) {
             $253 = ((($251)) + 4|0); //@line 1591 "load_abc.cpp"
             $254 = HEAPU8[$253>>0]|(HEAPU8[$253+1>>0]<<8)|(HEAPU8[$253+2>>0]<<16)|(HEAPU8[$253+3>>0]<<24); //@line 1591 "load_abc.cpp"
             $255 = ($254|0)==($247|0); //@line 1591 "load_abc.cpp"
             if ($255) {
              break L108;
             }
             $256 = ((($251)) + 8|0); //@line 1592 "load_abc.cpp"
             $257 = HEAP8[$256>>0]|0; //@line 1592 "load_abc.cpp"
             $258 = ($257<<24>>24)==(1); //@line 1592 "load_abc.cpp"
             if (!($258)) {
              break;
             }
             $259 = ((($251)) + 9|0); //@line 1592 "load_abc.cpp"
             $260 = HEAP8[$259>>0]|0; //@line 1592 "load_abc.cpp"
             $261 = ($260<<24>>24)==(83); //@line 1592 "load_abc.cpp"
             if (!($261)) {
              break;
             }
             HEAP8[$253>>0]=$247&255;HEAP8[$253+1>>0]=($247>>8)&255;HEAP8[$253+2>>0]=($247>>16)&255;HEAP8[$253+3>>0]=$247>>24; //@line 1593 "load_abc.cpp"
             break L108;
            }
           } while(0);
           $262 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
           HEAP8[$262>>0]=0&255;HEAP8[$262+1>>0]=(0>>8)&255;HEAP8[$262+2>>0]=(0>>16)&255;HEAP8[$262+3>>0]=0>>24; //@line 540 "load_abc.cpp"
           $263 = ((($262)) + 4|0); //@line 541 "load_abc.cpp"
           HEAP8[$263>>0]=$247&255;HEAP8[$263+1>>0]=($247>>8)&255;HEAP8[$263+2>>0]=($247>>16)&255;HEAP8[$263+3>>0]=$247>>24; //@line 541 "load_abc.cpp"
           $264 = ((($262)) + 8|0); //@line 543 "load_abc.cpp"
           HEAP8[$264>>0] = 1; //@line 543 "load_abc.cpp"
           $265 = ((($262)) + 9|0); //@line 543 "load_abc.cpp"
           HEAP8[$265>>0] = 83; //@line 543 "load_abc.cpp"
           $266 = ((($262)) + 10|0); //@line 543 "load_abc.cpp"
           HEAP8[$266>>0]=0&255;HEAP8[$266+1>>0]=(0>>8)&255;HEAP8[$266+2>>0]=(0>>16)&255;HEAP8[$266+3>>0]=0>>24; //@line 543 "load_abc.cpp"
           $267 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
           $268 = ((($262)) + 14|0); //@line 544 "load_abc.cpp"
           HEAP8[$268>>0] = $267; //@line 544 "load_abc.cpp"
           $269 = ((($262)) + 15|0); //@line 545 "load_abc.cpp"
           HEAP8[$269>>0] = 0; //@line 545 "load_abc.cpp"
           __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($249,$262); //@line 1600 "load_abc.cpp"
          }
         } while(0);
         $270 = HEAPU8[$42>>0]|(HEAPU8[$42+1>>0]<<8)|(HEAPU8[$42+2>>0]<<16)|(HEAPU8[$42+3>>0]<<24); //@line 3961 "load_abc.cpp"
         $271 = ($270|0)==(0|0); //@line 3961 "load_abc.cpp"
         do {
          if (!($271)) {
           $272 = HEAPU8[$39>>0]|(HEAPU8[$39+1>>0]<<8)|(HEAPU8[$39+2>>0]<<16)|(HEAPU8[$39+3>>0]<<24); //@line 3961 "load_abc.cpp"
           $273 = ($272|0)==(0); //@line 3961 "load_abc.cpp"
           if (!($273)) {
            break;
           }
           $274 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 3961 "load_abc.cpp"
           $275 = ((($270)) + 44|0); //@line 1608 "load_abc.cpp"
           $i$01$i = 1;
           while(1) {
            $276 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,$275,$i$01$i)|0); //@line 1608 "load_abc.cpp"
            $277 = ((($276)) + 8|0); //@line 1590 "load_abc.cpp"
            $278 = HEAPU8[$277>>0]|(HEAPU8[$277+1>>0]<<8)|(HEAPU8[$277+2>>0]<<16)|(HEAPU8[$277+3>>0]<<24); //@line 1590 "load_abc.cpp"
            $279 = ($278|0)==(0|0); //@line 1591 "load_abc.cpp"
            do {
             if ($279) {
              label = 85;
             } else {
              $280 = ((($278)) + 4|0); //@line 1591 "load_abc.cpp"
              $281 = HEAPU8[$280>>0]|(HEAPU8[$280+1>>0]<<8)|(HEAPU8[$280+2>>0]<<16)|(HEAPU8[$280+3>>0]<<24); //@line 1591 "load_abc.cpp"
              $282 = ($281|0)==($274|0); //@line 1591 "load_abc.cpp"
              if ($282) {
               break;
              }
              $283 = ((($278)) + 8|0); //@line 1592 "load_abc.cpp"
              $284 = HEAP8[$283>>0]|0; //@line 1592 "load_abc.cpp"
              $285 = ($284<<24>>24)==(1); //@line 1592 "load_abc.cpp"
              if (!($285)) {
               label = 85;
               break;
              }
              $286 = ((($278)) + 9|0); //@line 1592 "load_abc.cpp"
              $287 = HEAP8[$286>>0]|0; //@line 1592 "load_abc.cpp"
              $288 = ($287<<24>>24)==(83); //@line 1592 "load_abc.cpp"
              if (!($288)) {
               label = 85;
               break;
              }
              HEAP8[$280>>0]=$274&255;HEAP8[$280+1>>0]=($274>>8)&255;HEAP8[$280+2>>0]=($274>>16)&255;HEAP8[$280+3>>0]=$274>>24; //@line 1593 "load_abc.cpp"
             }
            } while(0);
            if ((label|0) == 85) {
             label = 0;
             $289 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
             HEAP8[$289>>0]=0&255;HEAP8[$289+1>>0]=(0>>8)&255;HEAP8[$289+2>>0]=(0>>16)&255;HEAP8[$289+3>>0]=0>>24; //@line 540 "load_abc.cpp"
             $290 = ((($289)) + 4|0); //@line 541 "load_abc.cpp"
             HEAP8[$290>>0]=$274&255;HEAP8[$290+1>>0]=($274>>8)&255;HEAP8[$290+2>>0]=($274>>16)&255;HEAP8[$290+3>>0]=$274>>24; //@line 541 "load_abc.cpp"
             $291 = ((($289)) + 8|0); //@line 543 "load_abc.cpp"
             HEAP8[$291>>0] = 1; //@line 543 "load_abc.cpp"
             $292 = ((($289)) + 9|0); //@line 543 "load_abc.cpp"
             HEAP8[$292>>0] = 83; //@line 543 "load_abc.cpp"
             $293 = ((($289)) + 10|0); //@line 543 "load_abc.cpp"
             HEAP8[$293>>0]=0&255;HEAP8[$293+1>>0]=(0>>8)&255;HEAP8[$293+2>>0]=(0>>16)&255;HEAP8[$293+3>>0]=0>>24; //@line 543 "load_abc.cpp"
             $294 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
             $295 = ((($289)) + 14|0); //@line 544 "load_abc.cpp"
             HEAP8[$295>>0] = $294; //@line 544 "load_abc.cpp"
             $296 = ((($289)) + 15|0); //@line 545 "load_abc.cpp"
             HEAP8[$296>>0] = 0; //@line 545 "load_abc.cpp"
             __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($276,$289); //@line 1600 "load_abc.cpp"
            }
            $297 = (($i$01$i) + 1)|0; //@line 1607 "load_abc.cpp"
            $exitcond$i132 = ($297|0)==(8); //@line 1607 "load_abc.cpp"
            if ($exitcond$i132) {
             break;
            } else {
             $i$01$i = $297;
            }
           }
          }
         } while(0);
         HEAP8[$39>>0]=1&255;HEAP8[$39+1>>0]=(1>>8)&255;HEAP8[$39+2>>0]=(1>>16)&255;HEAP8[$39+3>>0]=1>>24; //@line 3962 "load_abc.cpp"
         HEAP8[$38>>0]=0&255;HEAP8[$38+1>>0]=(0>>8)&255;HEAP8[$38+2>>0]=(0>>16)&255;HEAP8[$38+3>>0]=0>>24; //@line 3963 "load_abc.cpp"
         label = 88;
        } else {
         $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = 0;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$1$lcssa;$tpd$2 = $tpd$11121;
        }
       }
       break;
      }
      case 2:  {
       label = 88;
       break;
      }
      case 5:  {
       $p$4 = $p$1$lcssa;
       while(1) {
        $527 = HEAP8[$p$4>>0]|0; //@line 4080 "load_abc.cpp"
        $528 = ($527<<24>>24)==(0); //@line 4080 "load_abc.cpp"
        $529 = ($527<<24>>24)!=(34); //@line 4080 "load_abc.cpp"
        $not$832 = $528 ^ 1; //@line 4080 "load_abc.cpp"
        $$81 = $529 & $not$832; //@line 4080 "load_abc.cpp"
        $530 = ((($p$4)) + 1|0); //@line 4081 "load_abc.cpp"
        if ($$81) {
         $p$4 = $530;
        } else {
         break;
        }
       }
       if ($528) {
        $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = 5;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$4;$tpd$2 = $tpd$11121;
       } else {
        $abcstate$5 = 3;$orderlen$2 = $orderlen$11090;$p$51550 = $p$4;
        label = 218;
       }
       break;
      }
      case 3:  {
       $strlenfirst$pre = HEAP8[$p$1$lcssa>>0]|0; //@line 4087 "load_abc.cpp"
       $531 = ($strlenfirst$pre<<24>>24)==(0); //@line 4087 "load_abc.cpp"
       if ($531) {
        $532 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 4087 "load_abc.cpp"
        $533 = ($532|0)==(0|0); //@line 4087 "load_abc.cpp"
        if ($533) {
         $abcstate$5 = 3;$orderlen$2 = $orderlen$11090;$p$51550 = $p$1$lcssa;
         label = 218;
        } else {
         $534 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8)|(HEAPU8[$4+2>>0]<<16)|(HEAPU8[$4+3>>0]<<24); //@line 4088 "load_abc.cpp"
         $535 = ($534|0)!=(0); //@line 4088 "load_abc.cpp"
         $536 = $535 ? 4 : 1; //@line 4088 "load_abc.cpp"
         HEAP8[$31>>0]=0&255;HEAP8[$31+1>>0]=(0>>8)&255;HEAP8[$31+2>>0]=(0>>16)&255;HEAP8[$31+3>>0]=0>>24; //@line 3670 "load_abc.cpp"
         $543 = 0;$ttp$02$i = $532;
         while(1) {
          $537 = ((($ttp$02$i)) + 8|0); //@line 3672 "load_abc.cpp"
          $538 = HEAPU8[$537>>0]|(HEAPU8[$537+1>>0]<<8)|(HEAPU8[$537+2>>0]<<16)|(HEAPU8[$537+3>>0]<<24); //@line 3672 "load_abc.cpp"
          $539 = ($538|0)==(0|0); //@line 3672 "load_abc.cpp"
          do {
           if ($539) {
            $4054 = $543;
           } else {
            $540 = ((($538)) + 4|0); //@line 3672 "load_abc.cpp"
            $541 = HEAPU8[$540>>0]|(HEAPU8[$540+1>>0]<<8)|(HEAPU8[$540+2>>0]<<16)|(HEAPU8[$540+3>>0]<<24); //@line 3672 "load_abc.cpp"
            $542 = ($541>>>0)>($543>>>0); //@line 3672 "load_abc.cpp"
            if (!($542)) {
             $4054 = $543;
             break;
            }
            HEAP8[$31>>0]=$541&255;HEAP8[$31+1>>0]=($541>>8)&255;HEAP8[$31+2>>0]=($541>>16)&255;HEAP8[$31+3>>0]=$541>>24; //@line 3673 "load_abc.cpp"
            $4054 = $541;
           }
          } while(0);
          $ttp$0$i = HEAPU8[$ttp$02$i>>0]|(HEAPU8[$ttp$02$i+1>>0]<<8)|(HEAPU8[$ttp$02$i+2>>0]<<16)|(HEAPU8[$ttp$02$i+3>>0]<<24); //@line 3671 "load_abc.cpp"
          $544 = ($ttp$0$i|0)==(0|0); //@line 3671 "load_abc.cpp"
          if ($544) {
           $552 = $4054;$ttp$11061 = $532;
           break;
          } else {
           $543 = $4054;$ttp$02$i = $ttp$0$i;
          }
         }
         while(1) {
          $545 = ((($ttp$11061)) + 42|0); //@line 1174 "load_abc.cpp"
          $546 = HEAP8[$545>>0]|0; //@line 1174 "load_abc.cpp"
          $547 = $546&255; //@line 1174 "load_abc.cpp"
          $548 = (_pat_gmtosmp($547)|0); //@line 1174 "load_abc.cpp"
          $549 = $548&255; //@line 1174 "load_abc.cpp"
          $550 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$550>>0]=0&255;HEAP8[$550+1>>0]=(0>>8)&255;HEAP8[$550+2>>0]=(0>>16)&255;HEAP8[$550+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $551 = ((($550)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$551>>0]=$552&255;HEAP8[$551+1>>0]=($552>>8)&255;HEAP8[$551+2>>0]=($552>>16)&255;HEAP8[$551+3>>0]=$552>>24; //@line 541 "load_abc.cpp"
          $553 = ((($550)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$553>>0] = 0; //@line 543 "load_abc.cpp"
          $554 = ((($550)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$554>>0] = 0; //@line 543 "load_abc.cpp"
          $555 = ((($550)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$555>>0] = $549; //@line 543 "load_abc.cpp"
          $556 = ((($550)) + 11|0); //@line 543 "load_abc.cpp"
          HEAP8[$556>>0] = 0; //@line 543 "load_abc.cpp"
          $557 = ((($550)) + 12|0); //@line 543 "load_abc.cpp"
          HEAP8[$557>>0] = 0; //@line 543 "load_abc.cpp"
          $558 = ((($550)) + 13|0); //@line 543 "load_abc.cpp"
          HEAP8[$558>>0] = 0; //@line 543 "load_abc.cpp"
          $559 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $560 = ((($550)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$560>>0] = $559; //@line 544 "load_abc.cpp"
          $561 = ((($550)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$561>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($ttp$11061,$550); //@line 1179 "load_abc.cpp"
          $ttp$1 = HEAPU8[$ttp$11061>>0]|(HEAPU8[$ttp$11061+1>>0]<<8)|(HEAPU8[$ttp$11061+2>>0]<<16)|(HEAPU8[$ttp$11061+3>>0]<<24); //@line 4091 "load_abc.cpp"
          $562 = ($ttp$1|0)==(0|0); //@line 4091 "load_abc.cpp"
          if ($562) {
           break;
          }
          $$pre1509 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4092 "load_abc.cpp"
          $552 = $$pre1509;$ttp$11061 = $ttp$1;
         }
         $$pre1510 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 4093 "load_abc.cpp"
         $$pre1511 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4093 "load_abc.cpp"
         $563 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
         HEAP8[$563>>0]=0&255;HEAP8[$563+1>>0]=(0>>8)&255;HEAP8[$563+2>>0]=(0>>16)&255;HEAP8[$563+3>>0]=0>>24; //@line 540 "load_abc.cpp"
         $564 = ((($563)) + 4|0); //@line 541 "load_abc.cpp"
         HEAP8[$564>>0]=$$pre1511&255;HEAP8[$564+1>>0]=($$pre1511>>8)&255;HEAP8[$564+2>>0]=($$pre1511>>16)&255;HEAP8[$564+3>>0]=$$pre1511>>24; //@line 541 "load_abc.cpp"
         $565 = ((($563)) + 8|0); //@line 543 "load_abc.cpp"
         HEAP8[$565>>0] = 1; //@line 543 "load_abc.cpp"
         $566 = ((($563)) + 9|0); //@line 543 "load_abc.cpp"
         HEAP8[$566>>0] = 80; //@line 543 "load_abc.cpp"
         $567 = ((($563)) + 10|0); //@line 543 "load_abc.cpp"
         HEAP8[$567>>0]=0&255;HEAP8[$567+1>>0]=(0>>8)&255;HEAP8[$567+2>>0]=(0>>16)&255;HEAP8[$567+3>>0]=0>>24; //@line 543 "load_abc.cpp"
         $568 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
         $569 = ((($563)) + 14|0); //@line 544 "load_abc.cpp"
         HEAP8[$569>>0] = $568; //@line 544 "load_abc.cpp"
         $570 = ((($563)) + 15|0); //@line 545 "load_abc.cpp"
         HEAP8[$570>>0] = 0; //@line 545 "load_abc.cpp"
         __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($$pre1510,$563); //@line 1153 "load_abc.cpp"
         $571 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4094 "load_abc.cpp"
         $572 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 1928 "load_abc.cpp"
         $573 = ($572*12288)|0; //@line 339 "load_abc.cpp"
         $574 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 1929 "load_abc.cpp"
         $575 = ($574|0)==(0|0); //@line 1929 "load_abc.cpp"
         do {
          if ($575) {
           $$0$i154 = 0;$623 = 0;
          } else {
           $576 = ((($574)) + 4|0); //@line 1929 "load_abc.cpp"
           $577 = HEAPU8[$576>>0]|(HEAPU8[$576+1>>0]<<8)|(HEAPU8[$576+2>>0]<<16)|(HEAPU8[$576+3>>0]<<24); //@line 1929 "load_abc.cpp"
           $578 = ($577|0)==(0|0); //@line 1929 "load_abc.cpp"
           if ($578) {
            $4055 = 0;
           } else {
            $dt$03$i$i142 = 0;$e$02$i$i143 = $577;
            while(1) {
             $579 = ((($e$02$i$i143)) + 4|0); //@line 1931 "load_abc.cpp"
             $580 = HEAPU8[$579>>0]|(HEAPU8[$579+1>>0]<<8)|(HEAPU8[$579+2>>0]<<16)|(HEAPU8[$579+3>>0]<<24); //@line 1931 "load_abc.cpp"
             $581 = ($580>>>0)>($571>>>0); //@line 1931 "load_abc.cpp"
             if ($581) {
              $dt$0$lcssa$i$i146 = $dt$03$i$i142;
              break;
             }
             $582 = ((($e$02$i$i143)) + 8|0); //@line 1932 "load_abc.cpp"
             $583 = HEAP8[$582>>0]|0; //@line 1932 "load_abc.cpp"
             $584 = ($583<<24>>24)==(1); //@line 1932 "load_abc.cpp"
             do {
              if ($584) {
               $585 = ((($e$02$i$i143)) + 9|0); //@line 1932 "load_abc.cpp"
               $586 = HEAP8[$585>>0]|0; //@line 1932 "load_abc.cpp"
               $587 = ($586<<24>>24)==(80); //@line 1932 "load_abc.cpp"
               if (!($587)) {
                $dt$1$i$i145 = $dt$03$i$i142;
                break;
               }
               $588 = (($580) + ($dt$03$i$i142))|0; //@line 1933 "load_abc.cpp"
               $589 = (($588>>>0) % ($573>>>0))&-1; //@line 1934 "load_abc.cpp"
               $590 = ($589|0)==(0); //@line 1934 "load_abc.cpp"
               if ($590) {
                $dt$1$i$i145 = $dt$03$i$i142;
                break;
               }
               $591 = (($588) + ($573))|0; //@line 1935 "load_abc.cpp"
               $592 = (($591>>>0) % ($573>>>0))&-1; //@line 1937 "load_abc.cpp"
               $593 = (($dt$03$i$i142) + ($573))|0; //@line 1937 "load_abc.cpp"
               $594 = (($593) - ($592))|0; //@line 1938 "load_abc.cpp"
               $dt$1$i$i145 = $594;
              } else {
               $dt$1$i$i145 = $dt$03$i$i142;
              }
             } while(0);
             $595 = HEAPU8[$e$02$i$i143>>0]|(HEAPU8[$e$02$i$i143+1>>0]<<8)|(HEAPU8[$e$02$i$i143+2>>0]<<16)|(HEAPU8[$e$02$i$i143+3>>0]<<24); //@line 1931 "load_abc.cpp"
             $596 = ($595|0)==(0|0); //@line 1931 "load_abc.cpp"
             if ($596) {
              $dt$0$lcssa$i$i146 = $dt$1$i$i145;
              break;
             } else {
              $dt$03$i$i142 = $dt$1$i$i145;$e$02$i$i143 = $595;
             }
            }
            $597 = (($dt$0$lcssa$i$i146) + ($571))|0; //@line 1942 "load_abc.cpp"
            $phitmp$i147 = (($597>>>0) / 192)&-1; //@line 1942 "load_abc.cpp"
            $phitmp1$i148 = $phitmp$i147 >>> 6; //@line 1942 "load_abc.cpp"
            $598 = (($phitmp1$i148>>>0) / ($572>>>0))&-1; //@line 1947 "load_abc.cpp"
            if ($575) {
             $$0$i154 = 0;$623 = $598;
             break;
            } else {
             $4055 = $598;
            }
           }
           $599 = ((($574)) + 4|0); //@line 1929 "load_abc.cpp"
           $600 = HEAPU8[$599>>0]|(HEAPU8[$599+1>>0]<<8)|(HEAPU8[$599+2>>0]<<16)|(HEAPU8[$599+3>>0]<<24); //@line 1929 "load_abc.cpp"
           $601 = ($600|0)==(0|0); //@line 1929 "load_abc.cpp"
           if ($601) {
            $$0$i154 = 0;$623 = $4055;
            break;
           } else {
            $dt$03$i = 0;$e$02$i = $600;
           }
           while(1) {
            $602 = ((($e$02$i)) + 4|0); //@line 1931 "load_abc.cpp"
            $603 = HEAPU8[$602>>0]|(HEAPU8[$602+1>>0]<<8)|(HEAPU8[$602+2>>0]<<16)|(HEAPU8[$602+3>>0]<<24); //@line 1931 "load_abc.cpp"
            $604 = ($603>>>0)>($571>>>0); //@line 1931 "load_abc.cpp"
            if ($604) {
             $dt$0$lcssa$i = $dt$03$i;
             break;
            }
            $605 = ((($e$02$i)) + 8|0); //@line 1932 "load_abc.cpp"
            $606 = HEAP8[$605>>0]|0; //@line 1932 "load_abc.cpp"
            $607 = ($606<<24>>24)==(1); //@line 1932 "load_abc.cpp"
            do {
             if ($607) {
              $608 = ((($e$02$i)) + 9|0); //@line 1932 "load_abc.cpp"
              $609 = HEAP8[$608>>0]|0; //@line 1932 "load_abc.cpp"
              $610 = ($609<<24>>24)==(80); //@line 1932 "load_abc.cpp"
              if (!($610)) {
               $dt$1$i = $dt$03$i;
               break;
              }
              $611 = (($603) + ($dt$03$i))|0; //@line 1933 "load_abc.cpp"
              $612 = (($611>>>0) % ($573>>>0))&-1; //@line 1934 "load_abc.cpp"
              $613 = ($612|0)==(0); //@line 1934 "load_abc.cpp"
              if ($613) {
               $dt$1$i = $dt$03$i;
               break;
              }
              $614 = (($611) + ($573))|0; //@line 1935 "load_abc.cpp"
              $615 = (($614>>>0) % ($573>>>0))&-1; //@line 1937 "load_abc.cpp"
              $616 = (($dt$03$i) + ($573))|0; //@line 1937 "load_abc.cpp"
              $617 = (($616) - ($615))|0; //@line 1938 "load_abc.cpp"
              $dt$1$i = $617;
             } else {
              $dt$1$i = $dt$03$i;
             }
            } while(0);
            $618 = HEAPU8[$e$02$i>>0]|(HEAPU8[$e$02$i+1>>0]<<8)|(HEAPU8[$e$02$i+2>>0]<<16)|(HEAPU8[$e$02$i+3>>0]<<24); //@line 1931 "load_abc.cpp"
            $619 = ($618|0)==(0|0); //@line 1931 "load_abc.cpp"
            if ($619) {
             $dt$0$lcssa$i = $dt$1$i;
             break;
            } else {
             $dt$03$i = $dt$1$i;$e$02$i = $618;
            }
           }
           $620 = (($dt$0$lcssa$i) + ($571))|0; //@line 1942 "load_abc.cpp"
           $$0$i154 = $620;$623 = $4055;
          }
         } while(0);
         $621 = (($$0$i154>>>0) % ($573>>>0))&-1; //@line 4095 "load_abc.cpp"
         $not$831 = ($621|0)!=(0); //@line 4095 "load_abc.cpp"
         $622 = $not$831&1; //@line 4095 "load_abc.cpp"
         $$82 = (($622) + ($623))|0; //@line 4095 "load_abc.cpp"
         $624 = HEAP8[1278234]|0; //@line 4096 "load_abc.cpp"
         $625 = ($624<<24>>24)==(32); //@line 4096 "load_abc.cpp"
         $626 = $$82&255; //@line 4097 "load_abc.cpp"
         L171: do {
          if ($625) {
           HEAP8[$101>>0] = $626; //@line 4097 "load_abc.cpp"
           $627 = HEAP32[$abcparts>>2]|0; //@line 4098 "load_abc.cpp"
           $628 = ($627|0)==(0|0); //@line 4098 "load_abc.cpp"
           if ($628) {
            label = 216;
            break;
           } else {
            $t$41063 = 0;
           }
           while(1) {
            $629 = (($partpat) + ($t$41063<<1)|0); //@line 4100 "load_abc.cpp"
            $630 = HEAP8[$629>>0]|0; //@line 4100 "load_abc.cpp"
            $631 = (((($partpat) + ($t$41063<<1)|0)) + 1|0); //@line 4100 "load_abc.cpp"
            $632 = HEAP8[$631>>0]|0; //@line 4100 "load_abc.cpp"
            $633 = ($630&255)<($632&255); //@line 4100 "load_abc.cpp"
            if ($633) {
             $$pr = $627;
             label = 215;
             break L171;
            }
            $634 = (($t$41063) + 1)|0; //@line 4099 "load_abc.cpp"
            $635 = ($634>>>0)<(26); //@line 4099 "load_abc.cpp"
            if ($635) {
             $t$41063 = $634;
            } else {
             $$lcssa2127 = $634;
             break;
            }
           }
           $636 = ($$lcssa2127|0)==(26); //@line 4101 "load_abc.cpp"
           if (!($636)) {
            $$pr = $627;
            label = 215;
            break;
           }
           $637 = (_strlen($627)|0); //@line 323 "load_abc.cpp"
           $638 = (($637) + 27)|0; //@line 323 "load_abc.cpp"
           $639 = ($638>>>0)>(255); //@line 323 "load_abc.cpp"
           if ($639) {
            $641 = $627;
           } else {
            HEAP32[$vararg_buffer15>>2] = $627; //@line 324 "load_abc.cpp"
            (_sprintf($txt$i,526650,$vararg_buffer15)|0); //@line 324 "load_abc.cpp"
            HEAP32[$vararg_buffer18>>2] = $txt$i; //@line 328 "load_abc.cpp"
            (_fprintf($155,524747,$vararg_buffer18)|0); //@line 328 "load_abc.cpp"
            $$pre1512 = HEAP32[$abcparts>>2]|0; //@line 1820 "load_abc.cpp"
            $641 = $$pre1512;
           }
           $640 = ($641|0)==(0|0); //@line 1820 "load_abc.cpp"
           if (!($640)) {
            _free($641); //@line 1820 "load_abc.cpp"
           }
           HEAP32[$abcparts>>2] = 0; //@line 1822 "load_abc.cpp"
           label = 216;
          } else {
           $642 = $624 << 24 >> 24; //@line 4096 "load_abc.cpp"
           $643 = (($642) + -65)|0; //@line 4108 "load_abc.cpp"
           $644 = (((($partpat) + ($643<<1)|0)) + 1|0); //@line 4108 "load_abc.cpp"
           HEAP8[$644>>0] = $626; //@line 4108 "load_abc.cpp"
           $$pr$pre = HEAP32[$abcparts>>2]|0; //@line 4109 "load_abc.cpp"
           $$pr = $$pr$pre;
           label = 215;
          }
         } while(0);
         if ((label|0) == 215) {
          label = 0;
          $645 = ($$pr|0)==(0|0); //@line 4109 "load_abc.cpp"
          if ($645) {
           label = 216;
          } else {
           $646 = $$pr;
          }
         }
         if ((label|0) == 216) {
          label = 0;
          __ZL17abc_song_to_partsP10_ABCHANDLEPPcPA2_h($1,$abcparts,$partpat); //@line 4109 "load_abc.cpp"
          $$pre1514 = HEAP32[$abcparts>>2]|0; //@line 4110 "load_abc.cpp"
          $646 = $$pre1514;
         }
         $647 = (__ZL24abc_partpat_to_orderlistPA2_hPKcP10_ABCHANDLEPPhi($partpat,$646,$orderlist,$orderlen$11090)|0); //@line 4110 "load_abc.cpp"
         $abcstate$5 = $536;$orderlen$2 = $647;$p$51550 = $p$1$lcssa;
         label = 218;
        }
       } else {
        $abcstate$5 = 3;$orderlen$2 = $orderlen$11090;$p$51550 = $p$1$lcssa;
        label = 218;
       }
       break;
      }
      default: {
       $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = $abcstate$21092;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$1$lcssa;$tpd$2 = $tpd$11121;
      }
      }
      do {
       if ((label|0) == 43) {
        label = 0;
        $191 = (_strncmp($p$1$lcssa,525008,2)|0); //@line 3918 "load_abc.cpp"
        $192 = ($191|0)==(0); //@line 3918 "load_abc.cpp"
        if ($192) {
         dest=$157; stop=dest+32|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 3923 "load_abc.cpp"
         $193 = ((($p$1$lcssa)) + 2|0); //@line 3925 "load_abc.cpp"
         $p$2 = $193;
         while(1) {
          $194 = HEAP8[$p$2>>0]|0; //@line 3925 "load_abc.cpp"
          $195 = $194 << 24 >> 24; //@line 3925 "load_abc.cpp"
          $196 = (_isspace($195)|0); //@line 3925 "load_abc.cpp"
          $197 = ($196|0)==(0); //@line 3925 "load_abc.cpp"
          $198 = ((($p$2)) + 1|0); //@line 3925 "load_abc.cpp"
          if ($197) {
           $p$2$lcssa = $p$2;
           break;
          } else {
           $p$2 = $198;
          }
         }
         $199 = (_atoi($p$2$lcssa)|0); //@line 3926 "load_abc.cpp"
         HEAP8[$40>>0]=0&255;HEAP8[$40+1>>0]=(0>>8)&255;HEAP8[$40+2>>0]=(0>>16)&255;HEAP8[$40+3>>0]=0>>24; //@line 3928 "load_abc.cpp"
         HEAP8[$44>>0]=71&255;HEAP8[$44+1>>0]=(71>>8)&255;HEAP8[$44+2>>0]=(71>>16)&255;HEAP8[$44+3>>0]=71>>24; //@line 3929 "load_abc.cpp"
         HEAP8[$45>>0]=45&255;HEAP8[$45+1>>0]=(45>>8)&255;HEAP8[$45+2>>0]=(45>>16)&255;HEAP8[$45+3>>0]=45>>24; //@line 3930 "load_abc.cpp"
         HEAP8[$46>>0]=33&255;HEAP8[$46+1>>0]=(33>>8)&255;HEAP8[$46+2>>0]=(33>>16)&255;HEAP8[$46+3>>0]=33>>24; //@line 3931 "load_abc.cpp"
         HEAP8[$47>>0]=80&255;HEAP8[$47+1>>0]=(80>>8)&255;HEAP8[$47+2>>0]=(80>>16)&255;HEAP8[$47+3>>0]=80>>24; //@line 3932 "load_abc.cpp"
         HEAP8[$48>>0]=80&255;HEAP8[$48+1>>0]=(80>>8)&255;HEAP8[$48+2>>0]=(80>>16)&255;HEAP8[$48+3>>0]=80>>24; //@line 3933 "load_abc.cpp"
         $ttp$01072 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 3934 "load_abc.cpp"
         $200 = ($ttp$01072|0)==(0|0); //@line 3934 "load_abc.cpp"
         if (!($200)) {
          $ttp$01073 = $ttp$01072;
          while(1) {
           $201 = ((($ttp$01073)) + 36|0); //@line 3935 "load_abc.cpp"
           HEAP8[$201>>0] = 0; //@line 3935 "load_abc.cpp"
           $202 = ((($ttp$01073)) + 12|0); //@line 3936 "load_abc.cpp"
           HEAP8[$202>>0]=0&255;HEAP8[$202+1>>0]=(0>>8)&255;HEAP8[$202+2>>0]=(0>>16)&255;HEAP8[$202+3>>0]=0>>24; //@line 3936 "load_abc.cpp"
           $ttp$0 = HEAPU8[$ttp$01073>>0]|(HEAPU8[$ttp$01073+1>>0]<<8)|(HEAPU8[$ttp$01073+2>>0]<<16)|(HEAPU8[$ttp$01073+3>>0]<<24); //@line 3934 "load_abc.cpp"
           $203 = ($ttp$0|0)==(0|0); //@line 3934 "load_abc.cpp"
           if ($203) {
            break;
           } else {
            $ttp$01073 = $ttp$0;
           }
          }
         }
         HEAP8[$41>>0]=0&255;HEAP8[$41+1>>0]=(0>>8)&255;HEAP8[$41+2>>0]=(0>>16)&255;HEAP8[$41+3>>0]=0>>24; //@line 3938 "load_abc.cpp"
         HEAP8[$42>>0]=0&255;HEAP8[$42+1>>0]=(0>>8)&255;HEAP8[$42+2>>0]=(0>>16)&255;HEAP8[$42+3>>0]=0>>24; //@line 3939 "load_abc.cpp"
         HEAP8[$43>>0]=0&255;HEAP8[$43+1>>0]=(0>>8)&255;HEAP8[$43+2>>0]=(0>>16)&255;HEAP8[$43+3>>0]=0>>24; //@line 3940 "load_abc.cpp"
         HEAP32[299860] = 0; //@line 3941 "load_abc.cpp"
         $204 = HEAP32[$abcparts>>2]|0; //@line 1820 "load_abc.cpp"
         $205 = ($204|0)==(0|0); //@line 1820 "load_abc.cpp"
         if (!($205)) {
          _free($204); //@line 1820 "load_abc.cpp"
         }
         HEAP32[$abcparts>>2] = 0; //@line 1822 "load_abc.cpp"
         HEAP8[$37>>0]=0&255;HEAP8[$37+1>>0]=(0>>8)&255;HEAP8[$37+2>>0]=(0>>16)&255;HEAP8[$37+3>>0]=0>>24; //@line 3944 "load_abc.cpp"
         HEAP8[$39>>0]=1&255;HEAP8[$39+1>>0]=(1>>8)&255;HEAP8[$39+2>>0]=(1>>16)&255;HEAP8[$39+3>>0]=1>>24; //@line 3946 "load_abc.cpp"
         HEAP8[$38>>0]=0&255;HEAP8[$38+1>>0]=(0>>8)&255;HEAP8[$38+2>>0]=(0>>16)&255;HEAP8[$38+3>>0]=0>>24; //@line 3947 "load_abc.cpp"
         $206 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 3948 "load_abc.cpp"
         HEAP32[299861] = $206; //@line 3948 "load_abc.cpp"
         HEAP8[$102>>0] = 127; //@line 3507 "load_abc.cpp"
         HEAP8[$103>>0] = 125; //@line 3508 "load_abc.cpp"
         HEAP8[$104>>0] = 110; //@line 3509 "load_abc.cpp"
         HEAP8[$105>>0] = 1; //@line 3510 "load_abc.cpp"
         $$1$i111 = 1282466;
         while(1) {
          $207 = HEAP8[$$1$i111>>0]|0; //@line 3512 "load_abc.cpp"
          $208 = $207 << 24 >> 24; //@line 3512 "load_abc.cpp"
          $209 = (_isspace($208)|0); //@line 3512 "load_abc.cpp"
          $210 = ($209|0)==(0); //@line 3512 "load_abc.cpp"
          $211 = ((($$1$i111)) + 1|0); //@line 3512 "load_abc.cpp"
          if ($210) {
           $$1$i111$lcssa = $$1$i111;
           break;
          } else {
           $$1$i111 = $211;
          }
         }
         $212 = HEAP8[$$1$i111$lcssa>>0]|0; //@line 3513 "load_abc.cpp"
         $213 = ($212<<24>>24)==(0); //@line 3513 "load_abc.cpp"
         if ($213) {
          $$1$i111$1 = $$1$i111$lcssa;
         } else {
          $214 = $212 << 24 >> 24; //@line 1625 "load_abc.cpp"
          $isdigittmp1$i$i112 = (($214) + -48)|0; //@line 1625 "load_abc.cpp"
          $isdigit2$i$i113 = ($isdigittmp1$i$i112>>>0)<(10); //@line 1625 "load_abc.cpp"
          do {
           if ($isdigit2$i$i113) {
            $217 = $214;$h$04$i$i114 = 0;$i$03$i$i115 = 0;
            while(1) {
             $215 = ($h$04$i$i114*10)|0; //@line 1626 "load_abc.cpp"
             $216 = (($217) + -48)|0; //@line 1626 "load_abc.cpp"
             $218 = (($216) + ($215))|0; //@line 1626 "load_abc.cpp"
             $219 = (($i$03$i$i115) + 1)|0; //@line 1627 "load_abc.cpp"
             $220 = (($$1$i111$lcssa) + ($219)|0); //@line 1625 "load_abc.cpp"
             $221 = HEAP8[$220>>0]|0; //@line 1625 "load_abc.cpp"
             $222 = $221 << 24 >> 24; //@line 1625 "load_abc.cpp"
             $isdigittmp$i$i116 = (($222) + -48)|0; //@line 1625 "load_abc.cpp"
             $isdigit$i$i117 = ($isdigittmp$i$i116>>>0)<(10); //@line 1625 "load_abc.cpp"
             if ($isdigit$i$i117) {
              $217 = $222;$h$04$i$i114 = $218;$i$03$i$i115 = $219;
             } else {
              $$lcssa2133 = $218;$$lcssa2134 = $219;$$lcssa2135 = $220;
              break;
             }
            }
            $223 = ($$lcssa2134|0)==(0); //@line 1629 "load_abc.cpp"
            $$h$0$i$i119 = $223 ? 1 : $$lcssa2133; //@line 1630 "load_abc.cpp"
            $224 = ($$h$0$i$i119|0)<(0); //@line 3515 "load_abc.cpp"
            if ($224) {
             $227 = 0;$4053 = $$lcssa2135;
             break;
            }
            $225 = ($$h$0$i$i119|0)>(127); //@line 3516 "load_abc.cpp"
            $$$h$0$i$i121 = $225 ? 127 : $$h$0$i$i119; //@line 3516 "load_abc.cpp"
            $227 = $$$h$0$i$i121;$4053 = $$lcssa2135;
           } else {
            $227 = 1;$4053 = $$1$i111$lcssa;
           }
          } while(0);
          $226 = $227&255; //@line 3517 "load_abc.cpp"
          HEAP8[$167>>0] = $226; //@line 3517 "load_abc.cpp"
          $$1$i111$1 = $4053;
         }
         while(1) {
          $3909 = HEAP8[$$1$i111$1>>0]|0; //@line 3512 "load_abc.cpp"
          $3910 = $3909 << 24 >> 24; //@line 3512 "load_abc.cpp"
          $3911 = (_isspace($3910)|0); //@line 3512 "load_abc.cpp"
          $3912 = ($3911|0)==(0); //@line 3512 "load_abc.cpp"
          $3913 = ((($$1$i111$1)) + 1|0); //@line 3512 "load_abc.cpp"
          if ($3912) {
           $$1$i111$1$lcssa = $$1$i111$1;
           break;
          } else {
           $$1$i111$1 = $3913;
          }
         }
         $3914 = HEAP8[$$1$i111$1$lcssa>>0]|0; //@line 3513 "load_abc.cpp"
         $3915 = ($3914<<24>>24)==(0); //@line 3513 "load_abc.cpp"
         if ($3915) {
          $$1$i111$2 = $$1$i111$1$lcssa;
         } else {
          $3916 = $3914 << 24 >> 24; //@line 1625 "load_abc.cpp"
          $isdigittmp1$i$i112$1 = (($3916) + -48)|0; //@line 1625 "load_abc.cpp"
          $isdigit2$i$i113$1 = ($isdigittmp1$i$i112$1>>>0)<(10); //@line 1625 "load_abc.cpp"
          do {
           if ($isdigit2$i$i113$1) {
            $3919 = $3916;$h$04$i$i114$1 = 0;$i$03$i$i115$1 = 0;
            while(1) {
             $3917 = ($h$04$i$i114$1*10)|0; //@line 1626 "load_abc.cpp"
             $3918 = (($3919) + -48)|0; //@line 1626 "load_abc.cpp"
             $3920 = (($3918) + ($3917))|0; //@line 1626 "load_abc.cpp"
             $3921 = (($i$03$i$i115$1) + 1)|0; //@line 1627 "load_abc.cpp"
             $3922 = (($$1$i111$1$lcssa) + ($3921)|0); //@line 1625 "load_abc.cpp"
             $3923 = HEAP8[$3922>>0]|0; //@line 1625 "load_abc.cpp"
             $3924 = $3923 << 24 >> 24; //@line 1625 "load_abc.cpp"
             $isdigittmp$i$i116$1 = (($3924) + -48)|0; //@line 1625 "load_abc.cpp"
             $isdigit$i$i117$1 = ($isdigittmp$i$i116$1>>>0)<(10); //@line 1625 "load_abc.cpp"
             if ($isdigit$i$i117$1) {
              $3919 = $3924;$h$04$i$i114$1 = $3920;$i$03$i$i115$1 = $3921;
             } else {
              $$lcssa2136 = $3920;$$lcssa2137 = $3921;$$lcssa2138 = $3922;
              break;
             }
            }
            $3925 = ($$lcssa2137|0)==(0); //@line 1629 "load_abc.cpp"
            $$h$0$i$i119$1 = $3925 ? 1 : $$lcssa2136; //@line 1630 "load_abc.cpp"
            $3926 = ($$h$0$i$i119$1|0)<(0); //@line 3515 "load_abc.cpp"
            if ($3926) {
             $3929 = 0;$4070 = $$lcssa2138;
             break;
            }
            $3927 = ($$h$0$i$i119$1|0)>(127); //@line 3516 "load_abc.cpp"
            $$$h$0$i$i121$1 = $3927 ? 127 : $$h$0$i$i119$1; //@line 3516 "load_abc.cpp"
            $3929 = $$$h$0$i$i121$1;$4070 = $$lcssa2138;
           } else {
            $3929 = 1;$4070 = $$1$i111$1$lcssa;
           }
          } while(0);
          $3928 = $3929&255; //@line 3517 "load_abc.cpp"
          HEAP8[$168>>0] = $3928; //@line 3517 "load_abc.cpp"
          $$1$i111$2 = $4070;
         }
         while(1) {
          $3930 = HEAP8[$$1$i111$2>>0]|0; //@line 3512 "load_abc.cpp"
          $3931 = $3930 << 24 >> 24; //@line 3512 "load_abc.cpp"
          $3932 = (_isspace($3931)|0); //@line 3512 "load_abc.cpp"
          $3933 = ($3932|0)==(0); //@line 3512 "load_abc.cpp"
          $3934 = ((($$1$i111$2)) + 1|0); //@line 3512 "load_abc.cpp"
          if ($3933) {
           $$1$i111$2$lcssa = $$1$i111$2;
           break;
          } else {
           $$1$i111$2 = $3934;
          }
         }
         $3935 = HEAP8[$$1$i111$2$lcssa>>0]|0; //@line 3513 "load_abc.cpp"
         $3936 = ($3935<<24>>24)==(0); //@line 3513 "load_abc.cpp"
         if ($3936) {
          $$1$i111$3 = $$1$i111$2$lcssa;
         } else {
          $3937 = $3935 << 24 >> 24; //@line 1625 "load_abc.cpp"
          $isdigittmp1$i$i112$2 = (($3937) + -48)|0; //@line 1625 "load_abc.cpp"
          $isdigit2$i$i113$2 = ($isdigittmp1$i$i112$2>>>0)<(10); //@line 1625 "load_abc.cpp"
          do {
           if ($isdigit2$i$i113$2) {
            $3940 = $3937;$h$04$i$i114$2 = 0;$i$03$i$i115$2 = 0;
            while(1) {
             $3938 = ($h$04$i$i114$2*10)|0; //@line 1626 "load_abc.cpp"
             $3939 = (($3940) + -48)|0; //@line 1626 "load_abc.cpp"
             $3941 = (($3939) + ($3938))|0; //@line 1626 "load_abc.cpp"
             $3942 = (($i$03$i$i115$2) + 1)|0; //@line 1627 "load_abc.cpp"
             $3943 = (($$1$i111$2$lcssa) + ($3942)|0); //@line 1625 "load_abc.cpp"
             $3944 = HEAP8[$3943>>0]|0; //@line 1625 "load_abc.cpp"
             $3945 = $3944 << 24 >> 24; //@line 1625 "load_abc.cpp"
             $isdigittmp$i$i116$2 = (($3945) + -48)|0; //@line 1625 "load_abc.cpp"
             $isdigit$i$i117$2 = ($isdigittmp$i$i116$2>>>0)<(10); //@line 1625 "load_abc.cpp"
             if ($isdigit$i$i117$2) {
              $3940 = $3945;$h$04$i$i114$2 = $3941;$i$03$i$i115$2 = $3942;
             } else {
              $$lcssa2139 = $3941;$$lcssa2140 = $3942;$$lcssa2141 = $3943;
              break;
             }
            }
            $3946 = ($$lcssa2140|0)==(0); //@line 1629 "load_abc.cpp"
            $$h$0$i$i119$2 = $3946 ? 1 : $$lcssa2139; //@line 1630 "load_abc.cpp"
            $3947 = ($$h$0$i$i119$2|0)<(0); //@line 3515 "load_abc.cpp"
            if ($3947) {
             $3950 = 0;$4071 = $$lcssa2141;
             break;
            }
            $3948 = ($$h$0$i$i119$2|0)>(127); //@line 3516 "load_abc.cpp"
            $$$h$0$i$i121$2 = $3948 ? 127 : $$h$0$i$i119$2; //@line 3516 "load_abc.cpp"
            $3950 = $$$h$0$i$i121$2;$4071 = $$lcssa2141;
           } else {
            $3950 = 1;$4071 = $$1$i111$2$lcssa;
           }
          } while(0);
          $3949 = $3950&255; //@line 3517 "load_abc.cpp"
          HEAP8[$169>>0] = $3949; //@line 3517 "load_abc.cpp"
          $$1$i111$3 = $4071;
         }
         while(1) {
          $3951 = HEAP8[$$1$i111$3>>0]|0; //@line 3512 "load_abc.cpp"
          $3952 = $3951 << 24 >> 24; //@line 3512 "load_abc.cpp"
          $3953 = (_isspace($3952)|0); //@line 3512 "load_abc.cpp"
          $3954 = ($3953|0)==(0); //@line 3512 "load_abc.cpp"
          $3955 = ((($$1$i111$3)) + 1|0); //@line 3512 "load_abc.cpp"
          if ($3954) {
           $$1$i111$3$lcssa = $$1$i111$3;
           break;
          } else {
           $$1$i111$3 = $3955;
          }
         }
         $3956 = HEAP8[$$1$i111$3$lcssa>>0]|0; //@line 3513 "load_abc.cpp"
         $3957 = ($3956<<24>>24)==(0); //@line 3513 "load_abc.cpp"
         if ($3957) {
          $$pre1521 = HEAP8[$105>>0]|0; //@line 3520 "load_abc.cpp"
          $3973 = $$pre1521;
         } else {
          $3958 = $3956 << 24 >> 24; //@line 1625 "load_abc.cpp"
          $isdigittmp1$i$i112$3 = (($3958) + -48)|0; //@line 1625 "load_abc.cpp"
          $isdigit2$i$i113$3 = ($isdigittmp1$i$i112$3>>>0)<(10); //@line 1625 "load_abc.cpp"
          do {
           if ($isdigit2$i$i113$3) {
            $3961 = $3958;$h$04$i$i114$3 = 0;$i$03$i$i115$3 = 0;
            while(1) {
             $3959 = ($h$04$i$i114$3*10)|0; //@line 1626 "load_abc.cpp"
             $3960 = (($3961) + -48)|0; //@line 1626 "load_abc.cpp"
             $3962 = (($3960) + ($3959))|0; //@line 1626 "load_abc.cpp"
             $3963 = (($i$03$i$i115$3) + 1)|0; //@line 1627 "load_abc.cpp"
             $3964 = (($$1$i111$3$lcssa) + ($3963)|0); //@line 1625 "load_abc.cpp"
             $3965 = HEAP8[$3964>>0]|0; //@line 1625 "load_abc.cpp"
             $3966 = $3965 << 24 >> 24; //@line 1625 "load_abc.cpp"
             $isdigittmp$i$i116$3 = (($3966) + -48)|0; //@line 1625 "load_abc.cpp"
             $isdigit$i$i117$3 = ($isdigittmp$i$i116$3>>>0)<(10); //@line 1625 "load_abc.cpp"
             if ($isdigit$i$i117$3) {
              $3961 = $3966;$h$04$i$i114$3 = $3962;$i$03$i$i115$3 = $3963;
             } else {
              $$lcssa2142 = $3962;$$lcssa2143 = $3963;
              break;
             }
            }
            $3967 = ($$lcssa2143|0)==(0); //@line 1629 "load_abc.cpp"
            $$h$0$i$i119$3 = $3967 ? 1 : $$lcssa2142; //@line 1630 "load_abc.cpp"
            $3968 = ($$h$0$i$i119$3|0)<(0); //@line 3515 "load_abc.cpp"
            if ($3968) {
             $3971 = 0;
             break;
            }
            $3969 = ($$h$0$i$i119$3|0)>(127); //@line 3516 "load_abc.cpp"
            $$$h$0$i$i121$3 = $3969 ? 127 : $$h$0$i$i119$3; //@line 3516 "load_abc.cpp"
            $3971 = $$$h$0$i$i121$3;
           } else {
            $3971 = 1;
           }
          } while(0);
          $3970 = $3971&255; //@line 3517 "load_abc.cpp"
          HEAP8[$170>>0] = $3970; //@line 3517 "load_abc.cpp"
          $3973 = $3970;
         }
         $3972 = ($3973<<24>>24)==(0); //@line 3520 "load_abc.cpp"
         if ($3972) {
          HEAP8[$105>>0] = 1; //@line 3520 "load_abc.cpp"
          $$0$i126 = 1282466;
         } else {
          $$0$i126 = 1282466;
         }
         while(1) {
          $228 = HEAP8[$$0$i126>>0]|0; //@line 3535 "load_abc.cpp"
          $229 = $228 << 24 >> 24; //@line 3535 "load_abc.cpp"
          $230 = (_isspace($229)|0); //@line 3535 "load_abc.cpp"
          $231 = ($230|0)==(0); //@line 3535 "load_abc.cpp"
          $232 = ((($$0$i126)) + 1|0); //@line 3535 "load_abc.cpp"
          if ($231) {
           $$0$i126$lcssa = $$0$i126;
           break;
          } else {
           $$0$i126 = $232;
          }
         }
         $233 = HEAPU8[$3>>0]|(HEAPU8[$3+1>>0]<<8)|(HEAPU8[$3+2>>0]<<16)|(HEAPU8[$3+3>>0]<<24); //@line 3536 "load_abc.cpp"
         $234 = ($233|0)==(0|0); //@line 3536 "load_abc.cpp"
         if (!($234)) {
          _free($233); //@line 3536 "load_abc.cpp"
         }
         $strlenfirst$i127 = HEAP8[$$0$i126$lcssa>>0]|0; //@line 3537 "load_abc.cpp"
         $235 = ($strlenfirst$i127<<24>>24)==(0); //@line 3537 "load_abc.cpp"
         if ($235) {
          $storemerge$i128 = 0;
         } else {
          $236 = (___strdup($$0$i126$lcssa)|0); //@line 3538 "load_abc.cpp"
          $storemerge$i128 = $236;
         }
         HEAP8[$3>>0]=$storemerge$i128&255;HEAP8[$3+1>>0]=($storemerge$i128>>8)&255;HEAP8[$3+2>>0]=($storemerge$i128>>16)&255;HEAP8[$3+3>>0]=$storemerge$i128>>24; //@line 3538 "load_abc.cpp"
         HEAP8[$6>>0] = 0; //@line 3951 "load_abc.cpp"
         $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = 0;$abchornpipe$3 = 0;$abckey$3 = $abckey$11101;$abcnolegato$3 = 1;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = 2;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$3;$abcxnumber$2 = $199;$bartime$3 = $bartime$11119;$ktempo$2 = 0;$orderlen$3 = $orderlen$11090;$p$7 = $p$2$lcssa;$tpd$2 = $tpd$11121;
        } else {
         $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = $abcstate$21092;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$3;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$1$lcssa;$tpd$2 = $tpd$11121;
        }
       }
       else if ((label|0) == 88) {
        label = 0;
        $298 = (_strncmp($p$1$lcssa,525124,2)|0); //@line 3968 "load_abc.cpp"
        $299 = ($298|0)==(0); //@line 3968 "load_abc.cpp"
        if ($299) {
         $300 = ((($p$1$lcssa)) + 2|0); //@line 3969 "load_abc.cpp"
         HEAP32[$vararg_buffer4>>2] = $snotelen; //@line 3969 "load_abc.cpp"
         $vararg_ptr7 = ((($vararg_buffer4)) + 4|0); //@line 3969 "load_abc.cpp"
         HEAP32[$vararg_ptr7>>2] = $snotediv; //@line 3969 "load_abc.cpp"
         (_sscanf($300,525127,$vararg_buffer4)|0); //@line 3969 "load_abc.cpp"
         $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = 2;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$1$lcssa;$tpd$2 = $tpd$11121;
         break;
        }
        $301 = (_strncmp($p$1$lcssa,525136,2)|0); //@line 3972 "load_abc.cpp"
        $302 = ($301|0)==(0); //@line 3972 "load_abc.cpp"
        if ($302) {
         $303 = ((($p$1$lcssa)) + 2|0); //@line 3973 "load_abc.cpp"
         $304 = (_strncmp($303,525139,4)|0); //@line 3312 "load_abc.cpp"
         $305 = ($304|0)==(0); //@line 3312 "load_abc.cpp"
         if ($305) {
          HEAP32[$mnotelen>>2] = 1; //@line 3313 "load_abc.cpp"
          HEAP32[$mnotediv>>2] = 1; //@line 3314 "load_abc.cpp"
          $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = 2;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$1$lcssa;$tpd$2 = $tpd$11121;
          break;
         }
         $306 = (_strncmp($303,525144,2)|0); //@line 3317 "load_abc.cpp"
         $307 = ($306|0)==(0); //@line 3317 "load_abc.cpp"
         if ($307) {
          HEAP32[$mnotelen>>2] = 2; //@line 3318 "load_abc.cpp"
          HEAP32[$mnotediv>>2] = 2; //@line 3319 "load_abc.cpp"
          $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = 2;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$1$lcssa;$tpd$2 = $tpd$11121;
          break;
         }
         $308 = HEAP8[$303>>0]|0; //@line 3322 "load_abc.cpp"
         $309 = ($308<<24>>24)==(67); //@line 3322 "load_abc.cpp"
         if ($309) {
          HEAP32[$mnotelen>>2] = 4; //@line 3323 "load_abc.cpp"
          HEAP32[$mnotediv>>2] = 4; //@line 3324 "load_abc.cpp"
          $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = 2;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$1$lcssa;$tpd$2 = $tpd$11121;
          break;
         } else {
          $310 = (__ZL11abc_getexprPKcPi($303,$mnotelen)|0); //@line 3327 "load_abc.cpp"
          $$sum833 = (($310) + 2)|0; //@line 3327 "load_abc.cpp"
          $311 = (($p$1$lcssa) + ($$sum833)|0); //@line 3327 "load_abc.cpp"
          HEAP32[$vararg_buffer8>>2] = $mnotediv; //@line 3328 "load_abc.cpp"
          (_sscanf($311,525147,$vararg_buffer8)|0); //@line 3328 "load_abc.cpp"
          $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = 2;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$1$lcssa;$tpd$2 = $tpd$11121;
          break;
         }
        }
        $312 = (_strncmp($p$1$lcssa,525153,2)|0); //@line 3976 "load_abc.cpp"
        $313 = ($312|0)==(0); //@line 3976 "load_abc.cpp"
        if ($313) {
         $314 = ((($p$1$lcssa)) + 2|0); //@line 3977 "load_abc.cpp"
         __ZL13abc_set_partsPPcS_($abcparts,$314); //@line 3977 "load_abc.cpp"
         $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = 2;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$1$lcssa;$tpd$2 = $tpd$11121;
         break;
        }
        $315 = (_strncmp($p$1$lcssa,525350,2)|0); //@line 3980 "load_abc.cpp"
        $316 = ($315|0)==(0); //@line 3980 "load_abc.cpp"
        if ($316) {
         $317 = ((($p$1$lcssa)) + 2|0); //@line 3981 "load_abc.cpp"
         $318 = (__ZL17abc_extract_tempoPKci($317,0)|0); //@line 3981 "load_abc.cpp"
         $319 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 3983 "load_abc.cpp"
         $320 = ($319|0)==(0|0); //@line 3983 "load_abc.cpp"
         if (!($320)) {
          $321 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 3985 "load_abc.cpp"
          $322 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$322>>0]=0&255;HEAP8[$322+1>>0]=(0>>8)&255;HEAP8[$322+2>>0]=(0>>16)&255;HEAP8[$322+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $323 = ((($322)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$323>>0]=$321&255;HEAP8[$323+1>>0]=($321>>8)&255;HEAP8[$323+2>>0]=($321>>16)&255;HEAP8[$323+3>>0]=$321>>24; //@line 541 "load_abc.cpp"
          $324 = ((($322)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$324>>0] = 1; //@line 543 "load_abc.cpp"
          $325 = ((($322)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$325>>0] = 80; //@line 543 "load_abc.cpp"
          $326 = ((($322)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$326>>0]=0&255;HEAP8[$326+1>>0]=(0>>8)&255;HEAP8[$326+2>>0]=(0>>16)&255;HEAP8[$326+3>>0]=0>>24; //@line 543 "load_abc.cpp"
          $327 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $328 = ((($322)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$328>>0] = $327; //@line 544 "load_abc.cpp"
          $329 = ((($322)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$329>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($319,$322); //@line 1153 "load_abc.cpp"
          $330 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 3986 "load_abc.cpp"
          $331 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 3986 "load_abc.cpp"
          $332 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$332>>0]=0&255;HEAP8[$332+1>>0]=(0>>8)&255;HEAP8[$332+2>>0]=(0>>16)&255;HEAP8[$332+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $333 = ((($332)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$333>>0]=$331&255;HEAP8[$333+1>>0]=($331>>8)&255;HEAP8[$333+2>>0]=($331>>16)&255;HEAP8[$333+3>>0]=$331>>24; //@line 541 "load_abc.cpp"
          $334 = ((($332)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$334>>0] = 1; //@line 543 "load_abc.cpp"
          $335 = ((($332)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$335>>0] = 84; //@line 543 "load_abc.cpp"
          $336 = ((($332)) + 10|0); //@line 543 "load_abc.cpp"
          $337 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $338 = ((($332)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$338>>0] = $337; //@line 544 "load_abc.cpp"
          $339 = ((($332)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$339>>0] = 0; //@line 545 "load_abc.cpp"
          HEAP8[$336>>0]=$318&255;HEAP8[$336+1>>0]=($318>>8)&255;HEAP8[$336+2>>0]=($318>>16)&255;HEAP8[$336+3>>0]=$318>>24; //@line 1164 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($330,$332); //@line 1165 "load_abc.cpp"
         }
         $340 = HEAP32[$30>>2]|0; //@line 3988 "load_abc.cpp"
         $341 = ($340|0)==(0); //@line 3988 "load_abc.cpp"
         if (!($341)) {
          $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = 2;$abctempo$4 = $318;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = 1;$orderlen$3 = $orderlen$11090;$p$7 = $p$1$lcssa;$tpd$2 = $tpd$11121;
          break;
         }
         HEAP32[$30>>2] = $318; //@line 3988 "load_abc.cpp"
         $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = 2;$abctempo$4 = $318;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = 1;$orderlen$3 = $orderlen$11090;$p$7 = $p$1$lcssa;$tpd$2 = $tpd$11121;
         break;
        }
        $342 = (_strncmp($p$1$lcssa,525412,2)|0); //@line 3991 "load_abc.cpp"
        $343 = ($342|0)==(0); //@line 3991 "load_abc.cpp"
        if ($343) {
         $344 = (_strchr($p$1$lcssa,37)|0); //@line 3993 "load_abc.cpp"
         $345 = ($344|0)==(0|0); //@line 3993 "load_abc.cpp"
         if (!($345)) {
          HEAP8[$344>>0] = 0; //@line 3993 "load_abc.cpp"
         }
         $346 = (_strlen($p$1$lcssa)|0); //@line 3994 "load_abc.cpp"
         $t$01068 = (($346) + -1)|0; //@line 3994 "load_abc.cpp"
         $347 = (($p$1$lcssa) + ($t$01068)|0); //@line 3994 "load_abc.cpp"
         $348 = HEAP8[$347>>0]|0; //@line 3994 "load_abc.cpp"
         $349 = $348 << 24 >> 24; //@line 3994 "load_abc.cpp"
         $350 = (_isspace($349)|0); //@line 3994 "load_abc.cpp"
         $351 = ($350|0)==(0); //@line 3994 "load_abc.cpp"
         if ($351) {
          $t$1 = 2;
         } else {
          $352 = $347;$t$01069 = $t$01068;
          while(1) {
           HEAP8[$352>>0] = 0; //@line 3995 "load_abc.cpp"
           $t$0 = (($t$01069) + -1)|0; //@line 3994 "load_abc.cpp"
           $353 = (($p$1$lcssa) + ($t$0)|0); //@line 3994 "load_abc.cpp"
           $354 = HEAP8[$353>>0]|0; //@line 3994 "load_abc.cpp"
           $355 = $354 << 24 >> 24; //@line 3994 "load_abc.cpp"
           $356 = (_isspace($355)|0); //@line 3994 "load_abc.cpp"
           $357 = ($356|0)==(0); //@line 3994 "load_abc.cpp"
           if ($357) {
            $t$1 = 2;
            break;
           } else {
            $352 = $353;$t$01069 = $t$0;
           }
          }
         }
         while(1) {
          $358 = (($p$1$lcssa) + ($t$1)|0); //@line 3996 "load_abc.cpp"
          $359 = HEAP8[$358>>0]|0; //@line 3996 "load_abc.cpp"
          $360 = $359 << 24 >> 24; //@line 3996 "load_abc.cpp"
          $361 = (_isspace($360)|0); //@line 3996 "load_abc.cpp"
          $362 = ($361|0)==(0); //@line 3996 "load_abc.cpp"
          $363 = (($t$1) + 1)|0; //@line 3996 "load_abc.cpp"
          if ($362) {
           $$lcssa2132 = $358;$t$1$lcssa = $t$1;
           break;
          } else {
           $t$1 = $363;
          }
         }
         (_strcpy($buf,$157)|0); //@line 4003 "load_abc.cpp"
         $364 = (_strlen($buf)|0); //@line 4005 "load_abc.cpp"
         $365 = (_strlen($$lcssa2132)|0); //@line 4005 "load_abc.cpp"
         $366 = (($365) + ($364))|0; //@line 4005 "load_abc.cpp"
         $367 = ($366>>>0)>(199); //@line 4005 "load_abc.cpp"
         if ($367) {
          $368 = (($t$1$lcssa) + 199)|0; //@line 4005 "load_abc.cpp"
          $369 = (($368) - ($364))|0; //@line 4005 "load_abc.cpp"
          $370 = (($p$1$lcssa) + ($369)|0); //@line 4005 "load_abc.cpp"
          HEAP8[$370>>0] = 0; //@line 4005 "load_abc.cpp"
         }
         $strlenfirst72 = HEAP8[$buf>>0]|0; //@line 4006 "load_abc.cpp"
         $371 = ($strlenfirst72<<24>>24)==(0); //@line 4006 "load_abc.cpp"
         if (!($371)) {
          $strlen73 = (_strlen($buf)|0); //@line 4006 "load_abc.cpp"
          $endptr74 = (($buf) + ($strlen73)|0); //@line 4006 "load_abc.cpp"
          HEAP8[$endptr74>>0]=32&255;HEAP8[$endptr74+1>>0]=32>>8; //@line 4006 "load_abc.cpp"
         }
         (_strcat($buf,$$lcssa2132)|0); //@line 4007 "load_abc.cpp"
         $372 = (_strlen($buf)|0); //@line 4011 "load_abc.cpp"
         $373 = ($372>>>0)>(31); //@line 4011 "load_abc.cpp"
         if ($373) {
          HEAP8[$158>>0] = 0; //@line 4011 "load_abc.cpp"
         }
         (_strcpy($157,$buf)|0); //@line 4012 "load_abc.cpp"
         $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = 2;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$1$lcssa;$tpd$2 = $tpd$11121;
         break;
        }
        $374 = (_strncmp($p$1$lcssa,525415,2)|0); //@line 4016 "load_abc.cpp"
        $375 = ($374|0)==(0); //@line 4016 "load_abc.cpp"
        if ($375) {
         $376 = ((($p$1$lcssa)) + 2|0); //@line 4017 "load_abc.cpp"
         $p$3 = $376;
         while(1) {
          $377 = HEAP8[$p$3>>0]|0; //@line 4017 "load_abc.cpp"
          $378 = $377 << 24 >> 24; //@line 4017 "load_abc.cpp"
          $379 = (_isspace($378)|0); //@line 4017 "load_abc.cpp"
          $380 = ($379|0)==(0); //@line 4017 "load_abc.cpp"
          $381 = ((($p$3)) + 1|0); //@line 4017 "load_abc.cpp"
          if ($380) {
           $p$3$lcssa = $p$3;
           break;
          } else {
           $p$3 = $381;
          }
         }
         $382 = (_strncmp($p$3$lcssa,525418,8)|0); //@line 4018 "load_abc.cpp"
         $383 = ($382|0)==(0); //@line 4018 "load_abc.cpp"
         if ($383) {
          $384 = ((($p$3$lcssa)) + 8|0); //@line 4018 "load_abc.cpp"
          $385 = HEAP8[$384>>0]|0; //@line 4018 "load_abc.cpp"
          $386 = $385 << 24 >> 24; //@line 4018 "load_abc.cpp"
          $387 = (_isspace($386)|0); //@line 4018 "load_abc.cpp"
          $388 = ($387|0)==(0); //@line 4018 "load_abc.cpp"
          if (!($388)) {
           $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = 1;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = 2;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$3$lcssa;$tpd$2 = $tpd$11121;
           break;
          }
          $389 = HEAP8[$384>>0]|0; //@line 4018 "load_abc.cpp"
          $390 = ($389<<24>>24)==(0); //@line 4018 "load_abc.cpp"
          if ($390) {
           $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = 1;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = 2;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$3$lcssa;$tpd$2 = $tpd$11121;
           break;
          }
         }
         $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = 0;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = 2;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$3$lcssa;$tpd$2 = $tpd$11121;
         break;
        }
        $391 = (_strncmp($p$1$lcssa,525427,2)|0); //@line 4022 "load_abc.cpp"
        $392 = ($391|0)==(0); //@line 4022 "load_abc.cpp"
        if ($392) {
         $t$2 = 2;
         while(1) {
          $393 = (($p$1$lcssa) + ($t$2)|0); //@line 4023 "load_abc.cpp"
          $394 = HEAP8[$393>>0]|0; //@line 4023 "load_abc.cpp"
          $395 = ($394<<24>>24)==(32); //@line 4023 "load_abc.cpp"
          $396 = (($t$2) + 1)|0; //@line 4023 "load_abc.cpp"
          if ($395) {
           $t$2 = $396;
          } else {
           $$lcssa2131 = $393;
           break;
          }
         }
         $397 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,$$lcssa2131,0)|0); //@line 4024 "load_abc.cpp"
         HEAP8[$41>>0]=$397&255;HEAP8[$41+1>>0]=($397>>8)&255;HEAP8[$41+2>>0]=($397>>16)&255;HEAP8[$41+3>>0]=$397>>24; //@line 4024 "load_abc.cpp"
         $398 = ((($397)) + 41|0); //@line 4025 "load_abc.cpp"
         $399 = HEAP8[$398>>0]|0; //@line 4025 "load_abc.cpp"
         $400 = $399&255; //@line 4025 "load_abc.cpp"
         $401 = ((($397)) + 43|0); //@line 4026 "load_abc.cpp"
         $402 = HEAP8[$401>>0]|0; //@line 4026 "load_abc.cpp"
         $403 = ($402<<24>>24)==(0); //@line 4026 "load_abc.cpp"
         $404 = $403&1; //@line 4026 "load_abc.cpp"
         $abcnoslurs$1$ = $403 ? $abcnoslurs$11106 : 0; //@line 4027 "load_abc.cpp"
         $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$11101;$abcnolegato$3 = $404;$abcnoslurs$4 = $abcnoslurs$1$;$abcstate$6 = 2;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $400;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$1$lcssa;$tpd$2 = $tpd$11121;
         break;
        }
        $405 = (_strncmp($p$1$lcssa,525430,2)|0); //@line 4030 "load_abc.cpp"
        $406 = ($405|0)==(0); //@line 4030 "load_abc.cpp"
        if (!($406)) {
         $strlenfirst77 = HEAP8[$p$1$lcssa>>0]|0; //@line 4076 "load_abc.cpp"
         $526 = ($strlenfirst77<<24>>24)==(0); //@line 4076 "load_abc.cpp"
         $$abcstate$3 = $526 ? 1 : 2; //@line 4076 "load_abc.cpp"
         $4052 = $4051;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$11117;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$11101;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnoslurs$11106;$abcstate$6 = $$abcstate$3;$abctempo$4 = $abctempo$11111;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $p$1$lcssa;$tpd$2 = $tpd$11121;
         break;
        }
        $407 = ((($p$1$lcssa)) + 2|0); //@line 4032 "load_abc.cpp"
        $408 = (__ZL7ABC_KeyPKc($407)|0); //@line 4032 "load_abc.cpp"
        $409 = (47112 + ($408<<2)|0); //@line 4033 "load_abc.cpp"
        $410 = HEAP32[$409>>2]|0; //@line 4033 "load_abc.cpp"
        HEAP32[$vararg_buffer11>>2] = $410; //@line 4033 "load_abc.cpp"
        $vararg_ptr14 = ((($vararg_buffer11)) + 4|0); //@line 4033 "load_abc.cpp"
        HEAP32[$vararg_ptr14>>2] = $410; //@line 4033 "load_abc.cpp"
        (_sprintf($barsig,526499,$vararg_buffer11)|0); //@line 4033 "load_abc.cpp"
        $$0$i133 = $407;
        while(1) {
         $411 = HEAP8[$$0$i133>>0]|0; //@line 2917 "load_abc.cpp"
         $412 = $411 << 24 >> 24; //@line 2917 "load_abc.cpp"
         $413 = (_isspace($412)|0); //@line 2917 "load_abc.cpp"
         $414 = ($413|0)==(0); //@line 2917 "load_abc.cpp"
         $415 = ((($$0$i133)) + 1|0); //@line 2917 "load_abc.cpp"
         if ($414) {
          $$0$i133$lcssa = $$0$i133;
          break;
         } else {
          $$0$i133 = $415;
         }
        }
        $416 = HEAP8[$$0$i133$lcssa>>0]|0; //@line 2918 "load_abc.cpp"
        $417 = ($416<<24>>24)==(0); //@line 2918 "load_abc.cpp"
        L328: do {
         if ($417) {
          $$2$i135 = $$0$i133$lcssa;
         } else {
          $$12$i = $$0$i133$lcssa;$419 = $416;
          while(1) {
           $418 = $419 << 24 >> 24; //@line 2918 "load_abc.cpp"
           $420 = (_isspace($418)|0); //@line 2918 "load_abc.cpp"
           $421 = ($420|0)==(0); //@line 2918 "load_abc.cpp"
           if (!($421)) {
            $$2$i135 = $$12$i;
            break L328;
           }
           $422 = HEAP8[$$12$i>>0]|0; //@line 2918 "load_abc.cpp"
           $423 = ($422<<24>>24)==(93); //@line 2918 "load_abc.cpp"
           if ($423) {
            $$2$i135 = $$12$i;
            break L328;
           }
           $424 = ((($$12$i)) + 1|0); //@line 2918 "load_abc.cpp"
           $425 = HEAP8[$424>>0]|0; //@line 2918 "load_abc.cpp"
           $426 = ($425<<24>>24)==(0); //@line 2918 "load_abc.cpp"
           if ($426) {
            $$2$i135 = $424;
            break;
           } else {
            $$12$i = $424;$419 = $425;
           }
          }
         }
        } while(0);
        while(1) {
         $427 = HEAP8[$$2$i135>>0]|0; //@line 2919 "load_abc.cpp"
         $428 = $427 << 24 >> 24; //@line 2919 "load_abc.cpp"
         $429 = (_isspace($428)|0); //@line 2919 "load_abc.cpp"
         $430 = ($429|0)==(0); //@line 2919 "load_abc.cpp"
         $431 = ((($$2$i135)) + 1|0); //@line 2919 "load_abc.cpp"
         if ($430) {
          $$2$i135$lcssa = $$2$i135;
          break;
         } else {
          $$2$i135 = $431;
         }
        }
        $432 = (__ZL13abc_transposePKc($$2$i135$lcssa)|0); //@line 4035 "load_abc.cpp"
        HEAP8[$37>>0]=$432&255;HEAP8[$37+1>>0]=($432>>8)&255;HEAP8[$37+2>>0]=($432>>16)&255;HEAP8[$37+3>>0]=$432>>24; //@line 4035 "load_abc.cpp"
        HEAP8[$$2$i135$lcssa>>0] = 37; //@line 4036 "load_abc.cpp"
        $433 = HEAP32[$snotelen>>2]|0; //@line 4037 "load_abc.cpp"
        $434 = ($433|0)==(0); //@line 4037 "load_abc.cpp"
        if ($434) {
         $435 = HEAP32[$mnotediv>>2]|0; //@line 4038 "load_abc.cpp"
         $436 = ($435|0)==(0); //@line 4038 "load_abc.cpp"
         if ($436) {
          HEAP32[$mnotelen>>2] = 1; //@line 4038 "load_abc.cpp"
          HEAP32[$mnotediv>>2] = 1; //@line 4038 "load_abc.cpp"
          $438 = 1;$440 = 1;
         } else {
          $$pre1518 = HEAP32[$mnotelen>>2]|0; //@line 4039 "load_abc.cpp"
          $438 = $$pre1518;$440 = $435;
         }
         $437 = ($438*100)|0; //@line 4039 "load_abc.cpp"
         $439 = (($437|0) / ($440|0))&-1; //@line 4039 "load_abc.cpp"
         $441 = ($439|0)>(74); //@line 4040 "load_abc.cpp"
         $$80 = $441 ? 8 : 16; //@line 4041 "load_abc.cpp"
         HEAP32[$snotediv>>2] = $$80; //@line 4041 "load_abc.cpp"
         HEAP32[$snotelen>>2] = 1; //@line 4044 "load_abc.cpp"
         $443 = 1;$445 = $$80;
        } else {
         $$pre1519 = HEAP32[$snotediv>>2]|0;
         $443 = $433;$445 = $$pre1519;
        }
        $442 = ($443|0)==(1); //@line 4048 "load_abc.cpp"
        $444 = ($445|0)==(8); //@line 4048 "load_abc.cpp"
        $or$cond = $442 & $444; //@line 4048 "load_abc.cpp"
        $abchornpipe$1$ = $or$cond ? $abchornpipe$11113 : 0; //@line 4048 "load_abc.cpp"
        HEAP8[1278234] = 32; //@line 4050 "load_abc.cpp"
        HEAP8[$42>>0]=0&255;HEAP8[$42+1>>0]=(0>>8)&255;HEAP8[$42+2>>0]=(0>>16)&255;HEAP8[$42+3>>0]=0>>24; //@line 4054 "load_abc.cpp"
        HEAP8[$43>>0]=0&255;HEAP8[$43+1>>0]=(0>>8)&255;HEAP8[$43+2>>0]=(0>>16)&255;HEAP8[$43+3>>0]=0>>24; //@line 4056 "load_abc.cpp"
        $strlenfirst75 = HEAP8[$6>>0]|0; //@line 4057 "load_abc.cpp"
        $446 = ($strlenfirst75<<24>>24)==(0); //@line 4057 "load_abc.cpp"
        L344: do {
         if ($446) {
          $447 = HEAP32[$mnotelen>>2]|0; //@line 4057 "load_abc.cpp"
          $448 = HEAP32[$mnotediv>>2]|0; //@line 4057 "load_abc.cpp"
          $449 = $447 << 4; //@line 3269 "load_abc.cpp"
          $450 = (($449) + ($448))|0; //@line 3269 "load_abc.cpp"
          switch ($450|0) {
          case 34: case 68: case 36:  {
           (__ZL15abc_MIDI_gchordPKcP10_ABCHANDLE(526504,$1)|0); //@line 3273 "load_abc.cpp"
           break L344;
           break;
          }
          case 50: case 100:  {
           (__ZL15abc_MIDI_gchordPKcP10_ABCHANDLE(526543,$1)|0); //@line 3277 "load_abc.cpp"
           break L344;
           break;
          }
          case 56: case 52:  {
           (__ZL15abc_MIDI_gchordPKcP10_ABCHANDLE(526556,$1)|0); //@line 3281 "load_abc.cpp"
           break L344;
           break;
          }
          case 104:  {
           (__ZL15abc_MIDI_gchordPKcP10_ABCHANDLE(526563,$1)|0); //@line 3284 "load_abc.cpp"
           break L344;
           break;
          }
          case 152:  {
           (__ZL15abc_MIDI_gchordPKcP10_ABCHANDLE(526570,$1)|0); //@line 3287 "load_abc.cpp"
           break L344;
           break;
          }
          case 200:  {
           (__ZL15abc_MIDI_gchordPKcP10_ABCHANDLE(526580,$1)|0); //@line 3290 "load_abc.cpp"
           break L344;
           break;
          }
          default: {
           $451 = (($447|0) % 3)&-1; //@line 3293 "load_abc.cpp"
           $452 = ($451|0)==(0); //@line 3293 "load_abc.cpp"
           if ($452) {
            (__ZL15abc_MIDI_gchordPKcP10_ABCHANDLE(526593,$1)|0); //@line 3294 "load_abc.cpp"
           } else {
            (__ZL15abc_MIDI_gchordPKcP10_ABCHANDLE(526621,$1)|0); //@line 3296 "load_abc.cpp"
           }
           $453 = ($448|0)==(8); //@line 3297 "load_abc.cpp"
           if ($453) {
            $454 = $447 << 1; //@line 3297 "load_abc.cpp"
            $455 = (((($1)) + 36|0) + ($454)|0); //@line 3297 "load_abc.cpp"
            HEAP8[$455>>0] = 0; //@line 3297 "load_abc.cpp"
            break L344;
           } else {
            $456 = $447 << 2; //@line 3298 "load_abc.cpp"
            $457 = (((($1)) + 36|0) + ($456)|0); //@line 3298 "load_abc.cpp"
            HEAP8[$457>>0] = 0; //@line 3298 "load_abc.cpp"
            break L344;
           }
          }
          }
         }
        } while(0);
        $458 = HEAP32[$mnotediv>>2]|0; //@line 4058 "load_abc.cpp"
        $459 = ($458|0)==(0); //@line 345 "load_abc.cpp"
        if ($459) {
         $$0$i136 = 0;
        } else {
         $460 = HEAP32[$mnotelen>>2]|0; //@line 4058 "load_abc.cpp"
         $461 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4058 "load_abc.cpp"
         $462 = HEAP32[299862]|0; //@line 346 "load_abc.cpp"
         $463 = ($460*3072)|0; //@line 346 "load_abc.cpp"
         $464 = Math_imul($463, $461)|0; //@line 346 "load_abc.cpp"
         $465 = Math_imul($464, $462)|0; //@line 346 "load_abc.cpp"
         $466 = HEAP32[299863]|0; //@line 346 "load_abc.cpp"
         $467 = Math_imul($466, $458)|0; //@line 346 "load_abc.cpp"
         $468 = (($465|0) / ($467|0))&-1; //@line 346 "load_abc.cpp"
         $$0$i136 = $468;
        }
        HEAP8[$7>>0]=$$0$i136&255;HEAP8[$7+1>>0]=($$0$i136>>8)&255;HEAP8[$7+2>>0]=($$0$i136>>16)&255;HEAP8[$7+3>>0]=$$0$i136>>24; //@line 4058 "load_abc.cpp"
        $469 = ($abctempo$11111|0)==(0); //@line 4059 "load_abc.cpp"
        $470 = ($ktempo$11115|0)!=(0); //@line 4059 "load_abc.cpp"
        $or$cond3 = $469 | $470; //@line 4059 "load_abc.cpp"
        do {
         if ($or$cond3) {
          $abctempo$2 = $abctempo$11111;
         } else {
          HEAP32[299862] = 2; //@line 4061 "load_abc.cpp"
          HEAP32[299863] = 1; //@line 4062 "load_abc.cpp"
          $471 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 4063 "load_abc.cpp"
          $472 = ($471|0)==(0|0); //@line 4063 "load_abc.cpp"
          if (!($472)) {
           $473 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4065 "load_abc.cpp"
           $474 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
           HEAP8[$474>>0]=0&255;HEAP8[$474+1>>0]=(0>>8)&255;HEAP8[$474+2>>0]=(0>>16)&255;HEAP8[$474+3>>0]=0>>24; //@line 540 "load_abc.cpp"
           $475 = ((($474)) + 4|0); //@line 541 "load_abc.cpp"
           HEAP8[$475>>0]=$473&255;HEAP8[$475+1>>0]=($473>>8)&255;HEAP8[$475+2>>0]=($473>>16)&255;HEAP8[$475+3>>0]=$473>>24; //@line 541 "load_abc.cpp"
           $476 = ((($474)) + 8|0); //@line 543 "load_abc.cpp"
           HEAP8[$476>>0] = 1; //@line 543 "load_abc.cpp"
           $477 = ((($474)) + 9|0); //@line 543 "load_abc.cpp"
           HEAP8[$477>>0] = 80; //@line 543 "load_abc.cpp"
           $478 = ((($474)) + 10|0); //@line 543 "load_abc.cpp"
           HEAP8[$478>>0]=0&255;HEAP8[$478+1>>0]=(0>>8)&255;HEAP8[$478+2>>0]=(0>>16)&255;HEAP8[$478+3>>0]=0>>24; //@line 543 "load_abc.cpp"
           $479 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
           $480 = ((($474)) + 14|0); //@line 544 "load_abc.cpp"
           HEAP8[$480>>0] = $479; //@line 544 "load_abc.cpp"
           $481 = ((($474)) + 15|0); //@line 545 "load_abc.cpp"
           HEAP8[$481>>0] = 0; //@line 545 "load_abc.cpp"
           __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($471,$474); //@line 1153 "load_abc.cpp"
           $482 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 4066 "load_abc.cpp"
           $483 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4066 "load_abc.cpp"
           $484 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
           HEAP8[$484>>0]=0&255;HEAP8[$484+1>>0]=(0>>8)&255;HEAP8[$484+2>>0]=(0>>16)&255;HEAP8[$484+3>>0]=0>>24; //@line 540 "load_abc.cpp"
           $485 = ((($484)) + 4|0); //@line 541 "load_abc.cpp"
           HEAP8[$485>>0]=$483&255;HEAP8[$485+1>>0]=($483>>8)&255;HEAP8[$485+2>>0]=($483>>16)&255;HEAP8[$485+3>>0]=$483>>24; //@line 541 "load_abc.cpp"
           $486 = ((($484)) + 8|0); //@line 543 "load_abc.cpp"
           HEAP8[$486>>0] = 1; //@line 543 "load_abc.cpp"
           $487 = ((($484)) + 9|0); //@line 543 "load_abc.cpp"
           HEAP8[$487>>0] = 84; //@line 543 "load_abc.cpp"
           $488 = ((($484)) + 10|0); //@line 543 "load_abc.cpp"
           $489 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
           $490 = ((($484)) + 14|0); //@line 544 "load_abc.cpp"
           HEAP8[$490>>0] = $489; //@line 544 "load_abc.cpp"
           $491 = ((($484)) + 15|0); //@line 545 "load_abc.cpp"
           HEAP8[$491>>0] = 0; //@line 545 "load_abc.cpp"
           HEAP8[$488>>0]=240&255;HEAP8[$488+1>>0]=(240>>8)&255;HEAP8[$488+2>>0]=(240>>16)&255;HEAP8[$488+3>>0]=240>>24; //@line 1164 "load_abc.cpp"
           __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($482,$484); //@line 1165 "load_abc.cpp"
          }
          $492 = HEAP32[$30>>2]|0; //@line 4068 "load_abc.cpp"
          $493 = ($492|0)==(0); //@line 4068 "load_abc.cpp"
          if (!($493)) {
           $abctempo$2 = 0;
           break;
          }
          HEAP32[$30>>2] = 240; //@line 4068 "load_abc.cpp"
          $abctempo$2 = 0;
         }
        } while(0);
        HEAP8[$partpat>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$49>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$50>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$51>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$52>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$53>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$54>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$55>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$56>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$57>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$58>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$59>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$60>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$61>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$62>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$63>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$64>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$65>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$66>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$67>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$68>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$69>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$70>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$71>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$72>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$73>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$74>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$75>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$76>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$77>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$78>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$79>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$80>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$81>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$82>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$83>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$84>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$85>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$86>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$87>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$88>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$89>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$90>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$91>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$92>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$93>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$94>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$95>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$96>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$97>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$98>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$99>>0] = 0; //@line 3584 "load_abc.cpp"
        HEAP8[$100>>0] = -1; //@line 3583 "load_abc.cpp"
        HEAP8[$101>>0] = 0; //@line 3584 "load_abc.cpp"
        $494 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4071 "load_abc.cpp"
        $495 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 1928 "load_abc.cpp"
        $496 = ($495*12288)|0; //@line 339 "load_abc.cpp"
        $497 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 1929 "load_abc.cpp"
        $498 = ($497|0)==(0|0); //@line 1929 "load_abc.cpp"
        if ($498) {
         HEAP8[$100>>0] = 0; //@line 4071 "load_abc.cpp"
         HEAP8[$101>>0] = 0; //@line 4072 "load_abc.cpp"
         $4052 = 0;$abceffect$2 = 0;$abceffoper$2 = 0;$abcgrace$3 = 0;$abchornpipe$3 = $abchornpipe$1$;$abckey$3 = $408;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnolegato$11107;$abcstate$6 = 3;$abctempo$4 = $abctempo$2;$abcto$2 = 0;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $$2$i135$lcssa;$tpd$2 = 0;
         break;
        }
        $499 = ((($497)) + 4|0); //@line 1929 "load_abc.cpp"
        $500 = HEAPU8[$499>>0]|(HEAPU8[$499+1>>0]<<8)|(HEAPU8[$499+2>>0]<<16)|(HEAPU8[$499+3>>0]<<24); //@line 1929 "load_abc.cpp"
        $501 = ($500|0)==(0|0); //@line 1929 "load_abc.cpp"
        if ($501) {
         HEAP8[$100>>0] = 0; //@line 4071 "load_abc.cpp"
         HEAP8[$101>>0] = 0; //@line 4072 "load_abc.cpp"
         $tp$08$us11$i = $497;
        } else {
         $dt$03$i$i = 0;$e$02$i$i = $500;
         while(1) {
          $502 = ((($e$02$i$i)) + 4|0); //@line 1931 "load_abc.cpp"
          $503 = HEAPU8[$502>>0]|(HEAPU8[$502+1>>0]<<8)|(HEAPU8[$502+2>>0]<<16)|(HEAPU8[$502+3>>0]<<24); //@line 1931 "load_abc.cpp"
          $504 = ($503>>>0)>($494>>>0); //@line 1931 "load_abc.cpp"
          if ($504) {
           $dt$0$lcssa$i$i = $dt$03$i$i;
           break;
          }
          $505 = ((($e$02$i$i)) + 8|0); //@line 1932 "load_abc.cpp"
          $506 = HEAP8[$505>>0]|0; //@line 1932 "load_abc.cpp"
          $507 = ($506<<24>>24)==(1); //@line 1932 "load_abc.cpp"
          do {
           if ($507) {
            $508 = ((($e$02$i$i)) + 9|0); //@line 1932 "load_abc.cpp"
            $509 = HEAP8[$508>>0]|0; //@line 1932 "load_abc.cpp"
            $510 = ($509<<24>>24)==(80); //@line 1932 "load_abc.cpp"
            if (!($510)) {
             $dt$1$i$i = $dt$03$i$i;
             break;
            }
            $511 = (($503) + ($dt$03$i$i))|0; //@line 1933 "load_abc.cpp"
            $512 = (($511>>>0) % ($496>>>0))&-1; //@line 1934 "load_abc.cpp"
            $513 = ($512|0)==(0); //@line 1934 "load_abc.cpp"
            if ($513) {
             $dt$1$i$i = $dt$03$i$i;
             break;
            }
            $514 = (($511) + ($496))|0; //@line 1935 "load_abc.cpp"
            $515 = (($514>>>0) % ($496>>>0))&-1; //@line 1937 "load_abc.cpp"
            $516 = (($dt$03$i$i) + ($496))|0; //@line 1937 "load_abc.cpp"
            $517 = (($516) - ($515))|0; //@line 1938 "load_abc.cpp"
            $dt$1$i$i = $517;
           } else {
            $dt$1$i$i = $dt$03$i$i;
           }
          } while(0);
          $518 = HEAPU8[$e$02$i$i>>0]|(HEAPU8[$e$02$i$i+1>>0]<<8)|(HEAPU8[$e$02$i$i+2>>0]<<16)|(HEAPU8[$e$02$i$i+3>>0]<<24); //@line 1931 "load_abc.cpp"
          $519 = ($518|0)==(0|0); //@line 1931 "load_abc.cpp"
          if ($519) {
           $dt$0$lcssa$i$i = $dt$1$i$i;
           break;
          } else {
           $dt$03$i$i = $dt$1$i$i;$e$02$i$i = $518;
          }
         }
         $520 = (($dt$0$lcssa$i$i) + ($494))|0; //@line 1942 "load_abc.cpp"
         $phitmp$i = (($520>>>0) / 192)&-1; //@line 1942 "load_abc.cpp"
         $phitmp1$i = $phitmp$i >>> 6; //@line 1942 "load_abc.cpp"
         $521 = (($phitmp1$i>>>0) / ($495>>>0))&-1; //@line 1947 "load_abc.cpp"
         $522 = $521&255; //@line 4071 "load_abc.cpp"
         HEAP8[$100>>0] = $522; //@line 4071 "load_abc.cpp"
         HEAP8[$101>>0] = 0; //@line 4072 "load_abc.cpp"
         if ($498) {
          $4052 = 0;$abceffect$2 = 0;$abceffoper$2 = 0;$abcgrace$3 = 0;$abchornpipe$3 = $abchornpipe$1$;$abckey$3 = $408;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnolegato$11107;$abcstate$6 = 3;$abctempo$4 = $abctempo$2;$abcto$2 = 0;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $$2$i135$lcssa;$tpd$2 = 0;
          break;
         } else {
          $tp$08$us11$i = $497;
         }
        }
        while(1) {
         $523 = ((($tp$08$us11$i)) + 28|0); //@line 3636 "load_abc.cpp"
         HEAP8[$523>>0]=$494&255;HEAP8[$523+1>>0]=($494>>8)&255;HEAP8[$523+2>>0]=($494>>16)&255;HEAP8[$523+3>>0]=$494>>24; //@line 3636 "load_abc.cpp"
         $524 = ((($tp$08$us11$i)) + 32|0); //@line 3638 "load_abc.cpp"
         HEAP8[$524>>0]=0&255;HEAP8[$524+1>>0]=(0>>8)&255;HEAP8[$524+2>>0]=(0>>16)&255;HEAP8[$524+3>>0]=0>>24; //@line 3638 "load_abc.cpp"
         $tp$0$us15$i = HEAPU8[$tp$08$us11$i>>0]|(HEAPU8[$tp$08$us11$i+1>>0]<<8)|(HEAPU8[$tp$08$us11$i+2>>0]<<16)|(HEAPU8[$tp$08$us11$i+3>>0]<<24); //@line 3634 "load_abc.cpp"
         $525 = ($tp$0$us15$i|0)==(0|0); //@line 3634 "load_abc.cpp"
         if ($525) {
          $4052 = 0;$abceffect$2 = 0;$abceffoper$2 = 0;$abcgrace$3 = 0;$abchornpipe$3 = $abchornpipe$1$;$abckey$3 = $408;$abcnolegato$3 = $abcnolegato$11107;$abcnoslurs$4 = $abcnolegato$11107;$abcstate$6 = 3;$abctempo$4 = $abctempo$2;$abcto$2 = 0;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$11119;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$11090;$p$7 = $$2$i135$lcssa;$tpd$2 = 0;
          break;
         } else {
          $tp$08$us11$i = $tp$0$us15$i;
         }
        }
       }
       else if ((label|0) == 218) {
        label = 0;
        $648 = (_strncmp($p$51550,525427,2)|0); //@line 4112 "load_abc.cpp"
        $649 = ($648|0)==(0); //@line 4112 "load_abc.cpp"
        if ($649) {
         $t$5 = 2;
         while(1) {
          $650 = (($p$51550) + ($t$5)|0); //@line 4113 "load_abc.cpp"
          $651 = HEAP8[$650>>0]|0; //@line 4113 "load_abc.cpp"
          $652 = ($651<<24>>24)==(32); //@line 4113 "load_abc.cpp"
          $653 = (($t$5) + 1)|0; //@line 4113 "load_abc.cpp"
          if ($652) {
           $t$5 = $653;
          } else {
           $$lcssa2129 = $650;
           break;
          }
         }
         $654 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,$$lcssa2129,0)|0); //@line 4114 "load_abc.cpp"
         HEAP8[$41>>0]=$654&255;HEAP8[$41+1>>0]=($654>>8)&255;HEAP8[$41+2>>0]=($654>>16)&255;HEAP8[$41+3>>0]=$654>>24; //@line 4114 "load_abc.cpp"
         $655 = (47112 + ($abckey$11101<<2)|0); //@line 4115 "load_abc.cpp"
         $656 = HEAP32[$655>>2]|0; //@line 4115 "load_abc.cpp"
         HEAP32[$vararg_buffer21>>2] = $656; //@line 4115 "load_abc.cpp"
         $vararg_ptr24 = ((($vararg_buffer21)) + 4|0); //@line 4115 "load_abc.cpp"
         HEAP32[$vararg_ptr24>>2] = $656; //@line 4115 "load_abc.cpp"
         (_sprintf($barsig,526499,$vararg_buffer21)|0); //@line 4115 "load_abc.cpp"
         $657 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4118 "load_abc.cpp"
         $$idx102 = ((($657)) + 8|0);
         $$idx102$val = HEAPU8[$$idx102>>0]|(HEAPU8[$$idx102+1>>0]<<8)|(HEAPU8[$$idx102+2>>0]<<16)|(HEAPU8[$$idx102+3>>0]<<24);
         $658 = ($$idx102$val|0)==(0|0); //@line 2926 "load_abc.cpp"
         if ($658) {
          $tracktime$0$i = 0;
         } else {
          $659 = ((($$idx102$val)) + 4|0); //@line 2926 "load_abc.cpp"
          $660 = HEAPU8[$659>>0]|(HEAPU8[$659+1>>0]<<8)|(HEAPU8[$659+2>>0]<<16)|(HEAPU8[$659+3>>0]<<24); //@line 2926 "load_abc.cpp"
          $tracktime$0$i = $660;
         }
         $661 = HEAP32[299861]|0; //@line 2928 "load_abc.cpp"
         $662 = ($tracktime$0$i>>>0)<($661>>>0); //@line 2928 "load_abc.cpp"
         $$tracktime$0$i = $662 ? $661 : $tracktime$0$i; //@line 2928 "load_abc.cpp"
         HEAP8[$31>>0]=$$tracktime$0$i&255;HEAP8[$31+1>>0]=($$tracktime$0$i>>8)&255;HEAP8[$31+2>>0]=($$tracktime$0$i>>16)&255;HEAP8[$31+3>>0]=$$tracktime$0$i>>24; //@line 4118 "load_abc.cpp"
         $663 = ((($657)) + 43|0); //@line 4120 "load_abc.cpp"
         $664 = HEAP8[$663>>0]|0; //@line 4120 "load_abc.cpp"
         $665 = ($664<<24>>24)==(0); //@line 4120 "load_abc.cpp"
         $666 = $665&1; //@line 4120 "load_abc.cpp"
         $abcnoslurs$1$83 = $665 ? $abcnoslurs$11106 : 0; //@line 4121 "load_abc.cpp"
         HEAP8[$p$51550>>0] = 37; //@line 4122 "load_abc.cpp"
         $4056 = 0;$abcgrace$2 = 0;$abcnolegato$2 = $666;$abcnoslurs$3 = $abcnoslurs$1$83;$bartime$2 = $$tracktime$0$i;
        } else {
         $4056 = $4051;$abcgrace$2 = $abcgrace$11117;$abcnolegato$2 = $abcnolegato$11107;$abcnoslurs$3 = $abcnoslurs$11106;$bartime$2 = $bartime$11119;
        }
        $667 = (_strncmp($p$51550,525430,2)|0); //@line 4124 "load_abc.cpp"
        $668 = ($667|0)==(0); //@line 4124 "load_abc.cpp"
        if ($668) {
         $669 = ((($p$51550)) + 2|0); //@line 4125 "load_abc.cpp"
         $670 = (__ZL7ABC_KeyPKc($669)|0); //@line 4125 "load_abc.cpp"
         $671 = (47112 + ($670<<2)|0); //@line 4126 "load_abc.cpp"
         $672 = HEAP32[$671>>2]|0; //@line 4126 "load_abc.cpp"
         HEAP32[$vararg_buffer25>>2] = $672; //@line 4126 "load_abc.cpp"
         $vararg_ptr28 = ((($vararg_buffer25)) + 4|0); //@line 4126 "load_abc.cpp"
         HEAP32[$vararg_ptr28>>2] = $672; //@line 4126 "load_abc.cpp"
         (_sprintf($barsig,526499,$vararg_buffer25)|0); //@line 4126 "load_abc.cpp"
         $$0$i156 = $669;
         while(1) {
          $673 = HEAP8[$$0$i156>>0]|0; //@line 2917 "load_abc.cpp"
          $674 = $673 << 24 >> 24; //@line 2917 "load_abc.cpp"
          $675 = (_isspace($674)|0); //@line 2917 "load_abc.cpp"
          $676 = ($675|0)==(0); //@line 2917 "load_abc.cpp"
          $677 = ((($$0$i156)) + 1|0); //@line 2917 "load_abc.cpp"
          if ($676) {
           $$0$i156$lcssa = $$0$i156;
           break;
          } else {
           $$0$i156 = $677;
          }
         }
         $678 = HEAP8[$$0$i156$lcssa>>0]|0; //@line 2918 "load_abc.cpp"
         $679 = ($678<<24>>24)==(0); //@line 2918 "load_abc.cpp"
         L402: do {
          if ($679) {
           $$2$i160 = $$0$i156$lcssa;
          } else {
           $$12$i158 = $$0$i156$lcssa;$681 = $678;
           while(1) {
            $680 = $681 << 24 >> 24; //@line 2918 "load_abc.cpp"
            $682 = (_isspace($680)|0); //@line 2918 "load_abc.cpp"
            $683 = ($682|0)==(0); //@line 2918 "load_abc.cpp"
            if (!($683)) {
             $$2$i160 = $$12$i158;
             break L402;
            }
            $684 = HEAP8[$$12$i158>>0]|0; //@line 2918 "load_abc.cpp"
            $685 = ($684<<24>>24)==(93); //@line 2918 "load_abc.cpp"
            if ($685) {
             $$2$i160 = $$12$i158;
             break L402;
            }
            $686 = ((($$12$i158)) + 1|0); //@line 2918 "load_abc.cpp"
            $687 = HEAP8[$686>>0]|0; //@line 2918 "load_abc.cpp"
            $688 = ($687<<24>>24)==(0); //@line 2918 "load_abc.cpp"
            if ($688) {
             $$2$i160 = $686;
             break;
            } else {
             $$12$i158 = $686;$681 = $687;
            }
           }
          }
         } while(0);
         while(1) {
          $689 = HEAP8[$$2$i160>>0]|0; //@line 2919 "load_abc.cpp"
          $690 = $689 << 24 >> 24; //@line 2919 "load_abc.cpp"
          $691 = (_isspace($690)|0); //@line 2919 "load_abc.cpp"
          $692 = ($691|0)==(0); //@line 2919 "load_abc.cpp"
          $693 = ((($$2$i160)) + 1|0); //@line 2919 "load_abc.cpp"
          if ($692) {
           $$2$i160$lcssa = $$2$i160;
           break;
          } else {
           $$2$i160 = $693;
          }
         }
         $694 = (__ZL13abc_transposePKc($$2$i160$lcssa)|0); //@line 4128 "load_abc.cpp"
         HEAP8[$37>>0]=$694&255;HEAP8[$37+1>>0]=($694>>8)&255;HEAP8[$37+2>>0]=($694>>16)&255;HEAP8[$37+3>>0]=$694>>24; //@line 4128 "load_abc.cpp"
         HEAP8[$$2$i160$lcssa>>0] = 37; //@line 4129 "load_abc.cpp"
         $abckey$2 = $670;$p$6 = $$2$i160$lcssa;
        } else {
         $abckey$2 = $abckey$11101;$p$6 = $p$51550;
        }
        $695 = (_strncmp($p$6,525124,2)|0); //@line 4131 "load_abc.cpp"
        $696 = ($695|0)==(0); //@line 4131 "load_abc.cpp"
        if ($696) {
         $697 = ((($p$6)) + 2|0); //@line 4132 "load_abc.cpp"
         HEAP32[$vararg_buffer29>>2] = $snotelen; //@line 4132 "load_abc.cpp"
         $vararg_ptr32 = ((($vararg_buffer29)) + 4|0); //@line 4132 "load_abc.cpp"
         HEAP32[$vararg_ptr32>>2] = $snotediv; //@line 4132 "load_abc.cpp"
         (_sscanf($697,525127,$vararg_buffer29)|0); //@line 4132 "load_abc.cpp"
         HEAP8[$p$6>>0] = 37; //@line 4133 "load_abc.cpp"
        }
        $698 = (_strncmp($p$6,525136,2)|0); //@line 4135 "load_abc.cpp"
        $699 = ($698|0)==(0); //@line 4135 "load_abc.cpp"
        if ($699) {
         $700 = ((($p$6)) + 2|0); //@line 4136 "load_abc.cpp"
         $701 = (_strncmp($700,525139,4)|0); //@line 3312 "load_abc.cpp"
         $702 = ($701|0)==(0); //@line 3312 "load_abc.cpp"
         do {
          if ($702) {
           HEAP32[$mnotelen>>2] = 1; //@line 3313 "load_abc.cpp"
           HEAP32[$mnotediv>>2] = 1; //@line 3314 "load_abc.cpp"
           $713 = 1;$718 = 1;
           label = 243;
          } else {
           $703 = (_strncmp($700,525144,2)|0); //@line 3317 "load_abc.cpp"
           $704 = ($703|0)==(0); //@line 3317 "load_abc.cpp"
           if ($704) {
            HEAP32[$mnotelen>>2] = 2; //@line 3318 "load_abc.cpp"
            HEAP32[$mnotediv>>2] = 2; //@line 3319 "load_abc.cpp"
            $713 = 2;$718 = 2;
            label = 243;
            break;
           }
           $705 = HEAP8[$700>>0]|0; //@line 3322 "load_abc.cpp"
           $706 = ($705<<24>>24)==(67); //@line 3322 "load_abc.cpp"
           if ($706) {
            HEAP32[$mnotelen>>2] = 4; //@line 3323 "load_abc.cpp"
            HEAP32[$mnotediv>>2] = 4; //@line 3324 "load_abc.cpp"
            $713 = 4;$718 = 4;
            label = 243;
            break;
           }
           $707 = (__ZL11abc_getexprPKcPi($700,$mnotelen)|0); //@line 3327 "load_abc.cpp"
           $$sum830 = (($707) + 2)|0; //@line 3327 "load_abc.cpp"
           $708 = (($p$6) + ($$sum830)|0); //@line 3327 "load_abc.cpp"
           HEAP32[$vararg_buffer33>>2] = $mnotediv; //@line 3328 "load_abc.cpp"
           (_sscanf($708,525147,$vararg_buffer33)|0); //@line 3328 "load_abc.cpp"
           $$pr759 = HEAP32[$mnotediv>>2]|0; //@line 4137 "load_abc.cpp"
           $709 = ($$pr759|0)==(0); //@line 345 "load_abc.cpp"
           if ($709) {
            $$0$i164 = 0;
            break;
           }
           $$pre1515 = HEAP32[$mnotelen>>2]|0; //@line 4137 "load_abc.cpp"
           $713 = $$pre1515;$718 = $$pr759;
           label = 243;
          }
         } while(0);
         if ((label|0) == 243) {
          label = 0;
          $710 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4137 "load_abc.cpp"
          $711 = ($710*3072)|0; //@line 346 "load_abc.cpp"
          $712 = Math_imul($711, $713)|0; //@line 346 "load_abc.cpp"
          $714 = HEAP32[299862]|0; //@line 346 "load_abc.cpp"
          $715 = Math_imul($712, $714)|0; //@line 346 "load_abc.cpp"
          $716 = HEAP32[299863]|0; //@line 346 "load_abc.cpp"
          $717 = Math_imul($716, $718)|0; //@line 346 "load_abc.cpp"
          $719 = (($715|0) / ($717|0))&-1; //@line 346 "load_abc.cpp"
          $$0$i164 = $719;
         }
         HEAP8[$7>>0]=$$0$i164&255;HEAP8[$7+1>>0]=($$0$i164>>8)&255;HEAP8[$7+2>>0]=($$0$i164>>16)&255;HEAP8[$7+3>>0]=$$0$i164>>24; //@line 4137 "load_abc.cpp"
         HEAP8[$p$6>>0] = 37; //@line 4138 "load_abc.cpp"
        }
        $720 = (_strncmp($p$6,525350,2)|0); //@line 4140 "load_abc.cpp"
        $721 = ($720|0)==(0); //@line 4140 "load_abc.cpp"
        if ($721) {
         $722 = ((($p$6)) + 2|0); //@line 4141 "load_abc.cpp"
         $723 = (__ZL17abc_extract_tempoPKci($722,$181)|0); //@line 4141 "load_abc.cpp"
         $724 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 4142 "load_abc.cpp"
         $725 = ($724|0)==(0|0); //@line 4142 "load_abc.cpp"
         if ($725) {
          $726 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
          $727 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
          $728 = ((($726)) + 20|0); //@line 917 "load_abc.cpp"
          HEAP8[$728>>0]=$727&255;HEAP8[$728+1>>0]=($727>>8)&255;HEAP8[$728+2>>0]=($727>>16)&255;HEAP8[$728+3>>0]=$727>>24; //@line 917 "load_abc.cpp"
          HEAP8[$41>>0]=$726&255;HEAP8[$41+1>>0]=($726>>8)&255;HEAP8[$41+2>>0]=($726>>16)&255;HEAP8[$41+3>>0]=$726>>24; //@line 4143 "load_abc.cpp"
          $729 = ((($726)) + 36|0); //@line 4144 "load_abc.cpp"
          HEAP8[$729>>0] = 0; //@line 4144 "load_abc.cpp"
          $$pre1516 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 4146 "load_abc.cpp"
          $739 = $$pre1516;
         } else {
          $739 = $724;
         }
         $730 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4146 "load_abc.cpp"
         $731 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
         HEAP8[$731>>0]=0&255;HEAP8[$731+1>>0]=(0>>8)&255;HEAP8[$731+2>>0]=(0>>16)&255;HEAP8[$731+3>>0]=0>>24; //@line 540 "load_abc.cpp"
         $732 = ((($731)) + 4|0); //@line 541 "load_abc.cpp"
         HEAP8[$732>>0]=$730&255;HEAP8[$732+1>>0]=($730>>8)&255;HEAP8[$732+2>>0]=($730>>16)&255;HEAP8[$732+3>>0]=$730>>24; //@line 541 "load_abc.cpp"
         $733 = ((($731)) + 8|0); //@line 543 "load_abc.cpp"
         HEAP8[$733>>0] = 1; //@line 543 "load_abc.cpp"
         $734 = ((($731)) + 9|0); //@line 543 "load_abc.cpp"
         HEAP8[$734>>0] = 84; //@line 543 "load_abc.cpp"
         $735 = ((($731)) + 10|0); //@line 543 "load_abc.cpp"
         $736 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
         $737 = ((($731)) + 14|0); //@line 544 "load_abc.cpp"
         HEAP8[$737>>0] = $736; //@line 544 "load_abc.cpp"
         $738 = ((($731)) + 15|0); //@line 545 "load_abc.cpp"
         HEAP8[$738>>0] = 0; //@line 545 "load_abc.cpp"
         HEAP8[$735>>0]=$723&255;HEAP8[$735+1>>0]=($723>>8)&255;HEAP8[$735+2>>0]=($723>>16)&255;HEAP8[$735+3>>0]=$723>>24; //@line 1164 "load_abc.cpp"
         __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($739,$731); //@line 1165 "load_abc.cpp"
         HEAP8[$p$6>>0] = 37; //@line 4147 "load_abc.cpp"
         $abctempo$3 = $723;
        } else {
         $abctempo$3 = $abctempo$11111;
        }
        $740 = (_strncmp($p$6,525412,2)|0); //@line 4149 "load_abc.cpp"
        $741 = ($740|0)==(0); //@line 4149 "load_abc.cpp"
        if (!($741)) {
         $4052 = $4056;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$2;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$2;$abcnolegato$3 = $abcnolegato$2;$abcnoslurs$4 = $abcnoslurs$3;$abcstate$6 = $abcstate$5;$abctempo$4 = $abctempo$3;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$2;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$2;$p$7 = $p$6;$tpd$2 = $tpd$11121;
         break;
        }
        $742 = (_strchr($p$6,37)|0); //@line 4151 "load_abc.cpp"
        $743 = ($742|0)==(0|0); //@line 4151 "load_abc.cpp"
        if (!($743)) {
         HEAP8[$742>>0] = 0; //@line 4151 "load_abc.cpp"
        }
        $744 = (_strlen($p$6)|0); //@line 4152 "load_abc.cpp"
        $t$61064 = (($744) + -1)|0; //@line 4152 "load_abc.cpp"
        $745 = (($p$6) + ($t$61064)|0); //@line 4152 "load_abc.cpp"
        $746 = HEAP8[$745>>0]|0; //@line 4152 "load_abc.cpp"
        $747 = $746 << 24 >> 24; //@line 4152 "load_abc.cpp"
        $748 = (_isspace($747)|0); //@line 4152 "load_abc.cpp"
        $749 = ($748|0)==(0); //@line 4152 "load_abc.cpp"
        if ($749) {
         $t$7 = 2;
        } else {
         $750 = $745;$t$61065 = $t$61064;
         while(1) {
          HEAP8[$750>>0] = 0; //@line 4153 "load_abc.cpp"
          $t$6 = (($t$61065) + -1)|0; //@line 4152 "load_abc.cpp"
          $751 = (($p$6) + ($t$6)|0); //@line 4152 "load_abc.cpp"
          $752 = HEAP8[$751>>0]|0; //@line 4152 "load_abc.cpp"
          $753 = $752 << 24 >> 24; //@line 4152 "load_abc.cpp"
          $754 = (_isspace($753)|0); //@line 4152 "load_abc.cpp"
          $755 = ($754|0)==(0); //@line 4152 "load_abc.cpp"
          if ($755) {
           $t$7 = 2;
           break;
          } else {
           $750 = $751;$t$61065 = $t$6;
          }
         }
        }
        while(1) {
         $756 = (($p$6) + ($t$7)|0); //@line 4154 "load_abc.cpp"
         $757 = HEAP8[$756>>0]|0; //@line 4154 "load_abc.cpp"
         $758 = $757 << 24 >> 24; //@line 4154 "load_abc.cpp"
         $759 = (_isspace($758)|0); //@line 4154 "load_abc.cpp"
         $760 = ($759|0)==(0); //@line 4154 "load_abc.cpp"
         $761 = (($t$7) + 1)|0; //@line 4154 "load_abc.cpp"
         if ($760) {
          $$lcssa2130 = $756;$t$7$lcssa = $t$7;
          break;
         } else {
          $t$7 = $761;
         }
        }
        (_strcpy($buf1,$157)|0); //@line 4161 "load_abc.cpp"
        $762 = (_strlen($buf1)|0); //@line 4163 "load_abc.cpp"
        $763 = (_strlen($$lcssa2130)|0); //@line 4163 "load_abc.cpp"
        $764 = (($763) + ($762))|0; //@line 4163 "load_abc.cpp"
        $765 = ($764>>>0)>(198); //@line 4163 "load_abc.cpp"
        if ($765) {
         $766 = (($t$7$lcssa) + 198)|0; //@line 4163 "load_abc.cpp"
         $767 = (($766) - ($762))|0; //@line 4163 "load_abc.cpp"
         $768 = (($p$6) + ($767)|0); //@line 4163 "load_abc.cpp"
         HEAP8[$768>>0] = 0; //@line 4163 "load_abc.cpp"
        }
        $strlenfirst38 = HEAP8[$buf1>>0]|0; //@line 4164 "load_abc.cpp"
        $769 = ($strlenfirst38<<24>>24)==(0); //@line 4164 "load_abc.cpp"
        if (!($769)) {
         $strlen = (_strlen($buf1)|0); //@line 4164 "load_abc.cpp"
         $endptr = (($buf1) + ($strlen)|0); //@line 4164 "load_abc.cpp"
         HEAP8[$endptr>>0]=32&255;HEAP8[$endptr+1>>0]=32>>8; //@line 4164 "load_abc.cpp"
        }
        (_strcat($buf1,$$lcssa2130)|0); //@line 4165 "load_abc.cpp"
        $770 = (_strlen($buf1)|0); //@line 4169 "load_abc.cpp"
        $771 = ($770>>>0)>(31); //@line 4169 "load_abc.cpp"
        if ($771) {
         HEAP8[$159>>0] = 0; //@line 4169 "load_abc.cpp"
        }
        (_strcpy($157,$buf1)|0); //@line 4170 "load_abc.cpp"
        HEAP8[$p$6>>0] = 37; //@line 4172 "load_abc.cpp"
        $4052 = $4056;$abceffect$2 = $abceffect$11093;$abceffoper$2 = $abceffoper$11095;$abcgrace$3 = $abcgrace$2;$abchornpipe$3 = $abchornpipe$11113;$abckey$3 = $abckey$2;$abcnolegato$3 = $abcnolegato$2;$abcnoslurs$4 = $abcnoslurs$3;$abcstate$6 = $abcstate$5;$abctempo$4 = $abctempo$3;$abcto$2 = $abcto$11109;$abcvol$2 = $abcvol$11104;$abcxcount$4 = $abcxcount$21097;$abcxnumber$2 = $abcxnumber$11099;$bartime$3 = $bartime$2;$ktempo$2 = $ktempo$11115;$orderlen$3 = $orderlen$2;$p$7 = $p$6;$tpd$2 = $tpd$11121;
       }
      } while(0);
      $772 = (_strncmp($p$7,526678,2)|0); //@line 4176 "load_abc.cpp"
      $773 = ($772|0)==(0); //@line 4176 "load_abc.cpp"
      if ($773) {
       $774 = ($abcstate$6|0)==(4); //@line 4177 "load_abc.cpp"
       do {
        if ($774) {
         $abcnolegato$5 = $abcnolegato$3;$abcnoslurs$5 = $abcnoslurs$4;
        } else {
         $775 = HEAP32[$776>>2]|0; //@line 4179 "load_abc.cpp"
         $777 = (_strlen($p$7)|0); //@line 3733 "load_abc.cpp"
         $778 = (($777) + -1)|0; //@line 3733 "load_abc.cpp"
         $779 = (($p$7) + ($778)|0); //@line 3733 "load_abc.cpp"
         $780 = HEAP8[$779>>0]|0; //@line 3733 "load_abc.cpp"
         $781 = ($780<<24>>24)==(92); //@line 3733 "load_abc.cpp"
         L460: do {
          if ($781) {
           $p2$026$i = 0;$pm$025$i = $p$7;
           L461: while(1) {
            $782 = (___strdup($pm$025$i)|0); //@line 3734 "load_abc.cpp"
            $783 = ($p2$026$i|0)==(0|0); //@line 3735 "load_abc.cpp"
            if (!($783)) {
             _free($p2$026$i); //@line 3735 "load_abc.cpp"
            }
            $784 = (__ZL8abc_getsP10_ABCHANDLEP6MMFILE($1,$775)|0); //@line 3739 "load_abc.cpp"
            $785 = ($784|0)==(0|0); //@line 3740 "load_abc.cpp"
            if ($785) {
             $789 = $782;
             label = 269;
             break;
            } else {
             $787 = $784;
            }
            while(1) {
             $786 = HEAP8[$787>>0]|0; //@line 3744 "load_abc.cpp"
             $788 = ($786<<24>>24)==(0); //@line 3744 "load_abc.cpp"
             L468: do {
              if ($788) {
               $pm$2$lcssa$i = $787;
              } else {
               $794 = $786;$pm$214$i = $787;
               while(1) {
                $793 = $794 << 24 >> 24; //@line 3744 "load_abc.cpp"
                $795 = (_isspace($793)|0); //@line 3744 "load_abc.cpp"
                $796 = ($795|0)==(0); //@line 3744 "load_abc.cpp"
                if ($796) {
                 $pm$2$lcssa$i = $pm$214$i;
                 break L468;
                }
                $797 = ((($pm$214$i)) + 1|0); //@line 3744 "load_abc.cpp"
                $798 = HEAP8[$797>>0]|0; //@line 3744 "load_abc.cpp"
                $799 = ($798<<24>>24)==(0); //@line 3744 "load_abc.cpp"
                if ($799) {
                 $pm$2$lcssa$i = $797;
                 break;
                } else {
                 $794 = $798;$pm$214$i = $797;
                }
               }
              }
             } while(0);
             $800 = (_strncmp($pm$2$lcssa$i,526712,2)|0); //@line 3745 "load_abc.cpp"
             $801 = ($800|0)==(0); //@line 3745 "load_abc.cpp"
             if (!($801)) {
              $pm$2$lcssa$i$lcssa = $pm$2$lcssa$i;
              break;
             }
             $802 = ((($pm$2$lcssa$i)) + 2|0); //@line 3746 "load_abc.cpp"
             $803 = HEAP8[$802>>0]|0; //@line 3746 "load_abc.cpp"
             $804 = ($803<<24>>24)==(0); //@line 3746 "load_abc.cpp"
             L474: do {
              if ($804) {
               $p2$1$lcssa$i = $802;
              } else {
               $806 = $803;$p2$117$i = $802;
               while(1) {
                $805 = $806 << 24 >> 24; //@line 3746 "load_abc.cpp"
                $807 = (_isspace($805)|0); //@line 3746 "load_abc.cpp"
                $808 = ($807|0)==(0); //@line 3746 "load_abc.cpp"
                if ($808) {
                 $p2$1$lcssa$i = $p2$117$i;
                 break L474;
                }
                $809 = ((($p2$117$i)) + 1|0); //@line 3746 "load_abc.cpp"
                $810 = HEAP8[$809>>0]|0; //@line 3746 "load_abc.cpp"
                $811 = ($810<<24>>24)==(0); //@line 3746 "load_abc.cpp"
                if ($811) {
                 $p2$1$lcssa$i = $809;
                 break;
                } else {
                 $806 = $810;$p2$117$i = $809;
                }
               }
              }
             } while(0);
             $812 = (_strncmp($p2$1$lcssa$i,526715,4)|0); //@line 3747 "load_abc.cpp"
             $813 = ($812|0)==(0); //@line 3747 "load_abc.cpp"
             do {
              if ($813) {
               $816 = ((($p2$1$lcssa$i)) + 4|0); //@line 3747 "load_abc.cpp"
               $817 = HEAP8[$816>>0]|0; //@line 3747 "load_abc.cpp"
               $818 = ($817<<24>>24)==(61); //@line 3747 "load_abc.cpp"
               if (!($818)) {
                $819 = $817 << 24 >> 24; //@line 3747 "load_abc.cpp"
                $820 = (_isspace($819)|0); //@line 3747 "load_abc.cpp"
                $821 = ($820|0)==(0); //@line 3747 "load_abc.cpp"
                if ($821) {
                 break;
                }
               }
               $822 = ((($p2$1$lcssa$i)) + 5|0); //@line 3748 "load_abc.cpp"
               $823 = HEAP8[$822>>0]|0; //@line 3748 "load_abc.cpp"
               $824 = ($823<<24>>24)==(0); //@line 3748 "load_abc.cpp"
               L484: do {
                if ($824) {
                 $p2$4$i = $822;
                } else {
                 $826 = $823;$p2$222$i = $822;
                 while(1) {
                  $825 = $826 << 24 >> 24; //@line 3748 "load_abc.cpp"
                  $827 = (_isspace($825)|0); //@line 3748 "load_abc.cpp"
                  $828 = ($827|0)==(0); //@line 3748 "load_abc.cpp"
                  if ($828) {
                   $p2$222$i$lcssa = $p2$222$i;
                   break;
                  }
                  $829 = ((($p2$222$i)) + 1|0); //@line 3748 "load_abc.cpp"
                  $830 = HEAP8[$829>>0]|0; //@line 3748 "load_abc.cpp"
                  $831 = ($830<<24>>24)==(0); //@line 3748 "load_abc.cpp"
                  if ($831) {
                   $p2$4$i = $829;
                   break L484;
                  } else {
                   $826 = $830;$p2$222$i = $829;
                  }
                 }
                 $$pr$i = HEAP8[$p2$222$i$lcssa>>0]|0; //@line 3749 "load_abc.cpp"
                 $832 = ($$pr$i<<24>>24)==(61); //@line 3749 "load_abc.cpp"
                 if ($832) {
                  $p2$2$pn$i = $p2$222$i$lcssa;
                 } else {
                  $p2$4$i = $p2$222$i$lcssa;
                  break;
                 }
                 while(1) {
                  $p2$3$i = ((($p2$2$pn$i)) + 1|0); //@line 3750 "load_abc.cpp"
                  $833 = HEAP8[$p2$3$i>>0]|0; //@line 3750 "load_abc.cpp"
                  $834 = ($833<<24>>24)==(0); //@line 3750 "load_abc.cpp"
                  if ($834) {
                   $p2$4$i = $p2$3$i;
                   break L484;
                  }
                  $835 = $833 << 24 >> 24; //@line 3750 "load_abc.cpp"
                  $836 = (_isspace($835)|0); //@line 3750 "load_abc.cpp"
                  $837 = ($836|0)==(0); //@line 3750 "load_abc.cpp"
                  if ($837) {
                   $p2$4$i = $p2$3$i;
                   break;
                  } else {
                   $p2$2$pn$i = $p2$3$i;
                  }
                 }
                }
               } while(0);
               __ZL16abc_MIDI_commandP10_ABCHANDLEPcc($1,$p2$4$i,37); //@line 3751 "load_abc.cpp"
              }
             } while(0);
             $814 = (__ZL8abc_getsP10_ABCHANDLEP6MMFILE($1,$775)|0); //@line 3739 "load_abc.cpp"
             $815 = ($814|0)==(0|0); //@line 3740 "load_abc.cpp"
             if ($815) {
              $789 = $782;
              label = 269;
              break L461;
             } else {
              $787 = $814;
             }
            }
            $838 = (_strlen($782)|0); //@line 3756 "load_abc.cpp"
            $839 = (_strlen($pm$2$lcssa$i$lcssa)|0); //@line 3756 "load_abc.cpp"
            $840 = (($839) + ($838))|0; //@line 3756 "load_abc.cpp"
            $841 = (_malloc($840)|0); //@line 3756 "load_abc.cpp"
            $842 = ($841|0)==(0|0); //@line 3757 "load_abc.cpp"
            if ($842) {
             $$lcssa2148 = $782;
             label = 290;
             break;
            }
            $846 = (_strlen($782)|0); //@line 3761 "load_abc.cpp"
            $847 = (($846) + -1)|0; //@line 3761 "load_abc.cpp"
            $848 = (($782) + ($847)|0); //@line 3761 "load_abc.cpp"
            HEAP8[$848>>0] = 0; //@line 3761 "load_abc.cpp"
            (_strcpy($841,$782)|0); //@line 3762 "load_abc.cpp"
            (_strcat($841,$pm$2$lcssa$i$lcssa)|0); //@line 3763 "load_abc.cpp"
            _free($782); //@line 3765 "load_abc.cpp"
            $849 = (_strlen($841)|0); //@line 3733 "load_abc.cpp"
            $850 = (($849) + -1)|0; //@line 3733 "load_abc.cpp"
            $851 = (($841) + ($850)|0); //@line 3733 "load_abc.cpp"
            $852 = HEAP8[$851>>0]|0; //@line 3733 "load_abc.cpp"
            $853 = ($852<<24>>24)==(92); //@line 3733 "load_abc.cpp"
            if ($853) {
             $p2$026$i = $841;$pm$025$i = $841;
            } else {
             $$0$i170 = $841;
             break L460;
            }
           }
           if ((label|0) == 269) {
            label = 0;
            $790 = (_strlen($789)|0); //@line 323 "load_abc.cpp"
            $791 = (($790) + 30)|0; //@line 323 "load_abc.cpp"
            $792 = ($791>>>0)>(255); //@line 323 "load_abc.cpp"
            if (!($792)) {
             HEAP32[$vararg_buffer36>>2] = $789; //@line 324 "load_abc.cpp"
             (_sprintf($txt$i,526681,$vararg_buffer36)|0); //@line 324 "load_abc.cpp"
             HEAP32[$vararg_buffer39>>2] = $txt$i; //@line 328 "load_abc.cpp"
             (_fprintf($155,524747,$vararg_buffer39)|0); //@line 328 "load_abc.cpp"
            }
            $$0$i170 = $789;
            break;
           }
           else if ((label|0) == 290) {
            label = 0;
            $843 = (_strlen($$lcssa2148)|0); //@line 323 "load_abc.cpp"
            $844 = (($843) + 22)|0; //@line 323 "load_abc.cpp"
            $845 = ($844>>>0)>(255); //@line 323 "load_abc.cpp"
            if (!($845)) {
             HEAP32[$vararg_buffer42>>2] = $$lcssa2148; //@line 324 "load_abc.cpp"
             (_sprintf($txt$i,526962,$vararg_buffer42)|0); //@line 324 "load_abc.cpp"
             HEAP32[$vararg_buffer45>>2] = $txt$i; //@line 328 "load_abc.cpp"
             (_fprintf($155,524747,$vararg_buffer45)|0); //@line 328 "load_abc.cpp"
            }
            $$0$i170 = $$lcssa2148;
            break;
           }
          } else {
           $$0$i170 = $p$7;
          }
         } while(0);
         $854 = ((($$0$i170)) + 2|0); //@line 4180 "load_abc.cpp"
         __ZL19abc_extractkeyvaluePcjS_jPKc($txt$i,$value$i,$854); //@line 570 "load_abc.cpp"
         $855 = (_calloc(1,66)|0); //@line 572 "load_abc.cpp"
         $856 = (___strdup($txt$i)|0); //@line 573 "load_abc.cpp"
         $857 = ((($855)) + 4|0); //@line 573 "load_abc.cpp"
         HEAP8[$857>>0]=$856&255;HEAP8[$857+1>>0]=($856>>8)&255;HEAP8[$857+2>>0]=($856>>16)&255;HEAP8[$857+3>>0]=$856>>24; //@line 573 "load_abc.cpp"
         $858 = (_strrchr($856,110)|0); //@line 574 "load_abc.cpp"
         $859 = ((($855)) + 12|0); //@line 574 "load_abc.cpp"
         HEAP8[$859>>0]=$858&255;HEAP8[$859+1>>0]=($858>>8)&255;HEAP8[$859+2>>0]=($858>>16)&255;HEAP8[$859+3>>0]=$858>>24; //@line 574 "load_abc.cpp"
         $860 = (___strdup($value$i)|0); //@line 575 "load_abc.cpp"
         $861 = ((($855)) + 8|0); //@line 575 "load_abc.cpp"
         HEAP8[$861>>0]=$860&255;HEAP8[$861+1>>0]=($860>>8)&255;HEAP8[$861+2>>0]=($860>>16)&255;HEAP8[$861+3>>0]=$860>>24; //@line 575 "load_abc.cpp"
         $862 = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24); //@line 576 "load_abc.cpp"
         HEAP8[$855>>0]=$862&255;HEAP8[$855+1>>0]=($862>>8)&255;HEAP8[$855+2>>0]=($862>>16)&255;HEAP8[$855+3>>0]=$862>>24; //@line 576 "load_abc.cpp"
         HEAP8[$1>>0]=$855&255;HEAP8[$1+1>>0]=($855>>8)&255;HEAP8[$1+2>>0]=($855>>16)&255;HEAP8[$1+3>>0]=$855>>24; //@line 577 "load_abc.cpp"
         $863 = ($$0$i170|0)==($p$7|0); //@line 4181 "load_abc.cpp"
         if ($863) {
          $abcnolegato$5 = $abcnolegato$3;$abcnoslurs$5 = $abcnoslurs$4;
          break;
         }
         _free($$0$i170); //@line 4182 "load_abc.cpp"
         $864 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4183 "load_abc.cpp"
         $865 = ($864|0)==(0|0); //@line 4183 "load_abc.cpp"
         if ($865) {
          $abcnolegato$4 = $abcnolegato$3;
         } else {
          $866 = ((($864)) + 43|0); //@line 4183 "load_abc.cpp"
          $867 = HEAP8[$866>>0]|0; //@line 4183 "load_abc.cpp"
          $868 = ($867<<24>>24)==(0); //@line 4183 "load_abc.cpp"
          $869 = $868&1; //@line 4183 "load_abc.cpp"
          $abcnolegato$4 = $869;
         }
         $870 = ($abcnolegato$4|0)==(0); //@line 4184 "load_abc.cpp"
         $$abcnoslurs$4 = $870 ? 0 : $abcnoslurs$4; //@line 4184 "load_abc.cpp"
         $abcnolegato$5 = $abcnolegato$4;$abcnoslurs$5 = $$abcnoslurs$4;
        }
       } while(0);
       HEAP8[$p$7>>0] = 37; //@line 4187 "load_abc.cpp"
       $abcnolegato$6 = $abcnolegato$5;$abcnoslurs$6 = $abcnoslurs$5;
      } else {
       $abcnolegato$6 = $abcnolegato$3;$abcnoslurs$6 = $abcnoslurs$4;
      }
      $871 = (_strncmp($p$7,526985,2)|0); //@line 4189 "load_abc.cpp"
      $872 = ($871|0)==(0); //@line 4189 "load_abc.cpp"
      if ($872) {
       $873 = ((($p$7)) + 2|0); //@line 4190 "load_abc.cpp"
       __ZL14abc_new_umacroP10_ABCHANDLEPKc($1,$873); //@line 4190 "load_abc.cpp"
       HEAP8[$p$7>>0] = 37; //@line 4191 "load_abc.cpp"
      }
      $874 = (_strncmp($p$7,526988,2)|0); //@line 4193 "load_abc.cpp"
      $875 = ($874|0)==(0); //@line 4193 "load_abc.cpp"
      if ($875) {
       HEAP8[$p$7>>0] = 37; //@line 4194 "load_abc.cpp"
      }
      $876 = (_strncmp($p$7,526991,2)|0); //@line 4196 "load_abc.cpp"
      $877 = ($876|0)==(0); //@line 4196 "load_abc.cpp"
      if ($877) {
       HEAP8[$p$7>>0] = 37; //@line 4197 "load_abc.cpp"
      }
      $878 = (_strncmp($p$7,526994,2)|0); //@line 4199 "load_abc.cpp"
      $879 = ($878|0)==(0); //@line 4199 "load_abc.cpp"
      if ($879) {
       $880 = (_strlen($p$7)|0); //@line 323 "load_abc.cpp"
       $881 = (($880) + 45)|0; //@line 323 "load_abc.cpp"
       $882 = ($881>>>0)>(255); //@line 323 "load_abc.cpp"
       if (!($882)) {
        HEAP32[$vararg_buffer48>>2] = $p$7; //@line 324 "load_abc.cpp"
        (_sprintf($txt$i,526997,$vararg_buffer48)|0); //@line 324 "load_abc.cpp"
        HEAP32[$vararg_buffer51>>2] = $txt$i; //@line 328 "load_abc.cpp"
        (_fprintf($155,524747,$vararg_buffer51)|0); //@line 328 "load_abc.cpp"
       }
       HEAP8[$p$7>>0] = 37; //@line 4201 "load_abc.cpp"
      }
      $883 = (_strncmp($p$7,527043,2)|0); //@line 4203 "load_abc.cpp"
      $884 = ($883|0)==(0); //@line 4203 "load_abc.cpp"
      if ($884) {
       $885 = (_strlen($p$7)|0); //@line 323 "load_abc.cpp"
       $886 = (($885) + 45)|0; //@line 323 "load_abc.cpp"
       $887 = ($886>>>0)>(255); //@line 323 "load_abc.cpp"
       if (!($887)) {
        HEAP32[$vararg_buffer54>>2] = $p$7; //@line 324 "load_abc.cpp"
        (_sprintf($txt$i,527046,$vararg_buffer54)|0); //@line 324 "load_abc.cpp"
        HEAP32[$vararg_buffer57>>2] = $txt$i; //@line 328 "load_abc.cpp"
        (_fprintf($155,524747,$vararg_buffer57)|0); //@line 328 "load_abc.cpp"
       }
       HEAP8[$p$7>>0] = 37; //@line 4205 "load_abc.cpp"
      }
      $888 = (_strncmp($p$7,527092,2)|0); //@line 4207 "load_abc.cpp"
      $889 = ($888|0)==(0); //@line 4207 "load_abc.cpp"
      $890 = ($abcstate$6|0)!=(4); //@line 4207 "load_abc.cpp"
      $or$cond5 = $890 & $889; //@line 4207 "load_abc.cpp"
      if ($or$cond5) {
       HEAP8[$p$7>>0] = 37; //@line 4208 "load_abc.cpp"
       $891 = ((($p$7)) + 1|0); //@line 4209 "load_abc.cpp"
       HEAP8[$891>>0] = 37; //@line 4209 "load_abc.cpp"
      }
      $892 = (_strncmp($p$7,526712,2)|0); //@line 4211 "load_abc.cpp"
      $893 = ($892|0)==(0); //@line 4211 "load_abc.cpp"
      do {
       if ($893) {
        $894 = ((($p$7)) + 2|0); //@line 4212 "load_abc.cpp"
        $895 = HEAP8[$894>>0]|0; //@line 4212 "load_abc.cpp"
        $896 = ($895<<24>>24)==(0); //@line 4212 "load_abc.cpp"
        L538: do {
         if ($896) {
          $p$8$lcssa = $894;
         } else {
          $898 = $895;$p$81079 = $894;
          while(1) {
           $897 = $898 << 24 >> 24; //@line 4212 "load_abc.cpp"
           $899 = (_isspace($897)|0); //@line 4212 "load_abc.cpp"
           $900 = ($899|0)==(0); //@line 4212 "load_abc.cpp"
           if ($900) {
            $p$8$lcssa = $p$81079;
            break L538;
           }
           $901 = ((($p$81079)) + 1|0); //@line 4212 "load_abc.cpp"
           $902 = HEAP8[$901>>0]|0; //@line 4212 "load_abc.cpp"
           $903 = ($902<<24>>24)==(0); //@line 4212 "load_abc.cpp"
           if ($903) {
            $p$8$lcssa = $901;
            break;
           } else {
            $898 = $902;$p$81079 = $901;
           }
          }
         }
        } while(0);
        $904 = (_strncmp($p$8$lcssa,527095,11)|0); //@line 4213 "load_abc.cpp"
        $905 = ($904|0)==(0); //@line 4213 "load_abc.cpp"
        L543: do {
         if ($905) {
          $906 = ((($p$8$lcssa)) + 11|0); //@line 4213 "load_abc.cpp"
          $907 = HEAP8[$906>>0]|0; //@line 4213 "load_abc.cpp"
          $908 = $907 << 24 >> 24; //@line 4213 "load_abc.cpp"
          $909 = (_isspace($908)|0); //@line 4213 "load_abc.cpp"
          $910 = ($909|0)==(0); //@line 4213 "load_abc.cpp"
          if ($910) {
           $mmsp$2 = $mmsp$11123;
           break;
          } else {
           $t$8 = 12;
          }
          while(1) {
           $911 = (($p$8$lcssa) + ($t$8)|0); //@line 4214 "load_abc.cpp"
           $912 = HEAP8[$911>>0]|0; //@line 4214 "load_abc.cpp"
           $913 = $912 << 24 >> 24; //@line 4214 "load_abc.cpp"
           $914 = (_isspace($913)|0); //@line 4214 "load_abc.cpp"
           $915 = ($914|0)==(0); //@line 4214 "load_abc.cpp"
           $916 = (($t$8) + 1)|0; //@line 4214 "load_abc.cpp"
           if ($915) {
            $$lcssa2153 = $911;
            break;
           } else {
            $t$8 = $916;
           }
          }
          $917 = HEAP8[$$lcssa2153>>0]|0; //@line 4215 "load_abc.cpp"
          $918 = ($917<<24>>24)==(0); //@line 4215 "load_abc.cpp"
          if ($918) {
           $938 = (_strlen($p$8$lcssa)|0); //@line 323 "load_abc.cpp"
           $939 = (($938) + 42)|0; //@line 323 "load_abc.cpp"
           $940 = ($939>>>0)>(255); //@line 323 "load_abc.cpp"
           if (!($940)) {
            HEAP32[$vararg_buffer72>>2] = $p$8$lcssa; //@line 324 "load_abc.cpp"
            (_sprintf($txt$i,527184,$vararg_buffer72)|0); //@line 324 "load_abc.cpp"
            HEAP32[$vararg_buffer75>>2] = $txt$i; //@line 328 "load_abc.cpp"
            (_fprintf($155,524747,$vararg_buffer75)|0); //@line 328 "load_abc.cpp"
           }
           $mmsp$2 = $mmsp$11123;
           break;
          }
          $919 = (($mmsp$11123) + 1)|0; //@line 4216 "load_abc.cpp"
          $920 = ($919|0)==(8); //@line 4217 "load_abc.cpp"
          if ($920) {
           $921 = (_strlen($$lcssa2153)|0); //@line 323 "load_abc.cpp"
           $922 = (($921) + 35)|0; //@line 323 "load_abc.cpp"
           $923 = ($922>>>0)>(255); //@line 323 "load_abc.cpp"
           if (!($923)) {
            HEAP32[$vararg_buffer60>>2] = $$lcssa2153; //@line 324 "load_abc.cpp"
            (_sprintf($txt$i,527107,$vararg_buffer60)|0); //@line 324 "load_abc.cpp"
            HEAP32[$vararg_buffer63>>2] = $txt$i; //@line 328 "load_abc.cpp"
            (_fprintf($155,524747,$vararg_buffer63)|0); //@line 328 "load_abc.cpp"
           }
           $mmsp$2 = $mmsp$11123;
           break;
          }
          $924 = (_fopen($$lcssa2153,527143)|0); //@line 459 "load_abc.cpp"
          $925 = ($924|0)==(0|0); //@line 460 "load_abc.cpp"
          do {
           if (!($925)) {
            (_fseek($924,0,2)|0); //@line 461 "load_abc.cpp"
            $926 = (_ftell($924)|0); //@line 462 "load_abc.cpp"
            $927 = (($926) + 12)|0; //@line 463 "load_abc.cpp"
            $928 = (_malloc($927)|0); //@line 463 "load_abc.cpp"
            $929 = ($928|0)==(0|0); //@line 464 "load_abc.cpp"
            if ($929) {
             break;
            }
            (_fseek($924,0,0)|0); //@line 465 "load_abc.cpp"
            $934 = ((($928)) + 12|0); //@line 466 "load_abc.cpp"
            (_fread($934,1,$926,$924)|0); //@line 466 "load_abc.cpp"
            (_fclose($924)|0); //@line 467 "load_abc.cpp"
            HEAP8[$928>>0]=$934&255;HEAP8[$928+1>>0]=($934>>8)&255;HEAP8[$928+2>>0]=($934>>16)&255;HEAP8[$928+3>>0]=$934>>24; //@line 468 "load_abc.cpp"
            $935 = ((($928)) + 4|0); //@line 469 "load_abc.cpp"
            HEAP8[$935>>0]=$926&255;HEAP8[$935+1>>0]=($926>>8)&255;HEAP8[$935+2>>0]=($926>>16)&255;HEAP8[$935+3>>0]=$926>>24; //@line 469 "load_abc.cpp"
            $936 = ((($928)) + 8|0); //@line 470 "load_abc.cpp"
            HEAP8[$936>>0]=0&255;HEAP8[$936+1>>0]=(0>>8)&255;HEAP8[$936+2>>0]=(0>>16)&255;HEAP8[$936+3>>0]=0>>24; //@line 470 "load_abc.cpp"
            $937 = (($mmstack) + ($919<<2)|0); //@line 4221 "load_abc.cpp"
            HEAP32[$937>>2] = $928; //@line 4221 "load_abc.cpp"
            $mmsp$2 = $919;
            break L543;
           }
          } while(0);
          $930 = (($mmstack) + ($919<<2)|0); //@line 4221 "load_abc.cpp"
          HEAP32[$930>>2] = 0; //@line 4221 "load_abc.cpp"
          $931 = (_strlen($$lcssa2153)|0); //@line 323 "load_abc.cpp"
          $932 = (($931) + 38)|0; //@line 323 "load_abc.cpp"
          $933 = ($932>>>0)>(255); //@line 323 "load_abc.cpp"
          if (!($933)) {
           HEAP32[$vararg_buffer66>>2] = $$lcssa2153; //@line 324 "load_abc.cpp"
           (_sprintf($txt$i,527145,$vararg_buffer66)|0); //@line 324 "load_abc.cpp"
           HEAP32[$vararg_buffer69>>2] = $txt$i; //@line 328 "load_abc.cpp"
           (_fprintf($155,524747,$vararg_buffer69)|0); //@line 328 "load_abc.cpp"
          }
          $mmsp$2 = $mmsp$11123;
         } else {
          $mmsp$2 = $mmsp$11123;
         }
        } while(0);
        $941 = (_strncmp($p$8$lcssa,526715,4)|0); //@line 4230 "load_abc.cpp"
        $942 = ($941|0)==(0); //@line 4230 "load_abc.cpp"
        do {
         if ($942) {
          $943 = ((($p$8$lcssa)) + 4|0); //@line 4230 "load_abc.cpp"
          $944 = HEAP8[$943>>0]|0; //@line 4230 "load_abc.cpp"
          $945 = ($944<<24>>24)==(61); //@line 4230 "load_abc.cpp"
          if ($945) {
           if (!($890)) {
            $abcnolegato$8 = $abcnolegato$6;$abcnoslurs$7 = $abcnoslurs$6;$p$12 = $p$8$lcssa;
            break;
           }
          } else {
           $946 = $944 << 24 >> 24; //@line 4230 "load_abc.cpp"
           $947 = (_isspace($946)|0); //@line 4230 "load_abc.cpp"
           $948 = ($947|0)!=(0); //@line 4230 "load_abc.cpp"
           $or$cond7 = $890 & $948; //@line 4230 "load_abc.cpp"
           if (!($or$cond7)) {
            $abcnolegato$8 = $abcnolegato$6;$abcnoslurs$7 = $abcnoslurs$6;$p$12 = $p$8$lcssa;
            break;
           }
          }
          $949 = ((($p$8$lcssa)) + 5|0); //@line 4231 "load_abc.cpp"
          $950 = HEAP8[$949>>0]|0; //@line 4231 "load_abc.cpp"
          $951 = ($950<<24>>24)==(0); //@line 4231 "load_abc.cpp"
          L574: do {
           if ($951) {
            $p$11 = $949;
           } else {
            $953 = $950;$p$91085 = $949;
            while(1) {
             $952 = $953 << 24 >> 24; //@line 4231 "load_abc.cpp"
             $954 = (_isspace($952)|0); //@line 4231 "load_abc.cpp"
             $955 = ($954|0)==(0); //@line 4231 "load_abc.cpp"
             if ($955) {
              $p$91085$lcssa = $p$91085;
              break;
             }
             $956 = ((($p$91085)) + 1|0); //@line 4231 "load_abc.cpp"
             $957 = HEAP8[$956>>0]|0; //@line 4231 "load_abc.cpp"
             $958 = ($957<<24>>24)==(0); //@line 4231 "load_abc.cpp"
             if ($958) {
              $p$11 = $956;
              break L574;
             } else {
              $953 = $957;$p$91085 = $956;
             }
            }
            $$pr760 = HEAP8[$p$91085$lcssa>>0]|0; //@line 4232 "load_abc.cpp"
            $959 = ($$pr760<<24>>24)==(61); //@line 4232 "load_abc.cpp"
            if ($959) {
             $p$9$pn = $p$91085$lcssa;
            } else {
             $p$11 = $p$91085$lcssa;
             break;
            }
            while(1) {
             $p$10 = ((($p$9$pn)) + 1|0); //@line 4233 "load_abc.cpp"
             $960 = HEAP8[$p$10>>0]|0; //@line 4233 "load_abc.cpp"
             $961 = ($960<<24>>24)==(0); //@line 4233 "load_abc.cpp"
             if ($961) {
              $p$11 = $p$10;
              break L574;
             }
             $962 = $960 << 24 >> 24; //@line 4233 "load_abc.cpp"
             $963 = (_isspace($962)|0); //@line 4233 "load_abc.cpp"
             $964 = ($963|0)==(0); //@line 4233 "load_abc.cpp"
             if ($964) {
              $p$11 = $p$10;
              break;
             } else {
              $p$9$pn = $p$10;
             }
            }
           }
          } while(0);
          __ZL16abc_MIDI_commandP10_ABCHANDLEPcc($1,$p$11,37); //@line 4234 "load_abc.cpp"
          $965 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4235 "load_abc.cpp"
          $966 = ($965|0)==(0|0); //@line 4235 "load_abc.cpp"
          if ($966) {
           $abcnolegato$7 = $abcnolegato$6;
          } else {
           $967 = ((($965)) + 43|0); //@line 4235 "load_abc.cpp"
           $968 = HEAP8[$967>>0]|0; //@line 4235 "load_abc.cpp"
           $969 = ($968<<24>>24)==(0); //@line 4235 "load_abc.cpp"
           $970 = $969&1; //@line 4235 "load_abc.cpp"
           $abcnolegato$7 = $970;
          }
          $971 = ($abcnolegato$7|0)==(0); //@line 4236 "load_abc.cpp"
          $$abcnoslurs$6 = $971 ? 0 : $abcnoslurs$6; //@line 4236 "load_abc.cpp"
          $abcnolegato$8 = $abcnolegato$7;$abcnoslurs$7 = $$abcnoslurs$6;$p$12 = $p$11;
         } else {
          $abcnolegato$8 = $abcnolegato$6;$abcnoslurs$7 = $abcnoslurs$6;$p$12 = $p$8$lcssa;
         }
        } while(0);
        $972 = HEAP8[$p$12>>0]|0; //@line 4238 "load_abc.cpp"
        $973 = ($972<<24>>24)==(0); //@line 4238 "load_abc.cpp"
        if ($973) {
         $abcnolegato$9 = $abcnolegato$8;$abcnoslurs$8 = $abcnoslurs$7;$mmsp$3 = $mmsp$2;$p$13 = $p$12;
         break;
        }
        HEAP8[$p$12>>0] = 37; //@line 4238 "load_abc.cpp"
        $abcnolegato$9 = $abcnolegato$8;$abcnoslurs$8 = $abcnoslurs$7;$mmsp$3 = $mmsp$2;$p$13 = $p$12;
       } else {
        $abcnolegato$9 = $abcnolegato$6;$abcnoslurs$8 = $abcnoslurs$6;$mmsp$3 = $mmsp$11123;$p$13 = $p$7;
       }
      } while(0);
      $974 = ($abcstate$6|0)==(3); //@line 4240 "load_abc.cpp"
      if ($974) {
       $979 = HEAP8[$p$13>>0]|0; //@line 4241 "load_abc.cpp"
       $980 = ($979<<24>>24)==(80); //@line 4241 "load_abc.cpp"
       do {
        if ($980) {
         $981 = ((($p$13)) + 1|0); //@line 4241 "load_abc.cpp"
         $982 = HEAP8[$981>>0]|0; //@line 4241 "load_abc.cpp"
         $983 = ($982<<24>>24)==(58); //@line 4241 "load_abc.cpp"
         if (!($983)) {
          $p$16 = $p$13;
          break;
         }
         $984 = HEAP32[$abcparts>>2]|0; //@line 4242 "load_abc.cpp"
         $985 = ($984|0)==(0|0); //@line 4242 "load_abc.cpp"
         L594: do {
          if ($985) {
           $p$15 = $p$13;
          } else {
           $986 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 4244 "load_abc.cpp"
           $987 = ($986|0)==(0|0); //@line 4244 "load_abc.cpp"
           if ($987) {
            $988 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
            $989 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
            $990 = ((($988)) + 20|0); //@line 917 "load_abc.cpp"
            HEAP8[$990>>0]=$989&255;HEAP8[$990+1>>0]=($989>>8)&255;HEAP8[$990+2>>0]=($989>>16)&255;HEAP8[$990+3>>0]=$989>>24; //@line 917 "load_abc.cpp"
            HEAP8[$41>>0]=$988&255;HEAP8[$41+1>>0]=($988>>8)&255;HEAP8[$41+2>>0]=($988>>16)&255;HEAP8[$41+3>>0]=$988>>24; //@line 4245 "load_abc.cpp"
            $991 = ((($988)) + 36|0); //@line 4246 "load_abc.cpp"
            HEAP8[$991>>0] = 0; //@line 4246 "load_abc.cpp"
            $$pr761 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 4248 "load_abc.cpp"
            $992 = ($$pr761|0)==(0|0); //@line 4248 "load_abc.cpp"
            if ($992) {
             $1008 = 0;$999 = 0;
            } else {
             $993 = $$pr761;
             label = 358;
            }
           } else {
            $993 = $986;
            label = 358;
           }
           if ((label|0) == 358) {
            label = 0;
            $$idx101 = ((($993)) + 8|0);
            $$idx101$val = HEAPU8[$$idx101>>0]|(HEAPU8[$$idx101+1>>0]<<8)|(HEAPU8[$$idx101+2>>0]<<16)|(HEAPU8[$$idx101+3>>0]<<24);
            $994 = ($$idx101$val|0)==(0|0); //@line 2926 "load_abc.cpp"
            if ($994) {
             $tracktime$0$i184 = 0;
            } else {
             $995 = ((($$idx101$val)) + 4|0); //@line 2926 "load_abc.cpp"
             $996 = HEAPU8[$995>>0]|(HEAPU8[$995+1>>0]<<8)|(HEAPU8[$995+2>>0]<<16)|(HEAPU8[$995+3>>0]<<24); //@line 2926 "load_abc.cpp"
             $tracktime$0$i184 = $996;
            }
            $997 = HEAP32[299861]|0; //@line 2928 "load_abc.cpp"
            $998 = ($tracktime$0$i184>>>0)<($997>>>0); //@line 2928 "load_abc.cpp"
            $$tracktime$0$i185 = $998 ? $997 : $tracktime$0$i184; //@line 2928 "load_abc.cpp"
            $1008 = $993;$999 = $$tracktime$0$i185;
           }
           HEAP8[$31>>0]=$999&255;HEAP8[$31+1>>0]=($999>>8)&255;HEAP8[$31+2>>0]=($999>>16)&255;HEAP8[$31+3>>0]=$999>>24; //@line 4248 "load_abc.cpp"
           $1000 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
           HEAP8[$1000>>0]=0&255;HEAP8[$1000+1>>0]=(0>>8)&255;HEAP8[$1000+2>>0]=(0>>16)&255;HEAP8[$1000+3>>0]=0>>24; //@line 540 "load_abc.cpp"
           $1001 = ((($1000)) + 4|0); //@line 541 "load_abc.cpp"
           HEAP8[$1001>>0]=$999&255;HEAP8[$1001+1>>0]=($999>>8)&255;HEAP8[$1001+2>>0]=($999>>16)&255;HEAP8[$1001+3>>0]=$999>>24; //@line 541 "load_abc.cpp"
           $1002 = ((($1000)) + 8|0); //@line 543 "load_abc.cpp"
           HEAP8[$1002>>0] = 1; //@line 543 "load_abc.cpp"
           $1003 = ((($1000)) + 9|0); //@line 543 "load_abc.cpp"
           HEAP8[$1003>>0] = 80; //@line 543 "load_abc.cpp"
           $1004 = ((($1000)) + 10|0); //@line 543 "load_abc.cpp"
           HEAP8[$1004>>0]=0&255;HEAP8[$1004+1>>0]=(0>>8)&255;HEAP8[$1004+2>>0]=(0>>16)&255;HEAP8[$1004+3>>0]=0>>24; //@line 543 "load_abc.cpp"
           $1005 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
           $1006 = ((($1000)) + 14|0); //@line 544 "load_abc.cpp"
           HEAP8[$1006>>0] = $1005; //@line 544 "load_abc.cpp"
           $1007 = ((($1000)) + 15|0); //@line 545 "load_abc.cpp"
           HEAP8[$1007>>0] = 0; //@line 545 "load_abc.cpp"
           __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($1008,$1000); //@line 1153 "load_abc.cpp"
           $1009 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4250 "load_abc.cpp"
           $1010 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 1928 "load_abc.cpp"
           $1011 = ($1010*12288)|0; //@line 339 "load_abc.cpp"
           $1012 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 1929 "load_abc.cpp"
           $1013 = ($1012|0)==(0|0); //@line 1929 "load_abc.cpp"
           do {
            if ($1013) {
             $$0$i$i195 = 0;
            } else {
             $1014 = ((($1012)) + 4|0); //@line 1929 "load_abc.cpp"
             $1015 = HEAPU8[$1014>>0]|(HEAPU8[$1014+1>>0]<<8)|(HEAPU8[$1014+2>>0]<<16)|(HEAPU8[$1014+3>>0]<<24); //@line 1929 "load_abc.cpp"
             $1016 = ($1015|0)==(0|0); //@line 1929 "load_abc.cpp"
             if ($1016) {
              $$0$i$i195 = 0;
              break;
             } else {
              $dt$03$i$i187 = 0;$e$02$i$i188 = $1015;
             }
             while(1) {
              $1017 = ((($e$02$i$i188)) + 4|0); //@line 1931 "load_abc.cpp"
              $1018 = HEAPU8[$1017>>0]|(HEAPU8[$1017+1>>0]<<8)|(HEAPU8[$1017+2>>0]<<16)|(HEAPU8[$1017+3>>0]<<24); //@line 1931 "load_abc.cpp"
              $1019 = ($1018>>>0)>($1009>>>0); //@line 1931 "load_abc.cpp"
              if ($1019) {
               $dt$0$lcssa$i$i191 = $dt$03$i$i187;
               break;
              }
              $1020 = ((($e$02$i$i188)) + 8|0); //@line 1932 "load_abc.cpp"
              $1021 = HEAP8[$1020>>0]|0; //@line 1932 "load_abc.cpp"
              $1022 = ($1021<<24>>24)==(1); //@line 1932 "load_abc.cpp"
              do {
               if ($1022) {
                $1023 = ((($e$02$i$i188)) + 9|0); //@line 1932 "load_abc.cpp"
                $1024 = HEAP8[$1023>>0]|0; //@line 1932 "load_abc.cpp"
                $1025 = ($1024<<24>>24)==(80); //@line 1932 "load_abc.cpp"
                if (!($1025)) {
                 $dt$1$i$i190 = $dt$03$i$i187;
                 break;
                }
                $1026 = (($1018) + ($dt$03$i$i187))|0; //@line 1933 "load_abc.cpp"
                $1027 = (($1026>>>0) % ($1011>>>0))&-1; //@line 1934 "load_abc.cpp"
                $1028 = ($1027|0)==(0); //@line 1934 "load_abc.cpp"
                if ($1028) {
                 $dt$1$i$i190 = $dt$03$i$i187;
                 break;
                }
                $1029 = (($1026) + ($1011))|0; //@line 1935 "load_abc.cpp"
                $1030 = (($1029>>>0) % ($1011>>>0))&-1; //@line 1937 "load_abc.cpp"
                $1031 = (($dt$03$i$i187) + ($1011))|0; //@line 1937 "load_abc.cpp"
                $1032 = (($1031) - ($1030))|0; //@line 1938 "load_abc.cpp"
                $dt$1$i$i190 = $1032;
               } else {
                $dt$1$i$i190 = $dt$03$i$i187;
               }
              } while(0);
              $1033 = HEAPU8[$e$02$i$i188>>0]|(HEAPU8[$e$02$i$i188+1>>0]<<8)|(HEAPU8[$e$02$i$i188+2>>0]<<16)|(HEAPU8[$e$02$i$i188+3>>0]<<24); //@line 1931 "load_abc.cpp"
              $1034 = ($1033|0)==(0|0); //@line 1931 "load_abc.cpp"
              if ($1034) {
               $dt$0$lcssa$i$i191 = $dt$1$i$i190;
               break;
              } else {
               $dt$03$i$i187 = $dt$1$i$i190;$e$02$i$i188 = $1033;
              }
             }
             $1035 = (($dt$0$lcssa$i$i191) + ($1009))|0; //@line 1942 "load_abc.cpp"
             $phitmp$i192 = (($1035>>>0) / 192)&-1; //@line 1942 "load_abc.cpp"
             $phitmp1$i193 = $phitmp$i192 >>> 6; //@line 1942 "load_abc.cpp"
             $$0$i$i195 = $phitmp1$i193;
            }
           } while(0);
           $1036 = (($$0$i$i195>>>0) / ($1010>>>0))&-1; //@line 1947 "load_abc.cpp"
           $1037 = HEAP8[1278234]|0; //@line 4251 "load_abc.cpp"
           $1038 = ($1037<<24>>24)==(32); //@line 4251 "load_abc.cpp"
           $1039 = $1036&255; //@line 4252 "load_abc.cpp"
           L616: do {
            if ($1038) {
             HEAP8[$101>>0] = $1039; //@line 4252 "load_abc.cpp"
             $1040 = HEAP32[$abcparts>>2]|0; //@line 4253 "load_abc.cpp"
             $1041 = ($1040|0)==(0|0); //@line 4253 "load_abc.cpp"
             if ($1041) {
              $t$10$ph = $1036;
              break;
             } else {
              $t$91089 = 0;
             }
             while(1) {
              $1042 = (($partpat) + ($t$91089<<1)|0); //@line 4255 "load_abc.cpp"
              $1043 = HEAP8[$1042>>0]|0; //@line 4255 "load_abc.cpp"
              $1044 = (((($partpat) + ($t$91089<<1)|0)) + 1|0); //@line 4255 "load_abc.cpp"
              $1045 = HEAP8[$1044>>0]|0; //@line 4255 "load_abc.cpp"
              $1046 = ($1043&255)<($1045&255); //@line 4255 "load_abc.cpp"
              if ($1046) {
               $t$10$ph = $t$91089;
               break L616;
              }
              $1047 = (($t$91089) + 1)|0; //@line 4254 "load_abc.cpp"
              $1048 = ($1047>>>0)<(26); //@line 4254 "load_abc.cpp"
              if ($1048) {
               $t$91089 = $1047;
              } else {
               $$lcssa2157 = $1047;
               break;
              }
             }
             $1049 = ($$lcssa2157|0)==(26); //@line 4256 "load_abc.cpp"
             if (!($1049)) {
              $t$10$ph = $$lcssa2157;
              break;
             }
             $1050 = (_strlen($1040)|0); //@line 323 "load_abc.cpp"
             $1051 = (($1050) + 27)|0; //@line 323 "load_abc.cpp"
             $1052 = ($1051>>>0)>(255); //@line 323 "load_abc.cpp"
             if ($1052) {
              $1054 = $1040;
             } else {
              HEAP32[$vararg_buffer78>>2] = $1040; //@line 324 "load_abc.cpp"
              (_sprintf($txt$i,526650,$vararg_buffer78)|0); //@line 324 "load_abc.cpp"
              HEAP32[$vararg_buffer81>>2] = $txt$i; //@line 328 "load_abc.cpp"
              (_fprintf($155,524747,$vararg_buffer81)|0); //@line 328 "load_abc.cpp"
              $$pre1522 = HEAP32[$abcparts>>2]|0; //@line 1820 "load_abc.cpp"
              $1054 = $$pre1522;
             }
             $1053 = ($1054|0)==(0|0); //@line 1820 "load_abc.cpp"
             if (!($1053)) {
              _free($1054); //@line 1820 "load_abc.cpp"
             }
             HEAP32[$abcparts>>2] = 0; //@line 1822 "load_abc.cpp"
             $t$10$ph = 26;
            } else {
             $1055 = $1037 << 24 >> 24; //@line 4251 "load_abc.cpp"
             $1056 = (($1055) + -65)|0; //@line 4263 "load_abc.cpp"
             $1057 = (((($partpat) + ($1056<<1)|0)) + 1|0); //@line 4263 "load_abc.cpp"
             HEAP8[$1057>>0] = $1039; //@line 4263 "load_abc.cpp"
             $t$10$ph = $1036;
            }
           } while(0);
           $p$14 = $p$13;
           L631: while(1) {
            $1058 = ((($p$14)) + 2|0); //@line 4265 "load_abc.cpp"
            $1059 = HEAP8[$1058>>0]|0; //@line 4265 "load_abc.cpp"
            switch ($1059<<24>>24) {
            case 32: case 46:  {
             break;
            }
            default: {
             $$lcssa2158 = $1058;$$lcssa2159 = $1059;$p$14$lcssa = $p$14;
             break L631;
            }
            }
            $1060 = ((($p$14)) + 1|0); //@line 4265 "load_abc.cpp"
            $p$14 = $1060;
           }
           $1061 = $$lcssa2159 << 24 >> 24; //@line 4266 "load_abc.cpp"
           $1062 = (_isupper($1061)|0); //@line 4266 "load_abc.cpp"
           $1063 = ($1062|0)==(0); //@line 4266 "load_abc.cpp"
           do {
            if ($1063) {
             HEAP8[1278234] = 32; //@line 4267 "load_abc.cpp"
             $1064 = $t$10$ph&255; //@line 4271 "load_abc.cpp"
             $1068 = $1064;
            } else {
             $1065 = HEAP8[$$lcssa2158>>0]|0; //@line 4267 "load_abc.cpp"
             HEAP8[1278234] = $1065; //@line 4267 "load_abc.cpp"
             $1066 = ($1065<<24>>24)==(32); //@line 4270 "load_abc.cpp"
             $1067 = $t$10$ph&255; //@line 4271 "load_abc.cpp"
             if ($1066) {
              $1068 = $1067;
              break;
             }
             $1069 = $1065 << 24 >> 24; //@line 4270 "load_abc.cpp"
             $1070 = (($1069) + -65)|0; //@line 4273 "load_abc.cpp"
             $1071 = (($partpat) + ($1070<<1)|0); //@line 4273 "load_abc.cpp"
             HEAP8[$1071>>0] = $1067; //@line 4273 "load_abc.cpp"
             $p$15 = $p$14$lcssa;
             break L594;
            }
           } while(0);
           HEAP8[$100>>0] = $1068; //@line 4271 "load_abc.cpp"
           $p$15 = $p$14$lcssa;
          }
         } while(0);
         HEAP8[$p$15>>0] = 37; //@line 4275 "load_abc.cpp"
         $p$16 = $p$15;
        } else {
         $p$16 = $p$13;
        }
       } while(0);
       $1072 = HEAPU8[$40>>0]|(HEAPU8[$40+1>>0]<<8)|(HEAPU8[$40+2>>0]<<16)|(HEAPU8[$40+3>>0]<<24); //@line 4277 "load_abc.cpp"
       $1073 = ($1072|0)==(0); //@line 4277 "load_abc.cpp"
       $1074 = ($tpd$2|0)!=(0|0); //@line 4277 "load_abc.cpp"
       $or$cond12 = $1074 | $1073; //@line 4277 "load_abc.cpp"
       do {
        if ($or$cond12) {
         $tpd$3 = $tpd$2;
         label = 393;
        } else {
         $1075 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 4278 "load_abc.cpp"
         $1076 = ($1075|0)==(0|0); //@line 4279 "load_abc.cpp"
         if ($1076) {
          $tpd$4 = 0;
          break;
         }
         $1077 = ((($1075)) + 44|0); //@line 4280 "load_abc.cpp"
         $1078 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,$1077,9)|0); //@line 4280 "load_abc.cpp"
         $1079 = HEAPU8[$44>>0]|(HEAPU8[$44+1>>0]<<8)|(HEAPU8[$44+2>>0]<<16)|(HEAPU8[$44+3>>0]<<24); //@line 4281 "load_abc.cpp"
         $1080 = $1079&255; //@line 4281 "load_abc.cpp"
         $1081 = ((($1078)) + 42|0); //@line 4281 "load_abc.cpp"
         HEAP8[$1081>>0] = $1080; //@line 4281 "load_abc.cpp"
         $1082 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4282 "load_abc.cpp"
         $1083 = HEAPU8[$45>>0]|(HEAPU8[$45+1>>0]<<8)|(HEAPU8[$45+2>>0]<<16)|(HEAPU8[$45+3>>0]<<24); //@line 4282 "load_abc.cpp"
         $1084 = HEAPU8[$47>>0]|(HEAPU8[$47+1>>0]<<8)|(HEAPU8[$47+2>>0]<<16)|(HEAPU8[$47+3>>0]<<24); //@line 4282 "load_abc.cpp"
         __ZL17abc_add_dronenoteP10_ABCHANDLEP9_ABCTRACKjii($1078,$1082,$1083,$1084); //@line 4282 "load_abc.cpp"
         $1085 = ((($1078)) + 44|0); //@line 4283 "load_abc.cpp"
         $1086 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,$1085,10)|0); //@line 4283 "load_abc.cpp"
         $1087 = HEAPU8[$44>>0]|(HEAPU8[$44+1>>0]<<8)|(HEAPU8[$44+2>>0]<<16)|(HEAPU8[$44+3>>0]<<24); //@line 4284 "load_abc.cpp"
         $1088 = $1087&255; //@line 4284 "load_abc.cpp"
         $1089 = ((($1086)) + 42|0); //@line 4284 "load_abc.cpp"
         HEAP8[$1089>>0] = $1088; //@line 4284 "load_abc.cpp"
         $1090 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4285 "load_abc.cpp"
         $1091 = HEAPU8[$46>>0]|(HEAPU8[$46+1>>0]<<8)|(HEAPU8[$46+2>>0]<<16)|(HEAPU8[$46+3>>0]<<24); //@line 4285 "load_abc.cpp"
         $1092 = HEAPU8[$48>>0]|(HEAPU8[$48+1>>0]<<8)|(HEAPU8[$48+2>>0]<<16)|(HEAPU8[$48+3>>0]<<24); //@line 4285 "load_abc.cpp"
         __ZL17abc_add_dronenoteP10_ABCHANDLEP9_ABCTRACKjii($1086,$1090,$1091,$1092); //@line 4285 "load_abc.cpp"
         $tpd$3 = $1086;
         label = 393;
        }
       } while(0);
       do {
        if ((label|0) == 393) {
         label = 0;
         $1093 = ($tpd$3|0)==(0|0); //@line 4288 "load_abc.cpp"
         if ($1093) {
          $tpd$4 = 0;
          break;
         }
         $1094 = HEAPU8[$40>>0]|(HEAPU8[$40+1>>0]<<8)|(HEAPU8[$40+2>>0]<<16)|(HEAPU8[$40+3>>0]<<24); //@line 4288 "load_abc.cpp"
         $1095 = ($1094|0)==(0); //@line 4288 "load_abc.cpp"
         if (!($1095)) {
          $tpd$4 = $tpd$3;
          break;
         }
         $1096 = ((($tpd$3)) + 44|0); //@line 4289 "load_abc.cpp"
         $1097 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,$1096,9)|0); //@line 4289 "load_abc.cpp"
         $1098 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4290 "load_abc.cpp"
         $1099 = ((($1097)) + 42|0); //@line 1174 "load_abc.cpp"
         $1100 = HEAP8[$1099>>0]|0; //@line 1174 "load_abc.cpp"
         $1101 = $1100&255; //@line 1174 "load_abc.cpp"
         $1102 = (_pat_gmtosmp($1101)|0); //@line 1174 "load_abc.cpp"
         $1103 = $1102&255; //@line 1174 "load_abc.cpp"
         $1104 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
         HEAP8[$1104>>0]=0&255;HEAP8[$1104+1>>0]=(0>>8)&255;HEAP8[$1104+2>>0]=(0>>16)&255;HEAP8[$1104+3>>0]=0>>24; //@line 540 "load_abc.cpp"
         $1105 = ((($1104)) + 4|0); //@line 541 "load_abc.cpp"
         HEAP8[$1105>>0]=$1098&255;HEAP8[$1105+1>>0]=($1098>>8)&255;HEAP8[$1105+2>>0]=($1098>>16)&255;HEAP8[$1105+3>>0]=$1098>>24; //@line 541 "load_abc.cpp"
         $1106 = ((($1104)) + 8|0); //@line 543 "load_abc.cpp"
         HEAP8[$1106>>0] = 0; //@line 543 "load_abc.cpp"
         $1107 = ((($1104)) + 9|0); //@line 543 "load_abc.cpp"
         HEAP8[$1107>>0] = 0; //@line 543 "load_abc.cpp"
         $1108 = ((($1104)) + 10|0); //@line 543 "load_abc.cpp"
         HEAP8[$1108>>0] = $1103; //@line 543 "load_abc.cpp"
         $1109 = ((($1104)) + 11|0); //@line 543 "load_abc.cpp"
         HEAP8[$1109>>0] = 0; //@line 543 "load_abc.cpp"
         $1110 = ((($1104)) + 12|0); //@line 543 "load_abc.cpp"
         HEAP8[$1110>>0] = 0; //@line 543 "load_abc.cpp"
         $1111 = ((($1104)) + 13|0); //@line 543 "load_abc.cpp"
         HEAP8[$1111>>0] = 0; //@line 543 "load_abc.cpp"
         $1112 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
         $1113 = ((($1104)) + 14|0); //@line 544 "load_abc.cpp"
         HEAP8[$1113>>0] = $1112; //@line 544 "load_abc.cpp"
         $1114 = ((($1104)) + 15|0); //@line 545 "load_abc.cpp"
         HEAP8[$1114>>0] = 0; //@line 545 "load_abc.cpp"
         __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($1097,$1104); //@line 1179 "load_abc.cpp"
         $1115 = ((($1097)) + 44|0); //@line 4291 "load_abc.cpp"
         $1116 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,$1115,10)|0); //@line 4291 "load_abc.cpp"
         $1117 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4292 "load_abc.cpp"
         $1118 = ((($1116)) + 42|0); //@line 1174 "load_abc.cpp"
         $1119 = HEAP8[$1118>>0]|0; //@line 1174 "load_abc.cpp"
         $1120 = $1119&255; //@line 1174 "load_abc.cpp"
         $1121 = (_pat_gmtosmp($1120)|0); //@line 1174 "load_abc.cpp"
         $1122 = $1121&255; //@line 1174 "load_abc.cpp"
         $1123 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
         HEAP8[$1123>>0]=0&255;HEAP8[$1123+1>>0]=(0>>8)&255;HEAP8[$1123+2>>0]=(0>>16)&255;HEAP8[$1123+3>>0]=0>>24; //@line 540 "load_abc.cpp"
         $1124 = ((($1123)) + 4|0); //@line 541 "load_abc.cpp"
         HEAP8[$1124>>0]=$1117&255;HEAP8[$1124+1>>0]=($1117>>8)&255;HEAP8[$1124+2>>0]=($1117>>16)&255;HEAP8[$1124+3>>0]=$1117>>24; //@line 541 "load_abc.cpp"
         $1125 = ((($1123)) + 8|0); //@line 543 "load_abc.cpp"
         HEAP8[$1125>>0] = 0; //@line 543 "load_abc.cpp"
         $1126 = ((($1123)) + 9|0); //@line 543 "load_abc.cpp"
         HEAP8[$1126>>0] = 0; //@line 543 "load_abc.cpp"
         $1127 = ((($1123)) + 10|0); //@line 543 "load_abc.cpp"
         HEAP8[$1127>>0] = $1122; //@line 543 "load_abc.cpp"
         $1128 = ((($1123)) + 11|0); //@line 543 "load_abc.cpp"
         HEAP8[$1128>>0] = 0; //@line 543 "load_abc.cpp"
         $1129 = ((($1123)) + 12|0); //@line 543 "load_abc.cpp"
         HEAP8[$1129>>0] = 0; //@line 543 "load_abc.cpp"
         $1130 = ((($1123)) + 13|0); //@line 543 "load_abc.cpp"
         HEAP8[$1130>>0] = 0; //@line 543 "load_abc.cpp"
         $1131 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
         $1132 = ((($1123)) + 14|0); //@line 544 "load_abc.cpp"
         HEAP8[$1132>>0] = $1131; //@line 544 "load_abc.cpp"
         $1133 = ((($1123)) + 15|0); //@line 545 "load_abc.cpp"
         HEAP8[$1133>>0] = 0; //@line 545 "load_abc.cpp"
         __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($1116,$1123); //@line 1179 "load_abc.cpp"
         $tpd$4 = 0;
        }
       } while(0);
       $1134 = HEAPU8[$38>>0]|(HEAPU8[$38+1>>0]<<8)|(HEAPU8[$38+2>>0]<<16)|(HEAPU8[$38+3>>0]<<24); //@line 4295 "load_abc.cpp"
       $1135 = ($1134|0)==(0); //@line 4295 "load_abc.cpp"
       L650: do {
        if ($1135) {
         label = 405;
        } else {
         $1136 = HEAPU8[$43>>0]|(HEAPU8[$43+1>>0]<<8)|(HEAPU8[$43+2>>0]<<16)|(HEAPU8[$43+3>>0]<<24); //@line 4295 "load_abc.cpp"
         $1137 = ($1136|0)==(0|0); //@line 4295 "load_abc.cpp"
         if (!($1137)) {
          break;
         }
         $1138 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 4296 "load_abc.cpp"
         HEAP8[$43>>0]=$1138&255;HEAP8[$43+1>>0]=($1138>>8)&255;HEAP8[$43+2>>0]=($1138>>16)&255;HEAP8[$43+3>>0]=$1138>>24; //@line 4296 "load_abc.cpp"
         $1139 = ($1138|0)==(0|0); //@line 4297 "load_abc.cpp"
         if ($1139) {
          break;
         }
         $1140 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4297 "load_abc.cpp"
         $1141 = ((($1138)) + 44|0); //@line 1616 "load_abc.cpp"
         $1142 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,$1141,8)|0); //@line 1616 "load_abc.cpp"
         $1143 = ((($1142)) + 8|0); //@line 1590 "load_abc.cpp"
         $1144 = HEAPU8[$1143>>0]|(HEAPU8[$1143+1>>0]<<8)|(HEAPU8[$1143+2>>0]<<16)|(HEAPU8[$1143+3>>0]<<24); //@line 1590 "load_abc.cpp"
         $1145 = ($1144|0)==(0|0); //@line 1591 "load_abc.cpp"
         do {
          if (!($1145)) {
           $1146 = ((($1144)) + 4|0); //@line 1591 "load_abc.cpp"
           $1147 = HEAPU8[$1146>>0]|(HEAPU8[$1146+1>>0]<<8)|(HEAPU8[$1146+2>>0]<<16)|(HEAPU8[$1146+3>>0]<<24); //@line 1591 "load_abc.cpp"
           $1148 = ($1147|0)==($1140|0); //@line 1591 "load_abc.cpp"
           if ($1148) {
            label = 405;
            break L650;
           }
           $1149 = ((($1144)) + 8|0); //@line 1592 "load_abc.cpp"
           $1150 = HEAP8[$1149>>0]|0; //@line 1592 "load_abc.cpp"
           $1151 = ($1150<<24>>24)==(1); //@line 1592 "load_abc.cpp"
           if (!($1151)) {
            break;
           }
           $1152 = ((($1144)) + 9|0); //@line 1592 "load_abc.cpp"
           $1153 = HEAP8[$1152>>0]|0; //@line 1592 "load_abc.cpp"
           $1154 = ($1153<<24>>24)==(83); //@line 1592 "load_abc.cpp"
           if (!($1154)) {
            break;
           }
           HEAP8[$1146>>0]=$1140&255;HEAP8[$1146+1>>0]=($1140>>8)&255;HEAP8[$1146+2>>0]=($1140>>16)&255;HEAP8[$1146+3>>0]=$1140>>24; //@line 1593 "load_abc.cpp"
           label = 405;
           break L650;
          }
         } while(0);
         $1155 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
         HEAP8[$1155>>0]=0&255;HEAP8[$1155+1>>0]=(0>>8)&255;HEAP8[$1155+2>>0]=(0>>16)&255;HEAP8[$1155+3>>0]=0>>24; //@line 540 "load_abc.cpp"
         $1156 = ((($1155)) + 4|0); //@line 541 "load_abc.cpp"
         HEAP8[$1156>>0]=$1140&255;HEAP8[$1156+1>>0]=($1140>>8)&255;HEAP8[$1156+2>>0]=($1140>>16)&255;HEAP8[$1156+3>>0]=$1140>>24; //@line 541 "load_abc.cpp"
         $1157 = ((($1155)) + 8|0); //@line 543 "load_abc.cpp"
         HEAP8[$1157>>0] = 1; //@line 543 "load_abc.cpp"
         $1158 = ((($1155)) + 9|0); //@line 543 "load_abc.cpp"
         HEAP8[$1158>>0] = 83; //@line 543 "load_abc.cpp"
         $1159 = ((($1155)) + 10|0); //@line 543 "load_abc.cpp"
         HEAP8[$1159>>0]=0&255;HEAP8[$1159+1>>0]=(0>>8)&255;HEAP8[$1159+2>>0]=(0>>16)&255;HEAP8[$1159+3>>0]=0>>24; //@line 543 "load_abc.cpp"
         $1160 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
         $1161 = ((($1155)) + 14|0); //@line 544 "load_abc.cpp"
         HEAP8[$1161>>0] = $1160; //@line 544 "load_abc.cpp"
         $1162 = ((($1155)) + 15|0); //@line 545 "load_abc.cpp"
         HEAP8[$1162>>0] = 0; //@line 545 "load_abc.cpp"
         __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($1142,$1155); //@line 1600 "load_abc.cpp"
         label = 405;
        }
       } while(0);
       do {
        if ((label|0) == 405) {
         label = 0;
         $$pr766 = HEAPU8[$43>>0]|(HEAPU8[$43+1>>0]<<8)|(HEAPU8[$43+2>>0]<<16)|(HEAPU8[$43+3>>0]<<24); //@line 4299 "load_abc.cpp"
         $1163 = ($$pr766|0)==(0|0); //@line 4299 "load_abc.cpp"
         if ($1163) {
          break;
         }
         $$pre1523 = HEAPU8[$38>>0]|(HEAPU8[$38+1>>0]<<8)|(HEAPU8[$38+2>>0]<<16)|(HEAPU8[$38+3>>0]<<24); //@line 4299 "load_abc.cpp"
         $1164 = ($$pre1523|0)==(0); //@line 4299 "load_abc.cpp"
         if (!($1164)) {
          break;
         }
         HEAP8[$43>>0]=0&255;HEAP8[$43+1>>0]=(0>>8)&255;HEAP8[$43+2>>0]=(0>>16)&255;HEAP8[$43+3>>0]=0>>24; //@line 4299 "load_abc.cpp"
        }
       } while(0);
       $1165 = HEAP8[$p$16>>0]|0; //@line 4300 "load_abc.cpp"
       switch ($1165<<24>>24) {
       case 37: case 0:  {
        $tpd$1$be = $tpd$4;
        break;
       }
       default: {
        $$lcssa2160 = $4052;$$lcssa2179 = $1165;$abceffect$2$lcssa = $abceffect$2;$abceffoper$2$lcssa = $abceffoper$2;$abcgrace$3$lcssa = $abcgrace$3;$abchornpipe$3$lcssa = $abchornpipe$3;$abckey$3$lcssa = $abckey$3;$abcnolegato$9$lcssa = $abcnolegato$9;$abcnoslurs$8$lcssa = $abcnoslurs$8;$abctempo$4$lcssa = $abctempo$4;$abcto$2$lcssa = $abcto$2;$abcvol$2$lcssa = $abcvol$2;$abcxcount$4$lcssa = $abcxcount$4;$abcxnumber$2$lcssa = $abcxnumber$2;$bartime$3$lcssa = $bartime$3;$ktempo$2$lcssa = $ktempo$2;$mmsp$3$lcssa = $mmsp$3;$orderlen$3$lcssa = $orderlen$3;$p$16$lcssa = $p$16;$tpd$4$lcssa = $tpd$4;
        break L94;
       }
       }
      } else {
       $tpd$1$be = $tpd$2;
      }
      $975 = (($mmstack) + ($mmsp$3<<2)|0); //@line 3908 "load_abc.cpp"
      $976 = HEAP32[$975>>2]|0; //@line 3908 "load_abc.cpp"
      $977 = (__ZL8abc_getsP10_ABCHANDLEP6MMFILE($1,$976)|0); //@line 3908 "load_abc.cpp"
      $978 = ($977|0)==(0|0); //@line 3908 "load_abc.cpp"
      if ($978) {
       $$lcssa928 = $4052;$$lcssa948 = $975;$$ph892$lcssa = $$ph8921258;$$ph893$lcssa = $$ph8931259;$$ph894$lcssa = $$ph8941260;$abcarpeggio$1$ph$lcssa = $abcarpeggio$1$ph1273;$abceffect$1$lcssa = $abceffect$2;$abceffoper$1$lcssa = $abceffoper$2;$abcfermata$1$ph$lcssa = $abcfermata$1$ph1272;$abcgrace$1$lcssa = $abcgrace$3;$abchornpipe$1$lcssa = $abchornpipe$3;$abckey$1$lcssa = $abckey$3;$abcnolegato$1$lcssa = $abcnolegato$9;$abcnoslurs$1$lcssa = $abcnoslurs$8;$abcstate$2$lcssa = $abcstate$6;$abctempo$1$lcssa = $abctempo$4;$abcto$1$lcssa = $abcto$2;$abcvol$1$lcssa = $abcvol$2;$abcxcount$2$lcssa = $abcxcount$4;$abcxnumber$1$lcssa = $abcxnumber$2;$bartime$1$lcssa = $bartime$3;$ch0$1$ph$lcssa = $ch0$1$ph1261;$cnotediv$1$ph$lcssa = $cnotediv$1$ph1277;$cnotelen$1$ph$lcssa = $cnotelen$1$ph1276;$ktempo$1$lcssa = $ktempo$2;$mmsp$1$lcssa = $mmsp$3;$orderlen$1$lcssa = $orderlen$3;$tpd$1$lcssa = $tpd$1$be;
       break L91;
      } else {
       $4050 = $977;$4051 = $4052;$776 = $975;$abceffect$11093 = $abceffect$2;$abceffoper$11095 = $abceffoper$2;$abcgrace$11117 = $abcgrace$3;$abchornpipe$11113 = $abchornpipe$3;$abckey$11101 = $abckey$3;$abcnolegato$11107 = $abcnolegato$9;$abcnoslurs$11106 = $abcnoslurs$8;$abcstate$21092 = $abcstate$6;$abctempo$11111 = $abctempo$4;$abcto$11109 = $abcto$2;$abcvol$11104 = $abcvol$2;$abcxcount$21097 = $abcxcount$4;$abcxnumber$11099 = $abcxnumber$2;$bartime$11119 = $bartime$3;$ktempo$11115 = $ktempo$2;$mmsp$11123 = $mmsp$3;$orderlen$11090 = $orderlen$3;$tpd$11121 = $tpd$1$be;
      }
     }
     $1169 = $$lcssa2179;$1626 = $$ph8941260;$1628 = $$ph8931259;$1630 = $$ph8921258;$1637 = $$lcssa2160;$abcarpeggio$21229 = $abcarpeggio$1$ph1273;$abceffect$31222 = $abceffect$2$lcssa;$abceffoper$31223 = $abceffoper$2$lcssa;$abcfermata$21228 = $abcfermata$1$ph1272;$abcgrace$41234 = $abcgrace$3$lcssa;$abckey$41224 = $abckey$3$lcssa;$abcnolegato$101227 = $abcnolegato$9$lcssa;$abcnoslurs$91226 = $abcnoslurs$8$lcssa;$abcstate$71221 = 3;$abctempo$51231 = $abctempo$4$lcssa;$abcto$31230 = $abcto$2$lcssa;$abcvol$31225 = $abcvol$2$lcssa;$bartime$41235 = $bartime$3$lcssa;$ch0$21220 = 32;$cnotediv$21233 = $cnotediv$1$ph1277;$cnotelen$21232 = $cnotelen$1$ph1276;$p$16$pn = $p$16$lcssa;$pp$01219 = 0;
     while(1) {
      $1166 = ((($p$16$pn)) + 1|0); //@line 4304 "load_abc.cpp"
      $1167 = ($pp$01219|0)==(0|0); //@line 4305 "load_abc.cpp"
      L669: do {
       if ($1167) {
        $1168 = $1169 << 24 >> 24; //@line 4305 "load_abc.cpp"
        $1170 = (_isalpha($1168)|0); //@line 4305 "load_abc.cpp"
        $1171 = ($1170|0)==(0); //@line 4305 "load_abc.cpp"
        if ($1171) {
         $ch$0 = $1169;$p$18 = $1166;$pp$1 = 0;
        } else {
         $1172 = HEAP8[$1166>>0]|0; //@line 4305 "load_abc.cpp"
         $1173 = ($1172<<24>>24)==(58); //@line 4305 "load_abc.cpp"
         if ($1173) {
          $ch$0 = $1169;$p$18 = $1166;$pp$1 = 0;
          break;
         }
         $mp$01181 = HEAPU8[$160>>0]|(HEAPU8[$160+1>>0]<<8)|(HEAPU8[$160+2>>0]<<16)|(HEAPU8[$160+3>>0]<<24); //@line 4306 "load_abc.cpp"
         $1174 = ($mp$01181|0)==(0|0); //@line 4306 "load_abc.cpp"
         if ($1174) {
          $ch$0 = $1169;$p$18 = $1166;$pp$1 = 0;
          break;
         } else {
          $mp$01182 = $mp$01181;
         }
         while(1) {
          $1175 = ((($mp$01182)) + 4|0); //@line 4307 "load_abc.cpp"
          $1176 = HEAPU8[$1175>>0]|(HEAPU8[$1175+1>>0]<<8)|(HEAPU8[$1175+2>>0]<<16)|(HEAPU8[$1175+3>>0]<<24); //@line 4307 "load_abc.cpp"
          $1177 = HEAP8[$1176>>0]|0; //@line 4307 "load_abc.cpp"
          $1178 = ($1169<<24>>24)==($1177<<24>>24); //@line 4307 "load_abc.cpp"
          if ($1178) {
           $mp$01182$lcssa = $mp$01182;
           break;
          }
          $mp$0 = HEAPU8[$mp$01182>>0]|(HEAPU8[$mp$01182+1>>0]<<8)|(HEAPU8[$mp$01182+2>>0]<<16)|(HEAPU8[$mp$01182+3>>0]<<24); //@line 4306 "load_abc.cpp"
          $1184 = ($mp$0|0)==(0|0); //@line 4306 "load_abc.cpp"
          if ($1184) {
           $ch$0 = $1169;$p$18 = $1166;$pp$1 = 0;
           break L669;
          } else {
           $mp$01182 = $mp$0;
          }
         }
         $1179 = ((($mp$01182$lcssa)) + 8|0); //@line 4309 "load_abc.cpp"
         $1180 = HEAPU8[$1179>>0]|(HEAPU8[$1179+1>>0]<<8)|(HEAPU8[$1179+2>>0]<<16)|(HEAPU8[$1179+3>>0]<<24); //@line 4309 "load_abc.cpp"
         $1181 = HEAP8[$1180>>0]|0; //@line 4310 "load_abc.cpp"
         $1182 = ($1181<<24>>24)==(0); //@line 4311 "load_abc.cpp"
         $1183 = ((($1180)) + 1|0); //@line 4311 "load_abc.cpp"
         $$85 = $1182 ? $1180 : $1183; //@line 4311 "load_abc.cpp"
         $ch$0 = $1181;$p$18 = $$85;$pp$1 = $1166;
        }
       } else {
        $ch$0 = $1169;$p$18 = $1166;$pp$1 = $pp$01219;
       }
      } while(0);
      $1185 = $ch$0 << 24 >> 24; //@line 4316 "load_abc.cpp"
      L678: do {
       switch ($1185|0) {
       case 37:  {
        $p$19 = $p$18;
        while(1) {
         $1186 = HEAP8[$p$19>>0]|0; //@line 4319 "load_abc.cpp"
         $1187 = ($1186<<24>>24)==(0); //@line 4319 "load_abc.cpp"
         $1188 = ((($p$19)) + 1|0); //@line 4319 "load_abc.cpp"
         if ($1187) {
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$19;
          break;
         } else {
          $p$19 = $1188;
         }
        }
        break;
       }
       case 91:  {
        $1189 = HEAP8[$p$18>>0]|0; //@line 4323 "load_abc.cpp"
        $1190 = ($1189<<24>>24)==(124); //@line 4323 "load_abc.cpp"
        if ($1190) {
         $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$18;
        } else {
         $1191 = (_strncmp($p$18,525427,2)|0); //@line 4324 "load_abc.cpp"
         $1192 = ($1191|0)==(0); //@line 4324 "load_abc.cpp"
         if ($1192) {
          $t$11 = 2;
          while(1) {
           $1193 = (($p$18) + ($t$11)|0); //@line 4325 "load_abc.cpp"
           $1194 = HEAP8[$1193>>0]|0; //@line 4325 "load_abc.cpp"
           $1195 = $1194 << 24 >> 24; //@line 4325 "load_abc.cpp"
           $1196 = (_isspace($1195)|0); //@line 4325 "load_abc.cpp"
           $1197 = ($1196|0)==(0); //@line 4325 "load_abc.cpp"
           $1198 = (($t$11) + 1)|0; //@line 4325 "load_abc.cpp"
           if ($1197) {
            $$lcssa2250 = $1193;
            break;
           } else {
            $t$11 = $1198;
           }
          }
          $1199 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,$$lcssa2250,0)|0); //@line 4326 "load_abc.cpp"
          HEAP8[$41>>0]=$1199&255;HEAP8[$41+1>>0]=($1199>>8)&255;HEAP8[$41+2>>0]=($1199>>16)&255;HEAP8[$41+3>>0]=$1199>>24; //@line 4326 "load_abc.cpp"
          $p$20 = $p$18;
          L891: while(1) {
           $1200 = HEAP8[$p$20>>0]|0; //@line 4327 "load_abc.cpp"
           switch ($1200<<24>>24) {
           case 93: case 0:  {
            $p$20$lcssa = $p$20;
            break L891;
            break;
           }
           default: {
           }
           }
           $1201 = ((($p$20)) + 1|0); //@line 4327 "load_abc.cpp"
           $p$20 = $1201;
          }
          $1202 = (47112 + ($abckey$41224<<2)|0); //@line 4330 "load_abc.cpp"
          $1203 = HEAP32[$1202>>2]|0; //@line 4330 "load_abc.cpp"
          HEAP32[$vararg_buffer84>>2] = $1203; //@line 4330 "load_abc.cpp"
          $vararg_ptr87 = ((($vararg_buffer84)) + 4|0); //@line 4330 "load_abc.cpp"
          HEAP32[$vararg_ptr87>>2] = $1203; //@line 4330 "load_abc.cpp"
          (_sprintf($barsig,526499,$vararg_buffer84)|0); //@line 4330 "load_abc.cpp"
          $1204 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4331 "load_abc.cpp"
          $$idx = ((($1204)) + 8|0);
          $$idx$val = HEAPU8[$$idx>>0]|(HEAPU8[$$idx+1>>0]<<8)|(HEAPU8[$$idx+2>>0]<<16)|(HEAPU8[$$idx+3>>0]<<24);
          $1205 = ($$idx$val|0)==(0|0); //@line 2926 "load_abc.cpp"
          if ($1205) {
           $tracktime$0$i203 = 0;
          } else {
           $1206 = ((($$idx$val)) + 4|0); //@line 2926 "load_abc.cpp"
           $1207 = HEAPU8[$1206>>0]|(HEAPU8[$1206+1>>0]<<8)|(HEAPU8[$1206+2>>0]<<16)|(HEAPU8[$1206+3>>0]<<24); //@line 2926 "load_abc.cpp"
           $tracktime$0$i203 = $1207;
          }
          $1208 = HEAP32[299861]|0; //@line 2928 "load_abc.cpp"
          $1209 = ($tracktime$0$i203>>>0)<($1208>>>0); //@line 2928 "load_abc.cpp"
          $$tracktime$0$i204 = $1209 ? $1208 : $tracktime$0$i203; //@line 2928 "load_abc.cpp"
          HEAP8[$31>>0]=$$tracktime$0$i204&255;HEAP8[$31+1>>0]=($$tracktime$0$i204>>8)&255;HEAP8[$31+2>>0]=($$tracktime$0$i204>>16)&255;HEAP8[$31+3>>0]=$$tracktime$0$i204>>24; //@line 4331 "load_abc.cpp"
          $1210 = ((($1204)) + 41|0); //@line 4333 "load_abc.cpp"
          $1211 = HEAP8[$1210>>0]|0; //@line 4333 "load_abc.cpp"
          $1212 = $1211&255; //@line 4333 "load_abc.cpp"
          $1213 = ((($1204)) + 43|0); //@line 4334 "load_abc.cpp"
          $1214 = HEAP8[$1213>>0]|0; //@line 4334 "load_abc.cpp"
          $1215 = ($1214<<24>>24)==(0); //@line 4334 "load_abc.cpp"
          $1216 = $1215&1; //@line 4334 "load_abc.cpp"
          $abcnoslurs$9$ = $1215 ? $abcnoslurs$91226 : 0; //@line 4335 "load_abc.cpp"
          $4057 = 0;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = 0;$abckey$5 = $abckey$41224;$abcnolegato$14 = $1216;$abcnoslurs$12 = $abcnoslurs$9$;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $1212;$bartime$5 = $$tracktime$0$i204;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$20$lcssa;
          break L678;
         }
         $1217 = (_strncmp($p$18,525430,2)|0); //@line 4338 "load_abc.cpp"
         $1218 = ($1217|0)==(0); //@line 4338 "load_abc.cpp"
         if ($1218) {
          $1219 = ((($p$18)) + 2|0); //@line 4339 "load_abc.cpp"
          $1220 = (__ZL7ABC_KeyPKc($1219)|0); //@line 4339 "load_abc.cpp"
          $1221 = (47112 + ($1220<<2)|0); //@line 4340 "load_abc.cpp"
          $1222 = HEAP32[$1221>>2]|0; //@line 4340 "load_abc.cpp"
          HEAP32[$vararg_buffer88>>2] = $1222; //@line 4340 "load_abc.cpp"
          $vararg_ptr91 = ((($vararg_buffer88)) + 4|0); //@line 4340 "load_abc.cpp"
          HEAP32[$vararg_ptr91>>2] = $1222; //@line 4340 "load_abc.cpp"
          (_sprintf($barsig,526499,$vararg_buffer88)|0); //@line 4340 "load_abc.cpp"
          $$0$i206 = $1219;
          while(1) {
           $1223 = HEAP8[$$0$i206>>0]|0; //@line 2917 "load_abc.cpp"
           $1224 = $1223 << 24 >> 24; //@line 2917 "load_abc.cpp"
           $1225 = (_isspace($1224)|0); //@line 2917 "load_abc.cpp"
           $1226 = ($1225|0)==(0); //@line 2917 "load_abc.cpp"
           $1227 = ((($$0$i206)) + 1|0); //@line 2917 "load_abc.cpp"
           if ($1226) {
            $$0$i206$lcssa = $$0$i206;
            break;
           } else {
            $$0$i206 = $1227;
           }
          }
          $1228 = HEAP8[$$0$i206$lcssa>>0]|0; //@line 2918 "load_abc.cpp"
          $1229 = ($1228<<24>>24)==(0); //@line 2918 "load_abc.cpp"
          L904: do {
           if ($1229) {
            $$2$i210 = $$0$i206$lcssa;
           } else {
            $$12$i208 = $$0$i206$lcssa;$1231 = $1228;
            while(1) {
             $1230 = $1231 << 24 >> 24; //@line 2918 "load_abc.cpp"
             $1232 = (_isspace($1230)|0); //@line 2918 "load_abc.cpp"
             $1233 = ($1232|0)==(0); //@line 2918 "load_abc.cpp"
             if (!($1233)) {
              $$2$i210 = $$12$i208;
              break L904;
             }
             $1234 = HEAP8[$$12$i208>>0]|0; //@line 2918 "load_abc.cpp"
             $1235 = ($1234<<24>>24)==(93); //@line 2918 "load_abc.cpp"
             if ($1235) {
              $$2$i210 = $$12$i208;
              break L904;
             }
             $1236 = ((($$12$i208)) + 1|0); //@line 2918 "load_abc.cpp"
             $1237 = HEAP8[$1236>>0]|0; //@line 2918 "load_abc.cpp"
             $1238 = ($1237<<24>>24)==(0); //@line 2918 "load_abc.cpp"
             if ($1238) {
              $$2$i210 = $1236;
              break;
             } else {
              $$12$i208 = $1236;$1231 = $1237;
             }
            }
           }
          } while(0);
          while(1) {
           $1239 = HEAP8[$$2$i210>>0]|0; //@line 2919 "load_abc.cpp"
           $1240 = $1239 << 24 >> 24; //@line 2919 "load_abc.cpp"
           $1241 = (_isspace($1240)|0); //@line 2919 "load_abc.cpp"
           $1242 = ($1241|0)==(0); //@line 2919 "load_abc.cpp"
           $1243 = ((($$2$i210)) + 1|0); //@line 2919 "load_abc.cpp"
           if ($1242) {
            $$2$i210$lcssa = $$2$i210;
            break;
           } else {
            $$2$i210 = $1243;
           }
          }
          $1244 = (__ZL13abc_transposePKc($$2$i210$lcssa)|0); //@line 4342 "load_abc.cpp"
          HEAP8[$37>>0]=$1244&255;HEAP8[$37+1>>0]=($1244>>8)&255;HEAP8[$37+2>>0]=($1244>>16)&255;HEAP8[$37+3>>0]=$1244>>24; //@line 4342 "load_abc.cpp"
          $p$21 = $$2$i210$lcssa;
          while(1) {
           $1245 = HEAP8[$p$21>>0]|0; //@line 4343 "load_abc.cpp"
           switch ($1245<<24>>24) {
           case 93: case 0:  {
            $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $1220;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$21;
            break L678;
            break;
           }
           default: {
           }
           }
           $1246 = ((($p$21)) + 1|0); //@line 4343 "load_abc.cpp"
           $p$21 = $1246;
          }
         }
         $1247 = (_strncmp($p$18,525136,2)|0); //@line 4346 "load_abc.cpp"
         $1248 = ($1247|0)==(0); //@line 4346 "load_abc.cpp"
         if ($1248) {
          $1249 = ((($p$18)) + 2|0); //@line 4347 "load_abc.cpp"
          $1250 = (_strncmp($1249,525139,4)|0); //@line 3312 "load_abc.cpp"
          $1251 = ($1250|0)==(0); //@line 3312 "load_abc.cpp"
          do {
           if ($1251) {
            HEAP32[$mnotelen>>2] = 1; //@line 3313 "load_abc.cpp"
            HEAP32[$mnotediv>>2] = 1; //@line 3314 "load_abc.cpp"
            $p$22 = $p$18;
           } else {
            $1252 = (_strncmp($1249,525144,2)|0); //@line 3317 "load_abc.cpp"
            $1253 = ($1252|0)==(0); //@line 3317 "load_abc.cpp"
            if ($1253) {
             HEAP32[$mnotelen>>2] = 2; //@line 3318 "load_abc.cpp"
             HEAP32[$mnotediv>>2] = 2; //@line 3319 "load_abc.cpp"
             $p$22 = $p$18;
             break;
            }
            $1254 = HEAP8[$1249>>0]|0; //@line 3322 "load_abc.cpp"
            $1255 = ($1254<<24>>24)==(67); //@line 3322 "load_abc.cpp"
            if ($1255) {
             HEAP32[$mnotelen>>2] = 4; //@line 3323 "load_abc.cpp"
             HEAP32[$mnotediv>>2] = 4; //@line 3324 "load_abc.cpp"
             $p$22 = $p$18;
             break;
            } else {
             $1256 = (__ZL11abc_getexprPKcPi($1249,$mnotelen)|0); //@line 3327 "load_abc.cpp"
             $$sum822 = (($1256) + 2)|0; //@line 3327 "load_abc.cpp"
             $1257 = (($p$18) + ($$sum822)|0); //@line 3327 "load_abc.cpp"
             HEAP32[$vararg_buffer92>>2] = $mnotediv; //@line 3328 "load_abc.cpp"
             (_sscanf($1257,525147,$vararg_buffer92)|0); //@line 3328 "load_abc.cpp"
             $p$22 = $p$18;
             break;
            }
           }
          } while(0);
          L927: while(1) {
           $1258 = HEAP8[$p$22>>0]|0; //@line 4348 "load_abc.cpp"
           switch ($1258<<24>>24) {
           case 93: case 0:  {
            $p$22$lcssa = $p$22;
            break L927;
            break;
           }
           default: {
           }
           }
           $1259 = ((($p$22)) + 1|0); //@line 4348 "load_abc.cpp"
           $p$22 = $1259;
          }
          $1260 = HEAP32[$mnotediv>>2]|0; //@line 4349 "load_abc.cpp"
          $1261 = ($1260|0)==(0); //@line 345 "load_abc.cpp"
          if ($1261) {
           $$0$i214 = 0;
          } else {
           $1262 = HEAP32[$mnotelen>>2]|0; //@line 4349 "load_abc.cpp"
           $1263 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4349 "load_abc.cpp"
           $1264 = HEAP32[299862]|0; //@line 346 "load_abc.cpp"
           $1265 = ($1262*3072)|0; //@line 346 "load_abc.cpp"
           $1266 = Math_imul($1265, $1263)|0; //@line 346 "load_abc.cpp"
           $1267 = Math_imul($1266, $1264)|0; //@line 346 "load_abc.cpp"
           $1268 = HEAP32[299863]|0; //@line 346 "load_abc.cpp"
           $1269 = Math_imul($1268, $1260)|0; //@line 346 "load_abc.cpp"
           $1270 = (($1267|0) / ($1269|0))&-1; //@line 346 "load_abc.cpp"
           $$0$i214 = $1270;
          }
          HEAP8[$7>>0]=$$0$i214&255;HEAP8[$7+1>>0]=($$0$i214>>8)&255;HEAP8[$7+2>>0]=($$0$i214>>16)&255;HEAP8[$7+3>>0]=$$0$i214>>24; //@line 4349 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$22$lcssa;
          break L678;
         }
         $1271 = (_strncmp($p$18,525153,2)|0); //@line 4352 "load_abc.cpp"
         $1272 = ($1271|0)==(0); //@line 4352 "load_abc.cpp"
         if ($1272) {
          $1273 = HEAP32[$abcparts>>2]|0; //@line 4353 "load_abc.cpp"
          $1274 = ($1273|0)==(0|0); //@line 4353 "load_abc.cpp"
          L937: do {
           if ($1274) {
            $p$24 = $p$18;
           } else {
            $1275 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 4355 "load_abc.cpp"
            $1276 = ($1275|0)==(0|0); //@line 4355 "load_abc.cpp"
            do {
             if ($1276) {
              $1277 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 1928 "load_abc.cpp"
              $$0$i$i224 = 0;$1314 = $1277;
             } else {
              $1278 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4356 "load_abc.cpp"
              $1279 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
              HEAP8[$1279>>0]=0&255;HEAP8[$1279+1>>0]=(0>>8)&255;HEAP8[$1279+2>>0]=(0>>16)&255;HEAP8[$1279+3>>0]=0>>24; //@line 540 "load_abc.cpp"
              $1280 = ((($1279)) + 4|0); //@line 541 "load_abc.cpp"
              HEAP8[$1280>>0]=$1278&255;HEAP8[$1280+1>>0]=($1278>>8)&255;HEAP8[$1280+2>>0]=($1278>>16)&255;HEAP8[$1280+3>>0]=$1278>>24; //@line 541 "load_abc.cpp"
              $1281 = ((($1279)) + 8|0); //@line 543 "load_abc.cpp"
              HEAP8[$1281>>0] = 1; //@line 543 "load_abc.cpp"
              $1282 = ((($1279)) + 9|0); //@line 543 "load_abc.cpp"
              HEAP8[$1282>>0] = 80; //@line 543 "load_abc.cpp"
              $1283 = ((($1279)) + 10|0); //@line 543 "load_abc.cpp"
              HEAP8[$1283>>0]=0&255;HEAP8[$1283+1>>0]=(0>>8)&255;HEAP8[$1283+2>>0]=(0>>16)&255;HEAP8[$1283+3>>0]=0>>24; //@line 543 "load_abc.cpp"
              $1284 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
              $1285 = ((($1279)) + 14|0); //@line 544 "load_abc.cpp"
              HEAP8[$1285>>0] = $1284; //@line 544 "load_abc.cpp"
              $1286 = ((($1279)) + 15|0); //@line 545 "load_abc.cpp"
              HEAP8[$1286>>0] = 0; //@line 545 "load_abc.cpp"
              __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($1275,$1279); //@line 1153 "load_abc.cpp"
              $$pr768 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 1929 "load_abc.cpp"
              $1287 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4357 "load_abc.cpp"
              $1288 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 1928 "load_abc.cpp"
              $1289 = ($1288*12288)|0; //@line 339 "load_abc.cpp"
              $1290 = ($$pr768|0)==(0|0); //@line 1929 "load_abc.cpp"
              if ($1290) {
               $$0$i$i224 = 0;$1314 = $1288;
               break;
              }
              $1291 = ((($$pr768)) + 4|0); //@line 1929 "load_abc.cpp"
              $1292 = HEAPU8[$1291>>0]|(HEAPU8[$1291+1>>0]<<8)|(HEAPU8[$1291+2>>0]<<16)|(HEAPU8[$1291+3>>0]<<24); //@line 1929 "load_abc.cpp"
              $1293 = ($1292|0)==(0|0); //@line 1929 "load_abc.cpp"
              if ($1293) {
               $$0$i$i224 = 0;$1314 = $1288;
               break;
              } else {
               $dt$03$i$i216 = 0;$e$02$i$i217 = $1292;
              }
              while(1) {
               $1294 = ((($e$02$i$i217)) + 4|0); //@line 1931 "load_abc.cpp"
               $1295 = HEAPU8[$1294>>0]|(HEAPU8[$1294+1>>0]<<8)|(HEAPU8[$1294+2>>0]<<16)|(HEAPU8[$1294+3>>0]<<24); //@line 1931 "load_abc.cpp"
               $1296 = ($1295>>>0)>($1287>>>0); //@line 1931 "load_abc.cpp"
               if ($1296) {
                $dt$0$lcssa$i$i220 = $dt$03$i$i216;
                break;
               }
               $1297 = ((($e$02$i$i217)) + 8|0); //@line 1932 "load_abc.cpp"
               $1298 = HEAP8[$1297>>0]|0; //@line 1932 "load_abc.cpp"
               $1299 = ($1298<<24>>24)==(1); //@line 1932 "load_abc.cpp"
               do {
                if ($1299) {
                 $1300 = ((($e$02$i$i217)) + 9|0); //@line 1932 "load_abc.cpp"
                 $1301 = HEAP8[$1300>>0]|0; //@line 1932 "load_abc.cpp"
                 $1302 = ($1301<<24>>24)==(80); //@line 1932 "load_abc.cpp"
                 if (!($1302)) {
                  $dt$1$i$i219 = $dt$03$i$i216;
                  break;
                 }
                 $1303 = (($1295) + ($dt$03$i$i216))|0; //@line 1933 "load_abc.cpp"
                 $1304 = (($1303>>>0) % ($1289>>>0))&-1; //@line 1934 "load_abc.cpp"
                 $1305 = ($1304|0)==(0); //@line 1934 "load_abc.cpp"
                 if ($1305) {
                  $dt$1$i$i219 = $dt$03$i$i216;
                  break;
                 }
                 $1306 = (($1303) + ($1289))|0; //@line 1935 "load_abc.cpp"
                 $1307 = (($1306>>>0) % ($1289>>>0))&-1; //@line 1937 "load_abc.cpp"
                 $1308 = (($dt$03$i$i216) + ($1289))|0; //@line 1937 "load_abc.cpp"
                 $1309 = (($1308) - ($1307))|0; //@line 1938 "load_abc.cpp"
                 $dt$1$i$i219 = $1309;
                } else {
                 $dt$1$i$i219 = $dt$03$i$i216;
                }
               } while(0);
               $1310 = HEAPU8[$e$02$i$i217>>0]|(HEAPU8[$e$02$i$i217+1>>0]<<8)|(HEAPU8[$e$02$i$i217+2>>0]<<16)|(HEAPU8[$e$02$i$i217+3>>0]<<24); //@line 1931 "load_abc.cpp"
               $1311 = ($1310|0)==(0|0); //@line 1931 "load_abc.cpp"
               if ($1311) {
                $dt$0$lcssa$i$i220 = $dt$1$i$i219;
                break;
               } else {
                $dt$03$i$i216 = $dt$1$i$i219;$e$02$i$i217 = $1310;
               }
              }
              $1312 = (($dt$0$lcssa$i$i220) + ($1287))|0; //@line 1942 "load_abc.cpp"
              $phitmp$i221 = (($1312>>>0) / 192)&-1; //@line 1942 "load_abc.cpp"
              $phitmp1$i222 = $phitmp$i221 >>> 6; //@line 1942 "load_abc.cpp"
              $$0$i$i224 = $phitmp1$i222;$1314 = $1288;
             }
            } while(0);
            $1313 = (($$0$i$i224>>>0) / ($1314>>>0))&-1; //@line 1947 "load_abc.cpp"
            $1315 = HEAP8[1278234]|0; //@line 4358 "load_abc.cpp"
            $1316 = ($1315<<24>>24)==(32); //@line 4358 "load_abc.cpp"
            $1317 = $1313&255; //@line 4359 "load_abc.cpp"
            if ($1316) {
             HEAP8[$101>>0] = $1317; //@line 4359 "load_abc.cpp"
             $p$23 = $p$18;
            } else {
             $1318 = $1315 << 24 >> 24; //@line 4358 "load_abc.cpp"
             $1319 = (($1318) + -65)|0; //@line 4361 "load_abc.cpp"
             $1320 = (((($partpat) + ($1319<<1)|0)) + 1|0); //@line 4361 "load_abc.cpp"
             HEAP8[$1320>>0] = $1317; //@line 4361 "load_abc.cpp"
             $p$23 = $p$18;
            }
            while(1) {
             $1321 = ((($p$23)) + 2|0); //@line 4363 "load_abc.cpp"
             $1322 = HEAP8[$1321>>0]|0; //@line 4363 "load_abc.cpp"
             $1323 = $1322 << 24 >> 24; //@line 4363 "load_abc.cpp"
             $1324 = (_isspace($1323)|0); //@line 4363 "load_abc.cpp"
             $1325 = ($1324|0)==(0); //@line 4363 "load_abc.cpp"
             if ($1325) {
              $1326 = HEAP8[$1321>>0]|0; //@line 4363 "load_abc.cpp"
              $1327 = ($1326<<24>>24)==(46); //@line 4363 "load_abc.cpp"
              if (!($1327)) {
               $$lcssa2248 = $1321;$$lcssa2249 = $1326;$p$23$lcssa = $p$23;
               break;
              }
             }
             $1328 = ((($p$23)) + 1|0); //@line 4363 "load_abc.cpp"
             $p$23 = $1328;
            }
            $1329 = $$lcssa2249 << 24 >> 24; //@line 4364 "load_abc.cpp"
            $1330 = (_isupper($1329)|0); //@line 4364 "load_abc.cpp"
            $1331 = ($1330|0)==(0); //@line 4364 "load_abc.cpp"
            do {
             if ($1331) {
              HEAP8[1278234] = 32; //@line 4365 "load_abc.cpp"
             } else {
              $1332 = HEAP8[$$lcssa2248>>0]|0; //@line 4365 "load_abc.cpp"
              HEAP8[1278234] = $1332; //@line 4365 "load_abc.cpp"
              $1333 = ($1332<<24>>24)==(32); //@line 4368 "load_abc.cpp"
              if ($1333) {
               break;
              }
              $1334 = $1332 << 24 >> 24; //@line 4368 "load_abc.cpp"
              $1335 = (($1334) + -65)|0; //@line 4371 "load_abc.cpp"
              $1336 = (($partpat) + ($1335<<1)|0); //@line 4371 "load_abc.cpp"
              HEAP8[$1336>>0] = $1317; //@line 4371 "load_abc.cpp"
              $p$24 = $p$23$lcssa;
              break L937;
             }
            } while(0);
            HEAP8[$100>>0] = $1317; //@line 4369 "load_abc.cpp"
            $p$24 = $p$23$lcssa;
           }
          } while(0);
          while(1) {
           $1337 = HEAP8[$p$24>>0]|0; //@line 4373 "load_abc.cpp"
           switch ($1337<<24>>24) {
           case 93: case 0:  {
            $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$24;
            break L678;
            break;
           }
           default: {
           }
           }
           $1338 = ((($p$24)) + 1|0); //@line 4373 "load_abc.cpp"
           $p$24 = $1338;
          }
         }
         $1339 = (_strncmp($p$18,525350,2)|0); //@line 4376 "load_abc.cpp"
         $1340 = ($1339|0)==(0); //@line 4376 "load_abc.cpp"
         if ($1340) {
          $1341 = ((($p$18)) + 2|0); //@line 4377 "load_abc.cpp"
          $1342 = (__ZL17abc_extract_tempoPKci($1341,1)|0); //@line 4377 "load_abc.cpp"
          $p$25 = $p$18;
          L973: while(1) {
           $1343 = HEAP8[$p$25>>0]|0; //@line 4378 "load_abc.cpp"
           switch ($1343<<24>>24) {
           case 93: case 0:  {
            $p$25$lcssa = $p$25;
            break L973;
            break;
           }
           default: {
           }
           }
           $1344 = ((($p$25)) + 1|0); //@line 4378 "load_abc.cpp"
           $p$25 = $1344;
          }
          $1345 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 4379 "load_abc.cpp"
          $1346 = ($1345|0)==(0|0); //@line 4379 "load_abc.cpp"
          if ($1346) {
           $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $1342;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$25$lcssa;
           break L678;
          }
          $1347 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4380 "load_abc.cpp"
          $1348 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$1348>>0]=0&255;HEAP8[$1348+1>>0]=(0>>8)&255;HEAP8[$1348+2>>0]=(0>>16)&255;HEAP8[$1348+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $1349 = ((($1348)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$1349>>0]=$1347&255;HEAP8[$1349+1>>0]=($1347>>8)&255;HEAP8[$1349+2>>0]=($1347>>16)&255;HEAP8[$1349+3>>0]=$1347>>24; //@line 541 "load_abc.cpp"
          $1350 = ((($1348)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$1350>>0] = 1; //@line 543 "load_abc.cpp"
          $1351 = ((($1348)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$1351>>0] = 84; //@line 543 "load_abc.cpp"
          $1352 = ((($1348)) + 10|0); //@line 543 "load_abc.cpp"
          $1353 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $1354 = ((($1348)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$1354>>0] = $1353; //@line 544 "load_abc.cpp"
          $1355 = ((($1348)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$1355>>0] = 0; //@line 545 "load_abc.cpp"
          HEAP8[$1352>>0]=$1342&255;HEAP8[$1352+1>>0]=($1342>>8)&255;HEAP8[$1352+2>>0]=($1342>>16)&255;HEAP8[$1352+3>>0]=$1342>>24; //@line 1164 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($1345,$1348); //@line 1165 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $1342;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$25$lcssa;
          break L678;
         }
         $1356 = (_strncmp($p$18,527092,2)|0); //@line 4383 "load_abc.cpp"
         $1357 = ($1356|0)==(0); //@line 4383 "load_abc.cpp"
         L979: do {
          if ($1357) {
           $1358 = ((($p$18)) + 2|0); //@line 4384 "load_abc.cpp"
           $p$26 = $1358;
           while(1) {
            $1359 = HEAP8[$p$26>>0]|0; //@line 4384 "load_abc.cpp"
            $1360 = $1359 << 24 >> 24; //@line 4384 "load_abc.cpp"
            $1361 = (_isspace($1360)|0); //@line 4384 "load_abc.cpp"
            $1362 = ($1361|0)==(0); //@line 4384 "load_abc.cpp"
            $1363 = ((($p$26)) + 1|0); //@line 4384 "load_abc.cpp"
            if ($1362) {
             $p$26$lcssa = $p$26;
             break;
            } else {
             $p$26 = $1363;
            }
           }
           $1364 = (_strncmp($p$26$lcssa,526715,4)|0); //@line 4385 "load_abc.cpp"
           $1365 = ($1364|0)==(0); //@line 4385 "load_abc.cpp"
           do {
            if ($1365) {
             $1366 = ((($p$26$lcssa)) + 4|0); //@line 4385 "load_abc.cpp"
             $1367 = HEAP8[$1366>>0]|0; //@line 4385 "load_abc.cpp"
             $1368 = ($1367<<24>>24)==(61); //@line 4385 "load_abc.cpp"
             if ($1368) {
              $p$27 = $1366;
             } else {
              $1369 = $1367 << 24 >> 24; //@line 4385 "load_abc.cpp"
              $1370 = (_isspace($1369)|0); //@line 4385 "load_abc.cpp"
              $1371 = ($1370|0)==(0); //@line 4385 "load_abc.cpp"
              if ($1371) {
               $abcnolegato$12$ph = $abcnolegato$101227;$abcnoslurs$10$ph = $abcnoslurs$91226;$p$30$ph = $p$26$lcssa;
               break;
              } else {
               $p$27 = $1366;
              }
             }
             while(1) {
              $1372 = HEAP8[$p$27>>0]|0; //@line 4386 "load_abc.cpp"
              $1373 = $1372 << 24 >> 24; //@line 4386 "load_abc.cpp"
              $1374 = (_isspace($1373)|0); //@line 4386 "load_abc.cpp"
              $1375 = ($1374|0)==(0); //@line 4386 "load_abc.cpp"
              $1376 = ((($p$27)) + 1|0); //@line 4386 "load_abc.cpp"
              if ($1375) {
               $p$27$lcssa = $p$27;
               break;
              } else {
               $p$27 = $1376;
              }
             }
             $1377 = HEAP8[$p$27$lcssa>>0]|0; //@line 4387 "load_abc.cpp"
             $1378 = ($1377<<24>>24)==(61); //@line 4387 "load_abc.cpp"
             if ($1378) {
              $p$27$pn = $p$27$lcssa;
              while(1) {
               $p$28 = ((($p$27$pn)) + 1|0); //@line 4388 "load_abc.cpp"
               $1379 = HEAP8[$p$28>>0]|0; //@line 4388 "load_abc.cpp"
               $1380 = $1379 << 24 >> 24; //@line 4388 "load_abc.cpp"
               $1381 = (_isspace($1380)|0); //@line 4388 "load_abc.cpp"
               $1382 = ($1381|0)==(0); //@line 4388 "load_abc.cpp"
               if ($1382) {
                $p$29 = $p$28;
                break;
               } else {
                $p$27$pn = $p$28;
               }
              }
             } else {
              $p$29 = $p$27$lcssa;
             }
             __ZL16abc_MIDI_commandP10_ABCHANDLEPcc($1,$p$29,93); //@line 4389 "load_abc.cpp"
             $1383 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4390 "load_abc.cpp"
             $1384 = ($1383|0)==(0|0); //@line 4390 "load_abc.cpp"
             if ($1384) {
              $abcnolegato$11 = $abcnolegato$101227;
             } else {
              $1385 = ((($1383)) + 43|0); //@line 4390 "load_abc.cpp"
              $1386 = HEAP8[$1385>>0]|0; //@line 4390 "load_abc.cpp"
              $1387 = ($1386<<24>>24)==(0); //@line 4390 "load_abc.cpp"
              $1388 = $1387&1; //@line 4390 "load_abc.cpp"
              $abcnolegato$11 = $1388;
             }
             $1389 = ($abcnolegato$11|0)==(0); //@line 4391 "load_abc.cpp"
             $$abcnoslurs$9 = $1389 ? 0 : $abcnoslurs$91226; //@line 4391 "load_abc.cpp"
             $abcnolegato$12$ph = $abcnolegato$11;$abcnoslurs$10$ph = $$abcnoslurs$9;$p$30$ph = $p$29;
            } else {
             $abcnolegato$12$ph = $abcnolegato$101227;$abcnoslurs$10$ph = $abcnoslurs$91226;$p$30$ph = $p$26$lcssa;
            }
           } while(0);
           $p$30 = $p$30$ph;
           while(1) {
            $1390 = HEAP8[$p$30>>0]|0; //@line 4393 "load_abc.cpp"
            switch ($1390<<24>>24) {
            case 93: case 0:  {
             $1393 = $1390;$abcnolegato$13 = $abcnolegato$12$ph;$abcnoslurs$11 = $abcnoslurs$10$ph;$p$31 = $p$30;
             break L979;
             break;
            }
            default: {
            }
            }
            $1391 = ((($p$30)) + 1|0); //@line 4393 "load_abc.cpp"
            $p$30 = $1391;
           }
          } else {
           $1393 = $1189;$abcnolegato$13 = $abcnolegato$101227;$abcnoslurs$11 = $abcnoslurs$91226;$p$31 = $p$18;
          }
         } while(0);
         $1392 = ($1393<<24>>24)==(0); //@line 4395 "load_abc.cpp"
         do {
          if (!($1392)) {
           $1394 = ((($p$31)) + 1|0); //@line 4395 "load_abc.cpp"
           $1395 = HEAP8[$1394>>0]|0; //@line 4395 "load_abc.cpp"
           $1396 = ($1395<<24>>24)==(58); //@line 4395 "load_abc.cpp"
           if ($1396) {
            $1397 = $1393;$p$32 = $p$31;
            while(1) {
             switch ($1397<<24>>24) {
             case 93: case 0:  {
              $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$13;$abcnoslurs$12 = $abcnoslurs$11;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$32;
              break L678;
              break;
             }
             default: {
             }
             }
             $1398 = ((($p$32)) + 1|0); //@line 4396 "load_abc.cpp"
             $$pre1532 = HEAP8[$1398>>0]|0; //@line 4396 "load_abc.cpp"
             $1397 = $$pre1532;$p$32 = $1398;
            }
           }
           $1399 = $1393 << 24 >> 24; //@line 4399 "load_abc.cpp"
           $memchr60 = (_memchr(527227,$1399,18)|0); //@line 4399 "load_abc.cpp"
           $1400 = ($memchr60|0)==(0|0); //@line 4399 "load_abc.cpp"
           if ($1400) {
            break;
           }
           $1401 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4403 "load_abc.cpp"
           $1402 = ($1401|0)==(0|0); //@line 915 "load_abc.cpp"
           if ($1402) {
            $1403 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
            $1404 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
            $1405 = ((($1403)) + 20|0); //@line 917 "load_abc.cpp"
            HEAP8[$1405>>0]=$1404&255;HEAP8[$1405+1>>0]=($1404>>8)&255;HEAP8[$1405+2>>0]=($1404>>16)&255;HEAP8[$1405+3>>0]=$1404>>24; //@line 917 "load_abc.cpp"
            $$0$i226 = $1403;
           } else {
            $$0$i226 = $1401;
           }
           HEAP8[$41>>0]=$$0$i226&255;HEAP8[$41+1>>0]=($$0$i226>>8)&255;HEAP8[$41+2>>0]=($$0$i226>>16)&255;HEAP8[$41+3>>0]=$$0$i226>>24; //@line 4403 "load_abc.cpp"
           $1406 = ((($$0$i226)) + 36|0); //@line 1227 "load_abc.cpp"
           $1407 = HEAP8[$1406>>0]|0; //@line 1227 "load_abc.cpp"
           $tp$01$i = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 1228 "load_abc.cpp"
           $1408 = ($tp$01$i|0)==(0|0); //@line 1228 "load_abc.cpp"
           if (!($1408)) {
            $tp$02$i = $tp$01$i;
            while(1) {
             $1409 = ((($tp$02$i)) + 36|0); //@line 1229 "load_abc.cpp"
             $1410 = HEAP8[$1409>>0]|0; //@line 1229 "load_abc.cpp"
             $1411 = ($1410<<24>>24)==($1407<<24>>24); //@line 1229 "load_abc.cpp"
             if ($1411) {
              $1412 = ((($tp$02$i)) + 37|0); //@line 1229 "load_abc.cpp"
              $1413 = HEAP8[$1412>>0]|0; //@line 1229 "load_abc.cpp"
              $1414 = ((($tp$02$i)) + 38|0); //@line 1229 "load_abc.cpp"
              HEAP8[$1414>>0] = $1413; //@line 1229 "load_abc.cpp"
             }
             $tp$0$i = HEAPU8[$tp$02$i>>0]|(HEAPU8[$tp$02$i+1>>0]<<8)|(HEAPU8[$tp$02$i+2>>0]<<16)|(HEAPU8[$tp$02$i+3>>0]<<24); //@line 1228 "load_abc.cpp"
             $1415 = ($tp$0$i|0)==(0|0); //@line 1228 "load_abc.cpp"
             if ($1415) {
              break;
             } else {
              $tp$02$i = $tp$0$i;
             }
            }
           }
           $1416 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4405 "load_abc.cpp"
           $1417 = (($1416) - ($bartime$41235))|0; //@line 4405 "load_abc.cpp"
           $1418 = HEAP32[$mnotediv>>2]|0; //@line 4405 "load_abc.cpp"
           $1419 = ($1418|0)==(0); //@line 345 "load_abc.cpp"
           if ($1419) {
            $$0$i230 = 0;
           } else {
            $1420 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4405 "load_abc.cpp"
            $1421 = ($1420*3072)|0; //@line 346 "load_abc.cpp"
            $1422 = HEAP32[299862]|0; //@line 346 "load_abc.cpp"
            $1423 = Math_imul($1421, $1422)|0; //@line 346 "load_abc.cpp"
            $1424 = HEAP32[299863]|0; //@line 346 "load_abc.cpp"
            $1425 = Math_imul($1424, $1418)|0; //@line 346 "load_abc.cpp"
            $1426 = (($1423|0) / ($1425|0))&-1; //@line 346 "load_abc.cpp"
            $$0$i230 = $1426;
           }
           $1427 = (($1417>>>0) / ($$0$i230>>>0))&-1; //@line 4405 "load_abc.cpp"
           $1428 = HEAPU8[$3>>0]|(HEAPU8[$3+1>>0]<<8)|(HEAPU8[$3+2>>0]<<16)|(HEAPU8[$3+3>>0]<<24); //@line 3546 "load_abc.cpp"
           $1429 = ($1428|0)==(0|0); //@line 3546 "load_abc.cpp"
           L1024: do {
            if ($1429) {
             $1443 = HEAP8[$105>>0]|0; //@line 3565 "load_abc.cpp"
             $1444 = $1443&255; //@line 3565 "load_abc.cpp"
             $1445 = (($1427|0) % ($1444|0))&-1; //@line 3565 "load_abc.cpp"
             $1446 = ($1445|0)==(0); //@line 3565 "load_abc.cpp"
             if (!($1446)) {
              $1452 = HEAP8[$104>>0]|0; //@line 3572 "load_abc.cpp"
              $1453 = $1452&255; //@line 3572 "load_abc.cpp"
              $vol$0$i = $1453;
              break;
             }
             $1447 = ($1427|0)==(0); //@line 3566 "load_abc.cpp"
             if ($1447) {
              $1450 = HEAP8[$102>>0]|0; //@line 3569 "load_abc.cpp"
              $1451 = $1450&255; //@line 3569 "load_abc.cpp"
              $vol$0$i = $1451;
              break;
             } else {
              $1448 = HEAP8[$103>>0]|0; //@line 3567 "load_abc.cpp"
              $1449 = $1448&255; //@line 3567 "load_abc.cpp"
              $vol$0$i = $1449;
              break;
             }
            } else {
             $1430 = HEAP8[$104>>0]|0; //@line 3547 "load_abc.cpp"
             $1431 = $1430&255; //@line 3547 "load_abc.cpp"
             $1432 = ($1431*9)|0; //@line 3547 "load_abc.cpp"
             $1433 = (($1432>>>0) / 10)&-1; //@line 3547 "load_abc.cpp"
             $1434 = (_strlen($1428)|0); //@line 3548 "load_abc.cpp"
             $1435 = ($1434|0)>($1427|0); //@line 3548 "load_abc.cpp"
             if (!($1435)) {
              $vol$0$i = $1433;
              break;
             }
             $1436 = (($1428) + ($1427)|0); //@line 3549 "load_abc.cpp"
             $1437 = HEAP8[$1436>>0]|0; //@line 3549 "load_abc.cpp"
             $1438 = $1437 << 24 >> 24; //@line 3549 "load_abc.cpp"
             switch ($1438|0) {
             case 102:  {
              $1439 = HEAP8[$102>>0]|0; //@line 3551 "load_abc.cpp"
              $1440 = $1439&255; //@line 3551 "load_abc.cpp"
              $vol$0$i = $1440;
              break L1024;
              break;
             }
             case 109:  {
              $1441 = HEAP8[$103>>0]|0; //@line 3554 "load_abc.cpp"
              $1442 = $1441&255; //@line 3554 "load_abc.cpp"
              $vol$0$i = $1442;
              break L1024;
              break;
             }
             case 112:  {
              $vol$0$i = $1431;
              break L1024;
              break;
             }
             default: {
              $vol$0$i = $1433;
              break L1024;
             }
             }
            }
           } while(0);
           $1454 = Math_imul($vol$0$i, $abcvol$31225)|0; //@line 3574 "load_abc.cpp"
           $1455 = (($1454|0) / 128)&-1; //@line 3575 "load_abc.cpp"
           $abcchord$0 = 0;$cnotediv$3 = $cnotediv$21233;$cnotelen$3 = $cnotelen$21232;$nd0$0 = 0;$nl0$0 = 0;$p$33 = $p$31;$vnl$0 = 0;
           L1039: while(1) {
            $1456 = ((($p$33)) + 1|0); //@line 4406 "load_abc.cpp"
            $1457 = HEAP8[$p$33>>0]|0; //@line 4406 "load_abc.cpp"
            switch ($1457<<24>>24) {
            case 93: case 0:  {
             $$lcssa2232 = $1456;$abcchord$0$lcssa = $abcchord$0;$cnotediv$3$lcssa = $cnotediv$3;$cnotelen$3$lcssa = $cnotelen$3;$nd0$0$lcssa = $nd0$0;$nl0$0$lcssa = $nl0$0;$p$33$lcssa = $p$33;$vnl$0$lcssa = $vnl$0;
             break L1039;
             break;
            }
            default: {
            }
            }
            $1458 = $1457 << 24 >> 24; //@line 4406 "load_abc.cpp"
            $1459 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4407 "load_abc.cpp"
            $1460 = ((($1459)) + 44|0); //@line 4407 "load_abc.cpp"
            $1461 = ($abcchord$0|0)!=(0); //@line 4407 "load_abc.cpp"
            $1462 = (($abcchord$0) + 10)|0; //@line 4407 "load_abc.cpp"
            $1463 = $1461 ? $1462 : 0; //@line 4407 "load_abc.cpp"
            $1464 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,$1460,$1463)|0); //@line 4407 "load_abc.cpp"
            HEAP8[$41>>0]=$1464&255;HEAP8[$41+1>>0]=($1464>>8)&255;HEAP8[$41+2>>0]=($1464>>16)&255;HEAP8[$41+3>>0]=$1464>>24; //@line 4407 "load_abc.cpp"
            $1465 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4408 "load_abc.cpp"
            $1466 = (__ZL14abc_add_noteonP10_ABCHANDLEiPKcjPci17ABCEVENT_X_EFFECTi($1,$1458,$1456,$1465,$barsig,$1455,$abceffect$31222,$abceffoper$31223)|0); //@line 4408 "load_abc.cpp"
            $$sum63 = (($1466) + 1)|0; //@line 4408 "load_abc.cpp"
            $1467 = (($p$33) + ($$sum63)|0); //@line 4408 "load_abc.cpp"
            $1468 = HEAP8[$1467>>0]|0; //@line 1625 "load_abc.cpp"
            $1469 = $1468 << 24 >> 24; //@line 1625 "load_abc.cpp"
            $isdigittmp1$i$i232 = (($1469) + -48)|0; //@line 1625 "load_abc.cpp"
            $isdigit2$i$i233 = ($isdigittmp1$i$i232>>>0)<(10); //@line 1625 "load_abc.cpp"
            if ($isdigit2$i$i233) {
             $1472 = $1469;$h$04$i$i234 = 0;$i$03$i$i235 = 0;
             while(1) {
              $1470 = ($h$04$i$i234*10)|0; //@line 1626 "load_abc.cpp"
              $1471 = (($1472) + -48)|0; //@line 1626 "load_abc.cpp"
              $1473 = (($1471) + ($1470))|0; //@line 1626 "load_abc.cpp"
              $1474 = (($i$03$i$i235) + 1)|0; //@line 1627 "load_abc.cpp"
              $$sum819 = (($1474) + ($$sum63))|0; //@line 1625 "load_abc.cpp"
              $1475 = (($p$33) + ($$sum819)|0); //@line 1625 "load_abc.cpp"
              $1476 = HEAP8[$1475>>0]|0; //@line 1625 "load_abc.cpp"
              $1477 = $1476 << 24 >> 24; //@line 1625 "load_abc.cpp"
              $isdigittmp$i$i236 = (($1477) + -48)|0; //@line 1625 "load_abc.cpp"
              $isdigit$i$i237 = ($isdigittmp$i$i236>>>0)<(10); //@line 1625 "load_abc.cpp"
              if ($isdigit$i$i237) {
               $1472 = $1477;$h$04$i$i234 = $1473;$i$03$i$i235 = $1474;
              } else {
               $h$0$lcssa$i$i = $1473;$i$0$lcssa$i$i = $1474;
               break;
              }
             }
            } else {
             $h$0$lcssa$i$i = 0;$i$0$lcssa$i$i = 0;
            }
            $1478 = ($i$0$lcssa$i$i|0)==(0); //@line 1629 "load_abc.cpp"
            $$h$0$i$i239 = $1478 ? 1 : $h$0$lcssa$i$i; //@line 1630 "load_abc.cpp"
            $$sum816 = (($i$0$lcssa$i$i) + ($$sum63))|0; //@line 1666 "load_abc.cpp"
            $1479 = (($p$33) + ($$sum816)|0); //@line 1666 "load_abc.cpp"
            $1480 = HEAP8[$1479>>0]|0; //@line 1666 "load_abc.cpp"
            $1481 = ($1480<<24>>24)==(47); //@line 1666 "load_abc.cpp"
            if ($1481) {
             $h$017$i = 1;$i$016$i = $i$0$lcssa$i$i;
             while(1) {
              $1482 = $h$017$i << 1; //@line 1667 "load_abc.cpp"
              $1483 = (($i$016$i) + 1)|0; //@line 1668 "load_abc.cpp"
              $$sum818 = (($1483) + ($$sum63))|0; //@line 1666 "load_abc.cpp"
              $1484 = (($p$33) + ($$sum818)|0); //@line 1666 "load_abc.cpp"
              $1485 = HEAP8[$1484>>0]|0; //@line 1666 "load_abc.cpp"
              $1486 = ($1485<<24>>24)==(47); //@line 1666 "load_abc.cpp"
              if ($1486) {
               $h$017$i = $1482;$i$016$i = $1483;
              } else {
               $$lcssa13$i = $1485;$h$0$lcssa$i = $1482;$i$0$lcssa$i = $1483;
               break;
              }
             }
            } else {
             $$lcssa13$i = $1480;$h$0$lcssa$i = 1;$i$0$lcssa$i = $i$0$lcssa$i$i;
            }
            $1487 = $$lcssa13$i << 24 >> 24; //@line 1666 "load_abc.cpp"
            $isdigittmp$i242 = (($1487) + -48)|0; //@line 1670 "load_abc.cpp"
            $isdigit$i243 = ($isdigittmp$i242>>>0)<(10); //@line 1670 "load_abc.cpp"
            if ($isdigit$i243) {
             $1490 = $1487;$h$04$i3$i = 0;$i$03$i4$i = 0;
             while(1) {
              $1488 = ($h$04$i3$i*10)|0; //@line 1626 "load_abc.cpp"
              $1489 = (($1490) + -48)|0; //@line 1626 "load_abc.cpp"
              $1491 = (($1489) + ($1488))|0; //@line 1626 "load_abc.cpp"
              $1492 = (($i$03$i4$i) + 1)|0; //@line 1627 "load_abc.cpp"
              $$sum$i = (($1492) + ($i$0$lcssa$i))|0; //@line 1625 "load_abc.cpp"
              $$sum817 = (($$sum$i) + ($$sum63))|0; //@line 1625 "load_abc.cpp"
              $1493 = (($p$33) + ($$sum817)|0); //@line 1625 "load_abc.cpp"
              $1494 = HEAP8[$1493>>0]|0; //@line 1625 "load_abc.cpp"
              $1495 = $1494 << 24 >> 24; //@line 1625 "load_abc.cpp"
              $isdigittmp$i5$i = (($1495) + -48)|0; //@line 1625 "load_abc.cpp"
              $isdigit$i6$i = ($isdigittmp$i5$i>>>0)<(10); //@line 1625 "load_abc.cpp"
              if ($isdigit$i6$i) {
               $1490 = $1495;$h$04$i3$i = $1491;$i$03$i4$i = $1492;
              } else {
               $$lcssa2230 = $1491;$$lcssa2231 = $1492;$$sum$i$lcssa = $$sum$i;
               break;
              }
             }
             $1496 = (($h$0$lcssa$i|0) / 2)&-1; //@line 1671 "load_abc.cpp"
             $1497 = ($$lcssa2231|0)==(0); //@line 1629 "load_abc.cpp"
             $$h$0$i10$i = $1497 ? 1 : $$lcssa2230; //@line 1630 "load_abc.cpp"
             $1499 = $$h$0$i10$i;$h$1$i = $1496;$i$1$i = $$sum$i$lcssa;
            } else {
             $1499 = 1;$h$1$i = $h$0$lcssa$i;$i$1$i = $i$0$lcssa$i;
            }
            $1498 = Math_imul($h$1$i, $1499)|0; //@line 1675 "load_abc.cpp"
            $$sum64 = (($i$1$i) + ($$sum63))|0; //@line 4409 "load_abc.cpp"
            $1500 = (($p$33) + ($$sum64)|0); //@line 4409 "load_abc.cpp"
            $1501 = HEAP8[$1500>>0]|0; //@line 4410 "load_abc.cpp"
            $1502 = ($1501<<24>>24)==(45); //@line 4410 "load_abc.cpp"
            do {
             if ($1502) {
              $$sum65 = (($$sum64) + 1)|0; //@line 4411 "load_abc.cpp"
              $1503 = (($p$33) + ($$sum65)|0); //@line 4411 "load_abc.cpp"
              $1504 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4412 "load_abc.cpp"
              $1505 = ((($1504)) + 8|0); //@line 4412 "load_abc.cpp"
              $1506 = HEAPU8[$1505>>0]|(HEAPU8[$1505+1>>0]<<8)|(HEAPU8[$1505+2>>0]<<16)|(HEAPU8[$1505+3>>0]<<24); //@line 4412 "load_abc.cpp"
              $1507 = ($1506|0)==(0|0); //@line 4412 "load_abc.cpp"
              if ($1507) {
               $p$34 = $1503;
               break;
              }
              $1508 = ((($1506)) + 8|0); //@line 4412 "load_abc.cpp"
              $1509 = HEAP8[$1508>>0]|0; //@line 4412 "load_abc.cpp"
              $1510 = ($1509<<24>>24)==(1); //@line 4412 "load_abc.cpp"
              if ($1510) {
               $p$34 = $1503;
               break;
              }
              $1511 = ((($1504)) + 16|0); //@line 4413 "load_abc.cpp"
              HEAP8[$1511>>0]=$1506&255;HEAP8[$1511+1>>0]=($1506>>8)&255;HEAP8[$1511+2>>0]=($1506>>16)&255;HEAP8[$1511+3>>0]=$1506>>24; //@line 4413 "load_abc.cpp"
              $p$34 = $1503;
             } else {
              $p$34 = $1500;
             }
            } while(0);
            $1512 = ($abcchord$0|0)<(8); //@line 4415 "load_abc.cpp"
            if ($1512) {
             $1513 = (($cnl) + ($abcchord$0<<2)|0); //@line 4416 "load_abc.cpp"
             HEAP32[$1513>>2] = $$h$0$i$i239; //@line 4416 "load_abc.cpp"
             $1514 = (($cnd) + ($abcchord$0<<2)|0); //@line 4417 "load_abc.cpp"
             HEAP32[$1514>>2] = $1498; //@line 4417 "load_abc.cpp"
             $1515 = ($abcchord$0|0)==(0); //@line 4419 "load_abc.cpp"
             if ($1515) {
              $cnotediv$5 = $1498;$cnotelen$5 = $$h$0$i$i239;$nd0$1 = $1498;$nl0$1 = $$h$0$i$i239;$vnl$1 = $vnl$0;
             } else {
              label = 538;
             }
            } else {
             label = 538;
            }
            do {
             if ((label|0) == 538) {
              label = 0;
              $1516 = ($cnotelen$3|0)==($$h$0$i$i239|0); //@line 4426 "load_abc.cpp"
              $1517 = ($cnotediv$3|0)==($1498|0); //@line 4426 "load_abc.cpp"
              $or$cond88 = $1516 & $1517; //@line 4426 "load_abc.cpp"
              if ($or$cond88) {
               $cnotediv$5 = $cnotediv$3;$cnotelen$5 = $cnotelen$3;$nd0$1 = $nd0$0;$nl0$1 = $nl0$0;$vnl$1 = $vnl$0;
               break;
              }
              $1518 = Math_imul($1498, $cnotelen$3)|0; //@line 4429 "load_abc.cpp"
              $1519 = Math_imul($$h$0$i$i239, $cnotediv$3)|0; //@line 4429 "load_abc.cpp"
              $1520 = ($1518|0)<($1519|0); //@line 4429 "load_abc.cpp"
              L1065: do {
               if ($1520) {
                $1521 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 1203 "load_abc.cpp"
                $1522 = ((($1521)) + 36|0); //@line 1204 "load_abc.cpp"
                $1523 = HEAP8[$1522>>0]|0; //@line 1204 "load_abc.cpp"
                $tp$01$i245 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 1205 "load_abc.cpp"
                $1524 = ($tp$01$i245|0)==(0|0); //@line 1205 "load_abc.cpp"
                if ($1524) {
                 $cnotediv$4 = $1498;$cnotelen$4 = $$h$0$i$i239;
                 break;
                } else {
                 $1526 = $1521;$tp$02$i247 = $tp$01$i245;
                }
                while(1) {
                 $1525 = ($tp$02$i247|0)==($1526|0); //@line 1206 "load_abc.cpp"
                 do {
                  if (!($1525)) {
                   $1527 = ((($tp$02$i247)) + 36|0); //@line 1206 "load_abc.cpp"
                   $1528 = HEAP8[$1527>>0]|0; //@line 1206 "load_abc.cpp"
                   $1529 = ($1528<<24>>24)==($1523<<24>>24); //@line 1206 "load_abc.cpp"
                   if (!($1529)) {
                    break;
                   }
                   $1530 = ((($tp$02$i247)) + 16|0); //@line 1206 "load_abc.cpp"
                   $1531 = HEAPU8[$1530>>0]|(HEAPU8[$1530+1>>0]<<8)|(HEAPU8[$1530+2>>0]<<16)|(HEAPU8[$1530+3>>0]<<24); //@line 1206 "load_abc.cpp"
                   $1532 = ($1531|0)==(0|0); //@line 1206 "load_abc.cpp"
                   if ($1532) {
                    break;
                   }
                   $1533 = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8)|(HEAPU8[$5+2>>0]<<16)|(HEAPU8[$5+3>>0]<<24); //@line 1207 "load_abc.cpp"
                   $1534 = (_strlen($1533)|0); //@line 323 "load_abc.cpp"
                   $1535 = (($1534) + 40)|0; //@line 323 "load_abc.cpp"
                   $1536 = ($1535>>>0)>(255); //@line 323 "load_abc.cpp"
                   if (!($1536)) {
                    HEAP32[$vararg_buffer95>>2] = $1533; //@line 324 "load_abc.cpp"
                    (_sprintf($txt$i,527285,$vararg_buffer95)|0); //@line 324 "load_abc.cpp"
                    HEAP32[$vararg_buffer98>>2] = $txt$i; //@line 328 "load_abc.cpp"
                    (_fprintf($155,524747,$vararg_buffer98)|0); //@line 328 "load_abc.cpp"
                   }
                   HEAP8[$1530>>0]=0&255;HEAP8[$1530+1>>0]=(0>>8)&255;HEAP8[$1530+2>>0]=(0>>16)&255;HEAP8[$1530+3>>0]=0>>24; //@line 1208 "load_abc.cpp"
                  }
                 } while(0);
                 $tp$0$i249 = HEAPU8[$tp$02$i247>>0]|(HEAPU8[$tp$02$i247+1>>0]<<8)|(HEAPU8[$tp$02$i247+2>>0]<<16)|(HEAPU8[$tp$02$i247+3>>0]<<24); //@line 1205 "load_abc.cpp"
                 $1537 = ($tp$0$i249|0)==(0|0); //@line 1205 "load_abc.cpp"
                 if ($1537) {
                  $cnotediv$4 = $1498;$cnotelen$4 = $$h$0$i$i239;
                  break L1065;
                 }
                 $$pre$i = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 1206 "load_abc.cpp"
                 $1526 = $$pre$i;$tp$02$i247 = $tp$0$i249;
                }
               } else {
                $cnotediv$4 = $cnotediv$3;$cnotelen$4 = $cnotelen$3;
               }
              } while(0);
              $1538 = Math_imul($cnotelen$4, $1498)|0; //@line 4434 "load_abc.cpp"
              $1539 = Math_imul($cnotediv$4, $$h$0$i$i239)|0; //@line 4434 "load_abc.cpp"
              $1540 = ($1538|0)>($1539|0); //@line 4434 "load_abc.cpp"
              do {
               if ($1540) {
                $1541 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4435 "load_abc.cpp"
                $1542 = ((($1541)) + 16|0); //@line 4435 "load_abc.cpp"
                $1543 = HEAPU8[$1542>>0]|(HEAPU8[$1542+1>>0]<<8)|(HEAPU8[$1542+2>>0]<<16)|(HEAPU8[$1542+3>>0]<<24); //@line 4435 "load_abc.cpp"
                $1544 = ($1543|0)==(0|0); //@line 4435 "load_abc.cpp"
                if ($1544) {
                 break;
                }
                $1545 = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8)|(HEAPU8[$5+2>>0]<<16)|(HEAPU8[$5+3>>0]<<24); //@line 4436 "load_abc.cpp"
                $1546 = (_strlen($1545)|0); //@line 323 "load_abc.cpp"
                $1547 = (($1546) + 40)|0; //@line 323 "load_abc.cpp"
                $1548 = ($1547>>>0)>(255); //@line 323 "load_abc.cpp"
                if ($1548) {
                 $1550 = $1541;
                } else {
                 HEAP32[$vararg_buffer101>>2] = $1545; //@line 324 "load_abc.cpp"
                 (_sprintf($txt$i,527285,$vararg_buffer101)|0); //@line 324 "load_abc.cpp"
                 HEAP32[$vararg_buffer104>>2] = $txt$i; //@line 328 "load_abc.cpp"
                 (_fprintf($155,524747,$vararg_buffer104)|0); //@line 328 "load_abc.cpp"
                 $$pre1531 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4437 "load_abc.cpp"
                 $1550 = $$pre1531;
                }
                $1549 = ((($1550)) + 16|0); //@line 4437 "load_abc.cpp"
                HEAP8[$1549>>0]=0&255;HEAP8[$1549+1>>0]=(0>>8)&255;HEAP8[$1549+2>>0]=(0>>16)&255;HEAP8[$1549+3>>0]=0>>24; //@line 4437 "load_abc.cpp"
               }
              } while(0);
              $1551 = Math_imul($1498, $nl0$0)|0; //@line 4441 "load_abc.cpp"
              $1552 = Math_imul($$h$0$i$i239, $nd0$0)|0; //@line 4441 "load_abc.cpp"
              $1553 = ($1551|0)>($1552|0); //@line 4441 "load_abc.cpp"
              $$nl0$0 = $1553 ? $$h$0$i$i239 : $nl0$0; //@line 4441 "load_abc.cpp"
              $$nd0$0 = $1553 ? $1498 : $nd0$0; //@line 4441 "load_abc.cpp"
              $cnotediv$5 = $cnotediv$4;$cnotelen$5 = $cnotelen$4;$nd0$1 = $$nd0$0;$nl0$1 = $$nl0$0;$vnl$1 = 1;
             }
            } while(0);
            $1554 = (($abcchord$0) + 1)|0; //@line 4447 "load_abc.cpp"
            $abcchord$0 = $1554;$cnotediv$3 = $cnotediv$5;$cnotelen$3 = $cnotelen$5;$nd0$0 = $nd0$1;$nl0$0 = $nl0$1;$p$33 = $p$34;$vnl$0 = $vnl$1;
           }
           $1555 = HEAP8[$$lcssa2232>>0]|0; //@line 1625 "load_abc.cpp"
           $1556 = $1555 << 24 >> 24; //@line 1625 "load_abc.cpp"
           $isdigittmp1$i$i253 = (($1556) + -48)|0; //@line 1625 "load_abc.cpp"
           $isdigit2$i$i254 = ($isdigittmp1$i$i253>>>0)<(10); //@line 1625 "load_abc.cpp"
           if ($isdigit2$i$i254) {
            $1559 = $1556;$h$04$i$i255 = 0;$i$03$i$i256 = 0;
            while(1) {
             $1557 = ($h$04$i$i255*10)|0; //@line 1626 "load_abc.cpp"
             $1558 = (($1559) + -48)|0; //@line 1626 "load_abc.cpp"
             $1560 = (($1558) + ($1557))|0; //@line 1626 "load_abc.cpp"
             $1561 = (($i$03$i$i256) + 1)|0; //@line 1627 "load_abc.cpp"
             $$sum815 = (($i$03$i$i256) + 2)|0; //@line 1625 "load_abc.cpp"
             $1562 = (($p$33$lcssa) + ($$sum815)|0); //@line 1625 "load_abc.cpp"
             $1563 = HEAP8[$1562>>0]|0; //@line 1625 "load_abc.cpp"
             $1564 = $1563 << 24 >> 24; //@line 1625 "load_abc.cpp"
             $isdigittmp$i$i257 = (($1564) + -48)|0; //@line 1625 "load_abc.cpp"
             $isdigit$i$i258 = ($isdigittmp$i$i257>>>0)<(10); //@line 1625 "load_abc.cpp"
             if ($isdigit$i$i258) {
              $1559 = $1564;$h$04$i$i255 = $1560;$i$03$i$i256 = $1561;
             } else {
              $1567 = $1563;$h$0$lcssa$i$i260 = $1560;$i$0$lcssa$i$i261 = $1561;
              break;
             }
            }
           } else {
            $1567 = $1555;$h$0$lcssa$i$i260 = 0;$i$0$lcssa$i$i261 = 0;
           }
           $1565 = ($i$0$lcssa$i$i261|0)==(0); //@line 1629 "load_abc.cpp"
           $$h$0$i$i262 = $1565 ? 1 : $h$0$lcssa$i$i260; //@line 1630 "load_abc.cpp"
           $1566 = ($1567<<24>>24)==(47); //@line 1666 "load_abc.cpp"
           if ($1566) {
            $h$017$i264 = 1;$i$016$i265 = $i$0$lcssa$i$i261;
            while(1) {
             $1568 = $h$017$i264 << 1; //@line 1667 "load_abc.cpp"
             $1569 = (($i$016$i265) + 1)|0; //@line 1668 "load_abc.cpp"
             $$sum814 = (($i$016$i265) + 2)|0; //@line 1666 "load_abc.cpp"
             $1570 = (($p$33$lcssa) + ($$sum814)|0); //@line 1666 "load_abc.cpp"
             $1571 = HEAP8[$1570>>0]|0; //@line 1666 "load_abc.cpp"
             $1572 = ($1571<<24>>24)==(47); //@line 1666 "load_abc.cpp"
             if ($1572) {
              $h$017$i264 = $1568;$i$016$i265 = $1569;
             } else {
              $$lcssa13$i267 = $1571;$h$0$lcssa$i268 = $1568;$i$0$lcssa$i269 = $1569;
              break;
             }
            }
           } else {
            $$lcssa13$i267 = $1567;$h$0$lcssa$i268 = 1;$i$0$lcssa$i269 = $i$0$lcssa$i$i261;
           }
           $1573 = $$lcssa13$i267 << 24 >> 24; //@line 1666 "load_abc.cpp"
           $isdigittmp$i270 = (($1573) + -48)|0; //@line 1670 "load_abc.cpp"
           $isdigit$i271 = ($isdigittmp$i270>>>0)<(10); //@line 1670 "load_abc.cpp"
           if ($isdigit$i271) {
            $1576 = $1573;$h$04$i3$i273 = 0;$i$03$i4$i274 = 0;
            while(1) {
             $1574 = ($h$04$i3$i273*10)|0; //@line 1626 "load_abc.cpp"
             $1575 = (($1576) + -48)|0; //@line 1626 "load_abc.cpp"
             $1577 = (($1575) + ($1574))|0; //@line 1626 "load_abc.cpp"
             $1578 = (($i$03$i4$i274) + 1)|0; //@line 1627 "load_abc.cpp"
             $$sum$i275 = (($1578) + ($i$0$lcssa$i269))|0; //@line 1625 "load_abc.cpp"
             $$sum813 = (($$sum$i275) + 1)|0; //@line 1625 "load_abc.cpp"
             $1579 = (($p$33$lcssa) + ($$sum813)|0); //@line 1625 "load_abc.cpp"
             $1580 = HEAP8[$1579>>0]|0; //@line 1625 "load_abc.cpp"
             $1581 = $1580 << 24 >> 24; //@line 1625 "load_abc.cpp"
             $isdigittmp$i5$i276 = (($1581) + -48)|0; //@line 1625 "load_abc.cpp"
             $isdigit$i6$i277 = ($isdigittmp$i5$i276>>>0)<(10); //@line 1625 "load_abc.cpp"
             if ($isdigit$i6$i277) {
              $1576 = $1581;$h$04$i3$i273 = $1577;$i$03$i4$i274 = $1578;
             } else {
              $$lcssa2239 = $1577;$$lcssa2240 = $1578;$$sum$i275$lcssa = $$sum$i275;
              break;
             }
            }
            $1582 = (($h$0$lcssa$i268|0) / 2)&-1; //@line 1671 "load_abc.cpp"
            $1583 = ($$lcssa2240|0)==(0); //@line 1629 "load_abc.cpp"
            $$h$0$i10$i279 = $1583 ? 1 : $$lcssa2239; //@line 1630 "load_abc.cpp"
            $1585 = $$h$0$i10$i279;$h$1$i282 = $1582;$i$1$i281 = $$sum$i275$lcssa;
           } else {
            $1585 = 1;$h$1$i282 = $h$0$lcssa$i268;$i$1$i281 = $i$0$lcssa$i269;
           }
           $1584 = Math_imul($h$1$i282, $1585)|0; //@line 1675 "load_abc.cpp"
           $$sum61 = (($i$1$i281) + 1)|0; //@line 4449 "load_abc.cpp"
           $1586 = (($p$33$lcssa) + ($$sum61)|0); //@line 4449 "load_abc.cpp"
           $1587 = HEAP8[$1586>>0]|0; //@line 4450 "load_abc.cpp"
           $1588 = ($1587<<24>>24)==(45); //@line 4450 "load_abc.cpp"
           $$sum62 = (($i$1$i281) + 2)|0; //@line 4450 "load_abc.cpp"
           $1589 = (($p$33$lcssa) + ($$sum62)|0); //@line 4450 "load_abc.cpp"
           $p$35 = $1588 ? $1589 : $1586; //@line 4450 "load_abc.cpp"
           $1590 = ($abcarpeggio$21229|0)==(0); //@line 4451 "load_abc.cpp"
           do {
            if (!($1590)) {
             $1591 = Math_imul($1584, $nd0$0$lcssa)|0; //@line 4452 "load_abc.cpp"
             $1592 = HEAP32[$snotediv>>2]|0; //@line 4452 "load_abc.cpp"
             $1593 = Math_imul($1591, $1592)|0; //@line 4452 "load_abc.cpp"
             $1594 = ($1593|0)==(0); //@line 345 "load_abc.cpp"
             if ($1594) {
              $$0$i284 = 0;
             } else {
              $1595 = HEAP32[$snotelen>>2]|0; //@line 4452 "load_abc.cpp"
              $1596 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4452 "load_abc.cpp"
              $1597 = HEAP32[299862]|0; //@line 346 "load_abc.cpp"
              $1598 = ($nl0$0$lcssa*3072)|0; //@line 346 "load_abc.cpp"
              $1599 = Math_imul($1598, $$h$0$i$i262)|0; //@line 4452 "load_abc.cpp"
              $1600 = Math_imul($1599, $1595)|0; //@line 4452 "load_abc.cpp"
              $1601 = Math_imul($1600, $1596)|0; //@line 346 "load_abc.cpp"
              $1602 = Math_imul($1601, $1597)|0; //@line 346 "load_abc.cpp"
              $1603 = HEAP32[299863]|0; //@line 346 "load_abc.cpp"
              $1604 = Math_imul($1603, $1593)|0; //@line 346 "load_abc.cpp"
              $1605 = (($1602|0) / ($1604|0))&-1; //@line 346 "load_abc.cpp"
              $$0$i284 = $1605;
             }
             $1606 = (($$0$i284>>>0) / ($abcchord$0$lcssa>>>0))&-1; //@line 4452 "load_abc.cpp"
             $1607 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4453 "load_abc.cpp"
             $1608 = ($1607*192)|0; //@line 339 "load_abc.cpp"
             $1609 = ($1606>>>0)>($1608>>>0); //@line 4453 "load_abc.cpp"
             $$786 = $1609 ? $1608 : $1606; //@line 4453 "load_abc.cpp"
             $1610 = ($abcchord$0$lcssa|0)>(1); //@line 4454 "load_abc.cpp"
             if (!($1610)) {
              break;
             }
             $$pre1527 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4455 "load_abc.cpp"
             $1612 = $$pre1527;$nl0$21204 = 1;
             while(1) {
              $1611 = ((($1612)) + 44|0); //@line 4455 "load_abc.cpp"
              $1613 = (($nl0$21204) + 10)|0; //@line 4455 "load_abc.cpp"
              $1614 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,$1611,$1613)|0); //@line 4455 "load_abc.cpp"
              HEAP8[$41>>0]=$1614&255;HEAP8[$41+1>>0]=($1614>>8)&255;HEAP8[$41+2>>0]=($1614>>16)&255;HEAP8[$41+3>>0]=$1614>>24; //@line 4455 "load_abc.cpp"
              $1615 = ((($1614)) + 8|0); //@line 4456 "load_abc.cpp"
              $1616 = HEAPU8[$1615>>0]|(HEAPU8[$1615+1>>0]<<8)|(HEAPU8[$1615+2>>0]<<16)|(HEAPU8[$1615+3>>0]<<24); //@line 4456 "load_abc.cpp"
              $1617 = ($1616|0)==(0|0); //@line 4456 "load_abc.cpp"
              if (!($1617)) {
               $1618 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4457 "load_abc.cpp"
               $1619 = Math_imul($nl0$21204, $$786)|0; //@line 4457 "load_abc.cpp"
               $1620 = (($1618) + ($1619))|0; //@line 4457 "load_abc.cpp"
               $1621 = ((($1616)) + 4|0); //@line 4457 "load_abc.cpp"
               HEAP8[$1621>>0]=$1620&255;HEAP8[$1621+1>>0]=($1620>>8)&255;HEAP8[$1621+2>>0]=($1620>>16)&255;HEAP8[$1621+3>>0]=$1620>>24; //@line 4457 "load_abc.cpp"
              }
              $1622 = (($nl0$21204) + 1)|0; //@line 4454 "load_abc.cpp"
              $exitcond1501 = ($1622|0)==($abcchord$0$lcssa|0); //@line 4454 "load_abc.cpp"
              if ($exitcond1501) {
               break;
              } else {
               $1612 = $1614;$nl0$21204 = $1622;
              }
             }
            }
           } while(0);
           $1623 = Math_imul($$h$0$i$i262, $cnotelen$3$lcssa)|0; //@line 4460 "load_abc.cpp"
           $1624 = Math_imul($1584, $cnotediv$3$lcssa)|0; //@line 4461 "load_abc.cpp"
           $1625 = ($1626|0)==(0); //@line 1718 "load_abc.cpp"
           if ($1625) {
            $$0$i286 = 0;$1639 = $1623;$1641 = $1624;
           } else {
            $1627 = Math_imul($1623, $1628)|0; //@line 1719 "load_abc.cpp"
            $1629 = Math_imul($1624, $1630)|0; //@line 1720 "load_abc.cpp"
            $1631 = (($1626) + -1)|0; //@line 1721 "load_abc.cpp"
            $$0$i286 = $1631;$1639 = $1627;$1641 = $1629;
           }
           $p$36 = $p$35;
           while(1) {
            $1632 = HEAP8[$p$36>>0]|0; //@line 4463 "load_abc.cpp"
            $1633 = $1632 << 24 >> 24; //@line 4463 "load_abc.cpp"
            $1634 = (_isspace($1633)|0); //@line 4463 "load_abc.cpp"
            $1635 = ($1634|0)==(0); //@line 4463 "load_abc.cpp"
            $1636 = ((($p$36)) + 1|0); //@line 4463 "load_abc.cpp"
            if ($1635) {
             $p$36$lcssa = $p$36;
             break;
            } else {
             $p$36 = $1636;
            }
           }
           switch ($1637|0) {
           case 60:  {
            $1638 = ($1639*3)|0; //@line 1683 "load_abc.cpp"
            $1640 = $1641 << 1; //@line 1684 "load_abc.cpp"
            $$0$i287 = 0;$1646 = $1638;$1648 = $1640;
            break;
           }
           case 62:  {
            $1642 = $1641 << 1; //@line 1688 "load_abc.cpp"
            $$0$i287 = 0;$1646 = $1639;$1648 = $1642;
            break;
           }
           default: {
            $$0$i287 = $abchornpipe$3$lcssa;$1646 = $1639;$1648 = $1641;
           }
           }
           $1643 = HEAP8[$p$36$lcssa>>0]|0; //@line 1692 "load_abc.cpp"
           $1644 = $1643 << 24 >> 24; //@line 1692 "load_abc.cpp"
           L1123: do {
            switch ($1644|0) {
            case 62:  {
             $1645 = ($1646*3)|0; //@line 1695 "load_abc.cpp"
             $1647 = $1648 << 1; //@line 1696 "load_abc.cpp"
             $$01$i = 1;$1656 = $1647;$1662 = $1645;$4061 = 62;
             break;
            }
            case 60:  {
             $1649 = $1648 << 1; //@line 1699 "load_abc.cpp"
             $$01$i = 1;$1656 = $1649;$1662 = $1646;$4061 = 60;
             break;
            }
            default: {
             $1650 = ($$0$i287|0)!=(0); //@line 1705 "load_abc.cpp"
             $1651 = ($1646|0)==(1); //@line 1706 "load_abc.cpp"
             $or$cond787 = $1651 & $1650; //@line 1705 "load_abc.cpp"
             if (!($or$cond787)) {
              $$01$i = 0;$1656 = $1648;$1662 = $1646;$4061 = 0;
              break L1123;
             }
             $1652 = ($1648|0)==(1); //@line 1706 "load_abc.cpp"
             $$788 = $1652 ? 3 : 1; //@line 1706 "load_abc.cpp"
             $$789 = $1652 ? 2 : $1648; //@line 1706 "load_abc.cpp"
             $$790 = $1652 ? 62 : 0; //@line 1706 "load_abc.cpp"
             $$01$i = 0;$1656 = $$789;$1662 = $$788;$4061 = $$790;
            }
            }
           } while(0);
           $1653 = (($p$36$lcssa) + ($$01$i)|0); //@line 4464 "load_abc.cpp"
           $1654 = HEAP32[$snotediv>>2]|0; //@line 4465 "load_abc.cpp"
           $1655 = Math_imul($1654, $1656)|0; //@line 4465 "load_abc.cpp"
           $1657 = ($1655|0)==(0); //@line 345 "load_abc.cpp"
           if ($1657) {
            $$0$i288 = 0;
           } else {
            $1658 = HEAP32[$snotelen>>2]|0; //@line 4465 "load_abc.cpp"
            $1659 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4465 "load_abc.cpp"
            $1660 = HEAP32[299862]|0; //@line 346 "load_abc.cpp"
            $1661 = ($1662*3072)|0; //@line 346 "load_abc.cpp"
            $1663 = Math_imul($1661, $1658)|0; //@line 4465 "load_abc.cpp"
            $1664 = Math_imul($1663, $1659)|0; //@line 346 "load_abc.cpp"
            $1665 = Math_imul($1664, $1660)|0; //@line 346 "load_abc.cpp"
            $1666 = HEAP32[299863]|0; //@line 346 "load_abc.cpp"
            $1667 = Math_imul($1666, $1655)|0; //@line 346 "load_abc.cpp"
            $1668 = (($1665|0) / ($1667|0))&-1; //@line 346 "load_abc.cpp"
            $$0$i288 = $1668;
           }
           $not$812 = ($abcfermata$21228|0)!=(0); //@line 4466 "load_abc.cpp"
           $1669 = $not$812&1; //@line 4466 "load_abc.cpp"
           $$89 = $$0$i288 << $1669; //@line 4466 "load_abc.cpp"
           $1670 = ($$89>>>0)>($abcgrace$41234>>>0); //@line 4470 "load_abc.cpp"
           if ($1670) {
            $1671 = (($$89) - ($abcgrace$41234))|0; //@line 4471 "load_abc.cpp"
            $abcgrace$5 = 0;$thistime$2 = $1671;
           } else {
            $1672 = (($abcgrace$41234) - ($$89))|0; //@line 4475 "load_abc.cpp"
            $1673 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4476 "load_abc.cpp"
            $1674 = ($1673*192)|0; //@line 339 "load_abc.cpp"
            $1675 = (($1672) + ($1674))|0; //@line 4477 "load_abc.cpp"
            $abcgrace$5 = $1675;$thistime$2 = $1674;
           }
           $1676 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4479 "load_abc.cpp"
           $1677 = (($1676) + ($thistime$2))|0; //@line 4479 "load_abc.cpp"
           HEAP8[$31>>0]=$1677&255;HEAP8[$31+1>>0]=($1677>>8)&255;HEAP8[$31+2>>0]=($1677>>16)&255;HEAP8[$31+3>>0]=$1677>>24; //@line 4479 "load_abc.cpp"
           $1678 = ($abcchord$0$lcssa|0)>(0); //@line 4480 "load_abc.cpp"
           if ($1678) {
            $1679 = ($vnl$0$lcssa|0)!=(0); //@line 4483 "load_abc.cpp"
            $1680 = Math_imul($thistime$2, $cnotediv$3$lcssa)|0; //@line 4486 "load_abc.cpp"
            $abcchord$11206 = $abcchord$0$lcssa;
            while(1) {
             $1681 = (($abcchord$11206) + -1)|0; //@line 4481 "load_abc.cpp"
             $1682 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4482 "load_abc.cpp"
             $1683 = ((($1682)) + 44|0); //@line 4482 "load_abc.cpp"
             $1684 = ($1681|0)!=(0); //@line 4482 "load_abc.cpp"
             $1685 = (($abcchord$11206) + 9)|0; //@line 4482 "load_abc.cpp"
             $1686 = $1684 ? $1685 : 0; //@line 4482 "load_abc.cpp"
             $1687 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,$1683,$1686)|0); //@line 4482 "load_abc.cpp"
             HEAP8[$41>>0]=$1687&255;HEAP8[$41+1>>0]=($1687>>8)&255;HEAP8[$41+2>>0]=($1687>>16)&255;HEAP8[$41+3>>0]=$1687>>24; //@line 4482 "load_abc.cpp"
             $1688 = ($abcchord$11206|0)<(9); //@line 4483 "load_abc.cpp"
             $or$cond23 = $1679 & $1688; //@line 4483 "load_abc.cpp"
             do {
              if ($or$cond23) {
               $1689 = (($cnl) + ($1681<<2)|0); //@line 4483 "load_abc.cpp"
               $1690 = HEAP32[$1689>>2]|0; //@line 4483 "load_abc.cpp"
               $1691 = ($1690|0)==($cnotelen$3$lcssa|0); //@line 4483 "load_abc.cpp"
               $1692 = (($cnd) + ($1681<<2)|0); //@line 4483 "load_abc.cpp"
               $1693 = HEAP32[$1692>>2]|0; //@line 4483 "load_abc.cpp"
               $1694 = ($1693|0)==($cnotediv$3$lcssa|0); //@line 4483 "load_abc.cpp"
               $or$cond1743 = $1691 & $1694; //@line 4483 "load_abc.cpp"
               if ($or$cond1743) {
                label = 593;
                break;
               }
               $1695 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4485 "load_abc.cpp"
               $1696 = (($1695) - ($thistime$2))|0; //@line 4485 "load_abc.cpp"
               $1697 = Math_imul($1680, $1690)|0; //@line 4486 "load_abc.cpp"
               $1698 = Math_imul($1693, $cnotelen$3$lcssa)|0; //@line 4486 "load_abc.cpp"
               $1699 = (($1697>>>0) / ($1698>>>0))&-1; //@line 4486 "load_abc.cpp"
               $1700 = (($1696) + ($1699))|0; //@line 4486 "load_abc.cpp"
               $1701 = ((($1687)) + 42|0); //@line 1174 "load_abc.cpp"
               $1702 = HEAP8[$1701>>0]|0; //@line 1174 "load_abc.cpp"
               $1703 = $1702&255; //@line 1174 "load_abc.cpp"
               $1704 = (_pat_gmtosmp($1703)|0); //@line 1174 "load_abc.cpp"
               $1705 = $1704&255; //@line 1174 "load_abc.cpp"
               $1706 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
               HEAP8[$1706>>0]=0&255;HEAP8[$1706+1>>0]=(0>>8)&255;HEAP8[$1706+2>>0]=(0>>16)&255;HEAP8[$1706+3>>0]=0>>24; //@line 540 "load_abc.cpp"
               $1707 = ((($1706)) + 4|0); //@line 541 "load_abc.cpp"
               HEAP8[$1707>>0]=$1700&255;HEAP8[$1707+1>>0]=($1700>>8)&255;HEAP8[$1707+2>>0]=($1700>>16)&255;HEAP8[$1707+3>>0]=$1700>>24; //@line 541 "load_abc.cpp"
               $1708 = ((($1706)) + 8|0); //@line 543 "load_abc.cpp"
               HEAP8[$1708>>0] = 0; //@line 543 "load_abc.cpp"
               $1709 = ((($1706)) + 9|0); //@line 543 "load_abc.cpp"
               HEAP8[$1709>>0] = 0; //@line 543 "load_abc.cpp"
               $1710 = ((($1706)) + 10|0); //@line 543 "load_abc.cpp"
               HEAP8[$1710>>0] = $1705; //@line 543 "load_abc.cpp"
               $1711 = ((($1706)) + 11|0); //@line 543 "load_abc.cpp"
               HEAP8[$1711>>0] = 0; //@line 543 "load_abc.cpp"
               $1712 = ((($1706)) + 12|0); //@line 543 "load_abc.cpp"
               HEAP8[$1712>>0] = 0; //@line 543 "load_abc.cpp"
               $1713 = ((($1706)) + 13|0); //@line 543 "load_abc.cpp"
               HEAP8[$1713>>0] = 0; //@line 543 "load_abc.cpp"
               $1714 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
               $1715 = ((($1706)) + 14|0); //@line 544 "load_abc.cpp"
               HEAP8[$1715>>0] = $1714; //@line 544 "load_abc.cpp"
               $1716 = ((($1706)) + 15|0); //@line 545 "load_abc.cpp"
               HEAP8[$1716>>0] = 0; //@line 545 "load_abc.cpp"
               __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($1687,$1706); //@line 1179 "load_abc.cpp"
              } else {
               label = 593;
              }
             } while(0);
             do {
              if ((label|0) == 593) {
               label = 0;
               do {
                if ($1588) {
                 $1717 = ((($1687)) + 8|0); //@line 4489 "load_abc.cpp"
                 $1718 = HEAPU8[$1717>>0]|(HEAPU8[$1717+1>>0]<<8)|(HEAPU8[$1717+2>>0]<<16)|(HEAPU8[$1717+3>>0]<<24); //@line 4489 "load_abc.cpp"
                 $1719 = ($1718|0)==(0|0); //@line 4489 "load_abc.cpp"
                 if ($1719) {
                  break;
                 }
                 $1720 = ((($1718)) + 8|0); //@line 4489 "load_abc.cpp"
                 $1721 = HEAP8[$1720>>0]|0; //@line 4489 "load_abc.cpp"
                 $1722 = ($1721<<24>>24)==(1); //@line 4489 "load_abc.cpp"
                 if ($1722) {
                  break;
                 }
                 $1723 = ((($1687)) + 16|0); //@line 4490 "load_abc.cpp"
                 HEAP8[$1723>>0]=$1718&255;HEAP8[$1723+1>>0]=($1718>>8)&255;HEAP8[$1723+2>>0]=($1718>>16)&255;HEAP8[$1723+3>>0]=$1718>>24; //@line 4490 "load_abc.cpp"
                }
               } while(0);
               $1724 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4491 "load_abc.cpp"
               $1725 = ($1724*192)|0; //@line 339 "load_abc.cpp"
               $1726 = ($thistime$2>>>0)>($1725>>>0); //@line 4491 "load_abc.cpp"
               $1727 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4492 "load_abc.cpp"
               $1728 = ((($1687)) + 42|0); //@line 1174 "load_abc.cpp"
               $1729 = HEAP8[$1728>>0]|0; //@line 1174 "load_abc.cpp"
               if ($1726) {
                $1730 = (($1727) - ($abcnoslurs$11))|0; //@line 4492 "load_abc.cpp"
                $1731 = $1729&255; //@line 1174 "load_abc.cpp"
                $1732 = (_pat_gmtosmp($1731)|0); //@line 1174 "load_abc.cpp"
                $1733 = $1732&255; //@line 1174 "load_abc.cpp"
                $1734 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
                HEAP8[$1734>>0]=0&255;HEAP8[$1734+1>>0]=(0>>8)&255;HEAP8[$1734+2>>0]=(0>>16)&255;HEAP8[$1734+3>>0]=0>>24; //@line 540 "load_abc.cpp"
                $1735 = ((($1734)) + 4|0); //@line 541 "load_abc.cpp"
                HEAP8[$1735>>0]=$1730&255;HEAP8[$1735+1>>0]=($1730>>8)&255;HEAP8[$1735+2>>0]=($1730>>16)&255;HEAP8[$1735+3>>0]=$1730>>24; //@line 541 "load_abc.cpp"
                $1736 = ((($1734)) + 8|0); //@line 543 "load_abc.cpp"
                HEAP8[$1736>>0] = 0; //@line 543 "load_abc.cpp"
                $1737 = ((($1734)) + 9|0); //@line 543 "load_abc.cpp"
                HEAP8[$1737>>0] = 0; //@line 543 "load_abc.cpp"
                $1738 = ((($1734)) + 10|0); //@line 543 "load_abc.cpp"
                HEAP8[$1738>>0] = $1733; //@line 543 "load_abc.cpp"
                $1739 = ((($1734)) + 11|0); //@line 543 "load_abc.cpp"
                HEAP8[$1739>>0] = 0; //@line 543 "load_abc.cpp"
                $1740 = ((($1734)) + 12|0); //@line 543 "load_abc.cpp"
                HEAP8[$1740>>0] = 0; //@line 543 "load_abc.cpp"
                $1741 = ((($1734)) + 13|0); //@line 543 "load_abc.cpp"
                HEAP8[$1741>>0] = 0; //@line 543 "load_abc.cpp"
                $1742 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
                $1743 = ((($1734)) + 14|0); //@line 544 "load_abc.cpp"
                HEAP8[$1743>>0] = $1742; //@line 544 "load_abc.cpp"
                $1744 = ((($1734)) + 15|0); //@line 545 "load_abc.cpp"
                HEAP8[$1744>>0] = 0; //@line 545 "load_abc.cpp"
                __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($1687,$1734); //@line 1179 "load_abc.cpp"
                break;
               } else {
                $1746 = $1729&255; //@line 1174 "load_abc.cpp"
                $1747 = (_pat_gmtosmp($1746)|0); //@line 1174 "load_abc.cpp"
                $1748 = $1747&255; //@line 1174 "load_abc.cpp"
                $1749 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
                HEAP8[$1749>>0]=0&255;HEAP8[$1749+1>>0]=(0>>8)&255;HEAP8[$1749+2>>0]=(0>>16)&255;HEAP8[$1749+3>>0]=0>>24; //@line 540 "load_abc.cpp"
                $1750 = ((($1749)) + 4|0); //@line 541 "load_abc.cpp"
                HEAP8[$1750>>0]=$1727&255;HEAP8[$1750+1>>0]=($1727>>8)&255;HEAP8[$1750+2>>0]=($1727>>16)&255;HEAP8[$1750+3>>0]=$1727>>24; //@line 541 "load_abc.cpp"
                $1751 = ((($1749)) + 8|0); //@line 543 "load_abc.cpp"
                HEAP8[$1751>>0] = 0; //@line 543 "load_abc.cpp"
                $1752 = ((($1749)) + 9|0); //@line 543 "load_abc.cpp"
                HEAP8[$1752>>0] = 0; //@line 543 "load_abc.cpp"
                $1753 = ((($1749)) + 10|0); //@line 543 "load_abc.cpp"
                HEAP8[$1753>>0] = $1748; //@line 543 "load_abc.cpp"
                $1754 = ((($1749)) + 11|0); //@line 543 "load_abc.cpp"
                HEAP8[$1754>>0] = 0; //@line 543 "load_abc.cpp"
                $1755 = ((($1749)) + 12|0); //@line 543 "load_abc.cpp"
                HEAP8[$1755>>0] = 0; //@line 543 "load_abc.cpp"
                $1756 = ((($1749)) + 13|0); //@line 543 "load_abc.cpp"
                HEAP8[$1756>>0] = 0; //@line 543 "load_abc.cpp"
                $1757 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
                $1758 = ((($1749)) + 14|0); //@line 544 "load_abc.cpp"
                HEAP8[$1758>>0] = $1757; //@line 544 "load_abc.cpp"
                $1759 = ((($1749)) + 15|0); //@line 545 "load_abc.cpp"
                HEAP8[$1759>>0] = 0; //@line 545 "load_abc.cpp"
                __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($1687,$1749); //@line 1179 "load_abc.cpp"
                break;
               }
              }
             } while(0);
             $1745 = ($abcchord$11206|0)>(1); //@line 4480 "load_abc.cpp"
             if ($1745) {
              $abcchord$11206 = $1681;
             } else {
              break;
             }
            }
           }
           $1760 = HEAPU8[$39>>0]|(HEAPU8[$39+1>>0]<<8)|(HEAPU8[$39+2>>0]<<16)|(HEAPU8[$39+3>>0]<<24); //@line 4497 "load_abc.cpp"
           $1761 = ($1760|0)==(0); //@line 4497 "load_abc.cpp"
           do {
            if (!($1761)) {
             $1762 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4497 "load_abc.cpp"
             $1763 = HEAPU8[$42>>0]|(HEAPU8[$42+1>>0]<<8)|(HEAPU8[$42+2>>0]<<16)|(HEAPU8[$42+3>>0]<<24); //@line 4497 "load_abc.cpp"
             $1764 = ($1762|0)==($1763|0); //@line 4497 "load_abc.cpp"
             if (!($1764)) {
              break;
             }
             $1765 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4498 "load_abc.cpp"
             __ZL14abc_add_gchordP10_ABCHANDLEjj($1,$1765,$bartime$41235); //@line 4498 "load_abc.cpp"
            }
           } while(0);
           $1766 = HEAPU8[$38>>0]|(HEAPU8[$38+1>>0]<<8)|(HEAPU8[$38+2>>0]<<16)|(HEAPU8[$38+3>>0]<<24); //@line 4499 "load_abc.cpp"
           $1767 = ($1766|0)==(0); //@line 4499 "load_abc.cpp"
           do {
            if (!($1767)) {
             $1768 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4499 "load_abc.cpp"
             $1769 = HEAPU8[$43>>0]|(HEAPU8[$43+1>>0]<<8)|(HEAPU8[$43+2>>0]<<16)|(HEAPU8[$43+3>>0]<<24); //@line 4499 "load_abc.cpp"
             $1770 = ($1768|0)==($1769|0); //@line 4499 "load_abc.cpp"
             if (!($1770)) {
              break;
             }
             $1771 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4500 "load_abc.cpp"
             __ZL12abc_add_drumP10_ABCHANDLEjj($1,$1771,$bartime$41235); //@line 4500 "load_abc.cpp"
            }
           } while(0);
           $1772 = ($abceffoper$31223|0)==(255); //@line 4502 "load_abc.cpp"
           $abceffect$3$ = $1772 ? $abceffect$31222 : 0; //@line 4502 "load_abc.cpp"
           $4057 = $4061;$4058 = $1630;$4059 = $1628;$4060 = $$0$i286;$abcarpeggio$3 = 0;$abceffect$5 = $abceffect$3$;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = 0;$abcgrace$10 = $abcgrace$5;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$13;$abcnoslurs$12 = $abcnoslurs$11;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $1587;$cnotediv$6 = $cnotediv$3$lcssa;$cnotelen$6 = $cnotelen$3$lcssa;$p$50 = $1653;
           break L678;
          }
         } while(0);
         $1773 = $1393 << 24 >> 24; //@line 4505 "load_abc.cpp"
         $isdigittmp56 = (($1773) + -48)|0; //@line 4505 "load_abc.cpp"
         $isdigit57 = ($isdigittmp56>>>0)<(10); //@line 4505 "load_abc.cpp"
         if (!($isdigit57)) {
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$13;$abcnoslurs$12 = $abcnoslurs$11;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$31;
          break L678;
         }
         $1774 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4506 "load_abc.cpp"
         $1775 = ($1774|0)==(0|0); //@line 915 "load_abc.cpp"
         if ($1775) {
          $1776 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
          $1777 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
          $1778 = ((($1776)) + 20|0); //@line 917 "load_abc.cpp"
          HEAP8[$1778>>0]=$1777&255;HEAP8[$1778+1>>0]=($1777>>8)&255;HEAP8[$1778+2>>0]=($1777>>16)&255;HEAP8[$1778+3>>0]=$1777>>24; //@line 917 "load_abc.cpp"
          $$0$i290 = $1776;
         } else {
          $$0$i290 = $1774;
         }
         HEAP8[$41>>0]=$$0$i290&255;HEAP8[$41+1>>0]=($$0$i290>>8)&255;HEAP8[$41+2>>0]=($$0$i290>>16)&255;HEAP8[$41+3>>0]=$$0$i290>>24; //@line 4506 "load_abc.cpp"
         $1779 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4507 "load_abc.cpp"
         $1780 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
         HEAP8[$1780>>0]=0&255;HEAP8[$1780+1>>0]=(0>>8)&255;HEAP8[$1780+2>>0]=(0>>16)&255;HEAP8[$1780+3>>0]=0>>24; //@line 540 "load_abc.cpp"
         $1781 = ((($1780)) + 4|0); //@line 541 "load_abc.cpp"
         HEAP8[$1781>>0]=$1779&255;HEAP8[$1781+1>>0]=($1779>>8)&255;HEAP8[$1781+2>>0]=($1779>>16)&255;HEAP8[$1781+3>>0]=$1779>>24; //@line 541 "load_abc.cpp"
         $1782 = ((($1780)) + 8|0); //@line 543 "load_abc.cpp"
         HEAP8[$1782>>0] = 1; //@line 543 "load_abc.cpp"
         $1783 = ((($1780)) + 9|0); //@line 543 "load_abc.cpp"
         HEAP8[$1783>>0] = 80; //@line 543 "load_abc.cpp"
         $1784 = ((($1780)) + 10|0); //@line 543 "load_abc.cpp"
         HEAP8[$1784>>0]=0&255;HEAP8[$1784+1>>0]=(0>>8)&255;HEAP8[$1784+2>>0]=(0>>16)&255;HEAP8[$1784+3>>0]=0>>24; //@line 543 "load_abc.cpp"
         $1785 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
         $1786 = ((($1780)) + 14|0); //@line 544 "load_abc.cpp"
         HEAP8[$1786>>0] = $1785; //@line 544 "load_abc.cpp"
         $1787 = ((($1780)) + 15|0); //@line 545 "load_abc.cpp"
         HEAP8[$1787>>0] = 0; //@line 545 "load_abc.cpp"
         __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($$0$i290,$1780); //@line 1153 "load_abc.cpp"
         $1788 = HEAP8[$p$31>>0]|0; //@line 1625 "load_abc.cpp"
         $1789 = $1788 << 24 >> 24; //@line 1625 "load_abc.cpp"
         $isdigittmp1$i = (($1789) + -48)|0; //@line 1625 "load_abc.cpp"
         $isdigit2$i = ($isdigittmp1$i>>>0)<(10); //@line 1625 "load_abc.cpp"
         if ($isdigit2$i) {
          $1792 = $1789;$h$04$i = 0;$i$03$i = 0;
          while(1) {
           $1790 = ($h$04$i*10)|0; //@line 1626 "load_abc.cpp"
           $1791 = (($1792) + -48)|0; //@line 1626 "load_abc.cpp"
           $1793 = (($1791) + ($1790))|0; //@line 1626 "load_abc.cpp"
           $1794 = (($i$03$i) + 1)|0; //@line 1627 "load_abc.cpp"
           $1795 = (($p$31) + ($1794)|0); //@line 1625 "load_abc.cpp"
           $1796 = HEAP8[$1795>>0]|0; //@line 1625 "load_abc.cpp"
           $1797 = $1796 << 24 >> 24; //@line 1625 "load_abc.cpp"
           $isdigittmp$i292 = (($1797) + -48)|0; //@line 1625 "load_abc.cpp"
           $isdigit$i293 = ($isdigittmp$i292>>>0)<(10); //@line 1625 "load_abc.cpp"
           if ($isdigit$i293) {
            $1792 = $1797;$h$04$i = $1793;$i$03$i = $1794;
           } else {
            $h$0$lcssa$i295 = $1793;$i$0$lcssa$i296 = $1794;
            break;
           }
          }
         } else {
          $h$0$lcssa$i295 = 0;$i$0$lcssa$i296 = 0;
         }
         $1798 = ($i$0$lcssa$i296|0)==(0); //@line 1629 "load_abc.cpp"
         $$h$0$i = $1798 ? 1 : $h$0$lcssa$i295; //@line 1630 "load_abc.cpp"
         $1799 = (($p$31) + ($i$0$lcssa$i296)|0); //@line 4508 "load_abc.cpp"
         $1800 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4509 "load_abc.cpp"
         $1801 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4509 "load_abc.cpp"
         $1802 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
         HEAP8[$1802>>0]=0&255;HEAP8[$1802+1>>0]=(0>>8)&255;HEAP8[$1802+2>>0]=(0>>16)&255;HEAP8[$1802+3>>0]=0>>24; //@line 540 "load_abc.cpp"
         $1803 = ((($1802)) + 4|0); //@line 541 "load_abc.cpp"
         HEAP8[$1803>>0]=$1801&255;HEAP8[$1803+1>>0]=($1801>>8)&255;HEAP8[$1803+2>>0]=($1801>>16)&255;HEAP8[$1803+3>>0]=$1801>>24; //@line 541 "load_abc.cpp"
         $1804 = ((($1802)) + 8|0); //@line 543 "load_abc.cpp"
         HEAP8[$1804>>0] = 1; //@line 543 "load_abc.cpp"
         $1805 = ((($1802)) + 9|0); //@line 543 "load_abc.cpp"
         HEAP8[$1805>>0] = 86; //@line 543 "load_abc.cpp"
         $1806 = ((($1802)) + 10|0); //@line 543 "load_abc.cpp"
         $1807 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
         $1808 = ((($1802)) + 14|0); //@line 544 "load_abc.cpp"
         HEAP8[$1808>>0] = $1807; //@line 544 "load_abc.cpp"
         $1809 = ((($1802)) + 15|0); //@line 545 "load_abc.cpp"
         HEAP8[$1809>>0] = 0; //@line 545 "load_abc.cpp"
         $1810 = 1 << $$h$0$i; //@line 1482 "load_abc.cpp"
         HEAP8[$1806>>0]=$1810&255;HEAP8[$1806+1>>0]=($1810>>8)&255;HEAP8[$1806+2>>0]=($1810>>16)&255;HEAP8[$1806+3>>0]=$1810>>24; //@line 1482 "load_abc.cpp"
         __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($1800,$1802); //@line 1483 "load_abc.cpp"
         $1811 = HEAP8[$1799>>0]|0; //@line 4510 "load_abc.cpp"
         $1812 = $1811 & -2; //@line 4510 "load_abc.cpp"
         $switch1214 = ($1812<<24>>24)==(44); //@line 4510 "load_abc.cpp"
         if ($switch1214) {
          $1815 = $1811;$1845 = $$h$0$i;$p$371215 = $1799;
         } else {
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$13;$abcnoslurs$12 = $abcnoslurs$11;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $1799;
          break L678;
         }
         while(1) {
          $1814 = ($1815<<24>>24)==(44); //@line 4511 "load_abc.cpp"
          $1816 = ((($p$371215)) + 1|0); //@line 4512 "load_abc.cpp"
          $1817 = HEAP8[$1816>>0]|0; //@line 1625 "load_abc.cpp"
          $1818 = $1817 << 24 >> 24; //@line 1625 "load_abc.cpp"
          $isdigittmp1$i298 = (($1818) + -48)|0; //@line 1625 "load_abc.cpp"
          $isdigit2$i299 = ($isdigittmp1$i298>>>0)<(10); //@line 1625 "load_abc.cpp"
          do {
           if ($1814) {
            if ($isdigit2$i299) {
             $1821 = $1818;$h$04$i300 = 0;$i$03$i301 = 0;
             while(1) {
              $1819 = ($h$04$i300*10)|0; //@line 1626 "load_abc.cpp"
              $1820 = (($1821) + -48)|0; //@line 1626 "load_abc.cpp"
              $1822 = (($1820) + ($1819))|0; //@line 1626 "load_abc.cpp"
              $1823 = (($i$03$i301) + 1)|0; //@line 1627 "load_abc.cpp"
              $$sum821 = (($i$03$i301) + 2)|0; //@line 1625 "load_abc.cpp"
              $1824 = (($p$371215) + ($$sum821)|0); //@line 1625 "load_abc.cpp"
              $1825 = HEAP8[$1824>>0]|0; //@line 1625 "load_abc.cpp"
              $1826 = $1825 << 24 >> 24; //@line 1625 "load_abc.cpp"
              $isdigittmp$i302 = (($1826) + -48)|0; //@line 1625 "load_abc.cpp"
              $isdigit$i303 = ($isdigittmp$i302>>>0)<(10); //@line 1625 "load_abc.cpp"
              if ($isdigit$i303) {
               $1821 = $1826;$h$04$i300 = $1822;$i$03$i301 = $1823;
              } else {
               $h$0$lcssa$i305 = $1822;$i$0$lcssa$i306 = $1823;
               break;
              }
             }
            } else {
             $h$0$lcssa$i305 = 0;$i$0$lcssa$i306 = 0;
            }
            $1827 = ($i$0$lcssa$i306|0)==(0); //@line 1629 "load_abc.cpp"
            $$h$0$i307 = $1827 ? 1 : $h$0$lcssa$i305; //@line 1630 "load_abc.cpp"
            $1828 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4514 "load_abc.cpp"
            $$idx106 = ((($1828)) + 8|0);
            $$idx106$val = HEAPU8[$$idx106>>0]|(HEAPU8[$$idx106+1>>0]<<8)|(HEAPU8[$$idx106+2>>0]<<16)|(HEAPU8[$$idx106+3>>0]<<24);
            $1829 = 1 << $$h$0$i307; //@line 1488 "load_abc.cpp"
            $1830 = ((($$idx106$val)) + 10|0); //@line 1488 "load_abc.cpp"
            $1831 = HEAPU8[$1830>>0]|(HEAPU8[$1830+1>>0]<<8)|(HEAPU8[$1830+2>>0]<<16)|(HEAPU8[$1830+3>>0]<<24); //@line 1488 "load_abc.cpp"
            $1832 = $1831 | $1829; //@line 1488 "load_abc.cpp"
            HEAP8[$1830>>0]=$1832&255;HEAP8[$1830+1>>0]=($1832>>8)&255;HEAP8[$1830+2>>0]=($1832>>16)&255;HEAP8[$1830+3>>0]=$1832>>24; //@line 1488 "load_abc.cpp"
            $$be = $$h$0$i307;$$sum59$pn$in = $i$0$lcssa$i306;
           } else {
            if ($isdigit2$i299) {
             $1837 = $1818;$h$04$i312 = 0;$i$03$i313 = 0;
             while(1) {
              $1835 = ($h$04$i312*10)|0; //@line 1626 "load_abc.cpp"
              $1836 = (($1837) + -48)|0; //@line 1626 "load_abc.cpp"
              $1838 = (($1836) + ($1835))|0; //@line 1626 "load_abc.cpp"
              $1839 = (($i$03$i313) + 1)|0; //@line 1627 "load_abc.cpp"
              $$sum820 = (($i$03$i313) + 2)|0; //@line 1625 "load_abc.cpp"
              $1840 = (($p$371215) + ($$sum820)|0); //@line 1625 "load_abc.cpp"
              $1841 = HEAP8[$1840>>0]|0; //@line 1625 "load_abc.cpp"
              $1842 = $1841 << 24 >> 24; //@line 1625 "load_abc.cpp"
              $isdigittmp$i314 = (($1842) + -48)|0; //@line 1625 "load_abc.cpp"
              $isdigit$i315 = ($isdigittmp$i314>>>0)<(10); //@line 1625 "load_abc.cpp"
              if ($isdigit$i315) {
               $1837 = $1842;$h$04$i312 = $1838;$i$03$i313 = $1839;
              } else {
               $h$0$lcssa$i317 = $1838;$i$0$lcssa$i318 = $1839;
               break;
              }
             }
            } else {
             $h$0$lcssa$i317 = 0;$i$0$lcssa$i318 = 0;
            }
            $1843 = ($i$0$lcssa$i318|0)==(0); //@line 1629 "load_abc.cpp"
            $$h$0$i319 = $1843 ? 1 : $h$0$lcssa$i317; //@line 1630 "load_abc.cpp"
            $1844 = ($1845|0)<($$h$0$i319|0); //@line 4519 "load_abc.cpp"
            if (!($1844)) {
             $$be = $1845;$$sum59$pn$in = $i$0$lcssa$i318;
             break;
            }
            $1846 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4521 "load_abc.cpp"
            $$idx105 = ((($1846)) + 8|0);
            $$idx105$val = HEAPU8[$$idx105>>0]|(HEAPU8[$$idx105+1>>0]<<8)|(HEAPU8[$$idx105+2>>0]<<16)|(HEAPU8[$$idx105+3>>0]<<24);
            $1813 = ((($$idx105$val)) + 10|0); //@line 1488 "load_abc.cpp"
            $$promoted1212 = HEAPU8[$1813>>0]|(HEAPU8[$1813+1>>0]<<8)|(HEAPU8[$1813+2>>0]<<16)|(HEAPU8[$1813+3>>0]<<24); //@line 1488 "load_abc.cpp"
            $1848 = $1845;$1851 = $$promoted1212;
            while(1) {
             $1847 = (($1848) + 1)|0; //@line 4520 "load_abc.cpp"
             $1849 = 1 << $1847; //@line 1488 "load_abc.cpp"
             $1850 = $1851 | $1849; //@line 1488 "load_abc.cpp"
             $exitcond1502 = ($1847|0)==($$h$0$i319|0); //@line 4519 "load_abc.cpp"
             if ($exitcond1502) {
              $$lcssa2245 = $1850;
              break;
             } else {
              $1848 = $1847;$1851 = $1850;
             }
            }
            HEAP8[$1813>>0]=$$lcssa2245&255;HEAP8[$1813+1>>0]=($$lcssa2245>>8)&255;HEAP8[$1813+2>>0]=($$lcssa2245>>16)&255;HEAP8[$1813+3>>0]=$$lcssa2245>>24; //@line 1488 "load_abc.cpp"
            $$be = $$h$0$i319;$$sum59$pn$in = $i$0$lcssa$i318;
           }
          } while(0);
          $$sum59$pn = (($$sum59$pn$in) + 1)|0; //@line 4513 "load_abc.cpp"
          $p$37$be = (($p$371215) + ($$sum59$pn)|0); //@line 4513 "load_abc.cpp"
          $1833 = HEAP8[$p$37$be>>0]|0; //@line 4510 "load_abc.cpp"
          $1834 = $1833 & -2; //@line 4510 "load_abc.cpp"
          $switch = ($1834<<24>>24)==(44); //@line 4510 "load_abc.cpp"
          if ($switch) {
           $1815 = $1833;$1845 = $$be;$p$371215 = $p$37$be;
          } else {
           $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$13;$abcnoslurs$12 = $abcnoslurs$11;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$37$be;
           break;
          }
         }
        }
        break;
       }
       case 40:  {
        $1852 = HEAP8[$p$18>>0]|0; //@line 4531 "load_abc.cpp"
        $1853 = $1852 << 24 >> 24; //@line 4531 "load_abc.cpp"
        $isdigittmp47 = (($1853) + -48)|0; //@line 4531 "load_abc.cpp"
        $isdigit48 = ($isdigittmp47>>>0)<(10); //@line 4531 "load_abc.cpp"
        if ($isdigit48) {
         $1856 = $1853;$h$04$i324 = 0;$i$03$i325 = 0;
         while(1) {
          $1854 = ($h$04$i324*10)|0; //@line 1626 "load_abc.cpp"
          $1855 = (($1856) + -48)|0; //@line 1626 "load_abc.cpp"
          $1857 = (($1855) + ($1854))|0; //@line 1626 "load_abc.cpp"
          $1858 = (($i$03$i325) + 1)|0; //@line 1627 "load_abc.cpp"
          $1859 = (($p$18) + ($1858)|0); //@line 1625 "load_abc.cpp"
          $1860 = HEAP8[$1859>>0]|0; //@line 1625 "load_abc.cpp"
          $1861 = $1860 << 24 >> 24; //@line 1625 "load_abc.cpp"
          $isdigittmp$i326 = (($1861) + -48)|0; //@line 1625 "load_abc.cpp"
          $isdigit$i327 = ($isdigittmp$i326>>>0)<(10); //@line 1625 "load_abc.cpp"
          if ($isdigit$i327) {
           $1856 = $1861;$h$04$i324 = $1857;$i$03$i325 = $1858;
          } else {
           $$lcssa2214 = $1857;$$lcssa2215 = $1858;$$lcssa2216 = $1859;$$lcssa2217 = $1860;$i$03$i325$lcssa = $i$03$i325;
           break;
          }
         }
         $1862 = ($$lcssa2215|0)==(0); //@line 1629 "load_abc.cpp"
         $$h$0$i331 = $1862 ? 1 : $$lcssa2214; //@line 1630 "load_abc.cpp"
         switch ($$h$0$i331|0) {
         case 8: case 4: case 2:  {
          $4062 = 3;
          break;
         }
         case 6: case 3:  {
          $4062 = 2;
          break;
         }
         default: {
          $1863 = HEAP32[$snotediv>>2]|0; //@line 4542 "load_abc.cpp"
          $1864 = ($1863|0)==(8); //@line 4542 "load_abc.cpp"
          $$791 = $1864 ? 3 : 2; //@line 4542 "load_abc.cpp"
          $4062 = $$791;
         }
         }
         $1865 = ($$lcssa2217<<24>>24)==(58); //@line 4548 "load_abc.cpp"
         if (!($1865)) {
          $4057 = $1637;$4058 = $$h$0$i331;$4059 = $4062;$4060 = $$h$0$i331;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $$lcssa2216;
          break L678;
         }
         $$sum49 = (($i$03$i325$lcssa) + 2)|0; //@line 4549 "load_abc.cpp"
         $1866 = (($p$18) + ($$sum49)|0); //@line 4549 "load_abc.cpp"
         $1867 = HEAP8[$1866>>0]|0; //@line 4550 "load_abc.cpp"
         $1868 = $1867 << 24 >> 24; //@line 4550 "load_abc.cpp"
         $isdigittmp50 = (($1868) + -48)|0; //@line 4550 "load_abc.cpp"
         $isdigit51 = ($isdigittmp50>>>0)<(10); //@line 4550 "load_abc.cpp"
         if ($isdigit51) {
          $1871 = $1868;$h$04$i336 = 0;$i$03$i337 = 0;
          while(1) {
           $1869 = ($h$04$i336*10)|0; //@line 1626 "load_abc.cpp"
           $1870 = (($1871) + -48)|0; //@line 1626 "load_abc.cpp"
           $1872 = (($1870) + ($1869))|0; //@line 1626 "load_abc.cpp"
           $1873 = (($i$03$i337) + 1)|0; //@line 1627 "load_abc.cpp"
           $$sum810 = (($1873) + ($$sum49))|0; //@line 1625 "load_abc.cpp"
           $1874 = (($p$18) + ($$sum810)|0); //@line 1625 "load_abc.cpp"
           $1875 = HEAP8[$1874>>0]|0; //@line 1625 "load_abc.cpp"
           $1876 = $1875 << 24 >> 24; //@line 1625 "load_abc.cpp"
           $isdigittmp$i338 = (($1876) + -48)|0; //@line 1625 "load_abc.cpp"
           $isdigit$i339 = ($isdigittmp$i338>>>0)<(10); //@line 1625 "load_abc.cpp"
           if ($isdigit$i339) {
            $1871 = $1876;$h$04$i336 = $1872;$i$03$i337 = $1873;
           } else {
            $$lcssa2218 = $1872;$$lcssa2219 = $1873;$$lcssa2220 = $1874;$$lcssa2221 = $1875;$$sum810$lcssa = $$sum810;
            break;
           }
          }
          $1877 = ($$lcssa2219|0)==(0); //@line 1629 "load_abc.cpp"
          $$h$0$i343 = $1877 ? 1 : $$lcssa2218; //@line 1630 "load_abc.cpp"
          $1879 = $$lcssa2221;$1880 = $$sum810$lcssa;$4063 = $$h$0$i343;$p$38 = $$lcssa2220;
         } else {
          $1879 = $1867;$1880 = $$sum49;$4063 = $4062;$p$38 = $1866;
         }
         $1878 = ($1879<<24>>24)==(58); //@line 4551 "load_abc.cpp"
         if (!($1878)) {
          $4057 = $1637;$4058 = $$h$0$i331;$4059 = $4063;$4060 = $$h$0$i331;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$38;
          break L678;
         }
         $$sum78 = (($1880) + 1)|0; //@line 4552 "load_abc.cpp"
         $1881 = (($p$18) + ($$sum78)|0); //@line 4552 "load_abc.cpp"
         $1882 = HEAP8[$1881>>0]|0; //@line 4553 "load_abc.cpp"
         $1883 = $1882 << 24 >> 24; //@line 4553 "load_abc.cpp"
         $isdigittmp52 = (($1883) + -48)|0; //@line 4553 "load_abc.cpp"
         $isdigit53 = ($isdigittmp52>>>0)<(10); //@line 4553 "load_abc.cpp"
         if ($isdigit53) {
          $1886 = $1883;$h$04$i348 = 0;$i$03$i349 = 0;
         } else {
          $4057 = $1637;$4058 = $$h$0$i331;$4059 = $4063;$4060 = $$h$0$i331;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $1881;
          break L678;
         }
         while(1) {
          $1884 = ($h$04$i348*10)|0; //@line 1626 "load_abc.cpp"
          $1885 = (($1886) + -48)|0; //@line 1626 "load_abc.cpp"
          $1887 = (($1885) + ($1884))|0; //@line 1626 "load_abc.cpp"
          $1888 = (($i$03$i349) + 1)|0; //@line 1627 "load_abc.cpp"
          $$sum809 = (($1888) + ($$sum78))|0; //@line 1625 "load_abc.cpp"
          $1889 = (($p$18) + ($$sum809)|0); //@line 1625 "load_abc.cpp"
          $1890 = HEAP8[$1889>>0]|0; //@line 1625 "load_abc.cpp"
          $1891 = $1890 << 24 >> 24; //@line 1625 "load_abc.cpp"
          $isdigittmp$i350 = (($1891) + -48)|0; //@line 1625 "load_abc.cpp"
          $isdigit$i351 = ($isdigittmp$i350>>>0)<(10); //@line 1625 "load_abc.cpp"
          if ($isdigit$i351) {
           $1886 = $1891;$h$04$i348 = $1887;$i$03$i349 = $1888;
          } else {
           $$lcssa2222 = $1887;$$lcssa2223 = $1888;$i$03$i349$lcssa = $i$03$i349;
           break;
          }
         }
         $1892 = ($$lcssa2223|0)==(0); //@line 1629 "load_abc.cpp"
         $$h$0$i355 = $1892 ? 1 : $$lcssa2222; //@line 1630 "load_abc.cpp"
         $$sum54 = (($1880) + 2)|0; //@line 4553 "load_abc.cpp"
         $$sum79 = (($$sum54) + ($i$03$i349$lcssa))|0; //@line 4553 "load_abc.cpp"
         $1893 = (($p$18) + ($$sum79)|0); //@line 4553 "load_abc.cpp"
         $4057 = $1637;$4058 = $$h$0$i331;$4059 = $4063;$4060 = $$h$0$i355;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $1893;
        } else {
         $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = 0;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$18;
        }
        break;
       }
       case 41:  {
        $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnolegato$101227;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$18;
        break;
       }
       case 123:  {
        $1894 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4566 "load_abc.cpp"
        $1895 = ($1894|0)==(0|0); //@line 915 "load_abc.cpp"
        if ($1895) {
         $1896 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
         $1897 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
         $1898 = ((($1896)) + 20|0); //@line 917 "load_abc.cpp"
         HEAP8[$1898>>0]=$1897&255;HEAP8[$1898+1>>0]=($1897>>8)&255;HEAP8[$1898+2>>0]=($1897>>16)&255;HEAP8[$1898+3>>0]=$1897>>24; //@line 917 "load_abc.cpp"
         $$0$i358 = $1896;
        } else {
         $$0$i358 = $1894;
        }
        HEAP8[$41>>0]=$$0$i358&255;HEAP8[$41+1>>0]=($$0$i358>>8)&255;HEAP8[$41+2>>0]=($$0$i358>>16)&255;HEAP8[$41+3>>0]=$$0$i358>>24; //@line 4566 "load_abc.cpp"
        $1899 = ((($$0$i358)) + 36|0); //@line 1227 "load_abc.cpp"
        $1900 = HEAP8[$1899>>0]|0; //@line 1227 "load_abc.cpp"
        $tp$01$i360 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 1228 "load_abc.cpp"
        $1901 = ($tp$01$i360|0)==(0|0); //@line 1228 "load_abc.cpp"
        if (!($1901)) {
         $tp$02$i361 = $tp$01$i360;
         while(1) {
          $1902 = ((($tp$02$i361)) + 36|0); //@line 1229 "load_abc.cpp"
          $1903 = HEAP8[$1902>>0]|0; //@line 1229 "load_abc.cpp"
          $1904 = ($1903<<24>>24)==($1900<<24>>24); //@line 1229 "load_abc.cpp"
          if ($1904) {
           $1905 = ((($tp$02$i361)) + 37|0); //@line 1229 "load_abc.cpp"
           $1906 = HEAP8[$1905>>0]|0; //@line 1229 "load_abc.cpp"
           $1907 = ((($tp$02$i361)) + 38|0); //@line 1229 "load_abc.cpp"
           HEAP8[$1907>>0] = $1906; //@line 1229 "load_abc.cpp"
          }
          $tp$0$i363 = HEAPU8[$tp$02$i361>>0]|(HEAPU8[$tp$02$i361+1>>0]<<8)|(HEAPU8[$tp$02$i361+2>>0]<<16)|(HEAPU8[$tp$02$i361+3>>0]<<24); //@line 1228 "load_abc.cpp"
          $1908 = ($tp$0$i363|0)==(0|0); //@line 1228 "load_abc.cpp"
          if ($1908) {
           break;
          } else {
           $tp$02$i361 = $tp$0$i363;
          }
         }
        }
        $1909 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4569 "load_abc.cpp"
        $1910 = (($1909) - ($bartime$41235))|0; //@line 4569 "load_abc.cpp"
        $1911 = HEAP32[$mnotediv>>2]|0; //@line 4569 "load_abc.cpp"
        $1912 = ($1911|0)==(0); //@line 345 "load_abc.cpp"
        if ($1912) {
         $$0$i366 = 0;
        } else {
         $1913 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4569 "load_abc.cpp"
         $1914 = ($1913*3072)|0; //@line 346 "load_abc.cpp"
         $1915 = HEAP32[299862]|0; //@line 346 "load_abc.cpp"
         $1916 = Math_imul($1914, $1915)|0; //@line 346 "load_abc.cpp"
         $1917 = HEAP32[299863]|0; //@line 346 "load_abc.cpp"
         $1918 = Math_imul($1917, $1911)|0; //@line 346 "load_abc.cpp"
         $1919 = (($1916|0) / ($1918|0))&-1; //@line 346 "load_abc.cpp"
         $$0$i366 = $1919;
        }
        $1920 = (($1910>>>0) / ($$0$i366>>>0))&-1; //@line 4569 "load_abc.cpp"
        $1921 = HEAPU8[$3>>0]|(HEAPU8[$3+1>>0]<<8)|(HEAPU8[$3+2>>0]<<16)|(HEAPU8[$3+3>>0]<<24); //@line 3546 "load_abc.cpp"
        $1922 = ($1921|0)==(0|0); //@line 3546 "load_abc.cpp"
        L1223: do {
         if ($1922) {
          $1936 = HEAP8[$105>>0]|0; //@line 3565 "load_abc.cpp"
          $1937 = $1936&255; //@line 3565 "load_abc.cpp"
          $1938 = (($1920|0) % ($1937|0))&-1; //@line 3565 "load_abc.cpp"
          $1939 = ($1938|0)==(0); //@line 3565 "load_abc.cpp"
          if (!($1939)) {
           $1945 = HEAP8[$104>>0]|0; //@line 3572 "load_abc.cpp"
           $1946 = $1945&255; //@line 3572 "load_abc.cpp"
           $vol$0$i368 = $1946;
           break;
          }
          $1940 = ($1920|0)==(0); //@line 3566 "load_abc.cpp"
          if ($1940) {
           $1943 = HEAP8[$102>>0]|0; //@line 3569 "load_abc.cpp"
           $1944 = $1943&255; //@line 3569 "load_abc.cpp"
           $vol$0$i368 = $1944;
           break;
          } else {
           $1941 = HEAP8[$103>>0]|0; //@line 3567 "load_abc.cpp"
           $1942 = $1941&255; //@line 3567 "load_abc.cpp"
           $vol$0$i368 = $1942;
           break;
          }
         } else {
          $1923 = HEAP8[$104>>0]|0; //@line 3547 "load_abc.cpp"
          $1924 = $1923&255; //@line 3547 "load_abc.cpp"
          $1925 = ($1924*9)|0; //@line 3547 "load_abc.cpp"
          $1926 = (($1925>>>0) / 10)&-1; //@line 3547 "load_abc.cpp"
          $1927 = (_strlen($1921)|0); //@line 3548 "load_abc.cpp"
          $1928 = ($1927|0)>($1920|0); //@line 3548 "load_abc.cpp"
          if (!($1928)) {
           $vol$0$i368 = $1926;
           break;
          }
          $1929 = (($1921) + ($1920)|0); //@line 3549 "load_abc.cpp"
          $1930 = HEAP8[$1929>>0]|0; //@line 3549 "load_abc.cpp"
          $1931 = $1930 << 24 >> 24; //@line 3549 "load_abc.cpp"
          switch ($1931|0) {
          case 102:  {
           $1932 = HEAP8[$102>>0]|0; //@line 3551 "load_abc.cpp"
           $1933 = $1932&255; //@line 3551 "load_abc.cpp"
           $vol$0$i368 = $1933;
           break L1223;
           break;
          }
          case 109:  {
           $1934 = HEAP8[$103>>0]|0; //@line 3554 "load_abc.cpp"
           $1935 = $1934&255; //@line 3554 "load_abc.cpp"
           $vol$0$i368 = $1935;
           break L1223;
           break;
          }
          case 112:  {
           $vol$0$i368 = $1924;
           break L1223;
           break;
          }
          default: {
           $vol$0$i368 = $1926;
           break L1223;
          }
          }
         }
        } while(0);
        $1947 = Math_imul($vol$0$i368, $abcvol$31225)|0; //@line 3574 "load_abc.cpp"
        $1948 = (($1947|0) / 128)&-1; //@line 3575 "load_abc.cpp"
        $abcgrace$6 = 0;$p$39 = $p$18;
        L1238: while(1) {
         $1949 = ((($p$39)) + 1|0); //@line 4570 "load_abc.cpp"
         $1950 = HEAP8[$p$39>>0]|0; //@line 4570 "load_abc.cpp"
         switch ($1950<<24>>24) {
         case 125: case 0:  {
          $$lcssa2212 = $1949;$$lcssa2213 = $1950;$abcgrace$6$lcssa = $abcgrace$6;
          break L1238;
          break;
         }
         default: {
         }
         }
         $1951 = $1950 << 24 >> 24; //@line 4570 "load_abc.cpp"
         $1952 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4571 "load_abc.cpp"
         $1953 = (($1952) + ($abcgrace$6))|0; //@line 4571 "load_abc.cpp"
         $1954 = (__ZL14abc_add_noteonP10_ABCHANDLEiPKcjPci17ABCEVENT_X_EFFECTi($1,$1951,$1949,$1953,$barsig,$1948,0,0)|0); //@line 4571 "load_abc.cpp"
         $$sum44 = (($1954) + 1)|0; //@line 4571 "load_abc.cpp"
         $1955 = (($p$39) + ($$sum44)|0); //@line 4571 "load_abc.cpp"
         $1956 = HEAP8[$1955>>0]|0; //@line 1625 "load_abc.cpp"
         $1957 = $1956 << 24 >> 24; //@line 1625 "load_abc.cpp"
         $isdigittmp1$i$i370 = (($1957) + -48)|0; //@line 1625 "load_abc.cpp"
         $isdigit2$i$i371 = ($isdigittmp1$i$i370>>>0)<(10); //@line 1625 "load_abc.cpp"
         if ($isdigit2$i$i371) {
          $1960 = $1957;$h$04$i$i372 = 0;$i$03$i$i373 = 0;
          while(1) {
           $1958 = ($h$04$i$i372*10)|0; //@line 1626 "load_abc.cpp"
           $1959 = (($1960) + -48)|0; //@line 1626 "load_abc.cpp"
           $1961 = (($1959) + ($1958))|0; //@line 1626 "load_abc.cpp"
           $1962 = (($i$03$i$i373) + 1)|0; //@line 1627 "load_abc.cpp"
           $$sum808 = (($1962) + ($$sum44))|0; //@line 1625 "load_abc.cpp"
           $1963 = (($p$39) + ($$sum808)|0); //@line 1625 "load_abc.cpp"
           $1964 = HEAP8[$1963>>0]|0; //@line 1625 "load_abc.cpp"
           $1965 = $1964 << 24 >> 24; //@line 1625 "load_abc.cpp"
           $isdigittmp$i$i374 = (($1965) + -48)|0; //@line 1625 "load_abc.cpp"
           $isdigit$i$i375 = ($isdigittmp$i$i374>>>0)<(10); //@line 1625 "load_abc.cpp"
           if ($isdigit$i$i375) {
            $1960 = $1965;$h$04$i$i372 = $1961;$i$03$i$i373 = $1962;
           } else {
            $h$0$lcssa$i$i377 = $1961;$i$0$lcssa$i$i378 = $1962;
            break;
           }
          }
         } else {
          $h$0$lcssa$i$i377 = 0;$i$0$lcssa$i$i378 = 0;
         }
         $1966 = ($i$0$lcssa$i$i378|0)==(0); //@line 1629 "load_abc.cpp"
         $$sum805 = (($i$0$lcssa$i$i378) + ($$sum44))|0; //@line 1666 "load_abc.cpp"
         $1967 = (($p$39) + ($$sum805)|0); //@line 1666 "load_abc.cpp"
         $1968 = HEAP8[$1967>>0]|0; //@line 1666 "load_abc.cpp"
         $1969 = ($1968<<24>>24)==(47); //@line 1666 "load_abc.cpp"
         if ($1969) {
          $h$017$i381 = 1;$i$016$i382 = $i$0$lcssa$i$i378;
          while(1) {
           $1970 = $h$017$i381 << 1; //@line 1667 "load_abc.cpp"
           $1971 = (($i$016$i382) + 1)|0; //@line 1668 "load_abc.cpp"
           $$sum807 = (($1971) + ($$sum44))|0; //@line 1666 "load_abc.cpp"
           $1972 = (($p$39) + ($$sum807)|0); //@line 1666 "load_abc.cpp"
           $1973 = HEAP8[$1972>>0]|0; //@line 1666 "load_abc.cpp"
           $1974 = ($1973<<24>>24)==(47); //@line 1666 "load_abc.cpp"
           if ($1974) {
            $h$017$i381 = $1970;$i$016$i382 = $1971;
           } else {
            $$lcssa13$i384 = $1973;$h$0$lcssa$i385 = $1970;$i$0$lcssa$i386 = $1971;
            break;
           }
          }
         } else {
          $$lcssa13$i384 = $1968;$h$0$lcssa$i385 = 1;$i$0$lcssa$i386 = $i$0$lcssa$i$i378;
         }
         $1975 = $$lcssa13$i384 << 24 >> 24; //@line 1666 "load_abc.cpp"
         $isdigittmp$i387 = (($1975) + -48)|0; //@line 1670 "load_abc.cpp"
         $isdigit$i388 = ($isdigittmp$i387>>>0)<(10); //@line 1670 "load_abc.cpp"
         if ($isdigit$i388) {
          $1978 = $1975;$h$04$i3$i390 = 0;$i$03$i4$i391 = 0;
          while(1) {
           $1976 = ($h$04$i3$i390*10)|0; //@line 1626 "load_abc.cpp"
           $1977 = (($1978) + -48)|0; //@line 1626 "load_abc.cpp"
           $1979 = (($1977) + ($1976))|0; //@line 1626 "load_abc.cpp"
           $1980 = (($i$03$i4$i391) + 1)|0; //@line 1627 "load_abc.cpp"
           $$sum$i392 = (($1980) + ($i$0$lcssa$i386))|0; //@line 1625 "load_abc.cpp"
           $$sum806 = (($$sum$i392) + ($$sum44))|0; //@line 1625 "load_abc.cpp"
           $1981 = (($p$39) + ($$sum806)|0); //@line 1625 "load_abc.cpp"
           $1982 = HEAP8[$1981>>0]|0; //@line 1625 "load_abc.cpp"
           $1983 = $1982 << 24 >> 24; //@line 1625 "load_abc.cpp"
           $isdigittmp$i5$i393 = (($1983) + -48)|0; //@line 1625 "load_abc.cpp"
           $isdigit$i6$i394 = ($isdigittmp$i5$i393>>>0)<(10); //@line 1625 "load_abc.cpp"
           if ($isdigit$i6$i394) {
            $1978 = $1983;$h$04$i3$i390 = $1979;$i$03$i4$i391 = $1980;
           } else {
            $$lcssa2210 = $1979;$$lcssa2211 = $1980;$$sum$i392$lcssa = $$sum$i392;
            break;
           }
          }
          $1984 = (($h$0$lcssa$i385|0) / 2)&-1; //@line 1671 "load_abc.cpp"
          $1985 = ($$lcssa2211|0)==(0); //@line 1629 "load_abc.cpp"
          $$h$0$i10$i396 = $1985 ? 1 : $$lcssa2210; //@line 1630 "load_abc.cpp"
          $1999 = $$h$0$i10$i396;$h$1$i399 = $1984;$i$1$i398 = $$sum$i392$lcssa;
         } else {
          $1999 = 1;$h$1$i399 = $h$0$lcssa$i385;$i$1$i398 = $i$0$lcssa$i386;
         }
         $$sum45 = (($i$1$i398) + ($$sum44))|0; //@line 4572 "load_abc.cpp"
         $1986 = (($p$39) + ($$sum45)|0); //@line 4572 "load_abc.cpp"
         $1987 = HEAP8[$1986>>0]|0; //@line 4573 "load_abc.cpp"
         $1988 = ($1987<<24>>24)==(45); //@line 4573 "load_abc.cpp"
         do {
          if ($1988) {
           $$sum46 = (($$sum45) + 1)|0; //@line 4574 "load_abc.cpp"
           $1989 = (($p$39) + ($$sum46)|0); //@line 4574 "load_abc.cpp"
           $1990 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4575 "load_abc.cpp"
           $1991 = ((($1990)) + 8|0); //@line 4575 "load_abc.cpp"
           $1992 = HEAPU8[$1991>>0]|(HEAPU8[$1991+1>>0]<<8)|(HEAPU8[$1991+2>>0]<<16)|(HEAPU8[$1991+3>>0]<<24); //@line 4575 "load_abc.cpp"
           $1993 = ((($1992)) + 8|0); //@line 4575 "load_abc.cpp"
           $1994 = HEAP8[$1993>>0]|0; //@line 4575 "load_abc.cpp"
           $1995 = ($1994<<24>>24)==(1); //@line 4575 "load_abc.cpp"
           if ($1995) {
            $p$40 = $1989;
            break;
           }
           $1996 = ((($1990)) + 16|0); //@line 4576 "load_abc.cpp"
           HEAP8[$1996>>0]=$1992&255;HEAP8[$1996+1>>0]=($1992>>8)&255;HEAP8[$1996+2>>0]=($1992>>16)&255;HEAP8[$1996+3>>0]=$1992>>24; //@line 4576 "load_abc.cpp"
           $p$40 = $1989;
          } else {
           $p$40 = $1986;
          }
         } while(0);
         $1997 = HEAP32[$snotediv>>2]|0; //@line 4579 "load_abc.cpp"
         $1998 = $1999 << 2; //@line 1675 "load_abc.cpp"
         $2000 = Math_imul($1998, $h$1$i399)|0; //@line 4578 "load_abc.cpp"
         $2001 = Math_imul($2000, $1997)|0; //@line 4579 "load_abc.cpp"
         $2002 = ($2001|0)==(0); //@line 345 "load_abc.cpp"
         if ($2002) {
          $$0$i401 = 0;
         } else {
          $2003 = HEAP32[$snotelen>>2]|0; //@line 4579 "load_abc.cpp"
          $2004 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4579 "load_abc.cpp"
          $2005 = HEAP32[299862]|0; //@line 346 "load_abc.cpp"
          $h$0$lcssa$i$i377$op = ($h$0$lcssa$i$i377*3072)|0; //@line 346 "load_abc.cpp"
          $2006 = $1966 ? 3072 : $h$0$lcssa$i$i377$op; //@line 346 "load_abc.cpp"
          $2007 = Math_imul($2006, $2003)|0; //@line 4579 "load_abc.cpp"
          $2008 = Math_imul($2007, $2004)|0; //@line 346 "load_abc.cpp"
          $2009 = Math_imul($2008, $2005)|0; //@line 346 "load_abc.cpp"
          $2010 = HEAP32[299863]|0; //@line 346 "load_abc.cpp"
          $2011 = Math_imul($2010, $2001)|0; //@line 346 "load_abc.cpp"
          $2012 = (($2009|0) / ($2011|0))&-1; //@line 346 "load_abc.cpp"
          $$0$i401 = $2012;
         }
         $2013 = (($$0$i401) + ($abcgrace$6))|0; //@line 4579 "load_abc.cpp"
         $2014 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4580 "load_abc.cpp"
         $2015 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4580 "load_abc.cpp"
         $2016 = (($2015) + ($2013))|0; //@line 4580 "load_abc.cpp"
         $2017 = ((($2014)) + 42|0); //@line 1174 "load_abc.cpp"
         $2018 = HEAP8[$2017>>0]|0; //@line 1174 "load_abc.cpp"
         $2019 = $2018&255; //@line 1174 "load_abc.cpp"
         $2020 = (_pat_gmtosmp($2019)|0); //@line 1174 "load_abc.cpp"
         $2021 = $2020&255; //@line 1174 "load_abc.cpp"
         $2022 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
         HEAP8[$2022>>0]=0&255;HEAP8[$2022+1>>0]=(0>>8)&255;HEAP8[$2022+2>>0]=(0>>16)&255;HEAP8[$2022+3>>0]=0>>24; //@line 540 "load_abc.cpp"
         $2023 = ((($2022)) + 4|0); //@line 541 "load_abc.cpp"
         HEAP8[$2023>>0]=$2016&255;HEAP8[$2023+1>>0]=($2016>>8)&255;HEAP8[$2023+2>>0]=($2016>>16)&255;HEAP8[$2023+3>>0]=$2016>>24; //@line 541 "load_abc.cpp"
         $2024 = ((($2022)) + 8|0); //@line 543 "load_abc.cpp"
         HEAP8[$2024>>0] = 0; //@line 543 "load_abc.cpp"
         $2025 = ((($2022)) + 9|0); //@line 543 "load_abc.cpp"
         HEAP8[$2025>>0] = 0; //@line 543 "load_abc.cpp"
         $2026 = ((($2022)) + 10|0); //@line 543 "load_abc.cpp"
         HEAP8[$2026>>0] = $2021; //@line 543 "load_abc.cpp"
         $2027 = ((($2022)) + 11|0); //@line 543 "load_abc.cpp"
         HEAP8[$2027>>0] = 0; //@line 543 "load_abc.cpp"
         $2028 = ((($2022)) + 12|0); //@line 543 "load_abc.cpp"
         HEAP8[$2028>>0] = 0; //@line 543 "load_abc.cpp"
         $2029 = ((($2022)) + 13|0); //@line 543 "load_abc.cpp"
         HEAP8[$2029>>0] = 0; //@line 543 "load_abc.cpp"
         $2030 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
         $2031 = ((($2022)) + 14|0); //@line 544 "load_abc.cpp"
         HEAP8[$2031>>0] = $2030; //@line 544 "load_abc.cpp"
         $2032 = ((($2022)) + 15|0); //@line 545 "load_abc.cpp"
         HEAP8[$2032>>0] = 0; //@line 545 "load_abc.cpp"
         __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($2014,$2022); //@line 1179 "load_abc.cpp"
         $abcgrace$6 = $2013;$p$39 = $p$40;
        }
        $2033 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4582 "load_abc.cpp"
        $2034 = (($2033) + ($abcgrace$6$lcssa))|0; //@line 4582 "load_abc.cpp"
        HEAP8[$31>>0]=$2034&255;HEAP8[$31+1>>0]=($2034>>8)&255;HEAP8[$31+2>>0]=($2034>>16)&255;HEAP8[$31+3>>0]=$2034>>24; //@line 4582 "load_abc.cpp"
        $2035 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4583 "load_abc.cpp"
        $2036 = ((($2035)) + 8|0); //@line 1590 "load_abc.cpp"
        $2037 = HEAPU8[$2036>>0]|(HEAPU8[$2036+1>>0]<<8)|(HEAPU8[$2036+2>>0]<<16)|(HEAPU8[$2036+3>>0]<<24); //@line 1590 "load_abc.cpp"
        $2038 = ($2037|0)==(0|0); //@line 1591 "load_abc.cpp"
        do {
         if ($2038) {
          label = 678;
         } else {
          $2039 = ((($2037)) + 4|0); //@line 1591 "load_abc.cpp"
          $2040 = HEAPU8[$2039>>0]|(HEAPU8[$2039+1>>0]<<8)|(HEAPU8[$2039+2>>0]<<16)|(HEAPU8[$2039+3>>0]<<24); //@line 1591 "load_abc.cpp"
          $2041 = ($2040|0)==($2034|0); //@line 1591 "load_abc.cpp"
          if ($2041) {
           break;
          }
          $2042 = ((($2037)) + 8|0); //@line 1592 "load_abc.cpp"
          $2043 = HEAP8[$2042>>0]|0; //@line 1592 "load_abc.cpp"
          $2044 = ($2043<<24>>24)==(1); //@line 1592 "load_abc.cpp"
          if (!($2044)) {
           label = 678;
           break;
          }
          $2045 = ((($2037)) + 9|0); //@line 1592 "load_abc.cpp"
          $2046 = HEAP8[$2045>>0]|0; //@line 1592 "load_abc.cpp"
          $2047 = ($2046<<24>>24)==(83); //@line 1592 "load_abc.cpp"
          if (!($2047)) {
           label = 678;
           break;
          }
          HEAP8[$2039>>0]=$2034&255;HEAP8[$2039+1>>0]=($2034>>8)&255;HEAP8[$2039+2>>0]=($2034>>16)&255;HEAP8[$2039+3>>0]=$2034>>24; //@line 1593 "load_abc.cpp"
         }
        } while(0);
        if ((label|0) == 678) {
         label = 0;
         $2048 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
         HEAP8[$2048>>0]=0&255;HEAP8[$2048+1>>0]=(0>>8)&255;HEAP8[$2048+2>>0]=(0>>16)&255;HEAP8[$2048+3>>0]=0>>24; //@line 540 "load_abc.cpp"
         $2049 = ((($2048)) + 4|0); //@line 541 "load_abc.cpp"
         HEAP8[$2049>>0]=$2034&255;HEAP8[$2049+1>>0]=($2034>>8)&255;HEAP8[$2049+2>>0]=($2034>>16)&255;HEAP8[$2049+3>>0]=$2034>>24; //@line 541 "load_abc.cpp"
         $2050 = ((($2048)) + 8|0); //@line 543 "load_abc.cpp"
         HEAP8[$2050>>0] = 1; //@line 543 "load_abc.cpp"
         $2051 = ((($2048)) + 9|0); //@line 543 "load_abc.cpp"
         HEAP8[$2051>>0] = 83; //@line 543 "load_abc.cpp"
         $2052 = ((($2048)) + 10|0); //@line 543 "load_abc.cpp"
         HEAP8[$2052>>0]=0&255;HEAP8[$2052+1>>0]=(0>>8)&255;HEAP8[$2052+2>>0]=(0>>16)&255;HEAP8[$2052+3>>0]=0>>24; //@line 543 "load_abc.cpp"
         $2053 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
         $2054 = ((($2048)) + 14|0); //@line 544 "load_abc.cpp"
         HEAP8[$2054>>0] = $2053; //@line 544 "load_abc.cpp"
         $2055 = ((($2048)) + 15|0); //@line 545 "load_abc.cpp"
         HEAP8[$2055>>0] = 0; //@line 545 "load_abc.cpp"
         __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($2035,$2048); //@line 1600 "load_abc.cpp"
        }
        $2056 = HEAPU8[$39>>0]|(HEAPU8[$39+1>>0]<<8)|(HEAPU8[$39+2>>0]<<16)|(HEAPU8[$39+3>>0]<<24); //@line 4584 "load_abc.cpp"
        $2057 = ($2056|0)==(0); //@line 4584 "load_abc.cpp"
        do {
         if (!($2057)) {
          $2058 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4584 "load_abc.cpp"
          $2059 = HEAPU8[$42>>0]|(HEAPU8[$42+1>>0]<<8)|(HEAPU8[$42+2>>0]<<16)|(HEAPU8[$42+3>>0]<<24); //@line 4584 "load_abc.cpp"
          $2060 = ($2058|0)==($2059|0); //@line 4584 "load_abc.cpp"
          if (!($2060)) {
           break;
          }
          $2061 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4585 "load_abc.cpp"
          __ZL14abc_add_gchordP10_ABCHANDLEjj($1,$2061,$bartime$41235); //@line 4585 "load_abc.cpp"
         }
        } while(0);
        $2062 = HEAPU8[$38>>0]|(HEAPU8[$38+1>>0]<<8)|(HEAPU8[$38+2>>0]<<16)|(HEAPU8[$38+3>>0]<<24); //@line 4586 "load_abc.cpp"
        $2063 = ($2062|0)==(0); //@line 4586 "load_abc.cpp"
        if ($2063) {
         $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$6$lcssa;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $$lcssa2213;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $$lcssa2212;
         break L678;
        }
        $2064 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4586 "load_abc.cpp"
        $2065 = HEAPU8[$43>>0]|(HEAPU8[$43+1>>0]<<8)|(HEAPU8[$43+2>>0]<<16)|(HEAPU8[$43+3>>0]<<24); //@line 4586 "load_abc.cpp"
        $2066 = ($2064|0)==($2065|0); //@line 4586 "load_abc.cpp"
        if (!($2066)) {
         $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$6$lcssa;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $$lcssa2213;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $$lcssa2212;
         break L678;
        }
        $2067 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4587 "load_abc.cpp"
        __ZL12abc_add_drumP10_ABCHANDLEjj($1,$2067,$bartime$41235); //@line 4587 "load_abc.cpp"
        $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$6$lcssa;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $$lcssa2213;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $$lcssa2212;
        break;
       }
       case 124:  {
        $2068 = HEAPU8[$39>>0]|(HEAPU8[$39+1>>0]<<8)|(HEAPU8[$39+2>>0]<<16)|(HEAPU8[$39+3>>0]<<24); //@line 4591 "load_abc.cpp"
        $2069 = ($2068|0)==(0); //@line 4591 "load_abc.cpp"
        do {
         if (!($2069)) {
          $2070 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4591 "load_abc.cpp"
          $2071 = ($2070|0)==(0|0); //@line 4591 "load_abc.cpp"
          if ($2071) {
           break;
          }
          $2072 = HEAPU8[$42>>0]|(HEAPU8[$42+1>>0]<<8)|(HEAPU8[$42+2>>0]<<16)|(HEAPU8[$42+3>>0]<<24); //@line 4591 "load_abc.cpp"
          $2073 = ($2070|0)==($2072|0); //@line 4591 "load_abc.cpp"
          if (!($2073)) {
           break;
          }
          $2074 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4592 "load_abc.cpp"
          __ZL14abc_add_gchordP10_ABCHANDLEjj($1,$2074,$bartime$41235); //@line 4592 "load_abc.cpp"
         }
        } while(0);
        $2075 = HEAPU8[$38>>0]|(HEAPU8[$38+1>>0]<<8)|(HEAPU8[$38+2>>0]<<16)|(HEAPU8[$38+3>>0]<<24); //@line 4593 "load_abc.cpp"
        $2076 = ($2075|0)==(0); //@line 4593 "load_abc.cpp"
        do {
         if (!($2076)) {
          $2077 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4593 "load_abc.cpp"
          $2078 = HEAPU8[$43>>0]|(HEAPU8[$43+1>>0]<<8)|(HEAPU8[$43+2>>0]<<16)|(HEAPU8[$43+3>>0]<<24); //@line 4593 "load_abc.cpp"
          $2079 = ($2077|0)==($2078|0); //@line 4593 "load_abc.cpp"
          if (!($2079)) {
           break;
          }
          $2080 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4594 "load_abc.cpp"
          __ZL12abc_add_drumP10_ABCHANDLEjj($1,$2080,$bartime$41235); //@line 4594 "load_abc.cpp"
         }
        } while(0);
        $2081 = (47112 + ($abckey$41224<<2)|0); //@line 4595 "load_abc.cpp"
        $2082 = HEAP32[$2081>>2]|0; //@line 4595 "load_abc.cpp"
        HEAP32[$vararg_buffer107>>2] = $2082; //@line 4595 "load_abc.cpp"
        $vararg_ptr110 = ((($vararg_buffer107)) + 4|0); //@line 4595 "load_abc.cpp"
        HEAP32[$vararg_ptr110>>2] = $2082; //@line 4595 "load_abc.cpp"
        (_sprintf($barsig,526499,$vararg_buffer107)|0); //@line 4595 "load_abc.cpp"
        $2083 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4596 "load_abc.cpp"
        $2084 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4597 "load_abc.cpp"
        $2085 = ($2084|0)==(0|0); //@line 4597 "load_abc.cpp"
        do {
         if ($2085) {
          $2094 = 0;
         } else {
          $2086 = ((($2084)) + 37|0); //@line 4597 "load_abc.cpp"
          $2087 = HEAP8[$2086>>0]|0; //@line 4597 "load_abc.cpp"
          $2088 = ($2087<<24>>24)==(0); //@line 4597 "load_abc.cpp"
          if ($2088) {
           $2094 = $2084;
           break;
          }
          $2089 = ((($2084)) + 44|0); //@line 4597 "load_abc.cpp"
          $2090 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,$2089,0)|0); //@line 4597 "load_abc.cpp"
          HEAP8[$41>>0]=$2090&255;HEAP8[$41+1>>0]=($2090>>8)&255;HEAP8[$41+2>>0]=($2090>>16)&255;HEAP8[$41+3>>0]=$2090>>24; //@line 4597 "load_abc.cpp"
          $2094 = $2090;
         }
        } while(0);
        $2091 = HEAP8[$p$18>>0]|0; //@line 4598 "load_abc.cpp"
        $2092 = $2091 << 24 >> 24; //@line 4598 "load_abc.cpp"
        $isdigittmp41 = (($2092) + -48)|0; //@line 4598 "load_abc.cpp"
        $isdigit42 = ($isdigittmp41>>>0)<(10); //@line 4598 "load_abc.cpp"
        if (!($isdigit42)) {
         $2171 = ($2091<<24>>24)==(58); //@line 4620 "load_abc.cpp"
         if (!($2171)) {
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $2083;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$18;
          break L678;
         }
         $2172 = ((($p$18)) + 1|0); //@line 4621 "load_abc.cpp"
         $2173 = ($2094|0)==(0|0); //@line 915 "load_abc.cpp"
         if ($2173) {
          $2174 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
          $2175 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
          $2176 = ((($2174)) + 20|0); //@line 917 "load_abc.cpp"
          HEAP8[$2176>>0]=$2175&255;HEAP8[$2176+1>>0]=($2175>>8)&255;HEAP8[$2176+2>>0]=($2175>>16)&255;HEAP8[$2176+3>>0]=$2175>>24; //@line 917 "load_abc.cpp"
          $$0$i442 = $2174;
         } else {
          $$0$i442 = $2094;
         }
         HEAP8[$41>>0]=$$0$i442&255;HEAP8[$41+1>>0]=($$0$i442>>8)&255;HEAP8[$41+2>>0]=($$0$i442>>16)&255;HEAP8[$41+3>>0]=$$0$i442>>24; //@line 4622 "load_abc.cpp"
         $2177 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4623 "load_abc.cpp"
         $2178 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
         HEAP8[$2178>>0]=0&255;HEAP8[$2178+1>>0]=(0>>8)&255;HEAP8[$2178+2>>0]=(0>>16)&255;HEAP8[$2178+3>>0]=0>>24; //@line 540 "load_abc.cpp"
         $2179 = ((($2178)) + 4|0); //@line 541 "load_abc.cpp"
         HEAP8[$2179>>0]=$2177&255;HEAP8[$2179+1>>0]=($2177>>8)&255;HEAP8[$2179+2>>0]=($2177>>16)&255;HEAP8[$2179+3>>0]=$2177>>24; //@line 541 "load_abc.cpp"
         $2180 = ((($2178)) + 8|0); //@line 543 "load_abc.cpp"
         HEAP8[$2180>>0] = 1; //@line 543 "load_abc.cpp"
         $2181 = ((($2178)) + 9|0); //@line 543 "load_abc.cpp"
         HEAP8[$2181>>0] = 80; //@line 543 "load_abc.cpp"
         $2182 = ((($2178)) + 10|0); //@line 543 "load_abc.cpp"
         HEAP8[$2182>>0]=0&255;HEAP8[$2182+1>>0]=(0>>8)&255;HEAP8[$2182+2>>0]=(0>>16)&255;HEAP8[$2182+3>>0]=0>>24; //@line 543 "load_abc.cpp"
         $2183 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
         $2184 = ((($2178)) + 14|0); //@line 544 "load_abc.cpp"
         HEAP8[$2184>>0] = $2183; //@line 544 "load_abc.cpp"
         $2185 = ((($2178)) + 15|0); //@line 545 "load_abc.cpp"
         HEAP8[$2185>>0] = 0; //@line 545 "load_abc.cpp"
         __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($$0$i442,$2178); //@line 1153 "load_abc.cpp"
         $2186 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4624 "load_abc.cpp"
         $2187 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4624 "load_abc.cpp"
         $2188 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
         HEAP8[$2188>>0]=0&255;HEAP8[$2188+1>>0]=(0>>8)&255;HEAP8[$2188+2>>0]=(0>>16)&255;HEAP8[$2188+3>>0]=0>>24; //@line 540 "load_abc.cpp"
         $2189 = ((($2188)) + 4|0); //@line 541 "load_abc.cpp"
         HEAP8[$2189>>0]=$2187&255;HEAP8[$2189+1>>0]=($2187>>8)&255;HEAP8[$2189+2>>0]=($2187>>16)&255;HEAP8[$2189+3>>0]=$2187>>24; //@line 541 "load_abc.cpp"
         $2190 = ((($2188)) + 8|0); //@line 543 "load_abc.cpp"
         HEAP8[$2190>>0] = 1; //@line 543 "load_abc.cpp"
         $2191 = ((($2188)) + 9|0); //@line 543 "load_abc.cpp"
         HEAP8[$2191>>0] = 76; //@line 543 "load_abc.cpp"
         $2192 = ((($2188)) + 10|0); //@line 543 "load_abc.cpp"
         HEAP8[$2192>>0]=0&255;HEAP8[$2192+1>>0]=(0>>8)&255;HEAP8[$2192+2>>0]=(0>>16)&255;HEAP8[$2192+3>>0]=0>>24; //@line 543 "load_abc.cpp"
         $2193 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
         $2194 = ((($2188)) + 14|0); //@line 544 "load_abc.cpp"
         HEAP8[$2194>>0] = $2193; //@line 544 "load_abc.cpp"
         $2195 = ((($2188)) + 15|0); //@line 545 "load_abc.cpp"
         HEAP8[$2195>>0] = 0; //@line 545 "load_abc.cpp"
         __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($2186,$2188); //@line 1562 "load_abc.cpp"
         $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $2083;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2172;
         break L678;
        }
        $2093 = ($2094|0)==(0|0); //@line 915 "load_abc.cpp"
        if ($2093) {
         $2095 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
         $2096 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
         $2097 = ((($2095)) + 20|0); //@line 917 "load_abc.cpp"
         HEAP8[$2097>>0]=$2096&255;HEAP8[$2097+1>>0]=($2096>>8)&255;HEAP8[$2097+2>>0]=($2096>>16)&255;HEAP8[$2097+3>>0]=$2096>>24; //@line 917 "load_abc.cpp"
         $$0$i404 = $2095;
        } else {
         $$0$i404 = $2094;
        }
        HEAP8[$41>>0]=$$0$i404&255;HEAP8[$41+1>>0]=($$0$i404>>8)&255;HEAP8[$41+2>>0]=($$0$i404>>16)&255;HEAP8[$41+3>>0]=$$0$i404>>24; //@line 4599 "load_abc.cpp"
        $2098 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4600 "load_abc.cpp"
        $2099 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
        HEAP8[$2099>>0]=0&255;HEAP8[$2099+1>>0]=(0>>8)&255;HEAP8[$2099+2>>0]=(0>>16)&255;HEAP8[$2099+3>>0]=0>>24; //@line 540 "load_abc.cpp"
        $2100 = ((($2099)) + 4|0); //@line 541 "load_abc.cpp"
        HEAP8[$2100>>0]=$2098&255;HEAP8[$2100+1>>0]=($2098>>8)&255;HEAP8[$2100+2>>0]=($2098>>16)&255;HEAP8[$2100+3>>0]=$2098>>24; //@line 541 "load_abc.cpp"
        $2101 = ((($2099)) + 8|0); //@line 543 "load_abc.cpp"
        HEAP8[$2101>>0] = 1; //@line 543 "load_abc.cpp"
        $2102 = ((($2099)) + 9|0); //@line 543 "load_abc.cpp"
        HEAP8[$2102>>0] = 80; //@line 543 "load_abc.cpp"
        $2103 = ((($2099)) + 10|0); //@line 543 "load_abc.cpp"
        HEAP8[$2103>>0]=0&255;HEAP8[$2103+1>>0]=(0>>8)&255;HEAP8[$2103+2>>0]=(0>>16)&255;HEAP8[$2103+3>>0]=0>>24; //@line 543 "load_abc.cpp"
        $2104 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
        $2105 = ((($2099)) + 14|0); //@line 544 "load_abc.cpp"
        HEAP8[$2105>>0] = $2104; //@line 544 "load_abc.cpp"
        $2106 = ((($2099)) + 15|0); //@line 545 "load_abc.cpp"
        HEAP8[$2106>>0] = 0; //@line 545 "load_abc.cpp"
        __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($$0$i404,$2099); //@line 1153 "load_abc.cpp"
        $2107 = HEAP8[$p$18>>0]|0; //@line 1625 "load_abc.cpp"
        $2108 = $2107 << 24 >> 24; //@line 1625 "load_abc.cpp"
        $isdigittmp1$i406 = (($2108) + -48)|0; //@line 1625 "load_abc.cpp"
        $isdigit2$i407 = ($isdigittmp1$i406>>>0)<(10); //@line 1625 "load_abc.cpp"
        if ($isdigit2$i407) {
         $2111 = $2108;$h$04$i408 = 0;$i$03$i409 = 0;
         while(1) {
          $2109 = ($h$04$i408*10)|0; //@line 1626 "load_abc.cpp"
          $2110 = (($2111) + -48)|0; //@line 1626 "load_abc.cpp"
          $2112 = (($2110) + ($2109))|0; //@line 1626 "load_abc.cpp"
          $2113 = (($i$03$i409) + 1)|0; //@line 1627 "load_abc.cpp"
          $2114 = (($p$18) + ($2113)|0); //@line 1625 "load_abc.cpp"
          $2115 = HEAP8[$2114>>0]|0; //@line 1625 "load_abc.cpp"
          $2116 = $2115 << 24 >> 24; //@line 1625 "load_abc.cpp"
          $isdigittmp$i410 = (($2116) + -48)|0; //@line 1625 "load_abc.cpp"
          $isdigit$i411 = ($isdigittmp$i410>>>0)<(10); //@line 1625 "load_abc.cpp"
          if ($isdigit$i411) {
           $2111 = $2116;$h$04$i408 = $2112;$i$03$i409 = $2113;
          } else {
           $h$0$lcssa$i413 = $2112;$i$0$lcssa$i414 = $2113;
           break;
          }
         }
        } else {
         $h$0$lcssa$i413 = 0;$i$0$lcssa$i414 = 0;
        }
        $2117 = ($i$0$lcssa$i414|0)==(0); //@line 1629 "load_abc.cpp"
        $$h$0$i415 = $2117 ? 1 : $h$0$lcssa$i413; //@line 1630 "load_abc.cpp"
        $2118 = (($p$18) + ($i$0$lcssa$i414)|0); //@line 4601 "load_abc.cpp"
        $2119 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4602 "load_abc.cpp"
        $2120 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4602 "load_abc.cpp"
        $2121 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
        HEAP8[$2121>>0]=0&255;HEAP8[$2121+1>>0]=(0>>8)&255;HEAP8[$2121+2>>0]=(0>>16)&255;HEAP8[$2121+3>>0]=0>>24; //@line 540 "load_abc.cpp"
        $2122 = ((($2121)) + 4|0); //@line 541 "load_abc.cpp"
        HEAP8[$2122>>0]=$2120&255;HEAP8[$2122+1>>0]=($2120>>8)&255;HEAP8[$2122+2>>0]=($2120>>16)&255;HEAP8[$2122+3>>0]=$2120>>24; //@line 541 "load_abc.cpp"
        $2123 = ((($2121)) + 8|0); //@line 543 "load_abc.cpp"
        HEAP8[$2123>>0] = 1; //@line 543 "load_abc.cpp"
        $2124 = ((($2121)) + 9|0); //@line 543 "load_abc.cpp"
        HEAP8[$2124>>0] = 86; //@line 543 "load_abc.cpp"
        $2125 = ((($2121)) + 10|0); //@line 543 "load_abc.cpp"
        $2126 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
        $2127 = ((($2121)) + 14|0); //@line 544 "load_abc.cpp"
        HEAP8[$2127>>0] = $2126; //@line 544 "load_abc.cpp"
        $2128 = ((($2121)) + 15|0); //@line 545 "load_abc.cpp"
        HEAP8[$2128>>0] = 0; //@line 545 "load_abc.cpp"
        $2129 = 1 << $$h$0$i415; //@line 1482 "load_abc.cpp"
        HEAP8[$2125>>0]=$2129&255;HEAP8[$2125+1>>0]=($2129>>8)&255;HEAP8[$2125+2>>0]=($2129>>16)&255;HEAP8[$2125+3>>0]=$2129>>24; //@line 1482 "load_abc.cpp"
        __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($2119,$2121); //@line 1483 "load_abc.cpp"
        $2130 = HEAP8[$2118>>0]|0; //@line 4603 "load_abc.cpp"
        $2131 = $2130 & -2; //@line 4603 "load_abc.cpp"
        $switch1001199 = ($2131<<24>>24)==(44); //@line 4603 "load_abc.cpp"
        if ($switch1001199) {
         $2134 = $2130;$2164 = $$h$0$i415;$p$411200 = $2118;
        } else {
         $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $2083;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2118;
         break L678;
        }
        while(1) {
         $2133 = ($2134<<24>>24)==(44); //@line 4604 "load_abc.cpp"
         $2135 = ((($p$411200)) + 1|0); //@line 4605 "load_abc.cpp"
         $2136 = HEAP8[$2135>>0]|0; //@line 1625 "load_abc.cpp"
         $2137 = $2136 << 24 >> 24; //@line 1625 "load_abc.cpp"
         $isdigittmp1$i418 = (($2137) + -48)|0; //@line 1625 "load_abc.cpp"
         $isdigit2$i419 = ($isdigittmp1$i418>>>0)<(10); //@line 1625 "load_abc.cpp"
         do {
          if ($2133) {
           if ($isdigit2$i419) {
            $2140 = $2137;$h$04$i420 = 0;$i$03$i421 = 0;
            while(1) {
             $2138 = ($h$04$i420*10)|0; //@line 1626 "load_abc.cpp"
             $2139 = (($2140) + -48)|0; //@line 1626 "load_abc.cpp"
             $2141 = (($2139) + ($2138))|0; //@line 1626 "load_abc.cpp"
             $2142 = (($i$03$i421) + 1)|0; //@line 1627 "load_abc.cpp"
             $$sum804 = (($i$03$i421) + 2)|0; //@line 1625 "load_abc.cpp"
             $2143 = (($p$411200) + ($$sum804)|0); //@line 1625 "load_abc.cpp"
             $2144 = HEAP8[$2143>>0]|0; //@line 1625 "load_abc.cpp"
             $2145 = $2144 << 24 >> 24; //@line 1625 "load_abc.cpp"
             $isdigittmp$i422 = (($2145) + -48)|0; //@line 1625 "load_abc.cpp"
             $isdigit$i423 = ($isdigittmp$i422>>>0)<(10); //@line 1625 "load_abc.cpp"
             if ($isdigit$i423) {
              $2140 = $2145;$h$04$i420 = $2141;$i$03$i421 = $2142;
             } else {
              $h$0$lcssa$i425 = $2141;$i$0$lcssa$i426 = $2142;
              break;
             }
            }
           } else {
            $h$0$lcssa$i425 = 0;$i$0$lcssa$i426 = 0;
           }
           $2146 = ($i$0$lcssa$i426|0)==(0); //@line 1629 "load_abc.cpp"
           $$h$0$i427 = $2146 ? 1 : $h$0$lcssa$i425; //@line 1630 "load_abc.cpp"
           $2147 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4607 "load_abc.cpp"
           $$idx104 = ((($2147)) + 8|0);
           $$idx104$val = HEAPU8[$$idx104>>0]|(HEAPU8[$$idx104+1>>0]<<8)|(HEAPU8[$$idx104+2>>0]<<16)|(HEAPU8[$$idx104+3>>0]<<24);
           $2148 = 1 << $$h$0$i427; //@line 1488 "load_abc.cpp"
           $2149 = ((($$idx104$val)) + 10|0); //@line 1488 "load_abc.cpp"
           $2150 = HEAPU8[$2149>>0]|(HEAPU8[$2149+1>>0]<<8)|(HEAPU8[$2149+2>>0]<<16)|(HEAPU8[$2149+3>>0]<<24); //@line 1488 "load_abc.cpp"
           $2151 = $2150 | $2148; //@line 1488 "load_abc.cpp"
           HEAP8[$2149>>0]=$2151&255;HEAP8[$2149+1>>0]=($2151>>8)&255;HEAP8[$2149+2>>0]=($2151>>16)&255;HEAP8[$2149+3>>0]=$2151>>24; //@line 1488 "load_abc.cpp"
           $$be865 = $$h$0$i427;$$sum43$pn$in = $i$0$lcssa$i426;
          } else {
           if ($isdigit2$i419) {
            $2156 = $2137;$h$04$i432 = 0;$i$03$i433 = 0;
            while(1) {
             $2154 = ($h$04$i432*10)|0; //@line 1626 "load_abc.cpp"
             $2155 = (($2156) + -48)|0; //@line 1626 "load_abc.cpp"
             $2157 = (($2155) + ($2154))|0; //@line 1626 "load_abc.cpp"
             $2158 = (($i$03$i433) + 1)|0; //@line 1627 "load_abc.cpp"
             $$sum803 = (($i$03$i433) + 2)|0; //@line 1625 "load_abc.cpp"
             $2159 = (($p$411200) + ($$sum803)|0); //@line 1625 "load_abc.cpp"
             $2160 = HEAP8[$2159>>0]|0; //@line 1625 "load_abc.cpp"
             $2161 = $2160 << 24 >> 24; //@line 1625 "load_abc.cpp"
             $isdigittmp$i434 = (($2161) + -48)|0; //@line 1625 "load_abc.cpp"
             $isdigit$i435 = ($isdigittmp$i434>>>0)<(10); //@line 1625 "load_abc.cpp"
             if ($isdigit$i435) {
              $2156 = $2161;$h$04$i432 = $2157;$i$03$i433 = $2158;
             } else {
              $h$0$lcssa$i437 = $2157;$i$0$lcssa$i438 = $2158;
              break;
             }
            }
           } else {
            $h$0$lcssa$i437 = 0;$i$0$lcssa$i438 = 0;
           }
           $2162 = ($i$0$lcssa$i438|0)==(0); //@line 1629 "load_abc.cpp"
           $$h$0$i439 = $2162 ? 1 : $h$0$lcssa$i437; //@line 1630 "load_abc.cpp"
           $2163 = ($2164|0)<($$h$0$i439|0); //@line 4612 "load_abc.cpp"
           if (!($2163)) {
            $$be865 = $2164;$$sum43$pn$in = $i$0$lcssa$i438;
            break;
           }
           $2165 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4614 "load_abc.cpp"
           $$idx103 = ((($2165)) + 8|0);
           $$idx103$val = HEAPU8[$$idx103>>0]|(HEAPU8[$$idx103+1>>0]<<8)|(HEAPU8[$$idx103+2>>0]<<16)|(HEAPU8[$$idx103+3>>0]<<24);
           $2132 = ((($$idx103$val)) + 10|0); //@line 1488 "load_abc.cpp"
           $$promoted = HEAPU8[$2132>>0]|(HEAPU8[$2132+1>>0]<<8)|(HEAPU8[$2132+2>>0]<<16)|(HEAPU8[$2132+3>>0]<<24); //@line 1488 "load_abc.cpp"
           $2167 = $2164;$2170 = $$promoted;
           while(1) {
            $2166 = (($2167) + 1)|0; //@line 4613 "load_abc.cpp"
            $2168 = 1 << $2166; //@line 1488 "load_abc.cpp"
            $2169 = $2170 | $2168; //@line 1488 "load_abc.cpp"
            $exitcond1500 = ($2166|0)==($$h$0$i439|0); //@line 4612 "load_abc.cpp"
            if ($exitcond1500) {
             $$lcssa2202 = $2169;
             break;
            } else {
             $2167 = $2166;$2170 = $2169;
            }
           }
           HEAP8[$2132>>0]=$$lcssa2202&255;HEAP8[$2132+1>>0]=($$lcssa2202>>8)&255;HEAP8[$2132+2>>0]=($$lcssa2202>>16)&255;HEAP8[$2132+3>>0]=$$lcssa2202>>24; //@line 1488 "load_abc.cpp"
           $$be865 = $$h$0$i439;$$sum43$pn$in = $i$0$lcssa$i438;
          }
         } while(0);
         $$sum43$pn = (($$sum43$pn$in) + 1)|0; //@line 4606 "load_abc.cpp"
         $p$41$be = (($p$411200) + ($$sum43$pn)|0); //@line 4606 "load_abc.cpp"
         $2152 = HEAP8[$p$41$be>>0]|0; //@line 4603 "load_abc.cpp"
         $2153 = $2152 & -2; //@line 4603 "load_abc.cpp"
         $switch100 = ($2153<<24>>24)==(44); //@line 4603 "load_abc.cpp"
         if ($switch100) {
          $2134 = $2152;$2164 = $$be865;$p$411200 = $p$41$be;
         } else {
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $2083;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$41$be;
          break;
         }
        }
        break;
       }
       case 38:  {
        HEAP8[$31>>0]=$bartime$41235&255;HEAP8[$31+1>>0]=($bartime$41235>>8)&255;HEAP8[$31+2>>0]=($bartime$41235>>16)&255;HEAP8[$31+3>>0]=$bartime$41235>>24; //@line 4629 "load_abc.cpp"
        $2196 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4630 "load_abc.cpp"
        $2197 = ($2196|0)==(0|0); //@line 915 "load_abc.cpp"
        if ($2197) {
         $2198 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
         $2199 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
         $2200 = ((($2198)) + 20|0); //@line 917 "load_abc.cpp"
         HEAP8[$2200>>0]=$2199&255;HEAP8[$2200+1>>0]=($2199>>8)&255;HEAP8[$2200+2>>0]=($2199>>16)&255;HEAP8[$2200+3>>0]=$2199>>24; //@line 917 "load_abc.cpp"
         $$0$i444 = $2198;
        } else {
         $$0$i444 = $2196;
        }
        HEAP8[$41>>0]=$$0$i444&255;HEAP8[$41+1>>0]=($$0$i444>>8)&255;HEAP8[$41+2>>0]=($$0$i444>>16)&255;HEAP8[$41+3>>0]=$$0$i444>>24; //@line 4630 "load_abc.cpp"
        $2201 = ((($$0$i444)) + 37|0); //@line 4631 "load_abc.cpp"
        $2202 = HEAP8[$2201>>0]|0; //@line 4631 "load_abc.cpp"
        $2203 = $2202&255; //@line 4631 "load_abc.cpp"
        $2204 = ((($$0$i444)) + 44|0); //@line 4632 "load_abc.cpp"
        $2205 = ($2202<<24>>24)!=(0); //@line 4632 "load_abc.cpp"
        $2206 = (($2203) + 1)|0; //@line 4632 "load_abc.cpp"
        $2207 = $2205 ? $2206 : 11; //@line 4632 "load_abc.cpp"
        $2208 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,$2204,$2207)|0); //@line 4632 "load_abc.cpp"
        HEAP8[$41>>0]=$2208&255;HEAP8[$41+1>>0]=($2208>>8)&255;HEAP8[$41+2>>0]=($2208>>16)&255;HEAP8[$41+3>>0]=$2208>>24; //@line 4632 "load_abc.cpp"
        $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$18;
        break;
       }
       case 93:  {
        $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$18;
        break;
       }
       case 58:  {
        $2209 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4639 "load_abc.cpp"
        $2210 = ($2209|0)==(0|0); //@line 915 "load_abc.cpp"
        if ($2210) {
         $2211 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
         $2212 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
         $2213 = ((($2211)) + 20|0); //@line 917 "load_abc.cpp"
         HEAP8[$2213>>0]=$2212&255;HEAP8[$2213+1>>0]=($2212>>8)&255;HEAP8[$2213+2>>0]=($2212>>16)&255;HEAP8[$2213+3>>0]=$2212>>24; //@line 917 "load_abc.cpp"
         $$0$i446 = $2211;
        } else {
         $$0$i446 = $2209;
        }
        HEAP8[$41>>0]=$$0$i446&255;HEAP8[$41+1>>0]=($$0$i446>>8)&255;HEAP8[$41+2>>0]=($$0$i446>>16)&255;HEAP8[$41+3>>0]=$$0$i446>>24; //@line 4639 "load_abc.cpp"
        $2214 = ((($$0$i446)) + 32|0); //@line 4640 "load_abc.cpp"
        $2215 = HEAPU8[$2214>>0]|(HEAPU8[$2214+1>>0]<<8)|(HEAPU8[$2214+2>>0]<<16)|(HEAPU8[$2214+3>>0]<<24); //@line 4640 "load_abc.cpp"
        $2216 = ($2215|0)==(-2); //@line 4640 "load_abc.cpp"
        $2217 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4641 "load_abc.cpp"
        $2218 = $2216&1; //@line 1581 "load_abc.cpp"
        $2219 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
        HEAP8[$2219>>0]=0&255;HEAP8[$2219+1>>0]=(0>>8)&255;HEAP8[$2219+2>>0]=(0>>16)&255;HEAP8[$2219+3>>0]=0>>24; //@line 540 "load_abc.cpp"
        $2220 = ((($2219)) + 4|0); //@line 541 "load_abc.cpp"
        HEAP8[$2220>>0]=$2217&255;HEAP8[$2220+1>>0]=($2217>>8)&255;HEAP8[$2220+2>>0]=($2217>>16)&255;HEAP8[$2220+3>>0]=$2217>>24; //@line 541 "load_abc.cpp"
        $2221 = ((($2219)) + 8|0); //@line 543 "load_abc.cpp"
        HEAP8[$2221>>0] = 1; //@line 543 "load_abc.cpp"
        $2222 = ((($2219)) + 9|0); //@line 543 "load_abc.cpp"
        HEAP8[$2222>>0] = 74; //@line 543 "load_abc.cpp"
        $2223 = ((($2219)) + 10|0); //@line 543 "load_abc.cpp"
        HEAP8[$2223>>0] = 0; //@line 543 "load_abc.cpp"
        $2224 = ((($2219)) + 11|0); //@line 543 "load_abc.cpp"
        HEAP8[$2224>>0] = 0; //@line 543 "load_abc.cpp"
        $2225 = ((($2219)) + 12|0); //@line 543 "load_abc.cpp"
        HEAP8[$2225>>0] = 0; //@line 543 "load_abc.cpp"
        $2226 = ((($2219)) + 13|0); //@line 543 "load_abc.cpp"
        HEAP8[$2226>>0] = $2218; //@line 543 "load_abc.cpp"
        $2227 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
        $2228 = ((($2219)) + 14|0); //@line 544 "load_abc.cpp"
        HEAP8[$2228>>0] = $2227; //@line 544 "load_abc.cpp"
        $2229 = ((($2219)) + 15|0); //@line 545 "load_abc.cpp"
        HEAP8[$2229>>0] = 0; //@line 545 "load_abc.cpp"
        __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($$0$i446,$2219); //@line 1583 "load_abc.cpp"
        $2230 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4642 "load_abc.cpp"
        $2231 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4642 "load_abc.cpp"
        $2232 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
        HEAP8[$2232>>0]=0&255;HEAP8[$2232+1>>0]=(0>>8)&255;HEAP8[$2232+2>>0]=(0>>16)&255;HEAP8[$2232+3>>0]=0>>24; //@line 540 "load_abc.cpp"
        $2233 = ((($2232)) + 4|0); //@line 541 "load_abc.cpp"
        HEAP8[$2233>>0]=$2231&255;HEAP8[$2233+1>>0]=($2231>>8)&255;HEAP8[$2233+2>>0]=($2231>>16)&255;HEAP8[$2233+3>>0]=$2231>>24; //@line 541 "load_abc.cpp"
        $2234 = ((($2232)) + 8|0); //@line 543 "load_abc.cpp"
        HEAP8[$2234>>0] = 1; //@line 543 "load_abc.cpp"
        $2235 = ((($2232)) + 9|0); //@line 543 "load_abc.cpp"
        HEAP8[$2235>>0] = 80; //@line 543 "load_abc.cpp"
        $2236 = ((($2232)) + 10|0); //@line 543 "load_abc.cpp"
        HEAP8[$2236>>0]=0&255;HEAP8[$2236+1>>0]=(0>>8)&255;HEAP8[$2236+2>>0]=(0>>16)&255;HEAP8[$2236+3>>0]=0>>24; //@line 543 "load_abc.cpp"
        $2237 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
        $2238 = ((($2232)) + 14|0); //@line 544 "load_abc.cpp"
        HEAP8[$2238>>0] = $2237; //@line 544 "load_abc.cpp"
        $2239 = ((($2232)) + 15|0); //@line 545 "load_abc.cpp"
        HEAP8[$2239>>0] = 0; //@line 545 "load_abc.cpp"
        __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($2230,$2232); //@line 1153 "load_abc.cpp"
        $2240 = HEAP8[$p$18>>0]|0; //@line 4643 "load_abc.cpp"
        $2241 = ($2240<<24>>24)==(58); //@line 4643 "load_abc.cpp"
        if (!($2241)) {
         $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$18;
         break L678;
        }
        $2242 = ((($p$18)) + 1|0); //@line 4644 "load_abc.cpp"
        $2243 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4645 "load_abc.cpp"
        $2244 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4645 "load_abc.cpp"
        $2245 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
        HEAP8[$2245>>0]=0&255;HEAP8[$2245+1>>0]=(0>>8)&255;HEAP8[$2245+2>>0]=(0>>16)&255;HEAP8[$2245+3>>0]=0>>24; //@line 540 "load_abc.cpp"
        $2246 = ((($2245)) + 4|0); //@line 541 "load_abc.cpp"
        HEAP8[$2246>>0]=$2244&255;HEAP8[$2246+1>>0]=($2244>>8)&255;HEAP8[$2246+2>>0]=($2244>>16)&255;HEAP8[$2246+3>>0]=$2244>>24; //@line 541 "load_abc.cpp"
        $2247 = ((($2245)) + 8|0); //@line 543 "load_abc.cpp"
        HEAP8[$2247>>0] = 1; //@line 543 "load_abc.cpp"
        $2248 = ((($2245)) + 9|0); //@line 543 "load_abc.cpp"
        HEAP8[$2248>>0] = 76; //@line 543 "load_abc.cpp"
        $2249 = ((($2245)) + 10|0); //@line 543 "load_abc.cpp"
        HEAP8[$2249>>0]=0&255;HEAP8[$2249+1>>0]=(0>>8)&255;HEAP8[$2249+2>>0]=(0>>16)&255;HEAP8[$2249+3>>0]=0>>24; //@line 543 "load_abc.cpp"
        $2250 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
        $2251 = ((($2245)) + 14|0); //@line 544 "load_abc.cpp"
        HEAP8[$2251>>0] = $2250; //@line 544 "load_abc.cpp"
        $2252 = ((($2245)) + 15|0); //@line 545 "load_abc.cpp"
        HEAP8[$2252>>0] = 0; //@line 545 "load_abc.cpp"
        __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($2243,$2245); //@line 1562 "load_abc.cpp"
        $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2242;
        break;
       }
       case 34:  {
        $2253 = HEAP8[$p$18>>0]|0; //@line 4649 "load_abc.cpp"
        $2254 = $2253 << 24 >> 24; //@line 4649 "load_abc.cpp"
        $memchr39 = (_memchr(527331,$2254,6)|0); //@line 4649 "load_abc.cpp"
        $2255 = ($memchr39|0)!=(0|0); //@line 4649 "load_abc.cpp"
        $isdigittmp = (($2254) + -48)|0;
        $isdigit = ($isdigittmp>>>0)<(10); //@line 4649 "load_abc.cpp"
        $or$cond792 = $2255 | $isdigit; //@line 4649 "load_abc.cpp"
        do {
         if (!($or$cond792)) {
          $2256 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4650 "load_abc.cpp"
          $2257 = ($2256|0)==(0|0); //@line 915 "load_abc.cpp"
          if ($2257) {
           $2258 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
           $2259 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
           $2260 = ((($2258)) + 20|0); //@line 917 "load_abc.cpp"
           HEAP8[$2260>>0]=$2259&255;HEAP8[$2260+1>>0]=($2259>>8)&255;HEAP8[$2260+2>>0]=($2259>>16)&255;HEAP8[$2260+3>>0]=$2259>>24; //@line 917 "load_abc.cpp"
           $$0$i448 = $2258;
          } else {
           $$0$i448 = $2256;
          }
          HEAP8[$41>>0]=$$0$i448&255;HEAP8[$41+1>>0]=($$0$i448>>8)&255;HEAP8[$41+2>>0]=($$0$i448>>16)&255;HEAP8[$41+3>>0]=$$0$i448>>24; //@line 4650 "load_abc.cpp"
          $2261 = HEAPU8[$42>>0]|(HEAPU8[$42+1>>0]<<8)|(HEAPU8[$42+2>>0]<<16)|(HEAPU8[$42+3>>0]<<24); //@line 4651 "load_abc.cpp"
          $2262 = ($2261|0)==(0|0); //@line 4651 "load_abc.cpp"
          if ($2262) {
           $2263 = ((($$0$i448)) + 44|0); //@line 4651 "load_abc.cpp"
           $2264 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,$2263,0)|0); //@line 4651 "load_abc.cpp"
           HEAP8[$42>>0]=$2264&255;HEAP8[$42+1>>0]=($2264>>8)&255;HEAP8[$42+2>>0]=($2264>>16)&255;HEAP8[$42+3>>0]=$2264>>24; //@line 4651 "load_abc.cpp"
           $$pre1525 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4652 "load_abc.cpp"
           $2266 = $$pre1525;$2267 = $2264;
          } else {
           $2266 = $$0$i448;$2267 = $2261;
          }
          $2265 = ($2266|0)==($2267|0); //@line 4652 "load_abc.cpp"
          if (!($2265)) {
           break;
          }
          $2268 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4652 "load_abc.cpp"
          $2269 = HEAP8[$p$18>>0]|0; //@line 1501 "load_abc.cpp"
          $2270 = ($2269<<24>>24)==(40); //@line 1501 "load_abc.cpp"
          $2271 = ((($p$18)) + 1|0); //@line 1501 "load_abc.cpp"
          $$p$i = $2270 ? $2271 : $p$18; //@line 1501 "load_abc.cpp"
          $2272 = HEAP8[$$p$i>>0]|0; //@line 1503 "load_abc.cpp"
          $2273 = ($2272<<24>>24)==(32); //@line 1503 "load_abc.cpp"
          do {
           if ($2273) {
            $2279 = 0;
           } else {
            $2274 = ($2272<<24>>24)==(67); //@line 1503 "load_abc.cpp"
            if ($2274) {
             $2279 = 1;
             break;
            }
            $2347 = ($2272<<24>>24)==(68); //@line 1503 "load_abc.cpp"
            if ($2347) {
             $2279 = 3;
             break;
            }
            $2348 = ($2272<<24>>24)==(69); //@line 1503 "load_abc.cpp"
            if ($2348) {
             $2279 = 5;
             break;
            }
            $2349 = ($2272<<24>>24)==(70); //@line 1503 "load_abc.cpp"
            if ($2349) {
             $2279 = 6;
             break;
            }
            $2350 = ($2272<<24>>24)==(71); //@line 1503 "load_abc.cpp"
            if ($2350) {
             $2279 = 8;
             break;
            }
            $2351 = ($2272<<24>>24)==(65); //@line 1503 "load_abc.cpp"
            if ($2351) {
             $2279 = 10;
             break;
            }
            $2352 = ($2272<<24>>24)==(66); //@line 1503 "load_abc.cpp"
            if ($2352) {
             $2279 = 12;
             break;
            }
            $2353 = ($2272<<24>>24)==(99); //@line 1503 "load_abc.cpp"
            if ($2353) {
             $2279 = 13;
             break;
            }
            $2354 = ($2272<<24>>24)==(100); //@line 1503 "load_abc.cpp"
            if ($2354) {
             $2279 = 15;
             break;
            }
            $2355 = ($2272<<24>>24)==(101); //@line 1503 "load_abc.cpp"
            if ($2355) {
             $2279 = 17;
             break;
            }
            $2356 = ($2272<<24>>24)==(102); //@line 1503 "load_abc.cpp"
            if ($2356) {
             $2279 = 18;
             break;
            }
            $2357 = ($2272<<24>>24)==(103); //@line 1503 "load_abc.cpp"
            if ($2357) {
             $2279 = 20;
             break;
            }
            $2358 = ($2272<<24>>24)==(97); //@line 1503 "load_abc.cpp"
            if ($2358) {
             $2279 = 22;
             break;
            }
            $2359 = ($2272<<24>>24)==(98); //@line 1503 "load_abc.cpp"
            $$39$i = $2359 ? 24 : 0; //@line 1503 "load_abc.cpp"
            $2279 = $$39$i;
           }
          } while(0);
          $2275 = ((($$p$i)) + 1|0); //@line 1507 "load_abc.cpp"
          $2276 = HEAP8[$2275>>0]|0; //@line 1508 "load_abc.cpp"
          $2277 = $2276 << 24 >> 24; //@line 1508 "load_abc.cpp"
          switch ($2277|0) {
          case 98:  {
           $2278 = (($2279) + -1)<<24>>24; //@line 1510 "load_abc.cpp"
           $2280 = ((($$p$i)) + 2|0); //@line 1511 "load_abc.cpp"
           $$1$i450 = $2280;$2318 = $2278;
           break;
          }
          case 35:  {
           $2281 = (($2279) + 1)<<24>>24; //@line 1514 "load_abc.cpp"
           $2282 = ((($$p$i)) + 2|0); //@line 1515 "load_abc.cpp"
           $$1$i450 = $2282;$2318 = $2281;
           break;
          }
          default: {
           $$1$i450 = $2275;$2318 = $2279;
          }
          }
          $2283 = HEAP8[$$1$i450>>0]|0; //@line 1519 "load_abc.cpp"
          L1362: do {
           switch ($2283<<24>>24) {
           case 32: case 41: case 40: case 47: case 34: case 0:  {
            $2288 = $2283;$i$1$lcssa$i = 0;
            break;
           }
           default: {
            HEAP8[$s$i>>0] = $2283; //@line 1519 "load_abc.cpp"
            $2284 = ((($$1$i450)) + 1|0); //@line 1519 "load_abc.cpp"
            $2285 = HEAP8[$2284>>0]|0; //@line 1519 "load_abc.cpp"
            switch ($2285<<24>>24) {
            case 32: case 41: case 40: case 47: case 34: case 0:  {
             $2288 = $2285;$i$1$lcssa$i = 1;
             break L1362;
             break;
            }
            default: {
            }
            }
            HEAP8[$161>>0] = $2285; //@line 1519 "load_abc.cpp"
            $2337 = ((($$1$i450)) + 2|0); //@line 1519 "load_abc.cpp"
            $2338 = HEAP8[$2337>>0]|0; //@line 1519 "load_abc.cpp"
            switch ($2338<<24>>24) {
            case 32: case 41: case 40: case 47: case 34: case 0:  {
             $2288 = $2338;$i$1$lcssa$i = 2;
             break L1362;
             break;
            }
            default: {
            }
            }
            HEAP8[$162>>0] = $2338; //@line 1519 "load_abc.cpp"
            $2339 = ((($$1$i450)) + 3|0); //@line 1519 "load_abc.cpp"
            $2340 = HEAP8[$2339>>0]|0; //@line 1519 "load_abc.cpp"
            switch ($2340<<24>>24) {
            case 32: case 41: case 40: case 47: case 34: case 0:  {
             $2288 = $2340;$i$1$lcssa$i = 3;
             break L1362;
             break;
            }
            default: {
            }
            }
            HEAP8[$163>>0] = $2340; //@line 1519 "load_abc.cpp"
            $2341 = ((($$1$i450)) + 4|0); //@line 1519 "load_abc.cpp"
            $2342 = HEAP8[$2341>>0]|0; //@line 1519 "load_abc.cpp"
            switch ($2342<<24>>24) {
            case 32: case 41: case 40: case 47: case 34: case 0:  {
             $2288 = $2342;$i$1$lcssa$i = 4;
             break L1362;
             break;
            }
            default: {
            }
            }
            HEAP8[$164>>0] = $2342; //@line 1519 "load_abc.cpp"
            $2343 = ((($$1$i450)) + 5|0); //@line 1519 "load_abc.cpp"
            $2344 = HEAP8[$2343>>0]|0; //@line 1519 "load_abc.cpp"
            switch ($2344<<24>>24) {
            case 32: case 41: case 40: case 47: case 34: case 0:  {
             $2288 = $2344;$i$1$lcssa$i = 5;
             break L1362;
             break;
            }
            default: {
            }
            }
            HEAP8[$165>>0] = $2344; //@line 1519 "load_abc.cpp"
            $2345 = ((($$1$i450)) + 6|0); //@line 1519 "load_abc.cpp"
            $2346 = HEAP8[$2345>>0]|0; //@line 1519 "load_abc.cpp"
            switch ($2346<<24>>24) {
            case 32: case 41: case 40: case 47: case 34: case 0:  {
             $2288 = $2346;$i$1$lcssa$i = 6;
             break L1362;
             break;
            }
            default: {
            }
            }
            HEAP8[$166>>0] = $2346; //@line 1519 "load_abc.cpp"
            $$phi$trans$insert$i = ((($$1$i450)) + 7|0); //@line 4652 "load_abc.cpp"
            $$pre$i457 = HEAP8[$$phi$trans$insert$i>>0]|0; //@line 1522 "load_abc.cpp"
            $2288 = $$pre$i457;$i$1$lcssa$i = 7;
           }
           }
          } while(0);
          $2286 = (($s$i) + ($i$1$lcssa$i)|0); //@line 1520 "load_abc.cpp"
          HEAP8[$2286>>0] = 0; //@line 1520 "load_abc.cpp"
          $2287 = ($2288<<24>>24)==(47); //@line 1522 "load_abc.cpp"
          L1371: do {
           if ($2287) {
            $$sum$i453 = (($i$1$lcssa$i) + 1)|0; //@line 1523 "load_abc.cpp"
            $2289 = (($$1$i450) + ($$sum$i453)|0); //@line 1523 "load_abc.cpp"
            $2290 = HEAP8[$2289>>0]|0; //@line 1525 "load_abc.cpp"
            $2291 = ($2290<<24>>24)==(32); //@line 1525 "load_abc.cpp"
            do {
             if ($2291) {
              $2297 = 0;
             } else {
              $2292 = ($2290<<24>>24)==(67); //@line 1525 "load_abc.cpp"
              if ($2292) {
               $2297 = 1;
               break;
              }
              $2324 = ($2290<<24>>24)==(68); //@line 1525 "load_abc.cpp"
              if ($2324) {
               $2297 = 3;
               break;
              }
              $2325 = ($2290<<24>>24)==(69); //@line 1525 "load_abc.cpp"
              if ($2325) {
               $2297 = 5;
               break;
              }
              $2326 = ($2290<<24>>24)==(70); //@line 1525 "load_abc.cpp"
              if ($2326) {
               $2297 = 6;
               break;
              }
              $2327 = ($2290<<24>>24)==(71); //@line 1525 "load_abc.cpp"
              if ($2327) {
               $2297 = 8;
               break;
              }
              $2328 = ($2290<<24>>24)==(65); //@line 1525 "load_abc.cpp"
              if ($2328) {
               $2297 = 10;
               break;
              }
              $2329 = ($2290<<24>>24)==(66); //@line 1525 "load_abc.cpp"
              if ($2329) {
               $2297 = 12;
               break;
              }
              $2330 = ($2290<<24>>24)==(99); //@line 1525 "load_abc.cpp"
              if ($2330) {
               $2297 = 13;
               break;
              }
              $2331 = ($2290<<24>>24)==(100); //@line 1525 "load_abc.cpp"
              if ($2331) {
               $2297 = 15;
               break;
              }
              $2332 = ($2290<<24>>24)==(101); //@line 1525 "load_abc.cpp"
              if ($2332) {
               $2297 = 17;
               break;
              }
              $2333 = ($2290<<24>>24)==(102); //@line 1525 "load_abc.cpp"
              if ($2333) {
               $2297 = 18;
               break;
              }
              $2334 = ($2290<<24>>24)==(103); //@line 1525 "load_abc.cpp"
              if ($2334) {
               $2297 = 20;
               break;
              }
              $2335 = ($2290<<24>>24)==(97); //@line 1525 "load_abc.cpp"
              if ($2335) {
               $2297 = 22;
               break;
              }
              $2336 = ($2290<<24>>24)==(98); //@line 1525 "load_abc.cpp"
              $$$i = $2336 ? 24 : $2318; //@line 1525 "load_abc.cpp"
              $2297 = $$$i;
             }
            } while(0);
            $$sum1$i = (($i$1$lcssa$i) + 2)|0; //@line 1529 "load_abc.cpp"
            $2293 = (($$1$i450) + ($$sum1$i)|0); //@line 1529 "load_abc.cpp"
            $2294 = HEAP8[$2293>>0]|0; //@line 1530 "load_abc.cpp"
            $2295 = $2294 << 24 >> 24; //@line 1530 "load_abc.cpp"
            switch ($2295|0) {
            case 98:  {
             $2296 = (($2297) + -1)<<24>>24; //@line 1532 "load_abc.cpp"
             $$ph$i = $2296;
             break L1371;
             break;
            }
            case 35:  {
             $2298 = (($2297) + 1)<<24>>24; //@line 1536 "load_abc.cpp"
             $$ph$i = $2298;
             break L1371;
             break;
            }
            default: {
             $$ph$i = $2297;
             break L1371;
            }
            }
           } else {
            $$ph$i = $2318;
           }
          } while(0);
          $2299 = HEAP32[299864]|0; //@line 1541 "load_abc.cpp"
          $2300 = ($2299|0)>(0); //@line 1541 "load_abc.cpp"
          L1393: do {
           if ($2300) {
            $i$310$i = 0;
            while(1) {
             $2301 = (1277594 + ($i$310$i<<3)|0); //@line 1542 "load_abc.cpp"
             $2302 = (_strcmp($s$i,$2301)|0); //@line 1542 "load_abc.cpp"
             $2303 = ($2302|0)==(0); //@line 1542 "load_abc.cpp"
             if ($2303) {
              $i$310$i$lcssa = $i$310$i;
              break;
             }
             $2305 = (($i$310$i) + 1)|0; //@line 1541 "load_abc.cpp"
             $2306 = ($2305|0)<($2299|0); //@line 1541 "load_abc.cpp"
             if ($2306) {
              $i$310$i = $2305;
             } else {
              $2316 = 0;$i$37$i = $2305;
              break L1393;
             }
            }
            $2304 = $i$310$i$lcssa&255; //@line 1543 "load_abc.cpp"
            $2316 = $2304;$i$37$i = $i$310$i$lcssa;
           } else {
            $2316 = 0;$i$37$i = 0;
           }
          } while(0);
          $2307 = ($i$37$i|0)==($2299|0); //@line 1546 "load_abc.cpp"
          if ($2307) {
           $2308 = (_strlen($s$i)|0); //@line 323 "load_abc.cpp"
           $2309 = (($2308) + 34)|0; //@line 323 "load_abc.cpp"
           $2310 = ($2309>>>0)>(255); //@line 323 "load_abc.cpp"
           if (!($2310)) {
            HEAP32[$vararg_buffer111>>2] = $s$i; //@line 324 "load_abc.cpp"
            (_sprintf($txt$i,527337,$vararg_buffer111)|0); //@line 324 "load_abc.cpp"
            HEAP32[$vararg_buffer114>>2] = $txt$i; //@line 328 "load_abc.cpp"
            (_fprintf($155,524747,$vararg_buffer114)|0); //@line 328 "load_abc.cpp"
           }
          } else {
           $2311 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
           HEAP8[$2311>>0]=0&255;HEAP8[$2311+1>>0]=(0>>8)&255;HEAP8[$2311+2>>0]=(0>>16)&255;HEAP8[$2311+3>>0]=0>>24; //@line 540 "load_abc.cpp"
           $2312 = ((($2311)) + 4|0); //@line 541 "load_abc.cpp"
           HEAP8[$2312>>0]=$2268&255;HEAP8[$2312+1>>0]=($2268>>8)&255;HEAP8[$2312+2>>0]=($2268>>16)&255;HEAP8[$2312+3>>0]=$2268>>24; //@line 541 "load_abc.cpp"
           $2313 = ((($2311)) + 8|0); //@line 543 "load_abc.cpp"
           HEAP8[$2313>>0] = 1; //@line 543 "load_abc.cpp"
           $2314 = ((($2311)) + 9|0); //@line 543 "load_abc.cpp"
           HEAP8[$2314>>0] = 67; //@line 543 "load_abc.cpp"
           $2315 = ((($2311)) + 10|0); //@line 543 "load_abc.cpp"
           HEAP8[$2315>>0] = $2316; //@line 543 "load_abc.cpp"
           $2317 = ((($2311)) + 11|0); //@line 543 "load_abc.cpp"
           HEAP8[$2317>>0] = $2318; //@line 543 "load_abc.cpp"
           $2319 = ((($2311)) + 12|0); //@line 543 "load_abc.cpp"
           HEAP8[$2319>>0] = $$ph$i; //@line 543 "load_abc.cpp"
           $2320 = ((($2311)) + 13|0); //@line 543 "load_abc.cpp"
           HEAP8[$2320>>0] = 0; //@line 543 "load_abc.cpp"
           $2321 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
           $2322 = ((($2311)) + 14|0); //@line 544 "load_abc.cpp"
           HEAP8[$2322>>0] = $2321; //@line 544 "load_abc.cpp"
           $2323 = ((($2311)) + 15|0); //@line 545 "load_abc.cpp"
           HEAP8[$2323>>0] = 0; //@line 545 "load_abc.cpp"
           __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($2267,$2311); //@line 1551 "load_abc.cpp"
          }
         }
        } while(0);
        $2360 = ((($p$18)) + 1|0); //@line 4655 "load_abc.cpp"
        $2361 = HEAP8[$p$18>>0]|0; //@line 4655 "load_abc.cpp"
        $2362 = ($2361<<24>>24)!=(0); //@line 4655 "load_abc.cpp"
        $$not1187 = $2362 ^ 1; //@line 4655 "load_abc.cpp"
        $2363 = ($2361<<24>>24)==(34); //@line 4655 "load_abc.cpp"
        $or$cond901188 = $2363 | $$not1187; //@line 4655 "load_abc.cpp"
        if ($or$cond901188) {
         $$lcssa959 = $2360;$$lcssa960 = $2361;$$lcssa961 = $2362;$abcto$4$lcssa = 0;
        } else {
         $2364 = $2360;$abcto$41190 = 0;$p$421189 = $p$18;
         while(1) {
          $2365 = (_strncasecmp($2364,527372,4)|0); //@line 4656 "load_abc.cpp"
          $2366 = ($2365|0)==(0); //@line 4656 "load_abc.cpp"
          do {
           if ($2366) {
            $2367 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 4656 "load_abc.cpp"
            $2368 = ($2367|0)==(0|0); //@line 4656 "load_abc.cpp"
            if ($2368) {
             break;
            }
            $2369 = ((($2367)) + 32|0); //@line 4656 "load_abc.cpp"
            $2370 = HEAPU8[$2369>>0]|(HEAPU8[$2369+1>>0]<<8)|(HEAPU8[$2369+2>>0]<<16)|(HEAPU8[$2369+3>>0]<<24); //@line 4656 "load_abc.cpp"
            $2371 = ($2370|0)>(-2); //@line 4656 "load_abc.cpp"
            if (!($2371)) {
             break;
            }
            $2372 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4657 "load_abc.cpp"
            $2375 = $2370;$tp$08$i = $2367;
            while(1) {
             $2373 = ((($tp$08$i)) + 28|0); //@line 3636 "load_abc.cpp"
             HEAP8[$2373>>0]=$2372&255;HEAP8[$2373+1>>0]=($2372>>8)&255;HEAP8[$2373+2>>0]=($2372>>16)&255;HEAP8[$2373+3>>0]=$2372>>24; //@line 3636 "load_abc.cpp"
             $2374 = ($2375|0)>(-2); //@line 3640 "load_abc.cpp"
             if ($2374) {
              $2376 = ((($tp$08$i)) + 32|0); //@line 3640 "load_abc.cpp"
              HEAP8[$2376>>0]=-2&255;HEAP8[$2376+1>>0]=(-2>>8)&255;HEAP8[$2376+2>>0]=(-2>>16)&255;HEAP8[$2376+3>>0]=-2>>24; //@line 3641 "load_abc.cpp"
             }
             $tp$0$i459 = HEAPU8[$tp$08$i>>0]|(HEAPU8[$tp$08$i+1>>0]<<8)|(HEAPU8[$tp$08$i+2>>0]<<16)|(HEAPU8[$tp$08$i+3>>0]<<24); //@line 3634 "load_abc.cpp"
             $2377 = ($tp$0$i459|0)==(0|0); //@line 3634 "load_abc.cpp"
             if ($2377) {
              break;
             }
             $$phi$trans$insert = ((($tp$0$i459)) + 32|0);
             $$pre1526 = HEAPU8[$$phi$trans$insert>>0]|(HEAPU8[$$phi$trans$insert+1>>0]<<8)|(HEAPU8[$$phi$trans$insert+2>>0]<<16)|(HEAPU8[$$phi$trans$insert+3>>0]<<24); //@line 3640 "load_abc.cpp"
             $2375 = $$pre1526;$tp$08$i = $tp$0$i459;
            }
            $2378 = ((($2367)) + 8|0); //@line 3643 "load_abc.cpp"
            $2379 = HEAPU8[$2378>>0]|(HEAPU8[$2378+1>>0]<<8)|(HEAPU8[$2378+2>>0]<<16)|(HEAPU8[$2378+3>>0]<<24); //@line 3643 "load_abc.cpp"
            $2380 = ($2379|0)!=(0|0); //@line 3643 "load_abc.cpp"
            $2381 = ($2370|0)!=(-2); //@line 3644 "load_abc.cpp"
            $or$cond793 = $2381 & $2380; //@line 3644 "load_abc.cpp"
            if (!($or$cond793)) {
             break;
            }
            $2382 = ((($2379)) + 4|0); //@line 3645 "load_abc.cpp"
            $2383 = HEAPU8[$2382>>0]|(HEAPU8[$2382+1>>0]<<8)|(HEAPU8[$2382+2>>0]<<16)|(HEAPU8[$2382+3>>0]<<24); //@line 3645 "load_abc.cpp"
            $2384 = ($2383>>>0)<($2372>>>0); //@line 3645 "load_abc.cpp"
            if ($2384) {
             break;
            }
            $2385 = ((($2367)) + 4|0); //@line 3647 "load_abc.cpp"
            $e$05$i = HEAPU8[$2385>>0]|(HEAPU8[$2385+1>>0]<<8)|(HEAPU8[$2385+2>>0]<<16)|(HEAPU8[$2385+3>>0]<<24); //@line 3647 "load_abc.cpp"
            $2386 = ($e$05$i|0)==(0|0); //@line 3647 "load_abc.cpp"
            if ($2386) {
             break;
            } else {
             $e$06$i = $e$05$i;
            }
            while(1) {
             $2387 = ((($e$06$i)) + 8|0); //@line 3648 "load_abc.cpp"
             $2388 = HEAP8[$2387>>0]|0; //@line 3648 "load_abc.cpp"
             $2389 = ($2388<<24>>24)==(1); //@line 3648 "load_abc.cpp"
             L1425: do {
              if ($2389) {
               $2390 = ((($e$06$i)) + 9|0); //@line 3648 "load_abc.cpp"
               $2391 = HEAP8[$2390>>0]|0; //@line 3648 "load_abc.cpp"
               $2392 = ($2391<<24>>24)==(74); //@line 3648 "load_abc.cpp"
               if (!($2392)) {
                break;
               }
               $2393 = ((($e$06$i)) + 4|0); //@line 3648 "load_abc.cpp"
               $2394 = HEAPU8[$2393>>0]|(HEAPU8[$2393+1>>0]<<8)|(HEAPU8[$2393+2>>0]<<16)|(HEAPU8[$2393+3>>0]<<24); //@line 3648 "load_abc.cpp"
               $2395 = ($2394>>>0)<($2372>>>0); //@line 3648 "load_abc.cpp"
               if ($2395) {
                break;
               }
               $2396 = ((($e$06$i)) + 13|0); //@line 3649 "load_abc.cpp"
               $2397 = HEAP8[$2396>>0]|0; //@line 3649 "load_abc.cpp"
               $2398 = $2397&255; //@line 3649 "load_abc.cpp"
               switch ($2398|0) {
               case 1: case 0:  {
                HEAP8[$2396>>0] = 1; //@line 3652 "load_abc.cpp"
                break L1425;
                break;
               }
               case 3: case 2:  {
                HEAP8[$2396>>0] = 3; //@line 3656 "load_abc.cpp"
                break L1425;
                break;
               }
               case 5: case 4:  {
                HEAP8[$2396>>0] = 5; //@line 3660 "load_abc.cpp"
                break L1425;
                break;
               }
               default: {
                break L1425;
               }
               }
              }
             } while(0);
             $e$0$i = HEAPU8[$e$06$i>>0]|(HEAPU8[$e$06$i+1>>0]<<8)|(HEAPU8[$e$06$i+2>>0]<<16)|(HEAPU8[$e$06$i+3>>0]<<24); //@line 3647 "load_abc.cpp"
             $2399 = ($e$0$i|0)==(0|0); //@line 3647 "load_abc.cpp"
             if ($2399) {
              break;
             } else {
              $e$06$i = $e$0$i;
             }
            }
           }
          } while(0);
          $2400 = (_strncasecmp($2364,527377,7)|0); //@line 4658 "load_abc.cpp"
          $2401 = ($2400|0)==(0); //@line 4658 "load_abc.cpp"
          if ($2401) {
           $2402 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4659 "load_abc.cpp"
           $2403 = ($2402|0)==(0|0); //@line 915 "load_abc.cpp"
           if ($2403) {
            $2404 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
            $2405 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
            $2406 = ((($2404)) + 20|0); //@line 917 "load_abc.cpp"
            HEAP8[$2406>>0]=$2405&255;HEAP8[$2406+1>>0]=($2405>>8)&255;HEAP8[$2406+2>>0]=($2405>>16)&255;HEAP8[$2406+3>>0]=$2405>>24; //@line 917 "load_abc.cpp"
            $$0$i464 = $2404;
           } else {
            $$0$i464 = $2402;
           }
           HEAP8[$41>>0]=$$0$i464&255;HEAP8[$41+1>>0]=($$0$i464>>8)&255;HEAP8[$41+2>>0]=($$0$i464>>16)&255;HEAP8[$41+3>>0]=$$0$i464>>24; //@line 4659 "load_abc.cpp"
           $2407 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4660 "load_abc.cpp"
           $2408 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
           HEAP8[$2408>>0]=0&255;HEAP8[$2408+1>>0]=(0>>8)&255;HEAP8[$2408+2>>0]=(0>>16)&255;HEAP8[$2408+3>>0]=0>>24; //@line 540 "load_abc.cpp"
           $2409 = ((($2408)) + 4|0); //@line 541 "load_abc.cpp"
           HEAP8[$2409>>0]=$2407&255;HEAP8[$2409+1>>0]=($2407>>8)&255;HEAP8[$2409+2>>0]=($2407>>16)&255;HEAP8[$2409+3>>0]=$2407>>24; //@line 541 "load_abc.cpp"
           $2410 = ((($2408)) + 8|0); //@line 543 "load_abc.cpp"
           HEAP8[$2410>>0] = 1; //@line 543 "load_abc.cpp"
           $2411 = ((($2408)) + 9|0); //@line 543 "load_abc.cpp"
           HEAP8[$2411>>0] = 80; //@line 543 "load_abc.cpp"
           $2412 = ((($2408)) + 10|0); //@line 543 "load_abc.cpp"
           HEAP8[$2412>>0]=0&255;HEAP8[$2412+1>>0]=(0>>8)&255;HEAP8[$2412+2>>0]=(0>>16)&255;HEAP8[$2412+3>>0]=0>>24; //@line 543 "load_abc.cpp"
           $2413 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
           $2414 = ((($2408)) + 14|0); //@line 544 "load_abc.cpp"
           HEAP8[$2414>>0] = $2413; //@line 544 "load_abc.cpp"
           $2415 = ((($2408)) + 15|0); //@line 545 "load_abc.cpp"
           HEAP8[$2415>>0] = 0; //@line 545 "load_abc.cpp"
           __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($$0$i464,$2408); //@line 1153 "load_abc.cpp"
           $2416 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4661 "load_abc.cpp"
           $2417 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4661 "load_abc.cpp"
           $2418 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
           HEAP8[$2418>>0]=0&255;HEAP8[$2418+1>>0]=(0>>8)&255;HEAP8[$2418+2>>0]=(0>>16)&255;HEAP8[$2418+3>>0]=0>>24; //@line 540 "load_abc.cpp"
           $2419 = ((($2418)) + 4|0); //@line 541 "load_abc.cpp"
           HEAP8[$2419>>0]=$2417&255;HEAP8[$2419+1>>0]=($2417>>8)&255;HEAP8[$2419+2>>0]=($2417>>16)&255;HEAP8[$2419+3>>0]=$2417>>24; //@line 541 "load_abc.cpp"
           $2420 = ((($2418)) + 8|0); //@line 543 "load_abc.cpp"
           HEAP8[$2420>>0] = 1; //@line 543 "load_abc.cpp"
           $2421 = ((($2418)) + 9|0); //@line 543 "load_abc.cpp"
           HEAP8[$2421>>0] = 88; //@line 543 "load_abc.cpp"
           $2422 = ((($2418)) + 10|0); //@line 543 "load_abc.cpp"
           HEAP8[$2422>>0]=0&255;HEAP8[$2422+1>>0]=(0>>8)&255;HEAP8[$2422+2>>0]=(0>>16)&255;HEAP8[$2422+3>>0]=0>>24; //@line 543 "load_abc.cpp"
           $2423 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
           $2424 = ((($2418)) + 14|0); //@line 544 "load_abc.cpp"
           HEAP8[$2424>>0] = $2423; //@line 544 "load_abc.cpp"
           $2425 = ((($2418)) + 15|0); //@line 545 "load_abc.cpp"
           HEAP8[$2425>>0] = 0; //@line 545 "load_abc.cpp"
           __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($2416,$2418); //@line 975 "load_abc.cpp"
           $2426 = ((($p$421189)) + 8|0); //@line 4662 "load_abc.cpp"
           $2434 = 10;$abcto$5 = -1;$p$43 = $2426;
          } else {
           $2427 = HEAP8[$2364>>0]|0; //@line 4666 "load_abc.cpp"
           $2428 = $2427 << 24 >> 24; //@line 4666 "load_abc.cpp"
           $2429 = (_isspace($2428)|0); //@line 4666 "load_abc.cpp"
           $2430 = ($2429|0)==(0); //@line 4666 "load_abc.cpp"
           $$abcto$4 = $2430 ? 0 : $abcto$41190; //@line 4666 "load_abc.cpp"
           $2434 = 3;$abcto$5 = $$abcto$4;$p$43 = $2364;
          }
          $2431 = (_strncasecmp($p$43,527385,2)|0); //@line 4667 "load_abc.cpp"
          $2432 = ($2431|0)==(0); //@line 4667 "load_abc.cpp"
          do {
           if ($2432) {
            $2433 = (($p$421189) + ($2434)|0); //@line 4667 "load_abc.cpp"
            $2435 = HEAP8[$2433>>0]|0; //@line 4667 "load_abc.cpp"
            $2436 = $2435 << 24 >> 24; //@line 4667 "load_abc.cpp"
            $2437 = (_isspace($2436)|0); //@line 4667 "load_abc.cpp"
            $2438 = ($2437|0)==(0); //@line 4667 "load_abc.cpp"
            if ($2438) {
             $2439 = HEAP8[$2433>>0]|0; //@line 4667 "load_abc.cpp"
             $2440 = ($2439<<24>>24)==(34); //@line 4667 "load_abc.cpp"
             if (!($2440)) {
              $abcto$4$be = $abcto$5;
              break;
             }
            }
            $abcto$4$be = 1;
           } else {
            $abcto$4$be = $abcto$5;
           }
          } while(0);
          $2441 = ((($p$43)) + 1|0); //@line 4655 "load_abc.cpp"
          $2442 = HEAP8[$p$43>>0]|0; //@line 4655 "load_abc.cpp"
          $2443 = ($2442<<24>>24)!=(0); //@line 4655 "load_abc.cpp"
          $$not = $2443 ^ 1; //@line 4655 "load_abc.cpp"
          $2444 = ($2442<<24>>24)==(34); //@line 4655 "load_abc.cpp"
          $or$cond90 = $2444 | $$not; //@line 4655 "load_abc.cpp"
          if ($or$cond90) {
           $$lcssa959 = $2441;$$lcssa960 = $2442;$$lcssa961 = $2443;$abcto$4$lcssa = $abcto$4$be;
           break;
          } else {
           $2364 = $2441;$abcto$41190 = $abcto$4$be;$p$421189 = $p$43;
          }
         }
        }
        $abcstate$7$ = $$lcssa961 ? $abcstate$71221 : 5; //@line 4669 "load_abc.cpp"
        $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$7$;$abctempo$6 = $abctempo$51231;$abcto$6 = $abcto$4$lcssa;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $$lcssa960;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $$lcssa959;
        break;
       }
       case 92:  {
        $p$44 = $p$18;
        while(1) {
         $2445 = HEAP8[$p$44>>0]|0; //@line 4672 "load_abc.cpp"
         $2446 = ($2445<<24>>24)==(0); //@line 4672 "load_abc.cpp"
         $2447 = ((($p$44)) + 1|0); //@line 4672 "load_abc.cpp"
         if ($2446) {
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = $abcto$31230;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = 92;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$44;
          break;
         } else {
          $p$44 = $2447;
         }
        }
        break;
       }
       case 43: case 33:  {
        $2448 = (_strncmp($p$18,527388,4)|0); //@line 4677 "load_abc.cpp"
        $2449 = ($2448|0)==(0); //@line 4677 "load_abc.cpp"
        do {
         if ($2449) {
          $2450 = ((($p$18)) + 4|0); //@line 4677 "load_abc.cpp"
          $2451 = HEAP8[$2450>>0]|0; //@line 4677 "load_abc.cpp"
          $2452 = ($2451<<24>>24)==($ch$0<<24>>24); //@line 4677 "load_abc.cpp"
          if (!($2452)) {
           break;
          }
          $2453 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4678 "load_abc.cpp"
          $2454 = ($2453|0)==(0|0); //@line 915 "load_abc.cpp"
          if ($2454) {
           $2455 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
           $2456 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
           $2457 = ((($2455)) + 20|0); //@line 917 "load_abc.cpp"
           HEAP8[$2457>>0]=$2456&255;HEAP8[$2457+1>>0]=($2456>>8)&255;HEAP8[$2457+2>>0]=($2456>>16)&255;HEAP8[$2457+3>>0]=$2456>>24; //@line 917 "load_abc.cpp"
           $$0$i466 = $2455;
          } else {
           $$0$i466 = $2453;
          }
          HEAP8[$41>>0]=$$0$i466&255;HEAP8[$41+1>>0]=($$0$i466>>8)&255;HEAP8[$41+2>>0]=($$0$i466>>16)&255;HEAP8[$41+3>>0]=$$0$i466>>24; //@line 4678 "load_abc.cpp"
          $2458 = ($abcto$31230|0)==(0); //@line 4679 "load_abc.cpp"
          do {
           if ($2458) {
            $2479 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4686 "load_abc.cpp"
            $2480 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
            HEAP8[$2480>>0]=0&255;HEAP8[$2480+1>>0]=(0>>8)&255;HEAP8[$2480+2>>0]=(0>>16)&255;HEAP8[$2480+3>>0]=0>>24; //@line 540 "load_abc.cpp"
            $2481 = ((($2480)) + 4|0); //@line 541 "load_abc.cpp"
            HEAP8[$2481>>0]=$2479&255;HEAP8[$2481+1>>0]=($2479>>8)&255;HEAP8[$2481+2>>0]=($2479>>16)&255;HEAP8[$2481+3>>0]=$2479>>24; //@line 541 "load_abc.cpp"
            $2482 = ((($2480)) + 8|0); //@line 543 "load_abc.cpp"
            HEAP8[$2482>>0] = 1; //@line 543 "load_abc.cpp"
            $2483 = ((($2480)) + 9|0); //@line 543 "load_abc.cpp"
            HEAP8[$2483>>0] = 80; //@line 543 "load_abc.cpp"
            $2484 = ((($2480)) + 10|0); //@line 543 "load_abc.cpp"
            HEAP8[$2484>>0]=0&255;HEAP8[$2484+1>>0]=(0>>8)&255;HEAP8[$2484+2>>0]=(0>>16)&255;HEAP8[$2484+3>>0]=0>>24; //@line 543 "load_abc.cpp"
            $2485 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
            $2486 = ((($2480)) + 14|0); //@line 544 "load_abc.cpp"
            HEAP8[$2486>>0] = $2485; //@line 544 "load_abc.cpp"
            $2487 = ((($2480)) + 15|0); //@line 545 "load_abc.cpp"
            HEAP8[$2487>>0] = 0; //@line 545 "load_abc.cpp"
            __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($$0$i466,$2480); //@line 1153 "load_abc.cpp"
            $2488 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4687 "load_abc.cpp"
            $2489 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4687 "load_abc.cpp"
            $2490 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
            HEAP8[$2490>>0]=0&255;HEAP8[$2490+1>>0]=(0>>8)&255;HEAP8[$2490+2>>0]=(0>>16)&255;HEAP8[$2490+3>>0]=0>>24; //@line 540 "load_abc.cpp"
            $2491 = ((($2490)) + 4|0); //@line 541 "load_abc.cpp"
            HEAP8[$2491>>0]=$2489&255;HEAP8[$2491+1>>0]=($2489>>8)&255;HEAP8[$2491+2>>0]=($2489>>16)&255;HEAP8[$2491+3>>0]=$2489>>24; //@line 541 "load_abc.cpp"
            $2492 = ((($2490)) + 8|0); //@line 543 "load_abc.cpp"
            HEAP8[$2492>>0] = 1; //@line 543 "load_abc.cpp"
            $2493 = ((($2490)) + 9|0); //@line 543 "load_abc.cpp"
            HEAP8[$2493>>0] = 79; //@line 543 "load_abc.cpp"
            $2494 = ((($2490)) + 10|0); //@line 543 "load_abc.cpp"
            HEAP8[$2494>>0]=0&255;HEAP8[$2494+1>>0]=(0>>8)&255;HEAP8[$2494+2>>0]=(0>>16)&255;HEAP8[$2494+3>>0]=0>>24; //@line 543 "load_abc.cpp"
            $2495 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
            $2496 = ((($2490)) + 14|0); //@line 544 "load_abc.cpp"
            HEAP8[$2496>>0] = $2495; //@line 544 "load_abc.cpp"
            $2497 = ((($2490)) + 15|0); //@line 545 "load_abc.cpp"
            HEAP8[$2497>>0] = 0; //@line 545 "load_abc.cpp"
            __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($2488,$2490); //@line 953 "load_abc.cpp"
           } else {
            $2459 = ($abcto$31230|0)>(0); //@line 4680 "load_abc.cpp"
            if (!($2459)) {
             break;
            }
            $2460 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4681 "load_abc.cpp"
            $2461 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
            HEAP8[$2461>>0]=0&255;HEAP8[$2461+1>>0]=(0>>8)&255;HEAP8[$2461+2>>0]=(0>>16)&255;HEAP8[$2461+3>>0]=0>>24; //@line 540 "load_abc.cpp"
            $2462 = ((($2461)) + 4|0); //@line 541 "load_abc.cpp"
            HEAP8[$2462>>0]=$2460&255;HEAP8[$2462+1>>0]=($2460>>8)&255;HEAP8[$2462+2>>0]=($2460>>16)&255;HEAP8[$2462+3>>0]=$2460>>24; //@line 541 "load_abc.cpp"
            $2463 = ((($2461)) + 8|0); //@line 543 "load_abc.cpp"
            HEAP8[$2463>>0] = 1; //@line 543 "load_abc.cpp"
            $2464 = ((($2461)) + 9|0); //@line 543 "load_abc.cpp"
            HEAP8[$2464>>0] = 80; //@line 543 "load_abc.cpp"
            $2465 = ((($2461)) + 10|0); //@line 543 "load_abc.cpp"
            HEAP8[$2465>>0]=0&255;HEAP8[$2465+1>>0]=(0>>8)&255;HEAP8[$2465+2>>0]=(0>>16)&255;HEAP8[$2465+3>>0]=0>>24; //@line 543 "load_abc.cpp"
            $2466 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
            $2467 = ((($2461)) + 14|0); //@line 544 "load_abc.cpp"
            HEAP8[$2467>>0] = $2466; //@line 544 "load_abc.cpp"
            $2468 = ((($2461)) + 15|0); //@line 545 "load_abc.cpp"
            HEAP8[$2468>>0] = 0; //@line 545 "load_abc.cpp"
            __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($$0$i466,$2461); //@line 1153 "load_abc.cpp"
            $2469 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4682 "load_abc.cpp"
            $2470 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4682 "load_abc.cpp"
            $2471 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
            HEAP8[$2471>>0]=0&255;HEAP8[$2471+1>>0]=(0>>8)&255;HEAP8[$2471+2>>0]=(0>>16)&255;HEAP8[$2471+3>>0]=0>>24; //@line 540 "load_abc.cpp"
            $2472 = ((($2471)) + 4|0); //@line 541 "load_abc.cpp"
            HEAP8[$2472>>0]=$2470&255;HEAP8[$2472+1>>0]=($2470>>8)&255;HEAP8[$2472+2>>0]=($2470>>16)&255;HEAP8[$2472+3>>0]=$2470>>24; //@line 541 "load_abc.cpp"
            $2473 = ((($2471)) + 8|0); //@line 543 "load_abc.cpp"
            HEAP8[$2473>>0] = 1; //@line 543 "load_abc.cpp"
            $2474 = ((($2471)) + 9|0); //@line 543 "load_abc.cpp"
            HEAP8[$2474>>0] = 88; //@line 543 "load_abc.cpp"
            $2475 = ((($2471)) + 10|0); //@line 543 "load_abc.cpp"
            HEAP8[$2475>>0]=0&255;HEAP8[$2475+1>>0]=(0>>8)&255;HEAP8[$2475+2>>0]=(0>>16)&255;HEAP8[$2475+3>>0]=0>>24; //@line 543 "load_abc.cpp"
            $2476 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
            $2477 = ((($2471)) + 14|0); //@line 544 "load_abc.cpp"
            HEAP8[$2477>>0] = $2476; //@line 544 "load_abc.cpp"
            $2478 = ((($2471)) + 15|0); //@line 545 "load_abc.cpp"
            HEAP8[$2478>>0] = 0; //@line 545 "load_abc.cpp"
            __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($2469,$2471); //@line 975 "load_abc.cpp"
           }
          } while(0);
          $2498 = ((($p$18)) + 5|0); //@line 4689 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2498;
          break L678;
         }
        } while(0);
        $2499 = (_strncmp($p$18,527393,8)|0); //@line 4694 "load_abc.cpp"
        $2500 = ($2499|0)==(0); //@line 4694 "load_abc.cpp"
        do {
         if ($2500) {
          $2501 = ((($p$18)) + 8|0); //@line 4694 "load_abc.cpp"
          $2502 = HEAP8[$2501>>0]|0; //@line 4694 "load_abc.cpp"
          $2503 = ($2502<<24>>24)==($ch$0<<24>>24); //@line 4694 "load_abc.cpp"
          if (!($2503)) {
           break;
          }
          $2504 = ((($p$18)) + 9|0); //@line 4696 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = 1;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2504;
          break L678;
         }
        } while(0);
        $2505 = (_strncmp($p$18,527402,10)|0); //@line 4699 "load_abc.cpp"
        $2506 = ($2505|0)==(0); //@line 4699 "load_abc.cpp"
        do {
         if ($2506) {
          $2507 = ((($p$18)) + 10|0); //@line 4699 "load_abc.cpp"
          $2508 = HEAP8[$2507>>0]|0; //@line 4699 "load_abc.cpp"
          $2509 = ($2508<<24>>24)==($ch$0<<24>>24); //@line 4699 "load_abc.cpp"
          if (!($2509)) {
           break;
          }
          $2510 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4700 "load_abc.cpp"
          $2511 = ($2510|0)==(0|0); //@line 915 "load_abc.cpp"
          if ($2511) {
           $2512 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
           $2513 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
           $2514 = ((($2512)) + 20|0); //@line 917 "load_abc.cpp"
           HEAP8[$2514>>0]=$2513&255;HEAP8[$2514+1>>0]=($2513>>8)&255;HEAP8[$2514+2>>0]=($2513>>16)&255;HEAP8[$2514+3>>0]=$2513>>24; //@line 917 "load_abc.cpp"
           $$0$i468 = $2512;
          } else {
           $$0$i468 = $2510;
          }
          HEAP8[$41>>0]=$$0$i468&255;HEAP8[$41+1>>0]=($$0$i468>>8)&255;HEAP8[$41+2>>0]=($$0$i468>>16)&255;HEAP8[$41+3>>0]=$$0$i468>>24; //@line 4700 "load_abc.cpp"
          $2515 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4701 "load_abc.cpp"
          $2516 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 3633 "load_abc.cpp"
          $2517 = ($2516|0)==(0|0); //@line 3633 "load_abc.cpp"
          if (!($2517)) {
           $tp$08$i471 = $2516;
           while(1) {
            $2518 = ((($tp$08$i471)) + 28|0); //@line 3636 "load_abc.cpp"
            HEAP8[$2518>>0]=$2515&255;HEAP8[$2518+1>>0]=($2515>>8)&255;HEAP8[$2518+2>>0]=($2515>>16)&255;HEAP8[$2518+3>>0]=$2515>>24; //@line 3636 "load_abc.cpp"
            $2519 = ((($tp$08$i471)) + 32|0); //@line 3640 "load_abc.cpp"
            $2520 = HEAPU8[$2519>>0]|(HEAPU8[$2519+1>>0]<<8)|(HEAPU8[$2519+2>>0]<<16)|(HEAPU8[$2519+3>>0]<<24); //@line 3640 "load_abc.cpp"
            $2521 = ($2520|0)>(-2); //@line 3640 "load_abc.cpp"
            if ($2521) {
             HEAP8[$2519>>0]=1&255;HEAP8[$2519+1>>0]=(1>>8)&255;HEAP8[$2519+2>>0]=(1>>16)&255;HEAP8[$2519+3>>0]=1>>24; //@line 3641 "load_abc.cpp"
            }
            $tp$0$i473 = HEAPU8[$tp$08$i471>>0]|(HEAPU8[$tp$08$i471+1>>0]<<8)|(HEAPU8[$tp$08$i471+2>>0]<<16)|(HEAPU8[$tp$08$i471+3>>0]<<24); //@line 3634 "load_abc.cpp"
            $2522 = ($tp$0$i473|0)==(0|0); //@line 3634 "load_abc.cpp"
            if ($2522) {
             break;
            } else {
             $tp$08$i471 = $tp$0$i473;
            }
           }
          }
          $2523 = ((($p$18)) + 11|0); //@line 4702 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2523;
          break L678;
         }
        } while(0);
        $2524 = (_strncmp($p$18,527413,10)|0); //@line 4705 "load_abc.cpp"
        $2525 = ($2524|0)==(0); //@line 4705 "load_abc.cpp"
        do {
         if ($2525) {
          $2526 = ((($p$18)) + 10|0); //@line 4705 "load_abc.cpp"
          $2527 = HEAP8[$2526>>0]|0; //@line 4705 "load_abc.cpp"
          $2528 = ($2527<<24>>24)==($ch$0<<24>>24); //@line 4705 "load_abc.cpp"
          if (!($2528)) {
           break;
          }
          $2529 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4706 "load_abc.cpp"
          $2530 = ($2529|0)==(0|0); //@line 915 "load_abc.cpp"
          if ($2530) {
           $2531 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
           $2532 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
           $2533 = ((($2531)) + 20|0); //@line 917 "load_abc.cpp"
           HEAP8[$2533>>0]=$2532&255;HEAP8[$2533+1>>0]=($2532>>8)&255;HEAP8[$2533+2>>0]=($2532>>16)&255;HEAP8[$2533+3>>0]=$2532>>24; //@line 917 "load_abc.cpp"
           $$0$i477 = $2531;
          } else {
           $$0$i477 = $2529;
          }
          HEAP8[$41>>0]=$$0$i477&255;HEAP8[$41+1>>0]=($$0$i477>>8)&255;HEAP8[$41+2>>0]=($$0$i477>>16)&255;HEAP8[$41+3>>0]=$$0$i477>>24; //@line 4706 "load_abc.cpp"
          $2534 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 3633 "load_abc.cpp"
          $2535 = ($2534|0)==(0|0); //@line 3633 "load_abc.cpp"
          if (!($2535)) {
           $tp$08$us$i = $2534;
           while(1) {
            $2536 = ((($tp$08$us$i)) + 32|0); //@line 3640 "load_abc.cpp"
            $2537 = HEAPU8[$2536>>0]|(HEAPU8[$2536+1>>0]<<8)|(HEAPU8[$2536+2>>0]<<16)|(HEAPU8[$2536+3>>0]<<24); //@line 3640 "load_abc.cpp"
            $2538 = ($2537|0)>(-2); //@line 3640 "load_abc.cpp"
            if ($2538) {
             HEAP8[$2536>>0]=0&255;HEAP8[$2536+1>>0]=(0>>8)&255;HEAP8[$2536+2>>0]=(0>>16)&255;HEAP8[$2536+3>>0]=0>>24; //@line 3641 "load_abc.cpp"
            }
            $tp$0$us$i = HEAPU8[$tp$08$us$i>>0]|(HEAPU8[$tp$08$us$i+1>>0]<<8)|(HEAPU8[$tp$08$us$i+2>>0]<<16)|(HEAPU8[$tp$08$us$i+3>>0]<<24); //@line 3634 "load_abc.cpp"
            $2539 = ($tp$0$us$i|0)==(0|0); //@line 3634 "load_abc.cpp"
            if ($2539) {
             break;
            } else {
             $tp$08$us$i = $tp$0$us$i;
            }
           }
          }
          $2540 = ((($p$18)) + 11|0); //@line 4708 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2540;
          break L678;
         }
        } while(0);
        $2541 = (_strncmp($p$18,527424,2)|0); //@line 4711 "load_abc.cpp"
        $2542 = ($2541|0)==(0); //@line 4711 "load_abc.cpp"
        do {
         if ($2542) {
          $2543 = ((($p$18)) + 2|0); //@line 4711 "load_abc.cpp"
          $2544 = HEAP8[$2543>>0]|0; //@line 4711 "load_abc.cpp"
          $2545 = ($2544<<24>>24)==($ch$0<<24>>24); //@line 4711 "load_abc.cpp"
          if (!($2545)) {
           break;
          }
          $2546 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4712 "load_abc.cpp"
          $2547 = ($2546|0)==(0|0); //@line 915 "load_abc.cpp"
          if ($2547) {
           $2548 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
           $2549 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
           $2550 = ((($2548)) + 20|0); //@line 917 "load_abc.cpp"
           HEAP8[$2550>>0]=$2549&255;HEAP8[$2550+1>>0]=($2549>>8)&255;HEAP8[$2550+2>>0]=($2549>>16)&255;HEAP8[$2550+3>>0]=$2549>>24; //@line 917 "load_abc.cpp"
           $$0$i483 = $2548;
          } else {
           $$0$i483 = $2546;
          }
          HEAP8[$41>>0]=$$0$i483&255;HEAP8[$41+1>>0]=($$0$i483>>8)&255;HEAP8[$41+2>>0]=($$0$i483>>16)&255;HEAP8[$41+3>>0]=$$0$i483>>24; //@line 4712 "load_abc.cpp"
          $2551 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4713 "load_abc.cpp"
          $2552 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 3633 "load_abc.cpp"
          $2553 = ($2552|0)==(0|0); //@line 3633 "load_abc.cpp"
          if (!($2553)) {
           $tp$08$i486 = $2552;
           while(1) {
            $2554 = ((($tp$08$i486)) + 28|0); //@line 3636 "load_abc.cpp"
            HEAP8[$2554>>0]=$2551&255;HEAP8[$2554+1>>0]=($2551>>8)&255;HEAP8[$2554+2>>0]=($2551>>16)&255;HEAP8[$2554+3>>0]=$2551>>24; //@line 3636 "load_abc.cpp"
            $2555 = ((($tp$08$i486)) + 32|0); //@line 3640 "load_abc.cpp"
            $2556 = HEAPU8[$2555>>0]|(HEAPU8[$2555+1>>0]<<8)|(HEAPU8[$2555+2>>0]<<16)|(HEAPU8[$2555+3>>0]<<24); //@line 3640 "load_abc.cpp"
            $2557 = ($2556|0)>(-2); //@line 3640 "load_abc.cpp"
            if ($2557) {
             HEAP8[$2555>>0]=1&255;HEAP8[$2555+1>>0]=(1>>8)&255;HEAP8[$2555+2>>0]=(1>>16)&255;HEAP8[$2555+3>>0]=1>>24; //@line 3641 "load_abc.cpp"
            }
            $tp$0$i488 = HEAPU8[$tp$08$i486>>0]|(HEAPU8[$tp$08$i486+1>>0]<<8)|(HEAPU8[$tp$08$i486+2>>0]<<16)|(HEAPU8[$tp$08$i486+3>>0]<<24); //@line 3634 "load_abc.cpp"
            $2558 = ($tp$0$i488|0)==(0|0); //@line 3634 "load_abc.cpp"
            if ($2558) {
             break;
            } else {
             $tp$08$i486 = $tp$0$i488;
            }
           }
          }
          $2559 = ((($p$18)) + 3|0); //@line 4714 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2559;
          break L678;
         }
        } while(0);
        $2560 = (_strncmp($p$18,527427,2)|0); //@line 4717 "load_abc.cpp"
        $2561 = ($2560|0)==(0); //@line 4717 "load_abc.cpp"
        do {
         if ($2561) {
          $2562 = ((($p$18)) + 2|0); //@line 4717 "load_abc.cpp"
          $2563 = HEAP8[$2562>>0]|0; //@line 4717 "load_abc.cpp"
          $2564 = ($2563<<24>>24)==($ch$0<<24>>24); //@line 4717 "load_abc.cpp"
          if (!($2564)) {
           break;
          }
          $2565 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4718 "load_abc.cpp"
          $2566 = ($2565|0)==(0|0); //@line 915 "load_abc.cpp"
          if ($2566) {
           $2567 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
           $2568 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
           $2569 = ((($2567)) + 20|0); //@line 917 "load_abc.cpp"
           HEAP8[$2569>>0]=$2568&255;HEAP8[$2569+1>>0]=($2568>>8)&255;HEAP8[$2569+2>>0]=($2568>>16)&255;HEAP8[$2569+3>>0]=$2568>>24; //@line 917 "load_abc.cpp"
           $$0$i492 = $2567;
          } else {
           $$0$i492 = $2565;
          }
          HEAP8[$41>>0]=$$0$i492&255;HEAP8[$41+1>>0]=($$0$i492>>8)&255;HEAP8[$41+2>>0]=($$0$i492>>16)&255;HEAP8[$41+3>>0]=$$0$i492>>24; //@line 4718 "load_abc.cpp"
          $2570 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 3633 "load_abc.cpp"
          $2571 = ($2570|0)==(0|0); //@line 3633 "load_abc.cpp"
          if (!($2571)) {
           $tp$08$us$i495 = $2570;
           while(1) {
            $2572 = ((($tp$08$us$i495)) + 32|0); //@line 3640 "load_abc.cpp"
            $2573 = HEAPU8[$2572>>0]|(HEAPU8[$2572+1>>0]<<8)|(HEAPU8[$2572+2>>0]<<16)|(HEAPU8[$2572+3>>0]<<24); //@line 3640 "load_abc.cpp"
            $2574 = ($2573|0)>(-2); //@line 3640 "load_abc.cpp"
            if ($2574) {
             HEAP8[$2572>>0]=0&255;HEAP8[$2572+1>>0]=(0>>8)&255;HEAP8[$2572+2>>0]=(0>>16)&255;HEAP8[$2572+3>>0]=0>>24; //@line 3641 "load_abc.cpp"
            }
            $tp$0$us$i497 = HEAPU8[$tp$08$us$i495>>0]|(HEAPU8[$tp$08$us$i495+1>>0]<<8)|(HEAPU8[$tp$08$us$i495+2>>0]<<16)|(HEAPU8[$tp$08$us$i495+3>>0]<<24); //@line 3634 "load_abc.cpp"
            $2575 = ($tp$0$us$i497|0)==(0|0); //@line 3634 "load_abc.cpp"
            if ($2575) {
             break;
            } else {
             $tp$08$us$i495 = $tp$0$us$i497;
            }
           }
          }
          $2576 = ((($p$18)) + 3|0); //@line 4720 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2576;
          break L678;
         }
        } while(0);
        $2577 = (_strncmp($p$18,527430,11)|0); //@line 4723 "load_abc.cpp"
        $2578 = ($2577|0)==(0); //@line 4723 "load_abc.cpp"
        do {
         if ($2578) {
          $2579 = ((($p$18)) + 11|0); //@line 4723 "load_abc.cpp"
          $2580 = HEAP8[$2579>>0]|0; //@line 4723 "load_abc.cpp"
          $2581 = ($2580<<24>>24)==($ch$0<<24>>24); //@line 4723 "load_abc.cpp"
          if (!($2581)) {
           break;
          }
          $2582 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4724 "load_abc.cpp"
          $2583 = ($2582|0)==(0|0); //@line 915 "load_abc.cpp"
          if ($2583) {
           $2584 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
           $2585 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
           $2586 = ((($2584)) + 20|0); //@line 917 "load_abc.cpp"
           HEAP8[$2586>>0]=$2585&255;HEAP8[$2586+1>>0]=($2585>>8)&255;HEAP8[$2586+2>>0]=($2585>>16)&255;HEAP8[$2586+3>>0]=$2585>>24; //@line 917 "load_abc.cpp"
           $$0$i501 = $2584;
          } else {
           $$0$i501 = $2582;
          }
          HEAP8[$41>>0]=$$0$i501&255;HEAP8[$41+1>>0]=($$0$i501>>8)&255;HEAP8[$41+2>>0]=($$0$i501>>16)&255;HEAP8[$41+3>>0]=$$0$i501>>24; //@line 4724 "load_abc.cpp"
          $2587 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4725 "load_abc.cpp"
          $2588 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 3633 "load_abc.cpp"
          $2589 = ($2588|0)==(0|0); //@line 3633 "load_abc.cpp"
          if (!($2589)) {
           $tp$08$i504 = $2588;
           while(1) {
            $2590 = ((($tp$08$i504)) + 28|0); //@line 3636 "load_abc.cpp"
            HEAP8[$2590>>0]=$2587&255;HEAP8[$2590+1>>0]=($2587>>8)&255;HEAP8[$2590+2>>0]=($2587>>16)&255;HEAP8[$2590+3>>0]=$2587>>24; //@line 3636 "load_abc.cpp"
            $2591 = ((($tp$08$i504)) + 32|0); //@line 3640 "load_abc.cpp"
            $2592 = HEAPU8[$2591>>0]|(HEAPU8[$2591+1>>0]<<8)|(HEAPU8[$2591+2>>0]<<16)|(HEAPU8[$2591+3>>0]<<24); //@line 3640 "load_abc.cpp"
            $2593 = ($2592|0)>(-2); //@line 3640 "load_abc.cpp"
            if ($2593) {
             HEAP8[$2591>>0]=-1&255;HEAP8[$2591+1>>0]=(-1>>8)&255;HEAP8[$2591+2>>0]=(-1>>16)&255;HEAP8[$2591+3>>0]=-1>>24; //@line 3641 "load_abc.cpp"
            }
            $tp$0$i506 = HEAPU8[$tp$08$i504>>0]|(HEAPU8[$tp$08$i504+1>>0]<<8)|(HEAPU8[$tp$08$i504+2>>0]<<16)|(HEAPU8[$tp$08$i504+3>>0]<<24); //@line 3634 "load_abc.cpp"
            $2594 = ($tp$0$i506|0)==(0|0); //@line 3634 "load_abc.cpp"
            if ($2594) {
             break;
            } else {
             $tp$08$i504 = $tp$0$i506;
            }
           }
          }
          $2595 = ((($p$18)) + 12|0); //@line 4726 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2595;
          break L678;
         }
        } while(0);
        $2596 = (_strncmp($p$18,527442,11)|0); //@line 4729 "load_abc.cpp"
        $2597 = ($2596|0)==(0); //@line 4729 "load_abc.cpp"
        do {
         if ($2597) {
          $2598 = ((($p$18)) + 11|0); //@line 4729 "load_abc.cpp"
          $2599 = HEAP8[$2598>>0]|0; //@line 4729 "load_abc.cpp"
          $2600 = ($2599<<24>>24)==($ch$0<<24>>24); //@line 4729 "load_abc.cpp"
          if (!($2600)) {
           break;
          }
          $2601 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4730 "load_abc.cpp"
          $2602 = ($2601|0)==(0|0); //@line 915 "load_abc.cpp"
          if ($2602) {
           $2603 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
           $2604 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
           $2605 = ((($2603)) + 20|0); //@line 917 "load_abc.cpp"
           HEAP8[$2605>>0]=$2604&255;HEAP8[$2605+1>>0]=($2604>>8)&255;HEAP8[$2605+2>>0]=($2604>>16)&255;HEAP8[$2605+3>>0]=$2604>>24; //@line 917 "load_abc.cpp"
           $$0$i510 = $2603;
          } else {
           $$0$i510 = $2601;
          }
          HEAP8[$41>>0]=$$0$i510&255;HEAP8[$41+1>>0]=($$0$i510>>8)&255;HEAP8[$41+2>>0]=($$0$i510>>16)&255;HEAP8[$41+3>>0]=$$0$i510>>24; //@line 4730 "load_abc.cpp"
          $2606 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 3633 "load_abc.cpp"
          $2607 = ($2606|0)==(0|0); //@line 3633 "load_abc.cpp"
          if (!($2607)) {
           $tp$08$us$i513 = $2606;
           while(1) {
            $2608 = ((($tp$08$us$i513)) + 32|0); //@line 3640 "load_abc.cpp"
            $2609 = HEAPU8[$2608>>0]|(HEAPU8[$2608+1>>0]<<8)|(HEAPU8[$2608+2>>0]<<16)|(HEAPU8[$2608+3>>0]<<24); //@line 3640 "load_abc.cpp"
            $2610 = ($2609|0)>(-2); //@line 3640 "load_abc.cpp"
            if ($2610) {
             HEAP8[$2608>>0]=0&255;HEAP8[$2608+1>>0]=(0>>8)&255;HEAP8[$2608+2>>0]=(0>>16)&255;HEAP8[$2608+3>>0]=0>>24; //@line 3641 "load_abc.cpp"
            }
            $tp$0$us$i515 = HEAPU8[$tp$08$us$i513>>0]|(HEAPU8[$tp$08$us$i513+1>>0]<<8)|(HEAPU8[$tp$08$us$i513+2>>0]<<16)|(HEAPU8[$tp$08$us$i513+3>>0]<<24); //@line 3634 "load_abc.cpp"
            $2611 = ($tp$0$us$i515|0)==(0|0); //@line 3634 "load_abc.cpp"
            if ($2611) {
             break;
            } else {
             $tp$08$us$i513 = $tp$0$us$i515;
            }
           }
          }
          $2612 = ((($p$18)) + 12|0); //@line 4732 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2612;
          break L678;
         }
        } while(0);
        $2613 = (_strncmp($p$18,527454,2)|0); //@line 4735 "load_abc.cpp"
        $2614 = ($2613|0)==(0); //@line 4735 "load_abc.cpp"
        do {
         if ($2614) {
          $2615 = ((($p$18)) + 2|0); //@line 4735 "load_abc.cpp"
          $2616 = HEAP8[$2615>>0]|0; //@line 4735 "load_abc.cpp"
          $2617 = ($2616<<24>>24)==($ch$0<<24>>24); //@line 4735 "load_abc.cpp"
          if (!($2617)) {
           break;
          }
          $2618 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4736 "load_abc.cpp"
          $2619 = ($2618|0)==(0|0); //@line 915 "load_abc.cpp"
          if ($2619) {
           $2620 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
           $2621 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
           $2622 = ((($2620)) + 20|0); //@line 917 "load_abc.cpp"
           HEAP8[$2622>>0]=$2621&255;HEAP8[$2622+1>>0]=($2621>>8)&255;HEAP8[$2622+2>>0]=($2621>>16)&255;HEAP8[$2622+3>>0]=$2621>>24; //@line 917 "load_abc.cpp"
           $$0$i519 = $2620;
          } else {
           $$0$i519 = $2618;
          }
          HEAP8[$41>>0]=$$0$i519&255;HEAP8[$41+1>>0]=($$0$i519>>8)&255;HEAP8[$41+2>>0]=($$0$i519>>16)&255;HEAP8[$41+3>>0]=$$0$i519>>24; //@line 4736 "load_abc.cpp"
          $2623 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4737 "load_abc.cpp"
          $2624 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 3633 "load_abc.cpp"
          $2625 = ($2624|0)==(0|0); //@line 3633 "load_abc.cpp"
          if (!($2625)) {
           $tp$08$i522 = $2624;
           while(1) {
            $2626 = ((($tp$08$i522)) + 28|0); //@line 3636 "load_abc.cpp"
            HEAP8[$2626>>0]=$2623&255;HEAP8[$2626+1>>0]=($2623>>8)&255;HEAP8[$2626+2>>0]=($2623>>16)&255;HEAP8[$2626+3>>0]=$2623>>24; //@line 3636 "load_abc.cpp"
            $2627 = ((($tp$08$i522)) + 32|0); //@line 3640 "load_abc.cpp"
            $2628 = HEAPU8[$2627>>0]|(HEAPU8[$2627+1>>0]<<8)|(HEAPU8[$2627+2>>0]<<16)|(HEAPU8[$2627+3>>0]<<24); //@line 3640 "load_abc.cpp"
            $2629 = ($2628|0)>(-2); //@line 3640 "load_abc.cpp"
            if ($2629) {
             HEAP8[$2627>>0]=-1&255;HEAP8[$2627+1>>0]=(-1>>8)&255;HEAP8[$2627+2>>0]=(-1>>16)&255;HEAP8[$2627+3>>0]=-1>>24; //@line 3641 "load_abc.cpp"
            }
            $tp$0$i524 = HEAPU8[$tp$08$i522>>0]|(HEAPU8[$tp$08$i522+1>>0]<<8)|(HEAPU8[$tp$08$i522+2>>0]<<16)|(HEAPU8[$tp$08$i522+3>>0]<<24); //@line 3634 "load_abc.cpp"
            $2630 = ($tp$0$i524|0)==(0|0); //@line 3634 "load_abc.cpp"
            if ($2630) {
             break;
            } else {
             $tp$08$i522 = $tp$0$i524;
            }
           }
          }
          $2631 = ((($p$18)) + 3|0); //@line 4738 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2631;
          break L678;
         }
        } while(0);
        $2632 = (_strncmp($p$18,527457,2)|0); //@line 4741 "load_abc.cpp"
        $2633 = ($2632|0)==(0); //@line 4741 "load_abc.cpp"
        do {
         if ($2633) {
          $2634 = ((($p$18)) + 2|0); //@line 4741 "load_abc.cpp"
          $2635 = HEAP8[$2634>>0]|0; //@line 4741 "load_abc.cpp"
          $2636 = ($2635<<24>>24)==($ch$0<<24>>24); //@line 4741 "load_abc.cpp"
          if (!($2636)) {
           break;
          }
          $2637 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4742 "load_abc.cpp"
          $2638 = ($2637|0)==(0|0); //@line 915 "load_abc.cpp"
          if ($2638) {
           $2639 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
           $2640 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
           $2641 = ((($2639)) + 20|0); //@line 917 "load_abc.cpp"
           HEAP8[$2641>>0]=$2640&255;HEAP8[$2641+1>>0]=($2640>>8)&255;HEAP8[$2641+2>>0]=($2640>>16)&255;HEAP8[$2641+3>>0]=$2640>>24; //@line 917 "load_abc.cpp"
           $$0$i528 = $2639;
          } else {
           $$0$i528 = $2637;
          }
          HEAP8[$41>>0]=$$0$i528&255;HEAP8[$41+1>>0]=($$0$i528>>8)&255;HEAP8[$41+2>>0]=($$0$i528>>16)&255;HEAP8[$41+3>>0]=$$0$i528>>24; //@line 4742 "load_abc.cpp"
          $2642 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 3633 "load_abc.cpp"
          $2643 = ($2642|0)==(0|0); //@line 3633 "load_abc.cpp"
          if (!($2643)) {
           $tp$08$us$i531 = $2642;
           while(1) {
            $2644 = ((($tp$08$us$i531)) + 32|0); //@line 3640 "load_abc.cpp"
            $2645 = HEAPU8[$2644>>0]|(HEAPU8[$2644+1>>0]<<8)|(HEAPU8[$2644+2>>0]<<16)|(HEAPU8[$2644+3>>0]<<24); //@line 3640 "load_abc.cpp"
            $2646 = ($2645|0)>(-2); //@line 3640 "load_abc.cpp"
            if ($2646) {
             HEAP8[$2644>>0]=0&255;HEAP8[$2644+1>>0]=(0>>8)&255;HEAP8[$2644+2>>0]=(0>>16)&255;HEAP8[$2644+3>>0]=0>>24; //@line 3641 "load_abc.cpp"
            }
            $tp$0$us$i533 = HEAPU8[$tp$08$us$i531>>0]|(HEAPU8[$tp$08$us$i531+1>>0]<<8)|(HEAPU8[$tp$08$us$i531+2>>0]<<16)|(HEAPU8[$tp$08$us$i531+3>>0]<<24); //@line 3634 "load_abc.cpp"
            $2647 = ($tp$0$us$i533|0)==(0|0); //@line 3634 "load_abc.cpp"
            if ($2647) {
             break;
            } else {
             $tp$08$us$i531 = $tp$0$us$i533;
            }
           }
          }
          $2648 = ((($p$18)) + 3|0); //@line 4744 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2648;
          break L678;
         }
        } while(0);
        $2649 = (_strncmp($p$18,527460,5)|0); //@line 4747 "load_abc.cpp"
        $2650 = ($2649|0)==(0); //@line 4747 "load_abc.cpp"
        do {
         if ($2650) {
          $2651 = ((($p$18)) + 5|0); //@line 4747 "load_abc.cpp"
          $2652 = HEAP8[$2651>>0]|0; //@line 4747 "load_abc.cpp"
          $2653 = ($2652<<24>>24)==($ch$0<<24>>24); //@line 4747 "load_abc.cpp"
          if (!($2653)) {
           break;
          }
          $2654 = ((($p$18)) + 6|0); //@line 4750 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = 2;$abceffoper$4 = 1;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2654;
          break L678;
         }
        } while(0);
        $2655 = (_strncmp($p$18,527466,7)|0); //@line 4753 "load_abc.cpp"
        $2656 = ($2655|0)==(0); //@line 4753 "load_abc.cpp"
        do {
         if ($2656) {
          $2657 = ((($p$18)) + 7|0); //@line 4753 "load_abc.cpp"
          $2658 = HEAP8[$2657>>0]|0; //@line 4753 "load_abc.cpp"
          $2659 = ($2658<<24>>24)==($ch$0<<24>>24); //@line 4753 "load_abc.cpp"
          if (!($2659)) {
           break;
          }
          $2660 = ((($p$18)) + 8|0); //@line 4756 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = 2;$abceffoper$4 = 0;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2660;
          break L678;
         }
        } while(0);
        $2661 = (_strncmp($p$18,527474,5)|0); //@line 4759 "load_abc.cpp"
        $2662 = ($2661|0)==(0); //@line 4759 "load_abc.cpp"
        do {
         if ($2662) {
          $2663 = ((($p$18)) + 5|0); //@line 4759 "load_abc.cpp"
          $2664 = HEAP8[$2663>>0]|0; //@line 4759 "load_abc.cpp"
          $2665 = ($2664<<24>>24)==($ch$0<<24>>24); //@line 4759 "load_abc.cpp"
          if (!($2665)) {
           break;
          }
          $2666 = ((($p$18)) + 6|0); //@line 4762 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = 1;$abceffoper$4 = 0;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2666;
          break L678;
         }
        } while(0);
        $2667 = (_strncmp($p$18,527480,6)|0); //@line 4765 "load_abc.cpp"
        $2668 = ($2667|0)==(0); //@line 4765 "load_abc.cpp"
        do {
         if ($2668) {
          $2669 = ((($p$18)) + 6|0); //@line 4765 "load_abc.cpp"
          $2670 = HEAP8[$2669>>0]|0; //@line 4765 "load_abc.cpp"
          $2671 = ($2670<<24>>24)==($ch$0<<24>>24); //@line 4765 "load_abc.cpp"
          if (!($2671)) {
           break;
          }
          $2672 = ((($p$18)) + 7|0); //@line 4768 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = 1;$abceffoper$4 = 255;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2672;
          break L678;
         }
        } while(0);
        $2673 = (_strncmp($p$18,527487,6)|0); //@line 4771 "load_abc.cpp"
        $2674 = ($2673|0)==(0); //@line 4771 "load_abc.cpp"
        do {
         if ($2674) {
          $2675 = ((($p$18)) + 6|0); //@line 4771 "load_abc.cpp"
          $2676 = HEAP8[$2675>>0]|0; //@line 4771 "load_abc.cpp"
          $2677 = ($2676<<24>>24)==($ch$0<<24>>24); //@line 4771 "load_abc.cpp"
          if (!($2677)) {
           break;
          }
          $2678 = ((($p$18)) + 7|0); //@line 4774 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = 0;$abceffoper$4 = 0;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2678;
          break L678;
         }
        } while(0);
        $2679 = (_strncmp($p$18,527494,6)|0); //@line 4777 "load_abc.cpp"
        $2680 = ($2679|0)==(0); //@line 4777 "load_abc.cpp"
        do {
         if ($2680) {
          $2681 = ((($p$18)) + 6|0); //@line 4777 "load_abc.cpp"
          $2682 = HEAP8[$2681>>0]|0; //@line 4777 "load_abc.cpp"
          $2683 = ($2682<<24>>24)==($ch$0<<24>>24); //@line 4777 "load_abc.cpp"
          if (!($2683)) {
           break;
          }
          $2684 = ((($p$18)) + 7|0); //@line 4780 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = 3;$abceffoper$4 = 0;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2684;
          break L678;
         }
        } while(0);
        $2685 = (_strncmp($p$18,527501,8)|0); //@line 4783 "load_abc.cpp"
        $2686 = ($2685|0)==(0); //@line 4783 "load_abc.cpp"
        do {
         if ($2686) {
          $2687 = ((($p$18)) + 8|0); //@line 4783 "load_abc.cpp"
          $2688 = HEAP8[$2687>>0]|0; //@line 4783 "load_abc.cpp"
          $2689 = ($2688<<24>>24)==($ch$0<<24>>24); //@line 4783 "load_abc.cpp"
          if (!($2689)) {
           break;
          }
          $2690 = ((($p$18)) + 9|0); //@line 4786 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = 3;$abceffoper$4 = 0;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2690;
          break L678;
         }
        } while(0);
        $lhsc = HEAP8[$p$18>>0]|0; //@line 4789 "load_abc.cpp"
        $2691 = ($lhsc<<24>>24)==(62); //@line 4789 "load_abc.cpp"
        do {
         if ($2691) {
          $2692 = ((($p$18)) + 1|0); //@line 4789 "load_abc.cpp"
          $2693 = HEAP8[$2692>>0]|0; //@line 4789 "load_abc.cpp"
          $2694 = ($2693<<24>>24)==($ch$0<<24>>24); //@line 4789 "load_abc.cpp"
          if (!($2694)) {
           break;
          }
          $2695 = ((($p$18)) + 2|0); //@line 4792 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = 3;$abceffoper$4 = 0;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2695;
          break L678;
         }
        } while(0);
        $2696 = (_strncmp($p$18,527510,7)|0); //@line 4795 "load_abc.cpp"
        $2697 = ($2696|0)==(0); //@line 4795 "load_abc.cpp"
        do {
         if ($2697) {
          $2698 = ((($p$18)) + 7|0); //@line 4795 "load_abc.cpp"
          $2699 = HEAP8[$2698>>0]|0; //@line 4795 "load_abc.cpp"
          $2700 = ($2699<<24>>24)==($ch$0<<24>>24); //@line 4795 "load_abc.cpp"
          if (!($2700)) {
           break;
          }
          $2701 = ((($p$18)) + 8|0); //@line 4797 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = 1;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2701;
          break L678;
         }
        } while(0);
        $2702 = (_strncmp($p$18,527518,4)|0); //@line 4800 "load_abc.cpp"
        $2703 = ($2702|0)==(0); //@line 4800 "load_abc.cpp"
        do {
         if ($2703) {
          $2704 = ((($p$18)) + 4|0); //@line 4800 "load_abc.cpp"
          $2705 = HEAP8[$2704>>0]|0; //@line 4800 "load_abc.cpp"
          $2706 = ($2705<<24>>24)==($ch$0<<24>>24); //@line 4800 "load_abc.cpp"
          if (!($2706)) {
           break;
          }
          $2707 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4801 "load_abc.cpp"
          $2708 = ($2707|0)==(0|0); //@line 915 "load_abc.cpp"
          if ($2708) {
           $2709 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
           $2710 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
           $2711 = ((($2709)) + 20|0); //@line 917 "load_abc.cpp"
           HEAP8[$2711>>0]=$2710&255;HEAP8[$2711+1>>0]=($2710>>8)&255;HEAP8[$2711+2>>0]=($2710>>16)&255;HEAP8[$2711+3>>0]=$2710>>24; //@line 917 "load_abc.cpp"
           $$0$i537 = $2709;
          } else {
           $$0$i537 = $2707;
          }
          HEAP8[$41>>0]=$$0$i537&255;HEAP8[$41+1>>0]=($$0$i537>>8)&255;HEAP8[$41+2>>0]=($$0$i537>>16)&255;HEAP8[$41+3>>0]=$$0$i537>>24; //@line 4801 "load_abc.cpp"
          $2712 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4802 "load_abc.cpp"
          $2713 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$2713>>0]=0&255;HEAP8[$2713+1>>0]=(0>>8)&255;HEAP8[$2713+2>>0]=(0>>16)&255;HEAP8[$2713+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $2714 = ((($2713)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$2714>>0]=$2712&255;HEAP8[$2714+1>>0]=($2712>>8)&255;HEAP8[$2714+2>>0]=($2712>>16)&255;HEAP8[$2714+3>>0]=$2712>>24; //@line 541 "load_abc.cpp"
          $2715 = ((($2713)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$2715>>0] = 1; //@line 543 "load_abc.cpp"
          $2716 = ((($2713)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$2716>>0] = 80; //@line 543 "load_abc.cpp"
          $2717 = ((($2713)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$2717>>0]=0&255;HEAP8[$2717+1>>0]=(0>>8)&255;HEAP8[$2717+2>>0]=(0>>16)&255;HEAP8[$2717+3>>0]=0>>24; //@line 543 "load_abc.cpp"
          $2718 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $2719 = ((($2713)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$2719>>0] = $2718; //@line 544 "load_abc.cpp"
          $2720 = ((($2713)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$2720>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($$0$i537,$2713); //@line 1153 "load_abc.cpp"
          $2721 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4803 "load_abc.cpp"
          $2722 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4803 "load_abc.cpp"
          $2723 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$2723>>0]=0&255;HEAP8[$2723+1>>0]=(0>>8)&255;HEAP8[$2723+2>>0]=(0>>16)&255;HEAP8[$2723+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $2724 = ((($2723)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$2724>>0]=$2722&255;HEAP8[$2724+1>>0]=($2722>>8)&255;HEAP8[$2724+2>>0]=($2722>>16)&255;HEAP8[$2724+3>>0]=$2722>>24; //@line 541 "load_abc.cpp"
          $2725 = ((($2723)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$2725>>0] = 1; //@line 543 "load_abc.cpp"
          $2726 = ((($2723)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$2726>>0] = 70; //@line 543 "load_abc.cpp"
          $2727 = ((($2723)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$2727>>0]=0&255;HEAP8[$2727+1>>0]=(0>>8)&255;HEAP8[$2727+2>>0]=(0>>16)&255;HEAP8[$2727+3>>0]=0>>24; //@line 543 "load_abc.cpp"
          $2728 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $2729 = ((($2723)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$2729>>0] = $2728; //@line 544 "load_abc.cpp"
          $2730 = ((($2723)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$2730>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($2721,$2723); //@line 964 "load_abc.cpp"
          $2731 = ((($p$18)) + 5|0); //@line 4804 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2731;
          break L678;
         }
        } while(0);
        $2732 = (_strncmp($p$18,527523,5)|0); //@line 4807 "load_abc.cpp"
        $2733 = ($2732|0)==(0); //@line 4807 "load_abc.cpp"
        do {
         if ($2733) {
          $2734 = ((($p$18)) + 5|0); //@line 4807 "load_abc.cpp"
          $2735 = HEAP8[$2734>>0]|0; //@line 4807 "load_abc.cpp"
          $2736 = ($2735<<24>>24)==($ch$0<<24>>24); //@line 4807 "load_abc.cpp"
          if (!($2736)) {
           break;
          }
          $2737 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4808 "load_abc.cpp"
          $2738 = ($2737|0)==(0|0); //@line 915 "load_abc.cpp"
          if ($2738) {
           $2739 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
           $2740 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
           $2741 = ((($2739)) + 20|0); //@line 917 "load_abc.cpp"
           HEAP8[$2741>>0]=$2740&255;HEAP8[$2741+1>>0]=($2740>>8)&255;HEAP8[$2741+2>>0]=($2740>>16)&255;HEAP8[$2741+3>>0]=$2740>>24; //@line 917 "load_abc.cpp"
           $$0$i539 = $2739;
          } else {
           $$0$i539 = $2737;
          }
          HEAP8[$41>>0]=$$0$i539&255;HEAP8[$41+1>>0]=($$0$i539>>8)&255;HEAP8[$41+2>>0]=($$0$i539>>16)&255;HEAP8[$41+3>>0]=$$0$i539>>24; //@line 4808 "load_abc.cpp"
          $2742 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4809 "load_abc.cpp"
          $2743 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$2743>>0]=0&255;HEAP8[$2743+1>>0]=(0>>8)&255;HEAP8[$2743+2>>0]=(0>>16)&255;HEAP8[$2743+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $2744 = ((($2743)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$2744>>0]=$2742&255;HEAP8[$2744+1>>0]=($2742>>8)&255;HEAP8[$2744+2>>0]=($2742>>16)&255;HEAP8[$2744+3>>0]=$2742>>24; //@line 541 "load_abc.cpp"
          $2745 = ((($2743)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$2745>>0] = 1; //@line 543 "load_abc.cpp"
          $2746 = ((($2743)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$2746>>0] = 80; //@line 543 "load_abc.cpp"
          $2747 = ((($2743)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$2747>>0]=0&255;HEAP8[$2747+1>>0]=(0>>8)&255;HEAP8[$2747+2>>0]=(0>>16)&255;HEAP8[$2747+3>>0]=0>>24; //@line 543 "load_abc.cpp"
          $2748 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $2749 = ((($2743)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$2749>>0] = $2748; //@line 544 "load_abc.cpp"
          $2750 = ((($2743)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$2750>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($$0$i539,$2743); //@line 1153 "load_abc.cpp"
          $2751 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4810 "load_abc.cpp"
          $2752 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4810 "load_abc.cpp"
          $2753 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$2753>>0]=0&255;HEAP8[$2753+1>>0]=(0>>8)&255;HEAP8[$2753+2>>0]=(0>>16)&255;HEAP8[$2753+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $2754 = ((($2753)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$2754>>0]=$2752&255;HEAP8[$2754+1>>0]=($2752>>8)&255;HEAP8[$2754+2>>0]=($2752>>16)&255;HEAP8[$2754+3>>0]=$2752>>24; //@line 541 "load_abc.cpp"
          $2755 = ((($2753)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$2755>>0] = 1; //@line 543 "load_abc.cpp"
          $2756 = ((($2753)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$2756>>0] = 36; //@line 543 "load_abc.cpp"
          $2757 = ((($2753)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$2757>>0]=0&255;HEAP8[$2757+1>>0]=(0>>8)&255;HEAP8[$2757+2>>0]=(0>>16)&255;HEAP8[$2757+3>>0]=0>>24; //@line 543 "load_abc.cpp"
          $2758 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $2759 = ((($2753)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$2759>>0] = $2758; //@line 544 "load_abc.cpp"
          $2760 = ((($2753)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$2760>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($2751,$2753); //@line 942 "load_abc.cpp"
          $2761 = ((($p$18)) + 6|0); //@line 4811 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2761;
          break L678;
         }
        } while(0);
        $2762 = (_strncmp($p$18,527529,6)|0); //@line 4814 "load_abc.cpp"
        $2763 = ($2762|0)==(0); //@line 4814 "load_abc.cpp"
        do {
         if ($2763) {
          $2764 = ((($p$18)) + 6|0); //@line 4814 "load_abc.cpp"
          $2765 = HEAP8[$2764>>0]|0; //@line 4814 "load_abc.cpp"
          $2766 = ($2765<<24>>24)==($ch$0<<24>>24); //@line 4814 "load_abc.cpp"
          if (!($2766)) {
           break;
          }
          $2767 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4815 "load_abc.cpp"
          $2768 = ($2767|0)==(0|0); //@line 915 "load_abc.cpp"
          if ($2768) {
           $2769 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
           $2770 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
           $2771 = ((($2769)) + 20|0); //@line 917 "load_abc.cpp"
           HEAP8[$2771>>0]=$2770&255;HEAP8[$2771+1>>0]=($2770>>8)&255;HEAP8[$2771+2>>0]=($2770>>16)&255;HEAP8[$2771+3>>0]=$2770>>24; //@line 917 "load_abc.cpp"
           $$0$i541 = $2769;
          } else {
           $$0$i541 = $2767;
          }
          HEAP8[$41>>0]=$$0$i541&255;HEAP8[$41+1>>0]=($$0$i541>>8)&255;HEAP8[$41+2>>0]=($$0$i541>>16)&255;HEAP8[$41+3>>0]=$$0$i541>>24; //@line 4815 "load_abc.cpp"
          $2772 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4816 "load_abc.cpp"
          $2773 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$2773>>0]=0&255;HEAP8[$2773+1>>0]=(0>>8)&255;HEAP8[$2773+2>>0]=(0>>16)&255;HEAP8[$2773+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $2774 = ((($2773)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$2774>>0]=$2772&255;HEAP8[$2774+1>>0]=($2772>>8)&255;HEAP8[$2774+2>>0]=($2772>>16)&255;HEAP8[$2774+3>>0]=$2772>>24; //@line 541 "load_abc.cpp"
          $2775 = ((($2773)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$2775>>0] = 1; //@line 543 "load_abc.cpp"
          $2776 = ((($2773)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$2776>>0] = 80; //@line 543 "load_abc.cpp"
          $2777 = ((($2773)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$2777>>0]=0&255;HEAP8[$2777+1>>0]=(0>>8)&255;HEAP8[$2777+2>>0]=(0>>16)&255;HEAP8[$2777+3>>0]=0>>24; //@line 543 "load_abc.cpp"
          $2778 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $2779 = ((($2773)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$2779>>0] = $2778; //@line 544 "load_abc.cpp"
          $2780 = ((($2773)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$2780>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($$0$i541,$2773); //@line 1153 "load_abc.cpp"
          $2781 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4817 "load_abc.cpp"
          $2782 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4817 "load_abc.cpp"
          $2783 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$2783>>0]=0&255;HEAP8[$2783+1>>0]=(0>>8)&255;HEAP8[$2783+2>>0]=(0>>16)&255;HEAP8[$2783+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $2784 = ((($2783)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$2784>>0]=$2782&255;HEAP8[$2784+1>>0]=($2782>>8)&255;HEAP8[$2784+2>>0]=($2782>>16)&255;HEAP8[$2784+3>>0]=$2782>>24; //@line 541 "load_abc.cpp"
          $2785 = ((($2783)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$2785>>0] = 1; //@line 543 "load_abc.cpp"
          $2786 = ((($2783)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$2786>>0] = 88; //@line 543 "load_abc.cpp"
          $2787 = ((($2783)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$2787>>0]=0&255;HEAP8[$2787+1>>0]=(0>>8)&255;HEAP8[$2787+2>>0]=(0>>16)&255;HEAP8[$2787+3>>0]=0>>24; //@line 543 "load_abc.cpp"
          $2788 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $2789 = ((($2783)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$2789>>0] = $2788; //@line 544 "load_abc.cpp"
          $2790 = ((($2783)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$2790>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($2781,$2783); //@line 975 "load_abc.cpp"
          $2791 = ((($p$18)) + 7|0); //@line 4818 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2791;
          break L678;
         }
        } while(0);
        $2792 = (_strncmp($p$18,527536,4)|0); //@line 4821 "load_abc.cpp"
        $2793 = ($2792|0)==(0); //@line 4821 "load_abc.cpp"
        do {
         if ($2793) {
          $2794 = ((($p$18)) + 4|0); //@line 4821 "load_abc.cpp"
          $2795 = HEAP8[$2794>>0]|0; //@line 4821 "load_abc.cpp"
          $2796 = ($2795<<24>>24)==($ch$0<<24>>24); //@line 4821 "load_abc.cpp"
          if (!($2796)) {
           break;
          }
          $2797 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4822 "load_abc.cpp"
          $2798 = ($2797|0)==(0|0); //@line 915 "load_abc.cpp"
          if ($2798) {
           $2799 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
           $2800 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
           $2801 = ((($2799)) + 20|0); //@line 917 "load_abc.cpp"
           HEAP8[$2801>>0]=$2800&255;HEAP8[$2801+1>>0]=($2800>>8)&255;HEAP8[$2801+2>>0]=($2800>>16)&255;HEAP8[$2801+3>>0]=$2800>>24; //@line 917 "load_abc.cpp"
           $$0$i543 = $2799;
          } else {
           $$0$i543 = $2797;
          }
          HEAP8[$41>>0]=$$0$i543&255;HEAP8[$41+1>>0]=($$0$i543>>8)&255;HEAP8[$41+2>>0]=($$0$i543>>16)&255;HEAP8[$41+3>>0]=$$0$i543>>24; //@line 4822 "load_abc.cpp"
          $2802 = ((($$0$i543)) + 32|0); //@line 4823 "load_abc.cpp"
          $2803 = HEAPU8[$2802>>0]|(HEAPU8[$2802+1>>0]<<8)|(HEAPU8[$2802+2>>0]<<16)|(HEAPU8[$2802+3>>0]<<24); //@line 4823 "load_abc.cpp"
          $2804 = ($2803|0)==(-2); //@line 4823 "load_abc.cpp"
          $2805 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4824 "load_abc.cpp"
          $2806 = $2804 ? 3 : 2; //@line 1581 "load_abc.cpp"
          $2807 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$2807>>0]=0&255;HEAP8[$2807+1>>0]=(0>>8)&255;HEAP8[$2807+2>>0]=(0>>16)&255;HEAP8[$2807+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $2808 = ((($2807)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$2808>>0]=$2805&255;HEAP8[$2808+1>>0]=($2805>>8)&255;HEAP8[$2808+2>>0]=($2805>>16)&255;HEAP8[$2808+3>>0]=$2805>>24; //@line 541 "load_abc.cpp"
          $2809 = ((($2807)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$2809>>0] = 1; //@line 543 "load_abc.cpp"
          $2810 = ((($2807)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$2810>>0] = 74; //@line 543 "load_abc.cpp"
          $2811 = ((($2807)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$2811>>0] = 0; //@line 543 "load_abc.cpp"
          $2812 = ((($2807)) + 11|0); //@line 543 "load_abc.cpp"
          HEAP8[$2812>>0] = 0; //@line 543 "load_abc.cpp"
          $2813 = ((($2807)) + 12|0); //@line 543 "load_abc.cpp"
          HEAP8[$2813>>0] = 0; //@line 543 "load_abc.cpp"
          $2814 = ((($2807)) + 13|0); //@line 543 "load_abc.cpp"
          HEAP8[$2814>>0] = $2806; //@line 543 "load_abc.cpp"
          $2815 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $2816 = ((($2807)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$2816>>0] = $2815; //@line 544 "load_abc.cpp"
          $2817 = ((($2807)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$2817>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($$0$i543,$2807); //@line 1583 "load_abc.cpp"
          $2818 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4825 "load_abc.cpp"
          $2819 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4825 "load_abc.cpp"
          $2820 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$2820>>0]=0&255;HEAP8[$2820+1>>0]=(0>>8)&255;HEAP8[$2820+2>>0]=(0>>16)&255;HEAP8[$2820+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $2821 = ((($2820)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$2821>>0]=$2819&255;HEAP8[$2821+1>>0]=($2819>>8)&255;HEAP8[$2821+2>>0]=($2819>>16)&255;HEAP8[$2821+3>>0]=$2819>>24; //@line 541 "load_abc.cpp"
          $2822 = ((($2820)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$2822>>0] = 1; //@line 543 "load_abc.cpp"
          $2823 = ((($2820)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$2823>>0] = 80; //@line 543 "load_abc.cpp"
          $2824 = ((($2820)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$2824>>0]=0&255;HEAP8[$2824+1>>0]=(0>>8)&255;HEAP8[$2824+2>>0]=(0>>16)&255;HEAP8[$2824+3>>0]=0>>24; //@line 543 "load_abc.cpp"
          $2825 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $2826 = ((($2820)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$2826>>0] = $2825; //@line 544 "load_abc.cpp"
          $2827 = ((($2820)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$2827>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($2818,$2820); //@line 1153 "load_abc.cpp"
          $2828 = ((($p$18)) + 5|0); //@line 4826 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2828;
          break L678;
         }
        } while(0);
        $2829 = (_strncmp($p$18,527541,4)|0); //@line 4829 "load_abc.cpp"
        $2830 = ($2829|0)==(0); //@line 4829 "load_abc.cpp"
        do {
         if ($2830) {
          $2831 = ((($p$18)) + 4|0); //@line 4829 "load_abc.cpp"
          $2832 = HEAP8[$2831>>0]|0; //@line 4829 "load_abc.cpp"
          $2833 = ($2832<<24>>24)==($ch$0<<24>>24); //@line 4829 "load_abc.cpp"
          if (!($2833)) {
           break;
          }
          $2834 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4830 "load_abc.cpp"
          $2835 = ($2834|0)==(0|0); //@line 915 "load_abc.cpp"
          if ($2835) {
           $2836 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
           $2837 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
           $2838 = ((($2836)) + 20|0); //@line 917 "load_abc.cpp"
           HEAP8[$2838>>0]=$2837&255;HEAP8[$2838+1>>0]=($2837>>8)&255;HEAP8[$2838+2>>0]=($2837>>16)&255;HEAP8[$2838+3>>0]=$2837>>24; //@line 917 "load_abc.cpp"
           $$0$i545 = $2836;
          } else {
           $$0$i545 = $2834;
          }
          HEAP8[$41>>0]=$$0$i545&255;HEAP8[$41+1>>0]=($$0$i545>>8)&255;HEAP8[$41+2>>0]=($$0$i545>>16)&255;HEAP8[$41+3>>0]=$$0$i545>>24; //@line 4830 "load_abc.cpp"
          $2839 = ((($$0$i545)) + 32|0); //@line 4831 "load_abc.cpp"
          $2840 = HEAPU8[$2839>>0]|(HEAPU8[$2839+1>>0]<<8)|(HEAPU8[$2839+2>>0]<<16)|(HEAPU8[$2839+3>>0]<<24); //@line 4831 "load_abc.cpp"
          $2841 = ($2840|0)==(-2); //@line 4831 "load_abc.cpp"
          $2842 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4832 "load_abc.cpp"
          $2843 = $2841 ? 5 : 4; //@line 1581 "load_abc.cpp"
          $2844 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$2844>>0]=0&255;HEAP8[$2844+1>>0]=(0>>8)&255;HEAP8[$2844+2>>0]=(0>>16)&255;HEAP8[$2844+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $2845 = ((($2844)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$2845>>0]=$2842&255;HEAP8[$2845+1>>0]=($2842>>8)&255;HEAP8[$2845+2>>0]=($2842>>16)&255;HEAP8[$2845+3>>0]=$2842>>24; //@line 541 "load_abc.cpp"
          $2846 = ((($2844)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$2846>>0] = 1; //@line 543 "load_abc.cpp"
          $2847 = ((($2844)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$2847>>0] = 74; //@line 543 "load_abc.cpp"
          $2848 = ((($2844)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$2848>>0] = 0; //@line 543 "load_abc.cpp"
          $2849 = ((($2844)) + 11|0); //@line 543 "load_abc.cpp"
          HEAP8[$2849>>0] = 0; //@line 543 "load_abc.cpp"
          $2850 = ((($2844)) + 12|0); //@line 543 "load_abc.cpp"
          HEAP8[$2850>>0] = 0; //@line 543 "load_abc.cpp"
          $2851 = ((($2844)) + 13|0); //@line 543 "load_abc.cpp"
          HEAP8[$2851>>0] = $2843; //@line 543 "load_abc.cpp"
          $2852 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $2853 = ((($2844)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$2853>>0] = $2852; //@line 544 "load_abc.cpp"
          $2854 = ((($2844)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$2854>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($$0$i545,$2844); //@line 1583 "load_abc.cpp"
          $2855 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4833 "load_abc.cpp"
          $2856 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4833 "load_abc.cpp"
          $2857 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$2857>>0]=0&255;HEAP8[$2857+1>>0]=(0>>8)&255;HEAP8[$2857+2>>0]=(0>>16)&255;HEAP8[$2857+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $2858 = ((($2857)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$2858>>0]=$2856&255;HEAP8[$2858+1>>0]=($2856>>8)&255;HEAP8[$2858+2>>0]=($2856>>16)&255;HEAP8[$2858+3>>0]=$2856>>24; //@line 541 "load_abc.cpp"
          $2859 = ((($2857)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$2859>>0] = 1; //@line 543 "load_abc.cpp"
          $2860 = ((($2857)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$2860>>0] = 80; //@line 543 "load_abc.cpp"
          $2861 = ((($2857)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$2861>>0]=0&255;HEAP8[$2861+1>>0]=(0>>8)&255;HEAP8[$2861+2>>0]=(0>>16)&255;HEAP8[$2861+3>>0]=0>>24; //@line 543 "load_abc.cpp"
          $2862 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $2863 = ((($2857)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$2863>>0] = $2862; //@line 544 "load_abc.cpp"
          $2864 = ((($2857)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$2864>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($2855,$2857); //@line 1153 "load_abc.cpp"
          $2865 = ((($p$18)) + 5|0); //@line 4834 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2865;
          break L678;
         }
        } while(0);
        $2866 = (_strncmp($p$18,527546,6)|0); //@line 4837 "load_abc.cpp"
        $2867 = ($2866|0)==(0); //@line 4837 "load_abc.cpp"
        do {
         if ($2867) {
          $2868 = ((($p$18)) + 6|0); //@line 4837 "load_abc.cpp"
          $2869 = HEAP8[$2868>>0]|0; //@line 4837 "load_abc.cpp"
          $2870 = ($2869<<24>>24)==($ch$0<<24>>24); //@line 4837 "load_abc.cpp"
          if (!($2870)) {
           break;
          }
          $2871 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4838 "load_abc.cpp"
          $2872 = ($2871|0)==(0|0); //@line 915 "load_abc.cpp"
          if ($2872) {
           $2873 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
           $2874 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
           $2875 = ((($2873)) + 20|0); //@line 917 "load_abc.cpp"
           HEAP8[$2875>>0]=$2874&255;HEAP8[$2875+1>>0]=($2874>>8)&255;HEAP8[$2875+2>>0]=($2874>>16)&255;HEAP8[$2875+3>>0]=$2874>>24; //@line 917 "load_abc.cpp"
           $$0$i547 = $2873;
          } else {
           $$0$i547 = $2871;
          }
          HEAP8[$41>>0]=$$0$i547&255;HEAP8[$41+1>>0]=($$0$i547>>8)&255;HEAP8[$41+2>>0]=($$0$i547>>16)&255;HEAP8[$41+3>>0]=$$0$i547>>24; //@line 4838 "load_abc.cpp"
          $2876 = ((($$0$i547)) + 32|0); //@line 4839 "load_abc.cpp"
          $2877 = HEAPU8[$2876>>0]|(HEAPU8[$2876+1>>0]<<8)|(HEAPU8[$2876+2>>0]<<16)|(HEAPU8[$2876+3>>0]<<24); //@line 4839 "load_abc.cpp"
          $2878 = ($2877|0)==(-2); //@line 4839 "load_abc.cpp"
          $2879 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4840 "load_abc.cpp"
          $2880 = $2878 ? 3 : 2; //@line 1581 "load_abc.cpp"
          $2881 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$2881>>0]=0&255;HEAP8[$2881+1>>0]=(0>>8)&255;HEAP8[$2881+2>>0]=(0>>16)&255;HEAP8[$2881+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $2882 = ((($2881)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$2882>>0]=$2879&255;HEAP8[$2882+1>>0]=($2879>>8)&255;HEAP8[$2882+2>>0]=($2879>>16)&255;HEAP8[$2882+3>>0]=$2879>>24; //@line 541 "load_abc.cpp"
          $2883 = ((($2881)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$2883>>0] = 1; //@line 543 "load_abc.cpp"
          $2884 = ((($2881)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$2884>>0] = 74; //@line 543 "load_abc.cpp"
          $2885 = ((($2881)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$2885>>0] = 0; //@line 543 "load_abc.cpp"
          $2886 = ((($2881)) + 11|0); //@line 543 "load_abc.cpp"
          HEAP8[$2886>>0] = 0; //@line 543 "load_abc.cpp"
          $2887 = ((($2881)) + 12|0); //@line 543 "load_abc.cpp"
          HEAP8[$2887>>0] = 0; //@line 543 "load_abc.cpp"
          $2888 = ((($2881)) + 13|0); //@line 543 "load_abc.cpp"
          HEAP8[$2888>>0] = $2880; //@line 543 "load_abc.cpp"
          $2889 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $2890 = ((($2881)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$2890>>0] = $2889; //@line 544 "load_abc.cpp"
          $2891 = ((($2881)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$2891>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($$0$i547,$2881); //@line 1583 "load_abc.cpp"
          $2892 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4841 "load_abc.cpp"
          $2893 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4841 "load_abc.cpp"
          $2894 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$2894>>0]=0&255;HEAP8[$2894+1>>0]=(0>>8)&255;HEAP8[$2894+2>>0]=(0>>16)&255;HEAP8[$2894+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $2895 = ((($2894)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$2895>>0]=$2893&255;HEAP8[$2895+1>>0]=($2893>>8)&255;HEAP8[$2895+2>>0]=($2893>>16)&255;HEAP8[$2895+3>>0]=$2893>>24; //@line 541 "load_abc.cpp"
          $2896 = ((($2894)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$2896>>0] = 1; //@line 543 "load_abc.cpp"
          $2897 = ((($2894)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$2897>>0] = 80; //@line 543 "load_abc.cpp"
          $2898 = ((($2894)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$2898>>0]=0&255;HEAP8[$2898+1>>0]=(0>>8)&255;HEAP8[$2898+2>>0]=(0>>16)&255;HEAP8[$2898+3>>0]=0>>24; //@line 543 "load_abc.cpp"
          $2899 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $2900 = ((($2894)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$2900>>0] = $2899; //@line 544 "load_abc.cpp"
          $2901 = ((($2894)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$2901>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($2892,$2894); //@line 1153 "load_abc.cpp"
          $2902 = ((($p$18)) + 7|0); //@line 4842 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2902;
          break L678;
         }
        } while(0);
        $2903 = (_strncmp($p$18,527553,6)|0); //@line 4845 "load_abc.cpp"
        $2904 = ($2903|0)==(0); //@line 4845 "load_abc.cpp"
        do {
         if ($2904) {
          $2905 = ((($p$18)) + 6|0); //@line 4845 "load_abc.cpp"
          $2906 = HEAP8[$2905>>0]|0; //@line 4845 "load_abc.cpp"
          $2907 = ($2906<<24>>24)==($ch$0<<24>>24); //@line 4845 "load_abc.cpp"
          if (!($2907)) {
           break;
          }
          $2908 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4846 "load_abc.cpp"
          $2909 = ($2908|0)==(0|0); //@line 915 "load_abc.cpp"
          if ($2909) {
           $2910 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
           $2911 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
           $2912 = ((($2910)) + 20|0); //@line 917 "load_abc.cpp"
           HEAP8[$2912>>0]=$2911&255;HEAP8[$2912+1>>0]=($2911>>8)&255;HEAP8[$2912+2>>0]=($2911>>16)&255;HEAP8[$2912+3>>0]=$2911>>24; //@line 917 "load_abc.cpp"
           $$0$i549 = $2910;
          } else {
           $$0$i549 = $2908;
          }
          HEAP8[$41>>0]=$$0$i549&255;HEAP8[$41+1>>0]=($$0$i549>>8)&255;HEAP8[$41+2>>0]=($$0$i549>>16)&255;HEAP8[$41+3>>0]=$$0$i549>>24; //@line 4846 "load_abc.cpp"
          $2913 = ((($$0$i549)) + 32|0); //@line 4847 "load_abc.cpp"
          $2914 = HEAPU8[$2913>>0]|(HEAPU8[$2913+1>>0]<<8)|(HEAPU8[$2913+2>>0]<<16)|(HEAPU8[$2913+3>>0]<<24); //@line 4847 "load_abc.cpp"
          $2915 = ($2914|0)==(-2); //@line 4847 "load_abc.cpp"
          $2916 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4848 "load_abc.cpp"
          $2917 = $2915 ? 3 : 2; //@line 1581 "load_abc.cpp"
          $2918 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$2918>>0]=0&255;HEAP8[$2918+1>>0]=(0>>8)&255;HEAP8[$2918+2>>0]=(0>>16)&255;HEAP8[$2918+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $2919 = ((($2918)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$2919>>0]=$2916&255;HEAP8[$2919+1>>0]=($2916>>8)&255;HEAP8[$2919+2>>0]=($2916>>16)&255;HEAP8[$2919+3>>0]=$2916>>24; //@line 541 "load_abc.cpp"
          $2920 = ((($2918)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$2920>>0] = 1; //@line 543 "load_abc.cpp"
          $2921 = ((($2918)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$2921>>0] = 74; //@line 543 "load_abc.cpp"
          $2922 = ((($2918)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$2922>>0] = 0; //@line 543 "load_abc.cpp"
          $2923 = ((($2918)) + 11|0); //@line 543 "load_abc.cpp"
          HEAP8[$2923>>0] = 0; //@line 543 "load_abc.cpp"
          $2924 = ((($2918)) + 12|0); //@line 543 "load_abc.cpp"
          HEAP8[$2924>>0] = 0; //@line 543 "load_abc.cpp"
          $2925 = ((($2918)) + 13|0); //@line 543 "load_abc.cpp"
          HEAP8[$2925>>0] = $2917; //@line 543 "load_abc.cpp"
          $2926 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $2927 = ((($2918)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$2927>>0] = $2926; //@line 544 "load_abc.cpp"
          $2928 = ((($2918)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$2928>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($$0$i549,$2918); //@line 1583 "load_abc.cpp"
          $2929 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4849 "load_abc.cpp"
          $2930 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4849 "load_abc.cpp"
          $2931 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$2931>>0]=0&255;HEAP8[$2931+1>>0]=(0>>8)&255;HEAP8[$2931+2>>0]=(0>>16)&255;HEAP8[$2931+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $2932 = ((($2931)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$2932>>0]=$2930&255;HEAP8[$2932+1>>0]=($2930>>8)&255;HEAP8[$2932+2>>0]=($2930>>16)&255;HEAP8[$2932+3>>0]=$2930>>24; //@line 541 "load_abc.cpp"
          $2933 = ((($2931)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$2933>>0] = 1; //@line 543 "load_abc.cpp"
          $2934 = ((($2931)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$2934>>0] = 80; //@line 543 "load_abc.cpp"
          $2935 = ((($2931)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$2935>>0]=0&255;HEAP8[$2935+1>>0]=(0>>8)&255;HEAP8[$2935+2>>0]=(0>>16)&255;HEAP8[$2935+3>>0]=0>>24; //@line 543 "load_abc.cpp"
          $2936 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $2937 = ((($2931)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$2937>>0] = $2936; //@line 544 "load_abc.cpp"
          $2938 = ((($2931)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$2938>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($2929,$2931); //@line 1153 "load_abc.cpp"
          $2939 = ((($p$18)) + 7|0); //@line 4850 "load_abc.cpp"
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2939;
          break L678;
         }
        } while(0);
        $2940 = ($ch$0<<24>>24)==(33); //@line 4853 "load_abc.cpp"
        if (!($2940)) {
         $2960 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4862 "load_abc.cpp"
         $2961 = ($2960|0)==(0|0); //@line 915 "load_abc.cpp"
         if ($2961) {
          $2962 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
          $2963 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
          $2964 = ((($2962)) + 20|0); //@line 917 "load_abc.cpp"
          HEAP8[$2964>>0]=$2963&255;HEAP8[$2964+1>>0]=($2963>>8)&255;HEAP8[$2964+2>>0]=($2963>>16)&255;HEAP8[$2964+3>>0]=$2963>>24; //@line 917 "load_abc.cpp"
          $$0$i553 = $2962;
         } else {
          $$0$i553 = $2960;
         }
         HEAP8[$41>>0]=$$0$i553&255;HEAP8[$41+1>>0]=($$0$i553>>8)&255;HEAP8[$41+2>>0]=($$0$i553>>16)&255;HEAP8[$41+3>>0]=$$0$i553>>24; //@line 4862 "load_abc.cpp"
         $2965 = (__ZL21abc_parse_decorationsP10_ABCHANDLEP9_ABCTRACKPKc($1,$$0$i553,$p$18)|0); //@line 4863 "load_abc.cpp"
         $p$45 = $p$18;
         while(1) {
          $2966 = ((($p$45)) + 1|0); //@line 4864 "load_abc.cpp"
          $2967 = HEAP8[$p$45>>0]|0; //@line 4864 "load_abc.cpp"
          switch ($2967<<24>>24) {
          case 43: case 0:  {
           $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $2965;$bartime$5 = $bartime$41235;$ch$1 = $2967;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2966;
           break L678;
           break;
          }
          default: {
           $p$45 = $2966;
          }
          }
         }
        }
        $2941 = ($lhsc<<24>>24)==(0); //@line 4854 "load_abc.cpp"
        if ($2941) {
         $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = 33;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$18;
         break L678;
        } else {
         $2943 = $lhsc;$4064 = $p$18;$t$121185 = 0;
        }
        while(1) {
         $2942 = $2943 << 24 >> 24; //@line 4854 "load_abc.cpp"
         $memchr = (_memchr(527560,$2942,6)|0); //@line 4854 "load_abc.cpp"
         $2944 = ($memchr|0)==(0|0); //@line 4854 "load_abc.cpp"
         if (!($2944)) {
          $$pr774 = $2943;$t$1211852192 = $t$121185;
          break;
         }
         $2945 = (_isspace($2942)|0); //@line 4854 "load_abc.cpp"
         $2946 = ($2945|0)==(0); //@line 4854 "load_abc.cpp"
         if (!($2946)) {
          $$lcssa2188 = $4064;$t$121185$lcssa2190 = $t$121185;
          label = 967;
          break;
         }
         $2947 = (($t$121185) + 1)|0; //@line 4854 "load_abc.cpp"
         $2948 = (($p$18) + ($2947)|0); //@line 4854 "load_abc.cpp"
         $2949 = HEAP8[$2948>>0]|0; //@line 4854 "load_abc.cpp"
         $2950 = ($2949<<24>>24)==(0); //@line 4854 "load_abc.cpp"
         if ($2950) {
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = 33;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$18;
          break L678;
         } else {
          $2943 = $2949;$4064 = $2948;$t$121185 = $2947;
         }
        }
        if ((label|0) == 967) {
         label = 0;
         $$pr774$pre = HEAP8[$$lcssa2188>>0]|0; //@line 4855 "load_abc.cpp"
         $$pr774 = $$pr774$pre;$t$1211852192 = $t$121185$lcssa2190;
        }
        $2951 = ($$pr774<<24>>24)==(33); //@line 4855 "load_abc.cpp"
        if (!($2951)) {
         $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = 33;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$18;
         break L678;
        }
        $2952 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4856 "load_abc.cpp"
        $2953 = ($2952|0)==(0|0); //@line 915 "load_abc.cpp"
        if ($2953) {
         $2954 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
         $2955 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
         $2956 = ((($2954)) + 20|0); //@line 917 "load_abc.cpp"
         HEAP8[$2956>>0]=$2955&255;HEAP8[$2956+1>>0]=($2955>>8)&255;HEAP8[$2956+2>>0]=($2955>>16)&255;HEAP8[$2956+3>>0]=$2955>>24; //@line 917 "load_abc.cpp"
         $$0$i551 = $2954;
        } else {
         $$0$i551 = $2952;
        }
        HEAP8[$41>>0]=$$0$i551&255;HEAP8[$41+1>>0]=($$0$i551>>8)&255;HEAP8[$41+2>>0]=($$0$i551>>16)&255;HEAP8[$41+3>>0]=$$0$i551>>24; //@line 4856 "load_abc.cpp"
        $2957 = (__ZL21abc_parse_decorationsP10_ABCHANDLEP9_ABCTRACKPKc($1,$$0$i551,$p$18)|0); //@line 4857 "load_abc.cpp"
        $2958 = (($t$1211852192) + 1)|0; //@line 4858 "load_abc.cpp"
        $2959 = (($p$18) + ($2958)|0); //@line 4858 "load_abc.cpp"
        $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = 0;$abcvol$4 = $2957;$bartime$5 = $bartime$41235;$ch$1 = 33;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $2959;
        break;
       }
       case 46: case 96:  {
        $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = $abcto$31230;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$18;
        break;
       }
       default: {
        $memchr66 = (_memchr(527576,$1185,19)|0); //@line 4873 "load_abc.cpp"
        $2968 = ($memchr66|0)==(0|0); //@line 4873 "load_abc.cpp"
        if (!($2968)) {
         $2969 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4874 "load_abc.cpp"
         $2970 = ($2969|0)==(0|0); //@line 915 "load_abc.cpp"
         if ($2970) {
          $2971 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
          $2972 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
          $2973 = ((($2971)) + 20|0); //@line 917 "load_abc.cpp"
          HEAP8[$2973>>0]=$2972&255;HEAP8[$2973+1>>0]=($2972>>8)&255;HEAP8[$2973+2>>0]=($2972>>16)&255;HEAP8[$2973+3>>0]=$2972>>24; //@line 917 "load_abc.cpp"
          $$0$i555 = $2971;
         } else {
          $$0$i555 = $2969;
         }
         HEAP8[$41>>0]=$$0$i555&255;HEAP8[$41+1>>0]=($$0$i555>>8)&255;HEAP8[$41+2>>0]=($$0$i555>>16)&255;HEAP8[$41+3>>0]=$$0$i555>>24; //@line 4874 "load_abc.cpp"
         $2974 = ((($$0$i555)) + 36|0); //@line 1227 "load_abc.cpp"
         $2975 = HEAP8[$2974>>0]|0; //@line 1227 "load_abc.cpp"
         $tp$01$i557 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 1228 "load_abc.cpp"
         $2976 = ($tp$01$i557|0)==(0|0); //@line 1228 "load_abc.cpp"
         if (!($2976)) {
          $tp$02$i558 = $tp$01$i557;
          while(1) {
           $2977 = ((($tp$02$i558)) + 36|0); //@line 1229 "load_abc.cpp"
           $2978 = HEAP8[$2977>>0]|0; //@line 1229 "load_abc.cpp"
           $2979 = ($2978<<24>>24)==($2975<<24>>24); //@line 1229 "load_abc.cpp"
           if ($2979) {
            $2980 = ((($tp$02$i558)) + 37|0); //@line 1229 "load_abc.cpp"
            $2981 = HEAP8[$2980>>0]|0; //@line 1229 "load_abc.cpp"
            $2982 = ((($tp$02$i558)) + 38|0); //@line 1229 "load_abc.cpp"
            HEAP8[$2982>>0] = $2981; //@line 1229 "load_abc.cpp"
           }
           $tp$0$i560 = HEAPU8[$tp$02$i558>>0]|(HEAPU8[$tp$02$i558+1>>0]<<8)|(HEAPU8[$tp$02$i558+2>>0]<<16)|(HEAPU8[$tp$02$i558+3>>0]<<24); //@line 1228 "load_abc.cpp"
           $2983 = ($tp$0$i560|0)==(0|0); //@line 1228 "load_abc.cpp"
           if ($2983) {
            break;
           } else {
            $tp$02$i558 = $tp$0$i560;
           }
          }
         }
         $2984 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4876 "load_abc.cpp"
         $2985 = (($2984) - ($bartime$41235))|0; //@line 4876 "load_abc.cpp"
         $2986 = HEAP32[$mnotediv>>2]|0; //@line 4876 "load_abc.cpp"
         $2987 = ($2986|0)==(0); //@line 345 "load_abc.cpp"
         if ($2987) {
          $$0$i563 = 0;
         } else {
          $2988 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4876 "load_abc.cpp"
          $2989 = ($2988*3072)|0; //@line 346 "load_abc.cpp"
          $2990 = HEAP32[299862]|0; //@line 346 "load_abc.cpp"
          $2991 = Math_imul($2989, $2990)|0; //@line 346 "load_abc.cpp"
          $2992 = HEAP32[299863]|0; //@line 346 "load_abc.cpp"
          $2993 = Math_imul($2992, $2986)|0; //@line 346 "load_abc.cpp"
          $2994 = (($2991|0) / ($2993|0))&-1; //@line 346 "load_abc.cpp"
          $$0$i563 = $2994;
         }
         $2995 = (($2985>>>0) / ($$0$i563>>>0))&-1; //@line 4876 "load_abc.cpp"
         $2996 = HEAPU8[$3>>0]|(HEAPU8[$3+1>>0]<<8)|(HEAPU8[$3+2>>0]<<16)|(HEAPU8[$3+3>>0]<<24); //@line 3546 "load_abc.cpp"
         $2997 = ($2996|0)==(0|0); //@line 3546 "load_abc.cpp"
         L695: do {
          if ($2997) {
           $3011 = HEAP8[$105>>0]|0; //@line 3565 "load_abc.cpp"
           $3012 = $3011&255; //@line 3565 "load_abc.cpp"
           $3013 = (($2995|0) % ($3012|0))&-1; //@line 3565 "load_abc.cpp"
           $3014 = ($3013|0)==(0); //@line 3565 "load_abc.cpp"
           if (!($3014)) {
            $3020 = HEAP8[$104>>0]|0; //@line 3572 "load_abc.cpp"
            $3021 = $3020&255; //@line 3572 "load_abc.cpp"
            $vol$0$i565 = $3021;
            break;
           }
           $3015 = ($2995|0)==(0); //@line 3566 "load_abc.cpp"
           if ($3015) {
            $3018 = HEAP8[$102>>0]|0; //@line 3569 "load_abc.cpp"
            $3019 = $3018&255; //@line 3569 "load_abc.cpp"
            $vol$0$i565 = $3019;
            break;
           } else {
            $3016 = HEAP8[$103>>0]|0; //@line 3567 "load_abc.cpp"
            $3017 = $3016&255; //@line 3567 "load_abc.cpp"
            $vol$0$i565 = $3017;
            break;
           }
          } else {
           $2998 = HEAP8[$104>>0]|0; //@line 3547 "load_abc.cpp"
           $2999 = $2998&255; //@line 3547 "load_abc.cpp"
           $3000 = ($2999*9)|0; //@line 3547 "load_abc.cpp"
           $3001 = (($3000>>>0) / 10)&-1; //@line 3547 "load_abc.cpp"
           $3002 = (_strlen($2996)|0); //@line 3548 "load_abc.cpp"
           $3003 = ($3002|0)>($2995|0); //@line 3548 "load_abc.cpp"
           if (!($3003)) {
            $vol$0$i565 = $3001;
            break;
           }
           $3004 = (($2996) + ($2995)|0); //@line 3549 "load_abc.cpp"
           $3005 = HEAP8[$3004>>0]|0; //@line 3549 "load_abc.cpp"
           $3006 = $3005 << 24 >> 24; //@line 3549 "load_abc.cpp"
           switch ($3006|0) {
           case 102:  {
            $3007 = HEAP8[$102>>0]|0; //@line 3551 "load_abc.cpp"
            $3008 = $3007&255; //@line 3551 "load_abc.cpp"
            $vol$0$i565 = $3008;
            break L695;
            break;
           }
           case 109:  {
            $3009 = HEAP8[$103>>0]|0; //@line 3554 "load_abc.cpp"
            $3010 = $3009&255; //@line 3554 "load_abc.cpp"
            $vol$0$i565 = $3010;
            break L695;
            break;
           }
           case 112:  {
            $vol$0$i565 = $2999;
            break L695;
            break;
           }
           default: {
            $vol$0$i565 = $3001;
            break L695;
           }
           }
          }
         } while(0);
         $3022 = Math_imul($vol$0$i565, $abcvol$31225)|0; //@line 3574 "load_abc.cpp"
         $3023 = (($3022|0) / 128)&-1; //@line 3575 "load_abc.cpp"
         $3024 = (__ZL14abc_add_noteonP10_ABCHANDLEiPKcjPci17ABCEVENT_X_EFFECTi($1,$1185,$p$18,$2984,$barsig,$3023,$abceffect$31222,$abceffoper$31223)|0); //@line 4877 "load_abc.cpp"
         $3025 = (($p$18) + ($3024)|0); //@line 4877 "load_abc.cpp"
         $3026 = ($abceffoper$31223|0)==(255); //@line 4878 "load_abc.cpp"
         $abceffect$3$91 = $3026 ? $abceffect$31222 : 0; //@line 4878 "load_abc.cpp"
         $3027 = HEAP8[$3025>>0]|0; //@line 1625 "load_abc.cpp"
         $3028 = $3027 << 24 >> 24; //@line 1625 "load_abc.cpp"
         $isdigittmp1$i$i567 = (($3028) + -48)|0; //@line 1625 "load_abc.cpp"
         $isdigit2$i$i568 = ($isdigittmp1$i$i567>>>0)<(10); //@line 1625 "load_abc.cpp"
         if ($isdigit2$i$i568) {
          $3031 = $3028;$h$04$i$i569 = 0;$i$03$i$i570 = 0;
          while(1) {
           $3029 = ($h$04$i$i569*10)|0; //@line 1626 "load_abc.cpp"
           $3030 = (($3031) + -48)|0; //@line 1626 "load_abc.cpp"
           $3032 = (($3030) + ($3029))|0; //@line 1626 "load_abc.cpp"
           $3033 = (($i$03$i$i570) + 1)|0; //@line 1627 "load_abc.cpp"
           $$sum827 = (($3033) + ($3024))|0; //@line 1625 "load_abc.cpp"
           $3034 = (($p$18) + ($$sum827)|0); //@line 1625 "load_abc.cpp"
           $3035 = HEAP8[$3034>>0]|0; //@line 1625 "load_abc.cpp"
           $3036 = $3035 << 24 >> 24; //@line 1625 "load_abc.cpp"
           $isdigittmp$i$i571 = (($3036) + -48)|0; //@line 1625 "load_abc.cpp"
           $isdigit$i$i572 = ($isdigittmp$i$i571>>>0)<(10); //@line 1625 "load_abc.cpp"
           if ($isdigit$i$i572) {
            $3031 = $3036;$h$04$i$i569 = $3032;$i$03$i$i570 = $3033;
           } else {
            $h$0$lcssa$i$i574 = $3032;$i$0$lcssa$i$i575 = $3033;
            break;
           }
          }
         } else {
          $h$0$lcssa$i$i574 = 0;$i$0$lcssa$i$i575 = 0;
         }
         $3037 = ($i$0$lcssa$i$i575|0)==(0); //@line 1629 "load_abc.cpp"
         $$h$0$i$i576 = $3037 ? 1 : $h$0$lcssa$i$i574; //@line 1630 "load_abc.cpp"
         $$sum823 = (($i$0$lcssa$i$i575) + ($3024))|0; //@line 1666 "load_abc.cpp"
         $3038 = (($p$18) + ($$sum823)|0); //@line 1666 "load_abc.cpp"
         $3039 = HEAP8[$3038>>0]|0; //@line 1666 "load_abc.cpp"
         $3040 = ($3039<<24>>24)==(47); //@line 1666 "load_abc.cpp"
         if ($3040) {
          $h$017$i578 = 1;$i$016$i579 = $i$0$lcssa$i$i575;
          while(1) {
           $3041 = $h$017$i578 << 1; //@line 1667 "load_abc.cpp"
           $3042 = (($i$016$i579) + 1)|0; //@line 1668 "load_abc.cpp"
           $$sum826 = (($3042) + ($3024))|0; //@line 1666 "load_abc.cpp"
           $3043 = (($p$18) + ($$sum826)|0); //@line 1666 "load_abc.cpp"
           $3044 = HEAP8[$3043>>0]|0; //@line 1666 "load_abc.cpp"
           $3045 = ($3044<<24>>24)==(47); //@line 1666 "load_abc.cpp"
           if ($3045) {
            $h$017$i578 = $3041;$i$016$i579 = $3042;
           } else {
            $$lcssa13$i581 = $3044;$h$0$lcssa$i582 = $3041;$i$0$lcssa$i583 = $3042;
            break;
           }
          }
         } else {
          $$lcssa13$i581 = $3039;$h$0$lcssa$i582 = 1;$i$0$lcssa$i583 = $i$0$lcssa$i$i575;
         }
         $3046 = $$lcssa13$i581 << 24 >> 24; //@line 1666 "load_abc.cpp"
         $isdigittmp$i584 = (($3046) + -48)|0; //@line 1670 "load_abc.cpp"
         $isdigit$i585 = ($isdigittmp$i584>>>0)<(10); //@line 1670 "load_abc.cpp"
         if ($isdigit$i585) {
          $3049 = $3046;$h$04$i3$i587 = 0;$i$03$i4$i588 = 0;
          while(1) {
           $3047 = ($h$04$i3$i587*10)|0; //@line 1626 "load_abc.cpp"
           $3048 = (($3049) + -48)|0; //@line 1626 "load_abc.cpp"
           $3050 = (($3048) + ($3047))|0; //@line 1626 "load_abc.cpp"
           $3051 = (($i$03$i4$i588) + 1)|0; //@line 1627 "load_abc.cpp"
           $$sum$i589 = (($3051) + ($i$0$lcssa$i583))|0; //@line 1625 "load_abc.cpp"
           $$sum825 = (($$sum$i589) + ($3024))|0; //@line 1625 "load_abc.cpp"
           $3052 = (($p$18) + ($$sum825)|0); //@line 1625 "load_abc.cpp"
           $3053 = HEAP8[$3052>>0]|0; //@line 1625 "load_abc.cpp"
           $3054 = $3053 << 24 >> 24; //@line 1625 "load_abc.cpp"
           $isdigittmp$i5$i590 = (($3054) + -48)|0; //@line 1625 "load_abc.cpp"
           $isdigit$i6$i591 = ($isdigittmp$i5$i590>>>0)<(10); //@line 1625 "load_abc.cpp"
           if ($isdigit$i6$i591) {
            $3049 = $3054;$h$04$i3$i587 = $3050;$i$03$i4$i588 = $3051;
           } else {
            $$lcssa2256 = $3050;$$lcssa2257 = $3051;$$sum$i589$lcssa = $$sum$i589;
            break;
           }
          }
          $3055 = (($h$0$lcssa$i582|0) / 2)&-1; //@line 1671 "load_abc.cpp"
          $3056 = ($$lcssa2257|0)==(0); //@line 1629 "load_abc.cpp"
          $$h$0$i10$i593 = $3056 ? 1 : $$lcssa2256; //@line 1630 "load_abc.cpp"
          $3058 = $$h$0$i10$i593;$h$1$i596 = $3055;$i$1$i595 = $$sum$i589$lcssa;
         } else {
          $3058 = 1;$h$1$i596 = $h$0$lcssa$i582;$i$1$i595 = $i$0$lcssa$i583;
         }
         $3057 = Math_imul($h$1$i596, $3058)|0; //@line 1675 "load_abc.cpp"
         $$sum69 = (($i$1$i595) + ($3024))|0; //@line 4879 "load_abc.cpp"
         $3059 = (($p$18) + ($$sum69)|0); //@line 4879 "load_abc.cpp"
         $3060 = HEAP8[$3059>>0]|0; //@line 4880 "load_abc.cpp"
         $3061 = ($3060<<24>>24)==(45); //@line 4880 "load_abc.cpp"
         do {
          if ($3061) {
           $$sum70 = (($$sum69) + 1)|0; //@line 4881 "load_abc.cpp"
           $3062 = (($p$18) + ($$sum70)|0); //@line 4881 "load_abc.cpp"
           $3063 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4882 "load_abc.cpp"
           $3064 = ((($3063)) + 8|0); //@line 4882 "load_abc.cpp"
           $3065 = HEAPU8[$3064>>0]|(HEAPU8[$3064+1>>0]<<8)|(HEAPU8[$3064+2>>0]<<16)|(HEAPU8[$3064+3>>0]<<24); //@line 4882 "load_abc.cpp"
           $3066 = ($3065|0)==(0|0); //@line 4882 "load_abc.cpp"
           if ($3066) {
            $p$46 = $3062;
            break;
           }
           $3067 = ((($3065)) + 8|0); //@line 4882 "load_abc.cpp"
           $3068 = HEAP8[$3067>>0]|0; //@line 4882 "load_abc.cpp"
           $3069 = ($3068<<24>>24)==(1); //@line 4882 "load_abc.cpp"
           if ($3069) {
            $p$46 = $3062;
            break;
           }
           $3070 = ((($3063)) + 16|0); //@line 4883 "load_abc.cpp"
           HEAP8[$3070>>0]=$3065&255;HEAP8[$3070+1>>0]=($3065>>8)&255;HEAP8[$3070+2>>0]=($3065>>16)&255;HEAP8[$3070+3>>0]=$3065>>24; //@line 4883 "load_abc.cpp"
           $p$46 = $3062;
          } else {
           $p$46 = $3059;
          }
         } while(0);
         $3071 = ($1626|0)==(0); //@line 1718 "load_abc.cpp"
         if ($3071) {
          $$0$i598 = 0;$3081 = $$h$0$i$i576;$3083 = $3057;
         } else {
          $3072 = Math_imul($$h$0$i$i576, $1628)|0; //@line 1719 "load_abc.cpp"
          $3073 = Math_imul($3057, $1630)|0; //@line 1720 "load_abc.cpp"
          $3074 = (($1626) + -1)|0; //@line 1721 "load_abc.cpp"
          $$0$i598 = $3074;$3081 = $3072;$3083 = $3073;
         }
         $p$47 = $p$46;
         while(1) {
          $3075 = HEAP8[$p$47>>0]|0; //@line 4886 "load_abc.cpp"
          $3076 = $3075 << 24 >> 24; //@line 4886 "load_abc.cpp"
          $3077 = (_isspace($3076)|0); //@line 4886 "load_abc.cpp"
          $3078 = ($3077|0)==(0); //@line 4886 "load_abc.cpp"
          $3079 = ((($p$47)) + 1|0); //@line 4886 "load_abc.cpp"
          if ($3078) {
           $p$47$lcssa = $p$47;
           break;
          } else {
           $p$47 = $3079;
          }
         }
         switch ($1637|0) {
         case 60:  {
          $3080 = ($3081*3)|0; //@line 1683 "load_abc.cpp"
          $3082 = $3083 << 1; //@line 1684 "load_abc.cpp"
          $$0$i600 = 0;$3088 = $3080;$3090 = $3082;
          break;
         }
         case 62:  {
          $3084 = $3083 << 1; //@line 1688 "load_abc.cpp"
          $$0$i600 = 0;$3088 = $3081;$3090 = $3084;
          break;
         }
         default: {
          $$0$i600 = $abchornpipe$3$lcssa;$3088 = $3081;$3090 = $3083;
         }
         }
         $3085 = HEAP8[$p$47$lcssa>>0]|0; //@line 1692 "load_abc.cpp"
         $3086 = $3085 << 24 >> 24; //@line 1692 "load_abc.cpp"
         L738: do {
          switch ($3086|0) {
          case 62:  {
           $3087 = ($3088*3)|0; //@line 1695 "load_abc.cpp"
           $3089 = $3090 << 1; //@line 1696 "load_abc.cpp"
           $$01$i601 = 1;$3098 = $3089;$3104 = $3087;$4065 = 62;
           break;
          }
          case 60:  {
           $3091 = $3090 << 1; //@line 1699 "load_abc.cpp"
           $$01$i601 = 1;$3098 = $3091;$3104 = $3088;$4065 = 60;
           break;
          }
          default: {
           $3092 = ($$0$i600|0)!=(0); //@line 1705 "load_abc.cpp"
           $3093 = ($3088|0)==(1); //@line 1706 "load_abc.cpp"
           $or$cond794 = $3093 & $3092; //@line 1705 "load_abc.cpp"
           if (!($or$cond794)) {
            $$01$i601 = 0;$3098 = $3090;$3104 = $3088;$4065 = 0;
            break L738;
           }
           $3094 = ($3090|0)==(1); //@line 1706 "load_abc.cpp"
           $$795 = $3094 ? 3 : 1; //@line 1706 "load_abc.cpp"
           $$796 = $3094 ? 2 : $3090; //@line 1706 "load_abc.cpp"
           $$797 = $3094 ? 62 : 0; //@line 1706 "load_abc.cpp"
           $$01$i601 = 0;$3098 = $$796;$3104 = $$795;$4065 = $$797;
          }
          }
         } while(0);
         $3095 = (($p$47$lcssa) + ($$01$i601)|0); //@line 4887 "load_abc.cpp"
         $3096 = HEAP32[$snotediv>>2]|0; //@line 4888 "load_abc.cpp"
         $3097 = Math_imul($3096, $3098)|0; //@line 4888 "load_abc.cpp"
         $3099 = ($3097|0)==(0); //@line 345 "load_abc.cpp"
         if ($3099) {
          $$0$i603 = 0;
         } else {
          $3100 = HEAP32[$snotelen>>2]|0; //@line 4888 "load_abc.cpp"
          $3101 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4888 "load_abc.cpp"
          $3102 = HEAP32[299862]|0; //@line 346 "load_abc.cpp"
          $3103 = ($3104*3072)|0; //@line 346 "load_abc.cpp"
          $3105 = Math_imul($3103, $3100)|0; //@line 4888 "load_abc.cpp"
          $3106 = Math_imul($3105, $3101)|0; //@line 346 "load_abc.cpp"
          $3107 = Math_imul($3106, $3102)|0; //@line 346 "load_abc.cpp"
          $3108 = HEAP32[299863]|0; //@line 346 "load_abc.cpp"
          $3109 = Math_imul($3108, $3097)|0; //@line 346 "load_abc.cpp"
          $3110 = (($3107|0) / ($3109|0))&-1; //@line 346 "load_abc.cpp"
          $$0$i603 = $3110;
         }
         $not$824 = ($abcfermata$21228|0)!=(0); //@line 4889 "load_abc.cpp"
         $3111 = $not$824&1; //@line 4889 "load_abc.cpp"
         $$93 = $$0$i603 << $3111; //@line 4889 "load_abc.cpp"
         $3112 = ($$93>>>0)>($abcgrace$41234>>>0); //@line 4893 "load_abc.cpp"
         if ($3112) {
          $3113 = (($$93) - ($abcgrace$41234))|0; //@line 4894 "load_abc.cpp"
          $$pre1533 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4903 "load_abc.cpp"
          $3121 = $$pre1533;$abcgrace$7 = 0;$thistime$4 = $3113;
         } else {
          $3114 = (($abcgrace$41234) - ($$93))|0; //@line 4898 "load_abc.cpp"
          $3115 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4899 "load_abc.cpp"
          $3116 = ($3115*192)|0; //@line 339 "load_abc.cpp"
          $3117 = (($3114) + ($3116))|0; //@line 4900 "load_abc.cpp"
          $3121 = $3115;$abcgrace$7 = $3117;$thistime$4 = $3116;
         }
         $3118 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4902 "load_abc.cpp"
         $3119 = (($3118) + ($thistime$4))|0; //@line 4902 "load_abc.cpp"
         HEAP8[$31>>0]=$3119&255;HEAP8[$31+1>>0]=($3119>>8)&255;HEAP8[$31+2>>0]=($3119>>16)&255;HEAP8[$31+3>>0]=$3119>>24; //@line 4902 "load_abc.cpp"
         $3120 = ($3121*192)|0; //@line 339 "load_abc.cpp"
         $3122 = ($thistime$4>>>0)>($3120>>>0); //@line 4903 "load_abc.cpp"
         $3123 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4904 "load_abc.cpp"
         $3124 = ((($3123)) + 42|0); //@line 1174 "load_abc.cpp"
         $3125 = HEAP8[$3124>>0]|0; //@line 1174 "load_abc.cpp"
         if ($3122) {
          $3126 = (($3119) - ($abcnoslurs$91226))|0; //@line 4904 "load_abc.cpp"
          $3127 = ($ch0$21220<<24>>24)==(46); //@line 4904 "load_abc.cpp"
          $3128 = $thistime$4 >>> 1; //@line 4904 "load_abc.cpp"
          $$op = (0 - ($3128))|0;
          $$neg840 = $3127 ? $$op : 0;
          $3129 = (($3126) + ($$neg840))|0; //@line 4904 "load_abc.cpp"
          $3130 = $3125&255; //@line 1174 "load_abc.cpp"
          $3131 = (_pat_gmtosmp($3130)|0); //@line 1174 "load_abc.cpp"
          $3132 = $3131&255; //@line 1174 "load_abc.cpp"
          $3133 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$3133>>0]=0&255;HEAP8[$3133+1>>0]=(0>>8)&255;HEAP8[$3133+2>>0]=(0>>16)&255;HEAP8[$3133+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $3134 = ((($3133)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$3134>>0]=$3129&255;HEAP8[$3134+1>>0]=($3129>>8)&255;HEAP8[$3134+2>>0]=($3129>>16)&255;HEAP8[$3134+3>>0]=$3129>>24; //@line 541 "load_abc.cpp"
          $3135 = ((($3133)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$3135>>0] = 0; //@line 543 "load_abc.cpp"
          $3136 = ((($3133)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$3136>>0] = 0; //@line 543 "load_abc.cpp"
          $3137 = ((($3133)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$3137>>0] = $3132; //@line 543 "load_abc.cpp"
          $3138 = ((($3133)) + 11|0); //@line 543 "load_abc.cpp"
          HEAP8[$3138>>0] = 0; //@line 543 "load_abc.cpp"
          $3139 = ((($3133)) + 12|0); //@line 543 "load_abc.cpp"
          HEAP8[$3139>>0] = 0; //@line 543 "load_abc.cpp"
          $3140 = ((($3133)) + 13|0); //@line 543 "load_abc.cpp"
          HEAP8[$3140>>0] = 0; //@line 543 "load_abc.cpp"
          $3141 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $3142 = ((($3133)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$3142>>0] = $3141; //@line 544 "load_abc.cpp"
          $3143 = ((($3133)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$3143>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($3123,$3133); //@line 1179 "load_abc.cpp"
         } else {
          $3144 = $3125&255; //@line 1174 "load_abc.cpp"
          $3145 = (_pat_gmtosmp($3144)|0); //@line 1174 "load_abc.cpp"
          $3146 = $3145&255; //@line 1174 "load_abc.cpp"
          $3147 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$3147>>0]=0&255;HEAP8[$3147+1>>0]=(0>>8)&255;HEAP8[$3147+2>>0]=(0>>16)&255;HEAP8[$3147+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $3148 = ((($3147)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$3148>>0]=$3119&255;HEAP8[$3148+1>>0]=($3119>>8)&255;HEAP8[$3148+2>>0]=($3119>>16)&255;HEAP8[$3148+3>>0]=$3119>>24; //@line 541 "load_abc.cpp"
          $3149 = ((($3147)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$3149>>0] = 0; //@line 543 "load_abc.cpp"
          $3150 = ((($3147)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$3150>>0] = 0; //@line 543 "load_abc.cpp"
          $3151 = ((($3147)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$3151>>0] = $3146; //@line 543 "load_abc.cpp"
          $3152 = ((($3147)) + 11|0); //@line 543 "load_abc.cpp"
          HEAP8[$3152>>0] = 0; //@line 543 "load_abc.cpp"
          $3153 = ((($3147)) + 12|0); //@line 543 "load_abc.cpp"
          HEAP8[$3153>>0] = 0; //@line 543 "load_abc.cpp"
          $3154 = ((($3147)) + 13|0); //@line 543 "load_abc.cpp"
          HEAP8[$3154>>0] = 0; //@line 543 "load_abc.cpp"
          $3155 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $3156 = ((($3147)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$3156>>0] = $3155; //@line 544 "load_abc.cpp"
          $3157 = ((($3147)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$3157>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($3123,$3147); //@line 1179 "load_abc.cpp"
         }
         $3158 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4907 "load_abc.cpp"
         $3159 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4907 "load_abc.cpp"
         $3160 = ((($3158)) + 8|0); //@line 1590 "load_abc.cpp"
         $3161 = HEAPU8[$3160>>0]|(HEAPU8[$3160+1>>0]<<8)|(HEAPU8[$3160+2>>0]<<16)|(HEAPU8[$3160+3>>0]<<24); //@line 1590 "load_abc.cpp"
         $3162 = ($3161|0)==(0|0); //@line 1591 "load_abc.cpp"
         do {
          if ($3162) {
           label = 1033;
          } else {
           $3163 = ((($3161)) + 4|0); //@line 1591 "load_abc.cpp"
           $3164 = HEAPU8[$3163>>0]|(HEAPU8[$3163+1>>0]<<8)|(HEAPU8[$3163+2>>0]<<16)|(HEAPU8[$3163+3>>0]<<24); //@line 1591 "load_abc.cpp"
           $3165 = ($3164|0)==($3159|0); //@line 1591 "load_abc.cpp"
           if ($3165) {
            break;
           }
           $3166 = ((($3161)) + 8|0); //@line 1592 "load_abc.cpp"
           $3167 = HEAP8[$3166>>0]|0; //@line 1592 "load_abc.cpp"
           $3168 = ($3167<<24>>24)==(1); //@line 1592 "load_abc.cpp"
           if (!($3168)) {
            label = 1033;
            break;
           }
           $3169 = ((($3161)) + 9|0); //@line 1592 "load_abc.cpp"
           $3170 = HEAP8[$3169>>0]|0; //@line 1592 "load_abc.cpp"
           $3171 = ($3170<<24>>24)==(83); //@line 1592 "load_abc.cpp"
           if (!($3171)) {
            label = 1033;
            break;
           }
           HEAP8[$3163>>0]=$3159&255;HEAP8[$3163+1>>0]=($3159>>8)&255;HEAP8[$3163+2>>0]=($3159>>16)&255;HEAP8[$3163+3>>0]=$3159>>24; //@line 1593 "load_abc.cpp"
          }
         } while(0);
         if ((label|0) == 1033) {
          label = 0;
          $3172 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$3172>>0]=0&255;HEAP8[$3172+1>>0]=(0>>8)&255;HEAP8[$3172+2>>0]=(0>>16)&255;HEAP8[$3172+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $3173 = ((($3172)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$3173>>0]=$3159&255;HEAP8[$3173+1>>0]=($3159>>8)&255;HEAP8[$3173+2>>0]=($3159>>16)&255;HEAP8[$3173+3>>0]=$3159>>24; //@line 541 "load_abc.cpp"
          $3174 = ((($3172)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$3174>>0] = 1; //@line 543 "load_abc.cpp"
          $3175 = ((($3172)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$3175>>0] = 83; //@line 543 "load_abc.cpp"
          $3176 = ((($3172)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$3176>>0]=0&255;HEAP8[$3176+1>>0]=(0>>8)&255;HEAP8[$3176+2>>0]=(0>>16)&255;HEAP8[$3176+3>>0]=0>>24; //@line 543 "load_abc.cpp"
          $3177 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $3178 = ((($3172)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$3178>>0] = $3177; //@line 544 "load_abc.cpp"
          $3179 = ((($3172)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$3179>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($3158,$3172); //@line 1600 "load_abc.cpp"
         }
         $3180 = HEAPU8[$39>>0]|(HEAPU8[$39+1>>0]<<8)|(HEAPU8[$39+2>>0]<<16)|(HEAPU8[$39+3>>0]<<24); //@line 4908 "load_abc.cpp"
         $3181 = ($3180|0)==(0); //@line 4908 "load_abc.cpp"
         do {
          if (!($3181)) {
           $3182 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4908 "load_abc.cpp"
           $3183 = HEAPU8[$42>>0]|(HEAPU8[$42+1>>0]<<8)|(HEAPU8[$42+2>>0]<<16)|(HEAPU8[$42+3>>0]<<24); //@line 4908 "load_abc.cpp"
           $3184 = ($3182|0)==($3183|0); //@line 4908 "load_abc.cpp"
           if (!($3184)) {
            break;
           }
           $3185 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4909 "load_abc.cpp"
           __ZL14abc_add_gchordP10_ABCHANDLEjj($1,$3185,$bartime$41235); //@line 4909 "load_abc.cpp"
          }
         } while(0);
         $3186 = HEAPU8[$38>>0]|(HEAPU8[$38+1>>0]<<8)|(HEAPU8[$38+2>>0]<<16)|(HEAPU8[$38+3>>0]<<24); //@line 4910 "load_abc.cpp"
         $3187 = ($3186|0)==(0); //@line 4910 "load_abc.cpp"
         if ($3187) {
          $4057 = $4065;$4058 = $1630;$4059 = $1628;$4060 = $$0$i598;$abcarpeggio$3 = 0;$abceffect$5 = $abceffect$3$91;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = 0;$abcgrace$10 = $abcgrace$7;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = $abcto$31230;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $3095;
          break L678;
         }
         $3188 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4910 "load_abc.cpp"
         $3189 = HEAPU8[$43>>0]|(HEAPU8[$43+1>>0]<<8)|(HEAPU8[$43+2>>0]<<16)|(HEAPU8[$43+3>>0]<<24); //@line 4910 "load_abc.cpp"
         $3190 = ($3188|0)==($3189|0); //@line 4910 "load_abc.cpp"
         if (!($3190)) {
          $4057 = $4065;$4058 = $1630;$4059 = $1628;$4060 = $$0$i598;$abcarpeggio$3 = 0;$abceffect$5 = $abceffect$3$91;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = 0;$abcgrace$10 = $abcgrace$7;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = $abcto$31230;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $3095;
          break L678;
         }
         $3191 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4911 "load_abc.cpp"
         __ZL12abc_add_drumP10_ABCHANDLEjj($1,$3191,$bartime$41235); //@line 4911 "load_abc.cpp"
         $4057 = $4065;$4058 = $1630;$4059 = $1628;$4060 = $$0$i598;$abcarpeggio$3 = 0;$abceffect$5 = $abceffect$3$91;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = 0;$abcgrace$10 = $abcgrace$7;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = $abcto$31230;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $3095;
         break L678;
        }
        $memchr67 = (_memchr(527595,$1185,3)|0); //@line 4915 "load_abc.cpp"
        $3192 = ($memchr67|0)==(0|0); //@line 4915 "load_abc.cpp"
        if ($3192) {
         $memchr68 = (_memchr(527598,$1185,2)|0); //@line 4945 "load_abc.cpp"
         $3320 = ($memchr68|0)==(0|0); //@line 4945 "load_abc.cpp"
         if ($3320) {
          $3424 = (_isalpha($1185)|0); //@line 4973 "load_abc.cpp"
          $3425 = ($3424|0)==(0); //@line 4973 "load_abc.cpp"
          if ($3425) {
           $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = $abcto$31230;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$18;
           break L678;
          }
          $3426 = HEAP8[$p$18>>0]|0; //@line 4973 "load_abc.cpp"
          $3427 = ($3426<<24>>24)==(58); //@line 4973 "load_abc.cpp"
          if ($3427) {
           $p$18$pn = $p$18;
          } else {
           $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = $abcto$31230;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $p$18;
           break L678;
          }
          while(1) {
           $3428 = ((($p$18$pn)) + 1|0); //@line 4975 "load_abc.cpp"
           $$pre1535 = HEAP8[$3428>>0]|0; //@line 4975 "load_abc.cpp"
           $3429 = ($$pre1535<<24>>24)==(0); //@line 4975 "load_abc.cpp"
           if ($3429) {
            $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = $abcarpeggio$21229;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = $abcfermata$21228;$abcgrace$10 = $abcgrace$41234;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = $abcto$31230;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $3428;
            break L678;
           } else {
            $p$18$pn = $3428;
           }
          }
         }
         $3321 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4946 "load_abc.cpp"
         $3322 = ($3321|0)==(0|0); //@line 915 "load_abc.cpp"
         if ($3322) {
          $3323 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
          $3324 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
          $3325 = ((($3323)) + 20|0); //@line 917 "load_abc.cpp"
          HEAP8[$3325>>0]=$3324&255;HEAP8[$3325+1>>0]=($3324>>8)&255;HEAP8[$3325+2>>0]=($3324>>16)&255;HEAP8[$3325+3>>0]=$3324>>24; //@line 917 "load_abc.cpp"
          $$0$i654 = $3323;
         } else {
          $$0$i654 = $3321;
         }
         HEAP8[$41>>0]=$$0$i654&255;HEAP8[$41+1>>0]=($$0$i654>>8)&255;HEAP8[$41+2>>0]=($$0$i654>>16)&255;HEAP8[$41+3>>0]=$$0$i654>>24; //@line 4946 "load_abc.cpp"
         $3326 = ((($$0$i654)) + 36|0); //@line 1217 "load_abc.cpp"
         $3327 = HEAP8[$3326>>0]|0; //@line 1217 "load_abc.cpp"
         $tp$01$i656 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 1218 "load_abc.cpp"
         $3328 = ($tp$01$i656|0)==(0|0); //@line 1218 "load_abc.cpp"
         if (!($3328)) {
          $tp$02$i657 = $tp$01$i656;
          while(1) {
           $3329 = ((($tp$02$i657)) + 36|0); //@line 1219 "load_abc.cpp"
           $3330 = HEAP8[$3329>>0]|0; //@line 1219 "load_abc.cpp"
           $3331 = ($3330<<24>>24)==($3327<<24>>24); //@line 1219 "load_abc.cpp"
           if ($3331) {
            $3332 = ((($tp$02$i657)) + 16|0); //@line 1219 "load_abc.cpp"
            HEAP8[$3332>>0]=0&255;HEAP8[$3332+1>>0]=(0>>8)&255;HEAP8[$3332+2>>0]=(0>>16)&255;HEAP8[$3332+3>>0]=0>>24; //@line 1219 "load_abc.cpp"
           }
           $tp$0$i659 = HEAPU8[$tp$02$i657>>0]|(HEAPU8[$tp$02$i657+1>>0]<<8)|(HEAPU8[$tp$02$i657+2>>0]<<16)|(HEAPU8[$tp$02$i657+3>>0]<<24); //@line 1218 "load_abc.cpp"
           $3333 = ($tp$0$i659|0)==(0|0); //@line 1218 "load_abc.cpp"
           if ($3333) {
            break;
           } else {
            $tp$02$i657 = $tp$0$i659;
           }
          }
         }
         $3334 = HEAP8[$p$18>>0]|0; //@line 1625 "load_abc.cpp"
         $3335 = $3334 << 24 >> 24; //@line 1625 "load_abc.cpp"
         $isdigittmp1$i$i662 = (($3335) + -48)|0; //@line 1625 "load_abc.cpp"
         $isdigit2$i$i663 = ($isdigittmp1$i$i662>>>0)<(10); //@line 1625 "load_abc.cpp"
         if ($isdigit2$i$i663) {
          $3338 = $3335;$h$04$i$i664 = 0;$i$03$i$i665 = 0;
          while(1) {
           $3336 = ($h$04$i$i664*10)|0; //@line 1626 "load_abc.cpp"
           $3337 = (($3338) + -48)|0; //@line 1626 "load_abc.cpp"
           $3339 = (($3337) + ($3336))|0; //@line 1626 "load_abc.cpp"
           $3340 = (($i$03$i$i665) + 1)|0; //@line 1627 "load_abc.cpp"
           $3341 = (($p$18) + ($3340)|0); //@line 1625 "load_abc.cpp"
           $3342 = HEAP8[$3341>>0]|0; //@line 1625 "load_abc.cpp"
           $3343 = $3342 << 24 >> 24; //@line 1625 "load_abc.cpp"
           $isdigittmp$i$i666 = (($3343) + -48)|0; //@line 1625 "load_abc.cpp"
           $isdigit$i$i667 = ($isdigittmp$i$i666>>>0)<(10); //@line 1625 "load_abc.cpp"
           if ($isdigit$i$i667) {
            $3338 = $3343;$h$04$i$i664 = $3339;$i$03$i$i665 = $3340;
           } else {
            $3346 = $3342;$h$0$lcssa$i$i669 = $3339;$i$0$lcssa$i$i670 = $3340;
            break;
           }
          }
         } else {
          $3346 = $3334;$h$0$lcssa$i$i669 = 0;$i$0$lcssa$i$i670 = 0;
         }
         $3344 = ($i$0$lcssa$i$i670|0)==(0); //@line 1629 "load_abc.cpp"
         $3345 = ($3346<<24>>24)==(47); //@line 1666 "load_abc.cpp"
         if ($3345) {
          $h$017$i673 = 1;$i$016$i674 = $i$0$lcssa$i$i670;
          while(1) {
           $3347 = $h$017$i673 << 1; //@line 1667 "load_abc.cpp"
           $3348 = (($i$016$i674) + 1)|0; //@line 1668 "load_abc.cpp"
           $3349 = (($p$18) + ($3348)|0); //@line 1666 "load_abc.cpp"
           $3350 = HEAP8[$3349>>0]|0; //@line 1666 "load_abc.cpp"
           $3351 = ($3350<<24>>24)==(47); //@line 1666 "load_abc.cpp"
           if ($3351) {
            $h$017$i673 = $3347;$i$016$i674 = $3348;
           } else {
            $$lcssa13$i676 = $3350;$h$0$lcssa$i677 = $3347;$i$0$lcssa$i678 = $3348;
            break;
           }
          }
         } else {
          $$lcssa13$i676 = $3346;$h$0$lcssa$i677 = 1;$i$0$lcssa$i678 = $i$0$lcssa$i$i670;
         }
         $3352 = $$lcssa13$i676 << 24 >> 24; //@line 1666 "load_abc.cpp"
         $isdigittmp$i679 = (($3352) + -48)|0; //@line 1670 "load_abc.cpp"
         $isdigit$i680 = ($isdigittmp$i679>>>0)<(10); //@line 1670 "load_abc.cpp"
         if ($isdigit$i680) {
          $3355 = $3352;$h$04$i3$i682 = 0;$i$03$i4$i683 = 0;
          while(1) {
           $3353 = ($h$04$i3$i682*10)|0; //@line 1626 "load_abc.cpp"
           $3354 = (($3355) + -48)|0; //@line 1626 "load_abc.cpp"
           $3356 = (($3354) + ($3353))|0; //@line 1626 "load_abc.cpp"
           $3357 = (($i$03$i4$i683) + 1)|0; //@line 1627 "load_abc.cpp"
           $$sum$i684 = (($3357) + ($i$0$lcssa$i678))|0; //@line 1625 "load_abc.cpp"
           $3358 = (($p$18) + ($$sum$i684)|0); //@line 1625 "load_abc.cpp"
           $3359 = HEAP8[$3358>>0]|0; //@line 1625 "load_abc.cpp"
           $3360 = $3359 << 24 >> 24; //@line 1625 "load_abc.cpp"
           $isdigittmp$i5$i685 = (($3360) + -48)|0; //@line 1625 "load_abc.cpp"
           $isdigit$i6$i686 = ($isdigittmp$i5$i685>>>0)<(10); //@line 1625 "load_abc.cpp"
           if ($isdigit$i6$i686) {
            $3355 = $3360;$h$04$i3$i682 = $3356;$i$03$i4$i683 = $3357;
           } else {
            $$lcssa2272 = $3356;$$lcssa2273 = $3357;$$sum$i684$lcssa = $$sum$i684;
            break;
           }
          }
          $3361 = (($h$0$lcssa$i677|0) / 2)&-1; //@line 1671 "load_abc.cpp"
          $3362 = ($$lcssa2273|0)==(0); //@line 1629 "load_abc.cpp"
          $$h$0$i10$i688 = $3362 ? 1 : $$lcssa2272; //@line 1630 "load_abc.cpp"
          $3364 = $$h$0$i10$i688;$h$1$i691 = $3361;$i$1$i690 = $$sum$i684$lcssa;
         } else {
          $3364 = 1;$h$1$i691 = $h$0$lcssa$i677;$i$1$i690 = $i$0$lcssa$i678;
         }
         $3363 = Math_imul($h$1$i691, $3364)|0; //@line 1675 "load_abc.cpp"
         $3365 = (($p$18) + ($i$1$i690)|0); //@line 4948 "load_abc.cpp"
         $3366 = HEAP32[$mnotediv>>2]|0; //@line 4949 "load_abc.cpp"
         $3367 = Math_imul($3363, $3366)|0; //@line 4949 "load_abc.cpp"
         $3368 = ($3367|0)==(0); //@line 345 "load_abc.cpp"
         if ($3368) {
          $$0$i693 = 0;
         } else {
          $3369 = HEAP32[$mnotelen>>2]|0; //@line 4949 "load_abc.cpp"
          $3370 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4949 "load_abc.cpp"
          $3371 = HEAP32[299862]|0; //@line 346 "load_abc.cpp"
          $h$0$lcssa$i$i669$op = ($h$0$lcssa$i$i669*3072)|0; //@line 346 "load_abc.cpp"
          $3372 = $3344 ? 3072 : $h$0$lcssa$i$i669$op; //@line 346 "load_abc.cpp"
          $3373 = Math_imul($3372, $3369)|0; //@line 4949 "load_abc.cpp"
          $3374 = Math_imul($3373, $3370)|0; //@line 346 "load_abc.cpp"
          $3375 = Math_imul($3374, $3371)|0; //@line 346 "load_abc.cpp"
          $3376 = HEAP32[299863]|0; //@line 346 "load_abc.cpp"
          $3377 = Math_imul($3376, $3367)|0; //@line 346 "load_abc.cpp"
          $3378 = (($3375|0) / ($3377|0))&-1; //@line 346 "load_abc.cpp"
          $$0$i693 = $3378;
         }
         $not$829 = ($abcfermata$21228|0)!=(0); //@line 4950 "load_abc.cpp"
         $3379 = $not$829&1; //@line 4950 "load_abc.cpp"
         $$97 = $$0$i693 << $3379; //@line 4950 "load_abc.cpp"
         $3380 = ($$97>>>0)>($abcgrace$41234>>>0); //@line 4954 "load_abc.cpp"
         if ($3380) {
          $3381 = (($$97) - ($abcgrace$41234))|0; //@line 4955 "load_abc.cpp"
          $abcgrace$9 = 0;$thistime$8 = $3381;
         } else {
          $3382 = (($abcgrace$41234) - ($$97))|0; //@line 4959 "load_abc.cpp"
          $3383 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4960 "load_abc.cpp"
          $3384 = ($3383*192)|0; //@line 339 "load_abc.cpp"
          $3385 = (($3382) + ($3384))|0; //@line 4961 "load_abc.cpp"
          $abcgrace$9 = $3385;$thistime$8 = $3384;
         }
         $3386 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4963 "load_abc.cpp"
         $3387 = (($3386) + ($thistime$8))|0; //@line 4963 "load_abc.cpp"
         HEAP8[$31>>0]=$3387&255;HEAP8[$31+1>>0]=($3387>>8)&255;HEAP8[$31+2>>0]=($3387>>16)&255;HEAP8[$31+3>>0]=$3387>>24; //@line 4963 "load_abc.cpp"
         $3388 = (47112 + ($abckey$41224<<2)|0); //@line 4964 "load_abc.cpp"
         $3389 = HEAP32[$3388>>2]|0; //@line 4964 "load_abc.cpp"
         HEAP32[$vararg_buffer117>>2] = $3389; //@line 4964 "load_abc.cpp"
         $vararg_ptr120 = ((($vararg_buffer117)) + 4|0); //@line 4964 "load_abc.cpp"
         HEAP32[$vararg_ptr120>>2] = $3389; //@line 4964 "load_abc.cpp"
         (_sprintf($barsig,526499,$vararg_buffer117)|0); //@line 4964 "load_abc.cpp"
         $3390 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4965 "load_abc.cpp"
         $3391 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4965 "load_abc.cpp"
         $3392 = ((($3390)) + 8|0); //@line 1590 "load_abc.cpp"
         $3393 = HEAPU8[$3392>>0]|(HEAPU8[$3392+1>>0]<<8)|(HEAPU8[$3392+2>>0]<<16)|(HEAPU8[$3392+3>>0]<<24); //@line 1590 "load_abc.cpp"
         $3394 = ($3393|0)==(0|0); //@line 1591 "load_abc.cpp"
         do {
          if ($3394) {
           label = 1107;
          } else {
           $3395 = ((($3393)) + 4|0); //@line 1591 "load_abc.cpp"
           $3396 = HEAPU8[$3395>>0]|(HEAPU8[$3395+1>>0]<<8)|(HEAPU8[$3395+2>>0]<<16)|(HEAPU8[$3395+3>>0]<<24); //@line 1591 "load_abc.cpp"
           $3397 = ($3396|0)==($3391|0); //@line 1591 "load_abc.cpp"
           if ($3397) {
            break;
           }
           $3398 = ((($3393)) + 8|0); //@line 1592 "load_abc.cpp"
           $3399 = HEAP8[$3398>>0]|0; //@line 1592 "load_abc.cpp"
           $3400 = ($3399<<24>>24)==(1); //@line 1592 "load_abc.cpp"
           if (!($3400)) {
            label = 1107;
            break;
           }
           $3401 = ((($3393)) + 9|0); //@line 1592 "load_abc.cpp"
           $3402 = HEAP8[$3401>>0]|0; //@line 1592 "load_abc.cpp"
           $3403 = ($3402<<24>>24)==(83); //@line 1592 "load_abc.cpp"
           if (!($3403)) {
            label = 1107;
            break;
           }
           HEAP8[$3395>>0]=$3391&255;HEAP8[$3395+1>>0]=($3391>>8)&255;HEAP8[$3395+2>>0]=($3391>>16)&255;HEAP8[$3395+3>>0]=$3391>>24; //@line 1593 "load_abc.cpp"
          }
         } while(0);
         if ((label|0) == 1107) {
          label = 0;
          $3404 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$3404>>0]=0&255;HEAP8[$3404+1>>0]=(0>>8)&255;HEAP8[$3404+2>>0]=(0>>16)&255;HEAP8[$3404+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $3405 = ((($3404)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$3405>>0]=$3391&255;HEAP8[$3405+1>>0]=($3391>>8)&255;HEAP8[$3405+2>>0]=($3391>>16)&255;HEAP8[$3405+3>>0]=$3391>>24; //@line 541 "load_abc.cpp"
          $3406 = ((($3404)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$3406>>0] = 1; //@line 543 "load_abc.cpp"
          $3407 = ((($3404)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$3407>>0] = 83; //@line 543 "load_abc.cpp"
          $3408 = ((($3404)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$3408>>0]=0&255;HEAP8[$3408+1>>0]=(0>>8)&255;HEAP8[$3408+2>>0]=(0>>16)&255;HEAP8[$3408+3>>0]=0>>24; //@line 543 "load_abc.cpp"
          $3409 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $3410 = ((($3404)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$3410>>0] = $3409; //@line 544 "load_abc.cpp"
          $3411 = ((($3404)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$3411>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($3390,$3404); //@line 1600 "load_abc.cpp"
         }
         $3412 = HEAPU8[$39>>0]|(HEAPU8[$39+1>>0]<<8)|(HEAPU8[$39+2>>0]<<16)|(HEAPU8[$39+3>>0]<<24); //@line 4966 "load_abc.cpp"
         $3413 = ($3412|0)==(0); //@line 4966 "load_abc.cpp"
         do {
          if (!($3413)) {
           $3414 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4966 "load_abc.cpp"
           $3415 = HEAPU8[$42>>0]|(HEAPU8[$42+1>>0]<<8)|(HEAPU8[$42+2>>0]<<16)|(HEAPU8[$42+3>>0]<<24); //@line 4966 "load_abc.cpp"
           $3416 = ($3414|0)==($3415|0); //@line 4966 "load_abc.cpp"
           if (!($3416)) {
            break;
           }
           $3417 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4967 "load_abc.cpp"
           __ZL14abc_add_gchordP10_ABCHANDLEjj($1,$3417,$bartime$41235); //@line 4967 "load_abc.cpp"
          }
         } while(0);
         $3418 = HEAPU8[$38>>0]|(HEAPU8[$38+1>>0]<<8)|(HEAPU8[$38+2>>0]<<16)|(HEAPU8[$38+3>>0]<<24); //@line 4968 "load_abc.cpp"
         $3419 = ($3418|0)==(0); //@line 4968 "load_abc.cpp"
         if ($3419) {
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = 0;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = 0;$abcgrace$10 = $abcgrace$9;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = $abcto$31230;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $3365;
          break L678;
         }
         $3420 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4968 "load_abc.cpp"
         $3421 = HEAPU8[$43>>0]|(HEAPU8[$43+1>>0]<<8)|(HEAPU8[$43+2>>0]<<16)|(HEAPU8[$43+3>>0]<<24); //@line 4968 "load_abc.cpp"
         $3422 = ($3420|0)==($3421|0); //@line 4968 "load_abc.cpp"
         if (!($3422)) {
          $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = 0;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = 0;$abcgrace$10 = $abcgrace$9;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = $abcto$31230;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $3365;
          break L678;
         }
         $3423 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4969 "load_abc.cpp"
         __ZL12abc_add_drumP10_ABCHANDLEjj($1,$3423,$bartime$41235); //@line 4969 "load_abc.cpp"
         $4057 = $1637;$4058 = $1630;$4059 = $1628;$4060 = $1626;$abcarpeggio$3 = 0;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = 0;$abcgrace$10 = $abcgrace$9;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = $abcto$31230;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $3365;
         break L678;
        }
        $3193 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4916 "load_abc.cpp"
        $3194 = ($3193|0)==(0|0); //@line 915 "load_abc.cpp"
        if ($3194) {
         $3195 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
         $3196 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
         $3197 = ((($3195)) + 20|0); //@line 917 "load_abc.cpp"
         HEAP8[$3197>>0]=$3196&255;HEAP8[$3197+1>>0]=($3196>>8)&255;HEAP8[$3197+2>>0]=($3196>>16)&255;HEAP8[$3197+3>>0]=$3196>>24; //@line 917 "load_abc.cpp"
         $$0$i607 = $3195;
        } else {
         $$0$i607 = $3193;
        }
        HEAP8[$41>>0]=$$0$i607&255;HEAP8[$41+1>>0]=($$0$i607>>8)&255;HEAP8[$41+2>>0]=($$0$i607>>16)&255;HEAP8[$41+3>>0]=$$0$i607>>24; //@line 4916 "load_abc.cpp"
        $3198 = ((($$0$i607)) + 36|0); //@line 1217 "load_abc.cpp"
        $3199 = HEAP8[$3198>>0]|0; //@line 1217 "load_abc.cpp"
        $tp$01$i609 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 1218 "load_abc.cpp"
        $3200 = ($tp$01$i609|0)==(0|0); //@line 1218 "load_abc.cpp"
        if (!($3200)) {
         $tp$02$i610 = $tp$01$i609;
         while(1) {
          $3201 = ((($tp$02$i610)) + 36|0); //@line 1219 "load_abc.cpp"
          $3202 = HEAP8[$3201>>0]|0; //@line 1219 "load_abc.cpp"
          $3203 = ($3202<<24>>24)==($3199<<24>>24); //@line 1219 "load_abc.cpp"
          if ($3203) {
           $3204 = ((($tp$02$i610)) + 16|0); //@line 1219 "load_abc.cpp"
           HEAP8[$3204>>0]=0&255;HEAP8[$3204+1>>0]=(0>>8)&255;HEAP8[$3204+2>>0]=(0>>16)&255;HEAP8[$3204+3>>0]=0>>24; //@line 1219 "load_abc.cpp"
          }
          $tp$0$i612 = HEAPU8[$tp$02$i610>>0]|(HEAPU8[$tp$02$i610+1>>0]<<8)|(HEAPU8[$tp$02$i610+2>>0]<<16)|(HEAPU8[$tp$02$i610+3>>0]<<24); //@line 1218 "load_abc.cpp"
          $3205 = ($tp$0$i612|0)==(0|0); //@line 1218 "load_abc.cpp"
          if ($3205) {
           break;
          } else {
           $tp$02$i610 = $tp$0$i612;
          }
         }
        }
        $3206 = HEAP8[$p$18>>0]|0; //@line 1625 "load_abc.cpp"
        $3207 = $3206 << 24 >> 24; //@line 1625 "load_abc.cpp"
        $isdigittmp1$i$i614 = (($3207) + -48)|0; //@line 1625 "load_abc.cpp"
        $isdigit2$i$i615 = ($isdigittmp1$i$i614>>>0)<(10); //@line 1625 "load_abc.cpp"
        if ($isdigit2$i$i615) {
         $3210 = $3207;$h$04$i$i616 = 0;$i$03$i$i617 = 0;
         while(1) {
          $3208 = ($h$04$i$i616*10)|0; //@line 1626 "load_abc.cpp"
          $3209 = (($3210) + -48)|0; //@line 1626 "load_abc.cpp"
          $3211 = (($3209) + ($3208))|0; //@line 1626 "load_abc.cpp"
          $3212 = (($i$03$i$i617) + 1)|0; //@line 1627 "load_abc.cpp"
          $3213 = (($p$18) + ($3212)|0); //@line 1625 "load_abc.cpp"
          $3214 = HEAP8[$3213>>0]|0; //@line 1625 "load_abc.cpp"
          $3215 = $3214 << 24 >> 24; //@line 1625 "load_abc.cpp"
          $isdigittmp$i$i618 = (($3215) + -48)|0; //@line 1625 "load_abc.cpp"
          $isdigit$i$i619 = ($isdigittmp$i$i618>>>0)<(10); //@line 1625 "load_abc.cpp"
          if ($isdigit$i$i619) {
           $3210 = $3215;$h$04$i$i616 = $3211;$i$03$i$i617 = $3212;
          } else {
           $3218 = $3214;$h$0$lcssa$i$i621 = $3211;$i$0$lcssa$i$i622 = $3212;
           break;
          }
         }
        } else {
         $3218 = $3206;$h$0$lcssa$i$i621 = 0;$i$0$lcssa$i$i622 = 0;
        }
        $3216 = ($i$0$lcssa$i$i622|0)==(0); //@line 1629 "load_abc.cpp"
        $$h$0$i$i623 = $3216 ? 1 : $h$0$lcssa$i$i621; //@line 1630 "load_abc.cpp"
        $3217 = ($3218<<24>>24)==(47); //@line 1666 "load_abc.cpp"
        if ($3217) {
         $h$017$i625 = 1;$i$016$i626 = $i$0$lcssa$i$i622;
         while(1) {
          $3219 = $h$017$i625 << 1; //@line 1667 "load_abc.cpp"
          $3220 = (($i$016$i626) + 1)|0; //@line 1668 "load_abc.cpp"
          $3221 = (($p$18) + ($3220)|0); //@line 1666 "load_abc.cpp"
          $3222 = HEAP8[$3221>>0]|0; //@line 1666 "load_abc.cpp"
          $3223 = ($3222<<24>>24)==(47); //@line 1666 "load_abc.cpp"
          if ($3223) {
           $h$017$i625 = $3219;$i$016$i626 = $3220;
          } else {
           $$lcssa13$i628 = $3222;$h$0$lcssa$i629 = $3219;$i$0$lcssa$i630 = $3220;
           break;
          }
         }
        } else {
         $$lcssa13$i628 = $3218;$h$0$lcssa$i629 = 1;$i$0$lcssa$i630 = $i$0$lcssa$i$i622;
        }
        $3224 = $$lcssa13$i628 << 24 >> 24; //@line 1666 "load_abc.cpp"
        $isdigittmp$i631 = (($3224) + -48)|0; //@line 1670 "load_abc.cpp"
        $isdigit$i632 = ($isdigittmp$i631>>>0)<(10); //@line 1670 "load_abc.cpp"
        if ($isdigit$i632) {
         $3227 = $3224;$h$04$i3$i634 = 0;$i$03$i4$i635 = 0;
         while(1) {
          $3225 = ($h$04$i3$i634*10)|0; //@line 1626 "load_abc.cpp"
          $3226 = (($3227) + -48)|0; //@line 1626 "load_abc.cpp"
          $3228 = (($3226) + ($3225))|0; //@line 1626 "load_abc.cpp"
          $3229 = (($i$03$i4$i635) + 1)|0; //@line 1627 "load_abc.cpp"
          $$sum$i636 = (($3229) + ($i$0$lcssa$i630))|0; //@line 1625 "load_abc.cpp"
          $3230 = (($p$18) + ($$sum$i636)|0); //@line 1625 "load_abc.cpp"
          $3231 = HEAP8[$3230>>0]|0; //@line 1625 "load_abc.cpp"
          $3232 = $3231 << 24 >> 24; //@line 1625 "load_abc.cpp"
          $isdigittmp$i5$i637 = (($3232) + -48)|0; //@line 1625 "load_abc.cpp"
          $isdigit$i6$i638 = ($isdigittmp$i5$i637>>>0)<(10); //@line 1625 "load_abc.cpp"
          if ($isdigit$i6$i638) {
           $3227 = $3232;$h$04$i3$i634 = $3228;$i$03$i4$i635 = $3229;
          } else {
           $$lcssa2264 = $3228;$$lcssa2265 = $3229;$$sum$i636$lcssa = $$sum$i636;
           break;
          }
         }
         $3233 = (($h$0$lcssa$i629|0) / 2)&-1; //@line 1671 "load_abc.cpp"
         $3234 = ($$lcssa2265|0)==(0); //@line 1629 "load_abc.cpp"
         $$h$0$i10$i640 = $3234 ? 1 : $$lcssa2264; //@line 1630 "load_abc.cpp"
         $3236 = $$h$0$i10$i640;$h$1$i643 = $3233;$i$1$i642 = $$sum$i636$lcssa;
        } else {
         $3236 = 1;$h$1$i643 = $h$0$lcssa$i629;$i$1$i642 = $i$0$lcssa$i630;
        }
        $3235 = Math_imul($h$1$i643, $3236)|0; //@line 1675 "load_abc.cpp"
        $3237 = (($p$18) + ($i$1$i642)|0); //@line 4918 "load_abc.cpp"
        $3238 = ($1626|0)==(0); //@line 1718 "load_abc.cpp"
        if ($3238) {
         $$0$i645 = 0;$3248 = $$h$0$i$i623;$3250 = $3235;
        } else {
         $3239 = Math_imul($$h$0$i$i623, $1628)|0; //@line 1719 "load_abc.cpp"
         $3240 = Math_imul($3235, $1630)|0; //@line 1720 "load_abc.cpp"
         $3241 = (($1626) + -1)|0; //@line 1721 "load_abc.cpp"
         $$0$i645 = $3241;$3248 = $3239;$3250 = $3240;
        }
        $p$48 = $3237;
        while(1) {
         $3242 = HEAP8[$p$48>>0]|0; //@line 4920 "load_abc.cpp"
         $3243 = $3242 << 24 >> 24; //@line 4920 "load_abc.cpp"
         $3244 = (_isspace($3243)|0); //@line 4920 "load_abc.cpp"
         $3245 = ($3244|0)==(0); //@line 4920 "load_abc.cpp"
         $3246 = ((($p$48)) + 1|0); //@line 4920 "load_abc.cpp"
         if ($3245) {
          $p$48$lcssa = $p$48;
          break;
         } else {
          $p$48 = $3246;
         }
        }
        switch ($1637|0) {
        case 60:  {
         $3247 = ($3248*3)|0; //@line 1683 "load_abc.cpp"
         $3249 = $3250 << 1; //@line 1684 "load_abc.cpp"
         $$0$i647 = 0;$3255 = $3247;$3257 = $3249;
         break;
        }
        case 62:  {
         $3251 = $3250 << 1; //@line 1688 "load_abc.cpp"
         $$0$i647 = 0;$3255 = $3248;$3257 = $3251;
         break;
        }
        default: {
         $$0$i647 = $abchornpipe$3$lcssa;$3255 = $3248;$3257 = $3250;
        }
        }
        $3252 = HEAP8[$p$48$lcssa>>0]|0; //@line 1692 "load_abc.cpp"
        $3253 = $3252 << 24 >> 24; //@line 1692 "load_abc.cpp"
        L856: do {
         switch ($3253|0) {
         case 62:  {
          $3254 = ($3255*3)|0; //@line 1695 "load_abc.cpp"
          $3256 = $3257 << 1; //@line 1696 "load_abc.cpp"
          $$01$i648 = 1;$3265 = $3256;$3271 = $3254;$4066 = 62;
          break;
         }
         case 60:  {
          $3258 = $3257 << 1; //@line 1699 "load_abc.cpp"
          $$01$i648 = 1;$3265 = $3258;$3271 = $3255;$4066 = 60;
          break;
         }
         default: {
          $3259 = ($$0$i647|0)!=(0); //@line 1705 "load_abc.cpp"
          $3260 = ($3255|0)==(1); //@line 1706 "load_abc.cpp"
          $or$cond798 = $3260 & $3259; //@line 1705 "load_abc.cpp"
          if (!($or$cond798)) {
           $$01$i648 = 0;$3265 = $3257;$3271 = $3255;$4066 = 0;
           break L856;
          }
          $3261 = ($3257|0)==(1); //@line 1706 "load_abc.cpp"
          $$799 = $3261 ? 3 : 1; //@line 1706 "load_abc.cpp"
          $$800 = $3261 ? 2 : $3257; //@line 1706 "load_abc.cpp"
          $$801 = $3261 ? 62 : 0; //@line 1706 "load_abc.cpp"
          $$01$i648 = 0;$3265 = $$800;$3271 = $$799;$4066 = $$801;
         }
         }
        } while(0);
        $3262 = (($p$48$lcssa) + ($$01$i648)|0); //@line 4921 "load_abc.cpp"
        $3263 = HEAP32[$snotediv>>2]|0; //@line 4922 "load_abc.cpp"
        $3264 = Math_imul($3263, $3265)|0; //@line 4922 "load_abc.cpp"
        $3266 = ($3264|0)==(0); //@line 345 "load_abc.cpp"
        if ($3266) {
         $$0$i650 = 0;
        } else {
         $3267 = HEAP32[$snotelen>>2]|0; //@line 4922 "load_abc.cpp"
         $3268 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4922 "load_abc.cpp"
         $3269 = HEAP32[299862]|0; //@line 346 "load_abc.cpp"
         $3270 = ($3271*3072)|0; //@line 346 "load_abc.cpp"
         $3272 = Math_imul($3270, $3267)|0; //@line 4922 "load_abc.cpp"
         $3273 = Math_imul($3272, $3268)|0; //@line 346 "load_abc.cpp"
         $3274 = Math_imul($3273, $3269)|0; //@line 346 "load_abc.cpp"
         $3275 = HEAP32[299863]|0; //@line 346 "load_abc.cpp"
         $3276 = Math_imul($3275, $3264)|0; //@line 346 "load_abc.cpp"
         $3277 = (($3274|0) / ($3276|0))&-1; //@line 346 "load_abc.cpp"
         $$0$i650 = $3277;
        }
        $not$828 = ($abcfermata$21228|0)!=(0); //@line 4923 "load_abc.cpp"
        $3278 = $not$828&1; //@line 4923 "load_abc.cpp"
        $$95 = $$0$i650 << $3278; //@line 4923 "load_abc.cpp"
        $3279 = ($$95>>>0)>($abcgrace$41234>>>0); //@line 4927 "load_abc.cpp"
        if ($3279) {
         $3280 = (($$95) - ($abcgrace$41234))|0; //@line 4928 "load_abc.cpp"
         $abcgrace$8 = 0;$thistime$6 = $3280;
        } else {
         $3281 = (($abcgrace$41234) - ($$95))|0; //@line 4932 "load_abc.cpp"
         $3282 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 4933 "load_abc.cpp"
         $3283 = ($3282*192)|0; //@line 339 "load_abc.cpp"
         $3284 = (($3281) + ($3283))|0; //@line 4934 "load_abc.cpp"
         $abcgrace$8 = $3284;$thistime$6 = $3283;
        }
        $3285 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4936 "load_abc.cpp"
        $3286 = (($3285) + ($thistime$6))|0; //@line 4936 "load_abc.cpp"
        HEAP8[$31>>0]=$3286&255;HEAP8[$31+1>>0]=($3286>>8)&255;HEAP8[$31+2>>0]=($3286>>16)&255;HEAP8[$31+3>>0]=$3286>>24; //@line 4936 "load_abc.cpp"
        $3287 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4937 "load_abc.cpp"
        $3288 = ((($3287)) + 8|0); //@line 1590 "load_abc.cpp"
        $3289 = HEAPU8[$3288>>0]|(HEAPU8[$3288+1>>0]<<8)|(HEAPU8[$3288+2>>0]<<16)|(HEAPU8[$3288+3>>0]<<24); //@line 1590 "load_abc.cpp"
        $3290 = ($3289|0)==(0|0); //@line 1591 "load_abc.cpp"
        do {
         if ($3290) {
          label = 1076;
         } else {
          $3291 = ((($3289)) + 4|0); //@line 1591 "load_abc.cpp"
          $3292 = HEAPU8[$3291>>0]|(HEAPU8[$3291+1>>0]<<8)|(HEAPU8[$3291+2>>0]<<16)|(HEAPU8[$3291+3>>0]<<24); //@line 1591 "load_abc.cpp"
          $3293 = ($3292|0)==($3286|0); //@line 1591 "load_abc.cpp"
          if ($3293) {
           break;
          }
          $3294 = ((($3289)) + 8|0); //@line 1592 "load_abc.cpp"
          $3295 = HEAP8[$3294>>0]|0; //@line 1592 "load_abc.cpp"
          $3296 = ($3295<<24>>24)==(1); //@line 1592 "load_abc.cpp"
          if (!($3296)) {
           label = 1076;
           break;
          }
          $3297 = ((($3289)) + 9|0); //@line 1592 "load_abc.cpp"
          $3298 = HEAP8[$3297>>0]|0; //@line 1592 "load_abc.cpp"
          $3299 = ($3298<<24>>24)==(83); //@line 1592 "load_abc.cpp"
          if (!($3299)) {
           label = 1076;
           break;
          }
          HEAP8[$3291>>0]=$3286&255;HEAP8[$3291+1>>0]=($3286>>8)&255;HEAP8[$3291+2>>0]=($3286>>16)&255;HEAP8[$3291+3>>0]=$3286>>24; //@line 1593 "load_abc.cpp"
         }
        } while(0);
        if ((label|0) == 1076) {
         label = 0;
         $3300 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
         HEAP8[$3300>>0]=0&255;HEAP8[$3300+1>>0]=(0>>8)&255;HEAP8[$3300+2>>0]=(0>>16)&255;HEAP8[$3300+3>>0]=0>>24; //@line 540 "load_abc.cpp"
         $3301 = ((($3300)) + 4|0); //@line 541 "load_abc.cpp"
         HEAP8[$3301>>0]=$3286&255;HEAP8[$3301+1>>0]=($3286>>8)&255;HEAP8[$3301+2>>0]=($3286>>16)&255;HEAP8[$3301+3>>0]=$3286>>24; //@line 541 "load_abc.cpp"
         $3302 = ((($3300)) + 8|0); //@line 543 "load_abc.cpp"
         HEAP8[$3302>>0] = 1; //@line 543 "load_abc.cpp"
         $3303 = ((($3300)) + 9|0); //@line 543 "load_abc.cpp"
         HEAP8[$3303>>0] = 83; //@line 543 "load_abc.cpp"
         $3304 = ((($3300)) + 10|0); //@line 543 "load_abc.cpp"
         HEAP8[$3304>>0]=0&255;HEAP8[$3304+1>>0]=(0>>8)&255;HEAP8[$3304+2>>0]=(0>>16)&255;HEAP8[$3304+3>>0]=0>>24; //@line 543 "load_abc.cpp"
         $3305 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
         $3306 = ((($3300)) + 14|0); //@line 544 "load_abc.cpp"
         HEAP8[$3306>>0] = $3305; //@line 544 "load_abc.cpp"
         $3307 = ((($3300)) + 15|0); //@line 545 "load_abc.cpp"
         HEAP8[$3307>>0] = 0; //@line 545 "load_abc.cpp"
         __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($3287,$3300); //@line 1600 "load_abc.cpp"
        }
        $3308 = HEAPU8[$39>>0]|(HEAPU8[$39+1>>0]<<8)|(HEAPU8[$39+2>>0]<<16)|(HEAPU8[$39+3>>0]<<24); //@line 4938 "load_abc.cpp"
        $3309 = ($3308|0)==(0); //@line 4938 "load_abc.cpp"
        do {
         if (!($3309)) {
          $3310 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4938 "load_abc.cpp"
          $3311 = HEAPU8[$42>>0]|(HEAPU8[$42+1>>0]<<8)|(HEAPU8[$42+2>>0]<<16)|(HEAPU8[$42+3>>0]<<24); //@line 4938 "load_abc.cpp"
          $3312 = ($3310|0)==($3311|0); //@line 4938 "load_abc.cpp"
          if (!($3312)) {
           break;
          }
          $3313 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4939 "load_abc.cpp"
          __ZL14abc_add_gchordP10_ABCHANDLEjj($1,$3313,$bartime$41235); //@line 4939 "load_abc.cpp"
         }
        } while(0);
        $3314 = HEAPU8[$38>>0]|(HEAPU8[$38+1>>0]<<8)|(HEAPU8[$38+2>>0]<<16)|(HEAPU8[$38+3>>0]<<24); //@line 4940 "load_abc.cpp"
        $3315 = ($3314|0)==(0); //@line 4940 "load_abc.cpp"
        if ($3315) {
         $4057 = $4066;$4058 = $1630;$4059 = $1628;$4060 = $$0$i645;$abcarpeggio$3 = 0;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = 0;$abcgrace$10 = $abcgrace$8;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = $abcto$31230;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $3262;
         break L678;
        }
        $3316 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 4940 "load_abc.cpp"
        $3317 = HEAPU8[$43>>0]|(HEAPU8[$43+1>>0]<<8)|(HEAPU8[$43+2>>0]<<16)|(HEAPU8[$43+3>>0]<<24); //@line 4940 "load_abc.cpp"
        $3318 = ($3316|0)==($3317|0); //@line 4940 "load_abc.cpp"
        if (!($3318)) {
         $4057 = $4066;$4058 = $1630;$4059 = $1628;$4060 = $$0$i645;$abcarpeggio$3 = 0;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = 0;$abcgrace$10 = $abcgrace$8;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = $abcto$31230;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $3262;
         break L678;
        }
        $3319 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 4941 "load_abc.cpp"
        __ZL12abc_add_drumP10_ABCHANDLEjj($1,$3319,$bartime$41235); //@line 4941 "load_abc.cpp"
        $4057 = $4066;$4058 = $1630;$4059 = $1628;$4060 = $$0$i645;$abcarpeggio$3 = 0;$abceffect$5 = $abceffect$31222;$abceffoper$4 = $abceffoper$31223;$abcfermata$7 = 0;$abcgrace$10 = $abcgrace$8;$abckey$5 = $abckey$41224;$abcnolegato$14 = $abcnolegato$101227;$abcnoslurs$12 = $abcnoslurs$91226;$abcstate$8 = $abcstate$71221;$abctempo$6 = $abctempo$51231;$abcto$6 = $abcto$31230;$abcvol$4 = $abcvol$31225;$bartime$5 = $bartime$41235;$ch$1 = $ch$0;$cnotediv$6 = $cnotediv$21233;$cnotelen$6 = $cnotelen$21232;$p$50 = $3262;
       }
       }
      } while(0);
      $3430 = ($pp$1|0)==(0|0); //@line 4981 "load_abc.cpp"
      if ($3430) {
       $p$17$be = $p$50;$pp$0$be = 0;
      } else {
       $3433 = HEAP8[$p$50>>0]|0; //@line 4982 "load_abc.cpp"
       $3434 = ($3433<<24>>24)==(0); //@line 4982 "load_abc.cpp"
       $pp$1$p$50 = $3434 ? $pp$1 : $p$50; //@line 4982 "load_abc.cpp"
       $$pp$1 = $3434 ? 0 : $pp$1; //@line 4982 "load_abc.cpp"
       $p$17$be = $pp$1$p$50;$pp$0$be = $$pp$1;
      }
      $3431 = HEAP8[$p$17$be>>0]|0; //@line 4304 "load_abc.cpp"
      $3432 = ($3431<<24>>24)==(0); //@line 4304 "load_abc.cpp"
      if ($3432) {
       $$lcssa2275 = $4057;$$lcssa2276 = $4058;$$lcssa2277 = $4059;$$lcssa2278 = $4060;$abcarpeggio$3$lcssa = $abcarpeggio$3;$abceffect$5$lcssa = $abceffect$5;$abceffoper$4$lcssa = $abceffoper$4;$abcfermata$7$lcssa = $abcfermata$7;$abcgrace$10$lcssa = $abcgrace$10;$abckey$5$lcssa = $abckey$5;$abcnolegato$14$lcssa = $abcnolegato$14;$abcnoslurs$12$lcssa = $abcnoslurs$12;$abcstate$8$lcssa = $abcstate$8;$abctempo$6$lcssa = $abctempo$6;$abcto$6$lcssa = $abcto$6;$abcvol$4$lcssa = $abcvol$4;$bartime$5$lcssa = $bartime$5;$ch$1$lcssa = $ch$1;$cnotediv$6$lcssa = $cnotediv$6;$cnotelen$6$lcssa = $cnotelen$6;
       break;
      } else {
       $1169 = $3431;$1626 = $4060;$1628 = $4059;$1630 = $4058;$1637 = $4057;$abcarpeggio$21229 = $abcarpeggio$3;$abceffect$31222 = $abceffect$5;$abceffoper$31223 = $abceffoper$4;$abcfermata$21228 = $abcfermata$7;$abcgrace$41234 = $abcgrace$10;$abckey$41224 = $abckey$5;$abcnolegato$101227 = $abcnolegato$14;$abcnoslurs$91226 = $abcnoslurs$12;$abcstate$71221 = $abcstate$8;$abctempo$51231 = $abctempo$6;$abcto$31230 = $abcto$6;$abcvol$31225 = $abcvol$4;$bartime$41235 = $bartime$5;$ch0$21220 = $ch$1;$cnotediv$21233 = $cnotediv$6;$cnotelen$21232 = $cnotelen$6;$p$16$pn = $p$17$be;$pp$01219 = $pp$0$be;
      }
     }
     $176 = (($mmstack) + ($mmsp$3$lcssa<<2)|0); //@line 3908 "load_abc.cpp"
     $177 = HEAP32[$176>>2]|0; //@line 3908 "load_abc.cpp"
     $178 = (__ZL8abc_getsP10_ABCHANDLEP6MMFILE($1,$177)|0); //@line 3908 "load_abc.cpp"
     $179 = ($178|0)==(0|0); //@line 3908 "load_abc.cpp"
     if ($179) {
      $$lcssa928 = $$lcssa2275;$$lcssa948 = $176;$$ph892$lcssa = $$lcssa2276;$$ph893$lcssa = $$lcssa2277;$$ph894$lcssa = $$lcssa2278;$abcarpeggio$1$ph$lcssa = $abcarpeggio$3$lcssa;$abceffect$1$lcssa = $abceffect$5$lcssa;$abceffoper$1$lcssa = $abceffoper$4$lcssa;$abcfermata$1$ph$lcssa = $abcfermata$7$lcssa;$abcgrace$1$lcssa = $abcgrace$10$lcssa;$abchornpipe$1$lcssa = $abchornpipe$3$lcssa;$abckey$1$lcssa = $abckey$5$lcssa;$abcnolegato$1$lcssa = $abcnolegato$14$lcssa;$abcnoslurs$1$lcssa = $abcnoslurs$12$lcssa;$abcstate$2$lcssa = $abcstate$8$lcssa;$abctempo$1$lcssa = $abctempo$6$lcssa;$abcto$1$lcssa = $abcto$6$lcssa;$abcvol$1$lcssa = $abcvol$4$lcssa;$abcxcount$2$lcssa = $abcxcount$4$lcssa;$abcxnumber$1$lcssa = $abcxnumber$2$lcssa;$bartime$1$lcssa = $bartime$5$lcssa;$ch0$1$ph$lcssa = $ch$1$lcssa;$cnotediv$1$ph$lcssa = $cnotediv$6$lcssa;$cnotelen$1$ph$lcssa = $cnotelen$6$lcssa;$ktempo$1$lcssa = $ktempo$2$lcssa;$mmsp$1$lcssa = $mmsp$3$lcssa;$orderlen$1$lcssa = $orderlen$3$lcssa;$tpd$1$lcssa = $tpd$4$lcssa;
      break;
     } else {
      $$ph1257 = $$lcssa2275;$$ph8921258 = $$lcssa2276;$$ph8931259 = $$lcssa2277;$$ph8941260 = $$lcssa2278;$4048 = $178;$4049 = $176;$abcarpeggio$1$ph1273 = $abcarpeggio$3$lcssa;$abceffect$1$ph1264 = $abceffect$5$lcssa;$abceffoper$1$ph1265 = $abceffoper$4$lcssa;$abcfermata$1$ph1272 = $abcfermata$7$lcssa;$abcgrace$1$ph1280 = $abcgrace$10$lcssa;$abchornpipe$1$ph1278 = $abchornpipe$3$lcssa;$abckey$1$ph1268 = $abckey$5$lcssa;$abcnolegato$1$ph1271 = $abcnolegato$14$lcssa;$abcnoslurs$1$ph1270 = $abcnoslurs$12$lcssa;$abcstate$2$ph1263 = $abcstate$8$lcssa;$abctempo$1$ph1275 = $abctempo$6$lcssa;$abcto$1$ph1274 = $abcto$6$lcssa;$abcvol$1$ph1269 = $abcvol$4$lcssa;$abcxcount$2$ph1266 = $abcxcount$4$lcssa;$abcxnumber$1$ph1267 = $abcxnumber$2$lcssa;$bartime$1$ph1281 = $bartime$5$lcssa;$ch0$1$ph1261 = $ch$1$lcssa;$cnotediv$1$ph1277 = $cnotediv$6$lcssa;$cnotelen$1$ph1276 = $cnotelen$6$lcssa;$ktempo$1$ph1279 = $ktempo$2$lcssa;$mmsp$1$ph1283 = $mmsp$3$lcssa;$orderlen$1$ph1262 = $orderlen$3$lcssa;$tpd$1$ph1282 = $tpd$4$lcssa;
     }
    }
   }
  } while(0);
  $3435 = ($mmsp$1$lcssa|0)==(0); //@line 4991 "load_abc.cpp"
  if ($3435) {
   $abcstate$2$lcssa$lcssa = $abcstate$2$lcssa;$abctempo$1$lcssa$lcssa = $abctempo$1$lcssa;$abcxnumber$1$lcssa$lcssa = $abcxnumber$1$lcssa;$orderlen$1$lcssa$lcssa = $orderlen$1$lcssa;
   break;
  }
  $3436 = HEAP32[$$lcssa948>>2]|0; //@line 4991 "load_abc.cpp"
  _free($3436); //@line 476 "load_abc.cpp"
  $3437 = ($mmsp$1$lcssa|0)>(0); //@line 3906 "load_abc.cpp"
  if ($3437) {
   $4044 = $$ph894$lcssa;$4045 = $$ph893$lcssa;$4046 = $$ph892$lcssa;$4047 = $$lcssa928;$abcarpeggio$01324 = $abcarpeggio$1$ph$lcssa;$abceffect$01315 = $abceffect$1$lcssa;$abceffoper$01316 = $abceffoper$1$lcssa;$abcfermata$01323 = $abcfermata$1$ph$lcssa;$abcgrace$01331 = $abcgrace$1$lcssa;$abchornpipe$01329 = $abchornpipe$1$lcssa;$abckey$01319 = $abckey$1$lcssa;$abcnolegato$01322 = $abcnolegato$1$lcssa;$abcnoslurs$01321 = $abcnoslurs$1$lcssa;$abcstate$11314 = $abcstate$2$lcssa;$abctempo$01326 = $abctempo$1$lcssa;$abcto$01325 = $abcto$1$lcssa;$abcvol$01320 = $abcvol$1$lcssa;$abcxcount$11317 = $abcxcount$2$lcssa;$abcxnumber$01318 = $abcxnumber$1$lcssa;$bartime$01332 = $bartime$1$lcssa;$ch0$01312 = $ch0$1$ph$lcssa;$cnotediv$01328 = $cnotediv$1$ph$lcssa;$cnotelen$01327 = $cnotelen$1$ph$lcssa;$ktempo$01330 = $ktempo$1$lcssa;$mmsp$01334 = $mmsp$1$lcssa;$orderlen$01313 = $orderlen$1$lcssa;$tpd$01333 = $tpd$1$lcssa;
  } else {
   $abcstate$2$lcssa$lcssa = $abcstate$2$lcssa;$abctempo$1$lcssa$lcssa = $abctempo$1$lcssa;$abcxnumber$1$lcssa$lcssa = $abcxnumber$1$lcssa;$orderlen$1$lcssa$lcssa = $orderlen$1$lcssa;
   break;
  }
 }
 __ZL17ABC_CleanupMacrosP10_ABCHANDLE($1); //@line 4993 "load_abc.cpp"
 $3438 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 4994 "load_abc.cpp"
 $3439 = ($3438|0)==(0|0); //@line 4994 "load_abc.cpp"
 if ($3439) {
  HEAP32[$vararg_buffer121>>2] = $abcxnumber$1$lcssa$lcssa; //@line 4996 "load_abc.cpp"
  (_sprintf($buf2,603227,$vararg_buffer121)|0); //@line 4996 "load_abc.cpp"
  $3440 = (_strlen($buf2)|0); //@line 323 "load_abc.cpp"
  $3441 = (($3440) + 20)|0; //@line 323 "load_abc.cpp"
  $3442 = ($3441>>>0)>(255); //@line 323 "load_abc.cpp"
  if (!($3442)) {
   HEAP32[$vararg_buffer124>>2] = $buf2; //@line 324 "load_abc.cpp"
   (_sprintf($txt$i,527600,$vararg_buffer124)|0); //@line 324 "load_abc.cpp"
   HEAP32[$vararg_buffer127>>2] = $txt$i; //@line 328 "load_abc.cpp"
   (_fprintf($155,524747,$vararg_buffer127)|0); //@line 328 "load_abc.cpp"
  }
  $3443 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 4998 "load_abc.cpp"
  $3444 = ($3443|0)==(0|0); //@line 915 "load_abc.cpp"
  if ($3444) {
   $3445 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,0)|0); //@line 916 "load_abc.cpp"
   $3446 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24); //@line 917 "load_abc.cpp"
   $3447 = ((($3445)) + 20|0); //@line 917 "load_abc.cpp"
   HEAP8[$3447>>0]=$3446&255;HEAP8[$3447+1>>0]=($3446>>8)&255;HEAP8[$3447+2>>0]=($3446>>16)&255;HEAP8[$3447+3>>0]=$3446>>24; //@line 917 "load_abc.cpp"
   $$0$i699 = $3445;
  } else {
   $$0$i699 = $3443;
  }
  HEAP8[$156>>0]=$$0$i699&255;HEAP8[$156+1>>0]=($$0$i699>>8)&255;HEAP8[$156+2>>0]=($$0$i699>>16)&255;HEAP8[$156+3>>0]=$$0$i699>>24; //@line 4998 "load_abc.cpp"
  $ttp$21057 = $$0$i699;
 } else {
  $ttp$21057 = $3438;
 }
 $3448 = ($abcstate$2$lcssa$lcssa|0)==(3); //@line 5000 "load_abc.cpp"
 if ($3448) {
  HEAP8[$31>>0]=0&255;HEAP8[$31+1>>0]=(0>>8)&255;HEAP8[$31+2>>0]=(0>>16)&255;HEAP8[$31+3>>0]=0>>24; //@line 3670 "load_abc.cpp"
  $3449 = ($ttp$21057|0)==(0|0); //@line 3671 "load_abc.cpp"
  if ($3449) {
   $3478 = 0;$3485 = 0;
  } else {
   $3456 = 0;$ttp$02$i702 = $ttp$21057;
   while(1) {
    $3450 = ((($ttp$02$i702)) + 8|0); //@line 3672 "load_abc.cpp"
    $3451 = HEAPU8[$3450>>0]|(HEAPU8[$3450+1>>0]<<8)|(HEAPU8[$3450+2>>0]<<16)|(HEAPU8[$3450+3>>0]<<24); //@line 3672 "load_abc.cpp"
    $3452 = ($3451|0)==(0|0); //@line 3672 "load_abc.cpp"
    do {
     if ($3452) {
      $4067 = $3456;
     } else {
      $3453 = ((($3451)) + 4|0); //@line 3672 "load_abc.cpp"
      $3454 = HEAPU8[$3453>>0]|(HEAPU8[$3453+1>>0]<<8)|(HEAPU8[$3453+2>>0]<<16)|(HEAPU8[$3453+3>>0]<<24); //@line 3672 "load_abc.cpp"
      $3455 = ($3454>>>0)>($3456>>>0); //@line 3672 "load_abc.cpp"
      if (!($3455)) {
       $4067 = $3456;
       break;
      }
      HEAP8[$31>>0]=$3454&255;HEAP8[$31+1>>0]=($3454>>8)&255;HEAP8[$31+2>>0]=($3454>>16)&255;HEAP8[$31+3>>0]=$3454>>24; //@line 3673 "load_abc.cpp"
      $4067 = $3454;
     }
    } while(0);
    $ttp$0$i704 = HEAPU8[$ttp$02$i702>>0]|(HEAPU8[$ttp$02$i702+1>>0]<<8)|(HEAPU8[$ttp$02$i702+2>>0]<<16)|(HEAPU8[$ttp$02$i702+3>>0]<<24); //@line 3671 "load_abc.cpp"
    $3457 = ($ttp$0$i704|0)==(0|0); //@line 3671 "load_abc.cpp"
    if ($3457) {
     $$lcssa2124 = $4067;
     break;
    } else {
     $3456 = $4067;$ttp$02$i702 = $ttp$0$i704;
    }
   }
   if ($3449) {
    $3478 = $$lcssa2124;$3485 = 0;
   } else {
    $3465 = $$lcssa2124;$ttp$21058 = $ttp$21057;
    while(1) {
     $3458 = ((($ttp$21058)) + 42|0); //@line 1174 "load_abc.cpp"
     $3459 = HEAP8[$3458>>0]|0; //@line 1174 "load_abc.cpp"
     $3460 = $3459&255; //@line 1174 "load_abc.cpp"
     $3461 = (_pat_gmtosmp($3460)|0); //@line 1174 "load_abc.cpp"
     $3462 = $3461&255; //@line 1174 "load_abc.cpp"
     $3463 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
     HEAP8[$3463>>0]=0&255;HEAP8[$3463+1>>0]=(0>>8)&255;HEAP8[$3463+2>>0]=(0>>16)&255;HEAP8[$3463+3>>0]=0>>24; //@line 540 "load_abc.cpp"
     $3464 = ((($3463)) + 4|0); //@line 541 "load_abc.cpp"
     HEAP8[$3464>>0]=$3465&255;HEAP8[$3464+1>>0]=($3465>>8)&255;HEAP8[$3464+2>>0]=($3465>>16)&255;HEAP8[$3464+3>>0]=$3465>>24; //@line 541 "load_abc.cpp"
     $3466 = ((($3463)) + 8|0); //@line 543 "load_abc.cpp"
     HEAP8[$3466>>0] = 0; //@line 543 "load_abc.cpp"
     $3467 = ((($3463)) + 9|0); //@line 543 "load_abc.cpp"
     HEAP8[$3467>>0] = 0; //@line 543 "load_abc.cpp"
     $3468 = ((($3463)) + 10|0); //@line 543 "load_abc.cpp"
     HEAP8[$3468>>0] = $3462; //@line 543 "load_abc.cpp"
     $3469 = ((($3463)) + 11|0); //@line 543 "load_abc.cpp"
     HEAP8[$3469>>0] = 0; //@line 543 "load_abc.cpp"
     $3470 = ((($3463)) + 12|0); //@line 543 "load_abc.cpp"
     HEAP8[$3470>>0] = 0; //@line 543 "load_abc.cpp"
     $3471 = ((($3463)) + 13|0); //@line 543 "load_abc.cpp"
     HEAP8[$3471>>0] = 0; //@line 543 "load_abc.cpp"
     $3472 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
     $3473 = ((($3463)) + 14|0); //@line 544 "load_abc.cpp"
     HEAP8[$3473>>0] = $3472; //@line 544 "load_abc.cpp"
     $3474 = ((($3463)) + 15|0); //@line 545 "load_abc.cpp"
     HEAP8[$3474>>0] = 0; //@line 545 "load_abc.cpp"
     __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($ttp$21058,$3463); //@line 1179 "load_abc.cpp"
     $ttp$2 = HEAPU8[$ttp$21058>>0]|(HEAPU8[$ttp$21058+1>>0]<<8)|(HEAPU8[$ttp$21058+2>>0]<<16)|(HEAPU8[$ttp$21058+3>>0]<<24); //@line 5003 "load_abc.cpp"
     $3475 = ($ttp$2|0)==(0|0); //@line 5003 "load_abc.cpp"
     if ($3475) {
      break;
     }
     $$pre1537 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 5004 "load_abc.cpp"
     $3465 = $$pre1537;$ttp$21058 = $ttp$2;
    }
    $$pre1538 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 5005 "load_abc.cpp"
    $$pre1539 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 5005 "load_abc.cpp"
    $3478 = $$pre1539;$3485 = $$pre1538;
   }
  }
  $3476 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
  HEAP8[$3476>>0]=0&255;HEAP8[$3476+1>>0]=(0>>8)&255;HEAP8[$3476+2>>0]=(0>>16)&255;HEAP8[$3476+3>>0]=0>>24; //@line 540 "load_abc.cpp"
  $3477 = ((($3476)) + 4|0); //@line 541 "load_abc.cpp"
  HEAP8[$3477>>0]=$3478&255;HEAP8[$3477+1>>0]=($3478>>8)&255;HEAP8[$3477+2>>0]=($3478>>16)&255;HEAP8[$3477+3>>0]=$3478>>24; //@line 541 "load_abc.cpp"
  $3479 = ((($3476)) + 8|0); //@line 543 "load_abc.cpp"
  HEAP8[$3479>>0] = 1; //@line 543 "load_abc.cpp"
  $3480 = ((($3476)) + 9|0); //@line 543 "load_abc.cpp"
  HEAP8[$3480>>0] = 80; //@line 543 "load_abc.cpp"
  $3481 = ((($3476)) + 10|0); //@line 543 "load_abc.cpp"
  HEAP8[$3481>>0]=0&255;HEAP8[$3481+1>>0]=(0>>8)&255;HEAP8[$3481+2>>0]=(0>>16)&255;HEAP8[$3481+3>>0]=0>>24; //@line 543 "load_abc.cpp"
  $3482 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
  $3483 = ((($3476)) + 14|0); //@line 544 "load_abc.cpp"
  HEAP8[$3483>>0] = $3482; //@line 544 "load_abc.cpp"
  $3484 = ((($3476)) + 15|0); //@line 545 "load_abc.cpp"
  HEAP8[$3484>>0] = 0; //@line 545 "load_abc.cpp"
  __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($3485,$3476); //@line 1153 "load_abc.cpp"
  $3486 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 5006 "load_abc.cpp"
  $3487 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 1928 "load_abc.cpp"
  $3488 = ($3487*12288)|0; //@line 339 "load_abc.cpp"
  $3489 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 1929 "load_abc.cpp"
  $3490 = ($3489|0)==(0|0); //@line 1929 "load_abc.cpp"
  do {
   if ($3490) {
    $$0$i723 = 0;$3538 = 0;
   } else {
    $3491 = ((($3489)) + 4|0); //@line 1929 "load_abc.cpp"
    $3492 = HEAPU8[$3491>>0]|(HEAPU8[$3491+1>>0]<<8)|(HEAPU8[$3491+2>>0]<<16)|(HEAPU8[$3491+3>>0]<<24); //@line 1929 "load_abc.cpp"
    $3493 = ($3492|0)==(0|0); //@line 1929 "load_abc.cpp"
    if ($3493) {
     $4068 = 0;
    } else {
     $dt$03$i$i707 = 0;$e$02$i$i708 = $3492;
     while(1) {
      $3494 = ((($e$02$i$i708)) + 4|0); //@line 1931 "load_abc.cpp"
      $3495 = HEAPU8[$3494>>0]|(HEAPU8[$3494+1>>0]<<8)|(HEAPU8[$3494+2>>0]<<16)|(HEAPU8[$3494+3>>0]<<24); //@line 1931 "load_abc.cpp"
      $3496 = ($3495>>>0)>($3486>>>0); //@line 1931 "load_abc.cpp"
      if ($3496) {
       $dt$0$lcssa$i$i711 = $dt$03$i$i707;
       break;
      }
      $3497 = ((($e$02$i$i708)) + 8|0); //@line 1932 "load_abc.cpp"
      $3498 = HEAP8[$3497>>0]|0; //@line 1932 "load_abc.cpp"
      $3499 = ($3498<<24>>24)==(1); //@line 1932 "load_abc.cpp"
      do {
       if ($3499) {
        $3500 = ((($e$02$i$i708)) + 9|0); //@line 1932 "load_abc.cpp"
        $3501 = HEAP8[$3500>>0]|0; //@line 1932 "load_abc.cpp"
        $3502 = ($3501<<24>>24)==(80); //@line 1932 "load_abc.cpp"
        if (!($3502)) {
         $dt$1$i$i710 = $dt$03$i$i707;
         break;
        }
        $3503 = (($3495) + ($dt$03$i$i707))|0; //@line 1933 "load_abc.cpp"
        $3504 = (($3503>>>0) % ($3488>>>0))&-1; //@line 1934 "load_abc.cpp"
        $3505 = ($3504|0)==(0); //@line 1934 "load_abc.cpp"
        if ($3505) {
         $dt$1$i$i710 = $dt$03$i$i707;
         break;
        }
        $3506 = (($3503) + ($3488))|0; //@line 1935 "load_abc.cpp"
        $3507 = (($3506>>>0) % ($3488>>>0))&-1; //@line 1937 "load_abc.cpp"
        $3508 = (($dt$03$i$i707) + ($3488))|0; //@line 1937 "load_abc.cpp"
        $3509 = (($3508) - ($3507))|0; //@line 1938 "load_abc.cpp"
        $dt$1$i$i710 = $3509;
       } else {
        $dt$1$i$i710 = $dt$03$i$i707;
       }
      } while(0);
      $3510 = HEAPU8[$e$02$i$i708>>0]|(HEAPU8[$e$02$i$i708+1>>0]<<8)|(HEAPU8[$e$02$i$i708+2>>0]<<16)|(HEAPU8[$e$02$i$i708+3>>0]<<24); //@line 1931 "load_abc.cpp"
      $3511 = ($3510|0)==(0|0); //@line 1931 "load_abc.cpp"
      if ($3511) {
       $dt$0$lcssa$i$i711 = $dt$1$i$i710;
       break;
      } else {
       $dt$03$i$i707 = $dt$1$i$i710;$e$02$i$i708 = $3510;
      }
     }
     $3512 = (($dt$0$lcssa$i$i711) + ($3486))|0; //@line 1942 "load_abc.cpp"
     $phitmp$i712 = (($3512>>>0) / 192)&-1; //@line 1942 "load_abc.cpp"
     $phitmp1$i713 = $phitmp$i712 >>> 6; //@line 1942 "load_abc.cpp"
     $3513 = (($phitmp1$i713>>>0) / ($3487>>>0))&-1; //@line 1947 "load_abc.cpp"
     if ($3490) {
      $$0$i723 = 0;$3538 = $3513;
      break;
     } else {
      $4068 = $3513;
     }
    }
    $3514 = ((($3489)) + 4|0); //@line 1929 "load_abc.cpp"
    $3515 = HEAPU8[$3514>>0]|(HEAPU8[$3514+1>>0]<<8)|(HEAPU8[$3514+2>>0]<<16)|(HEAPU8[$3514+3>>0]<<24); //@line 1929 "load_abc.cpp"
    $3516 = ($3515|0)==(0|0); //@line 1929 "load_abc.cpp"
    if ($3516) {
     $$0$i723 = 0;$3538 = $4068;
    } else {
     $dt$03$i717 = 0;$e$02$i718 = $3515;
     while(1) {
      $3517 = ((($e$02$i718)) + 4|0); //@line 1931 "load_abc.cpp"
      $3518 = HEAPU8[$3517>>0]|(HEAPU8[$3517+1>>0]<<8)|(HEAPU8[$3517+2>>0]<<16)|(HEAPU8[$3517+3>>0]<<24); //@line 1931 "load_abc.cpp"
      $3519 = ($3518>>>0)>($3486>>>0); //@line 1931 "load_abc.cpp"
      if ($3519) {
       $dt$0$lcssa$i721 = $dt$03$i717;
       break;
      }
      $3520 = ((($e$02$i718)) + 8|0); //@line 1932 "load_abc.cpp"
      $3521 = HEAP8[$3520>>0]|0; //@line 1932 "load_abc.cpp"
      $3522 = ($3521<<24>>24)==(1); //@line 1932 "load_abc.cpp"
      do {
       if ($3522) {
        $3523 = ((($e$02$i718)) + 9|0); //@line 1932 "load_abc.cpp"
        $3524 = HEAP8[$3523>>0]|0; //@line 1932 "load_abc.cpp"
        $3525 = ($3524<<24>>24)==(80); //@line 1932 "load_abc.cpp"
        if (!($3525)) {
         $dt$1$i720 = $dt$03$i717;
         break;
        }
        $3526 = (($3518) + ($dt$03$i717))|0; //@line 1933 "load_abc.cpp"
        $3527 = (($3526>>>0) % ($3488>>>0))&-1; //@line 1934 "load_abc.cpp"
        $3528 = ($3527|0)==(0); //@line 1934 "load_abc.cpp"
        if ($3528) {
         $dt$1$i720 = $dt$03$i717;
         break;
        }
        $3529 = (($3526) + ($3488))|0; //@line 1935 "load_abc.cpp"
        $3530 = (($3529>>>0) % ($3488>>>0))&-1; //@line 1937 "load_abc.cpp"
        $3531 = (($dt$03$i717) + ($3488))|0; //@line 1937 "load_abc.cpp"
        $3532 = (($3531) - ($3530))|0; //@line 1938 "load_abc.cpp"
        $dt$1$i720 = $3532;
       } else {
        $dt$1$i720 = $dt$03$i717;
       }
      } while(0);
      $3533 = HEAPU8[$e$02$i718>>0]|(HEAPU8[$e$02$i718+1>>0]<<8)|(HEAPU8[$e$02$i718+2>>0]<<16)|(HEAPU8[$e$02$i718+3>>0]<<24); //@line 1931 "load_abc.cpp"
      $3534 = ($3533|0)==(0|0); //@line 1931 "load_abc.cpp"
      if ($3534) {
       $dt$0$lcssa$i721 = $dt$1$i720;
       break;
      } else {
       $dt$03$i717 = $dt$1$i720;$e$02$i718 = $3533;
      }
     }
     $3535 = (($dt$0$lcssa$i721) + ($3486))|0; //@line 1942 "load_abc.cpp"
     $$0$i723 = $3535;$3538 = $4068;
    }
   }
  } while(0);
  $3536 = (($$0$i723>>>0) % ($3488>>>0))&-1; //@line 5007 "load_abc.cpp"
  $not$ = ($3536|0)!=(0); //@line 5007 "load_abc.cpp"
  $3537 = $not$&1; //@line 5007 "load_abc.cpp"
  $$98 = (($3537) + ($3538))|0; //@line 5007 "load_abc.cpp"
  $3539 = HEAP8[1278234]|0; //@line 5008 "load_abc.cpp"
  $3540 = ($3539<<24>>24)==(32); //@line 5008 "load_abc.cpp"
  $3541 = $$98&255; //@line 5009 "load_abc.cpp"
  L1739: do {
   if ($3540) {
    HEAP8[$101>>0] = $3541; //@line 5009 "load_abc.cpp"
    $3542 = HEAP32[$abcparts>>2]|0; //@line 5010 "load_abc.cpp"
    $3543 = ($3542|0)==(0|0); //@line 5010 "load_abc.cpp"
    if ($3543) {
     label = 1167;
    } else {
     $t$141056 = 0;
     while(1) {
      $3544 = (($partpat) + ($t$141056<<1)|0); //@line 5012 "load_abc.cpp"
      $3545 = HEAP8[$3544>>0]|0; //@line 5012 "load_abc.cpp"
      $3546 = (((($partpat) + ($t$141056<<1)|0)) + 1|0); //@line 5012 "load_abc.cpp"
      $3547 = HEAP8[$3546>>0]|0; //@line 5012 "load_abc.cpp"
      $3548 = ($3545&255)<($3547&255); //@line 5012 "load_abc.cpp"
      if ($3548) {
       $$pr777 = $3542;
       label = 1166;
       break L1739;
      }
      $3549 = (($t$141056) + 1)|0; //@line 5011 "load_abc.cpp"
      $3550 = ($3549>>>0)<(26); //@line 5011 "load_abc.cpp"
      if ($3550) {
       $t$141056 = $3549;
      } else {
       $$lcssa2123 = $3549;
       break;
      }
     }
     $3551 = ($$lcssa2123|0)==(26); //@line 5013 "load_abc.cpp"
     if (!($3551)) {
      $$pr777 = $3542;
      label = 1166;
      break;
     }
     $3552 = (_strlen($3542)|0); //@line 323 "load_abc.cpp"
     $3553 = (($3552) + 27)|0; //@line 323 "load_abc.cpp"
     $3554 = ($3553>>>0)>(255); //@line 323 "load_abc.cpp"
     if ($3554) {
      $3556 = $3542;
     } else {
      HEAP32[$vararg_buffer130>>2] = $3542; //@line 324 "load_abc.cpp"
      (_sprintf($txt$i,526650,$vararg_buffer130)|0); //@line 324 "load_abc.cpp"
      HEAP32[$vararg_buffer133>>2] = $txt$i; //@line 328 "load_abc.cpp"
      (_fprintf($155,524747,$vararg_buffer133)|0); //@line 328 "load_abc.cpp"
      $$pre1540 = HEAP32[$abcparts>>2]|0; //@line 1820 "load_abc.cpp"
      $3556 = $$pre1540;
     }
     $3555 = ($3556|0)==(0|0); //@line 1820 "load_abc.cpp"
     if (!($3555)) {
      _free($3556); //@line 1820 "load_abc.cpp"
     }
     HEAP32[$abcparts>>2] = 0; //@line 1822 "load_abc.cpp"
     label = 1167;
    }
   } else {
    $3557 = $3539 << 24 >> 24; //@line 5008 "load_abc.cpp"
    $3558 = (($3557) + -65)|0; //@line 5020 "load_abc.cpp"
    $3559 = (((($partpat) + ($3558<<1)|0)) + 1|0); //@line 5020 "load_abc.cpp"
    HEAP8[$3559>>0] = $3541; //@line 5020 "load_abc.cpp"
    $$pr777$pre = HEAP32[$abcparts>>2]|0; //@line 5021 "load_abc.cpp"
    $$pr777 = $$pr777$pre;
    label = 1166;
   }
  } while(0);
  if ((label|0) == 1166) {
   $3560 = ($$pr777|0)==(0|0); //@line 5021 "load_abc.cpp"
   if ($3560) {
    label = 1167;
   } else {
    $3561 = $$pr777;
   }
  }
  if ((label|0) == 1167) {
   __ZL17abc_song_to_partsP10_ABCHANDLEPPcPA2_h($1,$abcparts,$partpat); //@line 5021 "load_abc.cpp"
   $$pre1542 = HEAP32[$abcparts>>2]|0; //@line 5022 "load_abc.cpp"
   $3561 = $$pre1542;
  }
  $3562 = (__ZL24abc_partpat_to_orderlistPA2_hPKcP10_ABCHANDLEPPhi($partpat,$3561,$orderlist,$orderlen$1$lcssa$lcssa)|0); //@line 5022 "load_abc.cpp"
  $$pre1543 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 1099 "load_abc.cpp"
  $3564 = $$pre1543;$orderlen$4783 = $3562;
 } else {
  $3564 = $ttp$21057;$orderlen$4783 = $orderlen$1$lcssa$lcssa;
 }
 $3563 = ($3564|0)==(0|0); //@line 1099 "load_abc.cpp"
 if ($3563) {
  label = 1220;
 } else {
  $ptp$07$i$i = 0;$tp$08$i$i = $3564;
  L1760: while(1) {
   $3565 = ((($tp$08$i$i)) + 4|0); //@line 995 "load_abc.cpp"
   $ep$01$i$i = HEAPU8[$3565>>0]|(HEAPU8[$3565+1>>0]<<8)|(HEAPU8[$3565+2>>0]<<16)|(HEAPU8[$3565+3>>0]<<24); //@line 995 "load_abc.cpp"
   $3566 = ($ep$01$i$i|0)==(0|0); //@line 998 "load_abc.cpp"
   if ($3566) {
    label = 1188;
   } else {
    $3567 = ((($tp$08$i$i)) + 8|0); //@line 1021 "load_abc.cpp"
    $ct$03$i$i = 0;$d$04$i$i = -3;$el$02$i$i = 0;$ep$05$i$i = $ep$01$i$i;
    while(1) {
     $3568 = ((($ep$05$i$i)) + 4|0); //@line 999 "load_abc.cpp"
     $3569 = HEAPU8[$3568>>0]|(HEAPU8[$3568+1>>0]<<8)|(HEAPU8[$3568+2>>0]<<16)|(HEAPU8[$3568+3>>0]<<24); //@line 999 "load_abc.cpp"
     $3570 = $3569 << 3; //@line 1000 "load_abc.cpp"
     $3571 = $3570 | 4; //@line 1001 "load_abc.cpp"
     HEAP8[$3568>>0]=$3571&255;HEAP8[$3568+1>>0]=($3571>>8)&255;HEAP8[$3568+2>>0]=($3571>>16)&255;HEAP8[$3568+3>>0]=$3571>>24; //@line 1001 "load_abc.cpp"
     $3572 = ((($ep$05$i$i)) + 8|0); //@line 1002 "load_abc.cpp"
     $3573 = HEAP8[$3572>>0]|0; //@line 1002 "load_abc.cpp"
     $3574 = ($3573<<24>>24)==(1); //@line 1002 "load_abc.cpp"
     L1766: do {
      if ($3574) {
       $3575 = (($3571) + ($d$04$i$i))|0; //@line 1003 "load_abc.cpp"
       HEAP8[$3568>>0]=$3575&255;HEAP8[$3568+1>>0]=($3575>>8)&255;HEAP8[$3568+2>>0]=($3575>>16)&255;HEAP8[$3568+3>>0]=$3575>>24; //@line 1003 "load_abc.cpp"
       $3576 = (($d$04$i$i) + 1)|0; //@line 1004 "load_abc.cpp"
       $3577 = ($3576|0)==(0); //@line 1005 "load_abc.cpp"
       $$$i$i = $3577 ? -1 : $3576; //@line 1005 "load_abc.cpp"
       $3578 = ($$$i$i|0)==(4); //@line 1006 "load_abc.cpp"
       $$$$i$i = $3578 ? 3 : $$$i$i; //@line 1006 "load_abc.cpp"
       $3579 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 1007 "load_abc.cpp"
       $3580 = ($tp$08$i$i|0)==($3579|0); //@line 1007 "load_abc.cpp"
       $$phi$trans$insert$i$i = ((($ep$05$i$i)) + 9|0); //@line 1100 "load_abc.cpp"
       if ($3580) {
        $$pre$i$i = HEAP8[$$phi$trans$insert$i$i>>0]|0; //@line 1008 "load_abc.cpp"
        $3582 = $$pre$i$i;
       } else {
        HEAP8[$$phi$trans$insert$i$i>>0] = 72; //@line 1007 "load_abc.cpp"
        $3582 = 72;
       }
       $3581 = $3582&255; //@line 1008 "load_abc.cpp"
       switch ($3581|0) {
       case 83: case 72:  {
        break;
       }
       default: {
        $d$1$i$i = $$$$i$i;$el$1$i$i = $ep$05$i$i;$ep$1$in$i$i = $ep$05$i$i;
        break L1766;
       }
       }
       $3583 = ($el$02$i$i|0)==(0|0); //@line 1011 "load_abc.cpp"
       $3584 = HEAPU8[$ep$05$i$i>>0]|(HEAPU8[$ep$05$i$i+1>>0]<<8)|(HEAPU8[$ep$05$i$i+2>>0]<<16)|(HEAPU8[$ep$05$i$i+3>>0]<<24); //@line 1012 "load_abc.cpp"
       if ($3583) {
        HEAP8[$3565>>0]=$3584&255;HEAP8[$3565+1>>0]=($3584>>8)&255;HEAP8[$3565+2>>0]=($3584>>16)&255;HEAP8[$3565+3>>0]=$3584>>24; //@line 1019 "load_abc.cpp"
        $3586 = ($3584|0)==(0|0); //@line 1020 "load_abc.cpp"
        if ($3586) {
         HEAP8[$3567>>0]=0&255;HEAP8[$3567+1>>0]=(0>>8)&255;HEAP8[$3567+2>>0]=(0>>16)&255;HEAP8[$3567+3>>0]=0>>24; //@line 1021 "load_abc.cpp"
        }
        _free($ep$05$i$i); //@line 1022 "load_abc.cpp"
        $d$1$i$i = $$$$i$i;$el$1$i$i = 0;$ep$1$in$i$i = $3565;
        break;
       } else {
        HEAP8[$el$02$i$i>>0]=$3584&255;HEAP8[$el$02$i$i+1>>0]=($3584>>8)&255;HEAP8[$el$02$i$i+2>>0]=($3584>>16)&255;HEAP8[$el$02$i$i+3>>0]=$3584>>24; //@line 1012 "load_abc.cpp"
        $3585 = ($3584|0)==(0|0); //@line 1013 "load_abc.cpp"
        if ($3585) {
         HEAP8[$3567>>0]=$el$02$i$i&255;HEAP8[$3567+1>>0]=($el$02$i$i>>8)&255;HEAP8[$3567+2>>0]=($el$02$i$i>>16)&255;HEAP8[$3567+3>>0]=$el$02$i$i>>24; //@line 1014 "load_abc.cpp"
        }
        _free($ep$05$i$i); //@line 1015 "load_abc.cpp"
        $d$1$i$i = $$$$i$i;$el$1$i$i = $el$02$i$i;$ep$1$in$i$i = $el$02$i$i;
        break;
       }
      } else {
       $d$1$i$i = 1;$el$1$i$i = $ep$05$i$i;$ep$1$in$i$i = $ep$05$i$i;
      }
     } while(0);
     $3587 = ($3569>>>0)>($ct$03$i$i>>>0); //@line 1037 "load_abc.cpp"
     $$d$1$i$i = $3587 ? -3 : $d$1$i$i; //@line 1037 "load_abc.cpp"
     $ep$0$i$i = HEAPU8[$ep$1$in$i$i>>0]|(HEAPU8[$ep$1$in$i$i+1>>0]<<8)|(HEAPU8[$ep$1$in$i$i+2>>0]<<16)|(HEAPU8[$ep$1$in$i$i+3>>0]<<24); //@line 995 "load_abc.cpp"
     $3588 = ($ep$0$i$i|0)==(0|0); //@line 998 "load_abc.cpp"
     if ($3588) {
      break;
     } else {
      $ct$03$i$i = $3569;$d$04$i$i = $$d$1$i$i;$el$02$i$i = $el$1$i$i;$ep$05$i$i = $ep$0$i$i;
     }
    }
    $$pre15$i$i = HEAPU8[$3565>>0]|(HEAPU8[$3565+1>>0]<<8)|(HEAPU8[$3565+2>>0]<<16)|(HEAPU8[$3565+3>>0]<<24); //@line 1041 "load_abc.cpp"
    $3589 = ($$pre15$i$i|0)==(0|0); //@line 1041 "load_abc.cpp"
    if ($3589) {
     label = 1188;
    } else {
     $tp$1$i$i = $tp$08$i$i;
    }
   }
   do {
    if ((label|0) == 1188) {
     label = 0;
     $3590 = ($ptp$07$i$i|0)==(0|0); //@line 1042 "load_abc.cpp"
     $3591 = HEAPU8[$tp$08$i$i>>0]|(HEAPU8[$tp$08$i$i+1>>0]<<8)|(HEAPU8[$tp$08$i$i+2>>0]<<16)|(HEAPU8[$tp$08$i$i+3>>0]<<24); //@line 1043 "load_abc.cpp"
     if (!($3590)) {
      HEAP8[$ptp$07$i$i>>0]=$3591&255;HEAP8[$ptp$07$i$i+1>>0]=($3591>>8)&255;HEAP8[$ptp$07$i$i+2>>0]=($3591>>16)&255;HEAP8[$ptp$07$i$i+3>>0]=$3591>>24; //@line 1043 "load_abc.cpp"
      _free($tp$08$i$i); //@line 1044 "load_abc.cpp"
      $tp$1$i$i = $ptp$07$i$i;
      break;
     }
     $3592 = ($3591|0)==(0|0); //@line 1047 "load_abc.cpp"
     if ($3592) {
      break L1760;
     }
     HEAP8[$156>>0]=$3591&255;HEAP8[$156+1>>0]=($3591>>8)&255;HEAP8[$156+2>>0]=($3591>>16)&255;HEAP8[$156+3>>0]=$3591>>24; //@line 1048 "load_abc.cpp"
     _free($tp$08$i$i); //@line 1049 "load_abc.cpp"
     $3593 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 1050 "load_abc.cpp"
     $tp$1$i$i = $3593;
    }
   } while(0);
   $tp$0$i$i = HEAPU8[$tp$1$i$i>>0]|(HEAPU8[$tp$1$i$i+1>>0]<<8)|(HEAPU8[$tp$1$i$i+2>>0]<<16)|(HEAPU8[$tp$1$i$i+3>>0]<<24); //@line 993 "load_abc.cpp"
   $3594 = ($tp$0$i$i|0)==(0|0); //@line 993 "load_abc.cpp"
   if ($3594) {
    break;
   } else {
    $ptp$07$i$i = $tp$1$i$i;$tp$08$i$i = $tp$0$i$i;
   }
  }
  $3595 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 1101 "load_abc.cpp"
  $tp$015$i = HEAPU8[$3595>>0]|(HEAPU8[$3595+1>>0]<<8)|(HEAPU8[$3595+2>>0]<<16)|(HEAPU8[$3595+3>>0]<<24); //@line 1101 "load_abc.cpp"
  $3596 = ($tp$015$i|0)==(0|0); //@line 1101 "load_abc.cpp"
  if ($3596) {
   $tp$08$i1$i = $3595;
  } else {
   $3597 = ((($3595)) + 4|0); //@line 1102 "load_abc.cpp"
   $tp$016$i = $tp$015$i;
   while(1) {
    $em$011$i = HEAPU8[$3597>>0]|(HEAPU8[$3597+1>>0]<<8)|(HEAPU8[$3597+2>>0]<<16)|(HEAPU8[$3597+3>>0]<<24); //@line 1102 "load_abc.cpp"
    $3599 = ($em$011$i|0)==(0|0); //@line 1102 "load_abc.cpp"
    if (!($3599)) {
     $3600 = ((($tp$016$i)) + 4|0); //@line 1113 "load_abc.cpp"
     $em$012$i = $em$011$i;
     while(1) {
      $3601 = ((($em$012$i)) + 8|0); //@line 1103 "load_abc.cpp"
      $3602 = HEAP8[$3601>>0]|0; //@line 1103 "load_abc.cpp"
      $3603 = ($3602<<24>>24)==(1); //@line 1103 "load_abc.cpp"
      L1803: do {
       if ($3603) {
        $3604 = ((($em$012$i)) + 9|0); //@line 1104 "load_abc.cpp"
        $3605 = HEAP8[$3604>>0]|0; //@line 1104 "load_abc.cpp"
        $3606 = $3605&255; //@line 1104 "load_abc.cpp"
        switch ($3606|0) {
        case 83: case 84: case 72: case 67:  {
         break L1803;
         break;
        }
        default: {
        }
        }
        $3607 = (_calloc(1,16)|0); //@line 554 "load_abc.cpp"
        HEAP8[$3607>>0]=0&255;HEAP8[$3607+1>>0]=(0>>8)&255;HEAP8[$3607+2>>0]=(0>>16)&255;HEAP8[$3607+3>>0]=0>>24; //@line 555 "load_abc.cpp"
        $3608 = ((($em$012$i)) + 4|0); //@line 556 "load_abc.cpp"
        $3609 = HEAPU8[$3608>>0]|(HEAPU8[$3608+1>>0]<<8)|(HEAPU8[$3608+2>>0]<<16)|(HEAPU8[$3608+3>>0]<<24); //@line 556 "load_abc.cpp"
        $3610 = ((($3607)) + 4|0); //@line 556 "load_abc.cpp"
        HEAP8[$3610>>0]=$3609&255;HEAP8[$3610+1>>0]=($3609>>8)&255;HEAP8[$3610+2>>0]=($3609>>16)&255;HEAP8[$3610+3>>0]=$3609>>24; //@line 556 "load_abc.cpp"
        $3611 = HEAP8[$3601>>0]|0; //@line 557 "load_abc.cpp"
        $3612 = ((($3607)) + 8|0); //@line 557 "load_abc.cpp"
        HEAP8[$3612>>0] = $3611; //@line 557 "load_abc.cpp"
        $3613 = HEAP8[$3604>>0]|0; //@line 558 "load_abc.cpp"
        $3614 = ((($3607)) + 9|0); //@line 558 "load_abc.cpp"
        HEAP8[$3614>>0] = $3613; //@line 558 "load_abc.cpp"
        $3615 = ((($em$012$i)) + 10|0); //@line 559 "load_abc.cpp"
        $3616 = HEAPU8[$3615>>0]|(HEAPU8[$3615+1>>0]<<8)|(HEAPU8[$3615+2>>0]<<16)|(HEAPU8[$3615+3>>0]<<24); //@line 559 "load_abc.cpp"
        $3617 = ((($3607)) + 10|0); //@line 559 "load_abc.cpp"
        HEAP8[$3617>>0]=$3616&255;HEAP8[$3617+1>>0]=($3616>>8)&255;HEAP8[$3617+2>>0]=($3616>>16)&255;HEAP8[$3617+3>>0]=$3616>>24; //@line 559 "load_abc.cpp"
        $3618 = ((($em$012$i)) + 14|0); //@line 560 "load_abc.cpp"
        $3619 = HEAP8[$3618>>0]|0; //@line 560 "load_abc.cpp"
        $3620 = ((($3607)) + 14|0); //@line 560 "load_abc.cpp"
        HEAP8[$3620>>0] = $3619; //@line 560 "load_abc.cpp"
        $3621 = HEAPU8[$3608>>0]|(HEAPU8[$3608+1>>0]<<8)|(HEAPU8[$3608+2>>0]<<16)|(HEAPU8[$3608+3>>0]<<24); //@line 1112 "load_abc.cpp"
        $3622 = HEAPU8[$3600>>0]|(HEAPU8[$3600+1>>0]<<8)|(HEAPU8[$3600+2>>0]<<16)|(HEAPU8[$3600+3>>0]<<24); //@line 1113 "load_abc.cpp"
        $3623 = ($3622|0)==(0|0); //@line 1114 "load_abc.cpp"
        do {
         if ($3623) {
          $es$0$lcssa22$i = 0;
         } else {
          $es$07$i = $3622;$et$08$i = $3622;
          while(1) {
           $3624 = ((($et$08$i)) + 4|0); //@line 1114 "load_abc.cpp"
           $3625 = HEAPU8[$3624>>0]|(HEAPU8[$3624+1>>0]<<8)|(HEAPU8[$3624+2>>0]<<16)|(HEAPU8[$3624+3>>0]<<24); //@line 1114 "load_abc.cpp"
           $3626 = ($3625>>>0)>($3621>>>0); //@line 1114 "load_abc.cpp"
           if ($3626) {
            $es$0$lcssa$ph$i = $es$07$i;
            break;
           }
           $3627 = HEAPU8[$et$08$i>>0]|(HEAPU8[$et$08$i+1>>0]<<8)|(HEAPU8[$et$08$i+2>>0]<<16)|(HEAPU8[$et$08$i+3>>0]<<24); //@line 1114 "load_abc.cpp"
           $3628 = ($3627|0)==(0|0); //@line 1114 "load_abc.cpp"
           if ($3628) {
            $es$0$lcssa$ph$i = $et$08$i;
            break;
           } else {
            $es$07$i$phi = $et$08$i;$et$08$i = $3627;$es$07$i = $es$07$i$phi;
           }
          }
          $3629 = ((($es$0$lcssa$ph$i)) + 4|0); //@line 1116 "load_abc.cpp"
          $3630 = HEAPU8[$3629>>0]|(HEAPU8[$3629+1>>0]<<8)|(HEAPU8[$3629+2>>0]<<16)|(HEAPU8[$3629+3>>0]<<24); //@line 1116 "load_abc.cpp"
          $3631 = ($3630>>>0)>($3621>>>0); //@line 1116 "load_abc.cpp"
          if ($3631) {
           $es$0$lcssa22$i = $es$0$lcssa$ph$i;
           break;
          }
          $3632 = HEAPU8[$es$0$lcssa$ph$i>>0]|(HEAPU8[$es$0$lcssa$ph$i+1>>0]<<8)|(HEAPU8[$es$0$lcssa$ph$i+2>>0]<<16)|(HEAPU8[$es$0$lcssa$ph$i+3>>0]<<24); //@line 1121 "load_abc.cpp"
          HEAP8[$3607>>0]=$3632&255;HEAP8[$3607+1>>0]=($3632>>8)&255;HEAP8[$3607+2>>0]=($3632>>16)&255;HEAP8[$3607+3>>0]=$3632>>24; //@line 1121 "load_abc.cpp"
          HEAP8[$es$0$lcssa$ph$i>>0]=$3607&255;HEAP8[$es$0$lcssa$ph$i+1>>0]=($3607>>8)&255;HEAP8[$es$0$lcssa$ph$i+2>>0]=($3607>>16)&255;HEAP8[$es$0$lcssa$ph$i+3>>0]=$3607>>24; //@line 1122 "load_abc.cpp"
          break L1803;
         }
        } while(0);
        HEAP8[$3607>>0]=$es$0$lcssa22$i&255;HEAP8[$3607+1>>0]=($es$0$lcssa22$i>>8)&255;HEAP8[$3607+2>>0]=($es$0$lcssa22$i>>16)&255;HEAP8[$3607+3>>0]=$es$0$lcssa22$i>>24; //@line 1117 "load_abc.cpp"
        HEAP8[$3600>>0]=$3607&255;HEAP8[$3600+1>>0]=($3607>>8)&255;HEAP8[$3600+2>>0]=($3607>>16)&255;HEAP8[$3600+3>>0]=$3607>>24; //@line 1118 "load_abc.cpp"
       }
      } while(0);
      $em$0$i = HEAPU8[$em$012$i>>0]|(HEAPU8[$em$012$i+1>>0]<<8)|(HEAPU8[$em$012$i+2>>0]<<16)|(HEAPU8[$em$012$i+3>>0]<<24); //@line 1102 "load_abc.cpp"
      $3633 = ($em$0$i|0)==(0|0); //@line 1102 "load_abc.cpp"
      if ($3633) {
       break;
      } else {
       $em$012$i = $em$0$i;
      }
     }
    }
    $tp$0$i729 = HEAPU8[$tp$016$i>>0]|(HEAPU8[$tp$016$i+1>>0]<<8)|(HEAPU8[$tp$016$i+2>>0]<<16)|(HEAPU8[$tp$016$i+3>>0]<<24); //@line 1101 "load_abc.cpp"
    $3598 = ($tp$0$i729|0)==(0|0); //@line 1101 "load_abc.cpp"
    if ($3598) {
     break;
    } else {
     $tp$016$i = $tp$0$i729;
    }
   }
   $tp$08$i1$pre$i = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 1065 "load_abc.cpp"
   $tp$08$i1$i = $tp$08$i1$pre$i;
  }
  $3634 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 1064 "load_abc.cpp"
  $3635 = ($3634*12288)|0; //@line 339 "load_abc.cpp"
  $3636 = ($tp$08$i1$i|0)==(0|0); //@line 1065 "load_abc.cpp"
  if ($3636) {
   label = 1220;
  } else {
   $3637 = Math_imul($3634, -192)|0; //@line 1083 "load_abc.cpp"
   $tp$09$i$i = $tp$08$i1$i;
   while(1) {
    $3638 = ((($tp$09$i$i)) + 4|0); //@line 1068 "load_abc.cpp"
    $ep$01$i2$i = HEAPU8[$3638>>0]|(HEAPU8[$3638+1>>0]<<8)|(HEAPU8[$3638+2>>0]<<16)|(HEAPU8[$3638+3>>0]<<24); //@line 1068 "load_abc.cpp"
    $3639 = ($ep$01$i2$i|0)==(0|0); //@line 1068 "load_abc.cpp"
    do {
     if (!($3639)) {
      $ep$03$i$i = $ep$01$i2$i;$tt$02$i$i = 0;
      while(1) {
       $3641 = ((($ep$03$i$i)) + 4|0); //@line 1069 "load_abc.cpp"
       $3642 = HEAPU8[$3641>>0]|(HEAPU8[$3641+1>>0]<<8)|(HEAPU8[$3641+2>>0]<<16)|(HEAPU8[$3641+3>>0]<<24); //@line 1069 "load_abc.cpp"
       $3643 = $3642 >>> 3; //@line 1069 "load_abc.cpp"
       $3644 = (($3643) - ($tt$02$i$i))|0; //@line 1070 "load_abc.cpp"
       HEAP8[$3641>>0]=$3644&255;HEAP8[$3641+1>>0]=($3644>>8)&255;HEAP8[$3641+2>>0]=($3644>>16)&255;HEAP8[$3641+3>>0]=$3644>>24; //@line 1070 "load_abc.cpp"
       $ep$0$i4$i = HEAPU8[$ep$03$i$i>>0]|(HEAPU8[$ep$03$i$i+1>>0]<<8)|(HEAPU8[$ep$03$i$i+2>>0]<<16)|(HEAPU8[$ep$03$i$i+3>>0]<<24); //@line 1068 "load_abc.cpp"
       $3645 = ($ep$0$i4$i|0)==(0|0); //@line 1068 "load_abc.cpp"
       if ($3645) {
        break;
       } else {
        $ep$03$i$i = $ep$0$i4$i;$tt$02$i$i = $3643;
       }
      }
      $ep$14$pre$i$i = HEAPU8[$3638>>0]|(HEAPU8[$3638+1>>0]<<8)|(HEAPU8[$3638+2>>0]<<16)|(HEAPU8[$3638+3>>0]<<24); //@line 1075 "load_abc.cpp"
      $3640 = ($ep$14$pre$i$i|0)==(0|0); //@line 1075 "load_abc.cpp"
      if ($3640) {
       break;
      } else {
       $ep$16$i$i = $ep$14$pre$i$i;$tt$15$i$i = 0;
      }
      while(1) {
       $3646 = ((($ep$16$i$i)) + 4|0); //@line 1076 "load_abc.cpp"
       $3647 = HEAPU8[$3646>>0]|(HEAPU8[$3646+1>>0]<<8)|(HEAPU8[$3646+2>>0]<<16)|(HEAPU8[$3646+3>>0]<<24); //@line 1076 "load_abc.cpp"
       $3648 = (($3647) + ($tt$15$i$i))|0; //@line 1076 "load_abc.cpp"
       HEAP8[$3646>>0]=$3648&255;HEAP8[$3646+1>>0]=($3648>>8)&255;HEAP8[$3646+2>>0]=($3648>>16)&255;HEAP8[$3646+3>>0]=$3648>>24; //@line 1076 "load_abc.cpp"
       $3649 = ((($ep$16$i$i)) + 8|0); //@line 1078 "load_abc.cpp"
       $3650 = HEAP8[$3649>>0]|0; //@line 1078 "load_abc.cpp"
       $3651 = ($3650<<24>>24)==(1); //@line 1078 "load_abc.cpp"
       do {
        if ($3651) {
         $3652 = ((($ep$16$i$i)) + 9|0); //@line 1078 "load_abc.cpp"
         $3653 = HEAP8[$3652>>0]|0; //@line 1078 "load_abc.cpp"
         $3654 = ($3653<<24>>24)==(80); //@line 1078 "load_abc.cpp"
         if (!($3654)) {
          $tt$2$i$i = $3648;
          break;
         }
         $3655 = (($3648>>>0) % ($3635>>>0))&-1; //@line 1079 "load_abc.cpp"
         $3656 = ($3655|0)==(0); //@line 1079 "load_abc.cpp"
         if ($3656) {
          HEAP8[$3652>>0] = 72; //@line 1085 "load_abc.cpp"
          $tt$2$i$i = $3648;
          break;
         } else {
          $3657 = (($3648) + ($3635))|0; //@line 1080 "load_abc.cpp"
          $3658 = (($3657>>>0) % ($3635>>>0))&-1; //@line 1082 "load_abc.cpp"
          $3659 = (($3657) - ($3658))|0; //@line 1082 "load_abc.cpp"
          $3660 = (($3648) + ($3637))|0; //@line 1083 "load_abc.cpp"
          HEAP8[$3646>>0]=$3660&255;HEAP8[$3646+1>>0]=($3660>>8)&255;HEAP8[$3646+2>>0]=($3660>>16)&255;HEAP8[$3646+3>>0]=$3660>>24; //@line 1083 "load_abc.cpp"
          $tt$2$i$i = $3659;
          break;
         }
        } else {
         $tt$2$i$i = $3648;
        }
       } while(0);
       $ep$1$i$i = HEAPU8[$ep$16$i$i>>0]|(HEAPU8[$ep$16$i$i+1>>0]<<8)|(HEAPU8[$ep$16$i$i+2>>0]<<16)|(HEAPU8[$ep$16$i$i+3>>0]<<24); //@line 1075 "load_abc.cpp"
       $3661 = ($ep$1$i$i|0)==(0|0); //@line 1075 "load_abc.cpp"
       if ($3661) {
        break;
       } else {
        $ep$16$i$i = $ep$1$i$i;$tt$15$i$i = $tt$2$i$i;
       }
      }
     }
    } while(0);
    $tp$0$i6$i = HEAPU8[$tp$09$i$i>>0]|(HEAPU8[$tp$09$i$i+1>>0]<<8)|(HEAPU8[$tp$09$i$i+2>>0]<<16)|(HEAPU8[$tp$09$i$i+3>>0]<<24); //@line 1065 "load_abc.cpp"
    $3662 = ($tp$0$i6$i|0)==(0|0); //@line 1065 "load_abc.cpp"
    if ($3662) {
     break;
    } else {
     $tp$09$i$i = $tp$0$i6$i;
    }
   }
   $ttp$01$i733$pre = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 3671 "load_abc.cpp"
   HEAP8[$31>>0]=0&255;HEAP8[$31+1>>0]=(0>>8)&255;HEAP8[$31+2>>0]=(0>>16)&255;HEAP8[$31+3>>0]=0>>24; //@line 3670 "load_abc.cpp"
   $3663 = ($ttp$01$i733$pre|0)==(0|0); //@line 3671 "load_abc.cpp"
   if (!($3663)) {
    $3670 = 0;$ttp$02$i734 = $ttp$01$i733$pre;
    while(1) {
     $3664 = ((($ttp$02$i734)) + 8|0); //@line 3672 "load_abc.cpp"
     $3665 = HEAPU8[$3664>>0]|(HEAPU8[$3664+1>>0]<<8)|(HEAPU8[$3664+2>>0]<<16)|(HEAPU8[$3664+3>>0]<<24); //@line 3672 "load_abc.cpp"
     $3666 = ($3665|0)==(0|0); //@line 3672 "load_abc.cpp"
     do {
      if ($3666) {
       $4069 = $3670;
      } else {
       $3667 = ((($3665)) + 4|0); //@line 3672 "load_abc.cpp"
       $3668 = HEAPU8[$3667>>0]|(HEAPU8[$3667+1>>0]<<8)|(HEAPU8[$3667+2>>0]<<16)|(HEAPU8[$3667+3>>0]<<24); //@line 3672 "load_abc.cpp"
       $3669 = ($3668>>>0)>($3670>>>0); //@line 3672 "load_abc.cpp"
       if (!($3669)) {
        $4069 = $3670;
        break;
       }
       HEAP8[$31>>0]=$3668&255;HEAP8[$31+1>>0]=($3668>>8)&255;HEAP8[$31+2>>0]=($3668>>16)&255;HEAP8[$31+3>>0]=$3668>>24; //@line 3673 "load_abc.cpp"
       $4069 = $3668;
      }
     } while(0);
     $ttp$0$i736 = HEAPU8[$ttp$02$i734>>0]|(HEAPU8[$ttp$02$i734+1>>0]<<8)|(HEAPU8[$ttp$02$i734+2>>0]<<16)|(HEAPU8[$ttp$02$i734+3>>0]<<24); //@line 3671 "load_abc.cpp"
     $3671 = ($ttp$0$i736|0)==(0|0); //@line 3671 "load_abc.cpp"
     if ($3671) {
      break;
     } else {
      $3670 = $4069;$ttp$02$i734 = $ttp$0$i736;
     }
    }
   }
  }
 }
 if ((label|0) == 1220) {
  HEAP8[$31>>0]=0&255;HEAP8[$31+1>>0]=(0>>8)&255;HEAP8[$31+2>>0]=(0>>16)&255;HEAP8[$31+3>>0]=0>>24; //@line 3670 "load_abc.cpp"
 }
 $3672 = (_getenv((527621|0))|0); //@line 5041 "load_abc.cpp"
 $3673 = ($3672|0)==(0|0); //@line 5041 "load_abc.cpp"
 if (!($3673)) {
  $3674 = HEAP32[$abcparts>>2]|0; //@line 5042 "load_abc.cpp"
  HEAP32[$vararg_buffer136>>2] = $3674; //@line 5042 "load_abc.cpp"
  (_printf(527638,$vararg_buffer136)|0); //@line 5042 "load_abc.cpp"
  $t$151055 = 0;
  while(1) {
   $3675 = (((($partpat) + ($t$151055<<1)|0)) + 1|0); //@line 5044 "load_abc.cpp"
   $3676 = HEAP8[$3675>>0]|0; //@line 5044 "load_abc.cpp"
   $3677 = (($partpat) + ($t$151055<<1)|0); //@line 5044 "load_abc.cpp"
   $3678 = HEAP8[$3677>>0]|0; //@line 5044 "load_abc.cpp"
   $3679 = ($3676&255)<($3678&255); //@line 5044 "load_abc.cpp"
   if (!($3679)) {
    $3680 = (($t$151055) + 65)|0; //@line 5045 "load_abc.cpp"
    HEAP32[$vararg_buffer139>>2] = $3680; //@line 5045 "load_abc.cpp"
    (_printf(527644,$vararg_buffer139)|0); //@line 5045 "load_abc.cpp"
   }
   $3681 = (($t$151055) + 1)|0; //@line 5043 "load_abc.cpp"
   $exitcond1496 = ($3681|0)==(26); //@line 5043 "load_abc.cpp"
   if ($exitcond1496) {
    break;
   } else {
    $t$151055 = $3681;
   }
  }
  $3682 = HEAP8[$101>>0]|0; //@line 5046 "load_abc.cpp"
  $3683 = HEAP8[$100>>0]|0; //@line 5046 "load_abc.cpp"
  $3684 = ($3682&255)<($3683&255); //@line 5046 "load_abc.cpp"
  if (!($3684)) {
   (_printf(527650,$vararg_buffer142)|0); //@line 5047 "load_abc.cpp"
  }
  (_putchar(10)|0); //@line 5048 "load_abc.cpp"
  $t$161054 = 0;
  while(1) {
   $3685 = (((($partpat) + ($t$161054<<1)|0)) + 1|0); //@line 5050 "load_abc.cpp"
   $3686 = HEAP8[$3685>>0]|0; //@line 5050 "load_abc.cpp"
   $3687 = (($partpat) + ($t$161054<<1)|0); //@line 5050 "load_abc.cpp"
   $3688 = HEAP8[$3687>>0]|0; //@line 5050 "load_abc.cpp"
   $3689 = ($3686&255)<($3688&255); //@line 5050 "load_abc.cpp"
   if (!($3689)) {
    $3690 = $3688&255; //@line 5050 "load_abc.cpp"
    HEAP32[$vararg_buffer144>>2] = $3690; //@line 5051 "load_abc.cpp"
    (_printf(527654,$vararg_buffer144)|0); //@line 5051 "load_abc.cpp"
   }
   $3691 = (($t$161054) + 1)|0; //@line 5049 "load_abc.cpp"
   $exitcond1495 = ($3691|0)==(27); //@line 5049 "load_abc.cpp"
   if ($exitcond1495) {
    break;
   } else {
    $t$161054 = $3691;
   }
  }
  (_putchar(10)|0); //@line 5052 "load_abc.cpp"
  $t$171053 = 0;
  while(1) {
   $3692 = (((($partpat) + ($t$171053<<1)|0)) + 1|0); //@line 5054 "load_abc.cpp"
   $3693 = HEAP8[$3692>>0]|0; //@line 5054 "load_abc.cpp"
   $3694 = (($partpat) + ($t$171053<<1)|0); //@line 5054 "load_abc.cpp"
   $3695 = HEAP8[$3694>>0]|0; //@line 5054 "load_abc.cpp"
   $3696 = ($3693&255)<($3695&255); //@line 5054 "load_abc.cpp"
   if (!($3696)) {
    $3697 = $3693&255; //@line 5054 "load_abc.cpp"
    HEAP32[$vararg_buffer147>>2] = $3697; //@line 5055 "load_abc.cpp"
    (_printf(527654,$vararg_buffer147)|0); //@line 5055 "load_abc.cpp"
   }
   $3698 = (($t$171053) + 1)|0; //@line 5053 "load_abc.cpp"
   $exitcond1494 = ($3698|0)==(27); //@line 5053 "load_abc.cpp"
   if ($exitcond1494) {
    break;
   } else {
    $t$171053 = $3698;
   }
  }
  (_putchar(10)|0); //@line 5056 "load_abc.cpp"
  $3699 = ($orderlen$4783|0)>(0); //@line 5057 "load_abc.cpp"
  do {
   if ($3699) {
    $t$181049 = 0;
    while(1) {
     HEAP32[$vararg_buffer150>>2] = $t$181049; //@line 5058 "load_abc.cpp"
     (_printf(527654,$vararg_buffer150)|0); //@line 5058 "load_abc.cpp"
     $3700 = (($t$181049) + 1)|0; //@line 5057 "load_abc.cpp"
     $exitcond1493 = ($3700|0)==($orderlen$4783|0); //@line 5057 "load_abc.cpp"
     if ($exitcond1493) {
      break;
     } else {
      $t$181049 = $3700;
     }
    }
    (_putchar(10)|0); //@line 5059 "load_abc.cpp"
    if (!($3699)) {
     break;
    }
    $3701 = HEAP32[$orderlist>>2]|0; //@line 5061 "load_abc.cpp"
    $t$191045 = 0;
    while(1) {
     $3702 = (($3701) + ($t$191045)|0); //@line 5061 "load_abc.cpp"
     $3703 = HEAP8[$3702>>0]|0; //@line 5061 "load_abc.cpp"
     $3704 = $3703&255; //@line 5061 "load_abc.cpp"
     HEAP32[$vararg_buffer153>>2] = $3704; //@line 5061 "load_abc.cpp"
     (_printf(527654,$vararg_buffer153)|0); //@line 5061 "load_abc.cpp"
     $3705 = (($t$191045) + 1)|0; //@line 5060 "load_abc.cpp"
     $exitcond1492 = ($3705|0)==($orderlen$4783|0); //@line 5060 "load_abc.cpp"
     if ($exitcond1492) {
      break;
     } else {
      $t$191045 = $3705;
     }
    }
   } else {
    (_putchar(10)|0); //@line 5059 "load_abc.cpp"
   }
  } while(0);
  (_putchar(10)|0); //@line 5062 "load_abc.cpp"
  $t$08$i = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 357 "load_abc.cpp"
  $3706 = ($t$08$i|0)==(0|0); //@line 357 "load_abc.cpp"
  if (!($3706)) {
   $3707 = ((($nn$i)) + 1|0); //@line 377 "load_abc.cpp"
   $3708 = ((($nn$i)) + 2|0); //@line 378 "load_abc.cpp"
   $t$09$i = $t$08$i;
   while(1) {
    $3709 = ((($t$09$i)) + 36|0); //@line 358 "load_abc.cpp"
    $3710 = HEAP8[$3709>>0]|0; //@line 358 "load_abc.cpp"
    $3711 = $3710&255; //@line 358 "load_abc.cpp"
    $3712 = ((($t$09$i)) + 37|0); //@line 358 "load_abc.cpp"
    $3713 = HEAP8[$3712>>0]|0; //@line 358 "load_abc.cpp"
    $3714 = $3713&255; //@line 358 "load_abc.cpp"
    $3715 = ((($t$09$i)) + 40|0); //@line 359 "load_abc.cpp"
    $3716 = HEAP8[$3715>>0]|0; //@line 359 "load_abc.cpp"
    $3717 = $3716&255; //@line 359 "load_abc.cpp"
    $3718 = ((($t$09$i)) + 44|0); //@line 359 "load_abc.cpp"
    HEAP32[$vararg_buffer156>>2] = $3711; //@line 358 "load_abc.cpp"
    $vararg_ptr159 = ((($vararg_buffer156)) + 4|0); //@line 358 "load_abc.cpp"
    HEAP32[$vararg_ptr159>>2] = $3714; //@line 358 "load_abc.cpp"
    $vararg_ptr160 = ((($vararg_buffer156)) + 8|0); //@line 358 "load_abc.cpp"
    HEAP32[$vararg_ptr160>>2] = $3717; //@line 358 "load_abc.cpp"
    $vararg_ptr161 = ((($vararg_buffer156)) + 12|0); //@line 358 "load_abc.cpp"
    HEAP32[$vararg_ptr161>>2] = $3718; //@line 358 "load_abc.cpp"
    (_printf(527659,$vararg_buffer156)|0); //@line 358 "load_abc.cpp"
    $3719 = ((($t$09$i)) + 4|0); //@line 364 "load_abc.cpp"
    $e$03$i = HEAPU8[$3719>>0]|(HEAPU8[$3719+1>>0]<<8)|(HEAPU8[$3719+2>>0]<<16)|(HEAPU8[$3719+3>>0]<<24); //@line 364 "load_abc.cpp"
    $3720 = ($e$03$i|0)==(0|0); //@line 364 "load_abc.cpp"
    if (!($3720)) {
     $3721 = (_strcmp($3672,527683)|0); //@line 360 "load_abc.cpp"
     $not$$i = ($3721|0)!=(0); //@line 5063 "load_abc.cpp"
     $$$i739 = $not$$i&1; //@line 5063 "load_abc.cpp"
     $e$05$i741 = $e$03$i;$n$14$i = $$$i739;
     while(1) {
      $3722 = ((($e$05$i741)) + 4|0); //@line 365 "load_abc.cpp"
      $3723 = HEAPU8[$3722>>0]|(HEAPU8[$3722+1>>0]<<8)|(HEAPU8[$3722+2>>0]<<16)|(HEAPU8[$3722+3>>0]<<24); //@line 365 "load_abc.cpp"
      $3724 = (($3723>>>0) / 192)&-1; //@line 334 "load_abc.cpp"
      $3725 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 366 "load_abc.cpp"
      $3726 = (($3724|0) / ($3725|0))&-1; //@line 366 "load_abc.cpp"
      $3727 = (($3726|0) / 64)&-1; //@line 367 "load_abc.cpp"
      $3728 = (($3724|0) % ($3725|0))&-1; //@line 368 "load_abc.cpp"
      $3729 = (($3726|0) % 64)&-1; //@line 369 "load_abc.cpp"
      $3730 = ($3725*12288)|0; //@line 339 "load_abc.cpp"
      $3731 = (($3723>>>0) % ($3730>>>0))&-1; //@line 370 "load_abc.cpp"
      $3732 = ($3731|0)!=(0); //@line 370 "load_abc.cpp"
      $3733 = $3732 ? 32 : 45; //@line 370 "load_abc.cpp"
      HEAP8[$nn$i>>0] = $3733; //@line 370 "load_abc.cpp"
      $3734 = ((($e$05$i741)) + 8|0); //@line 371 "load_abc.cpp"
      $3735 = HEAP8[$3734>>0]|0; //@line 371 "load_abc.cpp"
      $3736 = ($3735<<24>>24)==(1); //@line 371 "load_abc.cpp"
      do {
       if ($3736) {
        $3737 = $3733&255; //@line 373 "load_abc.cpp"
        $3738 = ((($e$05$i741)) + 14|0); //@line 373 "load_abc.cpp"
        $3739 = HEAP8[$3738>>0]|0; //@line 373 "load_abc.cpp"
        $3740 = $3739&255; //@line 373 "load_abc.cpp"
        $3741 = HEAP8[$3709>>0]|0; //@line 373 "load_abc.cpp"
        $3742 = $3741&255; //@line 373 "load_abc.cpp"
        $3743 = HEAP8[$3712>>0]|0; //@line 374 "load_abc.cpp"
        $3744 = $3743&255; //@line 374 "load_abc.cpp"
        HEAP32[$vararg_buffer162>>2] = $3727; //@line 372 "load_abc.cpp"
        $vararg_ptr165 = ((($vararg_buffer162)) + 4|0); //@line 372 "load_abc.cpp"
        HEAP32[$vararg_ptr165>>2] = $3729; //@line 372 "load_abc.cpp"
        $vararg_ptr166 = ((($vararg_buffer162)) + 8|0); //@line 372 "load_abc.cpp"
        HEAP32[$vararg_ptr166>>2] = $3728; //@line 372 "load_abc.cpp"
        $vararg_ptr167 = ((($vararg_buffer162)) + 12|0); //@line 372 "load_abc.cpp"
        HEAP32[$vararg_ptr167>>2] = $3737; //@line 372 "load_abc.cpp"
        $vararg_ptr168 = ((($vararg_buffer162)) + 16|0); //@line 372 "load_abc.cpp"
        HEAP32[$vararg_ptr168>>2] = $3740; //@line 372 "load_abc.cpp"
        $vararg_ptr169 = ((($vararg_buffer162)) + 20|0); //@line 372 "load_abc.cpp"
        HEAP32[$vararg_ptr169>>2] = $3742; //@line 372 "load_abc.cpp"
        $vararg_ptr170 = ((($vararg_buffer162)) + 24|0); //@line 372 "load_abc.cpp"
        HEAP32[$vararg_ptr170>>2] = $3744; //@line 372 "load_abc.cpp"
        $vararg_ptr171 = ((($vararg_buffer162)) + 28|0); //@line 372 "load_abc.cpp"
        HEAP32[$vararg_ptr171>>2] = $3718; //@line 372 "load_abc.cpp"
        (_printf(527691,$vararg_buffer162)|0); //@line 372 "load_abc.cpp"
        $3745 = ((($e$05$i741)) + 9|0); //@line 375 "load_abc.cpp"
        $3746 = HEAP8[$3745>>0]|0; //@line 375 "load_abc.cpp"
        $3747 = ($3746<<24>>24)==(67); //@line 375 "load_abc.cpp"
        if ($3747) {
         $3748 = ((($e$05$i741)) + 11|0); //@line 376 "load_abc.cpp"
         $3749 = HEAP8[$3748>>0]|0; //@line 376 "load_abc.cpp"
         $3750 = $3749&255; //@line 376 "load_abc.cpp"
         $3751 = (527719 + ($3750)|0); //@line 376 "load_abc.cpp"
         $3752 = HEAP8[$3751>>0]|0; //@line 376 "load_abc.cpp"
         HEAP8[$nn$i>>0] = $3752; //@line 376 "load_abc.cpp"
         $3753 = HEAP8[$3748>>0]|0; //@line 377 "load_abc.cpp"
         $3754 = $3753&255; //@line 377 "load_abc.cpp"
         $3755 = (527746 + ($3754)|0); //@line 377 "load_abc.cpp"
         $3756 = HEAP8[$3755>>0]|0; //@line 377 "load_abc.cpp"
         HEAP8[$3707>>0] = $3756; //@line 377 "load_abc.cpp"
         HEAP8[$3708>>0] = 0; //@line 378 "load_abc.cpp"
         $3757 = $3756 << 24 >> 24; //@line 379 "load_abc.cpp"
         $3758 = (_isspace($3757)|0); //@line 379 "load_abc.cpp"
         $3759 = ($3758|0)==(0); //@line 379 "load_abc.cpp"
         if (!($3759)) {
          HEAP8[$3707>>0] = 0; //@line 379 "load_abc.cpp"
         }
         $3760 = HEAP8[$3745>>0]|0; //@line 381 "load_abc.cpp"
         $3761 = $3760 << 24 >> 24; //@line 381 "load_abc.cpp"
         $3762 = ((($e$05$i741)) + 10|0); //@line 381 "load_abc.cpp"
         $3763 = HEAP8[$3762>>0]|0; //@line 381 "load_abc.cpp"
         $3764 = $3763&255; //@line 381 "load_abc.cpp"
         $3765 = (1277594 + ($3764<<3)|0); //@line 381 "load_abc.cpp"
         HEAP32[$vararg_buffer172>>2] = $3761; //@line 380 "load_abc.cpp"
         $vararg_ptr175 = ((($vararg_buffer172)) + 4|0); //@line 380 "load_abc.cpp"
         HEAP32[$vararg_ptr175>>2] = $nn$i; //@line 380 "load_abc.cpp"
         $vararg_ptr176 = ((($vararg_buffer172)) + 8|0); //@line 380 "load_abc.cpp"
         HEAP32[$vararg_ptr176>>2] = $3765; //@line 380 "load_abc.cpp"
         (_printf(527773,$vararg_buffer172)|0); //@line 380 "load_abc.cpp"
         $3766 = ((($e$05$i741)) + 12|0); //@line 382 "load_abc.cpp"
         $3767 = HEAP8[$3766>>0]|0; //@line 382 "load_abc.cpp"
         $3768 = HEAP8[$3748>>0]|0; //@line 382 "load_abc.cpp"
         $3769 = ($3767<<24>>24)==($3768<<24>>24); //@line 382 "load_abc.cpp"
         if (!($3769)) {
          $3770 = $3767&255; //@line 382 "load_abc.cpp"
          $3771 = (527719 + ($3770)|0); //@line 383 "load_abc.cpp"
          $3772 = HEAP8[$3771>>0]|0; //@line 383 "load_abc.cpp"
          HEAP8[$nn$i>>0] = $3772; //@line 383 "load_abc.cpp"
          $3773 = HEAP8[$3766>>0]|0; //@line 384 "load_abc.cpp"
          $3774 = $3773&255; //@line 384 "load_abc.cpp"
          $3775 = (527746 + ($3774)|0); //@line 384 "load_abc.cpp"
          $3776 = HEAP8[$3775>>0]|0; //@line 384 "load_abc.cpp"
          HEAP8[$3707>>0] = $3776; //@line 384 "load_abc.cpp"
          HEAP8[$3708>>0] = 0; //@line 385 "load_abc.cpp"
          HEAP32[$vararg_buffer177>>2] = $nn$i; //@line 386 "load_abc.cpp"
          (_printf(527793,$vararg_buffer177)|0); //@line 386 "load_abc.cpp"
         }
         (_putchar(10)|0); //@line 388 "load_abc.cpp"
        } else {
         $3777 = $3746 << 24 >> 24; //@line 392 "load_abc.cpp"
         $3778 = ((($e$05$i741)) + 10|0); //@line 393 "load_abc.cpp"
         $3779 = HEAPU8[$3778>>0]|(HEAPU8[$3778+1>>0]<<8)|(HEAPU8[$3778+2>>0]<<16)|(HEAPU8[$3778+3>>0]<<24); //@line 393 "load_abc.cpp"
         HEAP32[$vararg_buffer180>>2] = $3777; //@line 391 "load_abc.cpp"
         $vararg_ptr183 = ((($vararg_buffer180)) + 4|0); //@line 391 "load_abc.cpp"
         HEAP32[$vararg_ptr183>>2] = $e$05$i741; //@line 391 "load_abc.cpp"
         $vararg_ptr184 = ((($vararg_buffer180)) + 8|0); //@line 391 "load_abc.cpp"
         HEAP32[$vararg_ptr184>>2] = $3779; //@line 391 "load_abc.cpp"
         (_printf(527797,$vararg_buffer180)|0); //@line 391 "load_abc.cpp"
        }
        $3780 = (_strcmp($3672,527683)|0); //@line 394 "load_abc.cpp"
        $not$2$i = ($3780|0)!=(0); //@line 5063 "load_abc.cpp"
        $$1$i742 = $not$2$i&1; //@line 5063 "load_abc.cpp"
        $n$2$i = $$1$i742;
       } else {
        $3781 = ($n$14$i|0)==(0); //@line 399 "load_abc.cpp"
        if ($3781) {
         $n$2$i = 0;
         break;
        }
        $3782 = $3733&255; //@line 400 "load_abc.cpp"
        $3783 = ((($e$05$i741)) + 14|0); //@line 400 "load_abc.cpp"
        $3784 = HEAP8[$3783>>0]|0; //@line 400 "load_abc.cpp"
        $3785 = $3784&255; //@line 400 "load_abc.cpp"
        $3786 = HEAP8[$3709>>0]|0; //@line 400 "load_abc.cpp"
        $3787 = $3786&255; //@line 400 "load_abc.cpp"
        $3788 = HEAP8[$3712>>0]|0; //@line 400 "load_abc.cpp"
        $3789 = $3788&255; //@line 400 "load_abc.cpp"
        HEAP32[$vararg_buffer185>>2] = $3727; //@line 400 "load_abc.cpp"
        $vararg_ptr188 = ((($vararg_buffer185)) + 4|0); //@line 400 "load_abc.cpp"
        HEAP32[$vararg_ptr188>>2] = $3729; //@line 400 "load_abc.cpp"
        $vararg_ptr189 = ((($vararg_buffer185)) + 8|0); //@line 400 "load_abc.cpp"
        HEAP32[$vararg_ptr189>>2] = $3728; //@line 400 "load_abc.cpp"
        $vararg_ptr190 = ((($vararg_buffer185)) + 12|0); //@line 400 "load_abc.cpp"
        HEAP32[$vararg_ptr190>>2] = $3782; //@line 400 "load_abc.cpp"
        $vararg_ptr191 = ((($vararg_buffer185)) + 16|0); //@line 400 "load_abc.cpp"
        HEAP32[$vararg_ptr191>>2] = $3785; //@line 400 "load_abc.cpp"
        $vararg_ptr192 = ((($vararg_buffer185)) + 20|0); //@line 400 "load_abc.cpp"
        HEAP32[$vararg_ptr192>>2] = $3787; //@line 400 "load_abc.cpp"
        $vararg_ptr193 = ((($vararg_buffer185)) + 24|0); //@line 400 "load_abc.cpp"
        HEAP32[$vararg_ptr193>>2] = $3789; //@line 400 "load_abc.cpp"
        $vararg_ptr194 = ((($vararg_buffer185)) + 28|0); //@line 400 "load_abc.cpp"
        HEAP32[$vararg_ptr194>>2] = $3718; //@line 400 "load_abc.cpp"
        (_printf(527691,$vararg_buffer185)|0); //@line 400 "load_abc.cpp"
        $3790 = HEAP8[$3734>>0]|0; //@line 401 "load_abc.cpp"
        $3791 = ($3790<<24>>24)==(0); //@line 401 "load_abc.cpp"
        if ($3791) {
         ;HEAP8[$nn$i>>0]=HEAP8[527817>>0]|0;HEAP8[$nn$i+1>>0]=HEAP8[527817+1>>0]|0;HEAP8[$nn$i+2>>0]=HEAP8[527817+2>>0]|0; //@line 406 "load_abc.cpp"
        } else {
         $3792 = $3790&255; //@line 402 "load_abc.cpp"
         $3793 = (($3792) + -23)|0; //@line 402 "load_abc.cpp"
         $3794 = (527719 + ($3793)|0); //@line 402 "load_abc.cpp"
         $3795 = HEAP8[$3794>>0]|0; //@line 402 "load_abc.cpp"
         HEAP8[$nn$i>>0] = $3795; //@line 402 "load_abc.cpp"
         $3796 = HEAP8[$3734>>0]|0; //@line 403 "load_abc.cpp"
         $3797 = $3796&255; //@line 403 "load_abc.cpp"
         $3798 = (($3797) + -23)|0; //@line 403 "load_abc.cpp"
         $3799 = (527746 + ($3798)|0); //@line 403 "load_abc.cpp"
         $3800 = HEAP8[$3799>>0]|0; //@line 403 "load_abc.cpp"
         HEAP8[$3707>>0] = $3800; //@line 403 "load_abc.cpp"
         HEAP8[$3708>>0] = 0; //@line 404 "load_abc.cpp"
        }
        $3801 = ((($e$05$i741)) + 9|0); //@line 408 "load_abc.cpp"
        $3802 = HEAP8[$3801>>0]|0; //@line 408 "load_abc.cpp"
        $3803 = $3802&255; //@line 408 "load_abc.cpp"
        $3804 = ((($e$05$i741)) + 10|0); //@line 408 "load_abc.cpp"
        $3805 = HEAP8[$3804>>0]|0; //@line 408 "load_abc.cpp"
        $3806 = $3805&255; //@line 408 "load_abc.cpp"
        $3807 = (_pat_smptogm($3806)|0); //@line 408 "load_abc.cpp"
        $3808 = (_pat_gm_name($3807)|0); //@line 408 "load_abc.cpp"
        $3809 = ((($e$05$i741)) + 11|0); //@line 408 "load_abc.cpp"
        $3810 = HEAP8[$3809>>0]|0; //@line 408 "load_abc.cpp"
        $3811 = $3810&255; //@line 408 "load_abc.cpp"
        HEAP32[$vararg_buffer195>>2] = $nn$i; //@line 407 "load_abc.cpp"
        $vararg_ptr198 = ((($vararg_buffer195)) + 4|0); //@line 407 "load_abc.cpp"
        HEAP32[$vararg_ptr198>>2] = $3803; //@line 407 "load_abc.cpp"
        $vararg_ptr199 = ((($vararg_buffer195)) + 8|0); //@line 407 "load_abc.cpp"
        HEAP32[$vararg_ptr199>>2] = $3808; //@line 407 "load_abc.cpp"
        $vararg_ptr200 = ((($vararg_buffer195)) + 12|0); //@line 407 "load_abc.cpp"
        HEAP32[$vararg_ptr200>>2] = $3811; //@line 407 "load_abc.cpp"
        (_printf(527834,$vararg_buffer195)|0); //@line 407 "load_abc.cpp"
        $3812 = (_strcmp($3672,527870)|0); //@line 409 "load_abc.cpp"
        $3813 = ($3812|0)==(0); //@line 409 "load_abc.cpp"
        $n$1$$i = $3813 ? $n$14$i : 0; //@line 409 "load_abc.cpp"
        $n$2$i = $n$1$$i;
       }
      } while(0);
      $e$0$i743 = HEAPU8[$e$05$i741>>0]|(HEAPU8[$e$05$i741+1>>0]<<8)|(HEAPU8[$e$05$i741+2>>0]<<16)|(HEAPU8[$e$05$i741+3>>0]<<24); //@line 364 "load_abc.cpp"
      $3814 = ($e$0$i743|0)==(0|0); //@line 364 "load_abc.cpp"
      if ($3814) {
       break;
      } else {
       $e$05$i741 = $e$0$i743;$n$14$i = $n$2$i;
      }
     }
    }
    $t$0$i = HEAPU8[$t$09$i>>0]|(HEAPU8[$t$09$i+1>>0]<<8)|(HEAPU8[$t$09$i+2>>0]<<16)|(HEAPU8[$t$09$i+3>>0]<<24); //@line 357 "load_abc.cpp"
    $3815 = ($t$0$i|0)==(0|0); //@line 357 "load_abc.cpp"
    if ($3815) {
     break;
    } else {
     $t$09$i = $t$0$i;
    }
   }
  }
 }
 $3816 = HEAP32[$30>>2]|0; //@line 5067 "load_abc.cpp"
 $3817 = ($3816|0)==(0); //@line 5067 "load_abc.cpp"
 if ($3817) {
  $3818 = ($abctempo$1$lcssa$lcssa|0)==(0); //@line 5066 "load_abc.cpp"
  $$abctempo$0 = $3818 ? 240 : $abctempo$1$lcssa$lcssa; //@line 5066 "load_abc.cpp"
  HEAP32[$30>>2] = $$abctempo$0; //@line 5067 "load_abc.cpp"
 }
 $3819 = ((($this)) + 59680|0); //@line 5088 "load_abc.cpp"
 HEAP32[$3819>>2] = 16777216; //@line 5088 "load_abc.cpp"
 $3820 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 5089 "load_abc.cpp"
 $3821 = (($3820>>>0) / 192)&-1; //@line 334 "load_abc.cpp"
 $3822 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 5089 "load_abc.cpp"
 $3823 = (($3821>>>0) / ($3822>>>0))&-1; //@line 5089 "load_abc.cpp"
 $3824 = $3823 >>> 6; //@line 5089 "load_abc.cpp"
 $3825 = (($3824) + 1)|0; //@line 5089 "load_abc.cpp"
 $3826 = ($3825>>>0)>(240); //@line 5090 "load_abc.cpp"
 $$ = $3826 ? 240 : $3825; //@line 5090 "load_abc.cpp"
 $3827 = ((($this)) + 59648|0); //@line 5092 "load_abc.cpp"
 HEAP32[$3827>>2] = $3822; //@line 5092 "load_abc.cpp"
 $t$01$i = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 644 "load_abc.cpp"
 $3828 = ($t$01$i|0)==(0|0); //@line 644 "load_abc.cpp"
 if ($3828) {
  $n$0$lcssa$i = 0;
 } else {
  $n$02$i = 0;$t$03$i = $t$01$i;
  while(1) {
   $3829 = (($n$02$i) + 1)|0; //@line 645 "load_abc.cpp"
   $t$0$i746 = HEAPU8[$t$03$i>>0]|(HEAPU8[$t$03$i+1>>0]<<8)|(HEAPU8[$t$03$i+2>>0]<<16)|(HEAPU8[$t$03$i+3>>0]<<24); //@line 644 "load_abc.cpp"
   $3830 = ($t$0$i746|0)==(0|0); //@line 644 "load_abc.cpp"
   if ($3830) {
    $n$0$lcssa$i = $3829;
    break;
   } else {
    $n$02$i = $3829;$t$03$i = $t$0$i746;
   }
  }
 }
 $3831 = ((($this)) + 59664|0); //@line 5093 "load_abc.cpp"
 HEAP32[$3831>>2] = $n$0$lcssa$i; //@line 5093 "load_abc.cpp"
 $3832 = ((($this)) + 59660|0); //@line 5094 "load_abc.cpp"
 HEAP32[$3832>>2] = 16; //@line 5094 "load_abc.cpp"
 $3833 = ((($this)) + 59764|0); //@line 5095 "load_abc.cpp"
 HEAP32[$3833>>2] = 112; //@line 5095 "load_abc.cpp"
 $3834 = ((($this)) + 59768|0); //@line 5096 "load_abc.cpp"
 HEAP32[$3834>>2] = 13696; //@line 5096 "load_abc.cpp"
 $3835 = ($orderlen$4783|0)==(0); //@line 5098 "load_abc.cpp"
 $3836 = HEAP32[$orderlist>>2]|0; //@line 5103 "load_abc.cpp"
 if (!($3835)) {
  $3837 = (0 - ($orderlen$4783))|0; //@line 5098 "load_abc.cpp"
  $3838 = ($3837>>>0)>(4294967040);
  $3839 = $3838 ? $orderlen$4783 : 256; //@line 5098 "load_abc.cpp"
  $t$201042 = 0;
  while(1) {
   $3840 = (($3836) + ($t$201042)|0); //@line 5101 "load_abc.cpp"
   $3841 = HEAP8[$3840>>0]|0; //@line 5101 "load_abc.cpp"
   $3842 = (((($this)) + 53344|0) + ($t$201042)|0); //@line 5101 "load_abc.cpp"
   HEAP8[$3842>>0] = $3841; //@line 5101 "load_abc.cpp"
   $3843 = (($t$201042) + 1)|0; //@line 5098 "load_abc.cpp"
   $exitcond1491 = ($3843|0)==($3839|0); //@line 5098 "load_abc.cpp"
   if ($exitcond1491) {
    break;
   } else {
    $t$201042 = $3843;
   }
  }
 }
 _free($3836); //@line 5103 "load_abc.cpp"
 $3844 = ((($this)) + 51904|0); //@line 5127 "load_abc.cpp"
 $3845 = ((($this)) + 52864|0); //@line 5127 "load_abc.cpp"
 $3846 = HEAP32[$3831>>2]|0; //@line 5127 "load_abc.cpp"
 $3847 = (__ZL16ABC_ReadPatternsPP11_MODCOMMANDPtP10_ABCHANDLEii($3844,$3845,$1,$$,$3846)|0); //@line 5127 "load_abc.cpp"
 $3848 = ($3847|0)==(0); //@line 5127 "load_abc.cpp"
 if (!($3848)) {
  $3849 = HEAP32[$3831>>2]|0; //@line 5129 "load_abc.cpp"
  $3850 = (($3849) + 1)|0; //@line 5129 "load_abc.cpp"
  HEAP32[$3831>>2] = $3850; //@line 5129 "load_abc.cpp"
  $3851 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($1,1282466,99)|0); //@line 5130 "load_abc.cpp"
  HEAP8[$41>>0]=$3851&255;HEAP8[$41+1>>0]=($3851>>8)&255;HEAP8[$41+2>>0]=($3851>>16)&255;HEAP8[$41+3>>0]=$3851>>24; //@line 5130 "load_abc.cpp"
  $3852 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 5131 "load_abc.cpp"
  $3853 = ((($3851)) + 8|0); //@line 1590 "load_abc.cpp"
  $3854 = HEAPU8[$3853>>0]|(HEAPU8[$3853+1>>0]<<8)|(HEAPU8[$3853+2>>0]<<16)|(HEAPU8[$3853+3>>0]<<24); //@line 1590 "load_abc.cpp"
  $3855 = ($3854|0)==(0|0); //@line 1591 "load_abc.cpp"
  do {
   if ($3855) {
    label = 1281;
   } else {
    $3856 = ((($3854)) + 4|0); //@line 1591 "load_abc.cpp"
    $3857 = HEAPU8[$3856>>0]|(HEAPU8[$3856+1>>0]<<8)|(HEAPU8[$3856+2>>0]<<16)|(HEAPU8[$3856+3>>0]<<24); //@line 1591 "load_abc.cpp"
    $3858 = ($3857|0)==($3852|0); //@line 1591 "load_abc.cpp"
    if ($3858) {
     break;
    }
    $3859 = ((($3854)) + 8|0); //@line 1592 "load_abc.cpp"
    $3860 = HEAP8[$3859>>0]|0; //@line 1592 "load_abc.cpp"
    $3861 = ($3860<<24>>24)==(1); //@line 1592 "load_abc.cpp"
    if (!($3861)) {
     label = 1281;
     break;
    }
    $3862 = ((($3854)) + 9|0); //@line 1592 "load_abc.cpp"
    $3863 = HEAP8[$3862>>0]|0; //@line 1592 "load_abc.cpp"
    $3864 = ($3863<<24>>24)==(83); //@line 1592 "load_abc.cpp"
    if (!($3864)) {
     label = 1281;
     break;
    }
    HEAP8[$3856>>0]=$3852&255;HEAP8[$3856+1>>0]=($3852>>8)&255;HEAP8[$3856+2>>0]=($3852>>16)&255;HEAP8[$3856+3>>0]=$3852>>24; //@line 1593 "load_abc.cpp"
   }
  } while(0);
  if ((label|0) == 1281) {
   $3865 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
   HEAP8[$3865>>0]=0&255;HEAP8[$3865+1>>0]=(0>>8)&255;HEAP8[$3865+2>>0]=(0>>16)&255;HEAP8[$3865+3>>0]=0>>24; //@line 540 "load_abc.cpp"
   $3866 = ((($3865)) + 4|0); //@line 541 "load_abc.cpp"
   HEAP8[$3866>>0]=$3852&255;HEAP8[$3866+1>>0]=($3852>>8)&255;HEAP8[$3866+2>>0]=($3852>>16)&255;HEAP8[$3866+3>>0]=$3852>>24; //@line 541 "load_abc.cpp"
   $3867 = ((($3865)) + 8|0); //@line 543 "load_abc.cpp"
   HEAP8[$3867>>0] = 1; //@line 543 "load_abc.cpp"
   $3868 = ((($3865)) + 9|0); //@line 543 "load_abc.cpp"
   HEAP8[$3868>>0] = 83; //@line 543 "load_abc.cpp"
   $3869 = ((($3865)) + 10|0); //@line 543 "load_abc.cpp"
   HEAP8[$3869>>0]=0&255;HEAP8[$3869+1>>0]=(0>>8)&255;HEAP8[$3869+2>>0]=(0>>16)&255;HEAP8[$3869+3>>0]=0>>24; //@line 543 "load_abc.cpp"
   $3870 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
   $3871 = ((($3865)) + 14|0); //@line 544 "load_abc.cpp"
   HEAP8[$3871>>0] = $3870; //@line 544 "load_abc.cpp"
   $3872 = ((($3865)) + 15|0); //@line 545 "load_abc.cpp"
   HEAP8[$3872>>0] = 0; //@line 545 "load_abc.cpp"
   __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($3851,$3865); //@line 1600 "load_abc.cpp"
  }
  $3873 = (-2 - ($3824))|0; //@line 5132 "load_abc.cpp"
  $3874 = $3873 ^ -16;
  $3875 = ($3874>>>0)<(240);
  $$op1547 = (($3824) + 1)|0; //@line 5132 "load_abc.cpp"
  $3876 = $3875 ? $$op1547 : 240; //@line 5132 "load_abc.cpp"
  $t$211041 = 0;
  while(1) {
   $3877 = (((($this)) + 51904|0) + ($t$211041<<2)|0); //@line 5133 "load_abc.cpp"
   $3878 = HEAP32[$3877>>2]|0; //@line 5133 "load_abc.cpp"
   __ZN10CSoundFile11FreePatternEPv($3878); //@line 5133 "load_abc.cpp"
   HEAP32[$3877>>2] = 0; //@line 5134 "load_abc.cpp"
   $3879 = (($t$211041) + 1)|0; //@line 5132 "load_abc.cpp"
   $exitcond = ($3879|0)==($3876|0); //@line 5132 "load_abc.cpp"
   if ($exitcond) {
    break;
   } else {
    $t$211041 = $3879;
   }
  }
  $3880 = HEAP32[$3831>>2]|0; //@line 5136 "load_abc.cpp"
  (__ZL16ABC_ReadPatternsPP11_MODCOMMANDPtP10_ABCHANDLEii($3844,$3845,$1,$$,$3880)|0); //@line 5136 "load_abc.cpp"
 }
 $3881 = (_PAT_Load_Instruments($this)|0); //@line 5139 "load_abc.cpp"
 if (!($3881)) {
  HEAP32[299859] = 0; //@line 5140 "load_abc.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 5158 "load_abc.cpp"
 }
 $3882 = HEAP32[$3831>>2]|0; //@line 5145 "load_abc.cpp"
 $3883 = ($3882>>>0)>(64); //@line 5145 "load_abc.cpp"
 if ($3883) {
  HEAP32[$3831>>2] = 64; //@line 5146 "load_abc.cpp"
  $t$221040 = 0;
  label = 1290;
 } else {
  $3884 = ($3882|0)==(0); //@line 5147 "load_abc.cpp"
  if (!($3884)) {
   $t$221040 = 0;
   label = 1290;
  }
 }
 if ((label|0) == 1290) {
  while(1) {
   label = 0;
   $3885 = (($t$221040) + 2)|0; //@line 5148 "load_abc.cpp"
   $3886 = (($3885>>>0) % 5)&-1; //@line 5148 "load_abc.cpp"
   $3887 = $3886 << 5; //@line 5148 "load_abc.cpp"
   $3888 = (($3887) + 48)|0; //@line 5148 "load_abc.cpp"
   $3889 = (((($this)) + 49600|0) + (($t$221040*36)|0)|0); //@line 5148 "load_abc.cpp"
   HEAP32[$3889>>2] = $3888; //@line 5148 "load_abc.cpp"
   $3890 = (((((($this)) + 49600|0) + (($t$221040*36)|0)|0)) + 4|0); //@line 5149 "load_abc.cpp"
   HEAP32[$3890>>2] = 64; //@line 5149 "load_abc.cpp"
   $3891 = (($t$221040) + 1)|0; //@line 5147 "load_abc.cpp"
   $3892 = HEAP32[$3831>>2]|0; //@line 5147 "load_abc.cpp"
   $3893 = ($3891>>>0)<($3892>>>0); //@line 5147 "load_abc.cpp"
   if ($3893) {
    $t$221040 = $3891;
    label = 1290;
   } else {
    break;
   }
  }
 }
 HEAP32[299859] = 0; //@line 5152 "load_abc.cpp"
 $3894 = HEAP32[$abcparts>>2]|0; //@line 1820 "load_abc.cpp"
 $3895 = ($3894|0)==(0|0); //@line 1820 "load_abc.cpp"
 if (!($3895)) {
  _free($3894); //@line 1820 "load_abc.cpp"
 }
 HEAP32[$abcparts>>2] = 0; //@line 1822 "load_abc.cpp"
 __ZL17ABC_CleanupMacrosP10_ABCHANDLE($1); //@line 2503 "load_abc.cpp"
 $3896 = HEAPU8[$156>>0]|(HEAPU8[$156+1>>0]<<8)|(HEAPU8[$156+2>>0]<<16)|(HEAPU8[$156+3>>0]<<24); //@line 2459 "load_abc.cpp"
 $3897 = ($3896|0)==(0|0); //@line 2459 "load_abc.cpp"
 if (!($3897)) {
  $tp$01$i$i = $3896;
  while(1) {
   $3898 = HEAPU8[$tp$01$i$i>>0]|(HEAPU8[$tp$01$i$i+1>>0]<<8)|(HEAPU8[$tp$01$i$i+2>>0]<<16)|(HEAPU8[$tp$01$i$i+3>>0]<<24); //@line 2460 "load_abc.cpp"
   $3899 = ((($tp$01$i$i)) + 4|0); //@line 2429 "load_abc.cpp"
   $3900 = HEAPU8[$3899>>0]|(HEAPU8[$3899+1>>0]<<8)|(HEAPU8[$3899+2>>0]<<16)|(HEAPU8[$3899+3>>0]<<24); //@line 2429 "load_abc.cpp"
   $3901 = ($3900|0)==(0|0); //@line 2429 "load_abc.cpp"
   if (!($3901)) {
    $ep$01$i$i$i = $3900;
    while(1) {
     $3902 = HEAPU8[$ep$01$i$i$i>>0]|(HEAPU8[$ep$01$i$i$i+1>>0]<<8)|(HEAPU8[$ep$01$i$i$i+2>>0]<<16)|(HEAPU8[$ep$01$i$i$i+3>>0]<<24); //@line 2430 "load_abc.cpp"
     _free($ep$01$i$i$i); //@line 2431 "load_abc.cpp"
     $3903 = ($3902|0)==(0|0); //@line 2429 "load_abc.cpp"
     if ($3903) {
      break;
     } else {
      $ep$01$i$i$i = $3902;
     }
    }
   }
   HEAP8[$3899>>0]=0&255;HEAP8[$3899+1>>0]=(0>>8)&255;HEAP8[$3899+2>>0]=(0>>16)&255;HEAP8[$3899+3>>0]=0>>24; //@line 2433 "load_abc.cpp"
   $3904 = ($3898|0)==(0|0); //@line 2459 "load_abc.cpp"
   if ($3904) {
    break;
   } else {
    $tp$01$i$i = $3898;
   }
  }
 }
 HEAP8[$156>>0]=0&255;HEAP8[$156+1>>0]=(0>>8)&255;HEAP8[$156+2>>0]=(0>>16)&255;HEAP8[$156+3>>0]=0>>24; //@line 2463 "load_abc.cpp"
 $3905 = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8)|(HEAPU8[$5+2>>0]<<16)|(HEAPU8[$5+3>>0]<<24); //@line 2510 "load_abc.cpp"
 $3906 = ($3905|0)==(0|0); //@line 2510 "load_abc.cpp"
 if (!($3906)) {
  _free($3905); //@line 2511 "load_abc.cpp"
 }
 $3907 = HEAPU8[$3>>0]|(HEAPU8[$3+1>>0]<<8)|(HEAPU8[$3+2>>0]<<16)|(HEAPU8[$3+3>>0]<<24); //@line 2512 "load_abc.cpp"
 $3908 = ($3907|0)==(0|0); //@line 2512 "load_abc.cpp"
 if (!($3908)) {
  _free($3907); //@line 2513 "load_abc.cpp"
 }
 _free($1); //@line 2514 "load_abc.cpp"
 $$0 = 1;
 STACKTOP = sp;return ($$0|0); //@line 5158 "load_abc.cpp"
}
function __ZN10CSoundFile7TestABCEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$0 = 0, $$ph = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$04$i$i$ph = 0, $i$06$i$i = 0, $id = 0, $isdigit$i = 0, $isdigit$i3 = 0, $isdigittmp$i = 0, $isdigittmp$i2 = 0, $not$$i$i = 0, $not$$i$i$i$i = 0, $not$$i$i19 = 0, $not$$i1$i$i$i = 0, $or$cond = 0, $or$cond$i = 0, $or$cond$i4 = 0;
 var $or$cond$not$i$i$i = 0, $or$cond10 = 0, $or$cond11 = 0, $or$cond4$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $id = sp;
 $0 = ((($id)) + 2|0); //@line 2360 "load_abc.cpp"
 $not$$i$i19 = ($dwMemLength|0)>(0); //@line 482 "load_abc.cpp"
 if (!($not$$i$i19)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 2363 "load_abc.cpp"
 }
 $1 = ((($id)) + 1|0); //@line 2355 "load_abc.cpp"
 $3 = 0;$i$06$i$i = 0;
 L4: while(1) {
  $2 = ($3|0)>(-1); //@line 481 "load_abc.cpp"
  $not$$i$i$i$i = ($3|0)<($dwMemLength|0); //@line 482 "load_abc.cpp"
  $or$cond11 = $2 & $not$$i$i$i$i; //@line 481 "load_abc.cpp"
  do {
   if ($or$cond11) {
    $4 = (($lpStream) + ($3)|0); //@line 490 "load_abc.cpp"
    $5 = HEAP8[$4>>0]|0; //@line 490 "load_abc.cpp"
    $6 = (($3) + 1)|0; //@line 491 "load_abc.cpp"
    $or$cond$not$i$i$i = ($5<<24>>24)==(13); //@line 492 "load_abc.cpp"
    $not$$i1$i$i$i = ($6|0)<($dwMemLength|0); //@line 482 "load_abc.cpp"
    $or$cond4$i$i$i = $not$$i1$i$i$i & $or$cond$not$i$i$i; //@line 492 "load_abc.cpp"
    if ($or$cond4$i$i$i) {
     $7 = (($lpStream) + ($6)|0); //@line 492 "load_abc.cpp"
     $8 = HEAP8[$7>>0]|0; //@line 492 "load_abc.cpp"
     $9 = ($8<<24>>24)==(10); //@line 492 "load_abc.cpp"
     if ($9) {
      $10 = (($3) + 2)|0; //@line 494 "load_abc.cpp"
      $11 = (($id) + ($i$06$i$i)|0); //@line 505 "load_abc.cpp"
      HEAP8[$11>>0] = 10; //@line 505 "load_abc.cpp"
      $$ph = $10;$i$04$i$i$ph = $i$06$i$i;
     } else {
      label = 7;
     }
    } else {
     label = 7;
    }
    if ((label|0) == 7) {
     label = 0;
     $12 = ($5<<24>>24)==(-1); //@line 504 "load_abc.cpp"
     if ($12) {
      $43 = $6;
      label = 11;
      break;
     }
     $13 = (($id) + ($i$06$i$i)|0); //@line 505 "load_abc.cpp"
     HEAP8[$13>>0] = $5; //@line 505 "load_abc.cpp"
     $14 = ($5<<24>>24)==(10); //@line 506 "load_abc.cpp"
     if ($14) {
      $$ph = $6;$i$04$i$i$ph = $i$06$i$i;
     } else {
      $15 = (($i$06$i$i) + 1)|0; //@line 502 "load_abc.cpp"
      $16 = ($15|0)<(127); //@line 502 "load_abc.cpp"
      if ($16) {
       $3 = $6;$i$06$i$i = $15;
       continue L4;
      } else {
       $$ph = $6;$i$04$i$i$ph = $15;
      }
     }
    }
    $17 = (($id) + ($i$04$i$i$ph)|0); //@line 508 "load_abc.cpp"
    HEAP8[$17>>0] = 0; //@line 508 "load_abc.cpp"
    $22 = $$ph;
   } else {
    $43 = $3;
    label = 11;
   }
  } while(0);
  if ((label|0) == 11) {
   label = 0;
   $18 = (($id) + ($i$06$i$i)|0); //@line 508 "load_abc.cpp"
   HEAP8[$18>>0] = 0; //@line 508 "load_abc.cpp"
   $22 = $43;
  }
  $19 = HEAP8[$id>>0]|0; //@line 2352 "load_abc.cpp"
  $20 = ($19<<24>>24)==(0); //@line 2352 "load_abc.cpp"
  if ($20) {
   $21 = ($22|0)<(120); //@line 2352 "load_abc.cpp"
   if ($21) {
    $$0 = 0;
    label = 24;
    break;
   }
  } else {
   $24 = $19 << 24 >> 24; //@line 281 "load_abc.cpp"
   $25 = (_isalpha($24)|0); //@line 281 "load_abc.cpp"
   $26 = ($25|0)!=(0); //@line 281 "load_abc.cpp"
   $isdigittmp$i2 = (($24) + -48)|0; //@line 2355 "load_abc.cpp"
   $isdigit$i3 = ($isdigittmp$i2>>>0)<(10); //@line 281 "load_abc.cpp"
   $or$cond$i4 = $26 | $isdigit$i3; //@line 281 "load_abc.cpp"
   if (!($or$cond$i4)) {
    $27 = (_isspace($24)|0); //@line 281 "load_abc.cpp"
    $28 = ($27|0)==(0); //@line 281 "load_abc.cpp"
    if ($28) {
     switch ($19<<24>>24) {
     case 37: case 58:  {
      break;
     }
     default: {
      $$0 = 0;
      label = 24;
      break L4;
     }
     }
    }
   }
   $29 = HEAP8[$1>>0]|0; //@line 2355 "load_abc.cpp"
   $30 = $29 << 24 >> 24; //@line 281 "load_abc.cpp"
   $31 = (_isalpha($30)|0); //@line 281 "load_abc.cpp"
   $32 = ($31|0)!=(0); //@line 281 "load_abc.cpp"
   $isdigittmp$i = (($30) + -48)|0; //@line 2355 "load_abc.cpp"
   $isdigit$i = ($isdigittmp$i>>>0)<(10); //@line 281 "load_abc.cpp"
   $or$cond$i = $32 | $isdigit$i; //@line 281 "load_abc.cpp"
   if (!($or$cond$i)) {
    $33 = (_isspace($30)|0); //@line 281 "load_abc.cpp"
    $34 = ($33|0)==(0); //@line 281 "load_abc.cpp"
    if ($34) {
     switch ($29<<24>>24) {
     case 37: case 58:  {
      break;
     }
     default: {
      $$0 = 0;
      label = 24;
      break L4;
     }
     }
    }
   }
   $35 = ($19<<24>>24)==(75); //@line 2358 "load_abc.cpp"
   $36 = ($29<<24>>24)==(58); //@line 2359 "load_abc.cpp"
   $or$cond = $35 & $36; //@line 2359 "load_abc.cpp"
   if ($or$cond) {
    $37 = HEAP8[$0>>0]|0; //@line 2360 "load_abc.cpp"
    $38 = $37 << 24 >> 24; //@line 2360 "load_abc.cpp"
    $39 = (_isalpha($38)|0); //@line 2360 "load_abc.cpp"
    $40 = ($39|0)==(0); //@line 2360 "load_abc.cpp"
    if (!($40)) {
     $$0 = 1;
     label = 24;
     break;
    }
    $41 = (_isspace($38)|0); //@line 2360 "load_abc.cpp"
    $42 = ($41|0)==(0); //@line 2360 "load_abc.cpp"
    if (!($42)) {
     $$0 = 1;
     label = 24;
     break;
    }
   }
  }
  $23 = ($22|0)>(-1); //@line 481 "load_abc.cpp"
  $not$$i$i = ($22|0)<($dwMemLength|0); //@line 482 "load_abc.cpp"
  $or$cond10 = $23 & $not$$i$i; //@line 481 "load_abc.cpp"
  if ($or$cond10) {
   $3 = $22;$i$06$i$i = 0;
  } else {
   $$0 = 0;
   label = 24;
   break;
  }
 }
 if ((label|0) == 24) {
  STACKTOP = sp;return ($$0|0); //@line 2363 "load_abc.cpp"
 }
 return (0)|0;
}
function _pat_resetsmp() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 _memset((1252508|0),0,191)|0; //@line 194 "load_pat.cpp"
 _memset((1252699|0),0,191)|0; //@line 195 "load_pat.cpp"
 return; //@line 197 "load_pat.cpp"
}
function _pat_init_patnames() {
 var $$isdrumset$0 = 0, $$lcssa = 0, $$lcssa144 = 0, $$lcssa146 = 0, $$nskip$0 = 0, $$op = 0, $$pre = 0, $$pre100 = 0, $$pre99$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cfgsources = 0, $endptr = 0, $endptr26 = 0, $exitcond = 0, $exitcond95 = 0, $exitcond96 = 0;
 var $i$076 = 0, $i$138 = 0, $isdigit = 0, $isdigit22 = 0, $isdigit24 = 0, $isdigittmp = 0, $isdigittmp21 = 0, $isdigittmp23 = 0, $isdrumset$071 = 0, $line = 0, $notlhs = 0, $nskip$039 = 0, $nskip$1 = 0, $nskip$1$lcssa = 0, $nsources$074 = 0, $nsources$1$lcssa = 0, $nsources$170 = 0, $nsources$2 = 0, $nsources$3 = 0, $or$cond11 = 0;
 var $or$cond16 = 0, $or$cond18 = 0, $or$cond20 = 0, $or$cond7 = 0, $p$0 = 0, $p$0$lcssa = 0, $p$1 = 0, $p$1$lcssa = 0, $p$1$pn = 0, $p$2$lcssa = 0, $p$3$lcssa = 0, $p$342 = 0, $p$467 = 0, $p$5 = 0, $p$5$lcssa = 0, $p$656 = 0, $p$656$lcssa = 0, $p$7 = 0, $p$8 = 0, $p$8$lcssa = 0;
 var $p$9 = 0, $pfnlen$0$lcssa = 0, $pfnlen$041 = 0, $pfnlen$1$be = 0, $pfnlen$166 = 0, $pfnlen$255 = 0, $pfnlen$255$lcssa = 0, $pfnlen$3 = 0, $pfnlen$4 = 0, $q$0$lcssa = 0, $q$043 = 0, $q$1$be = 0, $q$168 = 0, $q$257 = 0, $q$257$lcssa = 0, $q$3 = 0, $q$4 = 0, $q$5 = 0, $q$5$lcssa = 0, $q$640 = 0;
 var $q$7 = 0, $q$7$lcssa = 0, $q$8$ph37 = 0, $strlen = 0, $strlen25 = 0, $txt$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $z$075 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 24848|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $txt$i = sp + 24592|0;
 $line = sp + 20496|0;
 $cfgsources = sp + 16|0;
 _memset(($cfgsources|0),0,20480)|0; //@line 389 "load_pat.cpp"
 dest=1252890; src=524301; stop=dest+38|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 391 "load_pat.cpp"
 dest=1253018; src=524339; stop=dest+39|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 392 "load_pat.cpp"
 $0 = (_getenv((524378|0))|0); //@line 393 "load_pat.cpp"
 $1 = ($0|0)==(0|0); //@line 394 "load_pat.cpp"
 if (!($1)) {
  (_strcpy(1253018,$0)|0); //@line 395 "load_pat.cpp"
  (_strcpy(1252890,$0)|0); //@line 396 "load_pat.cpp"
  $strlen = (_strlen(1253018)|0); //@line 397 "load_pat.cpp"
  $endptr = (1253018 + ($strlen)|0); //@line 397 "load_pat.cpp"
  dest=$endptr; src=524396; stop=dest+14|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 397 "load_pat.cpp"
  $strlen25 = (_strlen(1252890)|0); //@line 398 "load_pat.cpp"
  $endptr26 = (1252890 + ($strlen25)|0); //@line 398 "load_pat.cpp"
  dest=$endptr26; src=524410; stop=dest+13|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 398 "load_pat.cpp"
 }
 (_strncpy($cfgsources,1253018,4096)|0); //@line 400 "load_pat.cpp"
 $i$076 = 0;
 while(1) {
  $5 = (1253146 + ($i$076<<7)|0); //@line 403 "load_pat.cpp"
  HEAP8[$5>>0] = 0; //@line 403 "load_pat.cpp"
  $6 = (($i$076) + 1)|0; //@line 403 "load_pat.cpp"
  $exitcond96 = ($6|0)==(191); //@line 403 "load_pat.cpp"
  if ($exitcond96) {
   break;
  } else {
   $i$076 = $6;
  }
 }
 $2 = HEAP32[37544]|0; //@line 186 "load_pat.cpp"
 $3 = ((($line)) + 7|0); //@line 450 "load_pat.cpp"
 $4 = ((($line)) + 1|0); //@line 416 "load_pat.cpp"
 $nsources$074 = 1;$z$075 = 0;
 while(1) {
  $7 = (($cfgsources) + ($z$075<<12)|0); //@line 406 "load_pat.cpp"
  $8 = HEAP8[$7>>0]|0; //@line 406 "load_pat.cpp"
  $9 = ($8<<24>>24)==(0); //@line 406 "load_pat.cpp"
  do {
   if ($9) {
    $nsources$3 = $nsources$074;
   } else {
    $10 = (_fopen($7,527143)|0); //@line 407 "load_pat.cpp"
    $11 = ($10|0)==(0|0); //@line 408 "load_pat.cpp"
    if ($11) {
     $12 = (_strlen($7)|0); //@line 181 "load_pat.cpp"
     $13 = (($12) + 77)|0; //@line 181 "load_pat.cpp"
     $14 = ($13>>>0)>(255); //@line 181 "load_pat.cpp"
     if (!($14)) {
      HEAP32[$vararg_buffer>>2] = $7; //@line 182 "load_pat.cpp"
      (_sprintf($txt$i,524423,$vararg_buffer)|0); //@line 182 "load_pat.cpp"
      HEAP32[$vararg_buffer1>>2] = $txt$i; //@line 186 "load_pat.cpp"
      (_fprintf($2,524501,$vararg_buffer1)|0); //@line 186 "load_pat.cpp"
     }
     $nsources$3 = $nsources$074;
     break;
    }
    (_fgets($line,4096,$10)|0); //@line 414 "load_pat.cpp"
    $15 = (_feof($10)|0); //@line 415 "load_pat.cpp"
    $16 = ($15|0)==(0); //@line 415 "load_pat.cpp"
    if ($16) {
     $isdrumset$071 = 0;$nsources$170 = $nsources$074;
     while(1) {
      $17 = HEAP8[$line>>0]|0; //@line 416 "load_pat.cpp"
      $18 = $17 << 24 >> 24; //@line 416 "load_pat.cpp"
      $isdigittmp = (($18) + -48)|0; //@line 416 "load_pat.cpp"
      $isdigit = ($isdigittmp>>>0)<(10); //@line 416 "load_pat.cpp"
      if ($isdigit) {
       $p$0 = $line;
       label = 15;
      } else {
       $19 = (_isblank($18)|0); //@line 416 "load_pat.cpp"
       $20 = ($19|0)==(0); //@line 416 "load_pat.cpp"
       if (!($20)) {
        $21 = HEAP8[$4>>0]|0; //@line 416 "load_pat.cpp"
        $22 = $21 << 24 >> 24; //@line 416 "load_pat.cpp"
        $isdigittmp21 = (($22) + -48)|0; //@line 416 "load_pat.cpp"
        $isdigit22 = ($isdigittmp21>>>0)<(10); //@line 416 "load_pat.cpp"
        if ($isdigit22) {
         $p$0 = $line;
         label = 15;
        }
       }
      }
      L23: do {
       if ((label|0) == 15) {
        while(1) {
         label = 0;
         $23 = HEAP8[$p$0>>0]|0; //@line 419 "load_pat.cpp"
         $24 = $23 << 24 >> 24; //@line 419 "load_pat.cpp"
         $25 = (_isspace($24)|0); //@line 419 "load_pat.cpp"
         $26 = ($25|0)==(0); //@line 419 "load_pat.cpp"
         $27 = ((($p$0)) + 1|0); //@line 419 "load_pat.cpp"
         if ($26) {
          $p$0$lcssa = $p$0;
          break;
         } else {
          $p$0 = $27;
          label = 15;
         }
        }
        $28 = (_atoi($p$0$lcssa)|0); //@line 420 "load_pat.cpp"
        $p$1 = $p$0$lcssa;
        while(1) {
         $29 = HEAP8[$p$1>>0]|0; //@line 421 "load_pat.cpp"
         $30 = $29 << 24 >> 24; //@line 421 "load_pat.cpp"
         $isdigittmp23 = (($30) + -48)|0; //@line 421 "load_pat.cpp"
         $isdigit24 = ($isdigittmp23>>>0)<(10); //@line 421 "load_pat.cpp"
         $31 = ((($p$1)) + 1|0); //@line 421 "load_pat.cpp"
         if ($isdigit24) {
          $p$1 = $31;
         } else {
          $$lcssa146 = $29;$p$1$lcssa = $p$1;
          break;
         }
        }
        $32 = $$lcssa146 << 24 >> 24; //@line 422 "load_pat.cpp"
        $33 = (_isspace($32)|0); //@line 422 "load_pat.cpp"
        $34 = ($33|0)==(0); //@line 422 "load_pat.cpp"
        if ($34) {
         $p$2$lcssa = $p$1$lcssa;
        } else {
         $p$1$pn = $p$1$lcssa;
         while(1) {
          $35 = ((($p$1$pn)) + 1|0); //@line 422 "load_pat.cpp"
          $$pre = HEAP8[$35>>0]|0; //@line 422 "load_pat.cpp"
          $36 = $$pre << 24 >> 24; //@line 422 "load_pat.cpp"
          $37 = (_isspace($36)|0); //@line 422 "load_pat.cpp"
          $38 = ($37|0)==(0); //@line 422 "load_pat.cpp"
          if ($38) {
           $p$2$lcssa = $35;
           break;
          } else {
           $p$1$pn = $35;
          }
         }
        }
        $notlhs = ($28>>>0)>(190); //@line 424 "load_pat.cpp"
        if (!($notlhs)) {
         $39 = HEAP8[$p$2$lcssa>>0]|0; //@line 424 "load_pat.cpp"
         switch ($39<<24>>24) {
         case 0: case 35:  {
          break L23;
          break;
         }
         default: {
         }
         }
         $40 = ($isdrumset$071|0)==(0); //@line 425 "load_pat.cpp"
         if ($40) {
          $41 = (1253146 + ($28<<7)|0); //@line 425 "load_pat.cpp"
          $48 = $39;$p$342 = $p$2$lcssa;$pfnlen$041 = 0;$q$043 = $41;
         } else {
          $42 = ($28|0)<(25); //@line 247 "load_pat.cpp"
          if ($42) {
           $48 = $39;$p$342 = $p$2$lcssa;$pfnlen$041 = 0;$q$043 = (1269530);
          } else {
           $43 = (($28) + 104)|0; //@line 248 "load_pat.cpp"
           $44 = ($43|0)<(191); //@line 248 "load_pat.cpp"
           $$op = (($28) + 103)|0; //@line 425 "load_pat.cpp"
           $45 = $44 ? $$op : 190; //@line 425 "load_pat.cpp"
           $46 = (1253146 + ($45<<7)|0); //@line 425 "load_pat.cpp"
           $48 = $39;$p$342 = $p$2$lcssa;$pfnlen$041 = 0;$q$043 = $46;
          }
         }
         while(1) {
          $47 = $48 << 24 >> 24; //@line 427 "load_pat.cpp"
          $49 = (_isspace($47)|0); //@line 427 "load_pat.cpp"
          $50 = ($49|0)==(0); //@line 427 "load_pat.cpp"
          $51 = HEAP8[$p$342>>0]|0; //@line 427 "load_pat.cpp"
          if (!($50)) {
           $60 = $51;$p$3$lcssa = $p$342;$pfnlen$0$lcssa = $pfnlen$041;$q$0$lcssa = $q$043;
           break;
          }
          $52 = ($51<<24>>24)!=(35); //@line 427 "load_pat.cpp"
          $53 = ($pfnlen$041|0)<(128); //@line 427 "load_pat.cpp"
          $or$cond16 = $53 & $52; //@line 427 "load_pat.cpp"
          if (!($or$cond16)) {
           $60 = $51;$p$3$lcssa = $p$342;$pfnlen$0$lcssa = $pfnlen$041;$q$0$lcssa = $q$043;
           break;
          }
          $54 = (($pfnlen$041) + 1)|0; //@line 428 "load_pat.cpp"
          $55 = ((($p$342)) + 1|0); //@line 429 "load_pat.cpp"
          $56 = ((($q$043)) + 1|0); //@line 429 "load_pat.cpp"
          HEAP8[$q$043>>0] = $51; //@line 429 "load_pat.cpp"
          $57 = HEAP8[$55>>0]|0; //@line 427 "load_pat.cpp"
          $58 = ($57<<24>>24)==(0); //@line 427 "load_pat.cpp"
          if ($58) {
           $60 = 0;$p$3$lcssa = $55;$pfnlen$0$lcssa = $54;$q$0$lcssa = $56;
           break;
          } else {
           $48 = $57;$p$342 = $55;$pfnlen$041 = $54;$q$043 = $56;
          }
         }
         $59 = $60 << 24 >> 24; //@line 431 "load_pat.cpp"
         $61 = (_isblank($59)|0); //@line 431 "load_pat.cpp"
         $62 = ($61|0)==(0); //@line 431 "load_pat.cpp"
         if ($62) {
          $q$4 = $q$0$lcssa;
         } else {
          $63 = ((($p$3$lcssa)) + 1|0); //@line 431 "load_pat.cpp"
          $64 = HEAP8[$63>>0]|0; //@line 431 "load_pat.cpp"
          $65 = ($64<<24>>24)!=(35); //@line 431 "load_pat.cpp"
          $66 = ($pfnlen$0$lcssa|0)<(128); //@line 431 "load_pat.cpp"
          $or$cond7 = $66 & $65; //@line 431 "load_pat.cpp"
          if ($or$cond7) {
           $67 = ((($q$0$lcssa)) + 1|0); //@line 432 "load_pat.cpp"
           HEAP8[$q$0$lcssa>>0] = 58; //@line 432 "load_pat.cpp"
           $68 = HEAP8[$p$3$lcssa>>0]|0; //@line 433 "load_pat.cpp"
           $69 = $68 << 24 >> 24; //@line 433 "load_pat.cpp"
           $70 = (_isspace($69)|0); //@line 433 "load_pat.cpp"
           $71 = ($70|0)==(0); //@line 433 "load_pat.cpp"
           if ($71) {
            $q$4 = $67;
           } else {
            $72 = (($pfnlen$0$lcssa) + 1)|0; //@line 432 "load_pat.cpp"
            $p$467 = $p$3$lcssa;$pfnlen$166 = $72;$q$168 = $67;
            while(1) {
             $p$5 = $p$467;
             while(1) {
              $73 = HEAP8[$p$5>>0]|0; //@line 434 "load_pat.cpp"
              $74 = $73 << 24 >> 24; //@line 434 "load_pat.cpp"
              $75 = (_isspace($74)|0); //@line 434 "load_pat.cpp"
              $76 = ($75|0)==(0); //@line 434 "load_pat.cpp"
              $77 = ((($p$5)) + 1|0); //@line 434 "load_pat.cpp"
              if ($76) {
               $p$5$lcssa = $p$5;
               break;
              } else {
               $p$5 = $77;
              }
             }
             $78 = HEAP8[$p$5$lcssa>>0]|0; //@line 435 "load_pat.cpp"
             $79 = ($78<<24>>24)==(35); //@line 435 "load_pat.cpp"
             L54: do {
              if ($79) {
               $93 = 35;$p$7 = $p$5$lcssa;$pfnlen$3 = $pfnlen$166;$q$3 = $q$168;
              } else {
               $80 = ($78<<24>>24)==(0); //@line 437 "load_pat.cpp"
               if ($80) {
                $93 = 0;$p$7 = $p$5$lcssa;$pfnlen$3 = $pfnlen$166;$q$3 = $q$168;
                break;
               } else {
                $82 = $78;$p$656 = $p$5$lcssa;$pfnlen$255 = $pfnlen$166;$q$257 = $q$168;
               }
               while(1) {
                $81 = $82 << 24 >> 24; //@line 437 "load_pat.cpp"
                $83 = (_isspace($81)|0); //@line 437 "load_pat.cpp"
                $84 = ($83|0)==(0); //@line 437 "load_pat.cpp"
                $85 = ($pfnlen$255|0)<(128); //@line 437 "load_pat.cpp"
                $or$cond18 = $85 & $84; //@line 437 "load_pat.cpp"
                if (!($or$cond18)) {
                 $p$656$lcssa = $p$656;$pfnlen$255$lcssa = $pfnlen$255;$q$257$lcssa = $q$257;
                 break;
                }
                $86 = (($pfnlen$255) + 1)|0; //@line 438 "load_pat.cpp"
                $87 = ((($p$656)) + 1|0); //@line 439 "load_pat.cpp"
                $88 = HEAP8[$p$656>>0]|0; //@line 439 "load_pat.cpp"
                $89 = ((($q$257)) + 1|0); //@line 439 "load_pat.cpp"
                HEAP8[$q$257>>0] = $88; //@line 439 "load_pat.cpp"
                $90 = HEAP8[$87>>0]|0; //@line 437 "load_pat.cpp"
                $91 = ($90<<24>>24)==(0); //@line 437 "load_pat.cpp"
                if ($91) {
                 $93 = 0;$p$7 = $87;$pfnlen$3 = $86;$q$3 = $89;
                 break L54;
                } else {
                 $82 = $90;$p$656 = $87;$pfnlen$255 = $86;$q$257 = $89;
                }
               }
               $$pre99$pre = HEAP8[$p$656$lcssa>>0]|0; //@line 441 "load_pat.cpp"
               $93 = $$pre99$pre;$p$7 = $p$656$lcssa;$pfnlen$3 = $pfnlen$255$lcssa;$q$3 = $q$257$lcssa;
              }
             } while(0);
             $92 = $93 << 24 >> 24; //@line 441 "load_pat.cpp"
             $94 = (_isspace($92)|0); //@line 441 "load_pat.cpp"
             $95 = ($94|0)==(0); //@line 441 "load_pat.cpp"
             if ($95) {
              $pfnlen$1$be = $pfnlen$3;$q$1$be = $q$3;
             } else {
              $96 = ((($q$3)) + 1|0); //@line 441 "load_pat.cpp"
              HEAP8[$q$3>>0] = 32; //@line 441 "load_pat.cpp"
              $97 = (($pfnlen$3) + 1)|0; //@line 441 "load_pat.cpp"
              $pfnlen$1$be = $97;$q$1$be = $96;
             }
             $98 = HEAP8[$p$7>>0]|0; //@line 433 "load_pat.cpp"
             $99 = $98 << 24 >> 24; //@line 433 "load_pat.cpp"
             $100 = (_isspace($99)|0); //@line 433 "load_pat.cpp"
             $101 = ($100|0)==(0); //@line 433 "load_pat.cpp"
             if ($101) {
              $q$4 = $q$1$be;
              break;
             } else {
              $p$467 = $p$7;$pfnlen$166 = $pfnlen$1$be;$q$168 = $q$1$be;
             }
            }
           }
          } else {
           $q$4 = $q$0$lcssa;
          }
         }
         HEAP8[$q$4>>0] = 0; //@line 444 "load_pat.cpp"
        }
       }
      } while(0);
      $102 = (_strncmp($line,524516,7)|0); //@line 447 "load_pat.cpp"
      $103 = ($102|0)==(0); //@line 447 "load_pat.cpp"
      $$isdrumset$0 = $103 ? 1 : $isdrumset$071; //@line 447 "load_pat.cpp"
      $104 = (_strncmp($line,524524,6)|0); //@line 448 "load_pat.cpp"
      $105 = ($104|0)==(0); //@line 448 "load_pat.cpp"
      $106 = ($nsources$170|0)<(5); //@line 448 "load_pat.cpp"
      $or$cond11 = $106 & $105; //@line 448 "load_pat.cpp"
      if ($or$cond11) {
       $p$8 = $3;
       while(1) {
        $107 = HEAP8[$p$8>>0]|0; //@line 451 "load_pat.cpp"
        $108 = $107 << 24 >> 24; //@line 451 "load_pat.cpp"
        $109 = (_isspace($108)|0); //@line 451 "load_pat.cpp"
        $110 = ($109|0)==(0); //@line 451 "load_pat.cpp"
        $111 = ((($p$8)) + 1|0); //@line 451 "load_pat.cpp"
        if ($110) {
         $p$8$lcssa = $p$8;
         break;
        } else {
         $p$8 = $111;
        }
       }
       $112 = (($cfgsources) + ($nsources$170<<12)|0); //@line 449 "load_pat.cpp"
       $p$9 = $p$8$lcssa;$pfnlen$4 = 0;$q$5 = $112;
       L70: while(1) {
        $113 = HEAP8[$p$9>>0]|0; //@line 453 "load_pat.cpp"
        switch ($113<<24>>24) {
        case 35: case 0:  {
         $q$5$lcssa = $q$5;
         break L70;
         break;
        }
        default: {
        }
        }
        $114 = $113 << 24 >> 24; //@line 453 "load_pat.cpp"
        $115 = (_isspace($114)|0); //@line 453 "load_pat.cpp"
        $116 = ($115|0)==(0); //@line 453 "load_pat.cpp"
        $117 = ($pfnlen$4|0)<(128); //@line 453 "load_pat.cpp"
        $or$cond20 = $117 & $116; //@line 453 "load_pat.cpp"
        if (!($or$cond20)) {
         $q$5$lcssa = $q$5;
         break;
        }
        $118 = (($pfnlen$4) + 1)|0; //@line 454 "load_pat.cpp"
        $119 = ((($p$9)) + 1|0); //@line 455 "load_pat.cpp"
        $120 = HEAP8[$p$9>>0]|0; //@line 455 "load_pat.cpp"
        $121 = ((($q$5)) + 1|0); //@line 455 "load_pat.cpp"
        HEAP8[$q$5>>0] = $120; //@line 455 "load_pat.cpp"
        $p$9 = $119;$pfnlen$4 = $118;$q$5 = $121;
       }
       HEAP8[$q$5$lcssa>>0] = 0; //@line 457 "load_pat.cpp"
       $122 = (($nsources$170) + 1)|0; //@line 458 "load_pat.cpp"
       $nsources$2 = $122;
      } else {
       $nsources$2 = $nsources$170;
      }
      (_fgets($line,4096,$10)|0); //@line 460 "load_pat.cpp"
      $123 = (_feof($10)|0); //@line 415 "load_pat.cpp"
      $124 = ($123|0)==(0); //@line 415 "load_pat.cpp"
      if ($124) {
       $isdrumset$071 = $$isdrumset$0;$nsources$170 = $nsources$2;
      } else {
       $nsources$1$lcssa = $nsources$2;
       break;
      }
     }
    } else {
     $nsources$1$lcssa = $nsources$074;
    }
    (_fclose($10)|0); //@line 463 "load_pat.cpp"
    $nsources$3 = $nsources$1$lcssa;
   }
  } while(0);
  $125 = (($z$075) + 1)|0; //@line 405 "load_pat.cpp"
  $exitcond95 = ($125|0)==(5); //@line 405 "load_pat.cpp"
  if ($exitcond95) {
   $i$138 = 0;$nskip$039 = 0;$q$640 = 1253146;
   break;
  } else {
   $nsources$074 = $nsources$3;$z$075 = $125;
  }
 }
 while(1) {
  $126 = (1253146 + ($i$138<<7)|0); //@line 470 "load_pat.cpp"
  $127 = HEAP8[$126>>0]|0; //@line 470 "load_pat.cpp"
  $128 = ($127<<24>>24)==(0); //@line 470 "load_pat.cpp"
  if ($128) {
   $129 = ($126|0)==($q$640|0); //@line 472 "load_pat.cpp"
   if ($129) {
    $131 = 0;
   } else {
    (_strcpy($126,$q$640)|0); //@line 473 "load_pat.cpp"
    $$pre100 = HEAP8[$126>>0]|0; //@line 474 "load_pat.cpp"
    $131 = $$pre100;
   }
   $130 = ($131<<24>>24)==(0); //@line 474 "load_pat.cpp"
   $132 = $130&1; //@line 474 "load_pat.cpp"
   $$nskip$0 = (($132) + ($nskip$039))|0; //@line 474 "load_pat.cpp"
   $nskip$1 = $$nskip$0;$q$7 = $q$640;
  } else {
   $nskip$1 = $nskip$039;$q$7 = $126;
  }
  $133 = (($i$138) + 1)|0; //@line 469 "load_pat.cpp"
  $exitcond = ($133|0)==(191); //@line 469 "load_pat.cpp"
  if ($exitcond) {
   $nskip$1$lcssa = $nskip$1;$q$7$lcssa = $q$7;
   break;
  } else {
   $i$138 = $133;$nskip$039 = $nskip$1;$q$640 = $q$7;
  }
 }
 $134 = ($nskip$1$lcssa|0)==(0); //@line 477 "load_pat.cpp"
 if ($134) {
  STACKTOP = sp;return; //@line 484 "load_pat.cpp"
 } else {
  $144 = 190;$q$8$ph37 = $q$7$lcssa;
 }
 L89: while(1) {
  $138 = $144;
  while(1) {
   $137 = (1253146 + ($138<<7)|0); //@line 479 "load_pat.cpp"
   $139 = HEAP8[$137>>0]|0; //@line 479 "load_pat.cpp"
   $140 = ($139<<24>>24)==(0); //@line 479 "load_pat.cpp"
   if (!($140)) {
    $$lcssa = $138;$$lcssa144 = $137;
    break;
   }
   $141 = ($137|0)==($q$8$ph37|0); //@line 480 "load_pat.cpp"
   if (!($141)) {
    (_strcpy($137,$q$8$ph37)|0); //@line 481 "load_pat.cpp"
   }
   $142 = (($138) + -1)|0; //@line 478 "load_pat.cpp"
   $143 = ($138|0)>(0); //@line 478 "load_pat.cpp"
   if ($143) {
    $138 = $142;
   } else {
    label = 66;
    break L89;
   }
  }
  $135 = (($$lcssa) + -1)|0; //@line 478 "load_pat.cpp"
  $136 = ($$lcssa|0)>(0); //@line 478 "load_pat.cpp"
  if ($136) {
   $144 = $135;$q$8$ph37 = $$lcssa144;
  } else {
   label = 66;
   break;
  }
 }
 if ((label|0) == 66) {
  STACKTOP = sp;return; //@line 484 "load_pat.cpp"
 }
}
function __ZL14abc_new_umacroP10_ABCHANDLEPKc($h,$m) {
 $h = $h|0;
 $m = $m|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $key = 0, $memchr = 0, $mp$04 = 0, $mp$04$lcssa = 0, $mp$04$phi = 0, $retval$0 = 0, $retval$03 = 0, $retval$05 = 0, $retval$05$lcssa = 0;
 var $value = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 512|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $key = sp + 256|0;
 $value = sp;
 __ZL19abc_extractkeyvaluePcjS_jPKc($key,$value,$m); //@line 586 "load_abc.cpp"
 $0 = (_strlen($key)|0); //@line 587 "load_abc.cpp"
 $1 = ($0>>>0)>(1); //@line 587 "load_abc.cpp"
 if ($1) {
  STACKTOP = sp;return; //@line 609 "load_abc.cpp"
 }
 $2 = HEAP8[$key>>0]|0; //@line 587 "load_abc.cpp"
 $3 = $2 << 24 >> 24; //@line 587 "load_abc.cpp"
 $4 = (_toupper($3)|0); //@line 587 "load_abc.cpp"
 $memchr = (_memchr(524545,$4,20)|0); //@line 587 "load_abc.cpp"
 $5 = ($memchr|0)==(0|0); //@line 587 "load_abc.cpp"
 if ($5) {
  STACKTOP = sp;return; //@line 609 "load_abc.cpp"
 }
 $6 = (_strchr($key,33)|0); //@line 588 "load_abc.cpp"
 $7 = ($6|0)==(0|0); //@line 588 "load_abc.cpp"
 if (!($7)) {
  $8 = $6;
  while(1) {
   HEAP8[$8>>0] = 43; //@line 589 "load_abc.cpp"
   $9 = (_strchr($key,33)|0); //@line 588 "load_abc.cpp"
   $10 = ($9|0)==(0|0); //@line 588 "load_abc.cpp"
   if ($10) {
    break;
   } else {
    $8 = $9;
   }
  }
 }
 $11 = (_strcmp($key,524565)|0); //@line 590 "load_abc.cpp"
 $12 = ($11|0)==(0); //@line 590 "load_abc.cpp"
 if (!($12)) {
  $23 = (_calloc(1,66)|0); //@line 603 "load_abc.cpp"
  $24 = (___strdup($key)|0); //@line 604 "load_abc.cpp"
  $25 = ((($23)) + 4|0); //@line 604 "load_abc.cpp"
  HEAP8[$25>>0]=$24&255;HEAP8[$25+1>>0]=($24>>8)&255;HEAP8[$25+2>>0]=($24>>16)&255;HEAP8[$25+3>>0]=$24>>24; //@line 604 "load_abc.cpp"
  $26 = (___strdup($value)|0); //@line 605 "load_abc.cpp"
  $27 = ((($23)) + 8|0); //@line 605 "load_abc.cpp"
  HEAP8[$27>>0]=$26&255;HEAP8[$27+1>>0]=($26>>8)&255;HEAP8[$27+2>>0]=($26>>16)&255;HEAP8[$27+3>>0]=$26>>24; //@line 605 "load_abc.cpp"
  $28 = ((($23)) + 12|0); //@line 606 "load_abc.cpp"
  HEAP8[$28>>0]=0&255;HEAP8[$28+1>>0]=(0>>8)&255;HEAP8[$28+2>>0]=(0>>16)&255;HEAP8[$28+3>>0]=0>>24; //@line 606 "load_abc.cpp"
  $29 = ((($h)) + 4|0); //@line 607 "load_abc.cpp"
  $30 = HEAPU8[$29>>0]|(HEAPU8[$29+1>>0]<<8)|(HEAPU8[$29+2>>0]<<16)|(HEAPU8[$29+3>>0]<<24); //@line 607 "load_abc.cpp"
  HEAP8[$23>>0]=$30&255;HEAP8[$23+1>>0]=($30>>8)&255;HEAP8[$23+2>>0]=($30>>16)&255;HEAP8[$23+3>>0]=$30>>24; //@line 607 "load_abc.cpp"
  HEAP8[$29>>0]=$23&255;HEAP8[$29+1>>0]=($23>>8)&255;HEAP8[$29+2>>0]=($23>>16)&255;HEAP8[$29+3>>0]=$23>>24; //@line 608 "load_abc.cpp"
  STACKTOP = sp;return; //@line 609 "load_abc.cpp"
 }
 $13 = ((($h)) + 4|0); //@line 592 "load_abc.cpp"
 $retval$03 = HEAPU8[$13>>0]|(HEAPU8[$13+1>>0]<<8)|(HEAPU8[$13+2>>0]<<16)|(HEAPU8[$13+3>>0]<<24); //@line 592 "load_abc.cpp"
 $14 = ($retval$03|0)==(0|0); //@line 592 "load_abc.cpp"
 if ($14) {
  STACKTOP = sp;return; //@line 609 "load_abc.cpp"
 }
 $15 = HEAP8[$key>>0]|0; //@line 593 "load_abc.cpp"
 $mp$04 = 0;$retval$05 = $retval$03;
 while(1) {
  $16 = ((($retval$05)) + 4|0); //@line 593 "load_abc.cpp"
  $17 = HEAPU8[$16>>0]|(HEAPU8[$16+1>>0]<<8)|(HEAPU8[$16+2>>0]<<16)|(HEAPU8[$16+3>>0]<<24); //@line 593 "load_abc.cpp"
  $18 = HEAP8[$17>>0]|0; //@line 593 "load_abc.cpp"
  $19 = ($18<<24>>24)==($15<<24>>24); //@line 593 "load_abc.cpp"
  if ($19) {
   $mp$04$lcssa = $mp$04;$retval$05$lcssa = $retval$05;
   break;
  }
  $retval$0 = HEAPU8[$retval$05>>0]|(HEAPU8[$retval$05+1>>0]<<8)|(HEAPU8[$retval$05+2>>0]<<16)|(HEAPU8[$retval$05+3>>0]<<24); //@line 592 "load_abc.cpp"
  $22 = ($retval$0|0)==(0|0); //@line 592 "load_abc.cpp"
  if ($22) {
   label = 15;
   break;
  } else {
   $mp$04$phi = $retval$05;$retval$05 = $retval$0;$mp$04 = $mp$04$phi;
  }
 }
 if ((label|0) == 15) {
  STACKTOP = sp;return; //@line 609 "load_abc.cpp"
 }
 $20 = ($mp$04$lcssa|0)==(0|0); //@line 594 "load_abc.cpp"
 $21 = HEAPU8[$retval$05$lcssa>>0]|(HEAPU8[$retval$05$lcssa+1>>0]<<8)|(HEAPU8[$retval$05$lcssa+2>>0]<<16)|(HEAPU8[$retval$05$lcssa+3>>0]<<24); //@line 594 "load_abc.cpp"
 if ($20) {
  HEAP8[$13>>0]=$21&255;HEAP8[$13+1>>0]=($21>>8)&255;HEAP8[$13+2>>0]=($21>>16)&255;HEAP8[$13+3>>0]=$21>>24; //@line 595 "load_abc.cpp"
 } else {
  HEAP8[$mp$04$lcssa>>0]=$21&255;HEAP8[$mp$04$lcssa+1>>0]=($21>>8)&255;HEAP8[$mp$04$lcssa+2>>0]=($21>>16)&255;HEAP8[$mp$04$lcssa+3>>0]=$21>>24; //@line 594 "load_abc.cpp"
 }
 _free($retval$05$lcssa); //@line 596 "load_abc.cpp"
 STACKTOP = sp;return; //@line 609 "load_abc.cpp"
}
function __ZL19abc_extractkeyvaluePcjS_jPKc($key,$value,$src) {
 $key = $key|0;
 $value = $value|0;
 $src = $src|0;
 var $$013 = 0, $$111 = 0, $$2$ph = 0, $$38 = 0, $$46 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $key_size$0$lcssa = 0, $key_size$012 = 0, $key_size$1 = 0, $key_size$1$lcssa = 0, $value_size$07 = 0, $value_size$1 = 0, $value_size$1$lcssa = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$src>>0]|0; //@line 294 "load_abc.cpp"
 $1 = ($0<<24>>24)==(0); //@line 294 "load_abc.cpp"
 L1: do {
  if ($1) {
   $$111 = $src;$key_size$012 = 0;
  } else {
   $$013 = $src;$3 = $0;
   while(1) {
    $2 = $3 << 24 >> 24; //@line 294 "load_abc.cpp"
    $4 = (_isspace($2)|0); //@line 294 "load_abc.cpp"
    $5 = ($4|0)==(0); //@line 294 "load_abc.cpp"
    if ($5) {
     $$111 = $$013;$key_size$012 = 0;
     break L1;
    }
    $6 = ((($$013)) + 1|0); //@line 295 "load_abc.cpp"
    $7 = HEAP8[$6>>0]|0; //@line 294 "load_abc.cpp"
    $8 = ($7<<24>>24)==(0); //@line 294 "load_abc.cpp"
    if ($8) {
     $$111 = $6;$key_size$012 = 0;
     break;
    } else {
     $$013 = $6;$3 = $7;
    }
   }
  }
 } while(0);
 while(1) {
  $9 = HEAP8[$$111>>0]|0; //@line 298 "load_abc.cpp"
  $10 = ($9<<24>>24)==(0); //@line 298 "load_abc.cpp"
  if ($10) {
   $$2$ph = $$111;$key_size$0$lcssa = $key_size$012;
   break;
  }
  $11 = ($9<<24>>24)==(61); //@line 299 "load_abc.cpp"
  $12 = ((($$111)) + 1|0); //@line 300 "load_abc.cpp"
  if ($11) {
   $$2$ph = $12;$key_size$0$lcssa = $key_size$012;
   break;
  }
  $13 = (($key_size$012) + 1)|0; //@line 303 "load_abc.cpp"
  $14 = (($key) + ($key_size$012)|0); //@line 303 "load_abc.cpp"
  HEAP8[$14>>0] = $9; //@line 303 "load_abc.cpp"
  $15 = ($13>>>0)<(255); //@line 298 "load_abc.cpp"
  if ($15) {
   $$111 = $12;$key_size$012 = $13;
  } else {
   $$2$ph = $12;$key_size$0$lcssa = $13;
   break;
  }
 }
 $key_size$1 = $key_size$0$lcssa;
 while(1) {
  $16 = ($key_size$1|0)==(0); //@line 305 "load_abc.cpp"
  if ($16) {
   $key_size$1$lcssa = 0;
   break;
  }
  $17 = (($key_size$1) + -1)|0; //@line 305 "load_abc.cpp"
  $18 = (($key) + ($17)|0); //@line 305 "load_abc.cpp"
  $19 = HEAP8[$18>>0]|0; //@line 305 "load_abc.cpp"
  $20 = $19 << 24 >> 24; //@line 305 "load_abc.cpp"
  $21 = (_isspace($20)|0); //@line 305 "load_abc.cpp"
  $22 = ($21|0)==(0); //@line 305 "load_abc.cpp"
  if ($22) {
   $key_size$1$lcssa = $key_size$1;
   break;
  } else {
   $key_size$1 = $17;
  }
 }
 $23 = (($key) + ($key_size$1$lcssa)|0); //@line 307 "load_abc.cpp"
 HEAP8[$23>>0] = 0; //@line 307 "load_abc.cpp"
 $24 = HEAP8[$$2$ph>>0]|0; //@line 309 "load_abc.cpp"
 $25 = ($24<<24>>24)==(0); //@line 309 "load_abc.cpp"
 L14: do {
  if ($25) {
   $$46 = $$2$ph;$value_size$07 = 0;
  } else {
   $$38 = $$2$ph;$27 = $24;
   while(1) {
    $26 = $27 << 24 >> 24; //@line 309 "load_abc.cpp"
    $28 = (_isspace($26)|0); //@line 309 "load_abc.cpp"
    $29 = ($28|0)==(0); //@line 309 "load_abc.cpp"
    if ($29) {
     $$46 = $$38;$value_size$07 = 0;
     break L14;
    }
    $30 = ((($$38)) + 1|0); //@line 310 "load_abc.cpp"
    $31 = HEAP8[$30>>0]|0; //@line 309 "load_abc.cpp"
    $32 = ($31<<24>>24)==(0); //@line 309 "load_abc.cpp"
    if ($32) {
     $$46 = $30;$value_size$07 = 0;
     break;
    } else {
     $$38 = $30;$27 = $31;
    }
   }
  }
 } while(0);
 while(1) {
  $33 = HEAP8[$$46>>0]|0; //@line 313 "load_abc.cpp"
  $34 = ($33<<24>>24)==(0); //@line 313 "load_abc.cpp"
  if ($34) {
   $value_size$1 = $value_size$07;
   break;
  }
  $35 = ((($$46)) + 1|0); //@line 314 "load_abc.cpp"
  $36 = (($value_size$07) + 1)|0; //@line 314 "load_abc.cpp"
  $37 = (($value) + ($value_size$07)|0); //@line 314 "load_abc.cpp"
  HEAP8[$37>>0] = $33; //@line 314 "load_abc.cpp"
  $38 = ($36>>>0)<(255); //@line 313 "load_abc.cpp"
  if ($38) {
   $$46 = $35;$value_size$07 = $36;
  } else {
   $value_size$1 = $36;
   break;
  }
 }
 while(1) {
  $39 = ($value_size$1|0)==(0); //@line 315 "load_abc.cpp"
  if ($39) {
   $value_size$1$lcssa = 0;
   label = 17;
   break;
  }
  $40 = (($value_size$1) + -1)|0; //@line 315 "load_abc.cpp"
  $41 = (($value) + ($40)|0); //@line 315 "load_abc.cpp"
  $42 = HEAP8[$41>>0]|0; //@line 315 "load_abc.cpp"
  $43 = $42 << 24 >> 24; //@line 315 "load_abc.cpp"
  $44 = (_isspace($43)|0); //@line 315 "load_abc.cpp"
  $45 = ($44|0)==(0); //@line 315 "load_abc.cpp"
  if ($45) {
   $value_size$1$lcssa = $value_size$1;
   label = 17;
   break;
  } else {
   $value_size$1 = $40;
  }
 }
 if ((label|0) == 17) {
  $46 = (($value) + ($value_size$1$lcssa)|0); //@line 317 "load_abc.cpp"
  HEAP8[$46>>0] = 0; //@line 317 "load_abc.cpp"
  return; //@line 318 "load_abc.cpp"
 }
}
function __ZL16abc_addchordnamePKciPKi($s,$len,$notes) {
 $s = $s|0;
 $len = $len|0;
 $notes = $notes|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $exitcond15 = 0, $i$010 = 0, $i$010$lcssa = 0;
 var $j$07 = 0, $j$18 = 0, $txt$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer10 = 0, $vararg_buffer13 = 0, $vararg_buffer4 = 0, $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer13 = sp + 40|0;
 $vararg_buffer10 = sp + 32|0;
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $txt$i = sp + 48|0;
 $0 = (_strlen($s)|0); //@line 2937 "load_abc.cpp"
 $1 = ($0>>>0)>(7); //@line 2937 "load_abc.cpp"
 if ($1) {
  $2 = (($0) + 50)|0; //@line 323 "load_abc.cpp"
  $3 = ($2>>>0)>(255); //@line 323 "load_abc.cpp"
  if (!($3)) {
   HEAP32[$vararg_buffer>>2] = $s; //@line 324 "load_abc.cpp"
   (_sprintf($txt$i,524696,$vararg_buffer)|0); //@line 324 "load_abc.cpp"
   $4 = HEAP32[37544]|0; //@line 328 "load_abc.cpp"
   HEAP32[$vararg_buffer1>>2] = $txt$i; //@line 328 "load_abc.cpp"
   (_fprintf($4,524747,$vararg_buffer1)|0); //@line 328 "load_abc.cpp"
  }
  STACKTOP = sp;return; //@line 2961 "load_abc.cpp"
 }
 $5 = ($len|0)>(6); //@line 2941 "load_abc.cpp"
 if ($5) {
  $8 = (($0) + 54)|0; //@line 323 "load_abc.cpp"
  $9 = ($8>>>0)>(255); //@line 323 "load_abc.cpp"
  if (!($9)) {
   HEAP32[$vararg_buffer4>>2] = $s; //@line 324 "load_abc.cpp"
   (_sprintf($txt$i,524762,$vararg_buffer4)|0); //@line 324 "load_abc.cpp"
   $10 = HEAP32[37544]|0; //@line 328 "load_abc.cpp"
   HEAP32[$vararg_buffer7>>2] = $txt$i; //@line 328 "load_abc.cpp"
   (_fprintf($10,524747,$vararg_buffer7)|0); //@line 328 "load_abc.cpp"
  }
  STACKTOP = sp;return; //@line 2961 "load_abc.cpp"
 }
 $6 = HEAP32[299864]|0; //@line 2945 "load_abc.cpp"
 $7 = ($6|0)>(0); //@line 2945 "load_abc.cpp"
 if ($7) {
  $i$010 = 0;
  while(1) {
   $11 = (1277594 + ($i$010<<3)|0); //@line 2946 "load_abc.cpp"
   $12 = (_strcmp($s,$11)|0); //@line 2946 "load_abc.cpp"
   $13 = ($12|0)==(0); //@line 2946 "load_abc.cpp"
   if ($13) {
    $i$010$lcssa = $i$010;
    label = 11;
    break;
   }
   $20 = (($i$010) + 1)|0; //@line 2945 "load_abc.cpp"
   $21 = ($20|0)<($6|0); //@line 2945 "load_abc.cpp"
   if ($21) {
    $i$010 = $20;
   } else {
    break;
   }
  }
  if ((label|0) == 11) {
   $14 = (1199460 + ($i$010$lcssa<<2)|0); //@line 2948 "load_abc.cpp"
   HEAP32[$14>>2] = $len; //@line 2948 "load_abc.cpp"
   $15 = ($len|0)>(0); //@line 2949 "load_abc.cpp"
   if ($15) {
    $j$07 = 0;
   } else {
    STACKTOP = sp;return; //@line 2961 "load_abc.cpp"
   }
   while(1) {
    $16 = (($notes) + ($j$07<<2)|0); //@line 2949 "load_abc.cpp"
    $17 = HEAP32[$16>>2]|0; //@line 2949 "load_abc.cpp"
    $18 = ((1199780 + (($i$010$lcssa*24)|0)|0) + ($j$07<<2)|0); //@line 2949 "load_abc.cpp"
    HEAP32[$18>>2] = $17; //@line 2949 "load_abc.cpp"
    $19 = (($j$07) + 1)|0; //@line 2949 "load_abc.cpp"
    $exitcond = ($19|0)==($len|0); //@line 2949 "load_abc.cpp"
    if ($exitcond) {
     break;
    } else {
     $j$07 = $19;
    }
   }
   STACKTOP = sp;return; //@line 2961 "load_abc.cpp"
  }
  $22 = ($6|0)>(79); //@line 2953 "load_abc.cpp"
  if ($22) {
   $23 = (_strlen($s)|0); //@line 323 "load_abc.cpp"
   $24 = (($23) + 45)|0; //@line 323 "load_abc.cpp"
   $25 = ($24>>>0)>(255); //@line 323 "load_abc.cpp"
   if (!($25)) {
    HEAP32[$vararg_buffer10>>2] = $s; //@line 324 "load_abc.cpp"
    (_sprintf($txt$i,524817,$vararg_buffer10)|0); //@line 324 "load_abc.cpp"
    $26 = HEAP32[37544]|0; //@line 328 "load_abc.cpp"
    HEAP32[$vararg_buffer13>>2] = $txt$i; //@line 328 "load_abc.cpp"
    (_fprintf($26,524747,$vararg_buffer13)|0); //@line 328 "load_abc.cpp"
   }
   STACKTOP = sp;return; //@line 2961 "load_abc.cpp"
  }
 }
 $27 = (1277594 + ($6<<3)|0); //@line 2956 "load_abc.cpp"
 (_strcpy($27,$s)|0); //@line 2956 "load_abc.cpp"
 $28 = HEAP32[299864]|0; //@line 2957 "load_abc.cpp"
 $29 = (1199460 + ($28<<2)|0); //@line 2957 "load_abc.cpp"
 HEAP32[$29>>2] = $len; //@line 2957 "load_abc.cpp"
 $30 = ($len|0)>(0); //@line 2958 "load_abc.cpp"
 if ($30) {
  $j$18 = 0;
  while(1) {
   $31 = (($notes) + ($j$18<<2)|0); //@line 2958 "load_abc.cpp"
   $32 = HEAP32[$31>>2]|0; //@line 2958 "load_abc.cpp"
   $33 = ((1199780 + (($28*24)|0)|0) + ($j$18<<2)|0); //@line 2958 "load_abc.cpp"
   HEAP32[$33>>2] = $32; //@line 2958 "load_abc.cpp"
   $34 = (($j$18) + 1)|0; //@line 2958 "load_abc.cpp"
   $exitcond15 = ($34|0)==($len|0); //@line 2958 "load_abc.cpp"
   if ($exitcond15) {
    break;
   } else {
    $j$18 = $34;
   }
  }
 }
 $35 = (($28) + 1)|0; //@line 2959 "load_abc.cpp"
 HEAP32[299864] = $35; //@line 2959 "load_abc.cpp"
 STACKTOP = sp;return; //@line 2961 "load_abc.cpp"
}
function __ZL8abc_getsP10_ABCHANDLEP6MMFILE($h,$mmfile) {
 $h = $h|0;
 $mmfile = $mmfile|0;
 var $$0 = 0, $$i$0$i = 0, $$i$0$i22 = 0, $$lcssa = 0, $$lcssa142 = 0, $$lcssa146 = 0, $$lcssa148 = 0, $$lcssa150 = 0, $$lcssa153 = 0, $$lcssa158 = 0, $$lcssa58 = 0, $$lcssa59 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$i = 0, $$pre$phiZ2D = 0, $$pre7$i = 0, $$pre85 = 0, $$pre86 = 0, $$sum = 0;
 var $$sum49 = 0, $$sum50 = 0, $$sum51 = 0, $$sum52 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $brmerge = 0, $exitcond$i = 0;
 var $exitcond4$i = 0, $i$02$i = 0, $i$02$i20 = 0, $i$03$i = 0, $i$05$i = 0, $i$05$i$lcssa = 0, $i$05$i$lcssa157 = 0, $i$05$i16 = 0, $i$05$i16$lcssa = 0, $i$05$i16$lcssa144 = 0, $i$1$i23 = 0, $j$01$i = 0, $mmfile$idx$val$i = 0, $mp$0 = 0, $mp$066 = 0, $mp$067 = 0, $not$$i$i = 0, $not$$i$i1 = 0, $not$$i$i13 = 0, $not$$i$i26 = 0;
 var $not$$i$i4 = 0, $not$$i1$i = 0, $not$$i1$i29 = 0, $not$$i1$i29$lcssa149 = 0, $not$$i1$i29$lcssa95$not = 0, $not$$i1$i36 = 0, $not$$i1$i7 = 0, $or$cond = 0, $or$cond$not$i = 0, $or$cond$not$i28 = 0, $or$cond$not$i35 = 0, $or$cond$not$i6 = 0, $or$cond4$i = 0, $or$cond4$i30 = 0, $or$cond4$i37 = 0, $or$cond4$i8 = 0, $p$0$lcssa$i = 0, $p$02$i = 0, $p$1$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($h)) + 16|0); //@line 2280 "load_abc.cpp"
 $1 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 2280 "load_abc.cpp"
 $2 = ($1|0)==(0); //@line 2280 "load_abc.cpp"
 if ($2) {
  HEAP8[$0>>0]=64&255;HEAP8[$0+1>>0]=(64>>8)&255;HEAP8[$0+2>>0]=(64>>16)&255;HEAP8[$0+3>>0]=64>>24; //@line 2281 "load_abc.cpp"
  $3 = (_calloc(64,1)|0); //@line 2282 "load_abc.cpp"
  $4 = ((($h)) + 24|0); //@line 2282 "load_abc.cpp"
  HEAP8[$4>>0]=$3&255;HEAP8[$4+1>>0]=($3>>8)&255;HEAP8[$4+2>>0]=($3>>16)&255;HEAP8[$4+3>>0]=$3>>24; //@line 2282 "load_abc.cpp"
  $$pre$phiZ2D = $4;$11 = 64;$25 = $3;
 } else {
  $$phi$trans$insert = ((($h)) + 24|0);
  $$pre = HEAPU8[$$phi$trans$insert>>0]|(HEAPU8[$$phi$trans$insert+1>>0]<<8)|(HEAPU8[$$phi$trans$insert+2>>0]<<16)|(HEAPU8[$$phi$trans$insert+3>>0]<<24); //@line 2284 "load_abc.cpp"
  $$pre$phiZ2D = $$phi$trans$insert;$11 = $1;$25 = $$pre;
 }
 $5 = ((($mmfile)) + 8|0); //@line 481 "load_abc.cpp"
 $6 = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8)|(HEAPU8[$5+2>>0]<<16)|(HEAPU8[$5+3>>0]<<24); //@line 481 "load_abc.cpp"
 $7 = ($6|0)<(0); //@line 481 "load_abc.cpp"
 if ($7) {
  $$0 = 0;
  return ($$0|0); //@line 2298 "load_abc.cpp"
 }
 $8 = ((($mmfile)) + 4|0); //@line 482 "load_abc.cpp"
 $9 = HEAPU8[$8>>0]|(HEAPU8[$8+1>>0]<<8)|(HEAPU8[$8+2>>0]<<16)|(HEAPU8[$8+3>>0]<<24); //@line 482 "load_abc.cpp"
 $not$$i$i = ($6|0)<($9|0); //@line 482 "load_abc.cpp"
 if (!($not$$i$i)) {
  $$0 = 0;
  return ($$0|0); //@line 2298 "load_abc.cpp"
 }
 $10 = (($11) + -2)|0; //@line 2201 "load_abc.cpp"
 $12 = ($10|0)==(0); //@line 2201 "load_abc.cpp"
 L11: do {
  if ($12) {
   $i$02$i = 0;
  } else {
   $14 = $6;$i$05$i = 0;
   L13: while(1) {
    $13 = ($14|0)<(0); //@line 481 "load_abc.cpp"
    if ($13) {
     label = 13;
    } else {
     $15 = HEAPU8[$8>>0]|(HEAPU8[$8+1>>0]<<8)|(HEAPU8[$8+2>>0]<<16)|(HEAPU8[$8+3>>0]<<24); //@line 482 "load_abc.cpp"
     $not$$i$i1 = ($14|0)<($15|0); //@line 482 "load_abc.cpp"
     if ($not$$i$i1) {
      $16 = HEAPU8[$mmfile>>0]|(HEAPU8[$mmfile+1>>0]<<8)|(HEAPU8[$mmfile+2>>0]<<16)|(HEAPU8[$mmfile+3>>0]<<24); //@line 490 "load_abc.cpp"
      $17 = (($16) + ($14)|0); //@line 490 "load_abc.cpp"
      $18 = HEAP8[$17>>0]|0; //@line 490 "load_abc.cpp"
      $19 = (($14) + 1)|0; //@line 491 "load_abc.cpp"
      HEAP8[$5>>0]=$19&255;HEAP8[$5+1>>0]=($19>>8)&255;HEAP8[$5+2>>0]=($19>>16)&255;HEAP8[$5+3>>0]=$19>>24; //@line 491 "load_abc.cpp"
      $or$cond$not$i = ($18<<24>>24)==(13); //@line 492 "load_abc.cpp"
      $not$$i1$i = ($19|0)<($15|0); //@line 482 "load_abc.cpp"
      $or$cond4$i = $not$$i1$i & $or$cond$not$i; //@line 492 "load_abc.cpp"
      if ($or$cond4$i) {
       $20 = (($16) + ($19)|0); //@line 492 "load_abc.cpp"
       $21 = HEAP8[$20>>0]|0; //@line 492 "load_abc.cpp"
       $22 = ($21<<24>>24)==(10); //@line 492 "load_abc.cpp"
       if ($22) {
        $$lcssa153 = $14;$i$05$i$lcssa = $i$05$i;
        label = 12;
        break;
       }
      }
      $27 = (($25) + ($i$05$i)|0); //@line 2202 "load_abc.cpp"
      HEAP8[$27>>0] = $18; //@line 2202 "load_abc.cpp"
      $28 = $18 << 24 >> 24; //@line 2203 "load_abc.cpp"
      switch ($28|0) {
      case 10:  {
       $i$02$i = $i$05$i;
       break L11;
       break;
      }
      case 13:  {
       $$lcssa158 = $27;$i$05$i$lcssa157 = $i$05$i;
       break L13;
       break;
      }
      default: {
      }
      }
     } else {
      label = 13;
     }
    }
    if ((label|0) == 13) {
     label = 0;
     $26 = (($25) + ($i$05$i)|0); //@line 2202 "load_abc.cpp"
     HEAP8[$26>>0] = -1; //@line 2202 "load_abc.cpp"
    }
    $40 = (($i$05$i) + 1)|0; //@line 2201 "load_abc.cpp"
    $41 = ($40>>>0)<($10>>>0); //@line 2201 "load_abc.cpp"
    if (!($41)) {
     $i$02$i = $40;
     break L11;
    }
    $$pre85 = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8)|(HEAPU8[$5+2>>0]<<16)|(HEAPU8[$5+3>>0]<<24); //@line 481 "load_abc.cpp"
    $14 = $$pre85;$i$05$i = $40;
   }
   if ((label|0) == 12) {
    $23 = (($$lcssa153) + 2)|0; //@line 494 "load_abc.cpp"
    HEAP8[$5>>0]=$23&255;HEAP8[$5+1>>0]=($23>>8)&255;HEAP8[$5+2>>0]=($23>>16)&255;HEAP8[$5+3>>0]=$23>>24; //@line 494 "load_abc.cpp"
    $24 = (($25) + ($i$05$i$lcssa)|0); //@line 2202 "load_abc.cpp"
    HEAP8[$24>>0] = 10; //@line 2202 "load_abc.cpp"
    $i$02$i = $i$05$i$lcssa;
    break;
   }
   $mmfile$idx$val$i = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8)|(HEAPU8[$5+2>>0]<<16)|(HEAPU8[$5+3>>0]<<24); //@line 2284 "load_abc.cpp"
   $29 = ($mmfile$idx$val$i|0)<(0); //@line 481 "load_abc.cpp"
   do {
    if ($29) {
     label = 21;
    } else {
     $30 = HEAPU8[$8>>0]|(HEAPU8[$8+1>>0]<<8)|(HEAPU8[$8+2>>0]<<16)|(HEAPU8[$8+3>>0]<<24); //@line 482 "load_abc.cpp"
     $not$$i$i4 = ($mmfile$idx$val$i|0)<($30|0); //@line 482 "load_abc.cpp"
     if ($not$$i$i4) {
      $31 = HEAPU8[$mmfile>>0]|(HEAPU8[$mmfile+1>>0]<<8)|(HEAPU8[$mmfile+2>>0]<<16)|(HEAPU8[$mmfile+3>>0]<<24); //@line 490 "load_abc.cpp"
      $32 = (($31) + ($mmfile$idx$val$i)|0); //@line 490 "load_abc.cpp"
      $33 = HEAP8[$32>>0]|0; //@line 490 "load_abc.cpp"
      $34 = (($mmfile$idx$val$i) + 1)|0; //@line 491 "load_abc.cpp"
      HEAP8[$5>>0]=$34&255;HEAP8[$5+1>>0]=($34>>8)&255;HEAP8[$5+2>>0]=($34>>16)&255;HEAP8[$5+3>>0]=$34>>24; //@line 491 "load_abc.cpp"
      $or$cond$not$i6 = ($33<<24>>24)==(13); //@line 492 "load_abc.cpp"
      $not$$i1$i7 = ($34|0)<($30|0); //@line 482 "load_abc.cpp"
      $or$cond4$i8 = $not$$i1$i7 & $or$cond$not$i6; //@line 492 "load_abc.cpp"
      if ($or$cond4$i8) {
       $35 = (($31) + ($34)|0); //@line 492 "load_abc.cpp"
       $36 = HEAP8[$35>>0]|0; //@line 492 "load_abc.cpp"
       $37 = ($36<<24>>24)==(10); //@line 492 "load_abc.cpp"
       if ($37) {
        $38 = (($mmfile$idx$val$i) + 2)|0; //@line 494 "load_abc.cpp"
        HEAP8[$5>>0]=$38&255;HEAP8[$5+1>>0]=($38>>8)&255;HEAP8[$5+2>>0]=($38>>16)&255;HEAP8[$5+3>>0]=$38>>24; //@line 494 "load_abc.cpp"
        break;
       }
      }
      $39 = ($33<<24>>24)==(10); //@line 2206 "load_abc.cpp"
      if (!($39)) {
       label = 21;
      }
     } else {
      label = 21;
     }
    }
   } while(0);
   if ((label|0) == 21) {
    HEAP8[$5>>0]=$mmfile$idx$val$i&255;HEAP8[$5+1>>0]=($mmfile$idx$val$i>>8)&255;HEAP8[$5+2>>0]=($mmfile$idx$val$i>>16)&255;HEAP8[$5+3>>0]=$mmfile$idx$val$i>>24; //@line 520 "load_abc.cpp"
   }
   HEAP8[$$lcssa158>>0] = 10; //@line 2207 "load_abc.cpp"
   $i$02$i = $i$05$i$lcssa157;
  }
 } while(0);
 $42 = ($i$02$i|0)==($10|0); //@line 2211 "load_abc.cpp"
 if ($42) {
  $48 = (($25) + ($10)|0); //@line 2212 "load_abc.cpp"
  HEAP8[$48>>0] = 0; //@line 2212 "load_abc.cpp"
  $49 = ($25|0)==(0|0); //@line 2284 "load_abc.cpp"
  if ($49) {
   $$0 = 0;
   return ($$0|0); //@line 2298 "load_abc.cpp"
  }
 } else {
  $43 = (($25) + ($i$02$i)|0); //@line 2211 "load_abc.cpp"
  $44 = HEAP8[$43>>0]|0; //@line 2211 "load_abc.cpp"
  $45 = ($44<<24>>24)==(10); //@line 2211 "load_abc.cpp"
  $46 = $45&1; //@line 2211 "load_abc.cpp"
  $$i$0$i = (($46) + ($i$02$i))|0; //@line 2211 "load_abc.cpp"
  $47 = (($25) + ($$i$0$i)|0); //@line 2212 "load_abc.cpp"
  HEAP8[$47>>0] = 0; //@line 2212 "load_abc.cpp"
 }
 $50 = HEAPU8[$$pre$phiZ2D>>0]|(HEAPU8[$$pre$phiZ2D+1>>0]<<8)|(HEAPU8[$$pre$phiZ2D+2>>0]<<16)|(HEAPU8[$$pre$phiZ2D+3>>0]<<24); //@line 2285 "load_abc.cpp"
 $51 = (_strlen($50)|0); //@line 2285 "load_abc.cpp"
 $52 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 2285 "load_abc.cpp"
 $53 = (($52) + -3)|0; //@line 2285 "load_abc.cpp"
 $54 = ($51|0)>($53|0); //@line 2285 "load_abc.cpp"
 if ($54) {
  $56 = $52;$57 = $50;$60 = $51;
  while(1) {
   $55 = $56 << 1; //@line 2287 "load_abc.cpp"
   $58 = (_realloc($57,$55)|0); //@line 2287 "load_abc.cpp"
   HEAP8[$$pre$phiZ2D>>0]=$58&255;HEAP8[$$pre$phiZ2D+1>>0]=($58>>8)&255;HEAP8[$$pre$phiZ2D+2>>0]=($58>>16)&255;HEAP8[$$pre$phiZ2D+3>>0]=$58>>24; //@line 2287 "load_abc.cpp"
   $59 = (($60) + -1)|0; //@line 2288 "load_abc.cpp"
   $61 = (($58) + ($59)|0); //@line 2288 "load_abc.cpp"
   $62 = HEAP8[$61>>0]|0; //@line 2288 "load_abc.cpp"
   $63 = ($62<<24>>24)==(10); //@line 2288 "load_abc.cpp"
   $$pre86 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 2290 "load_abc.cpp"
   if (!($63)) {
    $64 = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8)|(HEAPU8[$5+2>>0]<<16)|(HEAPU8[$5+3>>0]<<24); //@line 481 "load_abc.cpp"
    $65 = ($64|0)<(0); //@line 481 "load_abc.cpp"
    if (!($65)) {
     $66 = HEAPU8[$8>>0]|(HEAPU8[$8+1>>0]<<8)|(HEAPU8[$8+2>>0]<<16)|(HEAPU8[$8+3>>0]<<24); //@line 482 "load_abc.cpp"
     $not$$i$i13 = ($64|0)<($66|0); //@line 482 "load_abc.cpp"
     if ($not$$i$i13) {
      $67 = (($$pre86) + -2)|0; //@line 2201 "load_abc.cpp"
      $68 = ($67|0)==(0); //@line 2201 "load_abc.cpp"
      L52: do {
       if ($68) {
        $i$02$i20 = 0;
       } else {
        $70 = $64;$i$05$i16 = 0;
        L53: while(1) {
         $69 = ($70|0)>(-1); //@line 481 "load_abc.cpp"
         $not$$i$i26 = ($70|0)<($66|0); //@line 482 "load_abc.cpp"
         $or$cond = $69 & $not$$i$i26; //@line 481 "load_abc.cpp"
         if ($or$cond) {
          $71 = HEAPU8[$mmfile>>0]|(HEAPU8[$mmfile+1>>0]<<8)|(HEAPU8[$mmfile+2>>0]<<16)|(HEAPU8[$mmfile+3>>0]<<24); //@line 490 "load_abc.cpp"
          $72 = (($71) + ($70)|0); //@line 490 "load_abc.cpp"
          $73 = HEAP8[$72>>0]|0; //@line 490 "load_abc.cpp"
          $74 = (($70) + 1)|0; //@line 491 "load_abc.cpp"
          HEAP8[$5>>0]=$74&255;HEAP8[$5+1>>0]=($74>>8)&255;HEAP8[$5+2>>0]=($74>>16)&255;HEAP8[$5+3>>0]=$74>>24; //@line 491 "load_abc.cpp"
          $or$cond$not$i28 = ($73<<24>>24)==(13); //@line 492 "load_abc.cpp"
          $not$$i1$i29 = ($74|0)<($66|0); //@line 482 "load_abc.cpp"
          $or$cond4$i30 = $not$$i1$i29 & $or$cond$not$i28; //@line 492 "load_abc.cpp"
          if ($or$cond4$i30) {
           $75 = (($71) + ($74)|0); //@line 492 "load_abc.cpp"
           $76 = HEAP8[$75>>0]|0; //@line 492 "load_abc.cpp"
           $77 = ($76<<24>>24)==(10); //@line 492 "load_abc.cpp"
           if ($77) {
            $$lcssa = $70;$i$05$i16$lcssa = $i$05$i16;
            label = 37;
            break;
           }
          }
          $$sum51 = (($i$05$i16) + ($60))|0; //@line 2202 "load_abc.cpp"
          $81 = (($58) + ($$sum51)|0); //@line 2202 "load_abc.cpp"
          HEAP8[$81>>0] = $73; //@line 2202 "load_abc.cpp"
          $82 = $73 << 24 >> 24; //@line 2203 "load_abc.cpp"
          switch ($82|0) {
          case 10:  {
           $i$02$i20 = $i$05$i16;
           break L52;
           break;
          }
          case 13:  {
           $$lcssa142 = $70;$$lcssa146 = $71;$$lcssa148 = $74;$$lcssa150 = $81;$i$05$i16$lcssa144 = $i$05$i16;$not$$i1$i29$lcssa149 = $not$$i1$i29;
           break L53;
           break;
          }
          default: {
           $153 = $74;
          }
          }
         } else {
          $$sum = (($i$05$i16) + ($60))|0; //@line 2202 "load_abc.cpp"
          $80 = (($58) + ($$sum)|0); //@line 2202 "load_abc.cpp"
          HEAP8[$80>>0] = -1; //@line 2202 "load_abc.cpp"
          $153 = $70;
         }
         $92 = (($i$05$i16) + 1)|0; //@line 2201 "load_abc.cpp"
         $93 = ($92>>>0)<($67>>>0); //@line 2201 "load_abc.cpp"
         if ($93) {
          $70 = $153;$i$05$i16 = $92;
         } else {
          $i$02$i20 = $92;
          break L52;
         }
        }
        if ((label|0) == 37) {
         label = 0;
         $78 = (($$lcssa) + 2)|0; //@line 494 "load_abc.cpp"
         HEAP8[$5>>0]=$78&255;HEAP8[$5+1>>0]=($78>>8)&255;HEAP8[$5+2>>0]=($78>>16)&255;HEAP8[$5+3>>0]=$78>>24; //@line 494 "load_abc.cpp"
         $$sum52 = (($i$05$i16$lcssa) + ($60))|0; //@line 2202 "load_abc.cpp"
         $79 = (($58) + ($$sum52)|0); //@line 2202 "load_abc.cpp"
         HEAP8[$79>>0] = 10; //@line 2202 "load_abc.cpp"
         $i$02$i20 = $i$05$i16$lcssa;
         break;
        }
        $83 = ($$lcssa142|0)<(-1); //@line 481 "load_abc.cpp"
        $not$$i1$i29$lcssa95$not = $not$$i1$i29$lcssa149 ^ 1; //@line 481 "load_abc.cpp"
        $brmerge = $83 | $not$$i1$i29$lcssa95$not; //@line 481 "load_abc.cpp"
        do {
         if ($brmerge) {
          label = 45;
         } else {
          $84 = (($$lcssa146) + ($$lcssa148)|0); //@line 490 "load_abc.cpp"
          $85 = HEAP8[$84>>0]|0; //@line 490 "load_abc.cpp"
          $86 = (($$lcssa142) + 2)|0; //@line 491 "load_abc.cpp"
          HEAP8[$5>>0]=$86&255;HEAP8[$5+1>>0]=($86>>8)&255;HEAP8[$5+2>>0]=($86>>16)&255;HEAP8[$5+3>>0]=$86>>24; //@line 491 "load_abc.cpp"
          $or$cond$not$i35 = ($85<<24>>24)==(13); //@line 492 "load_abc.cpp"
          $not$$i1$i36 = ($86|0)<($66|0); //@line 482 "load_abc.cpp"
          $or$cond4$i37 = $not$$i1$i36 & $or$cond$not$i35; //@line 492 "load_abc.cpp"
          if ($or$cond4$i37) {
           $87 = (($$lcssa146) + ($86)|0); //@line 492 "load_abc.cpp"
           $88 = HEAP8[$87>>0]|0; //@line 492 "load_abc.cpp"
           $89 = ($88<<24>>24)==(10); //@line 492 "load_abc.cpp"
           if ($89) {
            $90 = (($$lcssa142) + 3)|0; //@line 494 "load_abc.cpp"
            HEAP8[$5>>0]=$90&255;HEAP8[$5+1>>0]=($90>>8)&255;HEAP8[$5+2>>0]=($90>>16)&255;HEAP8[$5+3>>0]=$90>>24; //@line 494 "load_abc.cpp"
            break;
           }
          }
          $91 = ($85<<24>>24)==(10); //@line 2206 "load_abc.cpp"
          if (!($91)) {
           label = 45;
          }
         }
        } while(0);
        if ((label|0) == 45) {
         label = 0;
         HEAP8[$5>>0]=$$lcssa148&255;HEAP8[$5+1>>0]=($$lcssa148>>8)&255;HEAP8[$5+2>>0]=($$lcssa148>>16)&255;HEAP8[$5+3>>0]=$$lcssa148>>24; //@line 520 "load_abc.cpp"
        }
        HEAP8[$$lcssa150>>0] = 10; //@line 2207 "load_abc.cpp"
        $i$02$i20 = $i$05$i16$lcssa144;
       }
      } while(0);
      $94 = ($i$02$i20|0)==($67|0); //@line 2211 "load_abc.cpp"
      if ($94) {
       $i$1$i23 = $67;
      } else {
       $$sum49 = (($i$02$i20) + ($60))|0; //@line 2211 "load_abc.cpp"
       $95 = (($58) + ($$sum49)|0); //@line 2211 "load_abc.cpp"
       $96 = HEAP8[$95>>0]|0; //@line 2211 "load_abc.cpp"
       $97 = ($96<<24>>24)==(10); //@line 2211 "load_abc.cpp"
       $98 = $97&1; //@line 2211 "load_abc.cpp"
       $$i$0$i22 = (($98) + ($i$02$i20))|0; //@line 2211 "load_abc.cpp"
       $i$1$i23 = $$i$0$i22;
      }
      $$sum50 = (($i$1$i23) + ($60))|0; //@line 2212 "load_abc.cpp"
      $99 = (($58) + ($$sum50)|0); //@line 2212 "load_abc.cpp"
      HEAP8[$99>>0] = 0; //@line 2212 "load_abc.cpp"
     }
    }
   }
   $100 = $$pre86 << 1; //@line 2290 "load_abc.cpp"
   HEAP8[$0>>0]=$100&255;HEAP8[$0+1>>0]=($100>>8)&255;HEAP8[$0+2>>0]=($100>>16)&255;HEAP8[$0+3>>0]=$100>>24; //@line 2290 "load_abc.cpp"
   $101 = (_strlen($58)|0); //@line 2285 "load_abc.cpp"
   $102 = (($100) + -3)|0; //@line 2285 "load_abc.cpp"
   $103 = ($101|0)>($102|0); //@line 2285 "load_abc.cpp"
   if ($103) {
    $56 = $100;$57 = $58;$60 = $101;
   } else {
    $$lcssa58 = $58;$$lcssa59 = $101;
    break;
   }
  }
 } else {
  $$lcssa58 = $50;$$lcssa59 = $51;
 }
 $104 = (($$lcssa59) + -1)|0; //@line 2292 "load_abc.cpp"
 $105 = (($$lcssa58) + ($104)|0); //@line 2292 "load_abc.cpp"
 HEAP8[$105>>0] = 0; //@line 2292 "load_abc.cpp"
 $mp$066 = HEAPU8[$h>>0]|(HEAPU8[$h+1>>0]<<8)|(HEAPU8[$h+2>>0]<<16)|(HEAPU8[$h+3>>0]<<24); //@line 2293 "load_abc.cpp"
 $106 = ($mp$066|0)==(0|0); //@line 2293 "load_abc.cpp"
 if (!($106)) {
  $mp$067 = $mp$066;
  while(1) {
   $107 = ((($mp$067)) + 12|0); //@line 2244 "load_abc.cpp"
   $108 = HEAPU8[$107>>0]|(HEAPU8[$107+1>>0]<<8)|(HEAPU8[$107+2>>0]<<16)|(HEAPU8[$107+3>>0]<<24); //@line 2244 "load_abc.cpp"
   $109 = ($108|0)==(0|0); //@line 2244 "load_abc.cpp"
   $110 = ((($mp$067)) + 4|0); //@line 2273 "load_abc.cpp"
   $111 = HEAPU8[$110>>0]|(HEAPU8[$110+1>>0]<<8)|(HEAPU8[$110+2>>0]<<16)|(HEAPU8[$110+3>>0]<<24); //@line 2273 "load_abc.cpp"
   L83: do {
    if ($109) {
     $149 = ((($mp$067)) + 8|0); //@line 2273 "load_abc.cpp"
     $150 = HEAPU8[$149>>0]|(HEAPU8[$149+1>>0]<<8)|(HEAPU8[$149+2>>0]<<16)|(HEAPU8[$149+3>>0]<<24); //@line 2273 "load_abc.cpp"
     __ZL14abc_substituteP10_ABCHANDLEPcS1_($h,$111,$150); //@line 2273 "load_abc.cpp"
    } else {
     $112 = $108; //@line 2245 "load_abc.cpp"
     $113 = $111; //@line 2245 "load_abc.cpp"
     $114 = (($112) - ($113))|0; //@line 2245 "load_abc.cpp"
     $115 = ((($mp$067)) + 8|0); //@line 2250 "load_abc.cpp"
     $116 = $111;$i$03$i = 0;
     while(1) {
      $117 = (_strlen($116)|0); //@line 2247 "load_abc.cpp"
      $118 = (($117) + 1)|0; //@line 2247 "load_abc.cpp"
      $119 = (_malloc($118)|0); //@line 2247 "load_abc.cpp"
      (_strcpy($119,$116)|0); //@line 2248 "load_abc.cpp"
      $120 = (524964 + ($i$03$i)|0); //@line 2249 "load_abc.cpp"
      $121 = HEAP8[$120>>0]|0; //@line 2249 "load_abc.cpp"
      $122 = (($119) + ($114)|0); //@line 2249 "load_abc.cpp"
      HEAP8[$122>>0] = $121; //@line 2249 "load_abc.cpp"
      $123 = HEAPU8[$115>>0]|(HEAPU8[$115+1>>0]<<8)|(HEAPU8[$115+2>>0]<<16)|(HEAPU8[$115+3>>0]<<24); //@line 2250 "load_abc.cpp"
      $124 = (_strlen($123)|0); //@line 2250 "load_abc.cpp"
      $125 = $124 << 1; //@line 2251 "load_abc.cpp"
      $126 = $125 | 1; //@line 2251 "load_abc.cpp"
      $127 = (_malloc($126)|0); //@line 2251 "load_abc.cpp"
      $128 = ($124|0)>(0); //@line 2253 "load_abc.cpp"
      L88: do {
       if ($128) {
        $129 = (($i$03$i) + -110)|0; //@line 2256 "load_abc.cpp"
        $131 = $123;$j$01$i = 0;$p$02$i = $127;
        while(1) {
         $130 = (($131) + ($j$01$i)|0); //@line 2254 "load_abc.cpp"
         $132 = HEAP8[$130>>0]|0; //@line 2254 "load_abc.cpp"
         $133 = $132 << 24 >> 24; //@line 2254 "load_abc.cpp"
         $134 = ($132<<24>>24)>(103); //@line 2255 "load_abc.cpp"
         do {
          if ($134) {
           $135 = (_islower($133)|0); //@line 2255 "load_abc.cpp"
           $136 = ($135|0)==(0); //@line 2255 "load_abc.cpp"
           if ($136) {
            label = 63;
           } else {
            $137 = (($129) + ($133))|0; //@line 2257 "load_abc.cpp"
            $138 = (($137) + 7)|0; //@line 2257 "load_abc.cpp"
            $139 = (524979 + ($138)|0); //@line 2257 "load_abc.cpp"
            $140 = HEAP8[$139>>0]|0; //@line 2257 "load_abc.cpp"
            $141 = ((($p$02$i)) + 1|0); //@line 2258 "load_abc.cpp"
            HEAP8[$p$02$i>>0] = $140; //@line 2258 "load_abc.cpp"
            $142 = ($137|0)<(0); //@line 2259 "load_abc.cpp"
            if ($142) {
             $143 = ((($p$02$i)) + 2|0); //@line 2260 "load_abc.cpp"
             HEAP8[$141>>0] = 44; //@line 2260 "load_abc.cpp"
             $p$1$i = $143;
             break;
            }
            $144 = ($137|0)>(13); //@line 2261 "load_abc.cpp"
            if ($144) {
             $145 = ((($p$02$i)) + 2|0); //@line 2262 "load_abc.cpp"
             HEAP8[$141>>0] = 39; //@line 2262 "load_abc.cpp"
             $p$1$i = $145;
            } else {
             $p$1$i = $141;
            }
           }
          } else {
           label = 63;
          }
         } while(0);
         if ((label|0) == 63) {
          label = 0;
          $146 = ((($p$02$i)) + 1|0); //@line 2264 "load_abc.cpp"
          HEAP8[$p$02$i>>0] = $132; //@line 2264 "load_abc.cpp"
          $p$1$i = $146;
         }
         $147 = (($j$01$i) + 1)|0; //@line 2253 "load_abc.cpp"
         $exitcond$i = ($147|0)==($124|0); //@line 2253 "load_abc.cpp"
         if ($exitcond$i) {
          $p$0$lcssa$i = $p$1$i;
          break L88;
         }
         $$pre7$i = HEAPU8[$115>>0]|(HEAPU8[$115+1>>0]<<8)|(HEAPU8[$115+2>>0]<<16)|(HEAPU8[$115+3>>0]<<24); //@line 2254 "load_abc.cpp"
         $131 = $$pre7$i;$j$01$i = $147;$p$02$i = $p$1$i;
        }
       } else {
        $p$0$lcssa$i = $127;
       }
      } while(0);
      HEAP8[$p$0$lcssa$i>>0] = 0; //@line 2266 "load_abc.cpp"
      __ZL14abc_substituteP10_ABCHANDLEPcS1_($h,$119,$127); //@line 2267 "load_abc.cpp"
      _free($127); //@line 2268 "load_abc.cpp"
      _free($119); //@line 2269 "load_abc.cpp"
      $148 = (($i$03$i) + 1)|0; //@line 2246 "load_abc.cpp"
      $exitcond4$i = ($148|0)==(14); //@line 2246 "load_abc.cpp"
      if ($exitcond4$i) {
       break L83;
      }
      $$pre$i = HEAPU8[$110>>0]|(HEAPU8[$110+1>>0]<<8)|(HEAPU8[$110+2>>0]<<16)|(HEAPU8[$110+3>>0]<<24); //@line 2247 "load_abc.cpp"
      $116 = $$pre$i;$i$03$i = $148;
     }
    }
   } while(0);
   $mp$0 = HEAPU8[$mp$067>>0]|(HEAPU8[$mp$067+1>>0]<<8)|(HEAPU8[$mp$067+2>>0]<<16)|(HEAPU8[$mp$067+3>>0]<<24); //@line 2293 "load_abc.cpp"
   $151 = ($mp$0|0)==(0|0); //@line 2293 "load_abc.cpp"
   if ($151) {
    break;
   } else {
    $mp$067 = $mp$0;
   }
  }
 }
 $152 = HEAPU8[$$pre$phiZ2D>>0]|(HEAPU8[$$pre$phiZ2D+1>>0]<<8)|(HEAPU8[$$pre$phiZ2D+2>>0]<<16)|(HEAPU8[$$pre$phiZ2D+3>>0]<<24); //@line 2295 "load_abc.cpp"
 $$0 = $152;
 return ($$0|0); //@line 2298 "load_abc.cpp"
}
function __ZL14abc_substituteP10_ABCHANDLEPcS1_($h,$target,$s) {
 $h = $h|0;
 $target = $target|0;
 $s = $s|0;
 var $$pre = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $p$0 = 0, $p$0$us = 0, $p$16$us = 0;
 var $q$04$us = 0, $q$15$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strlen($target)|0); //@line 2220 "load_abc.cpp"
 $1 = (_strlen($s)|0); //@line 2221 "load_abc.cpp"
 $2 = ($0|0)<(1); //@line 2222 "load_abc.cpp"
 $3 = ($1|0)<(1); //@line 2222 "load_abc.cpp"
 $or$cond = $2 | $3; //@line 2222 "load_abc.cpp"
 if ($or$cond) {
  return; //@line 2239 "load_abc.cpp"
 }
 $4 = (_strstr($s,$target)|0); //@line 2222 "load_abc.cpp"
 $5 = ($4|0)==(0|0); //@line 2222 "load_abc.cpp"
 if (!($5)) {
  return; //@line 2239 "load_abc.cpp"
 }
 $6 = ((($h)) + 24|0); //@line 2224 "load_abc.cpp"
 $7 = HEAPU8[$6>>0]|(HEAPU8[$6+1>>0]<<8)|(HEAPU8[$6+2>>0]<<16)|(HEAPU8[$6+3>>0]<<24); //@line 2224 "load_abc.cpp"
 $8 = (_strstr($7,$target)|0); //@line 2224 "load_abc.cpp"
 $9 = ($8|0)==(0|0); //@line 2224 "load_abc.cpp"
 if ($9) {
  return; //@line 2239 "load_abc.cpp"
 }
 $10 = (($1) - ($0))|0; //@line 2225 "load_abc.cpp"
 $11 = ((($h)) + 16|0); //@line 2225 "load_abc.cpp"
 $12 = ($1|0)>($0|0); //@line 2230 "load_abc.cpp"
 if ($12) {
  $13 = $7;$52 = $8;
 } else {
  $39 = $7;$53 = $8;
  while(1) {
   $40 = (_strlen($39)|0); //@line 2225 "load_abc.cpp"
   $41 = (($10) + ($40))|0; //@line 2225 "load_abc.cpp"
   $42 = HEAPU8[$11>>0]|(HEAPU8[$11+1>>0]<<8)|(HEAPU8[$11+2>>0]<<16)|(HEAPU8[$11+3>>0]<<24); //@line 2225 "load_abc.cpp"
   $43 = ($41|0)<($42|0); //@line 2225 "load_abc.cpp"
   if ($43) {
    $p$0 = $53;
   } else {
    $44 = $42 << 1; //@line 2226 "load_abc.cpp"
    $45 = (_realloc($39,$44)|0); //@line 2226 "load_abc.cpp"
    HEAP8[$6>>0]=$45&255;HEAP8[$6+1>>0]=($45>>8)&255;HEAP8[$6+2>>0]=($45>>16)&255;HEAP8[$6+3>>0]=$45>>24; //@line 2226 "load_abc.cpp"
    $46 = HEAPU8[$11>>0]|(HEAPU8[$11+1>>0]<<8)|(HEAPU8[$11+2>>0]<<16)|(HEAPU8[$11+3>>0]<<24); //@line 2227 "load_abc.cpp"
    $47 = $46 << 1; //@line 2227 "load_abc.cpp"
    HEAP8[$11>>0]=$47&255;HEAP8[$11+1>>0]=($47>>8)&255;HEAP8[$11+2>>0]=($47>>16)&255;HEAP8[$11+3>>0]=$47>>24; //@line 2227 "load_abc.cpp"
    $48 = (_strstr($45,$target)|0); //@line 2228 "load_abc.cpp"
    $p$0 = $48;
   }
   (_strcpy($p$0,$s)|0); //@line 2235 "load_abc.cpp"
   $49 = (($p$0) + ($0)|0); //@line 2236 "load_abc.cpp"
   (_strcat($p$0,$49)|0); //@line 2236 "load_abc.cpp"
   $$pre = HEAPU8[$6>>0]|(HEAPU8[$6+1>>0]<<8)|(HEAPU8[$6+2>>0]<<16)|(HEAPU8[$6+3>>0]<<24); //@line 2224 "load_abc.cpp"
   $50 = (_strstr($$pre,$target)|0); //@line 2224 "load_abc.cpp"
   $51 = ($50|0)==(0|0); //@line 2224 "load_abc.cpp"
   if ($51) {
    break;
   } else {
    $39 = $$pre;$53 = $50;
   }
  }
  return; //@line 2239 "load_abc.cpp"
 }
 while(1) {
  $14 = (_strlen($13)|0); //@line 2225 "load_abc.cpp"
  $15 = (($10) + ($14))|0; //@line 2225 "load_abc.cpp"
  $16 = HEAPU8[$11>>0]|(HEAPU8[$11+1>>0]<<8)|(HEAPU8[$11+2>>0]<<16)|(HEAPU8[$11+3>>0]<<24); //@line 2225 "load_abc.cpp"
  $17 = ($15|0)<($16|0); //@line 2225 "load_abc.cpp"
  if ($17) {
   $24 = $13;$p$0$us = $52;
  } else {
   $18 = $16 << 1; //@line 2226 "load_abc.cpp"
   $19 = (_realloc($13,$18)|0); //@line 2226 "load_abc.cpp"
   HEAP8[$6>>0]=$19&255;HEAP8[$6+1>>0]=($19>>8)&255;HEAP8[$6+2>>0]=($19>>16)&255;HEAP8[$6+3>>0]=$19>>24; //@line 2226 "load_abc.cpp"
   $20 = HEAPU8[$11>>0]|(HEAPU8[$11+1>>0]<<8)|(HEAPU8[$11+2>>0]<<16)|(HEAPU8[$11+3>>0]<<24); //@line 2227 "load_abc.cpp"
   $21 = $20 << 1; //@line 2227 "load_abc.cpp"
   HEAP8[$11>>0]=$21&255;HEAP8[$11+1>>0]=($21>>8)&255;HEAP8[$11+2>>0]=($21>>16)&255;HEAP8[$11+3>>0]=$21>>24; //@line 2227 "load_abc.cpp"
   $22 = (_strstr($19,$target)|0); //@line 2228 "load_abc.cpp"
   $24 = $19;$p$0$us = $22;
  }
  $23 = (($24) + ($14)|0); //@line 2231 "load_abc.cpp"
  $25 = ($23>>>0)>($p$0$us>>>0); //@line 2231 "load_abc.cpp"
  if ($25) {
   $q$04$us = $23;
   while(1) {
    $31 = HEAP8[$q$04$us>>0]|0; //@line 2231 "load_abc.cpp"
    $32 = (($q$04$us) + ($10)|0); //@line 2231 "load_abc.cpp"
    HEAP8[$32>>0] = $31; //@line 2231 "load_abc.cpp"
    $33 = ((($q$04$us)) + -1|0); //@line 2231 "load_abc.cpp"
    $34 = ($33>>>0)>($p$0$us>>>0); //@line 2231 "load_abc.cpp"
    if ($34) {
     $q$04$us = $33;
    } else {
     break;
    }
   }
  }
  $35 = HEAP8[$s>>0]|0; //@line 2232 "load_abc.cpp"
  $36 = ($35<<24>>24)==(0); //@line 2232 "load_abc.cpp"
  if (!($36)) {
   $27 = $35;$p$16$us = $p$0$us;$q$15$us = $s;
   while(1) {
    $26 = ((($p$16$us)) + 1|0); //@line 2232 "load_abc.cpp"
    HEAP8[$p$16$us>>0] = $27; //@line 2232 "load_abc.cpp"
    $28 = ((($q$15$us)) + 1|0); //@line 2232 "load_abc.cpp"
    $29 = HEAP8[$28>>0]|0; //@line 2232 "load_abc.cpp"
    $30 = ($29<<24>>24)==(0); //@line 2232 "load_abc.cpp"
    if ($30) {
     break;
    } else {
     $27 = $29;$p$16$us = $26;$q$15$us = $28;
    }
   }
  }
  $$pre9 = HEAPU8[$6>>0]|(HEAPU8[$6+1>>0]<<8)|(HEAPU8[$6+2>>0]<<16)|(HEAPU8[$6+3>>0]<<24); //@line 2224 "load_abc.cpp"
  $37 = (_strstr($$pre9,$target)|0); //@line 2224 "load_abc.cpp"
  $38 = ($37|0)==(0|0); //@line 2224 "load_abc.cpp"
  if ($38) {
   break;
  } else {
   $13 = $$pre9;$52 = $37;
  }
 }
 return; //@line 2239 "load_abc.cpp"
}
function __ZL16abc_locate_trackP10_ABCHANDLEPKci($h,$voice,$pos) {
 $h = $h|0;
 $voice = $voice|0;
 $pos = $pos|0;
 var $$0 = 0, $$04 = 0, $$1$lcssa = 0, $$124 = 0, $$pre = 0, $$sink = 0, $$sink27 = 0, $$sink28 = 0, $$sink28$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $channo$0$lcssa30 = 0, $channo$017 = 0, $channo$1 = 0, $channo$1$lcssa = 0, $i$0$lcssa = 0, $i$025 = 0, $i$025$phi = 0, $instrno$0$lcssa31 = 0, $instrno$016 = 0, $instrno$1 = 0, $instrno$1$lcssa = 0, $prev$0$lcssa34 = 0, $tr$0 = 0, $tr$0$trunused$0 = 0, $tr$012 = 0;
 var $tr$018 = 0, $tr$018$lcssa38 = 0, $trans$0 = 0, $trans$1$lcssa33 = 0, $trans$114 = 0, $trans$2 = 0, $trans$2$lcssa = 0, $trunused$013 = 0, $trunused$1 = 0, $trunused$1$lcssa = 0, $vc = 0, $voiceno$0$lcssa32 = 0, $voiceno$015 = 0, $voiceno$1 = 0, $voiceno$1$lcssa = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vc = sp;
 $$04 = $voice;
 while(1) {
  $0 = HEAP8[$$04>>0]|0; //@line 849 "load_abc.cpp"
  $1 = ($0<<24>>24)==(32); //@line 849 "load_abc.cpp"
  $2 = ((($$04)) + 1|0); //@line 849 "load_abc.cpp"
  if ($1) {
   $$04 = $2;
  } else {
   $$124 = $$04;$11 = 1;$3 = $0;$i$025 = 0;
   break;
  }
 }
 L3: while(1) {
  switch ($3<<24>>24) {
  case 37: case 93: case 0:  {
   $$1$lcssa = $$124;$i$0$lcssa = $i$025;
   break L3;
   break;
  }
  default: {
  }
  }
  $4 = $3 << 24 >> 24; //@line 850 "load_abc.cpp"
  $5 = (_isspace($4)|0); //@line 850 "load_abc.cpp"
  $6 = ($5|0)==(0); //@line 850 "load_abc.cpp"
  if (!($6)) {
   $$1$lcssa = $$124;$i$0$lcssa = $i$025;
   break;
  }
  $7 = HEAP8[$$124>>0]|0; //@line 851 "load_abc.cpp"
  $8 = (($vc) + ($i$025)|0); //@line 851 "load_abc.cpp"
  HEAP8[$8>>0] = $7; //@line 851 "load_abc.cpp"
  $9 = ((($$124)) + 1|0); //@line 850 "load_abc.cpp"
  $10 = (($11) + 1)|0; //@line 850 "load_abc.cpp"
  $12 = ($10>>>0)<(21); //@line 850 "load_abc.cpp"
  if (!($12)) {
   $$1$lcssa = $9;$i$0$lcssa = $11;
   break;
  }
  $$pre = HEAP8[$9>>0]|0; //@line 850 "load_abc.cpp"
  $i$025$phi = $11;$$124 = $9;$11 = $10;$3 = $$pre;$i$025 = $i$025$phi;
 }
 $13 = (($vc) + ($i$0$lcssa)|0); //@line 852 "load_abc.cpp"
 HEAP8[$13>>0] = 0; //@line 852 "load_abc.cpp"
 $14 = ($pos|0)!=(0); //@line 855 "load_abc.cpp"
 if ($14) {
  $trans$0 = 0;
 } else {
  $15 = (__ZL13abc_transposePKc($$1$lcssa)|0); //@line 855 "load_abc.cpp"
  $trans$0 = $15;
 }
 $16 = ((($h)) + 8|0); //@line 856 "load_abc.cpp"
 $tr$012 = HEAPU8[$16>>0]|(HEAPU8[$16+1>>0]<<8)|(HEAPU8[$16+2>>0]<<16)|(HEAPU8[$16+3>>0]<<24); //@line 856 "load_abc.cpp"
 $17 = ($tr$012|0)==(0|0); //@line 856 "load_abc.cpp"
 if ($17) {
  $channo$0$lcssa30 = 0;$instrno$0$lcssa31 = 1;$prev$0$lcssa34 = 0;$trans$1$lcssa33 = $trans$0;$voiceno$0$lcssa32 = 0;
 } else {
  $channo$017 = 0;$instrno$016 = 1;$tr$018 = $tr$012;$trans$114 = $trans$0;$trunused$013 = 0;$voiceno$015 = 0;
  while(1) {
   $18 = ((($tr$018)) + 36|0); //@line 857 "load_abc.cpp"
   $19 = HEAP8[$18>>0]|0; //@line 857 "load_abc.cpp"
   $20 = ($19<<24>>24)==(0); //@line 857 "load_abc.cpp"
   if ($20) {
    $21 = ($trunused$013|0)==(0|0); //@line 858 "load_abc.cpp"
    $tr$0$trunused$0 = $21 ? $tr$018 : $trunused$013; //@line 858 "load_abc.cpp"
    $channo$1 = $channo$017;$instrno$1 = $instrno$016;$trans$2 = $trans$114;$trunused$1 = $tr$0$trunused$0;$voiceno$1 = $voiceno$015;
   } else {
    $22 = ((($tr$018)) + 44|0); //@line 861 "load_abc.cpp"
    $23 = (_strncasecmp($22,$vc,20)|0); //@line 861 "load_abc.cpp"
    $24 = ($23|0)==(0); //@line 861 "load_abc.cpp"
    if ($24) {
     $25 = ((($tr$018)) + 37|0); //@line 862 "load_abc.cpp"
     $26 = HEAP8[$25>>0]|0; //@line 862 "load_abc.cpp"
     $27 = $26&255; //@line 862 "load_abc.cpp"
     $28 = ($27|0)==($pos|0); //@line 862 "load_abc.cpp"
     if ($28) {
      $$0 = $tr$018;
      label = 31;
      break;
     }
     $29 = ((($tr$018)) + 20|0); //@line 864 "load_abc.cpp"
     $30 = HEAPU8[$29>>0]|(HEAPU8[$29+1>>0]<<8)|(HEAPU8[$29+2>>0]<<16)|(HEAPU8[$29+3>>0]<<24); //@line 864 "load_abc.cpp"
     $31 = ((($tr$018)) + 24|0); //@line 865 "load_abc.cpp"
     $32 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 865 "load_abc.cpp"
     HEAP32[300425] = $32; //@line 865 "load_abc.cpp"
     $33 = $19&255; //@line 866 "load_abc.cpp"
     $34 = ((($tr$018)) + 42|0); //@line 867 "load_abc.cpp"
     $35 = HEAP8[$34>>0]|0; //@line 867 "load_abc.cpp"
     $36 = $35&255; //@line 867 "load_abc.cpp"
     $37 = ((($tr$018)) + 40|0); //@line 868 "load_abc.cpp"
     $38 = HEAP8[$37>>0]|0; //@line 868 "load_abc.cpp"
     $39 = $38&255; //@line 868 "load_abc.cpp"
     $channo$1 = $39;$instrno$1 = $36;$trans$2 = $30;$trunused$1 = $trunused$013;$voiceno$1 = $33;
    } else {
     $channo$1 = $channo$017;$instrno$1 = $instrno$016;$trans$2 = $trans$114;$trunused$1 = $trunused$013;$voiceno$1 = $voiceno$015;
    }
   }
   $tr$0 = HEAPU8[$tr$018>>0]|(HEAPU8[$tr$018+1>>0]<<8)|(HEAPU8[$tr$018+2>>0]<<16)|(HEAPU8[$tr$018+3>>0]<<24); //@line 856 "load_abc.cpp"
   $40 = ($tr$0|0)==(0|0); //@line 856 "load_abc.cpp"
   if ($40) {
    $channo$1$lcssa = $channo$1;$instrno$1$lcssa = $instrno$1;$tr$018$lcssa38 = $tr$018;$trans$2$lcssa = $trans$2;$trunused$1$lcssa = $trunused$1;$voiceno$1$lcssa = $voiceno$1;
    break;
   } else {
    $channo$017 = $channo$1;$instrno$016 = $instrno$1;$tr$018 = $tr$0;$trans$114 = $trans$2;$trunused$013 = $trunused$1;$voiceno$015 = $voiceno$1;
   }
  }
  if ((label|0) == 31) {
   STACKTOP = sp;return ($$0|0); //@line 909 "load_abc.cpp"
  }
  $41 = ($trunused$1$lcssa|0)==(0|0); //@line 873 "load_abc.cpp"
  if ($41) {
   $channo$0$lcssa30 = $channo$1$lcssa;$instrno$0$lcssa31 = $instrno$1$lcssa;$prev$0$lcssa34 = $tr$018$lcssa38;$trans$1$lcssa33 = $trans$2$lcssa;$voiceno$0$lcssa32 = $voiceno$1$lcssa;
  } else {
   if ($14) {
    $42 = $instrno$1$lcssa&255; //@line 877 "load_abc.cpp"
    $43 = $channo$1$lcssa&255; //@line 878 "load_abc.cpp"
    $$sink = $43;$$sink27 = $42;$$sink28$in = $voiceno$1$lcssa;
   } else {
    $44 = HEAP32[299860]|0; //@line 881 "load_abc.cpp"
    $45 = (($44) + 1)|0; //@line 881 "load_abc.cpp"
    HEAP32[299860] = $45; //@line 881 "load_abc.cpp"
    $$sink = 0;$$sink27 = 1;$$sink28$in = $45;
   }
   $$sink28 = $$sink28$in&255; //@line 882 "load_abc.cpp"
   $46 = ((($trunused$1$lcssa)) + 36|0); //@line 882 "load_abc.cpp"
   HEAP8[$46>>0] = $$sink28; //@line 882 "load_abc.cpp"
   $47 = ((($trunused$1$lcssa)) + 42|0); //@line 883 "load_abc.cpp"
   HEAP8[$47>>0] = $$sink27; //@line 883 "load_abc.cpp"
   $48 = ((($trunused$1$lcssa)) + 40|0); //@line 884 "load_abc.cpp"
   HEAP8[$48>>0] = $$sink; //@line 884 "load_abc.cpp"
   $49 = $pos&255; //@line 886 "load_abc.cpp"
   $50 = ((($trunused$1$lcssa)) + 37|0); //@line 886 "load_abc.cpp"
   HEAP8[$50>>0] = $49; //@line 886 "load_abc.cpp"
   $51 = ((($trunused$1$lcssa)) + 38|0); //@line 887 "load_abc.cpp"
   HEAP8[$51>>0] = $49; //@line 887 "load_abc.cpp"
   $52 = ((($trunused$1$lcssa)) + 44|0); //@line 888 "load_abc.cpp"
   (_strncpy($52,$vc,20)|0); //@line 888 "load_abc.cpp"
   $53 = ((($trunused$1$lcssa)) + 64|0); //@line 889 "load_abc.cpp"
   HEAP8[$53>>0] = 0; //@line 889 "load_abc.cpp"
   $54 = ((($trunused$1$lcssa)) + 39|0); //@line 890 "load_abc.cpp"
   HEAP8[$54>>0] = 0; //@line 890 "load_abc.cpp"
   $55 = ((($trunused$1$lcssa)) + 20|0); //@line 891 "load_abc.cpp"
   HEAP8[$55>>0]=$trans$2$lcssa&255;HEAP8[$55+1>>0]=($trans$2$lcssa>>8)&255;HEAP8[$55+2>>0]=($trans$2$lcssa>>16)&255;HEAP8[$55+3>>0]=$trans$2$lcssa>>24; //@line 891 "load_abc.cpp"
   $56 = HEAP32[300425]|0; //@line 892 "load_abc.cpp"
   $57 = ((($trunused$1$lcssa)) + 24|0); //@line 892 "load_abc.cpp"
   HEAP8[$57>>0]=$56&255;HEAP8[$57+1>>0]=($56>>8)&255;HEAP8[$57+2>>0]=($56>>16)&255;HEAP8[$57+3>>0]=$56>>24; //@line 892 "load_abc.cpp"
   $58 = HEAPU8[$16>>0]|(HEAPU8[$16+1>>0]<<8)|(HEAPU8[$16+2>>0]<<16)|(HEAPU8[$16+3>>0]<<24); //@line 893 "load_abc.cpp"
   $59 = ((($58)) + 41|0); //@line 893 "load_abc.cpp"
   $60 = HEAP8[$59>>0]|0; //@line 893 "load_abc.cpp"
   $61 = ((($trunused$1$lcssa)) + 41|0); //@line 893 "load_abc.cpp"
   HEAP8[$61>>0] = $60; //@line 893 "load_abc.cpp"
   $62 = ((($trunused$1$lcssa)) + 16|0); //@line 894 "load_abc.cpp"
   HEAP8[$62>>0]=0&255;HEAP8[$62+1>>0]=(0>>8)&255;HEAP8[$62+2>>0]=(0>>16)&255;HEAP8[$62+3>>0]=0>>24; //@line 894 "load_abc.cpp"
   $63 = ((($trunused$1$lcssa)) + 43|0); //@line 895 "load_abc.cpp"
   HEAP8[$63>>0] = 0; //@line 895 "load_abc.cpp"
   $$0 = $trunused$1$lcssa;
   STACKTOP = sp;return ($$0|0); //@line 909 "load_abc.cpp"
  }
 }
 $64 = ($pos|0)==(0); //@line 616 "load_abc.cpp"
 $65 = HEAP32[299860]|0; //@line 616 "load_abc.cpp"
 if ($64) {
  $66 = (($65) + 1)|0; //@line 616 "load_abc.cpp"
  HEAP32[299860] = $66; //@line 616 "load_abc.cpp"
  $69 = $66;
 } else {
  $69 = $65;
 }
 $67 = (_calloc(1,66)|0); //@line 617 "load_abc.cpp"
 HEAP8[$67>>0]=0&255;HEAP8[$67+1>>0]=(0>>8)&255;HEAP8[$67+2>>0]=(0>>16)&255;HEAP8[$67+3>>0]=0>>24; //@line 618 "load_abc.cpp"
 $68 = $69&255; //@line 619 "load_abc.cpp"
 $70 = ((($67)) + 36|0); //@line 619 "load_abc.cpp"
 HEAP8[$70>>0] = $68; //@line 619 "load_abc.cpp"
 $71 = $pos&255; //@line 620 "load_abc.cpp"
 $72 = ((($67)) + 37|0); //@line 620 "load_abc.cpp"
 HEAP8[$72>>0] = $71; //@line 620 "load_abc.cpp"
 $73 = ((($67)) + 38|0); //@line 621 "load_abc.cpp"
 HEAP8[$73>>0] = $71; //@line 621 "load_abc.cpp"
 $74 = ((($67)) + 42|0); //@line 622 "load_abc.cpp"
 HEAP8[$74>>0] = 1; //@line 622 "load_abc.cpp"
 $75 = ((($67)) + 44|0); //@line 623 "load_abc.cpp"
 (_strncpy($75,$vc,20)|0); //@line 623 "load_abc.cpp"
 $76 = ((($67)) + 64|0); //@line 624 "load_abc.cpp"
 HEAP8[$76>>0] = 0; //@line 624 "load_abc.cpp"
 $77 = ((($67)) + 4|0); //@line 625 "load_abc.cpp"
 $78 = ((($67)) + 39|0); //@line 629 "load_abc.cpp"
 HEAP8[$78>>0] = 0; //@line 629 "load_abc.cpp"
 $79 = ((($67)) + 40|0); //@line 630 "load_abc.cpp"
 HEAP8[$79>>0] = 0; //@line 630 "load_abc.cpp"
 dest=$77; stop=dest+20|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 625 "load_abc.cpp"
 $80 = HEAPU8[$16>>0]|(HEAPU8[$16+1>>0]<<8)|(HEAPU8[$16+2>>0]<<16)|(HEAPU8[$16+3>>0]<<24); //@line 632 "load_abc.cpp"
 $81 = ($80|0)==(0|0); //@line 632 "load_abc.cpp"
 if ($81) {
  $85 = 120;
 } else {
  $82 = ((($80)) + 41|0); //@line 632 "load_abc.cpp"
  $83 = HEAP8[$82>>0]|0; //@line 632 "load_abc.cpp"
  $85 = $83;
 }
 $84 = ((($67)) + 41|0); //@line 632 "load_abc.cpp"
 HEAP8[$84>>0] = $85; //@line 632 "load_abc.cpp"
 $86 = ((($67)) + 28|0); //@line 633 "load_abc.cpp"
 HEAP8[$86>>0]=0&255;HEAP8[$86+1>>0]=(0>>8)&255;HEAP8[$86+2>>0]=(0>>16)&255;HEAP8[$86+3>>0]=0>>24; //@line 633 "load_abc.cpp"
 $87 = ((($67)) + 32|0); //@line 634 "load_abc.cpp"
 HEAP8[$87>>0]=0&255;HEAP8[$87+1>>0]=(0>>8)&255;HEAP8[$87+2>>0]=(0>>16)&255;HEAP8[$87+3>>0]=0>>24; //@line 634 "load_abc.cpp"
 $88 = ((($67)) + 43|0); //@line 635 "load_abc.cpp"
 HEAP8[$88>>0] = 0; //@line 635 "load_abc.cpp"
 if ($14) {
  $89 = $voiceno$0$lcssa32&255; //@line 900 "load_abc.cpp"
  HEAP8[$70>>0] = $89; //@line 900 "load_abc.cpp"
  $90 = $instrno$0$lcssa31&255; //@line 901 "load_abc.cpp"
  HEAP8[$74>>0] = $90; //@line 901 "load_abc.cpp"
  $91 = $channo$0$lcssa30&255; //@line 902 "load_abc.cpp"
  HEAP8[$79>>0] = $91; //@line 902 "load_abc.cpp"
 }
 $92 = ((($67)) + 20|0); //@line 904 "load_abc.cpp"
 HEAP8[$92>>0]=$trans$1$lcssa33&255;HEAP8[$92+1>>0]=($trans$1$lcssa33>>8)&255;HEAP8[$92+2>>0]=($trans$1$lcssa33>>16)&255;HEAP8[$92+3>>0]=$trans$1$lcssa33>>24; //@line 904 "load_abc.cpp"
 $93 = HEAP32[300425]|0; //@line 905 "load_abc.cpp"
 $94 = ((($67)) + 24|0); //@line 905 "load_abc.cpp"
 HEAP8[$94>>0]=$93&255;HEAP8[$94+1>>0]=($93>>8)&255;HEAP8[$94+2>>0]=($93>>16)&255;HEAP8[$94+3>>0]=$93>>24; //@line 905 "load_abc.cpp"
 $95 = ($prev$0$lcssa34|0)==(0|0); //@line 906 "load_abc.cpp"
 if ($95) {
  HEAP8[$16>>0]=$67&255;HEAP8[$16+1>>0]=($67>>8)&255;HEAP8[$16+2>>0]=($67>>16)&255;HEAP8[$16+3>>0]=$67>>24; //@line 907 "load_abc.cpp"
  $$0 = $67;
  STACKTOP = sp;return ($$0|0); //@line 909 "load_abc.cpp"
 } else {
  HEAP8[$prev$0$lcssa34>>0]=$67&255;HEAP8[$prev$0$lcssa34+1>>0]=($67>>8)&255;HEAP8[$prev$0$lcssa34+2>>0]=($67>>16)&255;HEAP8[$prev$0$lcssa34+3>>0]=$67>>24; //@line 906 "load_abc.cpp"
  $$0 = $67;
  STACKTOP = sp;return ($$0|0); //@line 909 "load_abc.cpp"
 }
 return (0)|0;
}
function __ZL13abc_transposePKc($v) {
 $v = $v|0;
 var $$ = 0, $$$1$i = 0, $$$12 = 0, $$$i = 0, $$0 = 0, $$1 = 0, $$10 = 0, $$11 = 0, $$11$lcssa = 0, $$12 = 0, $$15 = 0, $$18 = 0, $$2 = 0, $$21 = 0, $$24 = 0, $$28 = 0, $$29 = 0, $$3 = 0, $$4 = 0, $$5 = 0;
 var $$6 = 0, $$7 = 0, $$8 = 0, $$9 = 0, $$9$lcssa = 0, $$h$0$i = 0, $$h$0$i19 = 0, $$h$0$i30 = 0, $$h$0$i41 = 0, $$k$0$1$i = 0, $$k$0$i = 0, $$lcssa119 = 0, $$neg73 = 0, $$pre = 0, $$sum = 0, $$sum7 = 0, $$sum8 = 0, $$sum9 = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0;
 var $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $h$0$lcssa$i = 0, $h$0$lcssa$i17 = 0, $h$0$lcssa$i28 = 0;
 var $h$0$lcssa$i39 = 0, $h$04$i = 0, $h$04$i12 = 0, $h$04$i23 = 0, $h$04$i34 = 0, $i$0$1$i = 0, $i$0$i = 0, $i$0$lcssa$i = 0, $i$0$lcssa$i18 = 0, $i$0$lcssa$i29 = 0, $i$0$lcssa$i40 = 0, $i$03$i = 0, $i$03$i13 = 0, $i$03$i24 = 0, $i$03$i35 = 0, $i$1$1$i = 0, $i$1$i = 0, $isdigit$i = 0, $isdigit$i15 = 0, $isdigit$i26 = 0;
 var $isdigit$i37 = 0, $isdigit2$i = 0, $isdigit2$i11 = 0, $isdigit2$i22 = 0, $isdigit2$i33 = 0, $isdigittmp$i = 0, $isdigittmp$i14 = 0, $isdigittmp$i25 = 0, $isdigittmp$i36 = 0, $isdigittmp1$i = 0, $isdigittmp1$i10 = 0, $isdigittmp1$i21 = 0, $isdigittmp1$i32 = 0, $j$0 = 0, $j$2 = 0, $k$0$lcssa$1$i = 0, $k$0$lcssa$i = 0, $k$01$1$i = 0, $k$01$i = 0, $m$0 = 0;
 var $m$0$lcssa = 0, $m$1 = 0, $m$3 = 0, $m$5 = 0, $m$7 = 0, $m$86068 = 0, $memchr = 0, $mv$0 = 0, $mv$0$lcssa = 0, $mv$1 = 0, $mv$2 = 0, $n$0$1$i = 0, $n$0$i = 0, $oct$0$1$i = 0, $oct$0$1$i$lcssa = 0, $oct$0$i = 0, $oct$0$i$lcssa = 0, $oct$1$1$i = 0, $oct$1$i = 0, $oct$1$v$1$i = 0;
 var $oct$1$v$i = 0, $or$cond = 0, $strlenfirst = 0, $t$0 = 0, $t$0$lcssa = 0, $t$1 = 0, $t$2 = 0, $t$3 = 0, $t$4 = 0, $t$5 = 0, $t$670 = 0, $t$7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[300425] = 99; //@line 700 "load_abc.cpp"
 $$0 = $v;$212 = 99;$m$0 = 596899;$mv$0 = 1282466;$t$0 = 0;
 L1: while(1) {
  $0 = HEAP8[$$0>>0]|0; //@line 701 "load_abc.cpp"
  switch ($0<<24>>24) {
  case 93: case 0:  {
   $$lcssa119 = $212;$m$0$lcssa = $m$0;$mv$0$lcssa = $mv$0;$t$0$lcssa = $t$0;
   break L1;
   break;
  }
  default: {
  }
  }
  $1 = (_strncasecmp($$0,525011,2)|0); //@line 702 "load_abc.cpp"
  $2 = ($1|0)==(0); //@line 702 "load_abc.cpp"
  if ($2) {
   $3 = ((($$0)) + 2|0); //@line 703 "load_abc.cpp"
   $4 = HEAP8[$3>>0]|0; //@line 704 "load_abc.cpp"
   $5 = ($4<<24>>24)==(45); //@line 704 "load_abc.cpp"
   $6 = ((($$0)) + 3|0); //@line 706 "load_abc.cpp"
   $7 = $5 ? 3 : 2; //@line 704 "load_abc.cpp"
   $$1 = $5 ? $6 : $3; //@line 704 "load_abc.cpp"
   $j$0 = $5 ? -1 : 1; //@line 704 "load_abc.cpp"
   $8 = HEAP8[$$1>>0]|0; //@line 1625 "load_abc.cpp"
   $9 = $8 << 24 >> 24; //@line 1625 "load_abc.cpp"
   $isdigittmp1$i = (($9) + -48)|0; //@line 1625 "load_abc.cpp"
   $isdigit2$i = ($isdigittmp1$i>>>0)<(10); //@line 1625 "load_abc.cpp"
   if ($isdigit2$i) {
    $12 = $9;$h$04$i = 0;$i$03$i = 0;
    while(1) {
     $10 = ($h$04$i*10)|0; //@line 1626 "load_abc.cpp"
     $11 = (($12) + -48)|0; //@line 1626 "load_abc.cpp"
     $13 = (($11) + ($10))|0; //@line 1626 "load_abc.cpp"
     $14 = (($i$03$i) + 1)|0; //@line 1627 "load_abc.cpp"
     $15 = (($$1) + ($14)|0); //@line 1625 "load_abc.cpp"
     $16 = HEAP8[$15>>0]|0; //@line 1625 "load_abc.cpp"
     $17 = $16 << 24 >> 24; //@line 1625 "load_abc.cpp"
     $isdigittmp$i = (($17) + -48)|0; //@line 1625 "load_abc.cpp"
     $isdigit$i = ($isdigittmp$i>>>0)<(10); //@line 1625 "load_abc.cpp"
     if ($isdigit$i) {
      $12 = $17;$h$04$i = $13;$i$03$i = $14;
     } else {
      $h$0$lcssa$i = $13;$i$0$lcssa$i = $14;
      break;
     }
    }
   } else {
    $h$0$lcssa$i = 0;$i$0$lcssa$i = 0;
   }
   $18 = ($i$0$lcssa$i|0)==(0); //@line 1629 "load_abc.cpp"
   $$h$0$i = $18 ? 1 : $h$0$lcssa$i; //@line 1630 "load_abc.cpp"
   $$sum9 = (($i$0$lcssa$i) + ($7))|0; //@line 709 "load_abc.cpp"
   $19 = (($$0) + ($$sum9)|0); //@line 709 "load_abc.cpp"
   $20 = Math_imul($$h$0$i, $j$0)|0; //@line 710 "load_abc.cpp"
   $21 = (($20) + ($t$0))|0; //@line 710 "load_abc.cpp"
   HEAP32[300425] = 0; //@line 711 "load_abc.cpp"
   $$2 = $19;$213 = 0;$t$1 = $21;
  } else {
   $$2 = $$0;$213 = $212;$t$1 = $t$0;
  }
  $22 = (_strncasecmp($$2,525014,7)|0); //@line 713 "load_abc.cpp"
  $23 = ($22|0)==(0); //@line 713 "load_abc.cpp"
  if ($23) {
   $24 = ((($$2)) + 7|0); //@line 714 "load_abc.cpp"
   $25 = HEAP8[$24>>0]|0; //@line 715 "load_abc.cpp"
   $26 = ($25<<24>>24)==(45); //@line 715 "load_abc.cpp"
   $27 = ((($$2)) + 8|0); //@line 717 "load_abc.cpp"
   $28 = $26 ? 8 : 7; //@line 715 "load_abc.cpp"
   $$3 = $26 ? $27 : $24; //@line 715 "load_abc.cpp"
   $29 = HEAP8[$$3>>0]|0; //@line 1625 "load_abc.cpp"
   $30 = $29 << 24 >> 24; //@line 1625 "load_abc.cpp"
   $isdigittmp1$i32 = (($30) + -48)|0; //@line 1625 "load_abc.cpp"
   $isdigit2$i33 = ($isdigittmp1$i32>>>0)<(10); //@line 1625 "load_abc.cpp"
   if ($isdigit2$i33) {
    $33 = $30;$h$04$i34 = 0;$i$03$i35 = 0;
    while(1) {
     $31 = ($h$04$i34*10)|0; //@line 1626 "load_abc.cpp"
     $32 = (($33) + -48)|0; //@line 1626 "load_abc.cpp"
     $34 = (($32) + ($31))|0; //@line 1626 "load_abc.cpp"
     $35 = (($i$03$i35) + 1)|0; //@line 1627 "load_abc.cpp"
     $36 = (($$3) + ($35)|0); //@line 1625 "load_abc.cpp"
     $37 = HEAP8[$36>>0]|0; //@line 1625 "load_abc.cpp"
     $38 = $37 << 24 >> 24; //@line 1625 "load_abc.cpp"
     $isdigittmp$i36 = (($38) + -48)|0; //@line 1625 "load_abc.cpp"
     $isdigit$i37 = ($isdigittmp$i36>>>0)<(10); //@line 1625 "load_abc.cpp"
     if ($isdigit$i37) {
      $33 = $38;$h$04$i34 = $34;$i$03$i35 = $35;
     } else {
      $h$0$lcssa$i39 = $34;$i$0$lcssa$i40 = $35;
      break;
     }
    }
   } else {
    $h$0$lcssa$i39 = 0;$i$0$lcssa$i40 = 0;
   }
   $39 = ($i$0$lcssa$i40|0)==(0); //@line 1629 "load_abc.cpp"
   $$h$0$i41 = $39 ? 1 : $h$0$lcssa$i39; //@line 1630 "load_abc.cpp"
   $$sum8 = (($i$0$lcssa$i40) + ($28))|0; //@line 720 "load_abc.cpp"
   $40 = (($$2) + ($$sum8)|0); //@line 720 "load_abc.cpp"
   $41 = $26 ? -12 : 12; //@line 721 "load_abc.cpp"
   $42 = Math_imul($41, $$h$0$i41)|0; //@line 721 "load_abc.cpp"
   $43 = (($42) + ($t$1))|0; //@line 721 "load_abc.cpp"
   HEAP32[300425] = 0; //@line 722 "load_abc.cpp"
   $$4 = $40;$214 = 0;$t$2 = $43;
  } else {
   $$4 = $$2;$214 = $213;$t$2 = $t$1;
  }
  $44 = (_strncasecmp($$4,525022,10)|0); //@line 724 "load_abc.cpp"
  $45 = ($44|0)==(0); //@line 724 "load_abc.cpp"
  if ($45) {
   $46 = ((($$4)) + 10|0); //@line 725 "load_abc.cpp"
   $47 = HEAP8[$46>>0]|0; //@line 726 "load_abc.cpp"
   $48 = ($47<<24>>24)==(45); //@line 726 "load_abc.cpp"
   $49 = ((($$4)) + 11|0); //@line 728 "load_abc.cpp"
   $50 = $48 ? 11 : 10; //@line 726 "load_abc.cpp"
   $$5 = $48 ? $49 : $46; //@line 726 "load_abc.cpp"
   $j$2 = $48 ? -1 : 1; //@line 726 "load_abc.cpp"
   $51 = HEAP8[$$5>>0]|0; //@line 1625 "load_abc.cpp"
   $52 = $51 << 24 >> 24; //@line 1625 "load_abc.cpp"
   $isdigittmp1$i21 = (($52) + -48)|0; //@line 1625 "load_abc.cpp"
   $isdigit2$i22 = ($isdigittmp1$i21>>>0)<(10); //@line 1625 "load_abc.cpp"
   if ($isdigit2$i22) {
    $55 = $52;$h$04$i23 = 0;$i$03$i24 = 0;
    while(1) {
     $53 = ($h$04$i23*10)|0; //@line 1626 "load_abc.cpp"
     $54 = (($55) + -48)|0; //@line 1626 "load_abc.cpp"
     $56 = (($54) + ($53))|0; //@line 1626 "load_abc.cpp"
     $57 = (($i$03$i24) + 1)|0; //@line 1627 "load_abc.cpp"
     $58 = (($$5) + ($57)|0); //@line 1625 "load_abc.cpp"
     $59 = HEAP8[$58>>0]|0; //@line 1625 "load_abc.cpp"
     $60 = $59 << 24 >> 24; //@line 1625 "load_abc.cpp"
     $isdigittmp$i25 = (($60) + -48)|0; //@line 1625 "load_abc.cpp"
     $isdigit$i26 = ($isdigittmp$i25>>>0)<(10); //@line 1625 "load_abc.cpp"
     if ($isdigit$i26) {
      $55 = $60;$h$04$i23 = $56;$i$03$i24 = $57;
     } else {
      $h$0$lcssa$i28 = $56;$i$0$lcssa$i29 = $57;
      break;
     }
    }
   } else {
    $h$0$lcssa$i28 = 0;$i$0$lcssa$i29 = 0;
   }
   $61 = ($i$0$lcssa$i29|0)==(0); //@line 1629 "load_abc.cpp"
   $$h$0$i30 = $61 ? 1 : $h$0$lcssa$i28; //@line 1630 "load_abc.cpp"
   $$sum7 = (($i$0$lcssa$i29) + ($50))|0; //@line 731 "load_abc.cpp"
   $62 = (($$4) + ($$sum7)|0); //@line 731 "load_abc.cpp"
   $63 = Math_imul($$h$0$i30, $j$2)|0; //@line 732 "load_abc.cpp"
   $64 = (($63) + ($t$2))|0; //@line 732 "load_abc.cpp"
   HEAP32[300425] = 0; //@line 733 "load_abc.cpp"
   $$6 = $62;$215 = 0;$t$3 = $64;
  } else {
   $$6 = $$4;$215 = $214;$t$3 = $t$2;
  }
  $65 = (_strncasecmp($$6,525014,7)|0); //@line 735 "load_abc.cpp"
  $66 = ($65|0)==(0); //@line 735 "load_abc.cpp"
  if ($66) {
   $67 = ((($$6)) + 7|0); //@line 736 "load_abc.cpp"
   $68 = HEAP8[$67>>0]|0; //@line 737 "load_abc.cpp"
   $69 = ($68<<24>>24)==(45); //@line 737 "load_abc.cpp"
   $70 = ((($$6)) + 8|0); //@line 739 "load_abc.cpp"
   $71 = $69 ? 8 : 7; //@line 737 "load_abc.cpp"
   $$7 = $69 ? $70 : $67; //@line 737 "load_abc.cpp"
   $72 = HEAP8[$$7>>0]|0; //@line 1625 "load_abc.cpp"
   $73 = $72 << 24 >> 24; //@line 1625 "load_abc.cpp"
   $isdigittmp1$i10 = (($73) + -48)|0; //@line 1625 "load_abc.cpp"
   $isdigit2$i11 = ($isdigittmp1$i10>>>0)<(10); //@line 1625 "load_abc.cpp"
   if ($isdigit2$i11) {
    $76 = $73;$h$04$i12 = 0;$i$03$i13 = 0;
    while(1) {
     $74 = ($h$04$i12*10)|0; //@line 1626 "load_abc.cpp"
     $75 = (($76) + -48)|0; //@line 1626 "load_abc.cpp"
     $77 = (($75) + ($74))|0; //@line 1626 "load_abc.cpp"
     $78 = (($i$03$i13) + 1)|0; //@line 1627 "load_abc.cpp"
     $79 = (($$7) + ($78)|0); //@line 1625 "load_abc.cpp"
     $80 = HEAP8[$79>>0]|0; //@line 1625 "load_abc.cpp"
     $81 = $80 << 24 >> 24; //@line 1625 "load_abc.cpp"
     $isdigittmp$i14 = (($81) + -48)|0; //@line 1625 "load_abc.cpp"
     $isdigit$i15 = ($isdigittmp$i14>>>0)<(10); //@line 1625 "load_abc.cpp"
     if ($isdigit$i15) {
      $76 = $81;$h$04$i12 = $77;$i$03$i13 = $78;
     } else {
      $h$0$lcssa$i17 = $77;$i$0$lcssa$i18 = $78;
      break;
     }
    }
   } else {
    $h$0$lcssa$i17 = 0;$i$0$lcssa$i18 = 0;
   }
   $82 = ($i$0$lcssa$i18|0)==(0); //@line 1629 "load_abc.cpp"
   $$h$0$i19 = $82 ? 1 : $h$0$lcssa$i17; //@line 1630 "load_abc.cpp"
   $$sum = (($i$0$lcssa$i18) + ($71))|0; //@line 742 "load_abc.cpp"
   $83 = (($$6) + ($$sum)|0); //@line 742 "load_abc.cpp"
   $84 = $69 ? -12 : 12; //@line 743 "load_abc.cpp"
   $85 = Math_imul($84, $$h$0$i19)|0; //@line 743 "load_abc.cpp"
   $86 = (($85) + ($t$3))|0; //@line 743 "load_abc.cpp"
   HEAP32[300425] = 0; //@line 744 "load_abc.cpp"
   $$8 = $83;$216 = 0;$t$4 = $86;
  } else {
   $$8 = $$6;$216 = $215;$t$4 = $t$3;
  }
  $87 = (_strncasecmp($$8,525033,2)|0); //@line 746 "load_abc.cpp"
  $88 = ($87|0)==(0); //@line 746 "load_abc.cpp"
  if ($88) {
   $89 = ((($$8)) + 2|0); //@line 747 "load_abc.cpp"
   $$9 = $89;
   L34: while(1) {
    $90 = HEAP8[$$9>>0]|0; //@line 749 "load_abc.cpp"
    switch ($90<<24>>24) {
    case 93: case 32: case 0:  {
     $$9$lcssa = $$9;
     break L34;
     break;
    }
    default: {
    }
    }
    $91 = ((($$9)) + 1|0); //@line 749 "load_abc.cpp"
    $$9 = $91;
   }
   HEAP32[300425] = 0; //@line 750 "load_abc.cpp"
   $$10 = $$9$lcssa;$217 = 0;$mv$1 = $89;
  } else {
   $$10 = $$8;$217 = $216;$mv$1 = $mv$0;
  }
  $92 = (_strncasecmp($$10,525036,7)|0); //@line 752 "load_abc.cpp"
  $93 = ($92|0)==(0); //@line 752 "load_abc.cpp"
  if ($93) {
   $94 = ((($$10)) + 7|0); //@line 753 "load_abc.cpp"
   $$11 = $94;
   L41: while(1) {
    $95 = HEAP8[$$11>>0]|0; //@line 755 "load_abc.cpp"
    switch ($95<<24>>24) {
    case 93: case 32: case 0:  {
     $$11$lcssa = $$11;
     break L41;
     break;
    }
    default: {
    }
    }
    $96 = ((($$11)) + 1|0); //@line 755 "load_abc.cpp"
    $$11 = $96;
   }
   HEAP32[300425] = 0; //@line 756 "load_abc.cpp"
   $$12 = $$11$lcssa;$121 = 0;$mv$2 = $94;
  } else {
   $$12 = $$10;$121 = $217;$mv$2 = $mv$1;
  }
  $97 = (_strncasecmp($$12,525044,5)|0); //@line 758 "load_abc.cpp"
  $98 = ($97|0)==(0); //@line 758 "load_abc.cpp"
  $99 = ((($$12)) + 5|0); //@line 759 "load_abc.cpp"
  $$$12 = $98 ? $99 : $$12; //@line 758 "load_abc.cpp"
  $100 = (_strncasecmp($$$12,525050,6)|0); //@line 761 "load_abc.cpp"
  $101 = ($100|0)==(0); //@line 761 "load_abc.cpp"
  L46: do {
   if ($101) {
    $102 = ((($$$12)) + 6|0); //@line 763 "load_abc.cpp"
    $103 = HEAP8[$102>>0]|0; //@line 764 "load_abc.cpp"
    $104 = $103 << 24 >> 24; //@line 764 "load_abc.cpp"
    switch ($104|0) {
    case 49:  {
     $105 = ((($$$12)) + 7|0); //@line 765 "load_abc.cpp"
     $$15 = $105;$m$1 = 525057;
     break;
    }
    case 50:  {
     $106 = ((($$$12)) + 7|0); //@line 766 "load_abc.cpp"
     $$15 = $106;$m$1 = 596899;
     break;
    }
    case 51:  {
     $107 = ((($$$12)) + 7|0); //@line 768 "load_abc.cpp"
     $$15 = $107;$m$1 = 525059;
     break;
    }
    case 52:  {
     $108 = ((($$$12)) + 7|0); //@line 769 "load_abc.cpp"
     $$15 = $108;$m$1 = 525061;
     break;
    }
    case 53:  {
     $109 = ((($$$12)) + 7|0); //@line 770 "load_abc.cpp"
     $$15 = $109;$m$1 = 620519;
     break;
    }
    default: {
     $$15 = $102;$m$1 = 596899;
    }
    }
    HEAP32[300425] = 0; //@line 772 "load_abc.cpp"
    $$29 = $$15;$218 = 0;$m$86068 = $m$1;$t$670 = $t$4;
   } else {
    $110 = (_strncasecmp($$$12,525063,4)|0); //@line 774 "load_abc.cpp"
    $111 = ($110|0)==(0); //@line 774 "load_abc.cpp"
    if ($111) {
     $112 = ((($$$12)) + 4|0); //@line 777 "load_abc.cpp"
     $113 = HEAP8[$112>>0]|0; //@line 778 "load_abc.cpp"
     $114 = $113 << 24 >> 24; //@line 778 "load_abc.cpp"
     switch ($114|0) {
     case 49:  {
      $115 = ((($$$12)) + 5|0); //@line 779 "load_abc.cpp"
      $$18 = $115;$m$3 = 620519;
      break;
     }
     case 50:  {
      $116 = ((($$$12)) + 5|0); //@line 780 "load_abc.cpp"
      $$18 = $116;$m$3 = 525071;
      break;
     }
     case 51:  {
      $117 = ((($$$12)) + 5|0); //@line 781 "load_abc.cpp"
      $$18 = $117;$m$3 = 525074;
      break;
     }
     case 52:  {
      $118 = ((($$$12)) + 5|0); //@line 782 "load_abc.cpp"
      $$18 = $118;$m$3 = 525068;
      break;
     }
     case 53:  {
      $119 = ((($$$12)) + 5|0); //@line 784 "load_abc.cpp"
      $$18 = $119;$m$3 = 525077;
      break;
     }
     default: {
      $$18 = $112;$m$3 = 525068;
     }
     }
     $120 = ($121|0)==(99); //@line 786 "load_abc.cpp"
     if (!($120)) {
      $$29 = $$18;$218 = $121;$m$86068 = $m$3;$t$670 = $t$4;
      break;
     }
     HEAP32[300425] = -2; //@line 787 "load_abc.cpp"
     $$29 = $$18;$218 = -2;$m$86068 = $m$3;$t$670 = $t$4;
     break;
    }
    $122 = (_strncasecmp($$$12,525081,5)|0); //@line 789 "load_abc.cpp"
    $123 = ($122|0)==(0); //@line 789 "load_abc.cpp"
    if ($123) {
     $124 = ((($$$12)) + 5|0); //@line 791 "load_abc.cpp"
     $125 = HEAP8[$124>>0]|0; //@line 792 "load_abc.cpp"
     $126 = $125 << 24 >> 24; //@line 792 "load_abc.cpp"
     switch ($126|0) {
     case 49:  {
      $127 = ((($$$12)) + 6|0); //@line 793 "load_abc.cpp"
      $$21 = $127;$m$5 = 525059;
      break;
     }
     case 50:  {
      $128 = ((($$$12)) + 6|0); //@line 794 "load_abc.cpp"
      $$21 = $128;$m$5 = 525061;
      break;
     }
     case 51:  {
      $129 = ((($$$12)) + 6|0); //@line 795 "load_abc.cpp"
      $$21 = $129;$m$5 = 620519;
      break;
     }
     case 52:  {
      $130 = ((($$$12)) + 6|0); //@line 796 "load_abc.cpp"
      $$21 = $130;$m$5 = 525071;
      break;
     }
     case 53:  {
      $131 = ((($$$12)) + 6|0); //@line 798 "load_abc.cpp"
      $$21 = $131;$m$5 = 525074;
      break;
     }
     default: {
      $$21 = $124;$m$5 = 525071;
     }
     }
     $132 = ($121|0)==(99); //@line 800 "load_abc.cpp"
     if (!($132)) {
      $$29 = $$21;$218 = $121;$m$86068 = $m$5;$t$670 = $t$4;
      break;
     }
     HEAP32[300425] = 1; //@line 801 "load_abc.cpp"
     $$29 = $$21;$218 = 1;$m$86068 = $m$5;$t$670 = $t$4;
     break;
    }
    $133 = (_strncasecmp($$$12,525087,4)|0); //@line 803 "load_abc.cpp"
    $134 = ($133|0)==(0); //@line 803 "load_abc.cpp"
    if ($134) {
     $135 = ((($$$12)) + 4|0); //@line 805 "load_abc.cpp"
     $136 = HEAP8[$135>>0]|0; //@line 806 "load_abc.cpp"
     $137 = $136 << 24 >> 24; //@line 806 "load_abc.cpp"
     switch ($137|0) {
     case 49:  {
      $138 = ((($$$12)) + 5|0); //@line 807 "load_abc.cpp"
      $$24 = $138;$m$7 = 525059;
      break;
     }
     case 50:  {
      $139 = ((($$$12)) + 5|0); //@line 808 "load_abc.cpp"
      $$24 = $139;$m$7 = 525061;
      break;
     }
     case 51:  {
      $140 = ((($$$12)) + 5|0); //@line 809 "load_abc.cpp"
      $$24 = $140;$m$7 = 620519;
      break;
     }
     case 52:  {
      $141 = ((($$$12)) + 5|0); //@line 811 "load_abc.cpp"
      $$24 = $141;$m$7 = 525071;
      break;
     }
     case 53:  {
      $142 = ((($$$12)) + 5|0); //@line 812 "load_abc.cpp"
      $$24 = $142;$m$7 = 525074;
      break;
     }
     default: {
      $$24 = $135;$m$7 = 620519;
     }
     }
     $143 = ($121|0)==(99); //@line 814 "load_abc.cpp"
     if (!($143)) {
      $$29 = $$24;$218 = $121;$m$86068 = $m$7;$t$670 = $t$4;
      break;
     }
     HEAP32[300425] = 1; //@line 815 "load_abc.cpp"
     $$29 = $$24;$218 = 1;$m$86068 = $m$7;$t$670 = $t$4;
     break;
    } else {
     $144 = HEAP8[$$$12>>0]|0; //@line 817 "load_abc.cpp"
     $145 = $144 << 24 >> 24; //@line 817 "load_abc.cpp"
     $memchr = (_memchr(525092,$145,3)|0); //@line 817 "load_abc.cpp"
     $146 = ($memchr|0)==(0|0); //@line 817 "load_abc.cpp"
     $147 = ($144<<24>>24)==(0); //@line 817 "load_abc.cpp"
     $or$cond = $147 | $146; //@line 817 "load_abc.cpp"
     if ($or$cond) {
      $$28 = $$$12;$157 = $144;
     } else {
      $148 = ((($$$12)) + 1|0); //@line 817 "load_abc.cpp"
      $149 = HEAP8[$148>>0]|0; //@line 817 "load_abc.cpp"
      $150 = ($149<<24>>24)==(56); //@line 817 "load_abc.cpp"
      if ($150) {
       switch ($145|0) {
       case 43:  {
        $151 = (($t$4) + 12)|0; //@line 820 "load_abc.cpp"
        $t$5 = $151;
        break;
       }
       case 45:  {
        $152 = (($t$4) + -12)|0; //@line 823 "load_abc.cpp"
        $t$5 = $152;
        break;
       }
       default: {
        $t$5 = $t$4;
       }
       }
       $153 = ((($$$12)) + 2|0); //@line 826 "load_abc.cpp"
       $154 = (_strncasecmp($153,525095,2)|0); //@line 827 "load_abc.cpp"
       $155 = ($154|0)==(0); //@line 827 "load_abc.cpp"
       $156 = ((($$$12)) + 4|0); //@line 827 "load_abc.cpp"
       $$ = $155 ? $156 : $153; //@line 827 "load_abc.cpp"
       HEAP32[300425] = 0; //@line 828 "load_abc.cpp"
       $$29 = $$;$218 = 0;$m$86068 = $m$0;$t$670 = $t$5;
       break;
      } else {
       $$28 = $$$12;$157 = $144;
      }
     }
     while(1) {
      switch ($157<<24>>24) {
      case 93: case 32: case 0:  {
       $$29 = $$28;$218 = $121;$m$86068 = $m$0;$t$670 = $t$4;
       break L46;
       break;
      }
      default: {
      }
      }
      $158 = ((($$28)) + 1|0); //@line 832 "load_abc.cpp"
      $$pre = HEAP8[$158>>0]|0; //@line 832 "load_abc.cpp"
      $$28 = $158;$157 = $$pre;
     }
    }
   }
  } while(0);
  $159 = ((($$29)) + 1|0); //@line 701 "load_abc.cpp"
  $$0 = $159;$212 = $218;$m$0 = $m$86068;$mv$0 = $mv$2;$t$0 = $t$670;
 }
 $strlenfirst = HEAP8[$mv$0$lcssa>>0]|0; //@line 835 "load_abc.cpp"
 $160 = ($strlenfirst<<24>>24)==(0); //@line 835 "load_abc.cpp"
 if ($160) {
  $t$7 = $t$0$lcssa;
 } else {
  $161 = $strlenfirst << 24 >> 24; //@line 657 "load_abc.cpp"
  switch ($161|0) {
  case 94:  {
   $162 = ((($mv$0$lcssa)) + 1|0); //@line 659 "load_abc.cpp"
   $163 = HEAP8[$162>>0]|0; //@line 659 "load_abc.cpp"
   $164 = $163 << 24 >> 24; //@line 659 "load_abc.cpp"
   $i$0$i = 2;$n$0$i = $164;
   break;
  }
  case 95:  {
   $165 = ((($mv$0$lcssa)) + 1|0); //@line 663 "load_abc.cpp"
   $166 = HEAP8[$165>>0]|0; //@line 663 "load_abc.cpp"
   $167 = $166 << 24 >> 24; //@line 663 "load_abc.cpp"
   $i$0$i = 2;$n$0$i = $167;
   break;
  }
  case 61:  {
   $168 = ((($mv$0$lcssa)) + 1|0); //@line 667 "load_abc.cpp"
   $169 = HEAP8[$168>>0]|0; //@line 667 "load_abc.cpp"
   $170 = $169 << 24 >> 24; //@line 667 "load_abc.cpp"
   $i$0$i = 2;$n$0$i = $170;
   break;
  }
  default: {
   $i$0$i = 1;$n$0$i = $161;
  }
  }
  $k$01$i = 0;
  while(1) {
   $171 = (525098 + ($k$01$i)|0); //@line 676 "load_abc.cpp"
   $172 = HEAP8[$171>>0]|0; //@line 676 "load_abc.cpp"
   $173 = $172 << 24 >> 24; //@line 676 "load_abc.cpp"
   $174 = ($n$0$i|0)==($173|0); //@line 676 "load_abc.cpp"
   if ($174) {
    $k$0$lcssa$i = $k$01$i;
    break;
   }
   $175 = (($k$01$i) + 1)|0; //@line 675 "load_abc.cpp"
   $176 = ($175|0)<(25); //@line 675 "load_abc.cpp"
   if ($176) {
    $k$01$i = $175;
   } else {
    $k$0$lcssa$i = $175;
    break;
   }
  }
  $177 = ($k$0$lcssa$i|0)>(12); //@line 679 "load_abc.cpp"
  $$$i = $177 ? 5 : 4; //@line 679 "load_abc.cpp"
  $i$1$i = $i$0$i;$oct$0$i = $$$i;
  L112: while(1) {
   $178 = (($mv$0$lcssa) + ($i$1$i)|0); //@line 683 "load_abc.cpp"
   $179 = HEAP8[$178>>0]|0; //@line 683 "load_abc.cpp"
   switch ($179<<24>>24) {
   case 39: case 44:  {
    break;
   }
   default: {
    $oct$0$i$lcssa = $oct$0$i;
    break L112;
   }
   }
   $180 = ($179<<24>>24)==(44); //@line 684 "load_abc.cpp"
   $oct$1$v$i = $180 ? -1 : 1; //@line 684 "load_abc.cpp"
   $oct$1$i = (($oct$1$v$i) + ($oct$0$i))|0; //@line 684 "load_abc.cpp"
   $181 = (($i$1$i) + 1)|0; //@line 688 "load_abc.cpp"
   $i$1$i = $181;$oct$0$i = $oct$1$i;
  }
  $182 = (($k$0$lcssa$i) + -12)|0; //@line 681 "load_abc.cpp"
  $$k$0$i = $177 ? $182 : $k$0$lcssa$i; //@line 679 "load_abc.cpp"
  $183 = ($oct$0$i$lcssa*12)|0; //@line 690 "load_abc.cpp"
  $184 = HEAP8[$m$0$lcssa>>0]|0; //@line 657 "load_abc.cpp"
  $185 = $184 << 24 >> 24; //@line 657 "load_abc.cpp"
  switch ($185|0) {
  case 94:  {
   $192 = ((($m$0$lcssa)) + 1|0); //@line 659 "load_abc.cpp"
   $193 = HEAP8[$192>>0]|0; //@line 659 "load_abc.cpp"
   $194 = $193 << 24 >> 24; //@line 659 "load_abc.cpp"
   $i$0$1$i = 2;$n$0$1$i = $194;
   break;
  }
  case 95:  {
   $189 = ((($m$0$lcssa)) + 1|0); //@line 663 "load_abc.cpp"
   $190 = HEAP8[$189>>0]|0; //@line 663 "load_abc.cpp"
   $191 = $190 << 24 >> 24; //@line 663 "load_abc.cpp"
   $i$0$1$i = 2;$n$0$1$i = $191;
   break;
  }
  case 61:  {
   $186 = ((($m$0$lcssa)) + 1|0); //@line 667 "load_abc.cpp"
   $187 = HEAP8[$186>>0]|0; //@line 667 "load_abc.cpp"
   $188 = $187 << 24 >> 24; //@line 667 "load_abc.cpp"
   $i$0$1$i = 2;$n$0$1$i = $188;
   break;
  }
  default: {
   $i$0$1$i = 1;$n$0$1$i = $185;
  }
  }
  $k$01$1$i = 0;
  while(1) {
   $195 = (525098 + ($k$01$1$i)|0); //@line 676 "load_abc.cpp"
   $196 = HEAP8[$195>>0]|0; //@line 676 "load_abc.cpp"
   $197 = $196 << 24 >> 24; //@line 676 "load_abc.cpp"
   $198 = ($n$0$1$i|0)==($197|0); //@line 676 "load_abc.cpp"
   if ($198) {
    $k$0$lcssa$1$i = $k$01$1$i;
    break;
   }
   $199 = (($k$01$1$i) + 1)|0; //@line 675 "load_abc.cpp"
   $200 = ($199|0)<(25); //@line 675 "load_abc.cpp"
   if ($200) {
    $k$01$1$i = $199;
   } else {
    $k$0$lcssa$1$i = $199;
    break;
   }
  }
  $201 = ($k$0$lcssa$1$i|0)>(12); //@line 679 "load_abc.cpp"
  $$$1$i = $201 ? 5 : 4; //@line 679 "load_abc.cpp"
  $i$1$1$i = $i$0$1$i;$oct$0$1$i = $$$1$i;
  L125: while(1) {
   $202 = (($m$0$lcssa) + ($i$1$1$i)|0); //@line 683 "load_abc.cpp"
   $203 = HEAP8[$202>>0]|0; //@line 683 "load_abc.cpp"
   switch ($203<<24>>24) {
   case 39: case 44:  {
    break;
   }
   default: {
    $oct$0$1$i$lcssa = $oct$0$1$i;
    break L125;
   }
   }
   $204 = ($203<<24>>24)==(44); //@line 684 "load_abc.cpp"
   $oct$1$v$1$i = $204 ? -1 : 1; //@line 684 "load_abc.cpp"
   $oct$1$1$i = (($oct$1$v$1$i) + ($oct$0$1$i))|0; //@line 684 "load_abc.cpp"
   $205 = (($i$1$1$i) + 1)|0; //@line 688 "load_abc.cpp"
   $i$1$1$i = $205;$oct$0$1$i = $oct$1$1$i;
  }
  $206 = (($k$0$lcssa$1$i) + -12)|0; //@line 681 "load_abc.cpp"
  $$k$0$1$i = $201 ? $206 : $k$0$lcssa$1$i; //@line 679 "load_abc.cpp"
  $207 = Math_imul($oct$0$1$i$lcssa, -12)|0;
  $$neg73 = (($$k$0$i) + ($t$0$lcssa))|0; //@line 690 "load_abc.cpp"
  $208 = (($$neg73) + ($183))|0; //@line 690 "load_abc.cpp"
  $209 = (($208) - ($$k$0$1$i))|0; //@line 692 "load_abc.cpp"
  $210 = (($209) + ($207))|0; //@line 836 "load_abc.cpp"
  $t$7 = $210;
 }
 $211 = ($$lcssa119|0)==(99); //@line 837 "load_abc.cpp"
 if (!($211)) {
  return ($t$7|0); //@line 839 "load_abc.cpp"
 }
 HEAP32[300425] = 0; //@line 838 "load_abc.cpp"
 return ($t$7|0); //@line 839 "load_abc.cpp"
}
function __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($tp,$e) {
 $tp = $tp|0;
 $e = $e|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($tp)) + 12|0); //@line 1134 "load_abc.cpp"
 $1 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 1134 "load_abc.cpp"
 $2 = ($1|0)==(0|0); //@line 1134 "load_abc.cpp"
 if ($2) {
  $3 = HEAP32[299861]|0; //@line 1134 "load_abc.cpp"
  $4 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
  HEAP8[$4>>0]=0&255;HEAP8[$4+1>>0]=(0>>8)&255;HEAP8[$4+2>>0]=(0>>16)&255;HEAP8[$4+3>>0]=0>>24; //@line 540 "load_abc.cpp"
  $5 = ((($4)) + 4|0); //@line 541 "load_abc.cpp"
  HEAP8[$5>>0]=$3&255;HEAP8[$5+1>>0]=($3>>8)&255;HEAP8[$5+2>>0]=($3>>16)&255;HEAP8[$5+3>>0]=$3>>24; //@line 541 "load_abc.cpp"
  $6 = ((($4)) + 8|0); //@line 543 "load_abc.cpp"
  HEAP8[$6>>0] = 1; //@line 543 "load_abc.cpp"
  $7 = ((($4)) + 9|0); //@line 543 "load_abc.cpp"
  HEAP8[$7>>0] = 66; //@line 543 "load_abc.cpp"
  $8 = ((($4)) + 10|0); //@line 543 "load_abc.cpp"
  HEAP8[$8>>0]=0&255;HEAP8[$8+1>>0]=(0>>8)&255;HEAP8[$8+2>>0]=(0>>16)&255;HEAP8[$8+3>>0]=0>>24; //@line 543 "load_abc.cpp"
  $9 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
  $10 = ((($4)) + 14|0); //@line 544 "load_abc.cpp"
  HEAP8[$10>>0] = $9; //@line 544 "load_abc.cpp"
  $11 = ((($4)) + 15|0); //@line 545 "load_abc.cpp"
  HEAP8[$11>>0] = 0; //@line 545 "load_abc.cpp"
  HEAP8[$0>>0]=$4&255;HEAP8[$0+1>>0]=($4>>8)&255;HEAP8[$0+2>>0]=($4>>16)&255;HEAP8[$0+3>>0]=$4>>24; //@line 930 "load_abc.cpp"
  __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($tp,$4); //@line 931 "load_abc.cpp"
 }
 $12 = ((($tp)) + 8|0); //@line 1135 "load_abc.cpp"
 $13 = HEAPU8[$12>>0]|(HEAPU8[$12+1>>0]<<8)|(HEAPU8[$12+2>>0]<<16)|(HEAPU8[$12+3>>0]<<24); //@line 1135 "load_abc.cpp"
 $14 = ($13|0)==(0|0); //@line 1135 "load_abc.cpp"
 if ($14) {
  $15 = ((($tp)) + 4|0); //@line 1140 "load_abc.cpp"
  HEAP8[$15>>0]=$e&255;HEAP8[$15+1>>0]=($e>>8)&255;HEAP8[$15+2>>0]=($e>>16)&255;HEAP8[$15+3>>0]=$e>>24; //@line 1140 "load_abc.cpp"
  HEAP8[$12>>0]=$e&255;HEAP8[$12+1>>0]=($e>>8)&255;HEAP8[$12+2>>0]=($e>>16)&255;HEAP8[$12+3>>0]=$e>>24; //@line 1137 "load_abc.cpp"
  return; //@line 1143 "load_abc.cpp"
 } else {
  HEAP8[$13>>0]=$e&255;HEAP8[$13+1>>0]=($e>>8)&255;HEAP8[$13+2>>0]=($e>>16)&255;HEAP8[$13+3>>0]=$e>>24; //@line 1136 "load_abc.cpp"
  HEAP8[$12>>0]=$e&255;HEAP8[$12+1>>0]=($e>>8)&255;HEAP8[$12+2>>0]=($e>>16)&255;HEAP8[$12+3>>0]=$e>>24; //@line 1137 "load_abc.cpp"
  return; //@line 1143 "load_abc.cpp"
 }
}
function __ZL11abc_getexprPKcPi($p,$number) {
 $p = $p|0;
 $number = $number|0;
 var $$0 = 0, $$h$0$i = 0, $$lcssa = 0, $$lcssa34 = 0, $$lcssa36 = 0, $$lcssa39 = 0, $$lcssa40 = 0, $$sum = 0, $$sum2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $h$0$lcssa$i = 0, $h$04$i = 0, $i$0 = 0, $i$0$lcssa = 0;
 var $i$0$lcssa$i = 0, $i$03$i = 0, $i$1 = 0, $i$2 = 0, $i$2$lcssa = 0, $i$3$lcssa = 0, $i$37 = 0, $i$4 = 0, $i$4$lcssa = 0, $isdigit$i = 0, $isdigit2$i = 0, $isdigittmp$i = 0, $isdigittmp1$i = 0, $term = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $term = sp;
 $i$0 = 0;
 while(1) {
  $0 = (($p) + ($i$0)|0); //@line 1640 "load_abc.cpp"
  $1 = HEAP8[$0>>0]|0; //@line 1640 "load_abc.cpp"
  $2 = $1 << 24 >> 24; //@line 1640 "load_abc.cpp"
  $3 = (_isspace($2)|0); //@line 1640 "load_abc.cpp"
  $4 = ($3|0)==(0); //@line 1640 "load_abc.cpp"
  $5 = (($i$0) + 1)|0; //@line 1641 "load_abc.cpp"
  if ($4) {
   $$lcssa39 = $0;$$lcssa40 = $5;$i$0$lcssa = $i$0;
   break;
  } else {
   $i$0 = $5;
  }
 }
 $6 = HEAP8[$$lcssa39>>0]|0; //@line 1642 "load_abc.cpp"
 $7 = ($6<<24>>24)==(40); //@line 1642 "load_abc.cpp"
 if ($7) {
  $8 = (($p) + ($$lcssa40)|0); //@line 1643 "load_abc.cpp"
  $9 = (__ZL11abc_getexprPKcPi($8,$number)|0); //@line 1643 "load_abc.cpp"
  $10 = (($9) + ($i$0$lcssa))|0; //@line 1643 "load_abc.cpp"
  $i$1 = $10;
  L6: while(1) {
   $11 = (($p) + ($i$1)|0); //@line 1644 "load_abc.cpp"
   $12 = HEAP8[$11>>0]|0; //@line 1644 "load_abc.cpp"
   switch ($12<<24>>24) {
   case 0: case 41:  {
    $$0 = $i$1;
    break L6;
    break;
   }
   default: {
   }
   }
   $13 = (($i$1) + 1)|0; //@line 1645 "load_abc.cpp"
   $i$1 = $13;
  }
  STACKTOP = sp;return ($$0|0); //@line 1659 "load_abc.cpp"
 }
 $14 = $6 << 24 >> 24; //@line 1625 "load_abc.cpp"
 $isdigittmp1$i = (($14) + -48)|0; //@line 1625 "load_abc.cpp"
 $isdigit2$i = ($isdigittmp1$i>>>0)<(10); //@line 1625 "load_abc.cpp"
 if ($isdigit2$i) {
  $17 = $14;$h$04$i = 0;$i$03$i = 0;
  while(1) {
   $15 = ($h$04$i*10)|0; //@line 1626 "load_abc.cpp"
   $16 = (($17) + -48)|0; //@line 1626 "load_abc.cpp"
   $18 = (($16) + ($15))|0; //@line 1626 "load_abc.cpp"
   $19 = (($i$03$i) + 1)|0; //@line 1627 "load_abc.cpp"
   $$sum2 = (($19) + ($i$0$lcssa))|0; //@line 1625 "load_abc.cpp"
   $20 = (($p) + ($$sum2)|0); //@line 1625 "load_abc.cpp"
   $21 = HEAP8[$20>>0]|0; //@line 1625 "load_abc.cpp"
   $22 = $21 << 24 >> 24; //@line 1625 "load_abc.cpp"
   $isdigittmp$i = (($22) + -48)|0; //@line 1625 "load_abc.cpp"
   $isdigit$i = ($isdigittmp$i>>>0)<(10); //@line 1625 "load_abc.cpp"
   if ($isdigit$i) {
    $17 = $22;$h$04$i = $18;$i$03$i = $19;
   } else {
    $h$0$lcssa$i = $18;$i$0$lcssa$i = $19;
    break;
   }
  }
 } else {
  $h$0$lcssa$i = 0;$i$0$lcssa$i = 0;
 }
 $23 = ($i$0$lcssa$i|0)==(0); //@line 1629 "load_abc.cpp"
 $24 = (($i$0$lcssa$i) + ($i$0$lcssa))|0; //@line 1648 "load_abc.cpp"
 $i$2 = $24;
 while(1) {
  $25 = (($p) + ($i$2)|0); //@line 1649 "load_abc.cpp"
  $26 = HEAP8[$25>>0]|0; //@line 1649 "load_abc.cpp"
  $27 = $26 << 24 >> 24; //@line 1649 "load_abc.cpp"
  $28 = (_isspace($27)|0); //@line 1649 "load_abc.cpp"
  $29 = ($28|0)==(0); //@line 1649 "load_abc.cpp"
  $30 = (($i$2) + 1)|0; //@line 1650 "load_abc.cpp"
  if ($29) {
   $$lcssa36 = $25;$i$2$lcssa = $i$2;
   break;
  } else {
   $i$2 = $30;
  }
 }
 $$h$0$i = $23 ? 1 : $h$0$lcssa$i; //@line 1630 "load_abc.cpp"
 $31 = HEAP8[$$lcssa36>>0]|0; //@line 1651 "load_abc.cpp"
 $32 = ($31<<24>>24)==(43); //@line 1651 "load_abc.cpp"
 if ($32) {
  $35 = $$h$0$i;$i$37 = $i$2$lcssa;
  while(1) {
   $$sum = (($i$37) + 1)|0; //@line 1652 "load_abc.cpp"
   $38 = (($p) + ($$sum)|0); //@line 1652 "load_abc.cpp"
   $39 = (__ZL11abc_getexprPKcPi($38,$term)|0); //@line 1652 "load_abc.cpp"
   $40 = (($$sum) + ($39))|0; //@line 1652 "load_abc.cpp"
   $34 = HEAP32[$term>>2]|0; //@line 1653 "load_abc.cpp"
   $i$4 = $40;
   while(1) {
    $41 = (($p) + ($i$4)|0); //@line 1654 "load_abc.cpp"
    $42 = HEAP8[$41>>0]|0; //@line 1654 "load_abc.cpp"
    $43 = $42 << 24 >> 24; //@line 1654 "load_abc.cpp"
    $44 = (_isspace($43)|0); //@line 1654 "load_abc.cpp"
    $45 = ($44|0)==(0); //@line 1654 "load_abc.cpp"
    $46 = (($i$4) + 1)|0; //@line 1655 "load_abc.cpp"
    if ($45) {
     $$lcssa34 = $41;$i$4$lcssa = $i$4;
     break;
    } else {
     $i$4 = $46;
    }
   }
   $33 = (($34) + ($35))|0; //@line 1653 "load_abc.cpp"
   $36 = HEAP8[$$lcssa34>>0]|0; //@line 1651 "load_abc.cpp"
   $37 = ($36<<24>>24)==(43); //@line 1651 "load_abc.cpp"
   if ($37) {
    $35 = $33;$i$37 = $i$4$lcssa;
   } else {
    $$lcssa = $33;$i$3$lcssa = $i$4$lcssa;
    break;
   }
  }
 } else {
  $$lcssa = $$h$0$i;$i$3$lcssa = $i$2$lcssa;
 }
 HEAP32[$number>>2] = $$lcssa; //@line 1657 "load_abc.cpp"
 $$0 = $i$3$lcssa;
 STACKTOP = sp;return ($$0|0); //@line 1659 "load_abc.cpp"
}
function __ZL13abc_set_partsPPcS_($d,$p) {
 $d = $d|0;
 $p = $p|0;
 var $$ = 0, $$$h$0$i = 0, $$h$0$i = 0, $$in = 0, $$lcssa = 0, $$lcssa116 = 0, $$lcssa118 = 0, $$lcssa119 = 0, $$lcssa123 = 0, $$lcssa124 = 0, $$sum = 0, $$sum37 = 0, $$sum38 = 0, $$sum39 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $exitcond80 = 0, $h$04$i = 0, $h$04$i12 = 0, $h$04$i25 = 0, $i$0 = 0, $i$03$i = 0, $i$03$i$lcssa = 0, $i$03$i13 = 0, $i$03$i26 = 0;
 var $i$03$i26$lcssa = 0, $i$1 = 0, $i$2 = 0, $i$3 = 0, $i$4 = 0, $i$556 = 0, $isdigit = 0, $isdigit$i = 0, $isdigit$i15 = 0, $isdigit$i28 = 0, $isdigit2$i11 = 0, $isdigit5 = 0, $isdigit8 = 0, $isdigittmp = 0, $isdigittmp$i = 0, $isdigittmp$i14 = 0, $isdigittmp$i27 = 0, $isdigittmp1$i10 = 0, $isdigittmp4 = 0, $isdigittmp7 = 0;
 var $j$0 = 0, $j$0$ = 0, $j$0$lcssa = 0, $j$10 = 0, $j$2 = 0, $j$3 = 0, $j$4 = 0, $j$4$lcssa121 = 0, $j$4122 = 0, $j$568$us = 0, $j$664$us = 0, $j$760 = 0, $j$8 = 0, $j$955 = 0, $m$065$us = 0, $memchr = 0, $n$0 = 0, $n$0$lcssa117 = 0, $not$ = 0, $or$cond = 0;
 var $or$cond94 = 0, $or$cond95 = 0, $smax = 0, $storemerge54 = 0, $storemerge662 = 0, $switch = 0, $txt$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer10 = 0, $vararg_buffer13 = 0, $vararg_buffer4 = 0, $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer13 = sp + 40|0;
 $vararg_buffer10 = sp + 32|0;
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $txt$i = sp + 48|0;
 $0 = HEAP32[$d>>2]|0; //@line 1820 "load_abc.cpp"
 $1 = ($0|0)==(0|0); //@line 1820 "load_abc.cpp"
 if (!($1)) {
  _free($0); //@line 1820 "load_abc.cpp"
 }
 HEAP32[$d>>2] = 0; //@line 1822 "load_abc.cpp"
 $2 = ($p|0)==(0|0); //@line 1823 "load_abc.cpp"
 if ($2) {
  STACKTOP = sp;return; //@line 1901 "load_abc.cpp"
 } else {
  $i$0 = 0;
 }
 L6: while(1) {
  $3 = (($p) + ($i$0)|0); //@line 1824 "load_abc.cpp"
  $4 = HEAP8[$3>>0]|0; //@line 1824 "load_abc.cpp"
  switch ($4<<24>>24) {
  case 37: case 0:  {
   $i$1 = 0;$j$0 = 0;
   break L6;
   break;
  }
  default: {
  }
  }
  $5 = $4 << 24 >> 24; //@line 1824 "load_abc.cpp"
  $memchr = (_memchr(525156,$5,41)|0); //@line 1825 "load_abc.cpp"
  $6 = ($memchr|0)==(0|0); //@line 1825 "load_abc.cpp"
  $7 = (($i$0) + 1)|0; //@line 1824 "load_abc.cpp"
  if ($6) {
   label = 6;
   break;
  } else {
   $i$0 = $7;
  }
 }
 if ((label|0) == 6) {
  $8 = (_strlen($p)|0); //@line 323 "load_abc.cpp"
  $9 = (($8) + 47)|0; //@line 323 "load_abc.cpp"
  $10 = ($9>>>0)>(255); //@line 323 "load_abc.cpp"
  if (!($10)) {
   HEAP32[$vararg_buffer>>2] = $p; //@line 324 "load_abc.cpp"
   (_sprintf($txt$i,525197,$vararg_buffer)|0); //@line 324 "load_abc.cpp"
   $11 = HEAP32[37544]|0; //@line 328 "load_abc.cpp"
   HEAP32[$vararg_buffer1>>2] = $txt$i; //@line 328 "load_abc.cpp"
   (_fprintf($11,524747,$vararg_buffer1)|0); //@line 328 "load_abc.cpp"
  }
  STACKTOP = sp;return; //@line 1901 "load_abc.cpp"
 }
 L15: while(1) {
  $12 = (($p) + ($i$1)|0); //@line 1837 "load_abc.cpp"
  $13 = HEAP8[$12>>0]|0; //@line 1837 "load_abc.cpp"
  switch ($13<<24>>24) {
  case 37: case 0:  {
   $j$0$lcssa = $j$0;
   break L15;
   break;
  }
  default: {
  }
  }
  $14 = $13 << 24 >> 24; //@line 1837 "load_abc.cpp"
  $15 = (_isupper($14)|0); //@line 1838 "load_abc.cpp"
  $not$ = ($15|0)!=(0); //@line 1838 "load_abc.cpp"
  $16 = $not$&1; //@line 1838 "load_abc.cpp"
  $j$0$ = (($16) + ($j$0))|0; //@line 1838 "load_abc.cpp"
  $17 = HEAP8[$12>>0]|0; //@line 1841 "load_abc.cpp"
  $18 = $17 << 24 >> 24; //@line 1841 "load_abc.cpp"
  $isdigittmp7 = (($18) + -48)|0; //@line 1841 "load_abc.cpp"
  $isdigit8 = ($isdigittmp7>>>0)<(10); //@line 1841 "load_abc.cpp"
  if ($isdigit8) {
   $21 = $18;$h$04$i = 0;$i$03$i = 0;
   while(1) {
    $19 = ($h$04$i*10)|0; //@line 1626 "load_abc.cpp"
    $20 = (($21) + -48)|0; //@line 1626 "load_abc.cpp"
    $22 = (($20) + ($19))|0; //@line 1626 "load_abc.cpp"
    $23 = (($i$03$i) + 1)|0; //@line 1627 "load_abc.cpp"
    $$sum39 = (($23) + ($i$1))|0; //@line 1625 "load_abc.cpp"
    $24 = (($p) + ($$sum39)|0); //@line 1625 "load_abc.cpp"
    $25 = HEAP8[$24>>0]|0; //@line 1625 "load_abc.cpp"
    $26 = $25 << 24 >> 24; //@line 1625 "load_abc.cpp"
    $isdigittmp$i = (($26) + -48)|0; //@line 1625 "load_abc.cpp"
    $isdigit$i = ($isdigittmp$i>>>0)<(10); //@line 1625 "load_abc.cpp"
    if ($isdigit$i) {
     $21 = $26;$h$04$i = $22;$i$03$i = $23;
    } else {
     $$lcssa123 = $22;$$lcssa124 = $23;$i$03$i$lcssa = $i$03$i;
     break;
    }
   }
   $27 = ($$lcssa124|0)==(0); //@line 1629 "load_abc.cpp"
   $$h$0$i = $27 ? 1 : $$lcssa123; //@line 1630 "load_abc.cpp"
   $28 = ($$h$0$i|0)==(0); //@line 1843 "load_abc.cpp"
   $$$h$0$i = $28 ? 1 : $$h$0$i; //@line 1843 "load_abc.cpp"
   $29 = (($i$1) + -1)|0; //@line 1845 "load_abc.cpp"
   $30 = (($p) + ($29)|0); //@line 1845 "load_abc.cpp"
   $31 = HEAP8[$30>>0]|0; //@line 1845 "load_abc.cpp"
   $32 = ($31<<24>>24)==(41); //@line 1845 "load_abc.cpp"
   $33 = Math_imul($$$h$0$i, $j$0$)|0; //@line 1846 "load_abc.cpp"
   $34 = (($j$0$) + -1)|0; //@line 1848 "load_abc.cpp"
   $35 = (($34) + ($$$h$0$i))|0; //@line 1848 "load_abc.cpp"
   $j$2 = $32 ? $33 : $35; //@line 1845 "load_abc.cpp"
   $36 = (($i$03$i$lcssa) + ($i$1))|0; //@line 1849 "load_abc.cpp"
   $i$2 = $36;$j$3 = $j$2;
  } else {
   $i$2 = $i$1;$j$3 = $j$0$;
  }
  $37 = (($i$2) + 1)|0; //@line 1837 "load_abc.cpp"
  $i$1 = $37;$j$0 = $j$3;
 }
 $38 = (($j$0$lcssa) + 1)|0; //@line 1852 "load_abc.cpp"
 $39 = (_calloc($38,1)|0); //@line 1852 "load_abc.cpp"
 $i$3 = 0;$j$4 = 0;
 L24: while(1) {
  $40 = (($p) + ($i$3)|0); //@line 1855 "load_abc.cpp"
  $41 = HEAP8[$40>>0]|0; //@line 1855 "load_abc.cpp"
  switch ($41<<24>>24) {
  case 37: case 0:  {
   $j$4122 = $j$4;
   break L24;
   break;
  }
  default: {
  }
  }
  $42 = $41 << 24 >> 24; //@line 1855 "load_abc.cpp"
  $isdigittmp = (($42) + -48)|0; //@line 1856 "load_abc.cpp"
  $isdigit = ($isdigittmp>>>0)<(10); //@line 1856 "load_abc.cpp"
  if ($isdigit) {
   $48 = $41;
   label = 18;
  } else {
   $43 = (_isupper($42)|0); //@line 1856 "load_abc.cpp"
   $44 = ($43|0)!=(0); //@line 1856 "load_abc.cpp"
   $45 = HEAP8[$40>>0]|0; //@line 1856 "load_abc.cpp"
   $46 = $45 & -2;
   $switch = ($46<<24>>24)==(40); //@line 1856 "load_abc.cpp"
   $or$cond = $44 | $switch; //@line 1856 "load_abc.cpp"
   if ($or$cond) {
    $48 = $45;
    label = 18;
   } else {
    $i$4 = $i$3;$j$8 = $j$4;
   }
  }
  do {
   if ((label|0) == 18) {
    label = 0;
    $47 = ($48<<24>>24)==(41); //@line 1857 "load_abc.cpp"
    if ($47) {
     $n$0 = $j$4;
    } else {
     $94 = $48 << 24 >> 24; //@line 1857 "load_abc.cpp"
     $isdigittmp4 = (($94) + -48)|0; //@line 1878 "load_abc.cpp"
     $isdigit5 = ($isdigittmp4>>>0)<(10); //@line 1878 "load_abc.cpp"
     if ($isdigit5) {
      $97 = $94;$h$04$i25 = 0;$i$03$i26 = 0;
     } else {
      $115 = (($39) + ($j$4)|0); //@line 1887 "load_abc.cpp"
      HEAP8[$115>>0] = $48; //@line 1887 "load_abc.cpp"
      $116 = (($j$4) + 1)|0; //@line 1888 "load_abc.cpp"
      $i$4 = $i$3;$j$8 = $116;
      break;
     }
     while(1) {
      $95 = ($h$04$i25*10)|0; //@line 1626 "load_abc.cpp"
      $96 = (($97) + -48)|0; //@line 1626 "load_abc.cpp"
      $98 = (($96) + ($95))|0; //@line 1626 "load_abc.cpp"
      $99 = (($i$03$i26) + 1)|0; //@line 1627 "load_abc.cpp"
      $$sum37 = (($99) + ($i$3))|0; //@line 1625 "load_abc.cpp"
      $100 = (($p) + ($$sum37)|0); //@line 1625 "load_abc.cpp"
      $101 = HEAP8[$100>>0]|0; //@line 1625 "load_abc.cpp"
      $102 = $101 << 24 >> 24; //@line 1625 "load_abc.cpp"
      $isdigittmp$i27 = (($102) + -48)|0; //@line 1625 "load_abc.cpp"
      $isdigit$i28 = ($isdigittmp$i27>>>0)<(10); //@line 1625 "load_abc.cpp"
      if ($isdigit$i28) {
       $97 = $102;$h$04$i25 = $98;$i$03$i26 = $99;
      } else {
       $$lcssa = $98;$$lcssa116 = $99;$i$03$i26$lcssa = $i$03$i26;
       break;
      }
     }
     $103 = ($$lcssa116|0)==(0); //@line 1629 "load_abc.cpp"
     $104 = (($i$03$i26$lcssa) + ($i$3))|0; //@line 1880 "load_abc.cpp"
     $$ = $103 ? 1 : $$lcssa;
     $105 = ($$|0)>(1); //@line 1881 "load_abc.cpp"
     if (!($105)) {
      $i$4 = $104;$j$8 = $j$4;
      break;
     }
     $106 = (($$) + -1)|0; //@line 1881 "load_abc.cpp"
     $108 = $$;$j$760 = $j$4;
     while(1) {
      $107 = (($108) + -1)|0; //@line 1881 "load_abc.cpp"
      $109 = (($j$760) + -1)|0; //@line 1882 "load_abc.cpp"
      $110 = (($39) + ($109)|0); //@line 1882 "load_abc.cpp"
      $111 = HEAP8[$110>>0]|0; //@line 1882 "load_abc.cpp"
      $112 = (($39) + ($j$760)|0); //@line 1882 "load_abc.cpp"
      HEAP8[$112>>0] = $111; //@line 1882 "load_abc.cpp"
      $113 = (($j$760) + 1)|0; //@line 1883 "load_abc.cpp"
      $114 = ($107|0)>(1); //@line 1881 "load_abc.cpp"
      if ($114) {
       $108 = $107;$j$760 = $113;
      } else {
       break;
      }
     }
     $117 = (($106) + ($j$4))|0; //@line 1881 "load_abc.cpp"
     $i$4 = $104;$j$8 = $117;
     break;
    }
    while(1) {
     $49 = ($n$0|0)>(0); //@line 1858 "load_abc.cpp"
     if (!($49)) {
      $j$4$lcssa121 = $j$4;
      label = 26;
      break L24;
     }
     $50 = (($n$0) + -1)|0; //@line 1858 "load_abc.cpp"
     $51 = (($39) + ($50)|0); //@line 1858 "load_abc.cpp"
     $52 = HEAP8[$51>>0]|0; //@line 1858 "load_abc.cpp"
     $53 = ($52<<24>>24)==(40); //@line 1858 "load_abc.cpp"
     if ($53) {
      $n$0$lcssa117 = $n$0;
      break;
     } else {
      $n$0 = $50;
     }
    }
    $54 = ($j$4|0)>($n$0$lcssa117|0); //@line 1861 "load_abc.cpp"
    if ($54) {
     $storemerge662 = $n$0$lcssa117;
     while(1) {
      $55 = (($39) + ($storemerge662)|0); //@line 1861 "load_abc.cpp"
      $56 = HEAP8[$55>>0]|0; //@line 1861 "load_abc.cpp"
      $57 = (($storemerge662) + -1)|0; //@line 1861 "load_abc.cpp"
      $58 = (($39) + ($57)|0); //@line 1861 "load_abc.cpp"
      HEAP8[$58>>0] = $56; //@line 1861 "load_abc.cpp"
      $59 = (($storemerge662) + 1)|0; //@line 1861 "load_abc.cpp"
      $exitcond80 = ($59|0)==($j$4|0); //@line 1861 "load_abc.cpp"
      if ($exitcond80) {
       break;
      } else {
       $storemerge662 = $59;
      }
     }
    }
    $60 = (($j$4) + -1)|0; //@line 1862 "load_abc.cpp"
    $61 = (($60) - ($n$0$lcssa117))|0; //@line 1868 "load_abc.cpp"
    $62 = (($61) + 1)|0; //@line 1868 "load_abc.cpp"
    $$sum = (($i$3) + 1)|0; //@line 1869 "load_abc.cpp"
    $63 = (($p) + ($$sum)|0); //@line 1869 "load_abc.cpp"
    $64 = HEAP8[$63>>0]|0; //@line 1625 "load_abc.cpp"
    $65 = $64 << 24 >> 24; //@line 1625 "load_abc.cpp"
    $isdigittmp1$i10 = (($65) + -48)|0; //@line 1625 "load_abc.cpp"
    $isdigit2$i11 = ($isdigittmp1$i10>>>0)<(10); //@line 1625 "load_abc.cpp"
    if ($isdigit2$i11) {
     $68 = $65;$h$04$i12 = 0;$i$03$i13 = 0;
     while(1) {
      $66 = ($h$04$i12*10)|0; //@line 1626 "load_abc.cpp"
      $67 = (($68) + -48)|0; //@line 1626 "load_abc.cpp"
      $69 = (($67) + ($66))|0; //@line 1626 "load_abc.cpp"
      $70 = (($i$03$i13) + 1)|0; //@line 1627 "load_abc.cpp"
      $$sum38 = (($70) + ($$sum))|0; //@line 1625 "load_abc.cpp"
      $71 = (($p) + ($$sum38)|0); //@line 1625 "load_abc.cpp"
      $72 = HEAP8[$71>>0]|0; //@line 1625 "load_abc.cpp"
      $73 = $72 << 24 >> 24; //@line 1625 "load_abc.cpp"
      $isdigittmp$i14 = (($73) + -48)|0; //@line 1625 "load_abc.cpp"
      $isdigit$i15 = ($isdigittmp$i14>>>0)<(10); //@line 1625 "load_abc.cpp"
      if ($isdigit$i15) {
       $68 = $73;$h$04$i12 = $69;$i$03$i13 = $70;
      } else {
       $$lcssa118 = $69;$$lcssa119 = $70;
       break;
      }
     }
     $74 = ($$lcssa119|0)!=(0); //@line 1629 "load_abc.cpp"
     $75 = (($$lcssa119) + ($i$3))|0; //@line 1869 "load_abc.cpp"
     $76 = ($$lcssa118|0)>(1); //@line 1870 "load_abc.cpp"
     $or$cond94 = $74 & $76;
     $77 = ($61|0)>(-1); //@line 1871 "load_abc.cpp"
     $or$cond95 = $or$cond94 & $77;
     if ($or$cond95) {
      $82 = (($j$4) - ($n$0$lcssa117))|0; //@line 1870 "load_abc.cpp"
      $83 = ($82|0)>(1);
      $smax = $83 ? $82 : 1;
      $$in = $$lcssa118;$j$568$us = $60;
      while(1) {
       $93 = (($$in) + -1)|0; //@line 1870 "load_abc.cpp"
       $j$664$us = $j$568$us;$m$065$us = 0;
       while(1) {
        $84 = (($j$664$us) - ($62))|0; //@line 1872 "load_abc.cpp"
        $85 = (($39) + ($84)|0); //@line 1872 "load_abc.cpp"
        $86 = HEAP8[$85>>0]|0; //@line 1872 "load_abc.cpp"
        $87 = (($39) + ($j$664$us)|0); //@line 1872 "load_abc.cpp"
        HEAP8[$87>>0] = $86; //@line 1872 "load_abc.cpp"
        $88 = (($j$664$us) + 1)|0; //@line 1873 "load_abc.cpp"
        $89 = (($m$065$us) + 1)|0; //@line 1871 "load_abc.cpp"
        $90 = ($89|0)<($62|0); //@line 1871 "load_abc.cpp"
        if ($90) {
         $j$664$us = $88;$m$065$us = $89;
        } else {
         break;
        }
       }
       $91 = (($smax) + ($j$568$us))|0; //@line 1871 "load_abc.cpp"
       $92 = ($93|0)>(1); //@line 1870 "load_abc.cpp"
       if ($92) {
        $$in = $93;$j$568$us = $91;
       } else {
        $i$4 = $75;$j$8 = $91;
        break;
       }
      }
     } else {
      $i$4 = $75;$j$8 = $60;
     }
    } else {
     $i$4 = $i$3;$j$8 = $60;
    }
   }
  } while(0);
  $118 = (($i$4) + 1)|0; //@line 1855 "load_abc.cpp"
  $i$3 = $118;$j$4 = $j$8;
 }
 if ((label|0) == 26) {
  $78 = (_strlen($p)|0); //@line 323 "load_abc.cpp"
  $79 = (($78) + 52)|0; //@line 323 "load_abc.cpp"
  $80 = ($79>>>0)>(255); //@line 323 "load_abc.cpp"
  if (!($80)) {
   HEAP32[$vararg_buffer4>>2] = $p; //@line 324 "load_abc.cpp"
   (_sprintf($txt$i,525245,$vararg_buffer4)|0); //@line 324 "load_abc.cpp"
   $81 = HEAP32[37544]|0; //@line 328 "load_abc.cpp"
   HEAP32[$vararg_buffer7>>2] = $txt$i; //@line 328 "load_abc.cpp"
   (_fprintf($81,524747,$vararg_buffer7)|0); //@line 328 "load_abc.cpp"
  }
  $j$4122 = $j$4$lcssa121;
 }
 $119 = (($39) + ($j$4122)|0); //@line 1891 "load_abc.cpp"
 HEAP8[$119>>0] = 0; //@line 1891 "load_abc.cpp"
 $120 = ($j$4122|0)>(0); //@line 1893 "load_abc.cpp"
 if ($120) {
  $121 = HEAP32[37544]|0; //@line 328 "load_abc.cpp"
  $i$556 = 0;$j$955 = $j$4122;
  while(1) {
   $122 = (($39) + ($i$556)|0); //@line 1894 "load_abc.cpp"
   $123 = HEAP8[$122>>0]|0; //@line 1894 "load_abc.cpp"
   $124 = ($123<<24>>24)==(40); //@line 1894 "load_abc.cpp"
   if ($124) {
    $125 = (_strlen($p)|0); //@line 323 "load_abc.cpp"
    $126 = (($125) + 51)|0; //@line 323 "load_abc.cpp"
    $127 = ($126>>>0)>(255); //@line 323 "load_abc.cpp"
    if (!($127)) {
     HEAP32[$vararg_buffer10>>2] = $p; //@line 324 "load_abc.cpp"
     (_sprintf($txt$i,525298,$vararg_buffer10)|0); //@line 324 "load_abc.cpp"
     HEAP32[$vararg_buffer13>>2] = $txt$i; //@line 328 "load_abc.cpp"
     (_fprintf($121,524747,$vararg_buffer13)|0); //@line 328 "load_abc.cpp"
    }
    $128 = ($i$556|0)<($j$955|0); //@line 1896 "load_abc.cpp"
    if ($128) {
     $storemerge54 = $i$556;
     while(1) {
      $129 = (($storemerge54) + 1)|0; //@line 1896 "load_abc.cpp"
      $130 = (($39) + ($129)|0); //@line 1896 "load_abc.cpp"
      $131 = HEAP8[$130>>0]|0; //@line 1896 "load_abc.cpp"
      $132 = (($39) + ($storemerge54)|0); //@line 1896 "load_abc.cpp"
      HEAP8[$132>>0] = $131; //@line 1896 "load_abc.cpp"
      $exitcond = ($129|0)==($j$955|0); //@line 1896 "load_abc.cpp"
      if ($exitcond) {
       break;
      } else {
       $storemerge54 = $129;
      }
     }
    }
    $133 = (($j$955) + -1)|0; //@line 1897 "load_abc.cpp"
    $j$10 = $133;
   } else {
    $j$10 = $j$955;
   }
   $134 = (($i$556) + 1)|0; //@line 1893 "load_abc.cpp"
   $135 = ($134|0)<($j$10|0); //@line 1893 "load_abc.cpp"
   if ($135) {
    $i$556 = $134;$j$955 = $j$10;
   } else {
    break;
   }
  }
 }
 HEAP32[$d>>2] = $39; //@line 1900 "load_abc.cpp"
 STACKTOP = sp;return; //@line 1901 "load_abc.cpp"
}
function __ZL17abc_extract_tempoPKci($p,$invoice) {
 $p = $p|0;
 $invoice = $invoice|0;
 var $$h$0$i = 0, $$h$0$i12 = 0, $$in$0 = 0, $$lcssa = 0, $$lcssa19 = 0, $$lcssa1954 = 0, $$lcssa77 = 0, $$lcssa78 = 0, $$lcssa79 = 0, $$lcssa80 = 0, $$lcssa81 = 0, $$lcssa82 = 0, $$nd$0 = 0, $$nd$057 = 0, $$nd$058 = 0, $$nl$0 = 0, $$nl$055 = 0, $$nl$056 = 0, $$off = 0, $$tempo$0 = 0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $8 = 0, $9 = 0, $h$04$i = 0, $h$04$i5 = 0, $i$03$i = 0, $i$03$i$lcssa = 0, $i$03$i6 = 0, $i$03$i6$lcssa = 0;
 var $in$023 = 0, $in$1 = 0, $isdigit = 0, $isdigit$i = 0, $isdigit$i8 = 0, $isdigittmp = 0, $isdigittmp$i = 0, $isdigittmp$i7 = 0, $nd$0$lcssa = 0, $nd$022 = 0, $nd$1 = 0, $nd$3 = 0, $nl$0$lcssa = 0, $nl$021 = 0, $nl$1 = 0, $nl$3 = 0, $nl1$024 = 0, $nl1$1 = 0, $notes$0$lcssa = 0, $notes$025 = 0;
 var $notes$1 = 0, $or$cond = 0, $q$027 = 0, $q$1 = 0, $state$026 = 0, $state$1 = 0, $tempo$0 = 0, $tempo$2 = 0, $tempo$3 = 0, $txt$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 272|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $txt$i = sp + 16|0;
 $0 = HEAP8[$p>>0]|0; //@line 1738 "load_abc.cpp"
 $1 = ($0<<24>>24)==(0); //@line 1738 "load_abc.cpp"
 if ($1) {
  $$lcssa1954 = 120;$$nd$057 = 4;$$nl$055 = 1;
  label = 16;
 } else {
  $4 = $0;$72 = 120;$in$023 = 0;$nd$022 = 1;$nl$021 = 0;$nl1$024 = 0;$notes$025 = 0;$q$027 = $p;$state$026 = 0;
  L2: while(1) {
   $2 = ($in$023|0)==(0); //@line 1739 "load_abc.cpp"
   $3 = $4 << 24 >> 24; //@line 1740 "load_abc.cpp"
   L4: do {
    if ($2) {
     $6 = ($4<<24>>24)==(93); //@line 1744 "load_abc.cpp"
     if ($6) {
      $$lcssa19 = $72;$nd$0$lcssa = $nd$022;$nl$0$lcssa = $nl$021;$notes$0$lcssa = $notes$025;
      break L2;
     }
     switch ($3|0) {
     case 61:  {
      $73 = $72;$in$1 = 0;$nd$1 = $nd$022;$nl$1 = $nl$021;$nl1$1 = $nl1$024;$notes$1 = $notes$025;$q$1 = $q$027;$state$1 = $state$026;
      break L4;
      break;
     }
     case 34:  {
      $73 = $72;$in$1 = 1;$nd$1 = $nd$022;$nl$1 = $nl$021;$nl1$1 = $nl1$024;$notes$1 = $notes$025;$q$1 = $q$027;$state$1 = $state$026;
      break L4;
      break;
     }
     case 47:  {
      $7 = (($notes$025) + 1)|0; //@line 1750 "load_abc.cpp"
      $73 = $72;$in$1 = 0;$nd$1 = $nd$022;$nl$1 = $nl$021;$nl1$1 = $72;$notes$1 = $7;$q$1 = $q$027;$state$1 = 1;
      break L4;
      break;
     }
     default: {
      $isdigittmp = (($3) + -48)|0; //@line 1757 "load_abc.cpp"
      $isdigit = ($isdigittmp>>>0)<(10); //@line 1757 "load_abc.cpp"
      if (!($isdigit)) {
       $73 = $72;$in$1 = 0;$nd$1 = $nd$022;$nl$1 = $nl$021;$nl1$1 = $nl1$024;$notes$1 = $notes$025;$q$1 = $q$027;$state$1 = $state$026;
       break L4;
      }
      $8 = ($state$026|0)==(0); //@line 1758 "load_abc.cpp"
      if ($8) {
       $25 = $3;$h$04$i5 = 0;$i$03$i6 = 0;
       while(1) {
        $23 = ($h$04$i5*10)|0; //@line 1626 "load_abc.cpp"
        $24 = (($25) + -48)|0; //@line 1626 "load_abc.cpp"
        $26 = (($24) + ($23))|0; //@line 1626 "load_abc.cpp"
        $27 = (($i$03$i6) + 1)|0; //@line 1627 "load_abc.cpp"
        $28 = (($q$027) + ($27)|0); //@line 1625 "load_abc.cpp"
        $29 = HEAP8[$28>>0]|0; //@line 1625 "load_abc.cpp"
        $30 = $29 << 24 >> 24; //@line 1625 "load_abc.cpp"
        $isdigittmp$i7 = (($30) + -48)|0; //@line 1625 "load_abc.cpp"
        $isdigit$i8 = ($isdigittmp$i7>>>0)<(10); //@line 1625 "load_abc.cpp"
        if ($isdigit$i8) {
         $25 = $30;$h$04$i5 = $26;$i$03$i6 = $27;
        } else {
         $$lcssa81 = $26;$$lcssa82 = $27;$i$03$i6$lcssa = $i$03$i6;
         break;
        }
       }
       $31 = ($$lcssa82|0)==(0); //@line 1629 "load_abc.cpp"
       $$h$0$i12 = $31 ? 1 : $$lcssa81; //@line 1630 "load_abc.cpp"
       $32 = (($q$027) + ($i$03$i6$lcssa)|0); //@line 1765 "load_abc.cpp"
       $73 = $$h$0$i12;$in$1 = 0;$nd$1 = $nd$022;$nl$1 = $nl$021;$nl1$1 = $nl1$024;$notes$1 = $notes$025;$q$1 = $32;$state$1 = 0;
       break L4;
      } else {
       $11 = $3;$h$04$i = 0;$i$03$i = 0;
       while(1) {
        $9 = ($h$04$i*10)|0; //@line 1626 "load_abc.cpp"
        $10 = (($11) + -48)|0; //@line 1626 "load_abc.cpp"
        $12 = (($10) + ($9))|0; //@line 1626 "load_abc.cpp"
        $13 = (($i$03$i) + 1)|0; //@line 1627 "load_abc.cpp"
        $14 = (($q$027) + ($13)|0); //@line 1625 "load_abc.cpp"
        $15 = HEAP8[$14>>0]|0; //@line 1625 "load_abc.cpp"
        $16 = $15 << 24 >> 24; //@line 1625 "load_abc.cpp"
        $isdigittmp$i = (($16) + -48)|0; //@line 1625 "load_abc.cpp"
        $isdigit$i = ($isdigittmp$i>>>0)<(10); //@line 1625 "load_abc.cpp"
        if ($isdigit$i) {
         $11 = $16;$h$04$i = $12;$i$03$i = $13;
        } else {
         $$lcssa79 = $12;$$lcssa80 = $13;$i$03$i$lcssa = $i$03$i;
         break;
        }
       }
       $17 = ($$lcssa80|0)==(0); //@line 1629 "load_abc.cpp"
       $$h$0$i = $17 ? 1 : $$lcssa79; //@line 1630 "load_abc.cpp"
       $18 = (($q$027) + ($i$03$i$lcssa)|0); //@line 1759 "load_abc.cpp"
       $19 = Math_imul($$h$0$i, $nl$021)|0; //@line 1761 "load_abc.cpp"
       $20 = Math_imul($nl1$024, $nd$022)|0; //@line 1761 "load_abc.cpp"
       $21 = (($19) + ($20))|0; //@line 1761 "load_abc.cpp"
       $22 = Math_imul($$h$0$i, $nd$022)|0; //@line 1762 "load_abc.cpp"
       $73 = $72;$in$1 = 0;$nd$1 = $22;$nl$1 = $21;$nl1$1 = $nl1$024;$notes$1 = $notes$025;$q$1 = $18;$state$1 = 0;
       break L4;
      }
     }
     }
    } else {
     $5 = ($4<<24>>24)==(34); //@line 1740 "load_abc.cpp"
     $$in$0 = $5 ? 0 : $in$023; //@line 1740 "load_abc.cpp"
     $73 = $72;$in$1 = $$in$0;$nd$1 = $nd$022;$nl$1 = $nl$021;$nl1$1 = $nl1$024;$notes$1 = $notes$025;$q$1 = $q$027;$state$1 = $state$026;
    }
   } while(0);
   $33 = ((($q$1)) + 1|0); //@line 1738 "load_abc.cpp"
   $34 = HEAP8[$33>>0]|0; //@line 1738 "load_abc.cpp"
   $35 = ($34<<24>>24)==(0); //@line 1738 "load_abc.cpp"
   if ($35) {
    $$lcssa19 = $73;$nd$0$lcssa = $nd$1;$nl$0$lcssa = $nl$1;$notes$0$lcssa = $notes$1;
    break;
   } else {
    $4 = $34;$72 = $73;$in$023 = $in$1;$nd$022 = $nd$1;$nl$021 = $nl$1;$nl1$024 = $nl1$1;$notes$025 = $notes$1;$q$027 = $33;$state$026 = $state$1;
   }
  }
  $36 = ($notes$0$lcssa|0)==(0); //@line 1771 "load_abc.cpp"
  $$nl$0 = $36 ? 1 : $nl$0$lcssa; //@line 1771 "load_abc.cpp"
  $$nd$0 = $36 ? 4 : $nd$0$lcssa; //@line 1771 "load_abc.cpp"
  $37 = ($$nd$0|0)==(0); //@line 1775 "load_abc.cpp"
  if ($37) {
   $$nd$058 = 0;$$nl$056 = $$nl$0;$tempo$0 = 120;
  } else {
   $$lcssa1954 = $$lcssa19;$$nd$057 = $$nd$0;$$nl$055 = $$nl$0;
   label = 16;
  }
 }
 if ((label|0) == 16) {
  $38 = $$lcssa1954 << 2; //@line 1776 "load_abc.cpp"
  $39 = Math_imul($38, $$nl$055)|0; //@line 1776 "load_abc.cpp"
  $40 = (($39|0) / ($$nd$057|0))&-1; //@line 1776 "load_abc.cpp"
  $$nd$058 = $$nd$057;$$nl$056 = $$nl$055;$tempo$0 = $40;
 }
 $41 = ($tempo$0|0)<(1); //@line 1777 "load_abc.cpp"
 $$tempo$0 = $41 ? 120 : $tempo$0; //@line 1777 "load_abc.cpp"
 $42 = ($invoice|0)!=(0); //@line 1779 "load_abc.cpp"
 $43 = HEAP32[299862]|0; //@line 1780 "load_abc.cpp"
 $44 = HEAP32[299863]|0; //@line 1781 "load_abc.cpp"
 $nl$3 = $42 ? $43 : $$nl$056; //@line 1779 "load_abc.cpp"
 HEAP32[299862] = 1; //@line 1783 "load_abc.cpp"
 $46 = 1;
 while(1) {
  $45 = (($$tempo$0|0) / ($46|0))&-1; //@line 1785 "load_abc.cpp"
  $47 = ($45|0)>(255); //@line 1785 "load_abc.cpp"
  $48 = (($46) + 1)|0; //@line 1786 "load_abc.cpp"
  if ($47) {
   $46 = $48;
  } else {
   $$lcssa77 = $46;$$lcssa78 = $45;
   break;
  }
 }
 $nd$3 = $42 ? $44 : $$nd$058; //@line 1779 "load_abc.cpp"
 HEAP32[299863] = $$lcssa77; //@line 1786 "load_abc.cpp"
 $50 = 1;
 while(1) {
  $49 = (($50) + 1)|0; //@line 1789 "load_abc.cpp"
  $51 = Math_imul($49, $$lcssa78)|0; //@line 1788 "load_abc.cpp"
  $52 = ($51|0)<(256); //@line 1788 "load_abc.cpp"
  if ($52) {
   $50 = $49;
  } else {
   $$lcssa = $50;
   break;
  }
 }
 HEAP32[299862] = $$lcssa; //@line 1790 "load_abc.cpp"
 $53 = Math_imul($$lcssa, $$lcssa78)|0; //@line 1791 "load_abc.cpp"
 $54 = ($53*3)|0; //@line 1792 "load_abc.cpp"
 $55 = ($54|0)<(512); //@line 1792 "load_abc.cpp"
 if ($55) {
  $56 = ($$lcssa*3)|0; //@line 1793 "load_abc.cpp"
  HEAP32[299862] = $56; //@line 1793 "load_abc.cpp"
  $57 = $$lcssa77 << 1; //@line 1794 "load_abc.cpp"
  HEAP32[299863] = $57; //@line 1794 "load_abc.cpp"
  $58 = (($54|0) / 2)&-1; //@line 1795 "load_abc.cpp"
  $60 = $56;$62 = $57;$tempo$2 = $58;
 } else {
  $60 = $$lcssa;$62 = $$lcssa77;$tempo$2 = $53;
 }
 if (!($42)) {
  $tempo$3 = $tempo$2;
  STACKTOP = sp;return ($tempo$3|0); //@line 1809 "load_abc.cpp"
 }
 $59 = ($nl$3|0)==($60|0); //@line 1798 "load_abc.cpp"
 $61 = ($nd$3|0)==($62|0); //@line 1798 "load_abc.cpp"
 $or$cond = $59 & $61; //@line 1798 "load_abc.cpp"
 if ($or$cond) {
  $tempo$3 = $tempo$2;
  STACKTOP = sp;return ($tempo$3|0); //@line 1809 "load_abc.cpp"
 }
 $63 = Math_imul($tempo$2, $nl$3)|0; //@line 1799 "load_abc.cpp"
 $64 = Math_imul($63, $62)|0; //@line 1799 "load_abc.cpp"
 $65 = Math_imul($60, $nd$3)|0; //@line 1799 "load_abc.cpp"
 $66 = (($64|0) / ($65|0))&-1; //@line 1799 "load_abc.cpp"
 $$off = (($66) + -32)|0; //@line 1800 "load_abc.cpp"
 $67 = ($$off>>>0)<(224); //@line 1800 "load_abc.cpp"
 if ($67) {
  HEAP32[299862] = $nl$3; //@line 1802 "load_abc.cpp"
  HEAP32[299863] = $nd$3; //@line 1803 "load_abc.cpp"
  $tempo$3 = $66;
  STACKTOP = sp;return ($tempo$3|0); //@line 1809 "load_abc.cpp"
 }
 $68 = (_strlen($p)|0); //@line 323 "load_abc.cpp"
 $69 = (($68) + 58)|0; //@line 323 "load_abc.cpp"
 $70 = ($69>>>0)>(255); //@line 323 "load_abc.cpp"
 if (!($70)) {
  HEAP32[$vararg_buffer>>2] = $p; //@line 324 "load_abc.cpp"
  (_sprintf($txt$i,525353,$vararg_buffer)|0); //@line 324 "load_abc.cpp"
  $71 = HEAP32[37544]|0; //@line 328 "load_abc.cpp"
  HEAP32[$vararg_buffer1>>2] = $txt$i; //@line 328 "load_abc.cpp"
  (_fprintf($71,524747,$vararg_buffer1)|0); //@line 328 "load_abc.cpp"
 }
 $tempo$3 = $tempo$2;
 STACKTOP = sp;return ($tempo$3|0); //@line 1809 "load_abc.cpp"
}
function __ZL7ABC_KeyPKc($p) {
 $p = $p|0;
 var $$$i$6 = 0, $$0 = 0, $$02 = 0, $$02$lcssa = 0, $$19 = 0, $$2 = 0, $$2$lcssa = 0, $$3 = 0, $$i$0 = 0, $$i$2 = 0, $$i$4 = 0, $$i$6 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $9 = 0, $c = 0, $i$0$lcssa = 0, $i$010 = 0, $i$2 = 0, $i$6 = 0, $i$97 = 0;
 var $txt$i = 0, $uglygep = 0, $uglygep14 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 288|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 16|0;
 $vararg_buffer = sp + 8|0;
 $txt$i = sp + 24|0;
 $c = sp; //@line 2878 "load_abc.cpp"
 $0 = $c; //@line 2878 "load_abc.cpp"
 $1 = $0; //@line 2878 "load_abc.cpp"
 HEAP32[$1>>2] = 0; //@line 2878 "load_abc.cpp"
 $2 = (($0) + 4)|0; //@line 2878 "load_abc.cpp"
 $3 = $2; //@line 2878 "load_abc.cpp"
 HEAP32[$3>>2] = 0; //@line 2878 "load_abc.cpp"
 $$02 = $p;
 while(1) {
  $4 = HEAP8[$$02>>0]|0; //@line 2880 "load_abc.cpp"
  $5 = $4 << 24 >> 24; //@line 2880 "load_abc.cpp"
  $6 = (_isspace($5)|0); //@line 2880 "load_abc.cpp"
  $7 = ($6|0)==(0); //@line 2880 "load_abc.cpp"
  $8 = ((($$02)) + 1|0); //@line 2880 "load_abc.cpp"
  if ($7) {
   $$02$lcssa = $$02;
   break;
  } else {
   $$02 = $8;
  }
 }
 $$19 = $$02$lcssa;$i$010 = 0;
 L4: while(1) {
  $9 = HEAP8[$$19>>0]|0; //@line 2883 "load_abc.cpp"
  switch ($9<<24>>24) {
  case 93: case 0:  {
   $i$0$lcssa = $i$010;
   break L4;
   break;
  }
  default: {
  }
  }
  $10 = $9 << 24 >> 24; //@line 2883 "load_abc.cpp"
  $11 = (_isspace($10)|0); //@line 2884 "load_abc.cpp"
  $12 = ($11|0)==(0); //@line 2884 "load_abc.cpp"
  if ($12) {
   $$3 = $$19;
  } else {
   $$2 = $$19;
   while(1) {
    $13 = HEAP8[$$2>>0]|0; //@line 2885 "load_abc.cpp"
    $14 = $13 << 24 >> 24; //@line 2885 "load_abc.cpp"
    $15 = (_isspace($14)|0); //@line 2885 "load_abc.cpp"
    $16 = ($15|0)==(0); //@line 2885 "load_abc.cpp"
    $17 = ((($$2)) + 1|0); //@line 2885 "load_abc.cpp"
    if ($16) {
     $$2$lcssa = $$2;
     break;
    } else {
     $$2 = $17;
    }
   }
   $18 = (_strncasecmp($$2$lcssa,525433,3)|0); //@line 2886 "load_abc.cpp"
   $19 = ($18|0)==(0); //@line 2886 "load_abc.cpp"
   if ($19) {
    $$3 = $$2$lcssa;
   } else {
    $20 = (_strncasecmp($$2$lcssa,525437,3)|0); //@line 2886 "load_abc.cpp"
    $21 = ($20|0)==(0); //@line 2886 "load_abc.cpp"
    if ($21) {
     $$3 = $$2$lcssa;
    } else {
     $i$0$lcssa = $i$010;
     break;
    }
   }
  }
  $22 = HEAP8[$$3>>0]|0; //@line 2889 "load_abc.cpp"
  $23 = (($c) + ($i$010)|0); //@line 2889 "load_abc.cpp"
  HEAP8[$23>>0] = $22; //@line 2889 "load_abc.cpp"
  $24 = (($i$010) + 1)|0; //@line 2890 "load_abc.cpp"
  $25 = ((($$3)) + 1|0); //@line 2883 "load_abc.cpp"
  $26 = ($24|0)<(8); //@line 2883 "load_abc.cpp"
  if ($26) {
   $$19 = $25;$i$010 = $24;
  } else {
   $i$0$lcssa = $24;
   break;
  }
 }
 $27 = (($c) + ($i$0$lcssa)|0); //@line 2892 "load_abc.cpp"
 HEAP8[$27>>0] = 0; //@line 2892 "load_abc.cpp"
 $28 = (_strcmp($c,525441)|0); //@line 2893 "load_abc.cpp"
 $29 = ($28|0)==(0); //@line 2893 "load_abc.cpp"
 if ($29) {
  label = 12;
 } else {
  $30 = (_strcmp($c,525444)|0); //@line 2893 "load_abc.cpp"
  $31 = ($30|0)==(0); //@line 2893 "load_abc.cpp"
  if ($31) {
   label = 12;
  }
 }
 if ((label|0) == 12) {
  ;HEAP8[$c>>0]=HEAP8[525447>>0]|0;HEAP8[$c+1>>0]=HEAP8[525447+1>>0]|0;HEAP8[$c+2>>0]=HEAP8[525447+2>>0]|0; //@line 2894 "load_abc.cpp"
 }
 $32 = ((($c)) + 1|0); //@line 2895 "load_abc.cpp"
 $33 = (_strcasecmp($32,525450)|0); //@line 2895 "load_abc.cpp"
 $34 = ($33|0)==(0); //@line 2895 "load_abc.cpp"
 $$i$0 = $34 ? 2 : $i$0$lcssa; //@line 2895 "load_abc.cpp"
 $35 = ((($c)) + 2|0); //@line 2896 "load_abc.cpp"
 $36 = (_strcasecmp($35,525450)|0); //@line 2896 "load_abc.cpp"
 $37 = ($36|0)==(0); //@line 2896 "load_abc.cpp"
 $i$2 = $37 ? 3 : $$i$0; //@line 2896 "load_abc.cpp"
 $38 = (_strcasecmp($32,525456)|0); //@line 2897 "load_abc.cpp"
 $39 = ($38|0)==(0); //@line 2897 "load_abc.cpp"
 $$i$2 = $39 ? 1 : $i$2; //@line 2897 "load_abc.cpp"
 $40 = (_strcasecmp($35,525456)|0); //@line 2898 "load_abc.cpp"
 $41 = ($40|0)==(0); //@line 2898 "load_abc.cpp"
 $42 = (_strcasecmp($32,525433)|0); //@line 2899 "load_abc.cpp"
 $43 = ($42|0)==(0); //@line 2899 "load_abc.cpp"
 $44 = $41 | $43; //@line 2899 "load_abc.cpp"
 $$i$4 = $44 ? 2 : $$i$2; //@line 2899 "load_abc.cpp"
 $45 = (_strcasecmp($35,525433)|0); //@line 2900 "load_abc.cpp"
 $46 = ($45|0)==(0); //@line 2900 "load_abc.cpp"
 $i$6 = $46 ? 3 : $$i$4; //@line 2900 "load_abc.cpp"
 $47 = (_strcasecmp($32,525437)|0); //@line 2901 "load_abc.cpp"
 $48 = ($47|0)==(0); //@line 2901 "load_abc.cpp"
 $$i$6 = $48 ? 1 : $i$6; //@line 2901 "load_abc.cpp"
 $49 = (_strcasecmp($35,525437)|0); //@line 2902 "load_abc.cpp"
 $50 = ($49|0)==(0); //@line 2902 "load_abc.cpp"
 $$$i$6 = $50 ? 2 : $$i$6;
 $51 = ($$$i$6|0)<(6); //@line 2903 "load_abc.cpp"
 $52 = (($c) + ($$$i$6)|0); //@line 2904 "load_abc.cpp"
 if ($51) {
  $uglygep14 = (($c) + ($$$i$6)|0);
  $53 = (6 - ($$$i$6))|0; //@line 2903 "load_abc.cpp"
  _memset(($uglygep14|0),32,($53|0))|0; //@line 2904 "load_abc.cpp"
  $uglygep = ((($c)) + 6|0);
  $$lcssa = $uglygep;
 } else {
  $$lcssa = $52;
 }
 HEAP8[$$lcssa>>0] = 0; //@line 2905 "load_abc.cpp"
 $55 = 47048;$i$97 = 0;
 while(1) {
  $54 = HEAP32[$55>>2]|0; //@line 2908 "load_abc.cpp"
  $56 = ((($54)) + 10|0); //@line 2908 "load_abc.cpp"
  $57 = (_strncasecmp($56,$c,6)|0); //@line 2908 "load_abc.cpp"
  $58 = ($57|0)==(0); //@line 2908 "load_abc.cpp"
  if ($58) {
   $$0 = $i$97;
   label = 21;
   break;
  }
  $59 = ((($54)) + 16|0); //@line 2908 "load_abc.cpp"
  $60 = (_strncasecmp($59,$c,6)|0); //@line 2908 "load_abc.cpp"
  $61 = ($60|0)==(0); //@line 2908 "load_abc.cpp"
  if ($61) {
   $$0 = $i$97;
   label = 21;
   break;
  }
  $66 = ((($54)) + 22|0); //@line 2908 "load_abc.cpp"
  $67 = (_strncasecmp($66,$c,6)|0); //@line 2908 "load_abc.cpp"
  $68 = ($67|0)==(0); //@line 2908 "load_abc.cpp"
  if ($68) {
   $$0 = $i$97;
   label = 21;
   break;
  }
  $69 = ((($54)) + 28|0); //@line 2908 "load_abc.cpp"
  $70 = (_strncasecmp($69,$c,6)|0); //@line 2908 "load_abc.cpp"
  $71 = ($70|0)==(0); //@line 2908 "load_abc.cpp"
  if ($71) {
   $$0 = $i$97;
   label = 21;
   break;
  }
  $72 = ((($54)) + 34|0); //@line 2908 "load_abc.cpp"
  $73 = (_strncasecmp($72,$c,6)|0); //@line 2908 "load_abc.cpp"
  $74 = ($73|0)==(0); //@line 2908 "load_abc.cpp"
  if ($74) {
   $$0 = $i$97;
   label = 21;
   break;
  }
  $75 = ((($54)) + 40|0); //@line 2908 "load_abc.cpp"
  $76 = (_strncasecmp($75,$c,6)|0); //@line 2908 "load_abc.cpp"
  $77 = ($76|0)==(0); //@line 2908 "load_abc.cpp"
  if ($77) {
   $$0 = $i$97;
   label = 21;
   break;
  }
  $78 = (($i$97) + 1)|0; //@line 2906 "load_abc.cpp"
  $79 = (47048 + ($78<<2)|0); //@line 2906 "load_abc.cpp"
  $80 = ($78|0)==(15); //@line 2906 "load_abc.cpp"
  if ($80) {
   label = 18;
   break;
  } else {
   $55 = $79;$i$97 = $78;
  }
 }
 if ((label|0) == 18) {
  $62 = (_strlen($$02$lcssa)|0); //@line 323 "load_abc.cpp"
  $63 = (($62) + 33)|0; //@line 323 "load_abc.cpp"
  $64 = ($63>>>0)>(255); //@line 323 "load_abc.cpp"
  if (!($64)) {
   HEAP32[$vararg_buffer>>2] = $$02$lcssa; //@line 324 "load_abc.cpp"
   (_sprintf($txt$i,526257,$vararg_buffer)|0); //@line 324 "load_abc.cpp"
   $65 = HEAP32[37544]|0; //@line 328 "load_abc.cpp"
   HEAP32[$vararg_buffer1>>2] = $txt$i; //@line 328 "load_abc.cpp"
   (_fprintf($65,524747,$vararg_buffer1)|0); //@line 328 "load_abc.cpp"
  }
  $$0 = 7;
  STACKTOP = sp;return ($$0|0); //@line 2913 "load_abc.cpp"
 }
 else if ((label|0) == 21) {
  STACKTOP = sp;return ($$0|0); //@line 2913 "load_abc.cpp"
 }
 return (0)|0;
}
function __ZL15abc_MIDI_gchordPKcP10_ABCHANDLE($p,$h) {
 $p = $p|0;
 $h = $h|0;
 var $$0 = 0, $$01 = 0, $$01$lcssa = 0, $$14 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $isdigit = 0, $isdigit3 = 0, $isdigittmp = 0, $isdigittmp2 = 0, $memchr = 0, $q$0$lcssa = 0, $q$05 = 0, $q$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$01 = $p;
 while(1) {
  $0 = HEAP8[$$01>>0]|0; //@line 3255 "load_abc.cpp"
  $1 = $0 << 24 >> 24; //@line 3255 "load_abc.cpp"
  $2 = (_isspace($1)|0); //@line 3255 "load_abc.cpp"
  $3 = ($2|0)==(0); //@line 3255 "load_abc.cpp"
  $4 = ((($$01)) + 1|0); //@line 3255 "load_abc.cpp"
  if ($3) {
   $$01$lcssa = $$01;
   break;
  } else {
   $$01 = $4;
  }
 }
 $5 = (_strncmp($$01$lcssa,526513,2)|0); //@line 3256 "load_abc.cpp"
 $6 = ($5|0)==(0); //@line 3256 "load_abc.cpp"
 if ($6) {
  $7 = ((($$01$lcssa)) + 2|0); //@line 3256 "load_abc.cpp"
  $8 = HEAP8[$7>>0]|0; //@line 3256 "load_abc.cpp"
  $9 = $8 << 24 >> 24; //@line 3256 "load_abc.cpp"
  $10 = (_isspace($9)|0); //@line 3256 "load_abc.cpp"
  $11 = ($10|0)==(0); //@line 3256 "load_abc.cpp"
  if (!($11)) {
   $$0 = 2;
   return ($$0|0); //@line 3265 "load_abc.cpp"
  }
  $12 = HEAP8[$7>>0]|0; //@line 3256 "load_abc.cpp"
  $13 = ($12<<24>>24)==(0); //@line 3256 "load_abc.cpp"
  if ($13) {
   $$0 = 2;
   return ($$0|0); //@line 3265 "load_abc.cpp"
  }
 }
 $14 = (_strncmp($$01$lcssa,526516,3)|0); //@line 3257 "load_abc.cpp"
 $15 = ($14|0)==(0); //@line 3257 "load_abc.cpp"
 if ($15) {
  $16 = ((($$01$lcssa)) + 3|0); //@line 3257 "load_abc.cpp"
  $17 = HEAP8[$16>>0]|0; //@line 3257 "load_abc.cpp"
  $18 = $17 << 24 >> 24; //@line 3257 "load_abc.cpp"
  $19 = (_isspace($18)|0); //@line 3257 "load_abc.cpp"
  $20 = ($19|0)==(0); //@line 3257 "load_abc.cpp"
  if (!($20)) {
   $$0 = 1;
   return ($$0|0); //@line 3265 "load_abc.cpp"
  }
  $21 = HEAP8[$16>>0]|0; //@line 3257 "load_abc.cpp"
  $22 = ($21<<24>>24)==(0); //@line 3257 "load_abc.cpp"
  if ($22) {
   $$0 = 1;
   return ($$0|0); //@line 3265 "load_abc.cpp"
  }
 }
 $23 = ((($h)) + 36|0); //@line 3258 "load_abc.cpp"
 $24 = HEAP8[$$01$lcssa>>0]|0; //@line 3258 "load_abc.cpp"
 $25 = ($24<<24>>24)==(0); //@line 3258 "load_abc.cpp"
 L18: do {
  if ($25) {
   $q$0$lcssa = $23;
  } else {
   $$14 = $$01$lcssa;$27 = $24;$q$05 = $23;
   while(1) {
    $26 = $27 << 24 >> 24; //@line 3258 "load_abc.cpp"
    $28 = (_isspace($26)|0); //@line 3258 "load_abc.cpp"
    $29 = ($28|0)==(0); //@line 3258 "load_abc.cpp"
    if (!($29)) {
     $q$0$lcssa = $q$05;
     break L18;
    }
    $30 = HEAP8[$$14>>0]|0; //@line 3259 "load_abc.cpp"
    $31 = $30 << 24 >> 24; //@line 3259 "load_abc.cpp"
    $memchr = (_memchr(526520,$31,23)|0); //@line 3259 "load_abc.cpp"
    $32 = ($memchr|0)==(0|0); //@line 3259 "load_abc.cpp"
    if ($32) {
     $q$0$lcssa = $q$05;
     break L18;
    }
    $33 = ((($q$05)) + 1|0); //@line 3260 "load_abc.cpp"
    HEAP8[$q$05>>0] = $30; //@line 3260 "load_abc.cpp"
    $34 = HEAP8[$$14>>0]|0; //@line 3261 "load_abc.cpp"
    $35 = $34 << 24 >> 24; //@line 3261 "load_abc.cpp"
    $isdigittmp = (($35) + -48)|0; //@line 3261 "load_abc.cpp"
    $isdigit = ($isdigittmp>>>0)<(10); //@line 3261 "load_abc.cpp"
    $$pre = ((($$14)) + 1|0); //@line 3258 "load_abc.cpp"
    if ($isdigit) {
     $q$1 = $33;
    } else {
     $36 = HEAP8[$$pre>>0]|0; //@line 3261 "load_abc.cpp"
     $37 = $36 << 24 >> 24; //@line 3261 "load_abc.cpp"
     $isdigittmp2 = (($37) + -48)|0; //@line 3261 "load_abc.cpp"
     $isdigit3 = ($isdigittmp2>>>0)<(10); //@line 3261 "load_abc.cpp"
     if ($isdigit3) {
      $q$1 = $33;
     } else {
      $38 = ((($q$05)) + 2|0); //@line 3261 "load_abc.cpp"
      HEAP8[$33>>0] = 49; //@line 3261 "load_abc.cpp"
      $q$1 = $38;
     }
    }
    $39 = HEAP8[$$pre>>0]|0; //@line 3258 "load_abc.cpp"
    $40 = ($39<<24>>24)==(0); //@line 3258 "load_abc.cpp"
    if ($40) {
     $q$0$lcssa = $q$1;
     break;
    } else {
     $$14 = $$pre;$27 = $39;$q$05 = $q$1;
    }
   }
  }
 } while(0);
 HEAP8[$q$0$lcssa>>0] = 0; //@line 3263 "load_abc.cpp"
 $$0 = 0;
 return ($$0|0); //@line 3265 "load_abc.cpp"
}
function _pat_gmtosmp($gm) {
 $gm = $gm|0;
 var $$0 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $smp$0 = 0, $smp$0$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $smp$0 = 0;
 while(1) {
  $0 = (1252699 + ($smp$0)|0); //@line 219 "load_pat.cpp"
  $1 = HEAP8[$0>>0]|0; //@line 219 "load_pat.cpp"
  $2 = ($1<<24>>24)==(0); //@line 219 "load_pat.cpp"
  if ($2) {
   $$lcssa = $0;$smp$0$lcssa = $smp$0;
   break;
  }
  $3 = $1&255; //@line 220 "load_pat.cpp"
  $4 = ($3|0)==($gm|0); //@line 220 "load_pat.cpp"
  $5 = (($smp$0) + 1)|0; //@line 221 "load_pat.cpp"
  if ($4) {
   $$0 = $5;
   label = 6;
   break;
  } else {
   $smp$0 = $5;
  }
 }
 if ((label|0) == 6) {
  return ($$0|0); //@line 227 "load_pat.cpp"
 }
 $6 = ($smp$0$lcssa|0)<(191); //@line 222 "load_pat.cpp"
 if (!($6)) {
  $$0 = 1;
  return ($$0|0); //@line 227 "load_pat.cpp"
 }
 $7 = $gm&255; //@line 223 "load_pat.cpp"
 HEAP8[$$lcssa>>0] = $7; //@line 223 "load_pat.cpp"
 $8 = (($smp$0$lcssa) + 1)|0; //@line 224 "load_pat.cpp"
 $$0 = $8;
 return ($$0|0); //@line 227 "load_pat.cpp"
}
function __ZL17abc_song_to_partsP10_ABCHANDLEPPcPA2_h($h,$abcparts,$partp) {
 $h = $h|0;
 $abcparts = $abcparts|0;
 $partp = $partp|0;
 var $$$02$i$i = 0, $$$02$i$i$i = 0, $$$i$i = 0, $$$i$i$i = 0, $$0$i = 0, $$0$i$i = 0, $$0$i$i$i = 0, $$0$i$i12 = 0, $$0$i$i21 = 0, $$01$i$i = 0, $$02$i$i = 0, $$02$i$i$i = 0, $$02$p$i$i = 0, $$02$p$i$i$i = 0, $$2$i$i = 0, $$2$i$i$i = 0, $$lcssa = 0, $$partno$0 = 0, $$ph36$i = 0, $$phi$trans$insert$i6 = 0;
 var $$phi$trans$insert25$i = 0, $$phi$trans$insert28$i = 0, $$pr$i = 0, $$pr$pre$i = 0, $$pr24$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i7 = 0, $$pre$phi$iZ2D = 0, $$pre22$i = 0, $$pre26$i = 0, $$pre29$i = 0, $$pre30$i = 0, $$pre31$i = 0, $$pre33$i = 0, $$pre94 = 0, $$skip$1 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0;
 var $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0;
 var $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0;
 var $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0;
 var $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0;
 var $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0;
 var $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0;
 var $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0;
 var $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0;
 var $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0;
 var $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0;
 var $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0;
 var $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0;
 var $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0;
 var $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0;
 var $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0;
 var $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0;
 var $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0;
 var $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0;
 var $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $buf = 0, $dt$0$lcssa$i = 0, $dt$0$lcssa$i$i = 0, $dt$0$lcssa$i$i17 = 0;
 var $dt$03$i = 0, $dt$03$i$i = 0, $dt$03$i$i13 = 0, $dt$1$i = 0, $dt$1$i$i = 0, $dt$1$i$i16 = 0, $e$0 = 0, $e$0$i = 0, $e$0$i$i = 0, $e$0$lcssa$i = 0, $e$01$i$i = 0, $e$010$i = 0, $e$02$i = 0, $e$02$i$i = 0, $e$02$i$i10 = 0, $e$02$i$i14 = 0, $e$074 = 0, $e$08$i = 0, $e$1 = 0, $e$115$i = 0;
 var $e$115$i$lcssa = 0, $e$218$i42 = 0, $e1$0$i$i = 0, $e1$02$i$i = 0, $e1$04$i$i = 0, $e1$1$ph$i$i = 0, $e1$11$i$i = 0, $e2$0$lcssa12$i$i = 0, $e2$03$i$i = 0, $e2$03$i$i$lcssa = 0, $e2$03$i$i$phi = 0, $exitcond = 0, $f$0$e$2$i = 0, $fading$0 = 0, $fading$2 = 0, $fading$3 = 0, $i$0$lcssa = 0, $i$060 = 0, $i$1 = 0, $i$2$ph$lcssa = 0;
 var $i$2$ph53 = 0, $i$3 = 0, $i$4 = 0, $i$543 = 0, $i$641 = 0, $ispos$i = 0, $loop$061 = 0, $loop$1 = 0, $loop$2$ = 0, $loop$2$ph$lcssa = 0, $loop$2$ph54 = 0, $loop$4 = 0, $n$0$e$0$i = 0, $n$0$lcssa$i = 0, $n$09$i = 0, $neg$i = 0, $nextp = 0, $normal$0 = 0, $not$ = 0, $or$cond = 0;
 var $or$cond$i = 0, $or$cond$i$i = 0, $or$cond$i$i$i = 0, $or$cond3 = 0, $or$cond4 = 0, $or$cond42$i = 0, $p$010$i = 0, $partcoda$068 = 0, $partcoda$1 = 0, $partfine$072 = 0, $partfine$1 = 0, $partloop$066 = 0, $partloop$1 = 0, $partno$0$lcssa = 0, $partno$062 = 0, $partno$1 = 0, $partno$2 = 0, $partsegno$065 = 0, $partsegno$1 = 0, $parttocoda$071 = 0;
 var $parttocoda$1 = 0, $phitmp$i = 0, $phitmp$i18 = 0, $phitmp1$i = 0, $phitmp1$i19 = 0, $ptt = 0, $skip$0$ph55 = 0, $skip$045 = 0, $skip$1 = 0, $skip$3 = 0, $skip$4 = 0, $skip$5 = 0, $skip$5$lcssa = 0, $skip$525 = 0, $starttick$073 = 0, $starttick$1 = 0, $switch$i$i = 0, $tp$0$i = 0, $tp$0$i8 = 0, $tp$016$i = 0;
 var $tp$017$i = 0, $tp$026$i = 0, $tp$027$i = 0, $tp$1$i = 0, $tp$112$pre$i = 0, $tp$113$i = 0, $vmask = 0, $vol$011$ph$i = 0, $x$0 = 0, $x$1$ph56 = 0, $x$146 = 0, $x$146$lcssa = 0, $y$0 = 0, $y$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 592|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vmask = sp + 216|0;
 $nextp = sp + 108|0;
 $ptt = sp;
 $buf = sp + 328|0;
 $0 = ($h|0)==(0|0); //@line 2047 "load_abc.cpp"
 if ($0) {
  STACKTOP = sp;return; //@line 2183 "load_abc.cpp"
 }
 $1 = ((($h)) + 8|0); //@line 2047 "load_abc.cpp"
 $2 = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24); //@line 2047 "load_abc.cpp"
 $3 = ($2|0)==(0|0); //@line 2047 "load_abc.cpp"
 if ($3) {
  STACKTOP = sp;return; //@line 2183 "load_abc.cpp"
 }
 $4 = ((($2)) + 12|0); //@line 2047 "load_abc.cpp"
 $5 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8)|(HEAPU8[$4+2>>0]<<16)|(HEAPU8[$4+3>>0]<<24); //@line 2047 "load_abc.cpp"
 $6 = ($5|0)==(0|0); //@line 2047 "load_abc.cpp"
 if ($6) {
  STACKTOP = sp;return; //@line 2183 "load_abc.cpp"
 }
 HEAP16[$buf>>1] = 65; //@line 2048 "load_abc.cpp"
 $7 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8)|(HEAPU8[$4+2>>0]<<16)|(HEAPU8[$4+3>>0]<<24); //@line 2057 "load_abc.cpp"
 $8 = ((($7)) + 4|0); //@line 2057 "load_abc.cpp"
 $9 = HEAPU8[$8>>0]|(HEAPU8[$8+1>>0]<<8)|(HEAPU8[$8+2>>0]<<16)|(HEAPU8[$8+3>>0]<<24); //@line 2057 "load_abc.cpp"
 HEAP32[$ptt>>2] = $9; //@line 2058 "load_abc.cpp"
 HEAP32[$vmask>>2] = -1; //@line 2059 "load_abc.cpp"
 HEAP32[$nextp>>2] = 1; //@line 2060 "load_abc.cpp"
 $10 = ($7|0)==(0|0); //@line 2061 "load_abc.cpp"
 L10: do {
  if ($10) {
   $i$0$lcssa = 1;$partno$0$lcssa = 0;
  } else {
   $85 = $2;$e$074 = $7;$i$060 = 1;$loop$061 = 1;$partcoda$068 = -1;$partfine$072 = -1;$partloop$066 = 0;$partno$062 = 0;$partsegno$065 = 0;$parttocoda$071 = -1;$starttick$073 = $9;
   L11: while(1) {
    $11 = ((($e$074)) + 8|0); //@line 2062 "load_abc.cpp"
    $12 = HEAP8[$11>>0]|0; //@line 2062 "load_abc.cpp"
    $13 = ($12<<24>>24)==(1); //@line 2062 "load_abc.cpp"
    L13: do {
     if ($13) {
      $14 = ((($e$074)) + 9|0); //@line 2063 "load_abc.cpp"
      $15 = HEAP8[$14>>0]|0; //@line 2063 "load_abc.cpp"
      $16 = $15&255; //@line 2063 "load_abc.cpp"
      switch ($16|0) {
      case 80:  {
       $17 = ((($e$074)) + 4|0); //@line 2065 "load_abc.cpp"
       $18 = HEAPU8[$17>>0]|(HEAPU8[$17+1>>0]<<8)|(HEAPU8[$17+2>>0]<<16)|(HEAPU8[$17+3>>0]<<24); //@line 2065 "load_abc.cpp"
       $19 = ($18>>>0)>($starttick$073>>>0); //@line 2065 "load_abc.cpp"
       if (!($19)) {
        $449 = $85;$e$1 = $e$074;$i$4 = $i$060;$loop$4 = $loop$061;$partcoda$1 = $partcoda$068;$partfine$1 = $partfine$072;$partloop$1 = $partloop$066;$partno$2 = $partno$062;$partsegno$1 = $partsegno$065;$parttocoda$1 = $parttocoda$071;$starttick$1 = $starttick$073;
        break L13;
       }
       $20 = ($partno$062|0)<(26); //@line 2067 "load_abc.cpp"
       if ($20) {
        $21 = (($partno$062) + 1)|0; //@line 2068 "load_abc.cpp"
        $22 = (($ptt) + ($21<<2)|0); //@line 2069 "load_abc.cpp"
        HEAP32[$22>>2] = $18; //@line 2069 "load_abc.cpp"
        $partno$1 = $21;
       } else {
        $partno$1 = $partno$062;
       }
       $23 = ($i$060|0)<(255); //@line 2071 "load_abc.cpp"
       if ($23) {
        $24 = (($partno$1) + 65)|0; //@line 2071 "load_abc.cpp"
        $25 = $24&255; //@line 2071 "load_abc.cpp"
        $26 = (($i$060) + 1)|0; //@line 2071 "load_abc.cpp"
        $27 = (($buf) + ($i$060)|0); //@line 2071 "load_abc.cpp"
        HEAP8[$27>>0] = $25; //@line 2071 "load_abc.cpp"
        $i$1 = $26;
       } else {
        $i$1 = $i$060;
       }
       $28 = (($vmask) + ($partno$1<<2)|0); //@line 2072 "load_abc.cpp"
       HEAP32[$28>>2] = -1; //@line 2072 "load_abc.cpp"
       $29 = (($partno$1) + 1)|0; //@line 2073 "load_abc.cpp"
       $30 = (($nextp) + ($partno$1<<2)|0); //@line 2073 "load_abc.cpp"
       HEAP32[$30>>2] = $29; //@line 2073 "load_abc.cpp"
       $449 = $85;$e$1 = $e$074;$i$4 = $i$1;$loop$4 = $loop$061;$partcoda$1 = $partcoda$068;$partfine$1 = $partfine$072;$partloop$1 = $partloop$066;$partno$2 = $partno$1;$partsegno$1 = $partsegno$065;$parttocoda$1 = $parttocoda$071;$starttick$1 = $18;
       break L13;
       break;
      }
      case 76:  {
       $449 = $85;$e$1 = $e$074;$i$4 = $i$060;$loop$4 = 1;$partcoda$1 = $partcoda$068;$partfine$1 = $partfine$072;$partloop$1 = $partno$062;$partno$2 = $partno$062;$partsegno$1 = $partsegno$065;$parttocoda$1 = $parttocoda$071;$starttick$1 = $starttick$073;
       break L13;
       break;
      }
      case 86:  {
       $31 = ((($e$074)) + 10|0); //@line 2081 "load_abc.cpp"
       $32 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 2081 "load_abc.cpp"
       $33 = (($vmask) + ($partno$062<<2)|0); //@line 2081 "load_abc.cpp"
       HEAP32[$33>>2] = $32; //@line 2081 "load_abc.cpp"
       $449 = $85;$e$1 = $e$074;$i$4 = $i$060;$loop$4 = $loop$061;$partcoda$1 = $partcoda$068;$partfine$1 = $partfine$072;$partloop$1 = $partloop$066;$partno$2 = $partno$062;$partsegno$1 = $partsegno$065;$parttocoda$1 = $parttocoda$071;$starttick$1 = $starttick$073;
       break L13;
       break;
      }
      case 74:  {
       $34 = (($buf) + ($i$060)|0); //@line 2088 "load_abc.cpp"
       $35 = ((($e$074)) + 13|0); //@line 2089 "load_abc.cpp"
       $36 = HEAP8[$35>>0]|0; //@line 2089 "load_abc.cpp"
       $37 = $36&255; //@line 2089 "load_abc.cpp"
       switch ($37|0) {
       case 1:  {
        $fading$0 = 1;
        label = 17;
        break;
       }
       case 0:  {
        $fading$0 = 0;
        label = 17;
        break;
       }
       case 5:  {
        $fading$3 = 1;$loop$1 = $loop$061;$normal$0 = 0;$x$0 = $partsegno$065;
        break;
       }
       case 4:  {
        $fading$3 = 0;$loop$1 = $loop$061;$normal$0 = 0;$x$0 = $partsegno$065;
        break;
       }
       case 3:  {
        $fading$2 = 1;
        label = 20;
        break;
       }
       case 2:  {
        $fading$2 = 0;
        label = 20;
        break;
       }
       default: {
        $fading$3 = 0;$loop$1 = $loop$061;$normal$0 = 0;$x$0 = 0;
       }
       }
       if ((label|0) == 17) {
        label = 0;
        $38 = (($loop$061) + 1)|0; //@line 2095 "load_abc.cpp"
        $fading$3 = $fading$0;$loop$1 = $38;$normal$0 = 1;$x$0 = $partloop$066;
       }
       else if ((label|0) == 20) {
        label = 0;
        $fading$3 = $fading$2;$loop$1 = $loop$061;$normal$0 = 0;$x$0 = 0;
       }
       $39 = (($vmask) + ($partno$062<<2)|0); //@line 2111 "load_abc.cpp"
       $40 = HEAP32[$39>>2]|0; //@line 2111 "load_abc.cpp"
       $41 = ($40|0)==(-1); //@line 2111 "load_abc.cpp"
       if (!($41)) {
        $42 = (($nextp) + ($partno$062<<2)|0); //@line 2111 "load_abc.cpp"
        HEAP32[$42>>2] = $x$0; //@line 2111 "load_abc.cpp"
       }
       $43 = ($partno$062|0)<(26); //@line 2112 "load_abc.cpp"
       if ($43) {
        $44 = ((($e$074)) + 4|0); //@line 2112 "load_abc.cpp"
        $45 = HEAPU8[$44>>0]|(HEAPU8[$44+1>>0]<<8)|(HEAPU8[$44+2>>0]<<16)|(HEAPU8[$44+3>>0]<<24); //@line 2112 "load_abc.cpp"
        $46 = (($partno$062) + 1)|0; //@line 2112 "load_abc.cpp"
        $47 = (($ptt) + ($46<<2)|0); //@line 2112 "load_abc.cpp"
        HEAP32[$47>>2] = $45; //@line 2112 "load_abc.cpp"
       }
       $48 = ($x$0|0)>($partno$062|0); //@line 2113 "load_abc.cpp"
       L42: do {
        if ($48) {
         $i$2$ph$lcssa = $i$060;$loop$2$ph$lcssa = $loop$1;
        } else {
         $49 = ($normal$0|0)==(0); //@line 2117 "load_abc.cpp"
         $i$2$ph53 = $i$060;$loop$2$ph54 = $loop$1;$skip$0$ph55 = 0;$x$1$ph56 = $x$0;
         while(1) {
          $50 = 1 << $loop$2$ph54; //@line 2122 "load_abc.cpp"
          $skip$045 = $skip$0$ph55;$x$146 = $x$1$ph56;
          while(1) {
           $51 = ($skip$045|0)==(1); //@line 2114 "load_abc.cpp"
           $52 = ($x$146|0)==($partcoda$068|0); //@line 2114 "load_abc.cpp"
           $or$cond4 = $51 & $52; //@line 2114 "load_abc.cpp"
           $skip$1 = $or$cond4 ? 0 : $skip$045; //@line 2114 "load_abc.cpp"
           $53 = ($skip$1|0)==(0); //@line 2115 "load_abc.cpp"
           $54 = $53&1; //@line 2115 "load_abc.cpp"
           if ($53) {
            if ($49) {
             $55 = ($x$146|0)==($partfine$072|0); //@line 2118 "load_abc.cpp"
             $$skip$1 = $55 ? 2 : 0; //@line 2118 "load_abc.cpp"
             $56 = ($x$146|0)==($parttocoda$071|0); //@line 2119 "load_abc.cpp"
             $skip$3 = $56 ? 1 : $$skip$1; //@line 2119 "load_abc.cpp"
             $57 = ($skip$3|0)==(0); //@line 2120 "load_abc.cpp"
             $58 = $57&1; //@line 2120 "load_abc.cpp"
             $skip$4 = $skip$3;$y$0 = $58;
            } else {
             $skip$4 = 0;$y$0 = $54;
            }
            $59 = (($vmask) + ($x$146<<2)|0); //@line 2122 "load_abc.cpp"
            $60 = HEAP32[$59>>2]|0; //@line 2122 "load_abc.cpp"
            $61 = $60 & $50; //@line 2122 "load_abc.cpp"
            $62 = ($61|0)==(0); //@line 2122 "load_abc.cpp"
            if ($62) {
             $skip$525 = $skip$4;
            } else {
             $skip$5 = $skip$4;$y$1 = $y$0;
             label = 33;
            }
           } else {
            $skip$5 = $skip$1;$y$1 = $54;
            label = 33;
           }
           if ((label|0) == 33) {
            label = 0;
            $63 = ($y$1|0)==(0); //@line 2124 "load_abc.cpp"
            if ($63) {
             $skip$525 = $skip$5;
            } else {
             $skip$5$lcssa = $skip$5;$x$146$lcssa = $x$146;
             break;
            }
           }
           $74 = (($x$146) + 1)|0; //@line 2130 "load_abc.cpp"
           $75 = ($x$146|0)<($partno$062|0); //@line 2113 "load_abc.cpp"
           if ($75) {
            $skip$045 = $skip$525;$x$146 = $74;
           } else {
            $i$2$ph$lcssa = $i$2$ph53;$loop$2$ph$lcssa = $loop$2$ph54;
            break L42;
           }
          }
          $64 = ($i$2$ph53|0)<(255); //@line 2125 "load_abc.cpp"
          if ($64) {
           $65 = (($x$146$lcssa) + 65)|0; //@line 2125 "load_abc.cpp"
           $66 = $65&255; //@line 2125 "load_abc.cpp"
           $67 = (($i$2$ph53) + 1)|0; //@line 2125 "load_abc.cpp"
           $68 = (($buf) + ($i$2$ph53)|0); //@line 2125 "load_abc.cpp"
           HEAP8[$68>>0] = $66; //@line 2125 "load_abc.cpp"
           $i$3 = $67;
          } else {
           $i$3 = $i$2$ph53;
          }
          $69 = (($nextp) + ($x$146$lcssa<<2)|0); //@line 2126 "load_abc.cpp"
          $70 = HEAP32[$69>>2]|0; //@line 2126 "load_abc.cpp"
          $71 = (($x$146$lcssa) + 1)|0; //@line 2126 "load_abc.cpp"
          $not$ = ($70|0)!=($71|0); //@line 2126 "load_abc.cpp"
          $72 = $not$&1; //@line 2126 "load_abc.cpp"
          $loop$2$ = (($72) + ($loop$2$ph54))|0; //@line 2126 "load_abc.cpp"
          $73 = ($70|0)>($partno$062|0); //@line 2113 "load_abc.cpp"
          if ($73) {
           $i$2$ph$lcssa = $i$3;$loop$2$ph$lcssa = $loop$2$;
           break;
          } else {
           $i$2$ph53 = $i$3;$loop$2$ph54 = $loop$2$;$skip$0$ph55 = $skip$5$lcssa;$x$1$ph56 = $70;
          }
         }
        }
       } while(0);
       $76 = ($fading$3|0)!=(0); //@line 2132 "load_abc.cpp"
       $77 = ($partno$062|0)<(25); //@line 2132 "load_abc.cpp"
       $or$cond = $77 & $76; //@line 2132 "load_abc.cpp"
       $78 = ($i$2$ph$lcssa|0)<(254); //@line 2132 "load_abc.cpp"
       $or$cond3 = $or$cond & $78; //@line 2132 "load_abc.cpp"
       if (!($or$cond3)) {
        $449 = $85;$e$1 = $e$074;$i$4 = $i$2$ph$lcssa;$loop$4 = $loop$2$ph$lcssa;$partcoda$1 = $partcoda$068;$partfine$1 = $partfine$072;$partloop$1 = $partloop$066;$partno$2 = $partno$062;$partsegno$1 = $partsegno$065;$parttocoda$1 = $parttocoda$071;$starttick$1 = $starttick$073;
        break L13;
       }
       $79 = (($partno$062) + 1)|0; //@line 2133 "load_abc.cpp"
       $80 = ((($e$074)) + 4|0); //@line 2134 "load_abc.cpp"
       $81 = HEAPU8[$80>>0]|(HEAPU8[$80+1>>0]<<8)|(HEAPU8[$80+2>>0]<<16)|(HEAPU8[$80+3>>0]<<24); //@line 2134 "load_abc.cpp"
       $82 = (($ptt) + ($79<<2)|0); //@line 2134 "load_abc.cpp"
       HEAP32[$82>>2] = $81; //@line 2134 "load_abc.cpp"
       $83 = (($buf) + ($i$2$ph$lcssa)|0); //@line 2135 "load_abc.cpp"
       HEAP8[$83>>0] = 0; //@line 2135 "load_abc.cpp"
       $84 = ((($85)) + 8|0); //@line 2011 "load_abc.cpp"
       $86 = HEAPU8[$84>>0]|(HEAPU8[$84+1>>0]<<8)|(HEAPU8[$84+2>>0]<<16)|(HEAPU8[$84+3>>0]<<24); //@line 2011 "load_abc.cpp"
       $87 = ((($86)) + 4|0); //@line 2011 "load_abc.cpp"
       $88 = HEAPU8[$87>>0]|(HEAPU8[$87+1>>0]<<8)|(HEAPU8[$87+2>>0]<<16)|(HEAPU8[$87+3>>0]<<24); //@line 2011 "load_abc.cpp"
       $tp$016$i = HEAPU8[$85>>0]|(HEAPU8[$85+1>>0]<<8)|(HEAPU8[$85+2>>0]<<16)|(HEAPU8[$85+3>>0]<<24); //@line 2012 "load_abc.cpp"
       $89 = ($tp$016$i|0)==(0|0); //@line 2012 "load_abc.cpp"
       if ($89) {
        $tp$113$i = $85;
       } else {
        $tp$017$i = $tp$016$i;
        while(1) {
         $90 = ((($tp$017$i)) + 8|0); //@line 2013 "load_abc.cpp"
         $91 = HEAPU8[$90>>0]|(HEAPU8[$90+1>>0]<<8)|(HEAPU8[$90+2>>0]<<16)|(HEAPU8[$90+3>>0]<<24); //@line 2013 "load_abc.cpp"
         $92 = ($91|0)==(0|0); //@line 2013 "load_abc.cpp"
         if ($92) {
          $93 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
          HEAP8[$93>>0]=0&255;HEAP8[$93+1>>0]=(0>>8)&255;HEAP8[$93+2>>0]=(0>>16)&255;HEAP8[$93+3>>0]=0>>24; //@line 540 "load_abc.cpp"
          $94 = ((($93)) + 4|0); //@line 541 "load_abc.cpp"
          HEAP8[$94>>0]=$88&255;HEAP8[$94+1>>0]=($88>>8)&255;HEAP8[$94+2>>0]=($88>>16)&255;HEAP8[$94+3>>0]=$88>>24; //@line 541 "load_abc.cpp"
          $95 = ((($93)) + 8|0); //@line 543 "load_abc.cpp"
          HEAP8[$95>>0] = 1; //@line 543 "load_abc.cpp"
          $96 = ((($93)) + 9|0); //@line 543 "load_abc.cpp"
          HEAP8[$96>>0] = 83; //@line 543 "load_abc.cpp"
          $97 = ((($93)) + 10|0); //@line 543 "load_abc.cpp"
          HEAP8[$97>>0]=0&255;HEAP8[$97+1>>0]=(0>>8)&255;HEAP8[$97+2>>0]=(0>>16)&255;HEAP8[$97+3>>0]=0>>24; //@line 543 "load_abc.cpp"
          $98 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
          $99 = ((($93)) + 14|0); //@line 544 "load_abc.cpp"
          HEAP8[$99>>0] = $98; //@line 544 "load_abc.cpp"
          $100 = ((($93)) + 15|0); //@line 545 "load_abc.cpp"
          HEAP8[$100>>0] = 0; //@line 545 "load_abc.cpp"
          __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($tp$017$i,$93); //@line 1600 "load_abc.cpp"
          $$pre$i = HEAPU8[$90>>0]|(HEAPU8[$90+1>>0]<<8)|(HEAPU8[$90+2>>0]<<16)|(HEAPU8[$90+3>>0]<<24); //@line 2014 "load_abc.cpp"
          $102 = $$pre$i;
         } else {
          $102 = $91;
         }
         $101 = ((($102)) + 4|0); //@line 2014 "load_abc.cpp"
         $103 = HEAPU8[$101>>0]|(HEAPU8[$101+1>>0]<<8)|(HEAPU8[$101+2>>0]<<16)|(HEAPU8[$101+3>>0]<<24); //@line 2014 "load_abc.cpp"
         $104 = ($103>>>0)>($88>>>0); //@line 2014 "load_abc.cpp"
         if ($104) {
          $105 = ((($tp$017$i)) + 4|0); //@line 1954 "load_abc.cpp"
          $e1$02$i$i = HEAPU8[$105>>0]|(HEAPU8[$105+1>>0]<<8)|(HEAPU8[$105+2>>0]<<16)|(HEAPU8[$105+3>>0]<<24); //@line 1954 "load_abc.cpp"
          $106 = ($e1$02$i$i|0)==(0|0); //@line 1954 "load_abc.cpp"
          do {
           if ($106) {
            label = 50;
           } else {
            $e1$04$i$i = $e1$02$i$i;$e2$03$i$i = 0;
            while(1) {
             $108 = ((($e1$04$i$i)) + 4|0); //@line 1954 "load_abc.cpp"
             $109 = HEAPU8[$108>>0]|(HEAPU8[$108+1>>0]<<8)|(HEAPU8[$108+2>>0]<<16)|(HEAPU8[$108+3>>0]<<24); //@line 1954 "load_abc.cpp"
             $110 = ($109>>>0)>($88>>>0); //@line 1954 "load_abc.cpp"
             if ($110) {
              $e2$03$i$i$lcssa = $e2$03$i$i;
              label = 47;
              break;
             }
             $e1$0$i$i = HEAPU8[$e1$04$i$i>>0]|(HEAPU8[$e1$04$i$i+1>>0]<<8)|(HEAPU8[$e1$04$i$i+2>>0]<<16)|(HEAPU8[$e1$04$i$i+3>>0]<<24); //@line 1954 "load_abc.cpp"
             $107 = ($e1$0$i$i|0)==(0|0); //@line 1954 "load_abc.cpp"
             if ($107) {
              $450 = 0;$e2$0$lcssa12$i$i = $e1$04$i$i;
              break;
             } else {
              $e2$03$i$i$phi = $e1$04$i$i;$e1$04$i$i = $e1$0$i$i;$e2$03$i$i = $e2$03$i$i$phi;
             }
            }
            if ((label|0) == 47) {
             label = 0;
             $111 = ($e2$03$i$i$lcssa|0)==(0|0); //@line 1956 "load_abc.cpp"
             if ($111) {
              label = 50;
              break;
             }
             $$pre22$i = HEAPU8[$e2$03$i$i$lcssa>>0]|(HEAPU8[$e2$03$i$i$lcssa+1>>0]<<8)|(HEAPU8[$e2$03$i$i$lcssa+2>>0]<<16)|(HEAPU8[$e2$03$i$i$lcssa+3>>0]<<24); //@line 1957 "load_abc.cpp"
             $450 = $$pre22$i;$e2$0$lcssa12$i$i = $e2$03$i$i$lcssa;
            }
            HEAP8[$90>>0]=$e2$0$lcssa12$i$i&255;HEAP8[$90+1>>0]=($e2$0$lcssa12$i$i>>8)&255;HEAP8[$90+2>>0]=($e2$0$lcssa12$i$i>>16)&255;HEAP8[$90+3>>0]=$e2$0$lcssa12$i$i>>24; //@line 1958 "load_abc.cpp"
            HEAP8[$e2$0$lcssa12$i$i>>0]=0&255;HEAP8[$e2$0$lcssa12$i$i+1>>0]=(0>>8)&255;HEAP8[$e2$0$lcssa12$i$i+2>>0]=(0>>16)&255;HEAP8[$e2$0$lcssa12$i$i+3>>0]=0>>24; //@line 1959 "load_abc.cpp"
            $$pr24$i = $e2$0$lcssa12$i$i;$e1$1$ph$i$i = $450;
           }
          } while(0);
          if ((label|0) == 50) {
           label = 0;
           HEAP8[$105>>0]=0&255;HEAP8[$105+1>>0]=(0>>8)&255;HEAP8[$105+2>>0]=(0>>16)&255;HEAP8[$105+3>>0]=0>>24; //@line 1963 "load_abc.cpp"
           HEAP8[$90>>0]=0&255;HEAP8[$90+1>>0]=(0>>8)&255;HEAP8[$90+2>>0]=(0>>16)&255;HEAP8[$90+3>>0]=0>>24; //@line 1964 "load_abc.cpp"
           $$pr24$i = 0;$e1$1$ph$i$i = $102;
          }
          $112 = ($e1$1$ph$i$i|0)==(0|0); //@line 1966 "load_abc.cpp"
          if ($112) {
           $$pr$i = $$pr24$i;
          } else {
           $e1$11$i$i = $e1$1$ph$i$i;
           while(1) {
            $113 = HEAPU8[$e1$11$i$i>>0]|(HEAPU8[$e1$11$i$i+1>>0]<<8)|(HEAPU8[$e1$11$i$i+2>>0]<<16)|(HEAPU8[$e1$11$i$i+3>>0]<<24); //@line 1967 "load_abc.cpp"
            _free($e1$11$i$i); //@line 1968 "load_abc.cpp"
            $114 = ($113|0)==(0|0); //@line 1966 "load_abc.cpp"
            if ($114) {
             break;
            } else {
             $e1$11$i$i = $113;
            }
           }
           $$pr$pre$i = HEAPU8[$90>>0]|(HEAPU8[$90+1>>0]<<8)|(HEAPU8[$90+2>>0]<<16)|(HEAPU8[$90+3>>0]<<24); //@line 2015 "load_abc.cpp"
           $$pr$i = $$pr$pre$i;
          }
          $$phi$trans$insert25$i = ((($$pr$i)) + 4|0); //@line 2136 "load_abc.cpp"
          $$pre26$i = HEAPU8[$$phi$trans$insert25$i>>0]|(HEAPU8[$$phi$trans$insert25$i+1>>0]<<8)|(HEAPU8[$$phi$trans$insert25$i+2>>0]<<16)|(HEAPU8[$$phi$trans$insert25$i+3>>0]<<24); //@line 2015 "load_abc.cpp"
          $116 = $$pr$i;$118 = $$pre26$i;
         } else {
          $116 = $102;$118 = $103;
         }
         $115 = ((($116)) + 4|0); //@line 2015 "load_abc.cpp"
         $117 = ($118>>>0)<($88>>>0); //@line 2015 "load_abc.cpp"
         L87: do {
          if ($117) {
           $120 = ($116|0)==(0|0); //@line 1591 "load_abc.cpp"
           do {
            if (!($120)) {
             $121 = ($118|0)==($88|0); //@line 1591 "load_abc.cpp"
             if ($121) {
              break L87;
             }
             $122 = ((($116)) + 8|0); //@line 1592 "load_abc.cpp"
             $123 = HEAP8[$122>>0]|0; //@line 1592 "load_abc.cpp"
             $124 = ($123<<24>>24)==(1); //@line 1592 "load_abc.cpp"
             if (!($124)) {
              break;
             }
             $125 = ((($116)) + 9|0); //@line 1592 "load_abc.cpp"
             $126 = HEAP8[$125>>0]|0; //@line 1592 "load_abc.cpp"
             $127 = ($126<<24>>24)==(83); //@line 1592 "load_abc.cpp"
             if (!($127)) {
              break;
             }
             HEAP8[$115>>0]=$88&255;HEAP8[$115+1>>0]=($88>>8)&255;HEAP8[$115+2>>0]=($88>>16)&255;HEAP8[$115+3>>0]=$88>>24; //@line 1593 "load_abc.cpp"
             break L87;
            }
           } while(0);
           $128 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
           HEAP8[$128>>0]=0&255;HEAP8[$128+1>>0]=(0>>8)&255;HEAP8[$128+2>>0]=(0>>16)&255;HEAP8[$128+3>>0]=0>>24; //@line 540 "load_abc.cpp"
           $129 = ((($128)) + 4|0); //@line 541 "load_abc.cpp"
           HEAP8[$129>>0]=$88&255;HEAP8[$129+1>>0]=($88>>8)&255;HEAP8[$129+2>>0]=($88>>16)&255;HEAP8[$129+3>>0]=$88>>24; //@line 541 "load_abc.cpp"
           $130 = ((($128)) + 8|0); //@line 543 "load_abc.cpp"
           HEAP8[$130>>0] = 1; //@line 543 "load_abc.cpp"
           $131 = ((($128)) + 9|0); //@line 543 "load_abc.cpp"
           HEAP8[$131>>0] = 83; //@line 543 "load_abc.cpp"
           $132 = ((($128)) + 10|0); //@line 543 "load_abc.cpp"
           HEAP8[$132>>0]=0&255;HEAP8[$132+1>>0]=(0>>8)&255;HEAP8[$132+2>>0]=(0>>16)&255;HEAP8[$132+3>>0]=0>>24; //@line 543 "load_abc.cpp"
           $133 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
           $134 = ((($128)) + 14|0); //@line 544 "load_abc.cpp"
           HEAP8[$134>>0] = $133; //@line 544 "load_abc.cpp"
           $135 = ((($128)) + 15|0); //@line 545 "load_abc.cpp"
           HEAP8[$135>>0] = 0; //@line 545 "load_abc.cpp"
           __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($tp$017$i,$128); //@line 1600 "load_abc.cpp"
          }
         } while(0);
         $tp$0$i = HEAPU8[$tp$017$i>>0]|(HEAPU8[$tp$017$i+1>>0]<<8)|(HEAPU8[$tp$017$i+2>>0]<<16)|(HEAPU8[$tp$017$i+3>>0]<<24); //@line 2012 "load_abc.cpp"
         $119 = ($tp$0$i|0)==(0|0); //@line 2012 "load_abc.cpp"
         if ($119) {
          break;
         } else {
          $tp$017$i = $tp$0$i;
         }
        }
        $tp$112$pre$i = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24); //@line 2017 "load_abc.cpp"
        $tp$113$i = $tp$112$pre$i;
       }
       while(1) {
        $136 = ((($tp$113$i)) + 8|0); //@line 2019 "load_abc.cpp"
        $137 = HEAPU8[$136>>0]|(HEAPU8[$136+1>>0]<<8)|(HEAPU8[$136+2>>0]<<16)|(HEAPU8[$136+3>>0]<<24); //@line 2019 "load_abc.cpp"
        $138 = ((($tp$113$i)) + 32|0); //@line 2020 "load_abc.cpp"
        $139 = HEAPU8[$138>>0]|(HEAPU8[$138+1>>0]<<8)|(HEAPU8[$138+2>>0]<<16)|(HEAPU8[$138+3>>0]<<24); //@line 2020 "load_abc.cpp"
        $140 = ($139|0)==(-2); //@line 2020 "load_abc.cpp"
        if ($140) {
         $$pre33$i = ((($tp$113$i)) + 28|0); //@line 1188 "load_abc.cpp"
         $$pre$phi$iZ2D = $$pre33$i;
        } else {
         HEAP8[$138>>0]=-2&255;HEAP8[$138+1>>0]=(-2>>8)&255;HEAP8[$138+2>>0]=(-2>>16)&255;HEAP8[$138+3>>0]=-2>>24; //@line 2021 "load_abc.cpp"
         $141 = ((($137)) + 4|0); //@line 2022 "load_abc.cpp"
         $142 = HEAPU8[$141>>0]|(HEAPU8[$141+1>>0]<<8)|(HEAPU8[$141+2>>0]<<16)|(HEAPU8[$141+3>>0]<<24); //@line 2022 "load_abc.cpp"
         $143 = ((($tp$113$i)) + 28|0); //@line 2022 "load_abc.cpp"
         HEAP8[$143>>0]=$142&255;HEAP8[$143+1>>0]=($142>>8)&255;HEAP8[$143+2>>0]=($142>>16)&255;HEAP8[$143+3>>0]=$142>>24; //@line 2022 "load_abc.cpp"
         $$pre$phi$iZ2D = $143;
        }
        $144 = ((($tp$113$i)) + 39|0); //@line 2024 "load_abc.cpp"
        HEAP8[$144>>0] = 0; //@line 2024 "load_abc.cpp"
        $145 = ((($tp$113$i)) + 4|0); //@line 1908 "load_abc.cpp"
        $$ph36$i = $137;$vol$011$ph$i = 127;
        L103: while(1) {
         $146 = HEAP8[$34>>0]|0; //@line 2026 "load_abc.cpp"
         $147 = ($146<<24>>24)!=(0); //@line 2026 "load_abc.cpp"
         $148 = ($vol$011$ph$i|0)>(5); //@line 2026 "load_abc.cpp"
         $149 = $148 & $147; //@line 2026 "load_abc.cpp"
         if (!($149)) {
          break L11;
         }
         $$phi$trans$insert28$i = ((($$ph36$i)) + 4|0); //@line 2136 "load_abc.cpp"
         $$pre29$i = HEAPU8[$$phi$trans$insert28$i>>0]|(HEAPU8[$$phi$trans$insert28$i+1>>0]<<8)|(HEAPU8[$$phi$trans$insert28$i+2>>0]<<16)|(HEAPU8[$$phi$trans$insert28$i+3>>0]<<24); //@line 1907 "load_abc.cpp"
         $151 = $146;$159 = $$pre29$i;$451 = $$ph36$i;$p$010$i = $34;
         while(1) {
          $150 = $151 << 24 >> 24; //@line 2027 "load_abc.cpp"
          $152 = (($150) + -65)|0; //@line 2027 "load_abc.cpp"
          $153 = (($ptt) + ($152<<2)|0); //@line 2027 "load_abc.cpp"
          $154 = HEAP32[$153>>2]|0; //@line 2027 "load_abc.cpp"
          $155 = (($150) + -64)|0; //@line 2028 "load_abc.cpp"
          $156 = (($ptt) + ($155<<2)|0); //@line 2028 "load_abc.cpp"
          $157 = HEAP32[$156>>2]|0; //@line 2028 "load_abc.cpp"
          $158 = (($159) - ($154))|0; //@line 1907 "load_abc.cpp"
          $e$01$i$i = HEAPU8[$145>>0]|(HEAPU8[$145+1>>0]<<8)|(HEAPU8[$145+2>>0]<<16)|(HEAPU8[$145+3>>0]<<24); //@line 1908 "load_abc.cpp"
          $160 = ($e$01$i$i|0)==(0|0); //@line 1908 "load_abc.cpp"
          L108: do {
           if ($160) {
            $189 = $451;$452 = $451;
           } else {
            $171 = $451;$185 = $451;$453 = $451;$454 = $451;$e$02$i$i = $e$01$i$i;
            while(1) {
             $161 = ((($e$02$i$i)) + 4|0); //@line 1908 "load_abc.cpp"
             $162 = HEAPU8[$161>>0]|(HEAPU8[$161+1>>0]<<8)|(HEAPU8[$161+2>>0]<<16)|(HEAPU8[$161+3>>0]<<24); //@line 1908 "load_abc.cpp"
             $163 = ($162>>>0)>($157>>>0); //@line 1908 "load_abc.cpp"
             if ($163) {
              $189 = $454;$452 = $453;
              break L108;
             }
             $164 = ($162>>>0)<($154>>>0); //@line 1909 "load_abc.cpp"
             L113: do {
              if ($164) {
               $455 = $453;$456 = $454;$457 = $185;$458 = $171;
              } else {
               $165 = ((($e$02$i$i)) + 8|0); //@line 1910 "load_abc.cpp"
               $166 = HEAP8[$165>>0]|0; //@line 1910 "load_abc.cpp"
               $167 = ($166<<24>>24)==(1); //@line 1910 "load_abc.cpp"
               if ($167) {
                $168 = ((($e$02$i$i)) + 9|0); //@line 1910 "load_abc.cpp"
                $169 = HEAP8[$168>>0]|0; //@line 1910 "load_abc.cpp"
                switch ($169<<24>>24) {
                case 67: case 83:  {
                 break;
                }
                default: {
                 $455 = $453;$456 = $454;$457 = $185;$458 = $171;
                 break L113;
                }
                }
               }
               $170 = ($e$02$i$i|0)==($171|0); //@line 1911 "load_abc.cpp"
               if ($170) {
                $455 = $453;$456 = $454;$457 = $185;$458 = $171;
                break;
               }
               $172 = (_calloc(1,16)|0); //@line 554 "load_abc.cpp"
               HEAP8[$172>>0]=0&255;HEAP8[$172+1>>0]=(0>>8)&255;HEAP8[$172+2>>0]=(0>>16)&255;HEAP8[$172+3>>0]=0>>24; //@line 555 "load_abc.cpp"
               $173 = HEAPU8[$161>>0]|(HEAPU8[$161+1>>0]<<8)|(HEAPU8[$161+2>>0]<<16)|(HEAPU8[$161+3>>0]<<24); //@line 556 "load_abc.cpp"
               $174 = ((($172)) + 4|0); //@line 556 "load_abc.cpp"
               $175 = HEAP8[$165>>0]|0; //@line 557 "load_abc.cpp"
               $176 = ((($172)) + 8|0); //@line 557 "load_abc.cpp"
               HEAP8[$176>>0] = $175; //@line 557 "load_abc.cpp"
               $177 = ((($e$02$i$i)) + 9|0); //@line 558 "load_abc.cpp"
               $178 = HEAP8[$177>>0]|0; //@line 558 "load_abc.cpp"
               $179 = ((($172)) + 9|0); //@line 558 "load_abc.cpp"
               HEAP8[$179>>0] = $178; //@line 558 "load_abc.cpp"
               $180 = ((($e$02$i$i)) + 10|0); //@line 559 "load_abc.cpp"
               $181 = HEAPU8[$180>>0]|(HEAPU8[$180+1>>0]<<8)|(HEAPU8[$180+2>>0]<<16)|(HEAPU8[$180+3>>0]<<24); //@line 559 "load_abc.cpp"
               $182 = ((($172)) + 10|0); //@line 559 "load_abc.cpp"
               HEAP8[$182>>0]=$181&255;HEAP8[$182+1>>0]=($181>>8)&255;HEAP8[$182+2>>0]=($181>>16)&255;HEAP8[$182+3>>0]=$181>>24; //@line 559 "load_abc.cpp"
               $183 = ((($172)) + 14|0); //@line 560 "load_abc.cpp"
               $184 = (($173) + ($158))|0; //@line 1914 "load_abc.cpp"
               HEAP8[$174>>0]=$184&255;HEAP8[$174+1>>0]=($184>>8)&255;HEAP8[$174+2>>0]=($184>>16)&255;HEAP8[$174+3>>0]=$184>>24; //@line 1914 "load_abc.cpp"
               HEAP8[$183>>0] = 42; //@line 1915 "load_abc.cpp"
               HEAP8[$185>>0]=$172&255;HEAP8[$185+1>>0]=($172>>8)&255;HEAP8[$185+2>>0]=($172>>16)&255;HEAP8[$185+3>>0]=$172>>24; //@line 1916 "load_abc.cpp"
               HEAP8[$136>>0]=$172&255;HEAP8[$136+1>>0]=($172>>8)&255;HEAP8[$136+2>>0]=($172>>16)&255;HEAP8[$136+3>>0]=$172>>24; //@line 1917 "load_abc.cpp"
               $455 = $172;$456 = $172;$457 = $172;$458 = $172;
              }
             } while(0);
             $e$0$i$i = HEAPU8[$e$02$i$i>>0]|(HEAPU8[$e$02$i$i+1>>0]<<8)|(HEAPU8[$e$02$i$i+2>>0]<<16)|(HEAPU8[$e$02$i$i+3>>0]<<24); //@line 1908 "load_abc.cpp"
             $186 = ($e$0$i$i|0)==(0|0); //@line 1908 "load_abc.cpp"
             if ($186) {
              $189 = $456;$452 = $455;
              break;
             } else {
              $171 = $458;$185 = $457;$453 = $455;$454 = $456;$e$02$i$i = $e$0$i$i;
             }
            }
           }
          } while(0);
          $187 = (($158) + ($157))|0; //@line 1922 "load_abc.cpp"
          $188 = ($189|0)==(0|0); //@line 1591 "load_abc.cpp"
          do {
           if ($188) {
            label = 85;
           } else {
            $190 = ((($189)) + 4|0); //@line 1591 "load_abc.cpp"
            $191 = HEAPU8[$190>>0]|(HEAPU8[$190+1>>0]<<8)|(HEAPU8[$190+2>>0]<<16)|(HEAPU8[$190+3>>0]<<24); //@line 1591 "load_abc.cpp"
            $192 = ($191|0)==($187|0); //@line 1591 "load_abc.cpp"
            if ($192) {
             $208 = $452;
             break;
            }
            $193 = ((($189)) + 8|0); //@line 1592 "load_abc.cpp"
            $194 = HEAP8[$193>>0]|0; //@line 1592 "load_abc.cpp"
            $195 = ($194<<24>>24)==(1); //@line 1592 "load_abc.cpp"
            if (!($195)) {
             label = 85;
             break;
            }
            $196 = ((($189)) + 9|0); //@line 1592 "load_abc.cpp"
            $197 = HEAP8[$196>>0]|0; //@line 1592 "load_abc.cpp"
            $198 = ($197<<24>>24)==(83); //@line 1592 "load_abc.cpp"
            if (!($198)) {
             label = 85;
             break;
            }
            HEAP8[$190>>0]=$187&255;HEAP8[$190+1>>0]=($187>>8)&255;HEAP8[$190+2>>0]=($187>>16)&255;HEAP8[$190+3>>0]=$187>>24; //@line 1593 "load_abc.cpp"
            $208 = $452;
           }
          } while(0);
          if ((label|0) == 85) {
           label = 0;
           $199 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
           HEAP8[$199>>0]=0&255;HEAP8[$199+1>>0]=(0>>8)&255;HEAP8[$199+2>>0]=(0>>16)&255;HEAP8[$199+3>>0]=0>>24; //@line 540 "load_abc.cpp"
           $200 = ((($199)) + 4|0); //@line 541 "load_abc.cpp"
           HEAP8[$200>>0]=$187&255;HEAP8[$200+1>>0]=($187>>8)&255;HEAP8[$200+2>>0]=($187>>16)&255;HEAP8[$200+3>>0]=$187>>24; //@line 541 "load_abc.cpp"
           $201 = ((($199)) + 8|0); //@line 543 "load_abc.cpp"
           HEAP8[$201>>0] = 1; //@line 543 "load_abc.cpp"
           $202 = ((($199)) + 9|0); //@line 543 "load_abc.cpp"
           HEAP8[$202>>0] = 83; //@line 543 "load_abc.cpp"
           $203 = ((($199)) + 10|0); //@line 543 "load_abc.cpp"
           HEAP8[$203>>0]=0&255;HEAP8[$203+1>>0]=(0>>8)&255;HEAP8[$203+2>>0]=(0>>16)&255;HEAP8[$203+3>>0]=0>>24; //@line 543 "load_abc.cpp"
           $204 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
           $205 = ((($199)) + 14|0); //@line 544 "load_abc.cpp"
           HEAP8[$205>>0] = $204; //@line 544 "load_abc.cpp"
           $206 = ((($199)) + 15|0); //@line 545 "load_abc.cpp"
           HEAP8[$206>>0] = 0; //@line 545 "load_abc.cpp"
           __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($tp$113$i,$199); //@line 1600 "load_abc.cpp"
           $$pre30$i = HEAPU8[$136>>0]|(HEAPU8[$136+1>>0]<<8)|(HEAPU8[$136+2>>0]<<16)|(HEAPU8[$136+3>>0]<<24); //@line 2030 "load_abc.cpp"
           $208 = $$pre30$i;
          }
          $207 = ((($208)) + 4|0); //@line 2030 "load_abc.cpp"
          $209 = HEAPU8[$207>>0]|(HEAPU8[$207+1>>0]<<8)|(HEAPU8[$207+2>>0]<<16)|(HEAPU8[$207+3>>0]<<24); //@line 2030 "load_abc.cpp"
          $210 = HEAP8[$144>>0]|0; //@line 1186 "load_abc.cpp"
          $211 = ($210<<24>>24)==(0); //@line 1186 "load_abc.cpp"
          if (!($211)) {
           break L103;
          }
          $212 = HEAPU8[$138>>0]|(HEAPU8[$138+1>>0]<<8)|(HEAPU8[$138+2>>0]<<16)|(HEAPU8[$138+3>>0]<<24); //@line 1187 "load_abc.cpp"
          $213 = ($212|0)==(0); //@line 1187 "load_abc.cpp"
          do {
           if ($213) {
            $$0$i$i = 127;
           } else {
            $214 = HEAPU8[$$pre$phi$iZ2D>>0]|(HEAPU8[$$pre$phi$iZ2D+1>>0]<<8)|(HEAPU8[$$pre$phi$iZ2D+2>>0]<<16)|(HEAPU8[$$pre$phi$iZ2D+3>>0]<<24); //@line 1188 "load_abc.cpp"
            $215 = ($209>>>0)<($214>>>0); //@line 1188 "load_abc.cpp"
            if ($215) {
             $$0$i$i = 127;
             break;
            }
            $216 = (($209) - ($214))|0; //@line 1189 "load_abc.cpp"
            $217 = (($216>>>0) / 192)&-1; //@line 334 "load_abc.cpp"
            $218 = ($217*15)|0; //@line 1190 "load_abc.cpp"
            $219 = (($218>>>0) / 384)&-1; //@line 1190 "load_abc.cpp"
            $220 = ($212|0)>(-2); //@line 1191 "load_abc.cpp"
            $221 = ($218>>>0)>(6143); //@line 1191 "load_abc.cpp"
            $or$cond$i$i = $220 & $221; //@line 1191 "load_abc.cpp"
            $$$i$i = $or$cond$i$i ? 15 : $219; //@line 1191 "load_abc.cpp"
            $222 = ($212|0)>(0); //@line 1192 "load_abc.cpp"
            $223 = (0 - ($$$i$i))|0; //@line 1192 "load_abc.cpp"
            $$02$p$i$i = $222 ? $$$i$i : $223; //@line 1192 "load_abc.cpp"
            $$02$i$i = (($$02$p$i$i) + 127)|0; //@line 1192 "load_abc.cpp"
            $224 = ($$02$i$i|0)<(2); //@line 1194 "load_abc.cpp"
            $$$02$i$i = $224 ? 2 : $$02$i$i; //@line 1194 "load_abc.cpp"
            $225 = ($$$02$i$i|0)>(127); //@line 1195 "load_abc.cpp"
            $$2$i$i = $225 ? 127 : $$$02$i$i; //@line 1195 "load_abc.cpp"
            $$0$i$i = $$2$i$i;
           }
          } while(0);
          $226 = ((($p$010$i)) + 1|0); //@line 2026 "load_abc.cpp"
          $227 = HEAP8[$226>>0]|0; //@line 2026 "load_abc.cpp"
          $228 = ($227<<24>>24)!=(0); //@line 2026 "load_abc.cpp"
          $229 = ($$0$i$i|0)>(5); //@line 2026 "load_abc.cpp"
          $230 = $229 & $228; //@line 2026 "load_abc.cpp"
          if ($230) {
           $151 = $227;$159 = $209;$451 = $208;$p$010$i = $226;
          } else {
           break;
          }
         }
         if ($229) {
          $$ph36$i = $208;$vol$011$ph$i = $$0$i$i;
         } else {
          break;
         }
        }
        $231 = ($137|0)==(0|0); //@line 1567 "load_abc.cpp"
        if (!($231)) {
         $$01$i$i = $137;
         while(1) {
          $232 = ((($$01$i$i)) + 8|0); //@line 1568 "load_abc.cpp"
          $233 = HEAP8[$232>>0]|0; //@line 1568 "load_abc.cpp"
          $switch$i$i = ($233&255)<(2); //@line 1568 "load_abc.cpp"
          if (!($switch$i$i)) {
           $234 = ((($$01$i$i)) + 4|0); //@line 1569 "load_abc.cpp"
           $235 = HEAPU8[$234>>0]|(HEAPU8[$234+1>>0]<<8)|(HEAPU8[$234+2>>0]<<16)|(HEAPU8[$234+3>>0]<<24); //@line 1569 "load_abc.cpp"
           $236 = ((($$01$i$i)) + 11|0); //@line 1569 "load_abc.cpp"
           $237 = HEAP8[$236>>0]|0; //@line 1569 "load_abc.cpp"
           $238 = $237&255; //@line 1569 "load_abc.cpp"
           $239 = HEAP8[$144>>0]|0; //@line 1186 "load_abc.cpp"
           $240 = ($239<<24>>24)==(0); //@line 1186 "load_abc.cpp"
           do {
            if ($240) {
             $241 = HEAPU8[$138>>0]|(HEAPU8[$138+1>>0]<<8)|(HEAPU8[$138+2>>0]<<16)|(HEAPU8[$138+3>>0]<<24); //@line 1187 "load_abc.cpp"
             $242 = ($241|0)==(0); //@line 1187 "load_abc.cpp"
             if ($242) {
              $$0$i$i$i = $238;
              break;
             }
             $243 = HEAPU8[$$pre$phi$iZ2D>>0]|(HEAPU8[$$pre$phi$iZ2D+1>>0]<<8)|(HEAPU8[$$pre$phi$iZ2D+2>>0]<<16)|(HEAPU8[$$pre$phi$iZ2D+3>>0]<<24); //@line 1188 "load_abc.cpp"
             $244 = ($235>>>0)<($243>>>0); //@line 1188 "load_abc.cpp"
             if ($244) {
              $$0$i$i$i = $238;
              break;
             }
             $245 = (($235) - ($243))|0; //@line 1189 "load_abc.cpp"
             $246 = (($245>>>0) / 192)&-1; //@line 334 "load_abc.cpp"
             $247 = ($246*15)|0; //@line 1190 "load_abc.cpp"
             $248 = (($247>>>0) / 384)&-1; //@line 1190 "load_abc.cpp"
             $249 = ($241|0)>(-2); //@line 1191 "load_abc.cpp"
             $250 = ($247>>>0)>(6143); //@line 1191 "load_abc.cpp"
             $or$cond$i$i$i = $249 & $250; //@line 1191 "load_abc.cpp"
             $$$i$i$i = $or$cond$i$i$i ? 15 : $248; //@line 1191 "load_abc.cpp"
             $251 = ($241|0)>(0); //@line 1192 "load_abc.cpp"
             $252 = (0 - ($$$i$i$i))|0; //@line 1192 "load_abc.cpp"
             $$02$p$i$i$i = $251 ? $$$i$i$i : $252; //@line 1192 "load_abc.cpp"
             $$02$i$i$i = (($$02$p$i$i$i) + ($238))|0; //@line 1192 "load_abc.cpp"
             $253 = ($$02$i$i$i|0)<(2); //@line 1194 "load_abc.cpp"
             $$$02$i$i$i = $253 ? 2 : $$02$i$i$i; //@line 1194 "load_abc.cpp"
             $254 = ($$$02$i$i$i|0)>(127); //@line 1195 "load_abc.cpp"
             $$2$i$i$i = $254 ? 127 : $$$02$i$i$i; //@line 1195 "load_abc.cpp"
             $$0$i$i$i = $$2$i$i$i;
            } else {
             $$0$i$i$i = 0;
            }
           } while(0);
           $255 = $$0$i$i$i&255; //@line 1569 "load_abc.cpp"
           HEAP8[$236>>0] = $255; //@line 1569 "load_abc.cpp"
          }
          $256 = HEAPU8[$$01$i$i>>0]|(HEAPU8[$$01$i$i+1>>0]<<8)|(HEAPU8[$$01$i$i+2>>0]<<16)|(HEAPU8[$$01$i$i+3>>0]<<24); //@line 1570 "load_abc.cpp"
          $257 = ($256|0)==(0|0); //@line 1567 "load_abc.cpp"
          if ($257) {
           break;
          } else {
           $$01$i$i = $256;
          }
         }
        }
        $tp$1$i = HEAPU8[$tp$113$i>>0]|(HEAPU8[$tp$113$i+1>>0]<<8)|(HEAPU8[$tp$113$i+2>>0]<<16)|(HEAPU8[$tp$113$i+3>>0]<<24); //@line 2017 "load_abc.cpp"
        $258 = ($tp$1$i|0)==(0|0); //@line 2017 "load_abc.cpp"
        if ($258) {
         break;
        } else {
         $tp$113$i = $tp$1$i;
        }
       }
       $$pre31$i = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24); //@line 2035 "load_abc.cpp"
       $259 = ((($$pre31$i)) + 8|0); //@line 2035 "load_abc.cpp"
       $260 = HEAPU8[$259>>0]|(HEAPU8[$259+1>>0]<<8)|(HEAPU8[$259+2>>0]<<16)|(HEAPU8[$259+3>>0]<<24); //@line 2035 "load_abc.cpp"
       $261 = ((($260)) + 4|0); //@line 2035 "load_abc.cpp"
       $262 = HEAPU8[$261>>0]|(HEAPU8[$261+1>>0]<<8)|(HEAPU8[$261+2>>0]<<16)|(HEAPU8[$261+3>>0]<<24); //@line 2035 "load_abc.cpp"
       $263 = (($partno$062) + 66)|0; //@line 2137 "load_abc.cpp"
       $264 = $263&255; //@line 2137 "load_abc.cpp"
       $265 = (($i$2$ph$lcssa) + 1)|0; //@line 2137 "load_abc.cpp"
       HEAP8[$83>>0] = $264; //@line 2137 "load_abc.cpp"
       $266 = (($partno$062) + 2)|0; //@line 2138 "load_abc.cpp"
       $267 = (($ptt) + ($266<<2)|0); //@line 2139 "load_abc.cpp"
       HEAP32[$267>>2] = $262; //@line 2139 "load_abc.cpp"
       $268 = (($partno$062) + 67)|0; //@line 2140 "load_abc.cpp"
       $269 = $268&255; //@line 2140 "load_abc.cpp"
       $270 = (($i$2$ph$lcssa) + 2)|0; //@line 2140 "load_abc.cpp"
       $271 = (($buf) + ($265)|0); //@line 2140 "load_abc.cpp"
       HEAP8[$271>>0] = $269; //@line 2140 "load_abc.cpp"
       $272 = HEAPU8[$259>>0]|(HEAPU8[$259+1>>0]<<8)|(HEAPU8[$259+2>>0]<<16)|(HEAPU8[$259+3>>0]<<24); //@line 2141 "load_abc.cpp"
       $449 = $$pre31$i;$e$1 = $272;$i$4 = $270;$loop$4 = $loop$2$ph$lcssa;$partcoda$1 = $partcoda$068;$partfine$1 = $partfine$072;$partloop$1 = $partloop$066;$partno$2 = $266;$partsegno$1 = $partsegno$065;$parttocoda$1 = $parttocoda$071;$starttick$1 = $262;
       break L13;
       break;
      }
      case 88:  {
       $449 = $85;$e$1 = $e$074;$i$4 = $i$060;$loop$4 = $loop$061;$partcoda$1 = $partcoda$068;$partfine$1 = $partfine$072;$partloop$1 = $partloop$066;$partno$2 = $partno$062;$partsegno$1 = $partsegno$065;$parttocoda$1 = $partno$062;$starttick$1 = $starttick$073;
       break L13;
       break;
      }
      case 79:  {
       $449 = $85;$e$1 = $e$074;$i$4 = $i$060;$loop$4 = $loop$061;$partcoda$1 = $partno$062;$partfine$1 = $partfine$072;$partloop$1 = $partloop$066;$partno$2 = $partno$062;$partsegno$1 = $partsegno$065;$parttocoda$1 = $parttocoda$071;$starttick$1 = $starttick$073;
       break L13;
       break;
      }
      case 70:  {
       $449 = $85;$e$1 = $e$074;$i$4 = $i$060;$loop$4 = $loop$061;$partcoda$1 = $partcoda$068;$partfine$1 = $partno$062;$partloop$1 = $partloop$066;$partno$2 = $partno$062;$partsegno$1 = $partsegno$065;$parttocoda$1 = $parttocoda$071;$starttick$1 = $starttick$073;
       break L13;
       break;
      }
      case 36:  {
       $449 = $85;$e$1 = $e$074;$i$4 = $i$060;$loop$4 = $loop$061;$partcoda$1 = $partcoda$068;$partfine$1 = $partfine$072;$partloop$1 = $partloop$066;$partno$2 = $partno$062;$partsegno$1 = $partno$062;$parttocoda$1 = $parttocoda$071;$starttick$1 = $starttick$073;
       break L13;
       break;
      }
      default: {
       $449 = $85;$e$1 = $e$074;$i$4 = $i$060;$loop$4 = $loop$061;$partcoda$1 = $partcoda$068;$partfine$1 = $partfine$072;$partloop$1 = $partloop$066;$partno$2 = $partno$062;$partsegno$1 = $partsegno$065;$parttocoda$1 = $parttocoda$071;$starttick$1 = $starttick$073;
       break L13;
      }
      }
     } else {
      $449 = $85;$e$1 = $e$074;$i$4 = $i$060;$loop$4 = $loop$061;$partcoda$1 = $partcoda$068;$partfine$1 = $partfine$072;$partloop$1 = $partloop$066;$partno$2 = $partno$062;$partsegno$1 = $partsegno$065;$parttocoda$1 = $parttocoda$071;$starttick$1 = $starttick$073;
     }
    } while(0);
    $273 = (($partno$2) + 97)|0; //@line 2158 "load_abc.cpp"
    $274 = $273&255; //@line 2158 "load_abc.cpp"
    $275 = ((($e$1)) + 14|0); //@line 2158 "load_abc.cpp"
    HEAP8[$275>>0] = $274; //@line 2158 "load_abc.cpp"
    $e$0 = HEAPU8[$e$1>>0]|(HEAPU8[$e$1+1>>0]<<8)|(HEAPU8[$e$1+2>>0]<<16)|(HEAPU8[$e$1+3>>0]<<24); //@line 2061 "load_abc.cpp"
    $276 = ($e$0|0)==(0|0); //@line 2061 "load_abc.cpp"
    if ($276) {
     $i$0$lcssa = $i$4;$partno$0$lcssa = $partno$2;
     break L10;
    } else {
     $85 = $449;$e$074 = $e$0;$i$060 = $i$4;$loop$061 = $loop$4;$partcoda$068 = $partcoda$1;$partfine$072 = $partfine$1;$partloop$066 = $partloop$1;$partno$062 = $partno$2;$partsegno$065 = $partsegno$1;$parttocoda$071 = $parttocoda$1;$starttick$073 = $starttick$1;
    }
   }
   while(1) {
   }
  }
 } while(0);
 $277 = (($i$0$lcssa) + -1)|0; //@line 2160 "load_abc.cpp"
 $278 = ($partno$0$lcssa|0)>(0); //@line 2161 "load_abc.cpp"
 $279 = $278 << 31 >> 31; //@line 2161 "load_abc.cpp"
 $$partno$0 = (($279) + ($partno$0$lcssa))|0; //@line 2161 "load_abc.cpp"
 $280 = (($buf) + ($277)|0); //@line 2162 "load_abc.cpp"
 HEAP8[$280>>0] = 0; //@line 2162 "load_abc.cpp"
 $281 = ($277|0)>(1); //@line 2163 "load_abc.cpp"
 if ($281) {
  $282 = ((($buf)) + 1|0); //@line 2164 "load_abc.cpp"
  $283 = HEAP8[$282>>0]|0; //@line 2164 "load_abc.cpp"
  $284 = ($283<<24>>24)==(0); //@line 2164 "load_abc.cpp"
  if (!($284)) {
   $286 = $283;$i$543 = 1;
   while(1) {
    $285 = $286 << 24 >> 24; //@line 2165 "load_abc.cpp"
    $287 = (($i$543) + -1)|0; //@line 2165 "load_abc.cpp"
    $288 = (($buf) + ($287)|0); //@line 2165 "load_abc.cpp"
    $289 = HEAP8[$288>>0]|0; //@line 2165 "load_abc.cpp"
    $290 = $289 << 24 >> 24; //@line 2165 "load_abc.cpp"
    $291 = (($290) + 1)|0; //@line 2165 "load_abc.cpp"
    $292 = ($285|0)==($291|0); //@line 2165 "load_abc.cpp"
    if (!($292)) {
     $293 = (($285) + -65)|0; //@line 2167 "load_abc.cpp"
     $294 = (($290) + -64)|0; //@line 2168 "load_abc.cpp"
     $295 = (($ptt) + ($294<<2)|0); //@line 2168 "load_abc.cpp"
     $296 = HEAP32[$295>>2]|0; //@line 2168 "load_abc.cpp"
     $297 = (($ptt) + ($293<<2)|0); //@line 2168 "load_abc.cpp"
     $298 = HEAP32[$297>>2]|0; //@line 2168 "load_abc.cpp"
     $299 = ($298>>>0)>($296>>>0); //@line 1976 "load_abc.cpp"
     if ($299) {
      $tp$026$i = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24); //@line 1977 "load_abc.cpp"
      $300 = ($tp$026$i|0)==(0|0); //@line 1977 "load_abc.cpp"
      if (!($300)) {
       $tp$027$i = $tp$026$i;
       while(1) {
        $301 = ((($tp$027$i)) + 36|0); //@line 1978 "load_abc.cpp"
        $302 = HEAP8[$301>>0]|0; //@line 1978 "load_abc.cpp"
        $303 = ($302<<24>>24)==(0); //@line 1978 "load_abc.cpp"
        L166: do {
         if (!($303)) {
          $304 = ((($tp$027$i)) + 4|0); //@line 1980 "load_abc.cpp"
          $e$08$i = HEAPU8[$304>>0]|(HEAPU8[$304+1>>0]<<8)|(HEAPU8[$304+2>>0]<<16)|(HEAPU8[$304+3>>0]<<24); //@line 1980 "load_abc.cpp"
          $305 = ($e$08$i|0)==(0|0); //@line 1980 "load_abc.cpp"
          if (!($305)) {
           $e$010$i = $e$08$i;$n$09$i = 0;
           while(1) {
            $306 = ((($e$010$i)) + 4|0); //@line 1980 "load_abc.cpp"
            $307 = HEAPU8[$306>>0]|(HEAPU8[$306+1>>0]<<8)|(HEAPU8[$306+2>>0]<<16)|(HEAPU8[$306+3>>0]<<24); //@line 1980 "load_abc.cpp"
            $308 = ($307>>>0)<($296>>>0); //@line 1980 "load_abc.cpp"
            if (!($308)) {
             $e$0$lcssa$i = $e$010$i;$n$0$lcssa$i = $n$09$i;
             break;
            }
            $309 = ((($e$010$i)) + 8|0); //@line 1981 "load_abc.cpp"
            $310 = HEAP8[$309>>0]|0; //@line 1981 "load_abc.cpp"
            $311 = ($310<<24>>24)==(1); //@line 1981 "load_abc.cpp"
            $n$0$e$0$i = $311 ? $n$09$i : $e$010$i; //@line 1981 "load_abc.cpp"
            $e$0$i = HEAPU8[$e$010$i>>0]|(HEAPU8[$e$010$i+1>>0]<<8)|(HEAPU8[$e$010$i+2>>0]<<16)|(HEAPU8[$e$010$i+3>>0]<<24); //@line 1980 "load_abc.cpp"
            $312 = ($e$0$i|0)==(0|0); //@line 1980 "load_abc.cpp"
            if ($312) {
             $e$0$lcssa$i = 0;$n$0$lcssa$i = $n$0$e$0$i;
             break;
            } else {
             $e$010$i = $e$0$i;$n$09$i = $n$0$e$0$i;
            }
           }
           $313 = ($n$0$lcssa$i|0)==(0|0); //@line 1982 "load_abc.cpp"
           if (!($313)) {
            $314 = ((($n$0$lcssa$i)) + 15|0); //@line 1982 "load_abc.cpp"
            $315 = HEAP8[$314>>0]|0; //@line 1982 "load_abc.cpp"
            $316 = ($315<<24>>24)==(0); //@line 1982 "load_abc.cpp"
            $317 = ($e$0$lcssa$i|0)==(0|0); //@line 1983 "load_abc.cpp"
            $or$cond42$i = $317 | $316; //@line 1982 "load_abc.cpp"
            if (!($or$cond42$i)) {
             $e$115$i = $e$0$lcssa$i;
             while(1) {
              $318 = ((($e$115$i)) + 4|0); //@line 1983 "load_abc.cpp"
              $319 = HEAPU8[$318>>0]|(HEAPU8[$318+1>>0]<<8)|(HEAPU8[$318+2>>0]<<16)|(HEAPU8[$318+3>>0]<<24); //@line 1983 "load_abc.cpp"
              $320 = ($319>>>0)<($298>>>0); //@line 1983 "load_abc.cpp"
              if (!($320)) {
               $$lcssa = $319;$e$115$i$lcssa = $e$115$i;
               break;
              }
              $321 = HEAPU8[$e$115$i>>0]|(HEAPU8[$e$115$i+1>>0]<<8)|(HEAPU8[$e$115$i+2>>0]<<16)|(HEAPU8[$e$115$i+3>>0]<<24); //@line 1983 "load_abc.cpp"
              $322 = ($321|0)==(0|0); //@line 1983 "load_abc.cpp"
              if ($322) {
               break L166;
              } else {
               $e$115$i = $321;
              }
             }
             $323 = ($$lcssa|0)==($298|0); //@line 1984 "load_abc.cpp"
             if ($323) {
              $e$218$i42 = $e$115$i$lcssa;
              while(1) {
               $324 = ((($e$218$i42)) + 8|0); //@line 1987 "load_abc.cpp"
               $325 = HEAP8[$324>>0]|0; //@line 1987 "load_abc.cpp"
               $326 = ($325<<24>>24)==(1); //@line 1987 "load_abc.cpp"
               $f$0$e$2$i = $326 ? 0 : $e$218$i42; //@line 1987 "load_abc.cpp"
               $327 = HEAPU8[$e$218$i42>>0]|(HEAPU8[$e$218$i42+1>>0]<<8)|(HEAPU8[$e$218$i42+2>>0]<<16)|(HEAPU8[$e$218$i42+3>>0]<<24); //@line 1988 "load_abc.cpp"
               $328 = ($f$0$e$2$i|0)==(0|0); //@line 1986 "load_abc.cpp"
               $329 = ($327|0)!=(0|0); //@line 1986 "load_abc.cpp"
               $or$cond$i = $329 & $328; //@line 1986 "load_abc.cpp"
               if (!($or$cond$i)) {
                break;
               }
               $$phi$trans$insert$i6 = ((($327)) + 4|0); //@line 2168 "load_abc.cpp"
               $$pre$i7 = HEAPU8[$$phi$trans$insert$i6>>0]|(HEAPU8[$$phi$trans$insert$i6+1>>0]<<8)|(HEAPU8[$$phi$trans$insert$i6+2>>0]<<16)|(HEAPU8[$$phi$trans$insert$i6+3>>0]<<24); //@line 1986 "load_abc.cpp"
               $330 = ($$pre$i7|0)==($298|0); //@line 1986 "load_abc.cpp"
               if ($330) {
                $e$218$i42 = $327;
               } else {
                break L166;
               }
              }
              if (!($328)) {
               $331 = ((($f$0$e$2$i)) + 8|0); //@line 1990 "load_abc.cpp"
               $332 = HEAP8[$331>>0]|0; //@line 1990 "load_abc.cpp"
               $333 = ($332<<24>>24)==(0); //@line 1990 "load_abc.cpp"
               if ($333) {
                break;
               }
               $334 = ((($n$0$lcssa$i)) + 8|0); //@line 1991 "load_abc.cpp"
               $335 = HEAP8[$334>>0]|0; //@line 1991 "load_abc.cpp"
               $336 = $335&255; //@line 1991 "load_abc.cpp"
               $337 = $332&255; //@line 1991 "load_abc.cpp"
               $338 = (($336) - ($337))|0; //@line 1991 "load_abc.cpp"
               $ispos$i = ($338|0)>(-1); //@line 1991 "load_abc.cpp"
               $neg$i = (0 - ($338))|0; //@line 1991 "load_abc.cpp"
               $339 = $ispos$i ? $338 : $neg$i; //@line 1991 "load_abc.cpp"
               $340 = ($339|0)<(3); //@line 1991 "load_abc.cpp"
               if (!($340)) {
                break;
               }
               HEAP8[$331>>0] = 1; //@line 1992 "load_abc.cpp"
               $341 = ((($f$0$e$2$i)) + 9|0); //@line 1993 "load_abc.cpp"
               HEAP8[$341>>0] = 72; //@line 1993 "load_abc.cpp"
               $342 = ((($f$0$e$2$i)) + 10|0); //@line 1994 "load_abc.cpp"
               HEAP8[$342>>0]=0&255;HEAP8[$342+1>>0]=(0>>8)&255;HEAP8[$342+2>>0]=(0>>16)&255;HEAP8[$342+3>>0]=0>>24; //@line 1994 "load_abc.cpp"
              }
             }
            }
           }
          }
         }
        } while(0);
        $tp$0$i8 = HEAPU8[$tp$027$i>>0]|(HEAPU8[$tp$027$i+1>>0]<<8)|(HEAPU8[$tp$027$i+2>>0]<<16)|(HEAPU8[$tp$027$i+3>>0]<<24); //@line 1977 "load_abc.cpp"
        $343 = ($tp$0$i8|0)==(0|0); //@line 1977 "load_abc.cpp"
        if ($343) {
         break;
        } else {
         $tp$027$i = $tp$0$i8;
        }
       }
      }
     }
    }
    $344 = (($i$543) + 1)|0; //@line 2164 "load_abc.cpp"
    $345 = (($buf) + ($344)|0); //@line 2164 "load_abc.cpp"
    $346 = HEAP8[$345>>0]|0; //@line 2164 "load_abc.cpp"
    $347 = ($346<<24>>24)==(0); //@line 2164 "load_abc.cpp"
    if ($347) {
     break;
    } else {
     $286 = $346;$i$543 = $344;
    }
   }
  }
 }
 $348 = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24); //@line 2172 "load_abc.cpp"
 $349 = ((($348)) + 8|0); //@line 2172 "load_abc.cpp"
 $350 = HEAPU8[$349>>0]|(HEAPU8[$349+1>>0]<<8)|(HEAPU8[$349+2>>0]<<16)|(HEAPU8[$349+3>>0]<<24); //@line 2172 "load_abc.cpp"
 $351 = ((($350)) + 4|0); //@line 2172 "load_abc.cpp"
 $352 = HEAPU8[$351>>0]|(HEAPU8[$351+1>>0]<<8)|(HEAPU8[$351+2>>0]<<16)|(HEAPU8[$351+3>>0]<<24); //@line 2172 "load_abc.cpp"
 $353 = (($$partno$0) + 1)|0; //@line 2173 "load_abc.cpp"
 $354 = (($ptt) + ($353<<2)|0); //@line 2173 "load_abc.cpp"
 HEAP32[$354>>2] = $352; //@line 2173 "load_abc.cpp"
 $355 = ($$partno$0|0)<(0); //@line 2174 "load_abc.cpp"
 $$pre94 = ((($h)) + 20|0); //@line 1928 "load_abc.cpp"
 if (!($355)) {
  $$pre = HEAP32[$ptt>>2]|0; //@line 2175 "load_abc.cpp"
  $366 = $$pre;$i$641 = 0;
  while(1) {
   $356 = HEAPU8[$$pre94>>0]|(HEAPU8[$$pre94+1>>0]<<8)|(HEAPU8[$$pre94+2>>0]<<16)|(HEAPU8[$$pre94+3>>0]<<24); //@line 1928 "load_abc.cpp"
   $357 = ($356*12288)|0; //@line 339 "load_abc.cpp"
   $358 = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24); //@line 1929 "load_abc.cpp"
   $359 = ($358|0)==(0|0); //@line 1929 "load_abc.cpp"
   if ($359) {
    $$0$i$i12 = 0;
   } else {
    $360 = ((($358)) + 4|0); //@line 1929 "load_abc.cpp"
    $361 = HEAPU8[$360>>0]|(HEAPU8[$360+1>>0]<<8)|(HEAPU8[$360+2>>0]<<16)|(HEAPU8[$360+3>>0]<<24); //@line 1929 "load_abc.cpp"
    $362 = ($361|0)==(0|0); //@line 1929 "load_abc.cpp"
    if ($362) {
     $$0$i$i12 = 0;
    } else {
     $dt$03$i$i = 0;$e$02$i$i10 = $361;
     while(1) {
      $363 = ((($e$02$i$i10)) + 4|0); //@line 1931 "load_abc.cpp"
      $364 = HEAPU8[$363>>0]|(HEAPU8[$363+1>>0]<<8)|(HEAPU8[$363+2>>0]<<16)|(HEAPU8[$363+3>>0]<<24); //@line 1931 "load_abc.cpp"
      $365 = ($364>>>0)>($366>>>0); //@line 1931 "load_abc.cpp"
      if ($365) {
       $dt$0$lcssa$i$i = $dt$03$i$i;
       break;
      }
      $367 = ((($e$02$i$i10)) + 8|0); //@line 1932 "load_abc.cpp"
      $368 = HEAP8[$367>>0]|0; //@line 1932 "load_abc.cpp"
      $369 = ($368<<24>>24)==(1); //@line 1932 "load_abc.cpp"
      if ($369) {
       $370 = ((($e$02$i$i10)) + 9|0); //@line 1932 "load_abc.cpp"
       $371 = HEAP8[$370>>0]|0; //@line 1932 "load_abc.cpp"
       $372 = ($371<<24>>24)==(80); //@line 1932 "load_abc.cpp"
       if ($372) {
        $373 = (($364) + ($dt$03$i$i))|0; //@line 1933 "load_abc.cpp"
        $374 = (($373>>>0) % ($357>>>0))&-1; //@line 1934 "load_abc.cpp"
        $375 = ($374|0)==(0); //@line 1934 "load_abc.cpp"
        if ($375) {
         $dt$1$i$i = $dt$03$i$i;
        } else {
         $376 = (($373) + ($357))|0; //@line 1935 "load_abc.cpp"
         $377 = (($376>>>0) % ($357>>>0))&-1; //@line 1937 "load_abc.cpp"
         $378 = (($dt$03$i$i) + ($357))|0; //@line 1937 "load_abc.cpp"
         $379 = (($378) - ($377))|0; //@line 1938 "load_abc.cpp"
         $dt$1$i$i = $379;
        }
       } else {
        $dt$1$i$i = $dt$03$i$i;
       }
      } else {
       $dt$1$i$i = $dt$03$i$i;
      }
      $380 = HEAPU8[$e$02$i$i10>>0]|(HEAPU8[$e$02$i$i10+1>>0]<<8)|(HEAPU8[$e$02$i$i10+2>>0]<<16)|(HEAPU8[$e$02$i$i10+3>>0]<<24); //@line 1931 "load_abc.cpp"
      $381 = ($380|0)==(0|0); //@line 1931 "load_abc.cpp"
      if ($381) {
       $dt$0$lcssa$i$i = $dt$1$i$i;
       break;
      } else {
       $dt$03$i$i = $dt$1$i$i;$e$02$i$i10 = $380;
      }
     }
     $382 = (($dt$0$lcssa$i$i) + ($366))|0; //@line 1942 "load_abc.cpp"
     $phitmp$i = (($382>>>0) / 192)&-1; //@line 1942 "load_abc.cpp"
     $phitmp1$i = $phitmp$i >>> 6; //@line 1942 "load_abc.cpp"
     $$0$i$i12 = $phitmp1$i;
    }
   }
   $383 = (($$0$i$i12>>>0) / ($356>>>0))&-1; //@line 1947 "load_abc.cpp"
   $384 = $383&255; //@line 2175 "load_abc.cpp"
   $385 = (($partp) + ($i$641<<1)|0); //@line 2175 "load_abc.cpp"
   HEAP8[$385>>0] = $384; //@line 2175 "load_abc.cpp"
   $386 = (($i$641) + 1)|0; //@line 2176 "load_abc.cpp"
   $387 = (($ptt) + ($386<<2)|0); //@line 2176 "load_abc.cpp"
   $388 = HEAP32[$387>>2]|0; //@line 2176 "load_abc.cpp"
   $389 = HEAPU8[$$pre94>>0]|(HEAPU8[$$pre94+1>>0]<<8)|(HEAPU8[$$pre94+2>>0]<<16)|(HEAPU8[$$pre94+3>>0]<<24); //@line 1928 "load_abc.cpp"
   $390 = ($389*12288)|0; //@line 339 "load_abc.cpp"
   $391 = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24); //@line 1929 "load_abc.cpp"
   $392 = ($391|0)==(0|0); //@line 1929 "load_abc.cpp"
   if ($392) {
    $$0$i$i21 = 0;
   } else {
    $393 = ((($391)) + 4|0); //@line 1929 "load_abc.cpp"
    $394 = HEAPU8[$393>>0]|(HEAPU8[$393+1>>0]<<8)|(HEAPU8[$393+2>>0]<<16)|(HEAPU8[$393+3>>0]<<24); //@line 1929 "load_abc.cpp"
    $395 = ($394|0)==(0|0); //@line 1929 "load_abc.cpp"
    if ($395) {
     $$0$i$i21 = 0;
    } else {
     $dt$03$i$i13 = 0;$e$02$i$i14 = $394;
     while(1) {
      $396 = ((($e$02$i$i14)) + 4|0); //@line 1931 "load_abc.cpp"
      $397 = HEAPU8[$396>>0]|(HEAPU8[$396+1>>0]<<8)|(HEAPU8[$396+2>>0]<<16)|(HEAPU8[$396+3>>0]<<24); //@line 1931 "load_abc.cpp"
      $398 = ($397>>>0)>($388>>>0); //@line 1931 "load_abc.cpp"
      if ($398) {
       $dt$0$lcssa$i$i17 = $dt$03$i$i13;
       break;
      }
      $399 = ((($e$02$i$i14)) + 8|0); //@line 1932 "load_abc.cpp"
      $400 = HEAP8[$399>>0]|0; //@line 1932 "load_abc.cpp"
      $401 = ($400<<24>>24)==(1); //@line 1932 "load_abc.cpp"
      if ($401) {
       $402 = ((($e$02$i$i14)) + 9|0); //@line 1932 "load_abc.cpp"
       $403 = HEAP8[$402>>0]|0; //@line 1932 "load_abc.cpp"
       $404 = ($403<<24>>24)==(80); //@line 1932 "load_abc.cpp"
       if ($404) {
        $405 = (($397) + ($dt$03$i$i13))|0; //@line 1933 "load_abc.cpp"
        $406 = (($405>>>0) % ($390>>>0))&-1; //@line 1934 "load_abc.cpp"
        $407 = ($406|0)==(0); //@line 1934 "load_abc.cpp"
        if ($407) {
         $dt$1$i$i16 = $dt$03$i$i13;
        } else {
         $408 = (($405) + ($390))|0; //@line 1935 "load_abc.cpp"
         $409 = (($408>>>0) % ($390>>>0))&-1; //@line 1937 "load_abc.cpp"
         $410 = (($dt$03$i$i13) + ($390))|0; //@line 1937 "load_abc.cpp"
         $411 = (($410) - ($409))|0; //@line 1938 "load_abc.cpp"
         $dt$1$i$i16 = $411;
        }
       } else {
        $dt$1$i$i16 = $dt$03$i$i13;
       }
      } else {
       $dt$1$i$i16 = $dt$03$i$i13;
      }
      $412 = HEAPU8[$e$02$i$i14>>0]|(HEAPU8[$e$02$i$i14+1>>0]<<8)|(HEAPU8[$e$02$i$i14+2>>0]<<16)|(HEAPU8[$e$02$i$i14+3>>0]<<24); //@line 1931 "load_abc.cpp"
      $413 = ($412|0)==(0|0); //@line 1931 "load_abc.cpp"
      if ($413) {
       $dt$0$lcssa$i$i17 = $dt$1$i$i16;
       break;
      } else {
       $dt$03$i$i13 = $dt$1$i$i16;$e$02$i$i14 = $412;
      }
     }
     $414 = (($dt$0$lcssa$i$i17) + ($388))|0; //@line 1942 "load_abc.cpp"
     $phitmp$i18 = (($414>>>0) / 192)&-1; //@line 1942 "load_abc.cpp"
     $phitmp1$i19 = $phitmp$i18 >>> 6; //@line 1942 "load_abc.cpp"
     $$0$i$i21 = $phitmp1$i19;
    }
   }
   $415 = (($$0$i$i21>>>0) / ($389>>>0))&-1; //@line 1947 "load_abc.cpp"
   $416 = $415&255; //@line 2176 "load_abc.cpp"
   $417 = (((($partp) + ($i$641<<1)|0)) + 1|0); //@line 2176 "load_abc.cpp"
   HEAP8[$417>>0] = $416; //@line 2176 "load_abc.cpp"
   $exitcond = ($386|0)==($353|0); //@line 2174 "load_abc.cpp"
   if ($exitcond) {
    break;
   } else {
    $366 = $388;$i$641 = $386;
   }
  }
 }
 $418 = HEAPU8[$$pre94>>0]|(HEAPU8[$$pre94+1>>0]<<8)|(HEAPU8[$$pre94+2>>0]<<16)|(HEAPU8[$$pre94+3>>0]<<24); //@line 1928 "load_abc.cpp"
 $419 = ($418*12288)|0; //@line 339 "load_abc.cpp"
 $420 = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24); //@line 1929 "load_abc.cpp"
 $421 = ($420|0)==(0|0); //@line 1929 "load_abc.cpp"
 if ($421) {
  $$0$i = 0;
 } else {
  $422 = ((($420)) + 4|0); //@line 1929 "load_abc.cpp"
  $423 = HEAPU8[$422>>0]|(HEAPU8[$422+1>>0]<<8)|(HEAPU8[$422+2>>0]<<16)|(HEAPU8[$422+3>>0]<<24); //@line 1929 "load_abc.cpp"
  $424 = ($423|0)==(0|0); //@line 1929 "load_abc.cpp"
  if ($424) {
   $$0$i = 0;
  } else {
   $dt$03$i = 0;$e$02$i = $423;
   while(1) {
    $425 = ((($e$02$i)) + 4|0); //@line 1931 "load_abc.cpp"
    $426 = HEAPU8[$425>>0]|(HEAPU8[$425+1>>0]<<8)|(HEAPU8[$425+2>>0]<<16)|(HEAPU8[$425+3>>0]<<24); //@line 1931 "load_abc.cpp"
    $427 = ($426>>>0)>($352>>>0); //@line 1931 "load_abc.cpp"
    if ($427) {
     $dt$0$lcssa$i = $dt$03$i;
     break;
    }
    $428 = ((($e$02$i)) + 8|0); //@line 1932 "load_abc.cpp"
    $429 = HEAP8[$428>>0]|0; //@line 1932 "load_abc.cpp"
    $430 = ($429<<24>>24)==(1); //@line 1932 "load_abc.cpp"
    if ($430) {
     $431 = ((($e$02$i)) + 9|0); //@line 1932 "load_abc.cpp"
     $432 = HEAP8[$431>>0]|0; //@line 1932 "load_abc.cpp"
     $433 = ($432<<24>>24)==(80); //@line 1932 "load_abc.cpp"
     if ($433) {
      $434 = (($426) + ($dt$03$i))|0; //@line 1933 "load_abc.cpp"
      $435 = (($434>>>0) % ($419>>>0))&-1; //@line 1934 "load_abc.cpp"
      $436 = ($435|0)==(0); //@line 1934 "load_abc.cpp"
      if ($436) {
       $dt$1$i = $dt$03$i;
      } else {
       $437 = (($434) + ($419))|0; //@line 1935 "load_abc.cpp"
       $438 = (($437>>>0) % ($419>>>0))&-1; //@line 1937 "load_abc.cpp"
       $439 = (($dt$03$i) + ($419))|0; //@line 1937 "load_abc.cpp"
       $440 = (($439) - ($438))|0; //@line 1938 "load_abc.cpp"
       $dt$1$i = $440;
      }
     } else {
      $dt$1$i = $dt$03$i;
     }
    } else {
     $dt$1$i = $dt$03$i;
    }
    $441 = HEAPU8[$e$02$i>>0]|(HEAPU8[$e$02$i+1>>0]<<8)|(HEAPU8[$e$02$i+2>>0]<<16)|(HEAPU8[$e$02$i+3>>0]<<24); //@line 1931 "load_abc.cpp"
    $442 = ($441|0)==(0|0); //@line 1931 "load_abc.cpp"
    if ($442) {
     $dt$0$lcssa$i = $dt$1$i;
     break;
    } else {
     $dt$03$i = $dt$1$i;$e$02$i = $441;
    }
   }
   $443 = (($dt$0$lcssa$i) + ($352))|0; //@line 1942 "load_abc.cpp"
   $$0$i = $443;
  }
 }
 $444 = (($$0$i>>>0) % ($419>>>0))&-1; //@line 2180 "load_abc.cpp"
 $445 = ($444|0)==(0); //@line 2180 "load_abc.cpp"
 if (!($445)) {
  $446 = (((($partp) + ($$partno$0<<1)|0)) + 1|0); //@line 2181 "load_abc.cpp"
  $447 = HEAP8[$446>>0]|0; //@line 2181 "load_abc.cpp"
  $448 = (($447) + 1)<<24>>24; //@line 2181 "load_abc.cpp"
  HEAP8[$446>>0] = $448; //@line 2181 "load_abc.cpp"
 }
 __ZL13abc_set_partsPPcS_($abcparts,$buf); //@line 2182 "load_abc.cpp"
 STACKTOP = sp;return; //@line 2183 "load_abc.cpp"
}
function __ZL24abc_partpat_to_orderlistPA2_hPKcP10_ABCHANDLEPPhi($partp,$abcparts,$list,$orderlen) {
 $partp = $partp|0;
 $abcparts = $abcparts|0;
 $list = $list|0;
 $orderlen = $orderlen|0;
 var $$01 = 0, $$012 = 0, $$1$lcssa = 0, $$1$lcssa$lcssa = 0, $$14 = 0, $$2 = 0, $$31 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $orderlist$0 = 0;
 var $orderlist$115 = 0, $orderlist$2$lcssa = 0, $orderlist$2$lcssa$lcssa = 0, $orderlist$27 = 0, $orderlist$3 = 0, $orderlist$4 = 0, $orderlist$53 = 0, $orderlist$6 = 0, $p$014 = 0, $partsused$013 = 0, $partsused$1$lcssa = 0, $partsused$1$lcssa$lcssa = 0, $partsused$16 = 0, $t$05 = 0, $t$12 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$list>>2]|0; //@line 3592 "load_abc.cpp"
 $1 = ($0|0)==(0|0); //@line 3594 "load_abc.cpp"
 if ($1) {
  HEAP32[300426] = 128; //@line 3595 "load_abc.cpp"
  $2 = (_calloc(128,1)|0); //@line 3596 "load_abc.cpp"
  HEAP32[$list>>2] = $2; //@line 3597 "load_abc.cpp"
  $orderlist$0 = $2;
 } else {
  $orderlist$0 = $0;
 }
 $3 = ($abcparts|0)==(0|0); //@line 3599 "load_abc.cpp"
 if ($3) {
  $$2 = $orderlen;$orderlist$4 = $orderlist$0;
 } else {
  $4 = HEAP8[$abcparts>>0]|0; //@line 3601 "load_abc.cpp"
  $5 = ($4<<24>>24)==(0); //@line 3601 "load_abc.cpp"
  if ($5) {
   $$2 = $orderlen;$orderlist$4 = $orderlist$0;
  } else {
   $$012 = $orderlen;$7 = $4;$orderlist$115 = $orderlist$0;$p$014 = $abcparts;$partsused$013 = 0;
   while(1) {
    $6 = $7 << 24 >> 24; //@line 3602 "load_abc.cpp"
    $8 = (($6) + -65)|0; //@line 3602 "load_abc.cpp"
    $9 = (($partp) + ($8<<1)|0); //@line 3602 "load_abc.cpp"
    $10 = HEAP8[$9>>0]|0; //@line 3602 "load_abc.cpp"
    $11 = (((($partp) + ($8<<1)|0)) + 1|0); //@line 3602 "load_abc.cpp"
    $12 = HEAP8[$11>>0]|0; //@line 3602 "load_abc.cpp"
    $13 = ($10&255)<($12&255); //@line 3602 "load_abc.cpp"
    if ($13) {
     $14 = $10&255; //@line 3602 "load_abc.cpp"
     $$14 = $$012;$orderlist$27 = $orderlist$115;$partsused$16 = $partsused$013;$t$05 = $14;
     while(1) {
      $15 = HEAP32[300426]|0; //@line 3603 "load_abc.cpp"
      $16 = ($$14|0)==($15|0); //@line 3603 "load_abc.cpp"
      if ($16) {
       $17 = $$14 << 1; //@line 3604 "load_abc.cpp"
       HEAP32[300426] = $17; //@line 3604 "load_abc.cpp"
       $18 = (_realloc($orderlist$27,$17)|0); //@line 3605 "load_abc.cpp"
       HEAP32[$list>>2] = $18; //@line 3606 "load_abc.cpp"
       $orderlist$3 = $18;
      } else {
       $orderlist$3 = $orderlist$27;
      }
      $19 = $t$05&255; //@line 3608 "load_abc.cpp"
      $20 = (($orderlist$3) + ($$14)|0); //@line 3608 "load_abc.cpp"
      HEAP8[$20>>0] = $19; //@line 3608 "load_abc.cpp"
      $21 = (($$14) + 1)|0; //@line 3609 "load_abc.cpp"
      $22 = (($partsused$16) + 1)|0; //@line 3610 "load_abc.cpp"
      $23 = (($t$05) + 1)|0; //@line 3602 "load_abc.cpp"
      $24 = HEAP8[$p$014>>0]|0; //@line 3602 "load_abc.cpp"
      $25 = $24 << 24 >> 24; //@line 3602 "load_abc.cpp"
      $26 = (($25) + -65)|0; //@line 3602 "load_abc.cpp"
      $27 = (((($partp) + ($26<<1)|0)) + 1|0); //@line 3602 "load_abc.cpp"
      $28 = HEAP8[$27>>0]|0; //@line 3602 "load_abc.cpp"
      $29 = $28&255; //@line 3602 "load_abc.cpp"
      $30 = ($23|0)<($29|0); //@line 3602 "load_abc.cpp"
      if ($30) {
       $$14 = $21;$orderlist$27 = $orderlist$3;$partsused$16 = $22;$t$05 = $23;
      } else {
       $$1$lcssa = $21;$orderlist$2$lcssa = $orderlist$3;$partsused$1$lcssa = $22;
       break;
      }
     }
    } else {
     $$1$lcssa = $$012;$orderlist$2$lcssa = $orderlist$115;$partsused$1$lcssa = $partsused$013;
    }
    $31 = ((($p$014)) + 1|0); //@line 3601 "load_abc.cpp"
    $32 = HEAP8[$31>>0]|0; //@line 3601 "load_abc.cpp"
    $33 = ($32<<24>>24)==(0); //@line 3601 "load_abc.cpp"
    if ($33) {
     $$1$lcssa$lcssa = $$1$lcssa;$orderlist$2$lcssa$lcssa = $orderlist$2$lcssa;$partsused$1$lcssa$lcssa = $partsused$1$lcssa;
     break;
    } else {
     $$012 = $$1$lcssa;$7 = $32;$orderlist$115 = $orderlist$2$lcssa;$p$014 = $31;$partsused$013 = $partsused$1$lcssa;
    }
   }
   $34 = ($partsused$1$lcssa$lcssa|0)==(0); //@line 3613 "load_abc.cpp"
   if ($34) {
    $$2 = $$1$lcssa$lcssa;$orderlist$4 = $orderlist$2$lcssa$lcssa;
   } else {
    $$01 = $$1$lcssa$lcssa;
    return ($$01|0); //@line 3626 "load_abc.cpp"
   }
  }
 }
 $35 = ((($partp)) + 52|0); //@line 3616 "load_abc.cpp"
 $36 = HEAP8[$35>>0]|0; //@line 3616 "load_abc.cpp"
 $37 = ((($partp)) + 53|0); //@line 3616 "load_abc.cpp"
 $38 = HEAP8[$37>>0]|0; //@line 3616 "load_abc.cpp"
 $39 = ($36&255)<($38&255); //@line 3616 "load_abc.cpp"
 if (!($39)) {
  $$01 = $$2;
  return ($$01|0); //@line 3626 "load_abc.cpp"
 }
 $40 = $36&255; //@line 3616 "load_abc.cpp"
 $$31 = $$2;$orderlist$53 = $orderlist$4;$t$12 = $40;
 while(1) {
  $41 = HEAP32[300426]|0; //@line 3617 "load_abc.cpp"
  $42 = ($$31|0)==($41|0); //@line 3617 "load_abc.cpp"
  if ($42) {
   $43 = $$31 << 1; //@line 3618 "load_abc.cpp"
   HEAP32[300426] = $43; //@line 3618 "load_abc.cpp"
   $44 = (_realloc($orderlist$53,$43)|0); //@line 3619 "load_abc.cpp"
   HEAP32[$list>>2] = $44; //@line 3620 "load_abc.cpp"
   $orderlist$6 = $44;
  } else {
   $orderlist$6 = $orderlist$53;
  }
  $45 = $t$12&255; //@line 3622 "load_abc.cpp"
  $46 = (($orderlist$6) + ($$31)|0); //@line 3622 "load_abc.cpp"
  HEAP8[$46>>0] = $45; //@line 3622 "load_abc.cpp"
  $47 = (($$31) + 1)|0; //@line 3623 "load_abc.cpp"
  $48 = (($t$12) + 1)|0; //@line 3616 "load_abc.cpp"
  $49 = HEAP8[$37>>0]|0; //@line 3616 "load_abc.cpp"
  $50 = $49&255; //@line 3616 "load_abc.cpp"
  $51 = ($48|0)<($50|0); //@line 3616 "load_abc.cpp"
  if ($51) {
   $$31 = $47;$orderlist$53 = $orderlist$6;$t$12 = $48;
  } else {
   $$01 = $47;
   break;
  }
 }
 return ($$01|0); //@line 3626 "load_abc.cpp"
}
function __ZL16abc_MIDI_commandP10_ABCHANDLEPcc($h,$p,$delim) {
 $h = $h|0;
 $p = $p|0;
 $delim = $delim|0;
 var $$$h$0$i$i = 0, $$$h$0$i$i18 = 0, $$$h$0$i$i29 = 0, $$$h$0$i$i29$1 = 0, $$$h$0$i$i29$2 = 0, $$$h$0$i$i29$3 = 0, $$$h$0$i$i52 = 0, $$$h$0$i$i65 = 0, $$$i = 0, $$$i171 = 0, $$0$i = 0, $$0$i$lcssa = 0, $$0$i101177 = 0, $$0$i116178 = 0, $$0$i138 = 0, $$0$i166 = 0, $$0$i31 = 0, $$0$i31$lcssa = 0, $$0$i42 = 0, $$0$i42$lcssa = 0;
 var $$0$i55 = 0, $$0$i55$lcssa = 0, $$0$i68 = 0, $$0$i68$lcssa = 0, $$0$i8 = 0, $$0$i8$lcssa = 0, $$0$i98$ph = 0, $$0$lcssa$i = 0, $$0$lcssa$i120 = 0, $$0$lcssa$i140 = 0, $$0$lcssa$i170 = 0, $$01$i = 0, $$01$i$lcssa = 0, $$01$i168 = 0, $$015$i = 0, $$015$i$lcssa = 0, $$022$i = 0, $$022$i$lcssa = 0, $$040$i = 0, $$05$i = 0;
 var $$05$i$lcssa = 0, $$07$i = 0, $$1$i = 0, $$1$i$1 = 0, $$1$i$1$lcssa = 0, $$1$i$2 = 0, $$1$i$2$lcssa = 0, $$1$i$3 = 0, $$1$i$3$lcssa = 0, $$1$i$lcssa = 0, $$1$i80 = 0, $$1$i80$lcssa = 0, $$1$lcssa$i = 0, $$1$lcssa$i132 = 0, $$1$lcssa$i83 = 0, $$10$i = 0, $$11$i = 0, $$119$i = 0, $$119$i$lcssa = 0, $$135$i = 0;
 var $$135$i143 = 0, $$135$i143$lcssa = 0, $$2$i81 = 0, $$2$i81$lcssa = 0, $$2$ph$i = 0, $$211$i245 = 0, $$231$i = 0, $$231$i144 = 0, $$3$i = 0, $$3$i$lcssa = 0, $$3$i146 = 0, $$3$i85 = 0, $$3$i85$lcssa = 0, $$3$i85$pn = 0, $$3$lcssa$i = 0, $$38$i = 0, $$38$i$lcssa = 0, $$4$i = 0, $$4$i$lcssa = 0, $$4$lcssa$i = 0;
 var $$4$ph$i = 0, $$426$i = 0, $$5$i = 0, $$521$i = 0, $$521$i$lcssa = 0, $$6$i = 0, $$619$i = 0, $$7$i = 0, $$728$i = 0, $$8$i = 0, $$8$i$lcssa = 0, $$8$i$pn = 0, $$h$0$i$i = 0, $$h$0$i$i112 = 0, $$h$0$i$i129 = 0, $$h$0$i$i154 = 0, $$h$0$i$i16 = 0, $$h$0$i$i27 = 0, $$h$0$i$i27$1 = 0, $$h$0$i$i27$2 = 0;
 var $$h$0$i$i27$3 = 0, $$h$0$i$i40 = 0, $$h$0$i$i50 = 0, $$h$0$i$i63 = 0, $$h$0$i$i78 = 0, $$h$0$i$i93 = 0, $$h$0$i14$i = 0, $$h$0$i17$i = 0, $$h$0$i22$i = 0, $$h$0$i2268$i = 0, $$h$0$i33$$i = 0, $$h$0$i33$i = 0, $$h$0$i44$i = 0, $$h$0$i4462$i = 0, $$h$0$i55$$i = 0, $$h$0$i55$i = 0, $$lcssa526 = 0, $$lcssa527 = 0, $$lcssa528 = 0, $$lcssa529 = 0;
 var $$lcssa532 = 0, $$lcssa533 = 0, $$lcssa536 = 0, $$lcssa537 = 0, $$lcssa538 = 0, $$lcssa539 = 0, $$lcssa542 = 0, $$lcssa543 = 0, $$lcssa547 = 0, $$lcssa548 = 0, $$lcssa549 = 0, $$lcssa551 = 0, $$lcssa552 = 0, $$lcssa553 = 0, $$lcssa554 = 0, $$lcssa555 = 0, $$lcssa556 = 0, $$lcssa557 = 0, $$lcssa558 = 0, $$lcssa559 = 0;
 var $$lcssa560 = 0, $$lcssa561 = 0, $$lcssa562 = 0, $$lcssa563 = 0, $$lcssa564 = 0, $$lcssa567 = 0, $$lcssa568 = 0, $$lcssa569 = 0, $$lcssa570 = 0, $$lcssa578 = 0, $$lcssa579 = 0, $$lcssa580 = 0, $$lcssa581 = 0, $$lcssa582 = 0, $$lcssa583 = 0, $$lcssa584 = 0, $$lcssa585 = 0, $$lcssa586 = 0, $$lcssa587 = 0, $$lcssa588 = 0;
 var $$lcssa589 = 0, $$lcssa590 = 0, $$lcssa591 = 0, $$lcssa592 = 0, $$off$i = 0, $$off$i113 = 0, $$off10$i = 0, $$off11$i = 0, $$off12$i = 0, $$phi$trans$insert = 0, $$pr = 0, $$pr176 = 0, $$pre = 0, $$pre$i = 0, $$pre$i103 = 0, $$pre$i118 = 0, $$pre$i142 = 0, $$pre$i87 = 0, $$pre25$i = 0, $$pre26$i = 0;
 var $$pre313 = 0, $$pre34$i = 0, $$pre58$i = 0, $$pre62$i = 0, $$pre64$i = 0, $$sum17$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0;
 var $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0;
 var $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0;
 var $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0;
 var $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0;
 var $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0;
 var $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0;
 var $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0;
 var $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0;
 var $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0;
 var $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0;
 var $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0;
 var $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0;
 var $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0;
 var $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0;
 var $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0;
 var $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0;
 var $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0;
 var $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0;
 var $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0;
 var $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0;
 var $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0;
 var $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0;
 var $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0;
 var $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0;
 var $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0;
 var $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0;
 var $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0;
 var $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0;
 var $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0;
 var $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0;
 var $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0;
 var $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0;
 var $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0;
 var $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0;
 var $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0;
 var $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0;
 var $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0;
 var $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0;
 var $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0;
 var $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0;
 var $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0;
 var $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $99 = 0, $exitcond$i100 = 0, $exitcond$i97 = 0, $exitcond60$i = 0, $gm$off$i$i = 0, $h$0$lcssa$i$i = 0, $h$0$lcssa$i$i76 = 0;
 var $h$04$i$i = 0, $h$04$i$i107 = 0, $h$04$i$i11 = 0, $h$04$i$i124 = 0, $h$04$i$i149 = 0, $h$04$i$i22 = 0, $h$04$i$i22$1 = 0, $h$04$i$i22$2 = 0, $h$04$i$i22$3 = 0, $h$04$i$i35 = 0, $h$04$i$i45 = 0, $h$04$i$i58 = 0, $h$04$i$i71 = 0, $h$04$i$i88 = 0, $h$04$i10$i = 0, $h$04$i15$i = 0, $h$04$i26$i = 0, $h$04$i37$i = 0, $h$04$i48$i = 0, $h$04$i7$i = 0;
 var $i$0$lcssa$i = 0, $i$0$lcssa$i$i = 0, $i$0$lcssa$i$i77 = 0, $i$01$i = 0, $i$03$i$i = 0, $i$03$i$i108 = 0, $i$03$i$i12 = 0, $i$03$i$i125 = 0, $i$03$i$i150 = 0, $i$03$i$i23 = 0, $i$03$i$i23$1 = 0, $i$03$i$i23$2 = 0, $i$03$i$i23$3 = 0, $i$03$i$i36 = 0, $i$03$i$i46 = 0, $i$03$i$i59 = 0, $i$03$i$i72 = 0, $i$03$i$i89 = 0, $i$03$i11$i = 0, $i$03$i16$i = 0;
 var $i$03$i27$i = 0, $i$03$i38$i = 0, $i$03$i49$i = 0, $i$03$i8$i = 0, $i$032$i = 0, $i$1$lcssa$i = 0, $i$112$i244 = 0, $i$129$i = 0, $i1$0$i = 0, $isdigit$i = 0, $isdigit$i$i = 0, $isdigit$i$i110 = 0, $isdigit$i$i127 = 0, $isdigit$i$i14 = 0, $isdigit$i$i152 = 0, $isdigit$i$i25 = 0, $isdigit$i$i25$1 = 0, $isdigit$i$i25$2 = 0, $isdigit$i$i25$3 = 0, $isdigit$i$i38 = 0;
 var $isdigit$i$i48 = 0, $isdigit$i$i61 = 0, $isdigit$i$i74 = 0, $isdigit$i$i91 = 0, $isdigit$i10$i = 0, $isdigit$i105 = 0, $isdigit$i122 = 0, $isdigit$i13$i = 0, $isdigit$i148 = 0, $isdigit$i18$i = 0, $isdigit$i29$i = 0, $isdigit$i40$i = 0, $isdigit$i51$i = 0, $isdigit2$i$i = 0, $isdigit2$i$i10 = 0, $isdigit2$i$i21 = 0, $isdigit2$i$i21$1 = 0, $isdigit2$i$i21$2 = 0, $isdigit2$i$i21$3 = 0, $isdigit2$i$i34 = 0;
 var $isdigit2$i$i44 = 0, $isdigit2$i$i57 = 0, $isdigit2$i$i70 = 0, $isdigit2$i14$i = 0, $isdigit2$i25$i = 0, $isdigit2$i36$i = 0, $isdigit2$i47$i = 0, $isdigit3$i = 0, $isdigit3$i134 = 0, $isdigit5$i = 0, $isdigit7$i = 0, $isdigit9$i = 0, $isdigittmp$i = 0, $isdigittmp$i$i = 0, $isdigittmp$i$i109 = 0, $isdigittmp$i$i126 = 0, $isdigittmp$i$i13 = 0, $isdigittmp$i$i151 = 0, $isdigittmp$i$i24 = 0, $isdigittmp$i$i24$1 = 0;
 var $isdigittmp$i$i24$2 = 0, $isdigittmp$i$i24$3 = 0, $isdigittmp$i$i37 = 0, $isdigittmp$i$i47 = 0, $isdigittmp$i$i60 = 0, $isdigittmp$i$i73 = 0, $isdigittmp$i$i90 = 0, $isdigittmp$i104 = 0, $isdigittmp$i12$i = 0, $isdigittmp$i121 = 0, $isdigittmp$i147 = 0, $isdigittmp$i17$i = 0, $isdigittmp$i28$i = 0, $isdigittmp$i39$i = 0, $isdigittmp$i50$i = 0, $isdigittmp$i9$i = 0, $isdigittmp1$i$i = 0, $isdigittmp1$i$i20 = 0, $isdigittmp1$i$i20$1 = 0, $isdigittmp1$i$i20$2 = 0;
 var $isdigittmp1$i$i20$3 = 0, $isdigittmp1$i$i33 = 0, $isdigittmp1$i$i43 = 0, $isdigittmp1$i$i56 = 0, $isdigittmp1$i$i69 = 0, $isdigittmp1$i$i9 = 0, $isdigittmp1$i13$i = 0, $isdigittmp1$i24$i = 0, $isdigittmp1$i35$i = 0, $isdigittmp1$i46$i = 0, $isdigittmp2$i = 0, $isdigittmp2$i133 = 0, $isdigittmp4$i = 0, $isdigittmp6$i = 0, $isdigittmp8$i = 0, $memchr$i = 0, $n$0$lcssa$i = 0, $n$037$i = 0, $n$1$i = 0, $name$i = 0;
 var $not$$i = 0, $notes$i = 0, $or$cond = 0, $or$cond$i = 0, $or$cond$i114 = 0, $or$cond$i160 = 0, $or$cond1 = 0, $or$cond2 = 0, $or$cond3 = 0, $or$cond4 = 0, $or$cond5 = 0, $or$cond6 = 0, $or$cond7 = 0, $q$0$lcssa$i = 0, $q$036$i = 0, $q$1$i = 0, $q$2$i = 0, $storemerge$i = 0, $storemerge82$i = 0, $storemerge83$i = 0;
 var $strlenfirst$i = 0, $t$0$i$i = 0, $t$0$i15$i = 0, $t$01$i$i = 0, $t$01$i12$i = 0, $t$02$i$i = 0, $t$02$i14$i = 0, $tno$off$i$i = 0, $tno$off$i$i136 = 0, $tno$off$i$i162 = 0, $tp$0$i$i = 0, $tp$0$i$i137 = 0, $tp$0$i$i165 = 0, $tp$02$i$i = 0, $tp$03$i$i = 0, $tp$06$i$i = 0, $tp$06$i$i163 = 0, $tp$07$i$i = 0, $tp$07$i$i164 = 0, $txt$i$i = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 320|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $txt$i$i = sp + 64|0;
 $name$i = sp + 40|0;
 $notes$i = sp + 16|0;
 $0 = (_strncmp($p,526720,8)|0); //@line 3679 "load_abc.cpp"
 $1 = ($0|0)==(0); //@line 3679 "load_abc.cpp"
 if ($1) {
  $2 = ((($p)) + 8|0); //@line 3679 "load_abc.cpp"
  $3 = HEAP8[$2>>0]|0; //@line 3679 "load_abc.cpp"
  $4 = $3 << 24 >> 24; //@line 3679 "load_abc.cpp"
  $5 = (_isspace($4)|0); //@line 3679 "load_abc.cpp"
  $6 = ($5|0)==(0); //@line 3679 "load_abc.cpp"
  if (!($6)) {
   $$0$i = $2;
   while(1) {
    $7 = HEAP8[$$0$i>>0]|0; //@line 3038 "load_abc.cpp"
    $8 = $7 << 24 >> 24; //@line 3038 "load_abc.cpp"
    $9 = (_isspace($8)|0); //@line 3038 "load_abc.cpp"
    $10 = ($9|0)==(0); //@line 3038 "load_abc.cpp"
    $11 = ((($$0$i)) + 1|0); //@line 3038 "load_abc.cpp"
    if ($10) {
     $$0$i$lcssa = $$0$i;
     break;
    } else {
     $$0$i = $11;
    }
   }
   $12 = HEAP8[$$0$i$lcssa>>0]|0; //@line 1625 "load_abc.cpp"
   $13 = $12 << 24 >> 24; //@line 1625 "load_abc.cpp"
   $isdigittmp1$i$i = (($13) + -48)|0; //@line 1625 "load_abc.cpp"
   $isdigit2$i$i = ($isdigittmp1$i$i>>>0)<(10); //@line 1625 "load_abc.cpp"
   if ($isdigit2$i$i) {
    $16 = $13;$h$04$i$i = 0;$i$03$i$i = 0;
    while(1) {
     $14 = ($h$04$i$i*10)|0; //@line 1626 "load_abc.cpp"
     $15 = (($16) + -48)|0; //@line 1626 "load_abc.cpp"
     $17 = (($15) + ($14))|0; //@line 1626 "load_abc.cpp"
     $18 = (($i$03$i$i) + 1)|0; //@line 1627 "load_abc.cpp"
     $19 = (($$0$i$lcssa) + ($18)|0); //@line 1625 "load_abc.cpp"
     $20 = HEAP8[$19>>0]|0; //@line 1625 "load_abc.cpp"
     $21 = $20 << 24 >> 24; //@line 1625 "load_abc.cpp"
     $isdigittmp$i$i = (($21) + -48)|0; //@line 1625 "load_abc.cpp"
     $isdigit$i$i = ($isdigittmp$i$i>>>0)<(10); //@line 1625 "load_abc.cpp"
     if ($isdigit$i$i) {
      $16 = $21;$h$04$i$i = $17;$i$03$i$i = $18;
     } else {
      $$lcssa591 = $17;$$lcssa592 = $18;
      break;
     }
    }
    $22 = ($$lcssa592|0)==(0); //@line 1629 "load_abc.cpp"
    $$h$0$i$i = $22 ? 1 : $$lcssa591; //@line 1630 "load_abc.cpp"
    $23 = ($$h$0$i$i|0)<(1); //@line 3040 "load_abc.cpp"
    if ($23) {
     $26 = 1;
    } else {
     $24 = ($$h$0$i$i|0)>(128); //@line 3041 "load_abc.cpp"
     $$$h$0$i$i = $24 ? 128 : $$h$0$i$i; //@line 3041 "load_abc.cpp"
     $26 = $$$h$0$i$i;
    }
   } else {
    $26 = 1;
   }
   $25 = (($26) + 1)|0; //@line 3679 "load_abc.cpp"
   $27 = ((($h)) + 372|0); //@line 3679 "load_abc.cpp"
   HEAP8[$27>>0]=$25&255;HEAP8[$27+1>>0]=($25>>8)&255;HEAP8[$27+2>>0]=($25>>16)&255;HEAP8[$27+3>>0]=$25>>24; //@line 3679 "load_abc.cpp"
  }
 }
 $28 = (_strncmp($p,526729,7)|0); //@line 3680 "load_abc.cpp"
 $29 = ($28|0)==(0); //@line 3680 "load_abc.cpp"
 if ($29) {
  $30 = ((($p)) + 7|0); //@line 3680 "load_abc.cpp"
  $31 = HEAP8[$30>>0]|0; //@line 3680 "load_abc.cpp"
  $32 = $31 << 24 >> 24; //@line 3680 "load_abc.cpp"
  $33 = (_isspace($32)|0); //@line 3680 "load_abc.cpp"
  $34 = ($33|0)==(0); //@line 3680 "load_abc.cpp"
  if (!($34)) {
   $$0$i8 = $30;
   while(1) {
    $35 = HEAP8[$$0$i8>>0]|0; //@line 3028 "load_abc.cpp"
    $36 = $35 << 24 >> 24; //@line 3028 "load_abc.cpp"
    $37 = (_isspace($36)|0); //@line 3028 "load_abc.cpp"
    $38 = ($37|0)==(0); //@line 3028 "load_abc.cpp"
    $39 = ((($$0$i8)) + 1|0); //@line 3028 "load_abc.cpp"
    if ($38) {
     $$0$i8$lcssa = $$0$i8;
     break;
    } else {
     $$0$i8 = $39;
    }
   }
   $40 = HEAP8[$$0$i8$lcssa>>0]|0; //@line 1625 "load_abc.cpp"
   $41 = $40 << 24 >> 24; //@line 1625 "load_abc.cpp"
   $isdigittmp1$i$i9 = (($41) + -48)|0; //@line 1625 "load_abc.cpp"
   $isdigit2$i$i10 = ($isdigittmp1$i$i9>>>0)<(10); //@line 1625 "load_abc.cpp"
   if ($isdigit2$i$i10) {
    $44 = $41;$h$04$i$i11 = 0;$i$03$i$i12 = 0;
    while(1) {
     $42 = ($h$04$i$i11*10)|0; //@line 1626 "load_abc.cpp"
     $43 = (($44) + -48)|0; //@line 1626 "load_abc.cpp"
     $45 = (($43) + ($42))|0; //@line 1626 "load_abc.cpp"
     $46 = (($i$03$i$i12) + 1)|0; //@line 1627 "load_abc.cpp"
     $47 = (($$0$i8$lcssa) + ($46)|0); //@line 1625 "load_abc.cpp"
     $48 = HEAP8[$47>>0]|0; //@line 1625 "load_abc.cpp"
     $49 = $48 << 24 >> 24; //@line 1625 "load_abc.cpp"
     $isdigittmp$i$i13 = (($49) + -48)|0; //@line 1625 "load_abc.cpp"
     $isdigit$i$i14 = ($isdigittmp$i$i13>>>0)<(10); //@line 1625 "load_abc.cpp"
     if ($isdigit$i$i14) {
      $44 = $49;$h$04$i$i11 = $45;$i$03$i$i12 = $46;
     } else {
      $$lcssa589 = $45;$$lcssa590 = $46;
      break;
     }
    }
    $50 = ($$lcssa590|0)==(0); //@line 1629 "load_abc.cpp"
    $$h$0$i$i16 = $50 ? 1 : $$lcssa589; //@line 1630 "load_abc.cpp"
    $51 = ($$h$0$i$i16|0)<(0); //@line 3030 "load_abc.cpp"
    if ($51) {
     $54 = 0;
    } else {
     $52 = ($$h$0$i$i16|0)>(127); //@line 3031 "load_abc.cpp"
     $$$h$0$i$i18 = $52 ? 127 : $$h$0$i$i16; //@line 3031 "load_abc.cpp"
     $54 = $$$h$0$i$i18;
    }
   } else {
    $54 = 1;
   }
   $53 = ((($h)) + 368|0); //@line 3680 "load_abc.cpp"
   HEAP8[$53>>0]=$54&255;HEAP8[$53+1>>0]=($54>>8)&255;HEAP8[$53+2>>0]=($54>>16)&255;HEAP8[$53+3>>0]=$54>>24; //@line 3680 "load_abc.cpp"
  }
 }
 $55 = (_strncmp($p,526737,4)|0); //@line 3681 "load_abc.cpp"
 $56 = ($55|0)==(0); //@line 3681 "load_abc.cpp"
 if ($56) {
  $57 = ((($p)) + 4|0); //@line 3681 "load_abc.cpp"
  $58 = HEAP8[$57>>0]|0; //@line 3681 "load_abc.cpp"
  $59 = $58 << 24 >> 24; //@line 3681 "load_abc.cpp"
  $60 = (_isspace($59)|0); //@line 3681 "load_abc.cpp"
  $61 = ($60|0)==(0); //@line 3681 "load_abc.cpp"
  if (!($61)) {
   $62 = ((($h)) + 32|0); //@line 3507 "load_abc.cpp"
   HEAP8[$62>>0] = 127; //@line 3507 "load_abc.cpp"
   $63 = ((($h)) + 33|0); //@line 3508 "load_abc.cpp"
   HEAP8[$63>>0] = 125; //@line 3508 "load_abc.cpp"
   $64 = ((($h)) + 34|0); //@line 3509 "load_abc.cpp"
   HEAP8[$64>>0] = 110; //@line 3509 "load_abc.cpp"
   $65 = ((($h)) + 35|0); //@line 3510 "load_abc.cpp"
   HEAP8[$65>>0] = 1; //@line 3510 "load_abc.cpp"
   $$1$i = $57;
   while(1) {
    $66 = HEAP8[$$1$i>>0]|0; //@line 3512 "load_abc.cpp"
    $67 = $66 << 24 >> 24; //@line 3512 "load_abc.cpp"
    $68 = (_isspace($67)|0); //@line 3512 "load_abc.cpp"
    $69 = ($68|0)==(0); //@line 3512 "load_abc.cpp"
    $70 = ((($$1$i)) + 1|0); //@line 3512 "load_abc.cpp"
    if ($69) {
     $$1$i$lcssa = $$1$i;
     break;
    } else {
     $$1$i = $70;
    }
   }
   $71 = HEAP8[$$1$i$lcssa>>0]|0; //@line 3513 "load_abc.cpp"
   $72 = ($71<<24>>24)==(0); //@line 3513 "load_abc.cpp"
   if ($72) {
    $$1$i$1 = $$1$i$lcssa;
   } else {
    $73 = $71 << 24 >> 24; //@line 1625 "load_abc.cpp"
    $isdigittmp1$i$i20 = (($73) + -48)|0; //@line 1625 "load_abc.cpp"
    $isdigit2$i$i21 = ($isdigittmp1$i$i20>>>0)<(10); //@line 1625 "load_abc.cpp"
    if ($isdigit2$i$i21) {
     $76 = $73;$h$04$i$i22 = 0;$i$03$i$i23 = 0;
     while(1) {
      $74 = ($h$04$i$i22*10)|0; //@line 1626 "load_abc.cpp"
      $75 = (($76) + -48)|0; //@line 1626 "load_abc.cpp"
      $77 = (($75) + ($74))|0; //@line 1626 "load_abc.cpp"
      $78 = (($i$03$i$i23) + 1)|0; //@line 1627 "load_abc.cpp"
      $79 = (($$1$i$lcssa) + ($78)|0); //@line 1625 "load_abc.cpp"
      $80 = HEAP8[$79>>0]|0; //@line 1625 "load_abc.cpp"
      $81 = $80 << 24 >> 24; //@line 1625 "load_abc.cpp"
      $isdigittmp$i$i24 = (($81) + -48)|0; //@line 1625 "load_abc.cpp"
      $isdigit$i$i25 = ($isdigittmp$i$i24>>>0)<(10); //@line 1625 "load_abc.cpp"
      if ($isdigit$i$i25) {
       $76 = $81;$h$04$i$i22 = $77;$i$03$i$i23 = $78;
      } else {
       $$lcssa586 = $77;$$lcssa587 = $78;$$lcssa588 = $79;
       break;
      }
     }
     $82 = ($$lcssa587|0)==(0); //@line 1629 "load_abc.cpp"
     $$h$0$i$i27 = $82 ? 1 : $$lcssa586; //@line 1630 "load_abc.cpp"
     $83 = ($$h$0$i$i27|0)<(0); //@line 3515 "load_abc.cpp"
     if ($83) {
      $86 = 0;$983 = $$lcssa588;
     } else {
      $84 = ($$h$0$i$i27|0)>(127); //@line 3516 "load_abc.cpp"
      $$$h$0$i$i29 = $84 ? 127 : $$h$0$i$i27; //@line 3516 "load_abc.cpp"
      $86 = $$$h$0$i$i29;$983 = $$lcssa588;
     }
    } else {
     $86 = 1;$983 = $$1$i$lcssa;
    }
    $85 = $86&255; //@line 3517 "load_abc.cpp"
    HEAP8[$62>>0] = $85; //@line 3517 "load_abc.cpp"
    $$1$i$1 = $983;
   }
   while(1) {
    $918 = HEAP8[$$1$i$1>>0]|0; //@line 3512 "load_abc.cpp"
    $919 = $918 << 24 >> 24; //@line 3512 "load_abc.cpp"
    $920 = (_isspace($919)|0); //@line 3512 "load_abc.cpp"
    $921 = ($920|0)==(0); //@line 3512 "load_abc.cpp"
    $922 = ((($$1$i$1)) + 1|0); //@line 3512 "load_abc.cpp"
    if ($921) {
     $$1$i$1$lcssa = $$1$i$1;
     break;
    } else {
     $$1$i$1 = $922;
    }
   }
   $923 = HEAP8[$$1$i$1$lcssa>>0]|0; //@line 3513 "load_abc.cpp"
   $924 = ($923<<24>>24)==(0); //@line 3513 "load_abc.cpp"
   if ($924) {
    $$1$i$2 = $$1$i$1$lcssa;
   } else {
    $925 = $923 << 24 >> 24; //@line 1625 "load_abc.cpp"
    $isdigittmp1$i$i20$1 = (($925) + -48)|0; //@line 1625 "load_abc.cpp"
    $isdigit2$i$i21$1 = ($isdigittmp1$i$i20$1>>>0)<(10); //@line 1625 "load_abc.cpp"
    if ($isdigit2$i$i21$1) {
     $928 = $925;$h$04$i$i22$1 = 0;$i$03$i$i23$1 = 0;
     while(1) {
      $926 = ($h$04$i$i22$1*10)|0; //@line 1626 "load_abc.cpp"
      $927 = (($928) + -48)|0; //@line 1626 "load_abc.cpp"
      $929 = (($927) + ($926))|0; //@line 1626 "load_abc.cpp"
      $930 = (($i$03$i$i23$1) + 1)|0; //@line 1627 "load_abc.cpp"
      $931 = (($$1$i$1$lcssa) + ($930)|0); //@line 1625 "load_abc.cpp"
      $932 = HEAP8[$931>>0]|0; //@line 1625 "load_abc.cpp"
      $933 = $932 << 24 >> 24; //@line 1625 "load_abc.cpp"
      $isdigittmp$i$i24$1 = (($933) + -48)|0; //@line 1625 "load_abc.cpp"
      $isdigit$i$i25$1 = ($isdigittmp$i$i24$1>>>0)<(10); //@line 1625 "load_abc.cpp"
      if ($isdigit$i$i25$1) {
       $928 = $933;$h$04$i$i22$1 = $929;$i$03$i$i23$1 = $930;
      } else {
       $$lcssa583 = $929;$$lcssa584 = $930;$$lcssa585 = $931;
       break;
      }
     }
     $934 = ($$lcssa584|0)==(0); //@line 1629 "load_abc.cpp"
     $$h$0$i$i27$1 = $934 ? 1 : $$lcssa583; //@line 1630 "load_abc.cpp"
     $935 = ($$h$0$i$i27$1|0)<(0); //@line 3515 "load_abc.cpp"
     if ($935) {
      $938 = 0;$986 = $$lcssa585;
     } else {
      $936 = ($$h$0$i$i27$1|0)>(127); //@line 3516 "load_abc.cpp"
      $$$h$0$i$i29$1 = $936 ? 127 : $$h$0$i$i27$1; //@line 3516 "load_abc.cpp"
      $938 = $$$h$0$i$i29$1;$986 = $$lcssa585;
     }
    } else {
     $938 = 1;$986 = $$1$i$1$lcssa;
    }
    $937 = $938&255; //@line 3517 "load_abc.cpp"
    HEAP8[$63>>0] = $937; //@line 3517 "load_abc.cpp"
    $$1$i$2 = $986;
   }
   while(1) {
    $939 = HEAP8[$$1$i$2>>0]|0; //@line 3512 "load_abc.cpp"
    $940 = $939 << 24 >> 24; //@line 3512 "load_abc.cpp"
    $941 = (_isspace($940)|0); //@line 3512 "load_abc.cpp"
    $942 = ($941|0)==(0); //@line 3512 "load_abc.cpp"
    $943 = ((($$1$i$2)) + 1|0); //@line 3512 "load_abc.cpp"
    if ($942) {
     $$1$i$2$lcssa = $$1$i$2;
     break;
    } else {
     $$1$i$2 = $943;
    }
   }
   $944 = HEAP8[$$1$i$2$lcssa>>0]|0; //@line 3513 "load_abc.cpp"
   $945 = ($944<<24>>24)==(0); //@line 3513 "load_abc.cpp"
   if ($945) {
    $$1$i$3 = $$1$i$2$lcssa;
   } else {
    $946 = $944 << 24 >> 24; //@line 1625 "load_abc.cpp"
    $isdigittmp1$i$i20$2 = (($946) + -48)|0; //@line 1625 "load_abc.cpp"
    $isdigit2$i$i21$2 = ($isdigittmp1$i$i20$2>>>0)<(10); //@line 1625 "load_abc.cpp"
    if ($isdigit2$i$i21$2) {
     $949 = $946;$h$04$i$i22$2 = 0;$i$03$i$i23$2 = 0;
     while(1) {
      $947 = ($h$04$i$i22$2*10)|0; //@line 1626 "load_abc.cpp"
      $948 = (($949) + -48)|0; //@line 1626 "load_abc.cpp"
      $950 = (($948) + ($947))|0; //@line 1626 "load_abc.cpp"
      $951 = (($i$03$i$i23$2) + 1)|0; //@line 1627 "load_abc.cpp"
      $952 = (($$1$i$2$lcssa) + ($951)|0); //@line 1625 "load_abc.cpp"
      $953 = HEAP8[$952>>0]|0; //@line 1625 "load_abc.cpp"
      $954 = $953 << 24 >> 24; //@line 1625 "load_abc.cpp"
      $isdigittmp$i$i24$2 = (($954) + -48)|0; //@line 1625 "load_abc.cpp"
      $isdigit$i$i25$2 = ($isdigittmp$i$i24$2>>>0)<(10); //@line 1625 "load_abc.cpp"
      if ($isdigit$i$i25$2) {
       $949 = $954;$h$04$i$i22$2 = $950;$i$03$i$i23$2 = $951;
      } else {
       $$lcssa580 = $950;$$lcssa581 = $951;$$lcssa582 = $952;
       break;
      }
     }
     $955 = ($$lcssa581|0)==(0); //@line 1629 "load_abc.cpp"
     $$h$0$i$i27$2 = $955 ? 1 : $$lcssa580; //@line 1630 "load_abc.cpp"
     $956 = ($$h$0$i$i27$2|0)<(0); //@line 3515 "load_abc.cpp"
     if ($956) {
      $959 = 0;$987 = $$lcssa582;
     } else {
      $957 = ($$h$0$i$i27$2|0)>(127); //@line 3516 "load_abc.cpp"
      $$$h$0$i$i29$2 = $957 ? 127 : $$h$0$i$i27$2; //@line 3516 "load_abc.cpp"
      $959 = $$$h$0$i$i29$2;$987 = $$lcssa582;
     }
    } else {
     $959 = 1;$987 = $$1$i$2$lcssa;
    }
    $958 = $959&255; //@line 3517 "load_abc.cpp"
    HEAP8[$64>>0] = $958; //@line 3517 "load_abc.cpp"
    $$1$i$3 = $987;
   }
   while(1) {
    $960 = HEAP8[$$1$i$3>>0]|0; //@line 3512 "load_abc.cpp"
    $961 = $960 << 24 >> 24; //@line 3512 "load_abc.cpp"
    $962 = (_isspace($961)|0); //@line 3512 "load_abc.cpp"
    $963 = ($962|0)==(0); //@line 3512 "load_abc.cpp"
    $964 = ((($$1$i$3)) + 1|0); //@line 3512 "load_abc.cpp"
    if ($963) {
     $$1$i$3$lcssa = $$1$i$3;
     break;
    } else {
     $$1$i$3 = $964;
    }
   }
   $965 = HEAP8[$$1$i$3$lcssa>>0]|0; //@line 3513 "load_abc.cpp"
   $966 = ($965<<24>>24)==(0); //@line 3513 "load_abc.cpp"
   if ($966) {
    $$pre = HEAP8[$65>>0]|0; //@line 3520 "load_abc.cpp"
    $982 = $$pre;
   } else {
    $967 = $965 << 24 >> 24; //@line 1625 "load_abc.cpp"
    $isdigittmp1$i$i20$3 = (($967) + -48)|0; //@line 1625 "load_abc.cpp"
    $isdigit2$i$i21$3 = ($isdigittmp1$i$i20$3>>>0)<(10); //@line 1625 "load_abc.cpp"
    if ($isdigit2$i$i21$3) {
     $970 = $967;$h$04$i$i22$3 = 0;$i$03$i$i23$3 = 0;
     while(1) {
      $968 = ($h$04$i$i22$3*10)|0; //@line 1626 "load_abc.cpp"
      $969 = (($970) + -48)|0; //@line 1626 "load_abc.cpp"
      $971 = (($969) + ($968))|0; //@line 1626 "load_abc.cpp"
      $972 = (($i$03$i$i23$3) + 1)|0; //@line 1627 "load_abc.cpp"
      $973 = (($$1$i$3$lcssa) + ($972)|0); //@line 1625 "load_abc.cpp"
      $974 = HEAP8[$973>>0]|0; //@line 1625 "load_abc.cpp"
      $975 = $974 << 24 >> 24; //@line 1625 "load_abc.cpp"
      $isdigittmp$i$i24$3 = (($975) + -48)|0; //@line 1625 "load_abc.cpp"
      $isdigit$i$i25$3 = ($isdigittmp$i$i24$3>>>0)<(10); //@line 1625 "load_abc.cpp"
      if ($isdigit$i$i25$3) {
       $970 = $975;$h$04$i$i22$3 = $971;$i$03$i$i23$3 = $972;
      } else {
       $$lcssa578 = $971;$$lcssa579 = $972;
       break;
      }
     }
     $976 = ($$lcssa579|0)==(0); //@line 1629 "load_abc.cpp"
     $$h$0$i$i27$3 = $976 ? 1 : $$lcssa578; //@line 1630 "load_abc.cpp"
     $977 = ($$h$0$i$i27$3|0)<(0); //@line 3515 "load_abc.cpp"
     if ($977) {
      $980 = 0;
     } else {
      $978 = ($$h$0$i$i27$3|0)>(127); //@line 3516 "load_abc.cpp"
      $$$h$0$i$i29$3 = $978 ? 127 : $$h$0$i$i27$3; //@line 3516 "load_abc.cpp"
      $980 = $$$h$0$i$i29$3;
     }
    } else {
     $980 = 1;
    }
    $979 = $980&255; //@line 3517 "load_abc.cpp"
    HEAP8[$65>>0] = $979; //@line 3517 "load_abc.cpp"
    $982 = $979;
   }
   $981 = ($982<<24>>24)==(0); //@line 3520 "load_abc.cpp"
   if ($981) {
    HEAP8[$65>>0] = 1; //@line 3520 "load_abc.cpp"
   }
  }
 }
 $87 = (_strncmp($p,526742,10)|0); //@line 3682 "load_abc.cpp"
 $88 = ($87|0)==(0); //@line 3682 "load_abc.cpp"
 if ($88) {
  $89 = ((($p)) + 10|0); //@line 3682 "load_abc.cpp"
  $90 = HEAP8[$89>>0]|0; //@line 3682 "load_abc.cpp"
  $91 = $90 << 24 >> 24; //@line 3682 "load_abc.cpp"
  $92 = (_isspace($91)|0); //@line 3682 "load_abc.cpp"
  $93 = ($92|0)==(0); //@line 3682 "load_abc.cpp"
  if (!($93)) {
   $94 = ((($p)) + 4|0); //@line 3682 "load_abc.cpp"
   $$0$i31 = $94;
   while(1) {
    $95 = HEAP8[$$0$i31>>0]|0; //@line 3535 "load_abc.cpp"
    $96 = $95 << 24 >> 24; //@line 3535 "load_abc.cpp"
    $97 = (_isspace($96)|0); //@line 3535 "load_abc.cpp"
    $98 = ($97|0)==(0); //@line 3535 "load_abc.cpp"
    $99 = ((($$0$i31)) + 1|0); //@line 3535 "load_abc.cpp"
    if ($98) {
     $$0$i31$lcssa = $$0$i31;
     break;
    } else {
     $$0$i31 = $99;
    }
   }
   $100 = ((($h)) + 28|0); //@line 3536 "load_abc.cpp"
   $101 = HEAPU8[$100>>0]|(HEAPU8[$100+1>>0]<<8)|(HEAPU8[$100+2>>0]<<16)|(HEAPU8[$100+3>>0]<<24); //@line 3536 "load_abc.cpp"
   $102 = ($101|0)==(0|0); //@line 3536 "load_abc.cpp"
   if (!($102)) {
    _free($101); //@line 3536 "load_abc.cpp"
   }
   $strlenfirst$i = HEAP8[$$0$i31$lcssa>>0]|0; //@line 3537 "load_abc.cpp"
   $103 = ($strlenfirst$i<<24>>24)==(0); //@line 3537 "load_abc.cpp"
   if ($103) {
    $storemerge$i = 0;
   } else {
    $104 = (___strdup($$0$i31$lcssa)|0); //@line 3538 "load_abc.cpp"
    $storemerge$i = $104;
   }
   HEAP8[$100>>0]=$storemerge$i&255;HEAP8[$100+1>>0]=($storemerge$i>>8)&255;HEAP8[$100+2>>0]=($storemerge$i>>16)&255;HEAP8[$100+3>>0]=$storemerge$i>>24; //@line 3538 "load_abc.cpp"
  }
 }
 $105 = (_strncmp($p,526753,9)|0); //@line 3683 "load_abc.cpp"
 $106 = ($105|0)==(0); //@line 3683 "load_abc.cpp"
 if ($106) {
  $107 = ((($p)) + 9|0); //@line 3683 "load_abc.cpp"
  $108 = HEAP8[$107>>0]|0; //@line 3683 "load_abc.cpp"
  $109 = $108 << 24 >> 24; //@line 3683 "load_abc.cpp"
  $110 = (_isspace($109)|0); //@line 3683 "load_abc.cpp"
  $111 = ($110|0)==(0); //@line 3683 "load_abc.cpp"
  if (!($111)) {
   $112 = HEAP8[$107>>0]|0; //@line 3179 "load_abc.cpp"
   $113 = ($112<<24>>24)==(0); //@line 3179 "load_abc.cpp"
   L92: do {
    if ($113) {
     $$1$lcssa$i = $107;$i$0$lcssa$i = 0;
    } else {
     $$015$i = $107;$115 = $112;
     while(1) {
      $114 = $115 << 24 >> 24; //@line 3179 "load_abc.cpp"
      $116 = (_isspace($114)|0); //@line 3179 "load_abc.cpp"
      $117 = ($116|0)==(0); //@line 3179 "load_abc.cpp"
      if ($117) {
       $$015$i$lcssa = $$015$i;
       break;
      }
      $118 = ((($$015$i)) + 1|0); //@line 3179 "load_abc.cpp"
      $119 = HEAP8[$118>>0]|0; //@line 3179 "load_abc.cpp"
      $120 = ($119<<24>>24)==(0); //@line 3179 "load_abc.cpp"
      if ($120) {
       $$1$lcssa$i = $118;$i$0$lcssa$i = 0;
       break L92;
      } else {
       $$015$i = $118;$115 = $119;
      }
     }
     $$pre$i = HEAP8[$$015$i$lcssa>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($$pre$i<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $$015$i$lcssa;$i$0$lcssa$i = 0;
      break L92;
      break;
     }
     default: {
     }
     }
     HEAP8[$name$i>>0] = $$pre$i; //@line 3182 "load_abc.cpp"
     $121 = ((($$015$i$lcssa)) + 1|0); //@line 3183 "load_abc.cpp"
     $122 = HEAP8[$121>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($122<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $121;$i$0$lcssa$i = 1;
      break L92;
      break;
     }
     default: {
     }
     }
     $159 = ((($name$i)) + 1|0); //@line 3182 "load_abc.cpp"
     HEAP8[$159>>0] = $122; //@line 3182 "load_abc.cpp"
     $160 = ((($$015$i$lcssa)) + 2|0); //@line 3183 "load_abc.cpp"
     $161 = HEAP8[$160>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($161<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $160;$i$0$lcssa$i = 2;
      break L92;
      break;
     }
     default: {
     }
     }
     $162 = ((($name$i)) + 2|0); //@line 3182 "load_abc.cpp"
     HEAP8[$162>>0] = $161; //@line 3182 "load_abc.cpp"
     $163 = ((($$015$i$lcssa)) + 3|0); //@line 3183 "load_abc.cpp"
     $164 = HEAP8[$163>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($164<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $163;$i$0$lcssa$i = 3;
      break L92;
      break;
     }
     default: {
     }
     }
     $165 = ((($name$i)) + 3|0); //@line 3182 "load_abc.cpp"
     HEAP8[$165>>0] = $164; //@line 3182 "load_abc.cpp"
     $166 = ((($$015$i$lcssa)) + 4|0); //@line 3183 "load_abc.cpp"
     $167 = HEAP8[$166>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($167<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $166;$i$0$lcssa$i = 4;
      break L92;
      break;
     }
     default: {
     }
     }
     $168 = ((($name$i)) + 4|0); //@line 3182 "load_abc.cpp"
     HEAP8[$168>>0] = $167; //@line 3182 "load_abc.cpp"
     $169 = ((($$015$i$lcssa)) + 5|0); //@line 3183 "load_abc.cpp"
     $170 = HEAP8[$169>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($170<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $169;$i$0$lcssa$i = 5;
      break L92;
      break;
     }
     default: {
     }
     }
     $171 = ((($name$i)) + 5|0); //@line 3182 "load_abc.cpp"
     HEAP8[$171>>0] = $170; //@line 3182 "load_abc.cpp"
     $172 = ((($$015$i$lcssa)) + 6|0); //@line 3183 "load_abc.cpp"
     $173 = HEAP8[$172>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($173<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $172;$i$0$lcssa$i = 6;
      break L92;
      break;
     }
     default: {
     }
     }
     $174 = ((($name$i)) + 6|0); //@line 3182 "load_abc.cpp"
     HEAP8[$174>>0] = $173; //@line 3182 "load_abc.cpp"
     $175 = ((($$015$i$lcssa)) + 7|0); //@line 3183 "load_abc.cpp"
     $176 = HEAP8[$175>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($176<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $175;$i$0$lcssa$i = 7;
      break L92;
      break;
     }
     default: {
     }
     }
     $177 = ((($name$i)) + 7|0); //@line 3182 "load_abc.cpp"
     HEAP8[$177>>0] = $176; //@line 3182 "load_abc.cpp"
     $178 = ((($$015$i$lcssa)) + 8|0); //@line 3183 "load_abc.cpp"
     $179 = HEAP8[$178>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($179<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $178;$i$0$lcssa$i = 8;
      break L92;
      break;
     }
     default: {
     }
     }
     $180 = ((($name$i)) + 8|0); //@line 3182 "load_abc.cpp"
     HEAP8[$180>>0] = $179; //@line 3182 "load_abc.cpp"
     $181 = ((($$015$i$lcssa)) + 9|0); //@line 3183 "load_abc.cpp"
     $182 = HEAP8[$181>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($182<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $181;$i$0$lcssa$i = 9;
      break L92;
      break;
     }
     default: {
     }
     }
     $183 = ((($name$i)) + 9|0); //@line 3182 "load_abc.cpp"
     HEAP8[$183>>0] = $182; //@line 3182 "load_abc.cpp"
     $184 = ((($$015$i$lcssa)) + 10|0); //@line 3183 "load_abc.cpp"
     $185 = HEAP8[$184>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($185<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $184;$i$0$lcssa$i = 10;
      break L92;
      break;
     }
     default: {
     }
     }
     $186 = ((($name$i)) + 10|0); //@line 3182 "load_abc.cpp"
     HEAP8[$186>>0] = $185; //@line 3182 "load_abc.cpp"
     $187 = ((($$015$i$lcssa)) + 11|0); //@line 3183 "load_abc.cpp"
     $188 = HEAP8[$187>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($188<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $187;$i$0$lcssa$i = 11;
      break L92;
      break;
     }
     default: {
     }
     }
     $189 = ((($name$i)) + 11|0); //@line 3182 "load_abc.cpp"
     HEAP8[$189>>0] = $188; //@line 3182 "load_abc.cpp"
     $190 = ((($$015$i$lcssa)) + 12|0); //@line 3183 "load_abc.cpp"
     $191 = HEAP8[$190>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($191<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $190;$i$0$lcssa$i = 12;
      break L92;
      break;
     }
     default: {
     }
     }
     $192 = ((($name$i)) + 12|0); //@line 3182 "load_abc.cpp"
     HEAP8[$192>>0] = $191; //@line 3182 "load_abc.cpp"
     $193 = ((($$015$i$lcssa)) + 13|0); //@line 3183 "load_abc.cpp"
     $194 = HEAP8[$193>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($194<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $193;$i$0$lcssa$i = 13;
      break L92;
      break;
     }
     default: {
     }
     }
     $195 = ((($name$i)) + 13|0); //@line 3182 "load_abc.cpp"
     HEAP8[$195>>0] = $194; //@line 3182 "load_abc.cpp"
     $196 = ((($$015$i$lcssa)) + 14|0); //@line 3183 "load_abc.cpp"
     $197 = HEAP8[$196>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($197<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $196;$i$0$lcssa$i = 14;
      break L92;
      break;
     }
     default: {
     }
     }
     $198 = ((($name$i)) + 14|0); //@line 3182 "load_abc.cpp"
     HEAP8[$198>>0] = $197; //@line 3182 "load_abc.cpp"
     $199 = ((($$015$i$lcssa)) + 15|0); //@line 3183 "load_abc.cpp"
     $200 = HEAP8[$199>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($200<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $199;$i$0$lcssa$i = 15;
      break L92;
      break;
     }
     default: {
     }
     }
     $201 = ((($name$i)) + 15|0); //@line 3182 "load_abc.cpp"
     HEAP8[$201>>0] = $200; //@line 3182 "load_abc.cpp"
     $202 = ((($$015$i$lcssa)) + 16|0); //@line 3183 "load_abc.cpp"
     $203 = HEAP8[$202>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($203<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $202;$i$0$lcssa$i = 16;
      break L92;
      break;
     }
     default: {
     }
     }
     $204 = ((($name$i)) + 16|0); //@line 3182 "load_abc.cpp"
     HEAP8[$204>>0] = $203; //@line 3182 "load_abc.cpp"
     $205 = ((($$015$i$lcssa)) + 17|0); //@line 3183 "load_abc.cpp"
     $206 = HEAP8[$205>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($206<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $205;$i$0$lcssa$i = 17;
      break L92;
      break;
     }
     default: {
     }
     }
     $207 = ((($name$i)) + 17|0); //@line 3182 "load_abc.cpp"
     HEAP8[$207>>0] = $206; //@line 3182 "load_abc.cpp"
     $208 = ((($$015$i$lcssa)) + 18|0); //@line 3183 "load_abc.cpp"
     $209 = HEAP8[$208>>0]|0; //@line 3181 "load_abc.cpp"
     switch ($209<<24>>24) {
     case 0: case 32:  {
      $$1$lcssa$i = $208;$i$0$lcssa$i = 18;
      break L92;
      break;
     }
     default: {
     }
     }
     $210 = ((($name$i)) + 18|0); //@line 3182 "load_abc.cpp"
     HEAP8[$210>>0] = $209; //@line 3182 "load_abc.cpp"
     $211 = ((($$015$i$lcssa)) + 19|0); //@line 3183 "load_abc.cpp"
     $$1$lcssa$i = $211;$i$0$lcssa$i = 19;
    }
   } while(0);
   $123 = (($name$i) + ($i$0$lcssa$i)|0); //@line 3186 "load_abc.cpp"
   HEAP8[$123>>0] = 0; //@line 3186 "load_abc.cpp"
   $124 = HEAP8[$$1$lcssa$i>>0]|0; //@line 3187 "load_abc.cpp"
   $125 = ($124<<24>>24)==(32); //@line 3187 "load_abc.cpp"
   if ($125) {
    $126 = (_isspace(32)|0); //@line 3192 "load_abc.cpp"
    $127 = ($126|0)==(0); //@line 3192 "load_abc.cpp"
    L119: do {
     if ($127) {
      $i$1$lcssa$i = 0;
     } else {
      $$211$i245 = $$1$lcssa$i;$i$112$i244 = 0;
      while(1) {
       $132 = HEAP8[$$211$i245>>0]|0; //@line 3193 "load_abc.cpp"
       $133 = ($132<<24>>24)==(0); //@line 3193 "load_abc.cpp"
       L122: do {
        if ($133) {
         $$3$lcssa$i = $$211$i245;$143 = 0;
        } else {
         $$38$i = $$211$i245;$135 = $132;
         while(1) {
          $134 = $135 << 24 >> 24; //@line 3193 "load_abc.cpp"
          $136 = (_isspace($134)|0); //@line 3193 "load_abc.cpp"
          $137 = ($136|0)==(0); //@line 3193 "load_abc.cpp"
          if ($137) {
           $$38$i$lcssa = $$38$i;
           break;
          }
          $138 = ((($$38$i)) + 1|0); //@line 3193 "load_abc.cpp"
          $139 = HEAP8[$138>>0]|0; //@line 3193 "load_abc.cpp"
          $140 = ($139<<24>>24)==(0); //@line 3193 "load_abc.cpp"
          if ($140) {
           $$3$lcssa$i = $138;$143 = 0;
           break L122;
          } else {
           $$38$i = $138;$135 = $139;
          }
         }
         $$pre26$i = HEAP8[$$38$i$lcssa>>0]|0; //@line 1625 "load_abc.cpp"
         $$3$lcssa$i = $$38$i$lcssa;$143 = $$pre26$i;
        }
       } while(0);
       $141 = (($notes$i) + ($i$112$i244<<2)|0); //@line 3194 "load_abc.cpp"
       $142 = $143 << 24 >> 24; //@line 1625 "load_abc.cpp"
       $isdigittmp1$i$i33 = (($142) + -48)|0; //@line 1625 "load_abc.cpp"
       $isdigit2$i$i34 = ($isdigittmp1$i$i33>>>0)<(10); //@line 1625 "load_abc.cpp"
       if ($isdigit2$i$i34) {
        $146 = $142;$h$04$i$i35 = 0;$i$03$i$i36 = 0;
        while(1) {
         $144 = ($h$04$i$i35*10)|0; //@line 1626 "load_abc.cpp"
         $145 = (($146) + -48)|0; //@line 1626 "load_abc.cpp"
         $147 = (($145) + ($144))|0; //@line 1626 "load_abc.cpp"
         $148 = (($i$03$i$i36) + 1)|0; //@line 1627 "load_abc.cpp"
         $149 = (($$3$lcssa$i) + ($148)|0); //@line 1625 "load_abc.cpp"
         $150 = HEAP8[$149>>0]|0; //@line 1625 "load_abc.cpp"
         $151 = $150 << 24 >> 24; //@line 1625 "load_abc.cpp"
         $isdigittmp$i$i37 = (($151) + -48)|0; //@line 1625 "load_abc.cpp"
         $isdigit$i$i38 = ($isdigittmp$i$i37>>>0)<(10); //@line 1625 "load_abc.cpp"
         if ($isdigit$i$i38) {
          $146 = $151;$h$04$i$i35 = $147;$i$03$i$i36 = $148;
         } else {
          $h$0$lcssa$i$i = $147;$i$0$lcssa$i$i = $148;
          break;
         }
        }
       } else {
        $h$0$lcssa$i$i = 0;$i$0$lcssa$i$i = 0;
       }
       $152 = ($i$0$lcssa$i$i|0)==(0); //@line 1629 "load_abc.cpp"
       $$h$0$i$i40 = $152 ? 1 : $h$0$lcssa$i$i; //@line 1630 "load_abc.cpp"
       HEAP32[$141>>2] = $$h$0$i$i40; //@line 1630 "load_abc.cpp"
       $153 = (($$3$lcssa$i) + ($i$0$lcssa$i$i)|0); //@line 3194 "load_abc.cpp"
       $154 = (($i$112$i244) + 1)|0; //@line 3195 "load_abc.cpp"
       $155 = ($154|0)<(7); //@line 3192 "load_abc.cpp"
       if (!($155)) {
        $i$1$lcssa$i = $154;
        break L119;
       }
       $$pre25$i = HEAP8[$153>>0]|0; //@line 3192 "load_abc.cpp"
       $156 = $$pre25$i << 24 >> 24; //@line 3192 "load_abc.cpp"
       $157 = (_isspace($156)|0); //@line 3192 "load_abc.cpp"
       $158 = ($157|0)==(0); //@line 3192 "load_abc.cpp"
       if ($158) {
        $i$1$lcssa$i = $154;
        break;
       } else {
        $$211$i245 = $153;$i$112$i244 = $154;
       }
      }
     }
    } while(0);
    __ZL16abc_addchordnamePKciPKi($name$i,$i$1$lcssa$i,$notes$i); //@line 3197 "load_abc.cpp"
   } else {
    $128 = (_strlen($$1$lcssa$i)|0); //@line 323 "load_abc.cpp"
    $129 = (($128) + 45)|0; //@line 323 "load_abc.cpp"
    $130 = ($129>>>0)>(255); //@line 323 "load_abc.cpp"
    if (!($130)) {
     HEAP32[$vararg_buffer>>2] = $$1$lcssa$i; //@line 324 "load_abc.cpp"
     (_sprintf($txt$i$i,526763,$vararg_buffer)|0); //@line 324 "load_abc.cpp"
     $131 = HEAP32[37544]|0; //@line 328 "load_abc.cpp"
     HEAP32[$vararg_buffer1>>2] = $txt$i$i; //@line 328 "load_abc.cpp"
     (_fprintf($131,524747,$vararg_buffer1)|0); //@line 328 "load_abc.cpp"
    }
   }
  }
 }
 $212 = (_strncmp($p,526809,9)|0); //@line 3684 "load_abc.cpp"
 $213 = ($212|0)==(0); //@line 3684 "load_abc.cpp"
 if ($213) {
  $214 = ((($p)) + 9|0); //@line 3684 "load_abc.cpp"
  $215 = HEAP8[$214>>0]|0; //@line 3684 "load_abc.cpp"
  $216 = $215 << 24 >> 24; //@line 3684 "load_abc.cpp"
  $217 = (_isspace($216)|0); //@line 3684 "load_abc.cpp"
  $218 = ($217|0)==(0); //@line 3684 "load_abc.cpp"
  if (!($218)) {
   $$0$i42 = $214;
   while(1) {
    $219 = HEAP8[$$0$i42>>0]|0; //@line 3038 "load_abc.cpp"
    $220 = $219 << 24 >> 24; //@line 3038 "load_abc.cpp"
    $221 = (_isspace($220)|0); //@line 3038 "load_abc.cpp"
    $222 = ($221|0)==(0); //@line 3038 "load_abc.cpp"
    $223 = ((($$0$i42)) + 1|0); //@line 3038 "load_abc.cpp"
    if ($222) {
     $$0$i42$lcssa = $$0$i42;
     break;
    } else {
     $$0$i42 = $223;
    }
   }
   $224 = HEAP8[$$0$i42$lcssa>>0]|0; //@line 1625 "load_abc.cpp"
   $225 = $224 << 24 >> 24; //@line 1625 "load_abc.cpp"
   $isdigittmp1$i$i43 = (($225) + -48)|0; //@line 1625 "load_abc.cpp"
   $isdigit2$i$i44 = ($isdigittmp1$i$i43>>>0)<(10); //@line 1625 "load_abc.cpp"
   if ($isdigit2$i$i44) {
    $228 = $225;$h$04$i$i45 = 0;$i$03$i$i46 = 0;
    while(1) {
     $226 = ($h$04$i$i45*10)|0; //@line 1626 "load_abc.cpp"
     $227 = (($228) + -48)|0; //@line 1626 "load_abc.cpp"
     $229 = (($227) + ($226))|0; //@line 1626 "load_abc.cpp"
     $230 = (($i$03$i$i46) + 1)|0; //@line 1627 "load_abc.cpp"
     $231 = (($$0$i42$lcssa) + ($230)|0); //@line 1625 "load_abc.cpp"
     $232 = HEAP8[$231>>0]|0; //@line 1625 "load_abc.cpp"
     $233 = $232 << 24 >> 24; //@line 1625 "load_abc.cpp"
     $isdigittmp$i$i47 = (($233) + -48)|0; //@line 1625 "load_abc.cpp"
     $isdigit$i$i48 = ($isdigittmp$i$i47>>>0)<(10); //@line 1625 "load_abc.cpp"
     if ($isdigit$i$i48) {
      $228 = $233;$h$04$i$i45 = $229;$i$03$i$i46 = $230;
     } else {
      $$lcssa569 = $229;$$lcssa570 = $230;
      break;
     }
    }
    $234 = ($$lcssa570|0)==(0); //@line 1629 "load_abc.cpp"
    $$h$0$i$i50 = $234 ? 1 : $$lcssa569; //@line 1630 "load_abc.cpp"
    $235 = ($$h$0$i$i50|0)<(1); //@line 3040 "load_abc.cpp"
    if ($235) {
     $238 = 1;
    } else {
     $236 = ($$h$0$i$i50|0)>(128); //@line 3041 "load_abc.cpp"
     $$$h$0$i$i52 = $236 ? 128 : $$h$0$i$i50; //@line 3041 "load_abc.cpp"
     $238 = $$$h$0$i$i52;
    }
   } else {
    $238 = 1;
   }
   $237 = (($238) + 1)|0; //@line 3684 "load_abc.cpp"
   $239 = ((($h)) + 364|0); //@line 3684 "load_abc.cpp"
   HEAP8[$239>>0]=$237&255;HEAP8[$239+1>>0]=($237>>8)&255;HEAP8[$239+2>>0]=($237>>16)&255;HEAP8[$239+3>>0]=$237>>24; //@line 3684 "load_abc.cpp"
  }
 }
 $240 = (_strncmp($p,526819,8)|0); //@line 3685 "load_abc.cpp"
 $241 = ($240|0)==(0); //@line 3685 "load_abc.cpp"
 if ($241) {
  $242 = ((($p)) + 8|0); //@line 3685 "load_abc.cpp"
  $243 = HEAP8[$242>>0]|0; //@line 3685 "load_abc.cpp"
  $244 = $243 << 24 >> 24; //@line 3685 "load_abc.cpp"
  $245 = (_isspace($244)|0); //@line 3685 "load_abc.cpp"
  $246 = ($245|0)==(0); //@line 3685 "load_abc.cpp"
  if (!($246)) {
   $$0$i55 = $242;
   while(1) {
    $247 = HEAP8[$$0$i55>>0]|0; //@line 3028 "load_abc.cpp"
    $248 = $247 << 24 >> 24; //@line 3028 "load_abc.cpp"
    $249 = (_isspace($248)|0); //@line 3028 "load_abc.cpp"
    $250 = ($249|0)==(0); //@line 3028 "load_abc.cpp"
    $251 = ((($$0$i55)) + 1|0); //@line 3028 "load_abc.cpp"
    if ($250) {
     $$0$i55$lcssa = $$0$i55;
     break;
    } else {
     $$0$i55 = $251;
    }
   }
   $252 = HEAP8[$$0$i55$lcssa>>0]|0; //@line 1625 "load_abc.cpp"
   $253 = $252 << 24 >> 24; //@line 1625 "load_abc.cpp"
   $isdigittmp1$i$i56 = (($253) + -48)|0; //@line 1625 "load_abc.cpp"
   $isdigit2$i$i57 = ($isdigittmp1$i$i56>>>0)<(10); //@line 1625 "load_abc.cpp"
   if ($isdigit2$i$i57) {
    $256 = $253;$h$04$i$i58 = 0;$i$03$i$i59 = 0;
    while(1) {
     $254 = ($h$04$i$i58*10)|0; //@line 1626 "load_abc.cpp"
     $255 = (($256) + -48)|0; //@line 1626 "load_abc.cpp"
     $257 = (($255) + ($254))|0; //@line 1626 "load_abc.cpp"
     $258 = (($i$03$i$i59) + 1)|0; //@line 1627 "load_abc.cpp"
     $259 = (($$0$i55$lcssa) + ($258)|0); //@line 1625 "load_abc.cpp"
     $260 = HEAP8[$259>>0]|0; //@line 1625 "load_abc.cpp"
     $261 = $260 << 24 >> 24; //@line 1625 "load_abc.cpp"
     $isdigittmp$i$i60 = (($261) + -48)|0; //@line 1625 "load_abc.cpp"
     $isdigit$i$i61 = ($isdigittmp$i$i60>>>0)<(10); //@line 1625 "load_abc.cpp"
     if ($isdigit$i$i61) {
      $256 = $261;$h$04$i$i58 = $257;$i$03$i$i59 = $258;
     } else {
      $$lcssa567 = $257;$$lcssa568 = $258;
      break;
     }
    }
    $262 = ($$lcssa568|0)==(0); //@line 1629 "load_abc.cpp"
    $$h$0$i$i63 = $262 ? 1 : $$lcssa567; //@line 1630 "load_abc.cpp"
    $263 = ($$h$0$i$i63|0)<(0); //@line 3030 "load_abc.cpp"
    if ($263) {
     $266 = 0;
    } else {
     $264 = ($$h$0$i$i63|0)>(127); //@line 3031 "load_abc.cpp"
     $$$h$0$i$i65 = $264 ? 127 : $$h$0$i$i63; //@line 3031 "load_abc.cpp"
     $266 = $$$h$0$i$i65;
    }
   } else {
    $266 = 1;
   }
   $265 = ((($h)) + 360|0); //@line 3685 "load_abc.cpp"
   HEAP8[$265>>0]=$266&255;HEAP8[$265+1>>0]=($266>>8)&255;HEAP8[$265+2>>0]=($266>>16)&255;HEAP8[$265+3>>0]=$266>>24; //@line 3685 "load_abc.cpp"
  }
 }
 $267 = (_strncmp($p,526828,5)|0); //@line 3686 "load_abc.cpp"
 $268 = ($267|0)==(0); //@line 3686 "load_abc.cpp"
 do {
  if ($268) {
   $269 = ((($p)) + 5|0); //@line 3686 "load_abc.cpp"
   $270 = HEAP8[$269>>0]|0; //@line 3686 "load_abc.cpp"
   $271 = $270 << 24 >> 24; //@line 3686 "load_abc.cpp"
   $272 = (_isspace($271)|0); //@line 3686 "load_abc.cpp"
   $273 = ($272|0)==(0); //@line 3686 "load_abc.cpp"
   if (!($273)) {
    $274 = ((($h)) + 396|0); //@line 3686 "load_abc.cpp"
    $$0$i68 = $269;
    while(1) {
     $275 = HEAP8[$$0$i68>>0]|0; //@line 3049 "load_abc.cpp"
     $276 = $275 << 24 >> 24; //@line 3049 "load_abc.cpp"
     $277 = (_isspace($276)|0); //@line 3049 "load_abc.cpp"
     $278 = ($277|0)==(0); //@line 3049 "load_abc.cpp"
     $279 = ((($$0$i68)) + 1|0); //@line 3049 "load_abc.cpp"
     if ($278) {
      $$0$i68$lcssa = $$0$i68;
      break;
     } else {
      $$0$i68 = $279;
     }
    }
    $280 = ((($h)) + 392|0); //@line 3686 "load_abc.cpp"
    $281 = ((($h)) + 404|0); //@line 3686 "load_abc.cpp"
    $282 = HEAP8[$$0$i68$lcssa>>0]|0; //@line 1625 "load_abc.cpp"
    $283 = $282 << 24 >> 24; //@line 1625 "load_abc.cpp"
    $isdigittmp1$i$i69 = (($283) + -48)|0; //@line 1625 "load_abc.cpp"
    $isdigit2$i$i70 = ($isdigittmp1$i$i69>>>0)<(10); //@line 1625 "load_abc.cpp"
    if ($isdigit2$i$i70) {
     $286 = $283;$h$04$i$i71 = 0;$i$03$i$i72 = 0;
     while(1) {
      $284 = ($h$04$i$i71*10)|0; //@line 1626 "load_abc.cpp"
      $285 = (($286) + -48)|0; //@line 1626 "load_abc.cpp"
      $287 = (($285) + ($284))|0; //@line 1626 "load_abc.cpp"
      $288 = (($i$03$i$i72) + 1)|0; //@line 1627 "load_abc.cpp"
      $289 = (($$0$i68$lcssa) + ($288)|0); //@line 1625 "load_abc.cpp"
      $290 = HEAP8[$289>>0]|0; //@line 1625 "load_abc.cpp"
      $291 = $290 << 24 >> 24; //@line 1625 "load_abc.cpp"
      $isdigittmp$i$i73 = (($291) + -48)|0; //@line 1625 "load_abc.cpp"
      $isdigit$i$i74 = ($isdigittmp$i$i73>>>0)<(10); //@line 1625 "load_abc.cpp"
      if ($isdigit$i$i74) {
       $286 = $291;$h$04$i$i71 = $287;$i$03$i$i72 = $288;
      } else {
       $h$0$lcssa$i$i76 = $287;$i$0$lcssa$i$i77 = $288;
       break;
      }
     }
    } else {
     $h$0$lcssa$i$i76 = 0;$i$0$lcssa$i$i77 = 0;
    }
    $292 = ($i$0$lcssa$i$i77|0)==(0); //@line 1629 "load_abc.cpp"
    $$h$0$i$i78 = $292 ? 1 : $h$0$lcssa$i$i76; //@line 1630 "load_abc.cpp"
    $293 = (($$0$i68$lcssa) + ($i$0$lcssa$i$i77)|0); //@line 3050 "load_abc.cpp"
    $294 = (($$h$0$i$i78) + 1)|0; //@line 3051 "load_abc.cpp"
    $295 = ($$h$0$i$i78|0)>(-1); //@line 3052 "load_abc.cpp"
    $296 = ($294|0)<(129); //@line 3052 "load_abc.cpp"
    $or$cond$i = $295 & $296; //@line 3052 "load_abc.cpp"
    $$$i = $or$cond$i ? $294 : 71; //@line 3052 "load_abc.cpp"
    HEAP32[$280>>2] = $$$i; //@line 3053 "load_abc.cpp"
    $$1$i80 = $293;
    while(1) {
     $297 = HEAP8[$$1$i80>>0]|0; //@line 3056 "load_abc.cpp"
     $298 = $297 << 24 >> 24; //@line 3056 "load_abc.cpp"
     $299 = (_isspace($298)|0); //@line 3056 "load_abc.cpp"
     $300 = ($299|0)==(0); //@line 3056 "load_abc.cpp"
     $301 = ((($$1$i80)) + 1|0); //@line 3056 "load_abc.cpp"
     if ($300) {
      $$1$i80$lcssa = $$1$i80;
      break;
     } else {
      $$1$i80 = $301;
     }
    }
    $302 = HEAP8[$$1$i80$lcssa>>0]|0; //@line 1625 "load_abc.cpp"
    $303 = $302 << 24 >> 24; //@line 1625 "load_abc.cpp"
    $isdigittmp1$i46$i = (($303) + -48)|0; //@line 1625 "load_abc.cpp"
    $isdigit2$i47$i = ($isdigittmp1$i46$i>>>0)<(10); //@line 1625 "load_abc.cpp"
    if ($isdigit2$i47$i) {
     $306 = $303;$h$04$i48$i = 0;$i$03$i49$i = 0;
     while(1) {
      $304 = ($h$04$i48$i*10)|0; //@line 1626 "load_abc.cpp"
      $305 = (($306) + -48)|0; //@line 1626 "load_abc.cpp"
      $307 = (($305) + ($304))|0; //@line 1626 "load_abc.cpp"
      $308 = (($i$03$i49$i) + 1)|0; //@line 1627 "load_abc.cpp"
      $309 = (($$1$i80$lcssa) + ($308)|0); //@line 1625 "load_abc.cpp"
      $310 = HEAP8[$309>>0]|0; //@line 1625 "load_abc.cpp"
      $311 = $310 << 24 >> 24; //@line 1625 "load_abc.cpp"
      $isdigittmp$i50$i = (($311) + -48)|0; //@line 1625 "load_abc.cpp"
      $isdigit$i51$i = ($isdigittmp$i50$i>>>0)<(10); //@line 1625 "load_abc.cpp"
      if ($isdigit$i51$i) {
       $306 = $311;$h$04$i48$i = $307;$i$03$i49$i = $308;
      } else {
       $$lcssa562 = $307;$$lcssa563 = $308;$$lcssa564 = $309;
       break;
      }
     }
     $312 = ($$lcssa563|0)==(0); //@line 1629 "load_abc.cpp"
     $$h$0$i55$i = $312 ? 1 : $$lcssa562; //@line 1630 "load_abc.cpp"
     $$off$i = (($$h$0$i55$i) + -1)|0; //@line 3058 "load_abc.cpp"
     $313 = ($$off$i>>>0)<(126); //@line 3058 "load_abc.cpp"
     $$h$0$i55$$i = $313 ? $$h$0$i55$i : 45; //@line 3058 "load_abc.cpp"
     $$2$ph$i = $$lcssa564;$storemerge82$i = $$h$0$i55$$i;
    } else {
     $$2$ph$i = $$1$i80$lcssa;$storemerge82$i = 1;
    }
    HEAP32[$274>>2] = $storemerge82$i; //@line 3061 "load_abc.cpp"
    $$2$i81 = $$2$ph$i;
    while(1) {
     $314 = HEAP8[$$2$i81>>0]|0; //@line 3062 "load_abc.cpp"
     $315 = $314 << 24 >> 24; //@line 3062 "load_abc.cpp"
     $316 = (_isspace($315)|0); //@line 3062 "load_abc.cpp"
     $317 = ($316|0)==(0); //@line 3062 "load_abc.cpp"
     $318 = ((($$2$i81)) + 1|0); //@line 3062 "load_abc.cpp"
     if ($317) {
      $$2$i81$lcssa = $$2$i81;
      break;
     } else {
      $$2$i81 = $318;
     }
    }
    $319 = HEAP8[$$2$i81$lcssa>>0]|0; //@line 1625 "load_abc.cpp"
    $320 = $319 << 24 >> 24; //@line 1625 "load_abc.cpp"
    $isdigittmp1$i35$i = (($320) + -48)|0; //@line 1625 "load_abc.cpp"
    $isdigit2$i36$i = ($isdigittmp1$i35$i>>>0)<(10); //@line 1625 "load_abc.cpp"
    if ($isdigit2$i36$i) {
     $324 = $320;$h$04$i37$i = 0;$i$03$i38$i = 0;
     while(1) {
      $322 = ($h$04$i37$i*10)|0; //@line 1626 "load_abc.cpp"
      $323 = (($324) + -48)|0; //@line 1626 "load_abc.cpp"
      $325 = (($323) + ($322))|0; //@line 1626 "load_abc.cpp"
      $326 = (($i$03$i38$i) + 1)|0; //@line 1627 "load_abc.cpp"
      $327 = (($$2$i81$lcssa) + ($326)|0); //@line 1625 "load_abc.cpp"
      $328 = HEAP8[$327>>0]|0; //@line 1625 "load_abc.cpp"
      $329 = $328 << 24 >> 24; //@line 1625 "load_abc.cpp"
      $isdigittmp$i39$i = (($329) + -48)|0; //@line 1625 "load_abc.cpp"
      $isdigit$i40$i = ($isdigittmp$i39$i>>>0)<(10); //@line 1625 "load_abc.cpp"
      if ($isdigit$i40$i) {
       $324 = $329;$h$04$i37$i = $325;$i$03$i38$i = $326;
      } else {
       $$lcssa559 = $325;$$lcssa560 = $326;$$lcssa561 = $327;
       break;
      }
     }
     $330 = ($$lcssa560|0)==(0); //@line 1629 "load_abc.cpp"
     $$h$0$i44$i = $330 ? 1 : $$lcssa559; //@line 1630 "load_abc.cpp"
     $$off10$i = (($$h$0$i44$i) + -1)|0; //@line 3064 "load_abc.cpp"
     $331 = ($$off10$i>>>0)<(126); //@line 3064 "load_abc.cpp"
     $332 = ((($h)) + 400|0); //@line 3065 "load_abc.cpp"
     if ($331) {
      $$h$0$i4462$i = $$h$0$i44$i;$333 = $332;$984 = $$lcssa561;
      label = 110;
     } else {
      HEAP32[$332>>2] = 33; //@line 3067 "load_abc.cpp"
      $$3$i = $$lcssa561;
     }
    } else {
     $321 = ((($h)) + 400|0); //@line 3065 "load_abc.cpp"
     $$h$0$i4462$i = 1;$333 = $321;$984 = $$2$i81$lcssa;
     label = 110;
    }
    if ((label|0) == 110) {
     HEAP32[$333>>2] = $$h$0$i4462$i; //@line 3065 "load_abc.cpp"
     $$3$i = $984;
    }
    while(1) {
     $334 = HEAP8[$$3$i>>0]|0; //@line 3068 "load_abc.cpp"
     $335 = $334 << 24 >> 24; //@line 3068 "load_abc.cpp"
     $336 = (_isspace($335)|0); //@line 3068 "load_abc.cpp"
     $337 = ($336|0)==(0); //@line 3068 "load_abc.cpp"
     $338 = ((($$3$i)) + 1|0); //@line 3068 "load_abc.cpp"
     if ($337) {
      $$3$i$lcssa = $$3$i;
      break;
     } else {
      $$3$i = $338;
     }
    }
    $339 = HEAP8[$$3$i$lcssa>>0]|0; //@line 1625 "load_abc.cpp"
    $340 = $339 << 24 >> 24; //@line 1625 "load_abc.cpp"
    $isdigittmp1$i24$i = (($340) + -48)|0; //@line 1625 "load_abc.cpp"
    $isdigit2$i25$i = ($isdigittmp1$i24$i>>>0)<(10); //@line 1625 "load_abc.cpp"
    if ($isdigit2$i25$i) {
     $343 = $340;$h$04$i26$i = 0;$i$03$i27$i = 0;
     while(1) {
      $341 = ($h$04$i26$i*10)|0; //@line 1626 "load_abc.cpp"
      $342 = (($343) + -48)|0; //@line 1626 "load_abc.cpp"
      $344 = (($342) + ($341))|0; //@line 1626 "load_abc.cpp"
      $345 = (($i$03$i27$i) + 1)|0; //@line 1627 "load_abc.cpp"
      $346 = (($$3$i$lcssa) + ($345)|0); //@line 1625 "load_abc.cpp"
      $347 = HEAP8[$346>>0]|0; //@line 1625 "load_abc.cpp"
      $348 = $347 << 24 >> 24; //@line 1625 "load_abc.cpp"
      $isdigittmp$i28$i = (($348) + -48)|0; //@line 1625 "load_abc.cpp"
      $isdigit$i29$i = ($isdigittmp$i28$i>>>0)<(10); //@line 1625 "load_abc.cpp"
      if ($isdigit$i29$i) {
       $343 = $348;$h$04$i26$i = $344;$i$03$i27$i = $345;
      } else {
       $$lcssa556 = $344;$$lcssa557 = $345;$$lcssa558 = $346;
       break;
      }
     }
     $349 = ($$lcssa557|0)==(0); //@line 1629 "load_abc.cpp"
     $$h$0$i33$i = $349 ? 1 : $$lcssa556; //@line 1630 "load_abc.cpp"
     $$off11$i = (($$h$0$i33$i) + -1)|0; //@line 3070 "load_abc.cpp"
     $350 = ($$off11$i>>>0)<(126); //@line 3070 "load_abc.cpp"
     $$h$0$i33$$i = $350 ? $$h$0$i33$i : 80; //@line 3070 "load_abc.cpp"
     $$4$ph$i = $$lcssa558;$storemerge83$i = $$h$0$i33$$i;
    } else {
     $$4$ph$i = $$3$i$lcssa;$storemerge83$i = 1;
    }
    HEAP32[$281>>2] = $storemerge83$i; //@line 3073 "load_abc.cpp"
    $$4$i = $$4$ph$i;
    while(1) {
     $351 = HEAP8[$$4$i>>0]|0; //@line 3074 "load_abc.cpp"
     $352 = $351 << 24 >> 24; //@line 3074 "load_abc.cpp"
     $353 = (_isspace($352)|0); //@line 3074 "load_abc.cpp"
     $354 = ($353|0)==(0); //@line 3074 "load_abc.cpp"
     $355 = ((($$4$i)) + 1|0); //@line 3074 "load_abc.cpp"
     if ($354) {
      $$4$i$lcssa = $$4$i;
      break;
     } else {
      $$4$i = $355;
     }
    }
    $356 = HEAP8[$$4$i$lcssa>>0]|0; //@line 1625 "load_abc.cpp"
    $357 = $356 << 24 >> 24; //@line 1625 "load_abc.cpp"
    $isdigittmp1$i13$i = (($357) + -48)|0; //@line 1625 "load_abc.cpp"
    $isdigit2$i14$i = ($isdigittmp1$i13$i>>>0)<(10); //@line 1625 "load_abc.cpp"
    if ($isdigit2$i14$i) {
     $361 = $357;$h$04$i15$i = 0;$i$03$i16$i = 0;
     while(1) {
      $359 = ($h$04$i15$i*10)|0; //@line 1626 "load_abc.cpp"
      $360 = (($361) + -48)|0; //@line 1626 "load_abc.cpp"
      $362 = (($360) + ($359))|0; //@line 1626 "load_abc.cpp"
      $363 = (($i$03$i16$i) + 1)|0; //@line 1627 "load_abc.cpp"
      $364 = (($$4$i$lcssa) + ($363)|0); //@line 1625 "load_abc.cpp"
      $365 = HEAP8[$364>>0]|0; //@line 1625 "load_abc.cpp"
      $366 = $365 << 24 >> 24; //@line 1625 "load_abc.cpp"
      $isdigittmp$i17$i = (($366) + -48)|0; //@line 1625 "load_abc.cpp"
      $isdigit$i18$i = ($isdigittmp$i17$i>>>0)<(10); //@line 1625 "load_abc.cpp"
      if ($isdigit$i18$i) {
       $361 = $366;$h$04$i15$i = $362;$i$03$i16$i = $363;
      } else {
       $$lcssa554 = $362;$$lcssa555 = $363;
       break;
      }
     }
     $367 = ($$lcssa555|0)==(0); //@line 1629 "load_abc.cpp"
     $$h$0$i22$i = $367 ? 1 : $$lcssa554; //@line 1630 "load_abc.cpp"
     $$off12$i = (($$h$0$i22$i) + -1)|0; //@line 3076 "load_abc.cpp"
     $368 = ($$off12$i>>>0)<(126); //@line 3076 "load_abc.cpp"
     $369 = ((($h)) + 408|0); //@line 3077 "load_abc.cpp"
     if ($368) {
      $$h$0$i2268$i = $$h$0$i22$i;$370 = $369;
     } else {
      HEAP32[$369>>2] = 80; //@line 3079 "load_abc.cpp"
      break;
     }
    } else {
     $358 = ((($h)) + 408|0); //@line 3077 "load_abc.cpp"
     $$h$0$i2268$i = 1;$370 = $358;
    }
    HEAP32[$370>>2] = $$h$0$i2268$i; //@line 3077 "load_abc.cpp"
   }
  }
 } while(0);
 $371 = (_strncmp($p,526834,8)|0); //@line 3687 "load_abc.cpp"
 $372 = ($371|0)==(0); //@line 3687 "load_abc.cpp"
 do {
  if ($372) {
   $373 = ((($p)) + 8|0); //@line 3687 "load_abc.cpp"
   $374 = HEAP8[$373>>0]|0; //@line 3687 "load_abc.cpp"
   $375 = ($374<<24>>24)==(0); //@line 3687 "load_abc.cpp"
   $376 = ($374<<24>>24)==($delim<<24>>24); //@line 3687 "load_abc.cpp"
   $or$cond = $375 | $376; //@line 3687 "load_abc.cpp"
   if (!($or$cond)) {
    $377 = $374 << 24 >> 24; //@line 3687 "load_abc.cpp"
    $378 = (_isspace($377)|0); //@line 3687 "load_abc.cpp"
    $379 = ($378|0)==(0); //@line 3687 "load_abc.cpp"
    if ($379) {
     break;
    }
   }
   $380 = ((($h)) + 388|0); //@line 3687 "load_abc.cpp"
   HEAP8[$380>>0]=0&255;HEAP8[$380+1>>0]=(0>>8)&255;HEAP8[$380+2>>0]=(0>>16)&255;HEAP8[$380+3>>0]=0>>24; //@line 3687 "load_abc.cpp"
  }
 } while(0);
 $381 = (_strncmp($p,526843,7)|0); //@line 3688 "load_abc.cpp"
 $382 = ($381|0)==(0); //@line 3688 "load_abc.cpp"
 do {
  if ($382) {
   $383 = ((($p)) + 7|0); //@line 3688 "load_abc.cpp"
   $384 = HEAP8[$383>>0]|0; //@line 3688 "load_abc.cpp"
   $385 = ($384<<24>>24)==(0); //@line 3688 "load_abc.cpp"
   $386 = ($384<<24>>24)==($delim<<24>>24); //@line 3688 "load_abc.cpp"
   $or$cond1 = $385 | $386; //@line 3688 "load_abc.cpp"
   if (!($or$cond1)) {
    $387 = $384 << 24 >> 24; //@line 3688 "load_abc.cpp"
    $388 = (_isspace($387)|0); //@line 3688 "load_abc.cpp"
    $389 = ($388|0)==(0); //@line 3688 "load_abc.cpp"
    if ($389) {
     break;
    }
   }
   $390 = ((($h)) + 388|0); //@line 3688 "load_abc.cpp"
   HEAP8[$390>>0]=1&255;HEAP8[$390+1>>0]=(1>>8)&255;HEAP8[$390+2>>0]=(1>>16)&255;HEAP8[$390+3>>0]=1>>24; //@line 3688 "load_abc.cpp"
  }
 } while(0);
 $391 = ((($h)) + 380|0); //@line 3689 "load_abc.cpp"
 $392 = HEAPU8[$391>>0]|(HEAPU8[$391+1>>0]<<8)|(HEAPU8[$391+2>>0]<<16)|(HEAPU8[$391+3>>0]<<24); //@line 3689 "load_abc.cpp"
 $393 = (_strncmp($p,526851,4)|0); //@line 3690 "load_abc.cpp"
 $394 = ($393|0)==(0); //@line 3690 "load_abc.cpp"
 L224: do {
  if ($394) {
   $395 = ((($p)) + 4|0); //@line 3690 "load_abc.cpp"
   $396 = HEAP8[$395>>0]|0; //@line 3690 "load_abc.cpp"
   $397 = $396 << 24 >> 24; //@line 3690 "load_abc.cpp"
   $398 = (_isspace($397)|0); //@line 3690 "load_abc.cpp"
   $399 = ($398|0)==(0); //@line 3690 "load_abc.cpp"
   if (!($399)) {
    $$01$i = $395;
    while(1) {
     $400 = HEAP8[$$01$i>>0]|0; //@line 3207 "load_abc.cpp"
     $401 = $400 << 24 >> 24; //@line 3207 "load_abc.cpp"
     $402 = (_isspace($401)|0); //@line 3207 "load_abc.cpp"
     $403 = ($402|0)==(0); //@line 3207 "load_abc.cpp"
     $404 = ((($$01$i)) + 1|0); //@line 3207 "load_abc.cpp"
     if ($403) {
      $$01$i$lcssa = $$01$i;
      break;
     } else {
      $$01$i = $404;
     }
    }
    $405 = (_strncmp($$01$i$lcssa,526513,2)|0); //@line 3208 "load_abc.cpp"
    $406 = ($405|0)==(0); //@line 3208 "load_abc.cpp"
    if ($406) {
     $407 = ((($$01$i$lcssa)) + 2|0); //@line 3208 "load_abc.cpp"
     $408 = HEAP8[$407>>0]|0; //@line 3208 "load_abc.cpp"
     $409 = $408 << 24 >> 24; //@line 3208 "load_abc.cpp"
     $410 = (_isspace($409)|0); //@line 3208 "load_abc.cpp"
     $411 = ($410|0)==(0); //@line 3208 "load_abc.cpp"
     if ($411) {
      $412 = HEAP8[$407>>0]|0; //@line 3208 "load_abc.cpp"
      $413 = ($412<<24>>24)==(0); //@line 3208 "load_abc.cpp"
      if ($413) {
       $$0$i98$ph = 1;
      } else {
       label = 138;
      }
     } else {
      $$0$i98$ph = 1;
     }
    } else {
     label = 138;
    }
    do {
     if ((label|0) == 138) {
      $414 = (_strncmp($$01$i$lcssa,526516,3)|0); //@line 3209 "load_abc.cpp"
      $415 = ($414|0)==(0); //@line 3209 "load_abc.cpp"
      if ($415) {
       $416 = ((($$01$i$lcssa)) + 3|0); //@line 3209 "load_abc.cpp"
       $417 = HEAP8[$416>>0]|0; //@line 3209 "load_abc.cpp"
       $418 = $417 << 24 >> 24; //@line 3209 "load_abc.cpp"
       $419 = (_isspace($418)|0); //@line 3209 "load_abc.cpp"
       $420 = ($419|0)==(0); //@line 3209 "load_abc.cpp"
       if (!($420)) {
        $$0$i98$ph = 0;
        break;
       }
       $421 = HEAP8[$416>>0]|0; //@line 3209 "load_abc.cpp"
       $422 = ($421<<24>>24)==(0); //@line 3209 "load_abc.cpp"
       if ($422) {
        $$0$i98$ph = 0;
        break;
       }
      }
      $423 = ((($h)) + 116|0); //@line 3211 "load_abc.cpp"
      $424 = HEAP8[$$01$i$lcssa>>0]|0; //@line 3211 "load_abc.cpp"
      $425 = ($424<<24>>24)==(0); //@line 3211 "load_abc.cpp"
      if ($425) {
       HEAP8[$423>>0] = 0; //@line 3219 "load_abc.cpp"
      } else {
       $$135$i = $$01$i$lcssa;$427 = $424;$n$037$i = 0;$q$036$i = $423;
       while(1) {
        $426 = $427 << 24 >> 24; //@line 3211 "load_abc.cpp"
        $428 = (_isspace($426)|0); //@line 3211 "load_abc.cpp"
        $429 = ($428|0)==(0); //@line 3211 "load_abc.cpp"
        if (!($429)) {
         $$1$lcssa$i83 = $$135$i;$n$0$lcssa$i = $n$037$i;$q$0$lcssa$i = $q$036$i;
         break;
        }
        $430 = HEAP8[$$135$i>>0]|0; //@line 3212 "load_abc.cpp"
        $431 = $430 << 24 >> 24; //@line 3212 "load_abc.cpp"
        $memchr$i = (_memchr(526856,$431,13)|0); //@line 3212 "load_abc.cpp"
        $432 = ($memchr$i|0)==(0|0); //@line 3212 "load_abc.cpp"
        if ($432) {
         $$1$lcssa$i83 = $$135$i;$n$0$lcssa$i = $n$037$i;$q$0$lcssa$i = $q$036$i;
         break;
        }
        $433 = ((($q$036$i)) + 1|0); //@line 3213 "load_abc.cpp"
        HEAP8[$q$036$i>>0] = $430; //@line 3213 "load_abc.cpp"
        $434 = HEAP8[$$135$i>>0]|0; //@line 3214 "load_abc.cpp"
        $435 = $434 << 24 >> 24; //@line 3214 "load_abc.cpp"
        $isdigittmp4$i = (($435) + -48)|0; //@line 3214 "load_abc.cpp"
        $isdigit5$i = ($isdigittmp4$i>>>0)<(10); //@line 3214 "load_abc.cpp"
        $$pre64$i = ((($$135$i)) + 1|0); //@line 3211 "load_abc.cpp"
        if ($isdigit5$i) {
         $n$1$i = $n$037$i;$q$2$i = $433;
        } else {
         $436 = HEAP8[$$pre64$i>>0]|0; //@line 3215 "load_abc.cpp"
         $437 = $436 << 24 >> 24; //@line 3215 "load_abc.cpp"
         $isdigittmp6$i = (($437) + -48)|0; //@line 3215 "load_abc.cpp"
         $isdigit7$i = ($isdigittmp6$i>>>0)<(10); //@line 3215 "load_abc.cpp"
         if ($isdigit7$i) {
          $q$1$i = $433;
         } else {
          $438 = ((($q$036$i)) + 2|0); //@line 3215 "load_abc.cpp"
          HEAP8[$433>>0] = 49; //@line 3215 "load_abc.cpp"
          $q$1$i = $438;
         }
         $439 = (($n$037$i) + 1)|0; //@line 3216 "load_abc.cpp"
         $n$1$i = $439;$q$2$i = $q$1$i;
        }
        $440 = HEAP8[$$pre64$i>>0]|0; //@line 3211 "load_abc.cpp"
        $441 = ($440<<24>>24)==(0); //@line 3211 "load_abc.cpp"
        if ($441) {
         $$1$lcssa$i83 = $$pre64$i;$n$0$lcssa$i = $n$1$i;$q$0$lcssa$i = $q$2$i;
         break;
        } else {
         $$135$i = $$pre64$i;$427 = $440;$n$037$i = $n$1$i;$q$036$i = $q$2$i;
        }
       }
       HEAP8[$q$0$lcssa$i>>0] = 0; //@line 3219 "load_abc.cpp"
       $442 = ($n$0$lcssa$i|0)>(0); //@line 3221 "load_abc.cpp"
       if ($442) {
        $$231$i = $$1$lcssa$i83;$i$032$i = 0;
        while(1) {
         $443 = $i$032$i << 1; //@line 3222 "load_abc.cpp"
         $444 = (((($h)) + 116|0) + ($443)|0); //@line 3222 "load_abc.cpp"
         $445 = HEAP8[$444>>0]|0; //@line 3222 "load_abc.cpp"
         $446 = ($445<<24>>24)==(100); //@line 3222 "load_abc.cpp"
         if ($446) {
          $$3$i85 = $$231$i;
          while(1) {
           $447 = HEAP8[$$3$i85>>0]|0; //@line 3223 "load_abc.cpp"
           $448 = $447 << 24 >> 24; //@line 3223 "load_abc.cpp"
           $449 = (_isspace($448)|0); //@line 3223 "load_abc.cpp"
           $450 = ($449|0)==(0); //@line 3223 "load_abc.cpp"
           $451 = ((($$3$i85)) + 1|0); //@line 3223 "load_abc.cpp"
           if ($450) {
            $$3$i85$lcssa = $$3$i85;
            break;
           } else {
            $$3$i85 = $451;
           }
          }
          $452 = HEAP8[$$3$i85$lcssa>>0]|0; //@line 3224 "load_abc.cpp"
          $453 = $452 << 24 >> 24; //@line 3224 "load_abc.cpp"
          $isdigittmp2$i = (($453) + -48)|0; //@line 3224 "load_abc.cpp"
          $isdigit3$i = ($isdigittmp2$i>>>0)<(10); //@line 3224 "load_abc.cpp"
          do {
           if ($isdigit3$i) {
            $462 = $453;$h$04$i$i88 = 0;$i$03$i$i89 = 0;
            while(1) {
             $460 = ($h$04$i$i88*10)|0; //@line 1626 "load_abc.cpp"
             $461 = (($462) + -48)|0; //@line 1626 "load_abc.cpp"
             $463 = (($461) + ($460))|0; //@line 1626 "load_abc.cpp"
             $464 = (($i$03$i$i89) + 1)|0; //@line 1627 "load_abc.cpp"
             $465 = (($$3$i85$lcssa) + ($464)|0); //@line 1625 "load_abc.cpp"
             $466 = HEAP8[$465>>0]|0; //@line 1625 "load_abc.cpp"
             $467 = $466 << 24 >> 24; //@line 1625 "load_abc.cpp"
             $isdigittmp$i$i90 = (($467) + -48)|0; //@line 1625 "load_abc.cpp"
             $isdigit$i$i91 = ($isdigittmp$i$i90>>>0)<(10); //@line 1625 "load_abc.cpp"
             if ($isdigit$i$i91) {
              $462 = $467;$h$04$i$i88 = $463;$i$03$i$i89 = $464;
             } else {
              $$lcssa551 = $463;$$lcssa552 = $464;$$lcssa553 = $465;
              break;
             }
            }
            $468 = ($$lcssa552|0)==(0); //@line 1629 "load_abc.cpp"
            $$h$0$i$i93 = $468 ? 1 : $$lcssa551; //@line 1630 "load_abc.cpp"
            $$5$i = $$lcssa553;$470 = $$h$0$i$i93;
           } else {
            $454 = (_isspace($453)|0); //@line 3226 "load_abc.cpp"
            $455 = ($454|0)==(0); //@line 3226 "load_abc.cpp"
            if ($455) {
             $$3$i85$pn = $$3$i85$lcssa;
            } else {
             $$5$i = $$3$i85$lcssa;$470 = 0;
             break;
            }
            while(1) {
             $456 = ((($$3$i85$pn)) + 1|0); //@line 3226 "load_abc.cpp"
             $$pre$i87 = HEAP8[$456>>0]|0; //@line 3226 "load_abc.cpp"
             $457 = $$pre$i87 << 24 >> 24; //@line 3226 "load_abc.cpp"
             $458 = (_isspace($457)|0); //@line 3226 "load_abc.cpp"
             $459 = ($458|0)==(0); //@line 3226 "load_abc.cpp"
             if ($459) {
              $$3$i85$pn = $456;
             } else {
              $$5$i = $456;$470 = 0;
              break;
             }
            }
           }
          } while(0);
          $469 = (($470) + 1)|0; //@line 3230 "load_abc.cpp"
          $471 = $469&255; //@line 3230 "load_abc.cpp"
          $472 = (((($h)) + 196|0) + ($i$032$i)|0); //@line 3230 "load_abc.cpp"
          HEAP8[$472>>0] = $471; //@line 3230 "load_abc.cpp"
          $$6$i = $$5$i;
         } else {
          $473 = (((($h)) + 196|0) + ($i$032$i)|0); //@line 3232 "load_abc.cpp"
          HEAP8[$473>>0] = 0; //@line 3232 "load_abc.cpp"
          $$6$i = $$231$i;
         }
         $474 = (($i$032$i) + 1)|0; //@line 3221 "load_abc.cpp"
         $exitcond60$i = ($474|0)==($n$0$lcssa$i|0); //@line 3221 "load_abc.cpp"
         if ($exitcond60$i) {
          $$728$i = $$6$i;$i$129$i = 0;
          break;
         } else {
          $$231$i = $$6$i;$i$032$i = $474;
         }
        }
        while(1) {
         $475 = $i$129$i << 1; //@line 3236 "load_abc.cpp"
         $476 = (((($h)) + 116|0) + ($475)|0); //@line 3236 "load_abc.cpp"
         $477 = HEAP8[$476>>0]|0; //@line 3236 "load_abc.cpp"
         $478 = ($477<<24>>24)==(100); //@line 3236 "load_abc.cpp"
         if ($478) {
          $$8$i = $$728$i;
          while(1) {
           $479 = HEAP8[$$8$i>>0]|0; //@line 3237 "load_abc.cpp"
           $480 = $479 << 24 >> 24; //@line 3237 "load_abc.cpp"
           $481 = (_isspace($480)|0); //@line 3237 "load_abc.cpp"
           $482 = ($481|0)==(0); //@line 3237 "load_abc.cpp"
           $483 = ((($$8$i)) + 1|0); //@line 3237 "load_abc.cpp"
           if ($482) {
            $$8$i$lcssa = $$8$i;
            break;
           } else {
            $$8$i = $483;
           }
          }
          $484 = HEAP8[$$8$i$lcssa>>0]|0; //@line 3238 "load_abc.cpp"
          $485 = $484 << 24 >> 24; //@line 3238 "load_abc.cpp"
          $isdigittmp$i = (($485) + -48)|0; //@line 3238 "load_abc.cpp"
          $isdigit$i = ($isdigittmp$i>>>0)<(10); //@line 3238 "load_abc.cpp"
          do {
           if ($isdigit$i) {
            $494 = $485;$h$04$i10$i = 0;$i$03$i11$i = 0;
            while(1) {
             $492 = ($h$04$i10$i*10)|0; //@line 1626 "load_abc.cpp"
             $493 = (($494) + -48)|0; //@line 1626 "load_abc.cpp"
             $495 = (($493) + ($492))|0; //@line 1626 "load_abc.cpp"
             $496 = (($i$03$i11$i) + 1)|0; //@line 1627 "load_abc.cpp"
             $497 = (($$8$i$lcssa) + ($496)|0); //@line 1625 "load_abc.cpp"
             $498 = HEAP8[$497>>0]|0; //@line 1625 "load_abc.cpp"
             $499 = $498 << 24 >> 24; //@line 1625 "load_abc.cpp"
             $isdigittmp$i12$i = (($499) + -48)|0; //@line 1625 "load_abc.cpp"
             $isdigit$i13$i = ($isdigittmp$i12$i>>>0)<(10); //@line 1625 "load_abc.cpp"
             if ($isdigit$i13$i) {
              $494 = $499;$h$04$i10$i = $495;$i$03$i11$i = $496;
             } else {
              $$lcssa547 = $495;$$lcssa548 = $496;$$lcssa549 = $497;
              break;
             }
            }
            $500 = ($$lcssa548|0)==(0); //@line 1629 "load_abc.cpp"
            $$h$0$i17$i = $500 ? 1 : $$lcssa547; //@line 1630 "load_abc.cpp"
            $$10$i = $$lcssa549;$502 = $$h$0$i17$i;
           } else {
            $486 = (_isspace($485)|0); //@line 3240 "load_abc.cpp"
            $487 = ($486|0)==(0); //@line 3240 "load_abc.cpp"
            if ($487) {
             $$8$i$pn = $$8$i$lcssa;
            } else {
             $$10$i = $$8$i$lcssa;$502 = 0;
             break;
            }
            while(1) {
             $488 = ((($$8$i$pn)) + 1|0); //@line 3240 "load_abc.cpp"
             $$pre62$i = HEAP8[$488>>0]|0; //@line 3240 "load_abc.cpp"
             $489 = $$pre62$i << 24 >> 24; //@line 3240 "load_abc.cpp"
             $490 = (_isspace($489)|0); //@line 3240 "load_abc.cpp"
             $491 = ($490|0)==(0); //@line 3240 "load_abc.cpp"
             if ($491) {
              $$8$i$pn = $488;
             } else {
              $$10$i = $488;$502 = 0;
              break;
             }
            }
           }
          } while(0);
          $501 = $502&255; //@line 3244 "load_abc.cpp"
          $503 = (((($h)) + 276|0) + ($i$129$i)|0); //@line 3244 "load_abc.cpp"
          HEAP8[$503>>0] = $501; //@line 3244 "load_abc.cpp"
          $$11$i = $$10$i;
         } else {
          $504 = (((($h)) + 276|0) + ($i$129$i)|0); //@line 3246 "load_abc.cpp"
          HEAP8[$504>>0] = 0; //@line 3246 "load_abc.cpp"
          $$11$i = $$728$i;
         }
         $505 = (($i$129$i) + 1)|0; //@line 3235 "load_abc.cpp"
         $exitcond$i97 = ($505|0)==($n$0$lcssa$i|0); //@line 3235 "load_abc.cpp"
         if ($exitcond$i97) {
          break;
         } else {
          $$728$i = $$11$i;$i$129$i = $505;
         }
        }
       }
      }
      HEAP8[$391>>0]=$392&255;HEAP8[$391+1>>0]=($392>>8)&255;HEAP8[$391+2>>0]=($392>>16)&255;HEAP8[$391+3>>0]=$392>>24; //@line 3693 "load_abc.cpp"
      break L224;
     }
    } while(0);
    HEAP8[$391>>0]=$$0$i98$ph&255;HEAP8[$391+1>>0]=($$0$i98$ph>>8)&255;HEAP8[$391+2>>0]=($$0$i98$ph>>16)&255;HEAP8[$391+3>>0]=$$0$i98$ph>>24; //@line 3692 "load_abc.cpp"
   }
  }
 } while(0);
 $506 = (_strncmp($p,526869,7)|0); //@line 3695 "load_abc.cpp"
 $507 = ($506|0)==(0); //@line 3695 "load_abc.cpp"
 do {
  if ($507) {
   $508 = ((($p)) + 7|0); //@line 3695 "load_abc.cpp"
   $509 = HEAP8[$508>>0]|0; //@line 3695 "load_abc.cpp"
   $510 = ($509<<24>>24)==(0); //@line 3695 "load_abc.cpp"
   $511 = ($509<<24>>24)==($delim<<24>>24); //@line 3695 "load_abc.cpp"
   $or$cond2 = $510 | $511; //@line 3695 "load_abc.cpp"
   if (!($or$cond2)) {
    $512 = $509 << 24 >> 24; //@line 3695 "load_abc.cpp"
    $513 = (_isspace($512)|0); //@line 3695 "load_abc.cpp"
    $514 = ($513|0)==(0); //@line 3695 "load_abc.cpp"
    if ($514) {
     break;
    }
   }
   HEAP8[$391>>0]=0&255;HEAP8[$391+1>>0]=(0>>8)&255;HEAP8[$391+2>>0]=(0>>16)&255;HEAP8[$391+3>>0]=0>>24; //@line 3695 "load_abc.cpp"
  }
 } while(0);
 $515 = (_strncmp($p,526877,6)|0); //@line 3696 "load_abc.cpp"
 $516 = ($515|0)==(0); //@line 3696 "load_abc.cpp"
 do {
  if ($516) {
   $517 = ((($p)) + 6|0); //@line 3696 "load_abc.cpp"
   $518 = HEAP8[$517>>0]|0; //@line 3696 "load_abc.cpp"
   $519 = ($518<<24>>24)==(0); //@line 3696 "load_abc.cpp"
   $520 = ($518<<24>>24)==($delim<<24>>24); //@line 3696 "load_abc.cpp"
   $or$cond3 = $519 | $520; //@line 3696 "load_abc.cpp"
   if (!($or$cond3)) {
    $521 = $518 << 24 >> 24; //@line 3696 "load_abc.cpp"
    $522 = (_isspace($521)|0); //@line 3696 "load_abc.cpp"
    $523 = ($522|0)==(0); //@line 3696 "load_abc.cpp"
    if ($523) {
     label = 181;
     break;
    }
   }
   HEAP8[$391>>0]=1&255;HEAP8[$391+1>>0]=(1>>8)&255;HEAP8[$391+2>>0]=(1>>16)&255;HEAP8[$391+3>>0]=1>>24; //@line 3696 "load_abc.cpp"
   $525 = 1;
  } else {
   label = 181;
  }
 } while(0);
 if ((label|0) == 181) {
  $$pr = HEAPU8[$391>>0]|(HEAPU8[$391+1>>0]<<8)|(HEAPU8[$391+2>>0]<<16)|(HEAPU8[$391+3>>0]<<24); //@line 3697 "load_abc.cpp"
  $525 = $$pr;
 }
 $524 = ($392|0)==($525|0); //@line 3697 "load_abc.cpp"
 do {
  if (!($524)) {
   $526 = ($525|0)==(0); //@line 3698 "load_abc.cpp"
   $$phi$trans$insert = ((($h)) + 420|0);
   $$pre313 = HEAPU8[$$phi$trans$insert>>0]|(HEAPU8[$$phi$trans$insert+1>>0]<<8)|(HEAPU8[$$phi$trans$insert+2>>0]<<16)|(HEAPU8[$$phi$trans$insert+3>>0]<<24); //@line 3699 "load_abc.cpp"
   if ($526) {
    $532 = $$pre313;
    label = 186;
   } else {
    $527 = ($$pre313|0)==(0|0); //@line 3698 "load_abc.cpp"
    if ($527) {
     $528 = ((($h)) + 8|0); //@line 3698 "load_abc.cpp"
     $529 = HEAPU8[$528>>0]|(HEAPU8[$528+1>>0]<<8)|(HEAPU8[$528+2>>0]<<16)|(HEAPU8[$528+3>>0]<<24); //@line 3698 "load_abc.cpp"
     HEAP8[$$phi$trans$insert>>0]=$529&255;HEAP8[$$phi$trans$insert+1>>0]=($529>>8)&255;HEAP8[$$phi$trans$insert+2>>0]=($529>>16)&255;HEAP8[$$phi$trans$insert+3>>0]=$529>>24; //@line 3698 "load_abc.cpp"
     $530 = $529;
     $532 = $530;
     label = 186;
    } else {
     $536 = $$pre313;
    }
   }
   if ((label|0) == 186) {
    $531 = ($532|0)==(0|0); //@line 3699 "load_abc.cpp"
    if ($531) {
     break;
    } else {
     $536 = $532;
    }
   }
   $533 = ((($h)) + 424|0); //@line 3699 "load_abc.cpp"
   $534 = HEAPU8[$533>>0]|(HEAPU8[$533+1>>0]<<8)|(HEAPU8[$533+2>>0]<<16)|(HEAPU8[$533+3>>0]<<24); //@line 3699 "load_abc.cpp"
   $535 = ((($536)) + 44|0); //@line 1616 "load_abc.cpp"
   $537 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,$535,8)|0); //@line 1616 "load_abc.cpp"
   $538 = ((($537)) + 8|0); //@line 1590 "load_abc.cpp"
   $539 = HEAPU8[$538>>0]|(HEAPU8[$538+1>>0]<<8)|(HEAPU8[$538+2>>0]<<16)|(HEAPU8[$538+3>>0]<<24); //@line 1590 "load_abc.cpp"
   $540 = ($539|0)==(0|0); //@line 1591 "load_abc.cpp"
   do {
    if ($540) {
     label = 192;
    } else {
     $541 = ((($539)) + 4|0); //@line 1591 "load_abc.cpp"
     $542 = HEAPU8[$541>>0]|(HEAPU8[$541+1>>0]<<8)|(HEAPU8[$541+2>>0]<<16)|(HEAPU8[$541+3>>0]<<24); //@line 1591 "load_abc.cpp"
     $543 = ($542|0)==($534|0); //@line 1591 "load_abc.cpp"
     if (!($543)) {
      $544 = ((($539)) + 8|0); //@line 1592 "load_abc.cpp"
      $545 = HEAP8[$544>>0]|0; //@line 1592 "load_abc.cpp"
      $546 = ($545<<24>>24)==(1); //@line 1592 "load_abc.cpp"
      if ($546) {
       $547 = ((($539)) + 9|0); //@line 1592 "load_abc.cpp"
       $548 = HEAP8[$547>>0]|0; //@line 1592 "load_abc.cpp"
       $549 = ($548<<24>>24)==(83); //@line 1592 "load_abc.cpp"
       if (!($549)) {
        label = 192;
        break;
       }
       HEAP8[$541>>0]=$534&255;HEAP8[$541+1>>0]=($534>>8)&255;HEAP8[$541+2>>0]=($534>>16)&255;HEAP8[$541+3>>0]=$534>>24; //@line 1593 "load_abc.cpp"
      } else {
       label = 192;
      }
     }
    }
   } while(0);
   if ((label|0) == 192) {
    $550 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
    HEAP8[$550>>0]=0&255;HEAP8[$550+1>>0]=(0>>8)&255;HEAP8[$550+2>>0]=(0>>16)&255;HEAP8[$550+3>>0]=0>>24; //@line 540 "load_abc.cpp"
    $551 = ((($550)) + 4|0); //@line 541 "load_abc.cpp"
    HEAP8[$551>>0]=$534&255;HEAP8[$551+1>>0]=($534>>8)&255;HEAP8[$551+2>>0]=($534>>16)&255;HEAP8[$551+3>>0]=$534>>24; //@line 541 "load_abc.cpp"
    $552 = ((($550)) + 8|0); //@line 543 "load_abc.cpp"
    HEAP8[$552>>0] = 1; //@line 543 "load_abc.cpp"
    $553 = ((($550)) + 9|0); //@line 543 "load_abc.cpp"
    HEAP8[$553>>0] = 83; //@line 543 "load_abc.cpp"
    $554 = ((($550)) + 10|0); //@line 543 "load_abc.cpp"
    HEAP8[$554>>0]=0&255;HEAP8[$554+1>>0]=(0>>8)&255;HEAP8[$554+2>>0]=(0>>16)&255;HEAP8[$554+3>>0]=0>>24; //@line 543 "load_abc.cpp"
    $555 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
    $556 = ((($550)) + 14|0); //@line 544 "load_abc.cpp"
    HEAP8[$556>>0] = $555; //@line 544 "load_abc.cpp"
    $557 = ((($550)) + 15|0); //@line 545 "load_abc.cpp"
    HEAP8[$557>>0] = 0; //@line 545 "load_abc.cpp"
    __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($537,$550); //@line 1600 "load_abc.cpp"
   }
   $$pr176 = HEAPU8[$$phi$trans$insert>>0]|(HEAPU8[$$phi$trans$insert+1>>0]<<8)|(HEAPU8[$$phi$trans$insert+2>>0]<<16)|(HEAPU8[$$phi$trans$insert+3>>0]<<24); //@line 3700 "load_abc.cpp"
   $558 = ($$pr176|0)==(0|0); //@line 3700 "load_abc.cpp"
   if (!($558)) {
    $559 = HEAPU8[$391>>0]|(HEAPU8[$391+1>>0]<<8)|(HEAPU8[$391+2>>0]<<16)|(HEAPU8[$391+3>>0]<<24); //@line 3700 "load_abc.cpp"
    $560 = ($559|0)==(0); //@line 3700 "load_abc.cpp"
    if ($560) {
     HEAP8[$$phi$trans$insert>>0]=0&255;HEAP8[$$phi$trans$insert+1>>0]=(0>>8)&255;HEAP8[$$phi$trans$insert+2>>0]=(0>>16)&255;HEAP8[$$phi$trans$insert+3>>0]=0>>24; //@line 3700 "load_abc.cpp"
    }
   }
  }
 } while(0);
 $561 = ((($h)) + 384|0); //@line 3702 "load_abc.cpp"
 $562 = HEAPU8[$561>>0]|(HEAPU8[$561+1>>0]<<8)|(HEAPU8[$561+2>>0]<<16)|(HEAPU8[$561+3>>0]<<24); //@line 3702 "load_abc.cpp"
 $563 = (_strncmp($p,526884,6)|0); //@line 3703 "load_abc.cpp"
 $564 = ($563|0)==(0); //@line 3703 "load_abc.cpp"
 do {
  if ($564) {
   $565 = ((($p)) + 6|0); //@line 3703 "load_abc.cpp"
   $566 = HEAP8[$565>>0]|0; //@line 3703 "load_abc.cpp"
   $567 = ($566<<24>>24)==(0); //@line 3703 "load_abc.cpp"
   $568 = ($566<<24>>24)==($delim<<24>>24); //@line 3703 "load_abc.cpp"
   $or$cond4 = $567 | $568; //@line 3703 "load_abc.cpp"
   if (!($or$cond4)) {
    $569 = $566 << 24 >> 24; //@line 3703 "load_abc.cpp"
    $570 = (_isspace($569)|0); //@line 3703 "load_abc.cpp"
    $571 = ($570|0)==(0); //@line 3703 "load_abc.cpp"
    if ($571) {
     break;
    }
   }
   $572 = (__ZL15abc_MIDI_gchordPKcP10_ABCHANDLE($565,$h)|0); //@line 3704 "load_abc.cpp"
   HEAP8[$561>>0]=$572&255;HEAP8[$561+1>>0]=($572>>8)&255;HEAP8[$561+2>>0]=($572>>16)&255;HEAP8[$561+3>>0]=$572>>24; //@line 3704 "load_abc.cpp"
   $573 = ($572|0)==(0); //@line 3705 "load_abc.cpp"
   if ($573) {
    HEAP8[$561>>0]=$562&255;HEAP8[$561+1>>0]=($562>>8)&255;HEAP8[$561+2>>0]=($562>>16)&255;HEAP8[$561+3>>0]=$562>>24; //@line 3706 "load_abc.cpp"
    break;
   } else {
    $574 = (($572) + -1)|0; //@line 3705 "load_abc.cpp"
    HEAP8[$561>>0]=$574&255;HEAP8[$561+1>>0]=($574>>8)&255;HEAP8[$561+2>>0]=($574>>16)&255;HEAP8[$561+3>>0]=$574>>24; //@line 3705 "load_abc.cpp"
    break;
   }
  }
 } while(0);
 $575 = (_strncmp($p,526891,9)|0); //@line 3708 "load_abc.cpp"
 $576 = ($575|0)==(0); //@line 3708 "load_abc.cpp"
 do {
  if ($576) {
   $577 = ((($p)) + 9|0); //@line 3708 "load_abc.cpp"
   $578 = HEAP8[$577>>0]|0; //@line 3708 "load_abc.cpp"
   $579 = ($578<<24>>24)==(0); //@line 3708 "load_abc.cpp"
   $580 = ($578<<24>>24)==($delim<<24>>24); //@line 3708 "load_abc.cpp"
   $or$cond5 = $579 | $580; //@line 3708 "load_abc.cpp"
   if (!($or$cond5)) {
    $581 = $578 << 24 >> 24; //@line 3708 "load_abc.cpp"
    $582 = (_isspace($581)|0); //@line 3708 "load_abc.cpp"
    $583 = ($582|0)==(0); //@line 3708 "load_abc.cpp"
    if ($583) {
     break;
    }
   }
   HEAP8[$561>>0]=0&255;HEAP8[$561+1>>0]=(0>>8)&255;HEAP8[$561+2>>0]=(0>>16)&255;HEAP8[$561+3>>0]=0>>24; //@line 3708 "load_abc.cpp"
  }
 } while(0);
 $584 = (_strncmp($p,526901,8)|0); //@line 3709 "load_abc.cpp"
 $585 = ($584|0)==(0); //@line 3709 "load_abc.cpp"
 do {
  if ($585) {
   $586 = ((($p)) + 8|0); //@line 3709 "load_abc.cpp"
   $587 = HEAP8[$586>>0]|0; //@line 3709 "load_abc.cpp"
   $588 = ($587<<24>>24)==(0); //@line 3709 "load_abc.cpp"
   $589 = ($587<<24>>24)==($delim<<24>>24); //@line 3709 "load_abc.cpp"
   $or$cond6 = $588 | $589; //@line 3709 "load_abc.cpp"
   if (!($or$cond6)) {
    $590 = $587 << 24 >> 24; //@line 3709 "load_abc.cpp"
    $591 = (_isspace($590)|0); //@line 3709 "load_abc.cpp"
    $592 = ($591|0)==(0); //@line 3709 "load_abc.cpp"
    if ($592) {
     break;
    }
   }
   HEAP8[$561>>0]=1&255;HEAP8[$561+1>>0]=(1>>8)&255;HEAP8[$561+2>>0]=(1>>16)&255;HEAP8[$561+3>>0]=1>>24; //@line 3709 "load_abc.cpp"
  }
 } while(0);
 $593 = HEAPU8[$561>>0]|(HEAPU8[$561+1>>0]<<8)|(HEAPU8[$561+2>>0]<<16)|(HEAPU8[$561+3>>0]<<24); //@line 3710 "load_abc.cpp"
 $594 = ($562|0)==($593|0); //@line 3710 "load_abc.cpp"
 if (!($594)) {
  $595 = ((($h)) + 416|0); //@line 3711 "load_abc.cpp"
  $596 = HEAPU8[$595>>0]|(HEAPU8[$595+1>>0]<<8)|(HEAPU8[$595+2>>0]<<16)|(HEAPU8[$595+3>>0]<<24); //@line 3711 "load_abc.cpp"
  $597 = ($596|0)==(0|0); //@line 3711 "load_abc.cpp"
  if (!($597)) {
   $598 = ((($h)) + 424|0); //@line 3711 "load_abc.cpp"
   $599 = HEAPU8[$598>>0]|(HEAPU8[$598+1>>0]<<8)|(HEAPU8[$598+2>>0]<<16)|(HEAPU8[$598+3>>0]<<24); //@line 3711 "load_abc.cpp"
   $600 = ((($596)) + 44|0); //@line 1608 "load_abc.cpp"
   $i$01$i = 1;
   while(1) {
    $601 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,$600,$i$01$i)|0); //@line 1608 "load_abc.cpp"
    $602 = ((($601)) + 8|0); //@line 1590 "load_abc.cpp"
    $603 = HEAPU8[$602>>0]|(HEAPU8[$602+1>>0]<<8)|(HEAPU8[$602+2>>0]<<16)|(HEAPU8[$602+3>>0]<<24); //@line 1590 "load_abc.cpp"
    $604 = ($603|0)==(0|0); //@line 1591 "load_abc.cpp"
    do {
     if ($604) {
      label = 218;
     } else {
      $605 = ((($603)) + 4|0); //@line 1591 "load_abc.cpp"
      $606 = HEAPU8[$605>>0]|(HEAPU8[$605+1>>0]<<8)|(HEAPU8[$605+2>>0]<<16)|(HEAPU8[$605+3>>0]<<24); //@line 1591 "load_abc.cpp"
      $607 = ($606|0)==($599|0); //@line 1591 "load_abc.cpp"
      if ($607) {
       break;
      }
      $608 = ((($603)) + 8|0); //@line 1592 "load_abc.cpp"
      $609 = HEAP8[$608>>0]|0; //@line 1592 "load_abc.cpp"
      $610 = ($609<<24>>24)==(1); //@line 1592 "load_abc.cpp"
      if (!($610)) {
       label = 218;
       break;
      }
      $611 = ((($603)) + 9|0); //@line 1592 "load_abc.cpp"
      $612 = HEAP8[$611>>0]|0; //@line 1592 "load_abc.cpp"
      $613 = ($612<<24>>24)==(83); //@line 1592 "load_abc.cpp"
      if (!($613)) {
       label = 218;
       break;
      }
      HEAP8[$605>>0]=$599&255;HEAP8[$605+1>>0]=($599>>8)&255;HEAP8[$605+2>>0]=($599>>16)&255;HEAP8[$605+3>>0]=$599>>24; //@line 1593 "load_abc.cpp"
     }
    } while(0);
    if ((label|0) == 218) {
     label = 0;
     $614 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
     HEAP8[$614>>0]=0&255;HEAP8[$614+1>>0]=(0>>8)&255;HEAP8[$614+2>>0]=(0>>16)&255;HEAP8[$614+3>>0]=0>>24; //@line 540 "load_abc.cpp"
     $615 = ((($614)) + 4|0); //@line 541 "load_abc.cpp"
     HEAP8[$615>>0]=$599&255;HEAP8[$615+1>>0]=($599>>8)&255;HEAP8[$615+2>>0]=($599>>16)&255;HEAP8[$615+3>>0]=$599>>24; //@line 541 "load_abc.cpp"
     $616 = ((($614)) + 8|0); //@line 543 "load_abc.cpp"
     HEAP8[$616>>0] = 1; //@line 543 "load_abc.cpp"
     $617 = ((($614)) + 9|0); //@line 543 "load_abc.cpp"
     HEAP8[$617>>0] = 83; //@line 543 "load_abc.cpp"
     $618 = ((($614)) + 10|0); //@line 543 "load_abc.cpp"
     HEAP8[$618>>0]=0&255;HEAP8[$618+1>>0]=(0>>8)&255;HEAP8[$618+2>>0]=(0>>16)&255;HEAP8[$618+3>>0]=0>>24; //@line 543 "load_abc.cpp"
     $619 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
     $620 = ((($614)) + 14|0); //@line 544 "load_abc.cpp"
     HEAP8[$620>>0] = $619; //@line 544 "load_abc.cpp"
     $621 = ((($614)) + 15|0); //@line 545 "load_abc.cpp"
     HEAP8[$621>>0] = 0; //@line 545 "load_abc.cpp"
     __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($601,$614); //@line 1600 "load_abc.cpp"
    }
    $622 = (($i$01$i) + 1)|0; //@line 1607 "load_abc.cpp"
    $exitcond$i100 = ($622|0)==(8); //@line 1607 "load_abc.cpp"
    if ($exitcond$i100) {
     break;
    } else {
     $i$01$i = $622;
    }
   }
  }
 }
 $623 = (_strncmp($p,526910,7)|0); //@line 3713 "load_abc.cpp"
 $624 = ($623|0)==(0); //@line 3713 "load_abc.cpp"
 do {
  if ($624) {
   $625 = ((($p)) + 7|0); //@line 3713 "load_abc.cpp"
   $626 = HEAP8[$625>>0]|0; //@line 3713 "load_abc.cpp"
   $627 = $626 << 24 >> 24; //@line 3713 "load_abc.cpp"
   $628 = (_isspace($627)|0); //@line 3713 "load_abc.cpp"
   $629 = ($628|0)==(0); //@line 3713 "load_abc.cpp"
   if ($629) {
    break;
   }
   $630 = ((($p)) + 8|0); //@line 3714 "load_abc.cpp"
   $631 = ((($h)) + 412|0); //@line 3714 "load_abc.cpp"
   $632 = HEAPU8[$631>>0]|(HEAPU8[$631+1>>0]<<8)|(HEAPU8[$631+2>>0]<<16)|(HEAPU8[$631+3>>0]<<24); //@line 3714 "load_abc.cpp"
   $633 = ($632|0)==(0|0); //@line 915 "load_abc.cpp"
   if ($633) {
    $634 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,1282466,0)|0); //@line 916 "load_abc.cpp"
    $635 = ((($h)) + 376|0); //@line 917 "load_abc.cpp"
    $636 = HEAPU8[$635>>0]|(HEAPU8[$635+1>>0]<<8)|(HEAPU8[$635+2>>0]<<16)|(HEAPU8[$635+3>>0]<<24); //@line 917 "load_abc.cpp"
    $637 = ((($634)) + 20|0); //@line 917 "load_abc.cpp"
    HEAP8[$637>>0]=$636&255;HEAP8[$637+1>>0]=($636>>8)&255;HEAP8[$637+2>>0]=($636>>16)&255;HEAP8[$637+3>>0]=$636>>24; //@line 917 "load_abc.cpp"
    HEAP8[$631>>0]=$634&255;HEAP8[$631+1>>0]=($634>>8)&255;HEAP8[$631+2>>0]=($634>>16)&255;HEAP8[$631+3>>0]=$634>>24; //@line 3714 "load_abc.cpp"
    $638 = ($634|0)==(0|0); //@line 3098 "load_abc.cpp"
    if ($638) {
     $663 = 1;
    } else {
     $$0$i101177 = $634;
     label = 225;
    }
   } else {
    HEAP8[$631>>0]=$632&255;HEAP8[$631+1>>0]=($632>>8)&255;HEAP8[$631+2>>0]=($632>>16)&255;HEAP8[$631+3>>0]=$632>>24; //@line 3714 "load_abc.cpp"
    $$0$i101177 = $632;
    label = 225;
   }
   if ((label|0) == 225) {
    $639 = ((($$0$i101177)) + 36|0); //@line 3098 "load_abc.cpp"
    $640 = HEAP8[$639>>0]|0; //@line 3098 "load_abc.cpp"
    $641 = $640&255; //@line 3098 "load_abc.cpp"
    $663 = $641;
   }
   $642 = HEAP8[$630>>0]|0; //@line 3099 "load_abc.cpp"
   $643 = ($642<<24>>24)==(0); //@line 3099 "load_abc.cpp"
   L363: do {
    if ($643) {
     $$0$lcssa$i = $630;$652 = 0;
    } else {
     $$05$i = $630;$645 = $642;
     while(1) {
      $644 = $645 << 24 >> 24; //@line 3099 "load_abc.cpp"
      $646 = (_isspace($644)|0); //@line 3099 "load_abc.cpp"
      $647 = ($646|0)==(0); //@line 3099 "load_abc.cpp"
      if ($647) {
       $$05$i$lcssa = $$05$i;
       break;
      }
      $648 = ((($$05$i)) + 1|0); //@line 3099 "load_abc.cpp"
      $649 = HEAP8[$648>>0]|0; //@line 3099 "load_abc.cpp"
      $650 = ($649<<24>>24)==(0); //@line 3099 "load_abc.cpp"
      if ($650) {
       $$0$lcssa$i = $648;$652 = 0;
       break L363;
      } else {
       $$05$i = $648;$645 = $649;
      }
     }
     $$pre$i103 = HEAP8[$$05$i$lcssa>>0]|0; //@line 3100 "load_abc.cpp"
     $$0$lcssa$i = $$05$i$lcssa;$652 = $$pre$i103;
    }
   } while(0);
   $651 = $652 << 24 >> 24; //@line 3100 "load_abc.cpp"
   $isdigittmp$i104 = (($651) + -48)|0; //@line 3100 "load_abc.cpp"
   $isdigit$i105 = ($isdigittmp$i104>>>0)<(10); //@line 3100 "load_abc.cpp"
   if ($isdigit$i105) {
    $655 = $651;$h$04$i$i107 = 0;$i$03$i$i108 = 0;
   } else {
    break;
   }
   while(1) {
    $653 = ($h$04$i$i107*10)|0; //@line 1626 "load_abc.cpp"
    $654 = (($655) + -48)|0; //@line 1626 "load_abc.cpp"
    $656 = (($654) + ($653))|0; //@line 1626 "load_abc.cpp"
    $657 = (($i$03$i$i108) + 1)|0; //@line 1627 "load_abc.cpp"
    $658 = (($$0$lcssa$i) + ($657)|0); //@line 1625 "load_abc.cpp"
    $659 = HEAP8[$658>>0]|0; //@line 1625 "load_abc.cpp"
    $660 = $659 << 24 >> 24; //@line 1625 "load_abc.cpp"
    $isdigittmp$i$i109 = (($660) + -48)|0; //@line 1625 "load_abc.cpp"
    $isdigit$i$i110 = ($isdigittmp$i$i109>>>0)<(10); //@line 1625 "load_abc.cpp"
    if ($isdigit$i$i110) {
     $655 = $660;$h$04$i$i107 = $656;$i$03$i$i108 = $657;
    } else {
     $$lcssa542 = $656;$$lcssa543 = $657;
     break;
    }
   }
   $661 = ($$lcssa543|0)==(0); //@line 1629 "load_abc.cpp"
   $$h$0$i$i112 = $661 ? 1 : $$lcssa542; //@line 1630 "load_abc.cpp"
   $$off$i113 = (($$h$0$i$i112) + -1)|0; //@line 3102 "load_abc.cpp"
   $662 = ($$off$i113>>>0)<(16); //@line 3102 "load_abc.cpp"
   $tno$off$i$i = (($663) + -1)|0; //@line 3714 "load_abc.cpp"
   $664 = ($tno$off$i$i>>>0)<(32); //@line 3085 "load_abc.cpp"
   $or$cond$i114 = $664 & $662; //@line 3102 "load_abc.cpp"
   if (!($or$cond$i114)) {
    break;
   }
   $665 = ((($h)) + 8|0); //@line 3086 "load_abc.cpp"
   $tp$02$i$i = HEAPU8[$665>>0]|(HEAPU8[$665+1>>0]<<8)|(HEAPU8[$665+2>>0]<<16)|(HEAPU8[$665+3>>0]<<24); //@line 3086 "load_abc.cpp"
   $666 = ($tp$02$i$i|0)==(0|0); //@line 3086 "load_abc.cpp"
   if ($666) {
    break;
   }
   $667 = $$h$0$i$i112&255; //@line 3088 "load_abc.cpp"
   $tp$03$i$i = $tp$02$i$i;
   while(1) {
    $668 = ((($tp$03$i$i)) + 36|0); //@line 3087 "load_abc.cpp"
    $669 = HEAP8[$668>>0]|0; //@line 3087 "load_abc.cpp"
    $670 = $669&255; //@line 3087 "load_abc.cpp"
    $671 = ($670|0)==($663|0); //@line 3087 "load_abc.cpp"
    do {
     if ($671) {
      $672 = ((($tp$03$i$i)) + 37|0); //@line 3087 "load_abc.cpp"
      $673 = HEAP8[$672>>0]|0; //@line 3087 "load_abc.cpp"
      $674 = (($673) + -1)<<24>>24; //@line 3087 "load_abc.cpp"
      $675 = ($674&255)>(9); //@line 3087 "load_abc.cpp"
      if (!($675)) {
       break;
      }
      $676 = ((($tp$03$i$i)) + 40|0); //@line 3088 "load_abc.cpp"
      HEAP8[$676>>0] = $667; //@line 3088 "load_abc.cpp"
     }
    } while(0);
    $tp$0$i$i = HEAPU8[$tp$03$i$i>>0]|(HEAPU8[$tp$03$i$i+1>>0]<<8)|(HEAPU8[$tp$03$i$i+2>>0]<<16)|(HEAPU8[$tp$03$i$i+3>>0]<<24); //@line 3086 "load_abc.cpp"
    $677 = ($tp$0$i$i|0)==(0|0); //@line 3086 "load_abc.cpp"
    if ($677) {
     break;
    } else {
     $tp$03$i$i = $tp$0$i$i;
    }
   }
  }
 } while(0);
 $678 = (_strncmp($p,526918,7)|0); //@line 3715 "load_abc.cpp"
 $679 = ($678|0)==(0); //@line 3715 "load_abc.cpp"
 do {
  if ($679) {
   $680 = ((($p)) + 7|0); //@line 3715 "load_abc.cpp"
   $681 = HEAP8[$680>>0]|0; //@line 3715 "load_abc.cpp"
   $682 = $681 << 24 >> 24; //@line 3715 "load_abc.cpp"
   $683 = (_isspace($682)|0); //@line 3715 "load_abc.cpp"
   $684 = ($683|0)==(0); //@line 3715 "load_abc.cpp"
   if ($684) {
    break;
   }
   $685 = ((($p)) + 8|0); //@line 3716 "load_abc.cpp"
   $686 = ((($h)) + 412|0); //@line 3716 "load_abc.cpp"
   $687 = HEAPU8[$686>>0]|(HEAPU8[$686+1>>0]<<8)|(HEAPU8[$686+2>>0]<<16)|(HEAPU8[$686+3>>0]<<24); //@line 3716 "load_abc.cpp"
   $688 = ($687|0)==(0|0); //@line 915 "load_abc.cpp"
   if ($688) {
    $689 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,1282466,0)|0); //@line 916 "load_abc.cpp"
    $690 = ((($h)) + 376|0); //@line 917 "load_abc.cpp"
    $691 = HEAPU8[$690>>0]|(HEAPU8[$690+1>>0]<<8)|(HEAPU8[$690+2>>0]<<16)|(HEAPU8[$690+3>>0]<<24); //@line 917 "load_abc.cpp"
    $692 = ((($689)) + 20|0); //@line 917 "load_abc.cpp"
    HEAP8[$692>>0]=$691&255;HEAP8[$692+1>>0]=($691>>8)&255;HEAP8[$692+2>>0]=($691>>16)&255;HEAP8[$692+3>>0]=$691>>24; //@line 917 "load_abc.cpp"
    HEAP8[$686>>0]=$689&255;HEAP8[$686+1>>0]=($689>>8)&255;HEAP8[$686+2>>0]=($689>>16)&255;HEAP8[$686+3>>0]=$689>>24; //@line 3716 "load_abc.cpp"
    $693 = ($689|0)==(0|0); //@line 3123 "load_abc.cpp"
    if ($693) {
     $985 = 1;
    } else {
     $$0$i116178 = $689;
     label = 244;
    }
   } else {
    HEAP8[$686>>0]=$687&255;HEAP8[$686+1>>0]=($687>>8)&255;HEAP8[$686+2>>0]=($687>>16)&255;HEAP8[$686+3>>0]=$687>>24; //@line 3716 "load_abc.cpp"
    $$0$i116178 = $687;
    label = 244;
   }
   if ((label|0) == 244) {
    $694 = ((($$0$i116178)) + 36|0); //@line 3123 "load_abc.cpp"
    $695 = HEAP8[$694>>0]|0; //@line 3123 "load_abc.cpp"
    $696 = $695&255; //@line 3123 "load_abc.cpp"
    $985 = $696;
   }
   $697 = HEAP8[$685>>0]|0; //@line 3124 "load_abc.cpp"
   $698 = ($697<<24>>24)==(0); //@line 3124 "load_abc.cpp"
   L390: do {
    if ($698) {
     $$0$lcssa$i120 = $685;$707 = 0;
    } else {
     $$022$i = $685;$700 = $697;
     while(1) {
      $699 = $700 << 24 >> 24; //@line 3124 "load_abc.cpp"
      $701 = (_isspace($699)|0); //@line 3124 "load_abc.cpp"
      $702 = ($701|0)==(0); //@line 3124 "load_abc.cpp"
      if ($702) {
       $$022$i$lcssa = $$022$i;
       break;
      }
      $703 = ((($$022$i)) + 1|0); //@line 3124 "load_abc.cpp"
      $704 = HEAP8[$703>>0]|0; //@line 3124 "load_abc.cpp"
      $705 = ($704<<24>>24)==(0); //@line 3124 "load_abc.cpp"
      if ($705) {
       $$0$lcssa$i120 = $703;$707 = 0;
       break L390;
      } else {
       $$022$i = $703;$700 = $704;
      }
     }
     $$pre$i118 = HEAP8[$$022$i$lcssa>>0]|0; //@line 3125 "load_abc.cpp"
     $$0$lcssa$i120 = $$022$i$lcssa;$707 = $$pre$i118;
    }
   } while(0);
   $706 = $707 << 24 >> 24; //@line 3125 "load_abc.cpp"
   $isdigittmp$i121 = (($706) + -48)|0; //@line 3125 "load_abc.cpp"
   $isdigit$i122 = ($isdigittmp$i121>>>0)<(10); //@line 3125 "load_abc.cpp"
   if ($isdigit$i122) {
    $710 = $706;$h$04$i$i124 = 0;$i$03$i$i125 = 0;
   } else {
    break;
   }
   while(1) {
    $708 = ($h$04$i$i124*10)|0; //@line 1626 "load_abc.cpp"
    $709 = (($710) + -48)|0; //@line 1626 "load_abc.cpp"
    $711 = (($709) + ($708))|0; //@line 1626 "load_abc.cpp"
    $712 = (($i$03$i$i125) + 1)|0; //@line 1627 "load_abc.cpp"
    $713 = (($$0$lcssa$i120) + ($712)|0); //@line 1625 "load_abc.cpp"
    $714 = HEAP8[$713>>0]|0; //@line 1625 "load_abc.cpp"
    $715 = $714 << 24 >> 24; //@line 1625 "load_abc.cpp"
    $isdigittmp$i$i126 = (($715) + -48)|0; //@line 1625 "load_abc.cpp"
    $isdigit$i$i127 = ($isdigittmp$i$i126>>>0)<(10); //@line 1625 "load_abc.cpp"
    if ($isdigit$i$i127) {
     $710 = $715;$h$04$i$i124 = $711;$i$03$i$i125 = $712;
    } else {
     $$lcssa536 = $711;$$lcssa537 = $712;$$lcssa538 = $713;$$lcssa539 = $714;
     break;
    }
   }
   $716 = ($$lcssa537|0)==(0); //@line 1629 "load_abc.cpp"
   $$h$0$i$i129 = $716 ? 1 : $$lcssa536; //@line 1630 "load_abc.cpp"
   $717 = ($$lcssa539<<24>>24)==(0); //@line 3127 "load_abc.cpp"
   L399: do {
    if ($717) {
     $$1$lcssa$i132 = $$lcssa538;$726 = 0;
    } else {
     $$119$i = $$lcssa538;$719 = $$lcssa539;
     while(1) {
      $718 = $719 << 24 >> 24; //@line 3127 "load_abc.cpp"
      $720 = (_isspace($718)|0); //@line 3127 "load_abc.cpp"
      $721 = ($720|0)==(0); //@line 3127 "load_abc.cpp"
      if ($721) {
       $$119$i$lcssa = $$119$i;
       break;
      }
      $722 = ((($$119$i)) + 1|0); //@line 3127 "load_abc.cpp"
      $723 = HEAP8[$722>>0]|0; //@line 3127 "load_abc.cpp"
      $724 = ($723<<24>>24)==(0); //@line 3127 "load_abc.cpp"
      if ($724) {
       $$1$lcssa$i132 = $722;$726 = 0;
       break L399;
      } else {
       $$119$i = $722;$719 = $723;
      }
     }
     $$pre34$i = HEAP8[$$119$i$lcssa>>0]|0; //@line 3128 "load_abc.cpp"
     $$1$lcssa$i132 = $$119$i$lcssa;$726 = $$pre34$i;
    }
   } while(0);
   $725 = $726 << 24 >> 24; //@line 3128 "load_abc.cpp"
   $isdigittmp2$i133 = (($725) + -48)|0; //@line 3128 "load_abc.cpp"
   $isdigit3$i134 = ($isdigittmp2$i133>>>0)<(10); //@line 3128 "load_abc.cpp"
   if ($isdigit3$i134) {
    $729 = $725;$h$04$i7$i = 0;$i$03$i8$i = 0;
    while(1) {
     $727 = ($h$04$i7$i*10)|0; //@line 1626 "load_abc.cpp"
     $728 = (($729) + -48)|0; //@line 1626 "load_abc.cpp"
     $730 = (($728) + ($727))|0; //@line 1626 "load_abc.cpp"
     $731 = (($i$03$i8$i) + 1)|0; //@line 1627 "load_abc.cpp"
     $732 = (($$1$lcssa$i132) + ($731)|0); //@line 1625 "load_abc.cpp"
     $733 = HEAP8[$732>>0]|0; //@line 1625 "load_abc.cpp"
     $734 = $733 << 24 >> 24; //@line 1625 "load_abc.cpp"
     $isdigittmp$i9$i = (($734) + -48)|0; //@line 1625 "load_abc.cpp"
     $isdigit$i10$i = ($isdigittmp$i9$i>>>0)<(10); //@line 1625 "load_abc.cpp"
     if ($isdigit$i10$i) {
      $729 = $734;$h$04$i7$i = $730;$i$03$i8$i = $731;
     } else {
      $$lcssa532 = $730;$$lcssa533 = $731;
      break;
     }
    }
    $735 = ($$lcssa533|0)==(0); //@line 1629 "load_abc.cpp"
    $$h$0$i14$i = $735 ? 1 : $$lcssa532; //@line 1630 "load_abc.cpp"
    $737 = $$h$0$i14$i;$i1$0$i = $$h$0$i$i129;
   } else {
    $737 = $$h$0$i$i129;$i1$0$i = $985;
   }
   $736 = (($737) + 1)|0; //@line 3132 "load_abc.cpp"
   $tno$off$i$i136 = (($i1$0$i) + -1)|0; //@line 3110 "load_abc.cpp"
   $738 = ($tno$off$i$i136>>>0)<(32); //@line 3110 "load_abc.cpp"
   $739 = ($737>>>0)<(128); //@line 3110 "load_abc.cpp"
   $740 = $739 & $738; //@line 3110 "load_abc.cpp"
   if (!($740)) {
    break;
   }
   $741 = ((($h)) + 8|0); //@line 3111 "load_abc.cpp"
   $tp$06$i$i = HEAPU8[$741>>0]|(HEAPU8[$741+1>>0]<<8)|(HEAPU8[$741+2>>0]<<16)|(HEAPU8[$741+3>>0]<<24); //@line 3111 "load_abc.cpp"
   $742 = ($tp$06$i$i|0)==(0|0); //@line 3111 "load_abc.cpp"
   if ($742) {
    break;
   }
   $743 = $736&255; //@line 3113 "load_abc.cpp"
   $tp$07$i$i = $tp$06$i$i;
   while(1) {
    $744 = ((($tp$07$i$i)) + 36|0); //@line 3112 "load_abc.cpp"
    $745 = HEAP8[$744>>0]|0; //@line 3112 "load_abc.cpp"
    $746 = $745&255; //@line 3112 "load_abc.cpp"
    $747 = ($746|0)==($i1$0$i|0); //@line 3112 "load_abc.cpp"
    do {
     if ($747) {
      $748 = ((($tp$07$i$i)) + 37|0); //@line 3112 "load_abc.cpp"
      $749 = HEAP8[$748>>0]|0; //@line 3112 "load_abc.cpp"
      $750 = (($749) + -1)<<24>>24; //@line 3112 "load_abc.cpp"
      $751 = ($750&255)>(9); //@line 3112 "load_abc.cpp"
      if (!($751)) {
       break;
      }
      $752 = ((($tp$07$i$i)) + 42|0); //@line 3113 "load_abc.cpp"
      HEAP8[$752>>0] = $743; //@line 3113 "load_abc.cpp"
     }
    } while(0);
    $tp$0$i$i137 = HEAPU8[$tp$07$i$i>>0]|(HEAPU8[$tp$07$i$i+1>>0]<<8)|(HEAPU8[$tp$07$i$i+2>>0]<<16)|(HEAPU8[$tp$07$i$i+3>>0]<<24); //@line 3111 "load_abc.cpp"
    $753 = ($tp$0$i$i137|0)==(0|0); //@line 3111 "load_abc.cpp"
    if ($753) {
     break;
    } else {
     $tp$07$i$i = $tp$0$i$i137;
    }
   }
  }
 } while(0);
 $754 = (_strncmp($p,526926,5)|0); //@line 3717 "load_abc.cpp"
 $755 = ($754|0)==(0); //@line 3717 "load_abc.cpp"
 do {
  if ($755) {
   $756 = ((($p)) + 5|0); //@line 3717 "load_abc.cpp"
   $757 = HEAP8[$756>>0]|0; //@line 3717 "load_abc.cpp"
   $758 = $757 << 24 >> 24; //@line 3717 "load_abc.cpp"
   $759 = (_isspace($758)|0); //@line 3717 "load_abc.cpp"
   $760 = ($759|0)==(0); //@line 3717 "load_abc.cpp"
   if ($760) {
    break;
   }
   $761 = ((($p)) + 6|0); //@line 3718 "load_abc.cpp"
   $762 = ((($h)) + 412|0); //@line 3718 "load_abc.cpp"
   $763 = HEAPU8[$762>>0]|(HEAPU8[$762+1>>0]<<8)|(HEAPU8[$762+2>>0]<<16)|(HEAPU8[$762+3>>0]<<24); //@line 3718 "load_abc.cpp"
   $764 = ($763|0)==(0|0); //@line 915 "load_abc.cpp"
   if ($764) {
    $765 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,1282466,0)|0); //@line 916 "load_abc.cpp"
    $766 = ((($h)) + 376|0); //@line 917 "load_abc.cpp"
    $767 = HEAPU8[$766>>0]|(HEAPU8[$766+1>>0]<<8)|(HEAPU8[$766+2>>0]<<16)|(HEAPU8[$766+3>>0]<<24); //@line 917 "load_abc.cpp"
    $768 = ((($765)) + 20|0); //@line 917 "load_abc.cpp"
    HEAP8[$768>>0]=$767&255;HEAP8[$768+1>>0]=($767>>8)&255;HEAP8[$768+2>>0]=($767>>16)&255;HEAP8[$768+3>>0]=$767>>24; //@line 917 "load_abc.cpp"
    $$0$i138 = $765;
   } else {
    $$0$i138 = $763;
   }
   HEAP8[$762>>0]=$$0$i138&255;HEAP8[$762+1>>0]=($$0$i138>>8)&255;HEAP8[$762+2>>0]=($$0$i138>>16)&255;HEAP8[$762+3>>0]=$$0$i138>>24; //@line 3718 "load_abc.cpp"
   $769 = HEAP8[$761>>0]|0; //@line 3149 "load_abc.cpp"
   $770 = ($769<<24>>24)==(0); //@line 3149 "load_abc.cpp"
   L425: do {
    if ($770) {
     $$0$lcssa$i140 = $761;
    } else {
     $$040$i = $761;$772 = $769;
     while(1) {
      $771 = $772 << 24 >> 24; //@line 3149 "load_abc.cpp"
      $773 = (_isspace($771)|0); //@line 3149 "load_abc.cpp"
      $774 = ($773|0)==(0); //@line 3149 "load_abc.cpp"
      if ($774) {
       $$0$lcssa$i140 = $$040$i;
       break L425;
      }
      $775 = ((($$040$i)) + 1|0); //@line 3149 "load_abc.cpp"
      $776 = HEAP8[$775>>0]|0; //@line 3149 "load_abc.cpp"
      $777 = ($776<<24>>24)==(0); //@line 3149 "load_abc.cpp"
      if ($777) {
       $$0$lcssa$i140 = $775;
       break;
      } else {
       $$040$i = $775;$772 = $776;
      }
     }
    }
   } while(0);
   $778 = (_strncmp($$0$lcssa$i140,526932,11)|0); //@line 3150 "load_abc.cpp"
   $779 = ($778|0)==(0); //@line 3150 "load_abc.cpp"
   L430: do {
    if ($779) {
     $$07$i = $$0$i138;$$3$i146 = $$0$lcssa$i140;
    } else {
     $780 = (_strncmp($$0$lcssa$i140,526944,4)|0); //@line 3150 "load_abc.cpp"
     $781 = ($780|0)==(0); //@line 3150 "load_abc.cpp"
     if ($781) {
      $$07$i = $$0$i138;$$3$i146 = $$0$lcssa$i140;
      break;
     }
     $782 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,$$0$lcssa$i140,0)|0); //@line 3151 "load_abc.cpp"
     $783 = HEAP8[$$0$lcssa$i140>>0]|0; //@line 3152 "load_abc.cpp"
     $784 = ($783<<24>>24)==(0); //@line 3152 "load_abc.cpp"
     if ($784) {
      $$07$i = $782;$$3$i146 = $$0$lcssa$i140;
      break;
     } else {
      $$135$i143 = $$0$lcssa$i140;$787 = $783;
     }
     while(1) {
      $786 = $787 << 24 >> 24; //@line 3152 "load_abc.cpp"
      $788 = (_isspace($786)|0); //@line 3152 "load_abc.cpp"
      $789 = ($788|0)==(0); //@line 3152 "load_abc.cpp"
      if (!($789)) {
       $$135$i143$lcssa = $$135$i143;
       break;
      }
      $790 = ((($$135$i143)) + 1|0); //@line 3152 "load_abc.cpp"
      $791 = HEAP8[$790>>0]|0; //@line 3152 "load_abc.cpp"
      $792 = ($791<<24>>24)==(0); //@line 3152 "load_abc.cpp"
      if ($792) {
       $$07$i = $782;$$3$i146 = $790;
       break L430;
      } else {
       $$135$i143 = $790;$787 = $791;
      }
     }
     $$pre$i142 = HEAP8[$$135$i143$lcssa>>0]|0; //@line 3153 "load_abc.cpp"
     $785 = ($$pre$i142<<24>>24)==(0); //@line 3153 "load_abc.cpp"
     if ($785) {
      $$07$i = $782;$$3$i146 = $$135$i143$lcssa;
      break;
     } else {
      $$231$i144 = $$135$i143$lcssa;$794 = $$pre$i142;
     }
     while(1) {
      $793 = $794 << 24 >> 24; //@line 3153 "load_abc.cpp"
      $795 = (_isspace($793)|0); //@line 3153 "load_abc.cpp"
      $796 = ($795|0)==(0); //@line 3153 "load_abc.cpp"
      if ($796) {
       $$07$i = $782;$$3$i146 = $$231$i144;
       break L430;
      }
      $797 = ((($$231$i144)) + 1|0); //@line 3153 "load_abc.cpp"
      $798 = HEAP8[$797>>0]|0; //@line 3153 "load_abc.cpp"
      $799 = ($798<<24>>24)==(0); //@line 3153 "load_abc.cpp"
      if ($799) {
       $$07$i = $782;$$3$i146 = $797;
       break;
      } else {
       $$231$i144 = $797;$794 = $798;
      }
     }
    }
   } while(0);
   $800 = ($$07$i|0)!=(0|0); //@line 3155 "load_abc.cpp"
   if ($800) {
    $801 = ((($$07$i)) + 36|0); //@line 3155 "load_abc.cpp"
    $802 = HEAP8[$801>>0]|0; //@line 3155 "load_abc.cpp"
    $803 = $802&255; //@line 3155 "load_abc.cpp"
    $873 = $803;
   } else {
    $873 = 1;
   }
   $804 = (_strncmp($$3$i146,526932,11)|0); //@line 3157 "load_abc.cpp"
   $805 = ($804|0)==(0); //@line 3157 "load_abc.cpp"
   L444: do {
    if ($805) {
     $806 = ((($$3$i146)) + 11|0); //@line 3157 "load_abc.cpp"
     $807 = HEAP8[$806>>0]|0; //@line 3157 "load_abc.cpp"
     $808 = $807 << 24 >> 24; //@line 3157 "load_abc.cpp"
     $isdigittmp$i147 = (($808) + -48)|0; //@line 3157 "load_abc.cpp"
     $isdigit$i148 = ($isdigittmp$i147>>>0)<(10); //@line 3157 "load_abc.cpp"
     if ($isdigit$i148) {
      $811 = $808;$h$04$i$i149 = 0;$i$03$i$i150 = 0;
     } else {
      $$7$i = $$3$i146;$875 = 0;
      break;
     }
     while(1) {
      $809 = ($h$04$i$i149*10)|0; //@line 1626 "load_abc.cpp"
      $810 = (($811) + -48)|0; //@line 1626 "load_abc.cpp"
      $812 = (($810) + ($809))|0; //@line 1626 "load_abc.cpp"
      $813 = (($i$03$i$i150) + 1)|0; //@line 1627 "load_abc.cpp"
      $$sum17$i = (($i$03$i$i150) + 12)|0; //@line 1625 "load_abc.cpp"
      $814 = (($$3$i146) + ($$sum17$i)|0); //@line 1625 "load_abc.cpp"
      $815 = HEAP8[$814>>0]|0; //@line 1625 "load_abc.cpp"
      $816 = $815 << 24 >> 24; //@line 1625 "load_abc.cpp"
      $isdigittmp$i$i151 = (($816) + -48)|0; //@line 1625 "load_abc.cpp"
      $isdigit$i$i152 = ($isdigittmp$i$i151>>>0)<(10); //@line 1625 "load_abc.cpp"
      if ($isdigit$i$i152) {
       $811 = $816;$h$04$i$i149 = $812;$i$03$i$i150 = $813;
      } else {
       $$lcssa526 = $812;$$lcssa527 = $813;$$lcssa528 = $814;$$lcssa529 = $815;
       break;
      }
     }
     $817 = ($$lcssa527|0)==(0); //@line 1629 "load_abc.cpp"
     $$h$0$i$i154 = $817 ? 1 : $$lcssa526; //@line 1630 "load_abc.cpp"
     $818 = ($$lcssa529<<24>>24)==(0); //@line 3160 "load_abc.cpp"
     L449: do {
      if ($818) {
       $$4$lcssa$i = $$lcssa528;
      } else {
       $$426$i = $$lcssa528;$820 = $$lcssa529;
       while(1) {
        $819 = $820 << 24 >> 24; //@line 3160 "load_abc.cpp"
        $821 = (_isspace($819)|0); //@line 3160 "load_abc.cpp"
        $822 = ($821|0)==(0); //@line 3160 "load_abc.cpp"
        if ($822) {
         $$4$lcssa$i = $$426$i;
         break L449;
        }
        $823 = ((($$426$i)) + 1|0); //@line 3160 "load_abc.cpp"
        $824 = HEAP8[$823>>0]|0; //@line 3160 "load_abc.cpp"
        $825 = ($824<<24>>24)==(0); //@line 3160 "load_abc.cpp"
        if ($825) {
         $$4$lcssa$i = $823;
         break;
        } else {
         $$426$i = $823;$820 = $824;
        }
       }
      }
     } while(0);
     $826 = (_strncmp($$4$lcssa$i,526949,5)|0); //@line 3161 "load_abc.cpp"
     $827 = ($826|0)==(0); //@line 3161 "load_abc.cpp"
     if (!($827)) {
      $$7$i = $$4$lcssa$i;$875 = $$h$0$i$i154;
      break;
     }
     $828 = ((($$4$lcssa$i)) + 5|0); //@line 3161 "load_abc.cpp"
     $829 = HEAP8[$828>>0]|0; //@line 3161 "load_abc.cpp"
     $830 = $829 << 24 >> 24; //@line 3161 "load_abc.cpp"
     $isdigittmp8$i = (($830) + -48)|0; //@line 3161 "load_abc.cpp"
     $isdigit9$i = ($isdigittmp8$i>>>0)<(10); //@line 3161 "load_abc.cpp"
     if (!($isdigit9$i)) {
      $$7$i = $$4$lcssa$i;$875 = $$h$0$i$i154;
      break;
     }
     $831 = HEAP8[$$4$lcssa$i>>0]|0; //@line 3162 "load_abc.cpp"
     $832 = ($831<<24>>24)==(0); //@line 3162 "load_abc.cpp"
     if ($832) {
      $$7$i = $$4$lcssa$i;$875 = $$h$0$i$i154;
      break;
     } else {
      $$521$i = $$4$lcssa$i;$835 = $831;
     }
     while(1) {
      $834 = $835 << 24 >> 24; //@line 3162 "load_abc.cpp"
      $836 = (_isspace($834)|0); //@line 3162 "load_abc.cpp"
      $837 = ($836|0)==(0); //@line 3162 "load_abc.cpp"
      if (!($837)) {
       $$521$i$lcssa = $$521$i;
       break;
      }
      $838 = ((($$521$i)) + 1|0); //@line 3162 "load_abc.cpp"
      $839 = HEAP8[$838>>0]|0; //@line 3162 "load_abc.cpp"
      $840 = ($839<<24>>24)==(0); //@line 3162 "load_abc.cpp"
      if ($840) {
       $$7$i = $838;$875 = $$h$0$i$i154;
       break L444;
      } else {
       $$521$i = $838;$835 = $839;
      }
     }
     $$pre58$i = HEAP8[$$521$i$lcssa>>0]|0; //@line 3163 "load_abc.cpp"
     $833 = ($$pre58$i<<24>>24)==(0); //@line 3163 "load_abc.cpp"
     if ($833) {
      $$7$i = $$521$i$lcssa;$875 = $$h$0$i$i154;
      break;
     } else {
      $$619$i = $$521$i$lcssa;$842 = $$pre58$i;
     }
     while(1) {
      $841 = $842 << 24 >> 24; //@line 3163 "load_abc.cpp"
      $843 = (_isspace($841)|0); //@line 3163 "load_abc.cpp"
      $844 = ($843|0)==(0); //@line 3163 "load_abc.cpp"
      if ($844) {
       $$7$i = $$619$i;$875 = $$h$0$i$i154;
       break L444;
      }
      $845 = ((($$619$i)) + 1|0); //@line 3163 "load_abc.cpp"
      $846 = HEAP8[$845>>0]|0; //@line 3163 "load_abc.cpp"
      $847 = ($846<<24>>24)==(0); //@line 3163 "load_abc.cpp"
      if ($847) {
       $$7$i = $845;$875 = $$h$0$i$i154;
       break;
      } else {
       $$619$i = $845;$842 = $846;
      }
     }
    } else {
     $$7$i = $$3$i146;$875 = 0;
    }
   } while(0);
   do {
    if ($800) {
     $848 = ((($h)) + 8|0); //@line 3139 "load_abc.cpp"
     $t$01$i12$i = HEAPU8[$848>>0]|(HEAPU8[$848+1>>0]<<8)|(HEAPU8[$848+2>>0]<<16)|(HEAPU8[$848+3>>0]<<24); //@line 3139 "load_abc.cpp"
     $849 = ($t$01$i12$i|0)==(0|0); //@line 3139 "load_abc.cpp"
     if ($849) {
      break;
     }
     $850 = ((($$07$i)) + 36|0); //@line 3140 "load_abc.cpp"
     $851 = HEAP8[$850>>0]|0; //@line 3140 "load_abc.cpp"
     $t$02$i14$i = $t$01$i12$i;
     while(1) {
      $852 = ((($t$02$i14$i)) + 36|0); //@line 3140 "load_abc.cpp"
      $853 = HEAP8[$852>>0]|0; //@line 3140 "load_abc.cpp"
      $854 = ($853<<24>>24)==($851<<24>>24); //@line 3140 "load_abc.cpp"
      if ($854) {
       $855 = ((($t$02$i14$i)) + 39|0); //@line 3140 "load_abc.cpp"
       HEAP8[$855>>0] = 0; //@line 3140 "load_abc.cpp"
      }
      $t$0$i15$i = HEAPU8[$t$02$i14$i>>0]|(HEAPU8[$t$02$i14$i+1>>0]<<8)|(HEAPU8[$t$02$i14$i+2>>0]<<16)|(HEAPU8[$t$02$i14$i+3>>0]<<24); //@line 3139 "load_abc.cpp"
      $856 = ($t$0$i15$i|0)==(0|0); //@line 3139 "load_abc.cpp"
      if ($856) {
       break;
      } else {
       $t$02$i14$i = $t$0$i15$i;
      }
     }
    }
   } while(0);
   $857 = (_strncmp($$7$i,526944,4)|0); //@line 3167 "load_abc.cpp"
   $858 = ($857|0)==(0); //@line 3167 "load_abc.cpp"
   L473: do {
    if ($858) {
     $859 = ((($$7$i)) + 4|0); //@line 3167 "load_abc.cpp"
     $860 = HEAP8[$859>>0]|0; //@line 3167 "load_abc.cpp"
     switch ($860<<24>>24) {
     case 37: case 0:  {
      if (!($800)) {
       break L473;
      }
      break;
     }
     default: {
      $861 = $860 << 24 >> 24; //@line 3167 "load_abc.cpp"
      $862 = (_isspace($861)|0); //@line 3167 "load_abc.cpp"
      $863 = ($862|0)!=(0); //@line 3167 "load_abc.cpp"
      $or$cond$i160 = $800 & $863; //@line 3167 "load_abc.cpp"
      if (!($or$cond$i160)) {
       break L473;
      }
     }
     }
     $864 = ((($h)) + 8|0); //@line 3139 "load_abc.cpp"
     $t$01$i$i = HEAPU8[$864>>0]|(HEAPU8[$864+1>>0]<<8)|(HEAPU8[$864+2>>0]<<16)|(HEAPU8[$864+3>>0]<<24); //@line 3139 "load_abc.cpp"
     $865 = ($t$01$i$i|0)==(0|0); //@line 3139 "load_abc.cpp"
     if ($865) {
      break;
     }
     $866 = ((($$07$i)) + 36|0); //@line 3140 "load_abc.cpp"
     $867 = HEAP8[$866>>0]|0; //@line 3140 "load_abc.cpp"
     $t$02$i$i = $t$01$i$i;
     while(1) {
      $868 = ((($t$02$i$i)) + 36|0); //@line 3140 "load_abc.cpp"
      $869 = HEAP8[$868>>0]|0; //@line 3140 "load_abc.cpp"
      $870 = ($869<<24>>24)==($867<<24>>24); //@line 3140 "load_abc.cpp"
      if ($870) {
       $871 = ((($t$02$i$i)) + 39|0); //@line 3140 "load_abc.cpp"
       HEAP8[$871>>0] = 1; //@line 3140 "load_abc.cpp"
      }
      $t$0$i$i = HEAPU8[$t$02$i$i>>0]|(HEAPU8[$t$02$i$i+1>>0]<<8)|(HEAPU8[$t$02$i$i+2>>0]<<16)|(HEAPU8[$t$02$i$i+3>>0]<<24); //@line 3139 "load_abc.cpp"
      $872 = ($t$0$i$i|0)==(0|0); //@line 3139 "load_abc.cpp"
      if ($872) {
       break;
      } else {
       $t$02$i$i = $t$0$i$i;
      }
     }
    }
   } while(0);
   $tno$off$i$i162 = (($873) + -1)|0; //@line 3110 "load_abc.cpp"
   $874 = ($tno$off$i$i162>>>0)<(32); //@line 3110 "load_abc.cpp"
   $gm$off$i$i = (($875) + -1)|0; //@line 3110 "load_abc.cpp"
   $876 = ($gm$off$i$i>>>0)<(128); //@line 3110 "load_abc.cpp"
   $877 = $874 & $876; //@line 3110 "load_abc.cpp"
   if (!($877)) {
    break;
   }
   $878 = ((($h)) + 8|0); //@line 3111 "load_abc.cpp"
   $tp$06$i$i163 = HEAPU8[$878>>0]|(HEAPU8[$878+1>>0]<<8)|(HEAPU8[$878+2>>0]<<16)|(HEAPU8[$878+3>>0]<<24); //@line 3111 "load_abc.cpp"
   $879 = ($tp$06$i$i163|0)==(0|0); //@line 3111 "load_abc.cpp"
   if ($879) {
    break;
   }
   $880 = $875&255; //@line 3113 "load_abc.cpp"
   $tp$07$i$i164 = $tp$06$i$i163;
   while(1) {
    $881 = ((($tp$07$i$i164)) + 36|0); //@line 3112 "load_abc.cpp"
    $882 = HEAP8[$881>>0]|0; //@line 3112 "load_abc.cpp"
    $883 = $882&255; //@line 3112 "load_abc.cpp"
    $884 = ($883|0)==($873|0); //@line 3112 "load_abc.cpp"
    do {
     if ($884) {
      $885 = ((($tp$07$i$i164)) + 37|0); //@line 3112 "load_abc.cpp"
      $886 = HEAP8[$885>>0]|0; //@line 3112 "load_abc.cpp"
      $887 = (($886) + -1)<<24>>24; //@line 3112 "load_abc.cpp"
      $888 = ($887&255)>(9); //@line 3112 "load_abc.cpp"
      if (!($888)) {
       break;
      }
      $889 = ((($tp$07$i$i164)) + 42|0); //@line 3113 "load_abc.cpp"
      HEAP8[$889>>0] = $880; //@line 3113 "load_abc.cpp"
     }
    } while(0);
    $tp$0$i$i165 = HEAPU8[$tp$07$i$i164>>0]|(HEAPU8[$tp$07$i$i164+1>>0]<<8)|(HEAPU8[$tp$07$i$i164+2>>0]<<16)|(HEAPU8[$tp$07$i$i164+3>>0]<<24); //@line 3111 "load_abc.cpp"
    $890 = ($tp$0$i$i165|0)==(0|0); //@line 3111 "load_abc.cpp"
    if ($890) {
     break;
    } else {
     $tp$07$i$i164 = $tp$0$i$i165;
    }
   }
  }
 } while(0);
 $891 = (_strncmp($p,526955,6)|0); //@line 3719 "load_abc.cpp"
 $892 = ($891|0)==(0); //@line 3719 "load_abc.cpp"
 if (!($892)) {
  STACKTOP = sp;return; //@line 3721 "load_abc.cpp"
 }
 $893 = ((($p)) + 6|0); //@line 3719 "load_abc.cpp"
 $894 = HEAP8[$893>>0]|0; //@line 3719 "load_abc.cpp"
 $895 = ($894<<24>>24)==(0); //@line 3719 "load_abc.cpp"
 $896 = ($894<<24>>24)==($delim<<24>>24); //@line 3719 "load_abc.cpp"
 $or$cond7 = $895 | $896; //@line 3719 "load_abc.cpp"
 do {
  if (!($or$cond7)) {
   $897 = $894 << 24 >> 24; //@line 3719 "load_abc.cpp"
   $898 = (_isspace($897)|0); //@line 3719 "load_abc.cpp"
   $899 = ($898|0)==(0); //@line 3719 "load_abc.cpp"
   if (!($899)) {
    break;
   }
   STACKTOP = sp;return; //@line 3721 "load_abc.cpp"
  }
 } while(0);
 $900 = ((($h)) + 412|0); //@line 3720 "load_abc.cpp"
 $901 = HEAPU8[$900>>0]|(HEAPU8[$900+1>>0]<<8)|(HEAPU8[$900+2>>0]<<16)|(HEAPU8[$900+3>>0]<<24); //@line 3720 "load_abc.cpp"
 $902 = ($901|0)==(0|0); //@line 915 "load_abc.cpp"
 if ($902) {
  $903 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,1282466,0)|0); //@line 916 "load_abc.cpp"
  $904 = ((($h)) + 376|0); //@line 917 "load_abc.cpp"
  $905 = HEAPU8[$904>>0]|(HEAPU8[$904+1>>0]<<8)|(HEAPU8[$904+2>>0]<<16)|(HEAPU8[$904+3>>0]<<24); //@line 917 "load_abc.cpp"
  $906 = ((($903)) + 20|0); //@line 917 "load_abc.cpp"
  HEAP8[$906>>0]=$905&255;HEAP8[$906+1>>0]=($905>>8)&255;HEAP8[$906+2>>0]=($905>>16)&255;HEAP8[$906+3>>0]=$905>>24; //@line 917 "load_abc.cpp"
  $$0$i166 = $903;
 } else {
  $$0$i166 = $901;
 }
 HEAP8[$900>>0]=$$0$i166&255;HEAP8[$900+1>>0]=($$0$i166>>8)&255;HEAP8[$900+2>>0]=($$0$i166>>16)&255;HEAP8[$900+3>>0]=$$0$i166>>24; //@line 3720 "load_abc.cpp"
 $907 = HEAP8[$893>>0]|0; //@line 3305 "load_abc.cpp"
 $908 = ($907<<24>>24)==(0); //@line 3305 "load_abc.cpp"
 L505: do {
  if ($908) {
   $$0$lcssa$i170 = $893;
  } else {
   $$01$i168 = $893;$910 = $907;
   while(1) {
    $909 = $910 << 24 >> 24; //@line 3305 "load_abc.cpp"
    $911 = (_isspace($909)|0); //@line 3305 "load_abc.cpp"
    $912 = ($911|0)==(0); //@line 3305 "load_abc.cpp"
    if ($912) {
     $$0$lcssa$i170 = $$01$i168;
     break L505;
    }
    $913 = ((($$01$i168)) + 1|0); //@line 3305 "load_abc.cpp"
    $914 = HEAP8[$913>>0]|0; //@line 3305 "load_abc.cpp"
    $915 = ($914<<24>>24)==(0); //@line 3305 "load_abc.cpp"
    if ($915) {
     $$0$lcssa$i170 = $913;
     break;
    } else {
     $$01$i168 = $913;$910 = $914;
    }
   }
  }
 } while(0);
 $916 = (_strncmp($$0$lcssa$i170,526516,3)|0); //@line 3306 "load_abc.cpp"
 $917 = ((($$0$i166)) + 43|0); //@line 3307 "load_abc.cpp"
 $not$$i = ($916|0)!=(0); //@line 3306 "load_abc.cpp"
 $$$i171 = $not$$i&1; //@line 3306 "load_abc.cpp"
 HEAP8[$917>>0] = $$$i171; //@line 3307 "load_abc.cpp"
 STACKTOP = sp;return; //@line 3721 "load_abc.cpp"
}
function __ZL17abc_add_dronenoteP10_ABCHANDLEP9_ABCTRACKjii($tp,$tracktime,$nnum,$vol) {
 $tp = $tp|0;
 $tracktime = $tracktime|0;
 $nnum = $nnum|0;
 $vol = $vol|0;
 var $$$02$i = 0, $$$i = 0, $$0$i = 0, $$02$i = 0, $$02$p$i = 0, $$2$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $9 = 0, $d = 0, $k$0$lcssa = 0, $k$0$lcssa$op = 0, $k$1$lcssa = 0, $oct$0$lcssa = 0, $oct$1$lcssa = 0, $or$cond$i = 0, $scevgep$i = 0, $smax16 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $d = sp;
 $0 = (($nnum) + 1)|0; //@line 1435 "load_abc.cpp"
 $1 = ($nnum|0)>(11); //@line 1436 "load_abc.cpp"
 if ($1) {
  $2 = (10 - ($nnum))|0; //@line 1436 "load_abc.cpp"
  $3 = ($2|0)>(-13);
  $smax16 = $3 ? $2 : -13;
  $4 = (($smax16) + ($nnum))|0; //@line 1436 "load_abc.cpp"
  $5 = (($4) + 1)|0; //@line 1436 "load_abc.cpp"
  $6 = (($5>>>0) / 12)&-1; //@line 1436 "load_abc.cpp"
  $7 = (($nnum) + -11)|0; //@line 1436 "load_abc.cpp"
  $8 = Math_imul($6, -12)|0; //@line 1436 "load_abc.cpp"
  $9 = (($7) + ($8))|0; //@line 1436 "load_abc.cpp"
  $k$0$lcssa = $9;$oct$0$lcssa = $6;
 } else {
  $k$0$lcssa = $0;$oct$0$lcssa = -1;
 }
 $10 = ($k$0$lcssa|0)<(0); //@line 1440 "load_abc.cpp"
 if ($10) {
  $11 = ($k$0$lcssa|0)>(-12);
  $k$0$lcssa$op = (($k$0$lcssa) + 11)|0; //@line 1440 "load_abc.cpp"
  $12 = $11 ? $k$0$lcssa$op : -1; //@line 1440 "load_abc.cpp"
  $13 = (($12) - ($k$0$lcssa))|0; //@line 1440 "load_abc.cpp"
  $14 = (($13>>>0) / 12)&-1; //@line 1440 "load_abc.cpp"
  $15 = ($14*12)|0; //@line 1440 "load_abc.cpp"
  $16 = (($k$0$lcssa) + ($15))|0; //@line 1440 "load_abc.cpp"
  $17 = (($oct$0$lcssa) + -1)|0; //@line 1440 "load_abc.cpp"
  $18 = (($16) + 12)|0; //@line 1440 "load_abc.cpp"
  $19 = (($17) - ($14))|0; //@line 1440 "load_abc.cpp"
  $k$1$lcssa = $18;$oct$1$lcssa = $19;
 } else {
  $k$1$lcssa = $k$0$lcssa;$oct$1$lcssa = $oct$0$lcssa;
 }
 $20 = ($oct$1$lcssa|0)<(0); //@line 1444 "load_abc.cpp"
 $21 = (($k$1$lcssa) + 23)|0; //@line 1445 "load_abc.cpp"
 $22 = $21&255; //@line 1445 "load_abc.cpp"
 HEAP8[$d>>0] = $22; //@line 1445 "load_abc.cpp"
 $23 = $oct$1$lcssa&255; //@line 1446 "load_abc.cpp"
 $24 = $20 ? 0 : $23; //@line 1446 "load_abc.cpp"
 $25 = ((($d)) + 1|0); //@line 1446 "load_abc.cpp"
 HEAP8[$25>>0] = $24; //@line 1446 "load_abc.cpp"
 $26 = ((($tp)) + 42|0); //@line 1447 "load_abc.cpp"
 $27 = HEAP8[$26>>0]|0; //@line 1447 "load_abc.cpp"
 $28 = $27&255; //@line 1447 "load_abc.cpp"
 $29 = (_pat_gmtosmp($28)|0); //@line 1447 "load_abc.cpp"
 $30 = $29&255; //@line 1447 "load_abc.cpp"
 $31 = ((($d)) + 2|0); //@line 1447 "load_abc.cpp"
 HEAP8[$31>>0] = $30; //@line 1447 "load_abc.cpp"
 $32 = ((($tp)) + 39|0); //@line 1186 "load_abc.cpp"
 $33 = HEAP8[$32>>0]|0; //@line 1186 "load_abc.cpp"
 $34 = ($33<<24>>24)==(0); //@line 1186 "load_abc.cpp"
 if ($34) {
  $35 = ((($tp)) + 32|0); //@line 1187 "load_abc.cpp"
  $36 = HEAPU8[$35>>0]|(HEAPU8[$35+1>>0]<<8)|(HEAPU8[$35+2>>0]<<16)|(HEAPU8[$35+3>>0]<<24); //@line 1187 "load_abc.cpp"
  $37 = ($36|0)==(0); //@line 1187 "load_abc.cpp"
  if ($37) {
   $$0$i = $vol;
  } else {
   $38 = ((($tp)) + 28|0); //@line 1188 "load_abc.cpp"
   $39 = HEAPU8[$38>>0]|(HEAPU8[$38+1>>0]<<8)|(HEAPU8[$38+2>>0]<<16)|(HEAPU8[$38+3>>0]<<24); //@line 1188 "load_abc.cpp"
   $40 = ($39>>>0)>($tracktime>>>0); //@line 1188 "load_abc.cpp"
   if ($40) {
    $$0$i = $vol;
   } else {
    $41 = (($tracktime) - ($39))|0; //@line 1189 "load_abc.cpp"
    $42 = (($41>>>0) / 192)&-1; //@line 334 "load_abc.cpp"
    $43 = ($42*15)|0; //@line 1190 "load_abc.cpp"
    $44 = (($43>>>0) / 384)&-1; //@line 1190 "load_abc.cpp"
    $45 = ($36|0)>(-2); //@line 1191 "load_abc.cpp"
    $46 = ($43>>>0)>(6143); //@line 1191 "load_abc.cpp"
    $or$cond$i = $45 & $46; //@line 1191 "load_abc.cpp"
    $$$i = $or$cond$i ? 15 : $44; //@line 1191 "load_abc.cpp"
    $47 = ($36|0)>(0); //@line 1192 "load_abc.cpp"
    $48 = (0 - ($$$i))|0; //@line 1192 "load_abc.cpp"
    $$02$p$i = $47 ? $$$i : $48; //@line 1192 "load_abc.cpp"
    $$02$i = (($$02$p$i) + ($vol))|0; //@line 1192 "load_abc.cpp"
    $49 = ($$02$i|0)<(2); //@line 1194 "load_abc.cpp"
    $$$02$i = $49 ? 2 : $$02$i; //@line 1194 "load_abc.cpp"
    $50 = ($$$02$i|0)>(127); //@line 1195 "load_abc.cpp"
    $$2$i = $50 ? 127 : $$$02$i; //@line 1195 "load_abc.cpp"
    $$0$i = $$2$i;
   }
  }
 } else {
  $$0$i = 0;
 }
 $51 = $$0$i&255; //@line 1448 "load_abc.cpp"
 $52 = ((($d)) + 3|0); //@line 1448 "load_abc.cpp"
 HEAP8[$52>>0] = $51; //@line 1448 "load_abc.cpp"
 $53 = ((($d)) + 4|0); //@line 1449 "load_abc.cpp"
 HEAP8[$53>>0] = 0; //@line 1449 "load_abc.cpp"
 $54 = ((($d)) + 5|0); //@line 1450 "load_abc.cpp"
 HEAP8[$54>>0] = 0; //@line 1450 "load_abc.cpp"
 $55 = ((($tp)) + 8|0); //@line 1451 "load_abc.cpp"
 $56 = HEAPU8[$55>>0]|(HEAPU8[$55+1>>0]<<8)|(HEAPU8[$55+2>>0]<<16)|(HEAPU8[$55+3>>0]<<24); //@line 1451 "load_abc.cpp"
 $57 = ($56|0)==(0|0); //@line 1451 "load_abc.cpp"
 if (!($57)) {
  $58 = ((($56)) + 4|0); //@line 1452 "load_abc.cpp"
  $59 = HEAPU8[$58>>0]|(HEAPU8[$58+1>>0]<<8)|(HEAPU8[$58+2>>0]<<16)|(HEAPU8[$58+3>>0]<<24); //@line 1452 "load_abc.cpp"
  $60 = ($59|0)==($tracktime|0); //@line 1452 "load_abc.cpp"
  if ($60) {
   $61 = ((($56)) + 8|0); //@line 1453 "load_abc.cpp"
   $62 = HEAP8[$61>>0]|0; //@line 1453 "load_abc.cpp"
   $63 = ($62<<24>>24)==(0); //@line 1453 "load_abc.cpp"
   if ($63) {
    $64 = ((($56)) + 9|0); //@line 1454 "load_abc.cpp"
    $65 = HEAP8[$64>>0]|0; //@line 1454 "load_abc.cpp"
    $66 = ($65<<24>>24)==(0); //@line 1454 "load_abc.cpp"
    if ($66) {
     HEAP8[$61>>0] = $22; //@line 1456 "load_abc.cpp"
     $67 = HEAPU8[$55>>0]|(HEAPU8[$55+1>>0]<<8)|(HEAPU8[$55+2>>0]<<16)|(HEAPU8[$55+3>>0]<<24); //@line 1456 "load_abc.cpp"
     $68 = ((($67)) + 9|0); //@line 1456 "load_abc.cpp"
     HEAP8[$68>>0] = $24; //@line 1456 "load_abc.cpp"
     $69 = HEAPU8[$55>>0]|(HEAPU8[$55+1>>0]<<8)|(HEAPU8[$55+2>>0]<<16)|(HEAPU8[$55+3>>0]<<24); //@line 1456 "load_abc.cpp"
     $70 = ((($69)) + 10|0); //@line 1456 "load_abc.cpp"
     HEAP8[$70>>0] = $30; //@line 1456 "load_abc.cpp"
     $71 = HEAPU8[$55>>0]|(HEAPU8[$55+1>>0]<<8)|(HEAPU8[$55+2>>0]<<16)|(HEAPU8[$55+3>>0]<<24); //@line 1456 "load_abc.cpp"
     $72 = ((($71)) + 11|0); //@line 1456 "load_abc.cpp"
     HEAP8[$72>>0] = $51; //@line 1456 "load_abc.cpp"
     $73 = HEAPU8[$55>>0]|(HEAPU8[$55+1>>0]<<8)|(HEAPU8[$55+2>>0]<<16)|(HEAPU8[$55+3>>0]<<24); //@line 1456 "load_abc.cpp"
     $74 = ((($73)) + 12|0); //@line 1456 "load_abc.cpp"
     HEAP8[$74>>0] = 0; //@line 1456 "load_abc.cpp"
     $75 = HEAPU8[$55>>0]|(HEAPU8[$55+1>>0]<<8)|(HEAPU8[$55+2>>0]<<16)|(HEAPU8[$55+3>>0]<<24); //@line 1456 "load_abc.cpp"
     $76 = ((($75)) + 13|0); //@line 1456 "load_abc.cpp"
     HEAP8[$76>>0] = 0; //@line 1456 "load_abc.cpp"
     STACKTOP = sp;return; //@line 1462 "load_abc.cpp"
    }
   }
  }
 }
 $77 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
 HEAP8[$77>>0]=0&255;HEAP8[$77+1>>0]=(0>>8)&255;HEAP8[$77+2>>0]=(0>>16)&255;HEAP8[$77+3>>0]=0>>24; //@line 540 "load_abc.cpp"
 $78 = ((($77)) + 4|0); //@line 541 "load_abc.cpp"
 HEAP8[$78>>0]=$tracktime&255;HEAP8[$78+1>>0]=($tracktime>>8)&255;HEAP8[$78+2>>0]=($tracktime>>16)&255;HEAP8[$78+3>>0]=$tracktime>>24; //@line 541 "load_abc.cpp"
 $scevgep$i = ((($77)) + 8|0); //@line 1459 "load_abc.cpp"
 ;HEAP8[$scevgep$i>>0]=HEAP8[$d>>0]|0;HEAP8[$scevgep$i+1>>0]=HEAP8[$d+1>>0]|0;HEAP8[$scevgep$i+2>>0]=HEAP8[$d+2>>0]|0;HEAP8[$scevgep$i+3>>0]=HEAP8[$d+3>>0]|0;HEAP8[$scevgep$i+4>>0]=HEAP8[$d+4>>0]|0;HEAP8[$scevgep$i+5>>0]=HEAP8[$d+5>>0]|0; //@line 543 "load_abc.cpp"
 $79 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
 $80 = ((($77)) + 14|0); //@line 544 "load_abc.cpp"
 HEAP8[$80>>0] = $79; //@line 544 "load_abc.cpp"
 $81 = ((($77)) + 15|0); //@line 545 "load_abc.cpp"
 HEAP8[$81>>0] = 0; //@line 545 "load_abc.cpp"
 __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($tp,$77); //@line 1460 "load_abc.cpp"
 STACKTOP = sp;return; //@line 1462 "load_abc.cpp"
}
function __ZL14abc_add_noteonP10_ABCHANDLEiPKcjPci17ABCEVENT_X_EFFECTi($h,$ch,$p,$tracktime,$barkey,$vol,$fx,$fxop) {
 $h = $h|0;
 $ch = $ch|0;
 $p = $p|0;
 $tracktime = $tracktime|0;
 $barkey = $barkey|0;
 $vol = $vol|0;
 $fx = $fx|0;
 $fxop = $fxop|0;
 var $$ = 0, $$$02$i = 0, $$$i = 0, $$0 = 0, $$0$i = 0, $$0$i4 = 0, $$02 = 0, $$02$i = 0, $$02$p$i = 0, $$2$i = 0, $$k$0 = 0, $$lcssa = 0, $$op = 0, $$phi$trans$insert = 0, $$pr = 0, $$pre = 0, $$pre58 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0;
 var $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0;
 var $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0;
 var $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0;
 var $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0;
 var $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0;
 var $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0;
 var $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0;
 var $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0;
 var $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0;
 var $281 = 0, $282 = 0, $283 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $d = 0, $e$021 = 0, $e$021$lcssa = 0, $i$0 = 0, $i$1$lcssa = 0, $i$122 = 0, $isdigit = 0, $isdigittmp = 0, $ispos = 0, $ispos$i = 0, $j$0 = 0, $k$0$lcssa = 0, $k$041 = 0, $k$240 = 0, $k$240$lcssa = 0, $k$3 = 0, $k$4 = 0, $k$5 = 0, $k$6 = 0;
 var $k$7$lcssa = 0, $k$7$lcssa$op = 0, $k$8$lcssa = 0, $memchr = 0, $n$0 = 0, $neg = 0, $neg$i = 0, $oct$0$lcssa = 0, $oct$1$lcssa = 0, $oct$2$lcssa = 0, $oct$2$ph = 0, $oct$223 = 0, $oct$3 = 0, $oct$3$v = 0, $oct$5 = 0, $or$cond$i = 0, $scevgep$i = 0, $tp$0$i = 0, $tp$05$i = 0, $tp$06$i = 0;
 var $tp$06$i$lcssa = 0, $tp$1$i = 0, $tp$14$i = 0, $tp$14$i$lcssa = 0, $txt$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 288|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $txt$i = sp + 24|0;
 $d = sp + 12|0;
 $0 = ((($h)) + 412|0); //@line 1273 "load_abc.cpp"
 $1 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 1273 "load_abc.cpp"
 L1: do {
  switch ($ch|0) {
  case 94:  {
   $2 = HEAP8[$p>>0]|0; //@line 1276 "load_abc.cpp"
   $3 = ($2<<24>>24)==(94); //@line 1276 "load_abc.cpp"
   if ($3) {
    $4 = ((($p)) + 1|0); //@line 1277 "load_abc.cpp"
    $5 = HEAP8[$4>>0]|0; //@line 1277 "load_abc.cpp"
    $6 = $5 << 24 >> 24; //@line 1277 "load_abc.cpp"
    $$02 = 120;$i$0 = 2;$n$0 = $6;
    break L1;
   } else {
    $7 = $2 << 24 >> 24; //@line 1276 "load_abc.cpp"
    $$02 = 94;$i$0 = 1;$n$0 = $7;
    break L1;
   }
   break;
  }
  case 95:  {
   $8 = HEAP8[$p>>0]|0; //@line 1287 "load_abc.cpp"
   $9 = ($8<<24>>24)==(95); //@line 1287 "load_abc.cpp"
   if ($9) {
    $10 = ((($p)) + 1|0); //@line 1288 "load_abc.cpp"
    $11 = HEAP8[$10>>0]|0; //@line 1288 "load_abc.cpp"
    $12 = $11 << 24 >> 24; //@line 1288 "load_abc.cpp"
    $$02 = 98;$i$0 = 2;$n$0 = $12;
    break L1;
   } else {
    $13 = $8 << 24 >> 24; //@line 1287 "load_abc.cpp"
    $$02 = 95;$i$0 = 1;$n$0 = $13;
    break L1;
   }
   break;
  }
  case 61:  {
   $14 = HEAP8[$p>>0]|0; //@line 1298 "load_abc.cpp"
   $15 = $14 << 24 >> 24; //@line 1298 "load_abc.cpp"
   $$02 = 61;$i$0 = 1;$n$0 = $15;
   break;
  }
  default: {
   $$02 = $ch;$i$0 = 0;$n$0 = $ch;
  }
  }
 } while(0);
 $k$041 = 0;
 while(1) {
  $16 = (($barkey) + ($k$041)|0); //@line 1307 "load_abc.cpp"
  $17 = HEAP8[$16>>0]|0; //@line 1307 "load_abc.cpp"
  $18 = $17 << 24 >> 24; //@line 1307 "load_abc.cpp"
  $19 = ($n$0|0)==($18|0); //@line 1307 "load_abc.cpp"
  if ($19) {
   $k$0$lcssa = $k$041;
   break;
  }
  $20 = (($k$041) + 1)|0; //@line 1306 "load_abc.cpp"
  $21 = ($20|0)<(51); //@line 1306 "load_abc.cpp"
  if ($21) {
   $k$041 = $20;
  } else {
   $k$0$lcssa = $20;
   break;
  }
 }
 $22 = ($k$0$lcssa|0)>(24); //@line 1311 "load_abc.cpp"
 $23 = (($k$0$lcssa) + -25)|0; //@line 1312 "load_abc.cpp"
 $$k$0 = $22 ? $23 : $k$0$lcssa; //@line 1311 "load_abc.cpp"
 $24 = ($i$0|0)==(0); //@line 1313 "load_abc.cpp"
 L16: do {
  if ($24) {
   $k$6 = $$k$0;
   label = 23;
  } else {
   $k$240 = 0;
   while(1) {
    $25 = (525098 + ($k$240)|0); //@line 1317 "load_abc.cpp"
    $26 = HEAP8[$25>>0]|0; //@line 1317 "load_abc.cpp"
    $27 = $26 << 24 >> 24; //@line 1317 "load_abc.cpp"
    $28 = ($n$0|0)==($27|0); //@line 1317 "load_abc.cpp"
    if ($28) {
     $k$240$lcssa = $k$240;
     break;
    }
    $29 = (($k$240) + 1)|0; //@line 1316 "load_abc.cpp"
    $30 = ($29|0)<(25); //@line 1316 "load_abc.cpp"
    if ($30) {
     $k$240 = $29;
    } else {
     label = 29;
     break L16;
    }
   }
   switch ($$02|0) {
   case 120:  {
    $31 = (($k$240$lcssa) + 1)|0; //@line 1323 "load_abc.cpp"
    $k$3 = $31;
    label = 17;
    break;
   }
   case 94:  {
    $k$3 = $k$240$lcssa;
    label = 17;
    break;
   }
   case 98:  {
    $33 = (($k$240$lcssa) + -1)|0; //@line 1328 "load_abc.cpp"
    $k$4 = $33;
    label = 19;
    break;
   }
   case 95:  {
    $k$4 = $k$240$lcssa;
    label = 19;
    break;
   }
   default: {
    $k$5 = $k$240$lcssa;
   }
   }
   if ((label|0) == 17) {
    $32 = (($k$3) + 1)|0; //@line 1325 "load_abc.cpp"
    $k$5 = $32;
   }
   else if ((label|0) == 19) {
    $34 = (($k$4) + -1)|0; //@line 1330 "load_abc.cpp"
    $k$5 = $34;
   }
   $35 = ($k$0$lcssa|0)<(25); //@line 1335 "load_abc.cpp"
   if ($35) {
    $36 = (($barkey) + ($k$0$lcssa)|0); //@line 1336 "load_abc.cpp"
    HEAP8[$36>>0] = 32; //@line 1336 "load_abc.cpp"
   }
   $37 = $n$0&255; //@line 1337 "load_abc.cpp"
   $38 = (($barkey) + ($k$5)|0); //@line 1337 "load_abc.cpp"
   HEAP8[$38>>0] = $37; //@line 1337 "load_abc.cpp"
   $k$6 = $k$5;
   label = 23;
  }
 } while(0);
 if ((label|0) == 23) {
  $39 = ($k$6|0)<(25); //@line 1341 "load_abc.cpp"
  if ($39) {
   $40 = ((($1)) + 20|0); //@line 1342 "load_abc.cpp"
   $41 = HEAPU8[$40>>0]|(HEAPU8[$40+1>>0]<<8)|(HEAPU8[$40+2>>0]<<16)|(HEAPU8[$40+3>>0]<<24); //@line 1342 "load_abc.cpp"
   $42 = (($41) + ($k$6))|0; //@line 1342 "load_abc.cpp"
   $43 = ($42|0)>(12); //@line 1343 "load_abc.cpp"
   if ($43) {
    $44 = (($42) + -12)|0; //@line 1343 "load_abc.cpp"
    $45 = (($42) + -13)|0; //@line 1343 "load_abc.cpp"
    $46 = (($45>>>0) / 12)&-1; //@line 1343 "load_abc.cpp"
    $47 = Math_imul($46, -12)|0; //@line 1343 "load_abc.cpp"
    $48 = (($44) + ($47))|0; //@line 1343 "load_abc.cpp"
    $49 = (($46) + 4)|0; //@line 1343 "load_abc.cpp"
    $k$7$lcssa = $48;$oct$0$lcssa = $49;
   } else {
    $k$7$lcssa = $42;$oct$0$lcssa = 3;
   }
   $50 = ($k$7$lcssa|0)<(0); //@line 1347 "load_abc.cpp"
   if ($50) {
    $51 = ($k$7$lcssa|0)>(-12);
    $k$7$lcssa$op = (($k$7$lcssa) + 11)|0; //@line 1347 "load_abc.cpp"
    $52 = $51 ? $k$7$lcssa$op : -1; //@line 1347 "load_abc.cpp"
    $53 = (($52) - ($k$7$lcssa))|0; //@line 1347 "load_abc.cpp"
    $54 = (($53>>>0) / 12)&-1; //@line 1347 "load_abc.cpp"
    $55 = ($54*12)|0; //@line 1347 "load_abc.cpp"
    $56 = (($k$7$lcssa) + ($55))|0; //@line 1347 "load_abc.cpp"
    $57 = (($oct$0$lcssa) + -1)|0; //@line 1347 "load_abc.cpp"
    $58 = (($56) + 12)|0; //@line 1347 "load_abc.cpp"
    $59 = (($57) - ($54))|0; //@line 1347 "load_abc.cpp"
    $k$8$lcssa = $58;$oct$1$lcssa = $59;
   } else {
    $k$8$lcssa = $k$7$lcssa;$oct$1$lcssa = $oct$0$lcssa;
   }
   $60 = (($k$8$lcssa) + 23)|0; //@line 1351 "load_abc.cpp"
   $61 = $60&255; //@line 1351 "load_abc.cpp"
   HEAP8[$d>>0] = $61; //@line 1351 "load_abc.cpp"
   $116 = $61;$oct$2$ph = $oct$1$lcssa;
  } else {
   label = 29;
  }
 }
 if ((label|0) == 29) {
  HEAP8[$d>>0] = 0; //@line 1354 "load_abc.cpp"
  $116 = 0;$oct$2$ph = 3;
 }
 $62 = (($p) + ($i$0)|0); //@line 1355 "load_abc.cpp"
 $63 = HEAP8[$62>>0]|0; //@line 1355 "load_abc.cpp"
 $64 = ($63<<24>>24)==(0); //@line 1355 "load_abc.cpp"
 $$pre58 = ((($1)) + 24|0); //@line 1361 "load_abc.cpp"
 L43: do {
  if ($64) {
   $i$1$lcssa = $i$0;$oct$2$lcssa = $oct$2$ph;
  } else {
   $66 = $63;$i$122 = $i$0;$oct$223 = $oct$2$ph;
   while(1) {
    $65 = $66 << 24 >> 24; //@line 1355 "load_abc.cpp"
    $memchr = (_memchr(527245,$65,3)|0); //@line 1355 "load_abc.cpp"
    $67 = ($memchr|0)==(0|0); //@line 1355 "load_abc.cpp"
    if ($67) {
     $i$1$lcssa = $i$122;$oct$2$lcssa = $oct$223;
     break L43;
    }
    $68 = ($66<<24>>24)==(44); //@line 1356 "load_abc.cpp"
    $oct$3$v = $68 ? -1 : 1; //@line 1356 "load_abc.cpp"
    $oct$3 = (($oct$3$v) + ($oct$223))|0; //@line 1356 "load_abc.cpp"
    $69 = (($i$122) + 1)|0; //@line 1358 "load_abc.cpp"
    HEAP8[$$pre58>>0]=0&255;HEAP8[$$pre58+1>>0]=(0>>8)&255;HEAP8[$$pre58+2>>0]=(0>>16)&255;HEAP8[$$pre58+3>>0]=0>>24; //@line 1359 "load_abc.cpp"
    $70 = (($p) + ($69)|0); //@line 1355 "load_abc.cpp"
    $71 = HEAP8[$70>>0]|0; //@line 1355 "load_abc.cpp"
    $72 = ($71<<24>>24)==(0); //@line 1355 "load_abc.cpp"
    if ($72) {
     $i$1$lcssa = $69;$oct$2$lcssa = $oct$3;
     break;
    } else {
     $66 = $71;$i$122 = $69;$oct$223 = $oct$3;
    }
   }
  }
 } while(0);
 $73 = HEAPU8[$$pre58>>0]|(HEAPU8[$$pre58+1>>0]<<8)|(HEAPU8[$$pre58+2>>0]<<16)|(HEAPU8[$$pre58+3>>0]<<24); //@line 1361 "load_abc.cpp"
 $74 = ($73|0)==(0); //@line 1361 "load_abc.cpp"
 if (!($74)) {
  $75 = ($73*12)|0; //@line 1362 "load_abc.cpp"
  $76 = ((($1)) + 20|0); //@line 1362 "load_abc.cpp"
  $77 = HEAPU8[$76>>0]|(HEAPU8[$76+1>>0]<<8)|(HEAPU8[$76+2>>0]<<16)|(HEAPU8[$76+3>>0]<<24); //@line 1362 "load_abc.cpp"
  $78 = (($77) + ($75))|0; //@line 1362 "load_abc.cpp"
  HEAP8[$76>>0]=$78&255;HEAP8[$76+1>>0]=($78>>8)&255;HEAP8[$76+2>>0]=($78>>16)&255;HEAP8[$76+3>>0]=$78>>24; //@line 1362 "load_abc.cpp"
 }
 $79 = (($73) + ($oct$2$lcssa))|0; //@line 1363 "load_abc.cpp"
 HEAP8[$$pre58>>0]=0&255;HEAP8[$$pre58+1>>0]=(0>>8)&255;HEAP8[$$pre58+2>>0]=(0>>16)&255;HEAP8[$$pre58+3>>0]=0>>24; //@line 1364 "load_abc.cpp"
 $80 = ($79|0)<(0); //@line 1365 "load_abc.cpp"
 $$ = $80 ? 0 : $79; //@line 1365 "load_abc.cpp"
 $81 = ($$|0)>(9); //@line 1366 "load_abc.cpp"
 $oct$5 = $81 ? 9 : $$; //@line 1366 "load_abc.cpp"
 $82 = $oct$5&255; //@line 1367 "load_abc.cpp"
 $83 = ((($d)) + 1|0); //@line 1367 "load_abc.cpp"
 HEAP8[$83>>0] = $82; //@line 1367 "load_abc.cpp"
 $84 = ((($1)) + 42|0); //@line 1368 "load_abc.cpp"
 $85 = HEAP8[$84>>0]|0; //@line 1368 "load_abc.cpp"
 $86 = $85&255; //@line 1368 "load_abc.cpp"
 $87 = (_pat_gmtosmp($86)|0); //@line 1368 "load_abc.cpp"
 $88 = $87&255; //@line 1368 "load_abc.cpp"
 $89 = ((($d)) + 2|0); //@line 1368 "load_abc.cpp"
 HEAP8[$89>>0] = $88; //@line 1368 "load_abc.cpp"
 $90 = ((($1)) + 39|0); //@line 1186 "load_abc.cpp"
 $91 = HEAP8[$90>>0]|0; //@line 1186 "load_abc.cpp"
 $92 = ($91<<24>>24)==(0); //@line 1186 "load_abc.cpp"
 if ($92) {
  $93 = ((($1)) + 32|0); //@line 1187 "load_abc.cpp"
  $94 = HEAPU8[$93>>0]|(HEAPU8[$93+1>>0]<<8)|(HEAPU8[$93+2>>0]<<16)|(HEAPU8[$93+3>>0]<<24); //@line 1187 "load_abc.cpp"
  $95 = ($94|0)==(0); //@line 1187 "load_abc.cpp"
  if ($95) {
   $$0$i = $vol;
  } else {
   $96 = ((($1)) + 28|0); //@line 1188 "load_abc.cpp"
   $97 = HEAPU8[$96>>0]|(HEAPU8[$96+1>>0]<<8)|(HEAPU8[$96+2>>0]<<16)|(HEAPU8[$96+3>>0]<<24); //@line 1188 "load_abc.cpp"
   $98 = ($97>>>0)>($tracktime>>>0); //@line 1188 "load_abc.cpp"
   if ($98) {
    $$0$i = $vol;
   } else {
    $99 = (($tracktime) - ($97))|0; //@line 1189 "load_abc.cpp"
    $100 = (($99>>>0) / 192)&-1; //@line 334 "load_abc.cpp"
    $101 = ($100*15)|0; //@line 1190 "load_abc.cpp"
    $102 = (($101>>>0) / 384)&-1; //@line 1190 "load_abc.cpp"
    $103 = ($94|0)>(-2); //@line 1191 "load_abc.cpp"
    $104 = ($101>>>0)>(6143); //@line 1191 "load_abc.cpp"
    $or$cond$i = $103 & $104; //@line 1191 "load_abc.cpp"
    $$$i = $or$cond$i ? 15 : $102; //@line 1191 "load_abc.cpp"
    $105 = ($94|0)>(0); //@line 1192 "load_abc.cpp"
    $106 = (0 - ($$$i))|0; //@line 1192 "load_abc.cpp"
    $$02$p$i = $105 ? $$$i : $106; //@line 1192 "load_abc.cpp"
    $$02$i = (($$02$p$i) + ($vol))|0; //@line 1192 "load_abc.cpp"
    $107 = ($$02$i|0)<(2); //@line 1194 "load_abc.cpp"
    $$$02$i = $107 ? 2 : $$02$i; //@line 1194 "load_abc.cpp"
    $108 = ($$$02$i|0)>(127); //@line 1195 "load_abc.cpp"
    $$2$i = $108 ? 127 : $$$02$i; //@line 1195 "load_abc.cpp"
    $$0$i = $$2$i;
   }
  }
 } else {
  $$0$i = 0;
 }
 $109 = $$0$i&255; //@line 1369 "load_abc.cpp"
 $110 = ((($d)) + 3|0); //@line 1369 "load_abc.cpp"
 HEAP8[$110>>0] = $109; //@line 1369 "load_abc.cpp"
 $111 = $fx&255; //@line 1370 "load_abc.cpp"
 $112 = ((($d)) + 4|0); //@line 1370 "load_abc.cpp"
 HEAP8[$112>>0] = $111; //@line 1370 "load_abc.cpp"
 $113 = $fxop&255; //@line 1371 "load_abc.cpp"
 $114 = ((($d)) + 5|0); //@line 1371 "load_abc.cpp"
 HEAP8[$114>>0] = $113; //@line 1371 "load_abc.cpp"
 $115 = $116 << 24 >> 24; //@line 1372 "load_abc.cpp"
 $117 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 1237 "load_abc.cpp"
 $118 = ((($117)) + 36|0); //@line 1238 "load_abc.cpp"
 $119 = HEAP8[$118>>0]|0; //@line 1238 "load_abc.cpp"
 $120 = ((($117)) + 37|0); //@line 1239 "load_abc.cpp"
 $121 = HEAP8[$120>>0]|0; //@line 1239 "load_abc.cpp"
 $122 = ((($h)) + 8|0); //@line 1240 "load_abc.cpp"
 $tp$05$i = HEAPU8[$122>>0]|(HEAPU8[$122+1>>0]<<8)|(HEAPU8[$122+2>>0]<<16)|(HEAPU8[$122+3>>0]<<24); //@line 1240 "load_abc.cpp"
 $123 = ($tp$05$i|0)==(0|0); //@line 1240 "load_abc.cpp"
 L56: do {
  if ($123) {
   $$0$i4 = $117;
  } else {
   $tp$06$i = $tp$05$i;
   while(1) {
    $124 = ((($tp$06$i)) + 36|0); //@line 1241 "load_abc.cpp"
    $125 = HEAP8[$124>>0]|0; //@line 1241 "load_abc.cpp"
    $126 = ($125<<24>>24)==($119<<24>>24); //@line 1241 "load_abc.cpp"
    if ($126) {
     $127 = ((($tp$06$i)) + 16|0); //@line 1242 "load_abc.cpp"
     $128 = HEAPU8[$127>>0]|(HEAPU8[$127+1>>0]<<8)|(HEAPU8[$127+2>>0]<<16)|(HEAPU8[$127+3>>0]<<24); //@line 1242 "load_abc.cpp"
     $129 = ($128|0)==(0|0); //@line 1243 "load_abc.cpp"
     if (!($129)) {
      $130 = ((($128)) + 4|0); //@line 1243 "load_abc.cpp"
      $131 = HEAPU8[$130>>0]|(HEAPU8[$130+1>>0]<<8)|(HEAPU8[$130+2>>0]<<16)|(HEAPU8[$130+3>>0]<<24); //@line 1243 "load_abc.cpp"
      $132 = ($131>>>0)<($tracktime>>>0); //@line 1243 "load_abc.cpp"
      if ($132) {
       $133 = ((($128)) + 9|0); //@line 1244 "load_abc.cpp"
       $134 = HEAP8[$133>>0]|0; //@line 1244 "load_abc.cpp"
       $135 = $134&255; //@line 1244 "load_abc.cpp"
       $136 = ($135|0)==($oct$5|0); //@line 1244 "load_abc.cpp"
       if ($136) {
        $137 = ((($128)) + 8|0); //@line 1244 "load_abc.cpp"
        $138 = HEAP8[$137>>0]|0; //@line 1244 "load_abc.cpp"
        $139 = $138&255; //@line 1244 "load_abc.cpp"
        $140 = (($139) - ($115))|0; //@line 1244 "load_abc.cpp"
        $ispos$i = ($140|0)>(-1); //@line 1244 "load_abc.cpp"
        $neg$i = (0 - ($140))|0; //@line 1244 "load_abc.cpp"
        $141 = $ispos$i ? $140 : $neg$i; //@line 1244 "load_abc.cpp"
        $142 = ($141|0)<(3); //@line 1244 "load_abc.cpp"
        if ($142) {
         $tp$06$i$lcssa = $tp$06$i;
         label = 45;
         break;
        }
       }
      }
     }
    }
    $tp$0$i = HEAPU8[$tp$06$i>>0]|(HEAPU8[$tp$06$i+1>>0]<<8)|(HEAPU8[$tp$06$i+2>>0]<<16)|(HEAPU8[$tp$06$i+3>>0]<<24); //@line 1240 "load_abc.cpp"
    $147 = ($tp$0$i|0)==(0|0); //@line 1240 "load_abc.cpp"
    if ($147) {
     break;
    } else {
     $tp$06$i = $tp$0$i;
    }
   }
   if ((label|0) == 45) {
    $143 = ((($tp$06$i$lcssa)) + 37|0); //@line 1245 "load_abc.cpp"
    $144 = HEAP8[$143>>0]|0; //@line 1245 "load_abc.cpp"
    $145 = ($144<<24>>24)==($121<<24>>24); //@line 1245 "load_abc.cpp"
    if (!($145)) {
     $146 = ((($tp$06$i$lcssa)) + 38|0); //@line 1245 "load_abc.cpp"
     HEAP8[$146>>0] = $121; //@line 1245 "load_abc.cpp"
    }
    HEAP8[$0>>0]=$tp$06$i$lcssa&255;HEAP8[$0+1>>0]=($tp$06$i$lcssa>>8)&255;HEAP8[$0+2>>0]=($tp$06$i$lcssa>>16)&255;HEAP8[$0+3>>0]=$tp$06$i$lcssa>>24; //@line 1246 "load_abc.cpp"
    $$0$i4 = $tp$06$i$lcssa;
    break;
   }
   $148 = ((($117)) + 38|0); //@line 1252 "load_abc.cpp"
   $149 = HEAP8[$148>>0]|0; //@line 1252 "load_abc.cpp"
   $150 = ($121<<24>>24)==($149<<24>>24); //@line 1253 "load_abc.cpp"
   if ($150) {
    $$0$i4 = $117;
   } else {
    $tp$14$i = $tp$05$i;
    while(1) {
     $151 = ((($tp$14$i)) + 36|0); //@line 1256 "load_abc.cpp"
     $152 = HEAP8[$151>>0]|0; //@line 1256 "load_abc.cpp"
     $153 = ($152<<24>>24)==($119<<24>>24); //@line 1256 "load_abc.cpp"
     if ($153) {
      $154 = ((($tp$14$i)) + 37|0); //@line 1256 "load_abc.cpp"
      $155 = HEAP8[$154>>0]|0; //@line 1256 "load_abc.cpp"
      $156 = ($155<<24>>24)==($149<<24>>24); //@line 1256 "load_abc.cpp"
      if ($156) {
       $tp$14$i$lcssa = $tp$14$i;
       break;
      }
     }
     $tp$1$i = HEAPU8[$tp$14$i>>0]|(HEAPU8[$tp$14$i+1>>0]<<8)|(HEAPU8[$tp$14$i+2>>0]<<16)|(HEAPU8[$tp$14$i+3>>0]<<24); //@line 1255 "load_abc.cpp"
     $158 = ($tp$1$i|0)==(0|0); //@line 1255 "load_abc.cpp"
     if ($158) {
      $$0$i4 = $117;
      break L56;
     } else {
      $tp$14$i = $tp$1$i;
     }
    }
    $157 = ((($tp$14$i$lcssa)) + 38|0); //@line 1257 "load_abc.cpp"
    HEAP8[$157>>0] = $121; //@line 1257 "load_abc.cpp"
    HEAP8[$0>>0]=$tp$14$i$lcssa&255;HEAP8[$0+1>>0]=($tp$14$i$lcssa>>8)&255;HEAP8[$0+2>>0]=($tp$14$i$lcssa>>16)&255;HEAP8[$0+3>>0]=$tp$14$i$lcssa>>24; //@line 1258 "load_abc.cpp"
    $$0$i4 = $tp$14$i$lcssa;
   }
  }
 } while(0);
 $159 = ((($$0$i4)) + 16|0); //@line 1373 "load_abc.cpp"
 $160 = HEAPU8[$159>>0]|(HEAPU8[$159+1>>0]<<8)|(HEAPU8[$159+2>>0]<<16)|(HEAPU8[$159+3>>0]<<24); //@line 1373 "load_abc.cpp"
 $161 = ($160|0)==(0|0); //@line 1373 "load_abc.cpp"
 if (!($161)) {
  $162 = ((($160)) + 8|0); //@line 1374 "load_abc.cpp"
  $163 = HEAP8[$162>>0]|0; //@line 1374 "load_abc.cpp"
  $164 = $163&255; //@line 1374 "load_abc.cpp"
  $165 = ($164|0)==($115|0); //@line 1374 "load_abc.cpp"
  do {
   if ($165) {
    $218 = $116;$224 = $82;
   } else {
    $166 = (($164) - ($115))|0; //@line 1375 "load_abc.cpp"
    $ispos = ($166|0)>(-1); //@line 1375 "load_abc.cpp"
    $neg = (0 - ($166))|0; //@line 1375 "load_abc.cpp"
    $167 = $ispos ? $166 : $neg; //@line 1375 "load_abc.cpp"
    $168 = ($167|0)<(3); //@line 1375 "load_abc.cpp"
    if ($168) {
     $169 = (($164) + -23)|0; //@line 1377 "load_abc.cpp"
     $170 = ((($$0$i4)) + 20|0); //@line 1377 "load_abc.cpp"
     $171 = HEAPU8[$170>>0]|(HEAPU8[$170+1>>0]<<8)|(HEAPU8[$170+2>>0]<<16)|(HEAPU8[$170+3>>0]<<24); //@line 1377 "load_abc.cpp"
     $172 = (($169) - ($171))|0; //@line 1377 "load_abc.cpp"
     $173 = ($172|0)>(0);
     $174 = $173 ? $169 : $171; //@line 1378 "load_abc.cpp"
     $175 = (($174) + 34)|0; //@line 1378 "load_abc.cpp"
     $176 = (($175) - ($164))|0; //@line 1378 "load_abc.cpp"
     $177 = (($176>>>0) % 12)&-1; //@line 1378 "load_abc.cpp"
     $178 = (($176) - ($177))|0; //@line 1378 "load_abc.cpp"
     $179 = (($178) + ($164))|0; //@line 1378 "load_abc.cpp"
     $180 = (($179) + -23)|0; //@line 1378 "load_abc.cpp"
     $181 = (($180) - ($171))|0; //@line 1378 "load_abc.cpp"
     $182 = ($181|0)<(12);
     $$op = $181 ^ -1;
     $183 = $182 ? $$op : -13;
     $184 = (($181) + ($183))|0; //@line 1379 "load_abc.cpp"
     $185 = (($184) + 12)|0; //@line 1379 "load_abc.cpp"
     $186 = (($185>>>0) % 12)&-1; //@line 1379 "load_abc.cpp"
     $187 = (($186) - ($185))|0; //@line 1379 "load_abc.cpp"
     $188 = (($181) + ($187))|0; //@line 1379 "load_abc.cpp"
     $189 = (_isupper($n$0)|0); //@line 1380 "load_abc.cpp"
     $190 = ($189|0)==(0); //@line 1380 "load_abc.cpp"
     if ($190) {
      label = 59;
     } else {
      $191 = (($188) + 12)|0; //@line 1380 "load_abc.cpp"
      $192 = (($barkey) + ($191)|0); //@line 1380 "load_abc.cpp"
      $193 = HEAP8[$192>>0]|0; //@line 1380 "load_abc.cpp"
      $194 = ($193<<24>>24)==(32); //@line 1380 "load_abc.cpp"
      if (!($194)) {
       label = 59;
      }
     }
     if ((label|0) == 59) {
      $195 = (_islower($n$0)|0); //@line 1380 "load_abc.cpp"
      $196 = ($195|0)==(0); //@line 1380 "load_abc.cpp"
      if ($196) {
       $218 = $116;$224 = $82;
       break;
      }
      $197 = (($barkey) + ($188)|0); //@line 1380 "load_abc.cpp"
      $198 = HEAP8[$197>>0]|0; //@line 1380 "load_abc.cpp"
      $199 = ($198<<24>>24)==(32); //@line 1380 "load_abc.cpp"
      if (!($199)) {
       $218 = $116;$224 = $82;
       break;
      }
     }
     $200 = (($barkey) + ($k$0$lcssa)|0); //@line 1381 "load_abc.cpp"
     HEAP8[$200>>0] = 32; //@line 1381 "load_abc.cpp"
     $201 = (_isupper($n$0)|0); //@line 1382 "load_abc.cpp"
     $202 = ($201|0)==(0); //@line 1382 "load_abc.cpp"
     $203 = $n$0&255; //@line 1383 "load_abc.cpp"
     if ($202) {
      $205 = (($188) + 12)|0; //@line 1385 "load_abc.cpp"
      $206 = (($barkey) + ($205)|0); //@line 1385 "load_abc.cpp"
      HEAP8[$206>>0] = $203; //@line 1385 "load_abc.cpp"
     } else {
      $204 = (($barkey) + ($188)|0); //@line 1383 "load_abc.cpp"
      HEAP8[$204>>0] = $203; //@line 1383 "load_abc.cpp"
     }
     $207 = HEAPU8[$159>>0]|(HEAPU8[$159+1>>0]<<8)|(HEAPU8[$159+2>>0]<<16)|(HEAPU8[$159+3>>0]<<24); //@line 1386 "load_abc.cpp"
     $208 = ((($207)) + 8|0); //@line 1386 "load_abc.cpp"
     $209 = HEAP8[$208>>0]|0; //@line 1386 "load_abc.cpp"
     HEAP8[$d>>0] = $209; //@line 1386 "load_abc.cpp"
     $210 = HEAPU8[$159>>0]|(HEAPU8[$159+1>>0]<<8)|(HEAPU8[$159+2>>0]<<16)|(HEAPU8[$159+3>>0]<<24); //@line 1387 "load_abc.cpp"
     $211 = ((($210)) + 9|0); //@line 1387 "load_abc.cpp"
     $212 = HEAP8[$211>>0]|0; //@line 1387 "load_abc.cpp"
     HEAP8[$83>>0] = $212; //@line 1387 "load_abc.cpp"
     $218 = $209;$224 = $212;
    } else {
     $218 = $116;$224 = $82;
    }
   }
  } while(0);
  $$pr = HEAPU8[$159>>0]|(HEAPU8[$159+1>>0]<<8)|(HEAPU8[$159+2>>0]<<16)|(HEAPU8[$159+3>>0]<<24); //@line 1392 "load_abc.cpp"
  $213 = ($$pr|0)==(0|0); //@line 1392 "load_abc.cpp"
  if (!($213)) {
   $214 = ((($$pr)) + 8|0); //@line 1393 "load_abc.cpp"
   $215 = HEAP8[$214>>0]|0; //@line 1393 "load_abc.cpp"
   $216 = $215&255; //@line 1393 "load_abc.cpp"
   $217 = $218 << 24 >> 24; //@line 1393 "load_abc.cpp"
   $219 = ($216|0)==($217|0); //@line 1393 "load_abc.cpp"
   if ($219) {
    $220 = ((($$pr)) + 9|0); //@line 1394 "load_abc.cpp"
    $221 = HEAP8[$220>>0]|0; //@line 1394 "load_abc.cpp"
    $222 = $221&255; //@line 1394 "load_abc.cpp"
    $223 = $224 << 24 >> 24; //@line 1394 "load_abc.cpp"
    $225 = ($222|0)==($223|0); //@line 1394 "load_abc.cpp"
    if ($225) {
     $227 = $215;$e$021 = $$pr;
     while(1) {
      $226 = ($227<<24>>24)==(0); //@line 1396 "load_abc.cpp"
      if ($226) {
       $228 = ((($e$021)) + 9|0); //@line 1396 "load_abc.cpp"
       $229 = HEAP8[$228>>0]|0; //@line 1396 "load_abc.cpp"
       $230 = ($229<<24>>24)==(0); //@line 1396 "load_abc.cpp"
       if ($230) {
        $$lcssa = $228;$e$021$lcssa = $e$021;
        label = 70;
        break;
       }
      }
      $233 = HEAPU8[$e$021>>0]|(HEAPU8[$e$021+1>>0]<<8)|(HEAPU8[$e$021+2>>0]<<16)|(HEAPU8[$e$021+3>>0]<<24); //@line 1395 "load_abc.cpp"
      $234 = ($233|0)==(0|0); //@line 1395 "load_abc.cpp"
      if ($234) {
       break;
      }
      $$phi$trans$insert = ((($233)) + 8|0);
      $$pre = HEAP8[$$phi$trans$insert>>0]|0; //@line 1396 "load_abc.cpp"
      $227 = $$pre;$e$021 = $233;
     }
     if ((label|0) == 70) {
      $231 = ((($e$021$lcssa)) + 8|0); //@line 1396 "load_abc.cpp"
      HEAP8[$231>>0] = 1; //@line 1397 "load_abc.cpp"
      HEAP8[$$lcssa>>0] = 72; //@line 1398 "load_abc.cpp"
      $232 = ((($e$021$lcssa)) + 10|0); //@line 1399 "load_abc.cpp"
      HEAP8[$232>>0]=0&255;HEAP8[$232+1>>0]=(0>>8)&255;HEAP8[$232+2>>0]=(0>>16)&255;HEAP8[$232+3>>0]=0>>24; //@line 1399 "load_abc.cpp"
     }
     $235 = ((($$pr)) + 15|0); //@line 1403 "load_abc.cpp"
     HEAP8[$235>>0] = 1; //@line 1403 "load_abc.cpp"
     $j$0 = $i$1$lcssa;
     L105: while(1) {
      $236 = (($p) + ($j$0)|0); //@line 1404 "load_abc.cpp"
      $237 = HEAP8[$236>>0]|0; //@line 1404 "load_abc.cpp"
      $238 = $237 << 24 >> 24; //@line 1404 "load_abc.cpp"
      $isdigittmp = (($238) + -48)|0; //@line 1404 "load_abc.cpp"
      $isdigit = ($isdigittmp>>>0)<(10); //@line 1404 "load_abc.cpp"
      if (!($isdigit)) {
       switch ($237<<24>>24) {
       case 45:  {
        $$0 = $i$1$lcssa;
        label = 89;
        break L105;
        break;
       }
       case 47:  {
        break;
       }
       default: {
        break L105;
       }
       }
      }
      $239 = (($j$0) + 1)|0; //@line 1404 "load_abc.cpp"
      $j$0 = $239;
     }
     if ((label|0) == 89) {
      STACKTOP = sp;return ($$0|0); //@line 1426 "load_abc.cpp"
     }
     HEAP8[$159>>0]=0&255;HEAP8[$159+1>>0]=(0>>8)&255;HEAP8[$159+2>>0]=(0>>16)&255;HEAP8[$159+3>>0]=0>>24; //@line 1406 "load_abc.cpp"
     $$0 = $i$1$lcssa;
     STACKTOP = sp;return ($$0|0); //@line 1426 "load_abc.cpp"
    }
   }
  }
 }
 HEAP8[$159>>0]=0&255;HEAP8[$159+1>>0]=(0>>8)&255;HEAP8[$159+2>>0]=(0>>16)&255;HEAP8[$159+3>>0]=0>>24; //@line 1409 "load_abc.cpp"
 $240 = ((($$0$i4)) + 8|0); //@line 1410 "load_abc.cpp"
 $241 = HEAPU8[$240>>0]|(HEAPU8[$240+1>>0]<<8)|(HEAPU8[$240+2>>0]<<16)|(HEAPU8[$240+3>>0]<<24); //@line 1410 "load_abc.cpp"
 $242 = ($241|0)==(0|0); //@line 1410 "load_abc.cpp"
 if ($242) {
  label = 83;
 } else {
  $243 = ((($241)) + 4|0); //@line 1411 "load_abc.cpp"
  $244 = HEAPU8[$243>>0]|(HEAPU8[$243+1>>0]<<8)|(HEAPU8[$243+2>>0]<<16)|(HEAPU8[$243+3>>0]<<24); //@line 1411 "load_abc.cpp"
  $245 = ($244|0)==($tracktime|0); //@line 1411 "load_abc.cpp"
  if ($245) {
   $246 = ((($241)) + 8|0); //@line 1412 "load_abc.cpp"
   $247 = HEAP8[$246>>0]|0; //@line 1412 "load_abc.cpp"
   $248 = ($247<<24>>24)==(0); //@line 1412 "load_abc.cpp"
   if ($248) {
    $249 = ((($241)) + 9|0); //@line 1413 "load_abc.cpp"
    $250 = HEAP8[$249>>0]|0; //@line 1413 "load_abc.cpp"
    $251 = ($250<<24>>24)==(0); //@line 1413 "load_abc.cpp"
    if ($251) {
     $252 = HEAP8[$d>>0]|0; //@line 1415 "load_abc.cpp"
     HEAP8[$246>>0] = $252; //@line 1415 "load_abc.cpp"
     $253 = HEAP8[$83>>0]|0; //@line 1415 "load_abc.cpp"
     $254 = HEAPU8[$240>>0]|(HEAPU8[$240+1>>0]<<8)|(HEAPU8[$240+2>>0]<<16)|(HEAPU8[$240+3>>0]<<24); //@line 1415 "load_abc.cpp"
     $255 = ((($254)) + 9|0); //@line 1415 "load_abc.cpp"
     HEAP8[$255>>0] = $253; //@line 1415 "load_abc.cpp"
     $256 = HEAP8[$89>>0]|0; //@line 1415 "load_abc.cpp"
     $257 = HEAPU8[$240>>0]|(HEAPU8[$240+1>>0]<<8)|(HEAPU8[$240+2>>0]<<16)|(HEAPU8[$240+3>>0]<<24); //@line 1415 "load_abc.cpp"
     $258 = ((($257)) + 10|0); //@line 1415 "load_abc.cpp"
     HEAP8[$258>>0] = $256; //@line 1415 "load_abc.cpp"
     $259 = HEAP8[$110>>0]|0; //@line 1415 "load_abc.cpp"
     $260 = HEAPU8[$240>>0]|(HEAPU8[$240+1>>0]<<8)|(HEAPU8[$240+2>>0]<<16)|(HEAPU8[$240+3>>0]<<24); //@line 1415 "load_abc.cpp"
     $261 = ((($260)) + 11|0); //@line 1415 "load_abc.cpp"
     HEAP8[$261>>0] = $259; //@line 1415 "load_abc.cpp"
     $262 = HEAP8[$112>>0]|0; //@line 1415 "load_abc.cpp"
     $263 = HEAPU8[$240>>0]|(HEAPU8[$240+1>>0]<<8)|(HEAPU8[$240+2>>0]<<16)|(HEAPU8[$240+3>>0]<<24); //@line 1415 "load_abc.cpp"
     $264 = ((($263)) + 12|0); //@line 1415 "load_abc.cpp"
     HEAP8[$264>>0] = $262; //@line 1415 "load_abc.cpp"
     $265 = HEAP8[$114>>0]|0; //@line 1415 "load_abc.cpp"
     $266 = HEAPU8[$240>>0]|(HEAPU8[$240+1>>0]<<8)|(HEAPU8[$240+2>>0]<<16)|(HEAPU8[$240+3>>0]<<24); //@line 1415 "load_abc.cpp"
     $267 = ((($266)) + 13|0); //@line 1415 "load_abc.cpp"
     HEAP8[$267>>0] = $265; //@line 1415 "load_abc.cpp"
    } else {
     label = 83;
    }
   } else {
    label = 83;
   }
  } else {
   label = 83;
  }
 }
 if ((label|0) == 83) {
  $268 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
  HEAP8[$268>>0]=0&255;HEAP8[$268+1>>0]=(0>>8)&255;HEAP8[$268+2>>0]=(0>>16)&255;HEAP8[$268+3>>0]=0>>24; //@line 540 "load_abc.cpp"
  $269 = ((($268)) + 4|0); //@line 541 "load_abc.cpp"
  HEAP8[$269>>0]=$tracktime&255;HEAP8[$269+1>>0]=($tracktime>>8)&255;HEAP8[$269+2>>0]=($tracktime>>16)&255;HEAP8[$269+3>>0]=$tracktime>>24; //@line 541 "load_abc.cpp"
  $scevgep$i = ((($268)) + 8|0); //@line 1418 "load_abc.cpp"
  ;HEAP8[$scevgep$i>>0]=HEAP8[$d>>0]|0;HEAP8[$scevgep$i+1>>0]=HEAP8[$d+1>>0]|0;HEAP8[$scevgep$i+2>>0]=HEAP8[$d+2>>0]|0;HEAP8[$scevgep$i+3>>0]=HEAP8[$d+3>>0]|0;HEAP8[$scevgep$i+4>>0]=HEAP8[$d+4>>0]|0;HEAP8[$scevgep$i+5>>0]=HEAP8[$d+5>>0]|0; //@line 543 "load_abc.cpp"
  $270 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
  $271 = ((($268)) + 14|0); //@line 544 "load_abc.cpp"
  HEAP8[$271>>0] = $270; //@line 544 "load_abc.cpp"
  $272 = ((($268)) + 15|0); //@line 545 "load_abc.cpp"
  HEAP8[$272>>0] = 0; //@line 545 "load_abc.cpp"
  __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($$0$i4,$268); //@line 1419 "load_abc.cpp"
 }
 $273 = ($i$1$lcssa|0)>(0); //@line 1421 "load_abc.cpp"
 if (!($273)) {
  $$0 = $i$1$lcssa;
  STACKTOP = sp;return ($$0|0); //@line 1426 "load_abc.cpp"
 }
 $274 = (($i$1$lcssa) + -1)|0; //@line 1421 "load_abc.cpp"
 $275 = (($p) + ($274)|0); //@line 1421 "load_abc.cpp"
 $276 = HEAP8[$275>>0]|0; //@line 1421 "load_abc.cpp"
 $277 = ($276<<24>>24)==(34); //@line 1421 "load_abc.cpp"
 if (!($277)) {
  $$0 = $i$1$lcssa;
  STACKTOP = sp;return ($$0|0); //@line 1426 "load_abc.cpp"
 }
 $278 = ((($h)) + 24|0); //@line 1423 "load_abc.cpp"
 $279 = HEAPU8[$278>>0]|(HEAPU8[$278+1>>0]<<8)|(HEAPU8[$278+2>>0]<<16)|(HEAPU8[$278+3>>0]<<24); //@line 1423 "load_abc.cpp"
 $280 = (_strlen($279)|0); //@line 323 "load_abc.cpp"
 $281 = (($280) + 36)|0; //@line 323 "load_abc.cpp"
 $282 = ($281>>>0)>(255); //@line 323 "load_abc.cpp"
 if (!($282)) {
  HEAP32[$vararg_buffer>>2] = $279; //@line 324 "load_abc.cpp"
  (_sprintf($txt$i,527248,$vararg_buffer)|0); //@line 324 "load_abc.cpp"
  $283 = HEAP32[37544]|0; //@line 328 "load_abc.cpp"
  HEAP32[$vararg_buffer1>>2] = $txt$i; //@line 328 "load_abc.cpp"
  (_fprintf($283,524747,$vararg_buffer1)|0); //@line 328 "load_abc.cpp"
 }
 $$0 = $274;
 STACKTOP = sp;return ($$0|0); //@line 1426 "load_abc.cpp"
}
function __ZL14abc_add_gchordP10_ABCHANDLEjj($h,$tracktime,$bartime) {
 $h = $h|0;
 $tracktime = $tracktime|0;
 $bartime = $bartime|0;
 var $$ = 0, $$1 = 0, $$1$lcssa = 0, $$1$lcssa154 = 0, $$1$lcssa155 = 0, $$1$lcssa156 = 0, $$1$lcssa157 = 0, $$1159 = 0, $$1160 = 0, $$2 = 0, $$2$v = 0, $$i$0$i = 0, $$lcssa = 0, $$lcssa145 = 0, $$lcssa146 = 0, $$lcssa147 = 0, $$lcssa148 = 0, $$lcssa152 = 0, $$pn = 0, $$pn64 = 0;
 var $$pre = 0, $$pre89 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0;
 var $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0;
 var $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0;
 var $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0;
 var $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $c$0$ = 0;
 var $c$0$$1 = 0, $c$0$$2 = 0, $c$0$$3 = 0, $c$0$$4 = 0, $c$0$$5 = 0, $c$0$$6 = 0, $c$058 = 0, $c$1 = 0, $c$1$lcssa = 0, $e$0 = 0, $e$0$c$0 = 0, $e$057 = 0, $e$059 = 0, $etime$1$1 = 0, $etime$1$2 = 0, $etime$1$3 = 0, $etime$1$4 = 0, $etime$1$5 = 0, $etime$1$6 = 0, $g$0$lcssa = 0;
 var $g$050 = 0, $g$1 = 0, $g$1$ph = 0, $i$0$lcssa$i = 0, $i$02$i = 0, $i$148 = 0, $isdigit$i = 0, $isdigittmp$i = 0, $p$01$i = 0, $rtime$0 = 0, $rtime$149 = 0, $rtime$2$ph = 0, $rtime$2$ph$be = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($h)) + 416|0); //@line 3398 "load_abc.cpp"
 $1 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 3398 "load_abc.cpp"
 $2 = ((($1)) + 4|0); //@line 3398 "load_abc.cpp"
 $e$057 = HEAPU8[$2>>0]|(HEAPU8[$2+1>>0]<<8)|(HEAPU8[$2+2>>0]<<16)|(HEAPU8[$2+3>>0]<<24); //@line 3398 "load_abc.cpp"
 $3 = ($e$057|0)==(0|0); //@line 3398 "load_abc.cpp"
 if ($3) {
  return; //@line 3483 "load_abc.cpp"
 } else {
  $c$058 = 0;$e$059 = $e$057;
 }
 while(1) {
  $4 = ((($e$059)) + 8|0); //@line 3399 "load_abc.cpp"
  $5 = HEAP8[$4>>0]|0; //@line 3399 "load_abc.cpp"
  $6 = ($5<<24>>24)==(1); //@line 3399 "load_abc.cpp"
  if ($6) {
   $7 = ((($e$059)) + 9|0); //@line 3399 "load_abc.cpp"
   $8 = HEAP8[$7>>0]|0; //@line 3399 "load_abc.cpp"
   $9 = ($8<<24>>24)==(67); //@line 3399 "load_abc.cpp"
   $e$0$c$0 = $9 ? $e$059 : $c$058; //@line 3399 "load_abc.cpp"
   $c$1 = $e$0$c$0;
  } else {
   $c$1 = $c$058;
  }
  $e$0 = HEAPU8[$e$059>>0]|(HEAPU8[$e$059+1>>0]<<8)|(HEAPU8[$e$059+2>>0]<<16)|(HEAPU8[$e$059+3>>0]<<24); //@line 3398 "load_abc.cpp"
  $10 = ($e$0|0)==(0|0); //@line 3398 "load_abc.cpp"
  if ($10) {
   $c$1$lcssa = $c$1;
   break;
  } else {
   $c$058 = $c$1;$e$059 = $e$0;
  }
 }
 $11 = ($c$1$lcssa|0)==(0|0); //@line 3401 "load_abc.cpp"
 if ($11) {
  return; //@line 3483 "load_abc.cpp"
 }
 $12 = ((($c$1$lcssa)) + 10|0); //@line 3402 "load_abc.cpp"
 $13 = HEAP8[$12>>0]|0; //@line 3402 "load_abc.cpp"
 $14 = $13&255; //@line 3402 "load_abc.cpp"
 $15 = ((($h)) + 36|0); //@line 3403 "load_abc.cpp"
 $16 = HEAP8[$15>>0]|0; //@line 3385 "load_abc.cpp"
 $17 = ($16<<24>>24)==(0); //@line 3385 "load_abc.cpp"
 if ($17) {
  $i$0$lcssa$i = 0;
 } else {
  $19 = $16;$i$02$i = 0;$p$01$i = $15;
  while(1) {
   $18 = $19 << 24 >> 24; //@line 3386 "load_abc.cpp"
   $isdigittmp$i = (($18) + -48)|0; //@line 3386 "load_abc.cpp"
   $isdigit$i = ($isdigittmp$i>>>0)<(10); //@line 3386 "load_abc.cpp"
   $20 = $isdigit$i ? $isdigittmp$i : 0; //@line 3386 "load_abc.cpp"
   $$i$0$i = (($20) + ($i$02$i))|0; //@line 3386 "load_abc.cpp"
   $21 = ((($p$01$i)) + 1|0); //@line 3385 "load_abc.cpp"
   $22 = HEAP8[$21>>0]|0; //@line 3385 "load_abc.cpp"
   $23 = ($22<<24>>24)==(0); //@line 3385 "load_abc.cpp"
   if ($23) {
    $i$0$lcssa$i = $$i$0$i;
    break;
   } else {
    $19 = $22;$i$02$i = $$i$0$i;$p$01$i = $21;
   }
  }
 }
 $24 = ((($h)) + 356|0); //@line 3404 "load_abc.cpp"
 $25 = HEAPU8[$24>>0]|(HEAPU8[$24+1>>0]<<8)|(HEAPU8[$24+2>>0]<<16)|(HEAPU8[$24+3>>0]<<24); //@line 3404 "load_abc.cpp"
 $26 = ((($1)) + 44|0); //@line 3407 "load_abc.cpp"
 $27 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,$26,1)|0); //@line 3407 "load_abc.cpp"
 $28 = ((($27)) + 8|0); //@line 3408 "load_abc.cpp"
 $29 = HEAPU8[$28>>0]|(HEAPU8[$28+1>>0]<<8)|(HEAPU8[$28+2>>0]<<16)|(HEAPU8[$28+3>>0]<<24); //@line 3408 "load_abc.cpp"
 $30 = ($29|0)==(0|0); //@line 3409 "load_abc.cpp"
 $c$0$ = $30 ? $c$1$lcssa : $29; //@line 3409 "load_abc.cpp"
 $31 = ((($c$0$)) + 4|0); //@line 3410 "load_abc.cpp"
 $32 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 3410 "load_abc.cpp"
 $33 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 3407 "load_abc.cpp"
 $34 = ((($33)) + 44|0); //@line 3407 "load_abc.cpp"
 $35 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,$34,2)|0); //@line 3407 "load_abc.cpp"
 $36 = ((($35)) + 8|0); //@line 3408 "load_abc.cpp"
 $37 = HEAPU8[$36>>0]|(HEAPU8[$36+1>>0]<<8)|(HEAPU8[$36+2>>0]<<16)|(HEAPU8[$36+3>>0]<<24); //@line 3408 "load_abc.cpp"
 $38 = ($37|0)==(0|0); //@line 3409 "load_abc.cpp"
 $c$0$$1 = $38 ? $c$1$lcssa : $37; //@line 3409 "load_abc.cpp"
 $39 = ((($c$0$$1)) + 4|0); //@line 3410 "load_abc.cpp"
 $40 = HEAPU8[$39>>0]|(HEAPU8[$39+1>>0]<<8)|(HEAPU8[$39+2>>0]<<16)|(HEAPU8[$39+3>>0]<<24); //@line 3410 "load_abc.cpp"
 $41 = ($40>>>0)>($32>>>0); //@line 3411 "load_abc.cpp"
 $etime$1$1 = $41 ? $40 : $32; //@line 3411 "load_abc.cpp"
 $42 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 3407 "load_abc.cpp"
 $43 = ((($42)) + 44|0); //@line 3407 "load_abc.cpp"
 $44 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,$43,3)|0); //@line 3407 "load_abc.cpp"
 $45 = ((($44)) + 8|0); //@line 3408 "load_abc.cpp"
 $46 = HEAPU8[$45>>0]|(HEAPU8[$45+1>>0]<<8)|(HEAPU8[$45+2>>0]<<16)|(HEAPU8[$45+3>>0]<<24); //@line 3408 "load_abc.cpp"
 $47 = ($46|0)==(0|0); //@line 3409 "load_abc.cpp"
 $c$0$$2 = $47 ? $c$1$lcssa : $46; //@line 3409 "load_abc.cpp"
 $48 = ((($c$0$$2)) + 4|0); //@line 3410 "load_abc.cpp"
 $49 = HEAPU8[$48>>0]|(HEAPU8[$48+1>>0]<<8)|(HEAPU8[$48+2>>0]<<16)|(HEAPU8[$48+3>>0]<<24); //@line 3410 "load_abc.cpp"
 $50 = ($49>>>0)>($etime$1$1>>>0); //@line 3411 "load_abc.cpp"
 $etime$1$2 = $50 ? $49 : $etime$1$1; //@line 3411 "load_abc.cpp"
 $51 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 3407 "load_abc.cpp"
 $52 = ((($51)) + 44|0); //@line 3407 "load_abc.cpp"
 $53 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,$52,4)|0); //@line 3407 "load_abc.cpp"
 $54 = ((($53)) + 8|0); //@line 3408 "load_abc.cpp"
 $55 = HEAPU8[$54>>0]|(HEAPU8[$54+1>>0]<<8)|(HEAPU8[$54+2>>0]<<16)|(HEAPU8[$54+3>>0]<<24); //@line 3408 "load_abc.cpp"
 $56 = ($55|0)==(0|0); //@line 3409 "load_abc.cpp"
 $c$0$$3 = $56 ? $c$1$lcssa : $55; //@line 3409 "load_abc.cpp"
 $57 = ((($c$0$$3)) + 4|0); //@line 3410 "load_abc.cpp"
 $58 = HEAPU8[$57>>0]|(HEAPU8[$57+1>>0]<<8)|(HEAPU8[$57+2>>0]<<16)|(HEAPU8[$57+3>>0]<<24); //@line 3410 "load_abc.cpp"
 $59 = ($58>>>0)>($etime$1$2>>>0); //@line 3411 "load_abc.cpp"
 $etime$1$3 = $59 ? $58 : $etime$1$2; //@line 3411 "load_abc.cpp"
 $60 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 3407 "load_abc.cpp"
 $61 = ((($60)) + 44|0); //@line 3407 "load_abc.cpp"
 $62 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,$61,5)|0); //@line 3407 "load_abc.cpp"
 $63 = ((($62)) + 8|0); //@line 3408 "load_abc.cpp"
 $64 = HEAPU8[$63>>0]|(HEAPU8[$63+1>>0]<<8)|(HEAPU8[$63+2>>0]<<16)|(HEAPU8[$63+3>>0]<<24); //@line 3408 "load_abc.cpp"
 $65 = ($64|0)==(0|0); //@line 3409 "load_abc.cpp"
 $c$0$$4 = $65 ? $c$1$lcssa : $64; //@line 3409 "load_abc.cpp"
 $66 = ((($c$0$$4)) + 4|0); //@line 3410 "load_abc.cpp"
 $67 = HEAPU8[$66>>0]|(HEAPU8[$66+1>>0]<<8)|(HEAPU8[$66+2>>0]<<16)|(HEAPU8[$66+3>>0]<<24); //@line 3410 "load_abc.cpp"
 $68 = ($67>>>0)>($etime$1$3>>>0); //@line 3411 "load_abc.cpp"
 $etime$1$4 = $68 ? $67 : $etime$1$3; //@line 3411 "load_abc.cpp"
 $69 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 3407 "load_abc.cpp"
 $70 = ((($69)) + 44|0); //@line 3407 "load_abc.cpp"
 $71 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,$70,6)|0); //@line 3407 "load_abc.cpp"
 $72 = ((($71)) + 8|0); //@line 3408 "load_abc.cpp"
 $73 = HEAPU8[$72>>0]|(HEAPU8[$72+1>>0]<<8)|(HEAPU8[$72+2>>0]<<16)|(HEAPU8[$72+3>>0]<<24); //@line 3408 "load_abc.cpp"
 $74 = ($73|0)==(0|0); //@line 3409 "load_abc.cpp"
 $c$0$$5 = $74 ? $c$1$lcssa : $73; //@line 3409 "load_abc.cpp"
 $75 = ((($c$0$$5)) + 4|0); //@line 3410 "load_abc.cpp"
 $76 = HEAPU8[$75>>0]|(HEAPU8[$75+1>>0]<<8)|(HEAPU8[$75+2>>0]<<16)|(HEAPU8[$75+3>>0]<<24); //@line 3410 "load_abc.cpp"
 $77 = ($76>>>0)>($etime$1$4>>>0); //@line 3411 "load_abc.cpp"
 $etime$1$5 = $77 ? $76 : $etime$1$4; //@line 3411 "load_abc.cpp"
 $78 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 3407 "load_abc.cpp"
 $79 = ((($78)) + 44|0); //@line 3407 "load_abc.cpp"
 $80 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,$79,7)|0); //@line 3407 "load_abc.cpp"
 $81 = ((($80)) + 8|0); //@line 3408 "load_abc.cpp"
 $82 = HEAPU8[$81>>0]|(HEAPU8[$81+1>>0]<<8)|(HEAPU8[$81+2>>0]<<16)|(HEAPU8[$81+3>>0]<<24); //@line 3408 "load_abc.cpp"
 $83 = ($82|0)==(0|0); //@line 3409 "load_abc.cpp"
 $c$0$$6 = $83 ? $c$1$lcssa : $82; //@line 3409 "load_abc.cpp"
 $84 = ((($c$0$$6)) + 4|0); //@line 3410 "load_abc.cpp"
 $85 = HEAPU8[$84>>0]|(HEAPU8[$84+1>>0]<<8)|(HEAPU8[$84+2>>0]<<16)|(HEAPU8[$84+3>>0]<<24); //@line 3410 "load_abc.cpp"
 $86 = ($85>>>0)>($etime$1$5>>>0); //@line 3411 "load_abc.cpp"
 $etime$1$6 = $86 ? $85 : $etime$1$5; //@line 3411 "load_abc.cpp"
 $87 = ($etime$1$6>>>0)>($tracktime>>>0); //@line 3413 "load_abc.cpp"
 if ($87) {
  return; //@line 3483 "load_abc.cpp"
 }
 $88 = ($etime$1$6>>>0)<($bartime>>>0); //@line 3414 "load_abc.cpp"
 if ($88) {
  $89 = HEAPU8[$24>>0]|(HEAPU8[$24+1>>0]<<8)|(HEAPU8[$24+2>>0]<<16)|(HEAPU8[$24+3>>0]<<24); //@line 3414 "load_abc.cpp"
  $90 = (($bartime) - ($etime$1$6))|0; //@line 3414 "load_abc.cpp"
  $91 = (($90>>>0) % ($89>>>0))&-1; //@line 3414 "load_abc.cpp"
  $92 = (($89) - ($91))|0; //@line 3414 "load_abc.cpp"
  $rtime$0 = $92;
 } else {
  $93 = (($etime$1$6) - ($bartime))|0; //@line 3415 "load_abc.cpp"
  $94 = HEAPU8[$24>>0]|(HEAPU8[$24+1>>0]<<8)|(HEAPU8[$24+2>>0]<<16)|(HEAPU8[$24+3>>0]<<24); //@line 3415 "load_abc.cpp"
  $95 = (($93>>>0) % ($94>>>0))&-1; //@line 3415 "load_abc.cpp"
  $rtime$0 = $95;
 }
 $96 = Math_imul($25, $i$0$lcssa$i)|0; //@line 3416 "load_abc.cpp"
 $97 = Math_imul($rtime$0, $i$0$lcssa$i)|0; //@line 3417 "load_abc.cpp"
 $98 = (($97) + ($96))|0; //@line 3418 "load_abc.cpp"
 $99 = (_strlen($15)|0); //@line 3419 "load_abc.cpp"
 $100 = ($98>>>0)>($96>>>0); //@line 3421 "load_abc.cpp"
 if ($100) {
  $g$050 = 0;$rtime$149 = $98;
  while(1) {
   $101 = $g$050 << 1; //@line 3422 "load_abc.cpp"
   $102 = $101 | 1; //@line 3422 "load_abc.cpp"
   $103 = (((($h)) + 36|0) + ($102)|0); //@line 3422 "load_abc.cpp"
   $104 = HEAP8[$103>>0]|0; //@line 3422 "load_abc.cpp"
   $105 = $104 << 24 >> 24; //@line 3422 "load_abc.cpp"
   $106 = (($105) + -48)|0; //@line 3422 "load_abc.cpp"
   $107 = Math_imul($106, $25)|0; //@line 3423 "load_abc.cpp"
   $108 = (($rtime$149) - ($107))|0; //@line 3423 "load_abc.cpp"
   $109 = (($g$050) + 1)|0; //@line 3424 "load_abc.cpp"
   $110 = ($109|0)==($99|0); //@line 3424 "load_abc.cpp"
   $$ = $110 ? 0 : $109; //@line 3424 "load_abc.cpp"
   $111 = ($108>>>0)>($96>>>0); //@line 3421 "load_abc.cpp"
   if ($111) {
    $g$050 = $$;$rtime$149 = $108;
   } else {
    $g$0$lcssa = $$;
    break;
   }
  }
 } else {
  $g$0$lcssa = 0;
 }
 $112 = (($tracktime) - ($etime$1$6))|0; //@line 3426 "load_abc.cpp"
 $113 = Math_imul($112, $i$0$lcssa$i)|0; //@line 3426 "load_abc.cpp"
 $114 = (1199460 + ($14<<2)|0); //@line 3468 "load_abc.cpp"
 $115 = ((($h)) + 372|0); //@line 3436 "load_abc.cpp"
 $116 = ((($c$1$lcssa)) + 11|0); //@line 3437 "load_abc.cpp"
 $117 = (1199780 + (($14*24)|0)|0); //@line 3437 "load_abc.cpp"
 $118 = ((($h)) + 368|0); //@line 3438 "load_abc.cpp"
 $119 = ((($c$1$lcssa)) + 12|0); //@line 3453 "load_abc.cpp"
 $120 = ((($h)) + 364|0); //@line 3470 "load_abc.cpp"
 $121 = ((($h)) + 360|0); //@line 3473 "load_abc.cpp"
 $g$1$ph = $g$0$lcssa;$rtime$2$ph = 0;
 L27: while(1) {
  $122 = ($rtime$2$ph>>>0)<($113>>>0); //@line 3428 "load_abc.cpp"
  $g$1 = $g$1$ph;
  L29: while(1) {
   if (!($122)) {
    break L27;
   }
   $123 = $g$1 << 1; //@line 3429 "load_abc.cpp"
   $124 = (((($h)) + 36|0) + ($123)|0); //@line 3429 "load_abc.cpp"
   $125 = HEAP8[$124>>0]|0; //@line 3429 "load_abc.cpp"
   $126 = $125 << 24 >> 24; //@line 3429 "load_abc.cpp"
   $127 = (($g$1) + 1)|0; //@line 3431 "load_abc.cpp"
   $128 = ($127|0)==($99|0); //@line 3431 "load_abc.cpp"
   $$1 = $128 ? 0 : $127; //@line 3431 "load_abc.cpp"
   switch ($126|0) {
   case 98:  {
    $$1$lcssa = $$1;$$lcssa = $123;
    label = 18;
    break L29;
    break;
   }
   case 99:  {
    $$1$lcssa154 = $$1;$$lcssa145 = $123;
    label = 19;
    break L29;
    break;
   }
   case 102:  {
    $$1$lcssa155 = $$1;$$lcssa146 = $123;
    label = 24;
    break L29;
    break;
   }
   case 74: case 73: case 72: case 71: case 106: case 105: case 104: case 103:  {
    $$1$lcssa156 = $$1;$$lcssa147 = $123;$$lcssa152 = $126;
    label = 25;
    break L29;
    break;
   }
   case 122:  {
    $$1$lcssa157 = $$1;$$lcssa148 = $123;
    label = 28;
    break L29;
    break;
   }
   default: {
    $g$1 = $$1;
   }
   }
  }
  if ((label|0) == 18) {
   label = 0;
   $129 = $$lcssa | 1; //@line 3430 "load_abc.cpp"
   $130 = (((($h)) + 36|0) + ($129)|0); //@line 3430 "load_abc.cpp"
   $131 = HEAP8[$130>>0]|0; //@line 3430 "load_abc.cpp"
   $132 = $131 << 24 >> 24; //@line 3430 "load_abc.cpp"
   $133 = (($132) + -48)|0; //@line 3430 "load_abc.cpp"
   $134 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 3435 "load_abc.cpp"
   $135 = ((($134)) + 44|0); //@line 3435 "load_abc.cpp"
   $136 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,$135,2)|0); //@line 3435 "load_abc.cpp"
   $137 = HEAPU8[$115>>0]|(HEAPU8[$115+1>>0]<<8)|(HEAPU8[$115+2>>0]<<16)|(HEAPU8[$115+3>>0]<<24); //@line 3436 "load_abc.cpp"
   $138 = $137&255; //@line 3436 "load_abc.cpp"
   $139 = ((($136)) + 42|0); //@line 3436 "load_abc.cpp"
   HEAP8[$139>>0] = $138; //@line 3436 "load_abc.cpp"
   $140 = HEAP8[$116>>0]|0; //@line 3437 "load_abc.cpp"
   $141 = $140&255; //@line 3437 "load_abc.cpp"
   $142 = HEAP32[$117>>2]|0; //@line 3437 "load_abc.cpp"
   $143 = (($rtime$2$ph>>>0) / ($i$0$lcssa$i>>>0))&-1; //@line 3438 "load_abc.cpp"
   $144 = (($143) + ($etime$1$6))|0; //@line 3438 "load_abc.cpp"
   $145 = HEAPU8[$118>>0]|(HEAPU8[$118+1>>0]<<8)|(HEAPU8[$118+2>>0]<<16)|(HEAPU8[$118+3>>0]<<24); //@line 3438 "load_abc.cpp"
   $146 = (($142) + 47)|0; //@line 3437 "load_abc.cpp"
   $147 = (($146) + ($141))|0; //@line 1466 "load_abc.cpp"
   $148 = ((($136)) + 39|0); //@line 1466 "load_abc.cpp"
   $149 = HEAP8[$148>>0]|0; //@line 1466 "load_abc.cpp"
   $150 = ($149<<24>>24)!=(0); //@line 1466 "load_abc.cpp"
   $151 = $150 ? 0 : $145; //@line 1466 "load_abc.cpp"
   __ZL17abc_add_dronenoteP10_ABCHANDLEP9_ABCTRACKjii($136,$144,$147,$151); //@line 1466 "load_abc.cpp"
   $152 = Math_imul($133, $25)|0; //@line 3439 "load_abc.cpp"
   $153 = (($152) + ($rtime$2$ph))|0; //@line 3439 "load_abc.cpp"
   $154 = (($153>>>0) / ($i$0$lcssa$i>>>0))&-1; //@line 3439 "load_abc.cpp"
   $155 = (($154) + ($etime$1$6))|0; //@line 3439 "load_abc.cpp"
   $156 = HEAP8[$139>>0]|0; //@line 1174 "load_abc.cpp"
   $157 = $156&255; //@line 1174 "load_abc.cpp"
   $158 = (_pat_gmtosmp($157)|0); //@line 1174 "load_abc.cpp"
   $159 = $158&255; //@line 1174 "load_abc.cpp"
   $160 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
   HEAP8[$160>>0]=0&255;HEAP8[$160+1>>0]=(0>>8)&255;HEAP8[$160+2>>0]=(0>>16)&255;HEAP8[$160+3>>0]=0>>24; //@line 540 "load_abc.cpp"
   $161 = ((($160)) + 4|0); //@line 541 "load_abc.cpp"
   HEAP8[$161>>0]=$155&255;HEAP8[$161+1>>0]=($155>>8)&255;HEAP8[$161+2>>0]=($155>>16)&255;HEAP8[$161+3>>0]=$155>>24; //@line 541 "load_abc.cpp"
   $162 = ((($160)) + 8|0); //@line 543 "load_abc.cpp"
   HEAP8[$162>>0] = 0; //@line 543 "load_abc.cpp"
   $163 = ((($160)) + 9|0); //@line 543 "load_abc.cpp"
   HEAP8[$163>>0] = 0; //@line 543 "load_abc.cpp"
   $164 = ((($160)) + 10|0); //@line 543 "load_abc.cpp"
   HEAP8[$164>>0] = $159; //@line 543 "load_abc.cpp"
   $165 = ((($160)) + 11|0); //@line 543 "load_abc.cpp"
   HEAP8[$165>>0] = 0; //@line 543 "load_abc.cpp"
   $166 = ((($160)) + 12|0); //@line 543 "load_abc.cpp"
   HEAP8[$166>>0] = 0; //@line 543 "load_abc.cpp"
   $167 = ((($160)) + 13|0); //@line 543 "load_abc.cpp"
   HEAP8[$167>>0] = 0; //@line 543 "load_abc.cpp"
   $168 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
   $169 = ((($160)) + 14|0); //@line 544 "load_abc.cpp"
   HEAP8[$169>>0] = $168; //@line 544 "load_abc.cpp"
   $170 = ((($160)) + 15|0); //@line 545 "load_abc.cpp"
   HEAP8[$170>>0] = 0; //@line 545 "load_abc.cpp"
   __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($136,$160); //@line 1179 "load_abc.cpp"
   $$1159 = $$1$lcssa;$178 = $133;
  }
  else if ((label|0) == 19) {
   label = 0;
   $171 = $$lcssa145 | 1; //@line 3430 "load_abc.cpp"
   $172 = (((($h)) + 36|0) + ($171)|0); //@line 3430 "load_abc.cpp"
   $173 = HEAP8[$172>>0]|0; //@line 3430 "load_abc.cpp"
   $174 = $173 << 24 >> 24; //@line 3430 "load_abc.cpp"
   $175 = (($174) + -48)|0; //@line 3430 "load_abc.cpp"
   $$1159 = $$1$lcssa154;$178 = $175;
  }
  else if ((label|0) == 24) {
   label = 0;
   $221 = $$lcssa146 | 1; //@line 3430 "load_abc.cpp"
   $222 = (((($h)) + 36|0) + ($221)|0); //@line 3430 "load_abc.cpp"
   $223 = HEAP8[$222>>0]|0; //@line 3430 "load_abc.cpp"
   $224 = $223 << 24 >> 24; //@line 3430 "load_abc.cpp"
   $225 = (($224) + -48)|0; //@line 3430 "load_abc.cpp"
   $226 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 3451 "load_abc.cpp"
   $227 = ((($226)) + 44|0); //@line 3451 "load_abc.cpp"
   $228 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,$227,2)|0); //@line 3451 "load_abc.cpp"
   $229 = HEAPU8[$115>>0]|(HEAPU8[$115+1>>0]<<8)|(HEAPU8[$115+2>>0]<<16)|(HEAPU8[$115+3>>0]<<24); //@line 3452 "load_abc.cpp"
   $230 = $229&255; //@line 3452 "load_abc.cpp"
   $231 = ((($228)) + 42|0); //@line 3452 "load_abc.cpp"
   HEAP8[$231>>0] = $230; //@line 3452 "load_abc.cpp"
   $232 = HEAP8[$119>>0]|0; //@line 3453 "load_abc.cpp"
   $233 = $232&255; //@line 3453 "load_abc.cpp"
   $234 = (($rtime$2$ph>>>0) / ($i$0$lcssa$i>>>0))&-1; //@line 3454 "load_abc.cpp"
   $235 = (($234) + ($etime$1$6))|0; //@line 3454 "load_abc.cpp"
   $236 = HEAPU8[$118>>0]|(HEAPU8[$118+1>>0]<<8)|(HEAPU8[$118+2>>0]<<16)|(HEAPU8[$118+3>>0]<<24); //@line 3454 "load_abc.cpp"
   $237 = (($233) + 35)|0; //@line 1466 "load_abc.cpp"
   $238 = ((($228)) + 39|0); //@line 1466 "load_abc.cpp"
   $239 = HEAP8[$238>>0]|0; //@line 1466 "load_abc.cpp"
   $240 = ($239<<24>>24)!=(0); //@line 1466 "load_abc.cpp"
   $241 = $240 ? 0 : $236; //@line 1466 "load_abc.cpp"
   __ZL17abc_add_dronenoteP10_ABCHANDLEP9_ABCTRACKjii($228,$235,$237,$241); //@line 1466 "load_abc.cpp"
   $242 = Math_imul($225, $25)|0; //@line 3455 "load_abc.cpp"
   $243 = (($242) + ($rtime$2$ph))|0; //@line 3455 "load_abc.cpp"
   $244 = (($243>>>0) / ($i$0$lcssa$i>>>0))&-1; //@line 3456 "load_abc.cpp"
   $245 = (($244) + ($etime$1$6))|0; //@line 3456 "load_abc.cpp"
   $246 = HEAP8[$231>>0]|0; //@line 1174 "load_abc.cpp"
   $247 = $246&255; //@line 1174 "load_abc.cpp"
   $248 = (_pat_gmtosmp($247)|0); //@line 1174 "load_abc.cpp"
   $249 = $248&255; //@line 1174 "load_abc.cpp"
   $250 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
   HEAP8[$250>>0]=0&255;HEAP8[$250+1>>0]=(0>>8)&255;HEAP8[$250+2>>0]=(0>>16)&255;HEAP8[$250+3>>0]=0>>24; //@line 540 "load_abc.cpp"
   $251 = ((($250)) + 4|0); //@line 541 "load_abc.cpp"
   HEAP8[$251>>0]=$245&255;HEAP8[$251+1>>0]=($245>>8)&255;HEAP8[$251+2>>0]=($245>>16)&255;HEAP8[$251+3>>0]=$245>>24; //@line 541 "load_abc.cpp"
   $252 = ((($250)) + 8|0); //@line 543 "load_abc.cpp"
   HEAP8[$252>>0] = 0; //@line 543 "load_abc.cpp"
   $253 = ((($250)) + 9|0); //@line 543 "load_abc.cpp"
   HEAP8[$253>>0] = 0; //@line 543 "load_abc.cpp"
   $254 = ((($250)) + 10|0); //@line 543 "load_abc.cpp"
   HEAP8[$254>>0] = $249; //@line 543 "load_abc.cpp"
   $255 = ((($250)) + 11|0); //@line 543 "load_abc.cpp"
   HEAP8[$255>>0] = 0; //@line 543 "load_abc.cpp"
   $256 = ((($250)) + 12|0); //@line 543 "load_abc.cpp"
   HEAP8[$256>>0] = 0; //@line 543 "load_abc.cpp"
   $257 = ((($250)) + 13|0); //@line 543 "load_abc.cpp"
   HEAP8[$257>>0] = 0; //@line 543 "load_abc.cpp"
   $258 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
   $259 = ((($250)) + 14|0); //@line 544 "load_abc.cpp"
   HEAP8[$259>>0] = $258; //@line 544 "load_abc.cpp"
   $260 = ((($250)) + 15|0); //@line 545 "load_abc.cpp"
   HEAP8[$260>>0] = 0; //@line 545 "load_abc.cpp"
   __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($228,$250); //@line 1179 "load_abc.cpp"
   $g$1$ph = $$1$lcssa155;$rtime$2$ph = $243;
   continue;
  }
  else if ((label|0) == 25) {
   label = 0;
   $261 = $$lcssa147 | 1; //@line 3430 "load_abc.cpp"
   $262 = (((($h)) + 36|0) + ($261)|0); //@line 3430 "load_abc.cpp"
   $263 = HEAP8[$262>>0]|0; //@line 3430 "load_abc.cpp"
   $264 = $263 << 24 >> 24; //@line 3430 "load_abc.cpp"
   $265 = (($264) + -48)|0; //@line 3430 "load_abc.cpp"
   $266 = (_toupper($$lcssa152)|0); //@line 3466 "load_abc.cpp"
   $267 = (($266) + -71)|0; //@line 3466 "load_abc.cpp"
   $268 = HEAP32[$114>>2]|0; //@line 3468 "load_abc.cpp"
   $269 = ($267|0)<($268|0); //@line 3468 "load_abc.cpp"
   if ($269) {
    $270 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 3469 "load_abc.cpp"
    $271 = ((($270)) + 44|0); //@line 3469 "load_abc.cpp"
    $272 = (($266) + -68)|0; //@line 3469 "load_abc.cpp"
    $273 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,$271,$272)|0); //@line 3469 "load_abc.cpp"
    $274 = HEAPU8[$120>>0]|(HEAPU8[$120+1>>0]<<8)|(HEAPU8[$120+2>>0]<<16)|(HEAPU8[$120+3>>0]<<24); //@line 3470 "load_abc.cpp"
    $275 = $274&255; //@line 3470 "load_abc.cpp"
    $276 = ((($273)) + 42|0); //@line 3470 "load_abc.cpp"
    HEAP8[$276>>0] = $275; //@line 3470 "load_abc.cpp"
    $277 = HEAP8[$116>>0]|0; //@line 3471 "load_abc.cpp"
    $278 = $277&255; //@line 3471 "load_abc.cpp"
    $279 = ((1199780 + (($14*24)|0)|0) + ($267<<2)|0); //@line 3471 "load_abc.cpp"
    $280 = HEAP32[$279>>2]|0; //@line 3471 "load_abc.cpp"
    $281 = (($278) + ($280))|0; //@line 3471 "load_abc.cpp"
    $282 = (_isupper($$lcssa152)|0); //@line 3472 "load_abc.cpp"
    $283 = ($282|0)==(0); //@line 3472 "load_abc.cpp"
    $$2$v = $283 ? 24 : 12; //@line 3472 "load_abc.cpp"
    $$2 = (($281) + ($$2$v))|0; //@line 3472 "load_abc.cpp"
    $284 = (($rtime$2$ph>>>0) / ($i$0$lcssa$i>>>0))&-1; //@line 3473 "load_abc.cpp"
    $285 = (($284) + ($etime$1$6))|0; //@line 3473 "load_abc.cpp"
    $286 = HEAPU8[$121>>0]|(HEAPU8[$121+1>>0]<<8)|(HEAPU8[$121+2>>0]<<16)|(HEAPU8[$121+3>>0]<<24); //@line 3473 "load_abc.cpp"
    $287 = (($$2) + 23)|0; //@line 1466 "load_abc.cpp"
    $288 = ((($273)) + 39|0); //@line 1466 "load_abc.cpp"
    $289 = HEAP8[$288>>0]|0; //@line 1466 "load_abc.cpp"
    $290 = ($289<<24>>24)!=(0); //@line 1466 "load_abc.cpp"
    $291 = $290 ? 0 : $286; //@line 1466 "load_abc.cpp"
    __ZL17abc_add_dronenoteP10_ABCHANDLEP9_ABCTRACKjii($273,$285,$287,$291); //@line 1466 "load_abc.cpp"
    $292 = Math_imul($265, $25)|0; //@line 3475 "load_abc.cpp"
    $293 = (($292) + ($rtime$2$ph))|0; //@line 3475 "load_abc.cpp"
    $294 = ($$2|0)==(0); //@line 3476 "load_abc.cpp"
    if ($294) {
     $g$1$ph = $$1$lcssa156;$rtime$2$ph = $293;
     continue;
    }
    $295 = (($293>>>0) / ($i$0$lcssa$i>>>0))&-1; //@line 3476 "load_abc.cpp"
    $296 = (($295) + ($etime$1$6))|0; //@line 3476 "load_abc.cpp"
    $297 = HEAP8[$276>>0]|0; //@line 1174 "load_abc.cpp"
    $298 = $297&255; //@line 1174 "load_abc.cpp"
    $299 = (_pat_gmtosmp($298)|0); //@line 1174 "load_abc.cpp"
    $300 = $299&255; //@line 1174 "load_abc.cpp"
    $301 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
    HEAP8[$301>>0]=0&255;HEAP8[$301+1>>0]=(0>>8)&255;HEAP8[$301+2>>0]=(0>>16)&255;HEAP8[$301+3>>0]=0>>24; //@line 540 "load_abc.cpp"
    $302 = ((($301)) + 4|0); //@line 541 "load_abc.cpp"
    HEAP8[$302>>0]=$296&255;HEAP8[$302+1>>0]=($296>>8)&255;HEAP8[$302+2>>0]=($296>>16)&255;HEAP8[$302+3>>0]=$296>>24; //@line 541 "load_abc.cpp"
    $303 = ((($301)) + 8|0); //@line 543 "load_abc.cpp"
    HEAP8[$303>>0] = 0; //@line 543 "load_abc.cpp"
    $304 = ((($301)) + 9|0); //@line 543 "load_abc.cpp"
    HEAP8[$304>>0] = 0; //@line 543 "load_abc.cpp"
    $305 = ((($301)) + 10|0); //@line 543 "load_abc.cpp"
    HEAP8[$305>>0] = $300; //@line 543 "load_abc.cpp"
    $306 = ((($301)) + 11|0); //@line 543 "load_abc.cpp"
    HEAP8[$306>>0] = 0; //@line 543 "load_abc.cpp"
    $307 = ((($301)) + 12|0); //@line 543 "load_abc.cpp"
    HEAP8[$307>>0] = 0; //@line 543 "load_abc.cpp"
    $308 = ((($301)) + 13|0); //@line 543 "load_abc.cpp"
    HEAP8[$308>>0] = 0; //@line 543 "load_abc.cpp"
    $309 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
    $310 = ((($301)) + 14|0); //@line 544 "load_abc.cpp"
    HEAP8[$310>>0] = $309; //@line 544 "load_abc.cpp"
    $311 = ((($301)) + 15|0); //@line 545 "load_abc.cpp"
    HEAP8[$311>>0] = 0; //@line 545 "load_abc.cpp"
    __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($273,$301); //@line 1179 "load_abc.cpp"
    $g$1$ph = $$1$lcssa156;$rtime$2$ph = $293;
    continue;
   } else {
    $$1160 = $$1$lcssa156;$$pn64 = $265;
    label = 29;
   }
  }
  else if ((label|0) == 28) {
   label = 0;
   $312 = $$lcssa148 | 1; //@line 3430 "load_abc.cpp"
   $313 = (((($h)) + 36|0) + ($312)|0); //@line 3430 "load_abc.cpp"
   $314 = HEAP8[$313>>0]|0; //@line 3430 "load_abc.cpp"
   $315 = $314 << 24 >> 24; //@line 3430 "load_abc.cpp"
   $316 = (($315) + -48)|0; //@line 3430 "load_abc.cpp"
   $$1160 = $$1$lcssa157;$$pn64 = $316;
   label = 29;
  }
  if ((label|0) == 29) {
   label = 0;
   $$pn = Math_imul($$pn64, $25)|0; //@line 3475 "load_abc.cpp"
   $rtime$2$ph$be = (($$pn) + ($rtime$2$ph))|0; //@line 3475 "load_abc.cpp"
   $g$1$ph = $$1160;$rtime$2$ph = $rtime$2$ph$be;
   continue;
  }
  $176 = HEAP32[$114>>2]|0; //@line 3441 "load_abc.cpp"
  $177 = ($176|0)>(1); //@line 3441 "load_abc.cpp"
  if (!($177)) {
   $$pre = Math_imul($178, $25)|0; //@line 3448 "load_abc.cpp"
   $$pre89 = (($$pre) + ($rtime$2$ph))|0; //@line 3448 "load_abc.cpp"
   $g$1$ph = $$1159;$rtime$2$ph = $$pre89;
   continue;
  }
  $179 = (($rtime$2$ph>>>0) / ($i$0$lcssa$i>>>0))&-1; //@line 3445 "load_abc.cpp"
  $180 = (($179) + ($etime$1$6))|0; //@line 3445 "load_abc.cpp"
  $181 = Math_imul($178, $25)|0; //@line 3446 "load_abc.cpp"
  $182 = (($181) + ($rtime$2$ph))|0; //@line 3446 "load_abc.cpp"
  $183 = (($182>>>0) / ($i$0$lcssa$i>>>0))&-1; //@line 3446 "load_abc.cpp"
  $184 = (($183) + ($etime$1$6))|0; //@line 3446 "load_abc.cpp"
  $i$148 = 1;
  while(1) {
   $185 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 3442 "load_abc.cpp"
   $186 = ((($185)) + 44|0); //@line 3442 "load_abc.cpp"
   $187 = (($i$148) + 2)|0; //@line 3442 "load_abc.cpp"
   $188 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,$186,$187)|0); //@line 3442 "load_abc.cpp"
   $189 = HEAPU8[$120>>0]|(HEAPU8[$120+1>>0]<<8)|(HEAPU8[$120+2>>0]<<16)|(HEAPU8[$120+3>>0]<<24); //@line 3443 "load_abc.cpp"
   $190 = $189&255; //@line 3443 "load_abc.cpp"
   $191 = ((($188)) + 42|0); //@line 3443 "load_abc.cpp"
   HEAP8[$191>>0] = $190; //@line 3443 "load_abc.cpp"
   $192 = HEAP8[$116>>0]|0; //@line 3444 "load_abc.cpp"
   $193 = $192&255; //@line 3444 "load_abc.cpp"
   $194 = ((1199780 + (($14*24)|0)|0) + ($i$148<<2)|0); //@line 3444 "load_abc.cpp"
   $195 = HEAP32[$194>>2]|0; //@line 3444 "load_abc.cpp"
   $196 = HEAPU8[$121>>0]|(HEAPU8[$121+1>>0]<<8)|(HEAPU8[$121+2>>0]<<16)|(HEAPU8[$121+3>>0]<<24); //@line 3445 "load_abc.cpp"
   $197 = (($195) + 47)|0; //@line 3444 "load_abc.cpp"
   $198 = (($197) + ($193))|0; //@line 1466 "load_abc.cpp"
   $199 = ((($188)) + 39|0); //@line 1466 "load_abc.cpp"
   $200 = HEAP8[$199>>0]|0; //@line 1466 "load_abc.cpp"
   $201 = ($200<<24>>24)!=(0); //@line 1466 "load_abc.cpp"
   $202 = $201 ? 0 : $196; //@line 1466 "load_abc.cpp"
   __ZL17abc_add_dronenoteP10_ABCHANDLEP9_ABCTRACKjii($188,$180,$198,$202); //@line 1466 "load_abc.cpp"
   $203 = HEAP8[$191>>0]|0; //@line 1174 "load_abc.cpp"
   $204 = $203&255; //@line 1174 "load_abc.cpp"
   $205 = (_pat_gmtosmp($204)|0); //@line 1174 "load_abc.cpp"
   $206 = $205&255; //@line 1174 "load_abc.cpp"
   $207 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
   HEAP8[$207>>0]=0&255;HEAP8[$207+1>>0]=(0>>8)&255;HEAP8[$207+2>>0]=(0>>16)&255;HEAP8[$207+3>>0]=0>>24; //@line 540 "load_abc.cpp"
   $208 = ((($207)) + 4|0); //@line 541 "load_abc.cpp"
   HEAP8[$208>>0]=$184&255;HEAP8[$208+1>>0]=($184>>8)&255;HEAP8[$208+2>>0]=($184>>16)&255;HEAP8[$208+3>>0]=$184>>24; //@line 541 "load_abc.cpp"
   $209 = ((($207)) + 8|0); //@line 543 "load_abc.cpp"
   HEAP8[$209>>0] = 0; //@line 543 "load_abc.cpp"
   $210 = ((($207)) + 9|0); //@line 543 "load_abc.cpp"
   HEAP8[$210>>0] = 0; //@line 543 "load_abc.cpp"
   $211 = ((($207)) + 10|0); //@line 543 "load_abc.cpp"
   HEAP8[$211>>0] = $206; //@line 543 "load_abc.cpp"
   $212 = ((($207)) + 11|0); //@line 543 "load_abc.cpp"
   HEAP8[$212>>0] = 0; //@line 543 "load_abc.cpp"
   $213 = ((($207)) + 12|0); //@line 543 "load_abc.cpp"
   HEAP8[$213>>0] = 0; //@line 543 "load_abc.cpp"
   $214 = ((($207)) + 13|0); //@line 543 "load_abc.cpp"
   HEAP8[$214>>0] = 0; //@line 543 "load_abc.cpp"
   $215 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
   $216 = ((($207)) + 14|0); //@line 544 "load_abc.cpp"
   HEAP8[$216>>0] = $215; //@line 544 "load_abc.cpp"
   $217 = ((($207)) + 15|0); //@line 545 "load_abc.cpp"
   HEAP8[$217>>0] = 0; //@line 545 "load_abc.cpp"
   __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($188,$207); //@line 1179 "load_abc.cpp"
   $218 = (($i$148) + 1)|0; //@line 3441 "load_abc.cpp"
   $219 = HEAP32[$114>>2]|0; //@line 3441 "load_abc.cpp"
   $220 = ($218|0)<($219|0); //@line 3441 "load_abc.cpp"
   if ($220) {
    $i$148 = $218;
   } else {
    $g$1$ph = $$1159;$rtime$2$ph = $182;
    continue L27;
   }
  }
 }
 return; //@line 3483 "load_abc.cpp"
}
function __ZL12abc_add_drumP10_ABCHANDLEjj($h,$tracktime,$bartime) {
 $h = $h|0;
 $tracktime = $tracktime|0;
 $bartime = $bartime|0;
 var $$ = 0, $$1 = 0, $$i$0$i = 0, $$pre = 0, $$pre$phi9Z2D = 0, $$pre8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $g$0$lcssa = 0, $g$04 = 0, $g$12 = 0, $i$0$lcssa$i = 0, $i$02$i = 0, $isdigit$i = 0, $isdigittmp$i = 0, $p$01$i = 0, $rtime$0 = 0, $rtime$15 = 0, $rtime$23 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($h)) + 116|0); //@line 3347 "load_abc.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 3335 "load_abc.cpp"
 $2 = ($1<<24>>24)==(0); //@line 3335 "load_abc.cpp"
 if ($2) {
  $i$0$lcssa$i = 0;
 } else {
  $4 = $1;$i$02$i = 0;$p$01$i = $0;
  while(1) {
   $3 = $4 << 24 >> 24; //@line 3336 "load_abc.cpp"
   $isdigittmp$i = (($3) + -48)|0; //@line 3336 "load_abc.cpp"
   $isdigit$i = ($isdigittmp$i>>>0)<(10); //@line 3336 "load_abc.cpp"
   $5 = $isdigit$i ? $isdigittmp$i : 0; //@line 3336 "load_abc.cpp"
   $$i$0$i = (($5) + ($i$02$i))|0; //@line 3336 "load_abc.cpp"
   $6 = ((($p$01$i)) + 1|0); //@line 3335 "load_abc.cpp"
   $7 = HEAP8[$6>>0]|0; //@line 3335 "load_abc.cpp"
   $8 = ($7<<24>>24)==(0); //@line 3335 "load_abc.cpp"
   if ($8) {
    $i$0$lcssa$i = $$i$0$i;
    break;
   } else {
    $4 = $7;$i$02$i = $$i$0$i;$p$01$i = $6;
   }
  }
 }
 $9 = ((($h)) + 356|0); //@line 3348 "load_abc.cpp"
 $10 = HEAPU8[$9>>0]|(HEAPU8[$9+1>>0]<<8)|(HEAPU8[$9+2>>0]<<16)|(HEAPU8[$9+3>>0]<<24); //@line 3348 "load_abc.cpp"
 $11 = ((($h)) + 420|0); //@line 3350 "load_abc.cpp"
 $12 = HEAPU8[$11>>0]|(HEAPU8[$11+1>>0]<<8)|(HEAPU8[$11+2>>0]<<16)|(HEAPU8[$11+3>>0]<<24); //@line 3350 "load_abc.cpp"
 $13 = ((($12)) + 44|0); //@line 3350 "load_abc.cpp"
 $14 = (__ZL16abc_locate_trackP10_ABCHANDLEPKci($h,$13,8)|0); //@line 3350 "load_abc.cpp"
 $15 = ((($14)) + 8|0); //@line 3351 "load_abc.cpp"
 $16 = HEAPU8[$15>>0]|(HEAPU8[$15+1>>0]<<8)|(HEAPU8[$15+2>>0]<<16)|(HEAPU8[$15+3>>0]<<24); //@line 3351 "load_abc.cpp"
 $17 = ($16|0)==(0|0); //@line 3352 "load_abc.cpp"
 if ($17) {
  $21 = $bartime;
 } else {
  $18 = ((($16)) + 4|0); //@line 3352 "load_abc.cpp"
  $19 = HEAPU8[$18>>0]|(HEAPU8[$18+1>>0]<<8)|(HEAPU8[$18+2>>0]<<16)|(HEAPU8[$18+3>>0]<<24); //@line 3352 "load_abc.cpp"
  $21 = $19;
 }
 $20 = ($21>>>0)>($tracktime>>>0); //@line 3353 "load_abc.cpp"
 if ($20) {
  return; //@line 3379 "load_abc.cpp"
 }
 $22 = ($21>>>0)<($bartime>>>0); //@line 3354 "load_abc.cpp"
 if ($22) {
  $23 = HEAPU8[$9>>0]|(HEAPU8[$9+1>>0]<<8)|(HEAPU8[$9+2>>0]<<16)|(HEAPU8[$9+3>>0]<<24); //@line 3354 "load_abc.cpp"
  $24 = (($bartime) - ($21))|0; //@line 3354 "load_abc.cpp"
  $25 = (($24>>>0) % ($23>>>0))&-1; //@line 3354 "load_abc.cpp"
  $26 = (($23) - ($25))|0; //@line 3354 "load_abc.cpp"
  $rtime$0 = $26;
 } else {
  $27 = (($21) - ($bartime))|0; //@line 3355 "load_abc.cpp"
  $28 = HEAPU8[$9>>0]|(HEAPU8[$9+1>>0]<<8)|(HEAPU8[$9+2>>0]<<16)|(HEAPU8[$9+3>>0]<<24); //@line 3355 "load_abc.cpp"
  $29 = (($27>>>0) % ($28>>>0))&-1; //@line 3355 "load_abc.cpp"
  $rtime$0 = $29;
 }
 $30 = Math_imul($10, $i$0$lcssa$i)|0; //@line 3356 "load_abc.cpp"
 $31 = Math_imul($rtime$0, $i$0$lcssa$i)|0; //@line 3357 "load_abc.cpp"
 $32 = (($31) + ($30))|0; //@line 3358 "load_abc.cpp"
 $33 = (_strlen($0)|0); //@line 3359 "load_abc.cpp"
 $34 = $33 >>> 1; //@line 3359 "load_abc.cpp"
 $35 = ($32>>>0)>($30>>>0); //@line 3361 "load_abc.cpp"
 if ($35) {
  $g$04 = 0;$rtime$15 = $32;
  while(1) {
   $36 = $g$04 << 1; //@line 3362 "load_abc.cpp"
   $37 = $36 | 1; //@line 3362 "load_abc.cpp"
   $38 = (((($h)) + 116|0) + ($37)|0); //@line 3362 "load_abc.cpp"
   $39 = HEAP8[$38>>0]|0; //@line 3362 "load_abc.cpp"
   $40 = $39 << 24 >> 24; //@line 3362 "load_abc.cpp"
   $41 = (($40) + -48)|0; //@line 3362 "load_abc.cpp"
   $42 = Math_imul($41, $10)|0; //@line 3362 "load_abc.cpp"
   $43 = (($rtime$15) - ($42))|0; //@line 3362 "load_abc.cpp"
   $44 = (($g$04) + 1)|0; //@line 3363 "load_abc.cpp"
   $45 = ($44|0)==($34|0); //@line 3363 "load_abc.cpp"
   $$ = $45 ? 0 : $44; //@line 3363 "load_abc.cpp"
   $46 = ($43>>>0)>($30>>>0); //@line 3361 "load_abc.cpp"
   if ($46) {
    $g$04 = $$;$rtime$15 = $43;
   } else {
    $g$0$lcssa = $$;
    break;
   }
  }
 } else {
  $g$0$lcssa = 0;
 }
 $47 = (($tracktime) - ($21))|0; //@line 3365 "load_abc.cpp"
 $48 = Math_imul($47, $i$0$lcssa$i)|0; //@line 3365 "load_abc.cpp"
 $49 = ($48|0)==(0); //@line 3367 "load_abc.cpp"
 if ($49) {
  return; //@line 3379 "load_abc.cpp"
 }
 $50 = ((($14)) + 42|0); //@line 3371 "load_abc.cpp"
 $51 = ((($14)) + 39|0); //@line 1471 "load_abc.cpp"
 $g$12 = $g$0$lcssa;$rtime$23 = 0;
 while(1) {
  $52 = $g$12 << 1; //@line 3368 "load_abc.cpp"
  $53 = (((($h)) + 116|0) + ($52)|0); //@line 3368 "load_abc.cpp"
  $54 = HEAP8[$53>>0]|0; //@line 3368 "load_abc.cpp"
  $55 = $52 | 1; //@line 3369 "load_abc.cpp"
  $56 = (((($h)) + 116|0) + ($55)|0); //@line 3369 "load_abc.cpp"
  $57 = HEAP8[$56>>0]|0; //@line 3369 "load_abc.cpp"
  $58 = $57 << 24 >> 24; //@line 3369 "load_abc.cpp"
  $59 = (($58) + -48)|0; //@line 3369 "load_abc.cpp"
  $60 = ($54<<24>>24)==(100); //@line 3370 "load_abc.cpp"
  if ($60) {
   $61 = (((($h)) + 196|0) + ($g$12)|0); //@line 3371 "load_abc.cpp"
   $62 = HEAP8[$61>>0]|0; //@line 3371 "load_abc.cpp"
   $63 = $62 << 24 >> 24; //@line 3371 "load_abc.cpp"
   $64 = (($63) + -1)|0; //@line 3371 "load_abc.cpp"
   $65 = (_pat_gm_drumnr($64)|0); //@line 3371 "load_abc.cpp"
   $66 = $65&255; //@line 3371 "load_abc.cpp"
   HEAP8[$50>>0] = $66; //@line 3371 "load_abc.cpp"
   $67 = HEAP8[$61>>0]|0; //@line 3372 "load_abc.cpp"
   $68 = $67 << 24 >> 24; //@line 3372 "load_abc.cpp"
   $69 = (_pat_gm_drumnote($68)|0); //@line 3372 "load_abc.cpp"
   $70 = (($rtime$23>>>0) / ($i$0$lcssa$i>>>0))&-1; //@line 3373 "load_abc.cpp"
   $71 = (($70) + ($21))|0; //@line 3373 "load_abc.cpp"
   $72 = (((($h)) + 276|0) + ($g$12)|0); //@line 3373 "load_abc.cpp"
   $73 = HEAP8[$72>>0]|0; //@line 3373 "load_abc.cpp"
   $74 = $73 << 24 >> 24; //@line 3373 "load_abc.cpp"
   $75 = HEAP8[$51>>0]|0; //@line 1471 "load_abc.cpp"
   $76 = ($75<<24>>24)!=(0); //@line 1471 "load_abc.cpp"
   $77 = $76 ? 0 : $74; //@line 1471 "load_abc.cpp"
   __ZL17abc_add_dronenoteP10_ABCHANDLEP9_ABCTRACKjii($14,$71,$69,$77); //@line 1471 "load_abc.cpp"
   $78 = Math_imul($59, $10)|0; //@line 3374 "load_abc.cpp"
   $79 = (($78) + ($rtime$23))|0; //@line 3374 "load_abc.cpp"
   $80 = (($79>>>0) / ($i$0$lcssa$i>>>0))&-1; //@line 3374 "load_abc.cpp"
   $81 = (($80) + ($21))|0; //@line 3374 "load_abc.cpp"
   $82 = HEAP8[$50>>0]|0; //@line 1174 "load_abc.cpp"
   $83 = $82&255; //@line 1174 "load_abc.cpp"
   $84 = (_pat_gmtosmp($83)|0); //@line 1174 "load_abc.cpp"
   $85 = $84&255; //@line 1174 "load_abc.cpp"
   $86 = (_calloc(1,16)|0); //@line 539 "load_abc.cpp"
   HEAP8[$86>>0]=0&255;HEAP8[$86+1>>0]=(0>>8)&255;HEAP8[$86+2>>0]=(0>>16)&255;HEAP8[$86+3>>0]=0>>24; //@line 540 "load_abc.cpp"
   $87 = ((($86)) + 4|0); //@line 541 "load_abc.cpp"
   HEAP8[$87>>0]=$81&255;HEAP8[$87+1>>0]=($81>>8)&255;HEAP8[$87+2>>0]=($81>>16)&255;HEAP8[$87+3>>0]=$81>>24; //@line 541 "load_abc.cpp"
   $88 = ((($86)) + 8|0); //@line 543 "load_abc.cpp"
   HEAP8[$88>>0] = 0; //@line 543 "load_abc.cpp"
   $89 = ((($86)) + 9|0); //@line 543 "load_abc.cpp"
   HEAP8[$89>>0] = 0; //@line 543 "load_abc.cpp"
   $90 = ((($86)) + 10|0); //@line 543 "load_abc.cpp"
   HEAP8[$90>>0] = $85; //@line 543 "load_abc.cpp"
   $91 = ((($86)) + 11|0); //@line 543 "load_abc.cpp"
   HEAP8[$91>>0] = 0; //@line 543 "load_abc.cpp"
   $92 = ((($86)) + 12|0); //@line 543 "load_abc.cpp"
   HEAP8[$92>>0] = 0; //@line 543 "load_abc.cpp"
   $93 = ((($86)) + 13|0); //@line 543 "load_abc.cpp"
   HEAP8[$93>>0] = 0; //@line 543 "load_abc.cpp"
   $94 = HEAP8[1278234]|0; //@line 544 "load_abc.cpp"
   $95 = ((($86)) + 14|0); //@line 544 "load_abc.cpp"
   HEAP8[$95>>0] = $94; //@line 544 "load_abc.cpp"
   $96 = ((($86)) + 15|0); //@line 545 "load_abc.cpp"
   HEAP8[$96>>0] = 0; //@line 545 "load_abc.cpp"
   __ZL13abc_add_eventP10_ABCHANDLEP9_ABCTRACKP9_ABCEVENT($14,$86); //@line 1179 "load_abc.cpp"
   $$pre$phi9Z2D = $79;
  } else {
   $$pre = Math_imul($59, $10)|0; //@line 3377 "load_abc.cpp"
   $$pre8 = (($$pre) + ($rtime$23))|0; //@line 3377 "load_abc.cpp"
   $$pre$phi9Z2D = $$pre8;
  }
  $97 = (($g$12) + 1)|0; //@line 3376 "load_abc.cpp"
  $98 = ($97|0)==($34|0); //@line 3376 "load_abc.cpp"
  $$1 = $98 ? 0 : $97; //@line 3376 "load_abc.cpp"
  $99 = ($$pre$phi9Z2D>>>0)<($48>>>0); //@line 3367 "load_abc.cpp"
  if ($99) {
   $g$12 = $$1;$rtime$23 = $$pre$phi9Z2D;
  } else {
   break;
  }
 }
 return; //@line 3379 "load_abc.cpp"
}
function _pat_gm_drumnr($n) {
 $n = $n|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)<(25); //@line 247 "load_pat.cpp"
 if ($0) {
  $$0 = 129;
  return ($$0|0); //@line 251 "load_pat.cpp"
 }
 $1 = (($n) + 104)|0; //@line 248 "load_pat.cpp"
 $2 = ($1|0)<(191); //@line 248 "load_pat.cpp"
 $$ = $2 ? $1 : 191; //@line 249 "load_pat.cpp"
 $$0 = $$;
 return ($$0|0); //@line 251 "load_pat.cpp"
}
function _pat_gm_drumnote($n) {
 $n = $n|0;
 var $$0 = 0, $$0$i = 0, $$op = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)<(25); //@line 247 "load_pat.cpp"
 if ($0) {
  $$0$i = 128;
 } else {
  $1 = (($n) + 104)|0; //@line 248 "load_pat.cpp"
  $2 = ($1|0)<(191); //@line 248 "load_pat.cpp"
  $$op = (($n) + 103)|0; //@line 249 "load_pat.cpp"
  $phitmp = $2 ? $$op : 190; //@line 249 "load_pat.cpp"
  $$0$i = $phitmp;
 }
 $3 = (1253146 + ($$0$i<<7)|0); //@line 256 "load_pat.cpp"
 $4 = (_strchr($3,58)|0); //@line 256 "load_pat.cpp"
 $5 = ($4|0)==(0|0); //@line 257 "load_pat.cpp"
 if ($5) {
  $$0 = $n;
  return ($$0|0); //@line 259 "load_pat.cpp"
 }
 $6 = ((($4)) + 1|0); //@line 257 "load_pat.cpp"
 $7 = (_strstr($6,527326)|0); //@line 698 "load_pat.cpp"
 $8 = ($7|0)==(0|0); //@line 699 "load_pat.cpp"
 if ($8) {
  $$0 = $n;
  return ($$0|0); //@line 259 "load_pat.cpp"
 }
 $9 = (_strchr($7,61)|0); //@line 700 "load_pat.cpp"
 $10 = ((($9)) + 1|0); //@line 700 "load_pat.cpp"
 $11 = (_atoi($10)|0); //@line 700 "load_pat.cpp"
 $$0 = $11;
 return ($$0|0); //@line 259 "load_pat.cpp"
}
function __ZL21abc_parse_decorationsP10_ABCHANDLEP9_ABCTRACKPKc($h,$tp,$p) {
 $h = $h|0;
 $tp = $tp|0;
 $p = $p|0;
 var $$ = 0, $$016 = 0, $$1 = 0, $$27 = 0, $$lcssa = 0, $$lcssa21 = 0, $$lcssa22 = 0, $$pre = 0, $$pre$phiZ2D = 0, $$pre13 = 0, $$pre15 = 0, $$vol$1 = 0, $$vol$3 = 0, $$vol$5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vol$1 = 0, $vol$318 = 0, $vol$4 = 0, $vol$58 = 0, $vol$7 = 0, $vol$73 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strncmp($p,527566,2)|0); //@line 2303 "load_abc.cpp"
 $1 = ($0|0)==(0); //@line 2303 "load_abc.cpp"
 $$ = $1 ? 75 : 0; //@line 2303 "load_abc.cpp"
 $2 = (_strncmp($p,527569,2)|0); //@line 2304 "load_abc.cpp"
 $3 = ($2|0)==(0); //@line 2304 "load_abc.cpp"
 $vol$1 = $3 ? 90 : $$; //@line 2304 "load_abc.cpp"
 $4 = (_strncmp($p,527572,3)|0); //@line 2305 "load_abc.cpp"
 $5 = ($4|0)==(0); //@line 2305 "load_abc.cpp"
 $$vol$1 = $5 ? 100 : $vol$1; //@line 2305 "load_abc.cpp"
 $6 = HEAP8[$p>>0]|0; //@line 2306 "load_abc.cpp"
 $7 = ($6<<24>>24)==(112); //@line 2306 "load_abc.cpp"
 if ($7) {
  $8 = ((($p)) + 1|0); //@line 2308 "load_abc.cpp"
  $10 = $8;$vol$318 = 60;
  while(1) {
   $9 = (($vol$318) + -15)|0; //@line 2308 "load_abc.cpp"
   $$pre = HEAP8[$10>>0]|0; //@line 2308 "load_abc.cpp"
   $11 = ((($10)) + 1|0); //@line 2308 "load_abc.cpp"
   $12 = ($$pre<<24>>24)==(112); //@line 2308 "load_abc.cpp"
   if ($12) {
    $10 = $11;$vol$318 = $9;
   } else {
    $$lcssa21 = $9;$$lcssa22 = $11;
    break;
   }
  }
  $13 = ($$lcssa21|0)<(1); //@line 2309 "load_abc.cpp"
  $$vol$3 = $13 ? 1 : $$lcssa21; //@line 2309 "load_abc.cpp"
  $$pre13 = HEAP8[$$lcssa22>>0]|0; //@line 2311 "load_abc.cpp"
  $$1 = $$lcssa22;$15 = $$pre13;$vol$4 = $$vol$3;
 } else {
  $$1 = $p;$15 = $6;$vol$4 = $$vol$1;
 }
 $14 = ($15<<24>>24)==(102); //@line 2311 "load_abc.cpp"
 if ($14) {
  $$27 = $$1;$vol$58 = 105;
  while(1) {
   $16 = ((($$27)) + 1|0); //@line 2313 "load_abc.cpp"
   $17 = (($vol$58) + 15)|0; //@line 2313 "load_abc.cpp"
   $18 = HEAP8[$16>>0]|0; //@line 2313 "load_abc.cpp"
   $19 = ($18<<24>>24)==(102); //@line 2313 "load_abc.cpp"
   if ($19) {
    $$27 = $16;$vol$58 = $17;
   } else {
    $$lcssa = $17;
    break;
   }
  }
  $20 = ($$lcssa|0)>(135); //@line 2314 "load_abc.cpp"
  $$vol$5 = $20 ? 127 : $$lcssa; //@line 2314 "load_abc.cpp"
  $21 = ($$vol$5|0)>(127); //@line 2315 "load_abc.cpp"
  if ($21) {
   $vol$73 = 125;
  } else {
   $vol$7 = $$vol$5;
   label = 8;
  }
 } else {
  $vol$7 = $vol$4;
  label = 8;
 }
 if ((label|0) == 8) {
  $22 = ($vol$7|0)==(0); //@line 2317 "load_abc.cpp"
  if ($22) {
   $$pre15 = ((($tp)) + 41|0); //@line 2326 "load_abc.cpp"
   $$pre$phiZ2D = $$pre15;
   $35 = HEAP8[$$pre$phiZ2D>>0]|0; //@line 2326 "load_abc.cpp"
   $36 = $35&255; //@line 2326 "load_abc.cpp"
   return ($36|0); //@line 2326 "load_abc.cpp"
  } else {
   $vol$73 = $vol$7;
  }
 }
 $23 = $vol$73&255; //@line 2318 "load_abc.cpp"
 $24 = ((($tp)) + 41|0); //@line 2318 "load_abc.cpp"
 HEAP8[$24>>0] = $23; //@line 2318 "load_abc.cpp"
 $25 = ((($h)) + 8|0); //@line 2319 "load_abc.cpp"
 $26 = HEAPU8[$25>>0]|(HEAPU8[$25+1>>0]<<8)|(HEAPU8[$25+2>>0]<<16)|(HEAPU8[$25+3>>0]<<24); //@line 2319 "load_abc.cpp"
 $27 = ($26|0)==($tp|0); //@line 2319 "load_abc.cpp"
 if ($27) {
  $$016 = $tp;
 } else {
  $$pre$phiZ2D = $24;
  $35 = HEAP8[$$pre$phiZ2D>>0]|0; //@line 2326 "load_abc.cpp"
  $36 = $35&255; //@line 2326 "load_abc.cpp"
  return ($36|0); //@line 2326 "load_abc.cpp"
 }
 while(1) {
  $28 = ((($$016)) + 37|0); //@line 2321 "load_abc.cpp"
  $29 = HEAP8[$28>>0]|0; //@line 2321 "load_abc.cpp"
  $30 = (($29) + -1)<<24>>24; //@line 2321 "load_abc.cpp"
  $31 = ($30&255)>(9); //@line 2321 "load_abc.cpp"
  if ($31) {
   $32 = ((($$016)) + 41|0); //@line 2321 "load_abc.cpp"
   HEAP8[$32>>0] = $23; //@line 2321 "load_abc.cpp"
  }
  $33 = HEAPU8[$$016>>0]|(HEAPU8[$$016+1>>0]<<8)|(HEAPU8[$$016+2>>0]<<16)|(HEAPU8[$$016+3>>0]<<24); //@line 2320 "load_abc.cpp"
  $34 = ($33|0)==(0|0); //@line 2320 "load_abc.cpp"
  if ($34) {
   $$pre$phiZ2D = $24;
   break;
  } else {
   $$016 = $33;
  }
 }
 $35 = HEAP8[$$pre$phiZ2D>>0]|0; //@line 2326 "load_abc.cpp"
 $36 = $35&255; //@line 2326 "load_abc.cpp"
 return ($36|0); //@line 2326 "load_abc.cpp"
}
function __ZL17ABC_CleanupMacrosP10_ABCHANDLE($handle) {
 $handle = $handle|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $mp$03 = 0, $mp$12 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($handle|0)==(0|0); //@line 2479 "load_abc.cpp"
 if ($0) {
  return; //@line 2492 "load_abc.cpp"
 }
 $1 = HEAPU8[$handle>>0]|(HEAPU8[$handle+1>>0]<<8)|(HEAPU8[$handle+2>>0]<<16)|(HEAPU8[$handle+3>>0]<<24); //@line 2480 "load_abc.cpp"
 $2 = ($1|0)==(0|0); //@line 2480 "load_abc.cpp"
 if (!($2)) {
  $mp$03 = $1;
  while(1) {
   $3 = HEAPU8[$mp$03>>0]|(HEAPU8[$mp$03+1>>0]<<8)|(HEAPU8[$mp$03+2>>0]<<16)|(HEAPU8[$mp$03+3>>0]<<24); //@line 2481 "load_abc.cpp"
   $4 = ((($mp$03)) + 4|0); //@line 2439 "load_abc.cpp"
   $5 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8)|(HEAPU8[$4+2>>0]<<16)|(HEAPU8[$4+3>>0]<<24); //@line 2439 "load_abc.cpp"
   $6 = ($5|0)==(0|0); //@line 2439 "load_abc.cpp"
   if (!($6)) {
    _free($5); //@line 2440 "load_abc.cpp"
   }
   $7 = ((($mp$03)) + 8|0); //@line 2441 "load_abc.cpp"
   $8 = HEAPU8[$7>>0]|(HEAPU8[$7+1>>0]<<8)|(HEAPU8[$7+2>>0]<<16)|(HEAPU8[$7+3>>0]<<24); //@line 2441 "load_abc.cpp"
   $9 = ($8|0)==(0|0); //@line 2441 "load_abc.cpp"
   if (!($9)) {
    _free($8); //@line 2442 "load_abc.cpp"
   }
   _free($mp$03); //@line 2443 "load_abc.cpp"
   $10 = ($3|0)==(0|0); //@line 2480 "load_abc.cpp"
   if ($10) {
    break;
   } else {
    $mp$03 = $3;
   }
  }
 }
 $11 = ((($handle)) + 4|0); //@line 2484 "load_abc.cpp"
 $12 = HEAPU8[$11>>0]|(HEAPU8[$11+1>>0]<<8)|(HEAPU8[$11+2>>0]<<16)|(HEAPU8[$11+3>>0]<<24); //@line 2484 "load_abc.cpp"
 $13 = ($12|0)==(0|0); //@line 2484 "load_abc.cpp"
 if (!($13)) {
  $mp$12 = $12;
  while(1) {
   $14 = HEAPU8[$mp$12>>0]|(HEAPU8[$mp$12+1>>0]<<8)|(HEAPU8[$mp$12+2>>0]<<16)|(HEAPU8[$mp$12+3>>0]<<24); //@line 2485 "load_abc.cpp"
   $15 = ((($mp$12)) + 4|0); //@line 2439 "load_abc.cpp"
   $16 = HEAPU8[$15>>0]|(HEAPU8[$15+1>>0]<<8)|(HEAPU8[$15+2>>0]<<16)|(HEAPU8[$15+3>>0]<<24); //@line 2439 "load_abc.cpp"
   $17 = ($16|0)==(0|0); //@line 2439 "load_abc.cpp"
   if (!($17)) {
    _free($16); //@line 2440 "load_abc.cpp"
   }
   $18 = ((($mp$12)) + 8|0); //@line 2441 "load_abc.cpp"
   $19 = HEAPU8[$18>>0]|(HEAPU8[$18+1>>0]<<8)|(HEAPU8[$18+2>>0]<<16)|(HEAPU8[$18+3>>0]<<24); //@line 2441 "load_abc.cpp"
   $20 = ($19|0)==(0|0); //@line 2441 "load_abc.cpp"
   if (!($20)) {
    _free($19); //@line 2442 "load_abc.cpp"
   }
   _free($mp$12); //@line 2443 "load_abc.cpp"
   $21 = ($14|0)==(0|0); //@line 2484 "load_abc.cpp"
   if ($21) {
    break;
   } else {
    $mp$12 = $14;
   }
  }
 }
 HEAP8[$handle>>0]=0&255;HEAP8[$handle+1>>0]=(0>>8)&255;HEAP8[$handle+2>>0]=(0>>16)&255;HEAP8[$handle+3>>0]=0>>24; //@line 2488 "load_abc.cpp"
 HEAP8[$11>>0]=0&255;HEAP8[$11+1>>0]=(0>>8)&255;HEAP8[$11+2>>0]=(0>>16)&255;HEAP8[$11+3>>0]=0>>24; //@line 2489 "load_abc.cpp"
 return; //@line 2492 "load_abc.cpp"
}
function _pat_smptogm($smp) {
 $smp = $smp|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($smp|0)<(191); //@line 211 "load_pat.cpp"
 if (!($0)) {
  $$0 = 1;
  return ($$0|0); //@line 214 "load_pat.cpp"
 }
 $1 = (($smp) + -1)|0; //@line 212 "load_pat.cpp"
 $2 = (1252699 + ($1)|0); //@line 212 "load_pat.cpp"
 $3 = HEAP8[$2>>0]|0; //@line 212 "load_pat.cpp"
 $4 = $3&255; //@line 212 "load_pat.cpp"
 $$0 = $4;
 return ($$0|0); //@line 214 "load_pat.cpp"
}
function _pat_gm_name($gm) {
 $gm = $gm|0;
 var $$0 = 0, $0 = 0, $1 = 0, $gm$off = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp; //@line 235 "load_pat.cpp"
 $gm$off = (($gm) + -1)|0; //@line 238 "load_pat.cpp"
 $0 = ($gm$off>>>0)>(190); //@line 238 "load_pat.cpp"
 if ($0) {
  HEAP32[$vararg_buffer>>2] = $gm; //@line 239 "load_pat.cpp"
  (_sprintf(1278235,527820,$vararg_buffer)|0); //@line 239 "load_pat.cpp"
  $$0 = 1278235;
  STACKTOP = sp;return ($$0|0); //@line 243 "load_pat.cpp"
 } else {
  $1 = (1253146 + ($gm$off<<7)|0); //@line 242 "load_pat.cpp"
  $$0 = $1;
  STACKTOP = sp;return ($$0|0); //@line 243 "load_pat.cpp"
 }
 return (0)|0;
}
function __ZL16ABC_ReadPatternsPP11_MODCOMMANDPtP10_ABCHANDLEii($pattern,$psize,$h,$numpat,$channels) {
 $pattern = $pattern|0;
 $psize = $psize|0;
 $h = $h|0;
 $numpat = $numpat|0;
 $channels = $channels|0;
 var $$ = 0, $$0 = 0, $$01$i = 0, $$01$i$lcssa = 0, $$01$i10 = 0, $$01$i10$lcssa = 0, $$01$i14 = 0, $$01$i14$lcssa = 0, $$02$i = 0, $$02$i$lcssa = 0, $$02$i20 = 0, $$02$i20$lcssa = 0, $$numpat = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre84 = 0, $$pre84$pre = 0, $$vol$0 = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0;
 var $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0;
 var $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0;
 var $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $ch$057 = 0, $div = 0, $div7 = 0, $div8 = 0, $e$031134 = 0, $e$1$ef$0 = 0, $e$1$lcssa91 = 0, $e$137$lcssa = 0, $e$137139 = 0;
 var $e$2$ef$0 = 0, $e$246143 = 0, $ef$0$lcssa = 0, $ef$047142 = 0, $ef$2 = 0, $el$0$lcssa = 0, $el$048141 = 0, $el$1 = 0, $i$0$lcssa = 0, $i$045144 = 0, $i$1 = 0, $ins$0 = 0, $ins$1 = 0, $n$0 = 0, $n$0$in = 0, $n$1 = 0, $n$1$in = 0, $or$cond = 0, $pat$070 = 0, $patbrk$0$lcssa = 0;
 var $patbrk$032133 = 0, $patbrk$1 = 0, $patbrk$2$lcssa = 0, $patbrk$259 = 0, $patbrk$3 = 0, $row$067 = 0, $t$0 = 0, $t$073 = 0, $t$074 = 0, $t$1 = 0, $t$161 = 0, $tempo$0$lcssa = 0, $tempo$033132 = 0, $tempo$1 = 0, $tempo$2$lcssa = 0, $tempo$260 = 0, $tempo$3 = 0, $trillbits$069 = 0, $trillbits$168 = 0, $trillbits$2$lcssa = 0;
 var $trillbits$2$lcssa$lcssa185 = 0, $trillbits$258 = 0, $trillbits$3 = 0, $trillbits$4 = 0, $trillbits$498 = 0, $trillbits$499 = 0, $vol$0 = 0, $vol$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($numpat|0)>(240); //@line 2710 "load_abc.cpp"
 $$numpat = $0 ? 240 : $numpat; //@line 2710 "load_abc.cpp"
 $1 = ((($h)) + 8|0); //@line 2713 "load_abc.cpp"
 $t$073 = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24); //@line 2713 "load_abc.cpp"
 $2 = ($t$073|0)==(0|0); //@line 2713 "load_abc.cpp"
 if (!($2)) {
  $t$074 = $t$073;
  while(1) {
   $5 = ((($t$074)) + 4|0); //@line 2713 "load_abc.cpp"
   $6 = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8)|(HEAPU8[$5+2>>0]<<16)|(HEAPU8[$5+3>>0]<<24); //@line 2713 "load_abc.cpp"
   $7 = ((($t$074)) + 12|0); //@line 2713 "load_abc.cpp"
   HEAP8[$7>>0]=$6&255;HEAP8[$7+1>>0]=($6>>8)&255;HEAP8[$7+2>>0]=($6>>16)&255;HEAP8[$7+3>>0]=$6>>24; //@line 2713 "load_abc.cpp"
   $t$0 = HEAPU8[$t$074>>0]|(HEAPU8[$t$074+1>>0]<<8)|(HEAPU8[$t$074+2>>0]<<16)|(HEAPU8[$t$074+3>>0]<<24); //@line 2713 "load_abc.cpp"
   $8 = ($t$0|0)==(0|0); //@line 2713 "load_abc.cpp"
   if ($8) {
    break;
   } else {
    $t$074 = $t$0;
   }
  }
 }
 $3 = ($$numpat|0)>(0); //@line 2715 "load_abc.cpp"
 if (!($3)) {
  $$0 = 0;
  return ($$0|0); //@line 2871 "load_abc.cpp"
 }
 $4 = ((($h)) + 20|0); //@line 2720 "load_abc.cpp"
 $pat$070 = 0;$trillbits$069 = 0;
 L8: while(1) {
  $9 = (__ZN10CSoundFile15AllocatePatternEjj(64,$channels)|0); //@line 2716 "load_abc.cpp"
  $10 = (($pattern) + ($pat$070<<2)|0); //@line 2716 "load_abc.cpp"
  HEAP32[$10>>2] = $9; //@line 2716 "load_abc.cpp"
  $11 = ($9|0)==(0|0); //@line 2717 "load_abc.cpp"
  if ($11) {
   $$0 = 0;
   label = 75;
   break;
  }
  $12 = (($psize) + ($pat$070<<1)|0); //@line 2718 "load_abc.cpp"
  HEAP16[$12>>1] = 64; //@line 2718 "load_abc.cpp"
  $13 = $pat$070 << 6; //@line 2720 "load_abc.cpp"
  $row$067 = 0;$trillbits$168 = $trillbits$069;
  while(1) {
   $16 = (($row$067) + ($13))|0; //@line 2720 "load_abc.cpp"
   $17 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8)|(HEAPU8[$4+2>>0]<<16)|(HEAPU8[$4+3>>0]<<24); //@line 2720 "load_abc.cpp"
   $18 = ($16*192)|0; //@line 2720 "load_abc.cpp"
   $19 = Math_imul($18, $17)|0; //@line 339 "load_abc.cpp"
   $20 = ($17*192)|0; //@line 339 "load_abc.cpp"
   $21 = (($19) + ($20))|0; //@line 2721 "load_abc.cpp"
   $22 = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24); //@line 2725 "load_abc.cpp"
   $23 = ((($22)) + 12|0); //@line 2725 "load_abc.cpp"
   $24 = HEAPU8[$23>>0]|(HEAPU8[$23+1>>0]<<8)|(HEAPU8[$23+2>>0]<<16)|(HEAPU8[$23+3>>0]<<24); //@line 2725 "load_abc.cpp"
   $25 = ($24|0)==(0|0); //@line 2526 "load_abc.cpp"
   L13: do {
    if ($25) {
     $patbrk$0$lcssa = 0;$tempo$0$lcssa = 0;
    } else {
     $$02$i20 = $24;
     while(1) {
      $26 = ((($$02$i20)) + 8|0); //@line 2521 "load_abc.cpp"
      $27 = HEAP8[$26>>0]|0; //@line 2521 "load_abc.cpp"
      $28 = ($27<<24>>24)==(1); //@line 2521 "load_abc.cpp"
      if ($28) {
       $29 = ((($$02$i20)) + 9|0); //@line 2521 "load_abc.cpp"
       $30 = HEAP8[$29>>0]|0; //@line 2521 "load_abc.cpp"
       $31 = $30 & -5; //@line 2521 "load_abc.cpp"
       $32 = ($31<<24>>24)==(80); //@line 2521 "load_abc.cpp"
       if ($32) {
        $$02$i20$lcssa = $$02$i20;
        break;
       }
      }
      $36 = HEAPU8[$$02$i20>>0]|(HEAPU8[$$02$i20+1>>0]<<8)|(HEAPU8[$$02$i20+2>>0]<<16)|(HEAPU8[$$02$i20+3>>0]<<24); //@line 2526 "load_abc.cpp"
      $37 = ($36|0)==(0|0); //@line 2526 "load_abc.cpp"
      if ($37) {
       $patbrk$0$lcssa = 0;$tempo$0$lcssa = 0;
       break L13;
      } else {
       $$02$i20 = $36;
      }
     }
     $33 = ((($$02$i20$lcssa)) + 4|0); //@line 2725 "load_abc.cpp"
     $34 = HEAPU8[$33>>0]|(HEAPU8[$33+1>>0]<<8)|(HEAPU8[$33+2>>0]<<16)|(HEAPU8[$33+3>>0]<<24); //@line 2725 "load_abc.cpp"
     $35 = ($34>>>0)<($21>>>0); //@line 2725 "load_abc.cpp"
     if ($35) {
      $44 = $34;$e$031134 = $$02$i20$lcssa;$patbrk$032133 = 0;$tempo$033132 = 0;
      while(1) {
       $43 = ($44>>>0)<($19>>>0); //@line 2726 "load_abc.cpp"
       L22: do {
        if ($43) {
         $patbrk$1 = $patbrk$032133;$tempo$1 = $tempo$033132;
        } else {
         $45 = ((($e$031134)) + 9|0); //@line 2727 "load_abc.cpp"
         $46 = HEAP8[$45>>0]|0; //@line 2727 "load_abc.cpp"
         $47 = $46&255; //@line 2727 "load_abc.cpp"
         switch ($47|0) {
         case 84:  {
          $48 = ((($e$031134)) + 10|0); //@line 2729 "load_abc.cpp"
          $49 = HEAPU8[$48>>0]|(HEAPU8[$48+1>>0]<<8)|(HEAPU8[$48+2>>0]<<16)|(HEAPU8[$48+3>>0]<<24); //@line 2729 "load_abc.cpp"
          $patbrk$1 = $patbrk$032133;$tempo$1 = $49;
          break L22;
          break;
         }
         case 80:  {
          $patbrk$1 = 1;$tempo$1 = $tempo$033132;
          break L22;
          break;
         }
         default: {
          $patbrk$1 = $patbrk$032133;$tempo$1 = $tempo$033132;
          break L22;
         }
         }
        }
       } while(0);
       $50 = HEAPU8[$e$031134>>0]|(HEAPU8[$e$031134+1>>0]<<8)|(HEAPU8[$e$031134+2>>0]<<16)|(HEAPU8[$e$031134+3>>0]<<24); //@line 2725 "load_abc.cpp"
       $51 = ($50|0)==(0|0); //@line 2526 "load_abc.cpp"
       if ($51) {
        $patbrk$0$lcssa = $patbrk$1;$tempo$0$lcssa = $tempo$1;
        break L13;
       } else {
        $$02$i = $50;
       }
       while(1) {
        $52 = ((($$02$i)) + 8|0); //@line 2521 "load_abc.cpp"
        $53 = HEAP8[$52>>0]|0; //@line 2521 "load_abc.cpp"
        $54 = ($53<<24>>24)==(1); //@line 2521 "load_abc.cpp"
        if ($54) {
         $55 = ((($$02$i)) + 9|0); //@line 2521 "load_abc.cpp"
         $56 = HEAP8[$55>>0]|0; //@line 2521 "load_abc.cpp"
         $57 = $56 & -5; //@line 2521 "load_abc.cpp"
         $58 = ($57<<24>>24)==(80); //@line 2521 "load_abc.cpp"
         if ($58) {
          $$02$i$lcssa = $$02$i;
          break;
         }
        }
        $59 = HEAPU8[$$02$i>>0]|(HEAPU8[$$02$i+1>>0]<<8)|(HEAPU8[$$02$i+2>>0]<<16)|(HEAPU8[$$02$i+3>>0]<<24); //@line 2526 "load_abc.cpp"
        $60 = ($59|0)==(0|0); //@line 2526 "load_abc.cpp"
        if ($60) {
         $patbrk$0$lcssa = $patbrk$1;$tempo$0$lcssa = $tempo$1;
         break L13;
        } else {
         $$02$i = $59;
        }
       }
       $38 = ((($$02$i$lcssa)) + 4|0); //@line 2725 "load_abc.cpp"
       $39 = HEAPU8[$38>>0]|(HEAPU8[$38+1>>0]<<8)|(HEAPU8[$38+2>>0]<<16)|(HEAPU8[$38+3>>0]<<24); //@line 2725 "load_abc.cpp"
       $40 = ($39>>>0)<($21>>>0); //@line 2725 "load_abc.cpp"
       if ($40) {
        $44 = $39;$e$031134 = $$02$i$lcssa;$patbrk$032133 = $patbrk$1;$tempo$033132 = $tempo$1;
       } else {
        $patbrk$0$lcssa = $patbrk$1;$tempo$0$lcssa = $tempo$1;
        break;
       }
      }
     } else {
      $patbrk$0$lcssa = 0;$tempo$0$lcssa = 0;
     }
    }
   } while(0);
   $41 = ($22|0)==(0|0); //@line 2737 "load_abc.cpp"
   L35: do {
    if ($41) {
     $patbrk$2$lcssa = $patbrk$0$lcssa;$tempo$2$lcssa = $tempo$0$lcssa;$trillbits$2$lcssa = $trillbits$168;
    } else {
     $42 = Math_imul($row$067, $channels)|0; //@line 2751 "load_abc.cpp"
     $63 = $24;$ch$057 = 0;$patbrk$259 = $patbrk$0$lcssa;$t$161 = $22;$tempo$260 = $tempo$0$lcssa;$trillbits$258 = $trillbits$168;
     while(1) {
      $61 = ((($t$161)) + 12|0); //@line 2738 "load_abc.cpp"
      $62 = ($63|0)==(0|0); //@line 2532 "load_abc.cpp"
      L39: do {
       if ($62) {
        $144 = 0;$e$1$lcssa91 = 0;$ef$0$lcssa = 0;$el$0$lcssa = 0;$i$0$lcssa = 0;
       } else {
        $$01$i14 = $63;
        while(1) {
         $64 = ((($$01$i14)) + 8|0); //@line 2532 "load_abc.cpp"
         $65 = HEAP8[$64>>0]|0; //@line 2532 "load_abc.cpp"
         $66 = ($65<<24>>24)==(1); //@line 2532 "load_abc.cpp"
         if (!($66)) {
          $$01$i14$lcssa = $$01$i14;
          break;
         }
         $70 = HEAPU8[$$01$i14>>0]|(HEAPU8[$$01$i14+1>>0]<<8)|(HEAPU8[$$01$i14+2>>0]<<16)|(HEAPU8[$$01$i14+3>>0]<<24); //@line 2532 "load_abc.cpp"
         $71 = ($70|0)==(0|0); //@line 2532 "load_abc.cpp"
         if ($71) {
          $144 = $63;$e$1$lcssa91 = 0;$ef$0$lcssa = 0;$el$0$lcssa = 0;$i$0$lcssa = 0;
          break L39;
         } else {
          $$01$i14 = $70;
         }
        }
        $67 = ((($$01$i14$lcssa)) + 4|0); //@line 2738 "load_abc.cpp"
        $68 = HEAPU8[$67>>0]|(HEAPU8[$67+1>>0]<<8)|(HEAPU8[$67+2>>0]<<16)|(HEAPU8[$67+3>>0]<<24); //@line 2738 "load_abc.cpp"
        $69 = ($68>>>0)<($19>>>0); //@line 2738 "load_abc.cpp"
        if ($69) {
         $e$137139 = $$01$i14$lcssa;
         while(1) {
          $78 = HEAPU8[$e$137139>>0]|(HEAPU8[$e$137139+1>>0]<<8)|(HEAPU8[$e$137139+2>>0]<<16)|(HEAPU8[$e$137139+3>>0]<<24); //@line 2738 "load_abc.cpp"
          $79 = ($78|0)==(0|0); //@line 2532 "load_abc.cpp"
          if ($79) {
           $144 = $63;$e$1$lcssa91 = 0;$ef$0$lcssa = 0;$el$0$lcssa = 0;$i$0$lcssa = 0;
           break L39;
          } else {
           $$01$i10 = $78;
          }
          while(1) {
           $80 = ((($$01$i10)) + 8|0); //@line 2532 "load_abc.cpp"
           $81 = HEAP8[$80>>0]|0; //@line 2532 "load_abc.cpp"
           $82 = ($81<<24>>24)==(1); //@line 2532 "load_abc.cpp"
           if (!($82)) {
            $$01$i10$lcssa = $$01$i10;
            break;
           }
           $83 = HEAPU8[$$01$i10>>0]|(HEAPU8[$$01$i10+1>>0]<<8)|(HEAPU8[$$01$i10+2>>0]<<16)|(HEAPU8[$$01$i10+3>>0]<<24); //@line 2532 "load_abc.cpp"
           $84 = ($83|0)==(0|0); //@line 2532 "load_abc.cpp"
           if ($84) {
            $144 = $63;$e$1$lcssa91 = 0;$ef$0$lcssa = 0;$el$0$lcssa = 0;$i$0$lcssa = 0;
            break L39;
           } else {
            $$01$i10 = $83;
           }
          }
          $72 = ((($$01$i10$lcssa)) + 4|0); //@line 2738 "load_abc.cpp"
          $73 = HEAPU8[$72>>0]|(HEAPU8[$72+1>>0]<<8)|(HEAPU8[$72+2>>0]<<16)|(HEAPU8[$72+3>>0]<<24); //@line 2738 "load_abc.cpp"
          $74 = ($73>>>0)<($19>>>0); //@line 2738 "load_abc.cpp"
          if ($74) {
           $e$137139 = $$01$i10$lcssa;
          } else {
           $e$137$lcssa = $$01$i10$lcssa;
           break;
          }
         }
        } else {
         $e$137$lcssa = $$01$i14$lcssa;
        }
        $75 = ((($e$137$lcssa)) + 4|0); //@line 2743 "load_abc.cpp"
        $76 = HEAPU8[$75>>0]|(HEAPU8[$75+1>>0]<<8)|(HEAPU8[$75+2>>0]<<16)|(HEAPU8[$75+3>>0]<<24); //@line 2743 "load_abc.cpp"
        $77 = ($76>>>0)<($21>>>0); //@line 2743 "load_abc.cpp"
        if ($77) {
         $e$246143 = $e$137$lcssa;$ef$047142 = 0;$el$048141 = $e$137$lcssa;$i$045144 = 0;
         while(1) {
          HEAP8[$61>>0]=$e$246143&255;HEAP8[$61+1>>0]=($e$246143>>8)&255;HEAP8[$61+2>>0]=($e$246143>>16)&255;HEAP8[$61+3>>0]=$e$246143>>24; //@line 2744 "load_abc.cpp"
          $88 = (($i$045144) + 1)|0; //@line 2745 "load_abc.cpp"
          $89 = ((($e$246143)) + 11|0); //@line 2746 "load_abc.cpp"
          $90 = HEAP8[$89>>0]|0; //@line 2746 "load_abc.cpp"
          $91 = ($90<<24>>24)==(0); //@line 2746 "load_abc.cpp"
          $92 = ($ef$047142|0)==(0|0); //@line 2747 "load_abc.cpp"
          $e$2$ef$0 = $92 ? $e$246143 : $ef$047142; //@line 2747 "load_abc.cpp"
          $ef$2 = $91 ? $ef$047142 : $e$2$ef$0; //@line 2746 "load_abc.cpp"
          $el$1 = $91 ? $el$048141 : $e$246143; //@line 2746 "load_abc.cpp"
          $93 = HEAPU8[$e$246143>>0]|(HEAPU8[$e$246143+1>>0]<<8)|(HEAPU8[$e$246143+2>>0]<<16)|(HEAPU8[$e$246143+3>>0]<<24); //@line 2743 "load_abc.cpp"
          $94 = ($93|0)==(0|0); //@line 2532 "load_abc.cpp"
          if ($94) {
           $144 = $e$246143;$e$1$lcssa91 = $e$137$lcssa;$ef$0$lcssa = $ef$2;$el$0$lcssa = $el$1;$i$0$lcssa = $88;
           break L39;
          } else {
           $$01$i = $93;
          }
          while(1) {
           $95 = ((($$01$i)) + 8|0); //@line 2532 "load_abc.cpp"
           $96 = HEAP8[$95>>0]|0; //@line 2532 "load_abc.cpp"
           $97 = ($96<<24>>24)==(1); //@line 2532 "load_abc.cpp"
           if (!($97)) {
            $$01$i$lcssa = $$01$i;
            break;
           }
           $98 = HEAPU8[$$01$i>>0]|(HEAPU8[$$01$i+1>>0]<<8)|(HEAPU8[$$01$i+2>>0]<<16)|(HEAPU8[$$01$i+3>>0]<<24); //@line 2532 "load_abc.cpp"
           $99 = ($98|0)==(0|0); //@line 2532 "load_abc.cpp"
           if ($99) {
            $144 = $e$246143;$e$1$lcssa91 = $e$137$lcssa;$ef$0$lcssa = $ef$2;$el$0$lcssa = $el$1;$i$0$lcssa = $88;
            break L39;
           } else {
            $$01$i = $98;
           }
          }
          $85 = ((($$01$i$lcssa)) + 4|0); //@line 2743 "load_abc.cpp"
          $86 = HEAPU8[$85>>0]|(HEAPU8[$85+1>>0]<<8)|(HEAPU8[$85+2>>0]<<16)|(HEAPU8[$85+3>>0]<<24); //@line 2743 "load_abc.cpp"
          $87 = ($86>>>0)<($21>>>0); //@line 2743 "load_abc.cpp"
          if ($87) {
           $e$246143 = $$01$i$lcssa;$ef$047142 = $ef$2;$el$048141 = $el$1;$i$045144 = $88;
          } else {
           $144 = $e$246143;$e$1$lcssa91 = $e$137$lcssa;$ef$0$lcssa = $ef$2;$el$0$lcssa = $el$1;$i$0$lcssa = $88;
           break;
          }
         }
        } else {
         $144 = $63;$e$1$lcssa91 = $e$137$lcssa;$ef$0$lcssa = 0;$el$0$lcssa = $e$137$lcssa;$i$0$lcssa = 0;
        }
       }
      } while(0);
      $100 = (($ch$057) + ($42))|0; //@line 2751 "load_abc.cpp"
      $101 = HEAP32[$10>>2]|0; //@line 2751 "load_abc.cpp"
      $102 = (($101) + (($100*6)|0)|0); //@line 2751 "load_abc.cpp"
      $103 = (((($101) + (($100*6)|0)|0)) + 5|0); //@line 2752 "load_abc.cpp"
      HEAP8[$103>>0] = 0; //@line 2752 "load_abc.cpp"
      $104 = (((($101) + (($100*6)|0)|0)) + 3|0); //@line 2753 "load_abc.cpp"
      HEAP8[$104>>0] = 0; //@line 2753 "load_abc.cpp"
      $105 = ($i$0$lcssa|0)==(0); //@line 2754 "load_abc.cpp"
      $106 = 1 << $ch$057; //@line 2842 "load_abc.cpp"
      if ($105) {
       $234 = $trillbits$258 & $106; //@line 2842 "load_abc.cpp"
       $235 = ($234|0)==(0); //@line 2842 "load_abc.cpp"
       if ($235) {
        $trillbits$499 = $trillbits$258;
        label = 66;
       } else {
        HEAP8[$104>>0] = 5; //@line 2844 "load_abc.cpp"
        HEAP8[$103>>0] = 0; //@line 2845 "load_abc.cpp"
        $236 = (((($101) + (($100*6)|0)|0)) + 1|0); //@line 2846 "load_abc.cpp"
        HEAP8[$236>>0] = 0; //@line 2846 "load_abc.cpp"
        $237 = (($101) + (($100*6)|0)|0); //@line 2847 "load_abc.cpp"
        HEAP8[$237>>0] = 0; //@line 2847 "load_abc.cpp"
        $238 = (((($101) + (($100*6)|0)|0)) + 2|0); //@line 2848 "load_abc.cpp"
        HEAP8[$238>>0] = 0; //@line 2848 "load_abc.cpp"
        $239 = (((($101) + (($100*6)|0)|0)) + 4|0); //@line 2849 "load_abc.cpp"
        HEAP8[$239>>0] = 0; //@line 2849 "load_abc.cpp"
        $trillbits$499 = $trillbits$258;
        label = 66;
       }
      } else {
       $107 = $106 ^ -1; //@line 2755 "load_abc.cpp"
       $108 = $trillbits$258 & $107; //@line 2755 "load_abc.cpp"
       $109 = ($i$0$lcssa|0)==(1); //@line 2756 "load_abc.cpp"
       do {
        if ($109) {
         label = 40;
        } else {
         $110 = ($ef$0$lcssa|0)!=($el$0$lcssa|0); //@line 2756 "load_abc.cpp"
         $111 = ($ef$0$lcssa|0)!=(0|0); //@line 2756 "load_abc.cpp"
         $or$cond = $111 & $110; //@line 2756 "load_abc.cpp"
         if ($or$cond) {
          $173 = ((($t$161)) + 40|0); //@line 2808 "load_abc.cpp"
          $174 = HEAP8[$173>>0]|0; //@line 2808 "load_abc.cpp"
          $175 = ($174<<24>>24)==(10); //@line 2808 "load_abc.cpp"
          if ($175) {
           $176 = ((($el$0$lcssa)) + 8|0); //@line 2809 "load_abc.cpp"
           $177 = HEAP8[$176>>0]|0; //@line 2809 "load_abc.cpp"
           $178 = $177&255; //@line 2809 "load_abc.cpp"
           $179 = ((($el$0$lcssa)) + 9|0); //@line 2809 "load_abc.cpp"
           $180 = HEAP8[$179>>0]|0; //@line 2809 "load_abc.cpp"
           $181 = $180&255; //@line 2809 "load_abc.cpp"
           $182 = ($181*12)|0; //@line 2809 "load_abc.cpp"
           $183 = (($182) + ($178))|0; //@line 2809 "load_abc.cpp"
           $184 = (_pat_gm_drumnote($183)|0); //@line 2810 "load_abc.cpp"
           $185 = (($184) + 23)|0; //@line 2810 "load_abc.cpp"
           $186 = (_pat_gm_drumnr($183)|0); //@line 2811 "load_abc.cpp"
           $187 = (_pat_gmtosmp($186)|0); //@line 2811 "load_abc.cpp"
           $188 = $187&255; //@line 2811 "load_abc.cpp"
           $189 = $185 & 255; //@line 2812 "load_abc.cpp"
           $i$1 = $189;$ins$1 = $188;$n$1$in = $185;
          } else {
           $190 = ((($ef$0$lcssa)) + 8|0); //@line 2815 "load_abc.cpp"
           $191 = HEAP8[$190>>0]|0; //@line 2815 "load_abc.cpp"
           $192 = $191&255; //@line 2815 "load_abc.cpp"
           $193 = ((($ef$0$lcssa)) + 9|0); //@line 2815 "load_abc.cpp"
           $194 = HEAP8[$193>>0]|0; //@line 2815 "load_abc.cpp"
           $195 = $194&255; //@line 2815 "load_abc.cpp"
           $196 = ($195*12)|0; //@line 2815 "load_abc.cpp"
           $197 = (($196) + ($192))|0; //@line 2815 "load_abc.cpp"
           $198 = (_pat_modnote($197)|0); //@line 2816 "load_abc.cpp"
           $199 = ((($el$0$lcssa)) + 10|0); //@line 2817 "load_abc.cpp"
           $200 = HEAP8[$199>>0]|0; //@line 2817 "load_abc.cpp"
           $201 = ((($el$0$lcssa)) + 8|0); //@line 2818 "load_abc.cpp"
           $202 = HEAP8[$201>>0]|0; //@line 2818 "load_abc.cpp"
           $203 = $202&255; //@line 2818 "load_abc.cpp"
           $204 = ((($el$0$lcssa)) + 9|0); //@line 2818 "load_abc.cpp"
           $205 = HEAP8[$204>>0]|0; //@line 2818 "load_abc.cpp"
           $206 = $205&255; //@line 2818 "load_abc.cpp"
           $207 = ($206*12)|0; //@line 2818 "load_abc.cpp"
           $208 = (($207) + ($203))|0; //@line 2818 "load_abc.cpp"
           $209 = (_pat_modnote($208)|0); //@line 2818 "load_abc.cpp"
           $i$1 = $209;$ins$1 = $200;$n$1$in = $198;
          }
          $n$1 = $n$1$in&255; //@line 2810 "load_abc.cpp"
          $210 = ((($el$0$lcssa)) + 11|0); //@line 2820 "load_abc.cpp"
          $211 = HEAP8[$210>>0]|0; //@line 2820 "load_abc.cpp"
          $div8 = ($211&255) >>> 1; //@line 2820 "load_abc.cpp"
          $212 = ($div8&255)>(64); //@line 2821 "load_abc.cpp"
          $$ = $212 ? 64 : $div8; //@line 2821 "load_abc.cpp"
          $213 = (((($101) + (($100*6)|0)|0)) + 1|0); //@line 2822 "load_abc.cpp"
          HEAP8[$213>>0] = $ins$1; //@line 2822 "load_abc.cpp"
          HEAP8[$102>>0] = $n$1; //@line 2823 "load_abc.cpp"
          $214 = (((($101) + (($100*6)|0)|0)) + 2|0); //@line 2824 "load_abc.cpp"
          HEAP8[$214>>0] = 1; //@line 2824 "load_abc.cpp"
          $215 = (((($101) + (($100*6)|0)|0)) + 4|0); //@line 2825 "load_abc.cpp"
          HEAP8[$215>>0] = $$; //@line 2825 "load_abc.cpp"
          $216 = $n$1$in & 255; //@line 2826 "load_abc.cpp"
          $217 = ($i$1|0)>($216|0); //@line 2826 "load_abc.cpp"
          $218 = (($i$1) - ($216))|0; //@line 2826 "load_abc.cpp"
          $219 = (($216) - ($i$1))|0; //@line 2826 "load_abc.cpp"
          $220 = $217 ? $218 : $219; //@line 2826 "load_abc.cpp"
          $221 = $220&255; //@line 2826 "load_abc.cpp"
          HEAP8[$103>>0] = $221; //@line 2826 "load_abc.cpp"
          $222 = $220 & 240; //@line 2827 "load_abc.cpp"
          $223 = ($222>>>0)<(16); //@line 2827 "load_abc.cpp"
          if (!($223)) {
           $233 = $217 ? 2 : 3; //@line 2838 "load_abc.cpp"
           HEAP8[$104>>0] = $233; //@line 2838 "load_abc.cpp"
           $241 = $221;$trillbits$4 = $108;
           break;
          }
          $224 = ($221<<24>>24)==(0); //@line 2828 "load_abc.cpp"
          if ($224) {
           HEAP8[$104>>0] = 15; //@line 2833 "load_abc.cpp"
           $228 = ((($el$0$lcssa)) + 4|0); //@line 2834 "load_abc.cpp"
           $229 = HEAPU8[$228>>0]|(HEAPU8[$228+1>>0]<<8)|(HEAPU8[$228+2>>0]<<16)|(HEAPU8[$228+3>>0]<<24); //@line 2834 "load_abc.cpp"
           $230 = (($229) - ($19))|0; //@line 2834 "load_abc.cpp"
           $231 = (($230>>>0) / 192)&-1; //@line 334 "load_abc.cpp"
           $232 = $231&255; //@line 2834 "load_abc.cpp"
           HEAP8[$103>>0] = $232; //@line 2834 "load_abc.cpp"
           $241 = $232;$trillbits$4 = $108;
           break;
          } else {
           HEAP8[$104>>0] = 28; //@line 2829 "load_abc.cpp"
           $225 = $217 ? 16 : 32; //@line 2830 "load_abc.cpp"
           $226 = $220 | $225; //@line 2830 "load_abc.cpp"
           $227 = $226&255; //@line 2830 "load_abc.cpp"
           HEAP8[$103>>0] = $227; //@line 2830 "load_abc.cpp"
           $241 = $227;$trillbits$4 = $108;
           break;
          }
         } else {
          label = 40;
         }
        }
       } while(0);
       if ((label|0) == 40) {
        label = 0;
        $112 = ($ef$0$lcssa|0)==(0|0); //@line 2757 "load_abc.cpp"
        $e$1$ef$0 = $112 ? $e$1$lcssa91 : $ef$0$lcssa;
        $113 = ((($e$1$ef$0)) + 8|0); //@line 2760 "load_abc.cpp"
        $114 = HEAP8[$113>>0]|0; //@line 2760 "load_abc.cpp"
        $115 = $114&255; //@line 2760 "load_abc.cpp"
        $116 = ((($e$1$ef$0)) + 9|0); //@line 2760 "load_abc.cpp"
        $117 = HEAP8[$116>>0]|0; //@line 2760 "load_abc.cpp"
        $118 = $117&255; //@line 2760 "load_abc.cpp"
        $119 = ($118*12)|0; //@line 2760 "load_abc.cpp"
        $120 = (($119) + ($115))|0; //@line 2760 "load_abc.cpp"
        $121 = ((($t$161)) + 40|0); //@line 2761 "load_abc.cpp"
        $122 = HEAP8[$121>>0]|0; //@line 2761 "load_abc.cpp"
        $123 = ($122<<24>>24)==(10); //@line 2761 "load_abc.cpp"
        if ($123) {
         $124 = (_pat_gm_drumnote($120)|0); //@line 2762 "load_abc.cpp"
         $125 = (($124) + 23)|0; //@line 2762 "load_abc.cpp"
         $126 = (_pat_gm_drumnr($120)|0); //@line 2763 "load_abc.cpp"
         $127 = (_pat_gmtosmp($126)|0); //@line 2763 "load_abc.cpp"
         $128 = $127&255; //@line 2763 "load_abc.cpp"
         $ins$0 = $128;$n$0$in = $125;
        } else {
         $129 = (_pat_modnote($120)|0); //@line 2766 "load_abc.cpp"
         $130 = ((($e$1$ef$0)) + 10|0); //@line 2767 "load_abc.cpp"
         $131 = HEAP8[$130>>0]|0; //@line 2767 "load_abc.cpp"
         $ins$0 = $131;$n$0$in = $129;
        }
        $n$0 = $n$0$in&255; //@line 2762 "load_abc.cpp"
        $132 = ((($e$1$ef$0)) + 11|0); //@line 2769 "load_abc.cpp"
        $133 = HEAP8[$132>>0]|0; //@line 2769 "load_abc.cpp"
        $div = ($133&255) >>> 1; //@line 2769 "load_abc.cpp"
        $134 = ($133<<24>>24)==(0); //@line 2770 "load_abc.cpp"
        L82: do {
         if ($134) {
          $$pre84$pre = HEAP8[$103>>0]|0; //@line 2852 "load_abc.cpp"
          $$pre84 = $$pre84$pre;$trillbits$3 = $108;$vol$2 = $div;
         } else {
          $135 = ((($e$1$ef$0)) + 12|0); //@line 2771 "load_abc.cpp"
          $136 = HEAP8[$135>>0]|0; //@line 2771 "load_abc.cpp"
          $137 = ($136<<24>>24)==(3); //@line 2771 "load_abc.cpp"
          if ($137) {
           $138 = $div&255; //@line 2771 "load_abc.cpp"
           $div7 = (($133&255) / 40)&-1; //@line 2771 "load_abc.cpp"
           $139 = $div7&255; //@line 2771 "load_abc.cpp"
           $140 = (($139) + ($138))|0; //@line 2771 "load_abc.cpp"
           $141 = $140&255; //@line 2771 "load_abc.cpp"
           $vol$0 = $141;
          } else {
           $vol$0 = $div;
          }
          $142 = ($vol$0&255)>(64); //@line 2772 "load_abc.cpp"
          $$vol$0 = $142 ? 64 : $vol$0; //@line 2772 "load_abc.cpp"
          $143 = ((($144)) + 11|0); //@line 2773 "load_abc.cpp"
          $145 = HEAP8[$143>>0]|0; //@line 2773 "load_abc.cpp"
          $146 = ($145<<24>>24)==(0); //@line 2773 "load_abc.cpp"
          if ($146) {
           $147 = ((($144)) + 4|0); //@line 2774 "load_abc.cpp"
           $148 = HEAPU8[$147>>0]|(HEAPU8[$147+1>>0]<<8)|(HEAPU8[$147+2>>0]<<16)|(HEAPU8[$147+3>>0]<<24); //@line 2774 "load_abc.cpp"
           $149 = (($148) - ($19))|0; //@line 2774 "load_abc.cpp"
           HEAP8[$104>>0] = 20; //@line 2775 "load_abc.cpp"
           $150 = $149 | 192; //@line 2776 "load_abc.cpp"
           $151 = $150&255; //@line 2776 "load_abc.cpp"
           HEAP8[$103>>0] = $151; //@line 2776 "load_abc.cpp"
           $$pre84 = $151;$trillbits$3 = $108;$vol$2 = $$vol$0;
           break;
          }
          $152 = $136&255; //@line 2779 "load_abc.cpp"
          switch ($152|0) {
          case 1:  {
           HEAP8[$104>>0] = 5; //@line 2781 "load_abc.cpp"
           HEAP8[$103>>0] = -62; //@line 2782 "load_abc.cpp"
           $153 = $108 | $106; //@line 2783 "load_abc.cpp"
           $$pre84 = -62;$trillbits$3 = $153;$vol$2 = $$vol$0;
           break L82;
           break;
          }
          case 2:  {
           HEAP8[$104>>0] = 28; //@line 2786 "load_abc.cpp"
           $154 = ((($e$1$ef$0)) + 13|0); //@line 2787 "load_abc.cpp"
           $155 = HEAP8[$154>>0]|0; //@line 2787 "load_abc.cpp"
           $156 = ($155<<24>>24)!=(0); //@line 2787 "load_abc.cpp"
           $157 = $156 ? 18 : 34; //@line 2787 "load_abc.cpp"
           $158 = HEAP8[$103>>0]|0; //@line 2787 "load_abc.cpp"
           $159 = $158&255; //@line 2787 "load_abc.cpp"
           $160 = $159 | $157; //@line 2787 "load_abc.cpp"
           $161 = $160&255; //@line 2787 "load_abc.cpp"
           HEAP8[$103>>0] = $161; //@line 2787 "load_abc.cpp"
           $$pre84 = $161;$trillbits$3 = $108;$vol$2 = $$vol$0;
           break L82;
           break;
          }
          default: {
           $162 = ((($e$1$ef$0)) + 4|0); //@line 2790 "load_abc.cpp"
           $163 = HEAPU8[$162>>0]|(HEAPU8[$162+1>>0]<<8)|(HEAPU8[$162+2>>0]<<16)|(HEAPU8[$162+3>>0]<<24); //@line 2790 "load_abc.cpp"
           $164 = (($163) - ($19))|0; //@line 2790 "load_abc.cpp"
           $165 = (($164>>>0) / 192)&-1; //@line 334 "load_abc.cpp"
           $166 = $165&255; //@line 2790 "load_abc.cpp"
           HEAP8[$103>>0] = $166; //@line 2790 "load_abc.cpp"
           $167 = ($166<<24>>24)==(0); //@line 2791 "load_abc.cpp"
           if ($167) {
            $$pre84 = 0;$trillbits$3 = $108;$vol$2 = $$vol$0;
            break L82;
           }
           HEAP8[$104>>0] = 20; //@line 2792 "load_abc.cpp"
           $168 = $165 | 208; //@line 2793 "load_abc.cpp"
           $169 = $168&255; //@line 2793 "load_abc.cpp"
           HEAP8[$103>>0] = $169; //@line 2793 "load_abc.cpp"
           $$pre84 = $169;$trillbits$3 = $108;$vol$2 = $$vol$0;
           break L82;
          }
          }
         }
        } while(0);
        $170 = (((($101) + (($100*6)|0)|0)) + 1|0); //@line 2799 "load_abc.cpp"
        HEAP8[$170>>0] = $ins$0; //@line 2799 "load_abc.cpp"
        HEAP8[$102>>0] = $n$0; //@line 2800 "load_abc.cpp"
        $171 = (((($101) + (($100*6)|0)|0)) + 2|0); //@line 2801 "load_abc.cpp"
        HEAP8[$171>>0] = 1; //@line 2801 "load_abc.cpp"
        $172 = (((($101) + (($100*6)|0)|0)) + 4|0); //@line 2802 "load_abc.cpp"
        HEAP8[$172>>0] = $vol$2; //@line 2802 "load_abc.cpp"
        $241 = $$pre84;$trillbits$4 = $trillbits$3;
       }
       $240 = ($241<<24>>24)==(0); //@line 2852 "load_abc.cpp"
       if ($240) {
        $trillbits$499 = $trillbits$4;
        label = 66;
       } else {
        $patbrk$3 = $patbrk$259;$tempo$3 = $tempo$260;$trillbits$498 = $trillbits$4;
       }
      }
      do {
       if ((label|0) == 66) {
        label = 0;
        $242 = HEAP8[$104>>0]|0; //@line 2852 "load_abc.cpp"
        $243 = ($242<<24>>24)==(0); //@line 2852 "load_abc.cpp"
        if ($243) {
         $244 = ($tempo$260|0)==(0); //@line 2853 "load_abc.cpp"
         if (!($244)) {
          HEAP8[$104>>0] = 17; //@line 2854 "load_abc.cpp"
          $245 = $tempo$260&255; //@line 2855 "load_abc.cpp"
          HEAP8[$103>>0] = $245; //@line 2855 "load_abc.cpp"
          $patbrk$3 = $patbrk$259;$tempo$3 = 0;$trillbits$498 = $trillbits$499;
          break;
         }
         $246 = ($patbrk$259|0)==(0); //@line 2859 "load_abc.cpp"
         if ($246) {
          $patbrk$3 = 0;$tempo$3 = 0;$trillbits$498 = $trillbits$499;
         } else {
          HEAP8[$104>>0] = 14; //@line 2860 "load_abc.cpp"
          $patbrk$3 = 0;$tempo$3 = 0;$trillbits$498 = $trillbits$499;
         }
        } else {
         $patbrk$3 = $patbrk$259;$tempo$3 = $tempo$260;$trillbits$498 = $trillbits$499;
        }
       }
      } while(0);
      $t$1 = HEAPU8[$t$161>>0]|(HEAPU8[$t$161+1>>0]<<8)|(HEAPU8[$t$161+2>>0]<<16)|(HEAPU8[$t$161+3>>0]<<24); //@line 2737 "load_abc.cpp"
      $247 = ($t$1|0)==(0|0); //@line 2737 "load_abc.cpp"
      if ($247) {
       $patbrk$2$lcssa = $patbrk$3;$tempo$2$lcssa = $tempo$3;$trillbits$2$lcssa = $trillbits$498;
       break L35;
      }
      $248 = (($ch$057) + 1)|0; //@line 2865 "load_abc.cpp"
      $$phi$trans$insert = ((($t$1)) + 12|0);
      $$pre = HEAPU8[$$phi$trans$insert>>0]|(HEAPU8[$$phi$trans$insert+1>>0]<<8)|(HEAPU8[$$phi$trans$insert+2>>0]<<16)|(HEAPU8[$$phi$trans$insert+3>>0]<<24); //@line 2738 "load_abc.cpp"
      $63 = $$pre;$ch$057 = $248;$patbrk$259 = $patbrk$3;$t$161 = $t$1;$tempo$260 = $tempo$3;$trillbits$258 = $trillbits$498;
     }
    }
   } while(0);
   $249 = $tempo$2$lcssa | $patbrk$2$lcssa; //@line 2867 "load_abc.cpp"
   $250 = ($249|0)==(0); //@line 2867 "load_abc.cpp"
   $15 = (($row$067) + 1)|0; //@line 2719 "load_abc.cpp"
   if (!($250)) {
    $$0 = 1;
    label = 75;
    break L8;
   }
   $14 = ($15|0)<(64); //@line 2719 "load_abc.cpp"
   if ($14) {
    $row$067 = $15;$trillbits$168 = $trillbits$2$lcssa;
   } else {
    $trillbits$2$lcssa$lcssa185 = $trillbits$2$lcssa;
    break;
   }
  }
  $251 = (($pat$070) + 1)|0; //@line 2715 "load_abc.cpp"
  $252 = ($251|0)<($$numpat|0); //@line 2715 "load_abc.cpp"
  if ($252) {
   $pat$070 = $251;$trillbits$069 = $trillbits$2$lcssa$lcssa185;
  } else {
   $$0 = 0;
   label = 75;
   break;
  }
 }
 if ((label|0) == 75) {
  return ($$0|0); //@line 2871 "load_abc.cpp"
 }
 return (0)|0;
}
function _pat_modnote($midinote) {
 $midinote = $midinote|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($midinote) + 13)|0; //@line 912 "load_pat.cpp"
 return ($0|0); //@line 914 "load_pat.cpp"
}
function __ZN10CSoundFile11FreePatternEPv($pat) {
 $pat = $pat|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($pat|0)==(0|0); //@line 349 "sndfile.cpp"
 if (!($0)) {
  __ZdaPv($pat); //@line 349 "sndfile.cpp"
 }
 return; //@line 350 "sndfile.cpp"
}
function _PAT_Load_Instruments($c) {
 $c = $c|0;
 var $$$i1$i = 0, $$$i1$i5 = 0, $$$i15$i = 0, $$0$i1 = 0, $$0$i11 = 0, $$lcssa = 0, $$off1$i = 0, $$pre$i = 0, $$pre$i7 = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0.0, $152 = 0.0, $153 = 0.0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0;
 var $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $d$02$i$i$i = 0, $d$02$i$i11$i = 0, $exitcond$i$i = 0, $exitcond$i$i$i = 0, $exitcond$i$i13$i = 0, $exitcond$i3$i = 0, $exitcond$i6$i = 0;
 var $extract$i = 0, $extract$t$i = 0, $extract$t18$i = 0, $extract$t20$i = 0, $extract17$i = 0, $extract19$i = 0, $fname$i$i = 0, $hw$i = 0, $hw$i3 = 0, $i$01$i$i = 0, $i$01$i2$i = 0, $i$01$i5$i = 0, $not$$i$i = 0, $not$$i$i9 = 0, $p$0$i = 0, $s$01$i$i$i = 0, $s$01$i$i12$i = 0, $s$i = 0, $s$i4 = 0, $storemerge$i$i = 0;
 var $strlenfirst = 0, $t$018 = 0, $t$03$i$i$i = 0, $t$03$i$i10$i = 0, $t$116 = 0, $vararg_buffer = 0, $vararg_buffer12 = 0, $vararg_buffer17 = 0, $vararg_buffer3 = 0, $vararg_buffer8 = 0, $vararg_ptr1 = 0, $vararg_ptr11 = 0, $vararg_ptr15 = 0, $vararg_ptr16 = 0, $vararg_ptr2 = 0, $vararg_ptr20 = 0, $vararg_ptr21 = 0, $vararg_ptr6 = 0, $vararg_ptr7 = 0, dest = 0;
 var label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 688|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer17 = sp + 56|0;
 $vararg_buffer12 = sp + 40|0;
 $vararg_buffer8 = sp + 32|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $hw$i3 = sp + 584|0;
 $s$i4 = sp + 552|0;
 $fname$i$i = sp + 424|0;
 $hw$i = sp + 328|0;
 $s$i = sp + 72|0;
 $strlenfirst = HEAP8[1252699]|0; //@line 201 "load_pat.cpp"
 $0 = ($strlenfirst<<24>>24)==(0); //@line 1401 "load_pat.cpp"
 if ($0) {
  HEAP8[1252699] = 1; //@line 223 "load_pat.cpp"
 }
 $1 = (_strlen(1252699)|0); //@line 201 "load_pat.cpp"
 $2 = (($1) + 1)|0; //@line 1402 "load_pat.cpp"
 $3 = ((($c)) + 59684|0); //@line 1402 "load_pat.cpp"
 HEAP32[$3>>2] = $2; //@line 1402 "load_pat.cpp"
 $4 = (_strlen(1252699)|0); //@line 206 "load_pat.cpp"
 $5 = (($4) + 1)|0; //@line 1403 "load_pat.cpp"
 $6 = ((($c)) + 59688|0); //@line 1403 "load_pat.cpp"
 HEAP32[$6>>2] = $5; //@line 1403 "load_pat.cpp"
 $7 = ($5>>>0)>(1); //@line 1404 "load_pat.cpp"
 if ($7) {
  $8 = ((($c)) + 48640|0); //@line 1405 "load_pat.cpp"
  $$pre$i7 = ((($hw$i3)) + 8|0); //@line 1219 "load_pat.cpp"
  $9 = ((($hw$i3)) + 55|0); //@line 1216 "load_pat.cpp"
  $10 = ((($hw$i3)) + 12|0); //@line 1217 "load_pat.cpp"
  $11 = ((($hw$i3)) + 16|0); //@line 1218 "load_pat.cpp"
  $12 = ((($hw$i3)) + 37|0); //@line 1221 "load_pat.cpp"
  $13 = ((($hw$i3)) + 38|0); //@line 1222 "load_pat.cpp"
  $14 = ((($hw$i3)) + 39|0); //@line 1223 "load_pat.cpp"
  $15 = ((($hw$i3)) + 40|0); //@line 1224 "load_pat.cpp"
  $16 = ((($hw$i3)) + 41|0); //@line 1225 "load_pat.cpp"
  $17 = ((($hw$i3)) + 42|0); //@line 1226 "load_pat.cpp"
  $18 = ((($hw$i3)) + 43|0); //@line 1227 "load_pat.cpp"
  $19 = ((($hw$i3)) + 44|0); //@line 1228 "load_pat.cpp"
  $20 = ((($hw$i3)) + 45|0); //@line 1229 "load_pat.cpp"
  $21 = ((($hw$i3)) + 46|0); //@line 1230 "load_pat.cpp"
  $22 = ((($hw$i3)) + 47|0); //@line 1231 "load_pat.cpp"
  $23 = ((($hw$i3)) + 48|0); //@line 1232 "load_pat.cpp"
  $24 = ((($hw$i3)) + 64|0); //@line 1233 "load_pat.cpp"
  $25 = ((($s$i4)) + 31|0); //@line 1243 "load_pat.cpp"
  $26 = ((($s$i4)) + 11|0); //@line 1247 "load_pat.cpp"
  $t$018 = 1;
  while(1) {
   $37 = (__Znwj(632)|0); //@line 1405 "load_pat.cpp"
   $38 = (($8) + ($t$018<<2)|0); //@line 1405 "load_pat.cpp"
   HEAP32[$38>>2] = $37; //@line 1405 "load_pat.cpp"
   _memset(($37|0),0,632)|0; //@line 1406 "load_pat.cpp"
   $39 = ($t$018|0)<(191); //@line 211 "load_pat.cpp"
   if ($39) {
    $40 = (($t$018) + -1)|0; //@line 212 "load_pat.cpp"
    $41 = (1252699 + ($40)|0); //@line 212 "load_pat.cpp"
    $42 = HEAP8[$41>>0]|0; //@line 212 "load_pat.cpp"
    $43 = $42&255; //@line 212 "load_pat.cpp"
    $$0$i1 = $43;
   } else {
    $$0$i1 = 1;
   }
   dest=$s$i4; stop=dest+32|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 1211 "load_pat.cpp"
   $44 = (($$0$i1) + -1)|0; //@line 1212 "load_pat.cpp"
   $45 = (1253146 + ($44<<7)|0); //@line 489 "load_pat.cpp"
   $46 = HEAP8[$45>>0]|0; //@line 490 "load_pat.cpp"
   $47 = ($46<<24>>24)==(47); //@line 490 "load_pat.cpp"
   $48 = (((1253146 + ($44<<7)|0)) + 1|0); //@line 491 "load_pat.cpp"
   $$$i1$i5 = $47 ? $48 : $45; //@line 491 "load_pat.cpp"
   $49 = (_strrchr($$$i1$i5,58)|0); //@line 492 "load_pat.cpp"
   $50 = ($49|0)==(0|0); //@line 493 "load_pat.cpp"
   $51 = $47 ? 1282466 : 1252890; //@line 494 "load_pat.cpp"
   if ($50) {
    HEAP32[$vararg_buffer3>>2] = $51; //@line 498 "load_pat.cpp"
    $vararg_ptr6 = ((($vararg_buffer3)) + 4|0); //@line 498 "load_pat.cpp"
    HEAP32[$vararg_ptr6>>2] = 47; //@line 498 "load_pat.cpp"
    $vararg_ptr7 = ((($vararg_buffer3)) + 8|0); //@line 498 "load_pat.cpp"
    HEAP32[$vararg_ptr7>>2] = $$$i1$i5; //@line 498 "load_pat.cpp"
    (_sprintf($fname$i$i,527886,$vararg_buffer3)|0); //@line 498 "load_pat.cpp"
   } else {
    HEAP32[$vararg_buffer>>2] = $51; //@line 494 "load_pat.cpp"
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 494 "load_pat.cpp"
    HEAP32[$vararg_ptr1>>2] = 47; //@line 494 "load_pat.cpp"
    $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 494 "load_pat.cpp"
    HEAP32[$vararg_ptr2>>2] = $$$i1$i5; //@line 494 "load_pat.cpp"
    (_sprintf($fname$i$i,527874,$vararg_buffer)|0); //@line 494 "load_pat.cpp"
    $52 = (_strrchr($fname$i$i,58)|0); //@line 495 "load_pat.cpp"
    ;HEAP8[$52>>0]=HEAP8[527881>>0]|0;HEAP8[$52+1>>0]=HEAP8[527881+1>>0]|0;HEAP8[$52+2>>0]=HEAP8[527881+2>>0]|0;HEAP8[$52+3>>0]=HEAP8[527881+3>>0]|0;HEAP8[$52+4>>0]=HEAP8[527881+4>>0]|0; //@line 495 "load_pat.cpp"
   }
   $53 = (_fopen($fname$i$i,527143)|0); //@line 632 "load_pat.cpp"
   $54 = ($53|0)==(0|0); //@line 633 "load_pat.cpp"
   if ($54) {
    label = 17;
   } else {
    (_fseek($53,0,2)|0); //@line 378 "load_pat.cpp"
    $55 = (_ftell($53)|0); //@line 379 "load_pat.cpp"
    (_fseek($53,0,0)|0); //@line 380 "load_pat.cpp"
    __ZL19pat_read_waveheaderP8_IO_FILEP10WaveHeaderi($53,$hw$i3); //@line 636 "load_pat.cpp"
    (_fclose($53)|0); //@line 637 "load_pat.cpp"
    $56 = HEAPU8[$$pre$i7>>0]|(HEAPU8[$$pre$i7+1>>0]<<8)|(HEAPU8[$$pre$i7+2>>0]<<16)|(HEAPU8[$$pre$i7+3>>0]<<24); //@line 638 "load_pat.cpp"
    $not$$i$i9 = ($56>>>0)>($55>>>0); //@line 639 "load_pat.cpp"
    if ($not$$i$i9) {
     label = 17;
    } else {
     __ZL15pat_setpat_instP10WaveHeaderP17_INSTRUMENTHEADERi($hw$i3,$37,$t$018); //@line 1213 "load_pat.cpp"
    }
   }
   if ((label|0) == 17) {
    label = 0;
    HEAP8[$9>>0] = 101; //@line 1216 "load_pat.cpp"
    HEAP8[$10>>0]=0&255;HEAP8[$10+1>>0]=(0>>8)&255;HEAP8[$10+2>>0]=(0>>16)&255;HEAP8[$10+3>>0]=0>>24; //@line 1217 "load_pat.cpp"
    HEAP8[$11>>0]=30000&255;HEAP8[$11+1>>0]=(30000>>8)&255;HEAP8[$11+2>>0]=(30000>>16)&255;HEAP8[$11+3>>0]=30000>>24; //@line 1218 "load_pat.cpp"
    HEAP8[$$pre$i7>>0]=30000&255;HEAP8[$$pre$i7+1>>0]=(30000>>8)&255;HEAP8[$$pre$i7+2>>0]=(30000>>16)&255;HEAP8[$$pre$i7+3>>0]=30000>>24; //@line 1219 "load_pat.cpp"
    HEAP8[$12>>0] = 63; //@line 1221 "load_pat.cpp"
    HEAP8[$13>>0] = 63; //@line 1222 "load_pat.cpp"
    HEAP8[$14>>0] = 63; //@line 1223 "load_pat.cpp"
    HEAP8[$15>>0] = -56; //@line 1224 "load_pat.cpp"
    HEAP8[$16>>0] = 63; //@line 1225 "load_pat.cpp"
    HEAP8[$17>>0] = 63; //@line 1226 "load_pat.cpp"
    HEAP8[$18>>0] = -10; //@line 1227 "load_pat.cpp"
    HEAP8[$19>>0] = -10; //@line 1228 "load_pat.cpp"
    HEAP8[$20>>0] = -10; //@line 1229 "load_pat.cpp"
    HEAP8[$21>>0] = 0; //@line 1230 "load_pat.cpp"
    HEAP8[$22>>0] = 0; //@line 1231 "load_pat.cpp"
    HEAP8[$23>>0] = 0; //@line 1232 "load_pat.cpp"
    (_strncpy($24,$45,32)|0); //@line 1233 "load_pat.cpp"
    __ZL15pat_setpat_instP10WaveHeaderP17_INSTRUMENTHEADERi($hw$i3,$37,$t$018); //@line 1234 "load_pat.cpp"
   }
   $57 = HEAP8[$24>>0]|0; //@line 1236 "load_pat.cpp"
   $58 = ($57<<24>>24)==(0); //@line 1236 "load_pat.cpp"
   if ($58) {
    (_strncpy($s$i4,$45,32)|0); //@line 1239 "load_pat.cpp"
   } else {
    (_strncpy($s$i4,$24,32)|0); //@line 1237 "load_pat.cpp"
   }
   HEAP8[$25>>0] = 0; //@line 1243 "load_pat.cpp"
   $59 = ((($37)) + 585|0); //@line 1244 "load_pat.cpp"
   dest=$59; stop=dest+32|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 1244 "load_pat.cpp"
   (_strcpy($59,$s$i4)|0); //@line 1245 "load_pat.cpp"
   (_strncpy($s$i4,$45,12)|0); //@line 1246 "load_pat.cpp"
   HEAP8[$26>>0] = 0; //@line 1247 "load_pat.cpp"
   $60 = ((($37)) + 617|0); //@line 1248 "load_pat.cpp"
   dest=$60; stop=dest+12|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 1248 "load_pat.cpp"
   (_strcpy($60,$s$i4)|0); //@line 1249 "load_pat.cpp"
   $61 = (($t$018) + 1)|0; //@line 1404 "load_pat.cpp"
   $62 = HEAP32[$6>>2]|0; //@line 1404 "load_pat.cpp"
   $63 = ($61>>>0)<($62>>>0); //@line 1404 "load_pat.cpp"
   if ($63) {
    $t$018 = $61;
   } else {
    $$lcssa = $62;
    break;
   }
  }
  $$pre24 = HEAP32[$3>>2]|0; //@line 1409 "load_pat.cpp"
  $221 = $$lcssa;$28 = $$pre24;
 } else {
  $221 = $5;$28 = $2;
 }
 $27 = ($28>>>0)>(1); //@line 1409 "load_pat.cpp"
 $29 = ((($c)) + 33280|0); //@line 1410 "load_pat.cpp"
 if (!($27)) {
  $214 = $221;
  $211 = (__Znwj(632)|0); //@line 1414 "load_pat.cpp"
  $212 = ((($c)) + 48640|0); //@line 1414 "load_pat.cpp"
  HEAP32[$212>>2] = $211; //@line 1414 "load_pat.cpp"
  $213 = (($214) + -1)|0; //@line 1413 "load_pat.cpp"
  $215 = (($212) + ($213<<2)|0); //@line 1415 "load_pat.cpp"
  $216 = HEAP32[$215>>2]|0; //@line 1415 "load_pat.cpp"
  _memcpy(($211|0),($216|0),632)|0; //@line 1415 "load_pat.cpp"
  $217 = ((($211)) + 585|0); //@line 1416 "load_pat.cpp"
  dest=$217; stop=dest+32|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 1416 "load_pat.cpp"
  (_strncpy($217,527914,32)|0); //@line 1417 "load_pat.cpp"
  $218 = HEAP32[$3>>2]|0; //@line 1418 "load_pat.cpp"
  $219 = (($218) + -1)|0; //@line 1418 "load_pat.cpp"
  $220 = (($29) + ($219<<6)|0); //@line 1419 "load_pat.cpp"
  dest=$29; src=$220; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1419 "load_pat.cpp"
  STACKTOP = sp;return 1; //@line 1422 "load_pat.cpp"
 }
 $30 = ((($s$i)) + 31|0); //@line 1332 "load_pat.cpp"
 $31 = ((($hw$i)) + 8|0); //@line 638 "load_pat.cpp"
 $32 = ((($hw$i)) + 20|0); //@line 1279 "load_pat.cpp"
 $33 = ((($hw$i)) + 30|0); //@line 1279 "load_pat.cpp"
 $34 = ((($hw$i)) + 12|0); //@line 1281 "load_pat.cpp"
 $35 = ((($hw$i)) + 16|0); //@line 1282 "load_pat.cpp"
 $36 = ((($hw$i)) + 55|0); //@line 1284 "load_pat.cpp"
 $t$116 = 1;
 while(1) {
  $64 = (($29) + ($t$116<<6)|0); //@line 1410 "load_pat.cpp"
  $65 = ($t$116|0)<(191); //@line 211 "load_pat.cpp"
  $66 = (($t$116) + -1)|0; //@line 212 "load_pat.cpp"
  if ($65) {
   $67 = (1252699 + ($66)|0); //@line 212 "load_pat.cpp"
   $68 = HEAP8[$67>>0]|0; //@line 212 "load_pat.cpp"
   $69 = $68&255; //@line 212 "load_pat.cpp"
   $$0$i11 = $69;
  } else {
   $$0$i11 = 1;
  }
  $70 = (($$0$i11) + -1)|0; //@line 1307 "load_pat.cpp"
  $71 = (1253146 + ($70<<7)|0); //@line 1307 "load_pat.cpp"
  HEAP32[$vararg_buffer8>>2] = $66; //@line 1307 "load_pat.cpp"
  $vararg_ptr11 = ((($vararg_buffer8)) + 4|0); //@line 1307 "load_pat.cpp"
  HEAP32[$vararg_ptr11>>2] = $71; //@line 1307 "load_pat.cpp"
  (_sprintf($s$i,527904,$vararg_buffer8)|0); //@line 1307 "load_pat.cpp"
  HEAP8[$30>>0] = 0; //@line 1332 "load_pat.cpp"
  $72 = (((($c)) + 59804|0) + ($t$116<<5)|0); //@line 1333 "load_pat.cpp"
  dest=$72; stop=dest+32|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 1333 "load_pat.cpp"
  (_strcpy($72,$s$i)|0); //@line 1334 "load_pat.cpp"
  $73 = (((($29) + ($t$116<<6)|0)) + 32|0); //@line 1335 "load_pat.cpp"
  HEAP16[$73>>1] = 64; //@line 1335 "load_pat.cpp"
  $74 = (((($29) + ($t$116<<6)|0)) + 28|0); //@line 1336 "load_pat.cpp"
  HEAP16[$74>>1] = 128; //@line 1336 "load_pat.cpp"
  $75 = (((($29) + ($t$116<<6)|0)) + 34|0); //@line 1337 "load_pat.cpp"
  HEAP16[$75>>1] = 1; //@line 1337 "load_pat.cpp"
  $76 = HEAP8[$71>>0]|0; //@line 490 "load_pat.cpp"
  $77 = ($76<<24>>24)==(47); //@line 490 "load_pat.cpp"
  $78 = (((1253146 + ($70<<7)|0)) + 1|0); //@line 491 "load_pat.cpp"
  $$$i15$i = $77 ? $78 : $71; //@line 491 "load_pat.cpp"
  $79 = (_strrchr($$$i15$i,58)|0); //@line 492 "load_pat.cpp"
  $80 = ($79|0)==(0|0); //@line 493 "load_pat.cpp"
  $81 = $77 ? 1282466 : 1252890; //@line 494 "load_pat.cpp"
  if ($80) {
   HEAP32[$vararg_buffer17>>2] = $81; //@line 498 "load_pat.cpp"
   $vararg_ptr20 = ((($vararg_buffer17)) + 4|0); //@line 498 "load_pat.cpp"
   HEAP32[$vararg_ptr20>>2] = 47; //@line 498 "load_pat.cpp"
   $vararg_ptr21 = ((($vararg_buffer17)) + 8|0); //@line 498 "load_pat.cpp"
   HEAP32[$vararg_ptr21>>2] = $$$i15$i; //@line 498 "load_pat.cpp"
   (_sprintf($fname$i$i,527886,$vararg_buffer17)|0); //@line 498 "load_pat.cpp"
  } else {
   HEAP32[$vararg_buffer12>>2] = $81; //@line 494 "load_pat.cpp"
   $vararg_ptr15 = ((($vararg_buffer12)) + 4|0); //@line 494 "load_pat.cpp"
   HEAP32[$vararg_ptr15>>2] = 47; //@line 494 "load_pat.cpp"
   $vararg_ptr16 = ((($vararg_buffer12)) + 8|0); //@line 494 "load_pat.cpp"
   HEAP32[$vararg_ptr16>>2] = $$$i15$i; //@line 494 "load_pat.cpp"
   (_sprintf($fname$i$i,527874,$vararg_buffer12)|0); //@line 494 "load_pat.cpp"
   $82 = (_strrchr($fname$i$i,58)|0); //@line 495 "load_pat.cpp"
   ;HEAP8[$82>>0]=HEAP8[527881>>0]|0;HEAP8[$82+1>>0]=HEAP8[527881+1>>0]|0;HEAP8[$82+2>>0]=HEAP8[527881+2>>0]|0;HEAP8[$82+3>>0]=HEAP8[527881+3>>0]|0;HEAP8[$82+4>>0]=HEAP8[527881+4>>0]|0; //@line 495 "load_pat.cpp"
  }
  $83 = (_fopen($fname$i$i,527143)|0); //@line 632 "load_pat.cpp"
  $84 = ($83|0)==(0|0); //@line 633 "load_pat.cpp"
  if ($84) {
   label = 55;
  } else {
   (_fseek($83,0,2)|0); //@line 378 "load_pat.cpp"
   $85 = (_ftell($83)|0); //@line 379 "load_pat.cpp"
   (_fseek($83,0,0)|0); //@line 380 "load_pat.cpp"
   __ZL19pat_read_waveheaderP8_IO_FILEP10WaveHeaderi($83,$hw$i); //@line 636 "load_pat.cpp"
   (_fclose($83)|0); //@line 637 "load_pat.cpp"
   $86 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 638 "load_pat.cpp"
   $not$$i$i = ($86>>>0)>($85>>>0); //@line 639 "load_pat.cpp"
   if ($not$$i$i) {
    label = 55;
   } else {
    $87 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8); //@line 1279 "load_pat.cpp"
    $88 = HEAPU8[$33>>0]|(HEAPU8[$33+1>>0]<<8)|(HEAPU8[$33+2>>0]<<16)|(HEAPU8[$33+3>>0]<<24); //@line 1279 "load_pat.cpp"
    $89 = (+($87&65535)); //@line 1041 "load_pat.cpp"
    $90 = $89 * 52325.100000000006; //@line 1042 "load_pat.cpp"
    $91 = (+($88>>>0)); //@line 1043 "load_pat.cpp"
    $92 = $91 * 0.40000000000000002; //@line 1043 "load_pat.cpp"
    $93 = $90 / $92; //@line 1044 "load_pat.cpp"
    $94 = $93 + 0.5; //@line 1045 "load_pat.cpp"
    $95 = (~~(($94))>>>0); //@line 1045 "load_pat.cpp"
    $96 = (((($29) + ($t$116<<6)|0)) + 24|0); //@line 1279 "load_pat.cpp"
    HEAP32[$96>>2] = $95; //@line 1279 "load_pat.cpp"
    HEAP32[$64>>2] = $86; //@line 1280 "load_pat.cpp"
    $97 = HEAPU8[$34>>0]|(HEAPU8[$34+1>>0]<<8)|(HEAPU8[$34+2>>0]<<16)|(HEAPU8[$34+3>>0]<<24); //@line 1281 "load_pat.cpp"
    $98 = (((($29) + ($t$116<<6)|0)) + 4|0); //@line 1281 "load_pat.cpp"
    HEAP32[$98>>2] = $97; //@line 1281 "load_pat.cpp"
    $99 = HEAPU8[$35>>0]|(HEAPU8[$35+1>>0]<<8)|(HEAPU8[$35+2>>0]<<16)|(HEAPU8[$35+3>>0]<<24); //@line 1282 "load_pat.cpp"
    $100 = (((($29) + ($t$116<<6)|0)) + 8|0); //@line 1282 "load_pat.cpp"
    HEAP32[$100>>2] = $99; //@line 1282 "load_pat.cpp"
    $101 = (((($29) + ($t$116<<6)|0)) + 30|0); //@line 1283 "load_pat.cpp"
    HEAP16[$101>>1] = 256; //@line 1283 "load_pat.cpp"
    $102 = HEAP8[$36>>0]|0; //@line 1284 "load_pat.cpp"
    $103 = $102 & 1; //@line 1284 "load_pat.cpp"
    $104 = ($103<<24>>24)==(0); //@line 1284 "load_pat.cpp"
    if (!($104)) {
     $105 = $86 >>> 1; //@line 1285 "load_pat.cpp"
     HEAP32[$64>>2] = $105; //@line 1285 "load_pat.cpp"
     $106 = $97 >>> 1; //@line 1286 "load_pat.cpp"
     HEAP32[$98>>2] = $106; //@line 1286 "load_pat.cpp"
     $107 = $99 >>> 1; //@line 1287 "load_pat.cpp"
     HEAP32[$100>>2] = $107; //@line 1287 "load_pat.cpp"
    }
    $108 = $102 & 4; //@line 1289 "load_pat.cpp"
    $109 = ($108<<24>>24)==(0); //@line 1289 "load_pat.cpp"
    $$pre$i = HEAP16[$75>>1]|0; //@line 1341 "load_pat.cpp"
    if ($109) {
     $extract$i = ($$pre$i&65535) >>> 1; //@line 1410 "load_pat.cpp"
     $extract$t$i = $extract$i&255; //@line 1410 "load_pat.cpp"
     $$off1$i = $extract$t$i;
    } else {
     $110 = $$pre$i&65535; //@line 1290 "load_pat.cpp"
     $111 = $102 & 8; //@line 1291 "load_pat.cpp"
     $112 = $111&255; //@line 1291 "load_pat.cpp"
     $113 = $112 << 1; //@line 1291 "load_pat.cpp"
     $114 = $110 | $113; //@line 1291 "load_pat.cpp"
     $$$i1$i = $114 | 2; //@line 1291 "load_pat.cpp"
     $storemerge$i$i = $$$i1$i&65535; //@line 1291 "load_pat.cpp"
     HEAP16[$75>>1] = $storemerge$i$i; //@line 1290 "load_pat.cpp"
     $115 = $102 & 32; //@line 1292 "load_pat.cpp"
     $116 = ($115<<24>>24)==(0); //@line 1292 "load_pat.cpp"
     $extract17$i = ($storemerge$i$i&65535) >>> 1; //@line 1410 "load_pat.cpp"
     $extract$t18$i = $extract17$i&255; //@line 1410 "load_pat.cpp"
     if ($116) {
      $$off1$i = $extract$t18$i;
     } else {
      $117 = $114 | 10; //@line 1292 "load_pat.cpp"
      $118 = $117&65535; //@line 1292 "load_pat.cpp"
      HEAP16[$75>>1] = $118; //@line 1292 "load_pat.cpp"
      $extract19$i = ($118&65535) >>> 1; //@line 1410 "load_pat.cpp"
      $extract$t20$i = $extract19$i&255; //@line 1410 "load_pat.cpp"
      $$off1$i = $extract$t20$i;
     }
    }
    $119 = $$off1$i & 1; //@line 1341 "load_pat.cpp"
    $120 = (1252508 + ($66)|0); //@line 1341 "load_pat.cpp"
    HEAP8[$120>>0] = $119; //@line 1341 "load_pat.cpp"
    if ($104) {
     $122 = $86 << 1; //@line 1343 "load_pat.cpp"
     $123 = (_malloc($122)|0); //@line 1343 "load_pat.cpp"
     $p$0$i = $123;
    } else {
     $121 = (_malloc($86)|0); //@line 1342 "load_pat.cpp"
     $p$0$i = $121;
    }
    $124 = ($p$0$i|0)==(0|0); //@line 1344 "load_pat.cpp"
    if (!($124)) {
     if ($104) {
      $145 = ($$0$i11>>>0)<(192); //@line 804 "load_pat.cpp"
      if ($145) {
       __ZL11pat_readpatiPci($70,$p$0$i,$86); //@line 804 "load_pat.cpp"
       label = 50;
      } else {
       $146 = (($$0$i11) + -192)|0; //@line 806 "load_pat.cpp"
       $147 = (($146|0) % 3)&-1; //@line 806 "load_pat.cpp"
       $148 = (47172 + ($147<<2)|0); //@line 806 "load_pat.cpp"
       $149 = HEAP32[$148>>2]|0; //@line 806 "load_pat.cpp"
       $150 = ($86|0)>(0); //@line 807 "load_pat.cpp"
       if ($150) {
        $i$01$i2$i = 0;
        while(1) {
         $151 = (+FUNCTION_TABLE_di[$149 & 4095]($i$01$i2$i)); //@line 808 "load_pat.cpp"
         $152 = $151; //@line 808 "load_pat.cpp"
         $153 = $152 * 120.0; //@line 808 "load_pat.cpp"
         $154 = (~~(($153))); //@line 808 "load_pat.cpp"
         $155 = $154 << 24 >> 24; //@line 808 "load_pat.cpp"
         $156 = (($p$0$i) + ($i$01$i2$i<<1)|0); //@line 808 "load_pat.cpp"
         HEAP16[$156>>1] = $155; //@line 808 "load_pat.cpp"
         $157 = (($i$01$i2$i) + 1)|0; //@line 807 "load_pat.cpp"
         $exitcond$i3$i = ($157|0)==($86|0); //@line 807 "load_pat.cpp"
         if ($exitcond$i3$i) {
          label = 50;
          break;
         } else {
          $i$01$i2$i = $157;
         }
        }
       }
      }
      if ((label|0) == 50) {
       label = 0;
       $158 = ($86|0)>(0); //@line 787 "load_pat.cpp"
       if ($158) {
        $159 = (($p$0$i) + ($86<<1)|0); //@line 786 "load_pat.cpp"
        $160 = (($p$0$i) + ($86)|0); //@line 785 "load_pat.cpp"
        $d$02$i$i$i = $159;$s$01$i$i$i = $160;$t$03$i$i$i = 0;
        while(1) {
         $161 = ((($s$01$i$i$i)) + -1|0); //@line 788 "load_pat.cpp"
         $162 = ((($d$02$i$i$i)) + -2|0); //@line 789 "load_pat.cpp"
         $163 = HEAP8[$161>>0]|0; //@line 790 "load_pat.cpp"
         $164 = $163 << 24 >> 24; //@line 790 "load_pat.cpp"
         $165 = $164 << 8; //@line 790 "load_pat.cpp"
         $166 = $165&65535; //@line 790 "load_pat.cpp"
         HEAP16[$162>>1] = $166; //@line 790 "load_pat.cpp"
         $167 = (($t$03$i$i$i) + 1)|0; //@line 787 "load_pat.cpp"
         $exitcond$i$i$i = ($167|0)==($86|0); //@line 787 "load_pat.cpp"
         if ($exitcond$i$i$i) {
          break;
         } else {
          $d$02$i$i$i = $162;$s$01$i$i$i = $161;$t$03$i$i$i = $167;
         }
        }
       }
      }
      $168 = HEAP8[$36>>0]|0; //@line 1351 "load_pat.cpp"
      $169 = $168 & 2; //@line 1351 "load_pat.cpp"
      $170 = $169&255; //@line 1351 "load_pat.cpp"
      $171 = $170 >>> 1; //@line 1351 "load_pat.cpp"
      $172 = (($171) + 5)|0; //@line 1351 "load_pat.cpp"
      $173 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 1351 "load_pat.cpp"
      $174 = $173 << 1; //@line 1351 "load_pat.cpp"
      (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($c,$64,$172,$p$0$i,$174)|0); //@line 1351 "load_pat.cpp"
     } else {
      $125 = $86 >>> 1; //@line 1346 "load_pat.cpp"
      $126 = ($$0$i11>>>0)<(192); //@line 767 "load_pat.cpp"
      if ($126) {
       $127 = $125 << 1; //@line 767 "load_pat.cpp"
       __ZL11pat_readpatiPci($70,$p$0$i,$127); //@line 767 "load_pat.cpp"
      } else {
       $128 = (($$0$i11) + -192)|0; //@line 769 "load_pat.cpp"
       $129 = (($128|0) % 3)&-1; //@line 769 "load_pat.cpp"
       $130 = (47172 + ($129<<2)|0); //@line 769 "load_pat.cpp"
       $131 = HEAP32[$130>>2]|0; //@line 769 "load_pat.cpp"
       $132 = ($125|0)==(0); //@line 770 "load_pat.cpp"
       if (!($132)) {
        $i$01$i$i = 0;
        while(1) {
         $133 = (+FUNCTION_TABLE_di[$131 & 4095]($i$01$i$i)); //@line 771 "load_pat.cpp"
         $134 = $133; //@line 771 "load_pat.cpp"
         $135 = $134 * 32000.0; //@line 771 "load_pat.cpp"
         $136 = (~~(($135))); //@line 771 "load_pat.cpp"
         $137 = (($p$0$i) + ($i$01$i$i<<1)|0); //@line 771 "load_pat.cpp"
         HEAP16[$137>>1] = $136; //@line 771 "load_pat.cpp"
         $138 = (($i$01$i$i) + 1)|0; //@line 770 "load_pat.cpp"
         $exitcond$i$i = ($138|0)==($125|0); //@line 770 "load_pat.cpp"
         if ($exitcond$i$i) {
          break;
         } else {
          $i$01$i$i = $138;
         }
        }
       }
      }
      $139 = HEAP8[$36>>0]|0; //@line 1347 "load_pat.cpp"
      $140 = $139 & 2; //@line 1347 "load_pat.cpp"
      $141 = $140&255; //@line 1347 "load_pat.cpp"
      $142 = $141 >>> 1; //@line 1347 "load_pat.cpp"
      $143 = (($142) + 5)|0; //@line 1347 "load_pat.cpp"
      $144 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 1347 "load_pat.cpp"
      (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($c,$64,$143,$p$0$i,$144)|0); //@line 1347 "load_pat.cpp"
     }
     _free($p$0$i); //@line 1353 "load_pat.cpp"
    }
   }
  }
  if ((label|0) == 55) {
   label = 0;
   $175 = (((($29) + ($t$116<<6)|0)) + 24|0); //@line 1358 "load_pat.cpp"
   HEAP32[$175>>2] = 8363; //@line 1358 "load_pat.cpp"
   HEAP32[$64>>2] = 30000; //@line 1359 "load_pat.cpp"
   $176 = (((($29) + ($t$116<<6)|0)) + 4|0); //@line 1360 "load_pat.cpp"
   HEAP32[$176>>2] = 0; //@line 1360 "load_pat.cpp"
   $177 = (((($29) + ($t$116<<6)|0)) + 8|0); //@line 1361 "load_pat.cpp"
   HEAP32[$177>>2] = 30000; //@line 1361 "load_pat.cpp"
   $178 = (((($29) + ($t$116<<6)|0)) + 30|0); //@line 1362 "load_pat.cpp"
   HEAP16[$178>>1] = 256; //@line 1362 "load_pat.cpp"
   $179 = HEAP16[$75>>1]|0; //@line 1363 "load_pat.cpp"
   $180 = $179&65535; //@line 1363 "load_pat.cpp"
   $181 = $180 | 3; //@line 1364 "load_pat.cpp"
   $182 = $181&65535; //@line 1364 "load_pat.cpp"
   HEAP16[$75>>1] = $182; //@line 1364 "load_pat.cpp"
   $183 = (_malloc(60000)|0); //@line 1365 "load_pat.cpp"
   $184 = ($183|0)==(0|0); //@line 1366 "load_pat.cpp"
   if (!($184)) {
    $185 = (($t$116) + 190)|0; //@line 1367 "load_pat.cpp"
    $186 = ($185|0)<(191); //@line 804 "load_pat.cpp"
    if ($186) {
     __ZL11pat_readpatiPci($185,$183,30000); //@line 804 "load_pat.cpp"
    } else {
     $187 = (($66|0) % 3)&-1; //@line 806 "load_pat.cpp"
     $188 = (47172 + ($187<<2)|0); //@line 806 "load_pat.cpp"
     $189 = HEAP32[$188>>2]|0; //@line 806 "load_pat.cpp"
     $i$01$i5$i = 0;
     while(1) {
      $190 = (+FUNCTION_TABLE_di[$189 & 4095]($i$01$i5$i)); //@line 808 "load_pat.cpp"
      $191 = $190; //@line 808 "load_pat.cpp"
      $192 = $191 * 120.0; //@line 808 "load_pat.cpp"
      $193 = (~~(($192))); //@line 808 "load_pat.cpp"
      $194 = $193 << 24 >> 24; //@line 808 "load_pat.cpp"
      $195 = (($183) + ($i$01$i5$i<<1)|0); //@line 808 "load_pat.cpp"
      HEAP16[$195>>1] = $194; //@line 808 "load_pat.cpp"
      $196 = (($i$01$i5$i) + 1)|0; //@line 807 "load_pat.cpp"
      $exitcond$i6$i = ($196|0)==(30000); //@line 807 "load_pat.cpp"
      if ($exitcond$i6$i) {
       break;
      } else {
       $i$01$i5$i = $196;
      }
     }
    }
    $197 = ((($183)) + 60000|0); //@line 786 "load_pat.cpp"
    $198 = ((($183)) + 30000|0); //@line 785 "load_pat.cpp"
    $d$02$i$i11$i = $197;$s$01$i$i12$i = $198;$t$03$i$i10$i = 0;
    while(1) {
     $199 = ((($s$01$i$i12$i)) + -1|0); //@line 788 "load_pat.cpp"
     $200 = ((($d$02$i$i11$i)) + -2|0); //@line 789 "load_pat.cpp"
     $201 = HEAP8[$199>>0]|0; //@line 790 "load_pat.cpp"
     $202 = $201 << 24 >> 24; //@line 790 "load_pat.cpp"
     $203 = $202 << 8; //@line 790 "load_pat.cpp"
     $204 = $203&65535; //@line 790 "load_pat.cpp"
     HEAP16[$200>>1] = $204; //@line 790 "load_pat.cpp"
     $205 = (($t$03$i$i10$i) + 1)|0; //@line 787 "load_pat.cpp"
     $exitcond$i$i13$i = ($205|0)==(30000); //@line 787 "load_pat.cpp"
     if ($exitcond$i$i13$i) {
      break;
     } else {
      $d$02$i$i11$i = $200;$s$01$i$i12$i = $199;$t$03$i$i10$i = $205;
     }
    }
    $206 = HEAP32[$64>>2]|0; //@line 1368 "load_pat.cpp"
    $207 = $206 << 1; //@line 1368 "load_pat.cpp"
    (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($c,$64,5,$183,$207)|0); //@line 1368 "load_pat.cpp"
    _free($183); //@line 1369 "load_pat.cpp"
   }
  }
  $208 = (($t$116) + 1)|0; //@line 1409 "load_pat.cpp"
  $209 = HEAP32[$3>>2]|0; //@line 1409 "load_pat.cpp"
  $210 = ($208>>>0)<($209>>>0); //@line 1409 "load_pat.cpp"
  if ($210) {
   $t$116 = $208;
  } else {
   break;
  }
 }
 $$pre25 = HEAP32[$6>>2]|0; //@line 1413 "load_pat.cpp"
 $214 = $$pre25;
 $211 = (__Znwj(632)|0); //@line 1414 "load_pat.cpp"
 $212 = ((($c)) + 48640|0); //@line 1414 "load_pat.cpp"
 HEAP32[$212>>2] = $211; //@line 1414 "load_pat.cpp"
 $213 = (($214) + -1)|0; //@line 1413 "load_pat.cpp"
 $215 = (($212) + ($213<<2)|0); //@line 1415 "load_pat.cpp"
 $216 = HEAP32[$215>>2]|0; //@line 1415 "load_pat.cpp"
 _memcpy(($211|0),($216|0),632)|0; //@line 1415 "load_pat.cpp"
 $217 = ((($211)) + 585|0); //@line 1416 "load_pat.cpp"
 dest=$217; stop=dest+32|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 1416 "load_pat.cpp"
 (_strncpy($217,527914,32)|0); //@line 1417 "load_pat.cpp"
 $218 = HEAP32[$3>>2]|0; //@line 1418 "load_pat.cpp"
 $219 = (($218) + -1)|0; //@line 1418 "load_pat.cpp"
 $220 = (($29) + ($219<<6)|0); //@line 1419 "load_pat.cpp"
 dest=$29; src=$220; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1419 "load_pat.cpp"
 STACKTOP = sp;return 1; //@line 1422 "load_pat.cpp"
}
function __ZL19pat_read_waveheaderP8_IO_FILEP10WaveHeaderi($mmpat,$hw) {
 $mmpat = $mmpat|0;
 $hw = $hw|0;
 var $$bestpos$0 = 0, $$bestpos$0$lcssa = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $bestfreq$02 = 0, $bestpos$01 = 0, $freqdist$0 = 0, $freqdist$0$bestfreq$0 = 0, $hl = 0, $i$13 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $hl = sp;
 (_fseek($mmpat,192,0)|0); //@line 512 "load_pat.cpp"
 (_fread($hl,47,1,$mmpat)|0); //@line 513 "load_pat.cpp"
 $0 = ((($hl)) + 6|0); //@line 539 "load_pat.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 539 "load_pat.cpp"
 $2 = ($1&255)>(1); //@line 539 "load_pat.cpp"
 if ($2) {
  $3 = ((($hw)) + 30|0); //@line 552 "load_pat.cpp"
  $4 = ((($hw)) + 8|0); //@line 560 "load_pat.cpp"
  $5 = $1&255; //@line 549 "load_pat.cpp"
  $bestfreq$02 = 523251000;$bestpos$01 = 0;$i$13 = 0;
  while(1) {
   $6 = (_ftell($mmpat)|0); //@line 550 "load_pat.cpp"
   (_fread($hw,96,1,$mmpat)|0); //@line 551 "load_pat.cpp"
   $7 = HEAPU8[$3>>0]|(HEAPU8[$3+1>>0]<<8)|(HEAPU8[$3+2>>0]<<16)|(HEAPU8[$3+3>>0]<<24); //@line 552 "load_pat.cpp"
   $8 = ($7>>>0)>(523251); //@line 552 "load_pat.cpp"
   $9 = (($7) + -523251)|0; //@line 553 "load_pat.cpp"
   $10 = (523251 - ($7))|0; //@line 555 "load_pat.cpp"
   $11 = $10 << 1; //@line 555 "load_pat.cpp"
   $freqdist$0 = $8 ? $9 : $11; //@line 552 "load_pat.cpp"
   $12 = ($freqdist$0>>>0)<($bestfreq$02>>>0); //@line 556 "load_pat.cpp"
   $$bestpos$0 = $12 ? $6 : $bestpos$01; //@line 556 "load_pat.cpp"
   $freqdist$0$bestfreq$0 = $12 ? $freqdist$0 : $bestfreq$02; //@line 556 "load_pat.cpp"
   $13 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8)|(HEAPU8[$4+2>>0]<<16)|(HEAPU8[$4+3>>0]<<24); //@line 560 "load_pat.cpp"
   (_fseek($mmpat,$13,1)|0); //@line 560 "load_pat.cpp"
   $14 = (($i$13) + 1)|0; //@line 549 "load_pat.cpp"
   $15 = ($14|0)<($5|0); //@line 549 "load_pat.cpp"
   if ($15) {
    $bestfreq$02 = $freqdist$0$bestfreq$0;$bestpos$01 = $$bestpos$0;$i$13 = $14;
   } else {
    $$bestpos$0$lcssa = $$bestpos$0;
    break;
   }
  }
  (_fseek($mmpat,$$bestpos$0$lcssa,0)|0); //@line 562 "load_pat.cpp"
 }
 (_fread($hw,96,1,$mmpat)|0); //@line 565 "load_pat.cpp"
 $16 = ((($hw)) + 64|0); //@line 566 "load_pat.cpp"
 $17 = ((($hl)) + 7|0); //@line 566 "load_pat.cpp"
 (_strncpy($16,$17,36)|0); //@line 566 "load_pat.cpp"
 $18 = ((($hw)) + 12|0); //@line 567 "load_pat.cpp"
 $19 = HEAPU8[$18>>0]|(HEAPU8[$18+1>>0]<<8)|(HEAPU8[$18+2>>0]<<16)|(HEAPU8[$18+3>>0]<<24); //@line 567 "load_pat.cpp"
 $20 = ((($hw)) + 8|0); //@line 567 "load_pat.cpp"
 $21 = HEAPU8[$20>>0]|(HEAPU8[$20+1>>0]<<8)|(HEAPU8[$20+2>>0]<<16)|(HEAPU8[$20+3>>0]<<24); //@line 567 "load_pat.cpp"
 $22 = ($19>>>0)<($21>>>0); //@line 567 "load_pat.cpp"
 if (!($22)) {
  HEAP8[$18>>0]=0&255;HEAP8[$18+1>>0]=(0>>8)&255;HEAP8[$18+2>>0]=(0>>16)&255;HEAP8[$18+3>>0]=0>>24; //@line 568 "load_pat.cpp"
  $23 = ((($hw)) + 16|0); //@line 569 "load_pat.cpp"
  HEAP8[$23>>0]=0&255;HEAP8[$23+1>>0]=(0>>8)&255;HEAP8[$23+2>>0]=(0>>16)&255;HEAP8[$23+3>>0]=0>>24; //@line 569 "load_pat.cpp"
  $24 = ((($hw)) + 55|0); //@line 570 "load_pat.cpp"
  $25 = HEAP8[$24>>0]|0; //@line 570 "load_pat.cpp"
  $26 = $25&255; //@line 570 "load_pat.cpp"
  $27 = $26 & 251; //@line 570 "load_pat.cpp"
  $28 = $27&255; //@line 570 "load_pat.cpp"
  HEAP8[$24>>0] = $28; //@line 570 "load_pat.cpp"
  STACKTOP = sp;return; //@line 574 "load_pat.cpp"
 }
 $$phi$trans$insert = ((($hw)) + 16|0);
 $$pre = HEAPU8[$$phi$trans$insert>>0]|(HEAPU8[$$phi$trans$insert+1>>0]<<8)|(HEAPU8[$$phi$trans$insert+2>>0]<<16)|(HEAPU8[$$phi$trans$insert+3>>0]<<24); //@line 572 "load_pat.cpp"
 $29 = ($$pre>>>0)>($21>>>0); //@line 572 "load_pat.cpp"
 if (!($29)) {
  STACKTOP = sp;return; //@line 574 "load_pat.cpp"
 }
 HEAP8[$$phi$trans$insert>>0]=$21&255;HEAP8[$$phi$trans$insert+1>>0]=($21>>8)&255;HEAP8[$$phi$trans$insert+2>>0]=($21>>16)&255;HEAP8[$$phi$trans$insert+3>>0]=$21>>24; //@line 573 "load_pat.cpp"
 STACKTOP = sp;return; //@line 574 "load_pat.cpp"
}
function __ZL15pat_setpat_instP10WaveHeaderP17_INSTRUMENTHEADERi($hw,$d,$smp) {
 $hw = $hw|0;
 $d = $d|0;
 $smp = $smp|0;
 var $$$i = 0, $$$i$i = 0, $$0$i$i = 0, $$lcssa = 0, $$lobit = 0, $$phi$trans$insert$i = 0, $$pre16$i = 0, $$pre17$i = 0, $$pre18$i = 0, $$pre19$i = 0, $$pre20$i = 0, $$pre21$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $envpoint = 0, $envvolume = 0, $exitcond = 0, $exitcond$i = 0, $exitcond13$i = 0;
 var $exitcond4 = 0, $i$17$i = 0, $i$43$i = 0, $inuse$0$ = 0, $inuse$0$$lcssa = 0, $inuse$03 = 0, $sum$06$i = 0, $u$02 = 0, $u$11 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $envpoint = sp + 24|0;
 $envvolume = sp;
 $0 = ((($d)) + 580|0); //@line 1156 "load_pat.cpp"
 HEAP8[$0>>0] = 0; //@line 1156 "load_pat.cpp"
 HEAP32[$d>>2] = 0; //@line 1157 "load_pat.cpp"
 $1 = ((($d)) + 10|0); //@line 1158 "load_pat.cpp"
 HEAP16[$1>>1] = 128; //@line 1158 "load_pat.cpp"
 $2 = ((($d)) + 584|0); //@line 1159 "load_pat.cpp"
 HEAP8[$2>>0] = 60; //@line 1159 "load_pat.cpp"
 $3 = ((($d)) + 4|0); //@line 1160 "load_pat.cpp"
 HEAP32[$3>>2] = 0; //@line 1160 "load_pat.cpp"
 $4 = ((($hw)) + 55|0); //@line 1161 "load_pat.cpp"
 $5 = HEAP8[$4>>0]|0; //@line 1161 "load_pat.cpp"
 $6 = ($5&255) >>> 6; //@line 1161 "load_pat.cpp"
 $$lobit = $6 & 1; //@line 1161 "load_pat.cpp"
 $7 = $5 & 32; //@line 1162 "load_pat.cpp"
 $8 = ($7&255) >>> 4; //@line 1162 "load_pat.cpp"
 $9 = $8 | $$lobit; //@line 1162 "load_pat.cpp"
 $10 = $9&255; //@line 1162 "load_pat.cpp"
 $11 = ($7&255) >>> 4; //@line 1162 "load_pat.cpp"
 $12 = $11 | $$lobit; //@line 1162 "load_pat.cpp"
 $13 = $12&255; //@line 1162 "load_pat.cpp"
 HEAP32[$3>>2] = $10; //@line 1161 "load_pat.cpp"
 $14 = $5 & 4; //@line 1163 "load_pat.cpp"
 $15 = ($14<<24>>24)==(0); //@line 1163 "load_pat.cpp"
 if (!($15)) {
  $16 = ((($hw)) + 12|0); //@line 1163 "load_pat.cpp"
  $17 = HEAPU8[$16>>0]|(HEAPU8[$16+1>>0]<<8)|(HEAPU8[$16+2>>0]<<16)|(HEAPU8[$16+3>>0]<<24); //@line 1163 "load_pat.cpp"
  $18 = ((($hw)) + 16|0); //@line 1163 "load_pat.cpp"
  $19 = HEAPU8[$18>>0]|(HEAPU8[$18+1>>0]<<8)|(HEAPU8[$18+2>>0]<<16)|(HEAPU8[$18+3>>0]<<24); //@line 1163 "load_pat.cpp"
  $20 = ($17|0)==($19|0); //@line 1163 "load_pat.cpp"
  if (!($20)) {
   $21 = $13 | 4; //@line 1163 "load_pat.cpp"
   HEAP32[$3>>2] = $21; //@line 1163 "load_pat.cpp"
  }
 }
 $22 = ((($d)) + 556|0); //@line 1164 "load_pat.cpp"
 HEAP8[$22>>0] = 6; //@line 1164 "load_pat.cpp"
 $23 = ((($d)) + 557|0); //@line 1166 "load_pat.cpp"
 HEAP8[$23>>0] = 0; //@line 1166 "load_pat.cpp"
 $24 = ((($d)) + 561|0); //@line 1167 "load_pat.cpp"
 HEAP8[$24>>0] = 1; //@line 1167 "load_pat.cpp"
 $25 = ((($d)) + 562|0); //@line 1168 "load_pat.cpp"
 HEAP8[$25>>0] = 1; //@line 1168 "load_pat.cpp"
 $26 = ((($d)) + 559|0); //@line 1169 "load_pat.cpp"
 HEAP8[$26>>0] = 1; //@line 1169 "load_pat.cpp"
 $27 = ((($d)) + 560|0); //@line 1170 "load_pat.cpp"
 HEAP8[$27>>0] = 2; //@line 1170 "load_pat.cpp"
 $28 = ((($d)) + 563|0); //@line 1173 "load_pat.cpp"
 $29 = ((($d)) + 8|0); //@line 1175 "load_pat.cpp"
 HEAP8[$28>>0]=0&255;HEAP8[$28+1>>0]=(0>>8)&255;HEAP8[$28+2>>0]=(0>>16)&255;HEAP8[$28+3>>0]=0>>24; //@line 1173 "load_pat.cpp"
 HEAP16[$29>>1] = 64; //@line 1175 "load_pat.cpp"
 HEAP32[$envpoint>>2] = 0; //@line 1072 "load_pat.cpp"
 HEAP32[$envvolume>>2] = 64; //@line 1073 "load_pat.cpp"
 $30 = ((($envpoint)) + 4|0); //@line 1072 "load_pat.cpp"
 HEAP32[$30>>2] = 0; //@line 1072 "load_pat.cpp"
 $31 = ((($envvolume)) + 4|0); //@line 1073 "load_pat.cpp"
 HEAP32[$31>>2] = 64; //@line 1073 "load_pat.cpp"
 $32 = ((($envpoint)) + 8|0); //@line 1072 "load_pat.cpp"
 HEAP32[$32>>2] = 0; //@line 1072 "load_pat.cpp"
 $33 = ((($envvolume)) + 8|0); //@line 1073 "load_pat.cpp"
 HEAP32[$33>>2] = 64; //@line 1073 "load_pat.cpp"
 $34 = ((($envpoint)) + 12|0); //@line 1072 "load_pat.cpp"
 HEAP32[$34>>2] = 0; //@line 1072 "load_pat.cpp"
 $35 = ((($envvolume)) + 12|0); //@line 1073 "load_pat.cpp"
 HEAP32[$35>>2] = 64; //@line 1073 "load_pat.cpp"
 $36 = ((($envpoint)) + 16|0); //@line 1072 "load_pat.cpp"
 HEAP32[$36>>2] = 0; //@line 1072 "load_pat.cpp"
 $37 = ((($envvolume)) + 16|0); //@line 1073 "load_pat.cpp"
 HEAP32[$37>>2] = 64; //@line 1073 "load_pat.cpp"
 $38 = ((($envpoint)) + 20|0); //@line 1072 "load_pat.cpp"
 HEAP32[$38>>2] = 0; //@line 1072 "load_pat.cpp"
 $39 = ((($envvolume)) + 20|0); //@line 1073 "load_pat.cpp"
 HEAP32[$39>>2] = 64; //@line 1073 "load_pat.cpp"
 $40 = ((($hw)) + 37|0); //@line 1070 "load_pat.cpp"
 $41 = (_memcmp($40,527897,6)|0); //@line 1075 "load_pat.cpp"
 $42 = ($41|0)==(0); //@line 1075 "load_pat.cpp"
 if ($42) {
  $inuse$03 = 0;$u$02 = 0;
 } else {
  $43 = ((($hw)) + 48|0); //@line 1075 "load_pat.cpp"
  $44 = HEAP8[$43>>0]|0; //@line 1075 "load_pat.cpp"
  $45 = ($44&255)>(99); //@line 1075 "load_pat.cpp"
  if ($45) {
   $inuse$03 = 0;$u$02 = 0;
  } else {
   $46 = $5&255; //@line 1076 "load_pat.cpp"
   $47 = $46 & 32; //@line 1076 "load_pat.cpp"
   $48 = ($47|0)==(0); //@line 1076 "load_pat.cpp"
   if ($48) {
    $inuse$03 = 0;$u$02 = 0;
   } else {
    $49 = ((($hw)) + 8|0); //@line 1077 "load_pat.cpp"
    $50 = HEAPU8[$49>>0]|(HEAPU8[$49+1>>0]<<8)|(HEAPU8[$49+2>>0]<<16)|(HEAPU8[$49+3>>0]<<24); //@line 1077 "load_pat.cpp"
    $51 = ($50|0)==(0); //@line 1078 "load_pat.cpp"
    if ($51) {
     $inuse$03 = 0;$u$02 = 0;
    } else {
     $i$17$i = 0;$sum$06$i = 0;
     while(1) {
      $52 = (((($hw)) + 43|0) + ($i$17$i)|0); //@line 1084 "load_pat.cpp"
      $53 = HEAP8[$52>>0]|0; //@line 1084 "load_pat.cpp"
      $54 = $53&255; //@line 1084 "load_pat.cpp"
      $55 = (($envvolume) + ($i$17$i<<2)|0); //@line 1084 "load_pat.cpp"
      HEAP32[$55>>2] = $54; //@line 1084 "load_pat.cpp"
      $56 = (((($hw)) + 37|0) + ($i$17$i)|0); //@line 1085 "load_pat.cpp"
      $57 = HEAP8[$56>>0]|0; //@line 1085 "load_pat.cpp"
      $58 = $57&255; //@line 1085 "load_pat.cpp"
      $59 = ($i$17$i|0)==(0); //@line 1085 "load_pat.cpp"
      if ($59) {
       $70 = 0;
      } else {
       $60 = (($i$17$i) + -1)|0; //@line 1085 "load_pat.cpp"
       $61 = (((($hw)) + 43|0) + ($60)|0); //@line 1085 "load_pat.cpp"
       $62 = HEAP8[$61>>0]|0; //@line 1085 "load_pat.cpp"
       $63 = $62&255; //@line 1085 "load_pat.cpp"
       $70 = $63;
      }
      $64 = $58 >>> 6; //@line 1057 "load_pat.cpp"
      $65 = Math_imul($64, -3)|0; //@line 1057 "load_pat.cpp"
      $66 = (($65) + 3)|0; //@line 1057 "load_pat.cpp"
      $67 = $58 & 63; //@line 1058 "load_pat.cpp"
      $68 = ($67|0)==(0); //@line 1059 "load_pat.cpp"
      if ($68) {
       $$0$i$i = 0;
      } else {
       $69 = (($54) - ($70))|0; //@line 1060 "load_pat.cpp"
       $71 = ($54|0)==($70|0); //@line 1061 "load_pat.cpp"
       if ($71) {
        $$0$i$i = 0;
       } else {
        $72 = ($69|0)<(0); //@line 1062 "load_pat.cpp"
        $73 = (0 - ($69))|0; //@line 1062 "load_pat.cpp"
        $$$i$i = $72 ? $73 : $69; //@line 1062 "load_pat.cpp"
        $74 = $$$i$i << $66; //@line 1063 "load_pat.cpp"
        $75 = (($74|0) / ($67|0))&-1; //@line 1063 "load_pat.cpp"
        $$0$i$i = $75;
       }
      }
      $76 = (($envpoint) + ($i$17$i<<2)|0); //@line 1085 "load_pat.cpp"
      HEAP32[$76>>2] = $$0$i$i; //@line 1085 "load_pat.cpp"
      $77 = (($$0$i$i) + ($sum$06$i))|0; //@line 1086 "load_pat.cpp"
      $78 = (($i$17$i) + 1)|0; //@line 1083 "load_pat.cpp"
      $exitcond13$i = ($78|0)==(6); //@line 1083 "load_pat.cpp"
      if ($exitcond13$i) {
       $$lcssa = $77;
       break;
      } else {
       $i$17$i = $78;$sum$06$i = $77;
      }
     }
     $79 = $46 & 1; //@line 1079 "load_pat.cpp"
     $$$i = $50 >> $79; //@line 1079 "load_pat.cpp"
     $80 = ($$lcssa|0)==(0); //@line 1088 "load_pat.cpp"
     if ($80) {
      $inuse$03 = 0;$u$02 = 0;
     } else {
      $81 = ($$lcssa|0)>($$$i|0); //@line 1089 "load_pat.cpp"
      $82 = HEAP32[$envpoint>>2]|0; //@line 1091 "load_pat.cpp"
      if ($81) {
       $83 = Math_imul($82, $$$i)|0; //@line 1091 "load_pat.cpp"
       $84 = (($83|0) / ($$lcssa|0))&-1; //@line 1091 "load_pat.cpp"
       HEAP32[$envpoint>>2] = $84; //@line 1091 "load_pat.cpp"
       $85 = HEAP32[$30>>2]|0; //@line 1091 "load_pat.cpp"
       $86 = Math_imul($85, $$$i)|0; //@line 1091 "load_pat.cpp"
       $87 = (($86|0) / ($$lcssa|0))&-1; //@line 1091 "load_pat.cpp"
       HEAP32[$30>>2] = $87; //@line 1091 "load_pat.cpp"
       $88 = HEAP32[$32>>2]|0; //@line 1091 "load_pat.cpp"
       $89 = Math_imul($88, $$$i)|0; //@line 1091 "load_pat.cpp"
       $90 = (($89|0) / ($$lcssa|0))&-1; //@line 1091 "load_pat.cpp"
       HEAP32[$32>>2] = $90; //@line 1091 "load_pat.cpp"
       $91 = HEAP32[$34>>2]|0; //@line 1091 "load_pat.cpp"
       $92 = Math_imul($91, $$$i)|0; //@line 1091 "load_pat.cpp"
       $93 = (($92|0) / ($$lcssa|0))&-1; //@line 1091 "load_pat.cpp"
       HEAP32[$34>>2] = $93; //@line 1091 "load_pat.cpp"
       $94 = HEAP32[$36>>2]|0; //@line 1091 "load_pat.cpp"
       $95 = Math_imul($94, $$$i)|0; //@line 1091 "load_pat.cpp"
       $96 = (($95|0) / ($$lcssa|0))&-1; //@line 1091 "load_pat.cpp"
       HEAP32[$36>>2] = $96; //@line 1091 "load_pat.cpp"
       $97 = HEAP32[$38>>2]|0; //@line 1091 "load_pat.cpp"
       $98 = Math_imul($97, $$$i)|0; //@line 1091 "load_pat.cpp"
       $99 = (($98|0) / ($$lcssa|0))&-1; //@line 1091 "load_pat.cpp"
       HEAP32[$38>>2] = $99; //@line 1091 "load_pat.cpp"
       $101 = $84;$102 = $87;$104 = $90;$106 = $93;$108 = $96;$110 = $99;
      } else {
       $$pre16$i = HEAP32[$30>>2]|0; //@line 1094 "load_pat.cpp"
       $$pre17$i = HEAP32[$32>>2]|0; //@line 1094 "load_pat.cpp"
       $$pre18$i = HEAP32[$34>>2]|0; //@line 1094 "load_pat.cpp"
       $$pre19$i = HEAP32[$36>>2]|0; //@line 1094 "load_pat.cpp"
       $$pre20$i = HEAP32[$38>>2]|0; //@line 1094 "load_pat.cpp"
       $101 = $82;$102 = $$pre16$i;$104 = $$pre17$i;$106 = $$pre18$i;$108 = $$pre19$i;$110 = $$pre20$i;
      }
      $100 = (($101) + ($102))|0; //@line 1094 "load_pat.cpp"
      HEAP32[$30>>2] = $100; //@line 1094 "load_pat.cpp"
      $103 = (($100) + ($104))|0; //@line 1094 "load_pat.cpp"
      HEAP32[$32>>2] = $103; //@line 1094 "load_pat.cpp"
      $105 = (($103) + ($106))|0; //@line 1094 "load_pat.cpp"
      HEAP32[$34>>2] = $105; //@line 1094 "load_pat.cpp"
      $107 = (($105) + ($108))|0; //@line 1094 "load_pat.cpp"
      HEAP32[$36>>2] = $107; //@line 1094 "load_pat.cpp"
      $109 = (($107) + ($110))|0; //@line 1094 "load_pat.cpp"
      HEAP32[$38>>2] = $109; //@line 1094 "load_pat.cpp"
      $113 = $101;$i$43$i = 0;
      while(1) {
       $111 = (($envpoint) + ($i$43$i<<2)|0); //@line 1096 "load_pat.cpp"
       $112 = $113 << 8; //@line 1096 "load_pat.cpp"
       $114 = (($112|0) / ($$$i|0))&-1; //@line 1096 "load_pat.cpp"
       $115 = (($114) + 1)|0; //@line 1097 "load_pat.cpp"
       HEAP32[$111>>2] = $115; //@line 1097 "load_pat.cpp"
       $116 = ($i$43$i|0)>(0); //@line 1098 "load_pat.cpp"
       do {
        if ($116) {
         $117 = (($i$43$i) + -1)|0; //@line 1098 "load_pat.cpp"
         $118 = (($envpoint) + ($117<<2)|0); //@line 1098 "load_pat.cpp"
         $119 = HEAP32[$118>>2]|0; //@line 1098 "load_pat.cpp"
         $120 = ($114|0)<($119|0); //@line 1098 "load_pat.cpp"
         if ($120) {
          $121 = (($envvolume) + ($i$43$i<<2)|0); //@line 1099 "load_pat.cpp"
          $122 = HEAP32[$121>>2]|0; //@line 1099 "load_pat.cpp"
          $123 = (($envvolume) + ($117<<2)|0); //@line 1099 "load_pat.cpp"
          $124 = HEAP32[$123>>2]|0; //@line 1099 "load_pat.cpp"
          $125 = ($122|0)==($124|0); //@line 1099 "load_pat.cpp"
          if ($125) {
           HEAP32[$111>>2] = $119; //@line 1099 "load_pat.cpp"
           $128 = $119;
           break;
          } else {
           $126 = (($119) + 1)|0; //@line 1100 "load_pat.cpp"
           HEAP32[$111>>2] = $126; //@line 1100 "load_pat.cpp"
           $128 = $126;
           break;
          }
         } else {
          $128 = $115;
         }
        } else {
         $128 = $115;
        }
       } while(0);
       $127 = ($128|0)>(256); //@line 1102 "load_pat.cpp"
       if ($127) {
        HEAP32[$111>>2] = 256; //@line 1102 "load_pat.cpp"
       }
       $129 = (($i$43$i) + 1)|0; //@line 1095 "load_pat.cpp"
       $exitcond$i = ($129|0)==(6); //@line 1095 "load_pat.cpp"
       if ($exitcond$i) {
        break;
       }
       $$phi$trans$insert$i = (($envpoint) + ($129<<2)|0); //@line 1176 "load_pat.cpp"
       $$pre21$i = HEAP32[$$phi$trans$insert$i>>2]|0; //@line 1096 "load_pat.cpp"
       $113 = $$pre21$i;$i$43$i = $129;
      }
      HEAP32[$39>>2] = 0; //@line 1104 "load_pat.cpp"
      $inuse$03 = 0;$u$02 = 0;
     }
    }
   }
  }
 }
 while(1) {
  $130 = (($envvolume) + ($u$02<<2)|0); //@line 1180 "load_pat.cpp"
  $131 = HEAP32[$130>>2]|0; //@line 1180 "load_pat.cpp"
  $132 = ($131|0)==(64); //@line 1180 "load_pat.cpp"
  $inuse$0$ = $132 ? $inuse$03 : 1; //@line 1180 "load_pat.cpp"
  $133 = (($envpoint) + ($u$02<<2)|0); //@line 1181 "load_pat.cpp"
  $134 = HEAP32[$133>>2]|0; //@line 1181 "load_pat.cpp"
  $135 = $134&65535; //@line 1181 "load_pat.cpp"
  $136 = (((($d)) + 12|0) + ($u$02<<1)|0); //@line 1181 "load_pat.cpp"
  HEAP16[$136>>1] = $135; //@line 1181 "load_pat.cpp"
  $137 = $131&255; //@line 1182 "load_pat.cpp"
  $138 = (((($d)) + 204|0) + ($u$02)|0); //@line 1182 "load_pat.cpp"
  HEAP8[$138>>0] = $137; //@line 1182 "load_pat.cpp"
  $139 = (((($d)) + 76|0) + ($u$02<<1)|0); //@line 1183 "load_pat.cpp"
  HEAP16[$139>>1] = 0; //@line 1183 "load_pat.cpp"
  $140 = (((($d)) + 236|0) + ($u$02)|0); //@line 1184 "load_pat.cpp"
  HEAP8[$140>>0] = 0; //@line 1184 "load_pat.cpp"
  $141 = ($u$02|0)==(0); //@line 1185 "load_pat.cpp"
  if ($141) {
   $inuse$03 = $inuse$0$;$u$02 = 1;
   continue;
  }
  $142 = HEAP16[$136>>1]|0; //@line 1187 "load_pat.cpp"
  $143 = (($u$02) + -1)|0; //@line 1187 "load_pat.cpp"
  $144 = (((($d)) + 12|0) + ($143<<1)|0); //@line 1187 "load_pat.cpp"
  $145 = HEAP16[$144>>1]|0; //@line 1187 "load_pat.cpp"
  $146 = ($142&65535)<($145&65535); //@line 1187 "load_pat.cpp"
  if ($146) {
   $147 = $142&65535; //@line 1187 "load_pat.cpp"
   $148 = $147 & 255; //@line 1189 "load_pat.cpp"
   $149 = $145&65535; //@line 1190 "load_pat.cpp"
   $150 = $149 & 65280; //@line 1190 "load_pat.cpp"
   $151 = $150 | $148; //@line 1190 "load_pat.cpp"
   $152 = $151&65535; //@line 1190 "load_pat.cpp"
   HEAP16[$136>>1] = $152; //@line 1190 "load_pat.cpp"
   $153 = ($151>>>0)<($149>>>0); //@line 1191 "load_pat.cpp"
   if ($153) {
    $154 = (($151) + 256)|0; //@line 1191 "load_pat.cpp"
    $155 = $154&65535; //@line 1191 "load_pat.cpp"
    HEAP16[$136>>1] = $155; //@line 1191 "load_pat.cpp"
   }
  }
  $156 = (($u$02) + 1)|0; //@line 1178 "load_pat.cpp"
  $exitcond4 = ($156|0)==(6); //@line 1178 "load_pat.cpp"
  if ($exitcond4) {
   $inuse$0$$lcssa = $inuse$0$;
   break;
  } else {
   $inuse$03 = $inuse$0$;$u$02 = $156;
  }
 }
 $157 = ($inuse$0$$lcssa|0)==(0); //@line 1195 "load_pat.cpp"
 if ($157) {
  HEAP8[$22>>0] = 0; //@line 1195 "load_pat.cpp"
 }
 $158 = $smp&255; //@line 1199 "load_pat.cpp"
 $u$11 = 0;
 while(1) {
  $159 = (($u$11) + 1)|0; //@line 1198 "load_pat.cpp"
  $160 = $159&255; //@line 1198 "load_pat.cpp"
  $161 = (((($d)) + 428|0) + ($u$11)|0); //@line 1198 "load_pat.cpp"
  HEAP8[$161>>0] = $160; //@line 1198 "load_pat.cpp"
  $162 = (((($d)) + 300|0) + ($u$11)|0); //@line 1199 "load_pat.cpp"
  HEAP8[$162>>0] = $158; //@line 1199 "load_pat.cpp"
  $exitcond = ($159|0)==(128); //@line 1196 "load_pat.cpp"
  if ($exitcond) {
   break;
  } else {
   $u$11 = $159;
  }
 }
 STACKTOP = sp;return; //@line 1201 "load_pat.cpp"
}
function __ZL11pat_readpatiPci($pat,$dest,$num) {
 $pat = $pat|0;
 $dest = $dest|0;
 $num = $num|0;
 var $$$i = 0, $$$i2 = 0, $$$op$i = 0, $$0$i = 0, $$1$i = 0, $$2$i = 0, $$2$op$i = 0, $$3$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $9 = 0, $exitcond$i = 0, $exitcond21$i = 0, $exitcond22$i = 0, $exitcond23$i = 0, $fname = 0, $hw = 0, $i$019$i = 0, $i$116$i = 0, $i$213$i = 0, $i$311$i = 0;
 var $pb$010$i = 0, $pi$015$i = 0, $pu$012$i = 0, $pw$018$i = 0, $vararg_buffer = 0, $vararg_buffer3 = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr6 = 0, $vararg_ptr7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $fname = sp + 128|0;
 $hw = sp + 32|0;
 $0 = HEAP32[300427]|0; //@line 727 "load_pat.cpp"
 $1 = ($0|0)==(0); //@line 727 "load_pat.cpp"
 do {
  if ($1) {
   $2 = (1253146 + ($pat<<7)|0); //@line 489 "load_pat.cpp"
   $3 = HEAP8[$2>>0]|0; //@line 490 "load_pat.cpp"
   $4 = ($3<<24>>24)==(47); //@line 490 "load_pat.cpp"
   $5 = (((1253146 + ($pat<<7)|0)) + 1|0); //@line 491 "load_pat.cpp"
   $$$i = $4 ? $5 : $2; //@line 491 "load_pat.cpp"
   $6 = (_strrchr($$$i,58)|0); //@line 492 "load_pat.cpp"
   $7 = ($6|0)==(0|0); //@line 493 "load_pat.cpp"
   $8 = $4 ? 1282466 : 1252890; //@line 494 "load_pat.cpp"
   if ($7) {
    HEAP32[$vararg_buffer3>>2] = $8; //@line 498 "load_pat.cpp"
    $vararg_ptr6 = ((($vararg_buffer3)) + 4|0); //@line 498 "load_pat.cpp"
    HEAP32[$vararg_ptr6>>2] = 47; //@line 498 "load_pat.cpp"
    $vararg_ptr7 = ((($vararg_buffer3)) + 8|0); //@line 498 "load_pat.cpp"
    HEAP32[$vararg_ptr7>>2] = $$$i; //@line 498 "load_pat.cpp"
    (_sprintf($fname,527886,$vararg_buffer3)|0); //@line 498 "load_pat.cpp"
    $$0$i = 0;
   } else {
    HEAP32[$vararg_buffer>>2] = $8; //@line 494 "load_pat.cpp"
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 494 "load_pat.cpp"
    HEAP32[$vararg_ptr1>>2] = 47; //@line 494 "load_pat.cpp"
    $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 494 "load_pat.cpp"
    HEAP32[$vararg_ptr2>>2] = $$$i; //@line 494 "load_pat.cpp"
    (_sprintf($fname,527874,$vararg_buffer)|0); //@line 494 "load_pat.cpp"
    $9 = (_strrchr($fname,58)|0); //@line 495 "load_pat.cpp"
    ;HEAP8[$9>>0]=HEAP8[527881>>0]|0;HEAP8[$9+1>>0]=HEAP8[527881+1>>0]|0;HEAP8[$9+2>>0]=HEAP8[527881+2>>0]|0;HEAP8[$9+3>>0]=HEAP8[527881+3>>0]|0;HEAP8[$9+4>>0]=HEAP8[527881+4>>0]|0; //@line 495 "load_pat.cpp"
    $$0$i = $6;
   }
   HEAP32[300428] = $$0$i; //@line 728 "load_pat.cpp"
   $10 = (_fopen($fname,527143)|0); //@line 729 "load_pat.cpp"
   HEAP32[300429] = $10; //@line 729 "load_pat.cpp"
   $11 = ($10|0)==(0|0); //@line 730 "load_pat.cpp"
   if ($11) {
    STACKTOP = sp;return; //@line 743 "load_pat.cpp"
   } else {
    __ZL19pat_read_waveheaderP8_IO_FILEP10WaveHeaderi($10,$hw); //@line 732 "load_pat.cpp"
    $12 = ((($hw)) + 8|0); //@line 733 "load_pat.cpp"
    $13 = HEAPU8[$12>>0]|(HEAPU8[$12+1>>0]<<8)|(HEAPU8[$12+2>>0]<<16)|(HEAPU8[$12+3>>0]<<24); //@line 733 "load_pat.cpp"
    HEAP32[300430] = $13; //@line 733 "load_pat.cpp"
    break;
   }
  }
 } while(0);
 $14 = HEAP32[300429]|0; //@line 735 "load_pat.cpp"
 (_fread($dest,$num,1,$14)|0); //@line 735 "load_pat.cpp"
 $15 = HEAP32[300428]|0; //@line 736 "load_pat.cpp"
 $16 = ($15|0)==(0|0); //@line 697 "load_pat.cpp"
 do {
  if (!($16)) {
   $17 = (_strstr($15,527910)|0); //@line 698 "load_pat.cpp"
   $18 = ($17|0)==(0|0); //@line 699 "load_pat.cpp"
   if (!($18)) {
    $19 = (_strchr($17,61)|0); //@line 700 "load_pat.cpp"
    $20 = ((($19)) + 1|0); //@line 700 "load_pat.cpp"
    $21 = (_atoi($20)|0); //@line 700 "load_pat.cpp"
    $22 = ($21|0)==(100); //@line 737 "load_pat.cpp"
    if (!($22)) {
     $23 = ((($hw)) + 55|0); //@line 737 "load_pat.cpp"
     $24 = HEAP8[$23>>0]|0; //@line 737 "load_pat.cpp"
     $25 = $24&255; //@line 737 "load_pat.cpp"
     $26 = $25 & 1; //@line 651 "load_pat.cpp"
     $27 = ($26|0)==(0); //@line 651 "load_pat.cpp"
     if ($27) {
      $53 = $25 & 2; //@line 673 "load_pat.cpp"
      $54 = ($53|0)==(0); //@line 673 "load_pat.cpp"
      $55 = ($num|0)>(0); //@line 684 "load_pat.cpp"
      if ($54) {
       if ($55) {
        $i$311$i = 0;$pb$010$i = $dest;
       } else {
        break;
       }
       while(1) {
        $67 = HEAP8[$pb$010$i>>0]|0; //@line 685 "load_pat.cpp"
        $68 = $67 << 24 >> 24; //@line 685 "load_pat.cpp"
        $69 = Math_imul($68, $21)|0; //@line 685 "load_pat.cpp"
        $70 = (($69|0) / 100)&-1; //@line 685 "load_pat.cpp"
        $71 = ($70|0)<(-128); //@line 686 "load_pat.cpp"
        $$3$i = $71 ? -128 : $70; //@line 686 "load_pat.cpp"
        $72 = ($$3$i|0)>(127); //@line 687 "load_pat.cpp"
        $73 = $$3$i&255; //@line 688 "load_pat.cpp"
        $74 = $72 ? 127 : $73; //@line 688 "load_pat.cpp"
        $75 = ((($pb$010$i)) + 1|0); //@line 688 "load_pat.cpp"
        HEAP8[$pb$010$i>>0] = $74; //@line 688 "load_pat.cpp"
        $76 = (($i$311$i) + 1)|0; //@line 684 "load_pat.cpp"
        $exitcond$i = ($76|0)==($num|0); //@line 684 "load_pat.cpp"
        if ($exitcond$i) {
         break;
        } else {
         $i$311$i = $76;$pb$010$i = $75;
        }
       }
      } else {
       if ($55) {
        $i$213$i = 0;$pu$012$i = $dest;
       } else {
        break;
       }
       while(1) {
        $56 = HEAP8[$pu$012$i>>0]|0; //@line 676 "load_pat.cpp"
        $57 = $56&255; //@line 676 "load_pat.cpp"
        $58 = (($57) + -128)|0; //@line 676 "load_pat.cpp"
        $59 = Math_imul($58, $21)|0; //@line 676 "load_pat.cpp"
        $60 = (($59|0) / 100)&-1; //@line 676 "load_pat.cpp"
        $61 = ($60|0)<(-128); //@line 677 "load_pat.cpp"
        $$2$i = $61 ? -128 : $60; //@line 677 "load_pat.cpp"
        $62 = ($$2$i|0)>(127); //@line 678 "load_pat.cpp"
        $$2$op$i = (($$2$i) + 128)|0; //@line 679 "load_pat.cpp"
        $63 = $$2$op$i&255; //@line 679 "load_pat.cpp"
        $64 = $62 ? -1 : $63; //@line 679 "load_pat.cpp"
        $65 = ((($pu$012$i)) + 1|0); //@line 679 "load_pat.cpp"
        HEAP8[$pu$012$i>>0] = $64; //@line 679 "load_pat.cpp"
        $66 = (($i$213$i) + 1)|0; //@line 675 "load_pat.cpp"
        $exitcond21$i = ($66|0)==($num|0); //@line 675 "load_pat.cpp"
        if ($exitcond21$i) {
         break;
        } else {
         $i$213$i = $66;$pu$012$i = $65;
        }
       }
      }
     } else {
      $28 = $num >> 1; //@line 652 "load_pat.cpp"
      $29 = $25 & 2; //@line 653 "load_pat.cpp"
      $30 = ($29|0)==(0); //@line 653 "load_pat.cpp"
      $31 = ($28|0)>(0); //@line 664 "load_pat.cpp"
      if ($30) {
       if ($31) {
        $i$116$i = 0;$pi$015$i = $dest;
       } else {
        break;
       }
       while(1) {
        $43 = HEAP16[$pi$015$i>>1]|0; //@line 665 "load_pat.cpp"
        $44 = $43 << 16 >> 16; //@line 665 "load_pat.cpp"
        $45 = Math_imul($44, $21)|0; //@line 665 "load_pat.cpp"
        $46 = (($45|0) / 100)&-1; //@line 665 "load_pat.cpp"
        $47 = ($46|0)<(-32768); //@line 666 "load_pat.cpp"
        $$1$i = $47 ? -32768 : $46; //@line 666 "load_pat.cpp"
        $48 = ($$1$i|0)>(32767); //@line 667 "load_pat.cpp"
        $49 = $$1$i&65535; //@line 668 "load_pat.cpp"
        $50 = $48 ? 32767 : $49; //@line 668 "load_pat.cpp"
        $51 = ((($pi$015$i)) + 2|0); //@line 668 "load_pat.cpp"
        HEAP16[$pi$015$i>>1] = $50; //@line 668 "load_pat.cpp"
        $52 = (($i$116$i) + 1)|0; //@line 664 "load_pat.cpp"
        $exitcond22$i = ($52|0)==($28|0); //@line 664 "load_pat.cpp"
        if ($exitcond22$i) {
         break;
        } else {
         $i$116$i = $52;$pi$015$i = $51;
        }
       }
      } else {
       if ($31) {
        $i$019$i = 0;$pw$018$i = $dest;
       } else {
        break;
       }
       while(1) {
        $32 = HEAP16[$pw$018$i>>1]|0; //@line 656 "load_pat.cpp"
        $33 = $32&65535; //@line 656 "load_pat.cpp"
        $34 = (($33) + -32768)|0; //@line 656 "load_pat.cpp"
        $35 = Math_imul($34, $21)|0; //@line 656 "load_pat.cpp"
        $36 = (($35|0) / 100)&-1; //@line 656 "load_pat.cpp"
        $37 = ($36|0)<(-32768); //@line 657 "load_pat.cpp"
        $$$i2 = $37 ? -32768 : $36; //@line 657 "load_pat.cpp"
        $38 = ($$$i2|0)>(32767); //@line 658 "load_pat.cpp"
        $$$op$i = (($$$i2) + 32768)|0; //@line 659 "load_pat.cpp"
        $39 = $$$op$i&65535; //@line 659 "load_pat.cpp"
        $40 = $38 ? -1 : $39; //@line 659 "load_pat.cpp"
        $41 = ((($pw$018$i)) + 2|0); //@line 659 "load_pat.cpp"
        HEAP16[$pw$018$i>>1] = $40; //@line 659 "load_pat.cpp"
        $42 = (($i$019$i) + 1)|0; //@line 655 "load_pat.cpp"
        $exitcond23$i = ($42|0)==($28|0); //@line 655 "load_pat.cpp"
        if ($exitcond23$i) {
         break;
        } else {
         $i$019$i = $42;$pw$018$i = $41;
        }
       }
      }
     }
    }
   }
  }
 } while(0);
 $77 = HEAP32[300427]|0; //@line 738 "load_pat.cpp"
 $78 = (($77) + ($num))|0; //@line 738 "load_pat.cpp"
 HEAP32[300427] = $78; //@line 738 "load_pat.cpp"
 $79 = HEAP32[300430]|0; //@line 739 "load_pat.cpp"
 $80 = ($78|0)<($79|0); //@line 739 "load_pat.cpp"
 if ($80) {
  STACKTOP = sp;return; //@line 743 "load_pat.cpp"
 }
 HEAP32[300427] = 0; //@line 740 "load_pat.cpp"
 $81 = HEAP32[300429]|0; //@line 741 "load_pat.cpp"
 (_fclose($81)|0); //@line 741 "load_pat.cpp"
 HEAP32[300429] = 0; //@line 742 "load_pat.cpp"
 STACKTOP = sp;return; //@line 743 "load_pat.cpp"
}
function __ZL9pat_sinusi($i) {
 $i = $i|0;
 var $0 = 0.0, $1 = 0.0, $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+($i|0)); //@line 263 "load_pat.cpp"
 $1 = $0 * 0.39312243461608887; //@line 263 "load_pat.cpp"
 $2 = (+Math_sin((+$1))); //@line 263 "load_pat.cpp"
 return (+$2); //@line 264 "load_pat.cpp"
}
function __ZL10pat_squarei($i) {
 $i = $i|0;
 var $$ = 0.0, $$0 = 0.0, $0 = 0.0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+($i|0)); //@line 269 "load_pat.cpp"
 $1 = $0 * 0.39312243461608887; //@line 269 "load_pat.cpp"
 $2 = (+Math_sin((+$1))); //@line 269 "load_pat.cpp"
 $3 = $2 * 30.0; //@line 269 "load_pat.cpp"
 $4 = $3 > 0.99000000953674316; //@line 270 "load_pat.cpp"
 $5 = $3 < -0.99000000953674316; //@line 271 "load_pat.cpp"
 $$ = $5 ? -0.99000000953674316 : $3; //@line 271 "load_pat.cpp"
 $$0 = $4 ? 0.99000000953674316 : $$; //@line 270 "load_pat.cpp"
 return (+$$0); //@line 273 "load_pat.cpp"
}
function __ZL12pat_sawtoothi($i) {
 $i = $i|0;
 var $$0 = 0.0, $$01 = 0.0, $0 = 0.0, $1 = 0.0, $10 = 0.0, $11 = 0, $12 = 0.0, $2 = 0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0, $res$0$lcssa = 0.0, $res$02 = 0.0, $res$1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+($i|0)); //@line 277 "load_pat.cpp"
 $1 = $0 * 0.39312243461608887; //@line 277 "load_pat.cpp"
 $2 = $1 > 6.2831854820251465; //@line 278 "load_pat.cpp"
 if ($2) {
  $res$02 = $1;
  while(1) {
   $3 = $res$02 + -6.2831854820251465; //@line 279 "load_pat.cpp"
   $4 = $3 > 6.2831854820251465; //@line 278 "load_pat.cpp"
   if ($4) {
    $res$02 = $3;
   } else {
    $res$0$lcssa = $3;
    break;
   }
  }
 } else {
  $res$0$lcssa = $1;
 }
 $5 = $res$0$lcssa > 3.1415927410125732; //@line 281 "load_pat.cpp"
 $6 = 3.1415927410125732 - $res$0$lcssa; //@line 282 "load_pat.cpp"
 $$0 = $5 ? -2.0 : 2.0; //@line 281 "load_pat.cpp"
 $res$1 = $5 ? $6 : $res$0$lcssa; //@line 281 "load_pat.cpp"
 $7 = $res$1 * 0.31830987334251404; //@line 285 "load_pat.cpp"
 $8 = $7 * $$0; //@line 285 "load_pat.cpp"
 $9 = $8 > 0.89999997615814208; //@line 286 "load_pat.cpp"
 if ($9) {
  $10 = 1.0 - $8; //@line 286 "load_pat.cpp"
  $$01 = $10;
  return (+$$01); //@line 289 "load_pat.cpp"
 }
 $11 = $8 < -0.89999997615814208; //@line 287 "load_pat.cpp"
 if (!($11)) {
  $$01 = $8;
  return (+$$01); //@line 289 "load_pat.cpp"
 }
 $12 = $8 + 1.0; //@line 287 "load_pat.cpp"
 $$01 = $12;
 return (+$$01); //@line 289 "load_pat.cpp"
}
function __ZN10CSoundFile7ReadMIDEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$$$i126 = 0, $$$i = 0, $$$i126 = 0, $$$i162 = 0, $$$i173 = 0, $$0 = 0, $$0$i$i = 0, $$0$i$i176 = 0, $$0$i1$i$i = 0, $$01$i = 0, $$05$i = 0, $$05$i$lcssa = 0, $$401 = 0, $$cast$i$i = 0, $$cast$i$i191 = 0, $$cast$pn$i$i = 0, $$cast$pn$i$i193 = 0, $$delta$0 = 0, $$idx13$val = 0;
 var $$idx15$val = 0, $$idx20$val = 0, $$idx28 = 0, $$idx28$val = 0, $$idx28$val$lcssa = 0, $$idx29$val = 0, $$idx30$val = 0, $$in$i$i = 0, $$in$i$i194 = 0, $$lcssa = 0, $$lcssa461 = 0, $$lcssa462 = 0, $$lcssa463 = 0, $$lcssa464 = 0, $$lcssa465 = 0, $$lcssa466 = 0, $$lcssa467 = 0, $$lcssa468 = 0, $$lcssa471 = 0, $$lcssa473 = 0;
 var $$lcssa477 = 0, $$lcssa478 = 0, $$lcssa483 = 0, $$lcssa484 = 0, $$lcssa485 = 0, $$lcssa486 = 0, $$lobit$i = 0, $$lobit10$i = 0, $$lobit11$i = 0, $$lobit12$i = 0, $$lobit13$i = 0, $$lobit14$i = 0, $$lobit15$i = 0, $$lobit16$i = 0, $$lobit17$i = 0, $$lobit18$i = 0, $$lobit19$i = 0, $$lobit6$i = 0, $$lobit7$i = 0, $$lobit8$i = 0;
 var $$lobit9$i = 0, $$maxtempo$1 = 0, $$off$i$i = 0, $$op$i$i = 0, $$op$op$i$i = 0, $$op2$i$i = 0, $$panlow$1 = 0, $$phi$trans$insert$i = 0, $$phi$trans$insert$i$phi$trans$insert = 0, $$pr = 0, $$pre = 0, $$pre$i = 0, $$pre$i$pre = 0, $$pre$phi340Z2D = 0, $$pre$phi354Z2D = 0, $$pre$phi356Z2D = 0, $$pre323 = 0, $$pre324 = 0, $$pre328 = 0, $$pre329 = 0;
 var $$pre334 = 0, $$pre335 = 0, $$pre336 = 0, $$pre337 = 0, $$pre341 = 0, $$pre343 = 0, $$pre345 = 0, $$pre347 = 0, $$pre349 = 0, $$pre351 = 0, $$pre353 = 0, $$pre355 = 0, $$pre6$i$pre = 0, $$promoted = 0, $$sink$i$i$i = 0, $$sink$i$i$i$i = 0, $$sink$i$i$i$i196 = 0, $$sink$i$i$i105 = 0, $$sink$i$i$i123 = 0, $$sink$i$i$i70 = 0;
 var $$sink$i$i$i84 = 0, $$sink$i$i$us$i = 0, $$str56$$str57$i = 0, $$sum$i33 = 0, $$sum1$i = 0, $$sum2$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0;
 var $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0;
 var $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0;
 var $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0;
 var $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0;
 var $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0;
 var $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0;
 var $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0;
 var $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0;
 var $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0;
 var $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0;
 var $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0;
 var $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0;
 var $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0;
 var $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0;
 var $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0, $1267 = 0, $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0;
 var $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0, $1285 = 0, $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0;
 var $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0, $1302 = 0, $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0;
 var $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0, $1320 = 0, $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0;
 var $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0;
 var $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0;
 var $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0;
 var $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0;
 var $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0;
 var $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0;
 var $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0;
 var $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0;
 var $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0;
 var $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0;
 var $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0;
 var $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0;
 var $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0;
 var $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0;
 var $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0;
 var $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0;
 var $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0;
 var $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0;
 var $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0;
 var $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0;
 var $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0;
 var $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0;
 var $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0;
 var $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0;
 var $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0;
 var $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0;
 var $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0;
 var $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0;
 var $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0;
 var $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0;
 var $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0;
 var $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0;
 var $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0;
 var $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0;
 var $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0;
 var $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0;
 var $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0;
 var $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0;
 var $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0;
 var $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0;
 var $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0;
 var $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0;
 var $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0;
 var $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $buf = 0, $c$0$lcssa$i = 0, $c$0$lcssa$i44 = 0, $c$04$i = 0, $c$04$i47 = 0, $d$0$i = 0, $d$0$i138 = 0, $d$0$i148 = 0, $delta$0$lcssa368 = 0;
 var $delta$0$lcssa369 = 0, $delta$0274 = 0, $delta$1 = 0, $delta$1$lcssa = 0, $div = 0, $dmulti$0$lcssa = 0, $dmulti$0282 = 0, $e$0$i = 0, $e$0$i163 = 0, $e$0$i185 = 0, $e$0$i54 = 0, $e$0$idx$val$i = 0, $e$0$idx$val$i175 = 0, $e$0$ph4$i$i = 0, $e$0$ph4$i$i102 = 0, $e$0$ph4$i$i120 = 0, $e$0$ph4$i$i67 = 0, $e$0$ph4$i$i81 = 0, $e$0$ph4$i$us$i = 0, $e$01$i$i = 0;
 var $e$01$i$i104 = 0, $e$01$i$i122 = 0, $e$01$i$i69 = 0, $e$01$i$i83 = 0, $e$01$i$us$i = 0, $e$02$i = 0, $e$02$i158 = 0, $e$02$i168 = 0, $e$02$i51 = 0, $e$03$i = 0, $e$03$i52 = 0, $e$08$i = 0, $e$09$i = 0, $ep$01$i$i = 0, $ep$01$i$i$i = 0, $ep$01$i$i188 = 0, $exitcond = 0, $exitcond$i = 0, $exitcond321 = 0, $h$idx$val$pre$i$pre = 0;
 var $hits$02$i = 0, $hits$02$i113 = 0, $hits$06$i = 0, $hits$1$i = 0, $hits$1$i$lcssa = 0, $hits$1$i114 = 0, $hits$1$i114$lcssa = 0, $hits$1$i127 = 0, $hits$1$i127$lcssa = 0, $i$0$i = 0, $i$0$i137 = 0, $i$0$i147 = 0, $i$01$i = 0, $i$04$i = 0, $i$1$i = 0, $i$3$i = 0, $id$i = 0, $info$i = 0, $info$i167 = 0, $isdigit = 0;
 var $isdigittmp = 0, $lno$03$i = 0, $lno$03$i169 = 0, $lno$1$i = 0, $lno$1$i$lcssa = 0, $lno$1$i180 = 0, $lno$1$i180$lcssa = 0, $maxtempo$0 = 0, $maxtempo$0$ph = 0, $maxtempo$1$be361 = 0, $maxtempo$1$lcssa = 0, $maxtempo$1248 = 0, $maxtempo$2 = 0, $maxtempo$2205 = 0, $maxtempo$2205$lcssa474 = 0, $maxtempo$3 = 0, $metalen$0238 = 0, $metalen$2273 = 0, $midimod$off$i$i = 0, $miditracklen$0$be362 = 0;
 var $miditracklen$0249 = 0, $miditracklen$1 = 0, $miditracklen$2$lcssa = 0, $miditracklen$2242 = 0, $miditracklen$3$lcssa = 0, $miditracklen$4 = 0, $miditracklen$5$lcssa = 0, $miditracklen$6 = 0, $mm = 0, $n$0$$1$i = 0, $n$0$$10$i = 0, $n$0$$11$i = 0, $n$0$$12$i = 0, $n$0$$13$i = 0, $n$0$$14$i = 0, $n$0$$15$i = 0, $n$0$$2$i = 0, $n$0$$3$i = 0, $n$0$$4$i = 0, $n$0$$5$i = 0;
 var $n$0$$6$i = 0, $n$0$$7$i = 0, $n$0$$8$i = 0, $n$0$$9$i = 0, $n$0$$i = 0, $n$0$lcssa$i = 0, $n$0$lcssa$i42 = 0, $n$02$i = 0, $n$02$i40 = 0, $n$04$i = 0, $n$05$i = 0, $n$1$i = 0, $n$1$i$lcssa = 0, $n$1$i181 = 0, $n$1$i181$lcssa = 0, $not$$i = 0, $notlhs = 0, $notrhs = 0, $or$cond = 0, $or$cond$i = 0;
 var $or$cond$i139 = 0, $or$cond$i149 = 0, $or$cond$i161 = 0, $or$cond$i172 = 0, $or$cond$i34 = 0, $or$cond1$i = 0, $or$cond1$i179 = 0, $or$cond208 = 0, $or$cond3$not = 0, $or$cond402 = 0, $or$cond6 = 0, $panhigh$0 = 0, $panhigh$0$ph = 0, $panhigh$1$be359 = 0, $panhigh$1$lcssa = 0, $panhigh$1246 = 0, $panhigh$2 = 0, $panhigh$3 = 0, $panhigh$4 = 0, $panhigh$4203 = 0;
 var $panhigh$4203$lcssa476 = 0, $panlow$0 = 0, $panlow$0$ph = 0, $panlow$1$be360 = 0, $panlow$1$lcssa = 0, $panlow$1247 = 0, $panlow$3 = 0, $panlow$4 = 0, $panlow$4204 = 0, $panlow$4204$lcssa475 = 0, $runningstatus$0250 = 0, $runningstatus$1 = 0, $sext = 0, $sext209 = 0, $sext210 = 0, $smax318 = 0, $storemerge$i = 0, $storemerge$i$lcssa = 0, $storemerge$i159 = 0, $storemerge$i170 = 0;
 var $storemerge4$i = 0, $t$0 = 0, $t$0$i = 0, $t$0$i39 = 0, $t$0$i48 = 0, $t$0$lcssa12$i = 0, $t$0$ph = 0, $t$01$i = 0, $t$01$i$pre$pre = 0, $t$01$i$pre339 = 0, $t$03$i = 0, $t$03$i37 = 0, $t$03$i43$pr = 0, $t$05$i = 0, $t$05$i$lcssa = 0, $t$05$i38 = 0, $t$05$i46 = 0, $t$05$i50 = 0, $t$1261 = 0, $t$2260 = 0;
 var $t$3256 = 0, $toff$07$i = 0, $toff$08$i = 0, $toff$1$i = 0, $toff$1$i$lcssa = 0, $toff$1$i184 = 0, $toff$1$i184$lcssa = 0, $ton$06$i = 0, $ton$07$i = 0, $ton$1$i = 0, $ton$1$i$lcssa = 0, $ton$1$i183 = 0, $ton$1$i183$lcssa = 0, $tp$01$i$i = 0, $tr$0$i = 0, $tr$0$i108 = 0, $tr$0$i115 = 0, $tr$0$i128 = 0, $tr$0$i55 = 0, $tr$0$i60 = 0;
 var $tr$0$i73 = 0, $tr$0$i87 = 0, $tr$0$i92 = 0, $tr$0$i96 = 0, $tr$0$us$i = 0, $tr$01$i = 0, $tr$01$i110 = 0, $tr$01$i56 = 0, $tr$01$i63 = 0, $tr$01$i77 = 0, $tr$01$i89 = 0, $tr$01$i94 = 0, $tr$02$i = 0, $tr$02$i58 = 0, $tr$02$i58$lcssa = 0, $tr$02$i65 = 0, $tr$02$i65$lcssa = 0, $tr$02$i79 = 0, $tr$02$i91 = 0, $tr$02$i98 = 0;
 var $tr$02$us$i = 0, $tr$03$i = 0, $tr$03$i100 = 0, $tr$03$i112 = 0, $tr$04$i = 0, $tr$05$i = 0, $tr$06$i = 0, $tr$07$i = 0, $ttp$0 = 0, $ttp$0272 = 0, $ttp$0275 = 0, $ttp$1 = 0, $ttp$1265 = 0, $ttp$1265$pre = 0, $ttp$1266 = 0, $ttp$2 = 0, $ttp$2263 = 0, $ttp$3 = 0, $ttp$3$lcssa = 0, $ttp$4 = 0;
 var $ttp$4255 = 0, $ttp$4257 = 0, $txt$i = 0, $v$05$i = 0, $v$06$i = 0, $v$1$i = 0, $v$1$i182 = 0, $vararg_buffer = 0, $vararg_buffer10 = 0, $vararg_buffer103 = 0, $vararg_buffer107 = 0, $vararg_buffer110 = 0, $vararg_buffer113 = 0, $vararg_buffer116 = 0, $vararg_buffer121 = 0, $vararg_buffer126 = 0, $vararg_buffer13 = 0, $vararg_buffer132 = 0, $vararg_buffer135 = 0, $vararg_buffer138 = 0;
 var $vararg_buffer141 = 0, $vararg_buffer144 = 0, $vararg_buffer147 = 0, $vararg_buffer152 = 0, $vararg_buffer155 = 0, $vararg_buffer158 = 0, $vararg_buffer16 = 0, $vararg_buffer164 = 0, $vararg_buffer167 = 0, $vararg_buffer170 = 0, $vararg_buffer173 = 0, $vararg_buffer176 = 0, $vararg_buffer179 = 0, $vararg_buffer182 = 0, $vararg_buffer185 = 0, $vararg_buffer188 = 0, $vararg_buffer191 = 0, $vararg_buffer194 = 0, $vararg_buffer197 = 0, $vararg_buffer200 = 0;
 var $vararg_buffer208 = 0, $vararg_buffer216 = 0, $vararg_buffer219 = 0, $vararg_buffer222 = 0, $vararg_buffer225 = 0, $vararg_buffer228 = 0, $vararg_buffer23 = 0, $vararg_buffer231 = 0, $vararg_buffer234 = 0, $vararg_buffer237 = 0, $vararg_buffer242 = 0, $vararg_buffer3 = 0, $vararg_buffer30 = 0, $vararg_buffer36 = 0, $vararg_buffer43 = 0, $vararg_buffer47 = 0, $vararg_buffer54 = 0, $vararg_buffer60 = 0, $vararg_buffer66 = 0, $vararg_buffer7 = 0;
 var $vararg_buffer72 = 0, $vararg_buffer77 = 0, $vararg_buffer80 = 0, $vararg_buffer85 = 0, $vararg_buffer90 = 0, $vararg_buffer93 = 0, $vararg_buffer98 = 0, $vararg_ptr1 = 0, $vararg_ptr101 = 0, $vararg_ptr102 = 0, $vararg_ptr106 = 0, $vararg_ptr119 = 0, $vararg_ptr120 = 0, $vararg_ptr124 = 0, $vararg_ptr125 = 0, $vararg_ptr129 = 0, $vararg_ptr130 = 0, $vararg_ptr131 = 0, $vararg_ptr150 = 0, $vararg_ptr151 = 0;
 var $vararg_ptr161 = 0, $vararg_ptr162 = 0, $vararg_ptr163 = 0, $vararg_ptr19 = 0, $vararg_ptr2 = 0, $vararg_ptr20 = 0, $vararg_ptr203 = 0, $vararg_ptr204 = 0, $vararg_ptr205 = 0, $vararg_ptr206 = 0, $vararg_ptr207 = 0, $vararg_ptr21 = 0, $vararg_ptr211 = 0, $vararg_ptr212 = 0, $vararg_ptr213 = 0, $vararg_ptr214 = 0, $vararg_ptr215 = 0, $vararg_ptr22 = 0, $vararg_ptr240 = 0, $vararg_ptr241 = 0;
 var $vararg_ptr245 = 0, $vararg_ptr26 = 0, $vararg_ptr27 = 0, $vararg_ptr28 = 0, $vararg_ptr29 = 0, $vararg_ptr33 = 0, $vararg_ptr34 = 0, $vararg_ptr35 = 0, $vararg_ptr39 = 0, $vararg_ptr40 = 0, $vararg_ptr41 = 0, $vararg_ptr42 = 0, $vararg_ptr46 = 0, $vararg_ptr50 = 0, $vararg_ptr51 = 0, $vararg_ptr52 = 0, $vararg_ptr53 = 0, $vararg_ptr57 = 0, $vararg_ptr58 = 0, $vararg_ptr59 = 0;
 var $vararg_ptr6 = 0, $vararg_ptr63 = 0, $vararg_ptr64 = 0, $vararg_ptr65 = 0, $vararg_ptr69 = 0, $vararg_ptr70 = 0, $vararg_ptr71 = 0, $vararg_ptr75 = 0, $vararg_ptr76 = 0, $vararg_ptr83 = 0, $vararg_ptr84 = 0, $vararg_ptr88 = 0, $vararg_ptr89 = 0, $vararg_ptr96 = 0, $vararg_ptr97 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1312|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer242 = sp + 696|0;
 $vararg_buffer237 = sp + 680|0;
 $vararg_buffer234 = sp + 672|0;
 $vararg_buffer231 = sp + 664|0;
 $vararg_buffer228 = sp + 656|0;
 $vararg_buffer225 = sp + 648|0;
 $vararg_buffer222 = sp + 640|0;
 $vararg_buffer219 = sp + 632|0;
 $vararg_buffer216 = sp + 624|0;
 $vararg_buffer208 = sp + 600|0;
 $vararg_buffer200 = sp + 576|0;
 $vararg_buffer197 = sp + 568|0;
 $vararg_buffer194 = sp + 560|0;
 $vararg_buffer191 = sp + 552|0;
 $vararg_buffer188 = sp + 544|0;
 $vararg_buffer185 = sp + 536|0;
 $vararg_buffer182 = sp + 528|0;
 $vararg_buffer179 = sp + 520|0;
 $vararg_buffer176 = sp + 512|0;
 $vararg_buffer173 = sp + 504|0;
 $vararg_buffer170 = sp + 496|0;
 $vararg_buffer167 = sp + 488|0;
 $vararg_buffer164 = sp + 480|0;
 $vararg_buffer158 = sp + 464|0;
 $vararg_buffer155 = sp + 456|0;
 $vararg_buffer152 = sp + 448|0;
 $vararg_buffer147 = sp + 432|0;
 $vararg_buffer144 = sp + 424|0;
 $vararg_buffer141 = sp + 416|0;
 $vararg_buffer138 = sp + 408|0;
 $vararg_buffer135 = sp + 400|0;
 $vararg_buffer132 = sp + 392|0;
 $vararg_buffer126 = sp + 376|0;
 $vararg_buffer121 = sp + 360|0;
 $vararg_buffer116 = sp + 344|0;
 $vararg_buffer113 = sp + 336|0;
 $vararg_buffer110 = sp + 328|0;
 $vararg_buffer107 = sp + 320|0;
 $vararg_buffer103 = sp + 312|0;
 $vararg_buffer98 = sp + 296|0;
 $vararg_buffer93 = sp + 280|0;
 $vararg_buffer90 = sp + 272|0;
 $vararg_buffer85 = sp + 256|0;
 $vararg_buffer80 = sp + 240|0;
 $vararg_buffer77 = sp + 232|0;
 $vararg_buffer72 = sp + 216|0;
 $vararg_buffer66 = sp + 200|0;
 $vararg_buffer60 = sp + 184|0;
 $vararg_buffer54 = sp + 168|0;
 $vararg_buffer47 = sp + 144|0;
 $vararg_buffer43 = sp + 136|0;
 $vararg_buffer36 = sp + 112|0;
 $vararg_buffer30 = sp + 96|0;
 $vararg_buffer23 = sp + 72|0;
 $vararg_buffer16 = sp + 48|0;
 $vararg_buffer13 = sp + 40|0;
 $vararg_buffer10 = sp + 32|0;
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $info$i167 = sp + 1272|0;
 $info$i = sp + 1240|0;
 $txt$i = sp + 984|0;
 $id$i = sp + 972|0;
 $mm = sp + 960|0;
 $buf = sp + 704|0;
 $0 = HEAP32[300431]|0; //@line 1425 "load_mid.cpp"
 $1 = ($0|0)==(0); //@line 1425 "load_mid.cpp"
 if (!($1)) {
  while(1) {
   (_sleep(1)|0); //@line 1425 "load_mid.cpp"
   $2 = HEAP32[300431]|0; //@line 1425 "load_mid.cpp"
   $3 = ($2|0)==(0); //@line 1425 "load_mid.cpp"
   if ($3) {
    break;
   }
  }
 }
 HEAP32[300431] = 1; //@line 1426 "load_mid.cpp"
 $4 = HEAPU8[$lpStream>>0]|(HEAPU8[$lpStream+1>>0]<<8)|(HEAPU8[$lpStream+2>>0]<<16)|(HEAPU8[$lpStream+3>>0]<<24); //@line 172 "load_mid.cpp"
 HEAP32[$id$i>>2] = $4; //@line 172 "load_mid.cpp"
 $5 = ((($id$i)) + 4|0); //@line 803 "load_mid.cpp"
 HEAP8[$5>>0] = 0; //@line 803 "load_mid.cpp"
 $6 = (_strcmp($id$i,527934)|0); //@line 804 "load_mid.cpp"
 $7 = ($6|0)==(0); //@line 804 "load_mid.cpp"
 if ($7) {
  $8 = ((($lpStream)) + 4|0); //@line 166 "load_mid.cpp"
  $9 = HEAP8[$8>>0]|0; //@line 166 "load_mid.cpp"
  $10 = ((($lpStream)) + 5|0); //@line 166 "load_mid.cpp"
  $11 = HEAP8[$10>>0]|0; //@line 166 "load_mid.cpp"
  $12 = ((($lpStream)) + 6|0); //@line 166 "load_mid.cpp"
  $13 = HEAP8[$12>>0]|0; //@line 166 "load_mid.cpp"
  $14 = ((($lpStream)) + 7|0); //@line 166 "load_mid.cpp"
  $15 = HEAP8[$14>>0]|0; //@line 166 "load_mid.cpp"
  $16 = $9&255; //@line 752 "load_mid.cpp"
  $17 = $16 << 24; //@line 752 "load_mid.cpp"
  $18 = $11&255; //@line 752 "load_mid.cpp"
  $19 = $18 << 16; //@line 752 "load_mid.cpp"
  $20 = $19 | $17; //@line 752 "load_mid.cpp"
  $21 = $13&255; //@line 752 "load_mid.cpp"
  $22 = $21 << 8; //@line 752 "load_mid.cpp"
  $23 = $20 | $22; //@line 752 "load_mid.cpp"
  $24 = $15&255; //@line 752 "load_mid.cpp"
  $25 = $23 | $24; //@line 752 "load_mid.cpp"
  $26 = ($25|0)==(6); //@line 804 "load_mid.cpp"
  if ($26) {
   $27 = (_calloc(1,60)|0); //@line 821 "load_mid.cpp"
   $28 = ($27|0)==(0|0); //@line 822 "load_mid.cpp"
   if ($28) {
    HEAP32[300431] = 0; //@line 1436 "load_mid.cpp"
    $$0 = 0;
    STACKTOP = sp;return ($$0|0); //@line 1880 "load_mid.cpp"
   }
   $29 = ((($27)) + 4|0); //@line 824 "load_mid.cpp"
   HEAP32[$29>>2] = 0; //@line 824 "load_mid.cpp"
   $30 = ((($27)) + 52|0); //@line 825 "load_mid.cpp"
   HEAP32[$30>>2] = 0; //@line 825 "load_mid.cpp"
   $31 = ((($27)) + 16|0); //@line 826 "load_mid.cpp"
   HEAP32[$31>>2] = 0; //@line 826 "load_mid.cpp"
   HEAP32[$27>>2] = $mm; //@line 1439 "load_mid.cpp"
   HEAP32[$mm>>2] = $lpStream; //@line 1440 "load_mid.cpp"
   $32 = ((($mm)) + 4|0); //@line 1441 "load_mid.cpp"
   HEAP32[$32>>2] = $dwMemLength; //@line 1441 "load_mid.cpp"
   $33 = ((($mm)) + 8|0); //@line 1442 "load_mid.cpp"
   HEAP32[$33>>2] = 0; //@line 1442 "load_mid.cpp"
   $34 = (_getenv((527939|0))|0); //@line 1444 "load_mid.cpp"
   HEAP32[$31>>2] = $34; //@line 1444 "load_mid.cpp"
   $35 = (_getenv((527951|0))|0); //@line 1445 "load_mid.cpp"
   $36 = ((($27)) + 20|0); //@line 1445 "load_mid.cpp"
   HEAP32[$36>>2] = $35; //@line 1445 "load_mid.cpp"
   _pat_resetsmp(); //@line 1446 "load_mid.cpp"
   _pat_init_patnames(); //@line 1447 "load_mid.cpp"
   $37 = HEAP32[$27>>2]|0; //@line 1448 "load_mid.cpp"
   $38 = ((($37)) + 8|0); //@line 140 "load_mid.cpp"
   HEAP32[$38>>2] = 8; //@line 140 "load_mid.cpp"
   $39 = HEAP32[$37>>2]|0; //@line 166 "load_mid.cpp"
   $40 = ((($39)) + 8|0); //@line 166 "load_mid.cpp"
   $41 = HEAP8[$40>>0]|0; //@line 166 "load_mid.cpp"
   $42 = ((($39)) + 9|0); //@line 166 "load_mid.cpp"
   $43 = HEAP8[$42>>0]|0; //@line 166 "load_mid.cpp"
   HEAP32[$38>>2] = 10; //@line 167 "load_mid.cpp"
   $44 = $41&255; //@line 759 "load_mid.cpp"
   $45 = $44 << 8; //@line 759 "load_mid.cpp"
   $46 = $43&255; //@line 759 "load_mid.cpp"
   $47 = $45 | $46; //@line 759 "load_mid.cpp"
   $sext = $47 << 16; //@line 1449 "load_mid.cpp"
   $48 = $sext >> 16; //@line 1449 "load_mid.cpp"
   $49 = ((($27)) + 32|0); //@line 1449 "load_mid.cpp"
   HEAP32[$49>>2] = $48; //@line 1449 "load_mid.cpp"
   $50 = ((($39)) + 10|0); //@line 166 "load_mid.cpp"
   $51 = HEAP8[$50>>0]|0; //@line 166 "load_mid.cpp"
   $52 = ((($39)) + 11|0); //@line 166 "load_mid.cpp"
   $53 = HEAP8[$52>>0]|0; //@line 166 "load_mid.cpp"
   HEAP32[$38>>2] = 12; //@line 167 "load_mid.cpp"
   $54 = $51&255; //@line 759 "load_mid.cpp"
   $55 = $54 << 8; //@line 759 "load_mid.cpp"
   $56 = $53&255; //@line 759 "load_mid.cpp"
   $57 = $55 | $56; //@line 759 "load_mid.cpp"
   $sext209 = $57 << 16; //@line 1450 "load_mid.cpp"
   $58 = $sext209 >> 16; //@line 1450 "load_mid.cpp"
   $59 = ((($27)) + 40|0); //@line 1450 "load_mid.cpp"
   HEAP32[$59>>2] = $58; //@line 1450 "load_mid.cpp"
   $60 = ((($39)) + 12|0); //@line 166 "load_mid.cpp"
   $61 = HEAP8[$60>>0]|0; //@line 166 "load_mid.cpp"
   $62 = ((($39)) + 13|0); //@line 166 "load_mid.cpp"
   $63 = HEAP8[$62>>0]|0; //@line 166 "load_mid.cpp"
   HEAP32[$38>>2] = 14; //@line 167 "load_mid.cpp"
   $64 = $61&255; //@line 759 "load_mid.cpp"
   $65 = $64 << 8; //@line 759 "load_mid.cpp"
   $66 = $63&255; //@line 759 "load_mid.cpp"
   $67 = $65 | $66; //@line 759 "load_mid.cpp"
   $sext210 = $67 << 16; //@line 1451 "load_mid.cpp"
   $68 = $sext210 >> 16; //@line 1451 "load_mid.cpp"
   $69 = ((($27)) + 36|0); //@line 1451 "load_mid.cpp"
   HEAP32[$69>>2] = $68; //@line 1451 "load_mid.cpp"
   $70 = ($48|0)==(0); //@line 1453 "load_mid.cpp"
   if ($70) {
    HEAP32[$59>>2] = 1; //@line 1453 "load_mid.cpp"
    $115 = 1;
   } else {
    $115 = $58;
   }
   $71 = $68 & 32768; //@line 1454 "load_mid.cpp"
   $72 = ($71|0)==(0); //@line 1454 "load_mid.cpp"
   if ($72) {
    $78 = ((($27)) + 44|0); //@line 1457 "load_mid.cpp"
    HEAP32[$78>>2] = $68; //@line 1457 "load_mid.cpp"
    $$pre$phi340Z2D = $78;$80 = $68;
   } else {
    $73 = $68 >>> 8; //@line 1455 "load_mid.cpp"
    $74 = $73 & 127; //@line 1455 "load_mid.cpp"
    $75 = $68 & 255; //@line 1455 "load_mid.cpp"
    $76 = Math_imul($74, $75)|0; //@line 1455 "load_mid.cpp"
    $77 = ((($27)) + 44|0); //@line 1455 "load_mid.cpp"
    HEAP32[$77>>2] = $76; //@line 1455 "load_mid.cpp"
    $$pre$phi340Z2D = $77;$80 = $76;
   }
   $79 = $80 << 2; //@line 1458 "load_mid.cpp"
   HEAP32[$$pre$phi340Z2D>>2] = $79; //@line 1458 "load_mid.cpp"
   $81 = ((($27)) + 48|0); //@line 1459 "load_mid.cpp"
   HEAP32[$81>>2] = 122; //@line 1459 "load_mid.cpp"
   $82 = ((($this)) + 59652|0); //@line 1460 "load_mid.cpp"
   HEAP32[$82>>2] = 0; //@line 1460 "load_mid.cpp"
   $83 = ((($27)) + 12|0); //@line 1461 "load_mid.cpp"
   HEAP32[$83>>2] = 0; //@line 1461 "load_mid.cpp"
   $84 = ((($27)) + 24|0); //@line 1462 "load_mid.cpp"
   HEAP32[$84>>2] = 6; //@line 1462 "load_mid.cpp"
   $85 = (_getenv((527965|0))|0); //@line 1463 "load_mid.cpp"
   $86 = ($85|0)==(0|0); //@line 1464 "load_mid.cpp"
   if ($86) {
    $96 = $79;$97 = 6;
   } else {
    $87 = HEAP8[$85>>0]|0; //@line 1464 "load_mid.cpp"
    $88 = $87&255; //@line 1464 "load_mid.cpp"
    $isdigittmp = (($88) + -48)|0; //@line 1464 "load_mid.cpp"
    $isdigit = ($isdigittmp>>>0)>(9); //@line 1464 "load_mid.cpp"
    $89 = ($87<<24>>24)==(48); //@line 1464 "load_mid.cpp"
    $or$cond208 = $89 | $isdigit; //@line 1464 "load_mid.cpp"
    if ($or$cond208) {
     $96 = $79;$97 = 6;
    } else {
     $90 = ((($85)) + 1|0); //@line 1464 "load_mid.cpp"
     $91 = HEAP8[$90>>0]|0; //@line 1464 "load_mid.cpp"
     $92 = ($91<<24>>24)==(0); //@line 1464 "load_mid.cpp"
     if ($92) {
      $93 = Math_imul($79, $isdigittmp)|0; //@line 1468 "load_mid.cpp"
      HEAP32[$84>>2] = $isdigittmp; //@line 1469 "load_mid.cpp"
      $94 = (($93|0) / 6)&-1; //@line 1470 "load_mid.cpp"
      HEAP32[$$pre$phi340Z2D>>2] = $94; //@line 1470 "load_mid.cpp"
      $96 = $94;$97 = $isdigittmp;
     } else {
      $96 = $79;$97 = 6;
     }
    }
   }
   $95 = Math_imul($96, $97)|0; //@line 256 "load_mid.cpp"
   $98 = $95 >>> 4; //@line 256 "load_mid.cpp"
   $99 = (($98>>>0) / ($97>>>0))&-1; //@line 256 "load_mid.cpp"
   $100 = ((($27)) + 28|0); //@line 1478 "load_mid.cpp"
   HEAP32[$100>>2] = $99; //@line 1478 "load_mid.cpp"
   $101 = ($99*10)|0; //@line 1478 "load_mid.cpp"
   $102 = ($97*255)|0; //@line 1478 "load_mid.cpp"
   $103 = ($101|0)<($102|0); //@line 1478 "load_mid.cpp"
   if ($103) {
    $dmulti$0282 = 1;
    while(1) {
     $104 = (($dmulti$0282) + 1)|0; //@line 1479 "load_mid.cpp"
     $105 = Math_imul($104, $96)|0; //@line 1480 "load_mid.cpp"
     $106 = Math_imul($105, $97)|0; //@line 256 "load_mid.cpp"
     $107 = $106 >>> 4; //@line 256 "load_mid.cpp"
     $108 = (($107>>>0) / ($97>>>0))&-1; //@line 256 "load_mid.cpp"
     $109 = ($108*10)|0; //@line 1478 "load_mid.cpp"
     $110 = ($109|0)<($102|0); //@line 1478 "load_mid.cpp"
     if ($110) {
      $dmulti$0282 = $104;
     } else {
      $$lcssa484 = $104;$$lcssa485 = $105;$$lcssa486 = $108;
      break;
     }
    }
    HEAP32[$$pre$phi340Z2D>>2] = $$lcssa485; //@line 1480 "load_mid.cpp"
    HEAP32[$100>>2] = $$lcssa486; //@line 1478 "load_mid.cpp"
    $dmulti$0$lcssa = $$lcssa484;
   } else {
    $dmulti$0$lcssa = 1;
   }
   $111 = ((($27)) + 8|0); //@line 1482 "load_mid.cpp"
   HEAP32[$111>>2] = 0; //@line 1482 "load_mid.cpp"
   _memset(($buf|0),0,256)|0; //@line 1483 "load_mid.cpp"
   $112 = ((($this)) + 59804|0); //@line 1487 "load_mid.cpp"
   HEAP8[$112>>0] = 0; //@line 1487 "load_mid.cpp"
   $113 = HEAP32[$36>>2]|0; //@line 1492 "load_mid.cpp"
   $114 = ($113|0)==(0|0); //@line 1492 "load_mid.cpp"
   if ($114) {
    $122 = 0;$maxtempo$0 = 0;$panhigh$0 = 64;$panlow$0 = 64;$t$0 = 0;
   } else {
    HEAP32[$vararg_buffer>>2] = $48; //@line 1493 "load_mid.cpp"
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 1493 "load_mid.cpp"
    HEAP32[$vararg_ptr1>>2] = $68; //@line 1493 "load_mid.cpp"
    $vararg_ptr2 = ((($vararg_buffer)) + 8|0); //@line 1493 "load_mid.cpp"
    HEAP32[$vararg_ptr2>>2] = $115; //@line 1493 "load_mid.cpp"
    (_printf(527977,$vararg_buffer)|0); //@line 1493 "load_mid.cpp"
    $$pre = HEAP32[$36>>2]|0; //@line 1498 "load_mid.cpp"
    $116 = ($$pre|0)!=(0|0); //@line 1498 "load_mid.cpp"
    $117 = ($dmulti$0$lcssa|0)>(1); //@line 1498 "load_mid.cpp"
    $or$cond = $117 & $116; //@line 1498 "load_mid.cpp"
    if ($or$cond) {
     $118 = HEAP32[$100>>2]|0; //@line 1500 "load_mid.cpp"
     HEAP32[$vararg_buffer3>>2] = $dmulti$0$lcssa; //@line 1499 "load_mid.cpp"
     $vararg_ptr6 = ((($vararg_buffer3)) + 4|0); //@line 1499 "load_mid.cpp"
     HEAP32[$vararg_ptr6>>2] = $118; //@line 1499 "load_mid.cpp"
     (_printf(528034,$vararg_buffer3)|0); //@line 1499 "load_mid.cpp"
     $maxtempo$0$ph = 0;$panhigh$0$ph = 64;$panlow$0$ph = 64;$t$0$ph = 0;
     label = 24;
    } else {
     $122 = $$pre;$maxtempo$0 = 0;$panhigh$0 = 64;$panlow$0 = 64;$t$0 = 0;
    }
   }
   while(1) {
    if ((label|0) == 24) {
     label = 0;
     $$pr = HEAP32[$36>>2]|0; //@line 1503 "load_mid.cpp"
     $122 = $$pr;$maxtempo$0 = $maxtempo$0$ph;$panhigh$0 = $panhigh$0$ph;$panlow$0 = $panlow$0$ph;$t$0 = $t$0$ph;
    }
    $119 = HEAP32[$59>>2]|0; //@line 1502 "load_mid.cpp"
    $120 = ($t$0>>>0)<($119>>>0); //@line 1502 "load_mid.cpp"
    $121 = ($122|0)!=(0|0); //@line 1503 "load_mid.cpp"
    if (!($120)) {
     break;
    }
    if ($121) {
     $123 = (($t$0) + 1)|0; //@line 1503 "load_mid.cpp"
     HEAP32[$vararg_buffer7>>2] = $123; //@line 1503 "load_mid.cpp"
     (_printf(528114,$vararg_buffer7)|0); //@line 1503 "load_mid.cpp"
    }
    $124 = HEAP32[$27>>2]|0; //@line 1504 "load_mid.cpp"
    $125 = ((($124)) + 8|0); //@line 172 "load_mid.cpp"
    $126 = HEAP32[$125>>2]|0; //@line 172 "load_mid.cpp"
    $127 = HEAP32[$124>>2]|0; //@line 172 "load_mid.cpp"
    $128 = (($127) + ($126)|0); //@line 172 "load_mid.cpp"
    $129 = HEAPU8[$128>>0]|(HEAPU8[$128+1>>0]<<8)|(HEAPU8[$128+2>>0]<<16)|(HEAPU8[$128+3>>0]<<24); //@line 172 "load_mid.cpp"
    HEAP32[$buf>>2] = $129; //@line 172 "load_mid.cpp"
    $130 = HEAP32[$125>>2]|0; //@line 173 "load_mid.cpp"
    $131 = (($130) + 4)|0; //@line 173 "load_mid.cpp"
    HEAP32[$125>>2] = $131; //@line 173 "load_mid.cpp"
    $132 = ((($buf)) + 4|0); //@line 1505 "load_mid.cpp"
    HEAP8[$132>>0] = 0; //@line 1505 "load_mid.cpp"
    $133 = (_strcmp($buf,528132)|0); //@line 1506 "load_mid.cpp"
    $134 = ($133|0)==(0); //@line 1506 "load_mid.cpp"
    if (!($134)) {
     label = 29;
     break;
    }
    $139 = HEAP32[$125>>2]|0; //@line 166 "load_mid.cpp"
    $140 = HEAP32[$124>>2]|0; //@line 166 "load_mid.cpp"
    $141 = (($140) + ($139)|0); //@line 166 "load_mid.cpp"
    $142 = HEAP8[$141>>0]|0; //@line 166 "load_mid.cpp"
    $$sum$i33 = (($139) + 1)|0; //@line 166 "load_mid.cpp"
    $143 = (($140) + ($$sum$i33)|0); //@line 166 "load_mid.cpp"
    $144 = HEAP8[$143>>0]|0; //@line 166 "load_mid.cpp"
    $$sum1$i = (($139) + 2)|0; //@line 166 "load_mid.cpp"
    $145 = (($140) + ($$sum1$i)|0); //@line 166 "load_mid.cpp"
    $146 = HEAP8[$145>>0]|0; //@line 166 "load_mid.cpp"
    $$sum2$i = (($139) + 3)|0; //@line 166 "load_mid.cpp"
    $147 = (($140) + ($$sum2$i)|0); //@line 166 "load_mid.cpp"
    $148 = HEAP8[$147>>0]|0; //@line 166 "load_mid.cpp"
    $149 = (($139) + 4)|0; //@line 167 "load_mid.cpp"
    HEAP32[$125>>2] = $149; //@line 167 "load_mid.cpp"
    $150 = $142&255; //@line 752 "load_mid.cpp"
    $151 = $150 << 24; //@line 752 "load_mid.cpp"
    $152 = $144&255; //@line 752 "load_mid.cpp"
    $153 = $152 << 16; //@line 752 "load_mid.cpp"
    $154 = $153 | $151; //@line 752 "load_mid.cpp"
    $155 = $146&255; //@line 752 "load_mid.cpp"
    $156 = $155 << 8; //@line 752 "load_mid.cpp"
    $157 = $154 | $156; //@line 752 "load_mid.cpp"
    $158 = $148&255; //@line 752 "load_mid.cpp"
    $159 = $157 | $158; //@line 752 "load_mid.cpp"
    $160 = ($t$0|0)==(0); //@line 1513 "load_mid.cpp"
    if (!($160)) {
     $161 = HEAP32[$49>>2]|0; //@line 1513 "load_mid.cpp"
     $162 = ($161|0)==(1); //@line 1513 "load_mid.cpp"
     if ($162) {
      HEAP32[$83>>2] = 0; //@line 370 "load_mid.cpp"
      $tr$01$i = HEAP32[$29>>2]|0; //@line 371 "load_mid.cpp"
      $163 = ($tr$01$i|0)==(0|0); //@line 371 "load_mid.cpp"
      if (!($163)) {
       $tr$02$i = $tr$01$i;
       while(1) {
        $164 = ((($tr$02$i)) + 25|0); //@line 372 "load_mid.cpp"
        HEAP8[$164>>0] = -1; //@line 372 "load_mid.cpp"
        $165 = ((($tr$02$i)) + 4|0); //@line 373 "load_mid.cpp"
        $166 = HEAP32[$165>>2]|0; //@line 373 "load_mid.cpp"
        $167 = ((($tr$02$i)) + 12|0); //@line 373 "load_mid.cpp"
        HEAP32[$167>>2] = $166; //@line 373 "load_mid.cpp"
        $168 = ((($tr$02$i)) + 20|0); //@line 374 "load_mid.cpp"
        HEAP32[$168>>2] = 0; //@line 374 "load_mid.cpp"
        $tr$0$i = HEAP32[$tr$02$i>>2]|0; //@line 371 "load_mid.cpp"
        $169 = ($tr$0$i|0)==(0|0); //@line 371 "load_mid.cpp"
        if ($169) {
         break;
        } else {
         $tr$02$i = $tr$0$i;
        }
       }
      }
     }
    }
    $170 = ($159|0)>(0); //@line 1514 "load_mid.cpp"
    L47: do {
     if ($170) {
      $171 = ((($27)) + 56|0); //@line 779 "load_mid.cpp"
      $172 = HEAP32[37544]|0; //@line 250 "load_mid.cpp"
      $$idx20$val = $124;$$pre$i = $149;$174 = $140;$maxtempo$1248 = $maxtempo$0;$miditracklen$0249 = $159;$panhigh$1246 = $panhigh$0;$panlow$1247 = $panlow$0;$runningstatus$0250 = 0;
      while(1) {
       $$phi$trans$insert$i = ((($$idx20$val)) + 8|0); //@line 1515 "load_mid.cpp"
       $175 = $$pre$i;$d$0$i = 0;$i$0$i = 0;
       while(1) {
        $173 = (($174) + ($175)|0); //@line 159 "load_mid.cpp"
        $176 = HEAP8[$173>>0]|0; //@line 159 "load_mid.cpp"
        $177 = (($175) + 1)|0; //@line 160 "load_mid.cpp"
        HEAP32[$$phi$trans$insert$i>>2] = $177; //@line 160 "load_mid.cpp"
        $178 = (($i$0$i) + 1)|0; //@line 774 "load_mid.cpp"
        $179 = $d$0$i << 7; //@line 775 "load_mid.cpp"
        $180 = $176&255; //@line 775 "load_mid.cpp"
        $181 = $180 & 127; //@line 775 "load_mid.cpp"
        $182 = $181 | $179; //@line 775 "load_mid.cpp"
        $183 = $180 & 128; //@line 776 "load_mid.cpp"
        $184 = ($183|0)!=(0); //@line 776 "load_mid.cpp"
        $185 = ($178|0)<(4); //@line 772 "load_mid.cpp"
        $or$cond$i = $185 & $184; //@line 776 "load_mid.cpp"
        if ($or$cond$i) {
         $175 = $177;$d$0$i = $182;$i$0$i = $178;
        } else {
         $$lcssa = $175;$$lcssa461 = $177;$$lcssa462 = $178;$$lcssa463 = $182;
         break;
        }
       }
       HEAP32[$171>>2] = $$lcssa463; //@line 779 "load_mid.cpp"
       $186 = (($miditracklen$0249) - ($$lcssa462))|0; //@line 1515 "load_mid.cpp"
       $187 = (($174) + ($$lcssa461)|0); //@line 159 "load_mid.cpp"
       $188 = HEAP8[$187>>0]|0; //@line 159 "load_mid.cpp"
       $189 = (($$lcssa) + 2)|0; //@line 160 "load_mid.cpp"
       HEAP32[$$phi$trans$insert$i>>2] = $189; //@line 160 "load_mid.cpp"
       $190 = (($186) + -1)|0; //@line 1517 "load_mid.cpp"
       $191 = $188&255; //@line 1518 "load_mid.cpp"
       $192 = $191 & 128; //@line 1518 "load_mid.cpp"
       $193 = ($192|0)==(0); //@line 1518 "load_mid.cpp"
       L54: do {
        if ($193) {
         $205 = $189;$210 = $188;$miditracklen$1 = $190;$runningstatus$1 = $runningstatus$0250;
        } else {
         switch ($191|0) {
         case 254: case 253: case 252: case 251: case 250: case 249: case 248: case 247: case 246: case 245: case 244: case 241:  {
          $205 = $189;$210 = $188;$miditracklen$1 = $190;$runningstatus$1 = $188;
          break L54;
          break;
         }
         default: {
         }
         }
         $194 = (($174) + ($189)|0); //@line 159 "load_mid.cpp"
         $195 = HEAP8[$194>>0]|0; //@line 159 "load_mid.cpp"
         $196 = (($$lcssa) + 3)|0; //@line 160 "load_mid.cpp"
         HEAP32[$$phi$trans$insert$i>>2] = $196; //@line 160 "load_mid.cpp"
         $197 = (($186) + -2)|0; //@line 1536 "load_mid.cpp"
         $205 = $196;$210 = $195;$miditracklen$1 = $197;$runningstatus$1 = $188;
        }
       } while(0);
       $198 = Math_imul($$lcssa463, $dmulti$0$lcssa)|0; //@line 1540 "load_mid.cpp"
       $199 = HEAP32[$83>>2]|0; //@line 1540 "load_mid.cpp"
       $200 = (($199) + ($198))|0; //@line 1540 "load_mid.cpp"
       HEAP32[$83>>2] = $200; //@line 1540 "load_mid.cpp"
       $201 = $runningstatus$1&255; //@line 1541 "load_mid.cpp"
       $202 = $201 & 15; //@line 1541 "load_mid.cpp"
       $203 = $201 & 240; //@line 1542 "load_mid.cpp"
       L58: do {
        switch ($203|0) {
        case 128:  {
         $204 = (($174) + ($205)|0); //@line 159 "load_mid.cpp"
         $206 = HEAP8[$204>>0]|0; //@line 159 "load_mid.cpp"
         $207 = (($205) + 1)|0; //@line 160 "load_mid.cpp"
         HEAP32[$$phi$trans$insert$i>>2] = $207; //@line 160 "load_mid.cpp"
         $208 = (($miditracklen$1) + -1)|0; //@line 1546 "load_mid.cpp"
         $209 = $210&255; //@line 1547 "load_mid.cpp"
         $tr$01$i56 = HEAP32[$29>>2]|0; //@line 407 "load_mid.cpp"
         $211 = ($tr$01$i56|0)==(0|0); //@line 407 "load_mid.cpp"
         L62: do {
          if (!($211)) {
           $218 = $200;$tr$02$i58 = $tr$01$i56;
           while(1) {
            $212 = ((($tr$02$i58)) + 12|0); //@line 394 "load_mid.cpp"
            $213 = HEAP32[$212>>2]|0; //@line 394 "load_mid.cpp"
            $214 = ($213|0)==(0|0); //@line 395 "load_mid.cpp"
            L65: do {
             if (!($214)) {
              $215 = ((($213)) + 4|0); //@line 395 "load_mid.cpp"
              $216 = HEAP32[$215>>2]|0; //@line 395 "load_mid.cpp"
              $217 = ($216>>>0)>($218>>>0); //@line 395 "load_mid.cpp"
              if ($217) {
               $219 = ((($tr$02$i58)) + 4|0); //@line 395 "load_mid.cpp"
               $220 = HEAP32[$219>>2]|0; //@line 395 "load_mid.cpp"
               $221 = ($220|0)==(0|0); //@line 396 "load_mid.cpp"
               if ($221) {
                break;
               } else {
                $e$0$ph4$i$i = $220;
               }
              } else {
               $e$0$ph4$i$i = $213;
              }
              $222 = ((($tr$02$i58)) + 27|0); //@line 388 "load_mid.cpp"
              $223 = ((($tr$02$i58)) + 25|0); //@line 384 "load_mid.cpp"
              $224 = ((($tr$02$i58)) + 26|0); //@line 385 "load_mid.cpp"
              $225 = ((($tr$02$i58)) + 20|0); //@line 386 "load_mid.cpp"
              $e$01$i$i = $e$0$ph4$i$i;
              while(1) {
               $226 = ((($e$01$i$i)) + 4|0); //@line 396 "load_mid.cpp"
               $227 = HEAP32[$226>>2]|0; //@line 396 "load_mid.cpp"
               $228 = ($227>>>0)>($218>>>0); //@line 396 "load_mid.cpp"
               if ($228) {
                break L65;
               }
               HEAP32[$212>>2] = $e$01$i$i; //@line 397 "load_mid.cpp"
               $229 = ((($e$01$i$i)) + 8|0); //@line 382 "load_mid.cpp"
               $230 = HEAP8[$229>>0]|0; //@line 382 "load_mid.cpp"
               $231 = ($230<<24>>24)==(0); //@line 382 "load_mid.cpp"
               if (!($231)) {
                $232 = ((($e$01$i$i)) + 10|0); //@line 383 "load_mid.cpp"
                $233 = HEAP8[$232>>0]|0; //@line 383 "load_mid.cpp"
                $234 = ($233<<24>>24)==(0); //@line 383 "load_mid.cpp"
                if ($234) {
                 $$sink$i$i$i = -1;
                } else {
                 $235 = ((($e$01$i$i)) + 9|0); //@line 383 "load_mid.cpp"
                 $236 = HEAP8[$235>>0]|0; //@line 383 "load_mid.cpp"
                 $$sink$i$i$i = $236;
                }
                HEAP8[$223>>0] = $$sink$i$i$i; //@line 384 "load_mid.cpp"
                $237 = HEAP8[$232>>0]|0; //@line 385 "load_mid.cpp"
                HEAP8[$224>>0] = $237; //@line 385 "load_mid.cpp"
                $238 = HEAP32[$226>>2]|0; //@line 386 "load_mid.cpp"
                HEAP32[$225>>2] = $238; //@line 386 "load_mid.cpp"
               }
               $239 = ((($e$01$i$i)) + 12|0); //@line 388 "load_mid.cpp"
               $240 = HEAP8[$239>>0]|0; //@line 388 "load_mid.cpp"
               $241 = ($240<<24>>24)==(8); //@line 388 "load_mid.cpp"
               if ($241) {
                $242 = ((($e$01$i$i)) + 13|0); //@line 388 "load_mid.cpp"
                $243 = HEAP8[$242>>0]|0; //@line 388 "load_mid.cpp"
                HEAP8[$222>>0] = $243; //@line 388 "load_mid.cpp"
               }
               $244 = HEAP32[$e$01$i$i>>2]|0; //@line 396 "load_mid.cpp"
               $245 = ($244|0)==(0|0); //@line 396 "load_mid.cpp"
               if ($245) {
                break;
               } else {
                $e$01$i$i = $244;
               }
              }
             }
            } while(0);
            $246 = ((($tr$02$i58)) + 24|0); //@line 409 "load_mid.cpp"
            $247 = HEAP8[$246>>0]|0; //@line 409 "load_mid.cpp"
            $248 = $247&255; //@line 409 "load_mid.cpp"
            $249 = ($248|0)==($202|0); //@line 409 "load_mid.cpp"
            if ($249) {
             $250 = ((($tr$02$i58)) + 25|0); //@line 409 "load_mid.cpp"
             $251 = HEAP8[$250>>0]|0; //@line 409 "load_mid.cpp"
             $252 = ($251<<24>>24)==($210<<24>>24); //@line 409 "load_mid.cpp"
             if ($252) {
              $$lcssa473 = $250;$tr$02$i58$lcssa = $tr$02$i58;
              break;
             }
            }
            $tr$0$i60 = HEAP32[$tr$02$i58>>2]|0; //@line 407 "load_mid.cpp"
            $253 = ($tr$0$i60|0)==(0|0); //@line 407 "load_mid.cpp"
            if ($253) {
             break L62;
            }
            $$pre337 = HEAP32[$83>>2]|0; //@line 408 "load_mid.cpp"
            $218 = $$pre337;$tr$02$i58 = $tr$0$i60;
           }
           $254 = (_calloc(1,16)|0); //@line 298 "load_mid.cpp"
           HEAP32[$254>>2] = 0; //@line 299 "load_mid.cpp"
           $255 = HEAP32[$83>>2]|0; //@line 300 "load_mid.cpp"
           $256 = ((($254)) + 4|0); //@line 300 "load_mid.cpp"
           HEAP32[$256>>2] = $255; //@line 300 "load_mid.cpp"
           $257 = ((($254)) + 8|0); //@line 301 "load_mid.cpp"
           ;HEAP8[$257>>0]=0|0;HEAP8[$257+1>>0]=0|0;HEAP8[$257+2>>0]=0|0;HEAP8[$257+3>>0]=0|0;HEAP8[$257+4>>0]=0|0;HEAP8[$257+5>>0]=0|0; //@line 302 "load_mid.cpp"
           HEAP8[$257>>0] = 1; //@line 554 "load_mid.cpp"
           $258 = HEAP8[$$lcssa473>>0]|0; //@line 555 "load_mid.cpp"
           $259 = ((($254)) + 9|0); //@line 555 "load_mid.cpp"
           HEAP8[$259>>0] = $258; //@line 555 "load_mid.cpp"
           $260 = ((($tr$02$i58$lcssa)) + 27|0); //@line 556 "load_mid.cpp"
           $261 = HEAP8[$260>>0]|0; //@line 556 "load_mid.cpp"
           $262 = ((($254)) + 11|0); //@line 556 "load_mid.cpp"
           HEAP8[$262>>0] = $261; //@line 556 "load_mid.cpp"
           __ZL13mid_add_eventP10_MIDHANDLEP9_MIDTRACKP9_MIDEVENT($tr$02$i58$lcssa,$254); //@line 557 "load_mid.cpp"
          }
         } while(0);
         $263 = HEAP32[$31>>2]|0; //@line 1549 "load_mid.cpp"
         $264 = ($263|0)==(0|0); //@line 1549 "load_mid.cpp"
         if ($264) {
          $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $208;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
          label = 248;
          break L58;
         }
         $265 = HEAP32[$83>>2]|0; //@line 1551 "load_mid.cpp"
         $266 = (($202) + 1)|0; //@line 1552 "load_mid.cpp"
         $267 = $206&255; //@line 1552 "load_mid.cpp"
         HEAP32[$vararg_buffer16>>2] = $t$0; //@line 1550 "load_mid.cpp"
         $vararg_ptr19 = ((($vararg_buffer16)) + 4|0); //@line 1550 "load_mid.cpp"
         HEAP32[$vararg_ptr19>>2] = $265; //@line 1550 "load_mid.cpp"
         $vararg_ptr20 = ((($vararg_buffer16)) + 8|0); //@line 1550 "load_mid.cpp"
         HEAP32[$vararg_ptr20>>2] = $266; //@line 1550 "load_mid.cpp"
         $vararg_ptr21 = ((($vararg_buffer16)) + 12|0); //@line 1550 "load_mid.cpp"
         HEAP32[$vararg_ptr21>>2] = $209; //@line 1550 "load_mid.cpp"
         $vararg_ptr22 = ((($vararg_buffer16)) + 16|0); //@line 1550 "load_mid.cpp"
         HEAP32[$vararg_ptr22>>2] = $267; //@line 1550 "load_mid.cpp"
         (_printf(528191,$vararg_buffer16)|0); //@line 1550 "load_mid.cpp"
         $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $208;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
         label = 248;
         break;
        }
        case 144:  {
         $268 = (($174) + ($205)|0); //@line 159 "load_mid.cpp"
         $269 = HEAP8[$268>>0]|0; //@line 159 "load_mid.cpp"
         $270 = (($205) + 1)|0; //@line 160 "load_mid.cpp"
         HEAP32[$$phi$trans$insert$i>>2] = $270; //@line 160 "load_mid.cpp"
         $271 = (($miditracklen$1) + -1)|0; //@line 1556 "load_mid.cpp"
         $272 = ($269<<24>>24)==(0); //@line 1557 "load_mid.cpp"
         $273 = $210&255; //@line 1558 "load_mid.cpp"
         if (!($272)) {
          $274 = (__ZL16mid_locate_trackP10_MIDHANDLEii($27,$202,$273)|0); //@line 1558 "load_mid.cpp"
          $275 = (_calloc(1,16)|0); //@line 298 "load_mid.cpp"
          HEAP32[$275>>2] = 0; //@line 299 "load_mid.cpp"
          $276 = HEAP32[$83>>2]|0; //@line 300 "load_mid.cpp"
          $277 = ((($275)) + 4|0); //@line 300 "load_mid.cpp"
          HEAP32[$277>>2] = $276; //@line 300 "load_mid.cpp"
          $278 = ((($275)) + 8|0); //@line 301 "load_mid.cpp"
          ;HEAP8[$278>>0]=0|0;HEAP8[$278+1>>0]=0|0;HEAP8[$278+2>>0]=0|0;HEAP8[$278+3>>0]=0|0;HEAP8[$278+4>>0]=0|0;HEAP8[$278+5>>0]=0|0; //@line 302 "load_mid.cpp"
          HEAP8[$278>>0] = 1; //@line 564 "load_mid.cpp"
          $279 = ((($275)) + 9|0); //@line 565 "load_mid.cpp"
          HEAP8[$279>>0] = $210; //@line 565 "load_mid.cpp"
          $280 = ((($274)) + 27|0); //@line 566 "load_mid.cpp"
          $281 = HEAP8[$280>>0]|0; //@line 566 "load_mid.cpp"
          $282 = ((($275)) + 11|0); //@line 566 "load_mid.cpp"
          HEAP8[$282>>0] = $281; //@line 566 "load_mid.cpp"
          $283 = ((($275)) + 10|0); //@line 567 "load_mid.cpp"
          HEAP8[$283>>0] = $269; //@line 567 "load_mid.cpp"
          __ZL13mid_add_eventP10_MIDHANDLEP9_MIDTRACKP9_MIDEVENT($274,$275); //@line 568 "load_mid.cpp"
          $284 = HEAP32[$31>>2]|0; //@line 1560 "load_mid.cpp"
          $285 = ($284|0)==(0|0); //@line 1560 "load_mid.cpp"
          if ($285) {
           $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $271;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
           label = 248;
           break L58;
          }
          $286 = $269&255; //@line 1559 "load_mid.cpp"
          $287 = HEAP32[$83>>2]|0; //@line 1562 "load_mid.cpp"
          $288 = (($202) + 1)|0; //@line 1563 "load_mid.cpp"
          HEAP32[$vararg_buffer23>>2] = $t$0; //@line 1561 "load_mid.cpp"
          $vararg_ptr26 = ((($vararg_buffer23)) + 4|0); //@line 1561 "load_mid.cpp"
          HEAP32[$vararg_ptr26>>2] = $287; //@line 1561 "load_mid.cpp"
          $vararg_ptr27 = ((($vararg_buffer23)) + 8|0); //@line 1561 "load_mid.cpp"
          HEAP32[$vararg_ptr27>>2] = $288; //@line 1561 "load_mid.cpp"
          $vararg_ptr28 = ((($vararg_buffer23)) + 12|0); //@line 1561 "load_mid.cpp"
          HEAP32[$vararg_ptr28>>2] = $273; //@line 1561 "load_mid.cpp"
          $vararg_ptr29 = ((($vararg_buffer23)) + 16|0); //@line 1561 "load_mid.cpp"
          HEAP32[$vararg_ptr29>>2] = $286; //@line 1561 "load_mid.cpp"
          (_printf(528238,$vararg_buffer23)|0); //@line 1561 "load_mid.cpp"
          $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $271;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
          label = 248;
          break L58;
         }
         $tr$01$i63 = HEAP32[$29>>2]|0; //@line 407 "load_mid.cpp"
         $289 = ($tr$01$i63|0)==(0|0); //@line 407 "load_mid.cpp"
         L95: do {
          if (!($289)) {
           $296 = $200;$tr$02$i65 = $tr$01$i63;
           while(1) {
            $290 = ((($tr$02$i65)) + 12|0); //@line 394 "load_mid.cpp"
            $291 = HEAP32[$290>>2]|0; //@line 394 "load_mid.cpp"
            $292 = ($291|0)==(0|0); //@line 395 "load_mid.cpp"
            L98: do {
             if (!($292)) {
              $293 = ((($291)) + 4|0); //@line 395 "load_mid.cpp"
              $294 = HEAP32[$293>>2]|0; //@line 395 "load_mid.cpp"
              $295 = ($294>>>0)>($296>>>0); //@line 395 "load_mid.cpp"
              if ($295) {
               $297 = ((($tr$02$i65)) + 4|0); //@line 395 "load_mid.cpp"
               $298 = HEAP32[$297>>2]|0; //@line 395 "load_mid.cpp"
               $299 = ($298|0)==(0|0); //@line 396 "load_mid.cpp"
               if ($299) {
                break;
               } else {
                $e$0$ph4$i$i67 = $298;
               }
              } else {
               $e$0$ph4$i$i67 = $291;
              }
              $300 = ((($tr$02$i65)) + 27|0); //@line 388 "load_mid.cpp"
              $301 = ((($tr$02$i65)) + 25|0); //@line 384 "load_mid.cpp"
              $302 = ((($tr$02$i65)) + 26|0); //@line 385 "load_mid.cpp"
              $303 = ((($tr$02$i65)) + 20|0); //@line 386 "load_mid.cpp"
              $e$01$i$i69 = $e$0$ph4$i$i67;
              while(1) {
               $304 = ((($e$01$i$i69)) + 4|0); //@line 396 "load_mid.cpp"
               $305 = HEAP32[$304>>2]|0; //@line 396 "load_mid.cpp"
               $306 = ($305>>>0)>($296>>>0); //@line 396 "load_mid.cpp"
               if ($306) {
                break L98;
               }
               HEAP32[$290>>2] = $e$01$i$i69; //@line 397 "load_mid.cpp"
               $307 = ((($e$01$i$i69)) + 8|0); //@line 382 "load_mid.cpp"
               $308 = HEAP8[$307>>0]|0; //@line 382 "load_mid.cpp"
               $309 = ($308<<24>>24)==(0); //@line 382 "load_mid.cpp"
               if (!($309)) {
                $310 = ((($e$01$i$i69)) + 10|0); //@line 383 "load_mid.cpp"
                $311 = HEAP8[$310>>0]|0; //@line 383 "load_mid.cpp"
                $312 = ($311<<24>>24)==(0); //@line 383 "load_mid.cpp"
                if ($312) {
                 $$sink$i$i$i70 = -1;
                } else {
                 $313 = ((($e$01$i$i69)) + 9|0); //@line 383 "load_mid.cpp"
                 $314 = HEAP8[$313>>0]|0; //@line 383 "load_mid.cpp"
                 $$sink$i$i$i70 = $314;
                }
                HEAP8[$301>>0] = $$sink$i$i$i70; //@line 384 "load_mid.cpp"
                $315 = HEAP8[$310>>0]|0; //@line 385 "load_mid.cpp"
                HEAP8[$302>>0] = $315; //@line 385 "load_mid.cpp"
                $316 = HEAP32[$304>>2]|0; //@line 386 "load_mid.cpp"
                HEAP32[$303>>2] = $316; //@line 386 "load_mid.cpp"
               }
               $317 = ((($e$01$i$i69)) + 12|0); //@line 388 "load_mid.cpp"
               $318 = HEAP8[$317>>0]|0; //@line 388 "load_mid.cpp"
               $319 = ($318<<24>>24)==(8); //@line 388 "load_mid.cpp"
               if ($319) {
                $320 = ((($e$01$i$i69)) + 13|0); //@line 388 "load_mid.cpp"
                $321 = HEAP8[$320>>0]|0; //@line 388 "load_mid.cpp"
                HEAP8[$300>>0] = $321; //@line 388 "load_mid.cpp"
               }
               $322 = HEAP32[$e$01$i$i69>>2]|0; //@line 396 "load_mid.cpp"
               $323 = ($322|0)==(0|0); //@line 396 "load_mid.cpp"
               if ($323) {
                break;
               } else {
                $e$01$i$i69 = $322;
               }
              }
             }
            } while(0);
            $324 = ((($tr$02$i65)) + 24|0); //@line 409 "load_mid.cpp"
            $325 = HEAP8[$324>>0]|0; //@line 409 "load_mid.cpp"
            $326 = $325&255; //@line 409 "load_mid.cpp"
            $327 = ($326|0)==($202|0); //@line 409 "load_mid.cpp"
            if ($327) {
             $328 = ((($tr$02$i65)) + 25|0); //@line 409 "load_mid.cpp"
             $329 = HEAP8[$328>>0]|0; //@line 409 "load_mid.cpp"
             $330 = ($329<<24>>24)==($210<<24>>24); //@line 409 "load_mid.cpp"
             if ($330) {
              $$lcssa471 = $328;$tr$02$i65$lcssa = $tr$02$i65;
              break;
             }
            }
            $tr$0$i73 = HEAP32[$tr$02$i65>>2]|0; //@line 407 "load_mid.cpp"
            $331 = ($tr$0$i73|0)==(0|0); //@line 407 "load_mid.cpp"
            if ($331) {
             break L95;
            }
            $$pre336 = HEAP32[$83>>2]|0; //@line 408 "load_mid.cpp"
            $296 = $$pre336;$tr$02$i65 = $tr$0$i73;
           }
           $332 = (_calloc(1,16)|0); //@line 298 "load_mid.cpp"
           HEAP32[$332>>2] = 0; //@line 299 "load_mid.cpp"
           $333 = HEAP32[$83>>2]|0; //@line 300 "load_mid.cpp"
           $334 = ((($332)) + 4|0); //@line 300 "load_mid.cpp"
           HEAP32[$334>>2] = $333; //@line 300 "load_mid.cpp"
           $335 = ((($332)) + 8|0); //@line 301 "load_mid.cpp"
           ;HEAP8[$335>>0]=0|0;HEAP8[$335+1>>0]=0|0;HEAP8[$335+2>>0]=0|0;HEAP8[$335+3>>0]=0|0;HEAP8[$335+4>>0]=0|0;HEAP8[$335+5>>0]=0|0; //@line 302 "load_mid.cpp"
           HEAP8[$335>>0] = 1; //@line 554 "load_mid.cpp"
           $336 = HEAP8[$$lcssa471>>0]|0; //@line 555 "load_mid.cpp"
           $337 = ((($332)) + 9|0); //@line 555 "load_mid.cpp"
           HEAP8[$337>>0] = $336; //@line 555 "load_mid.cpp"
           $338 = ((($tr$02$i65$lcssa)) + 27|0); //@line 556 "load_mid.cpp"
           $339 = HEAP8[$338>>0]|0; //@line 556 "load_mid.cpp"
           $340 = ((($332)) + 11|0); //@line 556 "load_mid.cpp"
           HEAP8[$340>>0] = $339; //@line 556 "load_mid.cpp"
           __ZL13mid_add_eventP10_MIDHANDLEP9_MIDTRACKP9_MIDEVENT($tr$02$i65$lcssa,$332); //@line 557 "load_mid.cpp"
          }
         } while(0);
         $341 = HEAP32[$31>>2]|0; //@line 1568 "load_mid.cpp"
         $342 = ($341|0)==(0|0); //@line 1568 "load_mid.cpp"
         if ($342) {
          $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $271;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
          label = 248;
          break L58;
         }
         $343 = HEAP32[$83>>2]|0; //@line 1570 "load_mid.cpp"
         $344 = (($202) + 1)|0; //@line 1571 "load_mid.cpp"
         HEAP32[$vararg_buffer30>>2] = $t$0; //@line 1569 "load_mid.cpp"
         $vararg_ptr33 = ((($vararg_buffer30)) + 4|0); //@line 1569 "load_mid.cpp"
         HEAP32[$vararg_ptr33>>2] = $343; //@line 1569 "load_mid.cpp"
         $vararg_ptr34 = ((($vararg_buffer30)) + 8|0); //@line 1569 "load_mid.cpp"
         HEAP32[$vararg_ptr34>>2] = $344; //@line 1569 "load_mid.cpp"
         $vararg_ptr35 = ((($vararg_buffer30)) + 12|0); //@line 1569 "load_mid.cpp"
         HEAP32[$vararg_ptr35>>2] = $273; //@line 1569 "load_mid.cpp"
         (_printf(528279,$vararg_buffer30)|0); //@line 1569 "load_mid.cpp"
         $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $271;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
         label = 248;
         break;
        }
        case 160:  {
         $345 = (($174) + ($205)|0); //@line 159 "load_mid.cpp"
         $346 = HEAP8[$345>>0]|0; //@line 159 "load_mid.cpp"
         $347 = (($205) + 1)|0; //@line 160 "load_mid.cpp"
         HEAP32[$$phi$trans$insert$i>>2] = $347; //@line 160 "load_mid.cpp"
         $348 = (($miditracklen$1) + -1)|0; //@line 1576 "load_mid.cpp"
         $349 = HEAP32[$31>>2]|0; //@line 1577 "load_mid.cpp"
         $350 = ($349|0)==(0|0); //@line 1577 "load_mid.cpp"
         if ($350) {
          $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $348;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
          label = 248;
          break L58;
         }
         $351 = (($202) + 1)|0; //@line 1578 "load_mid.cpp"
         $352 = $210&255; //@line 1578 "load_mid.cpp"
         $353 = $346&255; //@line 1578 "load_mid.cpp"
         HEAP32[$vararg_buffer36>>2] = $t$0; //@line 1578 "load_mid.cpp"
         $vararg_ptr39 = ((($vararg_buffer36)) + 4|0); //@line 1578 "load_mid.cpp"
         HEAP32[$vararg_ptr39>>2] = $200; //@line 1578 "load_mid.cpp"
         $vararg_ptr40 = ((($vararg_buffer36)) + 8|0); //@line 1578 "load_mid.cpp"
         HEAP32[$vararg_ptr40>>2] = $351; //@line 1578 "load_mid.cpp"
         $vararg_ptr41 = ((($vararg_buffer36)) + 12|0); //@line 1578 "load_mid.cpp"
         HEAP32[$vararg_ptr41>>2] = $352; //@line 1578 "load_mid.cpp"
         $vararg_ptr42 = ((($vararg_buffer36)) + 16|0); //@line 1578 "load_mid.cpp"
         HEAP32[$vararg_ptr42>>2] = $353; //@line 1578 "load_mid.cpp"
         (_printf(528313,$vararg_buffer36)|0); //@line 1578 "load_mid.cpp"
         $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $348;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
         label = 248;
         break;
        }
        case 176:  {
         $354 = (($174) + ($205)|0); //@line 159 "load_mid.cpp"
         $355 = HEAP8[$354>>0]|0; //@line 159 "load_mid.cpp"
         $356 = (($205) + 1)|0; //@line 160 "load_mid.cpp"
         HEAP32[$$phi$trans$insert$i>>2] = $356; //@line 160 "load_mid.cpp"
         $357 = (($miditracklen$1) + -1)|0; //@line 1582 "load_mid.cpp"
         $358 = $210&255; //@line 1583 "load_mid.cpp"
         L126: do {
          switch ($358|0) {
          case 1:  {
           $359 = $355&255; //@line 1585 "load_mid.cpp"
           $tr$01$i77 = HEAP32[$29>>2]|0; //@line 591 "load_mid.cpp"
           $360 = ($tr$01$i77|0)==(0|0); //@line 591 "load_mid.cpp"
           if ($360) {
            $panhigh$3 = $panhigh$1246;$panlow$3 = $panlow$1247;
            break L126;
           }
           $361 = ($355<<24>>24)==(0); //@line 574 "load_mid.cpp"
           $362 = (128 - ($359))|0; //@line 576 "load_mid.cpp"
           if ($361) {
            $tr$02$us$i = $tr$01$i77;
            while(1) {
             $372 = ((($tr$02$us$i)) + 24|0); //@line 592 "load_mid.cpp"
             $373 = HEAP8[$372>>0]|0; //@line 592 "load_mid.cpp"
             $374 = $373&255; //@line 592 "load_mid.cpp"
             $375 = ($374|0)==($202|0); //@line 592 "load_mid.cpp"
             do {
              if ($375) {
               $376 = HEAP32[$83>>2]|0; //@line 593 "load_mid.cpp"
               $377 = ((($tr$02$us$i)) + 12|0); //@line 394 "load_mid.cpp"
               $378 = HEAP32[$377>>2]|0; //@line 394 "load_mid.cpp"
               $379 = ($378|0)==(0|0); //@line 395 "load_mid.cpp"
               L134: do {
                if (!($379)) {
                 $380 = ((($378)) + 4|0); //@line 395 "load_mid.cpp"
                 $381 = HEAP32[$380>>2]|0; //@line 395 "load_mid.cpp"
                 $382 = ($381>>>0)>($376>>>0); //@line 395 "load_mid.cpp"
                 if ($382) {
                  $383 = ((($tr$02$us$i)) + 4|0); //@line 395 "load_mid.cpp"
                  $384 = HEAP32[$383>>2]|0; //@line 395 "load_mid.cpp"
                  $385 = ($384|0)==(0|0); //@line 396 "load_mid.cpp"
                  if ($385) {
                   break;
                  } else {
                   $e$0$ph4$i$us$i = $384;
                  }
                 } else {
                  $e$0$ph4$i$us$i = $378;
                 }
                 $386 = ((($tr$02$us$i)) + 27|0); //@line 388 "load_mid.cpp"
                 $387 = ((($tr$02$us$i)) + 25|0); //@line 384 "load_mid.cpp"
                 $388 = ((($tr$02$us$i)) + 26|0); //@line 385 "load_mid.cpp"
                 $389 = ((($tr$02$us$i)) + 20|0); //@line 386 "load_mid.cpp"
                 $e$01$i$us$i = $e$0$ph4$i$us$i;
                 while(1) {
                  $390 = ((($e$01$i$us$i)) + 4|0); //@line 396 "load_mid.cpp"
                  $391 = HEAP32[$390>>2]|0; //@line 396 "load_mid.cpp"
                  $392 = ($391>>>0)>($376>>>0); //@line 396 "load_mid.cpp"
                  if ($392) {
                   break L134;
                  }
                  HEAP32[$377>>2] = $e$01$i$us$i; //@line 397 "load_mid.cpp"
                  $393 = ((($e$01$i$us$i)) + 8|0); //@line 382 "load_mid.cpp"
                  $394 = HEAP8[$393>>0]|0; //@line 382 "load_mid.cpp"
                  $395 = ($394<<24>>24)==(0); //@line 382 "load_mid.cpp"
                  if (!($395)) {
                   $396 = ((($e$01$i$us$i)) + 10|0); //@line 383 "load_mid.cpp"
                   $397 = HEAP8[$396>>0]|0; //@line 383 "load_mid.cpp"
                   $398 = ($397<<24>>24)==(0); //@line 383 "load_mid.cpp"
                   if ($398) {
                    $$sink$i$i$us$i = -1;
                   } else {
                    $399 = ((($e$01$i$us$i)) + 9|0); //@line 383 "load_mid.cpp"
                    $400 = HEAP8[$399>>0]|0; //@line 383 "load_mid.cpp"
                    $$sink$i$i$us$i = $400;
                   }
                   HEAP8[$387>>0] = $$sink$i$i$us$i; //@line 384 "load_mid.cpp"
                   $401 = HEAP8[$396>>0]|0; //@line 385 "load_mid.cpp"
                   HEAP8[$388>>0] = $401; //@line 385 "load_mid.cpp"
                   $402 = HEAP32[$390>>2]|0; //@line 386 "load_mid.cpp"
                   HEAP32[$389>>2] = $402; //@line 386 "load_mid.cpp"
                  }
                  $403 = ((($e$01$i$us$i)) + 12|0); //@line 388 "load_mid.cpp"
                  $404 = HEAP8[$403>>0]|0; //@line 388 "load_mid.cpp"
                  $405 = ($404<<24>>24)==(8); //@line 388 "load_mid.cpp"
                  if ($405) {
                   $406 = ((($e$01$i$us$i)) + 13|0); //@line 388 "load_mid.cpp"
                   $407 = HEAP8[$406>>0]|0; //@line 388 "load_mid.cpp"
                   HEAP8[$386>>0] = $407; //@line 388 "load_mid.cpp"
                  }
                  $408 = HEAP32[$e$01$i$us$i>>2]|0; //@line 396 "load_mid.cpp"
                  $409 = ($408|0)==(0|0); //@line 396 "load_mid.cpp"
                  if ($409) {
                   break;
                  } else {
                   $e$01$i$us$i = $408;
                  }
                 }
                }
               } while(0);
               $410 = ((($tr$02$us$i)) + 25|0); //@line 594 "load_mid.cpp"
               $411 = HEAP8[$410>>0]|0; //@line 594 "load_mid.cpp"
               $412 = ($411<<24>>24)==(-1); //@line 594 "load_mid.cpp"
               if ($412) {
                break;
               }
               $413 = (_calloc(1,16)|0); //@line 298 "load_mid.cpp"
               HEAP32[$413>>2] = 0; //@line 299 "load_mid.cpp"
               $414 = HEAP32[$83>>2]|0; //@line 300 "load_mid.cpp"
               $415 = ((($413)) + 4|0); //@line 300 "load_mid.cpp"
               HEAP32[$415>>2] = $414; //@line 300 "load_mid.cpp"
               $416 = ((($413)) + 8|0); //@line 301 "load_mid.cpp"
               $417 = ((($413)) + 12|0); //@line 597 "load_mid.cpp"
               HEAP8[$416>>0]=0&255;HEAP8[$416+1>>0]=(0>>8)&255;HEAP8[$416+2>>0]=(0>>16)&255;HEAP8[$416+3>>0]=0>>24; //@line 596 "load_mid.cpp"
               HEAP8[$417>>0] = 6; //@line 597 "load_mid.cpp"
               $418 = ((($413)) + 13|0); //@line 598 "load_mid.cpp"
               HEAP8[$418>>0] = 0; //@line 598 "load_mid.cpp"
               __ZL13mid_add_eventP10_MIDHANDLEP9_MIDTRACKP9_MIDEVENT($tr$02$us$i,$413); //@line 599 "load_mid.cpp"
              }
             } while(0);
             $tr$0$us$i = HEAP32[$tr$02$us$i>>2]|0; //@line 591 "load_mid.cpp"
             $419 = ($tr$0$us$i|0)==(0|0); //@line 591 "load_mid.cpp"
             if ($419) {
              $panhigh$3 = $panhigh$1246;$panlow$3 = $panlow$1247;
              break L126;
             } else {
              $tr$02$us$i = $tr$0$us$i;
             }
            }
           }
           $midimod$off$i$i = (($359) + 3)|0; //@line 581 "load_mid.cpp"
           $363 = ($midimod$off$i$i>>>0)<(7); //@line 581 "load_mid.cpp"
           $div = ($355&255) >>> 2; //@line 580 "load_mid.cpp"
           $364 = $div&255; //@line 580 "load_mid.cpp"
           $$op$i$i = $364 << 4; //@line 582 "load_mid.cpp"
           $$op$op$i$i = $$op$i$i | 15; //@line 582 "load_mid.cpp"
           $365 = $$op$op$i$i&255; //@line 582 "load_mid.cpp"
           $366 = $363 ? 31 : $365; //@line 582 "load_mid.cpp"
           $$off$i$i = (($362) + 3)|0; //@line 577 "load_mid.cpp"
           $367 = ($$off$i$i>>>0)<(7); //@line 577 "load_mid.cpp"
           $368 = (($362|0) / 4)&-1; //@line 576 "load_mid.cpp"
           $$op2$i$i = $368 | 240; //@line 578 "load_mid.cpp"
           $369 = $$op2$i$i&255; //@line 578 "load_mid.cpp"
           $370 = $367 ? -15 : $369; //@line 578 "load_mid.cpp"
           $371 = ($355&255)>(63); //@line 575 "load_mid.cpp"
           $$$i = $371 ? $370 : $366; //@line 575 "load_mid.cpp"
           $tr$02$i79 = $tr$01$i77;
           while(1) {
            $420 = ((($tr$02$i79)) + 24|0); //@line 592 "load_mid.cpp"
            $421 = HEAP8[$420>>0]|0; //@line 592 "load_mid.cpp"
            $422 = $421&255; //@line 592 "load_mid.cpp"
            $423 = ($422|0)==($202|0); //@line 592 "load_mid.cpp"
            do {
             if ($423) {
              $424 = HEAP32[$83>>2]|0; //@line 593 "load_mid.cpp"
              $425 = ((($tr$02$i79)) + 12|0); //@line 394 "load_mid.cpp"
              $426 = HEAP32[$425>>2]|0; //@line 394 "load_mid.cpp"
              $427 = ($426|0)==(0|0); //@line 395 "load_mid.cpp"
              L159: do {
               if (!($427)) {
                $428 = ((($426)) + 4|0); //@line 395 "load_mid.cpp"
                $429 = HEAP32[$428>>2]|0; //@line 395 "load_mid.cpp"
                $430 = ($429>>>0)>($424>>>0); //@line 395 "load_mid.cpp"
                if ($430) {
                 $431 = ((($tr$02$i79)) + 4|0); //@line 395 "load_mid.cpp"
                 $432 = HEAP32[$431>>2]|0; //@line 395 "load_mid.cpp"
                 $433 = ($432|0)==(0|0); //@line 396 "load_mid.cpp"
                 if ($433) {
                  break;
                 } else {
                  $e$0$ph4$i$i81 = $432;
                 }
                } else {
                 $e$0$ph4$i$i81 = $426;
                }
                $434 = ((($tr$02$i79)) + 27|0); //@line 388 "load_mid.cpp"
                $435 = ((($tr$02$i79)) + 25|0); //@line 384 "load_mid.cpp"
                $436 = ((($tr$02$i79)) + 26|0); //@line 385 "load_mid.cpp"
                $437 = ((($tr$02$i79)) + 20|0); //@line 386 "load_mid.cpp"
                $e$01$i$i83 = $e$0$ph4$i$i81;
                while(1) {
                 $438 = ((($e$01$i$i83)) + 4|0); //@line 396 "load_mid.cpp"
                 $439 = HEAP32[$438>>2]|0; //@line 396 "load_mid.cpp"
                 $440 = ($439>>>0)>($424>>>0); //@line 396 "load_mid.cpp"
                 if ($440) {
                  break L159;
                 }
                 HEAP32[$425>>2] = $e$01$i$i83; //@line 397 "load_mid.cpp"
                 $441 = ((($e$01$i$i83)) + 8|0); //@line 382 "load_mid.cpp"
                 $442 = HEAP8[$441>>0]|0; //@line 382 "load_mid.cpp"
                 $443 = ($442<<24>>24)==(0); //@line 382 "load_mid.cpp"
                 if (!($443)) {
                  $444 = ((($e$01$i$i83)) + 10|0); //@line 383 "load_mid.cpp"
                  $445 = HEAP8[$444>>0]|0; //@line 383 "load_mid.cpp"
                  $446 = ($445<<24>>24)==(0); //@line 383 "load_mid.cpp"
                  if ($446) {
                   $$sink$i$i$i84 = -1;
                  } else {
                   $447 = ((($e$01$i$i83)) + 9|0); //@line 383 "load_mid.cpp"
                   $448 = HEAP8[$447>>0]|0; //@line 383 "load_mid.cpp"
                   $$sink$i$i$i84 = $448;
                  }
                  HEAP8[$435>>0] = $$sink$i$i$i84; //@line 384 "load_mid.cpp"
                  $449 = HEAP8[$444>>0]|0; //@line 385 "load_mid.cpp"
                  HEAP8[$436>>0] = $449; //@line 385 "load_mid.cpp"
                  $450 = HEAP32[$438>>2]|0; //@line 386 "load_mid.cpp"
                  HEAP32[$437>>2] = $450; //@line 386 "load_mid.cpp"
                 }
                 $451 = ((($e$01$i$i83)) + 12|0); //@line 388 "load_mid.cpp"
                 $452 = HEAP8[$451>>0]|0; //@line 388 "load_mid.cpp"
                 $453 = ($452<<24>>24)==(8); //@line 388 "load_mid.cpp"
                 if ($453) {
                  $454 = ((($e$01$i$i83)) + 13|0); //@line 388 "load_mid.cpp"
                  $455 = HEAP8[$454>>0]|0; //@line 388 "load_mid.cpp"
                  HEAP8[$434>>0] = $455; //@line 388 "load_mid.cpp"
                 }
                 $456 = HEAP32[$e$01$i$i83>>2]|0; //@line 396 "load_mid.cpp"
                 $457 = ($456|0)==(0|0); //@line 396 "load_mid.cpp"
                 if ($457) {
                  break;
                 } else {
                  $e$01$i$i83 = $456;
                 }
                }
               }
              } while(0);
              $458 = ((($tr$02$i79)) + 25|0); //@line 594 "load_mid.cpp"
              $459 = HEAP8[$458>>0]|0; //@line 594 "load_mid.cpp"
              $460 = ($459<<24>>24)==(-1); //@line 594 "load_mid.cpp"
              if ($460) {
               break;
              }
              $461 = (_calloc(1,16)|0); //@line 298 "load_mid.cpp"
              HEAP32[$461>>2] = 0; //@line 299 "load_mid.cpp"
              $462 = HEAP32[$83>>2]|0; //@line 300 "load_mid.cpp"
              $463 = ((($461)) + 4|0); //@line 300 "load_mid.cpp"
              HEAP32[$463>>2] = $462; //@line 300 "load_mid.cpp"
              $464 = ((($461)) + 8|0); //@line 301 "load_mid.cpp"
              $465 = ((($461)) + 12|0); //@line 597 "load_mid.cpp"
              ;HEAP8[$464>>0]=0|0;HEAP8[$464+1>>0]=0|0;HEAP8[$464+2>>0]=0|0;HEAP8[$464+3>>0]=0|0;HEAP8[$464+4>>0]=0|0; //@line 596 "load_mid.cpp"
              HEAP8[$465>>0] = 6; //@line 597 "load_mid.cpp"
              $466 = ((($461)) + 13|0); //@line 598 "load_mid.cpp"
              HEAP8[$466>>0] = $$$i; //@line 598 "load_mid.cpp"
              __ZL13mid_add_eventP10_MIDHANDLEP9_MIDTRACKP9_MIDEVENT($tr$02$i79,$461); //@line 599 "load_mid.cpp"
             }
            } while(0);
            $tr$0$i87 = HEAP32[$tr$02$i79>>2]|0; //@line 591 "load_mid.cpp"
            $467 = ($tr$0$i87|0)==(0|0); //@line 591 "load_mid.cpp"
            if ($467) {
             $panhigh$3 = $panhigh$1246;$panlow$3 = $panlow$1247;
             break;
            } else {
             $tr$02$i79 = $tr$0$i87;
            }
           }
           break;
          }
          case 7:  {
           $tr$01$i89 = HEAP32[$29>>2]|0; //@line 611 "load_mid.cpp"
           $468 = ($tr$01$i89|0)==(0|0); //@line 611 "load_mid.cpp"
           if ($468) {
            $panhigh$3 = $panhigh$1246;$panlow$3 = $panlow$1247;
            break L126;
           } else {
            $tr$02$i91 = $tr$01$i89;
           }
           while(1) {
            $469 = ((($tr$02$i91)) + 24|0); //@line 612 "load_mid.cpp"
            $470 = HEAP8[$469>>0]|0; //@line 612 "load_mid.cpp"
            $471 = $470&255; //@line 612 "load_mid.cpp"
            $472 = ($471|0)==($202|0); //@line 612 "load_mid.cpp"
            if ($472) {
             $473 = (_calloc(1,16)|0); //@line 298 "load_mid.cpp"
             HEAP32[$473>>2] = 0; //@line 299 "load_mid.cpp"
             $474 = HEAP32[$83>>2]|0; //@line 300 "load_mid.cpp"
             $475 = ((($473)) + 4|0); //@line 300 "load_mid.cpp"
             HEAP32[$475>>2] = $474; //@line 300 "load_mid.cpp"
             $476 = ((($473)) + 8|0); //@line 301 "load_mid.cpp"
             $477 = ((($473)) + 12|0); //@line 615 "load_mid.cpp"
             HEAP8[$476>>0]=0&255;HEAP8[$476+1>>0]=(0>>8)&255;HEAP8[$476+2>>0]=(0>>16)&255;HEAP8[$476+3>>0]=0>>24; //@line 614 "load_mid.cpp"
             HEAP8[$477>>0] = 7; //@line 615 "load_mid.cpp"
             $478 = ((($473)) + 13|0); //@line 616 "load_mid.cpp"
             HEAP8[$478>>0] = $355; //@line 616 "load_mid.cpp"
             __ZL13mid_add_eventP10_MIDHANDLEP9_MIDTRACKP9_MIDEVENT($tr$02$i91,$473); //@line 617 "load_mid.cpp"
            }
            $tr$0$i92 = HEAP32[$tr$02$i91>>2]|0; //@line 611 "load_mid.cpp"
            $479 = ($tr$0$i92|0)==(0|0); //@line 611 "load_mid.cpp"
            if ($479) {
             $panhigh$3 = $panhigh$1246;$panlow$3 = $panlow$1247;
             break;
            } else {
             $tr$02$i91 = $tr$0$i92;
            }
           }
           break;
          }
          case 10:  {
           $480 = $355&255; //@line 1591 "load_mid.cpp"
           $481 = ($480|0)<($panlow$1247|0); //@line 1591 "load_mid.cpp"
           $$panlow$1 = $481 ? $480 : $panlow$1247; //@line 1591 "load_mid.cpp"
           $482 = ($480|0)>($panhigh$1246|0); //@line 1592 "load_mid.cpp"
           $panhigh$2 = $482 ? $480 : $panhigh$1246; //@line 1592 "load_mid.cpp"
           $tr$01$i94 = HEAP32[$29>>2]|0; //@line 639 "load_mid.cpp"
           $483 = ($tr$01$i94|0)==(0|0); //@line 639 "load_mid.cpp"
           if (!($483)) {
            $hits$02$i = 0;$tr$03$i = $tr$01$i94;
            while(1) {
             $484 = ((($tr$03$i)) + 24|0); //@line 640 "load_mid.cpp"
             $485 = HEAP8[$484>>0]|0; //@line 640 "load_mid.cpp"
             $486 = $485&255; //@line 640 "load_mid.cpp"
             $487 = ($486|0)==($202|0); //@line 640 "load_mid.cpp"
             if ($487) {
              $488 = (($hits$02$i) + 1)|0; //@line 641 "load_mid.cpp"
              $489 = ((($tr$03$i)) + 16|0); //@line 642 "load_mid.cpp"
              HEAP32[$489>>2] = $480; //@line 642 "load_mid.cpp"
              $hits$1$i = $488;
             } else {
              $hits$1$i = $hits$02$i;
             }
             $tr$0$i96 = HEAP32[$tr$03$i>>2]|0; //@line 639 "load_mid.cpp"
             $490 = ($tr$0$i96|0)==(0|0); //@line 639 "load_mid.cpp"
             if ($490) {
              $hits$1$i$lcssa = $hits$1$i;
              break;
             } else {
              $hits$02$i = $hits$1$i;$tr$03$i = $tr$0$i96;
             }
            }
            $491 = ($hits$1$i$lcssa|0)==(0); //@line 645 "load_mid.cpp"
            if (!($491)) {
             $panhigh$3 = $panhigh$2;$panlow$3 = $$panlow$1;
             break L126;
            }
           }
           $492 = (__ZL16mid_locate_trackP10_MIDHANDLEii($27,$202,255)|0); //@line 646 "load_mid.cpp"
           $493 = ((($492)) + 16|0); //@line 647 "load_mid.cpp"
           HEAP32[$493>>2] = $480; //@line 647 "load_mid.cpp"
           $panhigh$3 = $panhigh$2;$panlow$3 = $$panlow$1;
           break;
          }
          case 123:  {
           $494 = ($355<<24>>24)==(0); //@line 1598 "load_mid.cpp"
           if (!($494)) {
            $panhigh$3 = $panhigh$1246;$panlow$3 = $panlow$1247;
            break L126;
           }
           $495 = HEAP32[$31>>2]|0; //@line 684 "load_mid.cpp"
           $496 = ($495|0)==(0|0); //@line 684 "load_mid.cpp"
           if (!($496)) {
            $497 = (($202) + 1)|0; //@line 684 "load_mid.cpp"
            HEAP32[$vararg_buffer43>>2] = $200; //@line 684 "load_mid.cpp"
            $vararg_ptr46 = ((($vararg_buffer43)) + 4|0); //@line 684 "load_mid.cpp"
            HEAP32[$vararg_ptr46>>2] = $497; //@line 684 "load_mid.cpp"
            (_printf(528369,$vararg_buffer43)|0); //@line 684 "load_mid.cpp"
           }
           $tr$02$i98 = HEAP32[$29>>2]|0; //@line 685 "load_mid.cpp"
           $498 = ($tr$02$i98|0)==(0|0); //@line 685 "load_mid.cpp"
           if ($498) {
            $panhigh$3 = $panhigh$1246;$panlow$3 = $panlow$1247;
            break L126;
           } else {
            $tr$03$i100 = $tr$02$i98;
           }
           while(1) {
            $499 = ((($tr$03$i100)) + 24|0); //@line 686 "load_mid.cpp"
            $500 = HEAP8[$499>>0]|0; //@line 686 "load_mid.cpp"
            $501 = $500&255; //@line 686 "load_mid.cpp"
            $502 = ($501|0)==($202|0); //@line 686 "load_mid.cpp"
            do {
             if ($502) {
              $503 = HEAP32[$83>>2]|0; //@line 687 "load_mid.cpp"
              $504 = ((($tr$03$i100)) + 12|0); //@line 394 "load_mid.cpp"
              $505 = HEAP32[$504>>2]|0; //@line 394 "load_mid.cpp"
              $506 = ($505|0)==(0|0); //@line 395 "load_mid.cpp"
              L203: do {
               if (!($506)) {
                $507 = ((($505)) + 4|0); //@line 395 "load_mid.cpp"
                $508 = HEAP32[$507>>2]|0; //@line 395 "load_mid.cpp"
                $509 = ($508>>>0)>($503>>>0); //@line 395 "load_mid.cpp"
                if ($509) {
                 $510 = ((($tr$03$i100)) + 4|0); //@line 395 "load_mid.cpp"
                 $511 = HEAP32[$510>>2]|0; //@line 395 "load_mid.cpp"
                 $512 = ($511|0)==(0|0); //@line 396 "load_mid.cpp"
                 if ($512) {
                  break;
                 } else {
                  $e$0$ph4$i$i102 = $511;
                 }
                } else {
                 $e$0$ph4$i$i102 = $505;
                }
                $513 = ((($tr$03$i100)) + 27|0); //@line 388 "load_mid.cpp"
                $514 = ((($tr$03$i100)) + 25|0); //@line 384 "load_mid.cpp"
                $515 = ((($tr$03$i100)) + 26|0); //@line 385 "load_mid.cpp"
                $516 = ((($tr$03$i100)) + 20|0); //@line 386 "load_mid.cpp"
                $e$01$i$i104 = $e$0$ph4$i$i102;
                while(1) {
                 $517 = ((($e$01$i$i104)) + 4|0); //@line 396 "load_mid.cpp"
                 $518 = HEAP32[$517>>2]|0; //@line 396 "load_mid.cpp"
                 $519 = ($518>>>0)>($503>>>0); //@line 396 "load_mid.cpp"
                 if ($519) {
                  break L203;
                 }
                 HEAP32[$504>>2] = $e$01$i$i104; //@line 397 "load_mid.cpp"
                 $520 = ((($e$01$i$i104)) + 8|0); //@line 382 "load_mid.cpp"
                 $521 = HEAP8[$520>>0]|0; //@line 382 "load_mid.cpp"
                 $522 = ($521<<24>>24)==(0); //@line 382 "load_mid.cpp"
                 if (!($522)) {
                  $523 = ((($e$01$i$i104)) + 10|0); //@line 383 "load_mid.cpp"
                  $524 = HEAP8[$523>>0]|0; //@line 383 "load_mid.cpp"
                  $525 = ($524<<24>>24)==(0); //@line 383 "load_mid.cpp"
                  if ($525) {
                   $$sink$i$i$i105 = -1;
                  } else {
                   $526 = ((($e$01$i$i104)) + 9|0); //@line 383 "load_mid.cpp"
                   $527 = HEAP8[$526>>0]|0; //@line 383 "load_mid.cpp"
                   $$sink$i$i$i105 = $527;
                  }
                  HEAP8[$514>>0] = $$sink$i$i$i105; //@line 384 "load_mid.cpp"
                  $528 = HEAP8[$523>>0]|0; //@line 385 "load_mid.cpp"
                  HEAP8[$515>>0] = $528; //@line 385 "load_mid.cpp"
                  $529 = HEAP32[$517>>2]|0; //@line 386 "load_mid.cpp"
                  HEAP32[$516>>2] = $529; //@line 386 "load_mid.cpp"
                 }
                 $530 = ((($e$01$i$i104)) + 12|0); //@line 388 "load_mid.cpp"
                 $531 = HEAP8[$530>>0]|0; //@line 388 "load_mid.cpp"
                 $532 = ($531<<24>>24)==(8); //@line 388 "load_mid.cpp"
                 if ($532) {
                  $533 = ((($e$01$i$i104)) + 13|0); //@line 388 "load_mid.cpp"
                  $534 = HEAP8[$533>>0]|0; //@line 388 "load_mid.cpp"
                  HEAP8[$513>>0] = $534; //@line 388 "load_mid.cpp"
                 }
                 $535 = HEAP32[$e$01$i$i104>>2]|0; //@line 396 "load_mid.cpp"
                 $536 = ($535|0)==(0|0); //@line 396 "load_mid.cpp"
                 if ($536) {
                  break;
                 } else {
                  $e$01$i$i104 = $535;
                 }
                }
               }
              } while(0);
              $537 = ((($tr$03$i100)) + 25|0); //@line 688 "load_mid.cpp"
              $538 = HEAP8[$537>>0]|0; //@line 688 "load_mid.cpp"
              $539 = ($538<<24>>24)==(-1); //@line 688 "load_mid.cpp"
              if ($539) {
               break;
              }
              $540 = (_calloc(1,16)|0); //@line 298 "load_mid.cpp"
              HEAP32[$540>>2] = 0; //@line 299 "load_mid.cpp"
              $541 = HEAP32[$83>>2]|0; //@line 300 "load_mid.cpp"
              $542 = ((($540)) + 4|0); //@line 300 "load_mid.cpp"
              HEAP32[$542>>2] = $541; //@line 300 "load_mid.cpp"
              $543 = ((($540)) + 8|0); //@line 301 "load_mid.cpp"
              ;HEAP8[$543>>0]=0|0;HEAP8[$543+1>>0]=0|0;HEAP8[$543+2>>0]=0|0;HEAP8[$543+3>>0]=0|0;HEAP8[$543+4>>0]=0|0;HEAP8[$543+5>>0]=0|0; //@line 302 "load_mid.cpp"
              HEAP8[$543>>0] = 1; //@line 554 "load_mid.cpp"
              $544 = HEAP8[$537>>0]|0; //@line 555 "load_mid.cpp"
              $545 = ((($540)) + 9|0); //@line 555 "load_mid.cpp"
              HEAP8[$545>>0] = $544; //@line 555 "load_mid.cpp"
              $546 = ((($tr$03$i100)) + 27|0); //@line 556 "load_mid.cpp"
              $547 = HEAP8[$546>>0]|0; //@line 556 "load_mid.cpp"
              $548 = ((($540)) + 11|0); //@line 556 "load_mid.cpp"
              HEAP8[$548>>0] = $547; //@line 556 "load_mid.cpp"
              __ZL13mid_add_eventP10_MIDHANDLEP9_MIDTRACKP9_MIDEVENT($tr$03$i100,$540); //@line 557 "load_mid.cpp"
             }
            } while(0);
            $tr$0$i108 = HEAP32[$tr$03$i100>>2]|0; //@line 685 "load_mid.cpp"
            $549 = ($tr$0$i108|0)==(0|0); //@line 685 "load_mid.cpp"
            if ($549) {
             $panhigh$3 = $panhigh$1246;$panlow$3 = $panlow$1247;
             break;
            } else {
             $tr$03$i100 = $tr$0$i108;
            }
           }
           break;
          }
          default: {
           $panhigh$3 = $panhigh$1246;$panlow$3 = $panlow$1247;
          }
          }
         } while(0);
         $550 = HEAP32[$31>>2]|0; //@line 1604 "load_mid.cpp"
         $551 = ($550|0)==(0|0); //@line 1604 "load_mid.cpp"
         if ($551) {
          $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $357;$panhigh$4 = $panhigh$3;$panlow$4 = $panlow$3;
          label = 248;
          break L58;
         }
         $552 = HEAP32[$83>>2]|0; //@line 1606 "load_mid.cpp"
         $553 = (($202) + 1)|0; //@line 1606 "load_mid.cpp"
         $554 = $355&255; //@line 1606 "load_mid.cpp"
         HEAP32[$vararg_buffer47>>2] = $t$0; //@line 1605 "load_mid.cpp"
         $vararg_ptr50 = ((($vararg_buffer47)) + 4|0); //@line 1605 "load_mid.cpp"
         HEAP32[$vararg_ptr50>>2] = $552; //@line 1605 "load_mid.cpp"
         $vararg_ptr51 = ((($vararg_buffer47)) + 8|0); //@line 1605 "load_mid.cpp"
         HEAP32[$vararg_ptr51>>2] = $553; //@line 1605 "load_mid.cpp"
         $vararg_ptr52 = ((($vararg_buffer47)) + 12|0); //@line 1605 "load_mid.cpp"
         HEAP32[$vararg_ptr52>>2] = $358; //@line 1605 "load_mid.cpp"
         $vararg_ptr53 = ((($vararg_buffer47)) + 16|0); //@line 1605 "load_mid.cpp"
         HEAP32[$vararg_ptr53>>2] = $554; //@line 1605 "load_mid.cpp"
         (_printf(528391,$vararg_buffer47)|0); //@line 1605 "load_mid.cpp"
         $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $357;$panhigh$4 = $panhigh$3;$panlow$4 = $panlow$3;
         label = 248;
         break;
        }
        case 192:  {
         $555 = $210&255; //@line 1609 "load_mid.cpp"
         $tr$01$i110 = HEAP32[$29>>2]|0; //@line 659 "load_mid.cpp"
         $556 = ($tr$01$i110|0)==(0|0); //@line 659 "load_mid.cpp"
         $$pre341 = (($555) + 1)|0; //@line 674 "load_mid.cpp"
         if ($556) {
          label = 162;
         } else {
          $hits$02$i113 = 0;$tr$03$i112 = $tr$01$i110;
          while(1) {
           $557 = ((($tr$03$i112)) + 24|0); //@line 660 "load_mid.cpp"
           $558 = HEAP8[$557>>0]|0; //@line 660 "load_mid.cpp"
           $559 = $558&255; //@line 660 "load_mid.cpp"
           $560 = ($559|0)==($202|0); //@line 660 "load_mid.cpp"
           if ($560) {
            $561 = (($hits$02$i113) + 1)|0; //@line 661 "load_mid.cpp"
            $562 = (_calloc(1,16)|0); //@line 298 "load_mid.cpp"
            HEAP32[$562>>2] = 0; //@line 299 "load_mid.cpp"
            $563 = HEAP32[$83>>2]|0; //@line 300 "load_mid.cpp"
            $564 = ((($562)) + 4|0); //@line 300 "load_mid.cpp"
            HEAP32[$564>>2] = $563; //@line 300 "load_mid.cpp"
            $565 = ((($562)) + 8|0); //@line 301 "load_mid.cpp"
            $566 = ((($562)) + 12|0); //@line 664 "load_mid.cpp"
            ;HEAP8[$565>>0]=0|0;HEAP8[$565+1>>0]=0|0;HEAP8[$565+2>>0]=0|0;HEAP8[$565+3>>0]=0|0;HEAP8[$565+4>>0]=0|0;HEAP8[$565+5>>0]=0|0; //@line 663 "load_mid.cpp"
            HEAP8[$566>>0] = 8; //@line 664 "load_mid.cpp"
            $567 = (_pat_gmtosmp($$pre341)|0); //@line 665 "load_mid.cpp"
            $568 = $567&255; //@line 665 "load_mid.cpp"
            $569 = ((($562)) + 13|0); //@line 665 "load_mid.cpp"
            HEAP8[$569>>0] = $568; //@line 665 "load_mid.cpp"
            __ZL13mid_add_eventP10_MIDHANDLEP9_MIDTRACKP9_MIDEVENT($tr$03$i112,$562); //@line 666 "load_mid.cpp"
            $hits$1$i114 = $561;
           } else {
            $hits$1$i114 = $hits$02$i113;
           }
           $tr$0$i115 = HEAP32[$tr$03$i112>>2]|0; //@line 659 "load_mid.cpp"
           $570 = ($tr$0$i115|0)==(0|0); //@line 659 "load_mid.cpp"
           if ($570) {
            $hits$1$i114$lcssa = $hits$1$i114;
            break;
           } else {
            $hits$02$i113 = $hits$1$i114;$tr$03$i112 = $tr$0$i115;
           }
          }
          $571 = ($hits$1$i114$lcssa|0)==(0); //@line 669 "load_mid.cpp"
          if ($571) {
           label = 162;
          }
         }
         if ((label|0) == 162) {
          label = 0;
          $572 = (__ZL16mid_locate_trackP10_MIDHANDLEii($27,$202,255)|0); //@line 670 "load_mid.cpp"
          $573 = (_calloc(1,16)|0); //@line 298 "load_mid.cpp"
          HEAP32[$573>>2] = 0; //@line 299 "load_mid.cpp"
          $574 = HEAP32[$83>>2]|0; //@line 300 "load_mid.cpp"
          $575 = ((($573)) + 4|0); //@line 300 "load_mid.cpp"
          HEAP32[$575>>2] = $574; //@line 300 "load_mid.cpp"
          $576 = ((($573)) + 8|0); //@line 301 "load_mid.cpp"
          $577 = ((($573)) + 12|0); //@line 673 "load_mid.cpp"
          HEAP8[$576>>0]=0&255;HEAP8[$576+1>>0]=(0>>8)&255;HEAP8[$576+2>>0]=(0>>16)&255;HEAP8[$576+3>>0]=0>>24; //@line 672 "load_mid.cpp"
          HEAP8[$577>>0] = 8; //@line 673 "load_mid.cpp"
          $578 = (_pat_gmtosmp($$pre341)|0); //@line 674 "load_mid.cpp"
          $579 = $578&255; //@line 674 "load_mid.cpp"
          $580 = ((($573)) + 13|0); //@line 674 "load_mid.cpp"
          HEAP8[$580>>0] = $579; //@line 674 "load_mid.cpp"
          __ZL13mid_add_eventP10_MIDHANDLEP9_MIDTRACKP9_MIDEVENT($572,$573); //@line 675 "load_mid.cpp"
         }
         $581 = HEAP32[$31>>2]|0; //@line 1610 "load_mid.cpp"
         $582 = ($581|0)==(0|0); //@line 1610 "load_mid.cpp"
         if ($582) {
          $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $miditracklen$1;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
          label = 248;
          break L58;
         }
         $583 = HEAP32[$83>>2]|0; //@line 1612 "load_mid.cpp"
         $584 = (($202) + 1)|0; //@line 1612 "load_mid.cpp"
         HEAP32[$vararg_buffer54>>2] = $t$0; //@line 1611 "load_mid.cpp"
         $vararg_ptr57 = ((($vararg_buffer54)) + 4|0); //@line 1611 "load_mid.cpp"
         HEAP32[$vararg_ptr57>>2] = $583; //@line 1611 "load_mid.cpp"
         $vararg_ptr58 = ((($vararg_buffer54)) + 8|0); //@line 1611 "load_mid.cpp"
         HEAP32[$vararg_ptr58>>2] = $584; //@line 1611 "load_mid.cpp"
         $vararg_ptr59 = ((($vararg_buffer54)) + 12|0); //@line 1611 "load_mid.cpp"
         HEAP32[$vararg_ptr59>>2] = $555; //@line 1611 "load_mid.cpp"
         (_printf(528438,$vararg_buffer54)|0); //@line 1611 "load_mid.cpp"
         $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $miditracklen$1;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
         label = 248;
         break;
        }
        case 208:  {
         $585 = HEAP32[$31>>2]|0; //@line 1615 "load_mid.cpp"
         $586 = ($585|0)==(0|0); //@line 1615 "load_mid.cpp"
         if ($586) {
          $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $miditracklen$1;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
          label = 248;
          break L58;
         }
         $587 = (($202) + 1)|0; //@line 1616 "load_mid.cpp"
         $588 = $210&255; //@line 1616 "load_mid.cpp"
         HEAP32[$vararg_buffer60>>2] = $t$0; //@line 1616 "load_mid.cpp"
         $vararg_ptr63 = ((($vararg_buffer60)) + 4|0); //@line 1616 "load_mid.cpp"
         HEAP32[$vararg_ptr63>>2] = $200; //@line 1616 "load_mid.cpp"
         $vararg_ptr64 = ((($vararg_buffer60)) + 8|0); //@line 1616 "load_mid.cpp"
         HEAP32[$vararg_ptr64>>2] = $587; //@line 1616 "load_mid.cpp"
         $vararg_ptr65 = ((($vararg_buffer60)) + 12|0); //@line 1616 "load_mid.cpp"
         HEAP32[$vararg_ptr65>>2] = $588; //@line 1616 "load_mid.cpp"
         (_printf(528474,$vararg_buffer60)|0); //@line 1616 "load_mid.cpp"
         $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $miditracklen$1;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
         label = 248;
         break;
        }
        case 224:  {
         $589 = (($174) + ($205)|0); //@line 159 "load_mid.cpp"
         $590 = HEAP8[$589>>0]|0; //@line 159 "load_mid.cpp"
         $591 = (($205) + 1)|0; //@line 160 "load_mid.cpp"
         HEAP32[$$phi$trans$insert$i>>2] = $591; //@line 160 "load_mid.cpp"
         $592 = (($miditracklen$1) + -1)|0; //@line 1620 "load_mid.cpp"
         $593 = HEAP32[$31>>2]|0; //@line 1621 "load_mid.cpp"
         $594 = ($593|0)==(0|0); //@line 1621 "load_mid.cpp"
         if ($594) {
          $$pre343 = $210&255; //@line 1385 "load_mid.cpp"
          $$pre345 = $$pre343 & 127; //@line 1385 "load_mid.cpp"
          $$pre347 = $590&255; //@line 1385 "load_mid.cpp"
          $$pre349 = $$pre347 << 7; //@line 1385 "load_mid.cpp"
          $$pre351 = $$pre349 & 16256; //@line 1385 "load_mid.cpp"
          $$pre353 = $$pre351 | $$pre345; //@line 1385 "load_mid.cpp"
          $$pre355 = (($$pre353) + -8192)|0; //@line 1391 "load_mid.cpp"
          $$pre$phi354Z2D = $$pre353;$$pre$phi356Z2D = $$pre355;
         } else {
          $595 = (($202) + 1)|0; //@line 1623 "load_mid.cpp"
          $596 = $210&255; //@line 1385 "load_mid.cpp"
          $597 = $596 & 127; //@line 1385 "load_mid.cpp"
          $598 = $590&255; //@line 1385 "load_mid.cpp"
          $599 = $598 << 7; //@line 1385 "load_mid.cpp"
          $600 = $599 & 16256; //@line 1385 "load_mid.cpp"
          $601 = $600 | $597; //@line 1385 "load_mid.cpp"
          $602 = (($601) + -8192)|0; //@line 1391 "load_mid.cpp"
          HEAP32[$vararg_buffer66>>2] = $t$0; //@line 1622 "load_mid.cpp"
          $vararg_ptr69 = ((($vararg_buffer66)) + 4|0); //@line 1622 "load_mid.cpp"
          HEAP32[$vararg_ptr69>>2] = $200; //@line 1622 "load_mid.cpp"
          $vararg_ptr70 = ((($vararg_buffer66)) + 8|0); //@line 1622 "load_mid.cpp"
          HEAP32[$vararg_ptr70>>2] = $595; //@line 1622 "load_mid.cpp"
          $vararg_ptr71 = ((($vararg_buffer66)) + 12|0); //@line 1622 "load_mid.cpp"
          HEAP32[$vararg_ptr71>>2] = $602; //@line 1622 "load_mid.cpp"
          (_printf(528516,$vararg_buffer66)|0); //@line 1622 "load_mid.cpp"
          $$pre$phi354Z2D = $601;$$pre$phi356Z2D = $602;
         }
         $tr$05$i = HEAP32[$29>>2]|0; //@line 734 "load_mid.cpp"
         $603 = ($tr$05$i|0)==(0|0); //@line 734 "load_mid.cpp"
         if (!($603)) {
          $604 = ($$pre$phi354Z2D>>>0)<(8192); //@line 718 "load_mid.cpp"
          $605 = (8192 - ($$pre$phi354Z2D))|0; //@line 720 "load_mid.cpp"
          $606 = $605 >>> 10; //@line 709 "load_mid.cpp"
          $607 = (($606) + 1)|0; //@line 710 "load_mid.cpp"
          $608 = $607&255; //@line 710 "load_mid.cpp"
          $609 = ($$pre$phi356Z2D|0)==(0); //@line 708 "load_mid.cpp"
          $610 = $$pre$phi356Z2D >>> 10; //@line 709 "load_mid.cpp"
          $611 = (($610) + 1)|0; //@line 710 "load_mid.cpp"
          $612 = $611&255; //@line 710 "load_mid.cpp"
          $$$i126 = $609 ? 0 : $612; //@line 708 "load_mid.cpp"
          $$401 = $604 ? 1 : 2; //@line 718 "load_mid.cpp"
          $$$$i126 = $604 ? $608 : $$$i126; //@line 718 "load_mid.cpp"
          $hits$06$i = 0;$tr$07$i = $tr$05$i;
          while(1) {
           $613 = ((($tr$07$i)) + 24|0); //@line 735 "load_mid.cpp"
           $614 = HEAP8[$613>>0]|0; //@line 735 "load_mid.cpp"
           $615 = $614&255; //@line 735 "load_mid.cpp"
           $616 = ($615|0)==($202|0); //@line 735 "load_mid.cpp"
           do {
            if ($616) {
             $617 = (($hits$06$i) + 1)|0; //@line 736 "load_mid.cpp"
             $618 = HEAP32[$83>>2]|0; //@line 737 "load_mid.cpp"
             $619 = ((($tr$07$i)) + 12|0); //@line 394 "load_mid.cpp"
             $620 = HEAP32[$619>>2]|0; //@line 394 "load_mid.cpp"
             $621 = ($620|0)==(0|0); //@line 395 "load_mid.cpp"
             L250: do {
              if (!($621)) {
               $622 = ((($620)) + 4|0); //@line 395 "load_mid.cpp"
               $623 = HEAP32[$622>>2]|0; //@line 395 "load_mid.cpp"
               $624 = ($623>>>0)>($618>>>0); //@line 395 "load_mid.cpp"
               if ($624) {
                $625 = ((($tr$07$i)) + 4|0); //@line 395 "load_mid.cpp"
                $626 = HEAP32[$625>>2]|0; //@line 395 "load_mid.cpp"
                $627 = ($626|0)==(0|0); //@line 396 "load_mid.cpp"
                if ($627) {
                 break;
                } else {
                 $e$0$ph4$i$i120 = $626;
                }
               } else {
                $e$0$ph4$i$i120 = $620;
               }
               $628 = ((($tr$07$i)) + 27|0); //@line 388 "load_mid.cpp"
               $629 = ((($tr$07$i)) + 25|0); //@line 384 "load_mid.cpp"
               $630 = ((($tr$07$i)) + 26|0); //@line 385 "load_mid.cpp"
               $631 = ((($tr$07$i)) + 20|0); //@line 386 "load_mid.cpp"
               $e$01$i$i122 = $e$0$ph4$i$i120;
               while(1) {
                $632 = ((($e$01$i$i122)) + 4|0); //@line 396 "load_mid.cpp"
                $633 = HEAP32[$632>>2]|0; //@line 396 "load_mid.cpp"
                $634 = ($633>>>0)>($618>>>0); //@line 396 "load_mid.cpp"
                if ($634) {
                 break L250;
                }
                HEAP32[$619>>2] = $e$01$i$i122; //@line 397 "load_mid.cpp"
                $635 = ((($e$01$i$i122)) + 8|0); //@line 382 "load_mid.cpp"
                $636 = HEAP8[$635>>0]|0; //@line 382 "load_mid.cpp"
                $637 = ($636<<24>>24)==(0); //@line 382 "load_mid.cpp"
                if (!($637)) {
                 $638 = ((($e$01$i$i122)) + 10|0); //@line 383 "load_mid.cpp"
                 $639 = HEAP8[$638>>0]|0; //@line 383 "load_mid.cpp"
                 $640 = ($639<<24>>24)==(0); //@line 383 "load_mid.cpp"
                 if ($640) {
                  $$sink$i$i$i123 = -1;
                 } else {
                  $641 = ((($e$01$i$i122)) + 9|0); //@line 383 "load_mid.cpp"
                  $642 = HEAP8[$641>>0]|0; //@line 383 "load_mid.cpp"
                  $$sink$i$i$i123 = $642;
                 }
                 HEAP8[$629>>0] = $$sink$i$i$i123; //@line 384 "load_mid.cpp"
                 $643 = HEAP8[$638>>0]|0; //@line 385 "load_mid.cpp"
                 HEAP8[$630>>0] = $643; //@line 385 "load_mid.cpp"
                 $644 = HEAP32[$632>>2]|0; //@line 386 "load_mid.cpp"
                 HEAP32[$631>>2] = $644; //@line 386 "load_mid.cpp"
                }
                $645 = ((($e$01$i$i122)) + 12|0); //@line 388 "load_mid.cpp"
                $646 = HEAP8[$645>>0]|0; //@line 388 "load_mid.cpp"
                $647 = ($646<<24>>24)==(8); //@line 388 "load_mid.cpp"
                if ($647) {
                 $648 = ((($e$01$i$i122)) + 13|0); //@line 388 "load_mid.cpp"
                 $649 = HEAP8[$648>>0]|0; //@line 388 "load_mid.cpp"
                 HEAP8[$628>>0] = $649; //@line 388 "load_mid.cpp"
                }
                $650 = HEAP32[$e$01$i$i122>>2]|0; //@line 396 "load_mid.cpp"
                $651 = ($650|0)==(0|0); //@line 396 "load_mid.cpp"
                if ($651) {
                 break;
                } else {
                 $e$01$i$i122 = $650;
                }
               }
              }
             } while(0);
             $652 = ((($tr$07$i)) + 25|0); //@line 738 "load_mid.cpp"
             $653 = HEAP8[$652>>0]|0; //@line 738 "load_mid.cpp"
             $654 = ($653<<24>>24)==(-1); //@line 738 "load_mid.cpp"
             if ($654) {
              $hits$1$i127 = $617;
              break;
             }
             $655 = (_calloc(1,16)|0); //@line 298 "load_mid.cpp"
             HEAP32[$655>>2] = 0; //@line 299 "load_mid.cpp"
             $656 = HEAP32[$83>>2]|0; //@line 300 "load_mid.cpp"
             $657 = ((($655)) + 4|0); //@line 300 "load_mid.cpp"
             HEAP32[$657>>2] = $656; //@line 300 "load_mid.cpp"
             $658 = ((($655)) + 8|0); //@line 301 "load_mid.cpp"
             $659 = ((($655)) + 12|0); //@line 719 "load_mid.cpp"
             ;HEAP8[$658>>0]=0|0;HEAP8[$658+1>>0]=0|0;HEAP8[$658+2>>0]=0|0;HEAP8[$658+3>>0]=0|0;HEAP8[$658+4>>0]=0|0;HEAP8[$658+5>>0]=0|0; //@line 717 "load_mid.cpp"
             HEAP8[$659>>0] = $$401; //@line 719 "load_mid.cpp"
             $660 = ((($655)) + 13|0); //@line 724 "load_mid.cpp"
             HEAP8[$660>>0] = $$$$i126; //@line 724 "load_mid.cpp"
             __ZL13mid_add_eventP10_MIDHANDLEP9_MIDTRACKP9_MIDEVENT($tr$07$i,$655); //@line 726 "load_mid.cpp"
             $hits$1$i127 = $617;
            } else {
             $hits$1$i127 = $hits$06$i;
            }
           } while(0);
           $tr$0$i128 = HEAP32[$tr$07$i>>2]|0; //@line 734 "load_mid.cpp"
           $661 = ($tr$0$i128|0)==(0|0); //@line 734 "load_mid.cpp"
           if ($661) {
            $hits$1$i127$lcssa = $hits$1$i127;
            break;
           } else {
            $hits$06$i = $hits$1$i127;$tr$07$i = $tr$0$i128;
           }
          }
          $662 = ($hits$1$i127$lcssa|0)==(0); //@line 742 "load_mid.cpp"
          if (!($662)) {
           $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $592;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
           label = 248;
           break L58;
          }
         }
         $663 = (__ZL16mid_locate_trackP10_MIDHANDLEii($27,$202,255)|0); //@line 743 "load_mid.cpp"
         $664 = (_calloc(1,16)|0); //@line 298 "load_mid.cpp"
         HEAP32[$664>>2] = 0; //@line 299 "load_mid.cpp"
         $665 = HEAP32[$83>>2]|0; //@line 300 "load_mid.cpp"
         $666 = ((($664)) + 4|0); //@line 300 "load_mid.cpp"
         HEAP32[$666>>2] = $665; //@line 300 "load_mid.cpp"
         $667 = ((($664)) + 8|0); //@line 301 "load_mid.cpp"
         $668 = ($$pre$phi354Z2D>>>0)<(8192); //@line 718 "load_mid.cpp"
         $669 = ((($664)) + 12|0); //@line 719 "load_mid.cpp"
         ;HEAP8[$667>>0]=0|0;HEAP8[$667+1>>0]=0|0;HEAP8[$667+2>>0]=0|0;HEAP8[$667+3>>0]=0|0;HEAP8[$667+4>>0]=0|0;HEAP8[$667+5>>0]=0|0; //@line 717 "load_mid.cpp"
         if ($668) {
          HEAP8[$669>>0] = 1; //@line 719 "load_mid.cpp"
          $670 = (8192 - ($$pre$phi354Z2D))|0; //@line 720 "load_mid.cpp"
          $671 = $670 >>> 10; //@line 709 "load_mid.cpp"
          $672 = (($671) + 1)|0; //@line 710 "load_mid.cpp"
          $673 = $672&255; //@line 710 "load_mid.cpp"
          $674 = ((($664)) + 13|0); //@line 720 "load_mid.cpp"
          HEAP8[$674>>0] = $673; //@line 720 "load_mid.cpp"
         } else {
          HEAP8[$669>>0] = 2; //@line 723 "load_mid.cpp"
          $675 = ($$pre$phi356Z2D|0)==(0); //@line 708 "load_mid.cpp"
          if ($675) {
           $$0$i1$i$i = 0;
          } else {
           $676 = $$pre$phi356Z2D >>> 10; //@line 709 "load_mid.cpp"
           $677 = (($676) + 1)|0; //@line 710 "load_mid.cpp"
           $678 = $677&255; //@line 710 "load_mid.cpp"
           $$0$i1$i$i = $678;
          }
          $679 = ((($664)) + 13|0); //@line 724 "load_mid.cpp"
          HEAP8[$679>>0] = $$0$i1$i$i; //@line 724 "load_mid.cpp"
         }
         __ZL13mid_add_eventP10_MIDHANDLEP9_MIDTRACKP9_MIDEVENT($663,$664); //@line 726 "load_mid.cpp"
         $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $592;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
         label = 248;
         break;
        }
        case 240:  {
         switch ($201|0) {
         case 240:  {
          $680 = HEAP32[$31>>2]|0; //@line 1629 "load_mid.cpp"
          $681 = ($680|0)==(0|0); //@line 1629 "load_mid.cpp"
          if (!($681)) {
           $682 = $210&255; //@line 1630 "load_mid.cpp"
           HEAP32[$vararg_buffer72>>2] = $t$0; //@line 1629 "load_mid.cpp"
           $vararg_ptr75 = ((($vararg_buffer72)) + 4|0); //@line 1629 "load_mid.cpp"
           HEAP32[$vararg_ptr75>>2] = $200; //@line 1629 "load_mid.cpp"
           $vararg_ptr76 = ((($vararg_buffer72)) + 8|0); //@line 1629 "load_mid.cpp"
           HEAP32[$vararg_ptr76>>2] = $682; //@line 1629 "load_mid.cpp"
           (_printf(528556,$vararg_buffer72)|0); //@line 1629 "load_mid.cpp"
          }
          $683 = ($210<<24>>24)==(-9); //@line 1631 "load_mid.cpp"
          if ($683) {
           $miditracklen$2$lcssa = $miditracklen$1;
          } else {
           $miditracklen$2242 = $miditracklen$1;
           while(1) {
            $$idx15$val = HEAP32[$27>>2]|0;
            $684 = ((($$idx15$val)) + 8|0); //@line 159 "load_mid.cpp"
            $685 = HEAP32[$684>>2]|0; //@line 159 "load_mid.cpp"
            $686 = HEAP32[$$idx15$val>>2]|0; //@line 159 "load_mid.cpp"
            $687 = (($686) + ($685)|0); //@line 159 "load_mid.cpp"
            $688 = HEAP8[$687>>0]|0; //@line 159 "load_mid.cpp"
            $689 = (($685) + 1)|0; //@line 160 "load_mid.cpp"
            HEAP32[$684>>2] = $689; //@line 160 "load_mid.cpp"
            $690 = (($miditracklen$2242) + -1)|0; //@line 1633 "load_mid.cpp"
            $691 = HEAP32[$31>>2]|0; //@line 1634 "load_mid.cpp"
            $692 = ($691|0)==(0|0); //@line 1634 "load_mid.cpp"
            if (!($692)) {
             $693 = $688&255; //@line 1634 "load_mid.cpp"
             HEAP32[$vararg_buffer77>>2] = $693; //@line 1634 "load_mid.cpp"
             (_printf(528580,$vararg_buffer77)|0); //@line 1634 "load_mid.cpp"
            }
            $694 = ($688<<24>>24)==(-9); //@line 1631 "load_mid.cpp"
            if ($694) {
             $miditracklen$2$lcssa = $690;
             break;
            } else {
             $miditracklen$2242 = $690;
            }
           }
          }
          $695 = HEAP32[$31>>2]|0; //@line 1636 "load_mid.cpp"
          $696 = ($695|0)==(0|0); //@line 1636 "load_mid.cpp"
          if ($696) {
           $1323 = -9;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $miditracklen$2$lcssa;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
           label = 248;
           break L58;
          }
          (_putchar(10)|0); //@line 1636 "load_mid.cpp"
          $1323 = -9;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $miditracklen$2$lcssa;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
          label = 248;
          break L58;
          break;
         }
         case 242:  {
          $697 = (($174) + ($205)|0); //@line 159 "load_mid.cpp"
          $698 = HEAP8[$697>>0]|0; //@line 159 "load_mid.cpp"
          $699 = (($205) + 1)|0; //@line 160 "load_mid.cpp"
          HEAP32[$$phi$trans$insert$i>>2] = $699; //@line 160 "load_mid.cpp"
          $700 = (($miditracklen$1) + -1)|0; //@line 1640 "load_mid.cpp"
          $701 = HEAP32[$31>>2]|0; //@line 1641 "load_mid.cpp"
          $702 = ($701|0)==(0|0); //@line 1641 "load_mid.cpp"
          if ($702) {
           $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $700;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
           label = 248;
           break L58;
          }
          $703 = $210&255; //@line 1385 "load_mid.cpp"
          $704 = $703 & 127; //@line 1385 "load_mid.cpp"
          $705 = $698&255; //@line 1385 "load_mid.cpp"
          $706 = $705 << 7; //@line 1385 "load_mid.cpp"
          $707 = $706 & 16256; //@line 1385 "load_mid.cpp"
          $708 = $707 | $704; //@line 1385 "load_mid.cpp"
          $709 = (($708) + -8192)|0; //@line 1391 "load_mid.cpp"
          HEAP32[$vararg_buffer80>>2] = $t$0; //@line 1642 "load_mid.cpp"
          $vararg_ptr83 = ((($vararg_buffer80)) + 4|0); //@line 1642 "load_mid.cpp"
          HEAP32[$vararg_ptr83>>2] = $200; //@line 1642 "load_mid.cpp"
          $vararg_ptr84 = ((($vararg_buffer80)) + 8|0); //@line 1642 "load_mid.cpp"
          HEAP32[$vararg_ptr84>>2] = $709; //@line 1642 "load_mid.cpp"
          (_printf(528586,$vararg_buffer80)|0); //@line 1642 "load_mid.cpp"
          $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $700;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
          label = 248;
          break L58;
          break;
         }
         case 247:  {
          $711 = $205;$d$0$i138 = 0;$i$0$i137 = 0;
          while(1) {
           $710 = (($174) + ($711)|0); //@line 159 "load_mid.cpp"
           $712 = HEAP8[$710>>0]|0; //@line 159 "load_mid.cpp"
           $713 = (($711) + 1)|0; //@line 160 "load_mid.cpp"
           HEAP32[$$phi$trans$insert$i>>2] = $713; //@line 160 "load_mid.cpp"
           $714 = (($i$0$i137) + 1)|0; //@line 774 "load_mid.cpp"
           $715 = $d$0$i138 << 7; //@line 775 "load_mid.cpp"
           $716 = $712&255; //@line 775 "load_mid.cpp"
           $717 = $716 & 127; //@line 775 "load_mid.cpp"
           $718 = $717 | $715; //@line 775 "load_mid.cpp"
           $719 = $716 & 128; //@line 776 "load_mid.cpp"
           $720 = ($719|0)!=(0); //@line 776 "load_mid.cpp"
           $721 = ($714|0)<(4); //@line 772 "load_mid.cpp"
           $or$cond$i139 = $721 & $720; //@line 776 "load_mid.cpp"
           if ($or$cond$i139) {
            $711 = $713;$d$0$i138 = $718;$i$0$i137 = $714;
           } else {
            $$lcssa467 = $714;$$lcssa468 = $718;
            break;
           }
          }
          HEAP32[$171>>2] = $$lcssa468; //@line 779 "load_mid.cpp"
          $722 = (($miditracklen$1) - ($$lcssa467))|0; //@line 1647 "load_mid.cpp"
          $723 = HEAP32[$31>>2]|0; //@line 1649 "load_mid.cpp"
          $724 = ($723|0)==(0|0); //@line 1649 "load_mid.cpp"
          if (!($724)) {
           HEAP32[$vararg_buffer85>>2] = $t$0; //@line 1650 "load_mid.cpp"
           $vararg_ptr88 = ((($vararg_buffer85)) + 4|0); //@line 1650 "load_mid.cpp"
           HEAP32[$vararg_ptr88>>2] = $200; //@line 1650 "load_mid.cpp"
           $vararg_ptr89 = ((($vararg_buffer85)) + 8|0); //@line 1650 "load_mid.cpp"
           HEAP32[$vararg_ptr89>>2] = $$lcssa468; //@line 1650 "load_mid.cpp"
           (_printf(528622,$vararg_buffer85)|0); //@line 1650 "load_mid.cpp"
          }
          $725 = ($$lcssa468|0)>(0); //@line 1652 "load_mid.cpp"
          if ($725) {
           $726 = (($miditracklen$1) + -2)|0; //@line 1652 "load_mid.cpp"
           $727 = (($726) - ($$lcssa467))|0; //@line 1652 "load_mid.cpp"
           $metalen$0238 = $$lcssa468;
           while(1) {
            $$idx13$val = HEAP32[$27>>2]|0;
            $728 = ((($$idx13$val)) + 8|0); //@line 159 "load_mid.cpp"
            $729 = HEAP32[$728>>2]|0; //@line 159 "load_mid.cpp"
            $730 = HEAP32[$$idx13$val>>2]|0; //@line 159 "load_mid.cpp"
            $731 = (($730) + ($729)|0); //@line 159 "load_mid.cpp"
            $732 = HEAP8[$731>>0]|0; //@line 159 "load_mid.cpp"
            $733 = (($729) + 1)|0; //@line 160 "load_mid.cpp"
            HEAP32[$728>>2] = $733; //@line 160 "load_mid.cpp"
            $734 = (($metalen$0238) + -1)|0; //@line 1654 "load_mid.cpp"
            $735 = HEAP32[$31>>2]|0; //@line 1656 "load_mid.cpp"
            $736 = ($735|0)==(0|0); //@line 1656 "load_mid.cpp"
            if (!($736)) {
             $737 = $732&255; //@line 1656 "load_mid.cpp"
             HEAP32[$vararg_buffer90>>2] = $737; //@line 1656 "load_mid.cpp"
             (_printf(528580,$vararg_buffer90)|0); //@line 1656 "load_mid.cpp"
            }
            $738 = ($metalen$0238|0)>(1); //@line 1652 "load_mid.cpp"
            if ($738) {
             $metalen$0238 = $734;
            } else {
             break;
            }
           }
           $739 = (($727) - ($$lcssa468))|0; //@line 1652 "load_mid.cpp"
           $740 = (($739) + 2)|0; //@line 1652 "load_mid.cpp"
           $miditracklen$3$lcssa = $740;
          } else {
           $miditracklen$3$lcssa = $722;
          }
          HEAP32[$171>>2] = $$lcssa463; //@line 1658 "load_mid.cpp"
          $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $miditracklen$3$lcssa;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
          label = 248;
          break L58;
          break;
         }
         case 255:  {
          $742 = $205;$d$0$i148 = 0;$i$0$i147 = 0;
          while(1) {
           $741 = (($174) + ($742)|0); //@line 159 "load_mid.cpp"
           $743 = HEAP8[$741>>0]|0; //@line 159 "load_mid.cpp"
           $744 = (($742) + 1)|0; //@line 160 "load_mid.cpp"
           HEAP32[$$phi$trans$insert$i>>2] = $744; //@line 160 "load_mid.cpp"
           $745 = (($i$0$i147) + 1)|0; //@line 774 "load_mid.cpp"
           $746 = $d$0$i148 << 7; //@line 775 "load_mid.cpp"
           $747 = $743&255; //@line 775 "load_mid.cpp"
           $748 = $747 & 127; //@line 775 "load_mid.cpp"
           $749 = $748 | $746; //@line 775 "load_mid.cpp"
           $750 = $747 & 128; //@line 776 "load_mid.cpp"
           $751 = ($750|0)!=(0); //@line 776 "load_mid.cpp"
           $752 = ($745|0)<(4); //@line 772 "load_mid.cpp"
           $or$cond$i149 = $752 & $751; //@line 776 "load_mid.cpp"
           if ($or$cond$i149) {
            $742 = $744;$d$0$i148 = $749;$i$0$i147 = $745;
           } else {
            $$lcssa464 = $744;$$lcssa465 = $745;$$lcssa466 = $749;
            break;
           }
          }
          HEAP32[$171>>2] = $$lcssa466; //@line 779 "load_mid.cpp"
          $753 = (($miditracklen$1) - ($$lcssa465))|0; //@line 1662 "load_mid.cpp"
          $754 = ($$lcssa466|0)>(31); //@line 1664 "load_mid.cpp"
          $$ = $754 ? 31 : $$lcssa466; //@line 1664 "load_mid.cpp"
          $755 = ($$|0)==(0); //@line 1665 "load_mid.cpp"
          if ($755) {
           $miditracklen$4 = $753;
          } else {
           $756 = (($174) + ($$lcssa464)|0); //@line 172 "load_mid.cpp"
           _memcpy(($buf|0),($756|0),($$|0))|0; //@line 172 "load_mid.cpp"
           $757 = HEAP32[$$phi$trans$insert$i>>2]|0; //@line 173 "load_mid.cpp"
           $758 = (($757) + ($$))|0; //@line 173 "load_mid.cpp"
           HEAP32[$$phi$trans$insert$i>>2] = $758; //@line 173 "load_mid.cpp"
           $759 = (($753) - ($$))|0; //@line 1667 "load_mid.cpp"
           $miditracklen$4 = $759;
          }
          $760 = (($buf) + ($$)|0); //@line 1669 "load_mid.cpp"
          HEAP8[$760>>0] = 0; //@line 1669 "load_mid.cpp"
          if ($754) {
           $$promoted = HEAP32[$$phi$trans$insert$i>>2]|0; //@line 160 "load_mid.cpp"
           $761 = (($miditracklen$4) + -3)|0; //@line 1671 "load_mid.cpp"
           $762 = (($761) - ($$lcssa466))|0; //@line 1671 "load_mid.cpp"
           $763 = (-2 - ($$lcssa466))|0; //@line 1671 "load_mid.cpp"
           $764 = (($763) + 32)|0; //@line 1671 "load_mid.cpp"
           $765 = ($764|0)>(-2);
           $smax318 = $765 ? $764 : -2;
           $766 = (($762) - ($smax318))|0; //@line 1671 "load_mid.cpp"
           $767 = (($$lcssa466) + 3)|0; //@line 1671 "load_mid.cpp"
           $768 = (($767) + ($$promoted))|0; //@line 1671 "load_mid.cpp"
           $769 = (($768) + ($smax318))|0; //@line 1671 "load_mid.cpp"
           $770 = (($766) + 32)|0; //@line 1671 "load_mid.cpp"
           $771 = (($769) + -32)|0; //@line 1671 "load_mid.cpp"
           HEAP32[$$phi$trans$insert$i>>2] = $771; //@line 160 "load_mid.cpp"
           $miditracklen$5$lcssa = $770;
          } else {
           $miditracklen$5$lcssa = $miditracklen$4;
          }
          HEAP32[$171>>2] = $$lcssa463; //@line 1676 "load_mid.cpp"
          $772 = $210&255; //@line 1677 "load_mid.cpp"
          switch ($772|0) {
          case 3:  {
           $773 = HEAP32[$31>>2]|0; //@line 1679 "load_mid.cpp"
           $774 = ($773|0)==(0|0); //@line 1679 "load_mid.cpp"
           if (!($774)) {
            HEAP32[$vararg_buffer93>>2] = $t$0; //@line 1680 "load_mid.cpp"
            $vararg_ptr96 = ((($vararg_buffer93)) + 4|0); //@line 1680 "load_mid.cpp"
            HEAP32[$vararg_ptr96>>2] = $200; //@line 1680 "load_mid.cpp"
            $vararg_ptr97 = ((($vararg_buffer93)) + 8|0); //@line 1680 "load_mid.cpp"
            HEAP32[$vararg_ptr97>>2] = $buf; //@line 1680 "load_mid.cpp"
            (_printf(528653,$vararg_buffer93)|0); //@line 1680 "load_mid.cpp"
           }
           $775 = HEAP8[$112>>0]|0; //@line 1685 "load_mid.cpp"
           $776 = ($775<<24>>24)==(0); //@line 1685 "load_mid.cpp"
           if (!($776)) {
            $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $miditracklen$5$lcssa;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
            label = 248;
            break L58;
           }
           (_strcpy($112,$buf)|0); //@line 1686 "load_mid.cpp"
           $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $miditracklen$5$lcssa;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
           label = 248;
           break L58;
           break;
          }
          case 81:  {
           $777 = HEAP32[$buf>>2]|0; //@line 1691 "load_mid.cpp"
           $778 = $777 << 16; //@line 1691 "load_mid.cpp"
           $779 = $778 & 16711680; //@line 1691 "load_mid.cpp"
           $780 = $777 & 65280;
           $781 = $780 | $779; //@line 1691 "load_mid.cpp"
           $782 = $777 >>> 16;
           $783 = $782 & 255; //@line 1691 "load_mid.cpp"
           $784 = $781 | $783; //@line 1691 "load_mid.cpp"
           $785 = ($784|0)==(0); //@line 1692 "load_mid.cpp"
           if (!($785)) {
            $786 = (60000000 / ($784>>>0))&-1; //@line 1693 "load_mid.cpp"
            HEAP32[$81>>2] = $786; //@line 1693 "load_mid.cpp"
           }
           $787 = HEAP32[$31>>2]|0; //@line 1694 "load_mid.cpp"
           $788 = ($787|0)==(0|0); //@line 1694 "load_mid.cpp"
           if (!($788)) {
            $789 = HEAP32[$81>>2]|0; //@line 1694 "load_mid.cpp"
            HEAP32[$vararg_buffer98>>2] = $t$0; //@line 1694 "load_mid.cpp"
            $vararg_ptr101 = ((($vararg_buffer98)) + 4|0); //@line 1694 "load_mid.cpp"
            HEAP32[$vararg_ptr101>>2] = $200; //@line 1694 "load_mid.cpp"
            $vararg_ptr102 = ((($vararg_buffer98)) + 8|0); //@line 1694 "load_mid.cpp"
            HEAP32[$vararg_ptr102>>2] = $789; //@line 1694 "load_mid.cpp"
            (_printf(528682,$vararg_buffer98)|0); //@line 1694 "load_mid.cpp"
           }
           $790 = HEAP32[$82>>2]|0; //@line 1695 "load_mid.cpp"
           $791 = ($790|0)==(0); //@line 1695 "load_mid.cpp"
           if ($791) {
            $792 = HEAP32[$81>>2]|0; //@line 1695 "load_mid.cpp"
            HEAP32[$82>>2] = $792; //@line 1695 "load_mid.cpp"
            $805 = $792;
           } else {
            $793 = HEAP32[$29>>2]|0; //@line 1697 "load_mid.cpp"
            $794 = ($793|0)==(0|0); //@line 1698 "load_mid.cpp"
            if ($794) {
             (__ZL16mid_locate_trackP10_MIDHANDLEii($27,0,255)|0); //@line 1698 "load_mid.cpp"
             $$pre334 = HEAP32[$29>>2]|0; //@line 538 "load_mid.cpp"
             $803 = $$pre334;
            } else {
             $803 = $793;
            }
            $795 = HEAP32[$81>>2]|0; //@line 1699 "load_mid.cpp"
            $796 = (_calloc(1,16)|0); //@line 298 "load_mid.cpp"
            HEAP32[$796>>2] = 0; //@line 299 "load_mid.cpp"
            $797 = HEAP32[$83>>2]|0; //@line 300 "load_mid.cpp"
            $798 = ((($796)) + 4|0); //@line 300 "load_mid.cpp"
            HEAP32[$798>>2] = $797; //@line 300 "load_mid.cpp"
            $799 = ((($796)) + 8|0); //@line 301 "load_mid.cpp"
            $800 = ((($796)) + 12|0); //@line 536 "load_mid.cpp"
            HEAP8[$799>>0]=0&255;HEAP8[$799+1>>0]=(0>>8)&255;HEAP8[$799+2>>0]=(0>>16)&255;HEAP8[$799+3>>0]=0>>24; //@line 535 "load_mid.cpp"
            HEAP8[$800>>0] = 4; //@line 536 "load_mid.cpp"
            $801 = $795&255; //@line 537 "load_mid.cpp"
            $802 = ((($796)) + 13|0); //@line 537 "load_mid.cpp"
            HEAP8[$802>>0] = $801; //@line 537 "load_mid.cpp"
            __ZL13mid_add_eventP10_MIDHANDLEP9_MIDTRACKP9_MIDEVENT($803,$796); //@line 538 "load_mid.cpp"
            $$pre335 = HEAP32[$81>>2]|0; //@line 1701 "load_mid.cpp"
            $805 = $$pre335;
           }
           $804 = ($805|0)>($maxtempo$1248|0); //@line 1701 "load_mid.cpp"
           $$maxtempo$1 = $804 ? $805 : $maxtempo$1248; //@line 1701 "load_mid.cpp"
           $1323 = $210;$maxtempo$2 = $$maxtempo$1;$miditracklen$6 = $miditracklen$5$lcssa;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
           label = 248;
           break L58;
           break;
          }
          case 47:  {
           $806 = HEAP32[$31>>2]|0; //@line 1704 "load_mid.cpp"
           $807 = ($806|0)==(0|0); //@line 1704 "load_mid.cpp"
           if (!($807)) {
            HEAP32[$vararg_buffer103>>2] = $t$0; //@line 1704 "load_mid.cpp"
            $vararg_ptr106 = ((($vararg_buffer103)) + 4|0); //@line 1704 "load_mid.cpp"
            HEAP32[$vararg_ptr106>>2] = $200; //@line 1704 "load_mid.cpp"
            (_printf(528707,$vararg_buffer103)|0); //@line 1704 "load_mid.cpp"
           }
           $808 = ($miditracklen$5$lcssa|0)>(0); //@line 1705 "load_mid.cpp"
           if (!($808)) {
            $823 = $210;$maxtempo$2205 = $maxtempo$1248;$panhigh$4203 = $panhigh$1246;$panlow$4204 = $panlow$1247;
            label = 249;
            break L58;
           }
           HEAP32[$vararg_buffer107>>2] = $miditracklen$5$lcssa; //@line 1706 "load_mid.cpp"
           (_sprintf($buf,528736,$vararg_buffer107)|0); //@line 1706 "load_mid.cpp"
           $809 = (_strlen($buf)|0); //@line 245 "load_mid.cpp"
           $810 = (($809) + 54)|0; //@line 245 "load_mid.cpp"
           $811 = ($810>>>0)>(255); //@line 245 "load_mid.cpp"
           if (!($811)) {
            HEAP32[$vararg_buffer110>>2] = $buf; //@line 246 "load_mid.cpp"
            (_sprintf($txt$i,528740,$vararg_buffer110)|0); //@line 246 "load_mid.cpp"
            HEAP32[$vararg_buffer113>>2] = $txt$i; //@line 250 "load_mid.cpp"
            (_fprintf($172,528176,$vararg_buffer113)|0); //@line 250 "load_mid.cpp"
           }
           $823 = $210;$maxtempo$2205 = $maxtempo$1248;$panhigh$4203 = $panhigh$1246;$panlow$4204 = $panlow$1247;
           label = 249;
           break L58;
           break;
          }
          default: {
           $812 = HEAP32[$31>>2]|0; //@line 1712 "load_mid.cpp"
           $813 = ($812|0)==(0|0); //@line 1712 "load_mid.cpp"
           if ($813) {
            $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $miditracklen$5$lcssa;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
            label = 248;
            break L58;
           }
           HEAP32[$vararg_buffer116>>2] = $t$0; //@line 1712 "load_mid.cpp"
           $vararg_ptr119 = ((($vararg_buffer116)) + 4|0); //@line 1712 "load_mid.cpp"
           HEAP32[$vararg_ptr119>>2] = $200; //@line 1712 "load_mid.cpp"
           $vararg_ptr120 = ((($vararg_buffer116)) + 8|0); //@line 1712 "load_mid.cpp"
           HEAP32[$vararg_ptr120>>2] = $772; //@line 1712 "load_mid.cpp"
           (_printf(528795,$vararg_buffer116)|0); //@line 1712 "load_mid.cpp"
           $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $miditracklen$5$lcssa;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
           label = 248;
           break L58;
          }
          }
          break;
         }
         default: {
          $814 = HEAP32[$31>>2]|0; //@line 1717 "load_mid.cpp"
          $815 = ($814|0)==(0|0); //@line 1717 "load_mid.cpp"
          if ($815) {
           $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $miditracklen$1;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
           label = 248;
           break L58;
          }
          $816 = $210&255; //@line 1717 "load_mid.cpp"
          HEAP32[$vararg_buffer121>>2] = $t$0; //@line 1717 "load_mid.cpp"
          $vararg_ptr124 = ((($vararg_buffer121)) + 4|0); //@line 1717 "load_mid.cpp"
          HEAP32[$vararg_ptr124>>2] = $200; //@line 1717 "load_mid.cpp"
          $vararg_ptr125 = ((($vararg_buffer121)) + 8|0); //@line 1717 "load_mid.cpp"
          HEAP32[$vararg_ptr125>>2] = $816; //@line 1717 "load_mid.cpp"
          (_printf(528823,$vararg_buffer121)|0); //@line 1717 "load_mid.cpp"
          $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $miditracklen$1;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
          label = 248;
          break L58;
         }
         }
         break;
        }
        default: {
         $817 = HEAP32[$31>>2]|0; //@line 1722 "load_mid.cpp"
         $818 = ($817|0)==(0|0); //@line 1722 "load_mid.cpp"
         if ($818) {
          $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $miditracklen$1;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
          label = 248;
          break L58;
         }
         $819 = $210&255; //@line 1722 "load_mid.cpp"
         HEAP32[$vararg_buffer126>>2] = $t$0; //@line 1722 "load_mid.cpp"
         $vararg_ptr129 = ((($vararg_buffer126)) + 4|0); //@line 1722 "load_mid.cpp"
         HEAP32[$vararg_ptr129>>2] = $200; //@line 1722 "load_mid.cpp"
         $vararg_ptr130 = ((($vararg_buffer126)) + 8|0); //@line 1722 "load_mid.cpp"
         HEAP32[$vararg_ptr130>>2] = $201; //@line 1722 "load_mid.cpp"
         $vararg_ptr131 = ((($vararg_buffer126)) + 12|0); //@line 1722 "load_mid.cpp"
         HEAP32[$vararg_ptr131>>2] = $819; //@line 1722 "load_mid.cpp"
         (_printf(528853,$vararg_buffer126)|0); //@line 1722 "load_mid.cpp"
         $1323 = $210;$maxtempo$2 = $maxtempo$1248;$miditracklen$6 = $miditracklen$1;$panhigh$4 = $panhigh$1246;$panlow$4 = $panlow$1247;
         label = 248;
        }
        }
       } while(0);
       if ((label|0) == 248) {
        label = 0;
        $820 = ($miditracklen$6|0)<(1); //@line 1725 "load_mid.cpp"
        if ($820) {
         $823 = $1323;$maxtempo$2205 = $maxtempo$2;$panhigh$4203 = $panhigh$4;$panlow$4204 = $panlow$4;
         label = 249;
        } else {
         $maxtempo$1$be361 = $maxtempo$2;$miditracklen$0$be362 = $miditracklen$6;$panhigh$1$be359 = $panhigh$4;$panlow$1$be360 = $panlow$4;
        }
       }
       if ((label|0) == 249) {
        label = 0;
        $821 = ($runningstatus$1<<24>>24)==(-1); //@line 1725 "load_mid.cpp"
        $822 = ($823<<24>>24)==(47); //@line 1725 "load_mid.cpp"
        $or$cond6 = $821 & $822; //@line 1725 "load_mid.cpp"
        if ($or$cond6) {
         $maxtempo$1$lcssa = $maxtempo$2205;$panhigh$1$lcssa = $panhigh$4203;$panlow$1$lcssa = $panlow$4204;
         break L47;
        }
        $824 = HEAP32[$27>>2]|0; //@line 1726 "load_mid.cpp"
        $$idx28 = ((($824)) + 8|0);
        $$idx28$val = HEAP32[$$idx28>>2]|0;
        $825 = HEAP32[$824>>2]|0; //@line 172 "load_mid.cpp"
        $826 = (($825) + ($$idx28$val)|0); //@line 172 "load_mid.cpp"
        $827 = HEAPU8[$826>>0]|(HEAPU8[$826+1>>0]<<8)|(HEAPU8[$826+2>>0]<<16)|(HEAPU8[$826+3>>0]<<24); //@line 172 "load_mid.cpp"
        HEAP32[$buf>>2] = $827; //@line 172 "load_mid.cpp"
        $828 = HEAP32[$$idx28>>2]|0; //@line 173 "load_mid.cpp"
        $829 = (($828) + 4)|0; //@line 173 "load_mid.cpp"
        HEAP32[$$idx28>>2] = $829; //@line 173 "load_mid.cpp"
        HEAP8[$132>>0] = 0; //@line 1728 "load_mid.cpp"
        $830 = (_strcmp($buf,528132)|0); //@line 1729 "load_mid.cpp"
        $831 = ($830|0)==(0); //@line 1729 "load_mid.cpp"
        if ($831) {
         $$idx28$val$lcssa = $$idx28$val;$maxtempo$2205$lcssa474 = $maxtempo$2205;$panhigh$4203$lcssa476 = $panhigh$4203;$panlow$4204$lcssa475 = $panlow$4204;
         break;
        }
        HEAP32[$vararg_buffer132>>2] = 528909; //@line 246 "load_mid.cpp"
        (_sprintf($txt$i,528740,$vararg_buffer132)|0); //@line 246 "load_mid.cpp"
        HEAP32[$vararg_buffer135>>2] = $txt$i; //@line 250 "load_mid.cpp"
        (_fprintf($172,528176,$vararg_buffer135)|0); //@line 250 "load_mid.cpp"
        $832 = HEAP32[$27>>2]|0; //@line 1735 "load_mid.cpp"
        $833 = ((($832)) + 8|0); //@line 140 "load_mid.cpp"
        HEAP32[$833>>2] = $$idx28$val; //@line 140 "load_mid.cpp"
        $maxtempo$1$be361 = $maxtempo$2205;$miditracklen$0$be362 = 2147483647;$panhigh$1$be359 = $panhigh$4203;$panlow$1$be360 = $panlow$4204;
       }
       $h$idx$val$pre$i$pre = HEAP32[$27>>2]|0; //@line 1515 "load_mid.cpp"
       $$phi$trans$insert$i$phi$trans$insert = ((($h$idx$val$pre$i$pre)) + 8|0);
       $$pre$i$pre = HEAP32[$$phi$trans$insert$i$phi$trans$insert>>2]|0; //@line 159 "load_mid.cpp"
       $$pre6$i$pre = HEAP32[$h$idx$val$pre$i$pre>>2]|0; //@line 159 "load_mid.cpp"
       $$idx20$val = $h$idx$val$pre$i$pre;$$pre$i = $$pre$i$pre;$174 = $$pre6$i$pre;$maxtempo$1248 = $maxtempo$1$be361;$miditracklen$0249 = $miditracklen$0$be362;$panhigh$1246 = $panhigh$1$be359;$panlow$1247 = $panlow$1$be360;$runningstatus$0250 = $runningstatus$1;
      }
      HEAP32[$vararg_buffer138>>2] = 528921; //@line 246 "load_mid.cpp"
      (_sprintf($txt$i,528740,$vararg_buffer138)|0); //@line 246 "load_mid.cpp"
      HEAP32[$vararg_buffer141>>2] = $txt$i; //@line 250 "load_mid.cpp"
      (_fprintf($172,528176,$vararg_buffer141)|0); //@line 250 "load_mid.cpp"
      $834 = HEAP32[$27>>2]|0; //@line 1735 "load_mid.cpp"
      $835 = ((($834)) + 8|0); //@line 140 "load_mid.cpp"
      HEAP32[$835>>2] = $$idx28$val$lcssa; //@line 140 "load_mid.cpp"
      $maxtempo$1$lcssa = $maxtempo$2205$lcssa474;$panhigh$1$lcssa = $panhigh$4203$lcssa476;$panlow$1$lcssa = $panlow$4204$lcssa475;
     } else {
      $maxtempo$1$lcssa = $maxtempo$0;$panhigh$1$lcssa = $panhigh$0;$panlow$1$lcssa = $panlow$0;
     }
    } while(0);
    $836 = (($t$0) + 1)|0; //@line 1502 "load_mid.cpp"
    $maxtempo$0$ph = $maxtempo$1$lcssa;$panhigh$0$ph = $panhigh$1$lcssa;$panlow$0$ph = $panlow$1$lcssa;$t$0$ph = $836;
    label = 24;
   }
   if ((label|0) == 29) {
    $135 = (_strlen($buf)|0); //@line 245 "load_mid.cpp"
    $136 = (($135) + 38)|0; //@line 245 "load_mid.cpp"
    $137 = ($136>>>0)>(255); //@line 245 "load_mid.cpp"
    if (!($137)) {
     HEAP32[$vararg_buffer10>>2] = $buf; //@line 246 "load_mid.cpp"
     (_sprintf($txt$i,528137,$vararg_buffer10)|0); //@line 246 "load_mid.cpp"
     $138 = HEAP32[37544]|0; //@line 250 "load_mid.cpp"
     HEAP32[$vararg_buffer13>>2] = $txt$i; //@line 250 "load_mid.cpp"
     (_fprintf($138,528176,$vararg_buffer13)|0); //@line 250 "load_mid.cpp"
    }
    HEAP32[300431] = 0; //@line 1508 "load_mid.cpp"
    $$0 = 0;
    STACKTOP = sp;return ($$0|0); //@line 1880 "load_mid.cpp"
   }
   if ($121) {
    (_puts(528924)|0); //@line 1739 "load_mid.cpp"
   }
   $ttp$0272 = HEAP32[$29>>2]|0; //@line 1743 "load_mid.cpp"
   $837 = ($ttp$0272|0)==(0|0); //@line 1743 "load_mid.cpp"
   do {
    if ($837) {
     $delta$0$lcssa369 = 2147483647;
     label = 269;
    } else {
     $delta$0274 = 2147483647;$metalen$2273 = 0;$ttp$0275 = $ttp$0272;
     while(1) {
      $838 = ((($ttp$0275)) + 24|0); //@line 1744 "load_mid.cpp"
      $839 = HEAP8[$838>>0]|0; //@line 1744 "load_mid.cpp"
      $840 = $839&255; //@line 1744 "load_mid.cpp"
      $841 = 1 << $840; //@line 1744 "load_mid.cpp"
      $842 = $841 | $metalen$2273; //@line 1744 "load_mid.cpp"
      $843 = ((($ttp$0275)) + 4|0); //@line 1745 "load_mid.cpp"
      $844 = HEAP32[$843>>2]|0; //@line 1745 "load_mid.cpp"
      $845 = ($844|0)==(0|0); //@line 1745 "load_mid.cpp"
      if ($845) {
       $delta$1 = $delta$0274;
      } else {
       $$05$i = $844;
       while(1) {
        $846 = ((($$05$i)) + 8|0); //@line 1396 "load_mid.cpp"
        $847 = HEAP8[$846>>0]|0; //@line 1396 "load_mid.cpp"
        $848 = ($847<<24>>24)==(0); //@line 1396 "load_mid.cpp"
        if (!($848)) {
         $849 = ((($$05$i)) + 10|0); //@line 1396 "load_mid.cpp"
         $850 = HEAP8[$849>>0]|0; //@line 1396 "load_mid.cpp"
         $851 = ($850<<24>>24)==(0); //@line 1396 "load_mid.cpp"
         if (!($851)) {
          $$05$i$lcssa = $$05$i;
          label = 262;
          break;
         }
        }
        $852 = HEAP32[$$05$i>>2]|0; //@line 1396 "load_mid.cpp"
        $853 = ($852|0)==(0|0); //@line 1396 "load_mid.cpp"
        if ($853) {
         $$01$i = 2147483647;
         break;
        } else {
         $$05$i = $852;
        }
       }
       if ((label|0) == 262) {
        label = 0;
        $854 = ((($$05$i$lcssa)) + 4|0); //@line 1398 "load_mid.cpp"
        $855 = HEAP32[$854>>2]|0; //@line 1398 "load_mid.cpp"
        $$01$i = $855;
       }
       $856 = ($$01$i>>>0)<($delta$0274>>>0); //@line 1748 "load_mid.cpp"
       $$delta$0 = $856 ? $$01$i : $delta$0274; //@line 1748 "load_mid.cpp"
       $delta$1 = $$delta$0;
      }
      $ttp$0 = HEAP32[$ttp$0275>>2]|0; //@line 1743 "load_mid.cpp"
      $857 = ($ttp$0|0)==(0|0); //@line 1743 "load_mid.cpp"
      if ($857) {
       $$lcssa483 = $842;$delta$1$lcssa = $delta$1;
       break;
      } else {
       $delta$0274 = $delta$1;$metalen$2273 = $842;$ttp$0275 = $ttp$0;
      }
     }
     $858 = $$lcssa483 & 1023; //@line 1752 "load_mid.cpp"
     $859 = ($858|0)==(0); //@line 1752 "load_mid.cpp"
     if ($859) {
      $delta$0$lcssa369 = $delta$1$lcssa;
      label = 269;
     } else {
      $860 = $$lcssa483 & 3840; //@line 1753 "load_mid.cpp"
      $861 = ($860|0)==(1024); //@line 1753 "load_mid.cpp"
      if ($861) {
       HEAP32[$30>>2] = 10; //@line 1754 "load_mid.cpp"
       $864 = 11;$delta$0$lcssa368 = $delta$1$lcssa;
       break;
      } else {
       HEAP32[$30>>2] = 9; //@line 1756 "load_mid.cpp"
       $864 = 10;$delta$0$lcssa368 = $delta$1$lcssa;
       break;
      }
     }
    }
   } while(0);
   if ((label|0) == 269) {
    HEAP32[$30>>2] = 15; //@line 1758 "load_mid.cpp"
    $864 = 16;$delta$0$lcssa368 = $delta$0$lcssa369;
   }
   $862 = HEAP32[$36>>2]|0; //@line 1759 "load_mid.cpp"
   $863 = ($862|0)==(0|0); //@line 1759 "load_mid.cpp"
   if ($863) {
    $ttp$1265 = $ttp$0272;
   } else {
    HEAP32[$vararg_buffer144>>2] = $864; //@line 1760 "load_mid.cpp"
    (_printf(528955,$vararg_buffer144)|0); //@line 1760 "load_mid.cpp"
    $ttp$1265$pre = HEAP32[$29>>2]|0; //@line 1765 "load_mid.cpp"
    $ttp$1265 = $ttp$1265$pre;
   }
   HEAP32[$83>>2] = 0; //@line 1763 "load_mid.cpp"
   $865 = HEAP32[$100>>2]|0; //@line 1764 "load_mid.cpp"
   $866 = ($ttp$1265|0)==(0|0); //@line 1765 "load_mid.cpp"
   if ($866) {
    $1067 = 0;$1072 = 0;
   } else {
    $867 = HEAP32[37544]|0; //@line 250 "load_mid.cpp"
    $ttp$1266 = $ttp$1265;
    while(1) {
     $868 = ((($ttp$1266)) + 24|0); //@line 1766 "load_mid.cpp"
     $869 = HEAP8[$868>>0]|0; //@line 1766 "load_mid.cpp"
     $870 = $869&255; //@line 1766 "load_mid.cpp"
     $871 = HEAP32[$30>>2]|0; //@line 1766 "load_mid.cpp"
     $872 = ($870|0)==($871|0); //@line 1766 "load_mid.cpp"
     L398: do {
      if ($872) {
       $873 = ((($ttp$1266)) + 4|0); //@line 1281 "load_mid.cpp"
       $e$02$i158 = HEAP32[$873>>2]|0; //@line 1281 "load_mid.cpp"
       $874 = ($e$02$i158|0)==(0|0); //@line 1281 "load_mid.cpp"
       L400: do {
        if (!($874)) {
         $e$08$i = $e$02$i158;$lno$03$i = 0;$n$04$i = 0;$toff$07$i = 0;$ton$06$i = 0;$v$05$i = 127;
         while(1) {
          $875 = ((($e$08$i)) + 4|0); //@line 1282 "load_mid.cpp"
          $876 = HEAP32[$875>>2]|0; //@line 1282 "load_mid.cpp"
          $877 = ($876>>>0)<($delta$0$lcssa368>>>0); //@line 1282 "load_mid.cpp"
          $878 = (($876) - ($delta$0$lcssa368))|0; //@line 1283 "load_mid.cpp"
          $storemerge$i159 = $877 ? 0 : $878; //@line 1282 "load_mid.cpp"
          HEAP32[$875>>2] = $storemerge$i159; //@line 1282 "load_mid.cpp"
          $879 = ((($e$08$i)) + 8|0); //@line 1284 "load_mid.cpp"
          $880 = HEAP8[$879>>0]|0; //@line 1284 "load_mid.cpp"
          $881 = ($880<<24>>24)==(1); //@line 1284 "load_mid.cpp"
          do {
           if ($881) {
            $882 = ((($e$08$i)) + 10|0); //@line 1285 "load_mid.cpp"
            $883 = HEAP8[$882>>0]|0; //@line 1285 "load_mid.cpp"
            $884 = ($883<<24>>24)==(0); //@line 1285 "load_mid.cpp"
            if (!($884)) {
             $885 = ((($e$08$i)) + 9|0); //@line 1286 "load_mid.cpp"
             $886 = HEAP8[$885>>0]|0; //@line 1286 "load_mid.cpp"
             $887 = $886&255; //@line 1286 "load_mid.cpp"
             $888 = (_pat_gm_drumnr($887)|0); //@line 1287 "load_mid.cpp"
             $889 = (_pat_gmtosmp($888)|0); //@line 1287 "load_mid.cpp"
             $890 = $889&255; //@line 1287 "load_mid.cpp"
             $891 = ((($e$08$i)) + 11|0); //@line 1287 "load_mid.cpp"
             HEAP8[$891>>0] = $890; //@line 1287 "load_mid.cpp"
             $892 = (_pat_gm_drumnote($887)|0); //@line 1288 "load_mid.cpp"
             $893 = $892&255; //@line 1288 "load_mid.cpp"
             HEAP8[$885>>0] = $893; //@line 1288 "load_mid.cpp"
             $894 = HEAP8[$882>>0]|0; //@line 1289 "load_mid.cpp"
             $895 = $894&255; //@line 1289 "load_mid.cpp"
             $896 = Math_imul($895, $v$05$i)|0; //@line 1289 "load_mid.cpp"
             $897 = (($896|0) / 128)&-1; //@line 1289 "load_mid.cpp"
             $898 = $897&255; //@line 1289 "load_mid.cpp"
             $899 = ($v$05$i|0)!=(0); //@line 1290 "load_mid.cpp"
             $900 = ($898<<24>>24)==(0); //@line 1290 "load_mid.cpp"
             $or$cond$i161 = $899 & $900; //@line 1290 "load_mid.cpp"
             $$$i162 = $or$cond$i161 ? 1 : $898; //@line 1290 "load_mid.cpp"
             HEAP8[$882>>0] = $$$i162; //@line 1290 "load_mid.cpp"
             $901 = HEAP32[$875>>2]|0; //@line 1291 "load_mid.cpp"
             $lno$1$i = $lno$03$i;$n$1$i = $887;$toff$1$i = $toff$07$i;$ton$1$i = $901;$v$1$i = $v$05$i;
             break;
            }
            $902 = (($ton$06$i) + ($865))|0; //@line 1294 "load_mid.cpp"
            $903 = ($902>>>0)>($storemerge$i159>>>0); //@line 1295 "load_mid.cpp"
            if (!($903)) {
             $lno$1$i = $e$08$i;$n$1$i = $n$04$i;$toff$1$i = $storemerge$i159;$ton$1$i = $ton$06$i;$v$1$i = $v$05$i;
             break;
            }
            $e$0$idx$val$i = HEAP32[$e$08$i>>2]|0; //@line 1767 "load_mid.cpp"
            $904 = ($e$0$idx$val$i|0)==(0|0); //@line 1254 "load_mid.cpp"
            if ($904) {
             $$0$i$i = 2147483647;
            } else {
             $905 = ((($e$0$idx$val$i)) + 4|0); //@line 1254 "load_mid.cpp"
             $906 = HEAP32[$905>>2]|0; //@line 1254 "load_mid.cpp"
             $$0$i$i = $906;
            }
            $907 = (($902) + ($865))|0; //@line 1297 "load_mid.cpp"
            $908 = ($907>>>0)<($$0$i$i>>>0); //@line 1297 "load_mid.cpp"
            if ($908) {
             HEAP32[$875>>2] = $902; //@line 1297 "load_mid.cpp"
             $lno$1$i = $e$08$i;$n$1$i = $n$04$i;$toff$1$i = $902;$ton$1$i = $ton$06$i;$v$1$i = $v$05$i;
             break;
            }
            $909 = ($902>>>0)<($$0$i$i>>>0); //@line 1299 "load_mid.cpp"
            if ($909) {
             $910 = (($902) + -1)|0; //@line 1299 "load_mid.cpp"
             HEAP32[$875>>2] = $910; //@line 1299 "load_mid.cpp"
             $lno$1$i = $e$08$i;$n$1$i = $n$04$i;$toff$1$i = $910;$ton$1$i = $ton$06$i;$v$1$i = $v$05$i;
             break;
            } else {
             $911 = (($$0$i$i) + -1)|0; //@line 1300 "load_mid.cpp"
             HEAP32[$875>>2] = $911; //@line 1300 "load_mid.cpp"
             $lno$1$i = $e$08$i;$n$1$i = $n$04$i;$toff$1$i = $911;$ton$1$i = $ton$06$i;$v$1$i = $v$05$i;
             break;
            }
           } else {
            $912 = ((($e$08$i)) + 12|0); //@line 1308 "load_mid.cpp"
            $913 = HEAP8[$912>>0]|0; //@line 1308 "load_mid.cpp"
            $914 = ($913<<24>>24)==(7); //@line 1308 "load_mid.cpp"
            if (!($914)) {
             $lno$1$i = $lno$03$i;$n$1$i = $n$04$i;$toff$1$i = $toff$07$i;$ton$1$i = $ton$06$i;$v$1$i = $v$05$i;
             break;
            }
            $915 = ((($e$08$i)) + 13|0); //@line 1309 "load_mid.cpp"
            $916 = HEAP8[$915>>0]|0; //@line 1309 "load_mid.cpp"
            $917 = $916&255; //@line 1309 "load_mid.cpp"
            $918 = ($916<<24>>24)==(0); //@line 1310 "load_mid.cpp"
            $919 = ($ton$06$i>>>0)>($toff$07$i>>>0); //@line 1310 "load_mid.cpp"
            $or$cond1$i = $919 & $918; //@line 1310 "load_mid.cpp"
            if (!($or$cond1$i)) {
             $lno$1$i = $lno$03$i;$n$1$i = $n$04$i;$toff$1$i = $toff$07$i;$ton$1$i = $ton$06$i;$v$1$i = $917;
             break;
            }
            HEAP8[$879>>0] = 1; //@line 1311 "load_mid.cpp"
            $920 = ((($e$08$i)) + 10|0); //@line 1312 "load_mid.cpp"
            HEAP8[$920>>0] = 0; //@line 1312 "load_mid.cpp"
            $921 = (_pat_gm_drumnote($n$04$i)|0); //@line 1313 "load_mid.cpp"
            $922 = $921&255; //@line 1313 "load_mid.cpp"
            $923 = ((($e$08$i)) + 9|0); //@line 1313 "load_mid.cpp"
            HEAP8[$923>>0] = $922; //@line 1313 "load_mid.cpp"
            $924 = HEAP32[$875>>2]|0; //@line 1314 "load_mid.cpp"
            $lno$1$i = $e$08$i;$n$1$i = $n$04$i;$toff$1$i = $924;$ton$1$i = $ton$06$i;$v$1$i = $917;
           }
          } while(0);
          $e$0$i163 = HEAP32[$e$08$i>>2]|0; //@line 1281 "load_mid.cpp"
          $925 = ($e$0$i163|0)==(0|0); //@line 1281 "load_mid.cpp"
          if ($925) {
           $lno$1$i$lcssa = $lno$1$i;$n$1$i$lcssa = $n$1$i;$toff$1$i$lcssa = $toff$1$i;$ton$1$i$lcssa = $ton$1$i;
           break;
          } else {
           $e$08$i = $e$0$i163;$lno$03$i = $lno$1$i;$n$04$i = $n$1$i;$toff$07$i = $toff$1$i;$ton$06$i = $ton$1$i;$v$05$i = $v$1$i;
          }
         }
         $926 = ($ton$1$i$lcssa>>>0)>($toff$1$i$lcssa>>>0); //@line 1320 "load_mid.cpp"
         if ($926) {
          HEAP32[$vararg_buffer147>>2] = $ton$1$i$lcssa; //@line 1322 "load_mid.cpp"
          $vararg_ptr150 = ((($vararg_buffer147)) + 4|0); //@line 1322 "load_mid.cpp"
          HEAP32[$vararg_ptr150>>2] = $toff$1$i$lcssa; //@line 1322 "load_mid.cpp"
          $vararg_ptr151 = ((($vararg_buffer147)) + 8|0); //@line 1322 "load_mid.cpp"
          HEAP32[$vararg_ptr151>>2] = $n$1$i$lcssa; //@line 1322 "load_mid.cpp"
          (_sprintf($info$i,529030,$vararg_buffer147)|0); //@line 1322 "load_mid.cpp"
          $927 = (_strlen($info$i)|0); //@line 245 "load_mid.cpp"
          $928 = (($927) + 33)|0; //@line 245 "load_mid.cpp"
          $929 = ($928>>>0)>(255); //@line 245 "load_mid.cpp"
          if (!($929)) {
           HEAP32[$vararg_buffer152>>2] = $info$i; //@line 246 "load_mid.cpp"
           (_sprintf($txt$i,529048,$vararg_buffer152)|0); //@line 246 "load_mid.cpp"
           HEAP32[$vararg_buffer155>>2] = $txt$i; //@line 250 "load_mid.cpp"
           (_fprintf($867,528176,$vararg_buffer155)|0); //@line 250 "load_mid.cpp"
          }
         }
         $930 = ($lno$1$i$lcssa|0)==(0|0); //@line 1325 "load_mid.cpp"
         if ($930) {
          break;
         }
         $931 = HEAP32[$lno$1$i$lcssa>>2]|0; //@line 1325 "load_mid.cpp"
         $932 = ($931|0)==(0|0); //@line 1325 "load_mid.cpp"
         if ($932) {
          break;
         } else {
          $ep$01$i$i = $931;
         }
         while(1) {
          $933 = HEAP32[$ep$01$i$i>>2]|0; //@line 1264 "load_mid.cpp"
          _free($ep$01$i$i); //@line 1265 "load_mid.cpp"
          $934 = ($933|0)==(0|0); //@line 1263 "load_mid.cpp"
          if ($934) {
           break;
          } else {
           $ep$01$i$i = $933;
          }
         }
         HEAP32[$lno$1$i$lcssa>>2] = 0; //@line 1268 "load_mid.cpp"
         $935 = ((($ttp$1266)) + 8|0); //@line 1269 "load_mid.cpp"
         HEAP32[$935>>2] = $lno$1$i$lcssa; //@line 1269 "load_mid.cpp"
         $936 = HEAP32[$873>>2]|0; //@line 1270 "load_mid.cpp"
         $937 = ((($ttp$1266)) + 12|0); //@line 1270 "load_mid.cpp"
         HEAP32[$937>>2] = $936; //@line 1270 "load_mid.cpp"
         $938 = ((($lno$1$i$lcssa)) + 4|0); //@line 1271 "load_mid.cpp"
         $939 = HEAP32[$938>>2]|0; //@line 1271 "load_mid.cpp"
         $940 = ($936|0)==(0); //@line 395 "load_mid.cpp"
         if ($940) {
          break;
         }
         $$cast$i$i = $936; //@line 394 "load_mid.cpp"
         $941 = ((($ttp$1266)) + 27|0); //@line 388 "load_mid.cpp"
         $942 = ((($ttp$1266)) + 25|0); //@line 384 "load_mid.cpp"
         $943 = ((($ttp$1266)) + 26|0); //@line 385 "load_mid.cpp"
         $944 = ((($ttp$1266)) + 20|0); //@line 386 "load_mid.cpp"
         $$cast$pn$i$i = $$cast$i$i;
         while(1) {
          $$in$i$i = ((($$cast$pn$i$i)) + 4|0); //@line 395 "load_mid.cpp"
          $945 = HEAP32[$$in$i$i>>2]|0; //@line 395 "load_mid.cpp"
          $946 = ($945>>>0)>($939>>>0); //@line 396 "load_mid.cpp"
          if ($946) {
           break L400;
          }
          HEAP32[$937>>2] = $$cast$pn$i$i; //@line 397 "load_mid.cpp"
          $947 = ((($$cast$pn$i$i)) + 8|0); //@line 382 "load_mid.cpp"
          $948 = HEAP8[$947>>0]|0; //@line 382 "load_mid.cpp"
          $949 = ($948<<24>>24)==(0); //@line 382 "load_mid.cpp"
          if (!($949)) {
           $950 = ((($$cast$pn$i$i)) + 10|0); //@line 383 "load_mid.cpp"
           $951 = HEAP8[$950>>0]|0; //@line 383 "load_mid.cpp"
           $952 = ($951<<24>>24)==(0); //@line 383 "load_mid.cpp"
           if ($952) {
            $$sink$i$i$i$i = -1;
           } else {
            $953 = ((($$cast$pn$i$i)) + 9|0); //@line 383 "load_mid.cpp"
            $954 = HEAP8[$953>>0]|0; //@line 383 "load_mid.cpp"
            $$sink$i$i$i$i = $954;
           }
           HEAP8[$942>>0] = $$sink$i$i$i$i; //@line 384 "load_mid.cpp"
           $955 = HEAP8[$950>>0]|0; //@line 385 "load_mid.cpp"
           HEAP8[$943>>0] = $955; //@line 385 "load_mid.cpp"
           $956 = HEAP32[$$in$i$i>>2]|0; //@line 386 "load_mid.cpp"
           HEAP32[$944>>2] = $956; //@line 386 "load_mid.cpp"
          }
          $957 = ((($$cast$pn$i$i)) + 12|0); //@line 388 "load_mid.cpp"
          $958 = HEAP8[$957>>0]|0; //@line 388 "load_mid.cpp"
          $959 = ($958<<24>>24)==(8); //@line 388 "load_mid.cpp"
          if ($959) {
           $960 = ((($$cast$pn$i$i)) + 13|0); //@line 388 "load_mid.cpp"
           $961 = HEAP8[$960>>0]|0; //@line 388 "load_mid.cpp"
           HEAP8[$941>>0] = $961; //@line 388 "load_mid.cpp"
          }
          $962 = HEAP32[$$cast$pn$i$i>>2]|0; //@line 396 "load_mid.cpp"
          $963 = ($962|0)==(0|0); //@line 396 "load_mid.cpp"
          if ($963) {
           break;
          } else {
           $$cast$pn$i$i = $962;
          }
         }
        }
       } while(0);
      } else {
       $964 = ((($ttp$1266)) + 4|0); //@line 1334 "load_mid.cpp"
       $e$02$i168 = HEAP32[$964>>2]|0; //@line 1334 "load_mid.cpp"
       $965 = ($e$02$i168|0)==(0|0); //@line 1334 "load_mid.cpp"
       if (!($965)) {
        $e$09$i = $e$02$i168;$i$04$i = 0;$lno$03$i169 = 0;$n$05$i = 0;$toff$08$i = 0;$ton$07$i = 0;$v$06$i = 127;
        while(1) {
         $966 = ((($e$09$i)) + 4|0); //@line 1335 "load_mid.cpp"
         $967 = HEAP32[$966>>2]|0; //@line 1335 "load_mid.cpp"
         $968 = ($967>>>0)<($delta$0$lcssa368>>>0); //@line 1335 "load_mid.cpp"
         $969 = (($967) - ($delta$0$lcssa368))|0; //@line 1336 "load_mid.cpp"
         $storemerge$i170 = $968 ? 0 : $969; //@line 1335 "load_mid.cpp"
         HEAP32[$966>>2] = $storemerge$i170; //@line 1335 "load_mid.cpp"
         $970 = ((($e$09$i)) + 8|0); //@line 1337 "load_mid.cpp"
         $971 = HEAP8[$970>>0]|0; //@line 1337 "load_mid.cpp"
         $972 = ($971<<24>>24)==(1); //@line 1337 "load_mid.cpp"
         L450: do {
          if ($972) {
           $973 = ($i$04$i|0)==(0); //@line 1338 "load_mid.cpp"
           if ($973) {
            $974 = (_pat_gmtosmp(1)|0); //@line 1338 "load_mid.cpp"
            $i$1$i = $974;
           } else {
            $i$1$i = $i$04$i;
           }
           $975 = $i$1$i&255; //@line 1339 "load_mid.cpp"
           $976 = ((($e$09$i)) + 11|0); //@line 1339 "load_mid.cpp"
           HEAP8[$976>>0] = $975; //@line 1339 "load_mid.cpp"
           $977 = ((($e$09$i)) + 9|0); //@line 1340 "load_mid.cpp"
           $978 = HEAP8[$977>>0]|0; //@line 1340 "load_mid.cpp"
           $979 = $978&255; //@line 1340 "load_mid.cpp"
           $980 = ((($e$09$i)) + 10|0); //@line 1341 "load_mid.cpp"
           $981 = HEAP8[$980>>0]|0; //@line 1341 "load_mid.cpp"
           $982 = ($981<<24>>24)==(0); //@line 1341 "load_mid.cpp"
           if (!($982)) {
            $983 = $981&255; //@line 1341 "load_mid.cpp"
            $984 = Math_imul($983, $v$06$i)|0; //@line 1342 "load_mid.cpp"
            $985 = (($984|0) / 128)&-1; //@line 1342 "load_mid.cpp"
            $986 = $985&255; //@line 1342 "load_mid.cpp"
            $987 = ($v$06$i|0)!=(0); //@line 1343 "load_mid.cpp"
            $988 = ($986<<24>>24)==(0); //@line 1343 "load_mid.cpp"
            $or$cond$i172 = $987 & $988; //@line 1343 "load_mid.cpp"
            $$$i173 = $or$cond$i172 ? 1 : $986; //@line 1343 "load_mid.cpp"
            HEAP8[$980>>0] = $$$i173; //@line 1343 "load_mid.cpp"
            $989 = HEAP32[$966>>2]|0; //@line 1344 "load_mid.cpp"
            $i$3$i = $i$1$i;$lno$1$i180 = $lno$03$i169;$n$1$i181 = $979;$toff$1$i184 = $toff$08$i;$ton$1$i183 = $989;$v$1$i182 = $v$06$i;
            break;
           }
           $990 = (($ton$07$i) + ($865))|0; //@line 1347 "load_mid.cpp"
           $991 = HEAP32[$966>>2]|0; //@line 1348 "load_mid.cpp"
           $992 = ($990>>>0)>($991>>>0); //@line 1348 "load_mid.cpp"
           if (!($992)) {
            $i$3$i = $i$1$i;$lno$1$i180 = $e$09$i;$n$1$i181 = $979;$toff$1$i184 = $991;$ton$1$i183 = $ton$07$i;$v$1$i182 = $v$06$i;
            break;
           }
           $e$0$idx$val$i175 = HEAP32[$e$09$i>>2]|0; //@line 1769 "load_mid.cpp"
           $993 = ($e$0$idx$val$i175|0)==(0|0); //@line 1254 "load_mid.cpp"
           if ($993) {
            $$0$i$i176 = 2147483647;
           } else {
            $994 = ((($e$0$idx$val$i175)) + 4|0); //@line 1254 "load_mid.cpp"
            $995 = HEAP32[$994>>2]|0; //@line 1254 "load_mid.cpp"
            $$0$i$i176 = $995;
           }
           $996 = (($990) + ($865))|0; //@line 1350 "load_mid.cpp"
           $997 = ($996>>>0)<($$0$i$i176>>>0); //@line 1350 "load_mid.cpp"
           if ($997) {
            HEAP32[$966>>2] = $990; //@line 1350 "load_mid.cpp"
            $i$3$i = $i$1$i;$lno$1$i180 = $e$09$i;$n$1$i181 = $979;$toff$1$i184 = $990;$ton$1$i183 = $ton$07$i;$v$1$i182 = $v$06$i;
            break;
           }
           $998 = ($990>>>0)<($$0$i$i176>>>0); //@line 1352 "load_mid.cpp"
           if ($998) {
            $999 = (($990) + -1)|0; //@line 1352 "load_mid.cpp"
            HEAP32[$966>>2] = $999; //@line 1352 "load_mid.cpp"
            $i$3$i = $i$1$i;$lno$1$i180 = $e$09$i;$n$1$i181 = $979;$toff$1$i184 = $999;$ton$1$i183 = $ton$07$i;$v$1$i182 = $v$06$i;
            break;
           } else {
            $1000 = (($$0$i$i176) + -1)|0; //@line 1353 "load_mid.cpp"
            HEAP32[$966>>2] = $1000; //@line 1353 "load_mid.cpp"
            $i$3$i = $i$1$i;$lno$1$i180 = $e$09$i;$n$1$i181 = $979;$toff$1$i184 = $1000;$ton$1$i183 = $ton$07$i;$v$1$i182 = $v$06$i;
            break;
           }
          } else {
           $1001 = ((($e$09$i)) + 12|0); //@line 1361 "load_mid.cpp"
           $1002 = HEAP8[$1001>>0]|0; //@line 1361 "load_mid.cpp"
           switch ($1002<<24>>24) {
           case 8:  {
            $1003 = ((($e$09$i)) + 13|0); //@line 1361 "load_mid.cpp"
            $1004 = HEAP8[$1003>>0]|0; //@line 1361 "load_mid.cpp"
            $1005 = $1004&255; //@line 1361 "load_mid.cpp"
            $i$3$i = $1005;$lno$1$i180 = $lno$03$i169;$n$1$i181 = $n$05$i;$toff$1$i184 = $toff$08$i;$ton$1$i183 = $ton$07$i;$v$1$i182 = $v$06$i;
            break L450;
            break;
           }
           case 7:  {
            break;
           }
           default: {
            $i$3$i = $i$04$i;$lno$1$i180 = $lno$03$i169;$n$1$i181 = $n$05$i;$toff$1$i184 = $toff$08$i;$ton$1$i183 = $ton$07$i;$v$1$i182 = $v$06$i;
            break L450;
           }
           }
           $1006 = ((($e$09$i)) + 13|0); //@line 1363 "load_mid.cpp"
           $1007 = HEAP8[$1006>>0]|0; //@line 1363 "load_mid.cpp"
           $1008 = $1007&255; //@line 1363 "load_mid.cpp"
           $1009 = ($1007<<24>>24)==(0); //@line 1364 "load_mid.cpp"
           $1010 = ($ton$07$i>>>0)>($toff$08$i>>>0); //@line 1364 "load_mid.cpp"
           $or$cond1$i179 = $1010 & $1009; //@line 1364 "load_mid.cpp"
           if (!($or$cond1$i179)) {
            $i$3$i = $i$04$i;$lno$1$i180 = $lno$03$i169;$n$1$i181 = $n$05$i;$toff$1$i184 = $toff$08$i;$ton$1$i183 = $ton$07$i;$v$1$i182 = $1008;
            break;
           }
           HEAP8[$970>>0] = 1; //@line 1365 "load_mid.cpp"
           $1011 = ((($e$09$i)) + 10|0); //@line 1366 "load_mid.cpp"
           HEAP8[$1011>>0] = 0; //@line 1366 "load_mid.cpp"
           $1012 = $n$05$i&255; //@line 1367 "load_mid.cpp"
           $1013 = ((($e$09$i)) + 9|0); //@line 1367 "load_mid.cpp"
           HEAP8[$1013>>0] = $1012; //@line 1367 "load_mid.cpp"
           $i$3$i = $i$04$i;$lno$1$i180 = $e$09$i;$n$1$i181 = $n$05$i;$toff$1$i184 = $storemerge$i170;$ton$1$i183 = $ton$07$i;$v$1$i182 = $1008;
          }
         } while(0);
         $e$0$i185 = HEAP32[$e$09$i>>2]|0; //@line 1334 "load_mid.cpp"
         $1014 = ($e$0$i185|0)==(0|0); //@line 1334 "load_mid.cpp"
         if ($1014) {
          $lno$1$i180$lcssa = $lno$1$i180;$n$1$i181$lcssa = $n$1$i181;$toff$1$i184$lcssa = $toff$1$i184;$ton$1$i183$lcssa = $ton$1$i183;
          break;
         } else {
          $e$09$i = $e$0$i185;$i$04$i = $i$3$i;$lno$03$i169 = $lno$1$i180;$n$05$i = $n$1$i181;$toff$08$i = $toff$1$i184;$ton$07$i = $ton$1$i183;$v$06$i = $v$1$i182;
         }
        }
        $1015 = ($ton$1$i183$lcssa>>>0)>($toff$1$i184$lcssa>>>0); //@line 1374 "load_mid.cpp"
        if ($1015) {
         $1016 = HEAP8[$868>>0]|0; //@line 1376 "load_mid.cpp"
         $1017 = $1016&255; //@line 1376 "load_mid.cpp"
         $1018 = (($1017) + 1)|0; //@line 1376 "load_mid.cpp"
         HEAP32[$vararg_buffer158>>2] = $1018; //@line 1376 "load_mid.cpp"
         $vararg_ptr161 = ((($vararg_buffer158)) + 4|0); //@line 1376 "load_mid.cpp"
         HEAP32[$vararg_ptr161>>2] = $ton$1$i183$lcssa; //@line 1376 "load_mid.cpp"
         $vararg_ptr162 = ((($vararg_buffer158)) + 8|0); //@line 1376 "load_mid.cpp"
         HEAP32[$vararg_ptr162>>2] = $toff$1$i184$lcssa; //@line 1376 "load_mid.cpp"
         $vararg_ptr163 = ((($vararg_buffer158)) + 12|0); //@line 1376 "load_mid.cpp"
         HEAP32[$vararg_ptr163>>2] = $n$1$i181$lcssa; //@line 1376 "load_mid.cpp"
         (_sprintf($info$i167,529082,$vararg_buffer158)|0); //@line 1376 "load_mid.cpp"
         $1019 = (_strlen($info$i167)|0); //@line 245 "load_mid.cpp"
         $1020 = (($1019) + 35)|0; //@line 245 "load_mid.cpp"
         $1021 = ($1020>>>0)>(255); //@line 245 "load_mid.cpp"
         if (!($1021)) {
          HEAP32[$vararg_buffer164>>2] = $info$i167; //@line 246 "load_mid.cpp"
          (_sprintf($txt$i,529112,$vararg_buffer164)|0); //@line 246 "load_mid.cpp"
          HEAP32[$vararg_buffer167>>2] = $txt$i; //@line 250 "load_mid.cpp"
          (_fprintf($867,528176,$vararg_buffer167)|0); //@line 250 "load_mid.cpp"
         }
        }
        $1022 = ($lno$1$i180$lcssa|0)==(0|0); //@line 1379 "load_mid.cpp"
        if ($1022) {
         break;
        }
        $1023 = HEAP32[$lno$1$i180$lcssa>>2]|0; //@line 1379 "load_mid.cpp"
        $1024 = ($1023|0)==(0|0); //@line 1379 "load_mid.cpp"
        if ($1024) {
         break;
        } else {
         $ep$01$i$i188 = $1023;
        }
        while(1) {
         $1025 = HEAP32[$ep$01$i$i188>>2]|0; //@line 1264 "load_mid.cpp"
         _free($ep$01$i$i188); //@line 1265 "load_mid.cpp"
         $1026 = ($1025|0)==(0|0); //@line 1263 "load_mid.cpp"
         if ($1026) {
          break;
         } else {
          $ep$01$i$i188 = $1025;
         }
        }
        HEAP32[$lno$1$i180$lcssa>>2] = 0; //@line 1268 "load_mid.cpp"
        $1027 = ((($ttp$1266)) + 8|0); //@line 1269 "load_mid.cpp"
        HEAP32[$1027>>2] = $lno$1$i180$lcssa; //@line 1269 "load_mid.cpp"
        $1028 = HEAP32[$964>>2]|0; //@line 1270 "load_mid.cpp"
        $1029 = ((($ttp$1266)) + 12|0); //@line 1270 "load_mid.cpp"
        HEAP32[$1029>>2] = $1028; //@line 1270 "load_mid.cpp"
        $1030 = ((($lno$1$i180$lcssa)) + 4|0); //@line 1271 "load_mid.cpp"
        $1031 = HEAP32[$1030>>2]|0; //@line 1271 "load_mid.cpp"
        $1032 = ($1028|0)==(0); //@line 395 "load_mid.cpp"
        if ($1032) {
         break;
        }
        $$cast$i$i191 = $1028; //@line 394 "load_mid.cpp"
        $1033 = ((($ttp$1266)) + 27|0); //@line 388 "load_mid.cpp"
        $1034 = ((($ttp$1266)) + 25|0); //@line 384 "load_mid.cpp"
        $1035 = ((($ttp$1266)) + 26|0); //@line 385 "load_mid.cpp"
        $1036 = ((($ttp$1266)) + 20|0); //@line 386 "load_mid.cpp"
        $$cast$pn$i$i193 = $$cast$i$i191;
        while(1) {
         $$in$i$i194 = ((($$cast$pn$i$i193)) + 4|0); //@line 395 "load_mid.cpp"
         $1037 = HEAP32[$$in$i$i194>>2]|0; //@line 395 "load_mid.cpp"
         $1038 = ($1037>>>0)>($1031>>>0); //@line 396 "load_mid.cpp"
         if ($1038) {
          break L398;
         }
         HEAP32[$1029>>2] = $$cast$pn$i$i193; //@line 397 "load_mid.cpp"
         $1039 = ((($$cast$pn$i$i193)) + 8|0); //@line 382 "load_mid.cpp"
         $1040 = HEAP8[$1039>>0]|0; //@line 382 "load_mid.cpp"
         $1041 = ($1040<<24>>24)==(0); //@line 382 "load_mid.cpp"
         if (!($1041)) {
          $1042 = ((($$cast$pn$i$i193)) + 10|0); //@line 383 "load_mid.cpp"
          $1043 = HEAP8[$1042>>0]|0; //@line 383 "load_mid.cpp"
          $1044 = ($1043<<24>>24)==(0); //@line 383 "load_mid.cpp"
          if ($1044) {
           $$sink$i$i$i$i196 = -1;
          } else {
           $1045 = ((($$cast$pn$i$i193)) + 9|0); //@line 383 "load_mid.cpp"
           $1046 = HEAP8[$1045>>0]|0; //@line 383 "load_mid.cpp"
           $$sink$i$i$i$i196 = $1046;
          }
          HEAP8[$1034>>0] = $$sink$i$i$i$i196; //@line 384 "load_mid.cpp"
          $1047 = HEAP8[$1042>>0]|0; //@line 385 "load_mid.cpp"
          HEAP8[$1035>>0] = $1047; //@line 385 "load_mid.cpp"
          $1048 = HEAP32[$$in$i$i194>>2]|0; //@line 386 "load_mid.cpp"
          HEAP32[$1036>>2] = $1048; //@line 386 "load_mid.cpp"
         }
         $1049 = ((($$cast$pn$i$i193)) + 12|0); //@line 388 "load_mid.cpp"
         $1050 = HEAP8[$1049>>0]|0; //@line 388 "load_mid.cpp"
         $1051 = ($1050<<24>>24)==(8); //@line 388 "load_mid.cpp"
         if ($1051) {
          $1052 = ((($$cast$pn$i$i193)) + 13|0); //@line 388 "load_mid.cpp"
          $1053 = HEAP8[$1052>>0]|0; //@line 388 "load_mid.cpp"
          HEAP8[$1033>>0] = $1053; //@line 388 "load_mid.cpp"
         }
         $1054 = HEAP32[$$cast$pn$i$i193>>2]|0; //@line 396 "load_mid.cpp"
         $1055 = ($1054|0)==(0|0); //@line 396 "load_mid.cpp"
         if ($1055) {
          break;
         } else {
          $$cast$pn$i$i193 = $1054;
         }
        }
       }
      }
     } while(0);
     $1056 = ((($ttp$1266)) + 8|0); //@line 1770 "load_mid.cpp"
     $1057 = HEAP32[$1056>>2]|0; //@line 1770 "load_mid.cpp"
     $1058 = ($1057|0)==(0|0); //@line 1770 "load_mid.cpp"
     if (!($1058)) {
      $1059 = ((($1057)) + 4|0); //@line 1770 "load_mid.cpp"
      $1060 = HEAP32[$1059>>2]|0; //@line 1770 "load_mid.cpp"
      $1061 = HEAP32[$83>>2]|0; //@line 1770 "load_mid.cpp"
      $1062 = ($1060>>>0)>($1061>>>0); //@line 1770 "load_mid.cpp"
      if ($1062) {
       HEAP32[$83>>2] = $1060; //@line 1771 "load_mid.cpp"
      }
     }
     $ttp$1 = HEAP32[$ttp$1266>>2]|0; //@line 1765 "load_mid.cpp"
     $1063 = ($ttp$1|0)==(0|0); //@line 1765 "load_mid.cpp"
     if ($1063) {
      break;
     } else {
      $ttp$1266 = $ttp$1;
     }
    }
    $$pre323 = HEAP32[$83>>2]|0; //@line 1773 "load_mid.cpp"
    $$pre324 = HEAP32[$29>>2]|0; //@line 547 "load_mid.cpp"
    $1067 = $$pre323;$1072 = $$pre324;
   }
   $1064 = HEAP32[$$pre$phi340Z2D>>2]|0; //@line 1773 "load_mid.cpp"
   $1065 = $1064 >> 2; //@line 1773 "load_mid.cpp"
   $1066 = (($1067) + ($1065))|0; //@line 1773 "load_mid.cpp"
   HEAP32[$83>>2] = $1066; //@line 1773 "load_mid.cpp"
   $1068 = (_calloc(1,16)|0); //@line 298 "load_mid.cpp"
   HEAP32[$1068>>2] = 0; //@line 299 "load_mid.cpp"
   $1069 = ((($1068)) + 4|0); //@line 300 "load_mid.cpp"
   HEAP32[$1069>>2] = $1066; //@line 300 "load_mid.cpp"
   $1070 = ((($1068)) + 8|0); //@line 301 "load_mid.cpp"
   $1071 = ((($1068)) + 12|0); //@line 546 "load_mid.cpp"
   ;HEAP8[$1070>>0]=0|0;HEAP8[$1070+1>>0]=0|0;HEAP8[$1070+2>>0]=0|0;HEAP8[$1070+3>>0]=0|0;HEAP8[$1070+4>>0]=0|0;HEAP8[$1070+5>>0]=0|0; //@line 545 "load_mid.cpp"
   HEAP8[$1071>>0] = 3; //@line 546 "load_mid.cpp"
   __ZL13mid_add_eventP10_MIDHANDLEP9_MIDTRACKP9_MIDEVENT($1072,$1068); //@line 547 "load_mid.cpp"
   $1073 = HEAP32[$31>>2]|0; //@line 1775 "load_mid.cpp"
   $1074 = ($1073|0)==(0|0); //@line 1775 "load_mid.cpp"
   if ($1074) {
    label = 365;
   } else {
    $1075 = HEAP32[$83>>2]|0; //@line 209 "load_mid.cpp"
    HEAP32[$vararg_buffer170>>2] = $1075; //@line 209 "load_mid.cpp"
    (_printf(529148,$vararg_buffer170)|0); //@line 209 "load_mid.cpp"
    $1076 = HEAP32[$84>>2]|0; //@line 210 "load_mid.cpp"
    HEAP32[$vararg_buffer173>>2] = $1076; //@line 210 "load_mid.cpp"
    (_printf(529166,$vararg_buffer173)|0); //@line 210 "load_mid.cpp"
    $1077 = HEAP32[$100>>2]|0; //@line 211 "load_mid.cpp"
    HEAP32[$vararg_buffer176>>2] = $1077; //@line 211 "load_mid.cpp"
    (_printf(529183,$vararg_buffer176)|0); //@line 211 "load_mid.cpp"
    $1078 = HEAP32[$49>>2]|0; //@line 212 "load_mid.cpp"
    HEAP32[$vararg_buffer179>>2] = $1078; //@line 212 "load_mid.cpp"
    (_printf(529200,$vararg_buffer179)|0); //@line 212 "load_mid.cpp"
    $1079 = HEAP32[$69>>2]|0; //@line 213 "load_mid.cpp"
    HEAP32[$vararg_buffer182>>2] = $1079; //@line 213 "load_mid.cpp"
    (_printf(529217,$vararg_buffer182)|0); //@line 213 "load_mid.cpp"
    $1080 = HEAP32[$59>>2]|0; //@line 214 "load_mid.cpp"
    HEAP32[$vararg_buffer185>>2] = $1080; //@line 214 "load_mid.cpp"
    (_printf(529234,$vararg_buffer185)|0); //@line 214 "load_mid.cpp"
    $1081 = HEAP32[$$pre$phi340Z2D>>2]|0; //@line 215 "load_mid.cpp"
    HEAP32[$vararg_buffer188>>2] = $1081; //@line 215 "load_mid.cpp"
    (_printf(529251,$vararg_buffer188)|0); //@line 215 "load_mid.cpp"
    $1082 = HEAP32[$81>>2]|0; //@line 216 "load_mid.cpp"
    HEAP32[$vararg_buffer191>>2] = $1082; //@line 216 "load_mid.cpp"
    (_printf(529268,$vararg_buffer191)|0); //@line 216 "load_mid.cpp"
    $1083 = HEAP32[$30>>2]|0; //@line 217 "load_mid.cpp"
    HEAP32[$vararg_buffer194>>2] = $1083; //@line 217 "load_mid.cpp"
    (_printf(529285,$vararg_buffer194)|0); //@line 217 "load_mid.cpp"
    $1084 = ((($27)) + 56|0); //@line 218 "load_mid.cpp"
    $1085 = HEAP32[$1084>>2]|0; //@line 218 "load_mid.cpp"
    HEAP32[$vararg_buffer197>>2] = $1085; //@line 218 "load_mid.cpp"
    (_printf(529302,$vararg_buffer197)|0); //@line 218 "load_mid.cpp"
    $tr$04$i = HEAP32[$29>>2]|0; //@line 220 "load_mid.cpp"
    $1086 = ($tr$04$i|0)==(0|0); //@line 220 "load_mid.cpp"
    if ($1086) {
     $c$0$lcssa$i44 = 0;$t$01$i$pre339 = 0;
    } else {
     $t$05$i50 = 0;$tr$06$i = $tr$04$i;
     while(1) {
      $1087 = (($t$05$i50) + 1)|0; //@line 221 "load_mid.cpp"
      $1088 = ((($tr$06$i)) + 24|0); //@line 222 "load_mid.cpp"
      $1089 = HEAP8[$1088>>0]|0; //@line 222 "load_mid.cpp"
      $1090 = $1089&255; //@line 222 "load_mid.cpp"
      $1091 = (($1090) + 1)|0; //@line 222 "load_mid.cpp"
      $1092 = ((($tr$06$i)) + 25|0); //@line 222 "load_mid.cpp"
      $1093 = HEAP8[$1092>>0]|0; //@line 222 "load_mid.cpp"
      $1094 = $1093&255; //@line 222 "load_mid.cpp"
      $1095 = ((($tr$06$i)) + 16|0); //@line 222 "load_mid.cpp"
      $1096 = HEAP32[$1095>>2]|0; //@line 222 "load_mid.cpp"
      $1097 = ((($tr$06$i)) + 26|0); //@line 222 "load_mid.cpp"
      $1098 = HEAP8[$1097>>0]|0; //@line 222 "load_mid.cpp"
      $1099 = $1098&255; //@line 222 "load_mid.cpp"
      $1100 = ((($tr$06$i)) + 27|0); //@line 222 "load_mid.cpp"
      $1101 = HEAP8[$1100>>0]|0; //@line 222 "load_mid.cpp"
      $1102 = $1101&255; //@line 222 "load_mid.cpp"
      HEAP32[$vararg_buffer200>>2] = $1087; //@line 222 "load_mid.cpp"
      $vararg_ptr203 = ((($vararg_buffer200)) + 4|0); //@line 222 "load_mid.cpp"
      HEAP32[$vararg_ptr203>>2] = $1091; //@line 222 "load_mid.cpp"
      $vararg_ptr204 = ((($vararg_buffer200)) + 8|0); //@line 222 "load_mid.cpp"
      HEAP32[$vararg_ptr204>>2] = $1094; //@line 222 "load_mid.cpp"
      $vararg_ptr205 = ((($vararg_buffer200)) + 12|0); //@line 222 "load_mid.cpp"
      HEAP32[$vararg_ptr205>>2] = $1096; //@line 222 "load_mid.cpp"
      $vararg_ptr206 = ((($vararg_buffer200)) + 16|0); //@line 222 "load_mid.cpp"
      HEAP32[$vararg_ptr206>>2] = $1099; //@line 222 "load_mid.cpp"
      $vararg_ptr207 = ((($vararg_buffer200)) + 20|0); //@line 222 "load_mid.cpp"
      HEAP32[$vararg_ptr207>>2] = $1102; //@line 222 "load_mid.cpp"
      (_printf(529320,$vararg_buffer200)|0); //@line 222 "load_mid.cpp"
      $1103 = ((($tr$06$i)) + 4|0); //@line 223 "load_mid.cpp"
      $e$02$i51 = HEAP32[$1103>>2]|0; //@line 223 "load_mid.cpp"
      $1104 = ($e$02$i51|0)==(0|0); //@line 223 "load_mid.cpp"
      L509: do {
       if (!($1104)) {
        $e$03$i52 = $e$02$i51;
        while(1) {
         $1105 = ((($e$03$i52)) + 4|0); //@line 225 "load_mid.cpp"
         $1106 = HEAP32[$1105>>2]|0; //@line 225 "load_mid.cpp"
         $1107 = ((($e$03$i52)) + 8|0); //@line 226 "load_mid.cpp"
         $1108 = HEAP8[$1107>>0]|0; //@line 226 "load_mid.cpp"
         $1109 = ($1108<<24>>24)!=(0); //@line 226 "load_mid.cpp"
         $$str56$$str57$i = $1109 ? 529378 : 529383; //@line 226 "load_mid.cpp"
         $1110 = ((($e$03$i52)) + 9|0); //@line 226 "load_mid.cpp"
         $1111 = HEAP8[$1110>>0]|0; //@line 226 "load_mid.cpp"
         $1112 = $1111&255; //@line 226 "load_mid.cpp"
         $1113 = ((($e$03$i52)) + 10|0); //@line 226 "load_mid.cpp"
         $1114 = HEAP8[$1113>>0]|0; //@line 226 "load_mid.cpp"
         $1115 = $1114&255; //@line 226 "load_mid.cpp"
         $1116 = ((($e$03$i52)) + 11|0); //@line 226 "load_mid.cpp"
         $1117 = HEAP8[$1116>>0]|0; //@line 226 "load_mid.cpp"
         $1118 = $1117&255; //@line 226 "load_mid.cpp"
         HEAP32[$vararg_buffer208>>2] = $1087; //@line 224 "load_mid.cpp"
         $vararg_ptr211 = ((($vararg_buffer208)) + 4|0); //@line 224 "load_mid.cpp"
         HEAP32[$vararg_ptr211>>2] = $1106; //@line 224 "load_mid.cpp"
         $vararg_ptr212 = ((($vararg_buffer208)) + 8|0); //@line 224 "load_mid.cpp"
         HEAP32[$vararg_ptr212>>2] = $$str56$$str57$i; //@line 224 "load_mid.cpp"
         $vararg_ptr213 = ((($vararg_buffer208)) + 12|0); //@line 224 "load_mid.cpp"
         HEAP32[$vararg_ptr213>>2] = $1112; //@line 224 "load_mid.cpp"
         $vararg_ptr214 = ((($vararg_buffer208)) + 16|0); //@line 224 "load_mid.cpp"
         HEAP32[$vararg_ptr214>>2] = $1115; //@line 224 "load_mid.cpp"
         $vararg_ptr215 = ((($vararg_buffer208)) + 20|0); //@line 224 "load_mid.cpp"
         HEAP32[$vararg_ptr215>>2] = $1118; //@line 224 "load_mid.cpp"
         (_printf(529388,$vararg_buffer208)|0); //@line 224 "load_mid.cpp"
         $1119 = ((($e$03$i52)) + 12|0); //@line 227 "load_mid.cpp"
         $1120 = HEAP8[$1119>>0]|0; //@line 227 "load_mid.cpp"
         $1121 = $1120&255; //@line 227 "load_mid.cpp"
         switch ($1121|0) {
         case 3:  {
          (_puts(529413)|0); //@line 228 "load_mid.cpp"
          break;
         }
         case 5:  {
          (_puts(529419)|0); //@line 229 "load_mid.cpp"
          break;
         }
         case 8:  {
          $1122 = ((($e$03$i52)) + 13|0); //@line 230 "load_mid.cpp"
          $1123 = HEAP8[$1122>>0]|0; //@line 230 "load_mid.cpp"
          $1124 = $1123&255; //@line 230 "load_mid.cpp"
          HEAP32[$vararg_buffer216>>2] = $1124; //@line 230 "load_mid.cpp"
          (_printf(529426,$vararg_buffer216)|0); //@line 230 "load_mid.cpp"
          break;
         }
         case 7:  {
          $1125 = ((($e$03$i52)) + 13|0); //@line 231 "load_mid.cpp"
          $1126 = HEAP8[$1125>>0]|0; //@line 231 "load_mid.cpp"
          $1127 = $1126&255; //@line 231 "load_mid.cpp"
          HEAP32[$vararg_buffer219>>2] = $1127; //@line 231 "load_mid.cpp"
          (_printf(529435,$vararg_buffer219)|0); //@line 231 "load_mid.cpp"
          break;
         }
         case 6:  {
          $1128 = ((($e$03$i52)) + 13|0); //@line 232 "load_mid.cpp"
          $1129 = HEAP8[$1128>>0]|0; //@line 232 "load_mid.cpp"
          $1130 = $1129&255; //@line 232 "load_mid.cpp"
          HEAP32[$vararg_buffer222>>2] = $1130; //@line 232 "load_mid.cpp"
          (_printf(529447,$vararg_buffer222)|0); //@line 232 "load_mid.cpp"
          break;
         }
         case 1:  {
          $1131 = ((($e$03$i52)) + 13|0); //@line 233 "load_mid.cpp"
          $1132 = HEAP8[$1131>>0]|0; //@line 233 "load_mid.cpp"
          $1133 = $1132&255; //@line 233 "load_mid.cpp"
          HEAP32[$vararg_buffer225>>2] = $1133; //@line 233 "load_mid.cpp"
          (_printf(529460,$vararg_buffer225)|0); //@line 233 "load_mid.cpp"
          break;
         }
         case 2:  {
          $1134 = ((($e$03$i52)) + 13|0); //@line 234 "load_mid.cpp"
          $1135 = HEAP8[$1134>>0]|0; //@line 234 "load_mid.cpp"
          $1136 = $1135&255; //@line 234 "load_mid.cpp"
          HEAP32[$vararg_buffer228>>2] = $1136; //@line 234 "load_mid.cpp"
          (_printf(529474,$vararg_buffer228)|0); //@line 234 "load_mid.cpp"
          break;
         }
         case 4:  {
          $1137 = ((($e$03$i52)) + 13|0); //@line 235 "load_mid.cpp"
          $1138 = HEAP8[$1137>>0]|0; //@line 235 "load_mid.cpp"
          $1139 = $1138&255; //@line 235 "load_mid.cpp"
          HEAP32[$vararg_buffer231>>2] = $1139; //@line 235 "load_mid.cpp"
          (_printf(529486,$vararg_buffer231)|0); //@line 235 "load_mid.cpp"
          break;
         }
         default: {
          (_putchar(10)|0); //@line 236 "load_mid.cpp"
         }
         }
         $e$0$i54 = HEAP32[$e$03$i52>>2]|0; //@line 223 "load_mid.cpp"
         $1140 = ($e$0$i54|0)==(0|0); //@line 223 "load_mid.cpp"
         if ($1140) {
          break L509;
         } else {
          $e$03$i52 = $e$0$i54;
         }
        }
       }
      } while(0);
      $tr$0$i55 = HEAP32[$tr$06$i>>2]|0; //@line 220 "load_mid.cpp"
      $1141 = ($tr$0$i55|0)==(0|0); //@line 220 "load_mid.cpp"
      if ($1141) {
       label = 365;
       break;
      } else {
       $t$05$i50 = $1087;$tr$06$i = $tr$0$i55;
      }
     }
    }
   }
   if ((label|0) == 365) {
    $t$03$i43$pr = HEAP32[$29>>2]|0; //@line 345 "load_mid.cpp"
    $1142 = ($t$03$i43$pr|0)==(0|0); //@line 345 "load_mid.cpp"
    if ($1142) {
     $c$0$lcssa$i44 = 0;$t$01$i$pre339 = 0;
    } else {
     $c$04$i47 = 0;$t$05$i46 = $t$03$i43$pr;
     while(1) {
      $1143 = ((($t$05$i46)) + 24|0); //@line 346 "load_mid.cpp"
      $1144 = HEAP8[$1143>>0]|0; //@line 346 "load_mid.cpp"
      $1145 = $1144&255; //@line 346 "load_mid.cpp"
      $1146 = 1 << $1145; //@line 346 "load_mid.cpp"
      $1147 = $1146 | $c$04$i47; //@line 346 "load_mid.cpp"
      $t$0$i48 = HEAP32[$t$05$i46>>2]|0; //@line 345 "load_mid.cpp"
      $1148 = ($t$0$i48|0)==(0|0); //@line 345 "load_mid.cpp"
      if ($1148) {
       $c$0$lcssa$i44 = $1147;$t$01$i$pre339 = $t$03$i43$pr;
       break;
      } else {
       $c$04$i47 = $1147;$t$05$i46 = $t$0$i48;
      }
     }
    }
   }
   $1149 = $c$0$lcssa$i44 & 1; //@line 349 "load_mid.cpp"
   $1150 = $c$0$lcssa$i44 >>> 1; //@line 349 "load_mid.cpp"
   $$lobit$i = $1150 & 1; //@line 349 "load_mid.cpp"
   $n$0$$1$i = (($$lobit$i) + ($1149))|0; //@line 349 "load_mid.cpp"
   $1151 = $c$0$lcssa$i44 >>> 2; //@line 349 "load_mid.cpp"
   $$lobit6$i = $1151 & 1; //@line 349 "load_mid.cpp"
   $n$0$$2$i = (($n$0$$1$i) + ($$lobit6$i))|0; //@line 349 "load_mid.cpp"
   $1152 = $c$0$lcssa$i44 >>> 3; //@line 349 "load_mid.cpp"
   $$lobit7$i = $1152 & 1; //@line 349 "load_mid.cpp"
   $n$0$$3$i = (($n$0$$2$i) + ($$lobit7$i))|0; //@line 349 "load_mid.cpp"
   $1153 = $c$0$lcssa$i44 >>> 4; //@line 349 "load_mid.cpp"
   $$lobit8$i = $1153 & 1; //@line 349 "load_mid.cpp"
   $n$0$$4$i = (($n$0$$3$i) + ($$lobit8$i))|0; //@line 349 "load_mid.cpp"
   $1154 = $c$0$lcssa$i44 >>> 5; //@line 349 "load_mid.cpp"
   $$lobit9$i = $1154 & 1; //@line 349 "load_mid.cpp"
   $n$0$$5$i = (($n$0$$4$i) + ($$lobit9$i))|0; //@line 349 "load_mid.cpp"
   $1155 = $c$0$lcssa$i44 >>> 6; //@line 349 "load_mid.cpp"
   $$lobit10$i = $1155 & 1; //@line 349 "load_mid.cpp"
   $n$0$$6$i = (($n$0$$5$i) + ($$lobit10$i))|0; //@line 349 "load_mid.cpp"
   $1156 = $c$0$lcssa$i44 >>> 7; //@line 349 "load_mid.cpp"
   $$lobit11$i = $1156 & 1; //@line 349 "load_mid.cpp"
   $n$0$$7$i = (($n$0$$6$i) + ($$lobit11$i))|0; //@line 349 "load_mid.cpp"
   $1157 = $c$0$lcssa$i44 >>> 8; //@line 349 "load_mid.cpp"
   $$lobit12$i = $1157 & 1; //@line 349 "load_mid.cpp"
   $n$0$$8$i = (($n$0$$7$i) + ($$lobit12$i))|0; //@line 349 "load_mid.cpp"
   $1158 = $c$0$lcssa$i44 >>> 9; //@line 349 "load_mid.cpp"
   $$lobit13$i = $1158 & 1; //@line 349 "load_mid.cpp"
   $n$0$$9$i = (($n$0$$8$i) + ($$lobit13$i))|0; //@line 349 "load_mid.cpp"
   $1159 = $c$0$lcssa$i44 >>> 10; //@line 349 "load_mid.cpp"
   $$lobit14$i = $1159 & 1; //@line 349 "load_mid.cpp"
   $n$0$$10$i = (($n$0$$9$i) + ($$lobit14$i))|0; //@line 349 "load_mid.cpp"
   $1160 = $c$0$lcssa$i44 >>> 11; //@line 349 "load_mid.cpp"
   $$lobit15$i = $1160 & 1; //@line 349 "load_mid.cpp"
   $n$0$$11$i = (($n$0$$10$i) + ($$lobit15$i))|0; //@line 349 "load_mid.cpp"
   $1161 = $c$0$lcssa$i44 >>> 12; //@line 349 "load_mid.cpp"
   $$lobit16$i = $1161 & 1; //@line 349 "load_mid.cpp"
   $n$0$$12$i = (($n$0$$11$i) + ($$lobit16$i))|0; //@line 349 "load_mid.cpp"
   $1162 = $c$0$lcssa$i44 >>> 13; //@line 349 "load_mid.cpp"
   $$lobit17$i = $1162 & 1; //@line 349 "load_mid.cpp"
   $n$0$$13$i = (($n$0$$12$i) + ($$lobit17$i))|0; //@line 349 "load_mid.cpp"
   $1163 = $c$0$lcssa$i44 >>> 14; //@line 349 "load_mid.cpp"
   $$lobit18$i = $1163 & 1; //@line 349 "load_mid.cpp"
   $n$0$$14$i = (($n$0$$13$i) + ($$lobit18$i))|0; //@line 349 "load_mid.cpp"
   $1164 = $c$0$lcssa$i44 >>> 15; //@line 349 "load_mid.cpp"
   $$lobit19$i = $1164 & 1; //@line 349 "load_mid.cpp"
   $n$0$$15$i = (($n$0$$14$i) + ($$lobit19$i))|0; //@line 349 "load_mid.cpp"
   $notlhs = ($panhigh$0|0)>(79); //@line 1778 "load_mid.cpp"
   $notrhs = ($panlow$0|0)<(49); //@line 1778 "load_mid.cpp"
   $or$cond3$not = $notrhs & $notlhs; //@line 1778 "load_mid.cpp"
   $1165 = ($t$01$i$pre339|0)==(0|0); //@line 1779 "load_mid.cpp"
   $or$cond402 = $or$cond3$not | $1165; //@line 1778 "load_mid.cpp"
   if ($or$cond402) {
    $t$01$i = $t$01$i$pre339;
   } else {
    $1166 = $n$0$$15$i << 6; //@line 1780 "load_mid.cpp"
    $t$03$i37 = $t$01$i$pre339;$ttp$2263 = $t$01$i$pre339;
    while(1) {
     $1167 = ((($ttp$2263)) + 24|0); //@line 1780 "load_mid.cpp"
     $1168 = HEAP8[$1167>>0]|0; //@line 1780 "load_mid.cpp"
     $1169 = $1168&255; //@line 1780 "load_mid.cpp"
     $1170 = ($t$03$i37|0)==(0|0); //@line 359 "load_mid.cpp"
     if ($1170) {
      $c$0$lcssa$i = 0;
     } else {
      $c$04$i = 0;$t$05$i38 = $t$03$i37;
      while(1) {
       $1172 = ((($t$05$i38)) + 24|0); //@line 360 "load_mid.cpp"
       $1173 = HEAP8[$1172>>0]|0; //@line 360 "load_mid.cpp"
       $1174 = $1173&255; //@line 360 "load_mid.cpp"
       $1175 = 1 << $1174; //@line 360 "load_mid.cpp"
       $1176 = $1175 | $c$04$i; //@line 360 "load_mid.cpp"
       $t$0$i39 = HEAP32[$t$05$i38>>2]|0; //@line 359 "load_mid.cpp"
       $1177 = ($t$0$i39|0)==(0|0); //@line 359 "load_mid.cpp"
       if ($1177) {
        $c$0$lcssa$i = $1176;
        break;
       } else {
        $c$04$i = $1176;$t$05$i38 = $t$0$i39;
       }
      }
     }
     $1171 = ($1168<<24>>24)==(0); //@line 362 "load_mid.cpp"
     if ($1171) {
      $n$0$lcssa$i42 = 0;
     } else {
      $i$01$i = 0;$n$02$i40 = 0;
      while(1) {
       $1178 = 1 << $i$01$i; //@line 363 "load_mid.cpp"
       $1179 = $1178 & $c$0$lcssa$i; //@line 363 "load_mid.cpp"
       $not$$i = ($1179|0)!=(0); //@line 363 "load_mid.cpp"
       $1180 = $not$$i&1; //@line 363 "load_mid.cpp"
       $n$0$$i = (($1180) + ($n$02$i40))|0; //@line 363 "load_mid.cpp"
       $1181 = (($i$01$i) + 1)|0; //@line 362 "load_mid.cpp"
       $exitcond$i = ($1181|0)==($1169|0); //@line 362 "load_mid.cpp"
       if ($exitcond$i) {
        $n$0$lcssa$i42 = $n$0$$i;
        break;
       } else {
        $i$01$i = $1181;$n$02$i40 = $n$0$$i;
       }
      }
     }
     $1182 = $n$0$lcssa$i42 << 7; //@line 1780 "load_mid.cpp"
     $1183 = (($1182) + ($1166))|0; //@line 1780 "load_mid.cpp"
     $1184 = (($1183|0) / ($n$0$$15$i|0))&-1; //@line 1780 "load_mid.cpp"
     $1185 = $1184 & 127; //@line 1780 "load_mid.cpp"
     $1186 = ((($ttp$2263)) + 16|0); //@line 1780 "load_mid.cpp"
     HEAP32[$1186>>2] = $1185; //@line 1780 "load_mid.cpp"
     $ttp$2 = HEAP32[$ttp$2263>>2]|0; //@line 1779 "load_mid.cpp"
     $1187 = ($ttp$2|0)==(0|0); //@line 1779 "load_mid.cpp"
     $t$01$i$pre$pre = HEAP32[$29>>2]|0; //@line 334 "load_mid.cpp"
     if ($1187) {
      $t$01$i = $t$01$i$pre$pre;
      break;
     } else {
      $t$03$i37 = $t$01$i$pre$pre;$ttp$2263 = $ttp$2;
     }
    }
   }
   $1188 = ($t$01$i|0)==(0|0); //@line 334 "load_mid.cpp"
   if ($1188) {
    $n$0$lcssa$i = 0;
   } else {
    $n$02$i = 0;$t$03$i = $t$01$i;
    while(1) {
     $1189 = (($n$02$i) + 1)|0; //@line 335 "load_mid.cpp"
     $t$0$i = HEAP32[$t$03$i>>2]|0; //@line 334 "load_mid.cpp"
     $1190 = ($t$0$i|0)==(0|0); //@line 334 "load_mid.cpp"
     if ($1190) {
      $n$0$lcssa$i = $1189;
      break;
     } else {
      $n$02$i = $1189;$t$03$i = $t$0$i;
     }
    }
   }
   $1191 = HEAP32[$82>>2]|0; //@line 1785 "load_mid.cpp"
   $1192 = ($1191|0)==(0); //@line 1785 "load_mid.cpp"
   if ($1192) {
    $1193 = HEAP32[$81>>2]|0; //@line 1785 "load_mid.cpp"
    HEAP32[$82>>2] = $1193; //@line 1785 "load_mid.cpp"
    $1324 = $1193;
   } else {
    $1324 = $1191;
   }
   $1194 = ($maxtempo$0|0)==(0); //@line 1786 "load_mid.cpp"
   if ($1194) {
    $1195 = HEAP32[$81>>2]|0; //@line 1786 "load_mid.cpp"
    $maxtempo$3 = $1195;
   } else {
    $maxtempo$3 = $maxtempo$0;
   }
   $1196 = ($maxtempo$3|0)==(255); //@line 1787 "load_mid.cpp"
   do {
    if ($1196) {
     $1234 = $1324;
     label = 395;
    } else {
     $1197 = HEAP32[$36>>2]|0; //@line 1788 "load_mid.cpp"
     $1198 = ($1197|0)==(0|0); //@line 1788 "load_mid.cpp"
     if (!($1198)) {
      HEAP32[$vararg_buffer234>>2] = $maxtempo$3; //@line 1788 "load_mid.cpp"
      (_printf(529495,$vararg_buffer234)|0); //@line 1788 "load_mid.cpp"
     }
     $1199 = ($maxtempo$3|0)<(1); //@line 273 "load_mid.cpp"
     if ($1199) {
      break;
     }
     $1200 = HEAP32[$$pre$phi340Z2D>>2]|0; //@line 274 "load_mid.cpp"
     $1201 = Math_imul($1200, $maxtempo$3)|0; //@line 276 "load_mid.cpp"
     $storemerge4$i = (($1201|0) / 255)&-1; //@line 276 "load_mid.cpp"
     HEAP32[$$pre$phi340Z2D>>2] = $storemerge4$i; //@line 279 "load_mid.cpp"
     $1202 = HEAP32[$84>>2]|0; //@line 277 "load_mid.cpp"
     $1203 = Math_imul($1202, $storemerge4$i)|0; //@line 256 "load_mid.cpp"
     $1204 = $1203 >>> 4; //@line 256 "load_mid.cpp"
     $1205 = (($1204>>>0) / ($1202>>>0))&-1; //@line 256 "load_mid.cpp"
     HEAP32[$100>>2] = $1205; //@line 277 "load_mid.cpp"
     $1206 = ($1205|0)<($1202|0); //@line 277 "load_mid.cpp"
     do {
      if ($1206) {
       $t$05$i = $maxtempo$3;
       while(1) {
        $1207 = (($t$05$i) + 1)|0; //@line 278 "load_mid.cpp"
        $1208 = Math_imul($1207, $1200)|0; //@line 279 "load_mid.cpp"
        $storemerge$i = (($1208|0) / 255)&-1; //@line 276 "load_mid.cpp"
        $1209 = Math_imul($storemerge$i, $1202)|0; //@line 256 "load_mid.cpp"
        $1210 = $1209 >>> 4; //@line 256 "load_mid.cpp"
        $1211 = (($1210>>>0) / ($1202>>>0))&-1; //@line 256 "load_mid.cpp"
        $1212 = ($1211|0)<($1202|0); //@line 277 "load_mid.cpp"
        if ($1212) {
         $t$05$i = $1207;
        } else {
         $$lcssa477 = $1207;$$lcssa478 = $1211;$storemerge$i$lcssa = $storemerge$i;$t$05$i$lcssa = $t$05$i;
         break;
        }
       }
       HEAP32[$$pre$phi340Z2D>>2] = $storemerge$i$lcssa; //@line 279 "load_mid.cpp"
       HEAP32[$100>>2] = $$lcssa478; //@line 277 "load_mid.cpp"
       $1213 = HEAP32[$36>>2]|0; //@line 281 "load_mid.cpp"
       $1214 = ($1213|0)!=(0|0); //@line 281 "load_mid.cpp"
       $1215 = ($t$05$i$lcssa|0)>=($maxtempo$3|0); //@line 281 "load_mid.cpp"
       $or$cond$i34 = $1215 & $1214; //@line 281 "load_mid.cpp"
       if (!($or$cond$i34)) {
        $t$0$lcssa12$i = $$lcssa477;
        break;
       }
       $1216 = $maxtempo$3 << 1; //@line 283 "load_mid.cpp"
       $1217 = (($1216) - ($$lcssa477))|0; //@line 283 "load_mid.cpp"
       HEAP32[$vararg_buffer237>>2] = $maxtempo$3; //@line 282 "load_mid.cpp"
       $vararg_ptr240 = ((($vararg_buffer237)) + 4|0); //@line 282 "load_mid.cpp"
       HEAP32[$vararg_ptr240>>2] = $1217; //@line 282 "load_mid.cpp"
       $vararg_ptr241 = ((($vararg_buffer237)) + 8|0); //@line 282 "load_mid.cpp"
       HEAP32[$vararg_ptr241>>2] = $$lcssa478; //@line 282 "load_mid.cpp"
       (_printf(529522,$vararg_buffer237)|0); //@line 282 "load_mid.cpp"
       $t$0$lcssa12$i = $$lcssa477;
      } else {
       $t$0$lcssa12$i = $maxtempo$3;
      }
     } while(0);
     $1218 = HEAP32[$29>>2]|0; //@line 284 "load_mid.cpp"
     $1219 = ($1218|0)==(0|0); //@line 284 "load_mid.cpp"
     do {
      if (!($1219)) {
       $1220 = ((($1218)) + 4|0); //@line 285 "load_mid.cpp"
       $e$02$i = HEAP32[$1220>>2]|0; //@line 285 "load_mid.cpp"
       $1221 = ($e$02$i|0)==(0|0); //@line 285 "load_mid.cpp"
       if ($1221) {
        break;
       } else {
        $e$03$i = $e$02$i;
       }
       while(1) {
        $1222 = ((($e$03$i)) + 12|0); //@line 286 "load_mid.cpp"
        $1223 = HEAP8[$1222>>0]|0; //@line 286 "load_mid.cpp"
        $1224 = ($1223<<24>>24)==(4); //@line 286 "load_mid.cpp"
        if ($1224) {
         $1225 = ((($e$03$i)) + 13|0); //@line 287 "load_mid.cpp"
         $1226 = HEAP8[$1225>>0]|0; //@line 287 "load_mid.cpp"
         $1227 = $1226&255; //@line 287 "load_mid.cpp"
         $1228 = ($1227*255)|0; //@line 287 "load_mid.cpp"
         $1229 = (($1228|0) / ($t$0$lcssa12$i|0))&-1; //@line 287 "load_mid.cpp"
         $1230 = $1229&255; //@line 287 "load_mid.cpp"
         HEAP8[$1225>>0] = $1230; //@line 287 "load_mid.cpp"
        }
        $e$0$i = HEAP32[$e$03$i>>2]|0; //@line 285 "load_mid.cpp"
        $1231 = ($e$0$i|0)==(0|0); //@line 285 "load_mid.cpp"
        if ($1231) {
         break;
        } else {
         $e$03$i = $e$0$i;
        }
       }
      }
     } while(0);
     $1232 = ($maxtempo$3|0)>(0); //@line 1791 "load_mid.cpp"
     if (!($1232)) {
      break;
     }
     $$pre328 = HEAP32[$82>>2]|0; //@line 1791 "load_mid.cpp"
     $1234 = $$pre328;
     label = 395;
    }
   } while(0);
   if ((label|0) == 395) {
    $1233 = ($1234*255)|0; //@line 1791 "load_mid.cpp"
    $1235 = (($1233>>>0) / ($maxtempo$3>>>0))&-1; //@line 1791 "load_mid.cpp"
    HEAP32[$82>>2] = $1235; //@line 1791 "load_mid.cpp"
   }
   $1236 = HEAP32[$83>>2]|0; //@line 1792 "load_mid.cpp"
   $$idx29$val = HEAP32[$84>>2]|0;
   $$idx30$val = HEAP32[$$pre$phi340Z2D>>2]|0;
   $1237 = $1236 << 4; //@line 261 "load_mid.cpp"
   $1238 = Math_imul($1237, $$idx29$val)|0; //@line 261 "load_mid.cpp"
   $1239 = (($1238>>>0) / ($$idx30$val>>>0))&-1; //@line 261 "load_mid.cpp"
   $1240 = (($1239>>>0) / ($$idx29$val>>>0))&-1; //@line 1792 "load_mid.cpp"
   $1241 = $1240 >>> 6; //@line 1792 "load_mid.cpp"
   $1242 = (($1241) + 1)|0; //@line 1792 "load_mid.cpp"
   $1243 = HEAP32[$36>>2]|0; //@line 1793 "load_mid.cpp"
   $1244 = ($1243|0)==(0|0); //@line 1793 "load_mid.cpp"
   if ($1244) {
    $1247 = $$idx29$val;
   } else {
    HEAP32[$vararg_buffer242>>2] = $1242; //@line 1793 "load_mid.cpp"
    $vararg_ptr245 = ((($vararg_buffer242)) + 4|0); //@line 1793 "load_mid.cpp"
    HEAP32[$vararg_ptr245>>2] = $$idx29$val; //@line 1793 "load_mid.cpp"
    (_printf(529595,$vararg_buffer242)|0); //@line 1793 "load_mid.cpp"
    $$pre329 = HEAP32[$84>>2]|0; //@line 1840 "load_mid.cpp"
    $1247 = $$pre329;
   }
   $1245 = ((($this)) + 59680|0); //@line 1839 "load_mid.cpp"
   HEAP32[$1245>>2] = 65536; //@line 1839 "load_mid.cpp"
   $1246 = ((($this)) + 59648|0); //@line 1840 "load_mid.cpp"
   HEAP32[$1246>>2] = $1247; //@line 1840 "load_mid.cpp"
   $1248 = ((($this)) + 59664|0); //@line 1841 "load_mid.cpp"
   HEAP32[$1248>>2] = $n$0$lcssa$i; //@line 1841 "load_mid.cpp"
   $1249 = ((($this)) + 59660|0); //@line 1842 "load_mid.cpp"
   HEAP32[$1249>>2] = 16; //@line 1842 "load_mid.cpp"
   $1250 = ((($this)) + 59764|0); //@line 1843 "load_mid.cpp"
   HEAP32[$1250>>2] = 112; //@line 1843 "load_mid.cpp"
   $1251 = ((($this)) + 59768|0); //@line 1844 "load_mid.cpp"
   HEAP32[$1251>>2] = 13696; //@line 1844 "load_mid.cpp"
   $1252 = Math_imul($1236, $$idx29$val)|0; //@line 1846 "load_mid.cpp"
   $1253 = $1252 << 4; //@line 1846 "load_mid.cpp"
   $1254 = (($1253>>>0) / ($$idx30$val>>>0))&-1; //@line 1846 "load_mid.cpp"
   $1255 = (($1254>>>0) / ($$idx29$val>>>0))&-1; //@line 1846 "load_mid.cpp"
   $1256 = $1255 >>> 6; //@line 1846 "load_mid.cpp"
   $1257 = (($1256) + 1)|0; //@line 1846 "load_mid.cpp"
   $t$1261 = 0;
   while(1) {
    $1258 = $t$1261&255; //@line 1847 "load_mid.cpp"
    $1259 = (((($this)) + 53344|0) + ($t$1261)|0); //@line 1847 "load_mid.cpp"
    HEAP8[$1259>>0] = $1258; //@line 1847 "load_mid.cpp"
    $1260 = (($t$1261) + 1)|0; //@line 1846 "load_mid.cpp"
    $exitcond321 = ($1260|0)==($1257|0); //@line 1846 "load_mid.cpp"
    if ($exitcond321) {
     break;
    } else {
     $t$1261 = $1260;
    }
   }
   $1261 = (_PAT_Load_Instruments($this)|0); //@line 1848 "load_mid.cpp"
   if (!($1261)) {
    HEAP32[300431] = 0; //@line 1849 "load_mid.cpp"
    $$0 = 0;
    STACKTOP = sp;return ($$0|0); //@line 1880 "load_mid.cpp"
   }
   $1262 = ((($this)) + 51904|0); //@line 1854 "load_mid.cpp"
   $1263 = ((($this)) + 52864|0); //@line 1854 "load_mid.cpp"
   $1264 = HEAP32[$1248>>2]|0; //@line 1854 "load_mid.cpp"
   $1265 = (__ZL16MID_ReadPatternsPP11_MODCOMMANDPtP10_MIDHANDLEii($1262,$1263,$27,$1242,$1264)|0); //@line 1854 "load_mid.cpp"
   $1266 = ($1265|0)==(0); //@line 1854 "load_mid.cpp"
   if (!($1266)) {
    $1267 = HEAP32[$1248>>2]|0; //@line 1856 "load_mid.cpp"
    $1268 = (($1267) + 1)|0; //@line 1856 "load_mid.cpp"
    HEAP32[$1248>>2] = $1268; //@line 1856 "load_mid.cpp"
    $1269 = HEAP32[$29>>2]|0; //@line 1857 "load_mid.cpp"
    $1270 = ((($1269)) + 24|0); //@line 1857 "load_mid.cpp"
    $1271 = HEAP8[$1270>>0]|0; //@line 1857 "load_mid.cpp"
    $1272 = (_calloc(1,28)|0); //@line 315 "load_mid.cpp"
    HEAP32[$1272>>2] = 0; //@line 316 "load_mid.cpp"
    $1273 = ((($1272)) + 25|0); //@line 317 "load_mid.cpp"
    HEAP8[$1273>>0] = -1; //@line 317 "load_mid.cpp"
    $1274 = ((($1272)) + 27|0); //@line 318 "load_mid.cpp"
    HEAP8[$1274>>0] = 1; //@line 318 "load_mid.cpp"
    $1275 = ((($1272)) + 24|0); //@line 319 "load_mid.cpp"
    HEAP8[$1275>>0] = $1271; //@line 319 "load_mid.cpp"
    $1276 = ((($1272)) + 4|0); //@line 320 "load_mid.cpp"
    HEAP32[$1276>>2] = 0; //@line 320 "load_mid.cpp"
    $1277 = ((($1272)) + 8|0); //@line 321 "load_mid.cpp"
    HEAP32[$1277>>2] = 0; //@line 321 "load_mid.cpp"
    $1278 = ((($1272)) + 12|0); //@line 322 "load_mid.cpp"
    HEAP32[$1278>>2] = 0; //@line 322 "load_mid.cpp"
    $1279 = ((($1272)) + 20|0); //@line 323 "load_mid.cpp"
    HEAP32[$1279>>2] = 0; //@line 323 "load_mid.cpp"
    $1280 = ($1269|0)==(0|0); //@line 324 "load_mid.cpp"
    if ($1280) {
     $1284 = 120;
    } else {
     $1281 = ((($1269)) + 26|0); //@line 324 "load_mid.cpp"
     $1282 = HEAP8[$1281>>0]|0; //@line 324 "load_mid.cpp"
     $1284 = $1282;
    }
    $1283 = ((($1272)) + 26|0); //@line 324 "load_mid.cpp"
    HEAP8[$1283>>0] = $1284; //@line 324 "load_mid.cpp"
    $1285 = ((($1272)) + 16|0); //@line 325 "load_mid.cpp"
    HEAP32[$1285>>2] = 64; //@line 325 "load_mid.cpp"
    HEAP32[$111>>2] = $1272; //@line 1857 "load_mid.cpp"
    $1286 = $1272;
    $ttp$3 = $1269;
    while(1) {
     $1287 = HEAP32[$ttp$3>>2]|0; //@line 1858 "load_mid.cpp"
     $1288 = ($1287|0)==(0|0); //@line 1858 "load_mid.cpp"
     if ($1288) {
      $ttp$3$lcssa = $ttp$3;
      break;
     } else {
      $ttp$3 = $1287;
     }
    }
    HEAP32[$ttp$3$lcssa>>2] = $1286; //@line 1859 "load_mid.cpp"
    $1289 = (_calloc(1,16)|0); //@line 298 "load_mid.cpp"
    HEAP32[$1289>>2] = 0; //@line 299 "load_mid.cpp"
    $1290 = HEAP32[$83>>2]|0; //@line 300 "load_mid.cpp"
    $1291 = ((($1289)) + 4|0); //@line 300 "load_mid.cpp"
    HEAP32[$1291>>2] = $1290; //@line 300 "load_mid.cpp"
    $1292 = ((($1289)) + 8|0); //@line 301 "load_mid.cpp"
    $1293 = ((($1289)) + 12|0); //@line 700 "load_mid.cpp"
    ;HEAP8[$1292>>0]=0|0;HEAP8[$1292+1>>0]=0|0;HEAP8[$1292+2>>0]=0|0;HEAP8[$1292+3>>0]=0|0;HEAP8[$1292+4>>0]=0|0;HEAP8[$1292+5>>0]=0|0; //@line 699 "load_mid.cpp"
    HEAP8[$1293>>0] = 5; //@line 700 "load_mid.cpp"
    __ZL13mid_add_eventP10_MIDHANDLEP9_MIDTRACKP9_MIDEVENT($1272,$1289); //@line 701 "load_mid.cpp"
    $t$2260 = 0;
    while(1) {
     $1294 = (((($this)) + 51904|0) + ($t$2260<<2)|0); //@line 1862 "load_mid.cpp"
     $1295 = HEAP32[$1294>>2]|0; //@line 1862 "load_mid.cpp"
     __ZN10CSoundFile11FreePatternEPv($1295); //@line 1862 "load_mid.cpp"
     HEAP32[$1294>>2] = 0; //@line 1863 "load_mid.cpp"
     $1296 = (($t$2260) + 1)|0; //@line 1861 "load_mid.cpp"
     $exitcond = ($1296|0)==($1257|0); //@line 1861 "load_mid.cpp"
     if ($exitcond) {
      break;
     } else {
      $t$2260 = $1296;
     }
    }
    $1297 = HEAP32[$1248>>2]|0; //@line 1865 "load_mid.cpp"
    (__ZL16MID_ReadPatternsPP11_MODCOMMANDPtP10_MIDHANDLEii($1262,$1263,$27,$1242,$1297)|0); //@line 1865 "load_mid.cpp"
   }
   $ttp$4255 = HEAP32[$29>>2]|0; //@line 1870 "load_mid.cpp"
   $1298 = ($ttp$4255|0)==(0|0); //@line 1870 "load_mid.cpp"
   do {
    if (!($1298)) {
     $1299 = (($n$0$$15$i|0) / 2)&-1; //@line 1871 "load_mid.cpp"
     $1300 = $1299 << 1; //@line 626 "load_mid.cpp"
     $1301 = $1300 | 1; //@line 626 "load_mid.cpp"
     $1302 = ($1301*48)|0; //@line 628 "load_mid.cpp"
     $t$3256 = 0;$ttp$4257 = $ttp$4255;
     while(1) {
      $1303 = ((($ttp$4257)) + 16|0); //@line 1871 "load_mid.cpp"
      $1304 = HEAP32[$1303>>2]|0; //@line 1871 "load_mid.cpp"
      $1305 = ($1304*160)|0; //@line 627 "load_mid.cpp"
      $1306 = Math_imul($1305, $1301)|0; //@line 627 "load_mid.cpp"
      $1307 = $1306 >> 7; //@line 627 "load_mid.cpp"
      $1308 = (($1307) + ($1302))|0; //@line 628 "load_mid.cpp"
      $1309 = (($1308|0) / ($1301|0))&-1; //@line 628 "load_mid.cpp"
      $1310 = (((($this)) + 49600|0) + (($t$3256*36)|0)|0); //@line 1871 "load_mid.cpp"
      HEAP32[$1310>>2] = $1309; //@line 1871 "load_mid.cpp"
      $1311 = (((((($this)) + 49600|0) + (($t$3256*36)|0)|0)) + 4|0); //@line 1872 "load_mid.cpp"
      HEAP32[$1311>>2] = 64; //@line 1872 "load_mid.cpp"
      $1312 = (($t$3256) + 1)|0; //@line 1873 "load_mid.cpp"
      $ttp$4 = HEAP32[$ttp$4257>>2]|0; //@line 1870 "load_mid.cpp"
      $1313 = ($ttp$4|0)==(0|0); //@line 1870 "load_mid.cpp"
      if ($1313) {
       break;
      } else {
       $t$3256 = $1312;$ttp$4257 = $ttp$4;
      }
     }
     if ($1298) {
      break;
     } else {
      $tp$01$i$i = $ttp$4255;
     }
     while(1) {
      $1314 = HEAP32[$tp$01$i$i>>2]|0; //@line 857 "load_mid.cpp"
      $1315 = ((($tp$01$i$i)) + 4|0); //@line 835 "load_mid.cpp"
      $1316 = HEAP32[$1315>>2]|0; //@line 835 "load_mid.cpp"
      $1317 = ($1316|0)==(0|0); //@line 835 "load_mid.cpp"
      if (!($1317)) {
       $ep$01$i$i$i = $1316;
       while(1) {
        $1318 = HEAP32[$ep$01$i$i$i>>2]|0; //@line 836 "load_mid.cpp"
        _free($ep$01$i$i$i); //@line 837 "load_mid.cpp"
        $1319 = ($1318|0)==(0|0); //@line 835 "load_mid.cpp"
        if ($1319) {
         break;
        } else {
         $ep$01$i$i$i = $1318;
        }
       }
      }
      HEAP32[$1315>>2] = 0; //@line 839 "load_mid.cpp"
      $1320 = ($1314|0)==(0|0); //@line 856 "load_mid.cpp"
      if ($1320) {
       break;
      } else {
       $tp$01$i$i = $1314;
      }
     }
    }
   } while(0);
   _free($27); //@line 878 "load_mid.cpp"
   $1321 = HEAP32[$36>>2]|0; //@line 1877 "load_mid.cpp"
   $1322 = ($1321|0)==(0|0); //@line 1877 "load_mid.cpp"
   if (!($1322)) {
    (_puts(529633)|0); //@line 1877 "load_mid.cpp"
   }
   HEAP32[300431] = 0; //@line 1878 "load_mid.cpp"
   $$0 = 1;
   STACKTOP = sp;return ($$0|0); //@line 1880 "load_mid.cpp"
  }
 } else {
 }
 HEAP32[300431] = 0; //@line 1431 "load_mid.cpp"
 $$0 = 0;
 STACKTOP = sp;return ($$0|0); //@line 1880 "load_mid.cpp"
}
function __ZL13mid_add_eventP10_MIDHANDLEP9_MIDTRACKP9_MIDEVENT($tp,$e) {
 $tp = $tp|0;
 $e = $e|0;
 var $$sink$i = 0, $$sink$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, $ep$0$lcssa14 = 0, $ep$04 = 0, $ep$04$lcssa = 0, $ep$04$phi = 0, $ew$0$lcssa15 = 0, $ew$0$ph12 = 0, $ew$03 = 0, $ew$03$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($tp)) + 12|0); //@line 510 "load_mid.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 510 "load_mid.cpp"
 $2 = ($1|0)==(0|0); //@line 511 "load_mid.cpp"
 do {
  if ($2) {
   label = 15;
  } else {
   $3 = ((($1)) + 4|0); //@line 511 "load_mid.cpp"
   $4 = HEAP32[$3>>2]|0; //@line 511 "load_mid.cpp"
   $5 = ((($e)) + 4|0); //@line 511 "load_mid.cpp"
   $6 = HEAP32[$5>>2]|0; //@line 511 "load_mid.cpp"
   $7 = ($4>>>0)>($6>>>0); //@line 511 "load_mid.cpp"
   if ($7) {
    $8 = ((($tp)) + 4|0); //@line 511 "load_mid.cpp"
    $9 = HEAP32[$8>>2]|0; //@line 511 "load_mid.cpp"
    $10 = ($9|0)==(0|0); //@line 512 "load_mid.cpp"
    if ($10) {
     label = 15;
     break;
    } else {
     $ew$0$ph12 = $9;
    }
   } else {
    $ew$0$ph12 = $1;
   }
   $11 = ((($e)) + 4|0); //@line 512 "load_mid.cpp"
   $12 = ((($tp)) + 27|0); //@line 388 "load_mid.cpp"
   $13 = ((($tp)) + 25|0); //@line 384 "load_mid.cpp"
   $14 = ((($tp)) + 26|0); //@line 385 "load_mid.cpp"
   $15 = ((($tp)) + 20|0); //@line 386 "load_mid.cpp"
   $ep$04 = 0;$ew$03 = $ew$0$ph12;
   while(1) {
    $16 = ((($ew$03)) + 4|0); //@line 512 "load_mid.cpp"
    $17 = HEAP32[$16>>2]|0; //@line 512 "load_mid.cpp"
    $18 = HEAP32[$11>>2]|0; //@line 512 "load_mid.cpp"
    $19 = ($17>>>0)>($18>>>0); //@line 512 "load_mid.cpp"
    if ($19) {
     $ep$04$lcssa = $ep$04;$ew$03$lcssa = $ew$03;
     label = 13;
     break;
    }
    HEAP32[$0>>2] = $ew$03; //@line 514 "load_mid.cpp"
    $20 = ((($ew$03)) + 8|0); //@line 382 "load_mid.cpp"
    $21 = HEAP8[$20>>0]|0; //@line 382 "load_mid.cpp"
    $22 = ($21<<24>>24)==(0); //@line 382 "load_mid.cpp"
    if (!($22)) {
     $23 = ((($ew$03)) + 10|0); //@line 383 "load_mid.cpp"
     $24 = HEAP8[$23>>0]|0; //@line 383 "load_mid.cpp"
     $25 = ($24<<24>>24)==(0); //@line 383 "load_mid.cpp"
     if ($25) {
      $$sink$i = -1;
     } else {
      $26 = ((($ew$03)) + 9|0); //@line 383 "load_mid.cpp"
      $27 = HEAP8[$26>>0]|0; //@line 383 "load_mid.cpp"
      $$sink$i = $27;
     }
     HEAP8[$13>>0] = $$sink$i; //@line 384 "load_mid.cpp"
     $28 = HEAP8[$23>>0]|0; //@line 385 "load_mid.cpp"
     HEAP8[$14>>0] = $28; //@line 385 "load_mid.cpp"
     $29 = HEAP32[$16>>2]|0; //@line 386 "load_mid.cpp"
     HEAP32[$15>>2] = $29; //@line 386 "load_mid.cpp"
    }
    $30 = ((($ew$03)) + 12|0); //@line 388 "load_mid.cpp"
    $31 = HEAP8[$30>>0]|0; //@line 388 "load_mid.cpp"
    $32 = ($31<<24>>24)==(8); //@line 388 "load_mid.cpp"
    if ($32) {
     $33 = ((($ew$03)) + 13|0); //@line 388 "load_mid.cpp"
     $34 = HEAP8[$33>>0]|0; //@line 388 "load_mid.cpp"
     HEAP8[$12>>0] = $34; //@line 388 "load_mid.cpp"
    }
    $35 = HEAP32[$ew$03>>2]|0; //@line 512 "load_mid.cpp"
    $36 = ($35|0)==(0|0); //@line 512 "load_mid.cpp"
    if ($36) {
     $ep$0$lcssa14 = $ew$03;$ew$0$lcssa15 = 0;
     break;
    } else {
     $ep$04$phi = $ew$03;$ew$03 = $35;$ep$04 = $ep$04$phi;
    }
   }
   if ((label|0) == 13) {
    $37 = ($ep$04$lcssa|0)==(0|0); //@line 517 "load_mid.cpp"
    if ($37) {
     label = 15;
     break;
    } else {
     $ep$0$lcssa14 = $ep$04$lcssa;$ew$0$lcssa15 = $ew$03$lcssa;
    }
   }
   HEAP32[$ep$0$lcssa14>>2] = $e; //@line 518 "load_mid.cpp"
   HEAP32[$e>>2] = $ew$0$lcssa15; //@line 519 "load_mid.cpp"
   $42 = $ew$0$lcssa15;
  }
 } while(0);
 if ((label|0) == 15) {
  $38 = ((($tp)) + 4|0); //@line 522 "load_mid.cpp"
  $39 = HEAP32[$38>>2]|0; //@line 522 "load_mid.cpp"
  HEAP32[$e>>2] = $39; //@line 522 "load_mid.cpp"
  HEAP32[$38>>2] = $e; //@line 523 "load_mid.cpp"
  $40 = $39;
  $42 = $40;
 }
 $41 = ($42|0)==(0|0); //@line 525 "load_mid.cpp"
 if ($41) {
  $43 = ((($tp)) + 8|0); //@line 526 "load_mid.cpp"
  HEAP32[$43>>2] = $e; //@line 526 "load_mid.cpp"
 }
 HEAP32[$0>>2] = $e; //@line 527 "load_mid.cpp"
 $44 = ((($e)) + 8|0); //@line 382 "load_mid.cpp"
 $45 = HEAP8[$44>>0]|0; //@line 382 "load_mid.cpp"
 $46 = ($45<<24>>24)==(0); //@line 382 "load_mid.cpp"
 if (!($46)) {
  $47 = ((($e)) + 10|0); //@line 383 "load_mid.cpp"
  $48 = HEAP8[$47>>0]|0; //@line 383 "load_mid.cpp"
  $49 = ($48<<24>>24)==(0); //@line 383 "load_mid.cpp"
  if ($49) {
   $$sink$i1 = -1;
  } else {
   $50 = ((($e)) + 9|0); //@line 383 "load_mid.cpp"
   $51 = HEAP8[$50>>0]|0; //@line 383 "load_mid.cpp"
   $$sink$i1 = $51;
  }
  $52 = ((($tp)) + 25|0); //@line 384 "load_mid.cpp"
  HEAP8[$52>>0] = $$sink$i1; //@line 384 "load_mid.cpp"
  $53 = HEAP8[$47>>0]|0; //@line 385 "load_mid.cpp"
  $54 = ((($tp)) + 26|0); //@line 385 "load_mid.cpp"
  HEAP8[$54>>0] = $53; //@line 385 "load_mid.cpp"
  $55 = ((($e)) + 4|0); //@line 386 "load_mid.cpp"
  $56 = HEAP32[$55>>2]|0; //@line 386 "load_mid.cpp"
  $57 = ((($tp)) + 20|0); //@line 386 "load_mid.cpp"
  HEAP32[$57>>2] = $56; //@line 386 "load_mid.cpp"
 }
 $58 = ((($e)) + 12|0); //@line 388 "load_mid.cpp"
 $59 = HEAP8[$58>>0]|0; //@line 388 "load_mid.cpp"
 $60 = ($59<<24>>24)==(8); //@line 388 "load_mid.cpp"
 if (!($60)) {
  return; //@line 529 "load_mid.cpp"
 }
 $61 = ((($e)) + 13|0); //@line 388 "load_mid.cpp"
 $62 = HEAP8[$61>>0]|0; //@line 388 "load_mid.cpp"
 $63 = ((($tp)) + 27|0); //@line 388 "load_mid.cpp"
 HEAP8[$63>>0] = $62; //@line 388 "load_mid.cpp"
 return; //@line 529 "load_mid.cpp"
}
function __ZL16mid_locate_trackP10_MIDHANDLEii($h,$mch,$pos) {
 $h = $h|0;
 $mch = $mch|0;
 $pos = $pos|0;
 var $$0 = 0, $$sink$i$i = 0, $$tmin$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $bal$0$lcssa516675 = 0, $bal$035 = 0, $bal$1 = 0, $bal$1$lcssa = 0, $e$0$ph4$i = 0, $e$01$i = 0, $instrno$0$lcssa545979 = 0, $instrno$032 = 0, $instrno$1 = 0, $instrno$1$lcssa = 0, $numtracks$036 = 0, $numtracks$036$lcssa89 = 0, $or$cond = 0, $polyphony$033 = 0, $polyphony$1 = 0, $polyphony$1$lcssa = 0, $prev$0$lcssa555681 = 0, $tmin$025 = 0, $tmin$1 = 0, $tmin$2$ = 0;
 var $tmin$218 = 0, $tmin$3 = 0, $tmin$4$ = 0, $tmin$414 = 0, $tr$0 = 0, $tr$0$trunused$0 = 0, $tr$030 = 0, $tr$037 = 0, $tr$037$lcssa88 = 0, $tr$1 = 0, $tr$1$trunused$3 = 0, $tr$123 = 0, $tr$126 = 0, $tr$2 = 0, $tr$216 = 0, $tr$219 = 0, $tr$3 = 0, $tr$315 = 0, $trunused$031 = 0, $trunused$1 = 0;
 var $trunused$2 = 0, $trunused$2$lcssa = 0, $trunused$324 = 0, $trunused$4 = 0, $trunused$4$lcssa = 0, $trunused$6$tr$2 = 0, $trunused$617 = 0, $trunused$7 = 0, $trunused$7$lcssa = 0, $trunused$8$tr$3 = 0, $trunused$8$tr$3$lcssa = 0, $trunused$813 = 0, $vol$0$lcssa526377 = 0, $vol$034 = 0, $vol$1 = 0, $vol$2 = 0, $vol$2$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($h)) + 28|0); //@line 430 "load_mid.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 430 "load_mid.cpp"
 $2 = ((($h)) + 4|0); //@line 432 "load_mid.cpp"
 $tr$030 = HEAP32[$2>>2]|0; //@line 432 "load_mid.cpp"
 $3 = ($tr$030|0)==(0|0); //@line 432 "load_mid.cpp"
 if ($3) {
  $bal$0$lcssa516675 = 0;$instrno$0$lcssa545979 = 1;$prev$0$lcssa555681 = 0;$vol$0$lcssa526377 = 0;
 } else {
  $4 = ((($h)) + 12|0); //@line 433 "load_mid.cpp"
  $bal$035 = 0;$instrno$032 = 1;$numtracks$036 = 0;$polyphony$033 = 0;$tr$037 = $tr$030;$trunused$031 = 0;$vol$034 = 0;
  while(1) {
   $5 = HEAP32[$4>>2]|0; //@line 433 "load_mid.cpp"
   $6 = ((($tr$037)) + 12|0); //@line 394 "load_mid.cpp"
   $7 = HEAP32[$6>>2]|0; //@line 394 "load_mid.cpp"
   $8 = ($7|0)==(0|0); //@line 395 "load_mid.cpp"
   L5: do {
    if (!($8)) {
     $9 = ((($7)) + 4|0); //@line 395 "load_mid.cpp"
     $10 = HEAP32[$9>>2]|0; //@line 395 "load_mid.cpp"
     $11 = ($10>>>0)>($5>>>0); //@line 395 "load_mid.cpp"
     if ($11) {
      $12 = ((($tr$037)) + 4|0); //@line 395 "load_mid.cpp"
      $13 = HEAP32[$12>>2]|0; //@line 395 "load_mid.cpp"
      $14 = ($13|0)==(0|0); //@line 396 "load_mid.cpp"
      if ($14) {
       break;
      } else {
       $e$0$ph4$i = $13;
      }
     } else {
      $e$0$ph4$i = $7;
     }
     $15 = ((($tr$037)) + 27|0); //@line 388 "load_mid.cpp"
     $16 = ((($tr$037)) + 25|0); //@line 384 "load_mid.cpp"
     $17 = ((($tr$037)) + 26|0); //@line 385 "load_mid.cpp"
     $18 = ((($tr$037)) + 20|0); //@line 386 "load_mid.cpp"
     $e$01$i = $e$0$ph4$i;
     while(1) {
      $19 = ((($e$01$i)) + 4|0); //@line 396 "load_mid.cpp"
      $20 = HEAP32[$19>>2]|0; //@line 396 "load_mid.cpp"
      $21 = ($20>>>0)>($5>>>0); //@line 396 "load_mid.cpp"
      if ($21) {
       break L5;
      }
      HEAP32[$6>>2] = $e$01$i; //@line 397 "load_mid.cpp"
      $22 = ((($e$01$i)) + 8|0); //@line 382 "load_mid.cpp"
      $23 = HEAP8[$22>>0]|0; //@line 382 "load_mid.cpp"
      $24 = ($23<<24>>24)==(0); //@line 382 "load_mid.cpp"
      if (!($24)) {
       $25 = ((($e$01$i)) + 10|0); //@line 383 "load_mid.cpp"
       $26 = HEAP8[$25>>0]|0; //@line 383 "load_mid.cpp"
       $27 = ($26<<24>>24)==(0); //@line 383 "load_mid.cpp"
       if ($27) {
        $$sink$i$i = -1;
       } else {
        $28 = ((($e$01$i)) + 9|0); //@line 383 "load_mid.cpp"
        $29 = HEAP8[$28>>0]|0; //@line 383 "load_mid.cpp"
        $$sink$i$i = $29;
       }
       HEAP8[$16>>0] = $$sink$i$i; //@line 384 "load_mid.cpp"
       $30 = HEAP8[$25>>0]|0; //@line 385 "load_mid.cpp"
       HEAP8[$17>>0] = $30; //@line 385 "load_mid.cpp"
       $31 = HEAP32[$19>>2]|0; //@line 386 "load_mid.cpp"
       HEAP32[$18>>2] = $31; //@line 386 "load_mid.cpp"
      }
      $32 = ((($e$01$i)) + 12|0); //@line 388 "load_mid.cpp"
      $33 = HEAP8[$32>>0]|0; //@line 388 "load_mid.cpp"
      $34 = ($33<<24>>24)==(8); //@line 388 "load_mid.cpp"
      if ($34) {
       $35 = ((($e$01$i)) + 13|0); //@line 388 "load_mid.cpp"
       $36 = HEAP8[$35>>0]|0; //@line 388 "load_mid.cpp"
       HEAP8[$15>>0] = $36; //@line 388 "load_mid.cpp"
      }
      $37 = HEAP32[$e$01$i>>2]|0; //@line 396 "load_mid.cpp"
      $38 = ($37|0)==(0|0); //@line 396 "load_mid.cpp"
      if ($38) {
       break;
      } else {
       $e$01$i = $37;
      }
     }
    }
   } while(0);
   $39 = ((($tr$037)) + 24|0); //@line 434 "load_mid.cpp"
   $40 = HEAP8[$39>>0]|0; //@line 434 "load_mid.cpp"
   $41 = $40&255; //@line 434 "load_mid.cpp"
   $42 = ($41|0)==($mch|0); //@line 434 "load_mid.cpp"
   if ($42) {
    $43 = ((($tr$037)) + 25|0); //@line 435 "load_mid.cpp"
    $44 = HEAP8[$43>>0]|0; //@line 435 "load_mid.cpp"
    $45 = $44&255; //@line 435 "load_mid.cpp"
    $46 = ($45|0)==($pos|0); //@line 435 "load_mid.cpp"
    if ($46) {
     $$0 = $tr$037;
     label = 49;
     break;
    }
    $47 = ($44<<24>>24)==(-1); //@line 437 "load_mid.cpp"
    if ($47) {
     $48 = HEAP32[$4>>2]|0; //@line 439 "load_mid.cpp"
     $49 = ((($tr$037)) + 20|0); //@line 439 "load_mid.cpp"
     $50 = HEAP32[$49>>2]|0; //@line 439 "load_mid.cpp"
     $51 = (($50) + ($1))|0; //@line 439 "load_mid.cpp"
     $52 = ($48>>>0)>($51>>>0); //@line 439 "load_mid.cpp"
     $tr$0$trunused$0 = $52 ? $tr$037 : $trunused$031; //@line 439 "load_mid.cpp"
     $trunused$1 = $tr$0$trunused$0;$vol$1 = $vol$034;
    } else {
     $53 = ((($tr$037)) + 26|0); //@line 441 "load_mid.cpp"
     $54 = HEAP8[$53>>0]|0; //@line 441 "load_mid.cpp"
     $55 = $54&255; //@line 441 "load_mid.cpp"
     $trunused$1 = $trunused$031;$vol$1 = $55;
    }
    $56 = ((($tr$037)) + 27|0); //@line 442 "load_mid.cpp"
    $57 = HEAP8[$56>>0]|0; //@line 442 "load_mid.cpp"
    $58 = $57&255; //@line 442 "load_mid.cpp"
    $59 = ((($tr$037)) + 16|0); //@line 443 "load_mid.cpp"
    $60 = HEAP32[$59>>2]|0; //@line 443 "load_mid.cpp"
    $61 = (($polyphony$033) + 1)|0; //@line 444 "load_mid.cpp"
    $bal$1 = $60;$instrno$1 = $58;$polyphony$1 = $61;$trunused$2 = $trunused$1;$vol$2 = $vol$1;
   } else {
    $bal$1 = $bal$035;$instrno$1 = $instrno$032;$polyphony$1 = $polyphony$033;$trunused$2 = $trunused$031;$vol$2 = $vol$034;
   }
   $62 = (($numtracks$036) + 1)|0; //@line 446 "load_mid.cpp"
   $tr$0 = HEAP32[$tr$037>>2]|0; //@line 432 "load_mid.cpp"
   $63 = ($tr$0|0)==(0|0); //@line 432 "load_mid.cpp"
   if ($63) {
    $bal$1$lcssa = $bal$1;$instrno$1$lcssa = $instrno$1;$numtracks$036$lcssa89 = $numtracks$036;$polyphony$1$lcssa = $polyphony$1;$tr$037$lcssa88 = $tr$037;$trunused$2$lcssa = $trunused$2;$vol$2$lcssa = $vol$2;
    break;
   } else {
    $bal$035 = $bal$1;$instrno$032 = $instrno$1;$numtracks$036 = $62;$polyphony$033 = $polyphony$1;$tr$037 = $tr$0;$trunused$031 = $trunused$2;$vol$034 = $vol$2;
   }
  }
  if ((label|0) == 49) {
   return ($$0|0); //@line 504 "load_mid.cpp"
  }
  $64 = ($trunused$2$lcssa|0)==(0|0); //@line 449 "load_mid.cpp"
  if (!($64)) {
   $65 = $pos&255; //@line 450 "load_mid.cpp"
   $66 = ((($trunused$2$lcssa)) + 25|0); //@line 450 "load_mid.cpp"
   HEAP8[$66>>0] = $65; //@line 450 "load_mid.cpp"
   $$0 = $trunused$2$lcssa;
   return ($$0|0); //@line 504 "load_mid.cpp"
  }
  $67 = ($polyphony$1$lcssa|0)>(16); //@line 453 "load_mid.cpp"
  if ($67) {
   label = 26;
  } else {
   $68 = ($polyphony$1$lcssa|0)>(0); //@line 453 "load_mid.cpp"
   $69 = ($numtracks$036$lcssa89|0)>(57); //@line 453 "load_mid.cpp"
   $or$cond = $68 & $69; //@line 453 "load_mid.cpp"
   if ($or$cond) {
    label = 26;
   }
  }
  if ((label|0) == 26) {
   $tr$123 = HEAP32[$2>>2]|0; //@line 454 "load_mid.cpp"
   $70 = ($tr$123|0)==(0|0); //@line 454 "load_mid.cpp"
   if (!($70)) {
    $71 = ((($h)) + 12|0); //@line 457 "load_mid.cpp"
    $tmin$025 = $1;$tr$126 = $tr$123;$trunused$324 = 0;
    while(1) {
     $72 = ((($tr$126)) + 24|0); //@line 455 "load_mid.cpp"
     $73 = HEAP8[$72>>0]|0; //@line 455 "load_mid.cpp"
     $74 = $73&255; //@line 455 "load_mid.cpp"
     $75 = ($74|0)==($mch|0); //@line 455 "load_mid.cpp"
     if ($75) {
      $76 = ((($tr$126)) + 12|0); //@line 456 "load_mid.cpp"
      $77 = HEAP32[$76>>2]|0; //@line 456 "load_mid.cpp"
      $78 = HEAP32[$71>>2]|0; //@line 457 "load_mid.cpp"
      $79 = ((($77)) + 4|0); //@line 457 "load_mid.cpp"
      $80 = HEAP32[$79>>2]|0; //@line 457 "load_mid.cpp"
      $81 = (($80) + ($tmin$025))|0; //@line 457 "load_mid.cpp"
      $82 = ($78>>>0)>($81>>>0); //@line 457 "load_mid.cpp"
      $83 = (($78) - ($80))|0; //@line 458 "load_mid.cpp"
      $tr$1$trunused$3 = $82 ? $tr$126 : $trunused$324; //@line 457 "load_mid.cpp"
      $$tmin$0 = $82 ? $83 : $tmin$025; //@line 457 "load_mid.cpp"
      $tmin$1 = $$tmin$0;$trunused$4 = $tr$1$trunused$3;
     } else {
      $tmin$1 = $tmin$025;$trunused$4 = $trunused$324;
     }
     $tr$1 = HEAP32[$tr$126>>2]|0; //@line 454 "load_mid.cpp"
     $84 = ($tr$1|0)==(0|0); //@line 454 "load_mid.cpp"
     if ($84) {
      $trunused$4$lcssa = $trunused$4;
      break;
     } else {
      $tmin$025 = $tmin$1;$tr$126 = $tr$1;$trunused$324 = $trunused$4;
     }
    }
    $85 = ($trunused$4$lcssa|0)==(0|0); //@line 463 "load_mid.cpp"
    if (!($85)) {
     $86 = $pos&255; //@line 464 "load_mid.cpp"
     $87 = ((($trunused$4$lcssa)) + 25|0); //@line 464 "load_mid.cpp"
     HEAP8[$87>>0] = $86; //@line 464 "load_mid.cpp"
     $$0 = $trunused$4$lcssa;
     return ($$0|0); //@line 504 "load_mid.cpp"
    }
   }
  }
  $88 = ($numtracks$036$lcssa89|0)>(57); //@line 468 "load_mid.cpp"
  if ($88) {
   $tr$216 = HEAP32[$2>>2]|0; //@line 470 "load_mid.cpp"
   $89 = ($tr$216|0)==(0|0); //@line 470 "load_mid.cpp"
   if ($89) {
    $bal$0$lcssa516675 = $bal$1$lcssa;$instrno$0$lcssa545979 = $instrno$1$lcssa;$prev$0$lcssa555681 = $tr$037$lcssa88;$vol$0$lcssa526377 = $vol$2$lcssa;
   } else {
    $90 = ((($h)) + 12|0); //@line 473 "load_mid.cpp"
    $tmin$218 = 0;$tr$219 = $tr$216;$trunused$617 = 0;
    while(1) {
     $91 = ((($tr$219)) + 24|0); //@line 471 "load_mid.cpp"
     $92 = HEAP8[$91>>0]|0; //@line 471 "load_mid.cpp"
     $93 = $92&255; //@line 471 "load_mid.cpp"
     $94 = ($93|0)==($mch|0); //@line 471 "load_mid.cpp"
     if ($94) {
      $95 = ((($tr$219)) + 12|0); //@line 472 "load_mid.cpp"
      $96 = HEAP32[$95>>2]|0; //@line 472 "load_mid.cpp"
      $97 = HEAP32[$90>>2]|0; //@line 473 "load_mid.cpp"
      $98 = ((($96)) + 4|0); //@line 473 "load_mid.cpp"
      $99 = HEAP32[$98>>2]|0; //@line 473 "load_mid.cpp"
      $100 = (($99) + ($tmin$218))|0; //@line 473 "load_mid.cpp"
      $101 = ($97>>>0)<($100>>>0); //@line 473 "load_mid.cpp"
      $102 = (($97) - ($99))|0; //@line 474 "load_mid.cpp"
      $trunused$6$tr$2 = $101 ? $trunused$617 : $tr$219; //@line 473 "load_mid.cpp"
      $tmin$2$ = $101 ? $tmin$218 : $102; //@line 473 "load_mid.cpp"
      $tmin$3 = $tmin$2$;$trunused$7 = $trunused$6$tr$2;
     } else {
      $tmin$3 = $tmin$218;$trunused$7 = $trunused$617;
     }
     $tr$2 = HEAP32[$tr$219>>2]|0; //@line 470 "load_mid.cpp"
     $103 = ($tr$2|0)==(0|0); //@line 470 "load_mid.cpp"
     if ($103) {
      $trunused$7$lcssa = $trunused$7;
      break;
     } else {
      $tmin$218 = $tmin$3;$tr$219 = $tr$2;$trunused$617 = $trunused$7;
     }
    }
    $104 = ($trunused$7$lcssa|0)==(0|0); //@line 479 "load_mid.cpp"
    if (!($104)) {
     $107 = $pos&255; //@line 480 "load_mid.cpp"
     $108 = ((($trunused$7$lcssa)) + 25|0); //@line 480 "load_mid.cpp"
     HEAP8[$108>>0] = $107; //@line 480 "load_mid.cpp"
     $$0 = $trunused$7$lcssa;
     return ($$0|0); //@line 504 "load_mid.cpp"
    }
    if ($89) {
     $bal$0$lcssa516675 = $bal$1$lcssa;$instrno$0$lcssa545979 = $instrno$1$lcssa;$prev$0$lcssa555681 = $tr$037$lcssa88;$vol$0$lcssa526377 = $vol$2$lcssa;
    } else {
     $105 = ((($h)) + 12|0); //@line 486 "load_mid.cpp"
     $106 = HEAP32[$105>>2]|0; //@line 486 "load_mid.cpp"
     $tmin$414 = 0;$tr$315 = $tr$216;$trunused$813 = 0;
     while(1) {
      $109 = ((($tr$315)) + 12|0); //@line 485 "load_mid.cpp"
      $110 = HEAP32[$109>>2]|0; //@line 485 "load_mid.cpp"
      $111 = ((($110)) + 4|0); //@line 486 "load_mid.cpp"
      $112 = HEAP32[$111>>2]|0; //@line 486 "load_mid.cpp"
      $113 = (($112) + ($tmin$414))|0; //@line 486 "load_mid.cpp"
      $114 = ($106>>>0)<($113>>>0); //@line 486 "load_mid.cpp"
      $115 = (($106) - ($112))|0; //@line 487 "load_mid.cpp"
      $trunused$8$tr$3 = $114 ? $trunused$813 : $tr$315; //@line 486 "load_mid.cpp"
      $tmin$4$ = $114 ? $tmin$414 : $115; //@line 486 "load_mid.cpp"
      $tr$3 = HEAP32[$tr$315>>2]|0; //@line 484 "load_mid.cpp"
      $116 = ($tr$3|0)==(0|0); //@line 484 "load_mid.cpp"
      if ($116) {
       $trunused$8$tr$3$lcssa = $trunused$8$tr$3;
       break;
      } else {
       $tmin$414 = $tmin$4$;$tr$315 = $tr$3;$trunused$813 = $trunused$8$tr$3;
      }
     }
     $117 = ($trunused$8$tr$3$lcssa|0)==(0|0); //@line 491 "load_mid.cpp"
     if ($117) {
      $bal$0$lcssa516675 = $bal$1$lcssa;$instrno$0$lcssa545979 = $instrno$1$lcssa;$prev$0$lcssa555681 = $tr$037$lcssa88;$vol$0$lcssa526377 = $vol$2$lcssa;
     } else {
      $118 = $pos&255; //@line 492 "load_mid.cpp"
      $119 = ((($trunused$8$tr$3$lcssa)) + 25|0); //@line 492 "load_mid.cpp"
      HEAP8[$119>>0] = $118; //@line 492 "load_mid.cpp"
      $120 = $mch&255; //@line 493 "load_mid.cpp"
      $121 = ((($trunused$8$tr$3$lcssa)) + 24|0); //@line 493 "load_mid.cpp"
      HEAP8[$121>>0] = $120; //@line 493 "load_mid.cpp"
      $$0 = $trunused$8$tr$3$lcssa;
      return ($$0|0); //@line 504 "load_mid.cpp"
     }
    }
   }
  } else {
   $bal$0$lcssa516675 = $bal$1$lcssa;$instrno$0$lcssa545979 = $instrno$1$lcssa;$prev$0$lcssa555681 = $tr$037$lcssa88;$vol$0$lcssa526377 = $vol$2$lcssa;
  }
 }
 $122 = (_calloc(1,28)|0); //@line 315 "load_mid.cpp"
 HEAP32[$122>>2] = 0; //@line 316 "load_mid.cpp"
 $123 = $pos&255; //@line 317 "load_mid.cpp"
 $124 = ((($122)) + 25|0); //@line 317 "load_mid.cpp"
 HEAP8[$124>>0] = $123; //@line 317 "load_mid.cpp"
 $125 = ((($122)) + 27|0); //@line 318 "load_mid.cpp"
 HEAP8[$125>>0] = 1; //@line 318 "load_mid.cpp"
 $126 = $mch&255; //@line 319 "load_mid.cpp"
 $127 = ((($122)) + 24|0); //@line 319 "load_mid.cpp"
 HEAP8[$127>>0] = $126; //@line 319 "load_mid.cpp"
 $128 = ((($122)) + 4|0); //@line 320 "load_mid.cpp"
 HEAP32[$128>>2] = 0; //@line 320 "load_mid.cpp"
 $129 = ((($122)) + 8|0); //@line 321 "load_mid.cpp"
 HEAP32[$129>>2] = 0; //@line 321 "load_mid.cpp"
 $130 = ((($122)) + 12|0); //@line 322 "load_mid.cpp"
 HEAP32[$130>>2] = 0; //@line 322 "load_mid.cpp"
 $131 = ((($122)) + 20|0); //@line 323 "load_mid.cpp"
 HEAP32[$131>>2] = 0; //@line 323 "load_mid.cpp"
 $132 = ((($122)) + 26|0); //@line 324 "load_mid.cpp"
 $133 = ((($122)) + 16|0); //@line 325 "load_mid.cpp"
 $134 = $instrno$0$lcssa545979&255; //@line 498 "load_mid.cpp"
 HEAP8[$125>>0] = $134; //@line 498 "load_mid.cpp"
 $135 = $vol$0$lcssa526377&255; //@line 499 "load_mid.cpp"
 HEAP8[$132>>0] = $135; //@line 499 "load_mid.cpp"
 HEAP32[$133>>2] = $bal$0$lcssa516675; //@line 500 "load_mid.cpp"
 $136 = ($prev$0$lcssa555681|0)==(0|0); //@line 501 "load_mid.cpp"
 if ($136) {
  HEAP32[$2>>2] = $122; //@line 502 "load_mid.cpp"
  $$0 = $122;
  return ($$0|0); //@line 504 "load_mid.cpp"
 } else {
  HEAP32[$prev$0$lcssa555681>>2] = $122; //@line 501 "load_mid.cpp"
  $$0 = $122;
  return ($$0|0); //@line 504 "load_mid.cpp"
 }
 return (0)|0;
}
function __ZL16MID_ReadPatternsPP11_MODCOMMANDPtP10_MIDHANDLEii($pattern,$psize,$h,$numpat,$channels) {
 $pattern = $pattern|0;
 $psize = $psize|0;
 $h = $h|0;
 $numpat = $numpat|0;
 $channels = $channels|0;
 var $$ = 0, $$0 = 0, $$0$idx$i = 0, $$0$idx$i30 = 0, $$0$idx$val$i = 0, $$0$idx$val$i31 = 0, $$01$i = 0, $$01$i$lcssa = 0, $$01$i29 = 0, $$01$i29$lcssa = 0, $$01$i42 = 0, $$01$i42$lcssa = 0, $$01$i45 = 0, $$01$i45$lcssa = 0, $$02$i = 0, $$02$i$lcssa = 0, $$02$i23 = 0, $$02$i23$lcssa = 0, $$02$i36 = 0, $$02$i36$lcssa = 0;
 var $$numpat = 0, $$phi$trans$insert = 0, $$pr$pre = 0, $$pr104 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0;
 var $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0;
 var $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0;
 var $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $ch$082 = 0, $div = 0, $div8 = 0, $e$0$6$val$off$i$i = 0, $e$0$6$val$off$i$i32 = 0, $e$054159 = 0, $e$160164 = 0, $e$2$ef$0 = 0, $e$262$lcssa = 0, $e$262165 = 0, $e$3$ef$0 = 0, $e$371170 = 0, $ef$0$lcssa = 0, $ef$072169 = 0, $ef$2 = 0, $el$0$lcssa = 0, $el$073168 = 0, $el$1 = 0, $h$idx12 = 0;
 var $h$idx12$val = 0, $h$idx13$val = 0, $h$idx14$val = 0, $h$idx15$val = 0, $h$idx16$val = 0, $h$idx17$val = 0, $h$idx18$val = 0, $h$idx19$val = 0, $h$idx20$val = 0, $i$0$lcssa = 0, $i$070171 = 0, $or$cond = 0, $pat$090 = 0, $patbrk$0$lcssa = 0, $patbrk$055158 = 0, $patbrk$1 = 0, $patbrk$2$lcssa = 0, $patbrk$283 = 0, $patbrk$3 = 0, $row$089 = 0;
 var $t$0 = 0, $t$093 = 0, $t$094 = 0, $t$1 = 0, $t$185 = 0, $tempo$0$lcssa = 0, $tempo$056157 = 0, $tempo$1 = 0, $tempo$2$lcssa = 0, $tempo$284 = 0, $tempo$3 = 0, $vol$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($numpat|0)>(240); //@line 1096 "load_mid.cpp"
 $$numpat = $0 ? 240 : $numpat; //@line 1096 "load_mid.cpp"
 $1 = ((($h)) + 4|0); //@line 1099 "load_mid.cpp"
 $t$093 = HEAP32[$1>>2]|0; //@line 1099 "load_mid.cpp"
 $2 = ($t$093|0)==(0|0); //@line 1099 "load_mid.cpp"
 if (!($2)) {
  $t$094 = $t$093;
  while(1) {
   $6 = ((($t$094)) + 4|0); //@line 1099 "load_mid.cpp"
   $7 = HEAP32[$6>>2]|0; //@line 1099 "load_mid.cpp"
   $8 = ((($t$094)) + 12|0); //@line 1099 "load_mid.cpp"
   HEAP32[$8>>2] = $7; //@line 1099 "load_mid.cpp"
   $t$0 = HEAP32[$t$094>>2]|0; //@line 1099 "load_mid.cpp"
   $9 = ($t$0|0)==(0|0); //@line 1099 "load_mid.cpp"
   if ($9) {
    break;
   } else {
    $t$094 = $t$0;
   }
  }
 }
 $3 = ($$numpat|0)>(0); //@line 1100 "load_mid.cpp"
 if (!($3)) {
  $$0 = 0;
  return ($$0|0); //@line 1246 "load_mid.cpp"
 }
 $4 = ((($h)) + 24|0); //@line 1105 "load_mid.cpp"
 $h$idx12 = ((($h)) + 44|0);
 $5 = ((($h)) + 28|0); //@line 1106 "load_mid.cpp"
 $pat$090 = 0;
 L8: while(1) {
  $10 = (__ZN10CSoundFile15AllocatePatternEjj(64,$channels)|0); //@line 1101 "load_mid.cpp"
  $11 = (($pattern) + ($pat$090<<2)|0); //@line 1101 "load_mid.cpp"
  HEAP32[$11>>2] = $10; //@line 1101 "load_mid.cpp"
  $12 = ($10|0)==(0|0); //@line 1102 "load_mid.cpp"
  if ($12) {
   $$0 = 0;
   label = 82;
   break;
  }
  $13 = (($psize) + ($pat$090<<1)|0); //@line 1103 "load_mid.cpp"
  HEAP16[$13>>1] = 64; //@line 1103 "load_mid.cpp"
  $14 = $pat$090 << 6; //@line 1105 "load_mid.cpp"
  $row$089 = 0;
  while(1) {
   $17 = (($row$089) + ($14))|0; //@line 1105 "load_mid.cpp"
   $18 = HEAP32[$4>>2]|0; //@line 1105 "load_mid.cpp"
   $19 = Math_imul($18, $17)|0; //@line 1105 "load_mid.cpp"
   $h$idx12$val = HEAP32[$h$idx12>>2]|0;
   $20 = Math_imul($19, $h$idx12$val)|0; //@line 256 "load_mid.cpp"
   $21 = $20 >>> 4; //@line 256 "load_mid.cpp"
   $22 = (($21>>>0) / ($18>>>0))&-1; //@line 256 "load_mid.cpp"
   $23 = HEAP32[$5>>2]|0; //@line 1106 "load_mid.cpp"
   $24 = (($23) + ($22))|0; //@line 1106 "load_mid.cpp"
   $25 = HEAP32[$1>>2]|0; //@line 1110 "load_mid.cpp"
   $26 = ((($25)) + 12|0); //@line 1110 "load_mid.cpp"
   $27 = HEAP32[$26>>2]|0; //@line 1110 "load_mid.cpp"
   $28 = ($27|0)==(0|0); //@line 891 "load_mid.cpp"
   L13: do {
    if ($28) {
     $patbrk$0$lcssa = 0;$tempo$0$lcssa = 0;
    } else {
     $$01$i = $27;
     while(1) {
      $$0$idx$i = ((($$01$i)) + 12|0); //@line 1110 "load_mid.cpp"
      $$0$idx$val$i = HEAP8[$$0$idx$i>>0]|0; //@line 1110 "load_mid.cpp"
      $e$0$6$val$off$i$i = (($$0$idx$val$i) + -3)<<24>>24; //@line 886 "load_mid.cpp"
      $29 = ($e$0$6$val$off$i$i&255)<(2); //@line 886 "load_mid.cpp"
      if ($29) {
       $$01$i$lcssa = $$01$i;
       break;
      }
      $33 = HEAP32[$$01$i>>2]|0; //@line 891 "load_mid.cpp"
      $34 = ($33|0)==(0|0); //@line 891 "load_mid.cpp"
      if ($34) {
       $patbrk$0$lcssa = 0;$tempo$0$lcssa = 0;
       break L13;
      } else {
       $$01$i = $33;
      }
     }
     $30 = ((($$01$i$lcssa)) + 4|0); //@line 1110 "load_mid.cpp"
     $31 = HEAP32[$30>>2]|0; //@line 1110 "load_mid.cpp"
     $32 = ($31>>>0)<($24>>>0); //@line 1110 "load_mid.cpp"
     if ($32) {
      $41 = $31;$e$054159 = $$01$i$lcssa;$patbrk$055158 = 0;$tempo$056157 = 0;
      while(1) {
       $40 = ($41>>>0)<($22>>>0); //@line 1111 "load_mid.cpp"
       L20: do {
        if ($40) {
         $patbrk$1 = $patbrk$055158;$tempo$1 = $tempo$056157;
        } else {
         $42 = ((($e$054159)) + 12|0); //@line 1112 "load_mid.cpp"
         $43 = HEAP8[$42>>0]|0; //@line 1112 "load_mid.cpp"
         $44 = $43&255; //@line 1112 "load_mid.cpp"
         switch ($44|0) {
         case 4:  {
          $45 = ((($e$054159)) + 13|0); //@line 1114 "load_mid.cpp"
          $46 = HEAP8[$45>>0]|0; //@line 1114 "load_mid.cpp"
          $47 = $46&255; //@line 1114 "load_mid.cpp"
          $patbrk$1 = $patbrk$055158;$tempo$1 = $47;
          break L20;
          break;
         }
         case 3:  {
          $patbrk$1 = 1;$tempo$1 = $tempo$056157;
          break L20;
          break;
         }
         default: {
          $patbrk$1 = $patbrk$055158;$tempo$1 = $tempo$056157;
          break L20;
         }
         }
        }
       } while(0);
       $48 = HEAP32[$e$054159>>2]|0; //@line 1110 "load_mid.cpp"
       $49 = ($48|0)==(0|0); //@line 891 "load_mid.cpp"
       if ($49) {
        $patbrk$0$lcssa = $patbrk$1;$tempo$0$lcssa = $tempo$1;
        break L13;
       } else {
        $$01$i29 = $48;
       }
       while(1) {
        $$0$idx$i30 = ((($$01$i29)) + 12|0); //@line 1110 "load_mid.cpp"
        $$0$idx$val$i31 = HEAP8[$$0$idx$i30>>0]|0; //@line 1110 "load_mid.cpp"
        $e$0$6$val$off$i$i32 = (($$0$idx$val$i31) + -3)<<24>>24; //@line 886 "load_mid.cpp"
        $50 = ($e$0$6$val$off$i$i32&255)<(2); //@line 886 "load_mid.cpp"
        if ($50) {
         $$01$i29$lcssa = $$01$i29;
         break;
        }
        $51 = HEAP32[$$01$i29>>2]|0; //@line 891 "load_mid.cpp"
        $52 = ($51|0)==(0|0); //@line 891 "load_mid.cpp"
        if ($52) {
         $patbrk$0$lcssa = $patbrk$1;$tempo$0$lcssa = $tempo$1;
         break L13;
        } else {
         $$01$i29 = $51;
        }
       }
       $35 = ((($$01$i29$lcssa)) + 4|0); //@line 1110 "load_mid.cpp"
       $36 = HEAP32[$35>>2]|0; //@line 1110 "load_mid.cpp"
       $37 = ($36>>>0)<($24>>>0); //@line 1110 "load_mid.cpp"
       if ($37) {
        $41 = $36;$e$054159 = $$01$i29$lcssa;$patbrk$055158 = $patbrk$1;$tempo$056157 = $tempo$1;
       } else {
        $patbrk$0$lcssa = $patbrk$1;$tempo$0$lcssa = $tempo$1;
        break;
       }
      }
     } else {
      $patbrk$0$lcssa = 0;$tempo$0$lcssa = 0;
     }
    }
   } while(0);
   $38 = ($25|0)==(0|0); //@line 1122 "load_mid.cpp"
   L31: do {
    if ($38) {
     $patbrk$2$lcssa = $patbrk$0$lcssa;$tempo$2$lcssa = $tempo$0$lcssa;
    } else {
     $39 = Math_imul($row$089, $channels)|0; //@line 1123 "load_mid.cpp"
     $60 = $27;$ch$082 = 0;$patbrk$283 = $patbrk$0$lcssa;$t$185 = $25;$tempo$284 = $tempo$0$lcssa;
     while(1) {
      $53 = (($ch$082) + ($39))|0; //@line 1123 "load_mid.cpp"
      $54 = HEAP32[$11>>2]|0; //@line 1123 "load_mid.cpp"
      $55 = (($54) + (($53*6)|0)|0); //@line 1123 "load_mid.cpp"
      $56 = (((($54) + (($53*6)|0)|0)) + 5|0); //@line 1124 "load_mid.cpp"
      HEAP8[$56>>0] = 0; //@line 1124 "load_mid.cpp"
      $57 = (((($54) + (($53*6)|0)|0)) + 3|0); //@line 1125 "load_mid.cpp"
      HEAP8[$57>>0] = 0; //@line 1125 "load_mid.cpp"
      $58 = ((($t$185)) + 12|0); //@line 1126 "load_mid.cpp"
      $59 = ($60|0)==(0|0); //@line 897 "load_mid.cpp"
      L35: do {
       if ($59) {
        label = 55;
       } else {
        $$01$i42 = $60;
        while(1) {
         $61 = ((($$01$i42)) + 12|0); //@line 897 "load_mid.cpp"
         $62 = HEAP8[$61>>0]|0; //@line 897 "load_mid.cpp"
         $63 = ($62<<24>>24)==(0); //@line 897 "load_mid.cpp"
         if (!($63)) {
          $$01$i42$lcssa = $$01$i42;
          label = 24;
          break;
         }
         $67 = HEAP32[$$01$i42>>2]|0; //@line 897 "load_mid.cpp"
         $68 = ($67|0)==(0|0); //@line 897 "load_mid.cpp"
         if ($68) {
          break;
         } else {
          $$01$i42 = $67;
         }
        }
        L39: do {
         if ((label|0) == 24) {
          label = 0;
          $64 = ((($$01$i42$lcssa)) + 4|0); //@line 1126 "load_mid.cpp"
          $65 = HEAP32[$64>>2]|0; //@line 1126 "load_mid.cpp"
          $66 = ($65>>>0)<($24>>>0); //@line 1126 "load_mid.cpp"
          if ($66) {
           $73 = $65;$e$160164 = $$01$i42$lcssa;
           while(1) {
            $72 = ($73>>>0)<($22>>>0); //@line 1127 "load_mid.cpp"
            L43: do {
             if (!($72)) {
              $74 = ((($e$160164)) + 12|0); //@line 1128 "load_mid.cpp"
              $75 = HEAP8[$74>>0]|0; //@line 1128 "load_mid.cpp"
              $76 = $75&255; //@line 1128 "load_mid.cpp"
              switch ($76|0) {
              case 6:  {
               $77 = ((($e$160164)) + 13|0); //@line 1130 "load_mid.cpp"
               $78 = HEAP8[$77>>0]|0; //@line 1130 "load_mid.cpp"
               HEAP8[$56>>0] = $78; //@line 1130 "load_mid.cpp"
               HEAP8[$57>>0] = 11; //@line 1131 "load_mid.cpp"
               break L43;
               break;
              }
              case 2:  {
               $79 = ((($e$160164)) + 13|0); //@line 1134 "load_mid.cpp"
               $80 = HEAP8[$79>>0]|0; //@line 1134 "load_mid.cpp"
               $81 = $80&255; //@line 1134 "load_mid.cpp"
               $82 = $81 | 16; //@line 1134 "load_mid.cpp"
               $83 = $82&255; //@line 1134 "load_mid.cpp"
               HEAP8[$56>>0] = $83; //@line 1134 "load_mid.cpp"
               HEAP8[$57>>0] = 28; //@line 1135 "load_mid.cpp"
               break L43;
               break;
              }
              case 1:  {
               $84 = ((($e$160164)) + 13|0); //@line 1138 "load_mid.cpp"
               $85 = HEAP8[$84>>0]|0; //@line 1138 "load_mid.cpp"
               $86 = $85&255; //@line 1138 "load_mid.cpp"
               $87 = $86 | 32; //@line 1138 "load_mid.cpp"
               $88 = $87&255; //@line 1138 "load_mid.cpp"
               HEAP8[$56>>0] = $88; //@line 1138 "load_mid.cpp"
               HEAP8[$57>>0] = 28; //@line 1139 "load_mid.cpp"
               break L43;
               break;
              }
              default: {
               break L43;
              }
              }
             }
            } while(0);
            $89 = HEAP32[$e$160164>>2]|0; //@line 1126 "load_mid.cpp"
            $90 = ($89|0)==(0|0); //@line 897 "load_mid.cpp"
            if ($90) {
             break L39;
            } else {
             $$01$i45 = $89;
            }
            while(1) {
             $91 = ((($$01$i45)) + 12|0); //@line 897 "load_mid.cpp"
             $92 = HEAP8[$91>>0]|0; //@line 897 "load_mid.cpp"
             $93 = ($92<<24>>24)==(0); //@line 897 "load_mid.cpp"
             if (!($93)) {
              $$01$i45$lcssa = $$01$i45;
              break;
             }
             $94 = HEAP32[$$01$i45>>2]|0; //@line 897 "load_mid.cpp"
             $95 = ($94|0)==(0|0); //@line 897 "load_mid.cpp"
             if ($95) {
              break L39;
             } else {
              $$01$i45 = $94;
             }
            }
            $69 = ((($$01$i45$lcssa)) + 4|0); //@line 1126 "load_mid.cpp"
            $70 = HEAP32[$69>>2]|0; //@line 1126 "load_mid.cpp"
            $71 = ($70>>>0)<($24>>>0); //@line 1126 "load_mid.cpp"
            if ($71) {
             $73 = $70;$e$160164 = $$01$i45$lcssa;
            } else {
             break;
            }
           }
          }
         }
        } while(0);
        if ($59) {
         label = 55;
        } else {
         $$02$i36 = $60;
         while(1) {
          $96 = ((($$02$i36)) + 8|0); //@line 906 "load_mid.cpp"
          $97 = HEAP8[$96>>0]|0; //@line 906 "load_mid.cpp"
          $98 = ($97<<24>>24)==(0); //@line 906 "load_mid.cpp"
          if (!($98)) {
           $99 = ((($$02$i36)) + 10|0); //@line 907 "load_mid.cpp"
           $100 = HEAP8[$99>>0]|0; //@line 907 "load_mid.cpp"
           $101 = ($100<<24>>24)==(0); //@line 907 "load_mid.cpp"
           if (!($101)) {
            $$02$i36$lcssa = $$02$i36;
            break;
           }
           $105 = ((($$02$i36)) + 11|0); //@line 908 "load_mid.cpp"
           $106 = HEAP8[$105>>0]|0; //@line 908 "load_mid.cpp"
           $107 = $106&255; //@line 908 "load_mid.cpp"
           $108 = (_pat_smplooped($107)|0); //@line 908 "load_mid.cpp"
           $109 = ($108|0)==(0); //@line 914 "load_mid.cpp"
           if (!($109)) {
            $$02$i36$lcssa = $$02$i36;
            break;
           }
          }
          $110 = HEAP32[$$02$i36>>2]|0; //@line 914 "load_mid.cpp"
          $111 = ($110|0)==(0|0); //@line 914 "load_mid.cpp"
          if ($111) {
           label = 55;
           break L35;
          } else {
           $$02$i36 = $110;
          }
         }
         $102 = ((($$02$i36$lcssa)) + 4|0); //@line 1144 "load_mid.cpp"
         $103 = HEAP32[$102>>2]|0; //@line 1144 "load_mid.cpp"
         $104 = ($103>>>0)<($22>>>0); //@line 1144 "load_mid.cpp"
         if ($104) {
          $e$262165 = $$02$i36$lcssa;
          while(1) {
           HEAP32[$58>>2] = $e$262165; //@line 1145 "load_mid.cpp"
           $118 = HEAP32[$e$262165>>2]|0; //@line 1144 "load_mid.cpp"
           $119 = ($118|0)==(0|0); //@line 914 "load_mid.cpp"
           if ($119) {
            label = 55;
            break L35;
           } else {
            $$02$i23 = $118;
           }
           while(1) {
            $120 = ((($$02$i23)) + 8|0); //@line 906 "load_mid.cpp"
            $121 = HEAP8[$120>>0]|0; //@line 906 "load_mid.cpp"
            $122 = ($121<<24>>24)==(0); //@line 906 "load_mid.cpp"
            if (!($122)) {
             $123 = ((($$02$i23)) + 10|0); //@line 907 "load_mid.cpp"
             $124 = HEAP8[$123>>0]|0; //@line 907 "load_mid.cpp"
             $125 = ($124<<24>>24)==(0); //@line 907 "load_mid.cpp"
             if (!($125)) {
              $$02$i23$lcssa = $$02$i23;
              break;
             }
             $126 = ((($$02$i23)) + 11|0); //@line 908 "load_mid.cpp"
             $127 = HEAP8[$126>>0]|0; //@line 908 "load_mid.cpp"
             $128 = $127&255; //@line 908 "load_mid.cpp"
             $129 = (_pat_smplooped($128)|0); //@line 908 "load_mid.cpp"
             $130 = ($129|0)==(0); //@line 914 "load_mid.cpp"
             if (!($130)) {
              $$02$i23$lcssa = $$02$i23;
              break;
             }
            }
            $131 = HEAP32[$$02$i23>>2]|0; //@line 914 "load_mid.cpp"
            $132 = ($131|0)==(0|0); //@line 914 "load_mid.cpp"
            if ($132) {
             label = 55;
             break L35;
            } else {
             $$02$i23 = $131;
            }
           }
           $112 = ((($$02$i23$lcssa)) + 4|0); //@line 1144 "load_mid.cpp"
           $113 = HEAP32[$112>>2]|0; //@line 1144 "load_mid.cpp"
           $114 = ($113>>>0)<($22>>>0); //@line 1144 "load_mid.cpp"
           if ($114) {
            $e$262165 = $$02$i23$lcssa;
           } else {
            $e$262$lcssa = $$02$i23$lcssa;
            break;
           }
          }
         } else {
          $e$262$lcssa = $$02$i36$lcssa;
         }
         $115 = ((($e$262$lcssa)) + 4|0); //@line 1150 "load_mid.cpp"
         $116 = HEAP32[$115>>2]|0; //@line 1150 "load_mid.cpp"
         $117 = ($116>>>0)<($24>>>0); //@line 1150 "load_mid.cpp"
         L73: do {
          if ($117) {
           $e$371170 = $e$262$lcssa;$ef$072169 = 0;$el$073168 = $e$262$lcssa;$i$070171 = 0;
           while(1) {
            HEAP32[$58>>2] = $e$371170; //@line 1151 "load_mid.cpp"
            $136 = (($i$070171) + 1)|0; //@line 1152 "load_mid.cpp"
            $137 = ((($e$371170)) + 10|0); //@line 1153 "load_mid.cpp"
            $138 = HEAP8[$137>>0]|0; //@line 1153 "load_mid.cpp"
            $139 = ($138<<24>>24)==(0); //@line 1153 "load_mid.cpp"
            $140 = ($ef$072169|0)==(0|0); //@line 1154 "load_mid.cpp"
            $e$3$ef$0 = $140 ? $e$371170 : $ef$072169; //@line 1154 "load_mid.cpp"
            $ef$2 = $139 ? $ef$072169 : $e$3$ef$0; //@line 1153 "load_mid.cpp"
            $el$1 = $139 ? $el$073168 : $e$371170; //@line 1153 "load_mid.cpp"
            $141 = HEAP32[$e$371170>>2]|0; //@line 1150 "load_mid.cpp"
            $142 = ($141|0)==(0|0); //@line 914 "load_mid.cpp"
            if ($142) {
             $ef$0$lcssa = $ef$2;$el$0$lcssa = $el$1;$i$0$lcssa = $136;
             break L73;
            } else {
             $$02$i = $141;
            }
            while(1) {
             $143 = ((($$02$i)) + 8|0); //@line 906 "load_mid.cpp"
             $144 = HEAP8[$143>>0]|0; //@line 906 "load_mid.cpp"
             $145 = ($144<<24>>24)==(0); //@line 906 "load_mid.cpp"
             if (!($145)) {
              $146 = ((($$02$i)) + 10|0); //@line 907 "load_mid.cpp"
              $147 = HEAP8[$146>>0]|0; //@line 907 "load_mid.cpp"
              $148 = ($147<<24>>24)==(0); //@line 907 "load_mid.cpp"
              if (!($148)) {
               $$02$i$lcssa = $$02$i;
               break;
              }
              $149 = ((($$02$i)) + 11|0); //@line 908 "load_mid.cpp"
              $150 = HEAP8[$149>>0]|0; //@line 908 "load_mid.cpp"
              $151 = $150&255; //@line 908 "load_mid.cpp"
              $152 = (_pat_smplooped($151)|0); //@line 908 "load_mid.cpp"
              $153 = ($152|0)==(0); //@line 914 "load_mid.cpp"
              if (!($153)) {
               $$02$i$lcssa = $$02$i;
               break;
              }
             }
             $154 = HEAP32[$$02$i>>2]|0; //@line 914 "load_mid.cpp"
             $155 = ($154|0)==(0|0); //@line 914 "load_mid.cpp"
             if ($155) {
              $ef$0$lcssa = $ef$2;$el$0$lcssa = $el$1;$i$0$lcssa = $136;
              break L73;
             } else {
              $$02$i = $154;
             }
            }
            $133 = ((($$02$i$lcssa)) + 4|0); //@line 1150 "load_mid.cpp"
            $134 = HEAP32[$133>>2]|0; //@line 1150 "load_mid.cpp"
            $135 = ($134>>>0)<($24>>>0); //@line 1150 "load_mid.cpp"
            if ($135) {
             $e$371170 = $$02$i$lcssa;$ef$072169 = $ef$2;$el$073168 = $el$1;$i$070171 = $136;
            } else {
             $ef$0$lcssa = $ef$2;$el$0$lcssa = $el$1;$i$0$lcssa = $136;
             break;
            }
           }
          } else {
           $ef$0$lcssa = 0;$el$0$lcssa = $e$262$lcssa;$i$0$lcssa = 0;
          }
         } while(0);
         switch ($i$0$lcssa|0) {
         case 0:  {
          label = 55;
          break L35;
          break;
         }
         case 1:  {
          break;
         }
         default: {
          $156 = ($ef$0$lcssa|0)!=($el$0$lcssa|0); //@line 1159 "load_mid.cpp"
          $157 = ($ef$0$lcssa|0)!=(0|0); //@line 1159 "load_mid.cpp"
          $or$cond = $157 & $156; //@line 1159 "load_mid.cpp"
          if ($or$cond) {
           $206 = ((($ef$0$lcssa)) + 9|0); //@line 1203 "load_mid.cpp"
           $207 = HEAP8[$206>>0]|0; //@line 1203 "load_mid.cpp"
           $208 = $207&255; //@line 1203 "load_mid.cpp"
           $209 = (_pat_modnote($208)|0); //@line 1203 "load_mid.cpp"
           $210 = $209&255; //@line 1203 "load_mid.cpp"
           $211 = ((($el$0$lcssa)) + 9|0); //@line 1204 "load_mid.cpp"
           $212 = HEAP8[$211>>0]|0; //@line 1204 "load_mid.cpp"
           $213 = $212&255; //@line 1204 "load_mid.cpp"
           $214 = (_pat_modnote($213)|0); //@line 1204 "load_mid.cpp"
           $215 = ((($el$0$lcssa)) + 11|0); //@line 1205 "load_mid.cpp"
           $216 = HEAP8[$215>>0]|0; //@line 1205 "load_mid.cpp"
           $217 = ((($el$0$lcssa)) + 10|0); //@line 1206 "load_mid.cpp"
           $218 = HEAP8[$217>>0]|0; //@line 1206 "load_mid.cpp"
           $div8 = ($218&255) >>> 1; //@line 1206 "load_mid.cpp"
           $219 = ($div8&255)>(64); //@line 1207 "load_mid.cpp"
           $$ = $219 ? 64 : $div8; //@line 1207 "load_mid.cpp"
           $220 = (((($54) + (($53*6)|0)|0)) + 1|0); //@line 1208 "load_mid.cpp"
           HEAP8[$220>>0] = $216; //@line 1208 "load_mid.cpp"
           HEAP8[$55>>0] = $210; //@line 1209 "load_mid.cpp"
           $221 = (((($54) + (($53*6)|0)|0)) + 2|0); //@line 1210 "load_mid.cpp"
           HEAP8[$221>>0] = 1; //@line 1210 "load_mid.cpp"
           $222 = (((($54) + (($53*6)|0)|0)) + 4|0); //@line 1211 "load_mid.cpp"
           HEAP8[$222>>0] = $$; //@line 1211 "load_mid.cpp"
           $223 = $209 & 255; //@line 1212 "load_mid.cpp"
           $224 = ($214|0)>($223|0); //@line 1212 "load_mid.cpp"
           $225 = (($214) - ($223))|0; //@line 1212 "load_mid.cpp"
           $226 = (($223) - ($214))|0; //@line 1212 "load_mid.cpp"
           $227 = $224 ? $225 : $226; //@line 1212 "load_mid.cpp"
           $228 = $227&255; //@line 1212 "load_mid.cpp"
           HEAP8[$56>>0] = $228; //@line 1212 "load_mid.cpp"
           $229 = $227 & 240; //@line 1213 "load_mid.cpp"
           $230 = ($229>>>0)<(16); //@line 1213 "load_mid.cpp"
           if (!($230)) {
            $242 = $224 ? 2 : 3; //@line 1224 "load_mid.cpp"
            HEAP8[$57>>0] = $242; //@line 1224 "load_mid.cpp"
            $244 = $228;
            break L35;
           }
           $231 = ($228<<24>>24)==(0); //@line 1214 "load_mid.cpp"
           if ($231) {
            HEAP8[$57>>0] = 15; //@line 1219 "load_mid.cpp"
            $235 = ((($el$0$lcssa)) + 4|0); //@line 1220 "load_mid.cpp"
            $236 = HEAP32[$235>>2]|0; //@line 1220 "load_mid.cpp"
            $237 = (($236) - ($22))|0; //@line 1220 "load_mid.cpp"
            $h$idx13$val = HEAP32[$4>>2]|0;
            $h$idx14$val = HEAP32[$h$idx12>>2]|0;
            $238 = $h$idx13$val << 4; //@line 261 "load_mid.cpp"
            $239 = Math_imul($238, $237)|0; //@line 261 "load_mid.cpp"
            $240 = (($239>>>0) / ($h$idx14$val>>>0))&-1; //@line 261 "load_mid.cpp"
            $241 = $240&255; //@line 1220 "load_mid.cpp"
            HEAP8[$56>>0] = $241; //@line 1220 "load_mid.cpp"
            $244 = $241;
            break L35;
           } else {
            HEAP8[$57>>0] = 28; //@line 1215 "load_mid.cpp"
            $232 = $224 ? 16 : 32; //@line 1216 "load_mid.cpp"
            $233 = $227 | $232; //@line 1216 "load_mid.cpp"
            $234 = $233&255; //@line 1216 "load_mid.cpp"
            HEAP8[$56>>0] = $234; //@line 1216 "load_mid.cpp"
            $244 = $234;
            break L35;
           }
          }
         }
         }
         $158 = ($ef$0$lcssa|0)==(0|0); //@line 1160 "load_mid.cpp"
         $e$2$ef$0 = $158 ? $e$262$lcssa : $ef$0$lcssa;
         $159 = HEAP32[$58>>2]|0; //@line 1162 "load_mid.cpp"
         $160 = ((($e$2$ef$0)) + 9|0); //@line 1163 "load_mid.cpp"
         $161 = HEAP8[$160>>0]|0; //@line 1163 "load_mid.cpp"
         $162 = $161&255; //@line 1163 "load_mid.cpp"
         $163 = (_pat_modnote($162)|0); //@line 1163 "load_mid.cpp"
         $164 = $163&255; //@line 1163 "load_mid.cpp"
         $165 = ((($e$2$ef$0)) + 11|0); //@line 1164 "load_mid.cpp"
         $166 = HEAP8[$165>>0]|0; //@line 1164 "load_mid.cpp"
         $167 = ((($e$2$ef$0)) + 10|0); //@line 1165 "load_mid.cpp"
         $168 = HEAP8[$167>>0]|0; //@line 1165 "load_mid.cpp"
         $169 = ($168<<24>>24)==(0); //@line 1165 "load_mid.cpp"
         do {
          if ($169) {
           $170 = ((($e$2$ef$0)) + 4|0); //@line 1166 "load_mid.cpp"
           $171 = HEAP32[$170>>2]|0; //@line 1166 "load_mid.cpp"
           $172 = (($171) - ($22))|0; //@line 1166 "load_mid.cpp"
           $h$idx19$val = HEAP32[$4>>2]|0;
           $h$idx20$val = HEAP32[$h$idx12>>2]|0;
           $173 = $h$idx19$val << 4; //@line 261 "load_mid.cpp"
           $174 = Math_imul($173, $172)|0; //@line 261 "load_mid.cpp"
           $175 = (($174>>>0) / ($h$idx20$val>>>0))&-1; //@line 261 "load_mid.cpp"
           $176 = $175&255; //@line 1166 "load_mid.cpp"
           HEAP8[$56>>0] = $176; //@line 1166 "load_mid.cpp"
           $177 = ($176<<24>>24)==(0); //@line 1167 "load_mid.cpp"
           if ($177) {
            HEAP8[$56>>0] = 0; //@line 1172 "load_mid.cpp"
            HEAP8[$57>>0] = 25; //@line 1173 "load_mid.cpp"
            $$pr104 = 0;$vol$0 = 0;
            break;
           } else {
            HEAP8[$57>>0] = 20; //@line 1168 "load_mid.cpp"
            $178 = $175 | 192; //@line 1169 "load_mid.cpp"
            $179 = $178&255; //@line 1169 "load_mid.cpp"
            HEAP8[$56>>0] = $179; //@line 1169 "load_mid.cpp"
            $$pr104 = $179;$vol$0 = 0;
            break;
           }
          } else {
           $div = ($168&255) >>> 1; //@line 1178 "load_mid.cpp"
           $180 = ((($159)) + 10|0); //@line 1179 "load_mid.cpp"
           $181 = HEAP8[$180>>0]|0; //@line 1179 "load_mid.cpp"
           $182 = ($181<<24>>24)==(0); //@line 1179 "load_mid.cpp"
           if ($182) {
            $183 = ((($159)) + 4|0); //@line 1180 "load_mid.cpp"
            $184 = HEAP32[$183>>2]|0; //@line 1180 "load_mid.cpp"
            $185 = (($184) - ($22))|0; //@line 1180 "load_mid.cpp"
            $h$idx17$val = HEAP32[$4>>2]|0;
            $h$idx18$val = HEAP32[$h$idx12>>2]|0;
            $186 = $h$idx17$val << 4; //@line 261 "load_mid.cpp"
            $187 = Math_imul($186, $185)|0; //@line 261 "load_mid.cpp"
            $188 = (($187>>>0) / ($h$idx18$val>>>0))&-1; //@line 261 "load_mid.cpp"
            $189 = $188&255; //@line 1180 "load_mid.cpp"
            HEAP8[$56>>0] = $189; //@line 1180 "load_mid.cpp"
            $190 = ($189<<24>>24)==(0); //@line 1181 "load_mid.cpp"
            if ($190) {
             $$pr104 = 0;$vol$0 = $div;
             break;
            }
            HEAP8[$57>>0] = 20; //@line 1182 "load_mid.cpp"
            $191 = $188 | 192; //@line 1183 "load_mid.cpp"
            $192 = $191&255; //@line 1183 "load_mid.cpp"
            HEAP8[$56>>0] = $192; //@line 1183 "load_mid.cpp"
            $$pr104 = $192;$vol$0 = $div;
            break;
           } else {
            $193 = ((($e$2$ef$0)) + 4|0); //@line 1187 "load_mid.cpp"
            $194 = HEAP32[$193>>2]|0; //@line 1187 "load_mid.cpp"
            $195 = (($194) - ($22))|0; //@line 1187 "load_mid.cpp"
            $h$idx15$val = HEAP32[$4>>2]|0;
            $h$idx16$val = HEAP32[$h$idx12>>2]|0;
            $196 = $h$idx15$val << 4; //@line 261 "load_mid.cpp"
            $197 = Math_imul($196, $195)|0; //@line 261 "load_mid.cpp"
            $198 = (($197>>>0) / ($h$idx16$val>>>0))&-1; //@line 261 "load_mid.cpp"
            $199 = $198&255; //@line 1187 "load_mid.cpp"
            HEAP8[$56>>0] = $199; //@line 1187 "load_mid.cpp"
            $200 = ($199<<24>>24)==(0); //@line 1188 "load_mid.cpp"
            if ($200) {
             $$pr104 = 0;$vol$0 = $div;
             break;
            }
            HEAP8[$57>>0] = 20; //@line 1189 "load_mid.cpp"
            $201 = $198 | 208; //@line 1190 "load_mid.cpp"
            $202 = $201&255; //@line 1190 "load_mid.cpp"
            HEAP8[$56>>0] = $202; //@line 1190 "load_mid.cpp"
            $$pr104 = $202;$vol$0 = $div;
            break;
           }
          }
         } while(0);
         $203 = (((($54) + (($53*6)|0)|0)) + 1|0); //@line 1194 "load_mid.cpp"
         HEAP8[$203>>0] = $166; //@line 1194 "load_mid.cpp"
         HEAP8[$55>>0] = $164; //@line 1195 "load_mid.cpp"
         $204 = (((($54) + (($53*6)|0)|0)) + 2|0); //@line 1196 "load_mid.cpp"
         HEAP8[$204>>0] = 1; //@line 1196 "load_mid.cpp"
         $205 = (((($54) + (($53*6)|0)|0)) + 4|0); //@line 1197 "load_mid.cpp"
         HEAP8[$205>>0] = $vol$0; //@line 1197 "load_mid.cpp"
         $244 = $$pr104;
        }
       }
      } while(0);
      if ((label|0) == 55) {
       label = 0;
       $$pr$pre = HEAP8[$56>>0]|0; //@line 1227 "load_mid.cpp"
       $244 = $$pr$pre;
      }
      $243 = ($244<<24>>24)==(0); //@line 1227 "load_mid.cpp"
      do {
       if ($243) {
        $245 = HEAP8[$57>>0]|0; //@line 1227 "load_mid.cpp"
        $246 = ($245<<24>>24)==(0); //@line 1227 "load_mid.cpp"
        if ($246) {
         $247 = ($tempo$284|0)==(0); //@line 1228 "load_mid.cpp"
         if (!($247)) {
          HEAP8[$57>>0] = 17; //@line 1229 "load_mid.cpp"
          $248 = $tempo$284&255; //@line 1230 "load_mid.cpp"
          HEAP8[$56>>0] = $248; //@line 1230 "load_mid.cpp"
          $patbrk$3 = $patbrk$283;$tempo$3 = 0;
          break;
         }
         $249 = ($patbrk$283|0)==(0); //@line 1234 "load_mid.cpp"
         if ($249) {
          $patbrk$3 = 0;$tempo$3 = 0;
         } else {
          HEAP8[$57>>0] = 14; //@line 1235 "load_mid.cpp"
          $patbrk$3 = 0;$tempo$3 = 0;
         }
        } else {
         $patbrk$3 = $patbrk$283;$tempo$3 = $tempo$284;
        }
       } else {
        $patbrk$3 = $patbrk$283;$tempo$3 = $tempo$284;
       }
      } while(0);
      $t$1 = HEAP32[$t$185>>2]|0; //@line 1122 "load_mid.cpp"
      $250 = ($t$1|0)==(0|0); //@line 1122 "load_mid.cpp"
      if ($250) {
       $patbrk$2$lcssa = $patbrk$3;$tempo$2$lcssa = $tempo$3;
       break L31;
      }
      $251 = (($ch$082) + 1)|0; //@line 1240 "load_mid.cpp"
      $$phi$trans$insert = ((($t$1)) + 12|0);
      $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 1126 "load_mid.cpp"
      $60 = $$pre;$ch$082 = $251;$patbrk$283 = $patbrk$3;$t$185 = $t$1;$tempo$284 = $tempo$3;
     }
    }
   } while(0);
   $252 = $tempo$2$lcssa | $patbrk$2$lcssa; //@line 1242 "load_mid.cpp"
   $253 = ($252|0)==(0); //@line 1242 "load_mid.cpp"
   $16 = (($row$089) + 1)|0; //@line 1104 "load_mid.cpp"
   if (!($253)) {
    $$0 = 1;
    label = 82;
    break L8;
   }
   $15 = ($16|0)<(64); //@line 1104 "load_mid.cpp"
   if ($15) {
    $row$089 = $16;
   } else {
    break;
   }
  }
  $254 = (($pat$090) + 1)|0; //@line 1100 "load_mid.cpp"
  $255 = ($254|0)<($$numpat|0); //@line 1100 "load_mid.cpp"
  if ($255) {
   $pat$090 = $254;
  } else {
   $$0 = 0;
   label = 82;
   break;
  }
 }
 if ((label|0) == 82) {
  return ($$0|0); //@line 1246 "load_mid.cpp"
 }
 return (0)|0;
}
function _pat_smplooped($smp) {
 $smp = $smp|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($smp|0)<(191); //@line 231 "load_pat.cpp"
 if (!($0)) {
  $$0 = 1;
  return ($$0|0); //@line 233 "load_pat.cpp"
 }
 $1 = (($smp) + -1)|0; //@line 231 "load_pat.cpp"
 $2 = (1252508 + ($1)|0); //@line 231 "load_pat.cpp"
 $3 = HEAP8[$2>>0]|0; //@line 231 "load_pat.cpp"
 $4 = $3&255; //@line 231 "load_pat.cpp"
 $$0 = $4;
 return ($$0|0); //@line 233 "load_pat.cpp"
}
function __ZN10CSoundFile7ReadPATEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$$i = 0, $$$i4 = 0, $$$us$i = 0, $$0$i$i = 0, $$0$i$us$i = 0, $$2 = 0, $$lcssa40 = 0, $$numpat$i = 0, $$pre = 0, $$pre$i$i = 0, $$pre$i2 = 0, $$pre23 = 0, $$sz$i = 0, $$sz$i$i = 0, $$sz$i$i$i = 0, $$sz$i1$i$i = 0, $0 = 0, $1 = 0, $10 = 0;
 var $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0;
 var $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0;
 var $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0;
 var $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0;
 var $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0;
 var $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0;
 var $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0;
 var $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0;
 var $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0;
 var $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0;
 var $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0.0, $95 = 0.0;
 var $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $buf = 0, $ch$04$i = 0, $ch$04$us$i = 0, $d$02$i = 0, $exitcond$i = 0, $exitcond$i5 = 0, $hl$i = 0, $hw = 0, $hw1 = 0, $ih$i = 0, $ih$i$i = 0, $leninc = 0, $mm = 0, $or$cond$us$i = 0, $p$0 = 0, $pat$08$i = 0;
 var $ph$i = 0, $row$05$i = 0, $s = 0, $s$01$i = 0, $s2 = 0, $storemerge$i = 0, $strlen = 0, $strlenfirst = 0, $t$03$i = 0, $t$115 = 0, $t$213 = 0, $t$311 = 0, $t$410 = 0, $vararg_buffer = 0, $vararg_buffer13 = 0, $vararg_buffer2 = 0, $vararg_buffer5 = 0, $vararg_buffer9 = 0, $vararg_ptr1 = 0, $vararg_ptr12 = 0;
 var $vararg_ptr8 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 688|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer13 = sp + 32|0;
 $vararg_buffer9 = sp + 24|0;
 $vararg_buffer5 = sp + 16|0;
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $ih$i$i = sp + 611|0;
 $hl$i = sp + 564|0;
 $ih$i = sp + 501|0;
 $ph$i = sp + 372|0;
 $buf = sp + 312|0;
 $mm = sp + 40|0;
 $hw = sp + 216|0;
 $s = sp + 184|0;
 $hw1 = sp + 88|0;
 $s2 = sp + 56|0;
 $0 = ($dwMemLength>>>0)<(129); //@line 839 "load_pat.cpp"
 if (!($0)) {
  _memcpy(($ph$i|0),($lpStream|0),129)|0; //@line 840 "load_pat.cpp"
  $1 = (_strcmp($ph$i,529638)|0); //@line 842 "load_pat.cpp"
  $2 = ($1|0)==(0); //@line 842 "load_pat.cpp"
  if ($2) {
   $3 = ((($ph$i)) + 12|0); //@line 842 "load_pat.cpp"
   $4 = (_strcmp($3,529650)|0); //@line 842 "load_pat.cpp"
   $5 = ($4|0)==(0); //@line 842 "load_pat.cpp"
   if ($5) {
    $6 = (_calloc(1,20)|0); //@line 858 "load_pat.cpp"
    $7 = ($6|0)==(0|0); //@line 1445 "load_pat.cpp"
    if ($7) {
     $$2 = 0;
     STACKTOP = sp;return ($$2|0); //@line 1638 "load_pat.cpp"
    }
    HEAP32[$mm>>2] = $lpStream; //@line 1447 "load_pat.cpp"
    $8 = ((($mm)) + 4|0); //@line 1448 "load_pat.cpp"
    HEAP32[$8>>2] = $dwMemLength; //@line 1448 "load_pat.cpp"
    $9 = ((($mm)) + 8|0); //@line 1449 "load_pat.cpp"
    HEAP32[$9>>2] = 0; //@line 1449 "load_pat.cpp"
    $10 = ((($mm)) + 12|0); //@line 1450 "load_pat.cpp"
    HEAP32[$10>>2] = 0; //@line 1450 "load_pat.cpp"
    $11 = HEAP32[300432]|0; //@line 1452 "load_pat.cpp"
    $12 = ($11|0)==(0); //@line 1452 "load_pat.cpp"
    if (!($12)) {
     while(1) {
      (_sleep(1)|0); //@line 1452 "load_pat.cpp"
      $13 = HEAP32[300432]|0; //@line 1452 "load_pat.cpp"
      $14 = ($13|0)==(0); //@line 1452 "load_pat.cpp"
      if ($14) {
       break;
      }
     }
    }
    HEAP32[300432] = 1; //@line 1453 "load_pat.cpp"
    $15 = ($dwMemLength|0)>(129); //@line 348 "load_pat.cpp"
    if ($15) {
     HEAP32[$9>>2] = 129; //@line 349 "load_pat.cpp"
     $17 = 129;
    } else {
     HEAP32[$10>>2] = 1; //@line 351 "load_pat.cpp"
     $17 = 0;
    }
    $16 = (($dwMemLength) - ($17))|0; //@line 359 "load_pat.cpp"
    $18 = ($16|0)<(63); //@line 359 "load_pat.cpp"
    $$sz$i$i = $18 ? $16 : 63; //@line 359 "load_pat.cpp"
    $19 = (($lpStream) + ($17)|0); //@line 361 "load_pat.cpp"
    _memcpy(($ih$i|0),($19|0),($$sz$i$i|0))|0; //@line 361 "load_pat.cpp"
    $20 = (($$sz$i$i) + ($17))|0; //@line 362 "load_pat.cpp"
    HEAP32[$9>>2] = $20; //@line 362 "load_pat.cpp"
    $21 = ((($ih$i)) + 2|0); //@line 506 "load_pat.cpp"
    (_strncpy($6,$21,16)|0); //@line 506 "load_pat.cpp"
    $22 = ((($6)) + 15|0); //@line 507 "load_pat.cpp"
    HEAP8[$22>>0] = 0; //@line 507 "load_pat.cpp"
    $23 = HEAP32[$8>>2]|0; //@line 348 "load_pat.cpp"
    $24 = ($23|0)>(129); //@line 348 "load_pat.cpp"
    if ($24) {
     HEAP32[$9>>2] = 129; //@line 349 "load_pat.cpp"
     $26 = 129;
    } else {
     HEAP32[$10>>2] = 1; //@line 351 "load_pat.cpp"
     $$pre$i$i = HEAP32[$9>>2]|0; //@line 359 "load_pat.cpp"
     $26 = $$pre$i$i;
    }
    $25 = (($23) - ($26))|0; //@line 359 "load_pat.cpp"
    $27 = ($25|0)<(63); //@line 359 "load_pat.cpp"
    $$sz$i1$i$i = $27 ? $25 : 63; //@line 359 "load_pat.cpp"
    $28 = HEAP32[$mm>>2]|0; //@line 361 "load_pat.cpp"
    $29 = (($28) + ($26)|0); //@line 361 "load_pat.cpp"
    _memcpy(($ih$i$i|0),($29|0),($$sz$i1$i$i|0))|0; //@line 361 "load_pat.cpp"
    $30 = (($$sz$i1$i$i) + ($26))|0; //@line 362 "load_pat.cpp"
    HEAP32[$9>>2] = $30; //@line 362 "load_pat.cpp"
    $31 = (($23) - ($30))|0; //@line 359 "load_pat.cpp"
    $32 = ($31|0)<(47); //@line 359 "load_pat.cpp"
    $$sz$i$i$i = $32 ? $31 : 47; //@line 359 "load_pat.cpp"
    $33 = (($28) + ($30)|0); //@line 361 "load_pat.cpp"
    _memcpy(($hl$i|0),($33|0),($$sz$i$i$i|0))|0; //@line 361 "load_pat.cpp"
    $34 = (($$sz$i$i$i) + ($30))|0; //@line 362 "load_pat.cpp"
    HEAP32[$9>>2] = $34; //@line 362 "load_pat.cpp"
    $35 = ((($hl$i)) + 7|0); //@line 522 "load_pat.cpp"
    $36 = ((($ih$i$i)) + 2|0); //@line 522 "load_pat.cpp"
    (_strncpy($35,$36,40)|0); //@line 522 "load_pat.cpp"
    $37 = ((($hl$i)) + 6|0); //@line 528 "load_pat.cpp"
    $38 = HEAP8[$37>>0]|0; //@line 528 "load_pat.cpp"
    $39 = $38&255; //@line 528 "load_pat.cpp"
    $40 = ((($6)) + 16|0); //@line 1455 "load_pat.cpp"
    HEAP8[$40>>0]=$39&255;HEAP8[$40+1>>0]=($39>>8)&255;HEAP8[$40+2>>0]=($39>>16)&255;HEAP8[$40+3>>0]=$39>>24; //@line 1455 "load_pat.cpp"
    $strlenfirst = HEAP8[$6>>0]|0; //@line 1456 "load_pat.cpp"
    $41 = ($strlenfirst<<24>>24)==(0); //@line 1456 "load_pat.cpp"
    if ($41) {
     HEAP32[$vararg_buffer2>>2] = $39; //@line 1459 "load_pat.cpp"
     (_sprintf($buf,529688,$vararg_buffer2)|0); //@line 1459 "load_pat.cpp"
    } else {
     HEAP32[$vararg_buffer>>2] = $6; //@line 1457 "load_pat.cpp"
     $vararg_ptr1 = ((($vararg_buffer)) + 4|0); //@line 1457 "load_pat.cpp"
     HEAP32[$vararg_ptr1>>2] = $39; //@line 1457 "load_pat.cpp"
     (_sprintf($buf,529660,$vararg_buffer)|0); //@line 1457 "load_pat.cpp"
    }
    $42 = (_strlen($buf)|0); //@line 1463 "load_pat.cpp"
    $43 = ($42>>>0)>(31); //@line 1463 "load_pat.cpp"
    if ($43) {
     $44 = ((($buf)) + 31|0); //@line 1463 "load_pat.cpp"
     HEAP8[$44>>0] = 0; //@line 1463 "load_pat.cpp"
    }
    $45 = ((($this)) + 59804|0); //@line 1464 "load_pat.cpp"
    (_strcpy($45,$buf)|0); //@line 1464 "load_pat.cpp"
    $46 = ((($this)) + 59652|0); //@line 1466 "load_pat.cpp"
    HEAP32[$46>>2] = 60; //@line 1466 "load_pat.cpp"
    $47 = HEAPU8[$40>>0]|(HEAPU8[$40+1>>0]<<8)|(HEAPU8[$40+2>>0]<<16)|(HEAPU8[$40+3>>0]<<24); //@line 1467 "load_pat.cpp"
    $48 = $47 << 4; //@line 1467 "load_pat.cpp"
    $49 = (($48) + 112)|0; //@line 1467 "load_pat.cpp"
    $50 = $49 & 48; //@line 1468 "load_pat.cpp"
    $51 = ($50|0)==(0); //@line 1468 "load_pat.cpp"
    $52 = (($48) + 176)|0; //@line 1468 "load_pat.cpp"
    $$ = $51 ? $49 : $52; //@line 1468 "load_pat.cpp"
    $53 = (($$|0) / 64)&-1; //@line 1469 "load_pat.cpp"
    $54 = ((($this)) + 59680|0); //@line 1525 "load_pat.cpp"
    HEAP32[$54>>2] = 33554432; //@line 1525 "load_pat.cpp"
    $55 = (($47) + 1)|0; //@line 1526 "load_pat.cpp"
    $56 = ((($this)) + 59688|0); //@line 1526 "load_pat.cpp"
    HEAP32[$56>>2] = $55; //@line 1526 "load_pat.cpp"
    $57 = ((($this)) + 59684|0); //@line 1527 "load_pat.cpp"
    HEAP32[$57>>2] = $55; //@line 1527 "load_pat.cpp"
    $58 = ((($this)) + 59648|0); //@line 1528 "load_pat.cpp"
    HEAP32[$58>>2] = 6; //@line 1528 "load_pat.cpp"
    $59 = ((($this)) + 59664|0); //@line 1529 "load_pat.cpp"
    HEAP32[$59>>2] = $47; //@line 1529 "load_pat.cpp"
    $60 = ((($this)) + 59660|0); //@line 1532 "load_pat.cpp"
    HEAP32[$60>>2] = 16; //@line 1532 "load_pat.cpp"
    $61 = ((($this)) + 59764|0); //@line 1533 "load_pat.cpp"
    HEAP32[$61>>2] = 112; //@line 1533 "load_pat.cpp"
    $62 = ((($this)) + 59768|0); //@line 1534 "load_pat.cpp"
    HEAP32[$62>>2] = 13696; //@line 1534 "load_pat.cpp"
    $63 = ($$|0)>(63); //@line 1536 "load_pat.cpp"
    if ($63) {
     $t$115 = 0;
     while(1) {
      $68 = $t$115&255; //@line 1537 "load_pat.cpp"
      $69 = (((($this)) + 53344|0) + ($t$115)|0); //@line 1537 "load_pat.cpp"
      HEAP8[$69>>0] = $68; //@line 1537 "load_pat.cpp"
      $70 = (($t$115) + 1)|0; //@line 1536 "load_pat.cpp"
      $71 = ($70|0)<($53|0); //@line 1536 "load_pat.cpp"
      if ($71) {
       $t$115 = $70;
      } else {
       break;
      }
     }
     $$pre = HEAP32[$56>>2]|0; //@line 1538 "load_pat.cpp"
     $65 = $$pre;
    } else {
     $65 = $55;
    }
    $64 = ($65|0)>(1); //@line 1538 "load_pat.cpp"
    if ($64) {
     $66 = ((($s)) + 31|0); //@line 1548 "load_pat.cpp"
     $67 = ((($s)) + 11|0); //@line 1551 "load_pat.cpp"
     $t$213 = 1;
     while(1) {
      $81 = (__Znwj(632)|0); //@line 1541 "load_pat.cpp"
      _memset(($81|0),0,632)|0; //@line 1545 "load_pat.cpp"
      $82 = (((($this)) + 48640|0) + ($t$213<<2)|0); //@line 1546 "load_pat.cpp"
      HEAP32[$82>>2] = $81; //@line 1546 "load_pat.cpp"
      $strlen = (_strlen($6)|0); //@line 1547 "load_pat.cpp"
      $leninc = (($strlen) + 1)|0; //@line 1547 "load_pat.cpp"
      _memcpy(($s|0),($6|0),($leninc|0))|0; //@line 1547 "load_pat.cpp"
      HEAP8[$66>>0] = 0; //@line 1548 "load_pat.cpp"
      $83 = ((($81)) + 585|0); //@line 1549 "load_pat.cpp"
      dest=$83; stop=dest+32|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 1549 "load_pat.cpp"
      (_strcpy($83,$s)|0); //@line 1550 "load_pat.cpp"
      HEAP8[$67>>0] = 0; //@line 1551 "load_pat.cpp"
      $84 = ((($81)) + 617|0); //@line 1552 "load_pat.cpp"
      dest=$84; stop=dest+12|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 1552 "load_pat.cpp"
      (_strcpy($84,$s)|0); //@line 1553 "load_pat.cpp"
      __ZL18pat_get_waveheaderP6MMFILEP10WaveHeaderi($mm,$hw,$t$213); //@line 1554 "load_pat.cpp"
      __ZL15pat_setpat_instP10WaveHeaderP17_INSTRUMENTHEADERi($hw,$81,$t$213); //@line 1555 "load_pat.cpp"
      $85 = (($t$213) + 1)|0; //@line 1538 "load_pat.cpp"
      $86 = HEAP32[$56>>2]|0; //@line 1538 "load_pat.cpp"
      $87 = ($85|0)<($86|0); //@line 1538 "load_pat.cpp"
      if ($87) {
       $t$213 = $85;
      } else {
       $285 = $86;
       break;
      }
     }
    } else {
     $285 = $65;
    }
    $72 = HEAP32[$57>>2]|0; //@line 1557 "load_pat.cpp"
    $73 = ($72|0)>(1); //@line 1557 "load_pat.cpp"
    if ($73) {
     $74 = ((($hw1)) + 20|0); //@line 1279 "load_pat.cpp"
     $75 = ((($hw1)) + 30|0); //@line 1279 "load_pat.cpp"
     $76 = ((($hw1)) + 8|0); //@line 1280 "load_pat.cpp"
     $77 = ((($hw1)) + 12|0); //@line 1281 "load_pat.cpp"
     $78 = ((($hw1)) + 16|0); //@line 1282 "load_pat.cpp"
     $79 = ((($hw1)) + 55|0); //@line 1284 "load_pat.cpp"
     $80 = ((($s2)) + 31|0); //@line 1576 "load_pat.cpp"
     $t$311 = 1;
     while(1) {
      $88 = (((($this)) + 33280|0) + ($t$311<<6)|0); //@line 1561 "load_pat.cpp"
      $89 = (((((($this)) + 33280|0) + ($t$311<<6)|0)) + 32|0); //@line 1562 "load_pat.cpp"
      HEAP16[$89>>1] = 64; //@line 1562 "load_pat.cpp"
      $90 = (((((($this)) + 33280|0) + ($t$311<<6)|0)) + 28|0); //@line 1563 "load_pat.cpp"
      HEAP16[$90>>1] = 128; //@line 1563 "load_pat.cpp"
      $91 = (((((($this)) + 33280|0) + ($t$311<<6)|0)) + 34|0); //@line 1564 "load_pat.cpp"
      HEAP16[$91>>1] = 1; //@line 1564 "load_pat.cpp"
      __ZL18pat_get_waveheaderP6MMFILEP10WaveHeaderi($mm,$hw1,$t$311); //@line 1565 "load_pat.cpp"
      $92 = HEAPU8[$74>>0]|(HEAPU8[$74+1>>0]<<8); //@line 1279 "load_pat.cpp"
      $93 = HEAPU8[$75>>0]|(HEAPU8[$75+1>>0]<<8)|(HEAPU8[$75+2>>0]<<16)|(HEAPU8[$75+3>>0]<<24); //@line 1279 "load_pat.cpp"
      $94 = (+($92&65535)); //@line 1041 "load_pat.cpp"
      $95 = $94 * 52325.100000000006; //@line 1042 "load_pat.cpp"
      $96 = (+($93>>>0)); //@line 1043 "load_pat.cpp"
      $97 = $96 * 0.40000000000000002; //@line 1043 "load_pat.cpp"
      $98 = $95 / $97; //@line 1044 "load_pat.cpp"
      $99 = $98 + 0.5; //@line 1045 "load_pat.cpp"
      $100 = (~~(($99))>>>0); //@line 1045 "load_pat.cpp"
      $101 = (((((($this)) + 33280|0) + ($t$311<<6)|0)) + 24|0); //@line 1279 "load_pat.cpp"
      HEAP32[$101>>2] = $100; //@line 1279 "load_pat.cpp"
      $102 = HEAPU8[$76>>0]|(HEAPU8[$76+1>>0]<<8)|(HEAPU8[$76+2>>0]<<16)|(HEAPU8[$76+3>>0]<<24); //@line 1280 "load_pat.cpp"
      HEAP32[$88>>2] = $102; //@line 1280 "load_pat.cpp"
      $103 = HEAPU8[$77>>0]|(HEAPU8[$77+1>>0]<<8)|(HEAPU8[$77+2>>0]<<16)|(HEAPU8[$77+3>>0]<<24); //@line 1281 "load_pat.cpp"
      $104 = (((((($this)) + 33280|0) + ($t$311<<6)|0)) + 4|0); //@line 1281 "load_pat.cpp"
      HEAP32[$104>>2] = $103; //@line 1281 "load_pat.cpp"
      $105 = HEAPU8[$78>>0]|(HEAPU8[$78+1>>0]<<8)|(HEAPU8[$78+2>>0]<<16)|(HEAPU8[$78+3>>0]<<24); //@line 1282 "load_pat.cpp"
      $106 = (((((($this)) + 33280|0) + ($t$311<<6)|0)) + 8|0); //@line 1282 "load_pat.cpp"
      HEAP32[$106>>2] = $105; //@line 1282 "load_pat.cpp"
      $107 = (((((($this)) + 33280|0) + ($t$311<<6)|0)) + 30|0); //@line 1283 "load_pat.cpp"
      HEAP16[$107>>1] = 256; //@line 1283 "load_pat.cpp"
      $108 = HEAP8[$79>>0]|0; //@line 1284 "load_pat.cpp"
      $109 = $108 & 1; //@line 1284 "load_pat.cpp"
      $110 = ($109<<24>>24)==(0); //@line 1284 "load_pat.cpp"
      if (!($110)) {
       $111 = $102 >>> 1; //@line 1285 "load_pat.cpp"
       HEAP32[$88>>2] = $111; //@line 1285 "load_pat.cpp"
       $112 = $103 >>> 1; //@line 1286 "load_pat.cpp"
       HEAP32[$104>>2] = $112; //@line 1286 "load_pat.cpp"
       $113 = $105 >>> 1; //@line 1287 "load_pat.cpp"
       HEAP32[$106>>2] = $113; //@line 1287 "load_pat.cpp"
      }
      $114 = $108 & 4; //@line 1289 "load_pat.cpp"
      $115 = ($114<<24>>24)==(0); //@line 1289 "load_pat.cpp"
      if (!($115)) {
       $116 = HEAP16[$91>>1]|0; //@line 1290 "load_pat.cpp"
       $117 = $116&65535; //@line 1290 "load_pat.cpp"
       $118 = $108 & 8; //@line 1291 "load_pat.cpp"
       $119 = $118&255; //@line 1291 "load_pat.cpp"
       $120 = $119 << 1; //@line 1291 "load_pat.cpp"
       $121 = $120 | $117; //@line 1291 "load_pat.cpp"
       $$$i = $121 | 2; //@line 1291 "load_pat.cpp"
       $storemerge$i = $$$i&65535; //@line 1291 "load_pat.cpp"
       HEAP16[$91>>1] = $storemerge$i; //@line 1290 "load_pat.cpp"
       $122 = $108 & 32; //@line 1292 "load_pat.cpp"
       $123 = ($122<<24>>24)==(0); //@line 1292 "load_pat.cpp"
       if (!($123)) {
        $124 = $121 | 10; //@line 1292 "load_pat.cpp"
        $125 = $124&65535; //@line 1292 "load_pat.cpp"
        HEAP16[$91>>1] = $125; //@line 1292 "load_pat.cpp"
       }
      }
      dest=$s2; stop=dest+32|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 1567 "load_pat.cpp"
      $126 = HEAP8[$hw1>>0]|0; //@line 1568 "load_pat.cpp"
      $127 = ($126<<24>>24)==(0); //@line 1568 "load_pat.cpp"
      do {
       if ($127) {
        $128 = HEAP8[$6>>0]|0; //@line 1571 "load_pat.cpp"
        $129 = ($128<<24>>24)==(0); //@line 1571 "load_pat.cpp"
        if ($129) {
         HEAP32[$vararg_buffer13>>2] = $t$311; //@line 1574 "load_pat.cpp"
         (_sprintf($s2,529717,$vararg_buffer13)|0); //@line 1574 "load_pat.cpp"
         break;
        } else {
         HEAP32[$vararg_buffer9>>2] = $t$311; //@line 1572 "load_pat.cpp"
         $vararg_ptr12 = ((($vararg_buffer9)) + 4|0); //@line 1572 "load_pat.cpp"
         HEAP32[$vararg_ptr12>>2] = $6; //@line 1572 "load_pat.cpp"
         (_sprintf($s2,527904,$vararg_buffer9)|0); //@line 1572 "load_pat.cpp"
         break;
        }
       } else {
        HEAP32[$vararg_buffer5>>2] = $t$311; //@line 1569 "load_pat.cpp"
        $vararg_ptr8 = ((($vararg_buffer5)) + 4|0); //@line 1569 "load_pat.cpp"
        HEAP32[$vararg_ptr8>>2] = $hw1; //@line 1569 "load_pat.cpp"
        (_sprintf($s2,527904,$vararg_buffer5)|0); //@line 1569 "load_pat.cpp"
       }
      } while(0);
      HEAP8[$80>>0] = 0; //@line 1576 "load_pat.cpp"
      $130 = (((($this)) + 59804|0) + ($t$311<<5)|0); //@line 1577 "load_pat.cpp"
      dest=$130; stop=dest+32|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 1577 "load_pat.cpp"
      (_strcpy($130,$s2)|0); //@line 1578 "load_pat.cpp"
      $131 = HEAP8[$79>>0]|0; //@line 1579 "load_pat.cpp"
      $132 = $131 & 1; //@line 1579 "load_pat.cpp"
      $133 = ($132<<24>>24)==(0); //@line 1579 "load_pat.cpp"
      $134 = HEAPU8[$76>>0]|(HEAPU8[$76+1>>0]<<8)|(HEAPU8[$76+2>>0]<<16)|(HEAPU8[$76+3>>0]<<24); //@line 1579 "load_pat.cpp"
      if ($133) {
       $136 = $134 << 1; //@line 1580 "load_pat.cpp"
       $137 = (_malloc($136)|0); //@line 1580 "load_pat.cpp"
       $p$0 = $137;
      } else {
       $135 = (_malloc($134)|0); //@line 1579 "load_pat.cpp"
       $p$0 = $135;
      }
      $138 = ($p$0|0)==(0|0); //@line 1581 "load_pat.cpp"
      if (!($138)) {
       $139 = HEAP32[$8>>2]|0; //@line 368 "load_pat.cpp"
       $140 = HEAP32[$9>>2]|0; //@line 368 "load_pat.cpp"
       $141 = (($139) - ($140))|0; //@line 368 "load_pat.cpp"
       $142 = ($141|0)<($134|0); //@line 368 "load_pat.cpp"
       $$sz$i = $142 ? $141 : $134; //@line 368 "load_pat.cpp"
       $143 = HEAP32[$mm>>2]|0; //@line 370 "load_pat.cpp"
       $144 = (($143) + ($140)|0); //@line 370 "load_pat.cpp"
       _memcpy(($p$0|0),($144|0),($$sz$i|0))|0; //@line 370 "load_pat.cpp"
       $145 = (($140) + ($$sz$i))|0; //@line 371 "load_pat.cpp"
       HEAP32[$9>>2] = $145; //@line 371 "load_pat.cpp"
       $146 = $131&255; //@line 1583 "load_pat.cpp"
       $147 = $146 & 1; //@line 1583 "load_pat.cpp"
       $148 = ($147|0)==(0); //@line 1583 "load_pat.cpp"
       if ($148) {
        $152 = ($134|0)>(0); //@line 787 "load_pat.cpp"
        if ($152) {
         $153 = (($p$0) + ($134<<1)|0); //@line 786 "load_pat.cpp"
         $154 = (($p$0) + ($134)|0); //@line 785 "load_pat.cpp"
         $d$02$i = $153;$s$01$i = $154;$t$03$i = 0;
         while(1) {
          $155 = ((($s$01$i)) + -1|0); //@line 788 "load_pat.cpp"
          $156 = ((($d$02$i)) + -2|0); //@line 789 "load_pat.cpp"
          $157 = HEAP8[$155>>0]|0; //@line 790 "load_pat.cpp"
          $158 = $157 << 24 >> 24; //@line 790 "load_pat.cpp"
          $159 = $158 << 8; //@line 790 "load_pat.cpp"
          $160 = $159&65535; //@line 790 "load_pat.cpp"
          HEAP16[$156>>1] = $160; //@line 790 "load_pat.cpp"
          $161 = (($t$03$i) + 1)|0; //@line 787 "load_pat.cpp"
          $exitcond$i = ($161|0)==($134|0); //@line 787 "load_pat.cpp"
          if ($exitcond$i) {
           break;
          } else {
           $d$02$i = $156;$s$01$i = $155;$t$03$i = $161;
          }
         }
        }
        $162 = $131 & 2; //@line 1588 "load_pat.cpp"
        $163 = $162&255; //@line 1588 "load_pat.cpp"
        $164 = $163 >>> 1; //@line 1588 "load_pat.cpp"
        $165 = (($164) + 5)|0; //@line 1588 "load_pat.cpp"
        $166 = $134 << 1; //@line 1588 "load_pat.cpp"
        (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$88,$165,$p$0,$166)|0); //@line 1588 "load_pat.cpp"
       } else {
        $149 = $146 >>> 1; //@line 1584 "load_pat.cpp"
        $150 = $149 & 1; //@line 1584 "load_pat.cpp"
        $151 = (($150) + 5)|0; //@line 1584 "load_pat.cpp"
        (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$88,$151,$p$0,$134)|0); //@line 1584 "load_pat.cpp"
       }
       _free($p$0); //@line 1590 "load_pat.cpp"
      }
      $167 = (($t$311) + 1)|0; //@line 1557 "load_pat.cpp"
      $168 = HEAP32[$57>>2]|0; //@line 1557 "load_pat.cpp"
      $169 = ($167|0)<($168|0); //@line 1557 "load_pat.cpp"
      if ($169) {
       $t$311 = $167;
      } else {
       break;
      }
     }
     $$pre23 = HEAP32[$56>>2]|0; //@line 1594 "load_pat.cpp"
     $173 = $$pre23;
    } else {
     $173 = $285;
    }
    $170 = (__Znwj(632)|0); //@line 1595 "load_pat.cpp"
    $171 = ((($this)) + 48640|0); //@line 1595 "load_pat.cpp"
    HEAP32[$171>>2] = $170; //@line 1595 "load_pat.cpp"
    $172 = (($173) + -1)|0; //@line 1594 "load_pat.cpp"
    $174 = (((($this)) + 48640|0) + ($172<<2)|0); //@line 1599 "load_pat.cpp"
    $175 = HEAP32[$174>>2]|0; //@line 1599 "load_pat.cpp"
    _memcpy(($170|0),($175|0),632)|0; //@line 1599 "load_pat.cpp"
    $176 = ((($170)) + 585|0); //@line 1600 "load_pat.cpp"
    dest=$176; stop=dest+32|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 1600 "load_pat.cpp"
    $177 = HEAP8[$6>>0]|0; //@line 1601 "load_pat.cpp"
    $178 = ($177<<24>>24)==(0); //@line 1601 "load_pat.cpp"
    if ($178) {
     (_strncpy($176,529738,32)|0); //@line 1604 "load_pat.cpp"
    } else {
     (_strncpy($176,$6,32)|0); //@line 1602 "load_pat.cpp"
    }
    $179 = HEAP32[$57>>2]|0; //@line 1605 "load_pat.cpp"
    $180 = (($179) + -1)|0; //@line 1605 "load_pat.cpp"
    $181 = ((($this)) + 33280|0); //@line 1606 "load_pat.cpp"
    $182 = (((($this)) + 33280|0) + ($180<<6)|0); //@line 1606 "load_pat.cpp"
    dest=$181; src=$182; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 1606 "load_pat.cpp"
    $183 = ($53|0)>(240); //@line 984 "load_pat.cpp"
    $$numpat$i = $183 ? 240 : $53; //@line 984 "load_pat.cpp"
    $184 = HEAPU8[$40>>0]|(HEAPU8[$40+1>>0]<<8)|(HEAPU8[$40+2>>0]<<16)|(HEAPU8[$40+3>>0]<<24); //@line 986 "load_pat.cpp"
    $185 = $184 << 4; //@line 986 "load_pat.cpp"
    $186 = (($185) + 112)|0; //@line 986 "load_pat.cpp"
    $187 = ($$numpat$i|0)>(0); //@line 987 "load_pat.cpp"
    L77: do {
     if ($187) {
      $188 = $184;$pat$08$i = 0;
      while(1) {
       $189 = (__ZN10CSoundFile15AllocatePatternEjj(64,$188)|0); //@line 988 "load_pat.cpp"
       $190 = (((($this)) + 51904|0) + ($pat$08$i<<2)|0); //@line 988 "load_pat.cpp"
       HEAP32[$190>>2] = $189; //@line 988 "load_pat.cpp"
       $191 = ($189|0)==(0|0); //@line 989 "load_pat.cpp"
       if ($191) {
        break L77;
       }
       $192 = (((($this)) + 52864|0) + ($pat$08$i<<1)|0); //@line 990 "load_pat.cpp"
       HEAP16[$192>>1] = 64; //@line 990 "load_pat.cpp"
       $193 = $pat$08$i << 6; //@line 992 "load_pat.cpp"
       $$pre$i2 = HEAPU8[$40>>0]|(HEAPU8[$40+1>>0]<<8)|(HEAPU8[$40+2>>0]<<16)|(HEAPU8[$40+3>>0]<<24); //@line 993 "load_pat.cpp"
       $196 = $$pre$i2;$row$05$i = 0;
       while(1) {
        $194 = (($row$05$i) + ($193))|0; //@line 992 "load_pat.cpp"
        $195 = ($196|0)>(0); //@line 993 "load_pat.cpp"
        L83: do {
         if ($195) {
          $197 = ($row$05$i|0)<(63); //@line 1015 "load_pat.cpp"
          if ($197) {
           $201 = $196;$ch$04$us$i = 0;
          } else {
           $239 = $196;$ch$04$i = 0;
           while(1) {
            $236 = $ch$04$i << 4; //@line 994 "load_pat.cpp"
            $237 = (($194) - ($236))|0; //@line 994 "load_pat.cpp"
            $238 = Math_imul($239, $row$05$i)|0; //@line 995 "load_pat.cpp"
            $240 = (($238) + ($ch$04$i))|0; //@line 995 "load_pat.cpp"
            $241 = HEAP32[$190>>2]|0; //@line 995 "load_pat.cpp"
            $242 = (((($241) + (($240*6)|0)|0)) + 5|0); //@line 996 "load_pat.cpp"
            HEAP8[$242>>0] = 0; //@line 996 "load_pat.cpp"
            $243 = (((($241) + (($240*6)|0)|0)) + 3|0); //@line 997 "load_pat.cpp"
            HEAP8[$243>>0] = 0; //@line 997 "load_pat.cpp"
            $244 = ($237|0)>(-1); //@line 998 "load_pat.cpp"
            do {
             if ($244) {
              $245 = (($239) + 3)|0; //@line 999 "load_pat.cpp"
              $246 = (($245) - ($ch$04$i))|0; //@line 999 "load_pat.cpp"
              $247 = $246 << 4; //@line 999 "load_pat.cpp"
              $248 = $247 & 48; //@line 999 "load_pat.cpp"
              $249 = (($186) - ($248))|0; //@line 999 "load_pat.cpp"
              $250 = ($194|0)<($249|0); //@line 1000 "load_pat.cpp"
              if (!($250)) {
               $265 = ($194|0)==($249|0); //@line 1020 "load_pat.cpp"
               if (!($265)) {
                break;
               }
               HEAP8[$242>>0] = 0; //@line 1021 "load_pat.cpp"
               HEAP8[$243>>0] = 25; //@line 1022 "load_pat.cpp"
               $266 = (((($241) + (($240*6)|0)|0)) + 2|0); //@line 1023 "load_pat.cpp"
               HEAP8[$266>>0] = 1; //@line 1023 "load_pat.cpp"
               $267 = (((($241) + (($240*6)|0)|0)) + 4|0); //@line 1024 "load_pat.cpp"
               HEAP8[$267>>0] = 0; //@line 1024 "load_pat.cpp"
               break;
              }
              $251 = (($237|0) % 64)&-1; //@line 1001 "load_pat.cpp"
              $252 = (529756 + ($251)|0); //@line 1002 "load_pat.cpp"
              $253 = HEAP8[$252>>0]|0; //@line 1002 "load_pat.cpp"
              $254 = $253 << 24 >> 24; //@line 1002 "load_pat.cpp"
              $255 = (_isalpha($254)|0); //@line 1002 "load_pat.cpp"
              $256 = ($255|0)==(0); //@line 1002 "load_pat.cpp"
              if ($256) {
               break;
              }
              do {
               switch ($254|0) {
               case 67:  {
                $$0$i$i = 61;
                break;
               }
               case 68:  {
                $$0$i$i = 63;
                break;
               }
               case 69:  {
                $$0$i$i = 65;
                break;
               }
               case 70:  {
                $$0$i$i = 66;
                break;
               }
               case 71:  {
                $$0$i$i = 68;
                break;
               }
               case 65:  {
                $$0$i$i = 70;
                break;
               }
               case 66:  {
                $$0$i$i = 72;
                break;
               }
               case 99:  {
                $$0$i$i = 73;
                break;
               }
               case 100:  {
                $$0$i$i = 75;
                break;
               }
               case 101:  {
                $$0$i$i = 77;
                break;
               }
               case 102:  {
                $$0$i$i = 78;
                break;
               }
               case 103:  {
                $$0$i$i = 80;
                break;
               }
               case 97:  {
                $$0$i$i = 82;
                break;
               }
               case 98:  {
                $$0$i$i = 84;
                break;
               }
               default: {
                $$0$i$i = 13;
               }
               }
              } while(0);
              $257 = (($ch$04$i) + 1)|0; //@line 1004 "load_pat.cpp"
              $258 = $257&255; //@line 1004 "load_pat.cpp"
              $259 = $251 & 15; //@line 1006 "load_pat.cpp"
              $260 = ($259|0)==(0); //@line 1006 "load_pat.cpp"
              $$$i4 = $260 ? 44 : 40; //@line 1006 "load_pat.cpp"
              $261 = (((($241) + (($240*6)|0)|0)) + 1|0); //@line 1010 "load_pat.cpp"
              HEAP8[$261>>0] = $258; //@line 1010 "load_pat.cpp"
              $262 = (($241) + (($240*6)|0)|0); //@line 1011 "load_pat.cpp"
              HEAP8[$262>>0] = $$0$i$i; //@line 1011 "load_pat.cpp"
              $263 = (((($241) + (($240*6)|0)|0)) + 2|0); //@line 1012 "load_pat.cpp"
              HEAP8[$263>>0] = 1; //@line 1012 "load_pat.cpp"
              $264 = (((($241) + (($240*6)|0)|0)) + 4|0); //@line 1013 "load_pat.cpp"
              HEAP8[$264>>0] = $$$i4; //@line 1013 "load_pat.cpp"
             }
            } while(0);
            $268 = (($ch$04$i) + 1)|0; //@line 993 "load_pat.cpp"
            $269 = HEAPU8[$40>>0]|(HEAPU8[$40+1>>0]<<8)|(HEAPU8[$40+2>>0]<<16)|(HEAPU8[$40+3>>0]<<24); //@line 993 "load_pat.cpp"
            $270 = ($268|0)<($269|0); //@line 993 "load_pat.cpp"
            if ($270) {
             $239 = $269;$ch$04$i = $268;
            } else {
             $286 = $269;
             break L83;
            }
           }
          }
          while(1) {
           $198 = $ch$04$us$i << 4; //@line 994 "load_pat.cpp"
           $199 = (($194) - ($198))|0; //@line 994 "load_pat.cpp"
           $200 = Math_imul($201, $row$05$i)|0; //@line 995 "load_pat.cpp"
           $202 = (($200) + ($ch$04$us$i))|0; //@line 995 "load_pat.cpp"
           $203 = HEAP32[$190>>2]|0; //@line 995 "load_pat.cpp"
           $204 = (((($203) + (($202*6)|0)|0)) + 5|0); //@line 996 "load_pat.cpp"
           HEAP8[$204>>0] = 0; //@line 996 "load_pat.cpp"
           $205 = (((($203) + (($202*6)|0)|0)) + 3|0); //@line 997 "load_pat.cpp"
           HEAP8[$205>>0] = 0; //@line 997 "load_pat.cpp"
           $206 = ($199|0)>(-1); //@line 998 "load_pat.cpp"
           do {
            if ($206) {
             $207 = (($201) + 3)|0; //@line 999 "load_pat.cpp"
             $208 = (($207) - ($ch$04$us$i))|0; //@line 999 "load_pat.cpp"
             $209 = $208 << 4; //@line 999 "load_pat.cpp"
             $210 = $209 & 48; //@line 999 "load_pat.cpp"
             $211 = (($186) - ($210))|0; //@line 999 "load_pat.cpp"
             $212 = ($194|0)<($211|0); //@line 1000 "load_pat.cpp"
             if (!($212)) {
              $213 = ($194|0)==($211|0); //@line 1020 "load_pat.cpp"
              if (!($213)) {
               break;
              }
              HEAP8[$204>>0] = 0; //@line 1021 "load_pat.cpp"
              HEAP8[$205>>0] = 25; //@line 1022 "load_pat.cpp"
              $214 = (((($203) + (($202*6)|0)|0)) + 2|0); //@line 1023 "load_pat.cpp"
              HEAP8[$214>>0] = 1; //@line 1023 "load_pat.cpp"
              $215 = (((($203) + (($202*6)|0)|0)) + 4|0); //@line 1024 "load_pat.cpp"
              HEAP8[$215>>0] = 0; //@line 1024 "load_pat.cpp"
              break;
             }
             $216 = (($199|0) % 64)&-1; //@line 1001 "load_pat.cpp"
             $217 = (529756 + ($216)|0); //@line 1002 "load_pat.cpp"
             $218 = HEAP8[$217>>0]|0; //@line 1002 "load_pat.cpp"
             $219 = $218 << 24 >> 24; //@line 1002 "load_pat.cpp"
             $220 = (_isalpha($219)|0); //@line 1002 "load_pat.cpp"
             $221 = ($220|0)==(0); //@line 1002 "load_pat.cpp"
             if (!($221)) {
              do {
               switch ($219|0) {
               case 67:  {
                $$0$i$us$i = 61;
                break;
               }
               case 68:  {
                $$0$i$us$i = 63;
                break;
               }
               case 69:  {
                $$0$i$us$i = 65;
                break;
               }
               case 70:  {
                $$0$i$us$i = 66;
                break;
               }
               case 71:  {
                $$0$i$us$i = 68;
                break;
               }
               case 65:  {
                $$0$i$us$i = 70;
                break;
               }
               case 66:  {
                $$0$i$us$i = 72;
                break;
               }
               case 99:  {
                $$0$i$us$i = 73;
                break;
               }
               case 100:  {
                $$0$i$us$i = 75;
                break;
               }
               case 101:  {
                $$0$i$us$i = 77;
                break;
               }
               case 102:  {
                $$0$i$us$i = 78;
                break;
               }
               case 103:  {
                $$0$i$us$i = 80;
                break;
               }
               case 97:  {
                $$0$i$us$i = 82;
                break;
               }
               case 98:  {
                $$0$i$us$i = 84;
                break;
               }
               default: {
                $$0$i$us$i = 13;
               }
               }
              } while(0);
              $222 = (($ch$04$us$i) + 1)|0; //@line 1004 "load_pat.cpp"
              $223 = $222&255; //@line 1004 "load_pat.cpp"
              $224 = $216 & 15; //@line 1006 "load_pat.cpp"
              $225 = ($224|0)==(0); //@line 1006 "load_pat.cpp"
              $$$us$i = $225 ? 44 : 40; //@line 1006 "load_pat.cpp"
              $226 = (((($203) + (($202*6)|0)|0)) + 1|0); //@line 1010 "load_pat.cpp"
              HEAP8[$226>>0] = $223; //@line 1010 "load_pat.cpp"
              $227 = (($203) + (($202*6)|0)|0); //@line 1011 "load_pat.cpp"
              HEAP8[$227>>0] = $$0$i$us$i; //@line 1011 "load_pat.cpp"
              $228 = (((($203) + (($202*6)|0)|0)) + 2|0); //@line 1012 "load_pat.cpp"
              HEAP8[$228>>0] = 1; //@line 1012 "load_pat.cpp"
              $229 = (((($203) + (($202*6)|0)|0)) + 4|0); //@line 1013 "load_pat.cpp"
              HEAP8[$229>>0] = $$$us$i; //@line 1013 "load_pat.cpp"
             }
             $230 = (($211) + -1)|0; //@line 1015 "load_pat.cpp"
             $231 = ($194|0)==($230|0); //@line 1015 "load_pat.cpp"
             $232 = ($ch$04$us$i|0)==(0); //@line 1015 "load_pat.cpp"
             $or$cond$us$i = $232 & $231; //@line 1015 "load_pat.cpp"
             if (!($or$cond$us$i)) {
              break;
             }
             HEAP8[$205>>0] = 14; //@line 1016 "load_pat.cpp"
            }
           } while(0);
           $233 = (($ch$04$us$i) + 1)|0; //@line 993 "load_pat.cpp"
           $234 = HEAPU8[$40>>0]|(HEAPU8[$40+1>>0]<<8)|(HEAPU8[$40+2>>0]<<16)|(HEAPU8[$40+3>>0]<<24); //@line 993 "load_pat.cpp"
           $235 = ($233|0)<($234|0); //@line 993 "load_pat.cpp"
           if ($235) {
            $201 = $234;$ch$04$us$i = $233;
           } else {
            $286 = $234;
            break;
           }
          }
         } else {
          $286 = $196;
         }
        } while(0);
        $271 = (($row$05$i) + 1)|0; //@line 991 "load_pat.cpp"
        $exitcond$i5 = ($271|0)==(64); //@line 991 "load_pat.cpp"
        if ($exitcond$i5) {
         $$lcssa40 = $286;
         break;
        } else {
         $196 = $286;$row$05$i = $271;
        }
       }
       $272 = (($pat$08$i) + 1)|0; //@line 987 "load_pat.cpp"
       $273 = ($272|0)<($$numpat$i|0); //@line 987 "load_pat.cpp"
       if ($273) {
        $188 = $$lcssa40;$pat$08$i = $272;
       } else {
        break;
       }
      }
     }
    } while(0);
    $274 = HEAP32[$59>>2]|0; //@line 1628 "load_pat.cpp"
    $275 = ($274|0)>(0); //@line 1628 "load_pat.cpp"
    if ($275) {
     $t$410 = 0;
     while(1) {
      $276 = (($t$410) + 2)|0; //@line 1629 "load_pat.cpp"
      $277 = (($276|0) % 5)&-1; //@line 1629 "load_pat.cpp"
      $278 = $277 << 5; //@line 1629 "load_pat.cpp"
      $279 = (($278) + 48)|0; //@line 1629 "load_pat.cpp"
      $280 = (((($this)) + 49600|0) + (($t$410*36)|0)|0); //@line 1629 "load_pat.cpp"
      HEAP32[$280>>2] = $279; //@line 1629 "load_pat.cpp"
      $281 = (((((($this)) + 49600|0) + (($t$410*36)|0)|0)) + 4|0); //@line 1630 "load_pat.cpp"
      HEAP32[$281>>2] = 64; //@line 1630 "load_pat.cpp"
      $282 = (($t$410) + 1)|0; //@line 1628 "load_pat.cpp"
      $283 = HEAP32[$59>>2]|0; //@line 1628 "load_pat.cpp"
      $284 = ($282|0)<($283|0); //@line 1628 "load_pat.cpp"
      if ($284) {
       $t$410 = $282;
      } else {
       break;
      }
     }
    }
    HEAP32[300432] = 0; //@line 1633 "load_pat.cpp"
    _free($6); //@line 875 "load_pat.cpp"
    $$2 = 1;
    STACKTOP = sp;return ($$2|0); //@line 1638 "load_pat.cpp"
   }
  }
 }
 $$2 = 0;
 STACKTOP = sp;return ($$2|0); //@line 1638 "load_pat.cpp"
}
function __ZL18pat_get_waveheaderP6MMFILEP10WaveHeaderi($mmpat,$hw,$layer) {
 $mmpat = $mmpat|0;
 $hw = $hw|0;
 $layer = $layer|0;
 var $$bestpos$0 = 0, $$bestpos$0$lcssa = 0, $$layer = 0, $$lcssa = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$i = 0, $$pre11 = 0, $$pre12 = 0, $$pre15 = 0, $$pre17 = 0, $$sz$i = 0, $$sz$i$i = 0, $$sz$i1 = 0, $$sz$i1$i = 0, $$sz$i3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0;
 var $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $9 = 0, $bestfreq$06 = 0, $bestpos$05 = 0, $freqdist$0 = 0, $freqdist$0$bestfreq$0 = 0, $hl = 0, $i$08 = 0, $i$17 = 0, $ih$i = 0, $mmpat$idx$val = 0, $mmpat$idx$val$pre = 0, $mmpat$idx$val13 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ih$i = sp + 47|0;
 $hl = sp;
 $0 = ((($mmpat)) + 8|0); //@line 336 "load_pat.cpp"
 $1 = ((($mmpat)) + 4|0); //@line 348 "load_pat.cpp"
 $2 = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24); //@line 348 "load_pat.cpp"
 $3 = ($2|0)>(129); //@line 348 "load_pat.cpp"
 if ($3) {
  HEAP8[$0>>0]=129&255;HEAP8[$0+1>>0]=(129>>8)&255;HEAP8[$0+2>>0]=(129>>16)&255;HEAP8[$0+3>>0]=129>>24; //@line 349 "load_pat.cpp"
  $6 = 129;
 } else {
  $4 = ((($mmpat)) + 12|0); //@line 351 "load_pat.cpp"
  HEAP8[$4>>0]=1&255;HEAP8[$4+1>>0]=(1>>8)&255;HEAP8[$4+2>>0]=(1>>16)&255;HEAP8[$4+3>>0]=1>>24; //@line 351 "load_pat.cpp"
  $$pre$i = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 359 "load_pat.cpp"
  $6 = $$pre$i;
 }
 $5 = (($2) - ($6))|0; //@line 359 "load_pat.cpp"
 $7 = ($5|0)<(63); //@line 359 "load_pat.cpp"
 $$sz$i1$i = $7 ? $5 : 63; //@line 359 "load_pat.cpp"
 $8 = HEAPU8[$mmpat>>0]|(HEAPU8[$mmpat+1>>0]<<8)|(HEAPU8[$mmpat+2>>0]<<16)|(HEAPU8[$mmpat+3>>0]<<24); //@line 361 "load_pat.cpp"
 $9 = (($8) + ($6)|0); //@line 361 "load_pat.cpp"
 _memcpy(($ih$i|0),($9|0),($$sz$i1$i|0))|0; //@line 361 "load_pat.cpp"
 $10 = (($$sz$i1$i) + ($6))|0; //@line 362 "load_pat.cpp"
 HEAP8[$0>>0]=$10&255;HEAP8[$0+1>>0]=($10>>8)&255;HEAP8[$0+2>>0]=($10>>16)&255;HEAP8[$0+3>>0]=$10>>24; //@line 362 "load_pat.cpp"
 $11 = (($2) - ($10))|0; //@line 359 "load_pat.cpp"
 $12 = ($11|0)<(47); //@line 359 "load_pat.cpp"
 $$sz$i$i = $12 ? $11 : 47; //@line 359 "load_pat.cpp"
 $13 = (($8) + ($10)|0); //@line 361 "load_pat.cpp"
 _memcpy(($hl|0),($13|0),($$sz$i$i|0))|0; //@line 361 "load_pat.cpp"
 $14 = (($$sz$i$i) + ($10))|0; //@line 362 "load_pat.cpp"
 HEAP8[$0>>0]=$14&255;HEAP8[$0+1>>0]=($14>>8)&255;HEAP8[$0+2>>0]=($14>>16)&255;HEAP8[$0+3>>0]=$14>>24; //@line 362 "load_pat.cpp"
 $15 = ((($hl)) + 7|0); //@line 522 "load_pat.cpp"
 $16 = ((($ih$i)) + 2|0); //@line 522 "load_pat.cpp"
 (_strncpy($15,$16,40)|0); //@line 522 "load_pat.cpp"
 $17 = ((($hl)) + 6|0); //@line 585 "load_pat.cpp"
 $18 = HEAP8[$17>>0]|0; //@line 585 "load_pat.cpp"
 $19 = $18&255; //@line 585 "load_pat.cpp"
 $20 = ($18&255)>(1); //@line 585 "load_pat.cpp"
 L5: do {
  if ($20) {
   $21 = ($layer|0)==(0); //@line 586 "load_pat.cpp"
   if ($21) {
    $22 = ((($hw)) + 30|0); //@line 602 "load_pat.cpp"
    $23 = ((($hw)) + 8|0); //@line 610 "load_pat.cpp"
    $24 = ((($mmpat)) + 12|0); //@line 351 "load_pat.cpp"
    $mmpat$idx$val$pre = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24);
    $$pre15 = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24); //@line 359 "load_pat.cpp"
    $45 = $$pre15;$bestfreq$06 = 523251000;$bestpos$05 = 0;$i$17 = 0;$mmpat$idx$val = $mmpat$idx$val$pre;
    while(1) {
     $44 = (($45) - ($mmpat$idx$val))|0; //@line 359 "load_pat.cpp"
     $46 = ($44|0)<(96); //@line 359 "load_pat.cpp"
     $$sz$i3 = $46 ? $44 : 96; //@line 359 "load_pat.cpp"
     $47 = HEAPU8[$mmpat>>0]|(HEAPU8[$mmpat+1>>0]<<8)|(HEAPU8[$mmpat+2>>0]<<16)|(HEAPU8[$mmpat+3>>0]<<24); //@line 361 "load_pat.cpp"
     $48 = (($47) + ($mmpat$idx$val)|0); //@line 361 "load_pat.cpp"
     _memcpy(($hw|0),($48|0),($$sz$i3|0))|0; //@line 361 "load_pat.cpp"
     $49 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 362 "load_pat.cpp"
     $50 = (($49) + ($$sz$i3))|0; //@line 362 "load_pat.cpp"
     HEAP8[$0>>0]=$50&255;HEAP8[$0+1>>0]=($50>>8)&255;HEAP8[$0+2>>0]=($50>>16)&255;HEAP8[$0+3>>0]=$50>>24; //@line 362 "load_pat.cpp"
     $51 = HEAPU8[$22>>0]|(HEAPU8[$22+1>>0]<<8)|(HEAPU8[$22+2>>0]<<16)|(HEAPU8[$22+3>>0]<<24); //@line 602 "load_pat.cpp"
     $52 = ($51>>>0)>(523251); //@line 602 "load_pat.cpp"
     $53 = (($51) + -523251)|0; //@line 603 "load_pat.cpp"
     $54 = (523251 - ($51))|0; //@line 605 "load_pat.cpp"
     $55 = $54 << 1; //@line 605 "load_pat.cpp"
     $freqdist$0 = $52 ? $53 : $55; //@line 602 "load_pat.cpp"
     $56 = ($freqdist$0>>>0)<($bestfreq$06>>>0); //@line 606 "load_pat.cpp"
     $$bestpos$0 = $56 ? $mmpat$idx$val : $bestpos$05; //@line 606 "load_pat.cpp"
     $freqdist$0$bestfreq$0 = $56 ? $freqdist$0 : $bestfreq$06; //@line 606 "load_pat.cpp"
     $57 = HEAPU8[$23>>0]|(HEAPU8[$23+1>>0]<<8)|(HEAPU8[$23+2>>0]<<16)|(HEAPU8[$23+3>>0]<<24); //@line 610 "load_pat.cpp"
     $58 = (($57) + ($50))|0; //@line 342 "load_pat.cpp"
     $59 = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24); //@line 348 "load_pat.cpp"
     $60 = ($58|0)<($59|0); //@line 348 "load_pat.cpp"
     if ($60) {
      HEAP8[$0>>0]=$58&255;HEAP8[$0+1>>0]=($58>>8)&255;HEAP8[$0+2>>0]=($58>>16)&255;HEAP8[$0+3>>0]=$58>>24; //@line 349 "load_pat.cpp"
      $mmpat$idx$val13 = $58;
     } else {
      HEAP8[$24>>0]=1&255;HEAP8[$24+1>>0]=(1>>8)&255;HEAP8[$24+2>>0]=(1>>16)&255;HEAP8[$24+3>>0]=1>>24; //@line 351 "load_pat.cpp"
      $mmpat$idx$val13 = $50;
     }
     $61 = (($i$17) + 1)|0; //@line 599 "load_pat.cpp"
     $62 = ($61|0)<($19|0); //@line 599 "load_pat.cpp"
     if ($62) {
      $45 = $59;$bestfreq$06 = $freqdist$0$bestfreq$0;$bestpos$05 = $$bestpos$0;$i$17 = $61;$mmpat$idx$val = $mmpat$idx$val13;
     } else {
      $$bestpos$0$lcssa = $$bestpos$0;$$lcssa = $59;
      break;
     }
    }
    $63 = ($$bestpos$0$lcssa|0)<($$lcssa|0); //@line 348 "load_pat.cpp"
    if ($63) {
     HEAP8[$0>>0]=$$bestpos$0$lcssa&255;HEAP8[$0+1>>0]=($$bestpos$0$lcssa>>8)&255;HEAP8[$0+2>>0]=($$bestpos$0$lcssa>>16)&255;HEAP8[$0+3>>0]=$$bestpos$0$lcssa>>24; //@line 349 "load_pat.cpp"
     break;
    } else {
     $64 = ((($mmpat)) + 12|0); //@line 351 "load_pat.cpp"
     HEAP8[$64>>0]=1&255;HEAP8[$64+1>>0]=(1>>8)&255;HEAP8[$64+2>>0]=(1>>16)&255;HEAP8[$64+3>>0]=1>>24; //@line 351 "load_pat.cpp"
     break;
    }
   } else {
    $25 = ($19|0)<($layer|0); //@line 587 "load_pat.cpp"
    $$layer = $25 ? $19 : $layer; //@line 587 "load_pat.cpp"
    $26 = ($$layer|0)>(1); //@line 588 "load_pat.cpp"
    if (!($26)) {
     break;
    }
    $27 = ((($hw)) + 8|0); //@line 590 "load_pat.cpp"
    $28 = ((($mmpat)) + 12|0); //@line 591 "load_pat.cpp"
    $$pre = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24); //@line 359 "load_pat.cpp"
    $$pre11 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 359 "load_pat.cpp"
    $32 = $$pre;$33 = $$pre11;$i$08 = 1;
    while(1) {
     $31 = (($32) - ($33))|0; //@line 359 "load_pat.cpp"
     $34 = ($31|0)<(96); //@line 359 "load_pat.cpp"
     $$sz$i = $34 ? $31 : 96; //@line 359 "load_pat.cpp"
     $35 = HEAPU8[$mmpat>>0]|(HEAPU8[$mmpat+1>>0]<<8)|(HEAPU8[$mmpat+2>>0]<<16)|(HEAPU8[$mmpat+3>>0]<<24); //@line 361 "load_pat.cpp"
     $36 = (($35) + ($33)|0); //@line 361 "load_pat.cpp"
     _memcpy(($hw|0),($36|0),($$sz$i|0))|0; //@line 361 "load_pat.cpp"
     $37 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 362 "load_pat.cpp"
     $38 = (($37) + ($$sz$i))|0; //@line 362 "load_pat.cpp"
     HEAP8[$0>>0]=$38&255;HEAP8[$0+1>>0]=($38>>8)&255;HEAP8[$0+2>>0]=($38>>16)&255;HEAP8[$0+3>>0]=$38>>24; //@line 362 "load_pat.cpp"
     $39 = HEAPU8[$27>>0]|(HEAPU8[$27+1>>0]<<8)|(HEAPU8[$27+2>>0]<<16)|(HEAPU8[$27+3>>0]<<24); //@line 590 "load_pat.cpp"
     $40 = (($38) + ($39))|0; //@line 342 "load_pat.cpp"
     $41 = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24); //@line 348 "load_pat.cpp"
     $42 = ($40|0)<($41|0); //@line 348 "load_pat.cpp"
     if (!($42)) {
      label = 11;
      break;
     }
     HEAP8[$0>>0]=$40&255;HEAP8[$0+1>>0]=($40>>8)&255;HEAP8[$0+2>>0]=($40>>16)&255;HEAP8[$0+3>>0]=$40>>24; //@line 349 "load_pat.cpp"
     $$pre12 = HEAPU8[$28>>0]|(HEAPU8[$28+1>>0]<<8)|(HEAPU8[$28+2>>0]<<16)|(HEAPU8[$28+3>>0]<<24); //@line 591 "load_pat.cpp"
     $43 = ($$pre12|0)==(0); //@line 591 "load_pat.cpp"
     $30 = (($i$08) + 1)|0; //@line 588 "load_pat.cpp"
     if (!($43)) {
      break;
     }
     $29 = ($30|0)<($$layer|0); //@line 588 "load_pat.cpp"
     if ($29) {
      $32 = $41;$33 = $40;$i$08 = $30;
     } else {
      break L5;
     }
    }
    if ((label|0) == 11) {
     HEAP8[$28>>0]=1&255;HEAP8[$28+1>>0]=(1>>8)&255;HEAP8[$28+2>>0]=(1>>16)&255;HEAP8[$28+3>>0]=1>>24; //@line 351 "load_pat.cpp"
    }
    HEAP8[$27>>0]=0&255;HEAP8[$27+1>>0]=(0>>8)&255;HEAP8[$27+2>>0]=(0>>16)&255;HEAP8[$27+3>>0]=0>>24; //@line 592 "load_pat.cpp"
    STACKTOP = sp;return; //@line 623 "load_pat.cpp"
   }
  }
 } while(0);
 $65 = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24); //@line 359 "load_pat.cpp"
 $66 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 359 "load_pat.cpp"
 $67 = (($65) - ($66))|0; //@line 359 "load_pat.cpp"
 $68 = ($67|0)<(96); //@line 359 "load_pat.cpp"
 $$sz$i1 = $68 ? $67 : 96; //@line 359 "load_pat.cpp"
 $69 = HEAPU8[$mmpat>>0]|(HEAPU8[$mmpat+1>>0]<<8)|(HEAPU8[$mmpat+2>>0]<<16)|(HEAPU8[$mmpat+3>>0]<<24); //@line 361 "load_pat.cpp"
 $70 = (($69) + ($66)|0); //@line 361 "load_pat.cpp"
 _memcpy(($hw|0),($70|0),($$sz$i1|0))|0; //@line 361 "load_pat.cpp"
 $71 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24); //@line 362 "load_pat.cpp"
 $72 = (($71) + ($$sz$i1))|0; //@line 362 "load_pat.cpp"
 HEAP8[$0>>0]=$72&255;HEAP8[$0+1>>0]=($72>>8)&255;HEAP8[$0+2>>0]=($72>>16)&255;HEAP8[$0+3>>0]=$72>>24; //@line 362 "load_pat.cpp"
 $73 = ((($hw)) + 12|0); //@line 616 "load_pat.cpp"
 $74 = HEAPU8[$73>>0]|(HEAPU8[$73+1>>0]<<8)|(HEAPU8[$73+2>>0]<<16)|(HEAPU8[$73+3>>0]<<24); //@line 616 "load_pat.cpp"
 $75 = ((($hw)) + 8|0); //@line 616 "load_pat.cpp"
 $76 = HEAPU8[$75>>0]|(HEAPU8[$75+1>>0]<<8)|(HEAPU8[$75+2>>0]<<16)|(HEAPU8[$75+3>>0]<<24); //@line 616 "load_pat.cpp"
 $77 = ($74>>>0)<($76>>>0); //@line 616 "load_pat.cpp"
 if (!($77)) {
  HEAP8[$73>>0]=0&255;HEAP8[$73+1>>0]=(0>>8)&255;HEAP8[$73+2>>0]=(0>>16)&255;HEAP8[$73+3>>0]=0>>24; //@line 617 "load_pat.cpp"
  $78 = ((($hw)) + 16|0); //@line 618 "load_pat.cpp"
  HEAP8[$78>>0]=0&255;HEAP8[$78+1>>0]=(0>>8)&255;HEAP8[$78+2>>0]=(0>>16)&255;HEAP8[$78+3>>0]=0>>24; //@line 618 "load_pat.cpp"
  $79 = ((($hw)) + 55|0); //@line 619 "load_pat.cpp"
  $80 = HEAP8[$79>>0]|0; //@line 619 "load_pat.cpp"
  $81 = $80&255; //@line 619 "load_pat.cpp"
  $82 = $81 & 251; //@line 619 "load_pat.cpp"
  $83 = $82&255; //@line 619 "load_pat.cpp"
  HEAP8[$79>>0] = $83; //@line 619 "load_pat.cpp"
  STACKTOP = sp;return; //@line 623 "load_pat.cpp"
 }
 $$phi$trans$insert = ((($hw)) + 16|0);
 $$pre17 = HEAPU8[$$phi$trans$insert>>0]|(HEAPU8[$$phi$trans$insert+1>>0]<<8)|(HEAPU8[$$phi$trans$insert+2>>0]<<16)|(HEAPU8[$$phi$trans$insert+3>>0]<<24); //@line 621 "load_pat.cpp"
 $84 = ($$pre17>>>0)>($76>>>0); //@line 621 "load_pat.cpp"
 if (!($84)) {
  STACKTOP = sp;return; //@line 623 "load_pat.cpp"
 }
 HEAP8[$$phi$trans$insert>>0]=$76&255;HEAP8[$$phi$trans$insert+1>>0]=($76>>8)&255;HEAP8[$$phi$trans$insert+2>>0]=($76>>16)&255;HEAP8[$$phi$trans$insert+3>>0]=$76>>24; //@line 622 "load_pat.cpp"
 STACKTOP = sp;return; //@line 623 "load_pat.cpp"
}
function __ZN10CSoundFile7ReadSTMEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$0 = 0, $$35 = 0, $$8 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $dwMemPos$022 = 0, $dwMemPos$117 = 0, $dwMemPos$2 = 0, $dwMemPos$3 = 0, $exitcond = 0, $exitcond27 = 0, $exitcond28 = 0, $exitcond29 = 0, $m$018 = 0, $n$020 = 0, $nIns$024 = 0, $nOrd$023 = 0, $nPat$021 = 0, $nSmp$016 = 0, $or$cond = 0, $or$cond11 = 0, $or$cond34 = 0, $or$cond9 = 0, $p$019 = 0;
 var $storemerge6 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ($lpStream|0)==(0|0); //@line 65 "load_stm.cpp"
 $1 = ($dwMemLength>>>0)<(1168); //@line 65 "load_stm.cpp"
 $or$cond = $0 | $1; //@line 65 "load_stm.cpp"
 if ($or$cond) {
  $$0 = 0;
  return ($$0|0); //@line 185 "load_stm.cpp"
 }
 $2 = ((($lpStream)) + 29|0); //@line 66 "load_stm.cpp"
 $3 = HEAP8[$2>>0]|0; //@line 66 "load_stm.cpp"
 $4 = ($3<<24>>24)==(2); //@line 66 "load_stm.cpp"
 if (!($4)) {
  $$0 = 0;
  return ($$0|0); //@line 185 "load_stm.cpp"
 }
 $5 = ((($lpStream)) + 28|0); //@line 66 "load_stm.cpp"
 $6 = HEAP8[$5>>0]|0; //@line 66 "load_stm.cpp"
 $7 = ($6<<24>>24)==(26); //@line 66 "load_stm.cpp"
 if (!($7)) {
  $$0 = 0;
  return ($$0|0); //@line 185 "load_stm.cpp"
 }
 $8 = ((($lpStream)) + 20|0); //@line 67 "load_stm.cpp"
 $9 = (_strncasecmp($8,529821,8)|0); //@line 67 "load_stm.cpp"
 $10 = ($9|0)==(0); //@line 67 "load_stm.cpp"
 if (!($10)) {
  $11 = (_strncasecmp($8,529830,8)|0); //@line 68 "load_stm.cpp"
  $12 = ($11|0)==(0); //@line 68 "load_stm.cpp"
  if (!($12)) {
   $$0 = 0;
   return ($$0|0); //@line 185 "load_stm.cpp"
  }
 }
 $13 = ((($this)) + 59804|0); //@line 69 "load_stm.cpp"
 dest=$13; src=$lpStream; stop=dest+20|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 69 "load_stm.cpp"
 $14 = ((($this)) + 59680|0); //@line 71 "load_stm.cpp"
 HEAP32[$14>>2] = 256; //@line 71 "load_stm.cpp"
 $15 = ((($this)) + 59684|0); //@line 72 "load_stm.cpp"
 HEAP32[$15>>2] = 31; //@line 72 "load_stm.cpp"
 $16 = ((($this)) + 59664|0); //@line 73 "load_stm.cpp"
 HEAP32[$16>>2] = 4; //@line 73 "load_stm.cpp"
 $17 = ((($this)) + 59688|0); //@line 74 "load_stm.cpp"
 HEAP32[$17>>2] = 0; //@line 74 "load_stm.cpp"
 $18 = ((($this)) + 59764|0); //@line 75 "load_stm.cpp"
 HEAP32[$18>>2] = 64; //@line 75 "load_stm.cpp"
 $19 = ((($this)) + 59768|0); //@line 76 "load_stm.cpp"
 HEAP32[$19>>2] = 32767; //@line 76 "load_stm.cpp"
 $20 = ((($lpStream)) + 32|0); //@line 77 "load_stm.cpp"
 $21 = HEAP8[$20>>0]|0; //@line 77 "load_stm.cpp"
 $22 = $21&255; //@line 77 "load_stm.cpp"
 $23 = $22 >>> 4; //@line 77 "load_stm.cpp"
 $24 = ((($this)) + 59648|0); //@line 77 "load_stm.cpp"
 $25 = ($23|0)==(0); //@line 78 "load_stm.cpp"
 $$ = $25 ? 1 : $23; //@line 78 "load_stm.cpp"
 HEAP32[$24>>2] = $$; //@line 77 "load_stm.cpp"
 $26 = ((($this)) + 59652|0); //@line 79 "load_stm.cpp"
 HEAP32[$26>>2] = 125; //@line 79 "load_stm.cpp"
 $27 = ((($lpStream)) + 34|0); //@line 80 "load_stm.cpp"
 $28 = HEAP8[$27>>0]|0; //@line 80 "load_stm.cpp"
 $29 = $28&255; //@line 80 "load_stm.cpp"
 $30 = $29 << 2; //@line 80 "load_stm.cpp"
 $31 = ((($this)) + 59656|0); //@line 80 "load_stm.cpp"
 $32 = ($30>>>0)>(256); //@line 81 "load_stm.cpp"
 $storemerge6 = $32 ? 256 : $30; //@line 81 "load_stm.cpp"
 HEAP32[$31>>2] = $storemerge6; //@line 80 "load_stm.cpp"
 $33 = ((($this)) + 53344|0); //@line 82 "load_stm.cpp"
 $34 = ((($lpStream)) + 1040|0); //@line 82 "load_stm.cpp"
 dest=$33; src=$34; stop=dest+128|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 82 "load_stm.cpp"
 $35 = ((($this)) + 49608|0); //@line 86 "load_stm.cpp"
 HEAP32[$35>>2] = 0; //@line 86 "load_stm.cpp"
 $36 = ((($this)) + 49604|0); //@line 87 "load_stm.cpp"
 HEAP32[$36>>2] = 64; //@line 87 "load_stm.cpp"
 $37 = ((($this)) + 49600|0); //@line 88 "load_stm.cpp"
 HEAP32[$37>>2] = 192; //@line 88 "load_stm.cpp"
 $38 = ((($this)) + 49644|0); //@line 86 "load_stm.cpp"
 HEAP32[$38>>2] = 0; //@line 86 "load_stm.cpp"
 $39 = ((($this)) + 49640|0); //@line 87 "load_stm.cpp"
 HEAP32[$39>>2] = 64; //@line 87 "load_stm.cpp"
 $40 = ((($this)) + 49636|0); //@line 88 "load_stm.cpp"
 HEAP32[$40>>2] = 64; //@line 88 "load_stm.cpp"
 $41 = ((($this)) + 49680|0); //@line 86 "load_stm.cpp"
 HEAP32[$41>>2] = 0; //@line 86 "load_stm.cpp"
 $42 = ((($this)) + 49676|0); //@line 87 "load_stm.cpp"
 HEAP32[$42>>2] = 64; //@line 87 "load_stm.cpp"
 $43 = ((($this)) + 49672|0); //@line 88 "load_stm.cpp"
 HEAP32[$43>>2] = 192; //@line 88 "load_stm.cpp"
 $44 = ((($this)) + 49716|0); //@line 86 "load_stm.cpp"
 HEAP32[$44>>2] = 0; //@line 86 "load_stm.cpp"
 $45 = ((($this)) + 49712|0); //@line 87 "load_stm.cpp"
 HEAP32[$45>>2] = 64; //@line 87 "load_stm.cpp"
 $46 = ((($this)) + 49708|0); //@line 88 "load_stm.cpp"
 HEAP32[$46>>2] = 64; //@line 88 "load_stm.cpp"
 $47 = ((($lpStream)) + 48|0); //@line 94 "load_stm.cpp"
 $nIns$024 = 0;
 while(1) {
  $48 = (($nIns$024) + 1)|0; //@line 93 "load_stm.cpp"
  $49 = (((((($this)) + 33280|0) + ($48<<6)|0)) + 42|0); //@line 95 "load_stm.cpp"
  $50 = (($47) + ($nIns$024<<5)|0); //@line 95 "load_stm.cpp"
  dest=$49; src=$50; stop=dest+13|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 95 "load_stm.cpp"
  $51 = (((($this)) + 59804|0) + ($48<<5)|0); //@line 96 "load_stm.cpp"
  dest=$51; src=$50; stop=dest+12|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 96 "load_stm.cpp"
  $52 = (((($47) + ($nIns$024<<5)|0)) + 24|0); //@line 97 "load_stm.cpp"
  $53 = HEAPU8[$52>>0]|(HEAPU8[$52+1>>0]<<8); //@line 97 "load_stm.cpp"
  $54 = $53&65535; //@line 97 "load_stm.cpp"
  $55 = (((((($this)) + 33280|0) + ($48<<6)|0)) + 24|0); //@line 97 "load_stm.cpp"
  HEAP32[$55>>2] = $54; //@line 97 "load_stm.cpp"
  $56 = (((((($this)) + 33280|0) + ($48<<6)|0)) + 32|0); //@line 98 "load_stm.cpp"
  HEAP16[$56>>1] = 64; //@line 98 "load_stm.cpp"
  $57 = (((($47) + ($nIns$024<<5)|0)) + 22|0); //@line 99 "load_stm.cpp"
  $58 = HEAP8[$57>>0]|0; //@line 99 "load_stm.cpp"
  $59 = $58&255; //@line 99 "load_stm.cpp"
  $60 = $59 << 2; //@line 99 "load_stm.cpp"
  $61 = $60&65535; //@line 99 "load_stm.cpp"
  $62 = (((((($this)) + 33280|0) + ($48<<6)|0)) + 30|0); //@line 99 "load_stm.cpp"
  $63 = ($60>>>0)>(256); //@line 100 "load_stm.cpp"
  $$8 = $63 ? 256 : $61; //@line 100 "load_stm.cpp"
  HEAP16[$62>>1] = $$8; //@line 99 "load_stm.cpp"
  $64 = (((($47) + ($nIns$024<<5)|0)) + 16|0); //@line 101 "load_stm.cpp"
  $65 = HEAPU8[$64>>0]|(HEAPU8[$64+1>>0]<<8); //@line 101 "load_stm.cpp"
  $66 = $65&65535; //@line 101 "load_stm.cpp"
  $67 = (((($this)) + 33280|0) + ($48<<6)|0); //@line 101 "load_stm.cpp"
  $68 = ($65&65535)<(4); //@line 102 "load_stm.cpp"
  $69 = ($$8<<16>>16)==(0); //@line 102 "load_stm.cpp"
  $or$cond34 = $68 | $69; //@line 102 "load_stm.cpp"
  $$35 = $or$cond34 ? 0 : $66; //@line 102 "load_stm.cpp"
  HEAP32[$67>>2] = $$35; //@line 102 "load_stm.cpp"
  $70 = (((($47) + ($nIns$024<<5)|0)) + 18|0); //@line 103 "load_stm.cpp"
  $71 = HEAPU8[$70>>0]|(HEAPU8[$70+1>>0]<<8); //@line 103 "load_stm.cpp"
  $72 = $71&65535; //@line 103 "load_stm.cpp"
  $73 = (((((($this)) + 33280|0) + ($48<<6)|0)) + 4|0); //@line 103 "load_stm.cpp"
  HEAP32[$73>>2] = $72; //@line 103 "load_stm.cpp"
  $74 = (((($47) + ($nIns$024<<5)|0)) + 20|0); //@line 104 "load_stm.cpp"
  $75 = HEAPU8[$74>>0]|(HEAPU8[$74+1>>0]<<8); //@line 104 "load_stm.cpp"
  $76 = $75&65535; //@line 104 "load_stm.cpp"
  $77 = (((((($this)) + 33280|0) + ($48<<6)|0)) + 8|0); //@line 104 "load_stm.cpp"
  HEAP32[$77>>2] = $76; //@line 104 "load_stm.cpp"
  $78 = ($75&65535)<=($71&65535); //@line 105 "load_stm.cpp"
  $79 = ($75<<16>>16)==(-1); //@line 105 "load_stm.cpp"
  $or$cond9 = $79 | $78; //@line 105 "load_stm.cpp"
  if (!($or$cond9)) {
   $80 = (((((($this)) + 33280|0) + ($48<<6)|0)) + 34|0); //@line 105 "load_stm.cpp"
   $81 = HEAP16[$80>>1]|0; //@line 105 "load_stm.cpp"
   $82 = $81&65535; //@line 105 "load_stm.cpp"
   $83 = $82 | 2; //@line 105 "load_stm.cpp"
   $84 = $83&65535; //@line 105 "load_stm.cpp"
   HEAP16[$80>>1] = $84; //@line 105 "load_stm.cpp"
  }
  $exitcond29 = ($48|0)==(31); //@line 91 "load_stm.cpp"
  if ($exitcond29) {
   $nOrd$023 = 0;
   break;
  } else {
   $nIns$024 = $48;
  }
 }
 while(1) {
  $85 = (((($this)) + 53344|0) + ($nOrd$023)|0); //@line 108 "load_stm.cpp"
  $86 = HEAP8[$85>>0]|0; //@line 108 "load_stm.cpp"
  $87 = ($86&255)>(98); //@line 108 "load_stm.cpp"
  if ($87) {
   HEAP8[$85>>0] = -1; //@line 108 "load_stm.cpp"
  }
  $88 = (($nOrd$023) + 1)|0; //@line 108 "load_stm.cpp"
  $exitcond28 = ($88|0)==(256); //@line 108 "load_stm.cpp"
  if ($exitcond28) {
   break;
  } else {
   $nOrd$023 = $88;
  }
 }
 $89 = ((($lpStream)) + 33|0); //@line 109 "load_stm.cpp"
 $90 = HEAP8[$89>>0]|0; //@line 109 "load_stm.cpp"
 $91 = $90&255; //@line 109 "load_stm.cpp"
 $92 = ($90<<24>>24)==(0); //@line 110 "load_stm.cpp"
 L25: do {
  if ($92) {
   $dwMemPos$117 = 1168;$nSmp$016 = 1;
  } else {
   $dwMemPos$022 = 1168;$nPat$021 = 0;
   while(1) {
    $93 = (($dwMemPos$022) + 1024)|0; //@line 112 "load_stm.cpp"
    $94 = ($93>>>0)>($dwMemLength>>>0); //@line 112 "load_stm.cpp"
    if ($94) {
     $$0 = 1;
     label = 45;
     break;
    }
    $95 = (((($this)) + 52864|0) + ($nPat$021<<1)|0); //@line 113 "load_stm.cpp"
    HEAP16[$95>>1] = 64; //@line 113 "load_stm.cpp"
    $96 = HEAP32[$16>>2]|0; //@line 114 "load_stm.cpp"
    $97 = (__ZN10CSoundFile15AllocatePatternEjj(64,$96)|0); //@line 114 "load_stm.cpp"
    $98 = (((($this)) + 51904|0) + ($nPat$021<<2)|0); //@line 114 "load_stm.cpp"
    HEAP32[$98>>2] = $97; //@line 114 "load_stm.cpp"
    $99 = ($97|0)==(0|0); //@line 114 "load_stm.cpp"
    if ($99) {
     $$0 = 1;
     label = 45;
     break;
    }
    $100 = (($lpStream) + ($dwMemPos$022)|0); //@line 116 "load_stm.cpp"
    $m$018 = $97;$n$020 = 0;$p$019 = $100;
    while(1) {
     $101 = HEAP8[$p$019>>0]|0; //@line 122 "load_stm.cpp"
     $102 = $101&255; //@line 122 "load_stm.cpp"
     $103 = ((($p$019)) + 1|0); //@line 123 "load_stm.cpp"
     $104 = HEAP8[$103>>0]|0; //@line 123 "load_stm.cpp"
     $105 = $104&255; //@line 123 "load_stm.cpp"
     $106 = $105 >>> 3; //@line 123 "load_stm.cpp"
     $107 = $105 & 7; //@line 124 "load_stm.cpp"
     $108 = ((($p$019)) + 2|0); //@line 124 "load_stm.cpp"
     $109 = HEAP8[$108>>0]|0; //@line 124 "load_stm.cpp"
     $110 = $109&255; //@line 124 "load_stm.cpp"
     $111 = $110 >>> 1; //@line 124 "load_stm.cpp"
     $112 = (($111) + ($107))|0; //@line 124 "load_stm.cpp"
     $113 = $110 & 15; //@line 125 "load_stm.cpp"
     $114 = ($106|0)==(0); //@line 126 "load_stm.cpp"
     if (!($114)) {
      $115 = $106&255; //@line 126 "load_stm.cpp"
      $116 = ((($m$018)) + 1|0); //@line 126 "load_stm.cpp"
      HEAP8[$116>>0] = $115; //@line 126 "load_stm.cpp"
     }
     $117 = $101 & -3; //@line 130 "load_stm.cpp"
     $118 = ($117<<24>>24)==(-4); //@line 130 "load_stm.cpp"
     if ($118) {
      HEAP8[$m$018>>0] = -2; //@line 130 "load_stm.cpp"
     } else {
      $119 = ($101&255)<(252); //@line 132 "load_stm.cpp"
      if ($119) {
       $120 = $102 >>> 4; //@line 132 "load_stm.cpp"
       $121 = ($120*12)|0; //@line 132 "load_stm.cpp"
       $122 = $102 & 15; //@line 132 "load_stm.cpp"
       $123 = (($122) + 37)|0; //@line 132 "load_stm.cpp"
       $124 = (($123) + ($121))|0; //@line 132 "load_stm.cpp"
       $125 = $124&255; //@line 132 "load_stm.cpp"
       HEAP8[$m$018>>0] = $125; //@line 132 "load_stm.cpp"
      }
     }
     $126 = ($112>>>0)<(65); //@line 133 "load_stm.cpp"
     if ($126) {
      $127 = ((($m$018)) + 2|0); //@line 133 "load_stm.cpp"
      HEAP8[$127>>0] = 1; //@line 133 "load_stm.cpp"
      $128 = $112&255; //@line 133 "load_stm.cpp"
      $129 = ((($m$018)) + 4|0); //@line 133 "load_stm.cpp"
      HEAP8[$129>>0] = $128; //@line 133 "load_stm.cpp"
     }
     $130 = ((($p$019)) + 3|0); //@line 134 "load_stm.cpp"
     $131 = HEAP8[$130>>0]|0; //@line 134 "load_stm.cpp"
     $132 = ((($m$018)) + 5|0); //@line 134 "load_stm.cpp"
     HEAP8[$132>>0] = $131; //@line 134 "load_stm.cpp"
     do {
      switch ($113|0) {
      case 1:  {
       $133 = ((($m$018)) + 3|0); //@line 138 "load_stm.cpp"
       HEAP8[$133>>0] = 16; //@line 138 "load_stm.cpp"
       $134 = ($131&255) >>> 4; //@line 138 "load_stm.cpp"
       HEAP8[$132>>0] = $134; //@line 138 "load_stm.cpp"
       break;
      }
      case 2:  {
       $135 = ((($m$018)) + 3|0); //@line 140 "load_stm.cpp"
       HEAP8[$135>>0] = 12; //@line 140 "load_stm.cpp"
       break;
      }
      case 3:  {
       $136 = ((($m$018)) + 3|0); //@line 142 "load_stm.cpp"
       HEAP8[$136>>0] = 14; //@line 142 "load_stm.cpp"
       $137 = $131&255; //@line 142 "load_stm.cpp"
       $138 = $137 & 240; //@line 142 "load_stm.cpp"
       $139 = ($138*10)|0; //@line 142 "load_stm.cpp"
       $140 = $137 & 15; //@line 142 "load_stm.cpp"
       $141 = $139 | $140; //@line 142 "load_stm.cpp"
       $142 = $141&255; //@line 142 "load_stm.cpp"
       HEAP8[$132>>0] = $142; //@line 142 "load_stm.cpp"
       break;
      }
      case 4:  {
       $143 = ((($m$018)) + 3|0); //@line 144 "load_stm.cpp"
       HEAP8[$143>>0] = 11; //@line 144 "load_stm.cpp"
       break;
      }
      case 5:  {
       $144 = ((($m$018)) + 3|0); //@line 146 "load_stm.cpp"
       HEAP8[$144>>0] = 3; //@line 146 "load_stm.cpp"
       break;
      }
      case 6:  {
       $145 = ((($m$018)) + 3|0); //@line 148 "load_stm.cpp"
       HEAP8[$145>>0] = 2; //@line 148 "load_stm.cpp"
       break;
      }
      case 7:  {
       $146 = ((($m$018)) + 3|0); //@line 150 "load_stm.cpp"
       HEAP8[$146>>0] = 4; //@line 150 "load_stm.cpp"
       break;
      }
      case 8:  {
       $147 = ((($m$018)) + 3|0); //@line 152 "load_stm.cpp"
       HEAP8[$147>>0] = 5; //@line 152 "load_stm.cpp"
       break;
      }
      case 9:  {
       $148 = ((($m$018)) + 3|0); //@line 154 "load_stm.cpp"
       HEAP8[$148>>0] = 18; //@line 154 "load_stm.cpp"
       break;
      }
      case 10:  {
       $149 = ((($m$018)) + 3|0); //@line 156 "load_stm.cpp"
       HEAP8[$149>>0] = 1; //@line 156 "load_stm.cpp"
       break;
      }
      case 11:  {
       $150 = ((($m$018)) + 3|0); //@line 158 "load_stm.cpp"
       HEAP8[$150>>0] = 7; //@line 158 "load_stm.cpp"
       break;
      }
      case 12:  {
       $151 = ((($m$018)) + 3|0); //@line 160 "load_stm.cpp"
       HEAP8[$151>>0] = 6; //@line 160 "load_stm.cpp"
       break;
      }
      default: {
       HEAP8[$132>>0] = 0; //@line 164 "load_stm.cpp"
       $152 = ((($m$018)) + 3|0); //@line 164 "load_stm.cpp"
       HEAP8[$152>>0] = 0; //@line 164 "load_stm.cpp"
      }
      }
     } while(0);
     $153 = (($n$020) + 1)|0; //@line 117 "load_stm.cpp"
     $154 = ((($p$019)) + 4|0); //@line 117 "load_stm.cpp"
     $155 = ((($m$018)) + 6|0); //@line 117 "load_stm.cpp"
     $exitcond27 = ($153|0)==(256); //@line 117 "load_stm.cpp"
     if ($exitcond27) {
      break;
     } else {
      $m$018 = $155;$n$020 = $153;$p$019 = $154;
     }
    }
    $156 = (($nPat$021) + 1)|0; //@line 110 "load_stm.cpp"
    $157 = ($156>>>0)<($91>>>0); //@line 110 "load_stm.cpp"
    if ($157) {
     $dwMemPos$022 = $93;$nPat$021 = $156;
    } else {
     $dwMemPos$117 = $93;$nSmp$016 = 1;
     break L25;
    }
   }
   if ((label|0) == 45) {
    return ($$0|0); //@line 185 "load_stm.cpp"
   }
  }
 } while(0);
 while(1) {
  $158 = (((($this)) + 33280|0) + ($nSmp$016<<6)|0); //@line 172 "load_stm.cpp"
  $159 = (($dwMemPos$117) + 15)|0; //@line 173 "load_stm.cpp"
  $160 = $159 & -16; //@line 173 "load_stm.cpp"
  $161 = HEAP32[$158>>2]|0; //@line 174 "load_stm.cpp"
  $162 = ($161|0)==(0); //@line 174 "load_stm.cpp"
  if ($162) {
   $dwMemPos$3 = $160;
  } else {
   $163 = (($nSmp$016) + -1)|0; //@line 176 "load_stm.cpp"
   $164 = (((($47) + ($163<<5)|0)) + 14|0); //@line 176 "load_stm.cpp"
   $165 = HEAPU8[$164>>0]|(HEAPU8[$164+1>>0]<<8); //@line 176 "load_stm.cpp"
   $166 = $165&65535; //@line 176 "load_stm.cpp"
   $167 = $166 << 4; //@line 176 "load_stm.cpp"
   $168 = ($167>>>0)<(1168); //@line 177 "load_stm.cpp"
   $169 = (($167) + ($161))|0;
   $170 = ($169>>>0)>($dwMemLength>>>0); //@line 177 "load_stm.cpp"
   $or$cond11 = $168 | $170; //@line 177 "load_stm.cpp"
   $dwMemPos$2 = $or$cond11 ? $160 : $167; //@line 177 "load_stm.cpp"
   $171 = ($dwMemPos$2>>>0)<($dwMemLength>>>0); //@line 178 "load_stm.cpp"
   if ($171) {
    $172 = (($lpStream) + ($dwMemPos$2)|0); //@line 180 "load_stm.cpp"
    $173 = (($dwMemLength) - ($dwMemPos$2))|0; //@line 180 "load_stm.cpp"
    $174 = (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$158,0,$172,$173)|0); //@line 180 "load_stm.cpp"
    $175 = (($174) + ($dwMemPos$2))|0; //@line 180 "load_stm.cpp"
    $dwMemPos$3 = $175;
   } else {
    $dwMemPos$3 = $dwMemPos$2;
   }
  }
  $176 = (($nSmp$016) + 1)|0; //@line 170 "load_stm.cpp"
  $exitcond = ($176|0)==(32); //@line 170 "load_stm.cpp"
  if ($exitcond) {
   $$0 = 1;
   break;
  } else {
   $dwMemPos$117 = $dwMemPos$3;$nSmp$016 = $176;
  }
 }
 return ($$0|0); //@line 185 "load_stm.cpp"
}
function __ZN10CSoundFile7ReadMedEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$0 = 0, $$21 = 0, $$22 = 0, $$22$us = 0, $$59 = 0, $$60 = 0, $$7$ = 0, $$70 = 0, $$79 = 0, $$83 = 0, $$deftempo$2 = 0, $$lcssa = 0, $$lcssa223 = 0, $$lobit = 0, $$off = 0, $$off88 = 0, $$off89 = 0, $$off89$us = 0, $$pre = 0;
 var $$pre199 = 0, $$pre205 = 0, $$sum = 0, $$sum23 = 0, $$sum24 = 0, $$sum25 = 0, $$sum26 = 0, $$sum27 = 0, $$sum28 = 0, $$sum29 = 0, $$sum30 = 0, $$sum32 = 0, $$sum33 = 0, $$sum34 = 0, $$sum34206 = 0, $$sum35 = 0, $$sum36 = 0, $$sum37 = 0, $$sum38 = 0, $$sum39 = 0;
 var $$sum40 = 0, $$sum41 = 0, $$sum42 = 0, $$sum43 = 0, $$sum45 = 0, $$sum45$pn = 0, $$sum46 = 0, $$sum47 = 0, $$sum48 = 0, $$sum49 = 0, $$sum50 = 0, $$sum51 = 0, $$sum52 = 0, $$trknamelen$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0;
 var $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0;
 var $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0;
 var $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0;
 var $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0;
 var $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0;
 var $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0;
 var $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0;
 var $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0;
 var $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0;
 var $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0;
 var $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0;
 var $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0;
 var $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0;
 var $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0;
 var $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0;
 var $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0;
 var $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0;
 var $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0;
 var $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0;
 var $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0;
 var $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0;
 var $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0;
 var $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $deftempo$0 = 0, $deftempo$1 = 0, $deftempo$2$lcssa = 0, $deftempo$2175 = 0, $deftempo$3 = 0, $exitcond = 0, $exitcond186 = 0, $exitcond187 = 0, $exitcond190 = 0, $exitcond191 = 0, $exitcond192 = 0, $exitcond198 = 0, $flags$0 = 0, $flags$1 = 0, $i$0159 = 0, $i23$0152 = 0, $i25$0148 = 0, $iBlk$0127 = 0, $iCh$0104 = 0;
 var $iSHdr$0171 = 0, $iSection$0163 = 0, $iSmp$0133 = 0, $instr$0$us = 0, $instr$1$us = 0, $len$0 = 0, $len$1 = 0, $len$1$ = 0, $len$2 = 0, $nOrders$0$lcssa207 = 0, $nOrders$0166 = 0, $nOrders$1160 = 0, $nOrders$2 = 0, $nOrders$3 = 0, $nOrders$3$lcssa = 0, $nSections$0 = 0, $nSections$1165 = 0, $nSections$2 = 0, $nplayseq$0 = 0, $or$cond = 0;
 var $or$cond12 = 0, $or$cond14 = 0, $or$cond16 = 0, $or$cond18 = 0, $or$cond212 = 0, $or$cond213 = 0, $or$cond215 = 0, $or$cond216 = 0, $or$cond55 = 0, $or$cond58 = 0, $or$cond6 = 0, $or$cond62 = 0, $or$cond64 = 0, $or$cond65 = 0, $or$cond67 = 0, $or$cond68 = 0, $or$cond69 = 0, $or$cond72 = 0, $or$cond73 = 0, $or$cond74 = 0;
 var $or$cond76 = 0, $or$cond78 = 0, $or$cond80 = 0, $or$cond82 = 0, $p$0124$us = 0, $p$1116$us = 0, $p$2$us = 0, $p$2$us$lcssa = 0, $p40$0113 = 0, $p40$1$lcssa = 0, $p40$1105 = 0, $p40$1105$us = 0, $p40$2 = 0, $p40$2$us = 0, $pcmdext$0 = 0, $pcmdext$1$ = 0, $pcmdext$1110 = 0, $playtransp$0 = 0, $pmex$0 = 0, $psdata$0 = 0;
 var $s$0125$us = 0, $s$1117$us = 0, $s41$0114 = 0, $s41$1$lcssa = 0, $s41$1106 = 0, $s41$1106$us = 0, $scevgep189 = 0, $scevgep196 = 0, $sectiontable$0164 = 0, $sectiontable$1 = 0, $split = 0, $split109 = 0, $switch$split2D = 0, $tracks$0 = 0, $tracks$1 = 0, $trknamelen$0$lcssa = 0, $trknamelen$0139 = 0, $trknamelen$1 = 0, $trknameofs$0$lcssa = 0, $trknameofs$0140 = 0;
 var $trknameofs$1 = 0, $trktagofs$0138 = 0, $uglygep = 0, $x$0119$us = 0, $x44$0107 = 0, $x44$0107$us = 0, $y$0126$us = 0, $y43$0115 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ($lpStream|0)==(0|0); //@line 488 "load_med.cpp"
 $1 = ($dwMemLength>>>0)<(512); //@line 488 "load_med.cpp"
 $or$cond = $0 | $1; //@line 488 "load_med.cpp"
 if ($or$cond) {
  $$0 = 0;
  return ($$0|0); //@line 929 "load_med.cpp"
 }
 $2 = HEAPU8[$lpStream>>0]|(HEAPU8[$lpStream+1>>0]<<8)|(HEAPU8[$lpStream+2>>0]<<16)|(HEAPU8[$lpStream+3>>0]<<24); //@line 490 "load_med.cpp"
 $3 = $2 & 16777215; //@line 490 "load_med.cpp"
 $4 = ($3|0)==(4476237); //@line 490 "load_med.cpp"
 if (!($4)) {
  $$0 = 0;
  return ($$0|0); //@line 929 "load_med.cpp"
 }
 $5 = ((($lpStream)) + 8|0); //@line 490 "load_med.cpp"
 $6 = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8)|(HEAPU8[$5+2>>0]<<16)|(HEAPU8[$5+3>>0]<<24); //@line 490 "load_med.cpp"
 $7 = ($6|0)==(0); //@line 490 "load_med.cpp"
 if ($7) {
  $$0 = 0;
  return ($$0|0); //@line 929 "load_med.cpp"
 }
 $8 = (_llvm_bswap_i32(($6|0))|0); //@line 492 "load_med.cpp"
 $9 = ($8>>>0)<($dwMemLength>>>0); //@line 493 "load_med.cpp"
 $10 = (($8) + 788)|0;
 $11 = ($10>>>0)<($dwMemLength>>>0); //@line 493 "load_med.cpp"
 $or$cond55 = $9 & $11; //@line 493 "load_med.cpp"
 if (!($or$cond55)) {
  $$0 = 0;
  return ($$0|0); //@line 929 "load_med.cpp"
 }
 $12 = $2 >> 24; //@line 495 "load_med.cpp"
 $13 = $12 & -4; //@line 495 "load_med.cpp"
 $14 = ($13|0)==(48); //@line 495 "load_med.cpp"
 if (!($14)) {
  $$0 = 0;
  return ($$0|0); //@line 929 "load_med.cpp"
 }
 $15 = ((($this)) + 59680|0); //@line 517 "load_med.cpp"
 HEAP32[$15>>2] = 8; //@line 517 "load_med.cpp"
 $16 = ((($this)) + 59748|0); //@line 518 "load_med.cpp"
 HEAP32[$16>>2] = 32; //@line 518 "load_med.cpp"
 $17 = ((($lpStream)) + 16|0); //@line 519 "load_med.cpp"
 $18 = HEAPU8[$17>>0]|(HEAPU8[$17+1>>0]<<8)|(HEAPU8[$17+2>>0]<<16)|(HEAPU8[$17+3>>0]<<24); //@line 519 "load_med.cpp"
 $19 = (_llvm_bswap_i32(($18|0))|0); //@line 519 "load_med.cpp"
 $20 = ((($lpStream)) + 24|0); //@line 520 "load_med.cpp"
 $21 = HEAPU8[$20>>0]|(HEAPU8[$20+1>>0]<<8)|(HEAPU8[$20+2>>0]<<16)|(HEAPU8[$20+3>>0]<<24); //@line 520 "load_med.cpp"
 $22 = (_llvm_bswap_i32(($21|0))|0); //@line 520 "load_med.cpp"
 $23 = ((($lpStream)) + 32|0); //@line 521 "load_med.cpp"
 $24 = HEAPU8[$23>>0]|(HEAPU8[$23+1>>0]<<8)|(HEAPU8[$23+2>>0]<<16)|(HEAPU8[$23+3>>0]<<24); //@line 521 "load_med.cpp"
 $25 = (_llvm_bswap_i32(($24|0))|0); //@line 521 "load_med.cpp"
 $26 = ($24|0)!=(0); //@line 522 "load_med.cpp"
 $27 = (($dwMemLength) + -80)|0;
 $28 = ($25>>>0)<($27>>>0); //@line 522 "load_med.cpp"
 $or$cond58 = $26 & $28; //@line 522 "load_med.cpp"
 $29 = (($lpStream) + ($25)|0); //@line 523 "load_med.cpp"
 $pmex$0 = $or$cond58 ? $29 : 0; //@line 522 "load_med.cpp"
 $30 = (($lpStream) + ($8)|0); //@line 526 "load_med.cpp"
 $$sum = (($8) + 504)|0; //@line 573 "load_med.cpp"
 $31 = (($lpStream) + ($$sum)|0); //@line 573 "load_med.cpp"
 $32 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8); //@line 573 "load_med.cpp"
 $33 = $32&65535; //@line 573 "load_med.cpp"
 $34 = $33 >>> 8; //@line 573 "load_med.cpp"
 $35 = $33 << 8; //@line 573 "load_med.cpp"
 $36 = $35 & 65280; //@line 573 "load_med.cpp"
 $37 = $36 | $34; //@line 573 "load_med.cpp"
 $38 = ((($this)) + 59664|0); //@line 574 "load_med.cpp"
 HEAP32[$38>>2] = 4; //@line 574 "load_med.cpp"
 $$sum23 = (($8) + 787)|0; //@line 575 "load_med.cpp"
 $39 = (($lpStream) + ($$sum23)|0); //@line 575 "load_med.cpp"
 $40 = HEAP8[$39>>0]|0; //@line 575 "load_med.cpp"
 $41 = $40&255; //@line 575 "load_med.cpp"
 $42 = ((($this)) + 59684|0); //@line 575 "load_med.cpp"
 $43 = ($40&255)>(63); //@line 576 "load_med.cpp"
 $$59 = $43 ? 63 : $41; //@line 576 "load_med.cpp"
 HEAP32[$42>>2] = $$59; //@line 575 "load_med.cpp"
 $44 = ((($this)) + 59652|0); //@line 578 "load_med.cpp"
 HEAP32[$44>>2] = 125; //@line 578 "load_med.cpp"
 $$sum24 = (($8) + 764)|0; //@line 579 "load_med.cpp"
 $45 = (($lpStream) + ($$sum24)|0); //@line 579 "load_med.cpp"
 $46 = HEAPU8[$45>>0]|(HEAPU8[$45+1>>0]<<8); //@line 579 "load_med.cpp"
 $47 = $46&65535; //@line 579 "load_med.cpp"
 $48 = $47 >>> 8; //@line 579 "load_med.cpp"
 $49 = $47 << 8; //@line 579 "load_med.cpp"
 $50 = $49 & 65280; //@line 579 "load_med.cpp"
 $51 = $50 | $48; //@line 579 "load_med.cpp"
 $52 = ($51|0)!=(0); //@line 580 "load_med.cpp"
 $$ = $52 ? $51 : 125; //@line 580 "load_med.cpp"
 $$sum25 = (($8) + 768)|0; //@line 581 "load_med.cpp"
 $53 = (($lpStream) + ($$sum25)|0); //@line 581 "load_med.cpp"
 $54 = HEAP8[$53>>0]|0; //@line 581 "load_med.cpp"
 $55 = $54&255; //@line 581 "load_med.cpp"
 $56 = $55 & 32; //@line 581 "load_med.cpp"
 $57 = ($56|0)==(0); //@line 581 "load_med.cpp"
 do {
  if ($57) {
   $$sum26 = (($8) + 767)|0; //@line 592 "load_med.cpp"
   $62 = (($lpStream) + ($$sum26)|0); //@line 592 "load_med.cpp"
   $63 = HEAP8[$62>>0]|0; //@line 592 "load_med.cpp"
   $64 = $63 & 64; //@line 592 "load_med.cpp"
   $65 = ($64<<24>>24)!=(0); //@line 592 "load_med.cpp"
   $66 = ($$>>>0)<(11); //@line 592 "load_med.cpp"
   $or$cond6 = $66 & $65; //@line 592 "load_med.cpp"
   if ($or$cond6) {
    $67 = (($$) + -1)|0; //@line 594 "load_med.cpp"
    $68 = (529839 + ($67)|0); //@line 594 "load_med.cpp"
    $69 = HEAP8[$68>>0]|0; //@line 594 "load_med.cpp"
    $70 = $69&255; //@line 594 "load_med.cpp"
    $deftempo$0 = $70;
    break;
   } else {
    $71 = (__Z7_muldivlll($$,3579545,948652)|0); //@line 596 "load_med.cpp"
    $deftempo$0 = $71;
    break;
   }
  } else {
   $58 = $55 & 31; //@line 583 "load_med.cpp"
   $59 = (($58) + 1)|0; //@line 583 "load_med.cpp"
   $60 = Math_imul($$, $59)|0; //@line 585 "load_med.cpp"
   $61 = $60 >>> 2; //@line 586 "load_med.cpp"
   $deftempo$0 = $61;
  }
 } while(0);
 $$sum27 = (($8) + 769)|0; //@line 603 "load_med.cpp"
 $72 = (($lpStream) + ($$sum27)|0); //@line 603 "load_med.cpp"
 $73 = HEAP8[$72>>0]|0; //@line 603 "load_med.cpp"
 $74 = $73&255; //@line 603 "load_med.cpp"
 $75 = ((($this)) + 59648|0); //@line 603 "load_med.cpp"
 $76 = ($73<<24>>24)==(0); //@line 604 "load_med.cpp"
 $$60 = $76 ? 6 : $74; //@line 604 "load_med.cpp"
 HEAP32[$75>>2] = $$60; //@line 603 "load_med.cpp"
 $77 = ($deftempo$0>>>0)<(33); //@line 605 "load_med.cpp"
 $deftempo$1 = $77 ? 33 : $deftempo$0; //@line 605 "load_med.cpp"
 $78 = ($deftempo$1>>>0)>(255); //@line 606 "load_med.cpp"
 if ($78) {
  $79 = ($$60>>>0)>(3); //@line 608 "load_med.cpp"
  $80 = ($deftempo$1>>>0)>(260); //@line 608 "load_med.cpp"
  $81 = $80 & $79; //@line 608 "load_med.cpp"
  if ($81) {
   $83 = $$60;$deftempo$2175 = $deftempo$1;
   while(1) {
    $82 = (($83) + -1)|0; //@line 610 "load_med.cpp"
    $84 = Math_imul($82, $deftempo$2175)|0; //@line 610 "load_med.cpp"
    $85 = (($84>>>0) / ($83>>>0))&-1; //@line 610 "load_med.cpp"
    $86 = ($82>>>0)>(3); //@line 608 "load_med.cpp"
    $87 = ($85>>>0)>(260); //@line 608 "load_med.cpp"
    $88 = $87 & $86; //@line 608 "load_med.cpp"
    if ($88) {
     $83 = $82;$deftempo$2175 = $85;
    } else {
     $$lcssa = $82;$$lcssa223 = $85;
     break;
    }
   }
   HEAP32[$75>>2] = $$lcssa; //@line 611 "load_med.cpp"
   $deftempo$2$lcssa = $$lcssa223;
  } else {
   $deftempo$2$lcssa = $deftempo$1;
  }
  $89 = ($deftempo$2$lcssa>>>0)>(255); //@line 613 "load_med.cpp"
  $$deftempo$2 = $89 ? 255 : $deftempo$2$lcssa; //@line 613 "load_med.cpp"
  $deftempo$3 = $$deftempo$2;
 } else {
  $deftempo$3 = $deftempo$1;
 }
 HEAP32[$44>>2] = $deftempo$3; //@line 615 "load_med.cpp"
 $90 = HEAP32[$42>>2]|0; //@line 617 "load_med.cpp"
 $91 = ($90|0)==(0); //@line 617 "load_med.cpp"
 if (!($91)) {
  $iSHdr$0171 = 0;
  while(1) {
   $92 = (($iSHdr$0171) + 1)|0; //@line 619 "load_med.cpp"
   $93 = (($30) + ($iSHdr$0171<<3)|0); //@line 620 "load_med.cpp"
   $94 = HEAPU8[$93>>0]|(HEAPU8[$93+1>>0]<<8); //@line 620 "load_med.cpp"
   $95 = $94&65535; //@line 620 "load_med.cpp"
   $96 = $95 >>> 8; //@line 620 "load_med.cpp"
   $97 = $95 << 8; //@line 620 "load_med.cpp"
   $98 = $97 & 65280; //@line 620 "load_med.cpp"
   $99 = $98 | $96; //@line 620 "load_med.cpp"
   $100 = $99 << 1; //@line 620 "load_med.cpp"
   $101 = (((((($this)) + 33280|0) + ($92<<6)|0)) + 4|0); //@line 620 "load_med.cpp"
   HEAP32[$101>>2] = $100; //@line 620 "load_med.cpp"
   $102 = (((($30) + ($iSHdr$0171<<3)|0)) + 2|0); //@line 621 "load_med.cpp"
   $103 = HEAPU8[$102>>0]|(HEAPU8[$102+1>>0]<<8); //@line 621 "load_med.cpp"
   $104 = $103&65535; //@line 621 "load_med.cpp"
   $105 = $104 >>> 8; //@line 621 "load_med.cpp"
   $106 = $104 << 8; //@line 621 "load_med.cpp"
   $107 = $106 & 65280; //@line 621 "load_med.cpp"
   $108 = $107 | $105; //@line 621 "load_med.cpp"
   $109 = $108 << 1; //@line 621 "load_med.cpp"
   $110 = (($109) + ($100))|0; //@line 621 "load_med.cpp"
   $111 = (((((($this)) + 33280|0) + ($92<<6)|0)) + 8|0); //@line 621 "load_med.cpp"
   HEAP32[$111>>2] = $110; //@line 621 "load_med.cpp"
   $112 = (((($30) + ($iSHdr$0171<<3)|0)) + 6|0); //@line 622 "load_med.cpp"
   $113 = HEAP8[$112>>0]|0; //@line 622 "load_med.cpp"
   $114 = $113&255; //@line 622 "load_med.cpp"
   $115 = $114 << 2; //@line 622 "load_med.cpp"
   $116 = $115&65535; //@line 622 "load_med.cpp"
   $117 = (((((($this)) + 33280|0) + ($92<<6)|0)) + 30|0); //@line 622 "load_med.cpp"
   HEAP16[$117>>1] = $116; //@line 622 "load_med.cpp"
   $118 = (((((($this)) + 33280|0) + ($92<<6)|0)) + 32|0); //@line 623 "load_med.cpp"
   HEAP16[$118>>1] = 64; //@line 623 "load_med.cpp"
   $119 = ($116&65535)>(256); //@line 624 "load_med.cpp"
   if ($119) {
    HEAP16[$117>>1] = 256; //@line 624 "load_med.cpp"
   }
   $120 = (((($30) + ($iSHdr$0171<<3)|0)) + 7|0); //@line 625 "load_med.cpp"
   $121 = HEAP8[$120>>0]|0; //@line 625 "load_med.cpp"
   $122 = $121 << 24 >> 24; //@line 625 "load_med.cpp"
   $123 = Math_imul($122, -12)|0; //@line 625 "load_med.cpp"
   $124 = $123&255; //@line 625 "load_med.cpp"
   $125 = (((((($this)) + 33280|0) + ($92<<6)|0)) + 36|0); //@line 625 "load_med.cpp"
   HEAP8[$125>>0] = $124; //@line 625 "load_med.cpp"
   $126 = (((((($this)) + 33280|0) + ($92<<6)|0)) + 28|0); //@line 626 "load_med.cpp"
   HEAP16[$126>>1] = 128; //@line 626 "load_med.cpp"
   $127 = ($110|0)==(0); //@line 627 "load_med.cpp"
   if (!($127)) {
    $130 = (((((($this)) + 33280|0) + ($92<<6)|0)) + 34|0); //@line 627 "load_med.cpp"
    $131 = HEAP16[$130>>1]|0; //@line 627 "load_med.cpp"
    $132 = $131&65535; //@line 627 "load_med.cpp"
    $133 = $132 | 2; //@line 627 "load_med.cpp"
    $134 = $133&65535; //@line 627 "load_med.cpp"
    HEAP16[$130>>1] = $134; //@line 627 "load_med.cpp"
   }
   $128 = HEAP32[$42>>2]|0; //@line 617 "load_med.cpp"
   $129 = ($92>>>0)<($128>>>0); //@line 617 "load_med.cpp"
   if ($129) {
    $iSHdr$0171 = $92;
   } else {
    break;
   }
  }
 }
 $$sum28 = (($8) + 767)|0; //@line 630 "load_med.cpp"
 $135 = (($lpStream) + ($$sum28)|0); //@line 630 "load_med.cpp"
 $136 = HEAP8[$135>>0]|0; //@line 630 "load_med.cpp"
 $137 = $136 & 32; //@line 630 "load_med.cpp"
 $138 = ($137<<24>>24)==(0); //@line 630 "load_med.cpp"
 if ($138) {
  $139 = ((($this)) + 59660|0); //@line 630 "load_med.cpp"
  $140 = HEAP32[$139>>2]|0; //@line 630 "load_med.cpp"
  $141 = $140 | 2; //@line 630 "load_med.cpp"
  HEAP32[$139>>2] = $141; //@line 630 "load_med.cpp"
 }
 $142 = ($12|0)<(50); //@line 632 "load_med.cpp"
 do {
  if ($142) {
   $$sum50 = (($8) + 506)|0; //@line 634 "load_med.cpp"
   $143 = (($lpStream) + ($$sum50)|0); //@line 634 "load_med.cpp"
   $144 = HEAPU8[$143>>0]|(HEAPU8[$143+1>>0]<<8); //@line 634 "load_med.cpp"
   $145 = $144&65535; //@line 634 "load_med.cpp"
   $146 = $145 >>> 8; //@line 634 "load_med.cpp"
   $147 = ($146|0)!=(0); //@line 636 "load_med.cpp"
   $$7$ = $147 ? $146 : 1; //@line 636 "load_med.cpp"
   $148 = ((($this)) + 53344|0); //@line 637 "load_med.cpp"
   $$sum51 = (($8) + 508)|0; //@line 637 "load_med.cpp"
   $149 = (($lpStream) + ($$sum51)|0); //@line 637 "load_med.cpp"
   _memcpy(($148|0),($149|0),($$7$|0))|0; //@line 637 "load_med.cpp"
   $$sum52 = (($8) + 766)|0; //@line 638 "load_med.cpp"
   $150 = (($lpStream) + ($$sum52)|0); //@line 638 "load_med.cpp"
   $151 = HEAP8[$150>>0]|0; //@line 638 "load_med.cpp"
   $152 = $151 << 24 >> 24; //@line 638 "load_med.cpp"
   $playtransp$0 = $152;
  } else {
   $$sum29 = (($8) + 520)|0; //@line 642 "load_med.cpp"
   $153 = (($lpStream) + ($$sum29)|0); //@line 642 "load_med.cpp"
   $154 = HEAPU8[$153>>0]|(HEAPU8[$153+1>>0]<<8); //@line 642 "load_med.cpp"
   $155 = $154&65535; //@line 642 "load_med.cpp"
   $156 = $155 >>> 8; //@line 642 "load_med.cpp"
   $157 = $155 << 8; //@line 642 "load_med.cpp"
   $158 = $157 & 65280; //@line 642 "load_med.cpp"
   $159 = $158 | $156; //@line 642 "load_med.cpp"
   $$off = (($159) + -4)|0; //@line 643 "load_med.cpp"
   $160 = ($$off>>>0)<(29); //@line 643 "load_med.cpp"
   if ($160) {
    HEAP32[$38>>2] = $159; //@line 643 "load_med.cpp"
   }
   $$sum30 = (($8) + 508)|0; //@line 644 "load_med.cpp"
   $161 = (($lpStream) + ($$sum30)|0); //@line 644 "load_med.cpp"
   $162 = HEAPU8[$161>>0]|(HEAPU8[$161+1>>0]<<8)|(HEAPU8[$161+2>>0]<<16)|(HEAPU8[$161+3>>0]<<24); //@line 644 "load_med.cpp"
   $163 = (_llvm_bswap_i32(($162|0))|0); //@line 644 "load_med.cpp"
   $$sum32 = (($8) + 506)|0; //@line 648 "load_med.cpp"
   $164 = (($lpStream) + ($$sum32)|0); //@line 648 "load_med.cpp"
   $165 = HEAPU8[$164>>0]|(HEAPU8[$164+1>>0]<<8); //@line 648 "load_med.cpp"
   $166 = $165&65535; //@line 648 "load_med.cpp"
   $167 = $166 >>> 8; //@line 648 "load_med.cpp"
   $168 = $166 << 8; //@line 648 "load_med.cpp"
   $169 = $168 & 65280; //@line 648 "load_med.cpp"
   $170 = $169 | $167; //@line 648 "load_med.cpp"
   $$sum33 = (($8) + 512)|0; //@line 649 "load_med.cpp"
   $171 = (($lpStream) + ($$sum33)|0); //@line 649 "load_med.cpp"
   $172 = HEAPU8[$171>>0]|(HEAPU8[$171+1>>0]<<8)|(HEAPU8[$171+2>>0]<<16)|(HEAPU8[$171+3>>0]<<24); //@line 649 "load_med.cpp"
   $173 = (_llvm_bswap_i32(($172|0))|0); //@line 649 "load_med.cpp"
   $174 = ($170|0)!=(0); //@line 650 "load_med.cpp"
   $175 = ($172|0)!=(0); //@line 650 "load_med.cpp"
   $or$cond12 = $175 & $174; //@line 650 "load_med.cpp"
   $176 = (($dwMemLength) + -2)|0;
   $177 = ($173>>>0)<($176>>>0); //@line 650 "load_med.cpp"
   $or$cond62 = $177 & $or$cond12; //@line 650 "load_med.cpp"
   $nSections$0 = $or$cond62 ? $170 : 1; //@line 650 "load_med.cpp"
   $178 = ($nSections$0|0)==(0); //@line 652 "load_med.cpp"
   if ($178) {
    $$sum34206 = (($8) + 766)|0; //@line 687 "load_med.cpp"
    $179 = (($lpStream) + ($$sum34206)|0); //@line 687 "load_med.cpp"
    $180 = HEAP8[$179>>0]|0; //@line 687 "load_med.cpp"
    $181 = $180 << 24 >> 24; //@line 687 "load_med.cpp"
    $604 = $181;$nOrders$0$lcssa207 = 0;
   } else {
    $182 = ($162|0)!=(0); //@line 665 "load_med.cpp"
    $183 = ($163>>>0)<($dwMemLength>>>0); //@line 665 "load_med.cpp"
    $or$cond65 = $182 & $183; //@line 665 "load_med.cpp"
    $184 = (($dwMemLength) - ($163))|0; //@line 665 "load_med.cpp"
    $185 = (($lpStream) + ($163)|0); //@line 667 "load_med.cpp"
    $186 = (($dwMemLength) + -1066)|0;
    $187 = ((($this)) + 59804|0); //@line 672 "load_med.cpp"
    $188 = (($dwMemLength) + -42)|0; //@line 674 "load_med.cpp"
    $iSection$0163 = 0;$nOrders$0166 = 0;$nSections$1165 = $nSections$0;$sectiontable$0164 = $173;
    while(1) {
     $189 = ($sectiontable$0164|0)!=(0); //@line 655 "load_med.cpp"
     $190 = ($sectiontable$0164>>>0)<($176>>>0); //@line 655 "load_med.cpp"
     $or$cond64 = $189 & $190; //@line 655 "load_med.cpp"
     if ($or$cond64) {
      $191 = (($sectiontable$0164) + 1)|0; //@line 657 "load_med.cpp"
      $192 = (($lpStream) + ($191)|0); //@line 657 "load_med.cpp"
      $193 = HEAP8[$192>>0]|0; //@line 657 "load_med.cpp"
      $194 = $193&255; //@line 657 "load_med.cpp"
      $195 = (($sectiontable$0164) + 2)|0; //@line 658 "load_med.cpp"
      $nSections$2 = $nSections$1165;$nplayseq$0 = $194;$sectiontable$1 = $195;
     } else {
      $nSections$2 = 0;$nplayseq$0 = 0;$sectiontable$1 = $sectiontable$0164;
     }
     $196 = $nplayseq$0 << 2;
     $197 = ($196>>>0)<($184>>>0); //@line 665 "load_med.cpp"
     $or$cond212 = $or$cond65 & $197; //@line 665 "load_med.cpp"
     do {
      if ($or$cond212) {
       $198 = (($185) + ($nplayseq$0<<2)|0); //@line 667 "load_med.cpp"
       $199 = HEAP32[$198>>2]|0; //@line 667 "load_med.cpp"
       $200 = (_llvm_bswap_i32(($199|0))|0); //@line 667 "load_med.cpp"
       $201 = ($199|0)!=(0); //@line 669 "load_med.cpp"
       $202 = ($200>>>0)<($186>>>0); //@line 669 "load_med.cpp"
       $or$cond67 = $201 & $202; //@line 669 "load_med.cpp"
       if ($or$cond67) {
        $203 = HEAP8[$187>>0]|0; //@line 672 "load_med.cpp"
        $204 = ($203<<24>>24)==(0); //@line 672 "load_med.cpp"
        if ($204) {
         $205 = (($lpStream) + ($200)|0); //@line 671 "load_med.cpp"
         dest=$187; src=$205; stop=dest+31|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 672 "load_med.cpp"
        }
        $$sum48 = (($200) + 40)|0; //@line 673 "load_med.cpp"
        $206 = (($lpStream) + ($$sum48)|0); //@line 673 "load_med.cpp"
        $207 = HEAPU8[$206>>0]|(HEAPU8[$206+1>>0]<<8); //@line 673 "load_med.cpp"
        $208 = $207&65535; //@line 673 "load_med.cpp"
        $209 = $208 >>> 8; //@line 673 "load_med.cpp"
        $210 = $208 << 8; //@line 673 "load_med.cpp"
        $211 = $210 & 65280; //@line 673 "load_med.cpp"
        $212 = $211 | $209; //@line 673 "load_med.cpp"
        $213 = (($188) - ($200))|0; //@line 674 "load_med.cpp"
        $214 = $213 >>> 1; //@line 674 "load_med.cpp"
        $215 = ($212>>>0)>=($214>>>0); //@line 674 "load_med.cpp"
        $216 = ($212|0)==(0); //@line 676 "load_med.cpp"
        $or$cond213 = $215 | $216; //@line 674 "load_med.cpp"
        if ($or$cond213) {
         $nOrders$3 = $nOrders$0166;
         break;
        }
        $$sum49 = (($200) + 42)|0; //@line 678 "load_med.cpp"
        $217 = (($lpStream) + ($$sum49)|0); //@line 678 "load_med.cpp"
        $i$0159 = 0;$nOrders$1160 = $nOrders$0166;
        while(1) {
         $218 = (($217) + ($i$0159<<1)|0); //@line 678 "load_med.cpp"
         $219 = HEAPU8[$218>>0]|(HEAPU8[$218+1>>0]<<8); //@line 678 "load_med.cpp"
         $220 = $219&65535; //@line 678 "load_med.cpp"
         $221 = $220 >>> 8; //@line 678 "load_med.cpp"
         $222 = ($221>>>0)<($37>>>0); //@line 679 "load_med.cpp"
         $223 = ($nOrders$1160>>>0)<(255); //@line 679 "load_med.cpp"
         $or$cond14 = $223 & $222; //@line 679 "load_med.cpp"
         if ($or$cond14) {
          $224 = $221&255; //@line 681 "load_med.cpp"
          $225 = (($nOrders$1160) + 1)|0; //@line 681 "load_med.cpp"
          $226 = (((($this)) + 53344|0) + ($nOrders$1160)|0); //@line 681 "load_med.cpp"
          HEAP8[$226>>0] = $224; //@line 681 "load_med.cpp"
          $nOrders$2 = $225;
         } else {
          $nOrders$2 = $nOrders$1160;
         }
         $227 = (($i$0159) + 1)|0; //@line 676 "load_med.cpp"
         $exitcond198 = ($227|0)==($212|0); //@line 676 "load_med.cpp"
         if ($exitcond198) {
          $nOrders$3 = $nOrders$2;
          break;
         } else {
          $i$0159 = $227;$nOrders$1160 = $nOrders$2;
         }
        }
       } else {
        $nOrders$3 = $nOrders$0166;
       }
      } else {
       $nOrders$3 = $nOrders$0166;
      }
     } while(0);
     $228 = (($iSection$0163) + 1)|0; //@line 652 "load_med.cpp"
     $229 = ($228>>>0)<($nSections$2>>>0); //@line 652 "load_med.cpp"
     if ($229) {
      $iSection$0163 = $228;$nOrders$0166 = $nOrders$3;$nSections$1165 = $nSections$2;$sectiontable$0164 = $sectiontable$1;
     } else {
      $nOrders$3$lcssa = $nOrders$3;
      break;
     }
    }
    $$sum34 = (($8) + 766)|0; //@line 687 "load_med.cpp"
    $230 = (($lpStream) + ($$sum34)|0); //@line 687 "load_med.cpp"
    $231 = HEAP8[$230>>0]|0; //@line 687 "load_med.cpp"
    $232 = $231 << 24 >> 24; //@line 687 "load_med.cpp"
    $233 = ($nOrders$3$lcssa>>>0)<(256); //@line 688 "load_med.cpp"
    if ($233) {
     $604 = $232;$nOrders$0$lcssa207 = $nOrders$3$lcssa;
    } else {
     $playtransp$0 = $232;
     break;
    }
   }
   $scevgep196 = ((($this)) + 53344|0);
   $uglygep = (($scevgep196) + ($nOrders$0$lcssa207)|0);
   $234 = (256 - ($nOrders$0$lcssa207))|0; //@line 688 "load_med.cpp"
   _memset(($uglygep|0),-1,($234|0))|0; //@line 688 "load_med.cpp"
   $playtransp$0 = $604;
  }
 } while(0);
 $235 = ($pmex$0|0)==(0|0); //@line 691 "load_med.cpp"
 if (!($235)) {
  $236 = HEAP32[$38>>2]|0; //@line 694 "load_med.cpp"
  $237 = ($236|0)==(4); //@line 694 "load_med.cpp"
  if ($237) {
   $238 = HEAP8[$135>>0]|0; //@line 694 "load_med.cpp"
   $239 = $238 & 64; //@line 694 "load_med.cpp"
   $240 = ($239<<24>>24)==(0); //@line 694 "load_med.cpp"
   if (!($240)) {
    $241 = ((($pmex$0)) + 36|0); //@line 698 "load_med.cpp"
    $242 = HEAP8[$241>>0]|0; //@line 698 "load_med.cpp"
    $243 = ($242<<24>>24)==(0); //@line 698 "load_med.cpp"
    if ($243) {
     $593 = 4;
    } else {
     HEAP32[$38>>2] = 3; //@line 698 "load_med.cpp"
     $593 = 3;
    }
    $244 = ((($pmex$0)) + 37|0); //@line 698 "load_med.cpp"
    $245 = HEAP8[$244>>0]|0; //@line 698 "load_med.cpp"
    $246 = ($245<<24>>24)==(0); //@line 698 "load_med.cpp"
    if ($246) {
     $598 = $593;
    } else {
     $592 = (($593) + -1)|0; //@line 698 "load_med.cpp"
     HEAP32[$38>>2] = $592; //@line 698 "load_med.cpp"
     $598 = $592;
    }
    $594 = ((($pmex$0)) + 38|0); //@line 698 "load_med.cpp"
    $595 = HEAP8[$594>>0]|0; //@line 698 "load_med.cpp"
    $596 = ($595<<24>>24)==(0); //@line 698 "load_med.cpp"
    if ($596) {
     $603 = $598;
    } else {
     $597 = (($598) + -1)|0; //@line 698 "load_med.cpp"
     HEAP32[$38>>2] = $597; //@line 698 "load_med.cpp"
     $603 = $597;
    }
    $599 = ((($pmex$0)) + 39|0); //@line 698 "load_med.cpp"
    $600 = HEAP8[$599>>0]|0; //@line 698 "load_med.cpp"
    $601 = ($600<<24>>24)==(0); //@line 698 "load_med.cpp"
    if (!($601)) {
     $602 = (($603) + -1)|0; //@line 698 "load_med.cpp"
     HEAP32[$38>>2] = $602; //@line 698 "load_med.cpp"
    }
   }
  }
  $247 = ((($pmex$0)) + 12|0); //@line 702 "load_med.cpp"
  $248 = HEAPU8[$247>>0]|(HEAPU8[$247+1>>0]<<8)|(HEAPU8[$247+2>>0]<<16)|(HEAPU8[$247+3>>0]<<24); //@line 702 "load_med.cpp"
  $249 = (_llvm_bswap_i32(($248|0))|0); //@line 702 "load_med.cpp"
  $250 = ((($pmex$0)) + 16|0); //@line 703 "load_med.cpp"
  $251 = HEAPU8[$250>>0]|(HEAPU8[$250+1>>0]<<8)|(HEAPU8[$250+2>>0]<<16)|(HEAPU8[$250+3>>0]<<24); //@line 703 "load_med.cpp"
  $252 = (_llvm_bswap_i32(($251|0))|0); //@line 703 "load_med.cpp"
  $253 = ($248|0)!=(0); //@line 704 "load_med.cpp"
  $254 = ($251|0)!=(0); //@line 704 "load_med.cpp"
  $or$cond16 = $253 & $254; //@line 704 "load_med.cpp"
  if ($or$cond16) {
   $255 = (($252) + ($249))|0; //@line 704 "load_med.cpp"
   $256 = ($255>>>0)<=($249>>>0); //@line 704 "load_med.cpp"
   $257 = ($255>>>0)>($dwMemLength>>>0); //@line 705 "load_med.cpp"
   $or$cond68 = $256 | $257; //@line 705 "load_med.cpp"
   if (!($or$cond68)) {
    $258 = (($252) + 1)|0; //@line 707 "load_med.cpp"
    $259 = (__Znaj($258)|0); //@line 707 "load_med.cpp"
    $260 = ((($this)) + 59796|0); //@line 707 "load_med.cpp"
    HEAP32[$260>>2] = $259; //@line 707 "load_med.cpp"
    $261 = (($lpStream) + ($249)|0); //@line 708 "load_med.cpp"
    _memcpy(($259|0),($261|0),($252|0))|0; //@line 708 "load_med.cpp"
    $262 = (($259) + ($252)|0); //@line 709 "load_med.cpp"
    HEAP8[$262>>0] = 0; //@line 709 "load_med.cpp"
   }
  }
  $263 = ((($pmex$0)) + 44|0); //@line 712 "load_med.cpp"
  $264 = HEAPU8[$263>>0]|(HEAPU8[$263+1>>0]<<8)|(HEAPU8[$263+2>>0]<<16)|(HEAPU8[$263+3>>0]<<24); //@line 712 "load_med.cpp"
  $265 = (_llvm_bswap_i32(($264|0))|0); //@line 712 "load_med.cpp"
  $266 = ((($pmex$0)) + 48|0); //@line 713 "load_med.cpp"
  $267 = HEAPU8[$266>>0]|(HEAPU8[$266+1>>0]<<8)|(HEAPU8[$266+2>>0]<<16)|(HEAPU8[$266+3>>0]<<24); //@line 713 "load_med.cpp"
  $268 = (_llvm_bswap_i32(($267|0))|0); //@line 713 "load_med.cpp"
  $269 = ($264|0)!=(0); //@line 714 "load_med.cpp"
  $270 = ($267|0)!=(0); //@line 714 "load_med.cpp"
  $or$cond18 = $269 & $270; //@line 714 "load_med.cpp"
  if ($or$cond18) {
   $271 = (($268) + ($265))|0; //@line 714 "load_med.cpp"
   $272 = ($271>>>0)<=($265>>>0); //@line 714 "load_med.cpp"
   $273 = ($271>>>0)>($dwMemLength>>>0); //@line 715 "load_med.cpp"
   $or$cond69 = $272 | $273; //@line 715 "load_med.cpp"
   if (!($or$cond69)) {
    $274 = ($268>>>0)>(31); //@line 717 "load_med.cpp"
    $$70 = $274 ? 31 : $268; //@line 717 "load_med.cpp"
    $275 = ((($this)) + 59804|0); //@line 718 "load_med.cpp"
    $276 = (($lpStream) + ($265)|0); //@line 718 "load_med.cpp"
    _memcpy(($275|0),($276|0),($$70|0))|0; //@line 718 "load_med.cpp"
    $277 = ((($this)) + 59835|0); //@line 719 "load_med.cpp"
    HEAP8[$277>>0] = 0; //@line 719 "load_med.cpp"
   }
  }
  $278 = ((($pmex$0)) + 20|0); //@line 722 "load_med.cpp"
  $279 = HEAPU8[$278>>0]|(HEAPU8[$278+1>>0]<<8)|(HEAPU8[$278+2>>0]<<16)|(HEAPU8[$278+3>>0]<<24); //@line 722 "load_med.cpp"
  $280 = (_llvm_bswap_i32(($279|0))|0); //@line 722 "load_med.cpp"
  $281 = ($279|0)==(0); //@line 723 "load_med.cpp"
  if (!($281)) {
   $282 = ((($pmex$0)) + 24|0); //@line 726 "load_med.cpp"
   $283 = HEAPU8[$282>>0]|(HEAPU8[$282+1>>0]<<8); //@line 726 "load_med.cpp"
   $284 = $283&65535; //@line 726 "load_med.cpp"
   $285 = $284 >>> 8; //@line 726 "load_med.cpp"
   $286 = $284 << 8; //@line 726 "load_med.cpp"
   $287 = $286 & 65280; //@line 726 "load_med.cpp"
   $288 = $287 | $285; //@line 726 "load_med.cpp"
   $289 = ((($pmex$0)) + 26|0); //@line 727 "load_med.cpp"
   $290 = HEAPU8[$289>>0]|(HEAPU8[$289+1>>0]<<8); //@line 727 "load_med.cpp"
   $291 = $290&65535; //@line 727 "load_med.cpp"
   $292 = $291 >>> 8; //@line 727 "load_med.cpp"
   $293 = $291 << 8; //@line 727 "load_med.cpp"
   $294 = $293 & 65280; //@line 727 "load_med.cpp"
   $295 = $294 | $292; //@line 727 "load_med.cpp"
   $296 = ($295>>>0)<(256); //@line 729 "load_med.cpp"
   if ($296) {
    $297 = Math_imul($295, $288)|0; //@line 730 "load_med.cpp"
    $298 = ($297>>>0)<($dwMemLength>>>0); //@line 730 "load_med.cpp"
    $299 = (($dwMemLength) - ($297))|0;
    $300 = ($280>>>0)<($299>>>0); //@line 731 "load_med.cpp"
    $or$cond72 = $298 & $300; //@line 730 "load_med.cpp"
    if ($or$cond72) {
     $301 = ($295>>>0)>(32); //@line 736 "load_med.cpp"
     $$21 = $301 ? 32 : $295; //@line 736 "load_med.cpp"
     $302 = ($288|0)==(0); //@line 737 "load_med.cpp"
     if (!($302)) {
      $i23$0152 = 0;
      while(1) {
       $303 = HEAP32[$42>>2]|0; //@line 737 "load_med.cpp"
       $304 = ($i23$0152>>>0)<($303>>>0); //@line 737 "load_med.cpp"
       $305 = (($i23$0152) + 1)|0; //@line 739 "load_med.cpp"
       if ($304) {
        $306 = (((($this)) + 59804|0) + ($305<<5)|0); //@line 739 "load_med.cpp"
        $307 = Math_imul($i23$0152, $295)|0; //@line 739 "load_med.cpp"
        $$sum47 = (($307) + ($280))|0; //@line 739 "load_med.cpp"
        $308 = (($lpStream) + ($$sum47)|0); //@line 739 "load_med.cpp"
        (_strncpy($306,$308,$$21)|0); //@line 739 "load_med.cpp"
        $309 = (((((($this)) + 59804|0) + ($305<<5)|0)) + 31|0); //@line 740 "load_med.cpp"
        HEAP8[$309>>0] = 0; //@line 740 "load_med.cpp"
       }
       $exitcond192 = ($305|0)==($288|0); //@line 737 "load_med.cpp"
       if ($exitcond192) {
        break;
       } else {
        $i23$0152 = $305;
       }
      }
     }
    }
   }
  }
  $310 = ((($pmex$0)) + 68|0); //@line 745 "load_med.cpp"
  $311 = HEAPU8[$310>>0]|(HEAPU8[$310+1>>0]<<8)|(HEAPU8[$310+2>>0]<<16)|(HEAPU8[$310+3>>0]<<24); //@line 745 "load_med.cpp"
  $312 = (_llvm_bswap_i32(($311|0))|0); //@line 745 "load_med.cpp"
  $313 = ($311|0)!=(0); //@line 746 "load_med.cpp"
  $314 = ($312>>>0)<($dwMemLength>>>0); //@line 746 "load_med.cpp"
  $or$cond73 = $313 & $314; //@line 746 "load_med.cpp"
  if ($or$cond73) {
   $315 = HEAP32[$38>>2]|0; //@line 746 "load_med.cpp"
   $316 = $315 << 2; //@line 746 "load_med.cpp"
   $317 = (($dwMemLength) - ($312))|0; //@line 746 "load_med.cpp"
   $318 = ($316>>>0)<($317>>>0); //@line 746 "load_med.cpp"
   if ($318) {
    $319 = (($lpStream) + ($312)|0); //@line 748 "load_med.cpp"
    $320 = ($315|0)==(0); //@line 749 "load_med.cpp"
    if (!($320)) {
     $321 = (($dwMemLength) + -8)|0; //@line 755 "load_med.cpp"
     $605 = $315;$i25$0148 = 0;
     while(1) {
      $322 = (($319) + ($i25$0148<<2)|0); //@line 752 "load_med.cpp"
      $323 = HEAP32[$322>>2]|0; //@line 752 "load_med.cpp"
      $324 = (_llvm_bswap_i32(($323|0))|0); //@line 752 "load_med.cpp"
      $325 = ($323|0)==(0); //@line 753 "load_med.cpp"
      do {
       if ($325) {
        $346 = $605;
       } else {
        $326 = ($324>>>0)<($321>>>0); //@line 755 "load_med.cpp"
        L116: do {
         if ($326) {
          $trknamelen$0139 = 0;$trknameofs$0140 = 0;$trktagofs$0138 = $324;
          while(1) {
           $327 = (($lpStream) + ($trktagofs$0138)|0); //@line 757 "load_med.cpp"
           $328 = HEAP32[$327>>2]|0; //@line 757 "load_med.cpp"
           $329 = ($328|0)==(0); //@line 758 "load_med.cpp"
           if ($329) {
            $trknamelen$0$lcssa = $trknamelen$0139;$trknameofs$0$lcssa = $trknameofs$0140;
            break L116;
           }
           $330 = (_llvm_bswap_i32(($328|0))|0); //@line 757 "load_med.cpp"
           $$sum46 = (($trktagofs$0138) + 4)|0; //@line 759 "load_med.cpp"
           $331 = (($lpStream) + ($$sum46)|0); //@line 759 "load_med.cpp"
           $332 = HEAP32[$331>>2]|0; //@line 759 "load_med.cpp"
           $333 = (_llvm_bswap_i32(($332|0))|0); //@line 759 "load_med.cpp"
           $switch$split2D = ($330|0)<(2);
           L120: do {
            if ($switch$split2D) {
             switch ($330|0) {
             case -2147483647:  {
              break;
             }
             default: {
              $trknamelen$1 = $trknamelen$0139;$trknameofs$1 = $trknameofs$0140;
              break L120;
             }
             }
             $trknamelen$1 = $trknamelen$0139;$trknameofs$1 = $333;
            } else {
             switch ($330|0) {
             case 2:  {
              break;
             }
             default: {
              $trknamelen$1 = $trknamelen$0139;$trknameofs$1 = $trknameofs$0140;
              break L120;
             }
             }
             $trknamelen$1 = $333;$trknameofs$1 = $trknameofs$0140;
            }
           } while(0);
           $334 = (($trktagofs$0138) + 8)|0; //@line 765 "load_med.cpp"
           $335 = ($334>>>0)<($321>>>0); //@line 755 "load_med.cpp"
           if ($335) {
            $trknamelen$0139 = $trknamelen$1;$trknameofs$0140 = $trknameofs$1;$trktagofs$0138 = $334;
           } else {
            $trknamelen$0$lcssa = $trknamelen$1;$trknameofs$0$lcssa = $trknameofs$1;
            break;
           }
          }
         } else {
          $trknamelen$0$lcssa = 0;$trknameofs$0$lcssa = 0;
         }
        } while(0);
        $336 = ($trknamelen$0$lcssa>>>0)>(20); //@line 767 "load_med.cpp"
        $$trknamelen$0 = $336 ? 20 : $trknamelen$0$lcssa; //@line 767 "load_med.cpp"
        $337 = ($trknameofs$0$lcssa|0)!=(0); //@line 768 "load_med.cpp"
        $338 = ($$trknamelen$0>>>0)<($dwMemLength>>>0); //@line 768 "load_med.cpp"
        $or$cond74 = $337 & $338; //@line 768 "load_med.cpp"
        $339 = (($dwMemLength) - ($$trknamelen$0))|0;
        $340 = ($trknameofs$0$lcssa>>>0)<($339>>>0); //@line 768 "load_med.cpp"
        $or$cond76 = $or$cond74 & $340; //@line 768 "load_med.cpp"
        if (!($or$cond76)) {
         $346 = $605;
         break;
        }
        $341 = (((((($this)) + 49600|0) + (($i25$0148*36)|0)|0)) + 16|0); //@line 770 "load_med.cpp"
        $342 = (($lpStream) + ($trknameofs$0$lcssa)|0); //@line 770 "load_med.cpp"
        (_strncpy($341,$342,20)|0); //@line 770 "load_med.cpp"
        $343 = (((((($this)) + 49600|0) + (($i25$0148*36)|0)|0)) + 35|0); //@line 771 "load_med.cpp"
        HEAP8[$343>>0] = 0; //@line 771 "load_med.cpp"
        $$pre = HEAP32[$38>>2]|0; //@line 749 "load_med.cpp"
        $346 = $$pre;
       }
      } while(0);
      $344 = (($i25$0148) + 1)|0; //@line 749 "load_med.cpp"
      $345 = ($344>>>0)<($346>>>0); //@line 749 "load_med.cpp"
      if ($345) {
       $605 = $346;$i25$0148 = $344;
      } else {
       break;
      }
     }
    }
   }
  }
 }
 $347 = HEAP32[$42>>2]|0; //@line 778 "load_med.cpp"
 $348 = $347 << 2; //@line 778 "load_med.cpp"
 $349 = (($dwMemLength) - ($348))|0; //@line 778 "load_med.cpp"
 $350 = ($22>>>0)>($349>>>0); //@line 778 "load_med.cpp"
 if ($350) {
  $$0 = 1;
  return ($$0|0); //@line 929 "load_med.cpp"
 }
 $351 = (($lpStream) + ($22)|0); //@line 779 "load_med.cpp"
 $352 = ($347|0)==(0); //@line 780 "load_med.cpp"
 if (!($352)) {
  $353 = (($dwMemLength) + -6)|0; //@line 809 "load_med.cpp"
  $606 = $347;$iSmp$0133 = 0;
  while(1) {
   $354 = (($351) + ($iSmp$0133<<2)|0); //@line 780 "load_med.cpp"
   $355 = HEAP32[$354>>2]|0; //@line 780 "load_med.cpp"
   $356 = ($355|0)==(0); //@line 780 "load_med.cpp"
   if ($356) {
    $396 = $606;
   } else {
    $357 = (_llvm_bswap_i32(($355|0))|0); //@line 782 "load_med.cpp"
    $358 = ($357>>>0)<($dwMemLength>>>0); //@line 783 "load_med.cpp"
    $359 = (($357) + 21)|0;
    $360 = ($359>>>0)<($dwMemLength>>>0); //@line 783 "load_med.cpp"
    $or$cond78 = $358 & $360; //@line 783 "load_med.cpp"
    if ($or$cond78) {
     $361 = (($lpStream) + ($357)|0); //@line 784 "load_med.cpp"
     $362 = HEAPU8[$361>>0]|(HEAPU8[$361+1>>0]<<8)|(HEAPU8[$361+2>>0]<<16)|(HEAPU8[$361+3>>0]<<24); //@line 785 "load_med.cpp"
     $363 = (_llvm_bswap_i32(($362|0))|0); //@line 785 "load_med.cpp"
     $364 = ($363>>>0)>(16000000); //@line 789 "load_med.cpp"
     $$pre205 = (($357) + 6)|0; //@line 791 "load_med.cpp"
     $365 = (($$pre205) + ($363))|0;
     $366 = ($365>>>0)>($dwMemLength>>>0); //@line 789 "load_med.cpp"
     $or$cond215 = $364 | $366; //@line 789 "load_med.cpp"
     $len$0 = $or$cond215 ? 0 : $363; //@line 789 "load_med.cpp"
     $$sum43 = (($357) + 4)|0; //@line 790 "load_med.cpp"
     $367 = (($lpStream) + ($$sum43)|0); //@line 790 "load_med.cpp"
     $368 = HEAPU8[$367>>0]|(HEAPU8[$367+1>>0]<<8); //@line 790 "load_med.cpp"
     $369 = $368&65535; //@line 790 "load_med.cpp"
     $370 = $369 >>> 8; //@line 790 "load_med.cpp"
     $371 = $370 & 128; //@line 792 "load_med.cpp"
     $372 = ($371|0)==(0); //@line 792 "load_med.cpp"
     if ($372) {
      $376 = $370 & 16; //@line 797 "load_med.cpp"
      $377 = ($376|0)==(0); //@line 797 "load_med.cpp"
      if ($377) {
       $388 = $369 >>> 10; //@line 804 "load_med.cpp"
       $389 = $388 & 8; //@line 804 "load_med.cpp"
       $flags$0 = $389;$len$1 = $len$0;
      } else {
       $378 = (($iSmp$0133) + 1)|0; //@line 799 "load_med.cpp"
       $379 = (((((($this)) + 33280|0) + ($378<<6)|0)) + 34|0); //@line 799 "load_med.cpp"
       $380 = HEAP16[$379>>1]|0; //@line 799 "load_med.cpp"
       $381 = $380&65535; //@line 799 "load_med.cpp"
       $382 = $381 | 1; //@line 799 "load_med.cpp"
       $383 = $382&65535; //@line 799 "load_med.cpp"
       HEAP16[$379>>1] = $383; //@line 799 "load_med.cpp"
       $384 = $len$0 >>> 1; //@line 800 "load_med.cpp"
       $385 = $369 >>> 10; //@line 801 "load_med.cpp"
       $386 = $385 & 8; //@line 801 "load_med.cpp"
       $387 = $386 | 7; //@line 801 "load_med.cpp"
       $flags$0 = $387;$len$1 = $384;
      }
      $390 = $369 >>> 13; //@line 806 "load_med.cpp"
      $$lobit = $390 & 1; //@line 806 "load_med.cpp"
      $len$1$ = $len$1 >>> $$lobit; //@line 806 "load_med.cpp"
      $$sum45$pn = $$pre205;$flags$1 = $flags$0;$len$2 = $len$1$;
     } else {
      $373 = $369 >>> 10; //@line 794 "load_med.cpp"
      $374 = $373 & 8; //@line 794 "load_med.cpp"
      $375 = $374 | 6; //@line 794 "load_med.cpp"
      $$sum45 = (($375) + ($$pre205))|0; //@line 794 "load_med.cpp"
      $$sum45$pn = $$sum45;$flags$1 = 0;$len$2 = $len$0;
     }
     $psdata$0 = (($lpStream) + ($$sum45$pn)|0); //@line 794 "load_med.cpp"
     $391 = (($iSmp$0133) + 1)|0; //@line 808 "load_med.cpp"
     $392 = (((($this)) + 33280|0) + ($391<<6)|0); //@line 808 "load_med.cpp"
     HEAP32[$392>>2] = $len$2; //@line 808 "load_med.cpp"
     $393 = (($353) - ($357))|0; //@line 809 "load_med.cpp"
     (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$392,$flags$1,$psdata$0,$393)|0); //@line 809 "load_med.cpp"
     $$pre199 = HEAP32[$42>>2]|0; //@line 780 "load_med.cpp"
     $396 = $$pre199;
    } else {
     $396 = $606;
    }
   }
   $394 = (($iSmp$0133) + 1)|0; //@line 780 "load_med.cpp"
   $395 = ($394>>>0)<($396>>>0); //@line 780 "load_med.cpp"
   if ($395) {
    $606 = $396;$iSmp$0133 = $394;
   } else {
    break;
   }
  }
 }
 $397 = ($37>>>0)>(240); //@line 812 "load_med.cpp"
 $$79 = $397 ? 240 : $37; //@line 812 "load_med.cpp"
 $398 = ($18|0)==(0); //@line 813 "load_med.cpp"
 if ($398) {
  $$0 = 1;
  return ($$0|0); //@line 929 "load_med.cpp"
 }
 $399 = $$79 << 2; //@line 813 "load_med.cpp"
 $400 = (($dwMemLength) - ($399))|0; //@line 813 "load_med.cpp"
 $401 = ($19>>>0)>($400>>>0); //@line 813 "load_med.cpp"
 if ($401) {
  $$0 = 1;
  return ($$0|0); //@line 929 "load_med.cpp"
 }
 $402 = (($lpStream) + ($19)|0); //@line 814 "load_med.cpp"
 $403 = ($12|0)==(51); //@line 815 "load_med.cpp"
 $404 = $403 ? 24 : 48; //@line 815 "load_med.cpp"
 $405 = (($playtransp$0) + ($404))|0; //@line 815 "load_med.cpp"
 $406 = ($$79|0)==(0); //@line 816 "load_med.cpp"
 L156: do {
  if (!($406)) {
   $407 = (($dwMemLength) + -8)|0; //@line 819 "load_med.cpp"
   $408 = ($12|0)==(48); //@line 821 "load_med.cpp"
   $409 = (($dwMemLength) + -2)|0; //@line 832 "load_med.cpp"
   $410 = (($dwMemLength) + -36)|0;
   $411 = (($dwMemLength) + -4)|0; //@line 885 "load_med.cpp"
   $iBlk$0127 = 0;
   while(1) {
    $414 = (($402) + ($iBlk$0127<<2)|0); //@line 818 "load_med.cpp"
    $415 = HEAP32[$414>>2]|0; //@line 818 "load_med.cpp"
    $416 = (_llvm_bswap_i32(($415|0))|0); //@line 818 "load_med.cpp"
    $417 = ($415|0)!=(0); //@line 819 "load_med.cpp"
    $418 = ($416>>>0)<($dwMemLength>>>0); //@line 819 "load_med.cpp"
    $or$cond80 = $417 & $418; //@line 819 "load_med.cpp"
    $419 = ($416>>>0)<($407>>>0); //@line 819 "load_med.cpp"
    $or$cond216 = $or$cond80 & $419; //@line 819 "load_med.cpp"
    L160: do {
     if ($or$cond216) {
      $420 = (($lpStream) + ($416)|0); //@line 823 "load_med.cpp"
      if ($408) {
       $$sum41 = (($416) + 1)|0; //@line 824 "load_med.cpp"
       $421 = (($lpStream) + ($$sum41)|0); //@line 824 "load_med.cpp"
       $422 = HEAP8[$421>>0]|0; //@line 824 "load_med.cpp"
       $423 = $422&255; //@line 824 "load_med.cpp"
       $424 = (($423) + 1)|0; //@line 824 "load_med.cpp"
       $425 = HEAP8[$420>>0]|0; //@line 825 "load_med.cpp"
       $426 = $425&255; //@line 825 "load_med.cpp"
       $427 = ($425<<24>>24)==(0); //@line 826 "load_med.cpp"
       if ($427) {
        $428 = HEAP32[$38>>2]|0; //@line 826 "load_med.cpp"
        $tracks$0 = $428;
       } else {
        $tracks$0 = $426;
       }
       $429 = HEAP32[$38>>2]|0; //@line 827 "load_med.cpp"
       $430 = (__ZN10CSoundFile15AllocatePatternEjj($424,$429)|0); //@line 827 "load_med.cpp"
       $431 = (((($this)) + 51904|0) + ($iBlk$0127<<2)|0); //@line 827 "load_med.cpp"
       HEAP32[$431>>2] = $430; //@line 827 "load_med.cpp"
       $432 = ($430|0)==(0|0); //@line 827 "load_med.cpp"
       if ($432) {
        break;
       }
       $433 = $424&65535; //@line 828 "load_med.cpp"
       $434 = (((($this)) + 52864|0) + ($iBlk$0127<<1)|0); //@line 828 "load_med.cpp"
       HEAP16[$434>>1] = $433; //@line 828 "load_med.cpp"
       $435 = ($424*3)|0; //@line 831 "load_med.cpp"
       $436 = Math_imul($435, $tracks$0)|0; //@line 831 "load_med.cpp"
       $437 = (($436) + ($416))|0; //@line 832 "load_med.cpp"
       $438 = ($437>>>0)>($409>>>0); //@line 832 "load_med.cpp"
       if ($438) {
        break L156;
       }
       $439 = ($tracks$0|0)==(0); //@line 835 "load_med.cpp"
       if ($439) {
        break;
       }
       $440 = HEAP32[$431>>2]|0; //@line 829 "load_med.cpp"
       $$sum42 = (($416) + 2)|0; //@line 830 "load_med.cpp"
       $441 = (($lpStream) + ($$sum42)|0); //@line 830 "load_med.cpp"
       $442 = ($tracks$0*3)|0; //@line 833 "load_med.cpp"
       $p$0124$us = $440;$s$0125$us = $441;$y$0126$us = 0;
       while(1) {
        $p$1116$us = $p$0124$us;$s$1117$us = $s$0125$us;$x$0119$us = 0;
        while(1) {
         $444 = HEAP32[$38>>2]|0; //@line 835 "load_med.cpp"
         $445 = ($x$0119$us>>>0)<($444>>>0); //@line 835 "load_med.cpp"
         if ($445) {
          $446 = HEAP8[$s$1117$us>>0]|0; //@line 837 "load_med.cpp"
          $447 = $446&255; //@line 837 "load_med.cpp"
          $448 = $447 & 63; //@line 837 "load_med.cpp"
          $449 = ((($s$1117$us)) + 1|0); //@line 838 "load_med.cpp"
          $450 = HEAP8[$449>>0]|0; //@line 838 "load_med.cpp"
          $451 = ($450&255) >>> 4; //@line 838 "load_med.cpp"
          $452 = $447 & 128; //@line 839 "load_med.cpp"
          $453 = ($452|0)==(0); //@line 839 "load_med.cpp"
          if ($453) {
           $instr$0$us = $451;
          } else {
           $454 = $451&255; //@line 839 "load_med.cpp"
           $455 = $454 | 16; //@line 839 "load_med.cpp"
           $456 = $455&255; //@line 839 "load_med.cpp"
           $instr$0$us = $456;
          }
          $457 = $446 & 64; //@line 840 "load_med.cpp"
          $458 = ($457<<24>>24)==(0); //@line 840 "load_med.cpp"
          if ($458) {
           $instr$1$us = $instr$0$us;
          } else {
           $459 = $instr$0$us&255; //@line 840 "load_med.cpp"
           $460 = $459 | 32; //@line 840 "load_med.cpp"
           $461 = $460&255; //@line 840 "load_med.cpp"
           $instr$1$us = $461;
          }
          $462 = ($448|0)==(0); //@line 841 "load_med.cpp"
          if (!($462)) {
           $463 = (($448) + ($405))|0; //@line 841 "load_med.cpp"
           $464 = $463&255; //@line 841 "load_med.cpp"
           HEAP8[$p$1116$us>>0] = $464; //@line 841 "load_med.cpp"
          }
          $465 = ((($p$1116$us)) + 1|0); //@line 842 "load_med.cpp"
          HEAP8[$465>>0] = $instr$1$us; //@line 842 "load_med.cpp"
          $466 = HEAP8[$449>>0]|0; //@line 843 "load_med.cpp"
          $467 = $466&255; //@line 843 "load_med.cpp"
          $468 = $467 & 15; //@line 843 "load_med.cpp"
          $469 = $468&255; //@line 843 "load_med.cpp"
          $470 = ((($p$1116$us)) + 3|0); //@line 843 "load_med.cpp"
          HEAP8[$470>>0] = $469; //@line 843 "load_med.cpp"
          $471 = ((($s$1117$us)) + 2|0); //@line 844 "load_med.cpp"
          $472 = HEAP8[$471>>0]|0; //@line 844 "load_med.cpp"
          $473 = ((($p$1116$us)) + 5|0); //@line 844 "load_med.cpp"
          HEAP8[$473>>0] = $472; //@line 844 "load_med.cpp"
          __ZL10MedConvertP11_MODCOMMANDPK17tagMMD0SONGHEADER($p$1116$us,$30); //@line 846 "load_med.cpp"
          $474 = ((($p$1116$us)) + 6|0); //@line 847 "load_med.cpp"
          $p$2$us = $474;
         } else {
          $p$2$us = $p$1116$us;
         }
         $475 = (($x$0119$us) + 1)|0; //@line 835 "load_med.cpp"
         $476 = ((($s$1117$us)) + 3|0); //@line 835 "load_med.cpp"
         $exitcond190 = ($475|0)==($tracks$0|0); //@line 835 "load_med.cpp"
         if ($exitcond190) {
          $p$2$us$lcssa = $p$2$us;
          break;
         } else {
          $p$1116$us = $p$2$us;$s$1117$us = $476;$x$0119$us = $475;
         }
        }
        $scevgep189 = (($s$0125$us) + ($442)|0);
        $443 = (($y$0126$us) + 1)|0; //@line 833 "load_med.cpp"
        $exitcond191 = ($443|0)==($424|0); //@line 833 "load_med.cpp"
        if ($exitcond191) {
         break L160;
        } else {
         $p$0124$us = $p$2$us$lcssa;$s$0125$us = $scevgep189;$y$0126$us = $443;
        }
       }
      }
      $$sum35 = (($416) + 2)|0; //@line 860 "load_med.cpp"
      $477 = (($lpStream) + ($$sum35)|0); //@line 860 "load_med.cpp"
      $478 = HEAPU8[$477>>0]|(HEAPU8[$477+1>>0]<<8); //@line 860 "load_med.cpp"
      $479 = $478&65535; //@line 860 "load_med.cpp"
      $480 = $479 >>> 8; //@line 860 "load_med.cpp"
      $481 = (($480) + 1)|0; //@line 860 "load_med.cpp"
      $482 = HEAPU8[$420>>0]|(HEAPU8[$420+1>>0]<<8); //@line 861 "load_med.cpp"
      $483 = $482&65535; //@line 861 "load_med.cpp"
      $484 = $483 >>> 8; //@line 861 "load_med.cpp"
      $485 = ($484|0)==(0); //@line 862 "load_med.cpp"
      if ($485) {
       $486 = HEAP32[$38>>2]|0; //@line 862 "load_med.cpp"
       $tracks$1 = $486;
      } else {
       $tracks$1 = $484;
      }
      $487 = HEAP32[$38>>2]|0; //@line 863 "load_med.cpp"
      $488 = (__ZN10CSoundFile15AllocatePatternEjj($481,$487)|0); //@line 863 "load_med.cpp"
      $489 = (((($this)) + 51904|0) + ($iBlk$0127<<2)|0); //@line 863 "load_med.cpp"
      HEAP32[$489>>2] = $488; //@line 863 "load_med.cpp"
      $490 = ($488|0)==(0|0); //@line 863 "load_med.cpp"
      if ($490) {
       break;
      }
      $491 = $481&65535; //@line 864 "load_med.cpp"
      $492 = (((($this)) + 52864|0) + ($iBlk$0127<<1)|0); //@line 864 "load_med.cpp"
      HEAP16[$492>>1] = $491; //@line 864 "load_med.cpp"
      $$sum36 = (($416) + 4)|0; //@line 865 "load_med.cpp"
      $493 = (($lpStream) + ($$sum36)|0); //@line 865 "load_med.cpp"
      $494 = HEAPU8[$493>>0]|(HEAPU8[$493+1>>0]<<8)|(HEAPU8[$493+2>>0]<<16)|(HEAPU8[$493+3>>0]<<24); //@line 865 "load_med.cpp"
      $495 = (_llvm_bswap_i32(($494|0))|0); //@line 865 "load_med.cpp"
      $496 = ($494|0)!=(0); //@line 866 "load_med.cpp"
      $497 = ($495>>>0)<($410>>>0); //@line 866 "load_med.cpp"
      $or$cond82 = $496 & $497; //@line 866 "load_med.cpp"
      do {
       if ($or$cond82) {
        $$sum38 = (($495) + 4)|0; //@line 873 "load_med.cpp"
        $498 = (($lpStream) + ($$sum38)|0); //@line 873 "load_med.cpp"
        $499 = HEAPU8[$498>>0]|(HEAPU8[$498+1>>0]<<8)|(HEAPU8[$498+2>>0]<<16)|(HEAPU8[$498+3>>0]<<24); //@line 873 "load_med.cpp"
        $500 = ($499|0)==(0); //@line 873 "load_med.cpp"
        do {
         if (!($500)) {
          $$sum40 = (($495) + 8)|0; //@line 873 "load_med.cpp"
          $501 = (($lpStream) + ($$sum40)|0); //@line 873 "load_med.cpp"
          $502 = HEAPU8[$501>>0]|(HEAPU8[$501+1>>0]<<8)|(HEAPU8[$501+2>>0]<<16)|(HEAPU8[$501+3>>0]<<24); //@line 873 "load_med.cpp"
          $503 = ($502|0)==(0); //@line 873 "load_med.cpp"
          if ($503) {
           break;
          }
          $504 = (_llvm_bswap_i32(($499|0))|0); //@line 875 "load_med.cpp"
          $505 = ($504>>>0)<($dwMemLength>>>0); //@line 877 "load_med.cpp"
          if (!($505)) {
           break;
          }
          $506 = (_llvm_bswap_i32(($502|0))|0); //@line 876 "load_med.cpp"
          $507 = (($504) + ($dwMemLength))|0; //@line 877 "load_med.cpp"
          $508 = ($506>>>0)<($507>>>0); //@line 877 "load_med.cpp"
          if (!($508)) {
           break;
          }
          $509 = (($lpStream) + ($504)|0); //@line 879 "load_med.cpp"
          (__ZN10CSoundFile14SetPatternNameEjPKc($this,$iBlk$0127,$509)|0); //@line 879 "load_med.cpp"
         }
        } while(0);
        $$sum39 = (($495) + 16)|0; //@line 882 "load_med.cpp"
        $510 = (($lpStream) + ($$sum39)|0); //@line 882 "load_med.cpp"
        $511 = HEAPU8[$510>>0]|(HEAPU8[$510+1>>0]<<8)|(HEAPU8[$510+2>>0]<<16)|(HEAPU8[$510+3>>0]<<24); //@line 882 "load_med.cpp"
        $512 = ($511|0)==(0); //@line 882 "load_med.cpp"
        if ($512) {
         $pcmdext$0 = 0;
         break;
        }
        $513 = (_llvm_bswap_i32(($511|0))|0); //@line 884 "load_med.cpp"
        $514 = ($513>>>0)<($411>>>0); //@line 885 "load_med.cpp"
        if (!($514)) {
         $pcmdext$0 = 0;
         break;
        }
        $515 = (($lpStream) + ($513)|0); //@line 887 "load_med.cpp"
        $516 = HEAP32[$515>>2]|0; //@line 887 "load_med.cpp"
        $517 = (_llvm_bswap_i32(($516|0))|0); //@line 887 "load_med.cpp"
        $518 = ($516|0)==(0); //@line 888 "load_med.cpp"
        if ($518) {
         $pcmdext$0 = 0;
         break;
        }
        $519 = Math_imul($tracks$1, $481)|0; //@line 888 "load_med.cpp"
        $520 = (($dwMemLength) - ($519))|0; //@line 888 "load_med.cpp"
        $521 = ($517>>>0)>($520>>>0); //@line 888 "load_med.cpp"
        $522 = (($lpStream) + ($517)|0); //@line 890 "load_med.cpp"
        $$83 = $521 ? 0 : $522; //@line 888 "load_med.cpp"
        $pcmdext$0 = $$83;
       } else {
        $pcmdext$0 = 0;
       }
      } while(0);
      $523 = $481 << 2; //@line 897 "load_med.cpp"
      $524 = Math_imul($523, $tracks$1)|0; //@line 897 "load_med.cpp"
      $525 = (($524) + ($416))|0; //@line 898 "load_med.cpp"
      $526 = ($525>>>0)>($407>>>0); //@line 898 "load_med.cpp"
      if ($526) {
       break L156;
      }
      $$sum37 = (($416) + 8)|0; //@line 896 "load_med.cpp"
      $527 = (($lpStream) + ($$sum37)|0); //@line 896 "load_med.cpp"
      $528 = HEAP32[$489>>2]|0; //@line 895 "load_med.cpp"
      $529 = ($tracks$1|0)==(0); //@line 901 "load_med.cpp"
      $530 = $tracks$1 << 2; //@line 899 "load_med.cpp"
      $p40$0113 = $528;$pcmdext$1110 = $pcmdext$0;$s41$0114 = $527;$y43$0115 = 0;
      while(1) {
       if ($529) {
        $p40$1$lcssa = $p40$0113;$s41$1$lcssa = $s41$0114;
       } else {
        $531 = ($pcmdext$1110|0)==(0|0); //@line 914 "load_med.cpp"
        if ($531) {
         $p40$1105$us = $p40$0113;$s41$1106$us = $s41$0114;$x44$0107$us = 0;
         while(1) {
          $532 = HEAP32[$38>>2]|0; //@line 901 "load_med.cpp"
          $533 = ($x44$0107$us>>>0)<($532>>>0); //@line 901 "load_med.cpp"
          if ($533) {
           $534 = HEAP8[$s41$1106$us>>0]|0; //@line 903 "load_med.cpp"
           $$off89$us = (($534) + -1)<<24>>24; //@line 904 "load_med.cpp"
           $535 = ($$off89$us&255)<(132); //@line 904 "load_med.cpp"
           if ($535) {
            $536 = $534&255; //@line 904 "load_med.cpp"
            $537 = (($536) + ($405))|0; //@line 906 "load_med.cpp"
            $538 = ($537|0)<(1); //@line 907 "load_med.cpp"
            $$22$us = $538 ? 1 : $537; //@line 907 "load_med.cpp"
            $539 = ($$22$us|0)>(120); //@line 908 "load_med.cpp"
            $540 = $$22$us&255; //@line 909 "load_med.cpp"
            $541 = $539 ? 120 : $540; //@line 909 "load_med.cpp"
            HEAP8[$p40$1105$us>>0] = $541; //@line 909 "load_med.cpp"
           }
           $542 = ((($s41$1106$us)) + 1|0); //@line 911 "load_med.cpp"
           $543 = HEAP8[$542>>0]|0; //@line 911 "load_med.cpp"
           $544 = ((($p40$1105$us)) + 1|0); //@line 911 "load_med.cpp"
           HEAP8[$544>>0] = $543; //@line 911 "load_med.cpp"
           $545 = ((($s41$1106$us)) + 2|0); //@line 912 "load_med.cpp"
           $546 = HEAP8[$545>>0]|0; //@line 912 "load_med.cpp"
           $547 = ((($p40$1105$us)) + 3|0); //@line 912 "load_med.cpp"
           HEAP8[$547>>0] = $546; //@line 912 "load_med.cpp"
           $548 = ((($s41$1106$us)) + 3|0); //@line 913 "load_med.cpp"
           $549 = HEAP8[$548>>0]|0; //@line 913 "load_med.cpp"
           $550 = ((($p40$1105$us)) + 5|0); //@line 913 "load_med.cpp"
           HEAP8[$550>>0] = $549; //@line 913 "load_med.cpp"
           __ZL10MedConvertP11_MODCOMMANDPK17tagMMD0SONGHEADER($p40$1105$us,$30); //@line 915 "load_med.cpp"
           $551 = ((($p40$1105$us)) + 6|0); //@line 916 "load_med.cpp"
           $p40$2$us = $551;
          } else {
           $p40$2$us = $p40$1105$us;
          }
          $552 = (($x44$0107$us) + 1)|0; //@line 901 "load_med.cpp"
          $553 = ((($s41$1106$us)) + 4|0); //@line 901 "load_med.cpp"
          $exitcond186 = ($552|0)==($tracks$1|0); //@line 901 "load_med.cpp"
          if ($exitcond186) {
           $split109 = $p40$2$us;
           break;
          } else {
           $p40$1105$us = $p40$2$us;$s41$1106$us = $553;$x44$0107$us = $552;
          }
         }
        } else {
         $p40$1105 = $p40$0113;$s41$1106 = $s41$0114;$x44$0107 = 0;
         while(1) {
          $554 = HEAP32[$38>>2]|0; //@line 901 "load_med.cpp"
          $555 = ($x44$0107>>>0)<($554>>>0); //@line 901 "load_med.cpp"
          if ($555) {
           $556 = HEAP8[$s41$1106>>0]|0; //@line 903 "load_med.cpp"
           $$off89 = (($556) + -1)<<24>>24; //@line 904 "load_med.cpp"
           $557 = ($$off89&255)<(132); //@line 904 "load_med.cpp"
           if ($557) {
            $558 = $556&255; //@line 904 "load_med.cpp"
            $559 = (($558) + ($405))|0; //@line 906 "load_med.cpp"
            $560 = ($559|0)<(1); //@line 907 "load_med.cpp"
            $$22 = $560 ? 1 : $559; //@line 907 "load_med.cpp"
            $561 = ($$22|0)>(120); //@line 908 "load_med.cpp"
            $562 = $$22&255; //@line 909 "load_med.cpp"
            $563 = $561 ? 120 : $562; //@line 909 "load_med.cpp"
            HEAP8[$p40$1105>>0] = $563; //@line 909 "load_med.cpp"
           }
           $564 = ((($s41$1106)) + 1|0); //@line 911 "load_med.cpp"
           $565 = HEAP8[$564>>0]|0; //@line 911 "load_med.cpp"
           $566 = ((($p40$1105)) + 1|0); //@line 911 "load_med.cpp"
           HEAP8[$566>>0] = $565; //@line 911 "load_med.cpp"
           $567 = ((($s41$1106)) + 2|0); //@line 912 "load_med.cpp"
           $568 = HEAP8[$567>>0]|0; //@line 912 "load_med.cpp"
           $569 = ((($p40$1105)) + 3|0); //@line 912 "load_med.cpp"
           HEAP8[$569>>0] = $568; //@line 912 "load_med.cpp"
           $570 = ((($s41$1106)) + 3|0); //@line 913 "load_med.cpp"
           $571 = HEAP8[$570>>0]|0; //@line 913 "load_med.cpp"
           $572 = ((($p40$1105)) + 5|0); //@line 913 "load_med.cpp"
           HEAP8[$572>>0] = $571; //@line 913 "load_med.cpp"
           $573 = (($pcmdext$1110) + ($x44$0107)|0); //@line 914 "load_med.cpp"
           $574 = HEAP8[$573>>0]|0; //@line 914 "load_med.cpp"
           $575 = ((($p40$1105)) + 4|0); //@line 914 "load_med.cpp"
           HEAP8[$575>>0] = $574; //@line 914 "load_med.cpp"
           __ZL10MedConvertP11_MODCOMMANDPK17tagMMD0SONGHEADER($p40$1105,$30); //@line 915 "load_med.cpp"
           $576 = ((($p40$1105)) + 6|0); //@line 916 "load_med.cpp"
           $p40$2 = $576;
          } else {
           $p40$2 = $p40$1105;
          }
          $577 = (($x44$0107) + 1)|0; //@line 901 "load_med.cpp"
          $578 = ((($s41$1106)) + 4|0); //@line 901 "load_med.cpp"
          $exitcond = ($577|0)==($tracks$1|0); //@line 901 "load_med.cpp"
          if ($exitcond) {
           $split109 = $p40$2;
           break;
          } else {
           $p40$1105 = $p40$2;$s41$1106 = $578;$x44$0107 = $577;
          }
         }
        }
        $split = (($s41$0114) + ($530)|0);
        $p40$1$lcssa = $split109;$s41$1$lcssa = $split;
       }
       $579 = ($pcmdext$1110|0)==(0|0); //@line 918 "load_med.cpp"
       $580 = (($pcmdext$1110) + ($tracks$1)|0); //@line 918 "load_med.cpp"
       $pcmdext$1$ = $579 ? 0 : $580; //@line 918 "load_med.cpp"
       $581 = (($y43$0115) + 1)|0; //@line 899 "load_med.cpp"
       $exitcond187 = ($581|0)==($481|0); //@line 899 "load_med.cpp"
       if ($exitcond187) {
        break;
       } else {
        $p40$0113 = $p40$1$lcssa;$pcmdext$1110 = $pcmdext$1$;$s41$0114 = $s41$1$lcssa;$y43$0115 = $581;
       }
      }
     }
    } while(0);
    $582 = (($iBlk$0127) + 1)|0; //@line 816 "load_med.cpp"
    $583 = ($582>>>0)<($$79>>>0); //@line 816 "load_med.cpp"
    if ($583) {
     $iBlk$0127 = $582;
    } else {
     break;
    }
   }
  }
 } while(0);
 $412 = HEAP32[$38>>2]|0; //@line 923 "load_med.cpp"
 $413 = ($412|0)==(0); //@line 923 "load_med.cpp"
 if ($413) {
  $$0 = 1;
  return ($$0|0); //@line 929 "load_med.cpp"
 } else {
  $iCh$0104 = 0;
 }
 while(1) {
  $584 = $iCh$0104 & 3; //@line 925 "load_med.cpp"
  $$off88 = (($584) + -1)|0; //@line 925 "load_med.cpp"
  $585 = ($$off88>>>0)<(2); //@line 925 "load_med.cpp"
  $586 = $585 ? 192 : 64; //@line 925 "load_med.cpp"
  $587 = (((($this)) + 49600|0) + (($iCh$0104*36)|0)|0); //@line 925 "load_med.cpp"
  HEAP32[$587>>2] = $586; //@line 925 "load_med.cpp"
  $588 = (((((($this)) + 49600|0) + (($iCh$0104*36)|0)|0)) + 4|0); //@line 926 "load_med.cpp"
  HEAP32[$588>>2] = 64; //@line 926 "load_med.cpp"
  $589 = (($iCh$0104) + 1)|0; //@line 923 "load_med.cpp"
  $590 = HEAP32[$38>>2]|0; //@line 923 "load_med.cpp"
  $591 = ($589>>>0)<($590>>>0); //@line 923 "load_med.cpp"
  if ($591) {
   $iCh$0104 = $589;
  } else {
   $$0 = 1;
   break;
  }
 }
 return ($$0|0); //@line 929 "load_med.cpp"
}
function __ZL10MedConvertP11_MODCOMMANDPK17tagMMD0SONGHEADER($p,$pmsh) {
 $p = $p|0;
 $pmsh = $pmsh|0;
 var $$ = 0, $$$param$2 = 0, $$1 = 0, $$2 = 0, $$4 = 0, $$5 = 0, $$6 = 0, $$op = 0, $$op15 = 0, $$op16 = 0, $$op17 = 0, $$op18 = 0, $$op19 = 0, $$op20 = 0, $$op21 = 0, $$param$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0;
 var $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $command$0 = 0, $command$1 = 0, $not$ = 0, $param$11 = 0, $param$2 = 0, $param$214 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($p)) + 3|0); //@line 272 "load_med.cpp"
 $1 = HEAP8[$0>>0]|0; //@line 272 "load_med.cpp"
 $2 = $1&255; //@line 272 "load_med.cpp"
 $3 = ((($p)) + 5|0); //@line 273 "load_med.cpp"
 $4 = HEAP8[$3>>0]|0; //@line 273 "load_med.cpp"
 $5 = $4&255; //@line 273 "load_med.cpp"
 L1: do {
  switch ($2|0) {
  case 0:  {
   $not$ = ($4<<24>>24)!=(0);
   $$4 = $not$&1;
   $command$1 = $$4;$param$11 = $5;
   break;
  }
  case 1:  {
   $command$1 = 2;$param$11 = $5;
   break;
  }
  case 2:  {
   $command$1 = 3;$param$11 = $5;
   break;
  }
  case 3:  {
   $command$1 = 4;$param$11 = $5;
   break;
  }
  case 4:  {
   $command$1 = 5;$param$11 = $5;
   break;
  }
  case 5:  {
   $command$1 = 6;$param$11 = $5;
   break;
  }
  case 6:  {
   $command$1 = 7;$param$11 = $5;
   break;
  }
  case 7:  {
   $command$1 = 8;$param$11 = $5;
   break;
  }
  case 10:  {
   $6 = $5 & 240; //@line 284 "load_med.cpp"
   $7 = ($6|0)==(0); //@line 284 "load_med.cpp"
   $$5 = $7 ? $5 : $6; //@line 284 "load_med.cpp"
   $8 = ($$5|0)!=(0); //@line 284 "load_med.cpp"
   $$ = $8 ? 11 : 0; //@line 284 "load_med.cpp"
   $command$1 = $$;$param$11 = $$5;
   break;
  }
  case 11:  {
   $command$1 = 12;$param$11 = $5;
   break;
  }
  case 12:  {
   $9 = ((($pmsh)) + 767|0); //@line 287 "load_med.cpp"
   $10 = HEAP8[$9>>0]|0; //@line 287 "load_med.cpp"
   $11 = $10 & 16; //@line 287 "load_med.cpp"
   $12 = ($11<<24>>24)==(0); //@line 287 "load_med.cpp"
   if ($12) {
    $16 = ($4&255)<(154); //@line 295 "load_med.cpp"
    if (!($16)) {
     $command$1 = 0;$param$11 = $5;
     break L1;
    }
    $17 = $5 >>> 4; //@line 297 "load_med.cpp"
    $18 = ($17*10)|0; //@line 297 "load_med.cpp"
    $19 = $5 & 15; //@line 297 "load_med.cpp"
    $20 = (($19>>>0) % 10)&-1; //@line 297 "load_med.cpp"
    $21 = (($20) + ($18))|0; //@line 297 "load_med.cpp"
    $22 = ($21>>>0)>(64); //@line 298 "load_med.cpp"
    $$1 = $22 ? 64 : $21; //@line 298 "load_med.cpp"
    $command$1 = 13;$param$11 = $$1;
    break L1;
   } else {
    $13 = ($4<<24>>24)>(-1); //@line 289 "load_med.cpp"
    if (!($13)) {
     $command$1 = 0;$param$11 = $5;
     break L1;
    }
    $14 = (($5) + 1)|0; //@line 291 "load_med.cpp"
    $15 = $14 >>> 1; //@line 291 "load_med.cpp"
    $command$1 = 13;$param$11 = $15;
    break L1;
   }
   break;
  }
  case 9:  {
   $23 = ($4&255)<(32); //@line 302 "load_med.cpp"
   $24 = $23 ? 16 : 17; //@line 302 "load_med.cpp"
   $command$1 = $24;$param$11 = $5;
   break;
  }
  case 13:  {
   $25 = $5 & 240; //@line 303 "load_med.cpp"
   $26 = ($25|0)==(0); //@line 303 "load_med.cpp"
   $$6 = $26 ? $5 : $25; //@line 303 "load_med.cpp"
   $27 = ($$6|0)!=(0); //@line 303 "load_med.cpp"
   $$2 = $27 ? 11 : 0; //@line 303 "load_med.cpp"
   $command$1 = $$2;$param$11 = $$6;
   break;
  }
  case 15:  {
   $28 = ($4<<24>>24)==(0); //@line 306 "load_med.cpp"
   if ($28) {
    $command$1 = 14;$param$11 = $5;
   } else {
    $29 = ($4&255)<(241); //@line 308 "load_med.cpp"
    if ($29) {
     $30 = ((($pmsh)) + 767|0); //@line 310 "load_med.cpp"
     $31 = HEAP8[$30>>0]|0; //@line 310 "load_med.cpp"
     $32 = $31 & 64; //@line 310 "load_med.cpp"
     $33 = ($32<<24>>24)==(0); //@line 310 "load_med.cpp"
     if ($33) {
      $39 = ($4&255)<(11); //@line 315 "load_med.cpp"
      if ($39) {
       $command$0 = 16;$param$2 = $5;
       label = 26;
      } else {
       $40 = ((($pmsh)) + 768|0); //@line 320 "load_med.cpp"
       $41 = HEAP8[$40>>0]|0; //@line 320 "load_med.cpp"
       $42 = $41 & 32; //@line 320 "load_med.cpp"
       $43 = ($42<<24>>24)==(0); //@line 320 "load_med.cpp"
       if ($43) {
        $44 = (__Z7_muldivlll($5,3579545,948652)|0); //@line 322 "load_med.cpp"
        $command$0 = 15;$param$2 = $44;
        label = 26;
       } else {
        $command$0 = 15;$param$2 = $5;
        label = 26;
       }
      }
     } else {
      $34 = ($4&255)>(10); //@line 312 "load_med.cpp"
      if ($34) {
       $param$214 = 99;
      } else {
       $35 = (($5) + -1)|0; //@line 312 "load_med.cpp"
       $36 = (529839 + ($35)|0); //@line 312 "load_med.cpp"
       $37 = HEAP8[$36>>0]|0; //@line 312 "load_med.cpp"
       $38 = $37&255; //@line 312 "load_med.cpp"
       $command$0 = 15;$param$2 = $38;
       label = 26;
      }
     }
     if ((label|0) == 26) {
      $45 = ($param$2>>>0)>(10); //@line 325 "load_med.cpp"
      if ($45) {
       $param$214 = $param$2;
      } else {
       $command$1 = $command$0;$param$11 = $param$2;
       break L1;
      }
     }
     $46 = ($param$214>>>0)<(33); //@line 328 "load_med.cpp"
     $$param$2 = $46 ? 33 : $param$214; //@line 328 "load_med.cpp"
     $47 = ($$param$2>>>0)>(240); //@line 329 "load_med.cpp"
     $$$param$2 = $47 ? 240 : $$param$2; //@line 329 "load_med.cpp"
     $command$1 = 17;$param$11 = $$$param$2;
     break L1;
    }
    do {
     switch ($5|0) {
     case 241:  {
      $command$1 = 19;$param$11 = 147;
      break L1;
      break;
     }
     case 242:  {
      $command$1 = 19;$param$11 = 211;
      break L1;
      break;
     }
     case 243:  {
      $command$1 = 19;$param$11 = 146;
      break L1;
      break;
     }
     case 244:  {
      $command$1 = 19;$param$11 = 210;
      break L1;
      break;
     }
     case 245:  {
      $command$1 = 19;$param$11 = 212;
      break L1;
      break;
     }
     case 248:  {
      $command$1 = 19;$param$11 = 0;
      break L1;
      break;
     }
     case 249:  {
      $command$1 = 19;$param$11 = 1;
      break L1;
      break;
     }
     case 253:  {
      $command$1 = 4;$param$11 = 255;
      break L1;
      break;
     }
     case 254:  {
      $command$1 = 16;$param$11 = 0;
      break L1;
      break;
     }
     case 255:  {
      $command$1 = 19;$param$11 = 192;
      break L1;
      break;
     }
     default: {
      $command$1 = 0;$param$11 = 0;
      break L1;
     }
     }
    } while(0);
   }
   break;
  }
  case 17:  {
   $48 = ($4&255)>(15); //@line 394 "load_med.cpp"
   $$op21 = $5 | 16; //@line 395 "load_med.cpp"
   $49 = $48 ? 31 : $$op21; //@line 395 "load_med.cpp"
   $command$1 = 19;$param$11 = $49;
   break;
  }
  case 18:  {
   $50 = ($4&255)>(15); //@line 400 "load_med.cpp"
   $$op20 = $5 | 32; //@line 401 "load_med.cpp"
   $51 = $50 ? 47 : $$op20; //@line 401 "load_med.cpp"
   $command$1 = 19;$param$11 = $51;
   break;
  }
  case 20:  {
   $command$1 = 5;$param$11 = $5;
   break;
  }
  case 21:  {
   $52 = $5 & 15; //@line 410 "load_med.cpp"
   $53 = $52 | 80; //@line 411 "load_med.cpp"
   $command$1 = 19;$param$11 = $53;
   break;
  }
  case 22:  {
   $54 = ($4&255)>(15); //@line 416 "load_med.cpp"
   $$op19 = $5 | 96; //@line 417 "load_med.cpp"
   $55 = $54 ? 111 : $$op19; //@line 417 "load_med.cpp"
   $command$1 = 19;$param$11 = $55;
   break;
  }
  case 24:  {
   $56 = ($4&255)>(15); //@line 422 "load_med.cpp"
   $$op18 = $5 | 192; //@line 423 "load_med.cpp"
   $57 = $56 ? 207 : $$op18; //@line 423 "load_med.cpp"
   $command$1 = 19;$param$11 = $57;
   break;
  }
  case 25:  {
   $command$1 = 10;$param$11 = $5;
   break;
  }
  case 26:  {
   $58 = ($4&255)>(15); //@line 432 "load_med.cpp"
   $$op17 = $5 | 160; //@line 433 "load_med.cpp"
   $59 = $58 ? 175 : $$op17; //@line 433 "load_med.cpp"
   $command$1 = 19;$param$11 = $59;
   break;
  }
  case 27:  {
   $60 = ($4&255)>(15); //@line 438 "load_med.cpp"
   $$op16 = $5 | 176; //@line 439 "load_med.cpp"
   $61 = $60 ? 191 : $$op16; //@line 439 "load_med.cpp"
   $command$1 = 19;$param$11 = $61;
   break;
  }
  case 29:  {
   $command$1 = 14;$param$11 = $5;
   break;
  }
  case 30:  {
   $62 = ($4&255)>(15); //@line 448 "load_med.cpp"
   $$op15 = $5 | 224; //@line 449 "load_med.cpp"
   $63 = $62 ? 239 : $$op15; //@line 449 "load_med.cpp"
   $command$1 = 19;$param$11 = $63;
   break;
  }
  case 31:  {
   $64 = $5 & 15; //@line 454 "load_med.cpp"
   $command$1 = 15;$param$11 = $64;
   break;
  }
  case 46:  {
   $65 = (($5) + 16)|0; //@line 459 "load_med.cpp"
   $66 = $65 >>> 1; //@line 459 "load_med.cpp"
   $67 = $66 & 112; //@line 459 "load_med.cpp"
   $68 = ($67>>>0)>(15); //@line 460 "load_med.cpp"
   $$op = $66 | 128; //@line 461 "load_med.cpp"
   $69 = $68 ? 143 : $$op; //@line 461 "load_med.cpp"
   $command$1 = 19;$param$11 = $69;
   break;
  }
  default: {
   $command$1 = 0;$param$11 = 0;
  }
  }
 } while(0);
 $70 = $command$1&255; //@line 470 "load_med.cpp"
 HEAP8[$0>>0] = $70; //@line 470 "load_med.cpp"
 $71 = $param$11&255; //@line 471 "load_med.cpp"
 HEAP8[$3>>0] = $71; //@line 471 "load_med.cpp"
 return; //@line 472 "load_med.cpp"
}
function __ZN10CSoundFile14SetPatternNameEjPKc($this,$nPat,$lpszName) {
 $this = $this|0;
 $nPat = $nPat|0;
 $lpszName = $lpszName|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $szName = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $szName = sp;
 dest=$szName; stop=dest+32|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 1764 "sndfile.cpp"
 $0 = ($nPat>>>0)>(239); //@line 1765 "sndfile.cpp"
 if ($0) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 1787 "sndfile.cpp"
 }
 $1 = ($lpszName|0)==(0|0); //@line 1766 "sndfile.cpp"
 if (!($1)) {
  (_strncpy($szName,$lpszName,32)|0); //@line 1766 "sndfile.cpp"
 }
 $2 = ((($szName)) + 31|0); //@line 1767 "sndfile.cpp"
 HEAP8[$2>>0] = 0; //@line 1767 "sndfile.cpp"
 $3 = ((($this)) + 59800|0); //@line 1768 "sndfile.cpp"
 $4 = HEAP32[$3>>2]|0; //@line 1768 "sndfile.cpp"
 $5 = ($4|0)==(0|0); //@line 1768 "sndfile.cpp"
 $6 = ((($this)) + 59792|0); //@line 1768 "sndfile.cpp"
 if ($5) {
  HEAP32[$6>>2] = 0; //@line 1768 "sndfile.cpp"
  $14 = 0;
  label = 7;
 } else {
  $$pre = HEAP32[$6>>2]|0; //@line 1769 "sndfile.cpp"
  $7 = ($$pre>>>0)>($nPat>>>0); //@line 1769 "sndfile.cpp"
  if ($7) {
   $17 = $4;
  } else {
   $14 = $$pre;
   label = 7;
  }
 }
 if ((label|0) == 7) {
  $8 = HEAP8[$lpszName>>0]|0; //@line 1771 "sndfile.cpp"
  $9 = ($8<<24>>24)==(0); //@line 1771 "sndfile.cpp"
  if ($9) {
   $$0 = 1;
   STACKTOP = sp;return ($$0|0); //@line 1787 "sndfile.cpp"
  }
  $10 = (($nPat) + 1)|0; //@line 1772 "sndfile.cpp"
  $11 = $10 << 5; //@line 1772 "sndfile.cpp"
  $12 = (__Znaj($11)|0); //@line 1773 "sndfile.cpp"
  _memset(($12|0),0,($11|0))|0; //@line 1775 "sndfile.cpp"
  if (!($5)) {
   $13 = $14 << 5; //@line 1778 "sndfile.cpp"
   _memcpy(($12|0),($4|0),($13|0))|0; //@line 1778 "sndfile.cpp"
   __ZdaPv($4); //@line 1779 "sndfile.cpp"
   HEAP32[$3>>2] = 0; //@line 1780 "sndfile.cpp"
  }
  HEAP32[$3>>2] = $12; //@line 1782 "sndfile.cpp"
  HEAP32[$6>>2] = $10; //@line 1783 "sndfile.cpp"
  $17 = $12;
 }
 $15 = $nPat << 5; //@line 1785 "sndfile.cpp"
 $16 = (($17) + ($15)|0); //@line 1785 "sndfile.cpp"
 dest=$16; src=$szName; stop=dest+32|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 1785 "sndfile.cpp"
 $$0 = 1;
 STACKTOP = sp;return ($$0|0); //@line 1787 "sndfile.cpp"
}
function __ZN10CSoundFile7ReadMTMEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$0 = 0, $$17 = 0, $$lcssa58 = 0, $$off = 0, $$pre = 0, $$pre43 = 0, $$pre44 = 0, $$pre46 = 0, $$pre49 = 0, $$pre50$pre = 0, $$sum = 0, $$sum10 = 0, $$sum11 = 0, $$sum12 = 0, $$sum6 = 0, $$sum7 = 0, $$sum8 = 0, $$sum9 = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dwMemPos$0$lcssa = 0, $dwMemPos$037 = 0, $dwMemPos$121 = 0, $exitcond = 0, $exitcond41 = 0, $exitcond42 = 0, $i$035 = 0, $i1$028 = 0, $i3$023 = 0;
 var $ich$032 = 0, $ismp$022 = 0, $m$027 = 0, $n$029 = 0, $or$cond = 0, $or$cond14 = 0, $or$cond16 = 0, $or$cond1620 = 0, $p$026 = 0, $pSeq$030 = 0, $param$0 = 0, $pat$031 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($lpStream|0)==(0|0); //@line 55 "load_mtm.cpp"
 $1 = ($dwMemLength>>>0)<(256); //@line 55 "load_mtm.cpp"
 $or$cond = $0 | $1; //@line 55 "load_mtm.cpp"
 if ($or$cond) {
  $$0 = 0;
  return ($$0|0); //@line 167 "load_mtm.cpp"
 }
 $2 = (_strncmp($lpStream,529849,3)|0); //@line 56 "load_mtm.cpp"
 $3 = ($2|0)==(0); //@line 56 "load_mtm.cpp"
 if (!($3)) {
  $$0 = 0;
  return ($$0|0); //@line 167 "load_mtm.cpp"
 }
 $4 = ((($lpStream)) + 33|0); //@line 56 "load_mtm.cpp"
 $5 = HEAP8[$4>>0]|0; //@line 56 "load_mtm.cpp"
 $6 = ($5&255)>(32); //@line 56 "load_mtm.cpp"
 if ($6) {
  $$0 = 0;
  return ($$0|0); //@line 167 "load_mtm.cpp"
 }
 $7 = ((($lpStream)) + 30|0); //@line 57 "load_mtm.cpp"
 $8 = HEAP8[$7>>0]|0; //@line 57 "load_mtm.cpp"
 $9 = (($8) + -1)<<24>>24; //@line 57 "load_mtm.cpp"
 $10 = ($9&255)>(238); //@line 57 "load_mtm.cpp"
 if ($10) {
  $$0 = 0;
  return ($$0|0); //@line 167 "load_mtm.cpp"
 }
 $11 = ((($lpStream)) + 24|0); //@line 58 "load_mtm.cpp"
 $12 = HEAPU8[$11>>0]|(HEAPU8[$11+1>>0]<<8); //@line 58 "load_mtm.cpp"
 $13 = ($12<<16>>16)==(0); //@line 58 "load_mtm.cpp"
 $14 = ($5<<24>>24)==(0); //@line 58 "load_mtm.cpp"
 $or$cond14 = $14 | $13; //@line 58 "load_mtm.cpp"
 if ($or$cond14) {
  $$0 = 0;
  return ($$0|0); //@line 167 "load_mtm.cpp"
 }
 $15 = ((($lpStream)) + 26|0); //@line 59 "load_mtm.cpp"
 $16 = HEAP8[$15>>0]|0; //@line 59 "load_mtm.cpp"
 $17 = (($16) + -1)<<24>>24; //@line 59 "load_mtm.cpp"
 $18 = ($17&255)>(239); //@line 59 "load_mtm.cpp"
 if ($18) {
  $$0 = 0;
  return ($$0|0); //@line 167 "load_mtm.cpp"
 }
 $19 = ((($this)) + 59804|0); //@line 61 "load_mtm.cpp"
 $20 = ((($lpStream)) + 4|0); //@line 61 "load_mtm.cpp"
 (_strncpy($19,$20,20)|0); //@line 61 "load_mtm.cpp"
 $21 = ((($this)) + 59824|0); //@line 62 "load_mtm.cpp"
 HEAP8[$21>>0] = 0; //@line 62 "load_mtm.cpp"
 $22 = HEAP8[$7>>0]|0; //@line 63 "load_mtm.cpp"
 $23 = $22&255; //@line 63 "load_mtm.cpp"
 $24 = ($23*37)|0; //@line 63 "load_mtm.cpp"
 $25 = HEAPU8[$11>>0]|(HEAPU8[$11+1>>0]<<8); //@line 63 "load_mtm.cpp"
 $26 = $25&65535; //@line 63 "load_mtm.cpp"
 $27 = ($26*192)|0; //@line 63 "load_mtm.cpp"
 $28 = HEAP8[$15>>0]|0; //@line 64 "load_mtm.cpp"
 $29 = $28&255; //@line 64 "load_mtm.cpp"
 $30 = $29 << 6; //@line 64 "load_mtm.cpp"
 $31 = ((($lpStream)) + 28|0); //@line 64 "load_mtm.cpp"
 $32 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8); //@line 64 "load_mtm.cpp"
 $33 = $32&65535; //@line 64 "load_mtm.cpp"
 $34 = (($24) + 258)|0; //@line 63 "load_mtm.cpp"
 $35 = (($34) + ($27))|0; //@line 63 "load_mtm.cpp"
 $36 = (($35) + ($30))|0; //@line 64 "load_mtm.cpp"
 $37 = (($36) + ($33))|0; //@line 64 "load_mtm.cpp"
 $38 = ($37>>>0)<($dwMemLength>>>0); //@line 64 "load_mtm.cpp"
 if (!($38)) {
  $$0 = 0;
  return ($$0|0); //@line 167 "load_mtm.cpp"
 }
 $39 = ((($this)) + 59680|0); //@line 66 "load_mtm.cpp"
 HEAP32[$39>>2] = 16; //@line 66 "load_mtm.cpp"
 $40 = ((($this)) + 59684|0); //@line 67 "load_mtm.cpp"
 HEAP32[$40>>2] = $23; //@line 67 "load_mtm.cpp"
 $41 = HEAP8[$4>>0]|0; //@line 68 "load_mtm.cpp"
 $42 = $41&255; //@line 68 "load_mtm.cpp"
 $43 = ((($this)) + 59664|0); //@line 68 "load_mtm.cpp"
 HEAP32[$43>>2] = $42; //@line 68 "load_mtm.cpp"
 $44 = ($22<<24>>24)==(0); //@line 70 "load_mtm.cpp"
 if ($44) {
  $46 = $42;$dwMemPos$0$lcssa = 66;
 } else {
  $dwMemPos$037 = 66;$i$035 = 1;
  while(1) {
   $47 = (($lpStream) + ($dwMemPos$037)|0); //@line 72 "load_mtm.cpp"
   $48 = (((($this)) + 59804|0) + ($i$035<<5)|0); //@line 73 "load_mtm.cpp"
   (_strncpy($48,$47,22)|0); //@line 73 "load_mtm.cpp"
   $49 = (((((($this)) + 59804|0) + ($i$035<<5)|0)) + 22|0); //@line 74 "load_mtm.cpp"
   HEAP8[$49>>0] = 0; //@line 74 "load_mtm.cpp"
   $$sum7 = (($dwMemPos$037) + 35)|0; //@line 75 "load_mtm.cpp"
   $50 = (($lpStream) + ($$sum7)|0); //@line 75 "load_mtm.cpp"
   $51 = HEAP8[$50>>0]|0; //@line 75 "load_mtm.cpp"
   $52 = $51&255; //@line 75 "load_mtm.cpp"
   $53 = $52 << 2; //@line 75 "load_mtm.cpp"
   $54 = $53&65535; //@line 75 "load_mtm.cpp"
   $55 = (((((($this)) + 33280|0) + ($i$035<<6)|0)) + 30|0); //@line 75 "load_mtm.cpp"
   HEAP16[$55>>1] = $54; //@line 75 "load_mtm.cpp"
   $56 = (((((($this)) + 33280|0) + ($i$035<<6)|0)) + 32|0); //@line 76 "load_mtm.cpp"
   HEAP16[$56>>1] = 64; //@line 76 "load_mtm.cpp"
   $$sum8 = (($dwMemPos$037) + 22)|0; //@line 77 "load_mtm.cpp"
   $57 = (($lpStream) + ($$sum8)|0); //@line 77 "load_mtm.cpp"
   $58 = HEAPU8[$57>>0]|(HEAPU8[$57+1>>0]<<8)|(HEAPU8[$57+2>>0]<<16)|(HEAPU8[$57+3>>0]<<24); //@line 77 "load_mtm.cpp"
   $$off = (($58) + -5)|0; //@line 78 "load_mtm.cpp"
   $59 = ($$off>>>0)<(15999996); //@line 78 "load_mtm.cpp"
   if ($59) {
    $60 = (((($this)) + 33280|0) + ($i$035<<6)|0); //@line 80 "load_mtm.cpp"
    HEAP32[$60>>2] = $58; //@line 80 "load_mtm.cpp"
    $$sum9 = (($dwMemPos$037) + 26)|0; //@line 81 "load_mtm.cpp"
    $61 = (($lpStream) + ($$sum9)|0); //@line 81 "load_mtm.cpp"
    $62 = HEAPU8[$61>>0]|(HEAPU8[$61+1>>0]<<8)|(HEAPU8[$61+2>>0]<<16)|(HEAPU8[$61+3>>0]<<24); //@line 81 "load_mtm.cpp"
    $63 = (((((($this)) + 33280|0) + ($i$035<<6)|0)) + 4|0); //@line 81 "load_mtm.cpp"
    HEAP32[$63>>2] = $62; //@line 81 "load_mtm.cpp"
    $$sum10 = (($dwMemPos$037) + 30)|0; //@line 82 "load_mtm.cpp"
    $64 = (($lpStream) + ($$sum10)|0); //@line 82 "load_mtm.cpp"
    $65 = HEAPU8[$64>>0]|(HEAPU8[$64+1>>0]<<8)|(HEAPU8[$64+2>>0]<<16)|(HEAPU8[$64+3>>0]<<24); //@line 82 "load_mtm.cpp"
    $66 = (((((($this)) + 33280|0) + ($i$035<<6)|0)) + 8|0); //@line 82 "load_mtm.cpp"
    $67 = ($65>>>0)>($58>>>0); //@line 83 "load_mtm.cpp"
    $$ = $67 ? $58 : $65; //@line 83 "load_mtm.cpp"
    HEAP32[$66>>2] = $$; //@line 84 "load_mtm.cpp"
    $68 = (($62) + 4)|0; //@line 85 "load_mtm.cpp"
    $69 = ($68>>>0)<($$>>>0); //@line 85 "load_mtm.cpp"
    if ($69) {
     $70 = ($$|0)==(0); //@line 87 "load_mtm.cpp"
     if ($70) {
      $93 = $62;$95 = 0;
     } else {
      $71 = (((((($this)) + 33280|0) + ($i$035<<6)|0)) + 34|0); //@line 87 "load_mtm.cpp"
      $72 = HEAP16[$71>>1]|0; //@line 87 "load_mtm.cpp"
      $73 = $72&65535; //@line 87 "load_mtm.cpp"
      $74 = $73 | 2; //@line 87 "load_mtm.cpp"
      $75 = $74&65535; //@line 87 "load_mtm.cpp"
      HEAP16[$71>>1] = $75; //@line 87 "load_mtm.cpp"
      $93 = $62;$95 = $$;
     }
    } else {
     HEAP32[$66>>2] = 0; //@line 86 "load_mtm.cpp"
     HEAP32[$63>>2] = 0; //@line 86 "load_mtm.cpp"
     $93 = 0;$95 = 0;
    }
    $$sum11 = (($dwMemPos$037) + 34)|0; //@line 88 "load_mtm.cpp"
    $76 = (($lpStream) + ($$sum11)|0); //@line 88 "load_mtm.cpp"
    $77 = HEAP8[$76>>0]|0; //@line 88 "load_mtm.cpp"
    $78 = $77 << 24 >> 24; //@line 88 "load_mtm.cpp"
    $79 = $78 << 4; //@line 88 "load_mtm.cpp"
    $80 = $79&255; //@line 88 "load_mtm.cpp"
    $81 = (((((($this)) + 33280|0) + ($i$035<<6)|0)) + 37|0); //@line 88 "load_mtm.cpp"
    HEAP8[$81>>0] = $80; //@line 88 "load_mtm.cpp"
    $$sum12 = (($dwMemPos$037) + 36)|0; //@line 89 "load_mtm.cpp"
    $82 = (($lpStream) + ($$sum12)|0); //@line 89 "load_mtm.cpp"
    $83 = HEAP8[$82>>0]|0; //@line 89 "load_mtm.cpp"
    $84 = $83 & 1; //@line 89 "load_mtm.cpp"
    $85 = ($84<<24>>24)==(0); //@line 89 "load_mtm.cpp"
    if (!($85)) {
     $86 = (((((($this)) + 33280|0) + ($i$035<<6)|0)) + 34|0); //@line 91 "load_mtm.cpp"
     $87 = HEAP16[$86>>1]|0; //@line 91 "load_mtm.cpp"
     $88 = $87&65535; //@line 91 "load_mtm.cpp"
     $89 = $88 | 1; //@line 91 "load_mtm.cpp"
     $90 = $89&65535; //@line 91 "load_mtm.cpp"
     HEAP16[$86>>1] = $90; //@line 91 "load_mtm.cpp"
     $91 = $58 >>> 1; //@line 92 "load_mtm.cpp"
     HEAP32[$60>>2] = $91; //@line 92 "load_mtm.cpp"
     $92 = $93 >>> 1; //@line 93 "load_mtm.cpp"
     HEAP32[$63>>2] = $92; //@line 93 "load_mtm.cpp"
     $94 = $95 >>> 1; //@line 94 "load_mtm.cpp"
     HEAP32[$66>>2] = $94; //@line 94 "load_mtm.cpp"
    }
    $96 = (((((($this)) + 33280|0) + ($i$035<<6)|0)) + 28|0); //@line 96 "load_mtm.cpp"
    HEAP16[$96>>1] = 128; //@line 96 "load_mtm.cpp"
   }
   $97 = (($dwMemPos$037) + 37)|0; //@line 98 "load_mtm.cpp"
   $98 = (($i$035) + 1)|0; //@line 70 "load_mtm.cpp"
   $99 = HEAP32[$40>>2]|0; //@line 70 "load_mtm.cpp"
   $100 = ($98>>>0)>($99>>>0); //@line 70 "load_mtm.cpp"
   if ($100) {
    $$lcssa58 = $97;
    break;
   } else {
    $dwMemPos$037 = $97;$i$035 = $98;
   }
  }
  $$pre = HEAP32[$43>>2]|0; //@line 101 "load_mtm.cpp"
  $46 = $$pre;$dwMemPos$0$lcssa = $$lcssa58;
 }
 $45 = ($46|0)==(0); //@line 101 "load_mtm.cpp"
 if (!($45)) {
  $ich$032 = 0;
  while(1) {
   $$sum6 = (($ich$032) + 34)|0; //@line 103 "load_mtm.cpp"
   $101 = (($lpStream) + ($$sum6)|0); //@line 103 "load_mtm.cpp"
   $102 = HEAP8[$101>>0]|0; //@line 103 "load_mtm.cpp"
   $103 = $102&255; //@line 103 "load_mtm.cpp"
   $104 = $103 << 4; //@line 103 "load_mtm.cpp"
   $105 = $104 & 240; //@line 103 "load_mtm.cpp"
   $106 = $105 | 8; //@line 103 "load_mtm.cpp"
   $107 = (((($this)) + 49600|0) + (($ich$032*36)|0)|0); //@line 103 "load_mtm.cpp"
   HEAP32[$107>>2] = $106; //@line 103 "load_mtm.cpp"
   $108 = (((((($this)) + 49600|0) + (($ich$032*36)|0)|0)) + 4|0); //@line 104 "load_mtm.cpp"
   HEAP32[$108>>2] = 64; //@line 104 "load_mtm.cpp"
   $109 = (($ich$032) + 1)|0; //@line 101 "load_mtm.cpp"
   $110 = HEAP32[$43>>2]|0; //@line 101 "load_mtm.cpp"
   $111 = ($109>>>0)<($110>>>0); //@line 101 "load_mtm.cpp"
   if ($111) {
    $ich$032 = $109;
   } else {
    break;
   }
  }
 }
 $112 = ((($this)) + 53344|0); //@line 107 "load_mtm.cpp"
 $113 = (($lpStream) + ($dwMemPos$0$lcssa)|0); //@line 107 "load_mtm.cpp"
 $114 = ((($lpStream)) + 27|0); //@line 107 "load_mtm.cpp"
 $115 = HEAP8[$114>>0]|0; //@line 107 "load_mtm.cpp"
 $116 = $115&255; //@line 107 "load_mtm.cpp"
 $117 = (($116) + 1)|0; //@line 107 "load_mtm.cpp"
 _memcpy(($112|0),($113|0),($117|0))|0; //@line 107 "load_mtm.cpp"
 $118 = (($dwMemPos$0$lcssa) + 128)|0; //@line 108 "load_mtm.cpp"
 $119 = HEAPU8[$11>>0]|(HEAPU8[$11+1>>0]<<8); //@line 111 "load_mtm.cpp"
 $120 = $119&65535; //@line 111 "load_mtm.cpp"
 $121 = ($120*192)|0; //@line 111 "load_mtm.cpp"
 $122 = (($121) + ($118))|0; //@line 111 "load_mtm.cpp"
 $123 = (($lpStream) + ($122)|0); //@line 112 "load_mtm.cpp"
 $124 = (($dwMemPos$0$lcssa) + -64)|0; //@line 119 "load_mtm.cpp"
 $pSeq$030 = $123;$pat$031 = 0;
 while(1) {
  $125 = (((($this)) + 52864|0) + ($pat$031<<1)|0); //@line 115 "load_mtm.cpp"
  HEAP16[$125>>1] = 64; //@line 115 "load_mtm.cpp"
  $126 = HEAP32[$43>>2]|0; //@line 116 "load_mtm.cpp"
  $127 = (__ZN10CSoundFile15AllocatePatternEjj(64,$126)|0); //@line 116 "load_mtm.cpp"
  $128 = (((($this)) + 51904|0) + ($pat$031<<2)|0); //@line 116 "load_mtm.cpp"
  HEAP32[$128>>2] = $127; //@line 116 "load_mtm.cpp"
  $129 = ($127|0)==(0|0); //@line 116 "load_mtm.cpp"
  if ($129) {
   label = 23;
   break;
  } else {
   $n$029 = 0;
  }
  while(1) {
   $130 = (($pSeq$030) + ($n$029<<1)|0); //@line 117 "load_mtm.cpp"
   $131 = HEAP16[$130>>1]|0; //@line 117 "load_mtm.cpp"
   $132 = ($131<<16>>16)==(0); //@line 117 "load_mtm.cpp"
   if (!($132)) {
    $133 = $131&65535; //@line 117 "load_mtm.cpp"
    $134 = HEAPU8[$11>>0]|(HEAPU8[$11+1>>0]<<8); //@line 117 "load_mtm.cpp"
    $135 = ($131&65535)>($134&65535); //@line 117 "load_mtm.cpp"
    if (!($135)) {
     $136 = HEAP32[$43>>2]|0; //@line 117 "load_mtm.cpp"
     $137 = ($n$029>>>0)<($136>>>0); //@line 117 "load_mtm.cpp"
     if ($137) {
      $138 = ($133*192)|0; //@line 119 "load_mtm.cpp"
      $$sum = (($124) + ($138))|0; //@line 119 "load_mtm.cpp"
      $139 = (($lpStream) + ($$sum)|0); //@line 119 "load_mtm.cpp"
      $140 = HEAP32[$128>>2]|0; //@line 120 "load_mtm.cpp"
      $141 = (($140) + (($n$029*6)|0)|0); //@line 120 "load_mtm.cpp"
      $231 = $136;$i1$028 = 0;$m$027 = $141;$p$026 = $139;
      while(1) {
       $142 = HEAP8[$p$026>>0]|0; //@line 123 "load_mtm.cpp"
       $143 = $142&255; //@line 123 "load_mtm.cpp"
       $144 = $143 & 252; //@line 123 "load_mtm.cpp"
       $145 = ($144|0)==(0); //@line 123 "load_mtm.cpp"
       if ($145) {
        $150 = $142;
       } else {
        $146 = $143 >>> 2; //@line 123 "load_mtm.cpp"
        $147 = (($146) + 37)|0; //@line 123 "load_mtm.cpp"
        $148 = $147&255; //@line 123 "load_mtm.cpp"
        HEAP8[$m$027>>0] = $148; //@line 123 "load_mtm.cpp"
        $$pre43 = HEAP8[$p$026>>0]|0; //@line 124 "load_mtm.cpp"
        $150 = $$pre43;
       }
       $149 = $150&255; //@line 124 "load_mtm.cpp"
       $151 = $149 << 4; //@line 124 "load_mtm.cpp"
       $152 = $151 & 48; //@line 124 "load_mtm.cpp"
       $153 = ((($p$026)) + 1|0); //@line 124 "load_mtm.cpp"
       $154 = HEAP8[$153>>0]|0; //@line 124 "load_mtm.cpp"
       $155 = $154&255; //@line 124 "load_mtm.cpp"
       $156 = $155 >>> 4; //@line 124 "load_mtm.cpp"
       $157 = $152 | $156; //@line 124 "load_mtm.cpp"
       $158 = $157&255; //@line 124 "load_mtm.cpp"
       $159 = ((($m$027)) + 1|0); //@line 124 "load_mtm.cpp"
       HEAP8[$159>>0] = $158; //@line 124 "load_mtm.cpp"
       $160 = HEAP8[$153>>0]|0; //@line 125 "load_mtm.cpp"
       $161 = $160&255; //@line 125 "load_mtm.cpp"
       $162 = $161 & 15; //@line 125 "load_mtm.cpp"
       $163 = ((($p$026)) + 2|0); //@line 126 "load_mtm.cpp"
       $164 = HEAP8[$163>>0]|0; //@line 126 "load_mtm.cpp"
       $165 = $164&255; //@line 126 "load_mtm.cpp"
       $166 = ($162|0)==(10); //@line 127 "load_mtm.cpp"
       if ($166) {
        $167 = $165 & 240; //@line 129 "load_mtm.cpp"
        $168 = ($167|0)==(0); //@line 129 "load_mtm.cpp"
        $169 = $165 & 15; //@line 129 "load_mtm.cpp"
        $$17 = $168 ? $169 : $167; //@line 129 "load_mtm.cpp"
        $param$0 = $$17;
       } else {
        $param$0 = $165;
       }
       $170 = $162&255; //@line 131 "load_mtm.cpp"
       $171 = ((($m$027)) + 3|0); //@line 131 "load_mtm.cpp"
       HEAP8[$171>>0] = $170; //@line 131 "load_mtm.cpp"
       $172 = $param$0&255; //@line 132 "load_mtm.cpp"
       $173 = ((($m$027)) + 5|0); //@line 132 "load_mtm.cpp"
       HEAP8[$173>>0] = $172; //@line 132 "load_mtm.cpp"
       $174 = $param$0 | $162; //@line 133 "load_mtm.cpp"
       $175 = ($174|0)==(0); //@line 133 "load_mtm.cpp"
       if ($175) {
        $178 = $231;
       } else {
        __ZNK10CSoundFile17ConvertModCommandEP11_MODCOMMAND($this,$m$027); //@line 133 "load_mtm.cpp"
        $$pre44 = HEAP32[$43>>2]|0; //@line 121 "load_mtm.cpp"
        $178 = $$pre44;
       }
       $176 = (($i1$028) + 1)|0; //@line 121 "load_mtm.cpp"
       $177 = (($m$027) + (($178*6)|0)|0); //@line 121 "load_mtm.cpp"
       $179 = ((($p$026)) + 3|0); //@line 121 "load_mtm.cpp"
       $exitcond41 = ($176|0)==(64); //@line 121 "load_mtm.cpp"
       if ($exitcond41) {
        break;
       } else {
        $231 = $178;$i1$028 = $176;$m$027 = $177;$p$026 = $179;
       }
      }
     }
    }
   }
   $180 = (($n$029) + 1)|0; //@line 117 "load_mtm.cpp"
   $exitcond42 = ($180|0)==(32); //@line 117 "load_mtm.cpp"
   if ($exitcond42) {
    break;
   } else {
    $n$029 = $180;
   }
  }
  $181 = ((($pSeq$030)) + 64|0); //@line 136 "load_mtm.cpp"
  $182 = (($pat$031) + 1)|0; //@line 113 "load_mtm.cpp"
  $183 = HEAP8[$15>>0]|0; //@line 113 "load_mtm.cpp"
  $184 = $183&255; //@line 113 "load_mtm.cpp"
  $185 = ($182>>>0)>($184>>>0); //@line 113 "load_mtm.cpp"
  if ($185) {
   $187 = $183;
   break;
  } else {
   $pSeq$030 = $181;$pat$031 = $182;
  }
 }
 if ((label|0) == 23) {
  $$pre46 = HEAP8[$15>>0]|0; //@line 138 "load_mtm.cpp"
  $187 = $$pre46;
 }
 $186 = $187&255; //@line 138 "load_mtm.cpp"
 $188 = $186 << 6; //@line 138 "load_mtm.cpp"
 $189 = (($122) + 64)|0; //@line 138 "load_mtm.cpp"
 $190 = (($189) + ($188))|0; //@line 138 "load_mtm.cpp"
 $191 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8); //@line 139 "load_mtm.cpp"
 $192 = ($191<<16>>16)==(0); //@line 139 "load_mtm.cpp"
 if ($192) {
  $210 = 0;
 } else {
  $193 = $191&65535; //@line 139 "load_mtm.cpp"
  $194 = (($190) + ($193))|0; //@line 139 "load_mtm.cpp"
  $195 = ($194>>>0)<($dwMemLength>>>0); //@line 139 "load_mtm.cpp"
  if ($195) {
   $196 = (($193) + 1)|0; //@line 142 "load_mtm.cpp"
   $197 = (__Znaj($196)|0); //@line 142 "load_mtm.cpp"
   $198 = ((($this)) + 59796|0); //@line 142 "load_mtm.cpp"
   HEAP32[$198>>2] = $197; //@line 142 "load_mtm.cpp"
   $199 = (($lpStream) + ($190)|0); //@line 145 "load_mtm.cpp"
   _memcpy(($197|0),($199|0),($193|0))|0; //@line 145 "load_mtm.cpp"
   $200 = (($197) + ($193)|0); //@line 146 "load_mtm.cpp"
   HEAP8[$200>>0] = 0; //@line 146 "load_mtm.cpp"
   $202 = $197;$i3$023 = 0;
   while(1) {
    $201 = (($202) + ($i3$023)|0); //@line 149 "load_mtm.cpp"
    $203 = HEAP8[$201>>0]|0; //@line 149 "load_mtm.cpp"
    $204 = ($203<<24>>24)==(0); //@line 149 "load_mtm.cpp"
    $205 = (($i3$023) + 1)|0; //@line 151 "load_mtm.cpp"
    if ($204) {
     $206 = (($205>>>0) % 40)&-1; //@line 151 "load_mtm.cpp"
     $207 = ($206|0)!=(0); //@line 151 "load_mtm.cpp"
     $208 = $207 ? 32 : 13; //@line 151 "load_mtm.cpp"
     HEAP8[$201>>0] = $208; //@line 151 "load_mtm.cpp"
    }
    $exitcond = ($205|0)==($193|0); //@line 147 "load_mtm.cpp"
    if ($exitcond) {
     break;
    }
    $$pre49 = HEAP32[$198>>2]|0; //@line 149 "load_mtm.cpp"
    $202 = $$pre49;$i3$023 = $205;
   }
   $$pre50$pre = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8); //@line 156 "load_mtm.cpp"
   $210 = $$pre50$pre;
  } else {
   $210 = $191;
  }
 }
 $209 = $210&65535; //@line 156 "load_mtm.cpp"
 $211 = (($209) + ($190))|0; //@line 156 "load_mtm.cpp"
 $212 = HEAP32[$40>>2]|0; //@line 158 "load_mtm.cpp"
 $213 = ($212|0)!=(0); //@line 158 "load_mtm.cpp"
 $214 = ($211>>>0)<($dwMemLength>>>0); //@line 160 "load_mtm.cpp"
 $or$cond1620 = $214 & $213; //@line 158 "load_mtm.cpp"
 if ($or$cond1620) {
  $dwMemPos$121 = $211;$ismp$022 = 1;
  while(1) {
   $215 = (((($this)) + 33280|0) + ($ismp$022<<6)|0); //@line 161 "load_mtm.cpp"
   $216 = (((((($this)) + 33280|0) + ($ismp$022<<6)|0)) + 34|0); //@line 161 "load_mtm.cpp"
   $217 = HEAP16[$216>>1]|0; //@line 161 "load_mtm.cpp"
   $218 = $217 & 1; //@line 161 "load_mtm.cpp"
   $219 = ($218<<16>>16)!=(0); //@line 161 "load_mtm.cpp"
   $220 = $219 ? 6 : 1; //@line 161 "load_mtm.cpp"
   $221 = (($lpStream) + ($dwMemPos$121)|0); //@line 162 "load_mtm.cpp"
   $222 = (($dwMemLength) - ($dwMemPos$121))|0; //@line 162 "load_mtm.cpp"
   $223 = (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$215,$220,$221,$222)|0); //@line 161 "load_mtm.cpp"
   $224 = (($223) + ($dwMemPos$121))|0; //@line 161 "load_mtm.cpp"
   $225 = (($ismp$022) + 1)|0; //@line 158 "load_mtm.cpp"
   $226 = HEAP32[$40>>2]|0; //@line 158 "load_mtm.cpp"
   $227 = ($225>>>0)<=($226>>>0); //@line 158 "load_mtm.cpp"
   $228 = ($224>>>0)<($dwMemLength>>>0); //@line 160 "load_mtm.cpp"
   $or$cond16 = $228 & $227; //@line 158 "load_mtm.cpp"
   if ($or$cond16) {
    $dwMemPos$121 = $224;$ismp$022 = $225;
   } else {
    break;
   }
  }
 }
 $229 = ((($this)) + 59764|0); //@line 164 "load_mtm.cpp"
 HEAP32[$229>>2] = 64; //@line 164 "load_mtm.cpp"
 $230 = ((($this)) + 59768|0); //@line 165 "load_mtm.cpp"
 HEAP32[$230>>2] = 32767; //@line 165 "load_mtm.cpp"
 $$0 = 1;
 return ($$0|0); //@line 167 "load_mtm.cpp"
}
function __ZN10CSoundFile7ReadMDLEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$0 = 0, $$2 = 0, $$55 = 0, $$lcssa = 0, $$lcssa258 = 0, $$lcssa262 = 0, $$off = 0, $$off25 = 0, $$op = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phi186Z2D = 0, $$pre$phi190Z2D = 0, $$pre$phi196Z2D = 0, $$pre176 = 0, $$pre177 = 0, $$pre179 = 0, $$pre180 = 0, $$pre181 = 0, $$pre182 = 0;
 var $$pre183 = 0, $$pre184 = 0, $$pre185 = 0, $$pre189 = 0, $$pre191 = 0, $$pre193 = 0, $$pre195 = 0, $$pre197 = 0, $$pre199 = 0, $$sum = 0, $$sum17 = 0, $$sum18 = 0, $$sum19 = 0, $$sum20 = 0, $$sum21 = 0, $$sum22 = 0, $$sum23 = 0, $$sum24 = 0, $$sum26 = 0, $$sum27 = 0;
 var $$sum28 = 0, $$sum29 = 0, $$sum30 = 0, $$sum31 = 0, $$sum32 = 0, $$sum33 = 0, $$sum34 = 0, $$sum35 = 0, $$sum36 = 0, $$sum37 = 0, $$sum38 = 0, $$sum39 = 0, $$sum40 = 0, $$sum41 = 0, $$sum42 = 0, $$sum44 = 0, $$sum45 = 0, $$sum46 = 0, $$sum47 = 0, $$sum48 = 0;
 var $$sum49 = 0, $$sum50 = 0, $$sum51 = 0, $$sum52 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0;
 var $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0;
 var $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0;
 var $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0;
 var $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0;
 var $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0;
 var $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0;
 var $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0;
 var $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0;
 var $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0;
 var $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0;
 var $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0;
 var $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0;
 var $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0;
 var $509 = 0, $51 = 0, $510 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $chn$098 = 0, $dwMemPos$0147 = 0, $dwPos$0133 = 0, $dwPos$1121 = 0, $dwPos$2105 = 0;
 var $dwPos$3110 = 0, $dwPos$4 = 0, $dwTrackPos$0$be = 0, $dwTrackPos$0146 = 0, $dwTrackPos$0146$lcssa = 0, $dwTrackPos$088 = 0, $exitcond = 0, $exitcond168 = 0, $exitcond169 = 0, $exitcond172 = 0, $exitcond174 = 0, $i$0137 = 0, $i$1132 = 0, $i$2120 = 0, $i$3104 = 0, $i$4109 = 0, $iIns$096 = 0, $inspanenv = 0, $insvolenv = 0, $ipat$0101 = 0;
 var $iv$0 = 0, $iv5$0 = 0, $j$0138 = 0, $j$2117 = 0, $j$3125 = 0, $note$0116 = 0, $note$1$lcssa = 0, $note$1112 = 0, $npanenv$0$be = 0, $npanenv$0143 = 0, $npanenv$0143$lcssa = 0, $npanenv$082 = 0, $npanenv$082203 = 0, $npatterns$0$be = 0, $npatterns$0145 = 0, $npatterns$0145$lcssa = 0, $npatterns$086 = 0, $npe$094 = 0, $ntracks$0$be = 0, $ntracks$0144 = 0;
 var $ntracks$0144$lcssa = 0, $ntracks$084 = 0, $nve$092 = 0, $nvolenv$0$be = 0, $nvolenv$0142 = 0, $nvolenv$0142$lcssa = 0, $nvolenv$080 = 0, $nvolenv$080204 = 0, $or$cond = 0, $or$cond10 = 0, $or$cond12 = 0, $or$cond14 = 0, $or$cond16 = 0, $or$cond54 = 0, $or$cond54131 = 0, $or$cond56 = 0, $or$cond8 = 0, $patterntracks = 0, $ppanenv$0$be = 0, $ppanenv$0141 = 0;
 var $ppanenv$0141$lcssa = 0, $ppanenv$078 = 0, $ppanenv$078205 = 0, $ppe$093 = 0, $pve$091 = 0, $pvolenv$0$be = 0, $pvolenv$0140 = 0, $pvolenv$0140$lcssa = 0, $pvolenv$076 = 0, $pvolenv$076206 = 0, $scevgep = 0, $scevgep171 = 0, $smpinfo = 0, $vtick$0 = 0, $vtick$1 = 0, $vtick4$0 = 0, $vtick4$1 = 0, dest = 0, label = 0, sp = 0;
 var src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16080|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $patterntracks = sp;
 $smpinfo = sp + 15840|0;
 $insvolenv = sp + 15600|0;
 $inspanenv = sp + 15360|0;
 $0 = ($lpStream|0)==(0|0); //@line 196 "load_mdl.cpp"
 $1 = ($dwMemLength>>>0)<(1024); //@line 196 "load_mdl.cpp"
 $or$cond = $0 | $1; //@line 196 "load_mdl.cpp"
 if ($or$cond) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 482 "load_mdl.cpp"
 }
 $2 = HEAP32[$lpStream>>2]|0; //@line 197 "load_mdl.cpp"
 $3 = ($2|0)==(1279544644); //@line 197 "load_mdl.cpp"
 if (!($3)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 482 "load_mdl.cpp"
 }
 $4 = ((($lpStream)) + 4|0); //@line 197 "load_mdl.cpp"
 $5 = HEAP8[$4>>0]|0; //@line 197 "load_mdl.cpp"
 $6 = $5&255; //@line 197 "load_mdl.cpp"
 $7 = $6 & 240; //@line 197 "load_mdl.cpp"
 $8 = ($7>>>0)>(16); //@line 197 "load_mdl.cpp"
 if ($8) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 482 "load_mdl.cpp"
 }
 _memset(($patterntracks|0),0,15360)|0; //@line 198 "load_mdl.cpp"
 _memset(($smpinfo|0),0,240)|0; //@line 199 "load_mdl.cpp"
 _memset(($insvolenv|0),0,240)|0; //@line 200 "load_mdl.cpp"
 _memset(($inspanenv|0),0,240)|0; //@line 201 "load_mdl.cpp"
 $9 = ((($this)) + 59688|0); //@line 206 "load_mdl.cpp"
 HEAP32[$9>>2] = 0; //@line 206 "load_mdl.cpp"
 $10 = ((($this)) + 59684|0); //@line 206 "load_mdl.cpp"
 HEAP32[$10>>2] = 0; //@line 206 "load_mdl.cpp"
 $11 = ($dwMemLength>>>0)>(11); //@line 207 "load_mdl.cpp"
 L10: do {
  if ($11) {
   $12 = ((($this)) + 59804|0); //@line 222 "load_mdl.cpp"
   $13 = ((($this)) + 59736|0); //@line 225 "load_mdl.cpp"
   $14 = ((($this)) + 59656|0); //@line 226 "load_mdl.cpp"
   $15 = ((($this)) + 59652|0); //@line 227 "load_mdl.cpp"
   $16 = ((($this)) + 59648|0); //@line 228 "load_mdl.cpp"
   $17 = ((($this)) + 59664|0); //@line 229 "load_mdl.cpp"
   $18 = ((($this)) + 59796|0); //@line 245 "load_mdl.cpp"
   $22 = 11;$dwMemPos$0147 = 5;$dwTrackPos$0146 = 0;$npanenv$0143 = 0;$npatterns$0145 = 0;$ntracks$0144 = 0;$nvolenv$0142 = 0;$ppanenv$0141 = 0;$pvolenv$0140 = 0;
   while(1) {
    $$sum = (($dwMemPos$0147) + 2)|0; //@line 210 "load_mdl.cpp"
    $19 = (($lpStream) + ($$sum)|0); //@line 210 "load_mdl.cpp"
    $20 = HEAP32[$19>>2]|0; //@line 210 "load_mdl.cpp"
    $21 = (($20) + ($22))|0; //@line 212 "load_mdl.cpp"
    $23 = ($21>>>0)>($dwMemLength>>>0); //@line 212 "load_mdl.cpp"
    if ($23) {
     $$lcssa262 = $22;$dwTrackPos$0146$lcssa = $dwTrackPos$0146;$npanenv$0143$lcssa = $npanenv$0143;$npatterns$0145$lcssa = $npatterns$0145;$ntracks$0144$lcssa = $ntracks$0144;$nvolenv$0142$lcssa = $nvolenv$0142;$ppanenv$0141$lcssa = $ppanenv$0141;$pvolenv$0140$lcssa = $pvolenv$0140;
     label = 7;
     break;
    }
    $25 = (($lpStream) + ($dwMemPos$0147)|0); //@line 209 "load_mdl.cpp"
    $26 = HEAP16[$25>>1]|0; //@line 209 "load_mdl.cpp"
    $27 = $26&65535; //@line 217 "load_mdl.cpp"
    L15: do {
     switch ($27|0) {
     case 20041:  {
      $30 = (($lpStream) + ($22)|0); //@line 221 "load_mdl.cpp"
      dest=$12; src=$30; stop=dest+32|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 222 "load_mdl.cpp"
      $$sum44 = (($dwMemPos$0147) + 58)|0; //@line 223 "load_mdl.cpp"
      $31 = (($lpStream) + ($$sum44)|0); //@line 223 "load_mdl.cpp"
      $32 = HEAP16[$31>>1]|0; //@line 223 "load_mdl.cpp"
      $33 = $32&65535; //@line 223 "load_mdl.cpp"
      $$sum45 = (($dwMemPos$0147) + 60)|0; //@line 225 "load_mdl.cpp"
      $34 = (($lpStream) + ($$sum45)|0); //@line 225 "load_mdl.cpp"
      $35 = HEAP16[$34>>1]|0; //@line 225 "load_mdl.cpp"
      $36 = $35&65535; //@line 225 "load_mdl.cpp"
      HEAP32[$13>>2] = $36; //@line 225 "load_mdl.cpp"
      $$sum46 = (($dwMemPos$0147) + 62)|0; //@line 226 "load_mdl.cpp"
      $37 = (($lpStream) + ($$sum46)|0); //@line 226 "load_mdl.cpp"
      $38 = HEAP8[$37>>0]|0; //@line 226 "load_mdl.cpp"
      $39 = $38&255; //@line 226 "load_mdl.cpp"
      HEAP32[$14>>2] = $39; //@line 226 "load_mdl.cpp"
      $$sum47 = (($dwMemPos$0147) + 64)|0; //@line 227 "load_mdl.cpp"
      $40 = (($lpStream) + ($$sum47)|0); //@line 227 "load_mdl.cpp"
      $41 = HEAP8[$40>>0]|0; //@line 227 "load_mdl.cpp"
      $42 = $41&255; //@line 227 "load_mdl.cpp"
      HEAP32[$15>>2] = $42; //@line 227 "load_mdl.cpp"
      $$sum48 = (($dwMemPos$0147) + 63)|0; //@line 228 "load_mdl.cpp"
      $43 = (($lpStream) + ($$sum48)|0); //@line 228 "load_mdl.cpp"
      $44 = HEAP8[$43>>0]|0; //@line 228 "load_mdl.cpp"
      $45 = $44&255; //@line 228 "load_mdl.cpp"
      HEAP32[$16>>2] = $45; //@line 228 "load_mdl.cpp"
      HEAP32[$17>>2] = 4; //@line 229 "load_mdl.cpp"
      $$sum51 = (($dwMemPos$0147) + 65)|0; //@line 233 "load_mdl.cpp"
      $i$0137 = 0;
      while(1) {
       $50 = (((((($this)) + 49600|0) + (($i$0137*36)|0)|0)) + 4|0); //@line 232 "load_mdl.cpp"
       HEAP32[$50>>2] = 64; //@line 232 "load_mdl.cpp"
       $$sum52 = (($$sum51) + ($i$0137))|0; //@line 233 "load_mdl.cpp"
       $51 = (($lpStream) + ($$sum52)|0); //@line 233 "load_mdl.cpp"
       $52 = HEAP8[$51>>0]|0; //@line 233 "load_mdl.cpp"
       $53 = $52&255; //@line 233 "load_mdl.cpp"
       $54 = $53 << 1; //@line 233 "load_mdl.cpp"
       $55 = $54 & 254; //@line 233 "load_mdl.cpp"
       $56 = (((($this)) + 49600|0) + (($i$0137*36)|0)|0); //@line 233 "load_mdl.cpp"
       HEAP32[$56>>2] = $55; //@line 233 "load_mdl.cpp"
       $57 = HEAP8[$51>>0]|0; //@line 234 "load_mdl.cpp"
       $58 = ($57<<24>>24)<(0); //@line 234 "load_mdl.cpp"
       if ($58) {
        $59 = (((((($this)) + 49600|0) + (($i$0137*36)|0)|0)) + 8|0); //@line 235 "load_mdl.cpp"
        $60 = HEAP32[$59>>2]|0; //@line 235 "load_mdl.cpp"
        $61 = $60 | 256; //@line 235 "load_mdl.cpp"
        HEAP32[$59>>2] = $61; //@line 235 "load_mdl.cpp"
        $$pre185 = (($i$0137) + 1)|0; //@line 230 "load_mdl.cpp"
        $$pre$phi186Z2D = $$pre185;
       } else {
        $62 = (($i$0137) + 1)|0; //@line 237 "load_mdl.cpp"
        HEAP32[$17>>2] = $62; //@line 237 "load_mdl.cpp"
        $$pre$phi186Z2D = $62;
       }
       $exitcond172 = ($$pre$phi186Z2D|0)==(32); //@line 230 "load_mdl.cpp"
       if ($exitcond172) {
        break;
       } else {
        $i$0137 = $$pre$phi186Z2D;
       }
      }
      $46 = ($32<<16>>16)==(0); //@line 239 "load_mdl.cpp"
      if ($46) {
       $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
      } else {
       $$sum49 = (($dwMemPos$0147) + 97)|0; //@line 239 "load_mdl.cpp"
       $47 = $33 ^ -256; //@line 239 "load_mdl.cpp"
       $48 = ($47>>>0)>(4294967039);
       $49 = $48 ? $33 : 256; //@line 239 "load_mdl.cpp"
       $j$0138 = 0;
       while(1) {
        $$sum50 = (($$sum49) + ($j$0138))|0; //@line 239 "load_mdl.cpp"
        $63 = (($lpStream) + ($$sum50)|0); //@line 239 "load_mdl.cpp"
        $64 = HEAP8[$63>>0]|0; //@line 239 "load_mdl.cpp"
        $65 = (((($this)) + 53344|0) + ($j$0138)|0); //@line 239 "load_mdl.cpp"
        HEAP8[$65>>0] = $64; //@line 239 "load_mdl.cpp"
        $66 = (($j$0138) + 1)|0; //@line 239 "load_mdl.cpp"
        $exitcond174 = ($66|0)==($49|0); //@line 239 "load_mdl.cpp"
        if ($exitcond174) {
         $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
         break;
        } else {
         $j$0138 = $66;
        }
       }
      }
      break;
     }
     case 17741:  {
      $67 = ($20|0)==(0); //@line 243 "load_mdl.cpp"
      if ($67) {
       $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
      } else {
       $70 = HEAP32[$18>>2]|0; //@line 245 "load_mdl.cpp"
       $71 = ($70|0)==(0|0); //@line 245 "load_mdl.cpp"
       if (!($71)) {
        __ZdaPv($70); //@line 245 "load_mdl.cpp"
       }
       $72 = (__Znaj($20)|0); //@line 246 "load_mdl.cpp"
       HEAP32[$18>>2] = $72; //@line 246 "load_mdl.cpp"
       $73 = (($lpStream) + ($22)|0); //@line 249 "load_mdl.cpp"
       _memcpy(($72|0),($73|0),($20|0))|0; //@line 249 "load_mdl.cpp"
       $74 = (($20) + -1)|0; //@line 250 "load_mdl.cpp"
       $75 = (($72) + ($74)|0); //@line 250 "load_mdl.cpp"
       HEAP8[$75>>0] = 0; //@line 250 "load_mdl.cpp"
       $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
      }
      break;
     }
     case 16720:  {
      $76 = (($lpStream) + ($22)|0); //@line 256 "load_mdl.cpp"
      $77 = HEAP8[$76>>0]|0; //@line 256 "load_mdl.cpp"
      $78 = $77&255; //@line 256 "load_mdl.cpp"
      $79 = ($78>>>0)>(240); //@line 257 "load_mdl.cpp"
      $$2 = $79 ? 240 : $78; //@line 257 "load_mdl.cpp"
      $80 = ($$2|0)!=(0); //@line 259 "load_mdl.cpp"
      $81 = (($dwMemPos$0147) + 25)|0;
      $82 = ($81>>>0)<($dwMemLength>>>0); //@line 261 "load_mdl.cpp"
      $or$cond54131 = $80 & $82; //@line 259 "load_mdl.cpp"
      if ($or$cond54131) {
       $83 = (($dwMemPos$0147) + 7)|0; //@line 258 "load_mdl.cpp"
       $101 = $81;$dwPos$0133 = $83;$i$1132 = 0;
       while(1) {
        $84 = $i$1132 << 5; //@line 262 "load_mdl.cpp"
        $scevgep = (($patterntracks) + ($84<<1)|0);
        $85 = (($lpStream) + ($dwPos$0133)|0); //@line 262 "load_mdl.cpp"
        $86 = HEAP8[$85>>0]|0; //@line 263 "load_mdl.cpp"
        $87 = ($86&255)>(32); //@line 263 "load_mdl.cpp"
        if ($87) {
         $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $$2;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
         break L15;
        }
        $$sum42 = (($dwPos$0133) + 1)|0; //@line 264 "load_mdl.cpp"
        $88 = (($lpStream) + ($$sum42)|0); //@line 264 "load_mdl.cpp"
        $89 = HEAP8[$88>>0]|0; //@line 264 "load_mdl.cpp"
        $90 = $89&255; //@line 264 "load_mdl.cpp"
        $91 = (($90) + 1)|0; //@line 264 "load_mdl.cpp"
        $92 = $91&65535; //@line 264 "load_mdl.cpp"
        $93 = (((($this)) + 52864|0) + ($i$1132<<1)|0); //@line 264 "load_mdl.cpp"
        HEAP16[$93>>1] = $92; //@line 264 "load_mdl.cpp"
        $94 = HEAP32[$17>>2]|0; //@line 265 "load_mdl.cpp"
        $95 = $86&255; //@line 265 "load_mdl.cpp"
        $96 = ($94>>>0)<($95>>>0); //@line 265 "load_mdl.cpp"
        if ($96) {
         HEAP32[$17>>2] = $95; //@line 265 "load_mdl.cpp"
        }
        $97 = $95 << 1; //@line 266 "load_mdl.cpp"
        $98 = (($dwPos$0133) + 18)|0; //@line 266 "load_mdl.cpp"
        $99 = (($98) + ($97))|0; //@line 266 "load_mdl.cpp"
        $100 = ($86<<24>>24)==(0); //@line 267 "load_mdl.cpp"
        if (!($100)) {
         $scevgep171 = (($lpStream) + ($101)|0);
         $102 = ($86&255)>(1);
         $$op = $95 << 1; //@line 267 "load_mdl.cpp"
         $103 = $102 ? $$op : 2; //@line 267 "load_mdl.cpp"
         _memcpy(($scevgep|0),($scevgep171|0),($103|0))|0; //@line 269 "load_mdl.cpp"
        }
        $104 = (($i$1132) + 1)|0; //@line 259 "load_mdl.cpp"
        $105 = ($104>>>0)<($$2>>>0); //@line 259 "load_mdl.cpp"
        $106 = (($99) + 18)|0;
        $107 = ($106>>>0)<($dwMemLength>>>0); //@line 261 "load_mdl.cpp"
        $or$cond54 = $105 & $107; //@line 259 "load_mdl.cpp"
        if ($or$cond54) {
         $101 = $106;$dwPos$0133 = $99;$i$1132 = $104;
        } else {
         $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $$2;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
         break;
        }
       }
      } else {
       $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $$2;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
      }
      break;
     }
     case 21076:  {
      $108 = ($dwTrackPos$0146|0)==(0); //@line 275 "load_mdl.cpp"
      if ($108) {
       $109 = (($lpStream) + ($22)|0); //@line 276 "load_mdl.cpp"
       $110 = HEAP16[$109>>1]|0; //@line 276 "load_mdl.cpp"
       $111 = $110&65535; //@line 276 "load_mdl.cpp"
       $112 = (($dwMemPos$0147) + 8)|0; //@line 277 "load_mdl.cpp"
       $dwTrackPos$0$be = $112;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $111;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
      } else {
       $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
      }
      break;
     }
     case 18761:  {
      $113 = (($lpStream) + ($22)|0); //@line 281 "load_mdl.cpp"
      $114 = HEAP8[$113>>0]|0; //@line 281 "load_mdl.cpp"
      $115 = $114&255; //@line 281 "load_mdl.cpp"
      $116 = ($114<<24>>24)==(0); //@line 283 "load_mdl.cpp"
      L46: do {
       if (!($116)) {
        $117 = (($dwMemPos$0147) + 7)|0; //@line 282 "load_mdl.cpp"
        $dwPos$1121 = $117;$i$2120 = 0;
        while(1) {
         $120 = (($lpStream) + ($dwPos$1121)|0); //@line 285 "load_mdl.cpp"
         $121 = HEAP8[$120>>0]|0; //@line 285 "load_mdl.cpp"
         $122 = $121&255; //@line 285 "load_mdl.cpp"
         $$off25 = (($121) + -1)<<24>>24; //@line 286 "load_mdl.cpp"
         $123 = ($$off25&255)<(239); //@line 286 "load_mdl.cpp"
         if (!($123)) {
          break L46;
         }
         $124 = HEAP32[$9>>2]|0; //@line 287 "load_mdl.cpp"
         $125 = ($124>>>0)<($122>>>0); //@line 287 "load_mdl.cpp"
         if ($125) {
          HEAP32[$9>>2] = $122; //@line 287 "load_mdl.cpp"
         }
         $126 = (((($this)) + 48640|0) + ($122<<2)|0); //@line 288 "load_mdl.cpp"
         $127 = HEAP32[$126>>2]|0; //@line 288 "load_mdl.cpp"
         $128 = ($127|0)==(0|0); //@line 288 "load_mdl.cpp"
         if ($128) {
          $129 = (__Znwj(632)|0); //@line 291 "load_mdl.cpp"
          HEAP32[$126>>2] = $129; //@line 291 "load_mdl.cpp"
          _memset(($129|0),0,632)|0; //@line 293 "load_mdl.cpp"
          $130 = ((($129)) + 585|0); //@line 294 "load_mdl.cpp"
          $$sum26 = (($dwPos$1121) + 2)|0; //@line 294 "load_mdl.cpp"
          $131 = (($lpStream) + ($$sum26)|0); //@line 294 "load_mdl.cpp"
          dest=$130; src=$131; stop=dest+32|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 294 "load_mdl.cpp"
          $132 = ((($129)) + 8|0); //@line 295 "load_mdl.cpp"
          HEAP16[$132>>1] = 64; //@line 295 "load_mdl.cpp"
          $133 = ((($129)) + 584|0); //@line 296 "load_mdl.cpp"
          HEAP8[$133>>0] = 60; //@line 296 "load_mdl.cpp"
          $134 = (($dwPos$1121) + 1)|0; //@line 297 "load_mdl.cpp"
          $135 = (($lpStream) + ($134)|0); //@line 297 "load_mdl.cpp"
          $136 = HEAP8[$135>>0]|0; //@line 297 "load_mdl.cpp"
          $137 = ($136<<24>>24)==(0); //@line 297 "load_mdl.cpp"
          if ($137) {
           $217 = 0;
          } else {
           $$sum27 = (($dwPos$1121) + 34)|0; //@line 299 "load_mdl.cpp"
           $138 = ((($129)) + 4|0); //@line 321 "load_mdl.cpp"
           $139 = (($insvolenv) + ($122)|0); //@line 322 "load_mdl.cpp"
           $140 = (($inspanenv) + ($122)|0); //@line 328 "load_mdl.cpp"
           $j$2117 = 0;$note$0116 = 12;
           while(1) {
            $141 = ($j$2117*14)|0; //@line 299 "load_mdl.cpp"
            $$sum28 = (($$sum27) + ($141))|0; //@line 299 "load_mdl.cpp"
            $142 = (($lpStream) + ($$sum28)|0); //@line 299 "load_mdl.cpp"
            $$sum29 = (($$sum28) + 1)|0; //@line 300 "load_mdl.cpp"
            $143 = (($lpStream) + ($$sum29)|0); //@line 300 "load_mdl.cpp"
            $144 = HEAP8[$143>>0]|0; //@line 300 "load_mdl.cpp"
            $145 = $144&255; //@line 300 "load_mdl.cpp"
            $146 = (($145) + 12)|0; //@line 300 "load_mdl.cpp"
            $147 = ($note$0116>>>0)<($146>>>0); //@line 300 "load_mdl.cpp"
            $148 = ($note$0116>>>0)<(120); //@line 300 "load_mdl.cpp"
            $149 = $148 & $147; //@line 300 "load_mdl.cpp"
            if ($149) {
             $$sum36 = (($$sum28) + 2)|0; //@line 307 "load_mdl.cpp"
             $150 = (($lpStream) + ($$sum36)|0); //@line 307 "load_mdl.cpp"
             $$sum37 = (($$sum28) + 4)|0; //@line 308 "load_mdl.cpp"
             $151 = (($lpStream) + ($$sum37)|0); //@line 308 "load_mdl.cpp"
             $$sum38 = (($$sum28) + 11)|0; //@line 309 "load_mdl.cpp"
             $152 = (($lpStream) + ($$sum38)|0); //@line 309 "load_mdl.cpp"
             $$sum39 = (($$sum28) + 10)|0; //@line 310 "load_mdl.cpp"
             $153 = (($lpStream) + ($$sum39)|0); //@line 310 "load_mdl.cpp"
             $$sum40 = (($$sum28) + 9)|0; //@line 311 "load_mdl.cpp"
             $154 = (($lpStream) + ($$sum40)|0); //@line 311 "load_mdl.cpp"
             $$sum41 = (($$sum28) + 8)|0; //@line 312 "load_mdl.cpp"
             $155 = (($lpStream) + ($$sum41)|0); //@line 312 "load_mdl.cpp"
             $$sum33 = (($$sum28) + 7)|0; //@line 314 "load_mdl.cpp"
             $156 = (($lpStream) + ($$sum33)|0); //@line 314 "load_mdl.cpp"
             $$sum34 = (($$sum28) + 6)|0; //@line 314 "load_mdl.cpp"
             $157 = (($lpStream) + ($$sum34)|0); //@line 314 "load_mdl.cpp"
             $509 = $144;$note$1112 = $note$0116;
             while(1) {
              $158 = (($note$1112) + 1)|0; //@line 302 "load_mdl.cpp"
              $159 = $158&255; //@line 302 "load_mdl.cpp"
              $$sum32 = (($note$1112) + 428)|0; //@line 302 "load_mdl.cpp"
              $160 = (($129) + ($$sum32)|0); //@line 302 "load_mdl.cpp"
              HEAP8[$160>>0] = $159; //@line 302 "load_mdl.cpp"
              $161 = HEAP8[$142>>0]|0; //@line 303 "load_mdl.cpp"
              $162 = ($161&255)<(240); //@line 303 "load_mdl.cpp"
              if ($162) {
               $163 = $161&255; //@line 303 "load_mdl.cpp"
               $$sum35 = (($note$1112) + 300)|0; //@line 306 "load_mdl.cpp"
               $164 = (($129) + ($$sum35)|0); //@line 306 "load_mdl.cpp"
               HEAP8[$164>>0] = $161; //@line 306 "load_mdl.cpp"
               $165 = HEAP8[$150>>0]|0; //@line 307 "load_mdl.cpp"
               $166 = $165&255; //@line 307 "load_mdl.cpp"
               $167 = (((((($this)) + 33280|0) + ($163<<6)|0)) + 30|0); //@line 307 "load_mdl.cpp"
               HEAP16[$167>>1] = $166; //@line 307 "load_mdl.cpp"
               $168 = HEAP8[$151>>0]|0; //@line 308 "load_mdl.cpp"
               $169 = $168&255; //@line 308 "load_mdl.cpp"
               $170 = $169 << 1; //@line 308 "load_mdl.cpp"
               $171 = $170&65535; //@line 308 "load_mdl.cpp"
               $172 = (((((($this)) + 33280|0) + ($163<<6)|0)) + 28|0); //@line 308 "load_mdl.cpp"
               HEAP16[$172>>1] = $171; //@line 308 "load_mdl.cpp"
               $173 = HEAP8[$152>>0]|0; //@line 309 "load_mdl.cpp"
               $174 = (((((($this)) + 33280|0) + ($163<<6)|0)) + 38|0); //@line 309 "load_mdl.cpp"
               HEAP8[$174>>0] = $173; //@line 309 "load_mdl.cpp"
               $175 = HEAP8[$153>>0]|0; //@line 310 "load_mdl.cpp"
               $176 = (((((($this)) + 33280|0) + ($163<<6)|0)) + 39|0); //@line 310 "load_mdl.cpp"
               HEAP8[$176>>0] = $175; //@line 310 "load_mdl.cpp"
               $177 = HEAP8[$154>>0]|0; //@line 311 "load_mdl.cpp"
               $178 = (((((($this)) + 33280|0) + ($163<<6)|0)) + 40|0); //@line 311 "load_mdl.cpp"
               HEAP8[$178>>0] = $177; //@line 311 "load_mdl.cpp"
               $179 = HEAP8[$155>>0]|0; //@line 312 "load_mdl.cpp"
               $180 = (((((($this)) + 33280|0) + ($163<<6)|0)) + 41|0); //@line 312 "load_mdl.cpp"
               HEAP8[$180>>0] = $179; //@line 312 "load_mdl.cpp"
               $$pre177 = HEAP8[$143>>0]|0; //@line 300 "load_mdl.cpp"
               $189 = $$pre177;
              } else {
               $189 = $509;
              }
              $181 = HEAP8[$156>>0]|0; //@line 314 "load_mdl.cpp"
              $182 = $181&255; //@line 314 "load_mdl.cpp"
              $183 = $182 << 8; //@line 314 "load_mdl.cpp"
              $184 = HEAP8[$157>>0]|0; //@line 314 "load_mdl.cpp"
              $185 = $184&255; //@line 314 "load_mdl.cpp"
              $186 = $183 | $185; //@line 314 "load_mdl.cpp"
              $187 = ($186|0)==(65535); //@line 315 "load_mdl.cpp"
              $$55 = $187 ? 0 : $186; //@line 315 "load_mdl.cpp"
              HEAP32[$129>>2] = $$55; //@line 314 "load_mdl.cpp"
              $188 = $189&255; //@line 300 "load_mdl.cpp"
              $190 = (($188) + 12)|0; //@line 300 "load_mdl.cpp"
              $191 = ($158>>>0)<($190>>>0); //@line 300 "load_mdl.cpp"
              $192 = ($158>>>0)<(120); //@line 300 "load_mdl.cpp"
              $193 = $192 & $191; //@line 300 "load_mdl.cpp"
              if ($193) {
               $509 = $189;$note$1112 = $158;
              } else {
               $note$1$lcssa = $158;
               break;
              }
             }
            } else {
             $note$1$lcssa = $note$0116;
            }
            $$sum30 = (($$sum28) + 3)|0; //@line 319 "load_mdl.cpp"
            $194 = (($lpStream) + ($$sum30)|0); //@line 319 "load_mdl.cpp"
            $195 = HEAP8[$194>>0]|0; //@line 319 "load_mdl.cpp"
            $196 = ($195<<24>>24)<(0); //@line 319 "load_mdl.cpp"
            if ($196) {
             $197 = HEAP32[$138>>2]|0; //@line 321 "load_mdl.cpp"
             $198 = $197 | 1; //@line 321 "load_mdl.cpp"
             HEAP32[$138>>2] = $198; //@line 321 "load_mdl.cpp"
             $199 = $195&255; //@line 322 "load_mdl.cpp"
             $200 = $199 & 63; //@line 322 "load_mdl.cpp"
             $201 = (($200) + 1)|0; //@line 322 "load_mdl.cpp"
             $202 = $201&255; //@line 322 "load_mdl.cpp"
             HEAP8[$139>>0] = $202; //@line 322 "load_mdl.cpp"
            }
            $$sum31 = (($$sum28) + 5)|0; //@line 325 "load_mdl.cpp"
            $203 = (($lpStream) + ($$sum31)|0); //@line 325 "load_mdl.cpp"
            $204 = HEAP8[$203>>0]|0; //@line 325 "load_mdl.cpp"
            $205 = ($204<<24>>24)<(0); //@line 325 "load_mdl.cpp"
            if ($205) {
             $206 = HEAP32[$138>>2]|0; //@line 327 "load_mdl.cpp"
             $207 = $206 | 8; //@line 327 "load_mdl.cpp"
             HEAP32[$138>>2] = $207; //@line 327 "load_mdl.cpp"
             $208 = $204&255; //@line 328 "load_mdl.cpp"
             $209 = $208 & 63; //@line 328 "load_mdl.cpp"
             $210 = (($209) + 1)|0; //@line 328 "load_mdl.cpp"
             $211 = $210&255; //@line 328 "load_mdl.cpp"
             HEAP8[$140>>0] = $211; //@line 328 "load_mdl.cpp"
            }
            $212 = (($j$2117) + 1)|0; //@line 297 "load_mdl.cpp"
            $213 = HEAP8[$135>>0]|0; //@line 297 "load_mdl.cpp"
            $214 = $213&255; //@line 297 "load_mdl.cpp"
            $215 = ($212>>>0)<($214>>>0); //@line 297 "load_mdl.cpp"
            if ($215) {
             $j$2117 = $212;$note$0116 = $note$1$lcssa;
            } else {
             $217 = $213;
             break;
            }
           }
          }
         } else {
          $$pre183 = (($dwPos$1121) + 1)|0; //@line 332 "load_mdl.cpp"
          $$phi$trans$insert = (($lpStream) + ($$pre183)|0);
          $$pre184 = HEAP8[$$phi$trans$insert>>0]|0; //@line 332 "load_mdl.cpp"
          $217 = $$pre184;
         }
         $216 = $217&255; //@line 332 "load_mdl.cpp"
         $218 = ($216*14)|0; //@line 332 "load_mdl.cpp"
         $219 = (($dwPos$1121) + 34)|0; //@line 332 "load_mdl.cpp"
         $220 = (($219) + ($218))|0; //@line 332 "load_mdl.cpp"
         $221 = (($i$2120) + 1)|0; //@line 283 "load_mdl.cpp"
         $222 = ($221>>>0)<($115>>>0); //@line 283 "load_mdl.cpp"
         if ($222) {
          $dwPos$1121 = $220;$i$2120 = $221;
         } else {
          break;
         }
        }
       }
      } while(0);
      $118 = HEAP32[$9>>2]|0; //@line 334 "load_mdl.cpp"
      $119 = ($118|0)==(0); //@line 334 "load_mdl.cpp"
      if ($119) {
       $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
      } else {
       $508 = $118;$j$3125 = 1;
       while(1) {
        $223 = (((($this)) + 48640|0) + ($j$3125<<2)|0); //@line 334 "load_mdl.cpp"
        $224 = HEAP32[$223>>2]|0; //@line 334 "load_mdl.cpp"
        $225 = ($224|0)==(0|0); //@line 334 "load_mdl.cpp"
        if ($225) {
         $226 = (__Znwj(632)|0); //@line 336 "load_mdl.cpp"
         HEAP32[$223>>2] = $226; //@line 336 "load_mdl.cpp"
         _memset(($226|0),0,632)|0; //@line 337 "load_mdl.cpp"
         $$pre179 = HEAP32[$9>>2]|0; //@line 334 "load_mdl.cpp"
         $229 = $$pre179;
        } else {
         $229 = $508;
        }
        $227 = (($j$3125) + 1)|0; //@line 334 "load_mdl.cpp"
        $228 = ($227>>>0)>($229>>>0); //@line 334 "load_mdl.cpp"
        if ($228) {
         $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
         break;
        } else {
         $508 = $229;$j$3125 = $227;
        }
       }
      }
      break;
     }
     case 17750:  {
      $230 = (($lpStream) + ($22)|0); //@line 342 "load_mdl.cpp"
      $231 = HEAP8[$230>>0]|0; //@line 342 "load_mdl.cpp"
      $232 = $231&255; //@line 342 "load_mdl.cpp"
      $233 = ($231<<24>>24)==(0); //@line 342 "load_mdl.cpp"
      if ($233) {
       $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $232;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
      } else {
       $234 = $232 << 5; //@line 343 "load_mdl.cpp"
       $235 = (($22) + 1)|0; //@line 343 "load_mdl.cpp"
       $236 = (($235) + ($234))|0; //@line 343 "load_mdl.cpp"
       $237 = ($236>>>0)>($dwMemLength>>>0); //@line 343 "load_mdl.cpp"
       if ($237) {
        $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $232;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
       } else {
        $$sum24 = (($dwMemPos$0147) + 7)|0; //@line 343 "load_mdl.cpp"
        $238 = (($lpStream) + ($$sum24)|0); //@line 343 "load_mdl.cpp"
        $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $232;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $238;
       }
      }
      break;
     }
     case 17744:  {
      $239 = (($lpStream) + ($22)|0); //@line 347 "load_mdl.cpp"
      $240 = HEAP8[$239>>0]|0; //@line 347 "load_mdl.cpp"
      $241 = $240&255; //@line 347 "load_mdl.cpp"
      $242 = ($240<<24>>24)==(0); //@line 347 "load_mdl.cpp"
      if ($242) {
       $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $241;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
      } else {
       $243 = $241 << 5; //@line 348 "load_mdl.cpp"
       $244 = (($22) + 1)|0; //@line 348 "load_mdl.cpp"
       $245 = (($244) + ($243))|0; //@line 348 "load_mdl.cpp"
       $246 = ($245>>>0)>($dwMemLength>>>0); //@line 348 "load_mdl.cpp"
       if ($246) {
        $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $241;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
       } else {
        $$sum23 = (($dwMemPos$0147) + 7)|0; //@line 348 "load_mdl.cpp"
        $247 = (($lpStream) + ($$sum23)|0); //@line 348 "load_mdl.cpp"
        $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $241;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $247;$pvolenv$0$be = $pvolenv$0140;
       }
      }
      break;
     }
     case 16723:  {
      $28 = HEAP32[$10>>2]|0; //@line 387 "load_mdl.cpp"
      $29 = ($28|0)==(0); //@line 387 "load_mdl.cpp"
      if ($29) {
       $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
      } else {
       $dwPos$3110 = $22;$i$4109 = 1;
       while(1) {
        $322 = (((($this)) + 33280|0) + ($i$4109<<6)|0); //@line 387 "load_mdl.cpp"
        $323 = HEAP32[$322>>2]|0; //@line 387 "load_mdl.cpp"
        $324 = ($323|0)==(0); //@line 387 "load_mdl.cpp"
        do {
         if ($324) {
          $dwPos$4 = $dwPos$3110;
         } else {
          $325 = (((((($this)) + 33280|0) + ($i$4109<<6)|0)) + 20|0); //@line 387 "load_mdl.cpp"
          $326 = HEAP32[$325>>2]|0; //@line 387 "load_mdl.cpp"
          $327 = ($326|0)==(0|0); //@line 387 "load_mdl.cpp"
          if ($327) {
           $328 = (($smpinfo) + ($i$4109)|0); //@line 387 "load_mdl.cpp"
           $329 = HEAP8[$328>>0]|0; //@line 387 "load_mdl.cpp"
           $330 = ($329<<24>>24)!=(3); //@line 387 "load_mdl.cpp"
           $331 = ($dwPos$3110>>>0)<($dwMemLength>>>0); //@line 387 "load_mdl.cpp"
           $or$cond56 = $331 & $330; //@line 387 "load_mdl.cpp"
           if ($or$cond56) {
            $332 = (((((($this)) + 33280|0) + ($i$4109<<6)|0)) + 34|0); //@line 390 "load_mdl.cpp"
            $333 = HEAP16[$332>>1]|0; //@line 390 "load_mdl.cpp"
            $334 = $333 & 1; //@line 390 "load_mdl.cpp"
            $335 = ($329<<24>>24)==(0); //@line 391 "load_mdl.cpp"
            if ($335) {
             $336 = ($334<<16>>16)!=(0); //@line 390 "load_mdl.cpp"
             $337 = $336 ? 5 : 0; //@line 390 "load_mdl.cpp"
             $338 = (($lpStream) + ($dwPos$3110)|0); //@line 393 "load_mdl.cpp"
             $339 = (($dwMemLength) - ($dwPos$3110))|0; //@line 393 "load_mdl.cpp"
             $340 = (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$322,$337,$338,$339)|0); //@line 393 "load_mdl.cpp"
             $341 = (($340) + ($dwPos$3110))|0; //@line 393 "load_mdl.cpp"
             $dwPos$4 = $341;
             break;
            }
            $342 = (($lpStream) + ($dwPos$3110)|0); //@line 396 "load_mdl.cpp"
            $343 = HEAP32[$342>>2]|0; //@line 396 "load_mdl.cpp"
            $344 = (($dwPos$3110) + 4)|0; //@line 397 "load_mdl.cpp"
            $345 = ($343>>>0)<($dwMemLength>>>0); //@line 398 "load_mdl.cpp"
            if ($345) {
             $346 = (($dwMemLength) - ($344))|0; //@line 398 "load_mdl.cpp"
             $347 = ($343>>>0)<=($346>>>0); //@line 398 "load_mdl.cpp"
             $348 = ($343>>>0)>(4); //@line 398 "load_mdl.cpp"
             $or$cond8 = $347 & $348; //@line 398 "load_mdl.cpp"
             if ($or$cond8) {
              $349 = $334&65535; //@line 400 "load_mdl.cpp"
              $350 = $349 << 2; //@line 400 "load_mdl.cpp"
              $351 = $350 | 32; //@line 400 "load_mdl.cpp"
              $352 = (($lpStream) + ($344)|0); //@line 401 "load_mdl.cpp"
              (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$322,$351,$352,$343)|0); //@line 401 "load_mdl.cpp"
             }
            }
            $353 = (($343) + ($344))|0; //@line 403 "load_mdl.cpp"
            $dwPos$4 = $353;
           } else {
            $dwPos$4 = $dwPos$3110;
           }
          } else {
           $dwPos$4 = $dwPos$3110;
          }
         }
        } while(0);
        $354 = (($i$4109) + 1)|0; //@line 387 "load_mdl.cpp"
        $355 = HEAP32[$10>>2]|0; //@line 387 "load_mdl.cpp"
        $356 = ($354>>>0)>($355>>>0); //@line 387 "load_mdl.cpp"
        if ($356) {
         $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
         break;
        } else {
         $dwPos$3110 = $dwPos$4;$i$4109 = $354;
        }
       }
      }
      break;
     }
     case 21321:  {
      $248 = (($lpStream) + ($22)|0); //@line 357 "load_mdl.cpp"
      $249 = HEAP8[$248>>0]|0; //@line 357 "load_mdl.cpp"
      $250 = ($249<<24>>24)==(0); //@line 359 "load_mdl.cpp"
      if ($250) {
       $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
      } else {
       $251 = (($dwMemPos$0147) + 7)|0; //@line 358 "load_mdl.cpp"
       $252 = $249&255;
       $dwPos$2105 = $251;$i$3104 = 0;
       while(1) {
        $253 = (($lpStream) + ($dwPos$2105)|0); //@line 361 "load_mdl.cpp"
        $254 = HEAP8[$253>>0]|0; //@line 361 "load_mdl.cpp"
        $255 = $254&255; //@line 361 "load_mdl.cpp"
        $$off = (($254) + -1)<<24>>24; //@line 362 "load_mdl.cpp"
        $256 = ($$off&255)<(239); //@line 362 "load_mdl.cpp"
        if ($256) {
         $257 = HEAP32[$10>>2]|0; //@line 363 "load_mdl.cpp"
         $258 = ($257>>>0)<($255>>>0); //@line 363 "load_mdl.cpp"
         if ($258) {
          HEAP32[$10>>2] = $255; //@line 363 "load_mdl.cpp"
         }
         $259 = (((($this)) + 59804|0) + ($255<<5)|0); //@line 365 "load_mdl.cpp"
         $$sum17 = (($dwPos$2105) + 1)|0; //@line 365 "load_mdl.cpp"
         $260 = (($lpStream) + ($$sum17)|0); //@line 365 "load_mdl.cpp"
         dest=$259; src=$260; stop=dest+32|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 365 "load_mdl.cpp"
         $261 = (((((($this)) + 33280|0) + ($255<<6)|0)) + 42|0); //@line 366 "load_mdl.cpp"
         $$sum18 = (($dwPos$2105) + 33)|0; //@line 366 "load_mdl.cpp"
         $262 = (($lpStream) + ($$sum18)|0); //@line 366 "load_mdl.cpp"
         $263 = $262; //@line 366 "load_mdl.cpp"
         $264 = $263; //@line 366 "load_mdl.cpp"
         $265 = HEAPU8[$264>>0]|(HEAPU8[$264+1>>0]<<8)|(HEAPU8[$264+2>>0]<<16)|(HEAPU8[$264+3>>0]<<24); //@line 366 "load_mdl.cpp"
         $266 = (($263) + 4)|0; //@line 366 "load_mdl.cpp"
         $267 = $266; //@line 366 "load_mdl.cpp"
         $268 = HEAPU8[$267>>0]|(HEAPU8[$267+1>>0]<<8)|(HEAPU8[$267+2>>0]<<16)|(HEAPU8[$267+3>>0]<<24); //@line 366 "load_mdl.cpp"
         $269 = $261; //@line 366 "load_mdl.cpp"
         $270 = $269; //@line 366 "load_mdl.cpp"
         HEAP8[$270>>0]=$265&255;HEAP8[$270+1>>0]=($265>>8)&255;HEAP8[$270+2>>0]=($265>>16)&255;HEAP8[$270+3>>0]=$265>>24; //@line 366 "load_mdl.cpp"
         $271 = (($269) + 4)|0; //@line 366 "load_mdl.cpp"
         $272 = $271; //@line 366 "load_mdl.cpp"
         HEAP8[$272>>0]=$268&255;HEAP8[$272+1>>0]=($268>>8)&255;HEAP8[$272+2>>0]=($268>>16)&255;HEAP8[$272+3>>0]=$268>>24; //@line 366 "load_mdl.cpp"
         $$sum19 = (($dwPos$2105) + 41)|0; //@line 367 "load_mdl.cpp"
         $273 = (($lpStream) + ($$sum19)|0); //@line 367 "load_mdl.cpp"
         $274 = HEAP32[$273>>2]|0; //@line 367 "load_mdl.cpp"
         $275 = (((((($this)) + 33280|0) + ($255<<6)|0)) + 24|0); //@line 367 "load_mdl.cpp"
         HEAP32[$275>>2] = $274; //@line 367 "load_mdl.cpp"
         $$sum20 = (($dwPos$2105) + 45)|0; //@line 368 "load_mdl.cpp"
         $276 = (($lpStream) + ($$sum20)|0); //@line 368 "load_mdl.cpp"
         $277 = HEAP32[$276>>2]|0; //@line 368 "load_mdl.cpp"
         $278 = (((($this)) + 33280|0) + ($255<<6)|0); //@line 368 "load_mdl.cpp"
         HEAP32[$278>>2] = $277; //@line 368 "load_mdl.cpp"
         $$sum21 = (($dwPos$2105) + 49)|0; //@line 369 "load_mdl.cpp"
         $279 = (($lpStream) + ($$sum21)|0); //@line 369 "load_mdl.cpp"
         $280 = HEAP32[$279>>2]|0; //@line 369 "load_mdl.cpp"
         $281 = (((((($this)) + 33280|0) + ($255<<6)|0)) + 4|0); //@line 369 "load_mdl.cpp"
         HEAP32[$281>>2] = $280; //@line 369 "load_mdl.cpp"
         $$sum22 = (($dwPos$2105) + 53)|0; //@line 370 "load_mdl.cpp"
         $282 = (($lpStream) + ($$sum22)|0); //@line 370 "load_mdl.cpp"
         $283 = HEAP32[$282>>2]|0; //@line 370 "load_mdl.cpp"
         $284 = (($283) + ($280))|0; //@line 370 "load_mdl.cpp"
         $285 = (((((($this)) + 33280|0) + ($255<<6)|0)) + 8|0); //@line 370 "load_mdl.cpp"
         HEAP32[$285>>2] = $284; //@line 370 "load_mdl.cpp"
         $286 = ($284>>>0)>($280>>>0); //@line 371 "load_mdl.cpp"
         if ($286) {
          $287 = (((((($this)) + 33280|0) + ($255<<6)|0)) + 34|0); //@line 371 "load_mdl.cpp"
          $288 = HEAP16[$287>>1]|0; //@line 371 "load_mdl.cpp"
          $289 = $288&65535; //@line 371 "load_mdl.cpp"
          $290 = $289 | 2; //@line 371 "load_mdl.cpp"
          $291 = $290&65535; //@line 371 "load_mdl.cpp"
          HEAP16[$287>>1] = $291; //@line 371 "load_mdl.cpp"
         }
         $292 = (((((($this)) + 33280|0) + ($255<<6)|0)) + 32|0); //@line 372 "load_mdl.cpp"
         HEAP16[$292>>1] = 64; //@line 372 "load_mdl.cpp"
         $293 = (($dwPos$2105) + 58)|0; //@line 373 "load_mdl.cpp"
         $294 = (($lpStream) + ($293)|0); //@line 373 "load_mdl.cpp"
         $295 = HEAP8[$294>>0]|0; //@line 373 "load_mdl.cpp"
         $296 = $295 & 1; //@line 373 "load_mdl.cpp"
         $297 = ($296<<24>>24)==(0); //@line 373 "load_mdl.cpp"
         if ($297) {
          $307 = $295;
         } else {
          $298 = (((((($this)) + 33280|0) + ($255<<6)|0)) + 34|0); //@line 375 "load_mdl.cpp"
          $299 = HEAP16[$298>>1]|0; //@line 375 "load_mdl.cpp"
          $300 = $299&65535; //@line 375 "load_mdl.cpp"
          $301 = $300 | 1; //@line 375 "load_mdl.cpp"
          $302 = $301&65535; //@line 375 "load_mdl.cpp"
          HEAP16[$298>>1] = $302; //@line 375 "load_mdl.cpp"
          $303 = $277 >>> 1; //@line 376 "load_mdl.cpp"
          HEAP32[$278>>2] = $303; //@line 376 "load_mdl.cpp"
          $304 = $280 >>> 1; //@line 377 "load_mdl.cpp"
          HEAP32[$281>>2] = $304; //@line 377 "load_mdl.cpp"
          $305 = $284 >>> 1; //@line 378 "load_mdl.cpp"
          HEAP32[$285>>2] = $305; //@line 378 "load_mdl.cpp"
          $$pre = HEAP8[$294>>0]|0; //@line 380 "load_mdl.cpp"
          $307 = $$pre;
         }
         $306 = $307 & 2; //@line 380 "load_mdl.cpp"
         $308 = ($306<<24>>24)==(0); //@line 380 "load_mdl.cpp"
         if ($308) {
          $315 = $307;
         } else {
          $309 = (((((($this)) + 33280|0) + ($255<<6)|0)) + 34|0); //@line 380 "load_mdl.cpp"
          $310 = HEAP16[$309>>1]|0; //@line 380 "load_mdl.cpp"
          $311 = $310&65535; //@line 380 "load_mdl.cpp"
          $312 = $311 | 4; //@line 380 "load_mdl.cpp"
          $313 = $312&65535; //@line 380 "load_mdl.cpp"
          HEAP16[$309>>1] = $313; //@line 380 "load_mdl.cpp"
          $$pre176 = HEAP8[$294>>0]|0; //@line 381 "load_mdl.cpp"
          $315 = $$pre176;
         }
         $314 = $315&255; //@line 381 "load_mdl.cpp"
         $316 = $314 >>> 2; //@line 381 "load_mdl.cpp"
         $317 = $316 & 3; //@line 381 "load_mdl.cpp"
         $318 = $317&255; //@line 381 "load_mdl.cpp"
         $319 = (($smpinfo) + ($255)|0); //@line 381 "load_mdl.cpp"
         HEAP8[$319>>0] = $318; //@line 381 "load_mdl.cpp"
        }
        $320 = (($i$3104) + 1)|0; //@line 359 "load_mdl.cpp"
        $321 = (($dwPos$2105) + 59)|0; //@line 359 "load_mdl.cpp"
        $exitcond169 = ($320|0)==($252|0); //@line 359 "load_mdl.cpp"
        if ($exitcond169) {
         $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
         break;
        } else {
         $dwPos$2105 = $321;$i$3104 = $320;
        }
       }
      }
      break;
     }
     default: {
      $dwTrackPos$0$be = $dwTrackPos$0146;$npanenv$0$be = $npanenv$0143;$npatterns$0$be = $npatterns$0145;$ntracks$0$be = $ntracks$0144;$nvolenv$0$be = $nvolenv$0142;$ppanenv$0$be = $ppanenv$0141;$pvolenv$0$be = $pvolenv$0140;
     }
     }
    } while(0);
    $68 = (($21) + 6)|0; //@line 207 "load_mdl.cpp"
    $69 = ($68>>>0)<($dwMemLength>>>0); //@line 207 "load_mdl.cpp"
    if ($69) {
     $22 = $68;$dwMemPos$0147 = $21;$dwTrackPos$0146 = $dwTrackPos$0$be;$npanenv$0143 = $npanenv$0$be;$npatterns$0145 = $npatterns$0$be;$ntracks$0144 = $ntracks$0$be;$nvolenv$0142 = $nvolenv$0$be;$ppanenv$0141 = $ppanenv$0$be;$pvolenv$0140 = $pvolenv$0$be;
    } else {
     $dwTrackPos$088 = $dwTrackPos$0$be;$npanenv$082 = $npanenv$0$be;$npatterns$086 = $npatterns$0$be;$ntracks$084 = $ntracks$0$be;$nvolenv$080 = $nvolenv$0$be;$ppanenv$078 = $ppanenv$0$be;$pvolenv$076 = $pvolenv$0$be;
     break;
    }
   }
   if ((label|0) == 7) {
    $24 = ($$lcssa262|0)==(11); //@line 214 "load_mdl.cpp"
    if ($24) {
     $$0 = 0;
     STACKTOP = sp;return ($$0|0); //@line 482 "load_mdl.cpp"
    } else {
     $dwTrackPos$088 = $dwTrackPos$0146$lcssa;$npanenv$082 = $npanenv$0143$lcssa;$npatterns$086 = $npatterns$0145$lcssa;$ntracks$084 = $ntracks$0144$lcssa;$nvolenv$080 = $nvolenv$0142$lcssa;$ppanenv$078 = $ppanenv$0141$lcssa;$pvolenv$076 = $pvolenv$0140$lcssa;
    }
   }
   $357 = ($dwTrackPos$088|0)!=(0); //@line 411 "load_mdl.cpp"
   $358 = ($npatterns$086|0)!=(0); //@line 411 "load_mdl.cpp"
   $or$cond10 = $358 & $357; //@line 411 "load_mdl.cpp"
   if ($or$cond10) {
    $365 = ((($this)) + 59664|0); //@line 411 "load_mdl.cpp"
    $366 = HEAP32[$365>>2]|0; //@line 411 "load_mdl.cpp"
    $367 = ($366|0)!=(0); //@line 411 "load_mdl.cpp"
    $368 = ($ntracks$084|0)!=(0); //@line 411 "load_mdl.cpp"
    $or$cond12 = $368 & $367; //@line 411 "load_mdl.cpp"
    if ($or$cond12) {
     $369 = (($lpStream) + ($dwTrackPos$088)|0); //@line 419 "load_mdl.cpp"
     $373 = $366;$ipat$0101 = 0;
     while(1) {
      $370 = (((($this)) + 52864|0) + ($ipat$0101<<1)|0); //@line 415 "load_mdl.cpp"
      $371 = HEAP16[$370>>1]|0; //@line 415 "load_mdl.cpp"
      $372 = $371&65535; //@line 415 "load_mdl.cpp"
      $374 = (__ZN10CSoundFile15AllocatePatternEjj($372,$373)|0); //@line 415 "load_mdl.cpp"
      $375 = (((($this)) + 51904|0) + ($ipat$0101<<2)|0); //@line 415 "load_mdl.cpp"
      HEAP32[$375>>2] = $374; //@line 415 "load_mdl.cpp"
      $376 = ($374|0)==(0|0); //@line 415 "load_mdl.cpp"
      if ($376) {
       $npanenv$082203 = $npanenv$082;$nvolenv$080204 = $nvolenv$080;$ppanenv$078205 = $ppanenv$078;$pvolenv$076206 = $pvolenv$076;
       break L10;
      }
      $377 = HEAP32[$365>>2]|0; //@line 416 "load_mdl.cpp"
      $378 = ($377|0)==(0); //@line 416 "load_mdl.cpp"
      if ($378) {
       $510 = 0;
      } else {
       $379 = $ipat$0101 << 5; //@line 416 "load_mdl.cpp"
       $390 = $377;$chn$098 = 0;
       while(1) {
        $380 = (($chn$098) + ($379))|0; //@line 416 "load_mdl.cpp"
        $381 = (($patterntracks) + ($380<<1)|0); //@line 416 "load_mdl.cpp"
        $382 = HEAP16[$381>>1]|0; //@line 416 "load_mdl.cpp"
        $383 = ($382<<16>>16)==(0); //@line 416 "load_mdl.cpp"
        if ($383) {
         $393 = $390;
        } else {
         $384 = $382&65535; //@line 416 "load_mdl.cpp"
         $385 = ($384>>>0)>($ntracks$084>>>0); //@line 416 "load_mdl.cpp"
         if ($385) {
          $393 = $390;
         } else {
          $386 = HEAP32[$375>>2]|0; //@line 418 "load_mdl.cpp"
          $387 = (($386) + (($chn$098*6)|0)|0); //@line 418 "load_mdl.cpp"
          $388 = HEAP16[$370>>1]|0; //@line 419 "load_mdl.cpp"
          $389 = $388&65535; //@line 419 "load_mdl.cpp"
          __Z14UnpackMDLTrackP11_MODCOMMANDjjjPKh($387,$390,$389,$384,$369); //@line 419 "load_mdl.cpp"
          $$pre180 = HEAP32[$365>>2]|0; //@line 416 "load_mdl.cpp"
          $393 = $$pre180;
         }
        }
        $391 = (($chn$098) + 1)|0; //@line 416 "load_mdl.cpp"
        $392 = ($391>>>0)<($393>>>0); //@line 416 "load_mdl.cpp"
        if ($392) {
         $390 = $393;$chn$098 = $391;
        } else {
         $510 = $393;
         break;
        }
       }
      }
      $394 = (($ipat$0101) + 1)|0; //@line 413 "load_mdl.cpp"
      $395 = ($394>>>0)<($npatterns$086>>>0); //@line 413 "load_mdl.cpp"
      if ($395) {
       $373 = $510;$ipat$0101 = $394;
      } else {
       $npanenv$082203 = $npanenv$082;$nvolenv$080204 = $nvolenv$080;$ppanenv$078205 = $ppanenv$078;$pvolenv$076206 = $pvolenv$076;
       break;
      }
     }
    } else {
     $npanenv$082203 = $npanenv$082;$nvolenv$080204 = $nvolenv$080;$ppanenv$078205 = $ppanenv$078;$pvolenv$076206 = $pvolenv$076;
    }
   } else {
    $npanenv$082203 = $npanenv$082;$nvolenv$080204 = $nvolenv$080;$ppanenv$078205 = $ppanenv$078;$pvolenv$076206 = $pvolenv$076;
   }
  } else {
   $npanenv$082203 = 0;$nvolenv$080204 = 0;$ppanenv$078205 = 0;$pvolenv$076206 = 0;
  }
 } while(0);
 $359 = HEAP32[$9>>2]|0; //@line 424 "load_mdl.cpp"
 $360 = ($359|0)==(0); //@line 424 "load_mdl.cpp"
 if (!($360)) {
  $361 = ($nvolenv$080204|0)!=(0); //@line 428 "load_mdl.cpp"
  $362 = ($pvolenv$076206|0)!=(0|0); //@line 428 "load_mdl.cpp"
  $or$cond14 = $362 & $361; //@line 428 "load_mdl.cpp"
  $363 = ($npanenv$082203|0)!=(0); //@line 454 "load_mdl.cpp"
  $364 = ($ppanenv$078205|0)!=(0|0); //@line 454 "load_mdl.cpp"
  $or$cond16 = $364 & $363; //@line 454 "load_mdl.cpp"
  $iIns$096 = 1;
  while(1) {
   $396 = (((($this)) + 48640|0) + ($iIns$096<<2)|0); //@line 424 "load_mdl.cpp"
   $397 = HEAP32[$396>>2]|0; //@line 424 "load_mdl.cpp"
   $398 = ($397|0)==(0|0); //@line 424 "load_mdl.cpp"
   if (!($398)) {
    if ($or$cond14) {
     $399 = (($insvolenv) + ($iIns$096)|0); //@line 428 "load_mdl.cpp"
     $400 = HEAP8[$399>>0]|0; //@line 428 "load_mdl.cpp"
     $401 = ($400<<24>>24)==(0); //@line 428 "load_mdl.cpp"
     if (!($401)) {
      $402 = ((($397)) + 556|0); //@line 434 "load_mdl.cpp"
      $403 = ((($397)) + 562|0); //@line 446 "load_mdl.cpp"
      $404 = ((($397)) + 561|0); //@line 446 "load_mdl.cpp"
      $405 = ((($397)) + 559|0); //@line 449 "load_mdl.cpp"
      $406 = ((($397)) + 560|0); //@line 450 "load_mdl.cpp"
      $407 = ((($397)) + 4|0); //@line 448 "load_mdl.cpp"
      $408 = $400&255; //@line 431 "load_mdl.cpp"
      $nve$092 = 0;$pve$091 = $pvolenv$076206;
      while(1) {
       $409 = HEAP8[$pve$091>>0]|0; //@line 431 "load_mdl.cpp"
       $410 = $409&255; //@line 431 "load_mdl.cpp"
       $411 = (($410) + 1)|0; //@line 431 "load_mdl.cpp"
       $412 = ($411|0)==($408|0); //@line 431 "load_mdl.cpp"
       if ($412) {
        HEAP8[$402>>0] = 15; //@line 434 "load_mdl.cpp"
        $iv$0 = 0;$vtick$0 = 1;
        while(1) {
         $413 = ($iv$0>>>0)<(15); //@line 435 "load_mdl.cpp"
         if (!($413)) {
          break;
         }
         $414 = ($iv$0|0)==(0); //@line 437 "load_mdl.cpp"
         $$pre195 = $iv$0 << 1; //@line 439 "load_mdl.cpp"
         $$pre197 = $$pre195 | 1; //@line 440 "load_mdl.cpp"
         $$pre199 = (($pve$091) + ($$pre197)|0); //@line 440 "load_mdl.cpp"
         if ($414) {
          $$pre$phi196Z2D = 0;$vtick$1 = $vtick$0;
         } else {
          $415 = HEAP8[$$pre199>>0]|0; //@line 437 "load_mdl.cpp"
          $416 = $415&255; //@line 437 "load_mdl.cpp"
          $417 = $vtick$0&65535; //@line 437 "load_mdl.cpp"
          $418 = (($416) + ($417))|0; //@line 437 "load_mdl.cpp"
          $419 = $418&65535; //@line 437 "load_mdl.cpp"
          $$pre$phi196Z2D = $$pre195;$vtick$1 = $419;
         }
         $420 = (((($397)) + 12|0) + ($iv$0<<1)|0); //@line 438 "load_mdl.cpp"
         HEAP16[$420>>1] = $vtick$1; //@line 438 "load_mdl.cpp"
         $421 = (($$pre$phi196Z2D) + 2)|0; //@line 439 "load_mdl.cpp"
         $422 = (($pve$091) + ($421)|0); //@line 439 "load_mdl.cpp"
         $423 = HEAP8[$422>>0]|0; //@line 439 "load_mdl.cpp"
         $424 = (((($397)) + 204|0) + ($iv$0)|0); //@line 439 "load_mdl.cpp"
         HEAP8[$424>>0] = $423; //@line 439 "load_mdl.cpp"
         $425 = HEAP8[$$pre199>>0]|0; //@line 440 "load_mdl.cpp"
         $426 = ($425<<24>>24)==(0); //@line 440 "load_mdl.cpp"
         $427 = (($iv$0) + 1)|0; //@line 435 "load_mdl.cpp"
         if ($426) {
          $$lcssa = $427;
          label = 109;
          break;
         } else {
          $iv$0 = $427;$vtick$0 = $vtick$1;
         }
        }
        if ((label|0) == 109) {
         label = 0;
         $428 = $$lcssa&255; //@line 442 "load_mdl.cpp"
         HEAP8[$402>>0] = $428; //@line 442 "load_mdl.cpp"
        }
        $429 = ((($pve$091)) + 31|0); //@line 446 "load_mdl.cpp"
        $430 = HEAP8[$429>>0]|0; //@line 446 "load_mdl.cpp"
        $431 = $430&255; //@line 446 "load_mdl.cpp"
        $432 = $431 & 15; //@line 446 "load_mdl.cpp"
        $433 = $432&255; //@line 446 "load_mdl.cpp"
        HEAP8[$403>>0] = $433; //@line 446 "load_mdl.cpp"
        HEAP8[$404>>0] = $433; //@line 446 "load_mdl.cpp"
        $434 = HEAP8[$429>>0]|0; //@line 447 "load_mdl.cpp"
        $435 = $434 & 16; //@line 447 "load_mdl.cpp"
        $436 = ($435<<24>>24)==(0); //@line 447 "load_mdl.cpp"
        if ($436) {
         $440 = $434;
        } else {
         $437 = HEAP32[$407>>2]|0; //@line 447 "load_mdl.cpp"
         $438 = $437 | 2; //@line 447 "load_mdl.cpp"
         HEAP32[$407>>2] = $438; //@line 447 "load_mdl.cpp"
         $$pre181 = HEAP8[$429>>0]|0; //@line 448 "load_mdl.cpp"
         $440 = $$pre181;
        }
        $439 = $440 & 32; //@line 448 "load_mdl.cpp"
        $441 = ($439<<24>>24)==(0); //@line 448 "load_mdl.cpp"
        if (!($441)) {
         $442 = HEAP32[$407>>2]|0; //@line 448 "load_mdl.cpp"
         $443 = $442 | 4; //@line 448 "load_mdl.cpp"
         HEAP32[$407>>2] = $443; //@line 448 "load_mdl.cpp"
        }
        $444 = ((($pve$091)) + 32|0); //@line 449 "load_mdl.cpp"
        $445 = HEAP8[$444>>0]|0; //@line 449 "load_mdl.cpp"
        $446 = $445&255; //@line 449 "load_mdl.cpp"
        $447 = $446 & 15; //@line 449 "load_mdl.cpp"
        $448 = $447&255; //@line 449 "load_mdl.cpp"
        HEAP8[$405>>0] = $448; //@line 449 "load_mdl.cpp"
        $449 = HEAP8[$444>>0]|0; //@line 450 "load_mdl.cpp"
        $450 = ($449&255) >>> 4; //@line 450 "load_mdl.cpp"
        HEAP8[$406>>0] = $450; //@line 450 "load_mdl.cpp"
       }
       $451 = (($nve$092) + 1)|0; //@line 431 "load_mdl.cpp"
       $452 = ((($pve$091)) + 33|0); //@line 431 "load_mdl.cpp"
       $exitcond = ($451|0)==($nvolenv$080204|0); //@line 431 "load_mdl.cpp"
       if ($exitcond) {
        break;
       } else {
        $nve$092 = $451;$pve$091 = $452;
       }
      }
     }
    }
    if ($or$cond16) {
     $453 = (($inspanenv) + ($iIns$096)|0); //@line 454 "load_mdl.cpp"
     $454 = HEAP8[$453>>0]|0; //@line 454 "load_mdl.cpp"
     $455 = ($454<<24>>24)==(0); //@line 454 "load_mdl.cpp"
     if (!($455)) {
      $456 = ((($397)) + 557|0); //@line 460 "load_mdl.cpp"
      $457 = ((($397)) + 563|0); //@line 474 "load_mdl.cpp"
      $458 = ((($397)) + 564|0); //@line 475 "load_mdl.cpp"
      $459 = ((($397)) + 4|0); //@line 473 "load_mdl.cpp"
      $460 = $454&255; //@line 457 "load_mdl.cpp"
      $npe$094 = 0;$ppe$093 = $ppanenv$078205;
      while(1) {
       $461 = HEAP8[$ppe$093>>0]|0; //@line 457 "load_mdl.cpp"
       $462 = $461&255; //@line 457 "load_mdl.cpp"
       $463 = (($462) + 1)|0; //@line 457 "load_mdl.cpp"
       $464 = ($463|0)==($460|0); //@line 457 "load_mdl.cpp"
       if ($464) {
        HEAP8[$456>>0] = 15; //@line 460 "load_mdl.cpp"
        $iv5$0 = 0;$vtick4$0 = 1;
        while(1) {
         $465 = ($iv5$0>>>0)<(15); //@line 461 "load_mdl.cpp"
         if (!($465)) {
          break;
         }
         $466 = ($iv5$0|0)==(0); //@line 463 "load_mdl.cpp"
         $$pre189 = $iv5$0 << 1; //@line 465 "load_mdl.cpp"
         $$pre191 = $$pre189 | 1; //@line 466 "load_mdl.cpp"
         $$pre193 = (($ppe$093) + ($$pre191)|0); //@line 466 "load_mdl.cpp"
         if ($466) {
          $$pre$phi190Z2D = 0;$vtick4$1 = $vtick4$0;
         } else {
          $467 = HEAP8[$$pre193>>0]|0; //@line 463 "load_mdl.cpp"
          $468 = $467&255; //@line 463 "load_mdl.cpp"
          $469 = $vtick4$0&65535; //@line 463 "load_mdl.cpp"
          $470 = (($468) + ($469))|0; //@line 463 "load_mdl.cpp"
          $471 = $470&65535; //@line 463 "load_mdl.cpp"
          $$pre$phi190Z2D = $$pre189;$vtick4$1 = $471;
         }
         $472 = (((($397)) + 76|0) + ($iv5$0<<1)|0); //@line 464 "load_mdl.cpp"
         HEAP16[$472>>1] = $vtick4$1; //@line 464 "load_mdl.cpp"
         $473 = (($$pre$phi190Z2D) + 2)|0; //@line 465 "load_mdl.cpp"
         $474 = (($ppe$093) + ($473)|0); //@line 465 "load_mdl.cpp"
         $475 = HEAP8[$474>>0]|0; //@line 465 "load_mdl.cpp"
         $476 = (((($397)) + 236|0) + ($iv5$0)|0); //@line 465 "load_mdl.cpp"
         HEAP8[$476>>0] = $475; //@line 465 "load_mdl.cpp"
         $477 = HEAP8[$$pre193>>0]|0; //@line 466 "load_mdl.cpp"
         $478 = ($477<<24>>24)==(0); //@line 466 "load_mdl.cpp"
         $479 = (($iv5$0) + 1)|0; //@line 461 "load_mdl.cpp"
         if ($478) {
          $$lcssa258 = $479;
          label = 125;
          break;
         } else {
          $iv5$0 = $479;$vtick4$0 = $vtick4$1;
         }
        }
        if ((label|0) == 125) {
         label = 0;
         $480 = $$lcssa258&255; //@line 468 "load_mdl.cpp"
         HEAP8[$456>>0] = $480; //@line 468 "load_mdl.cpp"
        }
        $481 = ((($ppe$093)) + 31|0); //@line 472 "load_mdl.cpp"
        $482 = HEAP8[$481>>0]|0; //@line 472 "load_mdl.cpp"
        $483 = $482 & 16; //@line 472 "load_mdl.cpp"
        $484 = ($483<<24>>24)==(0); //@line 472 "load_mdl.cpp"
        if ($484) {
         $488 = $482;
        } else {
         $485 = HEAP32[$459>>2]|0; //@line 472 "load_mdl.cpp"
         $486 = $485 | 16; //@line 472 "load_mdl.cpp"
         HEAP32[$459>>2] = $486; //@line 472 "load_mdl.cpp"
         $$pre182 = HEAP8[$481>>0]|0; //@line 473 "load_mdl.cpp"
         $488 = $$pre182;
        }
        $487 = $488 & 32; //@line 473 "load_mdl.cpp"
        $489 = ($487<<24>>24)==(0); //@line 473 "load_mdl.cpp"
        if (!($489)) {
         $490 = HEAP32[$459>>2]|0; //@line 473 "load_mdl.cpp"
         $491 = $490 | 32; //@line 473 "load_mdl.cpp"
         HEAP32[$459>>2] = $491; //@line 473 "load_mdl.cpp"
        }
        $492 = ((($ppe$093)) + 32|0); //@line 474 "load_mdl.cpp"
        $493 = HEAP8[$492>>0]|0; //@line 474 "load_mdl.cpp"
        $494 = $493&255; //@line 474 "load_mdl.cpp"
        $495 = $494 & 15; //@line 474 "load_mdl.cpp"
        $496 = $495&255; //@line 474 "load_mdl.cpp"
        HEAP8[$457>>0] = $496; //@line 474 "load_mdl.cpp"
        $497 = HEAP8[$492>>0]|0; //@line 475 "load_mdl.cpp"
        $498 = ($497&255) >>> 4; //@line 475 "load_mdl.cpp"
        HEAP8[$458>>0] = $498; //@line 475 "load_mdl.cpp"
       }
       $499 = (($npe$094) + 1)|0; //@line 457 "load_mdl.cpp"
       $500 = ((($ppe$093)) + 33|0); //@line 457 "load_mdl.cpp"
       $exitcond168 = ($499|0)==($npanenv$082203|0); //@line 457 "load_mdl.cpp"
       if ($exitcond168) {
        break;
       } else {
        $npe$094 = $499;$ppe$093 = $500;
       }
      }
     }
    }
   }
   $501 = (($iIns$096) + 1)|0; //@line 424 "load_mdl.cpp"
   $502 = HEAP32[$9>>2]|0; //@line 424 "load_mdl.cpp"
   $503 = ($501>>>0)>($502>>>0); //@line 424 "load_mdl.cpp"
   if ($503) {
    break;
   } else {
    $iIns$096 = $501;
   }
  }
 }
 $504 = ((($this)) + 59660|0); //@line 479 "load_mdl.cpp"
 $505 = HEAP32[$504>>2]|0; //@line 479 "load_mdl.cpp"
 $506 = $505 | 16; //@line 479 "load_mdl.cpp"
 HEAP32[$504>>2] = $506; //@line 479 "load_mdl.cpp"
 $507 = ((($this)) + 59680|0); //@line 480 "load_mdl.cpp"
 HEAP32[$507>>2] = 16384; //@line 480 "load_mdl.cpp"
 $$0 = 1;
 STACKTOP = sp;return ($$0|0); //@line 482 "load_mdl.cpp"
}
function __Z14UnpackMDLTrackP11_MODCOMMANDjjjPKh($pat,$nChannels,$nRows,$nTrack,$lpTracks) {
 $pat = $pat|0;
 $nChannels = $nChannels|0;
 $nRows = $nRows|0;
 $nTrack = $nTrack|0;
 $lpTracks = $lpTracks|0;
 var $$0$sum = 0, $$0$sum2 = 0, $$0$sum3 = 0, $$0$sum4 = 0, $$0$sum5 = 0, $$0$sum6 = 0, $$0$sum7 = 0, $$0$sum8 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $11 = 0, $12 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cmd = 0, $command1$0 = 0, $command2$0 = 0, $exitcond = 0, $exitcond28 = 0;
 var $i$012 = 0, $len$0$lcssa = 0, $len$019 = 0, $lpTracks$pn$lcssa = 0, $lpTracks$pn20 = 0, $m$0$be = 0, $m$017 = 0, $m$113 = 0, $ntrk$018 = 0, $or$cond = 0, $or$cond9 = 0, $or$cond914 = 0, $param1$0 = 0, $param2$0 = 0, $pos$0$be = 0, $pos$016 = 0, $pos$1 = 0, $pos$2 = 0, $pos$3 = 0, $pos$4 = 0;
 var $pos$5 = 0, $pos$6 = 0, $row$0$be = 0, $row$015 = 0, $row$111 = 0, $scevgep = 0, $umax = 0, $umax27 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $cmd = sp;
 $0 = HEAP16[$lpTracks>>1]|0; //@line 98 "load_mdl.cpp"
 $1 = $0&65535; //@line 98 "load_mdl.cpp"
 $2 = ($nTrack>>>0)>(1); //@line 101 "load_mdl.cpp"
 if ($2) {
  $len$019 = $1;$lpTracks$pn20 = $lpTracks;$ntrk$018 = 1;
  while(1) {
   $$0$sum8 = (($len$019) + 2)|0; //@line 103 "load_mdl.cpp"
   $3 = (($lpTracks$pn20) + ($$0$sum8)|0); //@line 103 "load_mdl.cpp"
   $4 = HEAP16[$3>>1]|0; //@line 104 "load_mdl.cpp"
   $5 = $4&65535; //@line 104 "load_mdl.cpp"
   $6 = (($ntrk$018) + 1)|0; //@line 101 "load_mdl.cpp"
   $exitcond28 = ($6|0)==($nTrack|0); //@line 101 "load_mdl.cpp"
   if ($exitcond28) {
    $len$0$lcssa = $5;$lpTracks$pn$lcssa = $3;
    break;
   } else {
    $len$019 = $5;$lpTracks$pn20 = $3;$ntrk$018 = $6;
   }
  }
 } else {
  $len$0$lcssa = $1;$lpTracks$pn$lcssa = $lpTracks;
 }
 $7 = ((($cmd)) + 1|0); //@line 107 "load_mdl.cpp"
 $8 = ((($cmd)) + 4|0); //@line 108 "load_mdl.cpp"
 $9 = ((($cmd)) + 2|0); //@line 108 "load_mdl.cpp"
 $10 = ((($cmd)) + 3|0); //@line 109 "load_mdl.cpp"
 $11 = ($nRows|0)!=(0); //@line 110 "load_mdl.cpp"
 $12 = ($len$0$lcssa|0)!=(0); //@line 110 "load_mdl.cpp"
 $or$cond914 = $11 & $12; //@line 110 "load_mdl.cpp"
 ;HEAP8[$cmd>>0]=0|0;HEAP8[$cmd+1>>0]=0|0;HEAP8[$cmd+2>>0]=0|0;HEAP8[$cmd+3>>0]=0|0;HEAP8[$cmd+4>>0]=0|0;HEAP8[$cmd+5>>0]=0|0; //@line 107 "load_mdl.cpp"
 if (!($or$cond914)) {
  STACKTOP = sp;return; //@line 175 "load_mdl.cpp"
 }
 $13 = (0 - ($nChannels))|0; //@line 120 "load_mdl.cpp"
 $m$017 = $pat;$pos$016 = 0;$row$015 = 0;
 while(1) {
  $16 = (($pos$016) + 1)|0; //@line 113 "load_mdl.cpp"
  $$0$sum = (($pos$016) + 2)|0; //@line 113 "load_mdl.cpp"
  $17 = (($lpTracks$pn$lcssa) + ($$0$sum)|0); //@line 113 "load_mdl.cpp"
  $18 = HEAP8[$17>>0]|0; //@line 113 "load_mdl.cpp"
  $19 = $18&255; //@line 114 "load_mdl.cpp"
  $20 = $19 >>> 2; //@line 114 "load_mdl.cpp"
  $21 = $19 & 3; //@line 115 "load_mdl.cpp"
  switch ($21|0) {
  case 1:  {
   $22 = (($row$015) + 1)|0; //@line 118 "load_mdl.cpp"
   $23 = ($22>>>0)<($nRows>>>0);
   $umax = $23 ? $nRows : $22;
   $24 = (($row$015) - ($umax))|0; //@line 118 "load_mdl.cpp"
   $25 = $20 ^ -1; //@line 118 "load_mdl.cpp"
   $26 = ($24>>>0)>($25>>>0);
   $umax27 = $26 ? $24 : $25;
   $15 = Math_imul($umax27, $13)|0; //@line 118 "load_mdl.cpp"
   $27 = (0 - ($umax27))|0; //@line 118 "load_mdl.cpp"
   $i$012 = 0;$m$113 = $m$017;$row$111 = $row$015;
   while(1) {
    $28 = ($row$111|0)==(0); //@line 120 "load_mdl.cpp"
    if (!($28)) {
     $29 = (($m$113) + (($13*6)|0)|0); //@line 120 "load_mdl.cpp"
     ;HEAP8[$m$113>>0]=HEAP8[$29>>0]|0;HEAP8[$m$113+1>>0]=HEAP8[$29+1>>0]|0;HEAP8[$m$113+2>>0]=HEAP8[$29+2>>0]|0;HEAP8[$m$113+3>>0]=HEAP8[$29+3>>0]|0;HEAP8[$m$113+4>>0]=HEAP8[$29+4>>0]|0;HEAP8[$m$113+5>>0]=HEAP8[$29+5>>0]|0; //@line 120 "load_mdl.cpp"
    }
    $30 = (($m$113) + (($nChannels*6)|0)|0); //@line 121 "load_mdl.cpp"
    $31 = (($row$111) + 1)|0; //@line 122 "load_mdl.cpp"
    $32 = (($i$012) + 1)|0; //@line 118 "load_mdl.cpp"
    $exitcond = ($32|0)==($27|0); //@line 123 "load_mdl.cpp"
    if ($exitcond) {
     break;
    } else {
     $i$012 = $32;$m$113 = $30;$row$111 = $31;
    }
   }
   $14 = (($row$015) - ($umax27))|0; //@line 118 "load_mdl.cpp"
   $scevgep = (($m$017) + (($15*6)|0)|0);
   $m$0$be = $scevgep;$pos$0$be = $16;$row$0$be = $14;
   break;
  }
  case 2:  {
   $33 = ($20>>>0)<($row$015>>>0); //@line 128 "load_mdl.cpp"
   if ($33) {
    $34 = Math_imul($20, $nChannels)|0; //@line 128 "load_mdl.cpp"
    $35 = (($pat) + (($34*6)|0)|0); //@line 128 "load_mdl.cpp"
    ;HEAP8[$m$017>>0]=HEAP8[$35>>0]|0;HEAP8[$m$017+1>>0]=HEAP8[$35+1>>0]|0;HEAP8[$m$017+2>>0]=HEAP8[$35+2>>0]|0;HEAP8[$m$017+3>>0]=HEAP8[$35+3>>0]|0;HEAP8[$m$017+4>>0]=HEAP8[$35+4>>0]|0;HEAP8[$m$017+5>>0]=HEAP8[$35+5>>0]|0; //@line 128 "load_mdl.cpp"
   }
   $36 = (($m$017) + (($nChannels*6)|0)|0); //@line 129 "load_mdl.cpp"
   $37 = (($row$015) + 1)|0; //@line 130 "load_mdl.cpp"
   $m$0$be = $36;$pos$0$be = $16;$row$0$be = $37;
   break;
  }
  case 3:  {
   $38 = $20 & 1; //@line 135 "load_mdl.cpp"
   $39 = ($38|0)==(0); //@line 135 "load_mdl.cpp"
   if ($39) {
    $44 = 0;$pos$1 = $16;
   } else {
    $$0$sum7 = (($pos$016) + 3)|0; //@line 135 "load_mdl.cpp"
    $40 = (($lpTracks$pn$lcssa) + ($$0$sum7)|0); //@line 135 "load_mdl.cpp"
    $41 = HEAP8[$40>>0]|0; //@line 135 "load_mdl.cpp"
    $42 = $41&255; //@line 135 "load_mdl.cpp"
    $44 = $42;$pos$1 = $$0$sum;
   }
   $43 = $44&255; //@line 135 "load_mdl.cpp"
   HEAP8[$cmd>>0] = $43; //@line 135 "load_mdl.cpp"
   $45 = $20 & 2; //@line 136 "load_mdl.cpp"
   $46 = ($45|0)==(0); //@line 136 "load_mdl.cpp"
   if ($46) {
    $52 = 0;$pos$2 = $pos$1;
   } else {
    $47 = (($pos$1) + 1)|0; //@line 136 "load_mdl.cpp"
    $$0$sum6 = (($pos$1) + 2)|0; //@line 136 "load_mdl.cpp"
    $48 = (($lpTracks$pn$lcssa) + ($$0$sum6)|0); //@line 136 "load_mdl.cpp"
    $49 = HEAP8[$48>>0]|0; //@line 136 "load_mdl.cpp"
    $50 = $49&255; //@line 136 "load_mdl.cpp"
    $52 = $50;$pos$2 = $47;
   }
   $51 = $52&255; //@line 136 "load_mdl.cpp"
   HEAP8[$7>>0] = $51; //@line 136 "load_mdl.cpp"
   $53 = (($43) + -1)<<24>>24; //@line 139 "load_mdl.cpp"
   $54 = ($53&255)>(106); //@line 139 "load_mdl.cpp"
   HEAP8[$9>>0]=0&255;HEAP8[$9+1>>0]=(0>>8)&255;HEAP8[$9+2>>0]=(0>>16)&255;HEAP8[$9+3>>0]=0>>24; //@line 137 "load_mdl.cpp"
   if (!($54)) {
    $55 = (($44) + 12)|0; //@line 139 "load_mdl.cpp"
    $56 = $55&255; //@line 139 "load_mdl.cpp"
    HEAP8[$cmd>>0] = $56; //@line 139 "load_mdl.cpp"
   }
   $57 = $20 & 4; //@line 140 "load_mdl.cpp"
   $58 = ($57|0)==(0); //@line 140 "load_mdl.cpp"
   if ($58) {
    $94 = 0;$pos$3 = $pos$2;
   } else {
    $59 = (($pos$2) + 1)|0; //@line 140 "load_mdl.cpp"
    $$0$sum5 = (($pos$2) + 2)|0; //@line 140 "load_mdl.cpp"
    $60 = (($lpTracks$pn$lcssa) + ($$0$sum5)|0); //@line 140 "load_mdl.cpp"
    $61 = HEAP8[$60>>0]|0; //@line 140 "load_mdl.cpp"
    $62 = $61&255; //@line 140 "load_mdl.cpp"
    $94 = $62;$pos$3 = $59;
   }
   $63 = $20 & 8; //@line 141 "load_mdl.cpp"
   $64 = ($63|0)==(0); //@line 141 "load_mdl.cpp"
   if ($64) {
    $70 = 0;$pos$4 = $pos$3;
   } else {
    $65 = (($pos$3) + 1)|0; //@line 141 "load_mdl.cpp"
    $$0$sum4 = (($pos$3) + 2)|0; //@line 141 "load_mdl.cpp"
    $66 = (($lpTracks$pn$lcssa) + ($$0$sum4)|0); //@line 141 "load_mdl.cpp"
    $67 = HEAP8[$66>>0]|0; //@line 141 "load_mdl.cpp"
    $68 = $67&255; //@line 141 "load_mdl.cpp"
    $70 = $68;$pos$4 = $65;
   }
   $69 = $70 & 15; //@line 142 "load_mdl.cpp"
   $71 = $70 & 240; //@line 143 "load_mdl.cpp"
   $72 = $20 & 16; //@line 144 "load_mdl.cpp"
   $73 = ($72|0)==(0); //@line 144 "load_mdl.cpp"
   if ($73) {
    $86 = 0;$pos$5 = $pos$4;
   } else {
    $74 = (($pos$4) + 1)|0; //@line 144 "load_mdl.cpp"
    $$0$sum3 = (($pos$4) + 2)|0; //@line 144 "load_mdl.cpp"
    $75 = (($lpTracks$pn$lcssa) + ($$0$sum3)|0); //@line 144 "load_mdl.cpp"
    $76 = HEAP8[$75>>0]|0; //@line 144 "load_mdl.cpp"
    $77 = $76&255; //@line 144 "load_mdl.cpp"
    $86 = $77;$pos$5 = $74;
   }
   $78 = $20 & 32; //@line 145 "load_mdl.cpp"
   $79 = ($78|0)==(0); //@line 145 "load_mdl.cpp"
   if ($79) {
    $92 = 0;$pos$6 = $pos$5;
   } else {
    $80 = (($pos$5) + 1)|0; //@line 145 "load_mdl.cpp"
    $$0$sum2 = (($pos$5) + 2)|0; //@line 145 "load_mdl.cpp"
    $81 = (($lpTracks$pn$lcssa) + ($$0$sum2)|0); //@line 145 "load_mdl.cpp"
    $82 = HEAP8[$81>>0]|0; //@line 145 "load_mdl.cpp"
    $83 = $82&255; //@line 145 "load_mdl.cpp"
    $92 = $83;$pos$6 = $80;
   }
   $84 = ($69|0)==(14); //@line 146 "load_mdl.cpp"
   if ($84) {
    $85 = $86 & 240; //@line 146 "load_mdl.cpp"
    $87 = ($85|0)!=(240); //@line 146 "load_mdl.cpp"
    $88 = ($71|0)!=(0); //@line 146 "load_mdl.cpp"
    $or$cond = $88 | $87; //@line 146 "load_mdl.cpp"
    if ($or$cond) {
     $command1$0 = 14;$command2$0 = $71;$param1$0 = $86;$param2$0 = $92;
    } else {
     $89 = $86 << 8; //@line 148 "load_mdl.cpp"
     $90 = $89 & 3840; //@line 148 "load_mdl.cpp"
     $91 = $92 | $90; //@line 148 "load_mdl.cpp"
     $command1$0 = 239;$command2$0 = 0;$param1$0 = $91;$param2$0 = 0;
    }
   } else {
    $command1$0 = $69;$command2$0 = $71;$param1$0 = $86;$param2$0 = $92;
   }
   $93 = ($94|0)==(0); //@line 152 "load_mdl.cpp"
   if (!($93)) {
    HEAP8[$9>>0] = 1; //@line 154 "load_mdl.cpp"
    $95 = (($94) + 1)|0; //@line 155 "load_mdl.cpp"
    $96 = $95 >>> 2; //@line 155 "load_mdl.cpp"
    $97 = $96&255; //@line 155 "load_mdl.cpp"
    HEAP8[$8>>0] = $97; //@line 155 "load_mdl.cpp"
   }
   __Z17ConvertMDLCommandP11_MODCOMMANDjj($cmd,$command1$0,$param1$0); //@line 157 "load_mdl.cpp"
   $98 = HEAP8[$10>>0]|0; //@line 158 "load_mdl.cpp"
   switch ($98<<24>>24) {
   case 14: case 17: case 16:  {
    break;
   }
   default: {
    __Z17ConvertMDLCommandP11_MODCOMMANDjj($cmd,$command2$0,$param2$0); //@line 161 "load_mdl.cpp"
   }
   }
   ;HEAP8[$m$017>>0]=HEAP8[$cmd>>0]|0;HEAP8[$m$017+1>>0]=HEAP8[$cmd+1>>0]|0;HEAP8[$m$017+2>>0]=HEAP8[$cmd+2>>0]|0;HEAP8[$m$017+3>>0]=HEAP8[$cmd+3>>0]|0;HEAP8[$m$017+4>>0]=HEAP8[$cmd+4>>0]|0;HEAP8[$m$017+5>>0]=HEAP8[$cmd+5>>0]|0; //@line 162 "load_mdl.cpp"
   $99 = (($m$017) + (($nChannels*6)|0)|0); //@line 163 "load_mdl.cpp"
   $100 = (($row$015) + 1)|0; //@line 164 "load_mdl.cpp"
   $m$0$be = $99;$pos$0$be = $pos$6;$row$0$be = $100;
   break;
  }
  default: {
   $101 = (($20) + 1)|0; //@line 170 "load_mdl.cpp"
   $102 = (($101) + ($row$015))|0; //@line 170 "load_mdl.cpp"
   $103 = Math_imul($101, $nChannels)|0; //@line 171 "load_mdl.cpp"
   $104 = (($m$017) + (($103*6)|0)|0); //@line 171 "load_mdl.cpp"
   $m$0$be = $104;$pos$0$be = $16;$row$0$be = $102;
  }
  }
  $105 = ($row$0$be>>>0)<($nRows>>>0); //@line 110 "load_mdl.cpp"
  $106 = ($pos$0$be>>>0)<($len$0$lcssa>>>0); //@line 110 "load_mdl.cpp"
  $or$cond9 = $105 & $106; //@line 110 "load_mdl.cpp"
  if ($or$cond9) {
   $m$017 = $m$0$be;$pos$016 = $pos$0$be;$row$015 = $row$0$be;
  } else {
   break;
  }
 }
 STACKTOP = sp;return; //@line 175 "load_mdl.cpp"
}
function __Z17ConvertMDLCommandP11_MODCOMMANDjj($m,$eff,$data) {
 $m = $m|0;
 $eff = $eff|0;
 $data = $data|0;
 var $$data = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $command$0$ph = 0, $data$ = 0, $data$1 = 0, $param$1$ph = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 L1: do {
  switch ($eff|0) {
  case 1:  {
   $command$0$ph = 2;$param$1$ph = $data;
   break;
  }
  case 2:  {
   $command$0$ph = 3;$param$1$ph = $data;
   break;
  }
  case 3:  {
   $command$0$ph = 4;$param$1$ph = $data;
   break;
  }
  case 4:  {
   $command$0$ph = 5;$param$1$ph = $data;
   break;
  }
  case 5:  {
   $command$0$ph = 1;$param$1$ph = $data;
   break;
  }
  case 7:  {
   $0 = ($data>>>0)<(32); //@line 56 "load_mdl.cpp"
   $phitmp = $0 ? 16 : 17; //@line 56 "load_mdl.cpp"
   $command$0$ph = $phitmp;$param$1$ph = $data;
   break;
  }
  case 8:  {
   $1 = $data << 1; //@line 57 "load_mdl.cpp"
   $command$0$ph = 9;$param$1$ph = $1;
   break;
  }
  case 11:  {
   $command$0$ph = 12;$param$1$ph = $data;
   break;
  }
  case 12:  {
   $command$0$ph = 23;$param$1$ph = $data;
   break;
  }
  case 13:  {
   $2 = $data & 15; //@line 60 "load_mdl.cpp"
   $3 = $data >>> 4; //@line 60 "load_mdl.cpp"
   $4 = ($3*10)|0; //@line 60 "load_mdl.cpp"
   $5 = (($4) + ($2))|0; //@line 60 "load_mdl.cpp"
   $command$0$ph = 14;$param$1$ph = $5;
   break;
  }
  case 14:  {
   $6 = $data & 240; //@line 63 "load_mdl.cpp"
   do {
    switch ($6|0) {
    case 16:  {
     $7 = $data & 15; //@line 66 "load_mdl.cpp"
     $8 = ($7|0)==(0); //@line 66 "load_mdl.cpp"
     $9 = $data | 240; //@line 66 "load_mdl.cpp"
     $data$1 = $8 ? $data : $9; //@line 66 "load_mdl.cpp"
     if (!($8)) {
      $command$0$ph = 29;$param$1$ph = $data$1;
      break L1;
     }
     return; //@line 91 "load_mdl.cpp"
     break;
    }
    case 32:  {
     $10 = $data & 15; //@line 67 "load_mdl.cpp"
     $11 = ($10|0)==(0); //@line 67 "load_mdl.cpp"
     if ($11) {
      return; //@line 91 "load_mdl.cpp"
     } else {
      $12 = $data << 4; //@line 67 "load_mdl.cpp"
      $13 = $12 | 15; //@line 67 "load_mdl.cpp"
      $command$0$ph = 29;$param$1$ph = $13;
      break L1;
     }
     break;
    }
    case 48:  {
     $14 = $data & 15; //@line 68 "load_mdl.cpp"
     $15 = $14 | 16; //@line 68 "load_mdl.cpp"
     $command$0$ph = 20;$param$1$ph = $15;
     break L1;
     break;
    }
    case 64:  {
     $16 = $data & 15; //@line 69 "load_mdl.cpp"
     $17 = $16 | 48; //@line 69 "load_mdl.cpp"
     $command$0$ph = 20;$param$1$ph = $17;
     break L1;
     break;
    }
    case 96:  {
     $18 = $data & 15; //@line 70 "load_mdl.cpp"
     $19 = $18 | 176; //@line 70 "load_mdl.cpp"
     $command$0$ph = 20;$param$1$ph = $19;
     break L1;
     break;
    }
    case 112:  {
     $20 = $data & 15; //@line 71 "load_mdl.cpp"
     $21 = $20 | 64; //@line 71 "load_mdl.cpp"
     $command$0$ph = 20;$param$1$ph = $21;
     break L1;
     break;
    }
    case 144:  {
     $22 = $data & 15; //@line 72 "load_mdl.cpp"
     $command$0$ph = 15;$param$1$ph = $22;
     break L1;
     break;
    }
    case 160:  {
     $23 = $data << 4; //@line 73 "load_mdl.cpp"
     $24 = $23 & 240; //@line 73 "load_mdl.cpp"
     $command$0$ph = 24;$param$1$ph = $24;
     break L1;
     break;
    }
    case 176:  {
     $25 = $data & 15; //@line 74 "load_mdl.cpp"
     $command$0$ph = 24;$param$1$ph = $25;
     break L1;
     break;
    }
    case 240:  {
     $26 = $data >>> 8; //@line 75 "load_mdl.cpp"
     $27 = $26 & 15; //@line 75 "load_mdl.cpp"
     $28 = $27 | 160; //@line 75 "load_mdl.cpp"
     $command$0$ph = 20;$param$1$ph = $28;
     break L1;
     break;
    }
    case 0:  {
     return; //@line 91 "load_mdl.cpp"
     break;
    }
    default: {
     $command$0$ph = 20;$param$1$ph = $data;
     break L1;
    }
    }
   } while(0);
   break;
  }
  case 15:  {
   $command$0$ph = 16;$param$1$ph = $data;
   break;
  }
  case 16:  {
   $29 = $data & 240; //@line 79 "load_mdl.cpp"
   switch ($29|0) {
   case 240:  {
    $30 = $data << 4; //@line 79 "load_mdl.cpp"
    $31 = $30 | 15; //@line 79 "load_mdl.cpp"
    $command$0$ph = 11;$param$1$ph = $31;
    break L1;
    break;
   }
   case 224:  {
    return; //@line 91 "load_mdl.cpp"
    break;
   }
   default: {
    $32 = $data >>> 2; //@line 79 "load_mdl.cpp"
    $command$0$ph = 11;$param$1$ph = $32;
    break L1;
   }
   }
   break;
  }
  case 32:  {
   $33 = $data & 240; //@line 80 "load_mdl.cpp"
   $34 = ($33|0)==(224); //@line 80 "load_mdl.cpp"
   if ($34) {
    return; //@line 91 "load_mdl.cpp"
   } else {
    $35 = ($33|0)==(240); //@line 80 "load_mdl.cpp"
    $36 = $data >>> 2; //@line 80 "load_mdl.cpp"
    $data$ = $35 ? $data : $36; //@line 80 "load_mdl.cpp"
    $command$0$ph = 11;$param$1$ph = $data$;
    break L1;
   }
   break;
  }
  case 48:  {
   $command$0$ph = 15;$param$1$ph = $data;
   break;
  }
  case 64:  {
   $command$0$ph = 8;$param$1$ph = $data;
   break;
  }
  case 80:  {
   $command$0$ph = 18;$param$1$ph = $data;
   break;
  }
  case 239:  {
   $37 = ($data>>>0)>(255); //@line 84 "load_mdl.cpp"
   $$data = $37 ? 255 : $data; //@line 84 "load_mdl.cpp"
   $command$0$ph = 10;$param$1$ph = $$data;
   break;
  }
  default: {
   return; //@line 91 "load_mdl.cpp"
  }
  }
 } while(0);
 $38 = ((($m)) + 3|0); //@line 88 "load_mdl.cpp"
 HEAP8[$38>>0] = $command$0$ph; //@line 88 "load_mdl.cpp"
 $39 = $param$1$ph&255; //@line 89 "load_mdl.cpp"
 $40 = ((($m)) + 5|0); //@line 89 "load_mdl.cpp"
 HEAP8[$40>>0] = $39; //@line 89 "load_mdl.cpp"
 return; //@line 91 "load_mdl.cpp"
}
function __ZN10CSoundFile7ReadDBMEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$0 = 0, $$51 = 0, $$56 = 0, $$60 = 0, $$lcssa = 0, $$lcssa223 = 0, $$lcssa224 = 0, $$lcssa225 = 0, $$lcssa226 = 0, $$lcssa227 = 0, $$lcssa228 = 0, $$lobit = 0, $$lobit65 = 0, $$lobit66 = 0, $$lobit67 = 0, $$lobit68 = 0, $$nInstruments$0 = 0, $$nPatterns$0 = 0, $$nSamples$0 = 0;
 var $$off = 0, $$off24 = 0, $$off25 = 0, $$off33 = 0, $$sum = 0, $$sum12 = 0, $$sum14 = 0, $$sum16 = 0, $$sum17 = 0, $$sum18 = 0, $$sum19 = 0, $$sum20 = 0, $$sum21 = 0, $$sum22 = 0, $$sum23 = 0, $$sum26 = 0, $$sum27 = 0, $$sum28 = 0, $$sum29 = 0, $$sum30 = 0;
 var $$sum31 = 0, $$sum32 = 0, $$sum34 = 0, $$sum35 = 0, $$sum36 = 0, $$sum37 = 0, $$sum38 = 0, $$sum39 = 0, $$sum40 = 0, $$sum41 = 0, $$v = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0;
 var $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0;
 var $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0;
 var $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0;
 var $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0;
 var $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0;
 var $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0;
 var $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0;
 var $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0;
 var $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0;
 var $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0;
 var $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0;
 var $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $chunk_pos$0109 = 0;
 var $chunk_pos$194 = 0, $chunk_pos$2124 = 0, $chunk_pos$3102 = 0, $cmd1$0 = 0, $cmd1$263 = 0, $cmd2$0 = 0, $cmd2$1 = 0, $cmd2$161 = 0, $dwMemPos$0$ph130 = 0, $dwMemPos$0$ph78114 = 0, $dwMemPos$0$ph81106 = 0, $dwMemPos$097 = 0, $dwMemPos$097$lcssa221 = 0, $dwMemPos$097$lcssa222 = 0, $exitcond = 0, $i$0107 = 0, $i15$092 = 0, $i20$0$be = 0, $i20$0$ph121 = 0, $i20$0116 = 0;
 var $i20$0116$lcssa = 0, $i20$1 = 0, $i20$10 = 0, $i20$10$ = 0, $i20$10$$ = 0, $i20$2 = 0, $i20$3 = 0, $i20$4 = 0, $i20$5 = 0, $i20$6 = 0, $i20$8 = 0, $i20$8$ = 0, $iEnv$093 = 0, $iIns$0108 = 0, $iOrd$0131 = 0, $iPat$0123 = 0, $iSmp$0101 = 0, $m$0$ph119 = 0, $nInstruments$0$ph128 = 0, $nInstruments$0$ph76112 = 0;
 var $nInstruments$0$ph76112$lcssa232 = 0, $nPatterns$0$ph127 = 0, $nSamples$0$ph129 = 0, $nSamples$0$ph77113 = 0, $nSamples$0$ph80105 = 0, $nSamples$0$ph80105$lcssa229 = 0, $nSamples$0$ph80105$lcssa230 = 0, $note$0 = 0, $or$cond = 0, $or$cond11 = 0, $or$cond160 = 0, $or$cond44 = 0, $or$cond46 = 0, $or$cond48 = 0, $or$cond50 = 0, $or$cond53 = 0, $or$cond53122 = 0, $or$cond54 = 0, $or$cond55115 = 0, $or$cond58 = 0;
 var $or$cond58100 = 0, $or$cond59 = 0, $param1$0 = 0, $param1$164 = 0, $param2$0 = 0, $row$0$ph120 = 0, $samplesize$0 = 0, $storemerge43 = 0, $switch$split12D = 0, $switch$split2D = 0, $switch$split42D = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ($lpStream|0)==(0|0); //@line 104 "load_dbm.cpp"
 $1 = ($dwMemLength>>>0)<(133); //@line 104 "load_dbm.cpp"
 $or$cond = $0 | $1; //@line 104 "load_dbm.cpp"
 if ($or$cond) {
  $$0 = 0;
  return ($$0|0); //@line 367 "load_dbm.cpp"
 }
 $2 = ((($lpStream)) + 76|0); //@line 104 "load_dbm.cpp"
 $3 = HEAPU8[$2>>0]|(HEAPU8[$2+1>>0]<<8); //@line 104 "load_dbm.cpp"
 $4 = ($3<<16>>16)==(0); //@line 104 "load_dbm.cpp"
 if ($4) {
  $$0 = 0;
  return ($$0|0); //@line 367 "load_dbm.cpp"
 }
 $5 = HEAPU8[$lpStream>>0]|(HEAPU8[$lpStream+1>>0]<<8)|(HEAPU8[$lpStream+2>>0]<<16)|(HEAPU8[$lpStream+3>>0]<<24); //@line 105 "load_dbm.cpp"
 $6 = ($5|0)==(810369604); //@line 105 "load_dbm.cpp"
 if (!($6)) {
  $$0 = 0;
  return ($$0|0); //@line 367 "load_dbm.cpp"
 }
 $7 = ((($lpStream)) + 72|0); //@line 105 "load_dbm.cpp"
 $8 = HEAPU8[$7>>0]|(HEAPU8[$7+1>>0]<<8); //@line 105 "load_dbm.cpp"
 $9 = ($8<<16>>16)==(0); //@line 105 "load_dbm.cpp"
 if ($9) {
  $$0 = 0;
  return ($$0|0); //@line 367 "load_dbm.cpp"
 }
 $10 = ((($lpStream)) + 78|0); //@line 105 "load_dbm.cpp"
 $11 = HEAPU8[$10>>0]|(HEAPU8[$10+1>>0]<<8)|(HEAPU8[$10+2>>0]<<16)|(HEAPU8[$10+3>>0]<<24); //@line 105 "load_dbm.cpp"
 $12 = ($11|0)==(1196314451); //@line 105 "load_dbm.cpp"
 if (!($12)) {
  $$0 = 0;
  return ($$0|0); //@line 367 "load_dbm.cpp"
 }
 $13 = ((($lpStream)) + 8|0); //@line 106 "load_dbm.cpp"
 $14 = HEAPU8[$13>>0]|(HEAPU8[$13+1>>0]<<8)|(HEAPU8[$13+2>>0]<<16)|(HEAPU8[$13+3>>0]<<24); //@line 106 "load_dbm.cpp"
 $15 = ($14|0)==(1162690894); //@line 106 "load_dbm.cpp"
 if (!($15)) {
  $$0 = 0;
  return ($$0|0); //@line 367 "load_dbm.cpp"
 }
 $16 = ((($lpStream)) + 12|0); //@line 106 "load_dbm.cpp"
 $17 = HEAPU8[$16>>0]|(HEAPU8[$16+1>>0]<<8)|(HEAPU8[$16+2>>0]<<16)|(HEAPU8[$16+3>>0]<<24); //@line 106 "load_dbm.cpp"
 $18 = ($17|0)==(738197504); //@line 106 "load_dbm.cpp"
 if (!($18)) {
  $$0 = 0;
  return ($$0|0); //@line 367 "load_dbm.cpp"
 }
 $19 = ((($lpStream)) + 60|0); //@line 107 "load_dbm.cpp"
 $20 = HEAPU8[$19>>0]|(HEAPU8[$19+1>>0]<<8)|(HEAPU8[$19+2>>0]<<16)|(HEAPU8[$19+3>>0]<<24); //@line 107 "load_dbm.cpp"
 $21 = ($20|0)==(1330007625); //@line 107 "load_dbm.cpp"
 if (!($21)) {
  $$0 = 0;
  return ($$0|0); //@line 367 "load_dbm.cpp"
 }
 $22 = ((($lpStream)) + 64|0); //@line 107 "load_dbm.cpp"
 $23 = HEAPU8[$22>>0]|(HEAPU8[$22+1>>0]<<8)|(HEAPU8[$22+2>>0]<<16)|(HEAPU8[$22+3>>0]<<24); //@line 107 "load_dbm.cpp"
 $24 = ($23|0)==(167772160); //@line 107 "load_dbm.cpp"
 if (!($24)) {
  $$0 = 0;
  return ($$0|0); //@line 367 "load_dbm.cpp"
 }
 $25 = ((($lpStream)) + 130|0); //@line 109 "load_dbm.cpp"
 $26 = HEAPU8[$25>>0]|(HEAPU8[$25+1>>0]<<8); //@line 109 "load_dbm.cpp"
 $27 = $26&65535; //@line 109 "load_dbm.cpp"
 $28 = $27 >>> 8; //@line 109 "load_dbm.cpp"
 $29 = $27 << 8; //@line 109 "load_dbm.cpp"
 $30 = $29 & 65280; //@line 109 "load_dbm.cpp"
 $31 = $30 | $28; //@line 109 "load_dbm.cpp"
 $32 = $31 << 1; //@line 110 "load_dbm.cpp"
 $33 = (($32) + 132)|0; //@line 110 "load_dbm.cpp"
 $34 = (($32) + 156)|0; //@line 110 "load_dbm.cpp"
 $35 = ($34>>>0)<($dwMemLength>>>0); //@line 110 "load_dbm.cpp"
 if (!($35)) {
  $$0 = 0;
  return ($$0|0); //@line 367 "load_dbm.cpp"
 }
 $36 = ((($lpStream)) + 68|0); //@line 111 "load_dbm.cpp"
 $37 = HEAPU8[$36>>0]|(HEAPU8[$36+1>>0]<<8); //@line 111 "load_dbm.cpp"
 $38 = $37&65535; //@line 111 "load_dbm.cpp"
 $39 = $38 >>> 8; //@line 111 "load_dbm.cpp"
 $40 = $38 << 8; //@line 111 "load_dbm.cpp"
 $41 = $40 & 65280; //@line 111 "load_dbm.cpp"
 $42 = $41 | $39; //@line 111 "load_dbm.cpp"
 $43 = ((($lpStream)) + 70|0); //@line 112 "load_dbm.cpp"
 $44 = HEAPU8[$43>>0]|(HEAPU8[$43+1>>0]<<8); //@line 112 "load_dbm.cpp"
 $45 = $44&65535; //@line 112 "load_dbm.cpp"
 $46 = $45 >>> 8; //@line 112 "load_dbm.cpp"
 $47 = $45 << 8; //@line 112 "load_dbm.cpp"
 $48 = $47 & 65280; //@line 112 "load_dbm.cpp"
 $49 = $48 | $46; //@line 112 "load_dbm.cpp"
 $50 = ((($lpStream)) + 74|0); //@line 113 "load_dbm.cpp"
 $51 = HEAPU8[$50>>0]|(HEAPU8[$50+1>>0]<<8); //@line 113 "load_dbm.cpp"
 $52 = $51&65535; //@line 113 "load_dbm.cpp"
 $53 = $52 >>> 8; //@line 113 "load_dbm.cpp"
 $54 = $52 << 8; //@line 113 "load_dbm.cpp"
 $55 = $54 & 65280; //@line 113 "load_dbm.cpp"
 $56 = $55 | $53; //@line 113 "load_dbm.cpp"
 $57 = ((($this)) + 59680|0); //@line 114 "load_dbm.cpp"
 HEAP32[$57>>2] = 524288; //@line 114 "load_dbm.cpp"
 $58 = $3&65535; //@line 115 "load_dbm.cpp"
 $59 = $58 >>> 8; //@line 115 "load_dbm.cpp"
 $60 = $58 << 8; //@line 115 "load_dbm.cpp"
 $61 = $60 & 65280; //@line 115 "load_dbm.cpp"
 $62 = $61 | $59; //@line 115 "load_dbm.cpp"
 $63 = ((($this)) + 59664|0); //@line 115 "load_dbm.cpp"
 $64 = ($62>>>0)<(4); //@line 116 "load_dbm.cpp"
 $$ = $64 ? 4 : $62; //@line 116 "load_dbm.cpp"
 $65 = ($$>>>0)>(64); //@line 117 "load_dbm.cpp"
 $storemerge43 = $65 ? 64 : $$; //@line 117 "load_dbm.cpp"
 HEAP32[$63>>2] = $storemerge43; //@line 115 "load_dbm.cpp"
 $66 = ((($this)) + 59804|0); //@line 118 "load_dbm.cpp"
 $67 = ((($lpStream)) + 16|0); //@line 118 "load_dbm.cpp"
 $68 = HEAP8[$67>>0]|0; //@line 118 "load_dbm.cpp"
 $69 = ($68<<24>>24)!=(0); //@line 118 "load_dbm.cpp"
 $70 = ((($lpStream)) + 86|0); //@line 118 "load_dbm.cpp"
 $$v = $69 ? $67 : $70; //@line 118 "load_dbm.cpp"
 dest=$66; src=$$v; stop=dest+31|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 118 "load_dbm.cpp"
 $71 = ((($this)) + 59835|0); //@line 119 "load_dbm.cpp"
 HEAP8[$71>>0] = 0; //@line 119 "load_dbm.cpp"
 $72 = ($31|0)==(0); //@line 120 "load_dbm.cpp"
 if (!($72)) {
  $iOrd$0131 = 0;
  while(1) {
   $77 = $iOrd$0131 << 1; //@line 122 "load_dbm.cpp"
   $78 = (($77) + 133)|0; //@line 122 "load_dbm.cpp"
   $79 = (($lpStream) + ($78)|0); //@line 122 "load_dbm.cpp"
   $80 = HEAP8[$79>>0]|0; //@line 122 "load_dbm.cpp"
   $81 = (((($this)) + 53344|0) + ($iOrd$0131)|0); //@line 122 "load_dbm.cpp"
   HEAP8[$81>>0] = $80; //@line 122 "load_dbm.cpp"
   $82 = ($iOrd$0131>>>0)<(254); //@line 123 "load_dbm.cpp"
   $83 = (($iOrd$0131) + 1)|0; //@line 120 "load_dbm.cpp"
   $84 = ($83>>>0)<($31>>>0); //@line 120 "load_dbm.cpp"
   $or$cond160 = $82 & $84; //@line 123 "load_dbm.cpp"
   if ($or$cond160) {
    $iOrd$0131 = $83;
   } else {
    break;
   }
  }
 }
 $73 = (($32) + 142)|0; //@line 126 "load_dbm.cpp"
 $74 = ($73>>>0)<($dwMemLength>>>0); //@line 126 "load_dbm.cpp"
 if (!($74)) {
  $$0 = 1;
  return ($$0|0); //@line 367 "load_dbm.cpp"
 }
 $75 = ((($this)) + 59684|0); //@line 336 "load_dbm.cpp"
 $76 = ((($this)) + 59688|0); //@line 194 "load_dbm.cpp"
 $423 = $73;$dwMemPos$0$ph130 = $33;$nInstruments$0$ph128 = $42;$nPatterns$0$ph127 = $56;$nSamples$0$ph129 = $49;
 L38: while(1) {
  $424 = $423;$dwMemPos$0$ph78114 = $dwMemPos$0$ph130;$nInstruments$0$ph76112 = $nInstruments$0$ph128;$nSamples$0$ph77113 = $nSamples$0$ph129;
  L40: while(1) {
   $425 = $424;$dwMemPos$0$ph81106 = $dwMemPos$0$ph78114;$nSamples$0$ph80105 = $nSamples$0$ph77113;
   L42: while(1) {
    $426 = $425;$dwMemPos$097 = $dwMemPos$0$ph81106;
    L44: while(1) {
     $$sum = (($dwMemPos$097) + 4)|0; //@line 129 "load_dbm.cpp"
     $91 = (($lpStream) + ($$sum)|0); //@line 129 "load_dbm.cpp"
     $92 = HEAP32[$91>>2]|0; //@line 129 "load_dbm.cpp"
     $93 = (_llvm_bswap_i32(($92|0))|0); //@line 129 "load_dbm.cpp"
     $94 = (($dwMemPos$097) + 8)|0; //@line 132 "load_dbm.cpp"
     $95 = (($93) + ($94))|0; //@line 134 "load_dbm.cpp"
     $96 = ($95>>>0)>($dwMemLength>>>0); //@line 134 "load_dbm.cpp"
     $97 = ($93>>>0)>($dwMemLength>>>0); //@line 134 "load_dbm.cpp"
     $or$cond44 = $97 | $96; //@line 134 "load_dbm.cpp"
     if ($or$cond44) {
      $$0 = 1;
      label = 96;
      break L38;
     }
     $98 = (($lpStream) + ($dwMemPos$097)|0); //@line 128 "load_dbm.cpp"
     $99 = HEAP32[$98>>2]|0; //@line 128 "load_dbm.cpp"
     $switch$split2D = ($99|0)<(1414807888);
     L47: do {
      if ($switch$split2D) {
       $switch$split12D = ($99|0)<(1414745673);
       if ($switch$split12D) {
        switch ($99|0) {
        case 1280331091:  {
         $$lcssa225 = $94;$$lcssa228 = $95;$dwMemPos$097$lcssa222 = $dwMemPos$097;
         break L44;
         break;
        }
        default: {
         break L47;
        }
        }
       } else {
        switch ($99|0) {
        case 1414745673:  {
         $$lcssa223 = $94;$$lcssa226 = $95;$nSamples$0$ph80105$lcssa229 = $nSamples$0$ph80105;
         break L42;
         break;
        }
        default: {
         break L47;
        }
        }
       }
      } else {
       $switch$split42D = ($99|0)<(1447970134);
       if ($switch$split42D) {
        switch ($99|0) {
        case 1414807888:  {
         $$lcssa224 = $94;$$lcssa227 = $95;$dwMemPos$097$lcssa221 = $dwMemPos$097;$nInstruments$0$ph76112$lcssa232 = $nInstruments$0$ph76112;$nSamples$0$ph80105$lcssa230 = $nSamples$0$ph80105;
         break L40;
         break;
        }
        default: {
         break L47;
        }
        }
       }
       switch ($99|0) {
       case 1447970134:  {
        break;
       }
       default: {
        break L47;
       }
       }
       $193 = (($dwMemPos$097) + 9)|0; //@line 200 "load_dbm.cpp"
       $194 = (($lpStream) + ($193)|0); //@line 200 "load_dbm.cpp"
       $195 = HEAP8[$194>>0]|0; //@line 200 "load_dbm.cpp"
       $196 = $195&255; //@line 200 "load_dbm.cpp"
       $197 = ($195<<24>>24)==(0); //@line 203 "load_dbm.cpp"
       if ($197) {
        break;
       } else {
        $chunk_pos$194 = $426;$iEnv$093 = 0;
       }
       while(1) {
        $198 = (($chunk_pos$194) + 136)|0; //@line 208 "load_dbm.cpp"
        $199 = ($198>>>0)>($95>>>0); //@line 208 "load_dbm.cpp"
        if ($199) {
         break L47;
        }
        $200 = (($lpStream) + ($chunk_pos$194)|0); //@line 209 "load_dbm.cpp"
        $201 = HEAPU8[$200>>0]|(HEAPU8[$200+1>>0]<<8); //@line 210 "load_dbm.cpp"
        $202 = $201&65535; //@line 210 "load_dbm.cpp"
        $203 = $202 >>> 8; //@line 210 "load_dbm.cpp"
        $204 = $202 << 8; //@line 210 "load_dbm.cpp"
        $205 = $204 & 65280; //@line 210 "load_dbm.cpp"
        $206 = $205 | $203; //@line 210 "load_dbm.cpp"
        $$off25 = (($206) + -1)|0; //@line 211 "load_dbm.cpp"
        $207 = ($$off25>>>0)<(239); //@line 211 "load_dbm.cpp"
        do {
         if ($207) {
          $208 = (((($this)) + 48640|0) + ($206<<2)|0); //@line 211 "load_dbm.cpp"
          $209 = HEAP32[$208>>2]|0; //@line 211 "load_dbm.cpp"
          $210 = ($209|0)==(0|0); //@line 211 "load_dbm.cpp"
          if ($210) {
           break;
          }
          $$sum26 = (($chunk_pos$194) + 3)|0; //@line 211 "load_dbm.cpp"
          $211 = (($lpStream) + ($$sum26)|0); //@line 211 "load_dbm.cpp"
          $212 = HEAP8[$211>>0]|0; //@line 211 "load_dbm.cpp"
          $213 = ($212<<24>>24)==(0); //@line 211 "load_dbm.cpp"
          if ($213) {
           break;
          }
          $$sum27 = (($chunk_pos$194) + 2)|0; //@line 215 "load_dbm.cpp"
          $214 = (($lpStream) + ($$sum27)|0); //@line 215 "load_dbm.cpp"
          $215 = HEAP8[$214>>0]|0; //@line 215 "load_dbm.cpp"
          $216 = $215 & 1; //@line 215 "load_dbm.cpp"
          $217 = ($216<<24>>24)==(0); //@line 215 "load_dbm.cpp"
          if (!($217)) {
           $218 = ((($209)) + 4|0); //@line 215 "load_dbm.cpp"
           $219 = HEAP32[$218>>2]|0; //@line 215 "load_dbm.cpp"
           $220 = $219 | 1; //@line 215 "load_dbm.cpp"
           HEAP32[$218>>2] = $220; //@line 215 "load_dbm.cpp"
          }
          $221 = $215 & 2; //@line 216 "load_dbm.cpp"
          $222 = ($221<<24>>24)==(0); //@line 216 "load_dbm.cpp"
          if (!($222)) {
           $223 = ((($209)) + 4|0); //@line 216 "load_dbm.cpp"
           $224 = HEAP32[$223>>2]|0; //@line 216 "load_dbm.cpp"
           $225 = $224 | 2; //@line 216 "load_dbm.cpp"
           HEAP32[$223>>2] = $225; //@line 216 "load_dbm.cpp"
          }
          $226 = $215 & 4; //@line 217 "load_dbm.cpp"
          $227 = ($226<<24>>24)==(0); //@line 217 "load_dbm.cpp"
          if (!($227)) {
           $228 = ((($209)) + 4|0); //@line 217 "load_dbm.cpp"
           $229 = HEAP32[$228>>2]|0; //@line 217 "load_dbm.cpp"
           $230 = $229 | 4; //@line 217 "load_dbm.cpp"
           HEAP32[$228>>2] = $230; //@line 217 "load_dbm.cpp"
          }
          $231 = $212&255; //@line 218 "load_dbm.cpp"
          $232 = (($231) + 1)|0; //@line 218 "load_dbm.cpp"
          $233 = $232&255; //@line 218 "load_dbm.cpp"
          $234 = ((($209)) + 556|0); //@line 218 "load_dbm.cpp"
          $235 = $232 & 255; //@line 219 "load_dbm.cpp"
          $236 = ($235>>>0)>(32); //@line 219 "load_dbm.cpp"
          $$51 = $236 ? 32 : $233; //@line 219 "load_dbm.cpp"
          HEAP8[$234>>0] = $$51; //@line 218 "load_dbm.cpp"
          $$sum28 = (($chunk_pos$194) + 5)|0; //@line 220 "load_dbm.cpp"
          $237 = (($lpStream) + ($$sum28)|0); //@line 220 "load_dbm.cpp"
          $238 = HEAP8[$237>>0]|0; //@line 220 "load_dbm.cpp"
          $239 = ((($209)) + 559|0); //@line 220 "load_dbm.cpp"
          HEAP8[$239>>0] = $238; //@line 220 "load_dbm.cpp"
          $$sum29 = (($chunk_pos$194) + 6)|0; //@line 221 "load_dbm.cpp"
          $240 = (($lpStream) + ($$sum29)|0); //@line 221 "load_dbm.cpp"
          $241 = HEAP8[$240>>0]|0; //@line 221 "load_dbm.cpp"
          $242 = ((($209)) + 560|0); //@line 221 "load_dbm.cpp"
          HEAP8[$242>>0] = $241; //@line 221 "load_dbm.cpp"
          $$sum30 = (($chunk_pos$194) + 4)|0; //@line 222 "load_dbm.cpp"
          $243 = (($lpStream) + ($$sum30)|0); //@line 222 "load_dbm.cpp"
          $244 = HEAP8[$243>>0]|0; //@line 222 "load_dbm.cpp"
          $245 = ((($209)) + 562|0); //@line 222 "load_dbm.cpp"
          HEAP8[$245>>0] = $244; //@line 222 "load_dbm.cpp"
          $246 = ((($209)) + 561|0); //@line 222 "load_dbm.cpp"
          HEAP8[$246>>0] = $244; //@line 222 "load_dbm.cpp"
          $247 = ($$51<<24>>24)==(0); //@line 223 "load_dbm.cpp"
          if ($247) {
           break;
          }
          $$sum31 = (($chunk_pos$194) + 8)|0; //@line 225 "load_dbm.cpp"
          $248 = (($lpStream) + ($$sum31)|0); //@line 225 "load_dbm.cpp"
          $i15$092 = 0;
          while(1) {
           $249 = $i15$092 << 1; //@line 225 "load_dbm.cpp"
           $250 = (($248) + ($249<<1)|0); //@line 225 "load_dbm.cpp"
           $251 = HEAPU8[$250>>0]|(HEAPU8[$250+1>>0]<<8); //@line 225 "load_dbm.cpp"
           $252 = $251&65535; //@line 225 "load_dbm.cpp"
           $253 = $252 >>> 8; //@line 225 "load_dbm.cpp"
           $254 = $252 << 8; //@line 225 "load_dbm.cpp"
           $255 = $253 | $254; //@line 225 "load_dbm.cpp"
           $256 = $255&65535; //@line 225 "load_dbm.cpp"
           $257 = (((($209)) + 12|0) + ($i15$092<<1)|0); //@line 225 "load_dbm.cpp"
           HEAP16[$257>>1] = $256; //@line 225 "load_dbm.cpp"
           $258 = $249 | 1; //@line 226 "load_dbm.cpp"
           $259 = (($248) + ($258<<1)|0); //@line 226 "load_dbm.cpp"
           $260 = HEAPU8[$259>>0]|(HEAPU8[$259+1>>0]<<8); //@line 226 "load_dbm.cpp"
           $261 = ($260&65535) >>> 8; //@line 226 "load_dbm.cpp"
           $262 = $261&255; //@line 226 "load_dbm.cpp"
           $263 = (((($209)) + 204|0) + ($i15$092)|0); //@line 226 "load_dbm.cpp"
           HEAP8[$263>>0] = $262; //@line 226 "load_dbm.cpp"
           $264 = (($i15$092) + 1)|0; //@line 223 "load_dbm.cpp"
           $265 = HEAP8[$234>>0]|0; //@line 223 "load_dbm.cpp"
           $266 = $265&255; //@line 223 "load_dbm.cpp"
           $267 = ($264>>>0)<($266>>>0); //@line 223 "load_dbm.cpp"
           if ($267) {
            $i15$092 = $264;
           } else {
            break;
           }
          }
         }
        } while(0);
        $268 = (($iEnv$093) + 1)|0; //@line 203 "load_dbm.cpp"
        $269 = ($268>>>0)<($196>>>0); //@line 203 "load_dbm.cpp"
        if ($269) {
         $chunk_pos$194 = $198;$iEnv$093 = $268;
        } else {
         break;
        }
       }
      }
     } while(0);
     $100 = (($95) + 10)|0; //@line 126 "load_dbm.cpp"
     $101 = ($100>>>0)<($dwMemLength>>>0); //@line 126 "load_dbm.cpp"
     if ($101) {
      $426 = $100;$dwMemPos$097 = $95;
     } else {
      $$0 = 1;
      label = 96;
      break L38;
     }
    }
    $389 = ($nSamples$0$ph80105>>>0)>(239); //@line 335 "load_dbm.cpp"
    $$nSamples$0 = $389 ? 239 : $nSamples$0$ph80105; //@line 335 "load_dbm.cpp"
    HEAP32[$75>>2] = $$nSamples$0; //@line 336 "load_dbm.cpp"
    $390 = ($$nSamples$0|0)!=(0); //@line 337 "load_dbm.cpp"
    $391 = (($dwMemPos$097$lcssa222) + 18)|0;
    $392 = ($391>>>0)<($$lcssa228>>>0); //@line 344 "load_dbm.cpp"
    $or$cond58100 = $390 & $392; //@line 337 "load_dbm.cpp"
    L79: do {
     if ($or$cond58100) {
      $chunk_pos$3102 = $$lcssa225;$iSmp$0101 = 1;
      while(1) {
       $393 = (($lpStream) + ($chunk_pos$3102)|0); //@line 345 "load_dbm.cpp"
       $394 = (($chunk_pos$3102) + 8)|0; //@line 346 "load_dbm.cpp"
       $$sum12 = (($chunk_pos$3102) + 4)|0; //@line 347 "load_dbm.cpp"
       $395 = (($lpStream) + ($$sum12)|0); //@line 347 "load_dbm.cpp"
       $396 = HEAPU8[$395>>0]|(HEAPU8[$395+1>>0]<<8)|(HEAPU8[$395+2>>0]<<16)|(HEAPU8[$395+3>>0]<<24); //@line 347 "load_dbm.cpp"
       $397 = (_llvm_bswap_i32(($396|0))|0); //@line 347 "load_dbm.cpp"
       $398 = HEAPU8[$393>>0]|(HEAPU8[$393+1>>0]<<8)|(HEAPU8[$393+2>>0]<<16)|(HEAPU8[$393+3>>0]<<24); //@line 348 "load_dbm.cpp"
       $399 = (((($this)) + 33280|0) + ($iSmp$0101<<6)|0); //@line 349 "load_dbm.cpp"
       HEAP32[$399>>2] = $397; //@line 350 "load_dbm.cpp"
       $400 = $398 & 33554432; //@line 351 "load_dbm.cpp"
       $401 = ($400|0)==(0); //@line 351 "load_dbm.cpp"
       if ($401) {
        $samplesize$0 = $397;
       } else {
        $402 = (((((($this)) + 33280|0) + ($iSmp$0101<<6)|0)) + 34|0); //@line 353 "load_dbm.cpp"
        $403 = HEAP16[$402>>1]|0; //@line 353 "load_dbm.cpp"
        $404 = $403&65535; //@line 353 "load_dbm.cpp"
        $405 = $404 | 1; //@line 353 "load_dbm.cpp"
        $406 = $405&65535; //@line 353 "load_dbm.cpp"
        HEAP16[$402>>1] = $406; //@line 353 "load_dbm.cpp"
        $407 = $397 << 1; //@line 354 "load_dbm.cpp"
        $samplesize$0 = $407;
       }
       $408 = (($samplesize$0) + ($394))|0; //@line 356 "load_dbm.cpp"
       $409 = ($408>>>0)>($$lcssa228>>>0); //@line 356 "load_dbm.cpp"
       $410 = ($samplesize$0>>>0)>($dwMemLength>>>0); //@line 356 "load_dbm.cpp"
       $or$cond59 = $410 | $409; //@line 356 "load_dbm.cpp"
       if ($or$cond59) {
        break L79;
       }
       $411 = $398 & 50331648; //@line 357 "load_dbm.cpp"
       $412 = ($411|0)==(0); //@line 357 "load_dbm.cpp"
       if (!($412)) {
        $413 = (((((($this)) + 33280|0) + ($iSmp$0101<<6)|0)) + 34|0); //@line 359 "load_dbm.cpp"
        $414 = HEAP16[$413>>1]|0; //@line 359 "load_dbm.cpp"
        $415 = $414 & 1; //@line 359 "load_dbm.cpp"
        $416 = ($415<<16>>16)!=(0); //@line 359 "load_dbm.cpp"
        $417 = $416 ? 7 : 0; //@line 359 "load_dbm.cpp"
        $418 = (($lpStream) + ($394)|0); //@line 360 "load_dbm.cpp"
        (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$399,$417,$418,$samplesize$0)|0); //@line 359 "load_dbm.cpp"
       }
       $419 = (($iSmp$0101) + 1)|0; //@line 337 "load_dbm.cpp"
       $420 = ($419>>>0)<=($$nSamples$0>>>0); //@line 337 "load_dbm.cpp"
       $421 = (($408) + 10)|0;
       $422 = ($421>>>0)<($$lcssa228>>>0); //@line 344 "load_dbm.cpp"
       $or$cond58 = $420 & $422; //@line 337 "load_dbm.cpp"
       if ($or$cond58) {
        $chunk_pos$3102 = $408;$iSmp$0101 = $419;
       } else {
        break;
       }
      }
     }
    } while(0);
    $89 = (($$lcssa228) + 10)|0; //@line 126 "load_dbm.cpp"
    $90 = ($89>>>0)<($dwMemLength>>>0); //@line 126 "load_dbm.cpp"
    if ($90) {
     $425 = $89;$dwMemPos$0$ph81106 = $$lcssa228;$nSamples$0$ph80105 = $$nSamples$0;
    } else {
     $$0 = 1;
     label = 96;
     break L38;
    }
   }
   $102 = ($nInstruments$0$ph76112>>>0)>(239); //@line 139 "load_dbm.cpp"
   $$nInstruments$0 = $102 ? 239 : $nInstruments$0$ph76112; //@line 139 "load_dbm.cpp"
   $103 = ($$nInstruments$0|0)==(0); //@line 140 "load_dbm.cpp"
   L91: do {
    if (!($103)) {
     $chunk_pos$0109 = $$lcssa223;$iIns$0108 = 0;
     while(1) {
      $104 = (($chunk_pos$0109) + 50)|0; //@line 147 "load_dbm.cpp"
      $105 = ($104>>>0)>($$lcssa226>>>0); //@line 147 "load_dbm.cpp"
      if ($105) {
       break L91;
      }
      $106 = (__Znwj(632)|0); //@line 148 "load_dbm.cpp"
      $107 = (($lpStream) + ($chunk_pos$0109)|0); //@line 149 "load_dbm.cpp"
      $$sum32 = (($chunk_pos$0109) + 30)|0; //@line 150 "load_dbm.cpp"
      $108 = (($lpStream) + ($$sum32)|0); //@line 150 "load_dbm.cpp"
      $109 = HEAPU8[$108>>0]|(HEAPU8[$108+1>>0]<<8); //@line 150 "load_dbm.cpp"
      $110 = $109&65535; //@line 150 "load_dbm.cpp"
      $111 = $110 >>> 8; //@line 150 "load_dbm.cpp"
      $112 = $110 << 8; //@line 150 "load_dbm.cpp"
      $113 = $112 & 65280; //@line 150 "load_dbm.cpp"
      $114 = $113 | $111; //@line 150 "load_dbm.cpp"
      $$off33 = (($114) + -1)|0; //@line 151 "load_dbm.cpp"
      $115 = ($$off33>>>0)<(239); //@line 151 "load_dbm.cpp"
      $116 = (((($this)) + 33280|0) + ($114<<6)|0); //@line 151 "load_dbm.cpp"
      $117 = $115 ? $116 : 0; //@line 151 "load_dbm.cpp"
      _memset(($106|0),0,632)|0; //@line 152 "load_dbm.cpp"
      $118 = ((($106)) + 585|0); //@line 153 "load_dbm.cpp"
      dest=$118; src=$107; stop=dest+30|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 153 "load_dbm.cpp"
      if ($115) {
       $119 = (((($this)) + 59804|0) + ($114<<5)|0); //@line 156 "load_dbm.cpp"
       dest=$119; src=$107; stop=dest+30|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 156 "load_dbm.cpp"
       $120 = (((((($this)) + 59804|0) + ($114<<5)|0)) + 30|0); //@line 157 "load_dbm.cpp"
       HEAP8[$120>>0] = 0; //@line 157 "load_dbm.cpp"
      }
      $121 = (($iIns$0108) + 1)|0; //@line 159 "load_dbm.cpp"
      $122 = (((($this)) + 48640|0) + ($121<<2)|0); //@line 159 "load_dbm.cpp"
      HEAP32[$122>>2] = $106; //@line 159 "load_dbm.cpp"
      HEAP32[$106>>2] = 1024; //@line 160 "load_dbm.cpp"
      $123 = ((($106)) + 8|0); //@line 161 "load_dbm.cpp"
      HEAP16[$123>>1] = 64; //@line 161 "load_dbm.cpp"
      $$sum34 = (($chunk_pos$0109) + 46)|0; //@line 162 "load_dbm.cpp"
      $124 = (($lpStream) + ($$sum34)|0); //@line 162 "load_dbm.cpp"
      $125 = HEAPU8[$124>>0]|(HEAPU8[$124+1>>0]<<8); //@line 162 "load_dbm.cpp"
      $126 = $125&65535; //@line 162 "load_dbm.cpp"
      $127 = $126 >>> 8; //@line 162 "load_dbm.cpp"
      $128 = $126 << 8; //@line 162 "load_dbm.cpp"
      $129 = $127 | $128; //@line 162 "load_dbm.cpp"
      $130 = $129&65535; //@line 162 "load_dbm.cpp"
      $131 = ((($106)) + 10|0); //@line 162 "load_dbm.cpp"
      HEAP16[$131>>1] = $130; //@line 162 "load_dbm.cpp"
      $132 = ($130<<16>>16)!=(0); //@line 163 "load_dbm.cpp"
      $133 = $128 & 65280;
      $134 = ($133>>>0)<(256); //@line 163 "load_dbm.cpp"
      $or$cond46 = $134 & $132; //@line 163 "load_dbm.cpp"
      if ($or$cond46) {
       $135 = ((($106)) + 4|0); //@line 164 "load_dbm.cpp"
       HEAP32[$135>>2] = 512; //@line 164 "load_dbm.cpp"
      } else {
       HEAP16[$131>>1] = 128; //@line 166 "load_dbm.cpp"
      }
      $136 = ((($106)) + 584|0); //@line 167 "load_dbm.cpp"
      HEAP8[$136>>0] = 60; //@line 167 "load_dbm.cpp"
      $137 = $111&255; //@line 170 "load_dbm.cpp"
      $i$0107 = 0;
      while(1) {
       $$sum40 = (($i$0107) + 300)|0; //@line 170 "load_dbm.cpp"
       $138 = (($106) + ($$sum40)|0); //@line 170 "load_dbm.cpp"
       HEAP8[$138>>0] = $137; //@line 170 "load_dbm.cpp"
       $139 = (($i$0107) + 1)|0; //@line 171 "load_dbm.cpp"
       $140 = $139&255; //@line 171 "load_dbm.cpp"
       $$sum41 = (($i$0107) + 428)|0; //@line 171 "load_dbm.cpp"
       $141 = (($106) + ($$sum41)|0); //@line 171 "load_dbm.cpp"
       HEAP8[$141>>0] = $140; //@line 171 "load_dbm.cpp"
       $exitcond = ($139|0)==(120); //@line 168 "load_dbm.cpp"
       if ($exitcond) {
        break;
       } else {
        $i$0107 = $139;
       }
      }
      do {
       if ($115) {
        $$sum35 = (($chunk_pos$0109) + 48)|0; //@line 176 "load_dbm.cpp"
        $142 = (($lpStream) + ($$sum35)|0); //@line 176 "load_dbm.cpp"
        $143 = HEAPU8[$142>>0]|(HEAPU8[$142+1>>0]<<8); //@line 176 "load_dbm.cpp"
        $144 = $143&65535; //@line 176 "load_dbm.cpp"
        $145 = $144 >>> 8; //@line 176 "load_dbm.cpp"
        $$sum36 = (($chunk_pos$0109) + 32)|0; //@line 177 "load_dbm.cpp"
        $146 = (($lpStream) + ($$sum36)|0); //@line 177 "load_dbm.cpp"
        $147 = HEAPU8[$146>>0]|(HEAPU8[$146+1>>0]<<8); //@line 177 "load_dbm.cpp"
        $148 = $147&65535; //@line 177 "load_dbm.cpp"
        $149 = $148 >>> 8; //@line 177 "load_dbm.cpp"
        $150 = $148 << 8; //@line 177 "load_dbm.cpp"
        $151 = $149 | $150; //@line 177 "load_dbm.cpp"
        $152 = $151 << 2; //@line 177 "load_dbm.cpp"
        $153 = $152&65535; //@line 177 "load_dbm.cpp"
        $154 = ((($117)) + 30|0); //@line 177 "load_dbm.cpp"
        $155 = ($153<<16>>16)==(0); //@line 178 "load_dbm.cpp"
        $156 = $152 & 65532;
        $157 = ($156>>>0)>(256); //@line 178 "load_dbm.cpp"
        $or$cond48 = $155 | $157; //@line 178 "load_dbm.cpp"
        $$60 = $or$cond48 ? 256 : $153; //@line 178 "load_dbm.cpp"
        HEAP16[$154>>1] = $$60; //@line 178 "load_dbm.cpp"
        $158 = ((($117)) + 32|0); //@line 179 "load_dbm.cpp"
        HEAP16[$158>>1] = 64; //@line 179 "load_dbm.cpp"
        $$sum37 = (($chunk_pos$0109) + 34)|0; //@line 180 "load_dbm.cpp"
        $159 = (($lpStream) + ($$sum37)|0); //@line 180 "load_dbm.cpp"
        $160 = HEAPU8[$159>>0]|(HEAPU8[$159+1>>0]<<8)|(HEAPU8[$159+2>>0]<<16)|(HEAPU8[$159+3>>0]<<24); //@line 180 "load_dbm.cpp"
        $161 = (_llvm_bswap_i32(($160|0))|0); //@line 180 "load_dbm.cpp"
        $162 = ((($117)) + 24|0); //@line 180 "load_dbm.cpp"
        HEAP32[$162>>2] = $161; //@line 180 "load_dbm.cpp"
        $163 = (__ZN10CSoundFile20FrequencyToTransposeEj($161)|0); //@line 181 "load_dbm.cpp"
        $164 = $163 >>> 7; //@line 182 "load_dbm.cpp"
        $165 = $164&255; //@line 182 "load_dbm.cpp"
        $166 = ((($117)) + 36|0); //@line 182 "load_dbm.cpp"
        HEAP8[$166>>0] = $165; //@line 182 "load_dbm.cpp"
        $167 = $163 & 127; //@line 183 "load_dbm.cpp"
        $168 = $167&255; //@line 183 "load_dbm.cpp"
        $169 = ((($117)) + 37|0); //@line 183 "load_dbm.cpp"
        HEAP8[$169>>0] = $168; //@line 183 "load_dbm.cpp"
        $$sum38 = (($chunk_pos$0109) + 42)|0; //@line 184 "load_dbm.cpp"
        $170 = (($lpStream) + ($$sum38)|0); //@line 184 "load_dbm.cpp"
        $171 = HEAPU8[$170>>0]|(HEAPU8[$170+1>>0]<<8)|(HEAPU8[$170+2>>0]<<16)|(HEAPU8[$170+3>>0]<<24); //@line 184 "load_dbm.cpp"
        $172 = ($171|0)==(0); //@line 184 "load_dbm.cpp"
        $173 = $145 & 3;
        $174 = ($173|0)==(0); //@line 184 "load_dbm.cpp"
        $or$cond50 = $174 | $172; //@line 184 "load_dbm.cpp"
        if ($or$cond50) {
         break;
        }
        $$sum39 = (($chunk_pos$0109) + 38)|0; //@line 186 "load_dbm.cpp"
        $175 = (($lpStream) + ($$sum39)|0); //@line 186 "load_dbm.cpp"
        $176 = HEAPU8[$175>>0]|(HEAPU8[$175+1>>0]<<8)|(HEAPU8[$175+2>>0]<<16)|(HEAPU8[$175+3>>0]<<24); //@line 186 "load_dbm.cpp"
        $177 = (_llvm_bswap_i32(($176|0))|0); //@line 186 "load_dbm.cpp"
        $178 = ((($117)) + 4|0); //@line 186 "load_dbm.cpp"
        HEAP32[$178>>2] = $177; //@line 186 "load_dbm.cpp"
        $179 = (_llvm_bswap_i32(($171|0))|0); //@line 187 "load_dbm.cpp"
        $180 = (($179) + ($177))|0; //@line 187 "load_dbm.cpp"
        $181 = ((($117)) + 8|0); //@line 187 "load_dbm.cpp"
        HEAP32[$181>>2] = $180; //@line 187 "load_dbm.cpp"
        $182 = ((($117)) + 34|0); //@line 188 "load_dbm.cpp"
        $183 = HEAP16[$182>>1]|0; //@line 188 "load_dbm.cpp"
        $184 = $183&65535; //@line 188 "load_dbm.cpp"
        $185 = $184 & 65529; //@line 189 "load_dbm.cpp"
        $186 = $185 | 2; //@line 189 "load_dbm.cpp"
        $187 = $186&65535; //@line 189 "load_dbm.cpp"
        HEAP16[$182>>1] = $187; //@line 189 "load_dbm.cpp"
        $188 = $145 & 2; //@line 190 "load_dbm.cpp"
        $189 = ($188|0)==(0); //@line 190 "load_dbm.cpp"
        if ($189) {
         break;
        }
        $190 = $184 | 6; //@line 190 "load_dbm.cpp"
        $191 = $190&65535; //@line 190 "load_dbm.cpp"
        HEAP16[$182>>1] = $191; //@line 190 "load_dbm.cpp"
       }
      } while(0);
      HEAP32[$76>>2] = $121; //@line 194 "load_dbm.cpp"
      $192 = ($121>>>0)<($$nInstruments$0>>>0); //@line 140 "load_dbm.cpp"
      if ($192) {
       $chunk_pos$0109 = $104;$iIns$0108 = $121;
      } else {
       break;
      }
     }
    }
   } while(0);
   $85 = (($$lcssa226) + 10)|0; //@line 126 "load_dbm.cpp"
   $86 = ($85>>>0)<($dwMemLength>>>0); //@line 126 "load_dbm.cpp"
   if ($86) {
    $424 = $85;$dwMemPos$0$ph78114 = $$lcssa226;$nInstruments$0$ph76112 = $$nInstruments$0;$nSamples$0$ph77113 = $nSamples$0$ph80105$lcssa229;
   } else {
    $$0 = 1;
    label = 96;
    break L38;
   }
  }
  $270 = ($nPatterns$0$ph127>>>0)>(240); //@line 235 "load_dbm.cpp"
  $$nPatterns$0 = $270 ? 240 : $nPatterns$0$ph127; //@line 235 "load_dbm.cpp"
  $271 = ($$nPatterns$0|0)==(0); //@line 236 "load_dbm.cpp"
  $272 = (($dwMemPos$097$lcssa221) + 16)|0;
  $273 = ($272>>>0)>($$lcssa227>>>0); //@line 242 "load_dbm.cpp"
  $or$cond53122 = $271 | $273; //@line 236 "load_dbm.cpp"
  L112: do {
   if (!($or$cond53122)) {
    $chunk_pos$2124 = $$lcssa224;$iPat$0123 = 0;
    while(1) {
     $$sum14 = (($chunk_pos$2124) + 2)|0; //@line 244 "load_dbm.cpp"
     $274 = (($lpStream) + ($$sum14)|0); //@line 244 "load_dbm.cpp"
     $275 = HEAPU8[$274>>0]|(HEAPU8[$274+1>>0]<<8)|(HEAPU8[$274+2>>0]<<16)|(HEAPU8[$274+3>>0]<<24); //@line 244 "load_dbm.cpp"
     $276 = (_llvm_bswap_i32(($275|0))|0); //@line 244 "load_dbm.cpp"
     $277 = (($chunk_pos$2124) + 6)|0; //@line 245 "load_dbm.cpp"
     $278 = (($277) + ($276))|0; //@line 245 "load_dbm.cpp"
     $279 = ($278>>>0)>($$lcssa227>>>0); //@line 245 "load_dbm.cpp"
     $280 = ($276>>>0)>($$lcssa227>>>0); //@line 245 "load_dbm.cpp"
     $or$cond54 = $280 | $279; //@line 245 "load_dbm.cpp"
     if ($or$cond54) {
      break L112;
     }
     $281 = (($lpStream) + ($chunk_pos$2124)|0); //@line 243 "load_dbm.cpp"
     $282 = HEAPU8[$281>>0]|(HEAPU8[$281+1>>0]<<8); //@line 246 "load_dbm.cpp"
     $283 = $282&65535; //@line 246 "load_dbm.cpp"
     $284 = $283 >>> 8; //@line 246 "load_dbm.cpp"
     $285 = $283 << 8; //@line 246 "load_dbm.cpp"
     $286 = $285 & 65280; //@line 246 "load_dbm.cpp"
     $287 = $286 | $284; //@line 246 "load_dbm.cpp"
     $$off = (($287) + -4)|0; //@line 247 "load_dbm.cpp"
     $288 = ($$off>>>0)<(253); //@line 247 "load_dbm.cpp"
     L116: do {
      if ($288) {
       $289 = HEAP32[$63>>2]|0; //@line 249 "load_dbm.cpp"
       $290 = (__ZN10CSoundFile15AllocatePatternEjj($287,$289)|0); //@line 249 "load_dbm.cpp"
       $291 = ($290|0)==(0|0); //@line 250 "load_dbm.cpp"
       if (!($291)) {
        $292 = $287&65535; //@line 256 "load_dbm.cpp"
        $293 = (((($this)) + 52864|0) + ($iPat$0123<<1)|0); //@line 256 "load_dbm.cpp"
        HEAP16[$293>>1] = $292; //@line 256 "load_dbm.cpp"
        $294 = (((($this)) + 51904|0) + ($iPat$0123<<2)|0); //@line 257 "load_dbm.cpp"
        HEAP32[$294>>2] = $290; //@line 257 "load_dbm.cpp"
        $295 = ($276>>>0)>(3); //@line 258 "load_dbm.cpp"
        if ($295) {
         $427 = 3;$i20$0$ph121 = 0;$m$0$ph119 = $290;$row$0$ph120 = 0;
        } else {
         break;
        }
        while(1) {
         $373 = $427;$i20$0116 = $i20$0$ph121;
         while(1) {
          $296 = (($i20$0116) + 1)|0; //@line 260 "load_dbm.cpp"
          $$sum16 = (($i20$0116) + ($277))|0; //@line 260 "load_dbm.cpp"
          $297 = (($lpStream) + ($$sum16)|0); //@line 260 "load_dbm.cpp"
          $298 = HEAP8[$297>>0]|0; //@line 260 "load_dbm.cpp"
          $299 = ($298<<24>>24)==(0); //@line 262 "load_dbm.cpp"
          if ($299) {
           $$lcssa = $296;$i20$0116$lcssa = $i20$0116;
           break;
          }
          $300 = $298&255; //@line 260 "load_dbm.cpp"
          $301 = (($i20$0116) + 2)|0; //@line 264 "load_dbm.cpp"
          $$sum17 = (($296) + ($277))|0; //@line 264 "load_dbm.cpp"
          $302 = (($lpStream) + ($$sum17)|0); //@line 264 "load_dbm.cpp"
          $303 = HEAP8[$302>>0]|0; //@line 264 "load_dbm.cpp"
          $304 = (($300) + -1)|0; //@line 265 "load_dbm.cpp"
          $305 = HEAP32[$63>>2]|0; //@line 266 "load_dbm.cpp"
          $306 = ($304>>>0)<($305>>>0); //@line 266 "load_dbm.cpp"
          $307 = $303&255; //@line 268 "load_dbm.cpp"
          $308 = $307 & 1; //@line 268 "load_dbm.cpp"
          $309 = ($308|0)!=(0); //@line 268 "load_dbm.cpp"
          do {
           if ($306) {
            if ($309) {
             $$sum23 = (($301) + ($277))|0; //@line 270 "load_dbm.cpp"
             $310 = (($lpStream) + ($$sum23)|0); //@line 270 "load_dbm.cpp"
             $311 = HEAP8[$310>>0]|0; //@line 270 "load_dbm.cpp"
             $312 = $311&255; //@line 270 "load_dbm.cpp"
             $313 = ($311<<24>>24)==(31); //@line 272 "load_dbm.cpp"
             do {
              if ($313) {
               $note$0 = 255;
              } else {
               $$off24 = (($311) + -1)<<24>>24; //@line 273 "load_dbm.cpp"
               $314 = ($$off24&255)<(253); //@line 273 "load_dbm.cpp"
               if (!($314)) {
                $note$0 = $312;
                break;
               }
               $315 = $312 >>> 4; //@line 275 "load_dbm.cpp"
               $316 = ($315*12)|0; //@line 275 "load_dbm.cpp"
               $317 = $312 & 15; //@line 275 "load_dbm.cpp"
               $318 = (($317) + 13)|0; //@line 275 "load_dbm.cpp"
               $319 = (($318) + ($316))|0; //@line 275 "load_dbm.cpp"
               $note$0 = $319;
              }
             } while(0);
             $320 = $note$0&255; //@line 277 "load_dbm.cpp"
             $321 = (($m$0$ph119) + (($304*6)|0)|0); //@line 277 "load_dbm.cpp"
             HEAP8[$321>>0] = $320; //@line 277 "load_dbm.cpp"
             $i20$1 = $373;
            } else {
             $i20$1 = $301;
            }
            $322 = $307 & 2; //@line 279 "load_dbm.cpp"
            $323 = ($322|0)==(0); //@line 279 "load_dbm.cpp"
            if ($323) {
             $i20$2 = $i20$1;
            } else {
             $324 = (($i20$1) + 1)|0; //@line 279 "load_dbm.cpp"
             $$sum22 = (($i20$1) + ($277))|0; //@line 279 "load_dbm.cpp"
             $325 = (($lpStream) + ($$sum22)|0); //@line 279 "load_dbm.cpp"
             $326 = HEAP8[$325>>0]|0; //@line 279 "load_dbm.cpp"
             $327 = (((($m$0$ph119) + (($304*6)|0)|0)) + 1|0); //@line 279 "load_dbm.cpp"
             HEAP8[$327>>0] = $326; //@line 279 "load_dbm.cpp"
             $i20$2 = $324;
            }
            $328 = $307 & 60; //@line 280 "load_dbm.cpp"
            $329 = ($328|0)==(0); //@line 280 "load_dbm.cpp"
            if ($329) {
             $i20$0$be = $i20$2;
             break;
            }
            $330 = $307 & 4; //@line 283 "load_dbm.cpp"
            $331 = ($330|0)==(0); //@line 283 "load_dbm.cpp"
            if ($331) {
             $cmd1$0 = 255;$i20$3 = $i20$2;
            } else {
             $332 = (($i20$2) + 1)|0; //@line 283 "load_dbm.cpp"
             $$sum21 = (($i20$2) + ($277))|0; //@line 283 "load_dbm.cpp"
             $333 = (($lpStream) + ($$sum21)|0); //@line 283 "load_dbm.cpp"
             $334 = HEAP8[$333>>0]|0; //@line 283 "load_dbm.cpp"
             $335 = $334&255; //@line 283 "load_dbm.cpp"
             $cmd1$0 = $335;$i20$3 = $332;
            }
            $336 = $307 & 8; //@line 284 "load_dbm.cpp"
            $337 = ($336|0)==(0); //@line 284 "load_dbm.cpp"
            if ($337) {
             $i20$4 = $i20$3;$param1$0 = 0;
            } else {
             $338 = (($i20$3) + 1)|0; //@line 284 "load_dbm.cpp"
             $$sum20 = (($i20$3) + ($277))|0; //@line 284 "load_dbm.cpp"
             $339 = (($lpStream) + ($$sum20)|0); //@line 284 "load_dbm.cpp"
             $340 = HEAP8[$339>>0]|0; //@line 284 "load_dbm.cpp"
             $341 = $340&255; //@line 284 "load_dbm.cpp"
             $i20$4 = $338;$param1$0 = $341;
            }
            $342 = $307 & 16; //@line 285 "load_dbm.cpp"
            $343 = ($342|0)==(0); //@line 285 "load_dbm.cpp"
            if ($343) {
             $cmd2$0 = 255;$i20$5 = $i20$4;
            } else {
             $344 = (($i20$4) + 1)|0; //@line 285 "load_dbm.cpp"
             $$sum19 = (($i20$4) + ($277))|0; //@line 285 "load_dbm.cpp"
             $345 = (($lpStream) + ($$sum19)|0); //@line 285 "load_dbm.cpp"
             $346 = HEAP8[$345>>0]|0; //@line 285 "load_dbm.cpp"
             $347 = $346&255; //@line 285 "load_dbm.cpp"
             $cmd2$0 = $347;$i20$5 = $344;
            }
            $348 = $307 & 32; //@line 286 "load_dbm.cpp"
            $349 = ($348|0)==(0); //@line 286 "load_dbm.cpp"
            if ($349) {
             $i20$6 = $i20$5;$param2$0 = 0;
            } else {
             $350 = (($i20$5) + 1)|0; //@line 286 "load_dbm.cpp"
             $$sum18 = (($i20$5) + ($277))|0; //@line 286 "load_dbm.cpp"
             $351 = (($lpStream) + ($$sum18)|0); //@line 286 "load_dbm.cpp"
             $352 = HEAP8[$351>>0]|0; //@line 286 "load_dbm.cpp"
             $353 = $352&255; //@line 286 "load_dbm.cpp"
             $i20$6 = $350;$param2$0 = $353;
            }
            $354 = ($cmd1$0|0)==(12); //@line 287 "load_dbm.cpp"
            do {
             if ($354) {
              $355 = (((($m$0$ph119) + (($304*6)|0)|0)) + 2|0); //@line 289 "load_dbm.cpp"
              HEAP8[$355>>0] = 1; //@line 289 "load_dbm.cpp"
              $356 = $param1$0&255; //@line 290 "load_dbm.cpp"
              $357 = (((($m$0$ph119) + (($304*6)|0)|0)) + 4|0); //@line 290 "load_dbm.cpp"
              HEAP8[$357>>0] = $356; //@line 290 "load_dbm.cpp"
              $cmd2$161 = $cmd2$0;
              label = 83;
             } else {
              $358 = ($cmd2$0|0)==(12); //@line 293 "load_dbm.cpp"
              if ($358) {
               $359 = (((($m$0$ph119) + (($304*6)|0)|0)) + 2|0); //@line 295 "load_dbm.cpp"
               HEAP8[$359>>0] = 1; //@line 295 "load_dbm.cpp"
               $360 = $param2$0&255; //@line 296 "load_dbm.cpp"
               $361 = (((($m$0$ph119) + (($304*6)|0)|0)) + 4|0); //@line 296 "load_dbm.cpp"
               HEAP8[$361>>0] = $360; //@line 296 "load_dbm.cpp"
               $cmd2$1 = 255;
              } else {
               $cmd2$1 = $cmd2$0;
              }
              $362 = ($cmd1$0>>>0)>(19); //@line 299 "load_dbm.cpp"
              if ($362) {
               $cmd2$161 = $cmd2$1;
               label = 83;
               break;
              }
              $363 = ($cmd1$0>>>0)>(15); //@line 299 "load_dbm.cpp"
              $364 = ($cmd2$1>>>0)<(16); //@line 299 "load_dbm.cpp"
              $or$cond11 = $363 & $364; //@line 299 "load_dbm.cpp"
              if ($or$cond11) {
               $cmd2$161 = $cmd2$1;
               label = 83;
              } else {
               $cmd1$263 = $cmd1$0;$param1$164 = $param1$0;
              }
             }
            } while(0);
            if ((label|0) == 83) {
             label = 0;
             $365 = ($cmd2$161>>>0)<(20); //@line 305 "load_dbm.cpp"
             if ($365) {
              $cmd1$263 = $cmd2$161;$param1$164 = $param2$0;
             } else {
              $i20$0$be = $i20$6;
              break;
             }
            }
            $368 = $cmd1$263&255; //@line 307 "load_dbm.cpp"
            $369 = (($m$0$ph119) + (($304*6)|0)|0); //@line 307 "load_dbm.cpp"
            $370 = (((($m$0$ph119) + (($304*6)|0)|0)) + 3|0); //@line 307 "load_dbm.cpp"
            HEAP8[$370>>0] = $368; //@line 307 "load_dbm.cpp"
            $371 = $param1$164&255; //@line 308 "load_dbm.cpp"
            $372 = (((($m$0$ph119) + (($304*6)|0)|0)) + 5|0); //@line 308 "load_dbm.cpp"
            HEAP8[$372>>0] = $371; //@line 308 "load_dbm.cpp"
            __ZNK10CSoundFile17ConvertModCommandEP11_MODCOMMAND($this,$369); //@line 309 "load_dbm.cpp"
            $i20$0$be = $i20$6;
           } else {
            $$56 = $309 ? $373 : $301; //@line 314 "load_dbm.cpp"
            $374 = $307 >>> 1; //@line 315 "load_dbm.cpp"
            $$lobit = $374 & 1; //@line 315 "load_dbm.cpp"
            $375 = $307 >>> 2; //@line 316 "load_dbm.cpp"
            $$lobit65 = $375 & 1; //@line 316 "load_dbm.cpp"
            $376 = $307 >>> 3; //@line 317 "load_dbm.cpp"
            $$lobit66 = $376 & 1; //@line 317 "load_dbm.cpp"
            $377 = $307 >>> 4; //@line 318 "load_dbm.cpp"
            $$lobit67 = $377 & 1; //@line 318 "load_dbm.cpp"
            $378 = $307 >>> 5; //@line 319 "load_dbm.cpp"
            $$lobit68 = $378 & 1; //@line 319 "load_dbm.cpp"
            $i20$8 = (($$lobit65) + ($$lobit))|0; //@line 315 "load_dbm.cpp"
            $i20$8$ = (($i20$8) + ($$lobit66))|0; //@line 316 "load_dbm.cpp"
            $i20$10 = (($i20$8$) + ($$lobit67))|0; //@line 317 "load_dbm.cpp"
            $i20$10$ = (($i20$10) + ($$lobit68))|0; //@line 318 "load_dbm.cpp"
            $i20$10$$ = (($i20$10$) + ($$56))|0; //@line 319 "load_dbm.cpp"
            $i20$0$be = $i20$10$$;
           }
          } while(0);
          $366 = (($i20$0$be) + 3)|0; //@line 258 "load_dbm.cpp"
          $367 = ($366>>>0)<($276>>>0); //@line 258 "load_dbm.cpp"
          if ($367) {
           $373 = $366;$i20$0116 = $i20$0$be;
          } else {
           break L116;
          }
         }
         $379 = (($row$0$ph120) + 1)|0; //@line 323 "load_dbm.cpp"
         $380 = HEAP32[$63>>2]|0; //@line 324 "load_dbm.cpp"
         $381 = (($m$0$ph119) + (($380*6)|0)|0); //@line 324 "load_dbm.cpp"
         $382 = (($i20$0116$lcssa) + 4)|0; //@line 258 "load_dbm.cpp"
         $383 = ($382>>>0)<($276>>>0); //@line 258 "load_dbm.cpp"
         $384 = ($379>>>0)<($287>>>0); //@line 258 "load_dbm.cpp"
         $or$cond55115 = $384 & $383; //@line 258 "load_dbm.cpp"
         if ($or$cond55115) {
          $427 = $382;$i20$0$ph121 = $$lcssa;$m$0$ph119 = $381;$row$0$ph120 = $379;
         } else {
          break;
         }
        }
       }
      }
     } while(0);
     $385 = (($iPat$0123) + 1)|0; //@line 236 "load_dbm.cpp"
     $386 = ($385>>>0)>=($$nPatterns$0>>>0); //@line 236 "load_dbm.cpp"
     $387 = (($278) + 8)|0;
     $388 = ($387>>>0)>($$lcssa227>>>0); //@line 242 "load_dbm.cpp"
     $or$cond53 = $386 | $388; //@line 236 "load_dbm.cpp"
     if ($or$cond53) {
      break;
     } else {
      $chunk_pos$2124 = $278;$iPat$0123 = $385;
     }
    }
   }
  } while(0);
  $87 = (($$lcssa227) + 10)|0; //@line 126 "load_dbm.cpp"
  $88 = ($87>>>0)<($dwMemLength>>>0); //@line 126 "load_dbm.cpp"
  if ($88) {
   $423 = $87;$dwMemPos$0$ph130 = $$lcssa227;$nInstruments$0$ph128 = $nInstruments$0$ph76112$lcssa232;$nPatterns$0$ph127 = $$nPatterns$0;$nSamples$0$ph129 = $nSamples$0$ph80105$lcssa230;
  } else {
   $$0 = 1;
   label = 96;
   break;
  }
 }
 if ((label|0) == 96) {
  return ($$0|0); //@line 367 "load_dbm.cpp"
 }
 return (0)|0;
}
function __ZN10CSoundFile20FrequencyToTransposeEj($freq) {
 $freq = $freq|0;
 var $0 = 0.0, $1 = 0.0, $2 = 0.0, $3 = 0.0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+($freq>>>0)); //@line 1708 "sndfile.cpp"
 $1 = $0 * 1.1957431543704412E-4; //@line 1708 "sndfile.cpp"
 $2 = (+Math_log((+$1))); //@line 1708 "sndfile.cpp"
 $3 = $2 * 2215.9795828054475; //@line 1708 "sndfile.cpp"
 $4 = (~~(($3))); //@line 1708 "sndfile.cpp"
 return ($4|0); //@line 1708 "sndfile.cpp"
}
function __ZN10CSoundFile7Read669EPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$$loopend$0 = 0, $$0 = 0, $$10 = 0, $$lcssa50 = 0, $$lcssa68 = 0, $$pre = 0, $$pre61 = 0, $$pre62 = 0, $$sum = 0, $$sum8 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0;
 var $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0;
 var $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0;
 var $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0;
 var $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0;
 var $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $command$112 = 0, $dontfuckwithme$044 = 0, $dwMemPos$0$lcssa = 0, $dwMemPos$03563 = 0, $dwMemPos$122 = 0, $exitcond = 0, $exitcond58 = 0, $i2$030$lcssa = 0, $ichk$043 = 0, $loopend$0 = 0, $loopend$2 = 0, $loopstart$0 = 0, $m$031 = 0, $m$126 = 0, $mspeed$028 = 0, $mspeed$1 = 0, $mspeed$2 = 0, $mspeed$2$lcssa = 0;
 var $n$029 = 0, $n3$023 = 0, $nins$040 = 0, $npat$03464 = 0, $or$cond = 0, $or$cond3 = 0, $or$cond5 = 0, $or$cond7 = 0, $p$032 = 0, $p$127 = 0, $param$013 = 0, $psmp$041 = 0, $row$033 = 0, $scevgep = 0, $scevgep56 = 0, $switch$split2D = 0, dest = 0, label = 0, sp = 0, src = 0;
 var stop = 0;
 sp = STACKTOP;
 $0 = ((($lpStream)) + 497|0); //@line 53 "load_669.cpp"
 $1 = ($lpStream|0)==(0|0); //@line 56 "load_669.cpp"
 $2 = ($dwMemLength>>>0)<(498); //@line 56 "load_669.cpp"
 $or$cond = $1 | $2; //@line 56 "load_669.cpp"
 if ($or$cond) {
  $$0 = 0;
  return ($$0|0); //@line 193 "load_669.cpp"
 }
 $3 = HEAP16[$lpStream>>1]|0; //@line 57 "load_669.cpp"
 $switch$split2D = ($3<<16>>16)<(26217);
 L4: do {
  if ($switch$split2D) {
   switch ($3<<16>>16) {
   case 20042:  {
    break L4;
    break;
   }
   default: {
    $$0 = 0;
   }
   }
   return ($$0|0); //@line 193 "load_669.cpp"
  } else {
   switch ($3<<16>>16) {
   case 26217:  {
    break L4;
    break;
   }
   default: {
    $$0 = 0;
   }
   }
   return ($$0|0); //@line 193 "load_669.cpp"
  }
 } while(0);
 $4 = ((($lpStream)) + 110|0); //@line 59 "load_669.cpp"
 $5 = HEAP8[$4>>0]|0; //@line 59 "load_669.cpp"
 $6 = ($5<<24>>24)==(0); //@line 59 "load_669.cpp"
 if ($6) {
  $$0 = 0;
  return ($$0|0); //@line 193 "load_669.cpp"
 }
 $7 = $5&255; //@line 59 "load_669.cpp"
 $8 = ($5&255)>(64); //@line 59 "load_669.cpp"
 if ($8) {
  $$0 = 0;
  return ($$0|0); //@line 193 "load_669.cpp"
 }
 $9 = ((($lpStream)) + 112|0); //@line 59 "load_669.cpp"
 $10 = HEAP8[$9>>0]|0; //@line 59 "load_669.cpp"
 $11 = ($10<<24>>24)<(0); //@line 59 "load_669.cpp"
 if ($11) {
  $$0 = 0;
  return ($$0|0); //@line 193 "load_669.cpp"
 }
 $12 = ((($lpStream)) + 111|0); //@line 60 "load_669.cpp"
 $13 = HEAP8[$12>>0]|0; //@line 60 "load_669.cpp"
 $14 = (($13) + -1)<<24>>24; //@line 60 "load_669.cpp"
 $15 = ($14<<24>>24)<(0); //@line 60 "load_669.cpp"
 if ($15) {
  $$0 = 0;
  return ($$0|0); //@line 193 "load_669.cpp"
 }
 $16 = $13&255; //@line 60 "load_669.cpp"
 $17 = ($7*25)|0; //@line 61 "load_669.cpp"
 $18 = (($17) + 497)|0; //@line 61 "load_669.cpp"
 $19 = ($16*1536)|0; //@line 61 "load_669.cpp"
 $20 = (($18) + ($19))|0; //@line 61 "load_669.cpp"
 $21 = ($20>>>0)>($dwMemLength>>>0); //@line 62 "load_669.cpp"
 if ($21) {
  $$0 = 0;
  return ($$0|0); //@line 193 "load_669.cpp"
 } else {
  $dontfuckwithme$044 = $20;$ichk$043 = 0;
 }
 while(1) {
  $22 = (((($0) + (($ichk$043*25)|0)|0)) + 13|0); //@line 65 "load_669.cpp"
  $23 = (((($0) + (($ichk$043*25)|0)|0)) + 16|0); //@line 39 "load_669.cpp"
  $24 = HEAP8[$23>>0]|0; //@line 39 "load_669.cpp"
  $25 = $24&255; //@line 39 "load_669.cpp"
  $26 = $25 << 24; //@line 39 "load_669.cpp"
  $27 = (((($0) + (($ichk$043*25)|0)|0)) + 15|0); //@line 40 "load_669.cpp"
  $28 = HEAP8[$27>>0]|0; //@line 40 "load_669.cpp"
  $29 = $28&255; //@line 40 "load_669.cpp"
  $30 = $29 << 16; //@line 40 "load_669.cpp"
  $31 = $30 | $26; //@line 39 "load_669.cpp"
  $32 = (((($0) + (($ichk$043*25)|0)|0)) + 14|0); //@line 41 "load_669.cpp"
  $33 = HEAP8[$32>>0]|0; //@line 41 "load_669.cpp"
  $34 = $33&255; //@line 41 "load_669.cpp"
  $35 = $34 << 8; //@line 41 "load_669.cpp"
  $36 = $31 | $35; //@line 40 "load_669.cpp"
  $37 = HEAP8[$22>>0]|0; //@line 42 "load_669.cpp"
  $38 = $37&255; //@line 42 "load_669.cpp"
  $39 = $36 | $38; //@line 41 "load_669.cpp"
  $40 = (($39) + ($dontfuckwithme$044))|0; //@line 66 "load_669.cpp"
  $41 = (($ichk$043) + 1)|0; //@line 63 "load_669.cpp"
  $42 = ($41>>>0)<($7>>>0); //@line 63 "load_669.cpp"
  if ($42) {
   $dontfuckwithme$044 = $40;$ichk$043 = $41;
  } else {
   $$lcssa68 = $40;
   break;
  }
 }
 $43 = ($$lcssa68>>>0)>($dwMemLength>>>0); //@line 68 "load_669.cpp"
 if ($43) {
  $$0 = 0;
  return ($$0|0); //@line 193 "load_669.cpp"
 }
 $44 = ((($this)) + 59680|0); //@line 70 "load_669.cpp"
 HEAP32[$44>>2] = 64; //@line 70 "load_669.cpp"
 $45 = ((($this)) + 59660|0); //@line 71 "load_669.cpp"
 $46 = HEAP32[$45>>2]|0; //@line 71 "load_669.cpp"
 $47 = $46 | 16; //@line 71 "load_669.cpp"
 HEAP32[$45>>2] = $47; //@line 71 "load_669.cpp"
 $48 = ((($this)) + 59764|0); //@line 72 "load_669.cpp"
 HEAP32[$48>>2] = 112; //@line 72 "load_669.cpp"
 $49 = ((($this)) + 59768|0); //@line 73 "load_669.cpp"
 HEAP32[$49>>2] = 13696; //@line 73 "load_669.cpp"
 $50 = ((($this)) + 59652|0); //@line 74 "load_669.cpp"
 HEAP32[$50>>2] = 125; //@line 74 "load_669.cpp"
 $51 = ((($this)) + 59648|0); //@line 75 "load_669.cpp"
 HEAP32[$51>>2] = 6; //@line 75 "load_669.cpp"
 $52 = ((($this)) + 59664|0); //@line 76 "load_669.cpp"
 HEAP32[$52>>2] = 8; //@line 76 "load_669.cpp"
 $53 = ((($this)) + 59804|0); //@line 77 "load_669.cpp"
 $54 = ((($lpStream)) + 2|0); //@line 77 "load_669.cpp"
 dest=$53; src=$54; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 77 "load_669.cpp"
 $55 = HEAP8[$4>>0]|0; //@line 78 "load_669.cpp"
 $56 = $55&255; //@line 78 "load_669.cpp"
 $57 = ((($this)) + 59684|0); //@line 78 "load_669.cpp"
 HEAP32[$57>>2] = $56; //@line 78 "load_669.cpp"
 $58 = ($55<<24>>24)==(0); //@line 79 "load_669.cpp"
 if (!($58)) {
  $nins$040 = 1;$psmp$041 = $0;
  while(1) {
   $59 = ((($psmp$041)) + 13|0); //@line 81 "load_669.cpp"
   $60 = ((($psmp$041)) + 16|0); //@line 39 "load_669.cpp"
   $61 = HEAP8[$60>>0]|0; //@line 39 "load_669.cpp"
   $62 = $61&255; //@line 39 "load_669.cpp"
   $63 = $62 << 24; //@line 39 "load_669.cpp"
   $64 = ((($psmp$041)) + 15|0); //@line 40 "load_669.cpp"
   $65 = HEAP8[$64>>0]|0; //@line 40 "load_669.cpp"
   $66 = $65&255; //@line 40 "load_669.cpp"
   $67 = $66 << 16; //@line 40 "load_669.cpp"
   $68 = $67 | $63; //@line 39 "load_669.cpp"
   $69 = ((($psmp$041)) + 14|0); //@line 41 "load_669.cpp"
   $70 = HEAP8[$69>>0]|0; //@line 41 "load_669.cpp"
   $71 = $70&255; //@line 41 "load_669.cpp"
   $72 = $71 << 8; //@line 41 "load_669.cpp"
   $73 = $68 | $72; //@line 40 "load_669.cpp"
   $74 = HEAP8[$59>>0]|0; //@line 42 "load_669.cpp"
   $75 = $74&255; //@line 42 "load_669.cpp"
   $76 = $73 | $75; //@line 41 "load_669.cpp"
   $77 = ((($psmp$041)) + 17|0); //@line 82 "load_669.cpp"
   $78 = ((($psmp$041)) + 20|0); //@line 39 "load_669.cpp"
   $79 = HEAP8[$78>>0]|0; //@line 39 "load_669.cpp"
   $80 = $79&255; //@line 39 "load_669.cpp"
   $81 = $80 << 24; //@line 39 "load_669.cpp"
   $82 = ((($psmp$041)) + 19|0); //@line 40 "load_669.cpp"
   $83 = HEAP8[$82>>0]|0; //@line 40 "load_669.cpp"
   $84 = $83&255; //@line 40 "load_669.cpp"
   $85 = $84 << 16; //@line 40 "load_669.cpp"
   $86 = $85 | $81; //@line 39 "load_669.cpp"
   $87 = ((($psmp$041)) + 18|0); //@line 41 "load_669.cpp"
   $88 = HEAP8[$87>>0]|0; //@line 41 "load_669.cpp"
   $89 = $88&255; //@line 41 "load_669.cpp"
   $90 = $89 << 8; //@line 41 "load_669.cpp"
   $91 = $86 | $90; //@line 40 "load_669.cpp"
   $92 = HEAP8[$77>>0]|0; //@line 42 "load_669.cpp"
   $93 = $92&255; //@line 42 "load_669.cpp"
   $94 = $91 | $93; //@line 41 "load_669.cpp"
   $95 = ((($psmp$041)) + 21|0); //@line 83 "load_669.cpp"
   $96 = ((($psmp$041)) + 24|0); //@line 39 "load_669.cpp"
   $97 = HEAP8[$96>>0]|0; //@line 39 "load_669.cpp"
   $98 = $97&255; //@line 39 "load_669.cpp"
   $99 = $98 << 24; //@line 39 "load_669.cpp"
   $100 = ((($psmp$041)) + 23|0); //@line 40 "load_669.cpp"
   $101 = HEAP8[$100>>0]|0; //@line 40 "load_669.cpp"
   $102 = $101&255; //@line 40 "load_669.cpp"
   $103 = $102 << 16; //@line 40 "load_669.cpp"
   $104 = $103 | $99; //@line 39 "load_669.cpp"
   $105 = ((($psmp$041)) + 22|0); //@line 41 "load_669.cpp"
   $106 = HEAP8[$105>>0]|0; //@line 41 "load_669.cpp"
   $107 = $106&255; //@line 41 "load_669.cpp"
   $108 = $107 << 8; //@line 41 "load_669.cpp"
   $109 = $104 | $108; //@line 40 "load_669.cpp"
   $110 = HEAP8[$95>>0]|0; //@line 42 "load_669.cpp"
   $111 = $110&255; //@line 42 "load_669.cpp"
   $112 = $109 | $111; //@line 41 "load_669.cpp"
   $113 = ($76>>>0)>(16000000); //@line 84 "load_669.cpp"
   $$ = $113 ? 16000000 : $76; //@line 84 "load_669.cpp"
   $114 = ($112>>>0)<=($$>>>0); //@line 85 "load_669.cpp"
   $115 = ($94|0)!=(0); //@line 85 "load_669.cpp"
   $or$cond3 = $115 | $114; //@line 85 "load_669.cpp"
   $loopend$0 = $or$cond3 ? $112 : 0; //@line 85 "load_669.cpp"
   $116 = ($loopend$0>>>0)>($$>>>0); //@line 86 "load_669.cpp"
   $$$loopend$0 = $116 ? $$ : $loopend$0; //@line 86 "load_669.cpp"
   $117 = (($94) + 4)|0; //@line 87 "load_669.cpp"
   $118 = ($117>>>0)<($$$loopend$0>>>0); //@line 87 "load_669.cpp"
   $loopstart$0 = $118 ? $94 : 0; //@line 87 "load_669.cpp"
   $loopend$2 = $118 ? $$$loopend$0 : 0; //@line 87 "load_669.cpp"
   $119 = (((($this)) + 33280|0) + ($nins$040<<6)|0); //@line 88 "load_669.cpp"
   HEAP32[$119>>2] = $$; //@line 88 "load_669.cpp"
   $120 = (((((($this)) + 33280|0) + ($nins$040<<6)|0)) + 4|0); //@line 89 "load_669.cpp"
   HEAP32[$120>>2] = $loopstart$0; //@line 89 "load_669.cpp"
   $121 = (((((($this)) + 33280|0) + ($nins$040<<6)|0)) + 8|0); //@line 90 "load_669.cpp"
   HEAP32[$121>>2] = $loopend$2; //@line 90 "load_669.cpp"
   $122 = ($loopend$2|0)==(0); //@line 91 "load_669.cpp"
   if (!($122)) {
    $123 = (((((($this)) + 33280|0) + ($nins$040<<6)|0)) + 34|0); //@line 91 "load_669.cpp"
    $124 = HEAP16[$123>>1]|0; //@line 91 "load_669.cpp"
    $125 = $124&65535; //@line 91 "load_669.cpp"
    $126 = $125 | 2; //@line 91 "load_669.cpp"
    $127 = $126&65535; //@line 91 "load_669.cpp"
    HEAP16[$123>>1] = $127; //@line 91 "load_669.cpp"
   }
   $128 = (((($this)) + 59804|0) + ($nins$040<<5)|0); //@line 92 "load_669.cpp"
   dest=$128; src=$psmp$041; stop=dest+13|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 92 "load_669.cpp"
   $129 = (((((($this)) + 33280|0) + ($nins$040<<6)|0)) + 30|0); //@line 93 "load_669.cpp"
   HEAP16[$129>>1] = 256; //@line 93 "load_669.cpp"
   $130 = (((((($this)) + 33280|0) + ($nins$040<<6)|0)) + 32|0); //@line 94 "load_669.cpp"
   HEAP16[$130>>1] = 64; //@line 94 "load_669.cpp"
   $131 = (((((($this)) + 33280|0) + ($nins$040<<6)|0)) + 28|0); //@line 95 "load_669.cpp"
   HEAP16[$131>>1] = 128; //@line 95 "load_669.cpp"
   $132 = (($nins$040) + 1)|0; //@line 79 "load_669.cpp"
   $133 = ((($psmp$041)) + 25|0); //@line 79 "load_669.cpp"
   $134 = HEAP32[$57>>2]|0; //@line 79 "load_669.cpp"
   $135 = ($132>>>0)>($134>>>0); //@line 79 "load_669.cpp"
   if ($135) {
    break;
   } else {
    $nins$040 = $132;$psmp$041 = $133;
   }
  }
 }
 $136 = (__Znaj(109)|0); //@line 98 "load_669.cpp"
 $137 = ((($this)) + 59796|0); //@line 98 "load_669.cpp"
 HEAP32[$137>>2] = $136; //@line 98 "load_669.cpp"
 dest=$136; src=$54; stop=dest+108|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 99 "load_669.cpp"
 $138 = ((($136)) + 108|0); //@line 100 "load_669.cpp"
 HEAP8[$138>>0] = 0; //@line 100 "load_669.cpp"
 $139 = ((($this)) + 53344|0); //@line 102 "load_669.cpp"
 $140 = ((($lpStream)) + 113|0); //@line 102 "load_669.cpp"
 dest=$139; src=$140; stop=dest+128|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 102 "load_669.cpp"
 $141 = HEAP8[$9>>0]|0; //@line 103 "load_669.cpp"
 $142 = $141&255; //@line 103 "load_669.cpp"
 $143 = ((($this)) + 59736|0); //@line 103 "load_669.cpp"
 HEAP32[$143>>2] = $142; //@line 103 "load_669.cpp"
 $144 = (((($this)) + 53344|0) + ($142)|0); //@line 104 "load_669.cpp"
 $145 = HEAP8[$144>>0]|0; //@line 104 "load_669.cpp"
 $146 = HEAP8[$12>>0]|0; //@line 104 "load_669.cpp"
 $147 = ($145&255)<($146&255); //@line 104 "load_669.cpp"
 $$10 = $147 ? $142 : 0; //@line 104 "load_669.cpp"
 HEAP32[$143>>2] = $$10; //@line 104 "load_669.cpp"
 $148 = ((($this)) + 49600|0); //@line 108 "load_669.cpp"
 HEAP32[$148>>2] = 208; //@line 108 "load_669.cpp"
 $149 = ((($this)) + 49604|0); //@line 109 "load_669.cpp"
 HEAP32[$149>>2] = 64; //@line 109 "load_669.cpp"
 $150 = ((($this)) + 49636|0); //@line 108 "load_669.cpp"
 HEAP32[$150>>2] = 48; //@line 108 "load_669.cpp"
 $151 = ((($this)) + 49640|0); //@line 109 "load_669.cpp"
 HEAP32[$151>>2] = 64; //@line 109 "load_669.cpp"
 $152 = ((($this)) + 49672|0); //@line 108 "load_669.cpp"
 HEAP32[$152>>2] = 208; //@line 108 "load_669.cpp"
 $153 = ((($this)) + 49676|0); //@line 109 "load_669.cpp"
 HEAP32[$153>>2] = 64; //@line 109 "load_669.cpp"
 $154 = ((($this)) + 49708|0); //@line 108 "load_669.cpp"
 HEAP32[$154>>2] = 48; //@line 108 "load_669.cpp"
 $155 = ((($this)) + 49712|0); //@line 109 "load_669.cpp"
 HEAP32[$155>>2] = 64; //@line 109 "load_669.cpp"
 $156 = ((($this)) + 49744|0); //@line 108 "load_669.cpp"
 HEAP32[$156>>2] = 208; //@line 108 "load_669.cpp"
 $157 = ((($this)) + 49748|0); //@line 109 "load_669.cpp"
 HEAP32[$157>>2] = 64; //@line 109 "load_669.cpp"
 $158 = ((($this)) + 49780|0); //@line 108 "load_669.cpp"
 HEAP32[$158>>2] = 48; //@line 108 "load_669.cpp"
 $159 = ((($this)) + 49784|0); //@line 109 "load_669.cpp"
 HEAP32[$159>>2] = 64; //@line 109 "load_669.cpp"
 $160 = ((($this)) + 49816|0); //@line 108 "load_669.cpp"
 HEAP32[$160>>2] = 208; //@line 108 "load_669.cpp"
 $161 = ((($this)) + 49820|0); //@line 109 "load_669.cpp"
 HEAP32[$161>>2] = 64; //@line 109 "load_669.cpp"
 $162 = ((($this)) + 49852|0); //@line 108 "load_669.cpp"
 HEAP32[$162>>2] = 48; //@line 108 "load_669.cpp"
 $163 = ((($this)) + 49856|0); //@line 109 "load_669.cpp"
 HEAP32[$163>>2] = 64; //@line 109 "load_669.cpp"
 $164 = HEAP8[$4>>0]|0; //@line 112 "load_669.cpp"
 $165 = $164&255; //@line 112 "load_669.cpp"
 $166 = ($165*25)|0; //@line 112 "load_669.cpp"
 $167 = (($166) + 497)|0; //@line 112 "load_669.cpp"
 $168 = ($146<<24>>24)==(0); //@line 113 "load_669.cpp"
 L37: do {
  if ($168) {
   $dwMemPos$0$lcssa = $167;
  } else {
   $169 = (__ZN10CSoundFile15AllocatePatternEjj(64,8)|0); //@line 115 "load_669.cpp"
   $170 = ((($this)) + 51904|0); //@line 115 "load_669.cpp"
   HEAP32[$170>>2] = $169; //@line 115 "load_669.cpp"
   $171 = ($169|0)==(0|0); //@line 116 "load_669.cpp"
   if ($171) {
    $dwMemPos$0$lcssa = $167;
   } else {
    $176 = $170;$dwMemPos$03563 = $167;$npat$03464 = 0;
    while(1) {
     $174 = (((($this)) + 52864|0) + ($npat$03464<<1)|0); //@line 117 "load_669.cpp"
     HEAP16[$174>>1] = 64; //@line 117 "load_669.cpp"
     $175 = HEAP32[$176>>2]|0; //@line 118 "load_669.cpp"
     $177 = (($lpStream) + ($dwMemPos$03563)|0); //@line 119 "load_669.cpp"
     $$sum = (($npat$03464) + 369)|0; //@line 123 "load_669.cpp"
     $178 = (($lpStream) + ($$sum)|0); //@line 123 "load_669.cpp"
     $$sum8 = (($npat$03464) + 241)|0; //@line 177 "load_669.cpp"
     $179 = (($lpStream) + ($$sum8)|0); //@line 177 "load_669.cpp"
     $m$031 = $175;$p$032 = $177;$row$033 = 0;
     while(1) {
      $180 = HEAP8[$178>>0]|0; //@line 123 "load_669.cpp"
      $181 = $180&255; //@line 123 "load_669.cpp"
      $182 = ($row$033|0)==($181|0); //@line 123 "load_669.cpp"
      $183 = ($row$033|0)!=(63); //@line 123 "load_669.cpp"
      $or$cond5 = $183 & $182; //@line 123 "load_669.cpp"
      if ($or$cond5) {
       $184 = ((($m$031)) + 3|0); //@line 127 "load_669.cpp"
       HEAP8[$184>>0] = 14; //@line 127 "load_669.cpp"
       $185 = ((($m$031)) + 5|0); //@line 128 "load_669.cpp"
       HEAP8[$185>>0] = 0; //@line 128 "load_669.cpp"
       $186 = ((($m$031)) + 9|0); //@line 127 "load_669.cpp"
       HEAP8[$186>>0] = 14; //@line 127 "load_669.cpp"
       $187 = ((($m$031)) + 11|0); //@line 128 "load_669.cpp"
       HEAP8[$187>>0] = 0; //@line 128 "load_669.cpp"
       $188 = ((($m$031)) + 15|0); //@line 127 "load_669.cpp"
       HEAP8[$188>>0] = 14; //@line 127 "load_669.cpp"
       $189 = ((($m$031)) + 17|0); //@line 128 "load_669.cpp"
       HEAP8[$189>>0] = 0; //@line 128 "load_669.cpp"
       $190 = ((($m$031)) + 21|0); //@line 127 "load_669.cpp"
       HEAP8[$190>>0] = 14; //@line 127 "load_669.cpp"
       $191 = ((($m$031)) + 23|0); //@line 128 "load_669.cpp"
       HEAP8[$191>>0] = 0; //@line 128 "load_669.cpp"
       $192 = ((($m$031)) + 27|0); //@line 127 "load_669.cpp"
       HEAP8[$192>>0] = 14; //@line 127 "load_669.cpp"
       $193 = ((($m$031)) + 29|0); //@line 128 "load_669.cpp"
       HEAP8[$193>>0] = 0; //@line 128 "load_669.cpp"
       $194 = ((($m$031)) + 33|0); //@line 127 "load_669.cpp"
       HEAP8[$194>>0] = 14; //@line 127 "load_669.cpp"
       $195 = ((($m$031)) + 35|0); //@line 128 "load_669.cpp"
       HEAP8[$195>>0] = 0; //@line 128 "load_669.cpp"
       $196 = ((($m$031)) + 39|0); //@line 127 "load_669.cpp"
       HEAP8[$196>>0] = 14; //@line 127 "load_669.cpp"
       $197 = ((($m$031)) + 41|0); //@line 128 "load_669.cpp"
       HEAP8[$197>>0] = 0; //@line 128 "load_669.cpp"
       $198 = ((($m$031)) + 45|0); //@line 127 "load_669.cpp"
       HEAP8[$198>>0] = 14; //@line 127 "load_669.cpp"
       $199 = ((($m$031)) + 47|0); //@line 128 "load_669.cpp"
       HEAP8[$199>>0] = 0; //@line 128 "load_669.cpp"
      }
      $scevgep = ((($p$032)) + 24|0);
      $m$126 = $m$031;$mspeed$028 = $m$031;$n$029 = 0;$p$127 = $p$032;
      while(1) {
       $200 = HEAP8[$p$127>>0]|0; //@line 133 "load_669.cpp"
       $201 = $200&255; //@line 133 "load_669.cpp"
       $202 = ((($p$127)) + 1|0); //@line 134 "load_669.cpp"
       $203 = HEAP8[$202>>0]|0; //@line 134 "load_669.cpp"
       $204 = $203&255; //@line 134 "load_669.cpp"
       $205 = ($200&255)<(254); //@line 136 "load_669.cpp"
       if ($205) {
        $206 = $201 << 4; //@line 134 "load_669.cpp"
        $207 = $206 & 48; //@line 134 "load_669.cpp"
        $208 = $204 >>> 4; //@line 134 "load_669.cpp"
        $209 = $207 | $208; //@line 134 "load_669.cpp"
        $210 = $201 >>> 2; //@line 133 "load_669.cpp"
        $211 = (($210) + 37)|0; //@line 138 "load_669.cpp"
        $212 = $211&255; //@line 138 "load_669.cpp"
        HEAP8[$m$126>>0] = $212; //@line 138 "load_669.cpp"
        $213 = (($209) + 1)|0; //@line 139 "load_669.cpp"
        $214 = $213&255; //@line 139 "load_669.cpp"
        $215 = ((($m$126)) + 1|0); //@line 139 "load_669.cpp"
        HEAP8[$215>>0] = $214; //@line 139 "load_669.cpp"
        $$pre61 = HEAP8[$p$127>>0]|0; //@line 141 "load_669.cpp"
        $217 = $$pre61;
       } else {
        $217 = $200;
       }
       $216 = ($217<<24>>24)==(-1); //@line 141 "load_669.cpp"
       if (!($216)) {
        $218 = ((($m$126)) + 2|0); //@line 143 "load_669.cpp"
        HEAP8[$218>>0] = 1; //@line 143 "load_669.cpp"
        $219 = $204 << 2; //@line 144 "load_669.cpp"
        $220 = $219 & 60; //@line 144 "load_669.cpp"
        $221 = $220 | 2; //@line 144 "load_669.cpp"
        $222 = $221&255; //@line 144 "load_669.cpp"
        $223 = ((($m$126)) + 4|0); //@line 144 "load_669.cpp"
        HEAP8[$223>>0] = $222; //@line 144 "load_669.cpp"
       }
       $224 = ((($p$127)) + 2|0); //@line 146 "load_669.cpp"
       $225 = HEAP8[$224>>0]|0; //@line 146 "load_669.cpp"
       $226 = $225&255; //@line 146 "load_669.cpp"
       $227 = ($225<<24>>24)==(-1); //@line 146 "load_669.cpp"
       L54: do {
        if ($227) {
         $mspeed$2 = $mspeed$028;
        } else {
         $228 = $226 >>> 4; //@line 148 "load_669.cpp"
         $229 = $226 & 15; //@line 149 "load_669.cpp"
         L56: do {
          switch ($228|0) {
          case 0:  {
           $command$112 = 2;$mspeed$1 = $mspeed$028;$param$013 = $229;
           break;
          }
          case 1:  {
           $command$112 = 3;$mspeed$1 = $mspeed$028;$param$013 = $229;
           break;
          }
          case 2:  {
           $command$112 = 4;$mspeed$1 = $mspeed$028;$param$013 = $229;
           break;
          }
          case 3:  {
           $230 = $229 | 80; //@line 155 "load_669.cpp"
           $command$112 = 19;$mspeed$1 = $mspeed$028;$param$013 = $230;
           break;
          }
          case 4:  {
           $231 = $229 | 64; //@line 156 "load_669.cpp"
           $command$112 = 5;$mspeed$1 = $mspeed$028;$param$013 = $231;
           break;
          }
          case 5:  {
           $232 = ($229|0)==(0); //@line 157 "load_669.cpp"
           $233 = (($229) + 2)|0; //@line 157 "load_669.cpp"
           if ($232) {
            $mspeed$2 = $mspeed$028;
            break L54;
           } else {
            $command$112 = 16;$mspeed$1 = 0;$param$013 = $233;
           }
           break;
          }
          case 6:  {
           switch ($229|0) {
           case 0:  {
            $command$112 = 29;$mspeed$1 = $mspeed$028;$param$013 = 254;
            break L56;
            break;
           }
           case 1:  {
            break;
           }
           default: {
            $mspeed$2 = $mspeed$028;
            break L54;
           }
           }
           $command$112 = 29;$mspeed$1 = $mspeed$028;$param$013 = 239;
           break;
          }
          default: {
           $mspeed$2 = $mspeed$028;
           break L54;
          }
          }
         } while(0);
         $234 = ((($m$126)) + 3|0); //@line 167 "load_669.cpp"
         HEAP8[$234>>0] = $command$112; //@line 167 "load_669.cpp"
         $235 = $param$013&255; //@line 168 "load_669.cpp"
         $236 = ((($m$126)) + 5|0); //@line 168 "load_669.cpp"
         HEAP8[$236>>0] = $235; //@line 168 "load_669.cpp"
         $mspeed$2 = $mspeed$1;
        }
       } while(0);
       $237 = (($n$029) + 1)|0; //@line 131 "load_669.cpp"
       $238 = ((($m$126)) + 6|0); //@line 131 "load_669.cpp"
       $239 = ((($p$127)) + 3|0); //@line 131 "load_669.cpp"
       $exitcond = ($237|0)==(8); //@line 131 "load_669.cpp"
       if ($exitcond) {
        $mspeed$2$lcssa = $mspeed$2;
        break;
       } else {
        $m$126 = $238;$mspeed$028 = $mspeed$2;$n$029 = $237;$p$127 = $239;
       }
      }
      $scevgep56 = ((($m$031)) + 48|0);
      $240 = ($row$033|0)==(0); //@line 172 "load_669.cpp"
      $241 = ($mspeed$2$lcssa|0)!=(0|0); //@line 172 "load_669.cpp"
      $or$cond7 = $240 & $241; //@line 172 "load_669.cpp"
      L67: do {
       if ($or$cond7) {
        $242 = ((($mspeed$2$lcssa)) + 3|0); //@line 174 "load_669.cpp"
        $243 = HEAP8[$242>>0]|0; //@line 174 "load_669.cpp"
        $244 = ($243<<24>>24)==(0); //@line 174 "load_669.cpp"
        do {
         if ($244) {
          $$lcssa50 = $242;$i2$030$lcssa = 0;
         } else {
          $250 = ((($mspeed$2$lcssa)) + 9|0); //@line 174 "load_669.cpp"
          $251 = HEAP8[$250>>0]|0; //@line 174 "load_669.cpp"
          $252 = ($251<<24>>24)==(0); //@line 174 "load_669.cpp"
          if ($252) {
           $$lcssa50 = $250;$i2$030$lcssa = 1;
          } else {
           $272 = ((($mspeed$2$lcssa)) + 15|0); //@line 174 "load_669.cpp"
           $273 = HEAP8[$272>>0]|0; //@line 174 "load_669.cpp"
           $274 = ($273<<24>>24)==(0); //@line 174 "load_669.cpp"
           if ($274) {
            $$lcssa50 = $272;$i2$030$lcssa = 2;
            break;
           }
           $275 = ((($mspeed$2$lcssa)) + 21|0); //@line 174 "load_669.cpp"
           $276 = HEAP8[$275>>0]|0; //@line 174 "load_669.cpp"
           $277 = ($276<<24>>24)==(0); //@line 174 "load_669.cpp"
           if ($277) {
            $$lcssa50 = $275;$i2$030$lcssa = 3;
            break;
           }
           $278 = ((($mspeed$2$lcssa)) + 27|0); //@line 174 "load_669.cpp"
           $279 = HEAP8[$278>>0]|0; //@line 174 "load_669.cpp"
           $280 = ($279<<24>>24)==(0); //@line 174 "load_669.cpp"
           if ($280) {
            $$lcssa50 = $278;$i2$030$lcssa = 4;
            break;
           }
           $281 = ((($mspeed$2$lcssa)) + 33|0); //@line 174 "load_669.cpp"
           $282 = HEAP8[$281>>0]|0; //@line 174 "load_669.cpp"
           $283 = ($282<<24>>24)==(0); //@line 174 "load_669.cpp"
           if ($283) {
            $$lcssa50 = $281;$i2$030$lcssa = 5;
            break;
           }
           $284 = ((($mspeed$2$lcssa)) + 39|0); //@line 174 "load_669.cpp"
           $285 = HEAP8[$284>>0]|0; //@line 174 "load_669.cpp"
           $286 = ($285<<24>>24)==(0); //@line 174 "load_669.cpp"
           if ($286) {
            $$lcssa50 = $284;$i2$030$lcssa = 6;
            break;
           }
           $287 = ((($mspeed$2$lcssa)) + 45|0); //@line 174 "load_669.cpp"
           $288 = HEAP8[$287>>0]|0; //@line 174 "load_669.cpp"
           $289 = ($288<<24>>24)==(0); //@line 174 "load_669.cpp"
           if ($289) {
            $$lcssa50 = $287;$i2$030$lcssa = 7;
           } else {
            break L67;
           }
          }
         }
        } while(0);
        HEAP8[$$lcssa50>>0] = 16; //@line 176 "load_669.cpp"
        $245 = HEAP8[$179>>0]|0; //@line 177 "load_669.cpp"
        $246 = $245&255; //@line 177 "load_669.cpp"
        $247 = (($246) + 2)|0; //@line 177 "load_669.cpp"
        $248 = $247&255; //@line 177 "load_669.cpp"
        $249 = (((($mspeed$2$lcssa) + (($i2$030$lcssa*6)|0)|0)) + 5|0); //@line 177 "load_669.cpp"
        HEAP8[$249>>0] = $248; //@line 177 "load_669.cpp"
       }
      } while(0);
      $253 = (($row$033) + 1)|0; //@line 120 "load_669.cpp"
      $exitcond58 = ($253|0)==(64); //@line 120 "load_669.cpp"
      if ($exitcond58) {
       break;
      } else {
       $m$031 = $scevgep56;$p$032 = $scevgep;$row$033 = $253;
      }
     }
     $254 = (($dwMemPos$03563) + 1536)|0; //@line 182 "load_669.cpp"
     $255 = (($npat$03464) + 1)|0; //@line 113 "load_669.cpp"
     $256 = HEAP8[$12>>0]|0; //@line 113 "load_669.cpp"
     $257 = $256&255; //@line 113 "load_669.cpp"
     $258 = ($255>>>0)<($257>>>0); //@line 113 "load_669.cpp"
     if (!($258)) {
      $dwMemPos$0$lcssa = $254;
      break L37;
     }
     $$pre = HEAP32[$52>>2]|0; //@line 115 "load_669.cpp"
     $259 = (__ZN10CSoundFile15AllocatePatternEjj(64,$$pre)|0); //@line 115 "load_669.cpp"
     $260 = (((($this)) + 51904|0) + ($255<<2)|0); //@line 115 "load_669.cpp"
     HEAP32[$260>>2] = $259; //@line 115 "load_669.cpp"
     $261 = ($259|0)==(0|0); //@line 116 "load_669.cpp"
     if ($261) {
      $dwMemPos$0$lcssa = $254;
      break;
     } else {
      $176 = $260;$dwMemPos$03563 = $254;$npat$03464 = $255;
     }
    }
   }
  }
 } while(0);
 $172 = HEAP32[$57>>2]|0; //@line 185 "load_669.cpp"
 $173 = ($172|0)==(0); //@line 185 "load_669.cpp"
 if ($173) {
  $$0 = 1;
  return ($$0|0); //@line 193 "load_669.cpp"
 } else {
  $290 = $172;$dwMemPos$122 = $dwMemPos$0$lcssa;$n3$023 = 1;
 }
 while(1) {
  $262 = (((($this)) + 33280|0) + ($n3$023<<6)|0); //@line 187 "load_669.cpp"
  $263 = HEAP32[$262>>2]|0; //@line 187 "load_669.cpp"
  $264 = ($dwMemPos$122>>>0)<($dwMemLength>>>0); //@line 188 "load_669.cpp"
  if (!($264)) {
   $$0 = 1;
   label = 47;
   break;
  }
  $265 = ($263>>>0)>(4); //@line 189 "load_669.cpp"
  if ($265) {
   $266 = (($lpStream) + ($dwMemPos$122)|0); //@line 189 "load_669.cpp"
   $267 = (($dwMemLength) - ($dwMemPos$122))|0; //@line 189 "load_669.cpp"
   (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$262,1,$266,$267)|0); //@line 189 "load_669.cpp"
   $$pre62 = HEAP32[$57>>2]|0; //@line 185 "load_669.cpp"
   $271 = $$pre62;
  } else {
   $271 = $290;
  }
  $268 = (($263) + ($dwMemPos$122))|0; //@line 190 "load_669.cpp"
  $269 = (($n3$023) + 1)|0; //@line 185 "load_669.cpp"
  $270 = ($269>>>0)>($271>>>0); //@line 185 "load_669.cpp"
  if ($270) {
   $$0 = 1;
   label = 47;
   break;
  } else {
   $290 = $271;$dwMemPos$122 = $268;$n3$023 = $269;
  }
 }
 if ((label|0) == 47) {
  return ($$0|0); //@line 193 "load_669.cpp"
 }
 return (0)|0;
}
function __ZN10CSoundFile7ReadFAREPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$0 = 0, $$19 = 0, $$20 = 0, $$21 = 0, $$pre = 0, $$pre$phiZ2D = 0, $$sum = 0, $$sum10 = 0, $$sum11 = 0, $$sum12 = 0, $$sum13 = 0, $$sum14 = 0, $$sum15 = 0, $$sum16 = 0, $$sum17 = 0, $$sum18 = 0, $$sum4 = 0, $$sum5 = 0, $$sum6 = 0;
 var $$sum7 = 0, $$sum8 = 0, $$sum9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0;
 var $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dwMemPos$0 = 0, $dwMemPos$136 = 0, $dwMemPos$2 = 0, $dwMemPos$2$lcssa = 0, $dwMemPos$331 = 0, $dwMemPos$4 = 0, $exitcond = 0, $exitcond39 = 0, $iorder$037 = 0, $ipat$035 = 0;
 var $ismp$030 = 0, $len$034 = 0, $m$032 = 0, $nchpan$038 = 0, $or$cond = 0, $or$cond23 = 0, $or$cond25 = 0, $or$cond3 = 0, $or$cond41 = 0, $patbrk$033 = 0, $patbrk$1 = 0, $pins$029 = 0, $rows$1 = 0, $samplemap = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $samplemap = sp; //@line 229 "load_far.cpp"
 $0 = ($lpStream|0)==(0|0); //@line 67 "load_far.cpp"
 $1 = ($dwMemLength>>>0)<(1024); //@line 67 "load_far.cpp"
 $or$cond = $0 | $1; //@line 67 "load_far.cpp"
 if ($or$cond) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 264 "load_far.cpp"
 }
 $2 = HEAPU8[$lpStream>>0]|(HEAPU8[$lpStream+1>>0]<<8)|(HEAPU8[$lpStream+2>>0]<<16)|(HEAPU8[$lpStream+3>>0]<<24); //@line 67 "load_far.cpp"
 $3 = ($2|0)==(-28163770); //@line 67 "load_far.cpp"
 if (!($3)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 264 "load_far.cpp"
 }
 $4 = ((($lpStream)) + 44|0); //@line 68 "load_far.cpp"
 $5 = HEAP8[$4>>0]|0; //@line 68 "load_far.cpp"
 $6 = ($5<<24>>24)==(13); //@line 68 "load_far.cpp"
 if (!($6)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 264 "load_far.cpp"
 }
 $7 = ((($lpStream)) + 45|0); //@line 68 "load_far.cpp"
 $8 = HEAP8[$7>>0]|0; //@line 68 "load_far.cpp"
 $9 = ($8<<24>>24)==(10); //@line 68 "load_far.cpp"
 if (!($9)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 264 "load_far.cpp"
 }
 $10 = ((($lpStream)) + 46|0); //@line 68 "load_far.cpp"
 $11 = HEAP8[$10>>0]|0; //@line 68 "load_far.cpp"
 $12 = ($11<<24>>24)==(26); //@line 68 "load_far.cpp"
 if (!($12)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 264 "load_far.cpp"
 }
 $13 = ((($lpStream)) + 47|0); //@line 69 "load_far.cpp"
 $14 = HEAPU8[$13>>0]|(HEAPU8[$13+1>>0]<<8); //@line 69 "load_far.cpp"
 $15 = $14&65535; //@line 69 "load_far.cpp"
 $16 = ((($lpStream)) + 96|0); //@line 70 "load_far.cpp"
 $17 = HEAPU8[$16>>0]|(HEAPU8[$16+1>>0]<<8); //@line 70 "load_far.cpp"
 $18 = $17&65535; //@line 70 "load_far.cpp"
 $19 = ($15>>>0)<($dwMemLength>>>0); //@line 71 "load_far.cpp"
 if (!($19)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 264 "load_far.cpp"
 }
 $20 = (($18) + 98)|0; //@line 71 "load_far.cpp"
 $21 = (($18) + 869)|0; //@line 71 "load_far.cpp"
 $22 = ($21>>>0)<($dwMemLength>>>0); //@line 71 "load_far.cpp"
 if (!($22)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 264 "load_far.cpp"
 }
 $23 = ((($this)) + 59680|0); //@line 73 "load_far.cpp"
 HEAP32[$23>>2] = 512; //@line 73 "load_far.cpp"
 $24 = ((($this)) + 59664|0); //@line 74 "load_far.cpp"
 HEAP32[$24>>2] = 16; //@line 74 "load_far.cpp"
 $25 = ((($this)) + 59688|0); //@line 75 "load_far.cpp"
 HEAP32[$25>>2] = 0; //@line 75 "load_far.cpp"
 $26 = ((($this)) + 59684|0); //@line 76 "load_far.cpp"
 HEAP32[$26>>2] = 0; //@line 76 "load_far.cpp"
 $27 = ((($this)) + 59748|0); //@line 77 "load_far.cpp"
 HEAP32[$27>>2] = 32; //@line 77 "load_far.cpp"
 $28 = ((($lpStream)) + 75|0); //@line 78 "load_far.cpp"
 $29 = HEAP8[$28>>0]|0; //@line 78 "load_far.cpp"
 $30 = $29&255; //@line 78 "load_far.cpp"
 $31 = ((($this)) + 59648|0); //@line 78 "load_far.cpp"
 HEAP32[$31>>2] = $30; //@line 78 "load_far.cpp"
 $32 = ((($this)) + 59652|0); //@line 79 "load_far.cpp"
 HEAP32[$32>>2] = 80; //@line 79 "load_far.cpp"
 $33 = ((($this)) + 59656|0); //@line 80 "load_far.cpp"
 HEAP32[$33>>2] = 256; //@line 80 "load_far.cpp"
 $34 = ((($this)) + 59804|0); //@line 82 "load_far.cpp"
 $35 = ((($lpStream)) + 4|0); //@line 82 "load_far.cpp"
 dest=$34; src=$35; stop=dest+32|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 82 "load_far.cpp"
 $nchpan$038 = 0;
 while(1) {
  $36 = (((((($this)) + 49600|0) + (($nchpan$038*36)|0)|0)) + 8|0); //@line 86 "load_far.cpp"
  HEAP32[$36>>2] = 0; //@line 86 "load_far.cpp"
  $$sum18 = (($nchpan$038) + 76)|0; //@line 87 "load_far.cpp"
  $37 = (($lpStream) + ($$sum18)|0); //@line 87 "load_far.cpp"
  $38 = HEAP8[$37>>0]|0; //@line 87 "load_far.cpp"
  $39 = $38&255; //@line 87 "load_far.cpp"
  $40 = $39 << 4; //@line 87 "load_far.cpp"
  $41 = $40 & 240; //@line 87 "load_far.cpp"
  $42 = $41 | 8; //@line 87 "load_far.cpp"
  $43 = (((($this)) + 49600|0) + (($nchpan$038*36)|0)|0); //@line 87 "load_far.cpp"
  HEAP32[$43>>2] = $42; //@line 87 "load_far.cpp"
  $44 = (((((($this)) + 49600|0) + (($nchpan$038*36)|0)|0)) + 4|0); //@line 88 "load_far.cpp"
  HEAP32[$44>>2] = 64; //@line 88 "load_far.cpp"
  $45 = (($nchpan$038) + 1)|0; //@line 84 "load_far.cpp"
  $exitcond39 = ($45|0)==(16); //@line 84 "load_far.cpp"
  if ($exitcond39) {
   break;
  } else {
   $nchpan$038 = $45;
  }
 }
 $46 = ($17<<16>>16)==(0); //@line 91 "load_far.cpp"
 if ($46) {
  $dwMemPos$0 = 98;
 } else {
  $47 = (($dwMemLength) + -98)|0; //@line 94 "load_far.cpp"
  $48 = ($18>>>0)>($47>>>0); //@line 94 "load_far.cpp"
  $$ = $48 ? $47 : $18; //@line 94 "load_far.cpp"
  $49 = (($$) + 1)|0; //@line 95 "load_far.cpp"
  $50 = (__Znaj($49)|0); //@line 95 "load_far.cpp"
  $51 = ((($this)) + 59796|0); //@line 95 "load_far.cpp"
  HEAP32[$51>>2] = $50; //@line 95 "load_far.cpp"
  $52 = ((($lpStream)) + 98|0); //@line 97 "load_far.cpp"
  _memcpy(($50|0),($52|0),($$|0))|0; //@line 97 "load_far.cpp"
  $53 = (($50) + ($$)|0); //@line 98 "load_far.cpp"
  HEAP8[$53>>0] = 0; //@line 98 "load_far.cpp"
  $dwMemPos$0 = $20;
 }
 $54 = (($dwMemLength) - ($dwMemPos$0))|0; //@line 103 "load_far.cpp"
 $55 = ($54>>>0)<(771); //@line 103 "load_far.cpp"
 if ($55) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 264 "load_far.cpp"
 }
 $56 = (($dwMemPos$0) + 771)|0; //@line 105 "load_far.cpp"
 $57 = ($56>>>0)<($dwMemLength>>>0); //@line 106 "load_far.cpp"
 if (!($57)) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 264 "load_far.cpp"
 }
 $$sum16 = (($dwMemPos$0) + 257)|0; //@line 109 "load_far.cpp"
 $58 = (($lpStream) + ($$sum16)|0); //@line 109 "load_far.cpp"
 $iorder$037 = 0;
 while(1) {
  $59 = HEAP8[$58>>0]|0; //@line 109 "load_far.cpp"
  $60 = $59&255; //@line 109 "load_far.cpp"
  $61 = ($iorder$037>>>0)>($60>>>0); //@line 109 "load_far.cpp"
  if ($61) {
   $66 = 255;
  } else {
   $$sum17 = (($iorder$037) + ($dwMemPos$0))|0; //@line 109 "load_far.cpp"
   $62 = (($lpStream) + ($$sum17)|0); //@line 109 "load_far.cpp"
   $63 = HEAP8[$62>>0]|0; //@line 109 "load_far.cpp"
   $64 = $63&255; //@line 109 "load_far.cpp"
   $66 = $64;
  }
  $65 = $66&255; //@line 109 "load_far.cpp"
  $67 = (((($this)) + 53344|0) + ($iorder$037)|0); //@line 109 "load_far.cpp"
  HEAP8[$67>>0] = $65; //@line 109 "load_far.cpp"
  $68 = (($iorder$037) + 1)|0; //@line 107 "load_far.cpp"
  $exitcond = ($68|0)==(256); //@line 107 "load_far.cpp"
  if ($exitcond) {
   break;
  } else {
   $iorder$037 = $68;
  }
 }
 $$sum = (($dwMemPos$0) + 258)|0; //@line 111 "load_far.cpp"
 $69 = (($lpStream) + ($$sum)|0); //@line 111 "load_far.cpp"
 $70 = HEAP8[$69>>0]|0; //@line 111 "load_far.cpp"
 $71 = $70&255; //@line 111 "load_far.cpp"
 $72 = ((($this)) + 59736|0); //@line 111 "load_far.cpp"
 HEAP32[$72>>2] = $71; //@line 111 "load_far.cpp"
 $73 = (($15) - ($21))|0; //@line 113 "load_far.cpp"
 $74 = (($56) + ($73))|0; //@line 113 "load_far.cpp"
 $75 = ($74>>>0)<($dwMemLength>>>0); //@line 114 "load_far.cpp"
 if (!($75)) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 264 "load_far.cpp"
 }
 $$sum4 = (($dwMemPos$0) + 259)|0; //@line 118 "load_far.cpp"
 $76 = (($lpStream) + ($$sum4)|0); //@line 118 "load_far.cpp"
 $dwMemPos$136 = $74;$ipat$035 = 0;
 L43: while(1) {
  $77 = (($76) + ($ipat$035<<1)|0); //@line 119 "load_far.cpp"
  $78 = HEAP16[$77>>1]|0; //@line 119 "load_far.cpp"
  $79 = ($78<<16>>16)==(0); //@line 119 "load_far.cpp"
  L45: do {
   if ($79) {
    $dwMemPos$2 = $dwMemPos$136;
   } else {
    $80 = $78&65535; //@line 121 "load_far.cpp"
    $81 = ($ipat$035>>>0)>(239); //@line 122 "load_far.cpp"
    $82 = ($78&65535)<(2); //@line 122 "load_far.cpp"
    $or$cond3 = $81 | $82; //@line 122 "load_far.cpp"
    $83 = (($80) + ($dwMemPos$136))|0; //@line 124 "load_far.cpp"
    if ($or$cond3) {
     $dwMemPos$2 = $83;
    } else {
     $84 = ($83>>>0)<($dwMemLength>>>0); //@line 127 "load_far.cpp"
     if (!($84)) {
      $$0 = 1;
      label = 61;
      break L43;
     }
     $85 = (($80) + -2)|0; //@line 128 "load_far.cpp"
     $86 = $85 >>> 6; //@line 128 "load_far.cpp"
     $87 = ($86|0)==(0); //@line 129 "load_far.cpp"
     if ($87) {
      $dwMemPos$2 = $83;
     } else {
      $88 = ($85>>>0)>(16447); //@line 134 "load_far.cpp"
      $$19 = $88 ? 256 : $86; //@line 134 "load_far.cpp"
      $89 = ($$19>>>0)<(16); //@line 135 "load_far.cpp"
      $rows$1 = $89 ? 16 : $$19; //@line 135 "load_far.cpp"
      $90 = $rows$1&65535; //@line 136 "load_far.cpp"
      $91 = (((($this)) + 52864|0) + ($ipat$035<<1)|0); //@line 136 "load_far.cpp"
      HEAP16[$91>>1] = $90; //@line 136 "load_far.cpp"
      $92 = HEAP32[$24>>2]|0; //@line 137 "load_far.cpp"
      $93 = (__ZN10CSoundFile15AllocatePatternEjj($rows$1,$92)|0); //@line 137 "load_far.cpp"
      $94 = (((($this)) + 51904|0) + ($ipat$035<<2)|0); //@line 137 "load_far.cpp"
      HEAP32[$94>>2] = $93; //@line 137 "load_far.cpp"
      $95 = ($93|0)==(0|0); //@line 137 "load_far.cpp"
      if ($95) {
       $$0 = 1;
       label = 61;
       break L43;
      }
      $$sum11 = (($dwMemPos$136) + 2)|0; //@line 140 "load_far.cpp"
      $96 = $rows$1 << 6; //@line 141 "load_far.cpp"
      $97 = ($96>>>0)>($85>>>0); //@line 142 "load_far.cpp"
      $$20 = $97 ? $85 : $96; //@line 142 "load_far.cpp"
      $98 = ($$20|0)==(0); //@line 143 "load_far.cpp"
      if ($98) {
       $dwMemPos$2 = $83;
      } else {
       $99 = (($lpStream) + ($dwMemPos$136)|0); //@line 139 "load_far.cpp"
       $100 = HEAP8[$99>>0]|0; //@line 139 "load_far.cpp"
       $101 = $100&255; //@line 139 "load_far.cpp"
       $102 = (($rows$1) + -1)|0;
       $len$034 = 0;$m$032 = $93;$patbrk$033 = $101;
       while(1) {
        $$sum12 = (($len$034) + ($$sum11))|0; //@line 145 "load_far.cpp"
        $103 = (($lpStream) + ($$sum12)|0); //@line 145 "load_far.cpp"
        $104 = HEAP8[$103>>0]|0; //@line 145 "load_far.cpp"
        $105 = $len$034 | 2; //@line 147 "load_far.cpp"
        $$sum14 = (($105) + ($$sum11))|0; //@line 147 "load_far.cpp"
        $106 = (($lpStream) + ($$sum14)|0); //@line 147 "load_far.cpp"
        $107 = HEAP8[$106>>0]|0; //@line 147 "load_far.cpp"
        $108 = $len$034 | 3; //@line 148 "load_far.cpp"
        $$sum15 = (($108) + ($$sum11))|0; //@line 148 "load_far.cpp"
        $109 = (($lpStream) + ($$sum15)|0); //@line 148 "load_far.cpp"
        $110 = HEAP8[$109>>0]|0; //@line 148 "load_far.cpp"
        $111 = ($104<<24>>24)==(0); //@line 149 "load_far.cpp"
        if (!($111)) {
         $112 = $len$034 | 1; //@line 146 "load_far.cpp"
         $$sum13 = (($112) + ($$sum11))|0; //@line 146 "load_far.cpp"
         $113 = (($lpStream) + ($$sum13)|0); //@line 146 "load_far.cpp"
         $114 = HEAP8[$113>>0]|0; //@line 146 "load_far.cpp"
         $115 = $114&255; //@line 151 "load_far.cpp"
         $116 = (($115) + 1)|0; //@line 151 "load_far.cpp"
         $117 = $116&255; //@line 151 "load_far.cpp"
         $118 = ((($m$032)) + 1|0); //@line 151 "load_far.cpp"
         HEAP8[$118>>0] = $117; //@line 151 "load_far.cpp"
         $119 = $104&255; //@line 152 "load_far.cpp"
         $120 = (($119) + 36)|0; //@line 152 "load_far.cpp"
         $121 = $120&255; //@line 152 "load_far.cpp"
         HEAP8[$m$032>>0] = $121; //@line 152 "load_far.cpp"
        }
        $122 = $107&255; //@line 154 "load_far.cpp"
        $123 = $122 & 15; //@line 154 "load_far.cpp"
        $124 = ($123|0)==(0); //@line 154 "load_far.cpp"
        if (!($124)) {
         $125 = ((($m$032)) + 2|0); //@line 156 "load_far.cpp"
         HEAP8[$125>>0] = 1; //@line 156 "load_far.cpp"
         $126 = $123 << 2; //@line 157 "load_far.cpp"
         $127 = $126&255; //@line 157 "load_far.cpp"
         $128 = ((($m$032)) + 4|0); //@line 157 "load_far.cpp"
         $129 = ($126>>>0)<(5); //@line 158 "load_far.cpp"
         $$21 = $129 ? 0 : $127; //@line 158 "load_far.cpp"
         HEAP8[$128>>0] = $$21; //@line 158 "load_far.cpp"
        }
        $130 = $110&255; //@line 160 "load_far.cpp"
        $131 = $130 & 240; //@line 160 "load_far.cpp"
        L60: do {
         switch ($131|0) {
         case 16:  {
          $132 = ((($m$032)) + 3|0); //@line 164 "load_far.cpp"
          HEAP8[$132>>0] = 2; //@line 164 "load_far.cpp"
          $133 = $130 & 15; //@line 165 "load_far.cpp"
          $134 = $133&255; //@line 165 "load_far.cpp"
          $135 = ((($m$032)) + 5|0); //@line 165 "load_far.cpp"
          HEAP8[$135>>0] = $134; //@line 165 "load_far.cpp"
          $patbrk$1 = $patbrk$033;
          break;
         }
         case 32:  {
          $136 = ((($m$032)) + 3|0); //@line 169 "load_far.cpp"
          HEAP8[$136>>0] = 3; //@line 169 "load_far.cpp"
          $137 = $130 & 15; //@line 170 "load_far.cpp"
          $138 = $137&255; //@line 170 "load_far.cpp"
          $139 = ((($m$032)) + 5|0); //@line 170 "load_far.cpp"
          HEAP8[$139>>0] = $138; //@line 170 "load_far.cpp"
          $patbrk$1 = $patbrk$033;
          break;
         }
         case 48:  {
          $140 = ((($m$032)) + 3|0); //@line 174 "load_far.cpp"
          HEAP8[$140>>0] = 4; //@line 174 "load_far.cpp"
          $141 = $130 << 2; //@line 175 "load_far.cpp"
          $142 = $141 & 60; //@line 175 "load_far.cpp"
          $143 = $142&255; //@line 175 "load_far.cpp"
          $144 = ((($m$032)) + 5|0); //@line 175 "load_far.cpp"
          HEAP8[$144>>0] = $143; //@line 175 "load_far.cpp"
          $patbrk$1 = $patbrk$033;
          break;
         }
         case 64:  {
          $145 = ((($m$032)) + 3|0); //@line 179 "load_far.cpp"
          HEAP8[$145>>0] = 15; //@line 179 "load_far.cpp"
          $146 = $130 & 15; //@line 180 "load_far.cpp"
          $147 = (($146) + 1)|0; //@line 180 "load_far.cpp"
          $148 = (6 / ($147>>>0))&-1; //@line 180 "load_far.cpp"
          $149 = (($148) + 1)|0; //@line 180 "load_far.cpp"
          $150 = $149&255; //@line 180 "load_far.cpp"
          $151 = ((($m$032)) + 5|0); //@line 180 "load_far.cpp"
          HEAP8[$151>>0] = $150; //@line 180 "load_far.cpp"
          $patbrk$1 = $patbrk$033;
          break;
         }
         case 80:  {
          $152 = ((($m$032)) + 3|0); //@line 184 "load_far.cpp"
          HEAP8[$152>>0] = 5; //@line 184 "load_far.cpp"
          $153 = $130 & 15; //@line 185 "load_far.cpp"
          $154 = $153&255; //@line 185 "load_far.cpp"
          $155 = ((($m$032)) + 5|0); //@line 185 "load_far.cpp"
          HEAP8[$155>>0] = $154; //@line 185 "load_far.cpp"
          $patbrk$1 = $patbrk$033;
          break;
         }
         case 96:  {
          $156 = ((($m$032)) + 3|0); //@line 189 "load_far.cpp"
          HEAP8[$156>>0] = 5; //@line 189 "load_far.cpp"
          $157 = $130 << 4; //@line 190 "load_far.cpp"
          $158 = $157&255; //@line 190 "load_far.cpp"
          $159 = ((($m$032)) + 5|0); //@line 190 "load_far.cpp"
          HEAP8[$159>>0] = $158; //@line 190 "load_far.cpp"
          $patbrk$1 = $patbrk$033;
          break;
         }
         case 112:  {
          $160 = ((($m$032)) + 3|0); //@line 194 "load_far.cpp"
          HEAP8[$160>>0] = 11; //@line 194 "load_far.cpp"
          $161 = $130 << 4; //@line 195 "load_far.cpp"
          $162 = $161&255; //@line 195 "load_far.cpp"
          $163 = ((($m$032)) + 5|0); //@line 195 "load_far.cpp"
          HEAP8[$163>>0] = $162; //@line 195 "load_far.cpp"
          $patbrk$1 = $patbrk$033;
          break;
         }
         case 128:  {
          $164 = ((($m$032)) + 3|0); //@line 199 "load_far.cpp"
          HEAP8[$164>>0] = 11; //@line 199 "load_far.cpp"
          $165 = $130 & 15; //@line 200 "load_far.cpp"
          $166 = $165&255; //@line 200 "load_far.cpp"
          $167 = ((($m$032)) + 5|0); //@line 200 "load_far.cpp"
          HEAP8[$167>>0] = $166; //@line 200 "load_far.cpp"
          $patbrk$1 = $patbrk$033;
          break;
         }
         case 160:  {
          $168 = ((($m$032)) + 2|0); //@line 204 "load_far.cpp"
          HEAP8[$168>>0] = 1; //@line 204 "load_far.cpp"
          $169 = $130 << 2; //@line 205 "load_far.cpp"
          $170 = $169 & 60; //@line 205 "load_far.cpp"
          $171 = (($170) + 4)|0; //@line 205 "load_far.cpp"
          $172 = $171&255; //@line 205 "load_far.cpp"
          $173 = ((($m$032)) + 4|0); //@line 205 "load_far.cpp"
          HEAP8[$173>>0] = $172; //@line 205 "load_far.cpp"
          $patbrk$1 = $patbrk$033;
          break;
         }
         case 176:  {
          $174 = ((($m$032)) + 3|0); //@line 209 "load_far.cpp"
          HEAP8[$174>>0] = 9; //@line 209 "load_far.cpp"
          $175 = $130 << 4; //@line 210 "load_far.cpp"
          $176 = $175&255; //@line 210 "load_far.cpp"
          $177 = ((($m$032)) + 5|0); //@line 210 "load_far.cpp"
          HEAP8[$177>>0] = $176; //@line 210 "load_far.cpp"
          $patbrk$1 = $patbrk$033;
          break;
         }
         case 240:  {
          $178 = ((($m$032)) + 3|0); //@line 214 "load_far.cpp"
          HEAP8[$178>>0] = 16; //@line 214 "load_far.cpp"
          $179 = $130 & 15; //@line 215 "load_far.cpp"
          $180 = $179&255; //@line 215 "load_far.cpp"
          $181 = ((($m$032)) + 5|0); //@line 215 "load_far.cpp"
          HEAP8[$181>>0] = $180; //@line 215 "load_far.cpp"
          $patbrk$1 = $patbrk$033;
          break;
         }
         default: {
          $182 = ($patbrk$033|0)==(0); //@line 218 "load_far.cpp"
          if ($182) {
           $patbrk$1 = 0;
           break L60;
          }
          $183 = (($patbrk$033) + 1)|0; //@line 218 "load_far.cpp"
          $184 = $len$034 >>> 6; //@line 218 "load_far.cpp"
          $185 = ($183|0)!=($184|0); //@line 218 "load_far.cpp"
          $186 = ($183|0)==($102|0); //@line 218 "load_far.cpp"
          $or$cond23 = $186 | $185; //@line 218 "load_far.cpp"
          if ($or$cond23) {
           $patbrk$1 = $patbrk$033;
           break L60;
          }
          $187 = ((($m$032)) + 3|0); //@line 220 "load_far.cpp"
          HEAP8[$187>>0] = 14; //@line 220 "load_far.cpp"
          $patbrk$1 = 0;
         }
         }
        } while(0);
        $188 = (($len$034) + 4)|0; //@line 143 "load_far.cpp"
        $189 = ((($m$032)) + 6|0); //@line 143 "load_far.cpp"
        $190 = ($188>>>0)<($$20>>>0); //@line 143 "load_far.cpp"
        if ($190) {
         $len$034 = $188;$m$032 = $189;$patbrk$033 = $patbrk$1;
        } else {
         $dwMemPos$2 = $83;
         break L45;
        }
       }
      }
     }
    }
   }
  } while(0);
  $191 = (($ipat$035) + 1)|0; //@line 119 "load_far.cpp"
  $192 = ($191>>>0)<(256); //@line 119 "load_far.cpp"
  if ($192) {
   $dwMemPos$136 = $dwMemPos$2;$ipat$035 = $191;
  } else {
   $dwMemPos$2$lcssa = $dwMemPos$2;
   label = 48;
   break;
  }
 }
 if ((label|0) == 48) {
  $193 = (($dwMemPos$2$lcssa) + 8)|0; //@line 228 "load_far.cpp"
  $194 = ($193>>>0)<($dwMemLength>>>0); //@line 228 "load_far.cpp"
  if (!($194)) {
   $$0 = 1;
   STACKTOP = sp;return ($$0|0); //@line 264 "load_far.cpp"
  }
  $195 = (($lpStream) + ($dwMemPos$2$lcssa)|0); //@line 229 "load_far.cpp"
  $196 = $195; //@line 229 "load_far.cpp"
  $197 = $196; //@line 229 "load_far.cpp"
  $198 = HEAPU8[$197>>0]|(HEAPU8[$197+1>>0]<<8)|(HEAPU8[$197+2>>0]<<16)|(HEAPU8[$197+3>>0]<<24); //@line 229 "load_far.cpp"
  $199 = (($196) + 4)|0; //@line 229 "load_far.cpp"
  $200 = $199; //@line 229 "load_far.cpp"
  $201 = HEAPU8[$200>>0]|(HEAPU8[$200+1>>0]<<8)|(HEAPU8[$200+2>>0]<<16)|(HEAPU8[$200+3>>0]<<24); //@line 229 "load_far.cpp"
  $202 = $samplemap; //@line 229 "load_far.cpp"
  $203 = $202; //@line 229 "load_far.cpp"
  HEAP32[$203>>2] = $198; //@line 229 "load_far.cpp"
  $204 = (($202) + 4)|0; //@line 229 "load_far.cpp"
  $205 = $204; //@line 229 "load_far.cpp"
  HEAP32[$205>>2] = $201; //@line 229 "load_far.cpp"
  $206 = ((($this)) + 33344|0); //@line 231 "load_far.cpp"
  $dwMemPos$331 = $193;$ismp$030 = 0;$pins$029 = $206;
  while(1) {
   $207 = $ismp$030 >>> 3; //@line 232 "load_far.cpp"
   $208 = (($samplemap) + ($207)|0); //@line 232 "load_far.cpp"
   $209 = HEAP8[$208>>0]|0; //@line 232 "load_far.cpp"
   $210 = $209&255; //@line 232 "load_far.cpp"
   $211 = $ismp$030 & 7; //@line 232 "load_far.cpp"
   $212 = 1 << $211; //@line 232 "load_far.cpp"
   $213 = $210 & $212; //@line 232 "load_far.cpp"
   $214 = ($213|0)==(0); //@line 232 "load_far.cpp"
   if ($214) {
    $$pre = (($ismp$030) + 1)|0; //@line 232 "load_far.cpp"
    $$pre$phiZ2D = $$pre;$dwMemPos$4 = $dwMemPos$331;
   } else {
    $215 = (($dwMemPos$331) + 48)|0; //@line 234 "load_far.cpp"
    $216 = ($215>>>0)>($dwMemLength>>>0); //@line 234 "load_far.cpp"
    if ($216) {
     $$0 = 1;
     label = 61;
     break;
    }
    $217 = (($lpStream) + ($dwMemPos$331)|0); //@line 235 "load_far.cpp"
    $218 = (($ismp$030) + 1)|0; //@line 237 "load_far.cpp"
    HEAP32[$26>>2] = $218; //@line 237 "load_far.cpp"
    $219 = (((($this)) + 59804|0) + ($218<<5)|0); //@line 238 "load_far.cpp"
    dest=$219; src=$217; stop=dest+32|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 238 "load_far.cpp"
    $$sum5 = (($dwMemPos$331) + 32)|0; //@line 239 "load_far.cpp"
    $220 = (($lpStream) + ($$sum5)|0); //@line 239 "load_far.cpp"
    $221 = HEAPU8[$220>>0]|(HEAPU8[$220+1>>0]<<8)|(HEAPU8[$220+2>>0]<<16)|(HEAPU8[$220+3>>0]<<24); //@line 239 "load_far.cpp"
    HEAP32[$pins$029>>2] = $221; //@line 240 "load_far.cpp"
    $$sum6 = (($dwMemPos$331) + 38)|0; //@line 241 "load_far.cpp"
    $222 = (($lpStream) + ($$sum6)|0); //@line 241 "load_far.cpp"
    $223 = HEAPU8[$222>>0]|(HEAPU8[$222+1>>0]<<8)|(HEAPU8[$222+2>>0]<<16)|(HEAPU8[$222+3>>0]<<24); //@line 241 "load_far.cpp"
    $224 = ((($pins$029)) + 4|0); //@line 241 "load_far.cpp"
    HEAP32[$224>>2] = $223; //@line 241 "load_far.cpp"
    $$sum7 = (($dwMemPos$331) + 42)|0; //@line 242 "load_far.cpp"
    $225 = (($lpStream) + ($$sum7)|0); //@line 242 "load_far.cpp"
    $226 = HEAPU8[$225>>0]|(HEAPU8[$225+1>>0]<<8)|(HEAPU8[$225+2>>0]<<16)|(HEAPU8[$225+3>>0]<<24); //@line 242 "load_far.cpp"
    $227 = ((($pins$029)) + 8|0); //@line 242 "load_far.cpp"
    HEAP32[$227>>2] = $226; //@line 242 "load_far.cpp"
    $228 = ((($pins$029)) + 37|0); //@line 243 "load_far.cpp"
    HEAP8[$228>>0] = 0; //@line 243 "load_far.cpp"
    $229 = ((($pins$029)) + 24|0); //@line 244 "load_far.cpp"
    HEAP32[$229>>2] = 16726; //@line 244 "load_far.cpp"
    $230 = ((($pins$029)) + 32|0); //@line 245 "load_far.cpp"
    HEAP16[$230>>1] = 64; //@line 245 "load_far.cpp"
    $$sum8 = (($dwMemPos$331) + 37)|0; //@line 246 "load_far.cpp"
    $231 = (($lpStream) + ($$sum8)|0); //@line 246 "load_far.cpp"
    $232 = HEAP8[$231>>0]|0; //@line 246 "load_far.cpp"
    $233 = $232&255; //@line 246 "load_far.cpp"
    $234 = $233 << 4; //@line 246 "load_far.cpp"
    $235 = $234&65535; //@line 246 "load_far.cpp"
    $236 = ((($pins$029)) + 30|0); //@line 246 "load_far.cpp"
    HEAP16[$236>>1] = $235; //@line 246 "load_far.cpp"
    $237 = ((($pins$029)) + 34|0); //@line 247 "load_far.cpp"
    HEAP16[$237>>1] = 0; //@line 247 "load_far.cpp"
    $238 = ($221>>>0)>(3); //@line 248 "load_far.cpp"
    $239 = (($dwMemPos$331) + 52)|0;
    $240 = ($239>>>0)<($dwMemLength>>>0); //@line 248 "load_far.cpp"
    $or$cond25 = $240 & $238; //@line 248 "load_far.cpp"
    if ($or$cond25) {
     $$sum9 = (($dwMemPos$331) + 46)|0; //@line 250 "load_far.cpp"
     $241 = (($lpStream) + ($$sum9)|0); //@line 250 "load_far.cpp"
     $242 = HEAP8[$241>>0]|0; //@line 250 "load_far.cpp"
     $243 = $242 & 1; //@line 250 "load_far.cpp"
     $244 = ($243<<24>>24)==(0); //@line 250 "load_far.cpp"
     if ($244) {
      $253 = $226;$255 = 0;
     } else {
      HEAP16[$237>>1] = 1; //@line 252 "load_far.cpp"
      $245 = $221 >>> 1; //@line 253 "load_far.cpp"
      HEAP32[$pins$029>>2] = $245; //@line 253 "load_far.cpp"
      $246 = $223 >>> 1; //@line 254 "load_far.cpp"
      HEAP32[$224>>2] = $246; //@line 254 "load_far.cpp"
      $247 = $226 >>> 1; //@line 255 "load_far.cpp"
      HEAP32[$227>>2] = $247; //@line 255 "load_far.cpp"
      $253 = $247;$255 = 1;
     }
     $$sum10 = (($dwMemPos$331) + 47)|0; //@line 257 "load_far.cpp"
     $248 = (($lpStream) + ($$sum10)|0); //@line 257 "load_far.cpp"
     $249 = HEAP8[$248>>0]|0; //@line 257 "load_far.cpp"
     $250 = $249 & 8; //@line 257 "load_far.cpp"
     $251 = ($250<<24>>24)!=(0); //@line 257 "load_far.cpp"
     $252 = ($253>>>0)>(4); //@line 257 "load_far.cpp"
     $or$cond41 = $251 & $252; //@line 257 "load_far.cpp"
     if ($or$cond41) {
      $254 = $255&65535; //@line 257 "load_far.cpp"
      $256 = $254 | 2; //@line 257 "load_far.cpp"
      $257 = $256&65535; //@line 257 "load_far.cpp"
      HEAP16[$237>>1] = $257; //@line 257 "load_far.cpp"
      $259 = $257;
     } else {
      $259 = $255;
     }
     $258 = $259 & 1; //@line 258 "load_far.cpp"
     $260 = ($258<<16>>16)!=(0); //@line 258 "load_far.cpp"
     $261 = $260 ? 5 : 0; //@line 258 "load_far.cpp"
     $262 = (($lpStream) + ($215)|0); //@line 259 "load_far.cpp"
     $263 = (($dwMemLength) - ($215))|0; //@line 259 "load_far.cpp"
     (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$pins$029,$261,$262,$263)|0); //@line 258 "load_far.cpp"
    }
    $264 = (($221) + ($215))|0; //@line 261 "load_far.cpp"
    $$pre$phiZ2D = $218;$dwMemPos$4 = $264;
   }
   $265 = ((($pins$029)) + 64|0); //@line 232 "load_far.cpp"
   $266 = ($$pre$phiZ2D>>>0)<(64); //@line 232 "load_far.cpp"
   if ($266) {
    $dwMemPos$331 = $dwMemPos$4;$ismp$030 = $$pre$phiZ2D;$pins$029 = $265;
   } else {
    $$0 = 1;
    label = 61;
    break;
   }
  }
  if ((label|0) == 61) {
   STACKTOP = sp;return ($$0|0); //@line 264 "load_far.cpp"
  }
 }
 else if ((label|0) == 61) {
  STACKTOP = sp;return ($$0|0); //@line 264 "load_far.cpp"
 }
 return (0)|0;
}
function __ZN10CSoundFile7ReadAMSEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$0 = 0, $$17 = 0, $$18 = 0, $$lcssa109 = 0, $$lcssa110 = 0, $$lcssa111 = 0, $$lobit = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre68 = 0, $$pre69 = 0, $$pre70 = 0, $$pre71 = 0, $$sum = 0, $$sum10 = 0, $$sum11 = 0, $$sum12 = 0, $$sum2 = 0, $$sum3 = 0;
 var $$sum4 = 0, $$sum5 = 0, $$sum6 = 0, $$sum7 = 0, $$sum8 = 0, $$sum9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0;
 var $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0;
 var $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0;
 var $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0;
 var $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $b0$0 = 0, $b1$0 = 0, $b1$1 = 0, $cNam$046 = 0, $command$0 = 0, $dwMemPos$057 = 0, $dwMemPos$057$lcssa = 0, $dwMemPos$1$lcssa72 = 0, $dwMemPos$152 = 0, $dwMemPos$2$lcssa = 0, $dwMemPos$247 = 0, $dwMemPos$3$lcssa = 0;
 var $dwMemPos$343 = 0, $dwMemPos$5$lcssa = 0, $dwMemPos$5$ph = 0, $dwMemPos$539 = 0, $dwMemPos$6$lcssa = 0, $dwMemPos$636 = 0, $dwMemPos$732 = 0, $dwMemPos$8 = 0, $i$0 = 0, $i$0$ph = 0, $i$1 = 0, $i$2 = 0, $i$3 = 0, $i$4 = 0, $i$4$lcssa = 0, $iOrd$038 = 0, $iPat$035 = 0, $iSmp$033 = 0, $m$0$ph = 0, $nSmp$056 = 0;
 var $or$cond = 0, $or$cond13 = 0, $or$cond14 = 0, $or$cond16 = 0, $or$cond84 = 0, $pNam$042 = 0, $param$0 = 0, $row$0$ph = 0, $sNam$051 = 0, $storemerge = 0, $storemerge85 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($lpStream|0)==(0|0); //@line 53 "load_ams.cpp"
 $1 = ($dwMemLength>>>0)<(1024); //@line 53 "load_ams.cpp"
 $or$cond = $0 | $1; //@line 53 "load_ams.cpp"
 if ($or$cond) {
  $$0 = 0;
  return ($$0|0); //@line 246 "load_ams.cpp"
 }
 $2 = ((($lpStream)) + 8|0); //@line 54 "load_ams.cpp"
 $3 = HEAP8[$2>>0]|0; //@line 54 "load_ams.cpp"
 $4 = ($3<<24>>24)==(1); //@line 54 "load_ams.cpp"
 if ($4) {
  $5 = (_strncmp($lpStream,529853,7)|0); //@line 54 "load_ams.cpp"
  $6 = ($5|0)==(0); //@line 54 "load_ams.cpp"
  if ($6) {
   $7 = ((($lpStream)) + 11|0); //@line 55 "load_ams.cpp"
   $8 = HEAPU8[$7>>0]|(HEAPU8[$7+1>>0]<<8); //@line 55 "load_ams.cpp"
   $9 = ($8<<16>>16)==(0); //@line 55 "load_ams.cpp"
   if (!($9)) {
    $10 = ((($lpStream)) + 13|0); //@line 55 "load_ams.cpp"
    $11 = HEAPU8[$10>>0]|(HEAPU8[$10+1>>0]<<8); //@line 55 "load_ams.cpp"
    $12 = ($11<<16>>16)==(0); //@line 55 "load_ams.cpp"
    if (!($12)) {
     $13 = ((($lpStream)) + 10|0); //@line 55 "load_ams.cpp"
     $14 = HEAP8[$13>>0]|0; //@line 55 "load_ams.cpp"
     $15 = ($14<<24>>24)==(0); //@line 55 "load_ams.cpp"
     if (!($15)) {
      $16 = ($14&255)>(239); //@line 55 "load_ams.cpp"
      $17 = ($8&65535)>(240); //@line 56 "load_ams.cpp"
      $or$cond13 = $17 | $16; //@line 56 "load_ams.cpp"
      $18 = ($11&65535)>(256); //@line 56 "load_ams.cpp"
      $or$cond14 = $18 | $or$cond13; //@line 56 "load_ams.cpp"
      if (!($or$cond14)) {
       $20 = $14&255; //@line 55 "load_ams.cpp"
       $21 = ((($lpStream)) + 16|0); //@line 60 "load_ams.cpp"
       $22 = HEAPU8[$21>>0]|(HEAPU8[$21+1>>0]<<8); //@line 60 "load_ams.cpp"
       $23 = $22&65535; //@line 60 "load_ams.cpp"
       $24 = (($23) + 18)|0; //@line 60 "load_ams.cpp"
       $25 = ($20*17)|0; //@line 61 "load_ams.cpp"
       $26 = (($25) + 256)|0; //@line 61 "load_ams.cpp"
       $27 = (($26) + ($24))|0; //@line 61 "load_ams.cpp"
       $28 = ($27>>>0)<($dwMemLength>>>0); //@line 61 "load_ams.cpp"
       if (!($28)) {
        $$0 = 0;
        return ($$0|0); //@line 246 "load_ams.cpp"
       }
       $29 = ((($this)) + 59680|0); //@line 62 "load_ams.cpp"
       HEAP32[$29>>2] = 4096; //@line 62 "load_ams.cpp"
       $30 = ((($this)) + 59688|0); //@line 63 "load_ams.cpp"
       HEAP32[$30>>2] = 0; //@line 63 "load_ams.cpp"
       $31 = ((($lpStream)) + 9|0); //@line 64 "load_ams.cpp"
       $32 = HEAP8[$31>>0]|0; //@line 64 "load_ams.cpp"
       $33 = $32&255; //@line 64 "load_ams.cpp"
       $34 = $33 & 31; //@line 64 "load_ams.cpp"
       $35 = (($34) + 1)|0; //@line 64 "load_ams.cpp"
       $36 = ((($this)) + 59664|0); //@line 64 "load_ams.cpp"
       HEAP32[$36>>2] = $35; //@line 64 "load_ams.cpp"
       $37 = ((($this)) + 59684|0); //@line 65 "load_ams.cpp"
       HEAP32[$37>>2] = $20; //@line 65 "load_ams.cpp"
       $dwMemPos$057 = $24;$nSmp$056 = 1;
       while(1) {
        $38 = (($lpStream) + ($dwMemPos$057)|0); //@line 68 "load_ams.cpp"
        $39 = HEAPU8[$38>>0]|(HEAPU8[$38+1>>0]<<8)|(HEAPU8[$38+2>>0]<<16)|(HEAPU8[$38+3>>0]<<24); //@line 70 "load_ams.cpp"
        $40 = (((($this)) + 33280|0) + ($nSmp$056<<6)|0); //@line 70 "load_ams.cpp"
        HEAP32[$40>>2] = $39; //@line 70 "load_ams.cpp"
        $$sum7 = (($dwMemPos$057) + 4)|0; //@line 71 "load_ams.cpp"
        $41 = (($lpStream) + ($$sum7)|0); //@line 71 "load_ams.cpp"
        $42 = HEAPU8[$41>>0]|(HEAPU8[$41+1>>0]<<8)|(HEAPU8[$41+2>>0]<<16)|(HEAPU8[$41+3>>0]<<24); //@line 71 "load_ams.cpp"
        $43 = (((((($this)) + 33280|0) + ($nSmp$056<<6)|0)) + 4|0); //@line 71 "load_ams.cpp"
        HEAP32[$43>>2] = $42; //@line 71 "load_ams.cpp"
        $$sum8 = (($dwMemPos$057) + 8)|0; //@line 72 "load_ams.cpp"
        $44 = (($lpStream) + ($$sum8)|0); //@line 72 "load_ams.cpp"
        $45 = HEAPU8[$44>>0]|(HEAPU8[$44+1>>0]<<8)|(HEAPU8[$44+2>>0]<<16)|(HEAPU8[$44+3>>0]<<24); //@line 72 "load_ams.cpp"
        $46 = (((((($this)) + 33280|0) + ($nSmp$056<<6)|0)) + 8|0); //@line 72 "load_ams.cpp"
        HEAP32[$46>>2] = $45; //@line 72 "load_ams.cpp"
        $47 = (((((($this)) + 33280|0) + ($nSmp$056<<6)|0)) + 32|0); //@line 73 "load_ams.cpp"
        HEAP16[$47>>1] = 64; //@line 73 "load_ams.cpp"
        $$sum9 = (($dwMemPos$057) + 15)|0; //@line 74 "load_ams.cpp"
        $48 = (($lpStream) + ($$sum9)|0); //@line 74 "load_ams.cpp"
        $49 = HEAP8[$48>>0]|0; //@line 74 "load_ams.cpp"
        $50 = $49&255; //@line 74 "load_ams.cpp"
        $51 = $50 << 1; //@line 74 "load_ams.cpp"
        $52 = $51&65535; //@line 74 "load_ams.cpp"
        $53 = (((((($this)) + 33280|0) + ($nSmp$056<<6)|0)) + 30|0); //@line 74 "load_ams.cpp"
        HEAP16[$53>>1] = $52; //@line 74 "load_ams.cpp"
        $$sum10 = (($dwMemPos$057) + 13)|0; //@line 75 "load_ams.cpp"
        $54 = (($lpStream) + ($$sum10)|0); //@line 75 "load_ams.cpp"
        $55 = HEAPU8[$54>>0]|(HEAPU8[$54+1>>0]<<8); //@line 75 "load_ams.cpp"
        $56 = $55&65535; //@line 75 "load_ams.cpp"
        $57 = (((((($this)) + 33280|0) + ($nSmp$056<<6)|0)) + 24|0); //@line 75 "load_ams.cpp"
        HEAP32[$57>>2] = $56; //@line 75 "load_ams.cpp"
        $$sum11 = (($dwMemPos$057) + 12)|0; //@line 76 "load_ams.cpp"
        $58 = (($lpStream) + ($$sum11)|0); //@line 76 "load_ams.cpp"
        $59 = HEAP8[$58>>0]|0; //@line 76 "load_ams.cpp"
        $60 = $59&255; //@line 76 "load_ams.cpp"
        $61 = $60 & 240; //@line 76 "load_ams.cpp"
        $62 = (((((($this)) + 33280|0) + ($nSmp$056<<6)|0)) + 28|0); //@line 76 "load_ams.cpp"
        $63 = ($61>>>0)<(128); //@line 77 "load_ams.cpp"
        $64 = (($61) + 16)|0; //@line 77 "load_ams.cpp"
        $$ = $63 ? $64 : $61; //@line 77 "load_ams.cpp"
        $storemerge = $$&65535; //@line 77 "load_ams.cpp"
        HEAP16[$62>>1] = $storemerge; //@line 76 "load_ams.cpp"
        $65 = $60 << 4; //@line 78 "load_ams.cpp"
        $66 = $65&255; //@line 78 "load_ams.cpp"
        $67 = (((((($this)) + 33280|0) + ($nSmp$056<<6)|0)) + 37|0); //@line 78 "load_ams.cpp"
        HEAP8[$67>>0] = $66; //@line 78 "load_ams.cpp"
        $$sum12 = (($dwMemPos$057) + 16)|0; //@line 79 "load_ams.cpp"
        $68 = (($lpStream) + ($$sum12)|0); //@line 79 "load_ams.cpp"
        $69 = HEAP8[$68>>0]|0; //@line 79 "load_ams.cpp"
        $$lobit = ($69&255) >>> 7; //@line 79 "load_ams.cpp"
        $70 = $$lobit&255; //@line 79 "load_ams.cpp"
        $71 = (((((($this)) + 33280|0) + ($nSmp$056<<6)|0)) + 34|0); //@line 79 "load_ams.cpp"
        $72 = ($45>>>0)>($39>>>0); //@line 80 "load_ams.cpp"
        $73 = (($42) + 4)|0;
        $74 = ($73>>>0)>($45>>>0); //@line 80 "load_ams.cpp"
        $or$cond84 = $72 | $74; //@line 80 "load_ams.cpp"
        if ($or$cond84) {
         $storemerge85 = $70;
        } else {
         $75 = $$lobit&255; //@line 80 "load_ams.cpp"
         $76 = $75 | 2; //@line 80 "load_ams.cpp"
         $77 = $76&65535; //@line 80 "load_ams.cpp"
         $storemerge85 = $77;
        }
        HEAP16[$71>>1] = $storemerge85; //@line 79 "load_ams.cpp"
        $78 = (($nSmp$056) + 1)|0; //@line 66 "load_ams.cpp"
        $79 = (($dwMemPos$057) + 17)|0; //@line 66 "load_ams.cpp"
        $80 = HEAP32[$37>>2]|0; //@line 66 "load_ams.cpp"
        $81 = ($78>>>0)>($80>>>0); //@line 66 "load_ams.cpp"
        if ($81) {
         $$lcssa110 = $79;$$lcssa111 = $80;$dwMemPos$057$lcssa = $dwMemPos$057;
         break;
        } else {
         $dwMemPos$057 = $79;$nSmp$056 = $78;
        }
       }
       $$phi$trans$insert = (($lpStream) + ($$lcssa110)|0);
       $$pre = HEAP8[$$phi$trans$insert>>0]|0; //@line 84 "load_ams.cpp"
       $82 = (($dwMemPos$057$lcssa) + 18)|0; //@line 84 "load_ams.cpp"
       $83 = $$pre&255; //@line 84 "load_ams.cpp"
       $84 = (($83) + ($82))|0; //@line 85 "load_ams.cpp"
       $85 = (($84) + 1)|0; //@line 85 "load_ams.cpp"
       $86 = ($85>>>0)<($dwMemLength>>>0); //@line 85 "load_ams.cpp"
       if (!($86)) {
        $$0 = 1;
        return ($$0|0); //@line 246 "load_ams.cpp"
       }
       $87 = ($83>>>0)>(31); //@line 86 "load_ams.cpp"
       $88 = $87 ? 31 : $83; //@line 86 "load_ams.cpp"
       $89 = ($88|0)==(0); //@line 87 "load_ams.cpp"
       if (!($89)) {
        $90 = ((($this)) + 59804|0); //@line 87 "load_ams.cpp"
        $91 = (($lpStream) + ($82)|0); //@line 87 "load_ams.cpp"
        _memcpy(($90|0),($91|0),($88|0))|0; //@line 87 "load_ams.cpp"
       }
       $92 = (((($this)) + 59804|0) + ($88)|0); //@line 88 "load_ams.cpp"
       HEAP8[$92>>0] = 0; //@line 88 "load_ams.cpp"
       $93 = ($$lcssa111|0)==(0); //@line 91 "load_ams.cpp"
       do {
        if ($93) {
         $119 = $35;$dwMemPos$1$lcssa72 = $84;
         label = 19;
        } else {
         $299 = $$lcssa111;$dwMemPos$152 = $84;$sNam$051 = 1;
         while(1) {
          $95 = (($dwMemPos$152) + 32)|0; //@line 93 "load_ams.cpp"
          $96 = ($95>>>0)<($dwMemLength>>>0); //@line 93 "load_ams.cpp"
          if (!($96)) {
           $$0 = 1;
           label = 86;
           break;
          }
          $97 = (($dwMemPos$152) + 1)|0; //@line 94 "load_ams.cpp"
          $98 = (($lpStream) + ($dwMemPos$152)|0); //@line 94 "load_ams.cpp"
          $99 = HEAP8[$98>>0]|0; //@line 94 "load_ams.cpp"
          $100 = $99&255; //@line 94 "load_ams.cpp"
          $101 = ($100>>>0)>(31); //@line 95 "load_ams.cpp"
          $102 = $101 ? 31 : $100; //@line 95 "load_ams.cpp"
          $103 = ($102|0)==(0); //@line 96 "load_ams.cpp"
          if ($103) {
           $109 = $299;
          } else {
           $104 = (((($this)) + 59804|0) + ($sNam$051<<5)|0); //@line 96 "load_ams.cpp"
           $105 = (($lpStream) + ($97)|0); //@line 96 "load_ams.cpp"
           _memcpy(($104|0),($105|0),($102|0))|0; //@line 96 "load_ams.cpp"
           $$pre68 = HEAP32[$37>>2]|0; //@line 91 "load_ams.cpp"
           $109 = $$pre68;
          }
          $106 = (($100) + ($97))|0; //@line 97 "load_ams.cpp"
          $107 = (($sNam$051) + 1)|0; //@line 91 "load_ams.cpp"
          $108 = ($107>>>0)>($109>>>0); //@line 91 "load_ams.cpp"
          if ($108) {
           $$lcssa109 = $106;
           label = 18;
           break;
          } else {
           $299 = $109;$dwMemPos$152 = $106;$sNam$051 = $107;
          }
         }
         if ((label|0) == 18) {
          $$pre69 = HEAP32[$36>>2]|0; //@line 100 "load_ams.cpp"
          $94 = ($$pre69|0)==(0); //@line 100 "load_ams.cpp"
          if ($94) {
           $dwMemPos$2$lcssa = $$lcssa109;
           break;
          } else {
           $119 = $$pre69;$dwMemPos$1$lcssa72 = $$lcssa109;
           label = 19;
           break;
          }
         }
         else if ((label|0) == 86) {
          return ($$0|0); //@line 246 "load_ams.cpp"
         }
        }
       } while(0);
       L36: do {
        if ((label|0) == 19) {
         $cNam$046 = 0;$dwMemPos$247 = $dwMemPos$1$lcssa72;
         while(1) {
          $110 = (($dwMemPos$247) + 32)|0; //@line 102 "load_ams.cpp"
          $111 = ($110>>>0)<($dwMemLength>>>0); //@line 102 "load_ams.cpp"
          if (!($111)) {
           $$0 = 1;
           break;
          }
          $112 = (($dwMemPos$247) + 1)|0; //@line 103 "load_ams.cpp"
          $113 = (($lpStream) + ($dwMemPos$247)|0); //@line 103 "load_ams.cpp"
          $114 = HEAP8[$113>>0]|0; //@line 103 "load_ams.cpp"
          $115 = $114&255; //@line 103 "load_ams.cpp"
          $116 = (($112) + ($115))|0; //@line 104 "load_ams.cpp"
          $117 = (($cNam$046) + 1)|0; //@line 100 "load_ams.cpp"
          $118 = ($117>>>0)<($119>>>0); //@line 100 "load_ams.cpp"
          if ($118) {
           $cNam$046 = $117;$dwMemPos$247 = $116;
          } else {
           $dwMemPos$2$lcssa = $116;
           break L36;
          }
         }
         return ($$0|0); //@line 246 "load_ams.cpp"
        }
       } while(0);
       $120 = HEAPU8[$7>>0]|(HEAPU8[$7+1>>0]<<8); //@line 107 "load_ams.cpp"
       $121 = $120&65535; //@line 107 "load_ams.cpp"
       $122 = $121 << 5; //@line 107 "load_ams.cpp"
       $123 = (__Znaj($122)|0); //@line 107 "load_ams.cpp"
       $124 = ((($this)) + 59800|0); //@line 107 "load_ams.cpp"
       HEAP32[$124>>2] = $123; //@line 107 "load_ams.cpp"
       $125 = ((($this)) + 59792|0); //@line 109 "load_ams.cpp"
       HEAP32[$125>>2] = $121; //@line 109 "load_ams.cpp"
       _memset(($123|0),0,($122|0))|0; //@line 110 "load_ams.cpp"
       $126 = ($120<<16>>16)==(0); //@line 111 "load_ams.cpp"
       L43: do {
        if ($126) {
         $dwMemPos$3$lcssa = $dwMemPos$2$lcssa;
        } else {
         $300 = $121;$dwMemPos$343 = $dwMemPos$2$lcssa;$pNam$042 = 0;
         while(1) {
          $127 = (($dwMemPos$343) + 32)|0; //@line 113 "load_ams.cpp"
          $128 = ($127>>>0)<($dwMemLength>>>0); //@line 113 "load_ams.cpp"
          if (!($128)) {
           $$0 = 1;
           break;
          }
          $129 = (($dwMemPos$343) + 1)|0; //@line 114 "load_ams.cpp"
          $130 = (($lpStream) + ($dwMemPos$343)|0); //@line 114 "load_ams.cpp"
          $131 = HEAP8[$130>>0]|0; //@line 114 "load_ams.cpp"
          $132 = $131&255; //@line 114 "load_ams.cpp"
          $133 = ($132>>>0)>(31); //@line 115 "load_ams.cpp"
          $134 = $133 ? 31 : $132; //@line 115 "load_ams.cpp"
          $135 = ($134|0)==(0); //@line 116 "load_ams.cpp"
          if ($135) {
           $143 = $300;
          } else {
           $136 = HEAP32[$124>>2]|0; //@line 116 "load_ams.cpp"
           $137 = $pNam$042 << 5; //@line 116 "load_ams.cpp"
           $138 = (($136) + ($137)|0); //@line 116 "load_ams.cpp"
           $139 = (($lpStream) + ($129)|0); //@line 116 "load_ams.cpp"
           _memcpy(($138|0),($139|0),($134|0))|0; //@line 116 "load_ams.cpp"
           $$pre70 = HEAP32[$125>>2]|0; //@line 111 "load_ams.cpp"
           $143 = $$pre70;
          }
          $140 = (($132) + ($129))|0; //@line 117 "load_ams.cpp"
          $141 = (($pNam$042) + 1)|0; //@line 111 "load_ams.cpp"
          $142 = ($141>>>0)<($143>>>0); //@line 111 "load_ams.cpp"
          if ($142) {
           $300 = $143;$dwMemPos$343 = $140;$pNam$042 = $141;
          } else {
           $dwMemPos$3$lcssa = $140;
           break L43;
          }
         }
         return ($$0|0); //@line 246 "load_ams.cpp"
        }
       } while(0);
       $144 = (($lpStream) + ($dwMemPos$3$lcssa)|0); //@line 120 "load_ams.cpp"
       $145 = HEAP16[$144>>1]|0; //@line 120 "load_ams.cpp"
       $146 = $145&65535; //@line 120 "load_ams.cpp"
       $147 = (($dwMemPos$3$lcssa) + 2)|0; //@line 121 "load_ams.cpp"
       $148 = (($146) + ($147))|0; //@line 122 "load_ams.cpp"
       $149 = ($148>>>0)<($dwMemLength>>>0); //@line 122 "load_ams.cpp"
       if (!($149)) {
        $$0 = 1;
        return ($$0|0); //@line 246 "load_ams.cpp"
       }
       $150 = ($145<<16>>16)==(0); //@line 123 "load_ams.cpp"
       if ($150) {
        $dwMemPos$5$ph = $147;
       } else {
        $151 = (($146) + 1)|0; //@line 125 "load_ams.cpp"
        $152 = (__Znaj($151)|0); //@line 125 "load_ams.cpp"
        $153 = ((($this)) + 59796|0); //@line 125 "load_ams.cpp"
        HEAP32[$153>>2] = $152; //@line 125 "load_ams.cpp"
        _memset(($152|0),0,($151|0))|0; //@line 127 "load_ams.cpp"
        $154 = (($lpStream) + ($147)|0); //@line 128 "load_ams.cpp"
        _memcpy(($152|0),($154|0),($146|0))|0; //@line 128 "load_ams.cpp"
        $dwMemPos$5$ph = $148;
       }
       $155 = HEAPU8[$10>>0]|(HEAPU8[$10+1>>0]<<8); //@line 132 "load_ams.cpp"
       $156 = ($155<<16>>16)==(0); //@line 132 "load_ams.cpp"
       if ($156) {
        $dwMemPos$5$lcssa = $dwMemPos$5$ph;
       } else {
        $dwMemPos$539 = $dwMemPos$5$ph;$iOrd$038 = 0;
        while(1) {
         $159 = (($lpStream) + ($dwMemPos$539)|0); //@line 134 "load_ams.cpp"
         $160 = HEAP16[$159>>1]|0; //@line 134 "load_ams.cpp"
         $161 = $160&255; //@line 135 "load_ams.cpp"
         $162 = (((($this)) + 53344|0) + ($iOrd$038)|0); //@line 135 "load_ams.cpp"
         HEAP8[$162>>0] = $161; //@line 135 "load_ams.cpp"
         $163 = (($iOrd$038) + 1)|0; //@line 132 "load_ams.cpp"
         $164 = (($dwMemPos$539) + 2)|0; //@line 132 "load_ams.cpp"
         $165 = HEAPU8[$10>>0]|(HEAPU8[$10+1>>0]<<8); //@line 132 "load_ams.cpp"
         $166 = $165&65535; //@line 132 "load_ams.cpp"
         $167 = ($163>>>0)<($166>>>0); //@line 132 "load_ams.cpp"
         if ($167) {
          $dwMemPos$539 = $164;$iOrd$038 = $163;
         } else {
          $dwMemPos$5$lcssa = $164;
          break;
         }
        }
       }
       $157 = HEAPU8[$7>>0]|(HEAPU8[$7+1>>0]<<8); //@line 138 "load_ams.cpp"
       $158 = ($157<<16>>16)==(0); //@line 138 "load_ams.cpp"
       L62: do {
        if ($158) {
         $dwMemPos$6$lcssa = $dwMemPos$5$lcssa;
        } else {
         $dwMemPos$636 = $dwMemPos$5$lcssa;$iPat$035 = 0;
         while(1) {
          $171 = (($dwMemPos$636) + 4)|0; //@line 140 "load_ams.cpp"
          $172 = ($171>>>0)<($dwMemLength>>>0); //@line 140 "load_ams.cpp"
          if (!($172)) {
           $$0 = 1;
           label = 86;
           break;
          }
          $173 = (($lpStream) + ($dwMemPos$636)|0); //@line 141 "load_ams.cpp"
          $174 = HEAP32[$173>>2]|0; //@line 141 "load_ams.cpp"
          $175 = ($174>>>0)<($dwMemLength>>>0); //@line 143 "load_ams.cpp"
          if (!($175)) {
           $$0 = 1;
           label = 86;
           break;
          }
          $176 = (($174) + ($171))|0; //@line 143 "load_ams.cpp"
          $177 = ($176>>>0)>($dwMemLength>>>0); //@line 143 "load_ams.cpp"
          if ($177) {
           $$0 = 1;
           label = 86;
           break;
          }
          $178 = (((($this)) + 52864|0) + ($iPat$035<<1)|0); //@line 144 "load_ams.cpp"
          HEAP16[$178>>1] = 64; //@line 144 "load_ams.cpp"
          $179 = HEAP32[$36>>2]|0; //@line 145 "load_ams.cpp"
          $180 = (__ZN10CSoundFile15AllocatePatternEjj(64,$179)|0); //@line 145 "load_ams.cpp"
          $181 = ($180|0)==(0|0); //@line 146 "load_ams.cpp"
          if ($181) {
           $$0 = 1;
           label = 86;
           break;
          }
          $182 = (((($this)) + 51904|0) + ($iPat$035<<2)|0); //@line 147 "load_ams.cpp"
          HEAP32[$182>>2] = $180; //@line 147 "load_ams.cpp"
          $183 = (($dwMemPos$636) + 5)|0; //@line 152 "load_ams.cpp"
          $i$0$ph = 0;$m$0$ph = $180;$row$0$ph = 0;
          L69: while(1) {
           $i$0 = $i$0$ph;
           while(1) {
            $184 = HEAP16[$178>>1]|0; //@line 150 "load_ams.cpp"
            $185 = $184&65535; //@line 150 "load_ams.cpp"
            $186 = ($row$0$ph>>>0)<($185>>>0); //@line 150 "load_ams.cpp"
            $187 = (($i$0) + 2)|0;
            $188 = ($187>>>0)<($174>>>0); //@line 150 "load_ams.cpp"
            $or$cond16 = $188 & $186; //@line 150 "load_ams.cpp"
            if (!($or$cond16)) {
             break L69;
            }
            $$sum = (($i$0) + ($171))|0; //@line 152 "load_ams.cpp"
            $189 = (($lpStream) + ($$sum)|0); //@line 152 "load_ams.cpp"
            $190 = HEAP8[$189>>0]|0; //@line 152 "load_ams.cpp"
            $$sum2 = (($183) + ($i$0))|0; //@line 153 "load_ams.cpp"
            $191 = (($lpStream) + ($$sum2)|0); //@line 153 "load_ams.cpp"
            $192 = HEAP8[$191>>0]|0; //@line 153 "load_ams.cpp"
            $193 = $190&255; //@line 155 "load_ams.cpp"
            $194 = $193 & 63; //@line 155 "load_ams.cpp"
            $195 = $193 & 64; //@line 157 "load_ams.cpp"
            $196 = ($195|0)==(0); //@line 157 "load_ams.cpp"
            do {
             if ($196) {
              $197 = (($i$0) + 3)|0; //@line 159 "load_ams.cpp"
              $$sum3 = (($187) + ($171))|0; //@line 159 "load_ams.cpp"
              $198 = (($lpStream) + ($$sum3)|0); //@line 159 "load_ams.cpp"
              $199 = HEAP8[$198>>0]|0; //@line 159 "load_ams.cpp"
              $200 = HEAP32[$36>>2]|0; //@line 160 "load_ams.cpp"
              $201 = ($194>>>0)<($200>>>0); //@line 160 "load_ams.cpp"
              if ($201) {
               $202 = $192&255; //@line 162 "load_ams.cpp"
               $203 = $202 & 127; //@line 162 "load_ams.cpp"
               $204 = ($203|0)==(0); //@line 162 "load_ams.cpp"
               if (!($204)) {
                $205 = (($203) + 25)|0; //@line 162 "load_ams.cpp"
                $206 = $205&255; //@line 162 "load_ams.cpp"
                $207 = (($m$0$ph) + (($194*6)|0)|0); //@line 162 "load_ams.cpp"
                HEAP8[$207>>0] = $206; //@line 162 "load_ams.cpp"
               }
               $208 = (((($m$0$ph) + (($194*6)|0)|0)) + 1|0); //@line 163 "load_ams.cpp"
               HEAP8[$208>>0] = $199; //@line 163 "load_ams.cpp"
              }
              $209 = ($192<<24>>24)<(0); //@line 165 "load_ams.cpp"
              if (!($209)) {
               $b0$0 = $190;$b1$0 = $192;$i$1 = $197;
               break;
              }
              $210 = $193 | 64; //@line 167 "load_ams.cpp"
              $211 = $210&255; //@line 167 "load_ams.cpp"
              $212 = (($i$0) + 4)|0; //@line 168 "load_ams.cpp"
              $$sum6 = (($197) + ($171))|0; //@line 168 "load_ams.cpp"
              $213 = (($lpStream) + ($$sum6)|0); //@line 168 "load_ams.cpp"
              $214 = HEAP8[$213>>0]|0; //@line 168 "load_ams.cpp"
              $b0$0 = $211;$b1$0 = $214;$i$1 = $212;
             } else {
              $b0$0 = $190;$b1$0 = $192;$i$1 = $187;
             }
            } while(0);
            $215 = $b0$0&255; //@line 172 "load_ams.cpp"
            $216 = $215 & 64; //@line 172 "load_ams.cpp"
            $217 = ($216|0)==(0); //@line 172 "load_ams.cpp"
            L84: do {
             if ($217) {
              $i$4 = $i$1;
             } else {
              $218 = (($m$0$ph) + (($194*6)|0)|0); //@line 218 "load_ams.cpp"
              $219 = (((($m$0$ph) + (($194*6)|0)|0)) + 3|0); //@line 218 "load_ams.cpp"
              $220 = (((($m$0$ph) + (($194*6)|0)|0)) + 5|0); //@line 219 "load_ams.cpp"
              $221 = (((($m$0$ph) + (($194*6)|0)|0)) + 2|0); //@line 190 "load_ams.cpp"
              $222 = (((($m$0$ph) + (($194*6)|0)|0)) + 4|0); //@line 191 "load_ams.cpp"
              $b1$1 = $b1$0;$i$2 = $i$1;
              while(1) {
               $223 = $b1$1&255; //@line 175 "load_ams.cpp"
               $224 = $223 & 64; //@line 175 "load_ams.cpp"
               $225 = ($224|0)==(0); //@line 175 "load_ams.cpp"
               L88: do {
                if ($225) {
                 $230 = (($i$2) + 1)|0; //@line 184 "load_ams.cpp"
                 $$sum4 = (($i$2) + ($171))|0; //@line 184 "load_ams.cpp"
                 $231 = (($lpStream) + ($$sum4)|0); //@line 184 "load_ams.cpp"
                 $232 = HEAP8[$231>>0]|0; //@line 184 "load_ams.cpp"
                 $233 = HEAP32[$36>>2]|0; //@line 185 "load_ams.cpp"
                 $234 = ($194>>>0)<($233>>>0); //@line 185 "load_ams.cpp"
                 if (!($234)) {
                  $i$3 = $230;
                  break;
                 }
                 $235 = $223 & 63; //@line 187 "load_ams.cpp"
                 switch ($235|0) {
                 case 12:  {
                  HEAP8[$221>>0] = 1; //@line 190 "load_ams.cpp"
                  $236 = ($232&255) >>> 1; //@line 191 "load_ams.cpp"
                  HEAP8[$222>>0] = $236; //@line 191 "load_ams.cpp"
                  $i$3 = $230;
                  break L88;
                  break;
                 }
                 case 14:  {
                  $237 = HEAP8[$219>>0]|0; //@line 195 "load_ams.cpp"
                  $238 = ($237<<24>>24)==(0); //@line 195 "load_ams.cpp"
                  if (!($238)) {
                   $i$3 = $230;
                   break L88;
                  }
                  $239 = $232&255; //@line 198 "load_ams.cpp"
                  $240 = $239 & 240; //@line 199 "load_ams.cpp"
                  L96: do {
                   switch ($240|0) {
                   case 0:  {
                    $241 = $239 & 8; //@line 201 "load_ams.cpp"
                    $242 = ($241|0)==(0); //@line 201 "load_ams.cpp"
                    if ($242) {
                     $command$0 = 0;$param$0 = 0;
                     break L96;
                    }
                    $243 = $239 & 7; //@line 201 "load_ams.cpp"
                    $244 = $243 | 144; //@line 201 "load_ams.cpp"
                    $command$0 = 20;$param$0 = $244;
                    break;
                   }
                   case 16:  {
                    $245 = $239 | 240; //@line 202 "load_ams.cpp"
                    $command$0 = 2;$param$0 = $245;
                    break;
                   }
                   case 32:  {
                    $246 = $239 | 240; //@line 203 "load_ams.cpp"
                    $command$0 = 3;$param$0 = $246;
                    break;
                   }
                   case 48:  {
                    $247 = $239 & 15; //@line 204 "load_ams.cpp"
                    $248 = $247 | 16; //@line 204 "load_ams.cpp"
                    $command$0 = 20;$param$0 = $248;
                    break;
                   }
                   case 64:  {
                    $249 = $239 & 15; //@line 205 "load_ams.cpp"
                    $250 = $249 | 48; //@line 205 "load_ams.cpp"
                    $command$0 = 20;$param$0 = $250;
                    break;
                   }
                   case 80:  {
                    $251 = $239 & 15; //@line 206 "load_ams.cpp"
                    $252 = $251 | 32; //@line 206 "load_ams.cpp"
                    $command$0 = 20;$param$0 = $252;
                    break;
                   }
                   case 96:  {
                    $253 = $239 & 15; //@line 207 "load_ams.cpp"
                    $254 = $253 | 176; //@line 207 "load_ams.cpp"
                    $command$0 = 20;$param$0 = $254;
                    break;
                   }
                   case 112:  {
                    $255 = $239 & 15; //@line 208 "load_ams.cpp"
                    $256 = $255 | 64; //@line 208 "load_ams.cpp"
                    $command$0 = 20;$param$0 = $256;
                    break;
                   }
                   case 144:  {
                    $257 = $239 & 15; //@line 209 "load_ams.cpp"
                    $command$0 = 15;$param$0 = $257;
                    break;
                   }
                   case 160:  {
                    $258 = $239 & 15; //@line 210 "load_ams.cpp"
                    $259 = ($258|0)==(0); //@line 210 "load_ams.cpp"
                    if ($259) {
                     $command$0 = 0;$param$0 = 0;
                     break L96;
                    }
                    $260 = $239 << 4; //@line 210 "load_ams.cpp"
                    $261 = $260 | 15; //@line 210 "load_ams.cpp"
                    $command$0 = 11;$param$0 = $261;
                    break;
                   }
                   case 176:  {
                    $262 = $239 & 15; //@line 211 "load_ams.cpp"
                    $263 = ($262|0)==(0); //@line 211 "load_ams.cpp"
                    $264 = $239 | 240; //@line 211 "load_ams.cpp"
                    $$17 = $263 ? 0 : 11; //@line 211 "load_ams.cpp"
                    $$18 = $263 ? 0 : $264; //@line 211 "load_ams.cpp"
                    $command$0 = $$17;$param$0 = $$18;
                    break;
                   }
                   default: {
                    $command$0 = 20;$param$0 = $239;
                   }
                   }
                  } while(0);
                  HEAP8[$219>>0] = $command$0; //@line 213 "load_ams.cpp"
                  $265 = $param$0&255; //@line 214 "load_ams.cpp"
                  HEAP8[$220>>0] = $265; //@line 214 "load_ams.cpp"
                  $i$3 = $230;
                  break L88;
                  break;
                 }
                 default: {
                  $266 = $235&255; //@line 218 "load_ams.cpp"
                  HEAP8[$219>>0] = $266; //@line 218 "load_ams.cpp"
                  HEAP8[$220>>0] = $232; //@line 219 "load_ams.cpp"
                  __ZNK10CSoundFile17ConvertModCommandEP11_MODCOMMAND($this,$218); //@line 220 "load_ams.cpp"
                  $i$3 = $230;
                  break L88;
                 }
                 }
                } else {
                 $226 = HEAP32[$36>>2]|0; //@line 177 "load_ams.cpp"
                 $227 = ($194>>>0)<($226>>>0); //@line 177 "load_ams.cpp"
                 if (!($227)) {
                  $i$3 = $i$2;
                  break;
                 }
                 HEAP8[$221>>0] = 1; //@line 179 "load_ams.cpp"
                 $228 = $223 & 63; //@line 180 "load_ams.cpp"
                 $229 = $228&255; //@line 180 "load_ams.cpp"
                 HEAP8[$222>>0] = $229; //@line 180 "load_ams.cpp"
                 $i$3 = $i$2;
                }
               } while(0);
               $267 = $223 & 128; //@line 224 "load_ams.cpp"
               $268 = ($267|0)==(0); //@line 224 "load_ams.cpp"
               if ($268) {
                $i$4 = $i$3;
                break L84;
               }
               $269 = (($i$3) + 1)|0; //@line 226 "load_ams.cpp"
               $$sum5 = (($i$3) + ($171))|0; //@line 226 "load_ams.cpp"
               $270 = (($lpStream) + ($$sum5)|0); //@line 226 "load_ams.cpp"
               $271 = HEAP8[$270>>0]|0; //@line 226 "load_ams.cpp"
               $272 = ($269>>>0)>($174>>>0); //@line 227 "load_ams.cpp"
               if ($272) {
                $i$4 = $269;
                break;
               } else {
                $b1$1 = $271;$i$2 = $269;
               }
              }
             }
            } while(0);
            $273 = $215 & 128; //@line 230 "load_ams.cpp"
            $274 = ($273|0)==(0); //@line 230 "load_ams.cpp"
            if ($274) {
             $i$0 = $i$4;
            } else {
             $i$4$lcssa = $i$4;
             break;
            }
           }
           $275 = (($row$0$ph) + 1)|0; //@line 232 "load_ams.cpp"
           $276 = HEAP32[$36>>2]|0; //@line 233 "load_ams.cpp"
           $277 = (($m$0$ph) + (($276*6)|0)|0); //@line 233 "load_ams.cpp"
           $i$0$ph = $i$4$lcssa;$m$0$ph = $277;$row$0$ph = $275;
          }
          $278 = (($iPat$035) + 1)|0; //@line 138 "load_ams.cpp"
          $279 = HEAPU8[$7>>0]|(HEAPU8[$7+1>>0]<<8); //@line 138 "load_ams.cpp"
          $280 = $279&65535; //@line 138 "load_ams.cpp"
          $281 = ($278>>>0)<($280>>>0); //@line 138 "load_ams.cpp"
          if ($281) {
           $dwMemPos$636 = $176;$iPat$035 = $278;
          } else {
           $dwMemPos$6$lcssa = $176;
           break L62;
          }
         }
         if ((label|0) == 86) {
          return ($$0|0); //@line 246 "load_ams.cpp"
         }
        }
       } while(0);
       $168 = (($dwMemLength) + -9)|0; //@line 241 "load_ams.cpp"
       $169 = HEAP32[$37>>2]|0; //@line 239 "load_ams.cpp"
       $170 = ($169|0)==(0); //@line 239 "load_ams.cpp"
       if ($170) {
        $$0 = 1;
        return ($$0|0); //@line 246 "load_ams.cpp"
       } else {
        $301 = $169;$dwMemPos$732 = $dwMemPos$6$lcssa;$iSmp$033 = 1;
       }
       while(1) {
        $282 = (((($this)) + 33280|0) + ($iSmp$033<<6)|0); //@line 239 "load_ams.cpp"
        $283 = HEAP32[$282>>2]|0; //@line 239 "load_ams.cpp"
        $284 = ($283|0)==(0); //@line 239 "load_ams.cpp"
        if ($284) {
         $298 = $301;$dwMemPos$8 = $dwMemPos$732;
        } else {
         $285 = ($dwMemPos$732>>>0)<($168>>>0); //@line 241 "load_ams.cpp"
         if (!($285)) {
          $$0 = 1;
          label = 86;
          break;
         }
         $286 = (((((($this)) + 33280|0) + ($iSmp$033<<6)|0)) + 34|0); //@line 242 "load_ams.cpp"
         $287 = HEAP16[$286>>1]|0; //@line 242 "load_ams.cpp"
         $288 = $287 & 1; //@line 242 "load_ams.cpp"
         $289 = $288&65535; //@line 242 "load_ams.cpp"
         $290 = $289 << 2; //@line 242 "load_ams.cpp"
         $291 = $290 | 17; //@line 242 "load_ams.cpp"
         $292 = (($lpStream) + ($dwMemPos$732)|0); //@line 243 "load_ams.cpp"
         $293 = (($dwMemLength) - ($dwMemPos$732))|0; //@line 243 "load_ams.cpp"
         $294 = (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$282,$291,$292,$293)|0); //@line 243 "load_ams.cpp"
         $295 = (($294) + ($dwMemPos$732))|0; //@line 243 "load_ams.cpp"
         $$pre71 = HEAP32[$37>>2]|0; //@line 239 "load_ams.cpp"
         $298 = $$pre71;$dwMemPos$8 = $295;
        }
        $296 = (($iSmp$033) + 1)|0; //@line 239 "load_ams.cpp"
        $297 = ($296>>>0)>($298>>>0); //@line 239 "load_ams.cpp"
        if ($297) {
         $$0 = 1;
         label = 86;
         break;
        } else {
         $301 = $298;$dwMemPos$732 = $dwMemPos$8;$iSmp$033 = $296;
        }
       }
       if ((label|0) == 86) {
        return ($$0|0); //@line 246 "load_ams.cpp"
       }
      }
     }
    }
   }
  }
 }
 $19 = (__ZN10CSoundFile8ReadAMS2EPKhj($this,$lpStream,$dwMemLength)|0); //@line 58 "load_ams.cpp"
 $$0 = $19;
 return ($$0|0); //@line 246 "load_ams.cpp"
}
function __ZN10CSoundFile8ReadAMS2EPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$0 = 0, $$lcssa152 = 0, $$lcssa153 = 0, $$lobit = 0, $$off = 0, $$off25 = 0, $$off35 = 0, $$pre = 0, $$pre$phi123Z2D = 0, $$pre$phi126Z2D = 0, $$pre$phiZ2D = 0, $$pre113 = 0, $$pre115 = 0, $$pre116 = 0, $$pre121 = 0, $$pre122 = 0, $$pre124 = 0, $$pre125 = 0, $$sum = 0;
 var $$sum12 = 0, $$sum13 = 0, $$sum14 = 0, $$sum15 = 0, $$sum16 = 0, $$sum17 = 0, $$sum18 = 0, $$sum19 = 0, $$sum20$pre$phiZZ2D = 0, $$sum21 = 0, $$sum22 = 0, $$sum23 = 0, $$sum24 = 0, $$sum26 = 0, $$sum27 = 0, $$sum28 = 0, $$sum29 = 0, $$sum30 = 0, $$sum31 = 0, $$sum32 = 0;
 var $$sum33 = 0, $$sum34 = 0, $$sum36 = 0, $$sum37 = 0, $$sum38 = 0, $$sum39 = 0, $$sum40 = 0, $$sum42 = 0, $$sum43 = 0, $$sum44 = 0, $$sum45 = 0, $$sum46 = 0, $$sum47 = 0, $$sum48 = 0, $$sum49 = 0, $$sum50 = 0, $$sum51 = 0, $$sum52 = 0, $$sum53 = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0;
 var $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0;
 var $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0;
 var $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0;
 var $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0;
 var $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0;
 var $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0;
 var $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0;
 var $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0;
 var $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $div = 0, $div41 = 0, $dwMemPos$0$lcssa = 0, $dwMemPos$0105 = 0, $dwMemPos$1$lcssa = 0, $dwMemPos$199 = 0, $dwMemPos$287 = 0, $dwMemPos$385 = 0, $dwMemPos$4 = 0, $dwMemPos$4$lcssa = 0, $dwMemPos$5$lcssa = 0;
 var $dwMemPos$581 = 0, $dwMemPos$670 = 0, $dwMemPos$7 = 0, $exitcond = 0, $flags$0 = 0, $i$093 = 0, $i1$086 = 0, $i2$084 = 0, $iSmp$071 = 0, $inotemap$092 = 0, $ipat$079 = 0, $ismp$098 = 0, $ismpmap$089 = 0, $nIns$0104 = 0, $or$cond11 = 0, $or$cond55 = 0, $or$cond7 = 0, $packedsamples = 0, $pos$094 = 0, $pos3$076 = 0;
 var $pos3$1$be = 0, $pos3$173 = 0, $pos3$2 = 0, $row$0$ = 0, $row$077 = 0, $s = 0, $smpmap = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 288|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $smpmap = sp + 272|0;
 $packedsamples = sp + 32|0;
 $s = sp;
 $0 = HEAPU8[$lpStream>>0]|(HEAPU8[$lpStream+1>>0]<<8)|(HEAPU8[$lpStream+2>>0]<<16)|(HEAPU8[$lpStream+3>>0]<<24); //@line 319 "load_ams.cpp"
 $1 = ($0|0)==(1750289729); //@line 319 "load_ams.cpp"
 if (!($1)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 557 "load_ams.cpp"
 }
 $2 = ((($lpStream)) + 4|0); //@line 319 "load_ams.cpp"
 $3 = HEAPU8[$2>>0]|(HEAPU8[$2+1>>0]<<8); //@line 319 "load_ams.cpp"
 $4 = ($3<<16>>16)==(29284); //@line 319 "load_ams.cpp"
 if (!($4)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 557 "load_ams.cpp"
 }
 $5 = ((($lpStream)) + 6|0); //@line 320 "load_ams.cpp"
 $6 = HEAP8[$5>>0]|0; //@line 320 "load_ams.cpp"
 $7 = ($6<<24>>24)==(26); //@line 320 "load_ams.cpp"
 if (!($7)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 557 "load_ams.cpp"
 }
 $8 = ((($lpStream)) + 7|0); //@line 320 "load_ams.cpp"
 $9 = HEAP8[$8>>0]|0; //@line 320 "load_ams.cpp"
 $10 = $9&255; //@line 320 "load_ams.cpp"
 $11 = ($9&255)>(30); //@line 320 "load_ams.cpp"
 if ($11) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 557 "load_ams.cpp"
 }
 $12 = (($10) + 8)|0; //@line 321 "load_ams.cpp"
 $13 = (($lpStream) + ($12)|0); //@line 322 "load_ams.cpp"
 $14 = HEAPU8[$13>>0]|(HEAPU8[$13+1>>0]<<8); //@line 323 "load_ams.cpp"
 $15 = $14 & -256; //@line 323 "load_ams.cpp"
 $16 = ($15<<16>>16)==(512); //@line 323 "load_ams.cpp"
 if (!($16)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 557 "load_ams.cpp"
 }
 $$sum = (($10) + 10)|0; //@line 323 "load_ams.cpp"
 $17 = (($lpStream) + ($$sum)|0); //@line 323 "load_ams.cpp"
 $18 = HEAP8[$17>>0]|0; //@line 323 "load_ams.cpp"
 $19 = (($18) + -1)<<24>>24; //@line 324 "load_ams.cpp"
 $20 = ($19&255)>(238); //@line 324 "load_ams.cpp"
 if ($20) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 557 "load_ams.cpp"
 }
 $$sum12 = (($10) + 11)|0; //@line 324 "load_ams.cpp"
 $21 = (($lpStream) + ($$sum12)|0); //@line 324 "load_ams.cpp"
 $22 = HEAPU8[$21>>0]|(HEAPU8[$21+1>>0]<<8); //@line 324 "load_ams.cpp"
 $23 = ($22<<16>>16)==(0); //@line 324 "load_ams.cpp"
 if ($23) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 557 "load_ams.cpp"
 }
 $$sum13 = (($10) + 13)|0; //@line 324 "load_ams.cpp"
 $24 = (($lpStream) + ($$sum13)|0); //@line 324 "load_ams.cpp"
 $25 = HEAPU8[$24>>0]|(HEAPU8[$24+1>>0]<<8); //@line 324 "load_ams.cpp"
 $26 = ($25<<16>>16)==(0); //@line 324 "load_ams.cpp"
 if ($26) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 557 "load_ams.cpp"
 }
 $27 = (($10) + 23)|0; //@line 325 "load_ams.cpp"
 $28 = ($9<<24>>24)==(0); //@line 326 "load_ams.cpp"
 if ($28) {
  $46 = $18;
 } else {
  $29 = ((($this)) + 59804|0); //@line 328 "load_ams.cpp"
  $30 = ((($lpStream)) + 8|0); //@line 328 "load_ams.cpp"
  _memcpy(($29|0),($30|0),($10|0))|0; //@line 328 "load_ams.cpp"
  $31 = HEAP8[$8>>0]|0; //@line 329 "load_ams.cpp"
  $32 = $31&255; //@line 329 "load_ams.cpp"
  $33 = (((($this)) + 59804|0) + ($32)|0); //@line 329 "load_ams.cpp"
  HEAP8[$33>>0] = 0; //@line 329 "load_ams.cpp"
  $$pre = HEAP8[$17>>0]|0; //@line 335 "load_ams.cpp"
  $46 = $$pre;
 }
 $34 = ((($this)) + 59680|0); //@line 331 "load_ams.cpp"
 HEAP32[$34>>2] = 4096; //@line 331 "load_ams.cpp"
 $35 = ((($this)) + 59664|0); //@line 332 "load_ams.cpp"
 HEAP32[$35>>2] = 32; //@line 332 "load_ams.cpp"
 $$sum14 = (($10) + 15)|0; //@line 333 "load_ams.cpp"
 $36 = (($lpStream) + ($$sum14)|0); //@line 333 "load_ams.cpp"
 $37 = HEAPU8[$36>>0]|(HEAPU8[$36+1>>0]<<8); //@line 333 "load_ams.cpp"
 $38 = $37&65535; //@line 333 "load_ams.cpp"
 $39 = $38 >>> 8; //@line 333 "load_ams.cpp"
 $40 = ((($this)) + 59652|0); //@line 333 "load_ams.cpp"
 HEAP32[$40>>2] = $39; //@line 333 "load_ams.cpp"
 $$sum15 = (($10) + 17)|0; //@line 334 "load_ams.cpp"
 $41 = (($lpStream) + ($$sum15)|0); //@line 334 "load_ams.cpp"
 $42 = HEAP8[$41>>0]|0; //@line 334 "load_ams.cpp"
 $43 = $42&255; //@line 334 "load_ams.cpp"
 $44 = ((($this)) + 59648|0); //@line 334 "load_ams.cpp"
 HEAP32[$44>>2] = $43; //@line 334 "load_ams.cpp"
 $45 = $46&255; //@line 335 "load_ams.cpp"
 $47 = ((($this)) + 59688|0); //@line 335 "load_ams.cpp"
 HEAP32[$47>>2] = $45; //@line 335 "load_ams.cpp"
 $48 = ((($this)) + 59684|0); //@line 336 "load_ams.cpp"
 HEAP32[$48>>2] = 0; //@line 336 "load_ams.cpp"
 $$sum16 = (($10) + 21)|0; //@line 337 "load_ams.cpp"
 $49 = (($lpStream) + ($$sum16)|0); //@line 337 "load_ams.cpp"
 $50 = HEAPU8[$49>>0]|(HEAPU8[$49+1>>0]<<8); //@line 337 "load_ams.cpp"
 $51 = $50 & 64; //@line 337 "load_ams.cpp"
 $52 = ($51<<16>>16)==(0); //@line 337 "load_ams.cpp"
 if (!($52)) {
  $53 = ((($this)) + 59660|0); //@line 337 "load_ams.cpp"
  $54 = HEAP32[$53>>2]|0; //@line 337 "load_ams.cpp"
  $55 = $54 | 16; //@line 337 "load_ams.cpp"
  HEAP32[$53>>2] = $55; //@line 337 "load_ams.cpp"
 }
 $56 = ($46<<24>>24)==(0); //@line 338 "load_ams.cpp"
 L31: do {
  if ($56) {
   $dwMemPos$0$lcssa = $27;
  } else {
   $dwMemPos$0105 = $27;$nIns$0104 = 1;
   while(1) {
    $57 = (($lpStream) + ($dwMemPos$0105)|0); //@line 340 "load_ams.cpp"
    $58 = HEAP8[$57>>0]|0; //@line 340 "load_ams.cpp"
    $59 = $58&255; //@line 340 "load_ams.cpp"
    $$sum28 = (($dwMemPos$0105) + 1)|0; //@line 341 "load_ams.cpp"
    $60 = (($lpStream) + ($$sum28)|0); //@line 341 "load_ams.cpp"
    $61 = (($$sum28) + ($59))|0; //@line 342 "load_ams.cpp"
    $62 = (($lpStream) + ($61)|0); //@line 343 "load_ams.cpp"
    $63 = (($61) + 1145)|0; //@line 345 "load_ams.cpp"
    $64 = ($63>>>0)<($dwMemLength>>>0); //@line 345 "load_ams.cpp"
    if (!($64)) {
     $$0 = 1;
     break;
    }
    $$sum29 = (($61) + 125)|0; //@line 348 "load_ams.cpp"
    $65 = (($lpStream) + ($$sum29)|0); //@line 348 "load_ams.cpp"
    $66 = HEAP8[$65>>0]|0; //@line 348 "load_ams.cpp"
    $67 = $66&255; //@line 348 "load_ams.cpp"
    $68 = ($67*3)|0; //@line 348 "load_ams.cpp"
    $69 = (($68) + ($61))|0; //@line 348 "load_ams.cpp"
    $$sum30 = (($69) + 130)|0; //@line 350 "load_ams.cpp"
    $70 = (($lpStream) + ($$sum30)|0); //@line 350 "load_ams.cpp"
    $71 = HEAP8[$70>>0]|0; //@line 350 "load_ams.cpp"
    $72 = $71&255; //@line 350 "load_ams.cpp"
    $73 = ($72*3)|0; //@line 350 "load_ams.cpp"
    $74 = (($69) + ($73))|0; //@line 350 "load_ams.cpp"
    $$sum31 = (($74) + 135)|0; //@line 352 "load_ams.cpp"
    $75 = (($lpStream) + ($$sum31)|0); //@line 352 "load_ams.cpp"
    $76 = HEAP8[$75>>0]|0; //@line 352 "load_ams.cpp"
    $77 = $76&255; //@line 352 "load_ams.cpp"
    $78 = ($77*3)|0; //@line 352 "load_ams.cpp"
    $79 = (($74) + ($78))|0; //@line 352 "load_ams.cpp"
    $80 = (__Znwj(632)|0); //@line 353 "load_ams.cpp"
    dest=$smpmap; stop=dest+16|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 355 "load_ams.cpp"
    _memset(($80|0),0,632)|0; //@line 356 "load_ams.cpp"
    $81 = HEAP8[$62>>0]|0; //@line 357 "load_ams.cpp"
    $82 = ($81<<24>>24)==(0); //@line 357 "load_ams.cpp"
    L36: do {
     if (!($82)) {
      $$pre113 = HEAP32[$48>>2]|0; //@line 359 "load_ams.cpp"
      $83 = $81&255; //@line 357 "load_ams.cpp"
      $85 = $$pre113;$ismpmap$089 = 0;
      while(1) {
       $84 = (($85) + 1)|0; //@line 359 "load_ams.cpp"
       $86 = ($84>>>0)>(239); //@line 359 "load_ams.cpp"
       if ($86) {
        break L36;
       }
       HEAP32[$48>>2] = $84; //@line 360 "load_ams.cpp"
       $87 = $84&255; //@line 361 "load_ams.cpp"
       $88 = (($smpmap) + ($ismpmap$089)|0); //@line 361 "load_ams.cpp"
       HEAP8[$88>>0] = $87; //@line 361 "load_ams.cpp"
       $89 = (($ismpmap$089) + 1)|0; //@line 357 "load_ams.cpp"
       $90 = ($89>>>0)>=($83>>>0); //@line 357 "load_ams.cpp"
       $91 = ($89>>>0)>(15); //@line 359 "load_ams.cpp"
       $or$cond11 = $91 | $90; //@line 357 "load_ams.cpp"
       if ($or$cond11) {
        break;
       } else {
        $85 = $84;$ismpmap$089 = $89;
       }
      }
     }
    } while(0);
    $92 = ((($80)) + 8|0); //@line 363 "load_ams.cpp"
    HEAP16[$92>>1] = 64; //@line 363 "load_ams.cpp"
    $93 = ((($80)) + 10|0); //@line 364 "load_ams.cpp"
    HEAP16[$93>>1] = 128; //@line 364 "load_ams.cpp"
    $94 = ((($80)) + 584|0); //@line 365 "load_ams.cpp"
    HEAP8[$94>>0] = 60; //@line 365 "load_ams.cpp"
    $95 = (((($this)) + 48640|0) + ($nIns$0104<<2)|0); //@line 366 "load_ams.cpp"
    HEAP32[$95>>2] = $80; //@line 366 "load_ams.cpp"
    $96 = ($58<<24>>24)==(0); //@line 367 "load_ams.cpp"
    if (!($96)) {
     $97 = ($58&255)>(31); //@line 369 "load_ams.cpp"
     $$ = $97 ? 31 : $59; //@line 369 "load_ams.cpp"
     $98 = ((($80)) + 585|0); //@line 370 "load_ams.cpp"
     _memcpy(($98|0),($60|0),($$|0))|0; //@line 370 "load_ams.cpp"
     $$sum53 = (($$) + 585)|0; //@line 371 "load_ams.cpp"
     $99 = (($80) + ($$sum53)|0); //@line 371 "load_ams.cpp"
     HEAP8[$99>>0] = 0; //@line 371 "load_ams.cpp"
    }
    $$sum50 = (($61) + 1)|0; //@line 376 "load_ams.cpp"
    $inotemap$092 = 0;
    while(1) {
     $100 = (($inotemap$092) + 1)|0; //@line 375 "load_ams.cpp"
     $101 = $100&255; //@line 375 "load_ams.cpp"
     $$sum49 = (($inotemap$092) + 428)|0; //@line 375 "load_ams.cpp"
     $102 = (($80) + ($$sum49)|0); //@line 375 "load_ams.cpp"
     HEAP8[$102>>0] = $101; //@line 375 "load_ams.cpp"
     $$sum51 = (($$sum50) + ($inotemap$092))|0; //@line 376 "load_ams.cpp"
     $103 = (($lpStream) + ($$sum51)|0); //@line 376 "load_ams.cpp"
     $104 = HEAP8[$103>>0]|0; //@line 376 "load_ams.cpp"
     $105 = $104&255; //@line 376 "load_ams.cpp"
     $106 = $105 & 15; //@line 376 "load_ams.cpp"
     $107 = (($smpmap) + ($106)|0); //@line 376 "load_ams.cpp"
     $108 = HEAP8[$107>>0]|0; //@line 376 "load_ams.cpp"
     $$sum52 = (($inotemap$092) + 300)|0; //@line 376 "load_ams.cpp"
     $109 = (($80) + ($$sum52)|0); //@line 376 "load_ams.cpp"
     HEAP8[$109>>0] = $108; //@line 376 "load_ams.cpp"
     $exitcond = ($100|0)==(120); //@line 373 "load_ams.cpp"
     if ($exitcond) {
      break;
     } else {
      $inotemap$092 = $100;
     }
    }
    $110 = ($66&255)>(16); //@line 381 "load_ams.cpp"
    $111 = $110 ? 16 : $66; //@line 381 "load_ams.cpp"
    $112 = ((($80)) + 556|0); //@line 381 "load_ams.cpp"
    HEAP8[$112>>0] = $111; //@line 381 "load_ams.cpp"
    $$sum32 = (($61) + 122)|0; //@line 382 "load_ams.cpp"
    $113 = (($lpStream) + ($$sum32)|0); //@line 382 "load_ams.cpp"
    $114 = HEAP8[$113>>0]|0; //@line 382 "load_ams.cpp"
    $115 = ((($80)) + 562|0); //@line 382 "load_ams.cpp"
    HEAP8[$115>>0] = $114; //@line 382 "load_ams.cpp"
    $116 = ((($80)) + 561|0); //@line 382 "load_ams.cpp"
    HEAP8[$116>>0] = $114; //@line 382 "load_ams.cpp"
    $$sum33 = (($61) + 123)|0; //@line 383 "load_ams.cpp"
    $117 = (($lpStream) + ($$sum33)|0); //@line 383 "load_ams.cpp"
    $118 = HEAP8[$117>>0]|0; //@line 383 "load_ams.cpp"
    $119 = ((($80)) + 559|0); //@line 383 "load_ams.cpp"
    HEAP8[$119>>0] = $118; //@line 383 "load_ams.cpp"
    $$sum34 = (($61) + 124)|0; //@line 384 "load_ams.cpp"
    $120 = (($lpStream) + ($$sum34)|0); //@line 384 "load_ams.cpp"
    $121 = HEAP8[$120>>0]|0; //@line 384 "load_ams.cpp"
    $122 = ((($80)) + 560|0); //@line 384 "load_ams.cpp"
    HEAP8[$122>>0] = $121; //@line 384 "load_ams.cpp"
    $123 = ($111<<24>>24)==(0); //@line 385 "load_ams.cpp"
    if (!($123)) {
     $$sum44 = (($61) + 126)|0; //@line 387 "load_ams.cpp"
     $124 = (($61) + 128)|0; //@line 387 "load_ams.cpp"
     $125 = (($61) + 127)|0; //@line 388 "load_ams.cpp"
     $126 = ((($80)) + 12|0); //@line 389 "load_ams.cpp"
     $i$093 = 0;$pos$094 = 0;
     while(1) {
      $127 = ($i$093*3)|0; //@line 387 "load_ams.cpp"
      $$sum45 = (($124) + ($127))|0; //@line 387 "load_ams.cpp"
      $128 = (($lpStream) + ($$sum45)|0); //@line 387 "load_ams.cpp"
      $129 = HEAP8[$128>>0]|0; //@line 387 "load_ams.cpp"
      $130 = $129&255; //@line 387 "load_ams.cpp"
      $131 = $130 >>> 1; //@line 387 "load_ams.cpp"
      $132 = $131 & 63; //@line 387 "load_ams.cpp"
      $133 = $132&255; //@line 387 "load_ams.cpp"
      $$sum46 = (($i$093) + 204)|0; //@line 387 "load_ams.cpp"
      $134 = (($80) + ($$sum46)|0); //@line 387 "load_ams.cpp"
      HEAP8[$134>>0] = $133; //@line 387 "load_ams.cpp"
      $$sum47 = (($127) + ($$sum44))|0; //@line 388 "load_ams.cpp"
      $135 = (($lpStream) + ($$sum47)|0); //@line 388 "load_ams.cpp"
      $136 = HEAP8[$135>>0]|0; //@line 388 "load_ams.cpp"
      $137 = $136&255; //@line 388 "load_ams.cpp"
      $$sum48 = (($125) + ($127))|0; //@line 388 "load_ams.cpp"
      $138 = (($lpStream) + ($$sum48)|0); //@line 388 "load_ams.cpp"
      $139 = HEAP8[$138>>0]|0; //@line 388 "load_ams.cpp"
      $140 = $139&255; //@line 388 "load_ams.cpp"
      $141 = $140 << 8; //@line 388 "load_ams.cpp"
      $142 = $141 & 256; //@line 388 "load_ams.cpp"
      $143 = $142 | $137; //@line 388 "load_ams.cpp"
      $144 = (($143) + ($pos$094))|0; //@line 388 "load_ams.cpp"
      $145 = $144&65535; //@line 389 "load_ams.cpp"
      $146 = (($126) + ($i$093<<1)|0); //@line 389 "load_ams.cpp"
      HEAP16[$146>>1] = $145; //@line 389 "load_ams.cpp"
      $147 = (($i$093) + 1)|0; //@line 385 "load_ams.cpp"
      $148 = HEAP8[$112>>0]|0; //@line 385 "load_ams.cpp"
      $149 = $148&255; //@line 385 "load_ams.cpp"
      $150 = ($147>>>0)<($149>>>0); //@line 385 "load_ams.cpp"
      if ($150) {
       $i$093 = $147;$pos$094 = $144;
      } else {
       break;
      }
     }
    }
    $151 = (($79) + 138)|0; //@line 392 "load_ams.cpp"
    $152 = (($lpStream) + ($151)|0); //@line 392 "load_ams.cpp"
    $153 = HEAP8[$152>>0]|0; //@line 392 "load_ams.cpp"
    $154 = $153&255; //@line 392 "load_ams.cpp"
    $155 = $154 << 8; //@line 392 "load_ams.cpp"
    $156 = $155 & 3840; //@line 392 "load_ams.cpp"
    $157 = (($79) + 137)|0; //@line 392 "load_ams.cpp"
    $158 = (($lpStream) + ($157)|0); //@line 392 "load_ams.cpp"
    $159 = HEAP8[$158>>0]|0; //@line 392 "load_ams.cpp"
    $160 = $159&255; //@line 392 "load_ams.cpp"
    $161 = $156 | $160; //@line 392 "load_ams.cpp"
    $162 = $161 << 3; //@line 392 "load_ams.cpp"
    HEAP32[$80>>2] = $162; //@line 392 "load_ams.cpp"
    $163 = (($79) + 139)|0; //@line 393 "load_ams.cpp"
    $164 = (($lpStream) + ($163)|0); //@line 393 "load_ams.cpp"
    $165 = HEAP8[$164>>0]|0; //@line 393 "load_ams.cpp"
    $166 = $165&255; //@line 393 "load_ams.cpp"
    $167 = $166 & 1; //@line 394 "load_ams.cpp"
    $168 = ($167|0)==(0); //@line 394 "load_ams.cpp"
    if (!($168)) {
     $169 = ((($80)) + 4|0); //@line 394 "load_ams.cpp"
     $170 = HEAP32[$169>>2]|0; //@line 394 "load_ams.cpp"
     $171 = $170 | 4; //@line 394 "load_ams.cpp"
     HEAP32[$169>>2] = $171; //@line 394 "load_ams.cpp"
    }
    $172 = $166 & 2; //@line 395 "load_ams.cpp"
    $173 = ($172|0)==(0); //@line 395 "load_ams.cpp"
    if (!($173)) {
     $174 = ((($80)) + 4|0); //@line 395 "load_ams.cpp"
     $175 = HEAP32[$174>>2]|0; //@line 395 "load_ams.cpp"
     $176 = $175 | 2; //@line 395 "load_ams.cpp"
     HEAP32[$174>>2] = $176; //@line 395 "load_ams.cpp"
    }
    $177 = $166 & 4; //@line 396 "load_ams.cpp"
    $178 = ($177|0)==(0); //@line 396 "load_ams.cpp"
    if (!($178)) {
     $179 = ((($80)) + 4|0); //@line 396 "load_ams.cpp"
     $180 = HEAP32[$179>>2]|0; //@line 396 "load_ams.cpp"
     $181 = $180 | 1; //@line 396 "load_ams.cpp"
     HEAP32[$179>>2] = $181; //@line 396 "load_ams.cpp"
    }
    $182 = (($79) + 141)|0; //@line 397 "load_ams.cpp"
    if ($82) {
     $dwMemPos$1$lcssa = $182;
    } else {
     $dwMemPos$199 = $182;$ismp$098 = 0;
     while(1) {
      $183 = ($ismp$098>>>0)<(16); //@line 401 "load_ams.cpp"
      do {
       if ($183) {
        $184 = (($smpmap) + ($ismp$098)|0); //@line 401 "load_ams.cpp"
        $185 = HEAP8[$184>>0]|0; //@line 401 "load_ams.cpp"
        $186 = ($185<<24>>24)==(0); //@line 401 "load_ams.cpp"
        if ($186) {
         $193 = 0;
         break;
        }
        $187 = $185&255; //@line 401 "load_ams.cpp"
        $188 = (((($this)) + 33280|0) + ($187<<6)|0); //@line 401 "load_ams.cpp"
        $193 = $188;
       } else {
        $193 = 0;
       }
      } while(0);
      $189 = (($lpStream) + ($dwMemPos$199)|0); //@line 402 "load_ams.cpp"
      $190 = HEAP8[$189>>0]|0; //@line 402 "load_ams.cpp"
      $191 = $190&255; //@line 402 "load_ams.cpp"
      $192 = ($193|0)!=(0|0); //@line 403 "load_ams.cpp"
      $$off35 = (($190) + -1)<<24>>24; //@line 403 "load_ams.cpp"
      $194 = ($$off35&255)<(22); //@line 403 "load_ams.cpp"
      $195 = $192 & $194; //@line 403 "load_ams.cpp"
      if ($195) {
       $196 = (($smpmap) + ($ismp$098)|0); //@line 405 "load_ams.cpp"
       $197 = HEAP8[$196>>0]|0; //@line 405 "load_ams.cpp"
       $198 = $197&255; //@line 405 "load_ams.cpp"
       $199 = (((($this)) + 59804|0) + ($198<<5)|0); //@line 405 "load_ams.cpp"
       $$sum43 = (($dwMemPos$199) + 1)|0; //@line 405 "load_ams.cpp"
       $200 = (($lpStream) + ($$sum43)|0); //@line 405 "load_ams.cpp"
       _memcpy(($199|0),($200|0),($191|0))|0; //@line 405 "load_ams.cpp"
       $$pre$phi126Z2D = $$sum43;
      } else {
       $$pre125 = (($dwMemPos$199) + 1)|0; //@line 407 "load_ams.cpp"
       $$pre$phi126Z2D = $$pre125;
      }
      $201 = (($$pre$phi126Z2D) + ($191))|0; //@line 407 "load_ams.cpp"
      do {
       if ($192) {
        $202 = (($lpStream) + ($201)|0); //@line 410 "load_ams.cpp"
        $203 = ((($193)) + 32|0); //@line 411 "load_ams.cpp"
        HEAP16[$203>>1] = 64; //@line 411 "load_ams.cpp"
        $204 = ((($193)) + 28|0); //@line 412 "load_ams.cpp"
        HEAP16[$204>>1] = 128; //@line 412 "load_ams.cpp"
        $205 = HEAPU8[$202>>0]|(HEAPU8[$202+1>>0]<<8)|(HEAPU8[$202+2>>0]<<16)|(HEAPU8[$202+3>>0]<<24); //@line 413 "load_ams.cpp"
        HEAP32[$193>>2] = $205; //@line 413 "load_ams.cpp"
        $$sum36 = (($201) + 4)|0; //@line 414 "load_ams.cpp"
        $206 = (($lpStream) + ($$sum36)|0); //@line 414 "load_ams.cpp"
        $207 = HEAPU8[$206>>0]|(HEAPU8[$206+1>>0]<<8)|(HEAPU8[$206+2>>0]<<16)|(HEAPU8[$206+3>>0]<<24); //@line 414 "load_ams.cpp"
        $208 = ((($193)) + 4|0); //@line 414 "load_ams.cpp"
        HEAP32[$208>>2] = $207; //@line 414 "load_ams.cpp"
        $$sum37 = (($201) + 8)|0; //@line 415 "load_ams.cpp"
        $209 = (($lpStream) + ($$sum37)|0); //@line 415 "load_ams.cpp"
        $210 = HEAPU8[$209>>0]|(HEAPU8[$209+1>>0]<<8)|(HEAPU8[$209+2>>0]<<16)|(HEAPU8[$209+3>>0]<<24); //@line 415 "load_ams.cpp"
        $211 = ((($193)) + 8|0); //@line 415 "load_ams.cpp"
        HEAP32[$211>>2] = $210; //@line 415 "load_ams.cpp"
        $$sum38 = (($201) + 15)|0; //@line 416 "load_ams.cpp"
        $212 = (($lpStream) + ($$sum38)|0); //@line 416 "load_ams.cpp"
        $213 = HEAPU8[$212>>0]|(HEAPU8[$212+1>>0]<<8); //@line 416 "load_ams.cpp"
        $214 = $213&65535; //@line 416 "load_ams.cpp"
        $215 = ((($193)) + 24|0); //@line 416 "load_ams.cpp"
        HEAP32[$215>>2] = $214; //@line 416 "load_ams.cpp"
        $$sum39 = (($201) + 17)|0; //@line 417 "load_ams.cpp"
        $216 = (($lpStream) + ($$sum39)|0); //@line 417 "load_ams.cpp"
        $217 = HEAP8[$216>>0]|0; //@line 417 "load_ams.cpp"
        $218 = ((($193)) + 36|0); //@line 417 "load_ams.cpp"
        HEAP8[$218>>0] = $217; //@line 417 "load_ams.cpp"
        $$sum40 = (($201) + 18)|0; //@line 418 "load_ams.cpp"
        $219 = (($lpStream) + ($$sum40)|0); //@line 418 "load_ams.cpp"
        $220 = HEAP8[$219>>0]|0; //@line 418 "load_ams.cpp"
        $div41 = ($220&255) >>> 1; //@line 418 "load_ams.cpp"
        $221 = $div41&255; //@line 418 "load_ams.cpp"
        $222 = ((($193)) + 30|0); //@line 418 "load_ams.cpp"
        HEAP16[$222>>1] = $221; //@line 418 "load_ams.cpp"
        $$sum42 = (($201) + 19)|0; //@line 419 "load_ams.cpp"
        $223 = (($lpStream) + ($$sum42)|0); //@line 419 "load_ams.cpp"
        $224 = HEAP8[$223>>0]|0; //@line 419 "load_ams.cpp"
        $225 = (($smpmap) + ($ismp$098)|0); //@line 419 "load_ams.cpp"
        $226 = HEAP8[$225>>0]|0; //@line 419 "load_ams.cpp"
        $227 = $226&255; //@line 419 "load_ams.cpp"
        $228 = (($packedsamples) + ($227)|0); //@line 419 "load_ams.cpp"
        HEAP8[$228>>0] = $224; //@line 419 "load_ams.cpp"
        $229 = $224 & 4; //@line 420 "load_ams.cpp"
        $230 = ($229<<24>>24)==(0); //@line 420 "load_ams.cpp"
        if (!($230)) {
         $231 = ((($193)) + 34|0); //@line 420 "load_ams.cpp"
         $232 = HEAP16[$231>>1]|0; //@line 420 "load_ams.cpp"
         $233 = $232&65535; //@line 420 "load_ams.cpp"
         $234 = $233 | 1; //@line 420 "load_ams.cpp"
         $235 = $234&65535; //@line 420 "load_ams.cpp"
         HEAP16[$231>>1] = $235; //@line 420 "load_ams.cpp"
        }
        $236 = $224 & 8; //@line 421 "load_ams.cpp"
        $237 = ($236<<24>>24)==(0); //@line 421 "load_ams.cpp"
        if (!($237)) {
         $238 = ((($193)) + 34|0); //@line 421 "load_ams.cpp"
         $239 = HEAP16[$238>>1]|0; //@line 421 "load_ams.cpp"
         $240 = $239&65535; //@line 421 "load_ams.cpp"
         $241 = $240 | 2; //@line 421 "load_ams.cpp"
         $242 = $241&65535; //@line 421 "load_ams.cpp"
         HEAP16[$238>>1] = $242; //@line 421 "load_ams.cpp"
        }
        $243 = $224 & 16; //@line 422 "load_ams.cpp"
        $244 = ($243<<24>>24)==(0); //@line 422 "load_ams.cpp"
        if ($244) {
         break;
        }
        $245 = ((($193)) + 34|0); //@line 422 "load_ams.cpp"
        $246 = HEAP16[$245>>1]|0; //@line 422 "load_ams.cpp"
        $247 = $246&65535; //@line 422 "load_ams.cpp"
        $248 = $247 | 4; //@line 422 "load_ams.cpp"
        $249 = $248&65535; //@line 422 "load_ams.cpp"
        HEAP16[$245>>1] = $249; //@line 422 "load_ams.cpp"
       }
      } while(0);
      $250 = (($201) + 20)|0; //@line 424 "load_ams.cpp"
      $251 = (($ismp$098) + 1)|0; //@line 399 "load_ams.cpp"
      $252 = HEAP8[$62>>0]|0; //@line 399 "load_ams.cpp"
      $253 = $252&255; //@line 399 "load_ams.cpp"
      $254 = ($251>>>0)<($253>>>0); //@line 399 "load_ams.cpp"
      if ($254) {
       $dwMemPos$199 = $250;$ismp$098 = $251;
      } else {
       $dwMemPos$1$lcssa = $250;
       break;
      }
     }
    }
    $255 = (($nIns$0104) + 1)|0; //@line 338 "load_ams.cpp"
    $256 = HEAP32[$47>>2]|0; //@line 338 "load_ams.cpp"
    $257 = ($255>>>0)>($256>>>0); //@line 338 "load_ams.cpp"
    if ($257) {
     $dwMemPos$0$lcssa = $dwMemPos$1$lcssa;
     break L31;
    } else {
     $dwMemPos$0105 = $dwMemPos$1$lcssa;$nIns$0104 = $255;
    }
   }
   STACKTOP = sp;return ($$0|0); //@line 557 "load_ams.cpp"
  }
 } while(0);
 $258 = (($dwMemPos$0$lcssa) + 256)|0; //@line 427 "load_ams.cpp"
 $259 = ($258>>>0)<($dwMemLength>>>0); //@line 427 "load_ams.cpp"
 if (!($259)) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 557 "load_ams.cpp"
 }
 $260 = (($lpStream) + ($dwMemPos$0$lcssa)|0); //@line 430 "load_ams.cpp"
 $261 = HEAP8[$260>>0]|0; //@line 430 "load_ams.cpp"
 $262 = $261&255; //@line 430 "load_ams.cpp"
 $263 = ($261<<24>>24)==(0); //@line 431 "load_ams.cpp"
 if ($263) {
  $$pre121 = (($dwMemPos$0$lcssa) + 1)|0; //@line 440 "load_ams.cpp"
  $$pre$phiZ2D = $$pre121;
 } else {
  $264 = (($262) + 1)|0; //@line 433 "load_ams.cpp"
  $265 = (__Znaj($264)|0); //@line 433 "load_ams.cpp"
  $266 = ((($this)) + 59796|0); //@line 433 "load_ams.cpp"
  HEAP32[$266>>2] = $265; //@line 433 "load_ams.cpp"
  $$sum27 = (($dwMemPos$0$lcssa) + 1)|0; //@line 436 "load_ams.cpp"
  $267 = (($lpStream) + ($$sum27)|0); //@line 436 "load_ams.cpp"
  _memcpy(($265|0),($267|0),($262|0))|0; //@line 436 "load_ams.cpp"
  $268 = (($265) + ($262)|0); //@line 437 "load_ams.cpp"
  HEAP8[$268>>0] = 0; //@line 437 "load_ams.cpp"
  $$pre$phiZ2D = $$sum27;
 }
 $269 = (($$pre$phiZ2D) + ($262))|0; //@line 440 "load_ams.cpp"
 $270 = (($lpStream) + ($269)|0); //@line 444 "load_ams.cpp"
 $276 = $270;$dwMemPos$287 = $269;$i1$086 = 0;
 while(1) {
  $275 = HEAP8[$276>>0]|0; //@line 444 "load_ams.cpp"
  $277 = $275&255; //@line 444 "load_ams.cpp"
  $$off25 = (($275) + -1)<<24>>24; //@line 445 "load_ams.cpp"
  $278 = ($$off25&255)<(19); //@line 445 "load_ams.cpp"
  if ($278) {
   $279 = (((((($this)) + 49600|0) + (($i1$086*36)|0)|0)) + 16|0); //@line 447 "load_ams.cpp"
   $$sum26 = (($dwMemPos$287) + 1)|0; //@line 447 "load_ams.cpp"
   $280 = (($lpStream) + ($$sum26)|0); //@line 447 "load_ams.cpp"
   _memcpy(($279|0),($280|0),($277|0))|0; //@line 447 "load_ams.cpp"
   $$pre$phi123Z2D = $$sum26;
  } else {
   $$pre122 = (($dwMemPos$287) + 1)|0; //@line 449 "load_ams.cpp"
   $$pre$phi123Z2D = $$pre122;
  }
  $274 = (($$pre$phi123Z2D) + ($277))|0; //@line 449 "load_ams.cpp"
  $281 = $277 | 256; //@line 450 "load_ams.cpp"
  $282 = (($281) + ($274))|0; //@line 450 "load_ams.cpp"
  $283 = ($282>>>0)<($dwMemLength>>>0); //@line 450 "load_ams.cpp"
  $272 = (($i1$086) + 1)|0; //@line 442 "load_ams.cpp"
  if (!($283)) {
   $$0 = 1;
   label = 97;
   break;
  }
  $271 = ($272>>>0)<(32); //@line 442 "load_ams.cpp"
  $273 = (($lpStream) + ($274)|0); //@line 444 "load_ams.cpp"
  if ($271) {
   $276 = $273;$dwMemPos$287 = $274;$i1$086 = $272;
  } else {
   $$lcssa152 = $274;$$lcssa153 = $273;
   break;
  }
 }
 if ((label|0) == 97) {
  STACKTOP = sp;return ($$0|0); //@line 557 "load_ams.cpp"
 }
 $284 = HEAP32[$$lcssa153>>2]|0; //@line 453 "load_ams.cpp"
 $285 = (($284) + ($$lcssa152))|0; //@line 454 "load_ams.cpp"
 $286 = (($285) + 256)|0; //@line 455 "load_ams.cpp"
 $287 = ($286>>>0)<($dwMemLength>>>0); //@line 455 "load_ams.cpp"
 if ($287) {
  $dwMemPos$385 = $285;$i2$084 = 0;
 } else {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 557 "load_ams.cpp"
 }
 while(1) {
  $290 = (((($this)) + 53344|0) + ($i2$084)|0); //@line 461 "load_ams.cpp"
  HEAP8[$290>>0] = -1; //@line 461 "load_ams.cpp"
  $291 = (($dwMemPos$385) + 2)|0; //@line 462 "load_ams.cpp"
  $292 = ($291>>>0)<($dwMemLength>>>0); //@line 462 "load_ams.cpp"
  if (!($292)) {
   $$0 = 1;
   label = 97;
   break;
  }
  $293 = HEAPU8[$24>>0]|(HEAPU8[$24+1>>0]<<8); //@line 463 "load_ams.cpp"
  $294 = $293&65535; //@line 463 "load_ams.cpp"
  $295 = ($i2$084>>>0)<($294>>>0); //@line 463 "load_ams.cpp"
  if ($295) {
   $296 = (($lpStream) + ($dwMemPos$385)|0); //@line 465 "load_ams.cpp"
   $297 = HEAP8[$296>>0]|0; //@line 465 "load_ams.cpp"
   HEAP8[$290>>0] = $297; //@line 465 "load_ams.cpp"
   $dwMemPos$4 = $291;
  } else {
   $dwMemPos$4 = $dwMemPos$385;
  }
  $298 = (($i2$084) + 1)|0; //@line 459 "load_ams.cpp"
  $299 = ($298>>>0)<(256); //@line 459 "load_ams.cpp"
  if ($299) {
   $dwMemPos$385 = $dwMemPos$4;$i2$084 = $298;
  } else {
   $dwMemPos$4$lcssa = $dwMemPos$4;
   break;
  }
 }
 if ((label|0) == 97) {
  STACKTOP = sp;return ($$0|0); //@line 557 "load_ams.cpp"
 }
 $288 = HEAPU8[$21>>0]|(HEAPU8[$21+1>>0]<<8); //@line 471 "load_ams.cpp"
 $289 = ($288<<16>>16)==(0); //@line 471 "load_ams.cpp"
 L114: do {
  if ($289) {
   $dwMemPos$5$lcssa = $dwMemPos$4$lcssa;
  } else {
   $dwMemPos$581 = $dwMemPos$4$lcssa;$ipat$079 = 0;
   L116: while(1) {
    $303 = (($dwMemPos$581) + 8)|0; //@line 473 "load_ams.cpp"
    $304 = ($303>>>0)<($dwMemLength>>>0); //@line 473 "load_ams.cpp"
    if (!($304)) {
     $$0 = 1;
     label = 97;
     break;
    }
    $305 = (($lpStream) + ($dwMemPos$581)|0); //@line 474 "load_ams.cpp"
    $306 = HEAP32[$305>>2]|0; //@line 474 "load_ams.cpp"
    $307 = (($dwMemPos$581) + 4)|0; //@line 475 "load_ams.cpp"
    $308 = (($lpStream) + ($307)|0); //@line 475 "load_ams.cpp"
    $309 = HEAP8[$308>>0]|0; //@line 475 "load_ams.cpp"
    $310 = $309&255; //@line 475 "load_ams.cpp"
    $311 = (($310) + 1)|0; //@line 475 "load_ams.cpp"
    $312 = (($dwMemPos$581) + 6)|0; //@line 478 "load_ams.cpp"
    $313 = (($lpStream) + ($312)|0); //@line 478 "load_ams.cpp"
    $314 = HEAP8[$313>>0]|0; //@line 478 "load_ams.cpp"
    $315 = $314&255; //@line 478 "load_ams.cpp"
    $316 = ($ipat$079>>>0)<(240); //@line 480 "load_ams.cpp"
    L119: do {
     if ($316) {
      $317 = (($dwMemLength) - ($307))|0; //@line 480 "load_ams.cpp"
      $318 = ($306>>>0)<($317>>>0); //@line 480 "load_ams.cpp"
      $319 = ($311>>>0)>(7); //@line 480 "load_ams.cpp"
      $or$cond7 = $318 & $319; //@line 480 "load_ams.cpp"
      if (!($or$cond7)) {
       break;
      }
      $$off = (($314) + -1)<<24>>24; //@line 482 "load_ams.cpp"
      $320 = ($$off&255)<(31); //@line 482 "load_ams.cpp"
      if ($320) {
       $$sum24 = (($dwMemPos$581) + 7)|0; //@line 485 "load_ams.cpp"
       $321 = (($lpStream) + ($$sum24)|0); //@line 485 "load_ams.cpp"
       _memcpy(($s|0),($321|0),($315|0))|0; //@line 485 "load_ams.cpp"
       $322 = (($s) + ($315)|0); //@line 486 "load_ams.cpp"
       HEAP8[$322>>0] = 0; //@line 486 "load_ams.cpp"
       (__ZN10CSoundFile14SetPatternNameEjPKc($this,$ipat$079,$s)|0); //@line 487 "load_ams.cpp"
      }
      $323 = $311&65535; //@line 489 "load_ams.cpp"
      $324 = (((($this)) + 52864|0) + ($ipat$079<<1)|0); //@line 489 "load_ams.cpp"
      HEAP16[$324>>1] = $323; //@line 489 "load_ams.cpp"
      $325 = HEAP32[$35>>2]|0; //@line 490 "load_ams.cpp"
      $326 = (__ZN10CSoundFile15AllocatePatternEjj($311,$325)|0); //@line 490 "load_ams.cpp"
      $327 = (((($this)) + 51904|0) + ($ipat$079<<2)|0); //@line 490 "load_ams.cpp"
      HEAP32[$327>>2] = $326; //@line 490 "load_ams.cpp"
      $328 = ($326|0)==(0|0); //@line 491 "load_ams.cpp"
      if ($328) {
       $$0 = 1;
       label = 97;
       break L116;
      }
      $329 = (($315) + 3)|0; //@line 494 "load_ams.cpp"
      $330 = ($329>>>0)<($306>>>0); //@line 496 "load_ams.cpp"
      if ($330) {
       $350 = $326;$pos3$076 = $329;$row$077 = 0;
      } else {
       break;
      }
      while(1) {
       $331 = HEAP32[$35>>2]|0; //@line 498 "load_ams.cpp"
       $332 = Math_imul($331, $row$077)|0; //@line 498 "load_ams.cpp"
       $333 = (($pos3$076) + 1)|0; //@line 499 "load_ams.cpp"
       $$sum17 = (($pos3$076) + ($307))|0; //@line 499 "load_ams.cpp"
       $334 = (($lpStream) + ($$sum17)|0); //@line 499 "load_ams.cpp"
       $335 = HEAP8[$334>>0]|0; //@line 499 "load_ams.cpp"
       $336 = $335&255; //@line 499 "load_ams.cpp"
       $337 = $336 & 31; //@line 500 "load_ams.cpp"
       $338 = $336 & 64; //@line 502 "load_ams.cpp"
       $339 = ($338|0)==(0); //@line 502 "load_ams.cpp"
       do {
        if ($339) {
         $$sum18 = (($333) + ($307))|0; //@line 504 "load_ams.cpp"
         $340 = (($lpStream) + ($$sum18)|0); //@line 504 "load_ams.cpp"
         $341 = HEAP8[$340>>0]|0; //@line 504 "load_ams.cpp"
         $342 = $341&255; //@line 504 "load_ams.cpp"
         $343 = $342 & 127; //@line 505 "load_ams.cpp"
         $344 = ($343|0)==(0); //@line 506 "load_ams.cpp"
         if ($344) {
          $$pre124 = (($337) + ($332))|0; //@line 507 "load_ams.cpp"
          $$sum20$pre$phiZZ2D = $$pre124;
         } else {
          $345 = ($343>>>0)>(1); //@line 506 "load_ams.cpp"
          $346 = (($343) + 255)|0; //@line 506 "load_ams.cpp"
          $347 = $346&255; //@line 506 "load_ams.cpp"
          $348 = $345 ? $347 : -1; //@line 506 "load_ams.cpp"
          $$sum23 = (($337) + ($332))|0; //@line 506 "load_ams.cpp"
          $349 = (($350) + (($$sum23*6)|0)|0); //@line 506 "load_ams.cpp"
          HEAP8[$349>>0] = $348; //@line 506 "load_ams.cpp"
          $$sum20$pre$phiZZ2D = $$sum23;
         }
         $351 = (($pos3$076) + 3)|0; //@line 507 "load_ams.cpp"
         $$sum19 = (($312) + ($pos3$076))|0; //@line 507 "load_ams.cpp"
         $352 = (($lpStream) + ($$sum19)|0); //@line 507 "load_ams.cpp"
         $353 = HEAP8[$352>>0]|0; //@line 507 "load_ams.cpp"
         $354 = (($350) + (($$sum20$pre$phiZZ2D*6)|0)|0); //@line 507 "load_ams.cpp"
         $355 = (((($350) + (($$sum20$pre$phiZZ2D*6)|0)|0)) + 1|0); //@line 507 "load_ams.cpp"
         HEAP8[$355>>0] = $353; //@line 507 "load_ams.cpp"
         $356 = $342 & 128; //@line 509 "load_ams.cpp"
         $357 = ($356|0)==(0); //@line 509 "load_ams.cpp"
         if ($357) {
          $pos3$2 = $351;
          break;
         }
         $358 = (((($350) + (($$sum20$pre$phiZZ2D*6)|0)|0)) + 2|0); //@line 522 "load_ams.cpp"
         $359 = (((($350) + (($$sum20$pre$phiZZ2D*6)|0)|0)) + 4|0); //@line 523 "load_ams.cpp"
         $360 = (((($350) + (($$sum20$pre$phiZZ2D*6)|0)|0)) + 3|0); //@line 527 "load_ams.cpp"
         $361 = (((($350) + (($$sum20$pre$phiZZ2D*6)|0)|0)) + 5|0); //@line 528 "load_ams.cpp"
         $pos3$173 = $351;
         while(1) {
          $362 = (($pos3$173) + 1)|0; //@line 511 "load_ams.cpp"
          $$sum21 = (($pos3$173) + ($307))|0; //@line 511 "load_ams.cpp"
          $363 = (($lpStream) + ($$sum21)|0); //@line 511 "load_ams.cpp"
          $364 = HEAP8[$363>>0]|0; //@line 511 "load_ams.cpp"
          $365 = $364&255; //@line 511 "load_ams.cpp"
          $366 = $365 & 64; //@line 512 "load_ams.cpp"
          $367 = ($366|0)==(0); //@line 512 "load_ams.cpp"
          do {
           if ($367) {
            $370 = $365 & 63; //@line 518 "load_ams.cpp"
            $371 = (($pos3$173) + 2)|0; //@line 519 "load_ams.cpp"
            $$sum22 = (($362) + ($307))|0; //@line 519 "load_ams.cpp"
            $372 = (($lpStream) + ($$sum22)|0); //@line 519 "load_ams.cpp"
            $373 = HEAP8[$372>>0]|0; //@line 519 "load_ams.cpp"
            $374 = ($370|0)==(12); //@line 520 "load_ams.cpp"
            if ($374) {
             HEAP8[$358>>0] = 1; //@line 522 "load_ams.cpp"
             $div = ($373&255) >>> 1; //@line 523 "load_ams.cpp"
             HEAP8[$359>>0] = $div; //@line 523 "load_ams.cpp"
             $pos3$1$be = $371;
             break;
            }
            $377 = ($370>>>0)<(16); //@line 525 "load_ams.cpp"
            if (!($377)) {
             $pos3$1$be = $371;
             break;
            }
            $378 = $370&255; //@line 527 "load_ams.cpp"
            HEAP8[$360>>0] = $378; //@line 527 "load_ams.cpp"
            HEAP8[$361>>0] = $373; //@line 528 "load_ams.cpp"
            __ZNK10CSoundFile17ConvertModCommandEP11_MODCOMMAND($this,$354); //@line 529 "load_ams.cpp"
            $pos3$1$be = $371;
           } else {
            HEAP8[$358>>0] = 1; //@line 514 "load_ams.cpp"
            $368 = $365 & 63; //@line 515 "load_ams.cpp"
            $369 = $368&255; //@line 515 "load_ams.cpp"
            HEAP8[$359>>0] = $369; //@line 515 "load_ams.cpp"
            $pos3$1$be = $362;
           }
          } while(0);
          $375 = $365 & 128; //@line 509 "load_ams.cpp"
          $376 = ($375|0)==(0); //@line 509 "load_ams.cpp"
          if ($376) {
           $pos3$2 = $pos3$1$be;
           break;
          } else {
           $pos3$173 = $pos3$1$be;
          }
         }
        } else {
         $pos3$2 = $333;
        }
       } while(0);
       $$lobit = $336 >>> 7; //@line 537 "load_ams.cpp"
       $row$0$ = (($$lobit) + ($row$077))|0; //@line 537 "load_ams.cpp"
       $379 = ($pos3$2>>>0)<($306>>>0); //@line 496 "load_ams.cpp"
       $380 = ($row$0$>>>0)<($311>>>0); //@line 496 "load_ams.cpp"
       $or$cond55 = $379 & $380; //@line 496 "load_ams.cpp"
       if (!($or$cond55)) {
        break L119;
       }
       $$pre115 = HEAP32[$327>>2]|0; //@line 498 "load_ams.cpp"
       $350 = $$pre115;$pos3$076 = $pos3$2;$row$077 = $row$0$;
      }
     }
    } while(0);
    $381 = (($306) + ($307))|0; //@line 540 "load_ams.cpp"
    $382 = (($ipat$079) + 1)|0; //@line 471 "load_ams.cpp"
    $383 = HEAPU8[$21>>0]|(HEAPU8[$21+1>>0]<<8); //@line 471 "load_ams.cpp"
    $384 = $383&65535; //@line 471 "load_ams.cpp"
    $385 = ($382>>>0)<($384>>>0); //@line 471 "load_ams.cpp"
    if ($385) {
     $dwMemPos$581 = $381;$ipat$079 = $382;
    } else {
     $dwMemPos$5$lcssa = $381;
     break L114;
    }
   }
   if ((label|0) == 97) {
    STACKTOP = sp;return ($$0|0); //@line 557 "load_ams.cpp"
   }
  }
 } while(0);
 $300 = (($dwMemLength) + -9)|0; //@line 545 "load_ams.cpp"
 $301 = HEAP32[$48>>2]|0; //@line 543 "load_ams.cpp"
 $302 = ($301|0)==(0); //@line 543 "load_ams.cpp"
 if ($302) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 557 "load_ams.cpp"
 } else {
  $409 = $301;$dwMemPos$670 = $dwMemPos$5$lcssa;$iSmp$071 = 1;
 }
 while(1) {
  $386 = (((($this)) + 33280|0) + ($iSmp$071<<6)|0); //@line 543 "load_ams.cpp"
  $387 = HEAP32[$386>>2]|0; //@line 543 "load_ams.cpp"
  $388 = ($387|0)==(0); //@line 543 "load_ams.cpp"
  if ($388) {
   $408 = $409;$dwMemPos$7 = $dwMemPos$670;
  } else {
   $389 = ($dwMemPos$670>>>0)<($300>>>0); //@line 545 "load_ams.cpp"
   if (!($389)) {
    $$0 = 1;
    label = 97;
    break;
   }
   $390 = (($packedsamples) + ($iSmp$071)|0); //@line 547 "load_ams.cpp"
   $391 = HEAP8[$390>>0]|0; //@line 547 "load_ams.cpp"
   $392 = $391 & 3; //@line 547 "load_ams.cpp"
   $393 = ($392<<24>>24)==(0); //@line 547 "load_ams.cpp"
   $394 = (((((($this)) + 33280|0) + ($iSmp$071<<6)|0)) + 34|0); //@line 549 "load_ams.cpp"
   $395 = HEAP16[$394>>1]|0; //@line 549 "load_ams.cpp"
   $396 = $395 & 1; //@line 549 "load_ams.cpp"
   if ($393) {
    $400 = ($396<<16>>16)!=(0); //@line 549 "load_ams.cpp"
    $401 = $400 ? 5 : 0; //@line 552 "load_ams.cpp"
    $flags$0 = $401;
   } else {
    $397 = $396&65535; //@line 549 "load_ams.cpp"
    $398 = $397 << 2; //@line 549 "load_ams.cpp"
    $399 = $398 | 17; //@line 549 "load_ams.cpp"
    $flags$0 = $399;
   }
   $402 = (($lpStream) + ($dwMemPos$670)|0); //@line 554 "load_ams.cpp"
   $403 = (($dwMemLength) - ($dwMemPos$670))|0; //@line 554 "load_ams.cpp"
   $404 = (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$386,$flags$0,$402,$403)|0); //@line 554 "load_ams.cpp"
   $405 = (($404) + ($dwMemPos$670))|0; //@line 554 "load_ams.cpp"
   $$pre116 = HEAP32[$48>>2]|0; //@line 543 "load_ams.cpp"
   $408 = $$pre116;$dwMemPos$7 = $405;
  }
  $406 = (($iSmp$071) + 1)|0; //@line 543 "load_ams.cpp"
  $407 = ($406>>>0)>($408>>>0); //@line 543 "load_ams.cpp"
  if ($407) {
   $$0 = 1;
   label = 97;
   break;
  } else {
   $409 = $408;$dwMemPos$670 = $dwMemPos$7;$iSmp$071 = $406;
  }
 }
 if ((label|0) == 97) {
  STACKTOP = sp;return ($$0|0); //@line 557 "load_ams.cpp"
 }
 return (0)|0;
}
function __ZN10CSoundFile7ReadOKTEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$0 = 0, $$2 = 0, $$23 = 0, $$24 = 0, $$25 = 0, $$26 = 0, $$lcssa = 0, $$pre = 0, $$pre$phiZ2D = 0, $$sum = 0, $$sum10 = 0, $$sum11 = 0, $$sum1213 = 0, $$sum1415 = 0, $$sum1617 = 0, $$sum1819 = 0, $$sum2021 = 0, $$sum4 = 0, $$sum5 = 0;
 var $$sum6 = 0, $$sum7 = 0, $$sum8 = 0, $$sum9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $dwMemPos$0$lcssa = 0, $dwMemPos$054 = 0, $dwMemPos$1 = 0, $dwMemPos$2 = 0, $dwMemPos$3 = 0, $dwMemPos$5$lcssa = 0, $dwMemPos$5$ph = 0, $dwMemPos$543 = 0, $dwMemPos$635 = 0, $dwPos$036 = 0, $exitcond = 0, $i$050 = 0, $i8$039 = 0, $j$047 = 0, $m$037 = 0, $norders$0 = 0, $npat$042 = 0, $nsmp$034 = 0, $or$cond = 0;
 var $smp$052 = 0, $storemerge = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ($lpStream|0)==(0|0); //@line 48 "load_okt.cpp"
 $1 = ($dwMemLength>>>0)<(1024); //@line 48 "load_okt.cpp"
 $or$cond = $0 | $1; //@line 48 "load_okt.cpp"
 if ($or$cond) {
  $$0 = 0;
  return ($$0|0); //@line 196 "load_okt.cpp"
 }
 $2 = HEAP32[$lpStream>>2]|0; //@line 49 "load_okt.cpp"
 $3 = ($2|0)==(1096043343); //@line 49 "load_okt.cpp"
 if (!($3)) {
  $$0 = 0;
  return ($$0|0); //@line 196 "load_okt.cpp"
 }
 $4 = ((($lpStream)) + 4|0); //@line 49 "load_okt.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 49 "load_okt.cpp"
 $6 = ($5|0)==(1196314451); //@line 49 "load_okt.cpp"
 if (!($6)) {
  $$0 = 0;
  return ($$0|0); //@line 196 "load_okt.cpp"
 }
 $7 = ((($lpStream)) + 8|0); //@line 50 "load_okt.cpp"
 $8 = HEAP32[$7>>2]|0; //@line 50 "load_okt.cpp"
 $9 = ($8|0)==(1146047811); //@line 50 "load_okt.cpp"
 if (!($9)) {
  $$0 = 0;
  return ($$0|0); //@line 196 "load_okt.cpp"
 }
 $10 = ((($lpStream)) + 16|0); //@line 50 "load_okt.cpp"
 $11 = HEAP8[$10>>0]|0; //@line 50 "load_okt.cpp"
 $12 = ($11<<24>>24)==(0); //@line 50 "load_okt.cpp"
 if (!($12)) {
  $$0 = 0;
  return ($$0|0); //@line 196 "load_okt.cpp"
 }
 $13 = ((($lpStream)) + 18|0); //@line 50 "load_okt.cpp"
 $14 = HEAP8[$13>>0]|0; //@line 50 "load_okt.cpp"
 $15 = ($14<<24>>24)==(0); //@line 50 "load_okt.cpp"
 if (!($15)) {
  $$0 = 0;
  return ($$0|0); //@line 196 "load_okt.cpp"
 }
 $16 = ((($lpStream)) + 20|0); //@line 51 "load_okt.cpp"
 $17 = HEAP8[$16>>0]|0; //@line 51 "load_okt.cpp"
 $18 = ($17<<24>>24)==(0); //@line 51 "load_okt.cpp"
 if (!($18)) {
  $$0 = 0;
  return ($$0|0); //@line 196 "load_okt.cpp"
 }
 $19 = ((($lpStream)) + 22|0); //@line 51 "load_okt.cpp"
 $20 = HEAP8[$19>>0]|0; //@line 51 "load_okt.cpp"
 $21 = ($20<<24>>24)==(0); //@line 51 "load_okt.cpp"
 if (!($21)) {
  $$0 = 0;
  return ($$0|0); //@line 196 "load_okt.cpp"
 }
 $22 = ((($lpStream)) + 12|0); //@line 51 "load_okt.cpp"
 $23 = HEAP32[$22>>2]|0; //@line 51 "load_okt.cpp"
 $24 = ($23|0)==(134217728); //@line 51 "load_okt.cpp"
 if (!($24)) {
  $$0 = 0;
  return ($$0|0); //@line 196 "load_okt.cpp"
 }
 $25 = ((($lpStream)) + 24|0); //@line 52 "load_okt.cpp"
 $26 = HEAP32[$25>>2]|0; //@line 52 "load_okt.cpp"
 $27 = ($26|0)==(1347240275); //@line 52 "load_okt.cpp"
 if (!($27)) {
  $$0 = 0;
  return ($$0|0); //@line 196 "load_okt.cpp"
 }
 $28 = ((($this)) + 59680|0); //@line 53 "load_okt.cpp"
 HEAP32[$28>>2] = 32768; //@line 53 "load_okt.cpp"
 $29 = ((($lpStream)) + 17|0); //@line 54 "load_okt.cpp"
 $30 = HEAP8[$29>>0]|0; //@line 54 "load_okt.cpp"
 $31 = $30&255; //@line 54 "load_okt.cpp"
 $32 = (($31) + 4)|0; //@line 54 "load_okt.cpp"
 $33 = ((($lpStream)) + 19|0); //@line 54 "load_okt.cpp"
 $34 = HEAP8[$33>>0]|0; //@line 54 "load_okt.cpp"
 $35 = $34&255; //@line 54 "load_okt.cpp"
 $36 = (($32) + ($35))|0; //@line 54 "load_okt.cpp"
 $37 = ((($lpStream)) + 21|0); //@line 54 "load_okt.cpp"
 $38 = HEAP8[$37>>0]|0; //@line 54 "load_okt.cpp"
 $39 = $38&255; //@line 54 "load_okt.cpp"
 $40 = (($36) + ($39))|0; //@line 54 "load_okt.cpp"
 $41 = ((($lpStream)) + 23|0); //@line 54 "load_okt.cpp"
 $42 = HEAP8[$41>>0]|0; //@line 54 "load_okt.cpp"
 $43 = $42&255; //@line 54 "load_okt.cpp"
 $44 = (($40) + ($43))|0; //@line 54 "load_okt.cpp"
 $45 = ((($this)) + 59664|0); //@line 54 "load_okt.cpp"
 $46 = ($44>>>0)>(128); //@line 55 "load_okt.cpp"
 $$ = $46 ? 128 : $44; //@line 55 "load_okt.cpp"
 HEAP32[$45>>2] = $$; //@line 54 "load_okt.cpp"
 $47 = ((($lpStream)) + 28|0); //@line 56 "load_okt.cpp"
 $48 = HEAP32[$47>>2]|0; //@line 56 "load_okt.cpp"
 $49 = (_llvm_bswap_i32(($48|0))|0); //@line 56 "load_okt.cpp"
 $50 = $49 >>> 5; //@line 56 "load_okt.cpp"
 $51 = ((($this)) + 59684|0); //@line 57 "load_okt.cpp"
 $52 = ($49>>>0)>(7679); //@line 58 "load_okt.cpp"
 $storemerge = $52 ? 239 : $50; //@line 58 "load_okt.cpp"
 HEAP32[$51>>2] = $storemerge; //@line 57 "load_okt.cpp"
 $53 = ($50|0)==(0); //@line 60 "load_okt.cpp"
 $54 = ($dwMemLength>>>0)<(33); //@line 62 "load_okt.cpp"
 L31: do {
  if ($53) {
   $$lcssa = $54;$dwMemPos$0$lcssa = 32;
  } else {
   $255 = $54;$dwMemPos$054 = 32;$smp$052 = 1;
   while(1) {
    if ($255) {
     $$0 = 1;
     break;
    }
    $55 = ($smp$052>>>0)<(240); //@line 63 "load_okt.cpp"
    if ($55) {
     $56 = (($lpStream) + ($dwMemPos$054)|0); //@line 65 "load_okt.cpp"
     $57 = (((($this)) + 59804|0) + ($smp$052<<5)|0); //@line 68 "load_okt.cpp"
     dest=$57; src=$56; stop=dest+20|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 68 "load_okt.cpp"
     $58 = (((((($this)) + 33280|0) + ($smp$052<<6)|0)) + 34|0); //@line 69 "load_okt.cpp"
     HEAP16[$58>>1] = 0; //@line 69 "load_okt.cpp"
     $$sum1415 = $dwMemPos$054 | 20; //@line 70 "load_okt.cpp"
     $59 = (($lpStream) + ($$sum1415)|0); //@line 70 "load_okt.cpp"
     $60 = HEAP32[$59>>2]|0; //@line 70 "load_okt.cpp"
     $61 = $60 & -16777217; //@line 70 "load_okt.cpp"
     $62 = (_llvm_bswap_i32(($61|0))|0); //@line 70 "load_okt.cpp"
     $63 = (((($this)) + 33280|0) + ($smp$052<<6)|0); //@line 70 "load_okt.cpp"
     HEAP32[$63>>2] = $62; //@line 70 "load_okt.cpp"
     $$sum1617 = $dwMemPos$054 | 24; //@line 71 "load_okt.cpp"
     $64 = (($lpStream) + ($$sum1617)|0); //@line 71 "load_okt.cpp"
     $65 = HEAP16[$64>>1]|0; //@line 71 "load_okt.cpp"
     $66 = $65&65535; //@line 71 "load_okt.cpp"
     $67 = $66 >>> 8; //@line 71 "load_okt.cpp"
     $68 = $66 << 8; //@line 71 "load_okt.cpp"
     $69 = $68 & 65280; //@line 71 "load_okt.cpp"
     $70 = $69 | $67; //@line 71 "load_okt.cpp"
     $71 = (((((($this)) + 33280|0) + ($smp$052<<6)|0)) + 4|0); //@line 71 "load_okt.cpp"
     HEAP32[$71>>2] = $70; //@line 71 "load_okt.cpp"
     $$sum1819 = $dwMemPos$054 | 26; //@line 72 "load_okt.cpp"
     $72 = (($lpStream) + ($$sum1819)|0); //@line 72 "load_okt.cpp"
     $73 = HEAP16[$72>>1]|0; //@line 72 "load_okt.cpp"
     $74 = $73&65535; //@line 72 "load_okt.cpp"
     $75 = $74 >>> 8; //@line 72 "load_okt.cpp"
     $76 = $74 << 8; //@line 72 "load_okt.cpp"
     $77 = $76 & 65280; //@line 72 "load_okt.cpp"
     $78 = $77 | $75; //@line 72 "load_okt.cpp"
     $79 = (($78) + ($70))|0; //@line 72 "load_okt.cpp"
     $80 = (((((($this)) + 33280|0) + ($smp$052<<6)|0)) + 8|0); //@line 72 "load_okt.cpp"
     HEAP32[$80>>2] = $79; //@line 72 "load_okt.cpp"
     $81 = (($70) + 2)|0; //@line 73 "load_okt.cpp"
     $82 = ($81>>>0)<($79>>>0); //@line 73 "load_okt.cpp"
     if ($82) {
      HEAP16[$58>>1] = 2; //@line 73 "load_okt.cpp"
     }
     $83 = (((((($this)) + 33280|0) + ($smp$052<<6)|0)) + 32|0); //@line 74 "load_okt.cpp"
     HEAP16[$83>>1] = 64; //@line 74 "load_okt.cpp"
     $$sum2021 = $dwMemPos$054 | 29; //@line 75 "load_okt.cpp"
     $84 = (($lpStream) + ($$sum2021)|0); //@line 75 "load_okt.cpp"
     $85 = HEAP8[$84>>0]|0; //@line 75 "load_okt.cpp"
     $86 = $85&255; //@line 75 "load_okt.cpp"
     $87 = $86 << 2; //@line 75 "load_okt.cpp"
     $88 = $87&65535; //@line 75 "load_okt.cpp"
     $89 = (((((($this)) + 33280|0) + ($smp$052<<6)|0)) + 30|0); //@line 75 "load_okt.cpp"
     HEAP16[$89>>1] = $88; //@line 75 "load_okt.cpp"
     $90 = (((((($this)) + 33280|0) + ($smp$052<<6)|0)) + 24|0); //@line 76 "load_okt.cpp"
     HEAP32[$90>>2] = 8363; //@line 76 "load_okt.cpp"
    }
    $91 = (($dwMemPos$054) + 32)|0; //@line 78 "load_okt.cpp"
    $92 = (($smp$052) + 1)|0; //@line 60 "load_okt.cpp"
    $93 = ($92>>>0)>($50>>>0); //@line 60 "load_okt.cpp"
    $94 = ($91>>>0)>=($dwMemLength>>>0); //@line 62 "load_okt.cpp"
    if ($93) {
     $$lcssa = $94;$dwMemPos$0$lcssa = $91;
     break L31;
    } else {
     $255 = $94;$dwMemPos$054 = $91;$smp$052 = $92;
    }
   }
   return ($$0|0); //@line 196 "load_okt.cpp"
  }
 } while(0);
 if ($$lcssa) {
  $$0 = 1;
  return ($$0|0); //@line 196 "load_okt.cpp"
 }
 $95 = (($lpStream) + ($dwMemPos$0$lcssa)|0); //@line 82 "load_okt.cpp"
 $96 = HEAP32[$95>>2]|0; //@line 82 "load_okt.cpp"
 $97 = ($96|0)==(1162170451); //@line 82 "load_okt.cpp"
 if ($97) {
  $98 = $dwMemPos$0$lcssa | 9; //@line 84 "load_okt.cpp"
  $99 = (($lpStream) + ($98)|0); //@line 84 "load_okt.cpp"
  $100 = HEAP8[$99>>0]|0; //@line 84 "load_okt.cpp"
  $101 = $100&255; //@line 84 "load_okt.cpp"
  $102 = ((($this)) + 59648|0); //@line 84 "load_okt.cpp"
  HEAP32[$102>>2] = $101; //@line 84 "load_okt.cpp"
  $$sum1213 = $dwMemPos$0$lcssa | 4; //@line 85 "load_okt.cpp"
  $103 = (($lpStream) + ($$sum1213)|0); //@line 85 "load_okt.cpp"
  $104 = HEAP32[$103>>2]|0; //@line 85 "load_okt.cpp"
  $105 = (_llvm_bswap_i32(($104|0))|0); //@line 85 "load_okt.cpp"
  $106 = (($dwMemPos$0$lcssa) + 8)|0; //@line 85 "load_okt.cpp"
  $107 = (($106) + ($105))|0; //@line 85 "load_okt.cpp"
  $dwMemPos$1 = $107;
 } else {
  $dwMemPos$1 = $dwMemPos$0$lcssa;
 }
 $108 = ($dwMemPos$1>>>0)<($dwMemLength>>>0); //@line 88 "load_okt.cpp"
 if (!($108)) {
  $$0 = 1;
  return ($$0|0); //@line 196 "load_okt.cpp"
 }
 $109 = (($lpStream) + ($dwMemPos$1)|0); //@line 89 "load_okt.cpp"
 $110 = HEAP32[$109>>2]|0; //@line 89 "load_okt.cpp"
 $111 = ($110|0)==(1313164371); //@line 89 "load_okt.cpp"
 if ($111) {
  $$sum11 = (($dwMemPos$1) + 4)|0; //@line 92 "load_okt.cpp"
  $112 = (($lpStream) + ($$sum11)|0); //@line 92 "load_okt.cpp"
  $113 = HEAP32[$112>>2]|0; //@line 92 "load_okt.cpp"
  $114 = (_llvm_bswap_i32(($113|0))|0); //@line 92 "load_okt.cpp"
  $115 = (($dwMemPos$1) + 8)|0; //@line 92 "load_okt.cpp"
  $116 = (($115) + ($114))|0; //@line 92 "load_okt.cpp"
  $dwMemPos$2 = $116;
 } else {
  $dwMemPos$2 = $dwMemPos$1;
 }
 $117 = ($dwMemPos$2>>>0)<($dwMemLength>>>0); //@line 95 "load_okt.cpp"
 if (!($117)) {
  $$0 = 1;
  return ($$0|0); //@line 196 "load_okt.cpp"
 }
 $118 = (($lpStream) + ($dwMemPos$2)|0); //@line 96 "load_okt.cpp"
 $119 = HEAP32[$118>>2]|0; //@line 96 "load_okt.cpp"
 $120 = ($119|0)==(1313164368); //@line 96 "load_okt.cpp"
 if ($120) {
  $121 = (($dwMemPos$2) + 9)|0; //@line 98 "load_okt.cpp"
  $122 = (($lpStream) + ($121)|0); //@line 98 "load_okt.cpp"
  $123 = HEAP8[$122>>0]|0; //@line 98 "load_okt.cpp"
  $124 = $123&255; //@line 98 "load_okt.cpp"
  $$sum10 = (($dwMemPos$2) + 4)|0; //@line 99 "load_okt.cpp"
  $125 = (($lpStream) + ($$sum10)|0); //@line 99 "load_okt.cpp"
  $126 = HEAP32[$125>>2]|0; //@line 99 "load_okt.cpp"
  $127 = (_llvm_bswap_i32(($126|0))|0); //@line 99 "load_okt.cpp"
  $128 = (($dwMemPos$2) + 8)|0; //@line 99 "load_okt.cpp"
  $129 = (($128) + ($127))|0; //@line 99 "load_okt.cpp"
  $dwMemPos$3 = $129;$norders$0 = $124;
 } else {
  $dwMemPos$3 = $dwMemPos$2;$norders$0 = 0;
 }
 $130 = ($dwMemPos$3>>>0)<($dwMemLength>>>0); //@line 102 "load_okt.cpp"
 if (!($130)) {
  $$0 = 1;
  return ($$0|0); //@line 196 "load_okt.cpp"
 }
 $131 = (($lpStream) + ($dwMemPos$3)|0); //@line 103 "load_okt.cpp"
 $132 = HEAP32[$131>>2]|0; //@line 103 "load_okt.cpp"
 $133 = ($132|0)==(1414807888); //@line 103 "load_okt.cpp"
 if ($133) {
  $134 = ($norders$0|0)==(0); //@line 107 "load_okt.cpp"
  L66: do {
   if (!($134)) {
    $135 = (($dwMemPos$3) + 10)|0; //@line 107 "load_okt.cpp"
    $i$050 = 0;
    while(1) {
     $137 = (($135) + ($i$050))|0; //@line 107 "load_okt.cpp"
     $138 = (($lpStream) + ($137)|0); //@line 107 "load_okt.cpp"
     $139 = HEAP8[$138>>0]|0; //@line 107 "load_okt.cpp"
     $140 = (((($this)) + 53344|0) + ($i$050)|0); //@line 107 "load_okt.cpp"
     HEAP8[$140>>0] = $139; //@line 107 "load_okt.cpp"
     $141 = (($i$050) + 1)|0; //@line 107 "load_okt.cpp"
     $exitcond = ($141|0)==($norders$0|0); //@line 107 "load_okt.cpp"
     if ($exitcond) {
      break;
     } else {
      $i$050 = $141;
     }
    }
    $136 = ($norders$0>>>0)>(1); //@line 108 "load_okt.cpp"
    if ($136) {
     $j$047 = $norders$0;
    } else {
     break;
    }
    while(1) {
     $142 = (($j$047) + -1)|0; //@line 108 "load_okt.cpp"
     $143 = (((($this)) + 53344|0) + ($142)|0); //@line 108 "load_okt.cpp"
     $144 = HEAP8[$143>>0]|0; //@line 108 "load_okt.cpp"
     $145 = ($144<<24>>24)==(0); //@line 108 "load_okt.cpp"
     if (!($145)) {
      break L66;
     }
     HEAP8[$143>>0] = -1; //@line 108 "load_okt.cpp"
     $146 = ($142>>>0)>(1); //@line 108 "load_okt.cpp"
     if ($146) {
      $j$047 = $142;
     } else {
      break;
     }
    }
   }
  } while(0);
  $$sum9 = (($dwMemPos$3) + 4)|0; //@line 109 "load_okt.cpp"
  $147 = (($lpStream) + ($$sum9)|0); //@line 109 "load_okt.cpp"
  $148 = HEAP32[$147>>2]|0; //@line 109 "load_okt.cpp"
  $149 = (_llvm_bswap_i32(($148|0))|0); //@line 109 "load_okt.cpp"
  $150 = (($dwMemPos$3) + 8)|0; //@line 109 "load_okt.cpp"
  $151 = (($150) + ($149))|0; //@line 109 "load_okt.cpp"
  $dwMemPos$5$ph = $151;
 } else {
  $dwMemPos$5$ph = $dwMemPos$3;
 }
 $152 = (($dwMemPos$5$ph) + 10)|0; //@line 113 "load_okt.cpp"
 $153 = ($152>>>0)<($dwMemLength>>>0); //@line 113 "load_okt.cpp"
 L76: do {
  if ($153) {
   $256 = $152;$dwMemPos$543 = $dwMemPos$5$ph;$npat$042 = 0;
   L77: while(1) {
    $154 = (($lpStream) + ($dwMemPos$543)|0); //@line 113 "load_okt.cpp"
    $155 = HEAP32[$154>>2]|0; //@line 113 "load_okt.cpp"
    $156 = ($155|0)==(1146045008); //@line 113 "load_okt.cpp"
    if (!($156)) {
     $dwMemPos$5$lcssa = $dwMemPos$543;
     break L76;
    }
    $160 = (($dwMemPos$543) + 9)|0; //@line 116 "load_okt.cpp"
    $161 = (($lpStream) + ($160)|0); //@line 116 "load_okt.cpp"
    $162 = HEAP8[$161>>0]|0; //@line 116 "load_okt.cpp"
    $163 = $162&255; //@line 116 "load_okt.cpp"
    $164 = ($162<<24>>24)!=(0); //@line 117 "load_okt.cpp"
    $$2 = $164 ? $163 : 64; //@line 117 "load_okt.cpp"
    $165 = ($npat$042>>>0)<(240); //@line 118 "load_okt.cpp"
    L80: do {
     if ($165) {
      $166 = HEAP32[$45>>2]|0; //@line 120 "load_okt.cpp"
      $167 = (__ZN10CSoundFile15AllocatePatternEjj($$2,$166)|0); //@line 120 "load_okt.cpp"
      $168 = (((($this)) + 51904|0) + ($npat$042<<2)|0); //@line 120 "load_okt.cpp"
      HEAP32[$168>>2] = $167; //@line 120 "load_okt.cpp"
      $169 = ($167|0)==(0|0); //@line 120 "load_okt.cpp"
      if ($169) {
       $$0 = 1;
       break L77;
      }
      $170 = $$2&65535; //@line 122 "load_okt.cpp"
      $171 = (((($this)) + 52864|0) + ($npat$042<<1)|0); //@line 122 "load_okt.cpp"
      HEAP16[$171>>1] = $170; //@line 122 "load_okt.cpp"
      $172 = HEAP32[$45>>2]|0; //@line 123 "load_okt.cpp"
      $173 = Math_imul($172, $$2)|0; //@line 123 "load_okt.cpp"
      $174 = ($173|0)==(0); //@line 124 "load_okt.cpp"
      if ($174) {
       break;
      } else {
       $dwPos$036 = $256;$i8$039 = 0;$m$037 = $167;
      }
      while(1) {
       $175 = (($dwPos$036) + 4)|0; //@line 126 "load_okt.cpp"
       $176 = ($175>>>0)>($dwMemLength>>>0); //@line 126 "load_okt.cpp"
       if ($176) {
        break L80;
       }
       $177 = (($lpStream) + ($dwPos$036)|0); //@line 127 "load_okt.cpp"
       $178 = HEAP8[$177>>0]|0; //@line 128 "load_okt.cpp"
       $179 = ($178<<24>>24)==(0); //@line 129 "load_okt.cpp"
       if (!($179)) {
        $180 = $178&255; //@line 128 "load_okt.cpp"
        $181 = (($180) + 48)|0; //@line 131 "load_okt.cpp"
        $182 = $181&255; //@line 131 "load_okt.cpp"
        HEAP8[$m$037>>0] = $182; //@line 131 "load_okt.cpp"
        $$sum8 = (($dwPos$036) + 1)|0; //@line 132 "load_okt.cpp"
        $183 = (($lpStream) + ($$sum8)|0); //@line 132 "load_okt.cpp"
        $184 = HEAP8[$183>>0]|0; //@line 132 "load_okt.cpp"
        $185 = $184&255; //@line 132 "load_okt.cpp"
        $186 = (($185) + 1)|0; //@line 132 "load_okt.cpp"
        $187 = $186&255; //@line 132 "load_okt.cpp"
        $188 = ((($m$037)) + 1|0); //@line 132 "load_okt.cpp"
        HEAP8[$188>>0] = $187; //@line 132 "load_okt.cpp"
       }
       $$sum6 = (($dwPos$036) + 2)|0; //@line 134 "load_okt.cpp"
       $189 = (($lpStream) + ($$sum6)|0); //@line 134 "load_okt.cpp"
       $190 = HEAP8[$189>>0]|0; //@line 134 "load_okt.cpp"
       $191 = $190&255; //@line 134 "load_okt.cpp"
       $$sum7 = (($dwPos$036) + 3)|0; //@line 135 "load_okt.cpp"
       $192 = (($lpStream) + ($$sum7)|0); //@line 135 "load_okt.cpp"
       $193 = HEAP8[$192>>0]|0; //@line 135 "load_okt.cpp"
       $194 = $193&255; //@line 135 "load_okt.cpp"
       $195 = ((($m$037)) + 5|0); //@line 136 "load_okt.cpp"
       HEAP8[$195>>0] = $193; //@line 136 "load_okt.cpp"
       L89: do {
        switch ($191|0) {
        case 31:  {
         $206 = ($193&255)<(65); //@line 175 "load_okt.cpp"
         if ($206) {
          $207 = ((($m$037)) + 3|0); //@line 175 "load_okt.cpp"
          HEAP8[$207>>0] = 13; //@line 175 "load_okt.cpp"
          break L89;
         }
         $208 = ($193&255)<(81); //@line 176 "load_okt.cpp"
         if ($208) {
          $209 = ((($m$037)) + 3|0); //@line 176 "load_okt.cpp"
          HEAP8[$209>>0] = 11; //@line 176 "load_okt.cpp"
          $210 = $194 & 15; //@line 176 "load_okt.cpp"
          $211 = $210&255; //@line 176 "load_okt.cpp"
          $212 = ($211<<24>>24)==(0); //@line 176 "load_okt.cpp"
          $$23 = $212 ? 15 : $211; //@line 176 "load_okt.cpp"
          HEAP8[$195>>0] = $$23; //@line 176 "load_okt.cpp"
          break L89;
         }
         $213 = ($193&255)<(97); //@line 177 "load_okt.cpp"
         if ($213) {
          $214 = ((($m$037)) + 3|0); //@line 177 "load_okt.cpp"
          HEAP8[$214>>0] = 11; //@line 177 "load_okt.cpp"
          $215 = $194 << 4; //@line 177 "load_okt.cpp"
          $216 = $215&255; //@line 177 "load_okt.cpp"
          $217 = ($216<<24>>24)==(0); //@line 177 "load_okt.cpp"
          $$24 = $217 ? -16 : $216; //@line 177 "load_okt.cpp"
          HEAP8[$195>>0] = $$24; //@line 177 "load_okt.cpp"
          break L89;
         }
         $218 = ($193&255)<(113); //@line 178 "load_okt.cpp"
         if ($218) {
          $219 = ((($m$037)) + 3|0); //@line 178 "load_okt.cpp"
          HEAP8[$219>>0] = 19; //@line 178 "load_okt.cpp"
          $220 = $194 & 15; //@line 178 "load_okt.cpp"
          $221 = $220 | 176; //@line 178 "load_okt.cpp"
          $222 = $221&255; //@line 178 "load_okt.cpp"
          $223 = ($220|0)==(0); //@line 178 "load_okt.cpp"
          $$25 = $223 ? -65 : $222; //@line 178 "load_okt.cpp"
          HEAP8[$195>>0] = $$25; //@line 178 "load_okt.cpp"
          break L89;
         }
         $224 = ($193&255)<(129); //@line 179 "load_okt.cpp"
         if (!($224)) {
          break L89;
         }
         $225 = ((($m$037)) + 3|0); //@line 179 "load_okt.cpp"
         HEAP8[$225>>0] = 19; //@line 179 "load_okt.cpp"
         $226 = $194 & 15; //@line 179 "load_okt.cpp"
         $227 = $226 | 160; //@line 179 "load_okt.cpp"
         $228 = $227&255; //@line 179 "load_okt.cpp"
         $229 = ($226|0)==(0); //@line 179 "load_okt.cpp"
         $$26 = $229 ? -81 : $228; //@line 179 "load_okt.cpp"
         HEAP8[$195>>0] = $$26; //@line 179 "load_okt.cpp"
         break;
        }
        case 30: case 17: case 1:  {
         $196 = ($193<<24>>24)==(0); //@line 146 "load_okt.cpp"
         if ($196) {
          break L89;
         }
         $197 = ((($m$037)) + 3|0); //@line 146 "load_okt.cpp"
         HEAP8[$197>>0] = 2; //@line 146 "load_okt.cpp"
         break;
        }
        case 21: case 13: case 2:  {
         $198 = ($193<<24>>24)==(0); //@line 152 "load_okt.cpp"
         if ($198) {
          break L89;
         }
         $199 = ((($m$037)) + 3|0); //@line 152 "load_okt.cpp"
         HEAP8[$199>>0] = 3; //@line 152 "load_okt.cpp"
         break;
        }
        case 12: case 11: case 10:  {
         $200 = ((($m$037)) + 3|0); //@line 158 "load_okt.cpp"
         HEAP8[$200>>0] = 1; //@line 158 "load_okt.cpp"
         break;
        }
        case 15:  {
         $201 = ((($m$037)) + 3|0); //@line 162 "load_okt.cpp"
         HEAP8[$201>>0] = 19; //@line 162 "load_okt.cpp"
         $202 = $194 & 15; //@line 163 "load_okt.cpp"
         $203 = $202&255; //@line 163 "load_okt.cpp"
         HEAP8[$195>>0] = $203; //@line 163 "load_okt.cpp"
         break;
        }
        case 25:  {
         $204 = ((($m$037)) + 3|0); //@line 167 "load_okt.cpp"
         HEAP8[$204>>0] = 12; //@line 167 "load_okt.cpp"
         break;
        }
        case 28:  {
         $205 = ((($m$037)) + 3|0); //@line 171 "load_okt.cpp"
         HEAP8[$205>>0] = 16; //@line 171 "load_okt.cpp"
         break;
        }
        default: {
        }
        }
       } while(0);
       $230 = (($i8$039) + 1)|0; //@line 124 "load_okt.cpp"
       $231 = ((($m$037)) + 6|0); //@line 124 "load_okt.cpp"
       $232 = ($230>>>0)<($173>>>0); //@line 124 "load_okt.cpp"
       if ($232) {
        $dwPos$036 = $175;$i8$039 = $230;$m$037 = $231;
       } else {
        break L80;
       }
      }
     }
    } while(0);
    $233 = (($npat$042) + 1)|0; //@line 184 "load_okt.cpp"
    $$sum5 = (($dwMemPos$543) + 4)|0; //@line 185 "load_okt.cpp"
    $234 = (($lpStream) + ($$sum5)|0); //@line 185 "load_okt.cpp"
    $235 = HEAP32[$234>>2]|0; //@line 185 "load_okt.cpp"
    $236 = (_llvm_bswap_i32(($235|0))|0); //@line 185 "load_okt.cpp"
    $237 = (($dwMemPos$543) + 8)|0; //@line 185 "load_okt.cpp"
    $238 = (($237) + ($236))|0; //@line 185 "load_okt.cpp"
    $239 = (($238) + 10)|0; //@line 113 "load_okt.cpp"
    $240 = ($239>>>0)<($dwMemLength>>>0); //@line 113 "load_okt.cpp"
    if ($240) {
     $256 = $239;$dwMemPos$543 = $238;$npat$042 = $233;
    } else {
     $dwMemPos$5$lcssa = $238;
     break L76;
    }
   }
   return ($$0|0); //@line 196 "load_okt.cpp"
  } else {
   $dwMemPos$5$lcssa = $dwMemPos$5$ph;
  }
 } while(0);
 $157 = (($dwMemPos$5$lcssa) + 10)|0; //@line 189 "load_okt.cpp"
 $158 = ($157>>>0)<($dwMemLength>>>0); //@line 189 "load_okt.cpp"
 if (!($158)) {
  $$0 = 1;
  return ($$0|0); //@line 196 "load_okt.cpp"
 }
 $159 = (($dwMemLength) + -8)|0; //@line 191 "load_okt.cpp"
 $dwMemPos$635 = $dwMemPos$5$lcssa;$nsmp$034 = 1;
 while(1) {
  $241 = (($lpStream) + ($dwMemPos$635)|0); //@line 189 "load_okt.cpp"
  $242 = HEAP32[$241>>2]|0; //@line 189 "load_okt.cpp"
  $243 = ($242|0)==(1146045011); //@line 189 "load_okt.cpp"
  if (!($243)) {
   $$0 = 1;
   label = 72;
   break;
  }
  $244 = ($nsmp$034>>>0)<(240); //@line 191 "load_okt.cpp"
  if ($244) {
   $245 = (((($this)) + 33280|0) + ($nsmp$034<<6)|0); //@line 191 "load_okt.cpp"
   $$sum4 = (($dwMemPos$635) + 8)|0; //@line 191 "load_okt.cpp"
   $246 = (($lpStream) + ($$sum4)|0); //@line 191 "load_okt.cpp"
   $247 = (($159) - ($dwMemPos$635))|0; //@line 191 "load_okt.cpp"
   (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$245,0,$246,$247)|0); //@line 191 "load_okt.cpp"
   $$pre$phiZ2D = $$sum4;
  } else {
   $$pre = (($dwMemPos$635) + 8)|0; //@line 192 "load_okt.cpp"
   $$pre$phiZ2D = $$pre;
  }
  $$sum = (($dwMemPos$635) + 4)|0; //@line 192 "load_okt.cpp"
  $248 = (($lpStream) + ($$sum)|0); //@line 192 "load_okt.cpp"
  $249 = HEAP32[$248>>2]|0; //@line 192 "load_okt.cpp"
  $250 = (_llvm_bswap_i32(($249|0))|0); //@line 192 "load_okt.cpp"
  $251 = (($$pre$phiZ2D) + ($250))|0; //@line 192 "load_okt.cpp"
  $252 = (($nsmp$034) + 1)|0; //@line 193 "load_okt.cpp"
  $253 = (($251) + 10)|0; //@line 189 "load_okt.cpp"
  $254 = ($253>>>0)<($dwMemLength>>>0); //@line 189 "load_okt.cpp"
  if ($254) {
   $dwMemPos$635 = $251;$nsmp$034 = $252;
  } else {
   $$0 = 1;
   label = 72;
   break;
  }
 }
 if ((label|0) == 72) {
  return ($$0|0); //@line 196 "load_okt.cpp"
 }
 return (0)|0;
}
function __ZN10CSoundFile7ReadPTMEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$0 = 0, $$1 = 0, $$idx = 0, $$idx$val = 0, $$idx23 = 0, $$idx23$val = 0, $$idx25 = 0, $$idx25$val = 0, $$idx27 = 0, $$idx27$val = 0, $$lcssa89 = 0, $$sum = 0, $$sum10 = 0, $$sum11 = 0, $$sum12 = 0, $$sum13 = 0, $$sum28 = 0, $$sum29 = 0, $$sum30 = 0, $$sum31 = 0;
 var $$sum45 = 0, $$sum67 = 0, $$sum89 = 0, $$val = 0, $$val22 = 0, $$val24 = 0, $$val26 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0;
 var $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dwMemPos$049 = 0, $dwMemPos$1$be$us = 0, $dwMemPos$1$ph44 = 0;
 var $dwMemPos$135$us = 0, $dwMemPos$2$us = 0, $dwMemPos$3$us = 0, $ipan$051 = 0, $ipat$045 = 0, $ismp$048 = 0, $m$0$ph41 = 0, $or$cond = 0, $or$cond17 = 0, $or$cond18 = 0, $or$cond1934 = 0, $or$cond21$us = 0, $or$cond3 = 0, $pfh = 0, $row$0$ph43 = 0, $smpflg$0 = 0, dest = 0, label = 0, sp = 0, src = 0;
 var stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 608|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $pfh = sp;
 $0 = ($lpStream|0)==(0|0); //@line 73 "load_ptm.cpp"
 $1 = ($dwMemLength>>>0)<(608); //@line 73 "load_ptm.cpp"
 $or$cond = $0 | $1; //@line 73 "load_ptm.cpp"
 if ($or$cond) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 211 "load_ptm.cpp"
 }
 _memcpy(($pfh|0),($lpStream|0),608)|0; //@line 74 "load_ptm.cpp"
 $2 = ((($pfh)) + 34|0); //@line 77 "load_ptm.cpp"
 $3 = ((($pfh)) + 32|0); //@line 76 "load_ptm.cpp"
 $4 = ((($pfh)) + 36|0); //@line 78 "load_ptm.cpp"
 $5 = ((($pfh)) + 44|0); //@line 82 "load_ptm.cpp"
 $6 = HEAPU8[$5>>0]|(HEAPU8[$5+1>>0]<<8)|(HEAPU8[$5+2>>0]<<16)|(HEAPU8[$5+3>>0]<<24); //@line 88 "load_ptm.cpp"
 $7 = ($6|0)==(1179472976); //@line 88 "load_ptm.cpp"
 L4: do {
  if ($7) {
   $8 = ((($pfh)) + 38|0); //@line 79 "load_ptm.cpp"
   $9 = HEAPU8[$8>>0]|(HEAPU8[$8+1>>0]<<8); //@line 88 "load_ptm.cpp"
   $10 = (($9) + -1)<<16>>16; //@line 89 "load_ptm.cpp"
   $11 = ($10&65535)>(31); //@line 89 "load_ptm.cpp"
   if ($11) {
    $$0 = 0;
   } else {
    $12 = HEAPU8[$3>>0]|(HEAPU8[$3+1>>0]<<8); //@line 90 "load_ptm.cpp"
    $13 = (($12) + -1)<<16>>16; //@line 90 "load_ptm.cpp"
    $14 = ($13&65535)>(255); //@line 90 "load_ptm.cpp"
    if ($14) {
     $$0 = 0;
    } else {
     $15 = HEAPU8[$2>>0]|(HEAPU8[$2+1>>0]<<8); //@line 91 "load_ptm.cpp"
     $16 = ($15<<16>>16)==(0); //@line 91 "load_ptm.cpp"
     if ($16) {
      $$0 = 0;
     } else {
      $17 = $15&65535; //@line 91 "load_ptm.cpp"
      $18 = ($15&65535)>(255); //@line 91 "load_ptm.cpp"
      if ($18) {
       $$0 = 0;
      } else {
       $19 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8); //@line 92 "load_ptm.cpp"
       $20 = (($19) + -1)<<16>>16; //@line 92 "load_ptm.cpp"
       $21 = ($20&65535)>(127); //@line 92 "load_ptm.cpp"
       if ($21) {
        $$0 = 0;
       } else {
        $22 = ($17*80)|0; //@line 93 "load_ptm.cpp"
        $23 = (($22) + 608)|0; //@line 93 "load_ptm.cpp"
        $24 = ($23|0)<($dwMemLength|0); //@line 93 "load_ptm.cpp"
        if ($24) {
         $25 = ((($this)) + 59804|0); //@line 94 "load_ptm.cpp"
         dest=$25; src=$pfh; stop=dest+28|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 94 "load_ptm.cpp"
         $26 = ((($this)) + 59832|0); //@line 95 "load_ptm.cpp"
         HEAP8[$26>>0] = 0; //@line 95 "load_ptm.cpp"
         $27 = ((($this)) + 59680|0); //@line 96 "load_ptm.cpp"
         HEAP32[$27>>2] = 262144; //@line 96 "load_ptm.cpp"
         $28 = $9&65535; //@line 97 "load_ptm.cpp"
         $29 = ((($this)) + 59664|0); //@line 97 "load_ptm.cpp"
         HEAP32[$29>>2] = $28; //@line 97 "load_ptm.cpp"
         $30 = ($17>>>0)>(239); //@line 98 "load_ptm.cpp"
         $31 = $30 ? 239 : $17; //@line 98 "load_ptm.cpp"
         $32 = ((($this)) + 59684|0); //@line 98 "load_ptm.cpp"
         HEAP32[$32>>2] = $31; //@line 98 "load_ptm.cpp"
         $33 = $12&65535; //@line 100 "load_ptm.cpp"
         $34 = ($33>>>0)>(255); //@line 100 "load_ptm.cpp"
         $35 = $34 ? 255 : $33; //@line 100 "load_ptm.cpp"
         $36 = ((($this)) + 53344|0); //@line 101 "load_ptm.cpp"
         $37 = ((($pfh)) + 96|0); //@line 101 "load_ptm.cpp"
         _memcpy(($36|0),($37|0),($35|0))|0; //@line 101 "load_ptm.cpp"
         $38 = HEAP32[$29>>2]|0; //@line 102 "load_ptm.cpp"
         $39 = ($38|0)==(0); //@line 102 "load_ptm.cpp"
         if (!($39)) {
          $ipan$051 = 0;
          while(1) {
           $42 = (((((($this)) + 49600|0) + (($ipan$051*36)|0)|0)) + 4|0); //@line 104 "load_ptm.cpp"
           HEAP32[$42>>2] = 64; //@line 104 "load_ptm.cpp"
           $43 = (((($pfh)) + 64|0) + ($ipan$051)|0); //@line 105 "load_ptm.cpp"
           $44 = HEAP8[$43>>0]|0; //@line 105 "load_ptm.cpp"
           $45 = $44&255; //@line 105 "load_ptm.cpp"
           $46 = $45 << 4; //@line 105 "load_ptm.cpp"
           $47 = $46 & 240; //@line 105 "load_ptm.cpp"
           $48 = $47 | 4; //@line 105 "load_ptm.cpp"
           $49 = (((($this)) + 49600|0) + (($ipan$051*36)|0)|0); //@line 105 "load_ptm.cpp"
           HEAP32[$49>>2] = $48; //@line 105 "load_ptm.cpp"
           $50 = (($ipan$051) + 1)|0; //@line 102 "load_ptm.cpp"
           $51 = HEAP32[$29>>2]|0; //@line 102 "load_ptm.cpp"
           $52 = ($50>>>0)<($51>>>0); //@line 102 "load_ptm.cpp"
           if ($52) {
            $ipan$051 = $50;
           } else {
            break;
           }
          }
         }
         $40 = HEAP32[$32>>2]|0; //@line 107 "load_ptm.cpp"
         $41 = ($40|0)==(0); //@line 107 "load_ptm.cpp"
         if (!($41)) {
          $dwMemPos$049 = 608;$ismp$048 = 0;
          while(1) {
           $55 = (($ismp$048) + 1)|0; //@line 109 "load_ptm.cpp"
           $56 = (((($this)) + 33280|0) + ($55<<6)|0); //@line 109 "load_ptm.cpp"
           $57 = (($lpStream) + ($dwMemPos$049)|0); //@line 110 "load_ptm.cpp"
           $58 = (((($this)) + 59804|0) + ($55<<5)|0); //@line 112 "load_ptm.cpp"
           $$sum = (($dwMemPos$049) + 50)|0; //@line 112 "load_ptm.cpp"
           $59 = (($lpStream) + ($$sum)|0); //@line 112 "load_ptm.cpp"
           (_strncpy($58,$59,28)|0); //@line 112 "load_ptm.cpp"
           $60 = (((((($this)) + 33280|0) + ($55<<6)|0)) + 42|0); //@line 113 "load_ptm.cpp"
           $$sum45 = $dwMemPos$049 | 1; //@line 113 "load_ptm.cpp"
           $61 = (($lpStream) + ($$sum45)|0); //@line 113 "load_ptm.cpp"
           dest=$60; src=$61; stop=dest+12|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 113 "load_ptm.cpp"
           $62 = (((((($this)) + 33280|0) + ($55<<6)|0)) + 54|0); //@line 114 "load_ptm.cpp"
           HEAP8[$62>>0] = 0; //@line 114 "load_ptm.cpp"
           $63 = (((((($this)) + 33280|0) + ($55<<6)|0)) + 32|0); //@line 115 "load_ptm.cpp"
           HEAP16[$63>>1] = 64; //@line 115 "load_ptm.cpp"
           $64 = (((((($this)) + 33280|0) + ($55<<6)|0)) + 28|0); //@line 116 "load_ptm.cpp"
           HEAP16[$64>>1] = 128; //@line 116 "load_ptm.cpp"
           $$sum67 = $dwMemPos$049 | 13; //@line 117 "load_ptm.cpp"
           $65 = (($lpStream) + ($$sum67)|0); //@line 117 "load_ptm.cpp"
           $66 = HEAP8[$65>>0]|0; //@line 117 "load_ptm.cpp"
           $67 = $66&255; //@line 117 "load_ptm.cpp"
           $68 = $67 << 2; //@line 117 "load_ptm.cpp"
           $69 = $68&65535; //@line 117 "load_ptm.cpp"
           $70 = (((((($this)) + 33280|0) + ($55<<6)|0)) + 30|0); //@line 117 "load_ptm.cpp"
           HEAP16[$70>>1] = $69; //@line 117 "load_ptm.cpp"
           $$sum89 = $dwMemPos$049 | 14; //@line 118 "load_ptm.cpp"
           $71 = (($lpStream) + ($$sum89)|0); //@line 118 "load_ptm.cpp"
           $72 = HEAPU8[$71>>0]|(HEAPU8[$71+1>>0]<<8); //@line 118 "load_ptm.cpp"
           $73 = $72&65535; //@line 118 "load_ptm.cpp"
           $74 = $73 << 1; //@line 118 "load_ptm.cpp"
           $75 = (((((($this)) + 33280|0) + ($55<<6)|0)) + 24|0); //@line 118 "load_ptm.cpp"
           HEAP32[$75>>2] = $74; //@line 118 "load_ptm.cpp"
           $76 = (((((($this)) + 33280|0) + ($55<<6)|0)) + 34|0); //@line 119 "load_ptm.cpp"
           HEAP16[$76>>1] = 0; //@line 119 "load_ptm.cpp"
           $77 = HEAP8[$57>>0]|0; //@line 120 "load_ptm.cpp"
           $78 = $77 & 3; //@line 120 "load_ptm.cpp"
           $79 = ($78<<24>>24)==(1); //@line 120 "load_ptm.cpp"
           if ($79) {
            $$sum10 = (($dwMemPos$049) + 22)|0; //@line 123 "load_ptm.cpp"
            $80 = (($lpStream) + ($$sum10)|0); //@line 123 "load_ptm.cpp"
            $$val26 = HEAP16[$80>>1]|0;
            $$sum28 = (($dwMemPos$049) + 24)|0;
            $$idx27 = (($lpStream) + ($$sum28)|0);
            $$idx27$val = HEAP16[$$idx27>>1]|0;
            $81 = $$idx27$val&65535; //@line 63 "load_ptm.cpp"
            $82 = $81 << 16; //@line 63 "load_ptm.cpp"
            $83 = $$val26&65535; //@line 63 "load_ptm.cpp"
            $84 = $82 | $83; //@line 63 "load_ptm.cpp"
            HEAP32[$56>>2] = $84; //@line 123 "load_ptm.cpp"
            $$sum11 = (($dwMemPos$049) + 26)|0; //@line 124 "load_ptm.cpp"
            $85 = (($lpStream) + ($$sum11)|0); //@line 124 "load_ptm.cpp"
            $$val24 = HEAP16[$85>>1]|0;
            $$sum29 = (($dwMemPos$049) + 28)|0;
            $$idx25 = (($lpStream) + ($$sum29)|0);
            $$idx25$val = HEAP16[$$idx25>>1]|0;
            $86 = $$idx25$val&65535; //@line 63 "load_ptm.cpp"
            $87 = $86 << 16; //@line 63 "load_ptm.cpp"
            $88 = $$val24&65535; //@line 63 "load_ptm.cpp"
            $89 = $87 | $88; //@line 63 "load_ptm.cpp"
            $90 = (((((($this)) + 33280|0) + ($55<<6)|0)) + 4|0); //@line 124 "load_ptm.cpp"
            HEAP32[$90>>2] = $89; //@line 124 "load_ptm.cpp"
            $$sum12 = (($dwMemPos$049) + 30)|0; //@line 125 "load_ptm.cpp"
            $91 = (($lpStream) + ($$sum12)|0); //@line 125 "load_ptm.cpp"
            $$val22 = HEAP16[$91>>1]|0;
            $$sum30 = (($dwMemPos$049) + 32)|0;
            $$idx23 = (($lpStream) + ($$sum30)|0);
            $$idx23$val = HEAP16[$$idx23>>1]|0;
            $92 = $$idx23$val&65535; //@line 63 "load_ptm.cpp"
            $93 = $92 << 16; //@line 63 "load_ptm.cpp"
            $94 = $$val22&65535; //@line 63 "load_ptm.cpp"
            $95 = $93 | $94; //@line 63 "load_ptm.cpp"
            $96 = (((((($this)) + 33280|0) + ($55<<6)|0)) + 8|0); //@line 125 "load_ptm.cpp"
            HEAP32[$96>>2] = $95; //@line 125 "load_ptm.cpp"
            $$sum13 = (($dwMemPos$049) + 18)|0; //@line 126 "load_ptm.cpp"
            $97 = (($lpStream) + ($$sum13)|0); //@line 126 "load_ptm.cpp"
            $$val = HEAP16[$97>>1]|0;
            $$sum31 = (($dwMemPos$049) + 20)|0;
            $$idx = (($lpStream) + ($$sum31)|0);
            $$idx$val = HEAP16[$$idx>>1]|0;
            $98 = $$idx$val&65535; //@line 63 "load_ptm.cpp"
            $99 = $98 << 16; //@line 63 "load_ptm.cpp"
            $100 = $$val&65535; //@line 63 "load_ptm.cpp"
            $101 = $99 | $100; //@line 63 "load_ptm.cpp"
            $102 = $77 & 4; //@line 127 "load_ptm.cpp"
            $103 = ($102<<24>>24)==(0); //@line 127 "load_ptm.cpp"
            if ($103) {
             $107 = 0;
            } else {
             HEAP16[$76>>1] = 2; //@line 127 "load_ptm.cpp"
             $107 = 2;
            }
            $104 = $77 & 8; //@line 128 "load_ptm.cpp"
            $105 = ($104<<24>>24)==(0); //@line 128 "load_ptm.cpp"
            if ($105) {
             $113 = $107;
            } else {
             $106 = $107&65535; //@line 128 "load_ptm.cpp"
             $108 = $106 | 4; //@line 128 "load_ptm.cpp"
             $109 = $108&65535; //@line 128 "load_ptm.cpp"
             HEAP16[$76>>1] = $109; //@line 128 "load_ptm.cpp"
             $113 = $109;
            }
            $110 = $77 & 16; //@line 129 "load_ptm.cpp"
            $111 = ($110<<24>>24)==(0); //@line 129 "load_ptm.cpp"
            if ($111) {
             $120 = $84;$smpflg$0 = 2;
            } else {
             $112 = $113&65535; //@line 131 "load_ptm.cpp"
             $114 = $112 | 1; //@line 131 "load_ptm.cpp"
             $115 = $114&65535; //@line 131 "load_ptm.cpp"
             HEAP16[$76>>1] = $115; //@line 131 "load_ptm.cpp"
             $116 = $84 >>> 1; //@line 132 "load_ptm.cpp"
             HEAP32[$56>>2] = $116; //@line 132 "load_ptm.cpp"
             $117 = $89 >>> 1; //@line 133 "load_ptm.cpp"
             HEAP32[$90>>2] = $117; //@line 133 "load_ptm.cpp"
             $118 = $95 >>> 1; //@line 134 "load_ptm.cpp"
             HEAP32[$96>>2] = $118; //@line 134 "load_ptm.cpp"
             $120 = $116;$smpflg$0 = 37;
            }
            $119 = ($120|0)!=(0); //@line 137 "load_ptm.cpp"
            $121 = ($101|0)!=(0); //@line 137 "load_ptm.cpp"
            $or$cond3 = $121 & $119; //@line 137 "load_ptm.cpp"
            $122 = ($101>>>0)<($dwMemLength>>>0); //@line 137 "load_ptm.cpp"
            $or$cond17 = $122 & $or$cond3; //@line 137 "load_ptm.cpp"
            if ($or$cond17) {
             $123 = (($lpStream) + ($101)|0); //@line 139 "load_ptm.cpp"
             $124 = (($dwMemLength) - ($101))|0; //@line 139 "load_ptm.cpp"
             (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$56,$smpflg$0,$123,$124)|0); //@line 139 "load_ptm.cpp"
            }
           }
           $125 = (($dwMemPos$049) + 80)|0; //@line 107 "load_ptm.cpp"
           $126 = HEAP32[$32>>2]|0; //@line 107 "load_ptm.cpp"
           $127 = ($55>>>0)<($126>>>0); //@line 107 "load_ptm.cpp"
           if ($127) {
            $dwMemPos$049 = $125;$ismp$048 = $55;
           } else {
            break;
           }
          }
         }
         $53 = ($19<<16>>16)==(0); //@line 144 "load_ptm.cpp"
         if ($53) {
          $$0 = 1;
         } else {
          $54 = $19&65535; //@line 144 "load_ptm.cpp"
          $ipat$045 = 0;
          while(1) {
           $128 = (((($pfh)) + 352|0) + ($ipat$045<<1)|0); //@line 146 "load_ptm.cpp"
           $129 = HEAPU8[$128>>0]|(HEAPU8[$128+1>>0]<<8); //@line 146 "load_ptm.cpp"
           $130 = $129&65535; //@line 146 "load_ptm.cpp"
           $131 = $130 << 4; //@line 146 "load_ptm.cpp"
           $132 = ($129<<16>>16)!=(0); //@line 147 "load_ptm.cpp"
           $133 = ($131>>>0)<($dwMemLength>>>0); //@line 147 "load_ptm.cpp"
           $or$cond18 = $132 & $133; //@line 147 "load_ptm.cpp"
           L36: do {
            if ($or$cond18) {
             $134 = (((($this)) + 52864|0) + ($ipat$045<<1)|0); //@line 148 "load_ptm.cpp"
             HEAP16[$134>>1] = 64; //@line 148 "load_ptm.cpp"
             $135 = HEAP32[$29>>2]|0; //@line 149 "load_ptm.cpp"
             $136 = (__ZN10CSoundFile15AllocatePatternEjj(64,$135)|0); //@line 149 "load_ptm.cpp"
             $137 = (((($this)) + 51904|0) + ($ipat$045<<2)|0); //@line 149 "load_ptm.cpp"
             HEAP32[$137>>2] = $136; //@line 149 "load_ptm.cpp"
             $138 = ($136|0)==(0|0); //@line 149 "load_ptm.cpp"
             if ($138) {
              $$0 = 1;
              break L4;
             } else {
              $dwMemPos$1$ph44 = $131;$m$0$ph41 = $136;$row$0$ph43 = 0;
             }
             while(1) {
              $dwMemPos$135$us = $dwMemPos$1$ph44;
              while(1) {
               $139 = (($dwMemPos$135$us) + 1)|0; //@line 154 "load_ptm.cpp"
               $140 = (($lpStream) + ($dwMemPos$135$us)|0); //@line 154 "load_ptm.cpp"
               $141 = HEAP8[$140>>0]|0; //@line 154 "load_ptm.cpp"
               $142 = $141&255; //@line 154 "load_ptm.cpp"
               $143 = ($139>>>0)<($dwMemLength>>>0); //@line 156 "load_ptm.cpp"
               if (!($143)) {
                break L36;
               }
               $144 = ($141<<24>>24)==(0); //@line 157 "load_ptm.cpp"
               if ($144) {
                $$lcssa89 = $139;
                break;
               }
               $145 = $142 & 31; //@line 159 "load_ptm.cpp"
               $146 = $142 & 32; //@line 161 "load_ptm.cpp"
               $147 = ($146|0)==(0); //@line 161 "load_ptm.cpp"
               if ($147) {
                $dwMemPos$2$us = $139;
               } else {
                $148 = (($dwMemPos$135$us) + 3)|0; //@line 163 "load_ptm.cpp"
                $149 = ($148>>>0)>($dwMemLength>>>0); //@line 163 "load_ptm.cpp"
                if ($149) {
                 break L36;
                }
                $150 = (($dwMemPos$135$us) + 2)|0; //@line 164 "load_ptm.cpp"
                $151 = (($lpStream) + ($139)|0); //@line 164 "load_ptm.cpp"
                $152 = HEAP8[$151>>0]|0; //@line 164 "load_ptm.cpp"
                $153 = (($m$0$ph41) + (($145*6)|0)|0); //@line 164 "load_ptm.cpp"
                HEAP8[$153>>0] = $152; //@line 164 "load_ptm.cpp"
                $154 = (($lpStream) + ($150)|0); //@line 165 "load_ptm.cpp"
                $155 = HEAP8[$154>>0]|0; //@line 165 "load_ptm.cpp"
                $156 = (((($m$0$ph41) + (($145*6)|0)|0)) + 1|0); //@line 165 "load_ptm.cpp"
                HEAP8[$156>>0] = $155; //@line 165 "load_ptm.cpp"
                $dwMemPos$2$us = $148;
               }
               $157 = $142 & 64; //@line 167 "load_ptm.cpp"
               $158 = ($157|0)==(0); //@line 167 "load_ptm.cpp"
               L48: do {
                if ($158) {
                 $dwMemPos$3$us = $dwMemPos$2$us;
                } else {
                 $159 = (($dwMemPos$2$us) + 2)|0; //@line 169 "load_ptm.cpp"
                 $160 = ($159>>>0)>($dwMemLength>>>0); //@line 169 "load_ptm.cpp"
                 if ($160) {
                  break L36;
                 }
                 $161 = (($dwMemPos$2$us) + 1)|0; //@line 170 "load_ptm.cpp"
                 $162 = (($lpStream) + ($dwMemPos$2$us)|0); //@line 170 "load_ptm.cpp"
                 $163 = HEAP8[$162>>0]|0; //@line 170 "load_ptm.cpp"
                 $164 = (($m$0$ph41) + (($145*6)|0)|0); //@line 170 "load_ptm.cpp"
                 $165 = (((($m$0$ph41) + (($145*6)|0)|0)) + 3|0); //@line 170 "load_ptm.cpp"
                 HEAP8[$165>>0] = $163; //@line 170 "load_ptm.cpp"
                 $166 = (($lpStream) + ($161)|0); //@line 171 "load_ptm.cpp"
                 $167 = HEAP8[$166>>0]|0; //@line 171 "load_ptm.cpp"
                 $168 = (((($m$0$ph41) + (($145*6)|0)|0)) + 5|0); //@line 171 "load_ptm.cpp"
                 HEAP8[$168>>0] = $167; //@line 171 "load_ptm.cpp"
                 $169 = ($163<<24>>24)==(14); //@line 172 "load_ptm.cpp"
                 $170 = $167 & -16;
                 $171 = ($170<<24>>24)==(-128); //@line 172 "load_ptm.cpp"
                 $or$cond21$us = $169 & $171; //@line 172 "load_ptm.cpp"
                 if ($or$cond21$us) {
                  HEAP8[$165>>0] = 20; //@line 174 "load_ptm.cpp"
                  $dwMemPos$3$us = $159;
                  break;
                 }
                 $172 = ($163&255)<(16); //@line 176 "load_ptm.cpp"
                 if ($172) {
                  __ZNK10CSoundFile17ConvertModCommandEP11_MODCOMMAND($this,$164); //@line 178 "load_ptm.cpp"
                  $dwMemPos$3$us = $159;
                  break;
                 }
                 $173 = $163&255; //@line 176 "load_ptm.cpp"
                 switch ($173|0) {
                 case 16:  {
                  HEAP8[$165>>0] = 23; //@line 184 "load_ptm.cpp"
                  $dwMemPos$3$us = $159;
                  break L48;
                  break;
                 }
                 case 17:  {
                  HEAP8[$165>>0] = 15; //@line 187 "load_ptm.cpp"
                  $dwMemPos$3$us = $159;
                  break L48;
                  break;
                 }
                 case 18:  {
                  HEAP8[$165>>0] = 26; //@line 190 "load_ptm.cpp"
                  $dwMemPos$3$us = $159;
                  break L48;
                  break;
                 }
                 default: {
                  HEAP8[$165>>0] = 0; //@line 193 "load_ptm.cpp"
                  $dwMemPos$3$us = $159;
                  break L48;
                 }
                 }
                }
               } while(0);
               $174 = $142 & 128; //@line 197 "load_ptm.cpp"
               $175 = ($174|0)==(0); //@line 197 "load_ptm.cpp"
               if ($175) {
                $dwMemPos$1$be$us = $dwMemPos$3$us;
               } else {
                $176 = ($dwMemPos$3$us>>>0)<($dwMemLength>>>0); //@line 199 "load_ptm.cpp"
                if (!($176)) {
                 break L36;
                }
                $177 = (((($m$0$ph41) + (($145*6)|0)|0)) + 2|0); //@line 200 "load_ptm.cpp"
                HEAP8[$177>>0] = 1; //@line 200 "load_ptm.cpp"
                $178 = (($dwMemPos$3$us) + 1)|0; //@line 201 "load_ptm.cpp"
                $179 = (($lpStream) + ($dwMemPos$3$us)|0); //@line 201 "load_ptm.cpp"
                $180 = HEAP8[$179>>0]|0; //@line 201 "load_ptm.cpp"
                $181 = (((($m$0$ph41) + (($145*6)|0)|0)) + 4|0); //@line 201 "load_ptm.cpp"
                HEAP8[$181>>0] = $180; //@line 201 "load_ptm.cpp"
                $dwMemPos$1$be$us = $178;
               }
               $182 = ($dwMemPos$1$be$us>>>0)<($dwMemLength>>>0); //@line 152 "load_ptm.cpp"
               if ($182) {
                $dwMemPos$135$us = $dwMemPos$1$be$us;
               } else {
                break L36;
               }
              }
              $183 = (($row$0$ph43) + 1)|0; //@line 205 "load_ptm.cpp"
              $184 = HEAP32[$29>>2]|0; //@line 206 "load_ptm.cpp"
              $185 = (($m$0$ph41) + (($184*6)|0)|0); //@line 206 "load_ptm.cpp"
              $186 = ($183>>>0)<(64); //@line 152 "load_ptm.cpp"
              $187 = ($$lcssa89>>>0)<($dwMemLength>>>0); //@line 152 "load_ptm.cpp"
              $or$cond1934 = $186 & $187; //@line 152 "load_ptm.cpp"
              if ($or$cond1934) {
               $dwMemPos$1$ph44 = $$lcssa89;$m$0$ph41 = $185;$row$0$ph43 = $183;
              } else {
               break;
              }
             }
            }
           } while(0);
           $188 = (($ipat$045) + 1)|0; //@line 144 "load_ptm.cpp"
           $189 = ($188>>>0)<($54>>>0); //@line 144 "load_ptm.cpp"
           if ($189) {
            $ipat$045 = $188;
           } else {
            $$0 = 1;
            break;
           }
          }
         }
        } else {
         $$0 = 0;
        }
       }
      }
     }
    }
   }
  } else {
   $$0 = 0;
  }
 } while(0);
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0); //@line 211 "load_ptm.cpp"
}
function __ZN10CSoundFile7ReadUltEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$0 = 0, $$11 = 0, $$12 = 0, $$lcssa119 = 0, $$lcssa119$lcssa = 0, $$lcssa120 = 0, $$lcssa120$lcssa = 0, $$lcssa122 = 0, $$lcssa123 = 0, $$pre = 0, $$pre85 = 0, $$pre86 = 0, $$sum = 0, $$sum10 = 0, $$sum2 = 0, $$sum3 = 0, $$sum4 = 0, $$sum5 = 0, $$sum6 = 0;
 var $$sum7 = 0, $$sum8 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0;
 var $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0;
 var $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cmd1$0 = 0, $cmd2$0$ph = 0, $dat1$0 = 0;
 var $dat2$0$ph = 0, $dwMemPos$0 = 0, $dwMemPos$1$lcssa = 0, $dwMemPos$157 = 0, $dwMemPos$249 = 0, $dwMemPos$3 = 0, $dwMemPos$4$lcssa = 0, $dwMemPos$445 = 0, $dwMemPos$543 = 0, $dwMemPos$641 = 0, $dwMemPos$7 = 0, $dwMemPos$827 = 0, $dwMemPos$9 = 0, $exitcond = 0, $exitcond75 = 0, $exitcond77 = 0, $exitcond78 = 0, $ins$055 = 0, $l$061 = 0, $nAllocPat$047 = 0;
 var $nChn$044 = 0, $nPat$042 = 0, $nSet$052 = 0, $note$0 = 0, $or$cond = 0, $pat$139 = 0, $pat$2$lcssa = 0, $pat$230 = 0, $pat$230$us = 0, $pat$230$us37 = 0, $pat$3 = 0, $pat$3$us = 0, $pat$3$us38 = 0, $rep$0 = 0, $row$040 = 0, $row$1$lcssa = 0, $row$131 = 0, $row$131$us = 0, $row$131$us36 = 0, $smp$028 = 0;
 var $smpsize$0 = 0, $t$048 = 0, $umax74 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ($lpStream|0)==(0|0); //@line 53 "load_ult.cpp"
 $1 = ($dwMemLength>>>0)<(256); //@line 53 "load_ult.cpp"
 $or$cond = $0 | $1; //@line 53 "load_ult.cpp"
 if ($or$cond) {
  $$0 = 0;
  return ($$0|0); //@line 223 "load_ult.cpp"
 }
 $2 = (_strncmp($lpStream,529861,14)|0); //@line 54 "load_ult.cpp"
 $3 = ($2|0)==(0); //@line 54 "load_ult.cpp"
 if (!($3)) {
  $$0 = 0;
  return ($$0|0); //@line 223 "load_ult.cpp"
 }
 $4 = ((($this)) + 59680|0); //@line 57 "load_ult.cpp"
 HEAP32[$4>>2] = 128; //@line 57 "load_ult.cpp"
 $5 = ((($this)) + 59648|0); //@line 58 "load_ult.cpp"
 HEAP32[$5>>2] = 6; //@line 58 "load_ult.cpp"
 $6 = ((($this)) + 59652|0); //@line 59 "load_ult.cpp"
 HEAP32[$6>>2] = 125; //@line 59 "load_ult.cpp"
 $7 = ((($this)) + 59804|0); //@line 60 "load_ult.cpp"
 $8 = ((($lpStream)) + 15|0); //@line 60 "load_ult.cpp"
 dest=$7; src=$8; stop=dest+31|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 60 "load_ult.cpp"
 $9 = ((($this)) + 59835|0); //@line 61 "load_ult.cpp"
 HEAP8[$9>>0] = 0; //@line 61 "load_ult.cpp"
 $10 = ((($lpStream)) + 47|0); //@line 64 "load_ult.cpp"
 $11 = HEAP8[$10>>0]|0; //@line 64 "load_ult.cpp"
 $12 = ($11<<24>>24)==(0); //@line 64 "load_ult.cpp"
 if ($12) {
  $dwMemPos$0 = 48;
 } else {
  $13 = $11&255; //@line 64 "load_ult.cpp"
  $14 = $13 << 5; //@line 64 "load_ult.cpp"
  $15 = (($14) + 48)|0; //@line 64 "load_ult.cpp"
  $16 = ($15>>>0)<($dwMemLength>>>0); //@line 64 "load_ult.cpp"
  if ($16) {
   $17 = $14 | 1; //@line 67 "load_ult.cpp"
   $18 = (($17) + ($13))|0; //@line 67 "load_ult.cpp"
   $19 = (__Znaj($18)|0); //@line 67 "load_ult.cpp"
   $20 = ((($this)) + 59796|0); //@line 67 "load_ult.cpp"
   HEAP32[$20>>2] = $19; //@line 67 "load_ult.cpp"
   $23 = $19;$l$061 = 0;
   while(1) {
    $21 = ($l$061*33)|0; //@line 72 "load_ult.cpp"
    $22 = (($23) + ($21)|0); //@line 72 "load_ult.cpp"
    $24 = $l$061 << 5; //@line 72 "load_ult.cpp"
    $$sum10 = (($24) + 48)|0; //@line 72 "load_ult.cpp"
    $25 = (($lpStream) + ($$sum10)|0); //@line 72 "load_ult.cpp"
    dest=$22; src=$25; stop=dest+32|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 72 "load_ult.cpp"
    $26 = (($21) + 32)|0; //@line 73 "load_ult.cpp"
    $27 = HEAP32[$20>>2]|0; //@line 73 "load_ult.cpp"
    $28 = (($27) + ($26)|0); //@line 73 "load_ult.cpp"
    HEAP8[$28>>0] = 13; //@line 73 "load_ult.cpp"
    $29 = (($l$061) + 1)|0; //@line 70 "load_ult.cpp"
    $30 = HEAP8[$10>>0]|0; //@line 70 "load_ult.cpp"
    $31 = $30&255; //@line 70 "load_ult.cpp"
    $32 = ($29>>>0)<($31>>>0); //@line 70 "load_ult.cpp"
    $33 = HEAP32[$20>>2]|0; //@line 72 "load_ult.cpp"
    if ($32) {
     $23 = $33;$l$061 = $29;
    } else {
     $$lcssa123 = $33;
     break;
    }
   }
   $34 = (($$lcssa123) + ($14)|0); //@line 75 "load_ult.cpp"
   HEAP8[$34>>0] = 0; //@line 75 "load_ult.cpp"
   $dwMemPos$0 = $15;
  } else {
   $dwMemPos$0 = 48;
  }
 }
 $35 = ($dwMemPos$0>>>0)<($dwMemLength>>>0); //@line 79 "load_ult.cpp"
 if (!($35)) {
  $$0 = 1;
  return ($$0|0); //@line 223 "load_ult.cpp"
 }
 $36 = (($dwMemPos$0) + 1)|0; //@line 80 "load_ult.cpp"
 $37 = (($lpStream) + ($dwMemPos$0)|0); //@line 80 "load_ult.cpp"
 $38 = HEAP8[$37>>0]|0; //@line 80 "load_ult.cpp"
 $39 = $38&255; //@line 80 "load_ult.cpp"
 $40 = ((($this)) + 59684|0); //@line 81 "load_ult.cpp"
 $41 = ($38&255)>(239); //@line 82 "load_ult.cpp"
 $$ = $41 ? 239 : $39; //@line 82 "load_ult.cpp"
 HEAP32[$40>>2] = $$; //@line 81 "load_ult.cpp"
 $42 = ((($lpStream)) + 14|0); //@line 84 "load_ult.cpp"
 $43 = HEAP8[$42>>0]|0; //@line 84 "load_ult.cpp"
 $44 = ($43<<24>>24)>(51); //@line 84 "load_ult.cpp"
 $smpsize$0 = $44 ? 66 : 64; //@line 84 "load_ult.cpp"
 $45 = Math_imul($smpsize$0, $39)|0; //@line 85 "load_ult.cpp"
 $46 = (($dwMemPos$0) + 259)|0; //@line 85 "load_ult.cpp"
 $47 = (($46) + ($45))|0; //@line 85 "load_ult.cpp"
 $48 = ($47>>>0)>($dwMemLength>>>0); //@line 85 "load_ult.cpp"
 if ($48) {
  $$0 = 1;
  return ($$0|0); //@line 223 "load_ult.cpp"
 }
 $49 = ($38<<24>>24)==(0); //@line 86 "load_ult.cpp"
 if ($49) {
  $dwMemPos$1$lcssa = $36;
 } else {
  $50 = (($39) + 1)|0; //@line 86 "load_ult.cpp"
  $51 = ($50>>>0)>(2);
  $52 = $51 ? $39 : 1; //@line 86 "load_ult.cpp"
  $53 = Math_imul($smpsize$0, $52)|0; //@line 86 "load_ult.cpp"
  $54 = (($dwMemPos$0) + ($53))|0; //@line 86 "load_ult.cpp"
  $56 = $$;$dwMemPos$157 = $36;$ins$055 = 1;
  while(1) {
   $55 = ($ins$055>>>0)>($56>>>0); //@line 86 "load_ult.cpp"
   if (!($55)) {
    $57 = (($lpStream) + ($dwMemPos$157)|0); //@line 88 "load_ult.cpp"
    $58 = (((($this)) + 59804|0) + ($ins$055<<5)|0); //@line 90 "load_ult.cpp"
    dest=$58; src=$57; stop=dest+32|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 90 "load_ult.cpp"
    $59 = (((((($this)) + 59804|0) + ($ins$055<<5)|0)) + 31|0); //@line 91 "load_ult.cpp"
    HEAP8[$59>>0] = 0; //@line 91 "load_ult.cpp"
    $60 = (((((($this)) + 33280|0) + ($ins$055<<6)|0)) + 42|0); //@line 92 "load_ult.cpp"
    $$sum = (($dwMemPos$157) + 32)|0; //@line 92 "load_ult.cpp"
    $61 = (($lpStream) + ($$sum)|0); //@line 92 "load_ult.cpp"
    dest=$60; src=$61; stop=dest+12|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 92 "load_ult.cpp"
    $$sum2 = (($dwMemPos$157) + 44)|0; //@line 93 "load_ult.cpp"
    $62 = (($lpStream) + ($$sum2)|0); //@line 93 "load_ult.cpp"
    $63 = HEAPU8[$62>>0]|(HEAPU8[$62+1>>0]<<8)|(HEAPU8[$62+2>>0]<<16)|(HEAPU8[$62+3>>0]<<24); //@line 93 "load_ult.cpp"
    $64 = (((((($this)) + 33280|0) + ($ins$055<<6)|0)) + 4|0); //@line 93 "load_ult.cpp"
    HEAP32[$64>>2] = $63; //@line 93 "load_ult.cpp"
    $$sum3 = (($dwMemPos$157) + 48)|0; //@line 94 "load_ult.cpp"
    $65 = (($lpStream) + ($$sum3)|0); //@line 94 "load_ult.cpp"
    $66 = HEAPU8[$65>>0]|(HEAPU8[$65+1>>0]<<8)|(HEAPU8[$65+2>>0]<<16)|(HEAPU8[$65+3>>0]<<24); //@line 94 "load_ult.cpp"
    $67 = (((((($this)) + 33280|0) + ($ins$055<<6)|0)) + 8|0); //@line 94 "load_ult.cpp"
    HEAP32[$67>>2] = $66; //@line 94 "load_ult.cpp"
    $$sum4 = (($dwMemPos$157) + 56)|0; //@line 95 "load_ult.cpp"
    $68 = (($lpStream) + ($$sum4)|0); //@line 95 "load_ult.cpp"
    $69 = HEAPU8[$68>>0]|(HEAPU8[$68+1>>0]<<8)|(HEAPU8[$68+2>>0]<<16)|(HEAPU8[$68+3>>0]<<24); //@line 95 "load_ult.cpp"
    $$sum5 = (($dwMemPos$157) + 52)|0; //@line 95 "load_ult.cpp"
    $70 = (($lpStream) + ($$sum5)|0); //@line 95 "load_ult.cpp"
    $71 = HEAPU8[$70>>0]|(HEAPU8[$70+1>>0]<<8)|(HEAPU8[$70+2>>0]<<16)|(HEAPU8[$70+3>>0]<<24); //@line 95 "load_ult.cpp"
    $72 = (($69) - ($71))|0; //@line 95 "load_ult.cpp"
    $73 = (((($this)) + 33280|0) + ($ins$055<<6)|0); //@line 95 "load_ult.cpp"
    HEAP32[$73>>2] = $72; //@line 95 "load_ult.cpp"
    $$sum6 = (($dwMemPos$157) + 60)|0; //@line 96 "load_ult.cpp"
    $74 = (($lpStream) + ($$sum6)|0); //@line 96 "load_ult.cpp"
    $75 = HEAP8[$74>>0]|0; //@line 96 "load_ult.cpp"
    $76 = $75&255; //@line 96 "load_ult.cpp"
    $77 = (((((($this)) + 33280|0) + ($ins$055<<6)|0)) + 30|0); //@line 96 "load_ult.cpp"
    HEAP16[$77>>1] = $76; //@line 96 "load_ult.cpp"
    $78 = (((((($this)) + 33280|0) + ($ins$055<<6)|0)) + 32|0); //@line 97 "load_ult.cpp"
    HEAP16[$78>>1] = 64; //@line 97 "load_ult.cpp"
    $79 = (((((($this)) + 33280|0) + ($ins$055<<6)|0)) + 24|0); //@line 98 "load_ult.cpp"
    HEAP32[$79>>2] = 8363; //@line 98 "load_ult.cpp"
    $80 = HEAP8[$42>>0]|0; //@line 99 "load_ult.cpp"
    $81 = ($80<<24>>24)>(51); //@line 99 "load_ult.cpp"
    if ($81) {
     $$sum8 = (($dwMemPos$157) + 62)|0; //@line 101 "load_ult.cpp"
     $82 = (($lpStream) + ($$sum8)|0); //@line 101 "load_ult.cpp"
     $83 = HEAPU8[$82>>0]|(HEAPU8[$82+1>>0]<<8); //@line 101 "load_ult.cpp"
     $84 = $83&65535; //@line 101 "load_ult.cpp"
     HEAP32[$79>>2] = $84; //@line 101 "load_ult.cpp"
    }
    $$sum7 = (($dwMemPos$157) + 61)|0; //@line 103 "load_ult.cpp"
    $85 = (($lpStream) + ($$sum7)|0); //@line 103 "load_ult.cpp"
    $86 = HEAP8[$85>>0]|0; //@line 103 "load_ult.cpp"
    $87 = $86 & 8; //@line 103 "load_ult.cpp"
    $88 = ($87<<24>>24)==(0); //@line 103 "load_ult.cpp"
    if (!($88)) {
     $89 = (((((($this)) + 33280|0) + ($ins$055<<6)|0)) + 34|0); //@line 103 "load_ult.cpp"
     $90 = HEAP16[$89>>1]|0; //@line 103 "load_ult.cpp"
     $91 = $90&65535; //@line 103 "load_ult.cpp"
     $92 = $91 | 2; //@line 103 "load_ult.cpp"
     $93 = $92&65535; //@line 103 "load_ult.cpp"
     HEAP16[$89>>1] = $93; //@line 103 "load_ult.cpp"
    }
    $94 = $86 & 16; //@line 104 "load_ult.cpp"
    $95 = ($94<<24>>24)==(0); //@line 104 "load_ult.cpp"
    if (!($95)) {
     $96 = (((((($this)) + 33280|0) + ($ins$055<<6)|0)) + 34|0); //@line 104 "load_ult.cpp"
     $97 = HEAP16[$96>>1]|0; //@line 104 "load_ult.cpp"
     $98 = $97&65535; //@line 104 "load_ult.cpp"
     $99 = $98 | 4; //@line 104 "load_ult.cpp"
     $100 = $99&65535; //@line 104 "load_ult.cpp"
     HEAP16[$96>>1] = $100; //@line 104 "load_ult.cpp"
    }
    $101 = $86 & 4; //@line 105 "load_ult.cpp"
    $102 = ($101<<24>>24)==(0); //@line 105 "load_ult.cpp"
    if (!($102)) {
     $103 = (((((($this)) + 33280|0) + ($ins$055<<6)|0)) + 34|0); //@line 107 "load_ult.cpp"
     $104 = HEAP16[$103>>1]|0; //@line 107 "load_ult.cpp"
     $105 = $104&65535; //@line 107 "load_ult.cpp"
     $106 = $105 | 1; //@line 107 "load_ult.cpp"
     $107 = $106&65535; //@line 107 "load_ult.cpp"
     HEAP16[$103>>1] = $107; //@line 107 "load_ult.cpp"
     $108 = $63 >>> 1; //@line 108 "load_ult.cpp"
     HEAP32[$64>>2] = $108; //@line 108 "load_ult.cpp"
     $109 = $66 >>> 1; //@line 109 "load_ult.cpp"
     HEAP32[$67>>2] = $109; //@line 109 "load_ult.cpp"
    }
   }
   $110 = (($ins$055) + 1)|0; //@line 86 "load_ult.cpp"
   $111 = ($110>>>0)>($39>>>0); //@line 86 "load_ult.cpp"
   if ($111) {
    break;
   }
   $112 = (($dwMemPos$157) + ($smpsize$0))|0; //@line 86 "load_ult.cpp"
   $$pre = HEAP32[$40>>2]|0; //@line 86 "load_ult.cpp"
   $56 = $$pre;$dwMemPos$157 = $112;$ins$055 = $110;
  }
  $113 = (($54) + 1)|0; //@line 86 "load_ult.cpp"
  $dwMemPos$1$lcssa = $113;
 }
 $114 = ((($this)) + 53344|0); //@line 112 "load_ult.cpp"
 $115 = (($lpStream) + ($dwMemPos$1$lcssa)|0); //@line 112 "load_ult.cpp"
 _memcpy(($114|0),($115|0),256)|0; //@line 112 "load_ult.cpp"
 $116 = (($dwMemPos$1$lcssa) + 256)|0; //@line 113 "load_ult.cpp"
 $117 = (($lpStream) + ($116)|0); //@line 114 "load_ult.cpp"
 $118 = HEAP8[$117>>0]|0; //@line 114 "load_ult.cpp"
 $119 = $118&255; //@line 114 "load_ult.cpp"
 $120 = (($119) + 1)|0; //@line 114 "load_ult.cpp"
 $121 = ((($this)) + 59664|0); //@line 114 "load_ult.cpp"
 HEAP32[$121>>2] = $120; //@line 114 "load_ult.cpp"
 $122 = (($dwMemPos$1$lcssa) + 257)|0; //@line 115 "load_ult.cpp"
 $123 = (($lpStream) + ($122)|0); //@line 115 "load_ult.cpp"
 $124 = HEAP8[$123>>0]|0; //@line 115 "load_ult.cpp"
 $125 = $124&255; //@line 115 "load_ult.cpp"
 $126 = (($125) + 1)|0; //@line 115 "load_ult.cpp"
 $127 = (($dwMemPos$1$lcssa) + 258)|0; //@line 116 "load_ult.cpp"
 $128 = ($120>>>0)>(32); //@line 117 "load_ult.cpp"
 if ($128) {
  HEAP32[$121>>2] = 32; //@line 117 "load_ult.cpp"
  $nSet$052 = 0;
 } else {
  $nSet$052 = 0;
 }
 while(1) {
  $129 = (((((($this)) + 49600|0) + (($nSet$052*36)|0)|0)) + 4|0); //@line 121 "load_ult.cpp"
  HEAP32[$129>>2] = 64; //@line 121 "load_ult.cpp"
  $130 = $nSet$052 << 7; //@line 122 "load_ult.cpp"
  $131 = $130 & 128; //@line 122 "load_ult.cpp"
  $132 = $131 ^ 192; //@line 122 "load_ult.cpp"
  $133 = (((($this)) + 49600|0) + (($nSet$052*36)|0)|0); //@line 122 "load_ult.cpp"
  HEAP32[$133>>2] = $132; //@line 122 "load_ult.cpp"
  $134 = (($nSet$052) + 1)|0; //@line 119 "load_ult.cpp"
  $135 = HEAP32[$121>>2]|0; //@line 119 "load_ult.cpp"
  $136 = ($134>>>0)<($135>>>0); //@line 119 "load_ult.cpp"
  if ($136) {
   $nSet$052 = $134;
  } else {
   $$lcssa122 = $135;
   break;
  }
 }
 $137 = HEAP8[$42>>0]|0; //@line 125 "load_ult.cpp"
 $138 = ($137<<24>>24)>(50); //@line 125 "load_ult.cpp"
 if ($138) {
  $139 = (($$lcssa122) + ($127))|0; //@line 127 "load_ult.cpp"
  $140 = ($139>>>0)>($dwMemLength>>>0); //@line 127 "load_ult.cpp"
  if ($140) {
   $$0 = 1;
   return ($$0|0); //@line 223 "load_ult.cpp"
  }
  $141 = ($$lcssa122|0)==(0); //@line 128 "load_ult.cpp"
  if ($141) {
   $dwMemPos$3 = $127;
  } else {
   $dwMemPos$249 = $127;$t$048 = 0;
   while(1) {
    $142 = (($dwMemPos$249) + 1)|0; //@line 130 "load_ult.cpp"
    $143 = (($lpStream) + ($dwMemPos$249)|0); //@line 130 "load_ult.cpp"
    $144 = HEAP8[$143>>0]|0; //@line 130 "load_ult.cpp"
    $145 = $144&255; //@line 130 "load_ult.cpp"
    $146 = $145 << 4; //@line 130 "load_ult.cpp"
    $147 = $146 | 8; //@line 130 "load_ult.cpp"
    $148 = (((($this)) + 49600|0) + (($t$048*36)|0)|0); //@line 130 "load_ult.cpp"
    $149 = ($147>>>0)>(256); //@line 131 "load_ult.cpp"
    $$11 = $149 ? 256 : $147; //@line 131 "load_ult.cpp"
    HEAP32[$148>>2] = $$11; //@line 130 "load_ult.cpp"
    $150 = (($t$048) + 1)|0; //@line 128 "load_ult.cpp"
    $151 = HEAP32[$121>>2]|0; //@line 128 "load_ult.cpp"
    $152 = ($150>>>0)<($151>>>0); //@line 128 "load_ult.cpp"
    if ($152) {
     $dwMemPos$249 = $142;$t$048 = $150;
    } else {
     $dwMemPos$3 = $142;
     break;
    }
   }
  }
 } else {
  $dwMemPos$3 = $127;
 }
 $nAllocPat$047 = 0;
 while(1) {
  $155 = ($nAllocPat$047>>>0)<(240); //@line 137 "load_ult.cpp"
  if ($155) {
   $156 = (((($this)) + 52864|0) + ($nAllocPat$047<<1)|0); //@line 139 "load_ult.cpp"
   HEAP16[$156>>1] = 64; //@line 139 "load_ult.cpp"
   $157 = HEAP32[$121>>2]|0; //@line 140 "load_ult.cpp"
   $158 = (__ZN10CSoundFile15AllocatePatternEjj(64,$157)|0); //@line 140 "load_ult.cpp"
   $159 = (((($this)) + 51904|0) + ($nAllocPat$047<<2)|0); //@line 140 "load_ult.cpp"
   HEAP32[$159>>2] = $158; //@line 140 "load_ult.cpp"
  }
  $160 = (($nAllocPat$047) + 1)|0; //@line 135 "load_ult.cpp"
  $exitcond78 = ($160|0)==($126|0); //@line 135 "load_ult.cpp"
  if ($exitcond78) {
   break;
  } else {
   $nAllocPat$047 = $160;
  }
 }
 $153 = HEAP32[$121>>2]|0; //@line 144 "load_ult.cpp"
 $154 = ($153|0)==(0); //@line 144 "load_ult.cpp"
 L59: do {
  if ($154) {
   $dwMemPos$4$lcssa = $dwMemPos$3;
  } else {
   $284 = $153;$dwMemPos$445 = $dwMemPos$3;$nChn$044 = 0;
   L60: while(1) {
    $285 = $284;$dwMemPos$543 = $dwMemPos$445;$nPat$042 = 0;
    while(1) {
     $163 = ($nPat$042>>>0)<(240); //@line 150 "load_ult.cpp"
     if ($163) {
      $164 = (((($this)) + 51904|0) + ($nPat$042<<2)|0); //@line 152 "load_ult.cpp"
      $165 = HEAP32[$164>>2]|0; //@line 152 "load_ult.cpp"
      $166 = ($165|0)==(0|0); //@line 153 "load_ult.cpp"
      $167 = (($165) + (($nChn$044*6)|0)|0); //@line 153 "load_ult.cpp"
      $$12 = $166 ? 0 : $167; //@line 153 "load_ult.cpp"
      $287 = $285;$dwMemPos$641 = $dwMemPos$543;$pat$139 = $$12;$row$040 = 0;
     } else {
      $287 = $285;$dwMemPos$641 = $dwMemPos$543;$pat$139 = 0;$row$040 = 0;
     }
     while(1) {
      $169 = (($dwMemPos$641) + 6)|0; //@line 158 "load_ult.cpp"
      $170 = ($169>>>0)>($dwMemLength>>>0); //@line 158 "load_ult.cpp"
      if ($170) {
       $$0 = 1;
       break L60;
      }
      $171 = (($dwMemPos$641) + 1)|0; //@line 160 "load_ult.cpp"
      $172 = (($lpStream) + ($dwMemPos$641)|0); //@line 160 "load_ult.cpp"
      $173 = HEAP8[$172>>0]|0; //@line 160 "load_ult.cpp"
      $174 = $173&255; //@line 160 "load_ult.cpp"
      $175 = ($173<<24>>24)==(-4); //@line 161 "load_ult.cpp"
      if ($175) {
       $176 = (($lpStream) + ($171)|0); //@line 163 "load_ult.cpp"
       $177 = HEAP8[$176>>0]|0; //@line 163 "load_ult.cpp"
       $178 = $177&255; //@line 163 "load_ult.cpp"
       $179 = (($dwMemPos$641) + 2)|0; //@line 164 "load_ult.cpp"
       $180 = (($lpStream) + ($179)|0); //@line 164 "load_ult.cpp"
       $181 = HEAP8[$180>>0]|0; //@line 164 "load_ult.cpp"
       $182 = $181&255; //@line 164 "load_ult.cpp"
       $183 = (($dwMemPos$641) + 3)|0; //@line 165 "load_ult.cpp"
       $dwMemPos$7 = $183;$note$0 = $182;$rep$0 = $178;
      } else {
       $dwMemPos$7 = $171;$note$0 = $174;$rep$0 = 1;
      }
      $184 = (($dwMemPos$7) + 1)|0; //@line 167 "load_ult.cpp"
      $185 = (($lpStream) + ($dwMemPos$7)|0); //@line 167 "load_ult.cpp"
      $186 = HEAP8[$185>>0]|0; //@line 167 "load_ult.cpp"
      $187 = (($dwMemPos$7) + 2)|0; //@line 168 "load_ult.cpp"
      $188 = (($lpStream) + ($184)|0); //@line 168 "load_ult.cpp"
      $189 = HEAP8[$188>>0]|0; //@line 168 "load_ult.cpp"
      $190 = $189&255; //@line 168 "load_ult.cpp"
      $191 = (($dwMemPos$7) + 3)|0; //@line 169 "load_ult.cpp"
      $192 = (($lpStream) + ($187)|0); //@line 169 "load_ult.cpp"
      $193 = HEAP8[$192>>0]|0; //@line 169 "load_ult.cpp"
      $194 = $193&255; //@line 169 "load_ult.cpp"
      $195 = (($dwMemPos$7) + 4)|0; //@line 170 "load_ult.cpp"
      $196 = (($lpStream) + ($191)|0); //@line 170 "load_ult.cpp"
      $197 = HEAP8[$196>>0]|0; //@line 170 "load_ult.cpp"
      $198 = $197&255; //@line 170 "load_ult.cpp"
      $199 = $190 & 15; //@line 171 "load_ult.cpp"
      $200 = $190 >>> 4; //@line 172 "load_ult.cpp"
      switch ($199|0) {
      case 12:  {
       $201 = $194 >>> 2; //@line 173 "load_ult.cpp"
       $cmd1$0 = 12;$dat1$0 = $201;
       break;
      }
      case 11:  {
       $cmd1$0 = 0;$dat1$0 = 0;
       break;
      }
      default: {
       $cmd1$0 = $199;$dat1$0 = $194;
      }
      }
      switch ($200|0) {
      case 12:  {
       $202 = $198 >>> 2; //@line 175 "load_ult.cpp"
       $cmd2$0$ph = 12;$dat2$0$ph = $202;
       break;
      }
      case 11:  {
       $cmd2$0$ph = 0;$dat2$0$ph = 0;
       break;
      }
      default: {
       $cmd2$0$ph = $200;$dat2$0$ph = $198;
      }
      }
      $203 = ($rep$0|0)!=(0); //@line 177 "load_ult.cpp"
      $204 = ($row$040>>>0)<(64); //@line 177 "load_ult.cpp"
      $205 = $204 & $203; //@line 177 "load_ult.cpp"
      L80: do {
       if ($205) {
        $206 = ($note$0|0)==(0); //@line 182 "load_ult.cpp"
        $207 = $cmd1$0 | $dat1$0; //@line 183 "load_ult.cpp"
        $208 = ($207|0)==(0); //@line 183 "load_ult.cpp"
        $209 = ($cmd2$0$ph|0)==(12); //@line 196 "load_ult.cpp"
        $210 = $dat2$0$ph&255; //@line 199 "load_ult.cpp"
        $211 = $cmd2$0$ph | $dat2$0$ph; //@line 201 "load_ult.cpp"
        $212 = ($211|0)==(0); //@line 201 "load_ult.cpp"
        $213 = $cmd2$0$ph&255; //@line 203 "load_ult.cpp"
        $214 = ($cmd1$0|0)==(12); //@line 185 "load_ult.cpp"
        $215 = $dat1$0&255; //@line 188 "load_ult.cpp"
        $216 = $cmd1$0&255; //@line 191 "load_ult.cpp"
        $217 = (($note$0) + 36)|0; //@line 182 "load_ult.cpp"
        $218 = $217&255; //@line 182 "load_ult.cpp"
        $219 = (0 - ($rep$0))|0; //@line 177 "load_ult.cpp"
        $220 = (($row$040) + -64)|0; //@line 177 "load_ult.cpp"
        $221 = ($220>>>0)<($219>>>0);
        $umax74 = $221 ? $219 : $220;
        $222 = (($row$040) - ($umax74))|0; //@line 177 "load_ult.cpp"
        if ($206) {
         $289 = $287;$pat$230$us = $pat$139;$row$131$us = $row$040;
         while(1) {
          $223 = ($pat$230$us|0)==(0|0); //@line 179 "load_ult.cpp"
          if ($223) {
           $292 = $289;$pat$3$us = 0;
          } else {
           $224 = ((($pat$230$us)) + 1|0); //@line 181 "load_ult.cpp"
           HEAP8[$224>>0] = $186; //@line 181 "load_ult.cpp"
           do {
            if (!($208)) {
             if ($214) {
              $227 = ((($pat$230$us)) + 2|0); //@line 187 "load_ult.cpp"
              HEAP8[$227>>0] = 1; //@line 187 "load_ult.cpp"
              $228 = ((($pat$230$us)) + 4|0); //@line 188 "load_ult.cpp"
              HEAP8[$228>>0] = $215; //@line 188 "load_ult.cpp"
              break;
             } else {
              $225 = ((($pat$230$us)) + 3|0); //@line 191 "load_ult.cpp"
              HEAP8[$225>>0] = $216; //@line 191 "load_ult.cpp"
              $226 = ((($pat$230$us)) + 5|0); //@line 192 "load_ult.cpp"
              HEAP8[$226>>0] = $215; //@line 192 "load_ult.cpp"
              __ZNK10CSoundFile17ConvertModCommandEP11_MODCOMMAND($this,$pat$230$us); //@line 193 "load_ult.cpp"
              break;
             }
            }
           } while(0);
           do {
            if ($209) {
             $233 = ((($pat$230$us)) + 2|0); //@line 198 "load_ult.cpp"
             HEAP8[$233>>0] = 1; //@line 198 "load_ult.cpp"
             $234 = ((($pat$230$us)) + 4|0); //@line 199 "load_ult.cpp"
             HEAP8[$234>>0] = $210; //@line 199 "load_ult.cpp"
            } else {
             if ($212) {
              break;
             }
             $229 = ((($pat$230$us)) + 3|0); //@line 201 "load_ult.cpp"
             $230 = HEAP8[$229>>0]|0; //@line 201 "load_ult.cpp"
             $231 = ($230<<24>>24)==(0); //@line 201 "load_ult.cpp"
             if (!($231)) {
              break;
             }
             HEAP8[$229>>0] = $213; //@line 203 "load_ult.cpp"
             $232 = ((($pat$230$us)) + 5|0); //@line 204 "load_ult.cpp"
             HEAP8[$232>>0] = $210; //@line 204 "load_ult.cpp"
             __ZNK10CSoundFile17ConvertModCommandEP11_MODCOMMAND($this,$pat$230$us); //@line 205 "load_ult.cpp"
            }
           } while(0);
           $235 = HEAP32[$121>>2]|0; //@line 207 "load_ult.cpp"
           $236 = (($pat$230$us) + (($235*6)|0)|0); //@line 207 "load_ult.cpp"
           $292 = $235;$pat$3$us = $236;
          }
          $237 = (($row$131$us) + 1)|0; //@line 209 "load_ult.cpp"
          $exitcond75 = ($237|0)==($222|0); //@line 177 "load_ult.cpp"
          if ($exitcond75) {
           $288 = $292;$pat$2$lcssa = $pat$3$us;$row$1$lcssa = $222;
           break L80;
          } else {
           $289 = $292;$pat$230$us = $pat$3$us;$row$131$us = $237;
          }
         }
        }
        if ($208) {
         $290 = $287;$pat$230$us37 = $pat$139;$row$131$us36 = $row$040;
         while(1) {
          $238 = ($pat$230$us37|0)==(0|0); //@line 179 "load_ult.cpp"
          if ($238) {
           $293 = $290;$pat$3$us38 = 0;
          } else {
           $239 = ((($pat$230$us37)) + 1|0); //@line 181 "load_ult.cpp"
           HEAP8[$239>>0] = $186; //@line 181 "load_ult.cpp"
           HEAP8[$pat$230$us37>>0] = $218; //@line 182 "load_ult.cpp"
           do {
            if ($209) {
             $244 = ((($pat$230$us37)) + 2|0); //@line 198 "load_ult.cpp"
             HEAP8[$244>>0] = 1; //@line 198 "load_ult.cpp"
             $245 = ((($pat$230$us37)) + 4|0); //@line 199 "load_ult.cpp"
             HEAP8[$245>>0] = $210; //@line 199 "load_ult.cpp"
             $247 = $290;
            } else {
             if ($212) {
              $247 = $290;
              break;
             }
             $240 = ((($pat$230$us37)) + 3|0); //@line 201 "load_ult.cpp"
             $241 = HEAP8[$240>>0]|0; //@line 201 "load_ult.cpp"
             $242 = ($241<<24>>24)==(0); //@line 201 "load_ult.cpp"
             if (!($242)) {
              $247 = $290;
              break;
             }
             HEAP8[$240>>0] = $213; //@line 203 "load_ult.cpp"
             $243 = ((($pat$230$us37)) + 5|0); //@line 204 "load_ult.cpp"
             HEAP8[$243>>0] = $210; //@line 204 "load_ult.cpp"
             __ZNK10CSoundFile17ConvertModCommandEP11_MODCOMMAND($this,$pat$230$us37); //@line 205 "load_ult.cpp"
             $$pre85 = HEAP32[$121>>2]|0; //@line 207 "load_ult.cpp"
             $247 = $$pre85;
            }
           } while(0);
           $246 = (($pat$230$us37) + (($247*6)|0)|0); //@line 207 "load_ult.cpp"
           $293 = $247;$pat$3$us38 = $246;
          }
          $248 = (($row$131$us36) + 1)|0; //@line 209 "load_ult.cpp"
          $exitcond77 = ($248|0)==($222|0); //@line 177 "load_ult.cpp"
          if ($exitcond77) {
           $288 = $293;$pat$2$lcssa = $pat$3$us38;$row$1$lcssa = $222;
           break L80;
          } else {
           $290 = $293;$pat$230$us37 = $pat$3$us38;$row$131$us36 = $248;
          }
         }
        } else {
         $291 = $287;$pat$230 = $pat$139;$row$131 = $row$040;
        }
        while(1) {
         $249 = ($pat$230|0)==(0|0); //@line 179 "load_ult.cpp"
         if ($249) {
          $294 = $291;$pat$3 = 0;
         } else {
          $250 = ((($pat$230)) + 1|0); //@line 181 "load_ult.cpp"
          HEAP8[$250>>0] = $186; //@line 181 "load_ult.cpp"
          HEAP8[$pat$230>>0] = $218; //@line 182 "load_ult.cpp"
          if ($214) {
           $251 = ((($pat$230)) + 2|0); //@line 187 "load_ult.cpp"
           HEAP8[$251>>0] = 1; //@line 187 "load_ult.cpp"
           $252 = ((($pat$230)) + 4|0); //@line 188 "load_ult.cpp"
           HEAP8[$252>>0] = $215; //@line 188 "load_ult.cpp"
          } else {
           $253 = ((($pat$230)) + 3|0); //@line 191 "load_ult.cpp"
           HEAP8[$253>>0] = $216; //@line 191 "load_ult.cpp"
           $254 = ((($pat$230)) + 5|0); //@line 192 "load_ult.cpp"
           HEAP8[$254>>0] = $215; //@line 192 "load_ult.cpp"
           __ZNK10CSoundFile17ConvertModCommandEP11_MODCOMMAND($this,$pat$230); //@line 193 "load_ult.cpp"
          }
          do {
           if ($209) {
            $255 = ((($pat$230)) + 2|0); //@line 198 "load_ult.cpp"
            HEAP8[$255>>0] = 1; //@line 198 "load_ult.cpp"
            $256 = ((($pat$230)) + 4|0); //@line 199 "load_ult.cpp"
            HEAP8[$256>>0] = $210; //@line 199 "load_ult.cpp"
           } else {
            if ($212) {
             break;
            }
            $257 = ((($pat$230)) + 3|0); //@line 201 "load_ult.cpp"
            $258 = HEAP8[$257>>0]|0; //@line 201 "load_ult.cpp"
            $259 = ($258<<24>>24)==(0); //@line 201 "load_ult.cpp"
            if (!($259)) {
             break;
            }
            HEAP8[$257>>0] = $213; //@line 203 "load_ult.cpp"
            $260 = ((($pat$230)) + 5|0); //@line 204 "load_ult.cpp"
            HEAP8[$260>>0] = $210; //@line 204 "load_ult.cpp"
            __ZNK10CSoundFile17ConvertModCommandEP11_MODCOMMAND($this,$pat$230); //@line 205 "load_ult.cpp"
           }
          } while(0);
          $261 = HEAP32[$121>>2]|0; //@line 207 "load_ult.cpp"
          $262 = (($pat$230) + (($261*6)|0)|0); //@line 207 "load_ult.cpp"
          $294 = $261;$pat$3 = $262;
         }
         $263 = (($row$131) + 1)|0; //@line 209 "load_ult.cpp"
         $exitcond = ($263|0)==($222|0); //@line 177 "load_ult.cpp"
         if ($exitcond) {
          $288 = $294;$pat$2$lcssa = $pat$3;$row$1$lcssa = $222;
          break;
         } else {
          $291 = $294;$pat$230 = $pat$3;$row$131 = $263;
         }
        }
       } else {
        $288 = $287;$pat$2$lcssa = $pat$139;$row$1$lcssa = $row$040;
       }
      } while(0);
      $168 = ($row$1$lcssa>>>0)<(64); //@line 156 "load_ult.cpp"
      if ($168) {
       $287 = $288;$dwMemPos$641 = $195;$pat$139 = $pat$2$lcssa;$row$040 = $row$1$lcssa;
      } else {
       $$lcssa119 = $195;$$lcssa120 = $288;
       break;
      }
     }
     $264 = (($nPat$042) + 1)|0; //@line 146 "load_ult.cpp"
     $265 = ($264>>>0)<($126>>>0); //@line 146 "load_ult.cpp"
     if ($265) {
      $285 = $$lcssa120;$dwMemPos$543 = $$lcssa119;$nPat$042 = $264;
     } else {
      $$lcssa119$lcssa = $$lcssa119;$$lcssa120$lcssa = $$lcssa120;
      break;
     }
    }
    $266 = (($nChn$044) + 1)|0; //@line 144 "load_ult.cpp"
    $267 = ($266>>>0)<($$lcssa120$lcssa>>>0); //@line 144 "load_ult.cpp"
    if ($267) {
     $284 = $$lcssa120$lcssa;$dwMemPos$445 = $$lcssa119$lcssa;$nChn$044 = $266;
    } else {
     $dwMemPos$4$lcssa = $$lcssa119$lcssa;
     break L59;
    }
   }
   return ($$0|0); //@line 223 "load_ult.cpp"
  }
 } while(0);
 $161 = HEAP32[$40>>2]|0; //@line 216 "load_ult.cpp"
 $162 = ($161|0)==(0); //@line 216 "load_ult.cpp"
 if ($162) {
  $$0 = 1;
  return ($$0|0); //@line 223 "load_ult.cpp"
 } else {
  $286 = $161;$dwMemPos$827 = $dwMemPos$4$lcssa;$smp$028 = 1;
 }
 while(1) {
  $268 = (((($this)) + 33280|0) + ($smp$028<<6)|0); //@line 216 "load_ult.cpp"
  $269 = HEAP32[$268>>2]|0; //@line 216 "load_ult.cpp"
  $270 = ($269|0)==(0); //@line 216 "load_ult.cpp"
  if ($270) {
   $283 = $286;$dwMemPos$9 = $dwMemPos$827;
  } else {
   $271 = ($dwMemPos$827>>>0)<($dwMemLength>>>0); //@line 218 "load_ult.cpp"
   if (!($271)) {
    $$0 = 1;
    label = 90;
    break;
   }
   $272 = (((((($this)) + 33280|0) + ($smp$028<<6)|0)) + 34|0); //@line 219 "load_ult.cpp"
   $273 = HEAP16[$272>>1]|0; //@line 219 "load_ult.cpp"
   $274 = $273 & 1; //@line 219 "load_ult.cpp"
   $275 = ($274<<16>>16)!=(0); //@line 219 "load_ult.cpp"
   $276 = $275 ? 5 : 0; //@line 219 "load_ult.cpp"
   $277 = (($lpStream) + ($dwMemPos$827)|0); //@line 220 "load_ult.cpp"
   $278 = (($dwMemLength) - ($dwMemPos$827))|0; //@line 220 "load_ult.cpp"
   $279 = (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$268,$276,$277,$278)|0); //@line 220 "load_ult.cpp"
   $280 = (($279) + ($dwMemPos$827))|0; //@line 220 "load_ult.cpp"
   $$pre86 = HEAP32[$40>>2]|0; //@line 216 "load_ult.cpp"
   $283 = $$pre86;$dwMemPos$9 = $280;
  }
  $281 = (($smp$028) + 1)|0; //@line 216 "load_ult.cpp"
  $282 = ($281>>>0)>($283>>>0); //@line 216 "load_ult.cpp"
  if ($282) {
   $$0 = 1;
   label = 90;
   break;
  } else {
   $286 = $283;$dwMemPos$827 = $dwMemPos$9;$smp$028 = $281;
  }
 }
 if ((label|0) == 90) {
  return ($$0|0); //@line 223 "load_ult.cpp"
 }
 return (0)|0;
}
function __ZN10CSoundFile7ReadDMFEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$$14 = 0, $$0 = 0, $$14 = 0, $$2 = 0, $$3 = 0, $$42 = 0, $$47 = 0, $$48 = 0, $$49 = 0, $$5 = 0, $$50 = 0, $$7 = 0, $$cmd$sroa$0$5 = 0, $$cmd$sroa$0$6 = 0, $$in = 0, $$lcssa = 0, $$lcssa147 = 0, $$lobit = 0, $$op = 0;
 var $$op36 = 0, $$pr = 0, $$pr128 = 0, $$pre = 0, $$pre$phi127Z2D = 0, $$pre116 = 0, $$pre118 = 0, $$pre119$pre = 0, $$pre120 = 0, $$pre121 = 0, $$pre125 = 0, $$pre126 = 0, $$sum = 0, $$sum16 = 0, $$sum17 = 0, $$sum18 = 0, $$sum19 = 0, $$sum20 = 0, $$sum21 = 0, $$sum22 = 0;
 var $$sum24 = 0, $$sum26 = 0, $$sum27 = 0, $$sum28 = 0, $$sum34 = 0, $$sum35 = 0, $$sum37 = 0, $$sum38 = 0, $$sum40 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0;
 var $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0;
 var $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0;
 var $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0;
 var $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0;
 var $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0;
 var $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0;
 var $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0;
 var $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0;
 var $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0;
 var $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0;
 var $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0;
 var $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0;
 var $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0;
 var $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cmd$sroa$0$5 = 0, $cmd$sroa$0$6 = 0, $cmd$sroa$0$7 = 0, $cmd$sroa$15$4 = 0, $cmd$sroa$16$4 = 0, $cmd$sroa$16$5 = 0, $cmd$sroa$19$4 = 0, $cmd$sroa$19$4$ = 0, $cmd$sroa$19$5 = 0;
 var $cmd$sroa$19$6 = 0, $cmd$sroa$35$4 = 0, $cmd$sroa$35$5 = 0, $cmd$sroa$37$4 = 0, $cmd$sroa$37$4$ = 0, $cmd$sroa$37$5 = 0, $cmd$sroa$37$6 = 0, $d$088 = 0, $d$1 = 0, $d$10 = 0, $d$11 = 0, $d$12 = 0, $d$13 = 0, $d$2 = 0, $d$3 = 0, $d$468 = 0, $d$5 = 0, $d$6 = 0, $d$7 = 0, $d$8 = 0;
 var $d$9 = 0, $dwMemPos$0$ph$be = 0, $dwMemPos$0$ph111 = 0, $dwMemPos$1 = 0, $dwPos$094 = 0, $dwPos8$061142 = 0, $dwPos9$0103 = 0, $exitcond = 0, $flags$0 = 0, $glbinfobyte$083 = 0, $glbinfobyte$1 = 0, $glbinfobyte$2 = 0, $i$060 = 0, $i1$0100 = 0, $i2$066 = 0, $iSmp$062141 = 0, $iSmp10$0105 = 0, $ich$079 = 0, $infobyte = 0, $infoval$0 = 0;
 var $ismpd$0104 = 0, $modtempo$2$off = 0, $modtempo$278 = 0, $modtempo$3 = 0, $npat$093 = 0, $or$cond = 0, $or$cond11 = 0, $or$cond129 = 0, $or$cond135 = 0, $or$cond43 = 0, $or$cond44 = 0, $or$cond9 = 0, $pbeat$084 = 0, $pbeat$1 = 0, $row$086 = 0, $smplflags = 0, $speed$0$lcssa = 0, $speed$075 = 0, $speed$177 = 0, $speed$2 = 0;
 var $switch$split102D = 0, $switch$split12D = 0, $switch$split132D = 0, $switch$split2D = 0, $switch$split42D = 0, $switch$split72D = 0, $switch$tableidx = 0, $tempo$082 = 0, $tempo$1 = 0, $tempochange$0$off085 = 0, $tempochange$1$off0 = 0, $ttype$087 = 0, $ttype$1 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 272|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $infobyte = sp + 240|0;
 $smplflags = sp;
 $0 = ($lpStream|0)==(0|0); //@line 96 "load_dmf.cpp"
 $1 = ($dwMemLength>>>0)<(1024); //@line 96 "load_dmf.cpp"
 $or$cond = $0 | $1; //@line 96 "load_dmf.cpp"
 if ($or$cond) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 483 "load_dmf.cpp"
 }
 $2 = HEAPU8[$lpStream>>0]|(HEAPU8[$lpStream+1>>0]<<8)|(HEAPU8[$lpStream+2>>0]<<16)|(HEAPU8[$lpStream+3>>0]<<24); //@line 97 "load_dmf.cpp"
 $3 = ($2|0)==(1179468868); //@line 97 "load_dmf.cpp"
 if (!($3)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 483 "load_dmf.cpp"
 }
 $4 = ((($lpStream)) + 4|0); //@line 97 "load_dmf.cpp"
 $5 = HEAP8[$4>>0]|0; //@line 97 "load_dmf.cpp"
 $6 = (($5) + -1)<<24>>24; //@line 97 "load_dmf.cpp"
 $7 = ($6&255)>(14); //@line 97 "load_dmf.cpp"
 if ($7) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 483 "load_dmf.cpp"
 }
 $8 = ((($this)) + 59804|0); //@line 99 "load_dmf.cpp"
 $9 = ((($lpStream)) + 13|0); //@line 99 "load_dmf.cpp"
 dest=$8; src=$9; stop=dest+30|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 99 "load_dmf.cpp"
 $10 = ((($this)) + 59834|0); //@line 100 "load_dmf.cpp"
 HEAP8[$10>>0] = 0; //@line 100 "load_dmf.cpp"
 $11 = ((($this)) + 59680|0); //@line 101 "load_dmf.cpp"
 HEAP32[$11>>2] = 131072; //@line 101 "load_dmf.cpp"
 $12 = ((($this)) + 59664|0); //@line 102 "load_dmf.cpp"
 HEAP32[$12>>2] = 0; //@line 102 "load_dmf.cpp"
 $13 = ($dwMemLength>>>0)>(73); //@line 106 "load_dmf.cpp"
 if ($13) {
  $14 = ((($lpStream)) + 66|0); //@line 108 "load_dmf.cpp"
  $15 = ((($this)) + 59684|0); //@line 426 "load_dmf.cpp"
  $16 = ((($this)) + 59736|0); //@line 145 "load_dmf.cpp"
  $17 = ((($this)) + 59796|0); //@line 119 "load_dmf.cpp"
  $18 = ((($smplflags)) + 1|0); //@line 390 "load_dmf.cpp"
  $$in = $14;$dwMemPos$0$ph111 = 66;
  L12: while(1) {
   $19 = HEAP32[$$in>>2]|0; //@line 108 "load_dmf.cpp"
   L14: while(1) {
    $switch$split2D = ($19|0)<(1229999443);
    if ($switch$split2D) {
     label = 146;
     break;
    }
    $switch$split42D = ($19|0)<(1414807888);
    if ($switch$split42D) {
     label = 150;
     break;
    }
    $switch$split132D = ($19|0)<(1431389523);
    if (!($switch$split132D)) {
     label = 157;
     break;
    }
    switch ($19|0) {
    case 1414807888:  {
     break;
    }
    default: {
     label = 139;
     break L14;
    }
    }
    $71 = HEAP32[$12>>2]|0; //@line 153 "load_dmf.cpp"
    $72 = ($71|0)==(0); //@line 153 "load_dmf.cpp"
    if ($72) {
     label = 30;
     break;
    }
   }
   L20: do {
    if ((label|0) == 30) {
     label = 0;
     $73 = (($dwMemPos$0$ph111) + 11)|0; //@line 157 "load_dmf.cpp"
     $$sum22 = (($dwMemPos$0$ph111) + 4)|0; //@line 158 "load_dmf.cpp"
     $74 = (($lpStream) + ($$sum22)|0); //@line 158 "load_dmf.cpp"
     $75 = HEAPU8[$74>>0]|(HEAPU8[$74+1>>0]<<8)|(HEAPU8[$74+2>>0]<<16)|(HEAPU8[$74+3>>0]<<24); //@line 158 "load_dmf.cpp"
     $76 = ($75>>>0)<($dwMemLength>>>0); //@line 158 "load_dmf.cpp"
     if (!($76)) {
      break L12;
     }
     $77 = (($dwMemPos$0$ph111) + 8)|0; //@line 158 "load_dmf.cpp"
     $78 = (($77) + ($75))|0; //@line 158 "load_dmf.cpp"
     $79 = ($78>>>0)>($dwMemLength>>>0); //@line 158 "load_dmf.cpp"
     if ($79) {
      break L12;
     }
     $80 = (($lpStream) + ($77)|0); //@line 159 "load_dmf.cpp"
     $81 = HEAPU8[$80>>0]|(HEAPU8[$80+1>>0]<<8); //@line 159 "load_dmf.cpp"
     $82 = $81&65535; //@line 159 "load_dmf.cpp"
     $83 = ($82>>>0)>(240); //@line 160 "load_dmf.cpp"
     $$2 = $83 ? 240 : $82; //@line 160 "load_dmf.cpp"
     $$sum24 = (($dwMemPos$0$ph111) + 10)|0; //@line 161 "load_dmf.cpp"
     $84 = (($lpStream) + ($$sum24)|0); //@line 161 "load_dmf.cpp"
     $85 = HEAP8[$84>>0]|0; //@line 161 "load_dmf.cpp"
     $86 = $85&255; //@line 161 "load_dmf.cpp"
     HEAP32[$12>>2] = $86; //@line 161 "load_dmf.cpp"
     $87 = (($lpStream) + ($73)|0); //@line 162 "load_dmf.cpp"
     $88 = HEAP8[$87>>0]|0; //@line 162 "load_dmf.cpp"
     $89 = ($85&255)<($88&255); //@line 162 "load_dmf.cpp"
     if ($89) {
      $90 = $88&255; //@line 162 "load_dmf.cpp"
      HEAP32[$12>>2] = $90; //@line 162 "load_dmf.cpp"
      $$pr = $90;
     } else {
      $$pr = $86;
     }
     $91 = ($$pr>>>0)>(32); //@line 163 "load_dmf.cpp"
     if ($91) {
      HEAP32[$12>>2] = 32; //@line 163 "load_dmf.cpp"
     } else {
      $92 = ($$pr>>>0)<(4); //@line 164 "load_dmf.cpp"
      if ($92) {
       HEAP32[$12>>2] = 4; //@line 164 "load_dmf.cpp"
      }
     }
     $93 = ($$2|0)==(0); //@line 165 "load_dmf.cpp"
     L50: do {
      if (!($93)) {
       $dwPos$094 = $73;$npat$093 = 0;
       while(1) {
        $94 = (($lpStream) + ($dwPos$094)|0); //@line 167 "load_dmf.cpp"
        $95 = HEAP8[$94>>0]|0; //@line 171 "load_dmf.cpp"
        $96 = $95&255; //@line 171 "load_dmf.cpp"
        $97 = ($96>>>0)>(32); //@line 172 "load_dmf.cpp"
        $$3 = $97 ? 32 : $96; //@line 172 "load_dmf.cpp"
        $$sum26 = (($dwPos$094) + 2)|0; //@line 173 "load_dmf.cpp"
        $98 = (($lpStream) + ($$sum26)|0); //@line 173 "load_dmf.cpp"
        $99 = HEAPU8[$98>>0]|(HEAPU8[$98+1>>0]<<8); //@line 173 "load_dmf.cpp"
        $100 = $99&65535; //@line 173 "load_dmf.cpp"
        $101 = ($100>>>0)>(256); //@line 174 "load_dmf.cpp"
        $$14 = $101 ? 256 : $100; //@line 174 "load_dmf.cpp"
        $102 = ($$14>>>0)<(16); //@line 175 "load_dmf.cpp"
        $$$14 = $102 ? 16 : $$14; //@line 175 "load_dmf.cpp"
        $103 = (($dwPos$094) + 8)|0; //@line 176 "load_dmf.cpp"
        $$sum27 = (($dwPos$094) + 4)|0; //@line 177 "load_dmf.cpp"
        $104 = (($lpStream) + ($$sum27)|0); //@line 177 "load_dmf.cpp"
        $105 = HEAPU8[$104>>0]|(HEAPU8[$104+1>>0]<<8)|(HEAPU8[$104+2>>0]<<16)|(HEAPU8[$104+3>>0]<<24); //@line 177 "load_dmf.cpp"
        $106 = ($105>>>0)<($dwMemLength>>>0); //@line 177 "load_dmf.cpp"
        if (!($106)) {
         break L50;
        }
        $107 = (($dwPos$094) + 12)|0; //@line 177 "load_dmf.cpp"
        $108 = (($107) + ($105))|0; //@line 177 "load_dmf.cpp"
        $109 = ($108>>>0)<($dwMemLength>>>0); //@line 177 "load_dmf.cpp"
        if (!($109)) {
         break L50;
        }
        $110 = $$$14&65535; //@line 178 "load_dmf.cpp"
        $111 = (((($this)) + 52864|0) + ($npat$093<<1)|0); //@line 178 "load_dmf.cpp"
        HEAP16[$111>>1] = $110; //@line 178 "load_dmf.cpp"
        $112 = HEAP32[$12>>2]|0; //@line 179 "load_dmf.cpp"
        $113 = (__ZN10CSoundFile15AllocatePatternEjj($$$14,$112)|0); //@line 179 "load_dmf.cpp"
        $114 = ($113|0)==(0|0); //@line 180 "load_dmf.cpp"
        if ($114) {
         break L12;
        }
        $115 = (((($this)) + 51904|0) + ($npat$093<<2)|0); //@line 181 "load_dmf.cpp"
        HEAP32[$115>>2] = $113; //@line 181 "load_dmf.cpp"
        $116 = HEAPU8[$104>>0]|(HEAPU8[$104+1>>0]<<8)|(HEAPU8[$104+2>>0]<<16)|(HEAPU8[$104+3>>0]<<24); //@line 183 "load_dmf.cpp"
        $117 = (($116) + ($103))|0; //@line 183 "load_dmf.cpp"
        $$sum28 = (($dwPos$094) + 1)|0; //@line 187 "load_dmf.cpp"
        $118 = (($lpStream) + ($$sum28)|0); //@line 187 "load_dmf.cpp"
        $119 = HEAP8[$118>>0]|0; //@line 187 "load_dmf.cpp"
        $120 = $119&255; //@line 187 "load_dmf.cpp"
        dest=$infobyte; stop=dest+32|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 189 "load_dmf.cpp"
        $121 = ($$$14|0)==(0); //@line 190 "load_dmf.cpp"
        if (!($121)) {
         $122 = ($119&255)>(15); //@line 188 "load_dmf.cpp"
         $123 = $120 & 240; //@line 187 "load_dmf.cpp"
         $124 = ($123|0)==(0); //@line 187 "load_dmf.cpp"
         $125 = $120 >>> 4; //@line 187 "load_dmf.cpp"
         $$42 = $124 ? 8 : $125; //@line 187 "load_dmf.cpp"
         $126 = ($$3|0)==(0); //@line 215 "load_dmf.cpp"
         $d$088 = $103;$glbinfobyte$083 = 0;$pbeat$084 = $$42;$row$086 = 0;$tempo$082 = 125;$tempochange$0$off085 = $122;$ttype$087 = 1;
         while(1) {
          $127 = HEAP32[$12>>2]|0; //@line 192 "load_dmf.cpp"
          $128 = Math_imul($127, $row$086)|0; //@line 192 "load_dmf.cpp"
          $129 = ($glbinfobyte$083|0)==(0); //@line 194 "load_dmf.cpp"
          L60: do {
           if ($129) {
            $130 = (($d$088) + 1)|0; //@line 196 "load_dmf.cpp"
            $131 = (($lpStream) + ($d$088)|0); //@line 196 "load_dmf.cpp"
            $132 = HEAP8[$131>>0]|0; //@line 196 "load_dmf.cpp"
            $133 = $132&255; //@line 198 "load_dmf.cpp"
            $134 = $133 & 128; //@line 198 "load_dmf.cpp"
            $135 = ($134|0)!=(0); //@line 198 "load_dmf.cpp"
            $136 = ($130>>>0)<($117>>>0); //@line 198 "load_dmf.cpp"
            $or$cond43 = $136 & $135; //@line 198 "load_dmf.cpp"
            if ($or$cond43) {
             $137 = (($d$088) + 2)|0; //@line 198 "load_dmf.cpp"
             $138 = (($lpStream) + ($130)|0); //@line 198 "load_dmf.cpp"
             $139 = HEAP8[$138>>0]|0; //@line 198 "load_dmf.cpp"
             $140 = $139&255; //@line 198 "load_dmf.cpp"
             $d$1 = $137;$glbinfobyte$1 = $140;
            } else {
             $d$1 = $130;$glbinfobyte$1 = 0;
            }
            $141 = $133 & 127; //@line 199 "load_dmf.cpp"
            $142 = ($141|0)!=(0); //@line 200 "load_dmf.cpp"
            $143 = ($d$1>>>0)<($117>>>0); //@line 200 "load_dmf.cpp"
            $or$cond44 = $142 & $143; //@line 200 "load_dmf.cpp"
            if ($or$cond44) {
             $144 = (($d$1) + 1)|0; //@line 200 "load_dmf.cpp"
             $145 = (($lpStream) + ($d$1)|0); //@line 200 "load_dmf.cpp"
             $146 = HEAP8[$145>>0]|0; //@line 200 "load_dmf.cpp"
             $d$2 = $144;$infoval$0 = $146;
            } else {
             $d$2 = $d$1;$infoval$0 = 0;
            }
            switch ($141|0) {
            case 1:  {
             $147 = $infoval$0&255; //@line 203 "load_dmf.cpp"
             $d$3 = $d$2;$glbinfobyte$2 = $glbinfobyte$1;$pbeat$1 = $pbeat$084;$tempo$1 = $147;$tempochange$1$off0 = 1;$ttype$1 = 0;
             break L60;
             break;
            }
            case 2:  {
             $148 = $infoval$0&255; //@line 204 "load_dmf.cpp"
             $d$3 = $d$2;$glbinfobyte$2 = $glbinfobyte$1;$pbeat$1 = $pbeat$084;$tempo$1 = $148;$tempochange$1$off0 = 1;$ttype$1 = 1;
             break L60;
             break;
            }
            case 3:  {
             $149 = $infoval$0&255; //@line 205 "load_dmf.cpp"
             $150 = $149 >>> 4; //@line 205 "load_dmf.cpp"
             $151 = ($ttype$087|0)!=(0); //@line 205 "load_dmf.cpp"
             $d$3 = $d$2;$glbinfobyte$2 = $glbinfobyte$1;$pbeat$1 = $150;$tempo$1 = $tempo$082;$tempochange$1$off0 = $151;$ttype$1 = $ttype$087;
             break L60;
             break;
            }
            default: {
             $d$3 = $d$2;$glbinfobyte$2 = $glbinfobyte$1;$pbeat$1 = $pbeat$084;$tempo$1 = $tempo$082;$tempochange$1$off0 = $tempochange$0$off085;$ttype$1 = $ttype$087;
             break L60;
            }
            }
           } else {
            $152 = (($glbinfobyte$083) + -1)|0; //@line 212 "load_dmf.cpp"
            $d$3 = $d$088;$glbinfobyte$2 = $152;$pbeat$1 = $pbeat$084;$tempo$1 = $tempo$082;$tempochange$1$off0 = $tempochange$0$off085;$ttype$1 = $ttype$087;
           }
          } while(0);
          L74: do {
           if ($126) {
            $d$13 = $d$3;
           } else {
            $d$468 = $d$3;$i2$066 = 0;
            while(1) {
             $153 = (($infobyte) + ($i2$066)|0); //@line 215 "load_dmf.cpp"
             $154 = HEAP8[$153>>0]|0; //@line 215 "load_dmf.cpp"
             $155 = ($154<<24>>24)==(0); //@line 215 "load_dmf.cpp"
             if ($155) {
              $156 = (($d$468) + 1)|0; //@line 218 "load_dmf.cpp"
              $157 = (($lpStream) + ($d$468)|0); //@line 218 "load_dmf.cpp"
              $158 = HEAP8[$157>>0]|0; //@line 218 "load_dmf.cpp"
              $159 = $158&255; //@line 219 "load_dmf.cpp"
              $160 = $159 & 128; //@line 219 "load_dmf.cpp"
              $161 = ($160|0)==(0); //@line 219 "load_dmf.cpp"
              if ($161) {
               $d$5 = $156;
              } else {
               $162 = (($d$468) + 2)|0; //@line 219 "load_dmf.cpp"
               $163 = (($lpStream) + ($156)|0); //@line 219 "load_dmf.cpp"
               $164 = HEAP8[$163>>0]|0; //@line 219 "load_dmf.cpp"
               HEAP8[$153>>0] = $164; //@line 219 "load_dmf.cpp"
               $d$5 = $162;
              }
              $165 = $159 & 64; //@line 221 "load_dmf.cpp"
              $166 = ($165|0)==(0); //@line 221 "load_dmf.cpp"
              if ($166) {
               $cmd$sroa$15$4 = 0;$d$6 = $d$5;
              } else {
               $167 = (($d$5) + 1)|0; //@line 223 "load_dmf.cpp"
               $168 = (($lpStream) + ($d$5)|0); //@line 223 "load_dmf.cpp"
               $169 = HEAP8[$168>>0]|0; //@line 223 "load_dmf.cpp"
               $cmd$sroa$15$4 = $169;$d$6 = $167;
              }
              $170 = $159 & 32; //@line 226 "load_dmf.cpp"
              $171 = ($170|0)==(0); //@line 226 "load_dmf.cpp"
              do {
               if ($171) {
                $cmd$sroa$0$5 = 0;$d$7 = $d$6;
               } else {
                $172 = (($d$6) + 1)|0; //@line 228 "load_dmf.cpp"
                $173 = (($lpStream) + ($d$6)|0); //@line 228 "load_dmf.cpp"
                $174 = HEAP8[$173>>0]|0; //@line 228 "load_dmf.cpp"
                $switch$tableidx = (($174) + 2)<<24>>24; //@line 229 "load_dmf.cpp"
                $175 = ($switch$tableidx&255)<(3); //@line 229 "load_dmf.cpp"
                if ($175) {
                 $cmd$sroa$0$5 = $174;$d$7 = $172;
                 break;
                }
                $176 = $174&255; //@line 229 "load_dmf.cpp"
                $177 = $176 & 127; //@line 229 "load_dmf.cpp"
                $178 = ($177|0)==(0); //@line 230 "load_dmf.cpp"
                if ($178) {
                 $cmd$sroa$0$5 = 0;$d$7 = $172;
                 break;
                }
                $179 = (($177) + 24)|0; //@line 230 "load_dmf.cpp"
                $180 = $179&255; //@line 230 "load_dmf.cpp"
                $cmd$sroa$0$5 = $180;$d$7 = $172;
               }
              } while(0);
              $181 = $159 & 16; //@line 233 "load_dmf.cpp"
              $182 = ($181|0)==(0); //@line 233 "load_dmf.cpp"
              if ($182) {
               $cmd$sroa$16$4 = 0;$cmd$sroa$35$4 = 0;$d$8 = $d$7;
              } else {
               $183 = (($d$7) + 1)|0; //@line 236 "load_dmf.cpp"
               $184 = (($lpStream) + ($d$7)|0); //@line 236 "load_dmf.cpp"
               $185 = HEAP8[$184>>0]|0; //@line 236 "load_dmf.cpp"
               $186 = $185&255; //@line 236 "load_dmf.cpp"
               $187 = (($186) + 3)|0; //@line 236 "load_dmf.cpp"
               $188 = $187 >>> 2; //@line 236 "load_dmf.cpp"
               $189 = $188&255; //@line 236 "load_dmf.cpp"
               $cmd$sroa$16$4 = 1;$cmd$sroa$35$4 = $189;$d$8 = $183;
              }
              $190 = $159 & 8; //@line 239 "load_dmf.cpp"
              $191 = ($190|0)==(0); //@line 239 "load_dmf.cpp"
              L93: do {
               if ($191) {
                $cmd$sroa$0$6 = $cmd$sroa$0$5;$cmd$sroa$19$4 = 0;$cmd$sroa$37$4 = 0;$d$9 = $d$8;
               } else {
                $192 = (($d$8) + 1)|0; //@line 241 "load_dmf.cpp"
                $193 = (($lpStream) + ($d$8)|0); //@line 241 "load_dmf.cpp"
                $194 = HEAP8[$193>>0]|0; //@line 241 "load_dmf.cpp"
                $195 = (($d$8) + 2)|0; //@line 242 "load_dmf.cpp"
                $196 = (($lpStream) + ($192)|0); //@line 242 "load_dmf.cpp"
                $197 = HEAP8[$196>>0]|0; //@line 242 "load_dmf.cpp"
                $198 = $194&255; //@line 243 "load_dmf.cpp"
                switch ($198|0) {
                case 1:  {
                 $199 = ($cmd$sroa$0$5<<24>>24)==(0); //@line 246 "load_dmf.cpp"
                 $$cmd$sroa$0$5 = $199 ? -2 : $cmd$sroa$0$5; //@line 246 "load_dmf.cpp"
                 $cmd$sroa$0$6 = $$cmd$sroa$0$5;$cmd$sroa$19$4 = 0;$cmd$sroa$37$4 = 0;$d$9 = $195;
                 break L93;
                 break;
                }
                case 4:  {
                 $200 = $197&255; //@line 249 "load_dmf.cpp"
                 $201 = $200 & 224; //@line 249 "load_dmf.cpp"
                 $202 = ($201|0)==(0); //@line 249 "load_dmf.cpp"
                 if ($202) {
                  $cmd$sroa$0$6 = $cmd$sroa$0$5;$cmd$sroa$19$4 = 0;$cmd$sroa$37$4 = 0;$d$9 = $195;
                  break L93;
                 }
                 $203 = $200 >>> 5; //@line 249 "load_dmf.cpp"
                 $204 = $203 | 208; //@line 249 "load_dmf.cpp"
                 $205 = $204&255; //@line 249 "load_dmf.cpp"
                 $cmd$sroa$0$6 = $cmd$sroa$0$5;$cmd$sroa$19$4 = 20;$cmd$sroa$37$4 = $205;$d$9 = $195;
                 break L93;
                 break;
                }
                case 5:  {
                 $206 = ($197&255)<(32); //@line 251 "load_dmf.cpp"
                 $207 = ($197&255) >>> 5; //@line 251 "load_dmf.cpp"
                 $$47 = $206 ? 0 : 15; //@line 251 "load_dmf.cpp"
                 $$48 = $206 ? 0 : $207; //@line 251 "load_dmf.cpp"
                 $cmd$sroa$0$6 = $cmd$sroa$0$5;$cmd$sroa$19$4 = $$47;$cmd$sroa$37$4 = $$48;$d$9 = $195;
                 break L93;
                 break;
                }
                case 6:  {
                 $cmd$sroa$0$6 = $cmd$sroa$0$5;$cmd$sroa$19$4 = 10;$cmd$sroa$37$4 = $197;$d$9 = $195;
                 break L93;
                 break;
                }
                default: {
                 $cmd$sroa$0$6 = $cmd$sroa$0$5;$cmd$sroa$19$4 = 0;$cmd$sroa$37$4 = 0;$d$9 = $195;
                 break L93;
                }
                }
               }
              } while(0);
              $208 = $159 & 4; //@line 260 "load_dmf.cpp"
              $209 = ($208|0)==(0); //@line 260 "load_dmf.cpp"
              L102: do {
               if ($209) {
                $cmd$sroa$0$7 = $cmd$sroa$0$6;$cmd$sroa$19$5 = $cmd$sroa$19$4;$cmd$sroa$37$5 = $cmd$sroa$37$4;$d$10 = $d$9;
               } else {
                $210 = (($d$9) + 1)|0; //@line 262 "load_dmf.cpp"
                $211 = (($lpStream) + ($d$9)|0); //@line 262 "load_dmf.cpp"
                $212 = HEAP8[$211>>0]|0; //@line 262 "load_dmf.cpp"
                $213 = (($d$9) + 2)|0; //@line 263 "load_dmf.cpp"
                $214 = (($lpStream) + ($210)|0); //@line 263 "load_dmf.cpp"
                $215 = HEAP8[$214>>0]|0; //@line 263 "load_dmf.cpp"
                $216 = $212&255; //@line 264 "load_dmf.cpp"
                switch ($216|0) {
                case 1:  {
                 $217 = $215&255; //@line 267 "load_dmf.cpp"
                 $218 = $217 & 240; //@line 267 "load_dmf.cpp"
                 $219 = ($218|0)==(0); //@line 267 "load_dmf.cpp"
                 if ($219) {
                  $cmd$sroa$0$7 = $cmd$sroa$0$6;$cmd$sroa$19$5 = $cmd$sroa$19$4;$cmd$sroa$37$5 = $cmd$sroa$37$4;$d$10 = $213;
                  break L102;
                 }
                 $220 = $217 >>> 4; //@line 267 "load_dmf.cpp"
                 $221 = $220 | 32; //@line 267 "load_dmf.cpp"
                 $222 = $221&255; //@line 267 "load_dmf.cpp"
                 $cmd$sroa$0$7 = $cmd$sroa$0$6;$cmd$sroa$19$5 = 20;$cmd$sroa$37$5 = $222;$d$10 = $213;
                 break L102;
                 break;
                }
                case 2:  {
                 $223 = $215&255; //@line 269 "load_dmf.cpp"
                 $224 = $223 & 224; //@line 269 "load_dmf.cpp"
                 $225 = ($224|0)==(0); //@line 269 "load_dmf.cpp"
                 if ($225) {
                  $cmd$sroa$0$7 = $cmd$sroa$0$6;$cmd$sroa$19$5 = $cmd$sroa$19$4;$cmd$sroa$37$5 = $cmd$sroa$37$4;$d$10 = $213;
                  break L102;
                 }
                 $226 = $223 >>> 5; //@line 269 "load_dmf.cpp"
                 $227 = $226 | 208; //@line 269 "load_dmf.cpp"
                 $228 = $227&255; //@line 269 "load_dmf.cpp"
                 $cmd$sroa$0$7 = $cmd$sroa$0$6;$cmd$sroa$19$5 = 20;$cmd$sroa$37$5 = $228;$d$10 = $213;
                 break L102;
                 break;
                }
                case 3:  {
                 $229 = ($215<<24>>24)==(0); //@line 271 "load_dmf.cpp"
                 $cmd$sroa$19$4$ = $229 ? $cmd$sroa$19$4 : 1; //@line 271 "load_dmf.cpp"
                 $cmd$sroa$37$4$ = $229 ? $cmd$sroa$37$4 : $215; //@line 271 "load_dmf.cpp"
                 $cmd$sroa$0$7 = $cmd$sroa$0$6;$cmd$sroa$19$5 = $cmd$sroa$19$4$;$cmd$sroa$37$5 = $cmd$sroa$37$4$;$d$10 = $213;
                 break L102;
                 break;
                }
                case 4:  {
                 $230 = ($215&255)>(223); //@line 273 "load_dmf.cpp"
                 $231 = $230 ? -33 : $215; //@line 273 "load_dmf.cpp"
                 $cmd$sroa$0$7 = $cmd$sroa$0$6;$cmd$sroa$19$5 = 2;$cmd$sroa$37$5 = $231;$d$10 = $213;
                 break L102;
                 break;
                }
                case 5:  {
                 $232 = ($215&255)>(223); //@line 275 "load_dmf.cpp"
                 $233 = $232 ? -33 : $215; //@line 275 "load_dmf.cpp"
                 $cmd$sroa$0$7 = $cmd$sroa$0$6;$cmd$sroa$19$5 = 3;$cmd$sroa$37$5 = $233;$d$10 = $213;
                 break L102;
                 break;
                }
                case 6:  {
                 $cmd$sroa$0$7 = $cmd$sroa$0$6;$cmd$sroa$19$5 = 4;$cmd$sroa$37$5 = $215;$d$10 = $213;
                 break L102;
                 break;
                }
                case 8:  {
                 $cmd$sroa$0$7 = $cmd$sroa$0$6;$cmd$sroa$19$5 = 5;$cmd$sroa$37$5 = $215;$d$10 = $213;
                 break L102;
                 break;
                }
                case 12:  {
                 $234 = $215&255; //@line 281 "load_dmf.cpp"
                 $235 = $234 & 224; //@line 281 "load_dmf.cpp"
                 $236 = ($235|0)==(0); //@line 281 "load_dmf.cpp"
                 if ($236) {
                  $240 = ($cmd$sroa$0$6<<24>>24)==(0); //@line 282 "load_dmf.cpp"
                  $$cmd$sroa$0$6 = $240 ? -2 : $cmd$sroa$0$6; //@line 282 "load_dmf.cpp"
                  $cmd$sroa$0$7 = $$cmd$sroa$0$6;$cmd$sroa$19$5 = $cmd$sroa$19$4;$cmd$sroa$37$5 = $cmd$sroa$37$4;$d$10 = $213;
                  break L102;
                 } else {
                  $237 = $234 >>> 5; //@line 281 "load_dmf.cpp"
                  $238 = $237 | 192; //@line 281 "load_dmf.cpp"
                  $239 = $238&255; //@line 281 "load_dmf.cpp"
                  $cmd$sroa$0$7 = $cmd$sroa$0$6;$cmd$sroa$19$5 = 20;$cmd$sroa$37$5 = $239;$d$10 = $213;
                  break L102;
                 }
                 break;
                }
                default: {
                 $cmd$sroa$0$7 = $cmd$sroa$0$6;$cmd$sroa$19$5 = $cmd$sroa$19$4;$cmd$sroa$37$5 = $cmd$sroa$37$4;$d$10 = $213;
                 break L102;
                }
                }
               }
              } while(0);
              $241 = $159 & 2; //@line 289 "load_dmf.cpp"
              $242 = ($241|0)==(0); //@line 289 "load_dmf.cpp"
              L119: do {
               if ($242) {
                $cmd$sroa$16$5 = $cmd$sroa$16$4;$cmd$sroa$19$6 = $cmd$sroa$19$5;$cmd$sroa$35$5 = $cmd$sroa$35$4;$cmd$sroa$37$6 = $cmd$sroa$37$5;$d$11 = $d$10;
               } else {
                $243 = (($d$10) + 1)|0; //@line 291 "load_dmf.cpp"
                $244 = (($lpStream) + ($d$10)|0); //@line 291 "load_dmf.cpp"
                $245 = HEAP8[$244>>0]|0; //@line 291 "load_dmf.cpp"
                $246 = (($d$10) + 2)|0; //@line 292 "load_dmf.cpp"
                $247 = (($lpStream) + ($243)|0); //@line 292 "load_dmf.cpp"
                $248 = HEAP8[$247>>0]|0; //@line 292 "load_dmf.cpp"
                $249 = $245&255; //@line 293 "load_dmf.cpp"
                switch ($249|0) {
                case 1:  {
                 $250 = ($248<<24>>24)==(-1); //@line 296 "load_dmf.cpp"
                 if ($250) {
                  $cmd$sroa$16$5 = $cmd$sroa$16$4;$cmd$sroa$19$6 = $cmd$sroa$19$5;$cmd$sroa$35$5 = $cmd$sroa$35$4;$cmd$sroa$37$6 = $cmd$sroa$37$5;$d$11 = $246;
                  break L119;
                 }
                 $251 = $248&255; //@line 296 "load_dmf.cpp"
                 $252 = (($251) + 3)|0; //@line 297 "load_dmf.cpp"
                 $253 = $252 >>> 2; //@line 297 "load_dmf.cpp"
                 $254 = ($252>>>0)>(63); //@line 297 "load_dmf.cpp"
                 $$op36 = $253 << 4; //@line 298 "load_dmf.cpp"
                 $255 = $$op36&255; //@line 298 "load_dmf.cpp"
                 $256 = $254 ? -16 : $255; //@line 298 "load_dmf.cpp"
                 $cmd$sroa$16$5 = $cmd$sroa$16$4;$cmd$sroa$19$6 = 11;$cmd$sroa$35$5 = $cmd$sroa$35$4;$cmd$sroa$37$6 = $256;$d$11 = $246;
                 break L119;
                 break;
                }
                case 2:  {
                 $257 = ($248<<24>>24)==(-1); //@line 300 "load_dmf.cpp"
                 if ($257) {
                  $cmd$sroa$16$5 = $cmd$sroa$16$4;$cmd$sroa$19$6 = $cmd$sroa$19$5;$cmd$sroa$35$5 = $cmd$sroa$35$4;$cmd$sroa$37$6 = $cmd$sroa$37$5;$d$11 = $246;
                  break L119;
                 }
                 $258 = $248&255; //@line 300 "load_dmf.cpp"
                 $259 = (($258) + 3)|0; //@line 301 "load_dmf.cpp"
                 $260 = $259 >>> 2; //@line 301 "load_dmf.cpp"
                 $261 = $260&255; //@line 301 "load_dmf.cpp"
                 $262 = ($259>>>0)>(63); //@line 301 "load_dmf.cpp"
                 $$5 = $262 ? 15 : $261; //@line 301 "load_dmf.cpp"
                 $cmd$sroa$16$5 = $cmd$sroa$16$4;$cmd$sroa$19$6 = 11;$cmd$sroa$35$5 = $cmd$sroa$35$4;$cmd$sroa$37$6 = $$5;$d$11 = $246;
                 break L119;
                 break;
                }
                case 7:  {
                 $263 = ($cmd$sroa$16$4<<24>>24)==(0); //@line 304 "load_dmf.cpp"
                 if (!($263)) {
                  $cmd$sroa$16$5 = $cmd$sroa$16$4;$cmd$sroa$19$6 = 9;$cmd$sroa$35$5 = $cmd$sroa$35$4;$cmd$sroa$37$6 = $248;$d$11 = $246;
                  break L119;
                 }
                 $264 = $248&255; //@line 304 "load_dmf.cpp"
                 $265 = (($264) + 3)|0; //@line 304 "load_dmf.cpp"
                 $266 = $265 >>> 2; //@line 304 "load_dmf.cpp"
                 $267 = $266&255; //@line 304 "load_dmf.cpp"
                 $cmd$sroa$16$5 = 2;$cmd$sroa$19$6 = $cmd$sroa$19$5;$cmd$sroa$35$5 = $267;$cmd$sroa$37$6 = $cmd$sroa$37$5;$d$11 = $246;
                 break L119;
                 break;
                }
                case 8:  {
                 $268 = $248&255; //@line 307 "load_dmf.cpp"
                 $269 = (($268) + 3)|0; //@line 307 "load_dmf.cpp"
                 $270 = $269 >>> 2; //@line 307 "load_dmf.cpp"
                 $271 = ($269>>>0)>(63); //@line 307 "load_dmf.cpp"
                 $$op = $270 << 4; //@line 308 "load_dmf.cpp"
                 $272 = $$op&255; //@line 308 "load_dmf.cpp"
                 $273 = $271 ? -16 : $272; //@line 308 "load_dmf.cpp"
                 $cmd$sroa$16$5 = $cmd$sroa$16$4;$cmd$sroa$19$6 = 29;$cmd$sroa$35$5 = $cmd$sroa$35$4;$cmd$sroa$37$6 = $273;$d$11 = $246;
                 break L119;
                 break;
                }
                case 9:  {
                 $274 = $248&255; //@line 310 "load_dmf.cpp"
                 $275 = (($274) + 3)|0; //@line 310 "load_dmf.cpp"
                 $276 = $275 >>> 2; //@line 310 "load_dmf.cpp"
                 $277 = $276&255; //@line 310 "load_dmf.cpp"
                 $278 = ($275>>>0)>(63); //@line 310 "load_dmf.cpp"
                 $$7 = $278 ? 15 : $277; //@line 310 "load_dmf.cpp"
                 $cmd$sroa$16$5 = $cmd$sroa$16$4;$cmd$sroa$19$6 = 29;$cmd$sroa$35$5 = $cmd$sroa$35$4;$cmd$sroa$37$6 = $$7;$d$11 = $246;
                 break L119;
                 break;
                }
                default: {
                 $cmd$sroa$16$5 = $cmd$sroa$16$4;$cmd$sroa$19$6 = $cmd$sroa$19$5;$cmd$sroa$35$5 = $cmd$sroa$35$4;$cmd$sroa$37$6 = $cmd$sroa$37$5;$d$11 = $246;
                 break L119;
                }
                }
               }
              } while(0);
              $279 = HEAP32[$12>>2]|0; //@line 319 "load_dmf.cpp"
              $280 = ($i2$066>>>0)<($279>>>0); //@line 319 "load_dmf.cpp"
              if ($280) {
               $$sum35 = (($i2$066) + ($128))|0; //@line 319 "load_dmf.cpp"
               $281 = (($113) + (($$sum35*6)|0)|0); //@line 319 "load_dmf.cpp"
               HEAP8[$281>>0] = $cmd$sroa$0$7; //@line 319 "load_dmf.cpp"
               $282 = (((($113) + (($$sum35*6)|0)|0)) + 1|0); //@line 319 "load_dmf.cpp"
               HEAP8[$282>>0] = $cmd$sroa$15$4; //@line 319 "load_dmf.cpp"
               $283 = (((($113) + (($$sum35*6)|0)|0)) + 2|0); //@line 319 "load_dmf.cpp"
               HEAP8[$283>>0] = $cmd$sroa$16$5; //@line 319 "load_dmf.cpp"
               $284 = (((($113) + (($$sum35*6)|0)|0)) + 3|0); //@line 319 "load_dmf.cpp"
               HEAP8[$284>>0] = $cmd$sroa$19$6; //@line 319 "load_dmf.cpp"
               $285 = (((($113) + (($$sum35*6)|0)|0)) + 4|0); //@line 319 "load_dmf.cpp"
               HEAP8[$285>>0] = $cmd$sroa$35$5; //@line 319 "load_dmf.cpp"
               $286 = (((($113) + (($$sum35*6)|0)|0)) + 5|0); //@line 319 "load_dmf.cpp"
               HEAP8[$286>>0] = $cmd$sroa$37$6; //@line 319 "load_dmf.cpp"
              }
              $287 = ($d$11>>>0)>($117>>>0); //@line 320 "load_dmf.cpp"
              if ($287) {
               $d$13 = $d$11;
               break L74;
              } else {
               $d$12 = $d$11;
              }
             } else {
              $288 = (($154) + -1)<<24>>24; //@line 329 "load_dmf.cpp"
              HEAP8[$153>>0] = $288; //@line 329 "load_dmf.cpp"
              $d$12 = $d$468;
             }
             $289 = (($i2$066) + 1)|0; //@line 215 "load_dmf.cpp"
             $290 = ($289>>>0)<($$3>>>0); //@line 215 "load_dmf.cpp"
             if ($290) {
              $d$468 = $d$12;$i2$066 = $289;
             } else {
              $d$13 = $d$12;
              break;
             }
            }
           }
          } while(0);
          L137: do {
           if ($tempochange$1$off0) {
            $291 = ($ttype$1|0)!=(0); //@line 337 "load_dmf.cpp"
            $292 = ($pbeat$1|0)!=(0); //@line 337 "load_dmf.cpp"
            $or$cond9 = $292 & $291; //@line 337 "load_dmf.cpp"
            $293 = Math_imul($pbeat$1, $tempo$1)|0; //@line 337 "load_dmf.cpp"
            $294 = ($tempo$1*15)|0; //@line 337 "load_dmf.cpp"
            $295 = (($294) + 15)|0; //@line 337 "load_dmf.cpp"
            $296 = $or$cond9 ? $293 : $295; //@line 337 "load_dmf.cpp"
            $speed$075 = 30;
            while(1) {
             $300 = Math_imul($speed$075, $296)|0; //@line 340 "load_dmf.cpp"
             $301 = ($300>>>0)<(4824); //@line 341 "load_dmf.cpp"
             if ($301) {
              $$lcssa147 = $300;$speed$0$lcssa = $speed$075;
              break;
             }
             $302 = ($speed$075>>>0)<(6); //@line 342 "load_dmf.cpp"
             $303 = ($300>>>0)<(6144); //@line 342 "load_dmf.cpp"
             $or$cond11 = $302 & $303; //@line 342 "load_dmf.cpp"
             if ($or$cond11) {
              $$lcssa147 = $300;$speed$0$lcssa = $speed$075;
              break;
             }
             $304 = (($speed$075) + -1)|0; //@line 338 "load_dmf.cpp"
             $305 = ($304>>>0)>(1); //@line 338 "load_dmf.cpp"
             if ($305) {
              $speed$075 = $304;
             } else {
              $$lcssa147 = $300;$speed$0$lcssa = $304;
              break;
             }
            }
            $297 = HEAP32[$12>>2]|0; //@line 348 "load_dmf.cpp"
            $298 = ($297|0)==(0); //@line 348 "load_dmf.cpp"
            if ($298) {
             break;
            }
            $299 = (($$lcssa147>>>0) / 24)&-1; //@line 340 "load_dmf.cpp"
            $ich$079 = 0;$modtempo$278 = $299;$speed$177 = $speed$0$lcssa;
            while(1) {
             $$sum34 = (($ich$079) + ($128))|0; //@line 348 "load_dmf.cpp"
             $306 = (((($113) + (($$sum34*6)|0)|0)) + 3|0); //@line 348 "load_dmf.cpp"
             $307 = HEAP8[$306>>0]|0; //@line 348 "load_dmf.cpp"
             $308 = ($307<<24>>24)==(0); //@line 348 "load_dmf.cpp"
             do {
              if ($308) {
               $309 = ($speed$177|0)==(0); //@line 350 "load_dmf.cpp"
               if (!($309)) {
                HEAP8[$306>>0] = 16; //@line 352 "load_dmf.cpp"
                $310 = $speed$177&255; //@line 353 "load_dmf.cpp"
                $311 = (((($113) + (($$sum34*6)|0)|0)) + 5|0); //@line 353 "load_dmf.cpp"
                HEAP8[$311>>0] = $310; //@line 353 "load_dmf.cpp"
                $modtempo$3 = $modtempo$278;$speed$2 = 0;
                break;
               }
               $modtempo$2$off = (($modtempo$278) + -32)|0; //@line 356 "load_dmf.cpp"
               $312 = ($modtempo$2$off>>>0)<(224); //@line 356 "load_dmf.cpp"
               if (!($312)) {
                break L137;
               }
               HEAP8[$306>>0] = 17; //@line 358 "load_dmf.cpp"
               $313 = $modtempo$278&255; //@line 359 "load_dmf.cpp"
               $314 = (((($113) + (($$sum34*6)|0)|0)) + 5|0); //@line 359 "load_dmf.cpp"
               HEAP8[$314>>0] = $313; //@line 359 "load_dmf.cpp"
               $modtempo$3 = 0;$speed$2 = 0;
              } else {
               $modtempo$3 = $modtempo$278;$speed$2 = $speed$177;
              }
             } while(0);
             $315 = (($ich$079) + 1)|0; //@line 348 "load_dmf.cpp"
             $316 = ($315>>>0)<($297>>>0); //@line 348 "load_dmf.cpp"
             if ($316) {
              $ich$079 = $315;$modtempo$278 = $modtempo$3;$speed$177 = $speed$2;
             } else {
              break;
             }
            }
           }
          } while(0);
          $317 = ($d$13>>>0)<($117>>>0); //@line 367 "load_dmf.cpp"
          $318 = (($row$086) + 1)|0; //@line 190 "load_dmf.cpp"
          $319 = ($318>>>0)<($$$14>>>0); //@line 190 "load_dmf.cpp"
          $or$cond129 = $317 & $319; //@line 367 "load_dmf.cpp"
          if ($or$cond129) {
           $d$088 = $d$13;$glbinfobyte$083 = $glbinfobyte$2;$pbeat$084 = $pbeat$1;$row$086 = $318;$tempo$082 = $tempo$1;$tempochange$0$off085 = 0;$ttype$087 = $ttype$1;
          } else {
           break;
          }
         }
        }
        $320 = (($117) + 8)|0; //@line 372 "load_dmf.cpp"
        $321 = ($320>>>0)<($dwMemLength>>>0); //@line 372 "load_dmf.cpp"
        $322 = (($npat$093) + 1)|0; //@line 165 "load_dmf.cpp"
        $323 = ($322>>>0)<($$2>>>0); //@line 165 "load_dmf.cpp"
        $or$cond135 = $321 & $323; //@line 372 "load_dmf.cpp"
        if ($or$cond135) {
         $dwPos$094 = $117;$npat$093 = $322;
        } else {
         break;
        }
       }
      }
     } while(0);
     $324 = HEAPU8[$74>>0]|(HEAPU8[$74+1>>0]<<8)|(HEAPU8[$74+2>>0]<<16)|(HEAPU8[$74+3>>0]<<24); //@line 374 "load_dmf.cpp"
     $325 = (($77) + ($324))|0; //@line 374 "load_dmf.cpp"
     $dwMemPos$0$ph$be = $325;
    }
    else if ((label|0) == 146) {
     label = 0;
     $switch$split12D = ($19|0)<(1162104389);
     if (!($switch$split12D)) {
      $switch$split72D = ($19|0)<(1196641603);
      if ($switch$split72D) {
       switch ($19|0) {
       case 1162104389:  {
        break L12;
        break;
       }
       default: {
        label = 139;
        break L20;
       }
       }
      }
      switch ($19|0) {
      case 1196641603:  {
       break;
      }
      default: {
       label = 139;
       break L20;
      }
      }
      $20 = (($dwMemPos$0$ph111) + 1)|0; //@line 117 "load_dmf.cpp"
      $dwMemPos$1 = $20;
      label = 9;
      break;
     }
     switch ($19|0) {
     case 1146113363:  {
      break;
     }
     default: {
      label = 139;
      break L20;
     }
     }
     $404 = (($dwMemPos$0$ph111) + 8)|0; //@line 424 "load_dmf.cpp"
     $405 = HEAP32[$15>>2]|0; //@line 426 "load_dmf.cpp"
     $406 = ($405|0)==(0); //@line 426 "load_dmf.cpp"
     if ($406) {
      $dwMemPos$0$ph$be = $404;
     } else {
      $441 = $405;$dwPos9$0103 = $404;$iSmp10$0105 = 1;$ismpd$0104 = 0;
      while(1) {
       $407 = (($ismpd$0104) + 1)|0; //@line 428 "load_dmf.cpp"
       $408 = (($dwPos9$0103) + 4)|0; //@line 430 "load_dmf.cpp"
       $409 = ($408>>>0)<($dwMemLength>>>0); //@line 430 "load_dmf.cpp"
       if (!($409)) {
        $dwMemPos$0$ph$be = $dwPos9$0103;
        break L20;
       }
       $410 = (($lpStream) + ($dwPos9$0103)|0); //@line 437 "load_dmf.cpp"
       $411 = HEAP32[$410>>2]|0; //@line 437 "load_dmf.cpp"
       $412 = (($dwMemLength) - ($408))|0; //@line 443 "load_dmf.cpp"
       $413 = ($411>>>0)>($412>>>0); //@line 443 "load_dmf.cpp"
       $$50 = $413 ? $412 : $411; //@line 443 "load_dmf.cpp"
       $414 = ($$50|0)==(0); //@line 450 "load_dmf.cpp"
       if ($414) {
        $432 = $441;
       } else {
        $415 = (((($this)) + 33280|0) + ($iSmp10$0105<<6)|0); //@line 452 "load_dmf.cpp"
        $416 = (((((($this)) + 33280|0) + ($iSmp10$0105<<6)|0)) + 34|0); //@line 452 "load_dmf.cpp"
        $417 = HEAP16[$416>>1]|0; //@line 452 "load_dmf.cpp"
        $418 = $417 & 1; //@line 452 "load_dmf.cpp"
        $419 = ($418<<16>>16)!=(0); //@line 452 "load_dmf.cpp"
        $420 = $419 ? 5 : 0; //@line 452 "load_dmf.cpp"
        $421 = (($smplflags) + ($407)|0); //@line 453 "load_dmf.cpp"
        $422 = HEAP8[$421>>0]|0; //@line 453 "load_dmf.cpp"
        $423 = $422 & 4; //@line 453 "load_dmf.cpp"
        $424 = ($423<<24>>24)==(0); //@line 453 "load_dmf.cpp"
        if ($424) {
         $flags$0 = $420;
        } else {
         $425 = $418&65535; //@line 453 "load_dmf.cpp"
         $426 = $425 << 2; //@line 453 "load_dmf.cpp"
         $427 = $426 | 19; //@line 453 "load_dmf.cpp"
         $flags$0 = $427;
        }
        $428 = (($lpStream) + ($408)|0); //@line 454 "load_dmf.cpp"
        (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$415,$flags$0,$428,$$50)|0); //@line 454 "load_dmf.cpp"
        $$pre121 = HEAP32[$15>>2]|0; //@line 426 "load_dmf.cpp"
        $432 = $$pre121;
       }
       $429 = (($$50) + ($408))|0; //@line 456 "load_dmf.cpp"
       $430 = (($iSmp10$0105) + 1)|0; //@line 426 "load_dmf.cpp"
       $431 = ($430>>>0)>($432>>>0); //@line 426 "load_dmf.cpp"
       if ($431) {
        $dwMemPos$0$ph$be = $429;
        break;
       } else {
        $441 = $432;$dwPos9$0103 = $429;$iSmp10$0105 = $430;$ismpd$0104 = $407;
       }
      }
     }
    }
    else if ((label|0) == 150) {
     label = 0;
     $switch$split102D = ($19|0)<(1330007625);
     if (!($switch$split102D)) {
      switch ($19|0) {
      case 1330007625:  {
       $dwMemPos$1 = $dwMemPos$0$ph111;
       label = 9;
       break L20;
       break;
      }
      default: {
       label = 139;
       break L20;
      }
      }
     }
     switch ($19|0) {
     case 1229999443:  {
      break;
     }
     default: {
      label = 139;
      break L20;
     }
     }
     $$sum = (($dwMemPos$0$ph111) + 4)|0; //@line 382 "load_dmf.cpp"
     $326 = (($lpStream) + ($$sum)|0); //@line 382 "load_dmf.cpp"
     $327 = HEAPU8[$326>>0]|(HEAPU8[$326+1>>0]<<8)|(HEAPU8[$326+2>>0]<<16)|(HEAPU8[$326+3>>0]<<24); //@line 382 "load_dmf.cpp"
     $328 = (($dwMemLength) - ($dwMemPos$0$ph111))|0; //@line 382 "load_dmf.cpp"
     $329 = ($327>>>0)>($328>>>0); //@line 382 "load_dmf.cpp"
     $$pre125 = (($dwMemPos$0$ph111) + 8)|0; //@line 417 "load_dmf.cpp"
     L175: do {
      if ($329) {
       $403 = $327;
      } else {
       $330 = (($lpStream) + ($$pre125)|0); //@line 385 "load_dmf.cpp"
       $331 = HEAP8[$330>>0]|0; //@line 385 "load_dmf.cpp"
       $332 = $331&255; //@line 385 "load_dmf.cpp"
       $333 = ($331&255)>(239); //@line 386 "load_dmf.cpp"
       $$49 = $333 ? 239 : $332; //@line 386 "load_dmf.cpp"
       HEAP32[$15>>2] = $$49; //@line 385 "load_dmf.cpp"
       $334 = ($$49|0)==(0); //@line 387 "load_dmf.cpp"
       if ($334) {
        $403 = $327;
       } else {
        $335 = (($dwMemPos$0$ph111) + 9)|0; //@line 384 "load_dmf.cpp"
        $336 = (($lpStream) + ($335)|0); //@line 389 "load_dmf.cpp"
        $337 = HEAP8[$336>>0]|0; //@line 389 "load_dmf.cpp"
        $338 = $337&255; //@line 389 "load_dmf.cpp"
        HEAP8[$18>>0] = 0; //@line 390 "load_dmf.cpp"
        $339 = (($dwMemPos$0$ph111) + 26)|0; //@line 391 "load_dmf.cpp"
        $340 = (($339) + ($338))|0; //@line 391 "load_dmf.cpp"
        $341 = (($$pre125) + ($327))|0; //@line 391 "load_dmf.cpp"
        $342 = ($340>>>0)>($341>>>0); //@line 391 "load_dmf.cpp"
        if ($342) {
         $403 = $327;
        } else {
         $344 = $337;$346 = $338;$386 = $18;$dwPos8$061142 = $335;$iSmp$062141 = 1;
         while(1) {
          $343 = ($344<<24>>24)==(0); //@line 392 "load_dmf.cpp"
          if ($343) {
           $$pre126 = (($dwPos8$061142) + 1)|0; //@line 398 "load_dmf.cpp"
           $$pre$phi127Z2D = $$pre126;
          } else {
           $345 = ($346>>>0)>(31); //@line 394 "load_dmf.cpp"
           $347 = $345 ? 31 : $346; //@line 394 "load_dmf.cpp"
           $348 = (((($this)) + 59804|0) + ($iSmp$062141<<5)|0); //@line 395 "load_dmf.cpp"
           $$sum21 = (($dwPos8$061142) + 1)|0; //@line 395 "load_dmf.cpp"
           $349 = (($lpStream) + ($$sum21)|0); //@line 395 "load_dmf.cpp"
           _memcpy(($348|0),($349|0),($347|0))|0; //@line 395 "load_dmf.cpp"
           $350 = ((((($this)) + 59804|0) + ($iSmp$062141<<5)|0) + ($347)|0); //@line 396 "load_dmf.cpp"
           HEAP8[$350>>0] = 0; //@line 396 "load_dmf.cpp"
           $$pre$phi127Z2D = $$sum21;
          }
          $351 = (($$pre$phi127Z2D) + ($346))|0; //@line 398 "load_dmf.cpp"
          $352 = (($lpStream) + ($351)|0); //@line 399 "load_dmf.cpp"
          $353 = HEAPU8[$352>>0]|(HEAPU8[$352+1>>0]<<8)|(HEAPU8[$352+2>>0]<<16)|(HEAPU8[$352+3>>0]<<24); //@line 401 "load_dmf.cpp"
          $354 = (((($this)) + 33280|0) + ($iSmp$062141<<6)|0); //@line 401 "load_dmf.cpp"
          HEAP32[$354>>2] = $353; //@line 401 "load_dmf.cpp"
          $$sum16 = (($351) + 4)|0; //@line 402 "load_dmf.cpp"
          $355 = (($lpStream) + ($$sum16)|0); //@line 402 "load_dmf.cpp"
          $356 = HEAPU8[$355>>0]|(HEAPU8[$355+1>>0]<<8)|(HEAPU8[$355+2>>0]<<16)|(HEAPU8[$355+3>>0]<<24); //@line 402 "load_dmf.cpp"
          $357 = (((((($this)) + 33280|0) + ($iSmp$062141<<6)|0)) + 4|0); //@line 402 "load_dmf.cpp"
          HEAP32[$357>>2] = $356; //@line 402 "load_dmf.cpp"
          $$sum17 = (($351) + 8)|0; //@line 403 "load_dmf.cpp"
          $358 = (($lpStream) + ($$sum17)|0); //@line 403 "load_dmf.cpp"
          $359 = HEAPU8[$358>>0]|(HEAPU8[$358+1>>0]<<8)|(HEAPU8[$358+2>>0]<<16)|(HEAPU8[$358+3>>0]<<24); //@line 403 "load_dmf.cpp"
          $360 = (((((($this)) + 33280|0) + ($iSmp$062141<<6)|0)) + 8|0); //@line 403 "load_dmf.cpp"
          HEAP32[$360>>2] = $359; //@line 403 "load_dmf.cpp"
          $$sum18 = (($351) + 12)|0; //@line 404 "load_dmf.cpp"
          $361 = (($lpStream) + ($$sum18)|0); //@line 404 "load_dmf.cpp"
          $362 = HEAPU8[$361>>0]|(HEAPU8[$361+1>>0]<<8); //@line 404 "load_dmf.cpp"
          $363 = $362&65535; //@line 404 "load_dmf.cpp"
          $364 = (((((($this)) + 33280|0) + ($iSmp$062141<<6)|0)) + 24|0); //@line 404 "load_dmf.cpp"
          HEAP32[$364>>2] = $363; //@line 404 "load_dmf.cpp"
          $365 = (((((($this)) + 33280|0) + ($iSmp$062141<<6)|0)) + 32|0); //@line 405 "load_dmf.cpp"
          HEAP16[$365>>1] = 64; //@line 405 "load_dmf.cpp"
          $$sum19 = (($351) + 14)|0; //@line 406 "load_dmf.cpp"
          $366 = (($lpStream) + ($$sum19)|0); //@line 406 "load_dmf.cpp"
          $367 = HEAP8[$366>>0]|0; //@line 406 "load_dmf.cpp"
          $368 = ($367<<24>>24)==(0); //@line 406 "load_dmf.cpp"
          $369 = $367&255; //@line 406 "load_dmf.cpp"
          $370 = (($369) + 1)|0; //@line 406 "load_dmf.cpp"
          $371 = $370&65535; //@line 406 "load_dmf.cpp"
          $372 = $368 ? 256 : $371; //@line 406 "load_dmf.cpp"
          $373 = (((((($this)) + 33280|0) + ($iSmp$062141<<6)|0)) + 30|0); //@line 406 "load_dmf.cpp"
          HEAP16[$373>>1] = $372; //@line 406 "load_dmf.cpp"
          $$sum20 = (($351) + 15)|0; //@line 407 "load_dmf.cpp"
          $374 = (($lpStream) + ($$sum20)|0); //@line 407 "load_dmf.cpp"
          $375 = HEAP8[$374>>0]|0; //@line 407 "load_dmf.cpp"
          $376 = ($375&255) >>> 1; //@line 407 "load_dmf.cpp"
          $$lobit = $376 & 1; //@line 407 "load_dmf.cpp"
          $377 = $$lobit&255; //@line 407 "load_dmf.cpp"
          $378 = (((((($this)) + 33280|0) + ($iSmp$062141<<6)|0)) + 34|0); //@line 407 "load_dmf.cpp"
          HEAP16[$378>>1] = $377; //@line 407 "load_dmf.cpp"
          $379 = ($$lobit<<24>>24)==(0); //@line 408 "load_dmf.cpp"
          if (!($379)) {
           $380 = $353 >>> 1; //@line 408 "load_dmf.cpp"
           HEAP32[$354>>2] = $380; //@line 408 "load_dmf.cpp"
          }
          $381 = $375 & 1; //@line 409 "load_dmf.cpp"
          $382 = ($381<<24>>24)==(0); //@line 409 "load_dmf.cpp"
          if (!($382)) {
           $383 = $$lobit&255; //@line 409 "load_dmf.cpp"
           $384 = $383 | 2; //@line 409 "load_dmf.cpp"
           $385 = $384&65535; //@line 409 "load_dmf.cpp"
           HEAP16[$378>>1] = $385; //@line 409 "load_dmf.cpp"
          }
          HEAP8[$386>>0] = $375; //@line 410 "load_dmf.cpp"
          $387 = (($iSmp$062141) + 1)|0; //@line 387 "load_dmf.cpp"
          $388 = HEAP32[$15>>2]|0; //@line 387 "load_dmf.cpp"
          $389 = ($387>>>0)>($388>>>0); //@line 387 "load_dmf.cpp"
          if ($389) {
           break;
          }
          $390 = HEAP8[$4>>0]|0; //@line 411 "load_dmf.cpp"
          $391 = ($390&255)<(8); //@line 411 "load_dmf.cpp"
          $392 = $391 ? 22 : 30; //@line 411 "load_dmf.cpp"
          $393 = (($392) + ($351))|0; //@line 411 "load_dmf.cpp"
          $$pre118 = HEAPU8[$326>>0]|(HEAPU8[$326+1>>0]<<8)|(HEAPU8[$326+2>>0]<<16)|(HEAPU8[$326+3>>0]<<24); //@line 391 "load_dmf.cpp"
          $394 = (($lpStream) + ($393)|0); //@line 389 "load_dmf.cpp"
          $395 = HEAP8[$394>>0]|0; //@line 389 "load_dmf.cpp"
          $396 = $395&255; //@line 389 "load_dmf.cpp"
          $397 = (($smplflags) + ($387)|0); //@line 390 "load_dmf.cpp"
          HEAP8[$397>>0] = 0; //@line 390 "load_dmf.cpp"
          $398 = (($393) + 17)|0; //@line 391 "load_dmf.cpp"
          $399 = (($398) + ($396))|0; //@line 391 "load_dmf.cpp"
          $400 = (($$pre125) + ($$pre118))|0; //@line 391 "load_dmf.cpp"
          $401 = ($399>>>0)>($400>>>0); //@line 391 "load_dmf.cpp"
          if ($401) {
           $403 = $$pre118;
           break L175;
          } else {
           $344 = $395;$346 = $396;$386 = $397;$dwPos8$061142 = $393;$iSmp$062141 = $387;
          }
         }
         $$pre119$pre = HEAPU8[$326>>0]|(HEAPU8[$326+1>>0]<<8)|(HEAPU8[$326+2>>0]<<16)|(HEAPU8[$326+3>>0]<<24); //@line 417 "load_dmf.cpp"
         $403 = $$pre119$pre;
        }
       }
      }
     } while(0);
     $402 = (($$pre125) + ($403))|0; //@line 417 "load_dmf.cpp"
     $dwMemPos$0$ph$be = $402;
    }
    else if ((label|0) == 157) {
     label = 0;
     switch ($19|0) {
     case 1431389523:  {
      break;
     }
     default: {
      label = 139;
      break L20;
     }
     }
     $$sum37 = (($dwMemPos$0$ph111) + 4)|0; //@line 141 "load_dmf.cpp"
     $50 = (($lpStream) + ($$sum37)|0); //@line 141 "load_dmf.cpp"
     $51 = HEAPU8[$50>>0]|(HEAPU8[$50+1>>0]<<8)|(HEAPU8[$50+2>>0]<<16)|(HEAPU8[$50+3>>0]<<24); //@line 141 "load_dmf.cpp"
     $52 = ($51>>>0)<($dwMemLength>>>0); //@line 141 "load_dmf.cpp"
     if (!($52)) {
      break L12;
     }
     $53 = (($dwMemPos$0$ph111) + 12)|0; //@line 141 "load_dmf.cpp"
     $54 = (($53) + ($51))|0; //@line 141 "load_dmf.cpp"
     $55 = ($54>>>0)>($dwMemLength>>>0); //@line 141 "load_dmf.cpp"
     if ($55) {
      break L12;
     }
     $56 = $51 >>> 1; //@line 143 "load_dmf.cpp"
     $57 = ($51>>>0)>(509); //@line 144 "load_dmf.cpp"
     $$ = $57 ? 255 : $56; //@line 144 "load_dmf.cpp"
     $$sum38 = (($dwMemPos$0$ph111) + 8)|0; //@line 145 "load_dmf.cpp"
     $58 = (($lpStream) + ($$sum38)|0); //@line 145 "load_dmf.cpp"
     $59 = HEAPU8[$58>>0]|(HEAPU8[$58+1>>0]<<8); //@line 145 "load_dmf.cpp"
     $60 = $59&65535; //@line 145 "load_dmf.cpp"
     $61 = ($60>>>0)<($$>>>0); //@line 145 "load_dmf.cpp"
     if ($61) {
      HEAP32[$16>>2] = $60; //@line 145 "load_dmf.cpp"
     }
     $62 = ($$|0)==(0); //@line 146 "load_dmf.cpp"
     if ($62) {
      $70 = $51;
     } else {
      $63 = (($lpStream) + ($53)|0); //@line 146 "load_dmf.cpp"
      $i1$0100 = 0;
      while(1) {
       $64 = (($63) + ($i1$0100<<1)|0); //@line 146 "load_dmf.cpp"
       $65 = HEAPU8[$64>>0]|(HEAPU8[$64+1>>0]<<8); //@line 146 "load_dmf.cpp"
       $66 = $65&255; //@line 146 "load_dmf.cpp"
       $67 = (((($this)) + 53344|0) + ($i1$0100)|0); //@line 146 "load_dmf.cpp"
       HEAP8[$67>>0] = $66; //@line 146 "load_dmf.cpp"
       $68 = (($i1$0100) + 1)|0; //@line 146 "load_dmf.cpp"
       $exitcond = ($68|0)==($$|0); //@line 146 "load_dmf.cpp"
       if ($exitcond) {
        break;
       } else {
        $i1$0100 = $68;
       }
      }
      $$pre120 = HEAPU8[$50>>0]|(HEAPU8[$50+1>>0]<<8)|(HEAPU8[$50+2>>0]<<16)|(HEAPU8[$50+3>>0]<<24); //@line 148 "load_dmf.cpp"
      $70 = $$pre120;
     }
     $69 = (($$sum38) + ($70))|0; //@line 148 "load_dmf.cpp"
     $dwMemPos$0$ph$be = $69;
    }
   } while(0);
   if ((label|0) == 9) {
    label = 0;
    $$sum40 = (($dwMemPos$0$ph111) + 4)|0; //@line 118 "load_dmf.cpp"
    $21 = (($lpStream) + ($$sum40)|0); //@line 118 "load_dmf.cpp"
    $22 = HEAPU8[$21>>0]|(HEAPU8[$21+1>>0]<<8)|(HEAPU8[$21+2>>0]<<16)|(HEAPU8[$21+3>>0]<<24); //@line 118 "load_dmf.cpp"
    $23 = ($22>>>0)>($dwMemLength>>>0); //@line 118 "load_dmf.cpp"
    if ($23) {
     break;
    }
    $24 = (($dwMemPos$1) + 8)|0; //@line 118 "load_dmf.cpp"
    $25 = (($24) + ($22))|0; //@line 118 "load_dmf.cpp"
    $26 = ($25>>>0)>($dwMemLength>>>0); //@line 118 "load_dmf.cpp"
    if ($26) {
     break;
    }
    $27 = ($22>>>0)>(7); //@line 119 "load_dmf.cpp"
    if ($27) {
     $28 = HEAP32[$17>>2]|0; //@line 119 "load_dmf.cpp"
     $29 = ($28|0)==(0|0); //@line 119 "load_dmf.cpp"
     if ($29) {
      $30 = (__Znaj($22)|0); //@line 121 "load_dmf.cpp"
      HEAP32[$17>>2] = $30; //@line 121 "load_dmf.cpp"
      $i$060 = 0;
      while(1) {
       $31 = (($i$060>>>0) % 40)&-1; //@line 127 "load_dmf.cpp"
       $32 = ($31|0)==(39); //@line 127 "load_dmf.cpp"
       if ($32) {
        $33 = HEAP32[$17>>2]|0; //@line 128 "load_dmf.cpp"
        $34 = (($33) + ($i$060)|0); //@line 128 "load_dmf.cpp"
        HEAP8[$34>>0] = 13; //@line 128 "load_dmf.cpp"
       } else {
        $35 = (($24) + ($i$060))|0; //@line 126 "load_dmf.cpp"
        $36 = (($lpStream) + ($35)|0); //@line 126 "load_dmf.cpp"
        $37 = HEAP8[$36>>0]|0; //@line 126 "load_dmf.cpp"
        $38 = ($37<<24>>24)<(32); //@line 130 "load_dmf.cpp"
        $39 = $38 ? 32 : $37; //@line 130 "load_dmf.cpp"
        $40 = HEAP32[$17>>2]|0; //@line 130 "load_dmf.cpp"
        $41 = (($40) + ($i$060)|0); //@line 130 "load_dmf.cpp"
        HEAP8[$41>>0] = $39; //@line 130 "load_dmf.cpp"
       }
       $42 = (($i$060) + 1)|0; //@line 124 "load_dmf.cpp"
       $43 = HEAPU8[$21>>0]|(HEAPU8[$21+1>>0]<<8)|(HEAPU8[$21+2>>0]<<16)|(HEAPU8[$21+3>>0]<<24); //@line 124 "load_dmf.cpp"
       $44 = (($43) + -1)|0; //@line 124 "load_dmf.cpp"
       $45 = ($42>>>0)<($44>>>0); //@line 124 "load_dmf.cpp"
       if ($45) {
        $i$060 = $42;
       } else {
        $$lcssa = $44;
        break;
       }
      }
      $$pre = HEAP32[$17>>2]|0; //@line 132 "load_dmf.cpp"
      $46 = (($$pre) + ($$lcssa)|0); //@line 132 "load_dmf.cpp"
      HEAP8[$46>>0] = 0; //@line 132 "load_dmf.cpp"
      $$pre116 = HEAPU8[$21>>0]|(HEAPU8[$21+1>>0]<<8)|(HEAPU8[$21+2>>0]<<16)|(HEAPU8[$21+3>>0]<<24); //@line 135 "load_dmf.cpp"
      $49 = $$pre116;
     } else {
      $49 = $22;
     }
    } else {
     $49 = $22;
    }
    $47 = (($dwMemPos$1) + 7)|0; //@line 135 "load_dmf.cpp"
    $48 = (($47) + ($49))|0; //@line 135 "load_dmf.cpp"
    $dwMemPos$0$ph$be = $48;
   }
   else if ((label|0) == 139) {
    label = 0;
    $433 = (($dwMemPos$0$ph111) + 4)|0; //@line 468 "load_dmf.cpp"
    $dwMemPos$0$ph$be = $433;
   }
   $434 = (($dwMemPos$0$ph$be) + 7)|0; //@line 106 "load_dmf.cpp"
   $435 = ($434>>>0)<($dwMemLength>>>0); //@line 106 "load_dmf.cpp"
   $436 = (($lpStream) + ($dwMemPos$0$ph$be)|0); //@line 108 "load_dmf.cpp"
   if ($435) {
    $$in = $436;$dwMemPos$0$ph111 = $dwMemPos$0$ph$be;
   } else {
    break;
   }
  }
  $$pr128 = HEAP32[$12>>2]|0; //@line 473 "load_dmf.cpp"
  $437 = ($$pr128|0)==(0); //@line 473 "load_dmf.cpp"
  if (!($437)) {
   $$0 = 1;
   STACKTOP = sp;return ($$0|0); //@line 483 "load_dmf.cpp"
  }
 }
 $438 = ((($this)) + 59684|0); //@line 475 "load_dmf.cpp"
 $439 = HEAP32[$438>>2]|0; //@line 475 "load_dmf.cpp"
 $440 = ($439|0)==(0); //@line 475 "load_dmf.cpp"
 if ($440) {
  HEAP32[$11>>2] = 0; //@line 477 "load_dmf.cpp"
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 483 "load_dmf.cpp"
 } else {
  HEAP32[$12>>2] = 4; //@line 480 "load_dmf.cpp"
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 483 "load_dmf.cpp"
 }
 return (0)|0;
}
function __ZN10CSoundFile7ReadDSMEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$0 = 0, $$17 = 0, $$21 = 0, $$22 = 0, $$23 = 0, $$80 = 0, $$lcssa = 0, $$lcssa101 = 0, $$lobit = 0, $$sum = 0, $$sum10 = 0, $$sum11 = 0, $$sum12 = 0, $$sum2 = 0, $$sum4 = 0, $$sum5 = 0, $$sum6 = 0, $$sum7 = 0, $$sum8 = 0;
 var $$sum9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $command$0 = 0, $dwMemPos$0$ph47 = 0, $dwMemPos$041 = 0, $dwMemPos$041$lcssa = 0, $dwPos$0$be = 0, $dwPos$0$ph36 = 0, $dwPos$034 = 0, $dwPos$034$lcssa = 0, $dwPos$1 = 0, $dwPos$2 = 0;
 var $dwPos$3 = 0, $exitcond = 0, $exitcond71 = 0, $iOrd$049 = 0, $iPan$048 = 0, $m$0$ph37 = 0, $nPat$0$ph46 = 0, $nPat$040 = 0, $nPat$040$lcssa = 0, $nSmp$0$ph45 = 0, $or$cond = 0, $or$cond16 = 0, $or$cond18 = 0, $or$cond2033 = 0, $or$cond25 = 0, $param$0 = 0, $row$0$ph39 = 0, $storemerge15 = 0, dest = 0, label = 0;
 var sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ($lpStream|0)==(0|0); //@line 93 "load_dsm.cpp"
 $1 = ($dwMemLength>>>0)<(1024); //@line 93 "load_dsm.cpp"
 $or$cond = $0 | $1; //@line 93 "load_dsm.cpp"
 if ($or$cond) {
  $$0 = 0;
  return ($$0|0); //@line 235 "load_dsm.cpp"
 }
 $2 = HEAPU8[$lpStream>>0]|(HEAPU8[$lpStream+1>>0]<<8)|(HEAPU8[$lpStream+2>>0]<<16)|(HEAPU8[$lpStream+3>>0]<<24); //@line 93 "load_dsm.cpp"
 $3 = ($2|0)==(1179011410); //@line 93 "load_dsm.cpp"
 if (!($3)) {
  $$0 = 0;
  return ($$0|0); //@line 235 "load_dsm.cpp"
 }
 $4 = ((($lpStream)) + 4|0); //@line 94 "load_dsm.cpp"
 $5 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8)|(HEAPU8[$4+2>>0]<<16)|(HEAPU8[$4+3>>0]<<24); //@line 94 "load_dsm.cpp"
 $6 = (($5) + 8)|0; //@line 94 "load_dsm.cpp"
 $7 = ($6>>>0)>($dwMemLength>>>0); //@line 94 "load_dsm.cpp"
 $8 = ($5>>>0)<(1024); //@line 94 "load_dsm.cpp"
 $or$cond16 = $8 | $7; //@line 94 "load_dsm.cpp"
 if ($or$cond16) {
  $$0 = 0;
  return ($$0|0); //@line 235 "load_dsm.cpp"
 }
 $9 = ((($lpStream)) + 8|0); //@line 95 "load_dsm.cpp"
 $10 = HEAPU8[$9>>0]|(HEAPU8[$9+1>>0]<<8)|(HEAPU8[$9+2>>0]<<16)|(HEAPU8[$9+3>>0]<<24); //@line 95 "load_dsm.cpp"
 $11 = ($10|0)==(1179472708); //@line 95 "load_dsm.cpp"
 if (!($11)) {
  $$0 = 0;
  return ($$0|0); //@line 235 "load_dsm.cpp"
 }
 $12 = ((($lpStream)) + 12|0); //@line 95 "load_dsm.cpp"
 $13 = HEAPU8[$12>>0]|(HEAPU8[$12+1>>0]<<8)|(HEAPU8[$12+2>>0]<<16)|(HEAPU8[$12+3>>0]<<24); //@line 95 "load_dsm.cpp"
 $14 = ($13|0)==(1196314451); //@line 95 "load_dsm.cpp"
 if (!($14)) {
  $$0 = 0;
  return ($$0|0); //@line 235 "load_dsm.cpp"
 }
 $15 = ((($lpStream)) + 16|0); //@line 96 "load_dsm.cpp"
 $16 = HEAPU8[$15>>0]|(HEAPU8[$15+1>>0]<<8)|(HEAPU8[$15+2>>0]<<16)|(HEAPU8[$15+3>>0]<<24); //@line 96 "load_dsm.cpp"
 $17 = ($16>>>0)>($dwMemLength>>>0); //@line 96 "load_dsm.cpp"
 if ($17) {
  $$0 = 0;
  return ($$0|0); //@line 235 "load_dsm.cpp"
 }
 $18 = ((($lpStream)) + 20|0); //@line 97 "load_dsm.cpp"
 $19 = ((($this)) + 59680|0); //@line 99 "load_dsm.cpp"
 HEAP32[$19>>2] = 8192; //@line 99 "load_dsm.cpp"
 $20 = ((($lpStream)) + 62|0); //@line 100 "load_dsm.cpp"
 $21 = HEAPU8[$20>>0]|(HEAPU8[$20+1>>0]<<8); //@line 100 "load_dsm.cpp"
 $22 = $21&65535; //@line 100 "load_dsm.cpp"
 $23 = ((($this)) + 59664|0); //@line 100 "load_dsm.cpp"
 $24 = ($21&65535)<(4); //@line 101 "load_dsm.cpp"
 $$ = $24 ? 4 : $22; //@line 101 "load_dsm.cpp"
 $25 = ($$>>>0)>(16); //@line 102 "load_dsm.cpp"
 $storemerge15 = $25 ? 16 : $$; //@line 102 "load_dsm.cpp"
 HEAP32[$23>>2] = $storemerge15; //@line 100 "load_dsm.cpp"
 $26 = ((($lpStream)) + 58|0); //@line 103 "load_dsm.cpp"
 $27 = HEAPU8[$26>>0]|(HEAPU8[$26+1>>0]<<8); //@line 103 "load_dsm.cpp"
 $28 = $27&65535; //@line 103 "load_dsm.cpp"
 $29 = ((($this)) + 59684|0); //@line 103 "load_dsm.cpp"
 $30 = ($27&65535)>(239); //@line 104 "load_dsm.cpp"
 $$17 = $30 ? 239 : $28; //@line 104 "load_dsm.cpp"
 HEAP32[$29>>2] = $$17; //@line 103 "load_dsm.cpp"
 $31 = ((($lpStream)) + 66|0); //@line 105 "load_dsm.cpp"
 $32 = HEAP8[$31>>0]|0; //@line 105 "load_dsm.cpp"
 $33 = $32&255; //@line 105 "load_dsm.cpp"
 $34 = ((($this)) + 59648|0); //@line 105 "load_dsm.cpp"
 HEAP32[$34>>2] = $33; //@line 105 "load_dsm.cpp"
 $35 = ((($lpStream)) + 67|0); //@line 106 "load_dsm.cpp"
 $36 = HEAP8[$35>>0]|0; //@line 106 "load_dsm.cpp"
 $37 = $36&255; //@line 106 "load_dsm.cpp"
 $38 = ((($this)) + 59652|0); //@line 106 "load_dsm.cpp"
 HEAP32[$38>>2] = $37; //@line 106 "load_dsm.cpp"
 $39 = ((($lpStream)) + 64|0); //@line 107 "load_dsm.cpp"
 $40 = HEAP8[$39>>0]|0; //@line 107 "load_dsm.cpp"
 $41 = $40&255; //@line 107 "load_dsm.cpp"
 $42 = $41 << 2; //@line 107 "load_dsm.cpp"
 $43 = ((($this)) + 59656|0); //@line 107 "load_dsm.cpp"
 $44 = ($40<<24>>24)==(0); //@line 108 "load_dsm.cpp"
 $45 = ($42>>>0)>(256); //@line 108 "load_dsm.cpp"
 $or$cond18 = $44 | $45; //@line 108 "load_dsm.cpp"
 $$23 = $or$cond18 ? 256 : $42; //@line 108 "load_dsm.cpp"
 HEAP32[$43>>2] = $$23; //@line 108 "load_dsm.cpp"
 $46 = ((($lpStream)) + 65|0); //@line 109 "load_dsm.cpp"
 $47 = HEAP8[$46>>0]|0; //@line 109 "load_dsm.cpp"
 $48 = $47&255; //@line 109 "load_dsm.cpp"
 $49 = $48 & 127; //@line 109 "load_dsm.cpp"
 $50 = ((($this)) + 59748|0); //@line 109 "load_dsm.cpp"
 HEAP32[$50>>2] = $49; //@line 109 "load_dsm.cpp"
 $51 = ((($lpStream)) + 56|0); //@line 112 "load_dsm.cpp"
 $iOrd$049 = 0;
 while(1) {
  $53 = HEAPU8[$51>>0]|(HEAPU8[$51+1>>0]<<8); //@line 112 "load_dsm.cpp"
  $54 = $53&65535; //@line 112 "load_dsm.cpp"
  $55 = ($iOrd$049>>>0)<($54>>>0); //@line 112 "load_dsm.cpp"
  if ($55) {
   $$sum12 = (($iOrd$049) + 84)|0; //@line 112 "load_dsm.cpp"
   $56 = (($lpStream) + ($$sum12)|0); //@line 112 "load_dsm.cpp"
   $57 = HEAP8[$56>>0]|0; //@line 112 "load_dsm.cpp"
   $58 = $57&255; //@line 112 "load_dsm.cpp"
   $60 = $58;
  } else {
   $60 = 255;
  }
  $59 = $60&255; //@line 112 "load_dsm.cpp"
  $61 = (((($this)) + 53344|0) + ($iOrd$049)|0); //@line 112 "load_dsm.cpp"
  HEAP8[$61>>0] = $59; //@line 112 "load_dsm.cpp"
  $62 = (($iOrd$049) + 1)|0; //@line 110 "load_dsm.cpp"
  $exitcond71 = ($62|0)==(256); //@line 110 "load_dsm.cpp"
  if ($exitcond71) {
   break;
  } else {
   $iOrd$049 = $62;
  }
 }
 $52 = (($16) + 20)|0; //@line 98 "load_dsm.cpp"
 $iPan$048 = 0;
 while(1) {
  $63 = (((($this)) + 49600|0) + (($iPan$048*36)|0)|0); //@line 116 "load_dsm.cpp"
  HEAP32[$63>>2] = 128; //@line 116 "load_dsm.cpp"
  $$sum11 = (($iPan$048) + 68)|0; //@line 117 "load_dsm.cpp"
  $64 = (($lpStream) + ($$sum11)|0); //@line 117 "load_dsm.cpp"
  $65 = HEAP8[$64>>0]|0; //@line 117 "load_dsm.cpp"
  $66 = ($65&255)<(129); //@line 117 "load_dsm.cpp"
  if ($66) {
   $67 = $65&255; //@line 117 "load_dsm.cpp"
   $68 = $67 << 1; //@line 119 "load_dsm.cpp"
   HEAP32[$63>>2] = $68; //@line 119 "load_dsm.cpp"
  }
  $69 = (($iPan$048) + 1)|0; //@line 114 "load_dsm.cpp"
  $exitcond = ($69|0)==(16); //@line 114 "load_dsm.cpp"
  if ($exitcond) {
   break;
  } else {
   $iPan$048 = $69;
  }
 }
 $70 = ((($this)) + 59804|0); //@line 122 "load_dsm.cpp"
 dest=$70; src=$18; stop=dest+28|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 122 "load_dsm.cpp"
 $71 = (($dwMemLength) + -8)|0; //@line 125 "load_dsm.cpp"
 $72 = ($52>>>0)<($71>>>0); //@line 125 "load_dsm.cpp"
 if ($72) {
  $dwMemPos$0$ph47 = $52;$nPat$0$ph46 = 0;$nSmp$0$ph45 = 1;
 } else {
  $$0 = 1;
  return ($$0|0); //@line 235 "load_dsm.cpp"
 }
 L33: while(1) {
  $dwMemPos$041 = $dwMemPos$0$ph47;$nPat$040 = $nPat$0$ph46;
  while(1) {
   $73 = (($lpStream) + ($dwMemPos$041)|0); //@line 127 "load_dsm.cpp"
   $74 = HEAPU8[$73>>0]|(HEAPU8[$73+1>>0]<<8)|(HEAPU8[$73+2>>0]<<16)|(HEAPU8[$73+3>>0]<<24); //@line 130 "load_dsm.cpp"
   $75 = ($74|0)==(1414807888); //@line 130 "load_dsm.cpp"
   if (!($75)) {
    $$lcssa101 = $74;$dwMemPos$041$lcssa = $dwMemPos$041;$nPat$040$lcssa = $nPat$040;
    break;
   }
   $76 = (($dwMemPos$041) + 8)|0; //@line 132 "load_dsm.cpp"
   $$sum10 = (($dwMemPos$041) + 4)|0; //@line 133 "load_dsm.cpp"
   $77 = (($lpStream) + ($$sum10)|0); //@line 133 "load_dsm.cpp"
   $78 = HEAPU8[$77>>0]|(HEAPU8[$77+1>>0]<<8)|(HEAPU8[$77+2>>0]<<16)|(HEAPU8[$77+3>>0]<<24); //@line 133 "load_dsm.cpp"
   $79 = (($78) + ($76))|0; //@line 133 "load_dsm.cpp"
   $80 = ($79>>>0)<($dwMemLength>>>0); //@line 133 "load_dsm.cpp"
   if (!($80)) {
    $$0 = 1;
    label = 48;
    break L33;
   }
   $81 = HEAP32[$23>>2]|0; //@line 136 "load_dsm.cpp"
   $82 = (__ZN10CSoundFile15AllocatePatternEjj(64,$81)|0); //@line 136 "load_dsm.cpp"
   $83 = ($82|0)==(0|0); //@line 137 "load_dsm.cpp"
   if ($83) {
    $$0 = 1;
    label = 48;
    break L33;
   }
   $84 = (((($this)) + 52864|0) + ($nPat$040<<1)|0); //@line 138 "load_dsm.cpp"
   HEAP16[$84>>1] = 64; //@line 138 "load_dsm.cpp"
   $85 = (((($this)) + 51904|0) + ($nPat$040<<2)|0); //@line 139 "load_dsm.cpp"
   HEAP32[$85>>2] = $82; //@line 139 "load_dsm.cpp"
   $86 = (($dwMemPos$041) + 10)|0;
   $87 = ($86>>>0)>($79>>>0); //@line 141 "load_dsm.cpp"
   L40: do {
    if (!($87)) {
     $207 = $86;$dwPos$0$ph36 = $76;$m$0$ph37 = $82;$row$0$ph39 = 0;
     while(1) {
      $208 = $207;$dwPos$034 = $dwPos$0$ph36;
      while(1) {
       $88 = (($dwPos$034) + 1)|0; //@line 143 "load_dsm.cpp"
       $89 = (($lpStream) + ($dwPos$034)|0); //@line 143 "load_dsm.cpp"
       $90 = HEAP8[$89>>0]|0; //@line 143 "load_dsm.cpp"
       $91 = $90&255; //@line 143 "load_dsm.cpp"
       $92 = ($90<<24>>24)==(0); //@line 144 "load_dsm.cpp"
       if ($92) {
        $$lcssa = $88;$dwPos$034$lcssa = $dwPos$034;
        break;
       }
       $93 = $91 & 15; //@line 146 "load_dsm.cpp"
       $94 = HEAP32[$23>>2]|0; //@line 146 "load_dsm.cpp"
       $95 = (($93>>>0) % ($94>>>0))&-1; //@line 146 "load_dsm.cpp"
       $96 = $91 & 128; //@line 147 "load_dsm.cpp"
       $97 = ($96|0)==(0); //@line 147 "load_dsm.cpp"
       if ($97) {
        $dwPos$1 = $88;
       } else {
        $98 = (($lpStream) + ($88)|0); //@line 149 "load_dsm.cpp"
        $99 = HEAP8[$98>>0]|0; //@line 149 "load_dsm.cpp"
        $100 = $99&255; //@line 149 "load_dsm.cpp"
        $101 = ($99<<24>>24)==(0); //@line 150 "load_dsm.cpp"
        if ($101) {
         $dwPos$1 = $208;
        } else {
         $102 = ($99&255)<(109); //@line 152 "load_dsm.cpp"
         $103 = (($100) + 12)|0; //@line 152 "load_dsm.cpp"
         $$21 = $102 ? $103 : $100; //@line 152 "load_dsm.cpp"
         $104 = $$21&255; //@line 153 "load_dsm.cpp"
         $105 = (($m$0$ph37) + (($95*6)|0)|0); //@line 153 "load_dsm.cpp"
         HEAP8[$105>>0] = $104; //@line 153 "load_dsm.cpp"
         $dwPos$1 = $208;
        }
       }
       $106 = $91 & 64; //@line 156 "load_dsm.cpp"
       $107 = ($106|0)==(0); //@line 156 "load_dsm.cpp"
       if ($107) {
        $dwPos$2 = $dwPos$1;
       } else {
        $108 = (($dwPos$1) + 1)|0; //@line 158 "load_dsm.cpp"
        $109 = (($lpStream) + ($dwPos$1)|0); //@line 158 "load_dsm.cpp"
        $110 = HEAP8[$109>>0]|0; //@line 158 "load_dsm.cpp"
        $111 = (((($m$0$ph37) + (($95*6)|0)|0)) + 1|0); //@line 158 "load_dsm.cpp"
        HEAP8[$111>>0] = $110; //@line 158 "load_dsm.cpp"
        $dwPos$2 = $108;
       }
       $112 = $91 & 32; //@line 160 "load_dsm.cpp"
       $113 = ($112|0)==(0); //@line 160 "load_dsm.cpp"
       if ($113) {
        $dwPos$3 = $dwPos$2;
       } else {
        $114 = (((($m$0$ph37) + (($95*6)|0)|0)) + 2|0); //@line 162 "load_dsm.cpp"
        HEAP8[$114>>0] = 1; //@line 162 "load_dsm.cpp"
        $115 = (($dwPos$2) + 1)|0; //@line 163 "load_dsm.cpp"
        $116 = (($lpStream) + ($dwPos$2)|0); //@line 163 "load_dsm.cpp"
        $117 = HEAP8[$116>>0]|0; //@line 163 "load_dsm.cpp"
        $118 = (((($m$0$ph37) + (($95*6)|0)|0)) + 4|0); //@line 163 "load_dsm.cpp"
        HEAP8[$118>>0] = $117; //@line 163 "load_dsm.cpp"
        $dwPos$3 = $115;
       }
       $119 = $91 & 16; //@line 165 "load_dsm.cpp"
       $120 = ($119|0)==(0); //@line 165 "load_dsm.cpp"
       do {
        if ($120) {
         $dwPos$0$be = $dwPos$3;
        } else {
         $121 = (($dwPos$3) + 1)|0; //@line 167 "load_dsm.cpp"
         $122 = (($lpStream) + ($dwPos$3)|0); //@line 167 "load_dsm.cpp"
         $123 = HEAP8[$122>>0]|0; //@line 167 "load_dsm.cpp"
         $124 = $123&255; //@line 167 "load_dsm.cpp"
         $125 = (($dwPos$3) + 2)|0; //@line 168 "load_dsm.cpp"
         $126 = (($lpStream) + ($121)|0); //@line 168 "load_dsm.cpp"
         $127 = HEAP8[$126>>0]|0; //@line 168 "load_dsm.cpp"
         $128 = $127&255; //@line 168 "load_dsm.cpp"
         L58: do {
          switch ($124|0) {
          case 8:  {
           $129 = $128 & 240; //@line 173 "load_dsm.cpp"
           switch ($129|0) {
           case 0:  {
            $130 = $128 << 4; //@line 175 "load_dsm.cpp"
            $command$0 = 8;$param$0 = $130;
            break L58;
            break;
           }
           case 16:  {
            $131 = $128 << 4; //@line 176 "load_dsm.cpp"
            $132 = $131 & 240; //@line 176 "load_dsm.cpp"
            $command$0 = 10;$param$0 = $132;
            break L58;
            break;
           }
           case 32:  {
            $133 = $128 & 15; //@line 177 "load_dsm.cpp"
            $134 = $133 | 160; //@line 177 "load_dsm.cpp"
            $command$0 = 14;$param$0 = $134;
            break L58;
            break;
           }
           case 48:  {
            $135 = $128 & 15; //@line 178 "load_dsm.cpp"
            $136 = $135 | 16; //@line 178 "load_dsm.cpp"
            $command$0 = 14;$param$0 = $136;
            break L58;
            break;
           }
           case 64:  {
            $137 = $128 & 15; //@line 179 "load_dsm.cpp"
            $138 = $137 | 32; //@line 179 "load_dsm.cpp"
            $command$0 = 14;$param$0 = $138;
            break L58;
            break;
           }
           default: {
            $command$0 = 0;$param$0 = $128;
            break L58;
           }
           }
           break;
          }
          case 18: case 17:  {
           $139 = $124 & 15; //@line 186 "load_dsm.cpp"
           $command$0 = $139;$param$0 = $128;
           break;
          }
          case 19:  {
           $command$0 = 33;$param$0 = 145;
           break;
          }
          default: {
           $140 = $124 & 240; //@line 195 "load_dsm.cpp"
           $141 = ($140|0)==(32); //@line 195 "load_dsm.cpp"
           $$80 = $141 ? 9 : 0;
           $command$0 = $$80;$param$0 = $128;
          }
          }
         } while(0);
         $142 = $command$0&255; //@line 197 "load_dsm.cpp"
         $143 = (((($m$0$ph37) + (($95*6)|0)|0)) + 3|0); //@line 197 "load_dsm.cpp"
         HEAP8[$143>>0] = $142; //@line 197 "load_dsm.cpp"
         $144 = $param$0&255; //@line 198 "load_dsm.cpp"
         $145 = (((($m$0$ph37) + (($95*6)|0)|0)) + 5|0); //@line 198 "load_dsm.cpp"
         HEAP8[$145>>0] = $144; //@line 198 "load_dsm.cpp"
         $146 = ($command$0|0)==(0); //@line 199 "load_dsm.cpp"
         if ($146) {
          $dwPos$0$be = $125;
          break;
         }
         $147 = (($m$0$ph37) + (($95*6)|0)|0); //@line 197 "load_dsm.cpp"
         __ZNK10CSoundFile17ConvertModCommandEP11_MODCOMMAND($this,$147); //@line 199 "load_dsm.cpp"
         $dwPos$0$be = $125;
        }
       } while(0);
       $148 = (($dwPos$0$be) + 2)|0;
       $149 = ($148>>>0)>($79>>>0); //@line 141 "load_dsm.cpp"
       if ($149) {
        break L40;
       } else {
        $208 = $148;$dwPos$034 = $dwPos$0$be;
       }
      }
      $150 = HEAP32[$23>>2]|0; //@line 203 "load_dsm.cpp"
      $151 = (($m$0$ph37) + (($150*6)|0)|0); //@line 203 "load_dsm.cpp"
      $152 = (($row$0$ph39) + 1)|0; //@line 204 "load_dsm.cpp"
      $153 = ($152>>>0)>(63); //@line 141 "load_dsm.cpp"
      $154 = (($dwPos$034$lcssa) + 3)|0;
      $155 = ($154>>>0)>($79>>>0); //@line 141 "load_dsm.cpp"
      $or$cond2033 = $155 | $153; //@line 141 "load_dsm.cpp"
      if ($or$cond2033) {
       break;
      } else {
       $207 = $154;$dwPos$0$ph36 = $$lcssa;$m$0$ph37 = $151;$row$0$ph39 = $152;
      }
     }
    }
   } while(0);
   $156 = (($nPat$040) + 1)|0; //@line 207 "load_dsm.cpp"
   $157 = ($79>>>0)<($71>>>0); //@line 125 "load_dsm.cpp"
   if ($157) {
    $dwMemPos$041 = $79;$nPat$040 = $156;
   } else {
    $$0 = 1;
    label = 48;
    break L33;
   }
  }
  $158 = HEAP32[$29>>2]|0; //@line 210 "load_dsm.cpp"
  $159 = ($nSmp$0$ph45>>>0)<=($158>>>0); //@line 210 "load_dsm.cpp"
  $160 = ($$lcssa101|0)==(1414745673); //@line 210 "load_dsm.cpp"
  $or$cond25 = $160 & $159; //@line 210 "load_dsm.cpp"
  if (!($or$cond25)) {
   $$0 = 1;
   label = 48;
   break;
  }
  $$sum = (($dwMemPos$041$lcssa) + 4)|0; //@line 212 "load_dsm.cpp"
  $161 = (($lpStream) + ($$sum)|0); //@line 212 "load_dsm.cpp"
  $162 = HEAPU8[$161>>0]|(HEAPU8[$161+1>>0]<<8)|(HEAPU8[$161+2>>0]<<16)|(HEAPU8[$161+3>>0]<<24); //@line 212 "load_dsm.cpp"
  $163 = (($162) + ($dwMemPos$041$lcssa))|0; //@line 212 "load_dsm.cpp"
  $164 = ($163>>>0)<($71>>>0); //@line 212 "load_dsm.cpp"
  if (!($164)) {
   $$0 = 1;
   label = 48;
   break;
  }
  $165 = (($dwMemPos$041$lcssa) + 72)|0; //@line 213 "load_dsm.cpp"
  $166 = (($dwMemPos$041$lcssa) + 8)|0; //@line 214 "load_dsm.cpp"
  $167 = (($166) + ($162))|0; //@line 214 "load_dsm.cpp"
  $168 = (((($this)) + 59804|0) + ($nSmp$0$ph45<<5)|0); //@line 215 "load_dsm.cpp"
  $$sum2 = (($dwMemPos$041$lcssa) + 44)|0; //@line 215 "load_dsm.cpp"
  $169 = (($lpStream) + ($$sum2)|0); //@line 215 "load_dsm.cpp"
  dest=$168; src=$169; stop=dest+28|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 215 "load_dsm.cpp"
  $170 = (((($this)) + 33280|0) + ($nSmp$0$ph45<<6)|0); //@line 216 "load_dsm.cpp"
  $171 = (((((($this)) + 33280|0) + ($nSmp$0$ph45<<6)|0)) + 42|0); //@line 217 "load_dsm.cpp"
  $172 = (($lpStream) + ($166)|0); //@line 217 "load_dsm.cpp"
  dest=$171; src=$172; stop=dest+13|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 217 "load_dsm.cpp"
  $173 = (((((($this)) + 33280|0) + ($nSmp$0$ph45<<6)|0)) + 32|0); //@line 218 "load_dsm.cpp"
  HEAP16[$173>>1] = 64; //@line 218 "load_dsm.cpp"
  $$sum4 = (($dwMemPos$041$lcssa) + 40)|0; //@line 219 "load_dsm.cpp"
  $174 = (($lpStream) + ($$sum4)|0); //@line 219 "load_dsm.cpp"
  $175 = HEAPU8[$174>>0]|(HEAPU8[$174+1>>0]<<8); //@line 219 "load_dsm.cpp"
  $176 = $175&65535; //@line 219 "load_dsm.cpp"
  $177 = (((((($this)) + 33280|0) + ($nSmp$0$ph45<<6)|0)) + 24|0); //@line 219 "load_dsm.cpp"
  HEAP32[$177>>2] = $176; //@line 219 "load_dsm.cpp"
  $$sum5 = (($dwMemPos$041$lcssa) + 21)|0; //@line 220 "load_dsm.cpp"
  $178 = (($lpStream) + ($$sum5)|0); //@line 220 "load_dsm.cpp"
  $179 = HEAP8[$178>>0]|0; //@line 220 "load_dsm.cpp"
  $180 = $179 & 1; //@line 220 "load_dsm.cpp"
  $181 = $180&255; //@line 220 "load_dsm.cpp"
  $182 = $181 << 1; //@line 220 "load_dsm.cpp"
  $183 = $182&65535; //@line 220 "load_dsm.cpp"
  $184 = (((((($this)) + 33280|0) + ($nSmp$0$ph45<<6)|0)) + 34|0); //@line 220 "load_dsm.cpp"
  HEAP16[$184>>1] = $183; //@line 220 "load_dsm.cpp"
  $$sum6 = (($dwMemPos$041$lcssa) + 24)|0; //@line 221 "load_dsm.cpp"
  $185 = (($lpStream) + ($$sum6)|0); //@line 221 "load_dsm.cpp"
  $186 = HEAPU8[$185>>0]|(HEAPU8[$185+1>>0]<<8)|(HEAPU8[$185+2>>0]<<16)|(HEAPU8[$185+3>>0]<<24); //@line 221 "load_dsm.cpp"
  HEAP32[$170>>2] = $186; //@line 221 "load_dsm.cpp"
  $$sum7 = (($dwMemPos$041$lcssa) + 28)|0; //@line 222 "load_dsm.cpp"
  $187 = (($lpStream) + ($$sum7)|0); //@line 222 "load_dsm.cpp"
  $188 = HEAPU8[$187>>0]|(HEAPU8[$187+1>>0]<<8)|(HEAPU8[$187+2>>0]<<16)|(HEAPU8[$187+3>>0]<<24); //@line 222 "load_dsm.cpp"
  $189 = (((((($this)) + 33280|0) + ($nSmp$0$ph45<<6)|0)) + 4|0); //@line 222 "load_dsm.cpp"
  HEAP32[$189>>2] = $188; //@line 222 "load_dsm.cpp"
  $$sum8 = (($dwMemPos$041$lcssa) + 32)|0; //@line 223 "load_dsm.cpp"
  $190 = (($lpStream) + ($$sum8)|0); //@line 223 "load_dsm.cpp"
  $191 = HEAPU8[$190>>0]|(HEAPU8[$190+1>>0]<<8)|(HEAPU8[$190+2>>0]<<16)|(HEAPU8[$190+3>>0]<<24); //@line 223 "load_dsm.cpp"
  $192 = (((((($this)) + 33280|0) + ($nSmp$0$ph45<<6)|0)) + 8|0); //@line 223 "load_dsm.cpp"
  HEAP32[$192>>2] = $191; //@line 223 "load_dsm.cpp"
  $$sum9 = (($dwMemPos$041$lcssa) + 23)|0; //@line 224 "load_dsm.cpp"
  $193 = (($lpStream) + ($$sum9)|0); //@line 224 "load_dsm.cpp"
  $194 = HEAP8[$193>>0]|0; //@line 224 "load_dsm.cpp"
  $195 = $194&255; //@line 224 "load_dsm.cpp"
  $196 = $195 << 2; //@line 224 "load_dsm.cpp"
  $197 = $196&65535; //@line 224 "load_dsm.cpp"
  $198 = (((((($this)) + 33280|0) + ($nSmp$0$ph45<<6)|0)) + 30|0); //@line 224 "load_dsm.cpp"
  $199 = ($196>>>0)>(256); //@line 225 "load_dsm.cpp"
  $$22 = $199 ? 256 : $197; //@line 225 "load_dsm.cpp"
  HEAP16[$198>>1] = $$22; //@line 224 "load_dsm.cpp"
  $200 = ($179&255) >>> 1; //@line 226 "load_dsm.cpp"
  $$lobit = $200 & 1; //@line 226 "load_dsm.cpp"
  $201 = $$lobit ^ 1; //@line 226 "load_dsm.cpp"
  $202 = $201&255; //@line 226 "load_dsm.cpp"
  $203 = (($lpStream) + ($165)|0); //@line 227 "load_dsm.cpp"
  $204 = (($dwMemLength) - ($165))|0; //@line 227 "load_dsm.cpp"
  (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$170,$202,$203,$204)|0); //@line 227 "load_dsm.cpp"
  $205 = (($nSmp$0$ph45) + 1)|0; //@line 228 "load_dsm.cpp"
  $206 = ($167>>>0)<($71>>>0); //@line 125 "load_dsm.cpp"
  if ($206) {
   $dwMemPos$0$ph47 = $167;$nPat$0$ph46 = $nPat$040$lcssa;$nSmp$0$ph45 = $205;
  } else {
   $$0 = 1;
   label = 48;
   break;
  }
 }
 if ((label|0) == 48) {
  return ($$0|0); //@line 235 "load_dsm.cpp"
 }
 return (0)|0;
}
function __ZN10CSoundFile7ReadUMXEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa51 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond3 = 0, $or$cond5 = 0, $switch$split2D = 0, $uscan$013 = 0, $uscan$013$lcssa = 0;
 var $uscan$013$lcssa46 = 0, $uscan$013$lcssa47 = 0, $uscan$013$lcssa48 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($lpStream|0)==(0|0); //@line 16 "load_umx.cpp"
 $1 = ($dwMemLength>>>0)<(2048); //@line 16 "load_umx.cpp"
 $or$cond = $0 | $1; //@line 16 "load_umx.cpp"
 if ($or$cond) {
  $$0 = 0;
  return ($$0|0); //@line 52 "load_umx.cpp"
 }
 $2 = ((($lpStream)) + 32|0); //@line 18 "load_umx.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 18 "load_umx.cpp"
 $4 = ($3>>>0)<($dwMemLength>>>0); //@line 18 "load_umx.cpp"
 if (!($4)) {
  $$0 = 0;
  return ($$0|0); //@line 52 "load_umx.cpp"
 }
 $5 = ((($lpStream)) + 24|0); //@line 19 "load_umx.cpp"
 $6 = HEAP32[$5>>2]|0; //@line 19 "load_umx.cpp"
 $7 = (($dwMemLength) + -16)|0; //@line 19 "load_umx.cpp"
 $8 = ($6>>>0)>($7>>>0); //@line 19 "load_umx.cpp"
 $9 = (($dwMemLength) + -512)|0;
 $10 = ($6>>>0)<($9>>>0); //@line 20 "load_umx.cpp"
 $or$cond5 = $8 | $10; //@line 20 "load_umx.cpp"
 if ($or$cond5) {
  $$0 = 0;
  return ($$0|0); //@line 52 "load_umx.cpp"
 } else {
  $uscan$013 = 64;
 }
 L9: while(1) {
  $11 = (($lpStream) + ($uscan$013)|0); //@line 24 "load_umx.cpp"
  $12 = HEAP32[$11>>2]|0; //@line 24 "load_umx.cpp"
  $switch$split2D = ($12|0)<(1297236819);
  if ($switch$split2D) {
   switch ($12|0) {
   case 1297108297:  {
    $$lcssa = $11;$uscan$013$lcssa = $uscan$013;
    label = 5;
    break L9;
    break;
   }
   default: {
   }
   }
  } else {
   switch ($12|0) {
   case 1297236819:  {
    $uscan$013$lcssa46 = $uscan$013;
    label = 6;
    break L9;
    break;
   }
   default: {
   }
   }
  }
  $19 = (_strncasecmp($11,529876,15)|0); //@line 38 "load_umx.cpp"
  $20 = ($19|0)==(0); //@line 38 "load_umx.cpp"
  if ($20) {
   $$lcssa51 = $11;$uscan$013$lcssa47 = $uscan$013;
   label = 8;
   break;
  }
  $23 = ($uscan$013>>>0)>(1080); //@line 44 "load_umx.cpp"
  $24 = ($12|0)==(776678989); //@line 44 "load_umx.cpp"
  $or$cond3 = $23 & $24; //@line 44 "load_umx.cpp"
  if ($or$cond3) {
   $uscan$013$lcssa48 = $uscan$013;
   label = 10;
   break;
  }
  $29 = (($uscan$013) + 1)|0; //@line 22 "load_umx.cpp"
  $30 = ($29>>>0)<(1280); //@line 22 "load_umx.cpp"
  if ($30) {
   $uscan$013 = $29;
  } else {
   $$0 = 0;
   label = 12;
   break;
  }
 }
 if ((label|0) == 5) {
  $13 = (($dwMemLength) - ($uscan$013$lcssa))|0; //@line 29 "load_umx.cpp"
  $14 = (__ZN10CSoundFile6ReadITEPKhj($this,$$lcssa,$13)|0); //@line 29 "load_umx.cpp"
  $$0 = $14;
  return ($$0|0); //@line 52 "load_umx.cpp"
 }
 else if ((label|0) == 6) {
  $15 = (($uscan$013$lcssa46) + -44)|0; //@line 34 "load_umx.cpp"
  $16 = (($lpStream) + ($15)|0); //@line 35 "load_umx.cpp"
  $17 = (($dwMemLength) - ($15))|0; //@line 35 "load_umx.cpp"
  $18 = (__ZN10CSoundFile7ReadS3MEPKhj($this,$16,$17)|0); //@line 35 "load_umx.cpp"
  $$0 = $18;
  return ($$0|0); //@line 52 "load_umx.cpp"
 }
 else if ((label|0) == 8) {
  $21 = (($dwMemLength) - ($uscan$013$lcssa47))|0; //@line 41 "load_umx.cpp"
  $22 = (__ZN10CSoundFile6ReadXMEPKhj($this,$$lcssa51,$21)|0); //@line 41 "load_umx.cpp"
  $$0 = $22;
  return ($$0|0); //@line 52 "load_umx.cpp"
 }
 else if ((label|0) == 10) {
  $25 = (($uscan$013$lcssa48) + -1080)|0; //@line 46 "load_umx.cpp"
  $26 = (($lpStream) + ($25)|0); //@line 47 "load_umx.cpp"
  $27 = (($dwMemLength) - ($25))|0; //@line 47 "load_umx.cpp"
  $28 = (__ZN10CSoundFile7ReadModEPKhj($this,$26,$27)|0); //@line 47 "load_umx.cpp"
  $$0 = $28;
  return ($$0|0); //@line 52 "load_umx.cpp"
 }
 else if ((label|0) == 12) {
  return ($$0|0); //@line 52 "load_umx.cpp"
 }
 return (0)|0;
}
function __ZN10CSoundFile7ReadModEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$0 = 0, $$32 = 0, $$37 = 0, $$lcssa135 = 0, $$lcssa136 = 0, $$lcssa137 = 0, $$nbpbuggy$0 = 0, $$off = 0, $$off60 = 0, $$off62 = 0, $$off64 = 0, $$off66 = 0, $$off68 = 0, $$off70 = 0, $$pre = 0, $$sink = 0, $$sum = 0, $$sum12 = 0, $$sum13 = 0;
 var $$sum14 = 0, $$sum15 = 0, $$sum16 = 0, $$sum89 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0;
 var $275 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $dwErrCheck$0$lcssa = 0, $dwErrCheck$080 = 0, $dwErrCheck$1 = 0, $dwMemPos$0110 = 0, $dwMemPos$0110$lcssa = 0, $dwMemPos$2$lcssa = 0, $dwMemPos$290 = 0, $dwMemPos$379 = 0, $dwMemPos$4 = 0, $dwMemPos$5 = 0, $dwTotalSampleLen$0109 = 0, $exitcond = 0, $flags$0 = 0, $i$0107 = 0, $ich$097 = 0, $iord$0106 = 0, $ipat$089 = 0, $ismp$081 = 0, $j$086 = 0, $m$084 = 0;
 var $nErr$0$ = 0, $nErr$0108 = 0, $nErr$1 = 0, $nErr$1$lcssa = 0, $nErr$2 = 0, $nErr$3 = 0, $nbp$0103 = 0, $nbp$1 = 0, $nbp$1$lcssa = 0, $nbp$2 = 0, $nbpbuggy$0$nbp$0 = 0, $nbpbuggy$0104 = 0, $nbpbuggy$1 = 0, $nbpbuggy$1$lcssa = 0, $nbpbuggy2$0$ = 0, $nbpbuggy2$0$$lcssa = 0, $nbpbuggy2$0$nbp$0 = 0, $nbpbuggy2$0105 = 0, $norders$0 = 0, $norders$1 = 0;
 var $or$cond = 0, $or$cond25 = 0, $or$cond26 = 0, $or$cond28 = 0, $or$cond30 = 0, $or$cond31 = 0, $or$cond33 = 0, $or$cond48 = 0, $or$cond49 = 0, $or$cond5 = 0, $or$cond50 = 0, $or$cond51 = 0, $or$cond52 = 0, $or$cond53 = 0, $or$cond54 = 0, $or$cond55 = 0, $or$cond56 = 0, $or$cond57 = 0, $or$cond58 = 0, $or$cond72 = 0;
 var $or$cond74 = 0, $p$085 = 0, $p4$0 = 0, $scevgep = 0, $sext = 0, $sext63 = 0, $sext65 = 0, $sext67 = 0, $storemerge17 = 0, $switch$split102D = 0, $switch$split12D = 0, $switch$split132D = 0, $switch$split2D = 0, $switch$split42D = 0, $switch$split72D = 0, $uglygep = 0, dest = 0, label = 0, sp = 0, src = 0;
 var stop = 0;
 sp = STACKTOP;
 $0 = ($lpStream|0)==(0|0); //@line 185 "load_mod.cpp"
 $1 = ($dwMemLength>>>0)<(1536); //@line 185 "load_mod.cpp"
 $or$cond = $0 | $1; //@line 185 "load_mod.cpp"
 if ($or$cond) {
  $$0 = 0;
  return ($$0|0); //@line 363 "load_mod.cpp"
 }
 $2 = ((($this)) + 59684|0); //@line 187 "load_mod.cpp"
 HEAP32[$2>>2] = 31; //@line 187 "load_mod.cpp"
 $3 = ((($this)) + 59664|0); //@line 188 "load_mod.cpp"
 HEAP32[$3>>2] = 4; //@line 188 "load_mod.cpp"
 $4 = ((($lpStream)) + 1080|0); //@line 191 "load_mod.cpp"
 $5 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8)|(HEAPU8[$4+2>>0]<<16)|(HEAPU8[$4+3>>0]<<24); //@line 191 "load_mod.cpp"
 $6 = $5&255; //@line 191 "load_mod.cpp"
 $7 = $5 >>> 8; //@line 191 "load_mod.cpp"
 $8 = $7&255; //@line 191 "load_mod.cpp"
 $9 = $5 >>> 16; //@line 191 "load_mod.cpp"
 $10 = $9&255; //@line 191 "load_mod.cpp"
 $11 = $5 >>> 24; //@line 191 "load_mod.cpp"
 $12 = $11&255; //@line 191 "load_mod.cpp"
 $13 = $11 << 24; //@line 173 "load_mod.cpp"
 $14 = $5 & 16777215; //@line 173 "load_mod.cpp"
 $15 = $14 | $13; //@line 173 "load_mod.cpp"
 $switch$split2D = ($15|0)<(777268814);
 L4: do {
  if ($switch$split2D) {
   $switch$split42D = ($15|0)<(558573133);
   if ($switch$split42D) {
    switch ($15|0) {
    case 558571853:  {
     label = 3;
     break L4;
     break;
    }
    default: {
     label = 5;
     break L4;
    }
    }
   }
   $switch$split102D = ($15|0)<(776678989);
   if ($switch$split102D) {
    switch ($15|0) {
    case 558573133:  {
     label = 3;
     break L4;
     break;
    }
    default: {
     label = 5;
     break L4;
    }
    }
   } else {
    switch ($15|0) {
    case 776678989:  {
     label = 3;
     break L4;
     break;
    }
    default: {
     label = 5;
     break L4;
    }
    }
   }
  } else {
   $switch$split72D = ($15|0)<(825771075);
   if ($switch$split72D) {
    switch ($15|0) {
    case 777268814:  {
     label = 3;
     break L4;
     break;
    }
    default: {
     label = 5;
     break L4;
    }
    }
   }
   $switch$split132D = ($15|0)<(1096043343);
   if ($switch$split132D) {
    switch ($15|0) {
    case 825771075:  {
     break;
    }
    default: {
     label = 5;
     break L4;
    }
    }
   } else {
    switch ($15|0) {
    case 1096043343:  {
     break;
    }
    default: {
     label = 5;
     break L4;
    }
    }
   }
   HEAP32[$3>>2] = 8; //@line 194 "load_mod.cpp"
   $dwMemPos$0110 = 20;$dwTotalSampleLen$0109 = 0;$i$0107 = 1;$nErr$0108 = 0;
  }
 } while(0);
 L20: do {
  if ((label|0) == 3) {
   HEAP32[$3>>2] = 4; //@line 193 "load_mod.cpp"
   $dwMemPos$0110 = 20;$dwTotalSampleLen$0109 = 0;$i$0107 = 1;$nErr$0108 = 0;
  }
  else if ((label|0) == 5) {
   $16 = ($6<<24>>24)==(70); //@line 195 "load_mod.cpp"
   L22: do {
    if ($16) {
     $17 = ($8<<24>>24)==(76); //@line 195 "load_mod.cpp"
     $18 = ($10<<24>>24)==(84); //@line 195 "load_mod.cpp"
     $or$cond48 = $17 & $18; //@line 195 "load_mod.cpp"
     $$off70 = (($12) + -52)<<24>>24;
     $19 = ($$off70&255)<(6); //@line 195 "load_mod.cpp"
     $or$cond72 = $or$cond48 & $19; //@line 195 "load_mod.cpp"
     if ($or$cond72) {
      $20 = $5 >> 24; //@line 195 "load_mod.cpp"
      $21 = (($20) + -48)|0; //@line 195 "load_mod.cpp"
      HEAP32[$3>>2] = $21; //@line 195 "load_mod.cpp"
      $dwMemPos$0110 = 20;$dwTotalSampleLen$0109 = 0;$i$0107 = 1;$nErr$0108 = 0;
      break L20;
     }
    } else {
     $$off60 = (($6) + -50)<<24>>24; //@line 196 "load_mod.cpp"
     $22 = ($$off60&255)<(8); //@line 196 "load_mod.cpp"
     $23 = ($8<<24>>24)==(67); //@line 196 "load_mod.cpp"
     $or$cond49 = $22 & $23; //@line 196 "load_mod.cpp"
     $24 = ($10<<24>>24)==(72); //@line 196 "load_mod.cpp"
     $or$cond50 = $24 & $or$cond49; //@line 196 "load_mod.cpp"
     $25 = ($12<<24>>24)==(78); //@line 196 "load_mod.cpp"
     $or$cond51 = $25 & $or$cond50; //@line 196 "load_mod.cpp"
     if ($or$cond51) {
      $sext = $5 << 24; //@line 196 "load_mod.cpp"
      $26 = $sext >> 24; //@line 196 "load_mod.cpp"
      $27 = (($26) + -48)|0; //@line 196 "load_mod.cpp"
      HEAP32[$3>>2] = $27; //@line 196 "load_mod.cpp"
      $dwMemPos$0110 = 20;$dwTotalSampleLen$0109 = 0;$i$0107 = 1;$nErr$0108 = 0;
      break L20;
     }
     switch ($6<<24>>24) {
     case 49:  {
      $$off68 = (($8) + -48)<<24>>24; //@line 197 "load_mod.cpp"
      $28 = ($$off68&255)<(10); //@line 197 "load_mod.cpp"
      $29 = ($10<<24>>24)==(67); //@line 197 "load_mod.cpp"
      $or$cond52 = $29 & $28; //@line 197 "load_mod.cpp"
      $30 = ($12<<24>>24)==(72); //@line 197 "load_mod.cpp"
      $or$cond53 = $30 & $or$cond52; //@line 197 "load_mod.cpp"
      if (!($or$cond53)) {
       break L22;
      }
      $sext67 = $7 << 24; //@line 197 "load_mod.cpp"
      $31 = $sext67 >> 24; //@line 197 "load_mod.cpp"
      $32 = (($31) + -38)|0; //@line 197 "load_mod.cpp"
      HEAP32[$3>>2] = $32; //@line 197 "load_mod.cpp"
      $dwMemPos$0110 = 20;$dwTotalSampleLen$0109 = 0;$i$0107 = 1;$nErr$0108 = 0;
      break L20;
      break;
     }
     case 50:  {
      $$off66 = (($8) + -48)<<24>>24; //@line 198 "load_mod.cpp"
      $33 = ($$off66&255)<(10); //@line 198 "load_mod.cpp"
      $34 = ($10<<24>>24)==(67); //@line 198 "load_mod.cpp"
      $or$cond54 = $34 & $33; //@line 198 "load_mod.cpp"
      $35 = ($12<<24>>24)==(72); //@line 198 "load_mod.cpp"
      $or$cond55 = $35 & $or$cond54; //@line 198 "load_mod.cpp"
      if (!($or$cond55)) {
       break L22;
      }
      $sext65 = $7 << 24; //@line 198 "load_mod.cpp"
      $36 = $sext65 >> 24; //@line 198 "load_mod.cpp"
      $37 = (($36) + -28)|0; //@line 198 "load_mod.cpp"
      HEAP32[$3>>2] = $37; //@line 198 "load_mod.cpp"
      $dwMemPos$0110 = 20;$dwTotalSampleLen$0109 = 0;$i$0107 = 1;$nErr$0108 = 0;
      break L20;
      break;
     }
     case 51:  {
      $$off64 = (($8) + -48)<<24>>24; //@line 199 "load_mod.cpp"
      $38 = ($$off64&255)<(3); //@line 199 "load_mod.cpp"
      $39 = ($10<<24>>24)==(67); //@line 199 "load_mod.cpp"
      $or$cond56 = $39 & $38; //@line 199 "load_mod.cpp"
      $40 = ($12<<24>>24)==(72); //@line 199 "load_mod.cpp"
      $or$cond57 = $40 & $or$cond56; //@line 199 "load_mod.cpp"
      if (!($or$cond57)) {
       break L22;
      }
      $sext63 = $7 << 24; //@line 199 "load_mod.cpp"
      $41 = $sext63 >> 24; //@line 199 "load_mod.cpp"
      $42 = (($41) + -18)|0; //@line 199 "load_mod.cpp"
      HEAP32[$3>>2] = $42; //@line 199 "load_mod.cpp"
      $dwMemPos$0110 = 20;$dwTotalSampleLen$0109 = 0;$i$0107 = 1;$nErr$0108 = 0;
      break L20;
      break;
     }
     case 84:  {
      $43 = ($8<<24>>24)==(68); //@line 200 "load_mod.cpp"
      $44 = ($10<<24>>24)==(90); //@line 200 "load_mod.cpp"
      $or$cond58 = $43 & $44; //@line 200 "load_mod.cpp"
      $$off62 = (($12) + -52)<<24>>24;
      $45 = ($$off62&255)<(6); //@line 200 "load_mod.cpp"
      $or$cond74 = $or$cond58 & $45; //@line 200 "load_mod.cpp"
      if (!($or$cond74)) {
       break L22;
      }
      $46 = $5 >> 24; //@line 200 "load_mod.cpp"
      $47 = (($46) + -48)|0; //@line 200 "load_mod.cpp"
      HEAP32[$3>>2] = $47; //@line 200 "load_mod.cpp"
      $dwMemPos$0110 = 20;$dwTotalSampleLen$0109 = 0;$i$0107 = 1;$nErr$0108 = 0;
      break L20;
      break;
     }
     default: {
      break L22;
     }
     }
    }
   } while(0);
   switch ($15|0) {
   case 1313027633:  {
    HEAP32[$3>>2] = 16; //@line 201 "load_mod.cpp"
    $dwMemPos$0110 = 20;$dwTotalSampleLen$0109 = 0;$i$0107 = 1;$nErr$0108 = 0;
    break L20;
    break;
   }
   case 1313026611:  {
    HEAP32[$3>>2] = 32; //@line 202 "load_mod.cpp"
    $dwMemPos$0110 = 20;$dwTotalSampleLen$0109 = 0;$i$0107 = 1;$nErr$0108 = 0;
    break L20;
    break;
   }
   default: {
    HEAP32[$2>>2] = 15; //@line 202 "load_mod.cpp"
    $dwMemPos$0110 = 20;$dwTotalSampleLen$0109 = 0;$i$0107 = 1;$nErr$0108 = 0;
    break L20;
   }
   }
  }
 } while(0);
 while(1) {
  $48 = (($lpStream) + ($dwMemPos$0110)|0); //@line 208 "load_mod.cpp"
  $49 = (((($this)) + 59804|0) + ($i$0107<<5)|0); //@line 212 "load_mod.cpp"
  dest=$49; src=$48; stop=dest+22|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 212 "load_mod.cpp"
  $50 = (((((($this)) + 59804|0) + ($i$0107<<5)|0)) + 22|0); //@line 213 "load_mod.cpp"
  HEAP8[$50>>0] = 0; //@line 213 "load_mod.cpp"
  $51 = (((((($this)) + 33280|0) + ($i$0107<<6)|0)) + 34|0); //@line 214 "load_mod.cpp"
  HEAP16[$51>>1] = 0; //@line 214 "load_mod.cpp"
  $$sum12 = (($dwMemPos$0110) + 22)|0; //@line 215 "load_mod.cpp"
  $52 = (($lpStream) + ($$sum12)|0); //@line 215 "load_mod.cpp"
  $53 = HEAPU8[$52>>0]|(HEAPU8[$52+1>>0]<<8); //@line 215 "load_mod.cpp"
  $54 = $53&65535; //@line 215 "load_mod.cpp"
  $55 = $54 >>> 8; //@line 215 "load_mod.cpp"
  $56 = $54 << 8; //@line 215 "load_mod.cpp"
  $57 = $56 & 65280; //@line 215 "load_mod.cpp"
  $58 = $57 | $55; //@line 215 "load_mod.cpp"
  $59 = $58 << 1; //@line 215 "load_mod.cpp"
  $60 = (((($this)) + 33280|0) + ($i$0107<<6)|0); //@line 215 "load_mod.cpp"
  HEAP32[$60>>2] = $59; //@line 215 "load_mod.cpp"
  $61 = (($59) + ($dwTotalSampleLen$0109))|0; //@line 216 "load_mod.cpp"
  $$sum13 = (($dwMemPos$0110) + 24)|0; //@line 217 "load_mod.cpp"
  $62 = (($lpStream) + ($$sum13)|0); //@line 217 "load_mod.cpp"
  $63 = HEAP8[$62>>0]|0; //@line 217 "load_mod.cpp"
  $64 = $63&255; //@line 217 "load_mod.cpp"
  $65 = $64 << 4; //@line 217 "load_mod.cpp"
  $66 = $65&255; //@line 217 "load_mod.cpp"
  $67 = (((((($this)) + 33280|0) + ($i$0107<<6)|0)) + 37|0); //@line 217 "load_mod.cpp"
  HEAP8[$67>>0] = $66; //@line 217 "load_mod.cpp"
  $$sum14 = (($dwMemPos$0110) + 25)|0; //@line 218 "load_mod.cpp"
  $68 = (($lpStream) + ($$sum14)|0); //@line 218 "load_mod.cpp"
  $69 = HEAP8[$68>>0]|0; //@line 218 "load_mod.cpp"
  $70 = $69&255; //@line 218 "load_mod.cpp"
  $71 = $70 << 2; //@line 218 "load_mod.cpp"
  $72 = $71&65535; //@line 218 "load_mod.cpp"
  $73 = (((((($this)) + 33280|0) + ($i$0107<<6)|0)) + 30|0); //@line 218 "load_mod.cpp"
  HEAP16[$73>>1] = $72; //@line 218 "load_mod.cpp"
  $74 = ($71>>>0)>(256); //@line 219 "load_mod.cpp"
  if ($74) {
   HEAP16[$73>>1] = 256; //@line 219 "load_mod.cpp"
   $75 = (($nErr$0108) + 1)|0; //@line 219 "load_mod.cpp"
   $nErr$1 = $75;
  } else {
   $nErr$1 = $nErr$0108;
  }
  $76 = (((((($this)) + 33280|0) + ($i$0107<<6)|0)) + 32|0); //@line 220 "load_mod.cpp"
  HEAP16[$76>>1] = 64; //@line 220 "load_mod.cpp"
  $77 = (((((($this)) + 33280|0) + ($i$0107<<6)|0)) + 28|0); //@line 221 "load_mod.cpp"
  HEAP16[$77>>1] = 128; //@line 221 "load_mod.cpp"
  $$sum15 = (($dwMemPos$0110) + 26)|0; //@line 222 "load_mod.cpp"
  $78 = (($lpStream) + ($$sum15)|0); //@line 222 "load_mod.cpp"
  $79 = HEAPU8[$78>>0]|(HEAPU8[$78+1>>0]<<8); //@line 222 "load_mod.cpp"
  $80 = $79&65535; //@line 222 "load_mod.cpp"
  $81 = $80 >>> 8; //@line 222 "load_mod.cpp"
  $82 = $80 << 8; //@line 222 "load_mod.cpp"
  $83 = $82 & 65280; //@line 222 "load_mod.cpp"
  $84 = $83 | $81; //@line 222 "load_mod.cpp"
  $85 = $84 << 1; //@line 222 "load_mod.cpp"
  $$sum16 = (($dwMemPos$0110) + 28)|0; //@line 223 "load_mod.cpp"
  $86 = (($lpStream) + ($$sum16)|0); //@line 223 "load_mod.cpp"
  $87 = HEAPU8[$86>>0]|(HEAPU8[$86+1>>0]<<8); //@line 223 "load_mod.cpp"
  $88 = $87&65535; //@line 223 "load_mod.cpp"
  $89 = $88 >>> 8; //@line 223 "load_mod.cpp"
  $90 = $88 << 8; //@line 223 "load_mod.cpp"
  $91 = $90 & 65280; //@line 223 "load_mod.cpp"
  $92 = $91 | $89; //@line 223 "load_mod.cpp"
  $93 = $92 << 1; //@line 223 "load_mod.cpp"
  $94 = ($93>>>0)>(2); //@line 225 "load_mod.cpp"
  if ($94) {
   $95 = (($93) + ($85))|0; //@line 225 "load_mod.cpp"
   $96 = ($95>>>0)<=($59>>>0); //@line 225 "load_mod.cpp"
   $97 = (($93) + ($84))|0;
   $98 = ($97>>>0)>($59>>>0); //@line 226 "load_mod.cpp"
   $or$cond25 = $96 | $98; //@line 226 "load_mod.cpp"
   $$37 = $or$cond25 ? $85 : $84; //@line 226 "load_mod.cpp"
   $100 = $$37;
  } else {
   $100 = $85;
  }
  $99 = (((((($this)) + 33280|0) + ($i$0107<<6)|0)) + 4|0); //@line 230 "load_mod.cpp"
  HEAP32[$99>>2] = $100; //@line 230 "load_mod.cpp"
  $101 = (($100) + ($93))|0; //@line 231 "load_mod.cpp"
  $102 = (((((($this)) + 33280|0) + ($i$0107<<6)|0)) + 8|0); //@line 231 "load_mod.cpp"
  HEAP32[$102>>2] = $101; //@line 231 "load_mod.cpp"
  $103 = ($59>>>0)<(4); //@line 232 "load_mod.cpp"
  do {
   if ($103) {
    HEAP32[$60>>2] = 0; //@line 232 "load_mod.cpp"
   } else {
    $104 = ($100>>>0)<($59>>>0); //@line 236 "load_mod.cpp"
    if ($104) {
     $109 = $100;
    } else {
     $105 = (($59) + -1)|0; //@line 236 "load_mod.cpp"
     HEAP32[$99>>2] = $105; //@line 236 "load_mod.cpp"
     $109 = $105;
    }
    $106 = ($101>>>0)>($59>>>0); //@line 237 "load_mod.cpp"
    if ($106) {
     HEAP32[$102>>2] = $59; //@line 237 "load_mod.cpp"
     $108 = $59;
    } else {
     $108 = $101;
    }
    $107 = ($108>>>0)<($109>>>0); //@line 239 "load_mod.cpp"
    $110 = ($108>>>0)<(9); //@line 239 "load_mod.cpp"
    $or$cond26 = $107 | $110; //@line 239 "load_mod.cpp"
    $111 = (($108) - ($109))|0;
    $112 = ($111>>>0)<(5); //@line 240 "load_mod.cpp"
    $or$cond28 = $or$cond26 | $112; //@line 239 "load_mod.cpp"
    if ($or$cond28) {
     HEAP32[$99>>2] = 0; //@line 242 "load_mod.cpp"
     HEAP32[$102>>2] = 0; //@line 243 "load_mod.cpp"
     break;
    }
    $113 = ($108>>>0)>($109>>>0); //@line 245 "load_mod.cpp"
    if ($113) {
     HEAP16[$51>>1] = 2; //@line 247 "load_mod.cpp"
    }
   }
  } while(0);
  $114 = (($dwMemPos$0110) + 30)|0; //@line 250 "load_mod.cpp"
  $115 = (($i$0107) + 1)|0; //@line 206 "load_mod.cpp"
  $116 = HEAP32[$2>>2]|0; //@line 206 "load_mod.cpp"
  $117 = ($115>>>0)>($116>>>0); //@line 206 "load_mod.cpp"
  if ($117) {
   $$lcssa135 = $61;$$lcssa136 = $114;$$lcssa137 = $116;$dwMemPos$0110$lcssa = $dwMemPos$0110;$nErr$1$lcssa = $nErr$1;
   break;
  } else {
   $dwMemPos$0110 = $114;$dwTotalSampleLen$0109 = $61;$i$0107 = $115;$nErr$0108 = $nErr$1;
  }
 }
 $118 = ($$lcssa137|0)==(15); //@line 252 "load_mod.cpp"
 $119 = $dwMemLength << 2;
 $120 = ($$lcssa135>>>0)>($119>>>0); //@line 252 "load_mod.cpp"
 $or$cond30 = $120 & $118; //@line 252 "load_mod.cpp"
 if ($or$cond30) {
  $$0 = 0;
  return ($$0|0); //@line 363 "load_mod.cpp"
 }
 $121 = (($lpStream) + ($$lcssa136)|0); //@line 253 "load_mod.cpp"
 $122 = (($dwMemPos$0110$lcssa) + 164)|0; //@line 254 "load_mod.cpp"
 $123 = (($dwMemPos$0110$lcssa) + 160)|0; //@line 255 "load_mod.cpp"
 $$ = $118 ? $123 : $122; //@line 255 "load_mod.cpp"
 $124 = ((($this)) + 53344|0); //@line 256 "load_mod.cpp"
 _memset(($124|0),0,256)|0; //@line 256 "load_mod.cpp"
 $$sum = (($dwMemPos$0110$lcssa) + 32)|0; //@line 257 "load_mod.cpp"
 $125 = (($lpStream) + ($$sum)|0); //@line 257 "load_mod.cpp"
 dest=$124; src=$125; stop=dest+128|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 257 "load_mod.cpp"
 $126 = HEAP8[$121>>0]|0; //@line 261 "load_mod.cpp"
 $127 = $126&255; //@line 261 "load_mod.cpp"
 $128 = (($126) + -1)<<24>>24; //@line 262 "load_mod.cpp"
 $129 = ($128<<24>>24)<(0); //@line 262 "load_mod.cpp"
 L70: do {
  if ($129) {
   $norders$0 = 128;
   while(1) {
    $130 = ($norders$0>>>0)>(1); //@line 265 "load_mod.cpp"
    if (!($130)) {
     $norders$1 = $norders$0;
     break L70;
    }
    $131 = (($norders$0) + -1)|0; //@line 265 "load_mod.cpp"
    $132 = (((($this)) + 53344|0) + ($131)|0); //@line 265 "load_mod.cpp"
    $133 = HEAP8[$132>>0]|0; //@line 265 "load_mod.cpp"
    $134 = ($133<<24>>24)==(0); //@line 265 "load_mod.cpp"
    if ($134) {
     $norders$0 = $131;
    } else {
     $norders$1 = $norders$0;
     break;
    }
   }
  } else {
   $norders$1 = $127;
  }
 } while(0);
 $iord$0106 = 0;$nbp$0103 = 0;$nbpbuggy$0104 = 0;$nbpbuggy2$0105 = 0;
 while(1) {
  $137 = (((($this)) + 53344|0) + ($iord$0106)|0); //@line 272 "load_mod.cpp"
  $138 = HEAP8[$137>>0]|0; //@line 272 "load_mod.cpp"
  $139 = $138&255; //@line 272 "load_mod.cpp"
  $140 = ($138<<24>>24)<(0); //@line 273 "load_mod.cpp"
  $141 = ($nbp$0103>>>0)>($139>>>0); //@line 273 "load_mod.cpp"
  $or$cond31 = $140 | $141; //@line 273 "load_mod.cpp"
  $$pre = (($139) + 1)|0; //@line 278 "load_mod.cpp"
  $142 = ($iord$0106>>>0)<($norders$1>>>0); //@line 276 "load_mod.cpp"
  $$nbpbuggy$0 = $142 ? $$pre : $nbpbuggy$0104; //@line 276 "load_mod.cpp"
  $nbp$1 = $or$cond31 ? $nbp$0103 : $$pre; //@line 273 "load_mod.cpp"
  $nbpbuggy$1 = $or$cond31 ? $nbpbuggy$0104 : $$nbpbuggy$0; //@line 273 "load_mod.cpp"
  $143 = ($139>>>0)<($nbpbuggy2$0105>>>0); //@line 278 "load_mod.cpp"
  $nbpbuggy2$0$ = $143 ? $nbpbuggy2$0105 : $$pre; //@line 278 "load_mod.cpp"
  $144 = (($iord$0106) + 1)|0; //@line 270 "load_mod.cpp"
  $exitcond = ($144|0)==(128); //@line 270 "load_mod.cpp"
  if ($exitcond) {
   $nbp$1$lcssa = $nbp$1;$nbpbuggy$1$lcssa = $nbpbuggy$1;$nbpbuggy2$0$$lcssa = $nbpbuggy2$0$;
   break;
  } else {
   $iord$0106 = $144;$nbp$0103 = $nbp$1;$nbpbuggy$0104 = $nbpbuggy$1;$nbpbuggy2$0105 = $nbpbuggy2$0$;
  }
 }
 $135 = ($norders$1>>>0)<(256); //@line 280 "load_mod.cpp"
 if ($135) {
  $scevgep = ((($this)) + 53344|0);
  $uglygep = (($scevgep) + ($norders$1)|0);
  $136 = (256 - ($norders$1))|0; //@line 280 "load_mod.cpp"
  _memset(($uglygep|0),-1,($136|0))|0; //@line 280 "load_mod.cpp"
 }
 $145 = (($norders$1) + -1)|0; //@line 281 "load_mod.cpp"
 $$sum89 = $$lcssa136 | 1; //@line 282 "load_mod.cpp"
 $146 = (($lpStream) + ($$sum89)|0); //@line 282 "load_mod.cpp"
 $147 = HEAP8[$146>>0]|0; //@line 282 "load_mod.cpp"
 $148 = $147&255; //@line 282 "load_mod.cpp"
 $149 = ((($this)) + 59736|0); //@line 282 "load_mod.cpp"
 $150 = ($147&255)>(119); //@line 283 "load_mod.cpp"
 $$32 = $150 ? 0 : $148; //@line 283 "load_mod.cpp"
 $151 = (($$32) + 1)|0; //@line 284 "load_mod.cpp"
 $152 = ($151>>>0)<($145>>>0); //@line 284 "load_mod.cpp"
 $storemerge17 = $152 ? $$32 : 0; //@line 284 "load_mod.cpp"
 HEAP32[$149>>2] = $storemerge17; //@line 282 "load_mod.cpp"
 $153 = ($nbp$1$lcssa|0)==(0); //@line 285 "load_mod.cpp"
 if ($153) {
  $$0 = 0;
  return ($$0|0); //@line 363 "load_mod.cpp"
 }
 $154 = (($$) + ($$lcssa135))|0; //@line 286 "load_mod.cpp"
 $155 = (($lpStream) + ($123)|0); //@line 287 "load_mod.cpp"
 $156 = HEAP32[$155>>2]|0; //@line 173 "load_mod.cpp"
 $157 = ($156|0)==(776678989); //@line 173 "load_mod.cpp"
 if ($157) {
  $158 = $nbp$1$lcssa << 11; //@line 287 "load_mod.cpp"
  $159 = (($158) + ($154))|0; //@line 287 "load_mod.cpp"
  $160 = ($159|0)==($dwMemLength|0); //@line 287 "load_mod.cpp"
  if ($160) {
   HEAP32[$3>>2] = 8; //@line 287 "load_mod.cpp"
  }
 }
 $161 = ($nbp$1$lcssa|0)==($nbpbuggy$1$lcssa|0); //@line 288 "load_mod.cpp"
 if ($161) {
  label = 53;
 } else {
  $162 = HEAP32[$3>>2]|0; //@line 288 "load_mod.cpp"
  $163 = $nbp$1$lcssa << 8; //@line 288 "load_mod.cpp"
  $164 = Math_imul($163, $162)|0; //@line 288 "load_mod.cpp"
  $165 = (($164) + ($154))|0; //@line 288 "load_mod.cpp"
  $166 = ($165|0)==($dwMemLength|0); //@line 288 "load_mod.cpp"
  if ($166) {
   label = 53;
  } else {
   $167 = $nbpbuggy$1$lcssa << 8; //@line 290 "load_mod.cpp"
   $168 = Math_imul($167, $162)|0; //@line 290 "load_mod.cpp"
   $169 = (($168) + ($154))|0; //@line 290 "load_mod.cpp"
   $170 = ($169|0)==($dwMemLength|0); //@line 290 "load_mod.cpp"
   $171 = (($nErr$1$lcssa) + 8)|0; //@line 291 "load_mod.cpp"
   $nbpbuggy$0$nbp$0 = $170 ? $nbpbuggy$1$lcssa : $nbp$1$lcssa; //@line 290 "load_mod.cpp"
   $nErr$0$ = $170 ? $nErr$1$lcssa : $171; //@line 290 "load_mod.cpp"
   $nErr$2 = $nErr$0$;$nbp$2 = $nbpbuggy$0$nbp$0;
  }
 }
 if ((label|0) == 53) {
  $172 = ($nbpbuggy2$0$$lcssa>>>0)>($nbp$1$lcssa>>>0); //@line 293 "load_mod.cpp"
  if ($172) {
   $173 = HEAP32[$3>>2]|0; //@line 293 "load_mod.cpp"
   $174 = $nbpbuggy2$0$$lcssa << 8; //@line 293 "load_mod.cpp"
   $175 = Math_imul($174, $173)|0; //@line 293 "load_mod.cpp"
   $176 = (($175) + ($154))|0; //@line 293 "load_mod.cpp"
   $177 = ($176|0)==($dwMemLength|0); //@line 293 "load_mod.cpp"
   $nbpbuggy2$0$nbp$0 = $177 ? $nbpbuggy2$0$$lcssa : $nbp$1$lcssa; //@line 293 "load_mod.cpp"
   $nErr$2 = $nErr$1$lcssa;$nbp$2 = $nbpbuggy2$0$nbp$0;
  } else {
   $nErr$2 = $nErr$1$lcssa;$nbp$2 = $nbp$1$lcssa;
  }
 }
 $178 = ($154>>>0)<(1536); //@line 297 "load_mod.cpp"
 $179 = ($154>>>0)>($dwMemLength>>>0); //@line 297 "load_mod.cpp"
 $or$cond33 = $178 | $179; //@line 297 "load_mod.cpp"
 $180 = (($nErr$2) + 8)|0; //@line 297 "load_mod.cpp"
 $nErr$3 = $or$cond33 ? $180 : $nErr$2; //@line 297 "load_mod.cpp"
 $181 = HEAP32[$2>>2]|0; //@line 298 "load_mod.cpp"
 $182 = ($181|0)==(15); //@line 298 "load_mod.cpp"
 $183 = ($nErr$3>>>0)>(15); //@line 298 "load_mod.cpp"
 $or$cond5 = $182 & $183; //@line 298 "load_mod.cpp"
 if ($or$cond5) {
  $$0 = 0;
  return ($$0|0); //@line 363 "load_mod.cpp"
 }
 $184 = ((($this)) + 59680|0); //@line 300 "load_mod.cpp"
 HEAP32[$184>>2] = 1; //@line 300 "load_mod.cpp"
 $185 = ((($this)) + 59648|0); //@line 301 "load_mod.cpp"
 HEAP32[$185>>2] = 6; //@line 301 "load_mod.cpp"
 $186 = ((($this)) + 59652|0); //@line 302 "load_mod.cpp"
 HEAP32[$186>>2] = 125; //@line 302 "load_mod.cpp"
 $187 = ((($this)) + 59764|0); //@line 303 "load_mod.cpp"
 HEAP32[$187>>2] = 56; //@line 303 "load_mod.cpp"
 $188 = ((($this)) + 59768|0); //@line 304 "load_mod.cpp"
 HEAP32[$188>>2] = 13696; //@line 304 "load_mod.cpp"
 $189 = ((($this)) + 59804|0); //@line 305 "load_mod.cpp"
 dest=$189; src=$lpStream; stop=dest+20|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 305 "load_mod.cpp"
 $190 = HEAP32[$3>>2]|0; //@line 307 "load_mod.cpp"
 $191 = ($190|0)==(0); //@line 307 "load_mod.cpp"
 if ($191) {
  $275 = 0;
 } else {
  $ich$097 = 0;
  while(1) {
   $193 = (((((($this)) + 49600|0) + (($ich$097*36)|0)|0)) + 4|0); //@line 309 "load_mod.cpp"
   HEAP32[$193>>2] = 64; //@line 309 "load_mod.cpp"
   $194 = HEAP32[183080]|0; //@line 310 "load_mod.cpp"
   $195 = $194 & 524288; //@line 310 "load_mod.cpp"
   $196 = ($195|0)==(0); //@line 310 "load_mod.cpp"
   $197 = $ich$097 & 3; //@line 311 "load_mod.cpp"
   $$off = (($197) + -1)|0; //@line 313 "load_mod.cpp"
   $198 = ($$off>>>0)<(2); //@line 313 "load_mod.cpp"
   $199 = $198 ? 256 : 0; //@line 311 "load_mod.cpp"
   $200 = $198 ? 192 : 64; //@line 313 "load_mod.cpp"
   $$sink = $196 ? $200 : $199; //@line 310 "load_mod.cpp"
   $201 = (((($this)) + 49600|0) + (($ich$097*36)|0)|0); //@line 311 "load_mod.cpp"
   HEAP32[$201>>2] = $$sink; //@line 311 "load_mod.cpp"
   $202 = (($ich$097) + 1)|0; //@line 307 "load_mod.cpp"
   $203 = HEAP32[$3>>2]|0; //@line 307 "load_mod.cpp"
   $204 = ($202>>>0)<($203>>>0); //@line 307 "load_mod.cpp"
   if ($204) {
    $ich$097 = $202;
   } else {
    $275 = $203;
    break;
   }
  }
 }
 $192 = ($nbp$2|0)==(0); //@line 316 "load_mod.cpp"
 L102: do {
  if ($192) {
   $dwMemPos$2$lcssa = $$;
  } else {
   $208 = $275;$dwMemPos$290 = $$;$ipat$089 = 0;
   while(1) {
    $207 = ($ipat$089>>>0)<(240); //@line 318 "load_mod.cpp"
    if ($207) {
     $209 = (__ZN10CSoundFile15AllocatePatternEjj(64,$208)|0); //@line 320 "load_mod.cpp"
     $210 = (((($this)) + 51904|0) + ($ipat$089<<2)|0); //@line 320 "load_mod.cpp"
     HEAP32[$210>>2] = $209; //@line 320 "load_mod.cpp"
     $211 = ($209|0)==(0|0); //@line 320 "load_mod.cpp"
     if ($211) {
      $dwMemPos$2$lcssa = $dwMemPos$290;
      break L102;
     }
     $212 = (((($this)) + 52864|0) + ($ipat$089<<1)|0); //@line 321 "load_mod.cpp"
     HEAP16[$212>>1] = 64; //@line 321 "load_mod.cpp"
     $213 = HEAP32[$3>>2]|0; //@line 322 "load_mod.cpp"
     $214 = $213 << 8; //@line 322 "load_mod.cpp"
     $215 = (($214) + ($dwMemPos$290))|0; //@line 322 "load_mod.cpp"
     $216 = ($215>>>0)<($dwMemLength>>>0); //@line 322 "load_mod.cpp"
     if (!($216)) {
      $dwMemPos$2$lcssa = $dwMemPos$290;
      break L102;
     }
     $217 = $213 << 6; //@line 325 "load_mod.cpp"
     $218 = ($217|0)==(0); //@line 325 "load_mod.cpp"
     if (!($218)) {
      $219 = (($lpStream) + ($dwMemPos$290)|0); //@line 324 "load_mod.cpp"
      $220 = HEAP32[$210>>2]|0; //@line 323 "load_mod.cpp"
      $j$086 = $217;$m$084 = $220;$p$085 = $219;
      while(1) {
       $221 = HEAP8[$p$085>>0]|0; //@line 327 "load_mod.cpp"
       $222 = ((($p$085)) + 1|0); //@line 327 "load_mod.cpp"
       $223 = HEAP8[$222>>0]|0; //@line 327 "load_mod.cpp"
       $224 = ((($p$085)) + 2|0); //@line 327 "load_mod.cpp"
       $225 = HEAP8[$224>>0]|0; //@line 327 "load_mod.cpp"
       $226 = ((($p$085)) + 3|0); //@line 327 "load_mod.cpp"
       $227 = HEAP8[$226>>0]|0; //@line 327 "load_mod.cpp"
       $228 = $221&255; //@line 328 "load_mod.cpp"
       $229 = $228 << 8; //@line 328 "load_mod.cpp"
       $230 = $229 & 3840; //@line 328 "load_mod.cpp"
       $231 = $223&255; //@line 328 "load_mod.cpp"
       $232 = $230 | $231; //@line 328 "load_mod.cpp"
       $switch$split12D = ($232|0)<(4095);
       if ($switch$split12D) {
        switch ($232|0) {
        case 0:  {
         break;
        }
        default: {
         label = 66;
        }
        }
       } else {
        switch ($232|0) {
        case 4095:  {
         break;
        }
        default: {
         label = 66;
        }
        }
       }
       if ((label|0) == 66) {
        label = 0;
        $233 = $232 << 2; //@line 329 "load_mod.cpp"
        $234 = (__ZNK10CSoundFile17GetNoteFromPeriodEj($this,$233)|0); //@line 329 "load_mod.cpp"
        $235 = $234&255; //@line 329 "load_mod.cpp"
        HEAP8[$m$084>>0] = $235; //@line 329 "load_mod.cpp"
       }
       $236 = $225&255; //@line 330 "load_mod.cpp"
       $237 = $236 >>> 4; //@line 330 "load_mod.cpp"
       $238 = $228 & 16; //@line 330 "load_mod.cpp"
       $239 = $237 | $238; //@line 330 "load_mod.cpp"
       $240 = $239&255; //@line 330 "load_mod.cpp"
       $241 = ((($m$084)) + 1|0); //@line 330 "load_mod.cpp"
       HEAP8[$241>>0] = $240; //@line 330 "load_mod.cpp"
       $242 = $236 & 15; //@line 331 "load_mod.cpp"
       $243 = $242&255; //@line 331 "load_mod.cpp"
       $244 = ((($m$084)) + 3|0); //@line 331 "load_mod.cpp"
       HEAP8[$244>>0] = $243; //@line 331 "load_mod.cpp"
       $245 = ((($m$084)) + 5|0); //@line 332 "load_mod.cpp"
       HEAP8[$245>>0] = $227; //@line 332 "load_mod.cpp"
       $246 = $243 | $227; //@line 333 "load_mod.cpp"
       $247 = ($246<<24>>24)==(0); //@line 333 "load_mod.cpp"
       if (!($247)) {
        __ZNK10CSoundFile17ConvertModCommandEP11_MODCOMMAND($this,$m$084); //@line 333 "load_mod.cpp"
       }
       $248 = ((($m$084)) + 6|0); //@line 325 "load_mod.cpp"
       $249 = ((($p$085)) + 4|0); //@line 325 "load_mod.cpp"
       $250 = (($j$086) + -1)|0; //@line 325 "load_mod.cpp"
       $251 = ($250|0)==(0); //@line 325 "load_mod.cpp"
       if ($251) {
        break;
       } else {
        $j$086 = $250;$m$084 = $248;$p$085 = $249;
       }
      }
     }
    }
    $252 = HEAP32[$3>>2]|0; //@line 336 "load_mod.cpp"
    $253 = $252 << 8; //@line 336 "load_mod.cpp"
    $254 = (($253) + ($dwMemPos$290))|0; //@line 336 "load_mod.cpp"
    $255 = (($ipat$089) + 1)|0; //@line 316 "load_mod.cpp"
    $256 = ($255>>>0)<($nbp$2>>>0); //@line 316 "load_mod.cpp"
    if ($256) {
     $208 = $252;$dwMemPos$290 = $254;$ipat$089 = $255;
    } else {
     $dwMemPos$2$lcssa = $254;
     break;
    }
   }
  }
 } while(0);
 $205 = HEAP32[$2>>2]|0; //@line 340 "load_mod.cpp"
 $206 = ($205|0)==(0); //@line 340 "load_mod.cpp"
 L123: do {
  if ($206) {
   $dwErrCheck$0$lcssa = 0;
  } else {
   $dwErrCheck$080 = 0;$dwMemPos$379 = $dwMemPos$2$lcssa;$ismp$081 = 1;
   while(1) {
    $257 = (((($this)) + 33280|0) + ($ismp$081<<6)|0); //@line 340 "load_mod.cpp"
    $258 = HEAP32[$257>>2]|0; //@line 340 "load_mod.cpp"
    $259 = ($258|0)==(0); //@line 340 "load_mod.cpp"
    if ($259) {
     $dwErrCheck$1 = $dwErrCheck$080;$dwMemPos$5 = $dwMemPos$379;
    } else {
     $260 = (($lpStream) + ($dwMemPos$379)|0); //@line 342 "load_mod.cpp"
     $261 = (($dwMemPos$379) + 5)|0; //@line 344 "load_mod.cpp"
     $262 = ($261>>>0)<($dwMemLength>>>0); //@line 344 "load_mod.cpp"
     if (!($262)) {
      $dwErrCheck$0$lcssa = $dwErrCheck$080;
      break L123;
     }
     $263 = (_strncasecmp($260,529892,5)|0); //@line 345 "load_mod.cpp"
     $264 = ($263|0)==(0); //@line 345 "load_mod.cpp"
     $265 = (($lpStream) + ($261)|0); //@line 348 "load_mod.cpp"
     $dwMemPos$4 = $264 ? $261 : $dwMemPos$379; //@line 345 "load_mod.cpp"
     $p4$0 = $264 ? $265 : $260; //@line 345 "load_mod.cpp"
     $flags$0 = $264 ? 3 : 0; //@line 345 "load_mod.cpp"
     $266 = (($dwMemLength) - ($dwMemPos$4))|0; //@line 351 "load_mod.cpp"
     $267 = (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$257,$flags$0,$p4$0,$266)|0); //@line 351 "load_mod.cpp"
     $268 = ($267|0)==(0); //@line 352 "load_mod.cpp"
     if ($268) {
      $dwErrCheck$1 = $dwErrCheck$080;$dwMemPos$5 = $dwMemPos$4;
     } else {
      $269 = (($dwMemPos$4) + ($267))|0; //@line 354 "load_mod.cpp"
      $270 = (($dwErrCheck$080) + 1)|0; //@line 355 "load_mod.cpp"
      $dwErrCheck$1 = $270;$dwMemPos$5 = $269;
     }
    }
    $271 = (($ismp$081) + 1)|0; //@line 340 "load_mod.cpp"
    $272 = HEAP32[$2>>2]|0; //@line 340 "load_mod.cpp"
    $273 = ($271>>>0)>($272>>>0); //@line 340 "load_mod.cpp"
    if ($273) {
     $dwErrCheck$0$lcssa = $dwErrCheck$1;
     break;
    } else {
     $dwErrCheck$080 = $dwErrCheck$1;$dwMemPos$379 = $dwMemPos$5;$ismp$081 = $271;
    }
   }
  }
 } while(0);
 $274 = ($dwErrCheck$0$lcssa|0)!=(0); //@line 361 "load_mod.cpp"
 $$0 = $274;
 return ($$0|0); //@line 363 "load_mod.cpp"
}
function __ZN10CSoundFile7ReadAMFEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$0 = 0, $$1 = 0, $$arith = 0, $$lcssa = 0, $$lcssa172 = 0, $$lcssa173 = 0, $$lcssa175 = 0, $$lcssa177 = 0, $$lcssa179 = 0, $$lcssa181 = 0, $$lcssa184 = 0, $$lcssa186 = 0, $$maxsampleseekpos$0 = 0, $$off = 0, $$off25 = 0, $$overflow = 0, $$pre = 0, $$pre114 = 0, $$pre115 = 0;
 var $$realtrackcnt$0 = 0, $$sum = 0, $$sum10 = 0, $$sum11 = 0, $$sum12 = 0, $$sum13 = 0, $$sum14 = 0, $$sum15 = 0, $$sum16 = 0, $$sum17 = 0, $$sum18 = 0, $$sum19 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0;
 var $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0;
 var $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0;
 var $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0;
 var $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0;
 var $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0;
 var $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0;
 var $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0;
 var $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0;
 var $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0;
 var $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0;
 var $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0;
 var $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0;
 var $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0;
 var $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dwMemPos$054 = 0, $dwMemPos$10 = 0, $dwMemPos$10$lcssa = 0, $dwMemPos$11$lcssa = 0, $dwMemPos$1166 = 0, $dwMemPos$12 = 0;
 var $dwMemPos$1357 = 0, $dwMemPos$14 = 0, $dwMemPos$150 = 0, $dwMemPos$2$ph = 0, $dwMemPos$4 = 0, $dwMemPos$690 = 0, $dwMemPos$7 = 0, $dwMemPos$8 = 0, $dwMemPos$8$lcssa = 0, $dwMemPos$9$lcssa = 0, $dwMemPos$9$ph86 = 0, $dwMemPos$977 = 0, $exitcond = 0, $exitcond108 = 0, $exitcond109 = 0, $exitcond110 = 0, $exitcond111 = 0, $i$046 = 0, $i2$091 = 0, $iChn$059 = 0;
 var $iData$0 = 0, $iData$0$ph = 0, $iIns$0$ph85 = 0, $iIns$076 = 0, $iOrd$055 = 0, $iOrd4$089 = 0, $iPat$049 = 0, $iPat5$061 = 0, $iSeek$058 = 0, $iSmp$053 = 0, $iSmp7$0 = 0, $iTrack$065 = 0, $iTrkMap$071 = 0, $maxsampleseekpos$0$ph$lcssa = 0, $maxsampleseekpos$0$ph84 = 0, $or$cond = 0, $or$cond134 = 0, $or$cond135 = 0, $or$cond24 = 0, $or$cond2456 = 0;
 var $or$cond3 = 0, $or$cond5$not = 0, $or$cond7 = 0, $or$cond9 = 0, $p$048 = 0, $pan$0 = 0, $pin$047 = 0, $ptracks = 0, $realtrackcnt$0$lcssa = 0, $realtrackcnt$070 = 0, $sampleseekpos = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1920|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ptracks = sp + 960|0;
 $sampleseekpos = sp;
 $0 = ($lpStream|0)==(0|0); //@line 172 "load_amf.cpp"
 $1 = ($dwMemLength>>>0)<(2048); //@line 172 "load_amf.cpp"
 $or$cond = $0 | $1; //@line 172 "load_amf.cpp"
 if ($or$cond) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0); //@line 421 "load_amf.cpp"
 }
 $2 = (_strncmp($lpStream,529898,25)|0); //@line 173 "load_amf.cpp"
 $3 = ($2|0)==(0); //@line 173 "load_amf.cpp"
 $4 = ($dwMemLength>>>0)>(4096); //@line 173 "load_amf.cpp"
 $or$cond3 = $4 & $3; //@line 173 "load_amf.cpp"
 if ($or$cond3) {
  $5 = ((($lpStream)) + 35|0); //@line 178 "load_amf.cpp"
  $6 = HEAP8[$5>>0]|0; //@line 178 "load_amf.cpp"
  $7 = $6&255; //@line 178 "load_amf.cpp"
  $8 = ((($lpStream)) + 36|0); //@line 179 "load_amf.cpp"
  $9 = HEAP8[$8>>0]|0; //@line 179 "load_amf.cpp"
  $10 = $9&255; //@line 179 "load_amf.cpp"
  $11 = (($6) + -1)<<24>>24; //@line 182 "load_amf.cpp"
  $or$cond5$not = ($11&255)<(240); //@line 182 "load_amf.cpp"
  $12 = ($9<<24>>24)!=(0); //@line 182 "load_amf.cpp"
  $or$cond7 = $or$cond5$not & $12; //@line 182 "load_amf.cpp"
  if (!($or$cond7)) {
   $$1 = 0;
   STACKTOP = sp;return ($$1|0); //@line 421 "load_amf.cpp"
  }
  $13 = $7 << 11; //@line 183 "load_amf.cpp"
  $14 = $13 | 294; //@line 183 "load_amf.cpp"
  $15 = (($14) + 2368)|0; //@line 183 "load_amf.cpp"
  $16 = ($15>>>0)<($dwMemLength>>>0); //@line 183 "load_amf.cpp"
  if (!($16)) {
   $$1 = 0;
   STACKTOP = sp;return ($$1|0); //@line 421 "load_amf.cpp"
  }
  $17 = ((($this)) + 59680|0); //@line 184 "load_amf.cpp"
  HEAP32[$17>>2] = 2097152; //@line 184 "load_amf.cpp"
  $18 = ((($this)) + 59664|0); //@line 185 "load_amf.cpp"
  HEAP32[$18>>2] = 8; //@line 185 "load_amf.cpp"
  $19 = ((($this)) + 59688|0); //@line 186 "load_amf.cpp"
  HEAP32[$19>>2] = 0; //@line 186 "load_amf.cpp"
  $20 = ((($this)) + 59684|0); //@line 187 "load_amf.cpp"
  HEAP32[$20>>2] = 31; //@line 187 "load_amf.cpp"
  $21 = ((($this)) + 59652|0); //@line 188 "load_amf.cpp"
  HEAP32[$21>>2] = 125; //@line 188 "load_amf.cpp"
  $22 = ((($this)) + 59648|0); //@line 189 "load_amf.cpp"
  HEAP32[$22>>2] = 6; //@line 189 "load_amf.cpp"
  $iOrd$055 = 0;
  while(1) {
   $23 = ($iOrd$055>>>0)<($10>>>0); //@line 192 "load_amf.cpp"
   if ($23) {
    $24 = (($iOrd$055) + 38)|0; //@line 192 "load_amf.cpp"
    $25 = (($lpStream) + ($24)|0); //@line 192 "load_amf.cpp"
    $26 = HEAP8[$25>>0]|0; //@line 192 "load_amf.cpp"
    $27 = $26&255; //@line 192 "load_amf.cpp"
    $29 = $27;
   } else {
    $29 = 255;
   }
   $28 = $29&255; //@line 192 "load_amf.cpp"
   $30 = (((($this)) + 53344|0) + ($iOrd$055)|0); //@line 192 "load_amf.cpp"
   HEAP8[$30>>0] = $28; //@line 192 "load_amf.cpp"
   $31 = (($iOrd$055) + 1)|0; //@line 190 "load_amf.cpp"
   $exitcond109 = ($31|0)==(256); //@line 190 "load_amf.cpp"
   if ($exitcond109) {
    $dwMemPos$054 = 294;$iSmp$053 = 0;
    break;
   } else {
    $iOrd$055 = $31;
   }
  }
  while(1) {
   $33 = (($iSmp$053) + 1)|0; //@line 197 "load_amf.cpp"
   $34 = (((($this)) + 59804|0) + ($33<<5)|0); //@line 198 "load_amf.cpp"
   $35 = (($lpStream) + ($dwMemPos$054)|0); //@line 198 "load_amf.cpp"
   dest=$34; src=$35; stop=dest+22|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 198 "load_amf.cpp"
   $36 = (((((($this)) + 59804|0) + ($33<<5)|0)) + 21|0); //@line 199 "load_amf.cpp"
   HEAP8[$36>>0] = 0; //@line 199 "load_amf.cpp"
   $37 = (($dwMemPos$054) + 22)|0; //@line 200 "load_amf.cpp"
   $38 = (($lpStream) + ($37)|0); //@line 200 "load_amf.cpp"
   $39 = HEAP8[$38>>0]|0; //@line 200 "load_amf.cpp"
   $40 = $39&255; //@line 200 "load_amf.cpp"
   $41 = $40 << 4; //@line 200 "load_amf.cpp"
   $42 = $41&255; //@line 200 "load_amf.cpp"
   $43 = (((((($this)) + 33280|0) + ($33<<6)|0)) + 37|0); //@line 200 "load_amf.cpp"
   HEAP8[$43>>0] = $42; //@line 200 "load_amf.cpp"
   $44 = (($dwMemPos$054) + 23)|0; //@line 201 "load_amf.cpp"
   $45 = (($lpStream) + ($44)|0); //@line 201 "load_amf.cpp"
   $46 = HEAP8[$45>>0]|0; //@line 201 "load_amf.cpp"
   $47 = $46&255; //@line 201 "load_amf.cpp"
   $48 = (((((($this)) + 33280|0) + ($33<<6)|0)) + 30|0); //@line 201 "load_amf.cpp"
   HEAP16[$48>>1] = $47; //@line 201 "load_amf.cpp"
   $49 = (((((($this)) + 33280|0) + ($33<<6)|0)) + 32|0); //@line 202 "load_amf.cpp"
   HEAP16[$49>>1] = 64; //@line 202 "load_amf.cpp"
   $50 = ($46&255)>(64); //@line 203 "load_amf.cpp"
   if ($50) {
    HEAP16[$48>>1] = 64; //@line 203 "load_amf.cpp"
    $52 = 64;
   } else {
    $52 = $47;
   }
   $51 = $52&65535; //@line 204 "load_amf.cpp"
   $53 = $51 << 2; //@line 204 "load_amf.cpp"
   $54 = $53&65535; //@line 204 "load_amf.cpp"
   HEAP16[$48>>1] = $54; //@line 204 "load_amf.cpp"
   $$sum17 = (($dwMemPos$054) + 25)|0; //@line 205 "load_amf.cpp"
   $55 = (($lpStream) + ($$sum17)|0); //@line 205 "load_amf.cpp"
   $56 = HEAP32[$55>>2]|0; //@line 205 "load_amf.cpp"
   $57 = (((($this)) + 33280|0) + ($33<<6)|0); //@line 205 "load_amf.cpp"
   HEAP32[$57>>2] = $56; //@line 205 "load_amf.cpp"
   $$sum18 = (($dwMemPos$054) + 29)|0; //@line 206 "load_amf.cpp"
   $58 = (($lpStream) + ($$sum18)|0); //@line 206 "load_amf.cpp"
   $59 = HEAP32[$58>>2]|0; //@line 206 "load_amf.cpp"
   $60 = (((((($this)) + 33280|0) + ($33<<6)|0)) + 4|0); //@line 206 "load_amf.cpp"
   HEAP32[$60>>2] = $59; //@line 206 "load_amf.cpp"
   $$sum19 = (($dwMemPos$054) + 33)|0; //@line 207 "load_amf.cpp"
   $61 = (($lpStream) + ($$sum19)|0); //@line 207 "load_amf.cpp"
   $62 = HEAP32[$61>>2]|0; //@line 207 "load_amf.cpp"
   $63 = (($62) + ($59))|0; //@line 207 "load_amf.cpp"
   $64 = (((((($this)) + 33280|0) + ($33<<6)|0)) + 8|0); //@line 207 "load_amf.cpp"
   HEAP32[$64>>2] = $63; //@line 207 "load_amf.cpp"
   $65 = ($63>>>0)<=($59>>>0); //@line 208 "load_amf.cpp"
   $66 = ($63>>>0)>($56>>>0); //@line 208 "load_amf.cpp"
   $or$cond134 = $65 | $66; //@line 208 "load_amf.cpp"
   if ($or$cond134) {
    HEAP32[$64>>2] = 0; //@line 213 "load_amf.cpp"
    HEAP32[$60>>2] = 0; //@line 213 "load_amf.cpp"
   } else {
    $67 = (((((($this)) + 33280|0) + ($33<<6)|0)) + 34|0); //@line 210 "load_amf.cpp"
    HEAP16[$67>>1] = 2; //@line 210 "load_amf.cpp"
   }
   $68 = ($56|0)!=(0); //@line 215 "load_amf.cpp"
   $69 = ($iSmp$053>>>0)>(31); //@line 215 "load_amf.cpp"
   $or$cond9 = $69 & $68; //@line 215 "load_amf.cpp"
   if ($or$cond9) {
    HEAP32[$20>>2] = $33; //@line 215 "load_amf.cpp"
   }
   $70 = (($dwMemPos$054) + 37)|0; //@line 216 "load_amf.cpp"
   $exitcond108 = ($33|0)==(64); //@line 195 "load_amf.cpp"
   if ($exitcond108) {
    break;
   } else {
    $dwMemPos$054 = $70;$iSmp$053 = $33;
   }
  }
  $32 = ($6<<24>>24)==(0); //@line 218 "load_amf.cpp"
  L30: do {
   if ($32) {
    $dwMemPos$2$ph = 2662;$iData$0$ph = 0;
   } else {
    $dwMemPos$150 = 2662;$iPat$049 = 0;
    while(1) {
     $71 = HEAP32[$18>>2]|0; //@line 220 "load_amf.cpp"
     $72 = (__ZN10CSoundFile15AllocatePatternEjj(64,$71)|0); //@line 220 "load_amf.cpp"
     $73 = ($72|0)==(0|0); //@line 221 "load_amf.cpp"
     if ($73) {
      $dwMemPos$2$ph = $dwMemPos$150;$iData$0$ph = 0;
      break L30;
     }
     $74 = (((($this)) + 51904|0) + ($iPat$049<<2)|0); //@line 222 "load_amf.cpp"
     HEAP32[$74>>2] = $72; //@line 222 "load_amf.cpp"
     $75 = (((($this)) + 52864|0) + ($iPat$049<<1)|0); //@line 223 "load_amf.cpp"
     HEAP16[$75>>1] = 64; //@line 223 "load_amf.cpp"
     $76 = (($lpStream) + ($dwMemPos$150)|0); //@line 224 "load_amf.cpp"
     $i$046 = 0;$p$048 = $72;$pin$047 = $76;
     while(1) {
      HEAP8[$p$048>>0] = 0; //@line 227 "load_amf.cpp"
      $77 = HEAP8[$pin$047>>0]|0; //@line 229 "load_amf.cpp"
      $78 = ($77<<24>>24)==(0); //@line 229 "load_amf.cpp"
      if (!($78)) {
       $79 = $77&255; //@line 231 "load_amf.cpp"
       $80 = (($79) + 13)|0; //@line 231 "load_amf.cpp"
       $81 = $80&255; //@line 231 "load_amf.cpp"
       HEAP8[$p$048>>0] = $81; //@line 231 "load_amf.cpp"
      }
      $82 = ((($pin$047)) + 1|0); //@line 233 "load_amf.cpp"
      $83 = HEAP8[$82>>0]|0; //@line 233 "load_amf.cpp"
      $84 = ((($p$048)) + 1|0); //@line 233 "load_amf.cpp"
      HEAP8[$84>>0] = $83; //@line 233 "load_amf.cpp"
      $85 = ((($pin$047)) + 2|0); //@line 234 "load_amf.cpp"
      $86 = HEAP8[$85>>0]|0; //@line 234 "load_amf.cpp"
      $87 = ((($p$048)) + 3|0); //@line 234 "load_amf.cpp"
      HEAP8[$87>>0] = $86; //@line 234 "load_amf.cpp"
      $88 = ((($pin$047)) + 3|0); //@line 235 "load_amf.cpp"
      $89 = HEAP8[$88>>0]|0; //@line 235 "load_amf.cpp"
      $90 = ((($p$048)) + 5|0); //@line 235 "load_amf.cpp"
      HEAP8[$90>>0] = $89; //@line 235 "load_amf.cpp"
      $91 = ($86&255)>(15); //@line 236 "load_amf.cpp"
      if ($91) {
       HEAP8[$87>>0] = 0; //@line 241 "load_amf.cpp"
      }
      __ZNK10CSoundFile17ConvertModCommandEP11_MODCOMMAND($this,$p$048); //@line 243 "load_amf.cpp"
      $92 = ((($pin$047)) + 4|0); //@line 244 "load_amf.cpp"
      $93 = ((($p$048)) + 6|0); //@line 245 "load_amf.cpp"
      $94 = (($i$046) + 1)|0; //@line 225 "load_amf.cpp"
      $exitcond = ($94|0)==(512); //@line 225 "load_amf.cpp"
      if ($exitcond) {
       break;
      } else {
       $i$046 = $94;$p$048 = $93;$pin$047 = $92;
      }
     }
     $95 = (($dwMemPos$150) + 2048)|0; //@line 247 "load_amf.cpp"
     $96 = (($iPat$049) + 1)|0; //@line 218 "load_amf.cpp"
     $97 = ($96>>>0)<($7>>>0); //@line 218 "load_amf.cpp"
     if ($97) {
      $dwMemPos$150 = $95;$iPat$049 = $96;
     } else {
      $dwMemPos$2$ph = $95;$iData$0$ph = 0;
      break;
     }
    }
   }
  } while(0);
  L43: while(1) {
   $99 = HEAP32[$20>>2]|0; //@line 250 "load_amf.cpp"
   $iData$0 = $iData$0$ph;
   while(1) {
    $98 = ($iData$0>>>0)<($99>>>0); //@line 250 "load_amf.cpp"
    if (!($98)) {
     $$1 = 1;
     label = 98;
     break L43;
    }
    $100 = (($iData$0) + 1)|0; //@line 252 "load_amf.cpp"
    $101 = (((($this)) + 33280|0) + ($100<<6)|0); //@line 252 "load_amf.cpp"
    $102 = HEAP32[$101>>2]|0; //@line 253 "load_amf.cpp"
    $103 = ($102|0)==(0); //@line 253 "load_amf.cpp"
    if ($103) {
     $iData$0 = $100;
    } else {
     $$lcssa = $100;$$lcssa172 = $101;
     break;
    }
   }
   $104 = ($dwMemPos$2$ph>>>0)>($dwMemLength>>>0); //@line 255 "load_amf.cpp"
   if ($104) {
    $$1 = 0;
    label = 98;
    break;
   }
   $105 = (($lpStream) + ($dwMemPos$2$ph)|0); //@line 256 "load_amf.cpp"
   $106 = (($dwMemLength) - ($dwMemPos$2$ph))|0; //@line 256 "load_amf.cpp"
   $107 = (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$$lcssa172,0,$105,$106)|0); //@line 256 "load_amf.cpp"
   $108 = (($107) + ($dwMemPos$2$ph))|0; //@line 256 "load_amf.cpp"
   $dwMemPos$2$ph = $108;$iData$0$ph = $$lcssa;
  }
  if ((label|0) == 98) {
   STACKTOP = sp;return ($$1|0); //@line 421 "load_amf.cpp"
  }
 }
 $109 = HEAP8[$lpStream>>0]|0; //@line 266 "load_amf.cpp"
 $110 = ($109<<24>>24)==(65); //@line 266 "load_amf.cpp"
 if ($110) {
  $111 = ((($lpStream)) + 1|0); //@line 266 "load_amf.cpp"
  $112 = HEAP8[$111>>0]|0; //@line 266 "load_amf.cpp"
  $113 = ($112<<24>>24)==(77); //@line 266 "load_amf.cpp"
  if ($113) {
   $114 = ((($lpStream)) + 2|0); //@line 266 "load_amf.cpp"
   $115 = HEAP8[$114>>0]|0; //@line 266 "load_amf.cpp"
   $116 = ($115<<24>>24)==(70); //@line 266 "load_amf.cpp"
   if ($116) {
    $117 = ((($lpStream)) + 3|0); //@line 267 "load_amf.cpp"
    $118 = HEAP8[$117>>0]|0; //@line 267 "load_amf.cpp"
    $$off = (($118) + -10)<<24>>24; //@line 267 "load_amf.cpp"
    $119 = ($$off&255)>(4); //@line 267 "load_amf.cpp"
    if ($119) {
     $$0 = 0;
    } else {
     $120 = ((($lpStream)) + 38|0); //@line 267 "load_amf.cpp"
     $121 = HEAPU8[$120>>0]|(HEAPU8[$120+1>>0]<<8); //@line 267 "load_amf.cpp"
     $122 = ($121<<16>>16)==(0); //@line 267 "load_amf.cpp"
     if ($122) {
      $$0 = 0;
     } else {
      $123 = ((($lpStream)) + 37|0); //@line 268 "load_amf.cpp"
      $124 = HEAP8[$123>>0]|0; //@line 268 "load_amf.cpp"
      $125 = (($124) + -1)<<24>>24; //@line 268 "load_amf.cpp"
      $126 = ($125&255)>(239); //@line 268 "load_amf.cpp"
      if ($126) {
       $$0 = 0;
      } else {
       $127 = ((($lpStream)) + 36|0); //@line 269 "load_amf.cpp"
       $128 = HEAP8[$127>>0]|0; //@line 269 "load_amf.cpp"
       $129 = (($128) + -1)<<24>>24; //@line 269 "load_amf.cpp"
       $130 = ($129&255)>(239); //@line 269 "load_amf.cpp"
       if ($130) {
        $$0 = 0;
       } else {
        $131 = ((($lpStream)) + 40|0); //@line 270 "load_amf.cpp"
        $132 = HEAP8[$131>>0]|0; //@line 270 "load_amf.cpp"
        $$off25 = (($132) + -4)<<24>>24; //@line 270 "load_amf.cpp"
        $133 = ($$off25&255)>(28); //@line 270 "load_amf.cpp"
        if ($133) {
         $$0 = 0;
        } else {
         $134 = ((($this)) + 59804|0); //@line 272 "load_amf.cpp"
         $135 = ((($lpStream)) + 4|0); //@line 272 "load_amf.cpp"
         dest=$134; src=$135; stop=dest+31|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 272 "load_amf.cpp"
         $136 = ((($this)) + 59835|0); //@line 273 "load_amf.cpp"
         HEAP8[$136>>0] = 0; //@line 273 "load_amf.cpp"
         $137 = ((($this)) + 59680|0); //@line 275 "load_amf.cpp"
         HEAP32[$137>>2] = 2048; //@line 275 "load_amf.cpp"
         $138 = HEAP8[$131>>0]|0; //@line 276 "load_amf.cpp"
         $139 = $138&255; //@line 276 "load_amf.cpp"
         $140 = ((($this)) + 59664|0); //@line 276 "load_amf.cpp"
         HEAP32[$140>>2] = $139; //@line 276 "load_amf.cpp"
         $141 = HEAP8[$127>>0]|0; //@line 277 "load_amf.cpp"
         $142 = $141&255; //@line 277 "load_amf.cpp"
         $143 = ((($this)) + 59684|0); //@line 277 "load_amf.cpp"
         HEAP32[$143>>2] = $142; //@line 277 "load_amf.cpp"
         $144 = ((($this)) + 59688|0); //@line 278 "load_amf.cpp"
         HEAP32[$144>>2] = 0; //@line 278 "load_amf.cpp"
         $145 = HEAP8[$117>>0]|0; //@line 280 "load_amf.cpp"
         $146 = ($145&255)>(10); //@line 280 "load_amf.cpp"
         if ($146) {
          $243 = ($145&255)>(12); //@line 283 "load_amf.cpp"
          $244 = $243 ? 32 : 16; //@line 283 "load_amf.cpp"
          $i2$091 = 0;
          while(1) {
           $$sum16 = (($i2$091) + 41)|0; //@line 286 "load_amf.cpp"
           $245 = (($lpStream) + ($$sum16)|0); //@line 286 "load_amf.cpp"
           $246 = HEAP8[$245>>0]|0; //@line 286 "load_amf.cpp"
           $247 = $246 << 24 >> 24; //@line 286 "load_amf.cpp"
           $248 = $247 << 1; //@line 286 "load_amf.cpp"
           $249 = (($248) + 128)|0; //@line 286 "load_amf.cpp"
           $250 = ($249|0)<(0); //@line 287 "load_amf.cpp"
           $$ = $250 ? 0 : $249; //@line 287 "load_amf.cpp"
           $251 = ($$|0)>(256); //@line 288 "load_amf.cpp"
           if ($251) {
            $252 = (((((($this)) + 49600|0) + (($i2$091*36)|0)|0)) + 8|0); //@line 288 "load_amf.cpp"
            $253 = HEAP32[$252>>2]|0; //@line 288 "load_amf.cpp"
            $254 = $253 | 2048; //@line 288 "load_amf.cpp"
            HEAP32[$252>>2] = $254; //@line 288 "load_amf.cpp"
            $pan$0 = 128;
           } else {
            $pan$0 = $$;
           }
           $255 = (((($this)) + 49600|0) + (($i2$091*36)|0)|0); //@line 289 "load_amf.cpp"
           HEAP32[$255>>2] = $pan$0; //@line 289 "load_amf.cpp"
           $256 = (($i2$091) + 1)|0; //@line 284 "load_amf.cpp"
           $257 = ($256>>>0)<($244>>>0); //@line 284 "load_amf.cpp"
           if ($257) {
            $i2$091 = $256;
           } else {
            break;
           }
          }
          $258 = (($244) + 41)|0; //@line 291 "load_amf.cpp"
          $262 = $145;$dwMemPos$4 = $258;
         } else {
          $147 = ((($lpStream)) + 41|0); //@line 296 "load_amf.cpp"
          $148 = HEAP8[$147>>0]|0; //@line 296 "load_amf.cpp"
          $149 = $148 & 1; //@line 296 "load_amf.cpp"
          $150 = ($149<<24>>24)!=(0); //@line 296 "load_amf.cpp"
          $151 = $150 ? 48 : 208; //@line 296 "load_amf.cpp"
          $152 = ((($this)) + 49600|0); //@line 296 "load_amf.cpp"
          HEAP32[$152>>2] = $151; //@line 296 "load_amf.cpp"
          $153 = ((($lpStream)) + 42|0); //@line 296 "load_amf.cpp"
          $154 = HEAP8[$153>>0]|0; //@line 296 "load_amf.cpp"
          $155 = $154 & 1; //@line 296 "load_amf.cpp"
          $156 = ($155<<24>>24)!=(0); //@line 296 "load_amf.cpp"
          $157 = $156 ? 48 : 208; //@line 296 "load_amf.cpp"
          $158 = ((($this)) + 49636|0); //@line 296 "load_amf.cpp"
          HEAP32[$158>>2] = $157; //@line 296 "load_amf.cpp"
          $159 = ((($lpStream)) + 43|0); //@line 296 "load_amf.cpp"
          $160 = HEAP8[$159>>0]|0; //@line 296 "load_amf.cpp"
          $161 = $160 & 1; //@line 296 "load_amf.cpp"
          $162 = ($161<<24>>24)!=(0); //@line 296 "load_amf.cpp"
          $163 = $162 ? 48 : 208; //@line 296 "load_amf.cpp"
          $164 = ((($this)) + 49672|0); //@line 296 "load_amf.cpp"
          HEAP32[$164>>2] = $163; //@line 296 "load_amf.cpp"
          $165 = ((($lpStream)) + 44|0); //@line 296 "load_amf.cpp"
          $166 = HEAP8[$165>>0]|0; //@line 296 "load_amf.cpp"
          $167 = $166 & 1; //@line 296 "load_amf.cpp"
          $168 = ($167<<24>>24)!=(0); //@line 296 "load_amf.cpp"
          $169 = $168 ? 48 : 208; //@line 296 "load_amf.cpp"
          $170 = ((($this)) + 49708|0); //@line 296 "load_amf.cpp"
          HEAP32[$170>>2] = $169; //@line 296 "load_amf.cpp"
          $171 = ((($lpStream)) + 45|0); //@line 296 "load_amf.cpp"
          $172 = HEAP8[$171>>0]|0; //@line 296 "load_amf.cpp"
          $173 = $172 & 1; //@line 296 "load_amf.cpp"
          $174 = ($173<<24>>24)!=(0); //@line 296 "load_amf.cpp"
          $175 = $174 ? 48 : 208; //@line 296 "load_amf.cpp"
          $176 = ((($this)) + 49744|0); //@line 296 "load_amf.cpp"
          HEAP32[$176>>2] = $175; //@line 296 "load_amf.cpp"
          $177 = ((($lpStream)) + 46|0); //@line 296 "load_amf.cpp"
          $178 = HEAP8[$177>>0]|0; //@line 296 "load_amf.cpp"
          $179 = $178 & 1; //@line 296 "load_amf.cpp"
          $180 = ($179<<24>>24)!=(0); //@line 296 "load_amf.cpp"
          $181 = $180 ? 48 : 208; //@line 296 "load_amf.cpp"
          $182 = ((($this)) + 49780|0); //@line 296 "load_amf.cpp"
          HEAP32[$182>>2] = $181; //@line 296 "load_amf.cpp"
          $183 = ((($lpStream)) + 47|0); //@line 296 "load_amf.cpp"
          $184 = HEAP8[$183>>0]|0; //@line 296 "load_amf.cpp"
          $185 = $184 & 1; //@line 296 "load_amf.cpp"
          $186 = ($185<<24>>24)!=(0); //@line 296 "load_amf.cpp"
          $187 = $186 ? 48 : 208; //@line 296 "load_amf.cpp"
          $188 = ((($this)) + 49816|0); //@line 296 "load_amf.cpp"
          HEAP32[$188>>2] = $187; //@line 296 "load_amf.cpp"
          $189 = ((($lpStream)) + 48|0); //@line 296 "load_amf.cpp"
          $190 = HEAP8[$189>>0]|0; //@line 296 "load_amf.cpp"
          $191 = $190 & 1; //@line 296 "load_amf.cpp"
          $192 = ($191<<24>>24)!=(0); //@line 296 "load_amf.cpp"
          $193 = $192 ? 48 : 208; //@line 296 "load_amf.cpp"
          $194 = ((($this)) + 49852|0); //@line 296 "load_amf.cpp"
          HEAP32[$194>>2] = $193; //@line 296 "load_amf.cpp"
          $195 = ((($lpStream)) + 49|0); //@line 296 "load_amf.cpp"
          $196 = HEAP8[$195>>0]|0; //@line 296 "load_amf.cpp"
          $197 = $196 & 1; //@line 296 "load_amf.cpp"
          $198 = ($197<<24>>24)!=(0); //@line 296 "load_amf.cpp"
          $199 = $198 ? 48 : 208; //@line 296 "load_amf.cpp"
          $200 = ((($this)) + 49888|0); //@line 296 "load_amf.cpp"
          HEAP32[$200>>2] = $199; //@line 296 "load_amf.cpp"
          $201 = ((($lpStream)) + 50|0); //@line 296 "load_amf.cpp"
          $202 = HEAP8[$201>>0]|0; //@line 296 "load_amf.cpp"
          $203 = $202 & 1; //@line 296 "load_amf.cpp"
          $204 = ($203<<24>>24)!=(0); //@line 296 "load_amf.cpp"
          $205 = $204 ? 48 : 208; //@line 296 "load_amf.cpp"
          $206 = ((($this)) + 49924|0); //@line 296 "load_amf.cpp"
          HEAP32[$206>>2] = $205; //@line 296 "load_amf.cpp"
          $207 = ((($lpStream)) + 51|0); //@line 296 "load_amf.cpp"
          $208 = HEAP8[$207>>0]|0; //@line 296 "load_amf.cpp"
          $209 = $208 & 1; //@line 296 "load_amf.cpp"
          $210 = ($209<<24>>24)!=(0); //@line 296 "load_amf.cpp"
          $211 = $210 ? 48 : 208; //@line 296 "load_amf.cpp"
          $212 = ((($this)) + 49960|0); //@line 296 "load_amf.cpp"
          HEAP32[$212>>2] = $211; //@line 296 "load_amf.cpp"
          $213 = ((($lpStream)) + 52|0); //@line 296 "load_amf.cpp"
          $214 = HEAP8[$213>>0]|0; //@line 296 "load_amf.cpp"
          $215 = $214 & 1; //@line 296 "load_amf.cpp"
          $216 = ($215<<24>>24)!=(0); //@line 296 "load_amf.cpp"
          $217 = $216 ? 48 : 208; //@line 296 "load_amf.cpp"
          $218 = ((($this)) + 49996|0); //@line 296 "load_amf.cpp"
          HEAP32[$218>>2] = $217; //@line 296 "load_amf.cpp"
          $219 = ((($lpStream)) + 53|0); //@line 296 "load_amf.cpp"
          $220 = HEAP8[$219>>0]|0; //@line 296 "load_amf.cpp"
          $221 = $220 & 1; //@line 296 "load_amf.cpp"
          $222 = ($221<<24>>24)!=(0); //@line 296 "load_amf.cpp"
          $223 = $222 ? 48 : 208; //@line 296 "load_amf.cpp"
          $224 = ((($this)) + 50032|0); //@line 296 "load_amf.cpp"
          HEAP32[$224>>2] = $223; //@line 296 "load_amf.cpp"
          $225 = ((($lpStream)) + 54|0); //@line 296 "load_amf.cpp"
          $226 = HEAP8[$225>>0]|0; //@line 296 "load_amf.cpp"
          $227 = $226 & 1; //@line 296 "load_amf.cpp"
          $228 = ($227<<24>>24)!=(0); //@line 296 "load_amf.cpp"
          $229 = $228 ? 48 : 208; //@line 296 "load_amf.cpp"
          $230 = ((($this)) + 50068|0); //@line 296 "load_amf.cpp"
          HEAP32[$230>>2] = $229; //@line 296 "load_amf.cpp"
          $231 = ((($lpStream)) + 55|0); //@line 296 "load_amf.cpp"
          $232 = HEAP8[$231>>0]|0; //@line 296 "load_amf.cpp"
          $233 = $232 & 1; //@line 296 "load_amf.cpp"
          $234 = ($233<<24>>24)!=(0); //@line 296 "load_amf.cpp"
          $235 = $234 ? 48 : 208; //@line 296 "load_amf.cpp"
          $236 = ((($this)) + 50104|0); //@line 296 "load_amf.cpp"
          HEAP32[$236>>2] = $235; //@line 296 "load_amf.cpp"
          $237 = ((($lpStream)) + 56|0); //@line 296 "load_amf.cpp"
          $238 = HEAP8[$237>>0]|0; //@line 296 "load_amf.cpp"
          $239 = $238 & 1; //@line 296 "load_amf.cpp"
          $240 = ($239<<24>>24)!=(0); //@line 296 "load_amf.cpp"
          $241 = $240 ? 48 : 208; //@line 296 "load_amf.cpp"
          $242 = ((($this)) + 50140|0); //@line 296 "load_amf.cpp"
          HEAP32[$242>>2] = $241; //@line 296 "load_amf.cpp"
          $$pre = HEAP8[$117>>0]|0; //@line 303 "load_amf.cpp"
          $262 = $$pre;$dwMemPos$4 = 57;
         }
         $259 = ((($this)) + 59652|0); //@line 301 "load_amf.cpp"
         HEAP32[$259>>2] = 125; //@line 301 "load_amf.cpp"
         $260 = ((($this)) + 59648|0); //@line 302 "load_amf.cpp"
         HEAP32[$260>>2] = 6; //@line 302 "load_amf.cpp"
         $261 = ($262&255)>(12); //@line 303 "load_amf.cpp"
         if ($261) {
          $263 = (($lpStream) + ($dwMemPos$4)|0); //@line 305 "load_amf.cpp"
          $264 = HEAP8[$263>>0]|0; //@line 305 "load_amf.cpp"
          $265 = ($264&255)>(31); //@line 305 "load_amf.cpp"
          if ($265) {
           $266 = $264&255; //@line 305 "load_amf.cpp"
           HEAP32[$259>>2] = $266; //@line 305 "load_amf.cpp"
          }
          $267 = (($dwMemPos$4) + 1)|0; //@line 306 "load_amf.cpp"
          $268 = (($lpStream) + ($267)|0); //@line 306 "load_amf.cpp"
          $269 = HEAP8[$268>>0]|0; //@line 306 "load_amf.cpp"
          $270 = ($269&255)<(33); //@line 306 "load_amf.cpp"
          if ($270) {
           $271 = $269&255; //@line 306 "load_amf.cpp"
           HEAP32[$260>>2] = $271; //@line 306 "load_amf.cpp"
          }
          $272 = (($dwMemPos$4) + 2)|0; //@line 307 "load_amf.cpp"
          $dwMemPos$690 = $272;$iOrd4$089 = 0;
         } else {
          $dwMemPos$690 = $dwMemPos$4;$iOrd4$089 = 0;
         }
         while(1) {
          $273 = (((($this)) + 53344|0) + ($iOrd4$089)|0); //@line 312 "load_amf.cpp"
          HEAP8[$273>>0] = -1; //@line 312 "load_amf.cpp"
          $274 = HEAP8[$123>>0]|0; //@line 313 "load_amf.cpp"
          $275 = $274&255; //@line 313 "load_amf.cpp"
          $276 = ($iOrd4$089>>>0)<($275>>>0); //@line 313 "load_amf.cpp"
          if ($276) {
           $277 = $iOrd4$089&255; //@line 315 "load_amf.cpp"
           HEAP8[$273>>0] = $277; //@line 315 "load_amf.cpp"
           $278 = (((($this)) + 52864|0) + ($iOrd4$089<<1)|0); //@line 316 "load_amf.cpp"
           HEAP16[$278>>1] = 64; //@line 316 "load_amf.cpp"
           $279 = HEAP8[$117>>0]|0; //@line 317 "load_amf.cpp"
           $280 = ($279&255)>(13); //@line 317 "load_amf.cpp"
           if ($280) {
            $281 = (($lpStream) + ($dwMemPos$690)|0); //@line 319 "load_amf.cpp"
            $282 = HEAP16[$281>>1]|0; //@line 319 "load_amf.cpp"
            HEAP16[$278>>1] = $282; //@line 319 "load_amf.cpp"
            $283 = (($dwMemPos$690) + 2)|0; //@line 320 "load_amf.cpp"
            $dwMemPos$7 = $283;
           } else {
            $dwMemPos$7 = $dwMemPos$690;
           }
           $284 = (($lpStream) + ($dwMemPos$7)|0); //@line 322 "load_amf.cpp"
           $285 = (($ptracks) + ($iOrd4$089<<2)|0); //@line 322 "load_amf.cpp"
           HEAP32[$285>>2] = $284; //@line 322 "load_amf.cpp"
           $286 = HEAP32[$140>>2]|0; //@line 323 "load_amf.cpp"
           $287 = $286 << 1; //@line 323 "load_amf.cpp"
           $288 = (($287) + ($dwMemPos$7))|0; //@line 323 "load_amf.cpp"
           $dwMemPos$8 = $288;
          } else {
           $dwMemPos$8 = $dwMemPos$690;
          }
          $289 = (($iOrd4$089) + 1)|0; //@line 310 "load_amf.cpp"
          $exitcond111 = ($289|0)==(256); //@line 310 "load_amf.cpp"
          if ($exitcond111) {
           $dwMemPos$8$lcssa = $dwMemPos$8;
           break;
          } else {
           $dwMemPos$690 = $dwMemPos$8;$iOrd4$089 = $289;
          }
         }
         $290 = HEAP32[$143>>2]|0; //@line 326 "load_amf.cpp"
         $291 = ($290*65)|0; //@line 326 "load_amf.cpp"
         $292 = (($291) + ($dwMemPos$8$lcssa))|0; //@line 326 "load_amf.cpp"
         $293 = ($292>>>0)>($dwMemLength>>>0); //@line 326 "load_amf.cpp"
         if ($293) {
          $$0 = 1;
         } else {
          $294 = ($290|0)==(0); //@line 329 "load_amf.cpp"
          L91: do {
           if ($294) {
            $dwMemPos$9$lcssa = $dwMemPos$8$lcssa;$maxsampleseekpos$0$ph$lcssa = 0;
           } else {
            $295 = (($dwMemLength) + -1)|0; //@line 355 "load_amf.cpp"
            $dwMemPos$9$ph86 = $dwMemPos$8$lcssa;$iIns$0$ph85 = 0;$maxsampleseekpos$0$ph84 = 0;
            while(1) {
             $dwMemPos$977 = $dwMemPos$9$ph86;$iIns$076 = $iIns$0$ph85;
             while(1) {
              $296 = (($iIns$076) + 1)|0; //@line 331 "load_amf.cpp"
              $297 = (($lpStream) + ($dwMemPos$977)|0); //@line 332 "load_amf.cpp"
              $298 = (($dwMemPos$977) + 57)|0; //@line 334 "load_amf.cpp"
              $299 = (((($this)) + 59804|0) + ($296<<5)|0); //@line 335 "load_amf.cpp"
              $$sum = (($dwMemPos$977) + 1)|0; //@line 335 "load_amf.cpp"
              $300 = (($lpStream) + ($$sum)|0); //@line 335 "load_amf.cpp"
              dest=$299; src=$300; stop=dest+32|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 335 "load_amf.cpp"
              $301 = (((((($this)) + 59804|0) + ($296<<5)|0)) + 31|0); //@line 336 "load_amf.cpp"
              HEAP8[$301>>0] = 0; //@line 336 "load_amf.cpp"
              $302 = (((((($this)) + 33280|0) + ($296<<6)|0)) + 42|0); //@line 337 "load_amf.cpp"
              $$sum10 = (($dwMemPos$977) + 33)|0; //@line 337 "load_amf.cpp"
              $303 = (($lpStream) + ($$sum10)|0); //@line 337 "load_amf.cpp"
              dest=$302; src=$303; stop=dest+13|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 337 "load_amf.cpp"
              $304 = (((((($this)) + 33280|0) + ($296<<6)|0)) + 54|0); //@line 338 "load_amf.cpp"
              HEAP8[$304>>0] = 0; //@line 338 "load_amf.cpp"
              $$sum11 = (($dwMemPos$977) + 50)|0; //@line 339 "load_amf.cpp"
              $305 = (($lpStream) + ($$sum11)|0); //@line 339 "load_amf.cpp"
              $306 = HEAPU8[$305>>0]|(HEAPU8[$305+1>>0]<<8)|(HEAPU8[$305+2>>0]<<16)|(HEAPU8[$305+3>>0]<<24); //@line 339 "load_amf.cpp"
              $307 = (((($this)) + 33280|0) + ($296<<6)|0); //@line 339 "load_amf.cpp"
              HEAP32[$307>>2] = $306; //@line 339 "load_amf.cpp"
              $$sum12 = (($dwMemPos$977) + 54)|0; //@line 340 "load_amf.cpp"
              $308 = (($lpStream) + ($$sum12)|0); //@line 340 "load_amf.cpp"
              $309 = HEAPU8[$308>>0]|(HEAPU8[$308+1>>0]<<8); //@line 340 "load_amf.cpp"
              $310 = $309&65535; //@line 340 "load_amf.cpp"
              $311 = (((((($this)) + 33280|0) + ($296<<6)|0)) + 24|0); //@line 340 "load_amf.cpp"
              HEAP32[$311>>2] = $310; //@line 340 "load_amf.cpp"
              $312 = (((((($this)) + 33280|0) + ($296<<6)|0)) + 32|0); //@line 341 "load_amf.cpp"
              HEAP16[$312>>1] = 64; //@line 341 "load_amf.cpp"
              $$sum13 = (($dwMemPos$977) + 56)|0; //@line 342 "load_amf.cpp"
              $313 = (($lpStream) + ($$sum13)|0); //@line 342 "load_amf.cpp"
              $314 = HEAP8[$313>>0]|0; //@line 342 "load_amf.cpp"
              $315 = $314&255; //@line 342 "load_amf.cpp"
              $316 = $315 << 2; //@line 342 "load_amf.cpp"
              $317 = $316&65535; //@line 342 "load_amf.cpp"
              $318 = (((((($this)) + 33280|0) + ($296<<6)|0)) + 30|0); //@line 342 "load_amf.cpp"
              HEAP16[$318>>1] = $317; //@line 342 "load_amf.cpp"
              $319 = HEAP8[$117>>0]|0; //@line 343 "load_amf.cpp"
              $320 = ($319&255)>(10); //@line 343 "load_amf.cpp"
              $321 = (($lpStream) + ($298)|0); //@line 345 "load_amf.cpp"
              if ($320) {
               $322 = HEAP32[$321>>2]|0; //@line 345 "load_amf.cpp"
               $323 = (((((($this)) + 33280|0) + ($296<<6)|0)) + 4|0); //@line 345 "load_amf.cpp"
               HEAP32[$323>>2] = $322; //@line 345 "load_amf.cpp"
               $$sum15 = (($dwMemPos$977) + 61)|0; //@line 346 "load_amf.cpp"
               $324 = (($lpStream) + ($$sum15)|0); //@line 346 "load_amf.cpp"
               $325 = HEAP32[$324>>2]|0; //@line 346 "load_amf.cpp"
               $326 = (($dwMemPos$977) + 65)|0; //@line 347 "load_amf.cpp"
               $332 = $325;$440 = $322;$dwMemPos$10 = $326;
              } else {
               $327 = HEAP16[$321>>1]|0; //@line 350 "load_amf.cpp"
               $328 = $327&65535; //@line 350 "load_amf.cpp"
               $329 = (((((($this)) + 33280|0) + ($296<<6)|0)) + 4|0); //@line 350 "load_amf.cpp"
               HEAP32[$329>>2] = $328; //@line 350 "load_amf.cpp"
               $330 = (($dwMemPos$977) + 59)|0; //@line 352 "load_amf.cpp"
               $332 = $306;$440 = $328;$dwMemPos$10 = $330;
              }
              $331 = (((((($this)) + 33280|0) + ($296<<6)|0)) + 8|0); //@line 351 "load_amf.cpp"
              HEAP32[$331>>2] = $332; //@line 351 "load_amf.cpp"
              $333 = (($sampleseekpos) + ($iIns$076<<2)|0); //@line 354 "load_amf.cpp"
              HEAP32[$333>>2] = 0; //@line 354 "load_amf.cpp"
              $334 = HEAP8[$297>>0]|0; //@line 355 "load_amf.cpp"
              $335 = ($334<<24>>24)==(0); //@line 355 "load_amf.cpp"
              if (!($335)) {
               $$sum14 = (($dwMemPos$977) + 46)|0; //@line 355 "load_amf.cpp"
               $338 = (($lpStream) + ($$sum14)|0); //@line 355 "load_amf.cpp"
               $339 = HEAPU8[$338>>0]|(HEAPU8[$338+1>>0]<<8)|(HEAPU8[$338+2>>0]<<16)|(HEAPU8[$338+3>>0]<<24); //@line 355 "load_amf.cpp"
               $340 = ($339>>>0)<($295>>>0); //@line 355 "load_amf.cpp"
               if ($340) {
                $$lcssa175 = $296;$$lcssa177 = $306;$$lcssa179 = $440;$$lcssa181 = $332;$$lcssa184 = $333;$$lcssa186 = $339;$dwMemPos$10$lcssa = $dwMemPos$10;
                break;
               }
              }
              $336 = HEAP32[$143>>2]|0; //@line 329 "load_amf.cpp"
              $337 = ($296>>>0)<($336>>>0); //@line 329 "load_amf.cpp"
              if ($337) {
               $dwMemPos$977 = $dwMemPos$10;$iIns$076 = $296;
              } else {
               $dwMemPos$9$lcssa = $dwMemPos$10;$maxsampleseekpos$0$ph$lcssa = $maxsampleseekpos$0$ph84;
               break L91;
              }
             }
             HEAP32[$$lcssa184>>2] = $$lcssa186; //@line 357 "load_amf.cpp"
             $341 = ($$lcssa186>>>0)>($maxsampleseekpos$0$ph84>>>0); //@line 358 "load_amf.cpp"
             $$maxsampleseekpos$0 = $341 ? $$lcssa186 : $maxsampleseekpos$0$ph84; //@line 358 "load_amf.cpp"
             $342 = (($$lcssa179) + 2)|0; //@line 360 "load_amf.cpp"
             $343 = ($$lcssa181>>>0)<=($342>>>0); //@line 360 "load_amf.cpp"
             $344 = ($$lcssa181>>>0)>($$lcssa177>>>0); //@line 361 "load_amf.cpp"
             $or$cond135 = $343 | $344; //@line 361 "load_amf.cpp"
             if (!($or$cond135)) {
              $345 = (((((($this)) + 33280|0) + ($$lcssa175<<6)|0)) + 34|0); //@line 361 "load_amf.cpp"
              $346 = HEAP16[$345>>1]|0; //@line 361 "load_amf.cpp"
              $347 = $346&65535; //@line 361 "load_amf.cpp"
              $348 = $347 | 2; //@line 361 "load_amf.cpp"
              $349 = $348&65535; //@line 361 "load_amf.cpp"
              HEAP16[$345>>1] = $349; //@line 361 "load_amf.cpp"
             }
             $350 = HEAP32[$143>>2]|0; //@line 329 "load_amf.cpp"
             $351 = ($$lcssa175>>>0)<($350>>>0); //@line 329 "load_amf.cpp"
             if ($351) {
              $dwMemPos$9$ph86 = $dwMemPos$10$lcssa;$iIns$0$ph85 = $$lcssa175;$maxsampleseekpos$0$ph84 = $$maxsampleseekpos$0;
             } else {
              $dwMemPos$9$lcssa = $dwMemPos$10$lcssa;$maxsampleseekpos$0$ph$lcssa = $$maxsampleseekpos$0;
              break;
             }
            }
           }
          } while(0);
          $352 = (($lpStream) + ($dwMemPos$9$lcssa)|0); //@line 365 "load_amf.cpp"
          $353 = HEAPU8[$120>>0]|(HEAPU8[$120+1>>0]<<8); //@line 367 "load_amf.cpp"
          $354 = $353&65535; //@line 367 "load_amf.cpp"
          $355 = $354 << 1; //@line 367 "load_amf.cpp"
          $356 = (($355) + ($dwMemPos$9$lcssa))|0; //@line 367 "load_amf.cpp"
          $357 = ($353<<16>>16)==(0); //@line 368 "load_amf.cpp"
          if ($357) {
           $realtrackcnt$0$lcssa = 0;
          } else {
           $iTrkMap$071 = 0;$realtrackcnt$070 = 0;
           while(1) {
            $358 = (($352) + ($iTrkMap$071<<1)|0); //@line 370 "load_amf.cpp"
            $359 = HEAP16[$358>>1]|0; //@line 370 "load_amf.cpp"
            $360 = $359&65535; //@line 370 "load_amf.cpp"
            $361 = ($realtrackcnt$070>>>0)<($360>>>0); //@line 370 "load_amf.cpp"
            $$realtrackcnt$0 = $361 ? $360 : $realtrackcnt$070; //@line 370 "load_amf.cpp"
            $362 = (($iTrkMap$071) + 1)|0; //@line 368 "load_amf.cpp"
            $363 = ($362>>>0)<($354>>>0); //@line 368 "load_amf.cpp"
            if ($363) {
             $iTrkMap$071 = $362;$realtrackcnt$070 = $$realtrackcnt$0;
            } else {
             $realtrackcnt$0$lcssa = $$realtrackcnt$0;
             break;
            }
           }
          }
          $$arith = $realtrackcnt$0$lcssa<<2; //@line 373 "load_amf.cpp"
          $$overflow = ($realtrackcnt$0$lcssa>>>0)>(1073741823); //@line 373 "load_amf.cpp"
          $364 = $$overflow ? -1 : $$arith; //@line 373 "load_amf.cpp"
          $365 = (__Znaj($364)|0); //@line 373 "load_amf.cpp"
          HEAP32[$365>>2] = 0; //@line 374 "load_amf.cpp"
          $366 = ($realtrackcnt$0$lcssa|0)==(0); //@line 375 "load_amf.cpp"
          if ($366) {
           $dwMemPos$11$lcssa = $356;
          } else {
           $367 = (($dwMemLength) + -3)|0; //@line 375 "load_amf.cpp"
           $dwMemPos$1166 = $356;$iTrack$065 = 0;
           while(1) {
            $370 = ($dwMemPos$1166>>>0)>($367>>>0); //@line 375 "load_amf.cpp"
            do {
             if ($370) {
              $dwMemPos$12 = $dwMemPos$1166;
             } else {
              $371 = (($lpStream) + ($dwMemPos$1166)|0); //@line 377 "load_amf.cpp"
              $372 = HEAP16[$371>>1]|0; //@line 377 "load_amf.cpp"
              $373 = $372&65535; //@line 377 "load_amf.cpp"
              $374 = (($dwMemPos$1166) + 2)|0; //@line 378 "load_amf.cpp"
              $375 = (($lpStream) + ($374)|0); //@line 378 "load_amf.cpp"
              $376 = HEAP8[$375>>0]|0; //@line 378 "load_amf.cpp"
              $377 = $376&255; //@line 378 "load_amf.cpp"
              $378 = $377 << 16; //@line 378 "load_amf.cpp"
              $379 = $378 | $373; //@line 378 "load_amf.cpp"
              $380 = ($379*3)|0; //@line 379 "load_amf.cpp"
              $381 = (($dwMemPos$1166) + 3)|0; //@line 379 "load_amf.cpp"
              $382 = (($381) + ($380))|0; //@line 379 "load_amf.cpp"
              $383 = ($382>>>0)>($dwMemLength>>>0); //@line 379 "load_amf.cpp"
              if ($383) {
               $dwMemPos$12 = $382;
               break;
              }
              $384 = (($365) + ($iTrack$065<<2)|0); //@line 381 "load_amf.cpp"
              HEAP32[$384>>2] = $371; //@line 381 "load_amf.cpp"
              $dwMemPos$12 = $382;
             }
            } while(0);
            $385 = (($iTrack$065) + 1)|0; //@line 375 "load_amf.cpp"
            $exitcond110 = ($385|0)==($realtrackcnt$0$lcssa|0); //@line 375 "load_amf.cpp"
            if ($exitcond110) {
             $dwMemPos$11$lcssa = $dwMemPos$12;
             break;
            } else {
             $dwMemPos$1166 = $dwMemPos$12;$iTrack$065 = $385;
            }
           }
          }
          $368 = HEAP8[$123>>0]|0; //@line 386 "load_amf.cpp"
          $369 = ($368<<24>>24)==(0); //@line 386 "load_amf.cpp"
          L122: do {
           if (!($369)) {
            $$pre114 = HEAP32[$140>>2]|0; //@line 388 "load_amf.cpp"
            $389 = $$pre114;$iPat5$061 = 0;
            while(1) {
             $386 = (((($this)) + 52864|0) + ($iPat5$061<<1)|0); //@line 388 "load_amf.cpp"
             $387 = HEAP16[$386>>1]|0; //@line 388 "load_amf.cpp"
             $388 = $387&65535; //@line 388 "load_amf.cpp"
             $390 = (__ZN10CSoundFile15AllocatePatternEjj($388,$389)|0); //@line 388 "load_amf.cpp"
             $391 = ($390|0)==(0|0); //@line 389 "load_amf.cpp"
             if ($391) {
              break L122;
             }
             $392 = (((($this)) + 51904|0) + ($iPat5$061<<2)|0); //@line 390 "load_amf.cpp"
             HEAP32[$392>>2] = $390; //@line 390 "load_amf.cpp"
             $393 = HEAP32[$140>>2]|0; //@line 391 "load_amf.cpp"
             $394 = ($393|0)==(0); //@line 391 "load_amf.cpp"
             if ($394) {
              $441 = 0;
             } else {
              $395 = (($ptracks) + ($iPat5$061<<2)|0); //@line 393 "load_amf.cpp"
              $396 = HEAP32[$395>>2]|0; //@line 393 "load_amf.cpp"
              $416 = $393;$iChn$059 = 0;
              while(1) {
               $397 = (($396) + ($iChn$059<<1)|0); //@line 393 "load_amf.cpp"
               $398 = HEAP16[$397>>1]|0; //@line 393 "load_amf.cpp"
               $399 = $398&65535; //@line 393 "load_amf.cpp"
               $400 = ($398<<16>>16)==(0); //@line 394 "load_amf.cpp"
               do {
                if ($400) {
                 $419 = $416;
                } else {
                 $401 = HEAPU8[$120>>0]|(HEAPU8[$120+1>>0]<<8); //@line 394 "load_amf.cpp"
                 $402 = ($398&65535)>($401&65535); //@line 394 "load_amf.cpp"
                 if ($402) {
                  $419 = $416;
                  break;
                 }
                 $403 = (($399) + -1)|0; //@line 396 "load_amf.cpp"
                 $404 = (($352) + ($403<<1)|0); //@line 396 "load_amf.cpp"
                 $405 = HEAP16[$404>>1]|0; //@line 396 "load_amf.cpp"
                 $406 = ($405<<16>>16)==(0); //@line 397 "load_amf.cpp"
                 if ($406) {
                  $419 = $416;
                  break;
                 }
                 $407 = $405&65535; //@line 396 "load_amf.cpp"
                 $408 = (($407) + -1)|0; //@line 399 "load_amf.cpp"
                 $409 = ($408>>>0)<($realtrackcnt$0$lcssa>>>0); //@line 400 "load_amf.cpp"
                 if (!($409)) {
                  $419 = $416;
                  break;
                 }
                 $410 = (($365) + ($408<<2)|0); //@line 400 "load_amf.cpp"
                 $411 = HEAP32[$410>>2]|0; //@line 400 "load_amf.cpp"
                 $412 = ($411|0)==(0|0); //@line 400 "load_amf.cpp"
                 if ($412) {
                  $419 = $416;
                  break;
                 }
                 $413 = (($390) + (($iChn$059*6)|0)|0); //@line 402 "load_amf.cpp"
                 $414 = HEAP16[$386>>1]|0; //@line 402 "load_amf.cpp"
                 $415 = $414&65535; //@line 402 "load_amf.cpp"
                 __Z10AMF_UnpackP11_MODCOMMANDPKhjj($413,$411,$415,$416); //@line 402 "load_amf.cpp"
                 $$pre115 = HEAP32[$140>>2]|0; //@line 391 "load_amf.cpp"
                 $419 = $$pre115;
                }
               } while(0);
               $417 = (($iChn$059) + 1)|0; //@line 391 "load_amf.cpp"
               $418 = ($417>>>0)<($419>>>0); //@line 391 "load_amf.cpp"
               if ($418) {
                $416 = $419;$iChn$059 = $417;
               } else {
                $441 = $419;
                break;
               }
              }
             }
             $420 = (($iPat5$061) + 1)|0; //@line 386 "load_amf.cpp"
             $421 = HEAP8[$123>>0]|0; //@line 386 "load_amf.cpp"
             $422 = $421&255; //@line 386 "load_amf.cpp"
             $423 = ($420>>>0)<($422>>>0); //@line 386 "load_amf.cpp"
             if ($423) {
              $389 = $441;$iPat5$061 = $420;
             } else {
              break;
             }
            }
           }
          } while(0);
          __ZdaPv($365); //@line 408 "load_amf.cpp"
          $424 = ($maxsampleseekpos$0$ph$lcssa|0)!=(0); //@line 410 "load_amf.cpp"
          $425 = ($dwMemPos$11$lcssa>>>0)<($dwMemLength>>>0); //@line 412 "load_amf.cpp"
          $or$cond2456 = $425 & $424; //@line 410 "load_amf.cpp"
          if ($or$cond2456) {
           $dwMemPos$1357 = $dwMemPos$11$lcssa;$iSeek$058 = 1;
           while(1) {
            $426 = HEAP32[$143>>2]|0; //@line 413 "load_amf.cpp"
            $iSmp7$0 = 0;
            while(1) {
             $427 = ($iSmp7$0>>>0)<($426>>>0); //@line 413 "load_amf.cpp"
             if (!($427)) {
              $dwMemPos$14 = $dwMemPos$1357;
              break;
             }
             $428 = (($sampleseekpos) + ($iSmp7$0<<2)|0); //@line 413 "load_amf.cpp"
             $429 = HEAP32[$428>>2]|0; //@line 413 "load_amf.cpp"
             $430 = ($iSeek$058|0)==($429|0); //@line 413 "load_amf.cpp"
             $431 = (($iSmp7$0) + 1)|0; //@line 415 "load_amf.cpp"
             if ($430) {
              $$lcssa173 = $431;
              label = 95;
              break;
             } else {
              $iSmp7$0 = $431;
             }
            }
            if ((label|0) == 95) {
             label = 0;
             $432 = (((($this)) + 33280|0) + ($$lcssa173<<6)|0); //@line 415 "load_amf.cpp"
             $433 = (($lpStream) + ($dwMemPos$1357)|0); //@line 416 "load_amf.cpp"
             $434 = (($dwMemLength) - ($dwMemPos$1357))|0; //@line 416 "load_amf.cpp"
             $435 = (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$432,1,$433,$434)|0); //@line 416 "load_amf.cpp"
             $436 = (($435) + ($dwMemPos$1357))|0; //@line 416 "load_amf.cpp"
             $dwMemPos$14 = $436;
            }
            $437 = (($iSeek$058) + 1)|0; //@line 410 "load_amf.cpp"
            $438 = ($437>>>0)<=($maxsampleseekpos$0$ph$lcssa>>>0); //@line 410 "load_amf.cpp"
            $439 = ($dwMemPos$14>>>0)<($dwMemLength>>>0); //@line 412 "load_amf.cpp"
            $or$cond24 = $439 & $438; //@line 410 "load_amf.cpp"
            if ($or$cond24) {
             $dwMemPos$1357 = $dwMemPos$14;$iSeek$058 = $437;
            } else {
             $$0 = 1;
             break;
            }
           }
          } else {
           $$0 = 1;
          }
         }
        }
       }
      }
     }
    }
   } else {
    $$0 = 0;
   }
  } else {
   $$0 = 0;
  }
 } else {
  $$0 = 0;
 }
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0); //@line 421 "load_amf.cpp"
}
function __Z10AMF_UnpackP11_MODCOMMANDPKhjj($pPat,$pTrack,$nRows,$nChannels) {
 $pPat = $pPat|0;
 $pTrack = $pTrack|0;
 $nRows = $nRows|0;
 $nChannels = $nChannels|0;
 var $$010 = 0, $$command$0 = 0, $$command$1 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $command$0 = 0, $command$1 = 0;
 var $command$31 = 0, $lastinstr$0$be = 0, $lastinstr$07 = 0, $or$cond = 0, $pTrack$pn = 0, $param$0 = 0, $param$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP16[$pTrack>>1]|0; //@line 59 "load_amf.cpp"
 $1 = $0&65535; //@line 59 "load_amf.cpp"
 $2 = ((($pTrack)) + 2|0); //@line 60 "load_amf.cpp"
 $3 = HEAP8[$2>>0]|0; //@line 60 "load_amf.cpp"
 $4 = $3&255; //@line 60 "load_amf.cpp"
 $5 = $4 << 16; //@line 60 "load_amf.cpp"
 $6 = $5 | $1; //@line 60 "load_amf.cpp"
 $7 = ($6|0)==(0); //@line 62 "load_amf.cpp"
 if ($7) {
  return; //@line 162 "load_amf.cpp"
 } else {
  $$in = $6;$lastinstr$07 = 0;$pTrack$pn = $pTrack;
 }
 while(1) {
  $$010 = ((($pTrack$pn)) + 3|0); //@line 61 "load_amf.cpp"
  $8 = (($$in) + -1)|0; //@line 62 "load_amf.cpp"
  $9 = HEAP8[$$010>>0]|0; //@line 64 "load_amf.cpp"
  $10 = $9&255; //@line 64 "load_amf.cpp"
  $11 = ((($pTrack$pn)) + 4|0); //@line 65 "load_amf.cpp"
  $12 = HEAP8[$11>>0]|0; //@line 65 "load_amf.cpp"
  $13 = $12&255; //@line 65 "load_amf.cpp"
  $14 = ((($pTrack$pn)) + 5|0); //@line 66 "load_amf.cpp"
  $15 = HEAP8[$14>>0]|0; //@line 66 "load_amf.cpp"
  $16 = $15&255; //@line 66 "load_amf.cpp"
  $17 = ($10>>>0)<($nRows>>>0); //@line 67 "load_amf.cpp"
  if (!($17)) {
   label = 42;
   break;
  }
  $18 = Math_imul($10, $nChannels)|0; //@line 68 "load_amf.cpp"
  $19 = (($pPat) + (($18*6)|0)|0); //@line 68 "load_amf.cpp"
  $20 = ($12&255)<(127); //@line 69 "load_amf.cpp"
  L6: do {
   if ($20) {
    $21 = (($13) + 1)|0; //@line 71 "load_amf.cpp"
    $22 = $21&255; //@line 71 "load_amf.cpp"
    HEAP8[$19>>0] = $22; //@line 71 "load_amf.cpp"
    $23 = (((($pPat) + (($18*6)|0)|0)) + 1|0); //@line 72 "load_amf.cpp"
    $24 = HEAP8[$23>>0]|0; //@line 72 "load_amf.cpp"
    $25 = ($24<<24>>24)==(0); //@line 72 "load_amf.cpp"
    if ($25) {
     $26 = $lastinstr$07&255; //@line 72 "load_amf.cpp"
     HEAP8[$23>>0] = $26; //@line 72 "load_amf.cpp"
    }
    $27 = (((($pPat) + (($18*6)|0)|0)) + 2|0); //@line 73 "load_amf.cpp"
    HEAP8[$27>>0] = 1; //@line 73 "load_amf.cpp"
    $28 = (((($pPat) + (($18*6)|0)|0)) + 4|0); //@line 74 "load_amf.cpp"
    HEAP8[$28>>0] = $15; //@line 74 "load_amf.cpp"
    $lastinstr$0$be = $lastinstr$07;
   } else {
    switch ($12<<24>>24) {
    case 127:  {
     $29 = $15 << 24 >> 24; //@line 79 "load_amf.cpp"
     $30 = (($29) + ($10))|0; //@line 79 "load_amf.cpp"
     $31 = ($30|0)>(-1); //@line 80 "load_amf.cpp"
     $32 = ($30|0)<($nRows|0); //@line 80 "load_amf.cpp"
     $or$cond = $31 & $32; //@line 80 "load_amf.cpp"
     if (!($or$cond)) {
      $lastinstr$0$be = $lastinstr$07;
      break L6;
     }
     $33 = Math_imul($30, $nChannels)|0; //@line 80 "load_amf.cpp"
     $34 = (($pPat) + (($33*6)|0)|0); //@line 80 "load_amf.cpp"
     ;HEAP8[$19>>0]=HEAP8[$34>>0]|0;HEAP8[$19+1>>0]=HEAP8[$34+1>>0]|0;HEAP8[$19+2>>0]=HEAP8[$34+2>>0]|0;HEAP8[$19+3>>0]=HEAP8[$34+3>>0]|0;HEAP8[$19+4>>0]=HEAP8[$34+4>>0]|0;HEAP8[$19+5>>0]=HEAP8[$34+5>>0]|0; //@line 80 "load_amf.cpp"
     $lastinstr$0$be = $lastinstr$07;
     break L6;
     break;
    }
    case -128:  {
     $36 = (($16) + 1)|0; //@line 84 "load_amf.cpp"
     $37 = $36&255; //@line 84 "load_amf.cpp"
     $38 = (((($pPat) + (($18*6)|0)|0)) + 1|0); //@line 84 "load_amf.cpp"
     HEAP8[$38>>0] = $37; //@line 84 "load_amf.cpp"
     $39 = $36 & 255; //@line 85 "load_amf.cpp"
     $lastinstr$0$be = $39;
     break L6;
     break;
    }
    case -125:  {
     $40 = (((($pPat) + (($18*6)|0)|0)) + 2|0); //@line 89 "load_amf.cpp"
     HEAP8[$40>>0] = 1; //@line 89 "load_amf.cpp"
     $41 = (((($pPat) + (($18*6)|0)|0)) + 4|0); //@line 90 "load_amf.cpp"
     HEAP8[$41>>0] = $15; //@line 90 "load_amf.cpp"
     $lastinstr$0$be = $lastinstr$07;
     break L6;
     break;
    }
    default: {
     $42 = $13 & 127; //@line 94 "load_amf.cpp"
     L14: do {
      switch ($42|0) {
      case 1:  {
       $command$31 = 16;$param$02 = $16;
       break;
      }
      case 2:  {
       $command$0 = 11;
       label = 15;
       break;
      }
      case 10:  {
       $command$0 = 10;
       label = 15;
       break;
      }
      case 11:  {
       $command$1 = 11;
       label = 16;
       break;
      }
      case 4:  {
       $51 = $16 & 128; //@line 110 "load_amf.cpp"
       $52 = ($51|0)==(0); //@line 110 "load_amf.cpp"
       if ($52) {
        $command$31 = 3;$param$02 = $16;
       } else {
        $53 = (0 - ($16))|0; //@line 110 "load_amf.cpp"
        $54 = $53 & 127; //@line 110 "load_amf.cpp"
        $command$31 = 2;$param$02 = $54;
       }
       break;
      }
      case 6:  {
       $command$31 = 4;$param$02 = $16;
       break;
      }
      case 7:  {
       $command$31 = 18;$param$02 = $16;
       break;
      }
      case 8:  {
       $command$31 = 1;$param$02 = $16;
       break;
      }
      case 9:  {
       $command$31 = 5;$param$02 = $16;
       break;
      }
      case 12:  {
       $command$31 = 14;$param$02 = $16;
       break;
      }
      case 13:  {
       $command$31 = 12;$param$02 = $16;
       break;
      }
      case 15:  {
       $command$31 = 15;$param$02 = $16;
       break;
      }
      case 16:  {
       $command$31 = 10;$param$02 = $16;
       break;
      }
      case 17:  {
       $55 = ($15<<24>>24)==(0); //@line 129 "load_amf.cpp"
       if ($55) {
        $lastinstr$0$be = $lastinstr$07;
        break L6;
       }
       $56 = $16 & 128; //@line 130 "load_amf.cpp"
       $57 = ($56|0)==(0); //@line 130 "load_amf.cpp"
       if ($57) {
        $61 = $16 << 4; //@line 131 "load_amf.cpp"
        $62 = $61 & 240; //@line 131 "load_amf.cpp"
        $63 = $62 | 15; //@line 131 "load_amf.cpp"
        $command$31 = 11;$param$02 = $63;
        break L14;
       } else {
        $58 = (0 - ($16))|0; //@line 130 "load_amf.cpp"
        $59 = $58 & 15; //@line 130 "load_amf.cpp"
        $60 = $59 | 240; //@line 130 "load_amf.cpp"
        $command$31 = 11;$param$02 = $60;
        break L14;
       }
       break;
      }
      case 22: case 18:  {
       $64 = ($15<<24>>24)==(0); //@line 136 "load_amf.cpp"
       if ($64) {
        $lastinstr$0$be = $lastinstr$07;
        break L6;
       }
       $65 = ($42|0)==(22); //@line 136 "load_amf.cpp"
       $66 = $65 ? 224 : 240; //@line 136 "load_amf.cpp"
       $67 = $16 & 128; //@line 137 "load_amf.cpp"
       $68 = ($67|0)==(0); //@line 137 "load_amf.cpp"
       $69 = $16 >>> 7; //@line 137 "load_amf.cpp"
       $70 = $69 ^ 3; //@line 137 "load_amf.cpp"
       if ($68) {
        $74 = $66 | $16; //@line 139 "load_amf.cpp"
        $command$31 = $70;$param$02 = $74;
        break L14;
       } else {
        $71 = (0 - ($16))|0; //@line 138 "load_amf.cpp"
        $72 = $71 & 15; //@line 138 "load_amf.cpp"
        $73 = $66 | $72; //@line 138 "load_amf.cpp"
        $command$31 = $70;$param$02 = $73;
        break L14;
       }
       break;
      }
      case 19:  {
       $75 = $16 & 15; //@line 142 "load_amf.cpp"
       $76 = $75 | 208; //@line 142 "load_amf.cpp"
       $command$31 = 20;$param$02 = $76;
       break;
      }
      case 20:  {
       $77 = $16 & 15; //@line 144 "load_amf.cpp"
       $78 = $77 | 192; //@line 144 "load_amf.cpp"
       $command$31 = 20;$param$02 = $78;
       break;
      }
      case 21:  {
       $command$31 = 17;$param$02 = $16;
       break;
      }
      case 23:  {
       $79 = (($16) + 64)|0; //@line 148 "load_amf.cpp"
       $80 = (((($pPat) + (($18*6)|0)|0)) + 3|0); //@line 149 "load_amf.cpp"
       $81 = HEAP8[$80>>0]|0; //@line 149 "load_amf.cpp"
       $82 = ($81<<24>>24)==(0); //@line 149 "load_amf.cpp"
       if ($82) {
        $lastinstr$0$be = $lastinstr$07;
        break L6;
       }
       $83 = (((($pPat) + (($18*6)|0)|0)) + 2|0); //@line 149 "load_amf.cpp"
       $84 = HEAP8[$83>>0]|0; //@line 149 "load_amf.cpp"
       $85 = ($84<<24>>24)==(0); //@line 149 "load_amf.cpp"
       if (!($85)) {
        $lastinstr$0$be = $lastinstr$07;
        break L6;
       }
       HEAP8[$83>>0] = 2; //@line 149 "load_amf.cpp"
       $86 = $79 >>> 1; //@line 149 "load_amf.cpp"
       $87 = $86 & 63; //@line 149 "load_amf.cpp"
       $88 = $87&255; //@line 149 "load_amf.cpp"
       $89 = (((($pPat) + (($18*6)|0)|0)) + 4|0); //@line 149 "load_amf.cpp"
       HEAP8[$89>>0] = $88; //@line 149 "load_amf.cpp"
       $lastinstr$0$be = $lastinstr$07;
       break L6;
       break;
      }
      default: {
       $lastinstr$0$be = $lastinstr$07;
       break L6;
      }
      }
     } while(0);
     if ((label|0) == 15) {
      label = 0;
      $43 = ($command$0|0)==(10); //@line 104 "load_amf.cpp"
      $$command$0 = $43 ? 6 : $command$0; //@line 104 "load_amf.cpp"
      $command$1 = $$command$0;
      label = 16;
     }
     if ((label|0) == 16) {
      label = 0;
      $44 = ($command$1|0)==(11); //@line 105 "load_amf.cpp"
      $$command$1 = $44 ? 7 : $command$1; //@line 105 "load_amf.cpp"
      $45 = $16 & 128; //@line 106 "load_amf.cpp"
      $46 = ($45|0)==(0); //@line 106 "load_amf.cpp"
      $47 = (0 - ($16))|0; //@line 106 "load_amf.cpp"
      $48 = $47 & 15; //@line 106 "load_amf.cpp"
      $49 = $16 << 4; //@line 107 "load_amf.cpp"
      $50 = $49 & 240; //@line 107 "load_amf.cpp"
      $param$0 = $46 ? $50 : $48; //@line 106 "load_amf.cpp"
      $command$31 = $$command$1;$param$02 = $param$0;
     }
     $90 = $command$31&255; //@line 156 "load_amf.cpp"
     $91 = (((($pPat) + (($18*6)|0)|0)) + 3|0); //@line 156 "load_amf.cpp"
     HEAP8[$91>>0] = $90; //@line 156 "load_amf.cpp"
     $92 = $param$02&255; //@line 157 "load_amf.cpp"
     $93 = (((($pPat) + (($18*6)|0)|0)) + 5|0); //@line 157 "load_amf.cpp"
     HEAP8[$93>>0] = $92; //@line 157 "load_amf.cpp"
     $lastinstr$0$be = $lastinstr$07;
     break L6;
    }
    }
   }
  } while(0);
  $35 = ($8|0)==(0); //@line 62 "load_amf.cpp"
  if ($35) {
   label = 42;
   break;
  } else {
   $$in = $8;$lastinstr$07 = $lastinstr$0$be;$pTrack$pn = $$010;
  }
 }
 if ((label|0) == 42) {
  return; //@line 162 "load_amf.cpp"
 }
}
function __ZN10CSoundFile7ReadPSMEPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$0 = 0, $$52 = 0, $$55 = 0, $$lcssa = 0, $$lcssa221 = 0, $$lcssa233 = 0, $$lcssa73$pn = 0, $$lcssa75 = 0, $$lcssa77 = 0, $$lcssa96 = 0, $$off = 0, $$off58 = 0, $$off59 = 0, $$off59$us = 0, $$off60 = 0, $$off60$us = 0, $$pre = 0, $$sum12 = 0, $$sum13 = 0, $$sum14 = 0;
 var $$sum15 = 0, $$sum16 = 0, $$sum17 = 0, $$sum18 = 0, $$sum19 = 0, $$sum20 = 0, $$sum21 = 0, $$sum22 = 0, $$sum23 = 0, $$sum24 = 0, $$sum25 = 0, $$sum26 = 0, $$sum27 = 0, $$sum28$us = 0, $$sum30$us = 0, $$sum31$us = 0, $$sum32 = 0, $$sum32$us = 0, $$sum33 = 0, $$sum33$us = 0;
 var $$sum35$us = 0, $$sum36 = 0, $$sum38 = 0, $$sum39 = 0, $$sum40 = 0, $$sum41 = 0, $$sum42 = 0, $$sum43 = 0, $$sum44 = 0, $$sum45 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0;
 var $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0;
 var $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0;
 var $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0;
 var $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0;
 var $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0;
 var $288 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $command$0 = 0;
 var $cond = 0, $cond57 = 0, $div = 0, $dwMemPos$0109 = 0, $dwMemPos$1104 = 0, $dwSongPos$0$dwMemPos$0 = 0, $dwSongPos$0$lcssa = 0, $dwSongPos$0108 = 0, $dwSongPos$1 = 0, $i$092$us = 0, $i5$097$us = 0, $i5$097$us$lcssa = 0, $iOrd$0$ph103 = 0, $len$0 = 0, $m$0$ph$be = 0, $m$0$ph85 = 0, $nPat$089 = 0, $nPatterns$0$lcssa = 0, $nPatterns$0107 = 0, $nPatterns$1 = 0;
 var $nRows$0 = 0, $note$0 = 0, $oldch$0$ph88 = 0, $oldch$082194 = 0, $or$cond = 0, $or$cond11 = 0, $or$cond3 = 0, $or$cond48 = 0, $or$cond5 = 0, $or$cond53 = 0, $or$cond56 = 0, $param$0 = 0, $patptrs = 0, $pdata$0 = 0, $pos$094 = 0, $pos$094$us = 0, $pos$2$ph = 0, $pos$2$ph67102 = 0, $pos$299$us = 0, $pos$299$us$phi = 0;
 var $pos8$0$ph86 = 0, $pos8$081$be = 0, $pos8$081195 = 0, $pos8$1 = 0, $pos8$2 = 0, $row$0$ph$be = 0, $row$0$ph87 = 0, $samplemap = 0, $switch$split12D = 0, $switch$split2D = 0, $switch$split42D = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1200|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $patptrs = sp;
 $samplemap = sp + 960|0;
 $0 = ($dwMemLength>>>0)<(256); //@line 112 "load_psm.cpp"
 if ($0) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 402 "load_psm.cpp"
 }
 $1 = HEAPU8[$lpStream>>0]|(HEAPU8[$lpStream+1>>0]<<8)|(HEAPU8[$lpStream+2>>0]<<16)|(HEAPU8[$lpStream+3>>0]<<24); //@line 113 "load_psm.cpp"
 $cond57 = ($1|0)==(541938512); //@line 113 "load_psm.cpp"
 if (!($cond57)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 402 "load_psm.cpp"
 }
 $2 = ((($lpStream)) + 4|0); //@line 120 "load_psm.cpp"
 $3 = HEAPU8[$2>>0]|(HEAPU8[$2+1>>0]<<8)|(HEAPU8[$2+2>>0]<<16)|(HEAPU8[$2+3>>0]<<24); //@line 120 "load_psm.cpp"
 $4 = (($3) + 12)|0; //@line 120 "load_psm.cpp"
 $5 = ($4>>>0)>($dwMemLength>>>0); //@line 120 "load_psm.cpp"
 if ($5) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 402 "load_psm.cpp"
 }
 $6 = ((($lpStream)) + 8|0); //@line 120 "load_psm.cpp"
 $7 = HEAPU8[$6>>0]|(HEAPU8[$6+1>>0]<<8)|(HEAPU8[$6+2>>0]<<16)|(HEAPU8[$6+3>>0]<<24); //@line 120 "load_psm.cpp"
 $8 = ($7|0)==(1162627398); //@line 120 "load_psm.cpp"
 if (!($8)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 402 "load_psm.cpp"
 }
 $9 = ((($this)) + 59680|0); //@line 121 "load_psm.cpp"
 HEAP32[$9>>2] = 4194304; //@line 121 "load_psm.cpp"
 $10 = ((($this)) + 59664|0); //@line 122 "load_psm.cpp"
 HEAP32[$10>>2] = 16; //@line 122 "load_psm.cpp"
 $11 = ((($this)) + 59684|0); //@line 123 "load_psm.cpp"
 HEAP32[$11>>2] = 0; //@line 123 "load_psm.cpp"
 $12 = ((($this)) + 49600|0); //@line 130 "load_psm.cpp"
 HEAP32[$12>>2] = 64; //@line 130 "load_psm.cpp"
 $13 = ((($this)) + 49636|0); //@line 130 "load_psm.cpp"
 HEAP32[$13>>2] = 192; //@line 130 "load_psm.cpp"
 $14 = ((($this)) + 49672|0); //@line 130 "load_psm.cpp"
 HEAP32[$14>>2] = 192; //@line 130 "load_psm.cpp"
 $15 = ((($this)) + 49708|0); //@line 130 "load_psm.cpp"
 HEAP32[$15>>2] = 64; //@line 130 "load_psm.cpp"
 $16 = ((($this)) + 49744|0); //@line 130 "load_psm.cpp"
 HEAP32[$16>>2] = 64; //@line 130 "load_psm.cpp"
 $17 = ((($this)) + 49780|0); //@line 130 "load_psm.cpp"
 HEAP32[$17>>2] = 192; //@line 130 "load_psm.cpp"
 $18 = ((($this)) + 49816|0); //@line 130 "load_psm.cpp"
 HEAP32[$18>>2] = 192; //@line 130 "load_psm.cpp"
 $19 = ((($this)) + 49852|0); //@line 130 "load_psm.cpp"
 HEAP32[$19>>2] = 64; //@line 130 "load_psm.cpp"
 $20 = ((($this)) + 49888|0); //@line 130 "load_psm.cpp"
 HEAP32[$20>>2] = 64; //@line 130 "load_psm.cpp"
 $21 = ((($this)) + 49924|0); //@line 130 "load_psm.cpp"
 HEAP32[$21>>2] = 192; //@line 130 "load_psm.cpp"
 $22 = ((($this)) + 49960|0); //@line 130 "load_psm.cpp"
 HEAP32[$22>>2] = 192; //@line 130 "load_psm.cpp"
 $23 = ((($this)) + 49996|0); //@line 130 "load_psm.cpp"
 HEAP32[$23>>2] = 64; //@line 130 "load_psm.cpp"
 $24 = ((($this)) + 50032|0); //@line 130 "load_psm.cpp"
 HEAP32[$24>>2] = 64; //@line 130 "load_psm.cpp"
 $25 = ((($this)) + 50068|0); //@line 130 "load_psm.cpp"
 HEAP32[$25>>2] = 192; //@line 130 "load_psm.cpp"
 $26 = ((($this)) + 50104|0); //@line 130 "load_psm.cpp"
 HEAP32[$26>>2] = 192; //@line 130 "load_psm.cpp"
 $27 = ((($this)) + 50140|0); //@line 130 "load_psm.cpp"
 HEAP32[$27>>2] = 64; //@line 130 "load_psm.cpp"
 $28 = (($dwMemLength) + -8)|0; //@line 136 "load_psm.cpp"
 $29 = ((($this)) + 59804|0); //@line 145 "load_psm.cpp"
 $30 = ((($this)) + 59835|0); //@line 146 "load_psm.cpp"
 $39 = 20;$dwMemPos$0109 = 12;$dwSongPos$0108 = 0;$nPatterns$0107 = 0;
 while(1) {
  $31 = (($lpStream) + ($dwMemPos$0109)|0); //@line 134 "load_psm.cpp"
  $$sum36 = (($dwMemPos$0109) + 4)|0; //@line 136 "load_psm.cpp"
  $32 = (($lpStream) + ($$sum36)|0); //@line 136 "load_psm.cpp"
  $33 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 136 "load_psm.cpp"
  $34 = ($33>>>0)<($28>>>0); //@line 136 "load_psm.cpp"
  if (!($34)) {
   $dwSongPos$0$lcssa = $dwSongPos$0108;$nPatterns$0$lcssa = $nPatterns$0107;
   break;
  }
  $35 = (($dwMemPos$0109) + 8)|0; //@line 136 "load_psm.cpp"
  $36 = (($35) + ($33))|0; //@line 136 "load_psm.cpp"
  $37 = ($36>>>0)>($dwMemLength>>>0); //@line 136 "load_psm.cpp"
  if ($37) {
   $dwSongPos$0$lcssa = $dwSongPos$0108;$nPatterns$0$lcssa = $nPatterns$0107;
   break;
  }
  $38 = (($lpStream) + ($39)|0); //@line 138 "load_psm.cpp"
  $40 = ($33|0)==(0); //@line 140 "load_psm.cpp"
  L17: do {
   if ($40) {
    $dwSongPos$1 = $dwSongPos$0108;$nPatterns$1 = $nPatterns$0107;
   } else {
    $41 = HEAPU8[$31>>0]|(HEAPU8[$31+1>>0]<<8)|(HEAPU8[$31+2>>0]<<16)|(HEAPU8[$31+3>>0]<<24); //@line 140 "load_psm.cpp"
    $switch$split2D = ($41|0)<(1280592212);
    if ($switch$split2D) {
     $switch$split12D = ($41|0)<(1196314451);
     if (!($switch$split12D)) {
      switch ($41|0) {
      case 1196314451:  {
       break;
      }
      default: {
       $dwSongPos$1 = $dwSongPos$0108;$nPatterns$1 = $nPatterns$0107;
       break L17;
      }
      }
      $52 = ($33>>>0)<(19); //@line 157 "load_psm.cpp"
      $53 = ($dwSongPos$0108|0)!=(0); //@line 157 "load_psm.cpp"
      $or$cond3 = $53 | $52; //@line 157 "load_psm.cpp"
      $dwSongPos$0$dwMemPos$0 = $or$cond3 ? $dwSongPos$0108 : $dwMemPos$0109; //@line 157 "load_psm.cpp"
      $dwSongPos$1 = $dwSongPos$0$dwMemPos$0;$nPatterns$1 = $nPatterns$0107;
      break;
     }
     switch ($41|0) {
     case 1146045008:  {
      break;
     }
     default: {
      $dwSongPos$1 = $dwSongPos$0108;$nPatterns$1 = $nPatterns$0107;
      break L17;
     }
     }
     $48 = ($33>>>0)>(11); //@line 150 "load_psm.cpp"
     $49 = ($nPatterns$0107>>>0)<(240); //@line 150 "load_psm.cpp"
     $or$cond = $49 & $48; //@line 150 "load_psm.cpp"
     if (!($or$cond)) {
      $dwSongPos$1 = $dwSongPos$0108;$nPatterns$1 = $nPatterns$0107;
      break;
     }
     $50 = (($nPatterns$0107) + 1)|0; //@line 152 "load_psm.cpp"
     $51 = (($patptrs) + ($nPatterns$0107<<2)|0); //@line 152 "load_psm.cpp"
     HEAP32[$51>>2] = $dwMemPos$0109; //@line 152 "load_psm.cpp"
     $dwSongPos$1 = $dwSongPos$0108;$nPatterns$1 = $50;
     break;
    }
    $switch$split42D = ($41|0)<(1347244868);
    if ($switch$split42D) {
     switch ($41|0) {
     case 1280592212:  {
      break;
     }
     default: {
      $dwSongPos$1 = $dwSongPos$0108;$nPatterns$1 = $nPatterns$0107;
      break L17;
     }
     }
     $42 = HEAP8[$38>>0]|0; //@line 144 "load_psm.cpp"
     $43 = ($42<<24>>24)==(0); //@line 144 "load_psm.cpp"
     $$sum45 = (($dwMemPos$0109) + 9)|0; //@line 144 "load_psm.cpp"
     $44 = (($lpStream) + ($$sum45)|0); //@line 144 "load_psm.cpp"
     $45 = (($33) + -1)|0; //@line 144 "load_psm.cpp"
     $len$0 = $43 ? $45 : $33; //@line 144 "load_psm.cpp"
     $pdata$0 = $43 ? $44 : $38; //@line 144 "load_psm.cpp"
     $46 = ($len$0>>>0)>(31); //@line 145 "load_psm.cpp"
     $47 = $46 ? 31 : $len$0; //@line 145 "load_psm.cpp"
     _memcpy(($29|0),($pdata$0|0),($47|0))|0; //@line 145 "load_psm.cpp"
     HEAP8[$30>>0] = 0; //@line 146 "load_psm.cpp"
     $dwSongPos$1 = $dwSongPos$0108;$nPatterns$1 = $nPatterns$0107;
     break;
    }
    switch ($41|0) {
    case 1347244868:  {
     break;
    }
    default: {
     $dwSongPos$1 = $dwSongPos$0108;$nPatterns$1 = $nPatterns$0107;
     break L17;
    }
    }
    $54 = ($33>>>0)>(95); //@line 164 "load_psm.cpp"
    if ($54) {
     $55 = HEAP32[$11>>2]|0; //@line 164 "load_psm.cpp"
     $56 = (($55) + 1)|0; //@line 164 "load_psm.cpp"
     $57 = ($56>>>0)<(240); //@line 164 "load_psm.cpp"
     if ($57) {
      HEAP32[$11>>2] = $56; //@line 166 "load_psm.cpp"
      $58 = (((($this)) + 33280|0) + ($56<<6)|0); //@line 167 "load_psm.cpp"
      $59 = (((($this)) + 59804|0) + ($56<<5)|0); //@line 171 "load_psm.cpp"
      $$sum38 = (($dwMemPos$0109) + 21)|0; //@line 171 "load_psm.cpp"
      $60 = (($lpStream) + ($$sum38)|0); //@line 171 "load_psm.cpp"
      dest=$59; src=$60; stop=dest+31|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 171 "load_psm.cpp"
      $61 = HEAP32[$11>>2]|0; //@line 172 "load_psm.cpp"
      $62 = (((((($this)) + 59804|0) + ($61<<5)|0)) + 31|0); //@line 172 "load_psm.cpp"
      HEAP8[$62>>0] = 0; //@line 172 "load_psm.cpp"
      $63 = $61&255; //@line 173 "load_psm.cpp"
      $64 = (($61) + -1)|0; //@line 173 "load_psm.cpp"
      $65 = (($samplemap) + ($64)|0); //@line 173 "load_psm.cpp"
      HEAP8[$65>>0] = $63; //@line 173 "load_psm.cpp"
      $66 = (((((($this)) + 33280|0) + ($56<<6)|0)) + 32|0); //@line 175 "load_psm.cpp"
      HEAP16[$66>>1] = 64; //@line 175 "load_psm.cpp"
      $$sum39 = (($dwMemPos$0109) + 81)|0; //@line 176 "load_psm.cpp"
      $67 = (($lpStream) + ($$sum39)|0); //@line 176 "load_psm.cpp"
      $68 = HEAPU8[$67>>0]|(HEAPU8[$67+1>>0]<<8)|(HEAPU8[$67+2>>0]<<16)|(HEAPU8[$67+3>>0]<<24); //@line 176 "load_psm.cpp"
      $69 = (((((($this)) + 33280|0) + ($56<<6)|0)) + 24|0); //@line 176 "load_psm.cpp"
      HEAP32[$69>>2] = $68; //@line 176 "load_psm.cpp"
      $$sum40 = (($dwMemPos$0109) + 62)|0; //@line 177 "load_psm.cpp"
      $70 = (($lpStream) + ($$sum40)|0); //@line 177 "load_psm.cpp"
      $71 = HEAPU8[$70>>0]|(HEAPU8[$70+1>>0]<<8)|(HEAPU8[$70+2>>0]<<16)|(HEAPU8[$70+3>>0]<<24); //@line 177 "load_psm.cpp"
      HEAP32[$58>>2] = $71; //@line 177 "load_psm.cpp"
      $$sum41 = (($dwMemPos$0109) + 66)|0; //@line 178 "load_psm.cpp"
      $72 = (($lpStream) + ($$sum41)|0); //@line 178 "load_psm.cpp"
      $73 = HEAPU8[$72>>0]|(HEAPU8[$72+1>>0]<<8)|(HEAPU8[$72+2>>0]<<16)|(HEAPU8[$72+3>>0]<<24); //@line 178 "load_psm.cpp"
      $74 = (((((($this)) + 33280|0) + ($56<<6)|0)) + 4|0); //@line 178 "load_psm.cpp"
      HEAP32[$74>>2] = $73; //@line 178 "load_psm.cpp"
      $$sum42 = (($dwMemPos$0109) + 70)|0; //@line 179 "load_psm.cpp"
      $75 = (($lpStream) + ($$sum42)|0); //@line 179 "load_psm.cpp"
      $76 = HEAPU8[$75>>0]|(HEAPU8[$75+1>>0]<<8)|(HEAPU8[$75+2>>0]<<16)|(HEAPU8[$75+3>>0]<<24); //@line 179 "load_psm.cpp"
      $77 = (((((($this)) + 33280|0) + ($56<<6)|0)) + 8|0); //@line 179 "load_psm.cpp"
      HEAP32[$77>>2] = $76; //@line 179 "load_psm.cpp"
      $78 = (((((($this)) + 33280|0) + ($56<<6)|0)) + 28|0); //@line 180 "load_psm.cpp"
      HEAP16[$78>>1] = 128; //@line 180 "load_psm.cpp"
      $$sum43 = (($dwMemPos$0109) + 76)|0; //@line 181 "load_psm.cpp"
      $79 = (($lpStream) + ($$sum43)|0); //@line 181 "load_psm.cpp"
      $80 = HEAP8[$79>>0]|0; //@line 181 "load_psm.cpp"
      $81 = $80&255; //@line 181 "load_psm.cpp"
      $82 = $81 << 1; //@line 181 "load_psm.cpp"
      $83 = (($82) + 2)|0; //@line 181 "load_psm.cpp"
      $84 = $83&65535; //@line 181 "load_psm.cpp"
      $85 = (((((($this)) + 33280|0) + ($56<<6)|0)) + 30|0); //@line 181 "load_psm.cpp"
      HEAP16[$85>>1] = $84; //@line 181 "load_psm.cpp"
      $86 = HEAP8[$38>>0]|0; //@line 182 "load_psm.cpp"
      $87 = ($86<<24>>24)<(0); //@line 182 "load_psm.cpp"
      $88 = $87 ? 2 : 0; //@line 182 "load_psm.cpp"
      $89 = (((((($this)) + 33280|0) + ($56<<6)|0)) + 34|0); //@line 182 "load_psm.cpp"
      HEAP16[$89>>1] = $88; //@line 182 "load_psm.cpp"
      $90 = ($73|0)==(0); //@line 183 "load_psm.cpp"
      if (!($90)) {
       $91 = (($73) + -1)|0; //@line 183 "load_psm.cpp"
       HEAP32[$74>>2] = $91; //@line 183 "load_psm.cpp"
      }
      $92 = (($33) + -96)|0; //@line 186 "load_psm.cpp"
      $93 = ($71>>>0)>(3); //@line 188 "load_psm.cpp"
      $94 = ($92>>>0)>(3); //@line 188 "load_psm.cpp"
      $or$cond5 = $94 & $93; //@line 188 "load_psm.cpp"
      if ($or$cond5) {
       $$sum44 = (($dwMemPos$0109) + 104)|0; //@line 185 "load_psm.cpp"
       $95 = (($lpStream) + ($$sum44)|0); //@line 185 "load_psm.cpp"
       (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$58,2,$95,$92)|0); //@line 190 "load_psm.cpp"
       $dwSongPos$1 = $dwSongPos$0108;$nPatterns$1 = $nPatterns$0107;
       break;
      } else {
       HEAP32[$58>>2] = 0; //@line 193 "load_psm.cpp"
       $dwSongPos$1 = $dwSongPos$0108;$nPatterns$1 = $nPatterns$0107;
       break;
      }
     } else {
      $dwSongPos$1 = $dwSongPos$0108;$nPatterns$1 = $nPatterns$0107;
     }
    } else {
     $dwSongPos$1 = $dwSongPos$0108;$nPatterns$1 = $nPatterns$0107;
    }
   }
  } while(0);
  $96 = HEAPU8[$32>>0]|(HEAPU8[$32+1>>0]<<8)|(HEAPU8[$32+2>>0]<<16)|(HEAPU8[$32+3>>0]<<24); //@line 208 "load_psm.cpp"
  $97 = (($96) + ($39))|0; //@line 208 "load_psm.cpp"
  $98 = (($97) + 8)|0; //@line 132 "load_psm.cpp"
  $99 = ($98>>>0)<($dwMemLength>>>0); //@line 132 "load_psm.cpp"
  if ($99) {
   $39 = $98;$dwMemPos$0109 = $97;$dwSongPos$0108 = $dwSongPos$1;$nPatterns$0107 = $nPatterns$1;
  } else {
   $dwSongPos$0$lcssa = $dwSongPos$1;$nPatterns$0$lcssa = $nPatterns$1;
   break;
  }
 }
 $100 = ($dwSongPos$0$lcssa|0)==(0); //@line 212 "load_psm.cpp"
 if ($100) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 402 "load_psm.cpp"
 }
 $$sum12 = (($dwSongPos$0$lcssa) + 18)|0; //@line 212 "load_psm.cpp"
 $101 = (($lpStream) + ($$sum12)|0); //@line 212 "load_psm.cpp"
 $102 = HEAP8[$101>>0]|0; //@line 212 "load_psm.cpp"
 $$off58 = (($102) + -2)<<24>>24; //@line 212 "load_psm.cpp"
 $103 = ($$off58&255)>(30); //@line 212 "load_psm.cpp"
 if ($103) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 402 "load_psm.cpp"
 }
 $104 = $102&255; //@line 212 "load_psm.cpp"
 HEAP32[$10>>2] = $104; //@line 213 "load_psm.cpp"
 $105 = (($dwSongPos$0$lcssa) + 8)|0; //@line 216 "load_psm.cpp"
 $$sum13 = (($dwSongPos$0$lcssa) + 4)|0; //@line 216 "load_psm.cpp"
 $106 = (($lpStream) + ($$sum13)|0); //@line 216 "load_psm.cpp"
 $107 = HEAP32[$106>>2]|0; //@line 216 "load_psm.cpp"
 $108 = (($105) + ($107))|0; //@line 216 "load_psm.cpp"
 $109 = (($dwSongPos$0$lcssa) + 27)|0; //@line 218 "load_psm.cpp"
 $110 = ($109>>>0)<($108>>>0); //@line 218 "load_psm.cpp"
 L49: do {
  if ($110) {
   $111 = (($dwSongPos$0$lcssa) + 19)|0; //@line 217 "load_psm.cpp"
   $112 = ($nPatterns$0$lcssa|0)==(0); //@line 237 "load_psm.cpp"
   $113 = ((($this)) + 59648|0); //@line 249 "load_psm.cpp"
   $114 = ((($this)) + 59652|0); //@line 250 "load_psm.cpp"
   $120 = $109;$dwMemPos$1104 = $111;
   while(1) {
    $$sum27 = (($dwMemPos$1104) + 4)|0; //@line 223 "load_psm.cpp"
    $116 = (($lpStream) + ($$sum27)|0); //@line 223 "load_psm.cpp"
    $117 = HEAPU8[$116>>0]|(HEAPU8[$116+1>>0]<<8)|(HEAPU8[$116+2>>0]<<16)|(HEAPU8[$116+3>>0]<<24); //@line 223 "load_psm.cpp"
    $118 = ($117>>>0)>($108>>>0); //@line 223 "load_psm.cpp"
    $119 = (($117) + ($120))|0;
    $121 = ($119>>>0)>($108>>>0); //@line 223 "load_psm.cpp"
    $or$cond48 = $118 | $121; //@line 223 "load_psm.cpp"
    if ($or$cond48) {
     break L49;
    }
    $122 = (($lpStream) + ($dwMemPos$1104)|0); //@line 220 "load_psm.cpp"
    $123 = HEAPU8[$122>>0]|(HEAPU8[$122+1>>0]<<8)|(HEAPU8[$122+2>>0]<<16)|(HEAPU8[$122+3>>0]<<24); //@line 226 "load_psm.cpp"
    $cond = ($123|0)==(1212960847); //@line 226 "load_psm.cpp"
    $124 = ($117>>>0)>(31); //@line 229 "load_psm.cpp"
    $or$cond11 = $124 & $cond; //@line 226 "load_psm.cpp"
    if ($or$cond11) {
     $125 = (($117) + -3)|0; //@line 231 "load_psm.cpp"
     $126 = ($125>>>0)>(5); //@line 232 "load_psm.cpp"
     L56: do {
      if ($126) {
       $127 = (($120) + -4)|0; //@line 246 "load_psm.cpp"
       $128 = (($120) + -2)|0; //@line 247 "load_psm.cpp"
       L58: do {
        if ($112) {
         $pos$094 = $125;
         while(1) {
          $146 = (($pos$094) + -5)|0; //@line 235 "load_psm.cpp"
          $$sum32 = (($127) + ($pos$094))|0; //@line 246 "load_psm.cpp"
          $147 = (($lpStream) + ($$sum32)|0); //@line 246 "load_psm.cpp"
          $148 = HEAP8[$147>>0]|0; //@line 246 "load_psm.cpp"
          $$off59 = (($148) + -1)<<24>>24; //@line 246 "load_psm.cpp"
          $149 = ($$off59&255)<(16); //@line 246 "load_psm.cpp"
          if ($149) {
           $$sum33 = (($128) + ($pos$094))|0; //@line 247 "load_psm.cpp"
           $150 = (($lpStream) + ($$sum33)|0); //@line 247 "load_psm.cpp"
           $151 = HEAP8[$150>>0]|0; //@line 247 "load_psm.cpp"
           $$off60 = (($151) + -65)<<24>>24; //@line 247 "load_psm.cpp"
           $152 = ($$off60&255)<(127); //@line 247 "load_psm.cpp"
           if ($152) {
            $$lcssa75 = $146;$$lcssa77 = $150;$$lcssa96 = $148;
            break L58;
           }
          }
          $153 = ($146>>>0)>(5); //@line 232 "load_psm.cpp"
          if ($153) {
           $pos$094 = $146;
          } else {
           $pos$2$ph = $146;
           break L56;
          }
         }
        } else {
         $pos$094$us = $125;
         while(1) {
          $144 = (($pos$094$us) + -5)|0; //@line 235 "load_psm.cpp"
          $$sum31$us = (($144) + ($120))|0; //@line 236 "load_psm.cpp"
          $145 = (($lpStream) + ($$sum31$us)|0); //@line 236 "load_psm.cpp"
          $142 = HEAP32[$145>>2]|0; //@line 236 "load_psm.cpp"
          $i$092$us = 0;
          while(1) {
           $137 = (($patptrs) + ($i$092$us<<2)|0); //@line 239 "load_psm.cpp"
           $138 = HEAP32[$137>>2]|0; //@line 239 "load_psm.cpp"
           $$sum35$us = (($138) + 12)|0; //@line 239 "load_psm.cpp"
           $139 = (($lpStream) + ($$sum35$us)|0); //@line 239 "load_psm.cpp"
           $140 = HEAPU8[$139>>0]|(HEAPU8[$139+1>>0]<<8)|(HEAPU8[$139+2>>0]<<16)|(HEAPU8[$139+3>>0]<<24); //@line 239 "load_psm.cpp"
           $141 = ($142|0)==($140|0); //@line 240 "load_psm.cpp"
           $130 = (($i$092$us) + 1)|0; //@line 237 "load_psm.cpp"
           if ($141) {
            break;
           }
           $129 = ($130>>>0)<($nPatterns$0$lcssa>>>0); //@line 237 "load_psm.cpp"
           if ($129) {
            $i$092$us = $130;
           } else {
            label = 32;
            break;
           }
          }
          if ((label|0) == 32) {
           label = 0;
           $$sum32$us = (($127) + ($pos$094$us))|0; //@line 246 "load_psm.cpp"
           $131 = (($lpStream) + ($$sum32$us)|0); //@line 246 "load_psm.cpp"
           $132 = HEAP8[$131>>0]|0; //@line 246 "load_psm.cpp"
           $$off59$us = (($132) + -1)<<24>>24; //@line 246 "load_psm.cpp"
           $133 = ($$off59$us&255)<(16); //@line 246 "load_psm.cpp"
           if ($133) {
            $$sum33$us = (($128) + ($pos$094$us))|0; //@line 247 "load_psm.cpp"
            $134 = (($lpStream) + ($$sum33$us)|0); //@line 247 "load_psm.cpp"
            $135 = HEAP8[$134>>0]|0; //@line 247 "load_psm.cpp"
            $$off60$us = (($135) + -65)<<24>>24; //@line 247 "load_psm.cpp"
            $136 = ($$off60$us&255)<(127); //@line 247 "load_psm.cpp"
            if ($136) {
             $$lcssa75 = $144;$$lcssa77 = $134;$$lcssa96 = $132;
             break L58;
            }
           }
          }
          $143 = ($144>>>0)>(5); //@line 232 "load_psm.cpp"
          if ($143) {
           $pos$094$us = $144;
          } else {
           $pos$2$ph = $144;
           break L56;
          }
         }
        }
       } while(0);
       $154 = $$lcssa96&255; //@line 246 "load_psm.cpp"
       HEAP32[$113>>2] = $154; //@line 249 "load_psm.cpp"
       $155 = HEAP8[$$lcssa77>>0]|0; //@line 250 "load_psm.cpp"
       $156 = $155&255; //@line 250 "load_psm.cpp"
       HEAP32[$114>>2] = $156; //@line 250 "load_psm.cpp"
       $pos$2$ph = $$lcssa75;
      } else {
       $pos$2$ph = $125;
      }
     } while(0);
     $157 = (($pos$2$ph) + 5)|0; //@line 255 "load_psm.cpp"
     $158 = ($157>>>0)<($117>>>0); //@line 255 "load_psm.cpp"
     L75: do {
      if ($158) {
       $288 = $157;$iOrd$0$ph103 = 0;$pos$2$ph67102 = $pos$2$ph;
       while(1) {
        if ($112) {
         $160 = $288;
         break;
        } else {
         $178 = $288;$pos$299$us = $pos$2$ph67102;
        }
        L78: while(1) {
         $$sum28$us = (($pos$299$us) + ($120))|0; //@line 257 "load_psm.cpp"
         $180 = (($lpStream) + ($$sum28$us)|0); //@line 257 "load_psm.cpp"
         $174 = HEAP32[$180>>2]|0; //@line 257 "load_psm.cpp"
         $i5$097$us = 0;
         while(1) {
          $169 = (($patptrs) + ($i5$097$us<<2)|0); //@line 260 "load_psm.cpp"
          $170 = HEAP32[$169>>2]|0; //@line 260 "load_psm.cpp"
          $$sum30$us = (($170) + 12)|0; //@line 260 "load_psm.cpp"
          $171 = (($lpStream) + ($$sum30$us)|0); //@line 260 "load_psm.cpp"
          $172 = HEAPU8[$171>>0]|(HEAPU8[$171+1>>0]<<8)|(HEAPU8[$171+2>>0]<<16)|(HEAPU8[$171+3>>0]<<24); //@line 260 "load_psm.cpp"
          $173 = ($174|0)==($172|0); //@line 261 "load_psm.cpp"
          if ($173) {
           $$lcssa233 = $178;$i5$097$us$lcssa = $i5$097$us;
           break L78;
          }
          $175 = (($i5$097$us) + 1)|0; //@line 258 "load_psm.cpp"
          $176 = ($175>>>0)<($nPatterns$0$lcssa>>>0); //@line 258 "load_psm.cpp"
          if ($176) {
           $i5$097$us = $175;
          } else {
           break;
          }
         }
         $177 = (($178) + 5)|0; //@line 255 "load_psm.cpp"
         $179 = ($177>>>0)<($117>>>0); //@line 255 "load_psm.cpp"
         if ($179) {
          $pos$299$us$phi = $178;$178 = $177;$pos$299$us = $pos$299$us$phi;
         } else {
          break L75;
         }
        }
        $162 = $i5$097$us$lcssa&255; //@line 263 "load_psm.cpp"
        $163 = (($iOrd$0$ph103) + 1)|0; //@line 263 "load_psm.cpp"
        $164 = (((($this)) + 53344|0) + ($iOrd$0$ph103)|0); //@line 263 "load_psm.cpp"
        HEAP8[$164>>0] = $162; //@line 263 "load_psm.cpp"
        $165 = (($$lcssa233) + 5)|0; //@line 255 "load_psm.cpp"
        $166 = ($165>>>0)<($117>>>0); //@line 255 "load_psm.cpp"
        $167 = ($163>>>0)<(256); //@line 255 "load_psm.cpp"
        $168 = $166 & $167; //@line 255 "load_psm.cpp"
        if ($168) {
         $288 = $165;$iOrd$0$ph103 = $163;$pos$2$ph67102 = $$lcssa233;
        } else {
         break L75;
        }
       }
       while(1) {
        $159 = (($160) + 5)|0; //@line 255 "load_psm.cpp"
        $161 = ($159>>>0)<($117>>>0); //@line 255 "load_psm.cpp"
        if ($161) {
         $160 = $159;
        } else {
         break;
        }
       }
      }
     } while(0);
     $$pre = HEAPU8[$116>>0]|(HEAPU8[$116+1>>0]<<8)|(HEAPU8[$116+2>>0]<<16)|(HEAPU8[$116+3>>0]<<24); //@line 272 "load_psm.cpp"
     $182 = $$pre;
    } else {
     $182 = $117;
    }
    $181 = (($182) + ($120))|0; //@line 272 "load_psm.cpp"
    $183 = (($181) + 8)|0; //@line 218 "load_psm.cpp"
    $184 = ($183>>>0)<($108>>>0); //@line 218 "load_psm.cpp"
    if ($184) {
     $120 = $183;$dwMemPos$1104 = $181;
    } else {
     break;
    }
   }
  }
 } while(0);
 $115 = ($nPatterns$0$lcssa|0)==(0); //@line 277 "load_psm.cpp"
 if ($115) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 402 "load_psm.cpp"
 } else {
  $nPat$089 = 0;
 }
 while(1) {
  $185 = (($patptrs) + ($nPat$089<<2)|0); //@line 279 "load_psm.cpp"
  $186 = HEAP32[$185>>2]|0; //@line 279 "load_psm.cpp"
  $$sum14 = (($186) + 8)|0; //@line 279 "load_psm.cpp"
  $187 = (($lpStream) + ($$sum14)|0); //@line 279 "load_psm.cpp"
  $$sum15 = (($186) + 4)|0; //@line 281 "load_psm.cpp"
  $188 = (($lpStream) + ($$sum15)|0); //@line 281 "load_psm.cpp"
  $189 = HEAP32[$188>>2]|0; //@line 281 "load_psm.cpp"
  $190 = (($189) + -12)|0; //@line 281 "load_psm.cpp"
  $$sum16 = (($186) + 16)|0; //@line 282 "load_psm.cpp"
  $191 = (($lpStream) + ($$sum16)|0); //@line 282 "load_psm.cpp"
  $192 = HEAPU8[$191>>0]|(HEAPU8[$191+1>>0]<<8); //@line 282 "load_psm.cpp"
  $193 = $192&65535; //@line 282 "load_psm.cpp"
  $194 = HEAPU8[$187>>0]|(HEAPU8[$187+1>>0]<<8)|(HEAPU8[$187+2>>0]<<16)|(HEAPU8[$187+3>>0]<<24); //@line 283 "load_psm.cpp"
  $195 = ($190>>>0)>($194>>>0); //@line 283 "load_psm.cpp"
  $$52 = $195 ? $194 : $190; //@line 283 "load_psm.cpp"
  $$off = (($192) + -64)<<16>>16; //@line 284 "load_psm.cpp"
  $196 = ($$off&65535)>(192); //@line 284 "load_psm.cpp"
  $nRows$0 = $196 ? 64 : $193; //@line 284 "load_psm.cpp"
  $197 = $nRows$0&65535; //@line 285 "load_psm.cpp"
  $198 = (((($this)) + 52864|0) + ($nPat$089<<1)|0); //@line 285 "load_psm.cpp"
  HEAP16[$198>>1] = $197; //@line 285 "load_psm.cpp"
  $199 = HEAP32[$10>>2]|0; //@line 286 "load_psm.cpp"
  $200 = (__ZN10CSoundFile15AllocatePatternEjj($nRows$0,$199)|0); //@line 286 "load_psm.cpp"
  $201 = (((($this)) + 51904|0) + ($nPat$089<<2)|0); //@line 286 "load_psm.cpp"
  HEAP32[$201>>2] = $200; //@line 286 "load_psm.cpp"
  $202 = ($200|0)==(0|0); //@line 286 "load_psm.cpp"
  if ($202) {
   $$0 = 1;
   label = 82;
   break;
  }
  $$sum17 = (($186) + 20)|0; //@line 288 "load_psm.cpp"
  $203 = ($nRows$0|0)==(0); //@line 296 "load_psm.cpp"
  L95: do {
   if (!($203)) {
    $m$0$ph85 = $200;$oldch$0$ph88 = 0;$pos8$0$ph86 = 0;$row$0$ph87 = 0;
    while(1) {
     $204 = (($pos8$0$ph86) + 1)|0; //@line 296 "load_psm.cpp"
     $205 = ($204>>>0)<($$52>>>0); //@line 296 "load_psm.cpp"
     if ($205) {
      $210 = $204;$oldch$082194 = $oldch$0$ph88;$pos8$081195 = $pos8$0$ph86;
     } else {
      break L95;
     }
     while(1) {
      $$sum18 = (($pos8$081195) + ($$sum17))|0; //@line 298 "load_psm.cpp"
      $206 = (($lpStream) + ($$sum18)|0); //@line 298 "load_psm.cpp"
      $207 = HEAP8[$206>>0]|0; //@line 298 "load_psm.cpp"
      $208 = $207&255; //@line 298 "load_psm.cpp"
      $209 = (($pos8$081195) + 2)|0; //@line 299 "load_psm.cpp"
      $$sum19 = (($210) + ($$sum17))|0; //@line 299 "load_psm.cpp"
      $211 = (($lpStream) + ($$sum19)|0); //@line 299 "load_psm.cpp"
      $212 = HEAP8[$211>>0]|0; //@line 299 "load_psm.cpp"
      $213 = $212&255; //@line 299 "load_psm.cpp"
      $214 = $208 & 240; //@line 304 "load_psm.cpp"
      $215 = ($214|0)!=(16); //@line 304 "load_psm.cpp"
      $216 = ($213>>>0)>($oldch$082194>>>0); //@line 304 "load_psm.cpp"
      $or$cond53 = $215 | $216; //@line 304 "load_psm.cpp"
      if (!($or$cond53)) {
       $217 = (($pos8$081195) + 3)|0; //@line 306 "load_psm.cpp"
       $218 = ($217>>>0)<($$52>>>0); //@line 306 "load_psm.cpp"
       if ($218) {
        $$sum25 = (($209) + ($$sum17))|0; //@line 306 "load_psm.cpp"
        $219 = (($lpStream) + ($$sum25)|0); //@line 306 "load_psm.cpp"
        $220 = HEAP8[$219>>0]|0; //@line 306 "load_psm.cpp"
        $221 = $220 & 15; //@line 306 "load_psm.cpp"
        $222 = ($221<<24>>24)==(0); //@line 306 "load_psm.cpp"
        if ($222) {
         $$sum26 = (($217) + ($$sum17))|0; //@line 306 "load_psm.cpp"
         $223 = (($lpStream) + ($$sum26)|0); //@line 306 "load_psm.cpp"
         $224 = HEAP8[$223>>0]|0; //@line 306 "load_psm.cpp"
         $225 = $224&255; //@line 306 "load_psm.cpp"
         $226 = HEAP32[$10>>2]|0; //@line 306 "load_psm.cpp"
         $227 = ($225>>>0)<($226>>>0); //@line 306 "load_psm.cpp"
         if ($227) {
          $$lcssa = $209;$$lcssa221 = $213;$$lcssa73$pn = $226;
          break;
         }
        }
       }
      }
      $228 = ($209>>>0)<($$52>>>0); //@line 317 "load_psm.cpp"
      if (!($228)) {
       break L95;
      }
      $229 = ($214|0)==(0); //@line 318 "load_psm.cpp"
      $230 = HEAP32[$10>>2]|0; //@line 324 "load_psm.cpp"
      if ($229) {
       $$lcssa = $209;$$lcssa221 = $213;$$lcssa73$pn = $230;
       break;
      }
      $232 = ($213>>>0)<($230>>>0); //@line 330 "load_psm.cpp"
      $$55 = $232 ? $213 : 0; //@line 330 "load_psm.cpp"
      $233 = $208 & 64; //@line 338 "load_psm.cpp"
      $234 = ($233|0)==(0); //@line 338 "load_psm.cpp"
      do {
       if ($234) {
        $pos8$1 = $209;
       } else {
        $235 = (($pos8$081195) + 3)|0; //@line 338 "load_psm.cpp"
        $236 = ($235>>>0)<($$52>>>0); //@line 338 "load_psm.cpp"
        if (!($236)) {
         $pos8$1 = $209;
         break;
        }
        $$sum23 = (($209) + ($$sum17))|0; //@line 340 "load_psm.cpp"
        $237 = (($lpStream) + ($$sum23)|0); //@line 340 "load_psm.cpp"
        $238 = HEAP8[$237>>0]|0; //@line 340 "load_psm.cpp"
        $239 = $238&255; //@line 340 "load_psm.cpp"
        $240 = (($pos8$081195) + 4)|0; //@line 341 "load_psm.cpp"
        $$sum24 = (($235) + ($$sum17))|0; //@line 341 "load_psm.cpp"
        $241 = (($lpStream) + ($$sum24)|0); //@line 341 "load_psm.cpp"
        $242 = HEAP8[$241>>0]|0; //@line 341 "load_psm.cpp"
        $243 = $242&255; //@line 341 "load_psm.cpp"
        $244 = ($238<<24>>24)>(0); //@line 346 "load_psm.cpp"
        if ($244) {
         $245 = $239 >>> 4; //@line 346 "load_psm.cpp"
         $246 = ($245*12)|0; //@line 346 "load_psm.cpp"
         $247 = $239 & 15; //@line 346 "load_psm.cpp"
         $248 = (($247) + 13)|0; //@line 346 "load_psm.cpp"
         $249 = (($248) + ($246))|0; //@line 346 "load_psm.cpp"
         $note$0 = $249;
        } else {
         $note$0 = $239;
        }
        $250 = (($samplemap) + ($243)|0); //@line 347 "load_psm.cpp"
        $251 = HEAP8[$250>>0]|0; //@line 347 "load_psm.cpp"
        $252 = (((($m$0$ph85) + (($$55*6)|0)|0)) + 1|0); //@line 347 "load_psm.cpp"
        HEAP8[$252>>0] = $251; //@line 347 "load_psm.cpp"
        $253 = $note$0&255; //@line 348 "load_psm.cpp"
        $254 = (($m$0$ph85) + (($$55*6)|0)|0); //@line 348 "load_psm.cpp"
        HEAP8[$254>>0] = $253; //@line 348 "load_psm.cpp"
        $pos8$1 = $240;
       }
      } while(0);
      $255 = $208 & 32; //@line 351 "load_psm.cpp"
      $256 = ($255|0)!=(0); //@line 351 "load_psm.cpp"
      $257 = ($pos8$1>>>0)<($$52>>>0); //@line 351 "load_psm.cpp"
      $or$cond56 = $256 & $257; //@line 351 "load_psm.cpp"
      if ($or$cond56) {
       $258 = (((($m$0$ph85) + (($$55*6)|0)|0)) + 2|0); //@line 353 "load_psm.cpp"
       HEAP8[$258>>0] = 1; //@line 353 "load_psm.cpp"
       $259 = (($pos8$1) + 1)|0; //@line 354 "load_psm.cpp"
       $$sum22 = (($pos8$1) + ($$sum17))|0; //@line 354 "load_psm.cpp"
       $260 = (($lpStream) + ($$sum22)|0); //@line 354 "load_psm.cpp"
       $261 = HEAP8[$260>>0]|0; //@line 354 "load_psm.cpp"
       $div = ($261&255) >>> 1; //@line 354 "load_psm.cpp"
       $262 = (((($m$0$ph85) + (($$55*6)|0)|0)) + 4|0); //@line 354 "load_psm.cpp"
       HEAP8[$262>>0] = $div; //@line 354 "load_psm.cpp"
       $pos8$2 = $259;
      } else {
       $pos8$2 = $pos8$1;
      }
      $263 = $208 & 16; //@line 357 "load_psm.cpp"
      $264 = ($263|0)==(0); //@line 357 "load_psm.cpp"
      do {
       if ($264) {
        $pos8$081$be = $pos8$2;
       } else {
        $267 = (($pos8$2) + 1)|0; //@line 357 "load_psm.cpp"
        $268 = ($267>>>0)<($$52>>>0); //@line 357 "load_psm.cpp"
        if (!($268)) {
         $pos8$081$be = $pos8$2;
         break;
        }
        $$sum20 = (($pos8$2) + ($$sum17))|0; //@line 359 "load_psm.cpp"
        $269 = (($lpStream) + ($$sum20)|0); //@line 359 "load_psm.cpp"
        $270 = HEAP8[$269>>0]|0; //@line 359 "load_psm.cpp"
        $271 = $270&255; //@line 359 "load_psm.cpp"
        $272 = (($pos8$2) + 2)|0; //@line 360 "load_psm.cpp"
        $$sum21 = (($267) + ($$sum17))|0; //@line 360 "load_psm.cpp"
        $273 = (($lpStream) + ($$sum21)|0); //@line 360 "load_psm.cpp"
        $274 = HEAP8[$273>>0]|0; //@line 360 "load_psm.cpp"
        $275 = $274&255; //@line 360 "load_psm.cpp"
        switch ($271|0) {
        case 1:  {
         $276 = $275 | 15; //@line 365 "load_psm.cpp"
         $command$0 = 11;$param$0 = $276;
         break;
        }
        case 4:  {
         $277 = $275 >>> 4; //@line 367 "load_psm.cpp"
         $278 = $277 | 240; //@line 367 "load_psm.cpp"
         $command$0 = 11;$param$0 = $278;
         break;
        }
        case 12:  {
         $279 = (($275) + 1)|0; //@line 369 "load_psm.cpp"
         $280 = $279 >>> 1; //@line 369 "load_psm.cpp"
         $command$0 = 2;$param$0 = $280;
         break;
        }
        case 14:  {
         $281 = (($275) + 1)|0; //@line 371 "load_psm.cpp"
         $282 = $281 >>> 1; //@line 371 "load_psm.cpp"
         $command$0 = 3;$param$0 = $282;
         break;
        }
        case 51:  {
         $command$0 = 12;$param$0 = $275;
         break;
        }
        case 52:  {
         $command$0 = 14;$param$0 = $275;
         break;
        }
        case 61:  {
         $command$0 = 16;$param$0 = $275;
         break;
        }
        case 62:  {
         $command$0 = 17;$param$0 = $275;
         break;
        }
        default: {
         $command$0 = 0;$param$0 = 0;
        }
        }
        $283 = (((($m$0$ph85) + (($$55*6)|0)|0)) + 3|0); //@line 387 "load_psm.cpp"
        HEAP8[$283>>0] = $command$0; //@line 387 "load_psm.cpp"
        $284 = $param$0&255; //@line 388 "load_psm.cpp"
        $285 = (((($m$0$ph85) + (($$55*6)|0)|0)) + 5|0); //@line 388 "load_psm.cpp"
        HEAP8[$285>>0] = $284; //@line 388 "load_psm.cpp"
        $pos8$081$be = $272;
       }
      } while(0);
      $265 = (($pos8$081$be) + 1)|0; //@line 296 "load_psm.cpp"
      $266 = ($265>>>0)<($$52>>>0); //@line 296 "load_psm.cpp"
      if ($266) {
       $210 = $265;$oldch$082194 = $$55;$pos8$081195 = $pos8$081$be;
      } else {
       break L95;
      }
     }
     $row$0$ph$be = (($row$0$ph87) + 1)|0; //@line 311 "load_psm.cpp"
     $m$0$ph$be = (($m$0$ph85) + (($$lcssa73$pn*6)|0)|0); //@line 312 "load_psm.cpp"
     $231 = ($row$0$ph$be>>>0)<($nRows$0>>>0); //@line 296 "load_psm.cpp"
     if ($231) {
      $m$0$ph85 = $m$0$ph$be;$oldch$0$ph88 = $$lcssa221;$pos8$0$ph86 = $$lcssa;$row$0$ph87 = $row$0$ph$be;
     } else {
      break;
     }
    }
   }
  } while(0);
  $286 = (($nPat$089) + 1)|0; //@line 277 "load_psm.cpp"
  $287 = ($286>>>0)<($nPatterns$0$lcssa>>>0); //@line 277 "load_psm.cpp"
  if ($287) {
   $nPat$089 = $286;
  } else {
   $$0 = 1;
   label = 82;
   break;
  }
 }
 if ((label|0) == 82) {
  STACKTOP = sp;return ($$0|0); //@line 402 "load_psm.cpp"
 }
 return (0)|0;
}
function __ZN10CSoundFile7ReadMT2EPKhj($this,$lpStream,$dwMemLength) {
 $this = $this|0;
 $lpStream = $lpStream|0;
 $dwMemLength = $dwMemLength|0;
 var $$ = 0, $$0 = 0, $$48 = 0, $$49 = 0, $$50 = 0, $$56 = 0, $$57 = 0, $$lobit = 0, $$neg62 = 0, $$off = 0, $$off59 = 0, $$off60 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre158 = 0, $$pre160 = 0, $$sum = 0, $$sum12 = 0, $$sum13 = 0, $$sum15 = 0;
 var $$sum16 = 0, $$sum17 = 0, $$sum18 = 0, $$sum19 = 0, $$sum20 = 0, $$sum21 = 0, $$sum22 = 0, $$sum23 = 0, $$sum24 = 0, $$sum25 = 0, $$sum26 = 0, $$sum26$1 = 0, $$sum26$2 = 0, $$sum26$3 = 0, $$sum27 = 0, $$sum28 = 0, $$sum29 = 0, $$sum30 = 0, $$sum31 = 0, $$sum32 = 0;
 var $$sum33 = 0, $$sum34 = 0, $$sum35 = 0, $$sum36 = 0, $$sum37 = 0, $$sum38 = 0, $$sum39 = 0, $$sum40 = 0, $$sum41 = 0, $$sum42 = 0, $$sum43 = 0, $$sum44 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0;
 var $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0;
 var $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0;
 var $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0;
 var $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0;
 var $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0;
 var $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0;
 var $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0;
 var $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0;
 var $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0;
 var $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0;
 var $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0;
 var $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0;
 var $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0;
 var $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0;
 var $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0;
 var $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0;
 var $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0;
 var $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0;
 var $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0;
 var $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0;
 var $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0;
 var $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0;
 var $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0;
 var $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0;
 var $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $InstrMap = 0, $SampleMap = 0, $ch$0 = 0, $ch$1$lcssa = 0, $ch$1120 = 0, $cmd = 0, $cond = 0;
 var $dwEnvPos3$1 = 0, $dwEnvPos3$1$1 = 0, $dwEnvPos3$1$2 = 0, $dwMemPos$1 = 0, $dwMemPos$11 = 0, $dwMemPos$12109 = 0, $dwMemPos$13 = 0, $dwMemPos$13$lcssa = 0, $dwMemPos$14102 = 0, $dwMemPos$15 = 0, $dwMemPos$1698 = 0, $dwMemPos$1796 = 0, $dwMemPos$1796$us = 0, $dwMemPos$18 = 0, $dwMemPos$1993 = 0, $dwMemPos$20 = 0, $dwMemPos$21 = 0, $dwMemPos$2141 = 0, $dwMemPos$3133 = 0, $dwMemPos$4 = 0;
 var $dwMemPos$5 = 0, $dwMemPos$6117 = 0, $dwMemPos$7 = 0, $dwMemPos$8113 = 0, $dwMemPos$9$ = 0, $dwMemPos$9$$1 = 0, $dwMemPos$9$$10 = 0, $dwMemPos$9$$11 = 0, $dwMemPos$9$$12 = 0, $dwMemPos$9$$13 = 0, $dwMemPos$9$$2 = 0, $dwMemPos$9$$3 = 0, $dwMemPos$9$$4 = 0, $dwMemPos$9$$5 = 0, $dwMemPos$9$$6 = 0, $dwMemPos$9$$7 = 0, $dwMemPos$9$$8 = 0, $dwMemPos$9$$9 = 0, $exitcond = 0, $exitcond150 = 0;
 var $exitcond153 = 0, $exitcond154 = 0, $exitcond156 = 0, $i$0103 = 0, $i6$0106 = 0, $i9$094 = 0, $iAuto$0112 = 0, $iData$092 = 0, $iDrm$0116 = 0, $iEnv5$0107 = 0, $iGrp$095 = 0, $iGrp$095$us = 0, $iIns$0108 = 0, $iMap$097 = 0, $iOrd$0143 = 0, $iPat$0132 = 0, $iSmp$099 = 0, $infobyte$0 = 0, $len$0127 = 0, $m$0128 = 0;
 var $n$0126 = 0, $nAutoCount$1 = 0, $or$cond = 0, $or$cond11 = 0, $or$cond175 = 0, $or$cond52 = 0, $or$cond7 = 0, $p1$0125 = 0, $pEnvData$0 = 0, $pEnvPoints$0 = 0, $pdd$0 = 0, $pedata = 0, $pehdr = 0, $penv$0 = 0, $penv8$0 = 0, $pos$0 = 0, $pos$1 = 0, $pos$2 = 0, $pos$3 = 0, $pos$4 = 0;
 var $pos$5 = 0, $pos$6 = 0, $pos$7 = 0, $pos$8 = 0, $pos$9 = 0, $row$0 = 0, $row$1$lcssa = 0, $row$1121 = 0, $rptcount$0 = 0, $rsflags$0 = 0, $rsflags$0$v = 0, $sext = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 2096|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $InstrMap = sp + 1056|0;
 $SampleMap = sp + 32|0;
 $cmd = sp + 2076|0;
 $pehdr = sp + 16|0;
 $pedata = sp;
 $0 = ($lpStream|0)==(0|0); //@line 193 "load_mt2.cpp"
 $1 = ($dwMemLength>>>0)<(382); //@line 193 "load_mt2.cpp"
 $or$cond = $0 | $1; //@line 193 "load_mt2.cpp"
 if ($or$cond) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 635 "load_mt2.cpp"
 }
 $2 = HEAPU8[$lpStream>>0]|(HEAPU8[$lpStream+1>>0]<<8)|(HEAPU8[$lpStream+2>>0]<<16)|(HEAPU8[$lpStream+3>>0]<<24); //@line 194 "load_mt2.cpp"
 $3 = ($2|0)==(808604749); //@line 194 "load_mt2.cpp"
 if (!($3)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 635 "load_mt2.cpp"
 }
 $4 = ((($lpStream)) + 8|0); //@line 195 "load_mt2.cpp"
 $5 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8); //@line 195 "load_mt2.cpp"
 $6 = $5 & -256; //@line 195 "load_mt2.cpp"
 $7 = ($6<<16>>16)==(512); //@line 195 "load_mt2.cpp"
 if (!($7)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 635 "load_mt2.cpp"
 }
 $8 = ((($lpStream)) + 112|0); //@line 196 "load_mt2.cpp"
 $9 = HEAPU8[$8>>0]|(HEAPU8[$8+1>>0]<<8); //@line 196 "load_mt2.cpp"
 $$off59 = (($9) + -4)<<16>>16; //@line 196 "load_mt2.cpp"
 $10 = ($$off59&65535)>(60); //@line 196 "load_mt2.cpp"
 if ($10) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0); //@line 635 "load_mt2.cpp"
 }
 $11 = ((($this)) + 59680|0); //@line 198 "load_mt2.cpp"
 HEAP32[$11>>2] = 1048576; //@line 198 "load_mt2.cpp"
 $12 = $9&65535; //@line 199 "load_mt2.cpp"
 $13 = ((($this)) + 59664|0); //@line 199 "load_mt2.cpp"
 HEAP32[$13>>2] = $12; //@line 199 "load_mt2.cpp"
 $14 = ((($lpStream)) + 108|0); //@line 200 "load_mt2.cpp"
 $15 = HEAPU8[$14>>0]|(HEAPU8[$14+1>>0]<<8); //@line 200 "load_mt2.cpp"
 $16 = $15&65535; //@line 200 "load_mt2.cpp"
 $17 = ((($this)) + 59736|0); //@line 200 "load_mt2.cpp"
 HEAP32[$17>>2] = $16; //@line 200 "load_mt2.cpp"
 $18 = ((($lpStream)) + 116|0); //@line 201 "load_mt2.cpp"
 $19 = HEAP8[$18>>0]|0; //@line 201 "load_mt2.cpp"
 $20 = $19&255; //@line 201 "load_mt2.cpp"
 $21 = ((($this)) + 59648|0); //@line 201 "load_mt2.cpp"
 HEAP32[$21>>2] = $20; //@line 201 "load_mt2.cpp"
 $22 = ((($this)) + 59652|0); //@line 202 "load_mt2.cpp"
 HEAP32[$22>>2] = 125; //@line 202 "load_mt2.cpp"
 $23 = ((($lpStream)) + 114|0); //@line 203 "load_mt2.cpp"
 $24 = HEAPU8[$23>>0]|(HEAPU8[$23+1>>0]<<8); //@line 203 "load_mt2.cpp"
 $$off60 = (($24) + -101)<<16>>16; //@line 203 "load_mt2.cpp"
 $25 = ($$off60&65535)<(4899); //@line 203 "load_mt2.cpp"
 if ($25) {
  $26 = $24&65535; //@line 203 "load_mt2.cpp"
  $27 = (110250 / ($26>>>0))&-1; //@line 205 "load_mt2.cpp"
  HEAP32[$22>>2] = $27; //@line 205 "load_mt2.cpp"
 }
 $28 = ((($lpStream)) + 106|0); //@line 209 "load_mt2.cpp"
 $iOrd$0143 = 0;
 while(1) {
  $29 = HEAPU8[$28>>0]|(HEAPU8[$28+1>>0]<<8); //@line 209 "load_mt2.cpp"
  $30 = $29&65535; //@line 209 "load_mt2.cpp"
  $31 = ($iOrd$0143>>>0)<($30>>>0); //@line 209 "load_mt2.cpp"
  if ($31) {
   $$sum44 = (($iOrd$0143) + 126)|0; //@line 209 "load_mt2.cpp"
   $32 = (($lpStream) + ($$sum44)|0); //@line 209 "load_mt2.cpp"
   $33 = HEAP8[$32>>0]|0; //@line 209 "load_mt2.cpp"
   $34 = $33&255; //@line 209 "load_mt2.cpp"
   $36 = $34;
  } else {
   $36 = 255;
  }
  $35 = $36&255; //@line 209 "load_mt2.cpp"
  $37 = (((($this)) + 53344|0) + ($iOrd$0143)|0); //@line 209 "load_mt2.cpp"
  HEAP8[$37>>0] = $35; //@line 209 "load_mt2.cpp"
  $38 = (($iOrd$0143) + 1)|0; //@line 207 "load_mt2.cpp"
  $exitcond156 = ($38|0)==(256); //@line 207 "load_mt2.cpp"
  if ($exitcond156) {
   break;
  } else {
   $iOrd$0143 = $38;
  }
 }
 $39 = ((($this)) + 59804|0); //@line 211 "load_mt2.cpp"
 $40 = ((($lpStream)) + 42|0); //@line 211 "load_mt2.cpp"
 dest=$39; src=$40; stop=dest+31|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 211 "load_mt2.cpp"
 $41 = ((($this)) + 59835|0); //@line 212 "load_mt2.cpp"
 HEAP8[$41>>0] = 0; //@line 212 "load_mt2.cpp"
 $42 = ((($lpStream)) + 382|0); //@line 214 "load_mt2.cpp"
 $43 = HEAP16[$42>>1]|0; //@line 214 "load_mt2.cpp"
 $44 = $43&65535; //@line 214 "load_mt2.cpp"
 $45 = ($43&65535)>(1); //@line 216 "load_mt2.cpp"
 $46 = ((($lpStream)) + 384|0); //@line 216 "load_mt2.cpp"
 $pdd$0 = $45 ? $46 : 0; //@line 216 "load_mt2.cpp"
 $47 = (($44) + 384)|0; //@line 217 "load_mt2.cpp"
 $48 = (($dwMemLength) + -12)|0; //@line 224 "load_mt2.cpp"
 $49 = ($47>>>0)<($48>>>0); //@line 224 "load_mt2.cpp"
 if (!($49)) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 635 "load_mt2.cpp"
 }
 $50 = (($lpStream) + ($47)|0); //@line 225 "load_mt2.cpp"
 $51 = HEAP32[$50>>2]|0; //@line 225 "load_mt2.cpp"
 $52 = ($51|0)==(0); //@line 225 "load_mt2.cpp"
 $53 = (($44) + 388)|0; //@line 225 "load_mt2.cpp"
 $$48 = $52 ? $53 : $47; //@line 225 "load_mt2.cpp"
 $54 = (($lpStream) + ($$48)|0); //@line 226 "load_mt2.cpp"
 $55 = HEAP32[$54>>2]|0; //@line 226 "load_mt2.cpp"
 $56 = ($55|0)==(0); //@line 226 "load_mt2.cpp"
 $57 = (($$48) + 4)|0; //@line 226 "load_mt2.cpp"
 $dwMemPos$1 = $56 ? $57 : $$48; //@line 226 "load_mt2.cpp"
 $58 = (($lpStream) + ($dwMemPos$1)|0); //@line 227 "load_mt2.cpp"
 $59 = HEAP32[$58>>2]|0; //@line 227 "load_mt2.cpp"
 $60 = (($dwMemPos$1) + 4)|0; //@line 228 "load_mt2.cpp"
 $61 = (($60) + ($59))|0; //@line 233 "load_mt2.cpp"
 $62 = ($61>>>0)<($dwMemLength>>>0); //@line 233 "load_mt2.cpp"
 if (!($62)) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0); //@line 635 "load_mt2.cpp"
 }
 $63 = (($dwMemPos$1) + 12)|0; //@line 234 "load_mt2.cpp"
 $64 = ($63>>>0)<($61>>>0); //@line 234 "load_mt2.cpp"
 L28: do {
  if ($64) {
   $65 = ((($this)) + 59796|0); //@line 250 "load_mt2.cpp"
   $80 = $63;$dwMemPos$2141 = $60;
   while(1) {
    $$sum42 = (($dwMemPos$2141) + 4)|0; //@line 237 "load_mt2.cpp"
    $77 = (($lpStream) + ($$sum42)|0); //@line 237 "load_mt2.cpp"
    $78 = HEAP32[$77>>2]|0; //@line 237 "load_mt2.cpp"
    $79 = (($78) + ($80))|0; //@line 239 "load_mt2.cpp"
    $81 = ($79>>>0)>($dwMemLength>>>0); //@line 239 "load_mt2.cpp"
    if ($81) {
     $$0 = 1;
     break;
    }
    $82 = (($lpStream) + ($dwMemPos$2141)|0); //@line 236 "load_mt2.cpp"
    $83 = HEAP32[$82>>2]|0; //@line 236 "load_mt2.cpp"
    $cond = ($83|0)==(4674381); //@line 246 "load_mt2.cpp"
    $84 = ($78>>>0)>(3); //@line 250 "load_mt2.cpp"
    $or$cond11 = $84 & $cond; //@line 246 "load_mt2.cpp"
    if ($or$cond11) {
     $85 = HEAP32[$65>>2]|0; //@line 250 "load_mt2.cpp"
     $86 = ($85|0)==(0|0); //@line 250 "load_mt2.cpp"
     if ($86) {
      $89 = ($78>>>0)>(32000); //@line 253 "load_mt2.cpp"
      $$ = $89 ? 32000 : $78; //@line 253 "load_mt2.cpp"
      $90 = (__Znaj($$)|0); //@line 254 "load_mt2.cpp"
      HEAP32[$65>>2] = $90; //@line 254 "load_mt2.cpp"
      $$sum43 = (($dwMemPos$2141) + 9)|0; //@line 257 "load_mt2.cpp"
      $91 = (($lpStream) + ($$sum43)|0); //@line 257 "load_mt2.cpp"
      $92 = (($$) + -1)|0; //@line 257 "load_mt2.cpp"
      _memcpy(($90|0),($91|0),($92|0))|0; //@line 257 "load_mt2.cpp"
      $93 = (($90) + ($92)|0); //@line 258 "load_mt2.cpp"
      HEAP8[$93>>0] = 0; //@line 258 "load_mt2.cpp"
     }
    }
    $87 = (($79) + 8)|0; //@line 234 "load_mt2.cpp"
    $88 = ($87>>>0)<($61>>>0); //@line 234 "load_mt2.cpp"
    if ($88) {
     $80 = $87;$dwMemPos$2141 = $79;
    } else {
     break L28;
    }
   }
   STACKTOP = sp;return ($$0|0); //@line 635 "load_mt2.cpp"
  }
 } while(0);
 $66 = ((($lpStream)) + 110|0); //@line 272 "load_mt2.cpp"
 $67 = HEAPU8[$66>>0]|(HEAPU8[$66+1>>0]<<8); //@line 272 "load_mt2.cpp"
 $68 = ($67<<16>>16)==(0); //@line 272 "load_mt2.cpp"
 L39: do {
  if ($68) {
   $217 = 0;$dwMemPos$5 = $61;
  } else {
   $69 = (($dwMemLength) + -6)|0; //@line 272 "load_mt2.cpp"
   $70 = ((($lpStream)) + 118|0); //@line 289 "load_mt2.cpp"
   $71 = ((($cmd)) + 6|0); //@line 312 "load_mt2.cpp"
   $72 = ((($cmd)) + 5|0); //@line 312 "load_mt2.cpp"
   $73 = ((($cmd)) + 4|0); //@line 312 "load_mt2.cpp"
   $74 = ((($cmd)) + 3|0); //@line 312 "load_mt2.cpp"
   $75 = ((($cmd)) + 2|0); //@line 312 "load_mt2.cpp"
   $76 = ((($cmd)) + 1|0); //@line 312 "load_mt2.cpp"
   $638 = $67;$dwMemPos$3133 = $61;$iPat$0132 = 0;
   L41: while(1) {
    $94 = ($dwMemPos$3133>>>0)<($69>>>0); //@line 272 "load_mt2.cpp"
    L43: do {
     if ($94) {
      $$sum31 = (($dwMemPos$3133) + 2)|0; //@line 275 "load_mt2.cpp"
      $95 = (($lpStream) + ($$sum31)|0); //@line 275 "load_mt2.cpp"
      $96 = HEAPU8[$95>>0]|(HEAPU8[$95+1>>0]<<8)|(HEAPU8[$95+2>>0]<<16)|(HEAPU8[$95+3>>0]<<24); //@line 275 "load_mt2.cpp"
      $97 = (($96) + 1)|0; //@line 275 "load_mt2.cpp"
      $98 = $97 & -2; //@line 275 "load_mt2.cpp"
      $99 = (($dwMemPos$3133) + 6)|0; //@line 276 "load_mt2.cpp"
      $100 = (($98) + ($99))|0; //@line 277 "load_mt2.cpp"
      $101 = ($100>>>0)>($dwMemLength>>>0); //@line 277 "load_mt2.cpp"
      if ($101) {
       $217 = $638;$dwMemPos$5 = $99;
       break L39;
      }
      $102 = (($lpStream) + ($dwMemPos$3133)|0); //@line 274 "load_mt2.cpp"
      $103 = HEAPU8[$102>>0]|(HEAPU8[$102+1>>0]<<8); //@line 278 "load_mt2.cpp"
      $104 = $103&65535; //@line 278 "load_mt2.cpp"
      $105 = ($iPat$0132>>>0)<(240); //@line 279 "load_mt2.cpp"
      $$off = (($103) + -1)<<16>>16; //@line 279 "load_mt2.cpp"
      $106 = ($$off&65535)<(256); //@line 279 "load_mt2.cpp"
      $107 = $105 & $106; //@line 279 "load_mt2.cpp"
      if ($107) {
       $108 = (((($this)) + 52864|0) + ($iPat$0132<<1)|0); //@line 284 "load_mt2.cpp"
       HEAP16[$108>>1] = $103; //@line 284 "load_mt2.cpp"
       $109 = HEAP32[$13>>2]|0; //@line 285 "load_mt2.cpp"
       $110 = (__ZN10CSoundFile15AllocatePatternEjj($104,$109)|0); //@line 285 "load_mt2.cpp"
       $111 = (((($this)) + 51904|0) + ($iPat$0132<<2)|0); //@line 285 "load_mt2.cpp"
       HEAP32[$111>>2] = $110; //@line 285 "load_mt2.cpp"
       $112 = ($110|0)==(0|0); //@line 286 "load_mt2.cpp"
       if ($112) {
        $$0 = 1;
        break L41;
       }
       $113 = HEAPU8[$70>>0]|(HEAPU8[$70+1>>0]<<8)|(HEAPU8[$70+2>>0]<<16)|(HEAPU8[$70+3>>0]<<24); //@line 289 "load_mt2.cpp"
       $114 = $113 & 1; //@line 289 "load_mt2.cpp"
       $115 = ($114|0)==(0); //@line 289 "load_mt2.cpp"
       if ($115) {
        $179 = ($98>>>0)>(7); //@line 336 "load_mt2.cpp"
        if (!($179)) {
         $dwMemPos$4 = $100;
         break;
        }
        $180 = (($lpStream) + ($99)|0); //@line 291 "load_mt2.cpp"
        $len$0127 = $98;$m$0128 = $110;$n$0126 = 0;$p1$0125 = $180;
        while(1) {
         $181 = HEAP32[$13>>2]|0; //@line 336 "load_mt2.cpp"
         $182 = Math_imul($181, $104)|0; //@line 336 "load_mt2.cpp"
         $183 = ($n$0126>>>0)<($182>>>0); //@line 336 "load_mt2.cpp"
         if (!($183)) {
          $dwMemPos$4 = $100;
          break L43;
         }
         __ZL17ConvertMT2CommandP10CSoundFileP11_MODCOMMANDPK11_MT2COMMAND($this,$m$0128,$p1$0125); //@line 338 "load_mt2.cpp"
         $184 = (($len$0127) + -7)|0; //@line 339 "load_mt2.cpp"
         $185 = (($n$0126) + 1)|0; //@line 340 "load_mt2.cpp"
         $186 = ((($p1$0125)) + 7|0); //@line 341 "load_mt2.cpp"
         $187 = ((($m$0128)) + 6|0); //@line 342 "load_mt2.cpp"
         $188 = ($184>>>0)>(7); //@line 336 "load_mt2.cpp"
         if ($188) {
          $len$0127 = $184;$m$0128 = $187;$n$0126 = $185;$p1$0125 = $186;
         } else {
          $dwMemPos$4 = $100;
          break L43;
         }
        }
       }
       $116 = (($dwMemPos$3133) + 8)|0; //@line 300 "load_mt2.cpp"
       $ch$0 = 0;$pos$0 = 0;$row$0 = 0;
       while(1) {
        $117 = ($pos$0>>>0)<($98>>>0); //@line 293 "load_mt2.cpp"
        if (!($117)) {
         $dwMemPos$4 = $100;
         break L43;
        }
        $118 = (($pos$0) + 1)|0; //@line 296 "load_mt2.cpp"
        $$sum32 = (($pos$0) + ($99))|0; //@line 296 "load_mt2.cpp"
        $119 = (($lpStream) + ($$sum32)|0); //@line 296 "load_mt2.cpp"
        $120 = HEAP8[$119>>0]|0; //@line 296 "load_mt2.cpp"
        $121 = $120&255; //@line 296 "load_mt2.cpp"
        $122 = ($120<<24>>24)==(-1); //@line 298 "load_mt2.cpp"
        if ($122) {
         $$sum40 = (($118) + ($99))|0; //@line 300 "load_mt2.cpp"
         $123 = (($lpStream) + ($$sum40)|0); //@line 300 "load_mt2.cpp"
         $124 = HEAP8[$123>>0]|0; //@line 300 "load_mt2.cpp"
         $125 = $124&255; //@line 300 "load_mt2.cpp"
         $126 = (($pos$0) + 3)|0; //@line 301 "load_mt2.cpp"
         $$sum41 = (($116) + ($pos$0))|0; //@line 301 "load_mt2.cpp"
         $127 = (($lpStream) + ($$sum41)|0); //@line 301 "load_mt2.cpp"
         $128 = HEAP8[$127>>0]|0; //@line 301 "load_mt2.cpp"
         $129 = $128&255; //@line 301 "load_mt2.cpp"
         $infobyte$0 = $129;$pos$1 = $126;$rptcount$0 = $125;
        } else {
         $infobyte$0 = $121;$pos$1 = $118;$rptcount$0 = 0;
        }
        $130 = $infobyte$0 & 127; //@line 309 "load_mt2.cpp"
        $131 = ($130|0)==(0); //@line 309 "load_mt2.cpp"
        if ($131) {
         $pos$9 = $pos$1;
        } else {
         $132 = HEAP32[$13>>2]|0; //@line 311 "load_mt2.cpp"
         $133 = Math_imul($132, $row$0)|0; //@line 311 "load_mt2.cpp"
         $134 = (($133) + ($ch$0))|0; //@line 311 "load_mt2.cpp"
         $135 = $infobyte$0 & 1; //@line 313 "load_mt2.cpp"
         $136 = ($135|0)==(0); //@line 313 "load_mt2.cpp"
         ;HEAP8[$cmd>>0]=0|0;HEAP8[$cmd+1>>0]=0|0;HEAP8[$cmd+2>>0]=0|0;HEAP8[$cmd+3>>0]=0|0;HEAP8[$cmd+4>>0]=0|0;HEAP8[$cmd+5>>0]=0|0;HEAP8[$cmd+6>>0]=0|0; //@line 312 "load_mt2.cpp"
         if ($136) {
          $pos$2 = $pos$1;
         } else {
          $137 = (($pos$1) + 1)|0; //@line 313 "load_mt2.cpp"
          $$sum39 = (($pos$1) + ($99))|0; //@line 313 "load_mt2.cpp"
          $138 = (($lpStream) + ($$sum39)|0); //@line 313 "load_mt2.cpp"
          $139 = HEAP8[$138>>0]|0; //@line 313 "load_mt2.cpp"
          HEAP8[$cmd>>0] = $139; //@line 313 "load_mt2.cpp"
          $pos$2 = $137;
         }
         $140 = $infobyte$0 & 2; //@line 314 "load_mt2.cpp"
         $141 = ($140|0)==(0); //@line 314 "load_mt2.cpp"
         if ($141) {
          $pos$3 = $pos$2;
         } else {
          $142 = (($pos$2) + 1)|0; //@line 314 "load_mt2.cpp"
          $$sum38 = (($pos$2) + ($99))|0; //@line 314 "load_mt2.cpp"
          $143 = (($lpStream) + ($$sum38)|0); //@line 314 "load_mt2.cpp"
          $144 = HEAP8[$143>>0]|0; //@line 314 "load_mt2.cpp"
          HEAP8[$76>>0] = $144; //@line 314 "load_mt2.cpp"
          $pos$3 = $142;
         }
         $145 = $infobyte$0 & 4; //@line 315 "load_mt2.cpp"
         $146 = ($145|0)==(0); //@line 315 "load_mt2.cpp"
         if ($146) {
          $pos$4 = $pos$3;
         } else {
          $147 = (($pos$3) + 1)|0; //@line 315 "load_mt2.cpp"
          $$sum37 = (($pos$3) + ($99))|0; //@line 315 "load_mt2.cpp"
          $148 = (($lpStream) + ($$sum37)|0); //@line 315 "load_mt2.cpp"
          $149 = HEAP8[$148>>0]|0; //@line 315 "load_mt2.cpp"
          HEAP8[$75>>0] = $149; //@line 315 "load_mt2.cpp"
          $pos$4 = $147;
         }
         $150 = $infobyte$0 & 8; //@line 316 "load_mt2.cpp"
         $151 = ($150|0)==(0); //@line 316 "load_mt2.cpp"
         if ($151) {
          $pos$5 = $pos$4;
         } else {
          $152 = (($pos$4) + 1)|0; //@line 316 "load_mt2.cpp"
          $$sum36 = (($pos$4) + ($99))|0; //@line 316 "load_mt2.cpp"
          $153 = (($lpStream) + ($$sum36)|0); //@line 316 "load_mt2.cpp"
          $154 = HEAP8[$153>>0]|0; //@line 316 "load_mt2.cpp"
          HEAP8[$74>>0] = $154; //@line 316 "load_mt2.cpp"
          $pos$5 = $152;
         }
         $155 = $infobyte$0 & 16; //@line 317 "load_mt2.cpp"
         $156 = ($155|0)==(0); //@line 317 "load_mt2.cpp"
         if ($156) {
          $pos$6 = $pos$5;
         } else {
          $157 = (($pos$5) + 1)|0; //@line 317 "load_mt2.cpp"
          $$sum35 = (($pos$5) + ($99))|0; //@line 317 "load_mt2.cpp"
          $158 = (($lpStream) + ($$sum35)|0); //@line 317 "load_mt2.cpp"
          $159 = HEAP8[$158>>0]|0; //@line 317 "load_mt2.cpp"
          HEAP8[$73>>0] = $159; //@line 317 "load_mt2.cpp"
          $pos$6 = $157;
         }
         $160 = $infobyte$0 & 32; //@line 318 "load_mt2.cpp"
         $161 = ($160|0)==(0); //@line 318 "load_mt2.cpp"
         if ($161) {
          $pos$7 = $pos$6;
         } else {
          $162 = (($pos$6) + 1)|0; //@line 318 "load_mt2.cpp"
          $$sum34 = (($pos$6) + ($99))|0; //@line 318 "load_mt2.cpp"
          $163 = (($lpStream) + ($$sum34)|0); //@line 318 "load_mt2.cpp"
          $164 = HEAP8[$163>>0]|0; //@line 318 "load_mt2.cpp"
          HEAP8[$72>>0] = $164; //@line 318 "load_mt2.cpp"
          $pos$7 = $162;
         }
         $165 = $infobyte$0 & 64; //@line 319 "load_mt2.cpp"
         $166 = ($165|0)==(0); //@line 319 "load_mt2.cpp"
         if ($166) {
          $pos$8 = $pos$7;
         } else {
          $167 = (($pos$7) + 1)|0; //@line 319 "load_mt2.cpp"
          $$sum33 = (($pos$7) + ($99))|0; //@line 319 "load_mt2.cpp"
          $168 = (($lpStream) + ($$sum33)|0); //@line 319 "load_mt2.cpp"
          $169 = HEAP8[$168>>0]|0; //@line 319 "load_mt2.cpp"
          HEAP8[$71>>0] = $169; //@line 319 "load_mt2.cpp"
          $pos$8 = $167;
         }
         $170 = (($110) + (($134*6)|0)|0); //@line 326 "load_mt2.cpp"
         __ZL17ConvertMT2CommandP10CSoundFileP11_MODCOMMANDPK11_MT2COMMAND($this,$170,$cmd); //@line 326 "load_mt2.cpp"
         $pos$9 = $pos$8;
        }
        $171 = (($row$0) + 1)|0; //@line 328 "load_mt2.cpp"
        $172 = (($171) + ($rptcount$0))|0; //@line 328 "load_mt2.cpp"
        $173 = ($172>>>0)<($104>>>0); //@line 329 "load_mt2.cpp"
        if ($173) {
         $ch$1$lcssa = $ch$0;$row$1$lcssa = $172;
        } else {
         $ch$1120 = $ch$0;$row$1121 = $172;
         while(1) {
          $174 = (($row$1121) - ($104))|0; //@line 329 "load_mt2.cpp"
          $175 = (($ch$1120) + 1)|0; //@line 329 "load_mt2.cpp"
          $176 = ($174>>>0)<($104>>>0); //@line 329 "load_mt2.cpp"
          if ($176) {
           $ch$1$lcssa = $175;$row$1$lcssa = $174;
           break;
          } else {
           $ch$1120 = $175;$row$1121 = $174;
          }
         }
        }
        $177 = HEAP32[$13>>2]|0; //@line 330 "load_mt2.cpp"
        $178 = ($ch$1$lcssa>>>0)<($177>>>0); //@line 330 "load_mt2.cpp"
        if ($178) {
         $ch$0 = $ch$1$lcssa;$pos$0 = $pos$9;$row$0 = $row$1$lcssa;
        } else {
         $dwMemPos$4 = $100;
         break;
        }
       }
      } else {
       $dwMemPos$4 = $100;
      }
     } else {
      $dwMemPos$4 = $dwMemPos$3133;
     }
    } while(0);
    $189 = (($iPat$0132) + 1)|0; //@line 272 "load_mt2.cpp"
    $190 = HEAPU8[$66>>0]|(HEAPU8[$66+1>>0]<<8); //@line 272 "load_mt2.cpp"
    $191 = $190&65535; //@line 272 "load_mt2.cpp"
    $192 = ($189>>>0)<($191>>>0); //@line 272 "load_mt2.cpp"
    if ($192) {
     $638 = $190;$dwMemPos$3133 = $dwMemPos$4;$iPat$0132 = $189;
    } else {
     $217 = $190;$dwMemPos$5 = $dwMemPos$4;
     break L39;
    }
   }
   STACKTOP = sp;return ($$0|0); //@line 635 "load_mt2.cpp"
  }
 } while(0);
 $193 = ($pdd$0|0)!=(0|0); //@line 349 "load_mt2.cpp"
 L92: do {
  if ($193) {
   $194 = HEAPU8[$pdd$0>>0]|(HEAPU8[$pdd$0+1>>0]<<8); //@line 354 "load_mt2.cpp"
   $195 = ($194<<16>>16)==(0); //@line 354 "load_mt2.cpp"
   if ($195) {
    $dwMemPos$7 = $dwMemPos$5;
   } else {
    $196 = (($dwMemLength) + -2)|0; //@line 356 "load_mt2.cpp"
    $197 = $194&65535; //@line 354 "load_mt2.cpp"
    $dwMemPos$6117 = $dwMemPos$5;$iDrm$0116 = 0;
    while(1) {
     $198 = ($dwMemPos$6117>>>0)>($196>>>0); //@line 356 "load_mt2.cpp"
     if ($198) {
      $$0 = 1;
      break;
     }
     $199 = (($lpStream) + ($dwMemPos$6117)|0); //@line 357 "load_mt2.cpp"
     $200 = HEAP16[$199>>1]|0; //@line 357 "load_mt2.cpp"
     $201 = $200&65535; //@line 357 "load_mt2.cpp"
     $202 = $201 << 5; //@line 361 "load_mt2.cpp"
     $203 = $202 | 2; //@line 361 "load_mt2.cpp"
     $204 = (($203) + ($dwMemPos$6117))|0; //@line 361 "load_mt2.cpp"
     $205 = (($iDrm$0116) + 1)|0; //@line 354 "load_mt2.cpp"
     $206 = ($205>>>0)<($197>>>0); //@line 354 "load_mt2.cpp"
     if ($206) {
      $dwMemPos$6117 = $204;$iDrm$0116 = $205;
     } else {
      $dwMemPos$7 = $204;
      break L92;
     }
    }
    STACKTOP = sp;return ($$0|0); //@line 635 "load_mt2.cpp"
   }
  } else {
   $dwMemPos$7 = $dwMemPos$5;
  }
 } while(0);
 $207 = ((($lpStream)) + 118|0); //@line 365 "load_mt2.cpp"
 $208 = HEAPU8[$207>>0]|(HEAPU8[$207+1>>0]<<8)|(HEAPU8[$207+2>>0]<<16)|(HEAPU8[$207+3>>0]<<24); //@line 365 "load_mt2.cpp"
 $209 = $208 & 2; //@line 365 "load_mt2.cpp"
 $210 = ($209|0)==(0); //@line 365 "load_mt2.cpp"
 L100: do {
  if ($210) {
   $dwMemPos$11 = $dwMemPos$7;
  } else {
   $211 = HEAP32[$13>>2]|0; //@line 370 "load_mt2.cpp"
   $212 = $208 >>> 4; //@line 371 "load_mt2.cpp"
   $$lobit = $212 & 1; //@line 371 "load_mt2.cpp"
   $$49 = (($211) + ($$lobit))|0; //@line 371 "load_mt2.cpp"
   $213 = $208 & 8; //@line 372 "load_mt2.cpp"
   $214 = ($213|0)!=(0); //@line 372 "load_mt2.cpp"
   $or$cond7 = $193 & $214; //@line 372 "load_mt2.cpp"
   $215 = (($$49) + 8)|0; //@line 372 "load_mt2.cpp"
   $nAutoCount$1 = $or$cond7 ? $215 : $$49; //@line 372 "load_mt2.cpp"
   $216 = $217&65535; //@line 373 "load_mt2.cpp"
   $218 = Math_imul($nAutoCount$1, $216)|0; //@line 373 "load_mt2.cpp"
   $219 = ($218|0)==(0); //@line 374 "load_mt2.cpp"
   if ($219) {
    $dwMemPos$11 = $dwMemPos$7;
   } else {
    $dwMemPos$8113 = $dwMemPos$7;$iAuto$0112 = 0;
    while(1) {
     $220 = (($dwMemPos$8113) + 12)|0; //@line 376 "load_mt2.cpp"
     $221 = ($220>>>0)<($dwMemLength>>>0); //@line 376 "load_mt2.cpp"
     if (!($221)) {
      $$0 = 1;
      break;
     }
     $222 = (($lpStream) + ($dwMemPos$8113)|0); //@line 377 "load_mt2.cpp"
     $223 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8); //@line 378 "load_mt2.cpp"
     $224 = ($223&65535)<(514); //@line 378 "load_mt2.cpp"
     $225 = $224 ? 4 : 8; //@line 378 "load_mt2.cpp"
     $226 = (($225) + ($dwMemPos$8113))|0; //@line 378 "load_mt2.cpp"
     $227 = HEAPU8[$222>>0]|(HEAPU8[$222+1>>0]<<8)|(HEAPU8[$222+2>>0]<<16)|(HEAPU8[$222+3>>0]<<24); //@line 381 "load_mt2.cpp"
     $228 = $227 & 1; //@line 381 "load_mt2.cpp"
     $229 = ($228|0)==(0); //@line 381 "load_mt2.cpp"
     $230 = (($226) + 260)|0; //@line 387 "load_mt2.cpp"
     $dwMemPos$9$ = $229 ? $226 : $230; //@line 381 "load_mt2.cpp"
     $231 = $227 & 2; //@line 381 "load_mt2.cpp"
     $232 = ($231|0)==(0); //@line 381 "load_mt2.cpp"
     $233 = (($dwMemPos$9$) + 260)|0; //@line 387 "load_mt2.cpp"
     $dwMemPos$9$$1 = $232 ? $dwMemPos$9$ : $233; //@line 381 "load_mt2.cpp"
     $234 = $227 & 4; //@line 381 "load_mt2.cpp"
     $235 = ($234|0)==(0); //@line 381 "load_mt2.cpp"
     $236 = (($dwMemPos$9$$1) + 260)|0; //@line 387 "load_mt2.cpp"
     $dwMemPos$9$$2 = $235 ? $dwMemPos$9$$1 : $236; //@line 381 "load_mt2.cpp"
     $237 = $227 & 8; //@line 381 "load_mt2.cpp"
     $238 = ($237|0)==(0); //@line 381 "load_mt2.cpp"
     $239 = (($dwMemPos$9$$2) + 260)|0; //@line 387 "load_mt2.cpp"
     $dwMemPos$9$$3 = $238 ? $dwMemPos$9$$2 : $239; //@line 381 "load_mt2.cpp"
     $240 = $227 & 16; //@line 381 "load_mt2.cpp"
     $241 = ($240|0)==(0); //@line 381 "load_mt2.cpp"
     $242 = (($dwMemPos$9$$3) + 260)|0; //@line 387 "load_mt2.cpp"
     $dwMemPos$9$$4 = $241 ? $dwMemPos$9$$3 : $242; //@line 381 "load_mt2.cpp"
     $243 = $227 & 32; //@line 381 "load_mt2.cpp"
     $244 = ($243|0)==(0); //@line 381 "load_mt2.cpp"
     $245 = (($dwMemPos$9$$4) + 260)|0; //@line 387 "load_mt2.cpp"
     $dwMemPos$9$$5 = $244 ? $dwMemPos$9$$4 : $245; //@line 381 "load_mt2.cpp"
     $246 = $227 & 64; //@line 381 "load_mt2.cpp"
     $247 = ($246|0)==(0); //@line 381 "load_mt2.cpp"
     $248 = (($dwMemPos$9$$5) + 260)|0; //@line 387 "load_mt2.cpp"
     $dwMemPos$9$$6 = $247 ? $dwMemPos$9$$5 : $248; //@line 381 "load_mt2.cpp"
     $249 = $227 & 128; //@line 381 "load_mt2.cpp"
     $250 = ($249|0)==(0); //@line 381 "load_mt2.cpp"
     $251 = (($dwMemPos$9$$6) + 260)|0; //@line 387 "load_mt2.cpp"
     $dwMemPos$9$$7 = $250 ? $dwMemPos$9$$6 : $251; //@line 381 "load_mt2.cpp"
     $252 = $227 & 256; //@line 381 "load_mt2.cpp"
     $253 = ($252|0)==(0); //@line 381 "load_mt2.cpp"
     $254 = (($dwMemPos$9$$7) + 260)|0; //@line 387 "load_mt2.cpp"
     $dwMemPos$9$$8 = $253 ? $dwMemPos$9$$7 : $254; //@line 381 "load_mt2.cpp"
     $255 = $227 & 512; //@line 381 "load_mt2.cpp"
     $256 = ($255|0)==(0); //@line 381 "load_mt2.cpp"
     $257 = (($dwMemPos$9$$8) + 260)|0; //@line 387 "load_mt2.cpp"
     $dwMemPos$9$$9 = $256 ? $dwMemPos$9$$8 : $257; //@line 381 "load_mt2.cpp"
     $258 = $227 & 1024; //@line 381 "load_mt2.cpp"
     $259 = ($258|0)==(0); //@line 381 "load_mt2.cpp"
     $260 = (($dwMemPos$9$$9) + 260)|0; //@line 387 "load_mt2.cpp"
     $dwMemPos$9$$10 = $259 ? $dwMemPos$9$$9 : $260; //@line 381 "load_mt2.cpp"
     $261 = $227 & 2048; //@line 381 "load_mt2.cpp"
     $262 = ($261|0)==(0); //@line 381 "load_mt2.cpp"
     $263 = (($dwMemPos$9$$10) + 260)|0; //@line 387 "load_mt2.cpp"
     $dwMemPos$9$$11 = $262 ? $dwMemPos$9$$10 : $263; //@line 381 "load_mt2.cpp"
     $264 = $227 & 4096; //@line 381 "load_mt2.cpp"
     $265 = ($264|0)==(0); //@line 381 "load_mt2.cpp"
     $266 = (($dwMemPos$9$$11) + 260)|0; //@line 387 "load_mt2.cpp"
     $dwMemPos$9$$12 = $265 ? $dwMemPos$9$$11 : $266; //@line 381 "load_mt2.cpp"
     $267 = $227 & 8192; //@line 381 "load_mt2.cpp"
     $268 = ($267|0)==(0); //@line 381 "load_mt2.cpp"
     $269 = (($dwMemPos$9$$12) + 260)|0; //@line 387 "load_mt2.cpp"
     $dwMemPos$9$$13 = $268 ? $dwMemPos$9$$12 : $269; //@line 381 "load_mt2.cpp"
     $270 = (($iAuto$0112) + 1)|0; //@line 374 "load_mt2.cpp"
     $271 = ($270>>>0)<($218>>>0); //@line 374 "load_mt2.cpp"
     if ($271) {
      $dwMemPos$8113 = $dwMemPos$9$$13;$iAuto$0112 = $270;
     } else {
      $dwMemPos$11 = $dwMemPos$9$$13;
      break L100;
     }
    }
    STACKTOP = sp;return ($$0|0); //@line 635 "load_mt2.cpp"
   }
  }
 } while(0);
 _memset(($InstrMap|0),0,1020)|0; //@line 396 "load_mt2.cpp"
 $272 = ((($lpStream)) + 122|0); //@line 397 "load_mt2.cpp"
 $273 = HEAPU8[$272>>0]|(HEAPU8[$272+1>>0]<<8); //@line 397 "load_mt2.cpp"
 $274 = $273&65535; //@line 397 "load_mt2.cpp"
 $275 = ($274>>>0)>(239); //@line 397 "load_mt2.cpp"
 $$50 = $275 ? 239 : $274; //@line 397 "load_mt2.cpp"
 $276 = ((($this)) + 59688|0); //@line 397 "load_mt2.cpp"
 HEAP32[$276>>2] = $$50; //@line 397 "load_mt2.cpp"
 $277 = (($dwMemLength) + -40)|0;
 $278 = ((($pehdr)) + 4|0); //@line 437 "load_mt2.cpp"
 $279 = ((($pehdr)) + 12|0); //@line 438 "load_mt2.cpp"
 $280 = ((($pehdr)) + 8|0); //@line 438 "load_mt2.cpp"
 $281 = ((($pedata)) + 4|0); //@line 440 "load_mt2.cpp"
 $282 = ((($pedata)) + 12|0); //@line 441 "load_mt2.cpp"
 $283 = ((($pedata)) + 8|0); //@line 441 "load_mt2.cpp"
 $dwMemPos$12109 = $dwMemPos$11;$iIns$0108 = 1;
 while(1) {
  $284 = (($dwMemPos$12109) + 36)|0; //@line 400 "load_mt2.cpp"
  $285 = ($284>>>0)>($dwMemLength>>>0); //@line 400 "load_mt2.cpp"
  if ($285) {
   $$0 = 1;
   label = 148;
   break;
  }
  $286 = (($lpStream) + ($dwMemPos$12109)|0); //@line 401 "load_mt2.cpp"
  $287 = HEAP32[$276>>2]|0; //@line 403 "load_mt2.cpp"
  $288 = ($iIns$0108>>>0)>($287>>>0); //@line 403 "load_mt2.cpp"
  if ($288) {
   $penv$0 = 0;
  } else {
   $289 = (__Znwj(632)|0); //@line 405 "load_mt2.cpp"
   $290 = (((($this)) + 48640|0) + ($iIns$0108<<2)|0); //@line 406 "load_mt2.cpp"
   HEAP32[$290>>2] = $289; //@line 406 "load_mt2.cpp"
   _memset(($289|0),0,632)|0; //@line 409 "load_mt2.cpp"
   $291 = ((($289)) + 585|0); //@line 410 "load_mt2.cpp"
   dest=$291; src=$286; stop=dest+32|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 410 "load_mt2.cpp"
   $292 = ((($289)) + 8|0); //@line 411 "load_mt2.cpp"
   HEAP16[$292>>1] = 64; //@line 411 "load_mt2.cpp"
   $293 = ((($289)) + 10|0); //@line 412 "load_mt2.cpp"
   HEAP16[$293>>1] = 128; //@line 412 "load_mt2.cpp"
   $i$0103 = 0;
   while(1) {
    $294 = (($i$0103) + 1)|0; //@line 415 "load_mt2.cpp"
    $295 = $294&255; //@line 415 "load_mt2.cpp"
    $$sum30 = (($i$0103) + 428)|0; //@line 415 "load_mt2.cpp"
    $296 = (($289) + ($$sum30)|0); //@line 415 "load_mt2.cpp"
    HEAP8[$296>>0] = $295; //@line 415 "load_mt2.cpp"
    $exitcond150 = ($294|0)==(120); //@line 413 "load_mt2.cpp"
    if ($exitcond150) {
     break;
    } else {
     $i$0103 = $294;
    }
   }
   $penv$0 = $289;
  }
  $$sum22 = (($dwMemPos$12109) + 32)|0; //@line 422 "load_mt2.cpp"
  $297 = (($lpStream) + ($$sum22)|0); //@line 422 "load_mt2.cpp"
  $298 = HEAPU8[$297>>0]|(HEAPU8[$297+1>>0]<<8)|(HEAPU8[$297+2>>0]<<16)|(HEAPU8[$297+3>>0]<<24); //@line 422 "load_mt2.cpp"
  $299 = ($298|0)<(1); //@line 422 "load_mt2.cpp"
  $300 = ($277>>>0)<($dwMemPos$12109>>>0); //@line 422 "load_mt2.cpp"
  $or$cond52 = $300 | $299; //@line 422 "load_mt2.cpp"
  $301 = (($277) - ($dwMemPos$12109))|0;
  $302 = ($298>>>0)>($301>>>0); //@line 422 "load_mt2.cpp"
  $or$cond175 = $or$cond52 | $302; //@line 422 "load_mt2.cpp"
  if ($or$cond175) {
   $dwMemPos$13 = $284;
  } else {
   $303 = (($iIns$0108) + -1)|0; //@line 424 "load_mt2.cpp"
   $304 = (($InstrMap) + ($303<<2)|0); //@line 424 "load_mt2.cpp"
   HEAP32[$304>>2] = $286; //@line 424 "load_mt2.cpp"
   $305 = ($penv$0|0)==(0|0); //@line 425 "load_mt2.cpp"
   if ($305) {
    $451 = $298;
   } else {
    $$sum23 = (($dwMemPos$12109) + 138)|0; //@line 427 "load_mt2.cpp"
    $306 = (($lpStream) + ($$sum23)|0); //@line 427 "load_mt2.cpp"
    $307 = HEAPU8[$306>>0]|(HEAPU8[$306+1>>0]<<8); //@line 427 "load_mt2.cpp"
    $308 = $307&65535; //@line 427 "load_mt2.cpp"
    HEAP32[$penv$0>>2] = $308; //@line 427 "load_mt2.cpp"
    $$sum24 = (($dwMemPos$12109) + 140)|0; //@line 428 "load_mt2.cpp"
    $309 = (($lpStream) + ($$sum24)|0); //@line 428 "load_mt2.cpp"
    $310 = HEAPU8[$309>>0]|(HEAPU8[$309+1>>0]<<8); //@line 428 "load_mt2.cpp"
    $311 = $310&65535; //@line 428 "load_mt2.cpp"
    $312 = $311 & 3; //@line 428 "load_mt2.cpp"
    $313 = $312&255; //@line 428 "load_mt2.cpp"
    $314 = ((($penv$0)) + 571|0); //@line 428 "load_mt2.cpp"
    HEAP8[$314>>0] = $313; //@line 428 "load_mt2.cpp"
    $315 = $311 >>> 8; //@line 429 "load_mt2.cpp"
    $316 = $315 & 3; //@line 429 "load_mt2.cpp"
    $317 = $316&255; //@line 429 "load_mt2.cpp"
    $318 = ((($penv$0)) + 572|0); //@line 429 "load_mt2.cpp"
    HEAP8[$318>>0] = $317; //@line 429 "load_mt2.cpp"
    $319 = $311 >>> 12; //@line 430 "load_mt2.cpp"
    $320 = $319 & 3; //@line 430 "load_mt2.cpp"
    $321 = $320&255; //@line 430 "load_mt2.cpp"
    $322 = ((($penv$0)) + 573|0); //@line 430 "load_mt2.cpp"
    HEAP8[$322>>0] = $321; //@line 430 "load_mt2.cpp"
    $323 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8); //@line 433 "load_mt2.cpp"
    $324 = ($323&65535)<(514); //@line 433 "load_mt2.cpp"
    do {
     if ($324) {
      $325 = (($dwMemPos$12109) + 144)|0; //@line 435 "load_mt2.cpp"
      $326 = (($lpStream) + ($325)|0); //@line 436 "load_mt2.cpp"
      HEAP32[$pehdr>>2] = $326; //@line 436 "load_mt2.cpp"
      $$sum27 = (($dwMemPos$12109) + 152)|0; //@line 437 "load_mt2.cpp"
      $327 = (($lpStream) + ($$sum27)|0); //@line 437 "load_mt2.cpp"
      HEAP32[$278>>2] = $327; //@line 437 "load_mt2.cpp"
      HEAP32[$279>>2] = 0; //@line 438 "load_mt2.cpp"
      HEAP32[$280>>2] = 0; //@line 438 "load_mt2.cpp"
      $$sum28 = (($dwMemPos$12109) + 160)|0; //@line 439 "load_mt2.cpp"
      $328 = (($lpStream) + ($$sum28)|0); //@line 439 "load_mt2.cpp"
      HEAP32[$pedata>>2] = $328; //@line 439 "load_mt2.cpp"
      $$sum29 = (($dwMemPos$12109) + 224)|0; //@line 440 "load_mt2.cpp"
      $329 = (($lpStream) + ($$sum29)|0); //@line 440 "load_mt2.cpp"
      HEAP32[$281>>2] = $329; //@line 440 "load_mt2.cpp"
      HEAP32[$282>>2] = 0; //@line 441 "load_mt2.cpp"
      HEAP32[$283>>2] = 0; //@line 441 "load_mt2.cpp"
      $639 = $326;
     } else {
      $352 = (($dwMemPos$12109) + 148)|0; //@line 435 "load_mt2.cpp"
      $$sum25 = (($dwMemPos$12109) + 144)|0; //@line 447 "load_mt2.cpp"
      $353 = (($lpStream) + ($$sum25)|0); //@line 447 "load_mt2.cpp"
      $354 = HEAPU8[$353>>0]|(HEAPU8[$353+1>>0]<<8); //@line 447 "load_mt2.cpp"
      $355 = $354&65535; //@line 447 "load_mt2.cpp"
      $356 = $355 & 1; //@line 447 "load_mt2.cpp"
      $357 = ($356|0)==(0); //@line 447 "load_mt2.cpp"
      if ($357) {
       HEAP32[$pehdr>>2] = 0; //@line 454 "load_mt2.cpp"
       HEAP32[$pedata>>2] = 0; //@line 455 "load_mt2.cpp"
       $640 = 0;$dwEnvPos3$1 = $352;
      } else {
       $358 = (($lpStream) + ($352)|0); //@line 449 "load_mt2.cpp"
       HEAP32[$pehdr>>2] = $358; //@line 449 "load_mt2.cpp"
       $$sum26 = (($dwMemPos$12109) + 156)|0; //@line 450 "load_mt2.cpp"
       $359 = (($lpStream) + ($$sum26)|0); //@line 450 "load_mt2.cpp"
       HEAP32[$pedata>>2] = $359; //@line 450 "load_mt2.cpp"
       $360 = (($dwMemPos$12109) + 220)|0; //@line 451 "load_mt2.cpp"
       $640 = $358;$dwEnvPos3$1 = $360;
      }
      $361 = $355 & 2; //@line 447 "load_mt2.cpp"
      $362 = ($361|0)==(0); //@line 447 "load_mt2.cpp"
      if ($362) {
       HEAP32[$278>>2] = 0; //@line 454 "load_mt2.cpp"
       HEAP32[$281>>2] = 0; //@line 455 "load_mt2.cpp"
       $dwEnvPos3$1$1 = $dwEnvPos3$1;
      } else {
       $626 = (($lpStream) + ($dwEnvPos3$1)|0); //@line 449 "load_mt2.cpp"
       HEAP32[$278>>2] = $626; //@line 449 "load_mt2.cpp"
       $$sum26$1 = (($dwEnvPos3$1) + 8)|0; //@line 450 "load_mt2.cpp"
       $627 = (($lpStream) + ($$sum26$1)|0); //@line 450 "load_mt2.cpp"
       HEAP32[$281>>2] = $627; //@line 450 "load_mt2.cpp"
       $628 = (($dwEnvPos3$1) + 72)|0; //@line 451 "load_mt2.cpp"
       $dwEnvPos3$1$1 = $628;
      }
      $629 = $355 & 4; //@line 447 "load_mt2.cpp"
      $630 = ($629|0)==(0); //@line 447 "load_mt2.cpp"
      if ($630) {
       HEAP32[$280>>2] = 0; //@line 454 "load_mt2.cpp"
       HEAP32[$283>>2] = 0; //@line 455 "load_mt2.cpp"
       $dwEnvPos3$1$2 = $dwEnvPos3$1$1;
      } else {
       $631 = (($lpStream) + ($dwEnvPos3$1$1)|0); //@line 449 "load_mt2.cpp"
       HEAP32[$280>>2] = $631; //@line 449 "load_mt2.cpp"
       $$sum26$2 = (($dwEnvPos3$1$1) + 8)|0; //@line 450 "load_mt2.cpp"
       $632 = (($lpStream) + ($$sum26$2)|0); //@line 450 "load_mt2.cpp"
       HEAP32[$283>>2] = $632; //@line 450 "load_mt2.cpp"
       $633 = (($dwEnvPos3$1$1) + 72)|0; //@line 451 "load_mt2.cpp"
       $dwEnvPos3$1$2 = $633;
      }
      $634 = $355 & 8; //@line 447 "load_mt2.cpp"
      $635 = ($634|0)==(0); //@line 447 "load_mt2.cpp"
      if ($635) {
       HEAP32[$279>>2] = 0; //@line 454 "load_mt2.cpp"
       HEAP32[$282>>2] = 0; //@line 455 "load_mt2.cpp"
       $639 = $640;
       break;
      } else {
       $636 = (($lpStream) + ($dwEnvPos3$1$2)|0); //@line 449 "load_mt2.cpp"
       HEAP32[$279>>2] = $636; //@line 449 "load_mt2.cpp"
       $$sum26$3 = (($dwEnvPos3$1$2) + 8)|0; //@line 450 "load_mt2.cpp"
       $637 = (($lpStream) + ($$sum26$3)|0); //@line 450 "load_mt2.cpp"
       HEAP32[$282>>2] = $637; //@line 450 "load_mt2.cpp"
       $639 = $640;
       break;
      }
     }
    } while(0);
    $330 = ((($penv$0)) + 558|0); //@line 501 "load_mt2.cpp"
    $331 = ((($penv$0)) + 570|0); //@line 502 "load_mt2.cpp"
    $332 = ((($penv$0)) + 569|0); //@line 502 "load_mt2.cpp"
    $333 = ((($penv$0)) + 567|0); //@line 503 "load_mt2.cpp"
    $334 = ((($penv$0)) + 568|0); //@line 504 "load_mt2.cpp"
    $335 = ((($penv$0)) + 140|0); //@line 505 "load_mt2.cpp"
    $336 = ((($penv$0)) + 268|0); //@line 506 "load_mt2.cpp"
    $337 = ((($penv$0)) + 4|0); //@line 500 "load_mt2.cpp"
    $338 = ((($penv$0)) + 556|0); //@line 475 "load_mt2.cpp"
    $339 = ((($penv$0)) + 562|0); //@line 476 "load_mt2.cpp"
    $340 = ((($penv$0)) + 561|0); //@line 476 "load_mt2.cpp"
    $341 = ((($penv$0)) + 559|0); //@line 477 "load_mt2.cpp"
    $342 = ((($penv$0)) + 560|0); //@line 478 "load_mt2.cpp"
    $343 = ((($penv$0)) + 12|0); //@line 479 "load_mt2.cpp"
    $344 = ((($penv$0)) + 204|0); //@line 480 "load_mt2.cpp"
    $345 = ((($penv$0)) + 557|0); //@line 488 "load_mt2.cpp"
    $346 = ((($penv$0)) + 566|0); //@line 489 "load_mt2.cpp"
    $347 = ((($penv$0)) + 565|0); //@line 489 "load_mt2.cpp"
    $348 = ((($penv$0)) + 563|0); //@line 490 "load_mt2.cpp"
    $349 = ((($penv$0)) + 564|0); //@line 491 "load_mt2.cpp"
    $350 = ((($penv$0)) + 76|0); //@line 492 "load_mt2.cpp"
    $351 = ((($penv$0)) + 236|0); //@line 493 "load_mt2.cpp"
    $364 = $639;$iEnv5$0107 = 0;
    while(1) {
     $363 = ($364|0)==(0|0); //@line 460 "load_mt2.cpp"
     do {
      if (!($363)) {
       switch ($iEnv5$0107|0) {
       case 0:  {
        $365 = HEAP8[$364>>0]|0; //@line 472 "load_mt2.cpp"
        $366 = $365 & 1; //@line 472 "load_mt2.cpp"
        $367 = ($366<<24>>24)==(0); //@line 472 "load_mt2.cpp"
        if (!($367)) {
         $368 = HEAP32[$337>>2]|0; //@line 472 "load_mt2.cpp"
         $369 = $368 | 1; //@line 472 "load_mt2.cpp"
         HEAP32[$337>>2] = $369; //@line 472 "load_mt2.cpp"
        }
        $370 = $365 & 2; //@line 473 "load_mt2.cpp"
        $371 = ($370<<24>>24)==(0); //@line 473 "load_mt2.cpp"
        if (!($371)) {
         $372 = HEAP32[$337>>2]|0; //@line 473 "load_mt2.cpp"
         $373 = $372 | 2; //@line 473 "load_mt2.cpp"
         HEAP32[$337>>2] = $373; //@line 473 "load_mt2.cpp"
        }
        $374 = $365 & 4; //@line 474 "load_mt2.cpp"
        $375 = ($374<<24>>24)==(0); //@line 474 "load_mt2.cpp"
        if (!($375)) {
         $376 = HEAP32[$337>>2]|0; //@line 474 "load_mt2.cpp"
         $377 = $376 | 4; //@line 474 "load_mt2.cpp"
         HEAP32[$337>>2] = $377; //@line 474 "load_mt2.cpp"
        }
        $378 = ((($364)) + 1|0); //@line 475 "load_mt2.cpp"
        $379 = HEAP8[$378>>0]|0; //@line 475 "load_mt2.cpp"
        $380 = ($379&255)>(16); //@line 475 "load_mt2.cpp"
        $381 = $380 ? 16 : $379; //@line 475 "load_mt2.cpp"
        HEAP8[$338>>0] = $381; //@line 475 "load_mt2.cpp"
        $382 = ((($364)) + 2|0); //@line 476 "load_mt2.cpp"
        $383 = HEAP8[$382>>0]|0; //@line 476 "load_mt2.cpp"
        HEAP8[$339>>0] = $383; //@line 476 "load_mt2.cpp"
        HEAP8[$340>>0] = $383; //@line 476 "load_mt2.cpp"
        $384 = ((($364)) + 3|0); //@line 477 "load_mt2.cpp"
        $385 = HEAP8[$384>>0]|0; //@line 477 "load_mt2.cpp"
        HEAP8[$341>>0] = $385; //@line 477 "load_mt2.cpp"
        $386 = ((($364)) + 4|0); //@line 478 "load_mt2.cpp"
        $387 = HEAP8[$386>>0]|0; //@line 478 "load_mt2.cpp"
        HEAP8[$342>>0] = $387; //@line 478 "load_mt2.cpp"
        $pEnvData$0 = $344;$pEnvPoints$0 = $343;
        break;
       }
       case 1:  {
        $388 = HEAP8[$364>>0]|0; //@line 485 "load_mt2.cpp"
        $389 = $388 & 1; //@line 485 "load_mt2.cpp"
        $390 = ($389<<24>>24)==(0); //@line 485 "load_mt2.cpp"
        if (!($390)) {
         $391 = HEAP32[$337>>2]|0; //@line 485 "load_mt2.cpp"
         $392 = $391 | 8; //@line 485 "load_mt2.cpp"
         HEAP32[$337>>2] = $392; //@line 485 "load_mt2.cpp"
        }
        $393 = $388 & 2; //@line 486 "load_mt2.cpp"
        $394 = ($393<<24>>24)==(0); //@line 486 "load_mt2.cpp"
        if (!($394)) {
         $395 = HEAP32[$337>>2]|0; //@line 486 "load_mt2.cpp"
         $396 = $395 | 16; //@line 486 "load_mt2.cpp"
         HEAP32[$337>>2] = $396; //@line 486 "load_mt2.cpp"
        }
        $397 = $388 & 4; //@line 487 "load_mt2.cpp"
        $398 = ($397<<24>>24)==(0); //@line 487 "load_mt2.cpp"
        if (!($398)) {
         $399 = HEAP32[$337>>2]|0; //@line 487 "load_mt2.cpp"
         $400 = $399 | 32; //@line 487 "load_mt2.cpp"
         HEAP32[$337>>2] = $400; //@line 487 "load_mt2.cpp"
        }
        $401 = ((($364)) + 1|0); //@line 488 "load_mt2.cpp"
        $402 = HEAP8[$401>>0]|0; //@line 488 "load_mt2.cpp"
        $403 = ($402&255)>(16); //@line 488 "load_mt2.cpp"
        $404 = $403 ? 16 : $402; //@line 488 "load_mt2.cpp"
        HEAP8[$345>>0] = $404; //@line 488 "load_mt2.cpp"
        $405 = ((($364)) + 2|0); //@line 489 "load_mt2.cpp"
        $406 = HEAP8[$405>>0]|0; //@line 489 "load_mt2.cpp"
        HEAP8[$346>>0] = $406; //@line 489 "load_mt2.cpp"
        HEAP8[$347>>0] = $406; //@line 489 "load_mt2.cpp"
        $407 = ((($364)) + 3|0); //@line 490 "load_mt2.cpp"
        $408 = HEAP8[$407>>0]|0; //@line 490 "load_mt2.cpp"
        HEAP8[$348>>0] = $408; //@line 490 "load_mt2.cpp"
        $409 = ((($364)) + 4|0); //@line 491 "load_mt2.cpp"
        $410 = HEAP8[$409>>0]|0; //@line 491 "load_mt2.cpp"
        HEAP8[$349>>0] = $410; //@line 491 "load_mt2.cpp"
        $pEnvData$0 = $351;$pEnvPoints$0 = $350;
        break;
       }
       default: {
        $411 = HEAP8[$364>>0]|0; //@line 498 "load_mt2.cpp"
        $412 = $411 & 1; //@line 498 "load_mt2.cpp"
        $413 = ($412<<24>>24)==(0); //@line 498 "load_mt2.cpp"
        if (!($413)) {
         $414 = ($iEnv5$0107|0)==(3); //@line 498 "load_mt2.cpp"
         $415 = $414 ? 1088 : 64; //@line 498 "load_mt2.cpp"
         $416 = HEAP32[$337>>2]|0; //@line 498 "load_mt2.cpp"
         $417 = $416 | $415; //@line 498 "load_mt2.cpp"
         HEAP32[$337>>2] = $417; //@line 498 "load_mt2.cpp"
        }
        $418 = $411 & 2; //@line 499 "load_mt2.cpp"
        $419 = ($418<<24>>24)==(0); //@line 499 "load_mt2.cpp"
        if (!($419)) {
         $420 = HEAP32[$337>>2]|0; //@line 499 "load_mt2.cpp"
         $421 = $420 | 128; //@line 499 "load_mt2.cpp"
         HEAP32[$337>>2] = $421; //@line 499 "load_mt2.cpp"
        }
        $422 = $411 & 4; //@line 500 "load_mt2.cpp"
        $423 = ($422<<24>>24)==(0); //@line 500 "load_mt2.cpp"
        if (!($423)) {
         $424 = HEAP32[$337>>2]|0; //@line 500 "load_mt2.cpp"
         $425 = $424 | 256; //@line 500 "load_mt2.cpp"
         HEAP32[$337>>2] = $425; //@line 500 "load_mt2.cpp"
        }
        $426 = ((($364)) + 1|0); //@line 501 "load_mt2.cpp"
        $427 = HEAP8[$426>>0]|0; //@line 501 "load_mt2.cpp"
        $428 = ($427&255)>(16); //@line 501 "load_mt2.cpp"
        $429 = $428 ? 16 : $427; //@line 501 "load_mt2.cpp"
        HEAP8[$330>>0] = $429; //@line 501 "load_mt2.cpp"
        $430 = ((($364)) + 2|0); //@line 502 "load_mt2.cpp"
        $431 = HEAP8[$430>>0]|0; //@line 502 "load_mt2.cpp"
        HEAP8[$331>>0] = $431; //@line 502 "load_mt2.cpp"
        HEAP8[$332>>0] = $431; //@line 502 "load_mt2.cpp"
        $432 = ((($364)) + 3|0); //@line 503 "load_mt2.cpp"
        $433 = HEAP8[$432>>0]|0; //@line 503 "load_mt2.cpp"
        HEAP8[$333>>0] = $433; //@line 503 "load_mt2.cpp"
        $434 = ((($364)) + 4|0); //@line 504 "load_mt2.cpp"
        $435 = HEAP8[$434>>0]|0; //@line 504 "load_mt2.cpp"
        HEAP8[$334>>0] = $435; //@line 504 "load_mt2.cpp"
        $pEnvData$0 = $336;$pEnvPoints$0 = $335;
       }
       }
       $436 = (($pedata) + ($iEnv5$0107<<2)|0); //@line 509 "load_mt2.cpp"
       $437 = HEAP32[$436>>2]|0; //@line 509 "load_mt2.cpp"
       $438 = ($437|0)==(0|0); //@line 509 "load_mt2.cpp"
       if ($438) {
        break;
       } else {
        $i6$0106 = 0;
       }
       while(1) {
        $439 = $i6$0106 << 1; //@line 514 "load_mt2.cpp"
        $440 = (($437) + ($439<<1)|0); //@line 514 "load_mt2.cpp"
        $441 = HEAP16[$440>>1]|0; //@line 514 "load_mt2.cpp"
        $442 = (($pEnvPoints$0) + ($i6$0106<<1)|0); //@line 514 "load_mt2.cpp"
        HEAP16[$442>>1] = $441; //@line 514 "load_mt2.cpp"
        $443 = $439 | 1; //@line 515 "load_mt2.cpp"
        $444 = (($437) + ($443<<1)|0); //@line 515 "load_mt2.cpp"
        $445 = HEAP16[$444>>1]|0; //@line 515 "load_mt2.cpp"
        $446 = $445&255; //@line 515 "load_mt2.cpp"
        $447 = (($pEnvData$0) + ($i6$0106)|0); //@line 515 "load_mt2.cpp"
        HEAP8[$447>>0] = $446; //@line 515 "load_mt2.cpp"
        $448 = (($i6$0106) + 1)|0; //@line 512 "load_mt2.cpp"
        $exitcond153 = ($448|0)==(16); //@line 512 "load_mt2.cpp"
        if ($exitcond153) {
         break;
        } else {
         $i6$0106 = $448;
        }
       }
      }
     } while(0);
     $449 = (($iEnv5$0107) + 1)|0; //@line 460 "load_mt2.cpp"
     $exitcond154 = ($449|0)==(4); //@line 460 "load_mt2.cpp"
     if ($exitcond154) {
      break;
     }
     $$phi$trans$insert = (($pehdr) + ($449<<2)|0);
     $$pre = HEAP32[$$phi$trans$insert>>2]|0; //@line 460 "load_mt2.cpp"
     $364 = $$pre;$iEnv5$0107 = $449;
    }
    $$pre158 = HEAPU8[$297>>0]|(HEAPU8[$297+1>>0]<<8)|(HEAPU8[$297+2>>0]<<16)|(HEAPU8[$297+3>>0]<<24); //@line 520 "load_mt2.cpp"
    $451 = $$pre158;
   }
   $450 = (($284) + ($451))|0; //@line 520 "load_mt2.cpp"
   $452 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8); //@line 521 "load_mt2.cpp"
   $453 = ($452&65535)>(513); //@line 521 "load_mt2.cpp"
   $454 = (($450) + 4)|0; //@line 521 "load_mt2.cpp"
   $$56 = $453 ? $454 : $450; //@line 521 "load_mt2.cpp"
   $dwMemPos$13 = $$56;
  }
  $455 = (($iIns$0108) + 1)|0; //@line 398 "load_mt2.cpp"
  $456 = ($455>>>0)<(256); //@line 398 "load_mt2.cpp"
  if ($456) {
   $dwMemPos$12109 = $dwMemPos$13;$iIns$0108 = $455;
  } else {
   $dwMemPos$13$lcssa = $dwMemPos$13;
   break;
  }
 }
 if ((label|0) == 148) {
  STACKTOP = sp;return ($$0|0); //@line 635 "load_mt2.cpp"
 }
 _memset(($SampleMap|0),0,1024)|0; //@line 530 "load_mt2.cpp"
 $457 = ((($lpStream)) + 124|0); //@line 531 "load_mt2.cpp"
 $458 = HEAPU8[$457>>0]|(HEAPU8[$457+1>>0]<<8); //@line 531 "load_mt2.cpp"
 $459 = $458&65535; //@line 531 "load_mt2.cpp"
 $460 = ($459>>>0)>(239); //@line 531 "load_mt2.cpp"
 $$57 = $460 ? 239 : $459; //@line 531 "load_mt2.cpp"
 $461 = ((($this)) + 59684|0); //@line 531 "load_mt2.cpp"
 HEAP32[$461>>2] = $$57; //@line 531 "load_mt2.cpp"
 $dwMemPos$14102 = $dwMemPos$13$lcssa;$iSmp$099 = 1;
 while(1) {
  $462 = (($dwMemPos$14102) + 36)|0; //@line 534 "load_mt2.cpp"
  $463 = ($462>>>0)>($dwMemLength>>>0); //@line 534 "load_mt2.cpp"
  if ($463) {
   $$0 = 1;
   label = 148;
   break;
  }
  $464 = (($lpStream) + ($dwMemPos$14102)|0); //@line 535 "load_mt2.cpp"
  $465 = ($iSmp$099>>>0)<(240); //@line 539 "load_mt2.cpp"
  if ($465) {
   $466 = (((($this)) + 59804|0) + ($iSmp$099<<5)|0); //@line 541 "load_mt2.cpp"
   dest=$466; src=$464; stop=dest+32|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0)); //@line 541 "load_mt2.cpp"
  }
  $$sum = (($dwMemPos$14102) + 32)|0; //@line 543 "load_mt2.cpp"
  $467 = (($lpStream) + ($$sum)|0); //@line 543 "load_mt2.cpp"
  $468 = HEAPU8[$467>>0]|(HEAPU8[$467+1>>0]<<8)|(HEAPU8[$467+2>>0]<<16)|(HEAPU8[$467+3>>0]<<24); //@line 543 "load_mt2.cpp"
  $469 = ($468|0)==(0); //@line 543 "load_mt2.cpp"
  if ($469) {
   $dwMemPos$15 = $462;
  } else {
   $470 = (($iSmp$099) + -1)|0; //@line 545 "load_mt2.cpp"
   $471 = (($SampleMap) + ($470<<2)|0); //@line 545 "load_mt2.cpp"
   HEAP32[$471>>2] = $464; //@line 545 "load_mt2.cpp"
   L193: do {
    if ($465) {
     $472 = (((($this)) + 33280|0) + ($iSmp$099<<6)|0); //@line 548 "load_mt2.cpp"
     $473 = (((((($this)) + 33280|0) + ($iSmp$099<<6)|0)) + 32|0); //@line 549 "load_mt2.cpp"
     HEAP16[$473>>1] = 64; //@line 549 "load_mt2.cpp"
     $$sum12 = (($dwMemPos$14102) + 56)|0; //@line 550 "load_mt2.cpp"
     $474 = (($lpStream) + ($$sum12)|0); //@line 550 "load_mt2.cpp"
     $475 = HEAPU8[$474>>0]|(HEAPU8[$474+1>>0]<<8); //@line 550 "load_mt2.cpp"
     $476 = ($475&65535) >>> 7; //@line 550 "load_mt2.cpp"
     $477 = (((((($this)) + 33280|0) + ($iSmp$099<<6)|0)) + 30|0); //@line 550 "load_mt2.cpp"
     HEAP16[$477>>1] = $476; //@line 550 "load_mt2.cpp"
     $$sum13 = (($dwMemPos$14102) + 58)|0; //@line 551 "load_mt2.cpp"
     $478 = (($lpStream) + ($$sum13)|0); //@line 551 "load_mt2.cpp"
     $479 = HEAP8[$478>>0]|0; //@line 551 "load_mt2.cpp"
     $480 = ($479<<24>>24)==(-128); //@line 551 "load_mt2.cpp"
     $481 = $479&255; //@line 551 "load_mt2.cpp"
     $482 = $481 ^ 128; //@line 551 "load_mt2.cpp"
     $483 = $482&65535; //@line 551 "load_mt2.cpp"
     $484 = $480 ? 128 : $483; //@line 551 "load_mt2.cpp"
     $485 = (((((($this)) + 33280|0) + ($iSmp$099<<6)|0)) + 28|0); //@line 551 "load_mt2.cpp"
     HEAP16[$485>>1] = $484; //@line 551 "load_mt2.cpp"
     $486 = (($lpStream) + ($462)|0); //@line 552 "load_mt2.cpp"
     $487 = HEAPU8[$486>>0]|(HEAPU8[$486+1>>0]<<8)|(HEAPU8[$486+2>>0]<<16)|(HEAPU8[$486+3>>0]<<24); //@line 552 "load_mt2.cpp"
     HEAP32[$472>>2] = $487; //@line 552 "load_mt2.cpp"
     $$sum15 = (($dwMemPos$14102) + 40)|0; //@line 553 "load_mt2.cpp"
     $488 = (($lpStream) + ($$sum15)|0); //@line 553 "load_mt2.cpp"
     $489 = HEAPU8[$488>>0]|(HEAPU8[$488+1>>0]<<8)|(HEAPU8[$488+2>>0]<<16)|(HEAPU8[$488+3>>0]<<24); //@line 553 "load_mt2.cpp"
     $490 = (((((($this)) + 33280|0) + ($iSmp$099<<6)|0)) + 24|0); //@line 553 "load_mt2.cpp"
     HEAP32[$490>>2] = $489; //@line 553 "load_mt2.cpp"
     $$sum16 = (($dwMemPos$14102) + 48)|0; //@line 554 "load_mt2.cpp"
     $491 = (($lpStream) + ($$sum16)|0); //@line 554 "load_mt2.cpp"
     $492 = HEAPU8[$491>>0]|(HEAPU8[$491+1>>0]<<8)|(HEAPU8[$491+2>>0]<<16)|(HEAPU8[$491+3>>0]<<24); //@line 554 "load_mt2.cpp"
     $493 = (((((($this)) + 33280|0) + ($iSmp$099<<6)|0)) + 4|0); //@line 554 "load_mt2.cpp"
     HEAP32[$493>>2] = $492; //@line 554 "load_mt2.cpp"
     $$sum17 = (($dwMemPos$14102) + 52)|0; //@line 555 "load_mt2.cpp"
     $494 = (($lpStream) + ($$sum17)|0); //@line 555 "load_mt2.cpp"
     $495 = HEAPU8[$494>>0]|(HEAPU8[$494+1>>0]<<8)|(HEAPU8[$494+2>>0]<<16)|(HEAPU8[$494+3>>0]<<24); //@line 555 "load_mt2.cpp"
     $496 = (((((($this)) + 33280|0) + ($iSmp$099<<6)|0)) + 8|0); //@line 555 "load_mt2.cpp"
     HEAP32[$496>>2] = $495; //@line 555 "load_mt2.cpp"
     __ZN10CSoundFile20FrequencyToTransposeEP14_MODINSTRUMENT($472); //@line 556 "load_mt2.cpp"
     $$sum18 = (($dwMemPos$14102) + 59)|0; //@line 557 "load_mt2.cpp"
     $497 = (($lpStream) + ($$sum18)|0); //@line 557 "load_mt2.cpp"
     $498 = HEAP8[$497>>0]|0; //@line 557 "load_mt2.cpp"
     $499 = $498&255; //@line 557 "load_mt2.cpp"
     $500 = (((((($this)) + 33280|0) + ($iSmp$099<<6)|0)) + 36|0); //@line 557 "load_mt2.cpp"
     $501 = HEAP8[$500>>0]|0; //@line 557 "load_mt2.cpp"
     $502 = $501&255; //@line 557 "load_mt2.cpp"
     $$neg62 = (-207 - ($499))|0; //@line 557 "load_mt2.cpp"
     $503 = (($$neg62) + ($502))|0; //@line 557 "load_mt2.cpp"
     $504 = $503&255; //@line 557 "load_mt2.cpp"
     HEAP8[$500>>0] = $504; //@line 557 "load_mt2.cpp"
     $sext = $503 << 24; //@line 558 "load_mt2.cpp"
     $505 = $sext >> 24; //@line 558 "load_mt2.cpp"
     $506 = (((((($this)) + 33280|0) + ($iSmp$099<<6)|0)) + 37|0); //@line 558 "load_mt2.cpp"
     $507 = HEAP8[$506>>0]|0; //@line 558 "load_mt2.cpp"
     $508 = $507 << 24 >> 24; //@line 558 "load_mt2.cpp"
     $509 = (__ZN10CSoundFile20TransposeToFrequencyEii($505,$508)|0); //@line 558 "load_mt2.cpp"
     HEAP32[$490>>2] = $509; //@line 558 "load_mt2.cpp"
     $$sum19 = (($dwMemPos$14102) + 44)|0; //@line 559 "load_mt2.cpp"
     $510 = (($lpStream) + ($$sum19)|0); //@line 559 "load_mt2.cpp"
     $511 = HEAP8[$510>>0]|0; //@line 559 "load_mt2.cpp"
     $512 = ($511<<24>>24)==(2); //@line 559 "load_mt2.cpp"
     if ($512) {
      $513 = (((((($this)) + 33280|0) + ($iSmp$099<<6)|0)) + 34|0); //@line 559 "load_mt2.cpp"
      $514 = HEAP16[$513>>1]|0; //@line 559 "load_mt2.cpp"
      $515 = $514&65535; //@line 559 "load_mt2.cpp"
      $516 = $515 | 1; //@line 559 "load_mt2.cpp"
      $517 = $516&65535; //@line 559 "load_mt2.cpp"
      HEAP16[$513>>1] = $517; //@line 559 "load_mt2.cpp"
      $518 = HEAP32[$472>>2]|0; //@line 559 "load_mt2.cpp"
      $519 = $518 >>> 1; //@line 559 "load_mt2.cpp"
      HEAP32[$472>>2] = $519; //@line 559 "load_mt2.cpp"
     }
     $$sum20 = (($dwMemPos$14102) + 45)|0; //@line 560 "load_mt2.cpp"
     $520 = (($lpStream) + ($$sum20)|0); //@line 560 "load_mt2.cpp"
     $521 = HEAP8[$520>>0]|0; //@line 560 "load_mt2.cpp"
     $522 = ($521<<24>>24)==(2); //@line 560 "load_mt2.cpp"
     if ($522) {
      $523 = HEAP32[$472>>2]|0; //@line 560 "load_mt2.cpp"
      $524 = $523 >>> 1; //@line 560 "load_mt2.cpp"
      HEAP32[$472>>2] = $524; //@line 560 "load_mt2.cpp"
     }
     $$sum21 = (($dwMemPos$14102) + 47)|0; //@line 561 "load_mt2.cpp"
     $525 = (($lpStream) + ($$sum21)|0); //@line 561 "load_mt2.cpp"
     $526 = HEAP8[$525>>0]|0; //@line 561 "load_mt2.cpp"
     switch ($526<<24>>24) {
     case 1:  {
      $527 = (((((($this)) + 33280|0) + ($iSmp$099<<6)|0)) + 34|0); //@line 561 "load_mt2.cpp"
      $528 = HEAP16[$527>>1]|0; //@line 561 "load_mt2.cpp"
      $529 = $528&65535; //@line 561 "load_mt2.cpp"
      $530 = $529 | 2; //@line 561 "load_mt2.cpp"
      $531 = $530&65535; //@line 561 "load_mt2.cpp"
      HEAP16[$527>>1] = $531; //@line 561 "load_mt2.cpp"
      break L193;
      break;
     }
     case 2:  {
      $532 = (((((($this)) + 33280|0) + ($iSmp$099<<6)|0)) + 34|0); //@line 562 "load_mt2.cpp"
      $533 = HEAP16[$532>>1]|0; //@line 562 "load_mt2.cpp"
      $534 = $533&65535; //@line 562 "load_mt2.cpp"
      $535 = $534 | 6; //@line 562 "load_mt2.cpp"
      $536 = $535&65535; //@line 562 "load_mt2.cpp"
      HEAP16[$532>>1] = $536; //@line 562 "load_mt2.cpp"
      break L193;
      break;
     }
     default: {
      break L193;
     }
     }
    }
   } while(0);
   $537 = HEAPU8[$467>>0]|(HEAPU8[$467+1>>0]<<8)|(HEAPU8[$467+2>>0]<<16)|(HEAPU8[$467+3>>0]<<24); //@line 564 "load_mt2.cpp"
   $538 = (($462) + ($537))|0; //@line 564 "load_mt2.cpp"
   $dwMemPos$15 = $538;
  }
  $539 = (($iSmp$099) + 1)|0; //@line 532 "load_mt2.cpp"
  $540 = ($539>>>0)<(257); //@line 532 "load_mt2.cpp"
  if ($540) {
   $dwMemPos$14102 = $dwMemPos$15;$iSmp$099 = $539;
  } else {
   $dwMemPos$1698 = $dwMemPos$15;$iMap$097 = 0;
   break;
  }
 }
 if ((label|0) == 148) {
  STACKTOP = sp;return ($$0|0); //@line 635 "load_mt2.cpp"
 }
 L208: while(1) {
  $541 = (($InstrMap) + ($iMap$097<<2)|0); //@line 573 "load_mt2.cpp"
  $542 = HEAP32[$541>>2]|0; //@line 573 "load_mt2.cpp"
  $543 = ($542|0)==(0|0); //@line 573 "load_mt2.cpp"
  L210: do {
   if ($543) {
    $dwMemPos$18 = $dwMemPos$1698;
   } else {
    $544 = (($dwMemPos$1698) + 8)|0; //@line 575 "load_mt2.cpp"
    $545 = ($544>>>0)>($dwMemLength>>>0); //@line 575 "load_mt2.cpp"
    if ($545) {
     $$0 = 1;
     label = 148;
     break L208;
    }
    $546 = HEAP32[$276>>2]|0; //@line 578 "load_mt2.cpp"
    $547 = ($iMap$097>>>0)<($546>>>0); //@line 578 "load_mt2.cpp"
    if ($547) {
     $548 = (($iMap$097) + 1)|0; //@line 578 "load_mt2.cpp"
     $549 = (((($this)) + 48640|0) + ($548<<2)|0); //@line 578 "load_mt2.cpp"
     $550 = HEAP32[$549>>2]|0; //@line 578 "load_mt2.cpp"
     $penv8$0 = $550;
    } else {
     $penv8$0 = 0;
    }
    $551 = ((($542)) + 36|0); //@line 579 "load_mt2.cpp"
    $552 = HEAPU8[$551>>0]|(HEAPU8[$551+1>>0]<<8); //@line 579 "load_mt2.cpp"
    $553 = ($552<<16>>16)==(0); //@line 579 "load_mt2.cpp"
    if ($553) {
     $dwMemPos$18 = $dwMemPos$1698;
    } else {
     $554 = ($penv8$0|0)==(0|0); //@line 581 "load_mt2.cpp"
     $555 = ((($542)) + 134|0); //@line 592 "load_mt2.cpp"
     $556 = ((($542)) + 135|0); //@line 593 "load_mt2.cpp"
     $557 = ((($542)) + 136|0); //@line 594 "load_mt2.cpp"
     $558 = ((($542)) + 137|0); //@line 595 "load_mt2.cpp"
     if ($554) {
      $559 = $552&65535; //@line 579 "load_mt2.cpp"
      $dwMemPos$1796$us = $dwMemPos$1698;$iGrp$095$us = 0;
      while(1) {
       $560 = (($dwMemPos$1796$us) + 8)|0; //@line 600 "load_mt2.cpp"
       $561 = (($iGrp$095$us) + 1)|0; //@line 579 "load_mt2.cpp"
       $562 = ($561>>>0)<($559>>>0); //@line 579 "load_mt2.cpp"
       if ($562) {
        $dwMemPos$1796$us = $560;$iGrp$095$us = $561;
       } else {
        $dwMemPos$18 = $560;
        break L210;
       }
      }
     } else {
      $dwMemPos$1796 = $dwMemPos$1698;$iGrp$095 = 0;
     }
     while(1) {
      $563 = (($lpStream) + ($dwMemPos$1796)|0); //@line 583 "load_mt2.cpp"
      $i9$094 = 0;
      while(1) {
       $564 = (((($542)) + 38|0) + ($i9$094)|0); //@line 586 "load_mt2.cpp"
       $565 = HEAP8[$564>>0]|0; //@line 586 "load_mt2.cpp"
       $566 = $565&255; //@line 586 "load_mt2.cpp"
       $567 = ($566|0)==($iGrp$095|0); //@line 586 "load_mt2.cpp"
       do {
        if ($567) {
         $568 = HEAP8[$563>>0]|0; //@line 588 "load_mt2.cpp"
         $569 = $568&255; //@line 588 "load_mt2.cpp"
         $570 = (($569) + 1)|0; //@line 588 "load_mt2.cpp"
         $571 = $570&255; //@line 589 "load_mt2.cpp"
         $572 = (($i9$094) + 12)|0; //@line 589 "load_mt2.cpp"
         $573 = (((($penv8$0)) + 300|0) + ($572)|0); //@line 589 "load_mt2.cpp"
         HEAP8[$573>>0] = $571; //@line 589 "load_mt2.cpp"
         $574 = HEAP32[$461>>2]|0; //@line 590 "load_mt2.cpp"
         $575 = ($570>>>0)>($574>>>0); //@line 590 "load_mt2.cpp"
         if ($575) {
          break;
         }
         $576 = HEAP8[$555>>0]|0; //@line 592 "load_mt2.cpp"
         $577 = (((((($this)) + 33280|0) + ($570<<6)|0)) + 38|0); //@line 592 "load_mt2.cpp"
         HEAP8[$577>>0] = $576; //@line 592 "load_mt2.cpp"
         $578 = HEAP8[$556>>0]|0; //@line 593 "load_mt2.cpp"
         $579 = (((((($this)) + 33280|0) + ($570<<6)|0)) + 39|0); //@line 593 "load_mt2.cpp"
         HEAP8[$579>>0] = $578; //@line 593 "load_mt2.cpp"
         $580 = HEAP8[$557>>0]|0; //@line 594 "load_mt2.cpp"
         $581 = (((((($this)) + 33280|0) + ($570<<6)|0)) + 40|0); //@line 594 "load_mt2.cpp"
         HEAP8[$581>>0] = $580; //@line 594 "load_mt2.cpp"
         $582 = HEAP8[$558>>0]|0; //@line 595 "load_mt2.cpp"
         $583 = (((((($this)) + 33280|0) + ($570<<6)|0)) + 41|0); //@line 595 "load_mt2.cpp"
         HEAP8[$583>>0] = $582; //@line 595 "load_mt2.cpp"
        }
       } while(0);
       $584 = (($i9$094) + 1)|0; //@line 584 "load_mt2.cpp"
       $exitcond = ($584|0)==(96); //@line 584 "load_mt2.cpp"
       if ($exitcond) {
        break;
       } else {
        $i9$094 = $584;
       }
      }
      $$pre160 = HEAPU8[$551>>0]|(HEAPU8[$551+1>>0]<<8); //@line 579 "load_mt2.cpp"
      $585 = (($dwMemPos$1796) + 8)|0; //@line 600 "load_mt2.cpp"
      $586 = (($iGrp$095) + 1)|0; //@line 579 "load_mt2.cpp"
      $587 = $$pre160&65535; //@line 579 "load_mt2.cpp"
      $588 = ($586>>>0)<($587>>>0); //@line 579 "load_mt2.cpp"
      if ($588) {
       $dwMemPos$1796 = $585;$iGrp$095 = $586;
      } else {
       $dwMemPos$18 = $585;
       break;
      }
     }
    }
   }
  } while(0);
  $589 = (($iMap$097) + 1)|0; //@line 573 "load_mt2.cpp"
  $590 = ($589>>>0)<(255); //@line 573 "load_mt2.cpp"
  if ($590) {
   $dwMemPos$1698 = $dwMemPos$18;$iMap$097 = $589;
  } else {
   $dwMemPos$1993 = $dwMemPos$18;$iData$092 = 0;
   break;
  }
 }
 if ((label|0) == 148) {
  STACKTOP = sp;return ($$0|0); //@line 635 "load_mt2.cpp"
 }
 while(1) {
  $591 = HEAP32[$461>>2]|0; //@line 606 "load_mt2.cpp"
  $592 = ($iData$092>>>0)<($591>>>0); //@line 606 "load_mt2.cpp"
  if ($592) {
   $593 = (($SampleMap) + ($iData$092<<2)|0); //@line 606 "load_mt2.cpp"
   $594 = HEAP32[$593>>2]|0; //@line 606 "load_mt2.cpp"
   $595 = ($594|0)==(0|0); //@line 606 "load_mt2.cpp"
   if ($595) {
    $dwMemPos$21 = $dwMemPos$1993;
   } else {
    $596 = (($iData$092) + 1)|0; //@line 609 "load_mt2.cpp"
    $597 = (((($this)) + 33280|0) + ($596<<6)|0); //@line 609 "load_mt2.cpp"
    $598 = ((($594)) + 46|0); //@line 610 "load_mt2.cpp"
    $599 = HEAP8[$598>>0]|0; //@line 610 "load_mt2.cpp"
    $600 = $599 & 5; //@line 610 "load_mt2.cpp"
    $601 = ($600<<24>>24)==(0); //@line 610 "load_mt2.cpp"
    if ($601) {
     $602 = HEAP32[$597>>2]|0; //@line 612 "load_mt2.cpp"
     $603 = ($602|0)==(0); //@line 612 "load_mt2.cpp"
     if ($603) {
      $dwMemPos$20 = $dwMemPos$1993;
     } else {
      $604 = ((($594)) + 45|0); //@line 619 "load_mt2.cpp"
      $605 = HEAP8[$604>>0]|0; //@line 619 "load_mt2.cpp"
      $606 = ($605<<24>>24)==(2); //@line 619 "load_mt2.cpp"
      $607 = (((((($this)) + 33280|0) + ($596<<6)|0)) + 34|0); //@line 620 "load_mt2.cpp"
      $608 = HEAP16[$607>>1]|0; //@line 620 "load_mt2.cpp"
      $609 = $608 & 1; //@line 620 "load_mt2.cpp"
      $610 = $609&65535; //@line 620 "load_mt2.cpp"
      $611 = $610 << 1; //@line 620 "load_mt2.cpp"
      $rsflags$0$v = $606 ? 10 : 2; //@line 619 "load_mt2.cpp"
      $rsflags$0 = (($611) + ($rsflags$0$v))|0; //@line 619 "load_mt2.cpp"
      $612 = (($lpStream) + ($dwMemPos$1993)|0); //@line 624 "load_mt2.cpp"
      $613 = (($dwMemLength) - ($dwMemPos$1993))|0; //@line 624 "load_mt2.cpp"
      $614 = (__ZN10CSoundFile10ReadSampleEP14_MODINSTRUMENTjPKcj($this,$597,$rsflags$0,$612,$613)|0); //@line 624 "load_mt2.cpp"
      $615 = (($614) + ($dwMemPos$1993))|0; //@line 624 "load_mt2.cpp"
      $dwMemPos$20 = $615;
     }
    } else {
     $616 = (($dwMemPos$1993) + 4)|0; //@line 627 "load_mt2.cpp"
     $617 = ($616>>>0)<($dwMemLength>>>0); //@line 627 "load_mt2.cpp"
     if ($617) {
      $618 = (($lpStream) + ($dwMemPos$1993)|0); //@line 629 "load_mt2.cpp"
      $619 = HEAP32[$618>>2]|0; //@line 629 "load_mt2.cpp"
      $620 = (($dwMemPos$1993) + 16)|0; //@line 630 "load_mt2.cpp"
      $621 = (($620) + ($619))|0; //@line 630 "load_mt2.cpp"
      $dwMemPos$20 = $621;
     } else {
      $dwMemPos$20 = $dwMemPos$1993;
     }
    }
    $622 = (($dwMemPos$20) + 4)|0; //@line 632 "load_mt2.cpp"
    $623 = ($622>>>0)<($dwMemLength>>>0); //@line 632 "load_mt2.cpp"
    if ($623) {
     $dwMemPos$21 = $dwMemPos$20;
    } else {
     $$0 = 1;
     label = 148;
     break;
    }
   }
  } else {
   $dwMemPos$21 = $dwMemPos$1993;
  }
  $624 = (($iData$092) + 1)|0; //@line 606 "load_mt2.cpp"
  $625 = ($624>>>0)<(256); //@line 606 "load_mt2.cpp"
  if ($625) {
   $dwMemPos$1993 = $dwMemPos$21;$iData$092 = $624;
  } else {
   $$0 = 1;
   label = 148;
   break;
  }
 }
 if ((label|0) == 148) {
  STACKTOP = sp;return ($$0|0); //@line 635 "load_mt2.cpp"
 }
 return (0)|0;
}
function __ZL17ConvertMT2CommandP10CSoundFileP11_MODCOMMANDPK11_MT2COMMAND($that,$m,$p) {
 $that = $that|0;
 $m = $m|0;
 $p = $p|0;
 var $$off = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$m>>0] = 0; //@line 131 "load_mt2.cpp"
 $0 = HEAP8[$p>>0]|0; //@line 132 "load_mt2.cpp"
 $1 = ($0<<24>>24)==(0); //@line 132 "load_mt2.cpp"
 if (!($1)) {
  $2 = ($0&255)>(96); //@line 132 "load_mt2.cpp"
  if ($2) {
   $5 = -1;
  } else {
   $3 = $0&255; //@line 132 "load_mt2.cpp"
   $4 = (($3) + 12)|0; //@line 132 "load_mt2.cpp"
   $phitmp = $4&255; //@line 132 "load_mt2.cpp"
   $5 = $phitmp;
  }
  HEAP8[$m>>0] = $5; //@line 132 "load_mt2.cpp"
 }
 $6 = ((($p)) + 1|0); //@line 134 "load_mt2.cpp"
 $7 = HEAP8[$6>>0]|0; //@line 134 "load_mt2.cpp"
 $8 = ((($m)) + 1|0); //@line 134 "load_mt2.cpp"
 HEAP8[$8>>0] = $7; //@line 134 "load_mt2.cpp"
 $9 = ((($p)) + 2|0); //@line 136 "load_mt2.cpp"
 $10 = HEAP8[$9>>0]|0; //@line 136 "load_mt2.cpp"
 $$off = (($10) + -16)<<24>>24; //@line 136 "load_mt2.cpp"
 $11 = ($$off&255)<(129); //@line 136 "load_mt2.cpp"
 L7: do {
  if ($11) {
   $12 = ((($m)) + 2|0); //@line 138 "load_mt2.cpp"
   HEAP8[$12>>0] = 1; //@line 138 "load_mt2.cpp"
   $13 = $10&255; //@line 139 "load_mt2.cpp"
   $14 = (($13) + 496)|0; //@line 139 "load_mt2.cpp"
   $15 = $14 >>> 1; //@line 139 "load_mt2.cpp"
   $16 = $15&255; //@line 139 "load_mt2.cpp"
   $17 = ((($m)) + 4|0); //@line 139 "load_mt2.cpp"
   HEAP8[$17>>0] = $16; //@line 139 "load_mt2.cpp"
  } else {
   $18 = $10 & -16; //@line 141 "load_mt2.cpp"
   switch ($18<<24>>24) {
   case -96:  {
    $19 = ((($m)) + 2|0); //@line 143 "load_mt2.cpp"
    HEAP8[$19>>0] = 4; //@line 143 "load_mt2.cpp"
    $20 = $10&255; //@line 144 "load_mt2.cpp"
    $21 = $20 & 15; //@line 144 "load_mt2.cpp"
    $22 = $21&255; //@line 144 "load_mt2.cpp"
    $23 = ((($m)) + 4|0); //@line 144 "load_mt2.cpp"
    HEAP8[$23>>0] = $22; //@line 144 "load_mt2.cpp"
    break L7;
    break;
   }
   case -80:  {
    $24 = ((($m)) + 2|0); //@line 148 "load_mt2.cpp"
    HEAP8[$24>>0] = 3; //@line 148 "load_mt2.cpp"
    $25 = $10&255; //@line 149 "load_mt2.cpp"
    $26 = $25 & 15; //@line 149 "load_mt2.cpp"
    $27 = $26&255; //@line 149 "load_mt2.cpp"
    $28 = ((($m)) + 4|0); //@line 149 "load_mt2.cpp"
    HEAP8[$28>>0] = $27; //@line 149 "load_mt2.cpp"
    break L7;
    break;
   }
   case -64:  {
    $29 = ((($m)) + 2|0); //@line 153 "load_mt2.cpp"
    HEAP8[$29>>0] = 6; //@line 153 "load_mt2.cpp"
    $30 = $10&255; //@line 154 "load_mt2.cpp"
    $31 = $30 & 15; //@line 154 "load_mt2.cpp"
    $32 = $31&255; //@line 154 "load_mt2.cpp"
    $33 = ((($m)) + 4|0); //@line 154 "load_mt2.cpp"
    HEAP8[$33>>0] = $32; //@line 154 "load_mt2.cpp"
    break L7;
    break;
   }
   default: {
    $34 = ($18<<24>>24)==(-48); //@line 156 "load_mt2.cpp"
    $35 = ((($m)) + 2|0); //@line 158 "load_mt2.cpp"
    if ($34) {
     HEAP8[$35>>0] = 5; //@line 158 "load_mt2.cpp"
     $36 = $10&255; //@line 159 "load_mt2.cpp"
     $37 = $36 & 15; //@line 159 "load_mt2.cpp"
     $38 = $37&255; //@line 159 "load_mt2.cpp"
     $39 = ((($m)) + 4|0); //@line 159 "load_mt2.cpp"
     HEAP8[$39>>0] = $38; //@line 159 "load_mt2.cpp"
     break L7;
    } else {
     HEAP8[$35>>0] = 0; //@line 162 "load_mt2.cpp"
     $40 = ((($m)) + 4|0); //@line 163 "load_mt2.cpp"
     HEAP8[$40>>0] = 0; //@line 163 "load_mt2.cpp"
     break L7;
    }
   }
   }
  }
 } while(0);
 $41 = ((($m)) + 3|0); //@line 166 "load_mt2.cpp"
 HEAP8[$41>>0] = 0; //@line 166 "load_mt2.cpp"
 $42 = ((($m)) + 5|0); //@line 167 "load_mt2.cpp"
 HEAP8[$42>>0] = 0; //@line 167 "load_mt2.cpp"
 $43 = ((($p)) + 4|0); //@line 168 "load_mt2.cpp"
 $44 = HEAP8[$43>>0]|0; //@line 168 "load_mt2.cpp"
 $45 = ($44<<24>>24)==(0); //@line 168 "load_mt2.cpp"
 if (!($45)) {
  return; //@line 180 "load_mt2.cpp"
 }
 $46 = ((($p)) + 5|0); //@line 168 "load_mt2.cpp"
 $47 = HEAP8[$46>>0]|0; //@line 168 "load_mt2.cpp"
 $48 = ($47<<24>>24)==(0); //@line 168 "load_mt2.cpp"
 if ($48) {
  $49 = ((($p)) + 6|0); //@line 168 "load_mt2.cpp"
  $50 = HEAP8[$49>>0]|0; //@line 168 "load_mt2.cpp"
  $51 = ($50<<24>>24)==(0); //@line 168 "load_mt2.cpp"
  if ($51) {
   return; //@line 180 "load_mt2.cpp"
  }
 }
 $52 = ((($p)) + 6|0); //@line 172 "load_mt2.cpp"
 $53 = HEAP8[$52>>0]|0; //@line 172 "load_mt2.cpp"
 HEAP8[$41>>0] = $53; //@line 172 "load_mt2.cpp"
 HEAP8[$42>>0] = $47; //@line 173 "load_mt2.cpp"
 __ZNK10CSoundFile17ConvertModCommandEP11_MODCOMMAND($that,$m); //@line 174 "load_mt2.cpp"
 return; //@line 180 "load_mt2.cpp"
}
function __ZN10CSoundFile20FrequencyToTransposeEP14_MODINSTRUMENT($psmp) {
 $psmp = $psmp|0;
 var $$transp$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $ftune$0 = 0, $transp$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($psmp)) + 24|0); //@line 1716 "sndfile.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 1716 "sndfile.cpp"
 $2 = (+($1>>>0)); //@line 1708 "sndfile.cpp"
 $3 = $2 * 1.1957431543704412E-4; //@line 1708 "sndfile.cpp"
 $4 = (+Math_log((+$3))); //@line 1708 "sndfile.cpp"
 $5 = $4 * 2215.9795828054475; //@line 1708 "sndfile.cpp"
 $6 = (~~(($5))); //@line 1708 "sndfile.cpp"
 $7 = $6 >> 7; //@line 1717 "sndfile.cpp"
 $8 = $6 & 127; //@line 1718 "sndfile.cpp"
 $9 = ($8>>>0)>(80); //@line 1719 "sndfile.cpp"
 $10 = $6 | 128; //@line 1722 "sndfile.cpp"
 $11 = $9&1; //@line 1719 "sndfile.cpp"
 $transp$0 = (($11) + ($7))|0; //@line 1719 "sndfile.cpp"
 $ftune$0 = $9 ? $10 : $8; //@line 1719 "sndfile.cpp"
 $12 = ($transp$0|0)>(127); //@line 1724 "sndfile.cpp"
 $$transp$0 = $12 ? 127 : $transp$0; //@line 1724 "sndfile.cpp"
 $13 = ($$transp$0|0)<(-127); //@line 1725 "sndfile.cpp"
 $14 = $$transp$0&255; //@line 1726 "sndfile.cpp"
 $15 = $13 ? -127 : $14; //@line 1726 "sndfile.cpp"
 $16 = ((($psmp)) + 36|0); //@line 1726 "sndfile.cpp"
 HEAP8[$16>>0] = $15; //@line 1726 "sndfile.cpp"
 $17 = $ftune$0&255; //@line 1727 "sndfile.cpp"
 $18 = ((($psmp)) + 37|0); //@line 1727 "sndfile.cpp"
 HEAP8[$18>>0] = $17; //@line 1727 "sndfile.cpp"
 return; //@line 1728 "sndfile.cpp"
}
function __ZN10CSoundFile20TransposeToFrequencyEii($transp,$ftune) {
 $transp = $transp|0;
 $ftune = $ftune|0;
 var $0 = 0, $1 = 0, $2 = 0.0, $3 = 0.0, $4 = 0.0, $5 = 0, $exp2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $transp << 7; //@line 1681 "sndfile.cpp"
 $1 = (($0) + ($ftune))|0; //@line 1681 "sndfile.cpp"
 $2 = (+($1|0)); //@line 1681 "sndfile.cpp"
 $3 = $2 * 6.5104166666666663E-4; //@line 1681 "sndfile.cpp"
 $exp2 = (+_exp2($3)); //@line 1681 "sndfile.cpp"
 $4 = $exp2 * 8363.0; //@line 1681 "sndfile.cpp"
 $5 = (~~(($4))>>>0); //@line 1681 "sndfile.cpp"
 return ($5|0); //@line 1681 "sndfile.cpp"
}
function __ZN10CSoundFile13SetWaveConfigEjjjb($nRate,$nBits,$nChannels,$bMMX) {
 $nRate = $nRate|0;
 $nBits = $nBits|0;
 $nChannels = $nChannels|0;
 $bMMX = $bMMX|0;
 var $$ = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $not$or$cond5 = 0, $or$cond = 0, $or$cond3 = 0, $or$cond5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[183080]|0; //@line 472 "sndfile.cpp"
 $1 = $0 & -131073; //@line 472 "sndfile.cpp"
 $2 = $0 | 131072; //@line 473 "sndfile.cpp"
 $$ = $bMMX ? $2 : $1; //@line 473 "sndfile.cpp"
 $3 = HEAP32[10109]|0; //@line 474 "sndfile.cpp"
 $4 = ($3|0)==($nRate|0); //@line 474 "sndfile.cpp"
 $5 = HEAP32[10110]|0;
 $6 = ($5|0)==($nBits|0); //@line 474 "sndfile.cpp"
 $or$cond = $4 & $6; //@line 474 "sndfile.cpp"
 $7 = HEAP32[10108]|0;
 $8 = ($7|0)==($nChannels|0); //@line 474 "sndfile.cpp"
 $or$cond3 = $or$cond & $8; //@line 474 "sndfile.cpp"
 $9 = ($$|0)==($0|0); //@line 474 "sndfile.cpp"
 $or$cond5 = $9 & $or$cond3; //@line 474 "sndfile.cpp"
 $not$or$cond5 = $or$cond5 ^ 1; //@line 474 "sndfile.cpp"
 HEAP32[10108] = $nChannels; //@line 475 "sndfile.cpp"
 HEAP32[183080] = $$; //@line 476 "sndfile.cpp"
 HEAP32[10109] = $nRate; //@line 477 "sndfile.cpp"
 HEAP32[10110] = $nBits; //@line 478 "sndfile.cpp"
 (__ZN10CSoundFile10InitPlayerEb($not$or$cond5)|0); //@line 479 "sndfile.cpp"
 return 1; //@line 480 "sndfile.cpp"
}
function __ZN10CSoundFile12SetMixConfigEjj($nStereoSeparation,$nMaxMixChannels) {
 $nStereoSeparation = $nStereoSeparation|0;
 $nMaxMixChannels = $nMaxMixChannels|0;
 var $$0 = 0, $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($nMaxMixChannels>>>0)<(2); //@line 486 "sndfile.cpp"
 if ($0) {
  $$0 = 0;
  return ($$0|0); //@line 491 "sndfile.cpp"
 }
 HEAP32[10107] = $nMaxMixChannels; //@line 488 "sndfile.cpp"
 HEAP32[10106] = $nStereoSeparation; //@line 489 "sndfile.cpp"
 $$0 = 1;
 return ($$0|0); //@line 491 "sndfile.cpp"
}
function __ZN10CSoundFile17SetResamplingModeEj($nMode) {
 $nMode = $nMode|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $d$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[183080]|0; //@line 497 "sndfile.cpp"
 $1 = $0 & -1049; //@line 497 "sndfile.cpp"
 switch ($nMode|0) {
 case 0:  {
  $2 = $1 | 8; //@line 500 "sndfile.cpp"
  $d$0 = $2;
  break;
 }
 case 1:  {
  $d$0 = $1;
  break;
 }
 case 2:  {
  $3 = $1 | 16; //@line 502 "sndfile.cpp"
  $d$0 = $3;
  break;
 }
 case 3:  {
  $4 = $1 | 1040; //@line 503 "sndfile.cpp"
  $d$0 = $4;
  break;
 }
 default: {
  $$0 = 0;
  return ($$0|0); //@line 509 "sndfile.cpp"
 }
 }
 HEAP32[183080] = $d$0; //@line 507 "sndfile.cpp"
 $$0 = 1;
 return ($$0|0); //@line 509 "sndfile.cpp"
}
function __ZN10CSoundFile15SetMasterVolumeEjb($this,$nVol,$bAdjustAGC) {
 $this = $this|0;
 $nVol = $nVol|0;
 $bAdjustAGC = $bAdjustAGC|0;
 var $$ = 0, $$1 = 0, $$nVol = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $bAdjustAGC$not = 0, $brmerge = 0, $or$cond = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ($nVol|0)==(0); //@line 515 "sndfile.cpp"
 $$nVol = $0 ? 1 : $nVol; //@line 515 "sndfile.cpp"
 $1 = ($$nVol>>>0)>(512); //@line 516 "sndfile.cpp"
 $$1 = $1 ? 512 : $$nVol; //@line 516 "sndfile.cpp"
 $2 = ((($this)) + 59740|0); //@line 517 "sndfile.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 517 "sndfile.cpp"
 $4 = ($$1>>>0)<($3>>>0); //@line 517 "sndfile.cpp"
 $5 = ($$1|0)!=(0); //@line 517 "sndfile.cpp"
 $or$cond = $5 & $4; //@line 517 "sndfile.cpp"
 if (!($or$cond)) {
  HEAP32[$2>>2] = $$1; //@line 522 "sndfile.cpp"
  return 1; //@line 523 "sndfile.cpp"
 }
 $6 = HEAP32[183080]|0; //@line 517 "sndfile.cpp"
 $7 = $6 & 4; //@line 517 "sndfile.cpp"
 $8 = ($7|0)==(0); //@line 517 "sndfile.cpp"
 $bAdjustAGC$not = $bAdjustAGC ^ 1; //@line 517 "sndfile.cpp"
 $brmerge = $8 | $bAdjustAGC$not; //@line 517 "sndfile.cpp"
 if ($brmerge) {
  HEAP32[$2>>2] = $$1; //@line 522 "sndfile.cpp"
  return 1; //@line 523 "sndfile.cpp"
 }
 $9 = HEAP32[10111]|0; //@line 519 "sndfile.cpp"
 $10 = Math_imul($9, $3)|0; //@line 519 "sndfile.cpp"
 $11 = (($10>>>0) / ($$1>>>0))&-1; //@line 519 "sndfile.cpp"
 $12 = ($11>>>0)>(512); //@line 520 "sndfile.cpp"
 $$ = $12 ? 512 : $11; //@line 520 "sndfile.cpp"
 HEAP32[10111] = $$; //@line 520 "sndfile.cpp"
 HEAP32[$2>>2] = $$1; //@line 522 "sndfile.cpp"
 return 1; //@line 523 "sndfile.cpp"
}
function __ZNK10CSoundFile14GetMaxPositionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$02 = 0, $max$0$lcssa = 0, $max$01 = 0, $max$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $i$02 = 0;$max$01 = 0;
 while(1) {
  $0 = (((($this)) + 53344|0) + ($i$02)|0); //@line 565 "sndfile.cpp"
  $1 = HEAP8[$0>>0]|0; //@line 565 "sndfile.cpp"
  $2 = $1&255; //@line 565 "sndfile.cpp"
  $3 = ($1<<24>>24)==(-1); //@line 565 "sndfile.cpp"
  if ($3) {
   $max$0$lcssa = $max$01;
   label = 6;
   break;
  }
  $4 = ($1&255)<(240); //@line 567 "sndfile.cpp"
  if ($4) {
   $5 = (((($this)) + 52864|0) + ($2<<1)|0); //@line 567 "sndfile.cpp"
   $6 = HEAP16[$5>>1]|0; //@line 567 "sndfile.cpp"
   $7 = $6&65535; //@line 567 "sndfile.cpp"
   $8 = (($7) + ($max$01))|0; //@line 567 "sndfile.cpp"
   $max$1 = $8;
  } else {
   $max$1 = $max$01;
  }
  $9 = (($i$02) + 1)|0; //@line 568 "sndfile.cpp"
  $10 = ($9>>>0)<(256); //@line 565 "sndfile.cpp"
  if ($10) {
   $i$02 = $9;$max$01 = $max$1;
  } else {
   $max$0$lcssa = $max$1;
   label = 6;
   break;
  }
 }
 if ((label|0) == 6) {
  return ($max$0$lcssa|0); //@line 570 "sndfile.cpp"
 }
 return (0)|0;
}
function __ZN10CSoundFile13SetCurrentPosEj($this,$nPos) {
 $this = $this|0;
 $nPos = $nPos|0;
 var $$017 = 0, $$017$lcssa46 = 0, $$1 = 0, $$lcssa48 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $exitcond = 0, $exitcond22 = 0, $i$019 = 0, $i$118 = 0, $k$09 = 0, $n$08 = 0, $nPattern$016 = 0, $nPattern$016$lcssa47 = 0, $nPattern$13 = 0, $nRow$0$ = 0, $nRow$012 = 0, $nRow$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $i$019 = 0;
 while(1) {
  $0 = (((($this) + ($i$019<<8)|0)) + 215|0); //@line 592 "sndfile.cpp"
  HEAP8[$0>>0] = 0; //@line 592 "sndfile.cpp"
  $1 = (((($this) + ($i$019<<8)|0)) + 214|0); //@line 592 "sndfile.cpp"
  HEAP8[$1>>0] = 0; //@line 592 "sndfile.cpp"
  $2 = (((($this) + ($i$019<<8)|0)) + 212|0); //@line 592 "sndfile.cpp"
  HEAP8[$2>>0] = 0; //@line 592 "sndfile.cpp"
  $3 = (((($this) + ($i$019<<8)|0)) + 144|0); //@line 593 "sndfile.cpp"
  HEAP32[$3>>2] = 0; //@line 593 "sndfile.cpp"
  $4 = (((($this) + ($i$019<<8)|0)) + 140|0); //@line 594 "sndfile.cpp"
  HEAP32[$4>>2] = 0; //@line 594 "sndfile.cpp"
  $5 = (((($this) + ($i$019<<8)|0)) + 136|0); //@line 595 "sndfile.cpp"
  HEAP32[$5>>2] = 0; //@line 595 "sndfile.cpp"
  $6 = (((($this) + ($i$019<<8)|0)) + 216|0); //@line 596 "sndfile.cpp"
  HEAP8[$6>>0] = 0; //@line 596 "sndfile.cpp"
  $7 = (((($this) + ($i$019<<8)|0)) + 245|0); //@line 597 "sndfile.cpp"
  HEAP8[$7>>0] = 0; //@line 597 "sndfile.cpp"
  $8 = (((($this) + ($i$019<<8)|0)) + 244|0); //@line 598 "sndfile.cpp"
  HEAP8[$8>>0] = 0; //@line 598 "sndfile.cpp"
  $9 = (((($this) + ($i$019<<8)|0)) + 124|0); //@line 599 "sndfile.cpp"
  HEAP32[$9>>2] = 0; //@line 599 "sndfile.cpp"
  $10 = (((($this) + ($i$019<<8)|0)) + 36|0); //@line 600 "sndfile.cpp"
  $11 = HEAP32[$10>>2]|0; //@line 600 "sndfile.cpp"
  $12 = $11 | 1536; //@line 600 "sndfile.cpp"
  HEAP32[$10>>2] = $12; //@line 600 "sndfile.cpp"
  $13 = (((($this) + ($i$019<<8)|0)) + 242|0); //@line 601 "sndfile.cpp"
  HEAP8[$13>>0] = 0; //@line 601 "sndfile.cpp"
  $14 = (($i$019) + 1)|0; //@line 590 "sndfile.cpp"
  $exitcond22 = ($14|0)==(128); //@line 590 "sndfile.cpp"
  if ($exitcond22) {
   break;
  } else {
   $i$019 = $14;
  }
 }
 $15 = ($nPos|0)==(0); //@line 603 "sndfile.cpp"
 if ($15) {
  $i$118 = 0;
  while(1) {
   $16 = (((($this) + ($i$118<<8)|0)) + 128|0); //@line 607 "sndfile.cpp"
   HEAP32[$16>>2] = 0; //@line 607 "sndfile.cpp"
   $17 = (((($this) + ($i$118<<8)|0)) + 32|0); //@line 608 "sndfile.cpp"
   HEAP32[$17>>2] = 0; //@line 608 "sndfile.cpp"
   $18 = (((($this) + ($i$118<<8)|0)) + 4|0); //@line 608 "sndfile.cpp"
   HEAP32[$18>>2] = 0; //@line 608 "sndfile.cpp"
   $19 = (((($this) + ($i$118<<8)|0)) + 40|0); //@line 609 "sndfile.cpp"
   HEAP32[$19>>2] = 0; //@line 609 "sndfile.cpp"
   $20 = (((($this) + ($i$118<<8)|0)) + 44|0); //@line 610 "sndfile.cpp"
   HEAP32[$20>>2] = 0; //@line 610 "sndfile.cpp"
   $21 = (((($this) + ($i$118<<8)|0)) + 88|0); //@line 611 "sndfile.cpp"
   HEAP32[$21>>2] = 0; //@line 611 "sndfile.cpp"
   $22 = (((($this) + ($i$118<<8)|0)) + 84|0); //@line 611 "sndfile.cpp"
   HEAP32[$22>>2] = 0; //@line 611 "sndfile.cpp"
   $23 = (((($this) + ($i$118<<8)|0)) + 96|0); //@line 612 "sndfile.cpp"
   HEAP32[$23>>2] = 0; //@line 612 "sndfile.cpp"
   $24 = (((($this) + ($i$118<<8)|0)) + 144|0); //@line 613 "sndfile.cpp"
   HEAP32[$24>>2] = 0; //@line 613 "sndfile.cpp"
   $25 = (((($this) + ($i$118<<8)|0)) + 140|0); //@line 614 "sndfile.cpp"
   HEAP32[$25>>2] = 0; //@line 614 "sndfile.cpp"
   $26 = (((($this) + ($i$118<<8)|0)) + 238|0); //@line 615 "sndfile.cpp"
   HEAP8[$26>>0] = 127; //@line 615 "sndfile.cpp"
   $27 = (((($this) + ($i$118<<8)|0)) + 239|0); //@line 616 "sndfile.cpp"
   HEAP8[$27>>0] = 0; //@line 616 "sndfile.cpp"
   $28 = (((($this) + ($i$118<<8)|0)) + 16|0); //@line 617 "sndfile.cpp"
   $29 = (((($this) + ($i$118<<8)|0)) + 100|0); //@line 618 "sndfile.cpp"
   HEAP32[$29>>2] = 0; //@line 618 "sndfile.cpp"
   $30 = (((($this) + ($i$118<<8)|0)) + 104|0); //@line 618 "sndfile.cpp"
   HEAP32[$30>>2] = 0; //@line 618 "sndfile.cpp"
   $31 = (((($this) + ($i$118<<8)|0)) + 116|0); //@line 620 "sndfile.cpp"
   ;HEAP32[$28>>2]=0|0;HEAP32[$28+4>>2]=0|0;HEAP32[$28+8>>2]=0|0;HEAP32[$28+12>>2]=0|0; //@line 617 "sndfile.cpp"
   HEAP32[$31>>2] = 256; //@line 620 "sndfile.cpp"
   $32 = ($i$118>>>0)<(64); //@line 621 "sndfile.cpp"
   if ($32) {
    $33 = (((((($this)) + 49600|0) + (($i$118*36)|0)|0)) + 8|0); //@line 623 "sndfile.cpp"
    $34 = HEAP32[$33>>2]|0; //@line 623 "sndfile.cpp"
    $35 = (((($this) + ($i$118<<8)|0)) + 36|0); //@line 623 "sndfile.cpp"
    HEAP32[$35>>2] = $34; //@line 623 "sndfile.cpp"
    $36 = (((($this)) + 49600|0) + (($i$118*36)|0)|0); //@line 624 "sndfile.cpp"
    $37 = HEAP32[$36>>2]|0; //@line 624 "sndfile.cpp"
    $38 = (((($this) + ($i$118<<8)|0)) + 120|0); //@line 624 "sndfile.cpp"
    HEAP32[$38>>2] = $37; //@line 624 "sndfile.cpp"
    $39 = (((((($this)) + 49600|0) + (($i$118*36)|0)|0)) + 4|0); //@line 625 "sndfile.cpp"
    $40 = HEAP32[$39>>2]|0; //@line 625 "sndfile.cpp"
    $$sink = $40;
   } else {
    $41 = (((($this) + ($i$118<<8)|0)) + 36|0); //@line 628 "sndfile.cpp"
    HEAP32[$41>>2] = 0; //@line 628 "sndfile.cpp"
    $42 = (((($this) + ($i$118<<8)|0)) + 120|0); //@line 629 "sndfile.cpp"
    HEAP32[$42>>2] = 128; //@line 629 "sndfile.cpp"
    $$sink = 64;
   }
   $43 = (((($this) + ($i$118<<8)|0)) + 168|0); //@line 625 "sndfile.cpp"
   HEAP32[$43>>2] = $$sink; //@line 625 "sndfile.cpp"
   $44 = (($i$118) + 1)|0; //@line 605 "sndfile.cpp"
   $exitcond = ($44|0)==(128); //@line 605 "sndfile.cpp"
   if ($exitcond) {
    break;
   } else {
    $i$118 = $44;
   }
  }
  $45 = ((($this)) + 59656|0); //@line 633 "sndfile.cpp"
  $46 = HEAP32[$45>>2]|0; //@line 633 "sndfile.cpp"
  $47 = ((($this)) + 59744|0); //@line 633 "sndfile.cpp"
  HEAP32[$47>>2] = $46; //@line 633 "sndfile.cpp"
  $48 = ((($this)) + 59648|0); //@line 634 "sndfile.cpp"
  $49 = HEAP32[$48>>2]|0; //@line 634 "sndfile.cpp"
  $50 = ((($this)) + 59708|0); //@line 634 "sndfile.cpp"
  HEAP32[$50>>2] = $49; //@line 634 "sndfile.cpp"
  $51 = ((($this)) + 59652|0); //@line 635 "sndfile.cpp"
  $52 = HEAP32[$51>>2]|0; //@line 635 "sndfile.cpp"
  $53 = ((($this)) + 59712|0); //@line 635 "sndfile.cpp"
  HEAP32[$53>>2] = $52; //@line 635 "sndfile.cpp"
 }
 $54 = ((($this)) + 59660|0); //@line 637 "sndfile.cpp"
 $55 = HEAP32[$54>>2]|0; //@line 637 "sndfile.cpp"
 $56 = $55 & -3873; //@line 637 "sndfile.cpp"
 HEAP32[$54>>2] = $56; //@line 637 "sndfile.cpp"
 $$017 = $nPos;$nPattern$016 = 0;
 L13: while(1) {
  $57 = (((($this)) + 53344|0) + ($nPattern$016)|0); //@line 640 "sndfile.cpp"
  $58 = HEAP8[$57>>0]|0; //@line 640 "sndfile.cpp"
  $59 = $58&255; //@line 640 "sndfile.cpp"
  switch ($58<<24>>24) {
  case -1:  {
   $nPattern$13 = 0;$nRow$1 = 0;
   break L13;
   break;
  }
  case -2:  {
   $$1 = $$017;
   break;
  }
  default: {
   $60 = ($58&255)<(240); //@line 643 "sndfile.cpp"
   if ($60) {
    $61 = (((($this)) + 52864|0) + ($59<<1)|0); //@line 645 "sndfile.cpp"
    $62 = HEAP16[$61>>1]|0; //@line 645 "sndfile.cpp"
    $63 = $62&65535; //@line 645 "sndfile.cpp"
    $64 = ($$017>>>0)<($63>>>0); //@line 645 "sndfile.cpp"
    if ($64) {
     $$017$lcssa46 = $$017;$$lcssa48 = $59;$nPattern$016$lcssa47 = $nPattern$016;
     label = 15;
     break L13;
    }
    $65 = (($$017) - ($63))|0; //@line 646 "sndfile.cpp"
    $$1 = $65;
   } else {
    $$1 = $$017;
   }
  }
  }
  $66 = (($nPattern$016) + 1)|0; //@line 638 "sndfile.cpp"
  $67 = ($66>>>0)<(256); //@line 638 "sndfile.cpp"
  if ($67) {
   $$017 = $$1;$nPattern$016 = $66;
  } else {
   $nPattern$13 = 0;$nRow$1 = 0;
   break;
  }
 }
 L20: do {
  if ((label|0) == 15) {
   $68 = (((($this)) + 52864|0) + ($$lcssa48<<1)|0); //@line 652 "sndfile.cpp"
   $69 = HEAP16[$68>>1]|0; //@line 652 "sndfile.cpp"
   $70 = $69&65535; //@line 652 "sndfile.cpp"
   $71 = ($$017$lcssa46>>>0)<($70>>>0); //@line 652 "sndfile.cpp"
   if ($71) {
    $72 = ($$017$lcssa46|0)==(0); //@line 658 "sndfile.cpp"
    if ($72) {
     $nPattern$13 = $nPattern$016$lcssa47;$nRow$1 = 0;
    } else {
     $73 = (((($this)) + 51904|0) + ($$lcssa48<<2)|0); //@line 660 "sndfile.cpp"
     $74 = HEAP32[$73>>2]|0; //@line 660 "sndfile.cpp"
     $75 = ($74|0)==(0|0); //@line 661 "sndfile.cpp"
     if ($75) {
      $nPattern$13 = $nPattern$016$lcssa47;$nRow$1 = $$017$lcssa46;
     } else {
      $76 = ((($this)) + 59664|0); //@line 666 "sndfile.cpp"
      $77 = HEAP32[$76>>2]|0; //@line 666 "sndfile.cpp"
      $78 = ($77|0)==(0); //@line 667 "sndfile.cpp"
      $nRow$012 = $$017$lcssa46;
      while(1) {
       if (!($78)) {
        $79 = Math_imul($77, $nRow$012)|0; //@line 666 "sndfile.cpp"
        $k$09 = 0;$n$08 = $79;
        while(1) {
         $80 = (($74) + (($n$08*6)|0)|0); //@line 669 "sndfile.cpp"
         $81 = HEAP8[$80>>0]|0; //@line 669 "sndfile.cpp"
         $82 = ($81<<24>>24)==(0); //@line 669 "sndfile.cpp"
         if (!($82)) {
          $nPattern$13 = $nPattern$016$lcssa47;$nRow$1 = $nRow$012;
          break L20;
         }
         $83 = (($k$09) + 1)|0; //@line 667 "sndfile.cpp"
         $84 = (($n$08) + 1)|0; //@line 667 "sndfile.cpp"
         $85 = ($83>>>0)<($77>>>0); //@line 667 "sndfile.cpp"
         if ($85) {
          $k$09 = $83;$n$08 = $84;
         } else {
          break;
         }
        }
       }
       $nRow$0$ = (($nRow$012) + -1)|0; //@line 675 "sndfile.cpp"
       $86 = ($nRow$0$|0)==(0); //@line 664 "sndfile.cpp"
       if ($86) {
        $nPattern$13 = $nPattern$016$lcssa47;$nRow$1 = $nRow$0$;
        break;
       } else {
        $nRow$012 = $nRow$0$;
       }
      }
     }
    }
   } else {
    $nPattern$13 = 0;$nRow$1 = 0;
   }
  }
 } while(0);
 $87 = ((($this)) + 59732|0); //@line 679 "sndfile.cpp"
 HEAP32[$87>>2] = $nPattern$13; //@line 679 "sndfile.cpp"
 $88 = ((($this)) + 59716|0); //@line 680 "sndfile.cpp"
 HEAP32[$88>>2] = $nRow$1; //@line 680 "sndfile.cpp"
 $89 = ((($this)) + 59708|0); //@line 681 "sndfile.cpp"
 $90 = HEAP32[$89>>2]|0; //@line 681 "sndfile.cpp"
 $91 = ((($this)) + 59692|0); //@line 681 "sndfile.cpp"
 HEAP32[$91>>2] = $90; //@line 681 "sndfile.cpp"
 $92 = ((($this)) + 59676|0); //@line 682 "sndfile.cpp"
 HEAP32[$92>>2] = 0; //@line 682 "sndfile.cpp"
 $93 = ((($this)) + 59700|0); //@line 683 "sndfile.cpp"
 HEAP32[$93>>2] = 0; //@line 683 "sndfile.cpp"
 $94 = ((($this)) + 59704|0); //@line 684 "sndfile.cpp"
 HEAP32[$94>>2] = 0; //@line 684 "sndfile.cpp"
 return; //@line 685 "sndfile.cpp"
}
function __ZN10CSoundFile9GetLengthEbb($this,$bAdjust,$bTotal) {
 $this = $this|0;
 $bAdjust = $bAdjust|0;
 $bTotal = $bTotal|0;
 var $$ = 0, $$$nGlbVol$2 = 0, $$10 = 0, $$12 = 0, $$12$v = 0, $$13 = 0, $$13$v = 0, $$14 = 0, $$14$v = 0, $$18 = 0, $$19 = 0, $$4 = 0, $$lobit = 0, $$nGlbVol$2 = 0, $$off = 0, $$param$2 = 0, $$param$2$op = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0;
 var $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0;
 var $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bAdjust$not = 0, $brmerge = 0, $chnvols = 0;
 var $dwElapsedTime$0 = 0, $dwElapsedTime$0$lcssa209 = 0, $dwElapsedTime$1$lcssa = 0, $dwElapsedTime$184 = 0, $dwElapsedTime$2 = 0, $dwElapsedTime$3 = 0, $dwElapsedTime$4 = 0, $dwElapsedTime$434 = 0, $icv$096 = 0, $instr = 0, $ipck$069 = 0, $n$063 = 0, $nChn$081 = 0, $nGlbVol$0 = 0, $nGlbVol$0$lcssa203 = 0, $nGlbVol$1$lcssa = 0, $nGlbVol$175 = 0, $nGlbVol$2 = 0, $nGlbVol$4 = 0, $nGlbVol$5 = 0;
 var $nMusicSpeed$0 = 0, $nMusicSpeed$0$lcssa197 = 0, $nMusicSpeed$1$lcssa = 0, $nMusicSpeed$172 = 0, $nMusicSpeed$2 = 0, $nMusicSpeed$3 = 0, $nMusicTempo$0 = 0, $nMusicTempo$0$lcssa200 = 0, $nMusicTempo$1$lcssa = 0, $nMusicTempo$173 = 0, $nMusicTempo$2 = 0, $nMusicTempo$3 = 0, $nNextPattern$0 = 0, $nNextPattern$1$lcssa = 0, $nNextPattern$165163 = 0, $nNextPattern$2 = 0, $nNextPattern$3$lcssa = 0, $nNextPattern$371 = 0, $nNextPattern$5 = 0, $nNextRow$0 = 0;
 var $nNextRow$0$ = 0, $nNextRow$1 = 0, $nNextRow$2$lcssa = 0, $nNextRow$274 = 0, $nNextRow$4 = 0, $nOldGlbVolSlide$0 = 0, $nOldGlbVolSlide$0$lcssa206 = 0, $nOldGlbVolSlide$1$lcssa = 0, $nOldGlbVolSlide$1$param$2 = 0, $nOldGlbVolSlide$176 = 0, $nOldGlbVolSlide$3 = 0, $nOldGlbVolSlide$4 = 0, $nPattern$0$lcssa = 0, $nPattern$066$be = 0, $nSpeedCount$0$lcssa = 0, $nSpeedCount$077 = 0, $nSpeedCount$1 = 0, $nSpeedCount$2 = 0, $not$ = 0, $notes = 0;
 var $oldparam = 0, $or$cond = 0, $or$cond11 = 0, $or$cond140 = 0, $or$cond16 = 0, $or$cond162 = 0, $or$cond17 = 0, $or$cond6 = 0, $or$cond7 = 0, $or$cond9 = 0, $p$080 = 0, $pChn$078 = 0, $param$0 = 0, $param$1 = 0, $param$10 = 0, $param$11 = 0, $param$2 = 0, $patloop = 0, $vols = 0, dest = 0;
 var label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1152|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $instr = sp + 1024|0;
 $notes = sp + 896|0;
 $vols = sp + 768|0;
 $oldparam = sp + 640|0;
 $chnvols = sp + 512|0;
 $patloop = sp;
 $0 = ((($this)) + 59648|0); //@line 23 "snd_fx.cpp"
 $1 = HEAP32[$0>>2]|0; //@line 23 "snd_fx.cpp"
 $2 = ((($this)) + 59652|0); //@line 23 "snd_fx.cpp"
 $3 = HEAP32[$2>>2]|0; //@line 23 "snd_fx.cpp"
 $4 = ((($this)) + 59656|0); //@line 25 "snd_fx.cpp"
 $5 = HEAP32[$4>>2]|0; //@line 25 "snd_fx.cpp"
 dest=$instr; stop=dest+128|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 34 "snd_fx.cpp"
 dest=$notes; stop=dest+128|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 35 "snd_fx.cpp"
 dest=$vols; stop=dest+128|0; do { HEAP8[dest>>0]=255|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 36 "snd_fx.cpp"
 _memset(($patloop|0),0,512)|0; //@line 37 "snd_fx.cpp"
 dest=$oldparam; stop=dest+128|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 38 "snd_fx.cpp"
 dest=$chnvols; stop=dest+128|0; do { HEAP8[dest>>0]=64|0; dest=dest+1|0; } while ((dest|0) < (stop|0)); //@line 39 "snd_fx.cpp"
 $6 = ((($this)) + 59664|0); //@line 41 "snd_fx.cpp"
 $7 = HEAP32[$6>>2]|0; //@line 41 "snd_fx.cpp"
 $8 = ($7|0)==(0); //@line 41 "snd_fx.cpp"
 if (!($8)) {
  $icv$096 = 0;
  while(1) {
   $9 = (((((($this)) + 49600|0) + (($icv$096*36)|0)|0)) + 4|0); //@line 41 "snd_fx.cpp"
   $10 = HEAP32[$9>>2]|0; //@line 41 "snd_fx.cpp"
   $11 = $10&255; //@line 41 "snd_fx.cpp"
   $12 = (($chnvols) + ($icv$096)|0); //@line 41 "snd_fx.cpp"
   HEAP8[$12>>0] = $11; //@line 41 "snd_fx.cpp"
   $13 = (($icv$096) + 1)|0; //@line 41 "snd_fx.cpp"
   $14 = ($13>>>0)<($7>>>0); //@line 41 "snd_fx.cpp"
   if ($14) {
    $icv$096 = $13;
   } else {
    break;
   }
  }
 }
 $15 = ((($this)) + 59716|0); //@line 42 "snd_fx.cpp"
 $16 = HEAP32[$15>>2]|0; //@line 42 "snd_fx.cpp"
 $17 = ((($this)) + 59732|0); //@line 43 "snd_fx.cpp"
 $18 = HEAP32[$17>>2]|0; //@line 43 "snd_fx.cpp"
 $19 = ((($this)) + 59680|0); //@line 194 "snd_fx.cpp"
 $46 = $7;$dwElapsedTime$0 = 0;$nGlbVol$0 = $5;$nMusicSpeed$0 = $1;$nMusicTempo$0 = $3;$nNextPattern$0 = 0;$nNextRow$0 = 0;$nOldGlbVolSlide$0 = 0;
 L5: while(1) {
  $20 = (((($this)) + 53344|0) + ($nNextPattern$0)|0); //@line 53 "snd_fx.cpp"
  $21 = HEAP8[$20>>0]|0; //@line 53 "snd_fx.cpp"
  $22 = $21&255; //@line 53 "snd_fx.cpp"
  $23 = ($21&255)>(239); //@line 54 "snd_fx.cpp"
  L7: do {
   if ($23) {
    $24 = ($21<<24>>24)==(-1); //@line 57 "snd_fx.cpp"
    $25 = ($nNextPattern$0>>>0)>(255); //@line 57 "snd_fx.cpp"
    $or$cond162 = $25 | $24; //@line 57 "snd_fx.cpp"
    if ($or$cond162) {
     $dwElapsedTime$4 = $dwElapsedTime$0;$nGlbVol$5 = $nGlbVol$0;$nOldGlbVolSlide$4 = $nOldGlbVolSlide$0;
     break L5;
    } else {
     $nNextPattern$165163 = $nNextPattern$0;
    }
    while(1) {
     $26 = (($nNextPattern$165163) + 1)|0; //@line 62 "snd_fx.cpp"
     $27 = ($26>>>0)<(256); //@line 63 "snd_fx.cpp"
     if ($27) {
      $28 = (((($this)) + 53344|0) + ($26)|0); //@line 63 "snd_fx.cpp"
      $29 = HEAP8[$28>>0]|0; //@line 63 "snd_fx.cpp"
      $30 = $29&255; //@line 63 "snd_fx.cpp"
      $31 = ($29&255)>(239); //@line 54 "snd_fx.cpp"
      if ($31) {
       $nPattern$066$be = $30;
      } else {
       $nNextPattern$1$lcssa = $26;$nPattern$0$lcssa = $30;
       break L7;
      }
     } else {
      $nPattern$066$be = 255;
     }
     $32 = ($nPattern$066$be|0)==(255); //@line 57 "snd_fx.cpp"
     $33 = ($26>>>0)>(255); //@line 57 "snd_fx.cpp"
     $or$cond = $33 | $32; //@line 57 "snd_fx.cpp"
     if ($or$cond) {
      $dwElapsedTime$4 = $dwElapsedTime$0;$nGlbVol$5 = $nGlbVol$0;$nOldGlbVolSlide$4 = $nOldGlbVolSlide$0;
      break L5;
     } else {
      $nNextPattern$165163 = $26;
     }
    }
   } else {
    $nNextPattern$1$lcssa = $nNextPattern$0;$nPattern$0$lcssa = $22;
   }
  } while(0);
  $34 = (((($this)) + 51904|0) + ($nPattern$0$lcssa<<2)|0); //@line 68 "snd_fx.cpp"
  $35 = HEAP32[$34>>2]|0; //@line 68 "snd_fx.cpp"
  $36 = ($35|0)==(0|0); //@line 68 "snd_fx.cpp"
  if ($36) {
   $dwElapsedTime$4 = $dwElapsedTime$0;$nGlbVol$5 = $nGlbVol$0;$nOldGlbVolSlide$4 = $nOldGlbVolSlide$0;
   break;
  }
  $37 = (((($this)) + 52864|0) + ($nPattern$0$lcssa<<1)|0); //@line 70 "snd_fx.cpp"
  $38 = HEAP16[$37>>1]|0; //@line 70 "snd_fx.cpp"
  $39 = $38&65535; //@line 70 "snd_fx.cpp"
  $40 = ($nNextRow$0>>>0)<($39>>>0); //@line 70 "snd_fx.cpp"
  $nNextRow$0$ = $40 ? $nNextRow$0 : 0; //@line 70 "snd_fx.cpp"
  $41 = (($nNextRow$0$) + 1)|0; //@line 72 "snd_fx.cpp"
  $42 = ($41>>>0)<($39>>>0); //@line 73 "snd_fx.cpp"
  $43 = (($nNextPattern$1$lcssa) + 1)|0; //@line 75 "snd_fx.cpp"
  $nNextPattern$2 = $42 ? $nNextPattern$1$lcssa : $43; //@line 73 "snd_fx.cpp"
  $nNextRow$1 = $42 ? $41 : 0; //@line 73 "snd_fx.cpp"
  $44 = ($nNextRow$0$|0)!=(0); //@line 78 "snd_fx.cpp"
  $45 = ($46|0)==(0); //@line 80 "snd_fx.cpp"
  $or$cond140 = $44 | $45; //@line 78 "snd_fx.cpp"
  if (!($or$cond140)) {
   $ipck$069 = 0;
   while(1) {
    $47 = (($patloop) + ($ipck$069<<2)|0); //@line 80 "snd_fx.cpp"
    HEAP32[$47>>2] = $dwElapsedTime$0; //@line 80 "snd_fx.cpp"
    $48 = (($ipck$069) + 1)|0; //@line 80 "snd_fx.cpp"
    $49 = ($48>>>0)<($46>>>0); //@line 80 "snd_fx.cpp"
    if ($49) {
     $ipck$069 = $48;
    } else {
     break;
    }
   }
  }
  if (!($bTotal)) {
   $50 = ($nNextPattern$1$lcssa>>>0)>($18>>>0); //@line 84 "snd_fx.cpp"
   if ($50) {
    $dwElapsedTime$0$lcssa209 = $dwElapsedTime$0;$nGlbVol$0$lcssa203 = $nGlbVol$0;$nMusicSpeed$0$lcssa197 = $nMusicSpeed$0;$nMusicTempo$0$lcssa200 = $nMusicTempo$0;$nOldGlbVolSlide$0$lcssa206 = $nOldGlbVolSlide$0;
    label = 15;
    break;
   }
   $51 = ($nNextPattern$1$lcssa|0)!=($18|0); //@line 84 "snd_fx.cpp"
   $52 = ($nNextRow$0$>>>0)<($16>>>0); //@line 84 "snd_fx.cpp"
   $or$cond7 = $51 | $52; //@line 84 "snd_fx.cpp"
   if (!($or$cond7)) {
    $dwElapsedTime$0$lcssa209 = $dwElapsedTime$0;$nGlbVol$0$lcssa203 = $nGlbVol$0;$nMusicSpeed$0$lcssa197 = $nMusicSpeed$0;$nMusicTempo$0$lcssa200 = $nMusicTempo$0;$nOldGlbVolSlide$0$lcssa206 = $nOldGlbVolSlide$0;
    label = 15;
    break;
   }
  }
  $55 = ($46|0)==(0); //@line 96 "snd_fx.cpp"
  if ($55) {
   $238 = 0;$dwElapsedTime$1$lcssa = $dwElapsedTime$0;$nGlbVol$1$lcssa = $nGlbVol$0;$nMusicSpeed$1$lcssa = $nMusicSpeed$0;$nMusicTempo$1$lcssa = $nMusicTempo$0;$nNextPattern$3$lcssa = $nNextPattern$2;$nNextRow$2$lcssa = $nNextRow$1;$nOldGlbVolSlide$1$lcssa = $nOldGlbVolSlide$0;$nSpeedCount$0$lcssa = 0;
  } else {
   $56 = Math_imul($7, $nNextRow$0$)|0; //@line 95 "snd_fx.cpp"
   $57 = (($35) + (($56*6)|0)|0); //@line 95 "snd_fx.cpp"
   $dwElapsedTime$184 = $dwElapsedTime$0;$nChn$081 = 0;$nGlbVol$175 = $nGlbVol$0;$nMusicSpeed$172 = $nMusicSpeed$0;$nMusicTempo$173 = $nMusicTempo$0;$nNextPattern$371 = $nNextPattern$2;$nNextRow$274 = $nNextRow$1;$nOldGlbVolSlide$176 = $nOldGlbVolSlide$0;$nSpeedCount$077 = 0;$p$080 = $57;$pChn$078 = $this;
   while(1) {
    $58 = HEAP32[$p$080>>2]|0; //@line 96 "snd_fx.cpp"
    $59 = ($58|0)==(0); //@line 96 "snd_fx.cpp"
    $60 = $58&255;
    L28: do {
     if ($59) {
      $dwElapsedTime$3 = $dwElapsedTime$184;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$172;$nMusicTempo$3 = $nMusicTempo$173;$nNextPattern$5 = $nNextPattern$371;$nNextRow$4 = $nNextRow$274;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$077;
     } else {
      $61 = ((($p$080)) + 3|0); //@line 98 "snd_fx.cpp"
      $62 = HEAP8[$61>>0]|0; //@line 98 "snd_fx.cpp"
      $63 = $62&255; //@line 98 "snd_fx.cpp"
      $64 = ((($p$080)) + 5|0); //@line 99 "snd_fx.cpp"
      $65 = HEAP8[$64>>0]|0; //@line 99 "snd_fx.cpp"
      $66 = $65&255; //@line 99 "snd_fx.cpp"
      $67 = ((($p$080)) + 1|0); //@line 101 "snd_fx.cpp"
      $68 = HEAP8[$67>>0]|0; //@line 101 "snd_fx.cpp"
      $69 = ($68<<24>>24)==(0); //@line 101 "snd_fx.cpp"
      if (!($69)) {
       $70 = (($instr) + ($nChn$081)|0); //@line 101 "snd_fx.cpp"
       HEAP8[$70>>0] = $68; //@line 101 "snd_fx.cpp"
       $71 = (($notes) + ($nChn$081)|0); //@line 101 "snd_fx.cpp"
       HEAP8[$71>>0] = 0; //@line 101 "snd_fx.cpp"
       $72 = (($vols) + ($nChn$081)|0); //@line 101 "snd_fx.cpp"
       HEAP8[$72>>0] = -1; //@line 101 "snd_fx.cpp"
      }
      $$off = (($60) + -1)<<24>>24; //@line 102 "snd_fx.cpp"
      $73 = ($$off&255)<(120); //@line 102 "snd_fx.cpp"
      if ($73) {
       $74 = (($notes) + ($nChn$081)|0); //@line 102 "snd_fx.cpp"
       HEAP8[$74>>0] = $60; //@line 102 "snd_fx.cpp"
      }
      $75 = ((($p$080)) + 2|0); //@line 103 "snd_fx.cpp"
      $76 = HEAP8[$75>>0]|0; //@line 103 "snd_fx.cpp"
      $77 = ($76<<24>>24)==(1); //@line 103 "snd_fx.cpp"
      if ($77) {
       $78 = ((($p$080)) + 4|0); //@line 103 "snd_fx.cpp"
       $79 = HEAP8[$78>>0]|0; //@line 103 "snd_fx.cpp"
       $80 = (($vols) + ($nChn$081)|0); //@line 103 "snd_fx.cpp"
       HEAP8[$80>>0] = $79; //@line 103 "snd_fx.cpp"
      }
      $81 = ($62<<24>>24)==(0); //@line 104 "snd_fx.cpp"
      L39: do {
       if ($81) {
        $dwElapsedTime$2 = $dwElapsedTime$184;$nMusicSpeed$2 = $nMusicSpeed$172;$nMusicTempo$2 = $nMusicTempo$173;$nSpeedCount$1 = $nSpeedCount$077;$param$2 = $66;
       } else {
        L41: do {
         switch ($63|0) {
         case 12:  {
          $82 = ($66>>>0)>($nNextPattern$1$lcssa>>>0); //@line 108 "snd_fx.cpp"
          if (!($82)) {
           $dwElapsedTime$4 = $dwElapsedTime$184;$nGlbVol$5 = $nGlbVol$175;$nOldGlbVolSlide$4 = $nOldGlbVolSlide$176;
           break L5;
          }
          if (!($bAdjust)) {
           $dwElapsedTime$3 = $dwElapsedTime$184;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$172;$nMusicTempo$3 = $nMusicTempo$173;$nNextPattern$5 = $66;$nNextRow$4 = 0;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$077;
           break L28;
          }
          $83 = ((($pChn$078)) + 245|0); //@line 113 "snd_fx.cpp"
          HEAP8[$83>>0] = 0; //@line 113 "snd_fx.cpp"
          $84 = ((($pChn$078)) + 244|0); //@line 114 "snd_fx.cpp"
          HEAP8[$84>>0] = 0; //@line 114 "snd_fx.cpp"
          $dwElapsedTime$3 = $dwElapsedTime$184;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$172;$nMusicTempo$3 = $nMusicTempo$173;$nNextPattern$5 = $66;$nNextRow$4 = 0;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$077;
          break L28;
          break;
         }
         case 14:  {
          if (!($bAdjust)) {
           $dwElapsedTime$3 = $dwElapsedTime$184;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$172;$nMusicTempo$3 = $nMusicTempo$173;$nNextPattern$5 = $43;$nNextRow$4 = $66;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$077;
           break L28;
          }
          $85 = ((($pChn$078)) + 245|0); //@line 123 "snd_fx.cpp"
          HEAP8[$85>>0] = 0; //@line 123 "snd_fx.cpp"
          $86 = ((($pChn$078)) + 244|0); //@line 124 "snd_fx.cpp"
          HEAP8[$86>>0] = 0; //@line 124 "snd_fx.cpp"
          $dwElapsedTime$3 = $dwElapsedTime$184;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$172;$nMusicTempo$3 = $nMusicTempo$173;$nNextPattern$5 = $43;$nNextRow$4 = $66;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$077;
          break L28;
          break;
         }
         case 16:  {
          $87 = ($65<<24>>24)==(0); //@line 129 "snd_fx.cpp"
          if ($87) {
           $dwElapsedTime$2 = $dwElapsedTime$184;$nMusicSpeed$2 = $nMusicSpeed$172;$nMusicTempo$2 = $nMusicTempo$173;$nSpeedCount$1 = $nSpeedCount$077;$param$2 = $66;
           break L39;
          }
          $88 = ($65&255)<(33); //@line 130 "snd_fx.cpp"
          if (!($88)) {
           $89 = HEAP32[$19>>2]|0; //@line 130 "snd_fx.cpp"
           $90 = ($89|0)!=(1); //@line 130 "snd_fx.cpp"
           $91 = ($65<<24>>24)>(-1); //@line 132 "snd_fx.cpp"
           $or$cond6 = $91 & $90; //@line 130 "snd_fx.cpp"
           if (!($or$cond6)) {
            $dwElapsedTime$2 = $dwElapsedTime$184;$nMusicSpeed$2 = $nMusicSpeed$172;$nMusicTempo$2 = $nMusicTempo$173;$nSpeedCount$1 = $nSpeedCount$077;$param$2 = $66;
            break L39;
           }
          }
          $dwElapsedTime$2 = $dwElapsedTime$184;$nMusicSpeed$2 = $66;$nMusicTempo$2 = $nMusicTempo$173;$nSpeedCount$1 = $nSpeedCount$077;$param$2 = $66;
          break L39;
          break;
         }
         case 17:  {
          do {
           if ($bAdjust) {
            $92 = HEAP32[$19>>2]|0; //@line 137 "snd_fx.cpp"
            $93 = $92 & 34; //@line 137 "snd_fx.cpp"
            $94 = ($93|0)==(0); //@line 137 "snd_fx.cpp"
            if ($94) {
             $param$0 = $66;
            } else {
             $95 = ($65<<24>>24)==(0); //@line 139 "snd_fx.cpp"
             $96 = ((($pChn$078)) + 235|0); //@line 139 "snd_fx.cpp"
             if ($95) {
              $97 = HEAP8[$96>>0]|0; //@line 139 "snd_fx.cpp"
              $98 = $97&255; //@line 139 "snd_fx.cpp"
              $param$0 = $98;
              break;
             } else {
              HEAP8[$96>>0] = $65; //@line 139 "snd_fx.cpp"
              $param$0 = $66;
              break;
             }
            }
           } else {
            $param$0 = $66;
           }
          } while(0);
          $99 = ($param$0>>>0)>(31); //@line 141 "snd_fx.cpp"
          if ($99) {
           $dwElapsedTime$2 = $dwElapsedTime$184;$nMusicSpeed$2 = $nMusicSpeed$172;$nMusicTempo$2 = $param$0;$nSpeedCount$1 = $nSpeedCount$077;$param$2 = $param$0;
           break L39;
          }
          $100 = $param$0 & 240; //@line 143 "snd_fx.cpp"
          $101 = ($100|0)==(16); //@line 143 "snd_fx.cpp"
          $102 = $param$0 & 15; //@line 145 "snd_fx.cpp"
          if ($101) {
           $103 = (($102) + ($nMusicTempo$173))|0; //@line 145 "snd_fx.cpp"
           $104 = ($103>>>0)>(255); //@line 146 "snd_fx.cpp"
           $$ = $104 ? 255 : $103; //@line 146 "snd_fx.cpp"
           $dwElapsedTime$2 = $dwElapsedTime$184;$nMusicSpeed$2 = $nMusicSpeed$172;$nMusicTempo$2 = $$;$nSpeedCount$1 = $nSpeedCount$077;$param$2 = $param$0;
           break L39;
          } else {
           $105 = (($nMusicTempo$173) - ($102))|0; //@line 149 "snd_fx.cpp"
           $106 = ($105>>>0)<(32); //@line 150 "snd_fx.cpp"
           $$4 = $106 ? 32 : $105; //@line 150 "snd_fx.cpp"
           $dwElapsedTime$2 = $dwElapsedTime$184;$nMusicSpeed$2 = $nMusicSpeed$172;$nMusicTempo$2 = $$4;$nSpeedCount$1 = $nSpeedCount$077;$param$2 = $param$0;
           break L39;
          }
          break;
         }
         case 20:  {
          $107 = $66 & 240; //@line 155 "snd_fx.cpp"
          switch ($107|0) {
          case 96:  {
           $108 = $66 & 15; //@line 155 "snd_fx.cpp"
           $dwElapsedTime$2 = $dwElapsedTime$184;$nMusicSpeed$2 = $nMusicSpeed$172;$nMusicTempo$2 = $nMusicTempo$173;$nSpeedCount$1 = $108;$param$2 = $66;
           break L39;
           break;
          }
          case 176:  {
           $109 = $66 & 15; //@line 156 "snd_fx.cpp"
           $110 = $109 | 96; //@line 156 "snd_fx.cpp"
           $param$1 = $110;
           break L41;
           break;
          }
          default: {
           $param$1 = $66;
           break L41;
          }
          }
          break;
         }
         case 19:  {
          $param$1 = $66;
          break;
         }
         default: {
          $dwElapsedTime$2 = $dwElapsedTime$184;$nMusicSpeed$2 = $nMusicSpeed$172;$nMusicTempo$2 = $nMusicTempo$173;$nSpeedCount$1 = $nSpeedCount$077;$param$2 = $66;
          break L39;
         }
         }
        } while(0);
        $111 = $param$1 & 240; //@line 158 "snd_fx.cpp"
        switch ($111|0) {
        case 224:  {
         $112 = $param$1 & 15; //@line 158 "snd_fx.cpp"
         $113 = Math_imul($112, $nMusicSpeed$172)|0; //@line 158 "snd_fx.cpp"
         $dwElapsedTime$2 = $dwElapsedTime$184;$nMusicSpeed$2 = $nMusicSpeed$172;$nMusicTempo$2 = $nMusicTempo$173;$nSpeedCount$1 = $113;$param$2 = $param$1;
         break L39;
         break;
        }
        case 96:  {
         break;
        }
        default: {
         $dwElapsedTime$2 = $dwElapsedTime$184;$nMusicSpeed$2 = $nMusicSpeed$172;$nMusicTempo$2 = $nMusicTempo$173;$nSpeedCount$1 = $nSpeedCount$077;$param$2 = $param$1;
         break L39;
        }
        }
        $114 = $param$1 & 15; //@line 161 "snd_fx.cpp"
        $115 = ($114|0)==(0); //@line 161 "snd_fx.cpp"
        $116 = (($patloop) + ($nChn$081<<2)|0); //@line 161 "snd_fx.cpp"
        if ($115) {
         HEAP32[$116>>2] = $dwElapsedTime$184; //@line 162 "snd_fx.cpp"
         $dwElapsedTime$2 = $dwElapsedTime$184;$nMusicSpeed$2 = $nMusicSpeed$172;$nMusicTempo$2 = $nMusicTempo$173;$nSpeedCount$1 = $nSpeedCount$077;$param$2 = $param$1;
         break;
        } else {
         $117 = HEAP32[$116>>2]|0; //@line 161 "snd_fx.cpp"
         $118 = (($dwElapsedTime$184) - ($117))|0; //@line 161 "snd_fx.cpp"
         $119 = Math_imul($118, $114)|0; //@line 161 "snd_fx.cpp"
         $120 = (($119) + ($dwElapsedTime$184))|0; //@line 161 "snd_fx.cpp"
         $dwElapsedTime$2 = $120;$nMusicSpeed$2 = $nMusicSpeed$172;$nMusicTempo$2 = $nMusicTempo$173;$nSpeedCount$1 = $nSpeedCount$077;$param$2 = $param$1;
         break;
        }
       }
      } while(0);
      if ($bAdjust) {
       switch ($63|0) {
       case 3: case 2:  {
        $121 = ($param$2|0)==(0); //@line 172 "snd_fx.cpp"
        if ($121) {
         $dwElapsedTime$3 = $dwElapsedTime$2;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$2;$nMusicTempo$3 = $nMusicTempo$2;$nNextPattern$5 = $nNextPattern$371;$nNextRow$4 = $nNextRow$274;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$1;
         break L28;
        }
        $122 = $param$2&255; //@line 172 "snd_fx.cpp"
        $123 = ((($pChn$078)) + 220|0); //@line 172 "snd_fx.cpp"
        HEAP8[$123>>0] = $122; //@line 172 "snd_fx.cpp"
        $dwElapsedTime$3 = $dwElapsedTime$2;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$2;$nMusicTempo$3 = $nMusicTempo$2;$nNextPattern$5 = $nNextPattern$371;$nNextRow$4 = $nNextRow$274;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$1;
        break L28;
        break;
       }
       case 4:  {
        $124 = ($param$2|0)==(0); //@line 176 "snd_fx.cpp"
        if ($124) {
         $dwElapsedTime$3 = $dwElapsedTime$2;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$2;$nMusicTempo$3 = $nMusicTempo$2;$nNextPattern$5 = $nNextPattern$371;$nNextRow$4 = $nNextRow$274;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$1;
         break L28;
        }
        $125 = $param$2 << 2; //@line 176 "snd_fx.cpp"
        $126 = ((($pChn$078)) + 184|0); //@line 176 "snd_fx.cpp"
        HEAP32[$126>>2] = $125; //@line 176 "snd_fx.cpp"
        $dwElapsedTime$3 = $dwElapsedTime$2;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$2;$nMusicTempo$3 = $nMusicTempo$2;$nNextPattern$5 = $nNextPattern$371;$nNextRow$4 = $nNextRow$274;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$1;
        break L28;
        break;
       }
       case 10:  {
        $127 = ($param$2|0)==(0); //@line 180 "snd_fx.cpp"
        if ($127) {
         $dwElapsedTime$3 = $dwElapsedTime$2;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$2;$nMusicTempo$3 = $nMusicTempo$2;$nNextPattern$5 = $nNextPattern$371;$nNextRow$4 = $nNextRow$274;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$1;
         break L28;
        }
        $128 = $param$2&255; //@line 180 "snd_fx.cpp"
        $129 = ((($pChn$078)) + 236|0); //@line 180 "snd_fx.cpp"
        HEAP8[$129>>0] = $128; //@line 180 "snd_fx.cpp"
        $dwElapsedTime$3 = $dwElapsedTime$2;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$2;$nMusicTempo$3 = $nMusicTempo$2;$nNextPattern$5 = $nNextPattern$371;$nNextRow$4 = $nNextRow$274;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$1;
        break L28;
        break;
       }
       case 7: case 6: case 11:  {
        $130 = ($param$2|0)==(0); //@line 186 "snd_fx.cpp"
        if ($130) {
         $dwElapsedTime$3 = $dwElapsedTime$2;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$2;$nMusicTempo$3 = $nMusicTempo$2;$nNextPattern$5 = $nNextPattern$371;$nNextRow$4 = $nNextRow$274;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$1;
         break L28;
        }
        $131 = $param$2&255; //@line 186 "snd_fx.cpp"
        $132 = ((($pChn$078)) + 218|0); //@line 186 "snd_fx.cpp"
        HEAP8[$132>>0] = $131; //@line 186 "snd_fx.cpp"
        $dwElapsedTime$3 = $dwElapsedTime$2;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$2;$nMusicTempo$3 = $nMusicTempo$2;$nNextPattern$5 = $nNextPattern$371;$nNextRow$4 = $nNextRow$274;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$1;
        break L28;
        break;
       }
       case 13:  {
        $133 = $param$2&255; //@line 190 "snd_fx.cpp"
        $134 = (($vols) + ($nChn$081)|0); //@line 190 "snd_fx.cpp"
        HEAP8[$134>>0] = $133; //@line 190 "snd_fx.cpp"
        $dwElapsedTime$3 = $dwElapsedTime$2;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$2;$nMusicTempo$3 = $nMusicTempo$2;$nNextPattern$5 = $nNextPattern$371;$nNextRow$4 = $nNextRow$274;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$1;
        break L28;
        break;
       }
       case 23:  {
        $135 = HEAP32[$19>>2]|0; //@line 194 "snd_fx.cpp"
        $136 = $135 >>> 5; //@line 194 "snd_fx.cpp"
        $$lobit = $136 & 1; //@line 194 "snd_fx.cpp"
        $137 = $$lobit ^ 1; //@line 194 "snd_fx.cpp"
        $$param$2 = $param$2 << $137; //@line 194 "snd_fx.cpp"
        $138 = ($$param$2>>>0)>(128); //@line 195 "snd_fx.cpp"
        $$param$2$op = $$param$2 << 1; //@line 196 "snd_fx.cpp"
        $139 = $138 ? 256 : $$param$2$op; //@line 196 "snd_fx.cpp"
        $dwElapsedTime$3 = $dwElapsedTime$2;$nGlbVol$4 = $139;$nMusicSpeed$3 = $nMusicSpeed$2;$nMusicTempo$3 = $nMusicTempo$2;$nNextPattern$5 = $nNextPattern$371;$nNextRow$4 = $nNextRow$274;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$1;
        break L28;
        break;
       }
       case 24:  {
        $140 = ($param$2|0)==(0); //@line 200 "snd_fx.cpp"
        $nOldGlbVolSlide$1$param$2 = $140 ? $nOldGlbVolSlide$176 : $param$2;
        $141 = $nOldGlbVolSlide$1$param$2 & 15; //@line 201 "snd_fx.cpp"
        $142 = ($141|0)!=(15); //@line 201 "snd_fx.cpp"
        $143 = $nOldGlbVolSlide$1$param$2 & 240;
        $144 = ($143|0)==(0); //@line 201 "snd_fx.cpp"
        $or$cond9 = $142 | $144; //@line 201 "snd_fx.cpp"
        do {
         if ($or$cond9) {
          $150 = ($143|0)!=(240); //@line 207 "snd_fx.cpp"
          $151 = ($141|0)==(0); //@line 207 "snd_fx.cpp"
          $or$cond11 = $150 | $151; //@line 207 "snd_fx.cpp"
          if (!($or$cond11)) {
           $152 = HEAP32[$19>>2]|0; //@line 210 "snd_fx.cpp"
           $153 = ($152|0)==(32); //@line 210 "snd_fx.cpp"
           $$12$v = $153 ? 1 : 2; //@line 210 "snd_fx.cpp"
           $$12 = $141 << $$12$v; //@line 210 "snd_fx.cpp"
           $154 = (($nGlbVol$175) - ($$12))|0; //@line 211 "snd_fx.cpp"
           $nGlbVol$2 = $154;
           break;
          }
          if ($144) {
           $160 = HEAP32[$19>>2]|0; //@line 222 "snd_fx.cpp"
           $161 = ($160|0)==(32); //@line 222 "snd_fx.cpp"
           $$14$v = $161 ? 1 : 2; //@line 222 "snd_fx.cpp"
           $$14 = $141 << $$14$v; //@line 222 "snd_fx.cpp"
           $162 = Math_imul($$14, $nMusicSpeed$2)|0; //@line 223 "snd_fx.cpp"
           $163 = (($nGlbVol$175) - ($162))|0; //@line 223 "snd_fx.cpp"
           $nGlbVol$2 = $163;
           break;
          } else {
           $155 = $nOldGlbVolSlide$1$param$2 >>> 4; //@line 215 "snd_fx.cpp"
           $156 = HEAP32[$19>>2]|0; //@line 217 "snd_fx.cpp"
           $157 = ($156|0)==(32); //@line 217 "snd_fx.cpp"
           $$13$v = $157 ? 1 : 2; //@line 217 "snd_fx.cpp"
           $$13 = $155 << $$13$v; //@line 217 "snd_fx.cpp"
           $158 = Math_imul($$13, $nMusicSpeed$2)|0; //@line 218 "snd_fx.cpp"
           $159 = (($158) + ($nGlbVol$175))|0; //@line 218 "snd_fx.cpp"
           $nGlbVol$2 = $159;
           break;
          }
         } else {
          $145 = $nOldGlbVolSlide$1$param$2 >>> 4; //@line 203 "snd_fx.cpp"
          $146 = HEAP32[$19>>2]|0; //@line 204 "snd_fx.cpp"
          $not$ = ($146|0)!=(32); //@line 204 "snd_fx.cpp"
          $147 = $not$&1; //@line 204 "snd_fx.cpp"
          $$10 = $145 << $147; //@line 204 "snd_fx.cpp"
          $148 = $$10 << 1; //@line 205 "snd_fx.cpp"
          $149 = (($148) + ($nGlbVol$175))|0; //@line 205 "snd_fx.cpp"
          $nGlbVol$2 = $149;
         }
        } while(0);
        $164 = ($nGlbVol$2|0)<(0); //@line 225 "snd_fx.cpp"
        $$nGlbVol$2 = $164 ? 0 : $nGlbVol$2; //@line 225 "snd_fx.cpp"
        $165 = ($$nGlbVol$2|0)>(256); //@line 226 "snd_fx.cpp"
        $$$nGlbVol$2 = $165 ? 256 : $$nGlbVol$2; //@line 226 "snd_fx.cpp"
        $dwElapsedTime$3 = $dwElapsedTime$2;$nGlbVol$4 = $$$nGlbVol$2;$nMusicSpeed$3 = $nMusicSpeed$2;$nMusicTempo$3 = $nMusicTempo$2;$nNextPattern$5 = $nNextPattern$371;$nNextRow$4 = $nNextRow$274;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$1$param$2;$nSpeedCount$2 = $nSpeedCount$1;
        break L28;
        break;
       }
       case 21:  {
        $166 = ($param$2>>>0)<(65); //@line 229 "snd_fx.cpp"
        if (!($166)) {
         $dwElapsedTime$3 = $dwElapsedTime$2;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$2;$nMusicTempo$3 = $nMusicTempo$2;$nNextPattern$5 = $nNextPattern$371;$nNextRow$4 = $nNextRow$274;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$1;
         break L28;
        }
        $167 = $param$2&255; //@line 229 "snd_fx.cpp"
        $168 = (($chnvols) + ($nChn$081)|0); //@line 229 "snd_fx.cpp"
        HEAP8[$168>>0] = $167; //@line 229 "snd_fx.cpp"
        $dwElapsedTime$3 = $dwElapsedTime$2;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$2;$nMusicTempo$3 = $nMusicTempo$2;$nNextPattern$5 = $nNextPattern$371;$nNextRow$4 = $nNextRow$274;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$1;
        break L28;
        break;
       }
       case 22:  {
        $169 = ($param$2|0)==(0); //@line 232 "snd_fx.cpp"
        if ($169) {
         $172 = (($oldparam) + ($nChn$081)|0); //@line 232 "snd_fx.cpp"
         $173 = HEAP8[$172>>0]|0; //@line 232 "snd_fx.cpp"
         $174 = $173&255; //@line 232 "snd_fx.cpp"
         $param$10 = $174;
        } else {
         $170 = $param$2&255; //@line 232 "snd_fx.cpp"
         $171 = (($oldparam) + ($nChn$081)|0); //@line 232 "snd_fx.cpp"
         HEAP8[$171>>0] = $170; //@line 232 "snd_fx.cpp"
         $param$10 = $param$2;
        }
        $175 = $param$10&255; //@line 233 "snd_fx.cpp"
        $176 = ((($pChn$078)) + 223|0); //@line 233 "snd_fx.cpp"
        HEAP8[$176>>0] = $175; //@line 233 "snd_fx.cpp"
        $177 = $param$10 & 15; //@line 234 "snd_fx.cpp"
        $178 = ($177|0)!=(15); //@line 234 "snd_fx.cpp"
        $179 = $param$10 & 240;
        $180 = ($179|0)==(0); //@line 234 "snd_fx.cpp"
        $or$cond16 = $178 | $180; //@line 234 "snd_fx.cpp"
        do {
         if ($or$cond16) {
          $186 = ($179|0)!=(240); //@line 238 "snd_fx.cpp"
          $187 = ($177|0)==(0); //@line 238 "snd_fx.cpp"
          $or$cond17 = $186 | $187; //@line 238 "snd_fx.cpp"
          $188 = (($chnvols) + ($nChn$081)|0); //@line 247 "snd_fx.cpp"
          $189 = HEAP8[$188>>0]|0; //@line 247 "snd_fx.cpp"
          if (!($or$cond17)) {
           $190 = $189&255; //@line 240 "snd_fx.cpp"
           $191 = ($190>>>0)>($177>>>0); //@line 240 "snd_fx.cpp"
           $192 = (($190) - ($177))|0; //@line 240 "snd_fx.cpp"
           $$19 = $191 ? $192 : 0; //@line 240 "snd_fx.cpp"
           $$pre$phiZ2D = $188;$param$11 = $$19;
           break;
          }
          if ($187) {
           $197 = $179 >>> 4; //@line 247 "snd_fx.cpp"
           $198 = Math_imul($197, $nMusicSpeed$2)|0; //@line 247 "snd_fx.cpp"
           $199 = $189&255; //@line 247 "snd_fx.cpp"
           $200 = (($199) + ($198))|0; //@line 247 "snd_fx.cpp"
           $$pre$phiZ2D = $188;$param$11 = $200;
           break;
          } else {
           $193 = Math_imul($177, $nMusicSpeed$2)|0; //@line 245 "snd_fx.cpp"
           $194 = $189&255; //@line 246 "snd_fx.cpp"
           $195 = ($194>>>0)>($193>>>0); //@line 246 "snd_fx.cpp"
           $196 = (($194) - ($193))|0; //@line 246 "snd_fx.cpp"
           $$18 = $195 ? $196 : 0; //@line 246 "snd_fx.cpp"
           $$pre$phiZ2D = $188;$param$11 = $$18;
           break;
          }
         } else {
          $181 = $param$10 >>> 4; //@line 236 "snd_fx.cpp"
          $182 = (($chnvols) + ($nChn$081)|0); //@line 236 "snd_fx.cpp"
          $183 = HEAP8[$182>>0]|0; //@line 236 "snd_fx.cpp"
          $184 = $183&255; //@line 236 "snd_fx.cpp"
          $185 = (($184) + ($181))|0; //@line 236 "snd_fx.cpp"
          $$pre$phiZ2D = $182;$param$11 = $185;
         }
        } while(0);
        $201 = ($param$11>>>0)>(64); //@line 248 "snd_fx.cpp"
        $202 = $param$11&255; //@line 249 "snd_fx.cpp"
        $203 = $201 ? 64 : $202; //@line 249 "snd_fx.cpp"
        HEAP8[$$pre$phiZ2D>>0] = $203; //@line 249 "snd_fx.cpp"
        $dwElapsedTime$3 = $dwElapsedTime$2;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$2;$nMusicTempo$3 = $nMusicTempo$2;$nNextPattern$5 = $nNextPattern$371;$nNextRow$4 = $nNextRow$274;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$1;
        break L28;
        break;
       }
       default: {
        $dwElapsedTime$3 = $dwElapsedTime$2;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$2;$nMusicTempo$3 = $nMusicTempo$2;$nNextPattern$5 = $nNextPattern$371;$nNextRow$4 = $nNextRow$274;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$1;
        break L28;
       }
       }
      } else {
       $dwElapsedTime$3 = $dwElapsedTime$2;$nGlbVol$4 = $nGlbVol$175;$nMusicSpeed$3 = $nMusicSpeed$2;$nMusicTempo$3 = $nMusicTempo$2;$nNextPattern$5 = $nNextPattern$371;$nNextRow$4 = $nNextRow$274;$nOldGlbVolSlide$3 = $nOldGlbVolSlide$176;$nSpeedCount$2 = $nSpeedCount$1;
      }
     }
    } while(0);
    $204 = ((($p$080)) + 6|0); //@line 96 "snd_fx.cpp"
    $205 = ((($pChn$078)) + 256|0); //@line 96 "snd_fx.cpp"
    $206 = (($nChn$081) + 1)|0; //@line 96 "snd_fx.cpp"
    $207 = ($206>>>0)<($7>>>0); //@line 96 "snd_fx.cpp"
    if ($207) {
     $dwElapsedTime$184 = $dwElapsedTime$3;$nChn$081 = $206;$nGlbVol$175 = $nGlbVol$4;$nMusicSpeed$172 = $nMusicSpeed$3;$nMusicTempo$173 = $nMusicTempo$3;$nNextPattern$371 = $nNextPattern$5;$nNextRow$274 = $nNextRow$4;$nOldGlbVolSlide$176 = $nOldGlbVolSlide$3;$nSpeedCount$077 = $nSpeedCount$2;$p$080 = $204;$pChn$078 = $205;
    } else {
     $238 = $7;$dwElapsedTime$1$lcssa = $dwElapsedTime$3;$nGlbVol$1$lcssa = $nGlbVol$4;$nMusicSpeed$1$lcssa = $nMusicSpeed$3;$nMusicTempo$1$lcssa = $nMusicTempo$3;$nNextPattern$3$lcssa = $nNextPattern$5;$nNextRow$2$lcssa = $nNextRow$4;$nOldGlbVolSlide$1$lcssa = $nOldGlbVolSlide$3;$nSpeedCount$0$lcssa = $nSpeedCount$2;
     break;
    }
   }
  }
  $208 = (($nSpeedCount$0$lcssa) + ($nMusicSpeed$1$lcssa))|0; //@line 253 "snd_fx.cpp"
  $209 = ($208*2500)|0; //@line 254 "snd_fx.cpp"
  $210 = (($209>>>0) / ($nMusicTempo$1$lcssa>>>0))&-1; //@line 254 "snd_fx.cpp"
  $211 = (($210) + ($dwElapsedTime$1$lcssa))|0; //@line 254 "snd_fx.cpp"
  $46 = $238;$dwElapsedTime$0 = $211;$nGlbVol$0 = $nGlbVol$1$lcssa;$nMusicSpeed$0 = $nMusicSpeed$1$lcssa;$nMusicTempo$0 = $nMusicTempo$1$lcssa;$nNextPattern$0 = $nNextPattern$3$lcssa;$nNextRow$0 = $nNextRow$2$lcssa;$nOldGlbVolSlide$0 = $nOldGlbVolSlide$1$lcssa;
 }
 do {
  if ((label|0) == 15) {
   if ($bAdjust) {
    $53 = ((($this)) + 59708|0); //@line 88 "snd_fx.cpp"
    HEAP32[$53>>2] = $nMusicSpeed$0$lcssa197; //@line 88 "snd_fx.cpp"
    $54 = ((($this)) + 59712|0); //@line 89 "snd_fx.cpp"
    HEAP32[$54>>2] = $nMusicTempo$0$lcssa200; //@line 89 "snd_fx.cpp"
    $dwElapsedTime$4 = $dwElapsedTime$0$lcssa209;$nGlbVol$5 = $nGlbVol$0$lcssa203;$nOldGlbVolSlide$4 = $nOldGlbVolSlide$0$lcssa206;
    break;
   } else {
    $dwElapsedTime$434 = $dwElapsedTime$0$lcssa209;
    $236 = (($dwElapsedTime$434) + 500)|0; //@line 273 "snd_fx.cpp"
    $237 = (($236>>>0) / 1000)&-1; //@line 273 "snd_fx.cpp"
    STACKTOP = sp;return ($237|0); //@line 274 "snd_fx.cpp"
   }
  }
 } while(0);
 $bAdjust$not = $bAdjust ^ 1; //@line 257 "snd_fx.cpp"
 $brmerge = $bAdjust$not | $bTotal; //@line 257 "snd_fx.cpp"
 if ($brmerge) {
  $dwElapsedTime$434 = $dwElapsedTime$4;
  $236 = (($dwElapsedTime$434) + 500)|0; //@line 273 "snd_fx.cpp"
  $237 = (($236>>>0) / 1000)&-1; //@line 273 "snd_fx.cpp"
  STACKTOP = sp;return ($237|0); //@line 274 "snd_fx.cpp"
 }
 $212 = ((($this)) + 59744|0); //@line 259 "snd_fx.cpp"
 HEAP32[$212>>2] = $nGlbVol$5; //@line 259 "snd_fx.cpp"
 $213 = ((($this)) + 59760|0); //@line 260 "snd_fx.cpp"
 HEAP32[$213>>2] = $nOldGlbVolSlide$4; //@line 260 "snd_fx.cpp"
 if ($8) {
  $dwElapsedTime$434 = $dwElapsedTime$4;
  $236 = (($dwElapsedTime$434) + 500)|0; //@line 273 "snd_fx.cpp"
  $237 = (($236>>>0) / 1000)&-1; //@line 273 "snd_fx.cpp"
  STACKTOP = sp;return ($237|0); //@line 274 "snd_fx.cpp"
 } else {
  $n$063 = 0;
 }
 while(1) {
  $214 = (($chnvols) + ($n$063)|0); //@line 263 "snd_fx.cpp"
  $215 = HEAP8[$214>>0]|0; //@line 263 "snd_fx.cpp"
  $216 = $215&255; //@line 263 "snd_fx.cpp"
  $217 = (((($this) + ($n$063<<8)|0)) + 168|0); //@line 263 "snd_fx.cpp"
  HEAP32[$217>>2] = $216; //@line 263 "snd_fx.cpp"
  $218 = (($notes) + ($n$063)|0); //@line 264 "snd_fx.cpp"
  $219 = HEAP8[$218>>0]|0; //@line 264 "snd_fx.cpp"
  $220 = ($219<<24>>24)==(0); //@line 264 "snd_fx.cpp"
  if (!($220)) {
   $221 = (((($this) + ($n$063<<8)|0)) + 214|0); //@line 264 "snd_fx.cpp"
   HEAP8[$221>>0] = $219; //@line 264 "snd_fx.cpp"
  }
  $222 = (($instr) + ($n$063)|0); //@line 265 "snd_fx.cpp"
  $223 = HEAP8[$222>>0]|0; //@line 265 "snd_fx.cpp"
  $224 = ($223<<24>>24)==(0); //@line 265 "snd_fx.cpp"
  if (!($224)) {
   $225 = (((($this) + ($n$063<<8)|0)) + 215|0); //@line 265 "snd_fx.cpp"
   HEAP8[$225>>0] = $223; //@line 265 "snd_fx.cpp"
  }
  $226 = (($vols) + ($n$063)|0); //@line 266 "snd_fx.cpp"
  $227 = HEAP8[$226>>0]|0; //@line 266 "snd_fx.cpp"
  $228 = ($227<<24>>24)==(-1); //@line 266 "snd_fx.cpp"
  if (!($228)) {
   $229 = ($227&255)>(64); //@line 268 "snd_fx.cpp"
   if ($229) {
    HEAP8[$226>>0] = 64; //@line 268 "snd_fx.cpp"
    $231 = 64;
   } else {
    $231 = $227;
   }
   $230 = $231&255; //@line 269 "snd_fx.cpp"
   $232 = $230 << 2; //@line 269 "snd_fx.cpp"
   $233 = (((($this) + ($n$063<<8)|0)) + 116|0); //@line 269 "snd_fx.cpp"
   HEAP32[$233>>2] = $232; //@line 269 "snd_fx.cpp"
  }
  $234 = (($n$063) + 1)|0; //@line 261 "snd_fx.cpp"
  $235 = ($234>>>0)<($7>>>0); //@line 261 "snd_fx.cpp"
  if ($235) {
   $n$063 = $234;
  } else {
   $dwElapsedTime$434 = $dwElapsedTime$4;
   break;
  }
 }
 $236 = (($dwElapsedTime$434) + 500)|0; //@line 273 "snd_fx.cpp"
 $237 = (($236>>>0) / 1000)&-1; //@line 273 "snd_fx.cpp"
 STACKTOP = sp;return ($237|0); //@line 274 "snd_fx.cpp"
}
function __ZN10CSoundFile19SetReverbParametersEjj($nDepth,$nDelay) {
 $nDepth = $nDepth|0;
 $nDelay = $nDelay|0;
 var $$$01 = 0, $$01 = 0, $$nDepth = 0, $$op = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($nDepth>>>0)>(100); //@line 430 "snd_dsp.cpp"
 $$nDepth = $0 ? 100 : $nDepth; //@line 430 "snd_dsp.cpp"
 $1 = (($$nDepth>>>0) / 20)&-1; //@line 431 "snd_dsp.cpp"
 $2 = ($$nDepth>>>0)>(99); //@line 432 "snd_dsp.cpp"
 $$op = (4 - ($1))|0; //@line 433 "snd_dsp.cpp"
 $3 = $2 ? 0 : $$op; //@line 433 "snd_dsp.cpp"
 HEAP32[10115] = $3; //@line 433 "snd_dsp.cpp"
 $4 = ($nDelay>>>0)<(40); //@line 434 "snd_dsp.cpp"
 $$01 = $4 ? 40 : $nDelay; //@line 434 "snd_dsp.cpp"
 $5 = ($$01>>>0)>(250); //@line 435 "snd_dsp.cpp"
 $$$01 = $5 ? 250 : $$01; //@line 435 "snd_dsp.cpp"
 HEAP32[10116] = $$$01; //@line 436 "snd_dsp.cpp"
 return 1; //@line 437 "snd_dsp.cpp"
}
function __ZN10CSoundFile18SetXBassParametersEjj($nDepth,$nRange) {
 $nDepth = $nDepth|0;
 $nRange = $nRange|0;
 var $$nDepth = 0, $$op = 0, $$op1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($nDepth>>>0)>(100); //@line 445 "snd_dsp.cpp"
 $$nDepth = $0 ? 100 : $nDepth; //@line 445 "snd_dsp.cpp"
 $1 = (($$nDepth>>>0) / 20)&-1; //@line 446 "snd_dsp.cpp"
 $2 = ($$nDepth>>>0)>(99); //@line 447 "snd_dsp.cpp"
 $$op = (8 - ($1))|0; //@line 448 "snd_dsp.cpp"
 $3 = $2 ? 4 : $$op; //@line 448 "snd_dsp.cpp"
 HEAP32[10113] = $3; //@line 448 "snd_dsp.cpp"
 $4 = ($nRange>>>0)>(29); //@line 450 "snd_dsp.cpp"
 $5 = (($nRange>>>0) / 5)&-1; //@line 449 "snd_dsp.cpp"
 $$op1 = (26 - ($5))|0; //@line 452 "snd_dsp.cpp"
 $6 = $4 ? $$op1 : 21; //@line 452 "snd_dsp.cpp"
 HEAP32[10114] = $6; //@line 452 "snd_dsp.cpp"
 return 1; //@line 453 "snd_dsp.cpp"
}
function __ZN10CSoundFile21SetSurroundParametersEjj($nDepth,$nDelay) {
 $nDepth = $nDepth|0;
 $nDelay = $nDelay|0;
 var $$ = 0, $$$ = 0, $$1 = 0, $$nDelay = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $nDepth << 4; //@line 461 "snd_dsp.cpp"
 $1 = (($0>>>0) / 100)&-1; //@line 461 "snd_dsp.cpp"
 $2 = ($1>>>0)>(16); //@line 462 "snd_dsp.cpp"
 $$ = $2 ? 16 : $1; //@line 462 "snd_dsp.cpp"
 $3 = ($$|0)==(0); //@line 463 "snd_dsp.cpp"
 $$$ = $3 ? 1 : $$; //@line 463 "snd_dsp.cpp"
 HEAP32[10117] = $$$; //@line 464 "snd_dsp.cpp"
 $4 = ($nDelay>>>0)<(4); //@line 465 "snd_dsp.cpp"
 $$nDelay = $4 ? 4 : $nDelay; //@line 465 "snd_dsp.cpp"
 $5 = ($$nDelay>>>0)>(50); //@line 466 "snd_dsp.cpp"
 $$1 = $5 ? 50 : $$nDelay; //@line 466 "snd_dsp.cpp"
 HEAP32[10118] = $$1; //@line 467 "snd_dsp.cpp"
 return 1; //@line 468 "snd_dsp.cpp"
}
function __ZN10CSoundFile15SetWaveConfigExEbbbbbbb($bSurround,$bNoOverSampling,$bReverb,$hqido,$bMegaBass,$bNR,$bEQ) {
 $bSurround = $bSurround|0;
 $bNoOverSampling = $bNoOverSampling|0;
 $bReverb = $bReverb|0;
 $hqido = $hqido|0;
 $bMegaBass = $bMegaBass|0;
 $bNR = $bNR|0;
 $bEQ = $bEQ|0;
 var $$ = 0, $$d$1 = 0, $$d$3 = 0, $$d$5 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $d$1 = 0, $d$3 = 0, $d$5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[183080]|0; //@line 474 "snd_dsp.cpp"
 $1 = $0 & -507; //@line 474 "snd_dsp.cpp"
 $2 = $1 | 64; //@line 475 "snd_dsp.cpp"
 $$ = $bSurround ? $2 : $1; //@line 475 "snd_dsp.cpp"
 $3 = $$ | 8; //@line 476 "snd_dsp.cpp"
 $d$1 = $bNoOverSampling ? $3 : $$; //@line 476 "snd_dsp.cpp"
 $4 = $d$1 | 128; //@line 477 "snd_dsp.cpp"
 $$d$1 = $bReverb ? $4 : $d$1; //@line 477 "snd_dsp.cpp"
 $5 = $$d$1 | 16; //@line 478 "snd_dsp.cpp"
 $d$3 = $hqido ? $5 : $$d$1; //@line 478 "snd_dsp.cpp"
 $6 = $d$3 | 32; //@line 479 "snd_dsp.cpp"
 $$d$3 = $bMegaBass ? $6 : $d$3; //@line 479 "snd_dsp.cpp"
 $7 = $$d$3 | 2; //@line 480 "snd_dsp.cpp"
 $d$5 = $bNR ? $7 : $$d$3; //@line 480 "snd_dsp.cpp"
 $8 = $d$5 | 256; //@line 481 "snd_dsp.cpp"
 $$d$5 = $bEQ ? $8 : $d$5; //@line 481 "snd_dsp.cpp"
 HEAP32[183080] = $$d$5; //@line 482 "snd_dsp.cpp"
 (__ZN10CSoundFile10InitPlayerEb(0)|0); //@line 483 "snd_dsp.cpp"
 return 1; //@line 484 "snd_dsp.cpp"
}
function __ZN7ModPlug14UpdateSettingsEb($updateBasicConfig) {
 $updateBasicConfig = $updateBasicConfig|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[10119]|0; //@line 42 "modplug.cpp"
 $1 = $0 & 4; //@line 42 "modplug.cpp"
 $2 = ($1|0)==(0); //@line 42 "modplug.cpp"
 if ($2) {
  $6 = $0;
 } else {
  $3 = HEAP32[(40504)>>2]|0; //@line 44 "modplug.cpp"
  $4 = HEAP32[(40508)>>2]|0; //@line 45 "modplug.cpp"
  (__ZN10CSoundFile19SetReverbParametersEjj($3,$4)|0); //@line 44 "modplug.cpp"
  $$pre = HEAP32[10119]|0; //@line 48 "modplug.cpp"
  $6 = $$pre;
 }
 $5 = $6 & 8; //@line 48 "modplug.cpp"
 $7 = ($5|0)==(0); //@line 48 "modplug.cpp"
 if ($7) {
  (__ZN10CSoundFile18SetXBassParametersEjj(0,0)|0); //@line 54 "modplug.cpp"
 } else {
  $8 = HEAP32[(40512)>>2]|0; //@line 50 "modplug.cpp"
  $9 = HEAP32[(40516)>>2]|0; //@line 51 "modplug.cpp"
  (__ZN10CSoundFile18SetXBassParametersEjj($8,$9)|0); //@line 50 "modplug.cpp"
 }
 $10 = HEAP32[10119]|0; //@line 56 "modplug.cpp"
 $11 = $10 & 16; //@line 56 "modplug.cpp"
 $12 = ($11|0)==(0); //@line 56 "modplug.cpp"
 if (!($12)) {
  $13 = HEAP32[(40520)>>2]|0; //@line 58 "modplug.cpp"
  $14 = HEAP32[(40524)>>2]|0; //@line 59 "modplug.cpp"
  (__ZN10CSoundFile21SetSurroundParametersEjj($13,$14)|0); //@line 58 "modplug.cpp"
 }
 if ($updateBasicConfig) {
  $15 = HEAP32[(40488)>>2]|0; //@line 64 "modplug.cpp"
  $16 = HEAP32[(40484)>>2]|0; //@line 65 "modplug.cpp"
  $17 = HEAP32[(40480)>>2]|0; //@line 66 "modplug.cpp"
  (__ZN10CSoundFile13SetWaveConfigEjjjb($15,$16,$17,0)|0); //@line 64 "modplug.cpp"
  $18 = HEAP32[(40496)>>2]|0; //@line 67 "modplug.cpp"
  $19 = HEAP32[(40500)>>2]|0; //@line 68 "modplug.cpp"
  (__ZN10CSoundFile12SetMixConfigEjj($18,$19)|0); //@line 67 "modplug.cpp"
  $20 = HEAP32[(40484)>>2]|0; //@line 70 "modplug.cpp"
  $21 = (($20|0) / 8)&-1; //@line 70 "modplug.cpp"
  $22 = HEAP32[(40480)>>2]|0; //@line 70 "modplug.cpp"
  $23 = Math_imul($22, $21)|0; //@line 70 "modplug.cpp"
  HEAP32[187187] = $23; //@line 70 "modplug.cpp"
 }
 $24 = HEAP32[10119]|0; //@line 73 "modplug.cpp"
 $25 = $24 & 16; //@line 73 "modplug.cpp"
 $26 = ($25|0)!=(0); //@line 73 "modplug.cpp"
 $27 = $24 & 1; //@line 74 "modplug.cpp"
 $28 = ($27|0)==(0); //@line 74 "modplug.cpp"
 $29 = $24 & 4; //@line 75 "modplug.cpp"
 $30 = ($29|0)!=(0); //@line 75 "modplug.cpp"
 $31 = $24 & 8; //@line 77 "modplug.cpp"
 $32 = ($31|0)!=(0); //@line 77 "modplug.cpp"
 $33 = $24 & 2; //@line 78 "modplug.cpp"
 $34 = ($33|0)!=(0); //@line 78 "modplug.cpp"
 (__ZN10CSoundFile15SetWaveConfigExEbbbbbbb($26,$28,$30,1,$32,$34,0)|0); //@line 73 "modplug.cpp"
 $35 = HEAP32[(40492)>>2]|0; //@line 80 "modplug.cpp"
 (__ZN10CSoundFile17SetResamplingModeEj($35)|0); //@line 80 "modplug.cpp"
 return; //@line 81 "modplug.cpp"
}
function _ModPlug_Load($data,$size) {
 $data = $data|0;
 $size = $size|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj(67500)|0); //@line 86 "modplug.cpp"
 __ZN10CSoundFileC2Ev($0); //@line 11 "modplug.cpp"
 __ZN7ModPlug14UpdateSettingsEb(1); //@line 87 "modplug.cpp"
 $1 = (__ZN10CSoundFile6CreateEPKhj($0,$data,$size)|0); //@line 88 "modplug.cpp"
 if ($1) {
  $2 = HEAP32[(40528)>>2]|0; //@line 90 "modplug.cpp"
  $3 = ((($0)) + 59772|0); //@line 613 "../src/libmodplug/sndfile.h"
  HEAP32[$3>>2] = $2; //@line 613 "../src/libmodplug/sndfile.h"
  $4 = ((($0)) + 59776|0); //@line 613 "../src/libmodplug/sndfile.h"
  HEAP32[$4>>2] = $2; //@line 613 "../src/libmodplug/sndfile.h"
  $$0 = $0;
  return ($$0|0); //@line 98 "modplug.cpp"
 } else {
  __ZN10CSoundFileD2Ev($0); //@line 11 "modplug.cpp"
  __ZdlPv($0); //@line 95 "modplug.cpp"
  $$0 = 0;
  return ($$0|0); //@line 98 "modplug.cpp"
 }
 return (0)|0;
}
function _ModPlug_Unload($file) {
 $file = $file|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 (__ZN10CSoundFile7DestroyEv($file)|0); //@line 102 "modplug.cpp"
 $0 = ($file|0)==(0|0); //@line 103 "modplug.cpp"
 if ($0) {
  return; //@line 104 "modplug.cpp"
 }
 __ZN10CSoundFileD2Ev($file); //@line 11 "modplug.cpp"
 __ZdlPv($file); //@line 103 "modplug.cpp"
 return; //@line 104 "modplug.cpp"
}
function _ModPlug_Read($file,$buffer,$size) {
 $file = $file|0;
 $buffer = $buffer|0;
 $size = $size|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN10CSoundFile4ReadEPvj($file,$buffer,$size)|0); //@line 108 "modplug.cpp"
 $1 = HEAP32[187187]|0; //@line 108 "modplug.cpp"
 $2 = Math_imul($1, $0)|0; //@line 108 "modplug.cpp"
 return ($2|0); //@line 108 "modplug.cpp"
}
function _ModPlug_GetLength($file) {
 $file = $file|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN10CSoundFile9GetLengthEbb($file,0,1)|0); //@line 612 "../src/libmodplug/sndfile.h"
 $1 = ($0*1000)|0; //@line 118 "modplug.cpp"
 return ($1|0); //@line 118 "modplug.cpp"
}
function _ModPlug_SetMasterVolume($file,$cvol) {
 $file = $file|0;
 $cvol = $cvol|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 (__ZN10CSoundFile15SetMasterVolumeEjb($file,$cvol,0)|0); //@line 140 "modplug.cpp"
 return; //@line 142 "modplug.cpp"
}
function _ModPlug_Seek($file,$millisecond) {
 $file = $file|0;
 $millisecond = $millisecond|0;
 var $$millisecond = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN10CSoundFile9GetLengthEbb($file,0,1)|0); //@line 612 "../src/libmodplug/sndfile.h"
 $1 = ($0*1000)|0; //@line 255 "modplug.cpp"
 $2 = ($1|0)<($millisecond|0); //@line 258 "modplug.cpp"
 $$millisecond = $2 ? $1 : $millisecond; //@line 258 "modplug.cpp"
 $3 = (__ZNK10CSoundFile14GetMaxPositionEv($file)|0); //@line 260 "modplug.cpp"
 $4 = (+($3|0)); //@line 261 "modplug.cpp"
 $5 = (+($1|0)); //@line 261 "modplug.cpp"
 $6 = $4 / $5; //@line 261 "modplug.cpp"
 $7 = (+($$millisecond|0)); //@line 263 "modplug.cpp"
 $8 = $7 * $6; //@line 263 "modplug.cpp"
 $9 = (~~(($8))); //@line 263 "modplug.cpp"
 __ZN10CSoundFile13SetCurrentPosEj($file,$9); //@line 263 "modplug.cpp"
 return; //@line 264 "modplug.cpp"
}
function _ModPlug_SetSettings($settings) {
 $settings = $settings|0;
 var dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 dest=40476; src=$settings; stop=dest+56|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0)); //@line 273 "modplug.cpp"
 __ZN7ModPlug14UpdateSettingsEb(0); //@line 274 "modplug.cpp"
 return; //@line 275 "modplug.cpp"
}
function _th_decode_headerin($_info,$_tc,$_setup,$_op) {
 $_info = $_info|0;
 $_tc = $_tc|0;
 $_setup = $_setup|0;
 $_op = $_op|0;
 var $$0 = 0, $$0$i = 0, $$0$i35$i = 0, $$0$i46$i = 0, $$012$i$i$i = 0, $$012$i3$i$i = 0, $$03$i$i$i = 0, $$03$i2$i$i = 0, $$phi$trans$insert$i$i = 0, $$pre$i$i = 0, $$pre11$i$i = 0, $$pre12$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0;
 var $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0;
 var $213 = 0, $214 = 0, $215 = 0, $216 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $buffer$i = 0, $i$08$i$i = 0, $i$08$i$i$lcssa = 0, $i$08$i$i$lcssa12 = 0, $opb = 0, $or$cond$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $buffer$i = sp + 20|0;
 $opb = sp;
 $0 = ($_op|0)==(0|0);
 if ($0) {
  $$0 = -20;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ($_info|0)==(0|0);
 if ($1) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = HEAP32[$_op>>2]|0;
 $3 = ((($_op)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 _oc_pack_readinit($opb,$2,$4);
 $5 = (_oc_pack_read($opb,8)|0);
 $6 = $5 & 128;
 $7 = ($6|0)==(0);
 if ($7) {
  $8 = ((($_info)) + 4|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)==(0);
  if ($10) {
   label = 7;
  } else {
   $11 = ((($_tc)) + 12|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($12|0)==(0|0);
   if ($13) {
    label = 7;
   } else {
    $14 = HEAP32[$_setup>>2]|0;
    $15 = ($14|0)==(0|0);
    if ($15) {
     label = 7;
    } else {
     $$0$i = 0;
    }
   }
  }
 } else {
  label = 7;
 }
 L11: do {
  if ((label|0) == 7) {
   $16 = (_oc_pack_read($opb,8)|0);
   $17 = $16&255;
   $18 = ((($buffer$i)) + 1|0);
   HEAP8[$buffer$i>>0] = $17;
   $19 = (_oc_pack_read($opb,8)|0);
   $20 = $19&255;
   $21 = ((($buffer$i)) + 2|0);
   HEAP8[$18>>0] = $20;
   $22 = (_oc_pack_read($opb,8)|0);
   $23 = $22&255;
   $24 = ((($buffer$i)) + 3|0);
   HEAP8[$21>>0] = $23;
   $25 = (_oc_pack_read($opb,8)|0);
   $26 = $25&255;
   $27 = ((($buffer$i)) + 4|0);
   HEAP8[$24>>0] = $26;
   $28 = (_oc_pack_read($opb,8)|0);
   $29 = $28&255;
   $30 = ((($buffer$i)) + 5|0);
   HEAP8[$27>>0] = $29;
   $31 = (_oc_pack_read($opb,8)|0);
   $32 = $31&255;
   HEAP8[$30>>0] = $32;
   $33 = (_memcmp($buffer$i,529923,6)|0);
   $34 = ($33|0)==(0);
   if ($34) {
    switch ($5|0) {
    case 128:  {
     $35 = ((($_op)) + 8|0);
     $36 = HEAP32[$35>>2]|0;
     $37 = ($36|0)==(0);
     if ($37) {
      $$0$i = -20;
      break L11;
     }
     $38 = ((($_info)) + 4|0);
     $39 = HEAP32[$38>>2]|0;
     $40 = ($39|0)==(0);
     if (!($40)) {
      $$0$i = -20;
      break L11;
     }
     $41 = (_oc_pack_read($opb,8)|0);
     $42 = $41&255;
     HEAP8[$_info>>0] = $42;
     $43 = (_oc_pack_read($opb,8)|0);
     $44 = $43&255;
     $45 = ((($_info)) + 1|0);
     HEAP8[$45>>0] = $44;
     $46 = (_oc_pack_read($opb,8)|0);
     $47 = $46&255;
     $48 = ((($_info)) + 2|0);
     HEAP8[$48>>0] = $47;
     $49 = HEAP8[$_info>>0]|0;
     $50 = ($49&255)>(3);
     do {
      if ($50) {
       $216 = -22;
      } else {
       $51 = ($49<<24>>24)==(3);
       if ($51) {
        $52 = HEAP8[$45>>0]|0;
        $53 = ($52&255)>(2);
        if ($53) {
         $216 = -22;
         break;
        }
       }
       $54 = (_oc_pack_read($opb,16)|0);
       $55 = $54 << 4;
       HEAP32[$38>>2] = $55;
       $56 = (_oc_pack_read($opb,16)|0);
       $57 = $56 << 4;
       $58 = ((($_info)) + 8|0);
       HEAP32[$58>>2] = $57;
       $59 = (_oc_pack_read($opb,24)|0);
       $60 = ((($_info)) + 12|0);
       HEAP32[$60>>2] = $59;
       $61 = (_oc_pack_read($opb,24)|0);
       $62 = ((($_info)) + 16|0);
       HEAP32[$62>>2] = $61;
       $63 = (_oc_pack_read($opb,8)|0);
       $64 = ((($_info)) + 20|0);
       HEAP32[$64>>2] = $63;
       $65 = (_oc_pack_read($opb,8)|0);
       $66 = ((($_info)) + 24|0);
       HEAP32[$66>>2] = $65;
       $67 = (_oc_pack_read($opb,32)|0);
       $68 = ((($_info)) + 28|0);
       HEAP32[$68>>2] = $67;
       $69 = (_oc_pack_read($opb,32)|0);
       $70 = ((($_info)) + 32|0);
       HEAP32[$70>>2] = $69;
       $71 = HEAP32[$38>>2]|0;
       $72 = ($71|0)==(0);
       if ($72) {
        $216 = -20;
       } else {
        $73 = HEAP32[$58>>2]|0;
        $74 = ($73|0)==(0);
        if ($74) {
         $216 = -20;
        } else {
         $75 = HEAP32[$60>>2]|0;
         $76 = HEAP32[$64>>2]|0;
         $77 = (($76) + ($75))|0;
         $78 = ($77>>>0)>($71>>>0);
         if ($78) {
          $216 = -20;
         } else {
          $79 = HEAP32[$62>>2]|0;
          $80 = HEAP32[$66>>2]|0;
          $81 = (($80) + ($79))|0;
          $82 = ($81>>>0)>($73>>>0);
          if ($82) {
           $216 = -20;
          } else {
           $83 = HEAP32[$68>>2]|0;
           $84 = ($83|0)==(0);
           $85 = ($69|0)==(0);
           $or$cond$i$i = $85 | $84;
           if ($or$cond$i$i) {
            $216 = -20;
           } else {
            $86 = (($73) - ($79))|0;
            $87 = (($86) - ($80))|0;
            HEAP32[$66>>2] = $87;
            $88 = (_oc_pack_read($opb,24)|0);
            $89 = ((($_info)) + 36|0);
            HEAP32[$89>>2] = $88;
            $90 = (_oc_pack_read($opb,24)|0);
            $91 = ((($_info)) + 40|0);
            HEAP32[$91>>2] = $90;
            $92 = (_oc_pack_read($opb,8)|0);
            $93 = ((($_info)) + 44|0);
            HEAP32[$93>>2] = $92;
            $94 = (_oc_pack_read($opb,24)|0);
            $95 = ((($_info)) + 52|0);
            HEAP32[$95>>2] = $94;
            $96 = (_oc_pack_read($opb,6)|0);
            $97 = ((($_info)) + 56|0);
            HEAP32[$97>>2] = $96;
            $98 = (_oc_pack_read($opb,5)|0);
            $99 = ((($_info)) + 60|0);
            HEAP32[$99>>2] = $98;
            $100 = (_oc_pack_read($opb,2)|0);
            $101 = ((($_info)) + 48|0);
            HEAP32[$101>>2] = $100;
            $102 = ($100|0)==(1);
            if ($102) {
             $216 = -20;
            } else {
             $103 = (_oc_pack_read($opb,3)|0);
             $104 = ($103|0)==(0);
             if ($104) {
              $105 = (_oc_pack_bytes_left($opb)|0);
              $106 = $105 >> 31;
              $107 = $106 & -20;
              $108 = ($107|0)<(0);
              if ($108) {
               $216 = $107;
              } else {
               $$0$i = 3;
               break L11;
              }
             } else {
              $216 = -20;
             }
            }
           }
          }
         }
        }
       }
      }
     } while(0);
     _th_info_clear($_info);
     $$0$i = $216;
     break L11;
     break;
    }
    case 129:  {
     $109 = ($_tc|0)==(0|0);
     if ($109) {
      $$0$i = -1;
      break L11;
     }
     $110 = ((($_info)) + 4|0);
     $111 = HEAP32[$110>>2]|0;
     $112 = ($111|0)==(0);
     if ($112) {
      $$0$i = -20;
      break L11;
     }
     $113 = ((($_tc)) + 12|0);
     $114 = HEAP32[$113>>2]|0;
     $115 = ($114|0)==(0|0);
     if (!($115)) {
      $$0$i = -20;
      break L11;
     }
     $116 = (_oc_pack_read($opb,8)|0);
     $117 = (_oc_pack_read($opb,8)|0);
     $118 = (_oc_pack_read($opb,8)|0);
     $119 = (_oc_pack_read($opb,8)|0);
     $120 = $117 << 8;
     $121 = $120 | $116;
     $122 = $118 << 16;
     $123 = $121 | $122;
     $124 = $119 << 24;
     $125 = $123 | $124;
     $126 = ($125|0)<(0);
     L35: do {
      if ($126) {
       $$0$i35$i = -20;
      } else {
       $127 = (_oc_pack_bytes_left($opb)|0);
       $128 = ($125|0)>($127|0);
       if ($128) {
        $$0$i35$i = -20;
       } else {
        $129 = (($125) + 1)|0;
        $130 = (_malloc($129)|0);
        HEAP32[$113>>2] = $130;
        $131 = ($130|0)==(0|0);
        if ($131) {
         $$0$i35$i = -1;
        } else {
         $132 = ($125|0)==(0);
         if ($132) {
          $139 = $130;
         } else {
          $$012$i$i$i = $125;$$03$i$i$i = $130;
          while(1) {
           $133 = (($$012$i$i$i) + -1)|0;
           $134 = (_oc_pack_read($opb,8)|0);
           $135 = $134&255;
           $136 = ((($$03$i$i$i)) + 1|0);
           HEAP8[$$03$i$i$i>>0] = $135;
           $137 = ($133|0)==(0);
           if ($137) {
            break;
           } else {
            $$012$i$i$i = $133;$$03$i$i$i = $136;
           }
          }
          $$pre$i$i = HEAP32[$113>>2]|0;
          $139 = $$pre$i$i;
         }
         $138 = (($139) + ($125)|0);
         HEAP8[$138>>0] = 0;
         $140 = (_oc_pack_read($opb,8)|0);
         $141 = (_oc_pack_read($opb,8)|0);
         $142 = (_oc_pack_read($opb,8)|0);
         $143 = (_oc_pack_read($opb,8)|0);
         $144 = $141 << 8;
         $145 = $144 | $140;
         $146 = $142 << 16;
         $147 = $145 | $146;
         $148 = $143 << 24;
         $149 = $147 | $148;
         $150 = ((($_tc)) + 8|0);
         HEAP32[$150>>2] = $149;
         $151 = ($149>>>0)>(536870911);
         if (!($151)) {
          $152 = $149 << 2;
          $153 = (_oc_pack_bytes_left($opb)|0);
          $154 = ($152|0)>($153|0);
          if (!($154)) {
           $155 = HEAP32[$150>>2]|0;
           $156 = $155 << 2;
           $157 = (_malloc($156)|0);
           $158 = ((($_tc)) + 4|0);
           HEAP32[$158>>2] = $157;
           $159 = (_malloc($156)|0);
           HEAP32[$_tc>>2] = $159;
           $160 = ($155|0)>(0);
           L47: do {
            if ($160) {
             $i$08$i$i = 0;
             while(1) {
              $161 = (_oc_pack_read($opb,8)|0);
              $162 = (_oc_pack_read($opb,8)|0);
              $163 = (_oc_pack_read($opb,8)|0);
              $164 = (_oc_pack_read($opb,8)|0);
              $165 = $162 << 8;
              $166 = $165 | $161;
              $167 = $163 << 16;
              $168 = $166 | $167;
              $169 = $164 << 24;
              $170 = $168 | $169;
              $171 = ($170|0)<(0);
              if ($171) {
               $i$08$i$i$lcssa = $i$08$i$i;
               label = 38;
               break;
              }
              $172 = (_oc_pack_bytes_left($opb)|0);
              $173 = ($170|0)>($172|0);
              if ($173) {
               $i$08$i$i$lcssa = $i$08$i$i;
               label = 38;
               break;
              }
              $174 = HEAP32[$158>>2]|0;
              $175 = (($174) + ($i$08$i$i<<2)|0);
              HEAP32[$175>>2] = $170;
              $176 = (($170) + 1)|0;
              $177 = (_malloc($176)|0);
              $178 = HEAP32[$_tc>>2]|0;
              $179 = (($178) + ($i$08$i$i<<2)|0);
              HEAP32[$179>>2] = $177;
              $180 = HEAP32[$_tc>>2]|0;
              $181 = (($180) + ($i$08$i$i<<2)|0);
              $182 = HEAP32[$181>>2]|0;
              $183 = ($182|0)==(0|0);
              if ($183) {
               $i$08$i$i$lcssa12 = $i$08$i$i;
               label = 40;
               break;
              }
              $184 = ($170|0)==(0);
              if ($184) {
               $191 = $182;
              } else {
               $$012$i3$i$i = $170;$$03$i2$i$i = $182;
               while(1) {
                $185 = (($$012$i3$i$i) + -1)|0;
                $186 = (_oc_pack_read($opb,8)|0);
                $187 = $186&255;
                $188 = ((($$03$i2$i$i)) + 1|0);
                HEAP8[$$03$i2$i$i>>0] = $187;
                $189 = ($185|0)==(0);
                if ($189) {
                 break;
                } else {
                 $$012$i3$i$i = $185;$$03$i2$i$i = $188;
                }
               }
               $$pre11$i$i = HEAP32[$_tc>>2]|0;
               $$phi$trans$insert$i$i = (($$pre11$i$i) + ($i$08$i$i<<2)|0);
               $$pre12$i$i = HEAP32[$$phi$trans$insert$i$i>>2]|0;
               $191 = $$pre12$i$i;
              }
              $190 = (($191) + ($170)|0);
              HEAP8[$190>>0] = 0;
              $192 = (($i$08$i$i) + 1)|0;
              $193 = HEAP32[$150>>2]|0;
              $194 = ($192|0)<($193|0);
              if ($194) {
               $i$08$i$i = $192;
              } else {
               break L47;
              }
             }
             if ((label|0) == 38) {
              HEAP32[$150>>2] = $i$08$i$i$lcssa;
              $$0$i35$i = -20;
              break L35;
             }
             else if ((label|0) == 40) {
              HEAP32[$150>>2] = $i$08$i$i$lcssa12;
              $$0$i35$i = -1;
              break L35;
             }
            }
           } while(0);
           $195 = (_oc_pack_bytes_left($opb)|0);
           $196 = $195 >> 31;
           $197 = $196 & -20;
           $198 = ($197|0)<(0);
           if ($198) {
            $$0$i35$i = $197;
            break;
           } else {
            $$0$i = 2;
            break L11;
           }
          }
         }
         HEAP32[$150>>2] = 0;
         $$0$i35$i = -20;
        }
       }
      }
     } while(0);
     _th_comment_clear($_tc);
     $$0$i = $$0$i35$i;
     break L11;
     break;
    }
    case 130:  {
     $199 = ($_tc|0)==(0|0);
     $200 = ($_setup|0)==(0|0);
     $or$cond$i = $199 | $200;
     if ($or$cond$i) {
      $$0$i = -1;
      break L11;
     }
     $201 = ((($_info)) + 4|0);
     $202 = HEAP32[$201>>2]|0;
     $203 = ($202|0)==(0);
     if ($203) {
      $$0$i = -20;
      break L11;
     }
     $204 = ((($_tc)) + 12|0);
     $205 = HEAP32[$204>>2]|0;
     $206 = ($205|0)==(0|0);
     if ($206) {
      $$0$i = -20;
      break L11;
     }
     $207 = HEAP32[$_setup>>2]|0;
     $208 = ($207|0)==(0|0);
     if (!($208)) {
      $$0$i = -20;
      break L11;
     }
     $209 = (_calloc(1,712)|0);
     $210 = ($209|0)==(0|0);
     if ($210) {
      $$0$i = -1;
      break L11;
     }
     $211 = ((($209)) + 320|0);
     $212 = (_oc_quant_params_unpack($opb,$211)|0);
     $213 = ($212|0)<(0);
     if ($213) {
      $$0$i46$i = $212;
     } else {
      $214 = (_oc_huff_trees_unpack($opb,$209)|0);
      $215 = ($214|0)<(0);
      if ($215) {
       $$0$i46$i = $214;
      } else {
       HEAP32[$_setup>>2] = $209;
       $$0$i = 1;
       break L11;
      }
     }
     _oc_quant_params_clear($211);
     _oc_huff_trees_clear($209);
     _free($209);
     $$0$i = $$0$i46$i;
     break L11;
     break;
    }
    default: {
     $$0$i = -20;
     break L11;
    }
    }
   } else {
    $$0$i = -21;
   }
  }
 } while(0);
 $$0 = $$0$i;
 STACKTOP = sp;return ($$0|0);
}
function _th_setup_free($_setup) {
 $_setup = $_setup|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($_setup|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($_setup)) + 320|0);
 _oc_quant_params_clear($1);
 _oc_huff_trees_clear($_setup);
 _free($_setup);
 return;
}
function _th_decode_alloc($_info,$_setup) {
 $_info = $_info|0;
 $_setup = $_setup|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i = 0, $exitcond18$i = 0, $or$cond = 0, $qi$012$i = 0, $qi$19$i = 0;
 var $tmp = 0, $tmp2 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ($_info|0)==(0|0);
 $1 = ($_setup|0)==(0|0);
 $or$cond = $0 | $1;
 if ($or$cond) {
  $$0 = 0;
  return ($$0|0);
 }
 $2 = (_malloc(54032)|0);
 $3 = ($2|0)==(0|0);
 do {
  if (!($3)) {
   $4 = (_oc_state_init($2,$_info,3)|0);
   $5 = ($4|0)<(0);
   if (!($5)) {
    $6 = ((($2)) + 51576|0);
    $7 = (_oc_huff_trees_copy($6,$_setup)|0);
    $8 = ($7|0)<(0);
    if ($8) {
     _oc_state_clear($2);
     break;
    }
    $9 = ((($2)) + 216|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = ($10*129)|0;
    $12 = (_malloc($11)|0);
    $13 = ((($2)) + 53432|0);
    HEAP32[$13>>2] = $12;
    $14 = ($12|0)==(0|0);
    if ($14) {
     _oc_huff_trees_clear($6);
     _oc_state_clear($2);
     break;
    } else {
     $qi$012$i = 0;
    }
    while(1) {
     $15 = (((($2)) + 2336|0) + (($qi$012$i*768)|0)|0);
     $16 = (((($2)) + 800|0) + (($qi$012$i*24)|0)|0);
     HEAP32[$16>>2] = $15;
     $17 = (((((($2)) + 2336|0) + (($qi$012$i*768)|0)|0)) + 128|0);
     $18 = (((((($2)) + 800|0) + (($qi$012$i*24)|0)|0)) + 4|0);
     HEAP32[$18>>2] = $17;
     $19 = (((((($2)) + 2336|0) + (($qi$012$i*768)|0)|0)) + 256|0);
     $20 = (((((($2)) + 800|0) + (($qi$012$i*24)|0)|0)) + 8|0);
     HEAP32[$20>>2] = $19;
     $21 = (((((($2)) + 2336|0) + (($qi$012$i*768)|0)|0)) + 384|0);
     $22 = (((((($2)) + 800|0) + (($qi$012$i*24)|0)|0)) + 12|0);
     HEAP32[$22>>2] = $21;
     $23 = (((((($2)) + 2336|0) + (($qi$012$i*768)|0)|0)) + 512|0);
     $24 = (((((($2)) + 800|0) + (($qi$012$i*24)|0)|0)) + 16|0);
     HEAP32[$24>>2] = $23;
     $25 = (((((($2)) + 2336|0) + (($qi$012$i*768)|0)|0)) + 640|0);
     $26 = (((((($2)) + 800|0) + (($qi$012$i*24)|0)|0)) + 20|0);
     HEAP32[$26>>2] = $25;
     $27 = (($qi$012$i) + 1)|0;
     $exitcond18$i = ($27|0)==(64);
     if ($exitcond18$i) {
      break;
     } else {
      $qi$012$i = $27;
     }
    }
    $28 = ((($2)) + 800|0);
    $29 = ((($2)) + 53448|0);
    $30 = ((($_setup)) + 320|0);
    _oc_dequant_tables_init($28,$29,$30);
    $qi$19$i = 0;
    while(1) {
     $31 = (((($2)) + 800|0) + ($qi$19$i<<2)|0);
     $32 = HEAP32[$31>>2]|0;
     $33 = ((($32)) + 24|0);
     $34 = HEAP16[$33>>1]|0;
     $35 = $34&65535;
     $36 = ((($32)) + 34|0);
     $37 = HEAP16[$36>>1]|0;
     $38 = $37&65535;
     $39 = ((($32)) + 36|0);
     $40 = HEAP16[$39>>1]|0;
     $41 = $40&65535;
     $42 = ((($32)) + 48|0);
     $43 = HEAP16[$42>>1]|0;
     $44 = $43&65535;
     $45 = (((($2)) + 808|0) + ($qi$19$i<<2)|0);
     $46 = HEAP32[$45>>2]|0;
     $47 = ((($46)) + 24|0);
     $48 = HEAP16[$47>>1]|0;
     $49 = $48&65535;
     $50 = ((($46)) + 34|0);
     $51 = HEAP16[$50>>1]|0;
     $52 = $51&65535;
     $53 = ((($46)) + 36|0);
     $54 = HEAP16[$53>>1]|0;
     $55 = $54&65535;
     $56 = ((($46)) + 48|0);
     $57 = HEAP16[$56>>1]|0;
     $58 = $57&65535;
     $59 = (((($2)) + 816|0) + ($qi$19$i<<2)|0);
     $60 = HEAP32[$59>>2]|0;
     $61 = ((($60)) + 24|0);
     $62 = HEAP16[$61>>1]|0;
     $63 = $62&65535;
     $64 = ((($60)) + 34|0);
     $65 = HEAP16[$64>>1]|0;
     $66 = $65&65535;
     $67 = ((($60)) + 36|0);
     $68 = HEAP16[$67>>1]|0;
     $69 = $68&65535;
     $70 = ((($60)) + 48|0);
     $71 = HEAP16[$70>>1]|0;
     $72 = $71&65535;
     $73 = (((($2)) + 824|0) + ($qi$19$i<<2)|0);
     $74 = HEAP32[$73>>2]|0;
     $75 = ((($74)) + 24|0);
     $76 = HEAP16[$75>>1]|0;
     $77 = $76&65535;
     $78 = ((($74)) + 34|0);
     $79 = HEAP16[$78>>1]|0;
     $80 = $79&65535;
     $81 = ((($74)) + 36|0);
     $82 = HEAP16[$81>>1]|0;
     $83 = $82&65535;
     $84 = ((($74)) + 48|0);
     $85 = HEAP16[$84>>1]|0;
     $86 = $85&65535;
     $87 = (((($2)) + 832|0) + ($qi$19$i<<2)|0);
     $88 = HEAP32[$87>>2]|0;
     $89 = ((($88)) + 24|0);
     $90 = HEAP16[$89>>1]|0;
     $91 = $90&65535;
     $92 = ((($88)) + 34|0);
     $93 = HEAP16[$92>>1]|0;
     $94 = $93&65535;
     $95 = ((($88)) + 36|0);
     $96 = HEAP16[$95>>1]|0;
     $97 = $96&65535;
     $98 = ((($88)) + 48|0);
     $99 = HEAP16[$98>>1]|0;
     $100 = $99&65535;
     $101 = (((($2)) + 840|0) + ($qi$19$i<<2)|0);
     $102 = HEAP32[$101>>2]|0;
     $103 = ((($102)) + 24|0);
     $104 = HEAP16[$103>>1]|0;
     $105 = $104&65535;
     $106 = ((($102)) + 34|0);
     $107 = HEAP16[$106>>1]|0;
     $108 = $107&65535;
     $109 = ((($102)) + 36|0);
     $110 = HEAP16[$109>>1]|0;
     $111 = $110&65535;
     $112 = ((($102)) + 48|0);
     $113 = HEAP16[$112>>1]|0;
     $114 = $113&65535;
     $115 = (($38) + ($35))|0;
     $116 = (($115) + ($41))|0;
     $117 = (($116) + ($44))|0;
     $118 = (($117) + ($77))|0;
     $119 = (($118) + ($80))|0;
     $120 = (($119) + ($83))|0;
     $tmp = (($120) + ($86))|0;
     $tmp2 = $tmp << 1;
     $121 = (($52) + ($49))|0;
     $122 = (($121) + ($55))|0;
     $123 = (($122) + ($58))|0;
     $124 = (($123) + ($63))|0;
     $125 = (($124) + ($66))|0;
     $126 = (($125) + ($69))|0;
     $127 = (($126) + ($72))|0;
     $128 = (($127) + ($91))|0;
     $129 = (($128) + ($94))|0;
     $130 = (($129) + ($97))|0;
     $131 = (($130) + ($100))|0;
     $132 = (($131) + ($105))|0;
     $133 = (($132) + ($108))|0;
     $134 = (($133) + ($111))|0;
     $135 = (($134) + ($114))|0;
     $136 = (($135) + ($tmp2))|0;
     $137 = $136 >> 11;
     $138 = (0 - ($137))|0;
     $139 = (((($2)) + 53704|0) + ($qi$19$i<<2)|0);
     HEAP32[$139>>2] = $138;
     $140 = (($qi$19$i) + 1)|0;
     $exitcond$i = ($140|0)==(64);
     if ($exitcond$i) {
      break;
     } else {
      $qi$19$i = $140;
     }
    }
    $141 = ((($2)) + 51488|0);
    $142 = ((($_setup)) + 576|0);
    dest=$141; src=$142; stop=dest+64|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
    $143 = ((($2)) + 53444|0);
    HEAP32[$143>>2] = 0;
    $144 = ((($2)) + 53960|0);
    HEAP32[$144>>2] = 0;
    $145 = ((($2)) + 53964|0);
    HEAP32[$145>>2] = 0;
    $146 = ((($2)) + 53968|0);
    HEAP32[$146>>2] = 0;
    $147 = ((($2)) + 54024|0);
    HEAP32[$147>>2] = 0;
    $148 = ((($2)) + 54028|0);
    HEAP32[$148>>2] = 0;
    $149 = ((($2)) + 776|0);
    $150 = $149;
    $151 = $150;
    HEAP32[$151>>2] = 0;
    $152 = (($150) + 4)|0;
    $153 = $152;
    HEAP32[$153>>2] = 0;
    $$0 = $2;
    return ($$0|0);
   }
  }
 } while(0);
 _free($2);
 $$0 = 0;
 return ($$0|0);
}
function _th_decode_free($_dec) {
 $_dec = $_dec|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($_dec|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($_dec)) + 53968|0);
 $2 = HEAP32[$1>>2]|0;
 _free($2);
 $3 = ((($_dec)) + 53964|0);
 $4 = HEAP32[$3>>2]|0;
 _free($4);
 $5 = ((($_dec)) + 53960|0);
 $6 = HEAP32[$5>>2]|0;
 _free($6);
 $7 = ((($_dec)) + 53432|0);
 $8 = HEAP32[$7>>2]|0;
 _free($8);
 $9 = ((($_dec)) + 51576|0);
 _oc_huff_trees_clear($9);
 _oc_state_clear($_dec);
 _free($_dec);
 return;
}
function _th_decode_ctl($_dec,$_req,$_buf,$_buf_sz) {
 $_dec = $_dec|0;
 $_req = $_req|0;
 $_buf = $_buf|0;
 $_buf_sz = $_buf_sz|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $or$cond = 0, $or$cond3 = 0, $or$cond7 = 0, $or$cond9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 switch ($_req|0) {
 case 1:  {
  $0 = ($_dec|0)==(0|0);
  $1 = ($_buf|0)==(0|0);
  $or$cond = $0 | $1;
  if ($or$cond) {
   $$0 = -1;
   return ($$0|0);
  }
  $2 = ($_buf_sz|0)==(4);
  if (!($2)) {
   $$0 = -10;
   return ($$0|0);
  }
  HEAP32[$_buf>>2] = 7;
  $$0 = 0;
  return ($$0|0);
  break;
 }
 case 3:  {
  $3 = ($_dec|0)==(0|0);
  $4 = ($_buf|0)==(0|0);
  $or$cond3 = $3 | $4;
  if ($or$cond3) {
   $$0 = -1;
   return ($$0|0);
  }
  $5 = ($_buf_sz|0)==(4);
  if (!($5)) {
   $$0 = -10;
   return ($$0|0);
  }
  $6 = HEAP32[$_buf>>2]|0;
  $7 = ($6>>>0)>(7);
  if ($7) {
   $$0 = -10;
   return ($$0|0);
  }
  $8 = ((($_dec)) + 53444|0);
  HEAP32[$8>>2] = $6;
  $$0 = 0;
  return ($$0|0);
  break;
 }
 case 5:  {
  $9 = ($_dec|0)==(0|0);
  $10 = ($_buf|0)==(0|0);
  $or$cond7 = $9 | $10;
  if ($or$cond7) {
   $$0 = -1;
   return ($$0|0);
  }
  $11 = ($_buf_sz|0)==(8);
  if (!($11)) {
   $$0 = -10;
   return ($$0|0);
  }
  $12 = $_buf;
  $13 = $12;
  $14 = HEAP32[$13>>2]|0;
  $15 = (($12) + 4)|0;
  $16 = $15;
  $17 = HEAP32[$16>>2]|0;
  $18 = ($17|0)<(0);
  if ($18) {
   $$0 = -10;
   return ($$0|0);
  }
  $19 = ((($_dec)) + 784|0);
  $20 = $19;
  $21 = $20;
  HEAP32[$21>>2] = $14;
  $22 = (($20) + 4)|0;
  $23 = $22;
  HEAP32[$23>>2] = $17;
  $24 = ((($_dec)) + 60|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = (_bitshift64Ashr(($14|0),($17|0),($25|0))|0);
  $27 = tempRet0;
  $28 = ((($_dec)) + 793|0);
  $29 = HEAP8[$28>>0]|0;
  $30 = $29&255;
  $31 = (_i64Subtract(($26|0),($27|0),($30|0),0)|0);
  $32 = tempRet0;
  $33 = ((($_dec)) + 768|0);
  $34 = $33;
  $35 = $34;
  HEAP32[$35>>2] = $31;
  $36 = (($34) + 4)|0;
  $37 = $36;
  HEAP32[$37>>2] = $32;
  $38 = 1 << $25;
  $39 = (($38) + -1)|0;
  $40 = ($39|0)<(0);
  $41 = $40 << 31 >> 31;
  $42 = $39 & $14;
  $43 = $41 & $17;
  $44 = (_i64Add(($42|0),($43|0),($31|0),($32|0))|0);
  $45 = tempRet0;
  $46 = ((($_dec)) + 776|0);
  $47 = $46;
  $48 = $47;
  HEAP32[$48>>2] = $44;
  $49 = (($47) + 4)|0;
  $50 = $49;
  HEAP32[$50>>2] = $45;
  $$0 = 0;
  return ($$0|0);
  break;
 }
 case 7:  {
  $51 = ($_dec|0)==(0|0);
  $52 = ($_buf|0)==(0|0);
  $or$cond9 = $51 | $52;
  if ($or$cond9) {
   $$0 = -1;
   return ($$0|0);
  }
  $53 = ($_buf_sz|0)==(8);
  if (!($53)) {
   $$0 = -10;
   return ($$0|0);
  }
  $54 = HEAP32[$_buf>>2]|0;
  $55 = ((($_dec)) + 54024|0);
  HEAP32[$55>>2] = $54;
  $56 = ((($_buf)) + 4|0);
  $57 = HEAP32[$56>>2]|0;
  $58 = ((($_dec)) + 54028|0);
  HEAP32[$58>>2] = $57;
  $$0 = 0;
  return ($$0|0);
  break;
 }
 default: {
  $$0 = -23;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function _th_decode_packetin($_dec,$_op,$_granpos) {
 $_dec = $_dec|0;
 $_op = $_op|0;
 $_granpos = $_granpos|0;
 var $$$i = 0, $$$i$i = 0, $$$i$i$lcssa = 0, $$$i50 = 0, $$$i6$i = 0, $$$us$i$i = 0, $$0 = 0, $$0$i$i = 0, $$0$lcssa$i$i = 0, $$0$lcssa$i$i$lcssa = 0, $$014$i$i = 0, $$014$us$i$i = 0, $$1 = 0, $$1$be$i$i = 0, $$1$be$i$i$lcssa = 0, $$1$be$us$i$i = 0, $$1$be$us$i$i$lcssa = 0, $$1$lcssa$i$i = 0, $$1$lcssa$us$i$i = 0, $$13$i$i = 0;
 var $$13$us$i$i = 0, $$eobs$0$i$i = 0, $$in$i$i = 0, $$lcssa = 0, $$lcssa249 = 0, $$lcssa251 = 0, $$lcssa252 = 0, $$lcssa253 = 0, $$lcssa255 = 0, $$lcssa262 = 0, $$lcssa263 = 0, $$lcssa265 = 0, $$lcssa266 = 0, $$lcssa266$lcssa = 0, $$lcssa268 = 0, $$lcssa269 = 0, $$lobit$i = 0, $$lobit$i$i = 0, $$lobit$i31 = 0, $$lobit$i70 = 0;
 var $$lobit1$i$i = 0, $$neg$i = 0, $$not$i = 0, $$not$i$i = 0, $$op2$i$i = 0, $$phi$trans$insert = 0, $$phi$trans$insert$i = 0, $$phi$trans$insert$i$i = 0, $$phi$trans$insert$i38 = 0, $$phi$trans$insert143 = 0, $$phi$trans$insert16$i = 0, $$phi$trans$insert23$i = 0, $$phi$trans$insert25$i = 0, $$pr$i$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i12 = 0, $$pre$i26 = 0, $$pre$i39 = 0;
 var $$pre$i53 = 0, $$pre$i64 = 0, $$pre$phi$iZ2D = 0, $$pre$phi153Z2D = 0, $$pre$phi155Z2D = 0, $$pre$phi6$i$iZ2D = 0, $$pre$pre$i$i = 0, $$pre142 = 0, $$pre144 = 0, $$pre145 = 0, $$pre147 = 0, $$pre148 = 0, $$pre149 = 0, $$pre150 = 0, $$pre154 = 0, $$pre157 = 0, $$pre159 = 0, $$pre17$i = 0, $$pre19$i$i = 0, $$pre24$i = 0;
 var $$pre26$i = 0, $$pre3$i$i = 0, $$pre4$i$i = 0, $$sum$i = 0, $$sum$i$i = 0, $$sum$i32 = 0, $$sum$i41 = 0, $$sum$i55 = 0, $$sum1$i = 0, $$sum1$i58 = 0, $$sum10$i = 0, $$sum11$i = 0, $$sum12$i = 0, $$sum13$i = 0, $$sum14$i = 0, $$sum15$i = 0, $$sum16$i = 0, $$sum17$i = 0, $$sum19$i = 0, $$sum2$i = 0;
 var $$sum2$i61 = 0, $$sum20$i = 0, $$sum3$i = 0, $$sum4$i = 0, $$sum5$i = 0, $$sum6$i = 0, $$sum65$i = 0, $$sum66$i = 0, $$sum8$i = 0, $$sum9$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0;
 var $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0;
 var $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0;
 var $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0;
 var $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0;
 var $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0;
 var $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0;
 var $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0;
 var $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0;
 var $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0;
 var $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0;
 var $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0;
 var $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0;
 var $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0;
 var $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0;
 var $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0, $1267 = 0, $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0;
 var $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0, $1285 = 0, $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0;
 var $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0, $1302 = 0, $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0;
 var $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0, $1320 = 0, $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0;
 var $1330 = 0, $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0, $1339 = 0, $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0;
 var $1349 = 0, $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0, $1357 = 0, $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0;
 var $1367 = 0, $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0, $1375 = 0, $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0;
 var $1385 = 0, $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0, $1393 = 0, $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0;
 var $1402 = 0, $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0, $1410 = 0, $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0;
 var $1420 = 0, $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0, $1429 = 0, $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0;
 var $1439 = 0, $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0, $1447 = 0, $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0;
 var $1457 = 0, $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0, $1465 = 0, $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0;
 var $1475 = 0, $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0, $1483 = 0, $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0;
 var $1493 = 0, $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0, $1500 = 0, $1501 = 0, $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0;
 var $1510 = 0, $1511 = 0, $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0, $1518 = 0, $1519 = 0, $152 = 0, $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0;
 var $1529 = 0, $153 = 0, $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0, $1536 = 0, $1537 = 0, $1538 = 0, $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0;
 var $1547 = 0, $1548 = 0, $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0, $1554 = 0, $1555 = 0, $1556 = 0, $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0;
 var $1565 = 0, $1566 = 0, $1567 = 0, $1568 = 0, $1569 = 0, $157 = 0, $1570 = 0, $1571 = 0, $1572 = 0, $1573 = 0, $1574 = 0, $1575 = 0, $1576 = 0, $1577 = 0, $1578 = 0, $1579 = 0, $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0;
 var $1583 = 0, $1584 = 0, $1585 = 0, $1586 = 0, $1587 = 0, $1588 = 0, $1589 = 0, $159 = 0, $1590 = 0, $1591 = 0, $1592 = 0, $1593 = 0, $1594 = 0, $1595 = 0, $1596 = 0, $1597 = 0, $1598 = 0, $1599 = 0, $16 = 0, $160 = 0;
 var $1600 = 0, $1601 = 0, $1602 = 0, $1603 = 0, $1604 = 0, $1605 = 0, $1606 = 0, $1607 = 0, $1608 = 0, $1609 = 0, $161 = 0, $1610 = 0, $1611 = 0, $1612 = 0, $1613 = 0, $1614 = 0, $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0;
 var $1619 = 0, $162 = 0, $1620 = 0, $1621 = 0, $1622 = 0, $1623 = 0, $1624 = 0, $1625 = 0, $1626 = 0, $1627 = 0, $1628 = 0, $1629 = 0, $163 = 0, $1630 = 0, $1631 = 0, $1632 = 0, $1633 = 0, $1634 = 0, $1635 = 0, $1636 = 0;
 var $1637 = 0, $1638 = 0, $1639 = 0, $164 = 0, $1640 = 0, $1641 = 0, $1642 = 0, $1643 = 0, $1644 = 0, $1645 = 0, $1646 = 0, $1647 = 0, $1648 = 0, $1649 = 0, $165 = 0, $1650 = 0, $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0;
 var $1655 = 0, $1656 = 0, $1657 = 0, $1658 = 0, $1659 = 0, $166 = 0, $1660 = 0, $1661 = 0, $1662 = 0, $1663 = 0, $1664 = 0, $1665 = 0, $1666 = 0, $1667 = 0, $1668 = 0, $1669 = 0, $167 = 0, $1670 = 0, $1671 = 0, $1672 = 0;
 var $1673 = 0, $1674 = 0, $1675 = 0, $1676 = 0, $1677 = 0, $1678 = 0, $1679 = 0, $168 = 0, $1680 = 0, $1681 = 0, $1682 = 0, $1683 = 0, $1684 = 0, $1685 = 0, $1686 = 0, $1687 = 0, $1688 = 0, $1689 = 0, $169 = 0, $1690 = 0;
 var $1691 = 0, $1692 = 0, $1693 = 0, $1694 = 0, $1695 = 0, $1696 = 0, $1697 = 0, $1698 = 0, $1699 = 0, $17 = 0, $170 = 0, $1700 = 0, $1701 = 0, $1702 = 0, $1703 = 0, $1704 = 0, $1705 = 0, $1706 = 0, $1707 = 0, $1708 = 0;
 var $1709 = 0, $171 = 0, $1710 = 0, $1711 = 0, $1712 = 0, $1713 = 0, $1714 = 0, $1715 = 0, $1716 = 0, $1717 = 0, $1718 = 0, $1719 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0;
 var $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0;
 var $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0;
 var $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0;
 var $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0;
 var $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0;
 var $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0;
 var $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0;
 var $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0;
 var $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0;
 var $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0;
 var $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0;
 var $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0;
 var $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0;
 var $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0;
 var $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0;
 var $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0;
 var $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0;
 var $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0;
 var $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0;
 var $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0;
 var $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0;
 var $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0;
 var $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0;
 var $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0;
 var $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0;
 var $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0;
 var $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0;
 var $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0;
 var $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0;
 var $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0;
 var $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0;
 var $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0;
 var $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0;
 var $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0;
 var $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0;
 var $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0;
 var $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0;
 var $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0;
 var $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0;
 var $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0;
 var $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0;
 var $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0;
 var $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0;
 var $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0;
 var $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $alphabet$0$i = 0, $avail_fragy0$0107 = 0, $avail_fragy_end$0108 = 0, $bi$024$i = 0, $cbmvs$i = 0, $coded$013$i = 0, $coded$08$i = 0, $coded$i = 0, $codedi$1$1$i = 0, $codedi$1$2$i = 0;
 var $codedi$1$328$i = 0, $codedi$1$329$i = 0, $codedi$1$i = 0, $codedi$25$i = 0, $codedi$36$i = 0, $cw$01$i = 0, $dc_qi$0$lcssa$i = 0, $dc_qi$0$lcssa$pn$i = 0, $dc_qi$030$i = 0, $dc_qi$1$i = 0, $dc_qi$1$lcssa$i = 0, $dc_qi$1$lcssa$i$lcssa = 0, $dc_qi$118$i = 0, $dc_qi$122$i = 0, $dc_qi$212$i = 0, $dc_quant$i = 0, $dct_coeffs$i = 0, $dst$0$lcssa$i = 0, $dst$042$i = 0, $dst$1$lcssa$i = 0;
 var $dst$128$i = 0, $dst$2$lcssa$i = 0, $dst$214$i = 0, $eb$0$i$i = 0, $eb$0$i5$i = 0, $eb$0$us$i$i = 0, $edelay$0 = 0, $edelay$1 = 0, $eob$0$i = 0, $eob_count$0$lcssa$i$i = 0, $eob_count$0$lcssa$us$i$i = 0, $eob_count$0$ph$lcssa$i$i = 0, $eob_count$0$ph26$i$i = 0, $eob_count$01$i$i = 0, $eob_count$01$us$i$i = 0, $eobi$17$i$i = 0, $eobi$220$i$i = 0, $eobs$024$i = 0, $eobs$035$i$i = 0, $eobs$1$lcssa$i = 0;
 var $eobs$1$lcssa$i$i = 0, $eobs$1$lcssa$i$i$lcssa = 0, $eobs$120$i = 0, $exitcond = 0, $exitcond$i = 0, $exitcond$i$i = 0, $exitcond$i$i25 = 0, $exitcond$i$i34 = 0, $exitcond$i10 = 0, $exitcond$i14 = 0, $exitcond$i18 = 0, $exitcond$i30 = 0, $exitcond$i44 = 0, $exitcond$i5 = 0, $exitcond$i51 = 0, $exitcond$i60 = 0, $exitcond$i8$i = 0, $exitcond138 = 0, $exitcond139 = 0, $exitcond140 = 0;
 var $exitcond141 = 0, $exitcond15$i = 0, $exitcond16$i = 0, $exitcond18$i$i = 0, $exitcond21$i = 0, $exitcond22$i = 0, $exitcond40$1$i = 0, $exitcond40$2$i = 0, $exitcond40$i = 0, $exitcond41$i = 0, $exitcond42$i = 0, $exitcond50$i = 0, $exitcond51$i = 0, $exitcond52$i = 0, $exitcond64$i = 0, $flag$0$be$i = 0, $flag$0$be$i$i = 0, $flag$0$be$i2$i = 0, $flag$0$i = 0, $flag$012$i$i = 0;
 var $flag$022$i = 0, $flag$05$i$i = 0, $flag$1$be$i = 0, $flag$120$i = 0, $flag$140$i = 0, $flag$2$lcssa$i = 0, $flag$231$i = 0, $flag$325$i = 0, $flag$418$i = 0, $flag$5$i = 0, $flag$611$i = 0, $flag$64$i = 0, $flag$65$i = 0, $flag$7$i = 0, $flag$8$i = 0, $flag$8$i$lcssa = 0, $frag$010$i = 0, $frag$1$lcssa$i = 0, $frag$15$i = 0, $fragi$036$i = 0;
 var $fragi$130$i = 0, $fragi$224$i = 0, $fragi$3$i = 0, $fragii$015$i = 0, $fragii$02$i$i = 0, $fragii$036$i$i = 0, $fragii$07$i = 0, $fragii$1$lcssa$i$i = 0, $fragii$124$i = 0, $fragii$18$i$i = 0, $fragii$2$i = 0, $fragii$2$lcssa$i$i = 0, $fragii$2$ph27$i$i = 0, $fragii$210$i$i = 0, $fragii$210$i$i$lcssa = 0, $fragii$3$i = 0, $fragii$3$i$lcssa = 0, $fragii$3$lcssa$i$i = 0, $fragii$3$lcssa$i$i$lcssa = 0, $fragii$321$i$i = 0;
 var $fragii$421$i = 0, $fragii$517$i = 0, $fragii$517$i$lcssa = 0, $fragx$029$i = 0, $fragx$123$i = 0, $fragy$035$i = 0, $hgi$026$i = 0, $huff_idxs$i = 0, $idata$08$i = 0, $ispos$i = 0, $ispos4$i = 0, $ispos6$i = 0, $l_ref$020$i = 0, $l_ref$1$i = 0, $last_mv$sroa$0$sroa$0$017$i = 0, $last_mv$sroa$0$sroa$0$1$i = 0, $last_mv$sroa$0$sroa$0$2$i = 0, $last_mv$sroa$0$sroa$0$3$i = 0, $last_mv$sroa$0$sroa$0$4$i = 0, $last_mv$sroa$0$sroa$8$016$i = 0;
 var $last_mv$sroa$0$sroa$8$1$i = 0, $last_mv$sroa$0$sroa$8$2$i = 0, $last_mv$sroa$0$sroa$8$3$i = 0, $last_mv$sroa$0$sroa$8$4$i = 0, $last_mv$sroa$11$sroa$0$09$i = 0, $last_mv$sroa$11$sroa$0$1$i = 0, $last_mv$sroa$11$sroa$0$2$i = 0, $last_mv$sroa$11$sroa$0$3$i = 0, $last_mv$sroa$11$sroa$0$4$i = 0, $last_mv$sroa$11$sroa$5$08$i = 0, $last_mv$sroa$11$sroa$5$1$i = 0, $last_mv$sroa$11$sroa$5$2$i = 0, $last_mv$sroa$11$sroa$5$3$i = 0, $last_mv$sroa$11$sroa$5$4$i = 0, $lbmvs$i = 0, $lti$1$i = 0, $mapii$0$i = 0, $mbi$010$i = 0, $mbi$05$i = 0, $mbmv$sroa$0$018$i = 0;
 var $mbmv$sroa$0$1$i = 0, $mbmv$sroa$0$2$i = 0, $mbmv$sroa$0$3$i = 0, $mbmv$sroa$7$019$i = 0, $mbmv$sroa$7$1$i = 0, $mbmv$sroa$7$2$i = 0, $mbmv$sroa$7$3$i = 0, $ncoded$0$i = 0, $ncoded$1$i = 0, $ncoded$1$i$lcssa = 0, $ncoded_fragis$0$lcssa$i = 0, $ncoded_fragis$037$i = 0, $ncoded_fragis$037$i$i = 0, $ncoded_fragis$044$i = 0, $ncoded_fragis$1$lcssa$i = 0, $ncoded_fragis$1$lcssa$i$lcssa = 0, $ncoded_fragis$1$lcssa$i6 = 0, $ncoded_fragis$1$lcssa$i6$lcssa = 0, $ncoded_fragis$131$i = 0, $ncoded_fragis$134$i = 0;
 var $ncoded_fragis$18$i = 0, $ncoded_fragis$2$i = 0, $ncoded_fragis$2$i$lcssa = 0, $ncoded_fragis$227$i = 0, $ncoded_fragis$26$i = 0, $ncoded_fragis$320$i = 0, $ncoded_fragis$327$i = 0, $ncoded_fragis$4$1$i = 0, $ncoded_fragis$4$2$i = 0, $ncoded_fragis$4$i = 0, $ncoded_fragis$4$i3 = 0, $ncoded_fragis$4$i43 = 0, $ncoded_fragis$4$i43$lcssa = 0, $ncoded_fragis$5$i = 0, $ncoded_fragis$5$i$lcssa = 0, $ncoded_fragis$5$i4 = 0, $ncoded_fragis$5$i45 = 0, $ncoded_fragis$5$i45$lcssa = 0, $ncoded_fragis$6$i = 0, $ncoded_fragis$6$i$lcssa = 0;
 var $neg$i = 0, $neg5$i = 0, $neg7$i = 0, $notstart$0111 = 0, $npartial$04$i$i = 0, $npartial$04$i$i$lcssa = 0, $nqi1$023$i = 0, $nqi1$1$i = 0, $nsbs$011$i = 0, $nsbs$046$i = 0, $ntoks$0$be$i$i = 0, $ntoks$0$be$i$i$lcssa = 0, $ntoks$0$be$us$i$i = 0, $ntoks$0$be$us$i$i$lcssa = 0, $ntoks$0$lcssa$i$i = 0, $ntoks$0$lcssa$us$i$i = 0, $ntoks_left$i = 0, $nuncoded_fragis$045$i = 0, $nuncoded_fragis$1$lcssa$i = 0, $nuncoded_fragis$135$i = 0;
 var $nuncoded_fragis$228$i = 0, $nuncoded_fragis$322$i = 0, $nuncoded_fragis$4$i = 0, $nuncoded_fragis$5$i = 0, $nuncoded_fragis$6$i = 0, $nuncoded_fragis$6$i$lcssa = 0, $oc_clc_mode_unpack$oc_vlc_mode_unpack$i = 0, $or$cond = 0, $or$cond$i = 0, $or$cond$i$i = 0, $or$cond$i21 = 0, $or$cond$i67 = 0, $or$cond1$i$i = 0, $or$cond113 = 0, $or$cond2$i = 0, $phitmp$i = 0, $pipe = 0, $pli$0109 = 0, $pli$011$i$i = 0, $pli$011$us$i$i = 0;
 var $pli$013$i = 0, $pli$034$i$i = 0, $pli$042$i = 0, $pred$0$i = 0, $prev_ncoded_fragis$010$i = 0, $qii$06$1$i = 0, $qii$06$2$i = 0, $qii$06$i = 0, $quadi$029$i = 0, $quadi$04$i = 0, $refi$0 = 0, $refi$0$lcssa = 0, $refi$1 = 0, $run_count$0$i = 0, $run_count$041$i = 0, $run_count$09$i$i = 0, $run_count$1$i$i = 0, $run_count$1$lcssa$i = 0, $run_count$116$i = 0, $run_count$132$i = 0;
 var $run_count$2$i = 0, $run_count$226$i = 0, $run_count$319$i = 0, $run_count$4$i = 0, $run_count$512$i = 0, $run_count$56$i = 0, $run_count$57$i = 0, $run_count$6$i = 0, $run_count$7$i = 0, $run_count$7$i$lcssa = 0, $run_counts$i$i = 0, $sbi$0$i$i = 0, $sbi$0$i$i$lcssa = 0, $sbi$012$i = 0, $sbi$043$i = 0, $sbi$06$i$i = 0, $sbi$1$i$i = 0, $sbi$1$lcssa$i = 0, $sbi$1$lcssa$i7 = 0, $sbi$113$i$i = 0;
 var $sbi$133$i = 0, $sbi$17$i = 0, $sbi$210$i$i = 0, $sbi$210$i$i$lcssa = 0, $scevgep$i = 0, $scevgep$i65 = 0, $scevgep13$i = 0, $scevgep58$i = 0, $scevgep59$i = 0, $scevgep60$i = 0, $scevgep61$i = 0, $scevgep62$i = 0, $scevgep63$i = 0, $scheme0_alphabet$i = 0, $sdelay$0 = 0, $sdelay$1 = 0, $smax$i = 0, $smax$i$i = 0, $smax46$i$i = 0, $smax48$i$i = 0;
 var $smax49$i$i = 0, $src$0$lcssa$i = 0, $src$041$i = 0, $src$1$lcssa$i = 0, $src$126$i = 0, $src$213$i = 0, $stripe_buf = 0, $stripe_fragy$0112 = 0, $ti$0$lcssa$i$i = 0, $ti$013$i$i = 0, $ti$013$us$i$i = 0, $ti$033$i$i = 0, $ti$1$lcssa$i$i = 0, $ti$1$lcssa$i$i$lcssa = 0, $ti$1$lcssa$i7$i = 0, $ti$1$lcssa$us$i$i = 0, $ti$1$ph25$i$i = 0, $ti$12$i$i = 0, $ti$12$us$i$i = 0, $ti$19$i$i = 0;
 var $ti$2$i$i = 0, $ti$2$i3$i = 0, $ti$2$us$i$i = 0, $ti$3$i$i = 0, $ti$3$i$i$lcssa = 0, $ti$3$i$i$lcssa$lcssa = 0, $ti$3$i4$i = 0, $ti$3$i4$i$lcssa = 0, $ti$3$us$i$i = 0, $ti$3$us$i$i$lcssa = 0, $u_ref$0$ph$i = 0, $u_ref$022$i = 0, $ul_ref$021$i = 0, $ul_ref$021$i$phi = 0, $umax$i$i = 0, $umax9$i$i = 0, $ur_ref$0$i = 0, $variance$0$lcssa$i = 0, $variance$0$pn19$i = 0, $variance$0$pn19$i$phi = 0;
 var $variance$031$i = 0, $variance$09$i = 0, $variance$1$i = 0, $variance$1$lcssa$i = 0, $variance$1$lcssa$i$lcssa = 0, $variance$1$lcssa$i69 = 0, $variance$1$sum$i = 0, $variance$117$i = 0, $variance$12$i = 0, $variance$121$i = 0, $variance$210$i = 0, $x$01$i = 0, $x$020$i = 0, $x$111$i = 0, $y$0$lcssa$i = 0, $y$043$i = 0, $y$07$i = 0, $y$1$lcssa$i = 0, $y$132$i = 0, $y$215$i = 0;
 var $zzi$027$1$i = 0, $zzi$027$2$i = 0, $zzi$027$i = 0, $zzi$125$i = 0, $zzi$16$i = 0, $zzi$16$i$lcssa = 0, $zzi$16$i247 = 0, $zzi$2$lcssa$i = 0, $zzi$221$i = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 4048|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $dc_quant$i = sp + 4036|0;
 $dct_coeffs$i = sp + 3904|0;
 $run_counts$i$i = sp + 3640|0;
 $ntoks_left$i = sp + 2872|0;
 $huff_idxs$i = sp + 2864|0;
 $cbmvs$i = sp + 4040|0;
 $coded$i = sp + 2808|0;
 $lbmvs$i = sp + 3896|0;
 $scheme0_alphabet$i = sp;
 $pipe = sp + 56|0;
 $stripe_buf = sp + 8|0;
 $0 = ($_dec|0)==(0|0);
 $1 = ($_op|0)==(0|0);
 $or$cond = $0 | $1;
 if ($or$cond) {
  $$1 = -1;
  STACKTOP = sp;return ($$1|0);
 }
 $2 = ((($_op)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0);
 if ($4) {
  $1643 = ((($_dec)) + 272|0);
  $1644 = HEAP32[$1643>>2]|0;
  $1645 = ($1644|0)<(0);
  $$pre157 = ((($_dec)) + 276|0);
  if ($1645) {
   label = 361;
  } else {
   $1646 = HEAP32[$$pre157>>2]|0;
   $1647 = ($1646|0)<(0);
   if ($1647) {
    label = 361;
   }
  }
  if ((label|0) == 361) {
   HEAP32[$1643>>2] = 0;
   HEAP32[$$pre157>>2] = 0;
   $1648 = ((($_dec)) + 280|0);
   HEAP32[$1648>>2] = 1;
   $1649 = ((($_dec)) + 4|0);
   $1650 = HEAP32[$1649>>2]|0;
   $1651 = (($1650) + 32)|0;
   $1652 = ((($_dec)) + 8|0);
   $1653 = HEAP32[$1652>>2]|0;
   $1654 = (($1653) + 32)|0;
   $1655 = ((($_dec)) + 48|0);
   $1656 = HEAP32[$1655>>2]|0;
   $1657 = $1656 & 1;
   $1658 = $1657 ^ 1;
   $1659 = $1651 >> $1658;
   $1660 = $1656 >>> 1;
   $$lobit$i70 = $1660 & 1;
   $1661 = $$lobit$i70 ^ 1;
   $1662 = $1654 >> $1661;
   $1663 = Math_imul($1654, $1651)|0;
   $1664 = ((($_dec)) + 480|0);
   $1665 = HEAP32[$1664>>2]|0;
   $1666 = $1659 << 1;
   $1667 = Math_imul($1666, $1662)|0;
   $1668 = (($1667) + ($1663))|0;
   _memset(($1665|0),-128,($1668|0))|0;
   $1669 = HEAP32[$$pre157>>2]|0;
   HEAP32[$1648>>2] = $1669;
   $1670 = ((($_dec)) + 53976|0);
   $1671 = (((($_dec)) + 288|0) + (($1669*48)|0)|0);
   dest=$1670; src=$1671; stop=dest+48|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
  }
  $1672 = ((($_dec)) + 768|0);
  $1673 = $1672;
  $1674 = $1673;
  $1675 = HEAP32[$1674>>2]|0;
  $1676 = (($1673) + 4)|0;
  $1677 = $1676;
  $1678 = HEAP32[$1677>>2]|0;
  $1679 = ((($_dec)) + 793|0);
  $1680 = HEAP8[$1679>>0]|0;
  $1681 = $1680&255;
  $1682 = (_i64Add(($1681|0),0,($1675|0),($1678|0))|0);
  $1683 = tempRet0;
  $1684 = ((($_dec)) + 60|0);
  $1685 = HEAP32[$1684>>2]|0;
  $1686 = (_bitshift64Shl(($1682|0),($1683|0),($1685|0))|0);
  $1687 = tempRet0;
  $1688 = ((($_dec)) + 776|0);
  $1689 = $1688;
  $1690 = $1689;
  $1691 = HEAP32[$1690>>2]|0;
  $1692 = (($1689) + 4)|0;
  $1693 = $1692;
  $1694 = HEAP32[$1693>>2]|0;
  $1695 = (_i64Subtract(($1691|0),($1694|0),($1675|0),($1678|0))|0);
  $1696 = tempRet0;
  $1697 = (_i64Add(($1695|0),($1696|0),($1686|0),($1687|0))|0);
  $1698 = tempRet0;
  $1699 = ((($_dec)) + 784|0);
  $1700 = $1699;
  $1701 = $1700;
  HEAP32[$1701>>2] = $1697;
  $1702 = (($1700) + 4)|0;
  $1703 = $1702;
  HEAP32[$1703>>2] = $1698;
  $1704 = (_i64Add(($1691|0),($1694|0),1,0)|0);
  $1705 = tempRet0;
  $1706 = $1688;
  $1707 = $1706;
  HEAP32[$1707>>2] = $1704;
  $1708 = (($1706) + 4)|0;
  $1709 = $1708;
  HEAP32[$1709>>2] = $1705;
  $1710 = ($_granpos|0)==(0|0);
  if ($1710) {
   $$1 = 1;
   STACKTOP = sp;return ($$1|0);
  }
  $1711 = $_granpos;
  $1712 = $1711;
  HEAP32[$1712>>2] = $1697;
  $1713 = (($1711) + 4)|0;
  $1714 = $1713;
  HEAP32[$1714>>2] = $1698;
  $$1 = 1;
  STACKTOP = sp;return ($$1|0);
 }
 $5 = ((($_dec)) + 51556|0);
 $6 = HEAP32[$_op>>2]|0;
 _oc_pack_readinit($5,$6,$3);
 $7 = (_oc_pack_read1($5)|0);
 $8 = ($7|0)==(0);
 do {
  if ($8) {
   $9 = (_oc_pack_read1($5)|0);
   $10 = $9&255;
   $11 = ((($_dec)) + 792|0);
   HEAP8[$11>>0] = $10;
   $12 = (_oc_pack_read($5,6)|0);
   $13 = $12&255;
   $14 = ((($_dec)) + 795|0);
   HEAP8[$14>>0] = $13;
   $15 = (_oc_pack_read1($5)|0);
   $16 = ($15|0)==(0);
   do {
    if ($16) {
     $17 = ((($_dec)) + 794|0);
     HEAP8[$17>>0] = 1;
    } else {
     $18 = (_oc_pack_read($5,6)|0);
     $19 = $18&255;
     $20 = ((($_dec)) + 796|0);
     HEAP8[$20>>0] = $19;
     $21 = (_oc_pack_read1($5)|0);
     $22 = ($21|0)==(0);
     if ($22) {
      $23 = ((($_dec)) + 794|0);
      HEAP8[$23>>0] = 2;
      break;
     } else {
      $24 = (_oc_pack_read($5,6)|0);
      $25 = $24&255;
      $26 = ((($_dec)) + 797|0);
      HEAP8[$26>>0] = $25;
      $27 = ((($_dec)) + 794|0);
      HEAP8[$27>>0] = 3;
      break;
     }
    }
   } while(0);
   $28 = HEAP8[$11>>0]|0;
   $29 = ($28<<24>>24)==(0);
   if ($29) {
    $30 = (_oc_pack_read($5,3)|0);
    $31 = ($30|0)==(0);
    if (!($31)) {
     $$0 = -23;
     break;
    }
    $$pre = HEAP8[$11>>0]|0;
    $32 = ($$pre<<24>>24)==(0);
    if ($32) {
     $$phi$trans$insert = ((($_dec)) + 272|0);
     $$pre142 = HEAP32[$$phi$trans$insert>>2]|0;
     $$phi$trans$insert143 = ((($_dec)) + 276|0);
     $$pre144 = HEAP32[$$phi$trans$insert143>>2]|0;
     $1717 = $_dec;$1718 = 0;$61 = $$pre142;$63 = $$pre144;
     label = 16;
    } else {
     $1715 = $_dec;$1716 = $$pre;
     label = 14;
    }
   } else {
    $1715 = $_dec;$1716 = $28;
    label = 14;
   }
   do {
    if ((label|0) == 14) {
     $33 = ((($_dec)) + 272|0);
     $34 = HEAP32[$33>>2]|0;
     $35 = ($34|0)<(0);
     $$pre159 = ((($_dec)) + 276|0);
     if (!($35)) {
      $36 = HEAP32[$$pre159>>2]|0;
      $37 = ($36|0)<(0);
      if (!($37)) {
       $1717 = $1715;$1718 = $1716;$61 = $34;$63 = $36;
       label = 16;
       break;
      }
     }
     HEAP32[$33>>2] = 0;
     HEAP32[$$pre159>>2] = 0;
     $38 = ((($_dec)) + 280|0);
     HEAP32[$38>>2] = 1;
     $39 = ((($_dec)) + 4|0);
     $40 = HEAP32[$39>>2]|0;
     $41 = (($40) + 32)|0;
     $42 = ((($_dec)) + 8|0);
     $43 = HEAP32[$42>>2]|0;
     $44 = (($43) + 32)|0;
     $45 = ((($_dec)) + 48|0);
     $46 = HEAP32[$45>>2]|0;
     $47 = $46 & 1;
     $48 = $47 ^ 1;
     $49 = $41 >> $48;
     $50 = $46 >>> 1;
     $$lobit$i = $50 & 1;
     $51 = $$lobit$i ^ 1;
     $52 = $44 >> $51;
     $53 = Math_imul($44, $41)|0;
     $54 = ((($_dec)) + 480|0);
     $55 = HEAP32[$54>>2]|0;
     $56 = $49 << 1;
     $57 = Math_imul($56, $52)|0;
     $58 = (($57) + ($53))|0;
     _memset(($55|0),-128,($58|0))|0;
     $59 = HEAP32[$38>>2]|0;
     $$pre145 = HEAP8[$11>>0]|0;
     $1086 = $1715;$67 = $$pre145;$refi$1 = $59;
    }
   } while(0);
   if ((label|0) == 16) {
    $refi$0 = 0;
    while(1) {
     $60 = ($refi$0|0)==($61|0);
     $62 = ($refi$0|0)==($63|0);
     $or$cond113 = $60 | $62;
     $64 = (($refi$0) + 1)|0;
     if ($or$cond113) {
      $refi$0 = $64;
     } else {
      $refi$0$lcssa = $refi$0;
      break;
     }
    }
    $65 = ((($_dec)) + 280|0);
    HEAP32[$65>>2] = $refi$0$lcssa;
    $1086 = $1717;$67 = $1718;$refi$1 = $refi$0$lcssa;
   }
   $66 = ($67<<24>>24)==(0);
   if ($66) {
    $68 = ((($_dec)) + 252|0);
    $69 = HEAP32[$68>>2]|0;
    $70 = ((($_dec)) + 220|0);
    $71 = HEAP32[$70>>2]|0;
    $72 = ((($_dec)) + 224|0);
    $73 = HEAP32[$72>>2]|0;
    $74 = ((($_dec)) + 204|0);
    $75 = HEAP32[$74>>2]|0;
    $nsbs$011$i = 0;$pli$013$i = 0;$prev_ncoded_fragis$010$i = 0;$sbi$012$i = 0;
    while(1) {
     $76 = (((((($_dec)) + 108|0) + ($pli$013$i<<5)|0)) + 28|0);
     $77 = HEAP32[$76>>2]|0;
     $78 = (($77) + ($nsbs$011$i))|0;
     $79 = ($sbi$012$i>>>0)<($78>>>0);
     if ($79) {
      $ncoded_fragis$18$i = $prev_ncoded_fragis$010$i;$sbi$17$i = $sbi$012$i;
      while(1) {
       $80 = (($73) + ($sbi$17$i)|0);
       $ncoded_fragis$26$i = $ncoded_fragis$18$i;$quadi$04$i = 0;
       while(1) {
        $81 = HEAP8[$80>>0]|0;
        $82 = ($81&255) >>> 2;
        $83 = $82 & 15;
        $84 = $83&255;
        $85 = 1 << $quadi$04$i;
        $86 = $84 & $85;
        $87 = ($86|0)==(0);
        if ($87) {
         $ncoded_fragis$5$i = $ncoded_fragis$26$i;
        } else {
         $88 = ((($71) + ($sbi$17$i<<6)|0) + ($quadi$04$i<<4)|0);
         $89 = HEAP32[$88>>2]|0;
         $90 = ($89|0)>(-1);
         if ($90) {
          $91 = (($75) + ($89<<2)|0);
          $92 = HEAP32[$91>>2]|0;
          $93 = $92 & -1794;
          $94 = $93 | 257;
          HEAP32[$91>>2] = $94;
          $95 = (($ncoded_fragis$26$i) + 1)|0;
          $96 = (($69) + ($ncoded_fragis$26$i<<2)|0);
          HEAP32[$96>>2] = $89;
          $ncoded_fragis$4$i = $95;
         } else {
          $ncoded_fragis$4$i = $ncoded_fragis$26$i;
         }
         $97 = ((((($71) + ($sbi$17$i<<6)|0) + ($quadi$04$i<<4)|0)) + 4|0);
         $98 = HEAP32[$97>>2]|0;
         $99 = ($98|0)>(-1);
         if ($99) {
          $105 = (($75) + ($98<<2)|0);
          $106 = HEAP32[$105>>2]|0;
          $107 = $106 & -1794;
          $108 = $107 | 257;
          HEAP32[$105>>2] = $108;
          $109 = (($ncoded_fragis$4$i) + 1)|0;
          $110 = (($69) + ($ncoded_fragis$4$i<<2)|0);
          HEAP32[$110>>2] = $98;
          $ncoded_fragis$4$1$i = $109;
         } else {
          $ncoded_fragis$4$1$i = $ncoded_fragis$4$i;
         }
         $111 = ((((($71) + ($sbi$17$i<<6)|0) + ($quadi$04$i<<4)|0)) + 8|0);
         $112 = HEAP32[$111>>2]|0;
         $113 = ($112|0)>(-1);
         if ($113) {
          $114 = (($75) + ($112<<2)|0);
          $115 = HEAP32[$114>>2]|0;
          $116 = $115 & -1794;
          $117 = $116 | 257;
          HEAP32[$114>>2] = $117;
          $118 = (($ncoded_fragis$4$1$i) + 1)|0;
          $119 = (($69) + ($ncoded_fragis$4$1$i<<2)|0);
          HEAP32[$119>>2] = $112;
          $ncoded_fragis$4$2$i = $118;
         } else {
          $ncoded_fragis$4$2$i = $ncoded_fragis$4$1$i;
         }
         $120 = ((((($71) + ($sbi$17$i<<6)|0) + ($quadi$04$i<<4)|0)) + 12|0);
         $121 = HEAP32[$120>>2]|0;
         $122 = ($121|0)>(-1);
         if ($122) {
          $123 = (($75) + ($121<<2)|0);
          $124 = HEAP32[$123>>2]|0;
          $125 = $124 & -1794;
          $126 = $125 | 257;
          HEAP32[$123>>2] = $126;
          $127 = (($ncoded_fragis$4$2$i) + 1)|0;
          $128 = (($69) + ($ncoded_fragis$4$2$i<<2)|0);
          HEAP32[$128>>2] = $121;
          $ncoded_fragis$5$i = $127;
         } else {
          $ncoded_fragis$5$i = $ncoded_fragis$4$2$i;
         }
        }
        $100 = (($quadi$04$i) + 1)|0;
        $exitcond$i = ($100|0)==(4);
        if ($exitcond$i) {
         $ncoded_fragis$5$i$lcssa = $ncoded_fragis$5$i;
         break;
        } else {
         $ncoded_fragis$26$i = $ncoded_fragis$5$i;$quadi$04$i = $100;
        }
       }
       $101 = (($sbi$17$i) + 1)|0;
       $exitcond15$i = ($101|0)==($78|0);
       if ($exitcond15$i) {
        $ncoded_fragis$1$lcssa$i = $ncoded_fragis$5$i$lcssa;$sbi$1$lcssa$i = $78;
        break;
       } else {
        $ncoded_fragis$18$i = $ncoded_fragis$5$i$lcssa;$sbi$17$i = $101;
       }
      }
     } else {
      $ncoded_fragis$1$lcssa$i = $prev_ncoded_fragis$010$i;$sbi$1$lcssa$i = $sbi$012$i;
     }
     $102 = (($ncoded_fragis$1$lcssa$i) - ($prev_ncoded_fragis$010$i))|0;
     $103 = (((($_dec)) + 256|0) + ($pli$013$i<<2)|0);
     HEAP32[$103>>2] = $102;
     $104 = (($pli$013$i) + 1)|0;
     $exitcond16$i = ($104|0)==(3);
     if ($exitcond16$i) {
      $ncoded_fragis$1$lcssa$i$lcssa = $ncoded_fragis$1$lcssa$i;
      break;
     } else {
      $nsbs$011$i = $78;$pli$013$i = $104;$prev_ncoded_fragis$010$i = $ncoded_fragis$1$lcssa$i;$sbi$012$i = $sbi$1$lcssa$i;
     }
    }
    $129 = ((($_dec)) + 268|0);
    HEAP32[$129>>2] = $ncoded_fragis$1$lcssa$i$lcssa;
    $130 = ((($_dec)) + 776|0);
    $131 = $130;
    $132 = $131;
    $133 = HEAP32[$132>>2]|0;
    $134 = (($131) + 4)|0;
    $135 = $134;
    $136 = HEAP32[$135>>2]|0;
    $137 = ((($_dec)) + 768|0);
    $138 = $137;
    $139 = $138;
    HEAP32[$139>>2] = $133;
    $140 = (($138) + 4)|0;
    $141 = $140;
    HEAP32[$141>>2] = $136;
    $495 = $ncoded_fragis$1$lcssa$i$lcssa;
   } else {
    $142 = (_oc_pack_read1($5)|0);
    $143 = ((($_dec)) + 224|0);
    $144 = HEAP32[$143>>2]|0;
    $145 = ((($_dec)) + 228|0);
    $146 = HEAP32[$145>>2]|0;
    $147 = ($146|0)==(0);
    if ($147) {
     $flag$0$i = 0;
    } else {
     $flag$05$i$i = $142;$npartial$04$i$i = 0;$sbi$06$i$i = 0;
     while(1) {
      $148 = (_oc_sb_run_unpack($5)|0);
      $149 = $flag$05$i$i&255;
      $150 = ($149 << 1)&255;
      $151 = $150 & 2;
      $152 = (0 - ($148))|0;
      $153 = (($sbi$06$i$i) + 1)|0;
      $154 = ($146>>>0)>($153>>>0);
      $umax$i$i = $154 ? $146 : $153;
      $155 = (($sbi$06$i$i) - ($umax$i$i))|0;
      $156 = ($155>>>0)<($152>>>0);
      $umax9$i$i = $156 ? $152 : $155;
      $157 = (($sbi$06$i$i) - ($umax9$i$i))|0;
      $sbi$1$i$i = $sbi$06$i$i;
      while(1) {
       $158 = (($144) + ($sbi$1$i$i)|0);
       $159 = HEAP8[$158>>0]|0;
       $160 = $159 & -4;
       $161 = $160 | $151;
       HEAP8[$158>>0] = $161;
       $162 = (($sbi$1$i$i) + 1)|0;
       $exitcond$i$i = ($162|0)==($157|0);
       if ($exitcond$i$i) {
        break;
       } else {
        $sbi$1$i$i = $162;
       }
      }
      $163 = Math_imul($umax9$i$i, $flag$05$i$i)|0;
      $164 = ($157>>>0)<($146>>>0);
      $165 = ($148>>>0)>(4128);
      $166 = (($npartial$04$i$i) - ($163))|0;
      $or$cond1$i$i = $165 & $164;
      if ($or$cond1$i$i) {
       $167 = (_oc_pack_read1($5)|0);
       $flag$0$be$i$i = $167;
      } else {
       $168 = ($flag$05$i$i|0)==(0);
       $169 = $168&1;
       $flag$0$be$i$i = $169;
      }
      if ($164) {
       $flag$05$i$i = $flag$0$be$i$i;$npartial$04$i$i = $166;$sbi$06$i$i = $157;
      } else {
       $$lcssa268 = $163;$$lcssa269 = $166;$npartial$04$i$i$lcssa = $npartial$04$i$i;
       break;
      }
     }
     $$pre$i = HEAP32[$145>>2]|0;
     $170 = ($$lcssa269>>>0)<($$pre$i>>>0);
     L78: do {
      if ($170) {
       $171 = HEAP32[$143>>2]|0;
       $sbi$0$i$i = 0;
       while(1) {
        $172 = (($171) + ($sbi$0$i$i)|0);
        $173 = HEAP8[$172>>0]|0;
        $174 = $173 & 2;
        $175 = ($174<<24>>24)==(0);
        $176 = (($sbi$0$i$i) + 1)|0;
        if ($175) {
         $sbi$0$i$i$lcssa = $sbi$0$i$i;
         break;
        } else {
         $sbi$0$i$i = $176;
        }
       }
       $177 = (_oc_pack_read1($5)|0);
       $178 = (_oc_sb_run_unpack($5)|0);
       $179 = ($sbi$0$i$i$lcssa>>>0)<($$pre$i>>>0);
       if ($179) {
        $192 = $178;$flag$012$i$i = $177;$sbi$113$i$i = $sbi$0$i$i$lcssa;
        while(1) {
         $180 = $flag$012$i$i&255;
         $181 = $180 & 1;
         $run_count$09$i$i = $192;$sbi$210$i$i = $sbi$113$i$i;
         while(1) {
          $182 = (($171) + ($sbi$210$i$i)|0);
          $183 = HEAP8[$182>>0]|0;
          $184 = $183 & 2;
          $185 = ($184<<24>>24)==(0);
          if ($185) {
           $186 = ($run_count$09$i$i|0)==(0);
           if ($186) {
            $sbi$210$i$i$lcssa = $sbi$210$i$i;
            break;
           }
           $187 = (($run_count$09$i$i) + -1)|0;
           $188 = $183 & -2;
           $189 = $188 | $181;
           HEAP8[$182>>0] = $189;
           $run_count$1$i$i = $187;
          } else {
           $run_count$1$i$i = $run_count$09$i$i;
          }
          $190 = (($sbi$210$i$i) + 1)|0;
          $191 = ($190>>>0)<($$pre$i>>>0);
          if ($191) {
           $run_count$09$i$i = $run_count$1$i$i;$sbi$210$i$i = $190;
          } else {
           break L78;
          }
         }
         $$not$i$i = ($192>>>0)<(4129);
         if ($$not$i$i) {
          $193 = ($flag$012$i$i|0)==(0);
          $194 = $193&1;
          $flag$0$be$i2$i = $194;
         } else {
          $195 = (_oc_pack_read1($5)|0);
          $flag$0$be$i2$i = $195;
         }
         $196 = (_oc_sb_run_unpack($5)|0);
         $197 = ($sbi$210$i$i$lcssa>>>0)<($$pre$i>>>0);
         if ($197) {
          $192 = $196;$flag$012$i$i = $flag$0$be$i2$i;$sbi$113$i$i = $sbi$210$i$i$lcssa;
         } else {
          break;
         }
        }
       }
      }
     } while(0);
     $198 = ($npartial$04$i$i$lcssa|0)==($$lcssa268|0);
     if ($198) {
      $flag$0$i = 0;
     } else {
      $199 = (_oc_pack_read1($5)|0);
      $200 = ($199|0)==(0);
      $201 = $200&1;
      $flag$0$i = $201;
     }
    }
    $202 = ((($_dec)) + 220|0);
    $203 = HEAP32[$202>>2]|0;
    $204 = HEAP32[$143>>2]|0;
    $205 = ((($_dec)) + 204|0);
    $206 = HEAP32[$205>>2]|0;
    $207 = ((($_dec)) + 252|0);
    $208 = HEAP32[$207>>2]|0;
    $209 = ((($_dec)) + 216|0);
    $210 = HEAP32[$209>>2]|0;
    $flag$140$i = $flag$0$i;$ncoded_fragis$044$i = 0;$nsbs$046$i = 0;$nuncoded_fragis$045$i = 0;$pli$042$i = 0;$run_count$041$i = 0;$sbi$043$i = 0;
    while(1) {
     $211 = (((((($_dec)) + 108|0) + ($pli$042$i<<5)|0)) + 28|0);
     $212 = HEAP32[$211>>2]|0;
     $213 = (($212) + ($nsbs$046$i))|0;
     $214 = ($sbi$043$i>>>0)<($213>>>0);
     if ($214) {
      $flag$231$i = $flag$140$i;$ncoded_fragis$134$i = $ncoded_fragis$044$i;$nuncoded_fragis$135$i = $nuncoded_fragis$045$i;$run_count$132$i = $run_count$041$i;$sbi$133$i = $sbi$043$i;
      while(1) {
       $215 = (($204) + ($sbi$133$i)|0);
       $flag$325$i = $flag$231$i;$ncoded_fragis$227$i = $ncoded_fragis$134$i;$nuncoded_fragis$228$i = $nuncoded_fragis$135$i;$quadi$029$i = 0;$run_count$226$i = $run_count$132$i;
       while(1) {
        $216 = HEAP8[$215>>0]|0;
        $217 = ($216&255) >>> 2;
        $218 = $217 & 15;
        $219 = $218&255;
        $220 = 1 << $quadi$029$i;
        $221 = $219 & $220;
        $222 = ($221|0)==(0);
        if ($222) {
         $flag$8$i = $flag$325$i;$ncoded_fragis$6$i = $ncoded_fragis$227$i;$nuncoded_fragis$6$i = $nuncoded_fragis$228$i;$run_count$7$i = $run_count$226$i;
        } else {
         $bi$024$i = 0;$flag$418$i = $flag$325$i;$ncoded_fragis$320$i = $ncoded_fragis$227$i;$nuncoded_fragis$322$i = $nuncoded_fragis$228$i;$run_count$319$i = $run_count$226$i;
         while(1) {
          $223 = (((($203) + ($sbi$133$i<<6)|0) + ($quadi$029$i<<4)|0) + ($bi$024$i<<2)|0);
          $224 = HEAP32[$223>>2]|0;
          $225 = ($224|0)>(-1);
          if ($225) {
           $226 = HEAP8[$215>>0]|0;
           $227 = $226 & 1;
           $228 = ($227<<24>>24)==(0);
           do {
            if ($228) {
             $229 = $226 & 2;
             $230 = ($229<<24>>24)==(0);
             if ($230) {
              $flag$65$i = $flag$418$i;$run_count$57$i = $run_count$319$i;
             } else {
              $231 = ($run_count$319$i|0)<(1);
              if ($231) {
               $232 = (_oc_pack_read($5,2)|0);
               $233 = $232 & 2;
               $234 = ($233|0)==(0);
               do {
                if ($234) {
                 $235 = (($232) + 1)|0;
                 $$0$i$i = $235;
                } else {
                 $236 = $232 & 1;
                 $237 = ($236|0)==(0);
                 if ($237) {
                  $238 = (_oc_pack_read1($5)|0);
                  $239 = (($238) + 3)|0;
                  $$0$i$i = $239;
                  break;
                 }
                 $240 = (_oc_pack_read($5,2)|0);
                 $241 = $240 & 2;
                 $242 = ($241|0)==(0);
                 if ($242) {
                  $243 = (($240) + 5)|0;
                  $$0$i$i = $243;
                  break;
                 }
                 $244 = $240 & 1;
                 $245 = ($244|0)==(0);
                 if ($245) {
                  $246 = (_oc_pack_read($5,2)|0);
                  $247 = (($246) + 7)|0;
                  $$0$i$i = $247;
                  break;
                 }
                 $248 = (_oc_pack_read($5,3)|0);
                 $249 = $248 & 4;
                 $250 = ($249|0)==(0);
                 if ($250) {
                  $251 = (($248) + 11)|0;
                  $$0$i$i = $251;
                  break;
                 } else {
                  $252 = (_oc_pack_read($5,2)|0);
                  $253 = $248 << 2;
                  $254 = $253 & 12;
                  $255 = (($254) + 15)|0;
                  $256 = (($255) + ($252))|0;
                  $$0$i$i = $256;
                  break;
                 }
                }
               } while(0);
               $257 = ($flag$418$i|0)==(0);
               $258 = $257&1;
               $flag$5$i = $258;$run_count$4$i = $$0$i$i;
              } else {
               $flag$5$i = $flag$418$i;$run_count$4$i = $run_count$319$i;
              }
              $259 = (($run_count$4$i) + -1)|0;
              $260 = ($flag$5$i|0)==(0);
              if ($260) {
               $flag$65$i = 0;$run_count$57$i = $259;
              } else {
               $coded$013$i = $flag$5$i;$flag$611$i = $flag$5$i;$run_count$512$i = $259;
               label = 80;
               break;
              }
             }
             $263 = (($nuncoded_fragis$322$i) + 1)|0;
             $264 = $nuncoded_fragis$322$i ^ -1;
             $$sum$i = (($210) + ($264))|0;
             $265 = (($208) + ($$sum$i<<2)|0);
             HEAP32[$265>>2] = $224;
             $coded$08$i = 0;$flag$64$i = $flag$65$i;$ncoded_fragis$4$i3 = $ncoded_fragis$320$i;$nuncoded_fragis$4$i = $263;$run_count$56$i = $run_count$57$i;
            } else {
             $coded$013$i = 1;$flag$611$i = $flag$418$i;$run_count$512$i = $run_count$319$i;
             label = 80;
            }
           } while(0);
           if ((label|0) == 80) {
            label = 0;
            $261 = (($ncoded_fragis$320$i) + 1)|0;
            $262 = (($208) + ($ncoded_fragis$320$i<<2)|0);
            HEAP32[$262>>2] = $224;
            $coded$08$i = $coded$013$i;$flag$64$i = $flag$611$i;$ncoded_fragis$4$i3 = $261;$nuncoded_fragis$4$i = $nuncoded_fragis$322$i;$run_count$56$i = $run_count$512$i;
           }
           $266 = (($206) + ($224<<2)|0);
           $267 = HEAP32[$266>>2]|0;
           $268 = $coded$08$i & 1;
           $269 = $267 & -2;
           $270 = $269 | $268;
           HEAP32[$266>>2] = $270;
           $flag$7$i = $flag$64$i;$ncoded_fragis$5$i4 = $ncoded_fragis$4$i3;$nuncoded_fragis$5$i = $nuncoded_fragis$4$i;$run_count$6$i = $run_count$56$i;
          } else {
           $flag$7$i = $flag$418$i;$ncoded_fragis$5$i4 = $ncoded_fragis$320$i;$nuncoded_fragis$5$i = $nuncoded_fragis$322$i;$run_count$6$i = $run_count$319$i;
          }
          $271 = (($bi$024$i) + 1)|0;
          $exitcond$i5 = ($271|0)==(4);
          if ($exitcond$i5) {
           $flag$8$i = $flag$7$i;$ncoded_fragis$6$i = $ncoded_fragis$5$i4;$nuncoded_fragis$6$i = $nuncoded_fragis$5$i;$run_count$7$i = $run_count$6$i;
           break;
          } else {
           $bi$024$i = $271;$flag$418$i = $flag$7$i;$ncoded_fragis$320$i = $ncoded_fragis$5$i4;$nuncoded_fragis$322$i = $nuncoded_fragis$5$i;$run_count$319$i = $run_count$6$i;
          }
         }
        }
        $272 = (($quadi$029$i) + 1)|0;
        $exitcond50$i = ($272|0)==(4);
        if ($exitcond50$i) {
         $flag$8$i$lcssa = $flag$8$i;$ncoded_fragis$6$i$lcssa = $ncoded_fragis$6$i;$nuncoded_fragis$6$i$lcssa = $nuncoded_fragis$6$i;$run_count$7$i$lcssa = $run_count$7$i;
         break;
        } else {
         $flag$325$i = $flag$8$i;$ncoded_fragis$227$i = $ncoded_fragis$6$i;$nuncoded_fragis$228$i = $nuncoded_fragis$6$i;$quadi$029$i = $272;$run_count$226$i = $run_count$7$i;
        }
       }
       $273 = (($sbi$133$i) + 1)|0;
       $exitcond51$i = ($273|0)==($213|0);
       if ($exitcond51$i) {
        $flag$2$lcssa$i = $flag$8$i$lcssa;$ncoded_fragis$1$lcssa$i6 = $ncoded_fragis$6$i$lcssa;$nuncoded_fragis$1$lcssa$i = $nuncoded_fragis$6$i$lcssa;$run_count$1$lcssa$i = $run_count$7$i$lcssa;$sbi$1$lcssa$i7 = $213;
        break;
       } else {
        $flag$231$i = $flag$8$i$lcssa;$ncoded_fragis$134$i = $ncoded_fragis$6$i$lcssa;$nuncoded_fragis$135$i = $nuncoded_fragis$6$i$lcssa;$run_count$132$i = $run_count$7$i$lcssa;$sbi$133$i = $273;
       }
      }
     } else {
      $flag$2$lcssa$i = $flag$140$i;$ncoded_fragis$1$lcssa$i6 = $ncoded_fragis$044$i;$nuncoded_fragis$1$lcssa$i = $nuncoded_fragis$045$i;$run_count$1$lcssa$i = $run_count$041$i;$sbi$1$lcssa$i7 = $sbi$043$i;
     }
     $274 = (($ncoded_fragis$1$lcssa$i6) - ($ncoded_fragis$044$i))|0;
     $275 = (((($_dec)) + 256|0) + ($pli$042$i<<2)|0);
     HEAP32[$275>>2] = $274;
     $276 = (($pli$042$i) + 1)|0;
     $exitcond52$i = ($276|0)==(3);
     if ($exitcond52$i) {
      $ncoded_fragis$1$lcssa$i6$lcssa = $ncoded_fragis$1$lcssa$i6;
      break;
     } else {
      $flag$140$i = $flag$2$lcssa$i;$ncoded_fragis$044$i = $ncoded_fragis$1$lcssa$i6;$nsbs$046$i = $213;$nuncoded_fragis$045$i = $nuncoded_fragis$1$lcssa$i;$pli$042$i = $276;$run_count$041$i = $run_count$1$lcssa$i;$sbi$043$i = $sbi$1$lcssa$i7;
     }
    }
    $277 = ((($_dec)) + 268|0);
    HEAP32[$277>>2] = $ncoded_fragis$1$lcssa$i6$lcssa;
    $278 = (_oc_pack_read($5,3)|0);
    $279 = ($278|0)==(0);
    if ($279) {
     $280 = $scheme0_alphabet$i;
     $281 = $280;
     HEAP32[$281>>2] = 0;
     $282 = (($280) + 4)|0;
     $283 = $282;
     HEAP32[$283>>2] = 0;
     $284 = (_oc_pack_read($5,3)|0);
     $285 = (($scheme0_alphabet$i) + ($284)|0);
     HEAP8[$285>>0] = 0;
     $286 = (_oc_pack_read($5,3)|0);
     $287 = (($scheme0_alphabet$i) + ($286)|0);
     HEAP8[$287>>0] = 1;
     $288 = (_oc_pack_read($5,3)|0);
     $289 = (($scheme0_alphabet$i) + ($288)|0);
     HEAP8[$289>>0] = 2;
     $290 = (_oc_pack_read($5,3)|0);
     $291 = (($scheme0_alphabet$i) + ($290)|0);
     HEAP8[$291>>0] = 3;
     $292 = (_oc_pack_read($5,3)|0);
     $293 = (($scheme0_alphabet$i) + ($292)|0);
     HEAP8[$293>>0] = 4;
     $294 = (_oc_pack_read($5,3)|0);
     $295 = (($scheme0_alphabet$i) + ($294)|0);
     HEAP8[$295>>0] = 5;
     $296 = (_oc_pack_read($5,3)|0);
     $297 = (($scheme0_alphabet$i) + ($296)|0);
     HEAP8[$297>>0] = 6;
     $298 = (_oc_pack_read($5,3)|0);
     $299 = (($scheme0_alphabet$i) + ($298)|0);
     HEAP8[$299>>0] = 7;
     $alphabet$0$i = $scheme0_alphabet$i;
    } else {
     $300 = (($278) + -1)|0;
     $301 = (529930 + ($300<<3)|0);
     $alphabet$0$i = $301;
    }
    $302 = ($278|0)==(7);
    $oc_clc_mode_unpack$oc_vlc_mode_unpack$i = $302 ? 4330 : 4331;
    $303 = ((($_dec)) + 236|0);
    $304 = HEAP32[$303>>2]|0;
    $305 = ((($_dec)) + 232|0);
    $306 = HEAP32[$305>>2]|0;
    $307 = ((($_dec)) + 248|0);
    $308 = HEAP32[$307>>2]|0;
    $309 = HEAP32[$205>>2]|0;
    $310 = ($308|0)==(0);
    if (!($310)) {
     $mbi$05$i = 0;
     while(1) {
      $311 = (($304) + ($mbi$05$i)|0);
      $312 = HEAP8[$311>>0]|0;
      $313 = ($312<<24>>24)==(-1);
      do {
       if (!($313)) {
        $314 = (($306) + (($mbi$05$i*48)|0)|0);
        $315 = HEAP32[$314>>2]|0;
        $316 = (($309) + ($315<<2)|0);
        $317 = HEAP32[$316>>2]|0;
        $318 = $317 & 1;
        $319 = ($318|0)==(0);
        if ($319) {
         $320 = (((($306) + (($mbi$05$i*48)|0)|0)) + 4|0);
         $321 = HEAP32[$320>>2]|0;
         $322 = (($309) + ($321<<2)|0);
         $323 = HEAP32[$322>>2]|0;
         $324 = $323 & 1;
         $325 = ($324|0)==(0);
         if ($325) {
          $330 = (((($306) + (($mbi$05$i*48)|0)|0)) + 8|0);
          $331 = HEAP32[$330>>2]|0;
          $332 = (($309) + ($331<<2)|0);
          $333 = HEAP32[$332>>2]|0;
          $334 = $333 & 1;
          $335 = ($334|0)==(0);
          if ($335) {
           $336 = (((($306) + (($mbi$05$i*48)|0)|0)) + 12|0);
           $337 = HEAP32[$336>>2]|0;
           $338 = (($309) + ($337<<2)|0);
           $339 = HEAP32[$338>>2]|0;
           $340 = $339 & 1;
           $341 = ($340|0)==(0);
           if ($341) {
            HEAP8[$311>>0] = 0;
            break;
           }
          }
         }
        }
        $326 = (FUNCTION_TABLE_ii[$oc_clc_mode_unpack$oc_vlc_mode_unpack$i & 8191]($5)|0);
        $327 = (($alphabet$0$i) + ($326)|0);
        $328 = HEAP8[$327>>0]|0;
        HEAP8[$311>>0] = $328;
       }
      } while(0);
      $329 = (($mbi$05$i) + 1)|0;
      $exitcond$i10 = ($329|0)==($308|0);
      if ($exitcond$i10) {
       break;
      } else {
       $mbi$05$i = $329;
      }
     }
    }
    $342 = ((($_dec)) + 48|0);
    $343 = HEAP32[$342>>2]|0;
    $344 = (47552 + ($343<<2)|0);
    $345 = HEAP32[$344>>2]|0;
    $346 = (_oc_pack_read1($5)|0);
    $347 = ($346|0)!=(0);
    $348 = $347 ? 4332 : 4333;
    $349 = HEAP32[$342>>2]|0;
    $350 = (530254 + ($349)|0);
    $351 = HEAP8[$350>>0]|0;
    $352 = $351&255;
    $353 = HEAP32[$205>>2]|0;
    $354 = ((($_dec)) + 212|0);
    $355 = HEAP32[$354>>2]|0;
    $356 = HEAP32[$305>>2]|0;
    $357 = HEAP32[$303>>2]|0;
    $358 = HEAP32[$307>>2]|0;
    $359 = ($358|0)==(0);
    if (!($359)) {
     $360 = ((($lbmvs$i)) + 1|0);
     $$phi$trans$insert$i = ((($coded$i)) + 4|0);
     $361 = ((($lbmvs$i)) + 2|0);
     $362 = ((($lbmvs$i)) + 3|0);
     $363 = ((($lbmvs$i)) + 4|0);
     $364 = ((($lbmvs$i)) + 5|0);
     $365 = ((($lbmvs$i)) + 6|0);
     $366 = ((($lbmvs$i)) + 7|0);
     $367 = ((($lbmvs$i)) + 6|0);
     $368 = ((($lbmvs$i)) + 4|0);
     $369 = ((($lbmvs$i)) + 2|0);
     $last_mv$sroa$0$sroa$0$017$i = 0;$last_mv$sroa$0$sroa$8$016$i = 0;$last_mv$sroa$11$sroa$0$09$i = 0;$last_mv$sroa$11$sroa$5$08$i = 0;$mbi$010$i = 0;$mbmv$sroa$0$018$i = 0;$mbmv$sroa$7$019$i = 0;
     while(1) {
      $370 = (($357) + ($mbi$010$i)|0);
      $371 = HEAP8[$370>>0]|0;
      $372 = $371 << 24 >> 24;
      $373 = ($371<<24>>24)==(-1);
      if ($373) {
       $last_mv$sroa$0$sroa$0$4$i = $last_mv$sroa$0$sroa$0$017$i;$last_mv$sroa$0$sroa$8$4$i = $last_mv$sroa$0$sroa$8$016$i;$last_mv$sroa$11$sroa$0$4$i = $last_mv$sroa$11$sroa$0$09$i;$last_mv$sroa$11$sroa$5$4$i = $last_mv$sroa$11$sroa$5$08$i;$mbmv$sroa$0$3$i = $mbmv$sroa$0$018$i;$mbmv$sroa$7$3$i = $mbmv$sroa$7$019$i;
      } else {
       $mapii$0$i = 0;$ncoded$0$i = 0;
       while(1) {
        $374 = ((530206 + (($349*12)|0)|0) + ($mapii$0$i)|0);
        $375 = HEAP8[$374>>0]|0;
        $376 = $375&255;
        $377 = $376 & 3;
        $378 = $376 >>> 2;
        $379 = (((($356) + (($mbi$010$i*48)|0)|0) + ($378<<4)|0) + ($377<<2)|0);
        $380 = HEAP32[$379>>2]|0;
        $381 = (($353) + ($380<<2)|0);
        $382 = HEAP32[$381>>2]|0;
        $383 = $382 & 1;
        $384 = ($383|0)==(0);
        if ($384) {
         $ncoded$1$i = $ncoded$0$i;
        } else {
         $385 = (($ncoded$0$i) + 1)|0;
         $386 = (($coded$i) + ($ncoded$0$i<<2)|0);
         HEAP32[$386>>2] = $376;
         $ncoded$1$i = $385;
        }
        $387 = (($mapii$0$i) + 1)|0;
        $388 = ($387|0)<($352|0);
        if ($388) {
         $mapii$0$i = $387;$ncoded$0$i = $ncoded$1$i;
        } else {
         $ncoded$1$i$lcssa = $ncoded$1$i;
         break;
        }
       }
       $389 = ($ncoded$1$i$lcssa|0)<(1);
       if ($389) {
        $last_mv$sroa$0$sroa$0$3$i = $last_mv$sroa$0$sroa$0$017$i;$last_mv$sroa$0$sroa$8$3$i = $last_mv$sroa$0$sroa$8$016$i;$last_mv$sroa$11$sroa$0$3$i = $last_mv$sroa$11$sroa$0$09$i;$last_mv$sroa$11$sroa$5$3$i = $last_mv$sroa$11$sroa$5$08$i;$mbmv$sroa$0$2$i = $mbmv$sroa$0$018$i;$mbmv$sroa$7$2$i = $mbmv$sroa$7$019$i;
       } else {
        L174: do {
         switch ($372|0) {
         case 7:  {
          $390 = (($coded$i) + ($ncoded$1$i$lcssa<<2)|0);
          HEAP32[$390>>2] = -1;
          $391 = HEAP32[$coded$i>>2]|0;
          $392 = ($391|0)==(0);
          if ($392) {
           $393 = (($356) + (($mbi$010$i*48)|0)|0);
           $394 = HEAP32[$393>>2]|0;
           $395 = (($353) + ($394<<2)|0);
           $396 = HEAP32[$395>>2]|0;
           $397 = $396 | 1792;
           HEAP32[$395>>2] = $397;
           $398 = (FUNCTION_TABLE_ii[$348 & 8191]($5)|0);
           $399 = $398&255;
           HEAP8[$lbmvs$i>>0] = $399;
           $400 = (FUNCTION_TABLE_ii[$348 & 8191]($5)|0);
           $401 = $400&255;
           HEAP8[$360>>0] = $401;
           $402 = (($355) + ($394<<1)|0);
           $403 = HEAP16[$lbmvs$i>>1]|0;
           HEAP8[$402>>0]=$403&255;HEAP8[$402+1>>0]=$403>>8;
           $$pre$i12 = HEAP32[$$phi$trans$insert$i>>2]|0;
           $405 = $$pre$i12;$codedi$1$i = 1;
          } else {
           HEAP8[$360>>0] = 0;
           HEAP8[$lbmvs$i>>0] = 0;
           $405 = $391;$codedi$1$i = 0;
          }
          $404 = ($405|0)==(1);
          if ($404) {
           $453 = (($codedi$1$i) + 1)|0;
           $454 = (((($356) + (($mbi$010$i*48)|0)|0)) + 4|0);
           $455 = HEAP32[$454>>2]|0;
           $456 = (($353) + ($455<<2)|0);
           $457 = HEAP32[$456>>2]|0;
           $458 = $457 | 1792;
           HEAP32[$456>>2] = $458;
           $459 = (FUNCTION_TABLE_ii[$348 & 8191]($5)|0);
           $460 = $459&255;
           HEAP8[$361>>0] = $460;
           $461 = (FUNCTION_TABLE_ii[$348 & 8191]($5)|0);
           $462 = $461&255;
           HEAP8[$362>>0] = $462;
           $463 = (($355) + ($455<<1)|0);
           $464 = HEAP16[$361>>1]|0;
           HEAP8[$463>>0]=$464&255;HEAP8[$463+1>>0]=$464>>8;
           $$phi$trans$insert23$i = (($coded$i) + ($453<<2)|0);
           $$pre24$i = HEAP32[$$phi$trans$insert23$i>>2]|0;
           $466 = $$pre24$i;$codedi$1$1$i = $453;
          } else {
           HEAP8[$362>>0] = 0;
           HEAP8[$369>>0] = 0;
           $466 = $405;$codedi$1$1$i = $codedi$1$i;
          }
          $465 = ($466|0)==(2);
          if ($465) {
           $467 = (($codedi$1$1$i) + 1)|0;
           $468 = (((($356) + (($mbi$010$i*48)|0)|0)) + 8|0);
           $469 = HEAP32[$468>>2]|0;
           $470 = (($353) + ($469<<2)|0);
           $471 = HEAP32[$470>>2]|0;
           $472 = $471 | 1792;
           HEAP32[$470>>2] = $472;
           $473 = (FUNCTION_TABLE_ii[$348 & 8191]($5)|0);
           $474 = $473&255;
           HEAP8[$363>>0] = $474;
           $475 = (FUNCTION_TABLE_ii[$348 & 8191]($5)|0);
           $476 = $475&255;
           HEAP8[$364>>0] = $476;
           $477 = (($355) + ($469<<1)|0);
           $478 = HEAP16[$363>>1]|0;
           HEAP8[$477>>0]=$478&255;HEAP8[$477+1>>0]=$478>>8;
           $$phi$trans$insert25$i = (($coded$i) + ($467<<2)|0);
           $$pre26$i = HEAP32[$$phi$trans$insert25$i>>2]|0;
           $480 = $$pre26$i;$codedi$1$2$i = $467;
          } else {
           HEAP8[$364>>0] = 0;
           HEAP8[$368>>0] = 0;
           $480 = $466;$codedi$1$2$i = $codedi$1$1$i;
          }
          $479 = ($480|0)==(3);
          if ($479) {
           $481 = (($codedi$1$2$i) + 1)|0;
           $482 = (((($356) + (($mbi$010$i*48)|0)|0)) + 12|0);
           $483 = HEAP32[$482>>2]|0;
           $484 = (($353) + ($483<<2)|0);
           $485 = HEAP32[$484>>2]|0;
           $486 = $485 | 1792;
           HEAP32[$484>>2] = $486;
           $487 = (FUNCTION_TABLE_ii[$348 & 8191]($5)|0);
           $488 = $487&255;
           HEAP8[$365>>0] = $488;
           $489 = (FUNCTION_TABLE_ii[$348 & 8191]($5)|0);
           $490 = $489&255;
           HEAP8[$366>>0] = $490;
           $491 = (($355) + ($483<<1)|0);
           $492 = HEAP16[$365>>1]|0;
           HEAP8[$491>>0]=$492&255;HEAP8[$491+1>>0]=$492>>8;
           $codedi$1$329$i = $481;
           label = 112;
          } else {
           HEAP8[$366>>0] = 0;
           HEAP8[$367>>0] = 0;
           $493 = ($codedi$1$2$i|0)>(0);
           if ($493) {
            $codedi$1$329$i = $codedi$1$2$i;
            label = 112;
           } else {
            $codedi$1$328$i = 0;$last_mv$sroa$0$sroa$0$1$i = $last_mv$sroa$0$sroa$0$017$i;$last_mv$sroa$0$sroa$8$1$i = $last_mv$sroa$0$sroa$8$016$i;$last_mv$sroa$11$sroa$0$1$i = $last_mv$sroa$11$sroa$0$09$i;$last_mv$sroa$11$sroa$5$1$i = $last_mv$sroa$11$sroa$5$08$i;
           }
          }
          if ((label|0) == 112) {
           label = 0;
           $406 = (($codedi$1$329$i) + -1)|0;
           $407 = (($coded$i) + ($406<<2)|0);
           $408 = HEAP32[$407>>2]|0;
           $409 = (($lbmvs$i) + ($408<<1)|0);
           $410 = HEAP8[$409>>0]|0;
           $411 = (((($lbmvs$i) + ($408<<1)|0)) + 1|0);
           $412 = HEAP8[$411>>0]|0;
           $codedi$1$328$i = $codedi$1$329$i;$last_mv$sroa$0$sroa$0$1$i = $410;$last_mv$sroa$0$sroa$8$1$i = $412;$last_mv$sroa$11$sroa$0$1$i = $last_mv$sroa$0$sroa$0$017$i;$last_mv$sroa$11$sroa$5$1$i = $last_mv$sroa$0$sroa$8$016$i;
          }
          $413 = ($codedi$1$328$i|0)<($ncoded$1$i$lcssa|0);
          if (!($413)) {
           $last_mv$sroa$0$sroa$0$2$i = $last_mv$sroa$0$sroa$0$1$i;$last_mv$sroa$0$sroa$8$2$i = $last_mv$sroa$0$sroa$8$1$i;$last_mv$sroa$11$sroa$0$2$i = $last_mv$sroa$11$sroa$0$1$i;$last_mv$sroa$11$sroa$5$2$i = $last_mv$sroa$11$sroa$5$1$i;$mbmv$sroa$0$1$i = $mbmv$sroa$0$018$i;$mbmv$sroa$7$1$i = $mbmv$sroa$7$019$i;
           break L174;
          }
          FUNCTION_TABLE_vii[$345 & 8191]($cbmvs$i,$lbmvs$i);
          $codedi$25$i = $codedi$1$328$i;
          while(1) {
           $414 = (($coded$i) + ($codedi$25$i<<2)|0);
           $415 = HEAP32[$414>>2]|0;
           $416 = $415 & 3;
           $417 = $415 >> 2;
           $418 = (((($356) + (($mbi$010$i*48)|0)|0) + ($417<<4)|0) + ($416<<2)|0);
           $419 = HEAP32[$418>>2]|0;
           $420 = (($353) + ($419<<2)|0);
           $421 = HEAP32[$420>>2]|0;
           $422 = $421 | 1792;
           HEAP32[$420>>2] = $422;
           $423 = (($355) + ($419<<1)|0);
           $424 = (($cbmvs$i) + ($416<<1)|0);
           $425 = HEAPU8[$424>>0]|(HEAPU8[$424+1>>0]<<8);
           HEAP8[$423>>0]=$425&255;HEAP8[$423+1>>0]=$425>>8;
           $426 = (($codedi$25$i) + 1)|0;
           $exitcond$i14 = ($426|0)==($ncoded$1$i$lcssa|0);
           if ($exitcond$i14) {
            $last_mv$sroa$0$sroa$0$2$i = $last_mv$sroa$0$sroa$0$1$i;$last_mv$sroa$0$sroa$8$2$i = $last_mv$sroa$0$sroa$8$1$i;$last_mv$sroa$11$sroa$0$2$i = $last_mv$sroa$11$sroa$0$1$i;$last_mv$sroa$11$sroa$5$2$i = $last_mv$sroa$11$sroa$5$1$i;$mbmv$sroa$0$1$i = $mbmv$sroa$0$018$i;$mbmv$sroa$7$1$i = $mbmv$sroa$7$019$i;
            break;
           } else {
            $codedi$25$i = $426;
           }
          }
          break;
         }
         case 2:  {
          $427 = (FUNCTION_TABLE_ii[$348 & 8191]($5)|0);
          $428 = $427&255;
          $429 = (FUNCTION_TABLE_ii[$348 & 8191]($5)|0);
          $430 = $429&255;
          $last_mv$sroa$0$sroa$0$2$i = $428;$last_mv$sroa$0$sroa$8$2$i = $430;$last_mv$sroa$11$sroa$0$2$i = $last_mv$sroa$0$sroa$0$017$i;$last_mv$sroa$11$sroa$5$2$i = $last_mv$sroa$0$sroa$8$016$i;$mbmv$sroa$0$1$i = $428;$mbmv$sroa$7$1$i = $430;
          break;
         }
         case 3:  {
          $last_mv$sroa$0$sroa$0$2$i = $last_mv$sroa$0$sroa$0$017$i;$last_mv$sroa$0$sroa$8$2$i = $last_mv$sroa$0$sroa$8$016$i;$last_mv$sroa$11$sroa$0$2$i = $last_mv$sroa$11$sroa$0$09$i;$last_mv$sroa$11$sroa$5$2$i = $last_mv$sroa$11$sroa$5$08$i;$mbmv$sroa$0$1$i = $last_mv$sroa$0$sroa$0$017$i;$mbmv$sroa$7$1$i = $last_mv$sroa$0$sroa$8$016$i;
          break;
         }
         case 4:  {
          $last_mv$sroa$0$sroa$0$2$i = $last_mv$sroa$11$sroa$0$09$i;$last_mv$sroa$0$sroa$8$2$i = $last_mv$sroa$11$sroa$5$08$i;$last_mv$sroa$11$sroa$0$2$i = $last_mv$sroa$0$sroa$0$017$i;$last_mv$sroa$11$sroa$5$2$i = $last_mv$sroa$0$sroa$8$016$i;$mbmv$sroa$0$1$i = $last_mv$sroa$11$sroa$0$09$i;$mbmv$sroa$7$1$i = $last_mv$sroa$11$sroa$5$08$i;
          break;
         }
         case 6:  {
          $431 = (FUNCTION_TABLE_ii[$348 & 8191]($5)|0);
          $432 = $431&255;
          $433 = (FUNCTION_TABLE_ii[$348 & 8191]($5)|0);
          $434 = $433&255;
          $last_mv$sroa$0$sroa$0$2$i = $last_mv$sroa$0$sroa$0$017$i;$last_mv$sroa$0$sroa$8$2$i = $last_mv$sroa$0$sroa$8$016$i;$last_mv$sroa$11$sroa$0$2$i = $last_mv$sroa$11$sroa$0$09$i;$last_mv$sroa$11$sroa$5$2$i = $last_mv$sroa$11$sroa$5$08$i;$mbmv$sroa$0$1$i = $432;$mbmv$sroa$7$1$i = $434;
          break;
         }
         default: {
          $last_mv$sroa$0$sroa$0$2$i = $last_mv$sroa$0$sroa$0$017$i;$last_mv$sroa$0$sroa$8$2$i = $last_mv$sroa$0$sroa$8$016$i;$last_mv$sroa$11$sroa$0$2$i = $last_mv$sroa$11$sroa$0$09$i;$last_mv$sroa$11$sroa$5$2$i = $last_mv$sroa$11$sroa$5$08$i;$mbmv$sroa$0$1$i = 0;$mbmv$sroa$7$1$i = 0;
         }
         }
        } while(0);
        $435 = ($371<<24>>24)!=(7);
        $436 = ($ncoded$1$i$lcssa|0)>(0);
        $or$cond$i = $435 & $436;
        if ($or$cond$i) {
         $437 = $372 << 8;
         $438 = $437 & 1792;
         $codedi$36$i = 0;
         while(1) {
          $439 = (($coded$i) + ($codedi$36$i<<2)|0);
          $440 = HEAP32[$439>>2]|0;
          $441 = $440 & 3;
          $442 = $440 >> 2;
          $443 = (((($356) + (($mbi$010$i*48)|0)|0) + ($442<<4)|0) + ($441<<2)|0);
          $444 = HEAP32[$443>>2]|0;
          $445 = (($353) + ($444<<2)|0);
          $446 = HEAP32[$445>>2]|0;
          $447 = $446 & -1793;
          $448 = $447 | $438;
          HEAP32[$445>>2] = $448;
          $449 = (($355) + ($444<<1)|0);
          HEAP8[$449>>0] = $mbmv$sroa$0$1$i;
          $450 = (((($355) + ($444<<1)|0)) + 1|0);
          HEAP8[$450>>0] = $mbmv$sroa$7$1$i;
          $451 = (($codedi$36$i) + 1)|0;
          $exitcond21$i = ($451|0)==($ncoded$1$i$lcssa|0);
          if ($exitcond21$i) {
           $last_mv$sroa$0$sroa$0$3$i = $last_mv$sroa$0$sroa$0$2$i;$last_mv$sroa$0$sroa$8$3$i = $last_mv$sroa$0$sroa$8$2$i;$last_mv$sroa$11$sroa$0$3$i = $last_mv$sroa$11$sroa$0$2$i;$last_mv$sroa$11$sroa$5$3$i = $last_mv$sroa$11$sroa$5$2$i;$mbmv$sroa$0$2$i = $mbmv$sroa$0$1$i;$mbmv$sroa$7$2$i = $mbmv$sroa$7$1$i;
           break;
          } else {
           $codedi$36$i = $451;
          }
         }
        } else {
         $last_mv$sroa$0$sroa$0$3$i = $last_mv$sroa$0$sroa$0$2$i;$last_mv$sroa$0$sroa$8$3$i = $last_mv$sroa$0$sroa$8$2$i;$last_mv$sroa$11$sroa$0$3$i = $last_mv$sroa$11$sroa$0$2$i;$last_mv$sroa$11$sroa$5$3$i = $last_mv$sroa$11$sroa$5$2$i;$mbmv$sroa$0$2$i = $mbmv$sroa$0$1$i;$mbmv$sroa$7$2$i = $mbmv$sroa$7$1$i;
        }
       }
       $last_mv$sroa$0$sroa$0$4$i = $last_mv$sroa$0$sroa$0$3$i;$last_mv$sroa$0$sroa$8$4$i = $last_mv$sroa$0$sroa$8$3$i;$last_mv$sroa$11$sroa$0$4$i = $last_mv$sroa$11$sroa$0$3$i;$last_mv$sroa$11$sroa$5$4$i = $last_mv$sroa$11$sroa$5$3$i;$mbmv$sroa$0$3$i = $mbmv$sroa$0$2$i;$mbmv$sroa$7$3$i = $mbmv$sroa$7$2$i;
      }
      $452 = (($mbi$010$i) + 1)|0;
      $exitcond22$i = ($452|0)==($358|0);
      if ($exitcond22$i) {
       break;
      } else {
       $last_mv$sroa$0$sroa$0$017$i = $last_mv$sroa$0$sroa$0$4$i;$last_mv$sroa$0$sroa$8$016$i = $last_mv$sroa$0$sroa$8$4$i;$last_mv$sroa$11$sroa$0$09$i = $last_mv$sroa$11$sroa$0$4$i;$last_mv$sroa$11$sroa$5$08$i = $last_mv$sroa$11$sroa$5$4$i;$mbi$010$i = $452;$mbmv$sroa$0$018$i = $mbmv$sroa$0$3$i;$mbmv$sroa$7$019$i = $mbmv$sroa$7$3$i;
      }
     }
    }
    $$pre147 = HEAP32[$277>>2]|0;
    $495 = $$pre147;
   }
   $494 = ($495|0)<(1);
   L209: do {
    if (!($494)) {
     $496 = ((($_dec)) + 204|0);
     $497 = HEAP32[$496>>2]|0;
     $498 = ((($_dec)) + 252|0);
     $499 = HEAP32[$498>>2]|0;
     $500 = ((($_dec)) + 794|0);
     $501 = HEAP8[$500>>0]|0;
     $502 = ($501<<24>>24)==(1);
     if ($502) {
      $fragii$015$i = 0;
      while(1) {
       $503 = (($499) + ($fragii$015$i<<2)|0);
       $504 = HEAP32[$503>>2]|0;
       $505 = (($497) + ($504<<2)|0);
       $506 = HEAP32[$505>>2]|0;
       $507 = $506 & -253;
       HEAP32[$505>>2] = $507;
       $508 = (($fragii$015$i) + 1)|0;
       $exitcond$i18 = ($508|0)==($495|0);
       if ($exitcond$i18) {
        break L209;
       } else {
        $fragii$015$i = $508;
       }
      }
     }
     $509 = (_oc_pack_read1($5)|0);
     $flag$022$i = $509;$fragii$124$i = 0;$nqi1$023$i = 0;
     while(1) {
      $510 = (_oc_sb_run_unpack($5)|0);
      $511 = ($510|0)>(4128);
      $512 = $flag$022$i << 2;
      $513 = $512 & 252;
      $fragii$2$i = $fragii$124$i;$nqi1$1$i = $nqi1$023$i;$run_count$0$i = $510;
      while(1) {
       $514 = (($fragii$2$i) + 1)|0;
       $515 = (($499) + ($fragii$2$i<<2)|0);
       $516 = HEAP32[$515>>2]|0;
       $517 = (($497) + ($516<<2)|0);
       $518 = HEAP32[$517>>2]|0;
       $519 = $518 & -253;
       $520 = $519 | $513;
       HEAP32[$517>>2] = $520;
       $521 = (($nqi1$1$i) + ($flag$022$i))|0;
       $522 = ($run_count$0$i|0)>(1);
       if (!($522)) {
        $$lcssa265 = $514;$$lcssa266 = $521;
        break;
       }
       $523 = (($run_count$0$i) + -1)|0;
       $524 = ($514|0)<($495|0);
       if ($524) {
        $fragii$2$i = $514;$nqi1$1$i = $521;$run_count$0$i = $523;
       } else {
        $$lcssa265 = $514;$$lcssa266 = $521;
        break;
       }
      }
      $525 = ($$lcssa265|0)<($495|0);
      $or$cond2$i = $511 & $525;
      if ($or$cond2$i) {
       $526 = (_oc_pack_read1($5)|0);
       $flag$0$be$i = $526;
      } else {
       $527 = ($flag$022$i|0)==(0);
       $528 = $527&1;
       $flag$0$be$i = $528;
      }
      if ($525) {
       $flag$022$i = $flag$0$be$i;$fragii$124$i = $$lcssa265;$nqi1$023$i = $$lcssa266;
      } else {
       $$lcssa266$lcssa = $$lcssa266;
       break;
      }
     }
     $phitmp$i = ($$lcssa266$lcssa|0)>(0);
     $529 = HEAP8[$500>>0]|0;
     $530 = ($529<<24>>24)==(3);
     $or$cond$i21 = $phitmp$i & $530;
     if ($or$cond$i21) {
      $fragii$3$i = 0;
      while(1) {
       $531 = (($499) + ($fragii$3$i<<2)|0);
       $532 = HEAP32[$531>>2]|0;
       $533 = (($497) + ($532<<2)|0);
       $534 = HEAP32[$533>>2]|0;
       $535 = $534 & 252;
       $536 = ($535|0)==(0);
       $537 = (($fragii$3$i) + 1)|0;
       if ($536) {
        $fragii$3$i = $537;
       } else {
        $fragii$3$i$lcssa = $fragii$3$i;
        break;
       }
      }
      $538 = (_oc_pack_read1($5)|0);
      $539 = (_oc_sb_run_unpack($5)|0);
      $540 = ($fragii$3$i$lcssa|0)<($495|0);
      if ($540) {
       $557 = $539;$flag$120$i = $538;$fragii$421$i = $fragii$3$i$lcssa;
       while(1) {
        $fragii$517$i = $fragii$421$i;$run_count$116$i = $557;
        while(1) {
         $541 = (($499) + ($fragii$517$i<<2)|0);
         $542 = HEAP32[$541>>2]|0;
         $543 = (($497) + ($542<<2)|0);
         $544 = HEAP32[$543>>2]|0;
         $545 = $544 >>> 2;
         $546 = $545 & 63;
         $547 = ($546|0)==(0);
         if ($547) {
          $run_count$2$i = $run_count$116$i;
         } else {
          $548 = ($run_count$116$i|0)<(1);
          if ($548) {
           $fragii$517$i$lcssa = $fragii$517$i;
           break;
          }
          $549 = (($run_count$116$i) + -1)|0;
          $550 = (($545) + ($flag$120$i))|0;
          $551 = $550 << 2;
          $552 = $551 & 252;
          $553 = $544 & -253;
          $554 = $552 | $553;
          HEAP32[$543>>2] = $554;
          $run_count$2$i = $549;
         }
         $555 = (($fragii$517$i) + 1)|0;
         $556 = ($555|0)<($495|0);
         if ($556) {
          $fragii$517$i = $555;$run_count$116$i = $run_count$2$i;
         } else {
          break L209;
         }
        }
        $$not$i = ($557|0)<(4129);
        if ($$not$i) {
         $558 = ($flag$120$i|0)==(0);
         $559 = $558&1;
         $flag$1$be$i = $559;
        } else {
         $560 = (_oc_pack_read1($5)|0);
         $flag$1$be$i = $560;
        }
        $561 = (_oc_sb_run_unpack($5)|0);
        $562 = ($fragii$517$i$lcssa|0)<($495|0);
        if ($562) {
         $557 = $561;$flag$120$i = $flag$1$be$i;$fragii$421$i = $fragii$517$i$lcssa;
        } else {
         break;
        }
       }
      }
     }
    }
   } while(0);
   $563 = ((($_dec)) + 256|0);
   $564 = HEAP32[$563>>2]|0;
   $zzi$027$i = 0;
   while(1) {
    $565 = (($ntoks_left$i) + ($zzi$027$i<<2)|0);
    HEAP32[$565>>2] = $564;
    $566 = (($zzi$027$i) + 1)|0;
    $exitcond40$i = ($566|0)==(64);
    if ($exitcond40$i) {
     break;
    } else {
     $zzi$027$i = $566;
    }
   }
   $567 = ((($_dec)) + 260|0);
   $568 = HEAP32[$567>>2]|0;
   $zzi$027$1$i = 0;
   while(1) {
    $852 = (((($ntoks_left$i)) + 256|0) + ($zzi$027$1$i<<2)|0);
    HEAP32[$852>>2] = $568;
    $853 = (($zzi$027$1$i) + 1)|0;
    $exitcond40$1$i = ($853|0)==(64);
    if ($exitcond40$1$i) {
     break;
    } else {
     $zzi$027$1$i = $853;
    }
   }
   $854 = ((($_dec)) + 264|0);
   $855 = HEAP32[$854>>2]|0;
   $zzi$027$2$i = 0;
   while(1) {
    $856 = (((($ntoks_left$i)) + 512|0) + ($zzi$027$2$i<<2)|0);
    HEAP32[$856>>2] = $855;
    $857 = (($zzi$027$2$i) + 1)|0;
    $exitcond40$2$i = ($857|0)==(64);
    if ($exitcond40$2$i) {
     break;
    } else {
     $zzi$027$2$i = $857;
    }
   }
   $858 = (_oc_pack_read($5,4)|0);
   HEAP32[$huff_idxs$i>>2] = $858;
   $859 = (_oc_pack_read($5,4)|0);
   $701 = ((($huff_idxs$i)) + 4|0);
   HEAP32[$701>>2] = $859;
   $860 = ((($_dec)) + 52664|0);
   HEAP32[$860>>2] = 0;
   $712 = ((($_dec)) + 53432|0);
   $615 = HEAP32[$712>>2]|0;
   $861 = ((($_dec)) + 204|0);
   $601 = HEAP32[$861>>2]|0;
   $862 = ((($_dec)) + 252|0);
   $598 = HEAP32[$862>>2]|0;
   $680 = ((($run_counts$i$i)) + 252|0);
   $eobs$035$i$i = 0;$fragii$036$i$i = 0;$ncoded_fragis$037$i$i = 0;$pli$034$i$i = 0;$ti$033$i$i = 0;
   while(1) {
    $569 = (((($_dec)) + 256|0) + ($pli$034$i$i<<2)|0);
    $570 = HEAP32[$569>>2]|0;
    $571 = (($570) + ($ncoded_fragis$037$i$i))|0;
    _memset(($run_counts$i$i|0),0,256)|0;
    $572 = (((($_dec)) + 52664|0) + ($pli$034$i$i<<8)|0);
    HEAP32[$572>>2] = $eobs$035$i$i;
    $573 = (((($_dec)) + 51896|0) + ($pli$034$i$i<<8)|0);
    HEAP32[$573>>2] = $ti$033$i$i;
    $574 = (($571) - ($fragii$036$i$i))|0;
    $575 = ($574|0)<($eobs$035$i$i|0);
    $$eobs$0$i$i = $575 ? $574 : $eobs$035$i$i;
    $576 = (($eobs$035$i$i) - ($$eobs$0$i$i))|0;
    $577 = ($$eobs$0$i$i|0)>(0);
    if ($577) {
     $578 = $ncoded_fragis$037$i$i ^ -1;
     $579 = (($fragii$036$i$i) + ($578))|0;
     $580 = (($579) - ($570))|0;
     $581 = $eobs$035$i$i ^ -1;
     $582 = ($580|0)>($581|0);
     $smax$i$i = $582 ? $580 : $581;
     $eobi$17$i$i = $$eobs$0$i$i;$fragii$18$i$i = $fragii$036$i$i;
     while(1) {
      $595 = (($eobi$17$i$i) + -1)|0;
      $596 = (($fragii$18$i$i) + 1)|0;
      $597 = (($598) + ($fragii$18$i$i<<2)|0);
      $599 = HEAP32[$597>>2]|0;
      $600 = (($601) + ($599<<2)|0);
      $602 = HEAP32[$600>>2]|0;
      $603 = $602 & 65535;
      HEAP32[$600>>2] = $603;
      $604 = ($eobi$17$i$i|0)>(1);
      if ($604) {
       $eobi$17$i$i = $595;$fragii$18$i$i = $596;
      } else {
       break;
      }
     }
     $583 = ($smax$i$i|0)>(-2);
     $smax46$i$i = $583 ? $smax$i$i : -2;
     $584 = (($fragii$036$i$i) + 1)|0;
     $585 = (($584) - ($smax$i$i))|0;
     $586 = (($585) + ($smax46$i$i))|0;
     $fragii$1$lcssa$i$i = $586;
    } else {
     $fragii$1$lcssa$i$i = $fragii$036$i$i;
    }
    $587 = ($571|0)>($fragii$1$lcssa$i$i|0);
    L260: do {
     if ($587) {
      $588 = (($pli$034$i$i) + 1)|0;
      $589 = $588 >> 1;
      $590 = (($huff_idxs$i) + ($589<<2)|0);
      $591 = $ncoded_fragis$037$i$i ^ -1;
      $592 = (($591) - ($570))|0;
      $593 = HEAP32[$590>>2]|0;
      $594 = (((($_dec)) + 51576|0) + ($593<<2)|0);
      $eob_count$0$ph26$i$i = $$eobs$0$i$i;$fragii$2$ph27$i$i = $fragii$1$lcssa$i$i;$ti$1$ph25$i$i = $ti$033$i$i;
      while(1) {
       $fragii$210$i$i = $fragii$2$ph27$i$i;$ti$19$i$i = $ti$1$ph25$i$i;
       while(1) {
        $610 = HEAP32[$594>>2]|0;
        $611 = (_oc_huff_token_decode($5,$610)|0);
        $612 = $611&255;
        $613 = (($ti$19$i$i) + 1)|0;
        $614 = (($615) + ($ti$19$i$i)|0);
        HEAP8[$614>>0] = $612;
        $616 = ($611>>>0)<(15);
        if ($616) {
         $617 = (529986 + ($611)|0);
         $618 = HEAP8[$617>>0]|0;
         $619 = $618&255;
         $620 = (_oc_pack_read($5,$619)|0);
         $621 = $620&255;
         $622 = (($ti$19$i$i) + 2)|0;
         $623 = (($615) + ($613)|0);
         HEAP8[$623>>0] = $621;
         $624 = ($611|0)==(0);
         if ($624) {
          $625 = $620 >>> 8;
          $626 = $625&255;
          $627 = (($ti$19$i$i) + 3)|0;
          $628 = (($615) + ($622)|0);
          HEAP8[$628>>0] = $626;
          $ti$2$i$i = $627;
         } else {
          $ti$2$i$i = $622;
         }
         $629 = ($611|0)<(2);
         $630 = $629 ? -13 : 0;
         $631 = ($611|0)<(12);
         $632 = $631 ? 21 : 0;
         $633 = (($630) + ($632))|0;
         $634 = $620 << $633;
         $eb$0$i$i = $634;$ti$3$i$i = $ti$2$i$i;
        } else {
         $eb$0$i$i = 0;$ti$3$i$i = $613;
        }
        $635 = (47184 + ($611<<2)|0);
        $636 = HEAP32[$635>>2]|0;
        $637 = (($636) + ($eb$0$i$i))|0;
        $638 = $637 >>> 8;
        $639 = $638 & 4095;
        $640 = ($637|0)==(0);
        $$$i$i = $640 ? 2147483647 : $639;
        $641 = ($$$i$i|0)==(0);
        if (!($641)) {
         $$$i$i$lcssa = $$$i$i;$fragii$210$i$i$lcssa = $fragii$210$i$i;$ti$3$i$i$lcssa = $ti$3$i$i;
         break;
        }
        $660 = $637 & 255;
        $661 = $637 & 1048576;
        $662 = (0 - ($661))|0;
        $663 = $637 ^ $662;
        $664 = $663 >> 21;
        $665 = ($660|0)!=(0);
        $666 = (($run_counts$i$i) + ($660<<2)|0);
        $667 = HEAP32[$666>>2]|0;
        $668 = (($667) + 1)|0;
        HEAP32[$666>>2] = $668;
        $669 = (($fragii$210$i$i) + 1)|0;
        $670 = (($598) + ($fragii$210$i$i<<2)|0);
        $671 = HEAP32[$670>>2]|0;
        $672 = (($601) + ($671<<2)|0);
        $673 = HEAP32[$672>>2]|0;
        $$op2$i$i = $664 << 16;
        $674 = $665 ? 0 : $$op2$i$i;
        $675 = $673 & 65535;
        $676 = $675 | $674;
        HEAP32[$672>>2] = $676;
        $677 = ($571|0)>($669|0);
        if ($677) {
         $fragii$210$i$i = $669;$ti$19$i$i = $ti$3$i$i;
        } else {
         $eob_count$0$ph$lcssa$i$i = $eob_count$0$ph26$i$i;$eobs$1$lcssa$i$i = 0;$fragii$2$lcssa$i$i = $669;$ti$1$lcssa$i$i = $ti$3$i$i;
         break L260;
        }
       }
       $642 = (($571) - ($fragii$210$i$i$lcssa))|0;
       $643 = (($642) - ($$$i$i$lcssa))|0;
       $644 = ($642|0)<($$$i$i$lcssa|0);
       $645 = $644 ? $643 : 0;
       $646 = (($645) + ($$$i$i$lcssa))|0;
       $647 = (($646) + ($eob_count$0$ph26$i$i))|0;
       $648 = ($646|0)>(0);
       if ($648) {
        $649 = (($fragii$210$i$i$lcssa) + ($592))|0;
        $650 = $$$i$i$lcssa ^ -1;
        $651 = ($649|0)>($650|0);
        $smax48$i$i = $651 ? $649 : $650;
        $eobi$220$i$i = $646;$fragii$321$i$i = $fragii$210$i$i$lcssa;
        while(1) {
         $652 = (($eobi$220$i$i) + -1)|0;
         $653 = (($fragii$321$i$i) + 1)|0;
         $654 = (($598) + ($fragii$321$i$i<<2)|0);
         $655 = HEAP32[$654>>2]|0;
         $656 = (($601) + ($655<<2)|0);
         $657 = HEAP32[$656>>2]|0;
         $658 = $657 & 65535;
         HEAP32[$656>>2] = $658;
         $659 = ($eobi$220$i$i|0)>(1);
         if ($659) {
          $eobi$220$i$i = $652;$fragii$321$i$i = $653;
         } else {
          break;
         }
        }
        $605 = ($smax48$i$i|0)>(-2);
        $smax49$i$i = $605 ? $smax48$i$i : -2;
        $606 = (($fragii$210$i$i$lcssa) + 1)|0;
        $607 = (($606) - ($smax48$i$i))|0;
        $608 = (($607) + ($smax49$i$i))|0;
        $fragii$3$lcssa$i$i = $608;
       } else {
        $fragii$3$lcssa$i$i = $fragii$210$i$i$lcssa;
       }
       $609 = ($571|0)>($fragii$3$lcssa$i$i|0);
       if ($609) {
        $eob_count$0$ph26$i$i = $647;$fragii$2$ph27$i$i = $fragii$3$lcssa$i$i;$ti$1$ph25$i$i = $ti$3$i$i$lcssa;
       } else {
        $$lcssa262 = $645;$$lcssa263 = $647;$fragii$3$lcssa$i$i$lcssa = $fragii$3$lcssa$i$i;$ti$3$i$i$lcssa$lcssa = $ti$3$i$i$lcssa;
        break;
       }
      }
      $678 = (0 - ($$lcssa262))|0;
      $eob_count$0$ph$lcssa$i$i = $$lcssa263;$eobs$1$lcssa$i$i = $678;$fragii$2$lcssa$i$i = $fragii$3$lcssa$i$i$lcssa;$ti$1$lcssa$i$i = $ti$3$i$i$lcssa$lcssa;
     } else {
      $eob_count$0$ph$lcssa$i$i = $$eobs$0$i$i;$eobs$1$lcssa$i$i = $576;$fragii$2$lcssa$i$i = $fragii$1$lcssa$i$i;$ti$1$lcssa$i$i = $ti$033$i$i;
     }
    } while(0);
    $679 = HEAP32[$680>>2]|0;
    $681 = (($679) + ($eob_count$0$ph$lcssa$i$i))|0;
    HEAP32[$680>>2] = $681;
    $683 = 62;$686 = $681;
    while(1) {
     $682 = (($run_counts$i$i) + ($683<<2)|0);
     $684 = HEAP32[$682>>2]|0;
     $685 = (($684) + ($686))|0;
     HEAP32[$682>>2] = $685;
     $687 = (($683) + -1)|0;
     $688 = ($683|0)>(0);
     if ($688) {
      $683 = $687;$686 = $685;
     } else {
      $690 = 63;
      break;
     }
    }
    while(1) {
     $689 = (($run_counts$i$i) + ($690<<2)|0);
     $691 = HEAP32[$689>>2]|0;
     $692 = ((($ntoks_left$i) + ($pli$034$i$i<<8)|0) + ($690<<2)|0);
     $693 = HEAP32[$692>>2]|0;
     $694 = (($693) - ($691))|0;
     HEAP32[$692>>2] = $694;
     $695 = (($690) + -1)|0;
     $696 = ($690|0)>(0);
     if ($696) {
      $690 = $695;
     } else {
      break;
     }
    }
    $697 = (($pli$034$i$i) + 1)|0;
    $exitcond$i$i25 = ($697|0)==(3);
    if ($exitcond$i$i25) {
     $eobs$1$lcssa$i$i$lcssa = $eobs$1$lcssa$i$i;$ti$1$lcssa$i$i$lcssa = $ti$1$lcssa$i$i;
     break;
    } else {
     $eobs$035$i$i = $eobs$1$lcssa$i$i;$fragii$036$i$i = $fragii$2$lcssa$i$i;$ncoded_fragis$037$i$i = $571;$pli$034$i$i = $697;$ti$033$i$i = $ti$1$lcssa$i$i;
    }
   }
   $698 = ((($_dec)) + 53440|0);
   HEAP32[$698>>2] = $ti$1$lcssa$i$i$lcssa;
   $699 = (_oc_pack_read($5,4)|0);
   HEAP32[$huff_idxs$i>>2] = $699;
   $700 = (_oc_pack_read($5,4)|0);
   HEAP32[$701>>2] = $700;
   $703 = $699;$705 = $700;$eobs$024$i = $eobs$1$lcssa$i$i$lcssa;$hgi$026$i = 1;$zzi$125$i = 1;
   while(1) {
    $702 = (($703) + 16)|0;
    HEAP32[$huff_idxs$i>>2] = $702;
    $704 = (($705) + 16)|0;
    HEAP32[$701>>2] = $704;
    $706 = (530001 + ($hgi$026$i)|0);
    $707 = HEAP8[$706>>0]|0;
    $708 = $707&255;
    $709 = ($zzi$125$i|0)<($708|0);
    if ($709) {
     $710 = (($zzi$125$i) + 1)|0;
     $$pre$i26 = HEAP32[$698>>2]|0;
     $1719 = $$pre$i26;$eobs$120$i = $eobs$024$i;$zzi$221$i = $zzi$125$i;
     while(1) {
      $711 = HEAP32[$712>>2]|0;
      $713 = (64 - ($zzi$221$i))|0;
      $714 = ($713|0)>(0);
      if ($714) {
       $$014$us$i$i = $eobs$120$i;$pli$011$us$i$i = 0;$ti$013$us$i$i = $1719;
       while(1) {
        $715 = ((((($_dec)) + 52664|0) + ($pli$011$us$i$i<<8)|0) + ($zzi$221$i<<2)|0);
        HEAP32[$715>>2] = $$014$us$i$i;
        $716 = ((((($_dec)) + 51896|0) + ($pli$011$us$i$i<<8)|0) + ($zzi$221$i<<2)|0);
        HEAP32[$716>>2] = $ti$013$us$i$i;
        $717 = ((($ntoks_left$i) + ($pli$011$us$i$i<<8)|0) + ($zzi$221$i<<2)|0);
        $718 = HEAP32[$717>>2]|0;
        _memset(($run_counts$i$i|0),0,256)|0;
        $719 = ($$014$us$i$i>>>0)<($718>>>0);
        if ($719) {
         $782 = (($pli$011$us$i$i) + 1)|0;
         $783 = $782 >> 1;
         $784 = (($huff_idxs$i) + ($783<<2)|0);
         $785 = HEAP32[$784>>2]|0;
         $743 = (((($_dec)) + 51576|0) + ($785<<2)|0);
         $$13$us$i$i = $$014$us$i$i;$779 = $$014$us$i$i;$eob_count$01$us$i$i = 0;$ti$12$us$i$i = $ti$013$us$i$i;
         while(1) {
          $741 = (($eob_count$01$us$i$i) + ($$13$us$i$i))|0;
          $742 = HEAP32[$743>>2]|0;
          $744 = (_oc_huff_token_decode($5,$742)|0);
          $745 = $744&255;
          $746 = (($ti$12$us$i$i) + 1)|0;
          $747 = (($711) + ($ti$12$us$i$i)|0);
          HEAP8[$747>>0] = $745;
          $748 = ($744>>>0)<(15);
          if ($748) {
           $749 = (529986 + ($744)|0);
           $750 = HEAP8[$749>>0]|0;
           $751 = $750&255;
           $752 = (_oc_pack_read($5,$751)|0);
           $753 = $752&255;
           $754 = (($ti$12$us$i$i) + 2)|0;
           $755 = (($711) + ($746)|0);
           HEAP8[$755>>0] = $753;
           $756 = ($744|0)==(0);
           if ($756) {
            $757 = $752 >>> 8;
            $758 = $757&255;
            $759 = (($ti$12$us$i$i) + 3)|0;
            $760 = (($711) + ($754)|0);
            HEAP8[$760>>0] = $758;
            $ti$2$us$i$i = $759;
           } else {
            $ti$2$us$i$i = $754;
           }
           $761 = ($744|0)<(2);
           $762 = $761 ? -13 : 0;
           $763 = ($744|0)<(12);
           $764 = $763 ? 21 : 0;
           $765 = (($762) + ($764))|0;
           $766 = $752 << $765;
           $eb$0$us$i$i = $766;$ti$3$us$i$i = $ti$2$us$i$i;
          } else {
           $eb$0$us$i$i = 0;$ti$3$us$i$i = $746;
          }
          $767 = (47184 + ($744<<2)|0);
          $768 = HEAP32[$767>>2]|0;
          $769 = (($768) + ($eb$0$us$i$i))|0;
          $770 = $769 >>> 8;
          $771 = $770 & 4095;
          $772 = ($769|0)==(0);
          $$$us$i$i = $772 ? 2147483647 : $771;
          $773 = ($$$us$i$i|0)==(0);
          if ($773) {
           $774 = $769 & 255;
           $775 = (($run_counts$i$i) + ($774<<2)|0);
           $776 = HEAP32[$775>>2]|0;
           $777 = (($776) + 1)|0;
           HEAP32[$775>>2] = $777;
           $778 = (($779) + 1)|0;
           $$1$be$us$i$i = 0;$ntoks$0$be$us$i$i = $778;
          } else {
           $$1$be$us$i$i = $$$us$i$i;$ntoks$0$be$us$i$i = $779;
          }
          $780 = (($$1$be$us$i$i) + ($ntoks$0$be$us$i$i))|0;
          $781 = ($780>>>0)<($718>>>0);
          if ($781) {
           $$13$us$i$i = $$1$be$us$i$i;$779 = $780;$eob_count$01$us$i$i = $741;$ti$12$us$i$i = $ti$3$us$i$i;
          } else {
           $$1$be$us$i$i$lcssa = $$1$be$us$i$i;$$lcssa255 = $741;$ntoks$0$be$us$i$i$lcssa = $ntoks$0$be$us$i$i;$ti$3$us$i$i$lcssa = $ti$3$us$i$i;
           break;
          }
         }
         $$pre19$i$i = HEAP32[$680>>2]|0;
         $$1$lcssa$us$i$i = $$1$be$us$i$i$lcssa;$722 = $$pre19$i$i;$eob_count$0$lcssa$us$i$i = $$lcssa255;$ntoks$0$lcssa$us$i$i = $ntoks$0$be$us$i$i$lcssa;$ti$1$lcssa$us$i$i = $ti$3$us$i$i$lcssa;
        } else {
         $$1$lcssa$us$i$i = $$014$us$i$i;$722 = 0;$eob_count$0$lcssa$us$i$i = 0;$ntoks$0$lcssa$us$i$i = 0;$ti$1$lcssa$us$i$i = $ti$013$us$i$i;
        }
        $720 = (($718) - ($ntoks$0$lcssa$us$i$i))|0;
        $721 = (($eob_count$0$lcssa$us$i$i) + ($722))|0;
        $723 = (($721) + ($720))|0;
        HEAP32[$680>>2] = $723;
        $735 = 62;$738 = $723;
        while(1) {
         $734 = (($run_counts$i$i) + ($735<<2)|0);
         $736 = HEAP32[$734>>2]|0;
         $737 = (($736) + ($738))|0;
         HEAP32[$734>>2] = $737;
         $739 = (($735) + -1)|0;
         $740 = ($735|0)>(0);
         if ($740) {
          $735 = $739;$738 = $737;
         } else {
          $$in$i$i = $713;
          break;
         }
        }
        while(1) {
         $726 = (($$in$i$i) + -1)|0;
         $727 = (($run_counts$i$i) + ($726<<2)|0);
         $728 = HEAP32[$727>>2]|0;
         $729 = (($726) + ($zzi$221$i))|0;
         $730 = ((($ntoks_left$i) + ($pli$011$us$i$i<<8)|0) + ($729<<2)|0);
         $731 = HEAP32[$730>>2]|0;
         $732 = (($731) - ($728))|0;
         HEAP32[$730>>2] = $732;
         $733 = ($$in$i$i|0)>(1);
         if ($733) {
          $$in$i$i = $726;
         } else {
          break;
         }
        }
        $724 = (($$1$lcssa$us$i$i) - ($720))|0;
        $725 = (($pli$011$us$i$i) + 1)|0;
        $exitcond18$i$i = ($725|0)==(3);
        if ($exitcond18$i$i) {
         $$0$lcssa$i$i = $724;$ti$0$lcssa$i$i = $ti$1$lcssa$us$i$i;
         break;
        } else {
         $$014$us$i$i = $724;$pli$011$us$i$i = $725;$ti$013$us$i$i = $ti$1$lcssa$us$i$i;
        }
       }
      } else {
       $$014$i$i = $eobs$120$i;$pli$011$i$i = 0;$ti$013$i$i = $1719;
       while(1) {
        $786 = ((((($_dec)) + 52664|0) + ($pli$011$i$i<<8)|0) + ($zzi$221$i<<2)|0);
        HEAP32[$786>>2] = $$014$i$i;
        $787 = ((((($_dec)) + 51896|0) + ($pli$011$i$i<<8)|0) + ($zzi$221$i<<2)|0);
        HEAP32[$787>>2] = $ti$013$i$i;
        $788 = ((($ntoks_left$i) + ($pli$011$i$i<<8)|0) + ($zzi$221$i<<2)|0);
        $789 = HEAP32[$788>>2]|0;
        _memset(($run_counts$i$i|0),0,256)|0;
        $790 = ($$014$i$i>>>0)<($789>>>0);
        if ($790) {
         $791 = (($pli$011$i$i) + 1)|0;
         $792 = $791 >> 1;
         $793 = (($huff_idxs$i) + ($792<<2)|0);
         $794 = HEAP32[$793>>2]|0;
         $795 = (((($_dec)) + 51576|0) + ($794<<2)|0);
         $$13$i$i = $$014$i$i;$833 = $$014$i$i;$eob_count$01$i$i = 0;$ti$12$i$i = $ti$013$i$i;
         while(1) {
          $796 = (($eob_count$01$i$i) + ($$13$i$i))|0;
          $797 = HEAP32[$795>>2]|0;
          $798 = (_oc_huff_token_decode($5,$797)|0);
          $799 = $798&255;
          $800 = (($ti$12$i$i) + 1)|0;
          $801 = (($711) + ($ti$12$i$i)|0);
          HEAP8[$801>>0] = $799;
          $802 = ($798>>>0)<(15);
          if ($802) {
           $803 = (529986 + ($798)|0);
           $804 = HEAP8[$803>>0]|0;
           $805 = $804&255;
           $806 = (_oc_pack_read($5,$805)|0);
           $807 = $806&255;
           $808 = (($ti$12$i$i) + 2)|0;
           $809 = (($711) + ($800)|0);
           HEAP8[$809>>0] = $807;
           $810 = ($798|0)==(0);
           if ($810) {
            $811 = $806 >>> 8;
            $812 = $811&255;
            $813 = (($ti$12$i$i) + 3)|0;
            $814 = (($711) + ($808)|0);
            HEAP8[$814>>0] = $812;
            $ti$2$i3$i = $813;
           } else {
            $ti$2$i3$i = $808;
           }
           $815 = ($798|0)<(2);
           $816 = $815 ? -13 : 0;
           $817 = ($798|0)<(12);
           $818 = $817 ? 21 : 0;
           $819 = (($816) + ($818))|0;
           $820 = $806 << $819;
           $eb$0$i5$i = $820;$ti$3$i4$i = $ti$2$i3$i;
          } else {
           $eb$0$i5$i = 0;$ti$3$i4$i = $800;
          }
          $821 = (47184 + ($798<<2)|0);
          $822 = HEAP32[$821>>2]|0;
          $823 = (($822) + ($eb$0$i5$i))|0;
          $824 = $823 >>> 8;
          $825 = $824 & 4095;
          $826 = ($823|0)==(0);
          $$$i6$i = $826 ? 2147483647 : $825;
          $827 = ($$$i6$i|0)==(0);
          if ($827) {
           $828 = $823 & 255;
           $829 = (($run_counts$i$i) + ($828<<2)|0);
           $830 = HEAP32[$829>>2]|0;
           $831 = (($830) + 1)|0;
           HEAP32[$829>>2] = $831;
           $832 = (($833) + 1)|0;
           $$1$be$i$i = 0;$ntoks$0$be$i$i = $832;
          } else {
           $$1$be$i$i = $$$i6$i;$ntoks$0$be$i$i = $833;
          }
          $834 = (($$1$be$i$i) + ($ntoks$0$be$i$i))|0;
          $835 = ($834>>>0)<($789>>>0);
          if ($835) {
           $$13$i$i = $$1$be$i$i;$833 = $834;$eob_count$01$i$i = $796;$ti$12$i$i = $ti$3$i4$i;
          } else {
           $$1$be$i$i$lcssa = $$1$be$i$i;$$lcssa253 = $796;$ntoks$0$be$i$i$lcssa = $ntoks$0$be$i$i;$ti$3$i4$i$lcssa = $ti$3$i4$i;
           break;
          }
         }
         $$pre$i$i = HEAP32[$680>>2]|0;
         $$1$lcssa$i$i = $$1$be$i$i$lcssa;$838 = $$pre$i$i;$eob_count$0$lcssa$i$i = $$lcssa253;$ntoks$0$lcssa$i$i = $ntoks$0$be$i$i$lcssa;$ti$1$lcssa$i7$i = $ti$3$i4$i$lcssa;
        } else {
         $$1$lcssa$i$i = $$014$i$i;$838 = 0;$eob_count$0$lcssa$i$i = 0;$ntoks$0$lcssa$i$i = 0;$ti$1$lcssa$i7$i = $ti$013$i$i;
        }
        $836 = (($789) - ($ntoks$0$lcssa$i$i))|0;
        $837 = (($eob_count$0$lcssa$i$i) + ($838))|0;
        $839 = (($837) + ($836))|0;
        HEAP32[$680>>2] = $839;
        $841 = 62;$844 = $839;
        while(1) {
         $840 = (($run_counts$i$i) + ($841<<2)|0);
         $842 = HEAP32[$840>>2]|0;
         $843 = (($842) + ($844))|0;
         HEAP32[$840>>2] = $843;
         $845 = (($841) + -1)|0;
         $846 = ($841|0)>(0);
         if ($846) {
          $841 = $845;$844 = $843;
         } else {
          break;
         }
        }
        $847 = (($$1$lcssa$i$i) - ($836))|0;
        $848 = (($pli$011$i$i) + 1)|0;
        $exitcond$i8$i = ($848|0)==(3);
        if ($exitcond$i8$i) {
         $$0$lcssa$i$i = $847;$ti$0$lcssa$i$i = $ti$1$lcssa$i7$i;
         break;
        } else {
         $$014$i$i = $847;$pli$011$i$i = $848;$ti$013$i$i = $ti$1$lcssa$i7$i;
        }
       }
      }
      HEAP32[$698>>2] = $ti$0$lcssa$i$i;
      $849 = (($zzi$221$i) + 1)|0;
      $exitcond141 = ($849|0)==($708|0);
      if ($exitcond141) {
       $$0$lcssa$i$i$lcssa = $$0$lcssa$i$i;
       break;
      } else {
       $1719 = $ti$0$lcssa$i$i;$eobs$120$i = $$0$lcssa$i$i;$zzi$221$i = $849;
      }
     }
     $850 = ($710|0)>($708|0);
     $smax$i = $850 ? $710 : $708;
     $eobs$1$lcssa$i = $$0$lcssa$i$i$lcssa;$zzi$2$lcssa$i = $smax$i;
    } else {
     $eobs$1$lcssa$i = $eobs$024$i;$zzi$2$lcssa$i = $zzi$125$i;
    }
    $851 = (($hgi$026$i) + 1)|0;
    $exitcond$i30 = ($851|0)==(5);
    if ($exitcond$i30) {
     break;
    } else {
     $703 = $702;$705 = $704;$eobs$024$i = $eobs$1$lcssa$i;$hgi$026$i = $851;$zzi$125$i = $zzi$2$lcssa$i;
    }
   }
   $863 = ((($_dec)) + 768|0);
   $864 = $863;
   $865 = $864;
   $866 = HEAP32[$865>>2]|0;
   $867 = (($864) + 4)|0;
   $868 = $867;
   $869 = HEAP32[$868>>2]|0;
   $870 = ((($_dec)) + 793|0);
   $871 = HEAP8[$870>>0]|0;
   $872 = $871&255;
   $873 = (_i64Add(($872|0),0,($866|0),($869|0))|0);
   $874 = tempRet0;
   $875 = ((($_dec)) + 60|0);
   $876 = HEAP32[$875>>2]|0;
   $877 = (_bitshift64Shl(($873|0),($874|0),($876|0))|0);
   $878 = tempRet0;
   $879 = ((($_dec)) + 776|0);
   $880 = $879;
   $881 = $880;
   $882 = HEAP32[$881>>2]|0;
   $883 = (($880) + 4)|0;
   $884 = $883;
   $885 = HEAP32[$884>>2]|0;
   $886 = (_i64Subtract(($882|0),($885|0),($866|0),($869|0))|0);
   $887 = tempRet0;
   $888 = (_i64Add(($886|0),($887|0),($877|0),($878|0))|0);
   $889 = tempRet0;
   $890 = ((($_dec)) + 784|0);
   $891 = $890;
   $892 = $891;
   HEAP32[$892>>2] = $888;
   $893 = (($891) + 4)|0;
   $894 = $893;
   HEAP32[$894>>2] = $889;
   $895 = (_i64Add(($882|0),($885|0),1,0)|0);
   $896 = tempRet0;
   $897 = $879;
   $898 = $897;
   HEAP32[$898>>2] = $895;
   $899 = (($897) + 4)|0;
   $900 = $899;
   HEAP32[$900>>2] = $896;
   $901 = ($_granpos|0)==(0|0);
   if (!($901)) {
    $902 = $_granpos;
    $903 = $902;
    HEAP32[$903>>2] = $888;
    $904 = (($902) + 4)|0;
    $905 = $904;
    HEAP32[$905>>2] = $889;
   }
   $906 = ((($_dec)) + 48|0);
   $907 = HEAP32[$906>>2]|0;
   $908 = $907 >>> 1;
   $$lobit$i31 = $908 & 1;
   $909 = $$lobit$i31 ^ 1;
   $910 = 4 << $909;
   $911 = ((($pipe)) + 2740|0);
   HEAP32[$911>>2] = $910;
   $912 = ((($pipe)) + 1024|0);
   $913 = ((($_dec)) + 51896|0);
   _memcpy(($912|0),($913|0),768)|0;
   $914 = ((($pipe)) + 1792|0);
   $915 = ((($_dec)) + 52664|0);
   _memcpy(($914|0),($915|0),768)|0;
   $916 = HEAP32[$862>>2]|0;
   $917 = ((($_dec)) + 216|0);
   $918 = HEAP32[$917>>2]|0;
   $919 = (($916) + ($918<<2)|0);
   $920 = ((($pipe)) + 2560|0);
   HEAP32[$920>>2] = $916;
   $921 = ((($pipe)) + 2572|0);
   HEAP32[$921>>2] = $919;
   $922 = HEAP32[$563>>2]|0;
   $923 = (($916) + ($922<<2)|0);
   $924 = ((($_dec)) + 120|0);
   $925 = HEAP32[$924>>2]|0;
   $926 = (($922) - ($925))|0;
   $$sum$i32 = (($926) + ($918))|0;
   $927 = (($916) + ($$sum$i32<<2)|0);
   $928 = ((($pipe)) + 2564|0);
   HEAP32[$928>>2] = $923;
   $929 = ((($pipe)) + 2576|0);
   HEAP32[$929>>2] = $927;
   $930 = HEAP32[$567>>2]|0;
   $$sum19$i = (($930) + ($922))|0;
   $931 = (($916) + ($$sum19$i<<2)|0);
   $932 = ((($_dec)) + 152|0);
   $933 = HEAP32[$932>>2]|0;
   $934 = (($930) - ($933))|0;
   $$sum20$i = (($934) + ($$sum$i32))|0;
   $935 = (($916) + ($$sum20$i<<2)|0);
   $936 = ((($pipe)) + 2568|0);
   HEAP32[$936>>2] = $931;
   $937 = ((($pipe)) + 2580|0);
   HEAP32[$937>>2] = $935;
   $938 = ((($_dec)) + 794|0);
   $939 = HEAP8[$938>>0]|0;
   $940 = ($939<<24>>24)==(0);
   if (!($940)) {
    $941 = $939&255;
    $qii$06$i = 0;
    while(1) {
     $942 = (((($_dec)) + 795|0) + ($qii$06$i)|0);
     $943 = HEAP8[$942>>0]|0;
     $944 = $943&255;
     $945 = (((($_dec)) + 800|0) + (($944*24)|0)|0);
     $946 = HEAP32[$945>>2]|0;
     $947 = (((($pipe)) + 2608|0) + ($qii$06$i<<3)|0);
     HEAP32[$947>>2] = $946;
     $948 = (((((($_dec)) + 800|0) + (($944*24)|0)|0)) + 4|0);
     $949 = HEAP32[$948>>2]|0;
     $950 = (((((($pipe)) + 2608|0) + ($qii$06$i<<3)|0)) + 4|0);
     HEAP32[$950>>2] = $949;
     $951 = (($qii$06$i) + 1)|0;
     $exitcond140 = ($951|0)==($941|0);
     if ($exitcond140) {
      $qii$06$1$i = 0;
      break;
     } else {
      $qii$06$i = $951;
     }
    }
    while(1) {
     $1065 = (((($_dec)) + 795|0) + ($qii$06$1$i)|0);
     $1066 = HEAP8[$1065>>0]|0;
     $1067 = $1066&255;
     $1068 = (((((($_dec)) + 800|0) + (($1067*24)|0)|0)) + 8|0);
     $1069 = HEAP32[$1068>>2]|0;
     $1070 = (((($pipe)) + 2632|0) + ($qii$06$1$i<<3)|0);
     HEAP32[$1070>>2] = $1069;
     $1071 = (((((($_dec)) + 800|0) + (($1067*24)|0)|0)) + 12|0);
     $1072 = HEAP32[$1071>>2]|0;
     $1073 = (((((($pipe)) + 2632|0) + ($qii$06$1$i<<3)|0)) + 4|0);
     HEAP32[$1073>>2] = $1072;
     $1074 = (($qii$06$1$i) + 1)|0;
     $exitcond139 = ($1074|0)==($941|0);
     if ($exitcond139) {
      $qii$06$2$i = 0;
      break;
     } else {
      $qii$06$1$i = $1074;
     }
    }
    while(1) {
     $1075 = (((($_dec)) + 795|0) + ($qii$06$2$i)|0);
     $1076 = HEAP8[$1075>>0]|0;
     $1077 = $1076&255;
     $1078 = (((((($_dec)) + 800|0) + (($1077*24)|0)|0)) + 16|0);
     $1079 = HEAP32[$1078>>2]|0;
     $1080 = (((($pipe)) + 2656|0) + ($qii$06$2$i<<3)|0);
     HEAP32[$1080>>2] = $1079;
     $1081 = (((((($_dec)) + 800|0) + (($1077*24)|0)|0)) + 20|0);
     $1082 = HEAP32[$1081>>2]|0;
     $1083 = (((((($pipe)) + 2656|0) + ($qii$06$2$i<<3)|0)) + 4|0);
     HEAP32[$1083>>2] = $1082;
     $1084 = (($qii$06$2$i) + 1)|0;
     $exitcond138 = ($1084|0)==($941|0);
     if ($exitcond138) {
      break;
     } else {
      $qii$06$2$i = $1084;
     }
    }
   }
   $1085 = ((($pipe)) + 2704|0);
   dest=$1085; stop=dest+36|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
   $1087 = (_oc_state_loop_filter_init($1086,$pipe)|0);
   $1088 = ($1087|0)==(0);
   $1089 = $1088&1;
   $1090 = ((($pipe)) + 2744|0);
   HEAP32[$1090>>2] = $1089;
   $976 = ((($_dec)) + 53444|0);
   $1091 = HEAP32[$976>>2]|0;
   $1092 = ($1091|0)<(1);
   $953 = ((($_dec)) + 53960|0);
   $952 = HEAP32[$953>>2]|0;
   $1093 = ($952|0)==(0|0);
   L353: do {
    if ($1092) {
     if ($1093) {
      label = 251;
     } else {
      _free($952);
      HEAP32[$953>>2] = 0;
      $954 = ((($_dec)) + 53964|0);
      $955 = HEAP32[$954>>2]|0;
      _free($955);
      HEAP32[$954>>2] = 0;
      $956 = ((($_dec)) + 53968|0);
      $957 = HEAP32[$956>>2]|0;
      _free($957);
      HEAP32[$956>>2] = 0;
      label = 251;
     }
    } else {
     if ($1093) {
      $958 = HEAP8[$11>>0]|0;
      $959 = ($958<<24>>24)==(0);
      if (!($959)) {
       label = 251;
       break;
      }
      $960 = HEAP32[$917>>2]|0;
      $961 = (_malloc($960)|0);
      HEAP32[$953>>2] = $961;
      $962 = ($961|0)==(0|0);
      if ($962) {
       label = 251;
       break;
      }
      $963 = HEAP8[$14>>0]|0;
      _memset(($961|0),($963|0),($960|0))|0;
      $978 = $1091;
     } else {
      $964 = HEAP32[$862>>2]|0;
      $965 = HEAP32[$563>>2]|0;
      $966 = HEAP32[$567>>2]|0;
      $967 = (($966) + ($965))|0;
      $968 = HEAP32[$854>>2]|0;
      $969 = (($967) + ($968))|0;
      $970 = HEAP8[$14>>0]|0;
      $971 = ($969|0)>(0);
      if ($971) {
       $fragii$02$i$i = 0;
       while(1) {
        $972 = (($964) + ($fragii$02$i$i<<2)|0);
        $973 = HEAP32[$972>>2]|0;
        $974 = (($952) + ($973)|0);
        HEAP8[$974>>0] = $970;
        $975 = (($fragii$02$i$i) + 1)|0;
        $exitcond$i$i34 = ($975|0)==($969|0);
        if ($exitcond$i$i34) {
         break;
        } else {
         $fragii$02$i$i = $975;
        }
       }
       $$pre$pre$i$i = HEAP32[$976>>2]|0;
       $978 = $$pre$pre$i$i;
      } else {
       $978 = $1091;
      }
     }
     $977 = ($978|0)<(2);
     $979 = ((($_dec)) + 53964|0);
     $980 = HEAP32[$979>>2]|0;
     $981 = ($980|0)==(0|0);
     if ($977) {
      if ($981) {
       label = 251;
       break;
      }
      _free($980);
      HEAP32[$979>>2] = 0;
      $982 = ((($_dec)) + 53968|0);
      $983 = HEAP32[$982>>2]|0;
      _free($983);
      HEAP32[$982>>2] = 0;
      label = 251;
      break;
     }
     do {
      if ($981) {
       $984 = ((($_dec)) + 4|0);
       $985 = HEAP32[$984>>2]|0;
       $986 = ((($_dec)) + 8|0);
       $987 = HEAP32[$986>>2]|0;
       $988 = Math_imul($987, $985)|0;
       $989 = HEAP32[$906>>2]|0;
       $990 = $989 & 1;
       $991 = $990 ^ 1;
       $992 = $985 >>> $991;
       $993 = $989 >>> 1;
       $$lobit1$i$i = $993 & 1;
       $994 = $$lobit1$i$i ^ 1;
       $995 = $987 >>> $994;
       $996 = $992 << 1;
       $997 = Math_imul($996, $995)|0;
       $998 = (($997) + ($988))|0;
       $999 = (_malloc($998)|0);
       $1000 = ((($_dec)) + 53968|0);
       HEAP32[$1000>>2] = $999;
       $1001 = HEAP32[$917>>2]|0;
       $1002 = $1001 << 2;
       $1003 = (_malloc($1002)|0);
       HEAP32[$979>>2] = $1003;
       $1004 = ($1003|0)==(0|0);
       $1005 = ($999|0)==(0|0);
       $or$cond$i$i = $1005 | $1004;
       if ($or$cond$i$i) {
        _free($999);
        HEAP32[$1000>>2] = 0;
        $1006 = HEAP32[$979>>2]|0;
        _free($1006);
        HEAP32[$979>>2] = 0;
        label = 251;
        break L353;
       } else {
        $1007 = ((($_dec)) + 53972|0);
        HEAP32[$1007>>2] = 0;
        $$pre$phi6$i$iZ2D = $1007;$1013 = $985;$1015 = $987;
        label = 244;
        break;
       }
      } else {
       $$phi$trans$insert$i$i = ((($_dec)) + 53972|0);
       $$pre3$i$i = HEAP32[$$phi$trans$insert$i$i>>2]|0;
       $1008 = ($978|0)>(4);
       $1009 = $1008 ? 2 : 1;
       $1010 = ($$pre3$i$i|0)==($1009|0);
       if ($1010) {
        $1053 = $978;
       } else {
        $$phi$trans$insert$i38 = ((($_dec)) + 4|0);
        $$pre$i39 = HEAP32[$$phi$trans$insert$i38>>2]|0;
        $$phi$trans$insert16$i = ((($_dec)) + 8|0);
        $$pre17$i = HEAP32[$$phi$trans$insert16$i>>2]|0;
        $$pre$phi6$i$iZ2D = $$phi$trans$insert$i$i;$1013 = $$pre$i39;$1015 = $$pre17$i;
        label = 244;
       }
      }
     } while(0);
     if ((label|0) == 244) {
      $1011 = ($978|0)<(5);
      if ($1011) {
       $1012 = ((($_dec)) + 53976|0);
       HEAP32[$1012>>2] = $1013;
       $1014 = ((($_dec)) + 53980|0);
       HEAP32[$1014>>2] = $1015;
       $1016 = (0 - ($1013))|0;
       $1017 = ((($_dec)) + 53984|0);
       HEAP32[$1017>>2] = $1016;
       $1018 = ((($_dec)) + 53968|0);
       $1019 = HEAP32[$1018>>2]|0;
       $1020 = (1 - ($1015))|0;
       $1021 = Math_imul($1020, $1016)|0;
       $1022 = (($1019) + ($1021)|0);
       $1023 = ((($_dec)) + 53988|0);
       HEAP32[$1023>>2] = $1022;
       $$pr$i$i = $978;
      } else {
       $1024 = Math_imul($1013, $1015)|0;
       $1025 = HEAP32[$906>>2]|0;
       $1026 = $1025 & 1;
       $1027 = $1026 ^ 1;
       $1028 = $1013 >>> $1027;
       $1029 = $1025 >>> 1;
       $$lobit$i$i = $1029 & 1;
       $1030 = $$lobit$i$i ^ 1;
       $1031 = $1015 >>> $1030;
       $1032 = Math_imul($1031, $1028)|0;
       $1033 = ((($_dec)) + 53976|0);
       HEAP32[$1033>>2] = $1013;
       $1034 = ((($_dec)) + 53980|0);
       HEAP32[$1034>>2] = $1015;
       $1035 = ((($_dec)) + 53984|0);
       HEAP32[$1035>>2] = $1013;
       $1036 = ((($_dec)) + 53968|0);
       $1037 = HEAP32[$1036>>2]|0;
       $1038 = ((($_dec)) + 53988|0);
       HEAP32[$1038>>2] = $1037;
       $1039 = ((($_dec)) + 53992|0);
       HEAP32[$1039>>2] = $1028;
       $1040 = ((($_dec)) + 53996|0);
       HEAP32[$1040>>2] = $1031;
       $1041 = ((($_dec)) + 54000|0);
       HEAP32[$1041>>2] = $1028;
       $1042 = $1037;
       $1043 = (($1042) + ($1024)|0);
       $1044 = ((($_dec)) + 54004|0);
       HEAP32[$1044>>2] = $1043;
       $1045 = ((($_dec)) + 54008|0);
       HEAP32[$1045>>2] = $1028;
       $1046 = ((($_dec)) + 54012|0);
       HEAP32[$1046>>2] = $1031;
       $1047 = ((($_dec)) + 54016|0);
       HEAP32[$1047>>2] = $1028;
       $$sum$i$i = (($1032) + ($1024))|0;
       $1048 = (($1042) + ($$sum$i$i)|0);
       $1049 = ((($_dec)) + 54020|0);
       HEAP32[$1049>>2] = $1048;
       _oc_ycbcr_buffer_flip($1033,$1033);
       $$pre4$i$i = HEAP32[$976>>2]|0;
       $$pr$i$i = $$pre4$i$i;
      }
      $1050 = ($$pr$i$i|0)>(4);
      $1051 = $1050 ? 2 : 1;
      HEAP32[$$pre$phi6$i$iZ2D>>2] = $1051;
      $1053 = $$pr$i$i;
     }
     $1052 = ($1053|0)<(5);
     if ($1052) {
      $1054 = ((($_dec)) + 53992|0);
      $1055 = ((($_dec)) + 280|0);
      $1056 = HEAP32[$1055>>2]|0;
      $1057 = (((((($_dec)) + 288|0) + (($1056*48)|0)|0)) + 16|0);
      ;HEAP32[$1054>>2]=HEAP32[$1057>>2]|0;HEAP32[$1054+4>>2]=HEAP32[$1057+4>>2]|0;HEAP32[$1054+8>>2]=HEAP32[$1057+8>>2]|0;HEAP32[$1054+12>>2]=HEAP32[$1057+12>>2]|0;HEAP32[$1054+16>>2]=HEAP32[$1057+16>>2]|0;HEAP32[$1054+20>>2]=HEAP32[$1057+20>>2]|0;HEAP32[$1054+24>>2]=HEAP32[$1057+24>>2]|0;HEAP32[$1054+28>>2]=HEAP32[$1057+28>>2]|0;
     }
     $1058 = HEAP32[$976>>2]|0;
     $1059 = ((($pipe)) + 2748|0);
     HEAP32[$1059>>2] = $1058;
     $$pre$phi153Z2D = $1059;$$pre$phi155Z2D = $979;
    }
   } while(0);
   if ((label|0) == 251) {
    $1060 = ((($pipe)) + 2748|0);
    HEAP32[$1060>>2] = 0;
    $1061 = ((($_dec)) + 53976|0);
    $1062 = ((($_dec)) + 280|0);
    $1063 = HEAP32[$1062>>2]|0;
    $1064 = (((($_dec)) + 288|0) + (($1063*48)|0)|0);
    dest=$1061; src=$1064; stop=dest+48|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
    $$pre154 = ((($_dec)) + 53964|0);
    $$pre$phi153Z2D = $1060;$$pre$phi155Z2D = $$pre154;
   }
   $1094 = ((($_dec)) + 53976|0);
   _oc_ycbcr_buffer_flip($stripe_buf,$1094);
   $1095 = ((($_dec)) + 112|0);
   $1096 = ((($_dec)) + 100|0);
   $1097 = ((($dc_quant$i)) + 2|0);
   $1098 = ((($_dec)) + 54028|0);
   $1099 = ((($_dec)) + 54024|0);
   $$pre148 = HEAP32[$911>>2]|0;
   $1107 = $$pre148;$notstart$0111 = 0;$stripe_fragy$0112 = 0;
   while(1) {
    $1105 = HEAP32[$1095>>2]|0;
    $1106 = (($1107) + ($stripe_fragy$0112))|0;
    $1108 = ($1106|0)<($1105|0);
    $1109 = $1108&1;
    $avail_fragy0$0107 = $1105;$avail_fragy_end$0108 = $1105;$pli$0109 = 0;
    while(1) {
     $1110 = ($pli$0109|0)!=(0);
     if ($1110) {
      $1111 = HEAP32[$906>>2]|0;
      $1112 = $1111 & 2;
      $1113 = ($1112|0)==(0);
      $1115 = $1113;
     } else {
      $1115 = 0;
     }
     $1114 = $1115&1;
     $1116 = $stripe_fragy$0112 >> $1114;
     $1117 = (((($pipe)) + 2680|0) + ($pli$0109<<2)|0);
     HEAP32[$1117>>2] = $1116;
     $1118 = (((((($_dec)) + 108|0) + ($pli$0109<<5)|0)) + 4|0);
     $1119 = HEAP32[$1118>>2]|0;
     $1120 = HEAP32[$911>>2]|0;
     $1121 = $1120 >> $1114;
     $1122 = (($1121) + ($1116))|0;
     $1123 = ($1122|0)<($1119|0);
     $1124 = $1123 ? $1122 : $1119;
     $1125 = (((($pipe)) + 2692|0) + ($pli$0109<<2)|0);
     HEAP32[$1125>>2] = $1124;
     $1126 = (((($_dec)) + 108|0) + ($pli$0109<<5)|0);
     $1127 = HEAP32[$1126>>2]|0;
     $1128 = HEAP32[$861>>2]|0;
     $1129 = ($1124|0)>($1116|0);
     if ($1129) {
      $1130 = (((((($_dec)) + 108|0) + ($pli$0109<<5)|0)) + 8|0);
      $1131 = HEAP32[$1130>>2]|0;
      $1132 = Math_imul($1127, $1116)|0;
      $1133 = (($1131) + ($1132))|0;
      $1134 = ($1127|0)>(0);
      $1135 = (1 - ($1127))|0;
      $1136 = $1127 ^ -1;
      $fragi$036$i = $1133;$fragy$035$i = $1116;$ncoded_fragis$037$i = 0;
      while(1) {
       $1137 = ($fragy$035$i|0)==(0);
       do {
        if ($1137) {
         if ($1134) {
          $fragi$130$i = $fragi$036$i;$fragx$029$i = 0;$ncoded_fragis$131$i = $ncoded_fragis$037$i;
          while(1) {
           $1138 = (($1128) + ($fragi$130$i<<2)|0);
           $1139 = HEAP32[$1138>>2]|0;
           $1140 = $1139 & 1;
           $1141 = ($1140|0)==(0);
           if ($1141) {
            $ncoded_fragis$2$i = $ncoded_fragis$131$i;
           } else {
            $1142 = $1139 >>> 6;
            $1143 = $1142 & 28;
            $1144 = 268505377 >>> $1143;
            $1145 = $1144 & 15;
            $1146 = ((((($pipe)) + 2704|0) + (($pli$0109*12)|0)|0) + ($1145<<2)|0);
            $1147 = HEAP32[$1146>>2]|0;
            $1148 = $1147 << 16;
            $1149 = (($1148) + ($1139))|0;
            HEAP32[$1138>>2] = $1149;
            $1150 = $1149 >> 16;
            HEAP32[$1146>>2] = $1150;
            $1151 = (($ncoded_fragis$131$i) + 1)|0;
            $ncoded_fragis$2$i = $1151;
           }
           $1152 = (($fragx$029$i) + 1)|0;
           $1153 = (($fragi$130$i) + 1)|0;
           $exitcond41$i = ($1152|0)==($1127|0);
           if ($exitcond41$i) {
            $ncoded_fragis$2$i$lcssa = $ncoded_fragis$2$i;
            break;
           } else {
            $fragi$130$i = $1153;$fragx$029$i = $1152;$ncoded_fragis$131$i = $ncoded_fragis$2$i;
           }
          }
          $1281 = (($fragi$036$i) + ($1127))|0;
          $fragi$3$i = $1281;$ncoded_fragis$5$i45 = $ncoded_fragis$2$i$lcssa;
         } else {
          $fragi$3$i = $fragi$036$i;$ncoded_fragis$5$i45 = $ncoded_fragis$037$i;
         }
        } else {
         $$sum$i41 = (($fragi$036$i) - ($1127))|0;
         $1154 = (($1128) + ($$sum$i41<<2)|0);
         $1155 = HEAP32[$1154>>2]|0;
         $1156 = $1155 & 1;
         $1157 = ($1156|0)==(0);
         if ($1157) {
          $u_ref$0$ph$i = -1;
         } else {
          $1158 = $1155 >>> 6;
          $1159 = $1158 & 28;
          $1160 = 268505377 >>> $1159;
          $1161 = $1160 & 15;
          $u_ref$0$ph$i = $1161;
         }
         if ($1134) {
          $fragi$224$i = $fragi$036$i;$fragx$123$i = 0;$l_ref$020$i = -1;$ncoded_fragis$327$i = $ncoded_fragis$037$i;$u_ref$022$i = $u_ref$0$ph$i;$ul_ref$021$i = -1;
         } else {
          $fragi$3$i = $fragi$036$i;$ncoded_fragis$5$i45 = $ncoded_fragis$037$i;
          break;
         }
         while(1) {
          $1162 = (($fragx$123$i) + 1)|0;
          $1163 = ($1162|0)<($1127|0);
          do {
           if ($1163) {
            $$sum1$i = (($fragi$224$i) + ($1135))|0;
            $1164 = (($1128) + ($$sum1$i<<2)|0);
            $1165 = HEAP32[$1164>>2]|0;
            $1166 = $1165 & 1;
            $1167 = ($1166|0)==(0);
            if ($1167) {
             $ur_ref$0$i = -1;
             break;
            }
            $1168 = $1165 >>> 6;
            $1169 = $1168 & 28;
            $1170 = 268505377 >>> $1169;
            $1171 = $1170 & 15;
            $ur_ref$0$i = $1171;
           } else {
            $ur_ref$0$i = -1;
           }
          } while(0);
          $1172 = (($1128) + ($fragi$224$i<<2)|0);
          $1173 = HEAP32[$1172>>2]|0;
          $1174 = $1173 & 1;
          $1175 = ($1174|0)==(0);
          if ($1175) {
           $l_ref$1$i = -1;$ncoded_fragis$4$i43 = $ncoded_fragis$327$i;
          } else {
           $1176 = $1173 >>> 6;
           $1177 = $1176 & 28;
           $1178 = 268505377 >>> $1177;
           $1179 = $1178 & 15;
           $1180 = ($l_ref$020$i|0)==($1179|0);
           $1181 = $1180&1;
           $1182 = ($ul_ref$021$i|0)==($1179|0);
           $1183 = $1182&1;
           $1184 = $1183 << 1;
           $1185 = $1184 | $1181;
           $1186 = ($u_ref$022$i|0)==($1179|0);
           $1187 = $1186&1;
           $1188 = $1187 << 2;
           $1189 = $1185 | $1188;
           $1190 = ($ur_ref$0$i|0)==($1179|0);
           $1191 = $1190&1;
           $1192 = $1191 << 3;
           $1193 = $1189 | $1192;
           L421: do {
            switch ($1193|0) {
            case 3: case 1:  {
             $1196 = (($fragi$224$i) + -1)|0;
             $1197 = (($1128) + ($1196<<2)|0);
             $1198 = HEAP32[$1197>>2]|0;
             $1199 = $1198 >> 16;
             $pred$0$i = $1199;
             break;
            }
            case 2:  {
             $$sum17$i = (($fragi$224$i) + ($1136))|0;
             $1200 = (($1128) + ($$sum17$i<<2)|0);
             $1201 = HEAP32[$1200>>2]|0;
             $1202 = $1201 >> 16;
             $pred$0$i = $1202;
             break;
            }
            case 12: case 6: case 4:  {
             $$sum16$i = (($fragi$224$i) - ($1127))|0;
             $1203 = (($1128) + ($$sum16$i<<2)|0);
             $1204 = HEAP32[$1203>>2]|0;
             $1205 = $1204 >> 16;
             $pred$0$i = $1205;
             break;
            }
            case 5:  {
             $1206 = (($fragi$224$i) + -1)|0;
             $1207 = (($1128) + ($1206<<2)|0);
             $1208 = HEAP32[$1207>>2]|0;
             $1209 = $1208 >> 16;
             $$sum15$i = (($fragi$224$i) - ($1127))|0;
             $1210 = (($1128) + ($$sum15$i<<2)|0);
             $1211 = HEAP32[$1210>>2]|0;
             $1212 = $1211 >> 16;
             $1213 = (($1212) + ($1209))|0;
             $1214 = (($1213|0) / 2)&-1;
             $pred$0$i = $1214;
             break;
            }
            case 8:  {
             $$sum14$i = (($fragi$224$i) + ($1135))|0;
             $1215 = (($1128) + ($$sum14$i<<2)|0);
             $1216 = HEAP32[$1215>>2]|0;
             $1217 = $1216 >> 16;
             $pred$0$i = $1217;
             break;
            }
            case 13: case 11: case 9:  {
             $1218 = (($fragi$224$i) + -1)|0;
             $1219 = (($1128) + ($1218<<2)|0);
             $1220 = HEAP32[$1219>>2]|0;
             $1221 = $1220 >> 16;
             $1222 = ($1221*75)|0;
             $$sum13$i = (($fragi$224$i) + ($1135))|0;
             $1223 = (($1128) + ($$sum13$i<<2)|0);
             $1224 = HEAP32[$1223>>2]|0;
             $1225 = $1224 >> 16;
             $1226 = ($1225*53)|0;
             $1227 = (($1226) + ($1222))|0;
             $1228 = (($1227|0) / 128)&-1;
             $pred$0$i = $1228;
             break;
            }
            case 10:  {
             $$sum11$i = (($fragi$224$i) + ($1136))|0;
             $1229 = (($1128) + ($$sum11$i<<2)|0);
             $1230 = HEAP32[$1229>>2]|0;
             $1231 = $1230 >> 16;
             $$sum12$i = (($fragi$224$i) + ($1135))|0;
             $1232 = (($1128) + ($$sum12$i<<2)|0);
             $1233 = HEAP32[$1232>>2]|0;
             $1234 = $1233 >> 16;
             $1235 = (($1234) + ($1231))|0;
             $1236 = (($1235|0) / 2)&-1;
             $pred$0$i = $1236;
             break;
            }
            case 14:  {
             $$sum8$i = (($fragi$224$i) + ($1136))|0;
             $1237 = (($1128) + ($$sum8$i<<2)|0);
             $1238 = HEAP32[$1237>>2]|0;
             $1239 = $1238 >> 16;
             $$sum9$i = (($fragi$224$i) + ($1135))|0;
             $1240 = (($1128) + ($$sum9$i<<2)|0);
             $1241 = HEAP32[$1240>>2]|0;
             $1242 = $1241 >> 16;
             $1243 = (($1242) + ($1239))|0;
             $1244 = ($1243*3)|0;
             $$sum10$i = (($fragi$224$i) - ($1127))|0;
             $1245 = (($1128) + ($$sum10$i<<2)|0);
             $1246 = HEAP32[$1245>>2]|0;
             $1247 = $1246 >> 16;
             $1248 = ($1247*10)|0;
             $1249 = (($1244) + ($1248))|0;
             $1250 = (($1249|0) / 16)&-1;
             $pred$0$i = $1250;
             break;
            }
            case 15: case 7:  {
             $1251 = (($fragi$224$i) + -1)|0;
             $1252 = (($1128) + ($1251<<2)|0);
             $1253 = HEAP32[$1252>>2]|0;
             $1254 = $1253 >> 16;
             $$sum2$i = (($1251) - ($1127))|0;
             $1255 = (($1128) + ($$sum2$i<<2)|0);
             $1256 = HEAP32[$1255>>2]|0;
             $1257 = $1256 >> 16;
             $$sum3$i = (($fragi$224$i) - ($1127))|0;
             $1258 = (($1128) + ($$sum3$i<<2)|0);
             $1259 = HEAP32[$1258>>2]|0;
             $1260 = $1259 >> 16;
             $1261 = (($1260) + ($1254))|0;
             $1262 = ($1261*29)|0;
             $1263 = Math_imul($1257, -26)|0;
             $1264 = (($1262) + ($1263))|0;
             $1265 = (($1264|0) / 32)&-1;
             $1266 = (($1265) - ($1260))|0;
             $ispos$i = ($1266|0)>(-1);
             $neg$i = (0 - ($1266))|0;
             $1267 = $ispos$i ? $1266 : $neg$i;
             $1268 = ($1267|0)>(128);
             if ($1268) {
              $pred$0$i = $1260;
              break L421;
             }
             $1269 = (($1265) - ($1254))|0;
             $ispos4$i = ($1269|0)>(-1);
             $neg5$i = (0 - ($1269))|0;
             $1270 = $ispos4$i ? $1269 : $neg5$i;
             $1271 = ($1270|0)>(128);
             if ($1271) {
              $pred$0$i = $1254;
              break L421;
             }
             $1272 = (($1265) - ($1257))|0;
             $ispos6$i = ($1272|0)>(-1);
             $neg7$i = (0 - ($1272))|0;
             $1273 = $ispos6$i ? $1272 : $neg7$i;
             $1274 = ($1273|0)>(128);
             $$$i = $1274 ? $1257 : $1265;
             $pred$0$i = $$$i;
             break;
            }
            default: {
             $1194 = ((((($pipe)) + 2704|0) + (($pli$0109*12)|0)|0) + ($1179<<2)|0);
             $1195 = HEAP32[$1194>>2]|0;
             $pred$0$i = $1195;
            }
            }
           } while(0);
           $1275 = $pred$0$i << 16;
           $1276 = (($1275) + ($1173))|0;
           HEAP32[$1172>>2] = $1276;
           $1277 = $1276 >> 16;
           $1278 = ((((($pipe)) + 2704|0) + (($pli$0109*12)|0)|0) + ($1179<<2)|0);
           HEAP32[$1278>>2] = $1277;
           $1279 = (($ncoded_fragis$327$i) + 1)|0;
           $l_ref$1$i = $1179;$ncoded_fragis$4$i43 = $1279;
          }
          $1280 = (($fragi$224$i) + 1)|0;
          $exitcond$i44 = ($1162|0)==($1127|0);
          if ($exitcond$i44) {
           $ncoded_fragis$4$i43$lcssa = $ncoded_fragis$4$i43;
           break;
          } else {
           $ul_ref$021$i$phi = $u_ref$022$i;$fragi$224$i = $1280;$fragx$123$i = $1162;$l_ref$020$i = $l_ref$1$i;$ncoded_fragis$327$i = $ncoded_fragis$4$i43;$u_ref$022$i = $ur_ref$0$i;$ul_ref$021$i = $ul_ref$021$i$phi;
          }
         }
         $1282 = (($fragi$036$i) + ($1127))|0;
         $fragi$3$i = $1282;$ncoded_fragis$5$i45 = $ncoded_fragis$4$i43$lcssa;
        }
       } while(0);
       $1283 = (($fragy$035$i) + 1)|0;
       $exitcond42$i = ($1283|0)==($1124|0);
       if ($exitcond42$i) {
        $ncoded_fragis$5$i45$lcssa = $ncoded_fragis$5$i45;
        break;
       } else {
        $fragi$036$i = $fragi$3$i;$fragy$035$i = $1283;$ncoded_fragis$037$i = $ncoded_fragis$5$i45;
       }
      }
      $$pre149 = HEAP32[$861>>2]|0;
      $1303 = $$pre149;$ncoded_fragis$0$lcssa$i = $ncoded_fragis$5$i45$lcssa;
     } else {
      $1303 = $1128;$ncoded_fragis$0$lcssa$i = 0;
     }
     $1284 = (((($pipe)) + 2584|0) + ($pli$0109<<2)|0);
     HEAP32[$1284>>2] = $ncoded_fragis$0$lcssa$i;
     $1285 = (($1124) - ($1116))|0;
     $1286 = Math_imul($1285, $1127)|0;
     $1287 = (($1286) - ($ncoded_fragis$0$lcssa$i))|0;
     $1288 = (((($pipe)) + 2596|0) + ($pli$0109<<2)|0);
     HEAP32[$1288>>2] = $1287;
     $1289 = HEAP32[$712>>2]|0;
     $1290 = HEAP32[$1096>>2]|0;
     $1291 = (((($pipe)) + 2560|0) + ($pli$0109<<2)|0);
     $1292 = HEAP32[$1291>>2]|0;
     $1293 = (((($pipe)) + 2608|0) + (($pli$0109*24)|0)|0);
     $1294 = HEAP32[$1293>>2]|0;
     $1295 = HEAP16[$1294>>1]|0;
     HEAP16[$dc_quant$i>>1] = $1295;
     $1296 = (((((($pipe)) + 2608|0) + (($pli$0109*24)|0)|0)) + 4|0);
     $1297 = HEAP32[$1296>>2]|0;
     $1298 = HEAP16[$1297>>1]|0;
     HEAP16[$1097>>1] = $1298;
     $1299 = ($ncoded_fragis$0$lcssa$i|0)>(0);
     if ($1299) {
      $fragii$07$i = 0;
      while(1) {
       $1300 = (($1292) + ($fragii$07$i<<2)|0);
       $1301 = HEAP32[$1300>>2]|0;
       dest=$dct_coeffs$i; stop=dest+128|0; do { HEAP16[dest>>1]=0|0; dest=dest+2|0; } while ((dest|0) < (stop|0));
       $1302 = (($1303) + ($1301<<2)|0);
       $1304 = HEAP32[$1302>>2]|0;
       $1305 = $1304 & 1792;
       $1306 = ($1305|0)!=(256);
       $1307 = $1306&1;
       $1308 = $1304 >>> 2;
       $1309 = $1308 & 63;
       $1310 = (((((($pipe)) + 2608|0) + (($pli$0109*24)|0)|0) + ($1309<<3)|0) + ($1307<<2)|0);
       $1311 = HEAP32[$1310>>2]|0;
       $zzi$16$i = 0;
       while(1) {
        $1312 = ((((($pipe)) + 1792|0) + ($pli$0109<<8)|0) + ($zzi$16$i<<2)|0);
        $1313 = HEAP32[$1312>>2]|0;
        $1314 = ($1313|0)==(0);
        if (!($1314)) {
         $$lcssa = $1312;$$lcssa249 = $1313;$zzi$16$i$lcssa = $zzi$16$i;
         label = 296;
         break;
        }
        $1316 = ((((($pipe)) + 1024|0) + ($pli$0109<<8)|0) + ($zzi$16$i<<2)|0);
        $1317 = HEAP32[$1316>>2]|0;
        $1318 = (($1317) + 1)|0;
        $1319 = (($1289) + ($1317)|0);
        $1320 = HEAP8[$1319>>0]|0;
        $1321 = $1320&255;
        $1322 = (47184 + ($1321<<2)|0);
        $1323 = HEAP32[$1322>>2]|0;
        $1324 = ($1320&255)<(15);
        do {
         if ($1324) {
          $1327 = (($1317) + 2)|0;
          $1328 = (($1289) + ($1318)|0);
          $1329 = HEAP8[$1328>>0]|0;
          $1330 = $1329&255;
          $1331 = ($1320&255)<(2);
          $1332 = $1331 ? -13 : 0;
          $1333 = ($1320&255)<(12);
          $1334 = $1333 ? 21 : 0;
          $1335 = (($1332) + ($1334))|0;
          $1336 = $1330 << $1335;
          $1337 = (($1336) + ($1323))|0;
          $1338 = $1337 >>> 8;
          $1339 = $1338 & 4095;
          $1340 = ($1320<<24>>24)==(0);
          if (!($1340)) {
           $cw$01$i = $1337;$eob$0$i = $1339;$lti$1$i = $1327;
           break;
          }
          $1341 = (($1317) + 3)|0;
          $1342 = (($1289) + ($1327)|0);
          $1343 = HEAP8[$1342>>0]|0;
          $1344 = $1343&255;
          $1345 = $1344 << 8;
          $1346 = (($1345) + ($1339))|0;
          $1347 = ($1346|0)==(0);
          $$$i50 = $1347 ? 2147483647 : $1346;
          $cw$01$i = $1337;$eob$0$i = $$$i50;$lti$1$i = $1341;
         } else {
          $1325 = $1323 >>> 8;
          $1326 = $1325 & 4095;
          $cw$01$i = $1323;$eob$0$i = $1326;$lti$1$i = $1318;
         }
        } while(0);
        $1348 = $cw$01$i & 255;
        $1349 = $cw$01$i & 1048576;
        $1350 = (0 - ($1349))|0;
        $1351 = $cw$01$i ^ $1350;
        $1352 = $1351 >> 21;
        HEAP32[$1312>>2] = $eob$0$i;
        HEAP32[$1316>>2] = $lti$1$i;
        $1353 = (($1348) + ($zzi$16$i))|0;
        $1354 = (($1311) + ($1353<<1)|0);
        $1355 = HEAP16[$1354>>1]|0;
        $1356 = $1355&65535;
        $1357 = Math_imul($1356, $1352)|0;
        $1358 = $1357&65535;
        $1359 = (($1290) + ($1353)|0);
        $1360 = HEAP8[$1359>>0]|0;
        $1361 = $1360&255;
        $1362 = (($dct_coeffs$i) + ($1361<<1)|0);
        HEAP16[$1362>>1] = $1358;
        $1363 = ($eob$0$i|0)==(0);
        $1364 = $1363&1;
        $1365 = (($1364) + ($1353))|0;
        $1366 = ($1365|0)<(64);
        if ($1366) {
         $zzi$16$i = $1365;
        } else {
         $zzi$16$i247 = $zzi$16$i;
         break;
        }
       }
       if ((label|0) == 296) {
        label = 0;
        $1315 = (($$lcssa249) + -1)|0;
        HEAP32[$$lcssa>>2] = $1315;
        $zzi$16$i247 = $zzi$16$i$lcssa;
       }
       $1367 = HEAP32[$1302>>2]|0;
       $1368 = $1367 >>> 16;
       $1369 = $1368&65535;
       HEAP16[$dct_coeffs$i>>1] = $1369;
       $1370 = (($dc_quant$i) + ($1307<<1)|0);
       $1371 = HEAP16[$1370>>1]|0;
       _oc_state_frag_recon($1086,$1301,$pli$0109,$dct_coeffs$i,$zzi$16$i247,$1371);
       $1372 = (($fragii$07$i) + 1)|0;
       $exitcond$i51 = ($1372|0)==($ncoded_fragis$0$lcssa$i|0);
       if ($exitcond$i51) {
        break;
       } else {
        $fragii$07$i = $1372;
       }
      }
      $$pre$i53 = HEAP32[$1291>>2]|0;
      $$pre150 = HEAP32[$1288>>2]|0;
      $1374 = $$pre$i53;$1378 = $$pre150;
     } else {
      $1374 = $1292;$1378 = $1287;
     }
     $1373 = (($1374) + ($ncoded_fragis$0$lcssa$i<<2)|0);
     HEAP32[$1291>>2] = $1373;
     $1375 = (((($pipe)) + 2572|0) + ($pli$0109<<2)|0);
     $1376 = HEAP32[$1375>>2]|0;
     $1377 = (0 - ($1378))|0;
     $1379 = (($1376) + ($1377<<2)|0);
     HEAP32[$1375>>2] = $1379;
     _oc_state_frag_copy_list($1086,$1379,$1378,2,1,$pli$0109);
     $1380 = HEAP32[$1090>>2]|0;
     $1381 = ($1380|0)==(0);
     if ($1381) {
      $edelay$0 = 0;$sdelay$0 = 0;
     } else {
      $1382 = HEAP32[$1117>>2]|0;
      $1383 = (($1382) - ($notstart$0111))|0;
      $1384 = HEAP32[$1125>>2]|0;
      $1385 = (($1384) - ($1109))|0;
      _oc_state_loop_filter_frag_rows($1086,$pipe,$refi$1,$pli$0109,$1383,$1385);
      $edelay$0 = $1109;$sdelay$0 = $notstart$0111;
     }
     $1386 = HEAP32[$1117>>2]|0;
     $1387 = (($1386) - ($sdelay$0))|0;
     $1388 = $1387 << 3;
     $1389 = $sdelay$0 << 1;
     $1390 = (($1388) - ($1389))|0;
     $1391 = HEAP32[$1125>>2]|0;
     $1392 = (($1391) - ($edelay$0))|0;
     $1393 = $1392 << 3;
     $1394 = $edelay$0 << 1;
     $1395 = (($1393) - ($1394))|0;
     _oc_state_borders_fill_rows($1086,$refi$1,$pli$0109,$1390,$1395);
     $1396 = $1110 ? 3 : 0;
     $1397 = HEAP32[$$pre$phi153Z2D>>2]|0;
     $1398 = (($1396) + 2)|0;
     $1399 = ($1397|0)<($1398|0);
     do {
      if ($1399) {
       $1618 = HEAP32[$1090>>2]|0;
       $1619 = ($1618|0)==(0);
       if ($1619) {
        $edelay$1 = $edelay$0;$sdelay$1 = $sdelay$0;
       } else {
        $1620 = (($sdelay$0) + ($notstart$0111))|0;
        $1621 = (($edelay$0) + ($1109))|0;
        $edelay$1 = $1621;$sdelay$1 = $1620;
       }
      } else {
       $1400 = (($sdelay$0) + ($notstart$0111))|0;
       $1401 = (($edelay$0) + ($1109))|0;
       $1402 = HEAP32[$1117>>2]|0;
       $1403 = (($1402) - ($1400))|0;
       $1404 = HEAP32[$1125>>2]|0;
       $1405 = (($1404) - ($1401))|0;
       $1406 = HEAP32[$1126>>2]|0;
       $1407 = (((((($_dec)) + 108|0) + ($pli$0109<<5)|0)) + 8|0);
       $1408 = HEAP32[$1407>>2]|0;
       $1409 = Math_imul($1406, $1403)|0;
       $1410 = (($1408) + ($1409))|0;
       $1411 = HEAP32[$$pre$phi155Z2D>>2]|0;
       $1412 = (($1411) + ($1410<<2)|0);
       $1413 = HEAP32[$953>>2]|0;
       $1414 = (($1413) + ($1410)|0);
       $1415 = ($1403|0)>(0);
       $1416 = $1415&1;
       $1417 = HEAP32[$1118>>2]|0;
       $1418 = ($1417|0)>($1405|0);
       $1419 = $1418&1;
       $1420 = $1415 ? $1406 : 0;
       $$sum$i55 = (($1410) + ($1420))|0;
       $1421 = (($1411) + ($$sum$i55<<2)|0);
       $$neg$i = $1415 << 31 >> 31;
       $1422 = (($1405) - ($1403))|0;
       $1423 = (($1422) + ($$neg$i))|0;
       $1424 = (($1423) + ($1419))|0;
       $1425 = $1406 << 2;
       $1426 = Math_imul($1425, $1424)|0;
       _memset(($1421|0),0,($1426|0))|0;
       $1427 = $1403 << 3;
       $1428 = $1416 << 2;
       $1429 = $1428 | $1427;
       $1430 = (((((($_dec)) + 53976|0) + ($pli$0109<<4)|0)) + 8|0);
       $1431 = HEAP32[$1430>>2]|0;
       $1432 = ((((((($_dec)) + 288|0) + (($refi$1*48)|0)|0) + ($pli$0109<<4)|0)) + 8|0);
       $1433 = HEAP32[$1432>>2]|0;
       $1434 = (((((($_dec)) + 53976|0) + ($pli$0109<<4)|0)) + 12|0);
       $1435 = HEAP32[$1434>>2]|0;
       $1436 = Math_imul($1431, $1429)|0;
       $1437 = (($1435) + ($1436)|0);
       $1438 = ((((((($_dec)) + 288|0) + (($refi$1*48)|0)|0) + ($pli$0109<<4)|0)) + 12|0);
       $1439 = HEAP32[$1438>>2]|0;
       $1440 = Math_imul($1433, $1429)|0;
       $1441 = (($1439) + ($1440)|0);
       $1442 = (((($_dec)) + 53976|0) + ($pli$0109<<4)|0);
       $1443 = HEAP32[$1442>>2]|0;
       $1444 = ($1429|0)<(4);
       if ($1444) {
        $1445 = $1433 << 2;
        $dst$042$i = $1437;$src$041$i = $1441;$y$043$i = $1429;
        while(1) {
         _memcpy(($dst$042$i|0),($src$041$i|0),($1443|0))|0;
         $1446 = (($dst$042$i) + ($1431)|0);
         $1447 = (($src$041$i) + ($1433)|0);
         $1448 = (($y$043$i) + 1)|0;
         $exitcond64$i = ($1448|0)==(4);
         if ($exitcond64$i) {
          break;
         } else {
          $dst$042$i = $1446;$src$041$i = $1447;$y$043$i = $1448;
         }
        }
        $1449 = $1431 << 2;
        $scevgep62$i = (($1439) + ($1445)|0);
        $scevgep63$i = (($1435) + ($1449)|0);
        $1497 = $1445;$1498 = $1449;$dst$0$lcssa$i = $scevgep63$i;$src$0$lcssa$i = $scevgep62$i;$y$0$lcssa$i = 4;
       } else {
        $1497 = $1440;$1498 = $1436;$dst$0$lcssa$i = $1437;$src$0$lcssa$i = $1441;$y$0$lcssa$i = $1429;
       }
       $1450 = $1419 ^ 1;
       $1451 = (($1405) - ($1450))|0;
       $1452 = $1451 << 3;
       $1453 = ($y$0$lcssa$i|0)<($1452|0);
       if ($1453) {
        $1454 = (0 - ($1433))|0;
        $1455 = ($1443|0)>(8);
        $1456 = $1431 << 3;
        $1457 = $1433 << 3;
        $variance$1$sum$i = (($1406) + 1)|0;
        $1458 = $1431 << 2;
        $$sum5$i = (-4 - ($1458))|0;
        $1459 = (($1443) + -9)|0;
        $1460 = $1459 >>> 3;
        $1461 = (($1460) + 2)|0;
        $1462 = $1405 << 3;
        $1463 = (($1462) + -1)|0;
        $1464 = $1450 << 3;
        $1465 = (($1463) - ($1464))|0;
        $1466 = (($1465) - ($y$0$lcssa$i))|0;
        $1467 = $1466 & -8;
        $1468 = (($1467) + 8)|0;
        $1469 = Math_imul($1468, $1433)|0;
        $dc_qi$030$i = $1414;$dst$128$i = $dst$0$lcssa$i;$src$126$i = $src$0$lcssa$i;$variance$031$i = $1412;$y$132$i = $y$0$lcssa$i;
        while(1) {
         $1470 = HEAP8[$dc_qi$030$i>>0]|0;
         $1471 = $1470&255;
         $1472 = (((($_dec)) + 53448|0) + ($1471<<2)|0);
         $1473 = HEAP32[$1472>>2]|0;
         $1474 = ($1473*3)|0;
         $1475 = $1474 >> 2;
         $1476 = (($src$126$i) + ($1454)|0);
         $1477 = (($variance$031$i) + ($1406<<2)|0);
         _oc_filter_hedge($dst$128$i,$1431,$1476,$1433,$1473,$1475,$variance$031$i,$1477);
         $variance$117$i = ((($variance$031$i)) + 4|0);
         $dc_qi$118$i = ((($dc_qi$030$i)) + 1|0);
         if ($1455) {
          $dc_qi$122$i = $dc_qi$118$i;$variance$0$pn19$i = $variance$031$i;$variance$121$i = $variance$117$i;$x$020$i = 8;
          while(1) {
           $1478 = HEAP8[$dc_qi$122$i>>0]|0;
           $1479 = $1478&255;
           $1480 = (((($_dec)) + 53448|0) + ($1479<<2)|0);
           $1481 = HEAP32[$1480>>2]|0;
           $1482 = ($1481*3)|0;
           $1483 = $1482 >> 2;
           $1484 = (($dst$128$i) + ($x$020$i)|0);
           $$sum4$i = (($x$020$i) - ($1433))|0;
           $1485 = (($src$126$i) + ($$sum4$i)|0);
           $1486 = (($variance$0$pn19$i) + ($variance$1$sum$i<<2)|0);
           _oc_filter_hedge($1484,$1431,$1485,$1433,$1481,$1483,$variance$121$i,$1486);
           $$sum6$i = (($$sum5$i) + ($x$020$i))|0;
           $1487 = (($dst$128$i) + ($$sum6$i)|0);
           _oc_filter_vedge($1487,$1431,$1481,$1483,$variance$0$pn19$i);
           $1488 = (($x$020$i) + 8)|0;
           $variance$1$i = ((($variance$121$i)) + 4|0);
           $dc_qi$1$i = ((($dc_qi$122$i)) + 1|0);
           $1489 = ($1488|0)<($1443|0);
           if ($1489) {
            $variance$0$pn19$i$phi = $variance$121$i;$dc_qi$122$i = $dc_qi$1$i;$variance$121$i = $variance$1$i;$x$020$i = $1488;$variance$0$pn19$i = $variance$0$pn19$i$phi;
           } else {
            break;
           }
          }
          $scevgep58$i = (($dc_qi$030$i) + ($1461)|0);
          $scevgep59$i = (($variance$031$i) + ($1461<<2)|0);
          $dc_qi$1$lcssa$i = $scevgep58$i;$variance$1$lcssa$i = $scevgep59$i;
         } else {
          $dc_qi$1$lcssa$i = $dc_qi$118$i;$variance$1$lcssa$i = $variance$117$i;
         }
         $1490 = (($dst$128$i) + ($1456)|0);
         $1491 = (($src$126$i) + ($1457)|0);
         $1492 = (($y$132$i) + 8)|0;
         $1493 = ($1492|0)<($1452|0);
         if ($1493) {
          $dc_qi$030$i = $dc_qi$1$lcssa$i;$dst$128$i = $1490;$src$126$i = $1491;$variance$031$i = $variance$1$lcssa$i;$y$132$i = $1492;
         } else {
          $dc_qi$1$lcssa$i$lcssa = $dc_qi$1$lcssa$i;$variance$1$lcssa$i$lcssa = $variance$1$lcssa$i;
          break;
         }
        }
        $1494 = Math_imul($1468, $1431)|0;
        $1495 = (($y$0$lcssa$i) + 8)|0;
        $1496 = (($1495) + ($1467))|0;
        $$sum65$i = (($1469) + ($1497))|0;
        $scevgep60$i = (($1439) + ($$sum65$i)|0);
        $$sum66$i = (($1494) + ($1498))|0;
        $scevgep61$i = (($1435) + ($$sum66$i)|0);
        $dc_qi$0$lcssa$i = $dc_qi$1$lcssa$i$lcssa;$dst$1$lcssa$i = $scevgep61$i;$src$1$lcssa$i = $scevgep60$i;$variance$0$lcssa$i = $variance$1$lcssa$i$lcssa;$y$1$lcssa$i = $1496;
       } else {
        $dc_qi$0$lcssa$i = $1414;$dst$1$lcssa$i = $dst$0$lcssa$i;$src$1$lcssa$i = $src$0$lcssa$i;$variance$0$lcssa$i = $1412;$y$1$lcssa$i = $y$0$lcssa$i;
       }
       do {
        if (!($1418)) {
         $1499 = (((((($_dec)) + 53976|0) + ($pli$0109<<4)|0)) + 4|0);
         $1500 = HEAP32[$1499>>2]|0;
         $1501 = ($1500|0)>($y$1$lcssa$i|0);
         if ($1501) {
          $1502 = (($1500) - ($y$1$lcssa$i))|0;
          $dst$214$i = $dst$1$lcssa$i;$src$213$i = $src$1$lcssa$i;$y$215$i = $y$1$lcssa$i;
          while(1) {
           _memcpy(($dst$214$i|0),($src$213$i|0),($1443|0))|0;
           $1506 = (($dst$214$i) + ($1431)|0);
           $1507 = (($src$213$i) + ($1433)|0);
           $1508 = (($y$215$i) + 1)|0;
           $exitcond$i60 = ($1508|0)==($1500|0);
           if ($exitcond$i60) {
            break;
           } else {
            $dst$214$i = $1506;$src$213$i = $1507;$y$215$i = $1508;
           }
          }
          $1503 = Math_imul($1502, $1431)|0;
          $scevgep$i = (($dst$1$lcssa$i) + ($1503)|0);
          $dst$2$lcssa$i = $scevgep$i;
         } else {
          $dst$2$lcssa$i = $dst$1$lcssa$i;
         }
         $1504 = ($1443|0)>(8);
         if (!($1504)) {
          break;
         }
         $1505 = $1431 << 3;
         $$sum1$i58 = (-4 - ($1505))|0;
         $dc_qi$0$lcssa$pn$i = $dc_qi$0$lcssa$i;$variance$210$i = $variance$0$lcssa$i;$x$111$i = 8;
         while(1) {
          $dc_qi$212$i = ((($dc_qi$0$lcssa$pn$i)) + 1|0);
          $1509 = HEAP8[$dc_qi$212$i>>0]|0;
          $1510 = $1509&255;
          $1511 = (((($_dec)) + 53448|0) + ($1510<<2)|0);
          $1512 = HEAP32[$1511>>2]|0;
          $1513 = ($1512*3)|0;
          $1514 = $1513 >> 2;
          $$sum2$i61 = (($$sum1$i58) + ($x$111$i))|0;
          $1515 = (($dst$2$lcssa$i) + ($$sum2$i61)|0);
          $1516 = ((($variance$210$i)) + 4|0);
          _oc_filter_vedge($1515,$1431,$1512,$1514,$variance$210$i);
          $1517 = (($x$111$i) + 8)|0;
          $1518 = ($1517|0)<($1443|0);
          if ($1518) {
           $dc_qi$0$lcssa$pn$i = $dc_qi$212$i;$variance$210$i = $1516;$x$111$i = $1517;
          } else {
           break;
          }
         }
        }
       } while(0);
       $1519 = HEAP32[$$pre$phi153Z2D>>2]|0;
       $1520 = (($1396) + 3)|0;
       $1521 = ($1519|0)<($1520|0);
       if ($1521) {
        $edelay$1 = $1401;$sdelay$1 = $1400;
        break;
       }
       $1522 = (($1400) + ($notstart$0111))|0;
       $1523 = (($1401) + ($1109))|0;
       $1524 = HEAP32[$1117>>2]|0;
       $1525 = (($1524) - ($1522))|0;
       $1526 = HEAP32[$1125>>2]|0;
       $1527 = (($1526) - ($1523))|0;
       $1528 = HEAP32[$1126>>2]|0;
       $1529 = HEAP32[$1407>>2]|0;
       $1530 = Math_imul($1528, $1525)|0;
       $1531 = (($1529) + ($1530))|0;
       $1532 = HEAP32[$976>>2]|0;
       $1533 = $1110 ? 7 : 4;
       $1534 = ($1532|0)>=($1533|0);
       $1535 = $1110 ? 3840 : 1920;
       $1536 = $1525 << 3;
       $1537 = HEAP32[$1430>>2]|0;
       $1538 = $1527 << 3;
       $1539 = HEAP32[$1442>>2]|0;
       $1540 = (((((($_dec)) + 53976|0) + ($pli$0109<<4)|0)) + 4|0);
       $1541 = HEAP32[$1540>>2]|0;
       $1542 = ($1536|0)<($1538|0);
       if (!($1542)) {
        $edelay$1 = $1523;$sdelay$1 = $1522;
        break;
       }
       $1543 = HEAP32[$1434>>2]|0;
       $1544 = Math_imul($1537, $1536)|0;
       $1545 = (($1543) + ($1544)|0);
       $1546 = HEAP32[$861>>2]|0;
       $1547 = (($1546) + ($1531<<2)|0);
       $1548 = HEAP32[$$pre$phi155Z2D>>2]|0;
       $1549 = (($1548) + ($1531<<2)|0);
       $1550 = ($1539|0)>(0);
       $1551 = $1537 << 3;
       $1552 = (0 - ($1528))|0;
       $1553 = (($1539) + -1)|0;
       $1554 = $1553 >>> 3;
       $1555 = (($1554) + 1)|0;
       $frag$010$i = $1547;$idata$08$i = $1545;$variance$09$i = $1549;$y$07$i = $1536;
       while(1) {
        if ($1550) {
         $1556 = ($y$07$i|0)<(1);
         $1557 = $1556&1;
         $1558 = $1557 << 2;
         $1559 = (($y$07$i) + 8)|0;
         $1560 = ($1559|0)>=($1541|0);
         $1561 = $1560&1;
         $1562 = $1561 << 3;
         $1563 = $1562 | $1558;
         $frag$15$i = $frag$010$i;$variance$12$i = $variance$09$i;$x$01$i = 0;
         while(1) {
          $1564 = HEAP32[$frag$15$i>>2]|0;
          $1565 = $1564 >>> 2;
          $1566 = $1565 & 63;
          $1567 = (((($_dec)) + 795|0) + ($1566)|0);
          $1568 = HEAP8[$1567>>0]|0;
          $1569 = $1568&255;
          $1570 = HEAP32[$variance$12$i>>2]|0;
          $1571 = ($x$01$i|0)<(1);
          $1572 = $1571&1;
          $1573 = (($x$01$i) + 8)|0;
          $1574 = ($1573|0)>=($1539|0);
          $1575 = $1574&1;
          $1576 = $1575 << 1;
          $1577 = $1563 | $1572;
          $1578 = $1577 | $1576;
          $1579 = ($1570|0)>($1535|0);
          $or$cond$i67 = $1534 & $1579;
          L500: do {
           if ($or$cond$i67) {
            $1580 = (($idata$08$i) + ($x$01$i)|0);
            $1581 = (((($_dec)) + 53448|0) + ($1569<<2)|0);
            $1582 = HEAP32[$1581>>2]|0;
            $1583 = (((($_dec)) + 53704|0) + ($1569<<2)|0);
            $1584 = HEAP32[$1583>>2]|0;
            _oc_dering_block($1580,$1537,$1578,$1582,$1584,1);
            do {
             if (!($1110)) {
              if (!($1571)) {
               $1585 = ((($variance$12$i)) + -4|0);
               $1586 = HEAP32[$1585>>2]|0;
               $1587 = ($1586|0)>(3840);
               if ($1587) {
                break;
               }
              }
              if (!($1574)) {
               $1588 = ((($variance$12$i)) + 4|0);
               $1589 = HEAP32[$1588>>2]|0;
               $1590 = ($1589|0)>(3840);
               if ($1590) {
                break;
               }
              }
              if (!($1556)) {
               $1591 = (($variance$12$i) + ($1552<<2)|0);
               $1592 = HEAP32[$1591>>2]|0;
               $1593 = ($1592|0)>(3840);
               if ($1593) {
                break;
               }
              }
              if ($1560) {
               break L500;
              }
              $1594 = (($variance$12$i) + ($1528<<2)|0);
              $1595 = HEAP32[$1594>>2]|0;
              $1596 = ($1595|0)>(3840);
              if (!($1596)) {
               break L500;
              }
             }
            } while(0);
            $1597 = HEAP32[$1581>>2]|0;
            $1598 = HEAP32[$1583>>2]|0;
            _oc_dering_block($1580,$1537,$1578,$1597,$1598,1);
            $1599 = HEAP32[$1581>>2]|0;
            $1600 = HEAP32[$1583>>2]|0;
            _oc_dering_block($1580,$1537,$1578,$1599,$1600,1);
           } else {
            $1601 = ($1570|0)>(1536);
            if ($1601) {
             $1602 = (($idata$08$i) + ($x$01$i)|0);
             $1603 = (((($_dec)) + 53448|0) + ($1569<<2)|0);
             $1604 = HEAP32[$1603>>2]|0;
             $1605 = (((($_dec)) + 53704|0) + ($1569<<2)|0);
             $1606 = HEAP32[$1605>>2]|0;
             _oc_dering_block($1602,$1537,$1578,$1604,$1606,1);
             break;
            }
            $1607 = ($1570|0)>(384);
            if (!($1607)) {
             break;
            }
            $1608 = (($idata$08$i) + ($x$01$i)|0);
            $1609 = (((($_dec)) + 53448|0) + ($1569<<2)|0);
            $1610 = HEAP32[$1609>>2]|0;
            $1611 = (((($_dec)) + 53704|0) + ($1569<<2)|0);
            $1612 = HEAP32[$1611>>2]|0;
            _oc_dering_block($1608,$1537,$1578,$1610,$1612,0);
           }
          } while(0);
          $1613 = ((($frag$15$i)) + 4|0);
          $1614 = ((($variance$12$i)) + 4|0);
          $1615 = ($1573|0)<($1539|0);
          if ($1615) {
           $frag$15$i = $1613;$variance$12$i = $1614;$x$01$i = $1573;
          } else {
           break;
          }
         }
         $scevgep$i65 = (($variance$09$i) + ($1555<<2)|0);
         $scevgep13$i = (($frag$010$i) + ($1555<<2)|0);
         $$pre$phi$iZ2D = $1559;$frag$1$lcssa$i = $scevgep13$i;$variance$1$lcssa$i69 = $scevgep$i65;
        } else {
         $$pre$i64 = (($y$07$i) + 8)|0;
         $$pre$phi$iZ2D = $$pre$i64;$frag$1$lcssa$i = $frag$010$i;$variance$1$lcssa$i69 = $variance$09$i;
        }
        $1616 = (($idata$08$i) + ($1551)|0);
        $1617 = ($$pre$phi$iZ2D|0)<($1538|0);
        if ($1617) {
         $frag$010$i = $frag$1$lcssa$i;$idata$08$i = $1616;$variance$09$i = $variance$1$lcssa$i69;$y$07$i = $$pre$phi$iZ2D;
        } else {
         $edelay$1 = $1523;$sdelay$1 = $1522;
         break;
        }
       }
      }
     } while(0);
     $1622 = HEAP32[$1117>>2]|0;
     $1623 = (($1622) - ($sdelay$1))|0;
     $1624 = $1623 << $1114;
     $1625 = ($1624|0)<($avail_fragy0$0107|0);
     $1626 = $1625 ? $1624 : $avail_fragy0$0107;
     $1627 = HEAP32[$1125>>2]|0;
     $1628 = (($1627) - ($edelay$1))|0;
     $1629 = $1628 << $1114;
     $1630 = ($1629|0)<($avail_fragy_end$0108|0);
     $1631 = $1630 ? $1629 : $avail_fragy_end$0108;
     $1632 = (($pli$0109) + 1)|0;
     $exitcond = ($1632|0)==(3);
     if ($exitcond) {
      $$lcssa251 = $1626;$$lcssa252 = $1631;
      break;
     } else {
      $avail_fragy0$0107 = $1626;$avail_fragy_end$0108 = $1631;$pli$0109 = $1632;
     }
    }
    $1633 = HEAP32[$1098>>2]|0;
    $1634 = ($1633|0)==(0|0);
    if (!($1634)) {
     _oc_restore_fpu($1086);
     $1635 = HEAP32[$1098>>2]|0;
     $1636 = HEAP32[$1099>>2]|0;
     $1637 = HEAP32[$1095>>2]|0;
     $1638 = (($1637) - ($$lcssa252))|0;
     $1639 = (($1637) - ($$lcssa251))|0;
     FUNCTION_TABLE_viiii[$1635 & 8191]($1636,$stripe_buf,$1638,$1639);
    }
    $1640 = HEAP32[$911>>2]|0;
    $1641 = (($1640) + ($stripe_fragy$0112))|0;
    if ($1108) {
     $1107 = $1640;$notstart$0111 = 1;$stripe_fragy$0112 = $1641;
    } else {
     break;
    }
   }
   _oc_state_borders_fill_caps($1086,$refi$1,0);
   _oc_state_borders_fill_caps($1086,$refi$1,1);
   _oc_state_borders_fill_caps($1086,$refi$1,2);
   $1100 = HEAP8[$11>>0]|0;
   $1101 = ($1100<<24>>24)==(0);
   $1102 = ((($_dec)) + 280|0);
   $1103 = HEAP32[$1102>>2]|0;
   $1104 = ((($_dec)) + 276|0);
   HEAP32[$1104>>2] = $1103;
   if ($1101) {
    $1642 = ((($_dec)) + 272|0);
    HEAP32[$1642>>2] = $1103;
   }
   _oc_restore_fpu($1086);
   $$0 = 0;
  } else {
   $$0 = -24;
  }
 } while(0);
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0);
}
function _oc_sb_run_unpack($_opb) {
 $_opb = $_opb|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_oc_pack_read1($_opb)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  $$0 = 1;
  return ($$0|0);
 }
 $2 = (_oc_pack_read($_opb,2)|0);
 $3 = $2 & 2;
 $4 = ($3|0)==(0);
 if ($4) {
  $5 = (($2) + 2)|0;
  $$0 = $5;
  return ($$0|0);
 }
 $6 = $2 & 1;
 $7 = ($6|0)==(0);
 if ($7) {
  $8 = (_oc_pack_read1($_opb)|0);
  $9 = (($8) + 4)|0;
  $$0 = $9;
  return ($$0|0);
 }
 $10 = (_oc_pack_read($_opb,3)|0);
 $11 = $10 & 4;
 $12 = ($11|0)==(0);
 if ($12) {
  $13 = (($10) + 6)|0;
  $$0 = $13;
  return ($$0|0);
 }
 $14 = $10 & 2;
 $15 = ($14|0)==(0);
 $16 = $10 & 1;
 if ($15) {
  $17 = $16 << 2;
  $18 = $17 | 10;
  $19 = (_oc_pack_read($_opb,2)|0);
  $20 = (($19) + ($18))|0;
  $$0 = $20;
  return ($$0|0);
 }
 $21 = ($16|0)==(0);
 if ($21) {
  $22 = (_oc_pack_read($_opb,4)|0);
  $23 = (($22) + 18)|0;
  $$0 = $23;
  return ($$0|0);
 } else {
  $24 = (_oc_pack_read($_opb,12)|0);
  $25 = (($24) + 34)|0;
  $$0 = $25;
  return ($$0|0);
 }
 return (0)|0;
}
function _oc_clc_mode_unpack($_opb) {
 $_opb = $_opb|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_oc_pack_read($_opb,3)|0);
 return ($0|0);
}
function _oc_vlc_mode_unpack($_opb) {
 $_opb = $_opb|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_oc_pack_read1($_opb)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  $i$0$lcssa = 0;
 } else {
  $2 = (_oc_pack_read1($_opb)|0);
  $3 = ($2|0)==(0);
  if ($3) {
   $i$0$lcssa = 1;
  } else {
   $4 = (_oc_pack_read1($_opb)|0);
   $5 = ($4|0)==(0);
   if ($5) {
    $i$0$lcssa = 2;
   } else {
    $6 = (_oc_pack_read1($_opb)|0);
    $7 = ($6|0)==(0);
    if ($7) {
     $i$0$lcssa = 3;
    } else {
     $8 = (_oc_pack_read1($_opb)|0);
     $9 = ($8|0)==(0);
     if ($9) {
      $i$0$lcssa = 4;
     } else {
      $10 = (_oc_pack_read1($_opb)|0);
      $11 = ($10|0)==(0);
      if ($11) {
       $i$0$lcssa = 5;
      } else {
       $12 = (_oc_pack_read1($_opb)|0);
       $13 = ($12|0)==(0);
       $$ = $13 ? 6 : 7;
       return ($$|0);
      }
     }
    }
   }
  }
 }
 return ($i$0$lcssa|0);
}
function _oc_clc_mv_comp_unpack($_opb) {
 $_opb = $_opb|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_oc_pack_read($_opb,6)|0);
 $1 = $0 >> 1;
 $2 = $0 & 1;
 $3 = (0 - ($2))|0;
 $4 = (($1) - ($2))|0;
 $5 = $4 ^ $3;
 return ($5|0);
}
function _oc_vlc_mv_comp_unpack($_opb) {
 $_opb = $_opb|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $bits$0 = 0, $mv$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_oc_pack_read($_opb,3)|0);
 switch ($0|0) {
 case 2:  {
  $$0 = -1;
  return ($$0|0);
  break;
 }
 case 4: case 3:  {
  $1 = (($0) + -1)|0;
  $2 = (_oc_pack_read1($_opb)|0);
  $bits$0 = $2;$mv$0 = $1;
  break;
 }
 case 1: case 0:  {
  $$0 = $0;
  return ($$0|0);
  break;
 }
 default: {
  $3 = (($0) + -3)|0;
  $4 = 1 << $3;
  $5 = (($0) + -2)|0;
  $6 = (_oc_pack_read($_opb,$5)|0);
  $7 = $6 >> 1;
  $8 = (($7) + ($4))|0;
  $9 = $6 & 1;
  $bits$0 = $9;$mv$0 = $8;
 }
 }
 $10 = (0 - ($bits$0))|0;
 $11 = (($mv$0) - ($bits$0))|0;
 $12 = $11 ^ $10;
 $$0 = $12;
 return ($$0|0);
}
function _oc_filter_hedge($_dst,$_dst_ystride,$_src,$_src_ystride,$_qstep,$_flimit,$_variance0,$_variance1) {
 $_dst = $_dst|0;
 $_dst_ystride = $_dst_ystride|0;
 $_src = $_src|0;
 $_src_ystride = $_src_ystride|0;
 $_qstep = $_qstep|0;
 $_flimit = $_flimit|0;
 $_variance0 = $_variance0|0;
 $_variance1 = $_variance1|0;
 var $$sum = 0, $$sum32 = 0, $$sum33 = 0, $$sum34 = 0, $$sum35 = 0, $$sum36 = 0, $$sum37 = 0, $$sum38 = 0, $$sum39 = 0, $$sum40 = 0, $$sum41 = 0, $$sum42 = 0, $$sum43 = 0, $$sum44 = 0, $$sum45 = 0, $$sum46 = 0, $$sum47 = 0, $$sum48 = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0;
 var $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $bx$022 = 0, $exitcond = 0, $indvars$iv = 0, $indvars$iv30 = 0, $ispos = 0, $ispos$1 = 0, $ispos$2 = 0, $ispos$3 = 0, $ispos1 = 0, $ispos1$1 = 0, $ispos1$2 = 0, $ispos1$3 = 0, $neg = 0, $neg$1 = 0, $neg$2 = 0, $neg$3 = 0, $neg2 = 0, $neg2$1 = 0, $neg2$2 = 0;
 var $neg2$3 = 0, $or$cond = 0, $or$cond4 = 0, $r = 0, $rdst$019 = 0, $rsrc$021 = 0, $scevgep = 0, $scevgep28 = 0, $scevgep29 = 0, $scevgep31 = 0, $tmp = 0, $tmp5 = 0, $tmp6 = 0, $tmp7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $r = sp;
 $0 = ((($r)) + 20|0);
 $1 = ((($r)) + 16|0);
 $2 = ((($r)) + 4|0);
 $3 = ((($r)) + 8|0);
 $4 = ((($r)) + 12|0);
 $$sum = $_dst_ystride << 1;
 $5 = ((($r)) + 24|0);
 $6 = ((($r)) + 28|0);
 $7 = ((($r)) + 32|0);
 $8 = ((($r)) + 36|0);
 $9 = ($_dst_ystride*6)|0;
 $scevgep = (($_dst) + ($9)|0);
 $10 = ($_dst_ystride*5)|0;
 $scevgep29 = (($_dst) + ($10)|0);
 $$sum32 = $_src_ystride << 1;
 $$sum33 = ($_src_ystride*3)|0;
 $$sum34 = $_src_ystride << 2;
 $$sum35 = ($_src_ystride*5)|0;
 $$sum36 = ($_src_ystride*6)|0;
 $$sum37 = ($_src_ystride*7)|0;
 $$sum38 = $_src_ystride << 3;
 $$sum39 = ($_src_ystride*9)|0;
 $$sum46 = ($_dst_ystride*3)|0;
 $$sum47 = $_dst_ystride << 2;
 $$sum48 = ($_dst_ystride*5)|0;
 $$sum40 = $_dst_ystride << 1;
 $$sum41 = ($_dst_ystride*3)|0;
 $$sum42 = $_dst_ystride << 2;
 $$sum43 = ($_dst_ystride*5)|0;
 $$sum44 = ($_dst_ystride*6)|0;
 $$sum45 = ($_dst_ystride*7)|0;
 $bx$022 = 0;$indvars$iv = $scevgep;$indvars$iv30 = $scevgep29;$rdst$019 = $_dst;$rsrc$021 = $_src;
 while(1) {
  $11 = HEAP8[$rsrc$021>>0]|0;
  $12 = $11&255;
  HEAP32[$r>>2] = $12;
  $13 = (($rsrc$021) + ($_src_ystride)|0);
  $14 = HEAP8[$13>>0]|0;
  $15 = $14&255;
  HEAP32[$2>>2] = $15;
  $16 = (($rsrc$021) + ($$sum32)|0);
  $17 = HEAP8[$16>>0]|0;
  $18 = $17&255;
  HEAP32[$3>>2] = $18;
  $19 = (($rsrc$021) + ($$sum33)|0);
  $20 = HEAP8[$19>>0]|0;
  $21 = $20&255;
  HEAP32[$4>>2] = $21;
  $22 = (($rsrc$021) + ($$sum34)|0);
  $23 = HEAP8[$22>>0]|0;
  $24 = $23&255;
  HEAP32[$1>>2] = $24;
  $25 = (($rsrc$021) + ($$sum35)|0);
  $26 = HEAP8[$25>>0]|0;
  $27 = $26&255;
  HEAP32[$0>>2] = $27;
  $28 = (($rsrc$021) + ($$sum36)|0);
  $29 = HEAP8[$28>>0]|0;
  $30 = $29&255;
  HEAP32[$5>>2] = $30;
  $31 = (($rsrc$021) + ($$sum37)|0);
  $32 = HEAP8[$31>>0]|0;
  $33 = $32&255;
  HEAP32[$6>>2] = $33;
  $34 = (($rsrc$021) + ($$sum38)|0);
  $35 = HEAP8[$34>>0]|0;
  $36 = $35&255;
  HEAP32[$7>>2] = $36;
  $37 = (($rsrc$021) + ($$sum39)|0);
  $38 = HEAP8[$37>>0]|0;
  $39 = $38&255;
  HEAP32[$8>>2] = $39;
  $40 = (($15) - ($12))|0;
  $ispos = ($40|0)>(-1);
  $neg = (0 - ($40))|0;
  $41 = $ispos ? $40 : $neg;
  $42 = (($27) - ($30))|0;
  $ispos1 = ($42|0)>(-1);
  $neg2 = (0 - ($42))|0;
  $43 = $ispos1 ? $42 : $neg2;
  $44 = (($18) - ($15))|0;
  $ispos$1 = ($44|0)>(-1);
  $neg$1 = (0 - ($44))|0;
  $45 = $ispos$1 ? $44 : $neg$1;
  $46 = (($45) + ($41))|0;
  $47 = (($30) - ($33))|0;
  $ispos1$1 = ($47|0)>(-1);
  $neg2$1 = (0 - ($47))|0;
  $48 = $ispos1$1 ? $47 : $neg2$1;
  $49 = (($48) + ($43))|0;
  $50 = (($21) - ($18))|0;
  $ispos$2 = ($50|0)>(-1);
  $neg$2 = (0 - ($50))|0;
  $51 = $ispos$2 ? $50 : $neg$2;
  $52 = (($51) + ($46))|0;
  $53 = (($33) - ($36))|0;
  $ispos1$2 = ($53|0)>(-1);
  $neg2$2 = (0 - ($53))|0;
  $54 = $ispos1$2 ? $53 : $neg2$2;
  $55 = (($54) + ($49))|0;
  $56 = (($24) - ($21))|0;
  $ispos$3 = ($56|0)>(-1);
  $neg$3 = (0 - ($56))|0;
  $57 = $ispos$3 ? $56 : $neg$3;
  $58 = (($57) + ($52))|0;
  $59 = (($36) - ($39))|0;
  $ispos1$3 = ($59|0)>(-1);
  $neg2$3 = (0 - ($59))|0;
  $60 = $ispos1$3 ? $59 : $neg2$3;
  $61 = (($60) + ($55))|0;
  $62 = ($58|0)<(255);
  $63 = $62 ? $58 : 255;
  $64 = HEAP32[$_variance0>>2]|0;
  $65 = (($64) + ($63))|0;
  HEAP32[$_variance0>>2] = $65;
  $66 = ($61|0)<(255);
  $67 = $66 ? $61 : 255;
  $68 = HEAP32[$_variance1>>2]|0;
  $69 = (($68) + ($67))|0;
  HEAP32[$_variance1>>2] = $69;
  $70 = ($58|0)<($_flimit|0);
  $71 = ($61|0)<($_flimit|0);
  $or$cond = $70 & $71;
  if ($or$cond) {
   $93 = (($27) - ($24))|0;
   $94 = ($93|0)<($_qstep|0);
   $95 = (($24) - ($27))|0;
   $96 = ($95|0)<($_qstep|0);
   $or$cond4 = $94 & $96;
   if ($or$cond4) {
    $97 = ($12*3)|0;
    $98 = $15 << 1;
    $99 = (($24) + 4)|0;
    $100 = (($99) + ($97))|0;
    $101 = (($100) + ($98))|0;
    $102 = (($101) + ($18))|0;
    $103 = (($102) + ($21))|0;
    $104 = $103 >>> 3;
    $105 = $104&255;
    HEAP8[$rdst$019>>0] = $105;
    $106 = (($rdst$019) + ($_dst_ystride)|0);
    $107 = HEAP32[$r>>2]|0;
    $108 = HEAP32[$2>>2]|0;
    $109 = HEAP32[$3>>2]|0;
    $110 = HEAP32[$4>>2]|0;
    $111 = HEAP32[$1>>2]|0;
    $112 = HEAP32[$0>>2]|0;
    $tmp = (($109) + ($107))|0;
    $tmp5 = $tmp << 1;
    $113 = (($108) + 4)|0;
    $114 = (($113) + ($110))|0;
    $115 = (($114) + ($111))|0;
    $116 = (($115) + ($112))|0;
    $117 = (($116) + ($tmp5))|0;
    $118 = $117 >>> 3;
    $119 = $118&255;
    HEAP8[$106>>0] = $119;
    $120 = (($rdst$019) + ($$sum)|0);
    $121 = HEAP32[$r>>2]|0;
    $122 = HEAP32[$2>>2]|0;
    $123 = HEAP32[$3>>2]|0;
    $124 = HEAP32[$4>>2]|0;
    $125 = $124 << 1;
    $126 = HEAP32[$1>>2]|0;
    $127 = HEAP32[$0>>2]|0;
    $128 = HEAP32[$5>>2]|0;
    $129 = (($121) + 4)|0;
    $130 = (($129) + ($122))|0;
    $131 = (($130) + ($123))|0;
    $132 = (($131) + ($125))|0;
    $133 = (($132) + ($126))|0;
    $134 = (($133) + ($127))|0;
    $135 = (($134) + ($128))|0;
    $136 = $135 >>> 3;
    $137 = $136&255;
    HEAP8[$120>>0] = $137;
    $138 = (($rdst$019) + ($$sum46)|0);
    $139 = HEAP32[$2>>2]|0;
    $140 = HEAP32[$3>>2]|0;
    $141 = HEAP32[$4>>2]|0;
    $142 = HEAP32[$1>>2]|0;
    $143 = $142 << 1;
    $144 = HEAP32[$0>>2]|0;
    $145 = HEAP32[$5>>2]|0;
    $146 = HEAP32[$6>>2]|0;
    $147 = (($139) + 4)|0;
    $148 = (($147) + ($140))|0;
    $149 = (($148) + ($141))|0;
    $150 = (($149) + ($143))|0;
    $151 = (($150) + ($144))|0;
    $152 = (($151) + ($145))|0;
    $153 = (($152) + ($146))|0;
    $154 = $153 >>> 3;
    $155 = $154&255;
    HEAP8[$138>>0] = $155;
    $156 = (($rdst$019) + ($$sum47)|0);
    $157 = HEAP32[$3>>2]|0;
    $158 = HEAP32[$4>>2]|0;
    $159 = HEAP32[$1>>2]|0;
    $160 = HEAP32[$0>>2]|0;
    $161 = $160 << 1;
    $162 = HEAP32[$5>>2]|0;
    $163 = HEAP32[$6>>2]|0;
    $164 = HEAP32[$7>>2]|0;
    $165 = (($157) + 4)|0;
    $166 = (($165) + ($158))|0;
    $167 = (($166) + ($159))|0;
    $168 = (($167) + ($161))|0;
    $169 = (($168) + ($162))|0;
    $170 = (($169) + ($163))|0;
    $171 = (($170) + ($164))|0;
    $172 = $171 >>> 3;
    $173 = $172&255;
    HEAP8[$156>>0] = $173;
    $174 = (($rdst$019) + ($$sum48)|0);
    $175 = HEAP32[$4>>2]|0;
    $176 = HEAP32[$1>>2]|0;
    $177 = HEAP32[$0>>2]|0;
    $178 = HEAP32[$5>>2]|0;
    $179 = $178 << 1;
    $180 = HEAP32[$6>>2]|0;
    $181 = HEAP32[$7>>2]|0;
    $182 = HEAP32[$8>>2]|0;
    $183 = (($175) + 4)|0;
    $184 = (($183) + ($176))|0;
    $185 = (($184) + ($177))|0;
    $186 = (($185) + ($179))|0;
    $187 = (($186) + ($180))|0;
    $188 = (($187) + ($181))|0;
    $189 = (($188) + ($182))|0;
    $190 = $189 >>> 3;
    $191 = $190&255;
    HEAP8[$174>>0] = $191;
    $192 = HEAP32[$1>>2]|0;
    $193 = HEAP32[$0>>2]|0;
    $194 = HEAP32[$5>>2]|0;
    $195 = HEAP32[$6>>2]|0;
    $196 = HEAP32[$7>>2]|0;
    $197 = HEAP32[$8>>2]|0;
    $tmp6 = (($197) + ($195))|0;
    $tmp7 = $tmp6 << 1;
    $198 = (($192) + 4)|0;
    $199 = (($198) + ($193))|0;
    $200 = (($199) + ($194))|0;
    $201 = (($200) + ($196))|0;
    $202 = (($201) + ($tmp7))|0;
    $203 = $202 >>> 3;
    $204 = $203&255;
    HEAP8[$indvars$iv>>0] = $204;
    $205 = (($indvars$iv30) + ($$sum)|0);
    $206 = HEAP32[$0>>2]|0;
    $207 = HEAP32[$5>>2]|0;
    $208 = HEAP32[$6>>2]|0;
    $209 = HEAP32[$7>>2]|0;
    $210 = $209 << 1;
    $211 = HEAP32[$8>>2]|0;
    $212 = ($211*3)|0;
    $213 = (($206) + 4)|0;
    $214 = (($213) + ($207))|0;
    $215 = (($214) + ($208))|0;
    $216 = (($215) + ($210))|0;
    $217 = (($216) + ($212))|0;
    $218 = $217 >>> 3;
    $219 = $218&255;
    HEAP8[$205>>0] = $219;
   } else {
    label = 3;
   }
  } else {
   label = 3;
  }
  if ((label|0) == 3) {
   label = 0;
   HEAP8[$rdst$019>>0] = $14;
   $72 = (($rdst$019) + ($_dst_ystride)|0);
   $73 = HEAP32[$3>>2]|0;
   $74 = $73&255;
   HEAP8[$72>>0] = $74;
   $75 = (($rdst$019) + ($$sum40)|0);
   $76 = HEAP32[$4>>2]|0;
   $77 = $76&255;
   HEAP8[$75>>0] = $77;
   $78 = (($rdst$019) + ($$sum41)|0);
   $79 = HEAP32[$1>>2]|0;
   $80 = $79&255;
   HEAP8[$78>>0] = $80;
   $81 = (($rdst$019) + ($$sum42)|0);
   $82 = HEAP32[$0>>2]|0;
   $83 = $82&255;
   HEAP8[$81>>0] = $83;
   $84 = (($rdst$019) + ($$sum43)|0);
   $85 = HEAP32[$5>>2]|0;
   $86 = $85&255;
   HEAP8[$84>>0] = $86;
   $87 = (($rdst$019) + ($$sum44)|0);
   $88 = HEAP32[$6>>2]|0;
   $89 = $88&255;
   HEAP8[$87>>0] = $89;
   $90 = (($rdst$019) + ($$sum45)|0);
   $91 = HEAP32[$7>>2]|0;
   $92 = $91&255;
   HEAP8[$90>>0] = $92;
  }
  $220 = ((($rdst$019)) + 1|0);
  $221 = ((($rsrc$021)) + 1|0);
  $222 = (($bx$022) + 1)|0;
  $scevgep28 = ((($indvars$iv)) + 1|0);
  $scevgep31 = ((($indvars$iv30)) + 1|0);
  $exitcond = ($222|0)==(8);
  if ($exitcond) {
   break;
  } else {
   $bx$022 = $222;$indvars$iv = $scevgep28;$indvars$iv30 = $scevgep31;$rdst$019 = $220;$rsrc$021 = $221;
  }
 }
 STACKTOP = sp;return;
}
function _oc_filter_vedge($_dst,$_dst_ystride,$_qstep,$_flimit,$_variances) {
 $_dst = $_dst|0;
 $_dst_ystride = $_dst_ystride|0;
 $_qstep = $_qstep|0;
 $_flimit = $_flimit|0;
 $_variances = $_variances|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $by$016 = 0, $cdst$015 = 0, $exitcond = 0, $indvars$iv = 0, $indvars$iv23 = 0, $ispos = 0, $ispos$1 = 0, $ispos$2 = 0, $ispos$3 = 0, $ispos1 = 0, $ispos1$1 = 0, $ispos1$2 = 0, $ispos1$3 = 0, $neg = 0, $neg$1 = 0, $neg$2 = 0, $neg$3 = 0, $neg2 = 0, $neg2$1 = 0, $neg2$2 = 0;
 var $neg2$3 = 0, $or$cond = 0, $or$cond4 = 0, $r = 0, $scevgep = 0, $scevgep2126 = 0, $scevgep22 = 0, $scevgep2527 = 0, $tmp = 0, $tmp5 = 0, $tmp6 = 0, $tmp7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $r = sp;
 $0 = ((($_variances)) + 4|0);
 $1 = ((($r)) + 20|0);
 $2 = ((($r)) + 16|0);
 $3 = ((($r)) + 4|0);
 $4 = ((($r)) + 8|0);
 $5 = ((($r)) + 12|0);
 $6 = ((($r)) + 24|0);
 $7 = ((($r)) + 28|0);
 $8 = ((($r)) + 32|0);
 $9 = ((($r)) + 36|0);
 $scevgep = ((($_dst)) + 6|0);
 $scevgep22 = ((($_dst)) + 5|0);
 $by$016 = 0;$cdst$015 = $_dst;$indvars$iv = $scevgep;$indvars$iv23 = $scevgep22;
 while(1) {
  $10 = ((($cdst$015)) + -1|0);
  $11 = HEAP8[$10>>0]|0;
  $12 = $11&255;
  HEAP32[$r>>2] = $12;
  $13 = ((($cdst$015)) + 1|0);
  $14 = HEAP8[$cdst$015>>0]|0;
  $15 = $14&255;
  HEAP32[$3>>2] = $15;
  $16 = ((($cdst$015)) + 2|0);
  $17 = HEAP8[$13>>0]|0;
  $18 = $17&255;
  HEAP32[$4>>2] = $18;
  $19 = ((($cdst$015)) + 3|0);
  $20 = HEAP8[$16>>0]|0;
  $21 = $20&255;
  HEAP32[$5>>2] = $21;
  $22 = ((($cdst$015)) + 4|0);
  $23 = HEAP8[$19>>0]|0;
  $24 = $23&255;
  HEAP32[$2>>2] = $24;
  $25 = ((($cdst$015)) + 5|0);
  $26 = HEAP8[$22>>0]|0;
  $27 = $26&255;
  HEAP32[$1>>2] = $27;
  $28 = ((($cdst$015)) + 6|0);
  $29 = HEAP8[$25>>0]|0;
  $30 = $29&255;
  HEAP32[$6>>2] = $30;
  $31 = ((($cdst$015)) + 7|0);
  $32 = HEAP8[$28>>0]|0;
  $33 = $32&255;
  HEAP32[$7>>2] = $33;
  $34 = ((($cdst$015)) + 8|0);
  $35 = HEAP8[$31>>0]|0;
  $36 = $35&255;
  HEAP32[$8>>2] = $36;
  $37 = HEAP8[$34>>0]|0;
  $38 = $37&255;
  HEAP32[$9>>2] = $38;
  $39 = (($15) - ($12))|0;
  $ispos = ($39|0)>(-1);
  $neg = (0 - ($39))|0;
  $40 = $ispos ? $39 : $neg;
  $41 = (($27) - ($30))|0;
  $ispos1 = ($41|0)>(-1);
  $neg2 = (0 - ($41))|0;
  $42 = $ispos1 ? $41 : $neg2;
  $43 = (($18) - ($15))|0;
  $ispos$1 = ($43|0)>(-1);
  $neg$1 = (0 - ($43))|0;
  $44 = $ispos$1 ? $43 : $neg$1;
  $45 = (($44) + ($40))|0;
  $46 = (($30) - ($33))|0;
  $ispos1$1 = ($46|0)>(-1);
  $neg2$1 = (0 - ($46))|0;
  $47 = $ispos1$1 ? $46 : $neg2$1;
  $48 = (($47) + ($42))|0;
  $49 = (($21) - ($18))|0;
  $ispos$2 = ($49|0)>(-1);
  $neg$2 = (0 - ($49))|0;
  $50 = $ispos$2 ? $49 : $neg$2;
  $51 = (($50) + ($45))|0;
  $52 = (($33) - ($36))|0;
  $ispos1$2 = ($52|0)>(-1);
  $neg2$2 = (0 - ($52))|0;
  $53 = $ispos1$2 ? $52 : $neg2$2;
  $54 = (($53) + ($48))|0;
  $55 = (($24) - ($21))|0;
  $ispos$3 = ($55|0)>(-1);
  $neg$3 = (0 - ($55))|0;
  $56 = $ispos$3 ? $55 : $neg$3;
  $57 = (($56) + ($51))|0;
  $58 = (($36) - ($38))|0;
  $ispos1$3 = ($58|0)>(-1);
  $neg2$3 = (0 - ($58))|0;
  $59 = $ispos1$3 ? $58 : $neg2$3;
  $60 = (($59) + ($54))|0;
  $61 = ($57|0)<(255);
  $62 = $61 ? $57 : 255;
  $63 = HEAP32[$_variances>>2]|0;
  $64 = (($63) + ($62))|0;
  HEAP32[$_variances>>2] = $64;
  $65 = ($60|0)<(255);
  $66 = $65 ? $60 : 255;
  $67 = HEAP32[$0>>2]|0;
  $68 = (($67) + ($66))|0;
  HEAP32[$0>>2] = $68;
  $69 = ($57|0)<($_flimit|0);
  $70 = ($60|0)<($_flimit|0);
  $or$cond = $69 & $70;
  if ($or$cond) {
   $71 = (($27) - ($24))|0;
   $72 = ($71|0)<($_qstep|0);
   $73 = (($24) - ($27))|0;
   $74 = ($73|0)<($_qstep|0);
   $or$cond4 = $72 & $74;
   if ($or$cond4) {
    $75 = ($12*3)|0;
    $76 = $15 << 1;
    $77 = (($24) + 4)|0;
    $78 = (($77) + ($75))|0;
    $79 = (($78) + ($76))|0;
    $80 = (($79) + ($18))|0;
    $81 = (($80) + ($21))|0;
    $82 = $81 >>> 3;
    $83 = $82&255;
    $84 = ((($cdst$015)) + 1|0);
    HEAP8[$cdst$015>>0] = $83;
    $85 = HEAP32[$r>>2]|0;
    $86 = HEAP32[$3>>2]|0;
    $87 = HEAP32[$4>>2]|0;
    $88 = HEAP32[$5>>2]|0;
    $89 = HEAP32[$2>>2]|0;
    $90 = HEAP32[$1>>2]|0;
    $tmp = (($87) + ($85))|0;
    $tmp5 = $tmp << 1;
    $91 = (($86) + 4)|0;
    $92 = (($91) + ($88))|0;
    $93 = (($92) + ($89))|0;
    $94 = (($93) + ($90))|0;
    $95 = (($94) + ($tmp5))|0;
    $96 = $95 >>> 3;
    $97 = $96&255;
    $98 = ((($cdst$015)) + 2|0);
    HEAP8[$84>>0] = $97;
    $99 = HEAP32[$r>>2]|0;
    $100 = HEAP32[$3>>2]|0;
    $101 = HEAP32[$4>>2]|0;
    $102 = HEAP32[$5>>2]|0;
    $103 = $102 << 1;
    $104 = HEAP32[$2>>2]|0;
    $105 = HEAP32[$1>>2]|0;
    $106 = HEAP32[$6>>2]|0;
    $107 = (($99) + 4)|0;
    $108 = (($107) + ($100))|0;
    $109 = (($108) + ($101))|0;
    $110 = (($109) + ($103))|0;
    $111 = (($110) + ($104))|0;
    $112 = (($111) + ($105))|0;
    $113 = (($112) + ($106))|0;
    $114 = $113 >>> 3;
    $115 = $114&255;
    $116 = ((($cdst$015)) + 3|0);
    HEAP8[$98>>0] = $115;
    $117 = HEAP32[$3>>2]|0;
    $118 = HEAP32[$4>>2]|0;
    $119 = HEAP32[$5>>2]|0;
    $120 = HEAP32[$2>>2]|0;
    $121 = $120 << 1;
    $122 = HEAP32[$1>>2]|0;
    $123 = HEAP32[$6>>2]|0;
    $124 = HEAP32[$7>>2]|0;
    $125 = (($117) + 4)|0;
    $126 = (($125) + ($118))|0;
    $127 = (($126) + ($119))|0;
    $128 = (($127) + ($121))|0;
    $129 = (($128) + ($122))|0;
    $130 = (($129) + ($123))|0;
    $131 = (($130) + ($124))|0;
    $132 = $131 >>> 3;
    $133 = $132&255;
    $134 = ((($cdst$015)) + 4|0);
    HEAP8[$116>>0] = $133;
    $135 = HEAP32[$4>>2]|0;
    $136 = HEAP32[$5>>2]|0;
    $137 = HEAP32[$2>>2]|0;
    $138 = HEAP32[$1>>2]|0;
    $139 = $138 << 1;
    $140 = HEAP32[$6>>2]|0;
    $141 = HEAP32[$7>>2]|0;
    $142 = HEAP32[$8>>2]|0;
    $143 = (($135) + 4)|0;
    $144 = (($143) + ($136))|0;
    $145 = (($144) + ($137))|0;
    $146 = (($145) + ($139))|0;
    $147 = (($146) + ($140))|0;
    $148 = (($147) + ($141))|0;
    $149 = (($148) + ($142))|0;
    $150 = $149 >>> 3;
    $151 = $150&255;
    $152 = ((($cdst$015)) + 5|0);
    HEAP8[$134>>0] = $151;
    $153 = HEAP32[$5>>2]|0;
    $154 = HEAP32[$2>>2]|0;
    $155 = HEAP32[$1>>2]|0;
    $156 = HEAP32[$6>>2]|0;
    $157 = $156 << 1;
    $158 = HEAP32[$7>>2]|0;
    $159 = HEAP32[$8>>2]|0;
    $160 = HEAP32[$9>>2]|0;
    $161 = (($153) + 4)|0;
    $162 = (($161) + ($154))|0;
    $163 = (($162) + ($155))|0;
    $164 = (($163) + ($157))|0;
    $165 = (($164) + ($158))|0;
    $166 = (($165) + ($159))|0;
    $167 = (($166) + ($160))|0;
    $168 = $167 >>> 3;
    $169 = $168&255;
    HEAP8[$152>>0] = $169;
    $170 = HEAP32[$2>>2]|0;
    $171 = HEAP32[$1>>2]|0;
    $172 = HEAP32[$6>>2]|0;
    $173 = HEAP32[$7>>2]|0;
    $174 = HEAP32[$8>>2]|0;
    $175 = HEAP32[$9>>2]|0;
    $tmp6 = (($175) + ($173))|0;
    $tmp7 = $tmp6 << 1;
    $176 = (($170) + 4)|0;
    $177 = (($176) + ($171))|0;
    $178 = (($177) + ($172))|0;
    $179 = (($178) + ($174))|0;
    $180 = (($179) + ($tmp7))|0;
    $181 = $180 >>> 3;
    $182 = $181&255;
    $183 = ((($indvars$iv23)) + 2|0);
    HEAP8[$indvars$iv>>0] = $182;
    $184 = HEAP32[$1>>2]|0;
    $185 = HEAP32[$6>>2]|0;
    $186 = HEAP32[$7>>2]|0;
    $187 = HEAP32[$8>>2]|0;
    $188 = $187 << 1;
    $189 = HEAP32[$9>>2]|0;
    $190 = ($189*3)|0;
    $191 = (($184) + 4)|0;
    $192 = (($191) + ($185))|0;
    $193 = (($192) + ($186))|0;
    $194 = (($193) + ($188))|0;
    $195 = (($194) + ($190))|0;
    $196 = $195 >>> 3;
    $197 = $196&255;
    HEAP8[$183>>0] = $197;
   }
  }
  $198 = (($cdst$015) + ($_dst_ystride)|0);
  $199 = (($by$016) + 1)|0;
  $scevgep2126 = (($indvars$iv) + ($_dst_ystride)|0);
  $scevgep2527 = (($indvars$iv23) + ($_dst_ystride)|0);
  $exitcond = ($199|0)==(8);
  if ($exitcond) {
   break;
  } else {
   $by$016 = $199;$cdst$015 = $198;$indvars$iv = $scevgep2126;$indvars$iv23 = $scevgep2527;
  }
 }
 STACKTOP = sp;return;
}
function _oc_dering_block($_idata,$_ystride,$_b,$_dc_scale,$_sharp_mod,$_strong) {
 $_idata = $_idata|0;
 $_ystride = $_ystride|0;
 $_b = $_b|0;
 $_dc_scale = $_dc_scale|0;
 $_sharp_mod = $_sharp_mod|0;
 $_strong = $_strong|0;
 var $$lobit = 0, $$lobit1 = 0, $$lobit3 = 0, $$lobit5 = 0, $$lobit6 = 0, $$lobit7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bx$036 = 0, $bx$134 = 0, $bx$224 = 0, $by$037 = 0, $by$130 = 0;
 var $by$227 = 0, $dst$025 = 0, $exitcond = 0, $exitcond40 = 0, $exitcond41 = 0, $exitcond42 = 0, $exitcond43 = 0, $exitcond44 = 0, $hmod = 0, $ispos = 0, $ispos8 = 0, $neg = 0, $neg9 = 0, $nsrc$033 = 0, $nsrc$126 = 0, $psrc$039 = 0, $psrc$039$phi = 0, $psrc$135 = 0, $psrc$135$phi = 0, $psrc$232 = 0;
 var $psrc$329 = 0, $psrc$329$phi = 0, $src$038 = 0, $src$131 = 0, $src$228 = 0, $src$228$phi = 0, $vmod = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 576|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vmod = sp + 288|0;
 $hmod = sp;
 $0 = ($_dc_scale*3)|0;
 $1 = (530006 + ($_strong)|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = $2&255;
 $4 = ($3|0)<($0|0);
 $5 = $4 ? $3 : $0;
 $6 = $_b >>> 2;
 $$lobit = $6 & 1;
 $7 = $$lobit ^ 1;
 $8 = (0 - ($7))|0;
 $9 = $8 & $_ystride;
 $10 = (0 - ($9))|0;
 $11 = (($_idata) + ($10)|0);
 $12 = (($_dc_scale) + 32)|0;
 $13 = (530008 + ($_strong)|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = $14&255;
 $16 = $_b >>> 3;
 $$lobit7 = $16 & 1;
 $17 = $$lobit7 ^ 1;
 $by$037 = 0;$psrc$039 = $11;$src$038 = $_idata;
 while(1) {
  $18 = $by$037 << 3;
  $bx$036 = 0;
  while(1) {
   $19 = (($src$038) + ($bx$036)|0);
   $20 = HEAP8[$19>>0]|0;
   $21 = $20&255;
   $22 = (($psrc$039) + ($bx$036)|0);
   $23 = HEAP8[$22>>0]|0;
   $24 = $23&255;
   $25 = (($21) - ($24))|0;
   $ispos8 = ($25|0)>(-1);
   $neg9 = (0 - ($25))|0;
   $26 = $ispos8 ? $25 : $neg9;
   $27 = $26 << $15;
   $28 = (($12) - ($27))|0;
   $29 = ($28|0)<(-64);
   if ($29) {
    $36 = $_sharp_mod;
   } else {
    $30 = ($5|0)<($28|0);
    $31 = $30 ? $5 : $28;
    $32 = ($31|0)>(0);
    $33 = $32 ? $31 : 0;
    $36 = $33;
   }
   $34 = (($bx$036) + ($18))|0;
   $35 = (($vmod) + ($34<<2)|0);
   HEAP32[$35>>2] = $36;
   $37 = (($bx$036) + 1)|0;
   $exitcond43 = ($37|0)==(8);
   if ($exitcond43) {
    break;
   } else {
    $bx$036 = $37;
   }
  }
  $38 = ($by$037|0)<(7);
  $39 = $38&1;
  $40 = $39 | $17;
  $41 = (0 - ($40))|0;
  $42 = $41 & $_ystride;
  $43 = (($src$038) + ($42)|0);
  $44 = (($by$037) + 1)|0;
  $exitcond44 = ($44|0)==(9);
  if ($exitcond44) {
   break;
  } else {
   $psrc$039$phi = $src$038;$by$037 = $44;$src$038 = $43;$psrc$039 = $psrc$039$phi;
  }
 }
 $45 = $_b & 1;
 $46 = $45 ^ 1;
 $47 = (0 - ($46))|0;
 $48 = (($_idata) + ($47)|0);
 $49 = $_b >>> 1;
 $$lobit6 = $49 & 1;
 $50 = $$lobit6 ^ 1;
 $bx$134 = 0;$nsrc$033 = $_idata;$psrc$135 = $48;
 while(1) {
  $51 = $bx$134 << 3;
  $by$130 = 0;$psrc$232 = $psrc$135;$src$131 = $nsrc$033;
  while(1) {
   $52 = HEAP8[$src$131>>0]|0;
   $53 = $52&255;
   $54 = HEAP8[$psrc$232>>0]|0;
   $55 = $54&255;
   $56 = (($53) - ($55))|0;
   $ispos = ($56|0)>(-1);
   $neg = (0 - ($56))|0;
   $57 = $ispos ? $56 : $neg;
   $58 = $57 << $15;
   $59 = (($12) - ($58))|0;
   $60 = ($59|0)<(-64);
   if ($60) {
    $67 = $_sharp_mod;
   } else {
    $61 = ($5|0)<($59|0);
    $62 = $61 ? $5 : $59;
    $63 = ($62|0)>(0);
    $64 = $63 ? $62 : 0;
    $67 = $64;
   }
   $65 = (($by$130) + ($51))|0;
   $66 = (($hmod) + ($65<<2)|0);
   HEAP32[$66>>2] = $67;
   $68 = (($psrc$232) + ($_ystride)|0);
   $69 = (($src$131) + ($_ystride)|0);
   $70 = (($by$130) + 1)|0;
   $exitcond41 = ($70|0)==(8);
   if ($exitcond41) {
    break;
   } else {
    $by$130 = $70;$psrc$232 = $68;$src$131 = $69;
   }
  }
  $71 = ($bx$134|0)<(7);
  $72 = $71&1;
  $73 = $72 | $50;
  $74 = (($nsrc$033) + ($73)|0);
  $75 = (($bx$134) + 1)|0;
  $exitcond42 = ($75|0)==(9);
  if ($exitcond42) {
   break;
  } else {
   $psrc$135$phi = $nsrc$033;$bx$134 = $75;$nsrc$033 = $74;$psrc$135 = $psrc$135$phi;
  }
 }
 $76 = (($_idata) + ($_ystride)|0);
 $77 = (8 - ($$lobit6))|0;
 $by$227 = 0;$dst$025 = $_idata;$nsrc$126 = $76;$psrc$329 = $11;$src$228 = $_idata;
 while(1) {
  $78 = (($hmod) + ($by$227<<2)|0);
  $79 = HEAP32[$78>>2]|0;
  $80 = (128 - ($79))|0;
  $81 = (($src$228) + ($47)|0);
  $82 = HEAP8[$81>>0]|0;
  $83 = $82&255;
  $84 = Math_imul($83, $79)|0;
  $85 = (($84) + 64)|0;
  $86 = $by$227 << 3;
  $87 = (($vmod) + ($86<<2)|0);
  $88 = HEAP32[$87>>2]|0;
  $89 = (($80) - ($88))|0;
  $90 = HEAP8[$psrc$329>>0]|0;
  $91 = $90&255;
  $92 = Math_imul($91, $88)|0;
  $93 = (($85) + ($92))|0;
  $94 = (($by$227) + 1)|0;
  $95 = $94 << 3;
  $96 = (($vmod) + ($95<<2)|0);
  $97 = HEAP32[$96>>2]|0;
  $98 = (($89) - ($97))|0;
  $99 = HEAP8[$nsrc$126>>0]|0;
  $100 = $99&255;
  $101 = Math_imul($100, $97)|0;
  $102 = (($93) + ($101))|0;
  $103 = (($by$227) + 8)|0;
  $104 = (($hmod) + ($103<<2)|0);
  $105 = HEAP32[$104>>2]|0;
  $106 = (($98) - ($105))|0;
  $107 = ((($src$228)) + 1|0);
  $108 = HEAP8[$107>>0]|0;
  $109 = $108&255;
  $110 = Math_imul($109, $105)|0;
  $111 = (($102) + ($110))|0;
  $112 = HEAP8[$src$228>>0]|0;
  $113 = $112&255;
  $114 = Math_imul($113, $106)|0;
  $115 = (($111) + ($114))|0;
  $116 = $115 >> 7;
  $$lobit1 = $116 >>> 31;
  $117 = (($$lobit1) + 255)|0;
  $118 = ($116|0)>(255);
  $119 = $118 ? 255 : $116;
  $120 = $117 & $119;
  $121 = $120&255;
  HEAP8[$dst$025>>0] = $121;
  $bx$224 = 1;
  while(1) {
   $122 = $bx$224 << 3;
   $123 = (($122) + ($by$227))|0;
   $124 = (($hmod) + ($123<<2)|0);
   $125 = HEAP32[$124>>2]|0;
   $126 = (128 - ($125))|0;
   $127 = (($bx$224) + -1)|0;
   $128 = (($src$228) + ($127)|0);
   $129 = HEAP8[$128>>0]|0;
   $130 = $129&255;
   $131 = Math_imul($130, $125)|0;
   $132 = (($131) + 64)|0;
   $133 = (($bx$224) + ($86))|0;
   $134 = (($vmod) + ($133<<2)|0);
   $135 = HEAP32[$134>>2]|0;
   $136 = (($126) - ($135))|0;
   $137 = (($psrc$329) + ($bx$224)|0);
   $138 = HEAP8[$137>>0]|0;
   $139 = $138&255;
   $140 = Math_imul($139, $135)|0;
   $141 = (($132) + ($140))|0;
   $142 = (($bx$224) + ($95))|0;
   $143 = (($vmod) + ($142<<2)|0);
   $144 = HEAP32[$143>>2]|0;
   $145 = (($136) - ($144))|0;
   $146 = (($nsrc$126) + ($bx$224)|0);
   $147 = HEAP8[$146>>0]|0;
   $148 = $147&255;
   $149 = Math_imul($148, $144)|0;
   $150 = (($141) + ($149))|0;
   $151 = (($bx$224) + 1)|0;
   $152 = $151 << 3;
   $153 = (($152) + ($by$227))|0;
   $154 = (($hmod) + ($153<<2)|0);
   $155 = HEAP32[$154>>2]|0;
   $156 = (($145) - ($155))|0;
   $157 = (($src$228) + ($151)|0);
   $158 = HEAP8[$157>>0]|0;
   $159 = $158&255;
   $160 = Math_imul($159, $155)|0;
   $161 = (($150) + ($160))|0;
   $162 = (($src$228) + ($bx$224)|0);
   $163 = HEAP8[$162>>0]|0;
   $164 = $163&255;
   $165 = Math_imul($164, $156)|0;
   $166 = (($161) + ($165))|0;
   $167 = $166 >> 7;
   $$lobit5 = $167 >>> 31;
   $168 = (($$lobit5) + 255)|0;
   $169 = ($167|0)>(255);
   $170 = $169 ? 255 : $167;
   $171 = $168 & $170;
   $172 = $171&255;
   $173 = (($dst$025) + ($bx$224)|0);
   HEAP8[$173>>0] = $172;
   $exitcond = ($151|0)==(7);
   if ($exitcond) {
    break;
   } else {
    $bx$224 = $151;
   }
  }
  $174 = (($by$227) + 56)|0;
  $175 = (($hmod) + ($174<<2)|0);
  $176 = HEAP32[$175>>2]|0;
  $177 = (128 - ($176))|0;
  $178 = ((($src$228)) + 6|0);
  $179 = HEAP8[$178>>0]|0;
  $180 = $179&255;
  $181 = Math_imul($180, $176)|0;
  $182 = (($181) + 64)|0;
  $183 = $86 | 7;
  $184 = (($vmod) + ($183<<2)|0);
  $185 = HEAP32[$184>>2]|0;
  $186 = (($177) - ($185))|0;
  $187 = ((($psrc$329)) + 7|0);
  $188 = HEAP8[$187>>0]|0;
  $189 = $188&255;
  $190 = Math_imul($189, $185)|0;
  $191 = (($182) + ($190))|0;
  $192 = $95 | 7;
  $193 = (($vmod) + ($192<<2)|0);
  $194 = HEAP32[$193>>2]|0;
  $195 = (($186) - ($194))|0;
  $196 = ((($nsrc$126)) + 7|0);
  $197 = HEAP8[$196>>0]|0;
  $198 = $197&255;
  $199 = Math_imul($198, $194)|0;
  $200 = (($191) + ($199))|0;
  $201 = (($by$227) + 64)|0;
  $202 = (($hmod) + ($201<<2)|0);
  $203 = HEAP32[$202>>2]|0;
  $204 = (($195) - ($203))|0;
  $205 = (($src$228) + ($77)|0);
  $206 = HEAP8[$205>>0]|0;
  $207 = $206&255;
  $208 = Math_imul($207, $203)|0;
  $209 = (($200) + ($208))|0;
  $210 = ((($src$228)) + 7|0);
  $211 = HEAP8[$210>>0]|0;
  $212 = $211&255;
  $213 = Math_imul($212, $204)|0;
  $214 = (($209) + ($213))|0;
  $215 = $214 >> 7;
  $$lobit3 = $215 >>> 31;
  $216 = (($$lobit3) + 255)|0;
  $217 = ($215|0)>(255);
  $218 = $217 ? 255 : $215;
  $219 = $216 & $218;
  $220 = $219&255;
  $221 = ((($dst$025)) + 7|0);
  HEAP8[$221>>0] = $220;
  $222 = (($dst$025) + ($_ystride)|0);
  $223 = ($by$227|0)<(6);
  $224 = $223&1;
  $225 = $224 | $17;
  $226 = (0 - ($225))|0;
  $227 = $226 & $_ystride;
  $228 = (($nsrc$126) + ($227)|0);
  $exitcond40 = ($94|0)==(8);
  if ($exitcond40) {
   break;
  } else {
   $src$228$phi = $nsrc$126;$psrc$329$phi = $src$228;$by$227 = $94;$dst$025 = $222;$nsrc$126 = $228;$src$228 = $src$228$phi;$psrc$329 = $psrc$329$phi;
  }
 }
 STACKTOP = sp;return;
}
function _th_decode_ycbcr_out($_dec,$_ycbcr) {
 $_dec = $_dec|0;
 $_ycbcr = $_ycbcr|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($_dec|0)==(0|0);
 $1 = ($_ycbcr|0)==(0|0);
 $or$cond = $0 | $1;
 if ($or$cond) {
  $$0 = -1;
  return ($$0|0);
 }
 $2 = ((($_dec)) + 53976|0);
 _oc_ycbcr_buffer_flip($_ycbcr,$2);
 $$0 = 0;
 return ($$0|0);
}
function _oc_quant_params_unpack($_opb,$_qinfo) {
 $_opb = $_opb|0;
 $_qinfo = $_qinfo|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa26 = 0, $$lcssa27 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $8 = 0, $9 = 0, $bmi$06 = 0, $ci$05 = 0, $exitcond = 0, $exitcond11 = 0, $exitcond12 = 0, $exitcond13 = 0, $i$04 = 0, $indices = 0, $plj$0 = 0, $qi$09 = 0, $qi$18 = 0, $qi$27 = 0, $qi$33 = 0, $qri$02 = 0, $qri$119 = 0, $qtj$0 = 0, $sizes = 0;
 var dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 512|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $sizes = sp + 256|0;
 $indices = sp;
 $0 = (_oc_pack_read($_opb,3)|0);
 $qi$09 = 0;
 while(1) {
  $1 = (_oc_pack_read($_opb,$0)|0);
  $2 = $1&255;
  $3 = (((($_qinfo)) + 256|0) + ($qi$09)|0);
  HEAP8[$3>>0] = $2;
  $4 = (($qi$09) + 1)|0;
  $exitcond13 = ($4|0)==(64);
  if ($exitcond13) {
   break;
  } else {
   $qi$09 = $4;
  }
 }
 $5 = (_oc_pack_read($_opb,4)|0);
 $6 = (($5) + 1)|0;
 $qi$18 = 0;
 while(1) {
  $7 = (_oc_pack_read($_opb,$6)|0);
  $8 = $7&65535;
  $9 = (((($_qinfo)) + 128|0) + ($qi$18<<1)|0);
  HEAP16[$9>>1] = $8;
  $10 = (($qi$18) + 1)|0;
  $exitcond12 = ($10|0)==(64);
  if ($exitcond12) {
   break;
  } else {
   $qi$18 = $10;
  }
 }
 $11 = (_oc_pack_read($_opb,4)|0);
 $12 = (($11) + 1)|0;
 $qi$27 = 0;
 while(1) {
  $13 = (_oc_pack_read($_opb,$12)|0);
  $14 = $13&65535;
  $15 = (($_qinfo) + ($qi$27<<1)|0);
  HEAP16[$15>>1] = $14;
  $16 = (($qi$27) + 1)|0;
  $exitcond11 = ($16|0)==(64);
  if ($exitcond11) {
   break;
  } else {
   $qi$27 = $16;
  }
 }
 $17 = (_oc_pack_read($_opb,9)|0);
 $18 = (($17) + 1)|0;
 $19 = $18 << 6;
 $20 = (_malloc($19)|0);
 $21 = ($20|0)==(0|0);
 if ($21) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $22 = ($17|0)>(-1);
 if ($22) {
  $bmi$06 = 0;
  while(1) {
   $ci$05 = 0;
   while(1) {
    $23 = (_oc_pack_read($_opb,8)|0);
    $24 = $23&255;
    $25 = ((($20) + ($bmi$06<<6)|0) + ($ci$05)|0);
    HEAP8[$25>>0] = $24;
    $26 = (($ci$05) + 1)|0;
    $exitcond = ($26|0)==(64);
    if ($exitcond) {
     break;
    } else {
     $ci$05 = $26;
    }
   }
   $27 = (($bmi$06) + 1)|0;
   $28 = ($27|0)<($18|0);
   if ($28) {
    $bmi$06 = $27;
   } else {
    break;
   }
  }
 }
 $29 = (_oc_ilog($17)|0);
 $i$04 = 0;
 L20: while(1) {
  $30 = (($i$04|0) / 3)&-1;
  $31 = (($i$04|0) % 3)&-1;
  $32 = ((((($_qinfo)) + 320|0) + (($30*36)|0)|0) + (($31*12)|0)|0);
  $33 = ($i$04|0)>(0);
  if ($33) {
   $34 = (_oc_pack_read1($_opb)|0);
   $35 = ($34|0)==(0);
   if ($35) {
    $36 = ($i$04|0)>(2);
    do {
     if ($36) {
      $37 = (_oc_pack_read1($_opb)|0);
      $38 = ($37|0)==(0);
      if ($38) {
       $40 = (($i$04) + -1)|0;
       $41 = (($40|0) / 3)&-1;
       $42 = (($40|0) % 3)&-1;
       $plj$0 = $42;$qtj$0 = $41;
       break;
      } else {
       $39 = (($30) + -1)|0;
       $plj$0 = $31;$qtj$0 = $39;
       break;
      }
     } else {
      $43 = (($i$04) + -1)|0;
      $44 = (($43|0) / 3)&-1;
      $45 = (($43|0) % 3)&-1;
      $plj$0 = $45;$qtj$0 = $44;
     }
    } while(0);
    $46 = ((((($_qinfo)) + 320|0) + (($qtj$0*36)|0)|0) + (($plj$0*12)|0)|0);
    ;HEAP32[$32>>2]=HEAP32[$46>>2]|0;HEAP32[$32+4>>2]=HEAP32[$46+4>>2]|0;HEAP32[$32+8>>2]=HEAP32[$46+8>>2]|0;
   } else {
    label = 21;
   }
  } else {
   label = 21;
  }
  L32: do {
   if ((label|0) == 21) {
    label = 0;
    $47 = (_oc_pack_read($_opb,$29)|0);
    HEAP32[$indices>>2] = $47;
    $qi$33 = 0;$qri$02 = 0;
    while(1) {
     $48 = (62 - ($qi$33))|0;
     $49 = (_oc_ilog($48)|0);
     $50 = (_oc_pack_read($_opb,$49)|0);
     $51 = (($50) + 1)|0;
     $52 = (($sizes) + ($qri$02<<2)|0);
     HEAP32[$52>>2] = $51;
     $53 = (($51) + ($qi$33))|0;
     $54 = (_oc_pack_read($_opb,$29)|0);
     $55 = (($qri$02) + 1)|0;
     $56 = (($indices) + ($55<<2)|0);
     HEAP32[$56>>2] = $54;
     $57 = ($53|0)<(63);
     if ($57) {
      $qi$33 = $53;$qri$02 = $55;
     } else {
      $$lcssa = $53;$$lcssa26 = $54;$$lcssa27 = $55;
      break;
     }
    }
    $58 = ($$lcssa|0)>(63);
    if ($58) {
     label = 24;
     break L20;
    }
    HEAP32[$32>>2] = $$lcssa27;
    $59 = $$lcssa27 << 2;
    $60 = (_malloc($59)|0);
    $61 = ((((((($_qinfo)) + 320|0) + (($30*36)|0)|0) + (($31*12)|0)|0)) + 4|0);
    HEAP32[$61>>2] = $60;
    $62 = ($60|0)==(0|0);
    if ($62) {
     label = 26;
     break L20;
    }
    _memcpy(($60|0),($sizes|0),($59|0))|0;
    $63 = $$lcssa27 << 6;
    $64 = (($63) + 64)|0;
    $65 = (_malloc($64)|0);
    $66 = ($65|0)==(0|0);
    if ($66) {
     label = 28;
     break L20;
    }
    $67 = ((((((($_qinfo)) + 320|0) + (($30*36)|0)|0) + (($31*12)|0)|0)) + 8|0);
    HEAP32[$67>>2] = $65;
    $68 = ($$lcssa26|0)<($18|0);
    if ($68) {
     $71 = $$lcssa26;$qri$119 = $$lcssa27;
    } else {
     label = 30;
     break L20;
    }
    while(1) {
     $69 = (($65) + ($qri$119<<6)|0);
     $70 = (($20) + ($71<<6)|0);
     dest=$69; src=$70; stop=dest+64|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
     $72 = (($qri$119) + -1)|0;
     $73 = ($qri$119|0)>(0);
     if (!($73)) {
      break L32;
     }
     $$phi$trans$insert = (($indices) + ($72<<2)|0);
     $$pre = HEAP32[$$phi$trans$insert>>2]|0;
     $74 = ($$pre|0)<($18|0);
     if ($74) {
      $71 = $$pre;$qri$119 = $72;
     } else {
      label = 30;
      break L20;
     }
    }
   }
  } while(0);
  $75 = (($i$04) + 1)|0;
  $76 = ($75|0)<(6);
  if ($76) {
   $i$04 = $75;
  } else {
   label = 34;
   break;
  }
 }
 if ((label|0) == 24) {
  _free($20);
  $$0 = -20;
  STACKTOP = sp;return ($$0|0);
 }
 else if ((label|0) == 26) {
  _free($20);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 else if ((label|0) == 28) {
  _free($20);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 else if ((label|0) == 30) {
  _free($20);
  $$0 = -20;
  STACKTOP = sp;return ($$0|0);
 }
 else if ((label|0) == 34) {
  _free($20);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _oc_quant_params_clear($_qinfo) {
 $_qinfo = $_qinfo|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $i$01$phi = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = 5;$i$01 = 6;
 while(1) {
  $0 = (($1|0) / 3)&-1;
  $2 = (($1|0) % 3)&-1;
  $3 = ($i$01|0)>(1);
  if ($3) {
   $4 = (($i$01) + -2)|0;
   $5 = (($4|0) / 3)&-1;
   $6 = (($4|0) % 3)&-1;
   $7 = ((((((($_qinfo)) + 320|0) + (($0*36)|0)|0) + (($2*12)|0)|0)) + 4|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ((((((($_qinfo)) + 320|0) + (($5*36)|0)|0) + (($6*12)|0)|0)) + 4|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($8|0)==($10|0);
   if ($11) {
    HEAP32[$7>>2] = 0;
   }
   $12 = ((((((($_qinfo)) + 320|0) + (($0*36)|0)|0) + (($2*12)|0)|0)) + 8|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = ((((((($_qinfo)) + 320|0) + (($5*36)|0)|0) + (($6*12)|0)|0)) + 8|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($13|0)==($15|0);
   if ($16) {
    HEAP32[$12>>2] = 0;
   }
  }
  $17 = ($1|0)>(2);
  if ($17) {
   $18 = (((((($_qinfo)) + 356|0) + (($2*12)|0)|0)) + 4|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = (((((($_qinfo)) + 320|0) + (($2*12)|0)|0)) + 4|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = ($19|0)==($21|0);
   if ($22) {
    HEAP32[$18>>2] = 0;
   }
   $23 = (((((($_qinfo)) + 356|0) + (($2*12)|0)|0)) + 8|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = (((((($_qinfo)) + 320|0) + (($2*12)|0)|0)) + 8|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = ($24|0)==($26|0);
   if ($27) {
    HEAP32[$23>>2] = 0;
   }
  }
  $28 = ((((((($_qinfo)) + 320|0) + (($0*36)|0)|0) + (($2*12)|0)|0)) + 4|0);
  $29 = HEAP32[$28>>2]|0;
  _free($29);
  $30 = ((((((($_qinfo)) + 320|0) + (($0*36)|0)|0) + (($2*12)|0)|0)) + 8|0);
  $31 = HEAP32[$30>>2]|0;
  _free($31);
  $32 = (($1) + -1)|0;
  $33 = ($1|0)>(0);
  if ($33) {
   $i$01$phi = $1;$1 = $32;$i$01 = $i$01$phi;
  } else {
   break;
  }
 }
 return;
}
function _oc_frag_copy($_state,$_dst,$_src,$_ystride) {
 $_state = $_state|0;
 $_dst = $_dst|0;
 $_src = $_src|0;
 $_ystride = $_ystride|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_state)) + 64|0);
 $1 = HEAP32[$0>>2]|0;
 FUNCTION_TABLE_viii[$1 & 8191]($_dst,$_src,$_ystride);
 return;
}
function _oc_frag_copy_c($_dst,$_src,$_ystride) {
 $_dst = $_dst|0;
 $_src = $_src|0;
 $_ystride = $_ystride|0;
 var $$sum = 0, $$sum10 = 0, $$sum11 = 0, $$sum12 = 0, $$sum13 = 0, $$sum14 = 0, $$sum15 = 0, $$sum5 = 0, $$sum6 = 0, $$sum7 = 0, $$sum8 = 0, $$sum9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $_src;
 $1 = $0;
 $2 = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24);
 $3 = (($0) + 4)|0;
 $4 = $3;
 $5 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8)|(HEAPU8[$4+2>>0]<<16)|(HEAPU8[$4+3>>0]<<24);
 $6 = $_dst;
 $7 = $6;
 HEAP8[$7>>0]=$2&255;HEAP8[$7+1>>0]=($2>>8)&255;HEAP8[$7+2>>0]=($2>>16)&255;HEAP8[$7+3>>0]=$2>>24;
 $8 = (($6) + 4)|0;
 $9 = $8;
 HEAP8[$9>>0]=$5&255;HEAP8[$9+1>>0]=($5>>8)&255;HEAP8[$9+2>>0]=($5>>16)&255;HEAP8[$9+3>>0]=$5>>24;
 $10 = (($_dst) + ($_ystride)|0);
 $11 = (($_src) + ($_ystride)|0);
 $12 = $11;
 $13 = $12;
 $14 = HEAPU8[$13>>0]|(HEAPU8[$13+1>>0]<<8)|(HEAPU8[$13+2>>0]<<16)|(HEAPU8[$13+3>>0]<<24);
 $15 = (($12) + 4)|0;
 $16 = $15;
 $17 = HEAPU8[$16>>0]|(HEAPU8[$16+1>>0]<<8)|(HEAPU8[$16+2>>0]<<16)|(HEAPU8[$16+3>>0]<<24);
 $18 = $10;
 $19 = $18;
 HEAP8[$19>>0]=$14&255;HEAP8[$19+1>>0]=($14>>8)&255;HEAP8[$19+2>>0]=($14>>16)&255;HEAP8[$19+3>>0]=$14>>24;
 $20 = (($18) + 4)|0;
 $21 = $20;
 HEAP8[$21>>0]=$17&255;HEAP8[$21+1>>0]=($17>>8)&255;HEAP8[$21+2>>0]=($17>>16)&255;HEAP8[$21+3>>0]=$17>>24;
 $$sum = $_ystride << 1;
 $22 = (($_dst) + ($$sum)|0);
 $$sum5 = $_ystride << 1;
 $23 = (($_src) + ($$sum5)|0);
 $24 = $23;
 $25 = $24;
 $26 = HEAPU8[$25>>0]|(HEAPU8[$25+1>>0]<<8)|(HEAPU8[$25+2>>0]<<16)|(HEAPU8[$25+3>>0]<<24);
 $27 = (($24) + 4)|0;
 $28 = $27;
 $29 = HEAPU8[$28>>0]|(HEAPU8[$28+1>>0]<<8)|(HEAPU8[$28+2>>0]<<16)|(HEAPU8[$28+3>>0]<<24);
 $30 = $22;
 $31 = $30;
 HEAP8[$31>>0]=$26&255;HEAP8[$31+1>>0]=($26>>8)&255;HEAP8[$31+2>>0]=($26>>16)&255;HEAP8[$31+3>>0]=$26>>24;
 $32 = (($30) + 4)|0;
 $33 = $32;
 HEAP8[$33>>0]=$29&255;HEAP8[$33+1>>0]=($29>>8)&255;HEAP8[$33+2>>0]=($29>>16)&255;HEAP8[$33+3>>0]=$29>>24;
 $$sum6 = ($_ystride*3)|0;
 $34 = (($_dst) + ($$sum6)|0);
 $$sum7 = ($_ystride*3)|0;
 $35 = (($_src) + ($$sum7)|0);
 $36 = $35;
 $37 = $36;
 $38 = HEAPU8[$37>>0]|(HEAPU8[$37+1>>0]<<8)|(HEAPU8[$37+2>>0]<<16)|(HEAPU8[$37+3>>0]<<24);
 $39 = (($36) + 4)|0;
 $40 = $39;
 $41 = HEAPU8[$40>>0]|(HEAPU8[$40+1>>0]<<8)|(HEAPU8[$40+2>>0]<<16)|(HEAPU8[$40+3>>0]<<24);
 $42 = $34;
 $43 = $42;
 HEAP8[$43>>0]=$38&255;HEAP8[$43+1>>0]=($38>>8)&255;HEAP8[$43+2>>0]=($38>>16)&255;HEAP8[$43+3>>0]=$38>>24;
 $44 = (($42) + 4)|0;
 $45 = $44;
 HEAP8[$45>>0]=$41&255;HEAP8[$45+1>>0]=($41>>8)&255;HEAP8[$45+2>>0]=($41>>16)&255;HEAP8[$45+3>>0]=$41>>24;
 $$sum8 = $_ystride << 2;
 $46 = (($_dst) + ($$sum8)|0);
 $$sum9 = $_ystride << 2;
 $47 = (($_src) + ($$sum9)|0);
 $48 = $47;
 $49 = $48;
 $50 = HEAPU8[$49>>0]|(HEAPU8[$49+1>>0]<<8)|(HEAPU8[$49+2>>0]<<16)|(HEAPU8[$49+3>>0]<<24);
 $51 = (($48) + 4)|0;
 $52 = $51;
 $53 = HEAPU8[$52>>0]|(HEAPU8[$52+1>>0]<<8)|(HEAPU8[$52+2>>0]<<16)|(HEAPU8[$52+3>>0]<<24);
 $54 = $46;
 $55 = $54;
 HEAP8[$55>>0]=$50&255;HEAP8[$55+1>>0]=($50>>8)&255;HEAP8[$55+2>>0]=($50>>16)&255;HEAP8[$55+3>>0]=$50>>24;
 $56 = (($54) + 4)|0;
 $57 = $56;
 HEAP8[$57>>0]=$53&255;HEAP8[$57+1>>0]=($53>>8)&255;HEAP8[$57+2>>0]=($53>>16)&255;HEAP8[$57+3>>0]=$53>>24;
 $$sum10 = ($_ystride*5)|0;
 $58 = (($_dst) + ($$sum10)|0);
 $$sum11 = ($_ystride*5)|0;
 $59 = (($_src) + ($$sum11)|0);
 $60 = $59;
 $61 = $60;
 $62 = HEAPU8[$61>>0]|(HEAPU8[$61+1>>0]<<8)|(HEAPU8[$61+2>>0]<<16)|(HEAPU8[$61+3>>0]<<24);
 $63 = (($60) + 4)|0;
 $64 = $63;
 $65 = HEAPU8[$64>>0]|(HEAPU8[$64+1>>0]<<8)|(HEAPU8[$64+2>>0]<<16)|(HEAPU8[$64+3>>0]<<24);
 $66 = $58;
 $67 = $66;
 HEAP8[$67>>0]=$62&255;HEAP8[$67+1>>0]=($62>>8)&255;HEAP8[$67+2>>0]=($62>>16)&255;HEAP8[$67+3>>0]=$62>>24;
 $68 = (($66) + 4)|0;
 $69 = $68;
 HEAP8[$69>>0]=$65&255;HEAP8[$69+1>>0]=($65>>8)&255;HEAP8[$69+2>>0]=($65>>16)&255;HEAP8[$69+3>>0]=$65>>24;
 $$sum12 = ($_ystride*6)|0;
 $70 = (($_dst) + ($$sum12)|0);
 $$sum13 = ($_ystride*6)|0;
 $71 = (($_src) + ($$sum13)|0);
 $72 = $71;
 $73 = $72;
 $74 = HEAPU8[$73>>0]|(HEAPU8[$73+1>>0]<<8)|(HEAPU8[$73+2>>0]<<16)|(HEAPU8[$73+3>>0]<<24);
 $75 = (($72) + 4)|0;
 $76 = $75;
 $77 = HEAPU8[$76>>0]|(HEAPU8[$76+1>>0]<<8)|(HEAPU8[$76+2>>0]<<16)|(HEAPU8[$76+3>>0]<<24);
 $78 = $70;
 $79 = $78;
 HEAP8[$79>>0]=$74&255;HEAP8[$79+1>>0]=($74>>8)&255;HEAP8[$79+2>>0]=($74>>16)&255;HEAP8[$79+3>>0]=$74>>24;
 $80 = (($78) + 4)|0;
 $81 = $80;
 HEAP8[$81>>0]=$77&255;HEAP8[$81+1>>0]=($77>>8)&255;HEAP8[$81+2>>0]=($77>>16)&255;HEAP8[$81+3>>0]=$77>>24;
 $$sum14 = ($_ystride*7)|0;
 $82 = (($_dst) + ($$sum14)|0);
 $$sum15 = ($_ystride*7)|0;
 $83 = (($_src) + ($$sum15)|0);
 $84 = $83;
 $85 = $84;
 $86 = HEAPU8[$85>>0]|(HEAPU8[$85+1>>0]<<8)|(HEAPU8[$85+2>>0]<<16)|(HEAPU8[$85+3>>0]<<24);
 $87 = (($84) + 4)|0;
 $88 = $87;
 $89 = HEAPU8[$88>>0]|(HEAPU8[$88+1>>0]<<8)|(HEAPU8[$88+2>>0]<<16)|(HEAPU8[$88+3>>0]<<24);
 $90 = $82;
 $91 = $90;
 HEAP8[$91>>0]=$86&255;HEAP8[$91+1>>0]=($86>>8)&255;HEAP8[$91+2>>0]=($86>>16)&255;HEAP8[$91+3>>0]=$86>>24;
 $92 = (($90) + 4)|0;
 $93 = $92;
 HEAP8[$93>>0]=$89&255;HEAP8[$93+1>>0]=($89>>8)&255;HEAP8[$93+2>>0]=($89>>16)&255;HEAP8[$93+3>>0]=$89>>24;
 return;
}
function _oc_frag_recon_intra($_state,$_dst,$_ystride,$_residue) {
 $_state = $_state|0;
 $_dst = $_dst|0;
 $_ystride = $_ystride|0;
 $_residue = $_residue|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_state)) + 68|0);
 $1 = HEAP32[$0>>2]|0;
 FUNCTION_TABLE_viii[$1 & 8191]($_dst,$_ystride,$_residue);
 return;
}
function _oc_frag_recon_intra_c($_dst,$_ystride,$_residue) {
 $_dst = $_dst|0;
 $_ystride = $_ystride|0;
 $_residue = $_residue|0;
 var $$02 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $9 = 0, $exitcond = 0, $i$03 = 0, $not$ = 0, $not$$1 = 0, $not$$2 = 0, $not$$3 = 0, $not$$4 = 0, $not$$5 = 0, $not$$6 = 0, $not$$7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$02 = $_dst;$i$03 = 0;
 while(1) {
  $0 = $i$03 << 3;
  $1 = (($_residue) + ($0<<1)|0);
  $2 = HEAP16[$1>>1]|0;
  $3 = $2 << 16 >> 16;
  $4 = (($3) + 128)|0;
  $not$ = ($2<<16>>16)>(-129);
  $5 = ($4|0)>(255);
  $6 = $4&255;
  $7 = $5 ? -1 : $6;
  $8 = $not$ ? $7 : 0;
  HEAP8[$$02>>0] = $8;
  $9 = $0 | 1;
  $10 = (($_residue) + ($9<<1)|0);
  $11 = HEAP16[$10>>1]|0;
  $12 = $11 << 16 >> 16;
  $13 = (($12) + 128)|0;
  $not$$1 = ($11<<16>>16)>(-129);
  $14 = ($13|0)>(255);
  $15 = $13&255;
  $16 = $14 ? -1 : $15;
  $17 = $not$$1 ? $16 : 0;
  $18 = ((($$02)) + 1|0);
  HEAP8[$18>>0] = $17;
  $19 = $0 | 2;
  $20 = (($_residue) + ($19<<1)|0);
  $21 = HEAP16[$20>>1]|0;
  $22 = $21 << 16 >> 16;
  $23 = (($22) + 128)|0;
  $not$$2 = ($21<<16>>16)>(-129);
  $24 = ($23|0)>(255);
  $25 = $23&255;
  $26 = $24 ? -1 : $25;
  $27 = $not$$2 ? $26 : 0;
  $28 = ((($$02)) + 2|0);
  HEAP8[$28>>0] = $27;
  $29 = $0 | 3;
  $30 = (($_residue) + ($29<<1)|0);
  $31 = HEAP16[$30>>1]|0;
  $32 = $31 << 16 >> 16;
  $33 = (($32) + 128)|0;
  $not$$3 = ($31<<16>>16)>(-129);
  $34 = ($33|0)>(255);
  $35 = $33&255;
  $36 = $34 ? -1 : $35;
  $37 = $not$$3 ? $36 : 0;
  $38 = ((($$02)) + 3|0);
  HEAP8[$38>>0] = $37;
  $39 = $0 | 4;
  $40 = (($_residue) + ($39<<1)|0);
  $41 = HEAP16[$40>>1]|0;
  $42 = $41 << 16 >> 16;
  $43 = (($42) + 128)|0;
  $not$$4 = ($41<<16>>16)>(-129);
  $44 = ($43|0)>(255);
  $45 = $43&255;
  $46 = $44 ? -1 : $45;
  $47 = $not$$4 ? $46 : 0;
  $48 = ((($$02)) + 4|0);
  HEAP8[$48>>0] = $47;
  $49 = $0 | 5;
  $50 = (($_residue) + ($49<<1)|0);
  $51 = HEAP16[$50>>1]|0;
  $52 = $51 << 16 >> 16;
  $53 = (($52) + 128)|0;
  $not$$5 = ($51<<16>>16)>(-129);
  $54 = ($53|0)>(255);
  $55 = $53&255;
  $56 = $54 ? -1 : $55;
  $57 = $not$$5 ? $56 : 0;
  $58 = ((($$02)) + 5|0);
  HEAP8[$58>>0] = $57;
  $59 = $0 | 6;
  $60 = (($_residue) + ($59<<1)|0);
  $61 = HEAP16[$60>>1]|0;
  $62 = $61 << 16 >> 16;
  $63 = (($62) + 128)|0;
  $not$$6 = ($61<<16>>16)>(-129);
  $64 = ($63|0)>(255);
  $65 = $63&255;
  $66 = $64 ? -1 : $65;
  $67 = $not$$6 ? $66 : 0;
  $68 = ((($$02)) + 6|0);
  HEAP8[$68>>0] = $67;
  $69 = $0 | 7;
  $70 = (($_residue) + ($69<<1)|0);
  $71 = HEAP16[$70>>1]|0;
  $72 = $71 << 16 >> 16;
  $73 = (($72) + 128)|0;
  $not$$7 = ($71<<16>>16)>(-129);
  $74 = ($73|0)>(255);
  $75 = $73&255;
  $76 = $74 ? -1 : $75;
  $77 = $not$$7 ? $76 : 0;
  $78 = ((($$02)) + 7|0);
  HEAP8[$78>>0] = $77;
  $79 = (($$02) + ($_ystride)|0);
  $80 = (($i$03) + 1)|0;
  $exitcond = ($80|0)==(8);
  if ($exitcond) {
   break;
  } else {
   $$02 = $79;$i$03 = $80;
  }
 }
 return;
}
function _oc_frag_recon_inter($_state,$_dst,$_src,$_ystride,$_residue) {
 $_state = $_state|0;
 $_dst = $_dst|0;
 $_src = $_src|0;
 $_ystride = $_ystride|0;
 $_residue = $_residue|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_state)) + 72|0);
 $1 = HEAP32[$0>>2]|0;
 FUNCTION_TABLE_viiii[$1 & 8191]($_dst,$_src,$_ystride,$_residue);
 return;
}
function _oc_frag_recon_inter_c($_dst,$_src,$_ystride,$_residue) {
 $_dst = $_dst|0;
 $_src = $_src|0;
 $_ystride = $_ystride|0;
 $_residue = $_residue|0;
 var $$013 = 0, $$04 = 0, $$lobit = 0, $$lobit$1 = 0, $$lobit$2 = 0, $$lobit$3 = 0, $$lobit$4 = 0, $$lobit$5 = 0, $$lobit$6 = 0, $$lobit$7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $i$05 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$013 = $_src;$$04 = $_dst;$i$05 = 0;
 while(1) {
  $0 = $i$05 << 3;
  $1 = (($_residue) + ($0<<1)|0);
  $2 = HEAP16[$1>>1]|0;
  $3 = $2 << 16 >> 16;
  $4 = HEAP8[$$013>>0]|0;
  $5 = $4&255;
  $6 = (($5) + ($3))|0;
  $$lobit = $6 >>> 31;
  $7 = (($$lobit) + 255)|0;
  $8 = ($6|0)>(255);
  $9 = $8 ? 255 : $6;
  $10 = $7 & $9;
  $11 = $10&255;
  HEAP8[$$04>>0] = $11;
  $12 = $0 | 1;
  $13 = (($_residue) + ($12<<1)|0);
  $14 = HEAP16[$13>>1]|0;
  $15 = $14 << 16 >> 16;
  $16 = ((($$013)) + 1|0);
  $17 = HEAP8[$16>>0]|0;
  $18 = $17&255;
  $19 = (($18) + ($15))|0;
  $$lobit$1 = $19 >>> 31;
  $20 = (($$lobit$1) + 255)|0;
  $21 = ($19|0)>(255);
  $22 = $21 ? 255 : $19;
  $23 = $20 & $22;
  $24 = $23&255;
  $25 = ((($$04)) + 1|0);
  HEAP8[$25>>0] = $24;
  $26 = $0 | 2;
  $27 = (($_residue) + ($26<<1)|0);
  $28 = HEAP16[$27>>1]|0;
  $29 = $28 << 16 >> 16;
  $30 = ((($$013)) + 2|0);
  $31 = HEAP8[$30>>0]|0;
  $32 = $31&255;
  $33 = (($32) + ($29))|0;
  $$lobit$2 = $33 >>> 31;
  $34 = (($$lobit$2) + 255)|0;
  $35 = ($33|0)>(255);
  $36 = $35 ? 255 : $33;
  $37 = $34 & $36;
  $38 = $37&255;
  $39 = ((($$04)) + 2|0);
  HEAP8[$39>>0] = $38;
  $40 = $0 | 3;
  $41 = (($_residue) + ($40<<1)|0);
  $42 = HEAP16[$41>>1]|0;
  $43 = $42 << 16 >> 16;
  $44 = ((($$013)) + 3|0);
  $45 = HEAP8[$44>>0]|0;
  $46 = $45&255;
  $47 = (($46) + ($43))|0;
  $$lobit$3 = $47 >>> 31;
  $48 = (($$lobit$3) + 255)|0;
  $49 = ($47|0)>(255);
  $50 = $49 ? 255 : $47;
  $51 = $48 & $50;
  $52 = $51&255;
  $53 = ((($$04)) + 3|0);
  HEAP8[$53>>0] = $52;
  $54 = $0 | 4;
  $55 = (($_residue) + ($54<<1)|0);
  $56 = HEAP16[$55>>1]|0;
  $57 = $56 << 16 >> 16;
  $58 = ((($$013)) + 4|0);
  $59 = HEAP8[$58>>0]|0;
  $60 = $59&255;
  $61 = (($60) + ($57))|0;
  $$lobit$4 = $61 >>> 31;
  $62 = (($$lobit$4) + 255)|0;
  $63 = ($61|0)>(255);
  $64 = $63 ? 255 : $61;
  $65 = $62 & $64;
  $66 = $65&255;
  $67 = ((($$04)) + 4|0);
  HEAP8[$67>>0] = $66;
  $68 = $0 | 5;
  $69 = (($_residue) + ($68<<1)|0);
  $70 = HEAP16[$69>>1]|0;
  $71 = $70 << 16 >> 16;
  $72 = ((($$013)) + 5|0);
  $73 = HEAP8[$72>>0]|0;
  $74 = $73&255;
  $75 = (($74) + ($71))|0;
  $$lobit$5 = $75 >>> 31;
  $76 = (($$lobit$5) + 255)|0;
  $77 = ($75|0)>(255);
  $78 = $77 ? 255 : $75;
  $79 = $76 & $78;
  $80 = $79&255;
  $81 = ((($$04)) + 5|0);
  HEAP8[$81>>0] = $80;
  $82 = $0 | 6;
  $83 = (($_residue) + ($82<<1)|0);
  $84 = HEAP16[$83>>1]|0;
  $85 = $84 << 16 >> 16;
  $86 = ((($$013)) + 6|0);
  $87 = HEAP8[$86>>0]|0;
  $88 = $87&255;
  $89 = (($88) + ($85))|0;
  $$lobit$6 = $89 >>> 31;
  $90 = (($$lobit$6) + 255)|0;
  $91 = ($89|0)>(255);
  $92 = $91 ? 255 : $89;
  $93 = $90 & $92;
  $94 = $93&255;
  $95 = ((($$04)) + 6|0);
  HEAP8[$95>>0] = $94;
  $96 = $0 | 7;
  $97 = (($_residue) + ($96<<1)|0);
  $98 = HEAP16[$97>>1]|0;
  $99 = $98 << 16 >> 16;
  $100 = ((($$013)) + 7|0);
  $101 = HEAP8[$100>>0]|0;
  $102 = $101&255;
  $103 = (($102) + ($99))|0;
  $$lobit$7 = $103 >>> 31;
  $104 = (($$lobit$7) + 255)|0;
  $105 = ($103|0)>(255);
  $106 = $105 ? 255 : $103;
  $107 = $104 & $106;
  $108 = $107&255;
  $109 = ((($$04)) + 7|0);
  HEAP8[$109>>0] = $108;
  $110 = (($$04) + ($_ystride)|0);
  $111 = (($$013) + ($_ystride)|0);
  $112 = (($i$05) + 1)|0;
  $exitcond = ($112|0)==(8);
  if ($exitcond) {
   break;
  } else {
   $$013 = $111;$$04 = $110;$i$05 = $112;
  }
 }
 return;
}
function _oc_frag_recon_inter2($_state,$_dst,$_src1,$_src2,$_ystride,$_residue) {
 $_state = $_state|0;
 $_dst = $_dst|0;
 $_src1 = $_src1|0;
 $_src2 = $_src2|0;
 $_ystride = $_ystride|0;
 $_residue = $_residue|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_state)) + 76|0);
 $1 = HEAP32[$0>>2]|0;
 FUNCTION_TABLE_viiiii[$1 & 8191]($_dst,$_src1,$_src2,$_ystride,$_residue);
 return;
}
function _oc_frag_recon_inter2_c($_dst,$_src1,$_src2,$_ystride,$_residue) {
 $_dst = $_dst|0;
 $_src1 = $_src1|0;
 $_src2 = $_src2|0;
 $_ystride = $_ystride|0;
 $_residue = $_residue|0;
 var $$015 = 0, $$024 = 0, $$06 = 0, $$lobit = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $exitcond8 = 0, $i$07 = 0, $j$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$015 = $_src1;$$024 = $_src2;$$06 = $_dst;$i$07 = 0;
 while(1) {
  $0 = $i$07 << 3;
  $j$03 = 0;
  while(1) {
   $1 = (($j$03) + ($0))|0;
   $2 = (($_residue) + ($1<<1)|0);
   $3 = HEAP16[$2>>1]|0;
   $4 = $3 << 16 >> 16;
   $5 = (($$015) + ($j$03)|0);
   $6 = HEAP8[$5>>0]|0;
   $7 = $6&255;
   $8 = (($$024) + ($j$03)|0);
   $9 = HEAP8[$8>>0]|0;
   $10 = $9&255;
   $11 = (($10) + ($7))|0;
   $12 = $11 >>> 1;
   $13 = (($12) + ($4))|0;
   $$lobit = $13 >>> 31;
   $14 = (($$lobit) + 255)|0;
   $15 = ($13|0)>(255);
   $16 = $15 ? 255 : $13;
   $17 = $14 & $16;
   $18 = $17&255;
   $19 = (($$06) + ($j$03)|0);
   HEAP8[$19>>0] = $18;
   $20 = (($j$03) + 1)|0;
   $exitcond = ($20|0)==(8);
   if ($exitcond) {
    break;
   } else {
    $j$03 = $20;
   }
  }
  $21 = (($$06) + ($_ystride)|0);
  $22 = (($$015) + ($_ystride)|0);
  $23 = (($$024) + ($_ystride)|0);
  $24 = (($i$07) + 1)|0;
  $exitcond8 = ($24|0)==(8);
  if ($exitcond8) {
   break;
  } else {
   $$015 = $22;$$024 = $23;$$06 = $21;$i$07 = $24;
  }
 }
 return;
}
function _oc_restore_fpu($_state) {
 $_state = $_state|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_state)) + 96|0);
 $1 = HEAP32[$0>>2]|0;
 FUNCTION_TABLE_v[$1 & 8191]();
 return;
}
function _oc_restore_fpu_c() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _oc_huff_trees_unpack($_opb,$_nodes) {
 $_opb = $_opb|0;
 $_nodes = $_nodes|0;
 var $$0$i = 0, $$1$ph = 0, $$2 = 0, $$not$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $depth$0$i = 0, $depth$0$i$lcssa = 0, $i$02 = 0, $nodes = 0, $occupancy$0$i = 0, $storage = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 6144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $nodes = sp + 4|0;
 $storage = sp;
 $0 = ((($nodes)) + 2|0);
 $i$02 = 0;
 while(1) {
  $1 = (_oc_huff_tree_unpack($_opb,$nodes,511)|0);
  $2 = ($1|0)<(0);
  if ($2) {
   $$1$ph = $1;
   break;
  }
  $3 = (_oc_huff_tree_collapse_size($nodes,0)|0);
  $4 = (_calloc(1,$3)|0);
  HEAP32[$storage>>2] = $4;
  $5 = ($4|0)==(0|0);
  if ($5) {
   $$1$ph = -1;
   break;
  }
  $6 = (_oc_huff_tree_mindepth($nodes)|0);
  $7 = 1 << $6;
  $depth$0$i = $6;$occupancy$0$i = $7;
  while(1) {
   $8 = (($depth$0$i) + 1)|0;
   $9 = (_oc_huff_tree_occupancy($nodes,$8)|0);
   $10 = ($9|0)>($occupancy$0$i|0);
   if (!($10)) {
    $depth$0$i$lcssa = $depth$0$i;
    break;
   }
   $11 = $depth$0$i >> 31;
   $$not$i = $11 ^ -1;
   $12 = $depth$0$i & $$not$i;
   $13 = 1 << $12;
   $14 = ($9|0)<($13|0);
   if ($14) {
    $depth$0$i$lcssa = $depth$0$i;
    break;
   } else {
    $depth$0$i = $8;$occupancy$0$i = $9;
   }
  }
  $15 = ($depth$0$i$lcssa|0)<(2);
  if ($15) {
   $16 = (_oc_huff_tree_copy($nodes,$storage)|0);
   $$0$i = $16;
  } else {
   $17 = 4 << $depth$0$i$lcssa;
   $18 = (($17) + 4)|0;
   $19 = $depth$0$i$lcssa&255;
   HEAP8[$4>>0] = $19;
   $20 = (($4) + ($18)|0);
   HEAP32[$storage>>2] = $20;
   $21 = HEAP8[$0>>0]|0;
   $22 = ((($4)) + 2|0);
   HEAP8[$22>>0] = $21;
   $23 = ((($4)) + 4|0);
   _oc_huff_node_fill($23,$nodes,$depth$0$i$lcssa,$depth$0$i$lcssa,$storage);
   $$0$i = $4;
  }
  $24 = (($_nodes) + ($i$02<<2)|0);
  HEAP32[$24>>2] = $$0$i;
  $25 = (($i$02) + 1)|0;
  $26 = ($25|0)<(80);
  if ($26) {
   $i$02 = $25;
  } else {
   $$2 = 0;
   label = 12;
   break;
  }
 }
 if ((label|0) == 12) {
  STACKTOP = sp;return ($$2|0);
 }
 $$2 = $$1$ph;
 STACKTOP = sp;return ($$2|0);
}
function _oc_huff_tree_unpack($_opb,$_binodes,$_nbinodes) {
 $_opb = $_opb|0;
 $_binodes = $_binodes|0;
 $_nbinodes = $_nbinodes|0;
 var $$ = 0, $$0 = 0, $$sum = 0, $$sum1 = 0, $$sum2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $exitcond = 0, $i$08 = 0, $i$14 = 0, $j$05 = 0, $nused$1$lcssa = 0, $nused$17 = 0, $nused$23 = 0, $smax = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($_nbinodes|0)<(1);
 if ($0) {
  $$0 = -20;
  return ($$0|0);
 }
 $1 = (_oc_pack_read1($_opb)|0);
 $2 = (_oc_pack_bytes_left($_opb)|0);
 $3 = ($2|0)<(0);
 if ($3) {
  $$0 = -20;
  return ($$0|0);
 }
 $4 = ($1|0)==(0);
 if ($4) {
  HEAP8[$_binodes>>0] = 1;
  $5 = ((($_binodes)) + 2|0);
  HEAP8[$5>>0] = 1;
  $6 = ((($_binodes)) + 12|0);
  $7 = ((($_binodes)) + 4|0);
  HEAP32[$7>>2] = $6;
  $8 = (($_nbinodes) + -1)|0;
  $9 = (_oc_huff_tree_unpack($_opb,$6,$8)|0);
  $10 = ($9|0)>(-1);
  if ($10) {
   $11 = (($9) + 1)|0;
   $12 = (($_binodes) + (($11*12)|0)|0);
   $13 = ((($_binodes)) + 8|0);
   HEAP32[$13>>2] = $12;
   $14 = (($_nbinodes) - ($11))|0;
   $15 = (_oc_huff_tree_unpack($_opb,$12,$14)|0);
   $16 = ($15|0)<(0);
   $17 = $16 ? 0 : $11;
   $$ = (($17) + ($15))|0;
   return ($$|0);
  } else {
   $$0 = $9;
   return ($$0|0);
  }
 }
 $18 = (_oc_pack_read($_opb,5)|0);
 $19 = (_oc_pack_bytes_left($_opb)|0);
 $20 = ($19|0)<(0);
 if ($20) {
  $$0 = -20;
  return ($$0|0);
 }
 $21 = (530010 + ($18)|0);
 $22 = HEAP8[$21>>0]|0;
 $23 = $22&255;
 $24 = $23 << 1;
 $25 = (($24) + -1)|0;
 $26 = ($25|0)>($_nbinodes|0);
 if ($26) {
  $$0 = -20;
  return ($$0|0);
 }
 $27 = -8040 >>> $18;
 $28 = $27 & 1;
 $29 = ($28|0)==(0);
 if ($29) {
  $nused$1$lcssa = 0;
 } else {
  $i$08 = 1;$nused$17 = 0;
  while(1) {
   $30 = (($i$08) + ($nused$17))|0;
   $31 = ($i$08|0)>(0);
   if ($31) {
    $j$05 = 0;
    while(1) {
     $$sum = (($j$05) + ($nused$17))|0;
     $32 = (($_binodes) + (($$sum*12)|0)|0);
     HEAP8[$32>>0] = 1;
     $33 = (((($_binodes) + (($$sum*12)|0)|0)) + 2|0);
     HEAP8[$33>>0] = 1;
     $34 = $j$05 << 1;
     $$sum1 = (($34) + ($30))|0;
     $35 = (($_binodes) + (($$sum1*12)|0)|0);
     $36 = (((($_binodes) + (($$sum*12)|0)|0)) + 4|0);
     HEAP32[$36>>2] = $35;
     $$sum2 = (($$sum1) + 1)|0;
     $37 = (($_binodes) + (($$sum2*12)|0)|0);
     $38 = (((($_binodes) + (($$sum*12)|0)|0)) + 8|0);
     HEAP32[$38>>2] = $37;
     $39 = (($j$05) + 1)|0;
     $exitcond = ($39|0)==($i$08|0);
     if ($exitcond) {
      break;
     } else {
      $j$05 = $39;
     }
    }
   }
   $40 = $i$08 << 1;
   $41 = ($40|0)<($23|0);
   if ($41) {
    $i$08 = $40;$nused$17 = $30;
   } else {
    $nused$1$lcssa = $30;
    break;
   }
  }
 }
 $42 = (530042 + ($18)|0);
 $43 = HEAP8[$42>>0]|0;
 $44 = $43&255;
 $45 = ($23>>>0)>(1);
 $smax = $45 ? $23 : 1;
 $i$14 = 0;$nused$23 = $nused$1$lcssa;
 while(1) {
  $46 = (($nused$23) + 1)|0;
  $47 = (($_binodes) + (($nused$23*12)|0)|0);
  HEAP8[$47>>0] = 0;
  $48 = (((($_binodes) + (($nused$23*12)|0)|0)) + 2|0);
  HEAP8[$48>>0] = 1;
  $49 = (($i$14) + ($44))|0;
  $50 = $49&255;
  $51 = (((($_binodes) + (($nused$23*12)|0)|0)) + 1|0);
  HEAP8[$51>>0] = $50;
  $52 = (($i$14) + 1)|0;
  $53 = ($52|0)<($23|0);
  if ($53) {
   $i$14 = $52;$nused$23 = $46;
  } else {
   break;
  }
 }
 $54 = (($nused$1$lcssa) + ($smax))|0;
 $$0 = $54;
 return ($$0|0);
}
function _oc_huff_tree_collapse_size($_binode,$_depth) {
 $_binode = $_binode|0;
 $_depth = $_depth|0;
 var $$not = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $depth$0 = 0, $depth$0$lcssa = 0, $occupancy$0 = 0, $or$cond = 0, $size$0$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$_binode>>0]|0;
 $1 = ($0<<24>>24)!=(0);
 $2 = ($_depth|0)>(0);
 $or$cond = $2 & $1;
 if ($or$cond) {
  $3 = ((($_binode)) + 4|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = (($_depth) + -1)|0;
  $6 = (_oc_huff_tree_collapse_size($4,$5)|0);
  $7 = ((($_binode)) + 8|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = (_oc_huff_tree_collapse_size($8,$5)|0);
  $10 = (($9) + ($6))|0;
  return ($10|0);
 }
 $11 = (_oc_huff_tree_mindepth($_binode)|0);
 $12 = 1 << $11;
 $depth$0 = $11;$occupancy$0 = $12;
 while(1) {
  $13 = (($depth$0) + 1)|0;
  $14 = (_oc_huff_tree_occupancy($_binode,$13)|0);
  $15 = ($14|0)>($occupancy$0|0);
  if (!($15)) {
   $depth$0$lcssa = $depth$0;
   break;
  }
  $16 = $depth$0 >> 31;
  $$not = $16 ^ -1;
  $17 = $depth$0 & $$not;
  $18 = 1 << $17;
  $19 = ($14|0)<($18|0);
  if ($19) {
   $depth$0$lcssa = $depth$0;
   break;
  } else {
   $depth$0 = $13;$occupancy$0 = $14;
  }
 }
 $20 = ($depth$0$lcssa|0)>(0);
 $21 = 4 << $depth$0$lcssa;
 $22 = (($21) + 4)|0;
 $size$0$i = $20 ? $22 : 4;
 if ($20) {
  $23 = ((($_binode)) + 4|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = (($depth$0$lcssa) + -1)|0;
  $26 = (_oc_huff_tree_collapse_size($24,$25)|0);
  $27 = (($26) + ($size$0$i))|0;
  $28 = ((($_binode)) + 8|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = (_oc_huff_tree_collapse_size($29,$25)|0);
  $31 = (($27) + ($30))|0;
  return ($31|0);
 } else {
  return ($size$0$i|0);
 }
 return (0)|0;
}
function _oc_huff_tree_mindepth($_binode) {
 $_binode = $_binode|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$_binode>>0]|0;
 $1 = ($0<<24>>24)==(0);
 if ($1) {
  return 0;
 } else {
  $2 = ((($_binode)) + 4|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = (_oc_huff_tree_mindepth($3)|0);
  $5 = ((($_binode)) + 8|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (_oc_huff_tree_mindepth($6)|0);
  $8 = ($7|0)<($4|0);
  $9 = $8 ? $7 : $4;
  $10 = (($9) + 1)|0;
  return ($10|0);
 }
 return (0)|0;
}
function _oc_huff_tree_occupancy($_binode,$_depth) {
 $_binode = $_binode|0;
 $_depth = $_depth|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $_binode$tr4 = 0, $_depth$tr5 = 0, $accumulator$tr$lcssa = 0, $accumulator$tr3 = 0, $or$cond = 0, $or$cond2 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$_binode>>0]|0;
 $1 = ($0<<24>>24)==(0);
 $2 = ($_depth|0)<(1);
 $or$cond2 = $2 | $1;
 if ($or$cond2) {
  $accumulator$tr$lcssa = 1;
  return ($accumulator$tr$lcssa|0);
 } else {
  $_binode$tr4 = $_binode;$_depth$tr5 = $_depth;$accumulator$tr3 = 1;
 }
 while(1) {
  $3 = ((($_binode$tr4)) + 4|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = (($_depth$tr5) + -1)|0;
  $6 = (_oc_huff_tree_occupancy($4,$5)|0);
  $7 = ((($_binode$tr4)) + 8|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = (($6) + ($accumulator$tr3))|0;
  $10 = HEAP8[$8>>0]|0;
  $11 = ($10<<24>>24)==(0);
  $12 = ($_depth$tr5|0)<(2);
  $or$cond = $12 | $11;
  if ($or$cond) {
   $accumulator$tr$lcssa = $9;
   break;
  } else {
   $_binode$tr4 = $8;$_depth$tr5 = $5;$accumulator$tr3 = $9;
  }
 }
 return ($accumulator$tr$lcssa|0);
}
function _oc_huff_tree_copy($_node,$_storage) {
 $_node = $_node|0;
 $_storage = $_storage|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$03 = 0;
 var $size$0$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$_node>>0]|0;
 $1 = $0&255;
 $2 = ($0<<24>>24)!=(0);
 $3 = 4 << $1;
 $4 = (($3) + 4)|0;
 $size$0$i = $2 ? $4 : 4;
 $5 = HEAP32[$_storage>>2]|0;
 HEAP8[$5>>0] = $0;
 $6 = (($5) + ($size$0$i)|0);
 HEAP32[$_storage>>2] = $6;
 $7 = ((($_node)) + 2|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = ((($5)) + 2|0);
 HEAP8[$9>>0] = $8;
 $10 = HEAP8[$_node>>0]|0;
 $11 = ($10<<24>>24)==(0);
 if ($11) {
  $34 = ((($_node)) + 1|0);
  $35 = HEAP8[$34>>0]|0;
  $36 = ((($5)) + 1|0);
  HEAP8[$36>>0] = $35;
  return ($5|0);
 }
 $12 = $10&255;
 $13 = 1 << $12;
 $14 = ($10<<24>>24)==(31);
 if ($14) {
  return ($5|0);
 } else {
  $i$03 = 0;
 }
 while(1) {
  $16 = (((($_node)) + 4|0) + ($i$03<<2)|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = (_oc_huff_tree_copy($17,$_storage)|0);
  $19 = (((($5)) + 4|0) + ($i$03<<2)|0);
  HEAP32[$19>>2] = $18;
  $20 = HEAP8[$_node>>0]|0;
  $21 = $20&255;
  $22 = ((($18)) + 2|0);
  $23 = HEAP8[$22>>0]|0;
  $24 = $23&255;
  $25 = (($21) - ($24))|0;
  $26 = 1 << $25;
  $27 = (($26) + ($i$03))|0;
  $28 = (($i$03) + 1)|0;
  $29 = ($28|0)<($27|0);
  $30 = $18;
  if ($29) {
   $32 = $28;
   while(1) {
    $31 = (((($5)) + 4|0) + ($32<<2)|0);
    HEAP32[$31>>2] = $30;
    $33 = (($32) + 1)|0;
    $exitcond = ($33|0)==($27|0);
    if ($exitcond) {
     $$lcssa = $27;
     break;
    } else {
     $32 = $33;
    }
   }
  } else {
   $$lcssa = $28;
  }
  $15 = ($$lcssa|0)<($13|0);
  if ($15) {
   $i$03 = $$lcssa;
  } else {
   break;
  }
 }
 return ($5|0);
}
function _oc_huff_node_fill($_nodes,$_binode,$_level,$_depth,$_storage) {
 $_nodes = $_nodes|0;
 $_binode = $_binode|0;
 $_level = $_level|0;
 $_depth = $_depth|0;
 $_storage = $_storage|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $_binode$tr$lcssa = 0, $_binode$tr3 = 0, $_level$tr$lcssa = 0, $_level$tr4 = 0, $_nodes$tr$lcssa = 0, $_nodes$tr2 = 0, $exitcond = 0, $exitcond19 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($_level|0)<(1);
 L1: do {
  if ($0) {
   $_binode$tr$lcssa = $_binode;$_level$tr$lcssa = $_level;$_nodes$tr$lcssa = $_nodes;
  } else {
   $_binode$tr3 = $_binode;$_level$tr4 = $_level;$_nodes$tr2 = $_nodes;
   while(1) {
    $1 = HEAP8[$_binode$tr3>>0]|0;
    $2 = ($1<<24>>24)==(0);
    if ($2) {
     $_binode$tr$lcssa = $_binode$tr3;$_level$tr$lcssa = $_level$tr4;$_nodes$tr$lcssa = $_nodes$tr2;
     break L1;
    }
    $14 = (($_level$tr4) + -1)|0;
    $15 = ((($_binode$tr3)) + 4|0);
    $16 = HEAP32[$15>>2]|0;
    _oc_huff_node_fill($_nodes$tr2,$16,$14,$_depth,$_storage);
    $17 = 1 << $14;
    $18 = (($_nodes$tr2) + ($17<<2)|0);
    $19 = ((($_binode$tr3)) + 8|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = ($_level$tr4|0)<(2);
    if ($21) {
     $_binode$tr$lcssa = $20;$_level$tr$lcssa = $14;$_nodes$tr$lcssa = $18;
     break;
    } else {
     $_binode$tr3 = $20;$_level$tr4 = $14;$_nodes$tr2 = $18;
    }
   }
  }
 } while(0);
 $3 = (($_depth) - ($_level$tr$lcssa))|0;
 $4 = $3&255;
 $5 = ((($_binode$tr$lcssa)) + 2|0);
 HEAP8[$5>>0] = $4;
 $6 = (_oc_huff_tree_collapse($_binode$tr$lcssa,$_storage)|0);
 HEAP32[$_nodes$tr$lcssa>>2] = $6;
 $7 = 1 << $_level$tr$lcssa;
 $8 = ($7|0)>(1);
 if (!($8)) {
  return;
 }
 $9 = $6;
 $10 = ((($_nodes$tr$lcssa)) + 4|0);
 HEAP32[$10>>2] = $9;
 $exitcond19 = ($_level$tr$lcssa|0)==(1);
 if ($exitcond19) {
  return;
 } else {
  $12 = 2;
 }
 while(1) {
  $$pre = HEAP32[$_nodes$tr$lcssa>>2]|0;
  $11 = (($_nodes$tr$lcssa) + ($12<<2)|0);
  HEAP32[$11>>2] = $$pre;
  $13 = (($12) + 1)|0;
  $exitcond = ($13|0)==($7|0);
  if ($exitcond) {
   break;
  } else {
   $12 = $13;
  }
 }
 return;
}
function _oc_huff_tree_collapse($_binode,$_storage) {
 $_binode = $_binode|0;
 $_storage = $_storage|0;
 var $$0 = 0, $$not = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $depth$0 = 0, $depth$0$lcssa = 0, $occupancy$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_oc_huff_tree_mindepth($_binode)|0);
 $1 = 1 << $0;
 $depth$0 = $0;$occupancy$0 = $1;
 while(1) {
  $2 = (($depth$0) + 1)|0;
  $3 = (_oc_huff_tree_occupancy($_binode,$2)|0);
  $4 = ($3|0)>($occupancy$0|0);
  if (!($4)) {
   $depth$0$lcssa = $depth$0;
   break;
  }
  $5 = $depth$0 >> 31;
  $$not = $5 ^ -1;
  $6 = $depth$0 & $$not;
  $7 = 1 << $6;
  $8 = ($3|0)<($7|0);
  if ($8) {
   $depth$0$lcssa = $depth$0;
   break;
  } else {
   $depth$0 = $2;$occupancy$0 = $3;
  }
 }
 $9 = ($depth$0$lcssa|0)<(2);
 if ($9) {
  $10 = (_oc_huff_tree_copy($_binode,$_storage)|0);
  $$0 = $10;
  return ($$0|0);
 } else {
  $11 = 4 << $depth$0$lcssa;
  $12 = (($11) + 4)|0;
  $13 = HEAP32[$_storage>>2]|0;
  $14 = $depth$0$lcssa&255;
  HEAP8[$13>>0] = $14;
  $15 = (($13) + ($12)|0);
  HEAP32[$_storage>>2] = $15;
  $16 = ((($_binode)) + 2|0);
  $17 = HEAP8[$16>>0]|0;
  $18 = ((($13)) + 2|0);
  HEAP8[$18>>0] = $17;
  $19 = ((($13)) + 4|0);
  _oc_huff_node_fill($19,$_binode,$depth$0$lcssa,$depth$0$lcssa,$_storage);
  $$0 = $13;
  return ($$0|0);
 }
 return (0)|0;
}
function _oc_huff_trees_copy($_dst,$_src) {
 $_dst = $_dst|0;
 $_src = $_src|0;
 var $$0 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$03 = 0, $i$03$lcssa = 0, $storage = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $storage = sp;
 $i$03 = 0;
 while(1) {
  $0 = (($_src) + ($i$03<<2)|0);
  $1 = HEAP32[$0>>2]|0;
  $2 = (_oc_huff_tree_size($1)|0);
  $3 = (_calloc(1,$2)|0);
  HEAP32[$storage>>2] = $3;
  $4 = ($3|0)==(0|0);
  if ($4) {
   $i$03$lcssa = $i$03;
   break;
  }
  $10 = (_oc_huff_tree_copy($1,$storage)|0);
  $11 = (($_dst) + ($i$03<<2)|0);
  HEAP32[$11>>2] = $10;
  $12 = (($i$03) + 1)|0;
  $13 = ($12|0)<(80);
  if ($13) {
   $i$03 = $12;
  } else {
   $$0 = 0;
   label = 6;
   break;
  }
 }
 if ((label|0) == 6) {
  STACKTOP = sp;return ($$0|0);
 }
 $5 = ($i$03$lcssa|0)>(0);
 if ($5) {
  $$in = $i$03$lcssa;
 } else {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 while(1) {
  $6 = (($$in) + -1)|0;
  $7 = (($_dst) + ($6<<2)|0);
  $8 = HEAP32[$7>>2]|0;
  _free($8);
  $9 = ($$in|0)>(1);
  if ($9) {
   $$in = $6;
  } else {
   $$0 = -1;
   break;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _oc_huff_tree_size($_node) {
 $_node = $_node|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$02 = 0;
 var $size$0$i = 0, $size$01 = 0, $size$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$_node>>0]|0;
 $1 = $0&255;
 $2 = ($0<<24>>24)!=(0);
 $3 = 4 << $1;
 $4 = (($3) + 4)|0;
 $size$0$i = $2 ? $4 : 4;
 $5 = ($0<<24>>24)==(0);
 if ($5) {
  $size$1 = $size$0$i;
  return ($size$1|0);
 }
 $6 = 1 << $1;
 $7 = ($0<<24>>24)==(31);
 if ($7) {
  $size$1 = $size$0$i;
  return ($size$1|0);
 } else {
  $i$02 = 0;$size$01 = $size$0$i;
 }
 while(1) {
  $8 = (((($_node)) + 4|0) + ($i$02<<2)|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = (_oc_huff_tree_size($9)|0);
  $11 = (($10) + ($size$01))|0;
  $12 = ((($9)) + 2|0);
  $13 = HEAP8[$12>>0]|0;
  $14 = $13&255;
  $15 = (($1) - ($14))|0;
  $16 = 1 << $15;
  $17 = (($16) + ($i$02))|0;
  $18 = ($17|0)<($6|0);
  if ($18) {
   $i$02 = $17;$size$01 = $11;
  } else {
   $size$1 = $11;
   break;
  }
 }
 return ($size$1|0);
}
function _oc_huff_trees_clear($_nodes) {
 $_nodes = $_nodes|0;
 var $0 = 0, $1 = 0, $2 = 0, $exitcond = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $i$01 = 0;
 while(1) {
  $0 = (($_nodes) + ($i$01<<2)|0);
  $1 = HEAP32[$0>>2]|0;
  _free($1);
  $2 = (($i$01) + 1)|0;
  $exitcond = ($2|0)==(80);
  if ($exitcond) {
   break;
  } else {
   $i$01 = $2;
  }
 }
 return;
}
function _oc_huff_token_decode($_opb,$_node) {
 $_opb = $_opb|0;
 $_node = $_node|0;
 var $$$i$i = 0, $$0$lcssa = 0, $$02 = 0, $$3$i$i = 0, $$pre = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $available$0$lcssa$i$i = 0, $available$06$i$i = 0, $ptr$0$lcssa$i$i = 0, $ptr$04$i$i = 0, $window$0$lcssa$i$i = 0, $window$05$i$i = 0, $window$1$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$_node>>0]|0;
 $1 = ($0<<24>>24)==(0);
 if ($1) {
  $$0$lcssa = $_node;
  $41 = ((($$0$lcssa)) + 1|0);
  $42 = HEAP8[$41>>0]|0;
  $43 = $42&255;
  return ($43|0);
 }
 $2 = ((($_opb)) + 12|0);
 $3 = ((($_opb)) + 4|0);
 $4 = ((($_opb)) + 8|0);
 $$pre = HEAP32[$2>>2]|0;
 $$pre5 = HEAP32[$_opb>>2]|0;
 $$02 = $_node;$44 = $$pre5;$6 = $0;$8 = $$pre;
 while(1) {
  $5 = $6&255;
  $7 = ($8|0)<($5|0);
  if ($7) {
   $9 = HEAP32[$3>>2]|0;
   $10 = HEAP32[$4>>2]|0;
   $11 = ($9>>>0)<($10>>>0);
   $$$i$i = $11 ? $8 : 1073741824;
   $12 = ($$$i$i|0)<(25);
   if ($12) {
    $available$06$i$i = $8;$ptr$04$i$i = $9;$window$05$i$i = $44;
    while(1) {
     $13 = (($available$06$i$i) + 8)|0;
     $14 = ((($ptr$04$i$i)) + 1|0);
     $15 = HEAP8[$ptr$04$i$i>>0]|0;
     $16 = $15&255;
     $17 = (24 - ($available$06$i$i))|0;
     $18 = $16 << $17;
     $19 = $18 | $window$05$i$i;
     $20 = ($14>>>0)<($10>>>0);
     $$3$i$i = $20 ? $13 : 1073741824;
     $21 = ($$3$i$i|0)<(25);
     if ($21) {
      $available$06$i$i = $13;$ptr$04$i$i = $14;$window$05$i$i = $19;
     } else {
      $available$0$lcssa$i$i = $$3$i$i;$ptr$0$lcssa$i$i = $14;$window$0$lcssa$i$i = $19;
      break;
     }
    }
   } else {
    $available$0$lcssa$i$i = $$$i$i;$ptr$0$lcssa$i$i = $9;$window$0$lcssa$i$i = $44;
   }
   HEAP32[$3>>2] = $ptr$0$lcssa$i$i;
   $22 = ($available$0$lcssa$i$i|0)<($5|0);
   if ($22) {
    $23 = HEAP8[$ptr$0$lcssa$i$i>>0]|0;
    $24 = $23&255;
    $25 = $available$0$lcssa$i$i & 7;
    $26 = $24 >>> $25;
    $27 = $26 | $window$0$lcssa$i$i;
    $window$1$i$i = $27;
   } else {
    $window$1$i$i = $window$0$lcssa$i$i;
   }
   HEAP32[$2>>2] = $available$0$lcssa$i$i;
   HEAP32[$_opb>>2] = $window$1$i$i;
   $30 = $window$1$i$i;$38 = $available$0$lcssa$i$i;
  } else {
   $30 = $44;$38 = $8;
  }
  $28 = (32 - ($5))|0;
  $29 = $30 >>> $28;
  $31 = (((($$02)) + 4|0) + ($29<<2)|0);
  $32 = HEAP32[$31>>2]|0;
  $33 = ((($32)) + 2|0);
  $34 = HEAP8[$33>>0]|0;
  $35 = $34&255;
  $36 = $30 << $35;
  HEAP32[$_opb>>2] = $36;
  $37 = (($38) - ($35))|0;
  HEAP32[$2>>2] = $37;
  $39 = HEAP8[$32>>0]|0;
  $40 = ($39<<24>>24)==(0);
  if ($40) {
   $$0$lcssa = $32;
   break;
  } else {
   $$02 = $32;$44 = $36;$6 = $39;$8 = $37;
  }
 }
 $41 = ((($$0$lcssa)) + 1|0);
 $42 = HEAP8[$41>>0]|0;
 $43 = $42&255;
 return ($43|0);
}
function _th_info_init($_info) {
 $_info = $_info|0;
 var $0 = 0, $1 = 0, $2 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 dest=$_info; stop=dest+60|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 HEAP8[$_info>>0] = 3;
 $0 = ((($_info)) + 1|0);
 HEAP8[$0>>0] = 2;
 $1 = ((($_info)) + 2|0);
 HEAP8[$1>>0] = 1;
 $2 = ((($_info)) + 60|0);
 HEAP32[$2>>2] = 6;
 return;
}
function _th_info_clear($_info) {
 $_info = $_info|0;
 var dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 dest=$_info; stop=dest+64|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 return;
}
function _th_comment_init($_tc) {
 $_tc = $_tc|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$_tc>>2]=0|0;HEAP32[$_tc+4>>2]=0|0;HEAP32[$_tc+8>>2]=0|0;HEAP32[$_tc+12>>2]=0|0;
 return;
}
function _th_comment_clear($_tc) {
 $_tc = $_tc|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($_tc|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($_tc)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)>(0);
 if ($3) {
  $i$01 = 0;
  while(1) {
   $4 = HEAP32[$_tc>>2]|0;
   $5 = (($4) + ($i$01<<2)|0);
   $6 = HEAP32[$5>>2]|0;
   _free($6);
   $7 = (($i$01) + 1)|0;
   $8 = HEAP32[$1>>2]|0;
   $9 = ($7|0)<($8|0);
   if ($9) {
    $i$01 = $7;
   } else {
    break;
   }
  }
 }
 $10 = HEAP32[$_tc>>2]|0;
 _free($10);
 $11 = ((($_tc)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 _free($12);
 $13 = ((($_tc)) + 12|0);
 $14 = HEAP32[$13>>2]|0;
 _free($14);
 ;HEAP32[$_tc>>2]=0|0;HEAP32[$_tc+4>>2]=0|0;HEAP32[$_tc+8>>2]=0|0;HEAP32[$_tc+12>>2]=0|0;
 return;
}
function _oc_set_chroma_mvs00($_cbmvs,$_lbmvs) {
 $_cbmvs = $_cbmvs|0;
 $_lbmvs = $_lbmvs|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$_lbmvs>>0]|0;
 $1 = $0 << 24 >> 24;
 $2 = ((($_lbmvs)) + 2|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3 << 24 >> 24;
 $5 = (($4) + ($1))|0;
 $6 = ((($_lbmvs)) + 4|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7 << 24 >> 24;
 $9 = (($5) + ($8))|0;
 $10 = ((($_lbmvs)) + 6|0);
 $11 = HEAP8[$10>>0]|0;
 $12 = $11 << 24 >> 24;
 $13 = (($9) + ($12))|0;
 $14 = ((($_lbmvs)) + 1|0);
 $15 = HEAP8[$14>>0]|0;
 $16 = $15 << 24 >> 24;
 $17 = ((($_lbmvs)) + 3|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = $18 << 24 >> 24;
 $20 = (($19) + ($16))|0;
 $21 = ((($_lbmvs)) + 5|0);
 $22 = HEAP8[$21>>0]|0;
 $23 = $22 << 24 >> 24;
 $24 = (($20) + ($23))|0;
 $25 = ((($_lbmvs)) + 7|0);
 $26 = HEAP8[$25>>0]|0;
 $27 = $26 << 24 >> 24;
 $28 = (($24) + ($27))|0;
 $29 = $13 >> 31;
 $30 = (($13) + 2)|0;
 $31 = (($30) + ($29))|0;
 $32 = $31 >>> 2;
 $33 = $32&255;
 HEAP8[$_cbmvs>>0] = $33;
 $34 = $28 >> 31;
 $35 = (($28) + 2)|0;
 $36 = (($35) + ($34))|0;
 $37 = $36 >>> 2;
 $38 = $37&255;
 $39 = ((($_cbmvs)) + 1|0);
 HEAP8[$39>>0] = $38;
 return;
}
function _oc_set_chroma_mvs01($_cbmvs,$_lbmvs) {
 $_cbmvs = $_cbmvs|0;
 $_lbmvs = $_lbmvs|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$_lbmvs>>0]|0;
 $1 = $0 << 24 >> 24;
 $2 = ((($_lbmvs)) + 4|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3 << 24 >> 24;
 $5 = (($4) + ($1))|0;
 $6 = ((($_lbmvs)) + 1|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7 << 24 >> 24;
 $9 = ((($_lbmvs)) + 5|0);
 $10 = HEAP8[$9>>0]|0;
 $11 = $10 << 24 >> 24;
 $12 = (($11) + ($8))|0;
 $13 = $5 >> 31;
 $14 = (($5) + 1)|0;
 $15 = (($14) + ($13))|0;
 $16 = $15 >>> 1;
 $17 = $16&255;
 HEAP8[$_cbmvs>>0] = $17;
 $18 = $12 >> 31;
 $19 = (($12) + 1)|0;
 $20 = (($19) + ($18))|0;
 $21 = $20 >>> 1;
 $22 = $21&255;
 $23 = ((($_cbmvs)) + 1|0);
 HEAP8[$23>>0] = $22;
 $24 = ((($_lbmvs)) + 2|0);
 $25 = HEAP8[$24>>0]|0;
 $26 = $25 << 24 >> 24;
 $27 = ((($_lbmvs)) + 6|0);
 $28 = HEAP8[$27>>0]|0;
 $29 = $28 << 24 >> 24;
 $30 = (($29) + ($26))|0;
 $31 = ((($_lbmvs)) + 3|0);
 $32 = HEAP8[$31>>0]|0;
 $33 = $32 << 24 >> 24;
 $34 = ((($_lbmvs)) + 7|0);
 $35 = HEAP8[$34>>0]|0;
 $36 = $35 << 24 >> 24;
 $37 = (($36) + ($33))|0;
 $38 = $30 >> 31;
 $39 = (($30) + 1)|0;
 $40 = (($39) + ($38))|0;
 $41 = $40 >>> 1;
 $42 = $41&255;
 $43 = ((($_cbmvs)) + 2|0);
 HEAP8[$43>>0] = $42;
 $44 = $37 >> 31;
 $45 = (($37) + 1)|0;
 $46 = (($45) + ($44))|0;
 $47 = $46 >>> 1;
 $48 = $47&255;
 $49 = ((($_cbmvs)) + 3|0);
 HEAP8[$49>>0] = $48;
 return;
}
function _oc_set_chroma_mvs10($_cbmvs,$_lbmvs) {
 $_cbmvs = $_cbmvs|0;
 $_lbmvs = $_lbmvs|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$_lbmvs>>0]|0;
 $1 = $0 << 24 >> 24;
 $2 = ((($_lbmvs)) + 2|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3 << 24 >> 24;
 $5 = (($4) + ($1))|0;
 $6 = ((($_lbmvs)) + 1|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7 << 24 >> 24;
 $9 = ((($_lbmvs)) + 3|0);
 $10 = HEAP8[$9>>0]|0;
 $11 = $10 << 24 >> 24;
 $12 = (($11) + ($8))|0;
 $13 = $5 >> 31;
 $14 = (($5) + 1)|0;
 $15 = (($14) + ($13))|0;
 $16 = $15 >>> 1;
 $17 = $16&255;
 HEAP8[$_cbmvs>>0] = $17;
 $18 = $12 >> 31;
 $19 = (($12) + 1)|0;
 $20 = (($19) + ($18))|0;
 $21 = $20 >>> 1;
 $22 = $21&255;
 $23 = ((($_cbmvs)) + 1|0);
 HEAP8[$23>>0] = $22;
 $24 = ((($_lbmvs)) + 4|0);
 $25 = HEAP8[$24>>0]|0;
 $26 = $25 << 24 >> 24;
 $27 = ((($_lbmvs)) + 6|0);
 $28 = HEAP8[$27>>0]|0;
 $29 = $28 << 24 >> 24;
 $30 = (($29) + ($26))|0;
 $31 = ((($_lbmvs)) + 5|0);
 $32 = HEAP8[$31>>0]|0;
 $33 = $32 << 24 >> 24;
 $34 = ((($_lbmvs)) + 7|0);
 $35 = HEAP8[$34>>0]|0;
 $36 = $35 << 24 >> 24;
 $37 = (($36) + ($33))|0;
 $38 = $30 >> 31;
 $39 = (($30) + 1)|0;
 $40 = (($39) + ($38))|0;
 $41 = $40 >>> 1;
 $42 = $41&255;
 $43 = ((($_cbmvs)) + 4|0);
 HEAP8[$43>>0] = $42;
 $44 = $37 >> 31;
 $45 = (($37) + 1)|0;
 $46 = (($45) + ($44))|0;
 $47 = $46 >>> 1;
 $48 = $47&255;
 $49 = ((($_cbmvs)) + 5|0);
 HEAP8[$49>>0] = $48;
 return;
}
function _oc_set_chroma_mvs11($_cbmvs,$_lbmvs) {
 $_cbmvs = $_cbmvs|0;
 $_lbmvs = $_lbmvs|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $_lbmvs;
 $1 = $0;
 $2 = HEAPU8[$1>>0]|(HEAPU8[$1+1>>0]<<8)|(HEAPU8[$1+2>>0]<<16)|(HEAPU8[$1+3>>0]<<24);
 $3 = (($0) + 4)|0;
 $4 = $3;
 $5 = HEAPU8[$4>>0]|(HEAPU8[$4+1>>0]<<8)|(HEAPU8[$4+2>>0]<<16)|(HEAPU8[$4+3>>0]<<24);
 $6 = $_cbmvs;
 $7 = $6;
 HEAP8[$7>>0]=$2&255;HEAP8[$7+1>>0]=($2>>8)&255;HEAP8[$7+2>>0]=($2>>16)&255;HEAP8[$7+3>>0]=$2>>24;
 $8 = (($6) + 4)|0;
 $9 = $8;
 HEAP8[$9>>0]=$5&255;HEAP8[$9+1>>0]=($5>>8)&255;HEAP8[$9+2>>0]=($5>>16)&255;HEAP8[$9+3>>0]=$5>>24;
 return;
}
function _oc_ilog($_v) {
 $_v = $_v|0;
 var $$01 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $ret$0$lcssa = 0, $ret$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($_v|0)==(0);
 if ($0) {
  $ret$0$lcssa = 0;
 } else {
  $$01 = $_v;$ret$02 = 0;
  while(1) {
   $1 = $$01 >>> 1;
   $2 = (($ret$02) + 1)|0;
   $3 = ($1|0)==(0);
   if ($3) {
    $ret$0$lcssa = $2;
    break;
   } else {
    $$01 = $1;$ret$02 = $2;
   }
  }
 }
 return ($ret$0$lcssa|0);
}
function _oc_ycbcr_buffer_flip($_dst,$_src) {
 $_dst = $_dst|0;
 $_src = $_src|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$_src>>2]|0;
 HEAP32[$_dst>>2] = $0;
 $1 = ((($_src)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($_dst)) + 4|0);
 HEAP32[$3>>2] = $2;
 $4 = ((($_src)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (0 - ($5))|0;
 $7 = ((($_dst)) + 8|0);
 HEAP32[$7>>2] = $6;
 $8 = ((($_src)) + 12|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (1 - ($2))|0;
 $11 = Math_imul($10, $6)|0;
 $12 = (($9) + ($11)|0);
 $13 = ((($_dst)) + 12|0);
 HEAP32[$13>>2] = $12;
 $14 = ((($_src)) + 16|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($_dst)) + 16|0);
 HEAP32[$16>>2] = $15;
 $17 = ((($_src)) + 20|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($_dst)) + 20|0);
 HEAP32[$19>>2] = $18;
 $20 = ((($_src)) + 24|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = (0 - ($21))|0;
 $23 = ((($_dst)) + 24|0);
 HEAP32[$23>>2] = $22;
 $24 = ((($_src)) + 28|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = (1 - ($18))|0;
 $27 = Math_imul($26, $22)|0;
 $28 = (($25) + ($27)|0);
 $29 = ((($_dst)) + 28|0);
 HEAP32[$29>>2] = $28;
 $30 = ((($_src)) + 32|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = ((($_dst)) + 32|0);
 HEAP32[$32>>2] = $31;
 $33 = ((($_src)) + 36|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = ((($_dst)) + 36|0);
 HEAP32[$35>>2] = $34;
 $36 = ((($_src)) + 40|0);
 $37 = HEAP32[$36>>2]|0;
 $38 = (0 - ($37))|0;
 $39 = ((($_dst)) + 40|0);
 HEAP32[$39>>2] = $38;
 $40 = ((($_src)) + 44|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = (1 - ($34))|0;
 $43 = Math_imul($42, $38)|0;
 $44 = (($41) + ($43)|0);
 $45 = ((($_dst)) + 44|0);
 HEAP32[$45>>2] = $44;
 return;
}
function _oc_dequant_tables_init($_dequant,$_pp_dc_scale,$_qinfo) {
 $_dequant = $_dequant|0;
 $_pp_dc_scale = $_pp_dc_scale|0;
 $_qinfo = $_qinfo|0;
 var $$lcssa = 0, $$lcssa65 = 0, $$lcssa67 = 0, $$lcssa68 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $base = 0, $ci$018 = 0, $ci$018$us = 0;
 var $exitcond = 0, $exitcond40 = 0, $exitcond41 = 0, $exitcond42 = 0, $exitcond43 = 0, $exitcond44 = 0, $pli$026 = 0, $plj$514 = 0, $plj$514$us = 0, $qi$019 = 0, $qi$1 = 0, $qi$1$us = 0, $qi_end$0$ph = 0, $qri$020 = 0, $qti$029 = 0, $qtj$015 = 0, $qtj$015$us = 0, $zzi$013 = 0, $zzi$013$us = 0, dest = 0;
 var label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $base = sp;
 $0 = ($_pp_dc_scale|0)==(0|0);
 $qti$029 = 0;
 while(1) {
  $1 = (47568 + ($qti$029<<2)|0);
  $2 = (47576 + ($qti$029<<2)|0);
  $pli$026 = 0;
  while(1) {
   $3 = ((((($_qinfo)) + 320|0) + (($qti$029*36)|0)|0) + (($pli$026*12)|0)|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)<(0);
   if (!($5)) {
    $6 = ((((((($_qinfo)) + 320|0) + (($qti$029*36)|0)|0) + (($pli$026*12)|0)|0)) + 8|0);
    $7 = HEAP32[$1>>2]|0;
    $8 = HEAP32[$2>>2]|0;
    $9 = ((((((($_qinfo)) + 320|0) + (($qti$029*36)|0)|0) + (($pli$026*12)|0)|0)) + 4|0);
    $13 = $4;$qi$019 = 0;$qri$020 = 0;
    while(1) {
     $10 = HEAP32[$6>>2]|0;
     $11 = (($10) + ($qri$020<<6)|0);
     dest=$base; src=$11; stop=dest+64|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
     $12 = ($qri$020|0)==($13|0);
     if ($12) {
      $14 = (($qi$019) + 1)|0;
      $qi_end$0$ph = $14;
     } else {
      $15 = HEAP32[$9>>2]|0;
      $16 = (($15) + ($qri$020<<2)|0);
      $17 = HEAP32[$16>>2]|0;
      $18 = (($17) + ($qi$019))|0;
      $qi_end$0$ph = $18;
     }
     $19 = (($qri$020) + 1)|0;
     L13: do {
      if ($0) {
       $qi$1$us = $qi$019;
       L14: while(1) {
        $20 = (($_qinfo) + ($qi$1$us<<1)|0);
        $21 = HEAP16[$20>>1]|0;
        $22 = $21&65535;
        $23 = HEAP8[$base>>0]|0;
        $24 = $23&255;
        $25 = Math_imul($24, $22)|0;
        $26 = (($25>>>0) / 100)&-1;
        $27 = $26 << 2;
        $28 = ($27>>>0)>(4096);
        $29 = $28 ? 4096 : $27;
        $30 = (($7) - ($29))|0;
        $31 = ($7>>>0)<($29>>>0);
        $32 = $31 ? $30 : 0;
        $33 = (($7) - ($32))|0;
        $34 = $33&65535;
        $35 = (((($_dequant) + (($qi$1$us*24)|0)|0) + ($pli$026<<3)|0) + ($qti$029<<2)|0);
        $36 = HEAP32[$35>>2]|0;
        HEAP16[$36>>1] = $34;
        $37 = (((($_qinfo)) + 128|0) + ($qi$1$us<<1)|0);
        $zzi$013$us = 1;
        while(1) {
         $69 = HEAP16[$37>>1]|0;
         $70 = $69&65535;
         $71 = (530074 + ($zzi$013$us)|0);
         $72 = HEAP8[$71>>0]|0;
         $73 = $72&255;
         $74 = (($base) + ($73)|0);
         $75 = HEAP8[$74>>0]|0;
         $76 = $75&255;
         $77 = Math_imul($76, $70)|0;
         $78 = (($77>>>0) / 100)&-1;
         $79 = $78 << 2;
         $80 = ($79>>>0)>(4096);
         $81 = $80 ? 4096 : $79;
         $82 = (($8) - ($81))|0;
         $83 = ($8>>>0)<($81>>>0);
         $84 = $83 ? $82 : 0;
         $85 = (($8) - ($84))|0;
         $86 = $85&65535;
         $87 = (($36) + ($zzi$013$us<<1)|0);
         HEAP16[$87>>1] = $86;
         $88 = (($zzi$013$us) + 1)|0;
         $exitcond41 = ($88|0)==(64);
         if ($exitcond41) {
          $qtj$015$us = 0;
          break;
         } else {
          $zzi$013$us = $88;
         }
        }
        L18: while(1) {
         $89 = ($qtj$015$us|0)<($qti$029|0);
         $45 = $89 ? 3 : $pli$026;
         $90 = ($45|0)>(0);
         if ($90) {
          $plj$514$us = 0;
          while(1) {
           $39 = (((($_dequant) + (($qi$1$us*24)|0)|0) + ($plj$514$us<<3)|0) + ($qtj$015$us<<2)|0);
           $40 = HEAP32[$39>>2]|0;
           $41 = (_memcmp($36,$40,128)|0);
           $42 = ($41|0)==(0);
           if ($42) {
            $$lcssa68 = $40;
            label = 13;
            break L18;
           }
           $43 = (($plj$514$us) + 1)|0;
           $44 = ($43|0)<($45|0);
           if ($44) {
            $plj$514$us = $43;
           } else {
            break;
           }
          }
         }
         $38 = (($qtj$015$us) + 1)|0;
         if ($89) {
          $qtj$015$us = $38;
         } else {
          break;
         }
        }
        if ((label|0) == 13) {
         label = 0;
         $46 = $$lcssa68;
         HEAP32[$35>>2] = $46;
        }
        $47 = (($qi$1$us) + 1)|0;
        $48 = ($qi_end$0$ph|0)>($47|0);
        if (!($48)) {
         $$lcssa = $47;$169 = $13;
         break L13;
        }
        $54 = (($qi_end$0$ph) - ($47))|0;
        $50 = HEAP32[$6>>2]|0;
        $59 = (($47) - ($qi$019))|0;
        $91 = HEAP32[$9>>2]|0;
        $92 = (($91) + ($qri$020<<2)|0);
        $63 = HEAP32[$92>>2]|0;
        $65 = $63 << 1;
        $ci$018$us = 0;
        while(1) {
         $49 = ((($50) + ($qri$020<<6)|0) + ($ci$018$us)|0);
         $51 = HEAP8[$49>>0]|0;
         $52 = $51&255;
         $53 = Math_imul($52, $54)|0;
         $55 = ((($50) + ($19<<6)|0) + ($ci$018$us)|0);
         $56 = HEAP8[$55>>0]|0;
         $57 = $56&255;
         $58 = Math_imul($57, $59)|0;
         $60 = (($58) + ($53))|0;
         $61 = $60 << 1;
         $62 = (($61) + ($63))|0;
         $64 = (($62|0) / ($65|0))&-1;
         $66 = $64&255;
         $67 = (($base) + ($ci$018$us)|0);
         HEAP8[$67>>0] = $66;
         $68 = (($ci$018$us) + 1)|0;
         $exitcond42 = ($68|0)==(64);
         if ($exitcond42) {
          $qi$1$us = $47;
          continue L14;
         } else {
          $ci$018$us = $68;
         }
        }
       }
      } else {
       $qi$1 = $qi$019;
       L31: while(1) {
        $93 = (($_qinfo) + ($qi$1<<1)|0);
        $94 = HEAP16[$93>>1]|0;
        $95 = $94&65535;
        $96 = HEAP8[$base>>0]|0;
        $97 = $96&255;
        $98 = Math_imul($97, $95)|0;
        $99 = (($98>>>0) / 160)&-1;
        $100 = (($_pp_dc_scale) + ($qi$1<<2)|0);
        HEAP32[$100>>2] = $99;
        $101 = (($98>>>0) / 100)&-1;
        $102 = $101 << 2;
        $103 = ($102>>>0)>(4096);
        $104 = $103 ? 4096 : $102;
        $105 = (($7) - ($104))|0;
        $106 = ($7>>>0)<($104>>>0);
        $107 = $106 ? $105 : 0;
        $108 = (($7) - ($107))|0;
        $109 = $108&65535;
        $110 = (((($_dequant) + (($qi$1*24)|0)|0) + ($pli$026<<3)|0) + ($qti$029<<2)|0);
        $111 = HEAP32[$110>>2]|0;
        HEAP16[$111>>1] = $109;
        $112 = (((($_qinfo)) + 128|0) + ($qi$1<<1)|0);
        $zzi$013 = 1;
        while(1) {
         $113 = HEAP16[$112>>1]|0;
         $114 = $113&65535;
         $115 = (530074 + ($zzi$013)|0);
         $116 = HEAP8[$115>>0]|0;
         $117 = $116&255;
         $118 = (($base) + ($117)|0);
         $119 = HEAP8[$118>>0]|0;
         $120 = $119&255;
         $121 = Math_imul($120, $114)|0;
         $122 = (($121>>>0) / 100)&-1;
         $123 = $122 << 2;
         $124 = ($123>>>0)>(4096);
         $125 = $124 ? 4096 : $123;
         $126 = (($8) - ($125))|0;
         $127 = ($8>>>0)<($125>>>0);
         $128 = $127 ? $126 : 0;
         $129 = (($8) - ($128))|0;
         $130 = $129&65535;
         $131 = (($111) + ($zzi$013<<1)|0);
         HEAP16[$131>>1] = $130;
         $132 = (($zzi$013) + 1)|0;
         $exitcond = ($132|0)==(64);
         if ($exitcond) {
          $qtj$015 = 0;
          break;
         } else {
          $zzi$013 = $132;
         }
        }
        L35: while(1) {
         $133 = ($qtj$015|0)<($qti$029|0);
         $134 = $133 ? 3 : $pli$026;
         $135 = ($134|0)>(0);
         if ($135) {
          $plj$514 = 0;
          while(1) {
           $136 = (((($_dequant) + (($qi$1*24)|0)|0) + ($plj$514<<3)|0) + ($qtj$015<<2)|0);
           $137 = HEAP32[$136>>2]|0;
           $138 = (_memcmp($111,$137,128)|0);
           $139 = ($138|0)==(0);
           if ($139) {
            $$lcssa65 = $137;
            label = 25;
            break L35;
           }
           $140 = (($plj$514) + 1)|0;
           $141 = ($140|0)<($134|0);
           if ($141) {
            $plj$514 = $140;
           } else {
            break;
           }
          }
         }
         $142 = (($qtj$015) + 1)|0;
         if ($133) {
          $qtj$015 = $142;
         } else {
          break;
         }
        }
        if ((label|0) == 25) {
         label = 0;
         $143 = $$lcssa65;
         HEAP32[$110>>2] = $143;
        }
        $144 = (($qi$1) + 1)|0;
        $145 = ($qi_end$0$ph|0)>($144|0);
        if (!($145)) {
         $$lcssa67 = $144;
         break;
        }
        $146 = (($qi_end$0$ph) - ($144))|0;
        $147 = HEAP32[$6>>2]|0;
        $148 = (($144) - ($qi$019))|0;
        $149 = HEAP32[$9>>2]|0;
        $150 = (($149) + ($qri$020<<2)|0);
        $151 = HEAP32[$150>>2]|0;
        $152 = $151 << 1;
        $ci$018 = 0;
        while(1) {
         $153 = ((($147) + ($qri$020<<6)|0) + ($ci$018)|0);
         $154 = HEAP8[$153>>0]|0;
         $155 = $154&255;
         $156 = Math_imul($155, $146)|0;
         $157 = ((($147) + ($19<<6)|0) + ($ci$018)|0);
         $158 = HEAP8[$157>>0]|0;
         $159 = $158&255;
         $160 = Math_imul($159, $148)|0;
         $161 = (($160) + ($156))|0;
         $162 = $161 << 1;
         $163 = (($162) + ($151))|0;
         $164 = (($163|0) / ($152|0))&-1;
         $165 = $164&255;
         $166 = (($base) + ($ci$018)|0);
         HEAP8[$166>>0] = $165;
         $167 = (($ci$018) + 1)|0;
         $exitcond40 = ($167|0)==(64);
         if ($exitcond40) {
          $qi$1 = $144;
          continue L31;
         } else {
          $ci$018 = $167;
         }
        }
       }
       $$pre = HEAP32[$3>>2]|0;
       $$lcssa = $$lcssa67;$169 = $$pre;
      }
     } while(0);
     $168 = ($qri$020|0)<($169|0);
     if ($168) {
      $13 = $169;$qi$019 = $$lcssa;$qri$020 = $19;
     } else {
      break;
     }
    }
   }
   $170 = (($pli$026) + 1)|0;
   $exitcond43 = ($170|0)==(3);
   if ($exitcond43) {
    break;
   } else {
    $pli$026 = $170;
   }
  }
  $171 = (($qti$029) + 1)|0;
  $exitcond44 = ($171|0)==(2);
  if ($exitcond44) {
   break;
  } else {
   $qti$029 = $171;
  }
 }
 STACKTOP = sp;return;
}
function _oc_state_frag_recon_c($_state,$_fragi,$_pli,$_dct_coeffs,$_last_zzi,$_dc_quant) {
 $_state = $_state|0;
 $_fragi = $_fragi|0;
 $_pli = $_pli|0;
 $_dct_coeffs = $_dct_coeffs|0;
 $_last_zzi = $_last_zzi|0;
 $_dc_quant = $_dc_quant|0;
 var $$sum1 = 0, $$sum2 = 0, $$sum4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $9 = 0, $ci$05 = 0, $exitcond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($_last_zzi|0)<(2);
 $1 = HEAP16[$_dct_coeffs>>1]|0;
 $2 = $1 << 16 >> 16;
 $3 = $_dc_quant&65535;
 $4 = Math_imul($2, $3)|0;
 if ($0) {
  $5 = (($4) + 15)|0;
  $6 = $5 >>> 5;
  $7 = $6&65535;
  $ci$05 = 0;
  while(1) {
   $8 = (($_dct_coeffs) + ($ci$05<<1)|0);
   HEAP16[$8>>1] = $7;
   $9 = (($ci$05) + 1)|0;
   $exitcond = ($9|0)==(64);
   if ($exitcond) {
    break;
   } else {
    $ci$05 = $9;
   }
  }
 } else {
  $10 = $4&65535;
  HEAP16[$_dct_coeffs>>1] = $10;
  _oc_idct8x8($_state,$_dct_coeffs,$_last_zzi);
 }
 $11 = ((($_state)) + 208|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($12) + ($_fragi<<2)|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($_state)) + 204|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = (($16) + ($_fragi<<2)|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = $18 >>> 8;
 $20 = $19 & 7;
 $21 = (((($_state)) + 496|0) + ($_pli<<2)|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ((($_state)) + 280|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = (((($_state)) + 480|0) + ($24<<2)|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = (($26) + ($14)|0);
 $28 = ($20|0)==(1);
 if ($28) {
  _oc_frag_recon_intra($_state,$27,$22,$_dct_coeffs);
  return;
 }
 $29 = $20 << 2;
 $30 = 268505377 >>> $29;
 $31 = $30 & 15;
 $32 = (((($_state)) + 272|0) + ($31<<2)|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = (((($_state)) + 480|0) + ($33<<2)|0);
 $35 = HEAP32[$34>>2]|0;
 $36 = ((($_state)) + 212|0);
 $37 = HEAP32[$36>>2]|0;
 $38 = (($37) + ($_fragi<<1)|0);
 $39 = HEAP8[$38>>0]|0;
 $40 = $39 << 24 >> 24;
 $41 = (((($37) + ($_fragi<<1)|0)) + 1|0);
 $42 = HEAP8[$41>>0]|0;
 $43 = $42 << 24 >> 24;
 $44 = ($_pli|0)!=(0);
 if ($44) {
  $45 = ((($_state)) + 48|0);
  $46 = HEAP32[$45>>2]|0;
  $47 = $46 & 2;
  $48 = ($47|0)==(0);
  $50 = $48;
 } else {
  $50 = 0;
 }
 $49 = $50&1;
 $51 = (($43) + 31)|0;
 $52 = ((530258 + ($49<<6)|0) + ($51)|0);
 $53 = HEAP8[$52>>0]|0;
 $54 = $53 << 24 >> 24;
 $55 = ((530386 + ($49<<6)|0) + ($51)|0);
 $56 = HEAP8[$55>>0]|0;
 $57 = $56 << 24 >> 24;
 if ($44) {
  $58 = ((($_state)) + 48|0);
  $59 = HEAP32[$58>>2]|0;
  $60 = $59 & 1;
  $61 = ($60|0)==(0);
  $63 = $61;
 } else {
  $63 = 0;
 }
 $62 = $63&1;
 $64 = (($40) + 31)|0;
 $65 = ((530258 + ($62<<6)|0) + ($64)|0);
 $66 = HEAP8[$65>>0]|0;
 $67 = $66 << 24 >> 24;
 $68 = ((530386 + ($62<<6)|0) + ($64)|0);
 $69 = HEAP8[$68>>0]|0;
 $70 = Math_imul($54, $22)|0;
 $71 = (($67) + ($70))|0;
 $72 = $69 | $56;
 $73 = ($72<<24>>24)==(0);
 if ($73) {
  $$sum4 = (($71) + ($14))|0;
  $80 = (($35) + ($$sum4)|0);
  _oc_frag_recon_inter($_state,$27,$80,$22,$_dct_coeffs);
  return;
 } else {
  $74 = $69 << 24 >> 24;
  $75 = Math_imul($57, $22)|0;
  $$sum2 = (($71) + ($14))|0;
  $76 = (($35) + ($$sum2)|0);
  $77 = (($75) + ($14))|0;
  $78 = (($77) + ($71))|0;
  $$sum1 = (($78) + ($74))|0;
  $79 = (($35) + ($$sum1)|0);
  _oc_frag_recon_inter2($_state,$27,$76,$79,$22,$_dct_coeffs);
  return;
 }
}
function _oc_state_frag_copy_list_c($_state,$_fragis,$_nfragis,$_dst_frame,$_src_frame,$_pli) {
 $_state = $_state|0;
 $_fragis = $_fragis|0;
 $_nfragis = $_nfragis|0;
 $_dst_frame = $_dst_frame|0;
 $_src_frame = $_src_frame|0;
 $_pli = $_pli|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $exitcond = 0, $fragii$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (((($_state)) + 272|0) + ($_dst_frame<<2)|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (((($_state)) + 480|0) + ($1<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (((($_state)) + 272|0) + ($_src_frame<<2)|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (((($_state)) + 480|0) + ($5<<2)|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (((($_state)) + 496|0) + ($_pli<<2)|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($_state)) + 208|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($_nfragis|0)>(0);
 if ($12) {
  $fragii$01 = 0;
 } else {
  return;
 }
 while(1) {
  $13 = (($_fragis) + ($fragii$01<<2)|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = (($11) + ($14<<2)|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = (($3) + ($16)|0);
  $18 = (($7) + ($16)|0);
  _oc_frag_copy($_state,$17,$18,$9);
  $19 = (($fragii$01) + 1)|0;
  $exitcond = ($19|0)==($_nfragis|0);
  if ($exitcond) {
   break;
  } else {
   $fragii$01 = $19;
  }
 }
 return;
}
function _oc_state_loop_filter_frag_rows_c($_state,$_bv,$_refi,$_pli,$_fragy0,$_fragy_end) {
 $_state = $_state|0;
 $_bv = $_bv|0;
 $_refi = $_refi|0;
 $_pli = $_pli|0;
 $_fragy0 = $_fragy0|0;
 $_fragy_end = $_fragy_end|0;
 var $$03$i = 0, $$03$i$us = 0, $$03$i4 = 0, $$03$i4$us = 0, $$lobit$i = 0, $$lobit$i$us = 0, $$lobit$i14$us = 0, $$lobit$i2 = 0, $$lobit$i2$us = 0, $$lobit$i6 = 0, $$lobit$i6$us = 0, $$lobit1$i = 0, $$lobit1$i$us = 0, $$lobit1$i7 = 0, $$lobit1$i7$us = 0, $$lobit4$i = 0, $$lobit4$i$us = 0, $$lobit4$i15$us = 0, $$pre = 0, $$pre$phi43Z2D = 0;
 var $$pre$phiZ2D = 0, $$pre42 = 0, $$sum = 0, $$sum$i = 0, $$sum$i$us = 0, $$sum$i11$us = 0, $$sum$us = 0, $$sum1$i = 0, $$sum1$i$us = 0, $$sum1$i12$us = 0, $$sum18 = 0, $$sum18$us = 0, $$sum19 = 0, $$sum19$us = 0, $$sum20 = 0, $$sum20$us = 0, $$sum21 = 0, $$sum21$us = 0, $$sum22 = 0, $$sum22$us = 0;
 var $$sum23 = 0, $$sum23$us = 0, $$sum24 = 0, $$sum24$us = 0, $$sum25$us = 0, $$sum26$us = 0, $$sum27$us = 0, $$sum28$us = 0, $$sum29$us = 0, $$sum3$i = 0, $$sum3$i$us = 0, $$sum3$i13$us = 0, $$sum30 = 0, $$sum30$us = 0, $$sum31 = 0, $$sum31$us = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0;
 var $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0;
 var $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0;
 var $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0;
 var $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i = 0, $exitcond$i$us = 0, $exitcond$i16$us = 0, $exitcond$i3 = 0, $exitcond$i3$us = 0, $exitcond$i8 = 0, $exitcond$i8$us = 0, $fragi$038 = 0, $fragi$038$us = 0, $fragi0$040 = 0, $x$06$i = 0, $x$06$i$us = 0, $x$06$i10$us = 0, $y$02$i = 0, $y$02$i$us = 0, $y$02$i5 = 0;
 var $y$02$i5$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (((($_state)) + 108|0) + ($_pli<<5)|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (((((($_state)) + 108|0) + ($_pli<<5)|0)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (((((($_state)) + 108|0) + ($_pli<<5)|0)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($5) + ($3))|0;
 $7 = Math_imul($1, $_fragy0)|0;
 $8 = (($3) + ($7))|0;
 $9 = (($_fragy_end) - ($_fragy0))|0;
 $10 = Math_imul($1, $9)|0;
 $11 = (($8) + ($10))|0;
 $12 = (((($_state)) + 496|0) + ($_pli<<2)|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($_state)) + 204|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($_state)) + 208|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = (((($_state)) + 480|0) + ($_refi<<2)|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($10|0)>(0);
 if (!($20)) {
  return;
 }
 $21 = $13 << 1;
 $22 = $13 << 3;
 $23 = ($1|0)>(0);
 $fragi0$040 = $8;
 while(1) {
  $25 = (($fragi0$040) + ($1))|0;
  L6: do {
   if ($23) {
    $26 = ($fragi0$040|0)>($3|0);
    if ($26) {
     $fragi$038$us = $fragi0$040;
    } else {
     $fragi$038 = $fragi0$040;
     while(1) {
      $185 = (($15) + ($fragi$038<<2)|0);
      $186 = HEAP32[$185>>2]|0;
      $187 = $186 & 1;
      $188 = ($187|0)==(0);
      if ($188) {
       $$pre42 = (($fragi$038) + 1)|0;
       $$pre$phi43Z2D = $$pre42;
      } else {
       $189 = (($17) + ($fragi$038<<2)|0);
       $190 = HEAP32[$189>>2]|0;
       $191 = ($fragi$038|0)>($fragi0$040|0);
       if ($191) {
        $$sum30 = (($190) + -2)|0;
        $192 = (($19) + ($$sum30)|0);
        $$03$i = $192;$y$02$i = 0;
        while(1) {
         $193 = HEAP8[$$03$i>>0]|0;
         $194 = $193&255;
         $195 = ((($$03$i)) + 3|0);
         $196 = HEAP8[$195>>0]|0;
         $197 = $196&255;
         $198 = ((($$03$i)) + 2|0);
         $199 = HEAP8[$198>>0]|0;
         $200 = $199&255;
         $201 = ((($$03$i)) + 1|0);
         $202 = HEAP8[$201>>0]|0;
         $203 = $202&255;
         $204 = (($200) - ($203))|0;
         $205 = ($204*3)|0;
         $206 = (($194) + 4)|0;
         $207 = (($206) - ($197))|0;
         $208 = (($207) + ($205))|0;
         $209 = $208 >> 3;
         $$sum31 = (($209) + 127)|0;
         $210 = (($_bv) + ($$sum31<<2)|0);
         $211 = HEAP32[$210>>2]|0;
         $212 = (($203) + ($211))|0;
         $$lobit$i = $212 >>> 31;
         $213 = (($$lobit$i) + 255)|0;
         $214 = ($212|0)>(255);
         $215 = $214 ? 255 : $212;
         $216 = $213 & $215;
         $217 = $216&255;
         HEAP8[$201>>0] = $217;
         $218 = (($200) - ($211))|0;
         $$lobit1$i = $218 >>> 31;
         $219 = (($$lobit1$i) + 255)|0;
         $220 = ($218|0)>(255);
         $221 = $220 ? 255 : $218;
         $222 = $219 & $221;
         $223 = $222&255;
         HEAP8[$198>>0] = $223;
         $224 = (($$03$i) + ($13)|0);
         $225 = (($y$02$i) + 1)|0;
         $exitcond$i = ($225|0)==(8);
         if ($exitcond$i) {
          break;
         } else {
          $$03$i = $224;$y$02$i = $225;
         }
        }
       }
       $226 = (($fragi$038) + 1)|0;
       $227 = ($226|0)<($25|0);
       if ($227) {
        $228 = (($15) + ($226<<2)|0);
        $229 = HEAP32[$228>>2]|0;
        $230 = $229 & 1;
        $231 = ($230|0)==(0);
        if ($231) {
         $$sum23 = (($190) + 6)|0;
         $232 = (($19) + ($$sum23)|0);
         $$03$i4 = $232;$y$02$i5 = 0;
         while(1) {
          $233 = HEAP8[$$03$i4>>0]|0;
          $234 = $233&255;
          $235 = ((($$03$i4)) + 3|0);
          $236 = HEAP8[$235>>0]|0;
          $237 = $236&255;
          $238 = ((($$03$i4)) + 2|0);
          $239 = HEAP8[$238>>0]|0;
          $240 = $239&255;
          $241 = ((($$03$i4)) + 1|0);
          $242 = HEAP8[$241>>0]|0;
          $243 = $242&255;
          $244 = (($240) - ($243))|0;
          $245 = ($244*3)|0;
          $246 = (($234) + 4)|0;
          $247 = (($246) - ($237))|0;
          $248 = (($247) + ($245))|0;
          $249 = $248 >> 3;
          $$sum24 = (($249) + 127)|0;
          $250 = (($_bv) + ($$sum24<<2)|0);
          $251 = HEAP32[$250>>2]|0;
          $252 = (($243) + ($251))|0;
          $$lobit$i6 = $252 >>> 31;
          $253 = (($$lobit$i6) + 255)|0;
          $254 = ($252|0)>(255);
          $255 = $254 ? 255 : $252;
          $256 = $253 & $255;
          $257 = $256&255;
          HEAP8[$241>>0] = $257;
          $258 = (($240) - ($251))|0;
          $$lobit1$i7 = $258 >>> 31;
          $259 = (($$lobit1$i7) + 255)|0;
          $260 = ($258|0)>(255);
          $261 = $260 ? 255 : $258;
          $262 = $259 & $261;
          $263 = $262&255;
          HEAP8[$238>>0] = $263;
          $264 = (($$03$i4) + ($13)|0);
          $265 = (($y$02$i5) + 1)|0;
          $exitcond$i8 = ($265|0)==(8);
          if ($exitcond$i8) {
           break;
          } else {
           $$03$i4 = $264;$y$02$i5 = $265;
          }
         }
        }
       }
       $266 = (($fragi$038) + ($1))|0;
       $267 = ($266|0)<($6|0);
       if ($267) {
        $268 = (($15) + ($266<<2)|0);
        $269 = HEAP32[$268>>2]|0;
        $270 = $269 & 1;
        $271 = ($270|0)==(0);
        if ($271) {
         $$sum = (($190) + ($22))|0;
         $$sum$i = (($$sum) - ($21))|0;
         $$sum1$i = (($$sum) + ($13))|0;
         $$sum3$i = (($$sum) - ($13))|0;
         $x$06$i = 0;
         while(1) {
          $$sum18 = (($$sum$i) + ($x$06$i))|0;
          $272 = (($19) + ($$sum18)|0);
          $273 = HEAP8[$272>>0]|0;
          $274 = $273&255;
          $$sum19 = (($$sum1$i) + ($x$06$i))|0;
          $275 = (($19) + ($$sum19)|0);
          $276 = HEAP8[$275>>0]|0;
          $277 = $276&255;
          $$sum20 = (($x$06$i) + ($$sum))|0;
          $278 = (($19) + ($$sum20)|0);
          $279 = HEAP8[$278>>0]|0;
          $280 = $279&255;
          $$sum21 = (($$sum3$i) + ($x$06$i))|0;
          $281 = (($19) + ($$sum21)|0);
          $282 = HEAP8[$281>>0]|0;
          $283 = $282&255;
          $284 = (($280) - ($283))|0;
          $285 = ($284*3)|0;
          $286 = (($274) + 4)|0;
          $287 = (($286) - ($277))|0;
          $288 = (($287) + ($285))|0;
          $289 = $288 >> 3;
          $$sum22 = (($289) + 127)|0;
          $290 = (($_bv) + ($$sum22<<2)|0);
          $291 = HEAP32[$290>>2]|0;
          $292 = (($283) + ($291))|0;
          $$lobit$i2 = $292 >>> 31;
          $293 = (($$lobit$i2) + 255)|0;
          $294 = ($292|0)>(255);
          $295 = $294 ? 255 : $292;
          $296 = $293 & $295;
          $297 = $296&255;
          HEAP8[$281>>0] = $297;
          $298 = HEAP8[$278>>0]|0;
          $299 = $298&255;
          $300 = (($299) - ($291))|0;
          $$lobit4$i = $300 >>> 31;
          $301 = (($$lobit4$i) + 255)|0;
          $302 = ($300|0)>(255);
          $303 = $302 ? 255 : $300;
          $304 = $301 & $303;
          $305 = $304&255;
          HEAP8[$278>>0] = $305;
          $306 = (($x$06$i) + 1)|0;
          $exitcond$i3 = ($306|0)==(8);
          if ($exitcond$i3) {
           $$pre$phi43Z2D = $226;
           break;
          } else {
           $x$06$i = $306;
          }
         }
        } else {
         $$pre$phi43Z2D = $226;
        }
       } else {
        $$pre$phi43Z2D = $226;
       }
      }
      $307 = ($$pre$phi43Z2D|0)<($25|0);
      if ($307) {
       $fragi$038 = $$pre$phi43Z2D;
      } else {
       break L6;
      }
     }
    }
    while(1) {
     $27 = (($15) + ($fragi$038$us<<2)|0);
     $28 = HEAP32[$27>>2]|0;
     $29 = $28 & 1;
     $30 = ($29|0)==(0);
     if ($30) {
      $$pre = (($fragi$038$us) + 1)|0;
      $$pre$phiZ2D = $$pre;
     } else {
      $31 = (($17) + ($fragi$038$us<<2)|0);
      $32 = HEAP32[$31>>2]|0;
      $33 = ($fragi$038$us|0)>($fragi0$040|0);
      if ($33) {
       $$sum30$us = (($32) + -2)|0;
       $34 = (($19) + ($$sum30$us)|0);
       $$03$i$us = $34;$y$02$i$us = 0;
       while(1) {
        $35 = HEAP8[$$03$i$us>>0]|0;
        $36 = $35&255;
        $37 = ((($$03$i$us)) + 3|0);
        $38 = HEAP8[$37>>0]|0;
        $39 = $38&255;
        $40 = ((($$03$i$us)) + 2|0);
        $41 = HEAP8[$40>>0]|0;
        $42 = $41&255;
        $43 = ((($$03$i$us)) + 1|0);
        $44 = HEAP8[$43>>0]|0;
        $45 = $44&255;
        $46 = (($42) - ($45))|0;
        $47 = ($46*3)|0;
        $48 = (($36) + 4)|0;
        $49 = (($48) - ($39))|0;
        $50 = (($49) + ($47))|0;
        $51 = $50 >> 3;
        $$sum31$us = (($51) + 127)|0;
        $52 = (($_bv) + ($$sum31$us<<2)|0);
        $53 = HEAP32[$52>>2]|0;
        $54 = (($45) + ($53))|0;
        $$lobit$i$us = $54 >>> 31;
        $55 = (($$lobit$i$us) + 255)|0;
        $56 = ($54|0)>(255);
        $57 = $56 ? 255 : $54;
        $58 = $55 & $57;
        $59 = $58&255;
        HEAP8[$43>>0] = $59;
        $60 = (($42) - ($53))|0;
        $$lobit1$i$us = $60 >>> 31;
        $61 = (($$lobit1$i$us) + 255)|0;
        $62 = ($60|0)>(255);
        $63 = $62 ? 255 : $60;
        $64 = $61 & $63;
        $65 = $64&255;
        HEAP8[$40>>0] = $65;
        $66 = (($$03$i$us) + ($13)|0);
        $67 = (($y$02$i$us) + 1)|0;
        $exitcond$i$us = ($67|0)==(8);
        if ($exitcond$i$us) {
         break;
        } else {
         $$03$i$us = $66;$y$02$i$us = $67;
        }
       }
      }
      $$sum$i11$us = (($32) - ($21))|0;
      $$sum1$i12$us = (($32) + ($13))|0;
      $$sum3$i13$us = (($32) - ($13))|0;
      $x$06$i10$us = 0;
      while(1) {
       $$sum25$us = (($$sum$i11$us) + ($x$06$i10$us))|0;
       $68 = (($19) + ($$sum25$us)|0);
       $69 = HEAP8[$68>>0]|0;
       $70 = $69&255;
       $$sum26$us = (($$sum1$i12$us) + ($x$06$i10$us))|0;
       $71 = (($19) + ($$sum26$us)|0);
       $72 = HEAP8[$71>>0]|0;
       $73 = $72&255;
       $$sum27$us = (($x$06$i10$us) + ($32))|0;
       $74 = (($19) + ($$sum27$us)|0);
       $75 = HEAP8[$74>>0]|0;
       $76 = $75&255;
       $$sum28$us = (($$sum3$i13$us) + ($x$06$i10$us))|0;
       $77 = (($19) + ($$sum28$us)|0);
       $78 = HEAP8[$77>>0]|0;
       $79 = $78&255;
       $80 = (($76) - ($79))|0;
       $81 = ($80*3)|0;
       $82 = (($70) + 4)|0;
       $83 = (($82) - ($73))|0;
       $84 = (($83) + ($81))|0;
       $85 = $84 >> 3;
       $$sum29$us = (($85) + 127)|0;
       $86 = (($_bv) + ($$sum29$us<<2)|0);
       $87 = HEAP32[$86>>2]|0;
       $88 = (($79) + ($87))|0;
       $$lobit$i14$us = $88 >>> 31;
       $89 = (($$lobit$i14$us) + 255)|0;
       $90 = ($88|0)>(255);
       $91 = $90 ? 255 : $88;
       $92 = $89 & $91;
       $93 = $92&255;
       HEAP8[$77>>0] = $93;
       $94 = HEAP8[$74>>0]|0;
       $95 = $94&255;
       $96 = (($95) - ($87))|0;
       $$lobit4$i15$us = $96 >>> 31;
       $97 = (($$lobit4$i15$us) + 255)|0;
       $98 = ($96|0)>(255);
       $99 = $98 ? 255 : $96;
       $100 = $97 & $99;
       $101 = $100&255;
       HEAP8[$74>>0] = $101;
       $102 = (($x$06$i10$us) + 1)|0;
       $exitcond$i16$us = ($102|0)==(8);
       if ($exitcond$i16$us) {
        break;
       } else {
        $x$06$i10$us = $102;
       }
      }
      $103 = (($fragi$038$us) + 1)|0;
      $104 = ($103|0)<($25|0);
      if ($104) {
       $105 = (($15) + ($103<<2)|0);
       $106 = HEAP32[$105>>2]|0;
       $107 = $106 & 1;
       $108 = ($107|0)==(0);
       if ($108) {
        $$sum23$us = (($32) + 6)|0;
        $109 = (($19) + ($$sum23$us)|0);
        $$03$i4$us = $109;$y$02$i5$us = 0;
        while(1) {
         $110 = HEAP8[$$03$i4$us>>0]|0;
         $111 = $110&255;
         $112 = ((($$03$i4$us)) + 3|0);
         $113 = HEAP8[$112>>0]|0;
         $114 = $113&255;
         $115 = ((($$03$i4$us)) + 2|0);
         $116 = HEAP8[$115>>0]|0;
         $117 = $116&255;
         $118 = ((($$03$i4$us)) + 1|0);
         $119 = HEAP8[$118>>0]|0;
         $120 = $119&255;
         $121 = (($117) - ($120))|0;
         $122 = ($121*3)|0;
         $123 = (($111) + 4)|0;
         $124 = (($123) - ($114))|0;
         $125 = (($124) + ($122))|0;
         $126 = $125 >> 3;
         $$sum24$us = (($126) + 127)|0;
         $127 = (($_bv) + ($$sum24$us<<2)|0);
         $128 = HEAP32[$127>>2]|0;
         $129 = (($120) + ($128))|0;
         $$lobit$i6$us = $129 >>> 31;
         $130 = (($$lobit$i6$us) + 255)|0;
         $131 = ($129|0)>(255);
         $132 = $131 ? 255 : $129;
         $133 = $130 & $132;
         $134 = $133&255;
         HEAP8[$118>>0] = $134;
         $135 = (($117) - ($128))|0;
         $$lobit1$i7$us = $135 >>> 31;
         $136 = (($$lobit1$i7$us) + 255)|0;
         $137 = ($135|0)>(255);
         $138 = $137 ? 255 : $135;
         $139 = $136 & $138;
         $140 = $139&255;
         HEAP8[$115>>0] = $140;
         $141 = (($$03$i4$us) + ($13)|0);
         $142 = (($y$02$i5$us) + 1)|0;
         $exitcond$i8$us = ($142|0)==(8);
         if ($exitcond$i8$us) {
          break;
         } else {
          $$03$i4$us = $141;$y$02$i5$us = $142;
         }
        }
       }
      }
      $143 = (($fragi$038$us) + ($1))|0;
      $144 = ($143|0)<($6|0);
      if ($144) {
       $145 = (($15) + ($143<<2)|0);
       $146 = HEAP32[$145>>2]|0;
       $147 = $146 & 1;
       $148 = ($147|0)==(0);
       if ($148) {
        $$sum$us = (($32) + ($22))|0;
        $$sum$i$us = (($$sum$us) - ($21))|0;
        $$sum1$i$us = (($$sum$us) + ($13))|0;
        $$sum3$i$us = (($$sum$us) - ($13))|0;
        $x$06$i$us = 0;
        while(1) {
         $$sum18$us = (($$sum$i$us) + ($x$06$i$us))|0;
         $149 = (($19) + ($$sum18$us)|0);
         $150 = HEAP8[$149>>0]|0;
         $151 = $150&255;
         $$sum19$us = (($$sum1$i$us) + ($x$06$i$us))|0;
         $152 = (($19) + ($$sum19$us)|0);
         $153 = HEAP8[$152>>0]|0;
         $154 = $153&255;
         $$sum20$us = (($x$06$i$us) + ($$sum$us))|0;
         $155 = (($19) + ($$sum20$us)|0);
         $156 = HEAP8[$155>>0]|0;
         $157 = $156&255;
         $$sum21$us = (($$sum3$i$us) + ($x$06$i$us))|0;
         $158 = (($19) + ($$sum21$us)|0);
         $159 = HEAP8[$158>>0]|0;
         $160 = $159&255;
         $161 = (($157) - ($160))|0;
         $162 = ($161*3)|0;
         $163 = (($151) + 4)|0;
         $164 = (($163) - ($154))|0;
         $165 = (($164) + ($162))|0;
         $166 = $165 >> 3;
         $$sum22$us = (($166) + 127)|0;
         $167 = (($_bv) + ($$sum22$us<<2)|0);
         $168 = HEAP32[$167>>2]|0;
         $169 = (($160) + ($168))|0;
         $$lobit$i2$us = $169 >>> 31;
         $170 = (($$lobit$i2$us) + 255)|0;
         $171 = ($169|0)>(255);
         $172 = $171 ? 255 : $169;
         $173 = $170 & $172;
         $174 = $173&255;
         HEAP8[$158>>0] = $174;
         $175 = HEAP8[$155>>0]|0;
         $176 = $175&255;
         $177 = (($176) - ($168))|0;
         $$lobit4$i$us = $177 >>> 31;
         $178 = (($$lobit4$i$us) + 255)|0;
         $179 = ($177|0)>(255);
         $180 = $179 ? 255 : $177;
         $181 = $178 & $180;
         $182 = $181&255;
         HEAP8[$155>>0] = $182;
         $183 = (($x$06$i$us) + 1)|0;
         $exitcond$i3$us = ($183|0)==(8);
         if ($exitcond$i3$us) {
          $$pre$phiZ2D = $103;
          break;
         } else {
          $x$06$i$us = $183;
         }
        }
       } else {
        $$pre$phiZ2D = $103;
       }
      } else {
       $$pre$phiZ2D = $103;
      }
     }
     $184 = ($$pre$phiZ2D|0)<($25|0);
     if ($184) {
      $fragi$038$us = $$pre$phiZ2D;
     } else {
      break;
     }
    }
   }
  } while(0);
  $24 = ($25|0)<($11|0);
  if ($24) {
   $fragi0$040 = $25;
  } else {
   break;
  }
 }
 return;
}
function _oc_state_init($_state,$_info,$_nrefs) {
 $_state = $_state|0;
 $_info = $_info|0;
 $_nrefs = $_nrefs|0;
 var $$0 = 0, $$lcssa36 = 0, $$lcssa37 = 0, $$lobit$1$i$i = 0, $$lobit$2$i$i = 0, $$lobit$3$i$i = 0, $$lobit$i = 0, $$lobit$i$i = 0, $$lobit$i6 = 0, $$lobit$not$1$i$i = 0, $$lobit$not$2$i$i = 0, $$lobit$not$3$i$i = 0, $$lobit$not$i$i = 0, $$phi$trans$insert = 0, $$phi$trans$insert$i = 0, $$phi$trans$insert24 = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i9$i = 0;
 var $$pre$phi$i$iZ2D = 0, $$pre11$i$i = 0, $$pre25 = 0, $$pre28$i = 0, $$pre29$i = 0, $$sum$i = 0, $$sum$i9 = 0, $$sum4$i = 0, $$sum5$i = 0, $$sum6$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0;
 var $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0;
 var $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0;
 var $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0;
 var $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0;
 var $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0;
 var $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0;
 var $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0;
 var $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0;
 var $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0;
 var $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0;
 var $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0;
 var $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0;
 var $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0;
 var $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0;
 var $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0;
 var $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0;
 var $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0;
 var $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0;
 var $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0;
 var $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0;
 var $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0;
 var $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0;
 var $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0;
 var $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0;
 var $63 = 0, $630 = 0, $631 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $_nrefs$off$i = 0, $brmerge$i$i = 0, $brmerge58$i$i = 0, $brmerge60$i$i = 0, $brmerge62$i$i = 0, $brmerge64$i$i = 0, $brmerge66$i$i = 0, $brmerge68$i$i = 0, $brmerge70$i$i = 0, $brmerge72$i$i = 0, $crop_x0$0$i$i = 0, $crop_x0$1$i$i = 0, $crop_xf$0$i$i = 0, $crop_xf$1$i$i = 0, $crop_y0$0$i$i = 0, $crop_yf$0$i$i = 0, $exitcond$i = 0, $exitcond$i$i = 0;
 var $exitcond$i10 = 0, $exitcond$i13$i = 0, $exitcond$i8$i = 0, $exitcond25$i = 0, $exitcond55$i$i = 0, $exitcond7$i$i = 0, $frag$041$i$i = 0, $frag$1$lcssa$i$i = 0, $frag$133$i$i = 0, $frag$13351$i$i = 0, $frag$2$lcssa$i$i = 0, $frag$226$i$i = 0, $fragi$1$lcssa$1$i = 0, $fragi$1$lcssa$i = 0, $fragi$111$1$i = 0, $fragi$111$2$i = 0, $fragi$111$i = 0, $fragi$2$lcssa$1$i = 0, $fragi$2$lcssa$2$i = 0, $fragi$2$lcssa$i = 0;
 var $fragi$28$1$i = 0, $fragi$28$2$i = 0, $fragi$28$i = 0, $hpix$09$1$i = 0, $hpix$09$2$i = 0, $hpix$09$i = 0, $i$023$i$i = 0, $i$04$i$i = 0, $i$1$lcssa$i$i = 0, $i$116$i$i = 0, $i$124$i$i = 0, $imax$0$i$i = 0, $j$02$i$i = 0, $jmax$0$i$i = 0, $notlhs$i = 0, $notlhs$i$i = 0, $notlhs75$i$i = 0, $notlhs77$i$i = 0, $notlhs79$i$i = 0, $notlhs81$i$i = 0;
 var $notlhs83$i$i = 0, $notlhs85$i$i = 0, $notlhs87$i$i = 0, $notrhs$i = 0, $notrhs$i$i = 0, $notrhs76$i$i = 0, $notrhs78$i$i = 0, $notrhs80$i$i = 0, $notrhs82$i$i = 0, $notrhs84$i$i = 0, $notrhs86$i$i = 0, $notrhs88$i$i = 0, $npixels$022$i$i = 0, $npixels$2$1$i$i = 0, $npixels$2$2$i$i = 0, $npixels$2$3$i$i = 0, $npixels$2$4$i$i = 0, $npixels$2$5$i$i = 0, $npixels$2$6$i$i = 0, $npixels$2$7$i$i = 0;
 var $npixels$2$7$i$i$lcssa = 0, $npixels$2$i$i = 0, $or$cond$i = 0, $or$cond$i7 = 0, $or$cond$not$i$i = 0, $or$cond1$i = 0, $or$cond1$i$i = 0, $or$cond10$i$i = 0, $or$cond11$i$i = 0, $or$cond12$i$i = 0, $or$cond2$i = 0, $or$cond2$i$i = 0, $or$cond2$not$i = 0, $or$cond3$i = 0, $or$cond3$i$i = 0, $or$cond3$not$i$i = 0, $or$cond33$i = 0, $or$cond34$i = 0, $or$cond35$i = 0, $or$cond36$i = 0;
 var $or$cond4 = 0, $or$cond4$i = 0, $or$cond5$i = 0, $or$cond6$i$i = 0, $or$cond7$i = 0, $or$cond7$i$i = 0, $or$cond8$1$not$i$i = 0, $or$cond8$2$not$i$i = 0, $or$cond8$3$not$i$i = 0, $or$cond8$4$not$i$i = 0, $or$cond8$5$not$i$i = 0, $or$cond8$6$not$i$i = 0, $or$cond8$7$not$i$i = 0, $or$cond9$not$i$i = 0, $phitmp = 0, $pli$024$i = 0, $pli$042$i$i = 0, $ref_frame_data$019$i = 0, $ret$0$ph = 0, $rfi$022$i = 0;
 var $rfi$118$i = 0, $sbi$0$i$i = 0, $sbi$09$i$i = 0, $sbi$1$i$i = 0, $sbi$1$i$i$lcssa = 0, $sbi$1$lcssa$i$i = 0, $sbi$16$i$i = 0, $scevgep49$i$i = 0, $scevgep49$sum$i$i = 0, $scevgep52$i$i = 0, $scevgep5253$i$i = 0, $scevgep54$i$i = 0, $smax$1$i = 0, $smax$2$i = 0, $smax$i = 0, $uglygep$sum$i$i = 0, $umax$i$i = 0, $vpix$010$1$i = 0, $vpix$010$2$i = 0, $vpix$010$i = 0;
 var $x$0$i$i = 0, $x$028$i$i = 0, $x$05$i$i = 0, $xfrag$03$i$i = 0, $y$0$i$i = 0, $y$034$i$i = 0, $y$07$i$i = 0, $yfrag$0$i$i = 0, $yfrag_end$040$i$i = 0, $ymb$04$i$i = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ($_info|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ((($_info)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2 & 15;
 $4 = ($3|0)==(0);
 if (!($4)) {
  $$0 = -10;
  return ($$0|0);
 }
 $5 = ((($_info)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $6 & 15;
 $8 = ($7|0)!=(0);
 $9 = (($2) + -1)|0;
 $10 = ($9>>>0)>(1048574);
 $11 = $10 | $8;
 $12 = (($6) + -1)|0;
 $13 = ($12>>>0)>(1048574);
 $14 = $13 | $11;
 if ($14) {
  $$0 = -10;
  return ($$0|0);
 }
 $15 = ((($_info)) + 20|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($_info)) + 12|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = (($18) + ($16))|0;
 $20 = ($19>>>0)>($2>>>0);
 if ($20) {
  $$0 = -10;
  return ($$0|0);
 }
 $21 = ((($_info)) + 24|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ((($_info)) + 16|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = (($24) + ($22))|0;
 $26 = ($25>>>0)>($6>>>0);
 $27 = ($16>>>0)>(255);
 $or$cond4 = $27 | $26;
 if ($or$cond4) {
  $$0 = -10;
  return ($$0|0);
 }
 $28 = (($6) - ($22))|0;
 $29 = (($28) - ($24))|0;
 $30 = ($29>>>0)>(255);
 if ($30) {
  $$0 = -10;
  return ($$0|0);
 }
 $31 = ((($_info)) + 44|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ($32>>>0)>(2);
 if ($33) {
  $$0 = -10;
  return ($$0|0);
 }
 $34 = ((($_info)) + 48|0);
 $35 = HEAP32[$34>>2]|0;
 $36 = ($35>>>0)>(3);
 if ($36) {
  $$0 = -10;
  return ($$0|0);
 }
 _memset(($_state|0),0,51552)|0;
 dest=$_state; src=$_info; stop=dest+64|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 $37 = HEAP32[$5>>2]|0;
 $38 = HEAP32[$23>>2]|0;
 $39 = (($37) - ($38))|0;
 $40 = HEAP32[$21>>2]|0;
 $41 = (($39) - ($40))|0;
 $42 = ((($_state)) + 24|0);
 HEAP32[$42>>2] = $41;
 $43 = ((($_state)) + 792|0);
 HEAP8[$43>>0] = -1;
 $44 = ((($_state)) + 64|0);
 HEAP32[$44>>2] = 4334;
 $45 = ((($_state)) + 68|0);
 HEAP32[$45>>2] = 4335;
 $46 = ((($_state)) + 72|0);
 HEAP32[$46>>2] = 4336;
 $47 = ((($_state)) + 76|0);
 HEAP32[$47>>2] = 4337;
 $48 = ((($_state)) + 80|0);
 HEAP32[$48>>2] = 4338;
 $49 = ((($_state)) + 84|0);
 HEAP32[$49>>2] = 4339;
 $50 = ((($_state)) + 88|0);
 HEAP32[$50>>2] = 4340;
 $51 = ((($_state)) + 92|0);
 HEAP32[$51>>2] = 4341;
 $52 = ((($_state)) + 96|0);
 HEAP32[$52>>2] = 4342;
 $53 = ((($_state)) + 100|0);
 HEAP32[$53>>2] = 530074;
 $54 = ((($_state)) + 4|0);
 $55 = HEAP32[$54>>2]|0;
 $56 = $55 >>> 3;
 $57 = ((($_state)) + 8|0);
 $58 = HEAP32[$57>>2]|0;
 $59 = $58 >>> 3;
 $60 = ((($_state)) + 48|0);
 $61 = HEAP32[$60>>2]|0;
 $62 = $61 & 1;
 $63 = $62 ^ 1;
 $64 = $61 >>> 1;
 $$lobit$i = $64 & 1;
 $65 = $$lobit$i ^ 1;
 $66 = (($63) + ($56))|0;
 $67 = $66 >>> $63;
 $68 = (($65) + ($59))|0;
 $69 = $68 >>> $65;
 $70 = Math_imul($59, $56)|0;
 $71 = Math_imul($69, $67)|0;
 $72 = $71 << 1;
 $73 = (($72) + ($70))|0;
 $74 = (($56) + 3)|0;
 $75 = $74 >>> 2;
 $76 = (($59) + 3)|0;
 $77 = $76 >>> 2;
 $78 = (($67) + 3)|0;
 $79 = $78 >> 2;
 $80 = (($69) + 3)|0;
 $81 = $80 >> 2;
 $82 = Math_imul($77, $75)|0;
 $83 = Math_imul($81, $79)|0;
 $84 = $83 << 1;
 $85 = (($84) + ($82))|0;
 $86 = $82 << 2;
 $87 = ($72|0)<($71|0);
 $88 = ($71|0)<(0);
 $or$cond$i = $88 | $87;
 do {
  if ($or$cond$i) {
   $ret$0$ph = -23;
  } else {
   $89 = (($82>>>0) / ($75>>>0))&-1;
   $90 = ($89|0)!=($77|0);
   $91 = ($84>>>0)<($83>>>0);
   $or$cond1$i = $91 | $90;
   $92 = ($85>>>0)<($82>>>0);
   $or$cond2$i = $92 | $or$cond1$i;
   $or$cond2$not$i = $or$cond2$i ^ 1;
   $93 = $82 & 1073741823;
   $94 = ($93|0)==($82|0);
   $or$cond4$i = $94 & $or$cond2$not$i;
   if ($or$cond4$i) {
    $95 = ((($_state)) + 108|0);
    HEAP32[$95>>2] = $56;
    $96 = ((($_state)) + 112|0);
    HEAP32[$96>>2] = $59;
    $97 = ((($_state)) + 116|0);
    HEAP32[$97>>2] = 0;
    $98 = ((($_state)) + 120|0);
    HEAP32[$98>>2] = $70;
    $99 = ((($_state)) + 124|0);
    HEAP32[$99>>2] = $75;
    $100 = ((($_state)) + 128|0);
    HEAP32[$100>>2] = $77;
    $101 = ((($_state)) + 132|0);
    HEAP32[$101>>2] = 0;
    $102 = ((($_state)) + 136|0);
    HEAP32[$102>>2] = $82;
    $103 = ((($_state)) + 172|0);
    HEAP32[$103>>2] = $67;
    $104 = ((($_state)) + 140|0);
    HEAP32[$104>>2] = $67;
    $105 = ((($_state)) + 176|0);
    HEAP32[$105>>2] = $69;
    $106 = ((($_state)) + 144|0);
    HEAP32[$106>>2] = $69;
    $107 = ((($_state)) + 148|0);
    HEAP32[$107>>2] = $70;
    $108 = (($71) + ($70))|0;
    $109 = ((($_state)) + 180|0);
    HEAP32[$109>>2] = $108;
    $110 = ((($_state)) + 184|0);
    HEAP32[$110>>2] = $71;
    $111 = ((($_state)) + 152|0);
    HEAP32[$111>>2] = $71;
    $112 = ((($_state)) + 188|0);
    HEAP32[$112>>2] = $79;
    $113 = ((($_state)) + 156|0);
    HEAP32[$113>>2] = $79;
    $114 = ((($_state)) + 192|0);
    HEAP32[$114>>2] = $81;
    $115 = ((($_state)) + 160|0);
    HEAP32[$115>>2] = $81;
    $116 = ((($_state)) + 164|0);
    HEAP32[$116>>2] = $82;
    $117 = (($83) + ($82))|0;
    $118 = ((($_state)) + 196|0);
    HEAP32[$118>>2] = $117;
    $119 = ((($_state)) + 200|0);
    HEAP32[$119>>2] = $83;
    $120 = ((($_state)) + 168|0);
    HEAP32[$120>>2] = $83;
    $121 = ((($_state)) + 216|0);
    HEAP32[$121>>2] = $73;
    $122 = (_calloc($73,4)|0);
    $123 = ((($_state)) + 204|0);
    HEAP32[$123>>2] = $122;
    $124 = $73 << 1;
    $125 = (_malloc($124)|0);
    $126 = ((($_state)) + 212|0);
    HEAP32[$126>>2] = $125;
    $127 = ((($_state)) + 228|0);
    HEAP32[$127>>2] = $85;
    $128 = $85 << 6;
    $129 = (_malloc($128)|0);
    $130 = ((($_state)) + 220|0);
    HEAP32[$130>>2] = $129;
    $131 = (_calloc($85,1)|0);
    $132 = ((($_state)) + 224|0);
    HEAP32[$132>>2] = $131;
    $133 = $75 << 1;
    $134 = ((($_state)) + 240|0);
    HEAP32[$134>>2] = $133;
    $135 = $77 << 1;
    $136 = ((($_state)) + 244|0);
    HEAP32[$136>>2] = $135;
    $137 = ((($_state)) + 248|0);
    HEAP32[$137>>2] = $86;
    $138 = (_calloc($86,48)|0);
    $139 = ((($_state)) + 232|0);
    HEAP32[$139>>2] = $138;
    $140 = (_calloc($86,1)|0);
    $141 = ((($_state)) + 236|0);
    HEAP32[$141>>2] = $140;
    $142 = $73 << 2;
    $143 = (_malloc($142)|0);
    $144 = ((($_state)) + 252|0);
    HEAP32[$144>>2] = $143;
    $145 = ($122|0)==(0|0);
    $146 = ($125|0)==(0|0);
    $or$cond33$i = $145 | $146;
    $147 = ($129|0)==(0|0);
    $or$cond34$i = $or$cond33$i | $147;
    $148 = ($131|0)==(0|0);
    $or$cond35$i = $or$cond34$i | $148;
    $149 = ($138|0)==(0|0);
    $or$cond36$i = $or$cond35$i | $149;
    if ($or$cond36$i) {
     $ret$0$ph = -1;
    } else {
     $150 = ($140|0)==(0|0);
     $151 = ($143|0)==(0|0);
     $or$cond5$i = $150 | $151;
     if ($or$cond5$i) {
      $ret$0$ph = -1;
     } else {
      $156 = $59;$163 = 0;$165 = $129;$168 = $131;$627 = 0;$pli$024$i = 0;
      while(1) {
       $152 = (((($_state)) + 108|0) + ($pli$024$i<<5)|0);
       $153 = HEAP32[$152>>2]|0;
       $154 = $153 << 2;
       $sbi$0$i$i = 0;$y$0$i$i = 0;$yfrag$0$i$i = $627;
       while(1) {
        $155 = (($156) - ($y$0$i$i))|0;
        $157 = ($155|0)>(4);
        if ($157) {
         $imax$0$i$i = 4;
        } else {
         $158 = ($155|0)<(1);
         if ($158) {
          break;
         } else {
          $imax$0$i$i = $155;
         }
        }
        $159 = ($imax$0$i$i|0)>(0);
        $sbi$1$i$i = $sbi$0$i$i;$x$0$i$i = 0;
        while(1) {
         $160 = (($153) - ($x$0$i$i))|0;
         $161 = ($160|0)>(4);
         if ($161) {
          $jmax$0$i$i = 4;
         } else {
          $162 = ($160|0)<(1);
          if ($162) {
           $sbi$1$i$i$lcssa = $sbi$1$i$i;
           break;
          } else {
           $jmax$0$i$i = $160;
          }
         }
         $$sum$i = (($sbi$1$i$i) + ($163))|0;
         $164 = (($165) + ($$sum$i<<6)|0);
         dest=$164; stop=dest+64|0; do { HEAP32[dest>>2]=4294967295|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
         if ($159) {
          $166 = (($x$0$i$i) + ($yfrag$0$i$i))|0;
          $i$04$i$i = 0;$xfrag$03$i$i = $166;
          while(1) {
           $j$02$i$i = 0;
           while(1) {
            $212 = (($j$02$i$i) + ($xfrag$03$i$i))|0;
            $213 = ((((47584 + ($i$04$i$i<<5)|0) + ($j$02$i$i<<3)|0)) + 4|0);
            $214 = HEAP32[$213>>2]|0;
            $215 = ((47584 + ($i$04$i$i<<5)|0) + ($j$02$i$i<<3)|0);
            $216 = HEAP32[$215>>2]|0;
            $217 = (((($165) + ($$sum$i<<6)|0) + ($216<<4)|0) + ($214<<2)|0);
            HEAP32[$217>>2] = $212;
            $218 = (($j$02$i$i) + 1)|0;
            $exitcond$i$i = ($218|0)==($jmax$0$i$i|0);
            if ($exitcond$i$i) {
             break;
            } else {
             $j$02$i$i = $218;
            }
           }
           $219 = (($xfrag$03$i$i) + ($153))|0;
           $220 = (($i$04$i$i) + 1)|0;
           $exitcond7$i$i = ($220|0)==($imax$0$i$i|0);
           if ($exitcond7$i$i) {
            break;
           } else {
            $i$04$i$i = $220;$xfrag$03$i$i = $219;
           }
          }
         }
         $167 = (($168) + ($$sum$i)|0);
         $169 = (($165) + ($$sum$i<<6)|0);
         $170 = HEAP32[$169>>2]|0;
         $$lobit$i$i = $170 >>> 31;
         $$lobit$not$i$i = $$lobit$i$i ^ 1;
         $171 = HEAP8[$167>>0]|0;
         $172 = ($171&255) >>> 2;
         $173 = $172&255;
         $174 = $173 | $$lobit$not$i$i;
         $175 = $174&255;
         $176 = ($175 << 2)&255;
         $177 = $176 & 60;
         $178 = $171 & -61;
         $179 = $177 | $178;
         HEAP8[$167>>0] = $179;
         $180 = (((($165) + ($$sum$i<<6)|0)) + 16|0);
         $181 = HEAP32[$180>>2]|0;
         $$lobit$1$i$i = $181 >>> 31;
         $$lobit$not$1$i$i = $$lobit$1$i$i << 1;
         $182 = $$lobit$not$1$i$i ^ 2;
         $183 = ($179&255) >>> 2;
         $184 = $183&255;
         $185 = $184 | $182;
         $186 = $185&255;
         $187 = ($186 << 2)&255;
         $188 = $187 & 60;
         $189 = $188 | $178;
         HEAP8[$167>>0] = $189;
         $190 = (((($165) + ($$sum$i<<6)|0)) + 32|0);
         $191 = HEAP32[$190>>2]|0;
         $$lobit$2$i$i = $191 >>> 31;
         $$lobit$not$2$i$i = $$lobit$2$i$i << 2;
         $192 = $$lobit$not$2$i$i ^ 4;
         $193 = ($189&255) >>> 2;
         $194 = $193&255;
         $195 = $194 | $192;
         $196 = $195&255;
         $197 = ($196 << 2)&255;
         $198 = $197 & 60;
         $199 = $198 | $178;
         HEAP8[$167>>0] = $199;
         $200 = (((($165) + ($$sum$i<<6)|0)) + 56|0);
         $201 = HEAP32[$200>>2]|0;
         $$lobit$3$i$i = $201 >>> 31;
         $$lobit$not$3$i$i = $$lobit$3$i$i << 3;
         $202 = $$lobit$not$3$i$i ^ 8;
         $203 = ($199&255) >>> 2;
         $204 = $203&255;
         $205 = $204 | $202;
         $206 = $205&255;
         $207 = ($206 << 2)&255;
         $208 = $207 & 60;
         $209 = $208 | $178;
         HEAP8[$167>>0] = $209;
         $210 = (($x$0$i$i) + 4)|0;
         $211 = (($sbi$1$i$i) + 1)|0;
         $sbi$1$i$i = $211;$x$0$i$i = $210;
        }
        $221 = (($yfrag$0$i$i) + ($154))|0;
        $222 = (($y$0$i$i) + 4)|0;
        $sbi$0$i$i = $sbi$1$i$i$lcssa;$y$0$i$i = $222;$yfrag$0$i$i = $221;
       }
       $223 = (($pli$024$i) + 1)|0;
       $exitcond$i = ($223|0)==(3);
       if ($exitcond$i) {
        break;
       }
       $$pre$i = HEAP32[$130>>2]|0;
       $$phi$trans$insert$i = (((((($_state)) + 108|0) + ($223<<5)|0)) + 24|0);
       $$pre28$i = HEAP32[$$phi$trans$insert$i>>2]|0;
       $$pre29$i = HEAP32[$132>>2]|0;
       $$phi$trans$insert = (((((($_state)) + 108|0) + ($223<<5)|0)) + 8|0);
       $$pre = HEAP32[$$phi$trans$insert>>2]|0;
       $$phi$trans$insert24 = (((((($_state)) + 108|0) + ($223<<5)|0)) + 4|0);
       $$pre25 = HEAP32[$$phi$trans$insert24>>2]|0;
       $156 = $$pre25;$163 = $$pre28$i;$165 = $$pre$i;$168 = $$pre29$i;$627 = $$pre;$pli$024$i = $223;
      }
      $224 = HEAP32[$139>>2]|0;
      $225 = HEAP32[$141>>2]|0;
      $226 = HEAP32[$60>>2]|0;
      $227 = (47712 + ($226<<2)|0);
      $228 = HEAP32[$227>>2]|0;
      $229 = HEAP32[$96>>2]|0;
      $230 = ($229|0)>(0);
      if ($230) {
       $$pre$i9$i = HEAP32[$95>>2]|0;
       $232 = $$pre$i9$i;$628 = $229;$sbi$09$i$i = 0;$y$07$i$i = 0;
       while(1) {
        $231 = ($232|0)>(0);
        if ($231) {
         $sbi$16$i$i = $sbi$09$i$i;$x$05$i$i = 0;
         while(1) {
          $233 = $sbi$16$i$i << 2;
          $234 = $x$05$i$i | 1;
          $235 = $x$05$i$i | 2;
          $236 = $x$05$i$i | 3;
          $ymb$04$i$i = 0;
          while(1) {
           $237 = $ymb$04$i$i << 1;
           $238 = $237 | $y$07$i$i;
           $239 = $238 | 1;
           $240 = (530202 + ($ymb$04$i$i<<1)|0);
           $241 = HEAP8[$240>>0]|0;
           $242 = $241&255;
           $243 = $242 | $233;
           $244 = (($224) + (($243*48)|0)|0);
           dest=$244; stop=dest+48|0; do { HEAP32[dest>>2]=4294967295|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
           $245 = HEAP32[$95>>2]|0;
           $246 = ($x$05$i$i|0)<($245|0);
           do {
            if ($246) {
             $247 = HEAP32[$96>>2]|0;
             $248 = ($238|0)<($247|0);
             if (!($248)) {
              label = 35;
              break;
             }
             $250 = Math_imul($245, $238)|0;
             $251 = (($250) + ($x$05$i$i))|0;
             HEAP32[$244>>2] = $251;
             $252 = HEAP32[$95>>2]|0;
             $253 = Math_imul($252, $238)|0;
             $254 = (($253) + ($234))|0;
             $255 = (((($224) + (($243*48)|0)|0)) + 4|0);
             HEAP32[$255>>2] = $254;
             $256 = HEAP32[$95>>2]|0;
             $257 = Math_imul($256, $239)|0;
             $258 = (($257) + ($x$05$i$i))|0;
             $259 = (((($224) + (($243*48)|0)|0)) + 8|0);
             HEAP32[$259>>2] = $258;
             $260 = HEAP32[$95>>2]|0;
             $261 = Math_imul($260, $239)|0;
             $262 = (($261) + ($234))|0;
             $263 = (((($224) + (($243*48)|0)|0)) + 12|0);
             HEAP32[$263>>2] = $262;
             FUNCTION_TABLE_viiii[$228 & 8191]($244,$95,$x$05$i$i,$238);
            } else {
             label = 35;
            }
           } while(0);
           if ((label|0) == 35) {
            label = 0;
            $249 = (($225) + ($243)|0);
            HEAP8[$249>>0] = -1;
           }
           $264 = (((530202 + ($ymb$04$i$i<<1)|0)) + 1|0);
           $265 = HEAP8[$264>>0]|0;
           $266 = $265&255;
           $267 = $266 | $233;
           $268 = (($224) + (($267*48)|0)|0);
           dest=$268; stop=dest+48|0; do { HEAP32[dest>>2]=4294967295|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
           $269 = HEAP32[$95>>2]|0;
           $270 = ($235|0)<($269|0);
           do {
            if ($270) {
             $278 = HEAP32[$96>>2]|0;
             $279 = ($238|0)<($278|0);
             if (!($279)) {
              label = 42;
              break;
             }
             $281 = Math_imul($269, $238)|0;
             $282 = (($281) + ($235))|0;
             HEAP32[$268>>2] = $282;
             $283 = HEAP32[$95>>2]|0;
             $284 = Math_imul($283, $238)|0;
             $285 = (($284) + ($236))|0;
             $286 = (((($224) + (($267*48)|0)|0)) + 4|0);
             HEAP32[$286>>2] = $285;
             $287 = HEAP32[$95>>2]|0;
             $288 = Math_imul($287, $239)|0;
             $289 = (($288) + ($235))|0;
             $290 = (((($224) + (($267*48)|0)|0)) + 8|0);
             HEAP32[$290>>2] = $289;
             $291 = HEAP32[$95>>2]|0;
             $292 = Math_imul($291, $239)|0;
             $293 = (($292) + ($236))|0;
             $294 = (((($224) + (($267*48)|0)|0)) + 12|0);
             HEAP32[$294>>2] = $293;
             FUNCTION_TABLE_viiii[$228 & 8191]($268,$95,$235,$238);
            } else {
             label = 42;
            }
           } while(0);
           if ((label|0) == 42) {
            label = 0;
            $280 = (($225) + ($267)|0);
            HEAP8[$280>>0] = -1;
           }
           $295 = (($ymb$04$i$i) + 1)|0;
           $exitcond$i13$i = ($295|0)==(2);
           if ($exitcond$i13$i) {
            break;
           } else {
            $ymb$04$i$i = $295;
           }
          }
          $271 = (($x$05$i$i) + 4)|0;
          $272 = (($sbi$16$i$i) + 1)|0;
          $273 = HEAP32[$95>>2]|0;
          $274 = ($271|0)<($273|0);
          if ($274) {
           $sbi$16$i$i = $272;$x$05$i$i = $271;
          } else {
           $$lcssa36 = $272;$$lcssa37 = $273;
           break;
          }
         }
         $$pre11$i$i = HEAP32[$96>>2]|0;
         $277 = $$pre11$i$i;$629 = $$lcssa37;$sbi$1$lcssa$i$i = $$lcssa36;
        } else {
         $277 = $628;$629 = $232;$sbi$1$lcssa$i$i = $sbi$09$i$i;
        }
        $275 = (($y$07$i$i) + 4)|0;
        $276 = ($275|0)<($277|0);
        if ($276) {
         $232 = $629;$628 = $277;$sbi$09$i$i = $sbi$1$lcssa$i$i;$y$07$i$i = $275;
        } else {
         break;
        }
       }
      }
      $296 = ((($_state)) + 508|0);
      HEAP32[$296>>2] = 0;
      $297 = HEAP32[$123>>2]|0;
      $298 = ((($_state)) + 20|0);
      $299 = ((($_state)) + 12|0);
      $300 = ((($_state)) + 16|0);
      $frag$041$i$i = $297;$pli$042$i$i = 0;$yfrag_end$040$i$i = $297;
      while(1) {
       $301 = HEAP32[$298>>2]|0;
       $302 = HEAP32[$299>>2]|0;
       $303 = (($302) + ($301))|0;
       $304 = HEAP32[$42>>2]|0;
       $305 = HEAP32[$300>>2]|0;
       $306 = (($305) + ($304))|0;
       $307 = ($pli$042$i$i|0)>(0);
       if ($307) {
        $308 = HEAP32[$60>>2]|0;
        $309 = $308 & 1;
        $310 = ($309|0)==(0);
        if ($310) {
         $311 = $301 >> 1;
         $312 = (($303) + 1)|0;
         $313 = $312 >> 1;
         $crop_x0$0$i$i = $311;$crop_xf$0$i$i = $313;
        } else {
         $crop_x0$0$i$i = $301;$crop_xf$0$i$i = $303;
        }
        $314 = $308 & 2;
        $315 = ($314|0)==(0);
        if ($315) {
         $316 = $304 >> 1;
         $317 = (($306) + 1)|0;
         $318 = $317 >> 1;
         $crop_x0$1$i$i = $crop_x0$0$i$i;$crop_xf$1$i$i = $crop_xf$0$i$i;$crop_y0$0$i$i = $316;$crop_yf$0$i$i = $318;
        } else {
         $crop_x0$1$i$i = $crop_x0$0$i$i;$crop_xf$1$i$i = $crop_xf$0$i$i;$crop_y0$0$i$i = $304;$crop_yf$0$i$i = $306;
        }
       } else {
        $crop_x0$1$i$i = $301;$crop_xf$1$i$i = $303;$crop_y0$0$i$i = $304;$crop_yf$0$i$i = $306;
       }
       $319 = (((((($_state)) + 108|0) + ($pli$042$i$i<<5)|0)) + 12|0);
       $320 = HEAP32[$319>>2]|0;
       $321 = (($yfrag_end$040$i$i) + ($320<<2)|0);
       $322 = ($frag$041$i$i>>>0)<($321>>>0);
       if ($322) {
        $323 = (((($_state)) + 108|0) + ($pli$042$i$i<<5)|0);
        $324 = ($crop_x0$1$i$i|0)<($crop_xf$1$i$i|0);
        $325 = ($crop_y0$0$i$i|0)<($crop_yf$0$i$i|0);
        $frag$133$i$i = $frag$041$i$i;$y$034$i$i = 0;
        while(1) {
         $326 = HEAP32[$323>>2]|0;
         $327 = (($frag$133$i$i) + ($326<<2)|0);
         $328 = ($326|0)>(0);
         if ($328) {
          $frag$13351$i$i = $frag$133$i$i;
          $329 = (($y$034$i$i) + 8)|0;
          $330 = ($329|0)>($crop_y0$0$i$i|0);
          $331 = ($crop_yf$0$i$i|0)>($y$034$i$i|0);
          $or$cond1$i$i = $331 & $330;
          $or$cond2$i$i = $324 & $or$cond1$i$i;
          $or$cond3$i$i = $325 & $or$cond2$i$i;
          $332 = ($y$034$i$i|0)<($crop_y0$0$i$i|0);
          $or$cond6$i$i = $332 & $330;
          $333 = ($crop_yf$0$i$i|0)<($329|0);
          $or$cond7$i$i = $331 & $333;
          $scevgep49$i$i = ((($frag$133$i$i)) + 4|0);
          $334 = ($327>>>0)>($scevgep49$i$i>>>0);
          $umax$i$i = $334 ? $327 : $scevgep49$i$i;
          $uglygep$sum$i$i = $frag$13351$i$i ^ -1;
          $scevgep52$i$i = (($umax$i$i) + ($uglygep$sum$i$i)|0);
          $or$cond3$not$i$i = $or$cond3$i$i ^ 1;
          $frag$226$i$i = $frag$133$i$i;$x$028$i$i = 0;
          while(1) {
           $335 = (($x$028$i$i) + 8)|0;
           $notlhs$i$i = ($crop_xf$1$i$i|0)<=($x$028$i$i|0);
           $notrhs$i$i = ($335|0)<=($crop_x0$1$i$i|0);
           $or$cond$not$i$i = $notlhs$i$i | $notrhs$i$i;
           $brmerge$i$i = $or$cond$not$i$i | $or$cond3$not$i$i;
           do {
            if ($brmerge$i$i) {
             $336 = HEAP32[$frag$226$i$i>>2]|0;
             $337 = $336 | 2;
             HEAP32[$frag$226$i$i>>2] = $337;
            } else {
             $338 = ($x$028$i$i|0)<($crop_x0$1$i$i|0);
             $339 = ($crop_xf$1$i$i|0)<($335|0);
             $or$cond10$i$i = $338 | $339;
             $or$cond11$i$i = $or$cond6$i$i | $or$cond10$i$i;
             $or$cond12$i$i = $or$cond7$i$i | $or$cond11$i$i;
             if (!($or$cond12$i$i)) {
              $384 = HEAP32[$frag$226$i$i>>2]|0;
              $385 = $384 | 63488;
              HEAP32[$frag$226$i$i>>2] = $385;
              break;
             }
             $340 = $x$028$i$i | 1;
             $341 = $x$028$i$i | 2;
             $342 = $x$028$i$i | 3;
             $343 = $x$028$i$i | 4;
             $344 = $x$028$i$i | 5;
             $345 = $x$028$i$i | 6;
             $346 = $x$028$i$i | 7;
             $notlhs75$i$i = ($340|0)<($crop_x0$1$i$i|0);
             $notrhs76$i$i = ($340|0)>=($crop_xf$1$i$i|0);
             $or$cond8$1$not$i$i = $notrhs76$i$i | $notlhs75$i$i;
             $notlhs77$i$i = ($341|0)<($crop_x0$1$i$i|0);
             $notrhs78$i$i = ($341|0)>=($crop_xf$1$i$i|0);
             $or$cond8$2$not$i$i = $notrhs78$i$i | $notlhs77$i$i;
             $notlhs79$i$i = ($342|0)<($crop_x0$1$i$i|0);
             $notrhs80$i$i = ($342|0)>=($crop_xf$1$i$i|0);
             $or$cond8$3$not$i$i = $notrhs80$i$i | $notlhs79$i$i;
             $notlhs81$i$i = ($343|0)<($crop_x0$1$i$i|0);
             $notrhs82$i$i = ($343|0)>=($crop_xf$1$i$i|0);
             $or$cond8$4$not$i$i = $notrhs82$i$i | $notlhs81$i$i;
             $notlhs83$i$i = ($344|0)<($crop_x0$1$i$i|0);
             $notrhs84$i$i = ($344|0)>=($crop_xf$1$i$i|0);
             $or$cond8$5$not$i$i = $notrhs84$i$i | $notlhs83$i$i;
             $notlhs85$i$i = ($345|0)<($crop_x0$1$i$i|0);
             $notrhs86$i$i = ($345|0)>=($crop_xf$1$i$i|0);
             $or$cond8$6$not$i$i = $notrhs86$i$i | $notlhs85$i$i;
             $notlhs87$i$i = ($346|0)<($crop_x0$1$i$i|0);
             $notrhs88$i$i = ($346|0)>=($crop_xf$1$i$i|0);
             $or$cond8$7$not$i$i = $notrhs88$i$i | $notlhs87$i$i;
             $354 = 0;$356 = 0;$i$023$i$i = 0;$npixels$022$i$i = 0;
             while(1) {
              $349 = (($i$023$i$i) + ($y$034$i$i))|0;
              $350 = $i$023$i$i << 3;
              $notlhs$i = ($349|0)<($crop_y0$0$i$i|0);
              $notrhs$i = ($349|0)>=($crop_yf$0$i$i|0);
              $or$cond9$not$i$i = $notrhs$i | $notlhs$i;
              $brmerge58$i$i = $338 | $or$cond9$not$i$i;
              if ($brmerge58$i$i) {
               $395 = $354;$397 = $356;$npixels$2$i$i = $npixels$022$i$i;
              } else {
               $351 = (_bitshift64Shl(1,0,($350|0))|0);
               $352 = tempRet0;
               $353 = $351 | $354;
               $355 = $352 | $356;
               $357 = (($npixels$022$i$i) + 1)|0;
               $395 = $353;$397 = $355;$npixels$2$i$i = $357;
              }
              $brmerge60$i$i = $or$cond8$1$not$i$i | $or$cond9$not$i$i;
              if ($brmerge60$i$i) {
               $403 = $395;$405 = $397;$npixels$2$1$i$i = $npixels$2$i$i;
              } else {
               $391 = $350 | 1;
               $392 = (_bitshift64Shl(1,0,($391|0))|0);
               $393 = tempRet0;
               $394 = $395 | $392;
               $396 = $397 | $393;
               $398 = (($npixels$2$i$i) + 1)|0;
               $403 = $394;$405 = $396;$npixels$2$1$i$i = $398;
              }
              $brmerge62$i$i = $or$cond8$2$not$i$i | $or$cond9$not$i$i;
              if ($brmerge62$i$i) {
               $411 = $403;$413 = $405;$npixels$2$2$i$i = $npixels$2$1$i$i;
              } else {
               $399 = $350 | 2;
               $400 = (_bitshift64Shl(1,0,($399|0))|0);
               $401 = tempRet0;
               $402 = $403 | $400;
               $404 = $405 | $401;
               $406 = (($npixels$2$1$i$i) + 1)|0;
               $411 = $402;$413 = $404;$npixels$2$2$i$i = $406;
              }
              $brmerge64$i$i = $or$cond8$3$not$i$i | $or$cond9$not$i$i;
              if ($brmerge64$i$i) {
               $419 = $411;$421 = $413;$npixels$2$3$i$i = $npixels$2$2$i$i;
              } else {
               $407 = $350 | 3;
               $408 = (_bitshift64Shl(1,0,($407|0))|0);
               $409 = tempRet0;
               $410 = $411 | $408;
               $412 = $413 | $409;
               $414 = (($npixels$2$2$i$i) + 1)|0;
               $419 = $410;$421 = $412;$npixels$2$3$i$i = $414;
              }
              $brmerge66$i$i = $or$cond8$4$not$i$i | $or$cond9$not$i$i;
              if ($brmerge66$i$i) {
               $427 = $419;$429 = $421;$npixels$2$4$i$i = $npixels$2$3$i$i;
              } else {
               $415 = $350 | 4;
               $416 = (_bitshift64Shl(1,0,($415|0))|0);
               $417 = tempRet0;
               $418 = $419 | $416;
               $420 = $421 | $417;
               $422 = (($npixels$2$3$i$i) + 1)|0;
               $427 = $418;$429 = $420;$npixels$2$4$i$i = $422;
              }
              $brmerge68$i$i = $or$cond8$5$not$i$i | $or$cond9$not$i$i;
              if ($brmerge68$i$i) {
               $435 = $427;$437 = $429;$npixels$2$5$i$i = $npixels$2$4$i$i;
              } else {
               $423 = $350 | 5;
               $424 = (_bitshift64Shl(1,0,($423|0))|0);
               $425 = tempRet0;
               $426 = $427 | $424;
               $428 = $429 | $425;
               $430 = (($npixels$2$4$i$i) + 1)|0;
               $435 = $426;$437 = $428;$npixels$2$5$i$i = $430;
              }
              $brmerge70$i$i = $or$cond8$6$not$i$i | $or$cond9$not$i$i;
              if ($brmerge70$i$i) {
               $443 = $435;$445 = $437;$npixels$2$6$i$i = $npixels$2$5$i$i;
              } else {
               $431 = $350 | 6;
               $432 = (_bitshift64Shl(1,0,($431|0))|0);
               $433 = tempRet0;
               $434 = $435 | $432;
               $436 = $437 | $433;
               $438 = (($npixels$2$5$i$i) + 1)|0;
               $443 = $434;$445 = $436;$npixels$2$6$i$i = $438;
              }
              $brmerge72$i$i = $or$cond8$7$not$i$i | $or$cond9$not$i$i;
              if ($brmerge72$i$i) {
               $630 = $443;$631 = $445;$npixels$2$7$i$i = $npixels$2$6$i$i;
              } else {
               $439 = $350 | 7;
               $440 = (_bitshift64Shl(1,0,($439|0))|0);
               $441 = tempRet0;
               $442 = $443 | $440;
               $444 = $445 | $441;
               $446 = (($npixels$2$6$i$i) + 1)|0;
               $630 = $442;$631 = $444;$npixels$2$7$i$i = $446;
              }
              $447 = (($i$023$i$i) + 1)|0;
              $exitcond$i8$i = ($447|0)==(8);
              if ($exitcond$i8$i) {
               $362 = $630;$365 = $631;$npixels$2$7$i$i$lcssa = $npixels$2$7$i$i;
               break;
              } else {
               $354 = $630;$356 = $631;$i$023$i$i = $447;$npixels$022$i$i = $npixels$2$7$i$i;
              }
             }
             $347 = HEAP32[$296>>2]|0;
             $348 = ($347|0)>(0);
             L129: do {
              if ($348) {
               $i$124$i$i = 0;
               while(1) {
                $367 = (((($_state)) + 512|0) + ($i$124$i$i<<4)|0);
                $368 = $367;
                $369 = $368;
                $370 = HEAP32[$369>>2]|0;
                $371 = (($368) + 4)|0;
                $372 = $371;
                $373 = HEAP32[$372>>2]|0;
                $374 = ($370|0)==($362|0);
                $375 = ($373|0)==($365|0);
                $376 = $374 & $375;
                if ($376) {
                 $i$116$i$i = $i$124$i$i;
                 break L129;
                }
                $382 = (($i$124$i$i) + 1)|0;
                $383 = ($382|0)<($347|0);
                if ($383) {
                 $i$124$i$i = $382;
                } else {
                 $i$1$lcssa$i$i = $382;
                 label = 64;
                 break;
                }
               }
              } else {
               $i$1$lcssa$i$i = 0;
               label = 64;
              }
             } while(0);
             if ((label|0) == 64) {
              label = 0;
              $358 = (($347) + 1)|0;
              HEAP32[$296>>2] = $358;
              $359 = (((($_state)) + 512|0) + ($i$1$lcssa$i$i<<4)|0);
              $360 = $359;
              $361 = $360;
              HEAP32[$361>>2] = $362;
              $363 = (($360) + 4)|0;
              $364 = $363;
              HEAP32[$364>>2] = $365;
              $366 = (((((($_state)) + 512|0) + ($i$1$lcssa$i$i<<4)|0)) + 8|0);
              HEAP32[$366>>2] = $npixels$2$7$i$i$lcssa;
              $i$116$i$i = $i$1$lcssa$i$i;
             }
             $377 = HEAP32[$frag$226$i$i>>2]|0;
             $378 = $i$116$i$i << 11;
             $379 = $378 & 63488;
             $380 = $377 & -63489;
             $381 = $380 | $379;
             HEAP32[$frag$226$i$i>>2] = $381;
            }
           } while(0);
           $386 = ((($frag$226$i$i)) + 4|0);
           $387 = ($386>>>0)<($327>>>0);
           if ($387) {
            $frag$226$i$i = $386;$x$028$i$i = $335;
           } else {
            break;
           }
          }
          $scevgep5253$i$i = $scevgep52$i$i;
          $388 = $scevgep5253$i$i >>> 2;
          $scevgep49$sum$i$i = (($388) + 1)|0;
          $scevgep54$i$i = (($frag$133$i$i) + ($scevgep49$sum$i$i<<2)|0);
          $$pre$phi$i$iZ2D = $329;$frag$2$lcssa$i$i = $scevgep54$i$i;
         } else {
          $$pre$i$i = (($y$034$i$i) + 8)|0;
          $$pre$phi$i$iZ2D = $$pre$i$i;$frag$2$lcssa$i$i = $frag$133$i$i;
         }
         $389 = ($frag$2$lcssa$i$i>>>0)<($321>>>0);
         if ($389) {
          $frag$133$i$i = $frag$2$lcssa$i$i;$y$034$i$i = $$pre$phi$i$iZ2D;
         } else {
          $frag$1$lcssa$i$i = $frag$2$lcssa$i$i;
          break;
         }
        }
       } else {
        $frag$1$lcssa$i$i = $frag$041$i$i;
       }
       $390 = (($pli$042$i$i) + 1)|0;
       $exitcond55$i$i = ($390|0)==(3);
       if ($exitcond55$i$i) {
        break;
       } else {
        $frag$041$i$i = $frag$1$lcssa$i$i;$pli$042$i$i = $390;$yfrag_end$040$i$i = $321;
       }
      }
      $_nrefs$off$i = (($_nrefs) + -3)|0;
      $448 = ($_nrefs$off$i>>>0)>(1);
      if ($448) {
       $ret$0$ph = -10;
      } else {
       $449 = HEAP32[$60>>2]|0;
       $450 = $449 & 1;
       $451 = $450 ^ 1;
       $452 = $449 >>> 1;
       $$lobit$i6 = $452 & 1;
       $453 = $$lobit$i6 ^ 1;
       $454 = HEAP32[$54>>2]|0;
       $455 = (($454) + 32)|0;
       $456 = HEAP32[$57>>2]|0;
       $457 = (($456) + 32)|0;
       $458 = $455 >> $451;
       $459 = $457 >> $453;
       $460 = Math_imul($457, $455)|0;
       $461 = Math_imul($459, $458)|0;
       $462 = $455 << 4;
       $463 = (($462) + 16)|0;
       $464 = 16 >>> $451;
       $465 = 16 >>> $453;
       $466 = Math_imul($465, $458)|0;
       $467 = (($466) + ($464))|0;
       $468 = $461 << 1;
       $469 = (($468) + ($460))|0;
       $470 = Math_imul($469, $_nrefs)|0;
       $471 = (($460>>>0) / ($455>>>0))&-1;
       $472 = ($471|0)!=($457|0);
       $473 = ($468>>>0)<($461>>>0);
       $or$cond$i7 = $472 | $473;
       $474 = ($469>>>0)<($460>>>0);
       $or$cond7$i = $or$cond$i7 | $474;
       if ($or$cond7$i) {
        $ret$0$ph = -23;
       } else {
        $475 = (($470>>>0) / ($_nrefs>>>0))&-1;
        $476 = ($475|0)==($469|0);
        if ($476) {
         $477 = (_malloc($470)|0);
         $478 = HEAP32[$121>>2]|0;
         $479 = $478 << 2;
         $480 = (_malloc($479)|0);
         $481 = ((($_state)) + 208|0);
         HEAP32[$481>>2] = $480;
         $482 = ($477|0)==(0|0);
         $483 = ($480|0)==(0|0);
         $or$cond3$i = $482 | $483;
         if ($or$cond3$i) {
          _free($480);
          _free($477);
          $ret$0$ph = -1;
          break;
         }
         $484 = ((($_state)) + 288|0);
         HEAP32[$484>>2] = $454;
         $485 = ((($_state)) + 292|0);
         HEAP32[$485>>2] = $456;
         $486 = ((($_state)) + 296|0);
         HEAP32[$486>>2] = $455;
         $487 = $454 >>> $451;
         $488 = ((($_state)) + 320|0);
         HEAP32[$488>>2] = $487;
         $489 = ((($_state)) + 304|0);
         HEAP32[$489>>2] = $487;
         $490 = $456 >>> $453;
         $491 = ((($_state)) + 324|0);
         HEAP32[$491>>2] = $490;
         $492 = ((($_state)) + 308|0);
         HEAP32[$492>>2] = $490;
         $493 = ((($_state)) + 328|0);
         HEAP32[$493>>2] = $458;
         $494 = ((($_state)) + 312|0);
         HEAP32[$494>>2] = $458;
         $495 = ($_nrefs|0)>(1);
         if ($495) {
          $rfi$022$i = 1;
          while(1) {
           $497 = (((($_state)) + 288|0) + (($rfi$022$i*48)|0)|0);
           dest=$497; src=$484; stop=dest+48|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
           $498 = (($rfi$022$i) + 1)|0;
           $exitcond25$i = ($498|0)==($_nrefs|0);
           if ($exitcond25$i) {
            break;
           } else {
            $rfi$022$i = $498;
           }
          }
         }
         $496 = ($_nrefs|0)>(0);
         if ($496) {
          $$sum$i9 = (($467) + ($460))|0;
          $$sum4$i = (($461) + ($460))|0;
          $$sum5$i = (($$sum4$i) + ($467))|0;
          $$sum6$i = (($$sum4$i) + ($461))|0;
          $ref_frame_data$019$i = $477;$rfi$118$i = 0;
          while(1) {
           $499 = (((($_state)) + 480|0) + ($rfi$118$i<<2)|0);
           HEAP32[$499>>2] = $ref_frame_data$019$i;
           $500 = (($ref_frame_data$019$i) + ($463)|0);
           $501 = (((($_state)) + 288|0) + (($rfi$118$i*48)|0)|0);
           $502 = (((((($_state)) + 288|0) + (($rfi$118$i*48)|0)|0)) + 12|0);
           HEAP32[$502>>2] = $500;
           $503 = (($ref_frame_data$019$i) + ($$sum$i9)|0);
           $504 = (((((($_state)) + 288|0) + (($rfi$118$i*48)|0)|0)) + 28|0);
           HEAP32[$504>>2] = $503;
           $505 = (($ref_frame_data$019$i) + ($$sum5$i)|0);
           $506 = (((((($_state)) + 288|0) + (($rfi$118$i*48)|0)|0)) + 44|0);
           HEAP32[$506>>2] = $505;
           $507 = (($ref_frame_data$019$i) + ($$sum6$i)|0);
           _oc_ycbcr_buffer_flip($501,$501);
           $508 = (($rfi$118$i) + 1)|0;
           $exitcond$i10 = ($508|0)==($_nrefs|0);
           if ($exitcond$i10) {
            break;
           } else {
            $ref_frame_data$019$i = $507;$rfi$118$i = $508;
           }
          }
         }
         $509 = (-32 - ($454))|0;
         $510 = ((($_state)) + 496|0);
         HEAP32[$510>>2] = $509;
         $511 = (0 - ($458))|0;
         $512 = ((($_state)) + 504|0);
         HEAP32[$512>>2] = $511;
         $513 = ((($_state)) + 500|0);
         HEAP32[$513>>2] = $511;
         $514 = ((($_state)) + 480|0);
         $515 = HEAP32[$514>>2]|0;
         $516 = $515;
         $517 = HEAP32[$97>>2]|0;
         $518 = HEAP32[$98>>2]|0;
         $519 = (($518) + ($517))|0;
         $520 = ((($_state)) + 108|0);
         $521 = HEAP32[$520>>2]|0;
         $522 = ($519|0)>(0);
         if ($522) {
          $523 = HEAP32[$486>>2]|0;
          $524 = ((($_state)) + 300|0);
          $525 = HEAP32[$524>>2]|0;
          $526 = $523 << 3;
          $527 = ($521|0)>(0);
          $fragi$111$i = 0;$vpix$010$i = $525;
          while(1) {
           $528 = (($fragi$111$i) + ($521))|0;
           if ($527) {
            $529 = (($fragi$111$i) + 1)|0;
            $fragi$28$i = $fragi$111$i;$hpix$09$i = $vpix$010$i;
            while(1) {
             $530 = $hpix$09$i;
             $531 = (($530) - ($516))|0;
             $532 = (($480) + ($fragi$28$i<<2)|0);
             HEAP32[$532>>2] = $531;
             $533 = ((($hpix$09$i)) + 8|0);
             $534 = (($fragi$28$i) + 1)|0;
             $535 = ($534|0)<($528|0);
             if ($535) {
              $fragi$28$i = $534;$hpix$09$i = $533;
             } else {
              break;
             }
            }
            $536 = ($528|0)>($529|0);
            $smax$i = $536 ? $528 : $529;
            $fragi$2$lcssa$i = $smax$i;
           } else {
            $fragi$2$lcssa$i = $fragi$111$i;
           }
           $537 = (($vpix$010$i) + ($526)|0);
           $538 = ($fragi$2$lcssa$i|0)<($519|0);
           if ($538) {
            $fragi$111$i = $fragi$2$lcssa$i;$vpix$010$i = $537;
           } else {
            $fragi$1$lcssa$i = $fragi$2$lcssa$i;
            break;
           }
          }
         } else {
          $fragi$1$lcssa$i = 0;
         }
         $539 = HEAP32[$107>>2]|0;
         $540 = HEAP32[$111>>2]|0;
         $541 = (($540) + ($539))|0;
         $542 = HEAP32[$104>>2]|0;
         $543 = ($fragi$1$lcssa$i|0)<($541|0);
         if ($543) {
          $544 = HEAP32[$494>>2]|0;
          $545 = ((($_state)) + 316|0);
          $546 = HEAP32[$545>>2]|0;
          $547 = $544 << 3;
          $548 = ($542|0)>(0);
          $fragi$111$1$i = $fragi$1$lcssa$i;$vpix$010$1$i = $546;
          while(1) {
           $549 = (($fragi$111$1$i) + ($542))|0;
           if ($548) {
            $550 = (($fragi$111$1$i) + 1)|0;
            $fragi$28$1$i = $fragi$111$1$i;$hpix$09$1$i = $vpix$010$1$i;
            while(1) {
             $551 = $hpix$09$1$i;
             $552 = (($551) - ($516))|0;
             $553 = (($480) + ($fragi$28$1$i<<2)|0);
             HEAP32[$553>>2] = $552;
             $554 = ((($hpix$09$1$i)) + 8|0);
             $555 = (($fragi$28$1$i) + 1)|0;
             $556 = ($555|0)<($549|0);
             if ($556) {
              $fragi$28$1$i = $555;$hpix$09$1$i = $554;
             } else {
              break;
             }
            }
            $557 = ($549|0)>($550|0);
            $smax$1$i = $557 ? $549 : $550;
            $fragi$2$lcssa$1$i = $smax$1$i;
           } else {
            $fragi$2$lcssa$1$i = $fragi$111$1$i;
           }
           $558 = (($vpix$010$1$i) + ($547)|0);
           $559 = ($fragi$2$lcssa$1$i|0)<($541|0);
           if ($559) {
            $fragi$111$1$i = $fragi$2$lcssa$1$i;$vpix$010$1$i = $558;
           } else {
            $fragi$1$lcssa$1$i = $fragi$2$lcssa$1$i;
            break;
           }
          }
         } else {
          $fragi$1$lcssa$1$i = $fragi$1$lcssa$i;
         }
         $560 = HEAP32[$109>>2]|0;
         $561 = HEAP32[$110>>2]|0;
         $562 = (($561) + ($560))|0;
         $563 = HEAP32[$103>>2]|0;
         $564 = ($fragi$1$lcssa$1$i|0)<($562|0);
         if ($564) {
          $565 = HEAP32[$493>>2]|0;
          $566 = ((($_state)) + 332|0);
          $567 = HEAP32[$566>>2]|0;
          $568 = $565 << 3;
          $569 = ($563|0)>(0);
          $fragi$111$2$i = $fragi$1$lcssa$1$i;$vpix$010$2$i = $567;
          while(1) {
           $570 = (($fragi$111$2$i) + ($563))|0;
           if ($569) {
            $571 = (($fragi$111$2$i) + 1)|0;
            $fragi$28$2$i = $fragi$111$2$i;$hpix$09$2$i = $vpix$010$2$i;
            while(1) {
             $572 = $hpix$09$2$i;
             $573 = (($572) - ($516))|0;
             $574 = (($480) + ($fragi$28$2$i<<2)|0);
             HEAP32[$574>>2] = $573;
             $575 = ((($hpix$09$2$i)) + 8|0);
             $576 = (($fragi$28$2$i) + 1)|0;
             $577 = ($576|0)<($570|0);
             if ($577) {
              $fragi$28$2$i = $576;$hpix$09$2$i = $575;
             } else {
              break;
             }
            }
            $578 = ($570|0)>($571|0);
            $smax$2$i = $578 ? $570 : $571;
            $fragi$2$lcssa$2$i = $smax$2$i;
           } else {
            $fragi$2$lcssa$2$i = $fragi$111$2$i;
           }
           $579 = (($vpix$010$2$i) + ($568)|0);
           $580 = ($fragi$2$lcssa$2$i|0)<($562|0);
           if ($580) {
            $fragi$111$2$i = $fragi$2$lcssa$2$i;$vpix$010$2$i = $579;
           } else {
            break;
           }
          }
         }
         $595 = ((($_state)) + 280|0);
         HEAP32[$595>>2] = -1;
         $596 = ((($_state)) + 276|0);
         HEAP32[$596>>2] = -1;
         $597 = ((($_state)) + 272|0);
         HEAP32[$597>>2] = -1;
         $598 = ($_nrefs|0)>(3);
         $599 = $598 ? 3 : -1;
         $600 = ((($_state)) + 284|0);
         HEAP32[$600>>2] = $599;
         $601 = ((($_info)) + 60|0);
         $602 = HEAP32[$601>>2]|0;
         $603 = ($602>>>0)>(31);
         if ($603) {
          $604 = ((($_state)) + 60|0);
          HEAP32[$604>>2] = 31;
         }
         $605 = ((($_state)) + 768|0);
         $606 = $605;
         $607 = $606;
         HEAP32[$607>>2] = 0;
         $608 = (($606) + 4)|0;
         $609 = $608;
         HEAP32[$609>>2] = 0;
         $610 = ((($_state)) + 776|0);
         $611 = $610;
         $612 = $611;
         HEAP32[$612>>2] = -1;
         $613 = (($611) + 4)|0;
         $614 = $613;
         HEAP32[$614>>2] = -1;
         $615 = HEAP8[$_info>>0]|0;
         $616 = ($615&255)>(3);
         do {
          if ($616) {
           $626 = 1;
          } else {
           $617 = ($615<<24>>24)==(3);
           if (!($617)) {
            $626 = 0;
            break;
           }
           $618 = ((($_info)) + 1|0);
           $619 = HEAP8[$618>>0]|0;
           $620 = ($619&255)>(2);
           if ($620) {
            $626 = 1;
            break;
           }
           $621 = ($619<<24>>24)==(2);
           if (!($621)) {
            $626 = 0;
            break;
           }
           $622 = ((($_info)) + 2|0);
           $623 = HEAP8[$622>>0]|0;
           $624 = ($623<<24>>24)!=(0);
           $phitmp = $624&1;
           $626 = $phitmp;
          }
         } while(0);
         $625 = ((($_state)) + 793|0);
         HEAP8[$625>>0] = $626;
         $$0 = 0;
         return ($$0|0);
        } else {
         $ret$0$ph = -23;
        }
       }
      }
     }
    }
   } else {
    $ret$0$ph = -23;
   }
  }
 } while(0);
 $581 = ((($_state)) + 252|0);
 $582 = HEAP32[$581>>2]|0;
 _free($582);
 $583 = ((($_state)) + 236|0);
 $584 = HEAP32[$583>>2]|0;
 _free($584);
 $585 = ((($_state)) + 232|0);
 $586 = HEAP32[$585>>2]|0;
 _free($586);
 $587 = ((($_state)) + 224|0);
 $588 = HEAP32[$587>>2]|0;
 _free($588);
 $589 = ((($_state)) + 220|0);
 $590 = HEAP32[$589>>2]|0;
 _free($590);
 $591 = ((($_state)) + 212|0);
 $592 = HEAP32[$591>>2]|0;
 _free($592);
 $593 = ((($_state)) + 204|0);
 $594 = HEAP32[$593>>2]|0;
 _free($594);
 $$0 = $ret$0$ph;
 return ($$0|0);
}
function _oc_mb_fill_cmapping00($_mb_map,$_fplanes,$_xfrag0,$_yfrag0) {
 $_mb_map = $_mb_map|0;
 $_fplanes = $_fplanes|0;
 $_xfrag0 = $_xfrag0|0;
 $_yfrag0 = $_yfrag0|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $_xfrag0 >> 1;
 $1 = $_yfrag0 >> 1;
 $2 = ((($_fplanes)) + 32|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = Math_imul($3, $1)|0;
 $5 = (($4) + ($0))|0;
 $6 = ((($_fplanes)) + 40|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (($5) + ($7))|0;
 $9 = ((($_mb_map)) + 16|0);
 HEAP32[$9>>2] = $8;
 $10 = ((($_fplanes)) + 72|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($5) + ($11))|0;
 $13 = ((($_mb_map)) + 32|0);
 HEAP32[$13>>2] = $12;
 return;
}
function _oc_mb_fill_cmapping01($_mb_map,$_fplanes,$_xfrag0,$_yfrag0) {
 $_mb_map = $_mb_map|0;
 $_fplanes = $_fplanes|0;
 $_xfrag0 = $_xfrag0|0;
 $_yfrag0 = $_yfrag0|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $_yfrag0 >> 1;
 $1 = ((($_fplanes)) + 32|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = Math_imul($2, $0)|0;
 $4 = (($3) + ($_xfrag0))|0;
 $5 = ((($_fplanes)) + 40|0);
 $6 = ((($_fplanes)) + 72|0);
 $7 = HEAP32[$5>>2]|0;
 $8 = (($7) + ($4))|0;
 $9 = ((($_mb_map)) + 16|0);
 HEAP32[$9>>2] = $8;
 $10 = HEAP32[$6>>2]|0;
 $11 = (($10) + ($4))|0;
 $12 = ((($_mb_map)) + 32|0);
 HEAP32[$12>>2] = $11;
 $13 = (($4) + 1)|0;
 $14 = HEAP32[$5>>2]|0;
 $15 = (($14) + ($13))|0;
 $16 = ((($_mb_map)) + 20|0);
 HEAP32[$16>>2] = $15;
 $17 = HEAP32[$6>>2]|0;
 $18 = (($17) + ($13))|0;
 $19 = ((($_mb_map)) + 36|0);
 HEAP32[$19>>2] = $18;
 return;
}
function _oc_mb_fill_cmapping10($_mb_map,$_fplanes,$_xfrag0,$_yfrag0) {
 $_mb_map = $_mb_map|0;
 $_fplanes = $_fplanes|0;
 $_xfrag0 = $_xfrag0|0;
 $_yfrag0 = $_yfrag0|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $_xfrag0 >> 1;
 $1 = ((($_fplanes)) + 32|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = Math_imul($2, $_yfrag0)|0;
 $4 = (($3) + ($0))|0;
 $5 = ((($_fplanes)) + 40|0);
 $6 = ((($_fplanes)) + 72|0);
 $7 = HEAP32[$5>>2]|0;
 $8 = (($7) + ($4))|0;
 $9 = ((($_mb_map)) + 16|0);
 HEAP32[$9>>2] = $8;
 $10 = HEAP32[$6>>2]|0;
 $11 = (($10) + ($4))|0;
 $12 = ((($_mb_map)) + 32|0);
 HEAP32[$12>>2] = $11;
 $13 = HEAP32[$1>>2]|0;
 $14 = (($13) + ($4))|0;
 $15 = HEAP32[$5>>2]|0;
 $16 = (($15) + ($14))|0;
 $17 = ((($_mb_map)) + 24|0);
 HEAP32[$17>>2] = $16;
 $18 = HEAP32[$6>>2]|0;
 $19 = (($18) + ($14))|0;
 $20 = ((($_mb_map)) + 40|0);
 HEAP32[$20>>2] = $19;
 return;
}
function _oc_mb_fill_cmapping11($_mb_map,$_fplanes) {
 $_mb_map = $_mb_map|0;
 $_fplanes = $_fplanes|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_fplanes)) + 40|0);
 $1 = ((($_fplanes)) + 72|0);
 $2 = HEAP32[$_mb_map>>2]|0;
 $3 = HEAP32[$0>>2]|0;
 $4 = (($3) + ($2))|0;
 $5 = ((($_mb_map)) + 16|0);
 HEAP32[$5>>2] = $4;
 $6 = HEAP32[$1>>2]|0;
 $7 = (($6) + ($2))|0;
 $8 = ((($_mb_map)) + 32|0);
 HEAP32[$8>>2] = $7;
 $9 = ((($_mb_map)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = HEAP32[$0>>2]|0;
 $12 = (($11) + ($10))|0;
 $13 = ((($_mb_map)) + 20|0);
 HEAP32[$13>>2] = $12;
 $14 = HEAP32[$1>>2]|0;
 $15 = (($14) + ($10))|0;
 $16 = ((($_mb_map)) + 36|0);
 HEAP32[$16>>2] = $15;
 $17 = ((($_mb_map)) + 8|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = HEAP32[$0>>2]|0;
 $20 = (($19) + ($18))|0;
 $21 = ((($_mb_map)) + 24|0);
 HEAP32[$21>>2] = $20;
 $22 = HEAP32[$1>>2]|0;
 $23 = (($22) + ($18))|0;
 $24 = ((($_mb_map)) + 40|0);
 HEAP32[$24>>2] = $23;
 $25 = ((($_mb_map)) + 12|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = HEAP32[$0>>2]|0;
 $28 = (($27) + ($26))|0;
 $29 = ((($_mb_map)) + 28|0);
 HEAP32[$29>>2] = $28;
 $30 = HEAP32[$1>>2]|0;
 $31 = (($30) + ($26))|0;
 $32 = ((($_mb_map)) + 44|0);
 HEAP32[$32>>2] = $31;
 return;
}
function _oc_state_clear($_state) {
 $_state = $_state|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_state)) + 208|0);
 $1 = HEAP32[$0>>2]|0;
 _free($1);
 $2 = ((($_state)) + 480|0);
 $3 = HEAP32[$2>>2]|0;
 _free($3);
 $4 = ((($_state)) + 252|0);
 $5 = HEAP32[$4>>2]|0;
 _free($5);
 $6 = ((($_state)) + 236|0);
 $7 = HEAP32[$6>>2]|0;
 _free($7);
 $8 = ((($_state)) + 232|0);
 $9 = HEAP32[$8>>2]|0;
 _free($9);
 $10 = ((($_state)) + 224|0);
 $11 = HEAP32[$10>>2]|0;
 _free($11);
 $12 = ((($_state)) + 220|0);
 $13 = HEAP32[$12>>2]|0;
 _free($13);
 $14 = ((($_state)) + 212|0);
 $15 = HEAP32[$14>>2]|0;
 _free($15);
 $16 = ((($_state)) + 204|0);
 $17 = HEAP32[$16>>2]|0;
 _free($17);
 return;
}
function _oc_state_borders_fill_rows($_state,$_refi,$_pli,$_y0,$_yend) {
 $_state = $_state|0;
 $_refi = $_refi|0;
 $_pli = $_pli|0;
 $_y0 = $_y0|0;
 $_yend = $_yend|0;
 var $$sum = 0, $$sum1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $apix$03 = 0, $bpix$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($_pli|0)==(0);
 if ($0) {
  $6 = 0;
 } else {
  $1 = ((($_state)) + 48|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = $2 & 1;
  $4 = ($3|0)==(0);
  $6 = $4;
 }
 $5 = $6&1;
 $7 = 16 >>> $5;
 $8 = ((((((($_state)) + 288|0) + (($_refi*48)|0)|0) + ($_pli<<4)|0)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((((((($_state)) + 288|0) + (($_refi*48)|0)|0) + ($_pli<<4)|0)) + 12|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = Math_imul($9, $_y0)|0;
 $13 = Math_imul($9, $_yend)|0;
 $14 = (($11) + ($13)|0);
 $15 = ($12|0)==($13|0);
 if ($15) {
  return;
 }
 $16 = (($11) + ($12)|0);
 $$sum = (($12) + -1)|0;
 $17 = ((((($_state)) + 288|0) + (($_refi*48)|0)|0) + ($_pli<<4)|0);
 $18 = HEAP32[$17>>2]|0;
 $$sum1 = (($$sum) + ($18))|0;
 $19 = (($11) + ($$sum1)|0);
 $20 = (0 - ($7))|0;
 $apix$03 = $16;$bpix$02 = $19;
 while(1) {
  $21 = (($apix$03) + ($20)|0);
  $22 = HEAP8[$apix$03>>0]|0;
  _memset(($21|0),($22|0),($7|0))|0;
  $23 = ((($bpix$02)) + 1|0);
  $24 = HEAP8[$bpix$02>>0]|0;
  _memset(($23|0),($24|0),($7|0))|0;
  $25 = (($apix$03) + ($9)|0);
  $26 = (($bpix$02) + ($9)|0);
  $27 = ($25|0)==($14|0);
  if ($27) {
   break;
  } else {
   $apix$03 = $25;$bpix$02 = $26;
  }
 }
 return;
}
function _oc_state_borders_fill_caps($_state,$_refi,$_pli) {
 $_state = $_state|0;
 $_refi = $_refi|0;
 $_pli = $_pli|0;
 var $$sum = 0, $$sum1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $apix$03 = 0, $bpix$02 = 0, $sum = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ($_pli|0)==(0);
 if ($0) {
  $16 = 16;$9 = 0;
 } else {
  $1 = ((($_state)) + 48|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = $2 & 1;
  $4 = $3 ^ 1;
  $5 = 16 >>> $4;
  $6 = $2 & 2;
  $7 = ($6|0)==(0);
  $16 = $5;$9 = $7;
 }
 $8 = $9&1;
 $10 = 16 >>> $8;
 $11 = ((((((($_state)) + 288|0) + (($_refi*48)|0)|0) + ($_pli<<4)|0)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((((($_state)) + 288|0) + (($_refi*48)|0)|0) + ($_pli<<4)|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = $16 << 1;
 $17 = (($14) + ($15))|0;
 $18 = ((((((($_state)) + 288|0) + (($_refi*48)|0)|0) + ($_pli<<4)|0)) + 12|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (0 - ($16))|0;
 $21 = Math_imul($10, $12)|0;
 $sum = (($16) + ($21))|0;
 $$sum1 = (0 - ($sum))|0;
 $22 = (($19) + ($$sum1)|0);
 $23 = ($20|0)==($$sum1|0);
 if ($23) {
  return;
 }
 $24 = (($19) + ($20)|0);
 $25 = ((((((($_state)) + 288|0) + (($_refi*48)|0)|0) + ($_pli<<4)|0)) + 4|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = (($26) + -1)|0;
 $28 = Math_imul($27, $12)|0;
 $$sum = (($28) - ($16))|0;
 $29 = (($19) + ($$sum)|0);
 $30 = (0 - ($12))|0;
 $apix$03 = $24;$bpix$02 = $29;
 while(1) {
  $31 = (($apix$03) + ($30)|0);
  _memcpy(($31|0),($apix$03|0),($17|0))|0;
  $32 = (($bpix$02) + ($12)|0);
  _memcpy(($32|0),($bpix$02|0),($17|0))|0;
  $33 = ($31|0)==($22|0);
  if ($33) {
   break;
  } else {
   $apix$03 = $31;$bpix$02 = $32;
  }
 }
 return;
}
function _oc_state_frag_recon($_state,$_fragi,$_pli,$_dct_coeffs,$_last_zzi,$_dc_quant) {
 $_state = $_state|0;
 $_fragi = $_fragi|0;
 $_pli = $_pli|0;
 $_dct_coeffs = $_dct_coeffs|0;
 $_last_zzi = $_last_zzi|0;
 $_dc_quant = $_dc_quant|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_state)) + 84|0);
 $1 = HEAP32[$0>>2]|0;
 FUNCTION_TABLE_viiiiii[$1 & 8191]($_state,$_fragi,$_pli,$_dct_coeffs,$_last_zzi,$_dc_quant);
 return;
}
function _oc_state_frag_copy_list($_state,$_fragis,$_nfragis,$_dst_frame,$_src_frame,$_pli) {
 $_state = $_state|0;
 $_fragis = $_fragis|0;
 $_nfragis = $_nfragis|0;
 $_dst_frame = $_dst_frame|0;
 $_src_frame = $_src_frame|0;
 $_pli = $_pli|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_state)) + 88|0);
 $1 = HEAP32[$0>>2]|0;
 FUNCTION_TABLE_viiiiii[$1 & 8191]($_state,$_fragis,$_nfragis,$_dst_frame,$_src_frame,$_pli);
 return;
}
function _oc_state_loop_filter_init($_state,$_bv) {
 $_state = $_state|0;
 $_bv = $_bv|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_state)) + 795|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = $1&255;
 $3 = (((($_state)) + 51488|0) + ($2)|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = $4&255;
 $6 = ($4<<24>>24)==(0);
 if ($6) {
  $$0 = 1;
  return ($$0|0);
 }
 _memset(($_bv|0),0,1024)|0;
 $i$01 = 0;
 while(1) {
  $7 = (127 - ($i$01))|0;
  $8 = (($7) - ($5))|0;
  $9 = ($8|0)>(-1);
  if ($9) {
   $10 = (($i$01) - ($5))|0;
   $11 = (($_bv) + ($8<<2)|0);
   HEAP32[$11>>2] = $10;
  }
  $12 = (0 - ($i$01))|0;
  $13 = (($_bv) + ($7<<2)|0);
  HEAP32[$13>>2] = $12;
  $14 = (($i$01) + 127)|0;
  $15 = (($_bv) + ($14<<2)|0);
  HEAP32[$15>>2] = $i$01;
  $16 = (($14) + ($5))|0;
  $17 = ($16|0)<(256);
  if ($17) {
   $18 = (($5) - ($i$01))|0;
   $19 = (($_bv) + ($16<<2)|0);
   HEAP32[$19>>2] = $18;
  }
  $20 = (($i$01) + 1)|0;
  $exitcond = ($20|0)==($5|0);
  if ($exitcond) {
   $$0 = 0;
   break;
  } else {
   $i$01 = $20;
  }
 }
 return ($$0|0);
}
function _oc_state_loop_filter_frag_rows($_state,$_bv,$_refi,$_pli,$_fragy0,$_fragy_end) {
 $_state = $_state|0;
 $_bv = $_bv|0;
 $_refi = $_refi|0;
 $_pli = $_pli|0;
 $_fragy0 = $_fragy0|0;
 $_fragy_end = $_fragy_end|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_state)) + 92|0);
 $1 = HEAP32[$0>>2]|0;
 FUNCTION_TABLE_viiiiii[$1 & 8191]($_state,$_bv,$_refi,$_pli,$_fragy0,$_fragy_end);
 return;
}
function _th_granule_time($_encdec,$0,$1) {
 $_encdec = $_encdec|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0.0, $35 = 0, $36 = 0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($1|0)>(-1);
 $3 = ($0>>>0)>(4294967295);
 $4 = ($1|0)==(-1);
 $5 = $4 & $3;
 $6 = $2 | $5;
 if (!($6)) {
  $$0 = -1.0;
  return (+$$0);
 }
 $7 = ((($_encdec)) + 60|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (_bitshift64Ashr(($0|0),($1|0),($8|0))|0);
 $10 = tempRet0;
 $11 = (_bitshift64Shl(($9|0),($10|0),($8|0))|0);
 $12 = tempRet0;
 $13 = HEAP8[$_encdec>>0]|0;
 $14 = ($13&255)>(3);
 if ($14) {
  $30 = 1;$31 = 0;
 } else {
  $15 = ($13<<24>>24)==(3);
  if ($15) {
   $16 = ((($_encdec)) + 1|0);
   $17 = HEAP8[$16>>0]|0;
   $18 = ($17&255)>(2);
   if ($18) {
    $30 = 1;$31 = 0;
   } else {
    $19 = ($17<<24>>24)==(2);
    if ($19) {
     $20 = ((($_encdec)) + 2|0);
     $21 = HEAP8[$20>>0]|0;
     $22 = ($21<<24>>24)!=(0);
     $23 = $22&1;
     $30 = $23;$31 = 0;
    } else {
     $30 = 0;$31 = 0;
    }
   }
  } else {
   $30 = 0;$31 = 0;
  }
 }
 $24 = (_i64Add(($0|0),($1|0),1,0)|0);
 $25 = tempRet0;
 $26 = (_i64Add(($24|0),($25|0),($9|0),($10|0))|0);
 $27 = tempRet0;
 $28 = (_i64Subtract(($26|0),($27|0),($11|0),($12|0))|0);
 $29 = tempRet0;
 $32 = (_i64Subtract(($28|0),($29|0),($30|0),($31|0))|0);
 $33 = tempRet0;
 $34 = (+($32>>>0)) + (4294967296.0*(+($33|0)));
 $35 = ((($_encdec)) + 32|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = (+($36>>>0));
 $38 = ((($_encdec)) + 28|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = (+($39>>>0));
 $41 = $37 / $40;
 $42 = $34 * $41;
 $$0 = $42;
 return (+$$0);
}
function _oc_pack_readinit($_b,$_buf,$_bytes) {
 $_b = $_b|0;
 $_buf = $_buf|0;
 $_bytes = $_bytes|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$_b>>2]=0|0;HEAP32[$_b+4>>2]=0|0;HEAP32[$_b+8>>2]=0|0;HEAP32[$_b+12>>2]=0|0;HEAP32[$_b+16>>2]=0|0;
 $0 = ((($_b)) + 4|0);
 HEAP32[$0>>2] = $_buf;
 $1 = (($_buf) + ($_bytes)|0);
 $2 = ((($_b)) + 8|0);
 HEAP32[$2>>2] = $1;
 return;
}
function _oc_pack_read($_b,$_bits) {
 $_b = $_b|0;
 $_bits = $_bits|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $available$0 = 0, $available$0$lcssa$i = 0, $available$04$i = 0, $available$1$i = 0, $exitcond$i = 0, $lftr$limit$i = 0, $or$cond1$i = 0, $ptr$0$lcssa$i = 0, $ptr$02$i = 0, $scevgep = 0, $umax = 0, $umax$i = 0;
 var $window$0 = 0, $window$0$lcssa$i = 0, $window$03$i = 0, $window$1$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_b)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($_bits|0)==(0);
 if ($2) {
  $$0 = 0;
  return ($$0|0);
 }
 $3 = HEAP32[$_b>>2]|0;
 $4 = ($1|0)<($_bits|0);
 if ($4) {
  $5 = ((($_b)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ((($_b)) + 8|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($1|0)<(25);
  $10 = ($6>>>0)<($8>>>0);
  $or$cond1$i = $9 & $10;
  if ($or$cond1$i) {
   $11 = $8;
   $12 = $6;
   $13 = (($12) - ($11))|0;
   $14 = (24 - ($1))|0;
   $15 = $14 >>> 3;
   $16 = $15 ^ -1;
   $17 = ($13>>>0)>($16>>>0);
   $umax$i = $17 ? $13 : $16;
   $18 = (0 - ($umax$i))|0;
   $lftr$limit$i = (($6) + ($18)|0);
   $19 = $15 ^ -1;
   $20 = ($13>>>0)>($19>>>0);
   $umax = $20 ? $13 : $19;
   $21 = (0 - ($umax))|0;
   $available$04$i = $1;$ptr$02$i = $6;$window$03$i = $3;
   while(1) {
    $22 = (($available$04$i) + 8)|0;
    $23 = ((($ptr$02$i)) + 1|0);
    $24 = HEAP8[$ptr$02$i>>0]|0;
    $25 = $24&255;
    $26 = (24 - ($available$04$i))|0;
    $27 = $25 << $26;
    $28 = $27 | $window$03$i;
    $exitcond$i = ($23|0)==($lftr$limit$i|0);
    if ($exitcond$i) {
     $$lcssa = $28;
     break;
    } else {
     $available$04$i = $22;$ptr$02$i = $23;$window$03$i = $28;
    }
   }
   $29 = $umax$i << 3;
   $scevgep = (($6) + ($21)|0);
   $30 = ($scevgep>>>0)<($8>>>0);
   $31 = (($1) - ($29))|0;
   $$lcssa$i = $30;$available$0$lcssa$i = $31;$ptr$0$lcssa$i = $lftr$limit$i;$window$0$lcssa$i = $$lcssa;
  } else {
   $$lcssa$i = $10;$available$0$lcssa$i = $1;$ptr$0$lcssa$i = $6;$window$0$lcssa$i = $3;
  }
  HEAP32[$5>>2] = $ptr$0$lcssa$i;
  $32 = ($available$0$lcssa$i|0)<($_bits|0);
  do {
   if ($32) {
    if ($$lcssa$i) {
     $34 = HEAP8[$ptr$0$lcssa$i>>0]|0;
     $35 = $34&255;
     $36 = $available$0$lcssa$i & 7;
     $37 = $35 >>> $36;
     $38 = $37 | $window$0$lcssa$i;
     $available$1$i = $available$0$lcssa$i;$window$1$i = $38;
     break;
    } else {
     $33 = ((($_b)) + 16|0);
     HEAP32[$33>>2] = 1;
     $available$1$i = 1073741824;$window$1$i = $window$0$lcssa$i;
     break;
    }
   } else {
    $available$1$i = $available$0$lcssa$i;$window$1$i = $window$0$lcssa$i;
   }
  } while(0);
  HEAP32[$0>>2] = $available$1$i;
  $available$0 = $available$1$i;$window$0 = $window$1$i;
 } else {
  $available$0 = $1;$window$0 = $3;
 }
 $39 = (32 - ($_bits))|0;
 $40 = $window$0 >>> $39;
 $41 = (($available$0) - ($_bits))|0;
 $42 = $window$0 << 1;
 $43 = (($_bits) + -1)|0;
 $44 = $42 << $43;
 HEAP32[$0>>2] = $41;
 HEAP32[$_b>>2] = $44;
 $$0 = $40;
 return ($$0|0);
}
function _oc_pack_read1($_b) {
 $_b = $_b|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $available$0 = 0, $available$04$i = 0, $available$1$i = 0, $exitcond$i = 0, $lftr$limit$i = 0, $ptr$02$i = 0, $scevgep = 0, $umax = 0, $umax$i = 0, $window$0 = 0, $window$0$lcssa$i37 = 0, $window$03$i = 0, $window$1$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$_b>>2]|0;
 $1 = ((($_b)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)<(1);
 if (!($3)) {
  $available$0 = $2;$window$0 = $0;
  $37 = $window$0 >>> 31;
  $38 = (($available$0) + -1)|0;
  $39 = $window$0 << 1;
  HEAP32[$1>>2] = $38;
  HEAP32[$_b>>2] = $39;
  return ($37|0);
 }
 $4 = ((($_b)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($_b)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($5>>>0)<($7>>>0);
 if ($8) {
  $9 = $7;
  $10 = $5;
  $11 = (($10) - ($9))|0;
  $12 = (24 - ($2))|0;
  $13 = $12 >>> 3;
  $14 = $13 ^ -1;
  $15 = ($11>>>0)>($14>>>0);
  $umax$i = $15 ? $11 : $14;
  $16 = (0 - ($umax$i))|0;
  $lftr$limit$i = (($5) + ($16)|0);
  $17 = $13 ^ -1;
  $18 = ($11>>>0)>($17>>>0);
  $umax = $18 ? $11 : $17;
  $19 = (0 - ($umax))|0;
  $available$04$i = $2;$ptr$02$i = $5;$window$03$i = $0;
  while(1) {
   $20 = (($available$04$i) + 8)|0;
   $21 = ((($ptr$02$i)) + 1|0);
   $22 = HEAP8[$ptr$02$i>>0]|0;
   $23 = $22&255;
   $24 = (24 - ($available$04$i))|0;
   $25 = $23 << $24;
   $26 = $25 | $window$03$i;
   $exitcond$i = ($21|0)==($lftr$limit$i|0);
   if ($exitcond$i) {
    $$lcssa = $26;
    break;
   } else {
    $available$04$i = $20;$ptr$02$i = $21;$window$03$i = $26;
   }
  }
  $27 = $umax$i << 3;
  $28 = (($2) - ($27))|0;
  HEAP32[$4>>2] = $lftr$limit$i;
  $29 = ($28|0)<(1);
  if ($29) {
   $scevgep = (($5) + ($19)|0);
   $30 = ($scevgep>>>0)<($7>>>0);
   if ($30) {
    $32 = HEAP8[$lftr$limit$i>>0]|0;
    $33 = $32&255;
    $34 = $28 & 7;
    $35 = $33 >>> $34;
    $36 = $35 | $$lcssa;
    $available$1$i = $28;$window$1$i = $36;
   } else {
    $window$0$lcssa$i37 = $$lcssa;
    label = 8;
   }
  } else {
   $available$1$i = $28;$window$1$i = $$lcssa;
  }
 } else {
  HEAP32[$4>>2] = $5;
  $window$0$lcssa$i37 = $0;
  label = 8;
 }
 if ((label|0) == 8) {
  $31 = ((($_b)) + 16|0);
  HEAP32[$31>>2] = 1;
  $available$1$i = 1073741824;$window$1$i = $window$0$lcssa$i37;
 }
 HEAP32[$1>>2] = $available$1$i;
 $available$0 = $available$1$i;$window$0 = $window$1$i;
 $37 = $window$0 >>> 31;
 $38 = (($available$0) + -1)|0;
 $39 = $window$0 << 1;
 HEAP32[$1>>2] = $38;
 HEAP32[$_b>>2] = $39;
 return ($37|0);
}
function _oc_pack_bytes_left($_b) {
 $_b = $_b|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_b)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  $$0 = -1;
  return ($$0|0);
 }
 $3 = ((($_b)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($_b)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $4;
 $8 = $6;
 $9 = (($7) - ($8))|0;
 $10 = ((($_b)) + 12|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 >> 3;
 $13 = (($9) + ($12))|0;
 $$0 = $13;
 return ($$0|0);
}
function _oc_idct8x8($_state,$_y,$_last_zzi) {
 $_state = $_state|0;
 $_y = $_y|0;
 $_last_zzi = $_last_zzi|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_state)) + 80|0);
 $1 = HEAP32[$0>>2]|0;
 FUNCTION_TABLE_vii[$1 & 8191]($_y,$_last_zzi);
 return;
}
function _oc_idct8x8_c($_y,$_last_zzi) {
 $_y = $_y|0;
 $_last_zzi = $_last_zzi|0;
 var $$idx$i = 0, $$idx$val$i = 0, $$idx3$i = 0, $$idx3$val$i = 0, $$idx4$i = 0, $$idx4$val$i = 0, $$val$i = 0, $$val1$i = 0, $$val2$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0;
 var $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0;
 var $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0;
 var $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0;
 var $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0;
 var $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0;
 var $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0;
 var $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0;
 var $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0;
 var $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0;
 var $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $_y$idx = 0, $_y$idx$val = 0, $_y$idx1 = 0, $_y$idx1$val = 0, $_y$val = 0, $in$0$idx$i = 0, $in$0$idx$val$i = 0, $in$0$val$i = 0, $in$0$val$pre$i = 0, $in$03$i = 0, $in$08$i = 0, $out$04$i = 0, $out$09$i = 0, $out$12$i = 0, $out$17$i = 0, $out$21$i = 0, $sext$i$i = 0;
 var $sext$i5$i = 0, $sext1$i$i = 0, $sext1$i6$i = 0, $w$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $w$i = sp;
 $0 = ($_last_zzi|0)<(3);
 if ($0) {
  $_y$val = HEAP16[$_y>>1]|0;
  $_y$idx = ((($_y)) + 2|0);
  $_y$idx$val = HEAP16[$_y$idx>>1]|0;
  $_y$idx1 = ((($_y)) + 16|0);
  $_y$idx1$val = HEAP16[$_y$idx1>>1]|0;
  $1 = $_y$val << 16 >> 16;
  $2 = ($1*46341)|0;
  $3 = $2 >> 16;
  $4 = $_y$idx$val << 16 >> 16;
  $5 = ($4*12785)|0;
  $6 = $5 >> 16;
  $7 = ($4*64277)|0;
  $8 = $7 >> 16;
  $9 = ($6*46341)|0;
  $10 = $9 >> 16;
  $11 = ($8*46341)|0;
  $12 = $11 >> 16;
  $13 = (($12) + ($10))|0;
  $14 = (($12) - ($10))|0;
  $15 = (($8) + ($3))|0;
  $16 = $15&65535;
  HEAP16[$w$i>>1] = $16;
  $17 = (($13) + ($3))|0;
  $18 = $17&65535;
  $19 = ((($w$i)) + 16|0);
  HEAP16[$19>>1] = $18;
  $20 = (($14) + ($3))|0;
  $21 = $20&65535;
  $22 = ((($w$i)) + 32|0);
  HEAP16[$22>>1] = $21;
  $23 = (($6) + ($3))|0;
  $24 = $23&65535;
  $25 = ((($w$i)) + 48|0);
  HEAP16[$25>>1] = $24;
  $26 = (($3) - ($6))|0;
  $27 = $26&65535;
  $28 = ((($w$i)) + 64|0);
  HEAP16[$28>>1] = $27;
  $29 = (($3) - ($14))|0;
  $30 = $29&65535;
  $31 = ((($w$i)) + 80|0);
  HEAP16[$31>>1] = $30;
  $32 = (($3) - ($13))|0;
  $33 = $32&65535;
  $34 = ((($w$i)) + 96|0);
  HEAP16[$34>>1] = $33;
  $35 = (($3) - ($8))|0;
  $36 = $35&65535;
  $37 = ((($w$i)) + 112|0);
  HEAP16[$37>>1] = $36;
  $38 = ((($w$i)) + 2|0);
  $39 = $_y$idx1$val << 16 >> 16;
  $40 = ($39*46341)|0;
  $41 = $40 >>> 16;
  $42 = $41&65535;
  $43 = ((($w$i)) + 114|0);
  HEAP16[$43>>1] = $42;
  $44 = ((($w$i)) + 98|0);
  HEAP16[$44>>1] = $42;
  $45 = ((($w$i)) + 82|0);
  HEAP16[$45>>1] = $42;
  $46 = ((($w$i)) + 66|0);
  HEAP16[$46>>1] = $42;
  $47 = ((($w$i)) + 50|0);
  HEAP16[$47>>1] = $42;
  $48 = ((($w$i)) + 34|0);
  HEAP16[$48>>1] = $42;
  $49 = ((($w$i)) + 18|0);
  HEAP16[$49>>1] = $42;
  HEAP16[$38>>1] = $42;
  $in$0$val$i = $16;$in$03$i = $w$i;$out$04$i = $_y;
  while(1) {
   $in$0$idx$i = ((($in$03$i)) + 2|0);
   $in$0$idx$val$i = HEAP16[$in$0$idx$i>>1]|0;
   $50 = $in$0$val$i << 16 >> 16;
   $51 = ($50*46341)|0;
   $52 = $51 >> 16;
   $53 = $in$0$idx$val$i << 16 >> 16;
   $54 = ($53*12785)|0;
   $55 = $54 >> 16;
   $56 = ($53*64277)|0;
   $57 = $56 >> 16;
   $58 = ($55*46341)|0;
   $59 = $58 >> 16;
   $60 = ($57*46341)|0;
   $61 = $60 >> 16;
   $62 = (($61) + ($59))|0;
   $63 = (($61) - ($59))|0;
   $64 = (($57) + ($52))|0;
   $65 = $64&65535;
   HEAP16[$out$04$i>>1] = $65;
   $66 = (($62) + ($52))|0;
   $67 = $66&65535;
   $68 = ((($out$04$i)) + 16|0);
   HEAP16[$68>>1] = $67;
   $69 = (($63) + ($52))|0;
   $70 = $69&65535;
   $71 = ((($out$04$i)) + 32|0);
   HEAP16[$71>>1] = $70;
   $72 = (($55) + ($52))|0;
   $73 = $72&65535;
   $74 = ((($out$04$i)) + 48|0);
   HEAP16[$74>>1] = $73;
   $75 = (($52) - ($55))|0;
   $76 = $75&65535;
   $77 = ((($out$04$i)) + 64|0);
   HEAP16[$77>>1] = $76;
   $78 = (($52) - ($63))|0;
   $79 = $78&65535;
   $80 = ((($out$04$i)) + 80|0);
   HEAP16[$80>>1] = $79;
   $81 = (($52) - ($62))|0;
   $82 = $81&65535;
   $83 = ((($out$04$i)) + 96|0);
   HEAP16[$83>>1] = $82;
   $84 = (($52) - ($57))|0;
   $85 = $84&65535;
   $86 = ((($out$04$i)) + 112|0);
   HEAP16[$86>>1] = $85;
   $87 = ((($in$03$i)) + 16|0);
   $88 = ((($out$04$i)) + 2|0);
   $89 = ($88>>>0)<($_y$idx1>>>0);
   if (!($89)) {
    break;
   }
   $in$0$val$pre$i = HEAP16[$87>>1]|0;
   $in$0$val$i = $in$0$val$pre$i;$in$03$i = $87;$out$04$i = $88;
  }
  $90 = ((($_y)) + 128|0);
  $out$12$i = $_y;
  while(1) {
   $91 = HEAP16[$out$12$i>>1]|0;
   $92 = $91 << 16 >> 16;
   $93 = (($92) + 8)|0;
   $94 = $93 >>> 4;
   $95 = $94&65535;
   HEAP16[$out$12$i>>1] = $95;
   $96 = ((($out$12$i)) + 2|0);
   $97 = ($96>>>0)<($90>>>0);
   if ($97) {
    $out$12$i = $96;
   } else {
    break;
   }
  }
  STACKTOP = sp;return;
 }
 $98 = ($_last_zzi|0)<(10);
 if (!($98)) {
  _idct8($w$i,$_y);
  $338 = ((($_y)) + 16|0);
  $339 = ((($w$i)) + 2|0);
  _idct8($339,$338);
  $340 = ((($_y)) + 32|0);
  $341 = ((($w$i)) + 4|0);
  _idct8($341,$340);
  $342 = ((($_y)) + 48|0);
  $343 = ((($w$i)) + 6|0);
  _idct8($343,$342);
  $344 = ((($_y)) + 64|0);
  $345 = ((($w$i)) + 8|0);
  _idct8($345,$344);
  $346 = ((($_y)) + 80|0);
  $347 = ((($w$i)) + 10|0);
  _idct8($347,$346);
  $348 = ((($_y)) + 96|0);
  $349 = ((($w$i)) + 12|0);
  _idct8($349,$348);
  $350 = ((($_y)) + 112|0);
  $351 = ((($w$i)) + 14|0);
  _idct8($351,$350);
  _idct8($_y,$w$i);
  $352 = ((($w$i)) + 16|0);
  $353 = ((($_y)) + 2|0);
  _idct8($353,$352);
  $354 = ((($w$i)) + 32|0);
  $355 = ((($_y)) + 4|0);
  _idct8($355,$354);
  $356 = ((($w$i)) + 48|0);
  $357 = ((($_y)) + 6|0);
  _idct8($357,$356);
  $358 = ((($w$i)) + 64|0);
  $359 = ((($_y)) + 8|0);
  _idct8($359,$358);
  $360 = ((($w$i)) + 80|0);
  $361 = ((($_y)) + 10|0);
  _idct8($361,$360);
  $362 = ((($w$i)) + 96|0);
  $363 = ((($_y)) + 12|0);
  _idct8($363,$362);
  $364 = ((($w$i)) + 112|0);
  $365 = ((($_y)) + 14|0);
  _idct8($365,$364);
  $366 = ((($_y)) + 128|0);
  $out$21$i = $_y;
  while(1) {
   $367 = HEAP16[$out$21$i>>1]|0;
   $368 = $367 << 16 >> 16;
   $369 = (($368) + 8)|0;
   $370 = $369 >>> 4;
   $371 = $370&65535;
   HEAP16[$out$21$i>>1] = $371;
   $372 = ((($out$21$i)) + 2|0);
   $373 = ($372>>>0)<($366>>>0);
   if ($373) {
    $out$21$i = $372;
   } else {
    break;
   }
  }
  STACKTOP = sp;return;
 }
 $99 = HEAP16[$_y>>1]|0;
 $100 = $99 << 16 >> 16;
 $101 = ($100*46341)|0;
 $102 = $101 >> 16;
 $103 = ((($_y)) + 4|0);
 $104 = HEAP16[$103>>1]|0;
 $105 = $104 << 16 >> 16;
 $106 = ($105*25080)|0;
 $107 = $106 >> 16;
 $108 = ($105*60547)|0;
 $109 = $108 >> 16;
 $110 = ((($_y)) + 2|0);
 $111 = HEAP16[$110>>1]|0;
 $112 = $111 << 16 >> 16;
 $113 = ($112*12785)|0;
 $114 = $113 >> 16;
 $115 = ((($_y)) + 6|0);
 $116 = HEAP16[$115>>1]|0;
 $117 = $116 << 16 >> 16;
 $118 = ($117*36410)|0;
 $119 = $118 >> 16;
 $120 = ($117*54491)|0;
 $121 = $120 >> 16;
 $122 = ($112*64277)|0;
 $123 = $122 >> 16;
 $124 = (($114) - ($119))|0;
 $125 = (($119) + ($114))|0;
 $sext$i$i = $125 << 16;
 $126 = $sext$i$i >> 16;
 $127 = ($126*46341)|0;
 $128 = $127 >> 16;
 $129 = (($121) + ($123))|0;
 $130 = (($123) - ($121))|0;
 $sext1$i$i = $130 << 16;
 $131 = $sext1$i$i >> 16;
 $132 = ($131*46341)|0;
 $133 = $132 >> 16;
 $134 = (($107) + ($102))|0;
 $135 = (($102) - ($107))|0;
 $136 = (($109) + ($102))|0;
 $137 = (($102) - ($109))|0;
 $138 = (($133) + ($128))|0;
 $139 = (($133) - ($128))|0;
 $140 = (($129) + ($136))|0;
 $141 = $140&65535;
 HEAP16[$w$i>>1] = $141;
 $142 = (($138) + ($134))|0;
 $143 = $142&65535;
 $144 = ((($w$i)) + 16|0);
 HEAP16[$144>>1] = $143;
 $145 = (($139) + ($135))|0;
 $146 = $145&65535;
 $147 = ((($w$i)) + 32|0);
 HEAP16[$147>>1] = $146;
 $148 = (($124) + ($137))|0;
 $149 = $148&65535;
 $150 = ((($w$i)) + 48|0);
 HEAP16[$150>>1] = $149;
 $151 = (($137) - ($124))|0;
 $152 = $151&65535;
 $153 = ((($w$i)) + 64|0);
 HEAP16[$153>>1] = $152;
 $154 = (($135) - ($139))|0;
 $155 = $154&65535;
 $156 = ((($w$i)) + 80|0);
 HEAP16[$156>>1] = $155;
 $157 = (($134) - ($138))|0;
 $158 = $157&65535;
 $159 = ((($w$i)) + 96|0);
 HEAP16[$159>>1] = $158;
 $160 = (($136) - ($129))|0;
 $161 = $160&65535;
 $162 = ((($w$i)) + 112|0);
 HEAP16[$162>>1] = $161;
 $163 = ((($w$i)) + 2|0);
 $164 = ((($_y)) + 16|0);
 $$val2$i = HEAP16[$164>>1]|0;
 $$idx3$i = ((($_y)) + 18|0);
 $$idx3$val$i = HEAP16[$$idx3$i>>1]|0;
 $$idx4$i = ((($_y)) + 20|0);
 $$idx4$val$i = HEAP16[$$idx4$i>>1]|0;
 $165 = $$val2$i << 16 >> 16;
 $166 = ($165*46341)|0;
 $167 = $166 >> 16;
 $168 = $$idx4$val$i << 16 >> 16;
 $169 = ($168*25080)|0;
 $170 = $169 >> 16;
 $171 = ($168*60547)|0;
 $172 = $171 >> 16;
 $173 = $$idx3$val$i << 16 >> 16;
 $174 = ($173*12785)|0;
 $175 = $174 >> 16;
 $176 = ($173*64277)|0;
 $177 = $176 >> 16;
 $178 = ($175*46341)|0;
 $179 = $178 >> 16;
 $180 = ($177*46341)|0;
 $181 = $180 >> 16;
 $182 = (($170) + ($167))|0;
 $183 = (($167) - ($170))|0;
 $184 = (($172) + ($167))|0;
 $185 = (($167) - ($172))|0;
 $186 = (($181) + ($179))|0;
 $187 = (($181) - ($179))|0;
 $188 = (($184) + ($177))|0;
 $189 = $188&65535;
 HEAP16[$163>>1] = $189;
 $190 = (($186) + ($182))|0;
 $191 = $190&65535;
 $192 = ((($w$i)) + 18|0);
 HEAP16[$192>>1] = $191;
 $193 = (($187) + ($183))|0;
 $194 = $193&65535;
 $195 = ((($w$i)) + 34|0);
 HEAP16[$195>>1] = $194;
 $196 = (($185) + ($175))|0;
 $197 = $196&65535;
 $198 = ((($w$i)) + 50|0);
 HEAP16[$198>>1] = $197;
 $199 = (($185) - ($175))|0;
 $200 = $199&65535;
 $201 = ((($w$i)) + 66|0);
 HEAP16[$201>>1] = $200;
 $202 = (($183) - ($187))|0;
 $203 = $202&65535;
 $204 = ((($w$i)) + 82|0);
 HEAP16[$204>>1] = $203;
 $205 = (($182) - ($186))|0;
 $206 = $205&65535;
 $207 = ((($w$i)) + 98|0);
 HEAP16[$207>>1] = $206;
 $208 = (($184) - ($177))|0;
 $209 = $208&65535;
 $210 = ((($w$i)) + 114|0);
 HEAP16[$210>>1] = $209;
 $211 = ((($w$i)) + 4|0);
 $212 = ((($_y)) + 32|0);
 $$val$i = HEAP16[$212>>1]|0;
 $$idx$i = ((($_y)) + 34|0);
 $$idx$val$i = HEAP16[$$idx$i>>1]|0;
 $213 = $$val$i << 16 >> 16;
 $214 = ($213*46341)|0;
 $215 = $214 >> 16;
 $216 = $$idx$val$i << 16 >> 16;
 $217 = ($216*12785)|0;
 $218 = $217 >> 16;
 $219 = ($216*64277)|0;
 $220 = $219 >> 16;
 $221 = ($218*46341)|0;
 $222 = $221 >> 16;
 $223 = ($220*46341)|0;
 $224 = $223 >> 16;
 $225 = (($224) + ($222))|0;
 $226 = (($224) - ($222))|0;
 $227 = (($220) + ($215))|0;
 $228 = $227&65535;
 HEAP16[$211>>1] = $228;
 $229 = (($225) + ($215))|0;
 $230 = $229&65535;
 $231 = ((($w$i)) + 20|0);
 HEAP16[$231>>1] = $230;
 $232 = (($226) + ($215))|0;
 $233 = $232&65535;
 $234 = ((($w$i)) + 36|0);
 HEAP16[$234>>1] = $233;
 $235 = (($218) + ($215))|0;
 $236 = $235&65535;
 $237 = ((($w$i)) + 52|0);
 HEAP16[$237>>1] = $236;
 $238 = (($215) - ($218))|0;
 $239 = $238&65535;
 $240 = ((($w$i)) + 68|0);
 HEAP16[$240>>1] = $239;
 $241 = (($215) - ($226))|0;
 $242 = $241&65535;
 $243 = ((($w$i)) + 84|0);
 HEAP16[$243>>1] = $242;
 $244 = (($215) - ($225))|0;
 $245 = $244&65535;
 $246 = ((($w$i)) + 100|0);
 HEAP16[$246>>1] = $245;
 $247 = (($215) - ($220))|0;
 $248 = $247&65535;
 $249 = ((($w$i)) + 116|0);
 HEAP16[$249>>1] = $248;
 $250 = ((($w$i)) + 6|0);
 $251 = ((($_y)) + 48|0);
 $$val1$i = HEAP16[$251>>1]|0;
 $252 = $$val1$i << 16 >> 16;
 $253 = ($252*46341)|0;
 $254 = $253 >>> 16;
 $255 = $254&65535;
 $256 = ((($w$i)) + 118|0);
 HEAP16[$256>>1] = $255;
 $257 = ((($w$i)) + 102|0);
 HEAP16[$257>>1] = $255;
 $258 = ((($w$i)) + 86|0);
 HEAP16[$258>>1] = $255;
 $259 = ((($w$i)) + 70|0);
 HEAP16[$259>>1] = $255;
 $260 = ((($w$i)) + 54|0);
 HEAP16[$260>>1] = $255;
 $261 = ((($w$i)) + 38|0);
 HEAP16[$261>>1] = $255;
 $262 = ((($w$i)) + 22|0);
 HEAP16[$262>>1] = $255;
 HEAP16[$250>>1] = $255;
 $in$08$i = $w$i;$out$09$i = $_y;
 while(1) {
  $263 = HEAP16[$in$08$i>>1]|0;
  $264 = $263 << 16 >> 16;
  $265 = ($264*46341)|0;
  $266 = $265 >> 16;
  $267 = ((($in$08$i)) + 4|0);
  $268 = HEAP16[$267>>1]|0;
  $269 = $268 << 16 >> 16;
  $270 = ($269*25080)|0;
  $271 = $270 >> 16;
  $272 = ($269*60547)|0;
  $273 = $272 >> 16;
  $274 = ((($in$08$i)) + 2|0);
  $275 = HEAP16[$274>>1]|0;
  $276 = $275 << 16 >> 16;
  $277 = ($276*12785)|0;
  $278 = $277 >> 16;
  $279 = ((($in$08$i)) + 6|0);
  $280 = HEAP16[$279>>1]|0;
  $281 = $280 << 16 >> 16;
  $282 = ($281*36410)|0;
  $283 = $282 >> 16;
  $284 = ($281*54491)|0;
  $285 = $284 >> 16;
  $286 = ($276*64277)|0;
  $287 = $286 >> 16;
  $288 = (($278) - ($283))|0;
  $289 = (($283) + ($278))|0;
  $sext$i5$i = $289 << 16;
  $290 = $sext$i5$i >> 16;
  $291 = ($290*46341)|0;
  $292 = $291 >> 16;
  $293 = (($285) + ($287))|0;
  $294 = (($287) - ($285))|0;
  $sext1$i6$i = $294 << 16;
  $295 = $sext1$i6$i >> 16;
  $296 = ($295*46341)|0;
  $297 = $296 >> 16;
  $298 = (($271) + ($266))|0;
  $299 = (($266) - ($271))|0;
  $300 = (($273) + ($266))|0;
  $301 = (($266) - ($273))|0;
  $302 = (($297) + ($292))|0;
  $303 = (($297) - ($292))|0;
  $304 = (($293) + ($300))|0;
  $305 = $304&65535;
  HEAP16[$out$09$i>>1] = $305;
  $306 = (($302) + ($298))|0;
  $307 = $306&65535;
  $308 = ((($out$09$i)) + 16|0);
  HEAP16[$308>>1] = $307;
  $309 = (($303) + ($299))|0;
  $310 = $309&65535;
  $311 = ((($out$09$i)) + 32|0);
  HEAP16[$311>>1] = $310;
  $312 = (($288) + ($301))|0;
  $313 = $312&65535;
  $314 = ((($out$09$i)) + 48|0);
  HEAP16[$314>>1] = $313;
  $315 = (($301) - ($288))|0;
  $316 = $315&65535;
  $317 = ((($out$09$i)) + 64|0);
  HEAP16[$317>>1] = $316;
  $318 = (($299) - ($303))|0;
  $319 = $318&65535;
  $320 = ((($out$09$i)) + 80|0);
  HEAP16[$320>>1] = $319;
  $321 = (($298) - ($302))|0;
  $322 = $321&65535;
  $323 = ((($out$09$i)) + 96|0);
  HEAP16[$323>>1] = $322;
  $324 = (($300) - ($293))|0;
  $325 = $324&65535;
  $326 = ((($out$09$i)) + 112|0);
  HEAP16[$326>>1] = $325;
  $327 = ((($in$08$i)) + 16|0);
  $328 = ((($out$09$i)) + 2|0);
  $329 = ($328>>>0)<($164>>>0);
  if ($329) {
   $in$08$i = $327;$out$09$i = $328;
  } else {
   break;
  }
 }
 $330 = ((($_y)) + 128|0);
 $out$17$i = $_y;
 while(1) {
  $331 = HEAP16[$out$17$i>>1]|0;
  $332 = $331 << 16 >> 16;
  $333 = (($332) + 8)|0;
  $334 = $333 >>> 4;
  $335 = $334&65535;
  HEAP16[$out$17$i>>1] = $335;
  $336 = ((($out$17$i)) + 2|0);
  $337 = ($336>>>0)<($330>>>0);
  if ($337) {
   $out$17$i = $336;
  } else {
   break;
  }
 }
 STACKTOP = sp;return;
}
function _idct8($_y,$_x) {
 $_y = $_y|0;
 $_x = $_x|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $sext = 0, $sext1 = 0, $sext2 = 0, $sext3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP16[$_x>>1]|0;
 $1 = $0 << 16 >> 16;
 $2 = ((($_x)) + 8|0);
 $3 = HEAP16[$2>>1]|0;
 $4 = $3 << 16 >> 16;
 $5 = (($4) + ($1))|0;
 $sext = $5 << 16;
 $6 = $sext >> 16;
 $7 = ($6*46341)|0;
 $8 = $7 >> 16;
 $9 = (($1) - ($4))|0;
 $sext1 = $9 << 16;
 $10 = $sext1 >> 16;
 $11 = ($10*46341)|0;
 $12 = $11 >> 16;
 $13 = ((($_x)) + 4|0);
 $14 = HEAP16[$13>>1]|0;
 $15 = $14 << 16 >> 16;
 $16 = ($15*25080)|0;
 $17 = $16 >> 16;
 $18 = ((($_x)) + 12|0);
 $19 = HEAP16[$18>>1]|0;
 $20 = $19 << 16 >> 16;
 $21 = ($20*60547)|0;
 $22 = $21 >> 16;
 $23 = (($17) - ($22))|0;
 $24 = ($15*60547)|0;
 $25 = $24 >> 16;
 $26 = ($20*25080)|0;
 $27 = $26 >> 16;
 $28 = (($27) + ($25))|0;
 $29 = ((($_x)) + 2|0);
 $30 = HEAP16[$29>>1]|0;
 $31 = $30 << 16 >> 16;
 $32 = ($31*12785)|0;
 $33 = $32 >> 16;
 $34 = ((($_x)) + 14|0);
 $35 = HEAP16[$34>>1]|0;
 $36 = $35 << 16 >> 16;
 $37 = ($36*64277)|0;
 $38 = $37 >> 16;
 $39 = (($33) - ($38))|0;
 $40 = ((($_x)) + 10|0);
 $41 = HEAP16[$40>>1]|0;
 $42 = $41 << 16 >> 16;
 $43 = ($42*54491)|0;
 $44 = $43 >> 16;
 $45 = ((($_x)) + 6|0);
 $46 = HEAP16[$45>>1]|0;
 $47 = $46 << 16 >> 16;
 $48 = ($47*36410)|0;
 $49 = $48 >> 16;
 $50 = (($44) - ($49))|0;
 $51 = ($42*36410)|0;
 $52 = $51 >> 16;
 $53 = ($47*54491)|0;
 $54 = $53 >> 16;
 $55 = (($54) + ($52))|0;
 $56 = ($31*64277)|0;
 $57 = $56 >> 16;
 $58 = ($36*12785)|0;
 $59 = $58 >> 16;
 $60 = (($59) + ($57))|0;
 $61 = (($50) + ($39))|0;
 $62 = (($39) - ($50))|0;
 $sext2 = $62 << 16;
 $63 = $sext2 >> 16;
 $64 = ($63*46341)|0;
 $65 = $64 >> 16;
 $66 = (($55) + ($60))|0;
 $67 = (($60) - ($55))|0;
 $sext3 = $67 << 16;
 $68 = $sext3 >> 16;
 $69 = ($68*46341)|0;
 $70 = $69 >> 16;
 $71 = (($8) + ($28))|0;
 $72 = (($8) - ($28))|0;
 $73 = (($12) + ($23))|0;
 $74 = (($12) - ($23))|0;
 $75 = (($70) + ($65))|0;
 $76 = (($70) - ($65))|0;
 $77 = (($66) + ($71))|0;
 $78 = $77&65535;
 HEAP16[$_y>>1] = $78;
 $79 = (($75) + ($73))|0;
 $80 = $79&65535;
 $81 = ((($_y)) + 16|0);
 HEAP16[$81>>1] = $80;
 $82 = (($76) + ($74))|0;
 $83 = $82&65535;
 $84 = ((($_y)) + 32|0);
 HEAP16[$84>>1] = $83;
 $85 = (($61) + ($72))|0;
 $86 = $85&65535;
 $87 = ((($_y)) + 48|0);
 HEAP16[$87>>1] = $86;
 $88 = (($72) - ($61))|0;
 $89 = $88&65535;
 $90 = ((($_y)) + 64|0);
 HEAP16[$90>>1] = $89;
 $91 = (($74) - ($76))|0;
 $92 = $91&65535;
 $93 = ((($_y)) + 80|0);
 HEAP16[$93>>1] = $92;
 $94 = (($73) - ($75))|0;
 $95 = $94&65535;
 $96 = ((($_y)) + 96|0);
 HEAP16[$96>>1] = $95;
 $97 = (($71) - ($66))|0;
 $98 = $97&65535;
 $99 = ((($_y)) + 112|0);
 HEAP16[$99>>1] = $98;
 return;
}
function _ft_bitmap_glyph_init($bitmap_glyph,$slot) {
 $bitmap_glyph = $bitmap_glyph|0;
 $slot = $slot|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$bitmap_glyph>>2]|0;
 $1 = ((($slot)) + 72|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(1651078259);
 if (!($3)) {
  $error$0 = 18;
  return ($error$0|0);
 }
 $4 = ((($slot)) + 100|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($bitmap_glyph)) + 20|0);
 HEAP32[$6>>2] = $5;
 $7 = ((($slot)) + 104|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($bitmap_glyph)) + 24|0);
 HEAP32[$9>>2] = $8;
 $10 = ((($slot)) + 156|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($11)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = $13 & 1;
 $15 = ($14|0)==(0);
 $16 = ((($bitmap_glyph)) + 28|0);
 if ($15) {
  _FT_Bitmap_Init($16);
  $22 = ((($slot)) + 76|0);
  $23 = (_FT_Bitmap_Copy($0,$22,$16)|0);
  $error$0 = $23;
  return ($error$0|0);
 } else {
  $17 = ((($slot)) + 76|0);
  ;HEAP32[$16>>2]=HEAP32[$17>>2]|0;HEAP32[$16+4>>2]=HEAP32[$17+4>>2]|0;HEAP32[$16+8>>2]=HEAP32[$17+8>>2]|0;HEAP32[$16+12>>2]=HEAP32[$17+12>>2]|0;HEAP32[$16+16>>2]=HEAP32[$17+16>>2]|0;HEAP32[$16+20>>2]=HEAP32[$17+20>>2]|0;
  $18 = HEAP32[$10>>2]|0;
  $19 = ((($18)) + 4|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = $20 & -2;
  HEAP32[$19>>2] = $21;
  $error$0 = 0;
  return ($error$0|0);
 }
 return (0)|0;
}
function _ft_bitmap_glyph_done($bitmap_glyph) {
 $bitmap_glyph = $bitmap_glyph|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$bitmap_glyph>>2]|0;
 $1 = ((($bitmap_glyph)) + 28|0);
 (_FT_Bitmap_Done($0,$1)|0);
 return;
}
function _ft_bitmap_glyph_copy($bitmap_source,$bitmap_target) {
 $bitmap_source = $bitmap_source|0;
 $bitmap_target = $bitmap_target|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$bitmap_source>>2]|0;
 $1 = ((($bitmap_source)) + 20|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($bitmap_target)) + 20|0);
 HEAP32[$3>>2] = $2;
 $4 = ((($bitmap_source)) + 24|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($bitmap_target)) + 24|0);
 HEAP32[$6>>2] = $5;
 $7 = ((($bitmap_source)) + 28|0);
 $8 = ((($bitmap_target)) + 28|0);
 $9 = (_FT_Bitmap_Copy($0,$7,$8)|0);
 return ($9|0);
}
function _ft_bitmap_glyph_bbox($bitmap_glyph,$cbox) {
 $bitmap_glyph = $bitmap_glyph|0;
 $cbox = $cbox|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($bitmap_glyph)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 6;
 HEAP32[$cbox>>2] = $2;
 $3 = ((($bitmap_glyph)) + 28|0);
 $4 = ((($3)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $5 << 6;
 $7 = (($6) + ($2))|0;
 $8 = ((($cbox)) + 8|0);
 HEAP32[$8>>2] = $7;
 $9 = ((($bitmap_glyph)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 6;
 $12 = ((($cbox)) + 12|0);
 HEAP32[$12>>2] = $11;
 $13 = HEAP32[$3>>2]|0;
 $14 = $13 << 6;
 $15 = (($11) - ($14))|0;
 $16 = ((($cbox)) + 4|0);
 HEAP32[$16>>2] = $15;
 return;
}
function _ft_outline_glyph_init($outline_glyph,$slot) {
 $outline_glyph = $outline_glyph|0;
 $slot = $slot|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($slot)) + 108|0);
 $1 = ((($outline_glyph)) + 20|0);
 $2 = ((($slot)) + 72|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(1869968492);
 if (!($4)) {
  $error$0 = 18;
  return ($error$0|0);
 }
 $5 = HEAP32[$outline_glyph>>2]|0;
 $6 = ((($slot)) + 110|0);
 $7 = HEAP16[$6>>1]|0;
 $8 = $7 << 16 >> 16;
 $9 = HEAP16[$0>>1]|0;
 $10 = $9 << 16 >> 16;
 $11 = (_FT_Outline_New($5,$8,$10,$1)|0);
 $12 = ($11|0)==(0);
 if (!($12)) {
  $error$0 = $11;
  return ($error$0|0);
 }
 (_FT_Outline_Copy($0,$1)|0);
 $error$0 = 0;
 return ($error$0|0);
}
function _ft_outline_glyph_done($outline_glyph) {
 $outline_glyph = $outline_glyph|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$outline_glyph>>2]|0;
 $1 = ((($outline_glyph)) + 20|0);
 (_FT_Outline_Done($0,$1)|0);
 return;
}
function _ft_outline_glyph_copy($outline_source,$outline_target) {
 $outline_source = $outline_source|0;
 $outline_target = $outline_target|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$outline_source>>2]|0;
 $1 = ((($outline_source)) + 20|0);
 $2 = ((($1)) + 2|0);
 $3 = HEAP16[$2>>1]|0;
 $4 = $3 << 16 >> 16;
 $5 = HEAP16[$1>>1]|0;
 $6 = $5 << 16 >> 16;
 $7 = ((($outline_target)) + 20|0);
 $8 = (_FT_Outline_New($0,$4,$6,$7)|0);
 $9 = ($8|0)==(0);
 if (!($9)) {
  return ($8|0);
 }
 (_FT_Outline_Copy($1,$7)|0);
 return ($8|0);
}
function _ft_outline_glyph_transform($outline_glyph,$matrix,$delta) {
 $outline_glyph = $outline_glyph|0;
 $matrix = $matrix|0;
 $delta = $delta|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($matrix|0)==(0|0);
 if (!($0)) {
  $1 = ((($outline_glyph)) + 20|0);
  _FT_Outline_Transform($1,$matrix);
 }
 $2 = ($delta|0)==(0|0);
 if ($2) {
  return;
 }
 $3 = ((($outline_glyph)) + 20|0);
 $4 = HEAP32[$delta>>2]|0;
 $5 = ((($delta)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 _FT_Outline_Translate($3,$4,$6);
 return;
}
function _ft_outline_glyph_bbox($outline_glyph,$bbox) {
 $outline_glyph = $outline_glyph|0;
 $bbox = $bbox|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($outline_glyph)) + 20|0);
 _FT_Outline_Get_CBox($0,$bbox);
 return;
}
function _ft_outline_glyph_prepare($outline_glyph,$slot) {
 $outline_glyph = $outline_glyph|0;
 $slot = $slot|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($slot)) + 72|0);
 HEAP32[$0>>2] = 1869968492;
 $1 = ((($slot)) + 108|0);
 $2 = ((($outline_glyph)) + 20|0);
 ;HEAP32[$1>>2]=HEAP32[$2>>2]|0;HEAP32[$1+4>>2]=HEAP32[$2+4>>2]|0;HEAP32[$1+8>>2]=HEAP32[$2+8>>2]|0;HEAP32[$1+12>>2]=HEAP32[$2+12>>2]|0;HEAP32[$1+16>>2]=HEAP32[$2+16>>2]|0;
 $3 = ((($slot)) + 124|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 & -2;
 HEAP32[$3>>2] = $5;
 return 0;
}
function _FT_Done_Glyph($glyph) {
 $glyph = $glyph|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($glyph|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = HEAP32[$glyph>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($glyph)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 12|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0|0);
 if (!($7)) {
  FUNCTION_TABLE_vi[$6 & 8191]($glyph);
 }
 _ft_mem_free($2,$glyph);
 return;
}
function _FT_Get_Glyph($slot,$aglyph) {
 $slot = $slot|0;
 $aglyph = $aglyph|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $clazz$1$ph = 0, $error$i = 0, $switch$split2D = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp;
 $0 = ($slot|0)==(0|0);
 if ($0) {
  $$0 = 37;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = HEAP32[$slot>>2]|0;
 $2 = ($aglyph|0)==(0|0);
 if ($2) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ((($slot)) + 72|0);
 $4 = HEAP32[$3>>2]|0;
 $switch$split2D = ($4|0)<(1869968492);
 L7: do {
  if ($switch$split2D) {
   switch ($4|0) {
   case 1651078259:  {
    $clazz$1$ph = 47728;
    break;
   }
   default: {
    label = 5;
   }
   }
  } else {
   switch ($4|0) {
   case 1869968492:  {
    break;
   }
   default: {
    label = 5;
    break L7;
   }
   }
   $clazz$1$ph = 47760;
  }
 } while(0);
 if ((label|0) == 5) {
  $5 = (_FT_Lookup_Renderer($1,$4,0)|0);
  $6 = ($5|0)==(0|0);
  $7 = ((($5)) + 20|0);
  if ($6) {
   $$0 = 18;
   STACKTOP = sp;return ($$0|0);
  } else {
   $clazz$1$ph = $7;
  }
 }
 $8 = HEAP32[$1>>2]|0;
 $9 = HEAP32[$clazz$1$ph>>2]|0;
 $10 = (_ft_mem_alloc($8,$9,$error$i)|0);
 $11 = HEAP32[$error$i>>2]|0;
 $12 = ($11|0)==(0);
 if (!($12)) {
  $$0 = $11;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$10>>2] = $1;
 $13 = ((($10)) + 4|0);
 HEAP32[$13>>2] = $clazz$1$ph;
 $14 = ((($clazz$1$ph)) + 4|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($10)) + 8|0);
 HEAP32[$16>>2] = $15;
 $17 = ((($slot)) + 64|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = $18 << 10;
 $20 = ((($10)) + 12|0);
 HEAP32[$20>>2] = $19;
 $21 = ((($slot)) + 68|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = $22 << 10;
 $24 = ((($10)) + 16|0);
 HEAP32[$24>>2] = $23;
 $25 = ((($clazz$1$ph)) + 8|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = (FUNCTION_TABLE_iii[$26 & 8191]($10,$slot)|0);
 $28 = ($27|0)==(0);
 if ($28) {
  HEAP32[$aglyph>>2] = $10;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $29 = HEAP32[$10>>2]|0;
 $30 = HEAP32[$29>>2]|0;
 $31 = HEAP32[$13>>2]|0;
 $32 = ((($31)) + 12|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ($33|0)==(0|0);
 if (!($34)) {
  FUNCTION_TABLE_vi[$33 & 8191]($10);
 }
 _ft_mem_free($30,$10);
 $$0 = $27;
 STACKTOP = sp;return ($$0|0);
}
function _FT_Glyph_To_Bitmap($the_glyph,$render_mode,$origin,$destroy) {
 $the_glyph = $the_glyph|0;
 $render_mode = $render_mode|0;
 $origin = $origin|0;
 $destroy = $destroy|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $dummy = 0, $dummy_internal = 0, $error$0 = 0, $error$1 = 0, $error$113 = 0;
 var $error$i = 0, $or$cond = 0, $or$cond3 = 0, $or$cond5 = 0, $v = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp + 208|0;
 $dummy = sp + 48|0;
 $dummy_internal = sp + 8|0;
 $v = sp;
 $0 = ($the_glyph|0)==(0|0);
 if ($0) {
  $error$113 = 6;
  STACKTOP = sp;return ($error$113|0);
 }
 $1 = HEAP32[$the_glyph>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $error$113 = 6;
  STACKTOP = sp;return ($error$113|0);
 }
 $3 = ((($1)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = HEAP32[$1>>2]|0;
 $6 = ($5|0)!=(0|0);
 $7 = ($4|0)!=(0|0);
 $or$cond = $7 & $6;
 if (!($or$cond)) {
  $error$113 = 6;
  STACKTOP = sp;return ($error$113|0);
 }
 $8 = ($4|0)==(47728|0);
 if ($8) {
  $error$113 = 0;
  STACKTOP = sp;return ($error$113|0);
 }
 $9 = ((($4)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10|0)==(0|0);
 if ($11) {
  $error$113 = 6;
  STACKTOP = sp;return ($error$113|0);
 }
 _memset(($dummy|0),0,156)|0;
 dest=$dummy_internal; stop=dest+40|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $12 = ((($dummy)) + 156|0);
 HEAP32[$12>>2] = $dummy_internal;
 HEAP32[$dummy>>2] = $5;
 $13 = ((($4)) + 4|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($dummy)) + 72|0);
 HEAP32[$15>>2] = $14;
 $16 = HEAP32[$5>>2]|0;
 $17 = (_ft_mem_alloc($16,52,$error$i)|0);
 $18 = HEAP32[$error$i>>2]|0;
 $19 = ($18|0)==(0);
 if (!($19)) {
  $error$113 = $18;
  STACKTOP = sp;return ($error$113|0);
 }
 HEAP32[$17>>2] = $5;
 $20 = ((($17)) + 4|0);
 HEAP32[$20>>2] = 47728;
 $21 = ((($17)) + 8|0);
 HEAP32[$21>>2] = 1651078259;
 $22 = ($origin|0)!=(0|0);
 if ($22) {
  $23 = HEAP32[$3>>2]|0;
  $24 = ($23|0)==(0|0);
  if (!($24)) {
   $25 = ((($23)) + 20|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = ($26|0)==(0|0);
   if (!($27)) {
    FUNCTION_TABLE_viii[$26 & 8191]($1,0,$origin);
   }
  }
 }
 $28 = HEAP32[$9>>2]|0;
 $29 = (FUNCTION_TABLE_iii[$28 & 8191]($1,$dummy)|0);
 $30 = ($29|0)==(0);
 if ($30) {
  $31 = HEAP32[$1>>2]|0;
  $32 = (_FT_Render_Glyph_Internal($31,$dummy,$render_mode)|0);
  $error$0 = $32;
 } else {
  $error$0 = $29;
 }
 $33 = ($destroy<<24>>24)==(0);
 $or$cond3 = $22 & $33;
 if ($or$cond3) {
  $34 = HEAP32[$origin>>2]|0;
  $35 = (0 - ($34))|0;
  HEAP32[$v>>2] = $35;
  $36 = ((($origin)) + 4|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = (0 - ($37))|0;
  $39 = ((($v)) + 4|0);
  HEAP32[$39>>2] = $38;
  $40 = HEAP32[$3>>2]|0;
  $41 = ($40|0)==(0|0);
  if (!($41)) {
   $42 = ((($40)) + 20|0);
   $43 = HEAP32[$42>>2]|0;
   $44 = ($43|0)==(0|0);
   if (!($44)) {
    FUNCTION_TABLE_viii[$43 & 8191]($1,0,$v);
   }
  }
 }
 $45 = ($error$0|0)==(0);
 do {
  if ($45) {
   $46 = HEAP32[$17>>2]|0;
   $47 = HEAP32[$15>>2]|0;
   $48 = ($47|0)==(1651078259);
   if ($48) {
    $49 = ((($dummy)) + 100|0);
    $50 = HEAP32[$49>>2]|0;
    $51 = ((($17)) + 20|0);
    HEAP32[$51>>2] = $50;
    $52 = ((($dummy)) + 104|0);
    $53 = HEAP32[$52>>2]|0;
    $54 = ((($17)) + 24|0);
    HEAP32[$54>>2] = $53;
    $55 = HEAP32[$12>>2]|0;
    $56 = ((($55)) + 4|0);
    $57 = HEAP32[$56>>2]|0;
    $58 = $57 & 1;
    $59 = ($58|0)==(0);
    $60 = ((($17)) + 28|0);
    if ($59) {
     _FT_Bitmap_Init($60);
     $66 = ((($dummy)) + 76|0);
     $67 = (_FT_Bitmap_Copy($46,$66,$60)|0);
     $68 = ($67|0)==(0);
     if (!($68)) {
      $error$1 = $67;
      break;
     }
    } else {
     $61 = ((($dummy)) + 76|0);
     ;HEAP32[$60>>2]=HEAP32[$61>>2]|0;HEAP32[$60+4>>2]=HEAP32[$61+4>>2]|0;HEAP32[$60+8>>2]=HEAP32[$61+8>>2]|0;HEAP32[$60+12>>2]=HEAP32[$61+12>>2]|0;HEAP32[$60+16>>2]=HEAP32[$61+16>>2]|0;HEAP32[$60+20>>2]=HEAP32[$61+20>>2]|0;
     $62 = HEAP32[$12>>2]|0;
     $63 = ((($62)) + 4|0);
     $64 = HEAP32[$63>>2]|0;
     $65 = $64 & -2;
     HEAP32[$63>>2] = $65;
    }
    $69 = ((($17)) + 12|0);
    $70 = ((($1)) + 12|0);
    $71 = $70;
    $72 = $71;
    $73 = HEAP32[$72>>2]|0;
    $74 = (($71) + 4)|0;
    $75 = $74;
    $76 = HEAP32[$75>>2]|0;
    $77 = $69;
    $78 = $77;
    HEAP32[$78>>2] = $73;
    $79 = (($77) + 4)|0;
    $80 = $79;
    HEAP32[$80>>2] = $76;
    if (!($33)) {
     $81 = HEAP32[$1>>2]|0;
     $82 = HEAP32[$81>>2]|0;
     $83 = HEAP32[$3>>2]|0;
     $84 = ((($83)) + 12|0);
     $85 = HEAP32[$84>>2]|0;
     $86 = ($85|0)==(0|0);
     if (!($86)) {
      FUNCTION_TABLE_vi[$85 & 8191]($1);
     }
     _ft_mem_free($82,$1);
    }
    HEAP32[$the_glyph>>2] = $17;
    $error$113 = 0;
    STACKTOP = sp;return ($error$113|0);
   } else {
    $error$1 = 18;
   }
  } else {
   $error$1 = $error$0;
  }
 } while(0);
 $87 = ($error$1|0)!=(0);
 $88 = ($17|0)!=(0|0);
 $or$cond5 = $88 & $87;
 if (!($or$cond5)) {
  $error$113 = $error$1;
  STACKTOP = sp;return ($error$113|0);
 }
 $89 = HEAP32[$17>>2]|0;
 $90 = HEAP32[$89>>2]|0;
 $91 = HEAP32[$20>>2]|0;
 $92 = ((($91)) + 12|0);
 $93 = HEAP32[$92>>2]|0;
 $94 = ($93|0)==(0|0);
 if (!($94)) {
  FUNCTION_TABLE_vi[$93 & 8191]($17);
 }
 _ft_mem_free($90,$17);
 $error$113 = $error$1;
 STACKTOP = sp;return ($error$113|0);
}
function _FT_Add_Default_Modules($library) {
 $library = $library|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 (_FT_Add_Module($library,62568)|0);
 (_FT_Add_Module($library,55596)|0);
 (_FT_Add_Module($library,56252)|0);
 (_FT_Add_Module($library,49368)|0);
 (_FT_Add_Module($library,51160)|0);
 (_FT_Add_Module($library,53376)|0);
 (_FT_Add_Module($library,58152)|0);
 (_FT_Add_Module($library,59040)|0);
 (_FT_Add_Module($library,53168)|0);
 (_FT_Add_Module($library,54212)|0);
 (_FT_Add_Module($library,54324)|0);
 (_FT_Add_Module($library,54276)|0);
 (_FT_Add_Module($library,54432)|0);
 (_FT_Add_Module($library,54960)|0);
 (_FT_Add_Module($library,55416)|0);
 (_FT_Add_Module($library,55476)|0);
 (_FT_Add_Module($library,55536)|0);
 (_FT_Add_Module($library,47872)|0);
 return;
}
function _FT_Init_FreeType($alibrary) {
 $alibrary = $alibrary|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_FT_New_Memory()|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  $$0 = 7;
  return ($$0|0);
 }
 $2 = (_FT_New_Library($0,$alibrary)|0);
 $3 = ($2|0)==(0);
 if ($3) {
  $4 = HEAP32[$alibrary>>2]|0;
  _FT_Add_Default_Modules($4);
  $$0 = 0;
  return ($$0|0);
 } else {
  _FT_Done_Memory($0);
  $$0 = $2;
  return ($$0|0);
 }
 return (0)|0;
}
function _FT_Done_FreeType($library) {
 $library = $library|0;
 var $$0 = 0, $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($library|0)==(0|0);
 if ($0) {
  $$0 = 33;
  return ($$0|0);
 }
 $1 = HEAP32[$library>>2]|0;
 (_FT_Done_Library($library)|0);
 _FT_Done_Memory($1);
 $$0 = 0;
 return ($$0|0);
}
function _ft_service_list_lookup($service_descriptors,$service_id) {
 $service_descriptors = $service_descriptors|0;
 $service_id = $service_id|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $desc$03 = 0, $desc$03$lcssa = 0, $or$cond = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($service_descriptors|0)!=(0|0);
 $1 = ($service_id|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  $result$0 = 0;
  return ($result$0|0);
 }
 $2 = HEAP32[$service_descriptors>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $result$0 = 0;
  return ($result$0|0);
 } else {
  $4 = $2;$desc$03 = $service_descriptors;
 }
 while(1) {
  $5 = (_strcmp($4,$service_id)|0);
  $6 = ($5|0)==(0);
  if ($6) {
   $desc$03$lcssa = $desc$03;
   break;
  }
  $9 = ((($desc$03)) + 8|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ($10|0)==(0|0);
  if ($11) {
   $result$0 = 0;
   label = 6;
   break;
  } else {
   $4 = $10;$desc$03 = $9;
  }
 }
 if ((label|0) == 6) {
  return ($result$0|0);
 }
 $7 = ((($desc$03$lcssa)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $result$0 = $8;
 return ($result$0|0);
}
function _ft_validator_init($valid,$base,$limit,$level) {
 $valid = $valid|0;
 $base = $base|0;
 $limit = $limit|0;
 $level = $level|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($valid)) + 156|0);
 HEAP32[$0>>2] = $base;
 $1 = ((($valid)) + 160|0);
 HEAP32[$1>>2] = $limit;
 $2 = ((($valid)) + 164|0);
 HEAP32[$2>>2] = $level;
 $3 = ((($valid)) + 168|0);
 HEAP32[$3>>2] = 0;
 return;
}
function _ft_validator_error($valid,$error) {
 $valid = $valid|0;
 $error = $error|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($valid)) + 168|0);
 HEAP32[$0>>2] = $error;
 _longjmp(($valid|0),1);
 // unreachable;
}
function _FT_Stream_New($library,$args,$astream) {
 $library = $library|0;
 $args = $args|0;
 $astream = $astream|0;
 var $$0 = 0, $$pr = 0, $$pre = 0, $$pre$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $stream$0$ph = 0, $stream$01 = 0, $stream$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 HEAP32[$astream>>2] = 0;
 $0 = ($library|0)==(0|0);
 if ($0) {
  $$0 = 33;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ($args|0)==(0|0);
 if ($1) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = HEAP32[$library>>2]|0;
 $3 = (_ft_mem_alloc($2,40,$error)|0);
 $4 = HEAP32[$error>>2]|0;
 $5 = ($4|0)==(0);
 if (!($5)) {
  $$0 = $4;
  STACKTOP = sp;return ($$0|0);
 }
 $6 = ((($3)) + 28|0);
 HEAP32[$6>>2] = $2;
 $7 = HEAP32[$args>>2]|0;
 $8 = $7 & 1;
 $9 = ($8|0)==(0);
 do {
  if ($9) {
   $14 = $7 & 4;
   $15 = ($14|0)==(0);
   if (!($15)) {
    $16 = ((($args)) + 12|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = (_FT_Stream_Open($3,$17)|0);
    HEAP32[$error>>2] = $18;
    $19 = HEAP32[$16>>2]|0;
    $20 = ((($3)) + 16|0);
    HEAP32[$20>>2] = $19;
    $stream$0$ph = $3;
    label = 12;
    break;
   }
   $21 = $7 & 2;
   $22 = ($21|0)==(0);
   if (!($22)) {
    $23 = ((($args)) + 16|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = ($24|0)==(0|0);
    if (!($25)) {
     _ft_mem_free($2,$3);
     $26 = HEAP32[$23>>2]|0;
     $stream$0$ph = $26;
     label = 12;
     break;
    }
   }
   HEAP32[$error>>2] = 6;
   $stream$01 = $3;
   label = 13;
  } else {
   $10 = ((($args)) + 4|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = ((($args)) + 8|0);
   $13 = HEAP32[$12>>2]|0;
   _FT_Stream_OpenMemory($3,$11,$13);
   $stream$0$ph = $3;
   label = 12;
  }
 } while(0);
 if ((label|0) == 12) {
  $$pr = HEAP32[$error>>2]|0;
  $27 = ($$pr|0)==(0);
  if ($27) {
   $28 = ((($stream$0$ph)) + 28|0);
   HEAP32[$28>>2] = $2;
   $$pre = 0;$stream$1 = $stream$0$ph;
  } else {
   $stream$01 = $stream$0$ph;
   label = 13;
  }
 }
 if ((label|0) == 13) {
  _ft_mem_free($2,$stream$01);
  $$pre$pre = HEAP32[$error>>2]|0;
  $$pre = $$pre$pre;$stream$1 = 0;
 }
 HEAP32[$astream>>2] = $stream$1;
 $$0 = $$pre;
 STACKTOP = sp;return ($$0|0);
}
function _FT_Stream_Free($stream,$external) {
 $stream = $stream|0;
 $external = $external|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($stream|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($stream)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 _FT_Stream_Close($stream);
 $3 = ($external|0)==(0);
 if (!($3)) {
  return;
 }
 _ft_mem_free($2,$stream);
 return;
}
function _ft_glyphslot_free_bitmap($slot) {
 $slot = $slot|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($slot)) + 156|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = ((($1)) + 4|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = $4 & 1;
  $6 = ($5|0)==(0);
  if (!($6)) {
   $7 = ((($slot)) + 4|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ((($8)) + 100|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ((($slot)) + 88|0);
   $12 = HEAP32[$11>>2]|0;
   _ft_mem_free($10,$12);
   HEAP32[$11>>2] = 0;
   $13 = HEAP32[$0>>2]|0;
   $14 = ((($13)) + 4|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = $15 & -2;
   HEAP32[$14>>2] = $16;
   return;
  }
 }
 $17 = ((($slot)) + 88|0);
 HEAP32[$17>>2] = 0;
 return;
}
function _ft_glyphslot_set_bitmap($slot,$buffer) {
 $slot = $slot|0;
 $buffer = $buffer|0;
 var $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($slot)) + 156|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = ((($1)) + 4|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = $4 & 1;
  $6 = ($5|0)==(0);
  if (!($6)) {
   $7 = ((($slot)) + 4|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ((($8)) + 100|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ((($slot)) + 88|0);
   $12 = HEAP32[$11>>2]|0;
   _ft_mem_free($10,$12);
   HEAP32[$11>>2] = 0;
   $13 = HEAP32[$0>>2]|0;
   $14 = ((($13)) + 4|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = $15 & -2;
   HEAP32[$14>>2] = $16;
   $$pre$phiZ2D = $11;
   HEAP32[$$pre$phiZ2D>>2] = $buffer;
   return;
  }
 }
 $17 = ((($slot)) + 88|0);
 HEAP32[$17>>2] = 0;
 $$pre$phiZ2D = $17;
 HEAP32[$$pre$phiZ2D>>2] = $buffer;
 return;
}
function _ft_glyphslot_alloc_bitmap($slot,$size) {
 $slot = $slot|0;
 $size = $size|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($slot)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 100|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($slot)) + 156|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 & 1;
 $9 = ($8|0)==(0);
 if ($9) {
  $12 = $7 | 1;
  HEAP32[$6>>2] = $12;
  $$pre = ((($slot)) + 88|0);
  $$pre$phiZ2D = $$pre;
 } else {
  $10 = ((($slot)) + 88|0);
  $11 = HEAP32[$10>>2]|0;
  _ft_mem_free($3,$11);
  HEAP32[$10>>2] = 0;
  $$pre$phiZ2D = $10;
 }
 $13 = (_ft_mem_alloc($3,$size,$error)|0);
 HEAP32[$$pre$phiZ2D>>2] = $13;
 $14 = HEAP32[$error>>2]|0;
 STACKTOP = sp;return ($14|0);
}
function _FT_New_GlyphSlot($face,$aslot) {
 $face = $face|0;
 $aslot = $aslot|0;
 var $$0 = 0, $$in$i = 0, $$ph = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $8 = 0, $9 = 0;
 var $error = 0, $error$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp + 4|0;
 $error = sp;
 $0 = ($face|0)==(0|0);
 if ($0) {
  $$0 = 35;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ((($face)) + 96|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ((($2)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($2)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($5)) + 44|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (_ft_mem_alloc($7,$9,$error)|0);
 $11 = HEAP32[$error>>2]|0;
 $12 = ($11|0)==(0);
 if (!($12)) {
  $75 = ($aslot|0)==(0|0);
  if ($75) {
   $$0 = $11;
   STACKTOP = sp;return ($$0|0);
  }
  HEAP32[$aslot>>2] = 0;
  $$0 = $11;
  STACKTOP = sp;return ($$0|0);
 }
 $13 = ((($10)) + 4|0);
 HEAP32[$13>>2] = $face;
 $14 = HEAP32[$1>>2]|0;
 $15 = ((($14)) + 12|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($14)) + 8|0);
 $18 = HEAP32[$17>>2]|0;
 HEAP32[$error$i>>2] = 0;
 $19 = ((($14)) + 4|0);
 $20 = HEAP32[$19>>2]|0;
 HEAP32[$10>>2] = $20;
 $21 = (_ft_mem_alloc($18,40,$error$i)|0);
 $22 = HEAP32[$error$i>>2]|0;
 $23 = ($22|0)==(0);
 do {
  if ($23) {
   $24 = ((($10)) + 156|0);
   HEAP32[$24>>2] = $21;
   $25 = HEAP32[$14>>2]|0;
   $26 = HEAP32[$25>>2]|0;
   $27 = $26 & 512;
   $28 = ($27|0)==(0);
   if ($28) {
    $29 = (_FT_GlyphLoader_New($18,$21)|0);
    HEAP32[$error$i>>2] = $29;
    $30 = ($29|0)==(0);
    if (!($30)) {
     $$ph = $29;
     label = 9;
     break;
    }
   }
   $31 = ((($16)) + 64|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = ($32|0)==(0|0);
   if ($33) {
    HEAP32[$error>>2] = 0;
   } else {
    $34 = (FUNCTION_TABLE_ii[$32 & 8191]($10)|0);
    HEAP32[$error>>2] = $34;
    $35 = ($34|0)==(0);
    if (!($35)) {
     break;
    }
   }
   $71 = ((($face)) + 84|0);
   $72 = HEAP32[$71>>2]|0;
   $73 = ((($10)) + 8|0);
   HEAP32[$73>>2] = $72;
   HEAP32[$71>>2] = $10;
   $74 = ($aslot|0)==(0|0);
   if ($74) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   HEAP32[$aslot>>2] = $10;
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  } else {
   $$ph = $22;
   label = 9;
  }
 } while(0);
 if ((label|0) == 9) {
  HEAP32[$error>>2] = $$ph;
 }
 $36 = HEAP32[$13>>2]|0;
 $37 = ((($36)) + 96|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = ((($38)) + 12|0);
 $40 = HEAP32[$39>>2]|0;
 $41 = ((($38)) + 8|0);
 $42 = HEAP32[$41>>2]|0;
 $43 = ((($40)) + 68|0);
 $44 = HEAP32[$43>>2]|0;
 $45 = ($44|0)==(0|0);
 if (!($45)) {
  FUNCTION_TABLE_vi[$44 & 8191]($10);
 }
 $46 = ((($10)) + 156|0);
 $47 = HEAP32[$46>>2]|0;
 $48 = ($47|0)==(0|0);
 if ($48) {
  label = 16;
 } else {
  $49 = ((($47)) + 4|0);
  $50 = HEAP32[$49>>2]|0;
  $51 = $50 & 1;
  $52 = ($51|0)==(0);
  if ($52) {
   label = 16;
  } else {
   $53 = HEAP32[$13>>2]|0;
   $54 = ((($53)) + 100|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = ((($10)) + 88|0);
   $57 = HEAP32[$56>>2]|0;
   _ft_mem_free($55,$57);
   HEAP32[$56>>2] = 0;
   $58 = HEAP32[$46>>2]|0;
   $59 = ((($58)) + 4|0);
   $60 = HEAP32[$59>>2]|0;
   $61 = $60 & -2;
   HEAP32[$59>>2] = $61;
   $64 = $58;
  }
 }
 if ((label|0) == 16) {
  $62 = ((($10)) + 88|0);
  HEAP32[$62>>2] = 0;
  $64 = $47;
 }
 $63 = ($64|0)==(0|0);
 if (!($63)) {
  $65 = HEAP32[$38>>2]|0;
  $66 = HEAP32[$65>>2]|0;
  $67 = $66 & 512;
  $68 = ($67|0)==(0);
  if ($68) {
   $69 = HEAP32[$64>>2]|0;
   _FT_GlyphLoader_Done($69);
   $70 = HEAP32[$46>>2]|0;
   HEAP32[$70>>2] = 0;
   $$in$i = $70;
  } else {
   $$in$i = $64;
  }
  _ft_mem_free($42,$$in$i);
  HEAP32[$46>>2] = 0;
 }
 _ft_mem_free($7,$10);
 $$pre = HEAP32[$error>>2]|0;
 $$0 = $$pre;
 STACKTOP = sp;return ($$0|0);
}
function _FT_Done_GlyphSlot($slot) {
 $slot = $slot|0;
 var $$in$i = 0, $$phi$trans$insert = 0, $$phi$trans$insert9 = 0, $$pre = 0, $$pre10 = 0, $$pre8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cur$0 = 0, $cur$03 = 0, $cur$05 = 0;
 var $prev$04 = 0, $prev$04$lcssa = 0, $prev$04$phi = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($slot|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($slot)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 96|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($2)) + 84|0);
 $cur$03 = HEAP32[$7>>2]|0;
 $8 = ($cur$03|0)==(0|0);
 if ($8) {
  return;
 } else {
  $cur$05 = $cur$03;$prev$04 = 0;
 }
 while(1) {
  $9 = ($cur$05|0)==($slot|0);
  if ($9) {
   $prev$04$lcssa = $prev$04;
   break;
  }
  $49 = ((($cur$05)) + 8|0);
  $cur$0 = HEAP32[$49>>2]|0;
  $50 = ($cur$0|0)==(0|0);
  if ($50) {
   label = 21;
   break;
  } else {
   $prev$04$phi = $cur$05;$cur$05 = $cur$0;$prev$04 = $prev$04$phi;
  }
 }
 if ((label|0) == 21) {
  return;
 }
 $10 = ($prev$04$lcssa|0)==(0|0);
 $11 = ((($slot)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 if ($10) {
  HEAP32[$7>>2] = $12;
 } else {
  $13 = ((($prev$04$lcssa)) + 8|0);
  HEAP32[$13>>2] = $12;
 }
 $14 = ((($slot)) + 20|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)==(0|0);
 if ($16) {
  $18 = $4;$48 = $6;
 } else {
  FUNCTION_TABLE_vi[$15 & 8191]($slot);
  $$pre = HEAP32[$1>>2]|0;
  $$phi$trans$insert = ((($$pre)) + 96|0);
  $$pre8 = HEAP32[$$phi$trans$insert>>2]|0;
  $$phi$trans$insert9 = ((($$pre8)) + 8|0);
  $$pre10 = HEAP32[$$phi$trans$insert9>>2]|0;
  $18 = $$pre8;$48 = $$pre10;
 }
 $17 = ((($18)) + 12|0);
 $19 = HEAP32[$17>>2]|0;
 $20 = ((($19)) + 68|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ($21|0)==(0|0);
 if (!($22)) {
  FUNCTION_TABLE_vi[$21 & 8191]($slot);
 }
 $23 = ((($slot)) + 156|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ($24|0)==(0|0);
 if ($25) {
  label = 14;
 } else {
  $26 = ((($24)) + 4|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = $27 & 1;
  $29 = ($28|0)==(0);
  if ($29) {
   label = 14;
  } else {
   $30 = HEAP32[$1>>2]|0;
   $31 = ((($30)) + 100|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = ((($slot)) + 88|0);
   $34 = HEAP32[$33>>2]|0;
   _ft_mem_free($32,$34);
   HEAP32[$33>>2] = 0;
   $35 = HEAP32[$23>>2]|0;
   $36 = ((($35)) + 4|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = $37 & -2;
   HEAP32[$36>>2] = $38;
   $41 = $35;
  }
 }
 if ((label|0) == 14) {
  $39 = ((($slot)) + 88|0);
  HEAP32[$39>>2] = 0;
  $41 = $24;
 }
 $40 = ($41|0)==(0|0);
 if (!($40)) {
  $42 = HEAP32[$18>>2]|0;
  $43 = HEAP32[$42>>2]|0;
  $44 = $43 & 512;
  $45 = ($44|0)==(0);
  if ($45) {
   $46 = HEAP32[$41>>2]|0;
   _FT_GlyphLoader_Done($46);
   $47 = HEAP32[$23>>2]|0;
   HEAP32[$47>>2] = 0;
   $$in$i = $47;
  } else {
   $$in$i = $41;
  }
  _ft_mem_free($48,$$in$i);
  HEAP32[$23>>2] = 0;
 }
 _ft_mem_free($6,$slot);
 return;
}
function _FT_Load_Glyph($face,$glyph_index,$load_flags) {
 $face = $face|0;
 $glyph_index = $glyph_index|0;
 $load_flags = $load_flags|0;
 var $$0 = 0, $$1 = 0, $$lcssa = 0, $$old1$i$i = 0, $$pre$i = 0, $$pre19 = 0, $$sink = 0, $$sink18 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0;
 var $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0;
 var $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0;
 var $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0;
 var $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0;
 var $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cur$2$i$i = 0, $error$1 = 0, $error$2 = 0, $error$4 = 0, $load_flags$ = 0, $mode2$0 = 0, $or$cond11 = 0, $or$cond15 = 0, $or$cond9 = 0, $result$0$i13 = 0, $switch$split2D = 0, dest = 0, label = 0;
 var sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = ($face|0)==(0|0);
 if ($0) {
  $$0 = 35;
  return ($$0|0);
 }
 $1 = ((($face)) + 88|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $$0 = 35;
  return ($$0|0);
 }
 $4 = ((($face)) + 84|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if ($6) {
  $$0 = 35;
  return ($$0|0);
 }
 $7 = ((($5)) + 156|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 if ($9) {
  label = 7;
 } else {
  $10 = ((($8)) + 4|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = $11 & 1;
  $13 = ($12|0)==(0);
  if ($13) {
   label = 7;
  } else {
   $14 = ((($5)) + 4|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ((($15)) + 100|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($5)) + 88|0);
   $19 = HEAP32[$18>>2]|0;
   _ft_mem_free($17,$19);
   HEAP32[$18>>2] = 0;
   $20 = HEAP32[$7>>2]|0;
   $21 = ((($20)) + 4|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = $22 & -2;
   HEAP32[$21>>2] = $23;
  }
 }
 if ((label|0) == 7) {
  $24 = ((($5)) + 88|0);
  HEAP32[$24>>2] = 0;
 }
 $25 = ((($5)) + 24|0);
 $26 = ((($5)) + 94|0);
 HEAP8[$26>>0] = 0;
 $27 = ((($5)) + 100|0);
 $28 = ((($5)) + 72|0);
 dest=$25; stop=dest+40|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 ;HEAP32[$28>>2]=0|0;HEAP32[$28+4>>2]=0|0;HEAP32[$28+8>>2]=0|0;HEAP32[$28+12>>2]=0|0;
 dest=$27; stop=dest+56|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $29 = ((($face)) + 96|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ((($30)) + 4|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ((($32)) + 160|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = $load_flags & 1024;
 $36 = ($35|0)==(0);
 $37 = $load_flags | 2049;
 $load_flags$ = $36 ? $load_flags : $37;
 $38 = $load_flags$ & 1;
 $39 = ($38|0)==(0);
 $40 = $load_flags$ & -15;
 $41 = $40 | 10;
 $$1 = $39 ? $load_flags$ : $41;
 $42 = ($34|0)!=(0|0);
 $43 = $$1 & 2;
 $44 = ($43|0)==(0);
 $45 = $$1 & 32770;
 $46 = ($45|0)==(0);
 $47 = $46 & $42;
 L16: do {
  if ($47) {
   $48 = HEAP32[$30>>2]|0;
   $49 = HEAP32[$48>>2]|0;
   $50 = $49 & 768;
   $51 = ($50|0)==(256);
   if ($51) {
    $52 = ((($face)) + 8|0);
    $53 = HEAP32[$52>>2]|0;
    $54 = $53 & 8192;
    $55 = ($54|0)==(0);
    if ($55) {
     $56 = $$1 & 2048;
     $57 = ($56|0)==(0);
     do {
      if ($57) {
       $58 = ((($face)) + 128|0);
       $59 = HEAP32[$58>>2]|0;
       $60 = ((($59)) + 8|0);
       $61 = HEAP32[$60>>2]|0;
       $62 = ($61|0)==(0);
       $63 = HEAP32[$59>>2]|0;
       $64 = ($63|0)==(0);
       if ($62) {
        if ($64) {
         label = 27;
         break L16;
        } else {
         break;
        }
       } else {
        if ($64) {
         break;
        } else {
         label = 27;
         break L16;
        }
       }
      }
     } while(0);
     $65 = $$1 & 32;
     $66 = ($65|0)==(0);
     if ($66) {
      $67 = $49 & 1024;
      $68 = ($67|0)==(0);
      $69 = $$1 & 983040;
      $70 = ($69|0)==(65536);
      $or$cond15 = $70 | $68;
      if (!($or$cond15)) {
       $71 = ((($face)) + 128|0);
       $72 = HEAP32[$71>>2]|0;
       $73 = ((($72)) + 52|0);
       $74 = HEAP8[$73>>0]|0;
       $75 = ($74<<24>>24)==(0);
       if ($75) {
        $76 = $53 & 8;
        $77 = ($76|0)==(0);
        if ($77) {
         label = 27;
         break;
        }
        $78 = ((($face)) + 696|0);
        $79 = HEAP32[$78>>2]|0;
        $80 = ($79|0)==(0);
        if ($80) {
         label = 27;
         break;
        }
        $81 = ((($face)) + 260|0);
        $82 = ((($81)) + 26|0);
        $83 = HEAP16[$82>>1]|0;
        $84 = ($83<<16>>16)==(0);
        if (!($84)) {
         label = 27;
         break;
        }
        $85 = ((($face)) + 628|0);
        $86 = HEAP32[$85>>2]|0;
        $87 = ($86|0)==(0);
        if (!($87)) {
         label = 27;
         break;
        }
        $88 = ((($face)) + 636|0);
        $89 = HEAP32[$88>>2]|0;
        $90 = ($89|0)==(0);
        if (!($90)) {
         label = 27;
         break;
        }
       }
      }
     }
     $91 = $53 & 2;
     $92 = ($91|0)!=(0);
     $93 = $$1 & 8;
     $94 = ($93|0)==(0);
     $or$cond9 = $94 & $92;
     if ($or$cond9) {
      $95 = ((($30)) + 12|0);
      $96 = HEAP32[$95>>2]|0;
      $97 = ((($96)) + 72|0);
      $98 = HEAP32[$97>>2]|0;
      $99 = HEAP32[$1>>2]|0;
      $100 = $$1 | 16384;
      $101 = (FUNCTION_TABLE_iiiii[$98 & 8191]($5,$99,$glyph_index,$100)|0);
      $102 = ($101|0)==(0);
      if ($102) {
       $103 = HEAP32[$28>>2]|0;
       $104 = ($103|0)==(1651078259);
       if ($104) {
        $error$1 = 0;
        break;
       }
      }
     }
     $105 = ((($face)) + 128|0);
     $106 = HEAP32[$105>>2]|0;
     $107 = ((($106)) + 24|0);
     $108 = HEAP32[$107>>2]|0;
     HEAP32[$107>>2] = 0;
     $109 = HEAP32[$34>>2]|0;
     $110 = ((($109)) + 20|0);
     $111 = HEAP32[$110>>2]|0;
     $112 = ((($111)) + 12|0);
     $113 = HEAP32[$112>>2]|0;
     $114 = HEAP32[$1>>2]|0;
     $115 = (FUNCTION_TABLE_iiiiii[$113 & 8191]($34,$5,$114,$glyph_index,$$1)|0);
     HEAP32[$107>>2] = $108;
     $error$1 = $115;
    } else {
     label = 27;
    }
   } else {
    label = 27;
   }
  } else {
   label = 27;
  }
 } while(0);
 if ((label|0) == 27) {
  $116 = ((($30)) + 12|0);
  $117 = HEAP32[$116>>2]|0;
  $118 = ((($117)) + 72|0);
  $119 = HEAP32[$118>>2]|0;
  $120 = HEAP32[$1>>2]|0;
  $121 = (FUNCTION_TABLE_iiiii[$119 & 8191]($5,$120,$glyph_index,$$1)|0);
  $122 = ($121|0)==(0);
  if (!($122)) {
   $$0 = $121;
   return ($$0|0);
  }
  $123 = HEAP32[$28>>2]|0;
  $124 = ($123|0)==(1869968492);
  if ($124) {
   $125 = ((($5)) + 108|0);
   $126 = (_FT_Outline_Check($125)|0);
   $127 = ($126|0)==(0);
   if (!($127)) {
    $$0 = $126;
    return ($$0|0);
   }
   if ($44) {
    $128 = $$1 & 16;
    $129 = ($128|0)==(0);
    if ($129) {
     $154 = ((($5)) + 44|0);
     $155 = HEAP32[$154>>2]|0;
     $156 = $155 & -64;
     HEAP32[$154>>2] = $156;
     $157 = ((($5)) + 48|0);
     $158 = HEAP32[$157>>2]|0;
     $159 = $158 & -64;
     HEAP32[$157>>2] = $159;
     $160 = ((($5)) + 32|0);
     $161 = HEAP32[$160>>2]|0;
     $162 = HEAP32[$25>>2]|0;
     $163 = (($161) + 63)|0;
     $164 = (($163) + ($162))|0;
     $165 = $164 & -64;
     $166 = ((($5)) + 36|0);
     $167 = HEAP32[$166>>2]|0;
     $168 = ((($5)) + 28|0);
     $169 = HEAP32[$168>>2]|0;
     $170 = (($167) - ($169))|0;
     $171 = $170 & -64;
     $172 = $161 & -64;
     HEAP32[$160>>2] = $172;
     $173 = (($167) + 63)|0;
     $174 = $173 & -64;
     HEAP32[$166>>2] = $174;
     $175 = (($165) - ($172))|0;
     HEAP32[$25>>2] = $175;
     $176 = (($174) - ($171))|0;
     HEAP32[$168>>2] = $176;
    } else {
     $130 = ((($5)) + 32|0);
     $131 = HEAP32[$130>>2]|0;
     $132 = $131 & -64;
     HEAP32[$130>>2] = $132;
     $133 = ((($5)) + 36|0);
     $134 = HEAP32[$133>>2]|0;
     $135 = (($134) + 63)|0;
     $136 = $135 & -64;
     HEAP32[$133>>2] = $136;
     $137 = ((($5)) + 44|0);
     $138 = HEAP32[$137>>2]|0;
     $139 = HEAP32[$25>>2]|0;
     $140 = (($138) + 63)|0;
     $141 = (($140) + ($139))|0;
     $142 = $141 & -64;
     $143 = ((($5)) + 48|0);
     $144 = HEAP32[$143>>2]|0;
     $145 = ((($5)) + 28|0);
     $146 = HEAP32[$145>>2]|0;
     $147 = (($144) + 63)|0;
     $148 = (($147) + ($146))|0;
     $149 = $148 & -64;
     $150 = $138 & -64;
     HEAP32[$137>>2] = $150;
     $151 = $144 & -64;
     HEAP32[$143>>2] = $151;
     $152 = (($142) - ($150))|0;
     HEAP32[$25>>2] = $152;
     $153 = (($149) - ($151))|0;
     HEAP32[$145>>2] = $153;
    }
    $177 = ((($5)) + 40|0);
    $178 = HEAP32[$177>>2]|0;
    $179 = (($178) + 32)|0;
    $180 = $179 & -64;
    HEAP32[$177>>2] = $180;
    $181 = ((($5)) + 52|0);
    $182 = HEAP32[$181>>2]|0;
    $183 = (($182) + 32)|0;
    $184 = $183 & -64;
    HEAP32[$181>>2] = $184;
    $error$1 = 0;
   } else {
    $error$1 = 0;
   }
  } else {
   $error$1 = 0;
  }
 }
 $185 = $$1 & 16;
 $186 = ($185|0)==(0);
 if ($186) {
  $189 = ((($5)) + 40|0);
  $190 = HEAP32[$189>>2]|0;
  $$sink = 0;$$sink18 = $190;
 } else {
  $187 = ((($5)) + 52|0);
  $188 = HEAP32[$187>>2]|0;
  $$sink = $188;$$sink18 = 0;
 }
 $191 = ((($5)) + 64|0);
 HEAP32[$191>>2] = $$sink18;
 $192 = ((($5)) + 68|0);
 HEAP32[$192>>2] = $$sink;
 $193 = $$1 & 8192;
 $194 = ($193|0)==(0);
 if ($194) {
  $195 = ((($face)) + 8|0);
  $196 = HEAP32[$195>>2]|0;
  $197 = $196 & 1;
  $198 = ($197|0)==(0);
  if (!($198)) {
   $199 = HEAP32[$1>>2]|0;
   $200 = ((($5)) + 56|0);
   $201 = HEAP32[$200>>2]|0;
   $202 = ((($199)) + 16|0);
   $203 = HEAP32[$202>>2]|0;
   $204 = (_FT_MulDiv($201,$203,64)|0);
   HEAP32[$200>>2] = $204;
   $205 = ((($5)) + 60|0);
   $206 = HEAP32[$205>>2]|0;
   $207 = ((($199)) + 20|0);
   $208 = HEAP32[$207>>2]|0;
   $209 = (_FT_MulDiv($206,$208,64)|0);
   HEAP32[$205>>2] = $209;
  }
 }
 $210 = $$1 & 2048;
 $211 = ($210|0)==(0);
 if ($211) {
  $212 = ((($face)) + 128|0);
  $213 = HEAP32[$212>>2]|0;
  $214 = ((($213)) + 24|0);
  $215 = HEAP32[$214>>2]|0;
  $216 = ($215|0)==(0);
  if ($216) {
   $error$4 = $error$1;
  } else {
   $217 = ((($5)) + 4|0);
   $218 = HEAP32[$217>>2]|0;
   $219 = ((($218)) + 96|0);
   $220 = HEAP32[$219>>2]|0;
   $221 = ((($220)) + 4|0);
   $222 = HEAP32[$221>>2]|0;
   $223 = ((($222)) + 156|0);
   $224 = HEAP32[$223>>2]|0;
   $225 = ($224|0)==(0|0);
   if ($225) {
    $$pre$i = HEAP32[$28>>2]|0;
    $238 = $$pre$i;
    label = 46;
   } else {
    $226 = ((($224)) + 16|0);
    $227 = HEAP32[$226>>2]|0;
    $228 = HEAP32[$28>>2]|0;
    $229 = ($227|0)==($228|0);
    if ($229) {
     $result$0$i13 = $224;
     label = 51;
    } else {
     $238 = $228;
     label = 46;
    }
   }
   L68: do {
    if ((label|0) == 46) {
     $230 = ($222|0)==(0|0);
     L70: do {
      if (!($230)) {
       $231 = ((($222)) + 148|0);
       $232 = HEAP32[$231>>2]|0;
       $$old1$i$i = ($232|0)==(0|0);
       if (!($$old1$i$i)) {
        $cur$2$i$i = $232;
        while(1) {
         $233 = ((($cur$2$i$i)) + 8|0);
         $234 = HEAP32[$233>>2]|0;
         $235 = ((($234)) + 16|0);
         $236 = HEAP32[$235>>2]|0;
         $237 = ($236|0)==($238|0);
         if ($237) {
          $$lcssa = $234;
          break;
         }
         $239 = ((($cur$2$i$i)) + 4|0);
         $240 = HEAP32[$239>>2]|0;
         $241 = ($240|0)==(0|0);
         if ($241) {
          break L70;
         } else {
          $cur$2$i$i = $240;
         }
        }
        $242 = ($$lcssa|0)==(0|0);
        if (!($242)) {
         $result$0$i13 = $$lcssa;
         label = 51;
         break L68;
        }
       }
      }
     } while(0);
     $249 = ($238|0)==(1869968492);
     if ($249) {
      $250 = $215 & 1;
      $251 = ($250|0)==(0);
      if ($251) {
       $254 = $215;
      } else {
       $252 = ((($5)) + 108|0);
       _FT_Outline_Transform($252,$213);
       $$pre19 = HEAP32[$214>>2]|0;
       $254 = $$pre19;
      }
      $253 = $254 & 2;
      $255 = ($253|0)==(0);
      if ($255) {
       $error$2 = $error$1;
      } else {
       $256 = ((($5)) + 108|0);
       $257 = ((($213)) + 16|0);
       $258 = HEAP32[$257>>2]|0;
       $259 = ((($213)) + 20|0);
       $260 = HEAP32[$259>>2]|0;
       _FT_Outline_Translate($256,$258,$260);
       $error$2 = $error$1;
      }
     } else {
      $error$2 = $error$1;
     }
    }
   } while(0);
   if ((label|0) == 51) {
    $243 = ((($result$0$i13)) + 12|0);
    $244 = HEAP32[$243>>2]|0;
    $245 = ((($244)) + 44|0);
    $246 = HEAP32[$245>>2]|0;
    $247 = ((($213)) + 16|0);
    $248 = (FUNCTION_TABLE_iiiii[$246 & 8191]($result$0$i13,$5,$213,$247)|0);
    $error$2 = $248;
   }
   $261 = ((($5)) + 64|0);
   _FT_Vector_Transform($261,$213);
   $error$4 = $error$2;
  }
 } else {
  $error$4 = $error$1;
 }
 $262 = ($error$4|0)==(0);
 if (!($262)) {
  $$0 = $error$4;
  return ($$0|0);
 }
 $263 = HEAP32[$28>>2]|0;
 $switch$split2D = ($263|0)<(1668246896);
 L89: do {
  if ($switch$split2D) {
   switch ($263|0) {
   case 1651078259:  {
    $$0 = 0;
    break;
   }
   default: {
    break L89;
   }
   }
   return ($$0|0);
  } else {
   switch ($263|0) {
   case 1668246896:  {
    $$0 = 0;
    break;
   }
   default: {
    break L89;
   }
   }
   return ($$0|0);
  }
 } while(0);
 $264 = $$1 & 4;
 $265 = ($264|0)==(0);
 if ($265) {
  $$0 = 0;
  return ($$0|0);
 }
 $266 = $$1 >>> 16;
 $267 = $266 & 15;
 $268 = ((($5)) + 4|0);
 $269 = HEAP32[$268>>2]|0;
 $270 = ($269|0)==(0|0);
 if ($270) {
  $$0 = 6;
  return ($$0|0);
 }
 $271 = ($267|0)!=(0);
 $272 = $$1 & 4096;
 $273 = ($272|0)==(0);
 $or$cond11 = $273 | $271;
 $mode2$0 = $or$cond11 ? $267 : 2;
 $274 = ((($269)) + 96|0);
 $275 = HEAP32[$274>>2]|0;
 $276 = ((($275)) + 4|0);
 $277 = HEAP32[$276>>2]|0;
 $278 = (_FT_Render_Glyph_Internal($277,$5,$mode2$0)|0);
 $$0 = $278;
 return ($$0|0);
}
function _FT_Render_Glyph_Internal($library,$slot,$render_mode) {
 $library = $library|0;
 $slot = $slot|0;
 $render_mode = $render_mode|0;
 var $$lcssa = 0, $$lcssa66 = 0, $$old1$i = 0, $$old1$i3 = 0, $$pre$phi33Z2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cur$0$i2 = 0, $cur$0$i2$in = 0, $cur$2$i = 0, $cur$2$i$lcssa = 0, $cur$2$i4 = 0, $cur$2$i4$lcssa = 0, $error$0$us$us = 0, $error$2 = 0;
 var $node$sroa$0$1$ph = 0, $node$sroa$0$1$ph34 = 0, $node$sroa$0$1$ph8$be = 0, $node$sroa$0$1$ph852 = 0, $or$cond = 0, $or$cond$us$us = 0, $renderer$1$ph = 0, $renderer$1$ph35 = 0, $renderer$1$ph9$be = 0, $renderer$1$ph953 = 0, $renderer$1$us$us = 0, $switch$split2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($slot)) + 72|0);
 $1 = HEAP32[$0>>2]|0;
 $switch$split2D = ($1|0)<(1869968492);
 L1: do {
  if ($switch$split2D) {
   switch ($1|0) {
   case 1651078259:  {
    $error$2 = 0;
    break;
   }
   default: {
    label = 3;
    break L1;
   }
   }
   return ($error$2|0);
  } else {
   switch ($1|0) {
   case 1869968492:  {
    break;
   }
   default: {
    label = 3;
    break L1;
   }
   }
   $2 = ((($library)) + 156|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = ((($library)) + 148|0);
   $5 = HEAP32[$4>>2]|0;
   $$pre$phi33Z2D = $4;$node$sroa$0$1$ph34 = $5;$renderer$1$ph35 = $3;
  }
 } while(0);
 do {
  if ((label|0) == 3) {
   $6 = ($library|0)==(0|0);
   if ($6) {
    $error$0$us$us = 7;$renderer$1$us$us = 0;
   } else {
    $7 = ((($library)) + 148|0);
    $8 = HEAP32[$7>>2]|0;
    $$old1$i = ($8|0)==(0|0);
    if ($$old1$i) {
     $$pre$phi33Z2D = $7;$node$sroa$0$1$ph34 = 0;$renderer$1$ph35 = 0;
     break;
    } else {
     $cur$2$i = $8;
    }
    while(1) {
     $9 = ((($cur$2$i)) + 8|0);
     $10 = HEAP32[$9>>2]|0;
     $11 = ((($10)) + 16|0);
     $12 = HEAP32[$11>>2]|0;
     $13 = ($12|0)==($1|0);
     if ($13) {
      $$lcssa66 = $10;$cur$2$i$lcssa = $cur$2$i;
      label = 6;
      break;
     }
     $15 = ((($cur$2$i)) + 4|0);
     $16 = HEAP32[$15>>2]|0;
     $17 = ($16|0)==(0|0);
     if ($17) {
      $node$sroa$0$1$ph = 0;$renderer$1$ph = 0;
      break;
     } else {
      $cur$2$i = $16;
     }
    }
    if ((label|0) == 6) {
     $14 = $cur$2$i$lcssa;
     $node$sroa$0$1$ph = $14;$renderer$1$ph = $$lcssa66;
    }
    $18 = ($library|0)==(0|0);
    if ($18) {
     $error$0$us$us = 7;$renderer$1$us$us = $renderer$1$ph;
    } else {
     $$pre$phi33Z2D = $7;$node$sroa$0$1$ph34 = $node$sroa$0$1$ph;$renderer$1$ph35 = $renderer$1$ph;
     break;
    }
   }
   while(1) {
    $20 = ($renderer$1$us$us|0)==(0|0);
    if ($20) {
     $error$2 = $error$0$us$us;
     label = 18;
     break;
    }
    $21 = ((($renderer$1$us$us)) + 60|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = (FUNCTION_TABLE_iiiii[$22 & 8191]($renderer$1$us$us,$slot,$render_mode,0)|0);
    $24 = ($23|0)!=(0);
    $25 = $23 & 255;
    $26 = ($25|0)==(19);
    $or$cond$us$us = $24 & $26;
    if ($or$cond$us$us) {
     $error$0$us$us = $23;$renderer$1$us$us = 0;
    } else {
     $error$2 = $23;
     label = 18;
     break;
    }
   }
   if ((label|0) == 18) {
    return ($error$2|0);
   }
  }
 } while(0);
 $19 = ($renderer$1$ph35|0)==(0|0);
 if ($19) {
  $error$2 = 7;
  return ($error$2|0);
 } else {
  $node$sroa$0$1$ph852 = $node$sroa$0$1$ph34;$renderer$1$ph953 = $renderer$1$ph35;
 }
 while(1) {
  $27 = ((($renderer$1$ph953)) + 60|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = (FUNCTION_TABLE_iiiii[$28 & 8191]($renderer$1$ph953,$slot,$render_mode,0)|0);
  $30 = ($29|0)!=(0);
  $31 = $29 & 255;
  $32 = ($31|0)==(19);
  $or$cond = $30 & $32;
  if (!($or$cond)) {
   $error$2 = $29;
   label = 18;
   break;
  }
  $33 = HEAP32[$0>>2]|0;
  $34 = ($node$sroa$0$1$ph852|0)==(0);
  $35 = $node$sroa$0$1$ph852;
  $36 = ((($35)) + 4|0);
  $cur$0$i2$in = $34 ? $$pre$phi33Z2D : $36;
  $cur$0$i2 = HEAP32[$cur$0$i2$in>>2]|0;
  $$old1$i3 = ($cur$0$i2|0)==(0|0);
  L28: do {
   if ($$old1$i3) {
    $node$sroa$0$1$ph8$be = 0;$renderer$1$ph9$be = 0;
   } else {
    $cur$2$i4 = $cur$0$i2;
    while(1) {
     $38 = ((($cur$2$i4)) + 8|0);
     $39 = HEAP32[$38>>2]|0;
     $40 = ((($39)) + 16|0);
     $41 = HEAP32[$40>>2]|0;
     $42 = ($41|0)==($33|0);
     if ($42) {
      $$lcssa = $39;$cur$2$i4$lcssa = $cur$2$i4;
      break;
     }
     $44 = ((($cur$2$i4)) + 4|0);
     $45 = HEAP32[$44>>2]|0;
     $46 = ($45|0)==(0|0);
     if ($46) {
      $node$sroa$0$1$ph8$be = 0;$renderer$1$ph9$be = 0;
      break L28;
     } else {
      $cur$2$i4 = $45;
     }
    }
    $43 = $cur$2$i4$lcssa;
    $node$sroa$0$1$ph8$be = $43;$renderer$1$ph9$be = $$lcssa;
   }
  } while(0);
  $37 = ($renderer$1$ph9$be|0)==(0|0);
  if ($37) {
   $error$2 = $29;
   label = 18;
   break;
  } else {
   $node$sroa$0$1$ph852 = $node$sroa$0$1$ph8$be;$renderer$1$ph953 = $renderer$1$ph9$be;
  }
 }
 if ((label|0) == 18) {
  return ($error$2|0);
 }
 return (0)|0;
}
function _FT_Get_Char_Index($face,$charcode) {
 $face = $face|0;
 $charcode = $charcode|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($face|0)==(0|0);
 if ($0) {
  $result$0 = 0;
  return ($result$0|0);
 }
 $1 = ((($face)) + 92|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $result$0 = 0;
  return ($result$0|0);
 }
 $4 = ((($2)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (FUNCTION_TABLE_iii[$7 & 8191]($2,$charcode)|0);
 $result$0 = $8;
 return ($result$0|0);
}
function _FT_Open_Face($library,$args,$face_index,$aface) {
 $library = $library|0;
 $args = $args|0;
 $face_index = $face_index|0;
 $aface = $aface|0;
 var $$4 = 0, $$cast = 0, $$lcssa56 = 0, $$not$i$i = 0, $$pre = 0, $$pre44 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0;
 var $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0;
 var $255 = 0, $256 = 0, $257 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $args2$i$i = 0, $cur$034 = 0, $driver$036 = 0, $driver$1 = 0, $driver$2 = 0, $driver$247 = 0, $driver$248 = 0, $driver$3 = 0, $driver$4$ph = 0, $driver$422 = 0, $error = 0, $error$0$i$i32 = 0, $error$0$i13 = 0, $error$1$i$i = 0, $error$2$i$i = 0, $errors$i$i = 0;
 var $face = 0, $file_names$i$i = 0, $i$0$i$i30 = 0, $i$028 = 0, $memory$0 = 0, $memory$045 = 0, $memory$046 = 0, $memory$1$ph = 0, $memory$121 = 0, $num_params$0 = 0, $num_params1$0 = 0, $offsets$i$i = 0, $or$cond = 0, $or$cond$i$i = 0, $or$cond$not = 0, $or$cond2$i$i = 0, $or$cond26 = 0, $or$cond3 = 0, $params$0 = 0, $params2$0 = 0;
 var $sext = 0, $size = 0, $stream = 0, $stream2$i$i = 0, $vfs_rfork_has_no_font$0$i$i33 = 0, $vfs_rfork_has_no_font$1$$i$i = 0, $vfs_rfork_has_no_font$1$i$i = 0, $vfs_rfork_has_no_font$2$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $file_names$i$i = sp + 128|0;
 $offsets$i$i = sp + 92|0;
 $errors$i$i = sp + 56|0;
 $args2$i$i = sp + 24|0;
 $stream2$i$i = sp + 16|0;
 $error = sp + 12|0;
 $stream = sp + 8|0;
 $face = sp + 4|0;
 $size = sp;
 HEAP32[$stream>>2] = 0;
 HEAP32[$face>>2] = 0;
 $0 = ($aface|0)==(0|0);
 $1 = ($face_index|0)>(-1);
 $or$cond = $1 & $0;
 $or$cond$not = $or$cond ^ 1;
 $2 = ($args|0)!=(0|0);
 $or$cond3 = $2 & $or$cond$not;
 if (!($or$cond3)) {
  $$4 = 6;
  STACKTOP = sp;return ($$4|0);
 }
 $3 = HEAP32[$args>>2]|0;
 $4 = $3 & 2;
 $5 = ($4|0)==(0);
 if ($5) {
  $10 = 0;
 } else {
  $6 = ((($args)) + 16|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)!=(0|0);
  $10 = $8;
 }
 $9 = $10&1;
 $11 = (_FT_Stream_New($library,$args,$stream)|0);
 HEAP32[$error>>2] = $11;
 $12 = ($11|0)==(0);
 L7: do {
  if ($12) {
   $13 = HEAP32[$library>>2]|0;
   $14 = HEAP32[$args>>2]|0;
   $15 = $14 & 8;
   $16 = ($15|0)==(0);
   do {
    if ($16) {
     label = 15;
    } else {
     $17 = ((($args)) + 20|0);
     $18 = HEAP32[$17>>2]|0;
     $19 = ($18|0)==(0|0);
     if ($19) {
      label = 15;
     } else {
      $20 = HEAP32[$18>>2]|0;
      $21 = HEAP32[$20>>2]|0;
      $22 = $21 & 1;
      $23 = ($22|0)==(0);
      if ($23) {
       HEAP32[$error>>2] = 32;
      } else {
       $24 = $14 & 16;
       $25 = ($24|0)==(0);
       if ($25) {
        $num_params$0 = 0;$params$0 = 0;
       } else {
        $26 = ((($args)) + 24|0);
        $27 = HEAP32[$26>>2]|0;
        $28 = ((($args)) + 28|0);
        $29 = HEAP32[$28>>2]|0;
        $num_params$0 = $27;$params$0 = $29;
       }
       $30 = (_open_face($18,$stream,$9,$face_index,$num_params$0,$params$0,$face)|0);
       HEAP32[$error>>2] = $30;
       $31 = ($30|0)==(0);
       if ($31) {
        $driver$3 = $18;
        break;
       }
      }
      $32 = HEAP32[$stream>>2]|0;
      $33 = ($32|0)==(0|0);
      if ($33) {
       $driver$4$ph = $18;$memory$1$ph = $13;
       label = 88;
       break L7;
      }
      $34 = ((($32)) + 28|0);
      $35 = HEAP32[$34>>2]|0;
      _FT_Stream_Close($32);
      if ($10) {
       $driver$4$ph = $18;$memory$1$ph = $13;
       label = 88;
       break L7;
      }
      _ft_mem_free($35,$32);
      $driver$4$ph = $18;$memory$1$ph = $13;
      label = 88;
      break L7;
     }
    }
   } while(0);
   L21: do {
    if ((label|0) == 15) {
     HEAP32[$error>>2] = 11;
     $36 = ((($library)) + 16|0);
     $37 = HEAP32[$36>>2]|0;
     $38 = (((($library)) + 20|0) + ($37<<2)|0);
     $39 = ($37|0)>(0);
     if (!($39)) {
      $driver$248 = 0;$memory$046 = $13;
      label = 30;
      break L7;
     }
     $40 = ((($library)) + 20|0);
     $41 = ((($args)) + 24|0);
     $42 = ((($args)) + 28|0);
     $256 = 11;$cur$034 = $40;$driver$036 = 0;
     while(1) {
      $43 = HEAP32[$cur$034>>2]|0;
      $44 = HEAP32[$43>>2]|0;
      $45 = HEAP32[$44>>2]|0;
      $46 = $45 & 1;
      $47 = ($46|0)==(0);
      if ($47) {
       $257 = $256;$driver$1 = $driver$036;
      } else {
       $48 = HEAP32[$args>>2]|0;
       $49 = $48 & 16;
       $50 = ($49|0)==(0);
       if ($50) {
        $num_params1$0 = 0;$params2$0 = 0;
       } else {
        $51 = HEAP32[$41>>2]|0;
        $52 = HEAP32[$42>>2]|0;
        $num_params1$0 = $51;$params2$0 = $52;
       }
       $53 = (_open_face($43,$stream,$9,$face_index,$num_params1$0,$params2$0,$face)|0);
       HEAP32[$error>>2] = $53;
       $54 = ($53|0)==(0);
       if ($54) {
        $driver$3 = $43;
        break L21;
       }
       $55 = HEAP32[$cur$034>>2]|0;
       $56 = HEAP32[$55>>2]|0;
       $57 = ((($56)) + 8|0);
       $58 = HEAP32[$57>>2]|0;
       $59 = (_strcmp($58,594296)|0);
       $60 = ($59|0)==(0);
       $61 = $53 & 255;
       $62 = ($61|0)==(142);
       $or$cond26 = $62 & $60;
       if ($or$cond26) {
        $63 = HEAP32[$stream>>2]|0;
        $64 = (_FT_Stream_Seek($63,0)|0);
        $65 = ($64|0)==(0);
        if (!($65)) {
         $77 = $53;$driver$2 = $43;$memory$0 = $13;
         label = 29;
         break L7;
        }
        $66 = (_open_face_PS_from_sfnt_stream($library,$63,$face_index,0,0,$aface)|0);
        HEAP32[$error>>2] = $66;
        $67 = ($66|0)==(0);
        if ($67) {
         $$lcssa56 = $63;
         break;
        } else {
         $72 = $66;
        }
       } else {
        $72 = $53;
       }
       $71 = $72 & 255;
       $73 = ($71|0)==(2);
       if ($73) {
        $257 = $72;$driver$1 = $43;
       } else {
        $77 = $72;$driver$2 = $43;$memory$0 = $13;
        label = 29;
        break L7;
       }
      }
      $74 = ((($cur$034)) + 4|0);
      $75 = ($74>>>0)<($38>>>0);
      if ($75) {
       $256 = $257;$cur$034 = $74;$driver$036 = $driver$1;
      } else {
       $77 = $257;$driver$2 = $driver$1;$memory$0 = $13;
       label = 29;
       break L7;
      }
     }
     $68 = ($$lcssa56|0)==(0|0);
     if ($68) {
      $$4 = 0;
      STACKTOP = sp;return ($$4|0);
     }
     $69 = ((($$lcssa56)) + 28|0);
     $70 = HEAP32[$69>>2]|0;
     _FT_Stream_Close($$lcssa56);
     if ($10) {
      $$4 = 0;
      STACKTOP = sp;return ($$4|0);
     }
     _ft_mem_free($70,$$lcssa56);
     $$4 = 0;
     STACKTOP = sp;return ($$4|0);
    }
   } while(0);
   $132 = (_ft_mem_alloc($13,12,$error)|0);
   $133 = HEAP32[$error>>2]|0;
   $134 = ($133|0)==(0);
   L46: do {
    if ($134) {
     $135 = HEAP32[$face>>2]|0;
     $136 = ((($132)) + 8|0);
     HEAP32[$136>>2] = $135;
     $$cast = $135;
     $137 = ((($$cast)) + 96|0);
     $138 = HEAP32[$137>>2]|0;
     $139 = ((($138)) + 16|0);
     _FT_List_Add($139,$132);
     do {
      if ($1) {
       $140 = (_FT_New_GlyphSlot($$cast,0)|0);
       HEAP32[$error>>2] = $140;
       $141 = ($140|0)==(0);
       if (!($141)) {
        $216 = $$cast;
        break L46;
       }
       $142 = (_FT_New_Size($$cast,$size)|0);
       HEAP32[$error>>2] = $142;
       $143 = ($142|0)==(0);
       if ($143) {
        $144 = HEAP32[$size>>2]|0;
        $145 = ((($$cast)) + 88|0);
        HEAP32[$145>>2] = $144;
        $$pre = HEAP32[$face>>2]|0;
        $147 = $$pre;
        break;
       } else {
        $216 = $$cast;
        break L46;
       }
      } else {
       $147 = $$cast;
      }
     } while(0);
     $146 = ((($147)) + 8|0);
     $148 = HEAP32[$146>>2]|0;
     $149 = $148 & 1;
     $150 = ($149|0)==(0);
     if (!($150)) {
      $151 = ((($147)) + 74|0);
      $152 = HEAP16[$151>>1]|0;
      $153 = ($152<<16>>16)<(0);
      if ($153) {
       $154 = $152&65535;
       $155 = (0 - ($154))|0;
       $156 = $155&65535;
       HEAP16[$151>>1] = $156;
      }
      $157 = ((($147)) + 8|0);
      $158 = HEAP32[$157>>2]|0;
      $159 = $158 & 32;
      $160 = ($159|0)==(0);
      if ($160) {
       $161 = ((($147)) + 74|0);
       $162 = HEAP16[$161>>1]|0;
       $163 = ((($147)) + 78|0);
       HEAP16[$163>>1] = $162;
      }
     }
     $164 = ((($147)) + 8|0);
     $165 = HEAP32[$164>>2]|0;
     $166 = $165 & 2;
     $167 = ($166|0)==(0);
     if (!($167)) {
      $168 = ((($147)) + 28|0);
      $169 = HEAP32[$168>>2]|0;
      $170 = ($169|0)>(0);
      if ($170) {
       $171 = HEAP32[$face>>2]|0;
       $172 = ((($171)) + 28|0);
       $173 = HEAP32[$172>>2]|0;
       $175 = $147;$i$028 = 0;
       while(1) {
        $174 = ((($175)) + 32|0);
        $176 = HEAP32[$174>>2]|0;
        $177 = (($176) + ($i$028<<4)|0);
        $178 = HEAP16[$177>>1]|0;
        $179 = ($178<<16>>16)<(0);
        if ($179) {
         $180 = $178&65535;
         $181 = (0 - ($180))|0;
         $182 = $181&65535;
         HEAP16[$177>>1] = $182;
        }
        $183 = (((($176) + ($i$028<<4)|0)) + 8|0);
        $184 = HEAP32[$183>>2]|0;
        $185 = ($184|0)<(0);
        if ($185) {
         $186 = (0 - ($184))|0;
         $sext = $186 << 16;
         $187 = $sext >> 16;
         HEAP32[$183>>2] = $187;
        }
        $188 = (((($176) + ($i$028<<4)|0)) + 12|0);
        $189 = HEAP32[$188>>2]|0;
        $190 = ($189|0)<(0);
        if ($190) {
         $191 = (0 - ($189))|0;
         HEAP32[$188>>2] = $191;
        }
        $192 = (($i$028) + 1)|0;
        $193 = ($192|0)<($173|0);
        if ($193) {
         $175 = $171;$i$028 = $192;
        } else {
         break;
        }
       }
      }
     }
     $194 = HEAP32[$face>>2]|0;
     $195 = ((($194)) + 128|0);
     $196 = HEAP32[$195>>2]|0;
     HEAP32[$196>>2] = 65536;
     $197 = ((($196)) + 4|0);
     HEAP32[$197>>2] = 0;
     $198 = ((($196)) + 8|0);
     HEAP32[$198>>2] = 0;
     $199 = ((($196)) + 12|0);
     HEAP32[$199>>2] = 65536;
     $200 = ((($196)) + 16|0);
     HEAP32[$200>>2] = 0;
     $201 = ((($196)) + 20|0);
     HEAP32[$201>>2] = 0;
     $202 = ((($196)) + 56|0);
     HEAP32[$202>>2] = 1;
     if (!($0)) {
      HEAP32[$aface>>2] = $194;
      break L7;
     }
     $203 = ($194|0)==(0|0);
     if ($203) {
      break L7;
     }
     $204 = ((($194)) + 96|0);
     $205 = HEAP32[$204>>2]|0;
     $206 = ($205|0)==(0|0);
     if ($206) {
      break L7;
     }
     HEAP32[$202>>2] = 0;
     $207 = ((($205)) + 8|0);
     $208 = HEAP32[$207>>2]|0;
     $209 = ((($205)) + 16|0);
     $210 = (_FT_List_Find($209,$194)|0);
     $211 = ($210|0)==(0|0);
     if ($211) {
      break L7;
     }
     _FT_List_Remove($209,$210);
     _ft_mem_free($208,$210);
     _destroy_face($208,$194,$205);
     break L7;
    } else {
     $213 = ($132|0)==(0|0);
     $214 = HEAP32[$face>>2]|0;
     if ($213) {
      $232 = $214;$driver$422 = $driver$3;$memory$121 = $13;
      label = 95;
      break L7;
     } else {
      $216 = $214;
     }
    }
   } while(0);
   $215 = ($216|0)==(0|0);
   if (!($215)) {
    $217 = ((($216)) + 96|0);
    $218 = HEAP32[$217>>2]|0;
    $219 = ($218|0)==(0|0);
    if (!($219)) {
     $220 = ((($216)) + 128|0);
     $221 = HEAP32[$220>>2]|0;
     $222 = ((($221)) + 56|0);
     $223 = HEAP32[$222>>2]|0;
     $224 = (($223) + -1)|0;
     HEAP32[$222>>2] = $224;
     $225 = ($223|0)>(1);
     if (!($225)) {
      $226 = ((($218)) + 8|0);
      $227 = HEAP32[$226>>2]|0;
      $228 = ((($218)) + 16|0);
      $229 = (_FT_List_Find($228,$216)|0);
      $230 = ($229|0)==(0|0);
      if (!($230)) {
       _FT_List_Remove($228,$229);
       _ft_mem_free($227,$229);
       _destroy_face($227,$216,$218);
      }
     }
    }
   }
  } else {
   $77 = $11;$driver$2 = 0;$memory$0 = 0;
   label = 29;
  }
 } while(0);
 L89: do {
  if ((label|0) == 29) {
   $76 = $77 & 255;
   switch ($76|0) {
   case 85: case 2: case 81:  {
    break;
   }
   default: {
    $driver$248 = $driver$2;$memory$046 = $memory$0;
    label = 30;
    break L89;
   }
   }
   $78 = HEAP32[$stream>>2]|0;
   $79 = (_IsMacBinary($library,$78,$face_index,$aface)|0);
   $80 = $79 & 255;
   $81 = ($80|0)==(2);
   if ($81) {
    $82 = (_IsMacResource($library,$78,0,$face_index,$aface)|0);
    $error$0$i13 = $82;
   } else {
    $error$0$i13 = $79;
   }
   $83 = $error$0$i13 & 255;
   L95: do {
    switch ($83|0) {
    case 85: case 2:  {
     $84 = HEAP32[$args>>2]|0;
     $85 = $84 & 4;
     $86 = ($85|0)==(0);
     if ($86) {
      label = 51;
     } else {
      $87 = HEAP32[$library>>2]|0;
      HEAP32[$stream2$i$i>>2] = 0;
      $88 = ((($args)) + 12|0);
      $89 = HEAP32[$88>>2]|0;
      _FT_Raccess_Guess($library,$78,$89,$file_names$i$i,$offsets$i$i,$errors$i$i);
      $90 = ((($args2$i$i)) + 12|0);
      $error$0$i$i32 = 2;$i$0$i$i30 = 0;$vfs_rfork_has_no_font$0$i$i33 = 0;
      while(1) {
       $91 = (_ft_raccess_rule_by_darwin_vfs($library,$i$0$i$i30)|0);
       $92 = ($91<<24>>24)!=(0);
       $$not$i$i = $92 ^ 1;
       $93 = ($vfs_rfork_has_no_font$0$i$i33<<24>>24)==(0);
       $or$cond$i$i = $93 | $$not$i$i;
       if ($or$cond$i$i) {
        $94 = (($errors$i$i) + ($i$0$i$i30<<2)|0);
        $95 = HEAP32[$94>>2]|0;
        $96 = ($95|0)==(0);
        if ($96) {
         HEAP32[$args2$i$i>>2] = 4;
         $97 = (($file_names$i$i) + ($i$0$i$i30<<2)|0);
         $98 = HEAP32[$97>>2]|0;
         $99 = ($98|0)==(0|0);
         if ($99) {
          $100 = HEAP32[$88>>2]|0;
          $101 = $100;
         } else {
          $101 = $98;
         }
         HEAP32[$90>>2] = $101;
         $102 = (_FT_Stream_New($library,$args2$i$i,$stream2$i$i)|0);
         $103 = $102 & 255;
         $104 = ($103|0)==(81);
         $or$cond2$i$i = $92 & $104;
         $vfs_rfork_has_no_font$1$i$i = $or$cond2$i$i ? 1 : $vfs_rfork_has_no_font$0$i$i33;
         $105 = ($102|0)==(0);
         if ($105) {
          $106 = HEAP32[$stream2$i$i>>2]|0;
          $107 = (($offsets$i$i) + ($i$0$i$i30<<2)|0);
          $108 = HEAP32[$107>>2]|0;
          $109 = (_IsMacResource($library,$106,$108,$face_index,$aface)|0);
          $110 = ($106|0)==(0|0);
          if (!($110)) {
           $111 = ((($106)) + 28|0);
           $112 = HEAP32[$111>>2]|0;
           _FT_Stream_Close($106);
           _ft_mem_free($112,$106);
          }
          $113 = ($109|0)==(0);
          if ($113) {
           $error$2$i$i = 0;
           break;
          }
          $114 = ($91<<24>>24)==(0);
          $vfs_rfork_has_no_font$1$$i$i = $114 ? $vfs_rfork_has_no_font$1$i$i : 1;
          $error$1$i$i = $109;$vfs_rfork_has_no_font$2$i$i = $vfs_rfork_has_no_font$1$$i$i;
         } else {
          $error$1$i$i = $102;$vfs_rfork_has_no_font$2$i$i = $vfs_rfork_has_no_font$1$i$i;
         }
        } else {
         $error$1$i$i = $error$0$i$i32;$vfs_rfork_has_no_font$2$i$i = $vfs_rfork_has_no_font$0$i$i33;
        }
       } else {
        $error$1$i$i = $error$0$i$i32;$vfs_rfork_has_no_font$2$i$i = $vfs_rfork_has_no_font$0$i$i33;
       }
       $115 = (($i$0$i$i30) + 1)|0;
       $116 = ($115>>>0)<(9);
       if ($116) {
        $error$0$i$i32 = $error$1$i$i;$i$0$i$i30 = $115;$vfs_rfork_has_no_font$0$i$i33 = $vfs_rfork_has_no_font$2$i$i;
       } else {
        $error$2$i$i = $error$1$i$i;
        break;
       }
      }
      $117 = HEAP32[$file_names$i$i>>2]|0;
      $118 = ($117|0)==(0|0);
      if (!($118)) {
       _ft_mem_free($87,$117);
       HEAP32[$file_names$i$i>>2] = 0;
      }
      $119 = ((($file_names$i$i)) + 4|0);
      $120 = HEAP32[$119>>2]|0;
      $121 = ($120|0)==(0|0);
      if (!($121)) {
       _ft_mem_free($87,$120);
       HEAP32[$119>>2] = 0;
      }
      $234 = ((($file_names$i$i)) + 8|0);
      $235 = HEAP32[$234>>2]|0;
      $236 = ($235|0)==(0|0);
      if (!($236)) {
       _ft_mem_free($87,$235);
       HEAP32[$234>>2] = 0;
      }
      $237 = ((($file_names$i$i)) + 12|0);
      $238 = HEAP32[$237>>2]|0;
      $239 = ($238|0)==(0|0);
      if (!($239)) {
       _ft_mem_free($87,$238);
       HEAP32[$237>>2] = 0;
      }
      $240 = ((($file_names$i$i)) + 16|0);
      $241 = HEAP32[$240>>2]|0;
      $242 = ($241|0)==(0|0);
      if (!($242)) {
       _ft_mem_free($87,$241);
       HEAP32[$240>>2] = 0;
      }
      $243 = ((($file_names$i$i)) + 20|0);
      $244 = HEAP32[$243>>2]|0;
      $245 = ($244|0)==(0|0);
      if (!($245)) {
       _ft_mem_free($87,$244);
       HEAP32[$243>>2] = 0;
      }
      $246 = ((($file_names$i$i)) + 24|0);
      $247 = HEAP32[$246>>2]|0;
      $248 = ($247|0)==(0|0);
      if (!($248)) {
       _ft_mem_free($87,$247);
       HEAP32[$246>>2] = 0;
      }
      $249 = ((($file_names$i$i)) + 28|0);
      $250 = HEAP32[$249>>2]|0;
      $251 = ($250|0)==(0|0);
      if (!($251)) {
       _ft_mem_free($87,$250);
       HEAP32[$249>>2] = 0;
      }
      $252 = ((($file_names$i$i)) + 32|0);
      $253 = HEAP32[$252>>2]|0;
      $254 = ($253|0)==(0|0);
      if (!($254)) {
       _ft_mem_free($87,$253);
       HEAP32[$252>>2] = 0;
      }
      $255 = ($error$2$i$i|0)==(0);
      if ($255) {
       HEAP32[$error>>2] = 0;
       break L95;
      } else {
       HEAP32[$error>>2] = 2;
       label = 57;
       break L95;
      }
     }
     break;
    }
    default: {
     label = 51;
    }
    }
   } while(0);
   if ((label|0) == 51) {
    HEAP32[$error>>2] = $error$0$i13;
    $122 = ($error$0$i13|0)==(0);
    if (!($122)) {
     $127 = ($83|0)==(2);
     if ($127) {
      label = 57;
     } else {
      $129 = $78;$driver$247 = $driver$2;$memory$045 = $memory$0;
      label = 58;
      break;
     }
    }
   }
   if ((label|0) == 57) {
    HEAP32[$error>>2] = 2;
    $129 = $78;$driver$247 = $driver$2;$memory$045 = $memory$0;
    label = 58;
    break;
   }
   $123 = ($78|0)==(0|0);
   if (!($123)) {
    $124 = ((($78)) + 28|0);
    $125 = HEAP32[$124>>2]|0;
    _FT_Stream_Close($78);
    if (!($10)) {
     _ft_mem_free($125,$78);
    }
   }
   $126 = HEAP32[$error>>2]|0;
   $$4 = $126;
   STACKTOP = sp;return ($$4|0);
  }
 } while(0);
 if ((label|0) == 30) {
  $$pre44 = HEAP32[$stream>>2]|0;
  $129 = $$pre44;$driver$247 = $driver$248;$memory$045 = $memory$046;
  label = 58;
 }
 if ((label|0) == 58) {
  $128 = ($129|0)==(0|0);
  if ($128) {
   $driver$4$ph = $driver$247;$memory$1$ph = $memory$045;
   label = 88;
  } else {
   $130 = ((($129)) + 28|0);
   $131 = HEAP32[$130>>2]|0;
   _FT_Stream_Close($129);
   if ($10) {
    $driver$4$ph = $driver$247;$memory$1$ph = $memory$045;
    label = 88;
   } else {
    _ft_mem_free($131,$129);
    $driver$4$ph = $driver$247;$memory$1$ph = $memory$045;
    label = 88;
   }
  }
 }
 if ((label|0) == 88) {
  $212 = HEAP32[$face>>2]|0;
  $232 = $212;$driver$422 = $driver$4$ph;$memory$121 = $memory$1$ph;
  label = 95;
 }
 if ((label|0) == 95) {
  $231 = ($232|0)==(0|0);
  if (!($231)) {
   _destroy_face($memory$121,$232,$driver$422);
  }
 }
 $233 = HEAP32[$error>>2]|0;
 $$4 = $233;
 STACKTOP = sp;return ($$4|0);
}
function _open_face($driver,$astream,$external_stream,$face_index,$num_params,$params,$aface) {
 $driver = $driver|0;
 $astream = $astream|0;
 $external_stream = $external_stream|0;
 $face_index = $face_index|0;
 $num_params = $num_params|0;
 $params = $params|0;
 $aface = $aface|0;
 var $$lcssa = 0, $$lcssa22 = 0, $$pre = 0, $$pre14 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cur$1$i = 0, $error = 0, $i$010 = 0, $internal$056 = 0, $internal$057 = 0, $n$02$i = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($driver)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($driver)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 36|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (_ft_mem_alloc($3,$5,$error)|0);
 $7 = HEAP32[$error>>2]|0;
 $8 = ($7|0)==(0);
 do {
  if ($8) {
   $9 = ((($6)) + 96|0);
   HEAP32[$9>>2] = $driver;
   $10 = ((($6)) + 100|0);
   HEAP32[$10>>2] = $3;
   $11 = HEAP32[$astream>>2]|0;
   $12 = ((($6)) + 104|0);
   HEAP32[$12>>2] = $11;
   $13 = ($external_stream<<24>>24)==(0);
   if (!($13)) {
    $14 = ((($6)) + 8|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = $15 | 1024;
    HEAP32[$14>>2] = $16;
   }
   $17 = (_ft_mem_alloc($3,60,$error)|0);
   $18 = HEAP32[$error>>2]|0;
   $19 = ($18|0)==(0);
   if ($19) {
    $20 = ((($6)) + 128|0);
    HEAP32[$20>>2] = $17;
    $21 = ((($17)) + 48|0);
    HEAP32[$21>>2] = 0;
    $22 = ($num_params|0)>(0);
    L7: do {
     if ($22) {
      $23 = HEAP32[$20>>2]|0;
      $24 = ((($23)) + 48|0);
      $$pre = HEAP32[$24>>2]|0;
      $26 = $$pre;$i$010 = 0;
      while(1) {
       $25 = ($26|0)==(0|0);
       if (!($25)) {
        break L7;
       }
       $27 = (($params) + ($i$010<<3)|0);
       $28 = HEAP32[$27>>2]|0;
       $29 = ($28|0)==(1768842098);
       if ($29) {
        $30 = (((($params) + ($i$010<<3)|0)) + 4|0);
        $31 = HEAP32[$30>>2]|0;
        HEAP32[$24>>2] = $31;
        $32 = $31;
        $99 = $32;
       } else {
        $99 = 0;
       }
       $33 = (($i$010) + 1)|0;
       $34 = ($33|0)<($num_params|0);
       if ($34) {
        $26 = $99;$i$010 = $33;
       } else {
        break;
       }
      }
     }
    } while(0);
    $35 = ((($1)) + 48|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($36|0)==(0|0);
    if ($37) {
     $38 = HEAP32[$12>>2]|0;
     HEAP32[$astream>>2] = $38;
    } else {
     $39 = HEAP32[$astream>>2]|0;
     $40 = (FUNCTION_TABLE_iiiiii[$36 & 8191]($39,$6,$face_index,$num_params,$params)|0);
     HEAP32[$error>>2] = $40;
     $phitmp = ($40|0)==(0);
     $41 = HEAP32[$12>>2]|0;
     HEAP32[$astream>>2] = $41;
     if (!($phitmp)) {
      $internal$056 = $17;
      label = 27;
      break;
     }
    }
    $42 = ((($6)) + 40|0);
    $43 = HEAP32[$42>>2]|0;
    $44 = ($43|0)==(0|0);
    L20: do {
     if (!($44)) {
      $45 = ((($6)) + 36|0);
      $46 = HEAP32[$45>>2]|0;
      $47 = (($43) + ($46<<2)|0);
      $48 = ((($47)) + -4|0);
      $49 = ($48>>>0)<($43>>>0);
      L22: do {
       if ($49) {
        $cur$1$i = $47;
       } else {
        $51 = $48;
        L23: while(1) {
         $50 = HEAP32[$51>>2]|0;
         $52 = ((($50)) + 4|0);
         $53 = HEAP32[$52>>2]|0;
         $54 = ($53|0)==(1970170211);
         L25: do {
          if ($54) {
           $55 = ((($50)) + 8|0);
           $56 = HEAP32[$55>>2]|0;
           $57 = $56&65535;
           $58 = $56 >>> 16;
           $59 = $58&65535;
           switch ($57<<16>>16) {
           case 3:  {
            $60 = ($59<<16>>16)==(10);
            if ($60) {
             $$lcssa22 = $50;
             break L23;
            } else {
             break L25;
            }
            break;
           }
           case 0:  {
            $63 = ($59<<16>>16)==(4);
            if ($63) {
             $$lcssa22 = $50;
             break L23;
            } else {
             break L25;
            }
            break;
           }
           default: {
            break L25;
           }
           }
          }
         } while(0);
         $61 = ((($51)) + -4|0);
         $62 = ($61>>>0)<($43>>>0);
         if ($62) {
          $cur$1$i = $47;
          break L22;
         } else {
          $51 = $61;
         }
        }
        $64 = $$lcssa22;
        $65 = ((($6)) + 92|0);
        HEAP32[$65>>2] = $64;
        break L20;
       }
      } while(0);
      while(1) {
       $66 = ((($cur$1$i)) + -4|0);
       $67 = ($66>>>0)<($43>>>0);
       if ($67) {
        break L20;
       }
       $68 = HEAP32[$66>>2]|0;
       $69 = ((($68)) + 4|0);
       $70 = HEAP32[$69>>2]|0;
       $71 = ($70|0)==(1970170211);
       if ($71) {
        $$lcssa = $68;
        break;
       } else {
        $cur$1$i = $66;
       }
      }
      $72 = ((($6)) + 92|0);
      HEAP32[$72>>2] = $$lcssa;
     }
    } while(0);
    HEAP32[$aface>>2] = $6;
    $98 = 0;
    STACKTOP = sp;return ($98|0);
   } else {
    $internal$056 = $17;
    label = 27;
   }
  } else {
   $73 = ($6|0)==(0|0);
   if ($73) {
    $internal$057 = 0;
   } else {
    $internal$056 = 0;
    label = 27;
   }
  }
 } while(0);
 if ((label|0) == 27) {
  $74 = ((($6)) + 36|0);
  $75 = HEAP32[$74>>2]|0;
  $76 = ($75|0)>(0);
  $77 = ((($6)) + 40|0);
  if ($76) {
   $n$02$i = 0;
   while(1) {
    $78 = HEAP32[$77>>2]|0;
    $79 = (($78) + ($n$02$i<<2)|0);
    $80 = HEAP32[$79>>2]|0;
    $81 = ((($80)) + 12|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = HEAP32[$80>>2]|0;
    $84 = ((($83)) + 100|0);
    $85 = HEAP32[$84>>2]|0;
    $86 = ((($82)) + 8|0);
    $87 = HEAP32[$86>>2]|0;
    $88 = ($87|0)==(0|0);
    if (!($88)) {
     FUNCTION_TABLE_vi[$87 & 8191]($80);
    }
    _ft_mem_free($85,$80);
    $89 = HEAP32[$77>>2]|0;
    $90 = (($89) + ($n$02$i<<2)|0);
    HEAP32[$90>>2] = 0;
    $91 = (($n$02$i) + 1)|0;
    $92 = HEAP32[$74>>2]|0;
    $93 = ($91|0)<($92|0);
    if ($93) {
     $n$02$i = $91;
    } else {
     break;
    }
   }
  }
  $94 = HEAP32[$77>>2]|0;
  _ft_mem_free($3,$94);
  HEAP32[$77>>2] = 0;
  HEAP32[$74>>2] = 0;
  $internal$057 = $internal$056;
 }
 $95 = ((($1)) + 52|0);
 $96 = HEAP32[$95>>2]|0;
 $97 = ($96|0)==(0|0);
 if (!($97)) {
  FUNCTION_TABLE_vi[$96 & 8191]($6);
 }
 _ft_mem_free($3,$internal$057);
 _ft_mem_free($3,$6);
 HEAP32[$aface>>2] = 0;
 $$pre14 = HEAP32[$error>>2]|0;
 $98 = $$pre14;
 STACKTOP = sp;return ($98|0);
}
function _open_face_PS_from_sfnt_stream($library,$stream,$face_index,$num_params,$params,$aface) {
 $library = $library|0;
 $stream = $stream|0;
 $face_index = $face_index|0;
 $num_params = $num_params|0;
 $params = $params|0;
 $aface = $aface|0;
 var $$0 = 0, $$0$i$ph = 0, $$lcssa = 0, $$lcssa29 = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$pre = 0, $$pre$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $error$i = 0, $i$02$i$us = 0, $i$02$us$i = 0;
 var $is_sfnt_cid$0$us = 0, $is_sfnt_cid$1$us = 0, $is_sfnt_cid$2 = 0, $length$2$us = 0, $length$3 = 0, $offset$2$us = 0, $offset$3 = 0, $pstable_index$01$i$us = 0, $pstable_index$1$i$us = 0, $switch$split12D = 0, $switch$split2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp + 4|0;
 $error = sp;
 $0 = HEAP32[$library>>2]|0;
 $1 = (_FT_Stream_Pos($stream)|0);
 $2 = (_FT_Stream_ReadULong($stream,$error$i)|0);
 $3 = HEAP32[$error$i>>2]|0;
 $4 = ($3|0)==(0);
 L1: do {
  if ($4) {
   $5 = ($2|0)==(1954115633);
   if ($5) {
    $6 = (_FT_Stream_ReadUShort($stream,$error$i)|0);
    $7 = HEAP32[$error$i>>2]|0;
    $8 = ($7|0)==(0);
    if ($8) {
     $9 = (_FT_Stream_Skip($stream,6)|0);
     HEAP32[$error$i>>2] = $9;
     $10 = ($9|0)==(0);
     if ($10) {
      $11 = $6&65535;
      $12 = ($6<<16>>16)==(0);
      if ($12) {
       $$0$i$ph = 142;
       label = 24;
      } else {
       $13 = ($face_index|0)<(0);
       L7: do {
        if ($13) {
         $i$02$us$i = 0;
         L8: while(1) {
          $34 = (_FT_Stream_ReadULong($stream,$error$i)|0);
          $35 = HEAP32[$error$i>>2]|0;
          $36 = ($35|0)==(0);
          if (!($36)) {
           $$0$i$ph = $35;
           label = 24;
           break L1;
          }
          $39 = (_FT_Stream_Skip($stream,4)|0);
          HEAP32[$error$i>>2] = $39;
          $40 = ($39|0)==(0);
          if (!($40)) {
           $$0$i$ph = $39;
           label = 24;
           break L1;
          }
          $41 = (_FT_Stream_ReadULong($stream,$error$i)|0);
          $42 = HEAP32[$error$i>>2]|0;
          $43 = ($42|0)==(0);
          if (!($43)) {
           $$0$i$ph = $42;
           label = 24;
           break L1;
          }
          $44 = (_FT_Stream_ReadULong($stream,$error$i)|0);
          $45 = HEAP32[$error$i>>2]|0;
          $46 = ($45|0)==(0);
          if (!($46)) {
           $$0$i$ph = $45;
           label = 24;
           break L1;
          }
          $switch$split12D = ($34|0)<(1415139377);
          if ($switch$split12D) {
           switch ($34|0) {
           case 1128875040:  {
            $$lcssa = $41;$$lcssa30 = $44;
            label = 23;
            break L8;
            break;
           }
           default: {
           }
           }
          } else {
           switch ($34|0) {
           case 1415139377:  {
            $$lcssa29 = $41;$$lcssa31 = $44;
            label = 22;
            break L8;
            break;
           }
           default: {
           }
           }
          }
          $37 = (($i$02$us$i) + 1)|0;
          $38 = ($37|0)<($11|0);
          if ($38) {
           $i$02$us$i = $37;
          } else {
           $$0$i$ph = 142;
           label = 24;
           break L1;
          }
         }
         if ((label|0) == 22) {
          $47 = (($$lcssa29) + 24)|0;
          $48 = (($$lcssa31) + -24)|0;
          $is_sfnt_cid$2 = 0;$length$3 = $48;$offset$3 = $47;
          break;
         }
         else if ((label|0) == 23) {
          $49 = (($$lcssa) + 22)|0;
          $50 = (($$lcssa30) + -22)|0;
          $is_sfnt_cid$2 = 1;$length$3 = $50;$offset$3 = $49;
          break;
         }
        } else {
         $i$02$i$us = 0;$is_sfnt_cid$0$us = 0;$pstable_index$01$i$us = -1;
         while(1) {
          $14 = (_FT_Stream_ReadULong($stream,$error$i)|0);
          $15 = HEAP32[$error$i>>2]|0;
          $16 = ($15|0)==(0);
          if (!($16)) {
           $$0$i$ph = $15;
           label = 24;
           break L1;
          }
          $17 = (_FT_Stream_Skip($stream,4)|0);
          HEAP32[$error$i>>2] = $17;
          $18 = ($17|0)==(0);
          if (!($18)) {
           $$0$i$ph = $17;
           label = 24;
           break L1;
          }
          $19 = (_FT_Stream_ReadULong($stream,$error$i)|0);
          $20 = HEAP32[$error$i>>2]|0;
          $21 = ($20|0)==(0);
          if (!($21)) {
           $$0$i$ph = $20;
           label = 24;
           break L1;
          }
          $22 = (_FT_Stream_ReadULong($stream,$error$i)|0);
          $23 = HEAP32[$error$i>>2]|0;
          $24 = ($23|0)==(0);
          if (!($24)) {
           $$0$i$ph = $23;
           label = 24;
           break L1;
          }
          $switch$split2D = ($14|0)<(1415139377);
          L27: do {
           if ($switch$split2D) {
            switch ($14|0) {
            case 1128875040:  {
             break;
            }
            default: {
             $is_sfnt_cid$1$us = $is_sfnt_cid$0$us;$length$2$us = $22;$offset$2$us = $19;$pstable_index$1$i$us = $pstable_index$01$i$us;
             break L27;
            }
            }
            $28 = (($pstable_index$01$i$us) + 1)|0;
            $29 = (($19) + 22)|0;
            $30 = (($22) + -22)|0;
            $is_sfnt_cid$1$us = 1;$length$2$us = $30;$offset$2$us = $29;$pstable_index$1$i$us = $28;
           } else {
            switch ($14|0) {
            case 1415139377:  {
             break;
            }
            default: {
             $is_sfnt_cid$1$us = $is_sfnt_cid$0$us;$length$2$us = $22;$offset$2$us = $19;$pstable_index$1$i$us = $pstable_index$01$i$us;
             break L27;
            }
            }
            $25 = (($pstable_index$01$i$us) + 1)|0;
            $26 = (($19) + 24)|0;
            $27 = (($22) + -24)|0;
            $is_sfnt_cid$1$us = 0;$length$2$us = $27;$offset$2$us = $26;$pstable_index$1$i$us = $25;
           }
          } while(0);
          $31 = ($pstable_index$1$i$us|0)==($face_index|0);
          $32 = (($i$02$i$us) + 1)|0;
          if ($31) {
           $is_sfnt_cid$2 = $is_sfnt_cid$1$us;$length$3 = $length$2$us;$offset$3 = $offset$2$us;
           break L7;
          }
          $33 = ($32|0)<($11|0);
          if ($33) {
           $i$02$i$us = $32;$is_sfnt_cid$0$us = $is_sfnt_cid$1$us;$pstable_index$01$i$us = $pstable_index$1$i$us;
          } else {
           $$0$i$ph = 142;
           label = 24;
           break L1;
          }
         }
        }
       } while(0);
       HEAP32[$error>>2] = 0;
       $51 = (($offset$3) + ($1))|0;
       $52 = (_FT_Stream_Seek($stream,$51)|0);
       $53 = ($52|0)==(0);
       if (!($53)) {
        $$0 = 0;
        STACKTOP = sp;return ($$0|0);
       }
       $54 = (_ft_mem_alloc($0,$length$3,$error)|0);
       $55 = HEAP32[$error>>2]|0;
       $56 = ($55|0)==(0);
       if ($56) {
        $57 = (_FT_Stream_Read($stream,$54,$length$3)|0);
        HEAP32[$error>>2] = $57;
        $58 = ($57|0)==(0);
        if ($58) {
         $59 = $13 ? $face_index : 0;
         $60 = ($is_sfnt_cid$2<<24>>24)!=(0);
         $61 = $60 ? 530514 : 593865;
         $62 = (_open_face_from_buffer($library,$54,$length$3,$59,$61,$aface)|0);
         HEAP32[$error>>2] = $62;
         $64 = $62;
        } else {
         $64 = $57;
        }
       } else {
        $64 = $55;
       }
      }
     } else {
      $$0$i$ph = $9;
      label = 24;
     }
    } else {
     $$0$i$ph = $7;
     label = 24;
    }
   } else {
    $$0$i$ph = 2;
    label = 24;
   }
  } else {
   $$0$i$ph = $3;
   label = 24;
  }
 } while(0);
 if ((label|0) == 24) {
  HEAP32[$error>>2] = $$0$i$ph;
  $64 = $$0$i$ph;
 }
 $63 = $64 & 255;
 $65 = ($63|0)==(2);
 if (!($65)) {
  $$0 = $64;
  STACKTOP = sp;return ($$0|0);
 }
 $66 = (_FT_Stream_Seek($stream,$1)|0);
 $67 = ($66|0)==(0);
 $$pre = HEAP32[$error>>2]|0;
 $$pre$ = $67 ? $$pre : $66;
 $$0 = $$pre$;
 STACKTOP = sp;return ($$0|0);
}
function _open_face_from_buffer($library,$base,$size,$face_index,$driver_name,$aface) {
 $library = $library|0;
 $base = $base|0;
 $size = $size|0;
 $face_index = $face_index|0;
 $driver_name = $driver_name|0;
 $aface = $aface|0;
 var $$0 = 0, $$0$i$ph = 0, $$0$i1 = 0, $$0$i2 = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $args = 0, $cur$02$i = 0, $error$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp + 32|0;
 $args = sp;
 $0 = HEAP32[$library>>2]|0;
 $1 = ($library|0)==(0|0);
 if ($1) {
  $$0$i$ph = 33;
  label = 4;
 } else {
  $2 = ($base|0)==(0|0);
  if ($2) {
   $$0$i$ph = 6;
   label = 4;
  } else {
   $3 = (_ft_mem_alloc($0,40,$error$i)|0);
   $4 = HEAP32[$error$i>>2]|0;
   $5 = ($4|0)==(0);
   if ($5) {
    _FT_Stream_OpenMemory($3,$base,$size);
    $6 = ((($3)) + 24|0);
    HEAP32[$6>>2] = 4343;
    $$pre$i = HEAP32[$error$i>>2]|0;
    $7 = ($$pre$i|0)==(0);
    if ($7) {
     $8 = $3;
     HEAP32[$args>>2] = 2;
     $9 = ((($args)) + 16|0);
     HEAP32[$9>>2] = $8;
     $10 = ($driver_name|0)==(0|0);
     if (!($10)) {
      HEAP32[$args>>2] = 10;
      $11 = ((($library)) + 16|0);
      $12 = HEAP32[$11>>2]|0;
      $13 = (((($library)) + 20|0) + ($12<<2)|0);
      $14 = ($12|0)>(0);
      L8: do {
       if ($14) {
        $15 = ((($library)) + 20|0);
        $cur$02$i = $15;
        while(1) {
         $18 = HEAP32[$cur$02$i>>2]|0;
         $19 = HEAP32[$18>>2]|0;
         $20 = ((($19)) + 8|0);
         $21 = HEAP32[$20>>2]|0;
         $22 = (_strcmp($21,$driver_name)|0);
         $23 = ($22|0)==(0);
         $17 = ((($cur$02$i)) + 4|0);
         if ($23) {
          $$0$i1 = $18;
          break L8;
         }
         $16 = ($17>>>0)<($13>>>0);
         if ($16) {
          $cur$02$i = $17;
         } else {
          $$0$i1 = 0;
          break;
         }
        }
       } else {
        $$0$i1 = 0;
       }
      } while(0);
      $24 = ((($args)) + 20|0);
      HEAP32[$24>>2] = $$0$i1;
     }
     $25 = (_FT_Open_Face($library,$args,$face_index,$aface)|0);
     $26 = ($25|0)==(0);
     if ($26) {
      $27 = HEAP32[$aface>>2]|0;
      $28 = ((($27)) + 8|0);
      $29 = HEAP32[$28>>2]|0;
      $30 = $29 & -1025;
      HEAP32[$28>>2] = $30;
      $$0 = 0;
      STACKTOP = sp;return ($$0|0);
     } else {
      _FT_Stream_Close($3);
      _ft_mem_free($0,$3);
      $$0 = $25;
      STACKTOP = sp;return ($$0|0);
     }
    } else {
     $$0$i2 = $$pre$i;
    }
   } else {
    $$0$i$ph = $4;
    label = 4;
   }
  }
 }
 if ((label|0) == 4) {
  $$0$i2 = $$0$i$ph;
 }
 _ft_mem_free($0,$base);
 $$0 = $$0$i2;
 STACKTOP = sp;return ($$0|0);
}
function _memory_stream_close($stream) {
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$stream>>2]|0;
 _ft_mem_free($1,$2);
 $3 = ((($stream)) + 4|0);
 HEAP32[$3>>2] = 0;
 HEAP32[$stream>>2] = 0;
 $4 = ((($stream)) + 24|0);
 HEAP32[$4>>2] = 0;
 return;
}
function _IsMacBinary($library,$stream,$face_index,$aface) {
 $library = $library|0;
 $stream = $stream|0;
 $face_index = $face_index|0;
 $aface = $aface|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $header = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $header = sp;
 $0 = ($stream|0)==(0|0);
 if ($0) {
  $$0 = 85;
 } else {
  $1 = (_FT_Stream_Seek($stream,0)|0);
  $2 = ($1|0)==(0);
  if ($2) {
   $3 = (_FT_Stream_Read($stream,$header,128)|0);
   $4 = ($3|0)==(0);
   if ($4) {
    $5 = HEAP8[$header>>0]|0;
    $6 = ($5<<24>>24)==(0);
    if ($6) {
     $7 = ((($header)) + 74|0);
     $8 = HEAP8[$7>>0]|0;
     $9 = ($8<<24>>24)==(0);
     if ($9) {
      $10 = ((($header)) + 82|0);
      $11 = HEAP8[$10>>0]|0;
      $12 = ($11<<24>>24)==(0);
      if ($12) {
       $13 = ((($header)) + 1|0);
       $14 = HEAP8[$13>>0]|0;
       $15 = $14&255;
       $16 = (($14) + -1)<<24>>24;
       $17 = ($16&255)>(32);
       if ($17) {
        $$0 = 2;
       } else {
        $18 = ((($header)) + 63|0);
        $19 = HEAP8[$18>>0]|0;
        $20 = ($19<<24>>24)==(0);
        if ($20) {
         $21 = (($15) + 2)|0;
         $22 = (($header) + ($21)|0);
         $23 = HEAP8[$22>>0]|0;
         $24 = ($23<<24>>24)==(0);
         if ($24) {
          $25 = ((($header)) + 83|0);
          $26 = HEAP8[$25>>0]|0;
          $27 = $26&255;
          $28 = $27 << 24;
          $29 = ((($header)) + 84|0);
          $30 = HEAP8[$29>>0]|0;
          $31 = $30&255;
          $32 = $31 << 16;
          $33 = $32 | $28;
          $34 = ((($header)) + 85|0);
          $35 = HEAP8[$34>>0]|0;
          $36 = $35&255;
          $37 = $36 << 8;
          $38 = $33 | $37;
          $39 = ((($header)) + 86|0);
          $40 = HEAP8[$39>>0]|0;
          $41 = $40&255;
          $42 = $38 | $41;
          $43 = (($42) + 255)|0;
          $44 = $43 & -128;
          $45 = (_IsMacResource($library,$stream,$44,$face_index,$aface)|0);
          $$0 = $45;
         } else {
          $$0 = 2;
         }
        } else {
         $$0 = 2;
        }
       }
      } else {
       $$0 = 2;
      }
     } else {
      $$0 = 2;
     }
    } else {
     $$0 = 2;
    }
   } else {
    $$0 = $3;
   }
  } else {
   $$0 = $1;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _IsMacResource($library,$stream,$resource_offset,$face_index,$aface) {
 $library = $library|0;
 $stream = $stream|0;
 $resource_offset = $resource_offset|0;
 $face_index = $face_index|0;
 $aface = $aface|0;
 var $$0 = 0, $$0$i = 0, $$0$i3 = 0, $$face_index$i = 0, $$lcssa = 0, $$pr = 0, $$pre = 0, $$pre33 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $count = 0, $data_offsets = 0, $error$i = 0;
 var $error$i1 = 0, $i$0$i27 = 0, $i$1$i16 = 0, $len$0$i$lcssa = 0, $len$0$i18 = 0, $len$1$i = 0, $len$2$i = 0, $map_offset = 0, $pfb_len$0$i28 = 0, $pfb_lenpos$0$i$lcssa = 0, $pfb_lenpos$0$i20 = 0, $pfb_lenpos$1$i = 0, $pfb_lenpos$2$i = 0, $pfb_pos$0$i$lcssa = 0, $pfb_pos$0$i19 = 0, $pfb_pos$1$i = 0, $pfb_pos$2$i = 0, $phitmp$i = 0, $rdara_pos = 0, $rlen$0$i = 0;
 var $type$0$i17 = 0, $type$1$i = 0, $type$2$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i1 = sp + 20|0;
 $error$i = sp + 16|0;
 $map_offset = sp + 12|0;
 $rdara_pos = sp + 8|0;
 $data_offsets = sp + 4|0;
 $count = sp;
 $0 = HEAP32[$library>>2]|0;
 $1 = (_FT_Raccess_Get_HeaderInfo($library,$stream,$resource_offset,$map_offset,$rdara_pos)|0);
 $2 = ($1|0)==(0);
 if (!($2)) {
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = HEAP32[$map_offset>>2]|0;
 $4 = HEAP32[$rdara_pos>>2]|0;
 $5 = (_FT_Raccess_Get_DataOffsets($library,$stream,$3,$4,1347375956,1,$data_offsets,$count)|0);
 $6 = ($5|0)==(0);
 if (!($6)) {
  $112 = HEAP32[$map_offset>>2]|0;
  $113 = HEAP32[$rdara_pos>>2]|0;
  $114 = (_FT_Raccess_Get_DataOffsets($library,$stream,$112,$113,1936092788,0,$data_offsets,$count)|0);
  $115 = ($114|0)==(0);
  if (!($115)) {
   $$0 = $114;
   STACKTOP = sp;return ($$0|0);
  }
  $116 = HEAP32[$count>>2]|0;
  $117 = (($face_index|0) % ($116|0))&-1;
  $118 = HEAP32[$data_offsets>>2]|0;
  $119 = HEAP32[$library>>2]|0;
  $120 = ($117|0)==(-1);
  $$face_index$i = $120 ? 0 : $117;
  $121 = ($$face_index$i|0)<($116|0);
  do {
   if ($121) {
    $122 = (($118) + ($$face_index$i<<2)|0);
    $123 = HEAP32[$122>>2]|0;
    $124 = (_FT_Stream_Seek($stream,$123)|0);
    HEAP32[$error$i>>2] = $124;
    $125 = ($124|0)==(0);
    if ($125) {
     $126 = (_FT_Stream_ReadULong($stream,$error$i)|0);
     $127 = HEAP32[$error$i>>2]|0;
     $128 = ($127|0)==(0);
     if ($128) {
      $129 = ($126|0)==(-1);
      if ($129) {
       $$0$i = 1;
      } else {
       $130 = (_open_face_PS_from_sfnt_stream($library,$stream,$$face_index$i,0,0,$aface)|0);
       HEAP32[$error$i>>2] = $130;
       $131 = ($130|0)==(0);
       if ($131) {
        $$0$i = 0;
       } else {
        $132 = (($123) + 4)|0;
        $133 = (_FT_Stream_Seek($stream,$132)|0);
        $134 = ($133|0)==(0);
        if (!($134)) {
         $$pre = HEAP32[$error$i>>2]|0;
         $$0$i = $$pre;
         break;
        }
        $135 = (_ft_mem_alloc($119,$126,$error$i)|0);
        $136 = HEAP32[$error$i>>2]|0;
        $137 = ($136|0)==(0);
        if ($137) {
         $138 = (_FT_Stream_Read($stream,$135,$126)|0);
         HEAP32[$error$i>>2] = $138;
         $139 = ($138|0)==(0);
         if ($139) {
          $140 = ($126|0)>(4);
          if ($140) {
           $141 = (_memcmp($135,530518,4)|0);
           $142 = ($141|0)!=(0);
           $phitmp$i = $142 ? 594296 : 532012;
           $143 = $phitmp$i;
          } else {
           $143 = 594296;
          }
          $144 = (_open_face_from_buffer($library,$135,$126,0,$143,$aface)|0);
          HEAP32[$error$i>>2] = $144;
          $$0$i = $144;
         } else {
          $$0$i = $138;
         }
        } else {
         $$0$i = $136;
        }
       }
      }
     } else {
      $$0$i = $127;
     }
    } else {
     $$0$i = $124;
    }
   } else {
    $$0$i = 1;
   }
  } while(0);
  $145 = HEAP32[$data_offsets>>2]|0;
  _ft_mem_free($0,$145);
  HEAP32[$data_offsets>>2] = 0;
  $146 = ($$0$i|0)==(0);
  if (!($146)) {
   $$0 = $$0$i;
   STACKTOP = sp;return ($$0|0);
  }
  $147 = HEAP32[$count>>2]|0;
  $148 = HEAP32[$aface>>2]|0;
  HEAP32[$148>>2] = $147;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $7 = HEAP32[$data_offsets>>2]|0;
 $8 = HEAP32[$count>>2]|0;
 HEAP32[$error$i1>>2] = 1;
 $9 = HEAP32[$library>>2]|0;
 L29: do {
  switch ($face_index|0) {
  case 0: case -1:  {
   $10 = ($8|0)>(0);
   if ($10) {
    $i$0$i27 = 0;$pfb_len$0$i28 = 0;
    while(1) {
     $11 = (($7) + ($i$0$i27<<2)|0);
     $12 = HEAP32[$11>>2]|0;
     $13 = (_FT_Stream_Seek($stream,$12)|0);
     HEAP32[$error$i1>>2] = $13;
     $14 = ($13|0)==(0);
     if (!($14)) {
      $$0$i3 = $13;
      break L29;
     }
     $15 = (_FT_Stream_ReadULong($stream,$error$i1)|0);
     $16 = HEAP32[$error$i1>>2]|0;
     $17 = ($16|0)==(0);
     if (!($17)) {
      $$0$i3 = $16;
      break L29;
     }
     $18 = ($15|0)<(0);
     if ($18) {
      label = 9;
      break;
     }
     $19 = (($pfb_len$0$i28) + 6)|0;
     $20 = (($19) + ($15))|0;
     $21 = ($20>>>0)<($pfb_len$0$i28>>>0);
     if ($21) {
      label = 9;
      break;
     }
     $22 = (($i$0$i27) + 1)|0;
     $23 = ($22|0)<($8|0);
     if ($23) {
      $i$0$i27 = $22;$pfb_len$0$i28 = $20;
     } else {
      $$lcssa = $20;
      break;
     }
    }
    if ((label|0) == 9) {
     HEAP32[$error$i1>>2] = 9;
     $$0$i3 = 9;
     break L29;
    }
    $24 = (($$lcssa) + 2)|0;
    $25 = ($24>>>0)<(6);
    if (!($25)) {
     $26 = (_ft_mem_alloc($9,$24,$error$i1)|0);
     $27 = HEAP32[$error$i1>>2]|0;
     $28 = ($27|0)==(0);
     if (!($28)) {
      $$0$i3 = $27;
      break L29;
     }
     HEAP8[$26>>0] = -128;
     $29 = ((($26)) + 1|0);
     HEAP8[$29>>0] = 1;
     $30 = ((($26)) + 2|0);
     HEAP8[$30>>0]=0&255;HEAP8[$30+1>>0]=(0>>8)&255;HEAP8[$30+2>>0]=(0>>16)&255;HEAP8[$30+3>>0]=0>>24;
     L43: do {
      if ($10) {
       $i$1$i16 = 0;$len$0$i18 = 0;$pfb_lenpos$0$i20 = 2;$pfb_pos$0$i19 = 6;$type$0$i17 = 1;
       while(1) {
        $31 = (($7) + ($i$1$i16<<2)|0);
        $32 = HEAP32[$31>>2]|0;
        $33 = (_FT_Stream_Seek($stream,$32)|0);
        HEAP32[$error$i1>>2] = $33;
        $34 = ($33|0)==(0);
        if (!($34)) {
         label = 35;
         break L43;
        }
        $35 = (_FT_Stream_ReadULong($stream,$error$i1)|0);
        $36 = HEAP32[$error$i1>>2]|0;
        $37 = ($36|0)==(0);
        if (!($37)) {
         label = 35;
         break L43;
        }
        $38 = ($35|0)<(0);
        if ($38) {
         break;
        }
        $39 = (_FT_Stream_ReadUShort($stream,$error$i1)|0);
        $40 = HEAP32[$error$i1>>2]|0;
        $41 = ($40|0)==(0);
        if (!($41)) {
         label = 35;
         break L43;
        }
        $42 = $39&65535;
        HEAP32[$error$i1>>2] = 10;
        $43 = $42 >>> 8;
        $44 = ($43|0)==(0);
        if ($44) {
         $len$2$i = $len$0$i18;$pfb_lenpos$2$i = $pfb_lenpos$0$i20;$pfb_pos$2$i = $pfb_pos$0$i19;$type$2$i = $type$0$i17;
        } else {
         $45 = ($35>>>0)>(2);
         $46 = (($35) + -2)|0;
         $rlen$0$i = $45 ? $46 : 0;
         $47 = ($43|0)==($type$0$i17|0);
         if ($47) {
          $48 = (($rlen$0$i) + ($len$0$i18))|0;
          $len$1$i = $48;$pfb_lenpos$1$i = $pfb_lenpos$0$i20;$pfb_pos$1$i = $pfb_pos$0$i19;$type$1$i = $type$0$i17;
         } else {
          $49 = (($pfb_lenpos$0$i20) + 3)|0;
          $50 = ($49>>>0)>($24>>>0);
          if ($50) {
           label = 34;
           break L43;
          }
          $51 = $len$0$i18&255;
          $52 = (($26) + ($pfb_lenpos$0$i20)|0);
          HEAP8[$52>>0] = $51;
          $53 = $len$0$i18 >>> 8;
          $54 = $53&255;
          $55 = (($pfb_lenpos$0$i20) + 1)|0;
          $56 = (($26) + ($55)|0);
          HEAP8[$56>>0] = $54;
          $57 = $len$0$i18 >>> 16;
          $58 = $57&255;
          $59 = (($pfb_lenpos$0$i20) + 2)|0;
          $60 = (($26) + ($59)|0);
          HEAP8[$60>>0] = $58;
          $61 = $len$0$i18 >>> 24;
          $62 = $61&255;
          $63 = (($26) + ($49)|0);
          HEAP8[$63>>0] = $62;
          $64 = ($43|0)==(5);
          if ($64) {
           $len$0$i$lcssa = $len$0$i18;$pfb_lenpos$0$i$lcssa = $pfb_lenpos$0$i20;$pfb_pos$0$i$lcssa = $pfb_pos$0$i19;
           label = 31;
           break L43;
          }
          $65 = (($pfb_pos$0$i19) + 6)|0;
          $66 = ($65>>>0)>($24>>>0);
          if ($66) {
           label = 34;
           break L43;
          }
          $67 = (($pfb_pos$0$i19) + 1)|0;
          $68 = (($26) + ($pfb_pos$0$i19)|0);
          HEAP8[$68>>0] = -128;
          $69 = $43&255;
          $70 = (($pfb_pos$0$i19) + 2)|0;
          $71 = (($26) + ($67)|0);
          HEAP8[$71>>0] = $69;
          $72 = (($pfb_pos$0$i19) + 3)|0;
          $73 = (($26) + ($70)|0);
          HEAP8[$73>>0] = 0;
          $74 = (($pfb_pos$0$i19) + 4)|0;
          $75 = (($26) + ($72)|0);
          HEAP8[$75>>0] = 0;
          $76 = (($pfb_pos$0$i19) + 5)|0;
          $77 = (($26) + ($74)|0);
          HEAP8[$77>>0] = 0;
          $78 = (($26) + ($76)|0);
          HEAP8[$78>>0] = 0;
          $len$1$i = $rlen$0$i;$pfb_lenpos$1$i = $70;$pfb_pos$1$i = $65;$type$1$i = $43;
         }
         $79 = ($pfb_pos$1$i>>>0)>($$lcssa>>>0);
         if ($79) {
          label = 34;
          break L43;
         }
         $80 = (($pfb_pos$1$i) + ($rlen$0$i))|0;
         $81 = ($80>>>0)>($$lcssa>>>0);
         if ($81) {
          label = 34;
          break L43;
         }
         $82 = (($26) + ($pfb_pos$1$i)|0);
         $83 = (_FT_Stream_Read($stream,$82,$rlen$0$i)|0);
         HEAP32[$error$i1>>2] = $83;
         $84 = ($83|0)==(0);
         if ($84) {
          $len$2$i = $len$1$i;$pfb_lenpos$2$i = $pfb_lenpos$1$i;$pfb_pos$2$i = $80;$type$2$i = $type$1$i;
         } else {
          label = 35;
          break L43;
         }
        }
        $85 = (($i$1$i16) + 1)|0;
        $86 = ($85|0)<($8|0);
        if ($86) {
         $i$1$i16 = $85;$len$0$i18 = $len$2$i;$pfb_lenpos$0$i20 = $pfb_lenpos$2$i;$pfb_pos$0$i19 = $pfb_pos$2$i;$type$0$i17 = $type$2$i;
        } else {
         $len$0$i$lcssa = $len$2$i;$pfb_lenpos$0$i$lcssa = $pfb_lenpos$2$i;$pfb_pos$0$i$lcssa = $pfb_pos$2$i;
         label = 31;
         break L43;
        }
       }
       HEAP32[$error$i1>>2] = 9;
       label = 35;
      } else {
       $len$0$i$lcssa = 0;$pfb_lenpos$0$i$lcssa = 2;$pfb_pos$0$i$lcssa = 6;
       label = 31;
      }
     } while(0);
     if ((label|0) == 31) {
      HEAP32[$error$i1>>2] = 10;
      $87 = (($pfb_pos$0$i$lcssa) + 2)|0;
      $88 = ($87>>>0)>($24>>>0);
      if ($88) {
       label = 35;
      } else {
       $89 = (($pfb_pos$0$i$lcssa) + 1)|0;
       $90 = (($26) + ($pfb_pos$0$i$lcssa)|0);
       HEAP8[$90>>0] = -128;
       $91 = (($26) + ($89)|0);
       HEAP8[$91>>0] = 3;
       $92 = (($pfb_lenpos$0$i$lcssa) + 3)|0;
       $93 = ($92>>>0)>($24>>>0);
       if ($93) {
        label = 34;
       } else {
        $94 = $len$0$i$lcssa&255;
        $95 = (($26) + ($pfb_lenpos$0$i$lcssa)|0);
        HEAP8[$95>>0] = $94;
        $96 = $len$0$i$lcssa >>> 8;
        $97 = $96&255;
        $98 = (($pfb_lenpos$0$i$lcssa) + 1)|0;
        $99 = (($26) + ($98)|0);
        HEAP8[$99>>0] = $97;
        $100 = $len$0$i$lcssa >>> 16;
        $101 = $100&255;
        $102 = (($pfb_lenpos$0$i$lcssa) + 2)|0;
        $103 = (($26) + ($102)|0);
        HEAP8[$103>>0] = $101;
        $104 = $len$0$i$lcssa >>> 24;
        $105 = $104&255;
        $106 = (($26) + ($92)|0);
        HEAP8[$106>>0] = $105;
        $107 = (_open_face_from_buffer($library,$26,$87,0,593865,$aface)|0);
        $$0$i3 = $107;
        break L29;
       }
      }
     }
     if ((label|0) == 34) {
      $$pr = HEAP32[$error$i1>>2]|0;
      $108 = ($$pr|0)==(0);
      if (!($108)) {
       label = 35;
      }
     }
     if ((label|0) == 35) {
      HEAP32[$error$i1>>2] = 1;
     }
     _ft_mem_free($9,$26);
     $$pre33 = HEAP32[$error$i1>>2]|0;
     $$0$i3 = $$pre33;
     break L29;
    }
   }
   HEAP32[$error$i1>>2] = 10;
   $$0$i3 = 10;
   break;
  }
  default: {
   $$0$i3 = 1;
  }
  }
 } while(0);
 $109 = HEAP32[$data_offsets>>2]|0;
 _ft_mem_free($0,$109);
 HEAP32[$data_offsets>>2] = 0;
 $110 = ($$0$i3|0)==(0);
 if (!($110)) {
  $$0 = $$0$i3;
  STACKTOP = sp;return ($$0|0);
 }
 $111 = HEAP32[$aface>>2]|0;
 HEAP32[$111>>2] = 1;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _FT_New_Size($face,$asize) {
 $face = $face|0;
 $asize = $asize|0;
 var $$0 = 0, $$pr2$pre = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $node$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ($face|0)==(0|0);
 if ($0) {
  $$0 = 35;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ($asize|0)==(0|0);
 if ($1) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ((($face)) + 96|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  $$0 = 34;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$asize>>2] = 0;
 $5 = HEAP32[$2>>2]|0;
 $6 = ((($5)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($face)) + 100|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($7)) + 40|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (_ft_mem_alloc($9,$11,$error)|0);
 $13 = HEAP32[$error>>2]|0;
 $14 = ($13|0)==(0);
 do {
  if ($14) {
   $15 = (_ft_mem_alloc($9,12,$error)|0);
   $16 = HEAP32[$error>>2]|0;
   $17 = ($16|0)==(0);
   if ($17) {
    HEAP32[$12>>2] = $face;
    $18 = ((($12)) + 40|0);
    HEAP32[$18>>2] = 0;
    $19 = ((($7)) + 56|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = ($20|0)==(0|0);
    if (!($21)) {
     $22 = (FUNCTION_TABLE_ii[$20 & 8191]($12)|0);
     HEAP32[$error>>2] = $22;
     $23 = ($22|0)==(0);
     if (!($23)) {
      $node$03 = $15;
      break;
     }
    }
    HEAP32[$asize>>2] = $12;
    $24 = ((($15)) + 8|0);
    HEAP32[$24>>2] = $12;
    $25 = ((($face)) + 108|0);
    _FT_List_Add($25,$15);
    $$pr2$pre = HEAP32[$error>>2]|0;
    $26 = ($$pr2$pre|0)==(0);
    if ($26) {
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    } else {
     $node$03 = $15;
    }
   } else {
    $node$03 = $15;
   }
  } else {
   $node$03 = 0;
  }
 } while(0);
 _ft_mem_free($9,$node$03);
 _ft_mem_free($9,$12);
 $$pre = HEAP32[$error>>2]|0;
 $$0 = $$pre;
 STACKTOP = sp;return ($$0|0);
}
function _destroy_face($memory,$face,$driver) {
 $memory = $memory|0;
 $face = $face|0;
 $driver = $driver|0;
 var $$lobit = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $n$02$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($driver)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 120|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  $5 = ((($face)) + 116|0);
  $6 = HEAP32[$5>>2]|0;
  FUNCTION_TABLE_vi[$3 & 8191]($6);
 }
 $7 = ((($face)) + 84|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 if (!($9)) {
  $10 = $8;
  while(1) {
   _FT_Done_GlyphSlot($10);
   $11 = HEAP32[$7>>2]|0;
   $12 = ($11|0)==(0|0);
   if ($12) {
    break;
   } else {
    $10 = $11;
   }
  }
 }
 $13 = ((($face)) + 108|0);
 _FT_List_Finalize($13,4344,$memory,$driver);
 $14 = ((($face)) + 88|0);
 HEAP32[$14>>2] = 0;
 $15 = ((($face)) + 48|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ($16|0)==(0|0);
 if (!($17)) {
  FUNCTION_TABLE_vi[$16 & 8191]($face);
 }
 $18 = ((($face)) + 36|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($19|0)>(0);
 $21 = ((($face)) + 40|0);
 if ($20) {
  $n$02$i = 0;
  while(1) {
   $22 = HEAP32[$21>>2]|0;
   $23 = (($22) + ($n$02$i<<2)|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = ((($24)) + 12|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = HEAP32[$24>>2]|0;
   $28 = ((($27)) + 100|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = ((($26)) + 8|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = ($31|0)==(0|0);
   if (!($32)) {
    FUNCTION_TABLE_vi[$31 & 8191]($24);
   }
   _ft_mem_free($29,$24);
   $33 = HEAP32[$21>>2]|0;
   $34 = (($33) + ($n$02$i<<2)|0);
   HEAP32[$34>>2] = 0;
   $35 = (($n$02$i) + 1)|0;
   $36 = HEAP32[$18>>2]|0;
   $37 = ($35|0)<($36|0);
   if ($37) {
    $n$02$i = $35;
   } else {
    break;
   }
  }
 }
 $38 = HEAP32[$21>>2]|0;
 _ft_mem_free($memory,$38);
 HEAP32[$21>>2] = 0;
 HEAP32[$18>>2] = 0;
 $39 = ((($1)) + 52|0);
 $40 = HEAP32[$39>>2]|0;
 $41 = ($40|0)==(0|0);
 if (!($41)) {
  FUNCTION_TABLE_vi[$40 & 8191]($face);
 }
 $42 = ((($face)) + 104|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = ($43|0)==(0|0);
 if (!($44)) {
  $45 = ((($face)) + 8|0);
  $46 = HEAP32[$45>>2]|0;
  $$lobit = $46 & 1024;
  $47 = ((($43)) + 28|0);
  $48 = HEAP32[$47>>2]|0;
  _FT_Stream_Close($43);
  $49 = ($$lobit|0)==(0);
  if ($49) {
   _ft_mem_free($48,$43);
  }
 }
 HEAP32[$42>>2] = 0;
 $50 = ((($face)) + 128|0);
 $51 = HEAP32[$50>>2]|0;
 $52 = ($51|0)==(0|0);
 if ($52) {
  _ft_mem_free($memory,$face);
  return;
 }
 _ft_mem_free($memory,$51);
 HEAP32[$50>>2] = 0;
 _ft_mem_free($memory,$face);
 return;
}
function _destroy_size($memory,$size,$driver) {
 $memory = $memory|0;
 $size = $size|0;
 $driver = $driver|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($size)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  FUNCTION_TABLE_vi[$1 & 8191]($size);
 }
 $3 = ((($driver)) + 12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 60|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0|0);
 if (!($7)) {
  FUNCTION_TABLE_vi[$6 & 8191]($size);
 }
 $8 = ((($size)) + 40|0);
 $9 = HEAP32[$8>>2]|0;
 _ft_mem_free($memory,$9);
 HEAP32[$8>>2] = 0;
 _ft_mem_free($memory,$size);
 return;
}
function _FT_New_Memory_Face($library,$file_base,$file_size,$face_index,$aface) {
 $library = $library|0;
 $file_base = $file_base|0;
 $file_size = $file_size|0;
 $face_index = $face_index|0;
 $aface = $aface|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $args = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $args = sp;
 $0 = ($file_base|0)==(0|0);
 if ($0) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$args>>2] = 1;
 $1 = ((($args)) + 4|0);
 HEAP32[$1>>2] = $file_base;
 $2 = ((($args)) + 8|0);
 HEAP32[$2>>2] = $file_size;
 $3 = ((($args)) + 16|0);
 HEAP32[$3>>2] = 0;
 $4 = (_FT_Open_Face($library,$args,$face_index,$aface)|0);
 $$0 = $4;
 STACKTOP = sp;return ($$0|0);
}
function _FT_Get_Module($library,$module_name) {
 $library = $library|0;
 $module_name = $module_name|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cur$02 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($library|0)!=(0|0);
 $1 = ($module_name|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  $$0 = 0;
  return ($$0|0);
 }
 $2 = ((($library)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (((($library)) + 20|0) + ($3<<2)|0);
 $5 = ($3|0)>(0);
 if (!($5)) {
  $$0 = 0;
  return ($$0|0);
 }
 $6 = ((($library)) + 20|0);
 $cur$02 = $6;
 while(1) {
  $9 = HEAP32[$cur$02>>2]|0;
  $10 = HEAP32[$9>>2]|0;
  $11 = ((($10)) + 8|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = (_strcmp($12,$module_name)|0);
  $14 = ($13|0)==(0);
  $8 = ((($cur$02)) + 4|0);
  if ($14) {
   $$0 = $9;
   label = 6;
   break;
  }
  $7 = ($8>>>0)<($4>>>0);
  if ($7) {
   $cur$02 = $8;
  } else {
   $$0 = 0;
   label = 6;
   break;
  }
 }
 if ((label|0) == 6) {
  return ($$0|0);
 }
 return (0)|0;
}
function _FT_Done_Face($face) {
 $face = $face|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($face|0)==(0|0);
 if ($0) {
  $error$0 = 35;
  return ($error$0|0);
 }
 $1 = ((($face)) + 96|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $error$0 = 35;
  return ($error$0|0);
 }
 $4 = ((($face)) + 128|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 56|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (($7) + -1)|0;
 HEAP32[$6>>2] = $8;
 $9 = ($7|0)>(1);
 if ($9) {
  $error$0 = 0;
  return ($error$0|0);
 }
 $10 = ((($2)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($2)) + 16|0);
 $13 = (_FT_List_Find($12,$face)|0);
 $14 = ($13|0)==(0|0);
 if ($14) {
  $error$0 = 35;
  return ($error$0|0);
 }
 _FT_List_Remove($12,$13);
 _ft_mem_free($11,$13);
 _destroy_face($11,$face,$2);
 $error$0 = 0;
 return ($error$0|0);
}
function _FT_Done_Size($size) {
 $size = $size|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($size|0)==(0|0);
 if ($0) {
  $$0 = 36;
  return ($$0|0);
 }
 $1 = HEAP32[$size>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = 35;
  return ($$0|0);
 }
 $3 = ((($1)) + 96|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if ($5) {
  $$0 = 34;
  return ($$0|0);
 }
 $6 = ((($4)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($1)) + 108|0);
 $9 = (_FT_List_Find($8,$size)|0);
 $10 = ($9|0)==(0|0);
 if ($10) {
  $$0 = 36;
  return ($$0|0);
 }
 _FT_List_Remove($8,$9);
 _ft_mem_free($7,$9);
 $11 = ((($1)) + 88|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)==($size|0);
 if ($13) {
  HEAP32[$11>>2] = 0;
  $14 = HEAP32[$8>>2]|0;
  $15 = ($14|0)==(0|0);
  if (!($15)) {
   $16 = ((($14)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   HEAP32[$11>>2] = $17;
  }
 }
 $18 = ((($size)) + 8|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($19|0)==(0|0);
 if (!($20)) {
  FUNCTION_TABLE_vi[$19 & 8191]($size);
 }
 $21 = ((($4)) + 12|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ((($22)) + 60|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ($24|0)==(0|0);
 if (!($25)) {
  FUNCTION_TABLE_vi[$24 & 8191]($size);
 }
 $26 = ((($size)) + 40|0);
 $27 = HEAP32[$26>>2]|0;
 _ft_mem_free($7,$27);
 HEAP32[$26>>2] = 0;
 _ft_mem_free($7,$size);
 $$0 = 0;
 return ($$0|0);
}
function _FT_Match_Size($face,$req,$ignore_width,$size_index) {
 $face = $face|0;
 $req = $req|0;
 $ignore_width = $ignore_width|0;
 $size_index = $size_index|0;
 var $$ = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $h$0 = 0, $i$0$lcssa5 = 0, $i$06 = 0, $i$06$us = 0;
 var $w$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 2;
 $3 = ($2|0)==(0);
 if ($3) {
  $$2 = 35;
  return ($$2|0);
 }
 $4 = HEAP32[$req>>2]|0;
 $5 = ($4|0)==(0);
 if (!($5)) {
  $$2 = 7;
  return ($$2|0);
 }
 $6 = ((($req)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0);
 $9 = ((($req)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 if ($8) {
  $24 = $10;
 } else {
  $11 = Math_imul($10, $7)|0;
  $12 = (($11) + 36)|0;
  $13 = (($12|0) / 72)&-1;
  $24 = $13;
 }
 $14 = ((($req)) + 16|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)==(0);
 $17 = ((($req)) + 8|0);
 $18 = HEAP32[$17>>2]|0;
 if ($16) {
  $25 = $18;
 } else {
  $19 = Math_imul($18, $15)|0;
  $20 = (($19) + 36)|0;
  $21 = (($20|0) / 72)&-1;
  $25 = $21;
 }
 $22 = ($10|0)==(0);
 $23 = ($18|0)==(0);
 $$ = $23 ? $24 : $25;
 $w$0 = $22 ? $$ : $24;
 $h$0 = $22 ? $25 : $$;
 $26 = (($w$0) + 32)|0;
 $27 = (($h$0) + 32)|0;
 $28 = ((($face)) + 28|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = ($29|0)>(0);
 if (!($30)) {
  $$2 = 23;
  return ($$2|0);
 }
 $31 = ((($face)) + 32|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ($ignore_width<<24>>24)==(0);
 L16: do {
  if ($33) {
   $i$06$us = 0;
   while(1) {
    $34 = (((($32) + ($i$06$us<<4)|0)) + 12|0);
    $35 = HEAP32[$34>>2]|0;
    $36 = (($35) + 32)|0;
    $37 = $36 ^ $27;
    $38 = ($37>>>0)>(63);
    if (!($38)) {
     $39 = (((($32) + ($i$06$us<<4)|0)) + 8|0);
     $40 = HEAP32[$39>>2]|0;
     $41 = (($40) + 32)|0;
     $42 = $41 ^ $26;
     $43 = ($42>>>0)>(63);
     if (!($43)) {
      $i$0$lcssa5 = $i$06$us;
      break L16;
     }
    }
    $44 = (($i$06$us) + 1)|0;
    $45 = ($44|0)<($29|0);
    if ($45) {
     $i$06$us = $44;
    } else {
     $$2 = 23;
     break;
    }
   }
   return ($$2|0);
  } else {
   $i$06 = 0;
   while(1) {
    $46 = (((($32) + ($i$06<<4)|0)) + 12|0);
    $47 = HEAP32[$46>>2]|0;
    $48 = (($47) + 32)|0;
    $49 = $48 ^ $27;
    $50 = ($49>>>0)>(63);
    if (!($50)) {
     $i$0$lcssa5 = $i$06;
     break L16;
    }
    $52 = (($i$06) + 1)|0;
    $53 = ($52|0)<($29|0);
    if ($53) {
     $i$06 = $52;
    } else {
     $$2 = 23;
     break;
    }
   }
   return ($$2|0);
  }
 } while(0);
 $51 = ($size_index|0)==(0|0);
 if ($51) {
  $$2 = 0;
  return ($$2|0);
 }
 HEAP32[$size_index>>2] = $i$0$lcssa5;
 $$2 = 0;
 return ($$2|0);
}
function _ft_synthesize_vertical_metrics($metrics,$advance) {
 $metrics = $metrics|0;
 $advance = $advance|0;
 var $$ = 0, $$0 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $height$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($metrics)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($metrics)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)<(0);
 if ($4) {
  $5 = ($1|0)<($3|0);
  $$ = $5 ? $3 : $1;
  $height$0 = $$;
 } else {
  $6 = ($3|0)>(0);
  $7 = $6 ? $3 : 0;
  $$1 = (($1) - ($7))|0;
  $height$0 = $$1;
 }
 $8 = ($advance|0)==(0);
 if ($8) {
  $9 = ($height$0*12)|0;
  $10 = (($9|0) / 10)&-1;
  $$0 = $10;
 } else {
  $$0 = $advance;
 }
 $11 = ((($metrics)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($metrics)) + 16|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($14|0) / 2)&-1;
 $16 = (($12) - ($15))|0;
 $17 = ((($metrics)) + 20|0);
 HEAP32[$17>>2] = $16;
 $18 = (($$0) - ($height$0))|0;
 $19 = (($18|0) / 2)&-1;
 $20 = ((($metrics)) + 24|0);
 HEAP32[$20>>2] = $19;
 $21 = ((($metrics)) + 28|0);
 HEAP32[$21>>2] = $$0;
 return;
}
function _FT_Select_Metrics($face,$strike_index) {
 $face = $face|0;
 $strike_index = $strike_index|0;
 var $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 88|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 32|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (((($3) + ($strike_index<<4)|0)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($5) + 32)|0;
 $7 = $6 >>> 6;
 $8 = $7&65535;
 $9 = ((($1)) + 12|0);
 HEAP16[$9>>1] = $8;
 $10 = (((($3) + ($strike_index<<4)|0)) + 12|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + 32)|0;
 $13 = $12 >>> 6;
 $14 = $13&65535;
 $15 = ((($1)) + 14|0);
 HEAP16[$15>>1] = $14;
 $16 = ((($face)) + 8|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 & 1;
 $19 = ($18|0)==(0);
 if ($19) {
  $59 = ((($1)) + 16|0);
  HEAP32[$59>>2] = 65536;
  $60 = ((($1)) + 20|0);
  HEAP32[$60>>2] = 65536;
  $61 = ((($1)) + 24|0);
  HEAP32[$61>>2] = $11;
  $62 = ((($1)) + 28|0);
  HEAP32[$62>>2] = 0;
  $63 = (($3) + ($strike_index<<4)|0);
  $64 = HEAP16[$63>>1]|0;
  $65 = $64 << 16 >> 16;
  $66 = $65 << 6;
  $67 = ((($1)) + 32|0);
  HEAP32[$67>>2] = $66;
  $$sink = $5;
  $68 = ((($1)) + 36|0);
  HEAP32[$68>>2] = $$sink;
  return;
 } else {
  $20 = ((($face)) + 68|0);
  $21 = HEAP16[$20>>1]|0;
  $22 = $21&65535;
  $23 = (_FT_DivFix($5,$22)|0);
  $24 = ((($1)) + 16|0);
  HEAP32[$24>>2] = $23;
  $25 = HEAP32[$10>>2]|0;
  $26 = HEAP16[$20>>1]|0;
  $27 = $26&65535;
  $28 = (_FT_DivFix($25,$27)|0);
  $29 = ((($1)) + 20|0);
  HEAP32[$29>>2] = $28;
  $30 = ((($face)) + 70|0);
  $31 = HEAP16[$30>>1]|0;
  $32 = $31 << 16 >> 16;
  $33 = (_FT_MulFix($32,$28)|0);
  $34 = (($33) + 63)|0;
  $35 = $34 & -64;
  $36 = ((($1)) + 24|0);
  HEAP32[$36>>2] = $35;
  $37 = ((($face)) + 72|0);
  $38 = HEAP16[$37>>1]|0;
  $39 = $38 << 16 >> 16;
  $40 = HEAP32[$29>>2]|0;
  $41 = (_FT_MulFix($39,$40)|0);
  $42 = $41 & -64;
  $43 = ((($1)) + 28|0);
  HEAP32[$43>>2] = $42;
  $44 = ((($face)) + 74|0);
  $45 = HEAP16[$44>>1]|0;
  $46 = $45 << 16 >> 16;
  $47 = HEAP32[$29>>2]|0;
  $48 = (_FT_MulFix($46,$47)|0);
  $49 = (($48) + 32)|0;
  $50 = $49 & -64;
  $51 = ((($1)) + 32|0);
  HEAP32[$51>>2] = $50;
  $52 = ((($face)) + 76|0);
  $53 = HEAP16[$52>>1]|0;
  $54 = $53 << 16 >> 16;
  $55 = HEAP32[$24>>2]|0;
  $56 = (_FT_MulFix($54,$55)|0);
  $57 = (($56) + 32)|0;
  $58 = $57 & -64;
  $$sink = $58;
  $68 = ((($1)) + 36|0);
  HEAP32[$68>>2] = $$sink;
  return;
 }
}
function _FT_Request_Metrics($face,$req) {
 $face = $face|0;
 $req = $req|0;
 var $$pr = 0, $$pre = 0, $$pre$phi2Z2D = 0, $$pre$phiZ2D = 0, $$pre1 = 0, $$w$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $h$0 = 0, $h$1 = 0, $scaled_h$0 = 0, $scaled_h$0$ph = 0, $scaled_h$1 = 0, $scaled_w$0 = 0, $scaled_w$0$ph = 0, $scaled_w$1 = 0, $w$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 88|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 12|0);
 $3 = ((($face)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 & 1;
 $6 = ($5|0)==(0);
 if ($6) {
  ;HEAP32[$2>>2]=0|0;HEAP32[$2+4>>2]=0|0;HEAP32[$2+8>>2]=0|0;HEAP32[$2+12>>2]=0|0;HEAP32[$2+16>>2]=0|0;HEAP32[$2+20>>2]=0|0;HEAP32[$2+24>>2]=0|0;
  $136 = ((($1)) + 16|0);
  HEAP32[$136>>2] = 65536;
  $137 = ((($1)) + 20|0);
  HEAP32[$137>>2] = 65536;
  return;
 }
 $7 = HEAP32[$req>>2]|0;
 L5: do {
  switch ($7|0) {
  case 0:  {
   $8 = ((($face)) + 68|0);
   $9 = HEAP16[$8>>1]|0;
   $10 = $9&65535;
   $h$0 = $10;$w$0 = $10;
   label = 11;
   break;
  }
  case 1:  {
   $11 = ((($face)) + 70|0);
   $12 = HEAP16[$11>>1]|0;
   $13 = $12 << 16 >> 16;
   $14 = ((($face)) + 72|0);
   $15 = HEAP16[$14>>1]|0;
   $16 = $15 << 16 >> 16;
   $17 = (($13) - ($16))|0;
   $h$0 = $17;$w$0 = $17;
   label = 11;
   break;
  }
  case 2:  {
   $18 = ((($face)) + 60|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ((($face)) + 52|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = (($19) - ($21))|0;
   $23 = ((($face)) + 64|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = ((($face)) + 56|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = (($24) - ($26))|0;
   $h$0 = $27;$w$0 = $22;
   label = 11;
   break;
  }
  case 3:  {
   $28 = ((($face)) + 76|0);
   $29 = HEAP16[$28>>1]|0;
   $30 = $29 << 16 >> 16;
   $31 = ((($face)) + 70|0);
   $32 = HEAP16[$31>>1]|0;
   $33 = $32 << 16 >> 16;
   $34 = ((($face)) + 72|0);
   $35 = HEAP16[$34>>1]|0;
   $36 = $35 << 16 >> 16;
   $37 = (($33) - ($36))|0;
   $h$0 = $37;$w$0 = $30;
   label = 11;
   break;
  }
  case 4:  {
   $38 = ((($req)) + 4|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = ((($1)) + 16|0);
   HEAP32[$40>>2] = $39;
   $41 = ((($req)) + 8|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = ((($1)) + 20|0);
   HEAP32[$43>>2] = $42;
   $44 = ($39|0)==(0);
   if ($44) {
    HEAP32[$40>>2] = $42;
    $scaled_h$0$ph = 0;$scaled_w$0$ph = 0;
    label = 23;
    break L5;
   }
   $45 = ($42|0)==(0);
   if ($45) {
    HEAP32[$43>>2] = $39;
    $scaled_h$0$ph = 0;$scaled_w$0$ph = 0;
    label = 23;
   } else {
    $scaled_h$0$ph = 0;$scaled_w$0$ph = 0;
    label = 23;
   }
   break;
  }
  default: {
   $h$0 = 0;$w$0 = 0;
   label = 11;
  }
  }
 } while(0);
 do {
  if ((label|0) == 11) {
   $46 = ($w$0|0)<(0);
   $47 = (0 - ($w$0))|0;
   $$w$0 = $46 ? $47 : $w$0;
   $48 = ($h$0|0)<(0);
   $49 = (0 - ($h$0))|0;
   $h$1 = $48 ? $49 : $h$0;
   $50 = ((($req)) + 12|0);
   $51 = HEAP32[$50>>2]|0;
   $52 = ($51|0)==(0);
   $53 = ((($req)) + 4|0);
   $54 = HEAP32[$53>>2]|0;
   if ($52) {
    $67 = $54;
   } else {
    $55 = Math_imul($54, $51)|0;
    $56 = (($55) + 36)|0;
    $57 = (($56|0) / 72)&-1;
    $67 = $57;
   }
   $58 = ((($req)) + 16|0);
   $59 = HEAP32[$58>>2]|0;
   $60 = ($59|0)==(0);
   $61 = ((($req)) + 8|0);
   $62 = HEAP32[$61>>2]|0;
   if ($60) {
    $72 = $62;
   } else {
    $63 = Math_imul($62, $59)|0;
    $64 = (($63) + 36)|0;
    $65 = (($64|0) / 72)&-1;
    $72 = $65;
   }
   $66 = ($54|0)==(0);
   if ($66) {
    $81 = (_FT_DivFix($72,$h$1)|0);
    $82 = ((($1)) + 20|0);
    HEAP32[$82>>2] = $81;
    $83 = ((($1)) + 16|0);
    HEAP32[$83>>2] = $81;
    $84 = (_FT_MulDiv($72,$$w$0,$h$1)|0);
    $scaled_h$0$ph = $72;$scaled_w$0$ph = $84;
    label = 23;
    break;
   }
   $68 = (_FT_DivFix($67,$$w$0)|0);
   $69 = ((($1)) + 16|0);
   HEAP32[$69>>2] = $68;
   $70 = HEAP32[$61>>2]|0;
   $71 = ($70|0)==(0);
   if ($71) {
    $79 = ((($1)) + 20|0);
    HEAP32[$79>>2] = $68;
    $80 = (_FT_MulDiv($67,$h$1,$$w$0)|0);
    $scaled_h$0$ph = $80;$scaled_w$0$ph = $67;
    label = 23;
    break;
   }
   $73 = (_FT_DivFix($72,$h$1)|0);
   $74 = ((($1)) + 20|0);
   HEAP32[$74>>2] = $73;
   $75 = HEAP32[$req>>2]|0;
   $76 = ($75|0)==(3);
   if ($76) {
    $77 = HEAP32[$69>>2]|0;
    $78 = ($73|0)>($77|0);
    if ($78) {
     HEAP32[$74>>2] = $77;
     $scaled_h$0$ph = $72;$scaled_w$0$ph = $67;
     label = 23;
     break;
    } else {
     HEAP32[$69>>2] = $73;
     $scaled_h$0$ph = $72;$scaled_w$0$ph = $67;
     label = 23;
     break;
    }
   } else {
    $86 = $75;$scaled_h$0 = $72;$scaled_w$0 = $67;
   }
  }
 } while(0);
 if ((label|0) == 23) {
  $$pr = HEAP32[$req>>2]|0;
  $86 = $$pr;$scaled_h$0 = $scaled_h$0$ph;$scaled_w$0 = $scaled_w$0$ph;
 }
 $85 = ($86|0)==(0);
 if ($85) {
  $$pre = ((($1)) + 20|0);
  $$pre1 = ((($1)) + 16|0);
  $$pre$phi2Z2D = $$pre1;$$pre$phiZ2D = $$pre;$scaled_h$1 = $scaled_h$0;$scaled_w$1 = $scaled_w$0;
 } else {
  $87 = ((($face)) + 68|0);
  $88 = HEAP16[$87>>1]|0;
  $89 = $88&65535;
  $90 = ((($1)) + 16|0);
  $91 = HEAP32[$90>>2]|0;
  $92 = (_FT_MulFix($89,$91)|0);
  $93 = HEAP16[$87>>1]|0;
  $94 = $93&65535;
  $95 = ((($1)) + 20|0);
  $96 = HEAP32[$95>>2]|0;
  $97 = (_FT_MulFix($94,$96)|0);
  $$pre$phi2Z2D = $90;$$pre$phiZ2D = $95;$scaled_h$1 = $97;$scaled_w$1 = $92;
 }
 $98 = (($scaled_w$1) + 32)|0;
 $99 = $98 >>> 6;
 $100 = $99&65535;
 HEAP16[$2>>1] = $100;
 $101 = (($scaled_h$1) + 32)|0;
 $102 = $101 >>> 6;
 $103 = $102&65535;
 $104 = ((($1)) + 14|0);
 HEAP16[$104>>1] = $103;
 $105 = ((($face)) + 70|0);
 $106 = HEAP16[$105>>1]|0;
 $107 = $106 << 16 >> 16;
 $108 = HEAP32[$$pre$phiZ2D>>2]|0;
 $109 = (_FT_MulFix($107,$108)|0);
 $110 = (($109) + 63)|0;
 $111 = $110 & -64;
 $112 = ((($1)) + 24|0);
 HEAP32[$112>>2] = $111;
 $113 = ((($face)) + 72|0);
 $114 = HEAP16[$113>>1]|0;
 $115 = $114 << 16 >> 16;
 $116 = HEAP32[$$pre$phiZ2D>>2]|0;
 $117 = (_FT_MulFix($115,$116)|0);
 $118 = $117 & -64;
 $119 = ((($1)) + 28|0);
 HEAP32[$119>>2] = $118;
 $120 = ((($face)) + 74|0);
 $121 = HEAP16[$120>>1]|0;
 $122 = $121 << 16 >> 16;
 $123 = HEAP32[$$pre$phiZ2D>>2]|0;
 $124 = (_FT_MulFix($122,$123)|0);
 $125 = (($124) + 32)|0;
 $126 = $125 & -64;
 $127 = ((($1)) + 32|0);
 HEAP32[$127>>2] = $126;
 $128 = ((($face)) + 76|0);
 $129 = HEAP16[$128>>1]|0;
 $130 = $129 << 16 >> 16;
 $131 = HEAP32[$$pre$phi2Z2D>>2]|0;
 $132 = (_FT_MulFix($130,$131)|0);
 $133 = (($132) + 32)|0;
 $134 = $133 & -64;
 $135 = ((($1)) + 36|0);
 HEAP32[$135>>2] = $134;
 return;
}
function _FT_Select_Size($face,$strike_index) {
 $face = $face|0;
 $strike_index = $strike_index|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($face|0)==(0|0);
 if ($0) {
  $$0 = 35;
  return ($$0|0);
 }
 $1 = ((($face)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2 & 2;
 $4 = ($3|0)==(0);
 if ($4) {
  $$0 = 35;
  return ($$0|0);
 }
 $5 = ($strike_index|0)<(0);
 if ($5) {
  $$0 = 6;
  return ($$0|0);
 }
 $6 = ((($face)) + 28|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)>($strike_index|0);
 if (!($8)) {
  $$0 = 6;
  return ($$0|0);
 }
 $9 = ((($face)) + 96|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + 12|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($12)) + 92|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ($14|0)==(0|0);
 if ($15) {
  _FT_Select_Metrics($face,$strike_index);
  $$0 = 0;
  return ($$0|0);
 } else {
  $16 = ((($face)) + 88|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = (FUNCTION_TABLE_iii[$14 & 8191]($17,$strike_index)|0);
  $$0 = $18;
  return ($$0|0);
 }
 return (0)|0;
}
function _FT_Request_Size($face,$req) {
 $face = $face|0;
 $req = $req|0;
 var $$$i = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $8 = 0, $9 = 0, $h$0$i = 0, $i$06$us$i = 0, $i$06$us$i$lcssa = 0, $or$cond = 0, $w$0$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($face|0)==(0|0);
 if ($0) {
  $$1 = 35;
  return ($$1|0);
 }
 $1 = ($req|0)==(0|0);
 if ($1) {
  $$1 = 6;
  return ($$1|0);
 }
 $2 = ((($req)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)<(0);
 if ($4) {
  $$1 = 6;
  return ($$1|0);
 }
 $5 = ((($req)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)<(0);
 if ($7) {
  $$1 = 6;
  return ($$1|0);
 }
 $8 = HEAP32[$req>>2]|0;
 $9 = ($8>>>0)>(4);
 if ($9) {
  $$1 = 6;
  return ($$1|0);
 }
 $10 = ((($face)) + 96|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($11)) + 12|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($13)) + 88|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)==(0|0);
 if (!($16)) {
  $17 = ((($face)) + 88|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = (FUNCTION_TABLE_iii[$15 & 8191]($18,$req)|0);
  $$1 = $19;
  return ($$1|0);
 }
 $20 = ((($face)) + 8|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = $21 & 3;
 $23 = ($22|0)==(2);
 if (!($23)) {
  _FT_Request_Metrics($face,$req);
  $$1 = 0;
  return ($$1|0);
 }
 $24 = $21 & 2;
 $25 = ($24|0)==(0);
 if ($25) {
  $$1 = 35;
  return ($$1|0);
 }
 $26 = ($8|0)==(0);
 if (!($26)) {
  $$1 = 7;
  return ($$1|0);
 }
 $27 = ((($req)) + 12|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ($28|0)==(0);
 if ($29) {
  $41 = $3;
 } else {
  $30 = Math_imul($28, $3)|0;
  $31 = (($30) + 36)|0;
  $32 = (($31|0) / 72)&-1;
  $41 = $32;
 }
 $33 = ((($req)) + 16|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = ($34|0)==(0);
 if ($35) {
  $42 = $6;
 } else {
  $36 = Math_imul($6, $34)|0;
  $37 = (($36) + 36)|0;
  $38 = (($37|0) / 72)&-1;
  $42 = $38;
 }
 $39 = ($3|0)==(0);
 $40 = ($6|0)==(0);
 $$$i = $40 ? $41 : $42;
 $w$0$i = $39 ? $$$i : $41;
 $h$0$i = $39 ? $42 : $$$i;
 $43 = (($w$0$i) + 32)|0;
 $44 = (($h$0$i) + 32)|0;
 $45 = ((($face)) + 28|0);
 $46 = HEAP32[$45>>2]|0;
 $47 = ($46|0)>(0);
 if (!($47)) {
  $$1 = 23;
  return ($$1|0);
 }
 $48 = ((($face)) + 32|0);
 $49 = HEAP32[$48>>2]|0;
 $i$06$us$i = 0;
 while(1) {
  $50 = (((($49) + ($i$06$us$i<<4)|0)) + 12|0);
  $51 = HEAP32[$50>>2]|0;
  $52 = (($51) + 32)|0;
  $53 = $52 ^ $44;
  $54 = ($53>>>0)>(63);
  if (!($54)) {
   $55 = (((($49) + ($i$06$us$i<<4)|0)) + 8|0);
   $56 = HEAP32[$55>>2]|0;
   $57 = (($56) + 32)|0;
   $58 = $57 ^ $43;
   $59 = ($58>>>0)>(63);
   if (!($59)) {
    $i$06$us$i$lcssa = $i$06$us$i;
    break;
   }
  }
  $60 = (($i$06$us$i) + 1)|0;
  $61 = ($60|0)<($46|0);
  if ($61) {
   $i$06$us$i = $60;
  } else {
   $$1 = 23;
   label = 25;
   break;
  }
 }
 if ((label|0) == 25) {
  return ($$1|0);
 }
 $62 = ($i$06$us$i$lcssa|0)>(-1);
 $63 = ($46|0)>($i$06$us$i$lcssa|0);
 $or$cond = $62 & $63;
 if (!($or$cond)) {
  $$1 = 6;
  return ($$1|0);
 }
 $64 = ((($13)) + 92|0);
 $65 = HEAP32[$64>>2]|0;
 $66 = ($65|0)==(0|0);
 if ($66) {
  _FT_Select_Metrics($face,$i$06$us$i$lcssa);
  $$1 = 0;
  return ($$1|0);
 } else {
  $67 = ((($face)) + 88|0);
  $68 = HEAP32[$67>>2]|0;
  $69 = (FUNCTION_TABLE_iii[$65 & 8191]($68,$i$06$us$i$lcssa)|0);
  $$1 = $69;
  return ($$1|0);
 }
 return (0)|0;
}
function _FT_Set_Pixel_Sizes($face,$pixel_width,$pixel_height) {
 $face = $face|0;
 $pixel_width = $pixel_width|0;
 $pixel_height = $pixel_height|0;
 var $$$$01 = 0, $$$01 = 0, $$0 = 0, $$01 = 0, $$1 = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $pixel_width$pixel_height = 0;
 var $req = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $req = sp;
 $0 = ($pixel_width|0)==(0);
 $1 = ($pixel_height|0)==(0);
 $pixel_width$pixel_height = $1 ? $pixel_width : $pixel_height;
 $$01 = $0 ? $pixel_height : $pixel_width;
 $$0 = $0 ? $pixel_height : $pixel_width$pixel_height;
 $2 = ($$01|0)==(0);
 $$$01 = $2 ? 1 : $$01;
 $3 = ($$0|0)==(0);
 $$1 = $3 ? 1 : $$0;
 $4 = ($$$01>>>0)<(65535);
 $$$$01 = $4 ? $$$01 : 65535;
 $5 = ($$1>>>0)<(65535);
 $$2 = $5 ? $$1 : 65535;
 HEAP32[$req>>2] = 0;
 $6 = $$$$01 << 6;
 $7 = ((($req)) + 4|0);
 HEAP32[$7>>2] = $6;
 $8 = $$2 << 6;
 $9 = ((($req)) + 8|0);
 HEAP32[$9>>2] = $8;
 $10 = ((($req)) + 12|0);
 HEAP32[$10>>2] = 0;
 $11 = ((($req)) + 16|0);
 HEAP32[$11>>2] = 0;
 $12 = (_FT_Request_Size($face,$req)|0);
 STACKTOP = sp;return ($12|0);
}
function _FT_Get_Kerning($face,$left_glyph,$right_glyph,$kern_mode,$akerning) {
 $face = $face|0;
 $left_glyph = $left_glyph|0;
 $right_glyph = $right_glyph|0;
 $kern_mode = $kern_mode|0;
 $akerning = $akerning|0;
 var $$0 = 0, $$pre = 0, $$pre2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($face|0)==(0|0);
 if ($0) {
  $$0 = 35;
  return ($$0|0);
 }
 $1 = ($akerning|0)==(0|0);
 if ($1) {
  $$0 = 6;
  return ($$0|0);
 }
 $2 = ((($face)) + 96|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$akerning>>2] = 0;
 $4 = ((($akerning)) + 4|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($3)) + 12|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($6)) + 76|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 if ($9) {
  $$0 = 0;
  return ($$0|0);
 }
 $10 = (FUNCTION_TABLE_iiiii[$8 & 8191]($face,$left_glyph,$right_glyph,$akerning)|0);
 $11 = ($10|0)==(0);
 $12 = ($kern_mode|0)!=(2);
 $or$cond = $12 & $11;
 if (!($or$cond)) {
  $$0 = $10;
  return ($$0|0);
 }
 $13 = HEAP32[$akerning>>2]|0;
 $14 = ((($face)) + 88|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($15)) + 16|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = (_FT_MulFix($13,$17)|0);
 HEAP32[$akerning>>2] = $18;
 $19 = HEAP32[$4>>2]|0;
 $20 = HEAP32[$14>>2]|0;
 $21 = ((($20)) + 20|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = (_FT_MulFix($19,$22)|0);
 HEAP32[$4>>2] = $23;
 $24 = ($kern_mode|0)==(1);
 if ($24) {
  $$0 = 0;
  return ($$0|0);
 }
 $25 = HEAP32[$14>>2]|0;
 $26 = ((($25)) + 12|0);
 $27 = HEAP16[$26>>1]|0;
 $28 = ($27&65535)<(25);
 if ($28) {
  $29 = $27&65535;
  $30 = HEAP32[$akerning>>2]|0;
  $31 = (_FT_MulDiv($30,$29,25)|0);
  HEAP32[$akerning>>2] = $31;
  $$pre = HEAP32[$14>>2]|0;
  $33 = $$pre;
 } else {
  $33 = $25;
 }
 $32 = ((($33)) + 14|0);
 $34 = HEAP16[$32>>1]|0;
 $35 = ($34&65535)<(25);
 if ($35) {
  $36 = $34&65535;
  $37 = HEAP32[$4>>2]|0;
  $38 = (_FT_MulDiv($37,$36,25)|0);
  HEAP32[$4>>2] = $38;
  $43 = $38;
 } else {
  $$pre2 = HEAP32[$4>>2]|0;
  $43 = $$pre2;
 }
 $39 = HEAP32[$akerning>>2]|0;
 $40 = (($39) + 32)|0;
 $41 = $40 & -64;
 HEAP32[$akerning>>2] = $41;
 $42 = (($43) + 32)|0;
 $44 = $42 & -64;
 HEAP32[$4>>2] = $44;
 $$0 = 0;
 return ($$0|0);
}
function _FT_Select_Charmap($face,$encoding) {
 $face = $face|0;
 $encoding = $encoding|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa27 = 0, $$lcssa28 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cur$05 = 0, $cur$1$i = 0, $switch$split2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($face|0)==(0|0);
 if ($0) {
  $$0 = 35;
  return ($$0|0);
 }
 $switch$split2D = ($encoding|0)<(1970170211);
 L4: do {
  if ($switch$split2D) {
   switch ($encoding|0) {
   case 0:  {
    $$0 = 6;
    break;
   }
   default: {
    break L4;
   }
   }
   return ($$0|0);
  } else {
   switch ($encoding|0) {
   case 1970170211:  {
    break;
   }
   default: {
    break L4;
   }
   }
   $1 = ((($face)) + 40|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = ($2|0)==(0|0);
   if ($3) {
    $$0 = 38;
    return ($$0|0);
   }
   $4 = ((($face)) + 36|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = (($2) + ($5<<2)|0);
   $7 = ((($6)) + -4|0);
   $8 = ($7>>>0)<($2>>>0);
   L12: do {
    if ($8) {
     $cur$1$i = $6;
    } else {
     $10 = $7;
     L13: while(1) {
      $9 = HEAP32[$10>>2]|0;
      $11 = ((($9)) + 4|0);
      $12 = HEAP32[$11>>2]|0;
      $13 = ($12|0)==(1970170211);
      L15: do {
       if ($13) {
        $14 = ((($9)) + 8|0);
        $15 = HEAP32[$14>>2]|0;
        $16 = $15&65535;
        $17 = $15 >>> 16;
        $18 = $17&65535;
        switch ($16<<16>>16) {
        case 3:  {
         $19 = ($18<<16>>16)==(10);
         if ($19) {
          $$lcssa28 = $9;
          break L13;
         } else {
          break L15;
         }
         break;
        }
        case 0:  {
         $22 = ($18<<16>>16)==(4);
         if ($22) {
          $$lcssa28 = $9;
          break L13;
         } else {
          break L15;
         }
         break;
        }
        default: {
         break L15;
        }
        }
       }
      } while(0);
      $20 = ((($10)) + -4|0);
      $21 = ($20>>>0)<($2>>>0);
      if ($21) {
       $cur$1$i = $6;
       break L12;
      } else {
       $10 = $20;
      }
     }
     $23 = $$lcssa28;
     $24 = ((($face)) + 92|0);
     HEAP32[$24>>2] = $23;
     $$0 = 0;
     return ($$0|0);
    }
   } while(0);
   while(1) {
    $25 = ((($cur$1$i)) + -4|0);
    $26 = ($25>>>0)<($2>>>0);
    if ($26) {
     $$0 = 38;
     label = 19;
     break;
    }
    $27 = HEAP32[$25>>2]|0;
    $28 = ((($27)) + 4|0);
    $29 = HEAP32[$28>>2]|0;
    $30 = ($29|0)==(1970170211);
    if ($30) {
     $$lcssa27 = $27;
     break;
    } else {
     $cur$1$i = $25;
    }
   }
   if ((label|0) == 19) {
    return ($$0|0);
   }
   $31 = ((($face)) + 92|0);
   HEAP32[$31>>2] = $$lcssa27;
   $$0 = 0;
   return ($$0|0);
  }
 } while(0);
 $32 = ((($face)) + 40|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ($33|0)==(0|0);
 if ($34) {
  $$0 = 38;
  return ($$0|0);
 }
 $35 = ((($face)) + 36|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = (($33) + ($36<<2)|0);
 $38 = ($36|0)>(0);
 if ($38) {
  $cur$05 = $33;
 } else {
  $$0 = 6;
  return ($$0|0);
 }
 while(1) {
  $41 = HEAP32[$cur$05>>2]|0;
  $42 = ((($41)) + 4|0);
  $43 = HEAP32[$42>>2]|0;
  $44 = ($43|0)==($encoding|0);
  $40 = ((($cur$05)) + 4|0);
  if ($44) {
   $$lcssa = $41;
   break;
  }
  $39 = ($40>>>0)<($37>>>0);
  if ($39) {
   $cur$05 = $40;
  } else {
   $$0 = 6;
   label = 19;
   break;
  }
 }
 if ((label|0) == 19) {
  return ($$0|0);
 }
 $45 = ((($face)) + 92|0);
 HEAP32[$45>>2] = $$lcssa;
 $$0 = 0;
 return ($$0|0);
}
function _FT_Set_Charmap($face,$charmap) {
 $face = $face|0;
 $charmap = $charmap|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cmap_info$i = 0, $cur$02 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $cmap_info$i = sp;
 $0 = ($face|0)==(0|0);
 if ($0) {
  $$0 = 35;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ((($face)) + 40|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)!=(0|0);
 $4 = ($charmap|0)!=(0|0);
 $or$cond = $4 & $3;
 if (!($or$cond)) {
  $$0 = 38;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = HEAP32[$charmap>>2]|0;
 $6 = ($5|0)==(0|0);
 if ($6) {
  label = 7;
 } else {
  $7 = ((($5)) + 96|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($9)) + 32|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($11|0)==(0|0);
  if ($12) {
   label = 7;
  } else {
   $13 = (FUNCTION_TABLE_iii[$11 & 8191]($8,593070)|0);
   $14 = ($13|0)==(0|0);
   if ($14) {
    label = 7;
   } else {
    $15 = HEAP32[$13>>2]|0;
    $16 = (FUNCTION_TABLE_iii[$15 & 8191]($charmap,$cmap_info$i)|0);
    $17 = ($16|0)==(0);
    if ($17) {
     $18 = ((($cmap_info$i)) + 4|0);
     $19 = HEAP32[$18>>2]|0;
     $20 = ($19|0)==(14);
     if ($20) {
      $$0 = 6;
      STACKTOP = sp;return ($$0|0);
     }
    } else {
     label = 7;
    }
   }
  }
 }
 if ((label|0) == 7) {
 }
 $21 = ((($face)) + 36|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = (($2) + ($22<<2)|0);
 $24 = ($22|0)>(0);
 if ($24) {
  $cur$02 = $2;
 } else {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 while(1) {
  $27 = HEAP32[$cur$02>>2]|0;
  $28 = ($27|0)==($charmap|0);
  $26 = ((($cur$02)) + 4|0);
  if ($28) {
   break;
  }
  $25 = ($26>>>0)<($23>>>0);
  if ($25) {
   $cur$02 = $26;
  } else {
   $$0 = 6;
   label = 13;
   break;
  }
 }
 if ((label|0) == 13) {
  STACKTOP = sp;return ($$0|0);
 }
 $29 = ((($face)) + 92|0);
 HEAP32[$29>>2] = $charmap;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _FT_CMap_New($clazz,$init_data,$charmap,$acmap) {
 $clazz = $clazz|0;
 $init_data = $init_data|0;
 $charmap = $charmap|0;
 $acmap = $acmap|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cmap$0 = 0, $error = 0, $or$cond = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 HEAP32[$error>>2] = 0;
 $0 = ($clazz|0)==(0|0);
 $1 = ($charmap|0)==(0|0);
 $or$cond = $0 | $1;
 if ($or$cond) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = HEAP32[$charmap>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ((($2)) + 100|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = HEAP32[$clazz>>2]|0;
 $7 = (_ft_mem_alloc($5,$6,$error)|0);
 $8 = HEAP32[$error>>2]|0;
 $9 = ($8|0)==(0);
 do {
  if ($9) {
   ;HEAP32[$7>>2]=HEAP32[$charmap>>2]|0;HEAP32[$7+4>>2]=HEAP32[$charmap+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$charmap+8>>2]|0;
   $10 = ((($7)) + 12|0);
   HEAP32[$10>>2] = $clazz;
   $11 = ((($clazz)) + 4|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($12|0)==(0|0);
   if ($13) {
    label = 6;
   } else {
    $14 = (FUNCTION_TABLE_iii[$12 & 8191]($7,$init_data)|0);
    HEAP32[$error>>2] = $14;
    $15 = ($14|0)==(0);
    if ($15) {
     label = 6;
    }
   }
   if ((label|0) == 6) {
    $16 = ((($2)) + 36|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = (($17) + 1)|0;
    $19 = ((($2)) + 40|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = (_ft_mem_realloc($5,4,$17,$18,$20,$error)|0);
    HEAP32[$19>>2] = $21;
    $22 = HEAP32[$error>>2]|0;
    $23 = ($22|0)==(0);
    if ($23) {
     $24 = HEAP32[$16>>2]|0;
     $25 = (($24) + 1)|0;
     HEAP32[$16>>2] = $25;
     $26 = (($21) + ($24<<2)|0);
     HEAP32[$26>>2] = $7;
     $cmap$0 = $7;
     break;
    }
   }
   $29 = HEAP32[$10>>2]|0;
   $30 = HEAP32[$7>>2]|0;
   $31 = ((($30)) + 100|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = ((($29)) + 8|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = ($34|0)==(0|0);
   if (!($35)) {
    FUNCTION_TABLE_vi[$34 & 8191]($7);
   }
   _ft_mem_free($32,$7);
   $cmap$0 = 0;
  } else {
   $cmap$0 = $7;
  }
 } while(0);
 $27 = ($acmap|0)==(0|0);
 if (!($27)) {
  HEAP32[$acmap>>2] = $cmap$0;
 }
 $28 = HEAP32[$error>>2]|0;
 $$0 = $28;
 STACKTOP = sp;return ($$0|0);
}
function _FT_Get_Next_Char($face,$charcode,$agindex) {
 $face = $face|0;
 $charcode = $charcode|0;
 $agindex = $agindex|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $code = 0;
 var $gindex$0 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $code = sp;
 $0 = ($face|0)==(0|0);
 if ($0) {
  $gindex$0 = 0;$result$0 = 0;
 } else {
  $1 = ((($face)) + 92|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ($2|0)==(0|0);
  if ($3) {
   $gindex$0 = 0;$result$0 = 0;
  } else {
   $4 = ((($face)) + 16|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = ($5|0)==(0);
   if ($6) {
    $gindex$0 = 0;$result$0 = 0;
   } else {
    HEAP32[$code>>2] = $charcode;
    $7 = ((($2)) + 12|0);
    while(1) {
     $8 = HEAP32[$7>>2]|0;
     $9 = ((($8)) + 16|0);
     $10 = HEAP32[$9>>2]|0;
     $11 = (FUNCTION_TABLE_iii[$10 & 8191]($2,$code)|0);
     $12 = HEAP32[$4>>2]|0;
     $13 = ($11>>>0)<($12>>>0);
     if ($13) {
      $$lcssa = $11;
      break;
     }
    }
    $14 = ($$lcssa|0)==(0);
    $15 = HEAP32[$code>>2]|0;
    $16 = $14 ? 0 : $15;
    $gindex$0 = $$lcssa;$result$0 = $16;
   }
  }
 }
 $17 = ($agindex|0)==(0|0);
 if ($17) {
  STACKTOP = sp;return ($result$0|0);
 }
 HEAP32[$agindex>>2] = $gindex$0;
 STACKTOP = sp;return ($result$0|0);
}
function _FT_Get_Glyph_Name($face,$glyph_index,$buffer,$buffer_max) {
 $face = $face|0;
 $glyph_index = $glyph_index|0;
 $buffer = $buffer|0;
 $buffer_max = $buffer_max|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $magicptr = 0, $or$cond = 0, $svc$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($face|0)==(0|0);
 if ($0) {
  $$0 = 35;
  return ($$0|0);
 }
 $1 = ($buffer|0)==(0|0);
 $2 = ($buffer_max|0)==(0);
 $or$cond = $1 | $2;
 if ($or$cond) {
  $$0 = 6;
  return ($$0|0);
 }
 HEAP8[$buffer>>0] = 0;
 $3 = ((($face)) + 16|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)>($glyph_index|0);
 if (!($5)) {
  $$0 = 16;
  return ($$0|0);
 }
 $6 = ((($face)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 & 512;
 $9 = ($8|0)==(0);
 if ($9) {
  $$0 = 6;
  return ($$0|0);
 }
 $10 = ((($face)) + 128|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($11)) + 36|0);
 $13 = HEAP32[$12>>2]|0;
 $magicptr = $13;
 switch ($magicptr|0) {
 case 0:  {
  label = 6;
  break;
 }
 case -2:  {
  $$0 = 6;
  return ($$0|0);
  break;
 }
 default: {
  $svc$03 = $13;
 }
 }
 do {
  if ((label|0) == 6) {
   $14 = ((($face)) + 96|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = HEAP32[$15>>2]|0;
   $17 = ((($16)) + 32|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = ($18|0)==(0|0);
   if ($19) {
    HEAP32[$12>>2] = (-2);
    $$0 = 6;
    return ($$0|0);
   } else {
    $20 = (FUNCTION_TABLE_iii[$18 & 8191]($15,594305)|0);
    $21 = ($20|0)!=(0|0);
    $22 = $21 ? $20 : (-2);
    $23 = HEAP32[$10>>2]|0;
    $24 = ((($23)) + 36|0);
    HEAP32[$24>>2] = $22;
    $25 = ($20|0)==(0|0);
    if ($25) {
     $$0 = 6;
    } else {
     $svc$03 = $20;
     break;
    }
    return ($$0|0);
   }
  }
 } while(0);
 $26 = HEAP32[$svc$03>>2]|0;
 $27 = ($26|0)==(0|0);
 if ($27) {
  $$0 = 6;
  return ($$0|0);
 }
 $28 = (FUNCTION_TABLE_iiiii[$26 & 8191]($face,$glyph_index,$buffer,$buffer_max)|0);
 $$0 = $28;
 return ($$0|0);
}
function _FT_Activate_Size($size) {
 $size = $size|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($size|0)==(0|0);
 if ($0) {
  $$0 = 36;
 } else {
  $1 = HEAP32[$size>>2]|0;
  $2 = ($1|0)==(0|0);
  if ($2) {
   $$0 = 35;
  } else {
   $3 = ((($1)) + 96|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==(0|0);
   if ($5) {
    $$0 = 35;
   } else {
    $6 = ((($1)) + 88|0);
    HEAP32[$6>>2] = $size;
    $$0 = 0;
   }
  }
 }
 return ($$0|0);
}
function _FT_Lookup_Renderer($library,$format,$node) {
 $library = $library|0;
 $format = $format|0;
 $node = $node|0;
 var $$lcssa = 0, $$old1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cur$0 = 0, $cur$1 = 0;
 var $cur$2 = 0, $cur$2$lcssa = 0, $result$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($library|0)==(0|0);
 if ($0) {
  $result$2 = 0;
  return ($result$2|0);
 }
 $1 = ((($library)) + 148|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($node|0)!=(0|0);
 if ($3) {
  $4 = HEAP32[$node>>2]|0;
  $5 = ($4|0)==(0|0);
  if ($5) {
   $cur$0 = $2;
  } else {
   $6 = ((($4)) + 4|0);
   $7 = HEAP32[$6>>2]|0;
   $cur$0 = $7;
  }
  HEAP32[$node>>2] = 0;
  $cur$1 = $cur$0;
 } else {
  $cur$1 = $2;
 }
 $$old1 = ($cur$1|0)==(0|0);
 if ($$old1) {
  $result$2 = 0;
  return ($result$2|0);
 } else {
  $cur$2 = $cur$1;
 }
 while(1) {
  $8 = ((($cur$2)) + 8|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($9)) + 16|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($11|0)==($format|0);
  if ($12) {
   $$lcssa = $9;$cur$2$lcssa = $cur$2;
   break;
  }
  $13 = ((($cur$2)) + 4|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ($14|0)==(0|0);
  if ($15) {
   $result$2 = 0;
   label = 11;
   break;
  } else {
   $cur$2 = $14;
  }
 }
 if ((label|0) == 11) {
  return ($result$2|0);
 }
 if (!($3)) {
  $result$2 = $$lcssa;
  return ($result$2|0);
 }
 HEAP32[$node>>2] = $cur$2$lcssa;
 $result$2 = $$lcssa;
 return ($result$2|0);
}
function _FT_Add_Module($library,$clazz) {
 $library = $library|0;
 $clazz = $clazz|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa20 = 0, $$old1$i$i$i = 0, $$pr$i = 0, $$pre = 0, $$pre$i = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $cur$2$i$i$i = 0, $error = 0, $error$i = 0, $nn$05 = 0, $result$2$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp + 4|0;
 $error = sp;
 $0 = ($library|0)==(0|0);
 if ($0) {
  $$0 = 33;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ($clazz|0)==(0|0);
 if ($1) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ((($clazz)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)>(131078);
 if ($4) {
  $$0 = 4;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = ((($library)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if ($7) {
  $8 = HEAP32[$library>>2]|0;
  HEAP32[$error>>2] = 0;
  $30 = $8;$40 = $library;
 } else {
  $9 = ((($clazz)) + 8|0);
  $10 = HEAP32[$9>>2]|0;
  $nn$05 = 0;
  while(1) {
   $13 = (((($library)) + 20|0) + ($nn$05<<2)|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = HEAP32[$14>>2]|0;
   $16 = ((($15)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = (_strcmp($17,$10)|0);
   $19 = ($18|0)==(0);
   $12 = (($nn$05) + 1)|0;
   if ($19) {
    $$lcssa = $14;$$lcssa20 = $15;
    label = 9;
    break;
   }
   $11 = ($12>>>0)<($6>>>0);
   if ($11) {
    $nn$05 = $12;
   } else {
    $27 = $6;
    break;
   }
  }
  do {
   if ((label|0) == 9) {
    $20 = ((($clazz)) + 12|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = ((($$lcssa20)) + 12|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = ($21|0)>($23|0);
    if ($24) {
     (_FT_Remove_Module($library,$$lcssa)|0);
     $$pre = HEAP32[$5>>2]|0;
     $27 = $$pre;
     break;
    } else {
     $$0 = 5;
     STACKTOP = sp;return ($$0|0);
    }
   }
  } while(0);
  $25 = HEAP32[$library>>2]|0;
  HEAP32[$error>>2] = 0;
  $26 = ($27>>>0)>(31);
  if ($26) {
   HEAP32[$error>>2] = 48;
   $$0 = 48;
   STACKTOP = sp;return ($$0|0);
  } else {
   $30 = $25;$40 = $library;
  }
 }
 $28 = ((($clazz)) + 4|0);
 $29 = HEAP32[$28>>2]|0;
 $31 = (_ft_mem_alloc($30,$29,$error)|0);
 $32 = HEAP32[$error>>2]|0;
 $33 = ($32|0)==(0);
 if (!($33)) {
  $$0 = $32;
  STACKTOP = sp;return ($$0|0);
 }
 $34 = ((($31)) + 4|0);
 HEAP32[$34>>2] = $library;
 $35 = ((($31)) + 8|0);
 HEAP32[$35>>2] = $30;
 HEAP32[$31>>2] = $clazz;
 $36 = HEAP32[$clazz>>2]|0;
 $37 = $36 & 2;
 $38 = ($37|0)==(0);
 do {
  if ($38) {
   label = 27;
  } else {
   $39 = HEAP32[$40>>2]|0;
   $41 = (_ft_mem_alloc($39,12,$error$i)|0);
   $42 = HEAP32[$error$i>>2]|0;
   $43 = ($42|0)==(0);
   if (!($43)) {
    HEAP32[$error>>2] = $42;
    break;
   }
   $44 = HEAP32[$31>>2]|0;
   $45 = ((($31)) + 12|0);
   HEAP32[$45>>2] = $44;
   $46 = ((($44)) + 36|0);
   $47 = HEAP32[$46>>2]|0;
   $48 = ((($31)) + 16|0);
   HEAP32[$48>>2] = $47;
   $49 = ($47|0)==(1869968492);
   if ($49) {
    $50 = ((($44)) + 56|0);
    $51 = HEAP32[$50>>2]|0;
    $52 = ((($51)) + 4|0);
    $53 = HEAP32[$52>>2]|0;
    $54 = ($53|0)==(0|0);
    if ($54) {
     label = 21;
    } else {
     $55 = ((($31)) + 52|0);
     $56 = (FUNCTION_TABLE_iii[$53 & 8191]($39,$55)|0);
     HEAP32[$error$i>>2] = $56;
     $57 = ($56|0)==(0);
     if ($57) {
      $58 = HEAP32[$50>>2]|0;
      $59 = ((($58)) + 16|0);
      $60 = HEAP32[$59>>2]|0;
      $61 = ((($31)) + 56|0);
      HEAP32[$61>>2] = $60;
      $62 = ((($44)) + 40|0);
      $63 = HEAP32[$62>>2]|0;
      $64 = ((($31)) + 60|0);
      HEAP32[$64>>2] = $63;
      label = 21;
     }
    }
   } else {
    label = 21;
   }
   if ((label|0) == 21) {
    $65 = ((($41)) + 8|0);
    HEAP32[$65>>2] = $31;
    $66 = ((($library)) + 148|0);
    _FT_List_Add($66,$41);
    $67 = ((($library)) + 148|0);
    $68 = HEAP32[$67>>2]|0;
    $$old1$i$i$i = ($68|0)==(0|0);
    L39: do {
     if ($$old1$i$i$i) {
      $result$2$i$i$i = 0;
     } else {
      $cur$2$i$i$i = $68;
      while(1) {
       $69 = ((($cur$2$i$i$i)) + 8|0);
       $70 = HEAP32[$69>>2]|0;
       $71 = ((($70)) + 16|0);
       $72 = HEAP32[$71>>2]|0;
       $73 = ($72|0)==(1869968492);
       if ($73) {
        $result$2$i$i$i = $70;
        break L39;
       }
       $74 = ((($cur$2$i$i$i)) + 4|0);
       $75 = HEAP32[$74>>2]|0;
       $76 = ($75|0)==(0|0);
       if ($76) {
        $result$2$i$i$i = 0;
        break;
       } else {
        $cur$2$i$i$i = $75;
       }
      }
     }
    } while(0);
    $77 = ((($library)) + 156|0);
    HEAP32[$77>>2] = $result$2$i$i$i;
    $$pr$i = HEAP32[$error$i>>2]|0;
    $78 = ($$pr$i|0)==(0);
    if ($78) {
     HEAP32[$error>>2] = 0;
     label = 27;
     break;
    }
   }
   _ft_mem_free($39,$41);
   $$pre$i = HEAP32[$error$i>>2]|0;
   HEAP32[$error>>2] = $$pre$i;
   $79 = ($$pre$i|0)==(0);
   if ($79) {
    label = 27;
   }
  }
 } while(0);
 do {
  if ((label|0) == 27) {
   $80 = HEAP32[$31>>2]|0;
   $81 = HEAP32[$80>>2]|0;
   $82 = $81 & 4;
   $83 = ($82|0)==(0);
   if (!($83)) {
    $84 = ((($library)) + 160|0);
    HEAP32[$84>>2] = $31;
   }
   $85 = $81 & 1;
   $86 = ($85|0)==(0);
   if (!($86)) {
    $87 = ((($31)) + 12|0);
    HEAP32[$87>>2] = $80;
   }
   $88 = ((($clazz)) + 24|0);
   $89 = HEAP32[$88>>2]|0;
   $90 = ($89|0)==(0|0);
   if (!($90)) {
    $91 = (FUNCTION_TABLE_ii[$89 & 8191]($31)|0);
    HEAP32[$error>>2] = $91;
    $92 = ($91|0)==(0);
    if (!($92)) {
     break;
    }
   }
   $93 = HEAP32[$5>>2]|0;
   $94 = (($93) + 1)|0;
   HEAP32[$5>>2] = $94;
   $95 = (((($library)) + 20|0) + ($93<<2)|0);
   HEAP32[$95>>2] = $31;
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 $96 = HEAP32[$31>>2]|0;
 $97 = HEAP32[$96>>2]|0;
 $98 = $97 & 2;
 $99 = ($98|0)==(0);
 if (!($99)) {
  $100 = ((($31)) + 12|0);
  $101 = HEAP32[$100>>2]|0;
  $102 = ($101|0)==(0|0);
  if (!($102)) {
   $103 = ((($101)) + 36|0);
   $104 = HEAP32[$103>>2]|0;
   $105 = ($104|0)==(1869968492);
   if ($105) {
    $106 = ((($31)) + 52|0);
    $107 = HEAP32[$106>>2]|0;
    $108 = ($107|0)==(0|0);
    if (!($108)) {
     $109 = ((($101)) + 56|0);
     $110 = HEAP32[$109>>2]|0;
     $111 = ((($110)) + 20|0);
     $112 = HEAP32[$111>>2]|0;
     FUNCTION_TABLE_vi[$112 & 8191]($107);
    }
   }
  }
 }
 _ft_mem_free($30,$31);
 $$pre9 = HEAP32[$error>>2]|0;
 $$0 = $$pre9;
 STACKTOP = sp;return ($$0|0);
}
function _FT_Remove_Module($library,$module) {
 $library = $library|0;
 $module = $module|0;
 var $$1 = 0, $$old1$i$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $cur$05 = 0, $cur$05$lcssa = 0, $cur$14 = 0, $cur$2$i$i$i$i = 0, $result$2$i$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($library|0)==(0|0);
 if ($0) {
  $$1 = 33;
  return ($$1|0);
 }
 $1 = ($module|0)==(0|0);
 if ($1) {
  $$1 = 34;
  return ($$1|0);
 }
 $2 = ((($library)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (((($library)) + 20|0) + ($3<<2)|0);
 $5 = ($3|0)>(0);
 if (!($5)) {
  $$1 = 34;
  return ($$1|0);
 }
 $6 = ((($library)) + 20|0);
 $cur$05 = $6;
 while(1) {
  $7 = HEAP32[$cur$05>>2]|0;
  $8 = ($7|0)==($module|0);
  if ($8) {
   $cur$05$lcssa = $cur$05;
   break;
  }
  $63 = ((($cur$05)) + 4|0);
  $64 = ($63>>>0)<($4>>>0);
  if ($64) {
   $cur$05 = $63;
  } else {
   $$1 = 34;
   label = 26;
   break;
  }
 }
 if ((label|0) == 26) {
  return ($$1|0);
 }
 $9 = (($3) + -1)|0;
 HEAP32[$2>>2] = $9;
 $10 = ((($4)) + -4|0);
 $11 = ($cur$05$lcssa>>>0)<($10>>>0);
 if ($11) {
  $cur$14 = $cur$05$lcssa;
  while(1) {
   $12 = ((($cur$14)) + 4|0);
   $13 = HEAP32[$12>>2]|0;
   HEAP32[$cur$14>>2] = $13;
   $14 = ($12>>>0)<($10>>>0);
   if ($14) {
    $cur$14 = $12;
   } else {
    break;
   }
  }
 }
 HEAP32[$10>>2] = 0;
 $15 = ((($module)) + 8|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = HEAP32[$module>>2]|0;
 $18 = ((($module)) + 4|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($19|0)==(0|0);
 if (!($20)) {
  $21 = ((($19)) + 160|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = ($22|0)==($module|0);
  if ($23) {
   HEAP32[$21>>2] = 0;
  }
  $24 = HEAP32[$17>>2]|0;
  $25 = $24 & 2;
  $26 = ($25|0)==(0);
  if (!($26)) {
   $27 = HEAP32[$19>>2]|0;
   $28 = ((($19)) + 148|0);
   $29 = (_FT_List_Find($28,$module)|0);
   $30 = ($29|0)==(0|0);
   if (!($30)) {
    $31 = ((($module)) + 12|0);
    $32 = HEAP32[$31>>2]|0;
    $33 = ((($32)) + 36|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = ($34|0)==(1869968492);
    if ($35) {
     $36 = ((($module)) + 52|0);
     $37 = HEAP32[$36>>2]|0;
     $38 = ($37|0)==(0|0);
     if (!($38)) {
      $39 = ((($32)) + 56|0);
      $40 = HEAP32[$39>>2]|0;
      $41 = ((($40)) + 20|0);
      $42 = HEAP32[$41>>2]|0;
      FUNCTION_TABLE_vi[$42 & 8191]($37);
     }
    }
    _FT_List_Remove($28,$29);
    _ft_mem_free($27,$29);
    $43 = ((($19)) + 148|0);
    $44 = HEAP32[$43>>2]|0;
    $$old1$i$i$i$i = ($44|0)==(0|0);
    L31: do {
     if ($$old1$i$i$i$i) {
      $result$2$i$i$i$i = 0;
     } else {
      $cur$2$i$i$i$i = $44;
      while(1) {
       $45 = ((($cur$2$i$i$i$i)) + 8|0);
       $46 = HEAP32[$45>>2]|0;
       $47 = ((($46)) + 16|0);
       $48 = HEAP32[$47>>2]|0;
       $49 = ($48|0)==(1869968492);
       if ($49) {
        $result$2$i$i$i$i = $46;
        break L31;
       }
       $50 = ((($cur$2$i$i$i$i)) + 4|0);
       $51 = HEAP32[$50>>2]|0;
       $52 = ($51|0)==(0|0);
       if ($52) {
        $result$2$i$i$i$i = 0;
        break;
       } else {
        $cur$2$i$i$i$i = $51;
       }
      }
     }
    } while(0);
    $53 = ((($19)) + 156|0);
    HEAP32[$53>>2] = $result$2$i$i$i$i;
   }
  }
 }
 $54 = HEAP32[$module>>2]|0;
 $55 = HEAP32[$54>>2]|0;
 $56 = $55 & 1;
 $57 = ($56|0)==(0);
 if (!($57)) {
  $58 = ((($module)) + 16|0);
  $59 = HEAP32[$15>>2]|0;
  _FT_List_Finalize($58,4345,$59,$module);
 }
 $60 = ((($17)) + 28|0);
 $61 = HEAP32[$60>>2]|0;
 $62 = ($61|0)==(0|0);
 if (!($62)) {
  FUNCTION_TABLE_vi[$61 & 8191]($module);
 }
 _ft_mem_free($16,$module);
 $$1 = 0;
 return ($$1|0);
}
function _FT_Get_Module_Interface($library,$mod_name) {
 $library = $library|0;
 $mod_name = $mod_name|0;
 var $$lcssa = 0, $$lcssa13 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $cur$02$i = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($library|0)!=(0|0);
 $1 = ($mod_name|0)!=(0|0);
 $or$cond$i = $0 & $1;
 if (!($or$cond$i)) {
  $18 = 0;
  return ($18|0);
 }
 $2 = ((($library)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (((($library)) + 20|0) + ($3<<2)|0);
 $5 = ($3|0)>(0);
 if (!($5)) {
  $18 = 0;
  return ($18|0);
 }
 $6 = ((($library)) + 20|0);
 $cur$02$i = $6;
 while(1) {
  $9 = HEAP32[$cur$02$i>>2]|0;
  $10 = HEAP32[$9>>2]|0;
  $11 = ((($10)) + 8|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = (_strcmp($12,$mod_name)|0);
  $14 = ($13|0)==(0);
  $8 = ((($cur$02$i)) + 4|0);
  if ($14) {
   $$lcssa = $9;$$lcssa13 = $10;
   break;
  }
  $7 = ($8>>>0)<($4>>>0);
  if ($7) {
   $cur$02$i = $8;
  } else {
   $18 = 0;
   label = 8;
   break;
  }
 }
 if ((label|0) == 8) {
  return ($18|0);
 }
 $15 = ($$lcssa|0)==(0|0);
 if ($15) {
  $18 = 0;
  return ($18|0);
 }
 $16 = ((($$lcssa13)) + 20|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17;
 return ($18|0);
}
function _ft_module_get_service($module,$service_id) {
 $module = $module|0;
 $service_id = $service_id|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cur$02 = 0, $result$4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($module|0)==(0|0);
 if ($0) {
  $result$4 = 0;
  return ($result$4|0);
 }
 $1 = HEAP32[$module>>2]|0;
 $2 = ((($1)) + 32|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  $5 = (FUNCTION_TABLE_iii[$3 & 8191]($module,$service_id)|0);
  $6 = ($5|0)==(0|0);
  if (!($6)) {
   $result$4 = $5;
   return ($result$4|0);
  }
 }
 $7 = ((($module)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($8)) + 16|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (((($8)) + 20|0) + ($10<<2)|0);
 $12 = ($10|0)>(0);
 if (!($12)) {
  $result$4 = 0;
  return ($result$4|0);
 }
 $13 = ((($8)) + 20|0);
 $cur$02 = $13;
 while(1) {
  $14 = HEAP32[$cur$02>>2]|0;
  $15 = ($14|0)==($module|0);
  if (!($15)) {
   $16 = HEAP32[$14>>2]|0;
   $17 = ((($16)) + 32|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = ($18|0)==(0|0);
   if (!($19)) {
    $20 = (FUNCTION_TABLE_iii[$18 & 8191]($14,$service_id)|0);
    $21 = ($20|0)==(0|0);
    if (!($21)) {
     $result$4 = $20;
     label = 10;
     break;
    }
   }
  }
  $22 = ((($cur$02)) + 4|0);
  $23 = ($22>>>0)<($11>>>0);
  if ($23) {
   $cur$02 = $22;
  } else {
   $result$4 = 0;
   label = 10;
   break;
  }
 }
 if ((label|0) == 10) {
  return ($result$4|0);
 }
 return (0)|0;
}
function _FT_New_Library($memory,$alibrary) {
 $memory = $memory|0;
 $alibrary = $alibrary|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ($memory|0)!=(0|0);
 $1 = ($alibrary|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = (_ft_mem_alloc($memory,192,$error)|0);
 $3 = HEAP32[$error>>2]|0;
 $4 = ($3|0)==(0);
 if (!($4)) {
  $$0 = $3;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$2>>2] = $memory;
 $5 = ((($2)) + 168|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($2)) + 164|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($2)) + 4|0);
 HEAP32[$7>>2] = 2;
 $8 = ((($2)) + 8|0);
 HEAP32[$8>>2] = 6;
 $9 = ((($2)) + 12|0);
 HEAP32[$9>>2] = 0;
 $10 = ((($2)) + 188|0);
 HEAP32[$10>>2] = 1;
 HEAP32[$alibrary>>2] = $2;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _FT_Done_Library($library) {
 $library = $library|0;
 var $$0 = 0, $$lcssa23 = 0, $$pre = 0, $$pre11 = 0, $$pre12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $exitcond = 0, $m$08 = 0, $n$05 = 0, $n$05$us = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($library|0)==(0|0);
 if ($0) {
  $$0 = 33;
  return ($$0|0);
 }
 $1 = ((($library)) + 188|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($2) + -1)|0;
 HEAP32[$1>>2] = $3;
 $4 = ($2|0)>(1);
 if ($4) {
  $$0 = 0;
  return ($$0|0);
 }
 $5 = HEAP32[$library>>2]|0;
 $6 = ((($library)) + 16|0);
 $$pre = HEAP32[$6>>2]|0;
 $8 = $$pre;$87 = $$pre;$m$08 = 0;
 while(1) {
  $7 = ($8|0)==(0);
  L9: do {
   if ($7) {
    $88 = $87;$89 = 0;
   } else {
    $9 = ($m$08|0)==(0);
    if (!($9)) {
     $n$05 = 0;
     while(1) {
      $48 = (((($library)) + 20|0) + ($n$05<<2)|0);
      $49 = HEAP32[$48>>2]|0;
      $50 = HEAP32[$49>>2]|0;
      $51 = HEAP32[$50>>2]|0;
      $52 = $51 & 1;
      $53 = ($52|0)==(0);
      if (!($53)) {
       $54 = ((($49)) + 16|0);
       $55 = HEAP32[$54>>2]|0;
       $56 = ($55|0)==(0|0);
       if (!($56)) {
        $58 = $55;
        while(1) {
         $57 = ((($58)) + 8|0);
         $59 = HEAP32[$57>>2]|0;
         $60 = ($59|0)==(0|0);
         if (!($60)) {
          $63 = ((($59)) + 96|0);
          $64 = HEAP32[$63>>2]|0;
          $65 = ($64|0)==(0|0);
          if (!($65)) {
           $66 = ((($59)) + 128|0);
           $67 = HEAP32[$66>>2]|0;
           $68 = ((($67)) + 56|0);
           $69 = HEAP32[$68>>2]|0;
           $70 = (($69) + -1)|0;
           HEAP32[$68>>2] = $70;
           $71 = ($69|0)>(1);
           if (!($71)) {
            $72 = ((($64)) + 8|0);
            $73 = HEAP32[$72>>2]|0;
            $74 = ((($64)) + 16|0);
            $75 = (_FT_List_Find($74,$59)|0);
            $76 = ($75|0)==(0|0);
            if (!($76)) {
             _FT_List_Remove($74,$75);
             _ft_mem_free($73,$75);
             _destroy_face($73,$59,$64);
            }
           }
          }
         }
         $61 = HEAP32[$54>>2]|0;
         $62 = ($61|0)==(0|0);
         if ($62) {
          break;
         } else {
          $58 = $61;
         }
        }
       }
      }
      $77 = (($n$05) + 1)|0;
      $78 = HEAP32[$6>>2]|0;
      $79 = ($77>>>0)<($78>>>0);
      if ($79) {
       $n$05 = $77;
      } else {
       $88 = $78;$89 = $78;
       break L9;
      }
     }
    }
    $10 = (47792 + ($m$08<<2)|0);
    $$pre11 = HEAP32[$10>>2]|0;
    $90 = $87;$n$05$us = 0;
    while(1) {
     $11 = (((($library)) + 20|0) + ($n$05$us<<2)|0);
     $12 = HEAP32[$11>>2]|0;
     $13 = HEAP32[$12>>2]|0;
     $14 = ((($13)) + 8|0);
     $15 = HEAP32[$14>>2]|0;
     $16 = (_strcmp($15,$$pre11)|0);
     $17 = ($16|0)==(0);
     if ($17) {
      $18 = HEAP32[$13>>2]|0;
      $19 = $18 & 1;
      $20 = ($19|0)==(0);
      if ($20) {
       $44 = $90;
      } else {
       $21 = ((($12)) + 16|0);
       $22 = HEAP32[$21>>2]|0;
       $23 = ($22|0)==(0|0);
       if ($23) {
        $44 = $90;
       } else {
        $25 = $22;
        while(1) {
         $24 = ((($25)) + 8|0);
         $26 = HEAP32[$24>>2]|0;
         $27 = ($26|0)==(0|0);
         if (!($27)) {
          $28 = ((($26)) + 96|0);
          $29 = HEAP32[$28>>2]|0;
          $30 = ($29|0)==(0|0);
          if (!($30)) {
           $31 = ((($26)) + 128|0);
           $32 = HEAP32[$31>>2]|0;
           $33 = ((($32)) + 56|0);
           $34 = HEAP32[$33>>2]|0;
           $35 = (($34) + -1)|0;
           HEAP32[$33>>2] = $35;
           $36 = ($34|0)>(1);
           if (!($36)) {
            $37 = ((($29)) + 8|0);
            $38 = HEAP32[$37>>2]|0;
            $39 = ((($29)) + 16|0);
            $40 = (_FT_List_Find($39,$26)|0);
            $41 = ($40|0)==(0|0);
            if (!($41)) {
             _FT_List_Remove($39,$40);
             _ft_mem_free($38,$40);
             _destroy_face($38,$26,$29);
            }
           }
          }
         }
         $45 = HEAP32[$21>>2]|0;
         $46 = ($45|0)==(0|0);
         if ($46) {
          break;
         } else {
          $25 = $45;
         }
        }
        $$pre12 = HEAP32[$6>>2]|0;
        $44 = $$pre12;
       }
      }
     } else {
      $44 = $90;
     }
     $42 = (($n$05$us) + 1)|0;
     $43 = ($42>>>0)<($44>>>0);
     if ($43) {
      $90 = $44;$n$05$us = $42;
     } else {
      $88 = $44;$89 = $44;
      break;
     }
    }
   }
  } while(0);
  $80 = (($m$08) + 1)|0;
  $exitcond = ($80|0)==(2);
  if ($exitcond) {
   $$lcssa23 = $88;
   break;
  } else {
   $8 = $89;$87 = $88;$m$08 = $80;
  }
 }
 $47 = ($$lcssa23|0)==(0);
 if (!($47)) {
  $82 = $$lcssa23;
  while(1) {
   $81 = (($82) + -1)|0;
   $83 = (((($library)) + 20|0) + ($81<<2)|0);
   $84 = HEAP32[$83>>2]|0;
   (_FT_Remove_Module($library,$84)|0);
   $85 = HEAP32[$6>>2]|0;
   $86 = ($85|0)==(0);
   if ($86) {
    break;
   } else {
    $82 = $85;
   }
  }
 }
 _ft_mem_free($5,$library);
 $$0 = 0;
 return ($$0|0);
}
function _FT_Outline_Decompose($outline,$func_interface,$user) {
 $outline = $outline|0;
 $func_interface = $func_interface|0;
 $user = $user|0;
 var $$0 = 0, $$0$ph = 0, $$1$ph = 0, $$4 = 0, $$7 = 0, $$first$0 = 0, $$lcssa70 = 0, $$lcssa72 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $error$3$ph = 0, $error$5 = 0, $error$6 = 0, $first$036 = 0;
 var $first$1 = 0, $limit$0 = 0, $limit$1 = 0, $n$035 = 0, $point$0 = 0, $point$1$be = 0, $point$130 = 0, $point$226 = 0, $tags$0 = 0, $tags$1$be = 0, $tags$131 = 0, $tags$227 = 0, $v_control = 0, $v_middle = 0, $v_start = 0, $vec = 0, $vec1 = 0, $vec13 = 0, $vec2 = 0, $vec4 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v_control = sp + 16|0;
 $v_start = sp + 8|0;
 $vec = sp + 56|0;
 $vec1 = sp;
 $v_middle = sp + 48|0;
 $vec13 = sp + 40|0;
 $vec2 = sp + 32|0;
 $vec4 = sp + 24|0;
 $0 = ($outline|0)==(0|0);
 if ($0) {
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ($func_interface|0)==(0|0);
 if ($1) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ((($func_interface)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($func_interface)) + 20|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = HEAP16[$outline>>1]|0;
 $7 = ($6<<16>>16)>(0);
 if (!($7)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $8 = ((($outline)) + 12|0);
 $9 = ((($outline)) + 4|0);
 $10 = ((($v_start)) + 4|0);
 $11 = ((($outline)) + 8|0);
 $12 = ((($func_interface)) + 4|0);
 $13 = ((($vec13)) + 4|0);
 $14 = ((($vec2)) + 4|0);
 $15 = ((($vec4)) + 4|0);
 $16 = ((($func_interface)) + 12|0);
 $17 = ((($vec)) + 4|0);
 $18 = ((($v_control)) + 4|0);
 $19 = ((($vec1)) + 4|0);
 $20 = ((($v_middle)) + 4|0);
 $21 = ((($func_interface)) + 8|0);
 $first$036 = 0;$n$035 = 0;
 L10: while(1) {
  $22 = HEAP32[$8>>2]|0;
  $23 = (($22) + ($n$035<<1)|0);
  $24 = HEAP16[$23>>1]|0;
  $25 = $24 << 16 >> 16;
  $26 = ($24<<16>>16)<(0);
  if ($26) {
   $$0 = 20;
   label = 34;
   break;
  }
  $27 = HEAP32[$9>>2]|0;
  $28 = (($27) + ($25<<3)|0);
  $29 = (($27) + ($first$036<<3)|0);
  $30 = $29;
  $31 = $30;
  $32 = HEAP32[$31>>2]|0;
  $33 = (($30) + 4)|0;
  $34 = $33;
  $35 = HEAP32[$34>>2]|0;
  $36 = $v_start;
  $37 = $36;
  HEAP32[$37>>2] = $32;
  $38 = (($36) + 4)|0;
  $39 = $38;
  HEAP32[$39>>2] = $35;
  $40 = $32 << $3;
  $41 = (($40) - ($5))|0;
  HEAP32[$v_start>>2] = $41;
  $42 = $35 << $3;
  $43 = (($42) - ($5))|0;
  HEAP32[$10>>2] = $43;
  $44 = (($27) + ($25<<3)|0);
  $45 = HEAP32[$44>>2]|0;
  $46 = (((($27) + ($25<<3)|0)) + 4|0);
  $47 = HEAP32[$46>>2]|0;
  $48 = $45 << $3;
  $49 = (($48) - ($5))|0;
  $50 = $47 << $3;
  $51 = (($50) - ($5))|0;
  $52 = $v_start;
  $53 = $52;
  $54 = HEAP32[$53>>2]|0;
  $55 = (($52) + 4)|0;
  $56 = $55;
  $57 = HEAP32[$56>>2]|0;
  $58 = $v_control;
  $59 = $58;
  HEAP32[$59>>2] = $54;
  $60 = (($58) + 4)|0;
  $61 = $60;
  HEAP32[$61>>2] = $57;
  $62 = HEAP32[$11>>2]|0;
  $63 = (($62) + ($first$036)|0);
  $64 = HEAP8[$63>>0]|0;
  $65 = $64&255;
  $66 = $65 & 3;
  switch ($66|0) {
  case 2:  {
   $$0 = 20;
   label = 34;
   break L10;
   break;
  }
  case 0:  {
   $67 = (($62) + ($25)|0);
   $68 = HEAP8[$67>>0]|0;
   $69 = $68 & 3;
   $70 = ($69<<24>>24)==(1);
   if ($70) {
    HEAP32[$v_start>>2] = $49;
    HEAP32[$10>>2] = $51;
    $71 = ((($28)) + -8|0);
    $limit$0 = $71;
   } else {
    $72 = (($54) + ($49))|0;
    $73 = (($72|0) / 2)&-1;
    HEAP32[$v_start>>2] = $73;
    $74 = (($43) + ($51))|0;
    $75 = (($74|0) / 2)&-1;
    HEAP32[$10>>2] = $75;
    $limit$0 = $28;
   }
   $76 = ((($29)) + -8|0);
   $77 = ((($63)) + -1|0);
   $limit$1 = $limit$0;$point$0 = $76;$tags$0 = $77;
   break;
  }
  default: {
   $limit$1 = $28;$point$0 = $29;$tags$0 = $63;
  }
  }
  $78 = HEAP32[$func_interface>>2]|0;
  $79 = (FUNCTION_TABLE_iii[$78 & 8191]($v_start,$user)|0);
  $80 = ($79|0)==(0);
  if (!($80)) {
   $$0$ph = $79;
   label = 33;
   break;
  }
  $81 = ($point$0>>>0)<($limit$1>>>0);
  L21: do {
   if ($81) {
    $point$130 = $point$0;$tags$131 = $tags$0;
    L22: while(1) {
     $82 = ((($point$130)) + 8|0);
     $83 = ((($tags$131)) + 1|0);
     $84 = HEAP8[$83>>0]|0;
     $85 = $84&255;
     $86 = $85 & 3;
     switch ($86|0) {
     case 1:  {
      $87 = HEAP32[$82>>2]|0;
      $88 = $87 << $3;
      $89 = (($88) - ($5))|0;
      HEAP32[$vec>>2] = $89;
      $90 = ((($point$130)) + 12|0);
      $91 = HEAP32[$90>>2]|0;
      $92 = $91 << $3;
      $93 = (($92) - ($5))|0;
      HEAP32[$17>>2] = $93;
      $94 = HEAP32[$12>>2]|0;
      $95 = (FUNCTION_TABLE_iii[$94 & 8191]($vec,$user)|0);
      $96 = ($95|0)==(0);
      if ($96) {
       $point$1$be = $82;$tags$1$be = $83;
      } else {
       $$0 = $95;
       label = 34;
       break L10;
      }
      break;
     }
     case 0:  {
      $98 = HEAP32[$82>>2]|0;
      $99 = $98 << $3;
      $100 = (($99) - ($5))|0;
      HEAP32[$v_control>>2] = $100;
      $101 = ((($point$130)) + 12|0);
      $102 = HEAP32[$101>>2]|0;
      $103 = $102 << $3;
      $104 = (($103) - ($5))|0;
      HEAP32[$18>>2] = $104;
      $105 = ($82>>>0)<($limit$1>>>0);
      if ($105) {
       $119 = $100;$122 = $104;$point$226 = $82;$tags$227 = $83;
      } else {
       label = 22;
       break L22;
      }
      L31: while(1) {
       $106 = ((($point$226)) + 8|0);
       $107 = ((($tags$227)) + 1|0);
       $108 = HEAP8[$107>>0]|0;
       $109 = $108&255;
       $110 = $109 & 3;
       $111 = HEAP32[$106>>2]|0;
       $112 = $111 << $3;
       $113 = (($112) - ($5))|0;
       HEAP32[$vec1>>2] = $113;
       $114 = ((($point$226)) + 12|0);
       $115 = HEAP32[$114>>2]|0;
       $116 = $115 << $3;
       $117 = (($116) - ($5))|0;
       HEAP32[$19>>2] = $117;
       switch ($110|0) {
       case 1:  {
        $$lcssa70 = $106;$$lcssa72 = $107;
        break L31;
        break;
       }
       case 0:  {
        break;
       }
       default: {
        $$1$ph = 7;$error$3$ph = 0;
        label = 20;
        break L22;
       }
       }
       $118 = (($113) + ($119))|0;
       $120 = (($118|0) / 2)&-1;
       HEAP32[$v_middle>>2] = $120;
       $121 = (($117) + ($122))|0;
       $123 = (($121|0) / 2)&-1;
       HEAP32[$20>>2] = $123;
       $124 = HEAP32[$21>>2]|0;
       $125 = (FUNCTION_TABLE_iiii[$124 & 8191]($v_control,$v_middle,$user)|0);
       $126 = ($125|0)==(0);
       if (!($126)) {
        $$1$ph = 10;$error$3$ph = $125;
        label = 20;
        break L22;
       }
       $127 = $vec1;
       $128 = $127;
       $129 = HEAP32[$128>>2]|0;
       $130 = (($127) + 4)|0;
       $131 = $130;
       $132 = HEAP32[$131>>2]|0;
       $133 = $v_control;
       $134 = $133;
       HEAP32[$134>>2] = $129;
       $135 = (($133) + 4)|0;
       $136 = $135;
       HEAP32[$136>>2] = $132;
       $137 = ($106>>>0)<($limit$1>>>0);
       if ($137) {
        $119 = $129;$122 = $132;$point$226 = $106;$tags$227 = $107;
       } else {
        label = 22;
        break L22;
       }
      }
      $138 = HEAP32[$21>>2]|0;
      $139 = (FUNCTION_TABLE_iiii[$138 & 8191]($v_control,$vec1,$user)|0);
      $140 = ($139|0)==(0);
      if ($140) {
       $point$1$be = $$lcssa70;$tags$1$be = $$lcssa72;
      } else {
       $$4 = 10;$error$6 = $139;$first$1 = $first$036;
       break L21;
      }
      break;
     }
     default: {
      $143 = ((($point$130)) + 16|0);
      $144 = ($143>>>0)>($limit$1>>>0);
      if ($144) {
       label = 27;
       break L10;
      }
      $145 = ((($tags$131)) + 2|0);
      $146 = HEAP8[$145>>0]|0;
      $147 = $146 & 3;
      $148 = ($147<<24>>24)==(2);
      if (!($148)) {
       label = 27;
       break L10;
      }
      $149 = ((($point$130)) + 24|0);
      $150 = ((($point$130)) + 8|0);
      $151 = HEAP32[$150>>2]|0;
      $152 = $151 << $3;
      $153 = (($152) - ($5))|0;
      HEAP32[$vec13>>2] = $153;
      $154 = ((($point$130)) + 12|0);
      $155 = HEAP32[$154>>2]|0;
      $156 = $155 << $3;
      $157 = (($156) - ($5))|0;
      HEAP32[$13>>2] = $157;
      $158 = ((($point$130)) + 16|0);
      $159 = HEAP32[$158>>2]|0;
      $160 = $159 << $3;
      $161 = (($160) - ($5))|0;
      HEAP32[$vec2>>2] = $161;
      $162 = ((($point$130)) + 20|0);
      $163 = HEAP32[$162>>2]|0;
      $164 = $163 << $3;
      $165 = (($164) - ($5))|0;
      HEAP32[$14>>2] = $165;
      $166 = ($149>>>0)>($limit$1>>>0);
      if ($166) {
       label = 26;
       break L22;
      }
      $169 = ((($tags$131)) + 3|0);
      $170 = HEAP32[$149>>2]|0;
      $171 = $170 << $3;
      $172 = (($171) - ($5))|0;
      HEAP32[$vec4>>2] = $172;
      $173 = ((($point$130)) + 28|0);
      $174 = HEAP32[$173>>2]|0;
      $175 = $174 << $3;
      $176 = (($175) - ($5))|0;
      HEAP32[$15>>2] = $176;
      $177 = HEAP32[$16>>2]|0;
      $178 = (FUNCTION_TABLE_iiiii[$177 & 8191]($vec13,$vec2,$vec4,$user)|0);
      $179 = ($178|0)==(0);
      if ($179) {
       $point$1$be = $149;$tags$1$be = $169;
      } else {
       $$4 = 10;$error$6 = $178;$first$1 = $first$036;
       break L21;
      }
     }
     }
     $97 = ($point$1$be>>>0)<($limit$1>>>0);
     if ($97) {
      $point$130 = $point$1$be;$tags$131 = $tags$1$be;
     } else {
      label = 29;
      break L21;
     }
    }
    if ((label|0) == 20) {
     label = 0;
     $$4 = $$1$ph;$error$6 = $error$3$ph;$first$1 = $first$036;
     break;
    }
    else if ((label|0) == 22) {
     label = 0;
     $141 = HEAP32[$21>>2]|0;
     $142 = (FUNCTION_TABLE_iiii[$141 & 8191]($v_control,$v_start,$user)|0);
     $error$5 = $142;
     label = 30;
     break;
    }
    else if ((label|0) == 26) {
     label = 0;
     $167 = HEAP32[$16>>2]|0;
     $168 = (FUNCTION_TABLE_iiiii[$167 & 8191]($vec13,$vec2,$v_start,$user)|0);
     $error$5 = $168;
     label = 30;
     break;
    }
   } else {
    label = 29;
   }
  } while(0);
  if ((label|0) == 29) {
   label = 0;
   $180 = HEAP32[$12>>2]|0;
   $181 = (FUNCTION_TABLE_iii[$180 & 8191]($v_start,$user)|0);
   $error$5 = $181;
   label = 30;
  }
  if ((label|0) == 30) {
   label = 0;
   $182 = ($error$5|0)==(0);
   $183 = (($25) + 1)|0;
   $$first$0 = $182 ? $183 : $first$036;
   $$7 = $182 ? 0 : 10;
   $$4 = $$7;$error$6 = $error$5;$first$1 = $$first$0;
  }
  switch ($$4|0) {
  case 7:  {
   $$0 = 20;
   label = 34;
   break L10;
   break;
  }
  case 0:  {
   break;
  }
  default: {
   $$0$ph = $error$6;
   label = 33;
   break L10;
  }
  }
  $184 = (($n$035) + 1)|0;
  $185 = HEAP16[$outline>>1]|0;
  $186 = $185 << 16 >> 16;
  $187 = ($184|0)<($186|0);
  if ($187) {
   $first$036 = $first$1;$n$035 = $184;
  } else {
   $$0$ph = 0;
   label = 33;
   break;
  }
 }
 if ((label|0) == 27) {
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 else if ((label|0) == 33) {
  $$0 = $$0$ph;
  STACKTOP = sp;return ($$0|0);
 }
 else if ((label|0) == 34) {
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _FT_Outline_New_Internal($memory,$numPoints,$numContours,$anoutline) {
 $memory = $memory|0;
 $numPoints = $numPoints|0;
 $numContours = $numContours|0;
 $anoutline = $anoutline|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $or$cond = 0, $or$cond2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ($anoutline|0)!=(0|0);
 $1 = ($memory|0)!=(0|0);
 $or$cond = $1 & $0;
 if (!($or$cond)) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 ;HEAP32[$anoutline>>2]=0|0;HEAP32[$anoutline+4>>2]=0|0;HEAP32[$anoutline+8>>2]=0|0;HEAP32[$anoutline+12>>2]=0|0;HEAP32[$anoutline+16>>2]=0|0;
 $2 = ($numContours|0)<(0);
 $3 = ($numContours>>>0)>($numPoints>>>0);
 $or$cond2 = $2 | $3;
 if ($or$cond2) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ($numPoints>>>0)>(32767);
 if ($4) {
  $$0 = 10;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = (_ft_mem_realloc($memory,8,0,$numPoints,0,$error)|0);
 $6 = ((($anoutline)) + 4|0);
 HEAP32[$6>>2] = $5;
 $7 = HEAP32[$error>>2]|0;
 $8 = ($7|0)==(0);
 if ($8) {
  $9 = (_ft_mem_realloc($memory,1,0,$numPoints,0,$error)|0);
  $10 = ((($anoutline)) + 8|0);
  HEAP32[$10>>2] = $9;
  $11 = HEAP32[$error>>2]|0;
  $12 = ($11|0)==(0);
  if ($12) {
   $13 = (_ft_mem_realloc($memory,2,0,$numContours,0,$error)|0);
   $14 = ((($anoutline)) + 12|0);
   HEAP32[$14>>2] = $13;
   $15 = HEAP32[$error>>2]|0;
   $16 = ($15|0)==(0);
   if ($16) {
    $17 = $numPoints&65535;
    $18 = ((($anoutline)) + 2|0);
    HEAP16[$18>>1] = $17;
    $19 = $numContours&65535;
    HEAP16[$anoutline>>1] = $19;
    $20 = ((($anoutline)) + 16|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = $21 | 1;
    HEAP32[$20>>2] = $22;
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
  }
 }
 $23 = ((($anoutline)) + 16|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = $24 | 1;
 HEAP32[$23>>2] = $25;
 $26 = HEAP32[$6>>2]|0;
 _ft_mem_free($memory,$26);
 HEAP32[$6>>2] = 0;
 $27 = ((($anoutline)) + 8|0);
 $28 = HEAP32[$27>>2]|0;
 _ft_mem_free($memory,$28);
 HEAP32[$27>>2] = 0;
 $29 = ((($anoutline)) + 12|0);
 $30 = HEAP32[$29>>2]|0;
 _ft_mem_free($memory,$30);
 ;HEAP32[$anoutline>>2]=0|0;HEAP32[$anoutline+4>>2]=0|0;HEAP32[$anoutline+8>>2]=0|0;HEAP32[$anoutline+12>>2]=0|0;HEAP32[$anoutline+16>>2]=0|0;
 $$pre = HEAP32[$error>>2]|0;
 $$0 = $$pre;
 STACKTOP = sp;return ($$0|0);
}
function _FT_Outline_New($library,$numPoints,$numContours,$anoutline) {
 $library = $library|0;
 $numPoints = $numPoints|0;
 $numContours = $numContours|0;
 $anoutline = $anoutline|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($library|0)==(0|0);
 if ($0) {
  $$0 = 33;
  return ($$0|0);
 }
 $1 = HEAP32[$library>>2]|0;
 $2 = (_FT_Outline_New_Internal($1,$numPoints,$numContours,$anoutline)|0);
 $$0 = $2;
 return ($$0|0);
}
function _FT_Outline_Check($outline) {
 $outline = $outline|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $end0$0$lcssa = 0, $end0$08 = 0, $n$09 = 0, $or$cond = 0, $or$cond3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($outline|0)==(0|0);
 L1: do {
  if (!($0)) {
   $1 = ((($outline)) + 2|0);
   $2 = HEAP16[$1>>1]|0;
   $3 = $2 << 16 >> 16;
   $4 = HEAP16[$outline>>1]|0;
   $5 = $4 << 16 >> 16;
   $6 = $4 | $2;
   $7 = ($6<<16>>16)==(0);
   if ($7) {
    $$1 = 0;
    return ($$1|0);
   }
   $8 = ($2<<16>>16)<(1);
   $9 = ($4<<16>>16)<(1);
   $or$cond3 = $8 | $9;
   if (!($or$cond3)) {
    $10 = ($4<<16>>16)>(0);
    if ($10) {
     $11 = ((($outline)) + 12|0);
     $12 = HEAP32[$11>>2]|0;
     $end0$08 = -1;$n$09 = 0;
     while(1) {
      $15 = (($12) + ($n$09<<1)|0);
      $16 = HEAP16[$15>>1]|0;
      $17 = $16 << 16 >> 16;
      $18 = ($17|0)>($end0$08|0);
      $19 = ($16<<16>>16)<($2<<16>>16);
      $or$cond = $19 & $18;
      $14 = (($n$09) + 1)|0;
      if (!($or$cond)) {
       break L1;
      }
      $13 = ($14|0)<($5|0);
      if ($13) {
       $end0$08 = $17;$n$09 = $14;
      } else {
       $end0$0$lcssa = $17;
       break;
      }
     }
    } else {
     $end0$0$lcssa = -1;
    }
    $20 = (($3) + -1)|0;
    $21 = ($end0$0$lcssa|0)==($20|0);
    if ($21) {
     $$1 = 0;
     return ($$1|0);
    }
   }
  }
 } while(0);
 $$1 = 6;
 return ($$1|0);
}
function _FT_Outline_Copy($source,$target) {
 $source = $source|0;
 $target = $target|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ($source|0)!=(0|0);
 $1 = ($target|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  $$0 = 20;
  return ($$0|0);
 }
 $2 = ((($source)) + 2|0);
 $3 = HEAP16[$2>>1]|0;
 $4 = $3 << 16 >> 16;
 $5 = ((($target)) + 2|0);
 $6 = HEAP16[$5>>1]|0;
 $7 = ($3<<16>>16)==($6<<16>>16);
 if (!($7)) {
  $$0 = 6;
  return ($$0|0);
 }
 $8 = HEAP16[$source>>1]|0;
 $9 = HEAP16[$target>>1]|0;
 $10 = ($8<<16>>16)==($9<<16>>16);
 if (!($10)) {
  $$0 = 6;
  return ($$0|0);
 }
 $11 = ($source|0)==($target|0);
 if ($11) {
  $$0 = 0;
  return ($$0|0);
 }
 $12 = ((($target)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($source)) + 4|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = $4 << 3;
 _memcpy(($13|0),($15|0),($16|0))|0;
 $17 = ((($target)) + 8|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($source)) + 8|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = HEAP16[$2>>1]|0;
 $22 = $21 << 16 >> 16;
 _memcpy(($18|0),($20|0),($22|0))|0;
 $23 = ((($target)) + 12|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($source)) + 12|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = HEAP16[$source>>1]|0;
 $28 = $27 << 16 >> 16;
 $29 = $28 << 1;
 _memcpy(($24|0),($26|0),($29|0))|0;
 $30 = ((($target)) + 16|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = $31 & 1;
 $33 = ((($source)) + 16|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = $34 & -2;
 $36 = $35 | $32;
 HEAP32[$30>>2] = $36;
 $$0 = 0;
 return ($$0|0);
}
function _FT_Outline_Done($library,$outline) {
 $library = $library|0;
 $outline = $outline|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($library|0)==(0|0);
 if ($0) {
  $$0 = 33;
  return ($$0|0);
 }
 $1 = HEAP32[$library>>2]|0;
 $2 = ($outline|0)==(0|0);
 if ($2) {
  $$0 = 20;
  return ($$0|0);
 }
 $3 = ($1|0)==(0|0);
 if ($3) {
  $$0 = 6;
  return ($$0|0);
 }
 $4 = ((($outline)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $5 & 1;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = ((($outline)) + 4|0);
  $9 = HEAP32[$8>>2]|0;
  _ft_mem_free($1,$9);
  HEAP32[$8>>2] = 0;
  $10 = ((($outline)) + 8|0);
  $11 = HEAP32[$10>>2]|0;
  _ft_mem_free($1,$11);
  HEAP32[$10>>2] = 0;
  $12 = ((($outline)) + 12|0);
  $13 = HEAP32[$12>>2]|0;
  _ft_mem_free($1,$13);
  HEAP32[$12>>2] = 0;
 }
 ;HEAP32[$outline>>2]=0|0;HEAP32[$outline+4>>2]=0|0;HEAP32[$outline+8>>2]=0|0;HEAP32[$outline+12>>2]=0|0;HEAP32[$outline+16>>2]=0|0;
 $$0 = 0;
 return ($$0|0);
}
function _FT_Outline_Get_CBox($outline,$acbox) {
 $outline = $outline|0;
 $acbox = $acbox|0;
 var $$pn4 = 0, $$pn4$phi = 0, $$xMin$0 = 0, $$yMin$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $vec$0 = 0, $vec$02 = 0, $vec$08 = 0, $xMax$05 = 0, $xMax$1 = 0, $xMax$2 = 0, $xMin$07 = 0, $xMin$2 = 0, $yMax$03 = 0, $yMax$1 = 0, $yMax$2 = 0;
 var $yMin$06 = 0, $yMin$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($outline|0)!=(0|0);
 $1 = ($acbox|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  return;
 }
 $2 = ((($outline)) + 2|0);
 $3 = HEAP16[$2>>1]|0;
 $4 = ($3<<16>>16)==(0);
 if ($4) {
  $xMax$2 = 0;$xMin$2 = 0;$yMax$2 = 0;$yMin$2 = 0;
 } else {
  $5 = $3 << 16 >> 16;
  $6 = ((($outline)) + 4|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (($7) + ($5<<3)|0);
  $9 = HEAP32[$7>>2]|0;
  $10 = ((($7)) + 4|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($3<<16>>16)>(1);
  if ($12) {
   $vec$02 = ((($7)) + 8|0);
   $$pn4 = $7;$vec$08 = $vec$02;$xMax$05 = $9;$xMin$07 = $9;$yMax$03 = $11;$yMin$06 = $11;
   while(1) {
    $13 = HEAP32[$vec$08>>2]|0;
    $14 = ($13|0)<($xMin$07|0);
    $$xMin$0 = $14 ? $13 : $xMin$07;
    $15 = ($13|0)>($xMax$05|0);
    $xMax$1 = $15 ? $13 : $xMax$05;
    $16 = ((($$pn4)) + 12|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = ($17|0)<($yMin$06|0);
    $$yMin$0 = $18 ? $17 : $yMin$06;
    $19 = ($17|0)>($yMax$03|0);
    $yMax$1 = $19 ? $17 : $yMax$03;
    $vec$0 = ((($vec$08)) + 8|0);
    $20 = ($vec$0>>>0)<($8>>>0);
    if ($20) {
     $$pn4$phi = $vec$08;$vec$08 = $vec$0;$xMax$05 = $xMax$1;$xMin$07 = $$xMin$0;$yMax$03 = $yMax$1;$yMin$06 = $$yMin$0;$$pn4 = $$pn4$phi;
    } else {
     $xMax$2 = $xMax$1;$xMin$2 = $$xMin$0;$yMax$2 = $yMax$1;$yMin$2 = $$yMin$0;
     break;
    }
   }
  } else {
   $xMax$2 = $9;$xMin$2 = $9;$yMax$2 = $11;$yMin$2 = $11;
  }
 }
 HEAP32[$acbox>>2] = $xMin$2;
 $21 = ((($acbox)) + 8|0);
 HEAP32[$21>>2] = $xMax$2;
 $22 = ((($acbox)) + 4|0);
 HEAP32[$22>>2] = $yMin$2;
 $23 = ((($acbox)) + 12|0);
 HEAP32[$23>>2] = $yMax$2;
 return;
}
function _FT_Outline_Translate($outline,$xOffset,$yOffset) {
 $outline = $outline|0;
 $xOffset = $xOffset|0;
 $yOffset = $yOffset|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $n$01 = 0, $vec$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($outline|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($outline)) + 2|0);
 $2 = HEAP16[$1>>1]|0;
 $3 = ($2<<16>>16)>(0);
 if (!($3)) {
  return;
 }
 $4 = ((($outline)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $2 << 16 >> 16;
 $n$01 = 0;$vec$02 = $5;
 while(1) {
  $7 = HEAP32[$vec$02>>2]|0;
  $8 = (($7) + ($xOffset))|0;
  HEAP32[$vec$02>>2] = $8;
  $9 = ((($vec$02)) + 4|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = (($10) + ($yOffset))|0;
  HEAP32[$9>>2] = $11;
  $12 = ((($vec$02)) + 8|0);
  $13 = (($n$01) + 1)<<16>>16;
  $14 = $13&65535;
  $15 = ($14|0)<($6|0);
  if ($15) {
   $n$01 = $13;$vec$02 = $12;
  } else {
   break;
  }
 }
 return;
}
function _FT_Vector_Transform($vector,$matrix) {
 $vector = $vector|0;
 $matrix = $matrix|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($vector|0)!=(0|0);
 $1 = ($matrix|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  return;
 }
 $2 = HEAP32[$vector>>2]|0;
 $3 = HEAP32[$matrix>>2]|0;
 $4 = (_FT_MulFix($2,$3)|0);
 $5 = ((($vector)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($matrix)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (_FT_MulFix($6,$8)|0);
 $10 = (($9) + ($4))|0;
 $11 = HEAP32[$vector>>2]|0;
 $12 = ((($matrix)) + 8|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (_FT_MulFix($11,$13)|0);
 $15 = HEAP32[$5>>2]|0;
 $16 = ((($matrix)) + 12|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = (_FT_MulFix($15,$17)|0);
 $19 = (($18) + ($14))|0;
 HEAP32[$vector>>2] = $10;
 HEAP32[$5>>2] = $19;
 return;
}
function _FT_Outline_Transform($outline,$matrix) {
 $outline = $outline|0;
 $matrix = $matrix|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $vec$02$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($outline|0)!=(0|0);
 $1 = ($matrix|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  return;
 }
 $2 = ((($outline)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($outline)) + 2|0);
 $5 = HEAP16[$4>>1]|0;
 $6 = $5 << 16 >> 16;
 $7 = (($3) + ($6<<3)|0);
 $8 = ($5<<16>>16)>(0);
 if (!($8)) {
  return;
 }
 $9 = ((($matrix)) + 4|0);
 $10 = ((($matrix)) + 8|0);
 $11 = ((($matrix)) + 12|0);
 $vec$02$us = $3;
 while(1) {
  $12 = ($vec$02$us|0)==(0|0);
  if (!($12)) {
   $13 = HEAP32[$vec$02$us>>2]|0;
   $14 = HEAP32[$matrix>>2]|0;
   $15 = (_FT_MulFix($13,$14)|0);
   $16 = ((($vec$02$us)) + 4|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = HEAP32[$9>>2]|0;
   $19 = (_FT_MulFix($17,$18)|0);
   $20 = (($19) + ($15))|0;
   $21 = HEAP32[$vec$02$us>>2]|0;
   $22 = HEAP32[$10>>2]|0;
   $23 = (_FT_MulFix($21,$22)|0);
   $24 = HEAP32[$16>>2]|0;
   $25 = HEAP32[$11>>2]|0;
   $26 = (_FT_MulFix($24,$25)|0);
   $27 = (($26) + ($23))|0;
   HEAP32[$vec$02$us>>2] = $20;
   HEAP32[$16>>2] = $27;
  }
  $28 = ((($vec$02$us)) + 8|0);
  $29 = ($28>>>0)<($7>>>0);
  if ($29) {
   $vec$02$us = $28;
  } else {
   break;
  }
 }
 return;
}
function _FT_Outline_Get_Orientation($outline) {
 $outline = $outline|0;
 var $$0 = 0, $$pn4$i = 0, $$pn4$i$phi = 0, $$xMin$0$i = 0, $$xMin$0$i$lcssa = 0, $$yMin$0$i = 0, $$yMin$0$i$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $area$0$lcssa12 = 0;
 var $area$07 = 0, $area$1$lcssa = 0, $area$1$lcssa$lcssa = 0, $area$14 = 0, $c$05 = 0, $first$06 = 0, $n$03 = 0, $or$cond = 0, $v_prev$sroa$0$02 = 0, $v_prev$sroa$5$01 = 0, $vec$0$i = 0, $vec$02$i = 0, $vec$08$i = 0, $xMax$05$i = 0, $xMax$1$i = 0, $xMax$1$i$lcssa = 0, $xMin$07$i = 0, $yMax$03$i = 0, $yMax$1$i = 0, $yMax$1$i$lcssa = 0;
 var $yMin$06$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($outline|0)==(0|0);
 if ($0) {
  $$0 = 0;
  return ($$0|0);
 }
 $1 = ((($outline)) + 2|0);
 $2 = HEAP16[$1>>1]|0;
 $3 = ($2<<16>>16)<(1);
 if ($3) {
  $$0 = 0;
  return ($$0|0);
 }
 $4 = $2 << 16 >> 16;
 $5 = ((($outline)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + ($4<<3)|0);
 $8 = HEAP32[$6>>2]|0;
 $9 = ((($6)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($2<<16>>16)>(1);
 if (!($11)) {
  $$0 = 2;
  return ($$0|0);
 }
 $vec$02$i = ((($6)) + 8|0);
 $$pn4$i = $6;$vec$08$i = $vec$02$i;$xMax$05$i = $8;$xMin$07$i = $8;$yMax$03$i = $10;$yMin$06$i = $10;
 while(1) {
  $12 = HEAP32[$vec$08$i>>2]|0;
  $13 = ($12|0)<($xMin$07$i|0);
  $$xMin$0$i = $13 ? $12 : $xMin$07$i;
  $14 = ($12|0)>($xMax$05$i|0);
  $xMax$1$i = $14 ? $12 : $xMax$05$i;
  $15 = ((($$pn4$i)) + 12|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ($16|0)<($yMin$06$i|0);
  $$yMin$0$i = $17 ? $16 : $yMin$06$i;
  $18 = ($16|0)>($yMax$03$i|0);
  $yMax$1$i = $18 ? $16 : $yMax$03$i;
  $vec$0$i = ((($vec$08$i)) + 8|0);
  $19 = ($vec$0$i>>>0)<($7>>>0);
  if ($19) {
   $$pn4$i$phi = $vec$08$i;$vec$08$i = $vec$0$i;$xMax$05$i = $xMax$1$i;$xMin$07$i = $$xMin$0$i;$yMax$03$i = $yMax$1$i;$yMin$06$i = $$yMin$0$i;$$pn4$i = $$pn4$i$phi;
  } else {
   $$xMin$0$i$lcssa = $$xMin$0$i;$$yMin$0$i$lcssa = $$yMin$0$i;$xMax$1$i$lcssa = $xMax$1$i;$yMax$1$i$lcssa = $yMax$1$i;
   break;
  }
 }
 $20 = ($$xMin$0$i$lcssa|0)==($xMax$1$i$lcssa|0);
 $21 = ($yMax$1$i$lcssa|0)==($$yMin$0$i$lcssa|0);
 $or$cond = $20 | $21;
 if ($or$cond) {
  $$0 = 2;
  return ($$0|0);
 }
 $22 = ($xMax$1$i$lcssa|0)<(0);
 $23 = (0 - ($xMax$1$i$lcssa))|0;
 $24 = $22 ? $23 : $xMax$1$i$lcssa;
 $25 = ($$xMin$0$i$lcssa|0)<(0);
 $26 = (0 - ($$xMin$0$i$lcssa))|0;
 $27 = $25 ? $26 : $$xMin$0$i$lcssa;
 $28 = $24 | $27;
 $29 = (Math_clz32(($28|0))|0);
 $30 = (31 - ($29))|0;
 $31 = (($30) + -14)|0;
 $32 = ($31|0)>(0);
 $33 = $32 ? $31 : 0;
 $34 = (($yMax$1$i$lcssa) - ($$yMin$0$i$lcssa))|0;
 $35 = (Math_clz32(($34|0))|0);
 $36 = (31 - ($35))|0;
 $37 = (($36) + -14)|0;
 $38 = ($37|0)>(0);
 $39 = $38 ? $37 : 0;
 $40 = HEAP16[$outline>>1]|0;
 $41 = ($40<<16>>16)>(0);
 if ($41) {
  $42 = ((($outline)) + 12|0);
  $43 = HEAP32[$42>>2]|0;
  $44 = $40 << 16 >> 16;
  $area$07 = 0;$c$05 = 0;$first$06 = 0;
  while(1) {
   $45 = (($43) + ($c$05<<1)|0);
   $46 = HEAP16[$45>>1]|0;
   $47 = $46 << 16 >> 16;
   $48 = ($first$06|0)>($47|0);
   if ($48) {
    $area$1$lcssa = $area$07;
   } else {
    $49 = (((($6) + ($47<<3)|0)) + 4|0);
    $50 = HEAP32[$49>>2]|0;
    $51 = (($6) + ($47<<3)|0);
    $52 = HEAP32[$51>>2]|0;
    $area$14 = $area$07;$n$03 = $first$06;$v_prev$sroa$0$02 = $52;$v_prev$sroa$5$01 = $50;
    while(1) {
     $53 = (($6) + ($n$03<<3)|0);
     $54 = HEAP32[$53>>2]|0;
     $55 = (((($6) + ($n$03<<3)|0)) + 4|0);
     $56 = HEAP32[$55>>2]|0;
     $57 = (($56) - ($v_prev$sroa$5$01))|0;
     $58 = $57 >> $39;
     $59 = (($54) + ($v_prev$sroa$0$02))|0;
     $60 = $59 >> $33;
     $61 = Math_imul($58, $60)|0;
     $62 = (($61) + ($area$14))|0;
     $63 = (($n$03) + 1)|0;
     $64 = ($n$03|0)<($47|0);
     if ($64) {
      $area$14 = $62;$n$03 = $63;$v_prev$sroa$0$02 = $54;$v_prev$sroa$5$01 = $56;
     } else {
      $area$1$lcssa = $62;
      break;
     }
    }
   }
   $65 = (($47) + 1)|0;
   $66 = (($c$05) + 1)|0;
   $67 = ($66|0)<($44|0);
   if ($67) {
    $area$07 = $area$1$lcssa;$c$05 = $66;$first$06 = $65;
   } else {
    $area$1$lcssa$lcssa = $area$1$lcssa;
    break;
   }
  }
  $68 = ($area$1$lcssa$lcssa|0)>(0);
  if ($68) {
   $$0 = 1;
   return ($$0|0);
  } else {
   $area$0$lcssa12 = $area$1$lcssa$lcssa;
  }
 } else {
  $area$0$lcssa12 = 0;
 }
 $69 = $area$0$lcssa12 >> 31;
 $70 = (($69) + 2)|0;
 $71 = $70 & -2;
 $$0 = $71;
 return ($$0|0);
}
function _FT_Raccess_Get_HeaderInfo($library,$stream,$rfork_offset,$map_offset,$rdata_pos) {
 $library = $library|0;
 $stream = $stream|0;
 $rfork_offset = $rfork_offset|0;
 $map_offset = $map_offset|0;
 $rdata_pos = $rdata_pos|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $allmatch$1 = 0, $allmatch$1$15 = 0;
 var $allzeros$0$ = 0, $allzeros$0$$15 = 0, $error = 0, $head = 0, $head2 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $head = sp + 24|0;
 $head2 = sp + 8|0;
 $0 = (_FT_Stream_Seek($stream,$rfork_offset)|0);
 HEAP32[$error>>2] = $0;
 $1 = ($0|0)==(0);
 if (!($1)) {
  $$0 = $0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = (_FT_Stream_Read($stream,$head,16)|0);
 HEAP32[$error>>2] = $2;
 $3 = ($2|0)==(0);
 if (!($3)) {
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = HEAP8[$head>>0]|0;
 $5 = $4&255;
 $6 = $5 << 24;
 $7 = ((($head)) + 1|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = $8&255;
 $10 = $9 << 16;
 $11 = $10 | $6;
 $12 = ((($head)) + 2|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = $13&255;
 $15 = $14 << 8;
 $16 = $11 | $15;
 $17 = ((($head)) + 3|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = $18&255;
 $20 = $16 | $19;
 $21 = (($20) + ($rfork_offset))|0;
 HEAP32[$rdata_pos>>2] = $21;
 $22 = ((($head)) + 4|0);
 $23 = HEAP8[$22>>0]|0;
 $24 = $23&255;
 $25 = $24 << 24;
 $26 = ((($head)) + 5|0);
 $27 = HEAP8[$26>>0]|0;
 $28 = $27&255;
 $29 = $28 << 16;
 $30 = $29 | $25;
 $31 = ((($head)) + 6|0);
 $32 = HEAP8[$31>>0]|0;
 $33 = $32&255;
 $34 = $33 << 8;
 $35 = $30 | $34;
 $36 = ((($head)) + 7|0);
 $37 = HEAP8[$36>>0]|0;
 $38 = $37&255;
 $39 = $35 | $38;
 $40 = (($39) + ($rfork_offset))|0;
 $41 = ((($head)) + 8|0);
 $42 = HEAP8[$41>>0]|0;
 $43 = $42&255;
 $44 = $43 << 24;
 $45 = ((($head)) + 9|0);
 $46 = HEAP8[$45>>0]|0;
 $47 = $46&255;
 $48 = $47 << 16;
 $49 = $48 | $44;
 $50 = ((($head)) + 10|0);
 $51 = HEAP8[$50>>0]|0;
 $52 = $51&255;
 $53 = $52 << 8;
 $54 = $49 | $53;
 $55 = ((($head)) + 11|0);
 $56 = HEAP8[$55>>0]|0;
 $57 = $56&255;
 $58 = $54 | $57;
 $59 = (($58) + ($21))|0;
 $60 = ($59|0)!=($40|0);
 $61 = ($39|0)==(0);
 $or$cond = $61 | $60;
 if ($or$cond) {
  $$0 = 2;
  STACKTOP = sp;return ($$0|0);
 }
 $62 = (_FT_Stream_Seek($stream,$40)|0);
 HEAP32[$error>>2] = $62;
 $63 = ($62|0)==(0);
 if (!($63)) {
  $$0 = $62;
  STACKTOP = sp;return ($$0|0);
 }
 $64 = ((($head)) + 15|0);
 $65 = HEAP8[$64>>0]|0;
 $66 = $65&255;
 $67 = (($66) + 1)|0;
 $68 = $67&255;
 $69 = ((($head2)) + 15|0);
 HEAP8[$69>>0] = $68;
 $70 = (_FT_Stream_Read($stream,$head2,16)|0);
 HEAP32[$error>>2] = $70;
 $71 = ($70|0)==(0);
 if (!($71)) {
  $$0 = $70;
  STACKTOP = sp;return ($$0|0);
 }
 $72 = HEAP8[$head2>>0]|0;
 $73 = ($72<<24>>24)==(0);
 $allzeros$0$ = $73&1;
 $74 = HEAP8[$head>>0]|0;
 $75 = ($72<<24>>24)==($74<<24>>24);
 $allmatch$1 = $75&1;
 $76 = ((($head2)) + 1|0);
 $77 = HEAP8[$76>>0]|0;
 $78 = HEAP8[$7>>0]|0;
 $79 = ($77<<24>>24)==($78<<24>>24);
 $80 = ((($head2)) + 2|0);
 $81 = HEAP8[$80>>0]|0;
 $82 = $81 | $77;
 $83 = HEAP8[$12>>0]|0;
 $84 = ($81<<24>>24)==($83<<24>>24);
 $85 = $84 & $79;
 $86 = ((($head2)) + 3|0);
 $87 = HEAP8[$86>>0]|0;
 $88 = $87 | $82;
 $89 = HEAP8[$17>>0]|0;
 $90 = ($87<<24>>24)==($89<<24>>24);
 $91 = $90 & $85;
 $92 = ((($head2)) + 4|0);
 $93 = HEAP8[$92>>0]|0;
 $94 = $93 | $88;
 $95 = HEAP8[$22>>0]|0;
 $96 = ($93<<24>>24)==($95<<24>>24);
 $97 = $96 & $91;
 $98 = ((($head2)) + 5|0);
 $99 = HEAP8[$98>>0]|0;
 $100 = $99 | $94;
 $101 = HEAP8[$26>>0]|0;
 $102 = ($99<<24>>24)==($101<<24>>24);
 $103 = $102 & $97;
 $104 = ((($head2)) + 6|0);
 $105 = HEAP8[$104>>0]|0;
 $106 = $105 | $100;
 $107 = HEAP8[$31>>0]|0;
 $108 = ($105<<24>>24)==($107<<24>>24);
 $109 = $108 & $103;
 $110 = ((($head2)) + 7|0);
 $111 = HEAP8[$110>>0]|0;
 $112 = $111 | $106;
 $113 = HEAP8[$36>>0]|0;
 $114 = ($111<<24>>24)==($113<<24>>24);
 $115 = $114 & $109;
 $116 = ((($head2)) + 8|0);
 $117 = HEAP8[$116>>0]|0;
 $118 = $117 | $112;
 $119 = HEAP8[$41>>0]|0;
 $120 = ($117<<24>>24)==($119<<24>>24);
 $121 = $120 & $115;
 $122 = ((($head2)) + 9|0);
 $123 = HEAP8[$122>>0]|0;
 $124 = $123 | $118;
 $125 = HEAP8[$45>>0]|0;
 $126 = ($123<<24>>24)==($125<<24>>24);
 $127 = $126 & $121;
 $128 = ((($head2)) + 10|0);
 $129 = HEAP8[$128>>0]|0;
 $130 = $129 | $124;
 $131 = HEAP8[$50>>0]|0;
 $132 = ($129<<24>>24)==($131<<24>>24);
 $133 = $132 & $127;
 $134 = ((($head2)) + 11|0);
 $135 = HEAP8[$134>>0]|0;
 $136 = $135 | $130;
 $137 = HEAP8[$55>>0]|0;
 $138 = ($135<<24>>24)==($137<<24>>24);
 $139 = $138 & $133;
 $140 = ((($head2)) + 12|0);
 $141 = HEAP8[$140>>0]|0;
 $142 = $141 | $136;
 $143 = ((($head)) + 12|0);
 $144 = HEAP8[$143>>0]|0;
 $145 = ($141<<24>>24)==($144<<24>>24);
 $146 = $145 & $139;
 $147 = ((($head2)) + 13|0);
 $148 = HEAP8[$147>>0]|0;
 $149 = $148 | $142;
 $150 = ((($head)) + 13|0);
 $151 = HEAP8[$150>>0]|0;
 $152 = ($148<<24>>24)==($151<<24>>24);
 $153 = $152 & $146;
 $154 = ((($head2)) + 14|0);
 $155 = HEAP8[$154>>0]|0;
 $156 = $155 | $149;
 $157 = ((($head)) + 14|0);
 $158 = HEAP8[$157>>0]|0;
 $159 = ($155<<24>>24)==($158<<24>>24);
 $160 = $159 & $153;
 $161 = HEAP8[$69>>0]|0;
 $162 = $161 | $156;
 $163 = ($162<<24>>24)==(0);
 $allzeros$0$$15 = $163 ? $allzeros$0$ : 0;
 $164 = HEAP8[$64>>0]|0;
 $165 = ($161<<24>>24)==($164<<24>>24);
 $166 = $165 & $160;
 $allmatch$1$15 = $166 ? $allmatch$1 : 0;
 $167 = $allmatch$1$15 | $allzeros$0$$15;
 $168 = ($167|0)==(0);
 if ($168) {
  $$0 = 2;
  STACKTOP = sp;return ($$0|0);
 }
 $169 = (_FT_Stream_Skip($stream,8)|0);
 HEAP32[$error>>2] = $169;
 $170 = (_FT_Stream_ReadUShort($stream,$error)|0);
 $171 = HEAP32[$error>>2]|0;
 $172 = ($171|0)==(0);
 if (!($172)) {
  $$0 = $171;
  STACKTOP = sp;return ($$0|0);
 }
 $173 = $170&65535;
 $174 = (($173) + ($40))|0;
 $175 = (_FT_Stream_Seek($stream,$174)|0);
 HEAP32[$error>>2] = $175;
 $176 = ($175|0)==(0);
 if (!($176)) {
  $$0 = $175;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$map_offset>>2] = $174;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _FT_Raccess_Get_DataOffsets($library,$stream,$map_offset,$rdata_pos,$tag,$sort_by_res_id,$offsets,$count) {
 $library = $library|0;
 $stream = $stream|0;
 $map_offset = $map_offset|0;
 $rdata_pos = $rdata_pos|0;
 $tag = $tag|0;
 $sort_by_res_id = $sort_by_res_id|0;
 $offsets = $offsets|0;
 $count = $count|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa29 = 0, $$lcssa30 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $i$011 = 0, $j$08 = 0, $j$27 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = HEAP32[$library>>2]|0;
 $1 = (_FT_Stream_Seek($stream,$map_offset)|0);
 HEAP32[$error>>2] = $1;
 $2 = ($1|0)==(0);
 if (!($2)) {
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = (_FT_Stream_ReadUShort($stream,$error)|0);
 $4 = HEAP32[$error>>2]|0;
 $5 = ($4|0)==(0);
 if (!($5)) {
  $$0 = $4;
  STACKTOP = sp;return ($$0|0);
 }
 $6 = $3&65535;
 $i$011 = 0;
 while(1) {
  $9 = (_FT_Stream_ReadULong($stream,$error)|0);
  $10 = HEAP32[$error>>2]|0;
  $11 = ($10|0)==(0);
  if (!($11)) {
   $$0 = $10;
   label = 24;
   break;
  }
  $12 = (_FT_Stream_ReadUShort($stream,$error)|0);
  $13 = HEAP32[$error>>2]|0;
  $14 = ($13|0)==(0);
  if (!($14)) {
   $$0 = $13;
   label = 24;
   break;
  }
  $15 = (_FT_Stream_ReadUShort($stream,$error)|0);
  $16 = HEAP32[$error>>2]|0;
  $17 = ($16|0)==(0);
  if (!($17)) {
   $$0 = $16;
   label = 24;
   break;
  }
  $18 = ($9|0)==($tag|0);
  if ($18) {
   $$lcssa29 = $12;$$lcssa30 = $15;
   label = 9;
   break;
  }
  $7 = (($i$011) + 1)|0;
  $8 = ($i$011|0)<($6|0);
  if ($8) {
   $i$011 = $7;
  } else {
   $$0 = 1;
   label = 24;
   break;
  }
 }
 if ((label|0) == 9) {
  $19 = $$lcssa29&65535;
  $20 = $$lcssa30&65535;
  $21 = (($19) + 1)|0;
  HEAP32[$count>>2] = $21;
  $22 = (($20) + ($map_offset))|0;
  $23 = (_FT_Stream_Seek($stream,$22)|0);
  HEAP32[$error>>2] = $23;
  $24 = ($23|0)==(0);
  if (!($24)) {
   $$0 = $23;
   STACKTOP = sp;return ($$0|0);
  }
  $25 = HEAP32[$count>>2]|0;
  $26 = (_ft_mem_realloc($0,8,0,$25,0,$error)|0);
  $27 = HEAP32[$error>>2]|0;
  $28 = ($27|0)==(0);
  if (!($28)) {
   $$0 = $27;
   STACKTOP = sp;return ($$0|0);
  }
  $29 = HEAP32[$count>>2]|0;
  $30 = ($29|0)>(0);
  L23: do {
   if ($30) {
    $j$08 = 0;
    while(1) {
     $31 = (_FT_Stream_ReadUShort($stream,$error)|0);
     $32 = (($26) + ($j$08<<3)|0);
     HEAP16[$32>>1] = $31;
     $33 = HEAP32[$error>>2]|0;
     $34 = ($33|0)==(0);
     if (!($34)) {
      break L23;
     }
     $35 = (_FT_Stream_Skip($stream,2)|0);
     HEAP32[$error>>2] = $35;
     $36 = ($35|0)==(0);
     if (!($36)) {
      break L23;
     }
     $37 = (_FT_Stream_ReadULong($stream,$error)|0);
     $38 = HEAP32[$error>>2]|0;
     $39 = ($38|0)==(0);
     if (!($39)) {
      break L23;
     }
     $40 = (_FT_Stream_Skip($stream,4)|0);
     HEAP32[$error>>2] = $40;
     $41 = ($40|0)==(0);
     if (!($41)) {
      break L23;
     }
     $42 = $37 & 16777215;
     $43 = (((($26) + ($j$08<<3)|0)) + 4|0);
     HEAP32[$43>>2] = $42;
     $44 = (($j$08) + 1)|0;
     $45 = HEAP32[$count>>2]|0;
     $46 = ($44|0)<($45|0);
     if ($46) {
      $j$08 = $44;
     } else {
      $$lcssa = $45;
      label = 17;
      break;
     }
    }
   } else {
    $$lcssa = $29;
    label = 17;
   }
  } while(0);
  if ((label|0) == 17) {
   $47 = ($sort_by_res_id<<24>>24)==(0);
   if ($47) {
    $48 = $$lcssa;
   } else {
    _qsort($26,$$lcssa,8,4346);
    $$pre = HEAP32[$count>>2]|0;
    $48 = $$pre;
   }
   $49 = (_ft_mem_realloc($0,4,0,$48,0,$error)|0);
   $50 = HEAP32[$error>>2]|0;
   $51 = ($50|0)==(0);
   if ($51) {
    $52 = HEAP32[$count>>2]|0;
    $53 = ($52|0)>(0);
    if ($53) {
     $j$27 = 0;
     while(1) {
      $54 = (((($26) + ($j$27<<3)|0)) + 4|0);
      $55 = HEAP32[$54>>2]|0;
      $56 = (($55) + ($rdata_pos))|0;
      $57 = (($49) + ($j$27<<2)|0);
      HEAP32[$57>>2] = $56;
      $58 = (($j$27) + 1)|0;
      $59 = HEAP32[$count>>2]|0;
      $60 = ($58|0)<($59|0);
      if ($60) {
       $j$27 = $58;
      } else {
       break;
      }
     }
    }
    HEAP32[$offsets>>2] = $49;
    HEAP32[$error>>2] = 0;
   }
  }
  _ft_mem_free($0,$26);
  $61 = HEAP32[$error>>2]|0;
  $$0 = $61;
  STACKTOP = sp;return ($$0|0);
 }
 else if ((label|0) == 24) {
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _ft_raccess_sort_ref_by_id($a,$b) {
 $a = $a|0;
 $b = $b|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP16[$a>>1]|0;
 $1 = HEAP16[$b>>1]|0;
 $2 = ($0&65535)<($1&65535);
 $3 = ($0&65535)>($1&65535);
 $$ = $3&1;
 $$0 = $2 ? -1 : $$;
 return ($$0|0);
}
function _FT_Raccess_Guess($library,$stream,$base_name,$new_names,$offsets,$errors) {
 $library = $library|0;
 $stream = $stream|0;
 $base_name = $base_name|0;
 $new_names = $new_names|0;
 $offsets = $offsets|0;
 $errors = $errors|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $exitcond3 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($stream|0)==(0|0);
 if ($0) {
  HEAP32[$new_names>>2] = 0;
  HEAP32[$errors>>2] = 0;
  $1 = (_raccess_guess_apple_double($library,0,$base_name,$new_names,$offsets)|0);
  HEAP32[$errors>>2] = $1;
  $2 = ((($new_names)) + 4|0);
  HEAP32[$2>>2] = 0;
  $3 = ((($errors)) + 4|0);
  HEAP32[$3>>2] = 0;
  $4 = ((($offsets)) + 4|0);
  $5 = (_raccess_guess_apple_single($library,0,$base_name,$2,$4)|0);
  HEAP32[$3>>2] = $5;
  $6 = ((($new_names)) + 8|0);
  HEAP32[$6>>2] = 0;
  $7 = ((($errors)) + 8|0);
  HEAP32[$7>>2] = 0;
  $8 = ((($offsets)) + 8|0);
  $9 = (_raccess_guess_darwin_ufs_export($library,0,$base_name,$6,$8)|0);
  HEAP32[$7>>2] = $9;
  $10 = ((($new_names)) + 12|0);
  HEAP32[$10>>2] = 0;
  $11 = ((($errors)) + 12|0);
  HEAP32[$11>>2] = 0;
  $12 = ((($offsets)) + 12|0);
  $13 = (_raccess_guess_darwin_newvfs($library,0,$base_name,$10,$12)|0);
  HEAP32[$11>>2] = $13;
  $14 = ((($new_names)) + 16|0);
  HEAP32[$14>>2] = 0;
  $15 = ((($errors)) + 16|0);
  HEAP32[$15>>2] = 0;
  $16 = ((($offsets)) + 16|0);
  $17 = (_raccess_guess_darwin_hfsplus($library,0,$base_name,$14,$16)|0);
  HEAP32[$15>>2] = $17;
  $18 = ((($new_names)) + 20|0);
  HEAP32[$18>>2] = 0;
  $19 = ((($errors)) + 20|0);
  HEAP32[$19>>2] = 0;
  $20 = ((($offsets)) + 20|0);
  $21 = (_raccess_guess_vfat($library,0,$base_name,$18,$20)|0);
  HEAP32[$19>>2] = $21;
  $22 = ((($new_names)) + 24|0);
  HEAP32[$22>>2] = 0;
  $23 = ((($errors)) + 24|0);
  HEAP32[$23>>2] = 0;
  $24 = ((($offsets)) + 24|0);
  $25 = (_raccess_guess_linux_cap($library,0,$base_name,$22,$24)|0);
  HEAP32[$23>>2] = $25;
  $26 = ((($new_names)) + 28|0);
  HEAP32[$26>>2] = 0;
  $27 = ((($errors)) + 28|0);
  HEAP32[$27>>2] = 0;
  $28 = ((($offsets)) + 28|0);
  $29 = (_raccess_guess_linux_double($library,0,$base_name,$26,$28)|0);
  HEAP32[$27>>2] = $29;
  $30 = ((($new_names)) + 32|0);
  HEAP32[$30>>2] = 0;
  $31 = ((($errors)) + 32|0);
  HEAP32[$31>>2] = 0;
  $32 = ((($offsets)) + 32|0);
  $33 = (_raccess_guess_linux_netatalk($library,0,$base_name,$30,$32)|0);
  HEAP32[$31>>2] = $33;
  return;
 } else {
  $i$01 = 0;
 }
 while(1) {
  $34 = (($new_names) + ($i$01<<2)|0);
  HEAP32[$34>>2] = 0;
  $35 = (_FT_Stream_Seek($stream,0)|0);
  $36 = (($errors) + ($i$01<<2)|0);
  HEAP32[$36>>2] = $35;
  $37 = ($35|0)==(0);
  if ($37) {
   $38 = (47800 + ($i$01<<3)|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = (($offsets) + ($i$01<<2)|0);
   $41 = (FUNCTION_TABLE_iiiiii[$39 & 8191]($library,$stream,$base_name,$34,$40)|0);
   HEAP32[$36>>2] = $41;
  }
  $42 = (($i$01) + 1)|0;
  $exitcond3 = ($42|0)==(9);
  if ($exitcond3) {
   break;
  } else {
   $i$01 = $42;
  }
 }
 return;
}
function _raccess_guess_apple_double($library,$stream,$base_file_name,$result_file_name,$result_offset) {
 $library = $library|0;
 $stream = $stream|0;
 $base_file_name = $base_file_name|0;
 $result_file_name = $result_file_name|0;
 $result_offset = $result_offset|0;
 var $$0 = 0, $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$result_file_name>>2] = 0;
 $0 = ($stream|0)==(0|0);
 if ($0) {
  $$0 = 81;
  return ($$0|0);
 }
 $1 = (_raccess_guess_apple_generic($stream,333319,$result_offset)|0);
 $$0 = $1;
 return ($$0|0);
}
function _raccess_guess_apple_generic($stream,$magic,$result_offset) {
 $stream = $stream|0;
 $magic = $magic|0;
 $result_offset = $result_offset|0;
 var $$0 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = (_FT_Stream_ReadULong($stream,$error)|0);
 $1 = HEAP32[$error>>2]|0;
 $2 = ($1|0)==(0);
 L1: do {
  if ($2) {
   $3 = ($0|0)==($magic|0);
   if ($3) {
    (_FT_Stream_ReadULong($stream,$error)|0);
    $4 = HEAP32[$error>>2]|0;
    $5 = ($4|0)==(0);
    if ($5) {
     $6 = (_FT_Stream_Skip($stream,16)|0);
     HEAP32[$error>>2] = $6;
     $7 = ($6|0)==(0);
     if ($7) {
      $8 = (_FT_Stream_ReadUShort($stream,$error)|0);
      $9 = HEAP32[$error>>2]|0;
      $10 = ($9|0)==(0);
      if ($10) {
       $11 = $8&65535;
       $12 = ($8<<16>>16)==(0);
       if ($12) {
        $$0 = 2;
       } else {
        $i$01 = 0;
        while(1) {
         $13 = (_FT_Stream_ReadULong($stream,$error)|0);
         $14 = HEAP32[$error>>2]|0;
         $15 = ($14|0)==(0);
         if (!($15)) {
          $$0 = $14;
          break L1;
         }
         $16 = ($13|0)==(2);
         if ($16) {
          $17 = (_FT_Stream_ReadULong($stream,$error)|0);
          $18 = HEAP32[$error>>2]|0;
          $19 = ($18|0)==(0);
          if ($19) {
           (_FT_Stream_ReadULong($stream,$error)|0);
           $20 = HEAP32[$error>>2]|0;
           $21 = ($20|0)==(0);
           if ($21) {
            $$lcssa = $17;
            break;
           }
          }
         } else {
          $22 = (_FT_Stream_Skip($stream,8)|0);
          HEAP32[$error>>2] = $22;
          $23 = ($22|0)==(0);
          if (!($23)) {
           $$0 = $22;
           break L1;
          }
         }
         $24 = (($i$01) + 1)|0;
         $25 = ($24|0)<($11|0);
         if ($25) {
          $i$01 = $24;
         } else {
          $$0 = 2;
          break L1;
         }
        }
        HEAP32[$result_offset>>2] = $$lcssa;
        $$0 = 0;
       }
      } else {
       $$0 = $9;
      }
     } else {
      $$0 = $6;
     }
    } else {
     $$0 = $4;
    }
   } else {
    $$0 = 2;
   }
  } else {
   $$0 = $1;
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _raccess_guess_apple_single($library,$stream,$base_file_name,$result_file_name,$result_offset) {
 $library = $library|0;
 $stream = $stream|0;
 $base_file_name = $base_file_name|0;
 $result_file_name = $result_file_name|0;
 $result_offset = $result_offset|0;
 var $$0 = 0, $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$result_file_name>>2] = 0;
 $0 = ($stream|0)==(0|0);
 if ($0) {
  $$0 = 81;
  return ($$0|0);
 }
 $1 = (_raccess_guess_apple_generic($stream,333312,$result_offset)|0);
 $$0 = $1;
 return ($$0|0);
}
function _raccess_guess_darwin_ufs_export($library,$stream,$base_file_name,$result_file_name,$result_offset) {
 $library = $library|0;
 $stream = $stream|0;
 $base_file_name = $base_file_name|0;
 $result_file_name = $result_file_name|0;
 $result_offset = $result_offset|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i1 = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $args2$i = 0, $stream2$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $args2$i = sp + 8|0;
 $stream2$i = sp;
 $0 = HEAP32[$library>>2]|0;
 $1 = (_raccess_make_file_name($0,$base_file_name,530523)|0);
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = 64;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$args2$i>>2] = 4;
 $3 = ((($args2$i)) + 12|0);
 HEAP32[$3>>2] = $1;
 $4 = (_FT_Stream_New($library,$args2$i,$stream2$i)|0);
 $5 = ($4|0)==(0);
 if ($5) {
  $6 = HEAP32[$stream2$i>>2]|0;
  $7 = ($6|0)==(0|0);
  if ($7) {
   $$0$i$i = 81;$9 = 0;
  } else {
   $8 = (_raccess_guess_apple_generic($6,333319,$result_offset)|0);
   $$pre$i = HEAP32[$stream2$i>>2]|0;
   $$0$i$i = $8;$9 = $$pre$i;
  }
  _FT_Stream_Free($9,0);
  $10 = ($$0$i$i|0)==(0);
  if ($10) {
   HEAP32[$result_file_name>>2] = $1;
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  } else {
   $$0$i1 = $$0$i$i;
  }
 } else {
  $$0$i1 = $4;
 }
 _ft_mem_free($0,$1);
 $$0 = $$0$i1;
 STACKTOP = sp;return ($$0|0);
}
function _raccess_make_file_name($memory,$original_name,$insertion) {
 $memory = $memory|0;
 $original_name = $original_name|0;
 $insertion = $insertion|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $slash$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 HEAP32[$error>>2] = 0;
 $0 = (_strlen($original_name)|0);
 $1 = (_strlen($insertion)|0);
 $2 = (($0) + 1)|0;
 $3 = (($2) + ($1))|0;
 $4 = (_ft_mem_alloc($memory,$3,$error)|0);
 $5 = HEAP32[$error>>2]|0;
 $6 = ($5|0)==(0);
 if (!($6)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $7 = (_strrchr($original_name,47)|0);
 $8 = ($7|0)==(0|0);
 if ($8) {
  HEAP8[$4>>0] = 0;
  $slash$0 = $original_name;
 } else {
  $9 = $7;
  $10 = $original_name;
  $11 = (($9) - ($10))|0;
  $12 = (($11) + 1)|0;
  (_strncpy($4,$original_name,$12)|0);
  $13 = (($4) + ($12)|0);
  HEAP8[$13>>0] = 0;
  $14 = ((($7)) + 1|0);
  $slash$0 = $14;
 }
 (_strcat($4,$insertion)|0);
 (_strcat($4,$slash$0)|0);
 $$0 = $4;
 STACKTOP = sp;return ($$0|0);
}
function _raccess_guess_darwin_newvfs($library,$stream,$base_file_name,$result_file_name,$result_offset) {
 $library = $library|0;
 $stream = $stream|0;
 $base_file_name = $base_file_name|0;
 $result_file_name = $result_file_name|0;
 $result_offset = $result_offset|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $error = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = (_strlen($base_file_name)|0);
 $1 = (($0) + 18)|0;
 $2 = HEAP32[$library>>2]|0;
 $3 = (_ft_mem_alloc($2,$1,$error)|0);
 $4 = HEAP32[$error>>2]|0;
 $5 = ($4|0)==(0);
 if (!($5)) {
  $$0 = $4;
  STACKTOP = sp;return ($$0|0);
 }
 _memcpy(($3|0),($base_file_name|0),($0|0))|0;
 $6 = (($3) + ($0)|0);
 dest=$6; src=530526; stop=dest+18|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
 HEAP32[$result_file_name>>2] = $3;
 HEAP32[$result_offset>>2] = 0;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _raccess_guess_darwin_hfsplus($library,$stream,$base_file_name,$result_file_name,$result_offset) {
 $library = $library|0;
 $stream = $stream|0;
 $base_file_name = $base_file_name|0;
 $result_file_name = $result_file_name|0;
 $result_offset = $result_offset|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $error = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = (_strlen($base_file_name)|0);
 $1 = (($0) + 6)|0;
 $2 = HEAP32[$library>>2]|0;
 $3 = (_ft_mem_alloc($2,$1,$error)|0);
 $4 = HEAP32[$error>>2]|0;
 $5 = ($4|0)==(0);
 if (!($5)) {
  $$0 = $4;
  STACKTOP = sp;return ($$0|0);
 }
 _memcpy(($3|0),($base_file_name|0),($0|0))|0;
 $6 = (($3) + ($0)|0);
 ;HEAP8[$6>>0]=HEAP8[530544>>0]|0;HEAP8[$6+1>>0]=HEAP8[530544+1>>0]|0;HEAP8[$6+2>>0]=HEAP8[530544+2>>0]|0;HEAP8[$6+3>>0]=HEAP8[530544+3>>0]|0;HEAP8[$6+4>>0]=HEAP8[530544+4>>0]|0;HEAP8[$6+5>>0]=HEAP8[530544+5>>0]|0;
 HEAP32[$result_file_name>>2] = $3;
 HEAP32[$result_offset>>2] = 0;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _raccess_guess_vfat($library,$stream,$base_file_name,$result_file_name,$result_offset) {
 $library = $library|0;
 $stream = $stream|0;
 $base_file_name = $base_file_name|0;
 $result_file_name = $result_file_name|0;
 $result_offset = $result_offset|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$library>>2]|0;
 $1 = (_raccess_make_file_name($0,$base_file_name,530550)|0);
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = 64;
  return ($$0|0);
 }
 HEAP32[$result_file_name>>2] = $1;
 HEAP32[$result_offset>>2] = 0;
 $$0 = 0;
 return ($$0|0);
}
function _raccess_guess_linux_cap($library,$stream,$base_file_name,$result_file_name,$result_offset) {
 $library = $library|0;
 $stream = $stream|0;
 $base_file_name = $base_file_name|0;
 $result_file_name = $result_file_name|0;
 $result_offset = $result_offset|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$library>>2]|0;
 $1 = (_raccess_make_file_name($0,$base_file_name,530564)|0);
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = 64;
  return ($$0|0);
 }
 HEAP32[$result_file_name>>2] = $1;
 HEAP32[$result_offset>>2] = 0;
 $$0 = 0;
 return ($$0|0);
}
function _raccess_guess_linux_double($library,$stream,$base_file_name,$result_file_name,$result_offset) {
 $library = $library|0;
 $stream = $stream|0;
 $base_file_name = $base_file_name|0;
 $result_file_name = $result_file_name|0;
 $result_offset = $result_offset|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i1 = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $args2$i = 0, $stream2$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $args2$i = sp + 8|0;
 $stream2$i = sp;
 $0 = HEAP32[$library>>2]|0;
 $1 = (_raccess_make_file_name($0,$base_file_name,530575)|0);
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = 64;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$args2$i>>2] = 4;
 $3 = ((($args2$i)) + 12|0);
 HEAP32[$3>>2] = $1;
 $4 = (_FT_Stream_New($library,$args2$i,$stream2$i)|0);
 $5 = ($4|0)==(0);
 if ($5) {
  $6 = HEAP32[$stream2$i>>2]|0;
  $7 = ($6|0)==(0|0);
  if ($7) {
   $$0$i$i = 81;$9 = 0;
  } else {
   $8 = (_raccess_guess_apple_generic($6,333319,$result_offset)|0);
   $$pre$i = HEAP32[$stream2$i>>2]|0;
   $$0$i$i = $8;$9 = $$pre$i;
  }
  _FT_Stream_Free($9,0);
  $10 = ($$0$i$i|0)==(0);
  if ($10) {
   HEAP32[$result_file_name>>2] = $1;
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  } else {
   $$0$i1 = $$0$i$i;
  }
 } else {
  $$0$i1 = $4;
 }
 _ft_mem_free($0,$1);
 $$0 = $$0$i1;
 STACKTOP = sp;return ($$0|0);
}
function _raccess_guess_linux_netatalk($library,$stream,$base_file_name,$result_file_name,$result_offset) {
 $library = $library|0;
 $stream = $stream|0;
 $base_file_name = $base_file_name|0;
 $result_file_name = $result_file_name|0;
 $result_offset = $result_offset|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i1 = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $args2$i = 0, $stream2$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $args2$i = sp + 8|0;
 $stream2$i = sp;
 $0 = HEAP32[$library>>2]|0;
 $1 = (_raccess_make_file_name($0,$base_file_name,530577)|0);
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = 64;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$args2$i>>2] = 4;
 $3 = ((($args2$i)) + 12|0);
 HEAP32[$3>>2] = $1;
 $4 = (_FT_Stream_New($library,$args2$i,$stream2$i)|0);
 $5 = ($4|0)==(0);
 if ($5) {
  $6 = HEAP32[$stream2$i>>2]|0;
  $7 = ($6|0)==(0|0);
  if ($7) {
   $$0$i$i = 81;$9 = 0;
  } else {
   $8 = (_raccess_guess_apple_generic($6,333319,$result_offset)|0);
   $$pre$i = HEAP32[$stream2$i>>2]|0;
   $$0$i$i = $8;$9 = $$pre$i;
  }
  _FT_Stream_Free($9,0);
  $10 = ($$0$i$i|0)==(0);
  if ($10) {
   HEAP32[$result_file_name>>2] = $1;
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  } else {
   $$0$i1 = $$0$i$i;
  }
 } else {
  $$0$i1 = $4;
 }
 _ft_mem_free($0,$1);
 $$0 = $$0$i1;
 STACKTOP = sp;return ($$0|0);
}
function _ft_raccess_rule_by_darwin_vfs($library,$rule_index) {
 $library = $library|0;
 $rule_index = $rule_index|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $phitmp = 0, $phitmp1 = 0, $phitmp2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($rule_index>>>0)>(8);
 if ($0) {
  $$0$i = 0;
  return ($$0$i|0);
 }
 $1 = (((47800 + ($rule_index<<3)|0)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $phitmp = (($2) + -3)|0;
 $phitmp1 = ($phitmp>>>0)<(2);
 $phitmp2 = $phitmp1&1;
 $$0$i = $phitmp2;
 return ($$0$i|0);
}
function _FT_Stream_OpenMemory($stream,$base,$size) {
 $stream = $stream|0;
 $base = $base|0;
 $size = $size|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$stream>>2] = $base;
 $0 = ((($stream)) + 4|0);
 HEAP32[$0>>2] = $size;
 $1 = ((($stream)) + 8|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($stream)) + 32|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($stream)) + 20|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($stream)) + 24|0);
 HEAP32[$4>>2] = 0;
 return;
}
function _FT_Stream_Close($stream) {
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($stream|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($stream)) + 24|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  return;
 }
 FUNCTION_TABLE_vi[$2 & 8191]($stream);
 return;
}
function _FT_Stream_Seek($stream,$pos) {
 $stream = $stream|0;
 $pos = $pos|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $error$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $5 = ((($stream)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6>>>0)<($pos>>>0);
  if ($7) {
   $error$03 = 85;
   return ($error$03|0);
  }
 } else {
  $3 = (FUNCTION_TABLE_iiiii[$1 & 8191]($stream,$pos,0,0)|0);
  $4 = ($3|0)==(0);
  if (!($4)) {
   $error$03 = 85;
   return ($error$03|0);
  }
 }
 $8 = ((($stream)) + 8|0);
 HEAP32[$8>>2] = $pos;
 $error$03 = 0;
 return ($error$03|0);
}
function _FT_Stream_Skip($stream,$distance) {
 $stream = $stream|0;
 $distance = $distance|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($distance|0)<(0);
 if ($0) {
  $$0 = 85;
  return ($$0|0);
 }
 $1 = ((($stream)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($2) + ($distance))|0;
 $4 = ((($stream)) + 20|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if ($6) {
  $9 = ((($stream)) + 4|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ($10>>>0)<($3>>>0);
  if ($11) {
   $$0 = 85;
   return ($$0|0);
  }
 } else {
  $7 = (FUNCTION_TABLE_iiiii[$5 & 8191]($stream,$3,0,0)|0);
  $8 = ($7|0)==(0);
  if (!($8)) {
   $$0 = 85;
   return ($$0|0);
  }
 }
 HEAP32[$1>>2] = $3;
 $$0 = 0;
 return ($$0|0);
}
function _FT_Stream_Pos($stream) {
 $stream = $stream|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function _FT_Stream_Read($stream,$buffer,$count) {
 $stream = $stream|0;
 $buffer = $buffer|0;
 $count = $count|0;
 var $$$i = 0, $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $count$$i = 0, $read_bytes$1$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($stream)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3>>>0)>($1>>>0);
 if (!($4)) {
  $$0$i = 85;
  return ($$0$i|0);
 }
 $5 = ((($stream)) + 20|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0|0);
 if ($7) {
  $9 = (($3) - ($1))|0;
  $10 = ($9>>>0)>($count>>>0);
  $count$$i = $10 ? $count : $9;
  $11 = HEAP32[$stream>>2]|0;
  $12 = (($11) + ($1)|0);
  _memcpy(($buffer|0),($12|0),($count$$i|0))|0;
  $read_bytes$1$i = $count$$i;
 } else {
  $8 = (FUNCTION_TABLE_iiiii[$6 & 8191]($stream,$1,$buffer,$count)|0);
  $read_bytes$1$i = $8;
 }
 $13 = (($read_bytes$1$i) + ($1))|0;
 HEAP32[$0>>2] = $13;
 $14 = ($read_bytes$1$i>>>0)<($count>>>0);
 $$$i = $14 ? 85 : 0;
 $$0$i = $$$i;
 return ($$0$i|0);
}
function _FT_Stream_ReadAt($stream,$pos,$buffer,$count) {
 $stream = $stream|0;
 $pos = $pos|0;
 $buffer = $buffer|0;
 $count = $count|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $count$ = 0, $read_bytes$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1>>>0)>($pos>>>0);
 if (!($2)) {
  $$0 = 85;
  return ($$0|0);
 }
 $3 = ((($stream)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if ($5) {
  $7 = (($1) - ($pos))|0;
  $8 = ($7>>>0)>($count>>>0);
  $count$ = $8 ? $count : $7;
  $9 = HEAP32[$stream>>2]|0;
  $10 = (($9) + ($pos)|0);
  _memcpy(($buffer|0),($10|0),($count$|0))|0;
  $read_bytes$1 = $count$;
 } else {
  $6 = (FUNCTION_TABLE_iiiii[$4 & 8191]($stream,$pos,$buffer,$count)|0);
  $read_bytes$1 = $6;
 }
 $11 = (($read_bytes$1) + ($pos))|0;
 $12 = ((($stream)) + 8|0);
 HEAP32[$12>>2] = $11;
 $13 = ($read_bytes$1>>>0)<($count>>>0);
 $$ = $13 ? 85 : 0;
 $$0 = $$;
 return ($$0|0);
}
function _FT_Stream_TryRead($stream,$buffer,$count) {
 $stream = $stream|0;
 $buffer = $buffer|0;
 $count = $count|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $count$ = 0, $read_bytes$1 = 0, $read_bytes$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($stream)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3>>>0)>($1>>>0);
 if (!($4)) {
  $read_bytes$2 = 0;
  return ($read_bytes$2|0);
 }
 $5 = ((($stream)) + 20|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0|0);
 if ($7) {
  $9 = (($3) - ($1))|0;
  $10 = ($9>>>0)>($count>>>0);
  $count$ = $10 ? $count : $9;
  $11 = HEAP32[$stream>>2]|0;
  $12 = (($11) + ($1)|0);
  _memcpy(($buffer|0),($12|0),($count$|0))|0;
  $read_bytes$1 = $count$;
 } else {
  $8 = (FUNCTION_TABLE_iiiii[$6 & 8191]($stream,$1,$buffer,$count)|0);
  $read_bytes$1 = $8;
 }
 $13 = HEAP32[$0>>2]|0;
 $14 = (($13) + ($read_bytes$1))|0;
 HEAP32[$0>>2] = $14;
 $read_bytes$2 = $read_bytes$1;
 return ($read_bytes$2|0);
}
function _FT_Stream_ExtractFrame($stream,$count,$pbytes) {
 $stream = $stream|0;
 $count = $count|0;
 $pbytes = $pbytes|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_FT_Stream_EnterFrame($stream,$count)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  return ($0|0);
 }
 $2 = ((($stream)) + 32|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$pbytes>>2] = $3;
 HEAP32[$2>>2] = 0;
 $4 = ((($stream)) + 36|0);
 HEAP32[$4>>2] = 0;
 return ($0|0);
}
function _FT_Stream_EnterFrame($stream,$count) {
 $stream = $stream|0;
 $count = $count|0;
 var $$pre3 = 0, $$pre3$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0;
 var $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 HEAP32[$error>>2] = 0;
 $0 = ((($stream)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $23 = ((($stream)) + 8|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = ((($stream)) + 4|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = ($26>>>0)<=($24>>>0);
  $28 = (($26) - ($24))|0;
  $29 = ($28>>>0)<($count>>>0);
  $or$cond = $27 | $29;
  if ($or$cond) {
   HEAP32[$error>>2] = 85;
   $36 = 85;
   STACKTOP = sp;return ($36|0);
  } else {
   $30 = HEAP32[$stream>>2]|0;
   $31 = (($30) + ($24)|0);
   $32 = ((($stream)) + 32|0);
   HEAP32[$32>>2] = $31;
   $33 = (($31) + ($count)|0);
   $34 = ((($stream)) + 36|0);
   HEAP32[$34>>2] = $33;
   $35 = (($24) + ($count))|0;
   HEAP32[$23>>2] = $35;
   $36 = 0;
   STACKTOP = sp;return ($36|0);
  }
 }
 $3 = ((($stream)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($stream)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6>>>0)<($count>>>0);
 if ($7) {
  HEAP32[$error>>2] = 85;
  $36 = 85;
  STACKTOP = sp;return ($36|0);
 }
 $8 = (_ft_mem_qalloc($4,$count,$error)|0);
 HEAP32[$stream>>2] = $8;
 $9 = HEAP32[$error>>2]|0;
 $10 = ($9|0)==(0);
 if (!($10)) {
  $36 = $9;
  STACKTOP = sp;return ($36|0);
 }
 $11 = HEAP32[$0>>2]|0;
 $12 = ((($stream)) + 8|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (FUNCTION_TABLE_iiiii[$11 & 8191]($stream,$13,$8,$count)|0);
 $15 = ($14>>>0)<($count>>>0);
 $16 = HEAP32[$stream>>2]|0;
 if ($15) {
  _ft_mem_free($4,$16);
  HEAP32[$stream>>2] = 0;
  HEAP32[$error>>2] = 85;
  $$pre3 = 85;$18 = 0;
 } else {
  $$pre3$pre = HEAP32[$error>>2]|0;
  $$pre3 = $$pre3$pre;$18 = $16;
 }
 $17 = ((($stream)) + 32|0);
 HEAP32[$17>>2] = $18;
 $19 = (($18) + ($count)|0);
 $20 = ((($stream)) + 36|0);
 HEAP32[$20>>2] = $19;
 $21 = HEAP32[$12>>2]|0;
 $22 = (($21) + ($14))|0;
 HEAP32[$12>>2] = $22;
 $36 = $$pre3;
 STACKTOP = sp;return ($36|0);
}
function _FT_Stream_ReleaseFrame($stream,$pbytes) {
 $stream = $stream|0;
 $pbytes = $pbytes|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($stream|0)==(0|0);
 if (!($0)) {
  $1 = ((($stream)) + 20|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ($2|0)==(0|0);
  if (!($3)) {
   $4 = ((($stream)) + 28|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = HEAP32[$pbytes>>2]|0;
   _ft_mem_free($5,$6);
   HEAP32[$pbytes>>2] = 0;
  }
 }
 HEAP32[$pbytes>>2] = 0;
 return;
}
function _FT_Stream_ExitFrame($stream) {
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = ((($stream)) + 28|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = HEAP32[$stream>>2]|0;
  _ft_mem_free($4,$5);
  HEAP32[$stream>>2] = 0;
 }
 $6 = ((($stream)) + 32|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($stream)) + 36|0);
 HEAP32[$7>>2] = 0;
 return;
}
function _FT_Stream_GetChar($stream) {
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($stream)) + 36|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1>>>0)<($3>>>0);
 if (!($4)) {
  $result$0 = 0;
  return ($result$0|0);
 }
 $5 = ((($1)) + 1|0);
 HEAP32[$0>>2] = $5;
 $6 = HEAP8[$1>>0]|0;
 $result$0 = $6;
 return ($result$0|0);
}
function _FT_Stream_GetUShort($stream) {
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p$0 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 1|0);
 $3 = ((($stream)) + 36|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2>>>0)<($4>>>0);
 if (!($5)) {
  $p$0 = $1;$result$0 = 0;
  HEAP32[$0>>2] = $p$0;
  return ($result$0|0);
 }
 $6 = ((($1)) + 2|0);
 $7 = HEAP8[$1>>0]|0;
 $8 = $7&255;
 $9 = $8 << 8;
 $10 = HEAP8[$2>>0]|0;
 $11 = $10&255;
 $12 = $9 | $11;
 $13 = $12&65535;
 $p$0 = $6;$result$0 = $13;
 HEAP32[$0>>2] = $p$0;
 return ($result$0|0);
}
function _FT_Stream_GetUShortLE($stream) {
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p$0 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 1|0);
 $3 = ((($stream)) + 36|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2>>>0)<($4>>>0);
 if (!($5)) {
  $p$0 = $1;$result$0 = 0;
  HEAP32[$0>>2] = $p$0;
  return ($result$0|0);
 }
 $6 = ((($1)) + 2|0);
 $7 = HEAP8[$2>>0]|0;
 $8 = $7&255;
 $9 = $8 << 8;
 $10 = HEAP8[$1>>0]|0;
 $11 = $10&255;
 $12 = $9 | $11;
 $13 = $12&65535;
 $p$0 = $6;$result$0 = $13;
 HEAP32[$0>>2] = $p$0;
 return ($result$0|0);
}
function _FT_Stream_GetULong($stream) {
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $p$0 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 3|0);
 $3 = ((($stream)) + 36|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2>>>0)<($4>>>0);
 if (!($5)) {
  $p$0 = $1;$result$0 = 0;
  HEAP32[$0>>2] = $p$0;
  return ($result$0|0);
 }
 $6 = ((($1)) + 4|0);
 $7 = HEAP8[$1>>0]|0;
 $8 = $7&255;
 $9 = $8 << 24;
 $10 = ((($1)) + 1|0);
 $11 = HEAP8[$10>>0]|0;
 $12 = $11&255;
 $13 = $12 << 16;
 $14 = $13 | $9;
 $15 = ((($1)) + 2|0);
 $16 = HEAP8[$15>>0]|0;
 $17 = $16&255;
 $18 = $17 << 8;
 $19 = $14 | $18;
 $20 = HEAP8[$2>>0]|0;
 $21 = $20&255;
 $22 = $19 | $21;
 $p$0 = $6;$result$0 = $22;
 HEAP32[$0>>2] = $p$0;
 return ($result$0|0);
}
function _FT_Stream_GetULongLE($stream) {
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $p$0 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 3|0);
 $3 = ((($stream)) + 36|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2>>>0)<($4>>>0);
 if (!($5)) {
  $p$0 = $1;$result$0 = 0;
  HEAP32[$0>>2] = $p$0;
  return ($result$0|0);
 }
 $6 = ((($1)) + 4|0);
 $7 = HEAP8[$2>>0]|0;
 $8 = $7&255;
 $9 = $8 << 24;
 $10 = ((($1)) + 2|0);
 $11 = HEAP8[$10>>0]|0;
 $12 = $11&255;
 $13 = $12 << 16;
 $14 = $13 | $9;
 $15 = ((($1)) + 1|0);
 $16 = HEAP8[$15>>0]|0;
 $17 = $16&255;
 $18 = $17 << 8;
 $19 = $14 | $18;
 $20 = HEAP8[$1>>0]|0;
 $21 = $20&255;
 $22 = $19 | $21;
 $p$0 = $6;$result$0 = $22;
 HEAP32[$0>>2] = $p$0;
 return ($result$0|0);
}
function _FT_Stream_ReadChar($stream,$error) {
 $stream = $stream|0;
 $error = $error|0;
 var $$0 = 0, $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $result = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $result = sp;
 HEAP8[$result>>0] = 0;
 HEAP32[$error>>2] = 0;
 $0 = ((($stream)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 $3 = ((($stream)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 if ($2) {
  $7 = ((($stream)) + 4|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($4>>>0)<($8>>>0);
  if ($9) {
   $10 = HEAP32[$stream>>2]|0;
   $11 = (($10) + ($4)|0);
   $12 = HEAP8[$11>>0]|0;
   HEAP8[$result>>0] = $12;
   $14 = $4;$15 = $12;
   label = 6;
  } else {
   label = 7;
  }
 } else {
  $5 = (FUNCTION_TABLE_iiiii[$1 & 8191]($stream,$4,$result,1)|0);
  $6 = ($5|0)==(1);
  if ($6) {
   $$pre = HEAP32[$3>>2]|0;
   $$pre1 = HEAP8[$result>>0]|0;
   $14 = $$pre;$15 = $$pre1;
   label = 6;
  } else {
   label = 7;
  }
 }
 if ((label|0) == 6) {
  $13 = (($14) + 1)|0;
  HEAP32[$3>>2] = $13;
  $$0 = $15;
  STACKTOP = sp;return ($$0|0);
 }
 else if ((label|0) == 7) {
  HEAP32[$error>>2] = 85;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _FT_Stream_ReadUShort($stream,$error) {
 $stream = $stream|0;
 $error = $error|0;
 var $$0 = 0, $$pre = 0, $$pre$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p$01 = 0, $reads = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $reads = sp;
 HEAP32[$error>>2] = 0;
 $0 = ((($stream)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + 1)|0;
 $3 = ((($stream)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2>>>0)<($4>>>0);
 do {
  if ($5) {
   $6 = ((($stream)) + 20|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0|0);
   if ($8) {
    $11 = HEAP32[$stream>>2]|0;
    $12 = (($11) + ($1)|0);
    $13 = ($12|0)==(0|0);
    if ($13) {
     $23 = $1;$result$0 = 0;
    } else {
     $$pre = $1;$p$01 = $12;
     label = 6;
    }
   } else {
    $9 = (FUNCTION_TABLE_iiiii[$7 & 8191]($stream,$1,$reads,2)|0);
    $10 = ($9|0)==(2);
    if (!($10)) {
     break;
    }
    $$pre$pre = HEAP32[$0>>2]|0;
    $$pre = $$pre$pre;$p$01 = $reads;
    label = 6;
   }
   if ((label|0) == 6) {
    $14 = HEAP8[$p$01>>0]|0;
    $15 = $14&255;
    $16 = $15 << 8;
    $17 = ((($p$01)) + 1|0);
    $18 = HEAP8[$17>>0]|0;
    $19 = $18&255;
    $20 = $16 | $19;
    $21 = $20&65535;
    $23 = $$pre;$result$0 = $21;
   }
   $22 = (($23) + 2)|0;
   HEAP32[$0>>2] = $22;
   $$0 = $result$0;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 HEAP32[$error>>2] = 85;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _FT_Stream_ReadUShortLE($stream,$error) {
 $stream = $stream|0;
 $error = $error|0;
 var $$0 = 0, $$pre = 0, $$pre$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p$01 = 0, $reads = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $reads = sp;
 HEAP32[$error>>2] = 0;
 $0 = ((($stream)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + 1)|0;
 $3 = ((($stream)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2>>>0)<($4>>>0);
 do {
  if ($5) {
   $6 = ((($stream)) + 20|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0|0);
   if ($8) {
    $11 = HEAP32[$stream>>2]|0;
    $12 = (($11) + ($1)|0);
    $13 = ($12|0)==(0|0);
    if ($13) {
     $23 = $1;$result$0 = 0;
    } else {
     $$pre = $1;$p$01 = $12;
     label = 6;
    }
   } else {
    $9 = (FUNCTION_TABLE_iiiii[$7 & 8191]($stream,$1,$reads,2)|0);
    $10 = ($9|0)==(2);
    if (!($10)) {
     break;
    }
    $$pre$pre = HEAP32[$0>>2]|0;
    $$pre = $$pre$pre;$p$01 = $reads;
    label = 6;
   }
   if ((label|0) == 6) {
    $14 = ((($p$01)) + 1|0);
    $15 = HEAP8[$14>>0]|0;
    $16 = $15&255;
    $17 = $16 << 8;
    $18 = HEAP8[$p$01>>0]|0;
    $19 = $18&255;
    $20 = $17 | $19;
    $21 = $20&65535;
    $23 = $$pre;$result$0 = $21;
   }
   $22 = (($23) + 2)|0;
   HEAP32[$0>>2] = $22;
   $$0 = $result$0;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 HEAP32[$error>>2] = 85;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _FT_Stream_ReadUOffset($stream,$error) {
 $stream = $stream|0;
 $error = $error|0;
 var $$0 = 0, $$pre = 0, $$pre$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p$01 = 0, $reads = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $reads = sp;
 HEAP32[$error>>2] = 0;
 $0 = ((($stream)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + 2)|0;
 $3 = ((($stream)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2>>>0)<($4>>>0);
 do {
  if ($5) {
   $6 = ((($stream)) + 20|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0|0);
   if ($8) {
    $11 = HEAP32[$stream>>2]|0;
    $12 = (($11) + ($1)|0);
    $13 = ($12|0)==(0|0);
    if ($13) {
     $27 = $1;$result$0 = 0;
    } else {
     $$pre = $1;$p$01 = $12;
     label = 6;
    }
   } else {
    $9 = (FUNCTION_TABLE_iiiii[$7 & 8191]($stream,$1,$reads,3)|0);
    $10 = ($9|0)==(3);
    if (!($10)) {
     break;
    }
    $$pre$pre = HEAP32[$0>>2]|0;
    $$pre = $$pre$pre;$p$01 = $reads;
    label = 6;
   }
   if ((label|0) == 6) {
    $14 = HEAP8[$p$01>>0]|0;
    $15 = $14&255;
    $16 = $15 << 16;
    $17 = ((($p$01)) + 1|0);
    $18 = HEAP8[$17>>0]|0;
    $19 = $18&255;
    $20 = $19 << 8;
    $21 = $20 | $16;
    $22 = ((($p$01)) + 2|0);
    $23 = HEAP8[$22>>0]|0;
    $24 = $23&255;
    $25 = $21 | $24;
    $27 = $$pre;$result$0 = $25;
   }
   $26 = (($27) + 3)|0;
   HEAP32[$0>>2] = $26;
   $$0 = $result$0;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 HEAP32[$error>>2] = 85;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _FT_Stream_ReadULong($stream,$error) {
 $stream = $stream|0;
 $error = $error|0;
 var $$0 = 0, $$pre = 0, $$pre$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p$01 = 0, $reads = 0, $result$0 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $reads = sp;
 HEAP32[$error>>2] = 0;
 $0 = ((($stream)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + 3)|0;
 $3 = ((($stream)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2>>>0)<($4>>>0);
 do {
  if ($5) {
   $6 = ((($stream)) + 20|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0|0);
   if ($8) {
    $11 = HEAP32[$stream>>2]|0;
    $12 = (($11) + ($1)|0);
    $13 = ($12|0)==(0|0);
    if ($13) {
     $32 = $1;$result$0 = 0;
    } else {
     $$pre = $1;$p$01 = $12;
     label = 6;
    }
   } else {
    $9 = (FUNCTION_TABLE_iiiii[$7 & 8191]($stream,$1,$reads,4)|0);
    $10 = ($9|0)==(4);
    if (!($10)) {
     break;
    }
    $$pre$pre = HEAP32[$0>>2]|0;
    $$pre = $$pre$pre;$p$01 = $reads;
    label = 6;
   }
   if ((label|0) == 6) {
    $14 = HEAP8[$p$01>>0]|0;
    $15 = $14&255;
    $16 = $15 << 24;
    $17 = ((($p$01)) + 1|0);
    $18 = HEAP8[$17>>0]|0;
    $19 = $18&255;
    $20 = $19 << 16;
    $21 = $20 | $16;
    $22 = ((($p$01)) + 2|0);
    $23 = HEAP8[$22>>0]|0;
    $24 = $23&255;
    $25 = $24 << 8;
    $26 = $21 | $25;
    $27 = ((($p$01)) + 3|0);
    $28 = HEAP8[$27>>0]|0;
    $29 = $28&255;
    $30 = $26 | $29;
    $32 = $$pre;$result$0 = $30;
   }
   $31 = (($32) + 4)|0;
   HEAP32[$0>>2] = $31;
   $$0 = $result$0;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 HEAP32[$error>>2] = 85;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _FT_Stream_ReadULongLE($stream,$error) {
 $stream = $stream|0;
 $error = $error|0;
 var $$0 = 0, $$pre = 0, $$pre$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p$01 = 0, $reads = 0, $result$0 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $reads = sp;
 HEAP32[$error>>2] = 0;
 $0 = ((($stream)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + 3)|0;
 $3 = ((($stream)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2>>>0)<($4>>>0);
 do {
  if ($5) {
   $6 = ((($stream)) + 20|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0|0);
   if ($8) {
    $11 = HEAP32[$stream>>2]|0;
    $12 = (($11) + ($1)|0);
    $13 = ($12|0)==(0|0);
    if ($13) {
     $32 = $1;$result$0 = 0;
    } else {
     $$pre = $1;$p$01 = $12;
     label = 6;
    }
   } else {
    $9 = (FUNCTION_TABLE_iiiii[$7 & 8191]($stream,$1,$reads,4)|0);
    $10 = ($9|0)==(4);
    if (!($10)) {
     break;
    }
    $$pre$pre = HEAP32[$0>>2]|0;
    $$pre = $$pre$pre;$p$01 = $reads;
    label = 6;
   }
   if ((label|0) == 6) {
    $14 = ((($p$01)) + 3|0);
    $15 = HEAP8[$14>>0]|0;
    $16 = $15&255;
    $17 = $16 << 24;
    $18 = ((($p$01)) + 2|0);
    $19 = HEAP8[$18>>0]|0;
    $20 = $19&255;
    $21 = $20 << 16;
    $22 = $21 | $17;
    $23 = ((($p$01)) + 1|0);
    $24 = HEAP8[$23>>0]|0;
    $25 = $24&255;
    $26 = $25 << 8;
    $27 = $22 | $26;
    $28 = HEAP8[$p$01>>0]|0;
    $29 = $28&255;
    $30 = $27 | $29;
    $32 = $$pre;$result$0 = $30;
   }
   $31 = (($32) + 4)|0;
   HEAP32[$0>>2] = $31;
   $$0 = $result$0;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 HEAP32[$error>>2] = 85;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _FT_Stream_ReadFields($stream,$fields,$structure) {
 $stream = $stream|0;
 $fields = $fields|0;
 $structure = $structure|0;
 var $$0 = 0, $$02 = 0, $$02$lcssa46 = 0, $$02$ph = 0, $$mask = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cursor$0 = 0, $cursor$0$be = 0, $cursor$0$lcssa = 0, $cursor$0$ph = 0, $cursor$2 = 0, $error$2 = 0, $frame_accessed$0$ph = 0, $frame_accessed$0$ph$lcssa = 0, $frame_accessed$0$ph52 = 0, $sign_shift$0 = 0, $value$0 = 0, $value$1 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($fields|0)==(0|0);
 if ($0) {
  $$0 = 6;
  return ($$0|0);
 }
 $1 = ($stream|0)==(0|0);
 if ($1) {
  $$0 = 40;
  return ($$0|0);
 }
 $2 = ((($stream)) + 32|0);
 $3 = ((($stream)) + 36|0);
 $$02$ph = $fields;$frame_accessed$0$ph = 0;
 L7: while(1) {
  $cursor$0$ph = HEAP32[$2>>2]|0;
  $$02 = $$02$ph;$cursor$0 = $cursor$0$ph;
  L9: while(1) {
   $4 = HEAP16[$$02>>1]|0;
   $5 = $4&255;
   $$mask = $4 & 255;
   $6 = $$mask&65535;
   $7 = ($4&65535) >>> 8;
   $8 = $7&65535;
   switch ($6|0) {
   case 4:  {
    $$02$lcssa46 = $$02;
    break L9;
    break;
   }
   case 25: case 24:  {
    $15 = (($cursor$0) + ($8)|0);
    $16 = HEAP32[$3>>2]|0;
    $17 = ($15>>>0)>($16>>>0);
    if ($17) {
     $error$2 = 85;$frame_accessed$0$ph52 = $frame_accessed$0$ph;
     break L7;
    }
    $18 = ($5<<24>>24)==(24);
    if ($18) {
     $19 = ((($$02)) + 2|0);
     $20 = HEAP16[$19>>1]|0;
     $21 = $20&65535;
     $22 = (($structure) + ($21)|0);
     _memcpy(($22|0),($cursor$0|0),($8|0))|0;
     $cursor$0$be = $15;
    } else {
     $cursor$0$be = $15;
    }
    break;
   }
   case 9: case 8:  {
    $23 = ((($cursor$0)) + 1|0);
    $24 = HEAP8[$cursor$0>>0]|0;
    $25 = $24&255;
    $cursor$2 = $23;$sign_shift$0 = 24;$value$0 = $25;
    label = 18;
    break;
   }
   case 12: case 13:  {
    $26 = ((($cursor$0)) + 2|0);
    $27 = HEAP8[$cursor$0>>0]|0;
    $28 = $27&255;
    $29 = $28 << 8;
    $30 = ((($cursor$0)) + 1|0);
    $31 = HEAP8[$30>>0]|0;
    $32 = $31&255;
    $33 = $29 | $32;
    $cursor$2 = $26;$sign_shift$0 = 16;$value$0 = $33;
    label = 18;
    break;
   }
   case 14: case 15:  {
    $34 = ((($cursor$0)) + 2|0);
    $35 = ((($cursor$0)) + 1|0);
    $36 = HEAP8[$35>>0]|0;
    $37 = $36&255;
    $38 = $37 << 8;
    $39 = HEAP8[$cursor$0>>0]|0;
    $40 = $39&255;
    $41 = $38 | $40;
    $cursor$2 = $34;$sign_shift$0 = 16;$value$0 = $41;
    label = 18;
    break;
   }
   case 16: case 17:  {
    $42 = ((($cursor$0)) + 4|0);
    $43 = HEAP8[$cursor$0>>0]|0;
    $44 = $43&255;
    $45 = $44 << 24;
    $46 = ((($cursor$0)) + 1|0);
    $47 = HEAP8[$46>>0]|0;
    $48 = $47&255;
    $49 = $48 << 16;
    $50 = $49 | $45;
    $51 = ((($cursor$0)) + 2|0);
    $52 = HEAP8[$51>>0]|0;
    $53 = $52&255;
    $54 = $53 << 8;
    $55 = $50 | $54;
    $56 = ((($cursor$0)) + 3|0);
    $57 = HEAP8[$56>>0]|0;
    $58 = $57&255;
    $59 = $55 | $58;
    $cursor$2 = $42;$sign_shift$0 = 0;$value$0 = $59;
    label = 18;
    break;
   }
   case 18: case 19:  {
    $60 = ((($cursor$0)) + 4|0);
    $61 = ((($cursor$0)) + 3|0);
    $62 = HEAP8[$61>>0]|0;
    $63 = $62&255;
    $64 = $63 << 24;
    $65 = ((($cursor$0)) + 2|0);
    $66 = HEAP8[$65>>0]|0;
    $67 = $66&255;
    $68 = $67 << 16;
    $69 = $68 | $64;
    $70 = ((($cursor$0)) + 1|0);
    $71 = HEAP8[$70>>0]|0;
    $72 = $71&255;
    $73 = $72 << 8;
    $74 = $69 | $73;
    $75 = HEAP8[$cursor$0>>0]|0;
    $76 = $75&255;
    $77 = $74 | $76;
    $cursor$2 = $60;$sign_shift$0 = 0;$value$0 = $77;
    label = 18;
    break;
   }
   case 20: case 21:  {
    $78 = ((($cursor$0)) + 3|0);
    $79 = HEAP8[$cursor$0>>0]|0;
    $80 = $79&255;
    $81 = $80 << 16;
    $82 = ((($cursor$0)) + 1|0);
    $83 = HEAP8[$82>>0]|0;
    $84 = $83&255;
    $85 = $84 << 8;
    $86 = $85 | $81;
    $87 = ((($cursor$0)) + 2|0);
    $88 = HEAP8[$87>>0]|0;
    $89 = $88&255;
    $90 = $86 | $89;
    $cursor$2 = $78;$sign_shift$0 = 8;$value$0 = $90;
    label = 18;
    break;
   }
   case 22: case 23:  {
    $91 = ((($cursor$0)) + 3|0);
    $92 = ((($cursor$0)) + 2|0);
    $93 = HEAP8[$92>>0]|0;
    $94 = $93&255;
    $95 = $94 << 16;
    $96 = ((($cursor$0)) + 1|0);
    $97 = HEAP8[$96>>0]|0;
    $98 = $97&255;
    $99 = $98 << 8;
    $100 = $99 | $95;
    $101 = HEAP8[$cursor$0>>0]|0;
    $102 = $101&255;
    $103 = $100 | $102;
    $cursor$2 = $91;$sign_shift$0 = 8;$value$0 = $103;
    label = 18;
    break;
   }
   default: {
    $cursor$0$lcssa = $cursor$0;$frame_accessed$0$ph$lcssa = $frame_accessed$0$ph;
    label = 17;
    break L7;
   }
   }
   L22: do {
    if ((label|0) == 18) {
     label = 0;
     $104 = $5 & 1;
     $105 = ($104<<24>>24)==(0);
     $106 = $value$0 << $sign_shift$0;
     $107 = $106 >> $sign_shift$0;
     $value$1 = $105 ? $value$0 : $107;
     $108 = ((($$02)) + 2|0);
     $109 = HEAP16[$108>>1]|0;
     $110 = $109&65535;
     $111 = (($structure) + ($110)|0);
     $112 = $7&65535;
     switch ($112|0) {
     case 1:  {
      $113 = $value$1&255;
      HEAP8[$111>>0] = $113;
      $cursor$0$be = $cursor$2;
      break L22;
      break;
     }
     case 2:  {
      $114 = $value$1&65535;
      HEAP16[$111>>1] = $114;
      $cursor$0$be = $cursor$2;
      break L22;
      break;
     }
     case 4:  {
      HEAP32[$111>>2] = $value$1;
      $cursor$0$be = $cursor$2;
      break L22;
      break;
     }
     default: {
      HEAP32[$111>>2] = $value$1;
      $cursor$0$be = $cursor$2;
      break L22;
     }
     }
    }
   } while(0);
   $115 = ((($$02)) + 4|0);
   $$02 = $115;$cursor$0 = $cursor$0$be;
  }
  $9 = ((($$02$lcssa46)) + 2|0);
  $10 = HEAP16[$9>>1]|0;
  $11 = $10&65535;
  $12 = (_FT_Stream_EnterFrame($stream,$11)|0);
  $13 = ($12|0)==(0);
  $14 = ((($$02$lcssa46)) + 4|0);
  if ($13) {
   $$02$ph = $14;$frame_accessed$0$ph = 1;
  } else {
   $error$2 = $12;$frame_accessed$0$ph52 = $frame_accessed$0$ph;
   break;
  }
 }
 if ((label|0) == 17) {
  HEAP32[$2>>2] = $cursor$0$lcssa;
  $error$2 = 0;$frame_accessed$0$ph52 = $frame_accessed$0$ph$lcssa;
 }
 $116 = ($frame_accessed$0$ph52<<24>>24)==(0);
 if ($116) {
  $$0 = $error$2;
  return ($$0|0);
 }
 $117 = ((($stream)) + 20|0);
 $118 = HEAP32[$117>>2]|0;
 $119 = ($118|0)==(0|0);
 if (!($119)) {
  $120 = ((($stream)) + 28|0);
  $121 = HEAP32[$120>>2]|0;
  $122 = HEAP32[$stream>>2]|0;
  _ft_mem_free($121,$122);
  HEAP32[$stream>>2] = 0;
 }
 HEAP32[$2>>2] = 0;
 HEAP32[$3>>2] = 0;
 $$0 = $error$2;
 return ($$0|0);
}
function _FT_Stream_Open($stream,$filepathname) {
 $stream = $stream|0;
 $filepathname = $filepathname|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($stream|0)==(0|0);
 if ($0) {
  $$0 = 40;
  return ($$0|0);
 }
 $1 = ((($stream)) + 12|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($stream)) + 16|0);
 HEAP32[$2>>2] = $filepathname;
 HEAP32[$stream>>2] = 0;
 $3 = ((($stream)) + 8|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($stream)) + 20|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($stream)) + 24|0);
 HEAP32[$5>>2] = 0;
 $6 = (_fopen($filepathname,530591)|0);
 $7 = ($6|0)==(0|0);
 if ($7) {
  $$0 = 1;
  return ($$0|0);
 }
 (_fseek($6,0,2)|0);
 $8 = (_ftell($6)|0);
 $9 = ((($stream)) + 4|0);
 HEAP32[$9>>2] = $8;
 $10 = ($8|0)==(0);
 if ($10) {
  (_fclose($6)|0);
  $$0 = 81;
  return ($$0|0);
 } else {
  (_fseek($6,0,0)|0);
  HEAP32[$1>>2] = $6;
  HEAP32[$4>>2] = 4347;
  HEAP32[$5>>2] = 4348;
  $$0 = 0;
  return ($$0|0);
 }
 return (0)|0;
}
function _ft_ansi_stream_io($stream,$offset,$buffer,$count) {
 $stream = $stream|0;
 $offset = $offset|0;
 $buffer = $buffer|0;
 $count = $count|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($count|0)==(0);
 if ($0) {
  $1 = ((($stream)) + 4|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ($2>>>0)<($offset>>>0);
  if ($3) {
   $$0 = 1;
   return ($$0|0);
  }
 }
 $4 = ((($stream)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($stream)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==($offset|0);
 if (!($8)) {
  (_fseek($5,$offset,0)|0);
 }
 $9 = (_fread($buffer,1,$count,$5)|0);
 $$0 = $9;
 return ($$0|0);
}
function _ft_ansi_stream_close($stream) {
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 (_fclose($1)|0);
 HEAP32[$0>>2] = 0;
 $2 = ((($stream)) + 4|0);
 HEAP32[$2>>2] = 0;
 HEAP32[$stream>>2] = 0;
 return;
}
function _FT_New_Memory() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_malloc(16)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  return ($0|0);
 }
 HEAP32[$0>>2] = 0;
 $2 = ((($0)) + 4|0);
 HEAP32[$2>>2] = 4349;
 $3 = ((($0)) + 12|0);
 HEAP32[$3>>2] = 4350;
 $4 = ((($0)) + 8|0);
 HEAP32[$4>>2] = 4351;
 return ($0|0);
}
function _ft_alloc($memory,$size) {
 $memory = $memory|0;
 $size = $size|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_malloc($size)|0);
 return ($0|0);
}
function _ft_realloc($memory,$cur_size,$new_size,$block) {
 $memory = $memory|0;
 $cur_size = $cur_size|0;
 $new_size = $new_size|0;
 $block = $block|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_realloc($block,$new_size)|0);
 return ($0|0);
}
function _ft_free($memory,$block) {
 $memory = $memory|0;
 $block = $block|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _free($block);
 return;
}
function _FT_Done_Memory($memory) {
 $memory = $memory|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _free($memory);
 return;
}
function _FT_Vector_Length($vec) {
 $vec = $vec|0;
 var $$ = 0, $$0 = 0, $$1$i = 0, $$2$i = 0, $$val$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $9 = 0, $b$07$i = 0, $exitcond$i = 0, $i$04$i = 0, $shift$0$i = 0, $v$sroa$0$0 = 0, $v$sroa$11$0 = 0, $v$sroa$11$0$ = 0, $x$15$i = 0, $x$2$i = 0, $x$2$i$lcssa = 0, $y$16$i = 0, $y$2$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($vec|0)==(0|0);
 if ($0) {
  $$0 = 0;
  return ($$0|0);
 }
 $1 = $vec;
 $2 = $1;
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + 4)|0;
 $5 = $4;
 $6 = HEAP32[$5>>2]|0;
 $7 = ($3|0)==(0);
 if ($7) {
  $8 = ($6|0)<(0);
  $9 = (0 - ($6))|0;
  $10 = $8 ? $9 : $6;
  $$0 = $10;
  return ($$0|0);
 }
 $11 = ($6|0)==(0);
 $12 = ($3|0)<(0);
 $13 = (0 - ($3))|0;
 $14 = $12 ? $13 : $3;
 if ($11) {
  $$0 = $14;
  return ($$0|0);
 }
 $15 = ($6|0)<(0);
 $16 = (0 - ($6))|0;
 $17 = $15 ? $16 : $6;
 $18 = $17 | $14;
 $19 = (Math_clz32(($18|0))|0);
 $20 = (31 - ($19))|0;
 $21 = ($20|0)<(30);
 if ($21) {
  $22 = (29 - ($20))|0;
  $23 = $3 << $22;
  $24 = $6 << $22;
  $shift$0$i = $22;$v$sroa$0$0 = $23;$v$sroa$11$0 = $24;
 } else {
  $25 = (($20) + -29)|0;
  $26 = $3 >> $25;
  $27 = $6 >> $25;
  $28 = (29 - ($20))|0;
  $shift$0$i = $28;$v$sroa$0$0 = $26;$v$sroa$11$0 = $27;
 }
 $29 = ($v$sroa$11$0|0)>($v$sroa$0$0|0);
 $30 = (0 - ($v$sroa$0$0))|0;
 if ($29) {
  $31 = ($v$sroa$11$0|0)>($30|0);
  $32 = (0 - ($v$sroa$11$0))|0;
  $$ = $31 ? $30 : $32;
  $v$sroa$11$0$ = $31 ? $v$sroa$11$0 : $30;
  $b$07$i = 1;$i$04$i = 1;$x$15$i = $v$sroa$11$0$;$y$16$i = $$;
 } else {
  $33 = ($v$sroa$11$0|0)<($30|0);
  $34 = (0 - ($v$sroa$11$0))|0;
  $$1$i = $33 ? $34 : $v$sroa$0$0;
  $$2$i = $33 ? $v$sroa$0$0 : $v$sroa$11$0;
  $b$07$i = 1;$i$04$i = 1;$x$15$i = $$1$i;$y$16$i = $$2$i;
 }
 while(1) {
  $35 = ($y$16$i|0)>(0);
  $36 = (($y$16$i) + ($b$07$i))|0;
  $37 = $36 >> $i$04$i;
  if ($35) {
   $38 = (($37) + ($x$15$i))|0;
   $39 = (($x$15$i) + ($b$07$i))|0;
   $40 = $39 >> $i$04$i;
   $41 = (($y$16$i) - ($40))|0;
   $x$2$i = $38;$y$2$i = $41;
  } else {
   $42 = (($x$15$i) - ($37))|0;
   $43 = (($x$15$i) + ($b$07$i))|0;
   $44 = $43 >> $i$04$i;
   $45 = (($44) + ($y$16$i))|0;
   $x$2$i = $42;$y$2$i = $45;
  }
  $46 = $b$07$i << 1;
  $47 = (($i$04$i) + 1)|0;
  $exitcond$i = ($47|0)==(23);
  if ($exitcond$i) {
   $x$2$i$lcssa = $x$2$i;
   break;
  } else {
   $b$07$i = $46;$i$04$i = $47;$x$15$i = $x$2$i;$y$16$i = $y$2$i;
  }
 }
 $48 = ($x$2$i$lcssa|0)<(0);
 $49 = (0 - ($x$2$i$lcssa))|0;
 $$val$i = $48 ? $49 : $x$2$i$lcssa;
 $50 = $$val$i & 65535;
 $51 = $$val$i >>> 16;
 $52 = ($50*23318)|0;
 $53 = ($50*56281)|0;
 $54 = ($51*23318)|0;
 $55 = ($51*56281)|0;
 $56 = (($53) + ($54))|0;
 $57 = ($56>>>0)<($54>>>0);
 $58 = $57&1;
 $59 = $58 << 16;
 $60 = $56 >>> 16;
 $61 = $56 << 16;
 $62 = (($61) + ($52))|0;
 $63 = ($62>>>0)<($61>>>0);
 $64 = $63&1;
 $65 = ($62>>>0)>(3221225471);
 $66 = $65&1;
 $67 = (($60) + ($55))|0;
 $68 = (($67) + ($59))|0;
 $69 = (($68) + ($64))|0;
 $70 = (($69) + ($66))|0;
 $71 = (0 - ($70))|0;
 $72 = $48 ? $71 : $70;
 $73 = ($shift$0$i|0)>(0);
 if ($73) {
  $74 = (($shift$0$i) + -1)|0;
  $75 = 1 << $74;
  $76 = (($72) + ($75))|0;
  $77 = $76 >> $shift$0$i;
  $$0 = $77;
  return ($$0|0);
 } else {
  $78 = (0 - ($shift$0$i))|0;
  $79 = $72 << $78;
  $$0 = $79;
  return ($$0|0);
 }
 return (0)|0;
}
function _ft_mem_alloc($memory,$size,$p_error) {
 $memory = $memory|0;
 $size = $size|0;
 $p_error = $p_error|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $block$0$i4 = 0, $error$0$i3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($size|0)>(0);
 if ($0) {
  $3 = ((($memory)) + 4|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = (FUNCTION_TABLE_iii[$4 & 8191]($memory,$size)|0);
  $6 = ($5|0)==(0|0);
  if ($6) {
   $block$0$i4 = 0;$error$0$i3 = 64;
  } else {
   _memset(($5|0),0,($size|0))|0;
   $block$0$i4 = $5;$error$0$i3 = 0;
  }
 } else {
  $1 = $size >> 31;
  $2 = $1 & 6;
  $block$0$i4 = 0;$error$0$i3 = $2;
 }
 HEAP32[$p_error>>2] = $error$0$i3;
 return ($block$0$i4|0);
}
function _ft_mem_qalloc($memory,$size,$p_error) {
 $memory = $memory|0;
 $size = $size|0;
 $p_error = $p_error|0;
 var $$ = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $block$0 = 0, $error$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($size|0)>(0);
 if ($0) {
  $1 = ((($memory)) + 4|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = (FUNCTION_TABLE_iii[$2 & 8191]($memory,$size)|0);
  $4 = ($3|0)==(0|0);
  $$ = $4 ? 64 : 0;
  $block$0 = $3;$error$0 = $$;
 } else {
  $5 = $size >> 31;
  $6 = $5 & 6;
  $block$0 = 0;$error$0 = $6;
 }
 HEAP32[$p_error>>2] = $error$0;
 return ($block$0|0);
}
function _ft_mem_realloc($memory,$item_size,$cur_count,$new_count,$block,$p_error) {
 $memory = $memory|0;
 $item_size = $item_size|0;
 $cur_count = $cur_count|0;
 $new_count = $new_count|0;
 $block = $block|0;
 $p_error = $p_error|0;
 var $$$i = 0, $$1$i = 0, $$1$i3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $block$$i = 0, $error$0$i = 0, $error$0$i2 = 0, $or$cond = 0, $or$cond5$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = $cur_count | $item_size;
 $1 = $0 | $new_count;
 $2 = ($1|0)<(0);
 if ($2) {
  $$1$i3 = $block;$error$0$i2 = 6;
  HEAP32[$p_error>>2] = $error$0$i2;
  return ($$1$i3|0);
 }
 $3 = ($new_count|0)==(0);
 $4 = ($item_size|0)==(0);
 $or$cond5$i = $4 | $3;
 do {
  if ($or$cond5$i) {
   $5 = ($block|0)==(0|0);
   if ($5) {
    $$1$i = 0;$error$0$i = 0;
   } else {
    $6 = ((($memory)) + 8|0);
    $7 = HEAP32[$6>>2]|0;
    FUNCTION_TABLE_vii[$7 & 8191]($memory,$block);
    $$1$i = 0;$error$0$i = 0;
   }
  } else {
   $8 = (2147483647 / ($item_size|0))&-1;
   $9 = ($8|0)<($new_count|0);
   if ($9) {
    $$1$i3 = $block;$error$0$i2 = 10;
    HEAP32[$p_error>>2] = $error$0$i2;
    return ($$1$i3|0);
   }
   $10 = ($cur_count|0)==(0);
   if (!($10)) {
    $19 = Math_imul($cur_count, $item_size)|0;
    $20 = Math_imul($new_count, $item_size)|0;
    $21 = ((($memory)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = (FUNCTION_TABLE_iiiii[$22 & 8191]($memory,$19,$20,$block)|0);
    $24 = ($23|0)==(0|0);
    $$$i = $24 ? 64 : 0;
    $block$$i = $24 ? $block : $23;
    $$1$i = $block$$i;$error$0$i = $$$i;
    break;
   }
   $11 = Math_imul($new_count, $item_size)|0;
   $12 = ($11|0)>(0);
   if (!($12)) {
    $13 = $11 >> 31;
    $14 = $13 & 6;
    $$1$i = 0;$error$0$i = $14;
    break;
   }
   $15 = ((($memory)) + 4|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = (FUNCTION_TABLE_iii[$16 & 8191]($memory,$11)|0);
   $18 = ($17|0)==(0|0);
   if ($18) {
    $$1$i3 = 0;$error$0$i2 = 64;
    HEAP32[$p_error>>2] = $error$0$i2;
    return ($$1$i3|0);
   } else {
    _memset(($17|0),0,($11|0))|0;
    $$1$i = $17;$error$0$i = 0;
    break;
   }
  }
 } while(0);
 $25 = ($error$0$i|0)==(0);
 $26 = ($new_count|0)>($cur_count|0);
 $or$cond = $26 & $25;
 if (!($or$cond)) {
  $$1$i3 = $$1$i;$error$0$i2 = $error$0$i;
  HEAP32[$p_error>>2] = $error$0$i2;
  return ($$1$i3|0);
 }
 $27 = Math_imul($cur_count, $item_size)|0;
 $28 = (($$1$i) + ($27)|0);
 $29 = (($new_count) - ($cur_count))|0;
 $30 = Math_imul($29, $item_size)|0;
 _memset(($28|0),0,($30|0))|0;
 $$1$i3 = $$1$i;$error$0$i2 = 0;
 HEAP32[$p_error>>2] = $error$0$i2;
 return ($$1$i3|0);
}
function _ft_mem_qrealloc($memory,$item_size,$cur_count,$new_count,$block,$p_error) {
 $memory = $memory|0;
 $item_size = $item_size|0;
 $cur_count = $cur_count|0;
 $new_count = $new_count|0;
 $block = $block|0;
 $p_error = $p_error|0;
 var $$ = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $block$ = 0, $error$0 = 0, $or$cond5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $cur_count | $item_size;
 $1 = $0 | $new_count;
 $2 = ($1|0)<(0);
 do {
  if ($2) {
   $$1 = $block;$error$0 = 6;
  } else {
   $3 = ($new_count|0)==(0);
   $4 = ($item_size|0)==(0);
   $or$cond5 = $4 | $3;
   if ($or$cond5) {
    $5 = ($block|0)==(0|0);
    if ($5) {
     $$1 = 0;$error$0 = 0;
     break;
    }
    $6 = ((($memory)) + 8|0);
    $7 = HEAP32[$6>>2]|0;
    FUNCTION_TABLE_vii[$7 & 8191]($memory,$block);
    $$1 = 0;$error$0 = 0;
    break;
   }
   $8 = (2147483647 / ($item_size|0))&-1;
   $9 = ($8|0)<($new_count|0);
   if ($9) {
    $$1 = $block;$error$0 = 10;
   } else {
    $10 = ($cur_count|0)==(0);
    if (!($10)) {
     $19 = Math_imul($cur_count, $item_size)|0;
     $20 = Math_imul($new_count, $item_size)|0;
     $21 = ((($memory)) + 12|0);
     $22 = HEAP32[$21>>2]|0;
     $23 = (FUNCTION_TABLE_iiiii[$22 & 8191]($memory,$19,$20,$block)|0);
     $24 = ($23|0)==(0|0);
     $$ = $24 ? 64 : 0;
     $block$ = $24 ? $block : $23;
     $$1 = $block$;$error$0 = $$;
     break;
    }
    $11 = Math_imul($new_count, $item_size)|0;
    $12 = ($11|0)>(0);
    if (!($12)) {
     $13 = $11 >> 31;
     $14 = $13 & 6;
     $$1 = 0;$error$0 = $14;
     break;
    }
    $15 = ((($memory)) + 4|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = (FUNCTION_TABLE_iii[$16 & 8191]($memory,$11)|0);
    $18 = ($17|0)==(0|0);
    if ($18) {
     $$1 = 0;$error$0 = 64;
    } else {
     _memset(($17|0),0,($11|0))|0;
     $$1 = $17;$error$0 = 0;
    }
   }
  }
 } while(0);
 HEAP32[$p_error>>2] = $error$0;
 return ($$1|0);
}
function _ft_mem_free($memory,$P) {
 $memory = $memory|0;
 $P = $P|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($P|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($memory)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 FUNCTION_TABLE_vii[$2 & 8191]($memory,$P);
 return;
}
function _ft_mem_strdup($memory,$str,$p_error) {
 $memory = $memory|0;
 $str = $str|0;
 $p_error = $p_error|0;
 var $$$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $block$0$i$i = 0, $error$0$i$i = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($str|0)==(0|0);
 if ($0) {
  $9 = 0;
  label = 4;
 } else {
  $1 = (_strlen($str)|0);
  $2 = (($1) + 1)|0;
  $3 = ($2|0)>(0);
  if ($3) {
   $4 = ((($memory)) + 4|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = (FUNCTION_TABLE_iii[$5 & 8191]($memory,$2)|0);
   $7 = ($6|0)==(0|0);
   $$$i$i = $7 ? 64 : 0;
   $13 = $2;$block$0$i$i = $6;$error$0$i$i = $$$i$i;
  } else {
   $9 = $2;
   label = 4;
  }
 }
 if ((label|0) == 4) {
  $8 = $9 >> 31;
  $10 = $8 & 6;
  $13 = $9;$block$0$i$i = 0;$error$0$i$i = $10;
 }
 $11 = ($error$0$i$i|0)==(0);
 $12 = ($str|0)!=(0|0);
 $or$cond$i = $12 & $11;
 if (!($or$cond$i)) {
  HEAP32[$p_error>>2] = $error$0$i$i;
  return ($block$0$i$i|0);
 }
 _memcpy(($block$0$i$i|0),($str|0),($13|0))|0;
 HEAP32[$p_error>>2] = $error$0$i$i;
 return ($block$0$i$i|0);
}
function _ft_mem_strcpyn($dst,$src,$size) {
 $dst = $dst|0;
 $src = $src|0;
 $size = $size|0;
 var $$01$lcssa = 0, $$014 = 0, $$02$lcssa = 0, $$023 = 0, $$05 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($size>>>0)>(1);
 L1: do {
  if ($0) {
   $$014 = $src;$$023 = $dst;$$05 = $size;
   while(1) {
    $1 = HEAP8[$$014>>0]|0;
    $2 = ($1<<24>>24)==(0);
    if ($2) {
     $$01$lcssa = $$014;$$02$lcssa = $$023;
     break L1;
    }
    $3 = ((($$014)) + 1|0);
    $4 = ((($$023)) + 1|0);
    HEAP8[$$023>>0] = $1;
    $5 = (($$05) + -1)|0;
    $6 = ($5>>>0)>(1);
    if ($6) {
     $$014 = $3;$$023 = $4;$$05 = $5;
    } else {
     $$01$lcssa = $3;$$02$lcssa = $4;
     break;
    }
   }
  } else {
   $$01$lcssa = $src;$$02$lcssa = $dst;
  }
 } while(0);
 HEAP8[$$02$lcssa>>0] = 0;
 $7 = HEAP8[$$01$lcssa>>0]|0;
 $8 = ($7<<24>>24)!=(0);
 $9 = $8&1;
 return ($9|0);
}
function _FT_List_Find($list,$data) {
 $list = $list|0;
 $data = $data|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $cur$0 = 0, $cur$01 = 0, $cur$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($list|0)==(0|0);
 if ($0) {
  $$0 = 0;
  return ($$0|0);
 }
 $cur$01 = HEAP32[$list>>2]|0;
 $1 = ($cur$01|0)==(0|0);
 if ($1) {
  $$0 = 0;
  return ($$0|0);
 } else {
  $cur$02 = $cur$01;
 }
 while(1) {
  $2 = ((($cur$02)) + 8|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ($3|0)==($data|0);
  if ($4) {
   $$0 = $cur$02;
   label = 5;
   break;
  }
  $5 = ((($cur$02)) + 4|0);
  $cur$0 = HEAP32[$5>>2]|0;
  $6 = ($cur$0|0)==(0|0);
  if ($6) {
   $$0 = 0;
   label = 5;
   break;
  } else {
   $cur$02 = $cur$0;
  }
 }
 if ((label|0) == 5) {
  return ($$0|0);
 }
 return (0)|0;
}
function _FT_List_Add($list,$node) {
 $list = $list|0;
 $node = $node|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($list|0)!=(0|0);
 $1 = ($node|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  return;
 }
 $2 = ((($list)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($node)) + 4|0);
 HEAP32[$4>>2] = 0;
 HEAP32[$node>>2] = $3;
 $5 = ($3|0)==(0|0);
 if ($5) {
  HEAP32[$list>>2] = $node;
 } else {
  $6 = ((($3)) + 4|0);
  HEAP32[$6>>2] = $node;
 }
 HEAP32[$2>>2] = $node;
 return;
}
function _FT_List_Remove($list,$node) {
 $list = $list|0;
 $node = $node|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($list|0)!=(0|0);
 $1 = ($node|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  return;
 }
 $2 = HEAP32[$node>>2]|0;
 $3 = ((($node)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2|0)==(0|0);
 if ($5) {
  HEAP32[$list>>2] = $4;
 } else {
  $6 = ((($2)) + 4|0);
  HEAP32[$6>>2] = $4;
 }
 $7 = ($4|0)==(0|0);
 if ($7) {
  $8 = ((($list)) + 4|0);
  HEAP32[$8>>2] = $2;
  return;
 } else {
  HEAP32[$4>>2] = $2;
  return;
 }
}
function _FT_List_Finalize($list,$destroy,$memory,$user) {
 $list = $list|0;
 $destroy = $destroy|0;
 $memory = $memory|0;
 $user = $user|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cur$02 = 0, $cur$02$us = 0, $or$cond = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($list|0)!=(0|0);
 $1 = ($memory|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  return;
 }
 $2 = HEAP32[$list>>2]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
  $4 = ($destroy|0)==(0|0);
  $5 = ((($memory)) + 8|0);
  if ($4) {
   $cur$02$us = $2;
   while(1) {
    $6 = ((($cur$02$us)) + 4|0);
    $7 = HEAP32[$6>>2]|0;
    $8 = HEAP32[$5>>2]|0;
    FUNCTION_TABLE_vii[$8 & 8191]($memory,$cur$02$us);
    $9 = ($7|0)==(0|0);
    if ($9) {
     break;
    } else {
     $cur$02$us = $7;
    }
   }
  } else {
   $cur$02 = $2;
   while(1) {
    $10 = ((($cur$02)) + 4|0);
    $11 = HEAP32[$10>>2]|0;
    $12 = ((($cur$02)) + 8|0);
    $13 = HEAP32[$12>>2]|0;
    FUNCTION_TABLE_viii[$destroy & 8191]($memory,$13,$user);
    $14 = HEAP32[$5>>2]|0;
    FUNCTION_TABLE_vii[$14 & 8191]($memory,$cur$02);
    $15 = ($11|0)==(0|0);
    if ($15) {
     break;
    } else {
     $cur$02 = $11;
    }
   }
  }
 }
 HEAP32[$list>>2] = 0;
 $16 = ((($list)) + 4|0);
 HEAP32[$16>>2] = 0;
 return;
}
function _bdf_driver_requester($module,$name) {
 $module = $module|0;
 $name = $name|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_ft_service_list_lookup(47968,$name)|0);
 return ($0|0);
}
function _bdf_get_charset_id($face,$acharset_encoding,$acharset_registry) {
 $face = $face|0;
 $acharset_encoding = $acharset_encoding|0;
 $acharset_registry = $acharset_registry|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 132|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$acharset_encoding>>2] = $1;
 $2 = ((($face)) + 136|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$acharset_registry>>2] = $3;
 return 0;
}
function _bdf_get_bdf_property($face,$prop_name,$aproperty) {
 $face = $face|0;
 $prop_name = $prop_name|0;
 $aproperty = $aproperty|0;
 var $$$i$i$i = 0, $$0 = 0, $$lcssa = 0, $$lcssa12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $kp$05$i$i$i = 0;
 var $ndp$01$i$i$i = 0, $or$cond$i = 0, $res$04$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 140|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = 6;
  return ($$0|0);
 }
 $3 = ((($1)) + 72|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0);
 $6 = ($prop_name|0)==(0|0);
 $or$cond$i = $6 | $5;
 if ($or$cond$i) {
  $$0 = 6;
  return ($$0|0);
 }
 $7 = HEAP8[$prop_name>>0]|0;
 $8 = ($7<<24>>24)==(0);
 if ($8) {
  $$0 = 6;
  return ($$0|0);
 }
 $9 = ((($1)) + 128|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + 12|0);
 $12 = HEAP32[$11>>2]|0;
 $16 = $7;$kp$05$i$i$i = $prop_name;$res$04$i$i$i = 0;
 while(1) {
  $13 = ($res$04$i$i$i*31)|0;
  $14 = ((($kp$05$i$i$i)) + 1|0);
  $15 = $16 << 24 >> 24;
  $17 = (($13) + ($15))|0;
  $18 = HEAP8[$14>>0]|0;
  $19 = ($18<<24>>24)==(0);
  if ($19) {
   $$lcssa12 = $17;
   break;
  } else {
   $16 = $18;$kp$05$i$i$i = $14;$res$04$i$i$i = $17;
  }
 }
 $20 = ((($10)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = (($$lcssa12>>>0) % ($21>>>0))&-1;
 $23 = (($12) + ($22<<2)|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ($24|0)==(0|0);
 if ($25) {
  $$0 = 6;
  return ($$0|0);
 }
 $26 = (($21) + -1)|0;
 $27 = (($12) + ($26<<2)|0);
 $29 = $24;$ndp$01$i$i$i = $23;
 while(1) {
  $28 = HEAP32[$29>>2]|0;
  $30 = HEAP8[$28>>0]|0;
  $31 = ($30<<24>>24)==($7<<24>>24);
  if ($31) {
   $32 = (_strcmp($28,$prop_name)|0);
   $33 = ($32|0)==(0);
   if ($33) {
    $$lcssa = $29;
    break;
   }
  }
  $34 = ((($ndp$01$i$i$i)) + -4|0);
  $35 = ($34>>>0)<($12>>>0);
  $$$i$i$i = $35 ? $27 : $34;
  $36 = HEAP32[$$$i$i$i>>2]|0;
  $37 = ($36|0)==(0|0);
  if ($37) {
   $$0 = 6;
   label = 16;
   break;
  } else {
   $29 = $36;$ndp$01$i$i$i = $$$i$i$i;
  }
 }
 if ((label|0) == 16) {
  return ($$0|0);
 }
 $38 = ((($1)) + 80|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ((($$lcssa)) + 4|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = (($39) + ($41<<4)|0);
 $43 = ($42|0)==(0|0);
 if ($43) {
  $$0 = 6;
  return ($$0|0);
 }
 $44 = (((($39) + ($41<<4)|0)) + 4|0);
 $45 = HEAP32[$44>>2]|0;
 switch ($45|0) {
 case 1:  {
  HEAP32[$aproperty>>2] = 1;
  $46 = (((($39) + ($41<<4)|0)) + 12|0);
  $47 = HEAP32[$46>>2]|0;
  $48 = ((($aproperty)) + 4|0);
  HEAP32[$48>>2] = $47;
  $$0 = 0;
  return ($$0|0);
  break;
 }
 case 2:  {
  $49 = (((($39) + ($41<<4)|0)) + 12|0);
  HEAP32[$aproperty>>2] = 2;
  $50 = HEAP32[$49>>2]|0;
  $51 = ((($aproperty)) + 4|0);
  HEAP32[$51>>2] = $50;
  $$0 = 0;
  return ($$0|0);
  break;
 }
 case 3:  {
  $52 = (((($39) + ($41<<4)|0)) + 12|0);
  HEAP32[$aproperty>>2] = 3;
  $53 = HEAP32[$52>>2]|0;
  $54 = ((($aproperty)) + 4|0);
  HEAP32[$54>>2] = $53;
  $$0 = 0;
  return ($$0|0);
  break;
 }
 default: {
  $$0 = 6;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function _BDF_Face_Init($stream,$bdfface,$face_index,$num_params,$params) {
 $stream = $stream|0;
 $bdfface = $bdfface|0;
 $face_index = $face_index|0;
 $num_params = $num_params|0;
 $params = $params|0;
 var $$$i$i$i = 0, $$$i$i$i$i = 0, $$$i$i$i101 = 0, $$$i$i$i12$i = 0, $$$i$i$i15 = 0, $$$i$i$i24$i = 0, $$$i$i$i29 = 0, $$$i$i$i36$i = 0, $$$i$i$i41 = 0, $$$i$i$i53 = 0, $$$i$i$i65 = 0, $$$i$i$i77 = 0, $$$i$i$i89 = 0, $$0 = 0, $$0$i92 = 0, $$lcssa = 0, $$lcssa442 = 0, $$lcssa444 = 0, $$lcssa446 = 0, $$lcssa448 = 0;
 var $$lcssa450 = 0, $$lcssa452 = 0, $$lcssa454 = 0, $$lcssa456 = 0, $$lcssa458 = 0, $$lcssa460 = 0, $$lcssa462 = 0, $$lcssa464 = 0, $$lcssa466 = 0, $$lcssa482 = 0, $$phi$trans$insert$i = 0, $$pre = 0, $$pre$i = 0, $$pre$phi65$iZ2D = 0, $$pre$phi69$iZ2D = 0, $$pre61$i = 0, $$pre62$i = 0, $$pre64$i = 0, $$pre66$i = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0;
 var $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0;
 var $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0;
 var $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0;
 var $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0;
 var $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0;
 var $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0;
 var $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0;
 var $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0;
 var $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0;
 var $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0;
 var $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0;
 var $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0;
 var $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0;
 var $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0;
 var $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0;
 var $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0;
 var $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0;
 var $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0;
 var $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0;
 var $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0;
 var $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0;
 var $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0;
 var $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0;
 var $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0;
 var $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0;
 var $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0;
 var $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0;
 var $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0;
 var $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $avail$0$i$i = 0, $avail$0$i$i$phi = 0, $avail$0$ph$i$i = 0, $avail$0$ph8$i$i = 0, $avail$0$ph8$ph$i$i = 0, $avail$1$i$i = 0, $avail$1$i$i$lcssa = 0, $avail$1$i$i$lcssa478 = 0, $avail$1$i$i$lcssa480 = 0, $brmerge = 0, $brmerge336 = 0;
 var $buf$0$ph$i$i = 0, $buf$0$ph$i$i$lcssa488 = 0, $buf$3$i$i = 0, $buf_size$0$ph$i$i = 0, $bytes$0$i$i = 0, $bytes$0$ph$i$i = 0, $bytes$0$ph6$i$i = 0, $bytes$0$ph6$ph$i$i = 0, $bytes$1$i$i = 0, $bytes$1$i$i$lcssa = 0, $bytes$1$i$i$lcssa472 = 0, $bytes$1$i$i$lcssa474 = 0, $cb$i$i = 0, $charmap = 0, $charmap1 = 0, $cursor$0$i$i = 0, $cursor$0$ph$i$i = 0, $cursor$0$ph$i$i$phi = 0, $cursor$0$ph7$i$i = 0, $cursor$0$ph7$ph$i$i = 0;
 var $cursor$1$i$i = 0, $cursor$1$i$i$lcssa = 0, $cursor$1$i$i$lcssa475 = 0, $end$034$i$i = 0, $end$034$i$i$lcssa = 0, $error = 0, $error$i = 0, $error$i$i = 0, $error$i20 = 0, $exitcond$3$i = 0, $exitcond$i = 0, $font$sroa$0$0 = 0, $font$sroa$0$1 = 0, $len$1$1$i = 0, $len$1$2$i = 0, $len$1$3$i = 0, $len$1$i = 0, $len$2$i = 0, $lengths$i$sroa$11$0 = 0, $lengths$i$sroa$11$0119 = 0;
 var $lengths$i$sroa$11$0120 = 0, $lengths$i$sroa$11$0121 = 0, $lengths$i$sroa$8$0 = 0, $lengths$i$sroa$8$0112113 = 0, $lengths$i$sroa$8$0112114 = 0, $lineno$0$ph$i$i = 0, $lineno$0$ph10$ph$i$i = 0, $lineno$0$ph10$ph$i$i$lcssa486 = 0, $mm$056$3$i = 0, $mm$056$i = 0, $n$0200 = 0, $ndp$01$i$i$i = 0, $ndp$01$i$i$i$i = 0, $ndp$01$i$i$i10$i = 0, $ndp$01$i$i$i13 = 0, $ndp$01$i$i$i22$i = 0, $ndp$01$i$i$i27 = 0, $ndp$01$i$i$i34$i = 0, $ndp$01$i$i$i39 = 0, $ndp$01$i$i$i51 = 0;
 var $ndp$01$i$i$i63 = 0, $ndp$01$i$i$i75 = 0, $ndp$01$i$i$i87 = 0, $ndp$01$i$i$i99 = 0, $options = 0, $or$cond = 0, $or$cond138 = 0, $or$cond7 = 0, $refill$0$i$i = 0, $refill$0$ph$i$i = 0, $refill$0$ph$ph$i$i = 0, $resolution_x$0133 = 0, $resolution_x$0136 = 0, $resolution_x$0136277 = 0, $resolution_y$0 = 0, $s$1$1$i = 0, $s$1$2$i = 0, $s$1$3$i = 0, $s$2$1$i = 0, $s$2$2$i = 0;
 var $s$2$3$i = 0, $s$2$i = 0, $sext = 0, $sext$i = 0, $sext272 = 0, $sext273 = 0, $sext6 = 0, $start$0$i$i = 0, $start$0$i$i$lcssa = 0, $start$0$i$i$lcssa469 = 0, $start$0$ph$i$i = 0, $start$0$ph$ph$i$i = 0, $strings$i$sroa$7$1 = 0, $strings$i$sroa$7$3107108111115 = 0, $strings$i$sroa$7$3107108111116 = 0, $strings$i$sroa$9$1 = 0, $strings$i$sroa$9$3106109110117 = 0, $strings$i$sroa$9$3106109110118 = 0, $switch$select$i$i = 0, $switch$select36$i$i = 0;
 var $switch$selectcmp$i$i = 0, $switch$selectcmp35$i$i = 0, $to_skip$0$ph$i$i = 0, $to_skip$0$ph5$i$i = 0, $to_skip$0$ph5$i$i$lcssa484 = 0, $to_skip$0$ph5$ph$i$i = 0, $unicode_charmap$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i20 = sp + 56|0;
 $cb$i$i = sp + 52|0;
 $error$i$i = sp + 48|0;
 $error$i = sp + 44|0;
 $error = sp + 40|0;
 $options = sp + 24|0;
 $charmap = sp + 12|0;
 $charmap1 = sp;
 $0 = ((($bdfface)) + 100|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_FT_Stream_Seek($stream,0)|0);
 HEAP32[$error>>2] = $2;
 $3 = ($2|0)==(0);
 if (!($3)) {
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$options>>2] = 1;
 $4 = ((($options)) + 4|0);
 HEAP32[$4>>2] = 1;
 $5 = ((($options)) + 8|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($options)) + 12|0);
 HEAP32[$6>>2] = 8;
 HEAP32[$error$i>>2] = 0;
 $7 = (_ft_mem_alloc($1,139324,$error$i)|0);
 $8 = HEAP32[$error$i>>2]|0;
 $9 = ($8|0)==(0);
 do {
  if ($9) {
   $10 = ((($7)) + 36|0);
   HEAP32[$10>>2] = $options;
   $11 = ((($7)) + 12|0);
   HEAP16[$11>>1] = 32767;
   $12 = ((($7)) + 139320|0);
   HEAP32[$12>>2] = $1;
   $13 = ((($7)) + 139304|0);
   ;HEAP32[$13>>2]=0|0;HEAP32[$13+4>>2]=0|0;HEAP32[$13+8>>2]=0|0;
   $14 = ((($7)) + 139316|0);
   HEAP32[$14>>2] = $1;
   $15 = ((($stream)) + 28|0);
   $16 = HEAP32[$15>>2]|0;
   HEAP32[$error$i$i>>2] = 0;
   $17 = (_ft_mem_realloc($16,1,0,1024,0,$error$i$i)|0);
   $18 = HEAP32[$error$i$i>>2]|0;
   $19 = ($18|0)==(0);
   L6: do {
    if ($19) {
     HEAP32[$cb$i$i>>2] = 4352;
     HEAP8[$17>>0] = 0;
     $avail$0$ph$i$i = 0;$buf$0$ph$i$i = $17;$buf_size$0$ph$i$i = 1024;$bytes$0$ph$i$i = 0;$cursor$0$ph$i$i = 0;$lineno$0$ph$i$i = 1;$to_skip$0$ph$i$i = 256;
     while(1) {
      $avail$0$ph8$ph$i$i = $avail$0$ph$i$i;$bytes$0$ph6$ph$i$i = $bytes$0$ph$i$i;$cursor$0$ph7$ph$i$i = $cursor$0$ph$i$i;$lineno$0$ph10$ph$i$i = $lineno$0$ph$i$i;$refill$0$ph$ph$i$i = 0;$start$0$ph$ph$i$i = 0;$to_skip$0$ph5$ph$i$i = $to_skip$0$ph$i$i;
      L10: while(1) {
       $avail$0$ph8$i$i = $avail$0$ph8$ph$i$i;$bytes$0$ph6$i$i = $bytes$0$ph6$ph$i$i;$cursor$0$ph7$i$i = $cursor$0$ph7$ph$i$i;$refill$0$ph$i$i = $refill$0$ph$ph$i$i;$start$0$ph$i$i = $start$0$ph$ph$i$i;$to_skip$0$ph5$i$i = $to_skip$0$ph5$ph$i$i;
       L12: while(1) {
        $avail$0$i$i = $avail$0$ph8$i$i;$bytes$0$i$i = $bytes$0$ph6$i$i;$cursor$0$i$i = $cursor$0$ph7$i$i;$refill$0$i$i = $refill$0$ph$i$i;$start$0$i$i = $start$0$ph$i$i;
        L14: while(1) {
         if ($refill$0$i$i) {
          $avail$1$i$i = $avail$0$i$i;$bytes$1$i$i = $bytes$0$i$i;$cursor$1$i$i = $cursor$0$i$i;
         } else {
          $20 = (($buf$0$ph$i$i) + ($cursor$0$i$i)|0);
          $21 = (($buf_size$0$ph$i$i) - ($cursor$0$i$i))|0;
          $22 = (_FT_Stream_TryRead($stream,$20,$21)|0);
          $23 = (($22) + ($cursor$0$i$i))|0;
          $avail$1$i$i = $23;$bytes$1$i$i = $22;$cursor$1$i$i = 0;
         }
         $24 = ($avail$1$i$i|0)>($start$0$i$i|0);
         L19: do {
          if ($24) {
           $25 = (($buf$0$ph$i$i) + ($start$0$i$i)|0);
           $26 = HEAP8[$25>>0]|0;
           $27 = $26 << 24 >> 24;
           $28 = ($27|0)==($to_skip$0$ph5$i$i|0);
           if ($28) {
            $avail$1$i$i$lcssa = $avail$1$i$i;$bytes$1$i$i$lcssa = $bytes$1$i$i;$cursor$1$i$i$lcssa = $cursor$1$i$i;$start$0$i$i$lcssa = $start$0$i$i;
            break L14;
           } else {
            $30 = $26;$end$034$i$i = $start$0$i$i;
           }
           while(1) {
            switch ($30<<24>>24) {
            case 10: case 13:  {
             $$lcssa466 = $30;$$lcssa482 = $25;$avail$1$i$i$lcssa478 = $avail$1$i$i;$bytes$1$i$i$lcssa472 = $bytes$1$i$i;$cursor$1$i$i$lcssa475 = $cursor$1$i$i;$end$034$i$i$lcssa = $end$034$i$i;$start$0$i$i$lcssa469 = $start$0$i$i;
             break L12;
             break;
            }
            default: {
            }
            }
            $31 = (($end$034$i$i) + 1)|0;
            $32 = ($31|0)<($avail$1$i$i|0);
            if (!($32)) {
             break L19;
            }
            $$phi$trans$insert$i = (($buf$0$ph$i$i) + ($31)|0);
            $$pre$i = HEAP8[$$phi$trans$insert$i>>0]|0;
            $30 = $$pre$i;$end$034$i$i = $31;
           }
          }
         } while(0);
         $33 = ($bytes$1$i$i|0)==(0);
         if ($33) {
          $buf$3$i$i = $buf$0$ph$i$i;
          break L6;
         }
         $34 = ($start$0$i$i|0)==(0);
         if ($34) {
          $avail$1$i$i$lcssa480 = $avail$1$i$i;$bytes$1$i$i$lcssa474 = $bytes$1$i$i;$lineno$0$ph10$ph$i$i$lcssa486 = $lineno$0$ph10$ph$i$i;$to_skip$0$ph5$i$i$lcssa484 = $to_skip$0$ph5$i$i;
          break L10;
         }
         $40 = (($avail$1$i$i) - ($start$0$i$i))|0;
         $41 = (($buf$0$ph$i$i) + ($start$0$i$i)|0);
         _memmove(($buf$0$ph$i$i|0),($41|0),($40|0))|0;
         $avail$0$i$i$phi = $start$0$i$i;$bytes$0$i$i = $40;$cursor$0$i$i = $40;$refill$0$i$i = 0;$start$0$i$i = 0;$avail$0$i$i = $avail$0$i$i$phi;
        }
        $29 = (($start$0$i$i$lcssa) + 1)|0;
        $avail$0$ph8$i$i = $avail$1$i$i$lcssa;$bytes$0$ph6$i$i = $bytes$1$i$i$lcssa;$cursor$0$ph7$i$i = $cursor$1$i$i$lcssa;$refill$0$ph$i$i = 1;$start$0$ph$i$i = $29;$to_skip$0$ph5$i$i = 256;
       }
       $42 = (($buf$0$ph$i$i) + ($end$034$i$i$lcssa)|0);
       HEAP8[$42>>0] = 0;
       $43 = HEAP8[$$lcssa482>>0]|0;
       switch ($43<<24>>24) {
       case 26: case 35:  {
        break;
       }
       default: {
        $44 = ($end$034$i$i$lcssa|0)>($start$0$i$i$lcssa469|0);
        if ($44) {
         $45 = HEAP32[$cb$i$i>>2]|0;
         $46 = (($end$034$i$i$lcssa) - ($start$0$i$i$lcssa469))|0;
         $47 = (FUNCTION_TABLE_iiiiii[$45 & 8191]($$lcssa482,$46,$lineno$0$ph10$ph$i$i,$cb$i$i,$7)|0);
         HEAP32[$error$i$i>>2] = $47;
         $48 = ($47|0)==(-1);
         if ($48) {
          $49 = HEAP32[$cb$i$i>>2]|0;
          $50 = (FUNCTION_TABLE_iiiiii[$49 & 8191]($$lcssa482,$46,$lineno$0$ph10$ph$i$i,$cb$i$i,$7)|0);
          HEAP32[$error$i$i>>2] = $50;
          $52 = $50;
         } else {
          $52 = $47;
         }
         $51 = ($52|0)==(0);
         if (!($51)) {
          $buf$3$i$i = $buf$0$ph$i$i;
          break L6;
         }
        }
       }
       }
       $53 = (($lineno$0$ph10$ph$i$i) + 1)|0;
       HEAP8[$42>>0] = $$lcssa466;
       $54 = (($end$034$i$i$lcssa) + 1)|0;
       $switch$selectcmp$i$i = ($$lcssa466<<24>>24)==(13);
       $switch$select$i$i = $switch$selectcmp$i$i ? 10 : 256;
       $switch$selectcmp35$i$i = ($$lcssa466<<24>>24)==(10);
       $switch$select36$i$i = $switch$selectcmp35$i$i ? 13 : $switch$select$i$i;
       $avail$0$ph8$ph$i$i = $avail$1$i$i$lcssa478;$bytes$0$ph6$ph$i$i = $bytes$1$i$i$lcssa472;$cursor$0$ph7$ph$i$i = $cursor$1$i$i$lcssa475;$lineno$0$ph10$ph$i$i = $53;$refill$0$ph$ph$i$i = 1;$start$0$ph$ph$i$i = $54;$to_skip$0$ph5$ph$i$i = $switch$select36$i$i;
      }
      $35 = ($buf_size$0$ph$i$i>>>0)>(65535);
      if ($35) {
       $buf$0$ph$i$i$lcssa488 = $buf$0$ph$i$i;
       break;
      }
      $36 = $buf_size$0$ph$i$i << 1;
      $37 = (_ft_mem_realloc($16,1,$buf_size$0$ph$i$i,$36,$buf$0$ph$i$i,$error$i$i)|0);
      $38 = HEAP32[$error$i$i>>2]|0;
      $39 = ($38|0)==(0);
      if ($39) {
       $cursor$0$ph$i$i$phi = $buf_size$0$ph$i$i;$avail$0$ph$i$i = $avail$1$i$i$lcssa480;$buf$0$ph$i$i = $37;$buf_size$0$ph$i$i = $36;$bytes$0$ph$i$i = $bytes$1$i$i$lcssa474;$lineno$0$ph$i$i = $lineno$0$ph10$ph$i$i$lcssa486;$to_skip$0$ph$i$i = $to_skip$0$ph5$i$i$lcssa484;$cursor$0$ph$i$i = $cursor$0$ph$i$i$phi;
      } else {
       $buf$3$i$i = $37;
       break L6;
      }
     }
     HEAP32[$error$i$i>>2] = 6;
     $buf$3$i$i = $buf$0$ph$i$i$lcssa488;
    } else {
     $buf$3$i$i = $17;
    }
   } while(0);
   _ft_mem_free($16,$buf$3$i$i);
   $55 = HEAP32[$error$i$i>>2]|0;
   HEAP32[$error$i>>2] = $55;
   $56 = ($55|0)==(0);
   $57 = ((($7)) + 32|0);
   $58 = HEAP32[$57>>2]|0;
   if ($56) {
    $59 = ($58|0)==(0|0);
    $60 = $58;
    if (!($59)) {
     $61 = ((($58)) + 28|0);
     $62 = HEAP32[$61>>2]|0;
     $63 = ($62|0)==(8);
     if (!($63)) {
      $64 = ((($58)) + 4|0);
      $65 = HEAP16[$64>>1]|0;
      $66 = ((($58)) + 32|0);
      HEAP16[$66>>1] = $65;
     }
     $67 = ((($7)) + 4|0);
     $68 = HEAP32[$67>>2]|0;
     $69 = ((($58)) + 52|0);
     $70 = HEAP32[$69>>2]|0;
     $71 = ((($58)) + 64|0);
     $72 = HEAP32[$71>>2]|0;
     $73 = (($72) + ($70))|0;
     $74 = ($68|0)==($73|0);
     if (!($74)) {
      $75 = ((($58)) + 278660|0);
      HEAP16[$75>>1] = 1;
     }
     $76 = HEAP32[$10>>2]|0;
     $77 = HEAP32[$76>>2]|0;
     $78 = ($77|0)==(0);
     $79 = $72 | $70;
     $80 = ($79|0)==(0);
     $or$cond138 = $80 | $78;
     if (!($or$cond138)) {
      $81 = ((($7)) + 16|0);
      $82 = HEAP32[$81>>2]|0;
      $sext$i = $82 << 16;
      $83 = $sext$i >> 16;
      $84 = HEAP16[$11>>1]|0;
      $85 = $84 << 16 >> 16;
      $86 = (($83) - ($85))|0;
      $87 = ((($58)) + 4|0);
      $88 = HEAP32[$87>>2]|0;
      $89 = $88 & 65535;
      $90 = ($86|0)==($89|0);
      $91 = $82 >>> 16;
      $92 = $91&65535;
      $93 = $88 >>> 16;
      if (!($90)) {
       $94 = $86&65535;
       HEAP16[$87>>1] = $94;
       $95 = ((($58)) + 278660|0);
       HEAP16[$95>>1] = 1;
      }
      $96 = ((($58)) + 8|0);
      $97 = HEAP16[$96>>1]|0;
      $98 = ($97<<16>>16)==($84<<16>>16);
      if (!($98)) {
       HEAP16[$96>>1] = $84;
       $99 = ((($58)) + 278660|0);
       HEAP16[$99>>1] = 1;
      }
      $100 = ((($58)) + 12|0);
      $101 = HEAP32[$100>>2]|0;
      $102 = $101&65535;
      $103 = ($102<<16>>16)==($92<<16>>16);
      $104 = $101 >>> 16;
      $105 = $104&65535;
      if (!($103)) {
       HEAP16[$100>>1] = $92;
       $106 = ((($58)) + 278660|0);
       HEAP16[$106>>1] = 1;
      }
      $107 = ((($7)) + 20|0);
      $108 = HEAP16[$107>>1]|0;
      $109 = ($105<<16>>16)==($108<<16>>16);
      if (!($109)) {
       $110 = ((($58)) + 14|0);
       HEAP16[$110>>1] = $108;
       $111 = $108&65535;
       $112 = (0 - ($111))|0;
       $113 = $112&65535;
       $114 = ((($58)) + 10|0);
       HEAP16[$114>>1] = $113;
       $115 = ((($58)) + 278660|0);
       HEAP16[$115>>1] = 1;
      }
      $116 = $82 >> 16;
      $117 = $108 << 16 >> 16;
      $118 = (($117) + ($116))|0;
      $119 = ($118|0)==($93|0);
      if (!($119)) {
       $120 = ((($58)) + 6|0);
       $121 = $118&65535;
       HEAP16[$120>>1] = $121;
      }
     }
    }
    $122 = HEAP32[$7>>2]|0;
    $123 = $122 & 1;
    $124 = ($123|0)==(0);
    if (!($124)) {
     $125 = $122 & 32;
     $126 = ($125|0)==(0);
     if ($126) {
      HEAP32[$error$i>>2] = 185;
      $$pre$phi65$iZ2D = $13;$$pre$phi69$iZ2D = $14;$font$sroa$0$0 = 0;
      label = 56;
      break;
     } else {
      HEAP32[$error$i>>2] = 186;
      $$pre$phi65$iZ2D = $13;$$pre$phi69$iZ2D = $14;$font$sroa$0$0 = 0;
      label = 56;
      break;
     }
    }
    if ($59) {
     HEAP32[$error$i>>2] = 3;
     $$pre$phi65$iZ2D = $13;$$pre$phi69$iZ2D = $14;$font$sroa$0$0 = $60;
     label = 56;
     break;
    }
    $127 = ((($58)) + 88|0);
    $128 = HEAP32[$127>>2]|0;
    $129 = ($128|0)==(0);
    if ($129) {
     $$pre$phi65$iZ2D = $13;$$pre$phi69$iZ2D = $14;$font$sroa$0$0 = $60;
     label = 56;
     break;
    }
    $130 = ((($58)) + 278664|0);
    $131 = HEAP32[$130>>2]|0;
    $132 = (($128) + 1)|0;
    $133 = ((($58)) + 84|0);
    $134 = HEAP32[$133>>2]|0;
    $135 = (_ft_mem_realloc($131,1,$128,$132,$134,$error$i)|0);
    $136 = HEAP32[$57>>2]|0;
    $137 = ((($136)) + 84|0);
    HEAP32[$137>>2] = $135;
    $138 = HEAP32[$error$i>>2]|0;
    $139 = ($138|0)==(0);
    if ($139) {
     $140 = ((($136)) + 88|0);
     $141 = HEAP32[$140>>2]|0;
     $142 = (($135) + ($141)|0);
     HEAP8[$142>>0] = 0;
     $$pre61$i = HEAP32[$57>>2]|0;
     $$pre$phi65$iZ2D = $13;$$pre$phi69$iZ2D = $14;$font$sroa$0$0 = $$pre61$i;
     label = 56;
     break;
    } else {
     $147 = $136;
    }
   } else {
    $147 = $58;
   }
   _bdf_free_font($147);
   $148 = HEAP32[$57>>2]|0;
   _ft_mem_free($1,$148);
   HEAP32[$57>>2] = 0;
   $$pre$phi65$iZ2D = $13;$$pre$phi69$iZ2D = $14;$font$sroa$0$0 = 0;
   label = 56;
  } else {
   $143 = ($7|0)==(0|0);
   if ($143) {
    $149 = $8;$font$sroa$0$1 = 0;
   } else {
    $$pre64$i = ((($7)) + 139304|0);
    $$pre66$i = ((($7)) + 139316|0);
    $$pre$phi65$iZ2D = $$pre64$i;$$pre$phi69$iZ2D = $$pre66$i;$font$sroa$0$0 = 0;
    label = 56;
   }
  }
 } while(0);
 if ((label|0) == 56) {
  $144 = HEAP32[$$pre$phi69$iZ2D>>2]|0;
  $145 = ($144|0)==(0|0);
  if (!($145)) {
   $146 = HEAP32[$$pre$phi65$iZ2D>>2]|0;
   _ft_mem_free($144,$146);
   ;HEAP32[$$pre$phi65$iZ2D>>2]=0|0;HEAP32[$$pre$phi65$iZ2D+4>>2]=0|0;HEAP32[$$pre$phi65$iZ2D+8>>2]=0|0;HEAP32[$$pre$phi65$iZ2D+12>>2]=0|0;
  }
  _ft_mem_free($1,$7);
  $$pre62$i = HEAP32[$error$i>>2]|0;
  $149 = $$pre62$i;$font$sroa$0$1 = $font$sroa$0$0;
 }
 HEAP32[$error>>2] = $149;
 $150 = $149 & 255;
 $151 = ($150|0)==(176);
 if ($151) {
  _BDF_Face_Done($bdfface);
  $$0 = 2;
  STACKTOP = sp;return ($$0|0);
 }
 $152 = ($149|0)==(0);
 if (!($152)) {
  $$0 = $149;
  STACKTOP = sp;return ($$0|0);
 }
 $153 = ((($bdfface)) + 140|0);
 HEAP32[$153>>2] = $font$sroa$0$1;
 $154 = ($face_index|0)>(0);
 $155 = $font$sroa$0$1;
 if ($154) {
  _BDF_Face_Done($bdfface);
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$bdfface>>2] = 1;
 $156 = ((($bdfface)) + 4|0);
 HEAP32[$156>>2] = 0;
 $157 = ((($bdfface)) + 8|0);
 $158 = HEAP32[$157>>2]|0;
 $159 = $158 | 146;
 HEAP32[$157>>2] = $159;
 $160 = ($font$sroa$0$1|0)==(0);
 L97: do {
  if ($160) {
   label = 86;
  } else {
   $161 = ((($155)) + 72|0);
   $162 = HEAP32[$161>>2]|0;
   $163 = ($162|0)==(0);
   if ($163) {
    label = 86;
   } else {
    $164 = ((($155)) + 128|0);
    $165 = HEAP32[$164>>2]|0;
    $166 = ((($165)) + 12|0);
    $167 = HEAP32[$166>>2]|0;
    $168 = ((($165)) + 4|0);
    $169 = HEAP32[$168>>2]|0;
    $170 = (3000791075 % ($169>>>0))&-1;
    $171 = (($167) + ($170<<2)|0);
    $172 = HEAP32[$171>>2]|0;
    $173 = ($172|0)==(0|0);
    if (!($173)) {
     $174 = (($169) + -1)|0;
     $175 = (($167) + ($174<<2)|0);
     $177 = $172;$ndp$01$i$i$i = $171;
     while(1) {
      $176 = HEAP32[$177>>2]|0;
      $178 = HEAP8[$176>>0]|0;
      $179 = ($178<<24>>24)==(83);
      if ($179) {
       $180 = (_strcmp($176,531437)|0);
       $181 = ($180|0)==(0);
       if ($181) {
        $$lcssa464 = $177;
        label = 71;
        break;
       }
      }
      $182 = ((($ndp$01$i$i$i)) + -4|0);
      $183 = ($182>>>0)<($167>>>0);
      $$$i$i$i = $183 ? $175 : $182;
      $184 = HEAP32[$$$i$i$i>>2]|0;
      $185 = ($184|0)==(0|0);
      if ($185) {
       break;
      } else {
       $177 = $184;$ndp$01$i$i$i = $$$i$i$i;
      }
     }
     L107: do {
      if ((label|0) == 71) {
       $186 = ((($155)) + 80|0);
       $187 = HEAP32[$186>>2]|0;
       $188 = ((($$lcssa464)) + 4|0);
       $189 = HEAP32[$188>>2]|0;
       $190 = (($187) + ($189<<4)|0);
       $191 = ($190|0)==(0|0);
       if (!($191)) {
        $192 = (((($187) + ($189<<4)|0)) + 4|0);
        $193 = HEAP32[$192>>2]|0;
        $194 = ($193|0)==(1);
        if ($194) {
         $195 = (((($187) + ($189<<4)|0)) + 12|0);
         $196 = HEAP32[$195>>2]|0;
         $197 = ($196|0)==(0|0);
         if (!($197)) {
          $198 = HEAP8[$196>>0]|0;
          switch ($198<<24>>24) {
          case 99: case 67: case 109: case 77:  {
           break;
          }
          default: {
           break L107;
          }
          }
          $199 = $158 | 150;
          HEAP32[$157>>2] = $199;
         }
        }
       }
      }
     } while(0);
     if ($163) {
      label = 86;
      break;
     }
    }
    $200 = ((($155)) + 128|0);
    $201 = HEAP32[$200>>2]|0;
    $202 = ((($201)) + 12|0);
    $203 = HEAP32[$202>>2]|0;
    $204 = ((($201)) + 4|0);
    $205 = HEAP32[$204>>2]|0;
    $206 = (1183963782 % ($205>>>0))&-1;
    $207 = (($203) + ($206<<2)|0);
    $208 = HEAP32[$207>>2]|0;
    $209 = ($208|0)==(0|0);
    if ($209) {
     label = 86;
    } else {
     $210 = (($205) + -1)|0;
     $211 = (($203) + ($210<<2)|0);
     $213 = $208;$ndp$01$i$i$i13 = $207;
     while(1) {
      $212 = HEAP32[$213>>2]|0;
      $214 = HEAP8[$212>>0]|0;
      $215 = ($214<<24>>24)==(70);
      if ($215) {
       $216 = (_strcmp($212,532470)|0);
       $217 = ($216|0)==(0);
       if ($217) {
        $$lcssa462 = $213;
        break;
       }
      }
      $218 = ((($ndp$01$i$i$i13)) + -4|0);
      $219 = ($218>>>0)<($203>>>0);
      $$$i$i$i15 = $219 ? $211 : $218;
      $220 = HEAP32[$$$i$i$i15>>2]|0;
      $221 = ($220|0)==(0|0);
      if ($221) {
       label = 86;
       break L97;
      } else {
       $213 = $220;$ndp$01$i$i$i13 = $$$i$i$i15;
      }
     }
     $222 = ((($155)) + 80|0);
     $223 = HEAP32[$222>>2]|0;
     $224 = ((($$lcssa462)) + 4|0);
     $225 = HEAP32[$224>>2]|0;
     $226 = (($223) + ($225<<4)|0);
     $227 = ($226|0)==(0|0);
     if ($227) {
      label = 86;
     } else {
      $228 = (((($223) + ($225<<4)|0)) + 12|0);
      $229 = HEAP32[$228>>2]|0;
      $230 = ($229|0)==(0|0);
      if ($230) {
       label = 86;
      } else {
       $231 = (_ft_mem_strdup($1,$229,$error)|0);
       $232 = ((($bdfface)) + 20|0);
       HEAP32[$232>>2] = $231;
       $233 = HEAP32[$error>>2]|0;
       $234 = ($233|0)==(0);
       if ($234) {
        $$pre = HEAP32[$153>>2]|0;
        $239 = $$pre;
        break;
       } else {
        $$0 = $233;
        STACKTOP = sp;return ($$0|0);
       }
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 86) {
  $235 = ((($bdfface)) + 20|0);
  HEAP32[$235>>2] = 0;
  $239 = $155;
 }
 HEAP32[$error$i20>>2] = 0;
 $236 = HEAP32[$0>>2]|0;
 $237 = ((($bdfface)) + 12|0);
 HEAP32[$237>>2] = 0;
 $238 = ($239|0)==(0|0);
 if ($238) {
  $756 = 0;$757 = 0;$758 = 0;$759 = 1;$lengths$i$sroa$11$0120 = 0;$lengths$i$sroa$8$0112114 = 0;$strings$i$sroa$7$3107108111116 = 0;$strings$i$sroa$9$3106109110118 = 0;
  label = 130;
 } else {
  $240 = ((($239)) + 72|0);
  $241 = HEAP32[$240>>2]|0;
  $242 = ($241|0)==(0);
  if ($242) {
   $756 = 0;$757 = 0;$758 = 0;$759 = 1;$lengths$i$sroa$11$0120 = 0;$lengths$i$sroa$8$0112114 = 0;$strings$i$sroa$7$3107108111116 = 0;$strings$i$sroa$9$3106109110118 = 0;
   label = 130;
  } else {
   $243 = ((($239)) + 128|0);
   $244 = HEAP32[$243>>2]|0;
   $245 = ((($244)) + 12|0);
   $246 = HEAP32[$245>>2]|0;
   $247 = ((($244)) + 4|0);
   $248 = HEAP32[$247>>2]|0;
   $249 = (78981326 % ($248>>>0))&-1;
   $250 = (($246) + ($249<<2)|0);
   $251 = HEAP32[$250>>2]|0;
   $252 = ($251|0)==(0|0);
   L133: do {
    if ($252) {
     $311 = 2;$strings$i$sroa$7$1 = 0;
    } else {
     $253 = (($248) + -1)|0;
     $254 = (($246) + ($253<<2)|0);
     $256 = $251;$ndp$01$i$i$i$i = $250;
     while(1) {
      $255 = HEAP32[$256>>2]|0;
      $257 = HEAP8[$255>>0]|0;
      $258 = ($257<<24>>24)==(83);
      if ($258) {
       $259 = (_strcmp($255,532415)|0);
       $260 = ($259|0)==(0);
       if ($260) {
        $$lcssa460 = $256;
        break;
       }
      }
      $261 = ((($ndp$01$i$i$i$i)) + -4|0);
      $262 = ($261>>>0)<($246>>>0);
      $$$i$i$i$i = $262 ? $254 : $261;
      $263 = HEAP32[$$$i$i$i$i>>2]|0;
      $264 = ($263|0)==(0|0);
      if ($264) {
       $311 = 2;$strings$i$sroa$7$1 = 0;
       break L133;
      } else {
       $256 = $263;$ndp$01$i$i$i$i = $$$i$i$i$i;
      }
     }
     $265 = ((($239)) + 80|0);
     $266 = HEAP32[$265>>2]|0;
     $267 = ((($$lcssa460)) + 4|0);
     $268 = HEAP32[$267>>2]|0;
     $269 = (($266) + ($268<<4)|0);
     $270 = ($269|0)==(0|0);
     if ($270) {
      $311 = 2;$strings$i$sroa$7$1 = 0;
     } else {
      $271 = (((($266) + ($268<<4)|0)) + 4|0);
      $272 = HEAP32[$271>>2]|0;
      $273 = ($272|0)==(1);
      if ($273) {
       $274 = (((($266) + ($268<<4)|0)) + 12|0);
       $275 = HEAP32[$274>>2]|0;
       $276 = ($275|0)==(0|0);
       if ($276) {
        $311 = 2;$strings$i$sroa$7$1 = 0;
       } else {
        $277 = HEAP8[$275>>0]|0;
        switch ($277<<24>>24) {
        case 105: case 73: case 111: case 79:  {
         break;
        }
        default: {
         $311 = 2;$strings$i$sroa$7$1 = 0;
         break L133;
        }
        }
        HEAP32[$237>>2] = 1;
        $278 = HEAP8[$275>>0]|0;
        $279 = $278 | 32;
        $280 = ($279<<24>>24)==(111);
        $281 = $280 ? 532421 : 594736;
        $311 = 3;$strings$i$sroa$7$1 = $281;
       }
      } else {
       $311 = 2;$strings$i$sroa$7$1 = 0;
      }
     }
    }
   } while(0);
   $282 = (4260511730 % ($248>>>0))&-1;
   $283 = (($246) + ($282<<2)|0);
   $284 = HEAP32[$283>>2]|0;
   $285 = ($284|0)==(0|0);
   L146: do {
    if ($285) {
     $373 = 0;
    } else {
     $286 = (($248) + -1)|0;
     $287 = (($246) + ($286<<2)|0);
     $289 = $284;$ndp$01$i$i$i10$i = $283;
     while(1) {
      $288 = HEAP32[$289>>2]|0;
      $290 = HEAP8[$288>>0]|0;
      $291 = ($290<<24>>24)==(87);
      if ($291) {
       $292 = (_strcmp($288,532429)|0);
       $293 = ($292|0)==(0);
       if ($293) {
        $$lcssa458 = $289;
        break;
       }
      }
      $294 = ((($ndp$01$i$i$i10$i)) + -4|0);
      $295 = ($294>>>0)<($246>>>0);
      $$$i$i$i12$i = $295 ? $287 : $294;
      $296 = HEAP32[$$$i$i$i12$i>>2]|0;
      $297 = ($296|0)==(0|0);
      if ($297) {
       $373 = 0;
       break L146;
      } else {
       $289 = $296;$ndp$01$i$i$i10$i = $$$i$i$i12$i;
      }
     }
     $298 = ((($239)) + 80|0);
     $299 = HEAP32[$298>>2]|0;
     $300 = ((($$lcssa458)) + 4|0);
     $301 = HEAP32[$300>>2]|0;
     $302 = (($299) + ($301<<4)|0);
     $303 = ($302|0)==(0|0);
     if ($303) {
      $373 = 0;
     } else {
      $304 = (((($299) + ($301<<4)|0)) + 4|0);
      $305 = HEAP32[$304>>2]|0;
      $306 = ($305|0)==(1);
      if ($306) {
       $307 = (((($299) + ($301<<4)|0)) + 12|0);
       $308 = HEAP32[$307>>2]|0;
       $309 = ($308|0)==(0|0);
       if ($309) {
        $373 = 0;
       } else {
        $310 = HEAP8[$308>>0]|0;
        switch ($310<<24>>24) {
        case 98: case 66:  {
         break;
        }
        default: {
         $373 = 0;
         break L146;
        }
        }
        HEAP32[$237>>2] = $311;
        $373 = 594731;
       }
      } else {
       $373 = 0;
      }
     }
    }
   } while(0);
   $312 = (41411398 % ($248>>>0))&-1;
   $313 = (($246) + ($312<<2)|0);
   $314 = HEAP32[$313>>2]|0;
   $315 = ($314|0)==(0|0);
   L159: do {
    if ($315) {
     $strings$i$sroa$9$1 = 0;
    } else {
     $316 = (($248) + -1)|0;
     $317 = (($246) + ($316<<2)|0);
     $319 = $314;$ndp$01$i$i$i22$i = $313;
     while(1) {
      $318 = HEAP32[$319>>2]|0;
      $320 = HEAP8[$318>>0]|0;
      $321 = ($320<<24>>24)==(83);
      if ($321) {
       $322 = (_strcmp($318,532441)|0);
       $323 = ($322|0)==(0);
       if ($323) {
        $$lcssa456 = $319;
        break;
       }
      }
      $324 = ((($ndp$01$i$i$i22$i)) + -4|0);
      $325 = ($324>>>0)<($246>>>0);
      $$$i$i$i24$i = $325 ? $317 : $324;
      $326 = HEAP32[$$$i$i$i24$i>>2]|0;
      $327 = ($326|0)==(0|0);
      if ($327) {
       $strings$i$sroa$9$1 = 0;
       break L159;
      } else {
       $319 = $326;$ndp$01$i$i$i22$i = $$$i$i$i24$i;
      }
     }
     $328 = ((($239)) + 80|0);
     $329 = HEAP32[$328>>2]|0;
     $330 = ((($$lcssa456)) + 4|0);
     $331 = HEAP32[$330>>2]|0;
     $332 = (($329) + ($331<<4)|0);
     $333 = ($332|0)==(0|0);
     if ($333) {
      $strings$i$sroa$9$1 = 0;
     } else {
      $334 = (((($329) + ($331<<4)|0)) + 4|0);
      $335 = HEAP32[$334>>2]|0;
      $336 = ($335|0)==(1);
      if ($336) {
       $337 = (((($329) + ($331<<4)|0)) + 12|0);
       $338 = HEAP32[$337>>2]|0;
       $339 = ($338|0)==(0|0);
       if ($339) {
        $strings$i$sroa$9$1 = 0;
       } else {
        $340 = HEAP8[$338>>0]|0;
        switch ($340<<24>>24) {
        case 110: case 78: case 0:  {
         $strings$i$sroa$9$1 = 0;
         break L159;
         break;
        }
        default: {
        }
        }
        $strings$i$sroa$9$1 = $338;
       }
      } else {
       $strings$i$sroa$9$1 = 0;
      }
     }
    }
   } while(0);
   $341 = (657458935 % ($248>>>0))&-1;
   $342 = (($246) + ($341<<2)|0);
   $343 = HEAP32[$342>>2]|0;
   $344 = ($343|0)==(0|0);
   L172: do {
    if ($344) {
     $760 = 0;$761 = 0;$len$1$i = 0;
    } else {
     $345 = (($248) + -1)|0;
     $346 = (($246) + ($345<<2)|0);
     $348 = $343;$ndp$01$i$i$i34$i = $342;
     while(1) {
      $347 = HEAP32[$348>>2]|0;
      $349 = HEAP8[$347>>0]|0;
      $350 = ($349<<24>>24)==(65);
      if ($350) {
       $351 = (_strcmp($347,532455)|0);
       $352 = ($351|0)==(0);
       if ($352) {
        $$lcssa454 = $348;
        break;
       }
      }
      $353 = ((($ndp$01$i$i$i34$i)) + -4|0);
      $354 = ($353>>>0)<($246>>>0);
      $$$i$i$i36$i = $354 ? $346 : $353;
      $355 = HEAP32[$$$i$i$i36$i>>2]|0;
      $356 = ($355|0)==(0|0);
      if ($356) {
       $760 = 0;$761 = 0;$len$1$i = 0;
       break L172;
      } else {
       $348 = $355;$ndp$01$i$i$i34$i = $$$i$i$i36$i;
      }
     }
     $357 = ((($239)) + 80|0);
     $358 = HEAP32[$357>>2]|0;
     $359 = ((($$lcssa454)) + 4|0);
     $360 = HEAP32[$359>>2]|0;
     $361 = (($358) + ($360<<4)|0);
     $362 = ($361|0)==(0|0);
     if ($362) {
      $760 = 0;$761 = 0;$len$1$i = 0;
     } else {
      $363 = (((($358) + ($360<<4)|0)) + 4|0);
      $364 = HEAP32[$363>>2]|0;
      $365 = ($364|0)==(1);
      if ($365) {
       $366 = (((($358) + ($360<<4)|0)) + 12|0);
       $367 = HEAP32[$366>>2]|0;
       $368 = ($367|0)==(0|0);
       if ($368) {
        $760 = 0;$761 = 0;$len$1$i = 0;
       } else {
        $369 = HEAP8[$367>>0]|0;
        switch ($369<<24>>24) {
        case 110: case 78: case 0:  {
         $760 = 0;$761 = 0;$len$1$i = 0;
         break L172;
         break;
        }
        default: {
        }
        }
        $370 = (_strlen($367)|0);
        $371 = (($370) + 1)|0;
        $760 = $367;$761 = $370;$len$1$i = $371;
       }
      } else {
       $760 = 0;$761 = 0;$len$1$i = 0;
      }
     }
    }
   } while(0);
   $372 = ($373|0)==(0|0);
   if ($372) {
    $762 = 1;$763 = 0;$764 = 0;$len$1$1$i = $len$1$i;
   } else {
    $409 = (_strlen($373)|0);
    $410 = (($len$1$i) + 1)|0;
    $411 = (($410) + ($409))|0;
    $762 = 0;$763 = $373;$764 = $409;$len$1$1$i = $411;
   }
   $412 = ($strings$i$sroa$7$1|0)==(0|0);
   if ($412) {
    $767 = 0;$len$1$2$i = $len$1$1$i;$lengths$i$sroa$8$0 = 0;
   } else {
    $413 = (_strlen($strings$i$sroa$7$1)|0);
    $414 = (($len$1$1$i) + 1)|0;
    $415 = (($414) + ($413))|0;
    $767 = $strings$i$sroa$7$1;$len$1$2$i = $415;$lengths$i$sroa$8$0 = $413;
   }
   $416 = ($strings$i$sroa$9$1|0)==(0|0);
   if ($416) {
    $len$1$3$i = $len$1$2$i;$lengths$i$sroa$11$0 = 0;
   } else {
    $417 = (_strlen($strings$i$sroa$9$1)|0);
    $418 = (($len$1$2$i) + 1)|0;
    $419 = (($418) + ($417))|0;
    $len$1$3$i = $419;$lengths$i$sroa$11$0 = $417;
   }
   $420 = ($len$1$3$i|0)==(0);
   if ($420) {
    $756 = $764;$757 = $763;$758 = $767;$759 = $762;$lengths$i$sroa$11$0120 = $lengths$i$sroa$11$0;$lengths$i$sroa$8$0112114 = $lengths$i$sroa$8$0;$strings$i$sroa$7$3107108111116 = $strings$i$sroa$7$1;$strings$i$sroa$9$3106109110118 = $strings$i$sroa$9$1;
    label = 130;
   } else {
    $379 = $760;$380 = $761;$390 = $763;$391 = $764;$765 = $767;$766 = $762;$len$2$i = $len$1$3$i;$lengths$i$sroa$11$0119 = $lengths$i$sroa$11$0;$lengths$i$sroa$8$0112113 = $lengths$i$sroa$8$0;$strings$i$sroa$7$3107108111115 = $strings$i$sroa$7$1;$strings$i$sroa$9$3106109110117 = $strings$i$sroa$9$1;
   }
  }
 }
 if ((label|0) == 130) {
  $379 = 594711;$380 = 7;$390 = $757;$391 = $756;$765 = $758;$766 = $759;$len$2$i = 8;$lengths$i$sroa$11$0119 = $lengths$i$sroa$11$0120;$lengths$i$sroa$8$0112113 = $lengths$i$sroa$8$0112114;$strings$i$sroa$7$3107108111115 = $strings$i$sroa$7$3107108111116;$strings$i$sroa$9$3106109110117 = $strings$i$sroa$9$3106109110118;
 }
 $374 = (_ft_mem_alloc($236,$len$2$i,$error$i20)|0);
 $375 = ((($bdfface)) + 24|0);
 HEAP32[$375>>2] = $374;
 $376 = HEAP32[$error$i20>>2]|0;
 $377 = ($376|0)==(0);
 if (!($377)) {
  HEAP32[$error>>2] = $376;
  $$0 = $376;
  STACKTOP = sp;return ($$0|0);
 }
 $378 = ($379|0)==(0|0);
 if ($378) {
  $s$2$i = $374;
 } else {
  _memcpy(($374|0),($379|0),($380|0))|0;
  $381 = ($380|0)==(0);
  if (!($381)) {
   $mm$056$i = 0;
   while(1) {
    $382 = (($374) + ($mm$056$i)|0);
    $383 = HEAP8[$382>>0]|0;
    $384 = ($383<<24>>24)==(32);
    if ($384) {
     HEAP8[$382>>0] = 45;
    }
    $385 = (($mm$056$i) + 1)|0;
    $exitcond$i = ($385|0)==($380|0);
    if ($exitcond$i) {
     break;
    } else {
     $mm$056$i = $385;
    }
   }
  }
  $386 = (($374) + ($380)|0);
  $s$2$i = $386;
 }
 if ($766) {
  $394 = $765;$s$2$1$i = $s$2$i;
 } else {
  $387 = HEAP32[$375>>2]|0;
  $388 = ($s$2$i|0)==($387|0);
  if ($388) {
   $s$1$1$i = $s$2$i;
  } else {
   $389 = ((($s$2$i)) + 1|0);
   HEAP8[$s$2$i>>0] = 32;
   $s$1$1$i = $389;
  }
  _memcpy(($s$1$1$i|0),($390|0),($391|0))|0;
  $392 = (($s$1$1$i) + ($391)|0);
  $394 = $strings$i$sroa$7$3107108111115;$s$2$1$i = $392;
 }
 $393 = ($394|0)==(0|0);
 if ($393) {
  $s$2$2$i = $s$2$1$i;
 } else {
  $395 = HEAP32[$375>>2]|0;
  $396 = ($s$2$1$i|0)==($395|0);
  if ($396) {
   $s$1$2$i = $s$2$1$i;
  } else {
   $397 = ((($s$2$1$i)) + 1|0);
   HEAP8[$s$2$1$i>>0] = 32;
   $s$1$2$i = $397;
  }
  _memcpy(($s$1$2$i|0),($394|0),($lengths$i$sroa$8$0112113|0))|0;
  $398 = (($s$1$2$i) + ($lengths$i$sroa$8$0112113)|0);
  $s$2$2$i = $398;
 }
 $399 = ($strings$i$sroa$9$3106109110117|0)==(0|0);
 if ($399) {
  $s$2$3$i = $s$2$2$i;
 } else {
  $400 = HEAP32[$375>>2]|0;
  $401 = ($s$2$2$i|0)==($400|0);
  if ($401) {
   $s$1$3$i = $s$2$2$i;
  } else {
   $402 = ((($s$2$2$i)) + 1|0);
   HEAP8[$s$2$2$i>>0] = 32;
   $s$1$3$i = $402;
  }
  _memcpy(($s$1$3$i|0),($strings$i$sroa$9$3106109110117|0),($lengths$i$sroa$11$0119|0))|0;
  $403 = ($lengths$i$sroa$11$0119|0)==(0);
  if ($403) {
   $lengths$i$sroa$11$0121 = 0;
  } else {
   $mm$056$3$i = 0;
   while(1) {
    $404 = (($s$1$3$i) + ($mm$056$3$i)|0);
    $405 = HEAP8[$404>>0]|0;
    $406 = ($405<<24>>24)==(32);
    if ($406) {
     HEAP8[$404>>0] = 45;
    }
    $407 = (($mm$056$3$i) + 1)|0;
    $exitcond$3$i = ($407|0)==($lengths$i$sroa$11$0119|0);
    if ($exitcond$3$i) {
     $lengths$i$sroa$11$0121 = $lengths$i$sroa$11$0119;
     break;
    } else {
     $mm$056$3$i = $407;
    }
   }
  }
  $408 = (($s$1$3$i) + ($lengths$i$sroa$11$0121)|0);
  $s$2$3$i = $408;
 }
 HEAP8[$s$2$3$i>>0] = 0;
 $421 = HEAP32[$error$i20>>2]|0;
 HEAP32[$error>>2] = $421;
 $422 = ($421|0)==(0);
 if (!($422)) {
  $$0 = $421;
  STACKTOP = sp;return ($$0|0);
 }
 $423 = ((($155)) + 48|0);
 $424 = HEAP32[$423>>2]|0;
 $425 = (($424) + 1)|0;
 $426 = ((($bdfface)) + 16|0);
 HEAP32[$426>>2] = $425;
 $427 = ((($bdfface)) + 28|0);
 HEAP32[$427>>2] = 1;
 $428 = (_ft_mem_realloc($1,16,0,1,0,$error)|0);
 $429 = ((($bdfface)) + 32|0);
 HEAP32[$429>>2] = $428;
 $430 = HEAP32[$error>>2]|0;
 $431 = ($430|0)==(0);
 if (!($431)) {
  $$0 = $430;
  STACKTOP = sp;return ($$0|0);
 }
 ;HEAP32[$428>>2]=0|0;HEAP32[$428+4>>2]=0|0;HEAP32[$428+8>>2]=0|0;HEAP32[$428+12>>2]=0|0;
 $432 = ((($155)) + 40|0);
 $433 = HEAP32[$432>>2]|0;
 $434 = ((($155)) + 44|0);
 $435 = HEAP32[$434>>2]|0;
 $436 = (($435) + ($433))|0;
 $437 = $436&65535;
 HEAP16[$428>>1] = $437;
 $438 = ((($155)) + 72|0);
 $439 = HEAP32[$438>>2]|0;
 $440 = ($439|0)==(0);
 L242: do {
  if ($440) {
   label = 171;
  } else {
   $441 = ((($155)) + 128|0);
   $442 = HEAP32[$441>>2]|0;
   $443 = ((($442)) + 12|0);
   $444 = HEAP32[$443>>2]|0;
   $445 = ((($442)) + 4|0);
   $446 = HEAP32[$445>>2]|0;
   $447 = (3825651940 % ($446>>>0))&-1;
   $448 = (($444) + ($447<<2)|0);
   $449 = HEAP32[$448>>2]|0;
   $450 = ($449|0)==(0|0);
   if ($450) {
    label = 171;
   } else {
    $451 = (($446) + -1)|0;
    $452 = (($444) + ($451<<2)|0);
    $454 = $449;$ndp$01$i$i$i27 = $448;
    while(1) {
     $453 = HEAP32[$454>>2]|0;
     $455 = HEAP8[$453>>0]|0;
     $456 = ($455<<24>>24)==(65);
     if ($456) {
      $457 = (_strcmp($453,532482)|0);
      $458 = ($457|0)==(0);
      if ($458) {
       $$lcssa452 = $454;
       break;
      }
     }
     $459 = ((($ndp$01$i$i$i27)) + -4|0);
     $460 = ($459>>>0)<($444>>>0);
     $$$i$i$i29 = $460 ? $452 : $459;
     $461 = HEAP32[$$$i$i$i29>>2]|0;
     $462 = ($461|0)==(0|0);
     if ($462) {
      label = 171;
      break L242;
     } else {
      $454 = $461;$ndp$01$i$i$i27 = $$$i$i$i29;
     }
    }
    $463 = ((($155)) + 80|0);
    $464 = HEAP32[$463>>2]|0;
    $465 = ((($$lcssa452)) + 4|0);
    $466 = HEAP32[$465>>2]|0;
    $467 = (($464) + ($466<<4)|0);
    $468 = ($467|0)==(0|0);
    if ($468) {
     label = 171;
    } else {
     $469 = (((($464) + ($466<<4)|0)) + 12|0);
     $470 = HEAP32[$469>>2]|0;
     $471 = (($470) + 5)|0;
     $472 = (($471|0) / 10)&-1;
     $473 = $472&65535;
     $474 = ((($428)) + 2|0);
     HEAP16[$474>>1] = $473;
     $513 = $472;
    }
   }
  }
 } while(0);
 if ((label|0) == 171) {
  $sext272 = $436 << 16;
  $475 = $sext272 >> 15;
  $476 = (($475|0) / 3)&-1;
  $477 = $476&65535;
  $478 = ((($428)) + 2|0);
  HEAP16[$478>>1] = $477;
  $513 = $476;
 }
 $brmerge = $160 | $440;
 L255: do {
  if ($brmerge) {
   label = 180;
  } else {
   $479 = ((($155)) + 128|0);
   $480 = HEAP32[$479>>2]|0;
   $481 = ((($480)) + 12|0);
   $482 = HEAP32[$481>>2]|0;
   $483 = ((($480)) + 4|0);
   $484 = HEAP32[$483>>2]|0;
   $485 = (1780881776 % ($484>>>0))&-1;
   $486 = (($482) + ($485<<2)|0);
   $487 = HEAP32[$486>>2]|0;
   $488 = ($487|0)==(0|0);
   if ($488) {
    label = 180;
   } else {
    $489 = (($484) + -1)|0;
    $490 = (($482) + ($489<<2)|0);
    $492 = $487;$ndp$01$i$i$i39 = $486;
    while(1) {
     $491 = HEAP32[$492>>2]|0;
     $493 = HEAP8[$491>>0]|0;
     $494 = ($493<<24>>24)==(80);
     if ($494) {
      $495 = (_strcmp($491,532496)|0);
      $496 = ($495|0)==(0);
      if ($496) {
       $$lcssa450 = $492;
       break;
      }
     }
     $497 = ((($ndp$01$i$i$i39)) + -4|0);
     $498 = ($497>>>0)<($482>>>0);
     $$$i$i$i41 = $498 ? $490 : $497;
     $499 = HEAP32[$$$i$i$i41>>2]|0;
     $500 = ($499|0)==(0|0);
     if ($500) {
      label = 180;
      break L255;
     } else {
      $492 = $499;$ndp$01$i$i$i39 = $$$i$i$i41;
     }
    }
    $501 = ((($155)) + 80|0);
    $502 = HEAP32[$501>>2]|0;
    $503 = ((($$lcssa450)) + 4|0);
    $504 = HEAP32[$503>>2]|0;
    $505 = (($502) + ($504<<4)|0);
    $506 = ($505|0)==(0|0);
    if ($506) {
     label = 180;
    } else {
     $507 = (((($502) + ($504<<4)|0)) + 12|0);
     $508 = HEAP32[$507>>2]|0;
     $509 = ($508*460800)|0;
     $510 = (($509) + 36135)|0;
     $511 = (($510|0) / 72270)&-1;
     $512 = ((($428)) + 4|0);
     HEAP32[$512>>2] = $511;
     $517 = $511;
    }
   }
  }
 } while(0);
 if ((label|0) == 180) {
  $sext273 = $513 << 16;
  $514 = $sext273 >> 10;
  $515 = ((($428)) + 4|0);
  HEAP32[$515>>2] = $514;
  $517 = $514;
 }
 do {
  if ($160) {
   $516 = ((($428)) + 12|0);
   HEAP32[$516>>2] = $517;
   $628 = $517;
   label = 213;
  } else {
   if ($440) {
    $518 = ((($428)) + 12|0);
    HEAP32[$518>>2] = $517;
    $628 = $517;
    label = 213;
    break;
   }
   $519 = ((($155)) + 128|0);
   $520 = HEAP32[$519>>2]|0;
   $521 = ((($520)) + 12|0);
   $522 = HEAP32[$521>>2]|0;
   $523 = ((($520)) + 4|0);
   $524 = HEAP32[$523>>2]|0;
   $525 = (486426170 % ($524>>>0))&-1;
   $526 = (($522) + ($525<<2)|0);
   $527 = HEAP32[$526>>2]|0;
   $528 = ($527|0)==(0|0);
   if ($528) {
    $768 = 0;
    label = 193;
   } else {
    $529 = (($524) + -1)|0;
    $530 = (($522) + ($529<<2)|0);
    $532 = $527;$ndp$01$i$i$i51 = $526;
    while(1) {
     $531 = HEAP32[$532>>2]|0;
     $533 = HEAP8[$531>>0]|0;
     $534 = ($533<<24>>24)==(80);
     if ($534) {
      $535 = (_strcmp($531,532507)|0);
      $536 = ($535|0)==(0);
      if ($536) {
       $$lcssa448 = $532;
       label = 190;
       break;
      }
     }
     $537 = ((($ndp$01$i$i$i51)) + -4|0);
     $538 = ($537>>>0)<($522>>>0);
     $$$i$i$i53 = $538 ? $530 : $537;
     $539 = HEAP32[$$$i$i$i53>>2]|0;
     $540 = ($539|0)==(0|0);
     if ($540) {
      $769 = 0;
      break;
     } else {
      $532 = $539;$ndp$01$i$i$i51 = $$$i$i$i53;
     }
    }
    do {
     if ((label|0) == 190) {
      $541 = ((($155)) + 80|0);
      $542 = HEAP32[$541>>2]|0;
      $543 = ((($$lcssa448)) + 4|0);
      $544 = HEAP32[$543>>2]|0;
      $545 = (($542) + ($544<<4)|0);
      $546 = ($545|0)==(0|0);
      if ($546) {
       $769 = 0;
       break;
      }
      $547 = (((($542) + ($544<<4)|0)) + 12|0);
      $548 = HEAP32[$547>>2]|0;
      $sext6 = $548 << 16;
      $549 = $sext6 >> 10;
      $550 = ((($428)) + 12|0);
      HEAP32[$550>>2] = $549;
      $769 = $549;
     }
    } while(0);
    if ($440) {
     $614 = $769;$resolution_x$0133 = 0;$resolution_y$0 = 0;
    } else {
     $768 = $769;
     label = 193;
    }
   }
   L285: do {
    if ((label|0) == 193) {
     $551 = ((($155)) + 128|0);
     $552 = HEAP32[$551>>2]|0;
     $553 = ((($552)) + 12|0);
     $554 = HEAP32[$553>>2]|0;
     $555 = ((($552)) + 4|0);
     $556 = HEAP32[$555>>2]|0;
     $557 = (2286220677 % ($556>>>0))&-1;
     $558 = (($554) + ($557<<2)|0);
     $559 = HEAP32[$558>>2]|0;
     $560 = ($559|0)==(0|0);
     if ($560) {
      $resolution_x$0136277 = 0;
     } else {
      $561 = (($556) + -1)|0;
      $562 = (($554) + ($561<<2)|0);
      $564 = $559;$ndp$01$i$i$i63 = $558;
      while(1) {
       $563 = HEAP32[$564>>2]|0;
       $565 = HEAP8[$563>>0]|0;
       $566 = ($565<<24>>24)==(82);
       if ($566) {
        $567 = (_strcmp($563,532518)|0);
        $568 = ($567|0)==(0);
        if ($568) {
         $$lcssa446 = $564;
         label = 198;
         break;
        }
       }
       $569 = ((($ndp$01$i$i$i63)) + -4|0);
       $570 = ($569>>>0)<($554>>>0);
       $$$i$i$i65 = $570 ? $562 : $569;
       $571 = HEAP32[$$$i$i$i65>>2]|0;
       $572 = ($571|0)==(0|0);
       if ($572) {
        $resolution_x$0136 = 0;
        break;
       } else {
        $564 = $571;$ndp$01$i$i$i63 = $$$i$i$i65;
       }
      }
      do {
       if ((label|0) == 198) {
        $573 = ((($155)) + 80|0);
        $574 = HEAP32[$573>>2]|0;
        $575 = ((($$lcssa446)) + 4|0);
        $576 = HEAP32[$575>>2]|0;
        $577 = (($574) + ($576<<4)|0);
        $578 = ($577|0)==(0|0);
        if ($578) {
         $resolution_x$0136 = 0;
         break;
        }
        $579 = (((($574) + ($576<<4)|0)) + 12|0);
        $580 = HEAP32[$579>>2]|0;
        $resolution_x$0136 = $580;
       }
      } while(0);
      if ($440) {
       $614 = $768;$resolution_x$0133 = $resolution_x$0136;$resolution_y$0 = 0;
       break;
      } else {
       $resolution_x$0136277 = $resolution_x$0136;
      }
     }
     $581 = ((($155)) + 128|0);
     $582 = HEAP32[$581>>2]|0;
     $583 = ((($582)) + 12|0);
     $584 = HEAP32[$583>>2]|0;
     $585 = ((($582)) + 4|0);
     $586 = HEAP32[$585>>2]|0;
     $587 = (2286220678 % ($586>>>0))&-1;
     $588 = (($584) + ($587<<2)|0);
     $589 = HEAP32[$588>>2]|0;
     $590 = ($589|0)==(0|0);
     if ($590) {
      $614 = $768;$resolution_x$0133 = $resolution_x$0136277;$resolution_y$0 = 0;
      break;
     }
     $591 = (($586) + -1)|0;
     $592 = (($584) + ($591<<2)|0);
     $594 = $589;$ndp$01$i$i$i75 = $588;
     while(1) {
      $593 = HEAP32[$594>>2]|0;
      $595 = HEAP8[$593>>0]|0;
      $596 = ($595<<24>>24)==(82);
      if ($596) {
       $597 = (_strcmp($593,532531)|0);
       $598 = ($597|0)==(0);
       if ($598) {
        $$lcssa444 = $594;
        break;
       }
      }
      $599 = ((($ndp$01$i$i$i75)) + -4|0);
      $600 = ($599>>>0)<($584>>>0);
      $$$i$i$i77 = $600 ? $592 : $599;
      $601 = HEAP32[$$$i$i$i77>>2]|0;
      $602 = ($601|0)==(0|0);
      if ($602) {
       $614 = $768;$resolution_x$0133 = $resolution_x$0136277;$resolution_y$0 = 0;
       break L285;
      } else {
       $594 = $601;$ndp$01$i$i$i75 = $$$i$i$i77;
      }
     }
     $603 = ((($155)) + 80|0);
     $604 = HEAP32[$603>>2]|0;
     $605 = ((($$lcssa444)) + 4|0);
     $606 = HEAP32[$605>>2]|0;
     $607 = (($604) + ($606<<4)|0);
     $608 = ($607|0)==(0|0);
     if ($608) {
      $614 = $768;$resolution_x$0133 = $resolution_x$0136277;$resolution_y$0 = 0;
      break;
     }
     $609 = (((($604) + ($606<<4)|0)) + 12|0);
     $610 = HEAP32[$609>>2]|0;
     $611 = $610&65535;
     $614 = $768;$resolution_x$0133 = $resolution_x$0136277;$resolution_y$0 = $611;
    }
   } while(0);
   $612 = ((($428)) + 12|0);
   $613 = ($614|0)==(0);
   if ($613) {
    HEAP32[$612>>2] = $517;
    $615 = ($resolution_y$0<<16>>16)==(0);
    if ($615) {
     $628 = $517;
     label = 213;
     break;
    }
    $616 = $resolution_y$0 << 16 >> 16;
    $617 = Math_imul($517, $616)|0;
    $618 = (($617|0) / 72)&-1;
    HEAP32[$612>>2] = $618;
    $624 = $618;
   } else {
    $624 = $614;
   }
   $sext = $resolution_x$0133 << 16;
   $619 = $sext >> 16;
   $620 = ($619|0)==(0);
   $621 = ($resolution_y$0<<16>>16)==(0);
   $or$cond7 = $621 | $620;
   if ($or$cond7) {
    $628 = $624;
    label = 213;
    break;
   }
   $622 = $resolution_y$0 << 16 >> 16;
   $623 = Math_imul($624, $619)|0;
   $625 = (($623|0) / ($622|0))&-1;
   $626 = ((($428)) + 8|0);
   HEAP32[$626>>2] = $625;
  }
 } while(0);
 if ((label|0) == 213) {
  $627 = ((($428)) + 8|0);
  HEAP32[$627>>2] = $628;
 }
 $629 = ((($155)) + 56|0);
 $630 = HEAP32[$629>>2]|0;
 $631 = HEAP32[$423>>2]|0;
 $632 = (_ft_mem_realloc($1,8,0,$631,0,$error)|0);
 $633 = ((($bdfface)) + 144|0);
 HEAP32[$633>>2] = $632;
 $634 = HEAP32[$error>>2]|0;
 $635 = ($634|0)==(0);
 if (!($635)) {
  $$0 = $634;
  STACKTOP = sp;return ($$0|0);
 }
 $636 = ((($bdfface)) + 164|0);
 HEAP32[$636>>2] = 0;
 $637 = HEAP32[$423>>2]|0;
 $638 = ($637|0)==(0);
 if (!($638)) {
  $639 = ((($155)) + 36|0);
  $640 = HEAP32[$639>>2]|0;
  $n$0200 = 0;
  while(1) {
   $641 = (((($630) + (($n$0200*36)|0)|0)) + 4|0);
   $642 = HEAP32[$641>>2]|0;
   $643 = (($632) + ($n$0200<<3)|0);
   HEAP32[$643>>2] = $642;
   $644 = $n$0200&65535;
   $645 = (((($632) + ($n$0200<<3)|0)) + 4|0);
   HEAP16[$645>>1] = $644;
   $646 = ($642|0)==($640|0);
   if ($646) {
    HEAP32[$636>>2] = $n$0200;
   }
   $647 = (($n$0200) + 1)|0;
   $648 = ($647>>>0)<($637>>>0);
   if ($648) {
    $n$0200 = $647;
   } else {
    break;
   }
  }
 }
 $649 = HEAP32[$438>>2]|0;
 $650 = ($649|0)==(0);
 L327: do {
  if ($650) {
   $$0$i92 = 0;
  } else {
   $651 = ((($155)) + 128|0);
   $652 = HEAP32[$651>>2]|0;
   $653 = ((($652)) + 12|0);
   $654 = HEAP32[$653>>2]|0;
   $655 = ((($652)) + 4|0);
   $656 = HEAP32[$655>>2]|0;
   $657 = (3781719536 % ($656>>>0))&-1;
   $658 = (($654) + ($657<<2)|0);
   $659 = HEAP32[$658>>2]|0;
   $660 = ($659|0)==(0|0);
   if ($660) {
    $$0$i92 = 0;
    break;
   }
   $661 = (($656) + -1)|0;
   $662 = (($654) + ($661<<2)|0);
   $664 = $659;$ndp$01$i$i$i87 = $658;
   while(1) {
    $663 = HEAP32[$664>>2]|0;
    $665 = HEAP8[$663>>0]|0;
    $666 = ($665<<24>>24)==(67);
    if ($666) {
     $667 = (_strcmp($663,593036)|0);
     $668 = ($667|0)==(0);
     if ($668) {
      $$lcssa442 = $664;
      break;
     }
    }
    $669 = ((($ndp$01$i$i$i87)) + -4|0);
    $670 = ($669>>>0)<($654>>>0);
    $$$i$i$i89 = $670 ? $662 : $669;
    $671 = HEAP32[$$$i$i$i89>>2]|0;
    $672 = ($671|0)==(0|0);
    if ($672) {
     $$0$i92 = 0;
     break L327;
    } else {
     $664 = $671;$ndp$01$i$i$i87 = $$$i$i$i89;
    }
   }
   $673 = ((($155)) + 80|0);
   $674 = HEAP32[$673>>2]|0;
   $675 = ((($$lcssa442)) + 4|0);
   $676 = HEAP32[$675>>2]|0;
   $677 = (($674) + ($676<<4)|0);
   $$0$i92 = $677;
  }
 } while(0);
 $brmerge336 = $160 | $650;
 L337: do {
  if (!($brmerge336)) {
   $678 = ((($155)) + 128|0);
   $679 = HEAP32[$678>>2]|0;
   $680 = ((($679)) + 12|0);
   $681 = HEAP32[$680>>2]|0;
   $682 = ((($679)) + 4|0);
   $683 = HEAP32[$682>>2]|0;
   $684 = (1888187142 % ($683>>>0))&-1;
   $685 = (($681) + ($684<<2)|0);
   $686 = HEAP32[$685>>2]|0;
   $687 = ($686|0)==(0|0);
   if ($687) {
    break;
   }
   $688 = (($683) + -1)|0;
   $689 = (($681) + ($688<<2)|0);
   $691 = $686;$ndp$01$i$i$i99 = $685;
   while(1) {
    $690 = HEAP32[$691>>2]|0;
    $692 = HEAP8[$690>>0]|0;
    $693 = ($692<<24>>24)==(67);
    if ($693) {
     $694 = (_strcmp($690,593053)|0);
     $695 = ($694|0)==(0);
     if ($695) {
      $$lcssa = $691;
      break;
     }
    }
    $696 = ((($ndp$01$i$i$i99)) + -4|0);
    $697 = ($696>>>0)<($681>>>0);
    $$$i$i$i101 = $697 ? $689 : $696;
    $698 = HEAP32[$$$i$i$i101>>2]|0;
    $699 = ($698|0)==(0|0);
    if ($699) {
     break L337;
    } else {
     $691 = $698;$ndp$01$i$i$i99 = $$$i$i$i101;
    }
   }
   $700 = ((($155)) + 80|0);
   $701 = HEAP32[$700>>2]|0;
   $702 = ((($$lcssa)) + 4|0);
   $703 = HEAP32[$702>>2]|0;
   $704 = (($701) + ($703<<4)|0);
   $705 = ($$0$i92|0)!=(0|0);
   $706 = ($704|0)!=(0|0);
   $or$cond = $705 & $706;
   if (!($or$cond)) {
    break;
   }
   $707 = ((($$0$i92)) + 4|0);
   $708 = HEAP32[$707>>2]|0;
   $709 = ($708|0)==(1);
   if (!($709)) {
    break;
   }
   $710 = (((($701) + ($703<<4)|0)) + 4|0);
   $711 = HEAP32[$710>>2]|0;
   $712 = ($711|0)==(1);
   if (!($712)) {
    break;
   }
   $713 = ((($$0$i92)) + 12|0);
   $714 = HEAP32[$713>>2]|0;
   $715 = ($714|0)==(0|0);
   if ($715) {
    break;
   }
   $716 = (((($701) + ($703<<4)|0)) + 12|0);
   $717 = HEAP32[$716>>2]|0;
   $718 = ($717|0)==(0|0);
   if ($718) {
    break;
   }
   $719 = (_ft_mem_strdup($1,$717,$error)|0);
   $720 = ((($bdfface)) + 132|0);
   HEAP32[$720>>2] = $719;
   $721 = HEAP32[$error>>2]|0;
   $722 = ($721|0)==(0);
   if (!($722)) {
    $$0 = $721;
    STACKTOP = sp;return ($$0|0);
   }
   $723 = HEAP32[$713>>2]|0;
   $724 = (_ft_mem_strdup($1,$723,$error)|0);
   $725 = ((($bdfface)) + 136|0);
   HEAP32[$725>>2] = $724;
   $726 = HEAP32[$error>>2]|0;
   $727 = ($726|0)==(0);
   if (!($727)) {
    $$0 = $726;
    STACKTOP = sp;return ($$0|0);
   }
   $728 = HEAP8[$724>>0]|0;
   L357: do {
    switch ($728<<24>>24) {
    case 73: case 105:  {
     $729 = ((($724)) + 1|0);
     $730 = HEAP8[$729>>0]|0;
     switch ($730<<24>>24) {
     case 83: case 115:  {
      break;
     }
     default: {
      $unicode_charmap$0 = 0;
      break L357;
     }
     }
     $731 = ((($724)) + 2|0);
     $732 = HEAP8[$731>>0]|0;
     switch ($732<<24>>24) {
     case 79: case 111:  {
      break;
     }
     default: {
      $unicode_charmap$0 = 0;
      break L357;
     }
     }
     $733 = ((($724)) + 3|0);
     $734 = (_strcmp($733,532544)|0);
     $735 = ($734|0)==(0);
     if (!($735)) {
      $736 = (_strcmp($733,532550)|0);
      $737 = ($736|0)==(0);
      if (!($737)) {
       $unicode_charmap$0 = 0;
       break L357;
      }
      $738 = HEAP32[$720>>2]|0;
      $739 = (_strcmp($738,532555)|0);
      $740 = ($739|0)==(0);
      if (!($740)) {
       $unicode_charmap$0 = 0;
       break L357;
      }
     }
     $unicode_charmap$0 = 1;
     break;
    }
    default: {
     $unicode_charmap$0 = 0;
    }
    }
   } while(0);
   HEAP32[$charmap>>2] = $bdfface;
   $741 = ((($charmap)) + 4|0);
   HEAP32[$741>>2] = 0;
   $742 = ((($charmap)) + 8|0);
   HEAP16[$742>>1] = 0;
   $743 = ((($charmap)) + 10|0);
   HEAP16[$743>>1] = 0;
   if ($unicode_charmap$0) {
    HEAP32[$741>>2] = 1970170211;
    HEAP16[$742>>1] = 3;
    HEAP16[$743>>1] = 1;
   }
   $744 = (_FT_CMap_New(49328,0,$charmap,0)|0);
   HEAP32[$error>>2] = $744;
   $$0 = $744;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 HEAP32[$charmap1>>2] = $bdfface;
 $745 = ((($charmap1)) + 4|0);
 HEAP32[$745>>2] = 1094995778;
 $746 = ((($charmap1)) + 8|0);
 HEAP16[$746>>1] = 7;
 $747 = ((($charmap1)) + 10|0);
 HEAP16[$747>>1] = 0;
 $748 = (_FT_CMap_New(49328,0,$charmap1,0)|0);
 HEAP32[$error>>2] = $748;
 $749 = ((($bdfface)) + 36|0);
 $750 = HEAP32[$749>>2]|0;
 $751 = ($750|0)==(0);
 if (!($751)) {
  $752 = ((($bdfface)) + 40|0);
  $753 = HEAP32[$752>>2]|0;
  $754 = HEAP32[$753>>2]|0;
  $755 = ((($bdfface)) + 92|0);
  HEAP32[$755>>2] = $754;
 }
 $$0 = $748;
 STACKTOP = sp;return ($$0|0);
}
function __bdf_parse_start($line,$linelen,$lineno,$call_data,$client_data) {
 $line = $line|0;
 $linelen = $linelen|0;
 $lineno = $lineno|0;
 $call_data = $call_data|0;
 $client_data = $client_data|0;
 var $$0$i112 = 0, $$01$i = 0, $$01$i32 = 0, $$01$i55 = 0, $$01$i64 = 0, $$01$i73 = 0, $$01$i85 = 0, $$01$i97 = 0, $$02 = 0, $$02$i = 0, $$02$i26 = 0, $$02$i34 = 0, $$02$i49 = 0, $$02$i58 = 0, $$02$i67 = 0, $$02$v = 0, $$1$i = 0, $$1$i27 = 0, $$1$i35 = 0, $$1$i44 = 0;
 var $$1$i50 = 0, $$1$i59 = 0, $$1$i68 = 0, $$1$i78 = 0, $$1$i90 = 0, $$13$i = 0, $$13$i77 = 0, $$13$i89 = 0, $$131$i = 0, $$131$i30 = 0, $$131$i38 = 0, $$131$i53 = 0, $$131$i62 = 0, $$131$i71 = 0, $$21$i = 0, $$21$i80 = 0, $$21$i92 = 0, $$lcssa$i = 0, $$lcssa$i83 = 0, $$lcssa$i95 = 0;
 var $$lcssa177 = 0, $$lcssa180 = 0, $$lcssa181 = 0, $$phi$trans$insert$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i103 = 0, $$pre$i24 = 0, $$pre$i25 = 0, $$pre$i33 = 0, $$pre$i43 = 0, $$pre$i48 = 0, $$pre$i57 = 0, $$pre$i66 = 0, $$pre$i76 = 0, $$pre$i88 = 0, $$pre10$i = 0, $$pre11$i = 0, $$pre9$i = 0, $$s$i = 0;
 var $$s$i75 = 0, $$s$i87 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0;
 var $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0;
 var $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0;
 var $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0;
 var $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0;
 var $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0;
 var $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0;
 var $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0;
 var $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0;
 var $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0;
 var $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0;
 var $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0;
 var $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0;
 var $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0;
 var $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0;
 var $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0;
 var $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0;
 var $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0;
 var $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0;
 var $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0;
 var $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0;
 var $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bitcount$0$i1$0 = 0, $bitcount$0$i1$0$lcssa = 0, $bitcount$0$lcssa138 = 0, $bitcount$0125 = 0, $brmerge = 0, $dmap$1$i = 0, $dmap$1$i28 = 0, $dmap$1$i36 = 0, $dmap$1$i45 = 0;
 var $dmap$1$i51 = 0, $dmap$1$i60 = 0, $dmap$1$i69 = 0, $dmap$1$i79 = 0, $dmap$1$i91 = 0, $error = 0, $error$i = 0, $error$i22 = 0, $error$i23 = 0, $exitcond = 0, $fp$02$i = 0, $i$01$i = 0, $i$0123 = 0, $i$03$i = 0, $i1$0126 = 0, $j$0$be$i = 0, $j$0$be$i$lcssa = 0, $j$04$i = 0, $j$1$lcssa$i = 0, $j$11$i = 0;
 var $list$i = 0, $memory$0 = 0, $name$i = 0, $nbuf = 0, $phitmp5 = 0, $prop$0124 = 0, $s$0 = 0, $sext$i = 0, $sext$i81 = 0, $sext$i93 = 0, $shift$0127 = 0, $u$02$i = 0, $v$02$i = 0, $v$02$i29 = 0, $v$02$i37 = 0, $v$02$i52 = 0, $v$02$i61 = 0, $v$02$i70 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 432|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $name$i = sp + 176|0;
 $list$i = sp + 32|0;
 $error$i23 = sp + 24|0;
 $error$i22 = sp + 20|0;
 $error$i = sp + 16|0;
 $error = sp + 12|0;
 $nbuf = sp + 48|0;
 HEAP32[$error>>2] = 0;
 $0 = ((($client_data)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $memory$0 = 0;
 } else {
  $3 = ((($1)) + 278664|0);
  $4 = HEAP32[$3>>2]|0;
  $memory$0 = $4;
 }
 $5 = (_strncmp($line,530598,7)|0);
 $6 = ($5|0)==(0);
 L4: do {
  if ($6) {
   $7 = ((($line)) + 7|0);
   $8 = HEAP8[$7>>0]|0;
   switch ($8<<24>>24) {
   case 9: case 13: case 10: case 0: case 32:  {
    break;
   }
   default: {
    break L4;
   }
   }
   $9 = ((($client_data)) + 36|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ((($10)) + 8|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($12|0)==(0);
   $brmerge = $13 | $2;
   if ($brmerge) {
    $624 = 0;
    STACKTOP = sp;return ($624|0);
   }
   $14 = ($8<<24>>24)==(0);
   $$02$v = $14 ? -7 : -8;
   $$02 = (($$02$v) + ($linelen))|0;
   $15 = ((($1)) + 278664|0);
   $16 = HEAP32[$15>>2]|0;
   HEAP32[$error$i>>2] = 0;
   $17 = ((($1)) + 88|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = (($$02) + 1)|0;
   $20 = (($19) + ($18))|0;
   $21 = ((($1)) + 84|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = (_ft_mem_realloc($16,1,$18,$20,$22,$error$i)|0);
   HEAP32[$21>>2] = $23;
   $24 = HEAP32[$error$i>>2]|0;
   $25 = ($24|0)==(0);
   if ($25) {
    $26 = ((($line)) + 8|0);
    $s$0 = $14 ? $7 : $26;
    $27 = HEAP32[$17>>2]|0;
    $28 = (($23) + ($27)|0);
    _memcpy(($28|0),($s$0|0),($$02|0))|0;
    $29 = (($28) + ($$02)|0);
    HEAP8[$29>>0] = 10;
    $30 = HEAP32[$17>>2]|0;
    $31 = (($30) + ($19))|0;
    HEAP32[$17>>2] = $31;
    $$pre$i = HEAP32[$error$i>>2]|0;
    $32 = $$pre$i;
   } else {
    $32 = $24;
   }
   HEAP32[$error>>2] = $32;
   $624 = $32;
   STACKTOP = sp;return ($624|0);
  }
 } while(0);
 $33 = HEAP32[$client_data>>2]|0;
 $34 = $33 & 1;
 $35 = ($34|0)==(0);
 if ($35) {
  $36 = ((($client_data)) + 139320|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = (_strncmp($line,530606,9)|0);
  $39 = ($38|0)==(0);
  L17: do {
   if ($39) {
    $40 = ((($line)) + 9|0);
    $41 = HEAP8[$40>>0]|0;
    switch ($41<<24>>24) {
    case 9: case 13: case 10: case 0: case 32:  {
     break;
    }
    default: {
     break L17;
    }
    }
    HEAP32[$client_data>>2] = 1;
    HEAP32[$0>>2] = 0;
    $42 = (_ft_mem_alloc($37,278692,$error)|0);
    $43 = HEAP32[$error>>2]|0;
    $44 = ($43|0)==(0);
    if (!($44)) {
     $624 = $43;
     STACKTOP = sp;return ($624|0);
    }
    HEAP32[$0>>2] = $42;
    $45 = HEAP32[$36>>2]|0;
    $46 = ((($42)) + 278664|0);
    HEAP32[$46>>2] = $45;
    HEAP32[$36>>2] = 0;
    $47 = ((($42)) + 278676|0);
    HEAP32[$error$i22>>2] = 0;
    $48 = ((($42)) + 278680|0);
    HEAP32[$48>>2] = 241;
    HEAP32[$47>>2] = 80;
    $49 = ((($42)) + 278684|0);
    HEAP32[$49>>2] = 0;
    $50 = (_ft_mem_realloc($37,4,0,241,0,$error$i22)|0);
    $51 = ((($42)) + 278688|0);
    HEAP32[$51>>2] = $50;
    $52 = HEAP32[$error$i22>>2]|0;
    HEAP32[$error>>2] = $52;
    $53 = ($52|0)==(0);
    if ($53) {
     $i$0123 = 0;$prop$0124 = 48000;
    } else {
     $624 = $52;
     STACKTOP = sp;return ($624|0);
    }
    while(1) {
     $54 = HEAP32[$prop$0124>>2]|0;
     $55 = (_hash_insert($54,$i$0123,$47,$37)|0);
     HEAP32[$error>>2] = $55;
     $56 = ($55|0)==(0);
     if (!($56)) {
      $624 = $55;
      label = 155;
      break;
     }
     $57 = (($i$0123) + 1)|0;
     $58 = ((($prop$0124)) + 16|0);
     $59 = ($57>>>0)<(83);
     if ($59) {
      $i$0123 = $57;$prop$0124 = $58;
     } else {
      break;
     }
    }
    if ((label|0) == 155) {
     STACKTOP = sp;return ($624|0);
    }
    $60 = (_ft_mem_alloc($37,16,$error)|0);
    $61 = HEAP32[$0>>2]|0;
    $62 = ((($61)) + 128|0);
    HEAP32[$62>>2] = $60;
    $63 = HEAP32[$error>>2]|0;
    $64 = ($63|0)==(0);
    if (!($64)) {
     $624 = $63;
     STACKTOP = sp;return ($624|0);
    }
    HEAP32[$error$i23>>2] = 0;
    $65 = ((($60)) + 4|0);
    HEAP32[$65>>2] = 241;
    HEAP32[$60>>2] = 80;
    $66 = ((($60)) + 8|0);
    HEAP32[$66>>2] = 0;
    $67 = (_ft_mem_realloc($37,4,0,241,0,$error$i23)|0);
    $68 = ((($60)) + 12|0);
    HEAP32[$68>>2] = $67;
    $69 = HEAP32[$error$i23>>2]|0;
    HEAP32[$error>>2] = $69;
    $70 = ($69|0)==(0);
    if (!($70)) {
     $624 = $69;
     STACKTOP = sp;return ($624|0);
    }
    $71 = ((($client_data)) + 36|0);
    $72 = HEAP32[$71>>2]|0;
    $73 = ((($72)) + 12|0);
    $74 = HEAP32[$73>>2]|0;
    $75 = HEAP32[$0>>2]|0;
    $76 = ((($75)) + 28|0);
    HEAP32[$76>>2] = $74;
    $77 = ((($75)) + 36|0);
    HEAP32[$77>>2] = -1;
    $$pre = HEAP32[$error>>2]|0;
    $624 = $$pre;
    STACKTOP = sp;return ($624|0);
   }
  } while(0);
  HEAP32[$error>>2] = 176;
  $624 = 176;
  STACKTOP = sp;return ($624|0);
 }
 $78 = (_strncmp($line,531664,15)|0);
 $79 = ($78|0)==(0);
 L41: do {
  if ($79) {
   $80 = ((($line)) + 15|0);
   $81 = HEAP8[$80>>0]|0;
   switch ($81<<24>>24) {
   case 9: case 13: case 10: case 0: case 32:  {
    break;
   }
   default: {
    break L41;
   }
   }
   $82 = $33 & 8;
   $83 = ($82|0)==(0);
   if ($83) {
    HEAP32[$error>>2] = 179;
    $624 = 179;
    STACKTOP = sp;return ($624|0);
   }
   $84 = ((($client_data)) + 139304|0);
   $85 = (__bdf_list_split($84,531680,$line,$linelen)|0);
   HEAP32[$error>>2] = $85;
   $86 = ($85|0)==(0);
   if (!($86)) {
    $624 = $85;
    STACKTOP = sp;return ($624|0);
   }
   $87 = HEAP32[$84>>2]|0;
   $88 = ((($87)) + 4|0);
   $89 = HEAP32[$88>>2]|0;
   $90 = ($89|0)==(0|0);
   L51: do {
    if ($90) {
     $$01$i = 0;
    } else {
     $91 = HEAP8[$89>>0]|0;
     L53: do {
      switch ($91<<24>>24) {
      case 0:  {
       $$01$i = 0;
       break L51;
       break;
      }
      case 48:  {
       $92 = ((($89)) + 1|0);
       $93 = HEAP8[$92>>0]|0;
       switch ($93<<24>>24) {
       case 88: case 120:  {
        break;
       }
       default: {
        $$02$i = $89;$$1$i = 10;$96 = 48;$dmap$1$i = 531715;
        break L53;
       }
       }
       $94 = ((($89)) + 2|0);
       $$pre$i24 = HEAP8[$94>>0]|0;
       $$02$i = $94;$$1$i = 16;$96 = $$pre$i24;$dmap$1$i = 531683;
       break;
      }
      default: {
       $$02$i = $89;$$1$i = 10;$96 = $91;$dmap$1$i = 531715;
      }
      }
     } while(0);
     $95 = $96&255;
     $97 = $95 >>> 3;
     $98 = (($dmap$1$i) + ($97)|0);
     $99 = HEAP8[$98>>0]|0;
     $100 = $99&255;
     $101 = $96 << 24 >> 24;
     $102 = $101 & 7;
     $103 = 1 << $102;
     $104 = $100 & $103;
     $105 = ($104|0)==(0);
     if ($105) {
      $$01$i = 0;
     } else {
      $$131$i = $$02$i;$108 = $101;$v$02$i = 0;
      while(1) {
       $106 = Math_imul($v$02$i, $$1$i)|0;
       $107 = (531747 + ($108)|0);
       $109 = HEAP8[$107>>0]|0;
       $110 = $109&255;
       $111 = (($110) + ($106))|0;
       $112 = ((($$131$i)) + 1|0);
       $113 = HEAP8[$112>>0]|0;
       $114 = $113&255;
       $115 = $114 >>> 3;
       $116 = (($dmap$1$i) + ($115)|0);
       $117 = HEAP8[$116>>0]|0;
       $118 = $117&255;
       $119 = $113 << 24 >> 24;
       $120 = $119 & 7;
       $121 = 1 << $120;
       $122 = $121 & $118;
       $123 = ($122|0)==(0);
       if ($123) {
        $$01$i = $111;
        break;
       } else {
        $$131$i = $112;$108 = $119;$v$02$i = $111;
       }
      }
     }
    }
   } while(0);
   $124 = HEAP32[$0>>2]|0;
   $125 = ((($124)) + 72|0);
   HEAP32[$125>>2] = $$01$i;
   $126 = ((($client_data)) + 4|0);
   HEAP32[$126>>2] = $$01$i;
   $127 = (_ft_mem_realloc($memory$0,16,0,$$01$i,0,$error)|0);
   $128 = HEAP32[$0>>2]|0;
   $129 = ((($128)) + 80|0);
   HEAP32[$129>>2] = $127;
   $130 = HEAP32[$error>>2]|0;
   $131 = ($130|0)==(0);
   if ($131) {
    $133 = HEAP32[$client_data>>2]|0;
    $134 = $133 | 16;
    HEAP32[$client_data>>2] = $134;
    HEAP32[$call_data>>2] = 4353;
    $624 = 0;
    STACKTOP = sp;return ($624|0);
   } else {
    $132 = ((($128)) + 72|0);
    HEAP32[$132>>2] = 0;
    $624 = $130;
    STACKTOP = sp;return ($624|0);
   }
  }
 } while(0);
 $135 = (_strncmp($line,531989,15)|0);
 $136 = ($135|0)==(0);
 L66: do {
  if ($136) {
   $137 = ((($line)) + 15|0);
   $138 = HEAP8[$137>>0]|0;
   switch ($138<<24>>24) {
   case 9: case 13: case 10: case 0: case 32:  {
    break;
   }
   default: {
    break L66;
   }
   }
   $139 = $33 & 4;
   $140 = ($139|0)==(0);
   if ($140) {
    HEAP32[$error>>2] = 178;
    $624 = 178;
    STACKTOP = sp;return ($624|0);
   }
   $141 = ((($client_data)) + 139304|0);
   $142 = (__bdf_list_split($141,531680,$line,$linelen)|0);
   HEAP32[$error>>2] = $142;
   $143 = ($142|0)==(0);
   if (!($143)) {
    $624 = $142;
    STACKTOP = sp;return ($624|0);
   }
   $144 = HEAP32[$141>>2]|0;
   $145 = ((($144)) + 4|0);
   $146 = HEAP32[$145>>2]|0;
   $147 = ($146|0)==(0|0);
   L76: do {
    if ($147) {
     $$01$i32 = 0;
    } else {
     $148 = HEAP8[$146>>0]|0;
     L78: do {
      switch ($148<<24>>24) {
      case 0:  {
       $$01$i32 = 0;
       break L76;
       break;
      }
      case 48:  {
       $149 = ((($146)) + 1|0);
       $150 = HEAP8[$149>>0]|0;
       switch ($150<<24>>24) {
       case 88: case 120:  {
        break;
       }
       default: {
        $$02$i26 = $146;$$1$i27 = 10;$153 = 48;$dmap$1$i28 = 531715;
        break L78;
       }
       }
       $151 = ((($146)) + 2|0);
       $$pre$i25 = HEAP8[$151>>0]|0;
       $$02$i26 = $151;$$1$i27 = 16;$153 = $$pre$i25;$dmap$1$i28 = 531683;
       break;
      }
      default: {
       $$02$i26 = $146;$$1$i27 = 10;$153 = $148;$dmap$1$i28 = 531715;
      }
      }
     } while(0);
     $152 = $153&255;
     $154 = $152 >>> 3;
     $155 = (($dmap$1$i28) + ($154)|0);
     $156 = HEAP8[$155>>0]|0;
     $157 = $156&255;
     $158 = $153 << 24 >> 24;
     $159 = $158 & 7;
     $160 = 1 << $159;
     $161 = $157 & $160;
     $162 = ($161|0)==(0);
     if ($162) {
      $$01$i32 = 0;
     } else {
      $$131$i30 = $$02$i26;$166 = $158;$v$02$i29 = 0;
      while(1) {
       $163 = $v$02$i29 & 65535;
       $164 = Math_imul($163, $$1$i27)|0;
       $165 = (531747 + ($166)|0);
       $167 = HEAP8[$165>>0]|0;
       $168 = $167&255;
       $169 = (($168) + ($164))|0;
       $170 = ((($$131$i30)) + 1|0);
       $171 = HEAP8[$170>>0]|0;
       $172 = $171&255;
       $173 = $172 >>> 3;
       $174 = (($dmap$1$i28) + ($173)|0);
       $175 = HEAP8[$174>>0]|0;
       $176 = $175&255;
       $177 = $171 << 24 >> 24;
       $178 = $177 & 7;
       $179 = 1 << $178;
       $180 = $179 & $176;
       $181 = ($180|0)==(0);
       if ($181) {
        $$lcssa181 = $169;
        break;
       } else {
        $$131$i30 = $170;$166 = $177;$v$02$i29 = $169;
       }
      }
      $182 = $$lcssa181&65535;
      $$01$i32 = $182;
     }
    }
   } while(0);
   $183 = HEAP32[$0>>2]|0;
   $184 = ((($183)) + 4|0);
   HEAP16[$184>>1] = $$01$i32;
   $185 = ((($144)) + 8|0);
   $186 = HEAP32[$185>>2]|0;
   $187 = ($186|0)==(0|0);
   L86: do {
    if ($187) {
     $224 = 0;
    } else {
     $188 = HEAP8[$186>>0]|0;
     L88: do {
      switch ($188<<24>>24) {
      case 0:  {
       $224 = 0;
       break L86;
       break;
      }
      case 48:  {
       $189 = ((($186)) + 1|0);
       $190 = HEAP8[$189>>0]|0;
       switch ($190<<24>>24) {
       case 88: case 120:  {
        break;
       }
       default: {
        $$02$i34 = $186;$$1$i35 = 10;$193 = 48;$dmap$1$i36 = 531715;
        break L88;
       }
       }
       $191 = ((($186)) + 2|0);
       $$pre$i33 = HEAP8[$191>>0]|0;
       $$02$i34 = $191;$$1$i35 = 16;$193 = $$pre$i33;$dmap$1$i36 = 531683;
       break;
      }
      default: {
       $$02$i34 = $186;$$1$i35 = 10;$193 = $188;$dmap$1$i36 = 531715;
      }
      }
     } while(0);
     $192 = $193&255;
     $194 = $192 >>> 3;
     $195 = (($dmap$1$i36) + ($194)|0);
     $196 = HEAP8[$195>>0]|0;
     $197 = $196&255;
     $198 = $193 << 24 >> 24;
     $199 = $198 & 7;
     $200 = 1 << $199;
     $201 = $197 & $200;
     $202 = ($201|0)==(0);
     if ($202) {
      $224 = 0;
     } else {
      $$131$i38 = $$02$i34;$206 = $198;$v$02$i37 = 0;
      while(1) {
       $203 = $v$02$i37 & 65535;
       $204 = Math_imul($203, $$1$i35)|0;
       $205 = (531747 + ($206)|0);
       $207 = HEAP8[$205>>0]|0;
       $208 = $207&255;
       $209 = (($208) + ($204))|0;
       $210 = ((($$131$i38)) + 1|0);
       $211 = HEAP8[$210>>0]|0;
       $212 = $211&255;
       $213 = $212 >>> 3;
       $214 = (($dmap$1$i36) + ($213)|0);
       $215 = HEAP8[$214>>0]|0;
       $216 = $215&255;
       $217 = $211 << 24 >> 24;
       $218 = $217 & 7;
       $219 = 1 << $218;
       $220 = $219 & $216;
       $221 = ($220|0)==(0);
       if ($221) {
        $$lcssa180 = $209;
        break;
       } else {
        $$131$i38 = $210;$206 = $217;$v$02$i37 = $209;
       }
      }
      $222 = $$lcssa180&65535;
      $224 = $222;
     }
    }
   } while(0);
   $223 = ((($183)) + 6|0);
   HEAP16[$223>>1] = $224;
   $225 = ((($144)) + 12|0);
   $226 = HEAP32[$225>>2]|0;
   $227 = ($226|0)==(0|0);
   if ($227) {
    $$01$i85 = 0;
   } else {
    $228 = HEAP8[$226>>0]|0;
    $229 = ($228<<24>>24)==(0);
    if ($229) {
     $$01$i85 = 0;
    } else {
     $230 = ($228<<24>>24)==(45);
     $231 = ((($226)) + 1|0);
     $$s$i75 = $230 ? $231 : $226;
     $232 = HEAP8[$$s$i75>>0]|0;
     $233 = ($232<<24>>24)==(48);
     L99: do {
      if ($233) {
       $234 = ((($$s$i75)) + 1|0);
       $235 = HEAP8[$234>>0]|0;
       switch ($235<<24>>24) {
       case 88: case 120:  {
        break;
       }
       default: {
        $$1$i78 = 10;$$13$i77 = $$s$i75;$238 = 48;$dmap$1$i79 = 531715;
        break L99;
       }
       }
       $236 = ((($$s$i75)) + 2|0);
       $$pre$i76 = HEAP8[$236>>0]|0;
       $$1$i78 = 16;$$13$i77 = $236;$238 = $$pre$i76;$dmap$1$i79 = 531683;
      } else {
       $$1$i78 = 10;$$13$i77 = $$s$i75;$238 = $232;$dmap$1$i79 = 531715;
      }
     } while(0);
     $237 = $238&255;
     $239 = $237 >>> 3;
     $240 = (($dmap$1$i79) + ($239)|0);
     $241 = HEAP8[$240>>0]|0;
     $242 = $241&255;
     $243 = $238 << 24 >> 24;
     $244 = $243 & 7;
     $245 = 1 << $244;
     $246 = $242 & $245;
     $247 = ($246|0)==(0);
     if ($247) {
      $$lcssa$i83 = 0;
     } else {
      $$21$i80 = $$13$i77;$249 = 0;$251 = $243;
      while(1) {
       $248 = Math_imul($249, $$1$i78)|0;
       $250 = (531747 + ($251)|0);
       $252 = HEAP8[$250>>0]|0;
       $253 = $252&255;
       $254 = (($253) + ($248))|0;
       $255 = ((($$21$i80)) + 1|0);
       $256 = HEAP8[$255>>0]|0;
       $257 = $256&255;
       $258 = $257 >>> 3;
       $259 = (($dmap$1$i79) + ($258)|0);
       $260 = HEAP8[$259>>0]|0;
       $261 = $260&255;
       $262 = $256 << 24 >> 24;
       $263 = $262 & 7;
       $264 = 1 << $263;
       $265 = $264 & $261;
       $266 = ($265|0)==(0);
       $sext$i81 = $254 << 16;
       $267 = $sext$i81 >> 16;
       if ($266) {
        $$lcssa$i83 = $267;
        break;
       } else {
        $$21$i80 = $255;$249 = $267;$251 = $262;
       }
      }
     }
     $268 = (0 - ($$lcssa$i83))|0;
     $269 = $230 ? $268 : $$lcssa$i83;
     $270 = $269&65535;
     $$01$i85 = $270;
    }
   }
   $271 = ((($183)) + 8|0);
   HEAP16[$271>>1] = $$01$i85;
   $272 = ((($144)) + 16|0);
   $273 = HEAP32[$272>>2]|0;
   $274 = ($273|0)==(0|0);
   if ($274) {
    $$01$i97 = 0;
   } else {
    $275 = HEAP8[$273>>0]|0;
    $276 = ($275<<24>>24)==(0);
    if ($276) {
     $$01$i97 = 0;
    } else {
     $277 = ($275<<24>>24)==(45);
     $278 = ((($273)) + 1|0);
     $$s$i87 = $277 ? $278 : $273;
     $279 = HEAP8[$$s$i87>>0]|0;
     $280 = ($279<<24>>24)==(48);
     L111: do {
      if ($280) {
       $281 = ((($$s$i87)) + 1|0);
       $282 = HEAP8[$281>>0]|0;
       switch ($282<<24>>24) {
       case 88: case 120:  {
        break;
       }
       default: {
        $$1$i90 = 10;$$13$i89 = $$s$i87;$285 = 48;$dmap$1$i91 = 531715;
        break L111;
       }
       }
       $283 = ((($$s$i87)) + 2|0);
       $$pre$i88 = HEAP8[$283>>0]|0;
       $$1$i90 = 16;$$13$i89 = $283;$285 = $$pre$i88;$dmap$1$i91 = 531683;
      } else {
       $$1$i90 = 10;$$13$i89 = $$s$i87;$285 = $279;$dmap$1$i91 = 531715;
      }
     } while(0);
     $284 = $285&255;
     $286 = $284 >>> 3;
     $287 = (($dmap$1$i91) + ($286)|0);
     $288 = HEAP8[$287>>0]|0;
     $289 = $288&255;
     $290 = $285 << 24 >> 24;
     $291 = $290 & 7;
     $292 = 1 << $291;
     $293 = $289 & $292;
     $294 = ($293|0)==(0);
     if ($294) {
      $$lcssa$i95 = 0;
     } else {
      $$21$i92 = $$13$i89;$296 = 0;$298 = $290;
      while(1) {
       $295 = Math_imul($296, $$1$i90)|0;
       $297 = (531747 + ($298)|0);
       $299 = HEAP8[$297>>0]|0;
       $300 = $299&255;
       $301 = (($300) + ($295))|0;
       $302 = ((($$21$i92)) + 1|0);
       $303 = HEAP8[$302>>0]|0;
       $304 = $303&255;
       $305 = $304 >>> 3;
       $306 = (($dmap$1$i91) + ($305)|0);
       $307 = HEAP8[$306>>0]|0;
       $308 = $307&255;
       $309 = $303 << 24 >> 24;
       $310 = $309 & 7;
       $311 = 1 << $310;
       $312 = $311 & $308;
       $313 = ($312|0)==(0);
       $sext$i93 = $301 << 16;
       $314 = $sext$i93 >> 16;
       if ($313) {
        $$lcssa$i95 = $314;
        break;
       } else {
        $$21$i92 = $302;$296 = $314;$298 = $309;
       }
      }
     }
     $315 = (0 - ($$lcssa$i95))|0;
     $316 = $277 ? $315 : $$lcssa$i95;
     $317 = $316&65535;
     $$01$i97 = $317;
    }
   }
   $318 = ((($183)) + 10|0);
   HEAP16[$318>>1] = $$01$i97;
   $319 = $224&65535;
   $320 = $$01$i97&65535;
   $321 = (($320) + ($319))|0;
   $322 = $321&65535;
   $323 = ((($183)) + 12|0);
   HEAP16[$323>>1] = $322;
   $324 = (0 - ($320))|0;
   $325 = $324&65535;
   $326 = ((($183)) + 14|0);
   HEAP16[$326>>1] = $325;
   $327 = HEAP32[$client_data>>2]|0;
   $328 = $327 | 8;
   HEAP32[$client_data>>2] = $328;
   $624 = 0;
   STACKTOP = sp;return ($624|0);
  }
 } while(0);
 $329 = (_strncmp($line,530770,4)|0);
 $330 = ($329|0)==(0);
 L122: do {
  if ($330) {
   $331 = ((($line)) + 4|0);
   $332 = HEAP8[$331>>0]|0;
   switch ($332<<24>>24) {
   case 9: case 13: case 10: case 0: case 32:  {
    break;
   }
   default: {
    break L122;
   }
   }
   $333 = ((($client_data)) + 139304|0);
   $334 = (__bdf_list_split($333,531680,$line,$linelen)|0);
   HEAP32[$error>>2] = $334;
   $335 = ($334|0)==(0);
   if (!($335)) {
    $624 = $334;
    STACKTOP = sp;return ($624|0);
   }
   $336 = ((($client_data)) + 139312|0);
   $337 = HEAP32[$336>>2]|0;
   $338 = ($337|0)==(0);
   do {
    if (!($338)) {
     $339 = ($337>>>0)<(2);
     if ($339) {
      HEAP32[$336>>2] = 0;
      break;
     }
     $340 = (($337) + -1)|0;
     $i$01$i = 0;$u$02$i = 1;
     while(1) {
      $341 = HEAP32[$333>>2]|0;
      $342 = (($341) + ($u$02$i<<2)|0);
      $343 = HEAP32[$342>>2]|0;
      $344 = (($341) + ($i$01$i<<2)|0);
      HEAP32[$344>>2] = $343;
      $345 = (($i$01$i) + 1)|0;
      $346 = (($u$02$i) + 1)|0;
      $exitcond = ($345|0)==($340|0);
      if ($exitcond) {
       break;
      } else {
       $i$01$i = $345;$u$02$i = $346;
      }
     }
     HEAP32[$336>>2] = $340;
     $347 = ($340|0)==(0);
     if (!($347)) {
      $348 = HEAP32[$333>>2]|0;
      $349 = HEAP32[$348>>2]|0;
      $351 = $349;$625 = $340;$i$03$i = 0;$j$04$i = 0;
      while(1) {
       $350 = HEAP8[$351>>0]|0;
       $352 = ($350<<24>>24)==(0);
       if ($352) {
        $361 = $625;$j$1$lcssa$i = $j$04$i;
       } else {
        $356 = $350;$fp$02$i = $351;$j$11$i = $j$04$i;
        while(1) {
         $353 = ((($fp$02$i)) + 1|0);
         $354 = (($j$11$i) + 1)|0;
         $355 = (($349) + ($j$11$i)|0);
         HEAP8[$355>>0] = $356;
         $357 = HEAP8[$353>>0]|0;
         $358 = ($357<<24>>24)==(0);
         if ($358) {
          $$lcssa177 = $354;
          break;
         } else {
          $356 = $357;$fp$02$i = $353;$j$11$i = $354;
         }
        }
        $$pre10$i = HEAP32[$336>>2]|0;
        $361 = $$pre10$i;$j$1$lcssa$i = $$lcssa177;
       }
       $359 = (($i$03$i) + 1)|0;
       $360 = ($359>>>0)<($361>>>0);
       if ($360) {
        $362 = (($j$1$lcssa$i) + 1)|0;
        $363 = (($349) + ($j$1$lcssa$i)|0);
        HEAP8[$363>>0] = 32;
        $$pre11$i = HEAP32[$336>>2]|0;
        $365 = $$pre11$i;$j$0$be$i = $362;
       } else {
        $365 = $361;$j$0$be$i = $j$1$lcssa$i;
       }
       $364 = ($359>>>0)<($365>>>0);
       if (!($364)) {
        $j$0$be$i$lcssa = $j$0$be$i;
        break;
       }
       $$pre$i103 = HEAP32[$333>>2]|0;
       $$phi$trans$insert$i = (($$pre$i103) + ($359<<2)|0);
       $$pre9$i = HEAP32[$$phi$trans$insert$i>>2]|0;
       $351 = $$pre9$i;$625 = $365;$i$03$i = $359;$j$04$i = $j$0$be$i;
      }
      $366 = ($349|0)==(1282466|0);
      if ($366) {
       $$0$i112 = 1282466;
      } else {
       $367 = (($349) + ($j$0$be$i$lcssa)|0);
       HEAP8[$367>>0] = 0;
       $368 = ($349|0)==(0|0);
       if ($368) {
        break;
       } else {
        $$0$i112 = $349;
       }
      }
      $369 = HEAP32[$0>>2]|0;
      $370 = HEAP32[$369>>2]|0;
      _ft_mem_free($memory$0,$370);
      $371 = HEAP32[$0>>2]|0;
      HEAP32[$371>>2] = 0;
      $372 = (($j$0$be$i$lcssa) + 1)|0;
      $373 = (_ft_mem_realloc($memory$0,1,0,$372,0,$error)|0);
      $374 = HEAP32[$0>>2]|0;
      HEAP32[$374>>2] = $373;
      $375 = HEAP32[$error>>2]|0;
      $376 = ($375|0)==(0);
      if (!($376)) {
       $624 = $375;
       STACKTOP = sp;return ($624|0);
      }
      _memcpy(($373|0),($$0$i112|0),($372|0))|0;
      $377 = HEAP32[$0>>2]|0;
      $378 = ((($client_data)) + 36|0);
      $379 = HEAP32[$378>>2]|0;
      $380 = ($377|0)==(0|0);
      if (!($380)) {
       $381 = HEAP32[$377>>2]|0;
       $382 = ($381|0)==(0|0);
       if (!($382)) {
        $383 = HEAP8[$381>>0]|0;
        $384 = ($383<<24>>24)==(0);
        if (!($384)) {
         $385 = ((($377)) + 278664|0);
         $386 = HEAP32[$385>>2]|0;
         ;HEAP32[$list$i>>2]=0|0;HEAP32[$list$i+4>>2]=0|0;HEAP32[$list$i+8>>2]=0|0;
         $387 = ((($list$i)) + 12|0);
         HEAP32[$387>>2] = $386;
         $388 = ((($379)) + 12|0);
         $389 = HEAP32[$388>>2]|0;
         $390 = ((($377)) + 28|0);
         HEAP32[$390>>2] = $389;
         $391 = (_strlen($381)|0);
         $392 = (($391) + 1)|0;
         $393 = ($392>>>0)>(255);
         if (!($393)) {
          _memcpy(($name$i|0),($381|0),($392|0))|0;
          $394 = (__bdf_list_split($list$i,532005,$name$i,$392)|0);
          $395 = ($394|0)==(0);
          L160: do {
           if ($395) {
            $396 = ((($list$i)) + 8|0);
            $397 = HEAP32[$396>>2]|0;
            $398 = ($397|0)==(15);
            if (!($398)) {
             break;
            }
            $399 = HEAP32[$list$i>>2]|0;
            $400 = ((($399)) + 44|0);
            $401 = HEAP32[$400>>2]|0;
            $402 = HEAP8[$401>>0]|0;
            $403 = $402 << 24 >> 24;
            switch ($403|0) {
            case 99: case 67:  {
             HEAP32[$390>>2] = 32;
             break L160;
             break;
            }
            case 109: case 77:  {
             HEAP32[$390>>2] = 16;
             break L160;
             break;
            }
            case 112: case 80:  {
             HEAP32[$390>>2] = 8;
             break L160;
             break;
            }
            default: {
             break L160;
            }
            }
           }
          } while(0);
          $404 = HEAP32[$387>>2]|0;
          $405 = ($404|0)==(0|0);
          if (!($405)) {
           $406 = HEAP32[$list$i>>2]|0;
           _ft_mem_free($404,$406);
           ;HEAP32[$list$i>>2]=0|0;HEAP32[$list$i+4>>2]=0|0;HEAP32[$list$i+8>>2]=0|0;HEAP32[$list$i+12>>2]=0|0;
          }
          HEAP32[$error>>2] = $394;
          if (!($395)) {
           $624 = $394;
           STACKTOP = sp;return ($624|0);
          }
          $407 = HEAP32[$client_data>>2]|0;
          $408 = $407 | 2;
          HEAP32[$client_data>>2] = $408;
          $624 = 0;
          STACKTOP = sp;return ($624|0);
         }
        }
       }
      }
      HEAP32[$error>>2] = 6;
      $624 = 6;
      STACKTOP = sp;return ($624|0);
     }
    }
   } while(0);
   HEAP32[$error>>2] = 3;
   $624 = 3;
   STACKTOP = sp;return ($624|0);
  }
 } while(0);
 $409 = (_strncmp($line,532007,4)|0);
 $410 = ($409|0)==(0);
 L180: do {
  if ($410) {
   $411 = ((($line)) + 4|0);
   $412 = HEAP8[$411>>0]|0;
   switch ($412<<24>>24) {
   case 9: case 13: case 10: case 0: case 32:  {
    break;
   }
   default: {
    break L180;
   }
   }
   $413 = $33 & 2;
   $414 = ($413|0)==(0);
   if ($414) {
    HEAP32[$error>>2] = 177;
    $624 = 177;
    STACKTOP = sp;return ($624|0);
   }
   $415 = ((($client_data)) + 139304|0);
   $416 = (__bdf_list_split($415,531680,$line,$linelen)|0);
   HEAP32[$error>>2] = $416;
   $417 = ($416|0)==(0);
   if (!($417)) {
    $624 = $416;
    STACKTOP = sp;return ($624|0);
   }
   $418 = HEAP32[$415>>2]|0;
   $419 = ((($418)) + 4|0);
   $420 = HEAP32[$419>>2]|0;
   $421 = ($420|0)==(0|0);
   L190: do {
    if ($421) {
     $$01$i73 = 0;
    } else {
     $422 = HEAP8[$420>>0]|0;
     L192: do {
      switch ($422<<24>>24) {
      case 0:  {
       $$01$i73 = 0;
       break L190;
       break;
      }
      case 48:  {
       $423 = ((($420)) + 1|0);
       $424 = HEAP8[$423>>0]|0;
       switch ($424<<24>>24) {
       case 88: case 120:  {
        break;
       }
       default: {
        $$02$i67 = $420;$$1$i68 = 10;$427 = 48;$dmap$1$i69 = 531715;
        break L192;
       }
       }
       $425 = ((($420)) + 2|0);
       $$pre$i66 = HEAP8[$425>>0]|0;
       $$02$i67 = $425;$$1$i68 = 16;$427 = $$pre$i66;$dmap$1$i69 = 531683;
       break;
      }
      default: {
       $$02$i67 = $420;$$1$i68 = 10;$427 = $422;$dmap$1$i69 = 531715;
      }
      }
     } while(0);
     $426 = $427&255;
     $428 = $426 >>> 3;
     $429 = (($dmap$1$i69) + ($428)|0);
     $430 = HEAP8[$429>>0]|0;
     $431 = $430&255;
     $432 = $427 << 24 >> 24;
     $433 = $432 & 7;
     $434 = 1 << $433;
     $435 = $431 & $434;
     $436 = ($435|0)==(0);
     if ($436) {
      $$01$i73 = 0;
     } else {
      $$131$i71 = $$02$i67;$439 = $432;$v$02$i70 = 0;
      while(1) {
       $437 = Math_imul($v$02$i70, $$1$i68)|0;
       $438 = (531747 + ($439)|0);
       $440 = HEAP8[$438>>0]|0;
       $441 = $440&255;
       $442 = (($441) + ($437))|0;
       $443 = ((($$131$i71)) + 1|0);
       $444 = HEAP8[$443>>0]|0;
       $445 = $444&255;
       $446 = $445 >>> 3;
       $447 = (($dmap$1$i69) + ($446)|0);
       $448 = HEAP8[$447>>0]|0;
       $449 = $448&255;
       $450 = $444 << 24 >> 24;
       $451 = $450 & 7;
       $452 = 1 << $451;
       $453 = $452 & $449;
       $454 = ($453|0)==(0);
       if ($454) {
        $$01$i73 = $442;
        break;
       } else {
        $$131$i71 = $443;$439 = $450;$v$02$i70 = $442;
       }
      }
     }
    }
   } while(0);
   $455 = HEAP32[$0>>2]|0;
   $456 = ((($455)) + 16|0);
   HEAP32[$456>>2] = $$01$i73;
   $457 = ((($418)) + 8|0);
   $458 = HEAP32[$457>>2]|0;
   $459 = ($458|0)==(0|0);
   L199: do {
    if ($459) {
     $$01$i64 = 0;
    } else {
     $460 = HEAP8[$458>>0]|0;
     L201: do {
      switch ($460<<24>>24) {
      case 0:  {
       $$01$i64 = 0;
       break L199;
       break;
      }
      case 48:  {
       $461 = ((($458)) + 1|0);
       $462 = HEAP8[$461>>0]|0;
       switch ($462<<24>>24) {
       case 88: case 120:  {
        break;
       }
       default: {
        $$02$i58 = $458;$$1$i59 = 10;$465 = 48;$dmap$1$i60 = 531715;
        break L201;
       }
       }
       $463 = ((($458)) + 2|0);
       $$pre$i57 = HEAP8[$463>>0]|0;
       $$02$i58 = $463;$$1$i59 = 16;$465 = $$pre$i57;$dmap$1$i60 = 531683;
       break;
      }
      default: {
       $$02$i58 = $458;$$1$i59 = 10;$465 = $460;$dmap$1$i60 = 531715;
      }
      }
     } while(0);
     $464 = $465&255;
     $466 = $464 >>> 3;
     $467 = (($dmap$1$i60) + ($466)|0);
     $468 = HEAP8[$467>>0]|0;
     $469 = $468&255;
     $470 = $465 << 24 >> 24;
     $471 = $470 & 7;
     $472 = 1 << $471;
     $473 = $469 & $472;
     $474 = ($473|0)==(0);
     if ($474) {
      $$01$i64 = 0;
     } else {
      $$131$i62 = $$02$i58;$477 = $470;$v$02$i61 = 0;
      while(1) {
       $475 = Math_imul($v$02$i61, $$1$i59)|0;
       $476 = (531747 + ($477)|0);
       $478 = HEAP8[$476>>0]|0;
       $479 = $478&255;
       $480 = (($479) + ($475))|0;
       $481 = ((($$131$i62)) + 1|0);
       $482 = HEAP8[$481>>0]|0;
       $483 = $482&255;
       $484 = $483 >>> 3;
       $485 = (($dmap$1$i60) + ($484)|0);
       $486 = HEAP8[$485>>0]|0;
       $487 = $486&255;
       $488 = $482 << 24 >> 24;
       $489 = $488 & 7;
       $490 = 1 << $489;
       $491 = $490 & $487;
       $492 = ($491|0)==(0);
       if ($492) {
        $$01$i64 = $480;
        break;
       } else {
        $$131$i62 = $481;$477 = $488;$v$02$i61 = $480;
       }
      }
     }
    }
   } while(0);
   $493 = ((($455)) + 20|0);
   HEAP32[$493>>2] = $$01$i64;
   $494 = ((($418)) + 12|0);
   $495 = HEAP32[$494>>2]|0;
   $496 = ($495|0)==(0|0);
   L208: do {
    if ($496) {
     $$01$i55 = 0;
    } else {
     $497 = HEAP8[$495>>0]|0;
     L210: do {
      switch ($497<<24>>24) {
      case 0:  {
       $$01$i55 = 0;
       break L208;
       break;
      }
      case 48:  {
       $498 = ((($495)) + 1|0);
       $499 = HEAP8[$498>>0]|0;
       switch ($499<<24>>24) {
       case 88: case 120:  {
        break;
       }
       default: {
        $$02$i49 = $495;$$1$i50 = 10;$502 = 48;$dmap$1$i51 = 531715;
        break L210;
       }
       }
       $500 = ((($495)) + 2|0);
       $$pre$i48 = HEAP8[$500>>0]|0;
       $$02$i49 = $500;$$1$i50 = 16;$502 = $$pre$i48;$dmap$1$i51 = 531683;
       break;
      }
      default: {
       $$02$i49 = $495;$$1$i50 = 10;$502 = $497;$dmap$1$i51 = 531715;
      }
      }
     } while(0);
     $501 = $502&255;
     $503 = $501 >>> 3;
     $504 = (($dmap$1$i51) + ($503)|0);
     $505 = HEAP8[$504>>0]|0;
     $506 = $505&255;
     $507 = $502 << 24 >> 24;
     $508 = $507 & 7;
     $509 = 1 << $508;
     $510 = $506 & $509;
     $511 = ($510|0)==(0);
     if ($511) {
      $$01$i55 = 0;
     } else {
      $$131$i53 = $$02$i49;$514 = $507;$v$02$i52 = 0;
      while(1) {
       $512 = Math_imul($v$02$i52, $$1$i50)|0;
       $513 = (531747 + ($514)|0);
       $515 = HEAP8[$513>>0]|0;
       $516 = $515&255;
       $517 = (($516) + ($512))|0;
       $518 = ((($$131$i53)) + 1|0);
       $519 = HEAP8[$518>>0]|0;
       $520 = $519&255;
       $521 = $520 >>> 3;
       $522 = (($dmap$1$i51) + ($521)|0);
       $523 = HEAP8[$522>>0]|0;
       $524 = $523&255;
       $525 = $519 << 24 >> 24;
       $526 = $525 & 7;
       $527 = 1 << $526;
       $528 = $527 & $524;
       $529 = ($528|0)==(0);
       if ($529) {
        $$01$i55 = $517;
        break;
       } else {
        $$131$i53 = $518;$514 = $525;$v$02$i52 = $517;
       }
      }
     }
    }
   } while(0);
   $530 = ((($455)) + 24|0);
   HEAP32[$530>>2] = $$01$i55;
   $531 = ((($client_data)) + 139312|0);
   $532 = HEAP32[$531>>2]|0;
   $533 = ($532|0)==(5);
   if ($533) {
    $534 = ((($418)) + 16|0);
    $535 = HEAP32[$534>>2]|0;
    $536 = ($535|0)==(0|0);
    do {
     if ($536) {
      label = 137;
     } else {
      $537 = HEAP8[$535>>0]|0;
      $538 = ($537<<24>>24)==(0);
      if ($538) {
       label = 137;
      } else {
       $539 = ($537<<24>>24)==(45);
       $540 = ((($535)) + 1|0);
       $$s$i = $539 ? $540 : $535;
       $541 = HEAP8[$$s$i>>0]|0;
       $542 = ($541<<24>>24)==(48);
       L222: do {
        if ($542) {
         $543 = ((($$s$i)) + 1|0);
         $544 = HEAP8[$543>>0]|0;
         switch ($544<<24>>24) {
         case 88: case 120:  {
          break;
         }
         default: {
          $$1$i44 = 10;$$13$i = $$s$i;$547 = 48;$dmap$1$i45 = 531715;
          break L222;
         }
         }
         $545 = ((($$s$i)) + 2|0);
         $$pre$i43 = HEAP8[$545>>0]|0;
         $$1$i44 = 16;$$13$i = $545;$547 = $$pre$i43;$dmap$1$i45 = 531683;
        } else {
         $$1$i44 = 10;$$13$i = $$s$i;$547 = $541;$dmap$1$i45 = 531715;
        }
       } while(0);
       $546 = $547&255;
       $548 = $546 >>> 3;
       $549 = (($dmap$1$i45) + ($548)|0);
       $550 = HEAP8[$549>>0]|0;
       $551 = $550&255;
       $552 = $547 << 24 >> 24;
       $553 = $552 & 7;
       $554 = 1 << $553;
       $555 = $551 & $554;
       $556 = ($555|0)==(0);
       if ($556) {
        $$lcssa$i = 0;
       } else {
        $$21$i = $$13$i;$558 = 0;$560 = $552;
        while(1) {
         $557 = Math_imul($558, $$1$i44)|0;
         $559 = (531747 + ($560)|0);
         $561 = HEAP8[$559>>0]|0;
         $562 = $561&255;
         $563 = (($562) + ($557))|0;
         $564 = ((($$21$i)) + 1|0);
         $565 = HEAP8[$564>>0]|0;
         $566 = $565&255;
         $567 = $566 >>> 3;
         $568 = (($dmap$1$i45) + ($567)|0);
         $569 = HEAP8[$568>>0]|0;
         $570 = $569&255;
         $571 = $565 << 24 >> 24;
         $572 = $571 & 7;
         $573 = 1 << $572;
         $574 = $573 & $570;
         $575 = ($574|0)==(0);
         $sext$i = $563 << 16;
         $576 = $sext$i >> 16;
         if ($575) {
          $$lcssa$i = $576;
          break;
         } else {
          $$21$i = $564;$558 = $576;$560 = $571;
         }
        }
       }
       $578 = (0 - ($$lcssa$i))|0;
       $579 = $539 ? $578 : $$lcssa$i;
       $580 = $579&65535;
       $581 = ((($455)) + 278662|0);
       HEAP16[$581>>1] = $580;
       $582 = ($580<<16>>16)==(0);
       if ($582) {
        $626 = 0;$627 = $581;$bitcount$0$lcssa138 = 0;
        label = 141;
        break;
       } else {
        $bitcount$0125 = 0;$i1$0126 = 0;$shift$0127 = $580;
       }
       while(1) {
        $583 = $shift$0127 & 1;
        $584 = ($583<<16>>16)==(0);
        $bitcount$0$i1$0 = $584 ? $bitcount$0125 : $i1$0126;
        $585 = ($shift$0127&65535) >>> 1;
        $586 = (($i1$0126) + 1)<<16>>16;
        $587 = ($585<<16>>16)==(0);
        if ($587) {
         $bitcount$0$i1$0$lcssa = $bitcount$0$i1$0;
         break;
        } else {
         $bitcount$0125 = $bitcount$0$i1$0;$i1$0126 = $586;$shift$0127 = $585;
        }
       }
       $588 = ($bitcount$0$i1$0$lcssa&65535)>(3);
       if ($588) {
        $592 = $579;$594 = 8;$597 = $581;
       } else {
        $626 = $579;$627 = $581;$bitcount$0$lcssa138 = $bitcount$0$i1$0$lcssa;
        label = 141;
       }
      }
     }
    } while(0);
    if ((label|0) == 137) {
     $577 = ((($455)) + 278662|0);
     HEAP16[$577>>1] = 0;
     $626 = 0;$627 = $577;$bitcount$0$lcssa138 = 0;
     label = 141;
    }
    if ((label|0) == 141) {
     $589 = $bitcount$0$lcssa138&65535;
     $590 = 1 << $589;
     $phitmp5 = $590 & 65535;
     $592 = $626;$594 = $phitmp5;$597 = $627;
    }
    $591 = $592 & 65535;
    $593 = ($591|0)==($594|0);
    if (!($593)) {
     $595 = $594 << 1;
     $596 = $595&65535;
     HEAP16[$597>>1] = $596;
    }
   } else {
    $598 = ((($455)) + 278662|0);
    HEAP16[$598>>1] = 1;
   }
   $599 = HEAP32[$client_data>>2]|0;
   $600 = $599 | 4;
   HEAP32[$client_data>>2] = $600;
   $624 = 0;
   STACKTOP = sp;return ($624|0);
  }
 } while(0);
 $601 = (_strncmp($line,531893,5)|0);
 $602 = ($601|0)==(0);
 L243: do {
  if ($602) {
   $603 = ((($line)) + 5|0);
   $604 = HEAP8[$603>>0]|0;
   switch ($604<<24>>24) {
   case 9: case 13: case 10: case 0: case 32:  {
    break;
   }
   default: {
    break L243;
   }
   }
   $605 = $33 & 8;
   $606 = ($605|0)==(0);
   if ($606) {
    HEAP32[$error>>2] = 179;
    $628 = 179;
   } else {
    $607 = ((($1)) + 12|0);
    $608 = HEAP16[$607>>1]|0;
    $609 = $608 << 16 >> 16;
    $610 = ((($1)) + 40|0);
    HEAP32[$610>>2] = $609;
    HEAP32[$vararg_buffer>>2] = $609;
    (_sprintf($nbuf,531889,$vararg_buffer)|0);
    $611 = HEAP32[$0>>2]|0;
    $612 = (__bdf_add_property($611,530793,$nbuf)|0);
    HEAP32[$error>>2] = $612;
    $613 = ($612|0)==(0);
    if ($613) {
     $614 = HEAP32[$0>>2]|0;
     $615 = ((($614)) + 14|0);
     $616 = HEAP16[$615>>1]|0;
     $617 = $616 << 16 >> 16;
     $618 = ((($614)) + 44|0);
     HEAP32[$618>>2] = $617;
     HEAP32[$vararg_buffer1>>2] = $617;
     (_sprintf($nbuf,531889,$vararg_buffer1)|0);
     $619 = HEAP32[$0>>2]|0;
     $620 = (__bdf_add_property($619,530805,$nbuf)|0);
     HEAP32[$error>>2] = $620;
     $621 = ($620|0)==(0);
     if ($621) {
      $622 = HEAP32[$0>>2]|0;
      $623 = ((($622)) + 278660|0);
      HEAP16[$623>>1] = 1;
      HEAP32[$call_data>>2] = 4354;
      HEAP32[$error>>2] = -1;
      $628 = -1;
     } else {
      $628 = $620;
     }
    } else {
     $628 = $612;
    }
   }
   $624 = $628;
   STACKTOP = sp;return ($624|0);
  }
 } while(0);
 HEAP32[$error>>2] = 3;
 $624 = 3;
 STACKTOP = sp;return ($624|0);
}
function _hash_insert($key,$data,$ht,$memory) {
 $key = $key|0;
 $data = $data|0;
 $ht = $ht|0;
 $memory = $memory|0;
 var $$$i = 0, $$$i$i = 0, $$lcssa14 = 0, $$pre = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $9 = 0, $bp$01$i = 0, $error = 0, $error$i = 0, $exitcond$i = 0, $i$03$i = 0, $kp$05$i = 0, $kp$05$i$i = 0, $ndp$0$lcssa$i$i = 0, $ndp$0$lcssa$i$ph = 0, $ndp$01$i = 0, $ndp$01$i$i = 0;
 var $res$0$lcssa$i = 0, $res$0$lcssa$i$i = 0, $res$04$i = 0, $res$04$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp + 4|0;
 $error = sp;
 $0 = ((($ht)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP8[$key>>0]|0;
 $3 = ($2<<24>>24)==(0);
 if ($3) {
  $res$0$lcssa$i = 0;
 } else {
  $7 = $2;$kp$05$i = $key;$res$04$i = 0;
  while(1) {
   $4 = ($res$04$i*31)|0;
   $5 = ((($kp$05$i)) + 1|0);
   $6 = $7 << 24 >> 24;
   $8 = (($4) + ($6))|0;
   $9 = HEAP8[$5>>0]|0;
   $10 = ($9<<24>>24)==(0);
   if ($10) {
    $res$0$lcssa$i = $8;
    break;
   } else {
    $7 = $9;$kp$05$i = $5;$res$04$i = $8;
   }
  }
 }
 $11 = ((($ht)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($res$0$lcssa$i>>>0) % ($12>>>0))&-1;
 $14 = (($1) + ($13<<2)|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)==(0|0);
 L5: do {
  if ($16) {
   $ndp$0$lcssa$i$ph = $14;
  } else {
   $17 = (($12) + -1)|0;
   $18 = (($1) + ($17<<2)|0);
   $20 = $15;$ndp$01$i = $14;
   while(1) {
    $19 = HEAP32[$20>>2]|0;
    $21 = HEAP8[$19>>0]|0;
    $22 = ($21<<24>>24)==($2<<24>>24);
    if ($22) {
     $23 = (_strcmp($19,$key)|0);
     $24 = ($23|0)==(0);
     if ($24) {
      $$lcssa14 = $20;
      break;
     }
    }
    $25 = ((($ndp$01$i)) + -4|0);
    $26 = ($25>>>0)<($1>>>0);
    $$$i = $26 ? $18 : $25;
    $27 = HEAP32[$$$i>>2]|0;
    $28 = ($27|0)==(0|0);
    if ($28) {
     $ndp$0$lcssa$i$ph = $$$i;
     break L5;
    } else {
     $20 = $27;$ndp$01$i = $$$i;
    }
   }
   HEAP32[$error>>2] = 0;
   $81 = ((($$lcssa14)) + 4|0);
   HEAP32[$81>>2] = $data;
   $83 = $error;
   $82 = HEAP32[$error>>2]|0;
   STACKTOP = sp;return ($82|0);
  }
 } while(0);
 HEAP32[$error>>2] = 0;
 $29 = (_ft_mem_alloc($memory,8,$error)|0);
 $30 = HEAP32[$error>>2]|0;
 $31 = ($30|0)==(0);
 if (!($31)) {
  $83 = $error;
  $82 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($82|0);
 }
 HEAP32[$ndp$0$lcssa$i$ph>>2] = $29;
 HEAP32[$29>>2] = $key;
 $32 = ((($29)) + 4|0);
 HEAP32[$32>>2] = $data;
 $33 = ((($ht)) + 8|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = HEAP32[$ht>>2]|0;
 $36 = ($34>>>0)<($35>>>0);
 do {
  if ($36) {
   $80 = $34;
  } else {
   $37 = HEAP32[$0>>2]|0;
   $38 = HEAP32[$11>>2]|0;
   HEAP32[$error$i>>2] = 0;
   $39 = $38 << 1;
   HEAP32[$11>>2] = $39;
   $40 = (($39>>>0) / 3)&-1;
   HEAP32[$ht>>2] = $40;
   $41 = (_ft_mem_realloc($memory,4,0,$39,0,$error$i)|0);
   HEAP32[$0>>2] = $41;
   $42 = HEAP32[$error$i>>2]|0;
   $43 = ($42|0)==(0);
   if (!($43)) {
    HEAP32[$error>>2] = $42;
    $83 = $error;
    $82 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($82|0);
   }
   $44 = ($38|0)==(0);
   if (!($44)) {
    $bp$01$i = $37;$i$03$i = 0;
    while(1) {
     $45 = HEAP32[$bp$01$i>>2]|0;
     $46 = ($45|0)==(0|0);
     $47 = $45;
     if (!($46)) {
      $48 = HEAP32[$45>>2]|0;
      $49 = HEAP32[$0>>2]|0;
      $50 = HEAP8[$48>>0]|0;
      $51 = ($50<<24>>24)==(0);
      if ($51) {
       $res$0$lcssa$i$i = 0;
      } else {
       $55 = $50;$kp$05$i$i = $48;$res$04$i$i = 0;
       while(1) {
        $52 = ($res$04$i$i*31)|0;
        $53 = ((($kp$05$i$i)) + 1|0);
        $54 = $55 << 24 >> 24;
        $56 = (($52) + ($54))|0;
        $57 = HEAP8[$53>>0]|0;
        $58 = ($57<<24>>24)==(0);
        if ($58) {
         $res$0$lcssa$i$i = $56;
         break;
        } else {
         $55 = $57;$kp$05$i$i = $53;$res$04$i$i = $56;
        }
       }
      }
      $59 = HEAP32[$11>>2]|0;
      $60 = (($res$0$lcssa$i$i>>>0) % ($59>>>0))&-1;
      $61 = (($49) + ($60<<2)|0);
      $62 = HEAP32[$61>>2]|0;
      $63 = ($62|0)==(0|0);
      L33: do {
       if ($63) {
        $ndp$0$lcssa$i$i = $61;
       } else {
        $64 = (($59) + -1)|0;
        $65 = (($49) + ($64<<2)|0);
        $67 = $62;$ndp$01$i$i = $61;
        while(1) {
         $66 = HEAP32[$67>>2]|0;
         $68 = HEAP8[$66>>0]|0;
         $69 = ($68<<24>>24)==($50<<24>>24);
         if ($69) {
          $70 = (_strcmp($66,$48)|0);
          $71 = ($70|0)==(0);
          if ($71) {
           $ndp$0$lcssa$i$i = $ndp$01$i$i;
           break L33;
          }
         }
         $72 = ((($ndp$01$i$i)) + -4|0);
         $73 = ($72>>>0)<($49>>>0);
         $$$i$i = $73 ? $65 : $72;
         $74 = HEAP32[$$$i$i>>2]|0;
         $75 = ($74|0)==(0|0);
         if ($75) {
          $ndp$0$lcssa$i$i = $$$i$i;
          break;
         } else {
          $67 = $74;$ndp$01$i$i = $$$i$i;
         }
        }
       }
      } while(0);
      HEAP32[$ndp$0$lcssa$i$i>>2] = $47;
     }
     $76 = (($i$03$i) + 1)|0;
     $77 = ((($bp$01$i)) + 4|0);
     $exitcond$i = ($76|0)==($38|0);
     if ($exitcond$i) {
      break;
     } else {
      $bp$01$i = $77;$i$03$i = $76;
     }
    }
   }
   _ft_mem_free($memory,$37);
   $$pre$i = HEAP32[$error$i>>2]|0;
   HEAP32[$error>>2] = $$pre$i;
   $78 = ($$pre$i|0)==(0);
   if ($78) {
    $$pre = HEAP32[$33>>2]|0;
    $80 = $$pre;
    break;
   } else {
    $83 = $error;
    $82 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($82|0);
   }
  }
 } while(0);
 $79 = (($80) + 1)|0;
 HEAP32[$33>>2] = $79;
 $83 = $error;
 $82 = HEAP32[$error>>2]|0;
 STACKTOP = sp;return ($82|0);
}
function __bdf_list_split($list,$separators,$line,$linelen) {
 $list = $list|0;
 $separators = $separators|0;
 $line = $line|0;
 $linelen = $linelen|0;
 var $$ = 0, $$pre = 0, $$pre$phi49Z2D = 0, $$pre$phiZ2D = 0, $$pre42 = 0, $$pre42$pre = 0, $$pre43 = 0, $$pre43$pre = 0, $$pre44 = 0, $$pre48$pre$phiZZ2D = 0, $$pre50 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $ep$1$lcssa$ph = 0, $ep$1$lcssa$us$ph = 0, $ep$116 = 0;
 var $ep$116$us = 0, $ep$2$lcssa = 0, $ep$2$lcssa51 = 0, $ep$219 = 0, $ep$39$us = 0, $error$3 = 0, $error$i = 0, $error$i4 = 0, $final_empty$0$lcssa = 0, $final_empty$025 = 0, $final_empty$025$us = 0, $mult$030 = 0, $mult$030$lcssa = 0, $mult$1 = 0, $newsize$0$i = 0, $newsize$0$i$us = 0, $newsize$0$i6 = 0, $or$cond$i = 0, $or$cond$i$us = 0, $or$cond$i5 = 0;
 var $seps = 0, $sp$031 = 0, $sp$124 = 0, $sp$124$us = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i4 = sp + 4|0;
 $error$i = sp;
 $seps = sp + 8|0;
 $0 = ((($list)) + 8|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($list)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0);
 if (!($3)) {
  $4 = HEAP32[$list>>2]|0;
  HEAP32[$4>>2] = 1282466;
  $5 = HEAP32[$list>>2]|0;
  $6 = ((($5)) + 4|0);
  HEAP32[$6>>2] = 1282466;
  $7 = HEAP32[$list>>2]|0;
  $8 = ((($7)) + 8|0);
  HEAP32[$8>>2] = 1282466;
  $9 = HEAP32[$list>>2]|0;
  $10 = ((($9)) + 12|0);
  HEAP32[$10>>2] = 1282466;
  $11 = HEAP32[$list>>2]|0;
  $12 = ((($11)) + 16|0);
  HEAP32[$12>>2] = 1282466;
 }
 $13 = ($linelen|0)==(0);
 if ($13) {
  $error$3 = 0;
  STACKTOP = sp;return ($error$3|0);
 }
 $14 = HEAP8[$line>>0]|0;
 $15 = ($14<<24>>24)==(0);
 if ($15) {
  $error$3 = 0;
  STACKTOP = sp;return ($error$3|0);
 }
 $16 = ($separators|0)==(0|0);
 if ($16) {
  $error$3 = 6;
  STACKTOP = sp;return ($error$3|0);
 }
 $17 = HEAP8[$separators>>0]|0;
 $18 = ($17<<24>>24)==(0);
 if ($18) {
  $error$3 = 6;
  STACKTOP = sp;return ($error$3|0);
 }
 dest=$seps; stop=dest+32|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0));
 $19 = $17;$mult$030 = 0;$sp$031 = $separators;
 L16: while(1) {
  switch ($19<<24>>24) {
  case 0:  {
   $mult$030$lcssa = $mult$030;
   break L16;
   break;
  }
  case 43:  {
   $20 = ((($sp$031)) + 1|0);
   $21 = HEAP8[$20>>0]|0;
   $22 = ($21<<24>>24)==(0);
   if ($22) {
    $$pre$phi49Z2D = $20;$mult$1 = 1;
   } else {
    $$pre48$pre$phiZZ2D = $20;
    label = 11;
   }
   break;
  }
  default: {
   $$pre50 = ((($sp$031)) + 1|0);
   $$pre48$pre$phiZZ2D = $$pre50;
   label = 11;
  }
  }
  if ((label|0) == 11) {
   label = 0;
   $23 = $19&255;
   $24 = $23 & 7;
   $25 = 1 << $24;
   $26 = $23 >>> 3;
   $27 = (($seps) + ($26)|0);
   $28 = HEAP8[$27>>0]|0;
   $29 = $28&255;
   $30 = $25 | $29;
   $31 = $30&255;
   HEAP8[$27>>0] = $31;
   $$pre$phi49Z2D = $$pre48$pre$phiZZ2D;$mult$1 = $mult$030;
  }
  $$pre = HEAP8[$$pre$phi49Z2D>>0]|0;
  $19 = $$pre;$mult$030 = $mult$1;$sp$031 = $$pre$phi49Z2D;
 }
 $32 = (($line) + ($linelen)|0);
 $33 = ($linelen|0)>(0);
 L25: do {
  if ($33) {
   $34 = ((($list)) + 12|0);
   $35 = ($mult$030$lcssa|0)==(0);
   L27: do {
    if ($35) {
     $37 = $14;$final_empty$025$us = 0;$sp$124$us = $line;
     while(1) {
      $36 = ($37<<24>>24)==(0);
      if ($36) {
       $final_empty$0$lcssa = $final_empty$025$us;
       break L25;
      } else {
       $39 = $37;$ep$116$us = $sp$124$us;
      }
      while(1) {
       $38 = $39&255;
       $40 = $38 >>> 3;
       $41 = (($seps) + ($40)|0);
       $42 = HEAP8[$41>>0]|0;
       $43 = $42 << 24 >> 24;
       $44 = $38 & 7;
       $45 = 1 << $44;
       $46 = $43 & $45;
       $47 = ($46|0)==(0);
       if (!($47)) {
        $ep$1$lcssa$us$ph = $ep$116$us;
        break;
       }
       $48 = ((($ep$116$us)) + 1|0);
       $49 = HEAP8[$48>>0]|0;
       $50 = ($49<<24>>24)==(0);
       if ($50) {
        $ep$1$lcssa$us$ph = $48;
        break;
       } else {
        $39 = $49;$ep$116$us = $48;
       }
      }
      $51 = HEAP32[$0>>2]|0;
      $52 = HEAP32[$1>>2]|0;
      $53 = ($51|0)==($52|0);
      if ($53) {
       HEAP32[$error$i>>2] = 0;
       $54 = ($51|0)==(-1);
       if ($54) {
        $$pre43 = -1;
       } else {
        $55 = $51 >>> 1;
        $56 = (($51) + 5)|0;
        $57 = (($56) + ($55))|0;
        $58 = ($51|0)==(536870911);
        if ($58) {
         label = 33;
         break L27;
        }
        $59 = HEAP32[$34>>2]|0;
        $60 = ($57>>>0)<($51>>>0);
        $61 = ($57>>>0)>(536870911);
        $or$cond$i$us = $60 | $61;
        $newsize$0$i$us = $or$cond$i$us ? 536870911 : $57;
        $62 = HEAP32[$list>>2]|0;
        $63 = (_ft_mem_realloc($59,4,$51,$newsize$0$i$us,$62,$error$i)|0);
        HEAP32[$list>>2] = $63;
        $64 = HEAP32[$error$i>>2]|0;
        $65 = ($64|0)==(0);
        if (!($65)) {
         $162 = $64;
         break L27;
        }
        HEAP32[$1>>2] = $newsize$0$i$us;
        $$pre43$pre = HEAP32[$0>>2]|0;
        $$pre43 = $$pre43$pre;
       }
       $69 = $$pre43;
      } else {
       $69 = $51;
      }
      $66 = ($ep$1$lcssa$us$ph>>>0)>($sp$124$us>>>0);
      $67 = $66 ? $sp$124$us : 1282466;
      $68 = (($69) + 1)|0;
      HEAP32[$0>>2] = $68;
      $70 = HEAP32[$list>>2]|0;
      $71 = (($70) + ($69<<2)|0);
      HEAP32[$71>>2] = $67;
      $72 = HEAP8[$ep$1$lcssa$us$ph>>0]|0;
      $73 = ($72<<24>>24)==(0);
      if ($73) {
       $163 = 0;$78 = 0;$ep$39$us = $ep$1$lcssa$us$ph;
      } else {
       $74 = ((($ep$1$lcssa$us$ph)) + 1|0);
       HEAP8[$ep$1$lcssa$us$ph>>0] = 0;
       $75 = HEAP8[$74>>0]|0;
       $76 = ($75<<24>>24)==(0);
       $163 = $75;$78 = $76;$ep$39$us = $74;
      }
      $77 = $78&1;
      $79 = ($ep$39$us>>>0)<($32>>>0);
      if ($79) {
       $37 = $163;$final_empty$025$us = $77;$sp$124$us = $ep$39$us;
      } else {
       $final_empty$0$lcssa = $77;
       break L25;
      }
     }
    } else {
     $81 = $14;$final_empty$025 = 0;$sp$124 = $line;
     while(1) {
      $80 = ($81<<24>>24)==(0);
      if ($80) {
       $final_empty$0$lcssa = $final_empty$025;
       break L25;
      } else {
       $83 = $81;$ep$116 = $sp$124;
      }
      while(1) {
       $82 = $83&255;
       $84 = $82 >>> 3;
       $85 = (($seps) + ($84)|0);
       $86 = HEAP8[$85>>0]|0;
       $87 = $86 << 24 >> 24;
       $88 = $82 & 7;
       $89 = 1 << $88;
       $90 = $87 & $89;
       $91 = ($90|0)==(0);
       if (!($91)) {
        $ep$1$lcssa$ph = $ep$116;
        break;
       }
       $92 = ((($ep$116)) + 1|0);
       $93 = HEAP8[$92>>0]|0;
       $94 = ($93<<24>>24)==(0);
       if ($94) {
        $ep$1$lcssa$ph = $92;
        break;
       } else {
        $83 = $93;$ep$116 = $92;
       }
      }
      $95 = HEAP32[$0>>2]|0;
      $96 = HEAP32[$1>>2]|0;
      $97 = ($95|0)==($96|0);
      if ($97) {
       HEAP32[$error$i>>2] = 0;
       $98 = ($95|0)==(-1);
       if ($98) {
        $$pre42 = -1;
       } else {
        $99 = $95 >>> 1;
        $100 = (($95) + 5)|0;
        $101 = (($100) + ($99))|0;
        $102 = ($95|0)==(536870911);
        if ($102) {
         label = 33;
         break L27;
        }
        $103 = HEAP32[$34>>2]|0;
        $104 = ($101>>>0)<($95>>>0);
        $105 = ($101>>>0)>(536870911);
        $or$cond$i = $104 | $105;
        $newsize$0$i = $or$cond$i ? 536870911 : $101;
        $106 = HEAP32[$list>>2]|0;
        $107 = (_ft_mem_realloc($103,4,$95,$newsize$0$i,$106,$error$i)|0);
        HEAP32[$list>>2] = $107;
        $108 = HEAP32[$error$i>>2]|0;
        $109 = ($108|0)==(0);
        if (!($109)) {
         $162 = $108;
         break L27;
        }
        HEAP32[$1>>2] = $newsize$0$i;
        $$pre42$pre = HEAP32[$0>>2]|0;
        $$pre42 = $$pre42$pre;
       }
       $113 = $$pre42;
      } else {
       $113 = $95;
      }
      $110 = ($ep$1$lcssa$ph>>>0)>($sp$124>>>0);
      $111 = $110 ? $sp$124 : 1282466;
      $112 = (($113) + 1)|0;
      HEAP32[$0>>2] = $112;
      $114 = HEAP32[$list>>2]|0;
      $115 = (($114) + ($113<<2)|0);
      HEAP32[$115>>2] = $111;
      $116 = HEAP8[$ep$1$lcssa$ph>>0]|0;
      $117 = ($116<<24>>24)==(0);
      if ($117) {
       $135 = 0;$164 = 0;$ep$2$lcssa51 = $ep$1$lcssa$ph;
      } else {
       $119 = $116;$ep$219 = $ep$1$lcssa$ph;
       while(1) {
        $118 = $119&255;
        $120 = $118 >>> 3;
        $121 = (($seps) + ($120)|0);
        $122 = HEAP8[$121>>0]|0;
        $123 = $122 << 24 >> 24;
        $124 = $118 & 7;
        $125 = 1 << $124;
        $126 = $123 & $125;
        $127 = ($126|0)==(0);
        if ($127) {
         $133 = $119;$ep$2$lcssa = $ep$219;
         break;
        }
        HEAP8[$ep$219>>0] = 0;
        $128 = ((($ep$219)) + 1|0);
        $129 = HEAP8[$128>>0]|0;
        $130 = ($129<<24>>24)==(0);
        if ($130) {
         $133 = 0;$ep$2$lcssa = $128;
         break;
        } else {
         $119 = $129;$ep$219 = $128;
        }
       }
       $131 = ($ep$2$lcssa>>>0)>($ep$1$lcssa$ph>>>0);
       $132 = ($133<<24>>24)==(0);
       $$ = $131 & $132;
       $135 = $$;$164 = $133;$ep$2$lcssa51 = $ep$2$lcssa;
      }
      $134 = $135&1;
      $136 = ($ep$2$lcssa51>>>0)<($32>>>0);
      if ($136) {
       $81 = $164;$final_empty$025 = $134;$sp$124 = $ep$2$lcssa51;
      } else {
       $final_empty$0$lcssa = $134;
       break L25;
      }
     }
    }
   } while(0);
   if ((label|0) == 33) {
    HEAP32[$error$i>>2] = 64;
    $162 = 64;
   }
   $error$3 = $162;
   STACKTOP = sp;return ($error$3|0);
  } else {
   $final_empty$0$lcssa = 0;
  }
 } while(0);
 $137 = HEAP32[$0>>2]|0;
 $138 = (($137) + ($final_empty$0$lcssa))|0;
 $139 = HEAP32[$1>>2]|0;
 $140 = ($138>>>0)<($139>>>0);
 if (!($140)) {
  $141 = (($138) + 1)|0;
  HEAP32[$error$i4>>2] = 0;
  $142 = ($139>>>0)<($141>>>0);
  do {
   if ($142) {
    $143 = $139 >>> 1;
    $144 = (($139) + 5)|0;
    $145 = (($144) + ($143))|0;
    $146 = ($139|0)==(536870911);
    if ($146) {
     HEAP32[$error$i4>>2] = 64;
     $165 = 64;
    } else {
     $147 = ((($list)) + 12|0);
     $148 = HEAP32[$147>>2]|0;
     $149 = ($145>>>0)<($139>>>0);
     $150 = ($145>>>0)>(536870911);
     $or$cond$i5 = $149 | $150;
     $newsize$0$i6 = $or$cond$i5 ? 536870911 : $145;
     $151 = HEAP32[$list>>2]|0;
     $152 = (_ft_mem_realloc($148,4,$139,$newsize$0$i6,$151,$error$i4)|0);
     HEAP32[$list>>2] = $152;
     $153 = HEAP32[$error$i4>>2]|0;
     $154 = ($153|0)==(0);
     if ($154) {
      HEAP32[$1>>2] = $newsize$0$i6;
      break;
     } else {
      $165 = $153;
     }
    }
    $error$3 = $165;
    STACKTOP = sp;return ($error$3|0);
   }
  } while(0);
 }
 $155 = ($final_empty$0$lcssa|0)==(0);
 $$pre44 = HEAP32[$0>>2]|0;
 if ($155) {
  $$pre$phiZ2D = $list;$161 = $$pre44;
 } else {
  $156 = (($$pre44) + 1)|0;
  HEAP32[$0>>2] = $156;
  $157 = HEAP32[$list>>2]|0;
  $158 = (($157) + ($$pre44<<2)|0);
  HEAP32[$158>>2] = 1282466;
  $$pre$phiZ2D = $list;$161 = $156;
 }
 $159 = HEAP32[$$pre$phiZ2D>>2]|0;
 $160 = (($159) + ($161<<2)|0);
 HEAP32[$160>>2] = 0;
 $error$3 = 0;
 STACKTOP = sp;return ($error$3|0);
}
function __bdf_parse_properties($line,$linelen,$lineno,$call_data,$client_data) {
 $line = $line|0;
 $linelen = $linelen|0;
 $lineno = $lineno|0;
 $call_data = $call_data|0;
 $client_data = $client_data|0;
 var $$$i$i$i = 0, $$$i$i$i$i = 0, $$$i$i$i17 = 0, $$0$i$i = 0, $$0$i26 = 0, $$lcssa = 0, $$lcssa103 = 0, $$lcssa104 = 0, $$lcssa105 = 0, $$lcssa107 = 0, $$lcssa111 = 0, $$lcssa112 = 0, $$lcssa114 = 0, $$phi$trans$insert$i = 0, $$pre = 0, $$pre$i = 0, $$pre10$i = 0, $$pre11$i = 0, $$pre9$i = 0, $$sp$0$i = 0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $ep$0$i = 0, $ep$0$i$lcssa108 = 0, $ep$0$i109 = 0, $ep$113$i = 0, $error$2 = 0, $exitcond = 0, $exitcond88 = 0, $fp$02$i = 0, $hold$0$i = 0, $i$03$i = 0, $j$0$be$i = 0, $j$0$be$i$lcssa = 0, $j$04$i = 0;
 var $j$1$lcssa$i = 0, $j$11$i = 0, $kp$05$i$i$i$i = 0, $nbuf = 0, $ndp$01$i$i$i = 0, $ndp$01$i$i$i$i = 0, $ndp$01$i$i$i15 = 0, $res$04$i$i$i$i = 0, $sp$0$i = 0, $sp$0$i$lcssa = 0, $u$02$i89 = 0, $value$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $nbuf = sp + 16|0;
 $0 = (_strncmp($line,531875,13)|0);
 $1 = ($0|0)==(0);
 L1: do {
  if ($1) {
   $2 = ((($line)) + 13|0);
   $3 = HEAP8[$2>>0]|0;
   switch ($3<<24>>24) {
   case 9: case 13: case 10: case 0: case 32:  {
    break;
   }
   default: {
    break L1;
   }
   }
   $4 = ((($client_data)) + 32|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = ($5|0)==(0|0);
   L4: do {
    if ($6) {
     label = 11;
    } else {
     $7 = ((($5)) + 72|0);
     $8 = HEAP32[$7>>2]|0;
     $9 = ($8|0)==(0);
     if ($9) {
      label = 11;
     } else {
      $10 = ((($5)) + 128|0);
      $11 = HEAP32[$10>>2]|0;
      $12 = ((($11)) + 12|0);
      $13 = HEAP32[$12>>2]|0;
      $14 = ((($11)) + 4|0);
      $15 = HEAP32[$14>>2]|0;
      $16 = (1895527114 % ($15>>>0))&-1;
      $17 = (($13) + ($16<<2)|0);
      $18 = HEAP32[$17>>2]|0;
      $19 = ($18|0)==(0|0);
      if ($19) {
       label = 11;
      } else {
       $20 = (($15) + -1)|0;
       $21 = (($13) + ($20<<2)|0);
       $23 = $18;$ndp$01$i$i$i = $17;
       while(1) {
        $22 = HEAP32[$23>>2]|0;
        $24 = HEAP8[$22>>0]|0;
        $25 = ($24<<24>>24)==(70);
        if ($25) {
         $26 = (_strcmp($22,530793)|0);
         $27 = ($26|0)==(0);
         if ($27) {
          $$lcssa114 = $23;
          break;
         }
        }
        $28 = ((($ndp$01$i$i$i)) + -4|0);
        $29 = ($28>>>0)<($13>>>0);
        $$$i$i$i = $29 ? $21 : $28;
        $30 = HEAP32[$$$i$i$i>>2]|0;
        $31 = ($30|0)==(0|0);
        if ($31) {
         label = 11;
         break L4;
        } else {
         $23 = $30;$ndp$01$i$i$i = $$$i$i$i;
        }
       }
       $32 = ((($5)) + 80|0);
       $33 = HEAP32[$32>>2]|0;
       $34 = ((($$lcssa114)) + 4|0);
       $35 = HEAP32[$34>>2]|0;
       $36 = (($33) + ($35<<4)|0);
       $37 = ($36|0)==(0|0);
       if ($37) {
        label = 11;
       } else {
        $48 = $5;
       }
      }
     }
    }
   } while(0);
   do {
    if ((label|0) == 11) {
     $38 = ((($5)) + 12|0);
     $39 = HEAP16[$38>>1]|0;
     $40 = $39 << 16 >> 16;
     $41 = ((($5)) + 40|0);
     HEAP32[$41>>2] = $40;
     HEAP32[$vararg_buffer>>2] = $40;
     (_sprintf($nbuf,531889,$vararg_buffer)|0);
     $42 = HEAP32[$4>>2]|0;
     $43 = (__bdf_add_property($42,530793,$nbuf)|0);
     $44 = ($43|0)==(0);
     if ($44) {
      $45 = HEAP32[$4>>2]|0;
      $46 = ((($45)) + 278660|0);
      HEAP16[$46>>1] = 1;
      $48 = $45;
      break;
     } else {
      $error$2 = $43;
      STACKTOP = sp;return ($error$2|0);
     }
    }
   } while(0);
   $47 = ((($48)) + 72|0);
   $49 = HEAP32[$47>>2]|0;
   $50 = ($49|0)==(0);
   L20: do {
    if ($50) {
     label = 20;
    } else {
     $51 = ((($48)) + 128|0);
     $52 = HEAP32[$51>>2]|0;
     $53 = ((($52)) + 12|0);
     $54 = HEAP32[$53>>2]|0;
     $55 = ((($52)) + 4|0);
     $56 = HEAP32[$55>>2]|0;
     $57 = (908209322 % ($56>>>0))&-1;
     $58 = (($54) + ($57<<2)|0);
     $59 = HEAP32[$58>>2]|0;
     $60 = ($59|0)==(0|0);
     if ($60) {
      label = 20;
     } else {
      $61 = (($56) + -1)|0;
      $62 = (($54) + ($61<<2)|0);
      $64 = $59;$ndp$01$i$i$i15 = $58;
      while(1) {
       $63 = HEAP32[$64>>2]|0;
       $65 = HEAP8[$63>>0]|0;
       $66 = ($65<<24>>24)==(70);
       if ($66) {
        $67 = (_strcmp($63,530805)|0);
        $68 = ($67|0)==(0);
        if ($68) {
         $$lcssa112 = $64;
         break;
        }
       }
       $69 = ((($ndp$01$i$i$i15)) + -4|0);
       $70 = ($69>>>0)<($54>>>0);
       $$$i$i$i17 = $70 ? $62 : $69;
       $71 = HEAP32[$$$i$i$i17>>2]|0;
       $72 = ($71|0)==(0|0);
       if ($72) {
        label = 20;
        break L20;
       } else {
        $64 = $71;$ndp$01$i$i$i15 = $$$i$i$i17;
       }
      }
      $73 = ((($48)) + 80|0);
      $74 = HEAP32[$73>>2]|0;
      $75 = ((($$lcssa112)) + 4|0);
      $76 = HEAP32[$75>>2]|0;
      $77 = (($74) + ($76<<4)|0);
      $78 = ($77|0)==(0|0);
      if ($78) {
       label = 20;
      }
     }
    }
   } while(0);
   do {
    if ((label|0) == 20) {
     $79 = ((($48)) + 14|0);
     $80 = HEAP16[$79>>1]|0;
     $81 = $80 << 16 >> 16;
     $82 = ((($48)) + 44|0);
     HEAP32[$82>>2] = $81;
     HEAP32[$vararg_buffer1>>2] = $81;
     (_sprintf($nbuf,531889,$vararg_buffer1)|0);
     $83 = HEAP32[$4>>2]|0;
     $84 = (__bdf_add_property($83,530805,$nbuf)|0);
     $85 = ($84|0)==(0);
     if ($85) {
      $86 = HEAP32[$4>>2]|0;
      $87 = ((($86)) + 278660|0);
      HEAP16[$87>>1] = 1;
      break;
     } else {
      $error$2 = $84;
      STACKTOP = sp;return ($error$2|0);
     }
    }
   } while(0);
   $88 = HEAP32[$client_data>>2]|0;
   $89 = $88 & -17;
   HEAP32[$client_data>>2] = $89;
   HEAP32[$call_data>>2] = 4354;
   $error$2 = 0;
   STACKTOP = sp;return ($error$2|0);
  }
 } while(0);
 $90 = (_strncmp($line,531967,21)|0);
 $91 = ($90|0)==(0);
 L37: do {
  if ($91) {
   $92 = ((($line)) + 21|0);
   $93 = HEAP8[$92>>0]|0;
   switch ($93<<24>>24) {
   case 9: case 13: case 10: case 0: case 32:  {
    $error$2 = 0;
    break;
   }
   default: {
    break L37;
   }
   }
   STACKTOP = sp;return ($error$2|0);
  }
 } while(0);
 $94 = (_strncmp($line,530598,7)|0);
 $95 = ($94|0)==(0);
 L41: do {
  if ($95) {
   $96 = ((($line)) + 7|0);
   $97 = HEAP8[$96>>0]|0;
   switch ($97<<24>>24) {
   case 0:  {
    $value$0 = $96;
    break;
   }
   case 32: case 13: case 10: case 9:  {
    $98 = ((($line)) + 8|0);
    HEAP8[$96>>0] = 0;
    $value$0 = $98;
    break;
   }
   default: {
    break L41;
   }
   }
   $99 = ((($client_data)) + 32|0);
   $100 = HEAP32[$99>>2]|0;
   $101 = (__bdf_add_property($100,$line,$value$0)|0);
   $error$2 = $101;
   STACKTOP = sp;return ($error$2|0);
  }
 } while(0);
 $102 = ((($client_data)) + 32|0);
 $103 = HEAP32[$102>>2]|0;
 $ep$0$i = $line;
 L48: while(1) {
  $104 = HEAP8[$ep$0$i>>0]|0;
  switch ($104<<24>>24) {
  case 0:  {
   $ep$0$i109 = $ep$0$i;$hold$0$i = -1;
   break L48;
   break;
  }
  case 32: case 9:  {
   $$lcssa111 = $104;$ep$0$i$lcssa108 = $ep$0$i;
   label = 32;
   break L48;
   break;
  }
  default: {
  }
  }
  $105 = ((($ep$0$i)) + 1|0);
  $ep$0$i = $105;
 }
 if ((label|0) == 32) {
  $106 = $$lcssa111 << 24 >> 24;
  HEAP8[$ep$0$i$lcssa108>>0] = 0;
  $ep$0$i109 = $ep$0$i$lcssa108;$hold$0$i = $106;
 }
 $107 = ($line|0)==(0|0);
 L54: do {
  if ($107) {
   $$0$i$i = 0;
  } else {
   $108 = HEAP8[$line>>0]|0;
   $109 = ($108<<24>>24)==(0);
   if ($109) {
    $$0$i$i = 0;
   } else {
    $110 = ((($103)) + 278688|0);
    $111 = HEAP32[$110>>2]|0;
    $115 = $108;$kp$05$i$i$i$i = $line;$res$04$i$i$i$i = 0;
    while(1) {
     $112 = ($res$04$i$i$i$i*31)|0;
     $113 = ((($kp$05$i$i$i$i)) + 1|0);
     $114 = $115 << 24 >> 24;
     $116 = (($112) + ($114))|0;
     $117 = HEAP8[$113>>0]|0;
     $118 = ($117<<24>>24)==(0);
     if ($118) {
      $$lcssa107 = $116;
      break;
     } else {
      $115 = $117;$kp$05$i$i$i$i = $113;$res$04$i$i$i$i = $116;
     }
    }
    $119 = ((($103)) + 278680|0);
    $120 = HEAP32[$119>>2]|0;
    $121 = (($$lcssa107>>>0) % ($120>>>0))&-1;
    $122 = (($111) + ($121<<2)|0);
    $123 = HEAP32[$122>>2]|0;
    $124 = ($123|0)==(0|0);
    if ($124) {
     $$0$i$i = 0;
    } else {
     $125 = (($120) + -1)|0;
     $126 = (($111) + ($125<<2)|0);
     $128 = $123;$ndp$01$i$i$i$i = $122;
     while(1) {
      $127 = HEAP32[$128>>2]|0;
      $129 = HEAP8[$127>>0]|0;
      $130 = ($129<<24>>24)==($108<<24>>24);
      if ($130) {
       $131 = (_strcmp($127,$line)|0);
       $132 = ($131|0)==(0);
       if ($132) {
        $$lcssa105 = $128;
        break;
       }
      }
      $133 = ((($ndp$01$i$i$i$i)) + -4|0);
      $134 = ($133>>>0)<($111>>>0);
      $$$i$i$i$i = $134 ? $126 : $133;
      $135 = HEAP32[$$$i$i$i$i>>2]|0;
      $136 = ($135|0)==(0|0);
      if ($136) {
       $$0$i$i = 0;
       break L54;
      } else {
       $128 = $135;$ndp$01$i$i$i$i = $$$i$i$i$i;
      }
     }
     $137 = ((($$lcssa105)) + 4|0);
     $138 = HEAP32[$137>>2]|0;
     $139 = ($138>>>0)>(82);
     if ($139) {
      $140 = ((($103)) + 278668|0);
      $141 = HEAP32[$140>>2]|0;
      $142 = (($138) + -83)|0;
      $143 = (($141) + ($142<<4)|0);
      $$0$i$i = $143;
      break;
     } else {
      $144 = (48000 + ($138<<4)|0);
      $$0$i$i = $144;
      break;
     }
    }
   }
  }
 } while(0);
 $145 = ($hold$0$i|0)==(-1);
 if ($145) {
  $153 = 0;
 } else {
  $146 = $hold$0$i&255;
  HEAP8[$ep$0$i109>>0] = $146;
  $153 = $146;
 }
 $147 = ($$0$i$i|0)==(0|0);
 if (!($147)) {
  $148 = ((($$0$i$i)) + 4|0);
  $149 = HEAP32[$148>>2]|0;
  $150 = ($149|0)==(1);
  if (!($150)) {
   $165 = ((($client_data)) + 139304|0);
   $166 = (__bdf_list_split($165,531680,$line,$linelen)|0);
   $167 = ($166|0)==(0);
   if (!($167)) {
    $error$2 = $166;
    STACKTOP = sp;return ($error$2|0);
   }
   $168 = HEAP32[$165>>2]|0;
   $169 = HEAP32[$168>>2]|0;
   $170 = ((($client_data)) + 139312|0);
   $171 = HEAP32[$170>>2]|0;
   $172 = ($171|0)==(0);
   do {
    if ($172) {
     $$0$i26 = 0;
    } else {
     $173 = ($171>>>0)<(2);
     if ($173) {
      HEAP32[$170>>2] = 0;
      $$0$i26 = 0;
      break;
     }
     $174 = (($171) + -1)|0;
     $175 = ((($168)) + 4|0);
     $176 = HEAP32[$175>>2]|0;
     HEAP32[$168>>2] = $176;
     $exitcond88 = ($174|0)==(1);
     if (!($exitcond88)) {
      $181 = 1;$u$02$i89 = 1;
      while(1) {
       $177 = (($u$02$i89) + 1)|0;
       $$pre = HEAP32[$165>>2]|0;
       $178 = (($$pre) + ($177<<2)|0);
       $179 = HEAP32[$178>>2]|0;
       $180 = (($$pre) + ($181<<2)|0);
       HEAP32[$180>>2] = $179;
       $182 = (($181) + 1)|0;
       $exitcond = ($182|0)==($174|0);
       if ($exitcond) {
        break;
       } else {
        $181 = $182;$u$02$i89 = $177;
       }
      }
     }
     HEAP32[$170>>2] = $174;
     $183 = ($174|0)==(0);
     if ($183) {
      $$0$i26 = 0;
     } else {
      $184 = HEAP32[$165>>2]|0;
      $185 = HEAP32[$184>>2]|0;
      $187 = $185;$206 = $174;$i$03$i = 0;$j$04$i = 0;
      while(1) {
       $186 = HEAP8[$187>>0]|0;
       $188 = ($186<<24>>24)==(0);
       if ($188) {
        $197 = $206;$j$1$lcssa$i = $j$04$i;
       } else {
        $192 = $186;$fp$02$i = $187;$j$11$i = $j$04$i;
        while(1) {
         $189 = ((($fp$02$i)) + 1|0);
         $190 = (($j$11$i) + 1)|0;
         $191 = (($185) + ($j$11$i)|0);
         HEAP8[$191>>0] = $192;
         $193 = HEAP8[$189>>0]|0;
         $194 = ($193<<24>>24)==(0);
         if ($194) {
          $$lcssa104 = $190;
          break;
         } else {
          $192 = $193;$fp$02$i = $189;$j$11$i = $190;
         }
        }
        $$pre10$i = HEAP32[$170>>2]|0;
        $197 = $$pre10$i;$j$1$lcssa$i = $$lcssa104;
       }
       $195 = (($i$03$i) + 1)|0;
       $196 = ($195>>>0)<($197>>>0);
       if ($196) {
        $198 = (($j$1$lcssa$i) + 1)|0;
        $199 = (($185) + ($j$1$lcssa$i)|0);
        HEAP8[$199>>0] = 32;
        $$pre11$i = HEAP32[$170>>2]|0;
        $201 = $$pre11$i;$j$0$be$i = $198;
       } else {
        $201 = $197;$j$0$be$i = $j$1$lcssa$i;
       }
       $200 = ($195>>>0)<($201>>>0);
       if (!($200)) {
        $j$0$be$i$lcssa = $j$0$be$i;
        break;
       }
       $$pre$i = HEAP32[$165>>2]|0;
       $$phi$trans$insert$i = (($$pre$i) + ($195<<2)|0);
       $$pre9$i = HEAP32[$$phi$trans$insert$i>>2]|0;
       $187 = $$pre9$i;$206 = $201;$i$03$i = $195;$j$04$i = $j$0$be$i;
      }
      $202 = ($185|0)==(1282466|0);
      if ($202) {
       $$0$i26 = 1282466;
      } else {
       $203 = (($185) + ($j$0$be$i$lcssa)|0);
       HEAP8[$203>>0] = 0;
       $$0$i26 = $185;
      }
     }
    }
   } while(0);
   $204 = HEAP32[$102>>2]|0;
   $205 = (__bdf_add_property($204,$169,$$0$i26)|0);
   $error$2 = $205;
   STACKTOP = sp;return ($error$2|0);
  }
 }
 $151 = (($line) + ($linelen)|0);
 $152 = ($153<<24>>24)==(0);
 if ($152) {
  $sp$0$i = $ep$0$i109;
 } else {
  $154 = ((($ep$0$i109)) + 1|0);
  HEAP8[$ep$0$i109>>0] = 0;
  $sp$0$i = $154;
 }
 L108: while(1) {
  $155 = HEAP8[$sp$0$i>>0]|0;
  switch ($155<<24>>24) {
  case 32: case 9:  {
   break;
  }
  default: {
   $$lcssa = $155;$sp$0$i$lcssa = $sp$0$i;
   break L108;
  }
  }
  $156 = ((($sp$0$i)) + 1|0);
  $sp$0$i = $156;
 }
 $157 = ($$lcssa<<24>>24)==(34);
 $158 = ((($sp$0$i$lcssa)) + 1|0);
 $$sp$0$i = $157 ? $158 : $sp$0$i$lcssa;
 $159 = ($151>>>0)>($$sp$0$i>>>0);
 L112: do {
  if ($159) {
   $ep$113$i = $151;
   L113: while(1) {
    $160 = ((($ep$113$i)) + -1|0);
    $161 = HEAP8[$160>>0]|0;
    switch ($161<<24>>24) {
    case 34:  {
     $$lcssa103 = $160;
     break L113;
     break;
    }
    case 9: case 32:  {
     break;
    }
    default: {
     break L112;
    }
    }
    HEAP8[$160>>0] = 0;
    $162 = ($160>>>0)>($$sp$0$i>>>0);
    if ($162) {
     $ep$113$i = $160;
    } else {
     break L112;
    }
   }
   HEAP8[$$lcssa103>>0] = 0;
  }
 } while(0);
 $163 = HEAP32[$102>>2]|0;
 $164 = (__bdf_add_property($163,$line,$$sp$0$i)|0);
 $error$2 = $164;
 STACKTOP = sp;return ($error$2|0);
}
function __bdf_add_property($font,$name,$value) {
 $font = $font|0;
 $name = $name|0;
 $value = $value|0;
 var $$$i$i = 0, $$$i$i$i = 0, $$$i$i49 = 0, $$$i$i60 = 0, $$01$i = 0, $$01$i19 = 0, $$01$i27 = 0, $$01$i39 = 0, $$02$i = 0, $$02$i21 = 0, $$1$i = 0, $$1$i15 = 0, $$1$i22 = 0, $$1$i32 = 0, $$13$i = 0, $$13$i31 = 0, $$131$i = 0, $$131$i25 = 0, $$21$i = 0, $$21$i35 = 0;
 var $$in = 0, $$lcssa137 = 0, $$ph = 0, $$pre = 0, $$pre$i = 0, $$pre$i14 = 0, $$pre$i20 = 0, $$pre$i30 = 0, $$pre92 = 0, $$s$i = 0, $$s$i29 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0;
 var $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0;
 var $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0;
 var $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0;
 var $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0;
 var $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0;
 var $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0;
 var $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0;
 var $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0;
 var $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0;
 var $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $dmap$1$i = 0, $dmap$1$i16 = 0, $dmap$1$i23 = 0, $dmap$1$i33 = 0, $error = 0, $error$i = 0, $hn$0 = 0, $kp$05$i$i = 0, $kp$05$i$i$i = 0, $kp$05$i$i41 = 0, $kp$05$i$i52 = 0, $ndp$01$i$i = 0, $ndp$01$i$i$i = 0, $ndp$01$i$i47 = 0, $ndp$01$i$i58 = 0, $prop$0 = 0, $res$0$lcssa$i$i = 0;
 var $res$0$lcssa$i$i$i = 0, $res$0$lcssa$i$i44 = 0, $res$0$lcssa$i$i55 = 0, $res$04$i$i = 0, $res$04$i$i$i = 0, $res$04$i$i42 = 0, $res$04$i$i53 = 0, $v$0$lcssa$i = 0, $v$0$lcssa$i37 = 0, $v$02$i = 0, $v$02$i17 = 0, $v$02$i24 = 0, $v$02$i34 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp + 4|0;
 $error = sp;
 $0 = ((($font)) + 278664|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$error>>2] = 0;
 $2 = ((($font)) + 128|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = HEAP8[$name>>0]|0;
 $7 = ($6<<24>>24)==(0);
 if ($7) {
  $res$0$lcssa$i$i = 0;
 } else {
  $11 = $6;$kp$05$i$i = $name;$res$04$i$i = 0;
  while(1) {
   $8 = ($res$04$i$i*31)|0;
   $9 = ((($kp$05$i$i)) + 1|0);
   $10 = $11 << 24 >> 24;
   $12 = (($8) + ($10))|0;
   $13 = HEAP8[$9>>0]|0;
   $14 = ($13<<24>>24)==(0);
   if ($14) {
    $res$0$lcssa$i$i = $12;
    break;
   } else {
    $11 = $13;$kp$05$i$i = $9;$res$04$i$i = $12;
   }
  }
 }
 $15 = ((($3)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = (($res$0$lcssa$i$i>>>0) % ($16>>>0))&-1;
 $18 = (($5) + ($17<<2)|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($19|0)==(0|0);
 L5: do {
  if (!($20)) {
   $21 = (($16) + -1)|0;
   $22 = (($5) + ($21<<2)|0);
   $24 = $19;$ndp$01$i$i = $18;
   while(1) {
    $23 = HEAP32[$24>>2]|0;
    $25 = HEAP8[$23>>0]|0;
    $26 = ($25<<24>>24)==($6<<24>>24);
    if ($26) {
     $27 = (_strcmp($23,$name)|0);
     $28 = ($27|0)==(0);
     if ($28) {
      $$lcssa137 = $24;
      break;
     }
    }
    $29 = ((($ndp$01$i$i)) + -4|0);
    $30 = ($29>>>0)<($5>>>0);
    $$$i$i = $30 ? $22 : $29;
    $31 = HEAP32[$$$i$i>>2]|0;
    $32 = ($31|0)==(0|0);
    if ($32) {
     break L5;
    } else {
     $24 = $31;$ndp$01$i$i = $$$i$i;
    }
   }
   $33 = ((($font)) + 80|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = ((($$lcssa137)) + 4|0);
   $36 = HEAP32[$35>>2]|0;
   $37 = (((($34) + ($36<<4)|0)) + 4|0);
   $38 = HEAP32[$37>>2]|0;
   switch ($38|0) {
   case 1:  {
    $39 = (((($34) + ($36<<4)|0)) + 12|0);
    $40 = HEAP32[$39>>2]|0;
    _ft_mem_free($1,$40);
    HEAP32[$39>>2] = 0;
    $41 = ($value|0)==(0|0);
    if ($41) {
     $391 = HEAP32[$error>>2]|0;
     STACKTOP = sp;return ($391|0);
    }
    $42 = HEAP8[$value>>0]|0;
    $43 = ($42<<24>>24)==(0);
    if ($43) {
     $391 = HEAP32[$error>>2]|0;
     STACKTOP = sp;return ($391|0);
    }
    $44 = (_ft_mem_strdup($1,$value,$error)|0);
    HEAP32[$39>>2] = $44;
    $391 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($391|0);
    break;
   }
   case 2:  {
    $45 = ($value|0)==(0|0);
    if ($45) {
     $$01$i = 0;
    } else {
     $46 = HEAP8[$value>>0]|0;
     $47 = ($46<<24>>24)==(0);
     if ($47) {
      $$01$i = 0;
     } else {
      $48 = ($46<<24>>24)==(45);
      $49 = ((($value)) + 1|0);
      $$s$i = $48 ? $49 : $value;
      $50 = HEAP8[$$s$i>>0]|0;
      $51 = ($50<<24>>24)==(48);
      L26: do {
       if ($51) {
        $52 = ((($$s$i)) + 1|0);
        $53 = HEAP8[$52>>0]|0;
        switch ($53<<24>>24) {
        case 88: case 120:  {
         break;
        }
        default: {
         $$1$i = 10;$$13$i = $$s$i;$56 = 48;$dmap$1$i = 531715;
         break L26;
        }
        }
        $54 = ((($$s$i)) + 2|0);
        $$pre$i = HEAP8[$54>>0]|0;
        $$1$i = 16;$$13$i = $54;$56 = $$pre$i;$dmap$1$i = 531683;
       } else {
        $$1$i = 10;$$13$i = $$s$i;$56 = $50;$dmap$1$i = 531715;
       }
      } while(0);
      $55 = $56&255;
      $57 = $55 >>> 3;
      $58 = (($dmap$1$i) + ($57)|0);
      $59 = HEAP8[$58>>0]|0;
      $60 = $59&255;
      $61 = $56 << 24 >> 24;
      $62 = $61 & 7;
      $63 = 1 << $62;
      $64 = $60 & $63;
      $65 = ($64|0)==(0);
      if ($65) {
       $v$0$lcssa$i = 0;
      } else {
       $$21$i = $$13$i;$68 = $61;$v$02$i = 0;
       while(1) {
        $66 = Math_imul($v$02$i, $$1$i)|0;
        $67 = (531747 + ($68)|0);
        $69 = HEAP8[$67>>0]|0;
        $70 = $69&255;
        $71 = (($70) + ($66))|0;
        $72 = ((($$21$i)) + 1|0);
        $73 = HEAP8[$72>>0]|0;
        $74 = $73&255;
        $75 = $74 >>> 3;
        $76 = (($dmap$1$i) + ($75)|0);
        $77 = HEAP8[$76>>0]|0;
        $78 = $77&255;
        $79 = $73 << 24 >> 24;
        $80 = $79 & 7;
        $81 = 1 << $80;
        $82 = $81 & $78;
        $83 = ($82|0)==(0);
        if ($83) {
         $v$0$lcssa$i = $71;
         break;
        } else {
         $$21$i = $72;$68 = $79;$v$02$i = $71;
        }
       }
      }
      $84 = (0 - ($v$0$lcssa$i))|0;
      $85 = $48 ? $84 : $v$0$lcssa$i;
      $$01$i = $85;
     }
    }
    $86 = (((($34) + ($36<<4)|0)) + 12|0);
    HEAP32[$86>>2] = $$01$i;
    $391 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($391|0);
    break;
   }
   case 3:  {
    $87 = ($value|0)==(0|0);
    L37: do {
     if ($87) {
      $$01$i19 = 0;
     } else {
      $88 = HEAP8[$value>>0]|0;
      L39: do {
       switch ($88<<24>>24) {
       case 0:  {
        $$01$i19 = 0;
        break L37;
        break;
       }
       case 48:  {
        $89 = ((($value)) + 1|0);
        $90 = HEAP8[$89>>0]|0;
        switch ($90<<24>>24) {
        case 88: case 120:  {
         break;
        }
        default: {
         $$02$i = $value;$$1$i15 = 10;$93 = 48;$dmap$1$i16 = 531715;
         break L39;
        }
        }
        $91 = ((($value)) + 2|0);
        $$pre$i14 = HEAP8[$91>>0]|0;
        $$02$i = $91;$$1$i15 = 16;$93 = $$pre$i14;$dmap$1$i16 = 531683;
        break;
       }
       default: {
        $$02$i = $value;$$1$i15 = 10;$93 = $88;$dmap$1$i16 = 531715;
       }
       }
      } while(0);
      $92 = $93&255;
      $94 = $92 >>> 3;
      $95 = (($dmap$1$i16) + ($94)|0);
      $96 = HEAP8[$95>>0]|0;
      $97 = $96&255;
      $98 = $93 << 24 >> 24;
      $99 = $98 & 7;
      $100 = 1 << $99;
      $101 = $97 & $100;
      $102 = ($101|0)==(0);
      if ($102) {
       $$01$i19 = 0;
      } else {
       $$131$i = $$02$i;$105 = $98;$v$02$i17 = 0;
       while(1) {
        $103 = Math_imul($v$02$i17, $$1$i15)|0;
        $104 = (531747 + ($105)|0);
        $106 = HEAP8[$104>>0]|0;
        $107 = $106&255;
        $108 = (($107) + ($103))|0;
        $109 = ((($$131$i)) + 1|0);
        $110 = HEAP8[$109>>0]|0;
        $111 = $110&255;
        $112 = $111 >>> 3;
        $113 = (($dmap$1$i16) + ($112)|0);
        $114 = HEAP8[$113>>0]|0;
        $115 = $114&255;
        $116 = $110 << 24 >> 24;
        $117 = $116 & 7;
        $118 = 1 << $117;
        $119 = $118 & $115;
        $120 = ($119|0)==(0);
        if ($120) {
         $$01$i19 = $108;
         break;
        } else {
         $$131$i = $109;$105 = $116;$v$02$i17 = $108;
        }
       }
      }
     }
    } while(0);
    $121 = (((($34) + ($36<<4)|0)) + 12|0);
    HEAP32[$121>>2] = $$01$i19;
    $391 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($391|0);
    break;
   }
   default: {
    $391 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($391|0);
   }
   }
  }
 } while(0);
 $122 = ((($font)) + 278676|0);
 $123 = ((($font)) + 278688|0);
 $124 = HEAP32[$123>>2]|0;
 if ($7) {
  $res$0$lcssa$i$i55 = 0;
 } else {
  $128 = $6;$kp$05$i$i52 = $name;$res$04$i$i53 = 0;
  while(1) {
   $125 = ($res$04$i$i53*31)|0;
   $126 = ((($kp$05$i$i52)) + 1|0);
   $127 = $128 << 24 >> 24;
   $129 = (($125) + ($127))|0;
   $130 = HEAP8[$126>>0]|0;
   $131 = ($130<<24>>24)==(0);
   if ($131) {
    $res$0$lcssa$i$i55 = $129;
    break;
   } else {
    $128 = $130;$kp$05$i$i52 = $126;$res$04$i$i53 = $129;
   }
  }
 }
 $132 = ((($font)) + 278680|0);
 $133 = HEAP32[$132>>2]|0;
 $134 = (($res$0$lcssa$i$i55>>>0) % ($133>>>0))&-1;
 $135 = (($124) + ($134<<2)|0);
 $136 = HEAP32[$135>>2]|0;
 $137 = ($136|0)==(0|0);
 L53: do {
  if ($137) {
   label = 35;
  } else {
   $138 = (($133) + -1)|0;
   $139 = (($124) + ($138<<2)|0);
   $141 = $136;$ndp$01$i$i58 = $135;
   while(1) {
    $140 = HEAP32[$141>>2]|0;
    $142 = HEAP8[$140>>0]|0;
    $143 = ($142<<24>>24)==($6<<24>>24);
    if ($143) {
     $144 = (_strcmp($140,$name)|0);
     $145 = ($144|0)==(0);
     if ($145) {
      $hn$0 = $141;
      break L53;
     }
    }
    $146 = ((($ndp$01$i$i58)) + -4|0);
    $147 = ($146>>>0)<($124>>>0);
    $$$i$i60 = $147 ? $139 : $146;
    $148 = HEAP32[$$$i$i60>>2]|0;
    $149 = ($148|0)==(0|0);
    if ($149) {
     label = 35;
     break;
    } else {
     $141 = $148;$ndp$01$i$i58 = $$$i$i60;
    }
   }
  }
 } while(0);
 L60: do {
  if ((label|0) == 35) {
   HEAP32[$error$i>>2] = 0;
   if ($7) {
    $res$0$lcssa$i$i$i = 0;
   } else {
    $153 = $6;$kp$05$i$i$i = $name;$res$04$i$i$i = 0;
    while(1) {
     $150 = ($res$04$i$i$i*31)|0;
     $151 = ((($kp$05$i$i$i)) + 1|0);
     $152 = $153 << 24 >> 24;
     $154 = (($150) + ($152))|0;
     $155 = HEAP8[$151>>0]|0;
     $156 = ($155<<24>>24)==(0);
     if ($156) {
      $res$0$lcssa$i$i$i = $154;
      break;
     } else {
      $153 = $155;$kp$05$i$i$i = $151;$res$04$i$i$i = $154;
     }
    }
   }
   $157 = (($res$0$lcssa$i$i$i>>>0) % ($133>>>0))&-1;
   $158 = (($124) + ($157<<2)|0);
   $159 = HEAP32[$158>>2]|0;
   $160 = ($159|0)==(0|0);
   L66: do {
    if ($160) {
     label = 42;
    } else {
     $161 = (($133) + -1)|0;
     $162 = (($124) + ($161<<2)|0);
     $164 = $159;$ndp$01$i$i$i = $158;
     while(1) {
      $163 = HEAP32[$164>>2]|0;
      $165 = HEAP8[$163>>0]|0;
      $166 = ($165<<24>>24)==($6<<24>>24);
      if ($166) {
       $167 = (_strcmp($163,$name)|0);
       $168 = ($167|0)==(0);
       if ($168) {
        $198 = $6;$209 = $124;
        break L66;
       }
      }
      $169 = ((($ndp$01$i$i$i)) + -4|0);
      $170 = ($169>>>0)<($124>>>0);
      $$$i$i$i = $170 ? $162 : $169;
      $171 = HEAP32[$$$i$i$i>>2]|0;
      $172 = ($171|0)==(0|0);
      if ($172) {
       label = 42;
       break;
      } else {
       $164 = $171;$ndp$01$i$i$i = $$$i$i$i;
      }
     }
    }
   } while(0);
   do {
    if ((label|0) == 42) {
     $173 = ((($font)) + 278672|0);
     $174 = HEAP32[$173>>2]|0;
     $175 = (($174) + 1)|0;
     $176 = ((($font)) + 278668|0);
     $177 = HEAP32[$176>>2]|0;
     $178 = (_ft_mem_realloc($1,16,$174,$175,$177,$error$i)|0);
     HEAP32[$176>>2] = $178;
     $179 = HEAP32[$error$i>>2]|0;
     $180 = ($179|0)==(0);
     if ($180) {
      $181 = HEAP32[$173>>2]|0;
      $182 = (($178) + ($181<<4)|0);
      ;HEAP32[$182>>2]=0|0;HEAP32[$182+4>>2]=0|0;HEAP32[$182+8>>2]=0|0;HEAP32[$182+12>>2]=0|0;
      $183 = (_strlen($name)|0);
      $184 = (($183) + 1)|0;
      $185 = (_ft_mem_realloc($1,1,0,$184,0,$error$i)|0);
      HEAP32[$182>>2] = $185;
      $186 = HEAP32[$error$i>>2]|0;
      $187 = ($186|0)==(0);
      if ($187) {
       _memcpy(($185|0),($name|0),($184|0))|0;
       $188 = (((($178) + ($181<<4)|0)) + 4|0);
       HEAP32[$188>>2] = 1;
       $189 = (((($178) + ($181<<4)|0)) + 8|0);
       HEAP32[$189>>2] = 0;
       $190 = HEAP32[$173>>2]|0;
       $191 = (($190) + 83)|0;
       $192 = HEAP32[$182>>2]|0;
       $193 = (_hash_insert($192,$191,$122,$1)|0);
       HEAP32[$error$i>>2] = $193;
       $194 = ($193|0)==(0);
       if ($194) {
        $195 = HEAP32[$173>>2]|0;
        $196 = (($195) + 1)|0;
        HEAP32[$173>>2] = $196;
        $$pre = HEAP32[$123>>2]|0;
        $$pre92 = HEAP8[$name>>0]|0;
        $198 = $$pre92;$209 = $$pre;
        break;
       } else {
        $$ph = $193;
       }
      } else {
       $$ph = $186;
      }
     } else {
      $$ph = $179;
     }
     HEAP32[$error>>2] = $$ph;
     $391 = HEAP32[$error>>2]|0;
     STACKTOP = sp;return ($391|0);
    }
   } while(0);
   HEAP32[$error>>2] = 0;
   $197 = ($198<<24>>24)==(0);
   if ($197) {
    $res$0$lcssa$i$i44 = 0;
   } else {
    $202 = $198;$kp$05$i$i41 = $name;$res$04$i$i42 = 0;
    while(1) {
     $199 = ($res$04$i$i42*31)|0;
     $200 = ((($kp$05$i$i41)) + 1|0);
     $201 = $202 << 24 >> 24;
     $203 = (($199) + ($201))|0;
     $204 = HEAP8[$200>>0]|0;
     $205 = ($204<<24>>24)==(0);
     if ($205) {
      $res$0$lcssa$i$i44 = $203;
      break;
     } else {
      $202 = $204;$kp$05$i$i41 = $200;$res$04$i$i42 = $203;
     }
    }
   }
   $206 = HEAP32[$132>>2]|0;
   $207 = (($res$0$lcssa$i$i44>>>0) % ($206>>>0))&-1;
   $208 = (($209) + ($207<<2)|0);
   $210 = HEAP32[$208>>2]|0;
   $211 = ($210|0)==(0|0);
   if ($211) {
    $hn$0 = 0;
   } else {
    $212 = (($206) + -1)|0;
    $213 = (($209) + ($212<<2)|0);
    $215 = $210;$ndp$01$i$i47 = $208;
    while(1) {
     $214 = HEAP32[$215>>2]|0;
     $216 = HEAP8[$214>>0]|0;
     $217 = ($216<<24>>24)==($198<<24>>24);
     if ($217) {
      $218 = (_strcmp($214,$name)|0);
      $219 = ($218|0)==(0);
      if ($219) {
       $hn$0 = $215;
       break L60;
      }
     }
     $220 = ((($ndp$01$i$i47)) + -4|0);
     $221 = ($220>>>0)<($209>>>0);
     $$$i$i49 = $221 ? $213 : $220;
     $222 = HEAP32[$$$i$i49>>2]|0;
     $223 = ($222|0)==(0|0);
     if ($223) {
      $hn$0 = 0;
      break;
     } else {
      $215 = $222;$ndp$01$i$i47 = $$$i$i49;
     }
    }
   }
  }
 } while(0);
 $224 = ((($font)) + 76|0);
 $225 = HEAP32[$224>>2]|0;
 $226 = ((($font)) + 72|0);
 $227 = HEAP32[$226>>2]|0;
 $228 = ($225|0)==($227|0);
 if ($228) {
  $229 = ($225|0)==(0);
  if ($229) {
   $230 = (_ft_mem_realloc($1,16,0,1,0,$error)|0);
   $231 = ((($font)) + 80|0);
   HEAP32[$231>>2] = $230;
   $232 = HEAP32[$error>>2]|0;
   $233 = ($232|0)==(0);
   if ($233) {
    $$in = $230;
   } else {
    $391 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($391|0);
   }
  } else {
   $234 = (($225) + 1)|0;
   $235 = ((($font)) + 80|0);
   $236 = HEAP32[$235>>2]|0;
   $237 = (_ft_mem_realloc($1,16,$225,$234,$236,$error)|0);
   HEAP32[$235>>2] = $237;
   $238 = HEAP32[$error>>2]|0;
   $239 = ($238|0)==(0);
   if ($239) {
    $$in = $237;
   } else {
    $391 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($391|0);
   }
  }
  $240 = HEAP32[$226>>2]|0;
  $241 = (($$in) + ($240<<4)|0);
  ;HEAP32[$241>>2]=0|0;HEAP32[$241+4>>2]=0|0;HEAP32[$241+8>>2]=0|0;HEAP32[$241+12>>2]=0|0;
  $242 = HEAP32[$226>>2]|0;
  $243 = (($242) + 1)|0;
  HEAP32[$226>>2] = $243;
 }
 $244 = ((($hn$0)) + 4|0);
 $245 = HEAP32[$244>>2]|0;
 $246 = ($245>>>0)>(82);
 if ($246) {
  $247 = ((($font)) + 278668|0);
  $248 = HEAP32[$247>>2]|0;
  $249 = (($245) + -83)|0;
  $250 = (($248) + ($249<<4)|0);
  $prop$0 = $250;
 } else {
  $251 = (48000 + ($245<<4)|0);
  $prop$0 = $251;
 }
 $252 = ((($font)) + 80|0);
 $253 = HEAP32[$252>>2]|0;
 $254 = HEAP32[$224>>2]|0;
 $255 = (($253) + ($254<<4)|0);
 $256 = HEAP32[$prop$0>>2]|0;
 HEAP32[$255>>2] = $256;
 $257 = ((($prop$0)) + 4|0);
 $258 = HEAP32[$257>>2]|0;
 $259 = (((($253) + ($254<<4)|0)) + 4|0);
 HEAP32[$259>>2] = $258;
 $260 = ((($prop$0)) + 8|0);
 $261 = HEAP32[$260>>2]|0;
 $262 = (((($253) + ($254<<4)|0)) + 8|0);
 HEAP32[$262>>2] = $261;
 $263 = HEAP32[$257>>2]|0;
 switch ($263|0) {
 case 1:  {
  $264 = (((($253) + ($254<<4)|0)) + 12|0);
  HEAP32[$264>>2] = 0;
  $265 = ($value|0)==(0|0);
  if (!($265)) {
   $266 = HEAP8[$value>>0]|0;
   $267 = ($266<<24>>24)==(0);
   if (!($267)) {
    $268 = (_ft_mem_strdup($1,$value,$error)|0);
    HEAP32[$264>>2] = $268;
    $269 = HEAP32[$error>>2]|0;
    $270 = ($269|0)==(0);
    if (!($270)) {
     $391 = HEAP32[$error>>2]|0;
     STACKTOP = sp;return ($391|0);
    }
   }
  }
  break;
 }
 case 2:  {
  $271 = ($value|0)==(0|0);
  if ($271) {
   $$01$i39 = 0;
  } else {
   $272 = HEAP8[$value>>0]|0;
   $273 = ($272<<24>>24)==(0);
   if ($273) {
    $$01$i39 = 0;
   } else {
    $274 = ($272<<24>>24)==(45);
    $275 = ((($value)) + 1|0);
    $$s$i29 = $274 ? $275 : $value;
    $276 = HEAP8[$$s$i29>>0]|0;
    $277 = ($276<<24>>24)==(48);
    L115: do {
     if ($277) {
      $278 = ((($$s$i29)) + 1|0);
      $279 = HEAP8[$278>>0]|0;
      switch ($279<<24>>24) {
      case 88: case 120:  {
       break;
      }
      default: {
       $$1$i32 = 10;$$13$i31 = $$s$i29;$282 = 48;$dmap$1$i33 = 531715;
       break L115;
      }
      }
      $280 = ((($$s$i29)) + 2|0);
      $$pre$i30 = HEAP8[$280>>0]|0;
      $$1$i32 = 16;$$13$i31 = $280;$282 = $$pre$i30;$dmap$1$i33 = 531683;
     } else {
      $$1$i32 = 10;$$13$i31 = $$s$i29;$282 = $276;$dmap$1$i33 = 531715;
     }
    } while(0);
    $281 = $282&255;
    $283 = $281 >>> 3;
    $284 = (($dmap$1$i33) + ($283)|0);
    $285 = HEAP8[$284>>0]|0;
    $286 = $285&255;
    $287 = $282 << 24 >> 24;
    $288 = $287 & 7;
    $289 = 1 << $288;
    $290 = $286 & $289;
    $291 = ($290|0)==(0);
    if ($291) {
     $v$0$lcssa$i37 = 0;
    } else {
     $$21$i35 = $$13$i31;$294 = $287;$v$02$i34 = 0;
     while(1) {
      $292 = Math_imul($v$02$i34, $$1$i32)|0;
      $293 = (531747 + ($294)|0);
      $295 = HEAP8[$293>>0]|0;
      $296 = $295&255;
      $297 = (($296) + ($292))|0;
      $298 = ((($$21$i35)) + 1|0);
      $299 = HEAP8[$298>>0]|0;
      $300 = $299&255;
      $301 = $300 >>> 3;
      $302 = (($dmap$1$i33) + ($301)|0);
      $303 = HEAP8[$302>>0]|0;
      $304 = $303&255;
      $305 = $299 << 24 >> 24;
      $306 = $305 & 7;
      $307 = 1 << $306;
      $308 = $307 & $304;
      $309 = ($308|0)==(0);
      if ($309) {
       $v$0$lcssa$i37 = $297;
       break;
      } else {
       $$21$i35 = $298;$294 = $305;$v$02$i34 = $297;
      }
     }
    }
    $310 = (0 - ($v$0$lcssa$i37))|0;
    $311 = $274 ? $310 : $v$0$lcssa$i37;
    $$01$i39 = $311;
   }
  }
  $312 = (((($253) + ($254<<4)|0)) + 12|0);
  HEAP32[$312>>2] = $$01$i39;
  break;
 }
 case 3:  {
  $313 = ($value|0)==(0|0);
  L125: do {
   if ($313) {
    $$01$i27 = 0;
   } else {
    $314 = HEAP8[$value>>0]|0;
    L127: do {
     switch ($314<<24>>24) {
     case 0:  {
      $$01$i27 = 0;
      break L125;
      break;
     }
     case 48:  {
      $315 = ((($value)) + 1|0);
      $316 = HEAP8[$315>>0]|0;
      switch ($316<<24>>24) {
      case 88: case 120:  {
       break;
      }
      default: {
       $$02$i21 = $value;$$1$i22 = 10;$319 = 48;$dmap$1$i23 = 531715;
       break L127;
      }
      }
      $317 = ((($value)) + 2|0);
      $$pre$i20 = HEAP8[$317>>0]|0;
      $$02$i21 = $317;$$1$i22 = 16;$319 = $$pre$i20;$dmap$1$i23 = 531683;
      break;
     }
     default: {
      $$02$i21 = $value;$$1$i22 = 10;$319 = $314;$dmap$1$i23 = 531715;
     }
     }
    } while(0);
    $318 = $319&255;
    $320 = $318 >>> 3;
    $321 = (($dmap$1$i23) + ($320)|0);
    $322 = HEAP8[$321>>0]|0;
    $323 = $322&255;
    $324 = $319 << 24 >> 24;
    $325 = $324 & 7;
    $326 = 1 << $325;
    $327 = $323 & $326;
    $328 = ($327|0)==(0);
    if ($328) {
     $$01$i27 = 0;
    } else {
     $$131$i25 = $$02$i21;$331 = $324;$v$02$i24 = 0;
     while(1) {
      $329 = Math_imul($v$02$i24, $$1$i22)|0;
      $330 = (531747 + ($331)|0);
      $332 = HEAP8[$330>>0]|0;
      $333 = $332&255;
      $334 = (($333) + ($329))|0;
      $335 = ((($$131$i25)) + 1|0);
      $336 = HEAP8[$335>>0]|0;
      $337 = $336&255;
      $338 = $337 >>> 3;
      $339 = (($dmap$1$i23) + ($338)|0);
      $340 = HEAP8[$339>>0]|0;
      $341 = $340&255;
      $342 = $336 << 24 >> 24;
      $343 = $342 & 7;
      $344 = 1 << $343;
      $345 = $344 & $341;
      $346 = ($345|0)==(0);
      if ($346) {
       $$01$i27 = $334;
       break;
      } else {
       $$131$i25 = $335;$331 = $342;$v$02$i24 = $334;
      }
     }
    }
   }
  } while(0);
  $347 = (((($253) + ($254<<4)|0)) + 12|0);
  HEAP32[$347>>2] = $$01$i27;
  break;
 }
 default: {
 }
 }
 $348 = (_strncmp($name,530598,7)|0);
 $349 = ($348|0)==(0);
 if ($349) {
  $350 = ((($name)) + 7|0);
  $351 = HEAP8[$350>>0]|0;
  switch ($351<<24>>24) {
  case 9: case 13: case 10: case 0: case 32:  {
   break;
  }
  default: {
   label = 84;
  }
  }
 } else {
  label = 84;
 }
 if ((label|0) == 84) {
  $352 = HEAP32[$255>>2]|0;
  $353 = HEAP32[$224>>2]|0;
  $354 = HEAP32[$2>>2]|0;
  $355 = (_hash_insert($352,$353,$354,$1)|0);
  HEAP32[$error>>2] = $355;
  $356 = ($355|0)==(0);
  if (!($356)) {
   $391 = HEAP32[$error>>2]|0;
   STACKTOP = sp;return ($391|0);
  }
 }
 $357 = HEAP32[$224>>2]|0;
 $358 = (($357) + 1)|0;
 HEAP32[$224>>2] = $358;
 $359 = (_strncmp($name,530695,12)|0);
 $360 = ($359|0)==(0);
 L141: do {
  if ($360) {
   $361 = ((($name)) + 12|0);
   $362 = HEAP8[$361>>0]|0;
   switch ($362<<24>>24) {
   case 9: case 13: case 10: case 0: case 32:  {
    break;
   }
   default: {
    break L141;
   }
   }
   $363 = (((($253) + ($254<<4)|0)) + 12|0);
   $364 = HEAP32[$363>>2]|0;
   $365 = ((($font)) + 36|0);
   HEAP32[$365>>2] = $364;
   $391 = HEAP32[$error>>2]|0;
   STACKTOP = sp;return ($391|0);
  }
 } while(0);
 $366 = (_strncmp($name,530793,11)|0);
 $367 = ($366|0)==(0);
 L146: do {
  if ($367) {
   $368 = ((($name)) + 11|0);
   $369 = HEAP8[$368>>0]|0;
   switch ($369<<24>>24) {
   case 9: case 13: case 10: case 0: case 32:  {
    break;
   }
   default: {
    break L146;
   }
   }
   $370 = (((($253) + ($254<<4)|0)) + 12|0);
   $371 = HEAP32[$370>>2]|0;
   $372 = ((($font)) + 40|0);
   HEAP32[$372>>2] = $371;
   $391 = HEAP32[$error>>2]|0;
   STACKTOP = sp;return ($391|0);
  }
 } while(0);
 $373 = (_strncmp($name,530805,12)|0);
 $374 = ($373|0)==(0);
 L151: do {
  if ($374) {
   $375 = ((($name)) + 12|0);
   $376 = HEAP8[$375>>0]|0;
   switch ($376<<24>>24) {
   case 9: case 13: case 10: case 0: case 32:  {
    break;
   }
   default: {
    break L151;
   }
   }
   $377 = (((($253) + ($254<<4)|0)) + 12|0);
   $378 = HEAP32[$377>>2]|0;
   $379 = ((($font)) + 44|0);
   HEAP32[$379>>2] = $378;
   $391 = HEAP32[$error>>2]|0;
   STACKTOP = sp;return ($391|0);
  }
 } while(0);
 $380 = (_strncmp($name,531437,7)|0);
 $381 = ($380|0)==(0);
 if (!($381)) {
  $391 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($391|0);
 }
 $382 = ((($name)) + 7|0);
 $383 = HEAP8[$382>>0]|0;
 switch ($383<<24>>24) {
 case 9: case 13: case 10: case 0: case 32:  {
  break;
 }
 default: {
  $391 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($391|0);
 }
 }
 $384 = (((($253) + ($254<<4)|0)) + 12|0);
 $385 = HEAP32[$384>>2]|0;
 $386 = ($385|0)==(0|0);
 if ($386) {
  HEAP32[$error>>2] = 3;
  $391 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($391|0);
 }
 $387 = HEAP8[$385>>0]|0;
 switch ($387<<24>>24) {
 case 80: case 112:  {
  $388 = ((($font)) + 28|0);
  HEAP32[$388>>2] = 8;
  $391 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($391|0);
  break;
 }
 case 77: case 109:  {
  $389 = ((($font)) + 28|0);
  HEAP32[$389>>2] = 16;
  $391 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($391|0);
  break;
 }
 case 67: case 99:  {
  $390 = ((($font)) + 28|0);
  HEAP32[$390>>2] = 32;
  $391 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($391|0);
  break;
 }
 default: {
  $391 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($391|0);
 }
 }
 return (0)|0;
}
function __bdf_parse_glyphs($line,$linelen,$lineno,$call_data,$client_data) {
 $line = $line|0;
 $linelen = $linelen|0;
 $lineno = $lineno|0;
 $call_data = $call_data|0;
 $client_data = $client_data|0;
 var $$ = 0, $$0$i126 = 0, $$01$i106 = 0, $$01$i43 = 0, $$01$i83 = 0, $$01$i95 = 0, $$02$i = 0, $$02$i57 = 0, $$02$i65 = 0, $$02$i77 = 0, $$02$i89 = 0, $$07 = 0, $$07$v = 0, $$1$i = 0, $$1$i100 = 0, $$1$i110 = 0, $$1$i40 = 0, $$1$i47 = 0, $$1$i58 = 0, $$1$i66 = 0;
 var $$1$i78 = 0, $$1$i90 = 0, $$13$i = 0, $$13$i109 = 0, $$13$i46 = 0, $$13$i99 = 0, $$131$i = 0, $$131$i61 = 0, $$131$i69 = 0, $$131$i81 = 0, $$131$i93 = 0, $$21$i = 0, $$21$i103 = 0, $$21$i113 = 0, $$21$i49 = 0, $$31 = 0, $$32 = 0, $$33 = 0, $$lcssa = 0, $$lcssa$i = 0;
 var $$lcssa$i52 = 0, $$lcssa244 = 0, $$lcssa245 = 0, $$lcssa247 = 0, $$lcssa251 = 0, $$lcssa256 = 0, $$phi$trans$insert = 0, $$phi$trans$insert$i = 0, $$phi$trans$insert175 = 0, $$pr129 = 0, $$pr129192195 = 0, $$pr134 = 0, $$pr139$pre = 0, $$pre = 0, $$pre$i = 0, $$pre$i108 = 0, $$pre$i38 = 0, $$pre$i39 = 0, $$pre$i45 = 0, $$pre$i56 = 0;
 var $$pre$i64 = 0, $$pre$i76 = 0, $$pre$i87 = 0, $$pre$i88 = 0, $$pre$i98 = 0, $$pre$phi183Z2D = 0, $$pre$phi187Z2D = 0, $$pre10$i = 0, $$pre11$i = 0, $$pre172 = 0, $$pre173 = 0, $$pre176 = 0, $$pre177 = 0, $$pre178 = 0, $$pre180 = 0, $$pre181 = 0, $$pre9$i = 0, $$s$i = 0, $$s$i107 = 0, $$s$i44 = 0;
 var $$s$i97 = 0, $$sink1 = 0, $$sink3 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0;
 var $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0;
 var $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0;
 var $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0;
 var $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0;
 var $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0;
 var $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0;
 var $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0;
 var $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0;
 var $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0;
 var $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0;
 var $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0;
 var $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0;
 var $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0;
 var $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0;
 var $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0;
 var $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0;
 var $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0;
 var $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0;
 var $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0;
 var $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0;
 var $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0;
 var $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0;
 var $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0;
 var $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0;
 var $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0;
 var $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0;
 var $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0;
 var $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0;
 var $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0;
 var $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0;
 var $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0;
 var $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0;
 var $834 = 0, $835 = 0, $836 = 0, $837 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $bp$0$ph = 0, $bp$0$ph$lcssa252 = 0, $bp$0$ph253 = 0, $dmap$1$i = 0, $dmap$1$i101 = 0, $dmap$1$i111 = 0, $dmap$1$i41 = 0, $dmap$1$i48 = 0, $dmap$1$i59 = 0, $dmap$1$i67 = 0, $dmap$1$i79 = 0, $dmap$1$i91 = 0, $error = 0, $error$i = 0, $exitcond = 0, $fp$02$i = 0, $glyph$0 = 0, $i$0 = 0, $i$0$lcssa248 = 0;
 var $i$0$ph = 0, $i$01$i = 0, $i$0250 = 0, $i$03$i = 0, $j$0$be$i = 0, $j$0$be$i$lcssa = 0, $j$04$i = 0, $j$1$lcssa$i = 0, $j$11$i = 0, $or$cond = 0, $or$cond21 = 0, $or$cond23 = 0, $phitmp = 0, $s$0 = 0, $sext = 0, $sext$i = 0, $sext$i50 = 0, $u$02$i = 0, $v$0$lcssa$i = 0, $v$0$lcssa$i115 = 0;
 var $v$02$i = 0, $v$02$i102 = 0, $v$02$i112 = 0, $v$02$i60 = 0, $v$02$i68 = 0, $v$02$i80 = 0, $v$02$i92 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp + 4|0;
 $error = sp;
 HEAP32[$error>>2] = 0;
 $0 = ((($client_data)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 278664|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_strncmp($line,530598,7)|0);
 $5 = ($4|0)==(0);
 L1: do {
  if ($5) {
   $6 = ((($line)) + 7|0);
   $7 = HEAP8[$6>>0]|0;
   switch ($7<<24>>24) {
   case 9: case 13: case 10: case 0: case 32:  {
    break;
   }
   default: {
    label = 6;
    break L1;
   }
   }
   $8 = ($7<<24>>24)==(0);
   $$07$v = $8 ? -7 : -8;
   $$07 = (($$07$v) + ($linelen))|0;
   HEAP32[$error$i>>2] = 0;
   $9 = ((($1)) + 88|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = (($$07) + 1)|0;
   $12 = (($10) + ($11))|0;
   $13 = ((($1)) + 84|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = (_ft_mem_realloc($3,1,$10,$12,$14,$error$i)|0);
   HEAP32[$13>>2] = $15;
   $16 = HEAP32[$error$i>>2]|0;
   $17 = ($16|0)==(0);
   if ($17) {
    $18 = ((($line)) + 8|0);
    $s$0 = $8 ? $6 : $18;
    $19 = HEAP32[$9>>2]|0;
    $20 = (($15) + ($19)|0);
    _memcpy(($20|0),($s$0|0),($$07|0))|0;
    $21 = (($20) + ($$07)|0);
    HEAP8[$21>>0] = 10;
    $22 = HEAP32[$9>>2]|0;
    $23 = (($22) + ($11))|0;
    HEAP32[$9>>2] = $23;
    $$pre$i = HEAP32[$error$i>>2]|0;
    $24 = $$pre$i;
   } else {
    $24 = $16;
   }
   HEAP32[$error>>2] = $24;
   $829 = $24;
   label = 204;
  } else {
   label = 6;
  }
 } while(0);
 L7: do {
  if ((label|0) == 6) {
   $25 = HEAP32[$client_data>>2]|0;
   $26 = $25 & 32;
   $27 = ($26|0)==(0);
   L9: do {
    if ($27) {
     $28 = (_strncmp($line,531893,5)|0);
     $29 = ($28|0)==(0);
     L11: do {
      if ($29) {
       $30 = ((($line)) + 5|0);
       $31 = HEAP8[$30>>0]|0;
       switch ($31<<24>>24) {
       case 9: case 13: case 10: case 0: case 32:  {
        break;
       }
       default: {
        break L11;
       }
       }
       $32 = ((($client_data)) + 139304|0);
       $33 = (__bdf_list_split($32,531680,$line,$linelen)|0);
       HEAP32[$error>>2] = $33;
       $34 = ($33|0)==(0);
       if (!($34)) {
        $$pre$phi187Z2D = $client_data;$836 = $33;
        break L7;
       }
       $35 = HEAP32[$32>>2]|0;
       $36 = ((($35)) + 4|0);
       $37 = HEAP32[$36>>2]|0;
       $38 = ($37|0)==(0|0);
       L15: do {
        if ($38) {
         label = 17;
        } else {
         $39 = HEAP8[$37>>0]|0;
         L17: do {
          switch ($39<<24>>24) {
          case 0:  {
           label = 17;
           break L15;
           break;
          }
          case 48:  {
           $40 = ((($37)) + 1|0);
           $41 = HEAP8[$40>>0]|0;
           switch ($41<<24>>24) {
           case 88: case 120:  {
            break;
           }
           default: {
            $$02$i = $37;$$1$i = 10;$44 = 48;$dmap$1$i = 531715;
            break L17;
           }
           }
           $42 = ((($37)) + 2|0);
           $$pre$i38 = HEAP8[$42>>0]|0;
           $$02$i = $42;$$1$i = 16;$44 = $$pre$i38;$dmap$1$i = 531683;
           break;
          }
          default: {
           $$02$i = $37;$$1$i = 10;$44 = $39;$dmap$1$i = 531715;
          }
          }
         } while(0);
         $43 = $44&255;
         $45 = $43 >>> 3;
         $46 = (($dmap$1$i) + ($45)|0);
         $47 = HEAP8[$46>>0]|0;
         $48 = $47&255;
         $49 = $44 << 24 >> 24;
         $50 = $49 & 7;
         $51 = 1 << $50;
         $52 = $48 & $51;
         $53 = ($52|0)==(0);
         if ($53) {
          label = 17;
         } else {
          $$131$i = $$02$i;$56 = $49;$v$02$i = 0;
          while(1) {
           $54 = Math_imul($v$02$i, $$1$i)|0;
           $55 = (531747 + ($56)|0);
           $57 = HEAP8[$55>>0]|0;
           $58 = $57&255;
           $59 = (($58) + ($54))|0;
           $60 = ((($$131$i)) + 1|0);
           $61 = HEAP8[$60>>0]|0;
           $62 = $61&255;
           $63 = $62 >>> 3;
           $64 = (($dmap$1$i) + ($63)|0);
           $65 = HEAP8[$64>>0]|0;
           $66 = $65&255;
           $67 = $61 << 24 >> 24;
           $68 = $67 & 7;
           $69 = 1 << $68;
           $70 = $69 & $66;
           $71 = ($70|0)==(0);
           if ($71) {
            $$lcssa = $59;
            break;
           } else {
            $$131$i = $60;$56 = $67;$v$02$i = $59;
           }
          }
          $74 = ((($1)) + 48|0);
          HEAP32[$74>>2] = $$lcssa;
          $75 = ((($client_data)) + 4|0);
          HEAP32[$75>>2] = $$lcssa;
          $76 = ($$lcssa|0)==(0);
          if ($76) {
           $77 = $74;
           label = 19;
          } else {
           $78 = ($$lcssa>>>0)>(1114111);
           if ($78) {
            HEAP32[$error>>2] = 6;
            $$pre$phi187Z2D = $client_data;$836 = 6;
            break L7;
           } else {
            $79 = $$lcssa;
           }
          }
         }
        }
       } while(0);
       if ((label|0) == 17) {
        $72 = ((($1)) + 48|0);
        HEAP32[$72>>2] = 0;
        $73 = ((($client_data)) + 4|0);
        HEAP32[$73>>2] = 0;
        $77 = $72;
        label = 19;
       }
       if ((label|0) == 19) {
        HEAP32[$77>>2] = 64;
        $79 = 64;
       }
       $80 = (_ft_mem_realloc($3,36,0,$79,0,$error)|0);
       $81 = ((($1)) + 56|0);
       HEAP32[$81>>2] = $80;
       $82 = HEAP32[$error>>2]|0;
       $83 = ($82|0)==(0);
       if (!($83)) {
        $$pre$phi187Z2D = $client_data;$836 = $82;
        break L7;
       }
       $84 = HEAP32[$client_data>>2]|0;
       $85 = $84 | 32;
       HEAP32[$client_data>>2] = $85;
       break L9;
      }
     } while(0);
     HEAP32[$error>>2] = 180;
     $$pre$phi187Z2D = $client_data;$836 = 180;
     break L7;
    } else {
     $86 = (_strncmp($line,531899,7)|0);
     $87 = ($86|0)==(0);
     L34: do {
      if ($87) {
       $88 = ((($line)) + 7|0);
       $89 = HEAP8[$88>>0]|0;
       switch ($89<<24>>24) {
       case 9: case 13: case 10: case 0: case 32:  {
        break;
       }
       default: {
        break L34;
       }
       }
       $90 = $25 & 4032;
       $91 = ($90|0)==(0);
       if ($91) {
        $92 = ((($1)) + 56|0);
        $93 = HEAP32[$92>>2]|0;
        $94 = ((($1)) + 52|0);
        $95 = HEAP32[$94>>2]|0;
        _qsort($93,$95,36,4355);
        $96 = HEAP32[$client_data>>2]|0;
        $97 = $96 & -2;
        HEAP32[$client_data>>2] = $97;
        break L9;
       } else {
        HEAP32[$error>>2] = 186;
        $$pre$phi187Z2D = $client_data;$836 = 186;
        break L7;
       }
      }
     } while(0);
     $98 = (_strncmp($line,531907,7)|0);
     $99 = ($98|0)==(0);
     L41: do {
      if ($99) {
       $100 = ((($line)) + 7|0);
       $101 = HEAP8[$100>>0]|0;
       switch ($101<<24>>24) {
       case 9: case 13: case 10: case 0: case 32:  {
        break;
       }
       default: {
        break L41;
       }
       }
       $102 = ((($client_data)) + 28|0);
       HEAP32[$102>>2] = 0;
       $103 = $25 & -4033;
       HEAP32[$client_data>>2] = $103;
       break L9;
      }
     } while(0);
     $104 = $25 & 64;
     $105 = ($104|0)==(0);
     if (!($105)) {
      $106 = ((($client_data)) + 28|0);
      $107 = HEAP32[$106>>2]|0;
      $108 = ($107|0)==(-1);
      if ($108) {
       $109 = ((($client_data)) + 36|0);
       $110 = HEAP32[$109>>2]|0;
       $111 = ((($110)) + 4|0);
       $112 = HEAP32[$111>>2]|0;
       $113 = ($112|0)==(0);
       if ($113) {
        break;
       }
      }
     }
     $114 = (_strncmp($line,531915,9)|0);
     $115 = ($114|0)==(0);
     L49: do {
      if ($115) {
       $116 = ((($line)) + 9|0);
       $117 = HEAP8[$116>>0]|0;
       switch ($117<<24>>24) {
       case 9: case 13: case 10: case 0: case 32:  {
        break;
       }
       default: {
        break L49;
       }
       }
       $118 = ((($client_data)) + 24|0);
       $119 = HEAP32[$118>>2]|0;
       _ft_mem_free($3,$119);
       HEAP32[$118>>2] = 0;
       $120 = ((($client_data)) + 139304|0);
       $121 = (__bdf_list_split($120,531680,$line,$linelen)|0);
       HEAP32[$error>>2] = $121;
       $122 = ($121|0)==(0);
       if (!($122)) {
        $$pre$phi187Z2D = $client_data;$836 = $121;
        break L7;
       }
       $123 = ((($client_data)) + 139312|0);
       $124 = HEAP32[$123>>2]|0;
       $125 = ($124|0)==(0);
       do {
        if (!($125)) {
         $126 = ($124>>>0)<(2);
         if ($126) {
          HEAP32[$123>>2] = 0;
          break;
         }
         $127 = (($124) + -1)|0;
         $i$01$i = 0;$u$02$i = 1;
         while(1) {
          $128 = HEAP32[$120>>2]|0;
          $129 = (($128) + ($u$02$i<<2)|0);
          $130 = HEAP32[$129>>2]|0;
          $131 = (($128) + ($i$01$i<<2)|0);
          HEAP32[$131>>2] = $130;
          $132 = (($i$01$i) + 1)|0;
          $133 = (($u$02$i) + 1)|0;
          $exitcond = ($132|0)==($127|0);
          if ($exitcond) {
           break;
          } else {
           $i$01$i = $132;$u$02$i = $133;
          }
         }
         HEAP32[$123>>2] = $127;
         $134 = ($127|0)==(0);
         if (!($134)) {
          $135 = HEAP32[$120>>2]|0;
          $136 = HEAP32[$135>>2]|0;
          $138 = $136;$837 = $127;$i$03$i = 0;$j$04$i = 0;
          while(1) {
           $137 = HEAP8[$138>>0]|0;
           $139 = ($137<<24>>24)==(0);
           if ($139) {
            $148 = $837;$j$1$lcssa$i = $j$04$i;
           } else {
            $143 = $137;$fp$02$i = $138;$j$11$i = $j$04$i;
            while(1) {
             $140 = ((($fp$02$i)) + 1|0);
             $141 = (($j$11$i) + 1)|0;
             $142 = (($136) + ($j$11$i)|0);
             HEAP8[$142>>0] = $143;
             $144 = HEAP8[$140>>0]|0;
             $145 = ($144<<24>>24)==(0);
             if ($145) {
              $$lcssa256 = $141;
              break;
             } else {
              $143 = $144;$fp$02$i = $140;$j$11$i = $141;
             }
            }
            $$pre10$i = HEAP32[$123>>2]|0;
            $148 = $$pre10$i;$j$1$lcssa$i = $$lcssa256;
           }
           $146 = (($i$03$i) + 1)|0;
           $147 = ($146>>>0)<($148>>>0);
           if ($147) {
            $149 = (($j$1$lcssa$i) + 1)|0;
            $150 = (($136) + ($j$1$lcssa$i)|0);
            HEAP8[$150>>0] = 32;
            $$pre11$i = HEAP32[$123>>2]|0;
            $152 = $$pre11$i;$j$0$be$i = $149;
           } else {
            $152 = $148;$j$0$be$i = $j$1$lcssa$i;
           }
           $151 = ($146>>>0)<($152>>>0);
           if (!($151)) {
            $j$0$be$i$lcssa = $j$0$be$i;
            break;
           }
           $$pre$i87 = HEAP32[$120>>2]|0;
           $$phi$trans$insert$i = (($$pre$i87) + ($146<<2)|0);
           $$pre9$i = HEAP32[$$phi$trans$insert$i>>2]|0;
           $138 = $$pre9$i;$837 = $152;$i$03$i = $146;$j$04$i = $j$0$be$i;
          }
          $153 = ($136|0)==(1282466|0);
          if ($153) {
           $$0$i126 = 1282466;
          } else {
           $154 = (($136) + ($j$0$be$i$lcssa)|0);
           HEAP8[$154>>0] = 0;
           $155 = ($136|0)==(0|0);
           if ($155) {
            break;
           } else {
            $$0$i126 = $136;
           }
          }
          $156 = (($j$0$be$i$lcssa) + 1)|0;
          $157 = (_ft_mem_realloc($3,1,0,$156,0,$error)|0);
          HEAP32[$118>>2] = $157;
          $158 = HEAP32[$error>>2]|0;
          $159 = ($158|0)==(0);
          if (!($159)) {
           $$pre$phi187Z2D = $client_data;$836 = $158;
           break L7;
          }
          _memcpy(($157|0),($$0$i126|0),($156|0))|0;
          $160 = HEAP32[$client_data>>2]|0;
          $161 = $160 | 64;
          HEAP32[$client_data>>2] = $161;
          break L9;
         }
        }
       } while(0);
       HEAP32[$error>>2] = 3;
       $$pre$phi187Z2D = $client_data;$836 = 3;
       break L7;
      }
     } while(0);
     $162 = (_strncmp($line,531925,8)|0);
     $163 = ($162|0)==(0);
     L80: do {
      if ($163) {
       $164 = ((($line)) + 8|0);
       $165 = HEAP8[$164>>0]|0;
       switch ($165<<24>>24) {
       case 9: case 13: case 10: case 0: case 32:  {
        break;
       }
       default: {
        break L80;
       }
       }
       if ($105) {
        HEAP32[$error>>2] = 181;
        $$pre$phi187Z2D = $client_data;$836 = 181;
        break L7;
       }
       $166 = ((($client_data)) + 139304|0);
       $167 = (__bdf_list_split($166,531680,$line,$linelen)|0);
       HEAP32[$error>>2] = $167;
       $168 = ($167|0)==(0);
       if (!($168)) {
        $$pre$phi187Z2D = $client_data;$836 = $167;
        break L7;
       }
       $169 = HEAP32[$166>>2]|0;
       $170 = ((($169)) + 4|0);
       $171 = HEAP32[$170>>2]|0;
       $172 = ($171|0)==(0|0);
       if ($172) {
        $$01$i106 = 0;
       } else {
        $173 = HEAP8[$171>>0]|0;
        $174 = ($173<<24>>24)==(0);
        if ($174) {
         $$01$i106 = 0;
        } else {
         $175 = ($173<<24>>24)==(45);
         $176 = ((($171)) + 1|0);
         $$s$i97 = $175 ? $176 : $171;
         $177 = HEAP8[$$s$i97>>0]|0;
         $178 = ($177<<24>>24)==(48);
         L90: do {
          if ($178) {
           $179 = ((($$s$i97)) + 1|0);
           $180 = HEAP8[$179>>0]|0;
           switch ($180<<24>>24) {
           case 88: case 120:  {
            break;
           }
           default: {
            $$1$i100 = 10;$$13$i99 = $$s$i97;$183 = 48;$dmap$1$i101 = 531715;
            break L90;
           }
           }
           $181 = ((($$s$i97)) + 2|0);
           $$pre$i98 = HEAP8[$181>>0]|0;
           $$1$i100 = 16;$$13$i99 = $181;$183 = $$pre$i98;$dmap$1$i101 = 531683;
          } else {
           $$1$i100 = 10;$$13$i99 = $$s$i97;$183 = $177;$dmap$1$i101 = 531715;
          }
         } while(0);
         $182 = $183&255;
         $184 = $182 >>> 3;
         $185 = (($dmap$1$i101) + ($184)|0);
         $186 = HEAP8[$185>>0]|0;
         $187 = $186&255;
         $188 = $183 << 24 >> 24;
         $189 = $188 & 7;
         $190 = 1 << $189;
         $191 = $187 & $190;
         $192 = ($191|0)==(0);
         if ($192) {
          $v$0$lcssa$i = 0;
         } else {
          $$21$i103 = $$13$i99;$195 = $188;$v$02$i102 = 0;
          while(1) {
           $193 = Math_imul($v$02$i102, $$1$i100)|0;
           $194 = (531747 + ($195)|0);
           $196 = HEAP8[$194>>0]|0;
           $197 = $196&255;
           $198 = (($197) + ($193))|0;
           $199 = ((($$21$i103)) + 1|0);
           $200 = HEAP8[$199>>0]|0;
           $201 = $200&255;
           $202 = $201 >>> 3;
           $203 = (($dmap$1$i101) + ($202)|0);
           $204 = HEAP8[$203>>0]|0;
           $205 = $204&255;
           $206 = $200 << 24 >> 24;
           $207 = $206 & 7;
           $208 = 1 << $207;
           $209 = $208 & $205;
           $210 = ($209|0)==(0);
           if ($210) {
            $v$0$lcssa$i = $198;
            break;
           } else {
            $$21$i103 = $199;$195 = $206;$v$02$i102 = $198;
           }
          }
         }
         $211 = (0 - ($v$0$lcssa$i))|0;
         $212 = $175 ? $211 : $v$0$lcssa$i;
         $$01$i106 = $212;
        }
       }
       $213 = ((($client_data)) + 28|0);
       $214 = ($$01$i106|0)<(-1);
       $$ = $214 ? -1 : $$01$i106;
       HEAP32[$213>>2] = $$;
       $215 = ($$|0)==(-1);
       do {
        if ($215) {
         $216 = ((($client_data)) + 139312|0);
         $217 = HEAP32[$216>>2]|0;
         $218 = ($217>>>0)>(2);
         if ($218) {
          $219 = ((($169)) + 8|0);
          $220 = HEAP32[$219>>2]|0;
          $221 = ($220|0)==(0|0);
          if (!($221)) {
           $222 = HEAP8[$220>>0]|0;
           $223 = ($222<<24>>24)==(0);
           if (!($223)) {
            $224 = ($222<<24>>24)==(45);
            $225 = ((($220)) + 1|0);
            $$s$i107 = $224 ? $225 : $220;
            $226 = HEAP8[$$s$i107>>0]|0;
            $227 = ($226<<24>>24)==(48);
            L105: do {
             if ($227) {
              $228 = ((($$s$i107)) + 1|0);
              $229 = HEAP8[$228>>0]|0;
              switch ($229<<24>>24) {
              case 88: case 120:  {
               break;
              }
              default: {
               $$1$i110 = 10;$$13$i109 = $$s$i107;$232 = 48;$dmap$1$i111 = 531715;
               break L105;
              }
              }
              $230 = ((($$s$i107)) + 2|0);
              $$pre$i108 = HEAP8[$230>>0]|0;
              $$1$i110 = 16;$$13$i109 = $230;$232 = $$pre$i108;$dmap$1$i111 = 531683;
             } else {
              $$1$i110 = 10;$$13$i109 = $$s$i107;$232 = $226;$dmap$1$i111 = 531715;
             }
            } while(0);
            $231 = $232&255;
            $233 = $231 >>> 3;
            $234 = (($dmap$1$i111) + ($233)|0);
            $235 = HEAP8[$234>>0]|0;
            $236 = $235&255;
            $237 = $232 << 24 >> 24;
            $238 = $237 & 7;
            $239 = 1 << $238;
            $240 = $236 & $239;
            $241 = ($240|0)==(0);
            if ($241) {
             $v$0$lcssa$i115 = 0;
            } else {
             $$21$i113 = $$13$i109;$244 = $237;$v$02$i112 = 0;
             while(1) {
              $242 = Math_imul($v$02$i112, $$1$i110)|0;
              $243 = (531747 + ($244)|0);
              $245 = HEAP8[$243>>0]|0;
              $246 = $245&255;
              $247 = (($246) + ($242))|0;
              $248 = ((($$21$i113)) + 1|0);
              $249 = HEAP8[$248>>0]|0;
              $250 = $249&255;
              $251 = $250 >>> 3;
              $252 = (($dmap$1$i111) + ($251)|0);
              $253 = HEAP8[$252>>0]|0;
              $254 = $253&255;
              $255 = $249 << 24 >> 24;
              $256 = $255 & 7;
              $257 = 1 << $256;
              $258 = $257 & $254;
              $259 = ($258|0)==(0);
              if ($259) {
               $v$0$lcssa$i115 = $247;
               break;
              } else {
               $$21$i113 = $248;$244 = $255;$v$02$i112 = $247;
              }
             }
            }
            $260 = (0 - ($v$0$lcssa$i115))|0;
            $261 = $224 ? $260 : $v$0$lcssa$i115;
            HEAP32[$213>>2] = $261;
            $262 = ($261|0)<(-1);
            if (!($262)) {
             $$pr129 = $261;
             label = 82;
             break;
            }
            HEAP32[$213>>2] = -1;
            label = 93;
            break;
           }
          }
          HEAP32[$213>>2] = 0;
          $$pr129192195 = 0;
          label = 85;
         } else {
          label = 93;
         }
        } else {
         $$pr129 = $$;
         label = 82;
        }
       } while(0);
       do {
        if ((label|0) == 82) {
         $263 = ($$pr129|0)>(0);
         $264 = ($$pr129>>>0)>(1114111);
         $or$cond21 = $263 & $264;
         if ($or$cond21) {
          HEAP32[$error>>2] = 3;
          $$pre$phi187Z2D = $client_data;$836 = 3;
          break L7;
         } else {
          $265 = ($$pr129|0)>(-1);
          if ($265) {
           $$pr129192195 = $$pr129;
           label = 85;
           break;
          } else {
           label = 93;
           break;
          }
         }
        }
       } while(0);
       do {
        if ((label|0) == 85) {
         $266 = $$pr129192195 >> 5;
         $267 = ((($client_data)) + 40|0);
         $268 = (($267) + ($266<<2)|0);
         $269 = HEAP32[$268>>2]|0;
         $270 = $$pr129192195 & 31;
         $271 = 1 << $270;
         $272 = $269 & $271;
         $273 = ($272|0)==(0);
         if (!($273)) {
          HEAP32[$213>>2] = -1;
          $274 = ((($1)) + 278660|0);
          HEAP16[$274>>1] = 1;
          label = 93;
          break;
         }
         $275 = $269 | $271;
         HEAP32[$268>>2] = $275;
         $$pr139$pre = HEAP32[$213>>2]|0;
         $276 = ($$pr139$pre|0)>(-1);
         if ($276) {
          $277 = ((($1)) + 52|0);
          $278 = HEAP32[$277>>2]|0;
          $279 = ((($1)) + 48|0);
          $280 = HEAP32[$279>>2]|0;
          $281 = ($278|0)==($280|0);
          if ($281) {
           $282 = (($278) + 64)|0;
           $283 = ((($1)) + 56|0);
           $284 = HEAP32[$283>>2]|0;
           $285 = (_ft_mem_realloc($3,36,$278,$282,$284,$error)|0);
           HEAP32[$283>>2] = $285;
           $286 = HEAP32[$error>>2]|0;
           $287 = ($286|0)==(0);
           if (!($287)) {
            $$pre$phi187Z2D = $client_data;$836 = $286;
            break L7;
           }
           $288 = HEAP32[$279>>2]|0;
           $289 = (($288) + 64)|0;
           HEAP32[$279>>2] = $289;
           $$pre172 = HEAP32[$277>>2]|0;
           $$pre173 = HEAP32[$213>>2]|0;
           $291 = $$pre172;$293 = $285;$297 = $$pre173;
          } else {
           $$phi$trans$insert = ((($1)) + 56|0);
           $$pre = HEAP32[$$phi$trans$insert>>2]|0;
           $291 = $278;$293 = $$pre;$297 = $$pr139$pre;
          }
          $290 = (($291) + 1)|0;
          HEAP32[$277>>2] = $290;
          $292 = (($293) + (($291*36)|0)|0);
          $294 = ((($client_data)) + 24|0);
          $295 = HEAP32[$294>>2]|0;
          HEAP32[$292>>2] = $295;
          $296 = (((($293) + (($291*36)|0)|0)) + 4|0);
          HEAP32[$296>>2] = $297;
          HEAP32[$294>>2] = 0;
         } else {
          label = 93;
         }
        }
       } while(0);
       if ((label|0) == 93) {
        $298 = ((($client_data)) + 36|0);
        $299 = HEAP32[$298>>2]|0;
        $300 = ((($299)) + 4|0);
        $301 = HEAP32[$300>>2]|0;
        $302 = ($301|0)==(0);
        if ($302) {
         $323 = ((($client_data)) + 24|0);
         $324 = HEAP32[$323>>2]|0;
         _ft_mem_free($3,$324);
         HEAP32[$323>>2] = 0;
         $$pre$phi183Z2D = $323;
        } else {
         $303 = ((($1)) + 64|0);
         $304 = HEAP32[$303>>2]|0;
         $305 = ((($1)) + 60|0);
         $306 = HEAP32[$305>>2]|0;
         $307 = ($304|0)==($306|0);
         if ($307) {
          $308 = (($304) + 4)|0;
          $309 = ((($1)) + 68|0);
          $310 = HEAP32[$309>>2]|0;
          $311 = (_ft_mem_realloc($3,36,$304,$308,$310,$error)|0);
          HEAP32[$309>>2] = $311;
          $312 = HEAP32[$error>>2]|0;
          $313 = ($312|0)==(0);
          if (!($313)) {
           $$pre$phi187Z2D = $client_data;$836 = $312;
           break L7;
          }
          $314 = HEAP32[$305>>2]|0;
          $315 = (($314) + 4)|0;
          HEAP32[$305>>2] = $315;
          $$pre177 = HEAP32[$303>>2]|0;
          $317 = $311;$318 = $$pre177;
         } else {
          $$phi$trans$insert175 = ((($1)) + 68|0);
          $$pre176 = HEAP32[$$phi$trans$insert175>>2]|0;
          $317 = $$pre176;$318 = $304;
         }
         $316 = (($317) + (($318*36)|0)|0);
         $319 = ((($client_data)) + 24|0);
         $320 = HEAP32[$319>>2]|0;
         HEAP32[$316>>2] = $320;
         $321 = (($318) + 1)|0;
         HEAP32[$303>>2] = $321;
         $322 = (((($317) + (($318*36)|0)|0)) + 4|0);
         HEAP32[$322>>2] = $318;
         $$pre$phi183Z2D = $319;
        }
        HEAP32[$$pre$phi183Z2D>>2] = 0;
       }
       $325 = HEAP32[$client_data>>2]|0;
       $326 = $325 & 1073741695;
       $327 = $326 | 128;
       HEAP32[$client_data>>2] = $327;
       break L9;
      }
     } while(0);
     $328 = ((($client_data)) + 28|0);
     $329 = HEAP32[$328>>2]|0;
     $330 = ($329|0)==(-1);
     if ($330) {
      $331 = ((($1)) + 68|0);
      $332 = HEAP32[$331>>2]|0;
      $333 = ((($1)) + 64|0);
      $334 = HEAP32[$333>>2]|0;
      $335 = (($334) + -1)|0;
      $336 = (($332) + (($335*36)|0)|0);
      $glyph$0 = $336;
     } else {
      $337 = ((($1)) + 56|0);
      $338 = HEAP32[$337>>2]|0;
      $339 = ((($1)) + 52|0);
      $340 = HEAP32[$339>>2]|0;
      $341 = (($340) + -1)|0;
      $342 = (($338) + (($341*36)|0)|0);
      $glyph$0 = $342;
     }
     $343 = $25 & 2048;
     $344 = ($343|0)==(0);
     if (!($344)) {
      $345 = ((($client_data)) + 8|0);
      $346 = HEAP32[$345>>2]|0;
      $347 = ((($glyph$0)) + 14|0);
      $348 = HEAP16[$347>>1]|0;
      $349 = $348&65535;
      $350 = ($346>>>0)<($349>>>0);
      if (!($350)) {
       $351 = ($25|0)<(0);
       if ($351) {
        break;
       }
       $352 = $25 | -2147483648;
       HEAP32[$client_data>>2] = $352;
       $353 = ((($1)) + 278660|0);
       HEAP16[$353>>1] = 1;
       break;
      }
      $354 = ((($glyph$0)) + 28|0);
      $355 = HEAP32[$354>>2]|0;
      $356 = $355 << 1;
      $357 = ((($glyph$0)) + 24|0);
      $358 = HEAP32[$357>>2]|0;
      $359 = Math_imul($355, $346)|0;
      $360 = (($358) + ($359)|0);
      $bp$0$ph = $360;$i$0$ph = 0;
      L154: while(1) {
       $i$0 = $i$0$ph;
       while(1) {
        $361 = ($i$0>>>0)<($356>>>0);
        if (!($361)) {
         $bp$0$ph253 = $bp$0$ph;$i$0250 = $i$0;
         break L154;
        }
        $362 = (($line) + ($i$0)|0);
        $363 = HEAP8[$362>>0]|0;
        $364 = $363 << 24 >> 24;
        $365 = $363&255;
        $366 = $365 >>> 3;
        $367 = (531683 + ($366)|0);
        $368 = HEAP8[$367>>0]|0;
        $369 = $368&255;
        $370 = $364 & 7;
        $371 = 1 << $370;
        $372 = $371 & $369;
        $373 = ($372|0)==(0);
        if ($373) {
         $bp$0$ph$lcssa252 = $bp$0$ph;$i$0$lcssa248 = $i$0;
         label = 115;
         break L154;
        }
        $374 = HEAP8[$bp$0$ph>>0]|0;
        $375 = $374&255;
        $376 = $375 << 4;
        $377 = (531747 + ($364)|0);
        $378 = HEAP8[$377>>0]|0;
        $379 = $378&255;
        $380 = (($376) + ($379))|0;
        $381 = $380&255;
        HEAP8[$bp$0$ph>>0] = $381;
        $382 = (($i$0) + 1)|0;
        $383 = ($382>>>0)>=($356>>>0);
        $384 = $i$0 & 1;
        $385 = ($384|0)==(0);
        $or$cond23 = $383 | $385;
        if ($or$cond23) {
         $i$0 = $382;
        } else {
         $$lcssa251 = $382;
         break;
        }
       }
       $386 = ((($bp$0$ph)) + 1|0);
       HEAP8[$386>>0] = 0;
       $bp$0$ph = $386;$i$0$ph = $$lcssa251;
      }
      if ((label|0) == 115) {
       $387 = HEAP32[$client_data>>2]|0;
       $388 = $387 & 1073741824;
       $389 = ($388|0)==(0);
       if ($389) {
        $390 = $387 | 1073741824;
        HEAP32[$client_data>>2] = $390;
        $391 = ((($1)) + 278660|0);
        HEAP16[$391>>1] = 1;
        $bp$0$ph253 = $bp$0$ph$lcssa252;$i$0250 = $i$0$lcssa248;
       } else {
        $bp$0$ph253 = $bp$0$ph$lcssa252;$i$0250 = $i$0$lcssa248;
       }
      }
      $392 = ((($glyph$0)) + 12|0);
      $393 = HEAP16[$392>>1]|0;
      $394 = ($393<<16>>16)==(0);
      if (!($394)) {
       $395 = $393&65535;
       $396 = HEAP32[$0>>2]|0;
       $397 = ((($396)) + 278662|0);
       $398 = HEAP16[$397>>1]|0;
       $399 = $398&65535;
       $400 = Math_imul($399, $395)|0;
       $401 = $400 & 7;
       $402 = (531934 + ($401)|0);
       $403 = HEAP8[$402>>0]|0;
       $404 = HEAP8[$bp$0$ph253>>0]|0;
       $405 = $404 & $403;
       HEAP8[$bp$0$ph253>>0] = $405;
      }
      $406 = ($i$0250|0)==($356|0);
      if ($406) {
       $407 = (($line) + ($356)|0);
       $408 = HEAP8[$407>>0]|0;
       $409 = $408&255;
       $410 = $409 >>> 3;
       $411 = (531683 + ($410)|0);
       $412 = HEAP8[$411>>0]|0;
       $413 = $412&255;
       $414 = $409 & 7;
       $415 = 1 << $414;
       $416 = $415 & $413;
       $417 = ($416|0)==(0);
       if (!($417)) {
        $418 = HEAP32[$client_data>>2]|0;
        $419 = $418 & 1073741824;
        $420 = ($419|0)==(0);
        if ($420) {
         $421 = $418 | 1073741824;
         HEAP32[$client_data>>2] = $421;
         $422 = ((($1)) + 278660|0);
         HEAP16[$422>>1] = 1;
        }
       }
      }
      $423 = HEAP32[$345>>2]|0;
      $424 = (($423) + 1)|0;
      HEAP32[$345>>2] = $424;
      break;
     }
     $425 = (_strncmp($line,531942,6)|0);
     $426 = ($425|0)==(0);
     L174: do {
      if ($426) {
       $427 = ((($line)) + 6|0);
       $428 = HEAP8[$427>>0]|0;
       switch ($428<<24>>24) {
       case 9: case 13: case 10: case 0: case 32:  {
        break;
       }
       default: {
        label = 136;
        break L174;
       }
       }
       $429 = $25 & 128;
       $430 = ($429|0)==(0);
       if (!($430)) {
        $431 = ((($client_data)) + 139304|0);
        $432 = (__bdf_list_split($431,531680,$line,$linelen)|0);
        HEAP32[$error>>2] = $432;
        $433 = ($432|0)==(0);
        if (!($433)) {
         $$pre$phi187Z2D = $client_data;$836 = $432;
         break L7;
        }
        $434 = HEAP32[$431>>2]|0;
        $435 = ((($434)) + 4|0);
        $436 = HEAP32[$435>>2]|0;
        $437 = ($436|0)==(0|0);
        L179: do {
         if ($437) {
          $$01$i95 = 0;
         } else {
          $438 = HEAP8[$436>>0]|0;
          L181: do {
           switch ($438<<24>>24) {
           case 0:  {
            $$01$i95 = 0;
            break L179;
            break;
           }
           case 48:  {
            $439 = ((($436)) + 1|0);
            $440 = HEAP8[$439>>0]|0;
            switch ($440<<24>>24) {
            case 88: case 120:  {
             break;
            }
            default: {
             $$02$i89 = $436;$$1$i90 = 10;$443 = 48;$dmap$1$i91 = 531715;
             break L181;
            }
            }
            $441 = ((($436)) + 2|0);
            $$pre$i88 = HEAP8[$441>>0]|0;
            $$02$i89 = $441;$$1$i90 = 16;$443 = $$pre$i88;$dmap$1$i91 = 531683;
            break;
           }
           default: {
            $$02$i89 = $436;$$1$i90 = 10;$443 = $438;$dmap$1$i91 = 531715;
           }
           }
          } while(0);
          $442 = $443&255;
          $444 = $442 >>> 3;
          $445 = (($dmap$1$i91) + ($444)|0);
          $446 = HEAP8[$445>>0]|0;
          $447 = $446&255;
          $448 = $443 << 24 >> 24;
          $449 = $448 & 7;
          $450 = 1 << $449;
          $451 = $447 & $450;
          $452 = ($451|0)==(0);
          if ($452) {
           $$01$i95 = 0;
          } else {
           $$131$i93 = $$02$i89;$455 = $448;$v$02$i92 = 0;
           while(1) {
            $453 = Math_imul($v$02$i92, $$1$i90)|0;
            $454 = (531747 + ($455)|0);
            $456 = HEAP8[$454>>0]|0;
            $457 = $456&255;
            $458 = (($457) + ($453))|0;
            $459 = ((($$131$i93)) + 1|0);
            $460 = HEAP8[$459>>0]|0;
            $461 = $460&255;
            $462 = $461 >>> 3;
            $463 = (($dmap$1$i91) + ($462)|0);
            $464 = HEAP8[$463>>0]|0;
            $465 = $464&255;
            $466 = $460 << 24 >> 24;
            $467 = $466 & 7;
            $468 = 1 << $467;
            $469 = $468 & $465;
            $470 = ($469|0)==(0);
            if ($470) {
             $$lcssa247 = $458;
             break;
            } else {
             $$131$i93 = $459;$455 = $466;$v$02$i92 = $458;
            }
           }
           $phitmp = $$lcssa247&65535;
           $$01$i95 = $phitmp;
          }
         }
        } while(0);
        $471 = ((($glyph$0)) + 8|0);
        HEAP16[$471>>1] = $$01$i95;
        $472 = HEAP32[$client_data>>2]|0;
        $473 = $472 | 256;
        HEAP32[$client_data>>2] = $473;
        break L9;
       }
      } else {
       label = 136;
      }
     } while(0);
     L189: do {
      if ((label|0) == 136) {
       $474 = (_strncmp($line,531949,6)|0);
       $475 = ($474|0)==(0);
       L191: do {
        if ($475) {
         $476 = ((($line)) + 6|0);
         $477 = HEAP8[$476>>0]|0;
         switch ($477<<24>>24) {
         case 9: case 13: case 10: case 0: case 32:  {
          break;
         }
         default: {
          break L191;
         }
         }
         $478 = $25 & 128;
         $479 = ($478|0)==(0);
         if ($479) {
          break L189;
         }
         $480 = ((($client_data)) + 139304|0);
         $481 = (__bdf_list_split($480,531680,$line,$linelen)|0);
         HEAP32[$error>>2] = $481;
         $482 = ($481|0)==(0);
         if (!($482)) {
          $$pre$phi187Z2D = $client_data;$836 = $481;
          break L7;
         }
         $483 = HEAP32[$480>>2]|0;
         $484 = ((($483)) + 4|0);
         $485 = HEAP32[$484>>2]|0;
         $486 = ($485|0)==(0|0);
         L196: do {
          if ($486) {
           $$01$i83 = 0;
          } else {
           $487 = HEAP8[$485>>0]|0;
           L198: do {
            switch ($487<<24>>24) {
            case 0:  {
             $$01$i83 = 0;
             break L196;
             break;
            }
            case 48:  {
             $488 = ((($485)) + 1|0);
             $489 = HEAP8[$488>>0]|0;
             switch ($489<<24>>24) {
             case 88: case 120:  {
              break;
             }
             default: {
              $$02$i77 = $485;$$1$i78 = 10;$492 = 48;$dmap$1$i79 = 531715;
              break L198;
             }
             }
             $490 = ((($485)) + 2|0);
             $$pre$i76 = HEAP8[$490>>0]|0;
             $$02$i77 = $490;$$1$i78 = 16;$492 = $$pre$i76;$dmap$1$i79 = 531683;
             break;
            }
            default: {
             $$02$i77 = $485;$$1$i78 = 10;$492 = $487;$dmap$1$i79 = 531715;
            }
            }
           } while(0);
           $491 = $492&255;
           $493 = $491 >>> 3;
           $494 = (($dmap$1$i79) + ($493)|0);
           $495 = HEAP8[$494>>0]|0;
           $496 = $495&255;
           $497 = $492 << 24 >> 24;
           $498 = $497 & 7;
           $499 = 1 << $498;
           $500 = $496 & $499;
           $501 = ($500|0)==(0);
           if ($501) {
            $$01$i83 = 0;
           } else {
            $$131$i81 = $$02$i77;$504 = $497;$v$02$i80 = 0;
            while(1) {
             $502 = Math_imul($v$02$i80, $$1$i78)|0;
             $503 = (531747 + ($504)|0);
             $505 = HEAP8[$503>>0]|0;
             $506 = $505&255;
             $507 = (($506) + ($502))|0;
             $508 = ((($$131$i81)) + 1|0);
             $509 = HEAP8[$508>>0]|0;
             $510 = $509&255;
             $511 = $510 >>> 3;
             $512 = (($dmap$1$i79) + ($511)|0);
             $513 = HEAP8[$512>>0]|0;
             $514 = $513&255;
             $515 = $509 << 24 >> 24;
             $516 = $515 & 7;
             $517 = 1 << $516;
             $518 = $517 & $514;
             $519 = ($518|0)==(0);
             if ($519) {
              $$01$i83 = $507;
              break;
             } else {
              $$131$i81 = $508;$504 = $515;$v$02$i80 = $507;
             }
            }
           }
          }
         } while(0);
         $520 = $$01$i83&65535;
         $521 = ((($glyph$0)) + 10|0);
         HEAP16[$521>>1] = $520;
         $522 = HEAP32[$client_data>>2]|0;
         $523 = $522 & 256;
         $524 = ($523|0)==(0);
         if ($524) {
          $525 = $$01$i83 & 65535;
          $526 = ((($1)) + 16|0);
          $527 = HEAP32[$526>>2]|0;
          $528 = ((($1)) + 20|0);
          $529 = HEAP32[$528>>2]|0;
          $530 = Math_imul($529, $527)|0;
          $531 = (_FT_MulDiv($525,72000,$530)|0);
          $532 = $531&65535;
          $533 = ((($glyph$0)) + 8|0);
          HEAP16[$533>>1] = $532;
          $$pre178 = HEAP32[$client_data>>2]|0;
          $535 = $$pre178;
         } else {
          $535 = $522;
         }
         $534 = $535 | 512;
         HEAP32[$client_data>>2] = $534;
         break L9;
        }
       } while(0);
       $536 = (_strncmp($line,531956,3)|0);
       $537 = ($536|0)==(0);
       L209: do {
        if ($537) {
         $538 = ((($line)) + 3|0);
         $539 = HEAP8[$538>>0]|0;
         switch ($539<<24>>24) {
         case 9: case 13: case 10: case 0: case 32:  {
          break;
         }
         default: {
          break L209;
         }
         }
         $540 = $25 & 128;
         $541 = ($540|0)==(0);
         if ($541) {
          break L189;
         }
         $542 = ((($client_data)) + 139304|0);
         $543 = (__bdf_list_split($542,531680,$line,$linelen)|0);
         HEAP32[$error>>2] = $543;
         $544 = ($543|0)==(0);
         if (!($544)) {
          $$pre$phi187Z2D = $client_data;$836 = $543;
          break L7;
         }
         $545 = HEAP32[$542>>2]|0;
         $546 = ((($545)) + 4|0);
         $547 = HEAP32[$546>>2]|0;
         $548 = ($547|0)==(0|0);
         L214: do {
          if ($548) {
           $585 = 0;
          } else {
           $549 = HEAP8[$547>>0]|0;
           L216: do {
            switch ($549<<24>>24) {
            case 0:  {
             $585 = 0;
             break L214;
             break;
            }
            case 48:  {
             $550 = ((($547)) + 1|0);
             $551 = HEAP8[$550>>0]|0;
             switch ($551<<24>>24) {
             case 88: case 120:  {
              break;
             }
             default: {
              $$02$i65 = $547;$$1$i66 = 10;$554 = 48;$dmap$1$i67 = 531715;
              break L216;
             }
             }
             $552 = ((($547)) + 2|0);
             $$pre$i64 = HEAP8[$552>>0]|0;
             $$02$i65 = $552;$$1$i66 = 16;$554 = $$pre$i64;$dmap$1$i67 = 531683;
             break;
            }
            default: {
             $$02$i65 = $547;$$1$i66 = 10;$554 = $549;$dmap$1$i67 = 531715;
            }
            }
           } while(0);
           $553 = $554&255;
           $555 = $553 >>> 3;
           $556 = (($dmap$1$i67) + ($555)|0);
           $557 = HEAP8[$556>>0]|0;
           $558 = $557&255;
           $559 = $554 << 24 >> 24;
           $560 = $559 & 7;
           $561 = 1 << $560;
           $562 = $558 & $561;
           $563 = ($562|0)==(0);
           if ($563) {
            $585 = 0;
            break;
           } else {
            $$131$i69 = $$02$i65;$567 = $559;$v$02$i68 = 0;
           }
           while(1) {
            $564 = $v$02$i68 & 65535;
            $565 = Math_imul($564, $$1$i66)|0;
            $566 = (531747 + ($567)|0);
            $568 = HEAP8[$566>>0]|0;
            $569 = $568&255;
            $570 = (($569) + ($565))|0;
            $571 = ((($$131$i69)) + 1|0);
            $572 = HEAP8[$571>>0]|0;
            $573 = $572&255;
            $574 = $573 >>> 3;
            $575 = (($dmap$1$i67) + ($574)|0);
            $576 = HEAP8[$575>>0]|0;
            $577 = $576&255;
            $578 = $572 << 24 >> 24;
            $579 = $578 & 7;
            $580 = 1 << $579;
            $581 = $580 & $577;
            $582 = ($581|0)==(0);
            if ($582) {
             $$lcssa245 = $570;
             break;
            } else {
             $$131$i69 = $571;$567 = $578;$v$02$i68 = $570;
            }
           }
           $583 = $$lcssa245&65535;
           $585 = $583;
          }
         } while(0);
         $584 = ((($glyph$0)) + 12|0);
         HEAP16[$584>>1] = $585;
         $586 = ((($545)) + 8|0);
         $587 = HEAP32[$586>>2]|0;
         $588 = ($587|0)==(0|0);
         L224: do {
          if ($588) {
           $625 = 0;
          } else {
           $589 = HEAP8[$587>>0]|0;
           L226: do {
            switch ($589<<24>>24) {
            case 0:  {
             $625 = 0;
             break L224;
             break;
            }
            case 48:  {
             $590 = ((($587)) + 1|0);
             $591 = HEAP8[$590>>0]|0;
             switch ($591<<24>>24) {
             case 88: case 120:  {
              break;
             }
             default: {
              $$02$i57 = $587;$$1$i58 = 10;$594 = 48;$dmap$1$i59 = 531715;
              break L226;
             }
             }
             $592 = ((($587)) + 2|0);
             $$pre$i56 = HEAP8[$592>>0]|0;
             $$02$i57 = $592;$$1$i58 = 16;$594 = $$pre$i56;$dmap$1$i59 = 531683;
             break;
            }
            default: {
             $$02$i57 = $587;$$1$i58 = 10;$594 = $589;$dmap$1$i59 = 531715;
            }
            }
           } while(0);
           $593 = $594&255;
           $595 = $593 >>> 3;
           $596 = (($dmap$1$i59) + ($595)|0);
           $597 = HEAP8[$596>>0]|0;
           $598 = $597&255;
           $599 = $594 << 24 >> 24;
           $600 = $599 & 7;
           $601 = 1 << $600;
           $602 = $598 & $601;
           $603 = ($602|0)==(0);
           if ($603) {
            $625 = 0;
            break;
           } else {
            $$131$i61 = $$02$i57;$607 = $599;$v$02$i60 = 0;
           }
           while(1) {
            $604 = $v$02$i60 & 65535;
            $605 = Math_imul($604, $$1$i58)|0;
            $606 = (531747 + ($607)|0);
            $608 = HEAP8[$606>>0]|0;
            $609 = $608&255;
            $610 = (($609) + ($605))|0;
            $611 = ((($$131$i61)) + 1|0);
            $612 = HEAP8[$611>>0]|0;
            $613 = $612&255;
            $614 = $613 >>> 3;
            $615 = (($dmap$1$i59) + ($614)|0);
            $616 = HEAP8[$615>>0]|0;
            $617 = $616&255;
            $618 = $612 << 24 >> 24;
            $619 = $618 & 7;
            $620 = 1 << $619;
            $621 = $620 & $617;
            $622 = ($621|0)==(0);
            if ($622) {
             $$lcssa244 = $610;
             break;
            } else {
             $$131$i61 = $611;$607 = $618;$v$02$i60 = $610;
            }
           }
           $623 = $$lcssa244&65535;
           $625 = $623;
          }
         } while(0);
         $624 = ((($glyph$0)) + 14|0);
         HEAP16[$624>>1] = $625;
         $626 = ((($545)) + 12|0);
         $627 = HEAP32[$626>>2]|0;
         $628 = ($627|0)==(0|0);
         do {
          if ($628) {
           $673 = 0;
          } else {
           $629 = HEAP8[$627>>0]|0;
           $630 = ($629<<24>>24)==(0);
           if ($630) {
            $673 = 0;
            break;
           }
           $631 = ($629<<24>>24)==(45);
           $632 = ((($627)) + 1|0);
           $$s$i44 = $631 ? $632 : $627;
           $633 = HEAP8[$$s$i44>>0]|0;
           $634 = ($633<<24>>24)==(48);
           L237: do {
            if ($634) {
             $635 = ((($$s$i44)) + 1|0);
             $636 = HEAP8[$635>>0]|0;
             switch ($636<<24>>24) {
             case 88: case 120:  {
              break;
             }
             default: {
              $$1$i47 = 10;$$13$i46 = $$s$i44;$639 = 48;$dmap$1$i48 = 531715;
              break L237;
             }
             }
             $637 = ((($$s$i44)) + 2|0);
             $$pre$i45 = HEAP8[$637>>0]|0;
             $$1$i47 = 16;$$13$i46 = $637;$639 = $$pre$i45;$dmap$1$i48 = 531683;
            } else {
             $$1$i47 = 10;$$13$i46 = $$s$i44;$639 = $633;$dmap$1$i48 = 531715;
            }
           } while(0);
           $638 = $639&255;
           $640 = $638 >>> 3;
           $641 = (($dmap$1$i48) + ($640)|0);
           $642 = HEAP8[$641>>0]|0;
           $643 = $642&255;
           $644 = $639 << 24 >> 24;
           $645 = $644 & 7;
           $646 = 1 << $645;
           $647 = $643 & $646;
           $648 = ($647|0)==(0);
           if ($648) {
            $$lcssa$i52 = 0;
           } else {
            $$21$i49 = $$13$i46;$650 = 0;$652 = $644;
            while(1) {
             $649 = Math_imul($650, $$1$i47)|0;
             $651 = (531747 + ($652)|0);
             $653 = HEAP8[$651>>0]|0;
             $654 = $653&255;
             $655 = (($654) + ($649))|0;
             $656 = ((($$21$i49)) + 1|0);
             $657 = HEAP8[$656>>0]|0;
             $658 = $657&255;
             $659 = $658 >>> 3;
             $660 = (($dmap$1$i48) + ($659)|0);
             $661 = HEAP8[$660>>0]|0;
             $662 = $661&255;
             $663 = $657 << 24 >> 24;
             $664 = $663 & 7;
             $665 = 1 << $664;
             $666 = $665 & $662;
             $667 = ($666|0)==(0);
             $sext$i50 = $655 << 16;
             $668 = $sext$i50 >> 16;
             if ($667) {
              $$lcssa$i52 = $668;
              break;
             } else {
              $$21$i49 = $656;$650 = $668;$652 = $663;
             }
            }
           }
           $669 = (0 - ($$lcssa$i52))|0;
           $670 = $631 ? $669 : $$lcssa$i52;
           $671 = $670&65535;
           $673 = $671;
          }
         } while(0);
         $672 = ((($glyph$0)) + 16|0);
         HEAP16[$672>>1] = $673;
         $674 = ((($545)) + 16|0);
         $675 = HEAP32[$674>>2]|0;
         $676 = ($675|0)==(0|0);
         do {
          if ($676) {
           $$01$i43 = 0;
          } else {
           $677 = HEAP8[$675>>0]|0;
           $678 = ($677<<24>>24)==(0);
           if ($678) {
            $$01$i43 = 0;
            break;
           }
           $679 = ($677<<24>>24)==(45);
           $680 = ((($675)) + 1|0);
           $$s$i = $679 ? $680 : $675;
           $681 = HEAP8[$$s$i>>0]|0;
           $682 = ($681<<24>>24)==(48);
           L249: do {
            if ($682) {
             $683 = ((($$s$i)) + 1|0);
             $684 = HEAP8[$683>>0]|0;
             switch ($684<<24>>24) {
             case 88: case 120:  {
              break;
             }
             default: {
              $$1$i40 = 10;$$13$i = $$s$i;$687 = 48;$dmap$1$i41 = 531715;
              break L249;
             }
             }
             $685 = ((($$s$i)) + 2|0);
             $$pre$i39 = HEAP8[$685>>0]|0;
             $$1$i40 = 16;$$13$i = $685;$687 = $$pre$i39;$dmap$1$i41 = 531683;
            } else {
             $$1$i40 = 10;$$13$i = $$s$i;$687 = $681;$dmap$1$i41 = 531715;
            }
           } while(0);
           $686 = $687&255;
           $688 = $686 >>> 3;
           $689 = (($dmap$1$i41) + ($688)|0);
           $690 = HEAP8[$689>>0]|0;
           $691 = $690&255;
           $692 = $687 << 24 >> 24;
           $693 = $692 & 7;
           $694 = 1 << $693;
           $695 = $691 & $694;
           $696 = ($695|0)==(0);
           if ($696) {
            $$lcssa$i = 0;
           } else {
            $$21$i = $$13$i;$698 = 0;$700 = $692;
            while(1) {
             $697 = Math_imul($698, $$1$i40)|0;
             $699 = (531747 + ($700)|0);
             $701 = HEAP8[$699>>0]|0;
             $702 = $701&255;
             $703 = (($702) + ($697))|0;
             $704 = ((($$21$i)) + 1|0);
             $705 = HEAP8[$704>>0]|0;
             $706 = $705&255;
             $707 = $706 >>> 3;
             $708 = (($dmap$1$i41) + ($707)|0);
             $709 = HEAP8[$708>>0]|0;
             $710 = $709&255;
             $711 = $705 << 24 >> 24;
             $712 = $711 & 7;
             $713 = 1 << $712;
             $714 = $713 & $710;
             $715 = ($714|0)==(0);
             $sext$i = $703 << 16;
             $716 = $sext$i >> 16;
             if ($715) {
              $$lcssa$i = $716;
              break;
             } else {
              $$21$i = $704;$698 = $716;$700 = $711;
             }
            }
           }
           $717 = (0 - ($$lcssa$i))|0;
           $718 = $679 ? $717 : $$lcssa$i;
           $719 = $718&65535;
           $$01$i43 = $719;
          }
         } while(0);
         $720 = ((($glyph$0)) + 18|0);
         HEAP16[$720>>1] = $$01$i43;
         $721 = $625&65535;
         $722 = $$01$i43&65535;
         $723 = (($721) + ($722))|0;
         $724 = $723&65535;
         $725 = ((($glyph$0)) + 20|0);
         HEAP16[$725>>1] = $724;
         $726 = (0 - ($722))|0;
         $727 = $726&65535;
         $728 = ((($glyph$0)) + 22|0);
         HEAP16[$728>>1] = $727;
         $729 = ((($client_data)) + 18|0);
         $730 = HEAP16[$729>>1]|0;
         $731 = ($724<<16>>16)>($730<<16>>16);
         $$31 = $731 ? $724 : $730;
         HEAP16[$729>>1] = $$31;
         $732 = ((($client_data)) + 20|0);
         $733 = HEAP16[$732>>1]|0;
         $734 = ($727<<16>>16)>($733<<16>>16);
         $$sink1 = $734 ? $727 : $733;
         HEAP16[$732>>1] = $$sink1;
         $735 = $585&65535;
         $736 = $673&65535;
         $737 = (($736) + ($735))|0;
         $738 = $737&65535;
         $739 = ((($client_data)) + 22|0);
         HEAP16[$739>>1] = $738;
         $sext = $737 << 16;
         $740 = $sext >> 16;
         $741 = ((($client_data)) + 16|0);
         $742 = HEAP16[$741>>1]|0;
         $743 = $742 << 16 >> 16;
         $744 = ($740|0)>($743|0);
         $$32 = $744 ? $738 : $742;
         HEAP16[$741>>1] = $$32;
         $745 = ((($client_data)) + 12|0);
         $746 = HEAP32[$745>>2]|0;
         $747 = $746&65535;
         $748 = ($673<<16>>16)<($747<<16>>16);
         $$sink3 = $748 ? $673 : $747;
         HEAP16[$745>>1] = $$sink3;
         $749 = ((($client_data)) + 14|0);
         $750 = $746 >>> 16;
         $751 = $750&65535;
         $752 = ($673<<16>>16)>($751<<16>>16);
         $$33 = $752 ? $673 : $751;
         HEAP16[$749>>1] = $$33;
         $753 = HEAP32[$client_data>>2]|0;
         $754 = $753 & 512;
         $755 = ($754|0)==(0);
         if ($755) {
          $756 = ((($glyph$0)) + 10|0);
          HEAP16[$756>>1] = $585;
         }
         $757 = ((($client_data)) + 36|0);
         $758 = HEAP32[$757>>2]|0;
         $759 = HEAP32[$758>>2]|0;
         $760 = ($759|0)==(0);
         do {
          if ($760) {
           $797 = $753;
          } else {
           $761 = ((($glyph$0)) + 10|0);
           $762 = HEAP16[$761>>1]|0;
           $763 = $762&65535;
           $764 = ((($1)) + 16|0);
           $765 = HEAP32[$764>>2]|0;
           $766 = ((($1)) + 20|0);
           $767 = HEAP32[$766>>2]|0;
           $768 = Math_imul($767, $765)|0;
           $769 = (_FT_MulDiv($763,72000,$768)|0);
           $770 = ((($glyph$0)) + 8|0);
           $771 = HEAP16[$770>>1]|0;
           $772 = $769&65535;
           $773 = ($771<<16>>16)==($772<<16>>16);
           if ($773) {
            $$pre180 = HEAP32[$client_data>>2]|0;
            $797 = $$pre180;
            break;
           }
           HEAP16[$770>>1] = $772;
           $774 = HEAP32[$328>>2]|0;
           $775 = ($774|0)==(-1);
           if ($775) {
            $776 = ((($1)) + 64|0);
            $777 = HEAP32[$776>>2]|0;
            $778 = (($777) + -1)|0;
            $779 = $778 & 31;
            $780 = 1 << $779;
            $781 = $778 >>> 5;
            $782 = (((($1)) + 139396|0) + ($781<<2)|0);
            $783 = HEAP32[$782>>2]|0;
            $784 = $780 | $783;
            HEAP32[$782>>2] = $784;
           } else {
            $785 = ((($glyph$0)) + 4|0);
            $786 = HEAP32[$785>>2]|0;
            $787 = $786 & 31;
            $788 = 1 << $787;
            $789 = $786 >> 5;
            $790 = (((($1)) + 132|0) + ($789<<2)|0);
            $791 = HEAP32[$790>>2]|0;
            $792 = $788 | $791;
            HEAP32[$790>>2] = $792;
           }
           $793 = HEAP32[$client_data>>2]|0;
           $794 = $793 | 4096;
           HEAP32[$client_data>>2] = $794;
           $795 = ((($1)) + 278660|0);
           HEAP16[$795>>1] = 1;
           $797 = $794;
          }
         } while(0);
         $796 = $797 | 1024;
         HEAP32[$client_data>>2] = $796;
         break L9;
        }
       } while(0);
       $798 = (_strncmp($line,531960,6)|0);
       $799 = ($798|0)==(0);
       L272: do {
        if ($799) {
         $800 = ((($line)) + 6|0);
         $801 = HEAP8[$800>>0]|0;
         switch ($801<<24>>24) {
         case 9: case 13: case 10: case 0: case 32:  {
          break;
         }
         default: {
          break L272;
         }
         }
         $802 = $25 & 1024;
         $803 = ($802|0)==(0);
         if ($803) {
          HEAP32[$error>>2] = 183;
          $$pre$phi187Z2D = $client_data;$836 = 183;
          break L7;
         }
         $804 = ((($glyph$0)) + 12|0);
         $805 = HEAP32[$804>>2]|0;
         $806 = $805 & 65535;
         $807 = ((($1)) + 278662|0);
         $808 = HEAP16[$807>>1]|0;
         $809 = $808&65535;
         $810 = Math_imul($809, $806)|0;
         $811 = (($810) + 7)|0;
         $812 = $811 >>> 3;
         $813 = ((($glyph$0)) + 28|0);
         HEAP32[$813>>2] = $812;
         $814 = $805 >>> 16;
         $815 = Math_imul($812, $814)|0;
         $816 = ($811>>>0)>(524287);
         $817 = ($815>>>0)>(65535);
         $or$cond = $816 | $817;
         if ($or$cond) {
          HEAP32[$error>>2] = 184;
          $$pre$phi187Z2D = $client_data;$836 = 184;
          break L7;
         }
         $818 = $815&65535;
         $819 = ((($glyph$0)) + 32|0);
         HEAP16[$819>>1] = $818;
         $820 = $815 & 65535;
         $821 = (_ft_mem_realloc($3,1,0,$820,0,$error)|0);
         $822 = ((($glyph$0)) + 24|0);
         HEAP32[$822>>2] = $821;
         $823 = HEAP32[$error>>2]|0;
         $824 = ($823|0)==(0);
         if (!($824)) {
          $$pre$phi187Z2D = $client_data;$836 = $823;
          break L7;
         }
         $825 = ((($client_data)) + 8|0);
         HEAP32[$825>>2] = 0;
         $826 = HEAP32[$client_data>>2]|0;
         $827 = $826 | 2048;
         HEAP32[$client_data>>2] = $827;
         break L9;
        }
       } while(0);
       HEAP32[$error>>2] = 3;
       $$pre$phi187Z2D = $client_data;$836 = 3;
       break L7;
      }
     } while(0);
     HEAP32[$error>>2] = 182;
     $$pre$phi187Z2D = $client_data;$836 = 182;
     break L7;
    }
   } while(0);
   $$pr134 = HEAP32[$error>>2]|0;
   $829 = $$pr134;
   label = 204;
  }
 } while(0);
 do {
  if ((label|0) == 204) {
   $828 = ($829|0)==(0);
   if ($828) {
    $835 = 0;
    STACKTOP = sp;return ($835|0);
   } else {
    $$pre$phi187Z2D = $client_data;$836 = $829;
    break;
   }
  }
 } while(0);
 $830 = HEAP32[$$pre$phi187Z2D>>2]|0;
 $831 = $830 & 64;
 $832 = ($831|0)==(0);
 if ($832) {
  $835 = $836;
  STACKTOP = sp;return ($835|0);
 }
 $833 = ((($client_data)) + 24|0);
 $834 = HEAP32[$833>>2]|0;
 _ft_mem_free($3,$834);
 HEAP32[$833>>2] = 0;
 $$pre181 = HEAP32[$error>>2]|0;
 $835 = $$pre181;
 STACKTOP = sp;return ($835|0);
}
function _by_encoding($a,$b) {
 $a = $a|0;
 $b = $b|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($a)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($b)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)<($3|0);
 $5 = ($1|0)>($3|0);
 $$ = $5&1;
 $$0 = $4 ? -1 : $$;
 return ($$0|0);
}
function _bdf_free_font($font) {
 $font = $font|0;
 var $$pre = 0, $$pre$i = 0, $$pre$i7 = 0, $$pre32 = 0, $$pre34 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bp$02$i = 0, $bp$02$i2 = 0, $exitcond$i = 0, $exitcond$i4 = 0, $glyphs$024 = 0, $glyphs$119 = 0, $glyphs$214 = 0, $i$01$i = 0, $i$01$i3 = 0, $i$028 = 0, $i$123 = 0, $i$218 = 0, $i$313 = 0, $i$411 = 0, $prop$012 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($font|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($font)) + 278664|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = HEAP32[$font>>2]|0;
 _ft_mem_free($2,$3);
 HEAP32[$font>>2] = 0;
 $4 = ((($font)) + 128|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if (!($6)) {
  $7 = ((($5)) + 4|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ((($5)) + 12|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ($8|0)==(0);
  if ($11) {
   $15 = $10;
  } else {
   $bp$02$i = $10;$i$01$i = 0;
   while(1) {
    $12 = HEAP32[$bp$02$i>>2]|0;
    _ft_mem_free($2,$12);
    HEAP32[$bp$02$i>>2] = 0;
    $13 = (($i$01$i) + 1)|0;
    $14 = ((($bp$02$i)) + 4|0);
    $exitcond$i = ($13|0)==($8|0);
    if ($exitcond$i) {
     break;
    } else {
     $bp$02$i = $14;$i$01$i = $13;
    }
   }
   $$pre$i = HEAP32[$9>>2]|0;
   $15 = $$pre$i;
  }
  _ft_mem_free($2,$15);
  HEAP32[$9>>2] = 0;
  $16 = HEAP32[$4>>2]|0;
  _ft_mem_free($2,$16);
  HEAP32[$4>>2] = 0;
 }
 $17 = ((($font)) + 84|0);
 $18 = HEAP32[$17>>2]|0;
 _ft_mem_free($2,$18);
 HEAP32[$17>>2] = 0;
 $19 = ((($font)) + 72|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ($20|0)==(0);
 $22 = ((($font)) + 80|0);
 if (!($21)) {
  $99 = $20;$i$028 = 0;
  while(1) {
   $23 = HEAP32[$22>>2]|0;
   $24 = (((($23) + ($i$028<<4)|0)) + 4|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = ($25|0)==(1);
   if ($26) {
    $27 = (((($23) + ($i$028<<4)|0)) + 12|0);
    $28 = HEAP32[$27>>2]|0;
    _ft_mem_free($2,$28);
    $29 = HEAP32[$22>>2]|0;
    $30 = (((($29) + ($i$028<<4)|0)) + 12|0);
    HEAP32[$30>>2] = 0;
    $$pre = HEAP32[$19>>2]|0;
    $33 = $$pre;
   } else {
    $33 = $99;
   }
   $31 = (($i$028) + 1)|0;
   $32 = ($31>>>0)<($33>>>0);
   if ($32) {
    $99 = $33;$i$028 = $31;
   } else {
    break;
   }
  }
 }
 $34 = HEAP32[$22>>2]|0;
 _ft_mem_free($2,$34);
 HEAP32[$22>>2] = 0;
 $35 = ((($font)) + 56|0);
 $36 = ((($font)) + 52|0);
 $37 = HEAP32[$36>>2]|0;
 $38 = ($37|0)==(0);
 if (!($38)) {
  $39 = HEAP32[$35>>2]|0;
  $glyphs$024 = $39;$i$123 = 0;
  while(1) {
   $40 = HEAP32[$glyphs$024>>2]|0;
   _ft_mem_free($2,$40);
   HEAP32[$glyphs$024>>2] = 0;
   $41 = ((($glyphs$024)) + 24|0);
   $42 = HEAP32[$41>>2]|0;
   _ft_mem_free($2,$42);
   HEAP32[$41>>2] = 0;
   $43 = (($i$123) + 1)|0;
   $44 = ((($glyphs$024)) + 36|0);
   $45 = HEAP32[$36>>2]|0;
   $46 = ($43>>>0)<($45>>>0);
   if ($46) {
    $glyphs$024 = $44;$i$123 = $43;
   } else {
    break;
   }
  }
 }
 $47 = ((($font)) + 68|0);
 $48 = ((($font)) + 64|0);
 $49 = HEAP32[$48>>2]|0;
 $50 = ($49|0)==(0);
 if (!($50)) {
  $51 = HEAP32[$47>>2]|0;
  $glyphs$119 = $51;$i$218 = 0;
  while(1) {
   $52 = HEAP32[$glyphs$119>>2]|0;
   _ft_mem_free($2,$52);
   HEAP32[$glyphs$119>>2] = 0;
   $53 = ((($glyphs$119)) + 24|0);
   $54 = HEAP32[$53>>2]|0;
   _ft_mem_free($2,$54);
   HEAP32[$53>>2] = 0;
   $55 = (($i$218) + 1)|0;
   $56 = ((($glyphs$119)) + 36|0);
   $57 = HEAP32[$48>>2]|0;
   $58 = ($55>>>0)<($57>>>0);
   if ($58) {
    $glyphs$119 = $56;$i$218 = $55;
   } else {
    break;
   }
  }
 }
 $59 = HEAP32[$35>>2]|0;
 _ft_mem_free($2,$59);
 HEAP32[$35>>2] = 0;
 $60 = HEAP32[$47>>2]|0;
 _ft_mem_free($2,$60);
 HEAP32[$47>>2] = 0;
 $61 = ((($font)) + 104|0);
 $62 = HEAP32[$61>>2]|0;
 $63 = ((($font)) + 112|0);
 $64 = HEAP32[$63>>2]|0;
 $65 = ($64|0)==(0);
 if ($65) {
  $73 = $62;
 } else {
  $glyphs$214 = $62;$i$313 = 0;
  while(1) {
   $66 = HEAP32[$glyphs$214>>2]|0;
   _ft_mem_free($2,$66);
   HEAP32[$glyphs$214>>2] = 0;
   $67 = ((($glyphs$214)) + 24|0);
   $68 = HEAP32[$67>>2]|0;
   _ft_mem_free($2,$68);
   HEAP32[$67>>2] = 0;
   $69 = (($i$313) + 1)|0;
   $70 = ((($glyphs$214)) + 36|0);
   $71 = HEAP32[$63>>2]|0;
   $72 = ($69>>>0)<($71>>>0);
   if ($72) {
    $glyphs$214 = $70;$i$313 = $69;
   } else {
    break;
   }
  }
  $$pre32 = HEAP32[$61>>2]|0;
  $73 = $$pre32;
 }
 _ft_mem_free($2,$73);
 HEAP32[$61>>2] = 0;
 $74 = ((($font)) + 278680|0);
 $75 = HEAP32[$74>>2]|0;
 $76 = ((($font)) + 278688|0);
 $77 = HEAP32[$76>>2]|0;
 $78 = ($75|0)==(0);
 if ($78) {
  $82 = $77;
 } else {
  $bp$02$i2 = $77;$i$01$i3 = 0;
  while(1) {
   $79 = HEAP32[$bp$02$i2>>2]|0;
   _ft_mem_free($2,$79);
   HEAP32[$bp$02$i2>>2] = 0;
   $80 = (($i$01$i3) + 1)|0;
   $81 = ((($bp$02$i2)) + 4|0);
   $exitcond$i4 = ($80|0)==($75|0);
   if ($exitcond$i4) {
    break;
   } else {
    $bp$02$i2 = $81;$i$01$i3 = $80;
   }
  }
  $$pre$i7 = HEAP32[$76>>2]|0;
  $82 = $$pre$i7;
 }
 _ft_mem_free($2,$82);
 HEAP32[$76>>2] = 0;
 $83 = ((($font)) + 278668|0);
 $84 = HEAP32[$83>>2]|0;
 $85 = ((($font)) + 278672|0);
 $86 = HEAP32[$85>>2]|0;
 $87 = ($86|0)==(0);
 if ($87) {
  $98 = $84;
 } else {
  $i$411 = 0;$prop$012 = $84;
  while(1) {
   $88 = HEAP32[$prop$012>>2]|0;
   _ft_mem_free($2,$88);
   HEAP32[$prop$012>>2] = 0;
   $89 = ((($prop$012)) + 4|0);
   $90 = HEAP32[$89>>2]|0;
   $91 = ($90|0)==(1);
   if ($91) {
    $92 = ((($prop$012)) + 12|0);
    $93 = HEAP32[$92>>2]|0;
    _ft_mem_free($2,$93);
    HEAP32[$92>>2] = 0;
   }
   $94 = (($i$411) + 1)|0;
   $95 = ((($prop$012)) + 16|0);
   $96 = HEAP32[$85>>2]|0;
   $97 = ($94>>>0)<($96>>>0);
   if ($97) {
    $i$411 = $94;$prop$012 = $95;
   } else {
    break;
   }
  }
  $$pre34 = HEAP32[$83>>2]|0;
  $98 = $$pre34;
 }
 _ft_mem_free($2,$98);
 HEAP32[$83>>2] = 0;
 return;
}
function _BDF_Face_Done($bdfface) {
 $bdfface = $bdfface|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($bdfface|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($bdfface)) + 100|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($bdfface)) + 140|0);
 $4 = HEAP32[$3>>2]|0;
 _bdf_free_font($4);
 $5 = ((($bdfface)) + 144|0);
 $6 = HEAP32[$5>>2]|0;
 _ft_mem_free($2,$6);
 HEAP32[$5>>2] = 0;
 $7 = ((($bdfface)) + 132|0);
 $8 = HEAP32[$7>>2]|0;
 _ft_mem_free($2,$8);
 HEAP32[$7>>2] = 0;
 $9 = ((($bdfface)) + 136|0);
 $10 = HEAP32[$9>>2]|0;
 _ft_mem_free($2,$10);
 HEAP32[$9>>2] = 0;
 $11 = ((($bdfface)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 _ft_mem_free($2,$12);
 HEAP32[$11>>2] = 0;
 $13 = ((($bdfface)) + 24|0);
 $14 = HEAP32[$13>>2]|0;
 _ft_mem_free($2,$14);
 HEAP32[$13>>2] = 0;
 $15 = ((($bdfface)) + 32|0);
 $16 = HEAP32[$15>>2]|0;
 _ft_mem_free($2,$16);
 HEAP32[$15>>2] = 0;
 $17 = HEAP32[$3>>2]|0;
 _ft_mem_free($2,$17);
 HEAP32[$3>>2] = 0;
 return;
}
function _bdf_cmap_init($bdfcmap,$init_data) {
 $bdfcmap = $bdfcmap|0;
 $init_data = $init_data|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$bdfcmap>>2]|0;
 $1 = ((($0)) + 140|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 52|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($bdfcmap)) + 16|0);
 HEAP32[$5>>2] = $4;
 $6 = ((($0)) + 144|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($bdfcmap)) + 20|0);
 HEAP32[$8>>2] = $7;
 return 0;
}
function _bdf_cmap_done($bdfcmap) {
 $bdfcmap = $bdfcmap|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($bdfcmap)) + 20|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($bdfcmap)) + 16|0);
 HEAP32[$1>>2] = 0;
 return;
}
function _bdf_cmap_char_index($bdfcmap,$charcode) {
 $bdfcmap = $bdfcmap|0;
 $charcode = $charcode|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $max$01 = 0, $max$1 = 0;
 var $min$02 = 0, $min$1 = 0, $phitmp = 0, $result$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($bdfcmap)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($bdfcmap)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0);
 if ($4) {
  $result$2 = 0;
  return ($result$2|0);
 } else {
  $max$01 = $3;$min$02 = 0;
 }
 while(1) {
  $5 = (($min$02) + ($max$01))|0;
  $6 = $5 >>> 1;
  $7 = (($1) + ($6<<3)|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8|0)==($charcode|0);
  if ($9) {
   $$lcssa = $6;
   break;
  }
  $10 = ($8>>>0)>($charcode>>>0);
  $11 = (($6) + 1)|0;
  $max$1 = $10 ? $6 : $max$01;
  $min$1 = $10 ? $min$02 : $11;
  $12 = ($min$1>>>0)<($max$1>>>0);
  if ($12) {
   $max$01 = $max$1;$min$02 = $min$1;
  } else {
   $result$2 = 0;
   label = 5;
   break;
  }
 }
 if ((label|0) == 5) {
  return ($result$2|0);
 }
 $13 = (((($1) + ($$lcssa<<3)|0)) + 4|0);
 $14 = HEAP16[$13>>1]|0;
 $15 = $14&65535;
 $16 = (($15) + 1)|0;
 $phitmp = $16 & 65535;
 $result$2 = $phitmp;
 return ($result$2|0);
}
function _bdf_cmap_char_next($bdfcmap,$acharcode) {
 $bdfcmap = $bdfcmap|0;
 $acharcode = $acharcode|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $charcode$0 = 0, $max$04 = 0, $max$1 = 0, $min$0$lcssa = 0, $min$05 = 0, $min$1 = 0, $result$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($bdfcmap)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$acharcode>>2]|0;
 $3 = (($2) + 1)|0;
 $4 = ((($bdfcmap)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0);
 L1: do {
  if ($6) {
   $min$0$lcssa = 0;
  } else {
   $max$04 = $5;$min$05 = 0;
   while(1) {
    $7 = (($min$05) + ($max$04))|0;
    $8 = $7 >>> 1;
    $9 = (($1) + ($8<<3)|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = ($3|0)==($10|0);
    if ($11) {
     $$lcssa = $8;
     break;
    }
    $12 = ($3>>>0)<($10>>>0);
    $13 = (($8) + 1)|0;
    $max$1 = $12 ? $8 : $max$04;
    $min$1 = $12 ? $min$05 : $13;
    $14 = ($min$1>>>0)<($max$1>>>0);
    if ($14) {
     $max$04 = $max$1;$min$05 = $min$1;
    } else {
     $min$0$lcssa = $min$1;
     break L1;
    }
   }
   $15 = (((($1) + ($$lcssa<<3)|0)) + 4|0);
   $16 = HEAP16[$15>>1]|0;
   $17 = $16&65535;
   $18 = (($17) + 1)|0;
   $charcode$0 = $3;$result$2 = $18;
   HEAP32[$acharcode>>2] = $charcode$0;
   $26 = $result$2 & 65535;
   return ($26|0);
  }
 } while(0);
 $19 = ($min$0$lcssa>>>0)<($5>>>0);
 if (!($19)) {
  $charcode$0 = 0;$result$2 = 0;
  HEAP32[$acharcode>>2] = $charcode$0;
  $26 = $result$2 & 65535;
  return ($26|0);
 }
 $20 = (($1) + ($min$0$lcssa<<3)|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = (((($1) + ($min$0$lcssa<<3)|0)) + 4|0);
 $23 = HEAP16[$22>>1]|0;
 $24 = $23&65535;
 $25 = (($24) + 1)|0;
 $charcode$0 = $21;$result$2 = $25;
 HEAP32[$acharcode>>2] = $charcode$0;
 $26 = $result$2 & 65535;
 return ($26|0);
}
function _BDF_Glyph_Load($slot,$size,$glyph_index,$load_flags) {
 $slot = $slot|0;
 $size = $size|0;
 $glyph_index = $glyph_index|0;
 $load_flags = $load_flags|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $7 = 0, $8 = 0, $9 = 0, $error$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$size>>2]|0;
 $1 = ((($0)) + 140|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 278662|0);
 $4 = HEAP16[$3>>1]|0;
 $5 = $4&65535;
 $6 = ($0|0)==(0|0);
 if ($6) {
  $error$0 = 35;
  return ($error$0|0);
 }
 $7 = ((($0)) + 16|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8>>>0)>($glyph_index>>>0);
 if (!($9)) {
  $error$0 = 6;
  return ($error$0|0);
 }
 $10 = ($glyph_index|0)==(0);
 if ($10) {
  $11 = ((($0)) + 164|0);
  $12 = HEAP32[$11>>2]|0;
  $$0 = $12;
 } else {
  $13 = (($glyph_index) + -1)|0;
  $$0 = $13;
 }
 $14 = ((($2)) + 56|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = (((($15) + (($$0*36)|0)|0)) + 10|0);
 $17 = HEAP16[$16>>1]|0;
 $18 = (((($15) + (($$0*36)|0)|0)) + 12|0);
 $19 = HEAP16[$18>>1]|0;
 $20 = (((($15) + (($$0*36)|0)|0)) + 14|0);
 $21 = HEAP16[$20>>1]|0;
 $22 = (((($15) + (($$0*36)|0)|0)) + 16|0);
 $23 = HEAP16[$22>>1]|0;
 $24 = (((($15) + (($$0*36)|0)|0)) + 20|0);
 $25 = HEAP16[$24>>1]|0;
 $26 = (((($15) + (($$0*36)|0)|0)) + 24|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = (((($15) + (($$0*36)|0)|0)) + 28|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = $21&65535;
 $31 = ((($slot)) + 76|0);
 HEAP32[$31>>2] = $30;
 $32 = $19&65535;
 $33 = ((($slot)) + 80|0);
 HEAP32[$33>>2] = $32;
 $34 = ((($slot)) + 84|0);
 HEAP32[$34>>2] = $29;
 _ft_glyphslot_set_bitmap($slot,$27);
 switch ($5|0) {
 case 1:  {
  $35 = ((($slot)) + 94|0);
  HEAP8[$35>>0] = 1;
  break;
 }
 case 2:  {
  $36 = ((($slot)) + 94|0);
  HEAP8[$36>>0] = 3;
  break;
 }
 case 4:  {
  $37 = ((($slot)) + 94|0);
  HEAP8[$37>>0] = 4;
  break;
 }
 case 8:  {
  $38 = ((($slot)) + 94|0);
  HEAP8[$38>>0] = 2;
  $39 = ((($slot)) + 92|0);
  HEAP16[$39>>1] = 256;
  break;
 }
 default: {
 }
 }
 $40 = ((($slot)) + 72|0);
 HEAP32[$40>>2] = 1651078259;
 $41 = $23 << 16 >> 16;
 $42 = ((($slot)) + 100|0);
 HEAP32[$42>>2] = $41;
 $43 = $25 << 16 >> 16;
 $44 = ((($slot)) + 104|0);
 HEAP32[$44>>2] = $43;
 $45 = $17&65535;
 $46 = $45 << 6;
 $47 = ((($slot)) + 24|0);
 $48 = ((($slot)) + 40|0);
 HEAP32[$48>>2] = $46;
 $49 = $41 << 6;
 $50 = ((($slot)) + 32|0);
 HEAP32[$50>>2] = $49;
 $51 = $43 << 6;
 $52 = ((($slot)) + 36|0);
 HEAP32[$52>>2] = $51;
 $53 = HEAP32[$33>>2]|0;
 $54 = $53 << 6;
 HEAP32[$47>>2] = $54;
 $55 = HEAP32[$31>>2]|0;
 $56 = $55 << 6;
 $57 = ((($slot)) + 28|0);
 HEAP32[$57>>2] = $56;
 $58 = HEAP32[$1>>2]|0;
 $59 = ((($58)) + 6|0);
 $60 = HEAP16[$59>>1]|0;
 $61 = $60&65535;
 $62 = $61 << 6;
 _ft_synthesize_vertical_metrics($47,$62);
 $error$0 = 0;
 return ($error$0|0);
}
function _BDF_Size_Request($size,$req) {
 $size = $size|0;
 $req = $req|0;
 var $$0 = 0, $$pre = 0, $$pre$phi4Z2D = 0, $$pre$phiZ2D = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$size>>2]|0;
 $1 = ((($0)) + 32|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($0)) + 140|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($req)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 $8 = ((($req)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 if ($7) {
  $14 = $9;
 } else {
  $10 = Math_imul($9, $6)|0;
  $11 = (($10) + 36)|0;
  $12 = (($11|0) / 72)&-1;
  $14 = $12;
 }
 $13 = (($14) + 32)|0;
 $15 = $13 >> 6;
 $16 = HEAP32[$req>>2]|0;
 L4: do {
  switch ($16|0) {
  case 0:  {
   $17 = ((($2)) + 12|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = (($18) + 32)|0;
   $20 = $19 >> 6;
   $21 = ($15|0)==($20|0);
   if ($21) {
    $$pre = ((($4)) + 40|0);
    $$pre3 = ((($4)) + 44|0);
    $$pre$phi4Z2D = $$pre3;$$pre$phiZ2D = $$pre;
    break L4;
   } else {
    $$0 = 23;
    return ($$0|0);
   }
   break;
  }
  case 1:  {
   $22 = ((($4)) + 40|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = ((($4)) + 44|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = (($25) + ($23))|0;
   $27 = ($15|0)==($26|0);
   if ($27) {
    $$pre$phi4Z2D = $24;$$pre$phiZ2D = $22;
   } else {
    $$0 = 23;
    return ($$0|0);
   }
   break;
  }
  default: {
   $$0 = 7;
   return ($$0|0);
  }
  }
 } while(0);
 _FT_Select_Metrics($0,0);
 $28 = HEAP32[$$pre$phiZ2D>>2]|0;
 $29 = $28 << 6;
 $30 = ((($size)) + 24|0);
 HEAP32[$30>>2] = $29;
 $31 = HEAP32[$$pre$phi4Z2D>>2]|0;
 $32 = (0 - ($31))|0;
 $33 = $32 << 6;
 $34 = ((($size)) + 28|0);
 HEAP32[$34>>2] = $33;
 $35 = ((($4)) + 4|0);
 $36 = HEAP16[$35>>1]|0;
 $37 = $36&65535;
 $38 = $37 << 6;
 $39 = ((($size)) + 36|0);
 HEAP32[$39>>2] = $38;
 $$0 = 0;
 return ($$0|0);
}
function _BDF_Size_Select($size,$strike_index) {
 $size = $size|0;
 $strike_index = $strike_index|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$size>>2]|0;
 $1 = ((($0)) + 140|0);
 $2 = HEAP32[$1>>2]|0;
 _FT_Select_Metrics($0,$strike_index);
 $3 = ((($2)) + 40|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 << 6;
 $6 = ((($size)) + 24|0);
 HEAP32[$6>>2] = $5;
 $7 = ((($2)) + 44|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (0 - ($8))|0;
 $10 = $9 << 6;
 $11 = ((($size)) + 28|0);
 HEAP32[$11>>2] = $10;
 $12 = ((($2)) + 4|0);
 $13 = HEAP16[$12>>1]|0;
 $14 = $13&65535;
 $15 = $14 << 6;
 $16 = ((($size)) + 36|0);
 HEAP32[$16>>2] = $15;
 return 0;
}
function _cff_driver_init($module) {
 $module = $module|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($module)) + 28|0);
 HEAP32[$0>>2] = 1;
 $1 = ((($module)) + 32|0);
 HEAP8[$1>>0] = 0;
 $2 = ((($module)) + 36|0);
 HEAP32[$2>>2] = 500;
 $3 = ((($module)) + 40|0);
 HEAP32[$3>>2] = 400;
 $4 = ((($module)) + 44|0);
 HEAP32[$4>>2] = 1000;
 $5 = ((($module)) + 48|0);
 HEAP32[$5>>2] = 275;
 $6 = ((($module)) + 52|0);
 HEAP32[$6>>2] = 1667;
 $7 = ((($module)) + 56|0);
 HEAP32[$7>>2] = 275;
 $8 = ((($module)) + 60|0);
 HEAP32[$8>>2] = 2333;
 $9 = ((($module)) + 64|0);
 HEAP32[$9>>2] = 0;
 return 0;
}
function _cff_driver_done($module) {
 $module = $module|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _cff_get_interface($driver,$module_interface) {
 $driver = $driver|0;
 $module_interface = $module_interface|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_ft_service_list_lookup(49464,$module_interface)|0);
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  $$0 = $0;
  return ($$0|0);
 }
 $2 = ($driver|0)==(0|0);
 if ($2) {
  $$0 = 0;
  return ($$0|0);
 }
 $3 = ((($driver)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if ($5) {
  $$0 = 0;
  return ($$0|0);
 }
 $6 = (_FT_Get_Module($4,593168)|0);
 $7 = ($6|0)==(0|0);
 if ($7) {
  $$0 = 0;
  return ($$0|0);
 }
 $8 = HEAP32[$6>>2]|0;
 $9 = ((($8)) + 32|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (FUNCTION_TABLE_iii[$10 & 8191]($6,$module_interface)|0);
 $$0 = $11;
 return ($$0|0);
}
function _cff_ps_get_font_info($face,$afont_info) {
 $face = $face|0;
 $afont_info = $afont_info|0;
 var $$0$i = 0, $$0$i2 = 0, $$0$i4 = 0, $$0$i6 = 0, $$0$i8 = 0, $$pre = 0, $$pre$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $error = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($face)) + 656|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$error>>2] = 0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $120 = 0;
  STACKTOP = sp;return ($120|0);
 }
 $3 = ((($1)) + 2960|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if ($5) {
  $6 = ((($face)) + 100|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (_ft_mem_alloc($7,32,$error)|0);
  $9 = HEAP32[$error>>2]|0;
  $10 = ($9|0)==(0);
  if (!($10)) {
   $120 = $9;
   STACKTOP = sp;return ($120|0);
  }
  $11 = ((($1)) + 1324|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ($12|0)==(65535);
  do {
   if ($13) {
    $$0$i = 0;
   } else {
    $14 = ($12>>>0)>(390);
    if ($14) {
     $15 = (($12) + -391)|0;
     $16 = ((($1)) + 1312|0);
     $17 = HEAP32[$16>>2]|0;
     $18 = ($17>>>0)>($15>>>0);
     if (!($18)) {
      $$0$i = 0;
      break;
     }
     $19 = ((($1)) + 1316|0);
     $20 = HEAP32[$19>>2]|0;
     $21 = (($20) + ($15<<2)|0);
     $22 = HEAP32[$21>>2]|0;
     $$0$i = $22;
     break;
    } else {
     $23 = ((($1)) + 2956|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ($24|0)==(0|0);
     if ($25) {
      $$0$i = 0;
      break;
     }
     $26 = ((($24)) + 20|0);
     $27 = HEAP32[$26>>2]|0;
     $28 = (FUNCTION_TABLE_ii[$27 & 8191]($12)|0);
     $$0$i = $28;
     break;
    }
   }
  } while(0);
  HEAP32[$8>>2] = $$0$i;
  $29 = ((($1)) + 1328|0);
  $30 = HEAP32[$29>>2]|0;
  $31 = ($30|0)==(65535);
  do {
   if ($31) {
    $$0$i2 = 0;
   } else {
    $32 = ($30>>>0)>(390);
    if ($32) {
     $33 = (($30) + -391)|0;
     $34 = ((($1)) + 1312|0);
     $35 = HEAP32[$34>>2]|0;
     $36 = ($35>>>0)>($33>>>0);
     if (!($36)) {
      $$0$i2 = 0;
      break;
     }
     $37 = ((($1)) + 1316|0);
     $38 = HEAP32[$37>>2]|0;
     $39 = (($38) + ($33<<2)|0);
     $40 = HEAP32[$39>>2]|0;
     $$0$i2 = $40;
     break;
    } else {
     $41 = ((($1)) + 2956|0);
     $42 = HEAP32[$41>>2]|0;
     $43 = ($42|0)==(0|0);
     if ($43) {
      $$0$i2 = 0;
      break;
     }
     $44 = ((($42)) + 20|0);
     $45 = HEAP32[$44>>2]|0;
     $46 = (FUNCTION_TABLE_ii[$45 & 8191]($30)|0);
     $$0$i2 = $46;
     break;
    }
   }
  } while(0);
  $47 = ((($8)) + 4|0);
  HEAP32[$47>>2] = $$0$i2;
  $48 = ((($1)) + 1336|0);
  $49 = HEAP32[$48>>2]|0;
  $50 = ($49|0)==(65535);
  do {
   if ($50) {
    $$0$i4 = 0;
   } else {
    $51 = ($49>>>0)>(390);
    if ($51) {
     $52 = (($49) + -391)|0;
     $53 = ((($1)) + 1312|0);
     $54 = HEAP32[$53>>2]|0;
     $55 = ($54>>>0)>($52>>>0);
     if (!($55)) {
      $$0$i4 = 0;
      break;
     }
     $56 = ((($1)) + 1316|0);
     $57 = HEAP32[$56>>2]|0;
     $58 = (($57) + ($52<<2)|0);
     $59 = HEAP32[$58>>2]|0;
     $$0$i4 = $59;
     break;
    } else {
     $60 = ((($1)) + 2956|0);
     $61 = HEAP32[$60>>2]|0;
     $62 = ($61|0)==(0|0);
     if ($62) {
      $$0$i4 = 0;
      break;
     }
     $63 = ((($61)) + 20|0);
     $64 = HEAP32[$63>>2]|0;
     $65 = (FUNCTION_TABLE_ii[$64 & 8191]($49)|0);
     $$0$i4 = $65;
     break;
    }
   }
  } while(0);
  $66 = ((($8)) + 8|0);
  HEAP32[$66>>2] = $$0$i4;
  $67 = ((($1)) + 1340|0);
  $68 = HEAP32[$67>>2]|0;
  $69 = ($68|0)==(65535);
  do {
   if ($69) {
    $$0$i6 = 0;
   } else {
    $70 = ($68>>>0)>(390);
    if ($70) {
     $71 = (($68) + -391)|0;
     $72 = ((($1)) + 1312|0);
     $73 = HEAP32[$72>>2]|0;
     $74 = ($73>>>0)>($71>>>0);
     if (!($74)) {
      $$0$i6 = 0;
      break;
     }
     $75 = ((($1)) + 1316|0);
     $76 = HEAP32[$75>>2]|0;
     $77 = (($76) + ($71<<2)|0);
     $78 = HEAP32[$77>>2]|0;
     $$0$i6 = $78;
     break;
    } else {
     $79 = ((($1)) + 2956|0);
     $80 = HEAP32[$79>>2]|0;
     $81 = ($80|0)==(0|0);
     if ($81) {
      $$0$i6 = 0;
      break;
     }
     $82 = ((($80)) + 20|0);
     $83 = HEAP32[$82>>2]|0;
     $84 = (FUNCTION_TABLE_ii[$83 & 8191]($68)|0);
     $$0$i6 = $84;
     break;
    }
   }
  } while(0);
  $85 = ((($8)) + 12|0);
  HEAP32[$85>>2] = $$0$i6;
  $86 = ((($1)) + 1344|0);
  $87 = HEAP32[$86>>2]|0;
  $88 = ($87|0)==(65535);
  do {
   if ($88) {
    $$0$i8 = 0;
   } else {
    $89 = ($87>>>0)>(390);
    if ($89) {
     $90 = (($87) + -391)|0;
     $91 = ((($1)) + 1312|0);
     $92 = HEAP32[$91>>2]|0;
     $93 = ($92>>>0)>($90>>>0);
     if (!($93)) {
      $$0$i8 = 0;
      break;
     }
     $94 = ((($1)) + 1316|0);
     $95 = HEAP32[$94>>2]|0;
     $96 = (($95) + ($90<<2)|0);
     $97 = HEAP32[$96>>2]|0;
     $$0$i8 = $97;
     break;
    } else {
     $98 = ((($1)) + 2956|0);
     $99 = HEAP32[$98>>2]|0;
     $100 = ($99|0)==(0|0);
     if ($100) {
      $$0$i8 = 0;
      break;
     }
     $101 = ((($99)) + 20|0);
     $102 = HEAP32[$101>>2]|0;
     $103 = (FUNCTION_TABLE_ii[$102 & 8191]($87)|0);
     $$0$i8 = $103;
     break;
    }
   }
  } while(0);
  $104 = ((($8)) + 16|0);
  HEAP32[$104>>2] = $$0$i8;
  $105 = ((($1)) + 1352|0);
  $106 = HEAP32[$105>>2]|0;
  $107 = ((($8)) + 20|0);
  HEAP32[$107>>2] = $106;
  $108 = ((($1)) + 1348|0);
  $109 = HEAP8[$108>>0]|0;
  $110 = ((($8)) + 24|0);
  HEAP8[$110>>0] = $109;
  $111 = ((($1)) + 1356|0);
  $112 = HEAP32[$111>>2]|0;
  $113 = $112&65535;
  $114 = ((($8)) + 26|0);
  HEAP16[$114>>1] = $113;
  $115 = ((($1)) + 1360|0);
  $116 = HEAP32[$115>>2]|0;
  $117 = $116&65535;
  $118 = ((($8)) + 28|0);
  HEAP16[$118>>1] = $117;
  HEAP32[$3>>2] = $8;
  $$pre$pre = HEAP32[$error>>2]|0;
  $$pre = $$pre$pre;$119 = $8;
 } else {
  $$pre = 0;$119 = $4;
 }
 ;HEAP32[$afont_info>>2]=HEAP32[$119>>2]|0;HEAP32[$afont_info+4>>2]=HEAP32[$119+4>>2]|0;HEAP32[$afont_info+8>>2]=HEAP32[$119+8>>2]|0;HEAP32[$afont_info+12>>2]=HEAP32[$119+12>>2]|0;HEAP32[$afont_info+16>>2]=HEAP32[$119+16>>2]|0;HEAP32[$afont_info+20>>2]=HEAP32[$119+20>>2]|0;HEAP32[$afont_info+24>>2]=HEAP32[$119+24>>2]|0;HEAP32[$afont_info+28>>2]=HEAP32[$119+28>>2]|0;
 $120 = $$pre;
 STACKTOP = sp;return ($120|0);
}
function _cff_ps_has_glyph_names($face) {
 $face = $face|0;
 var $$lobit = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 >>> 9;
 $$lobit = $2 & 1;
 return ($$lobit|0);
}
function _cff_get_ps_name($face) {
 $face = $face|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 656|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 532|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  $5 = ((($face)) + 96|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ((($6)) + 4|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = (_FT_Get_Module($8,593168)|0);
  $10 = (_ft_module_get_service($9,594316)|0);
  $11 = ($10|0)==(0|0);
  if (!($11)) {
   $12 = HEAP32[$10>>2]|0;
   $13 = ($12|0)==(0|0);
   if (!($13)) {
    $14 = (FUNCTION_TABLE_ii[$12 & 8191]($face)|0);
    $$1 = $14;
    return ($$1|0);
   }
  }
 }
 $15 = ((($1)) + 1304|0);
 $16 = HEAP32[$15>>2]|0;
 $$1 = $16;
 return ($$1|0);
}
function _cff_get_glyph_name($face,$glyph_index,$buffer,$buffer_max) {
 $face = $face|0;
 $glyph_index = $glyph_index|0;
 $buffer = $buffer|0;
 $buffer_max = $buffer_max|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 656|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 2956|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  $error$0 = 11;
  return ($error$0|0);
 }
 $5 = ((($1)) + 1160|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + ($glyph_index<<1)|0);
 $8 = HEAP16[$7>>1]|0;
 $9 = $8&65535;
 $10 = ($8<<16>>16)==(-1);
 if ($10) {
  $error$0 = 0;
  return ($error$0|0);
 }
 $11 = ($8&65535)>(390);
 do {
  if ($11) {
   $12 = (($9) + -391)|0;
   $13 = ((($1)) + 1312|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ($14>>>0)>($12>>>0);
   if ($15) {
    $16 = ((($1)) + 1316|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = (($17) + ($12<<2)|0);
    $19 = HEAP32[$18>>2]|0;
    $$0$i = $19;
    break;
   } else {
    $error$0 = 0;
    return ($error$0|0);
   }
  } else {
   $20 = ((($3)) + 20|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = (FUNCTION_TABLE_ii[$21 & 8191]($9)|0);
   $$0$i = $22;
  }
 } while(0);
 $23 = ($$0$i|0)==(0|0);
 if ($23) {
  $error$0 = 0;
  return ($error$0|0);
 }
 (_ft_mem_strcpyn($buffer,$$0$i,$buffer_max)|0);
 $error$0 = 0;
 return ($error$0|0);
}
function _cff_get_name_index($face,$glyph_name) {
 $face = $face|0;
 $glyph_name = $glyph_name|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $name$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 656|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 96|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_ft_module_get_service($3,594361)|0);
 $5 = ($4|0)==(0|0);
 if ($5) {
  $$0 = 0;
  return ($$0|0);
 }
 $6 = ((($1)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0);
 if ($8) {
  $$0 = 0;
  return ($$0|0);
 }
 $9 = ((($1)) + 1160|0);
 $10 = ((($1)) + 1312|0);
 $11 = ((($1)) + 1316|0);
 $12 = ((($4)) + 20|0);
 $i$01 = 0;
 while(1) {
  $13 = HEAP32[$9>>2]|0;
  $14 = (($13) + ($i$01<<1)|0);
  $15 = HEAP16[$14>>1]|0;
  $16 = $15&65535;
  $17 = ($15&65535)>(390);
  if ($17) {
   $18 = (($16) + -391)|0;
   $19 = HEAP32[$10>>2]|0;
   $20 = ($19>>>0)>($18>>>0);
   if ($20) {
    $21 = HEAP32[$11>>2]|0;
    $22 = (($21) + ($18<<2)|0);
    $23 = HEAP32[$22>>2]|0;
    $name$0 = $23;
    label = 8;
   }
  } else {
   $24 = HEAP32[$12>>2]|0;
   $25 = (FUNCTION_TABLE_ii[$24 & 8191]($16)|0);
   $name$0 = $25;
   label = 8;
  }
  if ((label|0) == 8) {
   label = 0;
   $26 = ($name$0|0)==(0|0);
   if (!($26)) {
    $27 = (_strcmp($glyph_name,$name$0)|0);
    $28 = ($27|0)==(0);
    if ($28) {
     $$0 = $i$01;
     label = 11;
     break;
    }
   }
  }
  $29 = (($i$01) + 1)|0;
  $30 = HEAP32[$6>>2]|0;
  $31 = ($29>>>0)<($30>>>0);
  if ($31) {
   $i$01 = $29;
  } else {
   $$0 = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 11) {
  return ($$0|0);
 }
 return (0)|0;
}
function _cff_get_cmap_info($charmap,$cmap_info) {
 $charmap = $charmap|0;
 $cmap_info = $cmap_info|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$1 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$charmap>>2]|0;
 $1 = ((($0)) + 96|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 HEAP32[$cmap_info>>2] = 0;
 $5 = ((($cmap_info)) + 4|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($charmap)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(49564|0);
 $9 = ($7|0)==(49604|0);
 $or$cond = $8 | $9;
 if ($or$cond) {
  $error$1 = 0;
  return ($error$1|0);
 }
 $10 = (_FT_Get_Module($4,593168)|0);
 $11 = (_ft_module_get_service($10,593070)|0);
 $12 = ($11|0)==(0|0);
 if ($12) {
  $error$1 = 0;
  return ($error$1|0);
 }
 $13 = HEAP32[$11>>2]|0;
 $14 = ($13|0)==(0|0);
 if ($14) {
  $error$1 = 0;
  return ($error$1|0);
 }
 $15 = (FUNCTION_TABLE_iii[$13 & 8191]($charmap,$cmap_info)|0);
 $error$1 = $15;
 return ($error$1|0);
}
function _cff_cmap_encoding_init($cmap,$pointer) {
 $cmap = $cmap|0;
 $pointer = $pointer|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$cmap>>2]|0;
 $1 = ((($0)) + 656|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 640|0);
 $4 = ((($cmap)) + 16|0);
 HEAP32[$4>>2] = $3;
 return 0;
}
function _cff_cmap_encoding_done($cmap) {
 $cmap = $cmap|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 HEAP32[$0>>2] = 0;
 return;
}
function _cff_cmap_encoding_char_index($cmap,$char_code) {
 $cmap = $cmap|0;
 $char_code = $char_code|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($char_code>>>0)<(256);
 if (!($0)) {
  $result$0 = 0;
  return ($result$0|0);
 }
 $1 = ((($cmap)) + 16|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($2) + ($char_code<<1)|0);
 $4 = HEAP16[$3>>1]|0;
 $5 = $4&65535;
 $result$0 = $5;
 return ($result$0|0);
}
function _cff_cmap_encoding_char_next($cmap,$pchar_code) {
 $cmap = $cmap|0;
 $pchar_code = $pchar_code|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $code$0 = 0, $code$0$in = 0, $code$0$lcssa12 = 0, $result$0 = 0, $result$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$pchar_code>>2]|0;
 HEAP32[$pchar_code>>2] = 0;
 $1 = ($0>>>0)<(255);
 if (!($1)) {
  $result$2 = 0;
  return ($result$2|0);
 }
 $2 = ((($cmap)) + 16|0);
 $code$0$in = $0;$result$0 = 0;
 while(1) {
  $code$0 = (($code$0$in) + 1)|0;
  $3 = ($code$0>>>0)>(255);
  if ($3) {
   $result$2 = $result$0;
   label = 6;
   break;
  }
  $4 = HEAP32[$2>>2]|0;
  $5 = (($4) + ($code$0<<1)|0);
  $6 = HEAP16[$5>>1]|0;
  $7 = $6&65535;
  $8 = ($6<<16>>16)==(0);
  if ($8) {
   $code$0$in = $code$0;$result$0 = $7;
  } else {
   $$lcssa = $7;$code$0$lcssa12 = $code$0;
   break;
  }
 }
 if ((label|0) == 6) {
  return ($result$2|0);
 }
 HEAP32[$pchar_code>>2] = $code$0$lcssa12;
 $result$2 = $$lcssa;
 return ($result$2|0);
}
function _cff_cmap_unicode_init($unicodes,$pointer) {
 $unicodes = $unicodes|0;
 $pointer = $pointer|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$unicodes>>2]|0;
 $1 = ((($0)) + 656|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 1160|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if ($5) {
  $$0 = 163;
  return ($$0|0);
 }
 $6 = ((($2)) + 2956|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($0)) + 100|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($7)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($2)) + 12|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (FUNCTION_TABLE_iiiiiii[$11 & 8191]($9,$unicodes,$13,4356,0,$0)|0);
 $$0 = $14;
 return ($$0|0);
}
function _cff_sid_to_glyph_name($face,$idx) {
 $face = $face|0;
 $idx = $idx|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 656|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 1160|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + ($idx<<1)|0);
 $5 = HEAP16[$4>>1]|0;
 $6 = $5&65535;
 $7 = ($5<<16>>16)==(-1);
 if ($7) {
  $$0$i = 0;
  return ($$0$i|0);
 }
 $8 = ($5&65535)>(390);
 if ($8) {
  $9 = (($6) + -391)|0;
  $10 = ((($1)) + 1312|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($11>>>0)>($9>>>0);
  if (!($12)) {
   $$0$i = 0;
   return ($$0$i|0);
  }
  $13 = ((($1)) + 1316|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = (($14) + ($9<<2)|0);
  $16 = HEAP32[$15>>2]|0;
  $$0$i = $16;
  return ($$0$i|0);
 } else {
  $17 = ((($1)) + 2956|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = ($18|0)==(0|0);
  if ($19) {
   $$0$i = 0;
   return ($$0$i|0);
  }
  $20 = ((($18)) + 20|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = (FUNCTION_TABLE_ii[$21 & 8191]($6)|0);
  $$0$i = $22;
  return ($$0$i|0);
 }
 return (0)|0;
}
function _cff_cmap_unicode_done($unicodes) {
 $unicodes = $unicodes|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$unicodes>>2]|0;
 $1 = ((($0)) + 100|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($unicodes)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 _ft_mem_free($2,$4);
 HEAP32[$3>>2] = 0;
 $5 = ((($unicodes)) + 16|0);
 HEAP32[$5>>2] = 0;
 return;
}
function _cff_cmap_unicode_char_index($unicodes,$char_code) {
 $unicodes = $unicodes|0;
 $char_code = $char_code|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$unicodes>>2]|0;
 $1 = ((($0)) + 656|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 2956|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (FUNCTION_TABLE_iii[$6 & 8191]($unicodes,$char_code)|0);
 return ($7|0);
}
function _cff_cmap_unicode_char_next($unicodes,$pchar_code) {
 $unicodes = $unicodes|0;
 $pchar_code = $pchar_code|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$unicodes>>2]|0;
 $1 = ((($0)) + 656|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 2956|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 12|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (FUNCTION_TABLE_iii[$6 & 8191]($unicodes,$pchar_code)|0);
 return ($7|0);
}
function _cff_get_ros($face,$registry,$ordering,$supplement) {
 $face = $face|0;
 $registry = $registry|0;
 $ordering = $ordering|0;
 $supplement = $supplement|0;
 var $$0$i = 0, $$0$i2 = 0, $$01 = 0, $$in = 0, $$in4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 656|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$01 = 0;
  return ($$01|0);
 }
 $3 = ((($1)) + 1456|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(65535);
 if ($5) {
  $$01 = 6;
  return ($$01|0);
 }
 $6 = ($registry|0)==(0|0);
 if (!($6)) {
  $7 = ((($1)) + 2964|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8|0)==(0|0);
  if ($9) {
   $10 = ($4>>>0)>(390);
   if ($10) {
    $11 = (($4) + -391)|0;
    $12 = ((($1)) + 1312|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ($13>>>0)>($11>>>0);
    if ($14) {
     $15 = ((($1)) + 1316|0);
     $16 = HEAP32[$15>>2]|0;
     $17 = (($16) + ($11<<2)|0);
     $18 = HEAP32[$17>>2]|0;
     $$0$i = $18;
    } else {
     $$0$i = 0;
    }
   } else {
    $19 = ((($1)) + 2956|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = ($20|0)==(0|0);
    if ($21) {
     $$0$i = 0;
    } else {
     $22 = ((($20)) + 20|0);
     $23 = HEAP32[$22>>2]|0;
     $24 = (FUNCTION_TABLE_ii[$23 & 8191]($4)|0);
     $$0$i = $24;
    }
   }
   HEAP32[$7>>2] = $$0$i;
   $$in = $$0$i;
  } else {
   $$in = $8;
  }
  $25 = $$in;
  HEAP32[$registry>>2] = $25;
 }
 $26 = ($ordering|0)==(0|0);
 if (!($26)) {
  $27 = ((($1)) + 2968|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = ($28|0)==(0|0);
  if ($29) {
   $30 = ((($1)) + 1460|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = ($31|0)==(65535);
   do {
    if ($32) {
     $$0$i2 = 0;
    } else {
     $33 = ($31>>>0)>(390);
     if ($33) {
      $34 = (($31) + -391)|0;
      $35 = ((($1)) + 1312|0);
      $36 = HEAP32[$35>>2]|0;
      $37 = ($36>>>0)>($34>>>0);
      if (!($37)) {
       $$0$i2 = 0;
       break;
      }
      $38 = ((($1)) + 1316|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = (($39) + ($34<<2)|0);
      $41 = HEAP32[$40>>2]|0;
      $$0$i2 = $41;
      break;
     } else {
      $42 = ((($1)) + 2956|0);
      $43 = HEAP32[$42>>2]|0;
      $44 = ($43|0)==(0|0);
      if ($44) {
       $$0$i2 = 0;
       break;
      }
      $45 = ((($43)) + 20|0);
      $46 = HEAP32[$45>>2]|0;
      $47 = (FUNCTION_TABLE_ii[$46 & 8191]($31)|0);
      $$0$i2 = $47;
      break;
     }
    }
   } while(0);
   HEAP32[$27>>2] = $$0$i2;
   $$in4 = $$0$i2;
  } else {
   $$in4 = $28;
  }
  $48 = $$in4;
  HEAP32[$ordering>>2] = $48;
 }
 $49 = ($supplement|0)==(0|0);
 if ($49) {
  $$01 = 0;
  return ($$01|0);
 }
 $50 = ((($1)) + 1464|0);
 $51 = HEAP32[$50>>2]|0;
 HEAP32[$supplement>>2] = $51;
 $$01 = 0;
 return ($$01|0);
}
function _cff_get_is_cid($face,$is_cid) {
 $face = $face|0;
 $is_cid = $is_cid|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 656|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP8[$is_cid>>0] = 0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return 0;
 }
 $3 = ((($1)) + 1456|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(65535);
 if ($5) {
  return 0;
 }
 HEAP8[$is_cid>>0] = 1;
 return 0;
}
function _cff_get_cid_from_glyph_index($face,$glyph_index,$cid) {
 $face = $face|0;
 $glyph_index = $glyph_index|0;
 $cid = $cid|0;
 var $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 656|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$01 = 0;
  return ($$01|0);
 }
 $3 = ((($1)) + 1456|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(65535);
 if ($5) {
  $$01 = 6;
  return ($$01|0);
 }
 $6 = ((($1)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7>>>0)<($glyph_index>>>0);
 if ($8) {
  $$01 = 6;
  return ($$01|0);
 }
 $9 = ($cid|0)==(0|0);
 if ($9) {
  $$01 = 0;
  return ($$01|0);
 }
 $10 = ((($1)) + 1160|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + ($glyph_index<<1)|0);
 $13 = HEAP16[$12>>1]|0;
 $14 = $13&65535;
 HEAP32[$cid>>2] = $14;
 $$01 = 0;
 return ($$01|0);
}
function _cff_property_set($module,$property_name,$value) {
 $module = $module|0;
 $property_name = $property_name|0;
 $value = $value|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond15 = 0, $or$cond17 = 0, $or$cond19 = 0, $or$cond21 = 0, $or$cond22 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strcmp($property_name,532020)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  $2 = HEAP32[$value>>2]|0;
  $3 = ((($value)) + 4|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ((($value)) + 8|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ((($value)) + 12|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ((($value)) + 16|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ((($value)) + 20|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ((($value)) + 24|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ((($value)) + 28|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = $4 | $2;
  $18 = $17 | $6;
  $19 = $18 | $8;
  $20 = $19 | $10;
  $21 = $20 | $12;
  $22 = $21 | $14;
  $23 = $22 | $16;
  $24 = ($23|0)<(0);
  $25 = ($2|0)>($6|0);
  $or$cond = $25 | $24;
  $26 = ($6|0)>($10|0);
  $or$cond22 = $26 | $or$cond;
  if ($or$cond22) {
   $$1 = 6;
   return ($$1|0);
  }
  $27 = ($10|0)>($14|0);
  $28 = ($4|0)>(500);
  $or$cond15 = $28 | $27;
  $29 = ($8|0)>(500);
  $or$cond17 = $29 | $or$cond15;
  $30 = ($12|0)>(500);
  $or$cond19 = $30 | $or$cond17;
  $31 = ($16|0)>(500);
  $or$cond21 = $31 | $or$cond19;
  if ($or$cond21) {
   $$1 = 6;
   return ($$1|0);
  }
  $32 = ((($module)) + 36|0);
  HEAP32[$32>>2] = $2;
  $33 = ((($module)) + 40|0);
  HEAP32[$33>>2] = $4;
  $34 = ((($module)) + 44|0);
  HEAP32[$34>>2] = $6;
  $35 = ((($module)) + 48|0);
  HEAP32[$35>>2] = $8;
  $36 = ((($module)) + 52|0);
  HEAP32[$36>>2] = $10;
  $37 = ((($module)) + 56|0);
  HEAP32[$37>>2] = $12;
  $38 = ((($module)) + 60|0);
  HEAP32[$38>>2] = $14;
  $39 = ((($module)) + 64|0);
  HEAP32[$39>>2] = $16;
  $$1 = 0;
  return ($$1|0);
 }
 $40 = (_strcmp($property_name,532041)|0);
 $41 = ($40|0)==(0);
 if ($41) {
  $42 = HEAP32[$value>>2]|0;
  $43 = ($42|0)==(1);
  if (!($43)) {
   $$1 = 7;
   return ($$1|0);
  }
  $44 = ((($module)) + 28|0);
  HEAP32[$44>>2] = 1;
  $$1 = 0;
  return ($$1|0);
 } else {
  $45 = (_strcmp($property_name,532056)|0);
  $46 = ($45|0)==(0);
  if (!($46)) {
   $$1 = 12;
   return ($$1|0);
  }
  $47 = HEAP8[$value>>0]|0;
  $48 = ((($module)) + 32|0);
  HEAP8[$48>>0] = $47;
  $$1 = 0;
  return ($$1|0);
 }
 return (0)|0;
}
function _cff_property_get($module,$property_name,$value) {
 $module = $module|0;
 $property_name = $property_name|0;
 $value = $value|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strcmp($property_name,532020)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  $2 = ((($module)) + 36|0);
  $3 = HEAP32[$2>>2]|0;
  HEAP32[$value>>2] = $3;
  $4 = ((($module)) + 40|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ((($value)) + 4|0);
  HEAP32[$6>>2] = $5;
  $7 = ((($module)) + 44|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ((($value)) + 8|0);
  HEAP32[$9>>2] = $8;
  $10 = ((($module)) + 48|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ((($value)) + 12|0);
  HEAP32[$12>>2] = $11;
  $13 = ((($module)) + 52|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ((($value)) + 16|0);
  HEAP32[$15>>2] = $14;
  $16 = ((($module)) + 56|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = ((($value)) + 20|0);
  HEAP32[$18>>2] = $17;
  $19 = ((($module)) + 60|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = ((($value)) + 24|0);
  HEAP32[$21>>2] = $20;
  $22 = ((($module)) + 64|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = ((($value)) + 28|0);
  HEAP32[$24>>2] = $23;
  $$0 = 0;
  return ($$0|0);
 }
 $25 = (_strcmp($property_name,532041)|0);
 $26 = ($25|0)==(0);
 if ($26) {
  $27 = ((($module)) + 28|0);
  $28 = HEAP32[$27>>2]|0;
  HEAP32[$value>>2] = $28;
  $$0 = 0;
  return ($$0|0);
 }
 $29 = (_strcmp($property_name,532056)|0);
 $30 = ($29|0)==(0);
 if (!($30)) {
  $$0 = 12;
  return ($$0|0);
 }
 $31 = ((($module)) + 32|0);
 $32 = HEAP8[$31>>0]|0;
 HEAP8[$value>>0] = $32;
 $$0 = 0;
 return ($$0|0);
}
function _cff_face_init($stream,$cffface,$face_index,$num_params,$params) {
 $stream = $stream|0;
 $cffface = $cffface|0;
 $face_index = $face_index|0;
 $num_params = $num_params|0;
 $params = $params|0;
 var $$ = 0, $$0 = 0, $$0$i = 0, $$0$i14 = 0, $$0$i20 = 0, $$0$i24 = 0, $$121 = 0, $$5 = 0, $$6 = 0, $$7 = 0, $$mask = 0, $$max_cid$0$i$i$i = 0, $$max_cid$0$i$i$i$lcssa = 0, $$max_cid$0$i$i16$i = 0, $$nleft$0$i$i = 0, $$off$us$1$i = 0, $$off$us$2$i = 0, $$off$us$3$i = 0, $$off$us$4$i = 0, $$off$us$5$i = 0;
 var $$off$us$i = 0, $$ph$i = 0, $$phi$trans$insert = 0, $$phi$trans$insert79 = 0, $$pr$pre$i$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i10$i = 0, $$pre$phi$i$iZ2D = 0, $$pre$phi$i26$iZ2D = 0, $$pre$phi46$i$i$inZ2D = 0, $$pre$phi85Z2D = 0, $$pre$phi87Z2D = 0, $$pre$phi88Z2D = 0, $$pre$phi95Z2D = 0, $$pre$phi99Z2D = 0, $$pre$phiZ2D = 0, $$pre43$i$i = 0, $$pre44$i$i = 0;
 var $$pre45$i$i = 0, $$pre55$i = 0, $$pre80 = 0, $$pre81 = 0, $$pre82 = 0, $$pre86 = 0, $$pre92 = 0, $$pre98 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0;
 var $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0;
 var $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0;
 var $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0;
 var $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0;
 var $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0;
 var $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0;
 var $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0;
 var $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0;
 var $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0;
 var $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0;
 var $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0;
 var $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0;
 var $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0;
 var $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0;
 var $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0;
 var $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0;
 var $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0;
 var $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0;
 var $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0;
 var $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0;
 var $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0;
 var $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0;
 var $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0;
 var $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0;
 var $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0;
 var $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0;
 var $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0;
 var $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0;
 var $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0;
 var $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0;
 var $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0;
 var $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0;
 var $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0;
 var $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0;
 var $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0;
 var $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0;
 var $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0;
 var $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $byte_len$i = 0, $byte_len$i$i = 0, $bytes$i = 0, $bytes$i$i = 0, $cmaprec = 0, $continue_search$0$ph10$i55 = 0, $continue_search$2$us$5$i = 0, $error = 0, $error$i = 0, $error$i$i = 0, $error$i$i$i = 0, $error$i$i5$i = 0, $error$i11 = 0, $error$i16 = 0;
 var $error$i22 = 0, $error$i23 = 0, $error$i27$i = 0, $error$i4$i = 0, $error$i6$i = 0, $exitcond = 0, $exitcond$i$i = 0, $exitcond$i$i$i = 0, $exitcond$i$i17$i = 0, $exitcond$i8$i = 0, $exitcond11$i = 0, $exitcond39$i$i = 0, $exitcond40$i$i = 0, $family$0 = 0, $family$1$ph$be = 0, $family$1$ph52 = 0, $family$148 = 0, $family$148$lcssa = 0, $fd_index$i = 0, $flags$1 = 0;
 var $flags$3 = 0, $flags$4 = 0, $fullp$0$ph51 = 0, $fullp$0$ph51$lcssa158 = 0, $gindex$015$us$i$i = 0, $gindex$015$us$i$i$lcssa = 0, $glyph_code$023$i$i = 0, $glyph_sid$09$i$i = 0, $i$011$i$i = 0, $i$027$i$i = 0, $i$04$i$i$i = 0, $i$04$i$i15$i = 0, $i$056 = 0, $idx$010$i = 0, $idx$044$i = 0, $idx$143$i = 0, $idx$15$i = 0, $idx$19$i = 0, $idx$19$in$i = 0, $idx$19$in$i$lcssa = 0;
 var $indvars$iv$i53 = 0, $indvars$iv$next$i = 0, $j$03$i$i$i = 0, $j$03$i$i23$i = 0, $j$03$in$i$i$i = 0, $j$03$in$i$i22$i = 0, $j$033$i$i = 0, $j$07$i$i = 0, $j$114$i$i = 0, $j$120$i$i = 0, $j$2$lcssa$i$i = 0, $j$210$i$i = 0, $j$226$i$i = 0, $j$316$i$i = 0, $j$316$us$i$i = 0, $j$431$i$i = 0, $k$024$i$i = 0, $max_cid$0$lcssa$i$i18$i = 0, $max_cid$05$i$i$i = 0, $max_cid$05$i$i14$i = 0;
 var $name$0$i = 0, $name$0$i$i = 0, $nleft$0$i$i = 0, $nn$047 = 0, $not$ = 0, $or$cond = 0, $or$cond$i = 0, $or$cond$i$i = 0, $or$cond$i7$i = 0, $or$cond10 = 0, $or$cond3 = 0, $or$cond34$i$i = 0, $or$cond8 = 0, $p$021$i$i = 0, $phitmp$i$i = 0, $pure_cff$0 = 0, $pure_cff$1 = 0, $scaling$0 = 0, $sext = 0, $sfnt_format$0 = 0;
 var $storemerge = 0, $storemerge102 = 0, $string_index$i = 0, $subfont_index$0$i = 0, $switch$split2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i23 = sp + 136|0;
 $error$i22 = sp + 132|0;
 $error$i16 = sp + 128|0;
 $bytes$i = sp + 124|0;
 $byte_len$i = sp + 120|0;
 $error$i11 = sp + 116|0;
 $bytes$i$i = sp + 112|0;
 $byte_len$i$i = sp + 108|0;
 $error$i27$i = sp + 104|0;
 $error$i$i5$i = sp + 100|0;
 $error$i6$i = sp + 96|0;
 $error$i$i$i = sp + 92|0;
 $error$i4$i = sp + 88|0;
 $error$i$i = sp + 84|0;
 $error$i = sp + 80|0;
 $string_index$i = sp + 48|0;
 $fd_index$i = sp + 16|0;
 $error = sp + 12|0;
 $cmaprec = sp;
 $0 = ((($cffface)) + 96|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_FT_Get_Module_Interface($3,593168)|0);
 $5 = ($4|0)==(0|0);
 if ($5) {
  HEAP32[$error>>2] = 11;
  $$0 = 11;
  STACKTOP = sp;return ($$0|0);
 }
 $6 = HEAP32[$0>>2]|0;
 $7 = (_ft_module_get_service($6,594361)|0);
 $8 = (_FT_Get_Module_Interface($3,593918)|0);
 $9 = (_FT_Stream_Seek($stream,0)|0);
 HEAP32[$error>>2] = $9;
 $10 = ($9|0)==(0);
 if (!($10)) {
  $$0 = $9;
  STACKTOP = sp;return ($$0|0);
 }
 $11 = ((($4)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (FUNCTION_TABLE_iiiiii[$12 & 8191]($stream,$cffface,$face_index,$num_params,$params)|0);
 HEAP32[$error>>2] = $13;
 $14 = ($13|0)==(0);
 do {
  if ($14) {
   $15 = ((($cffface)) + 148|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = ($16|0)==(1330926671);
   if (!($17)) {
    HEAP32[$error>>2] = 2;
    $$0 = 2;
    STACKTOP = sp;return ($$0|0);
   }
   $18 = ($face_index|0)<(0);
   if ($18) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   $19 = ((($cffface)) + 508|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = (FUNCTION_TABLE_iiiii[$20 & 8191]($cffface,1751474532,$stream,0)|0);
   HEAP32[$error>>2] = $21;
   $22 = ($21|0)==(0);
   if ($22) {
    $23 = ((($4)) + 8|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = (FUNCTION_TABLE_iiiiii[$24 & 8191]($stream,$cffface,$face_index,$num_params,$params)|0);
    HEAP32[$error>>2] = $25;
    $26 = ($25|0)==(0);
    if ($26) {
     $pure_cff$0 = 0;
    } else {
     $$0 = $25;
     STACKTOP = sp;return ($$0|0);
    }
   } else {
    $27 = ((($4)) + 32|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = (FUNCTION_TABLE_iii[$28 & 8191]($cffface,$stream)|0);
    HEAP32[$error>>2] = $29;
    $30 = ($29|0)==(0);
    if ($30) {
     $pure_cff$0 = 1;
    } else {
     $$0 = $29;
     STACKTOP = sp;return ($$0|0);
    }
   }
   $31 = HEAP32[$19>>2]|0;
   $32 = (FUNCTION_TABLE_iiiii[$31 & 8191]($cffface,1128678944,$stream,0)|0);
   HEAP32[$error>>2] = $32;
   $33 = ($32|0)==(0);
   if ($33) {
    $pure_cff$1 = $pure_cff$0;$sfnt_format$0 = 1;
   } else {
    $$0 = $32;
    STACKTOP = sp;return ($$0|0);
   }
  } else {
   $34 = (_FT_Stream_Seek($stream,0)|0);
   HEAP32[$error>>2] = $34;
   $35 = ($34|0)==(0);
   if ($35) {
    HEAP32[$error>>2] = 0;
    $pure_cff$1 = 1;$sfnt_format$0 = 0;
    break;
   } else {
    $$0 = $34;
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 $36 = ((($cffface)) + 100|0);
 $37 = HEAP32[$36>>2]|0;
 $38 = (_ft_mem_alloc($37,2980,$error)|0);
 $39 = HEAP32[$error>>2]|0;
 $40 = ($39|0)==(0);
 if (!($40)) {
  $$0 = $39;
  STACKTOP = sp;return ($$0|0);
 }
 $41 = ((($cffface)) + 656|0);
 HEAP32[$41>>2] = $38;
 $42 = ((($stream)) + 28|0);
 $43 = HEAP32[$42>>2]|0;
 _memset(($38|0),0,2980)|0;
 ;HEAP32[$string_index$i>>2]=0|0;HEAP32[$string_index$i+4>>2]=0|0;HEAP32[$string_index$i+8>>2]=0|0;HEAP32[$string_index$i+12>>2]=0|0;HEAP32[$string_index$i+16>>2]=0|0;HEAP32[$string_index$i+20>>2]=0|0;HEAP32[$string_index$i+24>>2]=0|0;HEAP32[$string_index$i+28>>2]=0|0;
 HEAP32[$38>>2] = $stream;
 $44 = ((($38)) + 4|0);
 HEAP32[$44>>2] = $43;
 $45 = ((($38)) + 1324|0);
 $46 = (_FT_Stream_Pos($stream)|0);
 $47 = (_FT_Stream_ReadFields($stream,169836,$38)|0);
 HEAP32[$error$i>>2] = $47;
 $48 = ($47|0)==(0);
 L32: do {
  if ($48) {
   $49 = ((($38)) + 16|0);
   $50 = HEAP32[$49>>2]|0;
   $51 = $50&255;
   $52 = ($51<<24>>24)==(1);
   $53 = $50 >>> 16;
   if ($52) {
    $54 = $50 >>> 24;
    $55 = $54&255;
    $56 = $53&255;
    $57 = ($56&255)<(4);
    $58 = ($55&255)>(4);
    $or$cond$i = $57 | $58;
    if (!($or$cond$i)) {
     $59 = $53 & 255;
     $60 = (($59) + -4)|0;
     $61 = (_FT_Stream_Skip($stream,$60)|0);
     HEAP32[$error$i>>2] = $61;
     $62 = ($61|0)==(0);
     if (!($62)) {
      break;
     }
     $63 = ((($38)) + 20|0);
     $64 = (_cff_index_init($63,$stream,0)|0);
     HEAP32[$error$i>>2] = $64;
     $65 = ($64|0)==(0);
     if (!($65)) {
      break;
     }
     $66 = ((($38)) + 1208|0);
     $67 = (_cff_index_init($66,$stream,0)|0);
     HEAP32[$error$i>>2] = $67;
     $68 = ($67|0)==(0);
     if (!($68)) {
      break;
     }
     $69 = (_cff_index_init($string_index$i,$stream,1)|0);
     HEAP32[$error$i>>2] = $69;
     $70 = ($69|0)==(0);
     if (!($70)) {
      break;
     }
     $71 = ((($38)) + 84|0);
     $72 = (_cff_index_init($71,$stream,1)|0);
     HEAP32[$error$i>>2] = $72;
     $73 = ($72|0)==(0);
     if (!($73)) {
      break;
     }
     $74 = ((($38)) + 1316|0);
     $75 = ((($38)) + 1320|0);
     $76 = (_cff_index_get_pointers($string_index$i,$74,$75)|0);
     HEAP32[$error$i>>2] = $76;
     $77 = ($76|0)==(0);
     if (!($77)) {
      break;
     }
     $78 = ((($string_index$i)) + 8|0);
     $79 = HEAP32[$78>>2]|0;
     $80 = ((($38)) + 1312|0);
     HEAP32[$80>>2] = $79;
     $81 = ($pure_cff$1<<24>>24)==(0);
     $82 = ((($38)) + 28|0);
     $83 = HEAP32[$82>>2]|0;
     do {
      if ($81) {
       $86 = ($83>>>0)>(1);
       if ($86) {
        HEAP32[$error$i>>2] = 3;
        break L32;
       } else {
        $subfont_index$0$i = 0;
       }
      } else {
       $84 = ($83>>>0)>($face_index>>>0);
       if ($84) {
        $85 = ((($38)) + 8|0);
        HEAP32[$85>>2] = $83;
        $subfont_index$0$i = $face_index;
        break;
       } else {
        HEAP32[$error$i>>2] = 6;
        break L32;
       }
      }
     } while(0);
     $87 = ($face_index|0)<(0);
     if ($87) {
      break;
     }
     $88 = (_cff_subfont_load($45,$66,$subfont_index$0$i,$stream,$46,$3)|0);
     HEAP32[$error$i>>2] = $88;
     $89 = ($88|0)==(0);
     if (!($89)) {
      break;
     }
     $90 = ((($38)) + 1436|0);
     $91 = HEAP32[$90>>2]|0;
     $92 = (($91) + ($46))|0;
     $93 = (_FT_Stream_Seek($stream,$92)|0);
     HEAP32[$error$i>>2] = $93;
     $94 = ($93|0)==(0);
     if (!($94)) {
      break;
     }
     $95 = ((($38)) + 1176|0);
     $96 = (_cff_index_init($95,$stream,0)|0);
     HEAP32[$error$i>>2] = $96;
     $97 = ($96|0)==(0);
     if (!($97)) {
      break;
     }
     $98 = ((($38)) + 1456|0);
     $99 = HEAP32[$98>>2]|0;
     $100 = ($99|0)==(65535);
     L55: do {
      if ($100) {
       $164 = ((($38)) + 1896|0);
       HEAP32[$164>>2] = 0;
      } else {
       $101 = ((($38)) + 1488|0);
       $102 = HEAP32[$101>>2]|0;
       $103 = (($102) + ($46))|0;
       $104 = (_FT_Stream_Seek($stream,$103)|0);
       HEAP32[$error$i>>2] = $104;
       $105 = ($104|0)==(0);
       do {
        if ($105) {
         $106 = (_cff_index_init($fd_index$i,$stream,0)|0);
         HEAP32[$error$i>>2] = $106;
         $107 = ($106|0)==(0);
         if (!($107)) {
          break;
         }
         $108 = ((($fd_index$i)) + 8|0);
         $109 = HEAP32[$108>>2]|0;
         $110 = ($109>>>0)>(256);
         L61: do {
          if ($110) {
           $877 = 0;
          } else {
           $111 = ((($38)) + 1896|0);
           HEAP32[$111>>2] = $109;
           $112 = (_ft_mem_realloc($43,572,0,$109,0,$error$i)|0);
           $113 = HEAP32[$error$i>>2]|0;
           $114 = ($113|0)==(0);
           if (!($114)) {
            $877 = $113;
            break;
           }
           $115 = HEAP32[$108>>2]|0;
           $116 = ($115|0)==(0);
           if (!($116)) {
            $idx$044$i = 0;
            while(1) {
             $117 = (($112) + (($idx$044$i*572)|0)|0);
             $118 = (((($38)) + 1900|0) + ($idx$044$i<<2)|0);
             HEAP32[$118>>2] = $117;
             $119 = (($idx$044$i) + 1)|0;
             $exitcond = ($119|0)==($115|0);
             if ($exitcond) {
              $idx$143$i = 0;
              break;
             } else {
              $idx$044$i = $119;
             }
            }
            while(1) {
             $123 = (((($38)) + 1900|0) + ($idx$143$i<<2)|0);
             $124 = HEAP32[$123>>2]|0;
             $125 = (_cff_subfont_load($124,$fd_index$i,$idx$143$i,$stream,$46,$3)|0);
             HEAP32[$error$i>>2] = $125;
             $126 = ($125|0)==(0);
             $122 = (($idx$143$i) + 1)|0;
             if (!($126)) {
              $877 = $125;
              break L61;
             }
             $120 = HEAP32[$108>>2]|0;
             $121 = ($122>>>0)<($120>>>0);
             if ($121) {
              $idx$143$i = $122;
             } else {
              break;
             }
            }
           }
           $127 = ((($38)) + 1184|0);
           $128 = HEAP32[$127>>2]|0;
           $129 = ((($38)) + 1492|0);
           $130 = HEAP32[$129>>2]|0;
           $131 = (($130) + ($46))|0;
           $132 = (_FT_Stream_Seek($stream,$131)|0);
           HEAP32[$error$i$i>>2] = $132;
           $133 = ($132|0)==(0);
           L71: do {
            if ($133) {
             $134 = (_FT_Stream_ReadChar($stream,$error$i$i)|0);
             $135 = HEAP32[$error$i$i>>2]|0;
             $136 = ($135|0)==(0);
             if (!($136)) {
              $152 = $135;
              break;
             }
             $137 = ((($38)) + 2924|0);
             HEAP8[$137>>0] = $134;
             $138 = ((($38)) + 2944|0);
             HEAP32[$138>>2] = 0;
             $139 = $134&255;
             L74: do {
              switch ($139|0) {
              case 0:  {
               $140 = ((($38)) + 2936|0);
               HEAP32[$140>>2] = $128;
               $150 = $128;
               break;
              }
              case 3:  {
               $141 = (_FT_Stream_ReadUShort($stream,$error$i$i)|0);
               $142 = $141&65535;
               $143 = HEAP32[$error$i$i>>2]|0;
               $144 = ($143|0)==(0);
               if (!($144)) {
                $152 = $143;
                break L71;
               }
               $145 = ($141<<16>>16)==(0);
               if ($145) {
                HEAP32[$error$i$i>>2] = 3;
                $152 = 3;
                break L71;
               } else {
                $146 = ($142*3)|0;
                $147 = (($146) + 2)|0;
                $148 = ((($38)) + 2936|0);
                HEAP32[$148>>2] = $147;
                $150 = $147;
                break L74;
               }
               break;
              }
              default: {
               HEAP32[$error$i$i>>2] = 3;
               $152 = 3;
               break L71;
              }
              }
             } while(0);
             $149 = ((($38)) + 2932|0);
             $151 = (_FT_Stream_ExtractFrame($stream,$150,$149)|0);
             HEAP32[$error$i$i>>2] = $151;
             $152 = $151;
            } else {
             $152 = $132;
            }
           } while(0);
           HEAP32[$error$i>>2] = $152;
           $877 = $152;
          }
         } while(0);
         $153 = HEAP32[$fd_index$i>>2]|0;
         $154 = ($153|0)==(0|0);
         if ($154) {
          $163 = $877;
         } else {
          $155 = ((($153)) + 28|0);
          $156 = HEAP32[$155>>2]|0;
          $157 = ((($fd_index$i)) + 28|0);
          $158 = HEAP32[$157>>2]|0;
          $159 = ($158|0)==(0|0);
          if (!($159)) {
           _FT_Stream_ReleaseFrame($153,$157);
          }
          $160 = ((($fd_index$i)) + 24|0);
          $161 = HEAP32[$160>>2]|0;
          _ft_mem_free($156,$161);
          ;HEAP32[$fd_index$i>>2]=0|0;HEAP32[$fd_index$i+4>>2]=0|0;HEAP32[$fd_index$i+8>>2]=0|0;HEAP32[$fd_index$i+12>>2]=0|0;HEAP32[$fd_index$i+16>>2]=0|0;HEAP32[$fd_index$i+20>>2]=0|0;HEAP32[$fd_index$i+24>>2]=0|0;HEAP32[$fd_index$i+28>>2]=0|0;
          $$pre$i = HEAP32[$error$i>>2]|0;
          $163 = $$pre$i;
         }
         $162 = ($163|0)==(0);
         if (!($162)) {
          break;
         }
         break L55;
        }
       } while(0);
       break L32;
      }
     } while(0);
     $165 = HEAP32[$90>>2]|0;
     $166 = ($165|0)==(0);
     if ($166) {
      HEAP32[$error$i>>2] = 3;
      break;
     }
     $167 = ((($38)) + 1184|0);
     $168 = HEAP32[$167>>2]|0;
     $169 = ((($38)) + 12|0);
     HEAP32[$169>>2] = $168;
     $170 = ((($38)) + 1308|0);
     $171 = (_cff_index_get_pointers($71,$170,0)|0);
     HEAP32[$error$i>>2] = $171;
     $172 = ($171|0)==(0);
     if (!($172)) {
      break;
     }
     $173 = HEAP32[$169>>2]|0;
     $174 = ($173|0)==(0);
     L98: do {
      if (!($174)) {
       $175 = HEAP32[$98>>2]|0;
       $176 = ($175|0)==(65535);
       $177 = ((($38)) + 1152|0);
       $178 = ((($38)) + 1428|0);
       $179 = HEAP32[$178>>2]|0;
       $180 = HEAP32[$42>>2]|0;
       HEAP32[$error$i4$i>>2] = 0;
       $181 = ($179>>>0)>(2);
       L100: do {
        if ($181) {
         $182 = (($179) + ($46))|0;
         $183 = ((($38)) + 1156|0);
         HEAP32[$183>>2] = $182;
         $184 = (_FT_Stream_Seek($stream,$182)|0);
         HEAP32[$error$i4$i>>2] = $184;
         $185 = ($184|0)==(0);
         if (!($185)) {
          $$pre$phi$i$iZ2D = $183;
          label = 114;
          break;
         }
         $186 = (_FT_Stream_ReadChar($stream,$error$i4$i)|0);
         $187 = $186&255;
         HEAP32[$177>>2] = $187;
         $188 = HEAP32[$error$i4$i>>2]|0;
         $189 = ($188|0)==(0);
         if (!($189)) {
          $$pre$phi$i$iZ2D = $183;
          label = 114;
          break;
         }
         $190 = (_ft_mem_realloc($180,2,0,$173,0,$error$i4$i)|0);
         $191 = ((($38)) + 1160|0);
         HEAP32[$191>>2] = $190;
         $192 = HEAP32[$error$i4$i>>2]|0;
         $193 = ($192|0)==(0);
         if (!($193)) {
          $$pre$phi$i$iZ2D = $183;
          label = 114;
          break;
         }
         HEAP16[$190>>1] = 0;
         $194 = HEAP32[$177>>2]|0;
         switch ($194|0) {
         case 0:  {
          $196 = $173 << 1;
          $197 = (($196) + -2)|0;
          $198 = (_FT_Stream_EnterFrame($stream,$197)|0);
          HEAP32[$error$i4$i>>2] = $198;
          $199 = ($198|0)==(0);
          if (!($199)) {
           $$pre$phi$i$iZ2D = $183;
           label = 114;
           break L100;
          }
          $200 = ($173>>>0)>(1);
          if ($200) {
           $j$07$i$i = 1;
           while(1) {
            $201 = (_FT_Stream_GetUShort($stream)|0);
            $202 = HEAP32[$191>>2]|0;
            $203 = (($202) + ($j$07$i$i<<1)|0);
            HEAP16[$203>>1] = $201;
            $204 = (($j$07$i$i) + 1)|0;
            $exitcond$i$i = ($204|0)==($173|0);
            if ($exitcond$i$i) {
             break;
            } else {
             $j$07$i$i = $204;
            }
           }
          }
          _FT_Stream_ExitFrame($stream);
          label = 100;
          break L100;
          break;
         }
         case 2: case 1:  {
          $195 = ($173>>>0)>(1);
          if ($195) {
           $j$114$i$i = 1;
          } else {
           label = 100;
           break L100;
          }
          while(1) {
           $206 = (_FT_Stream_ReadUShort($stream,$error$i4$i)|0);
           $207 = HEAP32[$error$i4$i>>2]|0;
           $208 = ($207|0)==(0);
           if (!($208)) {
            $$pre$phi$i$iZ2D = $183;
            label = 114;
            break L100;
           }
           $209 = HEAP32[$177>>2]|0;
           $210 = ($209|0)==(2);
           if ($210) {
            $211 = (_FT_Stream_ReadUShort($stream,$error$i4$i)|0);
            $212 = $211&65535;
            $213 = HEAP32[$error$i4$i>>2]|0;
            $214 = ($213|0)==(0);
            if ($214) {
             $nleft$0$i$i = $212;
            } else {
             $$pre$phi$i$iZ2D = $183;
             label = 114;
             break L100;
            }
           } else {
            $215 = (_FT_Stream_ReadChar($stream,$error$i4$i)|0);
            $216 = $215&255;
            $217 = HEAP32[$error$i4$i>>2]|0;
            $218 = ($217|0)==(0);
            if ($218) {
             $nleft$0$i$i = $216;
            } else {
             $$pre$phi$i$iZ2D = $183;
             label = 114;
             break L100;
            }
           }
           $219 = $206&65535;
           $220 = $nleft$0$i$i ^ 65535;
           $221 = ($219>>>0)>($220>>>0);
           $222 = $219 ^ 65535;
           $$nleft$0$i$i = $221 ? $222 : $nleft$0$i$i;
           $223 = ($j$114$i$i>>>0)<($173>>>0);
           if ($223) {
            $224 = HEAP32[$191>>2]|0;
            $glyph_sid$09$i$i = $206;$i$011$i$i = 0;$j$210$i$i = $j$114$i$i;
            while(1) {
             $225 = (($224) + ($j$210$i$i<<1)|0);
             HEAP16[$225>>1] = $glyph_sid$09$i$i;
             $226 = (($i$011$i$i) + 1)|0;
             $227 = (($j$210$i$i) + 1)|0;
             $228 = (($glyph_sid$09$i$i) + 1)<<16>>16;
             $229 = ($227>>>0)>=($173>>>0);
             $230 = ($226>>>0)>($$nleft$0$i$i>>>0);
             $or$cond$i$i = $230 | $229;
             if ($or$cond$i$i) {
              $j$2$lcssa$i$i = $227;
              break;
             } else {
              $glyph_sid$09$i$i = $228;$i$011$i$i = $226;$j$210$i$i = $227;
             }
            }
           } else {
            $j$2$lcssa$i$i = $j$114$i$i;
           }
           $205 = ($j$2$lcssa$i$i>>>0)<($173>>>0);
           if ($205) {
            $j$114$i$i = $j$2$lcssa$i$i;
           } else {
            label = 100;
            break;
           }
          }
          break;
         }
         default: {
          HEAP32[$error$i4$i>>2] = 3;
          $$pre$phi$i$iZ2D = $183;
          label = 114;
          break L100;
         }
         }
        } else {
         $231 = ((($38)) + 1156|0);
         HEAP32[$231>>2] = $179;
         switch ($179|0) {
         case 0:  {
          $232 = ($173>>>0)>(229);
          if ($232) {
           HEAP32[$error$i4$i>>2] = 3;
           $$pre$phi$i$iZ2D = $231;
           label = 114;
           break L100;
          }
          $233 = (_ft_mem_realloc($180,2,0,$173,0,$error$i4$i)|0);
          $234 = ((($38)) + 1160|0);
          HEAP32[$234>>2] = $233;
          $235 = HEAP32[$error$i4$i>>2]|0;
          $236 = ($235|0)==(0);
          if (!($236)) {
           $$pre$phi$i$iZ2D = $231;
           label = 114;
           break L100;
          }
          $237 = $173 << 1;
          _memcpy(($233|0),(169860|0),($237|0))|0;
          label = 100;
          break L100;
          break;
         }
         case 1:  {
          $238 = ($173>>>0)>(166);
          if ($238) {
           HEAP32[$error$i4$i>>2] = 3;
           $$pre$phi$i$iZ2D = $231;
           label = 114;
           break L100;
          }
          $239 = (_ft_mem_realloc($180,2,0,$173,0,$error$i4$i)|0);
          $240 = ((($38)) + 1160|0);
          HEAP32[$240>>2] = $239;
          $241 = HEAP32[$error$i4$i>>2]|0;
          $242 = ($241|0)==(0);
          if (!($242)) {
           $$pre$phi$i$iZ2D = $231;
           label = 114;
           break L100;
          }
          $243 = $173 << 1;
          _memcpy(($239|0),(170318|0),($243|0))|0;
          label = 100;
          break L100;
          break;
         }
         case 2:  {
          $244 = ($173>>>0)>(87);
          if ($244) {
           HEAP32[$error$i4$i>>2] = 3;
           $$pre$phi$i$iZ2D = $231;
           label = 114;
           break L100;
          }
          $245 = (_ft_mem_realloc($180,2,0,$173,0,$error$i4$i)|0);
          $246 = ((($38)) + 1160|0);
          HEAP32[$246>>2] = $245;
          $247 = HEAP32[$error$i4$i>>2]|0;
          $248 = ($247|0)==(0);
          if (!($248)) {
           $$pre$phi$i$iZ2D = $231;
           label = 114;
           break L100;
          }
          $249 = $173 << 1;
          _memcpy(($245|0),(170650|0),($249|0))|0;
          label = 100;
          break L100;
          break;
         }
         default: {
          HEAP32[$error$i4$i>>2] = 3;
          $$pre$phi$i$iZ2D = $231;
          label = 114;
          break L100;
         }
         }
        }
       } while(0);
       do {
        if ((label|0) == 100) {
         $250 = $81 | $176;
         if ($250) {
          $$pr$pre$i$i = HEAP32[$error$i4$i>>2]|0;
          $277 = $$pr$pre$i$i;
         } else {
          HEAP32[$error$i$i$i>>2] = 0;
          $251 = ((($38)) + 1168|0);
          $252 = HEAP32[$251>>2]|0;
          $253 = ($252|0)==(0);
          do {
           if ($253) {
            $254 = ((($38)) + 1160|0);
            $255 = HEAP32[$254>>2]|0;
            $i$04$i$i$i = 0;$max_cid$05$i$i$i = 0;
            while(1) {
             $256 = (($255) + ($i$04$i$i$i<<1)|0);
             $257 = HEAP16[$256>>1]|0;
             $258 = ($257&65535)>($max_cid$05$i$i$i&65535);
             $$max_cid$0$i$i$i = $258 ? $257 : $max_cid$05$i$i$i;
             $259 = (($i$04$i$i$i) + 1)|0;
             $exitcond$i$i$i = ($259|0)==($173|0);
             if ($exitcond$i$i$i) {
              $$max_cid$0$i$i$i$lcssa = $$max_cid$0$i$i$i;
              break;
             } else {
              $i$04$i$i$i = $259;$max_cid$05$i$i$i = $$max_cid$0$i$i$i;
             }
            }
            $260 = $$max_cid$0$i$i$i$lcssa&65535;
            $261 = (($260) + 1)|0;
            $262 = (_ft_mem_realloc($180,2,0,$261,0,$error$i$i$i)|0);
            $263 = ((($38)) + 1164|0);
            HEAP32[$263>>2] = $262;
            $264 = HEAP32[$error$i$i$i>>2]|0;
            $265 = ($264|0)==(0);
            if (!($265)) {
             $275 = $264;
             break;
            }
            $266 = ($173|0)>(0);
            if ($266) {
             $267 = HEAP32[$254>>2]|0;
             $j$03$in$i$i$i = $173;
             while(1) {
              $j$03$i$i$i = (($j$03$in$i$i$i) + -1)|0;
              $268 = $j$03$i$i$i&65535;
              $269 = (($267) + ($j$03$i$i$i<<1)|0);
              $270 = HEAP16[$269>>1]|0;
              $271 = $270&65535;
              $272 = (($262) + ($271<<1)|0);
              HEAP16[$272>>1] = $268;
              $273 = ($j$03$in$i$i$i|0)>(1);
              if ($273) {
               $j$03$in$i$i$i = $j$03$i$i$i;
              } else {
               break;
              }
             }
            }
            HEAP32[$251>>2] = $260;
            $274 = ((($38)) + 1172|0);
            HEAP32[$274>>2] = $173;
            $275 = 0;
           } else {
            $275 = 0;
           }
          } while(0);
          HEAP32[$error$i4$i>>2] = $275;
          $277 = $275;
         }
         $276 = ($277|0)==(0);
         if ($276) {
          HEAP32[$error$i>>2] = 0;
          break;
         } else {
          $$pre55$i = ((($38)) + 1156|0);
          $$pre$phi$i$iZ2D = $$pre55$i;
          label = 114;
          break;
         }
        }
       } while(0);
       if ((label|0) == 114) {
        $278 = ((($38)) + 1160|0);
        $279 = HEAP32[$278>>2]|0;
        _ft_mem_free($180,$279);
        HEAP32[$278>>2] = 0;
        $280 = ((($38)) + 1164|0);
        $281 = HEAP32[$280>>2]|0;
        _ft_mem_free($180,$281);
        HEAP32[$280>>2] = 0;
        HEAP32[$177>>2] = 0;
        HEAP32[$$pre$phi$i$iZ2D>>2] = 0;
        HEAP32[$278>>2] = 0;
        $$pre$i$i = HEAP32[$error$i4$i>>2]|0;
        HEAP32[$error$i>>2] = $$pre$i$i;
        $282 = ($$pre$i$i|0)==(0);
        if (!($282)) {
         break L32;
        }
       }
       $283 = HEAP32[$98>>2]|0;
       $284 = ($283|0)==(65535);
       if (!($284)) {
        break;
       }
       $285 = HEAP32[$169>>2]|0;
       $286 = ((($38)) + 1432|0);
       $287 = HEAP32[$286>>2]|0;
       HEAP32[$error$i6$i>>2] = 0;
       $288 = ((($38)) + 1160|0);
       $289 = HEAP32[$288>>2]|0;
       $290 = ($289|0)==(0|0);
       L169: do {
        if ($290) {
         HEAP32[$error$i6$i>>2] = 3;
         $$ph$i = 3;
        } else {
         $j$033$i$i = 0;
         while(1) {
          $291 = (((($38)) + 128|0) + ($j$033$i$i<<1)|0);
          HEAP16[$291>>1] = 0;
          $292 = (((($38)) + 640|0) + ($j$033$i$i<<1)|0);
          HEAP16[$292>>1] = 0;
          $293 = (($j$033$i$i) + 1)|0;
          $exitcond40$i$i = ($293|0)==(256);
          if ($exitcond40$i$i) {
           break;
          } else {
           $j$033$i$i = $293;
          }
         }
         $294 = ($287>>>0)>(1);
         L174: do {
          if ($294) {
           $295 = (($287) + ($46))|0;
           $296 = ((($38)) + 120|0);
           HEAP32[$296>>2] = $295;
           $297 = (_FT_Stream_Seek($stream,$295)|0);
           HEAP32[$error$i6$i>>2] = $297;
           $298 = ($297|0)==(0);
           if (!($298)) {
            $$ph$i = $297;
            break L169;
           }
           $299 = (_FT_Stream_ReadChar($stream,$error$i6$i)|0);
           $300 = $299&255;
           $301 = ((($38)) + 116|0);
           HEAP32[$301>>2] = $300;
           $302 = HEAP32[$error$i6$i>>2]|0;
           $303 = ($302|0)==(0);
           if (!($303)) {
            $$ph$i = $302;
            break L169;
           }
           $304 = (_FT_Stream_ReadChar($stream,$error$i6$i)|0);
           $305 = $304&255;
           $306 = HEAP32[$error$i6$i>>2]|0;
           $307 = ($306|0)==(0);
           if (!($307)) {
            $$ph$i = $306;
            break L169;
           }
           $308 = HEAP32[$301>>2]|0;
           $309 = $308 & 127;
           L179: do {
            switch ($309|0) {
            case 0:  {
             $310 = (($305) + 1)|0;
             $311 = ((($38)) + 124|0);
             HEAP32[$311>>2] = $310;
             $312 = (_FT_Stream_EnterFrame($stream,$305)|0);
             HEAP32[$error$i6$i>>2] = $312;
             $313 = ($312|0)==(0);
             if (!($313)) {
              $$ph$i = $312;
              break L169;
             }
             $314 = ($304<<24>>24)==(0);
             if (!($314)) {
              $315 = ((($stream)) + 32|0);
              $316 = HEAP32[$315>>2]|0;
              $j$120$i$i = 1;$p$021$i$i = $316;
              while(1) {
               $317 = ((($p$021$i$i)) + 1|0);
               $318 = HEAP8[$p$021$i$i>>0]|0;
               $319 = $318&255;
               $320 = ($j$120$i$i>>>0)<($285>>>0);
               if ($320) {
                $321 = $j$120$i$i&65535;
                $322 = (((($38)) + 640|0) + ($319<<1)|0);
                HEAP16[$322>>1] = $321;
                $323 = HEAP32[$288>>2]|0;
                $324 = (($323) + ($j$120$i$i<<1)|0);
                $325 = HEAP16[$324>>1]|0;
                $326 = (((($38)) + 128|0) + ($319<<1)|0);
                HEAP16[$326>>1] = $325;
               }
               $327 = (($j$120$i$i) + 1)|0;
               $328 = ($327>>>0)>($305>>>0);
               if ($328) {
                break;
               } else {
                $j$120$i$i = $327;$p$021$i$i = $317;
               }
              }
             }
             _FT_Stream_ExitFrame($stream);
             break;
            }
            case 1:  {
             $329 = ((($38)) + 124|0);
             HEAP32[$329>>2] = 0;
             $330 = ($304<<24>>24)==(0);
             if ($330) {
              break L179;
             } else {
              $i$027$i$i = 1;$j$226$i$i = 0;
             }
             while(1) {
              $331 = (_FT_Stream_ReadChar($stream,$error$i6$i)|0);
              $332 = $331&255;
              $333 = HEAP32[$error$i6$i>>2]|0;
              $334 = ($333|0)==(0);
              if (!($334)) {
               $$ph$i = $333;
               break L169;
              }
              $335 = (_FT_Stream_ReadChar($stream,$error$i6$i)|0);
              $336 = HEAP32[$error$i6$i>>2]|0;
              $337 = ($336|0)==(0);
              if (!($337)) {
               $$ph$i = $336;
               break L169;
              }
              $338 = $335&255;
              $339 = (($338) + 1)|0;
              $340 = HEAP32[$329>>2]|0;
              $341 = ($339>>>0)>($340>>>0);
              if ($341) {
               HEAP32[$329>>2] = $339;
              }
              $342 = (($339) + ($i$027$i$i))|0;
              $343 = ($i$027$i$i>>>0)<($342>>>0);
              if ($343) {
               $344 = (($i$027$i$i) + 1)|0;
               $345 = (($344) + ($338))|0;
               $glyph_code$023$i$i = $332;$k$024$i$i = $i$027$i$i;
               while(1) {
                $346 = ($k$024$i$i>>>0)<($285>>>0);
                $347 = ($glyph_code$023$i$i>>>0)<(256);
                $or$cond$i7$i = $346 & $347;
                if ($or$cond$i7$i) {
                 $348 = $k$024$i$i&65535;
                 $349 = (((($38)) + 640|0) + ($glyph_code$023$i$i<<1)|0);
                 HEAP16[$349>>1] = $348;
                 $350 = HEAP32[$288>>2]|0;
                 $351 = (($350) + ($k$024$i$i<<1)|0);
                 $352 = HEAP16[$351>>1]|0;
                 $353 = (((($38)) + 128|0) + ($glyph_code$023$i$i<<1)|0);
                 HEAP16[$353>>1] = $352;
                }
                $354 = (($k$024$i$i) + 1)|0;
                $355 = (($glyph_code$023$i$i) + 1)|0;
                $exitcond$i8$i = ($354|0)==($345|0);
                if ($exitcond$i8$i) {
                 break;
                } else {
                 $glyph_code$023$i$i = $355;$k$024$i$i = $354;
                }
               }
              }
              $356 = (($j$226$i$i) + 1)|0;
              $357 = ($356>>>0)<($305>>>0);
              if ($357) {
               $i$027$i$i = $342;$j$226$i$i = $356;
              } else {
               break;
              }
             }
             $$pre$i10$i = HEAP32[$329>>2]|0;
             $phitmp$i$i = ($$pre$i10$i>>>0)>(256);
             if (!($phitmp$i$i)) {
              break L179;
             }
             HEAP32[$329>>2] = 256;
             break;
            }
            default: {
             HEAP32[$error$i6$i>>2] = 3;
             $$ph$i = 3;
             break L169;
            }
            }
           } while(0);
           $358 = HEAP32[$301>>2]|0;
           $359 = $358 & 128;
           $360 = ($359|0)==(0);
           do {
            if ($360) {
             $$pre43$i$i = HEAP32[$error$i6$i>>2]|0;
             $436 = $$pre43$i$i;
            } else {
             $361 = (_FT_Stream_ReadChar($stream,$error$i6$i)|0);
             $362 = $361&255;
             $363 = HEAP32[$error$i6$i>>2]|0;
             $364 = ($363|0)==(0);
             $365 = ($361<<24>>24)!=(0);
             $or$cond34$i$i = $365 & $364;
             if (!($or$cond34$i$i)) {
              $436 = $363;
              break;
             }
             $366 = ($285|0)==(0);
             if ($366) {
              $j$316$i$i = 0;
              while(1) {
               $385 = (_FT_Stream_ReadChar($stream,$error$i6$i)|0);
               $386 = $385&255;
               $387 = HEAP32[$error$i6$i>>2]|0;
               $388 = ($387|0)==(0);
               if (!($388)) {
                $$ph$i = $387;
                break L169;
               }
               $389 = (_FT_Stream_ReadUShort($stream,$error$i6$i)|0);
               $390 = HEAP32[$error$i6$i>>2]|0;
               $391 = ($390|0)==(0);
               if (!($391)) {
                $$ph$i = $390;
                break L169;
               }
               $392 = (((($38)) + 128|0) + ($386<<1)|0);
               HEAP16[$392>>1] = $389;
               $393 = (($j$316$i$i) + 1)|0;
               $394 = ($393>>>0)<($362>>>0);
               if ($394) {
                $j$316$i$i = $393;
               } else {
                break L174;
               }
              }
             } else {
              $j$316$us$i$i = 0;
             }
             while(1) {
              $367 = (_FT_Stream_ReadChar($stream,$error$i6$i)|0);
              $368 = $367&255;
              $369 = HEAP32[$error$i6$i>>2]|0;
              $370 = ($369|0)==(0);
              if (!($370)) {
               $$ph$i = $369;
               break L169;
              }
              $371 = (_FT_Stream_ReadUShort($stream,$error$i6$i)|0);
              $372 = HEAP32[$error$i6$i>>2]|0;
              $373 = ($372|0)==(0);
              if (!($373)) {
               $$ph$i = $372;
               break L169;
              }
              $374 = (((($38)) + 128|0) + ($368<<1)|0);
              HEAP16[$374>>1] = $371;
              $375 = HEAP32[$288>>2]|0;
              $gindex$015$us$i$i = 0;
              while(1) {
               $376 = (($375) + ($gindex$015$us$i$i<<1)|0);
               $377 = HEAP16[$376>>1]|0;
               $378 = ($377<<16>>16)==($371<<16>>16);
               if ($378) {
                $gindex$015$us$i$i$lcssa = $gindex$015$us$i$i;
                label = 154;
                break;
               }
               $379 = (($gindex$015$us$i$i) + 1)|0;
               $380 = ($379>>>0)<($285>>>0);
               if ($380) {
                $gindex$015$us$i$i = $379;
               } else {
                break;
               }
              }
              if ((label|0) == 154) {
               label = 0;
               $381 = $gindex$015$us$i$i$lcssa&65535;
               $382 = (((($38)) + 640|0) + ($368<<1)|0);
               HEAP16[$382>>1] = $381;
              }
              $383 = (($j$316$us$i$i) + 1)|0;
              $384 = ($383>>>0)<($362>>>0);
              if ($384) {
               $j$316$us$i$i = $383;
              } else {
               break L174;
              }
             }
            }
           } while(0);
           HEAP32[$error$i>>2] = $436;
           $437 = ($436|0)==(0);
           if ($437) {
            break L98;
           } else {
            break L32;
           }
          } else {
           switch ($287|0) {
           case 0:  {
            $395 = ((($38)) + 128|0);
            _memcpy(($395|0),(170824|0),512)|0;
            break;
           }
           case 1:  {
            $396 = ((($38)) + 128|0);
            _memcpy(($396|0),(171336|0),512)|0;
            break;
           }
           default: {
            HEAP32[$error$i6$i>>2] = 3;
            $$ph$i = 3;
            break L169;
           }
           }
           $397 = ((($38)) + 124|0);
           HEAP32[$397>>2] = 0;
           $398 = HEAP32[$42>>2]|0;
           HEAP32[$error$i$i5$i>>2] = 0;
           $399 = ((($38)) + 1168|0);
           $400 = HEAP32[$399>>2]|0;
           $401 = ($400|0)==(0);
           if ($401) {
            $402 = ($285|0)==(0);
            if ($402) {
             $max_cid$0$lcssa$i$i18$i = 0;
            } else {
             $403 = HEAP32[$288>>2]|0;
             $i$04$i$i15$i = 0;$max_cid$05$i$i14$i = 0;
             while(1) {
              $404 = (($403) + ($i$04$i$i15$i<<1)|0);
              $405 = HEAP16[$404>>1]|0;
              $406 = ($405&65535)>($max_cid$05$i$i14$i&65535);
              $$max_cid$0$i$i16$i = $406 ? $405 : $max_cid$05$i$i14$i;
              $407 = (($i$04$i$i15$i) + 1)|0;
              $exitcond$i$i17$i = ($407|0)==($285|0);
              if ($exitcond$i$i17$i) {
               $max_cid$0$lcssa$i$i18$i = $$max_cid$0$i$i16$i;
               break;
              } else {
               $i$04$i$i15$i = $407;$max_cid$05$i$i14$i = $$max_cid$0$i$i16$i;
              }
             }
            }
            $408 = $max_cid$0$lcssa$i$i18$i&65535;
            $409 = (($408) + 1)|0;
            $410 = (_ft_mem_realloc($398,2,0,$409,0,$error$i$i5$i)|0);
            $411 = ((($38)) + 1164|0);
            HEAP32[$411>>2] = $410;
            $412 = HEAP32[$error$i$i5$i>>2]|0;
            $413 = ($412|0)==(0);
            if (!($413)) {
             HEAP32[$error$i6$i>>2] = $412;
             $$ph$i = $412;
             break L169;
            }
            $414 = ($285|0)>(0);
            if ($414) {
             $415 = HEAP32[$288>>2]|0;
             $j$03$in$i$i22$i = $285;
             while(1) {
              $j$03$i$i23$i = (($j$03$in$i$i22$i) + -1)|0;
              $416 = $j$03$i$i23$i&65535;
              $417 = (($415) + ($j$03$i$i23$i<<1)|0);
              $418 = HEAP16[$417>>1]|0;
              $419 = $418&65535;
              $420 = (($410) + ($419<<1)|0);
              HEAP16[$420>>1] = $416;
              $421 = ($j$03$in$i$i22$i|0)>(1);
              if ($421) {
               $j$03$in$i$i22$i = $j$03$i$i23$i;
              } else {
               break;
              }
             }
            }
            HEAP32[$399>>2] = $408;
            $422 = ((($38)) + 1172|0);
            HEAP32[$422>>2] = $285;
            $$pre$phi46$i$i$inZ2D = $411;$428 = $408;
           } else {
            $$pre45$i$i = ((($38)) + 1164|0);
            $$pre$phi46$i$i$inZ2D = $$pre45$i$i;$428 = $400;
           }
           HEAP32[$error$i6$i>>2] = 0;
           $j$431$i$i = 0;
           while(1) {
            $423 = (((($38)) + 128|0) + ($j$431$i$i<<1)|0);
            $424 = HEAP16[$423>>1]|0;
            $425 = ($424<<16>>16)==(0);
            do {
             if ($425) {
              label = 178;
             } else {
              $426 = $424&65535;
              $427 = ($428>>>0)<($426>>>0);
              if ($427) {
               label = 178;
               break;
              }
              $429 = HEAP32[$$pre$phi46$i$i$inZ2D>>2]|0;
              $430 = (($429) + ($426<<1)|0);
              $431 = HEAP16[$430>>1]|0;
              $432 = ($431<<16>>16)==(0);
              if ($432) {
               label = 178;
               break;
              }
              $433 = (((($38)) + 640|0) + ($j$431$i$i<<1)|0);
              HEAP16[$433>>1] = $431;
              $434 = (($j$431$i$i) + 1)|0;
              HEAP32[$397>>2] = $434;
              $$pre$phi$i26$iZ2D = $434;
             }
            } while(0);
            if ((label|0) == 178) {
             label = 0;
             $435 = (((($38)) + 640|0) + ($j$431$i$i<<1)|0);
             HEAP16[$435>>1] = 0;
             HEAP16[$423>>1] = 0;
             $$pre44$i$i = (($j$431$i$i) + 1)|0;
             $$pre$phi$i26$iZ2D = $$pre44$i$i;
            }
            $exitcond39$i$i = ($$pre$phi$i26$iZ2D|0)==(256);
            if ($exitcond39$i$i) {
             break;
            } else {
             $j$431$i$i = $$pre$phi$i26$iZ2D;
            }
           }
          }
         } while(0);
         HEAP32[$error$i>>2] = 0;
         break L98;
        }
       } while(0);
       HEAP32[$error$i>>2] = $$ph$i;
       break L32;
      }
     } while(0);
     $438 = HEAP32[$63>>2]|0;
     $439 = ((($438)) + 28|0);
     $440 = HEAP32[$439>>2]|0;
     $441 = (_cff_index_access_element($63,$subfont_index$0$i,$bytes$i$i,$byte_len$i$i)|0);
     HEAP32[$error$i27$i>>2] = $441;
     $442 = ($441|0)==(0);
     do {
      if ($442) {
       $443 = HEAP32[$byte_len$i$i>>2]|0;
       $444 = (($443) + 1)|0;
       $445 = (_ft_mem_alloc($440,$444,$error$i27$i)|0);
       $446 = HEAP32[$error$i27$i>>2]|0;
       $447 = ($446|0)==(0);
       if ($447) {
        $448 = HEAP32[$bytes$i$i>>2]|0;
        _memcpy(($445|0),($448|0),($443|0))|0;
        $449 = (($445) + ($443)|0);
        HEAP8[$449>>0] = 0;
       }
       $450 = ((($38)) + 48|0);
       $451 = HEAP32[$450>>2]|0;
       $452 = ($451|0)==(0|0);
       if (!($452)) {
        $name$0$i$i = $445;
        break;
       }
       $453 = HEAP32[$63>>2]|0;
       _FT_Stream_ReleaseFrame($453,$bytes$i$i);
       $name$0$i$i = $445;
      } else {
       $name$0$i$i = 0;
      }
     } while(0);
     $454 = ((($38)) + 1304|0);
     HEAP32[$454>>2] = $name$0$i$i;
     break;
    }
   }
   HEAP32[$error$i>>2] = 2;
  }
 } while(0);
 $455 = HEAP32[$string_index$i>>2]|0;
 $456 = ($455|0)==(0|0);
 if (!($456)) {
  $457 = ((($455)) + 28|0);
  $458 = HEAP32[$457>>2]|0;
  $459 = ((($string_index$i)) + 28|0);
  $460 = HEAP32[$459>>2]|0;
  $461 = ($460|0)==(0|0);
  if (!($461)) {
   _FT_Stream_ReleaseFrame($455,$459);
  }
  $462 = ((($string_index$i)) + 24|0);
  $463 = HEAP32[$462>>2]|0;
  _ft_mem_free($458,$463);
  ;HEAP32[$string_index$i>>2]=0|0;HEAP32[$string_index$i+4>>2]=0|0;HEAP32[$string_index$i+8>>2]=0|0;HEAP32[$string_index$i+12>>2]=0|0;HEAP32[$string_index$i+16>>2]=0|0;HEAP32[$string_index$i+20>>2]=0|0;HEAP32[$string_index$i+24>>2]=0|0;HEAP32[$string_index$i+28>>2]=0|0;
 }
 $464 = HEAP32[$error$i>>2]|0;
 HEAP32[$error>>2] = $464;
 $465 = ($464|0)==(0);
 if (!($465)) {
  $$0 = $464;
  STACKTOP = sp;return ($$0|0);
 }
 $466 = ((($38)) + 2952|0);
 HEAP32[$466>>2] = $8;
 $467 = ((($38)) + 2956|0);
 HEAP32[$467>>2] = $7;
 $468 = ((($cffface)) + 4|0);
 HEAP32[$468>>2] = $face_index;
 $469 = ((($38)) + 12|0);
 $470 = HEAP32[$469>>2]|0;
 $471 = ((($cffface)) + 16|0);
 HEAP32[$471>>2] = $470;
 $472 = ((($38)) + 1456|0);
 $473 = HEAP32[$472>>2]|0;
 $474 = ($473|0)!=(65535);
 $475 = ($7|0)!=(0|0);
 $or$cond = $475 | $474;
 if (!($or$cond)) {
  HEAP32[$error>>2] = 11;
  $$0 = 11;
  STACKTOP = sp;return ($$0|0);
 }
 $476 = ((($38)) + 1388|0);
 $477 = HEAP8[$476>>0]|0;
 $478 = ($477<<24>>24)==(0);
 if ($478) {
  $479 = ($pure_cff$1<<24>>24)==(0);
  if ($479) {
   $480 = ((($cffface)) + 68|0);
   $481 = HEAP16[$480>>1]|0;
   $482 = $481&65535;
   $484 = $482;
  } else {
   $484 = 1000;
  }
  $483 = ((($38)) + 1392|0);
  HEAP32[$483>>2] = $484;
  $$pre$phi95Z2D = $483;
 } else {
  $$pre92 = ((($38)) + 1392|0);
  $$pre$phi95Z2D = $$pre92;
 }
 $485 = ((($38)) + 1372|0);
 $486 = ((($38)) + 1396|0);
 $487 = ((($38)) + 1384|0);
 $488 = HEAP32[$487>>2]|0;
 $489 = ($488|0)<(0);
 $490 = (0 - ($488))|0;
 $491 = $489 ? $490 : $488;
 $492 = ($491|0)==(65536);
 if ($492) {
  $$phi$trans$insert = ((($38)) + 1400|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $$pre$phi85Z2D = $$phi$trans$insert;$$pre$phiZ2D = $486;$513 = $$pre;
 } else {
  $493 = HEAP32[$$pre$phi95Z2D>>2]|0;
  $494 = (_FT_DivFix($493,$491)|0);
  HEAP32[$$pre$phi95Z2D>>2] = $494;
  $495 = HEAP32[$485>>2]|0;
  $496 = (_FT_DivFix($495,$491)|0);
  HEAP32[$485>>2] = $496;
  $497 = ((($38)) + 1380|0);
  $498 = HEAP32[$497>>2]|0;
  $499 = (_FT_DivFix($498,$491)|0);
  HEAP32[$497>>2] = $499;
  $500 = ((($38)) + 1376|0);
  $501 = HEAP32[$500>>2]|0;
  $502 = (_FT_DivFix($501,$491)|0);
  HEAP32[$500>>2] = $502;
  $503 = HEAP32[$487>>2]|0;
  $504 = (_FT_DivFix($503,$491)|0);
  HEAP32[$487>>2] = $504;
  $505 = HEAP32[$486>>2]|0;
  $506 = (_FT_DivFix($505,$491)|0);
  HEAP32[$486>>2] = $506;
  $507 = ((($38)) + 1400|0);
  $508 = HEAP32[$507>>2]|0;
  $509 = (_FT_DivFix($508,$491)|0);
  HEAP32[$507>>2] = $509;
  $$pre$phi85Z2D = $507;$$pre$phiZ2D = $486;$513 = $509;
 }
 $510 = HEAP32[$$pre$phiZ2D>>2]|0;
 $511 = $510 >> 16;
 HEAP32[$$pre$phiZ2D>>2] = $511;
 $512 = $513 >> 16;
 HEAP32[$$pre$phi85Z2D>>2] = $512;
 $514 = ((($38)) + 1896|0);
 $515 = HEAP32[$514>>2]|0;
 $516 = ($515|0)==(0);
 if (!($516)) {
  $517 = ((($38)) + 1900|0);
  $i$056 = $515;
  while(1) {
   $518 = (($i$056) + -1)|0;
   $519 = (($517) + ($518<<2)|0);
   $520 = HEAP32[$519>>2]|0;
   $521 = ((($520)) + 64|0);
   $522 = HEAP8[$521>>0]|0;
   $523 = ($522<<24>>24)==(0);
   do {
    if ($523) {
     $537 = ((($520)) + 48|0);
     ;HEAP32[$537>>2]=HEAP32[$485>>2]|0;HEAP32[$537+4>>2]=HEAP32[$485+4>>2]|0;HEAP32[$537+8>>2]=HEAP32[$485+8>>2]|0;HEAP32[$537+12>>2]=HEAP32[$485+12>>2]|0;
     $538 = ((($520)) + 72|0);
     $539 = $486;
     $540 = $539;
     $541 = HEAP32[$540>>2]|0;
     $542 = (($539) + 4)|0;
     $543 = $542;
     $544 = HEAP32[$543>>2]|0;
     $545 = $538;
     $546 = $545;
     HEAP32[$546>>2] = $541;
     $547 = (($545) + 4)|0;
     $548 = $547;
     HEAP32[$548>>2] = $544;
     $549 = HEAP32[$$pre$phi95Z2D>>2]|0;
     $550 = ((($520)) + 68|0);
     HEAP32[$550>>2] = $549;
     $$pre$phi99Z2D = $550;
    } else {
     $524 = HEAP8[$476>>0]|0;
     $525 = ($524<<24>>24)==(0);
     if ($525) {
      $$pre98 = ((($520)) + 68|0);
      $$pre$phi99Z2D = $$pre98;
      break;
     }
     $526 = HEAP32[$$pre$phi95Z2D>>2]|0;
     $527 = ($526>>>0)>(1);
     $528 = ((($520)) + 68|0);
     if ($527) {
      $529 = HEAP32[$528>>2]|0;
      $530 = ($529>>>0)>(1);
      if ($530) {
       $531 = ($526>>>0)<($529>>>0);
       $$ = $531 ? $526 : $529;
       $scaling$0 = $$;
      } else {
       $scaling$0 = 1;
      }
     } else {
      $scaling$0 = 1;
     }
     $532 = ((($520)) + 48|0);
     _FT_Matrix_Multiply_Scaled($485,$532,$scaling$0);
     $533 = ((($520)) + 72|0);
     _FT_Vector_Transform_Scaled($533,$485,$scaling$0);
     $534 = HEAP32[$528>>2]|0;
     $535 = HEAP32[$$pre$phi95Z2D>>2]|0;
     $536 = (_FT_MulDiv($534,$535,$scaling$0)|0);
     HEAP32[$528>>2] = $536;
     $$pre$phi99Z2D = $528;
    }
   } while(0);
   $551 = ((($520)) + 60|0);
   $552 = HEAP32[$551>>2]|0;
   $553 = ($552|0)<(0);
   $554 = (0 - ($552))|0;
   $555 = $553 ? $554 : $552;
   $556 = ($555|0)==(65536);
   if ($556) {
    $$phi$trans$insert79 = ((($520)) + 76|0);
    $$pre80 = HEAP32[$$phi$trans$insert79>>2]|0;
    $$pre86 = ((($520)) + 72|0);
    $$pre$phi87Z2D = $$pre86;$$pre$phi88Z2D = $$phi$trans$insert79;$579 = $$pre80;
   } else {
    $557 = HEAP32[$$pre$phi99Z2D>>2]|0;
    $558 = (_FT_DivFix($557,$555)|0);
    HEAP32[$$pre$phi99Z2D>>2] = $558;
    $559 = ((($520)) + 48|0);
    $560 = HEAP32[$559>>2]|0;
    $561 = (_FT_DivFix($560,$555)|0);
    HEAP32[$559>>2] = $561;
    $562 = ((($520)) + 56|0);
    $563 = HEAP32[$562>>2]|0;
    $564 = (_FT_DivFix($563,$555)|0);
    HEAP32[$562>>2] = $564;
    $565 = ((($520)) + 52|0);
    $566 = HEAP32[$565>>2]|0;
    $567 = (_FT_DivFix($566,$555)|0);
    HEAP32[$565>>2] = $567;
    $568 = HEAP32[$551>>2]|0;
    $569 = (_FT_DivFix($568,$555)|0);
    HEAP32[$551>>2] = $569;
    $570 = ((($520)) + 72|0);
    $571 = HEAP32[$570>>2]|0;
    $572 = (_FT_DivFix($571,$555)|0);
    HEAP32[$570>>2] = $572;
    $573 = ((($520)) + 76|0);
    $574 = HEAP32[$573>>2]|0;
    $575 = (_FT_DivFix($574,$555)|0);
    HEAP32[$573>>2] = $575;
    $$pre$phi87Z2D = $570;$$pre$phi88Z2D = $573;$579 = $575;
   }
   $576 = HEAP32[$$pre$phi87Z2D>>2]|0;
   $577 = $576 >> 16;
   HEAP32[$$pre$phi87Z2D>>2] = $577;
   $578 = $579 >> 16;
   HEAP32[$$pre$phi88Z2D>>2] = $578;
   $580 = ($518|0)==(0);
   if ($580) {
    break;
   } else {
    $i$056 = $518;
   }
  }
 }
 $581 = ($pure_cff$1<<24>>24)==(0);
 if (!($581)) {
  $582 = ((($38)) + 8|0);
  $583 = HEAP32[$582>>2]|0;
  HEAP32[$cffface>>2] = $583;
  $584 = HEAP32[$472>>2]|0;
  $585 = ($584|0)==(65535);
  if ($585) {
   $589 = ((($38)) + 1184|0);
   $590 = HEAP32[$589>>2]|0;
   $storemerge = $590;
  } else {
   $586 = ((($38)) + 1168|0);
   $587 = HEAP32[$586>>2]|0;
   $588 = (($587) + 1)|0;
   $storemerge = $588;
  }
  HEAP32[$471>>2] = $storemerge;
  $591 = ((($38)) + 1408|0);
  $592 = HEAP32[$591>>2]|0;
  $593 = $592 >> 16;
  $594 = ((($cffface)) + 52|0);
  HEAP32[$594>>2] = $593;
  $595 = ((($38)) + 1412|0);
  $596 = HEAP32[$595>>2]|0;
  $597 = $596 >> 16;
  $598 = ((($cffface)) + 56|0);
  HEAP32[$598>>2] = $597;
  $599 = ((($38)) + 1416|0);
  $600 = HEAP32[$599>>2]|0;
  $601 = (($600) + 65535)|0;
  $602 = $601 >> 16;
  $603 = ((($cffface)) + 60|0);
  HEAP32[$603>>2] = $602;
  $604 = ((($38)) + 1420|0);
  $605 = HEAP32[$604>>2]|0;
  $606 = (($605) + 65535)|0;
  $607 = $606 >> 16;
  $608 = ((($cffface)) + 64|0);
  HEAP32[$608>>2] = $607;
  $609 = HEAP32[$$pre$phi95Z2D>>2]|0;
  $610 = $609&65535;
  $611 = ((($cffface)) + 68|0);
  HEAP16[$611>>1] = $610;
  $612 = $607&65535;
  $613 = ((($cffface)) + 70|0);
  HEAP16[$613>>1] = $612;
  $614 = $597&65535;
  $615 = ((($cffface)) + 72|0);
  HEAP16[$615>>1] = $614;
  $616 = $609 & 65535;
  $617 = ($616*12)|0;
  $618 = (($617>>>0) / 10)&-1;
  $619 = ((($cffface)) + 74|0);
  $sext = $618 << 16;
  $620 = $sext >> 16;
  $621 = (($607) - ($597))|0;
  $622 = ($620|0)<($621|0);
  $$121 = $622 ? $621 : $618;
  $storemerge102 = $$121&65535;
  HEAP16[$619>>1] = $storemerge102;
  $623 = ((($38)) + 1356|0);
  $624 = HEAP32[$623>>2]|0;
  $625 = $624 >>> 16;
  $626 = $625&65535;
  $627 = ((($cffface)) + 80|0);
  HEAP16[$627>>1] = $626;
  $628 = ((($38)) + 1360|0);
  $629 = HEAP32[$628>>2]|0;
  $630 = $629 >>> 16;
  $631 = $630&65535;
  $632 = ((($cffface)) + 82|0);
  HEAP16[$632>>1] = $631;
  $633 = ((($38)) + 20|0);
  $634 = HEAP32[$633>>2]|0;
  $635 = ((($634)) + 28|0);
  $636 = HEAP32[$635>>2]|0;
  $637 = (_cff_index_access_element($633,$face_index,$bytes$i,$byte_len$i)|0);
  HEAP32[$error$i11>>2] = $637;
  $638 = ($637|0)==(0);
  if ($638) {
   $639 = HEAP32[$byte_len$i>>2]|0;
   $640 = (($639) + 1)|0;
   $641 = (_ft_mem_alloc($636,$640,$error$i11)|0);
   $642 = HEAP32[$error$i11>>2]|0;
   $643 = ($642|0)==(0);
   if ($643) {
    $644 = HEAP32[$bytes$i>>2]|0;
    _memcpy(($641|0),($644|0),($639|0))|0;
    $645 = (($641) + ($639)|0);
    HEAP8[$645>>0] = 0;
   }
   $646 = ((($38)) + 48|0);
   $647 = HEAP32[$646>>2]|0;
   $648 = ($647|0)==(0|0);
   if ($648) {
    $649 = HEAP32[$633>>2]|0;
    _FT_Stream_ReleaseFrame($649,$bytes$i);
    $name$0$i = $641;
   } else {
    $name$0$i = $641;
   }
  } else {
   $name$0$i = 0;
  }
  $650 = ((($cffface)) + 20|0);
  HEAP32[$650>>2] = $name$0$i;
  $651 = ($name$0$i|0)==(0|0);
  L331: do {
   if ($651) {
    $757 = ((($38)) + 1496|0);
    $758 = HEAP32[$757>>2]|0;
    $759 = ($758|0)==(65535);
    if ($759) {
     label = 276;
    } else {
     $760 = ($758>>>0)>(390);
     if ($760) {
      $761 = (($758) + -391)|0;
      $762 = ((($38)) + 1312|0);
      $763 = HEAP32[$762>>2]|0;
      $764 = ($763>>>0)>($761>>>0);
      if (!($764)) {
       label = 276;
       break;
      }
      $765 = ((($38)) + 1316|0);
      $766 = HEAP32[$765>>2]|0;
      $767 = (($766) + ($761<<2)|0);
      $768 = HEAP32[$767>>2]|0;
      $$0$i20 = $768;
     } else {
      $769 = HEAP32[$467>>2]|0;
      $770 = ($769|0)==(0|0);
      if ($770) {
       label = 276;
       break;
      }
      $771 = ((($769)) + 20|0);
      $772 = HEAP32[$771>>2]|0;
      $773 = (FUNCTION_TABLE_ii[$772 & 8191]($758)|0);
      $$0$i20 = $773;
     }
     $774 = ($$0$i20|0)==(0|0);
     if ($774) {
      label = 276;
     } else {
      $775 = (_ft_mem_strdup($37,$$0$i20,$error$i22)|0);
      HEAP32[$650>>2] = $775;
      label = 276;
     }
    }
   } else {
    $652 = ((($38)) + 1336|0);
    $653 = HEAP32[$652>>2]|0;
    $654 = ($653|0)==(65535);
    do {
     if ($654) {
      $$0$i = 0;$669 = $name$0$i;
     } else {
      $655 = ($653>>>0)>(390);
      if ($655) {
       $656 = (($653) + -391)|0;
       $657 = ((($38)) + 1312|0);
       $658 = HEAP32[$657>>2]|0;
       $659 = ($658>>>0)>($656>>>0);
       if (!($659)) {
        $$0$i = 0;$669 = $name$0$i;
        break;
       }
       $660 = ((($38)) + 1316|0);
       $661 = HEAP32[$660>>2]|0;
       $662 = (($661) + ($656<<2)|0);
       $663 = HEAP32[$662>>2]|0;
       $$0$i = $663;$669 = $name$0$i;
       break;
      } else {
       $664 = HEAP32[$467>>2]|0;
       $665 = ($664|0)==(0|0);
       if ($665) {
        $$0$i = 0;$669 = $name$0$i;
        break;
       }
       $666 = ((($664)) + 20|0);
       $667 = HEAP32[$666>>2]|0;
       $668 = (FUNCTION_TABLE_ii[$667 & 8191]($653)|0);
       $$pre81 = HEAP32[$650>>2]|0;
       $$0$i = $668;$669 = $$pre81;
       break;
      }
     }
    } while(0);
    $670 = (_strlen($669)|0);
    $671 = ((($669)) + 6|0);
    $672 = ((($669)) + 1|0);
    $673 = ((($669)) + 2|0);
    $674 = ((($669)) + 3|0);
    $675 = ((($669)) + 4|0);
    $676 = ((($669)) + 5|0);
    $677 = ($670|0)>(5);
    L350: do {
     if ($677) {
      $678 = (($670) + 1)|0;
      $continue_search$0$ph10$i55 = 1;$indvars$iv$i53 = $678;
      while(1) {
       $679 = HEAP8[$671>>0]|0;
       $680 = ($679<<24>>24)==(43);
       if (!($680)) {
        break L350;
       }
       $681 = HEAP8[$669>>0]|0;
       $$off$us$i = (($681) + -65)<<24>>24;
       $682 = ($$off$us$i&255)<(26);
       $683 = HEAP8[$672>>0]|0;
       $$off$us$1$i = (($683) + -65)<<24>>24;
       $684 = ($$off$us$1$i&255)<(26);
       $685 = $682 & $684;
       $686 = HEAP8[$673>>0]|0;
       $$off$us$2$i = (($686) + -65)<<24>>24;
       $687 = ($$off$us$2$i&255)<(26);
       $688 = $687 & $685;
       $689 = HEAP8[$674>>0]|0;
       $$off$us$3$i = (($689) + -65)<<24>>24;
       $690 = ($$off$us$3$i&255)<(26);
       $691 = $690 & $688;
       $692 = HEAP8[$675>>0]|0;
       $$off$us$4$i = (($692) + -65)<<24>>24;
       $693 = ($$off$us$4$i&255)<(26);
       $694 = $693 & $691;
       $695 = HEAP8[$676>>0]|0;
       $$off$us$5$i = (($695) + -65)<<24>>24;
       $696 = ($$off$us$5$i&255)<(26);
       $697 = $696 & $694;
       $continue_search$2$us$5$i = $697 ? $continue_search$0$ph10$i55 : 0;
       $698 = ($continue_search$2$us$5$i<<24>>24)==(0);
       if ($698) {
        break L350;
       }
       $699 = ($indvars$iv$i53|0)>(7);
       if ($699) {
        $idx$15$i = 7;
        while(1) {
         $700 = (($669) + ($idx$15$i)|0);
         $701 = HEAP8[$700>>0]|0;
         $702 = (($idx$15$i) + -7)|0;
         $703 = (($669) + ($702)|0);
         HEAP8[$703>>0] = $701;
         $704 = (($idx$15$i) + 1)|0;
         $exitcond11$i = ($704|0)==($indvars$iv$i53|0);
         if ($exitcond11$i) {
          break;
         } else {
          $idx$15$i = $704;
         }
        }
       }
       $indvars$iv$next$i = (($indvars$iv$i53) + -7)|0;
       $705 = ($indvars$iv$next$i|0)>(6);
       if ($705) {
        $continue_search$0$ph10$i55 = $continue_search$2$us$5$i;$indvars$iv$i53 = $indvars$iv$next$i;
       } else {
        break;
       }
      }
     }
    } while(0);
    $706 = ((($38)) + 1340|0);
    $707 = HEAP32[$706>>2]|0;
    $switch$split2D = ($707|0)<(65535);
    if ($switch$split2D) {
     switch ($707|0) {
     case 0:  {
      $family$0 = $669;
      break;
     }
     default: {
      label = 243;
     }
     }
    } else {
     switch ($707|0) {
     case 65535:  {
      $$0$i14 = 0;
      label = 248;
      break;
     }
     default: {
      label = 243;
     }
     }
    }
    do {
     if ((label|0) == 243) {
      $708 = ($707>>>0)>(390);
      if ($708) {
       $709 = (($707) + -391)|0;
       $710 = ((($38)) + 1312|0);
       $711 = HEAP32[$710>>2]|0;
       $712 = ($711>>>0)>($709>>>0);
       if (!($712)) {
        $$0$i14 = 0;
        label = 248;
        break;
       }
       $713 = ((($38)) + 1316|0);
       $714 = HEAP32[$713>>2]|0;
       $715 = (($714) + ($709<<2)|0);
       $716 = HEAP32[$715>>2]|0;
       $$0$i14 = $716;
       label = 248;
       break;
      } else {
       $717 = HEAP32[$467>>2]|0;
       $718 = ($717|0)==(0|0);
       if ($718) {
        $$0$i14 = 0;
        label = 248;
        break;
       }
       $719 = ((($717)) + 20|0);
       $720 = HEAP32[$719>>2]|0;
       $721 = (FUNCTION_TABLE_ii[$720 & 8191]($707)|0);
       $$0$i14 = $721;
       label = 248;
       break;
      }
     }
    } while(0);
    if ((label|0) == 248) {
     $722 = ($$0$i14|0)==(0|0);
     $$5 = $722 ? $669 : $$0$i14;
     $family$0 = $$5;
    }
    $723 = ($$0$i|0)!=(0|0);
    $724 = ($family$0|0)!=(0|0);
    $or$cond3 = $723 & $724;
    if (!($or$cond3)) {
     label = 276;
     break;
    }
    $725 = HEAP8[$$0$i>>0]|0;
    $726 = ($725<<24>>24)==(0);
    if ($726) {
     label = 276;
     break;
    } else {
     $728 = $725;$family$1$ph52 = $family$0;$fullp$0$ph51 = $$0$i;
    }
    L375: while(1) {
     $727 = ($728<<24>>24)==(0);
     $family$148 = $family$1$ph52;
     L377: while(1) {
      $729 = HEAP8[$family$148>>0]|0;
      $730 = ($728<<24>>24)==($729<<24>>24);
      if ($730) {
       $family$148$lcssa = $family$148;
       label = 253;
       break;
      }
      switch ($728<<24>>24) {
      case 45: case 32:  {
       $family$1$ph$be = $family$148;
       break L377;
       break;
      }
      default: {
      }
      }
      switch ($729<<24>>24) {
      case 0:  {
       $fullp$0$ph51$lcssa158 = $fullp$0$ph51;
       break L375;
       break;
      }
      case 45: case 32:  {
       break;
      }
      default: {
       label = 276;
       break L331;
      }
      }
      $735 = ((($family$148)) + 1|0);
      if ($727) {
       label = 276;
       break L331;
      } else {
       $family$148 = $735;
      }
     }
     if ((label|0) == 253) {
      label = 0;
      $731 = ((($family$148$lcssa)) + 1|0);
      $family$1$ph$be = $731;
     }
     $732 = ((($fullp$0$ph51)) + 1|0);
     $733 = HEAP8[$732>>0]|0;
     $734 = ($733<<24>>24)==(0);
     if ($734) {
      label = 276;
      break L331;
     } else {
      $728 = $733;$family$1$ph52 = $family$1$ph$be;$fullp$0$ph51 = $732;
     }
    }
    $736 = (_ft_mem_strdup($37,$fullp$0$ph51$lcssa158,$error$i16)|0);
    $737 = HEAP32[$650>>2]|0;
    $738 = (_strlen($737)|0);
    $739 = (_strlen($736)|0);
    $740 = ($738|0)>($739|0);
    L386: do {
     if ($740) {
      $741 = ($739|0)<(1);
      if (!($741)) {
       $idx$010$i = 1;
       while(1) {
        $744 = (($738) - ($idx$010$i))|0;
        $745 = (($737) + ($744)|0);
        $746 = HEAP8[$745>>0]|0;
        $747 = (($739) - ($idx$010$i))|0;
        $748 = (($736) + ($747)|0);
        $749 = HEAP8[$748>>0]|0;
        $750 = ($746<<24>>24)==($749<<24>>24);
        if (!($750)) {
         break L386;
        }
        $742 = (($idx$010$i) + 1)|0;
        $743 = ($739|0)>($idx$010$i|0);
        if ($743) {
         $idx$010$i = $742;
        } else {
         break;
        }
       }
      }
      $751 = (($738) - ($739))|0;
      $752 = ($751|0)>(1);
      if ($752) {
       $idx$19$in$i = $751;
      } else {
       label = 274;
       break;
      }
      L393: while(1) {
       $idx$19$i = (($idx$19$in$i) + -1)|0;
       $753 = (($737) + ($idx$19$i)|0);
       $754 = HEAP8[$753>>0]|0;
       switch ($754<<24>>24) {
       case 43: case 95: case 32: case 45:  {
        break;
       }
       default: {
        $idx$19$in$i$lcssa = $idx$19$in$i;
        break L393;
       }
       }
       $755 = ($idx$19$i|0)>(1);
       if ($755) {
        $idx$19$in$i = $idx$19$i;
       } else {
        label = 274;
        break L386;
       }
      }
      $756 = (($737) + ($idx$19$in$i$lcssa)|0);
      HEAP8[$756>>0] = 0;
      label = 274;
     } else {
      label = 274;
     }
    } while(0);
    if ((label|0) == 274) {
     $776 = ($736|0)==(0|0);
     if ($776) {
      label = 276;
      break;
     }
    }
    $777 = ((($cffface)) + 24|0);
    HEAP32[$777>>2] = $736;
   }
  } while(0);
  if ((label|0) == 276) {
   $778 = (_ft_mem_strdup($37,594711,$error$i23)|0);
   $779 = ((($cffface)) + 24|0);
   HEAP32[$779>>2] = $778;
  }
  $780 = ($sfnt_format$0<<24>>24)==(0);
  $$6 = $780 ? 2065 : 2073;
  $781 = ((($38)) + 1348|0);
  $782 = HEAP8[$781>>0]|0;
  $783 = ($782<<24>>24)==(0);
  $784 = $$6 | 4;
  $flags$1 = $783 ? $$6 : $784;
  $785 = ((($cffface)) + 8|0);
  $786 = HEAP32[$785>>2]|0;
  $787 = $flags$1 | $786;
  HEAP32[$785>>2] = $787;
  $788 = ((($38)) + 1352|0);
  $789 = HEAP32[$788>>2]|0;
  $not$ = ($789|0)!=(0);
  $$7 = $not$&1;
  $790 = ((($38)) + 1344|0);
  $791 = HEAP32[$790>>2]|0;
  $792 = ($791|0)==(65535);
  do {
   if ($792) {
    $flags$3 = $$7;
   } else {
    $793 = ($791>>>0)>(390);
    if ($793) {
     $794 = (($791) + -391)|0;
     $795 = ((($38)) + 1312|0);
     $796 = HEAP32[$795>>2]|0;
     $797 = ($796>>>0)>($794>>>0);
     if (!($797)) {
      $flags$3 = $$7;
      break;
     }
     $798 = ((($38)) + 1316|0);
     $799 = HEAP32[$798>>2]|0;
     $800 = (($799) + ($794<<2)|0);
     $801 = HEAP32[$800>>2]|0;
     $$0$i24 = $801;
    } else {
     $802 = HEAP32[$467>>2]|0;
     $803 = ($802|0)==(0|0);
     if ($803) {
      $flags$3 = $$7;
      break;
     }
     $804 = ((($802)) + 20|0);
     $805 = HEAP32[$804>>2]|0;
     $806 = (FUNCTION_TABLE_ii[$805 & 8191]($791)|0);
     $$0$i24 = $806;
    }
    $807 = ($$0$i24|0)==(0|0);
    if ($807) {
     $flags$3 = $$7;
    } else {
     $808 = (_strcmp($$0$i24,594731)|0);
     $809 = ($808|0)==(0);
     if (!($809)) {
      $810 = (_strcmp($$0$i24,594283)|0);
      $811 = ($810|0)==(0);
      if (!($811)) {
       $flags$3 = $$7;
       break;
      }
     }
     $812 = $$7 | 2;
     $flags$3 = $812;
    }
   }
  } while(0);
  $813 = $flags$3 & 2;
  $814 = ($813|0)==(0);
  do {
   if ($814) {
    $815 = ((($cffface)) + 24|0);
    $816 = HEAP32[$815>>2]|0;
    $817 = ($816|0)==(0|0);
    if ($817) {
     $flags$4 = $flags$3;
    } else {
     $818 = (_strncmp($816,594731,4)|0);
     $819 = ($818|0)==(0);
     if (!($819)) {
      $820 = (_strncmp($816,594283,5)|0);
      $821 = ($820|0)==(0);
      if (!($821)) {
       $flags$4 = $flags$3;
       break;
      }
     }
     $822 = $flags$3 | 2;
     $flags$4 = $822;
    }
   } else {
    $flags$4 = $flags$3;
   }
  } while(0);
  $823 = ((($cffface)) + 12|0);
  HEAP32[$823>>2] = $flags$4;
 }
 $824 = HEAP32[$472>>2]|0;
 $825 = ($824|0)==(65535);
 if ($825) {
  $826 = ((($cffface)) + 8|0);
  $827 = HEAP32[$826>>2]|0;
  $828 = $827 | 512;
  HEAP32[$826>>2] = $828;
 } else {
  $or$cond8 = $581 | $825;
  if (!($or$cond8)) {
   $829 = ((($cffface)) + 8|0);
   $830 = HEAP32[$829>>2]|0;
   $831 = $830 | 4096;
   HEAP32[$829>>2] = $831;
  }
 }
 $832 = ((($cffface)) + 36|0);
 $833 = HEAP32[$832>>2]|0;
 $834 = ($833|0)==(0);
 L429: do {
  if ($834) {
   label = 302;
  } else {
   $835 = ((($cffface)) + 40|0);
   $836 = HEAP32[$835>>2]|0;
   $nn$047 = 0;
   while(1) {
    $839 = (($836) + ($nn$047<<2)|0);
    $840 = HEAP32[$839>>2]|0;
    $841 = ((($840)) + 8|0);
    $842 = HEAP32[$841>>2]|0;
    $843 = $842&65535;
    switch ($843<<16>>16) {
    case 0:  {
     label = 308;
     break L429;
     break;
    }
    case 3:  {
     $$mask = $842 & -65536;
     $844 = ($$mask|0)==(65536);
     if ($844) {
      label = 308;
      break L429;
     }
     break;
    }
    default: {
    }
    }
    $837 = (($nn$047) + 1)|0;
    $838 = ($837>>>0)<($833>>>0);
    if ($838) {
     $nn$047 = $837;
    } else {
     label = 302;
     break L429;
    }
   }
  }
 } while(0);
 do {
  if ((label|0) == 302) {
   if (!($581)) {
    $845 = HEAP32[$472>>2]|0;
    $846 = ($845|0)==(65535);
    if (!($846)) {
     break;
    }
   }
   HEAP32[$cmaprec>>2] = $cffface;
   $847 = ((($cmaprec)) + 8|0);
   HEAP16[$847>>1] = 3;
   $848 = ((($cmaprec)) + 10|0);
   HEAP16[$848>>1] = 1;
   $849 = ((($cmaprec)) + 4|0);
   HEAP32[$849>>2] = 1970170211;
   $850 = (_FT_CMap_New(49604,0,$cmaprec,0)|0);
   HEAP32[$error>>2] = $850;
   $851 = ($850|0)==(0);
   $852 = $850 & 255;
   $853 = ($852|0)==(163);
   $or$cond10 = $851 | $853;
   if ($or$cond10) {
    HEAP32[$error>>2] = 0;
    $854 = ((($cffface)) + 92|0);
    $855 = HEAP32[$854>>2]|0;
    $856 = ($855|0)==(0|0);
    if ($856) {
     $857 = HEAP32[$832>>2]|0;
     $858 = ($833|0)==($857|0);
     if ($858) {
      label = 308;
     } else {
      $859 = ((($cffface)) + 40|0);
      $860 = HEAP32[$859>>2]|0;
      $861 = (($860) + ($833<<2)|0);
      $862 = HEAP32[$861>>2]|0;
      HEAP32[$854>>2] = $862;
      label = 308;
     }
    } else {
     label = 308;
    }
   }
  }
 } while(0);
 if ((label|0) == 308) {
  $863 = ((($38)) + 124|0);
  $864 = HEAP32[$863>>2]|0;
  $865 = ($864|0)==(0);
  if (!($865)) {
   HEAP32[$cmaprec>>2] = $cffface;
   $866 = ((($cmaprec)) + 8|0);
   HEAP16[$866>>1] = 7;
   $867 = ((($38)) + 120|0);
   $868 = HEAP32[$867>>2]|0;
   $869 = ($868|0)==(0);
   do {
    if ($869) {
     $870 = ((($cmaprec)) + 10|0);
     HEAP16[$870>>1] = 0;
     $871 = ((($cmaprec)) + 4|0);
     HEAP32[$871>>2] = 1094995778;
    } else {
     $872 = ($868|0)==(1);
     $873 = ((($cmaprec)) + 10|0);
     if ($872) {
      HEAP16[$873>>1] = 1;
      $874 = ((($cmaprec)) + 4|0);
      HEAP32[$874>>2] = 1094992453;
      break;
     } else {
      HEAP16[$873>>1] = 2;
      $875 = ((($cmaprec)) + 4|0);
      HEAP32[$875>>2] = 1094992451;
      break;
     }
    }
   } while(0);
   $876 = (_FT_CMap_New(49564,0,$cmaprec,0)|0);
   HEAP32[$error>>2] = $876;
  }
 }
 $$pre82 = HEAP32[$error>>2]|0;
 $$0 = $$pre82;
 STACKTOP = sp;return ($$0|0);
}
function _cff_index_init($idx,$stream,$load) {
 $idx = $idx|0;
 $stream = $stream|0;
 $load = $load|0;
 var $$lcssa = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $exitcond = 0, $nn$02$i = 0, $result$01$i = 0, $storemerge = 0, $tmp$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tmp$i = sp + 4|0;
 $error = sp;
 $0 = ((($stream)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 ;HEAP32[$idx>>2]=0|0;HEAP32[$idx+4>>2]=0|0;HEAP32[$idx+8>>2]=0|0;HEAP32[$idx+12>>2]=0|0;HEAP32[$idx+16>>2]=0|0;HEAP32[$idx+20>>2]=0|0;HEAP32[$idx+24>>2]=0|0;HEAP32[$idx+28>>2]=0|0;
 HEAP32[$idx>>2] = $stream;
 $2 = (_FT_Stream_Pos($stream)|0);
 $3 = ((($idx)) + 4|0);
 HEAP32[$3>>2] = $2;
 $4 = (_FT_Stream_ReadUShort($stream,$error)|0);
 $5 = HEAP32[$error>>2]|0;
 $6 = ($5|0)==(0);
 do {
  if ($6) {
   $7 = $4&65535;
   $8 = ($4<<16>>16)==(0);
   if ($8) {
    $50 = 0;
    STACKTOP = sp;return ($50|0);
   }
   $9 = (_FT_Stream_ReadChar($stream,$error)|0);
   $10 = HEAP32[$error>>2]|0;
   $11 = ($10|0)==(0);
   if ($11) {
    $12 = $9&255;
    $13 = (($9) + -1)<<24>>24;
    $14 = ($13&255)>(3);
    if ($14) {
     HEAP32[$error>>2] = 8;
     break;
    }
    $15 = ((($idx)) + 8|0);
    HEAP32[$15>>2] = $7;
    $16 = ((($idx)) + 12|0);
    HEAP8[$16>>0] = $9;
    $17 = (($7) + 1)|0;
    $18 = Math_imul($12, $17)|0;
    $19 = HEAP32[$3>>2]|0;
    $20 = (($18) + 3)|0;
    $21 = (($20) + ($19))|0;
    $22 = ((($idx)) + 16|0);
    HEAP32[$22>>2] = $21;
    $23 = Math_imul($12, $7)|0;
    $24 = (_FT_Stream_Skip($stream,$23)|0);
    HEAP32[$error>>2] = $24;
    $25 = ($24|0)==(0);
    if ($25) {
     $26 = HEAP32[$idx>>2]|0;
     $27 = HEAP8[$16>>0]|0;
     $28 = $27&255;
     $29 = (_FT_Stream_Read($26,$tmp$i,$28)|0);
     $30 = ($29|0)==(0);
     if (!($30)) {
      HEAP32[$error>>2] = $29;
      break;
     }
     $31 = HEAP8[$16>>0]|0;
     $32 = ($31<<24>>24)==(0);
     if ($32) {
      HEAP32[$error>>2] = $29;
     } else {
      $33 = $31&255;
      $nn$02$i = 0;$result$01$i = 0;
      while(1) {
       $34 = $result$01$i << 8;
       $35 = (($tmp$i) + ($nn$02$i)|0);
       $36 = HEAP8[$35>>0]|0;
       $37 = $36&255;
       $38 = $37 | $34;
       $39 = (($nn$02$i) + 1)|0;
       $exitcond = ($39|0)==($33|0);
       if ($exitcond) {
        $$lcssa = $38;
        break;
       } else {
        $nn$02$i = $39;$result$01$i = $38;
       }
      }
      HEAP32[$error>>2] = $29;
      if (!($30)) {
       break;
      }
      $40 = ($$lcssa|0)==(0);
      if (!($40)) {
       $41 = (($$lcssa) + -1)|0;
       $42 = ((($idx)) + 20|0);
       HEAP32[$42>>2] = $41;
       $43 = ($load<<24>>24)==(0);
       if ($43) {
        $46 = (_FT_Stream_Skip($stream,$41)|0);
        $storemerge = $46;
       } else {
        $44 = ((($idx)) + 28|0);
        $45 = (_FT_Stream_ExtractFrame($stream,$41,$44)|0);
        $storemerge = $45;
       }
       HEAP32[$error>>2] = $storemerge;
       $47 = ($storemerge|0)==(0);
       if ($47) {
        $50 = 0;
       } else {
        break;
       }
       STACKTOP = sp;return ($50|0);
      }
     }
     HEAP32[$error>>2] = 8;
    }
   }
  }
 } while(0);
 $48 = ((($idx)) + 24|0);
 $49 = HEAP32[$48>>2]|0;
 _ft_mem_free($1,$49);
 HEAP32[$48>>2] = 0;
 $$pre = HEAP32[$error>>2]|0;
 $50 = $$pre;
 STACKTOP = sp;return ($50|0);
}
function _cff_index_get_pointers($idx,$table,$pool) {
 $idx = $idx|0;
 $table = $table|0;
 $pool = $pool|0;
 var $$ = 0, $$$us = 0, $$pr$pre$i = 0, $$pre = 0, $$pre$i = 0, $$pre$phiZ2D = 0, $$pre12 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cur_offset$05 = 0, $cur_offset$05$us = 0, $error = 0, $error$i = 0, $extra$06$us = 0, $extra$1$us = 0;
 var $n$04 = 0, $n$04$us = 0, $new_bytes$0 = 0, $new_bytes$0$ = 0, $next_offset$0 = 0, $next_offset$0$us13 = 0, $p$010$i = 0, $p$113$i = 0, $p$216$i = 0, $p$38$i = 0, $phitmp$i = 0, $poff$011$i = 0, $poff$114$i = 0, $poff$217$i = 0, $poff$39$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp + 4|0;
 $error = sp;
 HEAP32[$error>>2] = 0;
 $0 = HEAP32[$idx>>2]|0;
 $1 = ((($0)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 HEAP32[$table>>2] = 0;
 $3 = ((($idx)) + 24|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 L1: do {
  if ($5) {
   HEAP32[$error$i>>2] = 0;
   $6 = HEAP32[$idx>>2]|0;
   $7 = ((($6)) + 28|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ((($idx)) + 8|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($10|0)==(0);
   do {
    if (!($11)) {
     $12 = ((($idx)) + 12|0);
     $13 = HEAP8[$12>>0]|0;
     $14 = (($10) + 1)|0;
     $15 = $13&255;
     $16 = Math_imul($15, $14)|0;
     $17 = (_ft_mem_realloc($8,4,0,$14,0,$error$i)|0);
     HEAP32[$3>>2] = $17;
     $18 = HEAP32[$error$i>>2]|0;
     $19 = ($18|0)==(0);
     if ($19) {
      $20 = ((($idx)) + 4|0);
      $21 = HEAP32[$20>>2]|0;
      $22 = (($21) + 3)|0;
      $23 = (_FT_Stream_Seek($6,$22)|0);
      HEAP32[$error$i>>2] = $23;
      $24 = ($23|0)==(0);
      if ($24) {
       $25 = (_FT_Stream_EnterFrame($6,$16)|0);
       HEAP32[$error$i>>2] = $25;
       $26 = ($25|0)==(0);
       if ($26) {
        $27 = HEAP32[$3>>2]|0;
        $28 = ((($6)) + 32|0);
        $29 = HEAP32[$28>>2]|0;
        $30 = (($29) + ($16)|0);
        switch ($15|0) {
        case 1:  {
         $33 = ($16|0)>(0);
         if ($33) {
          $p$010$i = $29;$poff$011$i = $27;
          while(1) {
           $35 = HEAP8[$p$010$i>>0]|0;
           $36 = $35&255;
           HEAP32[$poff$011$i>>2] = $36;
           $37 = ((($p$010$i)) + 1|0);
           $38 = ((($poff$011$i)) + 4|0);
           $39 = ($37>>>0)<($30>>>0);
           if ($39) {
            $p$010$i = $37;$poff$011$i = $38;
           } else {
            break;
           }
          }
         }
         break;
        }
        case 2:  {
         $32 = ($16|0)>(0);
         if ($32) {
          $p$113$i = $29;$poff$114$i = $27;
          while(1) {
           $40 = HEAP8[$p$113$i>>0]|0;
           $41 = $40&255;
           $42 = $41 << 8;
           $43 = ((($p$113$i)) + 1|0);
           $44 = HEAP8[$43>>0]|0;
           $45 = $44&255;
           $46 = $42 | $45;
           HEAP32[$poff$114$i>>2] = $46;
           $47 = ((($p$113$i)) + 2|0);
           $48 = ((($poff$114$i)) + 4|0);
           $49 = ($47>>>0)<($30>>>0);
           if ($49) {
            $p$113$i = $47;$poff$114$i = $48;
           } else {
            break;
           }
          }
         }
         break;
        }
        case 3:  {
         $31 = ($16|0)>(0);
         if ($31) {
          $p$216$i = $29;$poff$217$i = $27;
          while(1) {
           $50 = HEAP8[$p$216$i>>0]|0;
           $51 = $50&255;
           $52 = $51 << 16;
           $53 = ((($p$216$i)) + 1|0);
           $54 = HEAP8[$53>>0]|0;
           $55 = $54&255;
           $56 = $55 << 8;
           $57 = $56 | $52;
           $58 = ((($p$216$i)) + 2|0);
           $59 = HEAP8[$58>>0]|0;
           $60 = $59&255;
           $61 = $57 | $60;
           HEAP32[$poff$217$i>>2] = $61;
           $62 = ((($p$216$i)) + 3|0);
           $63 = ((($poff$217$i)) + 4|0);
           $64 = ($62>>>0)<($30>>>0);
           if ($64) {
            $p$216$i = $62;$poff$217$i = $63;
           } else {
            break;
           }
          }
         }
         break;
        }
        default: {
         $34 = ($16|0)>(0);
         if ($34) {
          $p$38$i = $29;$poff$39$i = $27;
          while(1) {
           $65 = HEAP8[$p$38$i>>0]|0;
           $66 = $65&255;
           $67 = $66 << 24;
           $68 = ((($p$38$i)) + 1|0);
           $69 = HEAP8[$68>>0]|0;
           $70 = $69&255;
           $71 = $70 << 16;
           $72 = $71 | $67;
           $73 = ((($p$38$i)) + 2|0);
           $74 = HEAP8[$73>>0]|0;
           $75 = $74&255;
           $76 = $75 << 8;
           $77 = $72 | $76;
           $78 = ((($p$38$i)) + 3|0);
           $79 = HEAP8[$78>>0]|0;
           $80 = $79&255;
           $81 = $77 | $80;
           HEAP32[$poff$39$i>>2] = $81;
           $82 = ((($p$38$i)) + 4|0);
           $83 = ((($poff$39$i)) + 4|0);
           $84 = ($82>>>0)<($30>>>0);
           if ($84) {
            $p$38$i = $82;$poff$39$i = $83;
           } else {
            break;
           }
          }
         }
        }
        }
        _FT_Stream_ExitFrame($6);
        $$pr$pre$i = HEAP32[$error$i>>2]|0;
        $phitmp$i = ($$pr$pre$i|0)==(0);
        if ($phitmp$i) {
         break;
        }
       }
      }
     }
     $85 = HEAP32[$3>>2]|0;
     _ft_mem_free($8,$85);
     HEAP32[$3>>2] = 0;
     $$pre$i = HEAP32[$error$i>>2]|0;
     HEAP32[$error>>2] = $$pre$i;
     $86 = ($$pre$i|0)==(0);
     if ($86) {
      $$pre$phiZ2D = $9;
      break L1;
     }
     $145 = HEAP32[$error>>2]|0;
     STACKTOP = sp;return ($145|0);
    }
   } while(0);
   HEAP32[$error>>2] = 0;
   $$pre$phiZ2D = $9;
  } else {
   $$pre12 = ((($idx)) + 8|0);
   $$pre$phiZ2D = $$pre12;
  }
 } while(0);
 $87 = HEAP32[$$pre$phiZ2D>>2]|0;
 $88 = ($87|0)==(0);
 if ($88) {
  $145 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($145|0);
 }
 $89 = (($87) + 1)|0;
 $90 = (_ft_mem_realloc($2,4,0,$89,0,$error)|0);
 $91 = HEAP32[$error>>2]|0;
 $92 = ($91|0)==(0);
 if (!($92)) {
  $145 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($145|0);
 }
 $93 = ($pool|0)!=(0|0);
 if ($93) {
  $94 = ((($idx)) + 20|0);
  $95 = HEAP32[$94>>2]|0;
  $96 = HEAP32[$$pre$phiZ2D>>2]|0;
  $97 = (($96) + ($95))|0;
  $98 = (_ft_mem_alloc($2,$97,$error)|0);
  $99 = HEAP32[$error>>2]|0;
  $100 = ($99|0)==(0);
  if ($100) {
   $new_bytes$0 = $98;
  } else {
   $145 = HEAP32[$error>>2]|0;
   STACKTOP = sp;return ($145|0);
  }
 } else {
  $new_bytes$0 = 0;
 }
 $101 = ((($idx)) + 28|0);
 $102 = HEAP32[$101>>2]|0;
 $new_bytes$0$ = $93 ? $new_bytes$0 : $102;
 HEAP32[$90>>2] = $new_bytes$0$;
 $103 = HEAP32[$$pre$phiZ2D>>2]|0;
 $104 = ($103|0)==(0);
 L38: do {
  if (!($104)) {
   $105 = ((($idx)) + 20|0);
   if ($93) {
    $146 = $103;$cur_offset$05$us = 0;$extra$06$us = 0;$n$04$us = 1;
   } else {
    $cur_offset$05 = 0;$n$04 = 1;
    while(1) {
     $134 = HEAP32[$3>>2]|0;
     $135 = (($134) + ($n$04<<2)|0);
     $136 = HEAP32[$135>>2]|0;
     $137 = (($136) + -1)|0;
     $138 = ($137>>>0)<($cur_offset$05>>>0);
     if ($138) {
      $next_offset$0 = $cur_offset$05;
     } else {
      $139 = HEAP32[$105>>2]|0;
      $140 = ($137>>>0)>($139>>>0);
      $$ = $140 ? $139 : $137;
      $next_offset$0 = $$;
     }
     $141 = (($102) + ($next_offset$0)|0);
     $142 = (($90) + ($n$04<<2)|0);
     HEAP32[$142>>2] = $141;
     $143 = (($n$04) + 1)|0;
     $144 = ($143>>>0)>($103>>>0);
     if ($144) {
      break L38;
     } else {
      $cur_offset$05 = $next_offset$0;$n$04 = $143;
     }
    }
   }
   while(1) {
    $106 = HEAP32[$3>>2]|0;
    $107 = (($106) + ($n$04$us<<2)|0);
    $108 = HEAP32[$107>>2]|0;
    $109 = (($108) + -1)|0;
    $110 = ($109>>>0)<($cur_offset$05$us>>>0);
    if ($110) {
     $111 = (($new_bytes$0) + ($cur_offset$05$us)|0);
     $112 = (($111) + ($extra$06$us)|0);
     $113 = (($90) + ($n$04$us<<2)|0);
     HEAP32[$113>>2] = $112;
     $133 = $146;$extra$1$us = $extra$06$us;$next_offset$0$us13 = $cur_offset$05$us;
    } else {
     $114 = HEAP32[$105>>2]|0;
     $115 = ($109>>>0)>($114>>>0);
     $$$us = $115 ? $114 : $109;
     $116 = (($new_bytes$0) + ($$$us)|0);
     $117 = (($116) + ($extra$06$us)|0);
     $118 = (($90) + ($n$04$us<<2)|0);
     HEAP32[$118>>2] = $117;
     $119 = ($$$us|0)==($cur_offset$05$us|0);
     if ($119) {
      $133 = $146;$extra$1$us = $extra$06$us;$next_offset$0$us13 = $$$us;
     } else {
      $120 = (($n$04$us) + -1)|0;
      $121 = (($90) + ($120<<2)|0);
      $122 = HEAP32[$121>>2]|0;
      $123 = (($102) + ($cur_offset$05$us)|0);
      $124 = $117;
      $125 = $122;
      $126 = (($124) - ($125))|0;
      _memcpy(($122|0),($123|0),($126|0))|0;
      $127 = HEAP32[$118>>2]|0;
      HEAP8[$127>>0] = 0;
      $128 = HEAP32[$118>>2]|0;
      $129 = ((($128)) + 1|0);
      HEAP32[$118>>2] = $129;
      $130 = (($extra$06$us) + 1)|0;
      $$pre = HEAP32[$$pre$phiZ2D>>2]|0;
      $133 = $$pre;$extra$1$us = $130;$next_offset$0$us13 = $$$us;
     }
    }
    $131 = (($n$04$us) + 1)|0;
    $132 = ($131>>>0)>($133>>>0);
    if ($132) {
     break;
    } else {
     $146 = $133;$cur_offset$05$us = $next_offset$0$us13;$extra$06$us = $extra$1$us;$n$04$us = $131;
    }
   }
  }
 } while(0);
 HEAP32[$table>>2] = $90;
 if (!($93)) {
  $145 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($145|0);
 }
 HEAP32[$pool>>2] = $new_bytes$0;
 $145 = HEAP32[$error>>2]|0;
 STACKTOP = sp;return ($145|0);
}
function _cff_subfont_load($font,$idx,$font_index,$stream,$base_offset,$library) {
 $font = $font|0;
 $idx = $idx|0;
 $font_index = $font_index|0;
 $stream = $stream|0;
 $base_offset = $base_offset|0;
 $library = $library|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0, $dict = 0, $dict_len = 0, $error$0 = 0, $error$2 = 0, $parser = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 432|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $parser = sp + 8|0;
 $dict = sp + 4|0;
 $dict_len = sp;
 HEAP32[$dict>>2] = 0;
 _memset(($parser|0),0,408)|0;
 $0 = ((($parser)) + 16|0);
 $1 = ((($parser)) + 404|0);
 HEAP32[$1>>2] = $0;
 $2 = ((($parser)) + 408|0);
 HEAP32[$2>>2] = 4096;
 $3 = ((($parser)) + 412|0);
 HEAP32[$3>>2] = $font;
 HEAP32[$parser>>2] = $library;
 _memset(($font|0),0,172)|0;
 $4 = ((($font)) + 32|0);
 HEAP32[$4>>2] = -6553600;
 $5 = ((($font)) + 36|0);
 HEAP32[$5>>2] = 3276800;
 $6 = ((($font)) + 44|0);
 HEAP32[$6>>2] = 2;
 $7 = ((($font)) + 48|0);
 HEAP32[$7>>2] = 65536;
 $8 = ((($font)) + 60|0);
 HEAP32[$8>>2] = 65536;
 $9 = ((($font)) + 156|0);
 HEAP32[$9>>2] = 8720;
 HEAP32[$font>>2] = 65535;
 $10 = ((($font)) + 4|0);
 HEAP32[$10>>2] = 65535;
 $11 = ((($font)) + 8|0);
 HEAP32[$11>>2] = 65535;
 $12 = ((($font)) + 12|0);
 HEAP32[$12>>2] = 65535;
 $13 = ((($font)) + 16|0);
 HEAP32[$13>>2] = 65535;
 $14 = ((($font)) + 20|0);
 HEAP32[$14>>2] = 65535;
 $15 = ((($font)) + 128|0);
 HEAP32[$15>>2] = 65535;
 $16 = ((($font)) + 132|0);
 HEAP32[$16>>2] = 65535;
 $17 = ((($font)) + 136|0);
 HEAP32[$17>>2] = 65535;
 $18 = ((($font)) + 172|0);
 HEAP32[$18>>2] = 65535;
 $19 = (_cff_index_access_element($idx,$font_index,$dict,$dict_len)|0);
 $20 = ($19|0)==(0);
 if ($20) {
  $21 = HEAP32[$dict>>2]|0;
  $22 = HEAP32[$dict_len>>2]|0;
  $23 = (($21) + ($22)|0);
  $24 = (_cff_parser_run($parser,$21,$23)|0);
  $error$0 = $24;
 } else {
  $error$0 = $19;
 }
 $25 = ((($idx)) + 28|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ($26|0)==(0|0);
 if ($27) {
  $28 = HEAP32[$idx>>2]|0;
  _FT_Stream_ReleaseFrame($28,$dict);
 }
 $29 = ($error$0|0)==(0);
 if (!($29)) {
  $error$2 = $error$0;
  STACKTOP = sp;return ($error$2|0);
 }
 $30 = HEAP32[$16>>2]|0;
 $31 = ($30|0)==(65535);
 if (!($31)) {
  $error$2 = 0;
  STACKTOP = sp;return ($error$2|0);
 }
 $32 = ((($font)) + 116|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ($33|0)==(0);
 do {
  if (!($34)) {
   $35 = ((($font)) + 120|0);
   $36 = HEAP32[$35>>2]|0;
   $37 = ($36|0)==(0);
   if (!($37)) {
    $38 = ((($font)) + 176|0);
    _memset(($38|0),0,360)|0;
    $39 = ((($font)) + 376|0);
    HEAP32[$39>>2] = 7;
    $40 = ((($font)) + 380|0);
    HEAP32[$40>>2] = 1;
    $41 = ((($font)) + 508|0);
    HEAP32[$41>>2] = -1;
    $42 = ((($font)) + 516|0);
    HEAP32[$42>>2] = 3932;
    $43 = ((($font)) + 372|0);
    HEAP32[$43>>2] = 2596864;
    _memset(($parser|0),0,408)|0;
    HEAP32[$1>>2] = $0;
    HEAP32[$2>>2] = 8192;
    HEAP32[$3>>2] = $38;
    HEAP32[$parser>>2] = $library;
    $44 = (($33) + ($base_offset))|0;
    $45 = (_FT_Stream_Seek($stream,$44)|0);
    $46 = ($45|0)==(0);
    if (!($46)) {
     $error$2 = $45;
     STACKTOP = sp;return ($error$2|0);
    }
    $47 = HEAP32[$35>>2]|0;
    $48 = (_FT_Stream_EnterFrame($stream,$47)|0);
    $49 = ($48|0)==(0);
    if (!($49)) {
     $error$2 = $48;
     STACKTOP = sp;return ($error$2|0);
    }
    $50 = ((($stream)) + 32|0);
    $51 = HEAP32[$50>>2]|0;
    $52 = ((($stream)) + 36|0);
    $53 = HEAP32[$52>>2]|0;
    $54 = (_cff_parser_run($parser,$51,$53)|0);
    _FT_Stream_ExitFrame($stream);
    $55 = ($54|0)==(0);
    if ($55) {
     $56 = HEAP8[$38>>0]|0;
     $57 = $56&255;
     $58 = $57 & 254;
     $59 = $58&255;
     HEAP8[$38>>0] = $59;
     break;
    } else {
     $error$2 = $54;
     STACKTOP = sp;return ($error$2|0);
    }
   }
  }
 } while(0);
 $60 = ((($font)) + 524|0);
 $61 = HEAP32[$60>>2]|0;
 $62 = ($61|0)==(0);
 if ($62) {
  $error$2 = 0;
  STACKTOP = sp;return ($error$2|0);
 }
 $63 = HEAP32[$32>>2]|0;
 $64 = (($61) + ($base_offset))|0;
 $65 = (($64) + ($63))|0;
 $66 = (_FT_Stream_Seek($stream,$65)|0);
 $67 = ($66|0)==(0);
 if (!($67)) {
  $error$2 = $66;
  STACKTOP = sp;return ($error$2|0);
 }
 $68 = ((($font)) + 536|0);
 $69 = (_cff_index_init($68,$stream,1)|0);
 $70 = ($69|0)==(0);
 if (!($70)) {
  $error$2 = $69;
  STACKTOP = sp;return ($error$2|0);
 }
 $71 = ((($font)) + 568|0);
 $72 = (_cff_index_get_pointers($68,$71,0)|0);
 $error$2 = $72;
 STACKTOP = sp;return ($error$2|0);
}
function _cff_index_access_element($idx,$element,$pbytes,$pbyte_len) {
 $idx = $idx|0;
 $element = $element|0;
 $pbytes = $pbytes|0;
 $pbyte_len = $pbyte_len|0;
 var $$03 = 0, $$14 = 0, $$lcssa = 0, $$lcssa61 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre36 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $error$1$ph = 0, $error$116 = 0, $error$117 = 0, $error$118 = 0, $error$4 = 0, $exitcond = 0, $exitcond35 = 0, $nn$02$i = 0, $nn$02$i8 = 0, $off1$1 = 0, $off1$1$ph = 0, $off1$119 = 0, $off1$120 = 0, $off1$121 = 0, $off2$2 = 0, $off2$222 = 0, $off2$3 = 0, $or$cond = 0, $result$01$i = 0;
 var $result$01$i9 = 0, $tmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tmp$i = sp;
 $0 = ($idx|0)==(0|0);
 if ($0) {
  $error$4 = 6;
  STACKTOP = sp;return ($error$4|0);
 }
 $1 = ((($idx)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2>>>0)>($element>>>0);
 if (!($3)) {
  $error$4 = 6;
  STACKTOP = sp;return ($error$4|0);
 }
 $4 = HEAP32[$idx>>2]|0;
 $5 = ((($idx)) + 24|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0|0);
 L7: do {
  if ($7) {
   $8 = ((($idx)) + 12|0);
   $9 = HEAP8[$8>>0]|0;
   $10 = $9&255;
   $11 = Math_imul($10, $element)|0;
   $12 = ((($idx)) + 4|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = (($13) + 3)|0;
   $15 = (($14) + ($11))|0;
   $16 = (_FT_Stream_Seek($4,$15)|0);
   $17 = ($16|0)==(0);
   if (!($17)) {
    $error$4 = $16;
    STACKTOP = sp;return ($error$4|0);
   }
   $18 = HEAP32[$idx>>2]|0;
   $19 = HEAP8[$8>>0]|0;
   $20 = $19&255;
   $21 = (_FT_Stream_Read($18,$tmp$i,$20)|0);
   $22 = ($21|0)==(0);
   if (!($22)) {
    $error$4 = $21;
    STACKTOP = sp;return ($error$4|0);
   }
   $23 = HEAP8[$8>>0]|0;
   $24 = ($23<<24>>24)==(0);
   if ($24) {
   } else {
    $25 = $23&255;
    $nn$02$i = 0;$result$01$i = 0;
    while(1) {
     $26 = $result$01$i << 8;
     $27 = (($tmp$i) + ($nn$02$i)|0);
     $28 = HEAP8[$27>>0]|0;
     $29 = $28&255;
     $30 = $29 | $26;
     $31 = (($nn$02$i) + 1)|0;
     $exitcond35 = ($31|0)==($25|0);
     if ($exitcond35) {
      $$lcssa61 = $30;
      break;
     } else {
      $nn$02$i = $31;$result$01$i = $30;
     }
    }
    $32 = ($$lcssa61|0)==(0);
    if (!($32)) {
     $$03 = $element;$36 = $23;
     while(1) {
      $33 = (($$03) + 1)|0;
      $34 = HEAP32[$idx>>2]|0;
      $35 = $36&255;
      $37 = (_FT_Stream_Read($34,$tmp$i,$35)|0);
      $38 = ($37|0)==(0);
      if ($38) {
       $39 = HEAP8[$8>>0]|0;
       $40 = ($39<<24>>24)==(0);
       if ($40) {
        label = 16;
       } else {
        $41 = $39&255;
        $nn$02$i8 = 0;$result$01$i9 = 0;
        while(1) {
         $42 = $result$01$i9 << 8;
         $43 = (($tmp$i) + ($nn$02$i8)|0);
         $44 = HEAP8[$43>>0]|0;
         $45 = $44&255;
         $46 = $45 | $42;
         $47 = (($nn$02$i8) + 1)|0;
         $exitcond = ($47|0)==($41|0);
         if ($exitcond) {
          $$lcssa = $46;
          break;
         } else {
          $nn$02$i8 = $47;$result$01$i9 = $46;
         }
        }
        $48 = ($$lcssa|0)==(0);
        if (!($48)) {
         $off1$1 = $$lcssa61;$off2$2 = $$lcssa;
         label = 26;
         break L7;
        }
       }
      } else {
       label = 16;
      }
      if ((label|0) == 16) {
       label = 0;
      }
      $49 = HEAP32[$1>>2]|0;
      $50 = ($33>>>0)<($49>>>0);
      if (!($50)) {
       $error$1$ph = $37;$off1$1$ph = $$lcssa61;
       label = 25;
       break L7;
      }
      $$pre = HEAP8[$8>>0]|0;
      $$03 = $33;$36 = $$pre;
     }
    }
   }
   $61 = ((($4)) + 4|0);
   $62 = HEAP32[$61>>2]|0;
   $72 = $62;$error$118 = 0;$off1$121 = 0;$off2$222 = 0;
   label = 28;
  } else {
   $51 = (($6) + ($element<<2)|0);
   $52 = HEAP32[$51>>2]|0;
   $53 = ($52|0)==(0);
   if ($53) {
    $54 = ((($4)) + 4|0);
    $55 = HEAP32[$54>>2]|0;
    $72 = $55;$error$118 = 0;$off1$121 = 0;$off2$222 = 0;
    label = 28;
    break;
   } else {
    $$14 = $element;
   }
   while(1) {
    $56 = (($$14) + 1)|0;
    $57 = (($6) + ($56<<2)|0);
    $58 = HEAP32[$57>>2]|0;
    $59 = ($58|0)==(0);
    if (!($59)) {
     $off1$1 = $52;$off2$2 = $58;
     label = 26;
     break L7;
    }
    $60 = ($56>>>0)<($2>>>0);
    if ($60) {
     $$14 = $56;
    } else {
     $error$1$ph = 0;$off1$1$ph = $52;
     label = 25;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 25) {
  $63 = ((($4)) + 4|0);
  $64 = HEAP32[$63>>2]|0;
  $72 = $64;$error$118 = $error$1$ph;$off1$121 = $off1$1$ph;$off2$222 = 0;
  label = 28;
 }
 else if ((label|0) == 26) {
  $65 = ((($4)) + 4|0);
  $66 = HEAP32[$65>>2]|0;
  $67 = (($66) + 1)|0;
  $68 = ($off2$2>>>0)>($67>>>0);
  if ($68) {
   $$phi$trans$insert = ((($idx)) + 16|0);
   $$pre36 = HEAP32[$$phi$trans$insert>>2]|0;
   $76 = $66;$78 = $$pre36;$error$117 = 0;$off1$120 = $off1$1;
   label = 29;
  } else {
   $72 = $66;$error$118 = 0;$off1$121 = $off1$1;$off2$222 = $off2$2;
   label = 28;
  }
 }
 if ((label|0) == 28) {
  $69 = ((($idx)) + 16|0);
  $70 = HEAP32[$69>>2]|0;
  $71 = (($72) + 1)|0;
  $73 = (($71) - ($off2$222))|0;
  $74 = ($70>>>0)>($73>>>0);
  if ($74) {
   $76 = $72;$78 = $70;$error$117 = $error$118;$off1$120 = $off1$121;
   label = 29;
  } else {
   $error$116 = $error$118;$off1$119 = $off1$121;$off2$3 = $off2$222;
  }
 }
 if ((label|0) == 29) {
  $75 = (($76) + 1)|0;
  $77 = (($75) - ($78))|0;
  $error$116 = $error$117;$off1$119 = $off1$120;$off2$3 = $77;
 }
 $79 = ($off1$119|0)!=(0);
 $80 = ($off2$3>>>0)>($off1$119>>>0);
 $or$cond = $79 & $80;
 if (!($or$cond)) {
  HEAP32[$pbytes>>2] = 0;
  HEAP32[$pbyte_len>>2] = 0;
  $error$4 = $error$116;
  STACKTOP = sp;return ($error$4|0);
 }
 $81 = (($off2$3) - ($off1$119))|0;
 HEAP32[$pbyte_len>>2] = $81;
 $82 = ((($idx)) + 28|0);
 $83 = HEAP32[$82>>2]|0;
 $84 = ($83|0)==(0|0);
 if (!($84)) {
  $85 = (($83) + ($off1$119)|0);
  $86 = ((($85)) + -1|0);
  HEAP32[$pbytes>>2] = $86;
  $error$4 = $error$116;
  STACKTOP = sp;return ($error$4|0);
 }
 $87 = ((($idx)) + 16|0);
 $88 = HEAP32[$87>>2]|0;
 $89 = (($off1$119) + -1)|0;
 $90 = (($89) + ($88))|0;
 $91 = (_FT_Stream_Seek($4,$90)|0);
 $92 = ($91|0)==(0);
 if (!($92)) {
  $error$4 = $91;
  STACKTOP = sp;return ($error$4|0);
 }
 $93 = (_FT_Stream_ExtractFrame($4,$81,$pbytes)|0);
 $error$4 = $93;
 STACKTOP = sp;return ($error$4|0);
}
function _cff_parser_run($parser,$start,$limit) {
 $parser = $parser|0;
 $start = $start|0;
 $limit = $limit|0;
 var $$ = 0, $$$i$i = 0, $$$i$i12 = 0, $$05 = 0, $$idx = 0, $$idx$val = 0, $$idx11$val = 0, $$idx9$val = 0, $$lcssa = 0, $$mask = 0, $$p$0 = 0, $$pre = 0, $$val = 0, $$val10 = 0, $$val8 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0;
 var $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0;
 var $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0;
 var $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0;
 var $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $code$0 = 0, $data$0$idx$val = 0, $data$0$val = 0, $data$044 = 0, $field$040 = 0, $field$040$lcssa = 0;
 var $num_args$141 = 0, $or$cond = 0, $or$cond4 = 0, $or$cond7 = 0, $p$0$pn = 0, $p$045 = 0, $p$1 = 0, $p$2 = 0, $p$4 = 0, $q$043 = 0, $sext$i = 0, $sext$i14 = 0, $sext$i17 = 0, $sext$i20 = 0, $val$0 = 0, $val$0$i18 = 0, $val$0$i1823 = 0, $val$0$i21 = 0, $val$0$i2160 = 0, $val$142 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($parser)) + 16|0);
 $1 = ((($parser)) + 404|0);
 HEAP32[$1>>2] = $0;
 $2 = ((($parser)) + 4|0);
 HEAP32[$2>>2] = $start;
 $3 = ((($parser)) + 8|0);
 HEAP32[$3>>2] = $limit;
 $4 = ((($parser)) + 12|0);
 HEAP32[$4>>2] = $start;
 $5 = ($start>>>0)<($limit>>>0);
 if (!($5)) {
  $$05 = 0;
  return ($$05|0);
 }
 $6 = $0;
 $7 = ((($parser)) + 408|0);
 $8 = ((($parser)) + 412|0);
 $$idx = ((($parser)) + 20|0);
 $p$045 = $start;
 L4: while(1) {
  $9 = HEAP8[$p$045>>0]|0;
  $10 = ($9&255)>(26);
  $11 = ($9<<24>>24)!=(31);
  $or$cond = $10 & $11;
  L6: do {
   if ($or$cond) {
    $12 = HEAP32[$1>>2]|0;
    $13 = $12;
    $14 = (($13) - ($6))|0;
    $15 = ($14|0)>(380);
    if ($15) {
     $$05 = 6;
     label = 95;
     break L4;
    }
    $16 = ((($12)) + 4|0);
    HEAP32[$1>>2] = $16;
    HEAP32[$12>>2] = $p$045;
    switch ($9<<24>>24) {
    case 30:  {
     $p$0$pn = $p$045;
     while(1) {
      $p$1 = ((($p$0$pn)) + 1|0);
      $17 = ($p$1>>>0)<($limit>>>0);
      if (!($17)) {
       $$05 = 0;
       label = 95;
       break L4;
      }
      $18 = HEAP8[$p$1>>0]|0;
      $19 = $18&255;
      $$mask = $19 & 240;
      $20 = ($$mask|0)==(240);
      $21 = $19 & 15;
      $22 = ($21|0)==(15);
      $or$cond7 = $20 | $22;
      if ($or$cond7) {
       $p$4 = $p$1;
       break;
      } else {
       $p$0$pn = $p$1;
      }
     }
     break;
    }
    case 28:  {
     $23 = ((($p$045)) + 2|0);
     $p$4 = $23;
     break L6;
     break;
    }
    case 29:  {
     $24 = ((($p$045)) + 4|0);
     $p$4 = $24;
     break L6;
     break;
    }
    default: {
     $25 = ($9&255)>(246);
     $26 = ((($p$045)) + 1|0);
     $$p$0 = $25 ? $26 : $p$045;
     $p$4 = $$p$0;
     break L6;
    }
    }
   } else {
    $27 = $9&255;
    $28 = HEAP32[$1>>2]|0;
    $29 = $28;
    $30 = (($29) - ($6))|0;
    $31 = $30 >> 2;
    HEAP32[$28>>2] = $p$045;
    $32 = ($9<<24>>24)==(12);
    if ($32) {
     $33 = ((($p$045)) + 1|0);
     $34 = ($33>>>0)<($limit>>>0);
     if (!($34)) {
      $$05 = 6;
      label = 95;
      break L4;
     }
     $35 = HEAP8[$33>>0]|0;
     $36 = $35&255;
     $37 = $36 | 256;
     $code$0 = $37;$p$2 = $33;
    } else {
     $code$0 = $27;$p$2 = $p$045;
    }
    $38 = HEAP32[$7>>2]|0;
    $39 = $38 | $code$0;
    $298 = 4;$field$040 = 49664;
    while(1) {
     $40 = ((($field$040)) + 4|0);
     $41 = HEAP32[$40>>2]|0;
     $42 = ($41|0)==($39|0);
     if ($42) {
      $$lcssa = $298;$field$040$lcssa = $field$040;
      label = 16;
      break;
     }
     $293 = ((($field$040)) + 28|0);
     $294 = HEAP32[$293>>2]|0;
     $295 = ($294|0)==(0);
     if ($295) {
      break;
     } else {
      $298 = $294;$field$040 = $293;
     }
    }
    L24: do {
     if ((label|0) == 16) {
      label = 0;
      $43 = HEAP32[$8>>2]|0;
      $44 = ((($field$040$lcssa)) + 8|0);
      $45 = HEAP32[$44>>2]|0;
      $46 = (($43) + ($45)|0);
      $47 = ($$lcssa|0)!=(6);
      $48 = ($31|0)==(0);
      $or$cond4 = $48 & $47;
      if ($or$cond4) {
       $$05 = 6;
       label = 95;
       break L4;
      }
      L27: do {
       switch ($$lcssa|0) {
       case 1: case 4: case 5:  {
        $$val = HEAP32[$0>>2]|0;
        $$idx$val = HEAP32[$$idx>>2]|0;
        $49 = HEAP8[$$val>>0]|0;
        $50 = ($49<<24>>24)==(30);
        if ($50) {
         $51 = (_cff_parse_real($$val,$$idx$val,0,0)|0);
         $52 = $51 >> 16;
         $val$0 = $52;
         break L27;
        }
        $53 = ((($$val)) + 1|0);
        $54 = $49&255;
        switch ($49<<24>>24) {
        case 28:  {
         $55 = ((($$val)) + 3|0);
         $56 = ($55>>>0)>($$idx$val>>>0);
         if ($56) {
          $val$0 = 0;
          break L27;
         }
         $57 = HEAP8[$53>>0]|0;
         $58 = $57&255;
         $59 = $58 << 8;
         $60 = ((($$val)) + 2|0);
         $61 = HEAP8[$60>>0]|0;
         $62 = $61&255;
         $63 = $59 | $62;
         $sext$i = $63 << 16;
         $64 = $sext$i >> 16;
         $val$0 = $64;
         break L27;
         break;
        }
        case 29:  {
         $65 = ((($$val)) + 5|0);
         $66 = ($65>>>0)>($$idx$val>>>0);
         if ($66) {
          $val$0 = 0;
          break L27;
         }
         $67 = HEAP8[$53>>0]|0;
         $68 = $67&255;
         $69 = $68 << 24;
         $70 = ((($$val)) + 2|0);
         $71 = HEAP8[$70>>0]|0;
         $72 = $71&255;
         $73 = $72 << 16;
         $74 = $73 | $69;
         $75 = ((($$val)) + 3|0);
         $76 = HEAP8[$75>>0]|0;
         $77 = $76&255;
         $78 = $77 << 8;
         $79 = $74 | $78;
         $80 = ((($$val)) + 4|0);
         $81 = HEAP8[$80>>0]|0;
         $82 = $81&255;
         $83 = $79 | $82;
         $val$0 = $83;
         break L27;
         break;
        }
        default: {
         $84 = ($49&255)<(247);
         if ($84) {
          $85 = (($54) + -139)|0;
          $val$0 = $85;
          break L27;
         }
         $86 = ($49&255)<(251);
         $87 = ((($$val)) + 2|0);
         $88 = ($87>>>0)>($$idx$val>>>0);
         if ($86) {
          if ($88) {
           $val$0 = 0;
           break L27;
          }
          $89 = $54 << 8;
          $90 = (($89) + -63232)|0;
          $91 = HEAP8[$53>>0]|0;
          $92 = $91&255;
          $93 = $92 | $90;
          $94 = (($93) + 108)|0;
          $val$0 = $94;
          break L27;
         } else {
          if ($88) {
           $val$0 = 0;
           break L27;
          }
          $95 = (251 - ($54))|0;
          $96 = $95 << 8;
          $97 = HEAP8[$53>>0]|0;
          $98 = $97&255;
          $99 = (($96) + -108)|0;
          $100 = (($99) - ($98))|0;
          $val$0 = $100;
          break L27;
         }
        }
        }
        break;
       }
       case 2:  {
        $$val8 = HEAP32[$0>>2]|0;
        $$idx9$val = HEAP32[$$idx>>2]|0;
        $101 = HEAP8[$$val8>>0]|0;
        $102 = ($101<<24>>24)==(30);
        if ($102) {
         $103 = (_cff_parse_real($$val8,$$idx9$val,0,0)|0);
         $val$0 = $103;
         break L27;
        }
        $104 = ((($$val8)) + 1|0);
        $105 = $101&255;
        L51: do {
         switch ($101<<24>>24) {
         case 28:  {
          $106 = ((($$val8)) + 3|0);
          $107 = ($106>>>0)>($$idx9$val>>>0);
          if ($107) {
           $val$0$i1823 = 0;
          } else {
           $108 = HEAP8[$104>>0]|0;
           $109 = $108&255;
           $110 = $109 << 8;
           $111 = ((($$val8)) + 2|0);
           $112 = HEAP8[$111>>0]|0;
           $113 = $112&255;
           $114 = $110 | $113;
           $sext$i17 = $114 << 16;
           $115 = $sext$i17 >> 16;
           $val$0$i1823 = $115;
          }
          break;
         }
         case 29:  {
          $116 = ((($$val8)) + 5|0);
          $117 = ($116>>>0)>($$idx9$val>>>0);
          if ($117) {
           $val$0$i1823 = 0;
          } else {
           $118 = HEAP8[$104>>0]|0;
           $119 = $118&255;
           $120 = $119 << 24;
           $121 = ((($$val8)) + 2|0);
           $122 = HEAP8[$121>>0]|0;
           $123 = $122&255;
           $124 = $123 << 16;
           $125 = $124 | $120;
           $126 = ((($$val8)) + 3|0);
           $127 = HEAP8[$126>>0]|0;
           $128 = $127&255;
           $129 = $128 << 8;
           $130 = $125 | $129;
           $131 = ((($$val8)) + 4|0);
           $132 = HEAP8[$131>>0]|0;
           $133 = $132&255;
           $134 = $130 | $133;
           $val$0$i18 = $134;
           label = 46;
          }
          break;
         }
         default: {
          $135 = ($101&255)<(247);
          if ($135) {
           $136 = (($105) + -139)|0;
           $val$0$i1823 = $136;
           break L51;
          }
          $137 = ($101&255)<(251);
          $138 = ((($$val8)) + 2|0);
          $139 = ($138>>>0)>($$idx9$val>>>0);
          if ($137) {
           if ($139) {
            $val$0$i1823 = 0;
            break L51;
           }
           $140 = $105 << 8;
           $141 = (($140) + -63232)|0;
           $142 = HEAP8[$104>>0]|0;
           $143 = $142&255;
           $144 = $143 | $141;
           $145 = (($144) + 108)|0;
           $val$0$i18 = $145;
           label = 46;
           break L51;
          } else {
           if ($139) {
            $val$0$i1823 = 0;
            break L51;
           }
           $146 = (251 - ($105))|0;
           $147 = $146 << 8;
           $148 = HEAP8[$104>>0]|0;
           $149 = $148&255;
           $150 = (($147) + -108)|0;
           $151 = (($150) - ($149))|0;
           $val$0$i18 = $151;
           label = 46;
           break L51;
          }
         }
         }
        } while(0);
        if ((label|0) == 46) {
         label = 0;
         $152 = ($val$0$i18|0)>(32767);
         if ($152) {
          $val$0 = 2147483647;
          break L27;
         } else {
          $val$0$i1823 = $val$0$i18;
         }
        }
        $153 = ($val$0$i1823|0)<(-32767);
        $154 = $val$0$i1823 << 16;
        $$$i$i = $153 ? -2147483647 : $154;
        $val$0 = $$$i$i;
        break;
       }
       case 3:  {
        $$val10 = HEAP32[$0>>2]|0;
        $$idx11$val = HEAP32[$$idx>>2]|0;
        $155 = HEAP8[$$val10>>0]|0;
        $156 = ($155<<24>>24)==(30);
        if ($156) {
         $157 = (_cff_parse_real($$val10,$$idx11$val,3,0)|0);
         $val$0 = $157;
         break L27;
        }
        $158 = ((($$val10)) + 1|0);
        $159 = $155&255;
        L72: do {
         switch ($155<<24>>24) {
         case 28:  {
          $160 = ((($$val10)) + 3|0);
          $161 = ($160>>>0)>($$idx11$val>>>0);
          if ($161) {
           $209 = 0;$val$0$i2160 = 0;
          } else {
           $162 = HEAP8[$158>>0]|0;
           $163 = $162&255;
           $164 = $163 << 8;
           $165 = ((($$val10)) + 2|0);
           $166 = HEAP8[$165>>0]|0;
           $167 = $166&255;
           $168 = $164 | $167;
           $sext$i20 = $168 << 16;
           $169 = $sext$i20 >> 16;
           $val$0$i21 = $169;
           label = 62;
          }
          break;
         }
         case 29:  {
          $170 = ((($$val10)) + 5|0);
          $171 = ($170>>>0)>($$idx11$val>>>0);
          if ($171) {
           $209 = 0;$val$0$i2160 = 0;
          } else {
           $172 = HEAP8[$158>>0]|0;
           $173 = $172&255;
           $174 = $173 << 24;
           $175 = ((($$val10)) + 2|0);
           $176 = HEAP8[$175>>0]|0;
           $177 = $176&255;
           $178 = $177 << 16;
           $179 = $178 | $174;
           $180 = ((($$val10)) + 3|0);
           $181 = HEAP8[$180>>0]|0;
           $182 = $181&255;
           $183 = $182 << 8;
           $184 = $179 | $183;
           $185 = ((($$val10)) + 4|0);
           $186 = HEAP8[$185>>0]|0;
           $187 = $186&255;
           $188 = $184 | $187;
           $val$0$i21 = $188;
           label = 62;
          }
          break;
         }
         default: {
          $189 = ($155&255)<(247);
          if ($189) {
           $190 = (($159) + -139)|0;
           $val$0$i21 = $190;
           label = 62;
           break L72;
          }
          $191 = ($155&255)<(251);
          $192 = ((($$val10)) + 2|0);
          $193 = ($192>>>0)>($$idx11$val>>>0);
          if ($191) {
           if ($193) {
            $209 = 0;$val$0$i2160 = 0;
            break L72;
           }
           $194 = $159 << 8;
           $195 = (($194) + -63232)|0;
           $196 = HEAP8[$158>>0]|0;
           $197 = $196&255;
           $198 = $197 | $195;
           $199 = (($198) + 108)|0;
           $val$0$i21 = $199;
           label = 62;
           break L72;
          } else {
           if ($193) {
            $209 = 0;$val$0$i2160 = 0;
            break L72;
           }
           $200 = (251 - ($159))|0;
           $201 = $200 << 8;
           $202 = HEAP8[$158>>0]|0;
           $203 = $202&255;
           $204 = (($201) + -108)|0;
           $205 = (($204) - ($203))|0;
           $val$0$i21 = $205;
           label = 62;
           break L72;
          }
         }
         }
        } while(0);
        if ((label|0) == 62) {
         label = 0;
         $206 = ($val$0$i21*1000)|0;
         $207 = ($206|0)>(32767);
         if ($207) {
          $val$0 = 2147483647;
          break L27;
         } else {
          $209 = $206;$val$0$i2160 = $val$0$i21;
         }
        }
        $208 = ($209|0)<(-32767);
        $210 = Math_imul($val$0$i2160, 65536000)|0;
        $$$i$i12 = $208 ? -2147483647 : $210;
        $val$0 = $$$i$i12;
        break;
       }
       case 6:  {
        $216 = ((($field$040$lcssa)) + 24|0);
        $217 = HEAP32[$216>>2]|0;
        $218 = (($43) + ($217)|0);
        $219 = ((($field$040$lcssa)) + 20|0);
        $220 = HEAP32[$219>>2]|0;
        $221 = ($31>>>0)>($220>>>0);
        $$ = $221 ? $220 : $31;
        $222 = $$&255;
        HEAP8[$218>>0] = $222;
        $223 = ($$|0)==(0);
        if ($223) {
         break L24;
        }
        $224 = ((($field$040$lcssa)) + 12|0);
        $data$044 = $0;$num_args$141 = $$;$q$043 = $46;$val$142 = 0;
        while(1) {
         $225 = ((($data$044)) + 4|0);
         $data$0$val = HEAP32[$data$044>>2]|0;
         $data$0$idx$val = HEAP32[$225>>2]|0;
         $226 = HEAP8[$data$0$val>>0]|0;
         $227 = ($226<<24>>24)==(30);
         L93: do {
          if ($227) {
           $228 = (_cff_parse_real($data$0$val,$data$0$idx$val,0,0)|0);
           $229 = $228 >> 16;
           $279 = $229;
          } else {
           $230 = ((($data$0$val)) + 1|0);
           $231 = $226&255;
           switch ($226<<24>>24) {
           case 28:  {
            $232 = ((($data$0$val)) + 3|0);
            $233 = ($232>>>0)>($data$0$idx$val>>>0);
            if ($233) {
             $279 = 0;
             break L93;
            }
            $234 = HEAP8[$230>>0]|0;
            $235 = $234&255;
            $236 = $235 << 8;
            $237 = ((($data$0$val)) + 2|0);
            $238 = HEAP8[$237>>0]|0;
            $239 = $238&255;
            $240 = $236 | $239;
            $sext$i14 = $240 << 16;
            $241 = $sext$i14 >> 16;
            $279 = $241;
            break L93;
            break;
           }
           case 29:  {
            $242 = ((($data$0$val)) + 5|0);
            $243 = ($242>>>0)>($data$0$idx$val>>>0);
            if ($243) {
             $279 = 0;
             break L93;
            }
            $244 = HEAP8[$230>>0]|0;
            $245 = $244&255;
            $246 = $245 << 24;
            $247 = ((($data$0$val)) + 2|0);
            $248 = HEAP8[$247>>0]|0;
            $249 = $248&255;
            $250 = $249 << 16;
            $251 = $250 | $246;
            $252 = ((($data$0$val)) + 3|0);
            $253 = HEAP8[$252>>0]|0;
            $254 = $253&255;
            $255 = $254 << 8;
            $256 = $251 | $255;
            $257 = ((($data$0$val)) + 4|0);
            $258 = HEAP8[$257>>0]|0;
            $259 = $258&255;
            $260 = $256 | $259;
            $279 = $260;
            break L93;
            break;
           }
           default: {
            $261 = ($226&255)<(247);
            if ($261) {
             $262 = (($231) + -139)|0;
             $279 = $262;
             break L93;
            }
            $263 = ($226&255)<(251);
            $264 = ((($data$0$val)) + 2|0);
            $265 = ($264>>>0)>($data$0$idx$val>>>0);
            if ($263) {
             if ($265) {
              $279 = 0;
              break L93;
             }
             $266 = $231 << 8;
             $267 = (($266) + -63232)|0;
             $268 = HEAP8[$230>>0]|0;
             $269 = $268&255;
             $270 = $269 | $267;
             $271 = (($270) + 108)|0;
             $279 = $271;
             break L93;
            } else {
             if ($265) {
              $279 = 0;
              break L93;
             }
             $272 = (251 - ($231))|0;
             $273 = $272 << 8;
             $274 = HEAP8[$230>>0]|0;
             $275 = $274&255;
             $276 = (($273) + -108)|0;
             $277 = (($276) - ($275))|0;
             $279 = $277;
             break L93;
            }
           }
           }
          }
         } while(0);
         $278 = (($279) + ($val$142))|0;
         $280 = HEAP8[$224>>0]|0;
         $281 = $280&255;
         switch ($281|0) {
         case 1:  {
          $282 = $278&255;
          HEAP8[$q$043>>0] = $282;
          $$pre = HEAP8[$224>>0]|0;
          $285 = $$pre;
          break;
         }
         case 2:  {
          $283 = $278&65535;
          HEAP16[$q$043>>1] = $283;
          $285 = $280;
          break;
         }
         case 4:  {
          HEAP32[$q$043>>2] = $278;
          $285 = $280;
          break;
         }
         default: {
          HEAP32[$q$043>>2] = $278;
          $285 = $280;
         }
         }
         $284 = $285&255;
         $286 = (($q$043) + ($284)|0);
         $287 = (($num_args$141) + -1)|0;
         $288 = ($287|0)==(0);
         if ($288) {
          break L24;
         } else {
          $data$044 = $225;$num_args$141 = $287;$q$043 = $286;$val$142 = $278;
         }
        }
        break;
       }
       default: {
        $289 = ((($field$040$lcssa)) + 16|0);
        $290 = HEAP32[$289>>2]|0;
        $291 = (FUNCTION_TABLE_ii[$290 & 8191]($parser)|0);
        $292 = ($291|0)==(0);
        if ($292) {
         break L24;
        } else {
         $$05 = $291;
         label = 95;
         break L4;
        }
       }
       }
      } while(0);
      $211 = ((($field$040$lcssa)) + 12|0);
      $212 = HEAP8[$211>>0]|0;
      $213 = $212&255;
      switch ($213|0) {
      case 1:  {
       $214 = $val$0&255;
       HEAP8[$46>>0] = $214;
       break L24;
       break;
      }
      case 2:  {
       $215 = $val$0&65535;
       HEAP16[$46>>1] = $215;
       break L24;
       break;
      }
      case 4:  {
       HEAP32[$46>>2] = $val$0;
       break L24;
       break;
      }
      default: {
       HEAP32[$46>>2] = $val$0;
       break L24;
      }
      }
     }
    } while(0);
    HEAP32[$1>>2] = $0;
    $p$4 = $p$2;
   }
  } while(0);
  $296 = ((($p$4)) + 1|0);
  $297 = ($296>>>0)<($limit>>>0);
  if ($297) {
   $p$045 = $296;
  } else {
   $$05 = 0;
   label = 95;
   break;
  }
 }
 if ((label|0) == 95) {
  return ($$05|0);
 }
 return (0)|0;
}
function _cff_parse_font_matrix($parser) {
 $parser = $parser|0;
 var $$$i = 0, $$$i$i = 0, $$$i$i15 = 0, $$$i$i18 = 0, $$$i$i21 = 0, $$$i$i24 = 0, $$02 = 0, $$1$i = 0, $$idx$val = 0, $$idx10$val = 0, $$idx12 = 0, $$idx12$val = 0, $$idx4$val = 0, $$idx6$val = 0, $$idx8$val = 0, $$off = 0, $$val = 0, $$val$pre = 0, $$val11 = 0, $$val3 = 0;
 var $$val3$pre = 0, $$val5 = 0, $$val5$pre = 0, $$val7 = 0, $$val7$pre = 0, $$val9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0;
 var $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0;
 var $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0;
 var $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0;
 var $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0;
 var $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0;
 var $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0;
 var $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0;
 var $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0;
 var $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $integer_length$0$lcssa$i = 0, $scaling = 0, $sext$i = 0, $sext$i$i = 0, $sext$i26 = 0, $sext$i29 = 0, $sext$i32 = 0, $sext$i35 = 0, $storemerge = 0, $val$0$i = 0, $val$0$i$i = 0, $val$0$i$i13 = 0, $val$0$i$i14 = 0, $val$0$i$i17 = 0, $val$0$i$i20 = 0, $val$0$i$i23 = 0, $val$0$i1$i = 0, $val$0$i27 = 0;
 var $val$0$i30 = 0, $val$0$i33 = 0, $val$0$i36 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $scaling = sp;
 $0 = ((($parser)) + 412|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 72|0);
 $3 = ((($1)) + 68|0);
 $4 = ((($parser)) + 404|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($parser)) + 40|0);
 $7 = ($5>>>0)<($6>>>0);
 if ($7) {
  $$02 = 161;
  STACKTOP = sp;return ($$02|0);
 }
 $8 = ((($parser)) + 16|0);
 $9 = ((($1)) + 64|0);
 HEAP8[$9>>0] = 1;
 $$val11 = HEAP32[$8>>2]|0;
 $$idx12 = ((($parser)) + 20|0);
 $$idx12$val = HEAP32[$$idx12>>2]|0;
 $10 = HEAP8[$$val11>>0]|0;
 $11 = ($10<<24>>24)==(30);
 do {
  if ($11) {
   $12 = (_cff_parse_real($$val11,$$idx12$val,0,$scaling)|0);
   $$1$i = $12;
  } else {
   $13 = ((($$val11)) + 1|0);
   $14 = $10&255;
   L7: do {
    switch ($10<<24>>24) {
    case 28:  {
     $15 = ((($$val11)) + 3|0);
     $16 = ($15>>>0)>($$idx12$val>>>0);
     if ($16) {
      $val$0$i1$i = 0;
     } else {
      $17 = HEAP8[$13>>0]|0;
      $18 = $17&255;
      $19 = $18 << 8;
      $20 = ((($$val11)) + 2|0);
      $21 = HEAP8[$20>>0]|0;
      $22 = $21&255;
      $23 = $19 | $22;
      $sext$i$i = $23 << 16;
      $24 = $sext$i$i >> 16;
      $val$0$i1$i = $24;
     }
     break;
    }
    case 29:  {
     $25 = ((($$val11)) + 5|0);
     $26 = ($25>>>0)>($$idx12$val>>>0);
     if ($26) {
      $val$0$i1$i = 0;
     } else {
      $27 = HEAP8[$13>>0]|0;
      $28 = $27&255;
      $29 = $28 << 24;
      $30 = ((($$val11)) + 2|0);
      $31 = HEAP8[$30>>0]|0;
      $32 = $31&255;
      $33 = $32 << 16;
      $34 = $33 | $29;
      $35 = ((($$val11)) + 3|0);
      $36 = HEAP8[$35>>0]|0;
      $37 = $36&255;
      $38 = $37 << 8;
      $39 = $34 | $38;
      $40 = ((($$val11)) + 4|0);
      $41 = HEAP8[$40>>0]|0;
      $42 = $41&255;
      $43 = $39 | $42;
      $val$0$i$i = $43;
      label = 16;
     }
     break;
    }
    default: {
     $44 = ($10&255)<(247);
     if ($44) {
      $45 = (($14) + -139)|0;
      $val$0$i1$i = $45;
      break L7;
     }
     $46 = ($10&255)<(251);
     $47 = ((($$val11)) + 2|0);
     $48 = ($47>>>0)>($$idx12$val>>>0);
     if ($46) {
      if ($48) {
       $val$0$i1$i = 0;
       break L7;
      }
      $49 = $14 << 8;
      $50 = (($49) + -63232)|0;
      $51 = HEAP8[$13>>0]|0;
      $52 = $51&255;
      $53 = $52 | $50;
      $54 = (($53) + 108)|0;
      $val$0$i$i = $54;
      label = 16;
      break L7;
     } else {
      if ($48) {
       $val$0$i1$i = 0;
       break L7;
      }
      $55 = (251 - ($14))|0;
      $56 = $55 << 8;
      $57 = HEAP8[$13>>0]|0;
      $58 = $57&255;
      $59 = (($56) + -108)|0;
      $60 = (($59) - ($58))|0;
      $val$0$i$i = $60;
      label = 16;
      break L7;
     }
    }
    }
   } while(0);
   if ((label|0) == 16) {
    $61 = ($val$0$i$i|0)>(32767);
    if ($61) {
     $62 = ($val$0$i$i|0)<(100000);
     if ($62) {
      $integer_length$0$lcssa$i = 5;
     } else {
      $63 = ($val$0$i$i|0)<(1000000);
      if ($63) {
       $integer_length$0$lcssa$i = 6;
      } else {
       $75 = ($val$0$i$i|0)<(10000000);
       if ($75) {
        $integer_length$0$lcssa$i = 7;
       } else {
        $76 = ($val$0$i$i|0)<(100000000);
        if ($76) {
         $integer_length$0$lcssa$i = 8;
        } else {
         $77 = ($val$0$i$i|0)<(1000000000);
         $$$i = $77 ? 9 : 10;
         $integer_length$0$lcssa$i = $$$i;
        }
       }
      }
     }
     $64 = (($integer_length$0$lcssa$i) + -5)|0;
     $65 = (51120 + ($64<<2)|0);
     $66 = HEAP32[$65>>2]|0;
     $67 = (($val$0$i$i|0) / ($66|0))&-1;
     $68 = ($67|0)>(32767);
     if ($68) {
      $69 = (($integer_length$0$lcssa$i) + -4)|0;
      HEAP32[$scaling>>2] = $69;
      $70 = (51120 + ($69<<2)|0);
      $71 = HEAP32[$70>>2]|0;
      $72 = (_FT_DivFix($val$0$i$i,$71)|0);
      $$1$i = $72;
      break;
     } else {
      HEAP32[$scaling>>2] = $64;
      $73 = (_FT_DivFix($val$0$i$i,$66)|0);
      $$1$i = $73;
      break;
     }
    } else {
     $val$0$i1$i = $val$0$i$i;
    }
   }
   HEAP32[$scaling>>2] = 0;
   $74 = $val$0$i1$i << 16;
   $$1$i = $74;
  }
 } while(0);
 $78 = ((($1)) + 48|0);
 HEAP32[$78>>2] = $$1$i;
 $79 = HEAP32[$scaling>>2]|0;
 $80 = (0 - ($79))|0;
 HEAP32[$scaling>>2] = $80;
 $$off = (($79) + 9)|0;
 $81 = ($$off>>>0)>(9);
 if ($81) {
  HEAP32[$78>>2] = 65536;
  $82 = ((($1)) + 56|0);
  HEAP32[$82>>2] = 0;
  $83 = ((($1)) + 52|0);
  HEAP32[$83>>2] = 0;
  $84 = ((($1)) + 60|0);
  HEAP32[$84>>2] = 65536;
  HEAP32[$2>>2] = 0;
  $85 = ((($1)) + 76|0);
  HEAP32[$85>>2] = 0;
  $storemerge = 1;
 } else {
  $86 = ((($parser)) + 24|0);
  $$val9 = HEAP32[$$idx12>>2]|0;
  $$idx10$val = HEAP32[$86>>2]|0;
  $87 = HEAP8[$$val9>>0]|0;
  $88 = ($87<<24>>24)==(30);
  if ($88) {
   $89 = (_cff_parse_real($$val9,$$idx10$val,$80,0)|0);
   $$val7$pre = HEAP32[$86>>2]|0;
   $$val7 = $$val7$pre;$146 = $89;
  } else {
   $90 = ((($$val9)) + 1|0);
   $91 = $87&255;
   L41: do {
    switch ($87<<24>>24) {
    case 28:  {
     $92 = ((($$val9)) + 3|0);
     $93 = ($92>>>0)>($$idx10$val>>>0);
     if ($93) {
      $val$0$i = 0;
     } else {
      $94 = HEAP8[$90>>0]|0;
      $95 = $94&255;
      $96 = $95 << 8;
      $97 = ((($$val9)) + 2|0);
      $98 = HEAP8[$97>>0]|0;
      $99 = $98&255;
      $100 = $96 | $99;
      $sext$i = $100 << 16;
      $101 = $sext$i >> 16;
      $val$0$i = $101;
     }
     break;
    }
    case 29:  {
     $102 = ((($$val9)) + 5|0);
     $103 = ($102>>>0)>($$idx10$val>>>0);
     if ($103) {
      $val$0$i = 0;
     } else {
      $104 = HEAP8[$90>>0]|0;
      $105 = $104&255;
      $106 = $105 << 24;
      $107 = ((($$val9)) + 2|0);
      $108 = HEAP8[$107>>0]|0;
      $109 = $108&255;
      $110 = $109 << 16;
      $111 = $110 | $106;
      $112 = ((($$val9)) + 3|0);
      $113 = HEAP8[$112>>0]|0;
      $114 = $113&255;
      $115 = $114 << 8;
      $116 = $111 | $115;
      $117 = ((($$val9)) + 4|0);
      $118 = HEAP8[$117>>0]|0;
      $119 = $118&255;
      $120 = $116 | $119;
      $val$0$i = $120;
     }
     break;
    }
    default: {
     $121 = ($87&255)<(247);
     if ($121) {
      $122 = (($91) + -139)|0;
      $val$0$i = $122;
      break L41;
     }
     $123 = ($87&255)<(251);
     $124 = ((($$val9)) + 2|0);
     $125 = ($124>>>0)>($$idx10$val>>>0);
     if ($123) {
      if ($125) {
       $val$0$i = 0;
       break L41;
      }
      $126 = $91 << 8;
      $127 = (($126) + -63232)|0;
      $128 = HEAP8[$90>>0]|0;
      $129 = $128&255;
      $130 = $129 | $127;
      $131 = (($130) + 108)|0;
      $val$0$i = $131;
      break L41;
     } else {
      if ($125) {
       $val$0$i = 0;
       break L41;
      }
      $132 = (251 - ($91))|0;
      $133 = $132 << 8;
      $134 = HEAP8[$90>>0]|0;
      $135 = $134&255;
      $136 = (($133) + -108)|0;
      $137 = (($136) - ($135))|0;
      $val$0$i = $137;
      break L41;
     }
    }
    }
   } while(0);
   $138 = ($79|0)==(0);
   if ($138) {
    $val$0$i$i13 = $val$0$i;
   } else {
    $139 = (51120 + ($80<<2)|0);
    $140 = HEAP32[$139>>2]|0;
    $141 = Math_imul($140, $val$0$i)|0;
    $val$0$i$i13 = $141;
   }
   $142 = ($val$0$i$i13|0)>(32767);
   if ($142) {
    $$val7 = $$idx10$val;$146 = 2147483647;
   } else {
    $143 = ($val$0$i$i13|0)<(-32767);
    $144 = $val$0$i$i13 << 16;
    $$$i$i = $143 ? -2147483647 : $144;
    $$val7 = $$idx10$val;$146 = $$$i$i;
   }
  }
  $145 = ((($1)) + 56|0);
  HEAP32[$145>>2] = $146;
  $147 = ((($parser)) + 28|0);
  $148 = HEAP32[$scaling>>2]|0;
  $$idx8$val = HEAP32[$147>>2]|0;
  $149 = HEAP8[$$val7>>0]|0;
  $150 = ($149<<24>>24)==(30);
  if ($150) {
   $151 = (_cff_parse_real($$val7,$$idx8$val,$148,0)|0);
   $$val5$pre = HEAP32[$147>>2]|0;
   $$val5 = $$val5$pre;$208 = $151;
  } else {
   $152 = ((($$val7)) + 1|0);
   $153 = $149&255;
   L64: do {
    switch ($149<<24>>24) {
    case 28:  {
     $154 = ((($$val7)) + 3|0);
     $155 = ($154>>>0)>($$idx8$val>>>0);
     if ($155) {
      $val$0$i36 = 0;
     } else {
      $156 = HEAP8[$152>>0]|0;
      $157 = $156&255;
      $158 = $157 << 8;
      $159 = ((($$val7)) + 2|0);
      $160 = HEAP8[$159>>0]|0;
      $161 = $160&255;
      $162 = $158 | $161;
      $sext$i35 = $162 << 16;
      $163 = $sext$i35 >> 16;
      $val$0$i36 = $163;
     }
     break;
    }
    case 29:  {
     $164 = ((($$val7)) + 5|0);
     $165 = ($164>>>0)>($$idx8$val>>>0);
     if ($165) {
      $val$0$i36 = 0;
     } else {
      $166 = HEAP8[$152>>0]|0;
      $167 = $166&255;
      $168 = $167 << 24;
      $169 = ((($$val7)) + 2|0);
      $170 = HEAP8[$169>>0]|0;
      $171 = $170&255;
      $172 = $171 << 16;
      $173 = $172 | $168;
      $174 = ((($$val7)) + 3|0);
      $175 = HEAP8[$174>>0]|0;
      $176 = $175&255;
      $177 = $176 << 8;
      $178 = $173 | $177;
      $179 = ((($$val7)) + 4|0);
      $180 = HEAP8[$179>>0]|0;
      $181 = $180&255;
      $182 = $178 | $181;
      $val$0$i36 = $182;
     }
     break;
    }
    default: {
     $183 = ($149&255)<(247);
     if ($183) {
      $184 = (($153) + -139)|0;
      $val$0$i36 = $184;
      break L64;
     }
     $185 = ($149&255)<(251);
     $186 = ((($$val7)) + 2|0);
     $187 = ($186>>>0)>($$idx8$val>>>0);
     if ($185) {
      if ($187) {
       $val$0$i36 = 0;
       break L64;
      }
      $188 = $153 << 8;
      $189 = (($188) + -63232)|0;
      $190 = HEAP8[$152>>0]|0;
      $191 = $190&255;
      $192 = $191 | $189;
      $193 = (($192) + 108)|0;
      $val$0$i36 = $193;
      break L64;
     } else {
      if ($187) {
       $val$0$i36 = 0;
       break L64;
      }
      $194 = (251 - ($153))|0;
      $195 = $194 << 8;
      $196 = HEAP8[$152>>0]|0;
      $197 = $196&255;
      $198 = (($195) + -108)|0;
      $199 = (($198) - ($197))|0;
      $val$0$i36 = $199;
      break L64;
     }
    }
    }
   } while(0);
   $200 = ($148|0)==(0);
   if ($200) {
    $val$0$i$i14 = $val$0$i36;
   } else {
    $201 = (51120 + ($148<<2)|0);
    $202 = HEAP32[$201>>2]|0;
    $203 = Math_imul($202, $val$0$i36)|0;
    $val$0$i$i14 = $203;
   }
   $204 = ($val$0$i$i14|0)>(32767);
   if ($204) {
    $$val5 = $$idx8$val;$208 = 2147483647;
   } else {
    $205 = ($val$0$i$i14|0)<(-32767);
    $206 = $val$0$i$i14 << 16;
    $$$i$i15 = $205 ? -2147483647 : $206;
    $$val5 = $$idx8$val;$208 = $$$i$i15;
   }
  }
  $207 = ((($1)) + 52|0);
  HEAP32[$207>>2] = $208;
  $209 = ((($parser)) + 32|0);
  $210 = HEAP32[$scaling>>2]|0;
  $$idx6$val = HEAP32[$209>>2]|0;
  $211 = HEAP8[$$val5>>0]|0;
  $212 = ($211<<24>>24)==(30);
  if ($212) {
   $213 = (_cff_parse_real($$val5,$$idx6$val,$210,0)|0);
   $$val3$pre = HEAP32[$209>>2]|0;
   $$val3 = $$val3$pre;$270 = $213;
  } else {
   $214 = ((($$val5)) + 1|0);
   $215 = $211&255;
   L87: do {
    switch ($211<<24>>24) {
    case 28:  {
     $216 = ((($$val5)) + 3|0);
     $217 = ($216>>>0)>($$idx6$val>>>0);
     if ($217) {
      $val$0$i33 = 0;
     } else {
      $218 = HEAP8[$214>>0]|0;
      $219 = $218&255;
      $220 = $219 << 8;
      $221 = ((($$val5)) + 2|0);
      $222 = HEAP8[$221>>0]|0;
      $223 = $222&255;
      $224 = $220 | $223;
      $sext$i32 = $224 << 16;
      $225 = $sext$i32 >> 16;
      $val$0$i33 = $225;
     }
     break;
    }
    case 29:  {
     $226 = ((($$val5)) + 5|0);
     $227 = ($226>>>0)>($$idx6$val>>>0);
     if ($227) {
      $val$0$i33 = 0;
     } else {
      $228 = HEAP8[$214>>0]|0;
      $229 = $228&255;
      $230 = $229 << 24;
      $231 = ((($$val5)) + 2|0);
      $232 = HEAP8[$231>>0]|0;
      $233 = $232&255;
      $234 = $233 << 16;
      $235 = $234 | $230;
      $236 = ((($$val5)) + 3|0);
      $237 = HEAP8[$236>>0]|0;
      $238 = $237&255;
      $239 = $238 << 8;
      $240 = $235 | $239;
      $241 = ((($$val5)) + 4|0);
      $242 = HEAP8[$241>>0]|0;
      $243 = $242&255;
      $244 = $240 | $243;
      $val$0$i33 = $244;
     }
     break;
    }
    default: {
     $245 = ($211&255)<(247);
     if ($245) {
      $246 = (($215) + -139)|0;
      $val$0$i33 = $246;
      break L87;
     }
     $247 = ($211&255)<(251);
     $248 = ((($$val5)) + 2|0);
     $249 = ($248>>>0)>($$idx6$val>>>0);
     if ($247) {
      if ($249) {
       $val$0$i33 = 0;
       break L87;
      }
      $250 = $215 << 8;
      $251 = (($250) + -63232)|0;
      $252 = HEAP8[$214>>0]|0;
      $253 = $252&255;
      $254 = $253 | $251;
      $255 = (($254) + 108)|0;
      $val$0$i33 = $255;
      break L87;
     } else {
      if ($249) {
       $val$0$i33 = 0;
       break L87;
      }
      $256 = (251 - ($215))|0;
      $257 = $256 << 8;
      $258 = HEAP8[$214>>0]|0;
      $259 = $258&255;
      $260 = (($257) + -108)|0;
      $261 = (($260) - ($259))|0;
      $val$0$i33 = $261;
      break L87;
     }
    }
    }
   } while(0);
   $262 = ($210|0)==(0);
   if ($262) {
    $val$0$i$i17 = $val$0$i33;
   } else {
    $263 = (51120 + ($210<<2)|0);
    $264 = HEAP32[$263>>2]|0;
    $265 = Math_imul($264, $val$0$i33)|0;
    $val$0$i$i17 = $265;
   }
   $266 = ($val$0$i$i17|0)>(32767);
   if ($266) {
    $$val3 = $$idx6$val;$270 = 2147483647;
   } else {
    $267 = ($val$0$i$i17|0)<(-32767);
    $268 = $val$0$i$i17 << 16;
    $$$i$i18 = $267 ? -2147483647 : $268;
    $$val3 = $$idx6$val;$270 = $$$i$i18;
   }
  }
  $269 = ((($1)) + 60|0);
  HEAP32[$269>>2] = $270;
  $271 = ((($parser)) + 36|0);
  $272 = HEAP32[$scaling>>2]|0;
  $$idx4$val = HEAP32[$271>>2]|0;
  $273 = HEAP8[$$val3>>0]|0;
  $274 = ($273<<24>>24)==(30);
  if ($274) {
   $275 = (_cff_parse_real($$val3,$$idx4$val,$272,0)|0);
   $$val$pre = HEAP32[$271>>2]|0;
   $$val = $$val$pre;$331 = $275;
  } else {
   $276 = ((($$val3)) + 1|0);
   $277 = $273&255;
   L110: do {
    switch ($273<<24>>24) {
    case 28:  {
     $278 = ((($$val3)) + 3|0);
     $279 = ($278>>>0)>($$idx4$val>>>0);
     if ($279) {
      $val$0$i30 = 0;
     } else {
      $280 = HEAP8[$276>>0]|0;
      $281 = $280&255;
      $282 = $281 << 8;
      $283 = ((($$val3)) + 2|0);
      $284 = HEAP8[$283>>0]|0;
      $285 = $284&255;
      $286 = $282 | $285;
      $sext$i29 = $286 << 16;
      $287 = $sext$i29 >> 16;
      $val$0$i30 = $287;
     }
     break;
    }
    case 29:  {
     $288 = ((($$val3)) + 5|0);
     $289 = ($288>>>0)>($$idx4$val>>>0);
     if ($289) {
      $val$0$i30 = 0;
     } else {
      $290 = HEAP8[$276>>0]|0;
      $291 = $290&255;
      $292 = $291 << 24;
      $293 = ((($$val3)) + 2|0);
      $294 = HEAP8[$293>>0]|0;
      $295 = $294&255;
      $296 = $295 << 16;
      $297 = $296 | $292;
      $298 = ((($$val3)) + 3|0);
      $299 = HEAP8[$298>>0]|0;
      $300 = $299&255;
      $301 = $300 << 8;
      $302 = $297 | $301;
      $303 = ((($$val3)) + 4|0);
      $304 = HEAP8[$303>>0]|0;
      $305 = $304&255;
      $306 = $302 | $305;
      $val$0$i30 = $306;
     }
     break;
    }
    default: {
     $307 = ($273&255)<(247);
     if ($307) {
      $308 = (($277) + -139)|0;
      $val$0$i30 = $308;
      break L110;
     }
     $309 = ($273&255)<(251);
     $310 = ((($$val3)) + 2|0);
     $311 = ($310>>>0)>($$idx4$val>>>0);
     if ($309) {
      if ($311) {
       $val$0$i30 = 0;
       break L110;
      }
      $312 = $277 << 8;
      $313 = (($312) + -63232)|0;
      $314 = HEAP8[$276>>0]|0;
      $315 = $314&255;
      $316 = $315 | $313;
      $317 = (($316) + 108)|0;
      $val$0$i30 = $317;
      break L110;
     } else {
      if ($311) {
       $val$0$i30 = 0;
       break L110;
      }
      $318 = (251 - ($277))|0;
      $319 = $318 << 8;
      $320 = HEAP8[$276>>0]|0;
      $321 = $320&255;
      $322 = (($319) + -108)|0;
      $323 = (($322) - ($321))|0;
      $val$0$i30 = $323;
      break L110;
     }
    }
    }
   } while(0);
   $324 = ($272|0)==(0);
   if ($324) {
    $val$0$i$i20 = $val$0$i30;
   } else {
    $325 = (51120 + ($272<<2)|0);
    $326 = HEAP32[$325>>2]|0;
    $327 = Math_imul($326, $val$0$i30)|0;
    $val$0$i$i20 = $327;
   }
   $328 = ($val$0$i$i20|0)>(32767);
   if ($328) {
    $$val = $$idx4$val;$331 = 2147483647;
   } else {
    $329 = ($val$0$i$i20|0)<(-32767);
    $330 = $val$0$i$i20 << 16;
    $$$i$i21 = $329 ? -2147483647 : $330;
    $$val = $$idx4$val;$331 = $$$i$i21;
   }
  }
  HEAP32[$2>>2] = $331;
  $332 = HEAP32[$scaling>>2]|0;
  $$idx$val = HEAP32[$6>>2]|0;
  $333 = HEAP8[$$val>>0]|0;
  $334 = ($333<<24>>24)==(30);
  if ($334) {
   $335 = (_cff_parse_real($$val,$$idx$val,$332,0)|0);
   $392 = $335;
  } else {
   $336 = ((($$val)) + 1|0);
   $337 = $333&255;
   L133: do {
    switch ($333<<24>>24) {
    case 28:  {
     $338 = ((($$val)) + 3|0);
     $339 = ($338>>>0)>($$idx$val>>>0);
     if ($339) {
      $val$0$i27 = 0;
     } else {
      $340 = HEAP8[$336>>0]|0;
      $341 = $340&255;
      $342 = $341 << 8;
      $343 = ((($$val)) + 2|0);
      $344 = HEAP8[$343>>0]|0;
      $345 = $344&255;
      $346 = $342 | $345;
      $sext$i26 = $346 << 16;
      $347 = $sext$i26 >> 16;
      $val$0$i27 = $347;
     }
     break;
    }
    case 29:  {
     $348 = ((($$val)) + 5|0);
     $349 = ($348>>>0)>($$idx$val>>>0);
     if ($349) {
      $val$0$i27 = 0;
     } else {
      $350 = HEAP8[$336>>0]|0;
      $351 = $350&255;
      $352 = $351 << 24;
      $353 = ((($$val)) + 2|0);
      $354 = HEAP8[$353>>0]|0;
      $355 = $354&255;
      $356 = $355 << 16;
      $357 = $356 | $352;
      $358 = ((($$val)) + 3|0);
      $359 = HEAP8[$358>>0]|0;
      $360 = $359&255;
      $361 = $360 << 8;
      $362 = $357 | $361;
      $363 = ((($$val)) + 4|0);
      $364 = HEAP8[$363>>0]|0;
      $365 = $364&255;
      $366 = $362 | $365;
      $val$0$i27 = $366;
     }
     break;
    }
    default: {
     $367 = ($333&255)<(247);
     if ($367) {
      $368 = (($337) + -139)|0;
      $val$0$i27 = $368;
      break L133;
     }
     $369 = ($333&255)<(251);
     $370 = ((($$val)) + 2|0);
     $371 = ($370>>>0)>($$idx$val>>>0);
     if ($369) {
      if ($371) {
       $val$0$i27 = 0;
       break L133;
      }
      $372 = $337 << 8;
      $373 = (($372) + -63232)|0;
      $374 = HEAP8[$336>>0]|0;
      $375 = $374&255;
      $376 = $375 | $373;
      $377 = (($376) + 108)|0;
      $val$0$i27 = $377;
      break L133;
     } else {
      if ($371) {
       $val$0$i27 = 0;
       break L133;
      }
      $378 = (251 - ($337))|0;
      $379 = $378 << 8;
      $380 = HEAP8[$336>>0]|0;
      $381 = $380&255;
      $382 = (($379) + -108)|0;
      $383 = (($382) - ($381))|0;
      $val$0$i27 = $383;
      break L133;
     }
    }
    }
   } while(0);
   $384 = ($332|0)==(0);
   if ($384) {
    $val$0$i$i23 = $val$0$i27;
   } else {
    $385 = (51120 + ($332<<2)|0);
    $386 = HEAP32[$385>>2]|0;
    $387 = Math_imul($386, $val$0$i27)|0;
    $val$0$i$i23 = $387;
   }
   $388 = ($val$0$i$i23|0)>(32767);
   if ($388) {
    $392 = 2147483647;
   } else {
    $389 = ($val$0$i$i23|0)<(-32767);
    $390 = $val$0$i$i23 << 16;
    $$$i$i24 = $389 ? -2147483647 : $390;
    $392 = $$$i$i24;
   }
  }
  $391 = ((($1)) + 76|0);
  HEAP32[$391>>2] = $392;
  $393 = HEAP32[$scaling>>2]|0;
  $394 = (51120 + ($393<<2)|0);
  $395 = HEAP32[$394>>2]|0;
  $storemerge = $395;
 }
 HEAP32[$3>>2] = $storemerge;
 $$02 = 0;
 STACKTOP = sp;return ($$02|0);
}
function _cff_parse_real($start,$limit,$power_ten,$scaling) {
 $start = $start|0;
 $limit = $limit|0;
 $power_ten = $power_ten|0;
 $scaling = $scaling|0;
 var $$6 = 0, $$lcssa = 0, $$lcssa314 = 0, $$lcssa449 = 0, $$lcssa451 = 0, $$lcssa453 = 0, $$lcssa458 = 0, $$lcssa463 = 0, $$lcssa466 = 0, $$lcssa468 = 0, $$lcssa492 = 0, $$lcssa495 = 0, $$lcssa495$lcssa = 0, $$lcssa502 = 0, $$lcssa503 = 0, $$lcssa53 = 0, $$lcssa57 = 0, $$pre = 0, $$pre191 = 0, $$pre193 = 0;
 var $$pre195 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exponent$0$ = 0, $exponent$0$ph294 = 0, $exponent$0$ph363 = 0, $exponent$1 = 0;
 var $exponent$3 = 0, $exponent_add$0$ph24$us = 0, $exponent_add$0$ph29$lcssa61 = 0, $exponent_add$0$ph29$us$us = 0, $exponent_add$1$ph$lcssa = 0, $exponent_add$1$ph15 = 0, $exponent_add$1$ph15$us = 0, $exponent_add$1$ph15$us$lcssa461 = 0, $exponent_add$1$ph367 = 0, $exponent_add$2 = 0, $fraction_length$0$ph$lcssa = 0, $fraction_length$0$ph368 = 0, $fraction_length$1 = 0, $fraction_length$2 = 0, $fraction_length$38 = 0, $have_overflow$0$lcssa32 = 0, $have_overflow$0$us = 0, $have_overflow$1 = 0, $integer_length$0$ph359 = 0, $integer_length$0$ph378 = 0;
 var $nib$0 = 0, $nib$0$off = 0, $number$0$ph354 = 0, $number$0$ph376 = 0, $number$1$ph$lcssa = 0, $number$1$ph366 = 0, $number$2 = 0, $number$4 = 0, $number$5 = 0, $number$67 = 0, $or$cond5 = 0, $p$0 = 0, $p$0$ph26 = 0, $p$0$ph26$us = 0, $p$0$ph31$us$us = 0, $p$0$ph379 = 0, $p$1 = 0, $p$1$lcssa486 = 0, $p$1$lcssa486$lcssa = 0, $p$1$lcssa49 = 0;
 var $p$1$us$us$us = 0, $p$2 = 0, $p$2$ph$lcssa = 0, $p$2$ph17 = 0, $p$2$ph17$us = 0, $p$2$ph369 = 0, $p$3 = 0, $p$3$lcssa454 = 0, $p$3$us$us = 0, $p$3$us$us$lcssa464 = 0, $p$4 = 0, $p$5$ph364 = 0, $p$5$us = 0, $p$6 = 0, $p$6$lcssa450 = 0, $p$6$us = 0, $phase$0 = 0, $phase$0$ph22 = 0, $phase$0$ph22$us = 0, $phase$0$ph28$us$us = 0;
 var $phase$0$ph375 = 0, $phase$1 = 0, $phase$1$ph$lcssa = 0, $phase$1$ph14 = 0, $phase$1$ph14$us = 0, $phase$1$ph365 = 0, $phase$2 = 0, $phase$3$ph362 = 0, $phase$3$us = 0, $result$0 = 0, $result$0$ = 0, $sign$0$ph23 = 0, $sign$0$ph23$lcssa497 = 0, $sign$0$ph23$lcssa497$lcssa = 0, $sign$0$ph23$lcssa63 = 0, $sign$0$ph23$us = 0, $sign$0$ph2365 = 0, $sign$0$ph377 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($scaling|0)!=(0|0);
 if ($0) {
  HEAP32[$scaling>>2] = 0;
  $integer_length$0$ph378 = 0;$number$0$ph376 = 0;$p$0$ph379 = $start;$phase$0$ph375 = 4;$sign$0$ph377 = 0;
 } else {
  $integer_length$0$ph378 = 0;$number$0$ph376 = 0;$p$0$ph379 = $start;$phase$0$ph375 = 4;$sign$0$ph377 = 0;
 }
 L3: while(1) {
  $p$0$ph26 = $p$0$ph379;$phase$0$ph22 = $phase$0$ph375;$sign$0$ph23 = $sign$0$ph377;
  L5: while(1) {
   $p$0 = $p$0$ph26;$phase$0 = $phase$0$ph22;
   while(1) {
    $12 = ($phase$0|0)==(0);
    if ($12) {
     $p$1 = $p$0;
    } else {
     $13 = ((($p$0)) + 1|0);
     $14 = ($13>>>0)<($limit>>>0);
     if ($14) {
      $p$1 = $13;
     } else {
      $result$0 = 0;$sign$0$ph2365 = $sign$0$ph23;
      break L3;
     }
    }
    $15 = HEAP8[$p$1>>0]|0;
    $16 = $15&255;
    $17 = $16 >>> $phase$0;
    $18 = $17 & 15;
    $19 = (4 - ($phase$0))|0;
    $20 = ($18|0)==(14);
    if ($20) {
     $p$0$ph26 = $p$1;$phase$0$ph22 = $19;$sign$0$ph23 = 1;
     continue L5;
    }
    $21 = ($18>>>0)>(9);
    if ($21) {
     $$lcssa53 = $18;$$lcssa57 = $19;$147 = $15;$exponent_add$0$ph29$lcssa61 = 0;$integer_length$0$ph359 = $integer_length$0$ph378;$number$0$ph354 = $number$0$ph376;$p$1$lcssa49 = $p$1;$sign$0$ph23$lcssa63 = $sign$0$ph23;
     label = 18;
     break L3;
    }
    $22 = $18 | $number$0$ph376;
    $23 = ($22|0)==(0);
    if ($23) {
     $p$0 = $p$1;$phase$0 = $19;
    } else {
     $$lcssa492 = $18;$$lcssa495 = $19;$p$1$lcssa486 = $p$1;$sign$0$ph23$lcssa497 = $sign$0$ph23;
     break L5;
    }
   }
  }
  $24 = (($integer_length$0$ph378) + 1)|0;
  $25 = ($number$0$ph376*10)|0;
  $26 = (($$lcssa492) + ($25))|0;
  $27 = ($26|0)>(214748363);
  if ($27) {
   $$lcssa495$lcssa = $$lcssa495;$$lcssa502 = $24;$$lcssa503 = $26;$p$1$lcssa486$lcssa = $p$1$lcssa486;$sign$0$ph23$lcssa497$lcssa = $sign$0$ph23$lcssa497;
   label = 17;
   break;
  } else {
   $integer_length$0$ph378 = $24;$number$0$ph376 = $26;$p$0$ph379 = $p$1$lcssa486;$phase$0$ph375 = $$lcssa495;$sign$0$ph377 = $sign$0$ph23$lcssa497;
  }
 }
 L15: do {
  if ((label|0) == 17) {
   $exponent_add$0$ph24$us = 0;$p$0$ph26$us = $p$1$lcssa486$lcssa;$phase$0$ph22$us = $$lcssa495$lcssa;$sign$0$ph23$us = $sign$0$ph23$lcssa497$lcssa;
   L17: while(1) {
    $exponent_add$0$ph29$us$us = $exponent_add$0$ph24$us;$p$0$ph31$us$us = $p$0$ph26$us;$phase$0$ph28$us$us = $phase$0$ph22$us;
    while(1) {
     $1 = ($phase$0$ph28$us$us|0)==(0);
     if ($1) {
      $p$1$us$us$us = $p$0$ph31$us$us;
     } else {
      $2 = ((($p$0$ph31$us$us)) + 1|0);
      $3 = ($2>>>0)<($limit>>>0);
      if ($3) {
       $p$1$us$us$us = $2;
      } else {
       $result$0 = 0;$sign$0$ph2365 = $sign$0$ph23$us;
       break L15;
      }
     }
     $4 = HEAP8[$p$1$us$us$us>>0]|0;
     $5 = $4&255;
     $6 = $5 >>> $phase$0$ph28$us$us;
     $7 = $6 & 15;
     $8 = (4 - ($phase$0$ph28$us$us))|0;
     $9 = ($7|0)==(14);
     if ($9) {
      $exponent_add$0$ph24$us = $exponent_add$0$ph29$us$us;$p$0$ph26$us = $p$1$us$us$us;$phase$0$ph22$us = $8;$sign$0$ph23$us = 1;
      continue L17;
     }
     $10 = ($7>>>0)>(9);
     if ($10) {
      $$lcssa53 = $7;$$lcssa57 = $8;$147 = $4;$exponent_add$0$ph29$lcssa61 = $exponent_add$0$ph29$us$us;$integer_length$0$ph359 = $$lcssa502;$number$0$ph354 = $$lcssa503;$p$1$lcssa49 = $p$1$us$us$us;$sign$0$ph23$lcssa63 = $sign$0$ph23$us;
      label = 18;
      break L15;
     }
     $11 = (($exponent_add$0$ph29$us$us) + 1)|0;
     $exponent_add$0$ph29$us$us = $11;$p$0$ph31$us$us = $p$1$us$us$us;$phase$0$ph28$us$us = $8;
    }
   }
  }
 } while(0);
 L26: do {
  if ((label|0) == 18) {
   $28 = ($$lcssa53|0)==(10);
   L28: do {
    if ($28) {
     $29 = ($number$0$ph354|0)<(214748364);
     if ($29) {
      $149 = $147;$exponent_add$1$ph367 = $exponent_add$0$ph29$lcssa61;$fraction_length$0$ph368 = 0;$number$1$ph366 = $number$0$ph354;$p$2$ph369 = $p$1$lcssa49;$phase$1$ph365 = $$lcssa57;
      while(1) {
       $150 = $149;$exponent_add$1$ph15$us = $exponent_add$1$ph367;$p$2$ph17$us = $p$2$ph369;$phase$1$ph14$us = $phase$1$ph365;
       while(1) {
        $40 = ($phase$1$ph14$us|0)==(0);
        if ($40) {
         $33 = $150;$p$3$us$us = $p$2$ph17$us;
        } else {
         $30 = ((($p$2$ph17$us)) + 1|0);
         $31 = ($30>>>0)<($limit>>>0);
         if (!($31)) {
          $result$0 = 0;$sign$0$ph2365 = $sign$0$ph23$lcssa63;
          break L26;
         }
         $$pre191 = HEAP8[$30>>0]|0;
         $33 = $$pre191;$p$3$us$us = $30;
        }
        $32 = $33&255;
        $34 = $32 >>> $phase$1$ph14$us;
        $35 = $34 & 15;
        $36 = (4 - ($phase$1$ph14$us))|0;
        $37 = ($35>>>0)>(9);
        if ($37) {
         $148 = $33;$exponent_add$2 = $exponent_add$1$ph15$us;$fraction_length$1 = $fraction_length$0$ph368;$nib$0 = $35;$number$2 = $number$1$ph366;$p$4 = $p$3$us$us;$phase$2 = $36;
         break L28;
        }
        $38 = $35 | $number$1$ph366;
        $39 = ($38|0)==(0);
        if (!($39)) {
         $$lcssa463 = $33;$$lcssa466 = $35;$$lcssa468 = $36;$exponent_add$1$ph15$us$lcssa461 = $exponent_add$1$ph15$us;$p$3$us$us$lcssa464 = $p$3$us$us;
         break;
        }
        $41 = (($exponent_add$1$ph15$us) + -1)|0;
        $150 = $33;$exponent_add$1$ph15$us = $41;$p$2$ph17$us = $p$3$us$us;$phase$1$ph14$us = $36;
       }
       $54 = (($fraction_length$0$ph368) + 1)|0;
       $55 = ($number$1$ph366*10)|0;
       $56 = (($$lcssa466) + ($55))|0;
       $57 = ($56|0)<(214748364);
       $58 = ($54|0)<(9);
       $or$cond5 = $57 & $58;
       if ($or$cond5) {
        $149 = $$lcssa463;$exponent_add$1$ph367 = $exponent_add$1$ph15$us$lcssa461;$fraction_length$0$ph368 = $54;$number$1$ph366 = $56;$p$2$ph369 = $p$3$us$us$lcssa464;$phase$1$ph365 = $$lcssa468;
       } else {
        $$lcssa314 = $$lcssa463;$exponent_add$1$ph$lcssa = $exponent_add$1$ph15$us$lcssa461;$fraction_length$0$ph$lcssa = $54;$number$1$ph$lcssa = $56;$p$2$ph$lcssa = $p$3$us$us$lcssa464;$phase$1$ph$lcssa = $$lcssa468;
        break;
       }
      }
     } else {
      $$lcssa314 = $147;$exponent_add$1$ph$lcssa = $exponent_add$0$ph29$lcssa61;$fraction_length$0$ph$lcssa = 0;$number$1$ph$lcssa = $number$0$ph354;$p$2$ph$lcssa = $p$1$lcssa49;$phase$1$ph$lcssa = $$lcssa57;
     }
     $151 = $$lcssa314;$exponent_add$1$ph15 = $exponent_add$1$ph$lcssa;$p$2$ph17 = $p$2$ph$lcssa;$phase$1$ph14 = $phase$1$ph$lcssa;
     while(1) {
      $152 = $151;$p$2 = $p$2$ph17;$phase$1 = $phase$1$ph14;
      while(1) {
       $42 = ($phase$1|0)==(0);
       if ($42) {
        $46 = $152;$p$3 = $p$2;
       } else {
        $43 = ((($p$2)) + 1|0);
        $44 = ($43>>>0)<($limit>>>0);
        if (!($44)) {
         $result$0 = 0;$sign$0$ph2365 = $sign$0$ph23$lcssa63;
         break L26;
        }
        $$pre = HEAP8[$43>>0]|0;
        $46 = $$pre;$p$3 = $43;
       }
       $45 = $46&255;
       $47 = $45 >>> $phase$1;
       $48 = $47 & 15;
       $49 = (4 - ($phase$1))|0;
       $50 = ($48>>>0)>(9);
       if ($50) {
        $148 = $46;$exponent_add$2 = $exponent_add$1$ph15;$fraction_length$1 = $fraction_length$0$ph$lcssa;$nib$0 = $48;$number$2 = $number$1$ph$lcssa;$p$4 = $p$3;$phase$2 = $49;
        break L28;
       }
       $51 = $48 | $number$1$ph$lcssa;
       $52 = ($51|0)==(0);
       if ($52) {
        $$lcssa453 = $46;$$lcssa458 = $49;$p$3$lcssa454 = $p$3;
        break;
       } else {
        $152 = $46;$p$2 = $p$3;$phase$1 = $49;
       }
      }
      $53 = (($exponent_add$1$ph15) + -1)|0;
      $151 = $$lcssa453;$exponent_add$1$ph15 = $53;$p$2$ph17 = $p$3$lcssa454;$phase$1$ph14 = $$lcssa458;
     }
    } else {
     $148 = $147;$exponent_add$2 = $exponent_add$0$ph29$lcssa61;$fraction_length$1 = 0;$nib$0 = $$lcssa53;$number$2 = $number$0$ph354;$p$4 = $p$1$lcssa49;$phase$2 = $$lcssa57;
    }
   } while(0);
   $59 = ($nib$0|0)==(12);
   $nib$0$off = (($nib$0) + -11)|0;
   $60 = ($nib$0$off>>>0)<(2);
   if ($60) {
    $153 = $148;$exponent$0$ph363 = 0;$p$5$ph364 = $p$4;$phase$3$ph362 = $phase$2;
    while(1) {
     $70 = ($phase$3$ph362|0)==(0);
     if ($70) {
      $74 = $153;$p$6 = $p$5$ph364;
     } else {
      $71 = ((($p$5$ph364)) + 1|0);
      $72 = ($71>>>0)<($limit>>>0);
      if (!($72)) {
       $result$0 = 0;$sign$0$ph2365 = $sign$0$ph23$lcssa63;
       break L26;
      }
      $$pre193 = HEAP8[$71>>0]|0;
      $74 = $$pre193;$p$6 = $71;
     }
     $73 = $74&255;
     $75 = $73 >>> $phase$3$ph362;
     $76 = $75 & 15;
     $77 = ($76>>>0)>(9);
     if ($77) {
      $exponent$0$ph294 = $exponent$0$ph363;$have_overflow$0$lcssa32 = 0;
      break;
     }
     $78 = (4 - ($phase$3$ph362))|0;
     $79 = ($exponent$0$ph363*10)|0;
     $80 = (($76) + ($79))|0;
     $81 = ($80|0)>(1000);
     if ($81) {
      $$lcssa = $78;$$lcssa449 = $74;$$lcssa451 = $80;$p$6$lcssa450 = $p$6;
      label = 46;
      break;
     } else {
      $153 = $74;$exponent$0$ph363 = $80;$p$5$ph364 = $p$6;$phase$3$ph362 = $78;
     }
    }
    if ((label|0) == 46) {
     $154 = $$lcssa449;$have_overflow$0$us = 0;$p$5$us = $p$6$lcssa450;$phase$3$us = $$lcssa;
     while(1) {
      $61 = ($phase$3$us|0)==(0);
      if ($61) {
       $65 = $154;$p$6$us = $p$5$us;
      } else {
       $62 = ((($p$5$us)) + 1|0);
       $63 = ($62>>>0)<($limit>>>0);
       if (!($63)) {
        $result$0 = 0;$sign$0$ph2365 = $sign$0$ph23$lcssa63;
        break L26;
       }
       $$pre195 = HEAP8[$62>>0]|0;
       $65 = $$pre195;$p$6$us = $62;
      }
      $64 = $65&255;
      $66 = $64 >>> $phase$3$us;
      $67 = $66 & 14;
      $68 = (4 - ($phase$3$us))|0;
      $69 = ($67>>>0)>(9);
      if ($69) {
       $exponent$0$ph294 = $$lcssa451;$have_overflow$0$lcssa32 = $have_overflow$0$us;
       break;
      } else {
       $154 = $65;$have_overflow$0$us = 1;$p$5$us = $p$6$us;$phase$3$us = $68;
      }
     }
    }
    $82 = (0 - ($exponent$0$ph294))|0;
    $exponent$0$ = $59 ? $82 : $exponent$0$ph294;
    $exponent$1 = $exponent$0$;$have_overflow$1 = $have_overflow$0$lcssa32;
   } else {
    $exponent$1 = 0;$have_overflow$1 = 0;
   }
   $83 = ($number$2|0)==(0);
   if ($83) {
    $result$0 = 0;$sign$0$ph2365 = $sign$0$ph23$lcssa63;
   } else {
    $84 = ($have_overflow$1|0)==(0);
    do {
     if ($84) {
      $85 = (($exponent_add$2) + ($power_ten))|0;
      $86 = (($85) + ($exponent$1))|0;
      if (!($0)) {
       $121 = (($86) + ($integer_length$0$ph359))|0;
       $122 = (($fraction_length$1) - ($86))|0;
       $123 = ($121|0)>(5);
       if ($123) {
        $result$0 = 2147483647;$sign$0$ph2365 = $sign$0$ph23$lcssa63;
        break L26;
       }
       $124 = ($121|0)<(-5);
       if ($124) {
        break;
       }
       $125 = ($121|0)<(0);
       if ($125) {
        $126 = (0 - ($121))|0;
        $127 = (51120 + ($126<<2)|0);
        $128 = HEAP32[$127>>2]|0;
        $129 = (($number$2|0) / ($128|0))&-1;
        $130 = (($fraction_length$1) + ($integer_length$0$ph359))|0;
        $fraction_length$2 = $130;$number$5 = $129;
       } else {
        $fraction_length$2 = $122;$number$5 = $number$2;
       }
       $131 = ($fraction_length$2|0)==(10);
       if ($131) {
        $132 = (($number$5|0) / 10)&-1;
        $fraction_length$38 = 9;$number$67 = $132;
       } else {
        $133 = ($fraction_length$2|0)>(0);
        if ($133) {
         $fraction_length$38 = $fraction_length$2;$number$67 = $number$5;
        } else {
         $139 = (0 - ($fraction_length$2))|0;
         $140 = (51120 + ($139<<2)|0);
         $141 = HEAP32[$140>>2]|0;
         $142 = Math_imul($141, $number$5)|0;
         $143 = ($142|0)>(32767);
         $144 = $142 << 16;
         $$6 = $143 ? 2147483647 : $144;
         $result$0 = $$6;$sign$0$ph2365 = $sign$0$ph23$lcssa63;
         break L26;
        }
       }
       $134 = (51120 + ($fraction_length$38<<2)|0);
       $135 = HEAP32[$134>>2]|0;
       $136 = (($number$67|0) / ($135|0))&-1;
       $137 = ($136|0)>(32767);
       if ($137) {
        $result$0 = 0;$sign$0$ph2365 = $sign$0$ph23$lcssa63;
        break L26;
       }
       $138 = (_FT_DivFix($number$67,$135)|0);
       $result$0 = $138;$sign$0$ph2365 = $sign$0$ph23$lcssa63;
       break L26;
      }
      $87 = (($fraction_length$1) + ($integer_length$0$ph359))|0;
      $88 = (($86) + ($integer_length$0$ph359))|0;
      $89 = ($87|0)<(6);
      if (!($89)) {
       $109 = (($87) + -5)|0;
       $110 = (51120 + ($109<<2)|0);
       $111 = HEAP32[$110>>2]|0;
       $112 = (($number$2|0) / ($111|0))&-1;
       $113 = ($112|0)>(32767);
       if ($113) {
        $114 = (($87) + -4)|0;
        $115 = (51120 + ($114<<2)|0);
        $116 = HEAP32[$115>>2]|0;
        $117 = (_FT_DivFix($number$2,$116)|0);
        $118 = (($88) + -4)|0;
        HEAP32[$scaling>>2] = $118;
        $result$0 = $117;$sign$0$ph2365 = $sign$0$ph23$lcssa63;
        break L26;
       } else {
        $119 = (_FT_DivFix($number$2,$111)|0);
        $120 = (($88) + -5)|0;
        HEAP32[$scaling>>2] = $120;
        $result$0 = $119;$sign$0$ph2365 = $sign$0$ph23$lcssa63;
        break L26;
       }
      }
      $90 = ($number$2|0)>(32767);
      if ($90) {
       $91 = (_FT_DivFix($number$2,10)|0);
       $92 = (1 - ($87))|0;
       $93 = (($92) + ($88))|0;
       HEAP32[$scaling>>2] = $93;
       $result$0 = $91;$sign$0$ph2365 = $sign$0$ph23$lcssa63;
       break L26;
      }
      $94 = ($88|0)>(0);
      do {
       if ($94) {
        $95 = ($88|0)<(5);
        $96 = $95 ? $88 : 5;
        $97 = (($96) - ($87))|0;
        $98 = ($97|0)>(0);
        if (!($98)) {
         $106 = (($88) - ($87))|0;
         $exponent$3 = $106;$number$4 = $number$2;
         break;
        }
        $99 = (($88) - ($96))|0;
        $100 = (51120 + ($97<<2)|0);
        $101 = HEAP32[$100>>2]|0;
        $102 = Math_imul($101, $number$2)|0;
        $103 = ($102|0)>(32767);
        if ($103) {
         $104 = (($102|0) / 10)&-1;
         $105 = (($99) + 1)|0;
         $exponent$3 = $105;$number$4 = $104;
        } else {
         $exponent$3 = $99;$number$4 = $102;
        }
       } else {
        $107 = (($88) - ($87))|0;
        $exponent$3 = $107;$number$4 = $number$2;
       }
      } while(0);
      $108 = $number$4 << 16;
      HEAP32[$scaling>>2] = $exponent$3;
      $result$0 = $108;$sign$0$ph2365 = $sign$0$ph23$lcssa63;
      break L26;
     } else {
      if (!($59)) {
       $result$0 = 2147483647;$sign$0$ph2365 = $sign$0$ph23$lcssa63;
       break L26;
      }
     }
    } while(0);
    $result$0 = 0;$sign$0$ph2365 = $sign$0$ph23$lcssa63;
   }
  }
 } while(0);
 $145 = ($sign$0$ph2365|0)==(0);
 $146 = (0 - ($result$0))|0;
 $result$0$ = $145 ? $result$0 : $146;
 return ($result$0$|0);
}
function _cff_parse_font_bbox($parser) {
 $parser = $parser|0;
 var $$$i$i = 0, $$$i$i11 = 0, $$$i$i7 = 0, $$$i$i9 = 0, $$idx$val = 0, $$idx2$val = 0, $$idx4$val = 0, $$idx6$val = 0, $$val = 0, $$val1 = 0, $$val3 = 0, $$val5 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0;
 var $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0;
 var $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0;
 var $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $error$0 = 0, $sext$i = 0, $sext$i13 = 0, $sext$i16 = 0, $sext$i19 = 0, $val$0$i = 0, $val$0$i14 = 0, $val$0$i1425 = 0, $val$0$i17 = 0, $val$0$i1724 = 0;
 var $val$0$i20 = 0, $val$0$i2023 = 0, $val$0$i22 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($parser)) + 412|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($parser)) + 404|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($parser)) + 32|0);
 $5 = ($3>>>0)<($4>>>0);
 if ($5) {
  $error$0 = 161;
  return ($error$0|0);
 }
 $6 = ((($parser)) + 16|0);
 $7 = ((($parser)) + 20|0);
 $$val5 = HEAP32[$6>>2]|0;
 $$idx6$val = HEAP32[$7>>2]|0;
 $8 = HEAP8[$$val5>>0]|0;
 $9 = ($8<<24>>24)==(30);
 do {
  if ($9) {
   $10 = (_cff_parse_real($$val5,$$idx6$val,0,0)|0);
   $62 = $10;
  } else {
   $11 = ((($$val5)) + 1|0);
   $12 = $8&255;
   L7: do {
    switch ($8<<24>>24) {
    case 28:  {
     $13 = ((($$val5)) + 3|0);
     $14 = ($13>>>0)>($$idx6$val>>>0);
     if ($14) {
      $val$0$i22 = 0;
     } else {
      $15 = HEAP8[$11>>0]|0;
      $16 = $15&255;
      $17 = $16 << 8;
      $18 = ((($$val5)) + 2|0);
      $19 = HEAP8[$18>>0]|0;
      $20 = $19&255;
      $21 = $17 | $20;
      $sext$i = $21 << 16;
      $22 = $sext$i >> 16;
      $val$0$i22 = $22;
     }
     break;
    }
    case 29:  {
     $23 = ((($$val5)) + 5|0);
     $24 = ($23>>>0)>($$idx6$val>>>0);
     if ($24) {
      $val$0$i22 = 0;
     } else {
      $25 = HEAP8[$11>>0]|0;
      $26 = $25&255;
      $27 = $26 << 24;
      $28 = ((($$val5)) + 2|0);
      $29 = HEAP8[$28>>0]|0;
      $30 = $29&255;
      $31 = $30 << 16;
      $32 = $31 | $27;
      $33 = ((($$val5)) + 3|0);
      $34 = HEAP8[$33>>0]|0;
      $35 = $34&255;
      $36 = $35 << 8;
      $37 = $32 | $36;
      $38 = ((($$val5)) + 4|0);
      $39 = HEAP8[$38>>0]|0;
      $40 = $39&255;
      $41 = $37 | $40;
      $val$0$i = $41;
      label = 16;
     }
     break;
    }
    default: {
     $42 = ($8&255)<(247);
     if ($42) {
      $43 = (($12) + -139)|0;
      $val$0$i22 = $43;
      break L7;
     }
     $44 = ($8&255)<(251);
     $45 = ((($$val5)) + 2|0);
     $46 = ($45>>>0)>($$idx6$val>>>0);
     if ($44) {
      if ($46) {
       $val$0$i22 = 0;
       break L7;
      }
      $47 = $12 << 8;
      $48 = (($47) + -63232)|0;
      $49 = HEAP8[$11>>0]|0;
      $50 = $49&255;
      $51 = $50 | $48;
      $52 = (($51) + 108)|0;
      $val$0$i = $52;
      label = 16;
      break L7;
     } else {
      if ($46) {
       $val$0$i22 = 0;
       break L7;
      }
      $53 = (251 - ($12))|0;
      $54 = $53 << 8;
      $55 = HEAP8[$11>>0]|0;
      $56 = $55&255;
      $57 = (($54) + -108)|0;
      $58 = (($57) - ($56))|0;
      $val$0$i = $58;
      label = 16;
      break L7;
     }
    }
    }
   } while(0);
   if ((label|0) == 16) {
    $59 = ($val$0$i|0)>(32767);
    if ($59) {
     $62 = 2147483647;
     break;
    } else {
     $val$0$i22 = $val$0$i;
    }
   }
   $60 = ($val$0$i22|0)<(-32767);
   $61 = $val$0$i22 << 16;
   $$$i$i = $60 ? -2147483647 : $61;
   $62 = $$$i$i;
  }
 } while(0);
 $63 = (_FT_RoundFix($62)|0);
 $64 = ((($1)) + 84|0);
 HEAP32[$64>>2] = $63;
 $65 = ((($parser)) + 24|0);
 $$val3 = HEAP32[$7>>2]|0;
 $$idx4$val = HEAP32[$65>>2]|0;
 $66 = HEAP8[$$val3>>0]|0;
 $67 = ($66<<24>>24)==(30);
 do {
  if ($67) {
   $68 = (_cff_parse_real($$val3,$$idx4$val,0,0)|0);
   $120 = $68;
  } else {
   $69 = ((($$val3)) + 1|0);
   $70 = $66&255;
   L28: do {
    switch ($66<<24>>24) {
    case 28:  {
     $71 = ((($$val3)) + 3|0);
     $72 = ($71>>>0)>($$idx4$val>>>0);
     if ($72) {
      $val$0$i2023 = 0;
     } else {
      $73 = HEAP8[$69>>0]|0;
      $74 = $73&255;
      $75 = $74 << 8;
      $76 = ((($$val3)) + 2|0);
      $77 = HEAP8[$76>>0]|0;
      $78 = $77&255;
      $79 = $75 | $78;
      $sext$i19 = $79 << 16;
      $80 = $sext$i19 >> 16;
      $val$0$i2023 = $80;
     }
     break;
    }
    case 29:  {
     $81 = ((($$val3)) + 5|0);
     $82 = ($81>>>0)>($$idx4$val>>>0);
     if ($82) {
      $val$0$i2023 = 0;
     } else {
      $83 = HEAP8[$69>>0]|0;
      $84 = $83&255;
      $85 = $84 << 24;
      $86 = ((($$val3)) + 2|0);
      $87 = HEAP8[$86>>0]|0;
      $88 = $87&255;
      $89 = $88 << 16;
      $90 = $89 | $85;
      $91 = ((($$val3)) + 3|0);
      $92 = HEAP8[$91>>0]|0;
      $93 = $92&255;
      $94 = $93 << 8;
      $95 = $90 | $94;
      $96 = ((($$val3)) + 4|0);
      $97 = HEAP8[$96>>0]|0;
      $98 = $97&255;
      $99 = $95 | $98;
      $val$0$i20 = $99;
      label = 32;
     }
     break;
    }
    default: {
     $100 = ($66&255)<(247);
     if ($100) {
      $101 = (($70) + -139)|0;
      $val$0$i2023 = $101;
      break L28;
     }
     $102 = ($66&255)<(251);
     $103 = ((($$val3)) + 2|0);
     $104 = ($103>>>0)>($$idx4$val>>>0);
     if ($102) {
      if ($104) {
       $val$0$i2023 = 0;
       break L28;
      }
      $105 = $70 << 8;
      $106 = (($105) + -63232)|0;
      $107 = HEAP8[$69>>0]|0;
      $108 = $107&255;
      $109 = $108 | $106;
      $110 = (($109) + 108)|0;
      $val$0$i20 = $110;
      label = 32;
      break L28;
     } else {
      if ($104) {
       $val$0$i2023 = 0;
       break L28;
      }
      $111 = (251 - ($70))|0;
      $112 = $111 << 8;
      $113 = HEAP8[$69>>0]|0;
      $114 = $113&255;
      $115 = (($112) + -108)|0;
      $116 = (($115) - ($114))|0;
      $val$0$i20 = $116;
      label = 32;
      break L28;
     }
    }
    }
   } while(0);
   if ((label|0) == 32) {
    $117 = ($val$0$i20|0)>(32767);
    if ($117) {
     $120 = 2147483647;
     break;
    } else {
     $val$0$i2023 = $val$0$i20;
    }
   }
   $118 = ($val$0$i2023|0)<(-32767);
   $119 = $val$0$i2023 << 16;
   $$$i$i7 = $118 ? -2147483647 : $119;
   $120 = $$$i$i7;
  }
 } while(0);
 $121 = (_FT_RoundFix($120)|0);
 $122 = ((($1)) + 88|0);
 HEAP32[$122>>2] = $121;
 $123 = ((($parser)) + 28|0);
 $$val1 = HEAP32[$65>>2]|0;
 $$idx2$val = HEAP32[$123>>2]|0;
 $124 = HEAP8[$$val1>>0]|0;
 $125 = ($124<<24>>24)==(30);
 do {
  if ($125) {
   $126 = (_cff_parse_real($$val1,$$idx2$val,0,0)|0);
   $178 = $126;
  } else {
   $127 = ((($$val1)) + 1|0);
   $128 = $124&255;
   L49: do {
    switch ($124<<24>>24) {
    case 28:  {
     $129 = ((($$val1)) + 3|0);
     $130 = ($129>>>0)>($$idx2$val>>>0);
     if ($130) {
      $val$0$i1724 = 0;
     } else {
      $131 = HEAP8[$127>>0]|0;
      $132 = $131&255;
      $133 = $132 << 8;
      $134 = ((($$val1)) + 2|0);
      $135 = HEAP8[$134>>0]|0;
      $136 = $135&255;
      $137 = $133 | $136;
      $sext$i16 = $137 << 16;
      $138 = $sext$i16 >> 16;
      $val$0$i1724 = $138;
     }
     break;
    }
    case 29:  {
     $139 = ((($$val1)) + 5|0);
     $140 = ($139>>>0)>($$idx2$val>>>0);
     if ($140) {
      $val$0$i1724 = 0;
     } else {
      $141 = HEAP8[$127>>0]|0;
      $142 = $141&255;
      $143 = $142 << 24;
      $144 = ((($$val1)) + 2|0);
      $145 = HEAP8[$144>>0]|0;
      $146 = $145&255;
      $147 = $146 << 16;
      $148 = $147 | $143;
      $149 = ((($$val1)) + 3|0);
      $150 = HEAP8[$149>>0]|0;
      $151 = $150&255;
      $152 = $151 << 8;
      $153 = $148 | $152;
      $154 = ((($$val1)) + 4|0);
      $155 = HEAP8[$154>>0]|0;
      $156 = $155&255;
      $157 = $153 | $156;
      $val$0$i17 = $157;
      label = 48;
     }
     break;
    }
    default: {
     $158 = ($124&255)<(247);
     if ($158) {
      $159 = (($128) + -139)|0;
      $val$0$i1724 = $159;
      break L49;
     }
     $160 = ($124&255)<(251);
     $161 = ((($$val1)) + 2|0);
     $162 = ($161>>>0)>($$idx2$val>>>0);
     if ($160) {
      if ($162) {
       $val$0$i1724 = 0;
       break L49;
      }
      $163 = $128 << 8;
      $164 = (($163) + -63232)|0;
      $165 = HEAP8[$127>>0]|0;
      $166 = $165&255;
      $167 = $166 | $164;
      $168 = (($167) + 108)|0;
      $val$0$i17 = $168;
      label = 48;
      break L49;
     } else {
      if ($162) {
       $val$0$i1724 = 0;
       break L49;
      }
      $169 = (251 - ($128))|0;
      $170 = $169 << 8;
      $171 = HEAP8[$127>>0]|0;
      $172 = $171&255;
      $173 = (($170) + -108)|0;
      $174 = (($173) - ($172))|0;
      $val$0$i17 = $174;
      label = 48;
      break L49;
     }
    }
    }
   } while(0);
   if ((label|0) == 48) {
    $175 = ($val$0$i17|0)>(32767);
    if ($175) {
     $178 = 2147483647;
     break;
    } else {
     $val$0$i1724 = $val$0$i17;
    }
   }
   $176 = ($val$0$i1724|0)<(-32767);
   $177 = $val$0$i1724 << 16;
   $$$i$i9 = $176 ? -2147483647 : $177;
   $178 = $$$i$i9;
  }
 } while(0);
 $179 = (_FT_RoundFix($178)|0);
 $180 = ((($1)) + 92|0);
 HEAP32[$180>>2] = $179;
 $$val = HEAP32[$123>>2]|0;
 $$idx$val = HEAP32[$4>>2]|0;
 $181 = HEAP8[$$val>>0]|0;
 $182 = ($181<<24>>24)==(30);
 do {
  if ($182) {
   $183 = (_cff_parse_real($$val,$$idx$val,0,0)|0);
   $235 = $183;
  } else {
   $184 = ((($$val)) + 1|0);
   $185 = $181&255;
   L70: do {
    switch ($181<<24>>24) {
    case 28:  {
     $186 = ((($$val)) + 3|0);
     $187 = ($186>>>0)>($$idx$val>>>0);
     if ($187) {
      $val$0$i1425 = 0;
     } else {
      $188 = HEAP8[$184>>0]|0;
      $189 = $188&255;
      $190 = $189 << 8;
      $191 = ((($$val)) + 2|0);
      $192 = HEAP8[$191>>0]|0;
      $193 = $192&255;
      $194 = $190 | $193;
      $sext$i13 = $194 << 16;
      $195 = $sext$i13 >> 16;
      $val$0$i1425 = $195;
     }
     break;
    }
    case 29:  {
     $196 = ((($$val)) + 5|0);
     $197 = ($196>>>0)>($$idx$val>>>0);
     if ($197) {
      $val$0$i1425 = 0;
     } else {
      $198 = HEAP8[$184>>0]|0;
      $199 = $198&255;
      $200 = $199 << 24;
      $201 = ((($$val)) + 2|0);
      $202 = HEAP8[$201>>0]|0;
      $203 = $202&255;
      $204 = $203 << 16;
      $205 = $204 | $200;
      $206 = ((($$val)) + 3|0);
      $207 = HEAP8[$206>>0]|0;
      $208 = $207&255;
      $209 = $208 << 8;
      $210 = $205 | $209;
      $211 = ((($$val)) + 4|0);
      $212 = HEAP8[$211>>0]|0;
      $213 = $212&255;
      $214 = $210 | $213;
      $val$0$i14 = $214;
      label = 64;
     }
     break;
    }
    default: {
     $215 = ($181&255)<(247);
     if ($215) {
      $216 = (($185) + -139)|0;
      $val$0$i1425 = $216;
      break L70;
     }
     $217 = ($181&255)<(251);
     $218 = ((($$val)) + 2|0);
     $219 = ($218>>>0)>($$idx$val>>>0);
     if ($217) {
      if ($219) {
       $val$0$i1425 = 0;
       break L70;
      }
      $220 = $185 << 8;
      $221 = (($220) + -63232)|0;
      $222 = HEAP8[$184>>0]|0;
      $223 = $222&255;
      $224 = $223 | $221;
      $225 = (($224) + 108)|0;
      $val$0$i14 = $225;
      label = 64;
      break L70;
     } else {
      if ($219) {
       $val$0$i1425 = 0;
       break L70;
      }
      $226 = (251 - ($185))|0;
      $227 = $226 << 8;
      $228 = HEAP8[$184>>0]|0;
      $229 = $228&255;
      $230 = (($227) + -108)|0;
      $231 = (($230) - ($229))|0;
      $val$0$i14 = $231;
      label = 64;
      break L70;
     }
    }
    }
   } while(0);
   if ((label|0) == 64) {
    $232 = ($val$0$i14|0)>(32767);
    if ($232) {
     $235 = 2147483647;
     break;
    } else {
     $val$0$i1425 = $val$0$i14;
    }
   }
   $233 = ($val$0$i1425|0)<(-32767);
   $234 = $val$0$i1425 << 16;
   $$$i$i11 = $233 ? -2147483647 : $234;
   $235 = $$$i$i11;
  }
 } while(0);
 $236 = (_FT_RoundFix($235)|0);
 $237 = ((($1)) + 96|0);
 HEAP32[$237>>2] = $236;
 $error$0 = 0;
 return ($error$0|0);
}
function _cff_parse_private_dict($parser) {
 $parser = $parser|0;
 var $$01 = 0, $$idx$val = 0, $$idx3 = 0, $$idx3$val = 0, $$val = 0, $$val$pre = 0, $$val2 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $sext$i = 0, $sext$i5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($parser)) + 412|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($parser)) + 404|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($parser)) + 24|0);
 $5 = ($3>>>0)<($4>>>0);
 if ($5) {
  $$01 = 161;
  return ($$01|0);
 }
 $6 = ((($parser)) + 16|0);
 $$val2 = HEAP32[$6>>2]|0;
 $$idx3 = ((($parser)) + 20|0);
 $$idx3$val = HEAP32[$$idx3>>2]|0;
 $7 = HEAP8[$$val2>>0]|0;
 $8 = ($7<<24>>24)==(30);
 L4: do {
  if ($8) {
   $9 = (_cff_parse_real($$val2,$$idx3$val,0,0)|0);
   $10 = $9 >> 16;
   $60 = $10;
   label = 16;
  } else {
   $11 = ((($$val2)) + 1|0);
   $12 = $7&255;
   switch ($7<<24>>24) {
   case 28:  {
    $13 = ((($$val2)) + 3|0);
    $14 = ($13>>>0)>($$idx3$val>>>0);
    if ($14) {
     $$val = $$idx3$val;$62 = 0;
     break L4;
    }
    $15 = HEAP8[$11>>0]|0;
    $16 = $15&255;
    $17 = $16 << 8;
    $18 = ((($$val2)) + 2|0);
    $19 = HEAP8[$18>>0]|0;
    $20 = $19&255;
    $21 = $17 | $20;
    $sext$i = $21 << 16;
    $22 = $sext$i >> 16;
    $60 = $22;
    label = 16;
    break L4;
    break;
   }
   case 29:  {
    $23 = ((($$val2)) + 5|0);
    $24 = ($23>>>0)>($$idx3$val>>>0);
    if ($24) {
     $$val = $$idx3$val;$62 = 0;
     break L4;
    }
    $25 = HEAP8[$11>>0]|0;
    $26 = $25&255;
    $27 = $26 << 24;
    $28 = ((($$val2)) + 2|0);
    $29 = HEAP8[$28>>0]|0;
    $30 = $29&255;
    $31 = $30 << 16;
    $32 = $31 | $27;
    $33 = ((($$val2)) + 3|0);
    $34 = HEAP8[$33>>0]|0;
    $35 = $34&255;
    $36 = $35 << 8;
    $37 = $32 | $36;
    $38 = ((($$val2)) + 4|0);
    $39 = HEAP8[$38>>0]|0;
    $40 = $39&255;
    $41 = $37 | $40;
    $60 = $41;
    label = 16;
    break L4;
    break;
   }
   default: {
    $42 = ($7&255)<(247);
    if ($42) {
     $43 = (($12) + -139)|0;
     $60 = $43;
     label = 16;
     break L4;
    }
    $44 = ($7&255)<(251);
    $45 = ((($$val2)) + 2|0);
    $46 = ($45>>>0)>($$idx3$val>>>0);
    if ($44) {
     if ($46) {
      $$val = $$idx3$val;$62 = 0;
      break L4;
     }
     $47 = $12 << 8;
     $48 = (($47) + -63232)|0;
     $49 = HEAP8[$11>>0]|0;
     $50 = $49&255;
     $51 = $50 | $48;
     $52 = (($51) + 108)|0;
     $60 = $52;
     label = 16;
     break L4;
    } else {
     if ($46) {
      $$val = $$idx3$val;$62 = 0;
      break L4;
     }
     $53 = (251 - ($12))|0;
     $54 = $53 << 8;
     $55 = HEAP8[$11>>0]|0;
     $56 = $55&255;
     $57 = (($54) + -108)|0;
     $58 = (($57) - ($56))|0;
     $60 = $58;
     label = 16;
     break L4;
    }
   }
   }
  }
 } while(0);
 do {
  if ((label|0) == 16) {
   $59 = ($60|0)<(0);
   if ($59) {
    $$01 = 3;
    return ($$01|0);
   } else {
    $$val$pre = HEAP32[$$idx3>>2]|0;
    $$val = $$val$pre;$62 = $60;
    break;
   }
  }
 } while(0);
 $61 = ((($1)) + 120|0);
 HEAP32[$61>>2] = $62;
 $$idx$val = HEAP32[$4>>2]|0;
 $63 = HEAP8[$$val>>0]|0;
 $64 = ($63<<24>>24)==(30);
 L27: do {
  if ($64) {
   $65 = (_cff_parse_real($$val,$$idx$val,0,0)|0);
   $66 = $65 >> 16;
   $116 = $66;
   label = 32;
  } else {
   $67 = ((($$val)) + 1|0);
   $68 = $63&255;
   switch ($63<<24>>24) {
   case 28:  {
    $69 = ((($$val)) + 3|0);
    $70 = ($69>>>0)>($$idx$val>>>0);
    if ($70) {
     $118 = 0;
     break L27;
    }
    $71 = HEAP8[$67>>0]|0;
    $72 = $71&255;
    $73 = $72 << 8;
    $74 = ((($$val)) + 2|0);
    $75 = HEAP8[$74>>0]|0;
    $76 = $75&255;
    $77 = $73 | $76;
    $sext$i5 = $77 << 16;
    $78 = $sext$i5 >> 16;
    $116 = $78;
    label = 32;
    break L27;
    break;
   }
   case 29:  {
    $79 = ((($$val)) + 5|0);
    $80 = ($79>>>0)>($$idx$val>>>0);
    if ($80) {
     $118 = 0;
     break L27;
    }
    $81 = HEAP8[$67>>0]|0;
    $82 = $81&255;
    $83 = $82 << 24;
    $84 = ((($$val)) + 2|0);
    $85 = HEAP8[$84>>0]|0;
    $86 = $85&255;
    $87 = $86 << 16;
    $88 = $87 | $83;
    $89 = ((($$val)) + 3|0);
    $90 = HEAP8[$89>>0]|0;
    $91 = $90&255;
    $92 = $91 << 8;
    $93 = $88 | $92;
    $94 = ((($$val)) + 4|0);
    $95 = HEAP8[$94>>0]|0;
    $96 = $95&255;
    $97 = $93 | $96;
    $116 = $97;
    label = 32;
    break L27;
    break;
   }
   default: {
    $98 = ($63&255)<(247);
    if ($98) {
     $99 = (($68) + -139)|0;
     $116 = $99;
     label = 32;
     break L27;
    }
    $100 = ($63&255)<(251);
    $101 = ((($$val)) + 2|0);
    $102 = ($101>>>0)>($$idx$val>>>0);
    if ($100) {
     if ($102) {
      $118 = 0;
      break L27;
     }
     $103 = $68 << 8;
     $104 = (($103) + -63232)|0;
     $105 = HEAP8[$67>>0]|0;
     $106 = $105&255;
     $107 = $106 | $104;
     $108 = (($107) + 108)|0;
     $116 = $108;
     label = 32;
     break L27;
    } else {
     if ($102) {
      $118 = 0;
      break L27;
     }
     $109 = (251 - ($68))|0;
     $110 = $109 << 8;
     $111 = HEAP8[$67>>0]|0;
     $112 = $111&255;
     $113 = (($110) + -108)|0;
     $114 = (($113) - ($112))|0;
     $116 = $114;
     label = 32;
     break L27;
    }
   }
   }
  }
 } while(0);
 if ((label|0) == 32) {
  $115 = ($116|0)<(0);
  if ($115) {
   $$01 = 3;
   return ($$01|0);
  } else {
   $118 = $116;
  }
 }
 $117 = ((($1)) + 116|0);
 HEAP32[$117>>2] = $118;
 $$01 = 0;
 return ($$01|0);
}
function _cff_parse_cid_ros($parser) {
 $parser = $parser|0;
 var $$idx$val = 0, $$idx2$val = 0, $$idx4$val = 0, $$val = 0, $$val$pre = 0, $$val1 = 0, $$val1$pre = 0, $$val3 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $error$0 = 0;
 var $sext$i = 0, $sext$i10 = 0, $sext$i6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($parser)) + 412|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($parser)) + 404|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($parser)) + 28|0);
 $5 = ($3>>>0)<($4>>>0);
 if ($5) {
  $error$0 = 161;
  return ($error$0|0);
 }
 $6 = ((($parser)) + 16|0);
 $7 = ((($parser)) + 20|0);
 $$val3 = HEAP32[$6>>2]|0;
 $$idx4$val = HEAP32[$7>>2]|0;
 $8 = HEAP8[$$val3>>0]|0;
 $9 = ($8<<24>>24)==(30);
 L4: do {
  if ($9) {
   $10 = (_cff_parse_real($$val3,$$idx4$val,0,0)|0);
   $11 = $10 >> 16;
   $$val1$pre = HEAP32[$7>>2]|0;
   $$val1 = $$val1$pre;$61 = $11;
  } else {
   $12 = ((($$val3)) + 1|0);
   $13 = $8&255;
   switch ($8<<24>>24) {
   case 28:  {
    $14 = ((($$val3)) + 3|0);
    $15 = ($14>>>0)>($$idx4$val>>>0);
    if ($15) {
     $$val1 = $$idx4$val;$61 = 0;
     break L4;
    }
    $16 = HEAP8[$12>>0]|0;
    $17 = $16&255;
    $18 = $17 << 8;
    $19 = ((($$val3)) + 2|0);
    $20 = HEAP8[$19>>0]|0;
    $21 = $20&255;
    $22 = $18 | $21;
    $sext$i = $22 << 16;
    $23 = $sext$i >> 16;
    $$val1 = $$idx4$val;$61 = $23;
    break L4;
    break;
   }
   case 29:  {
    $24 = ((($$val3)) + 5|0);
    $25 = ($24>>>0)>($$idx4$val>>>0);
    if ($25) {
     $$val1 = $$idx4$val;$61 = 0;
     break L4;
    }
    $26 = HEAP8[$12>>0]|0;
    $27 = $26&255;
    $28 = $27 << 24;
    $29 = ((($$val3)) + 2|0);
    $30 = HEAP8[$29>>0]|0;
    $31 = $30&255;
    $32 = $31 << 16;
    $33 = $32 | $28;
    $34 = ((($$val3)) + 3|0);
    $35 = HEAP8[$34>>0]|0;
    $36 = $35&255;
    $37 = $36 << 8;
    $38 = $33 | $37;
    $39 = ((($$val3)) + 4|0);
    $40 = HEAP8[$39>>0]|0;
    $41 = $40&255;
    $42 = $38 | $41;
    $$val1 = $$idx4$val;$61 = $42;
    break L4;
    break;
   }
   default: {
    $43 = ($8&255)<(247);
    if ($43) {
     $44 = (($13) + -139)|0;
     $$val1 = $$idx4$val;$61 = $44;
     break L4;
    }
    $45 = ($8&255)<(251);
    $46 = ((($$val3)) + 2|0);
    $47 = ($46>>>0)>($$idx4$val>>>0);
    if ($45) {
     if ($47) {
      $$val1 = $$idx4$val;$61 = 0;
      break L4;
     }
     $48 = $13 << 8;
     $49 = (($48) + -63232)|0;
     $50 = HEAP8[$12>>0]|0;
     $51 = $50&255;
     $52 = $51 | $49;
     $53 = (($52) + 108)|0;
     $$val1 = $$idx4$val;$61 = $53;
     break L4;
    } else {
     if ($47) {
      $$val1 = $$idx4$val;$61 = 0;
      break L4;
     }
     $54 = (251 - ($13))|0;
     $55 = $54 << 8;
     $56 = HEAP8[$12>>0]|0;
     $57 = $56&255;
     $58 = (($55) + -108)|0;
     $59 = (($58) - ($57))|0;
     $$val1 = $$idx4$val;$61 = $59;
     break L4;
    }
   }
   }
  }
 } while(0);
 $60 = ((($1)) + 132|0);
 HEAP32[$60>>2] = $61;
 $62 = ((($parser)) + 24|0);
 $$idx2$val = HEAP32[$62>>2]|0;
 $63 = HEAP8[$$val1>>0]|0;
 $64 = ($63<<24>>24)==(30);
 L22: do {
  if ($64) {
   $65 = (_cff_parse_real($$val1,$$idx2$val,0,0)|0);
   $66 = $65 >> 16;
   $$val$pre = HEAP32[$62>>2]|0;
   $$val = $$val$pre;$116 = $66;
  } else {
   $67 = ((($$val1)) + 1|0);
   $68 = $63&255;
   switch ($63<<24>>24) {
   case 28:  {
    $69 = ((($$val1)) + 3|0);
    $70 = ($69>>>0)>($$idx2$val>>>0);
    if ($70) {
     $$val = $$idx2$val;$116 = 0;
     break L22;
    }
    $71 = HEAP8[$67>>0]|0;
    $72 = $71&255;
    $73 = $72 << 8;
    $74 = ((($$val1)) + 2|0);
    $75 = HEAP8[$74>>0]|0;
    $76 = $75&255;
    $77 = $73 | $76;
    $sext$i6 = $77 << 16;
    $78 = $sext$i6 >> 16;
    $$val = $$idx2$val;$116 = $78;
    break L22;
    break;
   }
   case 29:  {
    $79 = ((($$val1)) + 5|0);
    $80 = ($79>>>0)>($$idx2$val>>>0);
    if ($80) {
     $$val = $$idx2$val;$116 = 0;
     break L22;
    }
    $81 = HEAP8[$67>>0]|0;
    $82 = $81&255;
    $83 = $82 << 24;
    $84 = ((($$val1)) + 2|0);
    $85 = HEAP8[$84>>0]|0;
    $86 = $85&255;
    $87 = $86 << 16;
    $88 = $87 | $83;
    $89 = ((($$val1)) + 3|0);
    $90 = HEAP8[$89>>0]|0;
    $91 = $90&255;
    $92 = $91 << 8;
    $93 = $88 | $92;
    $94 = ((($$val1)) + 4|0);
    $95 = HEAP8[$94>>0]|0;
    $96 = $95&255;
    $97 = $93 | $96;
    $$val = $$idx2$val;$116 = $97;
    break L22;
    break;
   }
   default: {
    $98 = ($63&255)<(247);
    if ($98) {
     $99 = (($68) + -139)|0;
     $$val = $$idx2$val;$116 = $99;
     break L22;
    }
    $100 = ($63&255)<(251);
    $101 = ((($$val1)) + 2|0);
    $102 = ($101>>>0)>($$idx2$val>>>0);
    if ($100) {
     if ($102) {
      $$val = $$idx2$val;$116 = 0;
      break L22;
     }
     $103 = $68 << 8;
     $104 = (($103) + -63232)|0;
     $105 = HEAP8[$67>>0]|0;
     $106 = $105&255;
     $107 = $106 | $104;
     $108 = (($107) + 108)|0;
     $$val = $$idx2$val;$116 = $108;
     break L22;
    } else {
     if ($102) {
      $$val = $$idx2$val;$116 = 0;
      break L22;
     }
     $109 = (251 - ($68))|0;
     $110 = $109 << 8;
     $111 = HEAP8[$67>>0]|0;
     $112 = $111&255;
     $113 = (($110) + -108)|0;
     $114 = (($113) - ($112))|0;
     $$val = $$idx2$val;$116 = $114;
     break L22;
    }
   }
   }
  }
 } while(0);
 $115 = ((($1)) + 136|0);
 HEAP32[$115>>2] = $116;
 $$idx$val = HEAP32[$4>>2]|0;
 $117 = HEAP8[$$val>>0]|0;
 $118 = ($117<<24>>24)==(30);
 L40: do {
  if ($118) {
   $119 = (_cff_parse_real($$val,$$idx$val,0,0)|0);
   $120 = $119 >> 16;
   $170 = $120;
  } else {
   $121 = ((($$val)) + 1|0);
   $122 = $117&255;
   switch ($117<<24>>24) {
   case 28:  {
    $123 = ((($$val)) + 3|0);
    $124 = ($123>>>0)>($$idx$val>>>0);
    if ($124) {
     $170 = 0;
     break L40;
    }
    $125 = HEAP8[$121>>0]|0;
    $126 = $125&255;
    $127 = $126 << 8;
    $128 = ((($$val)) + 2|0);
    $129 = HEAP8[$128>>0]|0;
    $130 = $129&255;
    $131 = $127 | $130;
    $sext$i10 = $131 << 16;
    $132 = $sext$i10 >> 16;
    $170 = $132;
    break L40;
    break;
   }
   case 29:  {
    $133 = ((($$val)) + 5|0);
    $134 = ($133>>>0)>($$idx$val>>>0);
    if ($134) {
     $170 = 0;
     break L40;
    }
    $135 = HEAP8[$121>>0]|0;
    $136 = $135&255;
    $137 = $136 << 24;
    $138 = ((($$val)) + 2|0);
    $139 = HEAP8[$138>>0]|0;
    $140 = $139&255;
    $141 = $140 << 16;
    $142 = $141 | $137;
    $143 = ((($$val)) + 3|0);
    $144 = HEAP8[$143>>0]|0;
    $145 = $144&255;
    $146 = $145 << 8;
    $147 = $142 | $146;
    $148 = ((($$val)) + 4|0);
    $149 = HEAP8[$148>>0]|0;
    $150 = $149&255;
    $151 = $147 | $150;
    $170 = $151;
    break L40;
    break;
   }
   default: {
    $152 = ($117&255)<(247);
    if ($152) {
     $153 = (($122) + -139)|0;
     $170 = $153;
     break L40;
    }
    $154 = ($117&255)<(251);
    $155 = ((($$val)) + 2|0);
    $156 = ($155>>>0)>($$idx$val>>>0);
    if ($154) {
     if ($156) {
      $170 = 0;
      break L40;
     }
     $157 = $122 << 8;
     $158 = (($157) + -63232)|0;
     $159 = HEAP8[$121>>0]|0;
     $160 = $159&255;
     $161 = $160 | $158;
     $162 = (($161) + 108)|0;
     $170 = $162;
     break L40;
    } else {
     if ($156) {
      $170 = 0;
      break L40;
     }
     $163 = (251 - ($122))|0;
     $164 = $163 << 8;
     $165 = HEAP8[$121>>0]|0;
     $166 = $165&255;
     $167 = (($164) + -108)|0;
     $168 = (($167) - ($166))|0;
     $170 = $168;
     break L40;
    }
   }
   }
  }
 } while(0);
 $169 = ((($1)) + 140|0);
 HEAP32[$169>>2] = $170;
 $error$0 = 0;
 return ($error$0|0);
}
function _cff_face_done($cffface) {
 $cffface = $cffface|0;
 var $$idx$i = 0, $$idx$val$i = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $idx$06$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($cffface|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($cffface)) + 100|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($cffface)) + 532|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if (!($5)) {
  $6 = ((($4)) + 12|0);
  $7 = HEAP32[$6>>2]|0;
  FUNCTION_TABLE_vi[$7 & 8191]($cffface);
 }
 $8 = ((($cffface)) + 656|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)==(0|0);
 if ($10) {
  return;
 }
 $11 = ((($9)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($9)) + 84|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ($14|0)==(0|0);
 if (!($15)) {
  $16 = ((($14)) + 28|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = ((($9)) + 112|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = ($19|0)==(0|0);
  if (!($20)) {
   _FT_Stream_ReleaseFrame($14,$18);
  }
  $21 = ((($9)) + 108|0);
  $22 = HEAP32[$21>>2]|0;
  _ft_mem_free($17,$22);
  ;HEAP32[$13>>2]=0|0;HEAP32[$13+4>>2]=0|0;HEAP32[$13+8>>2]=0|0;HEAP32[$13+12>>2]=0|0;HEAP32[$13+16>>2]=0|0;HEAP32[$13+20>>2]=0|0;HEAP32[$13+24>>2]=0|0;HEAP32[$13+28>>2]=0|0;
 }
 $23 = ((($9)) + 1208|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ($24|0)==(0|0);
 if (!($25)) {
  $26 = ((($24)) + 28|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = ((($9)) + 1236|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = ($29|0)==(0|0);
  if (!($30)) {
   _FT_Stream_ReleaseFrame($24,$28);
  }
  $31 = ((($9)) + 1232|0);
  $32 = HEAP32[$31>>2]|0;
  _ft_mem_free($27,$32);
  ;HEAP32[$23>>2]=0|0;HEAP32[$23+4>>2]=0|0;HEAP32[$23+8>>2]=0|0;HEAP32[$23+12>>2]=0|0;HEAP32[$23+16>>2]=0|0;HEAP32[$23+20>>2]=0|0;HEAP32[$23+24>>2]=0|0;HEAP32[$23+28>>2]=0|0;
 }
 $33 = ((($9)) + 20|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = ($34|0)==(0|0);
 if (!($35)) {
  $36 = ((($34)) + 28|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = ((($9)) + 48|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = ($39|0)==(0|0);
  if (!($40)) {
   _FT_Stream_ReleaseFrame($34,$38);
  }
  $41 = ((($9)) + 44|0);
  $42 = HEAP32[$41>>2]|0;
  _ft_mem_free($37,$42);
  ;HEAP32[$33>>2]=0|0;HEAP32[$33+4>>2]=0|0;HEAP32[$33+8>>2]=0|0;HEAP32[$33+12>>2]=0|0;HEAP32[$33+16>>2]=0|0;HEAP32[$33+20>>2]=0|0;HEAP32[$33+24>>2]=0|0;HEAP32[$33+28>>2]=0|0;
 }
 $43 = ((($9)) + 1176|0);
 $44 = HEAP32[$43>>2]|0;
 $45 = ($44|0)==(0|0);
 if (!($45)) {
  $46 = ((($44)) + 28|0);
  $47 = HEAP32[$46>>2]|0;
  $48 = ((($9)) + 1204|0);
  $49 = HEAP32[$48>>2]|0;
  $50 = ($49|0)==(0|0);
  if (!($50)) {
   _FT_Stream_ReleaseFrame($44,$48);
  }
  $51 = ((($9)) + 1200|0);
  $52 = HEAP32[$51>>2]|0;
  _ft_mem_free($47,$52);
  ;HEAP32[$43>>2]=0|0;HEAP32[$43+4>>2]=0|0;HEAP32[$43+8>>2]=0|0;HEAP32[$43+12>>2]=0|0;HEAP32[$43+16>>2]=0|0;HEAP32[$43+20>>2]=0|0;HEAP32[$43+24>>2]=0|0;HEAP32[$43+28>>2]=0|0;
 }
 $53 = ((($9)) + 1896|0);
 $54 = HEAP32[$53>>2]|0;
 $55 = ($54|0)==(0);
 if (!($55)) {
  $123 = $54;$idx$06$i = 0;
  while(1) {
   $56 = (((($9)) + 1900|0) + ($idx$06$i<<2)|0);
   $57 = HEAP32[$56>>2]|0;
   $58 = ($57|0)==(0|0);
   if ($58) {
    $73 = $123;
   } else {
    $59 = ((($57)) + 536|0);
    $60 = HEAP32[$59>>2]|0;
    $61 = ($60|0)==(0|0);
    if (!($61)) {
     $62 = ((($60)) + 28|0);
     $63 = HEAP32[$62>>2]|0;
     $64 = ((($57)) + 564|0);
     $65 = HEAP32[$64>>2]|0;
     $66 = ($65|0)==(0|0);
     if (!($66)) {
      _FT_Stream_ReleaseFrame($60,$64);
     }
     $67 = ((($57)) + 560|0);
     $68 = HEAP32[$67>>2]|0;
     _ft_mem_free($63,$68);
     ;HEAP32[$59>>2]=0|0;HEAP32[$59+4>>2]=0|0;HEAP32[$59+8>>2]=0|0;HEAP32[$59+12>>2]=0|0;HEAP32[$59+16>>2]=0|0;HEAP32[$59+20>>2]=0|0;HEAP32[$59+24>>2]=0|0;HEAP32[$59+28>>2]=0|0;
    }
    $69 = ((($57)) + 568|0);
    $70 = HEAP32[$69>>2]|0;
    _ft_mem_free($12,$70);
    HEAP32[$69>>2] = 0;
    $$pre$i = HEAP32[$53>>2]|0;
    $73 = $$pre$i;
   }
   $71 = (($idx$06$i) + 1)|0;
   $72 = ($71>>>0)<($73>>>0);
   if ($72) {
    $123 = $73;$idx$06$i = $71;
   } else {
    break;
   }
  }
  $74 = ((($9)) + 1900|0);
  $75 = HEAP32[$74>>2]|0;
  _ft_mem_free($12,$75);
  HEAP32[$74>>2] = 0;
 }
 $76 = ((($9)) + 116|0);
 HEAP32[$76>>2] = 0;
 $77 = ((($9)) + 120|0);
 HEAP32[$77>>2] = 0;
 $78 = ((($9)) + 124|0);
 HEAP32[$78>>2] = 0;
 $79 = HEAP32[$9>>2]|0;
 $$idx$i = ((($79)) + 28|0);
 $$idx$val$i = HEAP32[$$idx$i>>2]|0;
 $80 = ((($9)) + 1164|0);
 $81 = HEAP32[$80>>2]|0;
 _ft_mem_free($$idx$val$i,$81);
 HEAP32[$80>>2] = 0;
 $82 = ((($9)) + 1168|0);
 HEAP32[$82>>2] = 0;
 $83 = ((($9)) + 1160|0);
 $84 = HEAP32[$83>>2]|0;
 _ft_mem_free($$idx$val$i,$84);
 HEAP32[$83>>2] = 0;
 $85 = ((($9)) + 1152|0);
 HEAP32[$85>>2] = 0;
 $86 = ((($9)) + 1156|0);
 HEAP32[$86>>2] = 0;
 $87 = ((($9)) + 1860|0);
 $88 = HEAP32[$87>>2]|0;
 $89 = ($88|0)==(0|0);
 if (!($89)) {
  $90 = ((($88)) + 28|0);
  $91 = HEAP32[$90>>2]|0;
  $92 = ((($9)) + 1888|0);
  $93 = HEAP32[$92>>2]|0;
  $94 = ($93|0)==(0|0);
  if (!($94)) {
   _FT_Stream_ReleaseFrame($88,$92);
  }
  $95 = ((($9)) + 1884|0);
  $96 = HEAP32[$95>>2]|0;
  _ft_mem_free($91,$96);
  ;HEAP32[$87>>2]=0|0;HEAP32[$87+4>>2]=0|0;HEAP32[$87+8>>2]=0|0;HEAP32[$87+12>>2]=0|0;HEAP32[$87+16>>2]=0|0;HEAP32[$87+20>>2]=0|0;HEAP32[$87+24>>2]=0|0;HEAP32[$87+28>>2]=0|0;
 }
 $97 = ((($9)) + 1892|0);
 $98 = HEAP32[$97>>2]|0;
 _ft_mem_free($12,$98);
 HEAP32[$97>>2] = 0;
 $99 = ((($9)) + 2932|0);
 $100 = HEAP32[$99>>2]|0;
 $101 = ($100|0)==(0|0);
 if (!($101)) {
  $102 = HEAP32[$9>>2]|0;
  _FT_Stream_ReleaseFrame($102,$99);
 }
 $103 = ((($9)) + 2936|0);
 HEAP32[$103>>2] = 0;
 $104 = ((($9)) + 2924|0);
 HEAP8[$104>>0] = 0;
 $105 = ((($9)) + 2928|0);
 HEAP32[$105>>2] = 0;
 $106 = ((($9)) + 2960|0);
 $107 = HEAP32[$106>>2]|0;
 _ft_mem_free($12,$107);
 HEAP32[$106>>2] = 0;
 $108 = ((($9)) + 1304|0);
 $109 = HEAP32[$108>>2]|0;
 _ft_mem_free($12,$109);
 HEAP32[$108>>2] = 0;
 $110 = ((($9)) + 1308|0);
 $111 = HEAP32[$110>>2]|0;
 _ft_mem_free($12,$111);
 HEAP32[$110>>2] = 0;
 $112 = ((($9)) + 1316|0);
 $113 = HEAP32[$112>>2]|0;
 _ft_mem_free($12,$113);
 HEAP32[$112>>2] = 0;
 $114 = ((($9)) + 1320|0);
 $115 = HEAP32[$114>>2]|0;
 _ft_mem_free($12,$115);
 HEAP32[$114>>2] = 0;
 $116 = ((($9)) + 2976|0);
 $117 = HEAP32[$116>>2]|0;
 $118 = ($117|0)==(0|0);
 if (!($118)) {
  $119 = ((($9)) + 2972|0);
  $120 = HEAP32[$119>>2]|0;
  FUNCTION_TABLE_vi[$117 & 8191]($120);
  $121 = HEAP32[$119>>2]|0;
  _ft_mem_free($12,$121);
  HEAP32[$119>>2] = 0;
 }
 $122 = HEAP32[$8>>2]|0;
 _ft_mem_free($2,$122);
 HEAP32[$8>>2] = 0;
 return;
}
function _cff_size_init($cffsize) {
 $cffsize = $cffsize|0;
 var $$idx$val = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $i$0 = 0, $or$cond$i = 0, $priv = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 208|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp + 196|0;
 $priv = sp;
 HEAP32[$error>>2] = 0;
 $$idx$val = HEAP32[$cffsize>>2]|0;
 $0 = ((($$idx$val)) + 656|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 2952|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($$idx$val)) + 96|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (_FT_Get_Module($7,593918)|0);
 $9 = ($8|0)!=(0|0);
 $10 = ($3|0)!=(0|0);
 $or$cond$i = $10 & $9;
 do {
  if ($or$cond$i) {
   $11 = HEAP32[$3>>2]|0;
   $12 = ($11|0)==(0|0);
   if (!($12)) {
    $13 = (FUNCTION_TABLE_ii[$11 & 8191]($8)|0);
    $14 = ($13|0)==(0|0);
    if (!($14)) {
     $15 = HEAP32[$cffsize>>2]|0;
     $16 = ((($15)) + 656|0);
     $17 = HEAP32[$16>>2]|0;
     $18 = ((($15)) + 100|0);
     $19 = HEAP32[$18>>2]|0;
     $20 = (_ft_mem_alloc($19,1028,$error)|0);
     $21 = HEAP32[$error>>2]|0;
     $22 = ($21|0)==(0);
     if ($22) {
      $23 = ((($17)) + 1324|0);
      _cff_make_private_dict($23,$priv);
      $24 = HEAP32[$13>>2]|0;
      $25 = HEAP32[$cffsize>>2]|0;
      $26 = ((($25)) + 100|0);
      $27 = HEAP32[$26>>2]|0;
      $28 = (FUNCTION_TABLE_iiii[$24 & 8191]($27,$priv,$20)|0);
      HEAP32[$error>>2] = $28;
      $29 = ($28|0)==(0);
      if ($29) {
       $30 = ((($17)) + 1896|0);
       $31 = HEAP32[$30>>2]|0;
       $32 = ((($20)) + 4|0);
       $i$0 = $31;
       while(1) {
        $33 = ($i$0|0)==(0);
        if ($33) {
         label = 9;
         break;
        }
        $34 = (($i$0) + -1)|0;
        $35 = (((($17)) + 1900|0) + ($34<<2)|0);
        $36 = HEAP32[$35>>2]|0;
        _cff_make_private_dict($36,$priv);
        $37 = HEAP32[$13>>2]|0;
        $38 = HEAP32[$cffsize>>2]|0;
        $39 = ((($38)) + 100|0);
        $40 = HEAP32[$39>>2]|0;
        $41 = (($32) + ($34<<2)|0);
        $42 = (FUNCTION_TABLE_iiii[$37 & 8191]($40,$priv,$41)|0);
        HEAP32[$error>>2] = $42;
        $43 = ($42|0)==(0);
        if ($43) {
         $i$0 = $34;
        } else {
         $$lcssa = $42;
         break;
        }
       }
       if ((label|0) == 9) {
        $44 = ((($cffsize)) + 40|0);
        HEAP32[$44>>2] = $20;
        break;
       }
       $46 = $$lcssa;
       STACKTOP = sp;return ($46|0);
      } else {
       $47 = $28;
      }
     } else {
      $47 = $21;
     }
     $46 = $47;
     STACKTOP = sp;return ($46|0);
    }
   }
  }
 } while(0);
 $45 = ((($cffsize)) + 44|0);
 HEAP32[$45>>2] = -1;
 $46 = 0;
 STACKTOP = sp;return ($46|0);
}
function _cff_make_private_dict($subfont,$priv) {
 $subfont = $subfont|0;
 $priv = $priv|0;
 var $$mask = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $9 = 0, $exitcond = 0, $exitcond22 = 0, $exitcond23 = 0, $exitcond24 = 0, $exitcond25 = 0, $exitcond26 = 0, $n$018 = 0, $n$114 = 0, $n$210 = 0, $n$36 = 0, $n$42 = 0;
 var $n$51 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _memset(($priv|0),0,196)|0;
 $0 = ((($subfont)) + 176|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1&65535;
 $3 = $1&255;
 $4 = ((($priv)) + 8|0);
 HEAP8[$4>>0] = $3;
 $5 = ($3<<24>>24)==(0);
 $6 = ($2&65535) >>> 8;
 $7 = $6&255;
 $8 = $1 >>> 16;
 $9 = $8&255;
 $10 = $1 >>> 24;
 $11 = $10&255;
 if (!($5)) {
  $12 = $1 & 255;
  $n$018 = 0;
  while(1) {
   $13 = (((($subfont)) + 180|0) + ($n$018<<2)|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = $14&65535;
   $16 = (((($priv)) + 12|0) + ($n$018<<1)|0);
   HEAP16[$16>>1] = $15;
   $17 = (($n$018) + 1)|0;
   $exitcond26 = ($17|0)==($12|0);
   if ($exitcond26) {
    break;
   } else {
    $n$018 = $17;
   }
  }
 }
 $18 = ((($priv)) + 9|0);
 HEAP8[$18>>0] = $7;
 $19 = ($7<<24>>24)==(0);
 if (!($19)) {
  $20 = $6&65535;
  $n$114 = 0;
  while(1) {
   $21 = (((($subfont)) + 236|0) + ($n$114<<2)|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = $22&65535;
   $24 = (((($priv)) + 40|0) + ($n$114<<1)|0);
   HEAP16[$24>>1] = $23;
   $25 = (($n$114) + 1)|0;
   $exitcond25 = ($25|0)==($20|0);
   if ($exitcond25) {
    break;
   } else {
    $n$114 = $25;
   }
  }
 }
 $26 = ((($priv)) + 10|0);
 HEAP8[$26>>0] = $9;
 $27 = ($9<<24>>24)==(0);
 if (!($27)) {
  $28 = $8 & 255;
  $n$210 = 0;
  while(1) {
   $29 = (((($subfont)) + 276|0) + ($n$210<<2)|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = $30&65535;
   $32 = (((($priv)) + 60|0) + ($n$210<<1)|0);
   HEAP16[$32>>1] = $31;
   $33 = (($n$210) + 1)|0;
   $exitcond24 = ($33|0)==($28|0);
   if ($exitcond24) {
    break;
   } else {
    $n$210 = $33;
   }
  }
 }
 $34 = ((($priv)) + 11|0);
 HEAP8[$34>>0] = $11;
 $35 = ($11<<24>>24)==(0);
 if (!($35)) {
  $n$36 = 0;
  while(1) {
   $36 = (((($subfont)) + 332|0) + ($n$36<<2)|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = $37&65535;
   $39 = (((($priv)) + 88|0) + ($n$36<<1)|0);
   HEAP16[$39>>1] = $38;
   $40 = (($n$36) + 1)|0;
   $exitcond23 = ($40|0)==($10|0);
   if ($exitcond23) {
    break;
   } else {
    $n$36 = $40;
   }
  }
 }
 $41 = ((($subfont)) + 372|0);
 $42 = HEAP32[$41>>2]|0;
 $43 = ((($priv)) + 108|0);
 HEAP32[$43>>2] = $42;
 $44 = ((($subfont)) + 376|0);
 $45 = HEAP32[$44>>2]|0;
 $46 = ((($priv)) + 112|0);
 HEAP32[$46>>2] = $45;
 $47 = ((($subfont)) + 380|0);
 $48 = HEAP32[$47>>2]|0;
 $49 = ((($priv)) + 116|0);
 HEAP32[$49>>2] = $48;
 $50 = ((($subfont)) + 384|0);
 $51 = HEAP32[$50>>2]|0;
 $52 = $51&65535;
 $53 = ((($priv)) + 120|0);
 HEAP16[$53>>1] = $52;
 $54 = ((($subfont)) + 388|0);
 $55 = HEAP32[$54>>2]|0;
 $56 = $55&65535;
 $57 = ((($priv)) + 122|0);
 HEAP16[$57>>1] = $56;
 $58 = ((($subfont)) + 392|0);
 $59 = HEAP16[$58>>1]|0;
 $60 = $59&255;
 $61 = ((($priv)) + 124|0);
 HEAP8[$61>>0] = $60;
 $62 = ($60<<24>>24)==(0);
 $63 = ($59&65535) >>> 8;
 $64 = $63&255;
 if (!($62)) {
  $$mask = $59 & 255;
  $65 = $$mask&65535;
  $n$42 = 0;
  while(1) {
   $66 = (((($subfont)) + 396|0) + ($n$42<<2)|0);
   $67 = HEAP32[$66>>2]|0;
   $68 = $67&65535;
   $69 = (((($priv)) + 128|0) + ($n$42<<1)|0);
   HEAP16[$69>>1] = $68;
   $70 = (($n$42) + 1)|0;
   $exitcond22 = ($70|0)==($65|0);
   if ($exitcond22) {
    break;
   } else {
    $n$42 = $70;
   }
  }
 }
 $71 = ((($priv)) + 125|0);
 HEAP8[$71>>0] = $64;
 $72 = ($64<<24>>24)==(0);
 if ($72) {
  $79 = ((($subfont)) + 500|0);
  $80 = HEAP8[$79>>0]|0;
  $81 = ((($priv)) + 126|0);
  HEAP8[$81>>0] = $80;
  $82 = ((($subfont)) + 512|0);
  $83 = HEAP32[$82>>2]|0;
  $84 = ((($priv)) + 184|0);
  HEAP32[$84>>2] = $83;
  $85 = ((($subfont)) + 508|0);
  $86 = HEAP32[$85>>2]|0;
  $87 = ((($priv)) + 4|0);
  HEAP32[$87>>2] = $86;
  return;
 }
 $73 = $63&65535;
 $n$51 = 0;
 while(1) {
  $74 = (((($subfont)) + 448|0) + ($n$51<<2)|0);
  $75 = HEAP32[$74>>2]|0;
  $76 = $75&65535;
  $77 = (((($priv)) + 154|0) + ($n$51<<1)|0);
  HEAP16[$77>>1] = $76;
  $78 = (($n$51) + 1)|0;
  $exitcond = ($78|0)==($73|0);
  if ($exitcond) {
   break;
  } else {
   $n$51 = $78;
  }
 }
 $79 = ((($subfont)) + 500|0);
 $80 = HEAP8[$79>>0]|0;
 $81 = ((($priv)) + 126|0);
 HEAP8[$81>>0] = $80;
 $82 = ((($subfont)) + 512|0);
 $83 = HEAP32[$82>>2]|0;
 $84 = ((($priv)) + 184|0);
 HEAP32[$84>>2] = $83;
 $85 = ((($subfont)) + 508|0);
 $86 = HEAP32[$85>>2]|0;
 $87 = ((($priv)) + 4|0);
 HEAP32[$87>>2] = $86;
 return;
}
function _cff_size_done($cffsize) {
 $cffsize = $cffsize|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$cffsize>>2]|0;
 $1 = ((($0)) + 656|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($cffsize)) + 40|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if ($5) {
  return;
 }
 $6 = ((($0)) + 656|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($7)) + 2952|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($0)) + 96|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($11)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (_FT_Get_Module($13,593918)|0);
 $15 = ($14|0)!=(0|0);
 $16 = ($9|0)!=(0|0);
 $or$cond$i = $16 & $15;
 if (!($or$cond$i)) {
  return;
 }
 $17 = HEAP32[$9>>2]|0;
 $18 = ($17|0)==(0|0);
 if ($18) {
  return;
 }
 $19 = (FUNCTION_TABLE_ii[$17 & 8191]($14)|0);
 $20 = ($19|0)==(0|0);
 if ($20) {
  return;
 }
 $21 = ((($19)) + 8|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = HEAP32[$4>>2]|0;
 FUNCTION_TABLE_vi[$22 & 8191]($23);
 $24 = ((($2)) + 1896|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = ($25|0)==(0);
 if ($26) {
  return;
 } else {
  $i$01 = $25;
 }
 while(1) {
  $27 = HEAP32[$21>>2]|0;
  $28 = (($i$01) + -1)|0;
  $29 = (((($4)) + 4|0) + ($28<<2)|0);
  $30 = HEAP32[$29>>2]|0;
  FUNCTION_TABLE_vi[$27 & 8191]($30);
  $31 = ($28|0)==(0);
  if ($31) {
   break;
  } else {
   $i$01 = $28;
  }
 }
 return;
}
function _cff_slot_init($slot) {
 $slot = $slot|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($slot)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 656|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 2952|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if ($6) {
  return 0;
 }
 $7 = ((($1)) + 96|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($8)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (_FT_Get_Module($10,593918)|0);
 $12 = ($11|0)==(0|0);
 if ($12) {
  return 0;
 }
 $13 = ((($5)) + 8|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (FUNCTION_TABLE_ii[$14 & 8191]($11)|0);
 $16 = ((($slot)) + 156|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ((($17)) + 36|0);
 HEAP32[$18>>2] = $15;
 return 0;
}
function _cff_slot_done($slot) {
 $slot = $slot|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($slot)) + 156|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 36|0);
 HEAP32[$2>>2] = 0;
 return;
}
function _cff_glyph_load($cffslot,$cffsize,$glyph_index,$load_flags) {
 $cffslot = $cffslot|0;
 $cffsize = $cffsize|0;
 $glyph_index = $glyph_index|0;
 $load_flags = $load_flags|0;
 var $$$i = 0, $$$i$i$i = 0, $$$i$i11$i = 0, $$0 = 0, $$0$i$i = 0, $$02$i = 0, $$3$i = 0, $$48$i = 0, $$lcssa = 0, $$lcssa35 = 0, $$load_flags = 0, $$lobit$i = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phi$iZ2D = 0, $$sink38$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0;
 var $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0;
 var $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0;
 var $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0;
 var $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0;
 var $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0;
 var $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0;
 var $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0;
 var $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0;
 var $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0;
 var $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0;
 var $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0;
 var $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0;
 var $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0;
 var $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0;
 var $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0;
 var $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0;
 var $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0;
 var $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0;
 var $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $advance$i = 0, $advance4$i = 0, $cbox$i = 0, $charstring$i = 0;
 var $charstring_len$i = 0, $data$i$i = 0, $data$i13$i = 0, $decoder$i = 0, $dummy$i = 0, $error$0$i = 0, $error$1$ph$i = 0, $fd$1$i$i = 0, $fd$1$i$i$i = 0, $fd_index$0$i = 0, $first$0$i$i = 0, $first$0$i$i$i = 0, $first$0$i$i$i$lcssa32 = 0, $first$0$i$i$lcssa34 = 0, $font_matrix$i = 0, $font_offset$sroa$0$0$i = 0, $font_offset$sroa$8$0$i = 0, $force_scaling$1$i = 0, $force_scaling$2$i = 0, $hinting$0$i = 0;
 var $horiAdvance$i = 0, $horiBearingX$i = 0, $load_flags$$i = 0, $metrics$i = 0, $metrics2$i = 0, $n$029$i = 0, $or$cond$i = 0, $or$cond$i$i = 0, $or$cond$i$i$i = 0, $or$cond10$i = 0, $p$0$i$i = 0, $p$0$i$i$i = 0, $p$0$i$i$i$lcssa31 = 0, $p$0$i$i$lcssa33 = 0, $result$0$i$i$i = 0, $result$0$i$i12$i = 0, $sext$i = 0, $sext40$i = 0, $size$0 = 0, $sub$1$i$i = 0;
 var $vec$030$i = 0, $vertAdvance$i = 0, $vertBearingY$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 864|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $data$i13$i = sp + 816|0;
 $data$i$i = sp + 808|0;
 $decoder$i = sp + 64|0;
 $font_matrix$i = sp + 48|0;
 $metrics$i = sp + 840|0;
 $advance$i = sp + 834|0;
 $dummy$i = sp + 832|0;
 $charstring$i = sp + 44|0;
 $charstring_len$i = sp + 40|0;
 $metrics2$i = sp + 24|0;
 $cbox$i = sp + 8|0;
 $advance4$i = sp;
 $horiBearingX$i = sp + 830|0;
 $horiAdvance$i = sp + 828|0;
 $vertBearingY$i = sp + 826|0;
 $vertAdvance$i = sp + 824|0;
 $0 = ($cffslot|0)==(0|0);
 if ($0) {
  $$0 = 37;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ($cffsize|0)==(0|0);
 $2 = $load_flags | 3;
 $$load_flags = $1 ? $2 : $load_flags;
 $3 = $$load_flags & 1;
 $4 = ($3|0)==(0);
 $size$0 = $4 ? $cffsize : 0;
 $5 = ($size$0|0)==(0|0);
 do {
  if ($5) {
   $$phi$trans$insert = ((($cffslot)) + 4|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0;
   $11 = $$pre;
  } else {
   $6 = HEAP32[$cffsize>>2]|0;
   $7 = ((($cffslot)) + 4|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ($6|0)==($8|0);
   if ($9) {
    $11 = $8;
    break;
   } else {
    $$0 = 35;
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 $10 = ((($11)) + 656|0);
 $12 = HEAP32[$10>>2]|0;
 $13 = ((($12)) + 1456|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ($14|0)==(65535);
 if ($15) {
  label = 11;
 } else {
  $16 = ((($12)) + 1164|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = ($17|0)==(0|0);
  if ($18) {
   label = 11;
  } else {
   $19 = ($glyph_index|0)==(0);
   if ($19) {
    $$02$i = 0;
    label = 12;
   } else {
    $20 = ((($12)) + 1168|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = ($21>>>0)<($glyph_index>>>0);
    if ($22) {
     $$3$i = 6;
    } else {
     $23 = (($17) + ($glyph_index<<1)|0);
     $24 = HEAP16[$23>>1]|0;
     $25 = $24&65535;
     $26 = ($24<<16>>16)==(0);
     if ($26) {
      $$3$i = 6;
     } else {
      $$02$i = $25;
      label = 12;
     }
    }
   }
  }
 }
 if ((label|0) == 11) {
  $27 = ((($12)) + 12|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = ($28>>>0)>($glyph_index>>>0);
  if ($29) {
   $$02$i = $glyph_index;
   label = 12;
  } else {
   $$3$i = 6;
  }
 }
 L18: do {
  if ((label|0) == 12) {
   $30 = $$load_flags & 1024;
   $31 = ($30|0)==(0);
   $load_flags$$i = $31 ? $$load_flags : $2;
   $32 = ((($cffslot)) + 164|0);
   HEAP32[$32>>2] = 65536;
   $33 = ((($cffslot)) + 168|0);
   HEAP32[$33>>2] = 65536;
   $34 = ($size$0|0)!=(0|0);
   do {
    if ($34) {
     $35 = ((($size$0)) + 16|0);
     $36 = HEAP32[$35>>2]|0;
     HEAP32[$32>>2] = $36;
     $37 = ((($size$0)) + 20|0);
     $38 = HEAP32[$37>>2]|0;
     HEAP32[$33>>2] = $38;
     $39 = HEAP32[$size$0>>2]|0;
     $40 = ((($39)) + 532|0);
     $41 = HEAP32[$40>>2]|0;
     $42 = ((($39)) + 104|0);
     $43 = HEAP32[$42>>2]|0;
     $44 = ((($size$0)) + 44|0);
     $45 = HEAP32[$44>>2]|0;
     $46 = ($45|0)==(-1);
     if (!($46)) {
      $47 = ((($41)) + 96|0);
      $48 = HEAP32[$47>>2]|0;
      $49 = ($48|0)!=(0|0);
      $50 = $load_flags$$i & 8;
      $51 = ($50|0)==(0);
      $or$cond$i = $51 & $49;
      if ($or$cond$i) {
       $52 = ((($41)) + 72|0);
       $53 = HEAP32[$52>>2]|0;
       $54 = ((($cffslot)) + 76|0);
       $55 = (FUNCTION_TABLE_iiiiiiii[$53 & 8191]($11,$45,$$02$i,$load_flags$$i,$43,$54,$metrics$i)|0);
       $56 = ($55|0)==(0);
       if (!($56)) {
        break;
       }
       $57 = ((($cffslot)) + 110|0);
       HEAP16[$57>>1] = 0;
       $58 = ((($cffslot)) + 108|0);
       HEAP16[$58>>1] = 0;
       $59 = ((($metrics$i)) + 2|0);
       $60 = HEAP16[$59>>1]|0;
       $61 = $60&65535;
       $62 = $61 << 6;
       $63 = ((($cffslot)) + 24|0);
       HEAP32[$63>>2] = $62;
       $64 = HEAP16[$metrics$i>>1]|0;
       $65 = $64&65535;
       $66 = $65 << 6;
       $67 = ((($cffslot)) + 28|0);
       HEAP32[$67>>2] = $66;
       $68 = ((($metrics$i)) + 4|0);
       $69 = HEAP16[$68>>1]|0;
       $70 = $69 << 16 >> 16;
       $71 = $70 << 6;
       $72 = ((($cffslot)) + 32|0);
       HEAP32[$72>>2] = $71;
       $73 = ((($metrics$i)) + 6|0);
       $74 = HEAP16[$73>>1]|0;
       $75 = $74 << 16 >> 16;
       $76 = $75 << 6;
       $77 = ((($cffslot)) + 36|0);
       HEAP32[$77>>2] = $76;
       $78 = ((($metrics$i)) + 8|0);
       $79 = HEAP16[$78>>1]|0;
       $80 = $79&65535;
       $81 = $80 << 6;
       $82 = ((($cffslot)) + 40|0);
       HEAP32[$82>>2] = $81;
       $83 = ((($metrics$i)) + 10|0);
       $84 = HEAP16[$83>>1]|0;
       $85 = $84 << 16 >> 16;
       $86 = $85 << 6;
       $87 = ((($cffslot)) + 44|0);
       HEAP32[$87>>2] = $86;
       $88 = ((($metrics$i)) + 12|0);
       $89 = HEAP16[$88>>1]|0;
       $90 = $89 << 16 >> 16;
       $91 = $90 << 6;
       $92 = ((($cffslot)) + 48|0);
       HEAP32[$92>>2] = $91;
       $93 = ((($metrics$i)) + 14|0);
       $94 = HEAP16[$93>>1]|0;
       $95 = $94&65535;
       $96 = $95 << 6;
       $97 = ((($cffslot)) + 52|0);
       HEAP32[$97>>2] = $96;
       $98 = ((($cffslot)) + 72|0);
       HEAP32[$98>>2] = 1651078259;
       $99 = $load_flags$$i & 16;
       $100 = ($99|0)==(0);
       $$$i = $100 ? $70 : $85;
       $$48$i = $100 ? $75 : $90;
       $101 = ((($cffslot)) + 100|0);
       HEAP32[$101>>2] = $$$i;
       $102 = ((($cffslot)) + 104|0);
       HEAP32[$102>>2] = $$48$i;
       $103 = ((($11)) + 532|0);
       $104 = HEAP32[$103>>2]|0;
       $105 = ((($104)) + 112|0);
       $106 = HEAP32[$105>>2]|0;
       FUNCTION_TABLE_viiiii[$106 & 8191]($11,0,$$02$i,$dummy$i,$advance$i);
       $107 = HEAP16[$advance$i>>1]|0;
       $108 = $107&65535;
       $109 = ((($cffslot)) + 56|0);
       HEAP32[$109>>2] = $108;
       $110 = ((($11)) + 292|0);
       $111 = HEAP8[$110>>0]|0;
       $112 = ($111<<24>>24)==(0);
       if ($112) {
        label = 20;
       } else {
        $113 = ((($11)) + 330|0);
        $114 = HEAP16[$113>>1]|0;
        $115 = ($114<<16>>16)==(0);
        if ($115) {
         label = 20;
        } else {
         $116 = HEAP32[$103>>2]|0;
         $117 = ((($116)) + 112|0);
         $118 = HEAP32[$117>>2]|0;
         FUNCTION_TABLE_viiiii[$118 & 8191]($11,1,$$02$i,$dummy$i,$advance$i);
         $119 = HEAP16[$advance$i>>1]|0;
         $120 = $119&65535;
         $121 = ((($cffslot)) + 60|0);
         HEAP32[$121>>2] = $120;
        }
       }
       do {
        if ((label|0) == 20) {
         $122 = ((($11)) + 364|0);
         $123 = HEAP16[$122>>1]|0;
         $124 = ($123<<16>>16)==(-1);
         if ($124) {
          $133 = ((($11)) + 220|0);
          $134 = HEAP32[$133>>2]|0;
          $sext40$i = $134 << 16;
          $135 = $sext40$i >> 16;
          $136 = $134 >> 16;
          $137 = (($135) - ($136))|0;
          $138 = ((($cffslot)) + 60|0);
          HEAP32[$138>>2] = $137;
          break;
         } else {
          $125 = ((($11)) + 434|0);
          $126 = HEAP16[$125>>1]|0;
          $127 = $126 << 16 >> 16;
          $128 = ((($11)) + 436|0);
          $129 = HEAP16[$128>>1]|0;
          $130 = $129 << 16 >> 16;
          $131 = (($127) - ($130))|0;
          $132 = ((($cffslot)) + 60|0);
          HEAP32[$132>>2] = $131;
          break;
         }
        }
       } while(0);
       $$3$i = 0;
       break L18;
      }
     }
    }
   } while(0);
   $139 = $load_flags$$i & 16384;
   $140 = ($139|0)==(0);
   if ($140) {
    $141 = ((($12)) + 1896|0);
    $142 = HEAP32[$141>>2]|0;
    $143 = ($142|0)==(0);
    if ($143) {
     $209 = ((($12)) + 1372|0);
     ;HEAP32[$font_matrix$i>>2]=HEAP32[$209>>2]|0;HEAP32[$font_matrix$i+4>>2]=HEAP32[$209+4>>2]|0;HEAP32[$font_matrix$i+8>>2]=HEAP32[$209+8>>2]|0;HEAP32[$font_matrix$i+12>>2]=HEAP32[$209+12>>2]|0;
     $210 = ((($12)) + 1396|0);
     $211 = HEAP32[$210>>2]|0;
     $212 = ((($12)) + 1400|0);
     $213 = HEAP32[$212>>2]|0;
     $font_offset$sroa$0$0$i = $211;$font_offset$sroa$8$0$i = $213;$force_scaling$1$i = 0;
    } else {
     $144 = ((($12)) + 2924|0);
     $145 = HEAP8[$144>>0]|0;
     $146 = $145&255;
     L41: do {
      switch ($146|0) {
      case 0:  {
       $147 = ((($12)) + 2932|0);
       $148 = HEAP32[$147>>2]|0;
       $149 = (($148) + ($$02$i)|0);
       $150 = HEAP8[$149>>0]|0;
       $fd$1$i$i = $150;
       break;
      }
      case 3:  {
       $151 = ((($12)) + 2940|0);
       $152 = HEAP32[$151>>2]|0;
       $153 = (($$02$i) - ($152))|0;
       $154 = ((($12)) + 2944|0);
       $155 = HEAP32[$154>>2]|0;
       $156 = ($153>>>0)<($155>>>0);
       if ($156) {
        $157 = ((($12)) + 2948|0);
        $158 = HEAP8[$157>>0]|0;
        $fd$1$i$i = $158;
        break L41;
       }
       $159 = ((($12)) + 2932|0);
       $160 = HEAP32[$159>>2]|0;
       $161 = ((($12)) + 2936|0);
       $162 = HEAP32[$161>>2]|0;
       $163 = (($160) + ($162)|0);
       $164 = ((($160)) + 2|0);
       $165 = HEAP8[$160>>0]|0;
       $166 = $165&255;
       $167 = $166 << 8;
       $168 = ((($160)) + 1|0);
       $169 = HEAP8[$168>>0]|0;
       $170 = $169&255;
       $171 = $167 | $170;
       $first$0$i$i = $171;$p$0$i$i = $164;
       while(1) {
        $172 = ($first$0$i$i>>>0)>($$02$i>>>0);
        if ($172) {
         $fd$1$i$i = 0;
         break L41;
        }
        $173 = ((($p$0$i$i)) + 3|0);
        $174 = ((($p$0$i$i)) + 1|0);
        $175 = HEAP8[$174>>0]|0;
        $176 = $175&255;
        $177 = $176 << 8;
        $178 = ((($p$0$i$i)) + 2|0);
        $179 = HEAP8[$178>>0]|0;
        $180 = $179&255;
        $181 = $177 | $180;
        $182 = ($181>>>0)>($$02$i>>>0);
        if ($182) {
         $$lcssa35 = $181;$first$0$i$i$lcssa34 = $first$0$i$i;$p$0$i$i$lcssa33 = $p$0$i$i;
         break;
        }
        $186 = ($173>>>0)<($163>>>0);
        if ($186) {
         $first$0$i$i = $181;$p$0$i$i = $173;
        } else {
         $fd$1$i$i = 0;
         break L41;
        }
       }
       $183 = HEAP8[$p$0$i$i$lcssa33>>0]|0;
       HEAP32[$151>>2] = $first$0$i$i$lcssa34;
       $184 = (($$lcssa35) - ($first$0$i$i$lcssa34))|0;
       HEAP32[$154>>2] = $184;
       $185 = ((($12)) + 2948|0);
       HEAP8[$185>>0] = $183;
       $fd$1$i$i = $183;
       break;
      }
      default: {
       $fd$1$i$i = 0;
      }
      }
     } while(0);
     $187 = $fd$1$i$i&255;
     $188 = ($187>>>0)<($142>>>0);
     $189 = (($142) + 255)|0;
     $190 = $189&255;
     $fd_index$0$i = $188 ? $fd$1$i$i : $190;
     $191 = ((($12)) + 1392|0);
     $192 = HEAP32[$191>>2]|0;
     $193 = $fd_index$0$i&255;
     $194 = (((($12)) + 1900|0) + ($193<<2)|0);
     $195 = HEAP32[$194>>2]|0;
     $196 = ((($195)) + 68|0);
     $197 = HEAP32[$196>>2]|0;
     $198 = ((($195)) + 48|0);
     ;HEAP32[$font_matrix$i>>2]=HEAP32[$198>>2]|0;HEAP32[$font_matrix$i+4>>2]=HEAP32[$198+4>>2]|0;HEAP32[$font_matrix$i+8>>2]=HEAP32[$198+8>>2]|0;HEAP32[$font_matrix$i+12>>2]=HEAP32[$198+12>>2]|0;
     $199 = HEAP32[$194>>2]|0;
     $200 = ((($199)) + 72|0);
     $201 = HEAP32[$200>>2]|0;
     $202 = ((($199)) + 76|0);
     $203 = HEAP32[$202>>2]|0;
     $204 = ($192|0)==($197|0);
     if ($204) {
      $font_offset$sroa$0$0$i = $201;$font_offset$sroa$8$0$i = $203;$force_scaling$1$i = 0;
     } else {
      $205 = HEAP32[$32>>2]|0;
      $206 = (_FT_MulDiv($205,$192,$197)|0);
      HEAP32[$32>>2] = $206;
      $207 = HEAP32[$33>>2]|0;
      $208 = (_FT_MulDiv($207,$192,$197)|0);
      HEAP32[$33>>2] = $208;
      $font_offset$sroa$0$0$i = $201;$font_offset$sroa$8$0$i = $203;$force_scaling$1$i = 1;
     }
    }
    $214 = ((($cffslot)) + 108|0);
    $215 = ((($cffslot)) + 110|0);
    HEAP16[$215>>1] = 0;
    HEAP16[$214>>1] = 0;
    $216 = $load_flags$$i >>> 1;
    $$lobit$i = $216 & 1;
    $217 = $$lobit$i ^ 1;
    $218 = $217&255;
    $219 = $load_flags$$i & 1;
    $220 = ($219|0)!=(0);
    $221 = $219 ^ 1;
    $222 = $221&255;
    $223 = ((($cffslot)) + 160|0);
    HEAP8[$223>>0] = $218;
    $224 = ((($cffslot)) + 161|0);
    HEAP8[$224>>0] = $222;
    $225 = ((($cffslot)) + 72|0);
    HEAP32[$225>>2] = 1869968492;
    $226 = $load_flags$$i >>> 16;
    $227 = $226 & 15;
    $228 = HEAP32[$10>>2]|0;
    $229 = ((($decoder$i)) + 65|0);
    _memset(($decoder$i|0),0,740)|0;
    HEAP8[$229>>0] = 1;
    $230 = ((($decoder$i)) + 4|0);
    HEAP32[$230>>2] = $11;
    $231 = ((($decoder$i)) + 8|0);
    HEAP32[$231>>2] = $cffslot;
    $232 = ((($11)) + 100|0);
    $233 = HEAP32[$232>>2]|0;
    HEAP32[$decoder$i>>2] = $233;
    $234 = ((($cffslot)) + 156|0);
    $235 = HEAP32[$234>>2]|0;
    $236 = HEAP32[$235>>2]|0;
    $237 = ((($decoder$i)) + 12|0);
    HEAP32[$237>>2] = $236;
    $238 = ((($236)) + 20|0);
    $239 = ((($decoder$i)) + 16|0);
    HEAP32[$239>>2] = $238;
    $240 = ((($236)) + 56|0);
    $241 = ((($decoder$i)) + 20|0);
    HEAP32[$241>>2] = $240;
    _FT_GlyphLoader_Rewind($236);
    $242 = ((($decoder$i)) + 72|0);
    HEAP32[$242>>2] = 0;
    $243 = ((($decoder$i)) + 68|0);
    HEAP32[$243>>2] = 0;
    $244 = ($218<<24>>24)!=(0);
    $or$cond$i$i$i = $34 & $244;
    if ($or$cond$i$i$i) {
     $245 = ((($size$0)) + 40|0);
     $246 = HEAP32[$245>>2]|0;
     $247 = HEAP32[$246>>2]|0;
     HEAP32[$242>>2] = $247;
     $248 = HEAP32[$234>>2]|0;
     $249 = ((($248)) + 36|0);
     $250 = HEAP32[$249>>2]|0;
     HEAP32[$243>>2] = $250;
    }
    $251 = ((($decoder$i)) + 24|0);
    ;HEAP32[$251>>2]=0|0;HEAP32[$251+4>>2]=0|0;HEAP32[$251+8>>2]=0|0;HEAP32[$251+12>>2]=0|0;HEAP32[$251+16>>2]=0|0;HEAP32[$251+20>>2]=0|0;
    $252 = ((($decoder$i)) + 76|0);
    HEAP32[$252>>2] = $228;
    $253 = ((($228)) + 92|0);
    $254 = HEAP32[$253>>2]|0;
    $255 = ((($decoder$i)) + 700|0);
    HEAP32[$255>>2] = $254;
    $256 = ((($228)) + 1308|0);
    $257 = HEAP32[$256>>2]|0;
    $258 = ((($decoder$i)) + 716|0);
    HEAP32[$258>>2] = $257;
    $259 = ((($228)) + 1368|0);
    $260 = HEAP32[$259>>2]|0;
    $261 = ($260|0)==(1);
    if ($261) {
     $result$0$i$i$i = 0;
    } else {
     $262 = ($254>>>0)<(1240);
     if ($262) {
      $result$0$i$i$i = 107;
     } else {
      $263 = ($254>>>0)<(33900);
      $$$i$i$i = $263 ? 1131 : 32768;
      $result$0$i$i$i = $$$i$i$i;
     }
    }
    $264 = ((($decoder$i)) + 708|0);
    HEAP32[$264>>2] = $result$0$i$i$i;
    $265 = ((($decoder$i)) + 728|0);
    HEAP32[$265>>2] = $227;
    $266 = $load_flags$$i & 256;
    $267 = ($266|0)==(0);
    if (!($267)) {
     $268 = ((($decoder$i)) + 561|0);
     HEAP8[$268>>0] = 1;
    }
    $269 = $load_flags$$i & 1024;
    $270 = ((($decoder$i)) + 66|0);
    HEAP8[$270>>0] = 0;
    $271 = ((($11)) + 128|0);
    $272 = HEAP32[$271>>2]|0;
    $273 = ((($272)) + 48|0);
    $274 = HEAP32[$273>>2]|0;
    $275 = ($274|0)==(0|0);
    if ($275) {
     $284 = HEAP32[$10>>2]|0;
     $285 = ((($284)) + 1176|0);
     $286 = (_cff_index_access_element($285,$$02$i,$charstring$i,$charstring_len$i)|0);
     $$0$i$i = $286;
    } else {
     $276 = HEAP32[$274>>2]|0;
     $277 = HEAP32[$276>>2]|0;
     $278 = ((($274)) + 4|0);
     $279 = HEAP32[$278>>2]|0;
     $280 = (FUNCTION_TABLE_iiii[$277 & 8191]($279,$$02$i,$data$i$i)|0);
     $281 = HEAP32[$data$i$i>>2]|0;
     HEAP32[$charstring$i>>2] = $281;
     $282 = ((($data$i$i)) + 4|0);
     $283 = HEAP32[$282>>2]|0;
     HEAP32[$charstring_len$i>>2] = $283;
     $$0$i$i = $280;
    }
    $287 = ($$0$i$i|0)==(0);
    do {
     if ($287) {
      $288 = HEAP32[$230>>2]|0;
      $289 = ((($288)) + 656|0);
      $290 = HEAP32[$289>>2]|0;
      $291 = ((($290)) + 1324|0);
      $292 = ((($290)) + 1896|0);
      $293 = HEAP32[$292>>2]|0;
      $294 = ($293|0)==(0);
      if ($294) {
       $sub$1$i$i = $291;
      } else {
       $295 = ((($290)) + 2924|0);
       $296 = HEAP8[$295>>0]|0;
       $297 = $296&255;
       L73: do {
        switch ($297|0) {
        case 0:  {
         $298 = ((($290)) + 2932|0);
         $299 = HEAP32[$298>>2]|0;
         $300 = (($299) + ($$02$i)|0);
         $301 = HEAP8[$300>>0]|0;
         $fd$1$i$i$i = $301;
         break;
        }
        case 3:  {
         $302 = ((($290)) + 2940|0);
         $303 = HEAP32[$302>>2]|0;
         $304 = (($$02$i) - ($303))|0;
         $305 = ((($290)) + 2944|0);
         $306 = HEAP32[$305>>2]|0;
         $307 = ($304>>>0)<($306>>>0);
         if ($307) {
          $308 = ((($290)) + 2948|0);
          $309 = HEAP8[$308>>0]|0;
          $fd$1$i$i$i = $309;
          break L73;
         }
         $310 = ((($290)) + 2932|0);
         $311 = HEAP32[$310>>2]|0;
         $312 = ((($290)) + 2936|0);
         $313 = HEAP32[$312>>2]|0;
         $314 = (($311) + ($313)|0);
         $315 = ((($311)) + 2|0);
         $316 = HEAP8[$311>>0]|0;
         $317 = $316&255;
         $318 = $317 << 8;
         $319 = ((($311)) + 1|0);
         $320 = HEAP8[$319>>0]|0;
         $321 = $320&255;
         $322 = $318 | $321;
         $first$0$i$i$i = $322;$p$0$i$i$i = $315;
         while(1) {
          $323 = ($first$0$i$i$i>>>0)>($$02$i>>>0);
          if ($323) {
           $fd$1$i$i$i = 0;
           break L73;
          }
          $324 = ((($p$0$i$i$i)) + 3|0);
          $325 = ((($p$0$i$i$i)) + 1|0);
          $326 = HEAP8[$325>>0]|0;
          $327 = $326&255;
          $328 = $327 << 8;
          $329 = ((($p$0$i$i$i)) + 2|0);
          $330 = HEAP8[$329>>0]|0;
          $331 = $330&255;
          $332 = $328 | $331;
          $333 = ($332>>>0)>($$02$i>>>0);
          if ($333) {
           $$lcssa = $332;$first$0$i$i$i$lcssa32 = $first$0$i$i$i;$p$0$i$i$i$lcssa31 = $p$0$i$i$i;
           break;
          }
          $337 = ($324>>>0)<($314>>>0);
          if ($337) {
           $first$0$i$i$i = $332;$p$0$i$i$i = $324;
          } else {
           $fd$1$i$i$i = 0;
           break L73;
          }
         }
         $334 = HEAP8[$p$0$i$i$i$lcssa31>>0]|0;
         HEAP32[$302>>2] = $first$0$i$i$i$lcssa32;
         $335 = (($$lcssa) - ($first$0$i$i$i$lcssa32))|0;
         HEAP32[$305>>2] = $335;
         $336 = ((($290)) + 2948|0);
         HEAP8[$336>>0] = $334;
         $fd$1$i$i$i = $334;
         break;
        }
        default: {
         $fd$1$i$i$i = 0;
        }
        }
       } while(0);
       $338 = $fd$1$i$i$i&255;
       $339 = ($338>>>0)<($293>>>0);
       if (!($339)) {
        $error$1$ph$i = 3;
        break;
       }
       $340 = (((($290)) + 1900|0) + ($338<<2)|0);
       $341 = HEAP32[$340>>2]|0;
       $342 = ((($decoder$i)) + 68|0);
       $343 = HEAP32[$342>>2]|0;
       $344 = ($343|0)!=(0|0);
       $or$cond$i$i = $34 & $344;
       if ($or$cond$i$i) {
        $345 = ((($size$0)) + 40|0);
        $346 = HEAP32[$345>>2]|0;
        $347 = (((($346)) + 4|0) + ($338<<2)|0);
        $348 = HEAP32[$347>>2]|0;
        $349 = ((($decoder$i)) + 72|0);
        HEAP32[$349>>2] = $348;
        $sub$1$i$i = $341;
       } else {
        $sub$1$i$i = $341;
       }
      }
      $350 = ((($sub$1$i$i)) + 544|0);
      $351 = HEAP32[$350>>2]|0;
      $352 = ((($decoder$i)) + 696|0);
      HEAP32[$352>>2] = $351;
      $353 = ((($sub$1$i$i)) + 568|0);
      $354 = HEAP32[$353>>2]|0;
      $355 = ((($decoder$i)) + 712|0);
      HEAP32[$355>>2] = $354;
      $356 = HEAP32[$252>>2]|0;
      $357 = ((($356)) + 1368|0);
      $358 = HEAP32[$357>>2]|0;
      $359 = ($358|0)==(1);
      if ($359) {
       $result$0$i$i12$i = 0;
      } else {
       $360 = ($351>>>0)<(1240);
       if ($360) {
        $result$0$i$i12$i = 107;
       } else {
        $361 = ($351>>>0)<(33900);
        $$$i$i11$i = $361 ? 1131 : 32768;
        $result$0$i$i12$i = $$$i$i11$i;
       }
      }
      $362 = ((($decoder$i)) + 704|0);
      HEAP32[$362>>2] = $result$0$i$i12$i;
      $363 = ((($sub$1$i$i)) + 528|0);
      $364 = HEAP32[$363>>2]|0;
      $365 = ((($decoder$i)) + 552|0);
      HEAP32[$365>>2] = $364;
      $366 = ((($sub$1$i$i)) + 532|0);
      $367 = HEAP32[$366>>2]|0;
      $368 = ((($decoder$i)) + 556|0);
      HEAP32[$368>>2] = $367;
      $369 = ((($decoder$i)) + 736|0);
      HEAP32[$369>>2] = $sub$1$i$i;
      $370 = HEAP32[$charstring$i>>2]|0;
      $371 = HEAP32[$charstring_len$i>>2]|0;
      $372 = (_cf2_decoder_parse_charstrings($decoder$i,$370,$371)|0);
      $373 = $372 & 255;
      $374 = ($373|0)==(164);
      if ($374) {
       HEAP8[$223>>0] = 0;
       $375 = HEAP32[$charstring$i>>2]|0;
       $376 = (_cf2_decoder_parse_charstrings($decoder$i,$375,$371)|0);
       $error$0$i = $376;$force_scaling$2$i = 1;$hinting$0$i = 0;
      } else {
       $error$0$i = $372;$force_scaling$2$i = $force_scaling$1$i;$hinting$0$i = $218;
      }
      $377 = HEAP32[$271>>2]|0;
      $378 = ((($377)) + 48|0);
      $379 = HEAP32[$378>>2]|0;
      $380 = ($379|0)==(0|0);
      if ($380) {
       $388 = HEAP32[$10>>2]|0;
       $389 = ((($388)) + 1204|0);
       $390 = HEAP32[$389>>2]|0;
       $391 = ($390|0)==(0|0);
       if ($391) {
        $392 = ((($388)) + 1176|0);
        $393 = HEAP32[$392>>2]|0;
        _FT_Stream_ReleaseFrame($393,$charstring$i);
       }
      } else {
       $381 = HEAP32[$charstring$i>>2]|0;
       HEAP32[$data$i13$i>>2] = $381;
       $382 = ((($data$i13$i)) + 4|0);
       HEAP32[$382>>2] = $371;
       $383 = HEAP32[$379>>2]|0;
       $384 = ((($383)) + 4|0);
       $385 = HEAP32[$384>>2]|0;
       $386 = ((($379)) + 4|0);
       $387 = HEAP32[$386>>2]|0;
       FUNCTION_TABLE_vii[$385 & 8191]($387,$data$i13$i);
      }
      $394 = ($error$0$i|0)==(0);
      if ($394) {
       $395 = HEAP32[$271>>2]|0;
       $396 = ((($395)) + 48|0);
       $397 = HEAP32[$396>>2]|0;
       $398 = ($397|0)==(0|0);
       do {
        if ($398) {
         $401 = ((($12)) + 1200|0);
         $402 = HEAP32[$401>>2]|0;
         $403 = ($402|0)==(0|0);
         if ($403) {
          break;
         }
         $404 = ((($12)) + 1204|0);
         $405 = HEAP32[$404>>2]|0;
         $406 = (($402) + ($$02$i<<2)|0);
         $407 = HEAP32[$406>>2]|0;
         $408 = (($405) + ($407)|0);
         $409 = ((($408)) + -1|0);
         $410 = ((($cffslot)) + 136|0);
         HEAP32[$410>>2] = $409;
         $411 = ((($cffslot)) + 140|0);
         HEAP32[$411>>2] = $371;
        } else {
         $399 = ((($cffslot)) + 136|0);
         HEAP32[$399>>2] = 0;
         $400 = ((($cffslot)) + 140|0);
         HEAP32[$400>>2] = 0;
        }
       } while(0);
       $412 = HEAP32[$231>>2]|0;
       $413 = ($412|0)==(0|0);
       if (!($413)) {
        $414 = ((($412)) + 108|0);
        $415 = ((($decoder$i)) + 16|0);
        $416 = HEAP32[$415>>2]|0;
        ;HEAP32[$414>>2]=HEAP32[$416>>2]|0;HEAP32[$414+4>>2]=HEAP32[$416+4>>2]|0;HEAP32[$414+8>>2]=HEAP32[$416+8>>2]|0;HEAP32[$414+12>>2]=HEAP32[$416+12>>2]|0;HEAP32[$414+16>>2]=HEAP32[$416+16>>2]|0;
       }
       $417 = HEAP32[$271>>2]|0;
       $418 = ((($417)) + 48|0);
       $419 = HEAP32[$418>>2]|0;
       $420 = ($419|0)==(0|0);
       do {
        if (!($420)) {
         $421 = HEAP32[$419>>2]|0;
         $422 = ((($421)) + 8|0);
         $423 = HEAP32[$422>>2]|0;
         $424 = ($423|0)==(0|0);
         if ($424) {
          break;
         }
         $425 = ((($decoder$i)) + 32|0);
         $426 = HEAP32[$425>>2]|0;
         HEAP32[$metrics2$i>>2] = $426;
         $427 = ((($metrics2$i)) + 4|0);
         HEAP32[$427>>2] = 0;
         $428 = ((($decoder$i)) + 40|0);
         $429 = HEAP32[$428>>2]|0;
         $430 = ((($metrics2$i)) + 8|0);
         HEAP32[$430>>2] = $429;
         $431 = ((($decoder$i)) + 44|0);
         $432 = HEAP32[$431>>2]|0;
         $433 = ((($metrics2$i)) + 12|0);
         HEAP32[$433>>2] = $432;
         $434 = ((($419)) + 4|0);
         $435 = HEAP32[$434>>2]|0;
         $436 = (FUNCTION_TABLE_iiiii[$423 & 8191]($435,$$02$i,0,$metrics2$i)|0);
         $437 = HEAP32[$metrics2$i>>2]|0;
         HEAP32[$425>>2] = $437;
         $438 = HEAP32[$430>>2]|0;
         HEAP32[$428>>2] = $438;
         $439 = HEAP32[$433>>2]|0;
         HEAP32[$431>>2] = $439;
         $440 = ($436|0)==(0);
         if (!($440)) {
          $$3$i = $436;
          break L18;
         }
        }
       } while(0);
       $441 = ($269|0)==(0);
       if (!($441)) {
        $442 = ((($cffslot)) + 156|0);
        $443 = HEAP32[$442>>2]|0;
        $444 = ((($decoder$i)) + 32|0);
        $445 = HEAP32[$444>>2]|0;
        $446 = ((($cffslot)) + 32|0);
        HEAP32[$446>>2] = $445;
        $447 = HEAP32[$365>>2]|0;
        $448 = ((($cffslot)) + 40|0);
        HEAP32[$448>>2] = $447;
        $449 = ((($443)) + 12|0);
        ;HEAP32[$449>>2]=HEAP32[$font_matrix$i>>2]|0;HEAP32[$449+4>>2]=HEAP32[$font_matrix$i+4>>2]|0;HEAP32[$449+8>>2]=HEAP32[$font_matrix$i+8>>2]|0;HEAP32[$449+12>>2]=HEAP32[$font_matrix$i+12>>2]|0;
        $450 = ((($443)) + 28|0);
        HEAP32[$450>>2] = $font_offset$sroa$0$0$i;
        $451 = ((($443)) + 32|0);
        HEAP32[$451>>2] = $font_offset$sroa$8$0$i;
        $452 = ((($443)) + 8|0);
        HEAP8[$452>>0] = 1;
        $$3$i = 0;
        break L18;
       }
       $453 = ((($cffslot)) + 24|0);
       $454 = ((($11)) + 250|0);
       $455 = HEAP16[$454>>1]|0;
       $456 = ($455<<16>>16)==(0);
       if ($456) {
        $466 = HEAP32[$365>>2]|0;
        $$sink38$i = $466;
       } else {
        HEAP16[$horiBearingX$i>>1] = 0;
        HEAP16[$horiAdvance$i>>1] = 0;
        $457 = ((($11)) + 532|0);
        $458 = HEAP32[$457>>2]|0;
        $459 = ((($458)) + 112|0);
        $460 = HEAP32[$459>>2]|0;
        FUNCTION_TABLE_viiiii[$460 & 8191]($11,0,$$02$i,$horiBearingX$i,$horiAdvance$i);
        $461 = HEAP16[$horiAdvance$i>>1]|0;
        $462 = $461&65535;
        $463 = HEAP16[$horiBearingX$i>>1]|0;
        $464 = $463 << 16 >> 16;
        $465 = ((($cffslot)) + 32|0);
        HEAP32[$465>>2] = $464;
        $$sink38$i = $462;
       }
       $467 = ((($cffslot)) + 40|0);
       HEAP32[$467>>2] = $$sink38$i;
       $468 = ((($cffslot)) + 56|0);
       HEAP32[$468>>2] = $$sink38$i;
       $469 = ((($cffslot)) + 156|0);
       $470 = HEAP32[$469>>2]|0;
       $471 = ((($470)) + 8|0);
       HEAP8[$471>>0] = 0;
       $472 = ((($11)) + 292|0);
       $473 = HEAP8[$472>>0]|0;
       $474 = ($473<<24>>24)==(0);
       do {
        if ($474) {
         label = 90;
        } else {
         $475 = ((($11)) + 330|0);
         $476 = HEAP16[$475>>1]|0;
         $477 = ($476<<16>>16)==(0);
         if ($477) {
          label = 90;
          break;
         }
         HEAP16[$vertBearingY$i>>1] = 0;
         HEAP16[$vertAdvance$i>>1] = 0;
         $478 = ((($11)) + 532|0);
         $479 = HEAP32[$478>>2]|0;
         $480 = ((($479)) + 112|0);
         $481 = HEAP32[$480>>2]|0;
         FUNCTION_TABLE_viiiii[$481 & 8191]($11,1,$$02$i,$vertBearingY$i,$vertAdvance$i);
         $482 = HEAP16[$vertBearingY$i>>1]|0;
         $483 = $482 << 16 >> 16;
         $484 = ((($cffslot)) + 48|0);
         HEAP32[$484>>2] = $483;
         $485 = HEAP16[$vertAdvance$i>>1]|0;
         $486 = $485&65535;
         $487 = ((($cffslot)) + 52|0);
         HEAP32[$487>>2] = $486;
         $$pre$phi$iZ2D = $487;$506 = $486;$579 = 1;
        }
       } while(0);
       do {
        if ((label|0) == 90) {
         $488 = ((($11)) + 364|0);
         $489 = HEAP16[$488>>1]|0;
         $490 = ($489<<16>>16)==(-1);
         if ($490) {
          $499 = ((($11)) + 220|0);
          $500 = HEAP32[$499>>2]|0;
          $sext$i = $500 << 16;
          $501 = $sext$i >> 16;
          $502 = $500 >> 16;
          $503 = (($501) - ($502))|0;
          $504 = ((($cffslot)) + 52|0);
          HEAP32[$504>>2] = $503;
          $$pre$phi$iZ2D = $504;$506 = $503;$579 = 0;
          break;
         } else {
          $491 = ((($11)) + 434|0);
          $492 = HEAP16[$491>>1]|0;
          $493 = $492 << 16 >> 16;
          $494 = ((($11)) + 436|0);
          $495 = HEAP16[$494>>1]|0;
          $496 = $495 << 16 >> 16;
          $497 = (($493) - ($496))|0;
          $498 = ((($cffslot)) + 52|0);
          HEAP32[$498>>2] = $497;
          $$pre$phi$iZ2D = $498;$506 = $497;$579 = 0;
          break;
         }
        }
       } while(0);
       $505 = ((($cffslot)) + 60|0);
       HEAP32[$505>>2] = $506;
       HEAP32[$225>>2] = 1869968492;
       $507 = ((($cffslot)) + 124|0);
       HEAP32[$507>>2] = 0;
       do {
        if ($34) {
         $508 = ((($size$0)) + 14|0);
         $509 = HEAP16[$508>>1]|0;
         $510 = ($509&65535)<(24);
         if (!($510)) {
          $511 = 4;
          break;
         }
         HEAP32[$507>>2] = 256;
         $511 = 260;
        } else {
         $511 = 4;
        }
       } while(0);
       HEAP32[$507>>2] = $511;
       $512 = HEAP32[$font_matrix$i>>2]|0;
       $513 = ($512|0)==(65536);
       do {
        if ($513) {
         $514 = ((($font_matrix$i)) + 12|0);
         $515 = HEAP32[$514>>2]|0;
         $516 = ($515|0)==(65536);
         if (!($516)) {
          label = 100;
          break;
         }
         $517 = ((($font_matrix$i)) + 4|0);
         $518 = HEAP32[$517>>2]|0;
         $519 = ($518|0)==(0);
         if (!($519)) {
          label = 100;
          break;
         }
         $520 = ((($font_matrix$i)) + 8|0);
         $521 = HEAP32[$520>>2]|0;
         $522 = ($521|0)==(0);
         if (!($522)) {
          label = 100;
         }
        } else {
         label = 100;
        }
       } while(0);
       if ((label|0) == 100) {
        _FT_Outline_Transform($214,$font_matrix$i);
       }
       $523 = $font_offset$sroa$8$0$i | $font_offset$sroa$0$0$i;
       $524 = ($523|0)==(0);
       if (!($524)) {
        _FT_Outline_Translate($214,$font_offset$sroa$0$0$i,$font_offset$sroa$8$0$i);
       }
       $525 = HEAP32[$467>>2]|0;
       HEAP32[$advance4$i>>2] = $525;
       $526 = ((($advance4$i)) + 4|0);
       HEAP32[$526>>2] = 0;
       _FT_Vector_Transform($advance4$i,$font_matrix$i);
       $527 = HEAP32[$advance4$i>>2]|0;
       $528 = (($527) + ($font_offset$sroa$0$0$i))|0;
       HEAP32[$467>>2] = $528;
       HEAP32[$advance4$i>>2] = 0;
       $529 = HEAP32[$$pre$phi$iZ2D>>2]|0;
       HEAP32[$526>>2] = $529;
       _FT_Vector_Transform($advance4$i,$font_matrix$i);
       $530 = HEAP32[$526>>2]|0;
       $531 = (($530) + ($font_offset$sroa$8$0$i))|0;
       HEAP32[$$pre$phi$iZ2D>>2] = $531;
       $532 = ($force_scaling$2$i<<24>>24)==(0);
       $or$cond10$i = $220 & $532;
       if (!($or$cond10$i)) {
        $533 = ((($cffslot)) + 112|0);
        $534 = HEAP32[$533>>2]|0;
        $535 = HEAP32[$32>>2]|0;
        $536 = HEAP32[$33>>2]|0;
        $537 = ($hinting$0$i<<24>>24)==(0);
        if ($537) {
         label = 106;
        } else {
         $538 = ((($decoder$i)) + 68|0);
         $539 = HEAP32[$538>>2]|0;
         $540 = ($539|0)==(0|0);
         if ($540) {
          label = 106;
         }
        }
        do {
         if ((label|0) == 106) {
          $541 = HEAP16[$215>>1]|0;
          $542 = ($541<<16>>16)>(0);
          if (!($542)) {
           break;
          }
          $543 = $541 << 16 >> 16;
          $n$029$i = $543;$vec$030$i = $534;
          while(1) {
           $544 = HEAP32[$vec$030$i>>2]|0;
           $545 = (_FT_MulFix($544,$535)|0);
           HEAP32[$vec$030$i>>2] = $545;
           $546 = ((($vec$030$i)) + 4|0);
           $547 = HEAP32[$546>>2]|0;
           $548 = (_FT_MulFix($547,$536)|0);
           HEAP32[$546>>2] = $548;
           $549 = (($n$029$i) + -1)|0;
           $550 = ((($vec$030$i)) + 8|0);
           $551 = ($n$029$i|0)>(1);
           if ($551) {
            $n$029$i = $549;$vec$030$i = $550;
           } else {
            break;
           }
          }
         }
        } while(0);
        $552 = HEAP32[$467>>2]|0;
        $553 = (_FT_MulFix($552,$535)|0);
        HEAP32[$467>>2] = $553;
        $554 = HEAP32[$$pre$phi$iZ2D>>2]|0;
        $555 = (_FT_MulFix($554,$536)|0);
        HEAP32[$$pre$phi$iZ2D>>2] = $555;
       }
       _FT_Outline_Get_CBox($214,$cbox$i);
       $556 = ((($cbox$i)) + 8|0);
       $557 = HEAP32[$556>>2]|0;
       $558 = HEAP32[$cbox$i>>2]|0;
       $559 = (($557) - ($558))|0;
       HEAP32[$453>>2] = $559;
       $560 = ((($cbox$i)) + 12|0);
       $561 = HEAP32[$560>>2]|0;
       $562 = ((($cbox$i)) + 4|0);
       $563 = HEAP32[$562>>2]|0;
       $564 = (($561) - ($563))|0;
       $565 = ((($cffslot)) + 28|0);
       HEAP32[$565>>2] = $564;
       $566 = HEAP16[$454>>1]|0;
       $567 = ($566<<16>>16)==(0);
       if ($567) {
        $568 = ((($cffslot)) + 32|0);
        HEAP32[$568>>2] = $558;
       }
       $569 = ((($cffslot)) + 36|0);
       HEAP32[$569>>2] = $561;
       do {
        if ($579) {
         $570 = ((($cffslot)) + 32|0);
         $571 = HEAP32[$570>>2]|0;
         $572 = HEAP32[$467>>2]|0;
         $573 = (($572|0) / 2)&-1;
         $574 = (($571) - ($573))|0;
         $575 = ((($cffslot)) + 44|0);
         HEAP32[$575>>2] = $574;
        } else {
         $576 = $load_flags$$i & 16;
         $577 = ($576|0)==(0);
         if ($577) {
          break;
         }
         $578 = HEAP32[$$pre$phi$iZ2D>>2]|0;
         _ft_synthesize_vertical_metrics($453,$578);
        }
       } while(0);
       $$3$i = 0;
       break L18;
      } else {
       $error$1$ph$i = $error$0$i;
      }
     } else {
      $error$1$ph$i = $$0$i$i;
     }
    } while(0);
    $$3$i = $error$1$ph$i;
   } else {
    $$3$i = 6;
   }
  }
 } while(0);
 $$0 = $$3$i;
 STACKTOP = sp;return ($$0|0);
}
function _cf2_decoder_parse_charstrings($decoder,$charstring_base,$charstring_len) {
 $decoder = $decoder|0;
 $charstring_base = $charstring_base|0;
 $charstring_len = $charstring_len|0;
 var $$ = 0, $$$i$i = 0, $$$i$i$i = 0, $$0 = 0, $$1 = 0, $$1$i$i = 0, $$idx$i$i = 0, $$idx$i1$i = 0, $$idx$val = 0, $$idx$val$i = 0, $$idx$val$i$i = 0, $$idx$val$i2$i = 0, $$idx2$i$i = 0, $$idx2$val$i$i = 0, $$idx2$val$idx$i$i = 0, $$idx2$val$idx$val$i$i = 0, $$idx2$val$idx$val$idx$i$i = 0, $$idx2$val$idx$val$idx$val$i$i = 0, $$idx3$val$i$i = 0, $$idx3$val$idx$i$i = 0;
 var $$idx3$val$idx$val$i$i = 0, $$idx4$val$i$i = 0, $$idx4$val$idx$i$i = 0, $$idx4$val$idx$val$i$i = 0, $$idx6$val$idx$i$i$i = 0, $$idx6$val$idx$val$i$i$i = 0, $$maxZoneHeight$0$i$i$i = 0, $$maxZoneHeight$3$i$i$i = 0, $$pr = 0, $$pre$i = 0, $$pre$i$i$i = 0, $$pre14$i$i = 0, $$pre18$i = 0, $$pre38$i$i$i = 0, $$sink$i$i$i = 0, $$sink$in$i$i$i = 0, $$sink$in$p$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0;
 var $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0;
 var $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0;
 var $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0;
 var $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0;
 var $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0;
 var $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0;
 var $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0;
 var $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0;
 var $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0;
 var $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0;
 var $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0;
 var $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0;
 var $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0;
 var $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0;
 var $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0;
 var $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0;
 var $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0;
 var $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0;
 var $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0;
 var $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0;
 var $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0;
 var $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0;
 var $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0;
 var $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0;
 var $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0;
 var $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $advWidth$i = 0, $buf = 0, $decoder$idx$val = 0, $decoder$idx$val$idx = 0, $decoder$idx$val$idx$val = 0, $error = 0, $font$0$in = 0, $i$032$i$i$i = 0, $i$127$i$i$i = 0, $i$222$i$us$i$i = 0, $i$222$us$i$i$i = 0, $i$222$us$i$i$us$i = 0, $i$222$us$i$us$i$i = 0;
 var $i$222$us$i$us$i$us$i = 0, $i$312$i$i$i = 0, $j$013$i$us$i$i = 0, $j$013$us$i$us$i$i = 0, $j$013$us$i$us$i$us$i = 0, $j$118$us$i$i$us$i = 0, $j$118$us$i$us$i$us$i = 0, $maxZoneHeight$0$lcssa$i$i$i = 0, $maxZoneHeight$033$i$i$i = 0, $maxZoneHeight$2$i$i$i = 0, $maxZoneHeight$3$lcssa$i$i$i = 0, $maxZoneHeight$328$i$i$i = 0, $maxZoneHeight$5$i$i$i = 0, $minDiff$014$i$us$i$i = 0, $minDiff$014$us$i$us$i$i = 0, $minDiff$014$us$i$us$i$us$i = 0, $minDiff$1$i$us$i$i = 0, $minDiff$1$us$i$us$i$i = 0, $minDiff$1$us$i$us$i$us$i = 0, $minDiff$2$us$i$ph$us$i$i = 0;
 var $minDiff$2$us$i$ph$us$i$us$i = 0, $minDiff$319$us$i$i$us$i = 0, $minDiff$319$us$i$us$i$us$i = 0, $minDiff$4$us$i$i$us$i = 0, $minDiff$4$us$i$us$i$us$i = 0, $needExtraSetup$0$i$i = 0, $needExtraSetup$1$i$i = 0, $needExtraSetup$2$i$i = 0, $needWinding$0$i = 0, $or$cond = 0, $or$cond$i$us$i$i = 0, $or$cond$not$i$i$i = 0, $or$cond$us$i$us$i$i = 0, $or$cond$us$i$us$i$us$i = 0, $or$cond1$us$i$i$i = 0, $or$cond1$us$i$i$us$i = 0, $or$cond1$us$i$us$i$i = 0, $or$cond1$us$i$us$i$us$i = 0, $or$cond2$us$i$i$us$i = 0, $or$cond2$us$i$us$i$us$i = 0;
 var $or$cond6 = 0, $or$cond6$i$i$i = 0, $sext$i$i$i = 0, $sext2$i$i$i = 0, $storemerge$i = 0, $transform = 0, $translation$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $translation$i = sp + 48|0;
 $advWidth$i = sp + 44|0;
 $error = sp + 40|0;
 $buf = sp;
 $transform = sp + 16|0;
 HEAP32[$error>>2] = 0;
 $0 = HEAP32[$decoder>>2]|0;
 $1 = ((($decoder)) + 76|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 2972|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 do {
  if ($5) {
   $6 = ((($2)) + 2976|0);
   HEAP32[$6>>2] = 4357;
   $7 = (_ft_mem_alloc($0,504,$error)|0);
   $8 = HEAP32[$1>>2]|0;
   $9 = ((($8)) + 2972|0);
   HEAP32[$9>>2] = $7;
   $10 = HEAP32[$error>>2]|0;
   $11 = ($10|0)==(0);
   if ($11) {
    HEAP32[$7>>2] = $0;
    $12 = ((($7)) + 100|0);
    $13 = ((($7)) + 4|0);
    ;HEAP32[$12>>2]=0|0;HEAP32[$12+4>>2]=0|0;HEAP32[$12+8>>2]=0|0;HEAP32[$12+12>>2]=0|0;HEAP32[$12+16>>2]=0|0;HEAP32[$12+20>>2]=0|0;HEAP32[$12+24>>2]=0|0;HEAP32[$12+28>>2]=0|0;
    $14 = ((($7)) + 120|0);
    HEAP32[$14>>2] = $0;
    $15 = ((($7)) + 124|0);
    HEAP32[$15>>2] = $13;
    HEAP32[$12>>2] = 4358;
    $16 = ((($7)) + 104|0);
    HEAP32[$16>>2] = 4359;
    $17 = ((($7)) + 112|0);
    HEAP32[$17>>2] = 4360;
    $font$0$in = $7;
    break;
   } else {
    $$1 = 64;
    STACKTOP = sp;return ($$1|0);
   }
  } else {
   $font$0$in = $4;
  }
 } while(0);
 $18 = ((($font$0$in)) + 132|0);
 HEAP32[$18>>2] = $decoder;
 $19 = ((($font$0$in)) + 128|0);
 HEAP32[$19>>2] = $decoder;
 $20 = ((($decoder)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ((($21)) + 96|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = $buf;
 $25 = $24;
 HEAP32[$25>>2] = 0;
 $26 = (($24) + 4)|0;
 $27 = $26;
 HEAP32[$27>>2] = 0;
 $28 = ((($buf)) + 12|0);
 HEAP32[$28>>2] = $charstring_base;
 $29 = ((($buf)) + 4|0);
 HEAP32[$29>>2] = $charstring_base;
 $30 = (($charstring_base) + ($charstring_len)|0);
 $31 = ((($buf)) + 8|0);
 HEAP32[$31>>2] = $30;
 ;HEAP32[$transform>>2]=0|0;HEAP32[$transform+4>>2]=0|0;HEAP32[$transform+8>>2]=0|0;HEAP32[$transform+12>>2]=0|0;HEAP32[$transform+16>>2]=0|0;HEAP32[$transform+20>>2]=0|0;
 $32 = ((($transform)) + 12|0);
 $33 = ((($decoder)) + 8|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = ((($34)) + 160|0);
 $36 = HEAP16[$35>>1]|0;
 $37 = $36&255;
 $38 = ($37<<24>>24)==(0);
 if ($38) {
  HEAP32[$transform>>2] = 1024;
  $storemerge$i = 1024;
 } else {
  $39 = ((($34)) + 164|0);
  $40 = HEAP32[$39>>2]|0;
  $41 = (($40) + 32)|0;
  $42 = (($41|0) / 64)&-1;
  HEAP32[$transform>>2] = $42;
  $43 = ((($34)) + 168|0);
  $44 = HEAP32[$43>>2]|0;
  $45 = (($44) + 32)|0;
  $46 = (($45|0) / 64)&-1;
  $storemerge$i = $46;
 }
 HEAP32[$32>>2] = $storemerge$i;
 $47 = ((($font$0$in)) + 8|0);
 $48 = $38&1;
 $$ = $48 ^ 1;
 HEAP32[$47>>2] = $$;
 $49 = ($36&65535)<(256);
 if (!($49)) {
  $50 = ((($23)) + 32|0);
  $51 = HEAP8[$50>>0]|0;
  $52 = ($51<<24>>24)==(0);
  if ($52) {
   $53 = $$ | 2;
   HEAP32[$47>>2] = $53;
  }
 }
 $54 = ((($23)) + 36|0);
 $55 = HEAP32[$54>>2]|0;
 $56 = ((($font$0$in)) + 144|0);
 HEAP32[$56>>2] = $55;
 $57 = ((($23)) + 40|0);
 $58 = HEAP32[$57>>2]|0;
 $59 = ((($font$0$in)) + 148|0);
 HEAP32[$59>>2] = $58;
 $60 = ((($23)) + 44|0);
 $61 = HEAP32[$60>>2]|0;
 $62 = ((($font$0$in)) + 152|0);
 HEAP32[$62>>2] = $61;
 $63 = ((($23)) + 48|0);
 $64 = HEAP32[$63>>2]|0;
 $65 = ((($font$0$in)) + 156|0);
 HEAP32[$65>>2] = $64;
 $66 = ((($23)) + 52|0);
 $67 = HEAP32[$66>>2]|0;
 $68 = ((($font$0$in)) + 160|0);
 HEAP32[$68>>2] = $67;
 $69 = ((($23)) + 56|0);
 $70 = HEAP32[$69>>2]|0;
 $71 = ((($font$0$in)) + 164|0);
 HEAP32[$71>>2] = $70;
 $72 = ((($23)) + 60|0);
 $73 = HEAP32[$72>>2]|0;
 $74 = ((($font$0$in)) + 168|0);
 HEAP32[$74>>2] = $73;
 $75 = ((($23)) + 64|0);
 $76 = HEAP32[$75>>2]|0;
 $77 = ((($font$0$in)) + 172|0);
 HEAP32[$77>>2] = $76;
 $decoder$idx$val = HEAP32[$20>>2]|0;
 $decoder$idx$val$idx = ((($decoder$idx$val)) + 68|0);
 $decoder$idx$val$idx$val = HEAP16[$decoder$idx$val$idx>>1]|0;
 $78 = $decoder$idx$val$idx$val&65535;
 $79 = ((($font$0$in)) + 88|0);
 HEAP32[$79>>2] = $78;
 if ($49) {
  label = 14;
 } else {
  $80 = HEAP32[$transform>>2]|0;
  $81 = ($80|0)<(1);
  $82 = HEAP32[$32>>2]|0;
  $83 = ($82|0)<(1);
  $or$cond = $81 | $83;
  if ($or$cond) {
   $$0 = 36;
  } else {
   $84 = ($decoder$idx$val$idx$val<<16>>16)<(0);
   if ($84) {
    $$0 = 164;
   } else {
    $85 = $78 << 16;
    $86 = (_FT_DivFix(131072000,$85)|0);
    $87 = ($80|0)>($86|0);
    $88 = ($82|0)>($86|0);
    $or$cond6 = $87 | $88;
    if ($or$cond6) {
     $$0 = 164;
    } else {
     label = 14;
    }
   }
  }
 }
 L19: do {
  if ((label|0) == 14) {
   HEAP32[$advWidth$i>>2] = 0;
   $89 = ((($transform)) + 16|0);
   $90 = HEAP32[$89>>2]|0;
   HEAP32[$translation$i>>2] = $90;
   $91 = ((($transform)) + 20|0);
   $92 = HEAP32[$91>>2]|0;
   $93 = ((($translation$i)) + 4|0);
   HEAP32[$93>>2] = $92;
   $94 = HEAP32[$18>>2]|0;
   $95 = ((($font$0$in)) + 92|0);
   $96 = HEAP32[$95>>2]|0;
   $97 = ((($font$0$in)) + 96|0);
   $98 = HEAP32[$97>>2]|0;
   $99 = ((($font$0$in)) + 4|0);
   HEAP32[$99>>2] = 0;
   $$idx$i$i = ((($94)) + 736|0);
   $$idx$val$i$i = HEAP32[$$idx$i$i>>2]|0;
   $100 = ((($font$0$in)) + 136|0);
   $101 = HEAP32[$100>>2]|0;
   $102 = ($101|0)==($$idx$val$i$i|0);
   if ($102) {
    $needExtraSetup$0$i$i = 0;
   } else {
    HEAP32[$100>>2] = $$idx$val$i$i;
    $needExtraSetup$0$i$i = 1;
   }
   $$idx2$i$i = ((($94)) + 4|0);
   $$idx2$val$i$i = HEAP32[$$idx2$i$i>>2]|0;
   $$idx2$val$idx$i$i = ((($$idx2$val$i$i)) + 88|0);
   $$idx2$val$idx$val$i$i = HEAP32[$$idx2$val$idx$i$i>>2]|0;
   $$idx2$val$idx$val$idx$i$i = ((($$idx2$val$idx$val$i$i)) + 14|0);
   $$idx2$val$idx$val$idx$val$i$i = HEAP16[$$idx2$val$idx$val$idx$i$i>>1]|0;
   $103 = $$idx2$val$idx$val$idx$val$i$i&65535;
   $104 = $103 << 16;
   $105 = ((($font$0$in)) + 84|0);
   $106 = HEAP32[$105>>2]|0;
   $107 = ($106|0)==($104|0);
   if ($107) {
    $needExtraSetup$1$i$i = $needExtraSetup$0$i$i;
   } else {
    HEAP32[$105>>2] = $104;
    $needExtraSetup$1$i$i = 1;
   }
   $108 = HEAP32[$47>>2]|0;
   $109 = $108 & 1;
   $110 = $109&255;
   $111 = ((($font$0$in)) + 140|0);
   HEAP8[$111>>0] = $110;
   $112 = ((($font$0$in)) + 12|0);
   $113 = (_memcmp($transform,$112,16)|0);
   $114 = ($113|0)==(0);
   if ($114) {
    $needExtraSetup$2$i$i = $needExtraSetup$1$i$i;
   } else {
    ;HEAP32[$112>>2]=HEAP32[$transform>>2]|0;HEAP32[$112+4>>2]=HEAP32[$transform+4>>2]|0;HEAP32[$112+8>>2]=HEAP32[$transform+8>>2]|0;HEAP32[$112+12>>2]=HEAP32[$transform+12>>2]|0;
    $115 = ((($font$0$in)) + 32|0);
    HEAP32[$115>>2] = 0;
    $116 = ((($font$0$in)) + 28|0);
    HEAP32[$116>>2] = 0;
    $117 = ((($font$0$in)) + 36|0);
    ;HEAP32[$117>>2]=HEAP32[$transform>>2]|0;HEAP32[$117+4>>2]=HEAP32[$transform+4>>2]|0;HEAP32[$117+8>>2]=HEAP32[$transform+8>>2]|0;HEAP32[$117+12>>2]=HEAP32[$transform+12>>2]|0;HEAP32[$117+16>>2]=HEAP32[$transform+16>>2]|0;HEAP32[$117+20>>2]=HEAP32[$transform+20>>2]|0;
    $118 = ((($font$0$in)) + 72|0);
    HEAP32[$118>>2] = 65536;
    $119 = ((($font$0$in)) + 60|0);
    HEAP32[$119>>2] = 65536;
    $120 = ((($font$0$in)) + 68|0);
    HEAP32[$120>>2] = 0;
    $121 = ((($font$0$in)) + 64|0);
    HEAP32[$121>>2] = 0;
    $needExtraSetup$2$i$i = 1;
   }
   $122 = ((($font$0$in)) + 142|0);
   $123 = HEAP8[$122>>0]|0;
   $124 = $123&255;
   $125 = $108 & 2;
   $126 = ($124|0)==($125|0);
   if ($126) {
    $128 = ($needExtraSetup$2$i$i<<24>>24)==(0);
    if (!($128)) {
     label = 23;
    }
   } else {
    $127 = $125&255;
    HEAP8[$122>>0] = $127;
    label = 23;
   }
   L33: do {
    if ((label|0) == 23) {
     $129 = HEAP32[$79>>2]|0;
     $130 = ($129|0)==(0);
     $$$i$i = $130 ? 1000 : $129;
     $131 = ($104|0)<(262144);
     $$1$i$i = $131 ? 262144 : $104;
     $132 = (65536000 / ($$$i$i|0))&-1;
     $$idx3$val$i$i = HEAP32[$$idx$i$i>>2]|0;
     $$idx3$val$idx$i$i = ((($$idx3$val$i$i)) + 388|0);
     $$idx3$val$idx$val$i$i = HEAP32[$$idx3$val$idx$i$i>>2]|0;
     $133 = $$idx3$val$idx$val$i$i << 16;
     $134 = ((($font$0$in)) + 176|0);
     HEAP32[$134>>2] = $133;
     $135 = ($133|0)<(1);
     if ($135) {
      $136 = (_FT_DivFix(4915200,$132)|0);
      HEAP32[$134>>2] = $136;
      $148 = $136;
     } else {
      $148 = $133;
     }
     $137 = ($96|0)>(0);
     if ($137) {
      $138 = $$$i$i << 16;
      $139 = (_FT_DivFix($138,$$1$i$i)|0);
      $140 = ($96|0)>($139|0);
      if ($140) {
       $143 = $96;
      } else {
       $141 = (_FT_DivFix($138,$$1$i$i)|0);
       $143 = $141;
      }
      $142 = ((($font$0$in)) + 184|0);
      HEAP32[$142>>2] = 0;
      $or$cond$not$i$i$i = ($143|0)==(0);
      $144 = ($132|0)<(655);
      $or$cond6$i$i$i = $144 | $or$cond$not$i$i$i;
      if (!($or$cond6$i$i$i)) {
       $145 = (($143|0) / 2)&-1;
       HEAP32[$142>>2] = $145;
      }
     } else {
      $146 = ((($font$0$in)) + 184|0);
      $147 = HEAP8[$122>>0]|0;
      _cf2_computeDarkening($132,$$1$i$i,$148,$146,0,$147,$56);
     }
     $$idx4$val$i$i = HEAP32[$$idx$i$i>>2]|0;
     $$idx4$val$idx$i$i = ((($$idx4$val$i$i)) + 384|0);
     $$idx4$val$idx$val$i$i = HEAP32[$$idx4$val$idx$i$i>>2]|0;
     $149 = $$idx4$val$idx$val$i$i&65535;
     $150 = ($149<<16>>16)>(0);
     if ($150) {
      $151 = HEAP32[$134>>2]|0;
      $152 = $$idx4$val$idx$val$i$i << 17;
      $153 = ($151|0)>($152|0);
      if ($153) {
       $154 = (_FT_DivFix(4915200,$132)|0);
       $155 = ((($font$0$in)) + 180|0);
       HEAP32[$155>>2] = $154;
       $160 = $154;
      } else {
       label = 34;
      }
     } else {
      label = 34;
     }
     if ((label|0) == 34) {
      $156 = (_FT_DivFix(7208960,$132)|0);
      $157 = ((($font$0$in)) + 180|0);
      HEAP32[$157>>2] = $156;
      $160 = $156;
     }
     $158 = ((($font$0$in)) + 188|0);
     $159 = HEAP8[$122>>0]|0;
     _cf2_computeDarkening($132,$$1$i$i,$160,$158,$98,$159,$56);
     $161 = ((($font$0$in)) + 184|0);
     $162 = HEAP32[$161>>2]|0;
     $163 = ($162|0)==(0);
     if ($163) {
      $164 = HEAP32[$158>>2]|0;
      $165 = ($164|0)==(0);
      if ($165) {
       $167 = ((($font$0$in)) + 141|0);
       HEAP8[$167>>0] = 0;
      } else {
       label = 37;
      }
     } else {
      label = 37;
     }
     if ((label|0) == 37) {
      $166 = ((($font$0$in)) + 141|0);
      HEAP8[$166>>0] = 1;
     }
     $168 = ((($font$0$in)) + 192|0);
     HEAP8[$168>>0] = 0;
     $169 = ((($font$0$in)) + 196|0);
     $170 = HEAP32[$18>>2]|0;
     _memset(($169|0),0,308)|0;
     $171 = ((($font$0$in)) + 48|0);
     $172 = HEAP32[$171>>2]|0;
     HEAP32[$169>>2] = $172;
     $173 = ((($font$0$in)) + 208|0);
     $174 = ((($font$0$in)) + 212|0);
     $175 = ((($font$0$in)) + 216|0);
     $176 = ((($170)) + 736|0);
     $177 = HEAP32[$176>>2]|0;
     $178 = ((($177)) + 372|0);
     $179 = HEAP32[$178>>2]|0;
     $180 = (_FT_DivFix($179,65536000)|0);
     HEAP32[$173>>2] = $180;
     $181 = HEAP32[$176>>2]|0;
     $182 = ((($181)) + 376|0);
     $183 = HEAP32[$182>>2]|0;
     $184 = $183 << 16;
     HEAP32[$174>>2] = $184;
     $185 = ((($181)) + 380|0);
     $186 = HEAP32[$185>>2]|0;
     $187 = $186 << 16;
     HEAP32[$175>>2] = $187;
     $188 = ((($181)) + 176|0);
     $189 = HEAP32[$188>>2]|0;
     $190 = $189&65535;
     $191 = $189&255;
     $192 = $189 & 255;
     $193 = ((($181)) + 180|0);
     $194 = ($190&65535) >>> 8;
     $195 = $194&65535;
     $196 = $189 >>> 16;
     $197 = $196&255;
     $198 = $196 & 255;
     $199 = $189 >>> 24;
     $$idx6$val$idx$i$i$i = ((($181)) + 512|0);
     $$idx6$val$idx$val$i$i$i = HEAP32[$$idx6$val$idx$i$i$i>>2]|0;
     $200 = ($$idx6$val$idx$val$i$i$i|0)==(1);
     L58: do {
      if ($200) {
       switch ($192|0) {
       case 0:  {
        break;
       }
       case 4:  {
        $202 = HEAP32[$193>>2]|0;
        $203 = $202&65535;
        $204 = ($203<<16>>16)<(-120);
        if (!($204)) {
         break L58;
        }
        $205 = ((($181)) + 184|0);
        $206 = HEAP32[$205>>2]|0;
        $207 = $206&65535;
        $208 = ($207<<16>>16)<(-120);
        if (!($208)) {
         break L58;
        }
        $209 = ((($181)) + 188|0);
        $210 = HEAP32[$209>>2]|0;
        $211 = $210&65535;
        $212 = ($211<<16>>16)>(880);
        if (!($212)) {
         break L58;
        }
        $213 = ((($181)) + 192|0);
        $214 = HEAP32[$213>>2]|0;
        $215 = $214&65535;
        $216 = ($215<<16>>16)>(880);
        if (!($216)) {
         break L58;
        }
        break;
       }
       default: {
        break L58;
       }
       }
       $217 = ((($font$0$in)) + 252|0);
       HEAP32[$217>>2] = -7864321;
       $218 = HEAP32[$169>>2]|0;
       $219 = (_FT_MulFix(-7864321,$218)|0);
       $220 = (($219) + 32768)|0;
       $221 = $220 & -65536;
       $222 = (($221) + -32768)|0;
       $223 = ((($font$0$in)) + 256|0);
       HEAP32[$223>>2] = $222;
       $224 = HEAP32[$169>>2]|0;
       $225 = ((($font$0$in)) + 260|0);
       HEAP32[$225>>2] = $224;
       $226 = ((($font$0$in)) + 244|0);
       HEAP32[$226>>2] = 49;
       $227 = HEAP32[$158>>2]|0;
       $228 = $227 << 1;
       $229 = (($228) + 57671681)|0;
       $230 = ((($font$0$in)) + 232|0);
       HEAP32[$230>>2] = $229;
       $231 = (_FT_MulFix($229,$224)|0);
       $232 = (($231) + 32768)|0;
       $233 = $232 & -65536;
       $234 = $233 | 32768;
       $235 = ((($font$0$in)) + 236|0);
       HEAP32[$235>>2] = $234;
       $236 = HEAP32[$169>>2]|0;
       $237 = ((($font$0$in)) + 240|0);
       HEAP32[$237>>2] = $236;
       $238 = ((($font$0$in)) + 224|0);
       HEAP32[$238>>2] = 50;
       $239 = ((($font$0$in)) + 205|0);
       HEAP8[$239>>0] = 1;
       break L33;
      }
     } while(0);
     $201 = ($191<<24>>24)==(0);
     $$pre14$i$i = ((($font$0$in)) + 200|0);
     if ($201) {
      $maxZoneHeight$0$lcssa$i$i$i = 0;
     } else {
      $$pre$i$i$i = HEAP32[$$pre14$i$i>>2]|0;
      $245 = $$pre$i$i$i;$i$032$i$i$i = 0;$maxZoneHeight$033$i$i$i = 0;
      while(1) {
       $241 = (((($181)) + 180|0) + ($i$032$i$i$i<<2)|0);
       $242 = HEAP32[$241>>2]|0;
       $243 = $242 << 16;
       $244 = (((($font$0$in)) + 264|0) + (($245*20)|0)|0);
       HEAP32[$244>>2] = $243;
       $246 = $i$032$i$i$i | 1;
       $247 = (((($181)) + 180|0) + ($246<<2)|0);
       $248 = HEAP32[$247>>2]|0;
       $249 = $248 << 16;
       $250 = HEAP32[$$pre14$i$i>>2]|0;
       $251 = (((((($font$0$in)) + 264|0) + (($250*20)|0)|0)) + 4|0);
       HEAP32[$251>>2] = $249;
       $252 = HEAP32[$$pre14$i$i>>2]|0;
       $253 = (((((($font$0$in)) + 264|0) + (($252*20)|0)|0)) + 4|0);
       $254 = HEAP32[$253>>2]|0;
       $255 = (((($font$0$in)) + 264|0) + (($252*20)|0)|0);
       $256 = HEAP32[$255>>2]|0;
       $257 = (($254) - ($256))|0;
       $258 = ($257|0)<(0);
       if ($258) {
        $566 = $252;$maxZoneHeight$2$i$i$i = $maxZoneHeight$033$i$i$i;
       } else {
        $259 = ($257|0)>($maxZoneHeight$033$i$i$i|0);
        $$maxZoneHeight$0$i$i$i = $259 ? $257 : $maxZoneHeight$033$i$i$i;
        $260 = ($i$032$i$i$i|0)==(0);
        if ($260) {
         $261 = (((((($font$0$in)) + 264|0) + (($252*20)|0)|0)) + 16|0);
         HEAP8[$261>>0] = 1;
         $262 = HEAP32[$$pre14$i$i>>2]|0;
         $263 = (((((($font$0$in)) + 264|0) + (($262*20)|0)|0)) + 4|0);
         $264 = HEAP32[$263>>2]|0;
         $265 = (((((($font$0$in)) + 264|0) + (($262*20)|0)|0)) + 8|0);
         HEAP32[$265>>2] = $264;
        } else {
         $266 = HEAP32[$158>>2]|0;
         $267 = $266 << 1;
         $268 = (($267) + ($254))|0;
         HEAP32[$253>>2] = $268;
         $269 = HEAP32[$$pre14$i$i>>2]|0;
         $270 = (((($font$0$in)) + 264|0) + (($269*20)|0)|0);
         $271 = HEAP32[$270>>2]|0;
         $272 = (($271) + ($267))|0;
         HEAP32[$270>>2] = $272;
         $273 = HEAP32[$$pre14$i$i>>2]|0;
         $274 = (((((($font$0$in)) + 264|0) + (($273*20)|0)|0)) + 16|0);
         HEAP8[$274>>0] = 0;
         $275 = HEAP32[$$pre14$i$i>>2]|0;
         $276 = (((($font$0$in)) + 264|0) + (($275*20)|0)|0);
         $277 = HEAP32[$276>>2]|0;
         $278 = (((((($font$0$in)) + 264|0) + (($275*20)|0)|0)) + 8|0);
         HEAP32[$278>>2] = $277;
        }
        $279 = HEAP32[$$pre14$i$i>>2]|0;
        $280 = (($279) + 1)|0;
        HEAP32[$$pre14$i$i>>2] = $280;
        $566 = $280;$maxZoneHeight$2$i$i$i = $$maxZoneHeight$0$i$i$i;
       }
       $281 = (($i$032$i$i$i) + 2)|0;
       $282 = ($281>>>0)<($192>>>0);
       if ($282) {
        $245 = $566;$i$032$i$i$i = $281;$maxZoneHeight$033$i$i$i = $maxZoneHeight$2$i$i$i;
       } else {
        $maxZoneHeight$0$lcssa$i$i$i = $maxZoneHeight$2$i$i$i;
        break;
       }
      }
     }
     $240 = ($194<<16>>16)==(0);
     if ($240) {
      $maxZoneHeight$3$lcssa$i$i$i = $maxZoneHeight$0$lcssa$i$i$i;
     } else {
      $$pre38$i$i$i = HEAP32[$$pre14$i$i>>2]|0;
      $287 = $$pre38$i$i$i;$i$127$i$i$i = 0;$maxZoneHeight$328$i$i$i = $maxZoneHeight$0$lcssa$i$i$i;
      while(1) {
       $283 = (((($181)) + 236|0) + ($i$127$i$i$i<<2)|0);
       $284 = HEAP32[$283>>2]|0;
       $285 = $284 << 16;
       $286 = (((($font$0$in)) + 264|0) + (($287*20)|0)|0);
       HEAP32[$286>>2] = $285;
       $288 = $i$127$i$i$i | 1;
       $289 = (((($181)) + 236|0) + ($288<<2)|0);
       $290 = HEAP32[$289>>2]|0;
       $291 = $290 << 16;
       $292 = HEAP32[$$pre14$i$i>>2]|0;
       $293 = (((((($font$0$in)) + 264|0) + (($292*20)|0)|0)) + 4|0);
       HEAP32[$293>>2] = $291;
       $294 = HEAP32[$$pre14$i$i>>2]|0;
       $295 = (((((($font$0$in)) + 264|0) + (($294*20)|0)|0)) + 4|0);
       $296 = HEAP32[$295>>2]|0;
       $297 = (((($font$0$in)) + 264|0) + (($294*20)|0)|0);
       $298 = HEAP32[$297>>2]|0;
       $299 = (($296) - ($298))|0;
       $300 = ($299|0)<(0);
       if ($300) {
        $567 = $294;$maxZoneHeight$5$i$i$i = $maxZoneHeight$328$i$i$i;
       } else {
        $301 = ($299|0)>($maxZoneHeight$328$i$i$i|0);
        $$maxZoneHeight$3$i$i$i = $301 ? $299 : $maxZoneHeight$328$i$i$i;
        $302 = (((((($font$0$in)) + 264|0) + (($294*20)|0)|0)) + 16|0);
        HEAP8[$302>>0] = 1;
        $303 = HEAP32[$$pre14$i$i>>2]|0;
        $304 = (((((($font$0$in)) + 264|0) + (($303*20)|0)|0)) + 4|0);
        $305 = HEAP32[$304>>2]|0;
        $306 = (((((($font$0$in)) + 264|0) + (($303*20)|0)|0)) + 8|0);
        HEAP32[$306>>2] = $305;
        $307 = HEAP32[$$pre14$i$i>>2]|0;
        $308 = (($307) + 1)|0;
        HEAP32[$$pre14$i$i>>2] = $308;
        $567 = $308;$maxZoneHeight$5$i$i$i = $$maxZoneHeight$3$i$i$i;
       }
       $309 = (($i$127$i$i$i) + 2)|0;
       $310 = ($309>>>0)<($195>>>0);
       if ($310) {
        $287 = $567;$i$127$i$i$i = $309;$maxZoneHeight$328$i$i$i = $maxZoneHeight$5$i$i$i;
       } else {
        $maxZoneHeight$3$lcssa$i$i$i = $maxZoneHeight$5$i$i$i;
        break;
       }
      }
     }
     $311 = HEAP32[$169>>2]|0;
     $312 = (_FT_DivFix(65536,$311)|0);
     $313 = HEAP32[$$pre14$i$i>>2]|0;
     $314 = ($313|0)==(0);
     L87: do {
      if (!($314)) {
       $315 = ($197&255)>(2);
       $316 = ($199|0)!=(0);
       $317 = ($197&255)>(1);
       $318 = ((($181)) + 280|0);
       if (!($317)) {
        if ($316) {
         $568 = $313;$i$222$i$us$i$i = 0;
        } else {
         break;
        }
        while(1) {
         $319 = (((((($font$0$in)) + 264|0) + (($i$222$i$us$i$i*20)|0)|0)) + 8|0);
         $320 = HEAP32[$319>>2]|0;
         $321 = (((((($font$0$in)) + 264|0) + (($i$222$i$us$i$i*20)|0)|0)) + 16|0);
         $322 = HEAP8[$321>>0]|0;
         $323 = ($322<<24>>24)==(0);
         if ($323) {
          $339 = $568;
         } else {
          $j$013$i$us$i$i = 0;$minDiff$014$i$us$i$i = 2147483647;
          while(1) {
           $324 = $j$013$i$us$i$i | 1;
           $325 = (((($181)) + 332|0) + ($324<<2)|0);
           $326 = HEAP32[$325>>2]|0;
           $327 = $326 << 16;
           $328 = (($320) - ($327))|0;
           $329 = ($328|0)<(0);
           $330 = (0 - ($328))|0;
           $331 = $329 ? $330 : $328;
           $332 = ($331|0)<($minDiff$014$i$us$i$i|0);
           $333 = ($331|0)<($312|0);
           $or$cond$i$us$i$i = $332 & $333;
           if ($or$cond$i$us$i$i) {
            HEAP32[$319>>2] = $327;
            $334 = ($331|0)==(0);
            if ($334) {
             break;
            } else {
             $minDiff$1$i$us$i$i = $331;
            }
           } else {
            $minDiff$1$i$us$i$i = $minDiff$014$i$us$i$i;
           }
           $335 = (($j$013$i$us$i$i) + 2)|0;
           $336 = ($335>>>0)<($199>>>0);
           if ($336) {
            $j$013$i$us$i$i = $335;$minDiff$014$i$us$i$i = $minDiff$1$i$us$i$i;
           } else {
            break;
           }
          }
          $$pre$i = HEAP32[$$pre14$i$i>>2]|0;
          $339 = $$pre$i;
         }
         $337 = (($i$222$i$us$i$i) + 1)|0;
         $338 = ($337>>>0)<($339>>>0);
         if ($338) {
          $568 = $339;$i$222$i$us$i$i = $337;
         } else {
          break L87;
         }
        }
       }
       if (!($316)) {
        if ($315) {
         $i$222$us$i$i$us$i = 0;
        } else {
         $569 = $313;$i$222$us$i$i$i = 0;
         while(1) {
          $444 = (((((($font$0$in)) + 264|0) + (($i$222$us$i$i$i*20)|0)|0)) + 8|0);
          $445 = (((((($font$0$in)) + 264|0) + (($i$222$us$i$i$i*20)|0)|0)) + 16|0);
          $446 = HEAP8[$445>>0]|0;
          $447 = ($446<<24>>24)==(0);
          do {
           if ($447) {
            $451 = $569;
           } else {
            $452 = HEAP32[$444>>2]|0;
            $453 = HEAP32[$318>>2]|0;
            $448 = $453 << 16;
            $454 = (($452) - ($448))|0;
            $455 = ($454|0)<(0);
            $456 = (0 - ($454))|0;
            $457 = $455 ? $456 : $454;
            $458 = ($457|0)!=(2147483647);
            $459 = ($457|0)<($312|0);
            $or$cond1$us$i$i$i = $458 & $459;
            if (!($or$cond1$us$i$i$i)) {
             $451 = $569;
             break;
            }
            HEAP32[$444>>2] = $448;
            $$pre18$i = HEAP32[$$pre14$i$i>>2]|0;
            $451 = $$pre18$i;
           }
          } while(0);
          $449 = (($i$222$us$i$i$i) + 1)|0;
          $450 = ($449>>>0)<($451>>>0);
          if ($450) {
           $569 = $451;$i$222$us$i$i$i = $449;
          } else {
           break L87;
          }
         }
        }
        while(1) {
         $340 = (((((($font$0$in)) + 264|0) + (($i$222$us$i$i$us$i*20)|0)|0)) + 8|0);
         $341 = HEAP32[$340>>2]|0;
         $342 = (((((($font$0$in)) + 264|0) + (($i$222$us$i$i$us$i*20)|0)|0)) + 16|0);
         $343 = HEAP8[$342>>0]|0;
         $344 = ($343<<24>>24)==(0);
         L113: do {
          if ($344) {
           $353 = HEAP32[$158>>2]|0;
           $354 = $353 << 1;
           $j$118$us$i$i$us$i = 2;$minDiff$319$us$i$i$us$i = 2147483647;
           while(1) {
            $355 = (((($181)) + 276|0) + ($j$118$us$i$i$us$i<<2)|0);
            $356 = HEAP32[$355>>2]|0;
            $357 = $356 << 16;
            $358 = (($357) + ($354))|0;
            $359 = (($341) - ($358))|0;
            $360 = ($359|0)<(0);
            $361 = (0 - ($359))|0;
            $362 = $360 ? $361 : $359;
            $363 = ($362|0)<($minDiff$319$us$i$i$us$i|0);
            $364 = ($362|0)<($312|0);
            $or$cond2$us$i$i$us$i = $363 & $364;
            if ($or$cond2$us$i$i$us$i) {
             HEAP32[$340>>2] = $358;
             $365 = ($362|0)==(0);
             if ($365) {
              break L113;
             } else {
              $minDiff$4$us$i$i$us$i = $362;
             }
            } else {
             $minDiff$4$us$i$i$us$i = $minDiff$319$us$i$i$us$i;
            }
            $366 = (($j$118$us$i$i$us$i) + 2)|0;
            $367 = ($366>>>0)<($198>>>0);
            if ($367) {
             $j$118$us$i$i$us$i = $366;$minDiff$319$us$i$i$us$i = $minDiff$4$us$i$i$us$i;
            } else {
             break;
            }
           }
          } else {
           $345 = HEAP32[$318>>2]|0;
           $346 = $345 << 16;
           $347 = (($341) - ($346))|0;
           $348 = ($347|0)<(0);
           $349 = (0 - ($347))|0;
           $350 = $348 ? $349 : $347;
           $351 = ($350|0)!=(2147483647);
           $352 = ($350|0)<($312|0);
           $or$cond1$us$i$i$us$i = $351 & $352;
           if (!($or$cond1$us$i$i$us$i)) {
            break;
           }
           HEAP32[$340>>2] = $346;
          }
         } while(0);
         $368 = (($i$222$us$i$i$us$i) + 1)|0;
         $369 = HEAP32[$$pre14$i$i>>2]|0;
         $370 = ($368>>>0)<($369>>>0);
         if ($370) {
          $i$222$us$i$i$us$i = $368;
         } else {
          break L87;
         }
        }
       }
       if ($315) {
        $i$222$us$i$us$i$us$i = 0;
       } else {
        $i$222$us$i$us$i$i = 0;
        while(1) {
         $415 = (((((($font$0$in)) + 264|0) + (($i$222$us$i$us$i$i*20)|0)|0)) + 8|0);
         $416 = HEAP32[$415>>2]|0;
         $417 = (((((($font$0$in)) + 264|0) + (($i$222$us$i$us$i$i*20)|0)|0)) + 16|0);
         $418 = HEAP8[$417>>0]|0;
         $419 = ($418<<24>>24)==(0);
         do {
          if (!($419)) {
           $j$013$us$i$us$i$i = 0;$minDiff$014$us$i$us$i$i = 2147483647;
           while(1) {
            $420 = $j$013$us$i$us$i$i | 1;
            $421 = (((($181)) + 332|0) + ($420<<2)|0);
            $422 = HEAP32[$421>>2]|0;
            $423 = $422 << 16;
            $424 = (($416) - ($423))|0;
            $425 = ($424|0)<(0);
            $426 = (0 - ($424))|0;
            $427 = $425 ? $426 : $424;
            $428 = ($427|0)<($minDiff$014$us$i$us$i$i|0);
            $429 = ($427|0)<($312|0);
            $or$cond$us$i$us$i$i = $428 & $429;
            if ($or$cond$us$i$us$i$i) {
             HEAP32[$415>>2] = $423;
             $430 = ($427|0)==(0);
             if ($430) {
              $minDiff$2$us$i$ph$us$i$i = 0;
              break;
             } else {
              $minDiff$1$us$i$us$i$i = $427;
             }
            } else {
             $minDiff$1$us$i$us$i$i = $minDiff$014$us$i$us$i$i;
            }
            $431 = (($j$013$us$i$us$i$i) + 2)|0;
            $432 = ($431>>>0)<($199>>>0);
            if ($432) {
             $j$013$us$i$us$i$i = $431;$minDiff$014$us$i$us$i$i = $minDiff$1$us$i$us$i$i;
            } else {
             $minDiff$2$us$i$ph$us$i$i = $minDiff$1$us$i$us$i$i;
             break;
            }
           }
           $437 = HEAP32[$318>>2]|0;
           $433 = $437 << 16;
           $438 = (($416) - ($433))|0;
           $439 = ($438|0)<(0);
           $440 = (0 - ($438))|0;
           $441 = $439 ? $440 : $438;
           $442 = ($441|0)<($minDiff$2$us$i$ph$us$i$i|0);
           $443 = ($441|0)<($312|0);
           $or$cond1$us$i$us$i$i = $442 & $443;
           if (!($or$cond1$us$i$us$i$i)) {
            break;
           }
           HEAP32[$415>>2] = $433;
          }
         } while(0);
         $434 = (($i$222$us$i$us$i$i) + 1)|0;
         $435 = HEAP32[$$pre14$i$i>>2]|0;
         $436 = ($434>>>0)<($435>>>0);
         if ($436) {
          $i$222$us$i$us$i$i = $434;
         } else {
          break L87;
         }
        }
       }
       while(1) {
        $371 = (((((($font$0$in)) + 264|0) + (($i$222$us$i$us$i$us$i*20)|0)|0)) + 8|0);
        $372 = HEAP32[$371>>2]|0;
        $373 = (((((($font$0$in)) + 264|0) + (($i$222$us$i$us$i$us$i*20)|0)|0)) + 16|0);
        $374 = HEAP8[$373>>0]|0;
        $375 = ($374<<24>>24)==(0);
        L138: do {
         if ($375) {
          $397 = HEAP32[$158>>2]|0;
          $398 = $397 << 1;
          $j$118$us$i$us$i$us$i = 2;$minDiff$319$us$i$us$i$us$i = 2147483647;
          while(1) {
           $399 = (((($181)) + 276|0) + ($j$118$us$i$us$i$us$i<<2)|0);
           $400 = HEAP32[$399>>2]|0;
           $401 = $400 << 16;
           $402 = (($401) + ($398))|0;
           $403 = (($372) - ($402))|0;
           $404 = ($403|0)<(0);
           $405 = (0 - ($403))|0;
           $406 = $404 ? $405 : $403;
           $407 = ($406|0)<($minDiff$319$us$i$us$i$us$i|0);
           $408 = ($406|0)<($312|0);
           $or$cond2$us$i$us$i$us$i = $407 & $408;
           if ($or$cond2$us$i$us$i$us$i) {
            HEAP32[$371>>2] = $402;
            $409 = ($406|0)==(0);
            if ($409) {
             break L138;
            } else {
             $minDiff$4$us$i$us$i$us$i = $406;
            }
           } else {
            $minDiff$4$us$i$us$i$us$i = $minDiff$319$us$i$us$i$us$i;
           }
           $410 = (($j$118$us$i$us$i$us$i) + 2)|0;
           $411 = ($410>>>0)<($198>>>0);
           if ($411) {
            $j$118$us$i$us$i$us$i = $410;$minDiff$319$us$i$us$i$us$i = $minDiff$4$us$i$us$i$us$i;
           } else {
            break;
           }
          }
         } else {
          $j$013$us$i$us$i$us$i = 0;$minDiff$014$us$i$us$i$us$i = 2147483647;
          while(1) {
           $376 = $j$013$us$i$us$i$us$i | 1;
           $377 = (((($181)) + 332|0) + ($376<<2)|0);
           $378 = HEAP32[$377>>2]|0;
           $379 = $378 << 16;
           $380 = (($372) - ($379))|0;
           $381 = ($380|0)<(0);
           $382 = (0 - ($380))|0;
           $383 = $381 ? $382 : $380;
           $384 = ($383|0)<($minDiff$014$us$i$us$i$us$i|0);
           $385 = ($383|0)<($312|0);
           $or$cond$us$i$us$i$us$i = $384 & $385;
           if ($or$cond$us$i$us$i$us$i) {
            HEAP32[$371>>2] = $379;
            $386 = ($383|0)==(0);
            if ($386) {
             $minDiff$2$us$i$ph$us$i$us$i = 0;
             break;
            } else {
             $minDiff$1$us$i$us$i$us$i = $383;
            }
           } else {
            $minDiff$1$us$i$us$i$us$i = $minDiff$014$us$i$us$i$us$i;
           }
           $387 = (($j$013$us$i$us$i$us$i) + 2)|0;
           $388 = ($387>>>0)<($199>>>0);
           if ($388) {
            $j$013$us$i$us$i$us$i = $387;$minDiff$014$us$i$us$i$us$i = $minDiff$1$us$i$us$i$us$i;
           } else {
            $minDiff$2$us$i$ph$us$i$us$i = $minDiff$1$us$i$us$i$us$i;
            break;
           }
          }
          $389 = HEAP32[$318>>2]|0;
          $390 = $389 << 16;
          $391 = (($372) - ($390))|0;
          $392 = ($391|0)<(0);
          $393 = (0 - ($391))|0;
          $394 = $392 ? $393 : $391;
          $395 = ($394|0)<($minDiff$2$us$i$ph$us$i$us$i|0);
          $396 = ($394|0)<($312|0);
          $or$cond1$us$i$us$i$us$i = $395 & $396;
          if (!($or$cond1$us$i$us$i$us$i)) {
           break;
          }
          HEAP32[$371>>2] = $390;
         }
        } while(0);
        $412 = (($i$222$us$i$us$i$us$i) + 1)|0;
        $413 = HEAP32[$$pre14$i$i>>2]|0;
        $414 = ($412>>>0)<($413>>>0);
        if ($414) {
         $i$222$us$i$us$i$us$i = $412;
        } else {
         break;
        }
       }
      }
     } while(0);
     $460 = ($maxZoneHeight$3$lcssa$i$i$i|0)>(0);
     do {
      if ($460) {
       $461 = HEAP32[$173>>2]|0;
       $462 = (_FT_DivFix(65536,$maxZoneHeight$3$lcssa$i$i$i)|0);
       $463 = ($461|0)>($462|0);
       if (!($463)) {
        break;
       }
       $464 = (_FT_DivFix(65536,$maxZoneHeight$3$lcssa$i$i$i)|0);
       HEAP32[$173>>2] = $464;
      }
     } while(0);
     $465 = HEAP32[$169>>2]|0;
     $466 = HEAP32[$173>>2]|0;
     $467 = ($465|0)<($466|0);
     if ($467) {
      $468 = ((($font$0$in)) + 204|0);
      HEAP8[$468>>0] = 1;
      $469 = (_FT_MulDiv(39322,$465,$466)|0);
      $470 = (39322 - ($469))|0;
      $471 = ((($font$0$in)) + 220|0);
      $472 = ($470|0)>(32767);
      $$$i$i$i = $472 ? 32767 : $470;
      HEAP32[$471>>2] = $$$i$i$i;
     }
     $473 = HEAP8[$122>>0]|0;
     $474 = ($473<<24>>24)==(0);
     if (!($474)) {
      $475 = ((($font$0$in)) + 220|0);
      HEAP32[$475>>2] = 0;
     }
     $476 = HEAP32[$$pre14$i$i>>2]|0;
     $477 = ($476|0)==(0);
     if ($477) {
      break;
     }
     $478 = ((($font$0$in)) + 220|0);
     $i$312$i$i$i = 0;
     while(1) {
      $479 = (((((($font$0$in)) + 264|0) + (($i$312$i$i$i*20)|0)|0)) + 16|0);
      $480 = HEAP8[$479>>0]|0;
      $481 = ($480<<24>>24)==(0);
      $482 = (((((($font$0$in)) + 264|0) + (($i$312$i$i$i*20)|0)|0)) + 8|0);
      $483 = HEAP32[$482>>2]|0;
      $484 = HEAP32[$169>>2]|0;
      $485 = (_FT_MulFix($483,$484)|0);
      $486 = HEAP32[$478>>2]|0;
      $487 = (($485) + 32768)|0;
      $488 = (0 - ($486))|0;
      $$sink$in$p$i$i$i = $481 ? $486 : $488;
      $$sink$in$i$i$i = (($487) + ($$sink$in$p$i$i$i))|0;
      $$sink$i$i$i = $$sink$in$i$i$i & -65536;
      $489 = (((((($font$0$in)) + 264|0) + (($i$312$i$i$i*20)|0)|0)) + 12|0);
      HEAP32[$489>>2] = $$sink$i$i$i;
      $490 = (($i$312$i$i$i) + 1)|0;
      $491 = HEAP32[$$pre14$i$i>>2]|0;
      $492 = ($490>>>0)<($491>>>0);
      if ($492) {
       $i$312$i$i$i = $490;
      } else {
       break;
      }
     }
    }
   } while(0);
   $493 = HEAP32[$99>>2]|0;
   $494 = ($493|0)==(0);
   L168: do {
    if ($494) {
     $495 = ((($font$0$in)) + 192|0);
     HEAP8[$495>>0] = 0;
     $496 = ((($font$0$in)) + 141|0);
     $497 = HEAP8[$496>>0]|0;
     $498 = ((($font$0$in)) + 116|0);
     $499 = ((($font$0$in)) + 100|0);
     $needWinding$0$i = $497;
     while(1) {
      $500 = HEAP32[$19>>2]|0;
      HEAP32[$498>>2] = 0;
      $501 = ((($500)) + 12|0);
      $502 = HEAP32[$501>>2]|0;
      _FT_GlyphLoader_Rewind($502);
      _cf2_interpT2CharString($font$0$in,$buf,$499,$translation$i,0,0,0,$advWidth$i);
      $503 = HEAP32[$99>>2]|0;
      $504 = ($503|0)==(0);
      if (!($504)) {
       break L168;
      }
      $505 = ($needWinding$0$i<<24>>24)==(0);
      if ($505) {
       break;
      }
      $506 = HEAP32[$498>>2]|0;
      $507 = ($506|0)>(-1);
      if ($507) {
       break;
      }
      HEAP8[$495>>0] = 1;
      $needWinding$0$i = 0;
     }
     $$idx$val$i = HEAP32[$19>>2]|0;
     $$idx$i1$i = ((($$idx$val$i)) + 20|0);
     $$idx$val$i2$i = HEAP32[$$idx$i1$i>>2]|0;
     $508 = ($$idx$val$i2$i|0)==(0|0);
     do {
      if (!($508)) {
       $509 = HEAP32[$$idx$val$i2$i>>2]|0;
       $510 = $509&65535;
       $511 = ($510<<16>>16)<(2);
       $512 = $509 >>> 16;
       $513 = $512&65535;
       if ($511) {
        $534 = 0;
       } else {
        $sext$i$i$i = $509 << 16;
        $514 = $sext$i$i$i >> 16;
        $515 = (($514) + -2)|0;
        $516 = ((($$idx$val$i2$i)) + 12|0);
        $517 = HEAP32[$516>>2]|0;
        $518 = (($517) + ($515<<1)|0);
        $519 = HEAP16[$518>>1]|0;
        $520 = $519 << 16 >> 16;
        $521 = (($520) + 1)|0;
        $534 = $521;
       }
       $522 = ((($$idx$val$i2$i)) + 2|0);
       $523 = $509 >> 16;
       $524 = ($513<<16>>16)>(1);
       if ($524) {
        $525 = ((($$idx$val$i2$i)) + 4|0);
        $526 = HEAP32[$525>>2]|0;
        $527 = (($526) + ($523<<3)|0);
        $528 = ((($527)) + -8|0);
        $529 = ((($$idx$val$i2$i)) + 8|0);
        $530 = HEAP32[$529>>2]|0;
        $531 = (($530) + ($523)|0);
        $532 = ((($531)) + -1|0);
        $533 = (($526) + ($534<<3)|0);
        $535 = HEAP32[$533>>2]|0;
        $536 = HEAP32[$528>>2]|0;
        $537 = ($535|0)==($536|0);
        if ($537) {
         $538 = (((($526) + ($534<<3)|0)) + 4|0);
         $539 = HEAP32[$538>>2]|0;
         $540 = ((($528)) + 4|0);
         $541 = HEAP32[$540>>2]|0;
         $542 = ($539|0)==($541|0);
         if ($542) {
          $543 = HEAP8[$532>>0]|0;
          $544 = ($543<<24>>24)==(1);
          if ($544) {
           $545 = (($513) + -1)<<16>>16;
           HEAP16[$522>>1] = $545;
           $549 = $545;
          } else {
           $549 = $513;
          }
         } else {
          $549 = $513;
         }
        } else {
         $549 = $513;
        }
       } else {
        $549 = $513;
       }
       $sext2$i$i$i = $509 << 16;
       $546 = $sext2$i$i$i >> 16;
       $547 = ($510<<16>>16)>(0);
       if ($547) {
        $548 = $549 << 16 >> 16;
        $550 = (($548) + -1)|0;
        $551 = ($534|0)==($550|0);
        if ($551) {
         $552 = (($510) + -1)<<16>>16;
         HEAP16[$$idx$val$i2$i>>1] = $552;
         $553 = (($549) + -1)<<16>>16;
         HEAP16[$522>>1] = $553;
         break;
        } else {
         $554 = $550&65535;
         $555 = (($546) + -1)|0;
         $556 = ((($$idx$val$i2$i)) + 12|0);
         $557 = HEAP32[$556>>2]|0;
         $558 = (($557) + ($555<<1)|0);
         HEAP16[$558>>1] = $554;
         break;
        }
       }
      }
     } while(0);
     $559 = ((($$idx$val$i)) + 12|0);
     $560 = HEAP32[$559>>2]|0;
     _FT_GlyphLoader_Add($560);
     $$pr = HEAP32[$99>>2]|0;
     $561 = ($$pr|0)==(0);
     if ($561) {
      $562 = HEAP32[$advWidth$i>>2]|0;
      HEAP32[$99>>2] = 0;
      $$idx$val = HEAP32[$19>>2]|0;
      $563 = (($562) + 32768)|0;
      $564 = $563 >> 16;
      $565 = ((($$idx$val)) + 552|0);
      HEAP32[$565>>2] = $564;
      $$0 = 0;
      break L19;
     }
    }
   } while(0);
   $$0 = 3;
  }
 } while(0);
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0);
}
function _cf2_free_instance($ptr) {
 $ptr = $ptr|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _cf2_builder_moveTo($callbacks,$params) {
 $callbacks = $callbacks|0;
 $params = $params|0;
 var $$idx = 0, $$idx$val = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $sext$i = 0, $sext2$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($callbacks)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $$idx = ((($1)) + 20|0);
 $$idx$val = HEAP32[$$idx>>2]|0;
 $2 = ($$idx$val|0)==(0|0);
 if ($2) {
  $53 = ((($1)) + 64|0);
  HEAP8[$53>>0] = 0;
  return;
 }
 $3 = HEAP32[$$idx$val>>2]|0;
 $4 = $3&65535;
 $5 = ($4<<16>>16)<(2);
 $6 = $3 >>> 16;
 $7 = $6&65535;
 if ($5) {
  $28 = 0;
 } else {
  $sext$i = $3 << 16;
  $8 = $sext$i >> 16;
  $9 = (($8) + -2)|0;
  $10 = ((($$idx$val)) + 12|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + ($9<<1)|0);
  $13 = HEAP16[$12>>1]|0;
  $14 = $13 << 16 >> 16;
  $15 = (($14) + 1)|0;
  $28 = $15;
 }
 $16 = ((($$idx$val)) + 2|0);
 $17 = $3 >> 16;
 $18 = ($7<<16>>16)>(1);
 if ($18) {
  $19 = ((($$idx$val)) + 4|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = (($20) + ($17<<3)|0);
  $22 = ((($21)) + -8|0);
  $23 = ((($$idx$val)) + 8|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = (($24) + ($17)|0);
  $26 = ((($25)) + -1|0);
  $27 = (($20) + ($28<<3)|0);
  $29 = HEAP32[$27>>2]|0;
  $30 = HEAP32[$22>>2]|0;
  $31 = ($29|0)==($30|0);
  if ($31) {
   $32 = (((($20) + ($28<<3)|0)) + 4|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ((($22)) + 4|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = ($33|0)==($35|0);
   if ($36) {
    $37 = HEAP8[$26>>0]|0;
    $38 = ($37<<24>>24)==(1);
    if ($38) {
     $39 = (($7) + -1)<<16>>16;
     HEAP16[$16>>1] = $39;
     $43 = $39;
    } else {
     $43 = $7;
    }
   } else {
    $43 = $7;
   }
  } else {
   $43 = $7;
  }
 } else {
  $43 = $7;
 }
 $sext2$i = $3 << 16;
 $40 = $sext2$i >> 16;
 $41 = ($4<<16>>16)>(0);
 if (!($41)) {
  $53 = ((($1)) + 64|0);
  HEAP8[$53>>0] = 0;
  return;
 }
 $42 = $43 << 16 >> 16;
 $44 = (($42) + -1)|0;
 $45 = ($28|0)==($44|0);
 if ($45) {
  $46 = (($4) + -1)<<16>>16;
  HEAP16[$$idx$val>>1] = $46;
  $47 = (($43) + -1)<<16>>16;
  HEAP16[$16>>1] = $47;
  $53 = ((($1)) + 64|0);
  HEAP8[$53>>0] = 0;
  return;
 } else {
  $48 = $44&65535;
  $49 = (($40) + -1)|0;
  $50 = ((($$idx$val)) + 12|0);
  $51 = HEAP32[$50>>2]|0;
  $52 = (($51) + ($49<<1)|0);
  HEAP16[$52>>1] = $48;
  $53 = ((($1)) + 64|0);
  HEAP8[$53>>0] = 0;
  return;
 }
}
function _cf2_builder_lineTo($callbacks,$params) {
 $callbacks = $callbacks|0;
 $params = $params|0;
 var $$pre$i$i = 0, $$pre$phi$i$iZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $builder$idx$i = 0, $builder$idx$val$i = 0, $builder$idx1$i = 0, $builder$idx1$val$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($callbacks)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 64|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = ($3<<24>>24)==(0);
 if ($4) {
  $5 = HEAP32[$params>>2]|0;
  $6 = ((($params)) + 4|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (_cff_builder_start_point($1,$5,$7)|0);
  $9 = ($8|0)==(0);
  if (!($9)) {
   $10 = ((($callbacks)) + 24|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = HEAP32[$11>>2]|0;
   $13 = ($12|0)==(0);
   if (!($13)) {
    return;
   }
   HEAP32[$11>>2] = $8;
   return;
  }
 }
 $14 = ((($params)) + 8|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($params)) + 12|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ((($1)) + 12|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ((($19)) + 22|0);
 $21 = HEAP16[$20>>1]|0;
 $22 = $21 << 16 >> 16;
 $23 = ((($19)) + 58|0);
 $24 = HEAP16[$23>>1]|0;
 $25 = $24 << 16 >> 16;
 $26 = (($22) + 1)|0;
 $27 = (($26) + ($25))|0;
 $28 = ((($19)) + 4|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = ($27>>>0)>($29>>>0);
 if ($30) {
  $31 = (_FT_GlyphLoader_CheckPoints($19,1,0)|0);
  $32 = ($31|0)==(0);
  if (!($32)) {
   $48 = ((($callbacks)) + 24|0);
   $49 = HEAP32[$48>>2]|0;
   $50 = HEAP32[$49>>2]|0;
   $51 = ($50|0)==(0);
   if (!($51)) {
    return;
   }
   HEAP32[$49>>2] = $31;
   return;
  }
 }
 $builder$idx$i = ((($1)) + 20|0);
 $builder$idx$val$i = HEAP32[$builder$idx$i>>2]|0;
 $builder$idx1$i = ((($1)) + 65|0);
 $builder$idx1$val$i = HEAP8[$builder$idx1$i>>0]|0;
 $33 = ($builder$idx1$val$i<<24>>24)==(0);
 if ($33) {
  $$pre$i$i = ((($builder$idx$val$i)) + 2|0);
  $$pre$phi$i$iZ2D = $$pre$i$i;
 } else {
  $34 = ((($builder$idx$val$i)) + 4|0);
  $35 = HEAP32[$34>>2]|0;
  $36 = ((($builder$idx$val$i)) + 2|0);
  $37 = HEAP16[$36>>1]|0;
  $38 = $37 << 16 >> 16;
  $39 = ((($builder$idx$val$i)) + 8|0);
  $40 = HEAP32[$39>>2]|0;
  $41 = (($40) + ($38)|0);
  $42 = $15 >> 10;
  $43 = (($35) + ($38<<3)|0);
  HEAP32[$43>>2] = $42;
  $44 = $17 >> 10;
  $45 = (((($35) + ($38<<3)|0)) + 4|0);
  HEAP32[$45>>2] = $44;
  HEAP8[$41>>0] = 1;
  $$pre$phi$i$iZ2D = $36;
 }
 $46 = HEAP16[$$pre$phi$i$iZ2D>>1]|0;
 $47 = (($46) + 1)<<16>>16;
 HEAP16[$$pre$phi$i$iZ2D>>1] = $47;
 return;
}
function _cff_builder_start_point($builder,$x,$y) {
 $builder = $builder|0;
 $x = $x|0;
 $y = $y|0;
 var $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$phi$i$iZ2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, $builder$idx$val$i = 0, $builder$idx1$val$i = 0, $error$0 = 0, $sext$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($builder)) + 64|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = $1&255;
 $3 = ($2<<24>>24)==(0);
 if (!($3)) {
  $error$0 = 0;
  return ($error$0|0);
 }
 HEAP8[$0>>0] = 1;
 $4 = ((($builder)) + 20|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($builder)) + 65|0);
 $7 = ($1&65535)<(256);
 if ($7) {
  $8 = HEAP16[$5>>1]|0;
  $9 = (($8) + 1)<<16>>16;
  HEAP16[$5>>1] = $9;
  $$pre = ((($builder)) + 12|0);
  $$pre$phiZ2D = $$pre;
 } else {
  $10 = ((($builder)) + 12|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ((($11)) + 20|0);
  $13 = HEAP16[$12>>1]|0;
  $14 = $13 << 16 >> 16;
  $15 = ((($11)) + 56|0);
  $16 = HEAP16[$15>>1]|0;
  $17 = $16 << 16 >> 16;
  $18 = (($14) + 1)|0;
  $19 = (($18) + ($17))|0;
  $20 = ((($11)) + 8|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = ($19>>>0)>($21>>>0);
  if ($22) {
   $23 = (_FT_GlyphLoader_CheckPoints($11,0,1)|0);
   $24 = ($23|0)==(0);
   if (!($24)) {
    $error$0 = $23;
    return ($error$0|0);
   }
  }
  $25 = HEAP32[$5>>2]|0;
  $26 = $25&65535;
  $27 = ($26<<16>>16)>(0);
  if ($27) {
   $28 = $25 >>> 16;
   $sext$i = $25 << 16;
   $29 = $sext$i >> 16;
   $30 = (($28) + 65535)|0;
   $31 = $30&65535;
   $32 = (($29) + -1)|0;
   $33 = ((($5)) + 12|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = (($34) + ($32<<1)|0);
   HEAP16[$35>>1] = $31;
   $$pre$i = HEAP16[$5>>1]|0;
   $37 = $$pre$i;
  } else {
   $37 = $26;
  }
  $36 = (($37) + 1)<<16>>16;
  HEAP16[$5>>1] = $36;
  $$pre$phiZ2D = $10;
 }
 $38 = HEAP32[$$pre$phiZ2D>>2]|0;
 $39 = ((($38)) + 22|0);
 $40 = HEAP16[$39>>1]|0;
 $41 = $40 << 16 >> 16;
 $42 = ((($38)) + 58|0);
 $43 = HEAP16[$42>>1]|0;
 $44 = $43 << 16 >> 16;
 $45 = (($41) + 1)|0;
 $46 = (($45) + ($44))|0;
 $47 = ((($38)) + 4|0);
 $48 = HEAP32[$47>>2]|0;
 $49 = ($46>>>0)>($48>>>0);
 if ($49) {
  $50 = (_FT_GlyphLoader_CheckPoints($38,1,0)|0);
  $51 = ($50|0)==(0);
  if (!($51)) {
   $error$0 = $50;
   return ($error$0|0);
  }
 }
 $builder$idx$val$i = HEAP32[$4>>2]|0;
 $builder$idx1$val$i = HEAP8[$6>>0]|0;
 $52 = ($builder$idx1$val$i<<24>>24)==(0);
 if ($52) {
  $$pre$i$i = ((($builder$idx$val$i)) + 2|0);
  $$pre$phi$i$iZ2D = $$pre$i$i;
 } else {
  $53 = ((($builder$idx$val$i)) + 4|0);
  $54 = HEAP32[$53>>2]|0;
  $55 = ((($builder$idx$val$i)) + 2|0);
  $56 = HEAP16[$55>>1]|0;
  $57 = $56 << 16 >> 16;
  $58 = ((($builder$idx$val$i)) + 8|0);
  $59 = HEAP32[$58>>2]|0;
  $60 = (($59) + ($57)|0);
  $61 = $x >> 10;
  $62 = (($54) + ($57<<3)|0);
  HEAP32[$62>>2] = $61;
  $63 = $y >> 10;
  $64 = (((($54) + ($57<<3)|0)) + 4|0);
  HEAP32[$64>>2] = $63;
  HEAP8[$60>>0] = 1;
  $$pre$phi$i$iZ2D = $55;
 }
 $65 = HEAP16[$$pre$phi$i$iZ2D>>1]|0;
 $66 = (($65) + 1)<<16>>16;
 HEAP16[$$pre$phi$i$iZ2D>>1] = $66;
 $error$0 = 0;
 return ($error$0|0);
}
function _cf2_builder_cubeTo($callbacks,$params) {
 $callbacks = $callbacks|0;
 $params = $params|0;
 var $$idx$val$pre = 0, $$idx$val20 = 0, $$idx1$val$pre = 0, $$idx2$val$pre = 0, $$idx2$val19 = 0, $$idx3$val$pre = 0, $$idx4 = 0, $$idx4$val = 0, $$idx5 = 0, $$idx5$val = 0, $$pre$i = 0, $$pre$i10 = 0, $$pre$i6 = 0, $$pre$phi$i8Z2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $phitmp = 0, $phitmp18 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($callbacks)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 64|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = ($3<<24>>24)==(0);
 if ($4) {
  $5 = HEAP32[$params>>2]|0;
  $6 = ((($params)) + 4|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (_cff_builder_start_point($1,$5,$7)|0);
  $9 = ($8|0)==(0);
  if (!($9)) {
   $10 = ((($callbacks)) + 24|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = HEAP32[$11>>2]|0;
   $13 = ($12|0)==(0);
   if (!($13)) {
    return;
   }
   HEAP32[$11>>2] = $8;
   return;
  }
 }
 $14 = ((($1)) + 12|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($15)) + 22|0);
 $17 = HEAP16[$16>>1]|0;
 $18 = $17 << 16 >> 16;
 $19 = ((($15)) + 58|0);
 $20 = HEAP16[$19>>1]|0;
 $21 = $20 << 16 >> 16;
 $22 = (($18) + 3)|0;
 $23 = (($22) + ($21))|0;
 $24 = ((($15)) + 4|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = ($23>>>0)>($25>>>0);
 if ($26) {
  $27 = (_FT_GlyphLoader_CheckPoints($15,3,0)|0);
  $28 = ($27|0)==(0);
  if (!($28)) {
   $29 = ((($callbacks)) + 24|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = HEAP32[$30>>2]|0;
   $32 = ($31|0)==(0);
   if (!($32)) {
    return;
   }
   HEAP32[$30>>2] = $27;
   return;
  }
 }
 $$idx4 = ((($1)) + 20|0);
 $$idx4$val = HEAP32[$$idx4>>2]|0;
 $$idx5 = ((($1)) + 65|0);
 $$idx5$val = HEAP8[$$idx5>>0]|0;
 $33 = ($$idx5$val<<24>>24)==(0);
 if ($33) {
  $$pre$i = ((($$idx4$val)) + 2|0);
  $34 = HEAP16[$$pre$i>>1]|0;
  $35 = (($34) + 1)<<16>>16;
  HEAP16[$$pre$i>>1] = $35;
  $$idx2$val19 = $$idx4$val;
  label = 12;
 } else {
  $36 = ((($params)) + 12|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = ((($params)) + 8|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = ((($$idx4$val)) + 4|0);
  $41 = HEAP32[$40>>2]|0;
  $42 = ((($$idx4$val)) + 2|0);
  $43 = HEAP16[$42>>1]|0;
  $44 = $43 << 16 >> 16;
  $45 = ((($$idx4$val)) + 8|0);
  $46 = HEAP32[$45>>2]|0;
  $47 = (($46) + ($44)|0);
  $48 = $39 >> 10;
  $49 = (($41) + ($44<<3)|0);
  HEAP32[$49>>2] = $48;
  $50 = $37 >> 10;
  $51 = (((($41) + ($44<<3)|0)) + 4|0);
  HEAP32[$51>>2] = $50;
  HEAP8[$47>>0] = 2;
  $$idx2$val$pre = HEAP32[$$idx4>>2]|0;
  $$idx3$val$pre = HEAP8[$$idx5>>0]|0;
  $phitmp = ($$idx3$val$pre<<24>>24)==(0);
  $52 = HEAP16[$42>>1]|0;
  $53 = (($52) + 1)<<16>>16;
  HEAP16[$42>>1] = $53;
  if ($phitmp) {
   $$idx2$val19 = $$idx2$val$pre;
   label = 12;
  } else {
   $56 = ((($params)) + 20|0);
   $57 = HEAP32[$56>>2]|0;
   $58 = ((($params)) + 16|0);
   $59 = HEAP32[$58>>2]|0;
   $60 = ((($$idx2$val$pre)) + 4|0);
   $61 = HEAP32[$60>>2]|0;
   $62 = ((($$idx2$val$pre)) + 2|0);
   $63 = HEAP16[$62>>1]|0;
   $64 = $63 << 16 >> 16;
   $65 = ((($$idx2$val$pre)) + 8|0);
   $66 = HEAP32[$65>>2]|0;
   $67 = (($66) + ($64)|0);
   $68 = $59 >> 10;
   $69 = (($61) + ($64<<3)|0);
   HEAP32[$69>>2] = $68;
   $70 = $57 >> 10;
   $71 = (((($61) + ($64<<3)|0)) + 4|0);
   HEAP32[$71>>2] = $70;
   HEAP8[$67>>0] = 2;
   $$idx$val$pre = HEAP32[$$idx4>>2]|0;
   $$idx1$val$pre = HEAP8[$$idx5>>0]|0;
   $phitmp18 = ($$idx1$val$pre<<24>>24)==(0);
   $72 = HEAP16[$62>>1]|0;
   $73 = (($72) + 1)<<16>>16;
   HEAP16[$62>>1] = $73;
   if ($phitmp18) {
    $$idx$val20 = $$idx$val$pre;
    label = 14;
   } else {
    $74 = ((($params)) + 28|0);
    $75 = HEAP32[$74>>2]|0;
    $76 = ((($params)) + 24|0);
    $77 = HEAP32[$76>>2]|0;
    $78 = ((($$idx$val$pre)) + 4|0);
    $79 = HEAP32[$78>>2]|0;
    $80 = ((($$idx$val$pre)) + 2|0);
    $81 = HEAP16[$80>>1]|0;
    $82 = $81 << 16 >> 16;
    $83 = ((($$idx$val$pre)) + 8|0);
    $84 = HEAP32[$83>>2]|0;
    $85 = (($84) + ($82)|0);
    $86 = $77 >> 10;
    $87 = (($79) + ($82<<3)|0);
    HEAP32[$87>>2] = $86;
    $88 = $75 >> 10;
    $89 = (((($79) + ($82<<3)|0)) + 4|0);
    HEAP32[$89>>2] = $88;
    HEAP8[$85>>0] = 1;
    $$pre$phi$i8Z2D = $80;
   }
  }
 }
 if ((label|0) == 12) {
  $$pre$i10 = ((($$idx2$val19)) + 2|0);
  $54 = HEAP16[$$pre$i10>>1]|0;
  $55 = (($54) + 1)<<16>>16;
  HEAP16[$$pre$i10>>1] = $55;
  $$idx$val20 = $$idx2$val19;
  label = 14;
 }
 if ((label|0) == 14) {
  $$pre$i6 = ((($$idx$val20)) + 2|0);
  $$pre$phi$i8Z2D = $$pre$i6;
 }
 $90 = HEAP16[$$pre$phi$i8Z2D>>1]|0;
 $91 = (($90) + 1)<<16>>16;
 HEAP16[$$pre$phi$i8Z2D>>1] = $91;
 return;
}
function _cf2_computeDarkening($emRatio,$ppem,$stemWidth,$darkenAmount,$boldenAmount,$stemDarkened,$darkenParams) {
 $emRatio = $emRatio|0;
 $ppem = $ppem|0;
 $stemWidth = $stemWidth|0;
 $darkenAmount = $darkenAmount|0;
 $boldenAmount = $boldenAmount|0;
 $stemDarkened = $stemDarkened|0;
 $darkenParams = $darkenParams|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond$not = 0, $or$cond6 = 0, $scaledStem$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$darkenAmount>>2] = 0;
 $0 = ($boldenAmount|0)!=(0);
 $1 = ($stemDarkened<<24>>24)!=(0);
 $or$cond = $0 | $1;
 $or$cond$not = $or$cond ^ 1;
 $2 = ($emRatio|0)<(655);
 $or$cond6 = $2 | $or$cond$not;
 if ($or$cond6) {
  return;
 }
 if ($1) {
  $3 = HEAP32[$darkenParams>>2]|0;
  $4 = ((($darkenParams)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ((($darkenParams)) + 8|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ((($darkenParams)) + 12|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($darkenParams)) + 16|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ((($darkenParams)) + 20|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ((($darkenParams)) + 24|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ((($darkenParams)) + 28|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = (($boldenAmount) + ($stemWidth))|0;
  $19 = (_FT_MulFix($18,$emRatio)|0);
  $20 = (Math_clz32(($19|0))|0);
  $21 = (Math_clz32(($ppem|0))|0);
  $22 = (62 - ($21))|0;
  $23 = (($22) - ($20))|0;
  $24 = ($23|0)>(45);
  if ($24) {
   $25 = $15 << 16;
   $scaledStem$0 = $25;
  } else {
   $26 = (_FT_MulFix($19,$ppem)|0);
   $scaledStem$0 = $26;
  }
  $27 = $3 << 16;
  $28 = ($scaledStem$0|0)<($27|0);
  L10: do {
   if ($28) {
    $29 = $5 << 16;
    $30 = (_FT_DivFix($29,$ppem)|0);
    HEAP32[$darkenAmount>>2] = $30;
    $67 = $30;
   } else {
    $31 = $7 << 16;
    $32 = ($scaledStem$0|0)<($31|0);
    if ($32) {
     $33 = (_FT_DivFix($27,$ppem)|0);
     $34 = ($7|0)==($3|0);
     if ($34) {
      label = 12;
     } else {
      $35 = (($7) - ($3))|0;
      $36 = (($19) - ($33))|0;
      $37 = (($9) - ($5))|0;
      $38 = (_FT_MulDiv($36,$37,$35)|0);
      $39 = $5 << 16;
      $40 = (_FT_DivFix($39,$ppem)|0);
      $41 = (($40) + ($38))|0;
      HEAP32[$darkenAmount>>2] = $41;
      $67 = $41;
      break;
     }
    } else {
     $42 = $11 << 16;
     $43 = ($scaledStem$0|0)<($42|0);
     if ($43) {
      label = 12;
     } else {
      $53 = $15 << 16;
      $54 = ($scaledStem$0|0)<($53|0);
      if ($54) {
       $$pre$phiZ2D = $42;
       label = 16;
      }
     }
    }
    do {
     if ((label|0) == 12) {
      $44 = (_FT_DivFix($31,$ppem)|0);
      $45 = ($11|0)==($7|0);
      if ($45) {
       $$pre = $11 << 16;
       $$pre$phiZ2D = $$pre;
       label = 16;
       break;
      } else {
       $46 = (($11) - ($7))|0;
       $47 = (($19) - ($44))|0;
       $48 = (($13) - ($9))|0;
       $49 = (_FT_MulDiv($47,$48,$46)|0);
       $50 = $9 << 16;
       $51 = (_FT_DivFix($50,$ppem)|0);
       $52 = (($51) + ($49))|0;
       HEAP32[$darkenAmount>>2] = $52;
       $67 = $52;
       break L10;
      }
     }
    } while(0);
    if ((label|0) == 16) {
     $55 = (_FT_DivFix($$pre$phiZ2D,$ppem)|0);
     $56 = ($15|0)==($11|0);
     if (!($56)) {
      $57 = (($15) - ($11))|0;
      $58 = (($19) - ($55))|0;
      $59 = (($17) - ($13))|0;
      $60 = (_FT_MulDiv($58,$59,$57)|0);
      $61 = $13 << 16;
      $62 = (_FT_DivFix($61,$ppem)|0);
      $63 = (($62) + ($60))|0;
      HEAP32[$darkenAmount>>2] = $63;
      $67 = $63;
      break;
     }
    }
    $64 = $17 << 16;
    $65 = (_FT_DivFix($64,$ppem)|0);
    HEAP32[$darkenAmount>>2] = $65;
    $67 = $65;
   }
  } while(0);
  $66 = $emRatio << 1;
  $68 = (_FT_DivFix($67,$66)|0);
  HEAP32[$darkenAmount>>2] = $68;
  $71 = $68;
 } else {
  $71 = 0;
 }
 $69 = (($boldenAmount|0) / 2)&-1;
 $70 = (($71) + ($69))|0;
 HEAP32[$darkenAmount>>2] = $70;
 return;
}
function _cf2_interpT2CharString($font,$buf,$callbacks,$translation,$doingSeac,$curX,$curY,$width) {
 $font = $font|0;
 $buf = $buf|0;
 $callbacks = $callbacks|0;
 $translation = $translation|0;
 $doingSeac = $doingSeac|0;
 $curX = $curX|0;
 $curY = $curY|0;
 $width = $width|0;
 var $$ = 0, $$$i = 0, $$0$i = 0, $$0$i1$i = 0, $$0$i1$i143 = 0, $$0$i109 = 0, $$0$i124 = 0, $$0$i129 = 0, $$0$i131 = 0, $$0$i133 = 0, $$0$i135 = 0, $$0$i137 = 0, $$0$i146 = 0, $$0$i148 = 0, $$0$i150$ph = 0, $$0$i151 = 0, $$0$i152 = 0, $$0$i154 = 0, $$0$i158 = 0, $$0$i160 = 0;
 var $$0$i162 = 0, $$0$i173$ph = 0, $$0$i176 = 0, $$0$i178 = 0, $$0$i180 = 0, $$0$i182 = 0, $$0$i184 = 0, $$0$i186 = 0, $$0$i187 = 0, $$0$i19 = 0, $$0$i194 = 0, $$0$i195 = 0, $$0$i197 = 0, $$0$i201 = 0, $$0$i21 = 0, $$0$i23 = 0, $$0$i25 = 0, $$0$i28 = 0, $$0$i32 = 0, $$0$i34 = 0;
 var $$0$i36 = 0, $$0$i37 = 0, $$0$i39 = 0, $$0$i41 = 0, $$0$i43 = 0, $$0$i45 = 0, $$0$i47 = 0, $$0$i49 = 0, $$0$i51 = 0, $$0$i53 = 0, $$0$i55 = 0, $$0$i57 = 0, $$0$i59 = 0, $$0$i61 = 0, $$0$i63 = 0, $$0$i65 = 0, $$0$i67 = 0, $$0$i69 = 0, $$0$i71 = 0, $$0$i73 = 0;
 var $$0$i77 = 0, $$0$i79 = 0, $$0$i81 = 0, $$0$i83 = 0, $$0$i85 = 0, $$0$i87 = 0, $$0$i89 = 0, $$0$i91 = 0, $$0$i96 = 0, $$0$i98 = 0, $$0$i99 = 0, $$0$val$i = 0, $$0$val$i112 = 0, $$0$val$i125 = 0, $$0$val$i190 = 0, $$1$i126 = 0, $$1$val$i = 0, $$1$val$i114 = 0, $$1$val$i127 = 0, $$1$val$i192 = 0;
 var $$cast$i = 0, $$cast$i172 = 0, $$idx = 0, $$idx$i = 0, $$idx$i105 = 0, $$idx$i120 = 0, $$idx$val = 0, $$idx$val$i = 0, $$idx$val$i106 = 0, $$idx$val$i121 = 0, $$idx$val$idx = 0, $$idx$val$idx$val = 0, $$idx1$i116 = 0, $$idx1$val$i = 0, $$idx1$val$i102 = 0, $$idx1$val$i117 = 0, $$idx11$val = 0, $$idx11$val$idx = 0, $$idx11$val$idx$val = 0, $$idx17$val = 0;
 var $$idx18$val = 0, $$in = 0, $$old = 0, $$old6 = 0, $$ph = 0, $$pr = 0, $$pre = 0, $$pre$i104 = 0, $$pre$i119 = 0, $$pre$i93 = 0, $$pre249 = 0, $$pre250 = 0, $$pre251 = 0, $$pre252 = 0, $$pre253 = 0, $$pre254 = 0, $$pre255 = 0, $$pre256 = 0, $$pre257 = 0, $$pre258 = 0;
 var $$pre259 = 0, $$pre260 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0;
 var $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0;
 var $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0;
 var $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0;
 var $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0;
 var $1086 = 0, $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0;
 var $1103 = 0, $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0;
 var $1121 = 0, $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0;
 var $114 = 0, $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0;
 var $1158 = 0, $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0;
 var $1176 = 0, $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0;
 var $1194 = 0, $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0;
 var $1211 = 0, $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0;
 var $123 = 0, $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0;
 var $1248 = 0, $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0;
 var $1266 = 0, $1267 = 0, $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0;
 var $1284 = 0, $1285 = 0, $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0;
 var $1301 = 0, $1302 = 0, $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0;
 var $132 = 0, $1320 = 0, $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0;
 var $1338 = 0, $1339 = 0, $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0;
 var $1356 = 0, $1357 = 0, $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0;
 var $1374 = 0, $1375 = 0, $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0;
 var $1392 = 0, $1393 = 0, $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0;
 var $141 = 0, $1410 = 0, $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0;
 var $1428 = 0, $1429 = 0, $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0;
 var $1446 = 0, $1447 = 0, $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0;
 var $1464 = 0, $1465 = 0, $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0;
 var $1482 = 0, $1483 = 0, $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0;
 var $150 = 0, $1500 = 0, $1501 = 0, $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0, $1511 = 0, $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0;
 var $1518 = 0, $1519 = 0, $152 = 0, $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0, $153 = 0, $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0;
 var $1536 = 0, $1537 = 0, $1538 = 0, $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0, $1548 = 0, $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0;
 var $1554 = 0, $1555 = 0, $1556 = 0, $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $1565 = 0, $1566 = 0, $1567 = 0, $1568 = 0, $1569 = 0, $157 = 0, $1570 = 0, $1571 = 0;
 var $1572 = 0, $1573 = 0, $1574 = 0, $1575 = 0, $1576 = 0, $1577 = 0, $1578 = 0, $1579 = 0, $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0, $1583 = 0, $1584 = 0, $1585 = 0, $1586 = 0, $1587 = 0, $1588 = 0, $1589 = 0, $159 = 0;
 var $1590 = 0, $1591 = 0, $1592 = 0, $1593 = 0, $1594 = 0, $1595 = 0, $1596 = 0, $1597 = 0, $1598 = 0, $1599 = 0, $16 = 0, $160 = 0, $1600 = 0, $1601 = 0, $1602 = 0, $1603 = 0, $1604 = 0, $1605 = 0, $1606 = 0, $1607 = 0;
 var $1608 = 0, $1609 = 0, $161 = 0, $1610 = 0, $1611 = 0, $1612 = 0, $1613 = 0, $1614 = 0, $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0, $1619 = 0, $162 = 0, $1620 = 0, $1621 = 0, $1622 = 0, $1623 = 0, $1624 = 0, $1625 = 0;
 var $1626 = 0, $1627 = 0, $1628 = 0, $1629 = 0, $163 = 0, $1630 = 0, $1631 = 0, $1632 = 0, $1633 = 0, $1634 = 0, $1635 = 0, $1636 = 0, $1637 = 0, $1638 = 0, $1639 = 0, $164 = 0, $1640 = 0, $1641 = 0, $1642 = 0, $1643 = 0;
 var $1644 = 0, $1645 = 0, $1646 = 0, $1647 = 0, $1648 = 0, $1649 = 0, $165 = 0, $1650 = 0, $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0, $1655 = 0, $1656 = 0, $1657 = 0, $1658 = 0, $1659 = 0, $166 = 0, $1660 = 0, $1661 = 0;
 var $1662 = 0, $1663 = 0, $1664 = 0, $1665 = 0, $1666 = 0, $1667 = 0, $1668 = 0, $1669 = 0, $167 = 0, $1670 = 0, $1671 = 0, $1672 = 0, $1673 = 0, $1674 = 0, $1675 = 0, $1676 = 0, $1677 = 0, $1678 = 0, $1679 = 0, $168 = 0;
 var $1680 = 0, $1681 = 0, $1682 = 0, $1683 = 0, $1684 = 0, $1685 = 0, $1686 = 0, $1687 = 0, $1688 = 0, $1689 = 0, $169 = 0, $1690 = 0, $1691 = 0, $1692 = 0, $1693 = 0, $1694 = 0, $1695 = 0, $1696 = 0, $1697 = 0, $1698 = 0;
 var $1699 = 0, $17 = 0, $170 = 0, $1700 = 0, $1701 = 0, $1702 = 0, $1703 = 0, $1704 = 0, $1705 = 0, $1706 = 0, $1707 = 0, $1708 = 0, $1709 = 0, $171 = 0, $1710 = 0, $1711 = 0, $1712 = 0, $1713 = 0, $1714 = 0, $1715 = 0;
 var $1716 = 0, $1717 = 0, $1718 = 0, $1719 = 0, $172 = 0, $1720 = 0, $1721 = 0, $1722 = 0, $1723 = 0, $1724 = 0, $1725 = 0, $1726 = 0, $1727 = 0, $1728 = 0, $1729 = 0, $173 = 0, $1730 = 0, $1731 = 0, $1732 = 0, $1733 = 0;
 var $1734 = 0, $1735 = 0, $1736 = 0, $1737 = 0, $1738 = 0, $1739 = 0, $174 = 0, $1740 = 0, $1741 = 0, $1742 = 0, $1743 = 0, $1744 = 0, $1745 = 0, $1746 = 0, $1747 = 0, $1748 = 0, $1749 = 0, $175 = 0, $1750 = 0, $1751 = 0;
 var $1752 = 0, $1753 = 0, $1754 = 0, $1755 = 0, $1756 = 0, $1757 = 0, $1758 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0;
 var $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0;
 var $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0;
 var $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0;
 var $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0;
 var $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0;
 var $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0;
 var $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0;
 var $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0;
 var $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0;
 var $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0;
 var $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0;
 var $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0;
 var $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0;
 var $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0;
 var $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0;
 var $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0;
 var $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0;
 var $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0;
 var $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0;
 var $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0;
 var $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0;
 var $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0;
 var $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0;
 var $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0;
 var $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0;
 var $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0;
 var $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0;
 var $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0;
 var $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0;
 var $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0;
 var $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0;
 var $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0;
 var $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0;
 var $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0;
 var $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0;
 var $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0;
 var $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0;
 var $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0;
 var $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0;
 var $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0;
 var $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0;
 var $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0;
 var $999 = 0, $alternate$0213 = 0, $alternate$1 = 0, $charstring$0 = 0, $charstring$0$idx = 0, $charstring$0$idx$val = 0, $charstring$0$idx12 = 0, $charstring$0$idx12$val = 0, $charstring$0$phi = 0, $charstring$i = 0, $charstring$i164 = 0, $charstringIndex$0 = 0, $charstringIndex$0$phi = 0, $component = 0, $cond = 0, $counterHintMap = 0, $counterMask = 0, $data$i = 0, $data$i$i = 0, $data$i$i156 = 0;
 var $data$i175 = 0, $dummyWidth = 0, $error$i = 0, $error$i$i = 0, $exitcond = 0, $gid$0$i171261 = 0, $gid$0$i263 = 0, $glyphPath = 0, $hStemHintArray = 0, $hStemHintArray$idx$val = 0, $hStemHintArray$idx16$val = 0, $haveWidth = 0, $hintMask = 0, $hintMask$idx$val = 0, $hintMask$idx13$val = 0, $hintMask$idx14 = 0, $hintMask$idx14$val = 0, $i$02$i = 0, $i$02$i142 = 0, $index$0238 = 0;
 var $index1$0234 = 0, $index16$0218 = 0, $index16$1 = 0, $index25$0214 = 0, $index25$1 = 0, $index34$0212 = 0, $index34$3 = 0, $index4$0$lcssa = 0, $index4$0228 = 0, $index4$0228$phi = 0, $index8$0$lcssa = 0, $index8$0222 = 0, $index8$1225 = 0, $instructionLimit$0 = 0, $isX$0$in233 = 0, $lastError$2 = 0, $lastError$3$ph = 0, $len$i = 0, $len$i165 = 0, $n$02$i$i = 0;
 var $n$02$i$i$lcssa = 0, $n$02$i$i168 = 0, $n$02$i$i168$lcssa = 0, $op1$0 = 0, $or$cond = 0, $or$cond$i = 0, $or$cond$i$i = 0, $or$cond$i$i166 = 0, $or$cond$i139 = 0, $or$cond244 = 0, $or$cond3 = 0, $or$cond5 = 0, $or$cond7 = 0, $sext = 0, $subrStack$sroa$19$0 = 0, $subrStack$sroa$31$3$ph = 0, $subrStack$sroa$31$3209 = 0, $vStemHintArray = 0, $vStemHintArray$idx$val = 0, $vStemHintArray$idx15$val = 0;
 var $x117$0 = 0, $x135$0 = 0, $x236$0 = 0, $x337$1 = 0, $y127$0 = 0, $y138$0 = 0, $y239$0 = 0, $y340$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 15872|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $data$i175 = sp + 15856|0;
 $charstring$i164 = sp + 15852|0;
 $len$i165 = sp + 15848|0;
 $data$i$i156 = sp + 15840|0;
 $data$i = sp + 15832|0;
 $charstring$i = sp + 15828|0;
 $len$i = sp + 15824|0;
 $data$i$i = sp + 15816|0;
 $error$i$i = sp + 15812|0;
 $error$i = sp + 15808|0;
 $0 = sp + 15804|0;
 $1 = sp + 15800|0;
 $haveWidth = sp + 15864|0;
 $hStemHintArray = sp + 15768|0;
 $vStemHintArray = sp + 15736|0;
 $hintMask = sp + 15704|0;
 $glyphPath = sp + 3920|0;
 $component = sp + 3904|0;
 $dummyWidth = sp + 3896|0;
 $counterHintMap = sp + 28|0;
 $counterMask = sp;
 HEAP32[$0>>2] = $curX;
 HEAP32[$1>>2] = $curY;
 $2 = ((($font)) + 132|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($font)) + 4|0);
 $5 = HEAP32[$font>>2]|0;
 $6 = ((($font)) + 48|0);
 $7 = HEAP32[$6>>2]|0;
 $$idx = ((($3)) + 736|0);
 $$idx$val = HEAP32[$$idx>>2]|0;
 $$idx$val$idx = ((($$idx$val)) + 532|0);
 $$idx$val$idx$val = HEAP32[$$idx$val$idx>>2]|0;
 $8 = $$idx$val$idx$val << 16;
 HEAP32[$hStemHintArray>>2] = $5;
 $9 = ((($hStemHintArray)) + 4|0);
 HEAP32[$9>>2] = $4;
 $10 = ((($hStemHintArray)) + 8|0);
 HEAP32[$10>>2] = 20;
 $11 = ((($hStemHintArray)) + 12|0);
 HEAP32[$11>>2] = 0;
 $12 = ((($hStemHintArray)) + 16|0);
 HEAP32[$12>>2] = 10;
 $13 = ((($hStemHintArray)) + 20|0);
 HEAP32[$13>>2] = 0;
 $14 = ((($hStemHintArray)) + 24|0);
 HEAP32[$14>>2] = 0;
 $15 = ((($hStemHintArray)) + 28|0);
 HEAP32[$15>>2] = 0;
 HEAP32[$vStemHintArray>>2] = $5;
 $16 = ((($vStemHintArray)) + 4|0);
 HEAP32[$16>>2] = $4;
 $17 = ((($vStemHintArray)) + 8|0);
 HEAP32[$17>>2] = 20;
 $18 = ((($vStemHintArray)) + 12|0);
 HEAP32[$18>>2] = 0;
 $19 = ((($vStemHintArray)) + 16|0);
 HEAP32[$19>>2] = 10;
 $20 = ((($vStemHintArray)) + 20|0);
 HEAP32[$20>>2] = 0;
 $21 = ((($vStemHintArray)) + 24|0);
 HEAP32[$21>>2] = 0;
 $22 = ((($vStemHintArray)) + 28|0);
 HEAP32[$22>>2] = 0;
 ;HEAP32[$hintMask>>2]=0|0;HEAP32[$hintMask+4>>2]=0|0;HEAP32[$hintMask+8>>2]=0|0;HEAP32[$hintMask+12>>2]=0|0;HEAP32[$hintMask+16>>2]=0|0;HEAP32[$hintMask+20>>2]=0|0;HEAP32[$hintMask+24>>2]=0|0;
 HEAP32[$hintMask>>2] = $4;
 $23 = ((($font)) + 196|0);
 _memset(($glyphPath|0),0,11784)|0;
 HEAP32[$glyphPath>>2] = $font;
 $24 = ((($glyphPath)) + 4|0);
 HEAP32[$24>>2] = $callbacks;
 $25 = ((($glyphPath)) + 11612|0);
 $26 = $5;
 HEAP32[$25>>2] = $26;
 $27 = ((($glyphPath)) + 11616|0);
 HEAP32[$27>>2] = $4;
 $28 = ((($glyphPath)) + 11620|0);
 HEAP32[$28>>2] = 8;
 $29 = ((($glyphPath)) + 11624|0);
 HEAP32[$29>>2] = 0;
 $30 = ((($glyphPath)) + 11628|0);
 HEAP32[$30>>2] = 10;
 $31 = ((($glyphPath)) + 11632|0);
 HEAP32[$31>>2] = 0;
 $32 = ((($glyphPath)) + 11636|0);
 HEAP32[$32>>2] = 0;
 $33 = ((($glyphPath)) + 11640|0);
 HEAP32[$33>>2] = 0;
 $34 = ((($glyphPath)) + 7744|0);
 _memset(($34|0),0,3868)|0;
 $35 = ((($font)) + 140|0);
 $36 = HEAP16[$35>>1]|0;
 $37 = $36&255;
 $38 = ((($glyphPath)) + 7757|0);
 HEAP8[$38>>0] = $37;
 $39 = ((($glyphPath)) + 7760|0);
 HEAP32[$39>>2] = $7;
 HEAP32[$34>>2] = $font;
 $40 = ((($glyphPath)) + 7748|0);
 HEAP32[$40>>2] = $34;
 $41 = ((($glyphPath)) + 7752|0);
 HEAP32[$41>>2] = $25;
 $42 = ((($glyphPath)) + 3876|0);
 _memset(($42|0),0,3868)|0;
 $43 = ((($glyphPath)) + 3889|0);
 HEAP8[$43>>0] = $37;
 $44 = ((($glyphPath)) + 3892|0);
 HEAP32[$44>>2] = $7;
 HEAP32[$42>>2] = $font;
 $45 = ((($glyphPath)) + 3880|0);
 HEAP32[$45>>2] = $34;
 $46 = ((($glyphPath)) + 3884|0);
 HEAP32[$46>>2] = $25;
 $47 = ((($glyphPath)) + 8|0);
 _memset(($47|0),0,3868)|0;
 $48 = ((($glyphPath)) + 21|0);
 HEAP8[$48>>0] = $37;
 $49 = ((($glyphPath)) + 24|0);
 HEAP32[$49>>2] = $7;
 HEAP32[$47>>2] = $font;
 $50 = ((($glyphPath)) + 12|0);
 HEAP32[$50>>2] = $34;
 $51 = ((($glyphPath)) + 16|0);
 HEAP32[$51>>2] = $25;
 $52 = ((($font)) + 36|0);
 $53 = HEAP32[$52>>2]|0;
 $54 = ((($glyphPath)) + 11644|0);
 HEAP32[$54>>2] = $53;
 $55 = ((($font)) + 44|0);
 $56 = HEAP32[$55>>2]|0;
 $57 = ((($glyphPath)) + 11648|0);
 HEAP32[$57>>2] = $56;
 $58 = HEAP32[$6>>2]|0;
 $59 = ((($glyphPath)) + 11652|0);
 HEAP32[$59>>2] = $58;
 $60 = ((($glyphPath)) + 11656|0);
 $61 = $translation;
 $62 = $61;
 $63 = HEAP32[$62>>2]|0;
 $64 = (($61) + 4)|0;
 $65 = $64;
 $66 = HEAP32[$65>>2]|0;
 $67 = $60;
 $68 = $67;
 HEAP32[$68>>2] = $63;
 $69 = (($67) + 4)|0;
 $70 = $69;
 HEAP32[$70>>2] = $66;
 $71 = ((($glyphPath)) + 11668|0);
 HEAP32[$71>>2] = $hStemHintArray;
 $72 = ((($glyphPath)) + 11672|0);
 HEAP32[$72>>2] = $vStemHintArray;
 $73 = ((($glyphPath)) + 11676|0);
 HEAP32[$73>>2] = $hintMask;
 $74 = ((($glyphPath)) + 11680|0);
 HEAP32[$74>>2] = $curY;
 $75 = ((($glyphPath)) + 11684|0);
 HEAP32[$75>>2] = $23;
 $76 = ($36&65535) >>> 8;
 $77 = $76&255;
 $78 = ((($glyphPath)) + 11666|0);
 HEAP8[$78>>0] = $77;
 $79 = ((($font)) + 184|0);
 $80 = HEAP32[$79>>2]|0;
 $81 = ((($glyphPath)) + 11688|0);
 HEAP32[$81>>2] = $80;
 $82 = ((($font)) + 188|0);
 $83 = HEAP32[$82>>2]|0;
 $84 = ((($glyphPath)) + 11692|0);
 HEAP32[$84>>2] = $83;
 $85 = ($80|0)<(0);
 $86 = (0 - ($80))|0;
 $87 = $85 ? $86 : $80;
 $88 = ($83|0)<(0);
 $89 = (0 - ($83))|0;
 $90 = $88 ? $89 : $83;
 $91 = ($87|0)>($90|0);
 $$$i = $91 ? $87 : $90;
 $92 = $$$i << 1;
 $93 = ((($glyphPath)) + 11696|0);
 HEAP32[$93>>2] = $92;
 $94 = ((($glyphPath)) + 11700|0);
 HEAP32[$94>>2] = 6554;
 $95 = ((($glyphPath)) + 11667|0);
 HEAP8[$95>>0] = 1;
 $96 = ((($glyphPath)) + 11664|0);
 HEAP8[$96>>0] = 0;
 $97 = ((($glyphPath)) + 11665|0);
 HEAP8[$97>>0] = 0;
 $98 = ((($glyphPath)) + 11744|0);
 HEAP8[$98>>0] = 0;
 HEAP8[$haveWidth>>0] = 0;
 $$idx11$val = HEAP32[$$idx>>2]|0;
 $$idx11$val$idx = ((($$idx11$val)) + 528|0);
 $$idx11$val$idx$val = HEAP32[$$idx11$val$idx>>2]|0;
 $99 = $$idx11$val$idx$val << 16;
 HEAP32[$width>>2] = $99;
 HEAP32[$error$i>>2] = 0;
 $100 = (_ft_mem_qalloc($5,396,$error$i)|0);
 $101 = HEAP32[$error$i>>2]|0;
 $102 = ($101|0)==(0);
 if ($102) {
  HEAP32[$100>>2] = $5;
  $103 = ((($100)) + 4|0);
  HEAP32[$103>>2] = $4;
  $104 = ((($100)) + 8|0);
  $105 = ((($100)) + 392|0);
  HEAP32[$105>>2] = $104;
  $106 = ($100|0)==(0|0);
  $1756 = $106;
  label = 4;
 } else {
  $107 = ($100|0)==(0|0);
  if ($107) {
   $$ph = 1;$lastError$3$ph = 64;$subrStack$sroa$31$3$ph = 0;
   label = 779;
  } else {
   $1756 = 0;
   label = 4;
  }
 }
 L4: do {
  if ((label|0) == 4) {
   HEAP32[$error$i$i>>2] = 0;
   $108 = (_ft_mem_realloc($5,1,0,272,0,$error$i$i)|0);
   $109 = HEAP32[$error$i$i>>2]|0;
   $110 = ($109|0)==(0);
   if ($110) {
    $subrStack$sroa$19$0 = 17;
   } else {
    $111 = HEAP32[$4>>2]|0;
    $112 = ($111|0)==(0);
    if ($112) {
     HEAP32[$4>>2] = 64;
     $subrStack$sroa$19$0 = 0;
    } else {
     $subrStack$sroa$19$0 = 0;
    }
   }
   ;HEAP32[$108>>2]=HEAP32[$buf>>2]|0;HEAP32[$108+4>>2]=HEAP32[$buf+4>>2]|0;HEAP32[$108+8>>2]=HEAP32[$buf+8>>2]|0;HEAP32[$108+12>>2]=HEAP32[$buf+12>>2]|0;
   $113 = HEAP32[$4>>2]|0;
   $114 = ($113|0)==(0);
   if ($114) {
    $115 = ((($100)) + 392|0);
    $116 = ((($100)) + 4|0);
    $117 = ((($100)) + 8|0);
    $hintMask$idx14 = ((($hintMask)) + 4|0);
    $118 = $117;
    $119 = ((($glyphPath)) + 11736|0);
    $120 = ((($glyphPath)) + 11720|0);
    $121 = ((($glyphPath)) + 11740|0);
    $122 = ((($glyphPath)) + 11724|0);
    $$idx1$i116 = ((($glyphPath)) + 20|0);
    $123 = ((($glyphPath)) + 11704|0);
    $$0$i124 = ((($glyphPath)) + 11712|0);
    $$1$i126 = ((($glyphPath)) + 11716|0);
    $124 = ((($100)) + 12|0);
    $125 = ((($3)) + 708|0);
    $126 = ((($3)) + 700|0);
    $127 = ((($3)) + 716|0);
    $128 = ((($3)) + 704|0);
    $129 = ((($3)) + 696|0);
    $130 = ((($3)) + 712|0);
    $131 = ((($hintMask)) + 8|0);
    $132 = ((($hintMask)) + 12|0);
    $133 = ((($hintMask)) + 5|0);
    $134 = ((($counterHintMap)) + 13|0);
    $135 = ((($counterHintMap)) + 16|0);
    $136 = ((($counterHintMap)) + 4|0);
    $137 = ((($counterHintMap)) + 8|0);
    $138 = ((($counterMask)) + 8|0);
    $139 = ((($counterMask)) + 12|0);
    $140 = ((($counterMask)) + 4|0);
    $141 = ((($counterMask)) + 5|0);
    $charstring$0 = $108;$charstringIndex$0 = 0;$instructionLimit$0 = 20000000;
    L12: while(1) {
     $charstring$0$idx = ((($charstring$0)) + 8|0);
     $charstring$0$idx$val = HEAP32[$charstring$0$idx>>2]|0;
     $charstring$0$idx12 = ((($charstring$0)) + 12|0);
     $charstring$0$idx12$val = HEAP32[$charstring$0$idx12>>2]|0;
     $142 = ($charstring$0$idx12$val>>>0)<($charstring$0$idx$val>>>0);
     if ($142) {
      $144 = ((($charstring$0$idx12$val)) + 1|0);
      HEAP32[$charstring$0$idx12>>2] = $144;
      $145 = HEAP8[$charstring$0$idx12$val>>0]|0;
      $146 = $145&255;
      $535 = $144;$op1$0 = $146;
     } else {
      $143 = ($charstringIndex$0|0)==(0);
      $$ = $143 ? 14 : 11;
      $535 = $charstring$0$idx12$val;$op1$0 = $$;
     }
     $147 = HEAP32[$4>>2]|0;
     $148 = ($147|0)==(0);
     if (!($148)) {
      $1757 = $1756;$subrStack$sroa$31$3209 = $108;
      break L4;
     }
     $149 = (($instructionLimit$0) + -1)|0;
     $150 = ($149|0)==(0);
     if ($150) {
      $$ph = $1756;$lastError$3$ph = 18;$subrStack$sroa$31$3$ph = $108;
      label = 779;
      break L4;
     }
     L20: do {
      switch ($op1$0|0) {
      case 14:  {
       break L12;
       break;
      }
      case 17: case 16: case 15: case 13: case 9: case 2: case 0:  {
       break;
      }
      case 1: case 18:  {
       $hintMask$idx14$val = HEAP8[$hintMask$idx14>>0]|0;
       $151 = ($hintMask$idx14$val<<24>>24)==(0);
       if ($151) {
        _cf2_doStems($font,$100,$hStemHintArray,$width,$haveWidth);
        $152 = HEAP32[$2>>2]|0;
        $153 = ((($152)) + 561|0);
        $154 = HEAP8[$153>>0]|0;
        $155 = ($154<<24>>24)==(0);
        if (!($155)) {
         $$ph = $1756;$lastError$3$ph = 0;$subrStack$sroa$31$3$ph = $108;
         label = 779;
         break L4;
        }
       }
       break;
      }
      case 3: case 23:  {
       $hintMask$idx13$val = HEAP8[$hintMask$idx14>>0]|0;
       $156 = ($hintMask$idx13$val<<24>>24)==(0);
       if ($156) {
        _cf2_doStems($font,$100,$vStemHintArray,$width,$haveWidth);
        $157 = HEAP32[$2>>2]|0;
        $158 = ((($157)) + 561|0);
        $159 = HEAP8[$158>>0]|0;
        $160 = ($159<<24>>24)==(0);
        if (!($160)) {
         $$ph = $1756;$lastError$3$ph = 0;$subrStack$sroa$31$3$ph = $108;
         label = 779;
         break L4;
        }
       }
       break;
      }
      case 4:  {
       $161 = HEAP32[$115>>2]|0;
       $162 = (($161) - ($118))|0;
       $163 = $162 >> 3;
       $164 = ($163>>>0)<(2);
       $165 = HEAP8[$haveWidth>>0]|0;
       $166 = ($165<<24>>24)!=(0);
       $or$cond = $166 | $164;
       $167 = $161;
       if (!($or$cond)) {
        $168 = ($163|0)==(0);
        L94: do {
         if ($168) {
          $169 = HEAP32[$116>>2]|0;
          $170 = ($169|0)==(0|0);
          if ($170) {
           $$0$i96 = 0;
          } else {
           $171 = HEAP32[$169>>2]|0;
           $172 = ($171|0)==(0);
           if ($172) {
            HEAP32[$169>>2] = 130;
            $$0$i96 = 0;
           } else {
            $$0$i96 = 0;
           }
          }
         } else {
          $173 = HEAP32[$124>>2]|0;
          switch ($173|0) {
          case 2:  {
           $174 = HEAP32[$117>>2]|0;
           $175 = $174 << 16;
           $$0$i96 = $175;
           break L94;
           break;
          }
          case 1:  {
           $176 = HEAP32[$117>>2]|0;
           $177 = ($176|0)<(0);
           if ($177) {
            $178 = (8192 - ($176))|0;
            $179 = $178 >> 14;
            $180 = (0 - ($179))|0;
            $$0$i96 = $180;
            break L94;
           } else {
            $181 = (($176) + 8192)|0;
            $182 = $181 >> 14;
            $$0$i96 = $182;
            break L94;
           }
           break;
          }
          default: {
           $183 = HEAP32[$117>>2]|0;
           $$0$i96 = $183;
           break L94;
          }
          }
         }
        } while(0);
        $184 = (($$0$i96) + ($8))|0;
        HEAP32[$width>>2] = $184;
       }
       HEAP8[$haveWidth>>0] = 1;
       $185 = HEAP32[$2>>2]|0;
       $186 = ((($185)) + 561|0);
       $187 = HEAP8[$186>>0]|0;
       $188 = ($187<<24>>24)==(0);
       if (!($188)) {
        $$ph = $1756;$lastError$3$ph = 0;$subrStack$sroa$31$3$ph = $108;
        label = 779;
        break L4;
       }
       $189 = ($167|0)==($117|0);
       L109: do {
        if ($189) {
         $190 = HEAP32[$116>>2]|0;
         $191 = ($190|0)==(0|0);
         if ($191) {
          $$0$i109 = 0;
         } else {
          $192 = HEAP32[$190>>2]|0;
          $193 = ($192|0)==(0);
          if ($193) {
           HEAP32[$190>>2] = 161;
           $$0$i109 = 0;
          } else {
           $$0$i109 = 0;
          }
         }
        } else {
         $194 = ((($167)) + -8|0);
         HEAP32[$115>>2] = $194;
         $195 = ((($167)) + -4|0);
         $196 = HEAP32[$195>>2]|0;
         switch ($196|0) {
         case 2:  {
          $197 = HEAP32[$194>>2]|0;
          $198 = $197 << 16;
          $$0$i109 = $198;
          break L109;
          break;
         }
         case 1:  {
          $199 = HEAP32[$194>>2]|0;
          $200 = ($199|0)<(0);
          if ($200) {
           $201 = (8192 - ($199))|0;
           $202 = $201 >> 14;
           $203 = (0 - ($202))|0;
           $$0$i109 = $203;
           break L109;
          } else {
           $204 = (($199) + 8192)|0;
           $205 = $204 >> 14;
           $$0$i109 = $205;
           break L109;
          }
          break;
         }
         default: {
          $206 = HEAP32[$194>>2]|0;
          $$0$i109 = $206;
          break L109;
         }
         }
        }
       } while(0);
       $207 = HEAP32[$1>>2]|0;
       $208 = (($207) + ($$0$i109))|0;
       HEAP32[$1>>2] = $208;
       $209 = HEAP32[$0>>2]|0;
       $210 = HEAP8[$96>>0]|0;
       $211 = ($210<<24>>24)==(0);
       if (!($211)) {
        HEAP8[$97>>0] = 1;
        $212 = HEAP32[$119>>2]|0;
        $213 = HEAP32[$121>>2]|0;
        _cf2_glyphpath_lineTo($glyphPath,$212,$213);
        $214 = HEAP8[$98>>0]|0;
        $215 = ($214<<24>>24)==(0);
        if (!($215)) {
         $$0$val$i125 = HEAP32[$$0$i124>>2]|0;
         $$1$val$i127 = HEAP32[$$1$i126>>2]|0;
         _cf2_glyphpath_pushPrevElem($glyphPath,$47,$123,$$0$val$i125,$$1$val$i127,1);
        }
        HEAP8[$95>>0] = 1;
        HEAP8[$96>>0] = 0;
        HEAP8[$97>>0] = 0;
        HEAP8[$98>>0] = 0;
       }
       HEAP32[$119>>2] = $209;
       HEAP32[$120>>2] = $209;
       HEAP32[$121>>2] = $208;
       HEAP32[$122>>2] = $208;
       HEAP8[$95>>0] = 1;
       $$idx1$val$i117 = HEAP8[$$idx1$i116>>0]|0;
       $216 = ($$idx1$val$i117<<24>>24)==(0);
       $$pre$i119 = HEAP32[$73>>2]|0;
       if ($216) {
        label = 49;
       } else {
        $$idx$i120 = ((($$pre$i119)) + 5|0);
        $$idx$val$i121 = HEAP8[$$idx$i120>>0]|0;
        $217 = ($$idx$val$i121<<24>>24)==(0);
        if (!($217)) {
         label = 49;
        }
       }
       if ((label|0) == 49) {
        label = 0;
        $218 = HEAP32[$71>>2]|0;
        $219 = HEAP32[$72>>2]|0;
        $220 = HEAP32[$74>>2]|0;
        _cf2_hintmap_build($47,$218,$219,$$pre$i119,$220,0);
       }
       _memcpy(($42|0),($47|0),3868)|0;
       break;
      }
      case 5:  {
       $221 = HEAP32[$115>>2]|0;
       $222 = (($221) - ($118))|0;
       $223 = $222 >> 3;
       $224 = ($223|0)==(0);
       L134: do {
        if (!($224)) {
         $226 = $221;$index$0238 = 0;
         while(1) {
          $225 = (($226) - ($118))|0;
          $227 = $225 >> 3;
          $228 = ($227>>>0)>($index$0238>>>0);
          L137: do {
           if ($228) {
            $233 = (((($100)) + 8|0) + ($index$0238<<3)|0);
            $234 = (((((($100)) + 8|0) + ($index$0238<<3)|0)) + 4|0);
            $235 = HEAP32[$234>>2]|0;
            switch ($235|0) {
            case 2:  {
             $236 = HEAP32[$233>>2]|0;
             $237 = $236 << 16;
             $$0$i129 = $237;
             break L137;
             break;
            }
            case 1:  {
             $238 = HEAP32[$233>>2]|0;
             $239 = ($238|0)<(0);
             if ($239) {
              $240 = (8192 - ($238))|0;
              $241 = $240 >> 14;
              $242 = (0 - ($241))|0;
              $$0$i129 = $242;
              break L137;
             } else {
              $243 = (($238) + 8192)|0;
              $244 = $243 >> 14;
              $$0$i129 = $244;
              break L137;
             }
             break;
            }
            default: {
             $245 = HEAP32[$233>>2]|0;
             $$0$i129 = $245;
             break L137;
            }
            }
           } else {
            $229 = HEAP32[$116>>2]|0;
            $230 = ($229|0)==(0|0);
            if ($230) {
             $$0$i129 = 0;
            } else {
             $231 = HEAP32[$229>>2]|0;
             $232 = ($231|0)==(0);
             if ($232) {
              HEAP32[$229>>2] = 130;
              $$0$i129 = 0;
             } else {
              $$0$i129 = 0;
             }
            }
           }
          } while(0);
          $246 = HEAP32[$0>>2]|0;
          $247 = (($246) + ($$0$i129))|0;
          HEAP32[$0>>2] = $247;
          $248 = $index$0238 | 1;
          $249 = ($227>>>0)>($248>>>0);
          L150: do {
           if ($249) {
            $254 = (((($100)) + 8|0) + ($248<<3)|0);
            $255 = (((((($100)) + 8|0) + ($248<<3)|0)) + 4|0);
            $256 = HEAP32[$255>>2]|0;
            switch ($256|0) {
            case 2:  {
             $257 = HEAP32[$254>>2]|0;
             $258 = $257 << 16;
             $$0$i131 = $258;$269 = $247;
             break L150;
             break;
            }
            case 1:  {
             $259 = HEAP32[$254>>2]|0;
             $260 = ($259|0)<(0);
             if ($260) {
              $261 = (8192 - ($259))|0;
              $262 = $261 >> 14;
              $263 = (0 - ($262))|0;
              $$0$i131 = $263;$269 = $247;
              break L150;
             } else {
              $264 = (($259) + 8192)|0;
              $265 = $264 >> 14;
              $$0$i131 = $265;$269 = $247;
              break L150;
             }
             break;
            }
            default: {
             $266 = HEAP32[$254>>2]|0;
             $$0$i131 = $266;$269 = $247;
             break L150;
            }
            }
           } else {
            $250 = HEAP32[$116>>2]|0;
            $251 = ($250|0)==(0|0);
            if ($251) {
             $$0$i131 = 0;$269 = $247;
            } else {
             $252 = HEAP32[$250>>2]|0;
             $253 = ($252|0)==(0);
             if ($253) {
              HEAP32[$250>>2] = 130;
              $$pre260 = HEAP32[$0>>2]|0;
              $$0$i131 = 0;$269 = $$pre260;
             } else {
              $$0$i131 = 0;$269 = $247;
             }
            }
           }
          } while(0);
          $267 = HEAP32[$1>>2]|0;
          $268 = (($267) + ($$0$i131))|0;
          HEAP32[$1>>2] = $268;
          _cf2_glyphpath_lineTo($glyphPath,$269,$268);
          $270 = (($index$0238) + 2)|0;
          $271 = ($270>>>0)<($223>>>0);
          if (!($271)) {
           break L134;
          }
          $$pre259 = HEAP32[$115>>2]|0;
          $226 = $$pre259;$index$0238 = $270;
         }
        }
       } while(0);
       HEAP32[$115>>2] = $117;
       $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
       continue L12;
       break;
      }
      case 7: case 6:  {
       $272 = HEAP32[$115>>2]|0;
       $273 = (($272) - ($118))|0;
       $274 = $273 >> 3;
       $275 = ($274|0)==(0);
       L166: do {
        if (!($275)) {
         $276 = ($op1$0|0)==(6);
         $278 = $272;$index1$0234 = 0;$isX$0$in233 = $276;
         while(1) {
          $277 = (($278) - ($118))|0;
          $279 = $277 >> 3;
          $280 = ($279>>>0)>($index1$0234>>>0);
          L170: do {
           if ($280) {
            $285 = (((($100)) + 8|0) + ($index1$0234<<3)|0);
            $286 = (((((($100)) + 8|0) + ($index1$0234<<3)|0)) + 4|0);
            $287 = HEAP32[$286>>2]|0;
            switch ($287|0) {
            case 2:  {
             $288 = HEAP32[$285>>2]|0;
             $289 = $288 << 16;
             $$0$i146 = $289;
             break L170;
             break;
            }
            case 1:  {
             $290 = HEAP32[$285>>2]|0;
             $291 = ($290|0)<(0);
             if ($291) {
              $292 = (8192 - ($290))|0;
              $293 = $292 >> 14;
              $294 = (0 - ($293))|0;
              $$0$i146 = $294;
              break L170;
             } else {
              $295 = (($290) + 8192)|0;
              $296 = $295 >> 14;
              $$0$i146 = $296;
              break L170;
             }
             break;
            }
            default: {
             $297 = HEAP32[$285>>2]|0;
             $$0$i146 = $297;
             break L170;
            }
            }
           } else {
            $281 = HEAP32[$116>>2]|0;
            $282 = ($281|0)==(0|0);
            if ($282) {
             $$0$i146 = 0;
            } else {
             $283 = HEAP32[$281>>2]|0;
             $284 = ($283|0)==(0);
             if ($284) {
              HEAP32[$281>>2] = 130;
              $$0$i146 = 0;
             } else {
              $$0$i146 = 0;
             }
            }
           }
          } while(0);
          if ($isX$0$in233) {
           $298 = HEAP32[$0>>2]|0;
           $299 = (($298) + ($$0$i146))|0;
           HEAP32[$0>>2] = $299;
           $$pre258 = HEAP32[$1>>2]|0;
           $302 = $299;$303 = $$pre258;
          } else {
           $300 = HEAP32[$1>>2]|0;
           $301 = (($300) + ($$0$i146))|0;
           HEAP32[$1>>2] = $301;
           $$pre257 = HEAP32[$0>>2]|0;
           $302 = $$pre257;$303 = $301;
          }
          _cf2_glyphpath_lineTo($glyphPath,$302,$303);
          $304 = (($index1$0234) + 1)|0;
          $exitcond = ($304|0)==($274|0);
          if ($exitcond) {
           break L166;
          }
          $305 = $isX$0$in233 ^ 1;
          $$pre256 = HEAP32[$115>>2]|0;
          $278 = $$pre256;$index1$0234 = $304;$isX$0$in233 = $305;
         }
        }
       } while(0);
       HEAP32[$115>>2] = $117;
       $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
       continue L12;
       break;
      }
      case 8: case 24:  {
       $306 = HEAP32[$115>>2]|0;
       $307 = (($306) - ($118))|0;
       $308 = $307 >> 3;
       $309 = ($308>>>0)<(6);
       L190: do {
        if ($309) {
         $index4$0$lcssa = 0;
        } else {
         $311 = $306;$435 = 6;$index4$0228 = 0;
         while(1) {
          $310 = (($311) - ($118))|0;
          $312 = $310 >> 3;
          $313 = ($312>>>0)>($index4$0228>>>0);
          L193: do {
           if ($313) {
            $318 = (((($100)) + 8|0) + ($index4$0228<<3)|0);
            $319 = (((((($100)) + 8|0) + ($index4$0228<<3)|0)) + 4|0);
            $320 = HEAP32[$319>>2]|0;
            switch ($320|0) {
            case 2:  {
             $321 = HEAP32[$318>>2]|0;
             $322 = $321 << 16;
             $$0$i148 = $322;
             break L193;
             break;
            }
            case 1:  {
             $323 = HEAP32[$318>>2]|0;
             $324 = ($323|0)<(0);
             if ($324) {
              $325 = (8192 - ($323))|0;
              $326 = $325 >> 14;
              $327 = (0 - ($326))|0;
              $$0$i148 = $327;
              break L193;
             } else {
              $328 = (($323) + 8192)|0;
              $329 = $328 >> 14;
              $$0$i148 = $329;
              break L193;
             }
             break;
            }
            default: {
             $330 = HEAP32[$318>>2]|0;
             $$0$i148 = $330;
             break L193;
            }
            }
           } else {
            $314 = HEAP32[$116>>2]|0;
            $315 = ($314|0)==(0|0);
            if ($315) {
             $$0$i148 = 0;
            } else {
             $316 = HEAP32[$314>>2]|0;
             $317 = ($316|0)==(0);
             if ($317) {
              HEAP32[$314>>2] = 130;
              $$0$i148 = 0;
             } else {
              $$0$i148 = 0;
             }
            }
           }
          } while(0);
          $331 = HEAP32[$0>>2]|0;
          $332 = (($331) + ($$0$i148))|0;
          $333 = $index4$0228 | 1;
          $334 = ($312>>>0)>($333>>>0);
          L206: do {
           if ($334) {
            $339 = (((($100)) + 8|0) + ($333<<3)|0);
            $340 = (((((($100)) + 8|0) + ($333<<3)|0)) + 4|0);
            $341 = HEAP32[$340>>2]|0;
            switch ($341|0) {
            case 2:  {
             $342 = HEAP32[$339>>2]|0;
             $343 = $342 << 16;
             $$0$i152 = $343;
             break L206;
             break;
            }
            case 1:  {
             $344 = HEAP32[$339>>2]|0;
             $345 = ($344|0)<(0);
             if ($345) {
              $346 = (8192 - ($344))|0;
              $347 = $346 >> 14;
              $348 = (0 - ($347))|0;
              $$0$i152 = $348;
              break L206;
             } else {
              $349 = (($344) + 8192)|0;
              $350 = $349 >> 14;
              $$0$i152 = $350;
              break L206;
             }
             break;
            }
            default: {
             $351 = HEAP32[$339>>2]|0;
             $$0$i152 = $351;
             break L206;
            }
            }
           } else {
            $335 = HEAP32[$116>>2]|0;
            $336 = ($335|0)==(0|0);
            if ($336) {
             $$0$i152 = 0;
            } else {
             $337 = HEAP32[$335>>2]|0;
             $338 = ($337|0)==(0);
             if ($338) {
              HEAP32[$335>>2] = 130;
              $$0$i152 = 0;
             } else {
              $$0$i152 = 0;
             }
            }
           }
          } while(0);
          $352 = HEAP32[$1>>2]|0;
          $353 = (($352) + ($$0$i152))|0;
          $354 = (($index4$0228) + 2)|0;
          $355 = ($312>>>0)>($354>>>0);
          L219: do {
           if ($355) {
            $360 = (((($100)) + 8|0) + ($354<<3)|0);
            $361 = (((((($100)) + 8|0) + ($354<<3)|0)) + 4|0);
            $362 = HEAP32[$361>>2]|0;
            switch ($362|0) {
            case 2:  {
             $363 = HEAP32[$360>>2]|0;
             $364 = $363 << 16;
             $$0$i154 = $364;
             break L219;
             break;
            }
            case 1:  {
             $365 = HEAP32[$360>>2]|0;
             $366 = ($365|0)<(0);
             if ($366) {
              $367 = (8192 - ($365))|0;
              $368 = $367 >> 14;
              $369 = (0 - ($368))|0;
              $$0$i154 = $369;
              break L219;
             } else {
              $370 = (($365) + 8192)|0;
              $371 = $370 >> 14;
              $$0$i154 = $371;
              break L219;
             }
             break;
            }
            default: {
             $372 = HEAP32[$360>>2]|0;
             $$0$i154 = $372;
             break L219;
            }
            }
           } else {
            $356 = HEAP32[$116>>2]|0;
            $357 = ($356|0)==(0|0);
            if ($357) {
             $$0$i154 = 0;
            } else {
             $358 = HEAP32[$356>>2]|0;
             $359 = ($358|0)==(0);
             if ($359) {
              HEAP32[$356>>2] = 130;
              $$0$i154 = 0;
             } else {
              $$0$i154 = 0;
             }
            }
           }
          } while(0);
          $373 = (($$0$i154) + ($332))|0;
          $374 = (($index4$0228) + 3)|0;
          $375 = ($312>>>0)>($374>>>0);
          L232: do {
           if ($375) {
            $380 = (((($100)) + 8|0) + ($374<<3)|0);
            $381 = (((((($100)) + 8|0) + ($374<<3)|0)) + 4|0);
            $382 = HEAP32[$381>>2]|0;
            switch ($382|0) {
            case 2:  {
             $383 = HEAP32[$380>>2]|0;
             $384 = $383 << 16;
             $$0$i158 = $384;
             break L232;
             break;
            }
            case 1:  {
             $385 = HEAP32[$380>>2]|0;
             $386 = ($385|0)<(0);
             if ($386) {
              $387 = (8192 - ($385))|0;
              $388 = $387 >> 14;
              $389 = (0 - ($388))|0;
              $$0$i158 = $389;
              break L232;
             } else {
              $390 = (($385) + 8192)|0;
              $391 = $390 >> 14;
              $$0$i158 = $391;
              break L232;
             }
             break;
            }
            default: {
             $392 = HEAP32[$380>>2]|0;
             $$0$i158 = $392;
             break L232;
            }
            }
           } else {
            $376 = HEAP32[$116>>2]|0;
            $377 = ($376|0)==(0|0);
            if ($377) {
             $$0$i158 = 0;
            } else {
             $378 = HEAP32[$376>>2]|0;
             $379 = ($378|0)==(0);
             if ($379) {
              HEAP32[$376>>2] = 130;
              $$0$i158 = 0;
             } else {
              $$0$i158 = 0;
             }
            }
           }
          } while(0);
          $393 = (($$0$i158) + ($353))|0;
          $394 = (($index4$0228) + 4)|0;
          $395 = ($312>>>0)>($394>>>0);
          L245: do {
           if ($395) {
            $400 = (((($100)) + 8|0) + ($394<<3)|0);
            $401 = (((((($100)) + 8|0) + ($394<<3)|0)) + 4|0);
            $402 = HEAP32[$401>>2]|0;
            switch ($402|0) {
            case 2:  {
             $403 = HEAP32[$400>>2]|0;
             $404 = $403 << 16;
             $$0$i160 = $404;
             break L245;
             break;
            }
            case 1:  {
             $405 = HEAP32[$400>>2]|0;
             $406 = ($405|0)<(0);
             if ($406) {
              $407 = (8192 - ($405))|0;
              $408 = $407 >> 14;
              $409 = (0 - ($408))|0;
              $$0$i160 = $409;
              break L245;
             } else {
              $410 = (($405) + 8192)|0;
              $411 = $410 >> 14;
              $$0$i160 = $411;
              break L245;
             }
             break;
            }
            default: {
             $412 = HEAP32[$400>>2]|0;
             $$0$i160 = $412;
             break L245;
            }
            }
           } else {
            $396 = HEAP32[$116>>2]|0;
            $397 = ($396|0)==(0|0);
            if ($397) {
             $$0$i160 = 0;
            } else {
             $398 = HEAP32[$396>>2]|0;
             $399 = ($398|0)==(0);
             if ($399) {
              HEAP32[$396>>2] = 130;
              $$0$i160 = 0;
             } else {
              $$0$i160 = 0;
             }
            }
           }
          } while(0);
          $413 = (($$0$i160) + ($373))|0;
          $414 = (($index4$0228) + 5)|0;
          $415 = ($312>>>0)>($414>>>0);
          L258: do {
           if ($415) {
            $420 = (((($100)) + 8|0) + ($414<<3)|0);
            $421 = (((((($100)) + 8|0) + ($414<<3)|0)) + 4|0);
            $422 = HEAP32[$421>>2]|0;
            switch ($422|0) {
            case 2:  {
             $423 = HEAP32[$420>>2]|0;
             $424 = $423 << 16;
             $$0$i162 = $424;
             break L258;
             break;
            }
            case 1:  {
             $425 = HEAP32[$420>>2]|0;
             $426 = ($425|0)<(0);
             if ($426) {
              $427 = (8192 - ($425))|0;
              $428 = $427 >> 14;
              $429 = (0 - ($428))|0;
              $$0$i162 = $429;
              break L258;
             } else {
              $430 = (($425) + 8192)|0;
              $431 = $430 >> 14;
              $$0$i162 = $431;
              break L258;
             }
             break;
            }
            default: {
             $432 = HEAP32[$420>>2]|0;
             $$0$i162 = $432;
             break L258;
            }
            }
           } else {
            $416 = HEAP32[$116>>2]|0;
            $417 = ($416|0)==(0|0);
            if ($417) {
             $$0$i162 = 0;
            } else {
             $418 = HEAP32[$416>>2]|0;
             $419 = ($418|0)==(0);
             if (!($419)) {
              $$0$i162 = 0;
              break;
             }
             HEAP32[$416>>2] = 130;
             $$0$i162 = 0;
            }
           }
          } while(0);
          $433 = (($$0$i162) + ($393))|0;
          _cf2_glyphpath_curveTo($glyphPath,$332,$353,$373,$393,$413,$433);
          HEAP32[$0>>2] = $413;
          HEAP32[$1>>2] = $433;
          $434 = (($435) + 6)|0;
          $436 = ($434>>>0)>($308>>>0);
          if ($436) {
           $index4$0$lcssa = $435;
           break L190;
          }
          $$pre254 = HEAP32[$115>>2]|0;
          $index4$0228$phi = $435;$311 = $$pre254;$435 = $434;$index4$0228 = $index4$0228$phi;
         }
        }
       } while(0);
       $437 = ($op1$0|0)==(24);
       if ($437) {
        $438 = HEAP32[$115>>2]|0;
        $439 = (($438) - ($118))|0;
        $440 = $439 >> 3;
        $441 = ($440>>>0)>($index4$0$lcssa>>>0);
        L275: do {
         if ($441) {
          $446 = (((($100)) + 8|0) + ($index4$0$lcssa<<3)|0);
          $447 = (((((($100)) + 8|0) + ($index4$0$lcssa<<3)|0)) + 4|0);
          $448 = HEAP32[$447>>2]|0;
          switch ($448|0) {
          case 2:  {
           $449 = HEAP32[$446>>2]|0;
           $450 = $449 << 16;
           $$0$i178 = $450;
           break L275;
           break;
          }
          case 1:  {
           $451 = HEAP32[$446>>2]|0;
           $452 = ($451|0)<(0);
           if ($452) {
            $453 = (8192 - ($451))|0;
            $454 = $453 >> 14;
            $455 = (0 - ($454))|0;
            $$0$i178 = $455;
            break L275;
           } else {
            $456 = (($451) + 8192)|0;
            $457 = $456 >> 14;
            $$0$i178 = $457;
            break L275;
           }
           break;
          }
          default: {
           $458 = HEAP32[$446>>2]|0;
           $$0$i178 = $458;
           break L275;
          }
          }
         } else {
          $442 = HEAP32[$116>>2]|0;
          $443 = ($442|0)==(0|0);
          if ($443) {
           $$0$i178 = 0;
          } else {
           $444 = HEAP32[$442>>2]|0;
           $445 = ($444|0)==(0);
           if ($445) {
            HEAP32[$442>>2] = 130;
            $$0$i178 = 0;
           } else {
            $$0$i178 = 0;
           }
          }
         }
        } while(0);
        $459 = HEAP32[$0>>2]|0;
        $460 = (($459) + ($$0$i178))|0;
        HEAP32[$0>>2] = $460;
        $461 = $index4$0$lcssa | 1;
        $462 = ($440>>>0)>($461>>>0);
        L288: do {
         if ($462) {
          $467 = (((($100)) + 8|0) + ($461<<3)|0);
          $468 = (((((($100)) + 8|0) + ($461<<3)|0)) + 4|0);
          $469 = HEAP32[$468>>2]|0;
          switch ($469|0) {
          case 2:  {
           $470 = HEAP32[$467>>2]|0;
           $471 = $470 << 16;
           $$0$i180 = $471;$482 = $460;
           break L288;
           break;
          }
          case 1:  {
           $472 = HEAP32[$467>>2]|0;
           $473 = ($472|0)<(0);
           if ($473) {
            $474 = (8192 - ($472))|0;
            $475 = $474 >> 14;
            $476 = (0 - ($475))|0;
            $$0$i180 = $476;$482 = $460;
            break L288;
           } else {
            $477 = (($472) + 8192)|0;
            $478 = $477 >> 14;
            $$0$i180 = $478;$482 = $460;
            break L288;
           }
           break;
          }
          default: {
           $479 = HEAP32[$467>>2]|0;
           $$0$i180 = $479;$482 = $460;
           break L288;
          }
          }
         } else {
          $463 = HEAP32[$116>>2]|0;
          $464 = ($463|0)==(0|0);
          if ($464) {
           $$0$i180 = 0;$482 = $460;
          } else {
           $465 = HEAP32[$463>>2]|0;
           $466 = ($465|0)==(0);
           if ($466) {
            HEAP32[$463>>2] = 130;
            $$pre255 = HEAP32[$0>>2]|0;
            $$0$i180 = 0;$482 = $$pre255;
           } else {
            $$0$i180 = 0;$482 = $460;
           }
          }
         }
        } while(0);
        $480 = HEAP32[$1>>2]|0;
        $481 = (($480) + ($$0$i180))|0;
        HEAP32[$1>>2] = $481;
        _cf2_glyphpath_lineTo($glyphPath,$482,$481);
       }
       HEAP32[$115>>2] = $117;
       $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
       continue L12;
       break;
      }
      case 10: case 29:  {
       $483 = ($charstringIndex$0|0)>(16);
       if ($483) {
        $$ph = $1756;$lastError$3$ph = 18;$subrStack$sroa$31$3$ph = $108;
        label = 779;
        break L4;
       }
       $484 = (($charstringIndex$0) + 1)|0;
       $485 = ($subrStack$sroa$19$0>>>0)>($484>>>0);
       if ($485) {
        $$0$i194 = $484;
       } else {
        HEAP32[$4>>2] = 130;
        $$0$i194 = 0;
       }
       $486 = $$0$i194 << 4;
       $487 = (($108) + ($486)|0);
       $488 = HEAP32[$115>>2]|0;
       $489 = ($488|0)==($117|0);
       do {
        if ($489) {
         $490 = HEAP32[$116>>2]|0;
         $491 = ($490|0)==(0|0);
         if ($491) {
          $$0$i197 = 0;
         } else {
          $492 = HEAP32[$490>>2]|0;
          $493 = ($492|0)==(0);
          if ($493) {
           HEAP32[$490>>2] = 161;
           $$0$i197 = 0;
          } else {
           $$0$i197 = 0;
          }
         }
        } else {
         $494 = ((($488)) + -8|0);
         $495 = ((($488)) + -4|0);
         $496 = HEAP32[$495>>2]|0;
         $497 = ($496|0)==(2);
         if ($497) {
          HEAP32[$115>>2] = $494;
          $502 = HEAP32[$494>>2]|0;
          $$0$i197 = $502;
          break;
         }
         $498 = HEAP32[$116>>2]|0;
         $499 = ($498|0)==(0|0);
         if ($499) {
          $$0$i197 = 0;
         } else {
          $500 = HEAP32[$498>>2]|0;
          $501 = ($500|0)==(0);
          if ($501) {
           HEAP32[$498>>2] = 160;
           $$0$i197 = 0;
          } else {
           $$0$i197 = 0;
          }
         }
        }
       } while(0);
       $cond = ($op1$0|0)==(29);
       ;HEAP32[$487>>2]=0|0;HEAP32[$487+4>>2]=0|0;HEAP32[$487+8>>2]=0|0;HEAP32[$487+12>>2]=0|0;
       if ($cond) {
        $503 = HEAP32[$125>>2]|0;
        $504 = (($503) + ($$0$i197))|0;
        $505 = HEAP32[$126>>2]|0;
        $506 = ($504>>>0)<($505>>>0);
        if (!($506)) {
         $$ph = $1756;$lastError$3$ph = 18;$subrStack$sroa$31$3$ph = $108;
         label = 779;
         break L4;
        }
        $507 = HEAP32[$127>>2]|0;
        $508 = (($507) + ($504<<2)|0);
        $509 = HEAP32[$508>>2]|0;
        $510 = ((($487)) + 12|0);
        HEAP32[$510>>2] = $509;
        $511 = ((($487)) + 4|0);
        HEAP32[$511>>2] = $509;
        $512 = (($504) + 1)|0;
        $513 = (($507) + ($512<<2)|0);
        $514 = HEAP32[$513>>2]|0;
        $515 = ((($487)) + 8|0);
        HEAP32[$515>>2] = $514;
        $charstring$0 = $487;$charstringIndex$0 = $484;$instructionLimit$0 = $149;
        continue L12;
       } else {
        $516 = HEAP32[$128>>2]|0;
        $517 = (($516) + ($$0$i197))|0;
        $518 = HEAP32[$129>>2]|0;
        $519 = ($517>>>0)<($518>>>0);
        if (!($519)) {
         $$ph = $1756;$lastError$3$ph = 18;$subrStack$sroa$31$3$ph = $108;
         label = 779;
         break L4;
        }
        $520 = HEAP32[$130>>2]|0;
        $521 = (($520) + ($517<<2)|0);
        $522 = HEAP32[$521>>2]|0;
        $523 = ((($487)) + 12|0);
        HEAP32[$523>>2] = $522;
        $524 = ((($487)) + 4|0);
        HEAP32[$524>>2] = $522;
        $525 = (($517) + 1)|0;
        $526 = (($520) + ($525<<2)|0);
        $527 = HEAP32[$526>>2]|0;
        $528 = ((($487)) + 8|0);
        HEAP32[$528>>2] = $527;
        $charstring$0 = $487;$charstringIndex$0 = $484;$instructionLimit$0 = $149;
        continue L12;
       }
       break;
      }
      case 11:  {
       $529 = ($charstringIndex$0|0)<(1);
       if ($529) {
        $$ph = $1756;$lastError$3$ph = 18;$subrStack$sroa$31$3$ph = $108;
        label = 779;
        break L4;
       }
       $530 = (($charstringIndex$0) + -1)|0;
       $531 = ($subrStack$sroa$19$0>>>0)>($530>>>0);
       if ($531) {
        $$0$i201 = $530;
       } else {
        HEAP32[$4>>2] = 130;
        $$0$i201 = 0;
       }
       $532 = $$0$i201 << 4;
       $533 = (($108) + ($532)|0);
       $charstring$0 = $533;$charstringIndex$0 = $530;$instructionLimit$0 = $149;
       continue L12;
       break;
      }
      case 12:  {
       $534 = ($535>>>0)<($charstring$0$idx$val>>>0);
       if (!($534)) {
        $536 = HEAP32[$charstring$0>>2]|0;
        $537 = ($536|0)==(0|0);
        if ($537) {
         break L20;
        }
        $538 = HEAP32[$536>>2]|0;
        $539 = ($538|0)==(0);
        if (!($539)) {
         break L20;
        }
        HEAP32[$536>>2] = 85;
        break L20;
       }
       $540 = ((($535)) + 1|0);
       HEAP32[$charstring$0$idx12>>2] = $540;
       $541 = HEAP8[$535>>0]|0;
       $542 = $541&255;
       switch ($542|0) {
       case 37:  {
        _cf2_doFlex($100,$0,$1,$glyphPath,532110,1);
        $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
        continue L12;
        break;
       }
       case 36:  {
        _cf2_doFlex($100,$0,$1,$glyphPath,532098,0);
        $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
        continue L12;
        break;
       }
       case 35:  {
        _cf2_doFlex($100,$0,$1,$glyphPath,532086,0);
        break L20;
        break;
       }
       case 34:  {
        _cf2_doFlex($100,$0,$1,$glyphPath,532074,0);
        $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
        continue L12;
        break;
       }
       default: {
        break L20;
       }
       }
       break;
      }
      case 19: case 20:  {
       $778 = HEAP32[$115>>2]|0;
       $779 = (($778) - ($118))|0;
       $780 = $779 >> 3;
       $781 = ($780>>>0)<(2);
       $hintMask$idx$val = HEAP8[$hintMask$idx14>>0]|0;
       $782 = ($hintMask$idx$val<<24>>24)==(0);
       $or$cond244 = $781 | $782;
       if ($or$cond244) {
        _cf2_doStems($font,$100,$vStemHintArray,$width,$haveWidth);
        $783 = HEAP32[$2>>2]|0;
        $784 = ((($783)) + 561|0);
        $785 = HEAP8[$784>>0]|0;
        $786 = ($785<<24>>24)==(0);
        if (!($786)) {
         $$ph = $1756;$lastError$3$ph = 0;$subrStack$sroa$31$3$ph = $108;
         label = 779;
         break L4;
        }
        $787 = ($op1$0|0)==(19);
        if ($787) {
         $hStemHintArray$idx16$val = HEAP32[$13>>2]|0;
         $vStemHintArray$idx15$val = HEAP32[$20>>2]|0;
         $788 = (($vStemHintArray$idx15$val) + ($hStemHintArray$idx16$val))|0;
         $789 = ($788>>>0)>(96);
         if ($789) {
          $790 = HEAP32[$hintMask>>2]|0;
          $791 = ($790|0)==(0|0);
          if ($791) {
           break L20;
          }
          $792 = HEAP32[$790>>2]|0;
          $793 = ($792|0)==(0);
          if (!($793)) {
           break L20;
          }
          HEAP32[$790>>2] = 18;
          break L20;
         }
         HEAP32[$131>>2] = $788;
         $794 = (($788) + 7)|0;
         $795 = $794 >>> 3;
         HEAP32[$132>>2] = $795;
         HEAP8[$hintMask$idx14>>0] = 1;
         HEAP8[$133>>0] = 1;
         $796 = ($788|0)==(0);
         $797 = ($795|0)==(0);
         $or$cond$i139 = $796 | $797;
         if ($or$cond$i139) {
          break L20;
         }
         $i$02$i142 = 0;
         while(1) {
          $798 = HEAP32[$charstring$0$idx12>>2]|0;
          $799 = HEAP32[$charstring$0$idx>>2]|0;
          $800 = ($798>>>0)<($799>>>0);
          if ($800) {
           $801 = ((($798)) + 1|0);
           HEAP32[$charstring$0$idx12>>2] = $801;
           $802 = HEAP8[$798>>0]|0;
           $803 = $802&255;
           $$0$i1$i143 = $803;
          } else {
           $804 = HEAP32[$charstring$0>>2]|0;
           $805 = ($804|0)==(0|0);
           if ($805) {
            $$0$i1$i143 = 0;
           } else {
            $806 = HEAP32[$804>>2]|0;
            $807 = ($806|0)==(0);
            if ($807) {
             HEAP32[$804>>2] = 85;
             $$0$i1$i143 = 0;
            } else {
             $$0$i1$i143 = 0;
            }
           }
          }
          $808 = $$0$i1$i143&255;
          $809 = (((($hintMask)) + 16|0) + ($i$02$i142)|0);
          HEAP8[$809>>0] = $808;
          $810 = (($i$02$i142) + 1)|0;
          $811 = HEAP32[$132>>2]|0;
          $812 = ($810>>>0)<($811>>>0);
          if ($812) {
           $i$02$i142 = $810;
          } else {
           break;
          }
         }
        } else {
         _memset(($counterHintMap|0),0,3868)|0;
         $813 = HEAP8[$35>>0]|0;
         HEAP8[$134>>0] = $813;
         HEAP32[$135>>2] = $7;
         HEAP32[$counterHintMap>>2] = $font;
         HEAP32[$136>>2] = $34;
         HEAP32[$137>>2] = $25;
         ;HEAP32[$counterMask>>2]=0|0;HEAP32[$counterMask+4>>2]=0|0;HEAP32[$counterMask+8>>2]=0|0;HEAP32[$counterMask+12>>2]=0|0;HEAP32[$counterMask+16>>2]=0|0;HEAP32[$counterMask+20>>2]=0|0;HEAP32[$counterMask+24>>2]=0|0;
         HEAP32[$counterMask>>2] = $4;
         $hStemHintArray$idx$val = HEAP32[$13>>2]|0;
         $vStemHintArray$idx$val = HEAP32[$20>>2]|0;
         $814 = (($vStemHintArray$idx$val) + ($hStemHintArray$idx$val))|0;
         $815 = ($814>>>0)>(96);
         if ($815) {
          $816 = HEAP32[$4>>2]|0;
          $817 = ($816|0)==(0);
          if ($817) {
           HEAP32[$4>>2] = 18;
          }
         } else {
          HEAP32[$138>>2] = $814;
          $818 = (($814) + 7)|0;
          $819 = $818 >>> 3;
          HEAP32[$139>>2] = $819;
          HEAP8[$140>>0] = 1;
          HEAP8[$141>>0] = 1;
          $820 = ($814|0)==(0);
          $821 = ($819|0)==(0);
          $or$cond$i = $820 | $821;
          if (!($or$cond$i)) {
           $i$02$i = 0;
           while(1) {
            $822 = HEAP32[$charstring$0$idx12>>2]|0;
            $823 = HEAP32[$charstring$0$idx>>2]|0;
            $824 = ($822>>>0)<($823>>>0);
            do {
             if ($824) {
              $825 = ((($822)) + 1|0);
              HEAP32[$charstring$0$idx12>>2] = $825;
              $826 = HEAP8[$822>>0]|0;
              $827 = $826&255;
              $$0$i1$i = $827;
             } else {
              $828 = HEAP32[$charstring$0>>2]|0;
              $829 = ($828|0)==(0|0);
              if ($829) {
               $$0$i1$i = 0;
              } else {
               $830 = HEAP32[$828>>2]|0;
               $831 = ($830|0)==(0);
               if (!($831)) {
                $$0$i1$i = 0;
                break;
               }
               HEAP32[$828>>2] = 85;
               $$0$i1$i = 0;
              }
             }
            } while(0);
            $832 = $$0$i1$i&255;
            $833 = (((($counterMask)) + 16|0) + ($i$02$i)|0);
            HEAP8[$833>>0] = $832;
            $834 = (($i$02$i) + 1)|0;
            $835 = HEAP32[$139>>2]|0;
            $836 = ($834>>>0)<($835>>>0);
            if ($836) {
             $i$02$i = $834;
            } else {
             break;
            }
           }
          }
         }
         _cf2_hintmap_build($counterHintMap,$hStemHintArray,$vStemHintArray,$counterMask,0,0);
         break L20;
        }
       }
       break;
      }
      case 21:  {
       $837 = HEAP32[$115>>2]|0;
       $838 = (($837) - ($118))|0;
       $839 = $838 >> 3;
       $840 = ($839>>>0)<(3);
       $841 = HEAP8[$haveWidth>>0]|0;
       $842 = ($841<<24>>24)!=(0);
       $or$cond3 = $842 | $840;
       $843 = $837;
       if (!($or$cond3)) {
        $844 = ($839|0)==(0);
        L376: do {
         if ($844) {
          $845 = HEAP32[$116>>2]|0;
          $846 = ($845|0)==(0|0);
          if ($846) {
           $$0$i137 = 0;
          } else {
           $847 = HEAP32[$845>>2]|0;
           $848 = ($847|0)==(0);
           if ($848) {
            HEAP32[$845>>2] = 130;
            $$0$i137 = 0;
           } else {
            $$0$i137 = 0;
           }
          }
         } else {
          $849 = HEAP32[$124>>2]|0;
          switch ($849|0) {
          case 2:  {
           $850 = HEAP32[$117>>2]|0;
           $851 = $850 << 16;
           $$0$i137 = $851;
           break L376;
           break;
          }
          case 1:  {
           $852 = HEAP32[$117>>2]|0;
           $853 = ($852|0)<(0);
           if ($853) {
            $854 = (8192 - ($852))|0;
            $855 = $854 >> 14;
            $856 = (0 - ($855))|0;
            $$0$i137 = $856;
            break L376;
           } else {
            $857 = (($852) + 8192)|0;
            $858 = $857 >> 14;
            $$0$i137 = $858;
            break L376;
           }
           break;
          }
          default: {
           $859 = HEAP32[$117>>2]|0;
           $$0$i137 = $859;
           break L376;
          }
          }
         }
        } while(0);
        $860 = (($$0$i137) + ($8))|0;
        HEAP32[$width>>2] = $860;
       }
       HEAP8[$haveWidth>>0] = 1;
       $861 = HEAP32[$2>>2]|0;
       $862 = ((($861)) + 561|0);
       $863 = HEAP8[$862>>0]|0;
       $864 = ($863<<24>>24)==(0);
       if (!($864)) {
        $$ph = $1756;$lastError$3$ph = 0;$subrStack$sroa$31$3$ph = $108;
        label = 779;
        break L4;
       }
       $865 = ($843|0)==($117|0);
       L391: do {
        if ($865) {
         $866 = HEAP32[$116>>2]|0;
         $867 = ($866|0)==(0|0);
         if (!($867)) {
          $868 = HEAP32[$866>>2]|0;
          $869 = ($868|0)==(0);
          if ($869) {
           HEAP32[$866>>2] = 161;
          }
         }
         $883 = HEAP32[$1>>2]|0;
         $1758 = $883;
         label = 355;
        } else {
         $870 = ((($843)) + -8|0);
         HEAP32[$115>>2] = $870;
         $871 = ((($843)) + -4|0);
         $872 = HEAP32[$871>>2]|0;
         L398: do {
          switch ($872|0) {
          case 2:  {
           $873 = HEAP32[$870>>2]|0;
           $874 = $873 << 16;
           $$0$i135 = $874;
           break;
          }
          case 1:  {
           $875 = HEAP32[$870>>2]|0;
           $876 = ($875|0)<(0);
           if ($876) {
            $877 = (8192 - ($875))|0;
            $878 = $877 >> 14;
            $879 = (0 - ($878))|0;
            $$0$i135 = $879;
            break L398;
           } else {
            $880 = (($875) + 8192)|0;
            $881 = $880 >> 14;
            $$0$i135 = $881;
            break L398;
           }
           break;
          }
          default: {
           $882 = HEAP32[$870>>2]|0;
           $$0$i135 = $882;
          }
          }
         } while(0);
         $884 = HEAP32[$1>>2]|0;
         $885 = (($884) + ($$0$i135))|0;
         HEAP32[$1>>2] = $885;
         $886 = ($870|0)==($117|0);
         if ($886) {
          $1758 = $885;
          label = 355;
         } else {
          $891 = ((($843)) + -16|0);
          HEAP32[$115>>2] = $891;
          $892 = ((($843)) + -12|0);
          $893 = HEAP32[$892>>2]|0;
          switch ($893|0) {
          case 2:  {
           $894 = HEAP32[$891>>2]|0;
           $895 = $894 << 16;
           $$0$i133 = $895;$912 = $885;
           break L391;
           break;
          }
          case 1:  {
           $896 = HEAP32[$891>>2]|0;
           $897 = ($896|0)<(0);
           if ($897) {
            $898 = (8192 - ($896))|0;
            $899 = $898 >> 14;
            $900 = (0 - ($899))|0;
            $$0$i133 = $900;$912 = $885;
            break L391;
           } else {
            $901 = (($896) + 8192)|0;
            $902 = $901 >> 14;
            $$0$i133 = $902;$912 = $885;
            break L391;
           }
           break;
          }
          default: {
           $903 = HEAP32[$891>>2]|0;
           $$0$i133 = $903;$912 = $885;
           break L391;
          }
          }
         }
        }
       } while(0);
       if ((label|0) == 355) {
        label = 0;
        $887 = HEAP32[$116>>2]|0;
        $888 = ($887|0)==(0|0);
        if ($888) {
         $$0$i133 = 0;$912 = $1758;
        } else {
         $889 = HEAP32[$887>>2]|0;
         $890 = ($889|0)==(0);
         if ($890) {
          HEAP32[$887>>2] = 161;
          $$pre252 = HEAP32[$1>>2]|0;
          $$0$i133 = 0;$912 = $$pre252;
         } else {
          $$0$i133 = 0;$912 = $1758;
         }
        }
       }
       $904 = HEAP32[$0>>2]|0;
       $905 = (($904) + ($$0$i133))|0;
       HEAP32[$0>>2] = $905;
       $906 = HEAP8[$96>>0]|0;
       $907 = ($906<<24>>24)==(0);
       if (!($907)) {
        HEAP8[$97>>0] = 1;
        $908 = HEAP32[$119>>2]|0;
        $909 = HEAP32[$121>>2]|0;
        _cf2_glyphpath_lineTo($glyphPath,$908,$909);
        $910 = HEAP8[$98>>0]|0;
        $911 = ($910<<24>>24)==(0);
        if (!($911)) {
         $$0$val$i112 = HEAP32[$$0$i124>>2]|0;
         $$1$val$i114 = HEAP32[$$1$i126>>2]|0;
         _cf2_glyphpath_pushPrevElem($glyphPath,$47,$123,$$0$val$i112,$$1$val$i114,1);
        }
        HEAP8[$95>>0] = 1;
        HEAP8[$96>>0] = 0;
        HEAP8[$97>>0] = 0;
        HEAP8[$98>>0] = 0;
       }
       HEAP32[$119>>2] = $905;
       HEAP32[$120>>2] = $905;
       HEAP32[$121>>2] = $912;
       HEAP32[$122>>2] = $912;
       HEAP8[$95>>0] = 1;
       $$idx1$val$i102 = HEAP8[$$idx1$i116>>0]|0;
       $913 = ($$idx1$val$i102<<24>>24)==(0);
       $$pre$i104 = HEAP32[$73>>2]|0;
       if ($913) {
        label = 370;
       } else {
        $$idx$i105 = ((($$pre$i104)) + 5|0);
        $$idx$val$i106 = HEAP8[$$idx$i105>>0]|0;
        $914 = ($$idx$val$i106<<24>>24)==(0);
        if (!($914)) {
         label = 370;
        }
       }
       if ((label|0) == 370) {
        label = 0;
        $915 = HEAP32[$71>>2]|0;
        $916 = HEAP32[$72>>2]|0;
        $917 = HEAP32[$74>>2]|0;
        _cf2_hintmap_build($47,$915,$916,$$pre$i104,$917,0);
       }
       _memcpy(($42|0),($47|0),3868)|0;
       break;
      }
      case 22:  {
       $918 = HEAP32[$115>>2]|0;
       $919 = (($918) - ($118))|0;
       $920 = $919 >> 3;
       $921 = ($920>>>0)<(2);
       $922 = HEAP8[$haveWidth>>0]|0;
       $923 = ($922<<24>>24)!=(0);
       $or$cond5 = $923 | $921;
       $924 = $918;
       if (!($or$cond5)) {
        $925 = ($920|0)==(0);
        L433: do {
         if ($925) {
          $926 = HEAP32[$116>>2]|0;
          $927 = ($926|0)==(0|0);
          if ($927) {
           $$0$i99 = 0;
          } else {
           $928 = HEAP32[$926>>2]|0;
           $929 = ($928|0)==(0);
           if ($929) {
            HEAP32[$926>>2] = 130;
            $$0$i99 = 0;
           } else {
            $$0$i99 = 0;
           }
          }
         } else {
          $930 = HEAP32[$124>>2]|0;
          switch ($930|0) {
          case 2:  {
           $931 = HEAP32[$117>>2]|0;
           $932 = $931 << 16;
           $$0$i99 = $932;
           break L433;
           break;
          }
          case 1:  {
           $933 = HEAP32[$117>>2]|0;
           $934 = ($933|0)<(0);
           if ($934) {
            $935 = (8192 - ($933))|0;
            $936 = $935 >> 14;
            $937 = (0 - ($936))|0;
            $$0$i99 = $937;
            break L433;
           } else {
            $938 = (($933) + 8192)|0;
            $939 = $938 >> 14;
            $$0$i99 = $939;
            break L433;
           }
           break;
          }
          default: {
           $940 = HEAP32[$117>>2]|0;
           $$0$i99 = $940;
           break L433;
          }
          }
         }
        } while(0);
        $941 = (($$0$i99) + ($8))|0;
        HEAP32[$width>>2] = $941;
       }
       HEAP8[$haveWidth>>0] = 1;
       $942 = HEAP32[$2>>2]|0;
       $943 = ((($942)) + 561|0);
       $944 = HEAP8[$943>>0]|0;
       $945 = ($944<<24>>24)==(0);
       if (!($945)) {
        $$ph = $1756;$lastError$3$ph = 0;$subrStack$sroa$31$3$ph = $108;
        label = 779;
        break L4;
       }
       $946 = ($924|0)==($117|0);
       L448: do {
        if ($946) {
         $947 = HEAP32[$116>>2]|0;
         $948 = ($947|0)==(0|0);
         if ($948) {
          $$0$i98 = 0;
         } else {
          $949 = HEAP32[$947>>2]|0;
          $950 = ($949|0)==(0);
          if ($950) {
           HEAP32[$947>>2] = 161;
           $$0$i98 = 0;
          } else {
           $$0$i98 = 0;
          }
         }
        } else {
         $951 = ((($924)) + -8|0);
         HEAP32[$115>>2] = $951;
         $952 = ((($924)) + -4|0);
         $953 = HEAP32[$952>>2]|0;
         switch ($953|0) {
         case 2:  {
          $954 = HEAP32[$951>>2]|0;
          $955 = $954 << 16;
          $$0$i98 = $955;
          break L448;
          break;
         }
         case 1:  {
          $956 = HEAP32[$951>>2]|0;
          $957 = ($956|0)<(0);
          if ($957) {
           $958 = (8192 - ($956))|0;
           $959 = $958 >> 14;
           $960 = (0 - ($959))|0;
           $$0$i98 = $960;
           break L448;
          } else {
           $961 = (($956) + 8192)|0;
           $962 = $961 >> 14;
           $$0$i98 = $962;
           break L448;
          }
          break;
         }
         default: {
          $963 = HEAP32[$951>>2]|0;
          $$0$i98 = $963;
          break L448;
         }
         }
        }
       } while(0);
       $964 = HEAP32[$0>>2]|0;
       $965 = (($964) + ($$0$i98))|0;
       HEAP32[$0>>2] = $965;
       $966 = HEAP32[$1>>2]|0;
       $967 = HEAP8[$96>>0]|0;
       $968 = ($967<<24>>24)==(0);
       if (!($968)) {
        HEAP8[$97>>0] = 1;
        $969 = HEAP32[$119>>2]|0;
        $970 = HEAP32[$121>>2]|0;
        _cf2_glyphpath_lineTo($glyphPath,$969,$970);
        $971 = HEAP8[$98>>0]|0;
        $972 = ($971<<24>>24)==(0);
        if (!($972)) {
         $$0$val$i = HEAP32[$$0$i124>>2]|0;
         $$1$val$i = HEAP32[$$1$i126>>2]|0;
         _cf2_glyphpath_pushPrevElem($glyphPath,$47,$123,$$0$val$i,$$1$val$i,1);
        }
        HEAP8[$95>>0] = 1;
        HEAP8[$96>>0] = 0;
        HEAP8[$97>>0] = 0;
        HEAP8[$98>>0] = 0;
       }
       HEAP32[$119>>2] = $965;
       HEAP32[$120>>2] = $965;
       HEAP32[$121>>2] = $966;
       HEAP32[$122>>2] = $966;
       HEAP8[$95>>0] = 1;
       $$idx1$val$i = HEAP8[$$idx1$i116>>0]|0;
       $973 = ($$idx1$val$i<<24>>24)==(0);
       $$pre$i93 = HEAP32[$73>>2]|0;
       if ($973) {
        label = 401;
       } else {
        $$idx$i = ((($$pre$i93)) + 5|0);
        $$idx$val$i = HEAP8[$$idx$i>>0]|0;
        $974 = ($$idx$val$i<<24>>24)==(0);
        if (!($974)) {
         label = 401;
        }
       }
       if ((label|0) == 401) {
        label = 0;
        $975 = HEAP32[$71>>2]|0;
        $976 = HEAP32[$72>>2]|0;
        $977 = HEAP32[$74>>2]|0;
        _cf2_hintmap_build($47,$975,$976,$$pre$i93,$977,0);
       }
       _memcpy(($42|0),($47|0),3868)|0;
       break;
      }
      case 25:  {
       $978 = HEAP32[$115>>2]|0;
       $979 = (($978) - ($118))|0;
       $980 = $979 >> 3;
       $981 = ($980>>>0)>(6);
       L473: do {
        if ($981) {
         $984 = $978;$index8$0222 = 0;
         while(1) {
          $983 = (($984) - ($118))|0;
          $985 = $983 >> 3;
          $986 = ($985>>>0)>($index8$0222>>>0);
          L476: do {
           if ($986) {
            $991 = (((($100)) + 8|0) + ($index8$0222<<3)|0);
            $992 = (((((($100)) + 8|0) + ($index8$0222<<3)|0)) + 4|0);
            $993 = HEAP32[$992>>2]|0;
            switch ($993|0) {
            case 2:  {
             $994 = HEAP32[$991>>2]|0;
             $995 = $994 << 16;
             $$0$i91 = $995;
             break L476;
             break;
            }
            case 1:  {
             $996 = HEAP32[$991>>2]|0;
             $997 = ($996|0)<(0);
             if ($997) {
              $998 = (8192 - ($996))|0;
              $999 = $998 >> 14;
              $1000 = (0 - ($999))|0;
              $$0$i91 = $1000;
              break L476;
             } else {
              $1001 = (($996) + 8192)|0;
              $1002 = $1001 >> 14;
              $$0$i91 = $1002;
              break L476;
             }
             break;
            }
            default: {
             $1003 = HEAP32[$991>>2]|0;
             $$0$i91 = $1003;
             break L476;
            }
            }
           } else {
            $987 = HEAP32[$116>>2]|0;
            $988 = ($987|0)==(0|0);
            if ($988) {
             $$0$i91 = 0;
            } else {
             $989 = HEAP32[$987>>2]|0;
             $990 = ($989|0)==(0);
             if ($990) {
              HEAP32[$987>>2] = 130;
              $$0$i91 = 0;
             } else {
              $$0$i91 = 0;
             }
            }
           }
          } while(0);
          $1004 = HEAP32[$0>>2]|0;
          $1005 = (($1004) + ($$0$i91))|0;
          HEAP32[$0>>2] = $1005;
          $1006 = $index8$0222 | 1;
          $1007 = ($985>>>0)>($1006>>>0);
          L489: do {
           if ($1007) {
            $1012 = (((($100)) + 8|0) + ($1006<<3)|0);
            $1013 = (((((($100)) + 8|0) + ($1006<<3)|0)) + 4|0);
            $1014 = HEAP32[$1013>>2]|0;
            switch ($1014|0) {
            case 2:  {
             $1015 = HEAP32[$1012>>2]|0;
             $1016 = $1015 << 16;
             $$0$i89 = $1016;$1027 = $1005;
             break L489;
             break;
            }
            case 1:  {
             $1017 = HEAP32[$1012>>2]|0;
             $1018 = ($1017|0)<(0);
             if ($1018) {
              $1019 = (8192 - ($1017))|0;
              $1020 = $1019 >> 14;
              $1021 = (0 - ($1020))|0;
              $$0$i89 = $1021;$1027 = $1005;
              break L489;
             } else {
              $1022 = (($1017) + 8192)|0;
              $1023 = $1022 >> 14;
              $$0$i89 = $1023;$1027 = $1005;
              break L489;
             }
             break;
            }
            default: {
             $1024 = HEAP32[$1012>>2]|0;
             $$0$i89 = $1024;$1027 = $1005;
             break L489;
            }
            }
           } else {
            $1008 = HEAP32[$116>>2]|0;
            $1009 = ($1008|0)==(0|0);
            if ($1009) {
             $$0$i89 = 0;$1027 = $1005;
            } else {
             $1010 = HEAP32[$1008>>2]|0;
             $1011 = ($1010|0)==(0);
             if ($1011) {
              HEAP32[$1008>>2] = 130;
              $$pre251 = HEAP32[$0>>2]|0;
              $$0$i89 = 0;$1027 = $$pre251;
             } else {
              $$0$i89 = 0;$1027 = $1005;
             }
            }
           }
          } while(0);
          $1025 = HEAP32[$1>>2]|0;
          $1026 = (($1025) + ($$0$i89))|0;
          HEAP32[$1>>2] = $1026;
          _cf2_glyphpath_lineTo($glyphPath,$1027,$1026);
          $1028 = (($index8$0222) + 2)|0;
          $1029 = (($index8$0222) + 8)|0;
          $1030 = ($1029>>>0)<($980>>>0);
          if (!($1030)) {
           $index8$0$lcssa = $1028;
           break L473;
          }
          $$pre250 = HEAP32[$115>>2]|0;
          $984 = $$pre250;$index8$0222 = $1028;
         }
        } else {
         $index8$0$lcssa = 0;
        }
       } while(0);
       $982 = ($index8$0$lcssa>>>0)<($980>>>0);
       if ($982) {
        $index8$1225 = $index8$0$lcssa;
        while(1) {
         $1031 = HEAP32[$115>>2]|0;
         $1032 = (($1031) - ($118))|0;
         $1033 = $1032 >> 3;
         $1034 = ($1033>>>0)>($index8$1225>>>0);
         L507: do {
          if ($1034) {
           $1039 = (((($100)) + 8|0) + ($index8$1225<<3)|0);
           $1040 = (((((($100)) + 8|0) + ($index8$1225<<3)|0)) + 4|0);
           $1041 = HEAP32[$1040>>2]|0;
           switch ($1041|0) {
           case 2:  {
            $1042 = HEAP32[$1039>>2]|0;
            $1043 = $1042 << 16;
            $$0$i87 = $1043;
            break L507;
            break;
           }
           case 1:  {
            $1044 = HEAP32[$1039>>2]|0;
            $1045 = ($1044|0)<(0);
            if ($1045) {
             $1046 = (8192 - ($1044))|0;
             $1047 = $1046 >> 14;
             $1048 = (0 - ($1047))|0;
             $$0$i87 = $1048;
             break L507;
            } else {
             $1049 = (($1044) + 8192)|0;
             $1050 = $1049 >> 14;
             $$0$i87 = $1050;
             break L507;
            }
            break;
           }
           default: {
            $1051 = HEAP32[$1039>>2]|0;
            $$0$i87 = $1051;
            break L507;
           }
           }
          } else {
           $1035 = HEAP32[$116>>2]|0;
           $1036 = ($1035|0)==(0|0);
           if ($1036) {
            $$0$i87 = 0;
           } else {
            $1037 = HEAP32[$1035>>2]|0;
            $1038 = ($1037|0)==(0);
            if ($1038) {
             HEAP32[$1035>>2] = 130;
             $$0$i87 = 0;
            } else {
             $$0$i87 = 0;
            }
           }
          }
         } while(0);
         $1052 = HEAP32[$0>>2]|0;
         $1053 = (($1052) + ($$0$i87))|0;
         $1054 = $index8$1225 | 1;
         $1055 = ($1033>>>0)>($1054>>>0);
         L520: do {
          if ($1055) {
           $1060 = (((($100)) + 8|0) + ($1054<<3)|0);
           $1061 = (((((($100)) + 8|0) + ($1054<<3)|0)) + 4|0);
           $1062 = HEAP32[$1061>>2]|0;
           switch ($1062|0) {
           case 2:  {
            $1063 = HEAP32[$1060>>2]|0;
            $1064 = $1063 << 16;
            $$0$i85 = $1064;
            break L520;
            break;
           }
           case 1:  {
            $1065 = HEAP32[$1060>>2]|0;
            $1066 = ($1065|0)<(0);
            if ($1066) {
             $1067 = (8192 - ($1065))|0;
             $1068 = $1067 >> 14;
             $1069 = (0 - ($1068))|0;
             $$0$i85 = $1069;
             break L520;
            } else {
             $1070 = (($1065) + 8192)|0;
             $1071 = $1070 >> 14;
             $$0$i85 = $1071;
             break L520;
            }
            break;
           }
           default: {
            $1072 = HEAP32[$1060>>2]|0;
            $$0$i85 = $1072;
            break L520;
           }
           }
          } else {
           $1056 = HEAP32[$116>>2]|0;
           $1057 = ($1056|0)==(0|0);
           if ($1057) {
            $$0$i85 = 0;
           } else {
            $1058 = HEAP32[$1056>>2]|0;
            $1059 = ($1058|0)==(0);
            if ($1059) {
             HEAP32[$1056>>2] = 130;
             $$0$i85 = 0;
            } else {
             $$0$i85 = 0;
            }
           }
          }
         } while(0);
         $1073 = HEAP32[$1>>2]|0;
         $1074 = (($1073) + ($$0$i85))|0;
         $1075 = (($index8$1225) + 2)|0;
         $1076 = ($1033>>>0)>($1075>>>0);
         L533: do {
          if ($1076) {
           $1081 = (((($100)) + 8|0) + ($1075<<3)|0);
           $1082 = (((((($100)) + 8|0) + ($1075<<3)|0)) + 4|0);
           $1083 = HEAP32[$1082>>2]|0;
           switch ($1083|0) {
           case 2:  {
            $1084 = HEAP32[$1081>>2]|0;
            $1085 = $1084 << 16;
            $$0$i83 = $1085;
            break L533;
            break;
           }
           case 1:  {
            $1086 = HEAP32[$1081>>2]|0;
            $1087 = ($1086|0)<(0);
            if ($1087) {
             $1088 = (8192 - ($1086))|0;
             $1089 = $1088 >> 14;
             $1090 = (0 - ($1089))|0;
             $$0$i83 = $1090;
             break L533;
            } else {
             $1091 = (($1086) + 8192)|0;
             $1092 = $1091 >> 14;
             $$0$i83 = $1092;
             break L533;
            }
            break;
           }
           default: {
            $1093 = HEAP32[$1081>>2]|0;
            $$0$i83 = $1093;
            break L533;
           }
           }
          } else {
           $1077 = HEAP32[$116>>2]|0;
           $1078 = ($1077|0)==(0|0);
           if ($1078) {
            $$0$i83 = 0;
           } else {
            $1079 = HEAP32[$1077>>2]|0;
            $1080 = ($1079|0)==(0);
            if ($1080) {
             HEAP32[$1077>>2] = 130;
             $$0$i83 = 0;
            } else {
             $$0$i83 = 0;
            }
           }
          }
         } while(0);
         $1094 = (($$0$i83) + ($1053))|0;
         $1095 = (($index8$1225) + 3)|0;
         $1096 = ($1033>>>0)>($1095>>>0);
         L546: do {
          if ($1096) {
           $1101 = (((($100)) + 8|0) + ($1095<<3)|0);
           $1102 = (((((($100)) + 8|0) + ($1095<<3)|0)) + 4|0);
           $1103 = HEAP32[$1102>>2]|0;
           switch ($1103|0) {
           case 2:  {
            $1104 = HEAP32[$1101>>2]|0;
            $1105 = $1104 << 16;
            $$0$i81 = $1105;
            break L546;
            break;
           }
           case 1:  {
            $1106 = HEAP32[$1101>>2]|0;
            $1107 = ($1106|0)<(0);
            if ($1107) {
             $1108 = (8192 - ($1106))|0;
             $1109 = $1108 >> 14;
             $1110 = (0 - ($1109))|0;
             $$0$i81 = $1110;
             break L546;
            } else {
             $1111 = (($1106) + 8192)|0;
             $1112 = $1111 >> 14;
             $$0$i81 = $1112;
             break L546;
            }
            break;
           }
           default: {
            $1113 = HEAP32[$1101>>2]|0;
            $$0$i81 = $1113;
            break L546;
           }
           }
          } else {
           $1097 = HEAP32[$116>>2]|0;
           $1098 = ($1097|0)==(0|0);
           if ($1098) {
            $$0$i81 = 0;
           } else {
            $1099 = HEAP32[$1097>>2]|0;
            $1100 = ($1099|0)==(0);
            if ($1100) {
             HEAP32[$1097>>2] = 130;
             $$0$i81 = 0;
            } else {
             $$0$i81 = 0;
            }
           }
          }
         } while(0);
         $1114 = (($$0$i81) + ($1074))|0;
         $1115 = (($index8$1225) + 4)|0;
         $1116 = ($1033>>>0)>($1115>>>0);
         L559: do {
          if ($1116) {
           $1121 = (((($100)) + 8|0) + ($1115<<3)|0);
           $1122 = (((((($100)) + 8|0) + ($1115<<3)|0)) + 4|0);
           $1123 = HEAP32[$1122>>2]|0;
           switch ($1123|0) {
           case 2:  {
            $1124 = HEAP32[$1121>>2]|0;
            $1125 = $1124 << 16;
            $$0$i79 = $1125;
            break L559;
            break;
           }
           case 1:  {
            $1126 = HEAP32[$1121>>2]|0;
            $1127 = ($1126|0)<(0);
            if ($1127) {
             $1128 = (8192 - ($1126))|0;
             $1129 = $1128 >> 14;
             $1130 = (0 - ($1129))|0;
             $$0$i79 = $1130;
             break L559;
            } else {
             $1131 = (($1126) + 8192)|0;
             $1132 = $1131 >> 14;
             $$0$i79 = $1132;
             break L559;
            }
            break;
           }
           default: {
            $1133 = HEAP32[$1121>>2]|0;
            $$0$i79 = $1133;
            break L559;
           }
           }
          } else {
           $1117 = HEAP32[$116>>2]|0;
           $1118 = ($1117|0)==(0|0);
           if ($1118) {
            $$0$i79 = 0;
           } else {
            $1119 = HEAP32[$1117>>2]|0;
            $1120 = ($1119|0)==(0);
            if (!($1120)) {
             $$0$i79 = 0;
             break;
            }
            HEAP32[$1117>>2] = 130;
            $$0$i79 = 0;
           }
          }
         } while(0);
         $1134 = (($$0$i79) + ($1094))|0;
         $1135 = (($index8$1225) + 5)|0;
         $1136 = ($1033>>>0)>($1135>>>0);
         L572: do {
          if ($1136) {
           $1141 = (((($100)) + 8|0) + ($1135<<3)|0);
           $1142 = (((((($100)) + 8|0) + ($1135<<3)|0)) + 4|0);
           $1143 = HEAP32[$1142>>2]|0;
           switch ($1143|0) {
           case 2:  {
            $1144 = HEAP32[$1141>>2]|0;
            $1145 = $1144 << 16;
            $$0$i77 = $1145;
            break L572;
            break;
           }
           case 1:  {
            $1146 = HEAP32[$1141>>2]|0;
            $1147 = ($1146|0)<(0);
            if ($1147) {
             $1148 = (8192 - ($1146))|0;
             $1149 = $1148 >> 14;
             $1150 = (0 - ($1149))|0;
             $$0$i77 = $1150;
             break L572;
            } else {
             $1151 = (($1146) + 8192)|0;
             $1152 = $1151 >> 14;
             $$0$i77 = $1152;
             break L572;
            }
            break;
           }
           default: {
            $1153 = HEAP32[$1141>>2]|0;
            $$0$i77 = $1153;
            break L572;
           }
           }
          } else {
           $1137 = HEAP32[$116>>2]|0;
           $1138 = ($1137|0)==(0|0);
           if ($1138) {
            $$0$i77 = 0;
            break;
           }
           $1139 = HEAP32[$1137>>2]|0;
           $1140 = ($1139|0)==(0);
           if (!($1140)) {
            $$0$i77 = 0;
            break;
           }
           HEAP32[$1137>>2] = 130;
           $$0$i77 = 0;
          }
         } while(0);
         $1154 = (($$0$i77) + ($1114))|0;
         _cf2_glyphpath_curveTo($glyphPath,$1053,$1074,$1094,$1114,$1134,$1154);
         HEAP32[$0>>2] = $1134;
         HEAP32[$1>>2] = $1154;
         $1155 = (($index8$1225) + 6)|0;
         $1156 = ($1155>>>0)<($980>>>0);
         if ($1156) {
          $index8$1225 = $1155;
         } else {
          break;
         }
        }
       }
       HEAP32[$115>>2] = $117;
       $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
       continue L12;
       break;
      }
      case 26:  {
       $1157 = HEAP32[$115>>2]|0;
       $1158 = (($1157) - ($118))|0;
       $1159 = $1158 >> 3;
       $1160 = $1159 & -3;
       $1161 = (($1159) - ($1160))|0;
       $1162 = ($1160>>>0)>($1161>>>0);
       if ($1162) {
        $index16$0218 = $1161;
        while(1) {
         $1163 = (($1159) - ($index16$0218))|0;
         $1164 = $1163 & 1;
         $1165 = ($1164|0)==(0);
         if ($1165) {
          $1190 = HEAP32[$0>>2]|0;
          $$pre249 = HEAP32[$115>>2]|0;
          $1192 = $$pre249;$index16$1 = $index16$0218;$x117$0 = $1190;
         } else {
          $1166 = HEAP32[$115>>2]|0;
          $1167 = (($1166) - ($118))|0;
          $1168 = $1167 >> 3;
          $1169 = ($1168>>>0)>($index16$0218>>>0);
          L593: do {
           if ($1169) {
            $1174 = (((($100)) + 8|0) + ($index16$0218<<3)|0);
            $1175 = (((((($100)) + 8|0) + ($index16$0218<<3)|0)) + 4|0);
            $1176 = HEAP32[$1175>>2]|0;
            switch ($1176|0) {
            case 2:  {
             $1177 = HEAP32[$1174>>2]|0;
             $1178 = $1177 << 16;
             $$0$i73 = $1178;
             break L593;
             break;
            }
            case 1:  {
             $1179 = HEAP32[$1174>>2]|0;
             $1180 = ($1179|0)<(0);
             if ($1180) {
              $1181 = (8192 - ($1179))|0;
              $1182 = $1181 >> 14;
              $1183 = (0 - ($1182))|0;
              $$0$i73 = $1183;
              break L593;
             } else {
              $1184 = (($1179) + 8192)|0;
              $1185 = $1184 >> 14;
              $$0$i73 = $1185;
              break L593;
             }
             break;
            }
            default: {
             $1186 = HEAP32[$1174>>2]|0;
             $$0$i73 = $1186;
             break L593;
            }
            }
           } else {
            $1170 = HEAP32[$116>>2]|0;
            $1171 = ($1170|0)==(0|0);
            if ($1171) {
             $$0$i73 = 0;
            } else {
             $1172 = HEAP32[$1170>>2]|0;
             $1173 = ($1172|0)==(0);
             if ($1173) {
              HEAP32[$1170>>2] = 130;
              $$0$i73 = 0;
             } else {
              $$0$i73 = 0;
             }
            }
           }
          } while(0);
          $1187 = HEAP32[$0>>2]|0;
          $1188 = (($1187) + ($$0$i73))|0;
          $1189 = (($index16$0218) + 1)|0;
          $1192 = $1166;$index16$1 = $1189;$x117$0 = $1188;
         }
         $1191 = (($1192) - ($118))|0;
         $1193 = $1191 >> 3;
         $1194 = ($1193>>>0)>($index16$1>>>0);
         L607: do {
          if ($1194) {
           $1199 = (((($100)) + 8|0) + ($index16$1<<3)|0);
           $1200 = (((((($100)) + 8|0) + ($index16$1<<3)|0)) + 4|0);
           $1201 = HEAP32[$1200>>2]|0;
           switch ($1201|0) {
           case 2:  {
            $1202 = HEAP32[$1199>>2]|0;
            $1203 = $1202 << 16;
            $$0$i71 = $1203;
            break L607;
            break;
           }
           case 1:  {
            $1204 = HEAP32[$1199>>2]|0;
            $1205 = ($1204|0)<(0);
            if ($1205) {
             $1206 = (8192 - ($1204))|0;
             $1207 = $1206 >> 14;
             $1208 = (0 - ($1207))|0;
             $$0$i71 = $1208;
             break L607;
            } else {
             $1209 = (($1204) + 8192)|0;
             $1210 = $1209 >> 14;
             $$0$i71 = $1210;
             break L607;
            }
            break;
           }
           default: {
            $1211 = HEAP32[$1199>>2]|0;
            $$0$i71 = $1211;
            break L607;
           }
           }
          } else {
           $1195 = HEAP32[$116>>2]|0;
           $1196 = ($1195|0)==(0|0);
           if ($1196) {
            $$0$i71 = 0;
           } else {
            $1197 = HEAP32[$1195>>2]|0;
            $1198 = ($1197|0)==(0);
            if ($1198) {
             HEAP32[$1195>>2] = 130;
             $$0$i71 = 0;
            } else {
             $$0$i71 = 0;
            }
           }
          }
         } while(0);
         $1212 = HEAP32[$1>>2]|0;
         $1213 = (($1212) + ($$0$i71))|0;
         $1214 = (($index16$1) + 1)|0;
         $1215 = ($1193>>>0)>($1214>>>0);
         L620: do {
          if ($1215) {
           $1220 = (((($100)) + 8|0) + ($1214<<3)|0);
           $1221 = (((((($100)) + 8|0) + ($1214<<3)|0)) + 4|0);
           $1222 = HEAP32[$1221>>2]|0;
           switch ($1222|0) {
           case 2:  {
            $1223 = HEAP32[$1220>>2]|0;
            $1224 = $1223 << 16;
            $$0$i69 = $1224;
            break L620;
            break;
           }
           case 1:  {
            $1225 = HEAP32[$1220>>2]|0;
            $1226 = ($1225|0)<(0);
            if ($1226) {
             $1227 = (8192 - ($1225))|0;
             $1228 = $1227 >> 14;
             $1229 = (0 - ($1228))|0;
             $$0$i69 = $1229;
             break L620;
            } else {
             $1230 = (($1225) + 8192)|0;
             $1231 = $1230 >> 14;
             $$0$i69 = $1231;
             break L620;
            }
            break;
           }
           default: {
            $1232 = HEAP32[$1220>>2]|0;
            $$0$i69 = $1232;
            break L620;
           }
           }
          } else {
           $1216 = HEAP32[$116>>2]|0;
           $1217 = ($1216|0)==(0|0);
           if ($1217) {
            $$0$i69 = 0;
           } else {
            $1218 = HEAP32[$1216>>2]|0;
            $1219 = ($1218|0)==(0);
            if ($1219) {
             HEAP32[$1216>>2] = 130;
             $$0$i69 = 0;
            } else {
             $$0$i69 = 0;
            }
           }
          }
         } while(0);
         $1233 = (($$0$i69) + ($x117$0))|0;
         $1234 = (($index16$1) + 2)|0;
         $1235 = ($1193>>>0)>($1234>>>0);
         L633: do {
          if ($1235) {
           $1240 = (((($100)) + 8|0) + ($1234<<3)|0);
           $1241 = (((((($100)) + 8|0) + ($1234<<3)|0)) + 4|0);
           $1242 = HEAP32[$1241>>2]|0;
           switch ($1242|0) {
           case 2:  {
            $1243 = HEAP32[$1240>>2]|0;
            $1244 = $1243 << 16;
            $$0$i67 = $1244;
            break L633;
            break;
           }
           case 1:  {
            $1245 = HEAP32[$1240>>2]|0;
            $1246 = ($1245|0)<(0);
            if ($1246) {
             $1247 = (8192 - ($1245))|0;
             $1248 = $1247 >> 14;
             $1249 = (0 - ($1248))|0;
             $$0$i67 = $1249;
             break L633;
            } else {
             $1250 = (($1245) + 8192)|0;
             $1251 = $1250 >> 14;
             $$0$i67 = $1251;
             break L633;
            }
            break;
           }
           default: {
            $1252 = HEAP32[$1240>>2]|0;
            $$0$i67 = $1252;
            break L633;
           }
           }
          } else {
           $1236 = HEAP32[$116>>2]|0;
           $1237 = ($1236|0)==(0|0);
           if ($1237) {
            $$0$i67 = 0;
           } else {
            $1238 = HEAP32[$1236>>2]|0;
            $1239 = ($1238|0)==(0);
            if ($1239) {
             HEAP32[$1236>>2] = 130;
             $$0$i67 = 0;
            } else {
             $$0$i67 = 0;
            }
           }
          }
         } while(0);
         $1253 = (($$0$i67) + ($1213))|0;
         $1254 = (($index16$1) + 3)|0;
         $1255 = ($1193>>>0)>($1254>>>0);
         L646: do {
          if ($1255) {
           $1260 = (((($100)) + 8|0) + ($1254<<3)|0);
           $1261 = (((((($100)) + 8|0) + ($1254<<3)|0)) + 4|0);
           $1262 = HEAP32[$1261>>2]|0;
           switch ($1262|0) {
           case 2:  {
            $1263 = HEAP32[$1260>>2]|0;
            $1264 = $1263 << 16;
            $$0$i65 = $1264;
            break L646;
            break;
           }
           case 1:  {
            $1265 = HEAP32[$1260>>2]|0;
            $1266 = ($1265|0)<(0);
            if ($1266) {
             $1267 = (8192 - ($1265))|0;
             $1268 = $1267 >> 14;
             $1269 = (0 - ($1268))|0;
             $$0$i65 = $1269;
             break L646;
            } else {
             $1270 = (($1265) + 8192)|0;
             $1271 = $1270 >> 14;
             $$0$i65 = $1271;
             break L646;
            }
            break;
           }
           default: {
            $1272 = HEAP32[$1260>>2]|0;
            $$0$i65 = $1272;
            break L646;
           }
           }
          } else {
           $1256 = HEAP32[$116>>2]|0;
           $1257 = ($1256|0)==(0|0);
           if ($1257) {
            $$0$i65 = 0;
           } else {
            $1258 = HEAP32[$1256>>2]|0;
            $1259 = ($1258|0)==(0);
            if ($1259) {
             HEAP32[$1256>>2] = 130;
             $$0$i65 = 0;
            } else {
             $$0$i65 = 0;
            }
           }
          }
         } while(0);
         $1273 = (($$0$i65) + ($1253))|0;
         _cf2_glyphpath_curveTo($glyphPath,$x117$0,$1213,$1233,$1253,$1233,$1273);
         HEAP32[$0>>2] = $1233;
         HEAP32[$1>>2] = $1273;
         $1274 = (($index16$1) + 4)|0;
         $1275 = ($1160>>>0)>($1274>>>0);
         if ($1275) {
          $index16$0218 = $1274;
         } else {
          break;
         }
        }
       }
       HEAP32[$115>>2] = $117;
       $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
       continue L12;
       break;
      }
      case 27:  {
       $1276 = HEAP32[$115>>2]|0;
       $1277 = (($1276) - ($118))|0;
       $1278 = $1277 >> 3;
       $1279 = $1278 & -3;
       $1280 = (($1278) - ($1279))|0;
       $1281 = ($1279>>>0)>($1280>>>0);
       if ($1281) {
        $index25$0214 = $1280;
        while(1) {
         $1282 = (($1278) - ($index25$0214))|0;
         $1283 = $1282 & 1;
         $1284 = ($1283|0)==(0);
         if ($1284) {
          $1309 = HEAP32[$1>>2]|0;
          $$pre = HEAP32[$115>>2]|0;
          $1311 = $$pre;$index25$1 = $index25$0214;$y127$0 = $1309;
         } else {
          $1285 = HEAP32[$115>>2]|0;
          $1286 = (($1285) - ($118))|0;
          $1287 = $1286 >> 3;
          $1288 = ($1287>>>0)>($index25$0214>>>0);
          L667: do {
           if ($1288) {
            $1293 = (((($100)) + 8|0) + ($index25$0214<<3)|0);
            $1294 = (((((($100)) + 8|0) + ($index25$0214<<3)|0)) + 4|0);
            $1295 = HEAP32[$1294>>2]|0;
            switch ($1295|0) {
            case 2:  {
             $1296 = HEAP32[$1293>>2]|0;
             $1297 = $1296 << 16;
             $$0$i63 = $1297;
             break L667;
             break;
            }
            case 1:  {
             $1298 = HEAP32[$1293>>2]|0;
             $1299 = ($1298|0)<(0);
             if ($1299) {
              $1300 = (8192 - ($1298))|0;
              $1301 = $1300 >> 14;
              $1302 = (0 - ($1301))|0;
              $$0$i63 = $1302;
              break L667;
             } else {
              $1303 = (($1298) + 8192)|0;
              $1304 = $1303 >> 14;
              $$0$i63 = $1304;
              break L667;
             }
             break;
            }
            default: {
             $1305 = HEAP32[$1293>>2]|0;
             $$0$i63 = $1305;
             break L667;
            }
            }
           } else {
            $1289 = HEAP32[$116>>2]|0;
            $1290 = ($1289|0)==(0|0);
            if ($1290) {
             $$0$i63 = 0;
            } else {
             $1291 = HEAP32[$1289>>2]|0;
             $1292 = ($1291|0)==(0);
             if ($1292) {
              HEAP32[$1289>>2] = 130;
              $$0$i63 = 0;
             } else {
              $$0$i63 = 0;
             }
            }
           }
          } while(0);
          $1306 = HEAP32[$1>>2]|0;
          $1307 = (($1306) + ($$0$i63))|0;
          $1308 = (($index25$0214) + 1)|0;
          $1311 = $1285;$index25$1 = $1308;$y127$0 = $1307;
         }
         $1310 = (($1311) - ($118))|0;
         $1312 = $1310 >> 3;
         $1313 = ($1312>>>0)>($index25$1>>>0);
         L681: do {
          if ($1313) {
           $1318 = (((($100)) + 8|0) + ($index25$1<<3)|0);
           $1319 = (((((($100)) + 8|0) + ($index25$1<<3)|0)) + 4|0);
           $1320 = HEAP32[$1319>>2]|0;
           switch ($1320|0) {
           case 2:  {
            $1321 = HEAP32[$1318>>2]|0;
            $1322 = $1321 << 16;
            $$0$i61 = $1322;
            break L681;
            break;
           }
           case 1:  {
            $1323 = HEAP32[$1318>>2]|0;
            $1324 = ($1323|0)<(0);
            if ($1324) {
             $1325 = (8192 - ($1323))|0;
             $1326 = $1325 >> 14;
             $1327 = (0 - ($1326))|0;
             $$0$i61 = $1327;
             break L681;
            } else {
             $1328 = (($1323) + 8192)|0;
             $1329 = $1328 >> 14;
             $$0$i61 = $1329;
             break L681;
            }
            break;
           }
           default: {
            $1330 = HEAP32[$1318>>2]|0;
            $$0$i61 = $1330;
            break L681;
           }
           }
          } else {
           $1314 = HEAP32[$116>>2]|0;
           $1315 = ($1314|0)==(0|0);
           if ($1315) {
            $$0$i61 = 0;
           } else {
            $1316 = HEAP32[$1314>>2]|0;
            $1317 = ($1316|0)==(0);
            if ($1317) {
             HEAP32[$1314>>2] = 130;
             $$0$i61 = 0;
            } else {
             $$0$i61 = 0;
            }
           }
          }
         } while(0);
         $1331 = HEAP32[$0>>2]|0;
         $1332 = (($1331) + ($$0$i61))|0;
         $1333 = (($index25$1) + 1)|0;
         $1334 = ($1312>>>0)>($1333>>>0);
         L694: do {
          if ($1334) {
           $1339 = (((($100)) + 8|0) + ($1333<<3)|0);
           $1340 = (((((($100)) + 8|0) + ($1333<<3)|0)) + 4|0);
           $1341 = HEAP32[$1340>>2]|0;
           switch ($1341|0) {
           case 2:  {
            $1342 = HEAP32[$1339>>2]|0;
            $1343 = $1342 << 16;
            $$0$i59 = $1343;
            break L694;
            break;
           }
           case 1:  {
            $1344 = HEAP32[$1339>>2]|0;
            $1345 = ($1344|0)<(0);
            if ($1345) {
             $1346 = (8192 - ($1344))|0;
             $1347 = $1346 >> 14;
             $1348 = (0 - ($1347))|0;
             $$0$i59 = $1348;
             break L694;
            } else {
             $1349 = (($1344) + 8192)|0;
             $1350 = $1349 >> 14;
             $$0$i59 = $1350;
             break L694;
            }
            break;
           }
           default: {
            $1351 = HEAP32[$1339>>2]|0;
            $$0$i59 = $1351;
            break L694;
           }
           }
          } else {
           $1335 = HEAP32[$116>>2]|0;
           $1336 = ($1335|0)==(0|0);
           if ($1336) {
            $$0$i59 = 0;
           } else {
            $1337 = HEAP32[$1335>>2]|0;
            $1338 = ($1337|0)==(0);
            if ($1338) {
             HEAP32[$1335>>2] = 130;
             $$0$i59 = 0;
            } else {
             $$0$i59 = 0;
            }
           }
          }
         } while(0);
         $1352 = (($$0$i59) + ($1332))|0;
         $1353 = (($index25$1) + 2)|0;
         $1354 = ($1312>>>0)>($1353>>>0);
         L707: do {
          if ($1354) {
           $1359 = (((($100)) + 8|0) + ($1353<<3)|0);
           $1360 = (((((($100)) + 8|0) + ($1353<<3)|0)) + 4|0);
           $1361 = HEAP32[$1360>>2]|0;
           switch ($1361|0) {
           case 2:  {
            $1362 = HEAP32[$1359>>2]|0;
            $1363 = $1362 << 16;
            $$0$i57 = $1363;
            break L707;
            break;
           }
           case 1:  {
            $1364 = HEAP32[$1359>>2]|0;
            $1365 = ($1364|0)<(0);
            if ($1365) {
             $1366 = (8192 - ($1364))|0;
             $1367 = $1366 >> 14;
             $1368 = (0 - ($1367))|0;
             $$0$i57 = $1368;
             break L707;
            } else {
             $1369 = (($1364) + 8192)|0;
             $1370 = $1369 >> 14;
             $$0$i57 = $1370;
             break L707;
            }
            break;
           }
           default: {
            $1371 = HEAP32[$1359>>2]|0;
            $$0$i57 = $1371;
            break L707;
           }
           }
          } else {
           $1355 = HEAP32[$116>>2]|0;
           $1356 = ($1355|0)==(0|0);
           if ($1356) {
            $$0$i57 = 0;
           } else {
            $1357 = HEAP32[$1355>>2]|0;
            $1358 = ($1357|0)==(0);
            if ($1358) {
             HEAP32[$1355>>2] = 130;
             $$0$i57 = 0;
            } else {
             $$0$i57 = 0;
            }
           }
          }
         } while(0);
         $1372 = (($$0$i57) + ($y127$0))|0;
         $1373 = (($index25$1) + 3)|0;
         $1374 = ($1312>>>0)>($1373>>>0);
         L720: do {
          if ($1374) {
           $1379 = (((($100)) + 8|0) + ($1373<<3)|0);
           $1380 = (((((($100)) + 8|0) + ($1373<<3)|0)) + 4|0);
           $1381 = HEAP32[$1380>>2]|0;
           switch ($1381|0) {
           case 2:  {
            $1382 = HEAP32[$1379>>2]|0;
            $1383 = $1382 << 16;
            $$0$i55 = $1383;
            break L720;
            break;
           }
           case 1:  {
            $1384 = HEAP32[$1379>>2]|0;
            $1385 = ($1384|0)<(0);
            if ($1385) {
             $1386 = (8192 - ($1384))|0;
             $1387 = $1386 >> 14;
             $1388 = (0 - ($1387))|0;
             $$0$i55 = $1388;
             break L720;
            } else {
             $1389 = (($1384) + 8192)|0;
             $1390 = $1389 >> 14;
             $$0$i55 = $1390;
             break L720;
            }
            break;
           }
           default: {
            $1391 = HEAP32[$1379>>2]|0;
            $$0$i55 = $1391;
            break L720;
           }
           }
          } else {
           $1375 = HEAP32[$116>>2]|0;
           $1376 = ($1375|0)==(0|0);
           if ($1376) {
            $$0$i55 = 0;
           } else {
            $1377 = HEAP32[$1375>>2]|0;
            $1378 = ($1377|0)==(0);
            if ($1378) {
             HEAP32[$1375>>2] = 130;
             $$0$i55 = 0;
            } else {
             $$0$i55 = 0;
            }
           }
          }
         } while(0);
         $1392 = (($$0$i55) + ($1352))|0;
         _cf2_glyphpath_curveTo($glyphPath,$1332,$y127$0,$1352,$1372,$1392,$1372);
         HEAP32[$0>>2] = $1392;
         HEAP32[$1>>2] = $1372;
         $1393 = (($index25$1) + 4)|0;
         $1394 = ($1279>>>0)>($1393>>>0);
         if ($1394) {
          $index25$0214 = $1393;
         } else {
          break;
         }
        }
       }
       HEAP32[$115>>2] = $117;
       $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
       continue L12;
       break;
      }
      case 31: case 30:  {
       $1395 = HEAP32[$115>>2]|0;
       $1396 = (($1395) - ($118))|0;
       $1397 = $1396 >> 3;
       $1398 = $1397 & -3;
       $1399 = (($1397) - ($1398))|0;
       $1400 = ($1398>>>0)>($1399>>>0);
       if ($1400) {
        $1401 = ($op1$0|0)==(31);
        $1402 = $1401&1;
        $alternate$0213 = $1402;$index34$0212 = $1399;
        while(1) {
         $1403 = ($alternate$0213<<24>>24)==(0);
         if ($1403) {
          $1510 = HEAP32[$0>>2]|0;
          $1511 = HEAP32[$115>>2]|0;
          $1512 = (($1511) - ($118))|0;
          $1513 = $1512 >> 3;
          $1514 = ($1513>>>0)>($index34$0212>>>0);
          L741: do {
           if ($1514) {
            $1519 = (((($100)) + 8|0) + ($index34$0212<<3)|0);
            $1520 = (((((($100)) + 8|0) + ($index34$0212<<3)|0)) + 4|0);
            $1521 = HEAP32[$1520>>2]|0;
            switch ($1521|0) {
            case 2:  {
             $1522 = HEAP32[$1519>>2]|0;
             $1523 = $1522 << 16;
             $$0$i43 = $1523;
             break L741;
             break;
            }
            case 1:  {
             $1524 = HEAP32[$1519>>2]|0;
             $1525 = ($1524|0)<(0);
             if ($1525) {
              $1526 = (8192 - ($1524))|0;
              $1527 = $1526 >> 14;
              $1528 = (0 - ($1527))|0;
              $$0$i43 = $1528;
              break L741;
             } else {
              $1529 = (($1524) + 8192)|0;
              $1530 = $1529 >> 14;
              $$0$i43 = $1530;
              break L741;
             }
             break;
            }
            default: {
             $1531 = HEAP32[$1519>>2]|0;
             $$0$i43 = $1531;
             break L741;
            }
            }
           } else {
            $1515 = HEAP32[$116>>2]|0;
            $1516 = ($1515|0)==(0|0);
            if ($1516) {
             $$0$i43 = 0;
            } else {
             $1517 = HEAP32[$1515>>2]|0;
             $1518 = ($1517|0)==(0);
             if ($1518) {
              HEAP32[$1515>>2] = 130;
              $$0$i43 = 0;
             } else {
              $$0$i43 = 0;
             }
            }
           }
          } while(0);
          $1532 = HEAP32[$1>>2]|0;
          $1533 = (($1532) + ($$0$i43))|0;
          $1534 = (($index34$0212) + 1)|0;
          $1535 = ($1513>>>0)>($1534>>>0);
          L754: do {
           if ($1535) {
            $1540 = (((($100)) + 8|0) + ($1534<<3)|0);
            $1541 = (((((($100)) + 8|0) + ($1534<<3)|0)) + 4|0);
            $1542 = HEAP32[$1541>>2]|0;
            switch ($1542|0) {
            case 2:  {
             $1543 = HEAP32[$1540>>2]|0;
             $1544 = $1543 << 16;
             $$0$i41 = $1544;
             break L754;
             break;
            }
            case 1:  {
             $1545 = HEAP32[$1540>>2]|0;
             $1546 = ($1545|0)<(0);
             if ($1546) {
              $1547 = (8192 - ($1545))|0;
              $1548 = $1547 >> 14;
              $1549 = (0 - ($1548))|0;
              $$0$i41 = $1549;
              break L754;
             } else {
              $1550 = (($1545) + 8192)|0;
              $1551 = $1550 >> 14;
              $$0$i41 = $1551;
              break L754;
             }
             break;
            }
            default: {
             $1552 = HEAP32[$1540>>2]|0;
             $$0$i41 = $1552;
             break L754;
            }
            }
           } else {
            $1536 = HEAP32[$116>>2]|0;
            $1537 = ($1536|0)==(0|0);
            if ($1537) {
             $$0$i41 = 0;
            } else {
             $1538 = HEAP32[$1536>>2]|0;
             $1539 = ($1538|0)==(0);
             if ($1539) {
              HEAP32[$1536>>2] = 130;
              $$0$i41 = 0;
             } else {
              $$0$i41 = 0;
             }
            }
           }
          } while(0);
          $1553 = (($$0$i41) + ($1510))|0;
          $1554 = (($index34$0212) + 2)|0;
          $1555 = ($1513>>>0)>($1554>>>0);
          L767: do {
           if ($1555) {
            $1560 = (((($100)) + 8|0) + ($1554<<3)|0);
            $1561 = (((((($100)) + 8|0) + ($1554<<3)|0)) + 4|0);
            $1562 = HEAP32[$1561>>2]|0;
            switch ($1562|0) {
            case 2:  {
             $1563 = HEAP32[$1560>>2]|0;
             $1564 = $1563 << 16;
             $$0$i39 = $1564;
             break L767;
             break;
            }
            case 1:  {
             $1565 = HEAP32[$1560>>2]|0;
             $1566 = ($1565|0)<(0);
             if ($1566) {
              $1567 = (8192 - ($1565))|0;
              $1568 = $1567 >> 14;
              $1569 = (0 - ($1568))|0;
              $$0$i39 = $1569;
              break L767;
             } else {
              $1570 = (($1565) + 8192)|0;
              $1571 = $1570 >> 14;
              $$0$i39 = $1571;
              break L767;
             }
             break;
            }
            default: {
             $1572 = HEAP32[$1560>>2]|0;
             $$0$i39 = $1572;
             break L767;
            }
            }
           } else {
            $1556 = HEAP32[$116>>2]|0;
            $1557 = ($1556|0)==(0|0);
            if ($1557) {
             $$0$i39 = 0;
            } else {
             $1558 = HEAP32[$1556>>2]|0;
             $1559 = ($1558|0)==(0);
             if ($1559) {
              HEAP32[$1556>>2] = 130;
              $$0$i39 = 0;
             } else {
              $$0$i39 = 0;
             }
            }
           }
          } while(0);
          $1573 = (($$0$i39) + ($1533))|0;
          $1574 = (($index34$0212) + 3)|0;
          $1575 = ($1513>>>0)>($1574>>>0);
          L780: do {
           if ($1575) {
            $1580 = (((($100)) + 8|0) + ($1574<<3)|0);
            $1581 = (((((($100)) + 8|0) + ($1574<<3)|0)) + 4|0);
            $1582 = HEAP32[$1581>>2]|0;
            switch ($1582|0) {
            case 2:  {
             $1583 = HEAP32[$1580>>2]|0;
             $1584 = $1583 << 16;
             $$0$i37 = $1584;
             break L780;
             break;
            }
            case 1:  {
             $1585 = HEAP32[$1580>>2]|0;
             $1586 = ($1585|0)<(0);
             if ($1586) {
              $1587 = (8192 - ($1585))|0;
              $1588 = $1587 >> 14;
              $1589 = (0 - ($1588))|0;
              $$0$i37 = $1589;
              break L780;
             } else {
              $1590 = (($1585) + 8192)|0;
              $1591 = $1590 >> 14;
              $$0$i37 = $1591;
              break L780;
             }
             break;
            }
            default: {
             $1592 = HEAP32[$1580>>2]|0;
             $$0$i37 = $1592;
             break L780;
            }
            }
           } else {
            $1576 = HEAP32[$116>>2]|0;
            $1577 = ($1576|0)==(0|0);
            if ($1577) {
             $$0$i37 = 0;
            } else {
             $1578 = HEAP32[$1576>>2]|0;
             $1579 = ($1578|0)==(0);
             if (!($1579)) {
              $$0$i37 = 0;
              break;
             }
             HEAP32[$1576>>2] = 130;
             $$0$i37 = 0;
            }
           }
          } while(0);
          $1593 = (($$0$i37) + ($1553))|0;
          $1594 = (($1398) - ($index34$0212))|0;
          $1595 = ($1594|0)==(5);
          if ($1595) {
           $1596 = (($index34$0212) + 4)|0;
           $1597 = ($1513>>>0)>($1596>>>0);
           L794: do {
            if ($1597) {
             $1602 = (((($100)) + 8|0) + ($1596<<3)|0);
             $1603 = (((((($100)) + 8|0) + ($1596<<3)|0)) + 4|0);
             $1604 = HEAP32[$1603>>2]|0;
             switch ($1604|0) {
             case 2:  {
              $1605 = HEAP32[$1602>>2]|0;
              $1606 = $1605 << 16;
              $$0$i36 = $1606;
              break L794;
              break;
             }
             case 1:  {
              $1607 = HEAP32[$1602>>2]|0;
              $1608 = ($1607|0)<(0);
              if ($1608) {
               $1609 = (8192 - ($1607))|0;
               $1610 = $1609 >> 14;
               $1611 = (0 - ($1610))|0;
               $$0$i36 = $1611;
               break L794;
              } else {
               $1612 = (($1607) + 8192)|0;
               $1613 = $1612 >> 14;
               $$0$i36 = $1613;
               break L794;
              }
              break;
             }
             default: {
              $1614 = HEAP32[$1602>>2]|0;
              $$0$i36 = $1614;
              break L794;
             }
             }
            } else {
             $1598 = HEAP32[$116>>2]|0;
             $1599 = ($1598|0)==(0|0);
             if ($1599) {
              $$0$i36 = 0;
              break;
             }
             $1600 = HEAP32[$1598>>2]|0;
             $1601 = ($1600|0)==(0);
             if (!($1601)) {
              $$0$i36 = 0;
              break;
             }
             HEAP32[$1598>>2] = 130;
             $$0$i36 = 0;
            }
           } while(0);
           $1615 = (($$0$i36) + ($1573))|0;
           $alternate$1 = 1;$index34$3 = $1534;$x135$0 = $1510;$x236$0 = $1553;$x337$1 = $1593;$y138$0 = $1533;$y239$0 = $1573;$y340$1 = $1615;
          } else {
           $alternate$1 = 1;$index34$3 = $index34$0212;$x135$0 = $1510;$x236$0 = $1553;$x337$1 = $1593;$y138$0 = $1533;$y239$0 = $1573;$y340$1 = $1573;
          }
         } else {
          $1404 = HEAP32[$115>>2]|0;
          $1405 = (($1404) - ($118))|0;
          $1406 = $1405 >> 3;
          $1407 = ($1406>>>0)>($index34$0212>>>0);
          L808: do {
           if ($1407) {
            $1412 = (((($100)) + 8|0) + ($index34$0212<<3)|0);
            $1413 = (((((($100)) + 8|0) + ($index34$0212<<3)|0)) + 4|0);
            $1414 = HEAP32[$1413>>2]|0;
            switch ($1414|0) {
            case 2:  {
             $1415 = HEAP32[$1412>>2]|0;
             $1416 = $1415 << 16;
             $$0$i53 = $1416;
             break L808;
             break;
            }
            case 1:  {
             $1417 = HEAP32[$1412>>2]|0;
             $1418 = ($1417|0)<(0);
             if ($1418) {
              $1419 = (8192 - ($1417))|0;
              $1420 = $1419 >> 14;
              $1421 = (0 - ($1420))|0;
              $$0$i53 = $1421;
              break L808;
             } else {
              $1422 = (($1417) + 8192)|0;
              $1423 = $1422 >> 14;
              $$0$i53 = $1423;
              break L808;
             }
             break;
            }
            default: {
             $1424 = HEAP32[$1412>>2]|0;
             $$0$i53 = $1424;
             break L808;
            }
            }
           } else {
            $1408 = HEAP32[$116>>2]|0;
            $1409 = ($1408|0)==(0|0);
            if ($1409) {
             $$0$i53 = 0;
            } else {
             $1410 = HEAP32[$1408>>2]|0;
             $1411 = ($1410|0)==(0);
             if ($1411) {
              HEAP32[$1408>>2] = 130;
              $$0$i53 = 0;
             } else {
              $$0$i53 = 0;
             }
            }
           }
          } while(0);
          $1425 = HEAP32[$0>>2]|0;
          $1426 = (($1425) + ($$0$i53))|0;
          $1427 = HEAP32[$1>>2]|0;
          $1428 = (($index34$0212) + 1)|0;
          $1429 = ($1406>>>0)>($1428>>>0);
          L821: do {
           if ($1429) {
            $1434 = (((($100)) + 8|0) + ($1428<<3)|0);
            $1435 = (((((($100)) + 8|0) + ($1428<<3)|0)) + 4|0);
            $1436 = HEAP32[$1435>>2]|0;
            switch ($1436|0) {
            case 2:  {
             $1437 = HEAP32[$1434>>2]|0;
             $1438 = $1437 << 16;
             $$0$i51 = $1438;
             break L821;
             break;
            }
            case 1:  {
             $1439 = HEAP32[$1434>>2]|0;
             $1440 = ($1439|0)<(0);
             if ($1440) {
              $1441 = (8192 - ($1439))|0;
              $1442 = $1441 >> 14;
              $1443 = (0 - ($1442))|0;
              $$0$i51 = $1443;
              break L821;
             } else {
              $1444 = (($1439) + 8192)|0;
              $1445 = $1444 >> 14;
              $$0$i51 = $1445;
              break L821;
             }
             break;
            }
            default: {
             $1446 = HEAP32[$1434>>2]|0;
             $$0$i51 = $1446;
             break L821;
            }
            }
           } else {
            $1430 = HEAP32[$116>>2]|0;
            $1431 = ($1430|0)==(0|0);
            if ($1431) {
             $$0$i51 = 0;
            } else {
             $1432 = HEAP32[$1430>>2]|0;
             $1433 = ($1432|0)==(0);
             if ($1433) {
              HEAP32[$1430>>2] = 130;
              $$0$i51 = 0;
             } else {
              $$0$i51 = 0;
             }
            }
           }
          } while(0);
          $1447 = (($$0$i51) + ($1426))|0;
          $1448 = (($index34$0212) + 2)|0;
          $1449 = ($1406>>>0)>($1448>>>0);
          L834: do {
           if ($1449) {
            $1454 = (((($100)) + 8|0) + ($1448<<3)|0);
            $1455 = (((((($100)) + 8|0) + ($1448<<3)|0)) + 4|0);
            $1456 = HEAP32[$1455>>2]|0;
            switch ($1456|0) {
            case 2:  {
             $1457 = HEAP32[$1454>>2]|0;
             $1458 = $1457 << 16;
             $$0$i49 = $1458;
             break L834;
             break;
            }
            case 1:  {
             $1459 = HEAP32[$1454>>2]|0;
             $1460 = ($1459|0)<(0);
             if ($1460) {
              $1461 = (8192 - ($1459))|0;
              $1462 = $1461 >> 14;
              $1463 = (0 - ($1462))|0;
              $$0$i49 = $1463;
              break L834;
             } else {
              $1464 = (($1459) + 8192)|0;
              $1465 = $1464 >> 14;
              $$0$i49 = $1465;
              break L834;
             }
             break;
            }
            default: {
             $1466 = HEAP32[$1454>>2]|0;
             $$0$i49 = $1466;
             break L834;
            }
            }
           } else {
            $1450 = HEAP32[$116>>2]|0;
            $1451 = ($1450|0)==(0|0);
            if ($1451) {
             $$0$i49 = 0;
            } else {
             $1452 = HEAP32[$1450>>2]|0;
             $1453 = ($1452|0)==(0);
             if ($1453) {
              HEAP32[$1450>>2] = 130;
              $$0$i49 = 0;
             } else {
              $$0$i49 = 0;
             }
            }
           }
          } while(0);
          $1467 = (($$0$i49) + ($1427))|0;
          $1468 = (($index34$0212) + 3)|0;
          $1469 = ($1406>>>0)>($1468>>>0);
          L847: do {
           if ($1469) {
            $1474 = (((($100)) + 8|0) + ($1468<<3)|0);
            $1475 = (((((($100)) + 8|0) + ($1468<<3)|0)) + 4|0);
            $1476 = HEAP32[$1475>>2]|0;
            switch ($1476|0) {
            case 2:  {
             $1477 = HEAP32[$1474>>2]|0;
             $1478 = $1477 << 16;
             $$0$i47 = $1478;
             break L847;
             break;
            }
            case 1:  {
             $1479 = HEAP32[$1474>>2]|0;
             $1480 = ($1479|0)<(0);
             if ($1480) {
              $1481 = (8192 - ($1479))|0;
              $1482 = $1481 >> 14;
              $1483 = (0 - ($1482))|0;
              $$0$i47 = $1483;
              break L847;
             } else {
              $1484 = (($1479) + 8192)|0;
              $1485 = $1484 >> 14;
              $$0$i47 = $1485;
              break L847;
             }
             break;
            }
            default: {
             $1486 = HEAP32[$1474>>2]|0;
             $$0$i47 = $1486;
             break L847;
            }
            }
           } else {
            $1470 = HEAP32[$116>>2]|0;
            $1471 = ($1470|0)==(0|0);
            if ($1471) {
             $$0$i47 = 0;
            } else {
             $1472 = HEAP32[$1470>>2]|0;
             $1473 = ($1472|0)==(0);
             if (!($1473)) {
              $$0$i47 = 0;
              break;
             }
             HEAP32[$1470>>2] = 130;
             $$0$i47 = 0;
            }
           }
          } while(0);
          $1487 = (($$0$i47) + ($1467))|0;
          $1488 = (($1398) - ($index34$0212))|0;
          $1489 = ($1488|0)==(5);
          if ($1489) {
           $1490 = (($index34$0212) + 4)|0;
           $1491 = ($1406>>>0)>($1490>>>0);
           L861: do {
            if ($1491) {
             $1496 = (((($100)) + 8|0) + ($1490<<3)|0);
             $1497 = (((((($100)) + 8|0) + ($1490<<3)|0)) + 4|0);
             $1498 = HEAP32[$1497>>2]|0;
             switch ($1498|0) {
             case 2:  {
              $1499 = HEAP32[$1496>>2]|0;
              $1500 = $1499 << 16;
              $$0$i45 = $1500;
              break L861;
              break;
             }
             case 1:  {
              $1501 = HEAP32[$1496>>2]|0;
              $1502 = ($1501|0)<(0);
              if ($1502) {
               $1503 = (8192 - ($1501))|0;
               $1504 = $1503 >> 14;
               $1505 = (0 - ($1504))|0;
               $$0$i45 = $1505;
               break L861;
              } else {
               $1506 = (($1501) + 8192)|0;
               $1507 = $1506 >> 14;
               $$0$i45 = $1507;
               break L861;
              }
              break;
             }
             default: {
              $1508 = HEAP32[$1496>>2]|0;
              $$0$i45 = $1508;
              break L861;
             }
             }
            } else {
             $1492 = HEAP32[$116>>2]|0;
             $1493 = ($1492|0)==(0|0);
             if ($1493) {
              $$0$i45 = 0;
              break;
             }
             $1494 = HEAP32[$1492>>2]|0;
             $1495 = ($1494|0)==(0);
             if (!($1495)) {
              $$0$i45 = 0;
              break;
             }
             HEAP32[$1492>>2] = 130;
             $$0$i45 = 0;
            }
           } while(0);
           $1509 = (($$0$i45) + ($1447))|0;
           $alternate$1 = 0;$index34$3 = $1428;$x135$0 = $1426;$x236$0 = $1447;$x337$1 = $1509;$y138$0 = $1427;$y239$0 = $1467;$y340$1 = $1487;
          } else {
           $alternate$1 = 0;$index34$3 = $index34$0212;$x135$0 = $1426;$x236$0 = $1447;$x337$1 = $1447;$y138$0 = $1427;$y239$0 = $1467;$y340$1 = $1487;
          }
         }
         _cf2_glyphpath_curveTo($glyphPath,$x135$0,$y138$0,$x236$0,$y239$0,$x337$1,$y340$1);
         HEAP32[$0>>2] = $x337$1;
         HEAP32[$1>>2] = $y340$1;
         $1616 = (($index34$3) + 4)|0;
         $1617 = ($1398>>>0)>($1616>>>0);
         if ($1617) {
          $alternate$0213 = $alternate$1;$index34$0212 = $1616;
         } else {
          break;
         }
        }
       }
       HEAP32[$115>>2] = $117;
       $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
       continue L12;
       break;
      }
      case 28:  {
       $1618 = ($535>>>0)<($charstring$0$idx$val>>>0);
       if ($1618) {
        $1619 = ((($535)) + 1|0);
        HEAP32[$charstring$0$idx12>>2] = $1619;
        $1620 = HEAP8[$535>>0]|0;
        $1621 = $1620&255;
        $$0$i34 = $1621;$1628 = $1619;
       } else {
        $1622 = HEAP32[$charstring$0>>2]|0;
        $1623 = ($1622|0)==(0|0);
        if ($1623) {
         $$0$i34 = 0;$1628 = $535;
        } else {
         $1624 = HEAP32[$1622>>2]|0;
         $1625 = ($1624|0)==(0);
         if ($1625) {
          HEAP32[$1622>>2] = 85;
          $$0$i34 = 0;$1628 = $535;
         } else {
          $$0$i34 = 0;$1628 = $535;
         }
        }
       }
       $1626 = $$0$i34 << 8;
       $1627 = ($1628>>>0)<($charstring$0$idx$val>>>0);
       if ($1627) {
        $1629 = ((($1628)) + 1|0);
        HEAP32[$charstring$0$idx12>>2] = $1629;
        $1630 = HEAP8[$1628>>0]|0;
        $1631 = $1630&255;
        $$0$i32 = $1631;
       } else {
        $1632 = HEAP32[$charstring$0>>2]|0;
        $1633 = ($1632|0)==(0|0);
        if ($1633) {
         $$0$i32 = 0;
        } else {
         $1634 = HEAP32[$1632>>2]|0;
         $1635 = ($1634|0)==(0);
         if ($1635) {
          HEAP32[$1632>>2] = 85;
          $$0$i32 = 0;
         } else {
          $$0$i32 = 0;
         }
        }
       }
       $1636 = HEAP32[$115>>2]|0;
       $1637 = ($1636|0)==($115|0);
       if (!($1637)) {
        $1642 = $$0$i32 | $1626;
        $sext = $1642 << 16;
        $1643 = $sext >> 16;
        HEAP32[$1636>>2] = $1643;
        $1644 = ((($1636)) + 4|0);
        HEAP32[$1644>>2] = 2;
        $1645 = ((($1636)) + 8|0);
        HEAP32[$115>>2] = $1645;
        $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
        continue L12;
       }
       $1638 = HEAP32[$116>>2]|0;
       $1639 = ($1638|0)==(0|0);
       if ($1639) {
        $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
        continue L12;
       }
       $1640 = HEAP32[$1638>>2]|0;
       $1641 = ($1640|0)==(0);
       if (!($1641)) {
        $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
        continue L12;
       }
       HEAP32[$1638>>2] = 130;
       $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
       continue L12;
       break;
      }
      default: {
       $1646 = ($op1$0>>>0)<(247);
       if ($1646) {
        $1647 = HEAP32[$115>>2]|0;
        $1648 = ($1647|0)==($115|0);
        if (!($1648)) {
         $1653 = (($op1$0) + -139)|0;
         HEAP32[$1647>>2] = $1653;
         $1654 = ((($1647)) + 4|0);
         HEAP32[$1654>>2] = 2;
         $1655 = ((($1647)) + 8|0);
         HEAP32[$115>>2] = $1655;
         $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
         continue L12;
        }
        $1649 = HEAP32[$116>>2]|0;
        $1650 = ($1649|0)==(0|0);
        if ($1650) {
         $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
         continue L12;
        }
        $1651 = HEAP32[$1649>>2]|0;
        $1652 = ($1651|0)==(0);
        if (!($1652)) {
         $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
         continue L12;
        }
        HEAP32[$1649>>2] = 130;
        $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
        continue L12;
       }
       $1656 = ($op1$0>>>0)<(251);
       if ($1656) {
        $1657 = $op1$0 << 8;
        $1658 = (($1657) + -63232)|0;
        $1659 = ($535>>>0)<($charstring$0$idx$val>>>0);
        if ($1659) {
         $1660 = ((($535)) + 1|0);
         HEAP32[$charstring$0$idx12>>2] = $1660;
         $1661 = HEAP8[$535>>0]|0;
         $1662 = $1661&255;
         $$0$i28 = $1662;
        } else {
         $1663 = HEAP32[$charstring$0>>2]|0;
         $1664 = ($1663|0)==(0|0);
         if ($1664) {
          $$0$i28 = 0;
         } else {
          $1665 = HEAP32[$1663>>2]|0;
          $1666 = ($1665|0)==(0);
          if ($1666) {
           HEAP32[$1663>>2] = 85;
           $$0$i28 = 0;
          } else {
           $$0$i28 = 0;
          }
         }
        }
        $1667 = HEAP32[$115>>2]|0;
        $1668 = ($1667|0)==($115|0);
        if (!($1668)) {
         $1673 = $$0$i28 | $1658;
         $1674 = (($1673) + 108)|0;
         HEAP32[$1667>>2] = $1674;
         $1675 = ((($1667)) + 4|0);
         HEAP32[$1675>>2] = 2;
         $1676 = ((($1667)) + 8|0);
         HEAP32[$115>>2] = $1676;
         $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
         continue L12;
        }
        $1669 = HEAP32[$116>>2]|0;
        $1670 = ($1669|0)==(0|0);
        if ($1670) {
         $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
         continue L12;
        }
        $1671 = HEAP32[$1669>>2]|0;
        $1672 = ($1671|0)==(0);
        if (!($1672)) {
         $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
         continue L12;
        }
        HEAP32[$1669>>2] = 130;
        $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
        continue L12;
       }
       $1677 = ($op1$0|0)==(255);
       if (!($1677)) {
        $1678 = $op1$0 << 8;
        $1679 = (($1678) + -64256)|0;
        $1680 = ($535>>>0)<($charstring$0$idx$val>>>0);
        if ($1680) {
         $1681 = ((($535)) + 1|0);
         HEAP32[$charstring$0$idx12>>2] = $1681;
         $1682 = HEAP8[$535>>0]|0;
         $1683 = $1682&255;
         $$0$i25 = $1683;
        } else {
         $1684 = HEAP32[$charstring$0>>2]|0;
         $1685 = ($1684|0)==(0|0);
         if ($1685) {
          $$0$i25 = 0;
         } else {
          $1686 = HEAP32[$1684>>2]|0;
          $1687 = ($1686|0)==(0);
          if ($1687) {
           HEAP32[$1684>>2] = 85;
           $$0$i25 = 0;
          } else {
           $$0$i25 = 0;
          }
         }
        }
        $1688 = HEAP32[$115>>2]|0;
        $1689 = ($1688|0)==($115|0);
        if (!($1689)) {
         $1694 = $$0$i25 | $1679;
         $1695 = (-108 - ($1694))|0;
         HEAP32[$1688>>2] = $1695;
         $1696 = ((($1688)) + 4|0);
         HEAP32[$1696>>2] = 2;
         $1697 = ((($1688)) + 8|0);
         HEAP32[$115>>2] = $1697;
         $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
         continue L12;
        }
        $1690 = HEAP32[$116>>2]|0;
        $1691 = ($1690|0)==(0|0);
        if ($1691) {
         $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
         continue L12;
        }
        $1692 = HEAP32[$1690>>2]|0;
        $1693 = ($1692|0)==(0);
        if (!($1693)) {
         $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
         continue L12;
        }
        HEAP32[$1690>>2] = 130;
        $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
        continue L12;
       }
       $1698 = ($535>>>0)<($charstring$0$idx$val>>>0);
       if ($1698) {
        $1699 = ((($535)) + 1|0);
        HEAP32[$charstring$0$idx12>>2] = $1699;
        $1700 = HEAP8[$535>>0]|0;
        $1701 = $1700&255;
        $$0$i23 = $1701;$1708 = $1699;
       } else {
        $1702 = HEAP32[$charstring$0>>2]|0;
        $1703 = ($1702|0)==(0|0);
        if ($1703) {
         $$0$i23 = 0;$1708 = $535;
        } else {
         $1704 = HEAP32[$1702>>2]|0;
         $1705 = ($1704|0)==(0);
         if ($1705) {
          HEAP32[$1702>>2] = 85;
          $$0$i23 = 0;$1708 = $535;
         } else {
          $$0$i23 = 0;$1708 = $535;
         }
        }
       }
       $1706 = $$0$i23 << 24;
       $1707 = ($1708>>>0)<($charstring$0$idx$val>>>0);
       if ($1707) {
        $1709 = ((($1708)) + 1|0);
        HEAP32[$charstring$0$idx12>>2] = $1709;
        $1710 = HEAP8[$1708>>0]|0;
        $1711 = $1710&255;
        $$0$i21 = $1711;$1719 = $1709;
       } else {
        $1712 = HEAP32[$charstring$0>>2]|0;
        $1713 = ($1712|0)==(0|0);
        if ($1713) {
         $$0$i21 = 0;$1719 = $1708;
        } else {
         $1714 = HEAP32[$1712>>2]|0;
         $1715 = ($1714|0)==(0);
         if ($1715) {
          HEAP32[$1712>>2] = 85;
          $$0$i21 = 0;$1719 = $1708;
         } else {
          $$0$i21 = 0;$1719 = $1708;
         }
        }
       }
       $1716 = $$0$i21 << 16;
       $1717 = $1716 | $1706;
       $1718 = ($1719>>>0)<($charstring$0$idx$val>>>0);
       if ($1718) {
        $1720 = ((($1719)) + 1|0);
        HEAP32[$charstring$0$idx12>>2] = $1720;
        $1721 = HEAP8[$1719>>0]|0;
        $1722 = $1721&255;
        $$0$i19 = $1722;$1730 = $1720;
       } else {
        $1723 = HEAP32[$charstring$0>>2]|0;
        $1724 = ($1723|0)==(0|0);
        if ($1724) {
         $$0$i19 = 0;$1730 = $1719;
        } else {
         $1725 = HEAP32[$1723>>2]|0;
         $1726 = ($1725|0)==(0);
         if ($1726) {
          HEAP32[$1723>>2] = 85;
          $$0$i19 = 0;$1730 = $1719;
         } else {
          $$0$i19 = 0;$1730 = $1719;
         }
        }
       }
       $1727 = $$0$i19 << 8;
       $1728 = $1717 | $1727;
       $1729 = ($1730>>>0)<($charstring$0$idx$val>>>0);
       if ($1729) {
        $1731 = ((($1730)) + 1|0);
        HEAP32[$charstring$0$idx12>>2] = $1731;
        $1732 = HEAP8[$1730>>0]|0;
        $1733 = $1732&255;
        $$0$i = $1733;
       } else {
        $1734 = HEAP32[$charstring$0>>2]|0;
        $1735 = ($1734|0)==(0|0);
        if ($1735) {
         $$0$i = 0;
        } else {
         $1736 = HEAP32[$1734>>2]|0;
         $1737 = ($1736|0)==(0);
         if ($1737) {
          HEAP32[$1734>>2] = 85;
          $$0$i = 0;
         } else {
          $$0$i = 0;
         }
        }
       }
       $1738 = HEAP32[$115>>2]|0;
       $1739 = ($1738|0)==($115|0);
       if (!($1739)) {
        $1744 = $1728 | $$0$i;
        HEAP32[$1738>>2] = $1744;
        $1745 = ((($1738)) + 4|0);
        HEAP32[$1745>>2] = 0;
        $1746 = ((($1738)) + 8|0);
        HEAP32[$115>>2] = $1746;
        $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
        continue L12;
       }
       $1740 = HEAP32[$116>>2]|0;
       $1741 = ($1740|0)==(0|0);
       if ($1741) {
        $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
        continue L12;
       }
       $1742 = HEAP32[$1740>>2]|0;
       $1743 = ($1742|0)==(0);
       if (!($1743)) {
        $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
        continue L12;
       }
       HEAP32[$1740>>2] = 130;
       $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
       continue L12;
      }
      }
     } while(0);
     HEAP32[$115>>2] = $117;
     $charstringIndex$0$phi = $charstringIndex$0;$charstring$0$phi = $charstring$0;$instructionLimit$0 = $149;$charstringIndex$0 = $charstringIndex$0$phi;$charstring$0 = $charstring$0$phi;
    }
    $543 = HEAP32[$115>>2]|0;
    $544 = (($543) - ($118))|0;
    $545 = $544 >> 3;
    $546 = ($545|0)==(1);
    if ($546) {
     $$old = HEAP8[$haveWidth>>0]|0;
     $$old6 = ($$old<<24>>24)==(0);
     if ($$old6) {
      label = 212;
     }
    } else {
     $547 = ($545|0)!=(5);
     $548 = HEAP8[$haveWidth>>0]|0;
     $549 = ($548<<24>>24)!=(0);
     $or$cond7 = $547 | $549;
     if (!($or$cond7)) {
      label = 212;
     }
    }
    if ((label|0) == 212) {
     $550 = ($543|0)==($118|0);
     L901: do {
      if ($550) {
       $551 = HEAP32[$116>>2]|0;
       $552 = ($551|0)==(0|0);
       if ($552) {
        $$0$i195 = 0;
       } else {
        $553 = HEAP32[$551>>2]|0;
        $554 = ($553|0)==(0);
        if ($554) {
         HEAP32[$551>>2] = 130;
         $$0$i195 = 0;
        } else {
         $$0$i195 = 0;
        }
       }
      } else {
       $555 = HEAP32[$124>>2]|0;
       switch ($555|0) {
       case 2:  {
        $556 = HEAP32[$117>>2]|0;
        $557 = $556 << 16;
        $$0$i195 = $557;
        break L901;
        break;
       }
       case 1:  {
        $558 = HEAP32[$117>>2]|0;
        $559 = ($558|0)<(0);
        if ($559) {
         $560 = (8192 - ($558))|0;
         $561 = $560 >> 14;
         $562 = (0 - ($561))|0;
         $$0$i195 = $562;
         break L901;
        } else {
         $563 = (($558) + 8192)|0;
         $564 = $563 >> 14;
         $$0$i195 = $564;
         break L901;
        }
        break;
       }
       default: {
        $565 = HEAP32[$117>>2]|0;
        $$0$i195 = $565;
        break L901;
       }
       }
      }
     } while(0);
     $566 = (($$0$i195) + ($8))|0;
     HEAP32[$width>>2] = $566;
    }
    HEAP8[$haveWidth>>0] = 1;
    $567 = HEAP32[$2>>2]|0;
    $568 = ((($567)) + 561|0);
    $569 = HEAP8[$568>>0]|0;
    $570 = ($569<<24>>24)==(0);
    if ($570) {
     $571 = HEAP8[$96>>0]|0;
     $572 = ($571<<24>>24)==(0);
     if ($572) {
      $$in = $543;
     } else {
      HEAP8[$97>>0] = 1;
      $573 = HEAP32[$119>>2]|0;
      $574 = HEAP32[$121>>2]|0;
      _cf2_glyphpath_lineTo($glyphPath,$573,$574);
      $575 = HEAP8[$98>>0]|0;
      $576 = ($575<<24>>24)==(0);
      if (!($576)) {
       $$0$val$i190 = HEAP32[$$0$i124>>2]|0;
       $$1$val$i192 = HEAP32[$$1$i126>>2]|0;
       _cf2_glyphpath_pushPrevElem($glyphPath,$47,$123,$$0$val$i190,$$1$val$i192,1);
      }
      HEAP8[$95>>0] = 1;
      HEAP8[$96>>0] = 0;
      HEAP8[$97>>0] = 0;
      HEAP8[$98>>0] = 0;
      $$pre253 = HEAP32[$115>>2]|0;
      $$in = $$pre253;
     }
     $577 = $$in;
     $578 = (($$in) - ($118))|0;
     $579 = $578 >> 3;
     $580 = ($579>>>0)>(1);
     if ($580) {
      $581 = ($doingSeac<<24>>24)==(0);
      L923: do {
       if ($581) {
        $582 = ($577|0)==($117|0);
        do {
         if ($582) {
          $583 = HEAP32[$116>>2]|0;
          $584 = ($583|0)==(0|0);
          if ($584) {
           $$0$i187 = 0;$597 = $577;
          } else {
           $585 = HEAP32[$583>>2]|0;
           $586 = ($585|0)==(0);
           if ($586) {
            HEAP32[$583>>2] = 161;
            $$0$i187 = 0;$597 = $577;
           } else {
            $$0$i187 = 0;$597 = $577;
           }
          }
         } else {
          $587 = ((($577)) + -8|0);
          $588 = ((($577)) + -4|0);
          $589 = HEAP32[$588>>2]|0;
          $590 = ($589|0)==(2);
          if ($590) {
           HEAP32[$115>>2] = $587;
           $595 = HEAP32[$587>>2]|0;
           $$0$i187 = $595;$597 = $587;
           break;
          }
          $591 = HEAP32[$116>>2]|0;
          $592 = ($591|0)==(0|0);
          if ($592) {
           $$0$i187 = 0;$597 = $577;
          } else {
           $593 = HEAP32[$591>>2]|0;
           $594 = ($593|0)==(0);
           if ($594) {
            HEAP32[$591>>2] = 160;
            $$0$i187 = 0;$597 = $577;
           } else {
            $$0$i187 = 0;$597 = $577;
           }
          }
         }
        } while(0);
        $596 = ($597|0)==($117|0);
        do {
         if ($596) {
          $598 = HEAP32[$116>>2]|0;
          $599 = ($598|0)==(0|0);
          if ($599) {
           $$0$i186 = 0;$612 = $597;
          } else {
           $600 = HEAP32[$598>>2]|0;
           $601 = ($600|0)==(0);
           if ($601) {
            HEAP32[$598>>2] = 161;
            $$0$i186 = 0;$612 = $597;
           } else {
            $$0$i186 = 0;$612 = $597;
           }
          }
         } else {
          $602 = ((($597)) + -8|0);
          $603 = ((($597)) + -4|0);
          $604 = HEAP32[$603>>2]|0;
          $605 = ($604|0)==(2);
          if ($605) {
           HEAP32[$115>>2] = $602;
           $610 = HEAP32[$602>>2]|0;
           $$0$i186 = $610;$612 = $602;
           break;
          }
          $606 = HEAP32[$116>>2]|0;
          $607 = ($606|0)==(0|0);
          if ($607) {
           $$0$i186 = 0;$612 = $597;
          } else {
           $608 = HEAP32[$606>>2]|0;
           $609 = ($608|0)==(0);
           if ($609) {
            HEAP32[$606>>2] = 160;
            $$0$i186 = 0;$612 = $597;
           } else {
            $$0$i186 = 0;$612 = $597;
           }
          }
         }
        } while(0);
        $611 = ($612|0)==($117|0);
        L947: do {
         if ($611) {
          $613 = HEAP32[$116>>2]|0;
          $614 = ($613|0)==(0|0);
          if ($614) {
           $$0$i184 = 0;$631 = $612;
          } else {
           $615 = HEAP32[$613>>2]|0;
           $616 = ($615|0)==(0);
           if ($616) {
            HEAP32[$613>>2] = 161;
            $$0$i184 = 0;$631 = $612;
           } else {
            $$0$i184 = 0;$631 = $612;
           }
          }
         } else {
          $617 = ((($612)) + -8|0);
          HEAP32[$115>>2] = $617;
          $618 = ((($612)) + -4|0);
          $619 = HEAP32[$618>>2]|0;
          switch ($619|0) {
          case 2:  {
           $620 = HEAP32[$617>>2]|0;
           $621 = $620 << 16;
           $$0$i184 = $621;$631 = $617;
           break L947;
           break;
          }
          case 1:  {
           $622 = HEAP32[$617>>2]|0;
           $623 = ($622|0)<(0);
           if ($623) {
            $624 = (8192 - ($622))|0;
            $625 = $624 >> 14;
            $626 = (0 - ($625))|0;
            $$0$i184 = $626;$631 = $617;
            break L947;
           } else {
            $627 = (($622) + 8192)|0;
            $628 = $627 >> 14;
            $$0$i184 = $628;$631 = $617;
            break L947;
           }
           break;
          }
          default: {
           $629 = HEAP32[$617>>2]|0;
           $$0$i184 = $629;$631 = $617;
           break L947;
          }
          }
         }
        } while(0);
        HEAP32[$1>>2] = $$0$i184;
        $630 = ($631|0)==($117|0);
        L960: do {
         if ($630) {
          $632 = HEAP32[$116>>2]|0;
          $633 = ($632|0)==(0|0);
          if ($633) {
           $$0$i182 = 0;
          } else {
           $634 = HEAP32[$632>>2]|0;
           $635 = ($634|0)==(0);
           if ($635) {
            HEAP32[$632>>2] = 161;
            $$0$i182 = 0;
           } else {
            $$0$i182 = 0;
           }
          }
         } else {
          $636 = ((($631)) + -8|0);
          HEAP32[$115>>2] = $636;
          $637 = ((($631)) + -4|0);
          $638 = HEAP32[$637>>2]|0;
          switch ($638|0) {
          case 2:  {
           $639 = HEAP32[$636>>2]|0;
           $640 = $639 << 16;
           $$0$i182 = $640;
           break L960;
           break;
          }
          case 1:  {
           $641 = HEAP32[$636>>2]|0;
           $642 = ($641|0)<(0);
           if ($642) {
            $643 = (8192 - ($641))|0;
            $644 = $643 >> 14;
            $645 = (0 - ($644))|0;
            $$0$i182 = $645;
            break L960;
           } else {
            $646 = (($641) + 8192)|0;
            $647 = $646 >> 14;
            $$0$i182 = $647;
            break L960;
           }
           break;
          }
          default: {
           $648 = HEAP32[$636>>2]|0;
           $$0$i182 = $648;
           break L960;
          }
          }
         }
        } while(0);
        HEAP32[$0>>2] = $$0$i182;
        ;HEAP32[$component>>2]=0|0;HEAP32[$component+4>>2]=0|0;HEAP32[$component+8>>2]=0|0;HEAP32[$component+12>>2]=0|0;
        $649 = ((($3)) + 4|0);
        $650 = HEAP32[$649>>2]|0;
        $651 = ((($650)) + 128|0);
        $652 = HEAP32[$651>>2]|0;
        $653 = ((($652)) + 48|0);
        $654 = HEAP32[$653>>2]|0;
        $655 = ($654|0)==(0|0);
        L973: do {
         if ($655) {
          $656 = ((($3)) + 76|0);
          $657 = HEAP32[$656>>2]|0;
          $658 = ((($657)) + 1160|0);
          $659 = HEAP32[$658>>2]|0;
          $660 = ($659|0)==(0|0);
          $661 = ($$0$i187>>>0)>(255);
          $or$cond$i$i166 = $661 | $660;
          if ($or$cond$i$i166) {
           $$0$i173$ph = 18;
          } else {
           $662 = ($$0$i187>>>0)<(256);
           if ($662) {
            $663 = (170824 + ($$0$i187<<1)|0);
            $664 = HEAP16[$663>>1]|0;
            $671 = $664;
           } else {
            $671 = 0;
           }
           $665 = ((($657)) + 12|0);
           $666 = HEAP32[$665>>2]|0;
           $667 = ($666|0)==(0);
           if ($667) {
            $$0$i173$ph = 18;
           } else {
            $n$02$i$i168 = 0;
            while(1) {
             $668 = (($659) + ($n$02$i$i168<<1)|0);
             $669 = HEAP16[$668>>1]|0;
             $670 = ($669<<16>>16)==($671<<16>>16);
             if ($670) {
              $n$02$i$i168$lcssa = $n$02$i$i168;
              break;
             }
             $672 = (($n$02$i$i168) + 1)|0;
             $673 = ($672>>>0)<($666>>>0);
             if ($673) {
              $n$02$i$i168 = $672;
             } else {
              $$0$i173$ph = 18;
              break L973;
             }
            }
            $674 = ($n$02$i$i168$lcssa|0)<(0);
            if ($674) {
             $$0$i173$ph = 18;
            } else {
             if (!($655)) {
              $gid$0$i171261 = $n$02$i$i168$lcssa;
              label = 277;
              break;
             }
             $683 = ((($650)) + 656|0);
             $684 = HEAP32[$683>>2]|0;
             $685 = ((($684)) + 1176|0);
             $686 = (_cff_index_access_element($685,$n$02$i$i168$lcssa,$charstring$i164,$len$i165)|0);
             $$0$i176 = $686;
             label = 279;
            }
           }
          }
         } else {
          $gid$0$i171261 = $$0$i187;
          label = 277;
         }
        } while(0);
        if ((label|0) == 277) {
         $675 = HEAP32[$654>>2]|0;
         $676 = HEAP32[$675>>2]|0;
         $677 = ((($654)) + 4|0);
         $678 = HEAP32[$677>>2]|0;
         $679 = (FUNCTION_TABLE_iiii[$676 & 8191]($678,$gid$0$i171261,$data$i175)|0);
         $680 = HEAP32[$data$i175>>2]|0;
         HEAP32[$charstring$i164>>2] = $680;
         $681 = ((($data$i175)) + 4|0);
         $682 = HEAP32[$681>>2]|0;
         HEAP32[$len$i165>>2] = $682;
         $$0$i176 = $679;
         label = 279;
        }
        if ((label|0) == 279) {
         $687 = ($$0$i176|0)==(0);
         if ($687) {
          $688 = HEAP32[$charstring$i164>>2]|0;
          $689 = ((($component)) + 4|0);
          HEAP32[$689>>2] = $688;
          $$cast$i172 = $688;
          $690 = HEAP32[$len$i165>>2]|0;
          $691 = (($$cast$i172) + ($690)|0);
          $692 = ((($component)) + 8|0);
          HEAP32[$692>>2] = $691;
          $693 = ((($component)) + 12|0);
          HEAP32[$693>>2] = $688;
          $694 = HEAP32[$0>>2]|0;
          $695 = HEAP32[$1>>2]|0;
          _cf2_interpT2CharString($font,$component,$callbacks,$translation,1,$694,$695,$dummyWidth);
          $$idx18$val = HEAP32[$649>>2]|0;
          $696 = ((($$idx18$val)) + 128|0);
          $697 = HEAP32[$696>>2]|0;
          $698 = ((($697)) + 48|0);
          $699 = HEAP32[$698>>2]|0;
          $700 = ($699|0)==(0|0);
          if ($700) {
           $709 = ((($$idx18$val)) + 656|0);
           $710 = HEAP32[$709>>2]|0;
           $711 = ((($710)) + 1204|0);
           $712 = HEAP32[$711>>2]|0;
           $713 = ($712|0)==(0|0);
           if ($713) {
            $714 = ((($710)) + 1176|0);
            $715 = HEAP32[$714>>2]|0;
            _FT_Stream_ReleaseFrame($715,$689);
           }
          } else {
           $701 = $691;
           $702 = (($701) - ($688))|0;
           HEAP32[$data$i$i156>>2] = $688;
           $703 = ((($data$i$i156)) + 4|0);
           HEAP32[$703>>2] = $702;
           $704 = HEAP32[$699>>2]|0;
           $705 = ((($704)) + 4|0);
           $706 = HEAP32[$705>>2]|0;
           $707 = ((($699)) + 4|0);
           $708 = HEAP32[$707>>2]|0;
           FUNCTION_TABLE_vii[$706 & 8191]($708,$data$i$i156);
          }
          ;HEAP32[$component>>2]=0|0;HEAP32[$component+4>>2]=0|0;HEAP32[$component+8>>2]=0|0;HEAP32[$component+12>>2]=0|0;
          $716 = HEAP32[$649>>2]|0;
          $717 = ((($716)) + 128|0);
          $718 = HEAP32[$717>>2]|0;
          $719 = ((($718)) + 48|0);
          $720 = HEAP32[$719>>2]|0;
          $721 = ($720|0)==(0|0);
          L995: do {
           if ($721) {
            $722 = ((($3)) + 76|0);
            $723 = HEAP32[$722>>2]|0;
            $724 = ((($723)) + 1160|0);
            $725 = HEAP32[$724>>2]|0;
            $726 = ($725|0)==(0|0);
            $727 = ($$0$i186>>>0)>(255);
            $or$cond$i$i = $727 | $726;
            if ($or$cond$i$i) {
             $$0$i150$ph = 18;
             break;
            }
            $728 = ($$0$i186>>>0)<(256);
            if ($728) {
             $729 = (170824 + ($$0$i186<<1)|0);
             $730 = HEAP16[$729>>1]|0;
             $737 = $730;
            } else {
             $737 = 0;
            }
            $731 = ((($723)) + 12|0);
            $732 = HEAP32[$731>>2]|0;
            $733 = ($732|0)==(0);
            if ($733) {
             $$0$i150$ph = 18;
             break;
            } else {
             $n$02$i$i = 0;
            }
            while(1) {
             $734 = (($725) + ($n$02$i$i<<1)|0);
             $735 = HEAP16[$734>>1]|0;
             $736 = ($735<<16>>16)==($737<<16>>16);
             if ($736) {
              $n$02$i$i$lcssa = $n$02$i$i;
              break;
             }
             $738 = (($n$02$i$i) + 1)|0;
             $739 = ($738>>>0)<($732>>>0);
             if ($739) {
              $n$02$i$i = $738;
             } else {
              $$0$i150$ph = 18;
              break L995;
             }
            }
            $740 = ($n$02$i$i$lcssa|0)<(0);
            if ($740) {
             $$0$i150$ph = 18;
             break;
            }
            if (!($721)) {
             $gid$0$i263 = $n$02$i$i$lcssa;
             label = 294;
             break;
            }
            $749 = ((($716)) + 656|0);
            $750 = HEAP32[$749>>2]|0;
            $751 = ((($750)) + 1176|0);
            $752 = (_cff_index_access_element($751,$n$02$i$i$lcssa,$charstring$i,$len$i)|0);
            $$0$i151 = $752;
            label = 296;
           } else {
            $gid$0$i263 = $$0$i186;
            label = 294;
           }
          } while(0);
          if ((label|0) == 294) {
           $741 = HEAP32[$720>>2]|0;
           $742 = HEAP32[$741>>2]|0;
           $743 = ((($720)) + 4|0);
           $744 = HEAP32[$743>>2]|0;
           $745 = (FUNCTION_TABLE_iiii[$742 & 8191]($744,$gid$0$i263,$data$i)|0);
           $746 = HEAP32[$data$i>>2]|0;
           HEAP32[$charstring$i>>2] = $746;
           $747 = ((($data$i)) + 4|0);
           $748 = HEAP32[$747>>2]|0;
           HEAP32[$len$i>>2] = $748;
           $$0$i151 = $745;
           label = 296;
          }
          do {
           if ((label|0) == 296) {
            $753 = ($$0$i151|0)==(0);
            if (!($753)) {
             $$0$i150$ph = $$0$i151;
             break;
            }
            $754 = HEAP32[$charstring$i>>2]|0;
            HEAP32[$689>>2] = $754;
            $$cast$i = $754;
            $755 = HEAP32[$len$i>>2]|0;
            $756 = (($$cast$i) + ($755)|0);
            HEAP32[$692>>2] = $756;
            HEAP32[$693>>2] = $754;
            _cf2_interpT2CharString($font,$component,$callbacks,$translation,1,0,0,$dummyWidth);
            $$idx17$val = HEAP32[$649>>2]|0;
            $757 = ((($$idx17$val)) + 128|0);
            $758 = HEAP32[$757>>2]|0;
            $759 = ((($758)) + 48|0);
            $760 = HEAP32[$759>>2]|0;
            $761 = ($760|0)==(0|0);
            if (!($761)) {
             $762 = HEAP32[$692>>2]|0;
             $763 = HEAP32[$689>>2]|0;
             $764 = (($762) - ($763))|0;
             HEAP32[$data$i$i>>2] = $763;
             $765 = ((($data$i$i)) + 4|0);
             HEAP32[$765>>2] = $764;
             $766 = HEAP32[$760>>2]|0;
             $767 = ((($766)) + 4|0);
             $768 = HEAP32[$767>>2]|0;
             $769 = ((($760)) + 4|0);
             $770 = HEAP32[$769>>2]|0;
             FUNCTION_TABLE_vii[$768 & 8191]($770,$data$i$i);
             $lastError$2 = 0;
             break L923;
            }
            $771 = ((($$idx17$val)) + 656|0);
            $772 = HEAP32[$771>>2]|0;
            $773 = ((($772)) + 1204|0);
            $774 = HEAP32[$773>>2]|0;
            $775 = ($774|0)==(0|0);
            if (!($775)) {
             $lastError$2 = 0;
             break L923;
            }
            $776 = ((($772)) + 1176|0);
            $777 = HEAP32[$776>>2]|0;
            _FT_Stream_ReleaseFrame($777,$689);
            $lastError$2 = 0;
            break L923;
           }
          } while(0);
          $lastError$2 = $$0$i150$ph;
          break;
         } else {
          $$0$i173$ph = $$0$i176;
         }
        }
        $lastError$2 = $$0$i173$ph;
       } else {
        $lastError$2 = 18;
       }
      } while(0);
      $$ph = $1756;$lastError$3$ph = $lastError$2;$subrStack$sroa$31$3$ph = $108;
      label = 779;
     } else {
      $$ph = $1756;$lastError$3$ph = 0;$subrStack$sroa$31$3$ph = $108;
      label = 779;
     }
    } else {
     $$ph = $1756;$lastError$3$ph = 0;$subrStack$sroa$31$3$ph = $108;
     label = 779;
    }
   } else {
    $1757 = $1756;$subrStack$sroa$31$3209 = $108;
   }
  }
 } while(0);
 if ((label|0) == 779) {
  $$pr = HEAP32[$4>>2]|0;
  $1747 = ($$pr|0)==(0);
  if ($1747) {
   HEAP32[$4>>2] = $lastError$3$ph;
   $1757 = $$ph;$subrStack$sroa$31$3209 = $subrStack$sroa$31$3$ph;
  } else {
   $1757 = $$ph;$subrStack$sroa$31$3209 = $subrStack$sroa$31$3$ph;
  }
 }
 $1748 = ((($glyphPath)) + 11612|0);
 $1749 = HEAP32[$1748>>2]|0;
 HEAP32[$29>>2] = 0;
 HEAP32[$31>>2] = 0;
 HEAP32[$32>>2] = 0;
 $1750 = HEAP32[$33>>2]|0;
 _ft_mem_free($1749,$1750);
 HEAP32[$33>>2] = 0;
 $1751 = HEAP32[$vStemHintArray>>2]|0;
 HEAP32[$18>>2] = 0;
 HEAP32[$20>>2] = 0;
 HEAP32[$21>>2] = 0;
 $1752 = HEAP32[$22>>2]|0;
 _ft_mem_free($1751,$1752);
 HEAP32[$22>>2] = 0;
 $1753 = HEAP32[$hStemHintArray>>2]|0;
 HEAP32[$11>>2] = 0;
 HEAP32[$13>>2] = 0;
 HEAP32[$14>>2] = 0;
 $1754 = HEAP32[$15>>2]|0;
 _ft_mem_free($1753,$1754);
 HEAP32[$15>>2] = 0;
 _ft_mem_free($5,$subrStack$sroa$31$3209);
 if ($1757) {
  STACKTOP = sp;return;
 }
 $1755 = HEAP32[$100>>2]|0;
 _ft_mem_free($1755,$100);
 STACKTOP = sp;return;
}
function _cf2_doStems($font,$opStack,$stemHintArray,$width,$haveWidth) {
 $font = $font|0;
 $opStack = $opStack|0;
 $stemHintArray = $stemHintArray|0;
 $width = $width|0;
 $haveWidth = $haveWidth|0;
 var $$0$i = 0, $$0$i1 = 0, $$0$i3 = 0, $$idx = 0, $$idx$val = 0, $$idx$val$idx = 0, $$idx$val$idx$val = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $i$05 = 0, $position$06 = 0, $stemhint = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $stemhint = sp;
 $0 = ((($opStack)) + 392|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($opStack)) + 8|0);
 $3 = $2;
 $4 = (($1) - ($3))|0;
 $5 = $4 >> 3;
 $6 = $5 & 1;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = HEAP8[$haveWidth>>0]|0;
  $9 = ($8<<24>>24)==(0);
  if ($9) {
   $10 = ($5|0)==(0);
   L4: do {
    if ($10) {
     $11 = ((($opStack)) + 4|0);
     $12 = HEAP32[$11>>2]|0;
     $13 = ($12|0)==(0|0);
     if ($13) {
      $$0$i = 0;
     } else {
      $14 = HEAP32[$12>>2]|0;
      $15 = ($14|0)==(0);
      if ($15) {
       HEAP32[$12>>2] = 130;
       $$0$i = 0;
      } else {
       $$0$i = 0;
      }
     }
    } else {
     $16 = ((($opStack)) + 12|0);
     $17 = HEAP32[$16>>2]|0;
     switch ($17|0) {
     case 2:  {
      $18 = HEAP32[$2>>2]|0;
      $19 = $18 << 16;
      $$0$i = $19;
      break L4;
      break;
     }
     case 1:  {
      $20 = HEAP32[$2>>2]|0;
      $21 = ($20|0)<(0);
      if ($21) {
       $22 = (8192 - ($20))|0;
       $23 = $22 >> 14;
       $24 = (0 - ($23))|0;
       $$0$i = $24;
       break L4;
      } else {
       $25 = (($20) + 8192)|0;
       $26 = $25 >> 14;
       $$0$i = $26;
       break L4;
      }
      break;
     }
     default: {
      $27 = HEAP32[$2>>2]|0;
      $$0$i = $27;
      break L4;
     }
     }
    }
   } while(0);
   $28 = ((($font)) + 132|0);
   $29 = HEAP32[$28>>2]|0;
   $$idx = ((($29)) + 736|0);
   $$idx$val = HEAP32[$$idx>>2]|0;
   $$idx$val$idx = ((($$idx$val)) + 532|0);
   $$idx$val$idx$val = HEAP32[$$idx$val$idx>>2]|0;
   $30 = $$idx$val$idx$val << 16;
   $31 = (($30) + ($$0$i))|0;
   HEAP32[$width>>2] = $31;
  }
 }
 $32 = ((($font)) + 132|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($33)) + 561|0);
 $35 = HEAP8[$34>>0]|0;
 $36 = ($35<<24>>24)==(0);
 if (!($36)) {
  HEAP8[$haveWidth>>0] = 1;
  STACKTOP = sp;return;
 }
 $37 = ($6>>>0)<($5>>>0);
 L21: do {
  if ($37) {
   $38 = ((($stemhint)) + 4|0);
   $39 = ((($stemhint)) + 8|0);
   $40 = ((($stemhint)) + 12|0);
   $41 = ((($stemhint)) + 16|0);
   $42 = ((($opStack)) + 4|0);
   $44 = $1;$i$05 = $6;$position$06 = 0;
   while(1) {
    $43 = (($44) - ($3))|0;
    $45 = $43 >> 3;
    $46 = ($45>>>0)>($i$05>>>0);
    L25: do {
     if ($46) {
      $51 = (((($opStack)) + 8|0) + ($i$05<<3)|0);
      $52 = (((((($opStack)) + 8|0) + ($i$05<<3)|0)) + 4|0);
      $53 = HEAP32[$52>>2]|0;
      switch ($53|0) {
      case 2:  {
       $54 = HEAP32[$51>>2]|0;
       $55 = $54 << 16;
       $$0$i3 = $55;
       break L25;
       break;
      }
      case 1:  {
       $56 = HEAP32[$51>>2]|0;
       $57 = ($56|0)<(0);
       if ($57) {
        $58 = (8192 - ($56))|0;
        $59 = $58 >> 14;
        $60 = (0 - ($59))|0;
        $$0$i3 = $60;
        break L25;
       } else {
        $61 = (($56) + 8192)|0;
        $62 = $61 >> 14;
        $$0$i3 = $62;
        break L25;
       }
       break;
      }
      default: {
       $63 = HEAP32[$51>>2]|0;
       $$0$i3 = $63;
       break L25;
      }
      }
     } else {
      $47 = HEAP32[$42>>2]|0;
      $48 = ($47|0)==(0|0);
      if ($48) {
       $$0$i3 = 0;
      } else {
       $49 = HEAP32[$47>>2]|0;
       $50 = ($49|0)==(0);
       if ($50) {
        HEAP32[$47>>2] = 130;
        $$0$i3 = 0;
       } else {
        $$0$i3 = 0;
       }
      }
     }
    } while(0);
    $64 = (($$0$i3) + ($position$06))|0;
    HEAP32[$38>>2] = $64;
    $65 = (($i$05) + 1)|0;
    $66 = ($45>>>0)>($65>>>0);
    L38: do {
     if ($66) {
      $71 = (((($opStack)) + 8|0) + ($65<<3)|0);
      $72 = (((((($opStack)) + 8|0) + ($65<<3)|0)) + 4|0);
      $73 = HEAP32[$72>>2]|0;
      switch ($73|0) {
      case 2:  {
       $74 = HEAP32[$71>>2]|0;
       $75 = $74 << 16;
       $$0$i1 = $75;
       break L38;
       break;
      }
      case 1:  {
       $76 = HEAP32[$71>>2]|0;
       $77 = ($76|0)<(0);
       if ($77) {
        $78 = (8192 - ($76))|0;
        $79 = $78 >> 14;
        $80 = (0 - ($79))|0;
        $$0$i1 = $80;
        break L38;
       } else {
        $81 = (($76) + 8192)|0;
        $82 = $81 >> 14;
        $$0$i1 = $82;
        break L38;
       }
       break;
      }
      default: {
       $83 = HEAP32[$71>>2]|0;
       $$0$i1 = $83;
       break L38;
      }
      }
     } else {
      $67 = HEAP32[$42>>2]|0;
      $68 = ($67|0)==(0|0);
      if ($68) {
       $$0$i1 = 0;
      } else {
       $69 = HEAP32[$67>>2]|0;
       $70 = ($69|0)==(0);
       if ($70) {
        HEAP32[$67>>2] = 130;
        $$0$i1 = 0;
       } else {
        $$0$i1 = 0;
       }
      }
     }
    } while(0);
    $84 = (($$0$i1) + ($64))|0;
    HEAP32[$39>>2] = $84;
    HEAP8[$stemhint>>0] = 0;
    HEAP32[$40>>2] = 0;
    HEAP32[$41>>2] = 0;
    _cf2_arrstack_push($stemHintArray,$stemhint);
    $85 = (($i$05) + 2)|0;
    $86 = ($85>>>0)<($5>>>0);
    if (!($86)) {
     break L21;
    }
    $$pre = HEAP32[$0>>2]|0;
    $44 = $$pre;$i$05 = $85;$position$06 = $84;
   }
  }
 } while(0);
 HEAP32[$0>>2] = $2;
 HEAP8[$haveWidth>>0] = 1;
 STACKTOP = sp;return;
}
function _cf2_arrstack_push($arrstack,$ptr) {
 $arrstack = $arrstack|0;
 $ptr = $ptr|0;
 var $$phi$trans$insert = 0, $$pre = 0, $$pre$phiZ2D = 0, $$pre2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp;
 $0 = ((($arrstack)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($arrstack)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==($3|0);
 do {
  if ($4) {
   $5 = ((($arrstack)) + 16|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = (($6) + ($1))|0;
   HEAP32[$error$i>>2] = 0;
   $8 = ((($arrstack)) + 8|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = Math_imul($9, $7)|0;
   $11 = (2147483647 / ($9>>>0))&-1;
   $12 = ($11>>>0)<($7>>>0);
   if (!($12)) {
    $13 = HEAP32[$arrstack>>2]|0;
    $14 = ((($arrstack)) + 24|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = ((($arrstack)) + 28|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = (_ft_mem_realloc($13,1,$15,$10,$17,$error$i)|0);
    HEAP32[$16>>2] = $18;
    $19 = HEAP32[$error$i>>2]|0;
    $20 = ($19|0)==(0);
    if ($20) {
     HEAP32[$2>>2] = $7;
     HEAP32[$14>>2] = $10;
     $21 = HEAP32[$0>>2]|0;
     $22 = ($21>>>0)>($7>>>0);
     if (!($22)) {
      $$pre$phiZ2D = $8;$35 = $21;$37 = $18;
      break;
     }
     $23 = ((($arrstack)) + 4|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ($24|0)==(0|0);
     if (!($25)) {
      $26 = HEAP32[$24>>2]|0;
      $27 = ($26|0)==(0);
      if ($27) {
       HEAP32[$24>>2] = 130;
      }
     }
     HEAP32[$0>>2] = $7;
     STACKTOP = sp;return;
    }
   }
   $28 = ((($arrstack)) + 4|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = ($29|0)==(0|0);
   if ($30) {
    STACKTOP = sp;return;
   }
   $31 = HEAP32[$29>>2]|0;
   $32 = ($31|0)==(0);
   if (!($32)) {
    STACKTOP = sp;return;
   }
   HEAP32[$29>>2] = 64;
   STACKTOP = sp;return;
  } else {
   $$phi$trans$insert = ((($arrstack)) + 28|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0;
   $$pre2 = ((($arrstack)) + 8|0);
   $$pre$phiZ2D = $$pre2;$35 = $1;$37 = $$pre;
  }
 } while(0);
 $33 = HEAP32[$$pre$phiZ2D>>2]|0;
 $34 = Math_imul($33, $35)|0;
 $36 = (($37) + ($34)|0);
 _memcpy(($36|0),($ptr|0),($33|0))|0;
 $38 = HEAP32[$0>>2]|0;
 $39 = (($38) + 1)|0;
 HEAP32[$0>>2] = $39;
 STACKTOP = sp;return;
}
function _cf2_glyphpath_lineTo($glyphpath,$x,$y) {
 $glyphpath = $glyphpath|0;
 $x = $x|0;
 $y = $y|0;
 var $$idx = 0, $$idx$val = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $7 = 0, $8 = 0, $9 = 0, $P0 = 0, $or$cond = 0, $phitmp = 0, $xOffset = 0, $yOffset = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $xOffset = sp + 12|0;
 $yOffset = sp + 8|0;
 $P0 = sp;
 $0 = ((($glyphpath)) + 11676|0);
 $1 = HEAP32[$0>>2]|0;
 $$idx = ((($1)) + 5|0);
 $$idx$val = HEAP8[$$idx>>0]|0;
 $2 = ($$idx$val<<24>>24)==(0);
 if ($2) {
  $13 = 0;
 } else {
  $3 = ((($glyphpath)) + 11665|0);
  $4 = HEAP8[$3>>0]|0;
  $5 = ($4<<24>>24)==(0);
  $phitmp = $5&1;
  $13 = $phitmp;
 }
 $6 = ((($glyphpath)) + 11720|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==($x|0);
 $9 = ((($glyphpath)) + 11724|0);
 $10 = HEAP32[$9>>2]|0;
 if ($8) {
  $11 = ($10|0)!=($y|0);
  $12 = ($13<<24>>24)!=(0);
  $or$cond = $12 | $11;
  if (!($or$cond)) {
   STACKTOP = sp;return;
  }
 }
 _cf2_glyphpath_computeOffset($glyphpath,$7,$10,$x,$y,$xOffset,$yOffset);
 $14 = HEAP32[$6>>2]|0;
 $15 = HEAP32[$xOffset>>2]|0;
 $16 = (($15) + ($14))|0;
 HEAP32[$P0>>2] = $16;
 $17 = HEAP32[$9>>2]|0;
 $18 = HEAP32[$yOffset>>2]|0;
 $19 = (($18) + ($17))|0;
 $20 = ((($P0)) + 4|0);
 HEAP32[$20>>2] = $19;
 $21 = (($15) + ($x))|0;
 $22 = (($18) + ($y))|0;
 $23 = ((($glyphpath)) + 11667|0);
 $24 = HEAP8[$23>>0]|0;
 $25 = ($24<<24>>24)==(0);
 if (!($25)) {
  _cf2_glyphpath_pushMove($glyphpath,$16,$19);
  HEAP8[$23>>0] = 0;
  $26 = ((($glyphpath)) + 11664|0);
  HEAP8[$26>>0] = 1;
  $27 = ((($glyphpath)) + 11712|0);
  $28 = $27;
  $29 = $28;
  HEAP32[$29>>2] = $21;
  $30 = (($28) + 4)|0;
  $31 = $30;
  HEAP32[$31>>2] = $22;
 }
 $32 = ((($glyphpath)) + 11744|0);
 $33 = HEAP8[$32>>0]|0;
 $34 = ($33<<24>>24)==(0);
 if (!($34)) {
  $35 = ((($glyphpath)) + 8|0);
  _cf2_glyphpath_pushPrevElem($glyphpath,$35,$P0,$21,$22,0);
 }
 HEAP8[$32>>0] = 1;
 $36 = ((($glyphpath)) + 11748|0);
 HEAP32[$36>>2] = 2;
 $37 = ((($glyphpath)) + 11752|0);
 $38 = $P0;
 $39 = $38;
 $40 = HEAP32[$39>>2]|0;
 $41 = (($38) + 4)|0;
 $42 = $41;
 $43 = HEAP32[$42>>2]|0;
 $44 = $37;
 $45 = $44;
 HEAP32[$45>>2] = $40;
 $46 = (($44) + 4)|0;
 $47 = $46;
 HEAP32[$47>>2] = $43;
 $48 = ((($glyphpath)) + 11760|0);
 $49 = $48;
 $50 = $49;
 HEAP32[$50>>2] = $21;
 $51 = (($49) + 4)|0;
 $52 = $51;
 HEAP32[$52>>2] = $22;
 $53 = ($13<<24>>24)==(0);
 if (!($53)) {
  $54 = ((($glyphpath)) + 8|0);
  $55 = ((($glyphpath)) + 11668|0);
  $56 = HEAP32[$55>>2]|0;
  $57 = ((($glyphpath)) + 11672|0);
  $58 = HEAP32[$57>>2]|0;
  $59 = HEAP32[$0>>2]|0;
  $60 = ((($glyphpath)) + 11680|0);
  $61 = HEAP32[$60>>2]|0;
  _cf2_hintmap_build($54,$56,$58,$59,$61,0);
 }
 HEAP32[$6>>2] = $x;
 HEAP32[$9>>2] = $y;
 STACKTOP = sp;return;
}
function _cf2_glyphpath_computeOffset($glyphpath,$x1,$y1,$x2,$y2,$x,$y) {
 $glyphpath = $glyphpath|0;
 $x1 = $x1|0;
 $y1 = $y1|0;
 $x2 = $x2|0;
 $y2 = $y2|0;
 $x = $x|0;
 $y = $y|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $9 = 0, $dx$0 = 0, $dy$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($x2) - ($x1))|0;
 $1 = (($y2) - ($y1))|0;
 $2 = HEAP32[$glyphpath>>2]|0;
 $3 = ((($2)) + 192|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)==(0);
 $6 = (0 - ($0))|0;
 $7 = (0 - ($1))|0;
 $dx$0 = $5 ? $0 : $6;
 $dy$0 = $5 ? $1 : $7;
 HEAP32[$y>>2] = 0;
 HEAP32[$x>>2] = 0;
 $8 = ((($glyphpath)) + 11666|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = ($9<<24>>24)==(0);
 if ($10) {
  return;
 }
 $11 = $x1 >> 16;
 $12 = $1 >> 16;
 $13 = Math_imul($12, $11)|0;
 $14 = $y1 >> 16;
 $15 = $0 >> 16;
 $16 = Math_imul($15, $14)|0;
 $17 = (($13) - ($16))|0;
 $18 = ((($glyphpath)) + 4|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ((($19)) + 16|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = (($17) + ($21))|0;
 HEAP32[$20>>2] = $22;
 $23 = ($dx$0|0)>(-1);
 $24 = ($dy$0|0)>(-1);
 if ($23) {
  if ($24) {
   $25 = $dy$0 << 1;
   $26 = ($dx$0|0)>($25|0);
   if ($26) {
    HEAP32[$x>>2] = 0;
    HEAP32[$y>>2] = 0;
    return;
   }
   $27 = $dx$0 << 1;
   $28 = ($dy$0|0)>($27|0);
   $29 = ((($glyphpath)) + 11688|0);
   $30 = HEAP32[$29>>2]|0;
   if ($28) {
    HEAP32[$x>>2] = $30;
    $31 = ((($glyphpath)) + 11692|0);
    $32 = HEAP32[$31>>2]|0;
    HEAP32[$y>>2] = $32;
    return;
   } else {
    $33 = (_FT_MulFix(45875,$30)|0);
    HEAP32[$x>>2] = $33;
    $34 = ((($glyphpath)) + 11692|0);
    $35 = HEAP32[$34>>2]|0;
    $36 = (_FT_MulFix(19661,$35)|0);
    HEAP32[$y>>2] = $36;
    return;
   }
  } else {
   $37 = Math_imul($dy$0, -2)|0;
   $38 = ($dx$0|0)>($37|0);
   if ($38) {
    HEAP32[$x>>2] = 0;
    HEAP32[$y>>2] = 0;
    return;
   }
   $39 = (0 - ($dy$0))|0;
   $40 = $dx$0 << 1;
   $41 = ($40|0)<($39|0);
   $42 = ((($glyphpath)) + 11688|0);
   $43 = HEAP32[$42>>2]|0;
   if ($41) {
    $44 = (0 - ($43))|0;
    HEAP32[$x>>2] = $44;
    $45 = ((($glyphpath)) + 11692|0);
    $46 = HEAP32[$45>>2]|0;
    HEAP32[$y>>2] = $46;
    return;
   } else {
    $47 = (_FT_MulFix(-45874,$43)|0);
    HEAP32[$x>>2] = $47;
    $48 = ((($glyphpath)) + 11692|0);
    $49 = HEAP32[$48>>2]|0;
    $50 = (_FT_MulFix(19661,$49)|0);
    HEAP32[$y>>2] = $50;
    return;
   }
  }
 } else {
  $51 = (0 - ($dx$0))|0;
  if ($24) {
   $52 = $dy$0 << 1;
   $53 = ($52|0)<($51|0);
   if ($53) {
    HEAP32[$x>>2] = 0;
    $54 = ((($glyphpath)) + 11692|0);
    $55 = HEAP32[$54>>2]|0;
    $56 = $55 << 1;
    HEAP32[$y>>2] = $56;
    return;
   }
   $57 = Math_imul($dx$0, -2)|0;
   $58 = ($dy$0|0)>($57|0);
   $59 = ((($glyphpath)) + 11688|0);
   $60 = HEAP32[$59>>2]|0;
   if ($58) {
    HEAP32[$x>>2] = $60;
    $61 = ((($glyphpath)) + 11692|0);
    $62 = HEAP32[$61>>2]|0;
    HEAP32[$y>>2] = $62;
    return;
   } else {
    $63 = (_FT_MulFix(45875,$60)|0);
    HEAP32[$x>>2] = $63;
    $64 = ((($glyphpath)) + 11692|0);
    $65 = HEAP32[$64>>2]|0;
    $66 = (_FT_MulFix(111411,$65)|0);
    HEAP32[$y>>2] = $66;
    return;
   }
  } else {
   $67 = Math_imul($dy$0, -2)|0;
   $68 = ($67|0)<($51|0);
   if ($68) {
    HEAP32[$x>>2] = 0;
    $69 = ((($glyphpath)) + 11692|0);
    $70 = HEAP32[$69>>2]|0;
    $71 = $70 << 1;
    HEAP32[$y>>2] = $71;
    return;
   }
   $72 = (0 - ($dy$0))|0;
   $73 = Math_imul($dx$0, -2)|0;
   $74 = ($73|0)<($72|0);
   $75 = ((($glyphpath)) + 11688|0);
   $76 = HEAP32[$75>>2]|0;
   if ($74) {
    $77 = (0 - ($76))|0;
    HEAP32[$x>>2] = $77;
    $78 = ((($glyphpath)) + 11692|0);
    $79 = HEAP32[$78>>2]|0;
    HEAP32[$y>>2] = $79;
    return;
   } else {
    $80 = (_FT_MulFix(-45874,$76)|0);
    HEAP32[$x>>2] = $80;
    $81 = ((($glyphpath)) + 11692|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = (_FT_MulFix(111411,$82)|0);
    HEAP32[$y>>2] = $83;
    return;
   }
  }
 }
}
function _cf2_glyphpath_pushMove($glyphpath,$start$0,$start$1) {
 $glyphpath = $glyphpath|0;
 $start$0 = $start$0|0;
 $start$1 = $start$1|0;
 var $$0$i = 0, $$0$val$i = 0, $$1$i = 0, $$1$val$i = 0, $$idx = 0, $$idx$i = 0, $$idx$val = 0, $$idx$val$i = 0, $$idx1$val$i = 0, $$idx1$val$i$pre = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phi3Z2D = 0, $$pre2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $9 = 0, $params = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $params = sp;
 $0 = ((($params)) + 32|0);
 HEAP32[$0>>2] = 1;
 $1 = ((($glyphpath)) + 11728|0);
 $2 = $1;
 $3 = $2;
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2) + 4)|0;
 $6 = $5;
 $7 = HEAP32[$6>>2]|0;
 $8 = $params;
 $9 = $8;
 HEAP32[$9>>2] = $4;
 $10 = (($8) + 4)|0;
 $11 = $10;
 HEAP32[$11>>2] = $7;
 $12 = ((($glyphpath)) + 8|0);
 $$idx = ((($glyphpath)) + 20|0);
 $$idx$val = HEAP8[$$idx>>0]|0;
 $13 = ($$idx$val<<24>>24)==(0);
 if ($13) {
  $14 = ((($glyphpath)) + 11736|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ((($glyphpath)) + 11740|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = ((($glyphpath)) + 11664|0);
  $19 = HEAP8[$18>>0]|0;
  $20 = ($19<<24>>24)==(0);
  if ($20) {
   $$pre2 = ((($glyphpath)) + 11667|0);
   $$idx1$val$i = 1;$$pre$phi3Z2D = $$pre2;
  } else {
   $21 = ((($glyphpath)) + 11665|0);
   HEAP8[$21>>0] = 1;
   _cf2_glyphpath_lineTo($glyphpath,$15,$17);
   $22 = ((($glyphpath)) + 11744|0);
   $23 = HEAP8[$22>>0]|0;
   $24 = ($23<<24>>24)==(0);
   if (!($24)) {
    $25 = ((($glyphpath)) + 11704|0);
    $$0$i = ((($glyphpath)) + 11712|0);
    $$0$val$i = HEAP32[$$0$i>>2]|0;
    $$1$i = ((($glyphpath)) + 11716|0);
    $$1$val$i = HEAP32[$$1$i>>2]|0;
    _cf2_glyphpath_pushPrevElem($glyphpath,$12,$25,$$0$val$i,$$1$val$i,1);
   }
   $26 = ((($glyphpath)) + 11667|0);
   HEAP8[$26>>0] = 1;
   HEAP8[$18>>0] = 0;
   HEAP8[$21>>0] = 0;
   HEAP8[$22>>0] = 0;
   $$idx1$val$i$pre = HEAP8[$$idx>>0]|0;
   $phitmp = ($$idx1$val$i$pre<<24>>24)==(0);
   $$idx1$val$i = $phitmp;$$pre$phi3Z2D = $26;
  }
  HEAP32[$14>>2] = $15;
  $27 = ((($glyphpath)) + 11720|0);
  HEAP32[$27>>2] = $15;
  HEAP32[$16>>2] = $17;
  $28 = ((($glyphpath)) + 11724|0);
  HEAP32[$28>>2] = $17;
  HEAP8[$$pre$phi3Z2D>>0] = 1;
  $$phi$trans$insert = ((($glyphpath)) + 11676|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  if ($$idx1$val$i) {
   label = 9;
  } else {
   $$idx$i = ((($$pre)) + 5|0);
   $$idx$val$i = HEAP8[$$idx$i>>0]|0;
   $29 = ($$idx$val$i<<24>>24)==(0);
   if (!($29)) {
    label = 9;
   }
  }
  if ((label|0) == 9) {
   $30 = ((($glyphpath)) + 11668|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = ((($glyphpath)) + 11672|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ((($glyphpath)) + 11680|0);
   $35 = HEAP32[$34>>2]|0;
   _cf2_hintmap_build($12,$31,$33,$$pre,$35,0);
  }
  $36 = ((($glyphpath)) + 3876|0);
  _memcpy(($36|0),($12|0),3868)|0;
 }
 $37 = ((($params)) + 8|0);
 $38 = ((($glyphpath)) + 11644|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = (_FT_MulFix($39,$start$0)|0);
 $41 = ((($glyphpath)) + 11648|0);
 $42 = HEAP32[$41>>2]|0;
 $43 = (_FT_MulFix($42,$start$1)|0);
 $44 = (($43) + ($40))|0;
 $45 = (_cf2_hintmap_map($12,$start$1)|0);
 $46 = HEAP32[$glyphpath>>2]|0;
 $47 = ((($46)) + 60|0);
 $48 = HEAP32[$47>>2]|0;
 $49 = (_FT_MulFix($48,$44)|0);
 $50 = HEAP32[$glyphpath>>2]|0;
 $51 = ((($50)) + 68|0);
 $52 = HEAP32[$51>>2]|0;
 $53 = (_FT_MulFix($52,$45)|0);
 $54 = (($53) + ($49))|0;
 $55 = ((($glyphpath)) + 11656|0);
 $56 = HEAP32[$55>>2]|0;
 $57 = (($54) + ($56))|0;
 HEAP32[$37>>2] = $57;
 $58 = HEAP32[$glyphpath>>2]|0;
 $59 = ((($58)) + 64|0);
 $60 = HEAP32[$59>>2]|0;
 $61 = (_FT_MulFix($60,$44)|0);
 $62 = HEAP32[$glyphpath>>2]|0;
 $63 = ((($62)) + 72|0);
 $64 = HEAP32[$63>>2]|0;
 $65 = (_FT_MulFix($64,$45)|0);
 $66 = (($65) + ($61))|0;
 $67 = ((($glyphpath)) + 11660|0);
 $68 = HEAP32[$67>>2]|0;
 $69 = (($66) + ($68))|0;
 $70 = ((($params)) + 12|0);
 HEAP32[$70>>2] = $69;
 $71 = ((($glyphpath)) + 4|0);
 $72 = HEAP32[$71>>2]|0;
 $73 = HEAP32[$72>>2]|0;
 FUNCTION_TABLE_vii[$73 & 8191]($72,$params);
 $74 = $37;
 $75 = $74;
 $76 = HEAP32[$75>>2]|0;
 $77 = (($74) + 4)|0;
 $78 = $77;
 $79 = HEAP32[$78>>2]|0;
 $80 = $1;
 $81 = $80;
 HEAP32[$81>>2] = $76;
 $82 = (($80) + 4)|0;
 $83 = $82;
 HEAP32[$83>>2] = $79;
 $84 = ((($glyphpath)) + 11704|0);
 $85 = $84;
 $86 = $85;
 HEAP32[$86>>2] = $start$0;
 $87 = (($85) + 4)|0;
 $88 = $87;
 HEAP32[$88>>2] = $start$1;
 STACKTOP = sp;return;
}
function _cf2_glyphpath_pushPrevElem($glyphpath,$hintmap,$nextP0,$nextP1$0,$nextP1$1,$close) {
 $glyphpath = $glyphpath|0;
 $hintmap = $hintmap|0;
 $nextP0 = $nextP0|0;
 $nextP1$0 = $nextP1$0|0;
 $nextP1$1 = $nextP1$1|0;
 $close = $close|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$intersection$sroa$0$0 = 0, $$intersection$sroa$11$0 = 0, $$sink7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0;
 var $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0;
 var $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0;
 var $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0;
 var $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0;
 var $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0;
 var $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0;
 var $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0;
 var $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0;
 var $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0;
 var $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0;
 var $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0;
 var $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $intersection$sroa$0$0 = 0, $intersection$sroa$0$1 = 0, $intersection$sroa$0$4 = 0, $intersection$sroa$11$0 = 0, $intersection$sroa$11$1 = 0, $intersection$sroa$11$4 = 0, $not$$i = 0;
 var $or$cond = 0, $or$cond12 = 0, $or$cond13 = 0, $params = 0, $prevP0$0 = 0, $prevP1$0 = 0, $useIntersection$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $params = sp;
 $0 = ((($glyphpath)) + 11748|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(2);
 $3 = ((($glyphpath)) + 11752|0);
 $4 = ((($glyphpath)) + 11760|0);
 $5 = ((($glyphpath)) + 11768|0);
 $6 = ((($glyphpath)) + 11776|0);
 $prevP0$0 = $2 ? $3 : $5;
 $prevP1$0 = $2 ? $4 : $6;
 $7 = HEAP32[$prevP1$0>>2]|0;
 $8 = HEAP32[$nextP0>>2]|0;
 $9 = ($7|0)==($8|0);
 $10 = ((($prevP1$0)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($nextP0)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($11|0)==($13|0);
 $or$cond12 = $9 & $14;
 if ($or$cond12) {
  $intersection$sroa$0$4 = 0;$intersection$sroa$11$4 = 0;$useIntersection$0 = 0;
 } else {
  $15 = HEAP32[$prevP0$0>>2]|0;
  $16 = (($7) + 16)|0;
  $17 = (($16) - ($15))|0;
  $18 = $17 >> 5;
  $19 = ((($prevP0$0)) + 4|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = (($11) + 16)|0;
  $22 = (($21) - ($20))|0;
  $23 = $22 >> 5;
  $24 = (($nextP1$0) + 16)|0;
  $25 = (($24) - ($8))|0;
  $26 = $25 >> 5;
  $27 = (($nextP1$1) + 16)|0;
  $28 = (($27) - ($13))|0;
  $29 = $28 >> 5;
  $30 = (_FT_MulFix($18,$29)|0);
  $31 = (_FT_MulFix($23,$26)|0);
  $32 = ($30|0)==($31|0);
  if ($32) {
   $intersection$sroa$0$4 = 0;$intersection$sroa$11$4 = 0;$useIntersection$0 = 0;
  } else {
   $33 = (($30) - ($31))|0;
   $34 = (16 - ($20))|0;
   $35 = (($34) + ($13))|0;
   $36 = $35 >> 5;
   $37 = (16 - ($15))|0;
   $38 = (($37) + ($8))|0;
   $39 = $38 >> 5;
   $40 = (_FT_MulFix($39,$29)|0);
   $41 = (_FT_MulFix($36,$26)|0);
   $42 = (($40) - ($41))|0;
   $43 = (_FT_DivFix($42,$33)|0);
   $44 = HEAP32[$prevP0$0>>2]|0;
   $45 = HEAP32[$prevP1$0>>2]|0;
   $46 = (($45) - ($44))|0;
   $47 = (_FT_MulFix($43,$46)|0);
   $48 = (($47) + ($44))|0;
   $49 = HEAP32[$19>>2]|0;
   $50 = HEAP32[$10>>2]|0;
   $51 = (($50) - ($49))|0;
   $52 = (_FT_MulFix($43,$51)|0);
   $53 = (($52) + ($49))|0;
   $54 = HEAP32[$prevP0$0>>2]|0;
   $55 = HEAP32[$prevP1$0>>2]|0;
   $56 = ($54|0)==($55|0);
   if ($56) {
    $57 = (($48) - ($54))|0;
    $58 = ($57|0)<(0);
    $59 = (0 - ($57))|0;
    $60 = $58 ? $59 : $57;
    $61 = ((($glyphpath)) + 11700|0);
    $62 = HEAP32[$61>>2]|0;
    $63 = ($60|0)<($62|0);
    $$ = $63 ? $54 : $48;
    $intersection$sroa$0$0 = $$;
   } else {
    $intersection$sroa$0$0 = $48;
   }
   $64 = HEAP32[$19>>2]|0;
   $65 = HEAP32[$10>>2]|0;
   $66 = ($64|0)==($65|0);
   if ($66) {
    $67 = (($53) - ($64))|0;
    $68 = ($67|0)<(0);
    $69 = (0 - ($67))|0;
    $70 = $68 ? $69 : $67;
    $71 = ((($glyphpath)) + 11700|0);
    $72 = HEAP32[$71>>2]|0;
    $73 = ($70|0)<($72|0);
    $$1 = $73 ? $64 : $53;
    $intersection$sroa$11$0 = $$1;
   } else {
    $intersection$sroa$11$0 = $53;
   }
   $74 = HEAP32[$nextP0>>2]|0;
   $75 = ($74|0)==($nextP1$0|0);
   if ($75) {
    $76 = (($intersection$sroa$0$0) - ($nextP1$0))|0;
    $77 = ($76|0)<(0);
    $78 = (0 - ($76))|0;
    $79 = $77 ? $78 : $76;
    $80 = ((($glyphpath)) + 11700|0);
    $81 = HEAP32[$80>>2]|0;
    $82 = ($79|0)<($81|0);
    $$intersection$sroa$0$0 = $82 ? $nextP1$0 : $intersection$sroa$0$0;
    $intersection$sroa$0$1 = $$intersection$sroa$0$0;
   } else {
    $intersection$sroa$0$1 = $intersection$sroa$0$0;
   }
   $83 = HEAP32[$12>>2]|0;
   $84 = ($83|0)==($nextP1$1|0);
   if ($84) {
    $85 = (($intersection$sroa$11$0) - ($nextP1$1))|0;
    $86 = ($85|0)<(0);
    $87 = (0 - ($85))|0;
    $88 = $86 ? $87 : $85;
    $89 = ((($glyphpath)) + 11700|0);
    $90 = HEAP32[$89>>2]|0;
    $91 = ($88|0)<($90|0);
    $$intersection$sroa$11$0 = $91 ? $nextP1$1 : $intersection$sroa$11$0;
    $$3 = $91 ? $nextP1$1 : $intersection$sroa$11$0;
    $104 = $$3;$intersection$sroa$11$1 = $$intersection$sroa$11$0;
   } else {
    $104 = $intersection$sroa$11$0;$intersection$sroa$11$1 = $intersection$sroa$11$0;
   }
   $92 = (($74) + ($55))|0;
   $93 = (($92|0) / 2)&-1;
   $94 = (($intersection$sroa$0$1) - ($93))|0;
   $95 = ($94|0)<(0);
   $96 = (0 - ($94))|0;
   $97 = $95 ? $96 : $94;
   $98 = ((($glyphpath)) + 11696|0);
   $99 = HEAP32[$98>>2]|0;
   $100 = ($97|0)>($99|0);
   if ($100) {
    $intersection$sroa$0$4 = $intersection$sroa$0$1;$intersection$sroa$11$4 = $intersection$sroa$11$1;$useIntersection$0 = 0;
   } else {
    $101 = (($83) + ($65))|0;
    $102 = (($101|0) / 2)&-1;
    $103 = (($104) - ($102))|0;
    $105 = ($103|0)<(0);
    $106 = (0 - ($103))|0;
    $107 = $105 ? $106 : $103;
    $not$$i = ($107|0)>($99|0);
    if ($not$$i) {
     $intersection$sroa$0$4 = $intersection$sroa$0$1;$intersection$sroa$11$4 = $intersection$sroa$11$1;$useIntersection$0 = 0;
    } else {
     $108 = $prevP1$0;
     $109 = $108;
     HEAP32[$109>>2] = $intersection$sroa$0$1;
     $110 = (($108) + 4)|0;
     $111 = $110;
     HEAP32[$111>>2] = $intersection$sroa$11$1;
     $intersection$sroa$0$4 = $intersection$sroa$0$1;$intersection$sroa$11$4 = $intersection$sroa$11$1;$useIntersection$0 = 1;
    }
   }
  }
 }
 $112 = ((($glyphpath)) + 11728|0);
 $113 = $112;
 $114 = $113;
 $115 = HEAP32[$114>>2]|0;
 $116 = (($113) + 4)|0;
 $117 = $116;
 $118 = HEAP32[$117>>2]|0;
 $119 = $params;
 $120 = $119;
 HEAP32[$120>>2] = $115;
 $121 = (($119) + 4)|0;
 $122 = $121;
 HEAP32[$122>>2] = $118;
 $123 = HEAP32[$0>>2]|0;
 switch ($123|0) {
 case 2:  {
  $124 = ((($params)) + 32|0);
  HEAP32[$124>>2] = 2;
  $125 = ($close<<24>>24)==(0);
  $126 = ((($glyphpath)) + 11760|0);
  $127 = HEAP32[$126>>2]|0;
  $128 = ((($glyphpath)) + 11764|0);
  $129 = HEAP32[$128>>2]|0;
  $130 = ((($glyphpath)) + 11644|0);
  $131 = HEAP32[$130>>2]|0;
  if ($125) {
   $163 = (_FT_MulFix($131,$127)|0);
   $164 = ((($glyphpath)) + 11648|0);
   $165 = HEAP32[$164>>2]|0;
   $166 = (_FT_MulFix($165,$129)|0);
   $167 = (($166) + ($163))|0;
   $168 = (_cf2_hintmap_map($hintmap,$129)|0);
   $169 = HEAP32[$glyphpath>>2]|0;
   $170 = ((($169)) + 60|0);
   $171 = HEAP32[$170>>2]|0;
   $172 = (_FT_MulFix($171,$167)|0);
   $173 = HEAP32[$glyphpath>>2]|0;
   $174 = ((($173)) + 68|0);
   $175 = HEAP32[$174>>2]|0;
   $176 = (_FT_MulFix($175,$168)|0);
   $177 = (($176) + ($172))|0;
   $178 = ((($glyphpath)) + 11656|0);
   $179 = HEAP32[$178>>2]|0;
   $180 = (($177) + ($179))|0;
   $181 = HEAP32[$glyphpath>>2]|0;
   $182 = ((($181)) + 64|0);
   $183 = HEAP32[$182>>2]|0;
   $184 = (_FT_MulFix($183,$167)|0);
   $185 = HEAP32[$glyphpath>>2]|0;
   $186 = ((($185)) + 72|0);
   $187 = HEAP32[$186>>2]|0;
   $188 = (_FT_MulFix($187,$168)|0);
   $189 = (($188) + ($184))|0;
   $190 = ((($glyphpath)) + 11660|0);
   $191 = HEAP32[$190>>2]|0;
   $192 = (($189) + ($191))|0;
   $$sink7 = $180;$195 = $192;
  } else {
   $132 = ((($glyphpath)) + 3876|0);
   $133 = (_FT_MulFix($131,$127)|0);
   $134 = ((($glyphpath)) + 11648|0);
   $135 = HEAP32[$134>>2]|0;
   $136 = (_FT_MulFix($135,$129)|0);
   $137 = (($136) + ($133))|0;
   $138 = (_cf2_hintmap_map($132,$129)|0);
   $139 = HEAP32[$glyphpath>>2]|0;
   $140 = ((($139)) + 60|0);
   $141 = HEAP32[$140>>2]|0;
   $142 = (_FT_MulFix($141,$137)|0);
   $143 = HEAP32[$glyphpath>>2]|0;
   $144 = ((($143)) + 68|0);
   $145 = HEAP32[$144>>2]|0;
   $146 = (_FT_MulFix($145,$138)|0);
   $147 = (($146) + ($142))|0;
   $148 = ((($glyphpath)) + 11656|0);
   $149 = HEAP32[$148>>2]|0;
   $150 = (($147) + ($149))|0;
   $151 = HEAP32[$glyphpath>>2]|0;
   $152 = ((($151)) + 64|0);
   $153 = HEAP32[$152>>2]|0;
   $154 = (_FT_MulFix($153,$137)|0);
   $155 = HEAP32[$glyphpath>>2]|0;
   $156 = ((($155)) + 72|0);
   $157 = HEAP32[$156>>2]|0;
   $158 = (_FT_MulFix($157,$138)|0);
   $159 = (($158) + ($154))|0;
   $160 = ((($glyphpath)) + 11660|0);
   $161 = HEAP32[$160>>2]|0;
   $162 = (($159) + ($161))|0;
   $$sink7 = $150;$195 = $162;
  }
  $193 = ((($params)) + 8|0);
  HEAP32[$193>>2] = $$sink7;
  $194 = ((($params)) + 12|0);
  HEAP32[$194>>2] = $195;
  $196 = ($115|0)==($$sink7|0);
  $197 = ($118|0)==($195|0);
  $or$cond13 = $196 & $197;
  if (!($or$cond13)) {
   $198 = ((($params)) + 8|0);
   $199 = ((($glyphpath)) + 4|0);
   $200 = HEAP32[$199>>2]|0;
   $201 = ((($200)) + 4|0);
   $202 = HEAP32[$201>>2]|0;
   FUNCTION_TABLE_vii[$202 & 8191]($200,$params);
   $203 = $198;
   $204 = $203;
   $205 = HEAP32[$204>>2]|0;
   $206 = (($203) + 4)|0;
   $207 = $206;
   $208 = HEAP32[$207>>2]|0;
   $209 = $112;
   $210 = $209;
   HEAP32[$210>>2] = $205;
   $211 = (($209) + 4)|0;
   $212 = $211;
   HEAP32[$212>>2] = $208;
  }
  break;
 }
 case 4:  {
  $213 = ((($params)) + 32|0);
  HEAP32[$213>>2] = 4;
  $214 = ((($glyphpath)) + 11760|0);
  $215 = HEAP32[$214>>2]|0;
  $216 = ((($glyphpath)) + 11764|0);
  $217 = HEAP32[$216>>2]|0;
  $218 = ((($glyphpath)) + 11644|0);
  $219 = HEAP32[$218>>2]|0;
  $220 = (_FT_MulFix($219,$215)|0);
  $221 = ((($glyphpath)) + 11648|0);
  $222 = HEAP32[$221>>2]|0;
  $223 = (_FT_MulFix($222,$217)|0);
  $224 = (($223) + ($220))|0;
  $225 = (_cf2_hintmap_map($hintmap,$217)|0);
  $226 = HEAP32[$glyphpath>>2]|0;
  $227 = ((($226)) + 60|0);
  $228 = HEAP32[$227>>2]|0;
  $229 = (_FT_MulFix($228,$224)|0);
  $230 = HEAP32[$glyphpath>>2]|0;
  $231 = ((($230)) + 68|0);
  $232 = HEAP32[$231>>2]|0;
  $233 = (_FT_MulFix($232,$225)|0);
  $234 = (($233) + ($229))|0;
  $235 = ((($glyphpath)) + 11656|0);
  $236 = HEAP32[$235>>2]|0;
  $237 = (($234) + ($236))|0;
  $238 = ((($params)) + 8|0);
  HEAP32[$238>>2] = $237;
  $239 = HEAP32[$glyphpath>>2]|0;
  $240 = ((($239)) + 64|0);
  $241 = HEAP32[$240>>2]|0;
  $242 = (_FT_MulFix($241,$224)|0);
  $243 = HEAP32[$glyphpath>>2]|0;
  $244 = ((($243)) + 72|0);
  $245 = HEAP32[$244>>2]|0;
  $246 = (_FT_MulFix($245,$225)|0);
  $247 = (($246) + ($242))|0;
  $248 = ((($glyphpath)) + 11660|0);
  $249 = HEAP32[$248>>2]|0;
  $250 = (($247) + ($249))|0;
  $251 = ((($params)) + 12|0);
  HEAP32[$251>>2] = $250;
  $252 = ((($glyphpath)) + 11768|0);
  $253 = HEAP32[$252>>2]|0;
  $254 = ((($glyphpath)) + 11772|0);
  $255 = HEAP32[$254>>2]|0;
  $256 = HEAP32[$218>>2]|0;
  $257 = (_FT_MulFix($256,$253)|0);
  $258 = HEAP32[$221>>2]|0;
  $259 = (_FT_MulFix($258,$255)|0);
  $260 = (($259) + ($257))|0;
  $261 = (_cf2_hintmap_map($hintmap,$255)|0);
  $262 = HEAP32[$glyphpath>>2]|0;
  $263 = ((($262)) + 60|0);
  $264 = HEAP32[$263>>2]|0;
  $265 = (_FT_MulFix($264,$260)|0);
  $266 = HEAP32[$glyphpath>>2]|0;
  $267 = ((($266)) + 68|0);
  $268 = HEAP32[$267>>2]|0;
  $269 = (_FT_MulFix($268,$261)|0);
  $270 = (($269) + ($265))|0;
  $271 = HEAP32[$235>>2]|0;
  $272 = (($270) + ($271))|0;
  $273 = ((($params)) + 16|0);
  HEAP32[$273>>2] = $272;
  $274 = HEAP32[$glyphpath>>2]|0;
  $275 = ((($274)) + 64|0);
  $276 = HEAP32[$275>>2]|0;
  $277 = (_FT_MulFix($276,$260)|0);
  $278 = HEAP32[$glyphpath>>2]|0;
  $279 = ((($278)) + 72|0);
  $280 = HEAP32[$279>>2]|0;
  $281 = (_FT_MulFix($280,$261)|0);
  $282 = (($281) + ($277))|0;
  $283 = HEAP32[$248>>2]|0;
  $284 = (($282) + ($283))|0;
  $285 = ((($params)) + 20|0);
  HEAP32[$285>>2] = $284;
  $286 = ((($params)) + 24|0);
  $287 = ((($glyphpath)) + 11776|0);
  $288 = HEAP32[$287>>2]|0;
  $289 = ((($glyphpath)) + 11780|0);
  $290 = HEAP32[$289>>2]|0;
  $291 = HEAP32[$218>>2]|0;
  $292 = (_FT_MulFix($291,$288)|0);
  $293 = HEAP32[$221>>2]|0;
  $294 = (_FT_MulFix($293,$290)|0);
  $295 = (($294) + ($292))|0;
  $296 = (_cf2_hintmap_map($hintmap,$290)|0);
  $297 = HEAP32[$glyphpath>>2]|0;
  $298 = ((($297)) + 60|0);
  $299 = HEAP32[$298>>2]|0;
  $300 = (_FT_MulFix($299,$295)|0);
  $301 = HEAP32[$glyphpath>>2]|0;
  $302 = ((($301)) + 68|0);
  $303 = HEAP32[$302>>2]|0;
  $304 = (_FT_MulFix($303,$296)|0);
  $305 = (($304) + ($300))|0;
  $306 = HEAP32[$235>>2]|0;
  $307 = (($305) + ($306))|0;
  HEAP32[$286>>2] = $307;
  $308 = HEAP32[$glyphpath>>2]|0;
  $309 = ((($308)) + 64|0);
  $310 = HEAP32[$309>>2]|0;
  $311 = (_FT_MulFix($310,$295)|0);
  $312 = HEAP32[$glyphpath>>2]|0;
  $313 = ((($312)) + 72|0);
  $314 = HEAP32[$313>>2]|0;
  $315 = (_FT_MulFix($314,$296)|0);
  $316 = (($315) + ($311))|0;
  $317 = HEAP32[$248>>2]|0;
  $318 = (($316) + ($317))|0;
  $319 = ((($params)) + 28|0);
  HEAP32[$319>>2] = $318;
  $320 = ((($glyphpath)) + 4|0);
  $321 = HEAP32[$320>>2]|0;
  $322 = ((($321)) + 12|0);
  $323 = HEAP32[$322>>2]|0;
  FUNCTION_TABLE_vii[$323 & 8191]($321,$params);
  $324 = $286;
  $325 = $324;
  $326 = HEAP32[$325>>2]|0;
  $327 = (($324) + 4)|0;
  $328 = $327;
  $329 = HEAP32[$328>>2]|0;
  $330 = $112;
  $331 = $330;
  HEAP32[$331>>2] = $326;
  $332 = (($330) + 4)|0;
  $333 = $332;
  HEAP32[$333>>2] = $329;
  break;
 }
 default: {
 }
 }
 $334 = ($useIntersection$0<<24>>24)!=(0);
 $335 = ($close<<24>>24)==(0);
 $or$cond = $335 & $334;
 do {
  if (!($or$cond)) {
   if ($335) {
    $373 = HEAP32[$nextP0>>2]|0;
    $374 = ((($nextP0)) + 4|0);
    $375 = HEAP32[$374>>2]|0;
    $376 = ((($glyphpath)) + 11644|0);
    $377 = HEAP32[$376>>2]|0;
    $378 = (_FT_MulFix($377,$373)|0);
    $379 = ((($glyphpath)) + 11648|0);
    $380 = HEAP32[$379>>2]|0;
    $381 = (_FT_MulFix($380,$375)|0);
    $382 = (($381) + ($378))|0;
    $383 = (_cf2_hintmap_map($hintmap,$375)|0);
    $384 = HEAP32[$glyphpath>>2]|0;
    $385 = ((($384)) + 60|0);
    $386 = HEAP32[$385>>2]|0;
    $387 = (_FT_MulFix($386,$382)|0);
    $388 = HEAP32[$glyphpath>>2]|0;
    $389 = ((($388)) + 68|0);
    $390 = HEAP32[$389>>2]|0;
    $391 = (_FT_MulFix($390,$383)|0);
    $392 = (($391) + ($387))|0;
    $393 = ((($glyphpath)) + 11656|0);
    $394 = HEAP32[$393>>2]|0;
    $395 = (($392) + ($394))|0;
    $396 = ((($params)) + 8|0);
    HEAP32[$396>>2] = $395;
    $397 = HEAP32[$glyphpath>>2]|0;
    $398 = ((($397)) + 64|0);
    $399 = HEAP32[$398>>2]|0;
    $400 = (_FT_MulFix($399,$382)|0);
    $401 = HEAP32[$glyphpath>>2]|0;
    $402 = ((($401)) + 72|0);
    $403 = HEAP32[$402>>2]|0;
    $404 = (_FT_MulFix($403,$383)|0);
    $405 = (($404) + ($400))|0;
    $406 = ((($glyphpath)) + 11660|0);
    $407 = HEAP32[$406>>2]|0;
    $408 = (($405) + ($407))|0;
    $410 = $408;
   } else {
    $336 = ((($glyphpath)) + 3876|0);
    $337 = HEAP32[$nextP0>>2]|0;
    $338 = ((($nextP0)) + 4|0);
    $339 = HEAP32[$338>>2]|0;
    $340 = ((($glyphpath)) + 11644|0);
    $341 = HEAP32[$340>>2]|0;
    $342 = (_FT_MulFix($341,$337)|0);
    $343 = ((($glyphpath)) + 11648|0);
    $344 = HEAP32[$343>>2]|0;
    $345 = (_FT_MulFix($344,$339)|0);
    $346 = (($345) + ($342))|0;
    $347 = (_cf2_hintmap_map($336,$339)|0);
    $348 = HEAP32[$glyphpath>>2]|0;
    $349 = ((($348)) + 60|0);
    $350 = HEAP32[$349>>2]|0;
    $351 = (_FT_MulFix($350,$346)|0);
    $352 = HEAP32[$glyphpath>>2]|0;
    $353 = ((($352)) + 68|0);
    $354 = HEAP32[$353>>2]|0;
    $355 = (_FT_MulFix($354,$347)|0);
    $356 = (($355) + ($351))|0;
    $357 = ((($glyphpath)) + 11656|0);
    $358 = HEAP32[$357>>2]|0;
    $359 = (($356) + ($358))|0;
    $360 = ((($params)) + 8|0);
    HEAP32[$360>>2] = $359;
    $361 = HEAP32[$glyphpath>>2]|0;
    $362 = ((($361)) + 64|0);
    $363 = HEAP32[$362>>2]|0;
    $364 = (_FT_MulFix($363,$346)|0);
    $365 = HEAP32[$glyphpath>>2]|0;
    $366 = ((($365)) + 72|0);
    $367 = HEAP32[$366>>2]|0;
    $368 = (_FT_MulFix($367,$347)|0);
    $369 = (($368) + ($364))|0;
    $370 = ((($glyphpath)) + 11660|0);
    $371 = HEAP32[$370>>2]|0;
    $372 = (($369) + ($371))|0;
    $410 = $372;
   }
   $409 = ((($params)) + 12|0);
   HEAP32[$409>>2] = $410;
   $411 = ((($params)) + 8|0);
   $412 = HEAP32[$411>>2]|0;
   $413 = HEAP32[$112>>2]|0;
   $414 = ($412|0)==($413|0);
   if ($414) {
    $415 = ((($glyphpath)) + 11732|0);
    $416 = HEAP32[$415>>2]|0;
    $417 = ($410|0)==($416|0);
    if ($417) {
     break;
    }
   }
   $418 = ((($params)) + 32|0);
   HEAP32[$418>>2] = 2;
   $419 = $112;
   $420 = $419;
   $421 = HEAP32[$420>>2]|0;
   $422 = (($419) + 4)|0;
   $423 = $422;
   $424 = HEAP32[$423>>2]|0;
   $425 = $params;
   $426 = $425;
   HEAP32[$426>>2] = $421;
   $427 = (($425) + 4)|0;
   $428 = $427;
   HEAP32[$428>>2] = $424;
   $429 = ((($glyphpath)) + 4|0);
   $430 = HEAP32[$429>>2]|0;
   $431 = ((($430)) + 4|0);
   $432 = HEAP32[$431>>2]|0;
   FUNCTION_TABLE_vii[$432 & 8191]($430,$params);
   $433 = $411;
   $434 = $433;
   $435 = HEAP32[$434>>2]|0;
   $436 = (($433) + 4)|0;
   $437 = $436;
   $438 = HEAP32[$437>>2]|0;
   $439 = $112;
   $440 = $439;
   HEAP32[$440>>2] = $435;
   $441 = (($439) + 4)|0;
   $442 = $441;
   HEAP32[$442>>2] = $438;
  }
 } while(0);
 if (!($334)) {
  STACKTOP = sp;return;
 }
 $443 = $nextP0;
 $444 = $443;
 HEAP32[$444>>2] = $intersection$sroa$0$4;
 $445 = (($443) + 4)|0;
 $446 = $445;
 HEAP32[$446>>2] = $intersection$sroa$11$4;
 STACKTOP = sp;return;
}
function _cf2_hintmap_map($hintmap,$csCoord) {
 $hintmap = $hintmap|0;
 $csCoord = $csCoord|0;
 var $$1 = 0, $$lcssa = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0, $i$0$lcssa = 0, $i$13 = 0, $i$14 = 0, $i$14$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($hintmap)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  $3 = ((($hintmap)) + 13|0);
  $4 = HEAP8[$3>>0]|0;
  $5 = ($4<<24>>24)==(0);
  if (!($5)) {
   $9 = ((($hintmap)) + 24|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = (($1) + -1)|0;
   $i$0 = $10;
   while(1) {
    $12 = ($i$0>>>0)<($11>>>0);
    if (!($12)) {
     $i$0$lcssa = $i$0;
     break;
    }
    $13 = (($i$0) + 1)|0;
    $14 = (((((($hintmap)) + 28|0) + (($13*20)|0)|0)) + 8|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = ($15|0)>($csCoord|0);
    if ($16) {
     $i$0$lcssa = $i$0;
     break;
    } else {
     $i$0 = $13;
    }
   }
   $17 = ($i$0$lcssa|0)==(0);
   L8: do {
    if ($17) {
     label = 11;
    } else {
     $i$14 = $i$0$lcssa;
     while(1) {
      $18 = (((((($hintmap)) + 28|0) + (($i$14*20)|0)|0)) + 8|0);
      $19 = HEAP32[$18>>2]|0;
      $20 = ($19|0)>($csCoord|0);
      if (!($20)) {
       $$lcssa = $18;$i$14$lcssa = $i$14;
       break;
      }
      $21 = (($i$14) + -1)|0;
      $22 = ($21|0)==(0);
      if ($22) {
       label = 11;
       break L8;
      } else {
       $i$14 = $21;
      }
     }
     HEAP32[$9>>2] = $i$14$lcssa;
     $$pre = HEAP32[$$lcssa>>2]|0;
     $34 = $$pre;$i$13 = $i$14$lcssa;
    }
   } while(0);
   if ((label|0) == 11) {
    HEAP32[$9>>2] = 0;
    $23 = ((($hintmap)) + 36|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = ($24|0)>($csCoord|0);
    if ($25) {
     $26 = (($csCoord) - ($24))|0;
     $27 = ((($hintmap)) + 16|0);
     $28 = HEAP32[$27>>2]|0;
     $29 = (_FT_MulFix($26,$28)|0);
     $30 = ((($hintmap)) + 40|0);
     $31 = HEAP32[$30>>2]|0;
     $32 = (($31) + ($29))|0;
     $$1 = $32;
     return ($$1|0);
    } else {
     $34 = $24;$i$13 = 0;
    }
   }
   $33 = (($csCoord) - ($34))|0;
   $35 = (((((($hintmap)) + 28|0) + (($i$13*20)|0)|0)) + 16|0);
   $36 = HEAP32[$35>>2]|0;
   $37 = (_FT_MulFix($33,$36)|0);
   $38 = (((((($hintmap)) + 28|0) + (($i$13*20)|0)|0)) + 12|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = (($39) + ($37))|0;
   $$1 = $40;
   return ($$1|0);
  }
 }
 $6 = ((($hintmap)) + 16|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (_FT_MulFix($csCoord,$7)|0);
 $$1 = $8;
 return ($$1|0);
}
function _cf2_hintmap_build($hintmap,$hStemHintArray,$vStemHintArray,$hintMask,$hintOrigin,$initialMap) {
 $hintmap = $hintmap|0;
 $hStemHintArray = $hStemHintArray|0;
 $vStemHintArray = $vStemHintArray|0;
 $hintMask = $hintMask|0;
 $hintOrigin = $hintOrigin|0;
 $initialMap = $initialMap|0;
 var $$$i = 0, $$0$i = 0, $$0$i$i = 0, $$0$i$i10 = 0, $$0$i$i27 = 0, $$0$i$i42 = 0, $$0$i$i5 = 0, $$1$i = 0, $$idx = 0, $$idx$phi$trans$insert$i = 0, $$idx$val = 0, $$idx$val$pre$i = 0, $$idx1$i = 0, $$idx1$val$i = 0, $$idx3$i = 0, $$idx3$val = 0, $$idx3$val$i = 0, $$idx3$val$pre = 0, $$idx4 = 0, $$idx4$i = 0;
 var $$idx4$val = 0, $$idx4$val$i = 0, $$lcssa$i = 0, $$lcssa160 = 0, $$lcssa60$ph = 0, $$lcssa61 = 0, $$lobit$i = 0, $$lobit$i$i = 0, $$lobit$i5$i = 0, $$ph$i$ph = 0, $$ph$i45$ph = 0, $$ph$i45109 = 0, $$ph7$i = 0, $$ph7$i35 = 0, $$phi$trans$insert$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i13$ph = 0, $$pre$i47$ph = 0, $$pre14$i = 0;
 var $$pre16$i = 0, $$pre17$i = 0, $$sink$i = 0, $$sink$i19 = 0, $$sink$i40 = 0, $$sink$i53 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0;
 var $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0;
 var $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0;
 var $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0;
 var $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0;
 var $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0;
 var $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0;
 var $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0;
 var $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0;
 var $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0;
 var $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0;
 var $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0;
 var $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0;
 var $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bottomHintEdge = 0, $bottomHintEdge$idx$val$pre$i = 0, $bottomHintEdge1 = 0, $dsMove$0$ph$i = 0, $dsNew$0$i = 0, $dsNew$1$i = 0, $dummy = 0;
 var $edge = 0, $exitcond = 0, $exitcond96 = 0, $hStemHintArray$idx$val = 0, $hStemHintArray$idx2 = 0, $hStemHintArray$idx2$val = 0, $hintMask$idx$val$pre = 0, $hintMask$idx1 = 0, $hintMask$idx1$val = 0, $i$01$i = 0, $i$010$i = 0, $i$013$i = 0, $i$013$i$lcssa$ph = 0, $i$013$i$lcssa58 = 0, $i$013$i$us = 0, $i$013$i$us79 = 0, $i$071 = 0, $i$1$i = 0, $i$166 = 0, $i$264 = 0;
 var $i$29$i = 0, $invalid = 0, $maskByte$073 = 0, $maskByte$1 = 0, $maskByte$268 = 0, $maskByte$3 = 0, $maskPtr$076 = 0, $maskPtr$1 = 0, $maskPtr$269 = 0, $maskPtr$3 = 0, $move$06$i = 0, $move$08$i = 0, $phitmp$i = 0, $savedMove$i = 0, $switch$split12D = 0, $switch$split2D = 0, $switch$split42D = 0, $switch$split72D = 0, $tempHintMask = 0, $topHintEdge = 0;
 var $topHintEdge$idx$val$i = 0, $topHintEdge2 = 0, $vStemHintArray$idx = 0, $vStemHintArray$idx$val = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $savedMove$i = sp + 168|0;
 $tempHintMask = sp + 140|0;
 $dummy = sp + 120|0;
 $bottomHintEdge = sp + 100|0;
 $topHintEdge = sp + 80|0;
 $edge = sp + 60|0;
 $invalid = sp + 40|0;
 $bottomHintEdge1 = sp + 20|0;
 $topHintEdge2 = sp;
 $0 = HEAP32[$hintmap>>2]|0;
 $1 = ($initialMap<<24>>24)!=(0);
 if (!($1)) {
  $2 = ((($hintmap)) + 4|0);
  $3 = HEAP32[$2>>2]|0;
  $$idx = ((($3)) + 12|0);
  $$idx$val = HEAP8[$$idx>>0]|0;
  $4 = ($$idx$val<<24>>24)==(0);
  if ($4) {
   $5 = HEAP32[$hintMask>>2]|0;
   ;HEAP32[$tempHintMask>>2]=0|0;HEAP32[$tempHintMask+4>>2]=0|0;HEAP32[$tempHintMask+8>>2]=0|0;HEAP32[$tempHintMask+12>>2]=0|0;HEAP32[$tempHintMask+16>>2]=0|0;HEAP32[$tempHintMask+20>>2]=0|0;HEAP32[$tempHintMask+24>>2]=0|0;
   HEAP32[$tempHintMask>>2] = $5;
   _cf2_hintmap_build($3,$hStemHintArray,$vStemHintArray,$tempHintMask,$hintOrigin,1);
  }
 }
 $hintMask$idx1 = ((($hintMask)) + 4|0);
 $hintMask$idx1$val = HEAP8[$hintMask$idx1>>0]|0;
 $6 = ($hintMask$idx1$val<<24>>24)==(0);
 $hStemHintArray$idx2 = ((($hStemHintArray)) + 20|0);
 do {
  if ($6) {
   $hStemHintArray$idx2$val = HEAP32[$hStemHintArray$idx2>>2]|0;
   $vStemHintArray$idx = ((($vStemHintArray)) + 20|0);
   $vStemHintArray$idx$val = HEAP32[$vStemHintArray$idx>>2]|0;
   $7 = (($vStemHintArray$idx$val) + ($hStemHintArray$idx2$val))|0;
   $8 = (0 - ($7))|0;
   $9 = $8 & 7;
   $10 = ($7>>>0)>(96);
   if ($10) {
    $11 = HEAP32[$hintMask>>2]|0;
    $12 = ($11|0)==(0|0);
    if ($12) {
     STACKTOP = sp;return;
    }
    $13 = HEAP32[$11>>2]|0;
    $14 = ($13|0)==(0);
    if (!($14)) {
     STACKTOP = sp;return;
    }
    HEAP32[$11>>2] = 18;
    STACKTOP = sp;return;
   } else {
    $15 = ((($hintMask)) + 8|0);
    HEAP32[$15>>2] = $7;
    $16 = (($7) + 7)|0;
    $17 = $16 >>> 3;
    $18 = ((($hintMask)) + 12|0);
    HEAP32[$18>>2] = $17;
    HEAP8[$hintMask$idx1>>0] = 1;
    $19 = ((($hintMask)) + 5|0);
    HEAP8[$19>>0] = 1;
    $20 = ($7|0)==(0);
    if ($20) {
     break;
    }
    $21 = ($17|0)==(0);
    if ($21) {
     $$lcssa$i = -1;
    } else {
     $i$01$i = 0;
     while(1) {
      $22 = (((($hintMask)) + 16|0) + ($i$01$i)|0);
      HEAP8[$22>>0] = -1;
      $23 = (($i$01$i) + 1)|0;
      $24 = HEAP32[$18>>2]|0;
      $25 = ($23>>>0)<($24>>>0);
      if ($25) {
       $i$01$i = $23;
      } else {
       $$lcssa160 = $24;
       break;
      }
     }
     $phitmp$i = (($$lcssa160) + -1)|0;
     $$lcssa$i = $phitmp$i;
    }
    $26 = -1 << $9;
    $27 = (((($hintMask)) + 16|0) + ($$lcssa$i)|0);
    $28 = HEAP8[$27>>0]|0;
    $29 = $28&255;
    $30 = $29 & $26;
    $31 = $30&255;
    HEAP8[$27>>0] = $31;
    $hintMask$idx$val$pre = HEAP8[$hintMask$idx1>>0]|0;
    $32 = ($hintMask$idx$val$pre<<24>>24)==(0);
    if (!($32)) {
     break;
    }
    STACKTOP = sp;return;
   }
  }
 } while(0);
 $33 = ((($hintmap)) + 20|0);
 HEAP32[$33>>2] = 0;
 $34 = ((($hintmap)) + 24|0);
 HEAP32[$34>>2] = 0;
 ;HEAP32[$tempHintMask>>2]=HEAP32[$hintMask>>2]|0;HEAP32[$tempHintMask+4>>2]=HEAP32[$hintMask+4>>2]|0;HEAP32[$tempHintMask+8>>2]=HEAP32[$hintMask+8>>2]|0;HEAP32[$tempHintMask+12>>2]=HEAP32[$hintMask+12>>2]|0;HEAP32[$tempHintMask+16>>2]=HEAP32[$hintMask+16>>2]|0;HEAP32[$tempHintMask+20>>2]=HEAP32[$hintMask+20>>2]|0;HEAP32[$tempHintMask+24>>2]=HEAP32[$hintMask+24>>2]|0;
 $35 = ((($tempHintMask)) + 16|0);
 $hStemHintArray$idx$val = HEAP32[$hStemHintArray$idx2>>2]|0;
 $36 = ((($hintMask)) + 8|0);
 $37 = HEAP32[$36>>2]|0;
 $38 = ($hStemHintArray$idx$val>>>0)>($37>>>0);
 if ($38) {
  STACKTOP = sp;return;
 }
 $39 = ((($0)) + 205|0);
 $40 = HEAP8[$39>>0]|0;
 $41 = ($40<<24>>24)==(0);
 if (!($41)) {
  ;HEAP32[$dummy>>2]=0|0;HEAP32[$dummy+4>>2]=0|0;HEAP32[$dummy+8>>2]=0|0;HEAP32[$dummy+12>>2]=0|0;HEAP32[$dummy+16>>2]=0|0;
  $42 = ((($0)) + 244|0);
  _cf2_hintmap_insertHint($hintmap,$42,$dummy);
  $43 = ((($0)) + 224|0);
  _cf2_hintmap_insertHint($hintmap,$dummy,$43);
 }
 $44 = ($hStemHintArray$idx$val|0)==(0);
 if (!($44)) {
  $45 = ((($hintmap)) + 16|0);
  $46 = ((($hStemHintArray)) + 28|0);
  $47 = ((($hStemHintArray)) + 8|0);
  $48 = ((($bottomHintEdge)) + 8|0);
  $49 = ((($bottomHintEdge)) + 16|0);
  $50 = ((($bottomHintEdge)) + 4|0);
  $51 = ((($bottomHintEdge)) + 12|0);
  $52 = ((($topHintEdge)) + 8|0);
  $53 = ((($0)) + 188|0);
  $54 = ((($topHintEdge)) + 16|0);
  $55 = ((($topHintEdge)) + 4|0);
  $56 = ((($topHintEdge)) + 12|0);
  $57 = ((($0)) + 216|0);
  $58 = ((($0)) + 200|0);
  $59 = ((($0)) + 204|0);
  $60 = ((($0)) + 212|0);
  $61 = ((($hStemHintArray)) + 4|0);
  $i$071 = 0;$maskByte$073 = 128;$maskPtr$076 = $35;
  while(1) {
   $62 = HEAP8[$maskPtr$076>>0]|0;
   $63 = $62&255;
   $64 = $63 & $maskByte$073;
   $65 = ($64|0)==(0);
   if (!($65)) {
    $66 = HEAP32[$45>>2]|0;
    ;HEAP32[$bottomHintEdge>>2]=0|0;HEAP32[$bottomHintEdge+4>>2]=0|0;HEAP32[$bottomHintEdge+8>>2]=0|0;HEAP32[$bottomHintEdge+12>>2]=0|0;HEAP32[$bottomHintEdge+16>>2]=0|0;
    $67 = HEAP32[$hStemHintArray$idx2>>2]|0;
    $68 = ($67>>>0)>($i$071>>>0);
    if ($68) {
     $$0$i$i42 = $i$071;
    } else {
     $69 = HEAP32[$61>>2]|0;
     $70 = ($69|0)==(0|0);
     if ($70) {
      $$0$i$i42 = 0;
     } else {
      $71 = HEAP32[$69>>2]|0;
      $72 = ($71|0)==(0);
      if ($72) {
       HEAP32[$69>>2] = 130;
       $$0$i$i42 = 0;
      } else {
       $$0$i$i42 = 0;
      }
     }
    }
    $73 = HEAP32[$46>>2]|0;
    $74 = HEAP32[$47>>2]|0;
    $75 = Math_imul($74, $$0$i$i42)|0;
    $76 = (($73) + ($75)|0);
    $77 = ((($76)) + 8|0);
    $78 = HEAP32[$77>>2]|0;
    $79 = ((($76)) + 4|0);
    $80 = HEAP32[$79>>2]|0;
    $81 = (($78) - ($80))|0;
    $switch$split2D = ($81|0)<(-1310720);
    L42: do {
     if ($switch$split2D) {
      switch ($81|0) {
      case -1376256:  {
       break;
      }
      default: {
       label = 26;
       break L42;
      }
      }
      HEAP32[$48>>2] = $78;
      HEAP32[$bottomHintEdge>>2] = 1;
      $$ph$i45$ph = 1;$$pre$i47$ph = $78;
      label = 30;
     } else {
      switch ($81|0) {
      case -1310720:  {
       break;
      }
      default: {
       label = 26;
       break L42;
      }
      }
      HEAP32[$bottomHintEdge>>2] = 0;
      HEAP32[$48>>2] = $hintOrigin;
      HEAP32[$49>>2] = $66;
      HEAP32[$50>>2] = $i$071;
      $$ph$i45109 = 0;$88 = $hintOrigin;
      label = 32;
     }
    } while(0);
    do {
     if ((label|0) == 26) {
      label = 0;
      $82 = ($81|0)<(0);
      if ($82) {
       HEAP32[$48>>2] = $78;
       HEAP32[$bottomHintEdge>>2] = 4;
       $$ph$i45$ph = 4;$$pre$i47$ph = $78;
       label = 30;
       break;
      } else {
       HEAP32[$48>>2] = $80;
       HEAP32[$bottomHintEdge>>2] = 4;
       $$ph$i45$ph = 4;$$pre$i47$ph = $80;
       label = 30;
       break;
      }
     }
    } while(0);
    if ((label|0) == 30) {
     label = 0;
     $83 = (($$pre$i47$ph) + ($hintOrigin))|0;
     HEAP32[$48>>2] = $83;
     HEAP32[$49>>2] = $66;
     HEAP32[$50>>2] = $i$071;
     $84 = HEAP8[$76>>0]|0;
     $85 = ($84<<24>>24)==(0);
     if ($85) {
      $$ph$i45109 = $$ph$i45$ph;$88 = $83;
      label = 32;
     } else {
      $86 = ((($76)) + 12|0);
      $$sink$i53 = HEAP32[$86>>2]|0;
      HEAP32[$51>>2] = $$sink$i53;
      $87 = $$ph$i45$ph | 16;
      HEAP32[$bottomHintEdge>>2] = $87;
      $149 = $83;$175 = $$sink$i53;$bottomHintEdge$idx$val$pre$i = $87;
     }
    }
    if ((label|0) == 32) {
     label = 0;
     $89 = (_FT_MulFix($88,$66)|0);
     HEAP32[$51>>2] = $89;
     $149 = $88;$175 = $89;$bottomHintEdge$idx$val$pre$i = $$ph$i45109;
    }
    $90 = HEAP32[$45>>2]|0;
    ;HEAP32[$topHintEdge>>2]=0|0;HEAP32[$topHintEdge+4>>2]=0|0;HEAP32[$topHintEdge+8>>2]=0|0;HEAP32[$topHintEdge+12>>2]=0|0;HEAP32[$topHintEdge+16>>2]=0|0;
    $91 = HEAP32[$hStemHintArray$idx2>>2]|0;
    $92 = ($91>>>0)>($i$071>>>0);
    if ($92) {
     $$0$i$i27 = $i$071;
    } else {
     $93 = HEAP32[$61>>2]|0;
     $94 = ($93|0)==(0|0);
     if ($94) {
      $$0$i$i27 = 0;
     } else {
      $95 = HEAP32[$93>>2]|0;
      $96 = ($95|0)==(0);
      if ($96) {
       HEAP32[$93>>2] = 130;
       $$0$i$i27 = 0;
      } else {
       $$0$i$i27 = 0;
      }
     }
    }
    $97 = HEAP32[$46>>2]|0;
    $98 = HEAP32[$47>>2]|0;
    $99 = Math_imul($98, $$0$i$i27)|0;
    $100 = (($97) + ($99)|0);
    $101 = ((($100)) + 8|0);
    $102 = HEAP32[$101>>2]|0;
    $103 = ((($100)) + 4|0);
    $104 = HEAP32[$103>>2]|0;
    $105 = (($102) - ($104))|0;
    $switch$split12D = ($105|0)<(-1310720);
    L63: do {
     if ($switch$split12D) {
      switch ($105|0) {
      case -1376256:  {
       break;
      }
      default: {
       label = 40;
       break L63;
      }
      }
      HEAP32[$topHintEdge>>2] = 0;
      HEAP32[$52>>2] = $hintOrigin;
      HEAP32[$54>>2] = $90;
      HEAP32[$55>>2] = $i$071;
      $116 = $hintOrigin;$465 = 0;
      label = 45;
     } else {
      switch ($105|0) {
      case -1310720:  {
       break;
      }
      default: {
       label = 40;
       break L63;
      }
      }
      HEAP32[$52>>2] = $104;
      HEAP32[$topHintEdge>>2] = 2;
      $$ph7$i35 = 2;$110 = $104;
      label = 43;
     }
    } while(0);
    do {
     if ((label|0) == 40) {
      label = 0;
      $106 = ($105|0)<(0);
      if ($106) {
       HEAP32[$52>>2] = $104;
       HEAP32[$topHintEdge>>2] = 8;
       $$ph7$i35 = 8;$110 = $104;
       label = 43;
       break;
      } else {
       HEAP32[$52>>2] = $102;
       HEAP32[$topHintEdge>>2] = 8;
       $$ph7$i35 = 8;$110 = $102;
       label = 43;
       break;
      }
     }
    } while(0);
    if ((label|0) == 43) {
     label = 0;
     $107 = HEAP32[$53>>2]|0;
     $108 = $107 << 1;
     $109 = (($110) + ($108))|0;
     $111 = (($109) + ($hintOrigin))|0;
     HEAP32[$52>>2] = $111;
     HEAP32[$54>>2] = $90;
     HEAP32[$55>>2] = $i$071;
     $112 = HEAP8[$100>>0]|0;
     $113 = ($112<<24>>24)==(0);
     if ($113) {
      $116 = $111;$465 = $$ph7$i35;
      label = 45;
     } else {
      $114 = ((($100)) + 16|0);
      $$sink$i40 = HEAP32[$114>>2]|0;
      HEAP32[$56>>2] = $$sink$i40;
      $115 = $$ph7$i35 | 16;
      HEAP32[$topHintEdge>>2] = $115;
      $135 = $111;$195 = $$sink$i40;$topHintEdge$idx$val$i = $115;
     }
    }
    if ((label|0) == 45) {
     label = 0;
     $117 = (_FT_MulFix($116,$90)|0);
     HEAP32[$56>>2] = $117;
     $135 = $116;$195 = $117;$topHintEdge$idx$val$i = $465;
    }
    $118 = $bottomHintEdge$idx$val$pre$i | $topHintEdge$idx$val$i;
    $119 = $118 & 16;
    $120 = ($119|0)==(0);
    L79: do {
     if ($120) {
      $121 = HEAP32[$57>>2]|0;
      $122 = HEAP32[$58>>2]|0;
      $123 = ($122|0)==(0);
      if (!($123)) {
       $124 = $topHintEdge$idx$val$i & 10;
       $125 = ($124|0)==(0);
       $126 = $bottomHintEdge$idx$val$pre$i & 5;
       $127 = ($126|0)==(0);
       L82: do {
        if ($127) {
         if ($125) {
          break L79;
         } else {
          $i$013$i$us = 0;
         }
         while(1) {
          $128 = (((((($0)) + 264|0) + (($i$013$i$us*20)|0)|0)) + 16|0);
          $129 = HEAP8[$128>>0]|0;
          $130 = ($129<<24>>24)==(0);
          do {
           if ($130) {
            $131 = (((($0)) + 264|0) + (($i$013$i$us*20)|0)|0);
            $132 = HEAP32[$131>>2]|0;
            $133 = (($132) - ($121))|0;
            $134 = ($133|0)>($135|0);
            if ($134) {
             break;
            }
            $136 = (((((($0)) + 264|0) + (($i$013$i$us*20)|0)|0)) + 4|0);
            $137 = HEAP32[$136>>2]|0;
            $138 = (($137) + ($121))|0;
            $139 = ($135|0)>($138|0);
            if (!($139)) {
             $$lcssa61 = $132;$i$013$i$lcssa58 = $i$013$i$us;
             label = 69;
             break L82;
            }
           }
          } while(0);
          $140 = (($i$013$i$us) + 1)|0;
          $141 = ($140>>>0)<($122>>>0);
          if ($141) {
           $i$013$i$us = $140;
          } else {
           break L79;
          }
         }
        } else {
         L91: do {
          if ($125) {
           $i$013$i$us79 = 0;
           while(1) {
            $142 = (((((($0)) + 264|0) + (($i$013$i$us79*20)|0)|0)) + 16|0);
            $143 = HEAP8[$142>>0]|0;
            $144 = ($143<<24>>24)==(0);
            do {
             if (!($144)) {
              $145 = (((($0)) + 264|0) + (($i$013$i$us79*20)|0)|0);
              $146 = HEAP32[$145>>2]|0;
              $147 = (($146) - ($121))|0;
              $148 = ($147|0)>($149|0);
              if ($148) {
               break;
              }
              $150 = (((((($0)) + 264|0) + (($i$013$i$us79*20)|0)|0)) + 4|0);
              $151 = HEAP32[$150>>2]|0;
              $152 = (($151) + ($121))|0;
              $153 = ($149|0)>($152|0);
              if (!($153)) {
               $$lcssa60$ph = $151;$i$013$i$lcssa$ph = $i$013$i$us79;
               break L91;
              }
             }
            } while(0);
            $154 = (($i$013$i$us79) + 1)|0;
            $155 = ($154>>>0)<($122>>>0);
            if ($155) {
             $i$013$i$us79 = $154;
            } else {
             break L79;
            }
           }
          } else {
           $i$013$i = 0;
           while(1) {
            $156 = (((((($0)) + 264|0) + (($i$013$i*20)|0)|0)) + 16|0);
            $157 = HEAP8[$156>>0]|0;
            $158 = ($157<<24>>24)==(0);
            $159 = (((($0)) + 264|0) + (($i$013$i*20)|0)|0);
            $160 = HEAP32[$159>>2]|0;
            $161 = (($160) - ($121))|0;
            do {
             if ($158) {
              $182 = ($161|0)>($135|0);
              if ($182) {
               break;
              }
              $183 = (((((($0)) + 264|0) + (($i$013$i*20)|0)|0)) + 4|0);
              $184 = HEAP32[$183>>2]|0;
              $185 = (($184) + ($121))|0;
              $186 = ($135|0)>($185|0);
              if (!($186)) {
               $$lcssa61 = $160;$i$013$i$lcssa58 = $i$013$i;
               label = 69;
               break L82;
              }
             } else {
              $162 = ($161|0)>($149|0);
              if ($162) {
               break;
              }
              $163 = (((((($0)) + 264|0) + (($i$013$i*20)|0)|0)) + 4|0);
              $164 = HEAP32[$163>>2]|0;
              $165 = (($164) + ($121))|0;
              $166 = ($149|0)>($165|0);
              if (!($166)) {
               $$lcssa60$ph = $164;$i$013$i$lcssa$ph = $i$013$i;
               break L91;
              }
             }
            } while(0);
            $202 = (($i$013$i) + 1)|0;
            $203 = ($202>>>0)<($122>>>0);
            if ($203) {
             $i$013$i = $202;
            } else {
             break L79;
            }
           }
          }
         } while(0);
         $167 = HEAP8[$59>>0]|0;
         $168 = ($167<<24>>24)==(0);
         do {
          if ($168) {
           $171 = (($$lcssa60$ph) - ($149))|0;
           $172 = HEAP32[$60>>2]|0;
           $173 = ($171|0)<($172|0);
           $174 = (($175) + 32768)|0;
           $176 = $174 & -65536;
           if ($173) {
            $dsNew$0$i = $176;
            break;
           }
           $177 = (((((($0)) + 264|0) + (($i$013$i$lcssa$ph*20)|0)|0)) + 12|0);
           $178 = HEAP32[$177>>2]|0;
           $179 = (($178) + -65536)|0;
           $180 = ($176|0)<($179|0);
           $$$i = $180 ? $176 : $179;
           $dsNew$0$i = $$$i;
          } else {
           $169 = (((((($0)) + 264|0) + (($i$013$i$lcssa$ph*20)|0)|0)) + 12|0);
           $170 = HEAP32[$169>>2]|0;
           $dsNew$0$i = $170;
          }
         } while(0);
         $181 = (($dsNew$0$i) - ($175))|0;
         $dsMove$0$ph$i = $181;
        }
       } while(0);
       if ((label|0) == 69) {
        label = 0;
        $187 = HEAP8[$59>>0]|0;
        $188 = ($187<<24>>24)==(0);
        do {
         if ($188) {
          $191 = (($135) - ($$lcssa61))|0;
          $192 = HEAP32[$60>>2]|0;
          $193 = ($191|0)<($192|0);
          $194 = (($195) + 32768)|0;
          $196 = $194 & -65536;
          if ($193) {
           $dsNew$1$i = $196;
           break;
          }
          $197 = (((((($0)) + 264|0) + (($i$013$i$lcssa58*20)|0)|0)) + 12|0);
          $198 = HEAP32[$197>>2]|0;
          $199 = (($198) + 65536)|0;
          $200 = ($196|0)>($199|0);
          $$1$i = $200 ? $196 : $199;
          $dsNew$1$i = $$1$i;
         } else {
          $189 = (((((($0)) + 264|0) + (($i$013$i$lcssa58*20)|0)|0)) + 12|0);
          $190 = HEAP32[$189>>2]|0;
          $dsNew$1$i = $190;
         }
        } while(0);
        $201 = (($dsNew$1$i) - ($195))|0;
        $dsMove$0$ph$i = $201;
       }
       $204 = ($bottomHintEdge$idx$val$pre$i|0)==(0);
       if (!($204)) {
        $205 = (($175) + ($dsMove$0$ph$i))|0;
        HEAP32[$51>>2] = $205;
        $206 = $bottomHintEdge$idx$val$pre$i | 16;
        HEAP32[$bottomHintEdge>>2] = $206;
       }
       $207 = ($topHintEdge$idx$val$i|0)==(0);
       if ($207) {
        label = 79;
       } else {
        $208 = (($195) + ($dsMove$0$ph$i))|0;
        HEAP32[$56>>2] = $208;
        $209 = $topHintEdge$idx$val$i | 16;
        HEAP32[$topHintEdge>>2] = $209;
        label = 79;
       }
      }
     } else {
      label = 79;
     }
    } while(0);
    if ((label|0) == 79) {
     label = 0;
     _cf2_hintmap_insertHint($hintmap,$bottomHintEdge,$topHintEdge);
     $210 = $maskByte$073 ^ 255;
     $211 = HEAP8[$maskPtr$076>>0]|0;
     $212 = $211&255;
     $213 = $212 & $210;
     $214 = $213&255;
     HEAP8[$maskPtr$076>>0] = $214;
    }
   }
   $215 = $i$071 & 7;
   $216 = ($215|0)==(7);
   $217 = ((($maskPtr$076)) + 1|0);
   $218 = $maskByte$073 >>> 1;
   $219 = $218 & 127;
   $maskByte$1 = $216 ? 128 : $219;
   $maskPtr$1 = $216 ? $217 : $maskPtr$076;
   $220 = (($i$071) + 1)|0;
   $exitcond96 = ($220|0)==($hStemHintArray$idx$val|0);
   if ($exitcond96) {
    break;
   } else {
    $i$071 = $220;$maskByte$073 = $maskByte$1;$maskPtr$076 = $maskPtr$1;
   }
  }
 }
 do {
  if ($1) {
   $234 = HEAP32[$33>>2]|0;
   $235 = ($234|0)==(0);
   if (!($235)) {
    $236 = ((($hintmap)) + 36|0);
    $237 = HEAP32[$236>>2]|0;
    $238 = ($237|0)>(0);
    if (!($238)) {
     $239 = (($234) + -1)|0;
     $240 = (((((($hintmap)) + 28|0) + (($239*20)|0)|0)) + 8|0);
     $241 = HEAP32[$240>>2]|0;
     $242 = ($241|0)<(0);
     if (!($242)) {
      break;
     }
    }
   }
   ;HEAP32[$edge>>2]=0|0;HEAP32[$edge+4>>2]=0|0;HEAP32[$edge+8>>2]=0|0;HEAP32[$edge+12>>2]=0|0;
   HEAP32[$edge>>2] = 49;
   $243 = ((($hintmap)) + 16|0);
   $244 = HEAP32[$243>>2]|0;
   $245 = ((($edge)) + 16|0);
   HEAP32[$245>>2] = $244;
   ;HEAP32[$invalid>>2]=0|0;HEAP32[$invalid+4>>2]=0|0;HEAP32[$invalid+8>>2]=0|0;HEAP32[$invalid+12>>2]=0|0;HEAP32[$invalid+16>>2]=0|0;
   _cf2_hintmap_insertHint($hintmap,$edge,$invalid);
  } else {
   if (!($44)) {
    $221 = ((($hintmap)) + 16|0);
    $222 = ((($hStemHintArray)) + 28|0);
    $223 = ((($hStemHintArray)) + 8|0);
    $224 = ((($bottomHintEdge1)) + 8|0);
    $225 = ((($bottomHintEdge1)) + 16|0);
    $226 = ((($bottomHintEdge1)) + 4|0);
    $227 = ((($bottomHintEdge1)) + 12|0);
    $228 = ((($topHintEdge2)) + 8|0);
    $229 = ((($0)) + 188|0);
    $230 = ((($topHintEdge2)) + 16|0);
    $231 = ((($topHintEdge2)) + 4|0);
    $232 = ((($topHintEdge2)) + 12|0);
    $233 = ((($hStemHintArray)) + 4|0);
    $i$166 = 0;$maskByte$268 = 128;$maskPtr$269 = $35;
    while(1) {
     $246 = HEAP8[$maskPtr$269>>0]|0;
     $247 = $246&255;
     $248 = $247 & $maskByte$268;
     $249 = ($248|0)==(0);
     if (!($249)) {
      $250 = HEAP32[$221>>2]|0;
      ;HEAP32[$bottomHintEdge1>>2]=0|0;HEAP32[$bottomHintEdge1+4>>2]=0|0;HEAP32[$bottomHintEdge1+8>>2]=0|0;HEAP32[$bottomHintEdge1+12>>2]=0|0;HEAP32[$bottomHintEdge1+16>>2]=0|0;
      $251 = HEAP32[$hStemHintArray$idx2>>2]|0;
      $252 = ($251>>>0)>($i$166>>>0);
      if ($252) {
       $$0$i$i10 = $i$166;
      } else {
       $253 = HEAP32[$233>>2]|0;
       $254 = ($253|0)==(0|0);
       if ($254) {
        $$0$i$i10 = 0;
       } else {
        $255 = HEAP32[$253>>2]|0;
        $256 = ($255|0)==(0);
        if ($256) {
         HEAP32[$253>>2] = 130;
         $$0$i$i10 = 0;
        } else {
         $$0$i$i10 = 0;
        }
       }
      }
      $257 = HEAP32[$222>>2]|0;
      $258 = HEAP32[$223>>2]|0;
      $259 = Math_imul($258, $$0$i$i10)|0;
      $260 = (($257) + ($259)|0);
      $261 = ((($260)) + 8|0);
      $262 = HEAP32[$261>>2]|0;
      $263 = ((($260)) + 4|0);
      $264 = HEAP32[$263>>2]|0;
      $265 = (($262) - ($264))|0;
      $switch$split42D = ($265|0)<(-1310720);
      L146: do {
       if ($switch$split42D) {
        switch ($265|0) {
        case -1376256:  {
         break;
        }
        default: {
         label = 96;
         break L146;
        }
        }
        HEAP32[$224>>2] = $262;
        HEAP32[$bottomHintEdge1>>2] = 1;
        $$ph$i$ph = 17;$$pre$i13$ph = $262;
        label = 100;
       } else {
        switch ($265|0) {
        case -1310720:  {
         break;
        }
        default: {
         label = 96;
         break L146;
        }
        }
        HEAP32[$bottomHintEdge1>>2] = 0;
        HEAP32[$224>>2] = $hintOrigin;
        HEAP32[$225>>2] = $250;
        HEAP32[$226>>2] = $i$166;
        $271 = $hintOrigin;
        label = 102;
       }
      } while(0);
      do {
       if ((label|0) == 96) {
        label = 0;
        $266 = ($265|0)<(0);
        if ($266) {
         HEAP32[$224>>2] = $262;
         HEAP32[$bottomHintEdge1>>2] = 4;
         $$ph$i$ph = 20;$$pre$i13$ph = $262;
         label = 100;
         break;
        } else {
         HEAP32[$224>>2] = $264;
         HEAP32[$bottomHintEdge1>>2] = 4;
         $$ph$i$ph = 20;$$pre$i13$ph = $264;
         label = 100;
         break;
        }
       }
      } while(0);
      if ((label|0) == 100) {
       label = 0;
       $267 = (($$pre$i13$ph) + ($hintOrigin))|0;
       HEAP32[$224>>2] = $267;
       HEAP32[$225>>2] = $250;
       HEAP32[$226>>2] = $i$166;
       $268 = HEAP8[$260>>0]|0;
       $269 = ($268<<24>>24)==(0);
       if ($269) {
        $271 = $267;
        label = 102;
       } else {
        $270 = ((($260)) + 12|0);
        $$sink$i19 = HEAP32[$270>>2]|0;
        HEAP32[$227>>2] = $$sink$i19;
        HEAP32[$bottomHintEdge1>>2] = $$ph$i$ph;
       }
      }
      if ((label|0) == 102) {
       label = 0;
       $272 = (_FT_MulFix($271,$250)|0);
       HEAP32[$227>>2] = $272;
      }
      $273 = HEAP32[$221>>2]|0;
      ;HEAP32[$topHintEdge2>>2]=0|0;HEAP32[$topHintEdge2+4>>2]=0|0;HEAP32[$topHintEdge2+8>>2]=0|0;HEAP32[$topHintEdge2+12>>2]=0|0;HEAP32[$topHintEdge2+16>>2]=0|0;
      $274 = HEAP32[$hStemHintArray$idx2>>2]|0;
      $275 = ($274>>>0)>($i$166>>>0);
      if ($275) {
       $$0$i$i5 = $i$166;
      } else {
       $276 = HEAP32[$233>>2]|0;
       $277 = ($276|0)==(0|0);
       if ($277) {
        $$0$i$i5 = 0;
       } else {
        $278 = HEAP32[$276>>2]|0;
        $279 = ($278|0)==(0);
        if ($279) {
         HEAP32[$276>>2] = 130;
         $$0$i$i5 = 0;
        } else {
         $$0$i$i5 = 0;
        }
       }
      }
      $280 = HEAP32[$222>>2]|0;
      $281 = HEAP32[$223>>2]|0;
      $282 = Math_imul($281, $$0$i$i5)|0;
      $283 = (($280) + ($282)|0);
      $284 = ((($283)) + 8|0);
      $285 = HEAP32[$284>>2]|0;
      $286 = ((($283)) + 4|0);
      $287 = HEAP32[$286>>2]|0;
      $288 = (($285) - ($287))|0;
      $switch$split72D = ($288|0)<(-1310720);
      L167: do {
       if ($switch$split72D) {
        switch ($288|0) {
        case -1376256:  {
         break;
        }
        default: {
         label = 110;
         break L167;
        }
        }
        HEAP32[$topHintEdge2>>2] = 0;
        HEAP32[$228>>2] = $hintOrigin;
        HEAP32[$230>>2] = $273;
        HEAP32[$231>>2] = $i$166;
        $298 = $hintOrigin;
        label = 115;
       } else {
        switch ($288|0) {
        case -1310720:  {
         break;
        }
        default: {
         label = 110;
         break L167;
        }
        }
        HEAP32[$228>>2] = $287;
        HEAP32[$topHintEdge2>>2] = 2;
        $$ph7$i = 18;$293 = $287;
        label = 113;
       }
      } while(0);
      do {
       if ((label|0) == 110) {
        label = 0;
        $289 = ($288|0)<(0);
        if ($289) {
         HEAP32[$228>>2] = $287;
         HEAP32[$topHintEdge2>>2] = 8;
         $$ph7$i = 24;$293 = $287;
         label = 113;
         break;
        } else {
         HEAP32[$228>>2] = $285;
         HEAP32[$topHintEdge2>>2] = 8;
         $$ph7$i = 24;$293 = $285;
         label = 113;
         break;
        }
       }
      } while(0);
      if ((label|0) == 113) {
       label = 0;
       $290 = HEAP32[$229>>2]|0;
       $291 = $290 << 1;
       $292 = (($293) + ($291))|0;
       $294 = (($292) + ($hintOrigin))|0;
       HEAP32[$228>>2] = $294;
       HEAP32[$230>>2] = $273;
       HEAP32[$231>>2] = $i$166;
       $295 = HEAP8[$283>>0]|0;
       $296 = ($295<<24>>24)==(0);
       if ($296) {
        $298 = $294;
        label = 115;
       } else {
        $297 = ((($283)) + 16|0);
        $$sink$i = HEAP32[$297>>2]|0;
        HEAP32[$232>>2] = $$sink$i;
        HEAP32[$topHintEdge2>>2] = $$ph7$i;
       }
      }
      if ((label|0) == 115) {
       label = 0;
       $299 = (_FT_MulFix($298,$273)|0);
       HEAP32[$232>>2] = $299;
      }
      _cf2_hintmap_insertHint($hintmap,$bottomHintEdge1,$topHintEdge2);
     }
     $300 = $i$166 & 7;
     $301 = ($300|0)==(7);
     $302 = ((($maskPtr$269)) + 1|0);
     $303 = $maskByte$268 >>> 1;
     $304 = $303 & 127;
     $maskByte$3 = $301 ? 128 : $304;
     $maskPtr$3 = $301 ? $302 : $maskPtr$269;
     $305 = (($i$166) + 1)|0;
     $exitcond = ($305|0)==($hStemHintArray$idx$val|0);
     if ($exitcond) {
      break;
     } else {
      $i$166 = $305;$maskByte$268 = $maskByte$3;$maskPtr$269 = $maskPtr$3;
     }
    }
   }
  }
 } while(0);
 $306 = ((($hintmap)) + 8|0);
 $307 = HEAP32[$306>>2]|0;
 $308 = ((($307)) + 20|0);
 HEAP32[$308>>2] = 0;
 $309 = HEAP32[$33>>2]|0;
 $310 = ($309|0)==(0);
 L185: do {
  if (!($310)) {
   $311 = ((($savedMove$i)) + 4|0);
   $336 = $309;$i$010$i = 0;
   while(1) {
    $$idx4$i = (((($hintmap)) + 28|0) + (($i$010$i*20)|0)|0);
    $$idx4$val$i = HEAP32[$$idx4$i>>2]|0;
    $312 = $$idx4$val$i & 12;
    $313 = ($312|0)!=(0);
    $314 = (($i$010$i) + 1)|0;
    $315 = $313 ? $314 : $i$010$i;
    $$lobit$i5$i = $$idx4$val$i & 16;
    $316 = ($$lobit$i5$i|0)==(0);
    if ($316) {
     $317 = (((((($hintmap)) + 28|0) + (($i$010$i*20)|0)|0)) + 12|0);
     $318 = HEAP32[$317>>2]|0;
     $319 = $318 & 65535;
     $320 = (((((($hintmap)) + 28|0) + (($315*20)|0)|0)) + 12|0);
     $321 = HEAP32[$320>>2]|0;
     $322 = $321 & 65535;
     $323 = (0 - ($319))|0;
     $324 = (0 - ($322))|0;
     $325 = ($319|0)==(0);
     $326 = (65536 - ($319))|0;
     $327 = $325 ? 0 : $326;
     $328 = ($322|0)==(0);
     $329 = (65536 - ($322))|0;
     $330 = $328 ? 0 : $329;
     $331 = ($327>>>0)<($330>>>0);
     $332 = $331 ? $327 : $330;
     $333 = ($323|0)>($324|0);
     $334 = $333 ? $323 : $324;
     $335 = (($336) + -1)|0;
     $337 = ($315>>>0)<($335>>>0);
     do {
      if ($337) {
       $338 = (($315) + 1)|0;
       $339 = (((((($hintmap)) + 28|0) + (($338*20)|0)|0)) + 12|0);
       $340 = HEAP32[$339>>2]|0;
       $341 = (($321) + 32768)|0;
       $342 = (($341) + ($332))|0;
       $343 = ($340|0)<($342|0);
       if ($343) {
        $354 = ($i$010$i|0)==(0);
        if ($354) {
         label = 128;
        } else {
         $355 = (($i$010$i) + -1)|0;
         $356 = (((((($hintmap)) + 28|0) + (($355*20)|0)|0)) + 12|0);
         $357 = HEAP32[$356>>2]|0;
         $358 = (($318) + -32768)|0;
         $359 = (($358) + ($334))|0;
         $360 = ($357|0)>($359|0);
         if ($360) {
          $move$08$i = 0;
         } else {
          label = 128;
         }
        }
        if ((label|0) == 128) {
         label = 0;
         $361 = (0 - ($334))|0;
         $362 = ($332|0)<($361|0);
         if ($362) {
          $move$08$i = $334;
         } else {
          $367 = $318;$move$06$i = $334;
          break;
         }
        }
        $$idx1$i = (((($hintmap)) + 28|0) + (($338*20)|0)|0);
        $$idx1$val$i = HEAP32[$$idx1$i>>2]|0;
        $$lobit$i$i = $$idx1$val$i & 16;
        $363 = ($$lobit$i$i|0)==(0);
        if ($363) {
         HEAP32[$savedMove$i>>2] = $315;
         $364 = (($332) - ($move$08$i))|0;
         HEAP32[$311>>2] = $364;
         $365 = HEAP32[$306>>2]|0;
         _cf2_arrstack_push($365,$savedMove$i);
         $$pre$i = HEAP32[$317>>2]|0;
         $367 = $$pre$i;$move$06$i = $move$08$i;
        } else {
         $367 = $318;$move$06$i = $move$08$i;
        }
       } else {
        label = 123;
       }
      } else {
       label = 123;
      }
     } while(0);
     do {
      if ((label|0) == 123) {
       label = 0;
       $344 = ($i$010$i|0)==(0);
       if (!($344)) {
        $345 = (($i$010$i) + -1)|0;
        $346 = (((((($hintmap)) + 28|0) + (($345*20)|0)|0)) + 12|0);
        $347 = HEAP32[$346>>2]|0;
        $348 = (($318) + -32768)|0;
        $349 = (($348) + ($334))|0;
        $350 = ($347|0)>($349|0);
        if ($350) {
         $367 = $318;$move$06$i = $332;
         break;
        }
       }
       $351 = (0 - ($334))|0;
       $352 = ($332|0)>($351|0);
       $353 = $352 ? $334 : $332;
       $367 = $318;$move$06$i = $353;
      }
     } while(0);
     $366 = (($move$06$i) + ($367))|0;
     HEAP32[$317>>2] = $366;
     if ($313) {
      $368 = HEAP32[$320>>2]|0;
      $369 = (($368) + ($move$06$i))|0;
      HEAP32[$320>>2] = $369;
     }
    }
    $370 = ($i$010$i|0)==(0);
    if (!($370)) {
     $371 = (((((($hintmap)) + 28|0) + (($i$010$i*20)|0)|0)) + 8|0);
     $372 = HEAP32[$371>>2]|0;
     $373 = (($i$010$i) + -1)|0;
     $374 = (((((($hintmap)) + 28|0) + (($373*20)|0)|0)) + 8|0);
     $375 = HEAP32[$374>>2]|0;
     $376 = ($372|0)==($375|0);
     if (!($376)) {
      $377 = (((((($hintmap)) + 28|0) + (($i$010$i*20)|0)|0)) + 12|0);
      $378 = HEAP32[$377>>2]|0;
      $379 = (((((($hintmap)) + 28|0) + (($373*20)|0)|0)) + 12|0);
      $380 = HEAP32[$379>>2]|0;
      $381 = (($378) - ($380))|0;
      $382 = (($372) - ($375))|0;
      $383 = (_FT_DivFix($381,$382)|0);
      $384 = (((((($hintmap)) + 28|0) + (($373*20)|0)|0)) + 16|0);
      HEAP32[$384>>2] = $383;
     }
    }
    if ($313) {
     $385 = (((((($hintmap)) + 28|0) + (($315*20)|0)|0)) + 8|0);
     $386 = HEAP32[$385>>2]|0;
     $387 = (($315) + -1)|0;
     $388 = (((((($hintmap)) + 28|0) + (($387*20)|0)|0)) + 8|0);
     $389 = HEAP32[$388>>2]|0;
     $390 = ($386|0)==($389|0);
     if ($390) {
      $i$1$i = $314;
     } else {
      $391 = (((((($hintmap)) + 28|0) + (($315*20)|0)|0)) + 12|0);
      $392 = HEAP32[$391>>2]|0;
      $393 = (((((($hintmap)) + 28|0) + (($387*20)|0)|0)) + 12|0);
      $394 = HEAP32[$393>>2]|0;
      $395 = (($392) - ($394))|0;
      $396 = (($386) - ($389))|0;
      $397 = (_FT_DivFix($395,$396)|0);
      $398 = (((((($hintmap)) + 28|0) + (($387*20)|0)|0)) + 16|0);
      HEAP32[$398>>2] = $397;
      $i$1$i = $314;
     }
    } else {
     $i$1$i = $i$010$i;
    }
    $399 = (($i$1$i) + 1)|0;
    $400 = HEAP32[$33>>2]|0;
    $401 = ($399>>>0)<($400>>>0);
    if ($401) {
     $336 = $400;$i$010$i = $399;
    } else {
     break;
    }
   }
   $$pre14$i = HEAP32[$306>>2]|0;
   $$idx$phi$trans$insert$i = ((($$pre14$i)) + 20|0);
   $$idx$val$pre$i = HEAP32[$$idx$phi$trans$insert$i>>2]|0;
   $402 = ($$idx$val$pre$i|0)==(0);
   if (!($402)) {
    $405 = $$idx$val$pre$i;$407 = $$pre14$i;$i$29$i = $$idx$val$pre$i;
    while(1) {
     $403 = (($i$29$i) + -1)|0;
     $404 = ($405>>>0)>($403>>>0);
     if ($404) {
      $$0$i$i = $403;
     } else {
      $406 = ((($407)) + 4|0);
      $408 = HEAP32[$406>>2]|0;
      $409 = ($408|0)==(0|0);
      if ($409) {
       $$0$i$i = 0;
      } else {
       $410 = HEAP32[$408>>2]|0;
       $411 = ($410|0)==(0);
       if ($411) {
        HEAP32[$408>>2] = 130;
        $$0$i$i = 0;
       } else {
        $$0$i$i = 0;
       }
      }
     }
     $412 = ((($407)) + 28|0);
     $413 = HEAP32[$412>>2]|0;
     $414 = ((($407)) + 8|0);
     $415 = HEAP32[$414>>2]|0;
     $416 = Math_imul($415, $$0$i$i)|0;
     $417 = (($413) + ($416)|0);
     $418 = HEAP32[$417>>2]|0;
     $419 = (($418) + 1)|0;
     $420 = (((((($hintmap)) + 28|0) + (($419*20)|0)|0)) + 12|0);
     $421 = HEAP32[$420>>2]|0;
     $422 = (((((($hintmap)) + 28|0) + (($418*20)|0)|0)) + 12|0);
     $423 = HEAP32[$422>>2]|0;
     $424 = ((($417)) + 4|0);
     $425 = HEAP32[$424>>2]|0;
     $426 = (($425) + ($423))|0;
     $427 = (($426) + 32768)|0;
     $428 = ($421|0)<($427|0);
     if (!($428)) {
      HEAP32[$422>>2] = $426;
      $$idx3$i = (((($hintmap)) + 28|0) + (($418*20)|0)|0);
      $$idx3$val$i = HEAP32[$$idx3$i>>2]|0;
      $430 = $$idx3$val$i & 12;
      $431 = ($430|0)==(0);
      if (!($431)) {
       $432 = (($418) + -1)|0;
       $433 = (((((($hintmap)) + 28|0) + (($432*20)|0)|0)) + 12|0);
       $434 = HEAP32[$433>>2]|0;
       $435 = (($434) + ($425))|0;
       HEAP32[$433>>2] = $435;
      }
     }
     $429 = ($403|0)==(0);
     if ($429) {
      break L185;
     }
     $$pre16$i = HEAP32[$306>>2]|0;
     $$phi$trans$insert$i = ((($$pre16$i)) + 20|0);
     $$pre17$i = HEAP32[$$phi$trans$insert$i>>2]|0;
     $405 = $$pre17$i;$407 = $$pre16$i;$i$29$i = $403;
    }
   }
  }
 } while(0);
 if (!($1)) {
  $436 = HEAP32[$33>>2]|0;
  $437 = ($436|0)==(0);
  if (!($437)) {
   $438 = ((($hStemHintArray)) + 28|0);
   $439 = ((($hStemHintArray)) + 8|0);
   $440 = ((($hStemHintArray)) + 4|0);
   $466 = $436;$i$264 = 0;
   while(1) {
    $$idx4 = (((($hintmap)) + 28|0) + (($i$264*20)|0)|0);
    $$idx4$val = HEAP32[$$idx4>>2]|0;
    $$lobit$i = $$idx4$val & 32;
    $441 = ($$lobit$i|0)==(0);
    if ($441) {
     $442 = (((((($hintmap)) + 28|0) + (($i$264*20)|0)|0)) + 4|0);
     $443 = HEAP32[$442>>2]|0;
     $444 = HEAP32[$hStemHintArray$idx2>>2]|0;
     $445 = ($444>>>0)>($443>>>0);
     if ($445) {
      $$0$i = $443;$$idx3$val = $$idx4$val;
     } else {
      $446 = HEAP32[$440>>2]|0;
      $447 = ($446|0)==(0|0);
      if ($447) {
       $$0$i = 0;$$idx3$val = $$idx4$val;
      } else {
       $448 = HEAP32[$446>>2]|0;
       $449 = ($448|0)==(0);
       if ($449) {
        HEAP32[$446>>2] = 130;
        $$idx3$val$pre = HEAP32[$$idx4>>2]|0;
        $$0$i = 0;$$idx3$val = $$idx3$val$pre;
       } else {
        $$0$i = 0;$$idx3$val = $$idx4$val;
       }
      }
     }
     $450 = HEAP32[$438>>2]|0;
     $451 = HEAP32[$439>>2]|0;
     $452 = Math_imul($451, $$0$i)|0;
     $453 = (($450) + ($452)|0);
     $454 = $$idx3$val & 10;
     $455 = ($454|0)==(0);
     $456 = (((((($hintmap)) + 28|0) + (($i$264*20)|0)|0)) + 12|0);
     $457 = HEAP32[$456>>2]|0;
     if ($455) {
      $459 = ((($453)) + 12|0);
      HEAP32[$459>>2] = $457;
     } else {
      $458 = ((($453)) + 16|0);
      HEAP32[$458>>2] = $457;
     }
     HEAP8[$453>>0] = 1;
     $$pre = HEAP32[$33>>2]|0;
     $462 = $$pre;
    } else {
     $462 = $466;
    }
    $460 = (($i$264) + 1)|0;
    $461 = ($460>>>0)<($462>>>0);
    if ($461) {
     $466 = $462;$i$264 = $460;
    } else {
     break;
    }
   }
  }
 }
 $463 = ((($hintmap)) + 12|0);
 HEAP8[$463>>0] = 1;
 $464 = ((($hintMask)) + 5|0);
 HEAP8[$464>>0] = 0;
 STACKTOP = sp;return;
}
function _cf2_hintmap_insertHint($hintmap,$bottomHintEdge,$topHintEdge) {
 $hintmap = $hintmap|0;
 $bottomHintEdge = $bottomHintEdge|0;
 $topHintEdge = $topHintEdge|0;
 var $$ = 0, $$idx = 0, $$idx$val = 0, $$idx1 = 0, $$idx1$val = 0, $$lobit$i = 0, $$lobit$i2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $bottomHintEdge$idx$val = 0, $count$06 = 0, $firstHintEdge$0 = 0, $firstHintEdge$0$idx$val = 0, $iDst$05 = 0, $iSrc$07 = 0;
 var $iSrc$07$in = 0, $indexInsert$0$lcssa = 0, $indexInsert$08 = 0, $isPair$0 = 0, $not$ = 0, $topHintEdge$idx$val = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $bottomHintEdge$idx$val = HEAP32[$bottomHintEdge>>2]|0;
 $0 = ($bottomHintEdge$idx$val|0)==(0);
 if ($0) {
  $firstHintEdge$0 = $topHintEdge;$isPair$0 = 0;
 } else {
  $topHintEdge$idx$val = HEAP32[$topHintEdge>>2]|0;
  $not$ = ($topHintEdge$idx$val|0)!=(0);
  $$ = $not$&1;
  $firstHintEdge$0 = $bottomHintEdge;$isPair$0 = $$;
 }
 $1 = ((($hintmap)) + 20|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0);
 L4: do {
  if ($3) {
   $indexInsert$0$lcssa = 0;
  } else {
   $4 = ((($firstHintEdge$0)) + 8|0);
   $5 = HEAP32[$4>>2]|0;
   $indexInsert$08 = 0;
   while(1) {
    $6 = (((((($hintmap)) + 28|0) + (($indexInsert$08*20)|0)|0)) + 8|0);
    $7 = HEAP32[$6>>2]|0;
    $8 = ($7|0)<($5|0);
    if (!($8)) {
     $indexInsert$0$lcssa = $indexInsert$08;
     break L4;
    }
    $9 = (($indexInsert$08) + 1)|0;
    $10 = ($9>>>0)<($2>>>0);
    if ($10) {
     $indexInsert$08 = $9;
    } else {
     $indexInsert$0$lcssa = $9;
     break;
    }
   }
  }
 } while(0);
 $11 = ($indexInsert$0$lcssa>>>0)<($2>>>0);
 if ($11) {
  $12 = (((((($hintmap)) + 28|0) + (($indexInsert$0$lcssa*20)|0)|0)) + 8|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ((($firstHintEdge$0)) + 8|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ($13|0)==($15|0);
  if ($16) {
   return;
  }
  $17 = ($isPair$0<<24>>24)==(0);
  if (!($17)) {
   $18 = ((($topHintEdge)) + 8|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ($13|0)>($19|0);
   if (!($20)) {
    return;
   }
  }
  $$idx1 = (((($hintmap)) + 28|0) + (($indexInsert$0$lcssa*20)|0)|0);
  $$idx1$val = HEAP32[$$idx1>>2]|0;
  $$lobit$i = $$idx1$val & 8;
  $21 = ($$lobit$i|0)==(0);
  if (!($21)) {
   return;
  }
 }
 $22 = ((($hintmap)) + 4|0);
 $23 = HEAP32[$22>>2]|0;
 $$idx = ((($23)) + 12|0);
 $$idx$val = HEAP8[$$idx>>0]|0;
 $24 = ($$idx$val<<24>>24)==(0);
 do {
  if (!($24)) {
   $firstHintEdge$0$idx$val = HEAP32[$firstHintEdge$0>>2]|0;
   $$lobit$i2 = $firstHintEdge$0$idx$val & 16;
   $25 = ($$lobit$i2|0)==(0);
   if ($25) {
    $26 = ($isPair$0<<24>>24)==(0);
    if ($26) {
     $45 = ((($firstHintEdge$0)) + 8|0);
     $46 = HEAP32[$45>>2]|0;
     $47 = (_cf2_hintmap_map($23,$46)|0);
     $48 = ((($firstHintEdge$0)) + 12|0);
     HEAP32[$48>>2] = $47;
     break;
    } else {
     $27 = ((($topHintEdge)) + 8|0);
     $28 = HEAP32[$27>>2]|0;
     $29 = ((($firstHintEdge$0)) + 8|0);
     $30 = HEAP32[$29>>2]|0;
     $31 = (($30) + ($28))|0;
     $32 = (($31|0) / 2)&-1;
     $33 = (_cf2_hintmap_map($23,$32)|0);
     $34 = HEAP32[$27>>2]|0;
     $35 = HEAP32[$29>>2]|0;
     $36 = (($34) - ($35))|0;
     $37 = (($36|0) / 2)&-1;
     $38 = ((($hintmap)) + 16|0);
     $39 = HEAP32[$38>>2]|0;
     $40 = (_FT_MulFix($37,$39)|0);
     $41 = (($33) - ($40))|0;
     $42 = ((($firstHintEdge$0)) + 12|0);
     HEAP32[$42>>2] = $41;
     $43 = (($40) + ($33))|0;
     $44 = ((($topHintEdge)) + 12|0);
     HEAP32[$44>>2] = $43;
     break;
    }
   }
  }
 } while(0);
 $49 = ($indexInsert$0$lcssa|0)==(0);
 if (!($49)) {
  $50 = ((($firstHintEdge$0)) + 12|0);
  $51 = HEAP32[$50>>2]|0;
  $52 = (($indexInsert$0$lcssa) + -1)|0;
  $53 = (((((($hintmap)) + 28|0) + (($52*20)|0)|0)) + 12|0);
  $54 = HEAP32[$53>>2]|0;
  $55 = ($51|0)<($54|0);
  if ($55) {
   return;
  }
 }
 $56 = HEAP32[$1>>2]|0;
 $57 = ($56>>>0)>($indexInsert$0$lcssa>>>0);
 do {
  if ($57) {
   $58 = ($isPair$0<<24>>24)==(0);
   if ($58) {
    $64 = ((($firstHintEdge$0)) + 12|0);
    $65 = HEAP32[$64>>2]|0;
    $66 = (((((($hintmap)) + 28|0) + (($indexInsert$0$lcssa*20)|0)|0)) + 12|0);
    $67 = HEAP32[$66>>2]|0;
    $68 = ($65|0)>($67|0);
    if (!($68)) {
     break;
    }
    return;
   } else {
    $59 = ((($topHintEdge)) + 12|0);
    $60 = HEAP32[$59>>2]|0;
    $61 = (((((($hintmap)) + 28|0) + (($indexInsert$0$lcssa*20)|0)|0)) + 12|0);
    $62 = HEAP32[$61>>2]|0;
    $63 = ($60|0)>($62|0);
    if (!($63)) {
     break;
    }
    return;
   }
  }
 } while(0);
 $69 = $isPair$0&255;
 $70 = (($56) + ($69))|0;
 $71 = ($70>>>0)>(191);
 if ($71) {
  return;
 }
 $72 = ($56|0)==($indexInsert$0$lcssa|0);
 if (!($72)) {
  $73 = (($56) - ($indexInsert$0$lcssa))|0;
  $count$06 = $73;$iDst$05 = $70;$iSrc$07$in = $56;
  while(1) {
   $iSrc$07 = (($iSrc$07$in) + -1)|0;
   $74 = (($count$06) + -1)|0;
   $75 = (($iDst$05) + -1)|0;
   $76 = (((($hintmap)) + 28|0) + (($iDst$05*20)|0)|0);
   $77 = (((($hintmap)) + 28|0) + (($iSrc$07*20)|0)|0);
   ;HEAP32[$76>>2]=HEAP32[$77>>2]|0;HEAP32[$76+4>>2]=HEAP32[$77+4>>2]|0;HEAP32[$76+8>>2]=HEAP32[$77+8>>2]|0;HEAP32[$76+12>>2]=HEAP32[$77+12>>2]|0;HEAP32[$76+16>>2]=HEAP32[$77+16>>2]|0;
   $78 = ($74|0)==(0);
   if ($78) {
    break;
   } else {
    $count$06 = $74;$iDst$05 = $75;$iSrc$07$in = $iSrc$07;
   }
  }
 }
 $79 = (((($hintmap)) + 28|0) + (($indexInsert$0$lcssa*20)|0)|0);
 ;HEAP32[$79>>2]=HEAP32[$firstHintEdge$0>>2]|0;HEAP32[$79+4>>2]=HEAP32[$firstHintEdge$0+4>>2]|0;HEAP32[$79+8>>2]=HEAP32[$firstHintEdge$0+8>>2]|0;HEAP32[$79+12>>2]=HEAP32[$firstHintEdge$0+12>>2]|0;HEAP32[$79+16>>2]=HEAP32[$firstHintEdge$0+16>>2]|0;
 $80 = HEAP32[$1>>2]|0;
 $81 = (($80) + 1)|0;
 HEAP32[$1>>2] = $81;
 $82 = ($isPair$0<<24>>24)==(0);
 if ($82) {
  return;
 }
 $83 = (($indexInsert$0$lcssa) + 1)|0;
 $84 = (((($hintmap)) + 28|0) + (($83*20)|0)|0);
 ;HEAP32[$84>>2]=HEAP32[$topHintEdge>>2]|0;HEAP32[$84+4>>2]=HEAP32[$topHintEdge+4>>2]|0;HEAP32[$84+8>>2]=HEAP32[$topHintEdge+8>>2]|0;HEAP32[$84+12>>2]=HEAP32[$topHintEdge+12>>2]|0;HEAP32[$84+16>>2]=HEAP32[$topHintEdge+16>>2]|0;
 $85 = HEAP32[$1>>2]|0;
 $86 = (($85) + 1)|0;
 HEAP32[$1>>2] = $86;
 return;
}
function _cf2_glyphpath_curveTo($glyphpath,$x1,$y1,$x2,$y2,$x3,$y3) {
 $glyphpath = $glyphpath|0;
 $x1 = $x1|0;
 $y1 = $y1|0;
 $x2 = $x2|0;
 $y2 = $y2|0;
 $x3 = $x3|0;
 $y3 = $y3|0;
 var $$idx = 0, $$idx$val = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $8 = 0, $9 = 0, $P0 = 0;
 var $xOffset1 = 0, $xOffset3 = 0, $yOffset1 = 0, $yOffset3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $xOffset1 = sp + 20|0;
 $yOffset1 = sp + 16|0;
 $xOffset3 = sp + 12|0;
 $yOffset3 = sp + 8|0;
 $P0 = sp;
 $0 = ((($glyphpath)) + 11720|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($glyphpath)) + 11724|0);
 $3 = HEAP32[$2>>2]|0;
 _cf2_glyphpath_computeOffset($glyphpath,$1,$3,$x1,$y1,$xOffset1,$yOffset1);
 _cf2_glyphpath_computeOffset($glyphpath,$x2,$y2,$x3,$y3,$xOffset3,$yOffset3);
 $4 = $x1 >> 16;
 $5 = (($y2) - ($y1))|0;
 $6 = $5 >> 16;
 $7 = Math_imul($6, $4)|0;
 $8 = $y1 >> 16;
 $9 = (($x2) - ($x1))|0;
 $10 = $9 >> 16;
 $11 = Math_imul($10, $8)|0;
 $12 = (($7) - ($11))|0;
 $13 = ((($glyphpath)) + 4|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($14)) + 16|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = (($12) + ($16))|0;
 HEAP32[$15>>2] = $17;
 $18 = HEAP32[$0>>2]|0;
 $19 = HEAP32[$xOffset1>>2]|0;
 $20 = (($19) + ($18))|0;
 HEAP32[$P0>>2] = $20;
 $21 = HEAP32[$2>>2]|0;
 $22 = HEAP32[$yOffset1>>2]|0;
 $23 = (($22) + ($21))|0;
 $24 = ((($P0)) + 4|0);
 HEAP32[$24>>2] = $23;
 $25 = (($19) + ($x1))|0;
 $26 = (($22) + ($y1))|0;
 $27 = HEAP32[$xOffset3>>2]|0;
 $28 = (($27) + ($x2))|0;
 $29 = HEAP32[$yOffset3>>2]|0;
 $30 = (($29) + ($y2))|0;
 $31 = (($27) + ($x3))|0;
 $32 = (($29) + ($y3))|0;
 $33 = ((($glyphpath)) + 11667|0);
 $34 = HEAP8[$33>>0]|0;
 $35 = ($34<<24>>24)==(0);
 if (!($35)) {
  _cf2_glyphpath_pushMove($glyphpath,$20,$23);
  HEAP8[$33>>0] = 0;
  $36 = ((($glyphpath)) + 11664|0);
  HEAP8[$36>>0] = 1;
  $37 = ((($glyphpath)) + 11712|0);
  $38 = $37;
  $39 = $38;
  HEAP32[$39>>2] = $25;
  $40 = (($38) + 4)|0;
  $41 = $40;
  HEAP32[$41>>2] = $26;
 }
 $42 = ((($glyphpath)) + 11744|0);
 $43 = HEAP8[$42>>0]|0;
 $44 = ($43<<24>>24)==(0);
 if (!($44)) {
  $45 = ((($glyphpath)) + 8|0);
  _cf2_glyphpath_pushPrevElem($glyphpath,$45,$P0,$25,$26,0);
 }
 HEAP8[$42>>0] = 1;
 $46 = ((($glyphpath)) + 11748|0);
 HEAP32[$46>>2] = 4;
 $47 = ((($glyphpath)) + 11752|0);
 $48 = $P0;
 $49 = $48;
 $50 = HEAP32[$49>>2]|0;
 $51 = (($48) + 4)|0;
 $52 = $51;
 $53 = HEAP32[$52>>2]|0;
 $54 = $47;
 $55 = $54;
 HEAP32[$55>>2] = $50;
 $56 = (($54) + 4)|0;
 $57 = $56;
 HEAP32[$57>>2] = $53;
 $58 = ((($glyphpath)) + 11760|0);
 $59 = $58;
 $60 = $59;
 HEAP32[$60>>2] = $25;
 $61 = (($59) + 4)|0;
 $62 = $61;
 HEAP32[$62>>2] = $26;
 $63 = ((($glyphpath)) + 11768|0);
 HEAP32[$63>>2] = $28;
 $64 = ((($glyphpath)) + 11772|0);
 HEAP32[$64>>2] = $30;
 $65 = ((($glyphpath)) + 11776|0);
 HEAP32[$65>>2] = $31;
 $66 = ((($glyphpath)) + 11780|0);
 HEAP32[$66>>2] = $32;
 $67 = ((($glyphpath)) + 11676|0);
 $68 = HEAP32[$67>>2]|0;
 $$idx = ((($68)) + 5|0);
 $$idx$val = HEAP8[$$idx>>0]|0;
 $69 = ($$idx$val<<24>>24)==(0);
 if ($69) {
  HEAP32[$0>>2] = $x3;
  HEAP32[$2>>2] = $y3;
  STACKTOP = sp;return;
 }
 $70 = ((($glyphpath)) + 8|0);
 $71 = ((($glyphpath)) + 11668|0);
 $72 = HEAP32[$71>>2]|0;
 $73 = ((($glyphpath)) + 11672|0);
 $74 = HEAP32[$73>>2]|0;
 $75 = ((($glyphpath)) + 11680|0);
 $76 = HEAP32[$75>>2]|0;
 _cf2_hintmap_build($70,$72,$74,$68,$76,0);
 HEAP32[$0>>2] = $x3;
 HEAP32[$2>>2] = $y3;
 STACKTOP = sp;return;
}
function _cf2_doFlex($opStack,$curX,$curY,$glyphPath,$readFromStack,$doConditionalLastRead) {
 $opStack = $opStack|0;
 $curX = $curX|0;
 $curY = $curY|0;
 $glyphPath = $glyphPath|0;
 $readFromStack = $readFromStack|0;
 $doConditionalLastRead = $doConditionalLastRead|0;
 var $$0$i = 0, $$0$i1 = 0, $$0$i3 = 0, $$0$i5 = 0, $$phi$trans$insert = 0, $$phi$trans$insert12$phi$trans$insert = 0, $$pre = 0, $$pre$phiZ2D = 0, $$pre11 = 0, $$pre13 = 0, $$pre13$pre = 0, $$pre15 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$08 = 0, $index$09 = 0, $index$1 = 0, $index$1$lcssa = 0, $index$2 = 0, $vals = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vals = sp;
 $0 = HEAP32[$curX>>2]|0;
 HEAP32[$vals>>2] = $0;
 $1 = HEAP32[$curY>>2]|0;
 $2 = ((($vals)) + 4|0);
 HEAP32[$2>>2] = $1;
 $3 = ((($readFromStack)) + 9|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)==(0);
 $6 = $5 ? 9 : 10;
 $7 = ((($opStack)) + 392|0);
 $8 = ((($opStack)) + 8|0);
 $9 = $8;
 $10 = ((($opStack)) + 4|0);
 $13 = $0;$i$08 = 0;$index$09 = 0;
 while(1) {
  $11 = (($i$08) + 2)|0;
  $12 = (($vals) + ($11<<2)|0);
  HEAP32[$12>>2] = $13;
  $14 = (($readFromStack) + ($i$08)|0);
  $15 = HEAP8[$14>>0]|0;
  $16 = ($15<<24>>24)==(0);
  if ($16) {
   $index$1 = $index$09;
  } else {
   $17 = (($index$09) + 1)|0;
   $18 = HEAP32[$7>>2]|0;
   $19 = (($18) - ($9))|0;
   $20 = $19 >> 3;
   $21 = ($20>>>0)>($index$09>>>0);
   L5: do {
    if ($21) {
     $26 = (((($opStack)) + 8|0) + ($index$09<<3)|0);
     $27 = (((((($opStack)) + 8|0) + ($index$09<<3)|0)) + 4|0);
     $28 = HEAP32[$27>>2]|0;
     switch ($28|0) {
     case 2:  {
      $29 = HEAP32[$26>>2]|0;
      $30 = $29 << 16;
      $$0$i = $30;$40 = $13;
      break L5;
      break;
     }
     case 1:  {
      $31 = HEAP32[$26>>2]|0;
      $32 = ($31|0)<(0);
      if ($32) {
       $33 = (8192 - ($31))|0;
       $34 = $33 >> 14;
       $35 = (0 - ($34))|0;
       $$0$i = $35;$40 = $13;
       break L5;
      } else {
       $36 = (($31) + 8192)|0;
       $37 = $36 >> 14;
       $$0$i = $37;$40 = $13;
       break L5;
      }
      break;
     }
     default: {
      $38 = HEAP32[$26>>2]|0;
      $$0$i = $38;$40 = $13;
      break L5;
     }
     }
    } else {
     $22 = HEAP32[$10>>2]|0;
     $23 = ($22|0)==(0|0);
     if ($23) {
      $$0$i = 0;$40 = $13;
     } else {
      $24 = HEAP32[$22>>2]|0;
      $25 = ($24|0)==(0);
      if ($25) {
       HEAP32[$22>>2] = 130;
       $$pre11 = HEAP32[$12>>2]|0;
       $$0$i = 0;$40 = $$pre11;
      } else {
       $$0$i = 0;$40 = $13;
      }
     }
    }
   } while(0);
   $39 = (($40) + ($$0$i))|0;
   HEAP32[$12>>2] = $39;
   $index$1 = $17;
  }
  $41 = (($i$08) + 1)|0;
  $42 = ($41|0)<($6|0);
  if (!($42)) {
   $index$1$lcssa = $index$1;
   break;
  }
  $$phi$trans$insert = (($vals) + ($41<<2)|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $13 = $$pre;$i$08 = $41;$index$09 = $index$1;
 }
 if ($5) {
  $43 = HEAP32[$curY>>2]|0;
  $44 = ((($vals)) + 44|0);
  HEAP32[$44>>2] = $43;
 }
 $45 = ($doConditionalLastRead<<24>>24)==(0);
 do {
  if ($45) {
   $92 = ((($readFromStack)) + 10|0);
   $93 = HEAP8[$92>>0]|0;
   $94 = ($93<<24>>24)==(0);
   if ($94) {
    $121 = HEAP32[$curX>>2]|0;
    $122 = ((($vals)) + 48|0);
    HEAP32[$122>>2] = $121;
    $174 = $121;$index$2 = $index$1$lcssa;
   } else {
    $95 = ((($vals)) + 40|0);
    $96 = HEAP32[$95>>2]|0;
    $97 = (($index$1$lcssa) + 1)|0;
    $98 = HEAP32[$7>>2]|0;
    $99 = (($98) - ($9))|0;
    $100 = $99 >> 3;
    $101 = ($100>>>0)>($index$1$lcssa>>>0);
    L29: do {
     if ($101) {
      $106 = (((($opStack)) + 8|0) + ($index$1$lcssa<<3)|0);
      $107 = (((((($opStack)) + 8|0) + ($index$1$lcssa<<3)|0)) + 4|0);
      $108 = HEAP32[$107>>2]|0;
      switch ($108|0) {
      case 2:  {
       $109 = HEAP32[$106>>2]|0;
       $110 = $109 << 16;
       $$0$i5 = $110;
       break L29;
       break;
      }
      case 1:  {
       $111 = HEAP32[$106>>2]|0;
       $112 = ($111|0)<(0);
       if ($112) {
        $113 = (8192 - ($111))|0;
        $114 = $113 >> 14;
        $115 = (0 - ($114))|0;
        $$0$i5 = $115;
        break L29;
       } else {
        $116 = (($111) + 8192)|0;
        $117 = $116 >> 14;
        $$0$i5 = $117;
        break L29;
       }
       break;
      }
      default: {
       $118 = HEAP32[$106>>2]|0;
       $$0$i5 = $118;
       break L29;
      }
      }
     } else {
      $102 = HEAP32[$10>>2]|0;
      $103 = ($102|0)==(0|0);
      if ($103) {
       $$0$i5 = 0;
      } else {
       $104 = HEAP32[$102>>2]|0;
       $105 = ($104|0)==(0);
       if ($105) {
        HEAP32[$102>>2] = 130;
        $$0$i5 = 0;
       } else {
        $$0$i5 = 0;
       }
      }
     }
    } while(0);
    $119 = (($$0$i5) + ($96))|0;
    $120 = ((($vals)) + 48|0);
    HEAP32[$120>>2] = $119;
    $174 = $119;$index$2 = $97;
   }
   $123 = ((($readFromStack)) + 11|0);
   $124 = HEAP8[$123>>0]|0;
   $125 = ($124<<24>>24)==(0);
   if ($125) {
    $151 = HEAP32[$curY>>2]|0;
    $152 = ((($vals)) + 52|0);
    HEAP32[$152>>2] = $151;
    $$pre15 = ((($vals)) + 44|0);
    $$pre$phiZ2D = $$pre15;$172 = $174;$173 = $151;
    break;
   }
   $126 = ((($vals)) + 44|0);
   $127 = HEAP32[$126>>2]|0;
   $128 = HEAP32[$7>>2]|0;
   $129 = (($128) - ($9))|0;
   $130 = $129 >> 3;
   $131 = ($130>>>0)>($index$2>>>0);
   L46: do {
    if ($131) {
     $136 = (((($opStack)) + 8|0) + ($index$2<<3)|0);
     $137 = (((((($opStack)) + 8|0) + ($index$2<<3)|0)) + 4|0);
     $138 = HEAP32[$137>>2]|0;
     switch ($138|0) {
     case 2:  {
      $139 = HEAP32[$136>>2]|0;
      $140 = $139 << 16;
      $$0$i3 = $140;$$pre13 = $174;
      break L46;
      break;
     }
     case 1:  {
      $141 = HEAP32[$136>>2]|0;
      $142 = ($141|0)<(0);
      if ($142) {
       $143 = (8192 - ($141))|0;
       $144 = $143 >> 14;
       $145 = (0 - ($144))|0;
       $$0$i3 = $145;$$pre13 = $174;
       break L46;
      } else {
       $146 = (($141) + 8192)|0;
       $147 = $146 >> 14;
       $$0$i3 = $147;$$pre13 = $174;
       break L46;
      }
      break;
     }
     default: {
      $148 = HEAP32[$136>>2]|0;
      $$0$i3 = $148;$$pre13 = $174;
      break L46;
     }
     }
    } else {
     $132 = HEAP32[$10>>2]|0;
     $133 = ($132|0)==(0|0);
     if ($133) {
      $$0$i3 = 0;$$pre13 = $174;
     } else {
      $134 = HEAP32[$132>>2]|0;
      $135 = ($134|0)==(0);
      if ($135) {
       HEAP32[$132>>2] = 130;
       $$phi$trans$insert12$phi$trans$insert = ((($vals)) + 48|0);
       $$pre13$pre = HEAP32[$$phi$trans$insert12$phi$trans$insert>>2]|0;
       $$0$i3 = 0;$$pre13 = $$pre13$pre;
      } else {
       $$0$i3 = 0;$$pre13 = $174;
      }
     }
    }
   } while(0);
   $149 = (($$0$i3) + ($127))|0;
   $150 = ((($vals)) + 52|0);
   HEAP32[$150>>2] = $149;
   $$pre$phiZ2D = $126;$172 = $$pre13;$173 = $149;
  } else {
   $46 = ((($vals)) + 40|0);
   $47 = HEAP32[$46>>2]|0;
   $48 = HEAP32[$curX>>2]|0;
   $49 = (($47) - ($48))|0;
   $50 = ($49|0)<(0);
   $51 = (0 - ($49))|0;
   $52 = $50 ? $51 : $49;
   $53 = ((($vals)) + 44|0);
   $54 = HEAP32[$53>>2]|0;
   $55 = HEAP32[$curY>>2]|0;
   $56 = (($54) - ($55))|0;
   $57 = ($56|0)<(0);
   $58 = (0 - ($56))|0;
   $59 = $57 ? $58 : $56;
   $60 = ($52|0)>($59|0);
   $61 = HEAP32[$7>>2]|0;
   $62 = (($61) - ($9))|0;
   $63 = $62 >> 3;
   $64 = ($63>>>0)>($index$1$lcssa>>>0);
   L60: do {
    if ($64) {
     $69 = (((($opStack)) + 8|0) + ($index$1$lcssa<<3)|0);
     $70 = (((((($opStack)) + 8|0) + ($index$1$lcssa<<3)|0)) + 4|0);
     $71 = HEAP32[$70>>2]|0;
     switch ($71|0) {
     case 2:  {
      $72 = HEAP32[$69>>2]|0;
      $73 = $72 << 16;
      $$0$i1 = $73;
      break L60;
      break;
     }
     case 1:  {
      $74 = HEAP32[$69>>2]|0;
      $75 = ($74|0)<(0);
      if ($75) {
       $76 = (8192 - ($74))|0;
       $77 = $76 >> 14;
       $78 = (0 - ($77))|0;
       $$0$i1 = $78;
       break L60;
      } else {
       $79 = (($74) + 8192)|0;
       $80 = $79 >> 14;
       $$0$i1 = $80;
       break L60;
      }
      break;
     }
     default: {
      $81 = HEAP32[$69>>2]|0;
      $$0$i1 = $81;
      break L60;
     }
     }
    } else {
     $65 = HEAP32[$10>>2]|0;
     $66 = ($65|0)==(0|0);
     if ($66) {
      $$0$i1 = 0;
     } else {
      $67 = HEAP32[$65>>2]|0;
      $68 = ($67|0)==(0);
      if ($68) {
       HEAP32[$65>>2] = 130;
       $$0$i1 = 0;
      } else {
       $$0$i1 = 0;
      }
     }
    }
   } while(0);
   if ($60) {
    $82 = HEAP32[$46>>2]|0;
    $83 = (($82) + ($$0$i1))|0;
    $84 = ((($vals)) + 48|0);
    HEAP32[$84>>2] = $83;
    $85 = HEAP32[$curY>>2]|0;
    $86 = ((($vals)) + 52|0);
    HEAP32[$86>>2] = $85;
    $$pre$phiZ2D = $53;$172 = $83;$173 = $85;
    break;
   } else {
    $87 = HEAP32[$curX>>2]|0;
    $88 = ((($vals)) + 48|0);
    HEAP32[$88>>2] = $87;
    $89 = HEAP32[$53>>2]|0;
    $90 = (($89) + ($$0$i1))|0;
    $91 = ((($vals)) + 52|0);
    HEAP32[$91>>2] = $90;
    $$pre$phiZ2D = $53;$172 = $87;$173 = $90;
    break;
   }
  }
 } while(0);
 $153 = ((($vals)) + 8|0);
 $154 = HEAP32[$153>>2]|0;
 $155 = ((($vals)) + 12|0);
 $156 = HEAP32[$155>>2]|0;
 $157 = ((($vals)) + 16|0);
 $158 = HEAP32[$157>>2]|0;
 $159 = ((($vals)) + 20|0);
 $160 = HEAP32[$159>>2]|0;
 $161 = ((($vals)) + 24|0);
 $162 = HEAP32[$161>>2]|0;
 $163 = ((($vals)) + 28|0);
 $164 = HEAP32[$163>>2]|0;
 _cf2_glyphpath_curveTo($glyphPath,$154,$156,$158,$160,$162,$164);
 $165 = ((($vals)) + 32|0);
 $166 = HEAP32[$165>>2]|0;
 $167 = ((($vals)) + 36|0);
 $168 = HEAP32[$167>>2]|0;
 $169 = ((($vals)) + 40|0);
 $170 = HEAP32[$169>>2]|0;
 $171 = HEAP32[$$pre$phiZ2D>>2]|0;
 _cf2_glyphpath_curveTo($glyphPath,$166,$168,$170,$171,$172,$173);
 HEAP32[$7>>2] = $8;
 HEAP32[$curX>>2] = $172;
 HEAP32[$curY>>2] = $173;
 STACKTOP = sp;return;
}
function _cff_get_kerning($ttface,$left_glyph,$right_glyph,$kerning) {
 $ttface = $ttface|0;
 $left_glyph = $left_glyph|0;
 $right_glyph = $right_glyph|0;
 $kerning = $kerning|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($ttface)) + 532|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$kerning>>2] = 0;
 $2 = ((($kerning)) + 4|0);
 HEAP32[$2>>2] = 0;
 $3 = ($1|0)==(0|0);
 if ($3) {
  return 0;
 }
 $4 = ((($1)) + 84|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiii[$5 & 8191]($ttface,$left_glyph,$right_glyph)|0);
 HEAP32[$kerning>>2] = $6;
 return 0;
}
function _cff_get_advances($face,$start,$count,$flags,$advances) {
 $face = $face|0;
 $start = $start|0;
 $count = $count|0;
 $flags = $flags|0;
 $advances = $advances|0;
 var $$1 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ah = 0, $aw = 0, $dummy = 0, $exitcond = 0, $exitcond14 = 0, $nn$011 = 0, $nn$19 = 0, $nn$26 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $dummy = sp + 4|0;
 $ah = sp + 2|0;
 $aw = sp;
 $0 = ((($face)) + 84|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 8;
 $5 = ($4|0)==(0);
 do {
  if (!($5)) {
   $6 = $flags & 16;
   $7 = ($6|0)==(0);
   if ($7) {
    $21 = ((($face)) + 216|0);
    $22 = ((($21)) + 34|0);
    $23 = HEAP16[$22>>1]|0;
    $24 = ($23<<16>>16)==(0);
    if ($24) {
     break;
    }
    $25 = ($count|0)==(0);
    if (!($25)) {
     $26 = ((($face)) + 532|0);
     $nn$19 = 0;
     while(1) {
      $27 = HEAP32[$26>>2]|0;
      $28 = ((($27)) + 112|0);
      $29 = HEAP32[$28>>2]|0;
      $30 = (($nn$19) + ($start))|0;
      FUNCTION_TABLE_viiiii[$29 & 8191]($face,0,$30,$dummy,$aw);
      $31 = HEAP16[$aw>>1]|0;
      $32 = $31&65535;
      $33 = (($advances) + ($nn$19<<2)|0);
      HEAP32[$33>>2] = $32;
      $34 = (($nn$19) + 1)|0;
      $exitcond = ($34|0)==($count|0);
      if ($exitcond) {
       break;
      } else {
       $nn$19 = $34;
      }
     }
    }
   } else {
    $8 = ((($face)) + 292|0);
    $9 = HEAP8[$8>>0]|0;
    $10 = ($9<<24>>24)==(0);
    if ($10) {
     break;
    }
    $11 = ($count|0)==(0);
    if (!($11)) {
     $12 = ((($face)) + 532|0);
     $nn$011 = 0;
     while(1) {
      $13 = HEAP32[$12>>2]|0;
      $14 = ((($13)) + 112|0);
      $15 = HEAP32[$14>>2]|0;
      $16 = (($nn$011) + ($start))|0;
      FUNCTION_TABLE_viiiii[$15 & 8191]($face,1,$16,$dummy,$ah);
      $17 = HEAP16[$ah>>1]|0;
      $18 = $17&65535;
      $19 = (($advances) + ($nn$011<<2)|0);
      HEAP32[$19>>2] = $18;
      $20 = (($nn$011) + 1)|0;
      $exitcond14 = ($20|0)==($count|0);
      if ($exitcond14) {
       break;
      } else {
       $nn$011 = $20;
      }
     }
    }
   }
   $$1 = 0;
   STACKTOP = sp;return ($$1|0);
  }
 } while(0);
 $35 = $flags | 256;
 $36 = ((($face)) + 88|0);
 $37 = ($count|0)==(0);
 if ($37) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $38 = $flags & 16;
 $39 = ($38|0)==(0);
 $40 = ((($1)) + 60|0);
 $41 = ((($1)) + 56|0);
 $$in = $39 ? $41 : $40;
 $nn$26 = 0;
 while(1) {
  $42 = HEAP32[$36>>2]|0;
  $43 = (($nn$26) + ($start))|0;
  $44 = (_cff_glyph_load($1,$42,$43,$35)|0);
  $45 = ($44|0)==(0);
  if (!($45)) {
   $$1 = $44;
   label = 18;
   break;
  }
  $46 = HEAP32[$$in>>2]|0;
  $47 = (($advances) + ($nn$26<<2)|0);
  HEAP32[$47>>2] = $46;
  $48 = (($nn$26) + 1)|0;
  $49 = ($48>>>0)<($count>>>0);
  if ($49) {
   $nn$26 = $48;
  } else {
   $$1 = 0;
   label = 18;
   break;
  }
 }
 if ((label|0) == 18) {
  STACKTOP = sp;return ($$1|0);
 }
 return (0)|0;
}
function _cff_size_request($size,$req) {
 $size = $size|0;
 $req = $req|0;
 var $$idx$val = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $i$02 = 0, $or$cond$i = 0, $strike_index = 0, $x_scale$0 = 0, $y_scale$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $strike_index = sp;
 $0 = HEAP32[$size>>2]|0;
 $1 = ((($0)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2 & 2;
 $4 = ($3|0)==(0);
 do {
  if ($4) {
   $13 = $0;
  } else {
   $5 = ((($0)) + 532|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ((($6)) + 104|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = (FUNCTION_TABLE_iiii[$8 & 8191]($0,$req,$strike_index)|0);
   $10 = ($9|0)==(0);
   if (!($10)) {
    $11 = ((($size)) + 44|0);
    HEAP32[$11>>2] = -1;
    $$pre = HEAP32[$size>>2]|0;
    $13 = $$pre;
    break;
   }
   $12 = HEAP32[$strike_index>>2]|0;
   (_cff_size_select($size,$12)|0);
   STACKTOP = sp;return 0;
  }
 } while(0);
 _FT_Request_Metrics($13,$req);
 $$idx$val = HEAP32[$size>>2]|0;
 $14 = ((($$idx$val)) + 656|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($15)) + 2952|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ((($$idx$val)) + 96|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ((($19)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = (_FT_Get_Module($21,593918)|0);
 $23 = ($22|0)!=(0|0);
 $24 = ($17|0)!=(0|0);
 $or$cond$i = $24 & $23;
 if (!($or$cond$i)) {
  STACKTOP = sp;return 0;
 }
 $25 = HEAP32[$17>>2]|0;
 $26 = ($25|0)==(0|0);
 if ($26) {
  STACKTOP = sp;return 0;
 }
 $27 = (FUNCTION_TABLE_ii[$25 & 8191]($22)|0);
 $28 = ($27|0)==(0|0);
 if ($28) {
  STACKTOP = sp;return 0;
 }
 $29 = HEAP32[$size>>2]|0;
 $30 = ((($29)) + 656|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = ((($size)) + 40|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($31)) + 1392|0);
 $35 = HEAP32[$34>>2]|0;
 $36 = ((($27)) + 4|0);
 $37 = HEAP32[$36>>2]|0;
 $38 = HEAP32[$33>>2]|0;
 $39 = ((($size)) + 16|0);
 $40 = HEAP32[$39>>2]|0;
 $41 = ((($size)) + 20|0);
 $42 = HEAP32[$41>>2]|0;
 FUNCTION_TABLE_viiiii[$37 & 8191]($38,$40,$42,0,0);
 $43 = ((($31)) + 1896|0);
 $44 = HEAP32[$43>>2]|0;
 $45 = ($44|0)==(0);
 if ($45) {
  STACKTOP = sp;return 0;
 } else {
  $i$02 = $44;
 }
 while(1) {
  $46 = (($i$02) + -1)|0;
  $47 = (((($31)) + 1900|0) + ($46<<2)|0);
  $48 = HEAP32[$47>>2]|0;
  $49 = ((($48)) + 68|0);
  $50 = HEAP32[$49>>2]|0;
  $51 = ($35|0)==($50|0);
  $52 = HEAP32[$39>>2]|0;
  if ($51) {
   $56 = HEAP32[$41>>2]|0;
   $x_scale$0 = $52;$y_scale$0 = $56;
  } else {
   $53 = (_FT_MulDiv($52,$35,$50)|0);
   $54 = HEAP32[$41>>2]|0;
   $55 = (_FT_MulDiv($54,$35,$50)|0);
   $x_scale$0 = $53;$y_scale$0 = $55;
  }
  $57 = HEAP32[$36>>2]|0;
  $58 = (((($33)) + 4|0) + ($46<<2)|0);
  $59 = HEAP32[$58>>2]|0;
  FUNCTION_TABLE_viiiii[$57 & 8191]($59,$x_scale$0,$y_scale$0,0,0);
  $60 = ($46|0)==(0);
  if ($60) {
   break;
  } else {
   $i$02 = $46;
  }
 }
 STACKTOP = sp;return 0;
}
function _cff_size_select($size,$strike_index) {
 $size = $size|0;
 $strike_index = $strike_index|0;
 var $$idx$val = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $or$cond$i = 0, $x_scale$0 = 0, $y_scale$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($size)) + 44|0);
 HEAP32[$0>>2] = $strike_index;
 $1 = HEAP32[$size>>2]|0;
 _FT_Select_Metrics($1,$strike_index);
 $$idx$val = HEAP32[$size>>2]|0;
 $2 = ((($$idx$val)) + 656|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 2952|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($$idx$val)) + 96|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($7)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (_FT_Get_Module($9,593918)|0);
 $11 = ($10|0)!=(0|0);
 $12 = ($5|0)!=(0|0);
 $or$cond$i = $12 & $11;
 if (!($or$cond$i)) {
  return 0;
 }
 $13 = HEAP32[$5>>2]|0;
 $14 = ($13|0)==(0|0);
 if ($14) {
  return 0;
 }
 $15 = (FUNCTION_TABLE_ii[$13 & 8191]($10)|0);
 $16 = ($15|0)==(0|0);
 if ($16) {
  return 0;
 }
 $17 = HEAP32[$size>>2]|0;
 $18 = ((($17)) + 656|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ((($size)) + 40|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ((($19)) + 1392|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = ((($15)) + 4|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = HEAP32[$21>>2]|0;
 $27 = ((($size)) + 16|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ((($size)) + 20|0);
 $30 = HEAP32[$29>>2]|0;
 FUNCTION_TABLE_viiiii[$25 & 8191]($26,$28,$30,0,0);
 $31 = ((($19)) + 1896|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ($32|0)==(0);
 if ($33) {
  return 0;
 } else {
  $i$01 = $32;
 }
 while(1) {
  $34 = (($i$01) + -1)|0;
  $35 = (((($19)) + 1900|0) + ($34<<2)|0);
  $36 = HEAP32[$35>>2]|0;
  $37 = ((($36)) + 68|0);
  $38 = HEAP32[$37>>2]|0;
  $39 = ($23|0)==($38|0);
  $40 = HEAP32[$27>>2]|0;
  if ($39) {
   $44 = HEAP32[$29>>2]|0;
   $x_scale$0 = $40;$y_scale$0 = $44;
  } else {
   $41 = (_FT_MulDiv($40,$23,$38)|0);
   $42 = HEAP32[$29>>2]|0;
   $43 = (_FT_MulDiv($42,$23,$38)|0);
   $x_scale$0 = $41;$y_scale$0 = $43;
  }
  $45 = HEAP32[$24>>2]|0;
  $46 = (((($21)) + 4|0) + ($34<<2)|0);
  $47 = HEAP32[$46>>2]|0;
  FUNCTION_TABLE_viiiii[$45 & 8191]($47,$x_scale$0,$y_scale$0,0,0);
  $48 = ($34|0)==(0);
  if ($48) {
   break;
  } else {
   $i$01 = $34;
  }
 }
 return 0;
}
function _cid_driver_init($driver) {
 $driver = $driver|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _cid_driver_done($driver) {
 $driver = $driver|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _cid_get_interface($module,$cid_interface) {
 $module = $module|0;
 $cid_interface = $cid_interface|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_ft_service_list_lookup(51256,$cid_interface)|0);
 return ($0|0);
}
function _cid_get_postscript_name($face) {
 $face = $face|0;
 var $$ = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 140|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return (0|0);
 } else {
  $3 = HEAP8[$1>>0]|0;
  $4 = ($3<<24>>24)==(47);
  $5 = ((($1)) + 1|0);
  $$ = $4 ? $5 : $1;
  return ($$|0);
 }
 return (0)|0;
}
function _cid_ps_get_font_info($face,$afont_info) {
 $face = $face|0;
 $afont_info = $afont_info|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 140|0);
 $1 = ((($0)) + 24|0);
 ;HEAP32[$afont_info>>2]=HEAP32[$1>>2]|0;HEAP32[$afont_info+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$afont_info+8>>2]=HEAP32[$1+8>>2]|0;HEAP32[$afont_info+12>>2]=HEAP32[$1+12>>2]|0;HEAP32[$afont_info+16>>2]=HEAP32[$1+16>>2]|0;HEAP32[$afont_info+20>>2]=HEAP32[$1+20>>2]|0;HEAP32[$afont_info+24>>2]=HEAP32[$1+24>>2]|0;HEAP32[$afont_info+28>>2]=HEAP32[$1+28>>2]|0;
 return 0;
}
function _cid_ps_get_font_extra($face,$afont_extra) {
 $face = $face|0;
 $afont_extra = $afont_extra|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 312|0);
 $1 = HEAP16[$0>>1]|0;
 HEAP16[$afont_extra>>1] = $1;
 return 0;
}
function _cid_get_ros($face,$registry,$ordering,$supplement) {
 $face = $face|0;
 $registry = $registry|0;
 $ordering = $ordering|0;
 $supplement = $supplement|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($registry|0)==(0|0);
 if (!($0)) {
  $1 = ((($face)) + 152|0);
  $2 = HEAP32[$1>>2]|0;
  HEAP32[$registry>>2] = $2;
 }
 $3 = ($ordering|0)==(0|0);
 if (!($3)) {
  $4 = ((($face)) + 156|0);
  $5 = HEAP32[$4>>2]|0;
  HEAP32[$ordering>>2] = $5;
 }
 $6 = ($supplement|0)==(0|0);
 if ($6) {
  return 0;
 }
 $7 = ((($face)) + 160|0);
 $8 = HEAP32[$7>>2]|0;
 HEAP32[$supplement>>2] = $8;
 return 0;
}
function _cid_get_is_cid($face,$is_cid) {
 $face = $face|0;
 $is_cid = $is_cid|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($is_cid|0)==(0|0);
 if ($0) {
  return 0;
 }
 HEAP8[$is_cid>>0] = 1;
 return 0;
}
function _cid_get_cid_from_glyph_index($face,$glyph_index,$cid) {
 $face = $face|0;
 $glyph_index = $glyph_index|0;
 $cid = $cid|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($cid|0)==(0|0);
 if ($0) {
  return 0;
 }
 HEAP32[$cid>>2] = $glyph_index;
 return 0;
}
function _cid_face_init($stream,$cidface,$face_index,$num_params,$params) {
 $stream = $stream|0;
 $cidface = $cidface|0;
 $face_index = $face_index|0;
 $num_params = $num_params|0;
 $params = $params|0;
 var $$ = 0, $$0$i$ph$i = 0, $$01$i$i$i = 0, $$01$i$i9$i = 0, $$146 = 0, $$9$i$ph$i = 0, $$cast = 0, $$cast$i$i = 0, $$in$i$i = 0, $$lcssa246 = 0, $$lcssa260 = 0, $$lcssa261 = 0, $$lcssa262 = 0, $$off$i$i$i = 0, $$off$us$i$i = 0, $$off2$us$i$i = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i5$i = 0, $$pre$i8$i = 0;
 var $$pre$phi$iZ2D = 0, $$pre39$i$i = 0, $$pre40$i$i = 0, $$pre41$i$i = 0, $$pre95$i = 0, $$pre96$i = 0, $$sink$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0;
 var $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0;
 var $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0;
 var $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0;
 var $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0;
 var $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0;
 var $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0;
 var $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0;
 var $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0;
 var $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0;
 var $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0;
 var $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0;
 var $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0;
 var $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0;
 var $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0;
 var $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0;
 var $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0;
 var $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0;
 var $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0;
 var $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $arg1$043$i$i = 0, $arg1$043$i$i$lcssa278 = 0;
 var $arg1$043$i$i$phi = 0, $arg2$044$i$i = 0, $arg2$044$i$i$lcssa276 = 0, $arg2$044$i$i$phi = 0, $buffer$i$i = 0, $buffer$i6$i = 0, $count$016$i$i = 0, $count$016$i$us$i = 0, $count$117$i$i = 0, $count$221$i$i = 0, $cur$0$i$i = 0, $cur$0$i2$i = 0, $cur$042$i$i = 0, $cur$045$i$i = 0, $cur$135$i$i = 0, $d$0$ph25$i$i = 0, $d$1$i$i = 0, $done$1$i$i = 0, $dummy_object$i$i$i = 0, $error$0 = 0;
 var $error$1$ph$i$i = 0, $error$i = 0, $error$i$i = 0, $exitcond$i$i = 0, $exitcond$i11$i = 0, $exitcond$i11$i51 = 0, $face$idx$val$i = 0, $family$0$ph$be = 0, $family$0$ph50 = 0, $family$047 = 0, $family$047$lcssa = 0, $full$0$ph49 = 0, $full$0$ph49$lcssa244 = 0, $keyword$0$lcssa34$i$i = 0, $keyword$040$i$i = 0, $keyword$040$us$i$i = 0, $loader$i = 0, $max_offsets$027$i$i = 0, $max_offsets$2$i$i = 0, $n$029$i$i = 0;
 var $n$036$us$i$i = 0, $n$114$i$i = 0, $newlimit$0$ph$i$i = 0, $not$ = 0, $object$2$i$i$i = 0, $offset$027$i$i = 0, $offset$027$i$i$lcssa273 = 0, $offset$027$i$i$lcssa274 = 0, $offset$027$i$i275 = 0, $offsets$026$i$i = 0, $offsets$026$i$i$lcssa = 0, $offsets$026$i$i$lcssa247 = 0, $offsets$2$i$i = 0, $offsets$3$ph$i$i = 0, $offsets$4$i$i = 0, $or$cond = 0, $or$cond$i = 0, $or$cond$i$i = 0, $or$cond12$i$i = 0, $or$cond19$i = 0;
 var $p$0$ph26$i$i = 0, $p$016$us$i$i = 0, $p$03$i$i$i = 0, $p$030$i$i = 0, $p$1$lcssa$i$i = 0, $p$126$i$i = 0, $p$126$i$i$lcssa = 0, $p$126$i$i$lcssa269 = 0, $p$2$us$i$i = 0, $p$2$us$i$i$lcssa = 0, $p$2$us$i$i$lcssa256 = 0, $p$2$us$i$i$lcssa257 = 0, $p$2$us$i$i259 = 0, $p$sroa$0$015$i$i = 0, $plimit$0$ph24$i$i = 0, $plimit$015$us$i$i = 0, $plimit$2$us$i$i = 0, $plimit$2$us$i$i$lcssa = 0, $plimit$2$us$i$i$lcssa252 = 0, $plimit$2$us$i$i$lcssa253 = 0;
 var $plimit$2$us$i$i255 = 0, $read_len$029$i$i = 0, $result$02$i$i$i = 0, $scevgep$i$i$i = 0, $sext = 0, $storemerge = 0, $subr$028$i$i = 0, $switch2$i$i$i = 0, $upper_nibble$0$ph27$i$i = 0, $val$0$i$i = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 656|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i$i = sp + 116|0;
 $buffer$i6$i = sp + 392|0;
 $dummy_object$i$i$i = sp + 112|0;
 $buffer$i$i = sp + 120|0;
 $loader$i = sp + 8|0;
 $error$i = sp;
 HEAP32[$cidface>>2] = 1;
 $0 = ((($cidface)) + 136|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 do {
  if ($2) {
   $3 = ((($cidface)) + 96|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ((($4)) + 4|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = (_FT_Get_Module_Interface($6,594378)|0);
   $8 = ($7|0)==(0|0);
   if ($8) {
    $error$0 = 11;
    STACKTOP = sp;return ($error$0|0);
   } else {
    HEAP32[$0>>2] = $7;
    break;
   }
  }
 } while(0);
 $9 = ((($cidface)) + 320|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10|0)==(0|0);
 if ($11) {
  $12 = ((($cidface)) + 96|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ((($13)) + 4|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = (_FT_Get_Module_Interface($15,593918)|0);
  HEAP32[$9>>2] = $16;
 }
 $17 = (_FT_Stream_Seek($stream,0)|0);
 $18 = ($17|0)==(0);
 if (!($18)) {
  $error$0 = $17;
  STACKTOP = sp;return ($error$0|0);
 }
 $19 = ((($cidface)) + 100|0);
 $20 = HEAP32[$19>>2]|0;
 dest=$loader$i; stop=dest+104|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $21 = ((($cidface)) + 104|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = HEAP32[$0>>2]|0;
 dest=$loader$i; stop=dest+100|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $24 = ((($23)) + 4|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = HEAP32[$25>>2]|0;
 FUNCTION_TABLE_viiii[$26 & 8191]($loader$i,0,0,$20);
 $27 = ((($loader$i)) + 72|0);
 HEAP32[$27>>2] = $22;
 $28 = (_FT_Stream_Pos($22)|0);
 $29 = (_FT_Stream_EnterFrame($22,31)|0);
 $30 = ($29|0)==(0);
 L13: do {
  if ($30) {
   $31 = ((($22)) + 32|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = (_strncmp($32,532143,31)|0);
   $34 = ($33|0)==(0);
   _FT_Stream_ExitFrame($22);
   if ($34) {
    $35 = ((($22)) + 4|0);
    $36 = ((($buffer$i$i)) + 10|0);
    $37 = $buffer$i$i;
    $38 = (10 - ($37))|0;
    $39 = ((($loader$i)) + 76|0);
    $40 = ((($loader$i)) + 84|0);
    $41 = ((($loader$i)) + 80|0);
    $42 = ((($loader$i)) + 4|0);
    $43 = ((($loader$i)) + 8|0);
    $44 = ((($loader$i)) + 96|0);
    $45 = ((($loader$i)) + 32|0);
    $46 = ((($loader$i)) + 28|0);
    $47 = ((($loader$i)) + 12|0);
    $48 = (7 - ($37))|0;
    L16: while(1) {
     $49 = (_FT_Stream_Pos($22)|0);
     $50 = HEAP32[$35>>2]|0;
     $51 = (_FT_Stream_Pos($22)|0);
     $52 = ($50|0)==($51|0);
     if ($52) {
      $error$1$ph$i$i = 3;
      label = 25;
      break;
     }
     $53 = (($50) - ($51))|0;
     $55 = $53;$offset$027$i$i = $49;$p$030$i$i = $buffer$i$i;$read_len$029$i$i = 266;
     L19: while(1) {
      $54 = ($read_len$029$i$i>>>0)<($55>>>0);
      $56 = $54 ? $read_len$029$i$i : $55;
      $57 = (_FT_Stream_Read($22,$p$030$i$i,$56)|0);
      $58 = ($57|0)==(0);
      if (!($58)) {
       $error$1$ph$i$i = $57;
       label = 25;
       break L16;
      }
      $59 = ($56>>>0)<(256);
      if ($59) {
       $60 = (($p$030$i$i) + ($56)|0);
       HEAP8[$60>>0] = 0;
      }
      $61 = (($p$030$i$i) + ($56)|0);
      $62 = ((($61)) + -10|0);
      $63 = ($buffer$i$i>>>0)<($62>>>0);
      if ($63) {
       $$pre$i$i = HEAP8[$buffer$i$i>>0]|0;
       $65 = $$pre$i$i;$p$126$i$i = $buffer$i$i;
       while(1) {
        $64 = ($65<<24>>24)==(83);
        if ($64) {
         $66 = (_strncmp($p$126$i$i,532175,9)|0);
         $67 = ($66|0)==(0);
         if ($67) {
          $offset$027$i$i$lcssa273 = $offset$027$i$i;$p$126$i$i$lcssa = $p$126$i$i;
          label = 19;
          break L19;
         }
        }
        $70 = ((($p$126$i$i)) + 1|0);
        $71 = HEAP8[$70>>0]|0;
        $72 = ($71<<24>>24)==(115);
        if ($72) {
         $73 = (_strncmp($p$126$i$i,532185,6)|0);
         $74 = ($73|0)==(0);
         if ($74) {
          $offset$027$i$i$lcssa274 = $offset$027$i$i;$p$126$i$i$lcssa269 = $p$126$i$i;
          label = 23;
          break L19;
         }
        }
        $75 = ($70>>>0)<($62>>>0);
        if ($75) {
         $65 = $71;$p$126$i$i = $70;
        } else {
         $p$1$lcssa$i$i = $70;
         break;
        }
       }
      } else {
       $p$1$lcssa$i$i = $buffer$i$i;
      }
      _memmove(($buffer$i$i|0),($p$1$lcssa$i$i|0),10)|0;
      $78 = (($offset$027$i$i) + 256)|0;
      $79 = HEAP32[$35>>2]|0;
      $80 = (_FT_Stream_Pos($22)|0);
      $81 = (($79) - ($80))|0;
      $82 = ($79|0)==($80|0);
      if ($82) {
       $error$1$ph$i$i = 3;
       label = 25;
       break L16;
      } else {
       $55 = $81;$offset$027$i$i = $78;$p$030$i$i = $36;$read_len$029$i$i = 256;
      }
     }
     if ((label|0) == 19) {
      label = 0;
      $68 = $p$126$i$i$lcssa;
      $69 = (($38) + ($68))|0;
      $$sink$i$i = $69;$offset$027$i$i275 = $offset$027$i$i$lcssa273;
     }
     else if ((label|0) == 23) {
      label = 0;
      $76 = $p$126$i$i$lcssa269;
      $77 = (($48) + ($76))|0;
      $$sink$i$i = $77;$offset$027$i$i275 = $offset$027$i$i$lcssa274;
     }
     $83 = (($$sink$i$i) + ($offset$027$i$i275))|0;
     $84 = (($83) - ($28))|0;
     $85 = (_FT_Stream_Seek($22,$28)|0);
     $86 = ($85|0)==(0);
     if (!($86)) {
      $$0$i$ph$i = $85;
      label = 39;
      break L13;
     }
     $87 = (_FT_Stream_ExtractFrame($22,$84,$39)|0);
     $88 = ($87|0)==(0);
     if (!($88)) {
      $$0$i$ph$i = $87;
      label = 39;
      break L13;
     }
     HEAP32[$40>>2] = $83;
     HEAP32[$41>>2] = $84;
     $89 = HEAP32[$39>>2]|0;
     HEAP32[$42>>2] = $89;
     $$cast$i$i = $89;
     HEAP32[$loader$i>>2] = $$cast$i$i;
     $90 = (($$cast$i$i) + ($84)|0);
     HEAP32[$43>>2] = $90;
     HEAP32[$44>>2] = -1;
     $91 = HEAP32[$45>>2]|0;
     FUNCTION_TABLE_vi[$91 & 8191]($loader$i);
     $92 = HEAP32[$46>>2]|0;
     FUNCTION_TABLE_vi[$92 & 8191]($loader$i);
     $93 = HEAP32[$loader$i>>2]|0;
     $94 = HEAP32[$45>>2]|0;
     FUNCTION_TABLE_vi[$94 & 8191]($loader$i);
     $95 = HEAP32[$46>>2]|0;
     FUNCTION_TABLE_vi[$95 & 8191]($loader$i);
     $96 = HEAP32[$43>>2]|0;
     $cur$042$i$i = HEAP32[$loader$i>>2]|0;
     $97 = ($cur$042$i$i>>>0)<($96>>>0);
     if ($97) {
      $arg1$043$i$i = $$cast$i$i;$arg2$044$i$i = $93;$cur$045$i$i = $cur$042$i$i;
      while(1) {
       $98 = HEAP32[$47>>2]|0;
       $99 = ($98|0)==(0);
       if (!($99)) {
        $$0$i$ph$i = $98;
        label = 39;
        break L13;
       }
       $100 = HEAP8[$cur$045$i$i>>0]|0;
       $101 = ($100<<24>>24)==(83);
       if ($101) {
        $102 = (_strncmp($cur$045$i$i,532175,9)|0);
        $103 = ($102|0)==(0);
        if ($103) {
         $arg1$043$i$i$lcssa278 = $arg1$043$i$i;$arg2$044$i$i$lcssa276 = $arg2$044$i$i;
         break L16;
        }
       }
       $109 = ((($cur$045$i$i)) + 1|0);
       $110 = HEAP8[$109>>0]|0;
       $111 = ($110<<24>>24)==(115);
       if ($111) {
        $112 = (_strncmp($cur$045$i$i,532185,6)|0);
        $113 = ($112|0)==(0);
        if ($113) {
         $$0$i$ph$i = 2;
         label = 39;
         break L13;
        }
       }
       $114 = HEAP32[$45>>2]|0;
       FUNCTION_TABLE_vi[$114 & 8191]($loader$i);
       $115 = HEAP32[$46>>2]|0;
       FUNCTION_TABLE_vi[$115 & 8191]($loader$i);
       $cur$0$i$i = HEAP32[$loader$i>>2]|0;
       $116 = ($cur$0$i$i>>>0)<($96>>>0);
       if ($116) {
        $arg2$044$i$i$phi = $cur$045$i$i;$arg1$043$i$i$phi = $arg2$044$i$i;$cur$045$i$i = $cur$0$i$i;$arg2$044$i$i = $arg2$044$i$i$phi;$arg1$043$i$i = $arg1$043$i$i$phi;
       } else {
        break;
       }
      }
     }
     _FT_Stream_ReleaseFrame($22,$39);
     $117 = (_FT_Stream_Seek($22,$83)|0);
     $118 = ($117|0)==(0);
     if (!($118)) {
      $$0$i$ph$i = $117;
      label = 39;
      break L13;
     }
    }
    if ((label|0) == 25) {
     $$0$i$ph$i = $error$1$ph$i$i;
     label = 39;
     break;
    }
    $104 = (_strncmp($arg1$043$i$i$lcssa278,532192,5)|0);
    $105 = ($104|0)==(0);
    if ($105) {
     $106 = (_atol($arg2$044$i$i$lcssa276)|0);
     $107 = ($106|0)<(0);
     if ($107) {
      $$0$i$ph$i = 3;
      label = 39;
      break;
     }
     $108 = ((($loader$i)) + 88|0);
     HEAP32[$108>>2] = $106;
    }
    HEAP32[$error$i>>2] = 0;
    $119 = HEAP32[$39>>2]|0;
    $120 = HEAP32[$41>>2]|0;
    HEAP32[$loader$i>>2] = $119;
    $121 = (($119) + ($120)|0);
    HEAP32[$43>>2] = $121;
    HEAP32[$47>>2] = 0;
    $122 = ((($121)) + -1|0);
    $123 = ((($122)) + -17|0);
    $124 = ((($cidface)) + 140|0);
    $125 = ((($cidface)) + 300|0);
    $126 = ((($cidface)) + 304|0);
    $127 = ((($loader$i)) + 68|0);
    $128 = ((($loader$i)) + 64|0);
    $129 = ((($cidface)) + 164|0);
    $130 = ((($cidface)) + 312|0);
    $131 = ((($cidface)) + 196|0);
    $cur$0$i2$i = $119;
    L60: while(1) {
     HEAP32[$loader$i>>2] = $cur$0$i2$i;
     $132 = HEAP32[$46>>2]|0;
     FUNCTION_TABLE_vi[$132 & 8191]($loader$i);
     $133 = HEAP32[$loader$i>>2]|0;
     $134 = ($133>>>0)<($121>>>0);
     $135 = ((($133)) + -17|0);
     $newlimit$0$ph$i$i = $134 ? $135 : $123;
     $136 = ($cur$0$i2$i>>>0)<($newlimit$0$ph$i$i>>>0);
     if ($136) {
      $cur$135$i$i = $cur$0$i2$i;
      while(1) {
       $137 = HEAP8[$cur$135$i$i>>0]|0;
       $138 = ($137<<24>>24)==(37);
       if ($138) {
        $139 = (_strncmp($cur$135$i$i,532198,17)|0);
        $140 = ($139|0)==(0);
        if ($140) {
         $141 = HEAP32[$125>>2]|0;
         $142 = ($141|0)>(0);
         if ($142) {
          $143 = HEAP32[$44>>2]|0;
          $144 = (($143) + 1)|0;
          HEAP32[$44>>2] = $144;
         }
        }
       }
       $145 = ((($cur$135$i$i)) + 1|0);
       $exitcond$i$i = ($145|0)==($newlimit$0$ph$i$i|0);
       if ($exitcond$i$i) {
        break;
       } else {
        $cur$135$i$i = $145;
       }
      }
     }
     if (!($134)) {
      label = 48;
      break;
     }
     $146 = HEAP32[$45>>2]|0;
     FUNCTION_TABLE_vi[$146 & 8191]($loader$i);
     $147 = HEAP32[$loader$i>>2]|0;
     $148 = ($147>>>0)<($121>>>0);
     $149 = HEAP32[$47>>2]|0;
     $150 = ($149|0)==(0);
     $or$cond19$i = $148 & $150;
     if (!($or$cond19$i)) {
      $205 = $149;
      break;
     }
     $151 = HEAP8[$133>>0]|0;
     $152 = ($151<<24>>24)==(47);
     $153 = ((($133)) + 2|0);
     $154 = ($153>>>0)<($121>>>0);
     $or$cond12$i$i = $154 & $152;
     if (!($or$cond12$i$i)) {
      $cur$0$i2$i = $147;
      continue;
     }
     $155 = ((($133)) + 1|0);
     $156 = $147;
     $157 = $155;
     $158 = (($156) - ($157))|0;
     $159 = ($147|0)!=($155|0);
     $160 = ($158>>>0)<(22);
     $or$cond$i$i = $159 & $160;
     if (!($or$cond$i$i)) {
      $cur$0$i2$i = $147;
      continue;
     }
     $161 = HEAP8[$155>>0]|0;
     $162 = ($158>>>0)>(1);
     L75: do {
      if ($162) {
       $164 = 532216;$keyword$040$us$i$i = 51332;
       while(1) {
        $163 = HEAP8[$164>>0]|0;
        $165 = ($161<<24>>24)==($163<<24>>24);
        L78: do {
         if ($165) {
          $166 = (_strlen($164)|0);
          $167 = ($158|0)==($166|0);
          if ($167) {
           $n$036$us$i$i = 1;
           while(1) {
            $170 = (($155) + ($n$036$us$i$i)|0);
            $171 = HEAP8[$170>>0]|0;
            $172 = (($164) + ($n$036$us$i$i)|0);
            $173 = HEAP8[$172>>0]|0;
            $174 = ($171<<24>>24)==($173<<24>>24);
            $169 = (($n$036$us$i$i) + 1)|0;
            if (!($174)) {
             break L78;
            }
            $168 = ($169>>>0)<($158>>>0);
            if ($168) {
             $n$036$us$i$i = $169;
            } else {
             $keyword$0$lcssa34$i$i = $keyword$040$us$i$i;
             break L75;
            }
           }
          }
         }
        } while(0);
        $175 = ((($keyword$040$us$i$i)) + 36|0);
        $176 = HEAP32[$175>>2]|0;
        $177 = ($176|0)==(0|0);
        if ($177) {
         $cur$0$i2$i = $147;
         continue L60;
        } else {
         $164 = $176;$keyword$040$us$i$i = $175;
        }
       }
      } else {
       $179 = 532216;$keyword$040$i$i = 51332;
       while(1) {
        $178 = HEAP8[$179>>0]|0;
        $180 = ($161<<24>>24)==($178<<24>>24);
        if ($180) {
         $181 = (_strlen($179)|0);
         $182 = ($158|0)==($181|0);
         if ($182) {
          $keyword$0$lcssa34$i$i = $keyword$040$i$i;
          break L75;
         }
        }
        $202 = ((($keyword$040$i$i)) + 36|0);
        $203 = HEAP32[$202>>2]|0;
        $204 = ($203|0)==(0|0);
        if ($204) {
         $cur$0$i2$i = $147;
         continue L60;
        } else {
         $179 = $203;$keyword$040$i$i = $202;
        }
       }
      }
     } while(0);
     $183 = ((($keyword$0$lcssa34$i$i)) + 8|0);
     $184 = HEAP32[$183>>2]|0;
     $185 = ($184|0)==(11);
     do {
      if ($185) {
       $186 = ((($keyword$0$lcssa34$i$i)) + 12|0);
       $187 = HEAP32[$186>>2]|0;
       FUNCTION_TABLE_vii[$187 & 8191]($cidface,$loader$i);
       $188 = HEAP32[$47>>2]|0;
       $$01$i$i$i = $188;
      } else {
       $189 = ((($keyword$0$lcssa34$i$i)) + 4|0);
       $190 = HEAP32[$189>>2]|0;
       switch ($190|0) {
       case 0:  {
        $object$2$i$i$i = $124;
        break;
       }
       case 3:  {
        $object$2$i$i$i = $129;
        break;
       }
       case 2:  {
        $object$2$i$i$i = $130;
        break;
       }
       case 5:  {
        $object$2$i$i$i = $131;
        break;
       }
       default: {
        $191 = HEAP32[$44>>2]|0;
        $192 = ($191|0)<(0);
        if ($192) {
         label = 72;
         break L60;
        }
        $193 = HEAP32[$125>>2]|0;
        $194 = ($191|0)<($193|0);
        if (!($194)) {
         label = 72;
         break L60;
        }
        $195 = HEAP32[$126>>2]|0;
        $196 = (($195) + (($191*252)|0)|0);
        $object$2$i$i$i = $196;
       }
       }
       HEAP32[$dummy_object$i$i$i>>2] = $object$2$i$i$i;
       $$off$i$i$i = (($184) + -9)|0;
       $switch2$i$i$i = ($$off$i$i$i>>>0)<(2);
       if ($switch2$i$i$i) {
        $197 = HEAP32[$127>>2]|0;
        $198 = (FUNCTION_TABLE_iiiiii[$197 & 8191]($loader$i,$keyword$0$lcssa34$i$i,$dummy_object$i$i$i,0,0)|0);
        $$01$i$i$i = $198;
        break;
       } else {
        $199 = HEAP32[$128>>2]|0;
        $200 = (FUNCTION_TABLE_iiiiii[$199 & 8191]($loader$i,$keyword$0$lcssa34$i$i,$dummy_object$i$i$i,0,0)|0);
        $$01$i$i$i = $200;
        break;
       }
      }
     } while(0);
     HEAP32[$47>>2] = $$01$i$i$i;
     $201 = ($$01$i$i$i|0)==(0);
     if (!($201)) {
      $$9$i$ph$i = $$01$i$i$i;
      label = 76;
      break;
     }
     $$pre$i5$i = HEAP32[$loader$i>>2]|0;
     $cur$0$i2$i = $$pre$i5$i;
    }
    if ((label|0) == 48) {
     $$pre$i = HEAP32[$47>>2]|0;
     $205 = $$pre$i;
    }
    else if ((label|0) == 72) {
     HEAP32[$47>>2] = 160;
     $$9$i$ph$i = 160;
     label = 76;
    }
    if ((label|0) == 76) {
     HEAP32[$error$i>>2] = $$9$i$ph$i;
     $$pre$phi$iZ2D = $39;
     break;
    }
    HEAP32[$error$i>>2] = $205;
    $206 = ($205|0)!=(0);
    $207 = ($face_index|0)<(0);
    $or$cond$i = $207 | $206;
    if ($or$cond$i) {
     $$pre$phi$iZ2D = $39;
    } else {
     $208 = (_ft_mem_alloc($20,40,$error$i)|0);
     $209 = ((($cidface)) + 328|0);
     HEAP32[$209>>2] = $208;
     $210 = HEAP32[$error$i>>2]|0;
     $211 = ($210|0)==(0);
     if ($211) {
      $212 = ((($loader$i)) + 88|0);
      $213 = HEAP32[$212>>2]|0;
      $214 = ($213|0)==(0);
      L114: do {
       if ($214) {
        $272 = HEAP32[$21>>2]|0;
        dest=$208; src=$272; stop=dest+40|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
        $273 = HEAP32[$40>>2]|0;
        $274 = ((($cidface)) + 308|0);
        HEAP32[$274>>2] = $273;
       } else {
        $215 = (_ft_mem_alloc($20,$213,$error$i)|0);
        $216 = ((($cidface)) + 324|0);
        HEAP32[$216>>2] = $215;
        $217 = HEAP32[$error$i>>2]|0;
        $218 = ($217|0)==(0);
        if (!($218)) {
         $$pre$phi$iZ2D = $39;
         break L13;
        }
        $219 = HEAP32[$212>>2]|0;
        $220 = HEAP32[$40>>2]|0;
        $face$idx$val$i = HEAP32[$21>>2]|0;
        $221 = (_FT_Stream_Seek($face$idx$val$i,$220)|0);
        $222 = ($221|0)==(0);
        L118: do {
         if ($222) {
          $223 = (($215) + ($219)|0);
          $224 = ($219|0)>(0);
          L120: do {
           if ($224) {
            $225 = ((($face$idx$val$i)) + 4|0);
            $d$0$ph25$i$i = $215;$p$0$ph26$i$i = $buffer$i6$i;$plimit$0$ph24$i$i = $buffer$i6$i;$upper_nibble$0$ph27$i$i = 1;
            while(1) {
             $p$016$us$i$i = $p$0$ph26$i$i;$plimit$015$us$i$i = $plimit$0$ph24$i$i;
             L124: while(1) {
              $226 = ($p$016$us$i$i>>>0)<($plimit$015$us$i$i>>>0);
              if ($226) {
               $p$2$us$i$i = $p$016$us$i$i;$plimit$2$us$i$i = $plimit$015$us$i$i;
              } else {
               $227 = (_FT_Stream_Pos($face$idx$val$i)|0);
               $228 = HEAP32[$225>>2]|0;
               $229 = (($228) - ($227))|0;
               $230 = ($228|0)==($227|0);
               if ($230) {
                break L118;
               }
               $231 = ($229>>>0)<(256);
               $232 = $231 ? $229 : 256;
               $233 = (_FT_Stream_Read($face$idx$val$i,$buffer$i6$i,$232)|0);
               $234 = ($233|0)==(0);
               if (!($234)) {
                break L118;
               }
               $235 = (_FT_Stream_Pos($face$idx$val$i)|0);
               $236 = (($buffer$i6$i) + ($235)|0);
               $237 = (0 - ($227))|0;
               $238 = (($236) + ($237)|0);
               $p$2$us$i$i = $buffer$i6$i;$plimit$2$us$i$i = $238;
              }
              $239 = HEAP8[$p$2$us$i$i>>0]|0;
              $240 = $239&255;
              $241 = (($240) + -48)|0;
              $242 = ($241>>>0)<(10);
              if ($242) {
               $$lcssa260 = $240;$p$2$us$i$i$lcssa = $p$2$us$i$i;$plimit$2$us$i$i$lcssa = $plimit$2$us$i$i;
               label = 94;
               break;
              }
              $$off$us$i$i = (($239) + -97)<<24>>24;
              $243 = ($$off$us$i$i&255)<(6);
              if ($243) {
               $$lcssa261 = $240;$p$2$us$i$i$lcssa256 = $p$2$us$i$i;$plimit$2$us$i$i$lcssa252 = $plimit$2$us$i$i;
               label = 95;
               break;
              }
              $$off2$us$i$i = (($239) + -65)<<24>>24;
              $244 = ($$off2$us$i$i&255)<(6);
              if ($244) {
               $$lcssa262 = $240;$p$2$us$i$i$lcssa257 = $p$2$us$i$i;$plimit$2$us$i$i$lcssa253 = $plimit$2$us$i$i;
               label = 96;
               break;
              }
              switch ($239<<24>>24) {
              case 62:  {
               $done$1$i$i = 1;$p$2$us$i$i259 = $p$2$us$i$i;$plimit$2$us$i$i255 = $plimit$2$us$i$i;$val$0$i$i = 0;
               break L124;
               break;
              }
              case 0: case 12: case 10: case 13: case 9: case 32:  {
               break;
              }
              default: {
               break L118;
              }
              }
              $245 = ((($p$2$us$i$i)) + 1|0);
              $p$016$us$i$i = $245;$plimit$015$us$i$i = $plimit$2$us$i$i;
             }
             if ((label|0) == 94) {
              label = 0;
              $246 = (($$lcssa260) + 208)|0;
              $247 = $246&255;
              $done$1$i$i = 0;$p$2$us$i$i259 = $p$2$us$i$i$lcssa;$plimit$2$us$i$i255 = $plimit$2$us$i$i$lcssa;$val$0$i$i = $247;
             }
             else if ((label|0) == 95) {
              label = 0;
              $248 = (($$lcssa261) + 159)|0;
              $249 = $248&255;
              $done$1$i$i = 0;$p$2$us$i$i259 = $p$2$us$i$i$lcssa256;$plimit$2$us$i$i255 = $plimit$2$us$i$i$lcssa252;$val$0$i$i = $249;
             }
             else if ((label|0) == 96) {
              label = 0;
              $250 = (($$lcssa262) + 201)|0;
              $251 = $250&255;
              $done$1$i$i = 0;$p$2$us$i$i259 = $p$2$us$i$i$lcssa257;$plimit$2$us$i$i255 = $plimit$2$us$i$i$lcssa253;$val$0$i$i = $251;
             }
             $252 = ($upper_nibble$0$ph27$i$i<<24>>24)==(0);
             if ($252) {
              $256 = HEAP8[$d$0$ph25$i$i>>0]|0;
              $257 = $256&255;
              $258 = $val$0$i$i&255;
              $259 = (($257) + ($258))|0;
              $260 = $259&255;
              HEAP8[$d$0$ph25$i$i>>0] = $260;
              $261 = ((($d$0$ph25$i$i)) + 1|0);
              $d$1$i$i = $261;
             } else {
              $253 = $val$0$i$i&255;
              $254 = $253 << 4;
              $255 = $254&255;
              HEAP8[$d$0$ph25$i$i>>0] = $255;
              $d$1$i$i = $d$0$ph25$i$i;
             }
             $262 = ($done$1$i$i<<24>>24)==(0);
             if (!($262)) {
              break L120;
             }
             $263 = $upper_nibble$0$ph27$i$i&255;
             $264 = (1 - ($263))|0;
             $265 = $264&255;
             $266 = ((($p$2$us$i$i259)) + 1|0);
             $267 = ($d$1$i$i>>>0)<($223>>>0);
             if ($267) {
              $d$0$ph25$i$i = $d$1$i$i;$p$0$ph26$i$i = $266;$plimit$0$ph24$i$i = $plimit$2$us$i$i255;$upper_nibble$0$ph27$i$i = $265;
             } else {
              break;
             }
            }
           }
          } while(0);
          $268 = HEAP32[$209>>2]|0;
          $269 = HEAP32[$216>>2]|0;
          $270 = HEAP32[$212>>2]|0;
          _FT_Stream_OpenMemory($268,$269,$270);
          $271 = ((($cidface)) + 308|0);
          HEAP32[$271>>2] = 0;
          break L114;
         }
        } while(0);
        $$pre$phi$iZ2D = $39;
        break L13;
       }
      } while(0);
      $275 = HEAP32[$19>>2]|0;
      $276 = HEAP32[$209>>2]|0;
      $277 = HEAP32[$0>>2]|0;
      $278 = HEAP32[$125>>2]|0;
      $279 = (_ft_mem_realloc($275,8,0,$278,0,$error$i$i)|0);
      $280 = ((($cidface)) + 316|0);
      HEAP32[$280>>2] = $279;
      $281 = HEAP32[$error$i$i>>2]|0;
      $282 = ($281|0)==(0);
      L148: do {
       if ($282) {
        $283 = HEAP32[$125>>2]|0;
        $284 = ($283|0)>(0);
        if ($284) {
         $285 = ((($cidface)) + 308|0);
         $286 = ((($276)) + 32|0);
         $287 = ((($277)) + 16|0);
         $max_offsets$027$i$i = 0;$n$029$i$i = 0;$offsets$026$i$i = 0;$subr$028$i$i = $279;
         L151: while(1) {
          $288 = HEAP32[$126>>2]|0;
          $289 = (((($288) + (($n$029$i$i*252)|0)|0)) + 4|0);
          $290 = HEAP32[$289>>2]|0;
          $291 = (((($288) + (($n$029$i$i*252)|0)|0)) + 240|0);
          $292 = HEAP32[$291>>2]|0;
          $293 = ($292|0)==(-1);
          if ($293) {
           $offsets$026$i$i$lcssa = $offsets$026$i$i;
           label = 109;
           break;
          }
          $294 = (($292) + 1)|0;
          $295 = ($294>>>0)>($max_offsets$027$i$i>>>0);
          if ($295) {
           $296 = (($292) + 4)|0;
           $297 = $296 & -4;
           $298 = ($297>>>0)>($max_offsets$027$i$i>>>0);
           if (!($298)) {
            $offsets$026$i$i$lcssa247 = $offsets$026$i$i;
            label = 112;
            break;
           }
           $299 = (_ft_mem_realloc($275,4,$max_offsets$027$i$i,$297,$offsets$026$i$i,$error$i$i)|0);
           $300 = HEAP32[$error$i$i>>2]|0;
           $301 = ($300|0)==(0);
           if ($301) {
            $max_offsets$2$i$i = $297;$offsets$2$i$i = $299;
           } else {
            $offsets$3$ph$i$i = $299;
            break;
           }
          } else {
           $max_offsets$2$i$i = $max_offsets$027$i$i;$offsets$2$i$i = $offsets$026$i$i;
          }
          $302 = HEAP32[$285>>2]|0;
          $303 = (((($288) + (($n$029$i$i*252)|0)|0)) + 244|0);
          $304 = HEAP32[$303>>2]|0;
          $305 = (($304) + ($302))|0;
          $306 = (_FT_Stream_Seek($276,$305)|0);
          HEAP32[$error$i$i>>2] = $306;
          $307 = ($306|0)==(0);
          if (!($307)) {
           $offsets$3$ph$i$i = $offsets$2$i$i;
           break;
          }
          $308 = (((($288) + (($n$029$i$i*252)|0)|0)) + 248|0);
          $309 = HEAP32[$308>>2]|0;
          $310 = Math_imul($309, $294)|0;
          $311 = (_FT_Stream_EnterFrame($276,$310)|0);
          HEAP32[$error$i$i>>2] = $311;
          $312 = ($311|0)==(0);
          if (!($312)) {
           $offsets$3$ph$i$i = $offsets$2$i$i;
           break;
          }
          $$pre$i8$i = HEAP32[$308>>2]|0;
          $313 = $$pre$i8$i&255;
          $314 = ($313<<24>>24)==(0);
          $315 = (($$pre$i8$i) + 255)|0;
          $316 = $315 & 255;
          $317 = (($316) + 1)|0;
          if ($314) {
           $count$016$i$us$i = 0;
           while(1) {
            $319 = (($offsets$2$i$i) + ($count$016$i$us$i<<2)|0);
            HEAP32[$319>>2] = 0;
            $320 = (($count$016$i$us$i) + 1)|0;
            $321 = ($320>>>0)>($292>>>0);
            if ($321) {
             break;
            } else {
             $count$016$i$us$i = $320;
            }
           }
          } else {
           $318 = HEAP32[$286>>2]|0;
           $count$016$i$i = 0;$p$sroa$0$015$i$i = $318;
           while(1) {
            $322 = $p$sroa$0$015$i$i;
            $$01$i$i9$i = $313;$p$03$i$i$i = $322;$result$02$i$i$i = 0;
            while(1) {
             $323 = $result$02$i$i$i << 8;
             $324 = ((($p$03$i$i$i)) + 1|0);
             $325 = HEAP8[$p$03$i$i$i>>0]|0;
             $326 = $325&255;
             $327 = $326 | $323;
             $328 = (($$01$i$i9$i) + -1)<<24>>24;
             $329 = ($328<<24>>24)==(0);
             if ($329) {
              $$lcssa246 = $327;
              break;
             } else {
              $$01$i$i9$i = $328;$p$03$i$i$i = $324;$result$02$i$i$i = $327;
             }
            }
            $scevgep$i$i$i = (($322) + ($317)|0);
            $330 = $scevgep$i$i$i;
            $331 = (($offsets$2$i$i) + ($count$016$i$i<<2)|0);
            HEAP32[$331>>2] = $$lcssa246;
            $332 = (($count$016$i$i) + 1)|0;
            $333 = ($332>>>0)>($292>>>0);
            if ($333) {
             break;
            } else {
             $count$016$i$i = $332;$p$sroa$0$015$i$i = $330;
            }
           }
          }
          _FT_Stream_ExitFrame($276);
          $334 = ($292|0)==(0);
          $$pre39$i$i = HEAP32[$offsets$2$i$i>>2]|0;
          if (!($334)) {
           $340 = $$pre39$i$i;$count$117$i$i = 1;
           while(1) {
            $337 = (($offsets$2$i$i) + ($count$117$i$i<<2)|0);
            $338 = HEAP32[$337>>2]|0;
            $339 = ($340>>>0)>($338>>>0);
            $336 = (($count$117$i$i) + 1)|0;
            if ($339) {
             $offsets$3$ph$i$i = $offsets$2$i$i;
             break L151;
            }
            $335 = ($336>>>0)>($292>>>0);
            if ($335) {
             break;
            } else {
             $340 = $338;$count$117$i$i = $336;
            }
           }
          }
          $341 = (($offsets$2$i$i) + ($292<<2)|0);
          $342 = HEAP32[$341>>2]|0;
          $343 = (($342) - ($$pre39$i$i))|0;
          $344 = (_ft_mem_realloc($275,4,0,$294,0,$error$i$i)|0);
          $345 = ((($subr$028$i$i)) + 4|0);
          HEAP32[$345>>2] = $344;
          $346 = HEAP32[$error$i$i>>2]|0;
          $347 = ($346|0)==(0);
          if (!($347)) {
           $offsets$3$ph$i$i = $offsets$2$i$i;
           break;
          }
          $348 = (_ft_mem_alloc($275,$343,$error$i$i)|0);
          $349 = HEAP32[$345>>2]|0;
          HEAP32[$349>>2] = $348;
          $350 = HEAP32[$error$i$i>>2]|0;
          $351 = ($350|0)==(0);
          if (!($351)) {
           $offsets$3$ph$i$i = $offsets$2$i$i;
           break;
          }
          $352 = HEAP32[$285>>2]|0;
          $353 = HEAP32[$offsets$2$i$i>>2]|0;
          $354 = (($353) + ($352))|0;
          $355 = (_FT_Stream_Seek($276,$354)|0);
          HEAP32[$error$i$i>>2] = $355;
          $356 = ($355|0)==(0);
          if (!($356)) {
           $offsets$3$ph$i$i = $offsets$2$i$i;
           break;
          }
          $357 = HEAP32[$345>>2]|0;
          $358 = HEAP32[$357>>2]|0;
          $359 = (_FT_Stream_Read($276,$358,$343)|0);
          HEAP32[$error$i$i>>2] = $359;
          $360 = ($359|0)==(0);
          if (!($360)) {
           $offsets$3$ph$i$i = $offsets$2$i$i;
           break;
          }
          do {
           if (!($334)) {
            $$pre40$i$i = HEAP32[$offsets$2$i$i>>2]|0;
            $365 = $$pre40$i$i;$count$221$i$i = 1;
            while(1) {
             $361 = (($offsets$2$i$i) + ($count$221$i$i<<2)|0);
             $362 = HEAP32[$361>>2]|0;
             $363 = (($count$221$i$i) + -1)|0;
             $364 = (($362) - ($365))|0;
             $366 = HEAP32[$345>>2]|0;
             $367 = (($366) + ($363<<2)|0);
             $368 = HEAP32[$367>>2]|0;
             $369 = (($368) + ($364)|0);
             $370 = (($366) + ($count$221$i$i<<2)|0);
             HEAP32[$370>>2] = $369;
             $371 = (($count$221$i$i) + 1)|0;
             $372 = ($371>>>0)>($292>>>0);
             if ($372) {
              break;
             } else {
              $365 = $362;$count$221$i$i = $371;
             }
            }
            $373 = ($290|0)>(-1);
            if (!($373)) {
             break;
            }
            $374 = ((($offsets$2$i$i)) + 4|0);
            $375 = HEAP32[$374>>2]|0;
            $376 = (($375) - ($$pre40$i$i))|0;
            $377 = HEAP32[$287>>2]|0;
            $378 = HEAP32[$345>>2]|0;
            $379 = HEAP32[$378>>2]|0;
            FUNCTION_TABLE_viii[$377 & 8191]($379,$376,4330);
            $exitcond$i11$i51 = ($292|0)==(1);
            if ($exitcond$i11$i51) {
             break;
            } else {
             $380 = $374;$382 = 1;
            }
            while(1) {
             $$pre95$i = HEAP32[$380>>2]|0;
             $381 = (($382) + 1)|0;
             $383 = (($offsets$2$i$i) + ($381<<2)|0);
             $384 = HEAP32[$383>>2]|0;
             $385 = (($384) - ($$pre95$i))|0;
             $386 = HEAP32[$287>>2]|0;
             $387 = HEAP32[$345>>2]|0;
             $388 = (($387) + ($382<<2)|0);
             $389 = HEAP32[$388>>2]|0;
             FUNCTION_TABLE_viii[$386 & 8191]($389,$385,4330);
             $exitcond$i11$i = ($381|0)==($292|0);
             if ($exitcond$i11$i) {
              break;
             } else {
              $380 = $383;$382 = $381;
             }
            }
           }
          } while(0);
          HEAP32[$subr$028$i$i>>2] = $292;
          $390 = (($n$029$i$i) + 1)|0;
          $391 = ((($subr$028$i$i)) + 8|0);
          $392 = HEAP32[$125>>2]|0;
          $393 = ($390|0)<($392|0);
          if ($393) {
           $max_offsets$027$i$i = $max_offsets$2$i$i;$n$029$i$i = $390;$offsets$026$i$i = $offsets$2$i$i;$subr$028$i$i = $391;
          } else {
           $offsets$4$i$i = $offsets$2$i$i;
           break L148;
          }
         }
         if ((label|0) == 109) {
          HEAP32[$error$i$i>>2] = 160;
          $offsets$3$ph$i$i = $offsets$026$i$i$lcssa;
         }
         else if ((label|0) == 112) {
          HEAP32[$error$i$i>>2] = 160;
          $offsets$3$ph$i$i = $offsets$026$i$i$lcssa247;
         }
         $394 = HEAP32[$280>>2]|0;
         $395 = ($394|0)==(0|0);
         if ($395) {
          $offsets$4$i$i = $offsets$3$ph$i$i;
         } else {
          $396 = HEAP32[$125>>2]|0;
          $397 = ($396|0)>(0);
          if ($397) {
           $399 = $394;$n$114$i$i = 0;
           while(1) {
            $398 = (((($399) + ($n$114$i$i<<3)|0)) + 4|0);
            $400 = HEAP32[$398>>2]|0;
            $401 = ($400|0)==(0|0);
            if ($401) {
             $407 = $399;
            } else {
             $402 = HEAP32[$400>>2]|0;
             _ft_mem_free($275,$402);
             $403 = HEAP32[$280>>2]|0;
             $404 = (((($403) + ($n$114$i$i<<3)|0)) + 4|0);
             $405 = HEAP32[$404>>2]|0;
             HEAP32[$405>>2] = 0;
             $$pre41$i$i = HEAP32[$280>>2]|0;
             $407 = $$pre41$i$i;
            }
            $406 = (((($407) + ($n$114$i$i<<3)|0)) + 4|0);
            $408 = HEAP32[$406>>2]|0;
            _ft_mem_free($275,$408);
            $409 = HEAP32[$280>>2]|0;
            $410 = (((($409) + ($n$114$i$i<<3)|0)) + 4|0);
            HEAP32[$410>>2] = 0;
            $411 = (($n$114$i$i) + 1)|0;
            $412 = HEAP32[$125>>2]|0;
            $413 = ($411|0)<($412|0);
            if ($413) {
             $399 = $409;$n$114$i$i = $411;
            } else {
             $$in$i$i = $409;
             break;
            }
           }
          } else {
           $$in$i$i = $394;
          }
          _ft_mem_free($275,$$in$i$i);
          HEAP32[$280>>2] = 0;
          $offsets$4$i$i = $offsets$3$ph$i$i;
         }
        } else {
         $offsets$4$i$i = 0;
        }
       } else {
        $offsets$4$i$i = 0;
       }
      } while(0);
      _ft_mem_free($275,$offsets$4$i$i);
      $414 = HEAP32[$error$i$i>>2]|0;
      HEAP32[$error$i>>2] = $414;
      $$pre$phi$iZ2D = $39;
     } else {
      $$pre$phi$iZ2D = $39;
     }
    }
   } else {
    $$0$i$ph$i = 2;
    label = 39;
   }
  } else {
   $$0$i$ph$i = $29;
   label = 39;
  }
 } while(0);
 if ((label|0) == 39) {
  HEAP32[$error$i>>2] = $$0$i$ph$i;
  $$pre96$i = ((($loader$i)) + 76|0);
  $$pre$phi$iZ2D = $$pre96$i;
 }
 $415 = HEAP32[$$pre$phi$iZ2D>>2]|0;
 $416 = ($415|0)==(0|0);
 if (!($416)) {
  $417 = HEAP32[$27>>2]|0;
  _FT_Stream_ReleaseFrame($417,$$pre$phi$iZ2D);
 }
 $418 = ((($loader$i)) + 24|0);
 $419 = HEAP32[$418>>2]|0;
 FUNCTION_TABLE_vi[$419 & 8191]($loader$i);
 $420 = HEAP32[$error$i>>2]|0;
 $421 = ($420|0)!=(0);
 $422 = ($face_index|0)<(0);
 $or$cond = $422 | $421;
 if ($or$cond) {
  $error$0 = $420;
  STACKTOP = sp;return ($error$0|0);
 }
 $423 = ($face_index|0)==(0);
 if (!($423)) {
  $error$0 = 6;
  STACKTOP = sp;return ($error$0|0);
 }
 $424 = ((($cidface)) + 140|0);
 $425 = ((($424)) + 156|0);
 $426 = HEAP32[$425>>2]|0;
 $427 = ((($cidface)) + 16|0);
 HEAP32[$427>>2] = $426;
 $428 = ((($cidface)) + 36|0);
 HEAP32[$428>>2] = 0;
 $429 = ((($cidface)) + 4|0);
 HEAP32[$429>>2] = 0;
 $430 = ((($cidface)) + 8|0);
 $431 = HEAP32[$430>>2]|0;
 $432 = $431 | 2065;
 HEAP32[$430>>2] = $432;
 $433 = ((($424)) + 48|0);
 $434 = HEAP32[$433>>2]|0;
 $435 = $434&255;
 $436 = ($435<<24>>24)==(0);
 $437 = $434 >>> 16;
 $438 = $437&65535;
 if (!($436)) {
  $439 = $431 | 2069;
  HEAP32[$430>>2] = $439;
 }
 $440 = ((($424)) + 36|0);
 $441 = HEAP32[$440>>2]|0;
 $442 = ((($cidface)) + 20|0);
 HEAP32[$442>>2] = $441;
 $443 = ((($cidface)) + 24|0);
 HEAP32[$443>>2] = 594711;
 $444 = ($441|0)==(0);
 L216: do {
  if ($444) {
   $459 = HEAP32[$424>>2]|0;
   $460 = ($459|0)==(0|0);
   if (!($460)) {
    HEAP32[$442>>2] = $459;
   }
  } else {
   $445 = ((($424)) + 32|0);
   $446 = HEAP32[$445>>2]|0;
   $447 = ($446|0)==(0|0);
   if (!($447)) {
    $448 = HEAP8[$446>>0]|0;
    $449 = ($448<<24>>24)==(0);
    if (!($449)) {
     $$cast = $441;
     $451 = $448;$family$0$ph50 = $$cast;$full$0$ph49 = $446;
     L222: while(1) {
      $450 = ($451<<24>>24)==(0);
      $family$047 = $family$0$ph50;
      L224: while(1) {
       $452 = HEAP8[$family$047>>0]|0;
       $453 = ($451<<24>>24)==($452<<24>>24);
       if ($453) {
        $family$047$lcssa = $family$047;
        label = 155;
        break;
       }
       switch ($451<<24>>24) {
       case 45: case 32:  {
        $family$0$ph$be = $family$047;
        break L224;
        break;
       }
       default: {
       }
       }
       switch ($452<<24>>24) {
       case 0:  {
        $full$0$ph49$lcssa244 = $full$0$ph49;
        break L222;
        break;
       }
       case 45: case 32:  {
        break;
       }
       default: {
        break L216;
       }
       }
       $458 = ((($family$047)) + 1|0);
       if ($450) {
        break L216;
       } else {
        $family$047 = $458;
       }
      }
      if ((label|0) == 155) {
       label = 0;
       $454 = ((($family$047$lcssa)) + 1|0);
       $family$0$ph$be = $454;
      }
      $455 = ((($full$0$ph49)) + 1|0);
      $456 = HEAP8[$455>>0]|0;
      $457 = ($456<<24>>24)==(0);
      if ($457) {
       break L216;
      } else {
       $451 = $456;$family$0$ph50 = $family$0$ph$be;$full$0$ph49 = $455;
      }
     }
     HEAP32[$443>>2] = $full$0$ph49$lcssa244;
    }
   }
  }
 } while(0);
 $461 = ((($cidface)) + 12|0);
 $462 = ((($424)) + 44|0);
 $463 = HEAP32[$462>>2]|0;
 $not$ = ($463|0)!=(0);
 $$ = $not$&1;
 HEAP32[$461>>2] = $$;
 $464 = ((($424)) + 40|0);
 $465 = HEAP32[$464>>2]|0;
 $466 = ($465|0)==(0|0);
 do {
  if (!($466)) {
   $467 = (_strcmp($465,594731)|0);
   $468 = ($467|0)==(0);
   if (!($468)) {
    $469 = (_strcmp($465,594283)|0);
    $470 = ($469|0)==(0);
    if (!($470)) {
     break;
    }
   }
   $471 = $$ | 2;
   HEAP32[$461>>2] = $471;
  }
 } while(0);
 $472 = ((($cidface)) + 28|0);
 HEAP32[$472>>2] = 0;
 $473 = ((($cidface)) + 32|0);
 HEAP32[$473>>2] = 0;
 $474 = ((($424)) + 56|0);
 $475 = HEAP32[$474>>2]|0;
 $476 = $475 >> 16;
 $477 = ((($cidface)) + 52|0);
 HEAP32[$477>>2] = $476;
 $478 = ((($424)) + 60|0);
 $479 = HEAP32[$478>>2]|0;
 $480 = $479 >> 16;
 $481 = ((($cidface)) + 56|0);
 HEAP32[$481>>2] = $480;
 $482 = ((($424)) + 64|0);
 $483 = HEAP32[$482>>2]|0;
 $484 = (($483) + 65535)|0;
 $485 = $484 >> 16;
 $486 = ((($cidface)) + 60|0);
 HEAP32[$486>>2] = $485;
 $487 = ((($424)) + 68|0);
 $488 = HEAP32[$487>>2]|0;
 $489 = (($488) + 65535)|0;
 $490 = $489 >> 16;
 $491 = ((($cidface)) + 64|0);
 HEAP32[$491>>2] = $490;
 $492 = ((($cidface)) + 68|0);
 $493 = HEAP16[$492>>1]|0;
 $494 = ($493<<16>>16)==(0);
 if ($494) {
  HEAP16[$492>>1] = 1000;
  $500 = 1000;
 } else {
  $500 = $493;
 }
 $495 = $490&65535;
 $496 = ((($cidface)) + 70|0);
 HEAP16[$496>>1] = $495;
 $497 = $480&65535;
 $498 = ((($cidface)) + 72|0);
 HEAP16[$498>>1] = $497;
 $499 = $500&65535;
 $501 = ($499*12)|0;
 $502 = (($501>>>0) / 10)&-1;
 $503 = ((($cidface)) + 74|0);
 $sext = $502 << 16;
 $504 = $sext >> 16;
 $505 = (($490) - ($480))|0;
 $506 = ($504|0)<($505|0);
 $$146 = $506 ? $505 : $502;
 $storemerge = $$146&65535;
 HEAP16[$503>>1] = $storemerge;
 $507 = ((($cidface)) + 80|0);
 HEAP16[$507>>1] = $438;
 $508 = ((($424)) + 52|0);
 $509 = HEAP16[$508>>1]|0;
 $510 = ((($cidface)) + 82|0);
 HEAP16[$510>>1] = $509;
 $error$0 = 0;
 STACKTOP = sp;return ($error$0|0);
}
function _parse_fd_array($face,$parser) {
 $face = $face|0;
 $parser = $parser|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $n$01 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($face)) + 100|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$error>>2] = 0;
 $2 = ((($parser)) + 36|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($parser)|0);
 $5 = ($4|0)<(0);
 if ($5) {
  HEAP32[$error>>2] = 3;
  STACKTOP = sp;return;
 }
 $6 = ((($face)) + 304|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 if (!($8)) {
  STACKTOP = sp;return;
 }
 $9 = (_ft_mem_realloc($1,252,0,$4,0,$error)|0);
 HEAP32[$6>>2] = $9;
 $10 = HEAP32[$error>>2]|0;
 $11 = ($10|0)==(0);
 if (!($11)) {
  STACKTOP = sp;return;
 }
 $12 = ((($face)) + 300|0);
 HEAP32[$12>>2] = $4;
 $13 = ($4|0)>(0);
 if ($13) {
  $n$01 = 0;
 } else {
  STACKTOP = sp;return;
 }
 while(1) {
  $14 = (((($9) + (($n$01*252)|0)|0)) + 4|0);
  HEAP32[$14>>2] = 4;
  $15 = (($n$01) + 1)|0;
  $16 = ($15|0)<($4|0);
  if ($16) {
   $n$01 = $15;
  } else {
   break;
  }
 }
 STACKTOP = sp;return;
}
function _cid_parse_font_matrix($face,$parser) {
 $face = $face|0;
 $parser = $parser|0;
 var $$pre = 0, $$pre$phi2Z2D = 0, $$pre$phi4Z2D = 0, $$pre$phi6Z2D = 0, $$pre$phiZ2D = 0, $$pre1 = 0, $$pre3 = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $temp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $temp = sp;
 $0 = ((($parser)) + 96|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(-1);
 if (!($2)) {
  STACKTOP = sp;return;
 }
 $3 = ((($face)) + 300|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($1|0)<($4|0);
 if (!($5)) {
  STACKTOP = sp;return;
 }
 $6 = ((($face)) + 304|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($parser)) + 52|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (FUNCTION_TABLE_iiiii[$9 & 8191]($parser,6,$temp,3)|0);
 $11 = ($10|0)<(6);
 if ($11) {
  STACKTOP = sp;return;
 }
 $12 = ((($temp)) + 12|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13|0)<(0);
 $15 = (0 - ($13))|0;
 $16 = $14 ? $15 : $13;
 $17 = ($16|0)==(0);
 if ($17) {
  STACKTOP = sp;return;
 }
 $18 = (_FT_DivFix(1000,$16)|0);
 $19 = $18&65535;
 $20 = ((($face)) + 68|0);
 HEAP16[$20>>1] = $19;
 $21 = ($16|0)==(65536);
 if ($21) {
  $$pre = ((($temp)) + 4|0);
  $$pre1 = ((($temp)) + 8|0);
  $$pre3 = ((($temp)) + 16|0);
  $$pre5 = ((($temp)) + 20|0);
  $$pre$phi2Z2D = $$pre1;$$pre$phi4Z2D = $$pre3;$$pre$phi6Z2D = $$pre5;$$pre$phiZ2D = $$pre;
 } else {
  $22 = HEAP32[$temp>>2]|0;
  $23 = (_FT_DivFix($22,$16)|0);
  HEAP32[$temp>>2] = $23;
  $24 = ((($temp)) + 4|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = (_FT_DivFix($25,$16)|0);
  HEAP32[$24>>2] = $26;
  $27 = ((($temp)) + 8|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = (_FT_DivFix($28,$16)|0);
  HEAP32[$27>>2] = $29;
  $30 = ((($temp)) + 16|0);
  $31 = HEAP32[$30>>2]|0;
  $32 = (_FT_DivFix($31,$16)|0);
  HEAP32[$30>>2] = $32;
  $33 = ((($temp)) + 20|0);
  $34 = HEAP32[$33>>2]|0;
  $35 = (_FT_DivFix($34,$16)|0);
  HEAP32[$33>>2] = $35;
  $36 = HEAP32[$12>>2]|0;
  $37 = $36 >> 31;
  $38 = $37 & -131072;
  $39 = $38 | 65536;
  HEAP32[$12>>2] = $39;
  $$pre$phi2Z2D = $27;$$pre$phi4Z2D = $30;$$pre$phi6Z2D = $33;$$pre$phiZ2D = $24;
 }
 $40 = HEAP32[$temp>>2]|0;
 $41 = (((($7) + (($1*252)|0)|0)) + 216|0);
 HEAP32[$41>>2] = $40;
 $42 = HEAP32[$$pre$phiZ2D>>2]|0;
 $43 = (((($7) + (($1*252)|0)|0)) + 224|0);
 HEAP32[$43>>2] = $42;
 $44 = HEAP32[$$pre$phi2Z2D>>2]|0;
 $45 = (((($7) + (($1*252)|0)|0)) + 220|0);
 HEAP32[$45>>2] = $44;
 $46 = HEAP32[$12>>2]|0;
 $47 = (((($7) + (($1*252)|0)|0)) + 228|0);
 HEAP32[$47>>2] = $46;
 $48 = HEAP32[$$pre$phi4Z2D>>2]|0;
 $49 = $48 >> 16;
 $50 = (((($7) + (($1*252)|0)|0)) + 232|0);
 HEAP32[$50>>2] = $49;
 $51 = HEAP32[$$pre$phi6Z2D>>2]|0;
 $52 = $51 >> 16;
 $53 = (((($7) + (($1*252)|0)|0)) + 236|0);
 HEAP32[$53>>2] = $52;
 STACKTOP = sp;return;
}
function _parse_expansion_factor($face,$parser) {
 $face = $face|0;
 $parser = $parser|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($parser)) + 96|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(-1);
 if (!($2)) {
  return;
 }
 $3 = ((($face)) + 300|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($1|0)<($4|0);
 if (!($5)) {
  return;
 }
 $6 = ((($face)) + 304|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($parser)) + 40|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (FUNCTION_TABLE_iii[$9 & 8191]($parser,0)|0);
 $11 = (((($7) + (($1*252)|0)|0)) + 208|0);
 HEAP32[$11>>2] = $10;
 $12 = (((($7) + (($1*252)|0)|0)) + 180|0);
 HEAP32[$12>>2] = $10;
 return;
}
function _cid_face_done($cidface) {
 $cidface = $cidface|0;
 var $$pre = 0, $$pre2 = 0, $$pre3 = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $n$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($cidface|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($cidface)) + 140|0);
 $2 = ((($1)) + 24|0);
 $3 = ((($cidface)) + 100|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($cidface)) + 316|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0|0);
 $$pre5 = ((($1)) + 160|0);
 if (!($7)) {
  $8 = HEAP32[$$pre5>>2]|0;
  $9 = ($8|0)>(0);
  if ($9) {
   $11 = $6;$43 = $8;$n$01 = 0;
   while(1) {
    $10 = (((($11) + ($n$01<<3)|0)) + 4|0);
    $12 = HEAP32[$10>>2]|0;
    $13 = ($12|0)==(0|0);
    if ($13) {
     $19 = $43;
    } else {
     $14 = HEAP32[$12>>2]|0;
     _ft_mem_free($4,$14);
     $15 = HEAP32[$10>>2]|0;
     HEAP32[$15>>2] = 0;
     $16 = HEAP32[$10>>2]|0;
     _ft_mem_free($4,$16);
     HEAP32[$10>>2] = 0;
     $$pre2 = HEAP32[$$pre5>>2]|0;
     $19 = $$pre2;
    }
    $17 = (($n$01) + 1)|0;
    $18 = ($17|0)<($19|0);
    if (!($18)) {
     break;
    }
    $$pre = HEAP32[$5>>2]|0;
    $11 = $$pre;$43 = $19;$n$01 = $17;
   }
   $$pre3 = HEAP32[$5>>2]|0;
   $20 = $$pre3;
  } else {
   $20 = $6;
  }
  _ft_mem_free($4,$20);
  HEAP32[$5>>2] = 0;
 }
 $21 = HEAP32[$2>>2]|0;
 _ft_mem_free($4,$21);
 HEAP32[$2>>2] = 0;
 $22 = ((($1)) + 28|0);
 $23 = HEAP32[$22>>2]|0;
 _ft_mem_free($4,$23);
 HEAP32[$22>>2] = 0;
 $24 = ((($1)) + 32|0);
 $25 = HEAP32[$24>>2]|0;
 _ft_mem_free($4,$25);
 HEAP32[$24>>2] = 0;
 $26 = ((($1)) + 36|0);
 $27 = HEAP32[$26>>2]|0;
 _ft_mem_free($4,$27);
 HEAP32[$26>>2] = 0;
 $28 = ((($1)) + 40|0);
 $29 = HEAP32[$28>>2]|0;
 _ft_mem_free($4,$29);
 HEAP32[$28>>2] = 0;
 $30 = ((($1)) + 164|0);
 $31 = HEAP32[$30>>2]|0;
 _ft_mem_free($4,$31);
 HEAP32[$30>>2] = 0;
 HEAP32[$$pre5>>2] = 0;
 $32 = HEAP32[$1>>2]|0;
 _ft_mem_free($4,$32);
 HEAP32[$1>>2] = 0;
 $33 = ((($1)) + 12|0);
 $34 = HEAP32[$33>>2]|0;
 _ft_mem_free($4,$34);
 HEAP32[$33>>2] = 0;
 $35 = ((($1)) + 16|0);
 $36 = HEAP32[$35>>2]|0;
 _ft_mem_free($4,$36);
 HEAP32[$35>>2] = 0;
 $37 = ((($cidface)) + 20|0);
 HEAP32[$37>>2] = 0;
 $38 = ((($cidface)) + 24|0);
 HEAP32[$38>>2] = 0;
 $39 = ((($cidface)) + 324|0);
 $40 = HEAP32[$39>>2]|0;
 _ft_mem_free($4,$40);
 HEAP32[$39>>2] = 0;
 $41 = ((($cidface)) + 328|0);
 $42 = HEAP32[$41>>2]|0;
 _ft_mem_free($4,$42);
 HEAP32[$41>>2] = 0;
 return;
}
function _cid_size_init($cidsize) {
 $cidsize = $cidsize|0;
 var $$idx$val = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$0 = 0, $globals = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $globals = sp;
 $$idx$val = HEAP32[$cidsize>>2]|0;
 $0 = ((($$idx$val)) + 320|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($$idx$val)) + 96|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (_FT_Get_Module($5,593918)|0);
 $7 = ($6|0)!=(0|0);
 $8 = ($1|0)!=(0|0);
 $or$cond$i = $8 & $7;
 if (!($or$cond$i)) {
  $error$0 = 0;
  STACKTOP = sp;return ($error$0|0);
 }
 $9 = HEAP32[$1>>2]|0;
 $10 = ($9|0)==(0|0);
 if ($10) {
  $error$0 = 0;
  STACKTOP = sp;return ($error$0|0);
 }
 $11 = (FUNCTION_TABLE_ii[$9 & 8191]($6)|0);
 $12 = ($11|0)==(0|0);
 if ($12) {
  $error$0 = 0;
  STACKTOP = sp;return ($error$0|0);
 }
 $13 = HEAP32[$cidsize>>2]|0;
 $14 = ((($13)) + 140|0);
 $15 = ((($14)) + 164|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($13)) + 4|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = (($16) + (($18*252)|0)|0);
 $20 = HEAP32[$11>>2]|0;
 $21 = ((($13)) + 100|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = (FUNCTION_TABLE_iiii[$20 & 8191]($22,$19,$globals)|0);
 $24 = ($23|0)==(0);
 if ($24) {
  $25 = HEAP32[$globals>>2]|0;
  $26 = ((($cidsize)) + 40|0);
  HEAP32[$26>>2] = $25;
 }
 $error$0 = $23;
 STACKTOP = sp;return ($error$0|0);
}
function _cid_size_done($cidsize) {
 $cidsize = $cidsize|0;
 var $$idx$val = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cidsize)) + 40|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 $$idx$val = HEAP32[$cidsize>>2]|0;
 $3 = ((($$idx$val)) + 320|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($$idx$val)) + 96|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($6)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (_FT_Get_Module($8,593918)|0);
 $10 = ($9|0)!=(0|0);
 $11 = ($4|0)!=(0|0);
 $or$cond$i = $11 & $10;
 if ($or$cond$i) {
  $12 = HEAP32[$4>>2]|0;
  $13 = ($12|0)==(0|0);
  if (!($13)) {
   $14 = (FUNCTION_TABLE_ii[$12 & 8191]($9)|0);
   $15 = ($14|0)==(0|0);
   if (!($15)) {
    $16 = ((($14)) + 8|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = HEAP32[$0>>2]|0;
    FUNCTION_TABLE_vi[$17 & 8191]($18);
   }
  }
 }
 HEAP32[$0>>2] = 0;
 return;
}
function _cid_slot_init($slot) {
 $slot = $slot|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($slot)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 320|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  return 0;
 }
 $5 = ((($1)) + 96|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($6)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (_FT_Get_Module($8,593918)|0);
 $10 = ($9|0)==(0|0);
 if ($10) {
  return 0;
 }
 $11 = ((($3)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (FUNCTION_TABLE_ii[$12 & 8191]($9)|0);
 $14 = ((($slot)) + 156|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($15)) + 36|0);
 HEAP32[$16>>2] = $13;
 return 0;
}
function _cid_slot_done($slot) {
 $slot = $slot|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($slot)) + 156|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 36|0);
 HEAP32[$2>>2] = 0;
 return;
}
function _cid_slot_load_glyph($cidglyph,$cidsize,$glyph_index,$load_flags) {
 $cidglyph = $cidglyph|0;
 $cidsize = $cidsize|0;
 $glyph_index = $glyph_index|0;
 $load_flags = $load_flags|0;
 var $$lobit = 0, $$lobit1 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $advance = 0, $cbox = 0, $decoder = 0, $error$0 = 0, $font_matrix = 0, $load_flags$ = 0, $n$02 = 0, $phitmp = 0, $vec$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1552|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $decoder = sp + 40|0;
 $font_matrix = sp + 24|0;
 $cbox = sp + 8|0;
 $advance = sp;
 $0 = ((($cidglyph)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 136|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5>>>0)>($glyph_index>>>0);
 if (!($6)) {
  $error$0 = 6;
  STACKTOP = sp;return ($error$0|0);
 }
 $7 = $load_flags & 1024;
 $8 = ($7|0)==(0);
 $9 = $load_flags | 3;
 $load_flags$ = $8 ? $load_flags : $9;
 $10 = ((($cidsize)) + 16|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($cidglyph)) + 164|0);
 HEAP32[$12>>2] = $11;
 $13 = ((($cidsize)) + 20|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($cidglyph)) + 168|0);
 HEAP32[$15>>2] = $14;
 $16 = ((($cidglyph)) + 108|0);
 $17 = ((($cidglyph)) + 110|0);
 HEAP16[$17>>1] = 0;
 HEAP16[$16>>1] = 0;
 $18 = $load_flags$ & 1;
 $19 = ($18|0)==(0);
 if ($19) {
  $20 = $load_flags$ >>> 1;
  $$lobit = $20 & 1;
  $21 = $$lobit ^ 1;
  $phitmp = $21&255;
  $28 = $phitmp;
 } else {
  $28 = 0;
 }
 $22 = ((($cidglyph)) + 72|0);
 HEAP32[$22>>2] = 1869968492;
 $23 = ((($3)) + 12|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = HEAP32[$24>>2]|0;
 $26 = $load_flags$ >>> 16;
 $27 = $26 & 15;
 $29 = (FUNCTION_TABLE_iiiiiiiiii[$25 & 8191]($decoder,$1,$cidsize,$cidglyph,0,0,$28,$27,4361)|0);
 $30 = ($29|0)==(0);
 if (!($30)) {
  $error$0 = $29;
  STACKTOP = sp;return ($error$0|0);
 }
 $31 = $load_flags$ & 1024;
 $$lobit1 = $31 >>> 10;
 $32 = $$lobit1&255;
 $33 = ((($decoder)) + 69|0);
 HEAP8[$33>>0] = $32;
 $34 = (_cid_load_glyph($decoder,$glyph_index)|0);
 $35 = ($34|0)==(0);
 if (!($35)) {
  $error$0 = $34;
  STACKTOP = sp;return ($error$0|0);
 }
 $36 = ($31|0)==(0);
 $37 = ((($decoder)) + 1376|0);
 ;HEAP32[$font_matrix>>2]=HEAP32[$37>>2]|0;HEAP32[$font_matrix+4>>2]=HEAP32[$37+4>>2]|0;HEAP32[$font_matrix+8>>2]=HEAP32[$37+8>>2]|0;HEAP32[$font_matrix+12>>2]=HEAP32[$37+12>>2]|0;
 $38 = ((($decoder)) + 1392|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ((($decoder)) + 1396|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = HEAP32[$23>>2]|0;
 $43 = ((($42)) + 4|0);
 $44 = HEAP32[$43>>2]|0;
 FUNCTION_TABLE_vi[$44 & 8191]($decoder);
 $45 = ((($cidglyph)) + 124|0);
 $46 = HEAP32[$45>>2]|0;
 $47 = $46 & 1;
 $48 = $47 | 4;
 HEAP32[$45>>2] = $48;
 if (!($36)) {
  $49 = ((($cidglyph)) + 156|0);
  $50 = HEAP32[$49>>2]|0;
  $51 = ((($decoder)) + 32|0);
  $52 = HEAP32[$51>>2]|0;
  $53 = (_FT_RoundFix($52)|0);
  $54 = $53 >> 16;
  $55 = ((($cidglyph)) + 32|0);
  HEAP32[$55>>2] = $54;
  $56 = ((($decoder)) + 40|0);
  $57 = HEAP32[$56>>2]|0;
  $58 = (_FT_RoundFix($57)|0);
  $59 = $58 >> 16;
  $60 = ((($cidglyph)) + 40|0);
  HEAP32[$60>>2] = $59;
  $61 = ((($50)) + 12|0);
  ;HEAP32[$61>>2]=HEAP32[$font_matrix>>2]|0;HEAP32[$61+4>>2]=HEAP32[$font_matrix+4>>2]|0;HEAP32[$61+8>>2]=HEAP32[$font_matrix+8>>2]|0;HEAP32[$61+12>>2]=HEAP32[$font_matrix+12>>2]|0;
  $62 = ((($50)) + 28|0);
  HEAP32[$62>>2] = $39;
  $63 = ((($50)) + 32|0);
  HEAP32[$63>>2] = $41;
  $64 = ((($50)) + 8|0);
  HEAP8[$64>>0] = 1;
  $error$0 = 0;
  STACKTOP = sp;return ($error$0|0);
 }
 $65 = ((($cidglyph)) + 24|0);
 $66 = ((($decoder)) + 40|0);
 $67 = HEAP32[$66>>2]|0;
 $68 = (_FT_RoundFix($67)|0);
 $69 = $68 >> 16;
 $70 = ((($cidglyph)) + 40|0);
 HEAP32[$70>>2] = $69;
 $71 = HEAP32[$66>>2]|0;
 $72 = (_FT_RoundFix($71)|0);
 $73 = $72 >> 16;
 $74 = ((($cidglyph)) + 56|0);
 HEAP32[$74>>2] = $73;
 $75 = ((($cidglyph)) + 156|0);
 $76 = HEAP32[$75>>2]|0;
 $77 = ((($76)) + 8|0);
 HEAP8[$77>>0] = 0;
 $78 = ((($1)) + 208|0);
 $79 = HEAP32[$78>>2]|0;
 $80 = ((($1)) + 200|0);
 $81 = HEAP32[$80>>2]|0;
 $82 = (($79) - ($81))|0;
 $83 = $82 >> 16;
 $84 = ((($cidglyph)) + 52|0);
 HEAP32[$84>>2] = $83;
 $85 = ((($cidglyph)) + 60|0);
 HEAP32[$85>>2] = $83;
 HEAP32[$22>>2] = 1869968492;
 $86 = ((($cidsize)) + 14|0);
 $87 = HEAP16[$86>>1]|0;
 $88 = ($87&65535)<(24);
 if ($88) {
  $89 = HEAP32[$45>>2]|0;
  $90 = $89 | 256;
  HEAP32[$45>>2] = $90;
 }
 _FT_Outline_Transform($16,$font_matrix);
 _FT_Outline_Translate($16,$39,$41);
 $91 = HEAP32[$70>>2]|0;
 HEAP32[$advance>>2] = $91;
 $92 = ((($advance)) + 4|0);
 HEAP32[$92>>2] = 0;
 _FT_Vector_Transform($advance,$font_matrix);
 $93 = HEAP32[$advance>>2]|0;
 $94 = (($93) + ($39))|0;
 HEAP32[$70>>2] = $94;
 HEAP32[$advance>>2] = 0;
 $95 = HEAP32[$84>>2]|0;
 HEAP32[$92>>2] = $95;
 _FT_Vector_Transform($advance,$font_matrix);
 $96 = HEAP32[$92>>2]|0;
 $97 = (($96) + ($41))|0;
 HEAP32[$84>>2] = $97;
 if ($19) {
  $98 = ((($decoder)) + 16|0);
  $99 = HEAP32[$98>>2]|0;
  $100 = ((($99)) + 4|0);
  $101 = HEAP32[$100>>2]|0;
  $102 = HEAP32[$12>>2]|0;
  $103 = HEAP32[$15>>2]|0;
  $104 = ($28<<24>>24)==(0);
  if ($104) {
   label = 13;
  } else {
   $105 = ((($decoder)) + 72|0);
   $106 = HEAP32[$105>>2]|0;
   $107 = ($106|0)==(0|0);
   if ($107) {
    label = 13;
   }
  }
  if ((label|0) == 13) {
   $108 = ((($99)) + 2|0);
   $109 = HEAP16[$108>>1]|0;
   $110 = ($109<<16>>16)>(0);
   if ($110) {
    $111 = $109 << 16 >> 16;
    $n$02 = $111;$vec$03 = $101;
    while(1) {
     $112 = HEAP32[$vec$03>>2]|0;
     $113 = (_FT_MulFix($112,$102)|0);
     HEAP32[$vec$03>>2] = $113;
     $114 = ((($vec$03)) + 4|0);
     $115 = HEAP32[$114>>2]|0;
     $116 = (_FT_MulFix($115,$103)|0);
     HEAP32[$114>>2] = $116;
     $117 = (($n$02) + -1)|0;
     $118 = ((($vec$03)) + 8|0);
     $119 = ($n$02|0)>(1);
     if ($119) {
      $n$02 = $117;$vec$03 = $118;
     } else {
      break;
     }
    }
   }
  }
  $120 = HEAP32[$70>>2]|0;
  $121 = (_FT_MulFix($120,$102)|0);
  HEAP32[$70>>2] = $121;
  $122 = HEAP32[$84>>2]|0;
  $123 = (_FT_MulFix($122,$103)|0);
  HEAP32[$84>>2] = $123;
 }
 _FT_Outline_Get_CBox($16,$cbox);
 $124 = ((($cbox)) + 8|0);
 $125 = HEAP32[$124>>2]|0;
 $126 = HEAP32[$cbox>>2]|0;
 $127 = (($125) - ($126))|0;
 HEAP32[$65>>2] = $127;
 $128 = ((($cbox)) + 12|0);
 $129 = HEAP32[$128>>2]|0;
 $130 = ((($cbox)) + 4|0);
 $131 = HEAP32[$130>>2]|0;
 $132 = (($129) - ($131))|0;
 $133 = ((($cidglyph)) + 28|0);
 HEAP32[$133>>2] = $132;
 $134 = ((($cidglyph)) + 32|0);
 HEAP32[$134>>2] = $126;
 $135 = ((($cidglyph)) + 36|0);
 HEAP32[$135>>2] = $129;
 $136 = $load_flags$ & 16;
 $137 = ($136|0)==(0);
 if (!($137)) {
  $138 = HEAP32[$84>>2]|0;
  _ft_synthesize_vertical_metrics($65,$138);
 }
 $error$0 = 0;
 STACKTOP = sp;return ($error$0|0);
}
function _cid_load_glyph($decoder,$glyph_index) {
 $decoder = $decoder|0;
 $glyph_index = $glyph_index|0;
 var $$01$i = 0, $$01$i10 = 0, $$01$i20 = 0, $$01$i30 = 0, $$6 = 0, $$lcssa58 = 0, $$lcssa59 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0;
 var $182 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $charstring$0 = 0, $charstring$3 = 0, $error = 0, $fd_select$2 = 0, $glyph_data = 0, $glyph_length$0 = 0, $glyph_length$3 = 0, $metrics = 0, $or$cond = 0, $p$0$lcssa$i14 = 0;
 var $p$03$i = 0, $p$03$i18 = 0, $p$03$i28 = 0, $p$03$i8 = 0, $result$0$lcssa$i = 0, $result$0$lcssa$i15 = 0, $result$0$lcssa$i2547 = 0, $result$0$lcssa$i35 = 0, $result$02$i = 0, $result$02$i19 = 0, $result$02$i29 = 0, $result$02$i9 = 0, $scevgep$i12 = 0, $scevgep$i22 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp + 24|0;
 $glyph_data = sp + 16|0;
 $metrics = sp;
 $0 = ((($decoder)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 328|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$error>>2] = 0;
 $4 = ((($1)) + 100|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($1)) + 136|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($1)) + 128|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 48|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)!=(0|0);
 do {
  if ($12) {
   $13 = HEAP32[$11>>2]|0;
   $14 = HEAP32[$13>>2]|0;
   $15 = ((($11)) + 4|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = (FUNCTION_TABLE_iiii[$14 & 8191]($16,$glyph_index,$glyph_data)|0);
   HEAP32[$error>>2] = $17;
   $18 = ($17|0)==(0);
   if ($18) {
    $19 = ((($1)) + 288|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = $20&255;
    $22 = ($21<<24>>24)==(0);
    if ($22) {
     $result$0$lcssa$i = 0;
    } else {
     $23 = HEAP32[$glyph_data>>2]|0;
     $$01$i = $21;$p$03$i = $23;$result$02$i = 0;
     while(1) {
      $24 = $result$02$i << 8;
      $25 = ((($p$03$i)) + 1|0);
      $26 = HEAP8[$p$03$i>>0]|0;
      $27 = $26&255;
      $28 = $27 | $24;
      $29 = (($$01$i) + -1)<<24>>24;
      $30 = ($29<<24>>24)==(0);
      if ($30) {
       $result$0$lcssa$i = $28;
       break;
      } else {
       $$01$i = $29;$p$03$i = $25;$result$02$i = $28;
      }
     }
    }
    $31 = ((($glyph_data)) + 4|0);
    $32 = HEAP32[$31>>2]|0;
    $33 = ($32|0)==(0);
    if ($33) {
     $charstring$0 = 0;$glyph_length$0 = 0;
    } else {
     $34 = (($32) - ($20))|0;
     $35 = (_ft_mem_alloc($5,$34,$error)|0);
     $36 = HEAP32[$error>>2]|0;
     $37 = ($36|0)==(0);
     if ($37) {
      $38 = HEAP32[$glyph_data>>2]|0;
      $39 = HEAP32[$19>>2]|0;
      $40 = (($38) + ($39)|0);
      _memcpy(($35|0),($40|0),($34|0))|0;
      $charstring$0 = $35;$glyph_length$0 = $34;
     } else {
      $charstring$0 = $35;$glyph_length$0 = $34;
     }
    }
    $41 = HEAP32[$11>>2]|0;
    $42 = ((($41)) + 4|0);
    $43 = HEAP32[$42>>2]|0;
    $44 = HEAP32[$15>>2]|0;
    FUNCTION_TABLE_vii[$43 & 8191]($44,$glyph_data);
    $45 = HEAP32[$error>>2]|0;
    $46 = ($45|0)==(0);
    if ($46) {
     $charstring$3 = $charstring$0;$fd_select$2 = $result$0$lcssa$i;$glyph_length$3 = $glyph_length$0;
     break;
    } else {
     $182 = $45;
    }
   } else {
    $182 = $17;
   }
   $181 = $182;
   STACKTOP = sp;return ($181|0);
  } else {
   $47 = ((($1)) + 288|0);
   $48 = HEAP32[$47>>2]|0;
   $49 = ((($1)) + 292|0);
   $50 = HEAP32[$49>>2]|0;
   $51 = (($50) + ($48))|0;
   $52 = ((($1)) + 308|0);
   $53 = HEAP32[$52>>2]|0;
   $54 = ((($1)) + 284|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = (($55) + ($53))|0;
   $57 = Math_imul($51, $glyph_index)|0;
   $58 = (($56) + ($57))|0;
   $59 = (_FT_Stream_Seek($3,$58)|0);
   HEAP32[$error>>2] = $59;
   $60 = ($59|0)==(0);
   if (!($60)) {
    $181 = $59;
    STACKTOP = sp;return ($181|0);
   }
   $61 = $51 << 1;
   $62 = (_FT_Stream_EnterFrame($3,$61)|0);
   HEAP32[$error>>2] = $62;
   $63 = ($62|0)==(0);
   if (!($63)) {
    $181 = $62;
    STACKTOP = sp;return ($181|0);
   }
   $64 = ((($3)) + 32|0);
   $65 = HEAP32[$64>>2]|0;
   $66 = HEAP32[$47>>2]|0;
   $67 = $66&255;
   $68 = ($67<<24>>24)==(0);
   if ($68) {
    $p$0$lcssa$i14 = $65;$result$0$lcssa$i15 = 0;
   } else {
    $69 = (($66) + 255)|0;
    $70 = $69 & 255;
    $$01$i10 = $67;$p$03$i8 = $65;$result$02$i9 = 0;
    while(1) {
     $71 = $result$02$i9 << 8;
     $72 = ((($p$03$i8)) + 1|0);
     $73 = HEAP8[$p$03$i8>>0]|0;
     $74 = $73&255;
     $75 = $74 | $71;
     $76 = (($$01$i10) + -1)<<24>>24;
     $77 = ($76<<24>>24)==(0);
     if ($77) {
      $$lcssa59 = $75;
      break;
     } else {
      $$01$i10 = $76;$p$03$i8 = $72;$result$02$i9 = $75;
     }
    }
    $78 = (($70) + 1)|0;
    $scevgep$i12 = (($65) + ($78)|0);
    $p$0$lcssa$i14 = $scevgep$i12;$result$0$lcssa$i15 = $$lcssa59;
   }
   $79 = HEAP32[$49>>2]|0;
   $80 = $79&255;
   $81 = ($80<<24>>24)==(0);
   if ($81) {
    $result$0$lcssa$i2547 = 0;$result$0$lcssa$i35 = 0;
   } else {
    $82 = (($79) + 255)|0;
    $83 = $82 & 255;
    $$01$i20 = $80;$p$03$i18 = $p$0$lcssa$i14;$result$02$i19 = 0;
    while(1) {
     $84 = $result$02$i19 << 8;
     $85 = ((($p$03$i18)) + 1|0);
     $86 = HEAP8[$p$03$i18>>0]|0;
     $87 = $86&255;
     $88 = $87 | $84;
     $89 = (($$01$i20) + -1)<<24>>24;
     $90 = ($89<<24>>24)==(0);
     if ($90) {
      $$lcssa58 = $88;
      break;
     } else {
      $$01$i20 = $89;$p$03$i18 = $85;$result$02$i19 = $88;
     }
    }
    if ($81) {
     $result$0$lcssa$i2547 = $$lcssa58;$result$0$lcssa$i35 = 0;
    } else {
     $91 = (($83) + 1)|0;
     $scevgep$i22 = (($p$0$lcssa$i14) + ($91)|0);
     $92 = (($scevgep$i22) + ($66)|0);
     $$01$i30 = $80;$p$03$i28 = $92;$result$02$i29 = 0;
     while(1) {
      $93 = $result$02$i29 << 8;
      $94 = ((($p$03$i28)) + 1|0);
      $95 = HEAP8[$p$03$i28>>0]|0;
      $96 = $95&255;
      $97 = $96 | $93;
      $98 = (($$01$i30) + -1)<<24>>24;
      $99 = ($98<<24>>24)==(0);
      if ($99) {
       $result$0$lcssa$i2547 = $$lcssa58;$result$0$lcssa$i35 = $97;
       break;
      } else {
       $$01$i30 = $98;$p$03$i28 = $94;$result$02$i29 = $97;
      }
     }
    }
   }
   $100 = (($result$0$lcssa$i35) - ($result$0$lcssa$i2547))|0;
   _FT_Stream_ExitFrame($3);
   $101 = ((($1)) + 300|0);
   $102 = HEAP32[$101>>2]|0;
   $103 = ($result$0$lcssa$i15>>>0)<($102>>>0);
   if (!($103)) {
    HEAP32[$error>>2] = 9;
    $181 = 9;
    STACKTOP = sp;return ($181|0);
   }
   $104 = ($result$0$lcssa$i35|0)==($result$0$lcssa$i2547|0);
   if ($104) {
    $181 = 0;
    STACKTOP = sp;return ($181|0);
   }
   $105 = (_ft_mem_alloc($5,$100,$error)|0);
   $106 = HEAP32[$error>>2]|0;
   $107 = ($106|0)==(0);
   if (!($107)) {
    $181 = $106;
    STACKTOP = sp;return ($181|0);
   }
   $108 = HEAP32[$52>>2]|0;
   $109 = (($108) + ($result$0$lcssa$i2547))|0;
   $110 = (_FT_Stream_ReadAt($3,$109,$105,$100)|0);
   HEAP32[$error>>2] = $110;
   $111 = ($110|0)==(0);
   if ($111) {
    $charstring$3 = $105;$fd_select$2 = $result$0$lcssa$i15;$glyph_length$3 = $100;
   } else {
    $181 = $110;
    STACKTOP = sp;return ($181|0);
   }
  }
 } while(0);
 $112 = ((($1)) + 316|0);
 $113 = HEAP32[$112>>2]|0;
 $114 = (($113) + ($fd_select$2<<3)|0);
 $115 = HEAP32[$114>>2]|0;
 $116 = ((($decoder)) + 1364|0);
 HEAP32[$116>>2] = $115;
 $117 = (((($113) + ($fd_select$2<<3)|0)) + 4|0);
 $118 = HEAP32[$117>>2]|0;
 $119 = ((($decoder)) + 1368|0);
 HEAP32[$119>>2] = $118;
 $120 = ((($decoder)) + 1372|0);
 HEAP32[$120>>2] = 0;
 $121 = ((($1)) + 304|0);
 $122 = HEAP32[$121>>2]|0;
 $123 = ((($decoder)) + 1376|0);
 $124 = (((($122) + (($fd_select$2*252)|0)|0)) + 216|0);
 ;HEAP32[$123>>2]=HEAP32[$124>>2]|0;HEAP32[$123+4>>2]=HEAP32[$124+4>>2]|0;HEAP32[$123+8>>2]=HEAP32[$124+8>>2]|0;HEAP32[$123+12>>2]=HEAP32[$124+12>>2]|0;
 $125 = ((($decoder)) + 1392|0);
 $126 = (((($122) + (($fd_select$2*252)|0)|0)) + 232|0);
 $127 = $126;
 $128 = $127;
 $129 = HEAP32[$128>>2]|0;
 $130 = (($127) + 4)|0;
 $131 = $130;
 $132 = HEAP32[$131>>2]|0;
 $133 = $125;
 $134 = $133;
 HEAP32[$134>>2] = $129;
 $135 = (($133) + 4)|0;
 $136 = $135;
 HEAP32[$136>>2] = $132;
 $137 = (((($122) + (($fd_select$2*252)|0)|0)) + 4|0);
 $138 = HEAP32[$137>>2]|0;
 $139 = ((($decoder)) + 1360|0);
 HEAP32[$139>>2] = $138;
 $140 = ($138|0)>(-1);
 $$6 = $140 ? $138 : 0;
 if ($140) {
  $141 = ((($7)) + 16|0);
  $142 = HEAP32[$141>>2]|0;
  FUNCTION_TABLE_viii[$142 & 8191]($charstring$3,$glyph_length$3,4330);
 }
 $143 = ((($decoder)) + 1484|0);
 $144 = HEAP32[$143>>2]|0;
 $145 = (($charstring$3) + ($$6)|0);
 $146 = (($glyph_length$3) - ($$6))|0;
 $147 = (FUNCTION_TABLE_iiii[$144 & 8191]($decoder,$145,$146)|0);
 HEAP32[$error>>2] = $147;
 _ft_mem_free($5,$charstring$3);
 $148 = HEAP32[$error>>2]|0;
 $149 = ($148|0)==(0);
 $or$cond = $12 & $149;
 if (!($or$cond)) {
  $181 = $148;
  STACKTOP = sp;return ($181|0);
 }
 $150 = HEAP32[$11>>2]|0;
 $151 = ((($150)) + 8|0);
 $152 = HEAP32[$151>>2]|0;
 $153 = ($152|0)==(0|0);
 if ($153) {
  $181 = 0;
  STACKTOP = sp;return ($181|0);
 }
 $154 = ((($decoder)) + 32|0);
 $155 = HEAP32[$154>>2]|0;
 $156 = (_FT_RoundFix($155)|0);
 $157 = $156 >> 16;
 HEAP32[$metrics>>2] = $157;
 $158 = ((($metrics)) + 4|0);
 HEAP32[$158>>2] = 0;
 $159 = ((($decoder)) + 40|0);
 $160 = HEAP32[$159>>2]|0;
 $161 = (_FT_RoundFix($160)|0);
 $162 = $161 >> 16;
 $163 = ((($metrics)) + 8|0);
 HEAP32[$163>>2] = $162;
 $164 = ((($decoder)) + 44|0);
 $165 = HEAP32[$164>>2]|0;
 $166 = (_FT_RoundFix($165)|0);
 $167 = $166 >> 16;
 $168 = ((($metrics)) + 12|0);
 HEAP32[$168>>2] = $167;
 $169 = HEAP32[$11>>2]|0;
 $170 = ((($169)) + 8|0);
 $171 = HEAP32[$170>>2]|0;
 $172 = ((($11)) + 4|0);
 $173 = HEAP32[$172>>2]|0;
 $174 = (FUNCTION_TABLE_iiiii[$171 & 8191]($173,$glyph_index,0,$metrics)|0);
 HEAP32[$error>>2] = $174;
 $175 = HEAP32[$metrics>>2]|0;
 $176 = $175 << 16;
 HEAP32[$154>>2] = $176;
 $177 = HEAP32[$163>>2]|0;
 $178 = $177 << 16;
 HEAP32[$159>>2] = $178;
 $179 = HEAP32[$168>>2]|0;
 $180 = $179 << 16;
 HEAP32[$164>>2] = $180;
 $181 = $174;
 STACKTOP = sp;return ($181|0);
}
function _cid_size_request($size,$req) {
 $size = $size|0;
 $req = $req|0;
 var $$idx$val = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$size>>2]|0;
 _FT_Request_Metrics($0,$req);
 $$idx$val = HEAP32[$size>>2]|0;
 $1 = ((($$idx$val)) + 320|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($$idx$val)) + 96|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (_FT_Get_Module($6,593918)|0);
 $8 = ($7|0)!=(0|0);
 $9 = ($2|0)!=(0|0);
 $or$cond$i = $9 & $8;
 if (!($or$cond$i)) {
  return 0;
 }
 $10 = HEAP32[$2>>2]|0;
 $11 = ($10|0)==(0|0);
 if ($11) {
  return 0;
 }
 $12 = (FUNCTION_TABLE_ii[$10 & 8191]($7)|0);
 $13 = ($12|0)==(0|0);
 if ($13) {
  return 0;
 }
 $14 = ((($12)) + 4|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($size)) + 40|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ((($size)) + 16|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ((($size)) + 20|0);
 $21 = HEAP32[$20>>2]|0;
 FUNCTION_TABLE_viiiii[$15 & 8191]($17,$19,$21,0,0);
 return 0;
}
function _pcf_driver_requester($module,$name) {
 $module = $module|0;
 $name = $name|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_ft_service_list_lookup(53264,$name)|0);
 return ($0|0);
}
function _pcf_get_charset_id($face,$acharset_encoding,$acharset_registry) {
 $face = $face|0;
 $acharset_encoding = $acharset_encoding|0;
 $acharset_registry = $acharset_registry|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 176|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$acharset_encoding>>2] = $1;
 $2 = ((($face)) + 180|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$acharset_registry>>2] = $3;
 return 0;
}
function _pcf_get_bdf_property($face,$prop_name,$aproperty) {
 $face = $face|0;
 $prop_name = $prop_name|0;
 $aproperty = $aproperty|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $i$02$i = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 284|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 280|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)>(0);
 if ($4) {
  $i$02$i = 0;
 } else {
  $$0 = 6;
  return ($$0|0);
 }
 while(1) {
  $5 = (($1) + (($i$02$i*12)|0)|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (_strcmp($6,$prop_name)|0);
  $8 = (($i$02$i) + 1)|0;
  $9 = ($8|0)<($3|0);
  $10 = ($7|0)!=(0);
  $or$cond$i = $9 & $10;
  if ($or$cond$i) {
   $i$02$i = $8;
  } else {
   break;
  }
 }
 $11 = (($1) + (($8*12)|0)|0);
 $12 = ((($11)) + -12|0);
 if ($10) {
  $$0 = 6;
  return ($$0|0);
 }
 $13 = ((($12)) + 4|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = ($14<<24>>24)==(0);
 if ($15) {
  $19 = ((($12)) + 8|0);
  HEAP32[$aproperty>>2] = 2;
  $20 = HEAP32[$19>>2]|0;
  $21 = ((($aproperty)) + 4|0);
  HEAP32[$21>>2] = $20;
  $$0 = 0;
  return ($$0|0);
 } else {
  HEAP32[$aproperty>>2] = 1;
  $16 = ((($12)) + 8|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = ((($aproperty)) + 4|0);
  HEAP32[$18>>2] = $17;
  $$0 = 0;
  return ($$0|0);
 }
 return (0)|0;
}
function _PCF_Face_Init($stream,$pcfface,$face_index,$num_params,$params) {
 $stream = $stream|0;
 $pcfface = $pcfface|0;
 $face_index = $face_index|0;
 $num_params = $num_params|0;
 $params = $params|0;
 var $$0 = 0, $$not = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $charmap = 0;
 var $or$cond = 0, $or$cond5 = 0, $unicode_charmap$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $charmap = sp;
 $0 = (_pcf_load_font($stream,$pcfface)|0);
 $1 = ($0|0)==(0);
 L1: do {
  if (!($1)) {
   _PCF_Face_Done($pcfface);
   $2 = ((($pcfface)) + 132|0);
   $3 = (_FT_Stream_OpenGzip($2,$stream)|0);
   $4 = $3 & 255;
   $5 = ($4|0)==(7);
   do {
    if (!($5)) {
     $6 = ($3|0)==(0);
     if (!($6)) {
      $7 = (_FT_Stream_OpenLZW($2,$stream)|0);
      $8 = $7 & 255;
      $$not = ($8|0)!=(7);
      $9 = ($7|0)==(0);
      $or$cond5 = $9 & $$not;
      if (!($or$cond5)) {
       break;
      }
     }
     $10 = ((($pcfface)) + 172|0);
     HEAP32[$10>>2] = $stream;
     $11 = ((($pcfface)) + 104|0);
     HEAP32[$11>>2] = $2;
     $12 = (_pcf_load_font($2,$pcfface)|0);
     $13 = ($12|0)==(0);
     if ($13) {
      break L1;
     }
    }
   } while(0);
   _PCF_Face_Done($pcfface);
   $$0 = 2;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 $14 = ($face_index|0)>(0);
 if ($14) {
  _PCF_Face_Done($pcfface);
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $15 = ((($pcfface)) + 180|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($pcfface)) + 176|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ($16|0)!=(0|0);
 $20 = ($18|0)!=(0|0);
 $or$cond = $19 & $20;
 L15: do {
  if ($or$cond) {
   $21 = HEAP8[$16>>0]|0;
   switch ($21<<24>>24) {
   case 73: case 105:  {
    break;
   }
   default: {
    $unicode_charmap$1 = 1;
    break L15;
   }
   }
   $22 = ((($16)) + 1|0);
   $23 = HEAP8[$22>>0]|0;
   switch ($23<<24>>24) {
   case 83: case 115:  {
    break;
   }
   default: {
    $unicode_charmap$1 = 1;
    break L15;
   }
   }
   $24 = ((($16)) + 2|0);
   $25 = HEAP8[$24>>0]|0;
   switch ($25<<24>>24) {
   case 79: case 111:  {
    break;
   }
   default: {
    $unicode_charmap$1 = 1;
    break L15;
   }
   }
   $26 = ((($16)) + 3|0);
   $27 = (_strcmp($26,532544)|0);
   $28 = ($27|0)==(0);
   if (!($28)) {
    $29 = (_strcmp($26,532550)|0);
    $30 = ($29|0)==(0);
    if (!($30)) {
     $unicode_charmap$1 = 1;
     break;
    }
    $31 = (_strcmp($18,532555)|0);
    $32 = ($31|0)==(0);
    if (!($32)) {
     $unicode_charmap$1 = 1;
     break;
    }
   }
   $unicode_charmap$1 = 0;
  } else {
   $unicode_charmap$1 = 1;
  }
 } while(0);
 HEAP32[$charmap>>2] = $pcfface;
 $33 = ((($charmap)) + 4|0);
 HEAP32[$33>>2] = 0;
 $34 = ((($charmap)) + 8|0);
 HEAP16[$34>>1] = 0;
 $35 = ((($charmap)) + 10|0);
 HEAP16[$35>>1] = 0;
 if (!($unicode_charmap$1)) {
  HEAP32[$33>>2] = 1970170211;
  HEAP16[$34>>1] = 3;
  HEAP16[$35>>1] = 1;
 }
 $36 = (_FT_CMap_New(53296,0,$charmap,0)|0);
 $$0 = $36;
 STACKTOP = sp;return ($$0|0);
}
function _pcf_load_font($stream,$face) {
 $stream = $stream|0;
 $face = $face|0;
 var $$0$i$ph$i = 0, $$0$i$ph$i15 = 0, $$0$i$ph$i25 = 0, $$0$i$ph$i37 = 0, $$0$i20$ph = 0, $$0$i33$ph = 0, $$0$i44$ph = 0, $$01$i$ph = 0, $$lcssa373 = 0, $$lcssa379 = 0, $$lcssa380 = 0, $$phi$trans$insert = 0, $$pr$pre = 0, $$pre = 0, $$pre$i = 0, $$pre$i18 = 0, $$pre$phi12$iZ2D = 0, $$pre10$i = 0, $$pre11$i = 0, $$pre216 = 0;
 var $$pre217 = 0, $$pre218 = 0, $$pre22$i = 0, $$pre23$i = 0, $$sink$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0;
 var $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0;
 var $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0;
 var $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0;
 var $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0;
 var $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0;
 var $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0;
 var $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0;
 var $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0;
 var $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0;
 var $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0;
 var $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0;
 var $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0;
 var $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0;
 var $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0;
 var $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0;
 var $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0;
 var $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0;
 var $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0;
 var $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0;
 var $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0;
 var $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0;
 var $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0;
 var $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0;
 var $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0;
 var $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0;
 var $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0;
 var $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0;
 var $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0;
 var $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0;
 var $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0;
 var $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $compr$i$i = 0, $encoding$0$i = 0, $error = 0, $error$i = 0, $error$i12 = 0, $error$i22 = 0, $error$i34 = 0, $error$i45 = 0, $error$i5 = 0, $exitcond$3$i = 0, $exitcond$i = 0, $exitcond$i50 = 0, $exitcond14$i = 0, $exitcond15$i = 0, $firstCol$0$i = 0, $firstRow$0$i = 0;
 var $have_change$09$i = 0, $have_change$1$i = 0, $have_change$1$i$lcssa378 = 0, $i$01$i = 0, $i$010$i = 0, $i$010$i41 = 0, $i$010$us$i = 0, $i$010$us24$i = 0, $i$02$i = 0, $i$02$i$i = 0, $i$02$i$i$lcssa = 0, $i$02$i$i13 = 0, $i$02$i$i13$lcssa = 0, $i$02$i$i23 = 0, $i$02$i$i23$lcssa = 0, $i$02$i$i35 = 0, $i$02$i$i35$lcssa = 0, $i$02$i$i46 = 0, $i$02$i11$i = 0, $i$02$i17$i = 0;
 var $i$02$i5$i = 0, $i$02$i55 = 0, $i$02$i61 = 0, $i$02$i67 = 0, $i$02$i73 = 0, $i$02$i79 = 0, $i$02$i85 = 0, $i$02$i91 = 0, $i$03$i = 0, $i$03$us$i = 0, $i$07$i = 0, $i$07$us$i = 0, $i$08$i = 0, $i$19$i = 0, $i$25$i = 0, $j$08$i = 0, $j$08$us$us38$i = 0, $k$0$lcssa$i = 0, $k$012$i = 0, $k$012$us23$i = 0;
 var $k$19$i = 0, $k$19$us$us37$i = 0, $k$2$i = 0, $k$2$i$lcssa = 0, $k$2$us$us40$i = 0, $k$2$us$us40$i$lcssa = 0, $lastCol$0$i = 0, $lastRow$0$i = 0, $len$1$1$i = 0, $len$1$2$i = 0, $len$1$3$i = 0, $len$1$i = 0, $len$2$i = 0, $lengths$i$sroa$10$0 = 0, $lengths$i$sroa$10$0108109 = 0, $lengths$i$sroa$10$0108110 = 0, $lengths$i$sroa$13$0 = 0, $lengths$i$sroa$13$0115 = 0, $lengths$i$sroa$13$0116 = 0, $lengths$i$sroa$13$0117 = 0;
 var $metrics$02$i = 0, $metrics$02$i$lcssa = 0, $metrics$02$us$i = 0, $mm$037$3$i = 0, $mm$037$i = 0, $n$017$i = 0, $n$113$i = 0, $n$27$i = 0, $nbitmaps$0$i = 0, $nmetrics$0$i = 0, $nprops$0$i = 0, $or$cond = 0, $or$cond$i = 0, $or$cond$i$i = 0, $or$cond$i12$i = 0, $or$cond$i18$i = 0, $or$cond$i30 = 0, $or$cond$i39 = 0, $or$cond$i53 = 0, $or$cond$i56 = 0;
 var $or$cond$i6 = 0, $or$cond$i6$i = 0, $or$cond$i62 = 0, $or$cond$i68 = 0, $or$cond$i74 = 0, $or$cond$i80 = 0, $or$cond$i86 = 0, $or$cond$i92 = 0, $or$cond2$i = 0, $or$cond3$i = 0, $or$cond3$i40 = 0, $or$cond4 = 0, $or$cond5$i = 0, $phitmp = 0, $phitmp221 = 0, $phitmp3 = 0, $props$0$i = 0, $resolution_x$0$ph$ph = 0, $resolution_y$0 = 0, $s$1$1$i = 0;
 var $s$1$2$i = 0, $s$1$3$i = 0, $s$2$1$i = 0, $s$2$2$i = 0, $s$2$3$i = 0, $s$2$i = 0, $sext = 0, $sext220 = 0, $string_size$0$i = 0, $strings$0$i = 0, $strings$i$sroa$10$0 = 0, $strings$i$sroa$10$2102105106113 = 0, $strings$i$sroa$10$2102105106114 = 0, $strings$i$sroa$8$0 = 0, $strings$i$sroa$8$2103104107111 = 0, $strings$i$sroa$8$2103104107112 = 0, $tables$016$i = 0, $tables$1$lcssa$i = 0, $tables$16$i = 0, $tmp$i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i45 = sp + 44|0;
 $error$i34 = sp + 40|0;
 $error$i22 = sp + 36|0;
 $compr$i$i = sp + 48|0;
 $error$i12 = sp + 32|0;
 $error$i5 = sp + 28|0;
 $error$i = sp + 24|0;
 $tmp$i = sp + 8|0;
 $error = sp;
 $0 = ((($face)) + 100|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 184|0);
 $3 = (_FT_Stream_Seek($stream,0)|0);
 HEAP32[$error$i>>2] = $3;
 $4 = ($3|0)==(0);
 L1: do {
  if ($4) {
   $5 = (_FT_Stream_ReadFields($stream,171848,$2)|0);
   HEAP32[$error$i>>2] = $5;
   $6 = ($5|0)==(0);
   if ($6) {
    $7 = HEAP32[$2>>2]|0;
    $8 = ($7|0)==(1885562369);
    if ($8) {
     $9 = ((($face)) + 188|0);
     $10 = HEAP32[$9>>2]|0;
     $11 = (($10) + -1)|0;
     $12 = ($11>>>0)>(134217726);
     if ($12) {
      $$01$i$ph = 3;
      label = 31;
     } else {
      $13 = (_ft_mem_realloc($1,16,0,$10,0,$error$i)|0);
      $14 = ((($face)) + 192|0);
      HEAP32[$14>>2] = $13;
      $15 = HEAP32[$error$i>>2]|0;
      $16 = ($15|0)==(0);
      if ($16) {
       $17 = HEAP32[$9>>2]|0;
       $18 = ($17|0)==(0);
       L7: do {
        if ($18) {
         $28 = $13;$743 = 0;
         label = 11;
        } else {
         $n$017$i = 0;$tables$016$i = $13;
         while(1) {
          $19 = (_FT_Stream_ReadFields($stream,171864,$tables$016$i)|0);
          HEAP32[$error$i>>2] = $19;
          $20 = ($19|0)==(0);
          if (!($20)) {
           label = 33;
           break L7;
          }
          $21 = ((($tables$016$i)) + 16|0);
          $22 = (($n$017$i) + 1)|0;
          $23 = HEAP32[$9>>2]|0;
          $24 = ($22>>>0)<($23>>>0);
          if ($24) {
           $n$017$i = $22;$tables$016$i = $21;
          } else {
           $$lcssa380 = $23;
           break;
          }
         }
         $$pre$i = HEAP32[$14>>2]|0;
         $25 = ($$lcssa380|0)==(1);
         if ($25) {
          $26 = ((($stream)) + 4|0);
          $27 = HEAP32[$26>>2]|0;
          $72 = $27;$744 = 1;$745 = $$pre$i;$tables$1$lcssa$i = $$pre$i;
          label = 27;
         } else {
          $28 = $$pre$i;$743 = $$lcssa380;
          label = 11;
         }
        }
       } while(0);
       L13: do {
        if ((label|0) == 11) {
         $$phi$trans$insert = ((($28)) + 12|0);
         $32 = $743;$n$113$i = 0;
         L15: while(1) {
          $29 = $n$113$i ^ -1;
          $30 = (($n$113$i) + 1)|0;
          $31 = ($32|0)==($30|0);
          if ($31) {
           $57 = $32;
           break;
          }
          $$pre = HEAP32[$$phi$trans$insert>>2]|0;
          $38 = $$pre;$have_change$09$i = 0;$i$08$i = 0;
          while(1) {
           $33 = (((($28) + ($i$08$i<<4)|0)) + 12|0);
           $34 = (($i$08$i) + 1)|0;
           $35 = (((($28) + ($34<<4)|0)) + 12|0);
           $36 = HEAP32[$35>>2]|0;
           $37 = ($38>>>0)>($36>>>0);
           if ($37) {
            $39 = (($28) + ($34<<4)|0);
            $40 = (($28) + ($i$08$i<<4)|0);
            ;HEAP32[$tmp$i>>2]=HEAP32[$40>>2]|0;HEAP32[$tmp$i+4>>2]=HEAP32[$40+4>>2]|0;HEAP32[$tmp$i+8>>2]=HEAP32[$40+8>>2]|0;HEAP32[$tmp$i+12>>2]=HEAP32[$40+12>>2]|0;
            ;HEAP32[$40>>2]=HEAP32[$39>>2]|0;HEAP32[$40+4>>2]=HEAP32[$39+4>>2]|0;HEAP32[$40+8>>2]=HEAP32[$39+8>>2]|0;HEAP32[$40+12>>2]=HEAP32[$39+12>>2]|0;
            ;HEAP32[$39>>2]=HEAP32[$tmp$i>>2]|0;HEAP32[$39+4>>2]=HEAP32[$tmp$i+4>>2]|0;HEAP32[$39+8>>2]=HEAP32[$tmp$i+8>>2]|0;HEAP32[$39+12>>2]=HEAP32[$tmp$i+12>>2]|0;
            $$pre22$i = HEAP32[$35>>2]|0;
            $44 = $$pre22$i;$have_change$1$i = 1;
           } else {
            $44 = $36;$have_change$1$i = $have_change$09$i;
           }
           $41 = (((($28) + ($i$08$i<<4)|0)) + 8|0);
           $42 = HEAP32[$41>>2]|0;
           $43 = ($44>>>0)<($42>>>0);
           if ($43) {
            label = 19;
            break L15;
           }
           $45 = HEAP32[$33>>2]|0;
           $46 = (($44) - ($42))|0;
           $47 = ($45>>>0)>($46>>>0);
           if ($47) {
            label = 19;
            break L15;
           }
           $48 = HEAP32[$9>>2]|0;
           $49 = (($48) + ($29))|0;
           $50 = ($34>>>0)<($49>>>0);
           if ($50) {
            $38 = $44;$have_change$09$i = $have_change$1$i;$i$08$i = $34;
           } else {
            $$lcssa379 = $48;$have_change$1$i$lcssa378 = $have_change$1$i;
            break;
           }
          }
          $51 = ($have_change$1$i$lcssa378|0)!=(0);
          $52 = (($$lcssa379) + -1)|0;
          $53 = ($30>>>0)<($52>>>0);
          $or$cond$i = $51 & $53;
          if ($or$cond$i) {
           $32 = $$lcssa379;$n$113$i = $30;
          } else {
           $57 = $$lcssa379;
           break;
          }
         }
         if ((label|0) == 19) {
          HEAP32[$error$i>>2] = 9;
          label = 33;
          break;
         }
         $$pre23$i = HEAP32[$14>>2]|0;
         $54 = ((($stream)) + 4|0);
         $55 = HEAP32[$54>>2]|0;
         $56 = ($57|0)==(1);
         if ($56) {
          $72 = $55;$744 = $57;$745 = $$pre23$i;$tables$1$lcssa$i = $$pre23$i;
          label = 27;
         } else {
          $58 = (($57) + -1)|0;
          $n$27$i = 0;$tables$16$i = $$pre23$i;
          while(1) {
           $59 = ((($tables$16$i)) + 8|0);
           $60 = HEAP32[$59>>2]|0;
           $61 = ($55>>>0)<($60>>>0);
           if ($61) {
            break;
           }
           $62 = ((($tables$16$i)) + 12|0);
           $63 = HEAP32[$62>>2]|0;
           $64 = (($55) - ($60))|0;
           $65 = ($63>>>0)>($64>>>0);
           if ($65) {
            break;
           }
           $66 = ((($tables$16$i)) + 16|0);
           $67 = (($n$27$i) + 1)|0;
           $68 = ($67>>>0)<($58>>>0);
           if ($68) {
            $n$27$i = $67;$tables$16$i = $66;
           } else {
            $72 = $55;$744 = $57;$745 = $$pre23$i;$tables$1$lcssa$i = $66;
            label = 27;
            break L13;
           }
          }
          HEAP32[$error$i>>2] = 8;
          label = 33;
         }
        }
       } while(0);
       do {
        if ((label|0) == 27) {
         $69 = ((($tables$1$lcssa$i)) + 12|0);
         $70 = HEAP32[$69>>2]|0;
         $71 = ($72>>>0)<($70>>>0);
         if ($71) {
          HEAP32[$error$i>>2] = 8;
          label = 33;
          break;
         }
         $73 = ((($tables$1$lcssa$i)) + 8|0);
         $74 = HEAP32[$73>>2]|0;
         $75 = (($72) - ($70))|0;
         $76 = ($74>>>0)>($75>>>0);
         if ($76) {
          HEAP32[$73>>2] = $75;
         }
         HEAP32[$error>>2] = 0;
         $82 = $744;$84 = $745;
        }
       } while(0);
       if ((label|0) == 33) {
        $77 = HEAP32[$14>>2]|0;
        _ft_mem_free($1,$77);
        HEAP32[$14>>2] = 0;
        $78 = HEAP32[$error$i>>2]|0;
        HEAP32[$error>>2] = $78;
        $79 = ($78|0)==(0);
        if (!($79)) {
         break;
        }
        $$pre216 = HEAP32[$9>>2]|0;
        $82 = $$pre216;$84 = 0;
       }
       $80 = HEAP32[$0>>2]|0;
       $81 = ($82|0)==(0);
       L47: do {
        if ($81) {
         $$0$i$ph$i = 3;
         label = 40;
        } else {
         $i$02$i$i = 0;
         while(1) {
          $83 = (($84) + ($i$02$i$i<<4)|0);
          $85 = HEAP32[$83>>2]|0;
          $86 = ($85|0)==(1);
          if ($86) {
           $i$02$i$i$lcssa = $i$02$i$i;
           break;
          }
          $95 = (($i$02$i$i) + 1)|0;
          $96 = ($95>>>0)<($82>>>0);
          if ($96) {
           $i$02$i$i = $95;
          } else {
           $$0$i$ph$i = 3;
           label = 40;
           break L47;
          }
         }
         $87 = ((($stream)) + 8|0);
         $88 = HEAP32[$87>>2]|0;
         $89 = (((($84) + ($i$02$i$i$lcssa<<4)|0)) + 12|0);
         $90 = HEAP32[$89>>2]|0;
         $91 = ($90>>>0)<($88>>>0);
         if ($91) {
          $$0$i$ph$i = 83;
          label = 40;
         } else {
          $92 = (($90) - ($88))|0;
          $93 = (_FT_Stream_Skip($stream,$92)|0);
          $94 = ($93|0)==(0);
          if ($94) {
           $97 = (((($84) + ($i$02$i$i$lcssa<<4)|0)) + 8|0);
           $98 = HEAP32[$97>>2]|0;
           HEAP32[$error$i5>>2] = 0;
           $99 = (_FT_Stream_ReadULongLE($stream,$error$i5)|0);
           $100 = HEAP32[$error$i5>>2]|0;
           $101 = ($100|0)==(0);
           $102 = ($99>>>0)<(256);
           $or$cond$i6 = $102 & $101;
           if ($or$cond$i6) {
            $103 = $99 & 4;
            $104 = ($103|0)==(0);
            if ($104) {
             $106 = (_FT_Stream_ReadULongLE($stream,$error$i5)|0);
             $nprops$0$i = $106;
            } else {
             $105 = (_FT_Stream_ReadULong($stream,$error$i5)|0);
             $nprops$0$i = $105;
            }
            $107 = HEAP32[$error$i5>>2]|0;
            $108 = ($107|0)==(0);
            if ($108) {
             $109 = (($98>>>0) / 9)&-1;
             $110 = ($nprops$0$i>>>0)>($109>>>0);
             if ($110) {
              HEAP32[$error$i5>>2] = 8;
              $props$0$i = 0;$strings$0$i = 0;
              break;
             }
             $111 = ((($face)) + 280|0);
             HEAP32[$111>>2] = $nprops$0$i;
             $112 = (_ft_mem_realloc($80,12,0,$nprops$0$i,0,$error$i5)|0);
             $113 = HEAP32[$error$i5>>2]|0;
             $114 = ($113|0)==(0);
             if ($114) {
              $115 = ($nprops$0$i|0)==(0);
              do {
               if (!($115)) {
                if ($104) {
                 $i$010$us$i = 0;
                 while(1) {
                  $116 = (($112) + (($i$010$us$i*12)|0)|0);
                  $117 = (_FT_Stream_ReadFields($stream,171888,$116)|0);
                  HEAP32[$error$i5>>2] = $117;
                  $118 = ($117|0)==(0);
                  if (!($118)) {
                   $props$0$i = $112;$strings$0$i = 0;
                   break L47;
                  }
                  $119 = (($i$010$us$i) + 1)|0;
                  $120 = ($119>>>0)<($nprops$0$i>>>0);
                  if ($120) {
                   $i$010$us$i = $119;
                  } else {
                   break;
                  }
                 }
                } else {
                 $i$010$i = 0;
                 while(1) {
                  $121 = (($112) + (($i$010$i*12)|0)|0);
                  $122 = (_FT_Stream_ReadFields($stream,171908,$121)|0);
                  HEAP32[$error$i5>>2] = $122;
                  $123 = ($122|0)==(0);
                  if (!($123)) {
                   $props$0$i = $112;$strings$0$i = 0;
                   break L47;
                  }
                  $124 = (($i$010$i) + 1)|0;
                  $125 = ($124>>>0)<($nprops$0$i>>>0);
                  if ($125) {
                   $i$010$i = $124;
                  } else {
                   break;
                  }
                 }
                }
                $126 = $nprops$0$i & 3;
                $127 = ($126|0)==(0);
                if (!($127)) {
                 $128 = (4 - ($126))|0;
                 $129 = (_FT_Stream_Skip($stream,$128)|0);
                 HEAP32[$error$i5>>2] = $129;
                 $130 = ($129|0)==(0);
                 if ($130) {
                  break;
                 }
                 HEAP32[$error$i5>>2] = 83;
                 $props$0$i = $112;$strings$0$i = 0;
                 break L47;
                }
               }
              } while(0);
              if ($104) {
               $132 = (_FT_Stream_ReadULongLE($stream,$error$i5)|0);
               $string_size$0$i = $132;
              } else {
               $131 = (_FT_Stream_ReadULong($stream,$error$i5)|0);
               $string_size$0$i = $131;
              }
              $133 = HEAP32[$error$i5>>2]|0;
              $134 = ($133|0)==(0);
              if ($134) {
               $135 = Math_imul($nprops$0$i, -9)|0;
               $136 = (($135) + ($98))|0;
               $137 = ($string_size$0$i>>>0)>($136>>>0);
               if ($137) {
                HEAP32[$error$i5>>2] = 8;
                $props$0$i = $112;$strings$0$i = 0;
                break;
               }
               $138 = (($string_size$0$i) + 1)|0;
               $139 = (_ft_mem_realloc($80,1,0,$138,0,$error$i5)|0);
               $140 = HEAP32[$error$i5>>2]|0;
               $141 = ($140|0)==(0);
               if (!($141)) {
                $props$0$i = $112;$strings$0$i = $139;
                break;
               }
               $142 = (_FT_Stream_Read($stream,$139,$string_size$0$i)|0);
               HEAP32[$error$i5>>2] = $142;
               $143 = ($142|0)==(0);
               if (!($143)) {
                $props$0$i = $112;$strings$0$i = $139;
                break;
               }
               $144 = (_ft_mem_realloc($80,12,0,$nprops$0$i,0,$error$i5)|0);
               $145 = HEAP32[$error$i5>>2]|0;
               $146 = ($145|0)==(0);
               if (!($146)) {
                $props$0$i = $112;$strings$0$i = $139;
                break;
               }
               $147 = ((($face)) + 284|0);
               HEAP32[$147>>2] = $144;
               L88: do {
                if (!($115)) {
                 $i$19$i = 0;
                 while(1) {
                  $148 = (($112) + (($i$19$i*12)|0)|0);
                  $149 = HEAP32[$148>>2]|0;
                  $150 = ($149|0)<(0);
                  $151 = ($149>>>0)>($string_size$0$i>>>0);
                  $or$cond2$i = $150 | $151;
                  if ($or$cond2$i) {
                   label = 69;
                   break;
                  }
                  $152 = (($139) + ($149)|0);
                  $153 = (_ft_mem_strdup($80,$152,$error$i5)|0);
                  $154 = (($144) + (($i$19$i*12)|0)|0);
                  HEAP32[$154>>2] = $153;
                  $155 = HEAP32[$error$i5>>2]|0;
                  $156 = ($155|0)==(0);
                  if (!($156)) {
                   $props$0$i = $112;$strings$0$i = $139;
                   break L47;
                  }
                  $157 = (((($112) + (($i$19$i*12)|0)|0)) + 4|0);
                  $158 = HEAP8[$157>>0]|0;
                  $159 = (((($144) + (($i$19$i*12)|0)|0)) + 4|0);
                  HEAP8[$159>>0] = $158;
                  $160 = ($158<<24>>24)==(0);
                  $161 = (((($112) + (($i$19$i*12)|0)|0)) + 8|0);
                  $162 = HEAP32[$161>>2]|0;
                  if ($160) {
                   $170 = (((($144) + (($i$19$i*12)|0)|0)) + 8|0);
                   HEAP32[$170>>2] = $162;
                  } else {
                   $163 = ($162|0)<(0);
                   $164 = ($162>>>0)>($string_size$0$i>>>0);
                   $or$cond3$i = $163 | $164;
                   if ($or$cond3$i) {
                    label = 73;
                    break;
                   }
                   $165 = (($139) + ($162)|0);
                   $166 = (_ft_mem_strdup($80,$165,$error$i5)|0);
                   $167 = (((($144) + (($i$19$i*12)|0)|0)) + 8|0);
                   HEAP32[$167>>2] = $166;
                   $168 = HEAP32[$error$i5>>2]|0;
                   $169 = ($168|0)==(0);
                   if (!($169)) {
                    $props$0$i = $112;$strings$0$i = $139;
                    break L47;
                   }
                  }
                  $171 = (($i$19$i) + 1)|0;
                  $172 = ($171>>>0)<($nprops$0$i>>>0);
                  if ($172) {
                   $i$19$i = $171;
                  } else {
                   break L88;
                  }
                 }
                 if ((label|0) == 69) {
                  HEAP32[$error$i5>>2] = 9;
                  $props$0$i = $112;$strings$0$i = $139;
                  break L47;
                 }
                 else if ((label|0) == 73) {
                  HEAP32[$error$i5>>2] = 9;
                  $props$0$i = $112;$strings$0$i = $139;
                  break L47;
                 }
                }
               } while(0);
               HEAP32[$error$i5>>2] = 0;
               $props$0$i = $112;$strings$0$i = $139;
              } else {
               $props$0$i = $112;$strings$0$i = 0;
              }
             } else {
              $props$0$i = $112;$strings$0$i = 0;
             }
            } else {
             $props$0$i = 0;$strings$0$i = 0;
            }
           } else {
            $props$0$i = 0;$strings$0$i = 0;
           }
          } else {
           $$0$i$ph$i = 83;
           label = 40;
          }
         }
        }
       } while(0);
       if ((label|0) == 40) {
        HEAP32[$error$i5>>2] = $$0$i$ph$i;
        $props$0$i = 0;$strings$0$i = 0;
       }
       _ft_mem_free($80,$props$0$i);
       _ft_mem_free($80,$strings$0$i);
       $173 = HEAP32[$error$i5>>2]|0;
       HEAP32[$error>>2] = $173;
       $174 = ($173|0)==(0);
       if ($174) {
        $175 = HEAP32[$14>>2]|0;
        $176 = HEAP32[$9>>2]|0;
        $177 = ($176|0)==(0);
        L106: do {
         if ($177) {
          label = 82;
         } else {
          $i$01$i = 0;
          while(1) {
           $180 = (($175) + ($i$01$i<<4)|0);
           $181 = HEAP32[$180>>2]|0;
           $182 = ($181|0)==(256);
           $179 = (($i$01$i) + 1)|0;
           if ($182) {
            $187 = $176;$189 = $175;$746 = 1;
            break L106;
           }
           $178 = ($179>>>0)<($176>>>0);
           if ($178) {
            $i$01$i = $179;
           } else {
            label = 82;
            break;
           }
          }
         }
        } while(0);
        if ((label|0) == 82) {
         $183 = (_pcf_get_accel($stream,$face,2)|0);
         HEAP32[$error>>2] = $183;
         $184 = ($183|0)==(0);
         if (!($184)) {
          break;
         }
         $$pre217 = HEAP32[$14>>2]|0;
         $$pre218 = HEAP32[$9>>2]|0;
         $187 = $$pre218;$189 = $$pre217;$746 = 0;
        }
        $185 = HEAP32[$0>>2]|0;
        $186 = ($187|0)==(0);
        L114: do {
         if ($186) {
          $$0$i$ph$i15 = 3;
          label = 89;
         } else {
          $i$02$i$i13 = 0;
          while(1) {
           $188 = (($189) + ($i$02$i$i13<<4)|0);
           $190 = HEAP32[$188>>2]|0;
           $191 = ($190|0)==(4);
           if ($191) {
            $i$02$i$i13$lcssa = $i$02$i$i13;
            break;
           }
           $200 = (($i$02$i$i13) + 1)|0;
           $201 = ($200>>>0)<($187>>>0);
           if ($201) {
            $i$02$i$i13 = $200;
           } else {
            $$0$i$ph$i15 = 3;
            label = 89;
            break L114;
           }
          }
          $192 = ((($stream)) + 8|0);
          $193 = HEAP32[$192>>2]|0;
          $194 = (((($189) + ($i$02$i$i13$lcssa<<4)|0)) + 12|0);
          $195 = HEAP32[$194>>2]|0;
          $196 = ($195>>>0)<($193>>>0);
          if ($196) {
           $$0$i$ph$i15 = 83;
           label = 89;
          } else {
           $197 = (($195) - ($193))|0;
           $198 = (_FT_Stream_Skip($stream,$197)|0);
           $199 = ($198|0)==(0);
           if ($199) {
            $202 = (((($189) + ($i$02$i$i13$lcssa<<4)|0)) + 8|0);
            $203 = HEAP32[$202>>2]|0;
            HEAP32[$error$i12>>2] = 0;
            $204 = (_FT_Stream_ReadULongLE($stream,$error$i12)|0);
            $205 = HEAP32[$error$i12>>2]|0;
            $206 = ($205|0)==(0);
            if ($206) {
             $207 = $204 & -256;
             switch ($207|0) {
             case 256: case 0:  {
              break;
             }
             default: {
              $$0$i20$ph = 3;
              break L114;
             }
             }
             $208 = ($204>>>0)<(256);
             $209 = $204 & 4;
             $210 = ($209|0)!=(0);
             do {
              if ($208) {
               if ($210) {
                $211 = (_FT_Stream_ReadULong($stream,$error$i12)|0);
                $nmetrics$0$i = $211;
                break;
               } else {
                $212 = (_FT_Stream_ReadULongLE($stream,$error$i12)|0);
                $nmetrics$0$i = $212;
                break;
               }
              } else {
               if ($210) {
                $213 = (_FT_Stream_ReadUShort($stream,$error$i12)|0);
                $214 = $213&65535;
                $nmetrics$0$i = $214;
                break;
               } else {
                $215 = (_FT_Stream_ReadUShortLE($stream,$error$i12)|0);
                $216 = $215&65535;
                $nmetrics$0$i = $216;
                break;
               }
              }
             } while(0);
             $217 = HEAP32[$error$i12>>2]|0;
             $218 = ($217|0)==(0);
             if ($218) {
              $219 = ((($face)) + 288|0);
              HEAP32[$219>>2] = $nmetrics$0$i;
              $220 = ($nmetrics$0$i|0)==(0);
              if ($220) {
               $$0$i20$ph = 8;
              } else {
               if ($208) {
                $221 = (($203>>>0) / 12)&-1;
                $222 = ($nmetrics$0$i>>>0)>($221>>>0);
                if ($222) {
                 $$0$i20$ph = 8;
                 break;
                }
               } else {
                $223 = (($203>>>0) / 5)&-1;
                $224 = ($nmetrics$0$i>>>0)>($223>>>0);
                if ($224) {
                 $$0$i20$ph = 8;
                 break;
                }
               }
               $225 = (_ft_mem_realloc($185,16,0,$nmetrics$0$i,0,$error$i12)|0);
               $226 = ((($face)) + 292|0);
               HEAP32[$226>>2] = $225;
               $227 = HEAP32[$error$i12>>2]|0;
               $228 = ($227|0)==(0);
               if ($228) {
                $229 = $210 ? 172032 : 172064;
                $230 = ((($compr$i$i)) + 1|0);
                $231 = ((($compr$i$i)) + 2|0);
                $232 = ((($compr$i$i)) + 3|0);
                $233 = ((($compr$i$i)) + 4|0);
                L140: do {
                 if ($208) {
                  $i$03$us$i = 0;$metrics$02$us$i = $225;
                  while(1) {
                   $234 = (_FT_Stream_ReadFields($stream,$229,$metrics$02$us$i)|0);
                   HEAP32[$error$i12>>2] = $234;
                   $235 = ((($metrics$02$us$i)) + 12|0);
                   HEAP32[$235>>2] = 0;
                   $236 = ($234|0)==(0);
                   if (!($236)) {
                    label = 120;
                    break L140;
                   }
                   $237 = ((($metrics$02$us$i)) + 2|0);
                   $238 = HEAP16[$237>>1]|0;
                   $239 = HEAP16[$metrics$02$us$i>>1]|0;
                   $240 = ($238<<16>>16)<($239<<16>>16);
                   if ($240) {
                    $$pre11$i = ((($metrics$02$us$i)) + 8|0);
                    $$pre$phi12$iZ2D = $$pre11$i;
                    label = 110;
                   } else {
                    $$pre10$i = ((($metrics$02$us$i)) + 6|0);
                    $241 = HEAP16[$$pre10$i>>1]|0;
                    $242 = $241 << 16 >> 16;
                    $243 = ((($metrics$02$us$i)) + 8|0);
                    $244 = HEAP16[$243>>1]|0;
                    $245 = $244 << 16 >> 16;
                    $246 = (($245) + ($242))|0;
                    $247 = ($246|0)<(0);
                    if ($247) {
                     $$pre$phi12$iZ2D = $243;
                     label = 110;
                    }
                   }
                   if ((label|0) == 110) {
                    label = 0;
                    $248 = $metrics$02$us$i;
                    $249 = $248;
                    HEAP32[$249>>2] = 0;
                    $250 = (($248) + 4)|0;
                    $251 = $250;
                    HEAP32[$251>>2] = 0;
                    HEAP16[$$pre$phi12$iZ2D>>1] = 0;
                   }
                   $252 = (($i$03$us$i) + 1)|0;
                   $253 = ((($metrics$02$us$i)) + 16|0);
                   $254 = ($252>>>0)<($nmetrics$0$i>>>0);
                   if ($254) {
                    $i$03$us$i = $252;$metrics$02$us$i = $253;
                   } else {
                    label = 119;
                    break;
                   }
                  }
                 } else {
                  $i$03$i = 0;$metrics$02$i = $225;
                  while(1) {
                   $255 = (_FT_Stream_ReadFields($stream,172096,$compr$i$i)|0);
                   $256 = ($255|0)==(0);
                   if (!($256)) {
                    $$lcssa373 = $255;$metrics$02$i$lcssa = $metrics$02$i;
                    break;
                   }
                   $257 = HEAP8[$compr$i$i>>0]|0;
                   $258 = $257&255;
                   $259 = (($258) + 65408)|0;
                   $260 = $259&65535;
                   HEAP16[$metrics$02$i>>1] = $260;
                   $261 = HEAP8[$230>>0]|0;
                   $262 = $261&255;
                   $263 = (($262) + 65408)|0;
                   $264 = $263&65535;
                   $265 = ((($metrics$02$i)) + 2|0);
                   HEAP16[$265>>1] = $264;
                   $266 = HEAP8[$231>>0]|0;
                   $267 = $266&255;
                   $268 = (($267) + 65408)|0;
                   $269 = $268&65535;
                   $270 = ((($metrics$02$i)) + 4|0);
                   HEAP16[$270>>1] = $269;
                   $271 = HEAP8[$232>>0]|0;
                   $272 = $271&255;
                   $273 = (($272) + 65408)|0;
                   $274 = $273&65535;
                   $275 = ((($metrics$02$i)) + 6|0);
                   HEAP16[$275>>1] = $274;
                   $276 = HEAP8[$233>>0]|0;
                   $277 = $276&255;
                   $278 = (($277) + 65408)|0;
                   $279 = $278&65535;
                   $280 = ((($metrics$02$i)) + 8|0);
                   HEAP16[$280>>1] = $279;
                   $281 = ((($metrics$02$i)) + 10|0);
                   HEAP16[$281>>1] = 0;
                   HEAP32[$error$i12>>2] = 0;
                   $282 = ((($metrics$02$i)) + 12|0);
                   HEAP32[$282>>2] = 0;
                   $283 = HEAP16[$265>>1]|0;
                   $284 = HEAP16[$metrics$02$i>>1]|0;
                   $285 = ($283<<16>>16)<($284<<16>>16);
                   if ($285) {
                    label = 115;
                   } else {
                    $286 = HEAP16[$275>>1]|0;
                    $287 = $286 << 16 >> 16;
                    $288 = HEAP16[$280>>1]|0;
                    $289 = $288 << 16 >> 16;
                    $290 = (($289) + ($287))|0;
                    $291 = ($290|0)<(0);
                    if ($291) {
                     label = 115;
                    }
                   }
                   if ((label|0) == 115) {
                    label = 0;
                    ;HEAP32[$metrics$02$i>>2]=0|0;HEAP32[$metrics$02$i+4>>2]=0|0;HEAP16[$metrics$02$i+8>>1]=0|0;
                   }
                   $292 = (($i$03$i) + 1)|0;
                   $293 = ((($metrics$02$i)) + 16|0);
                   $294 = ($292>>>0)<($nmetrics$0$i>>>0);
                   if ($294) {
                    $i$03$i = $292;$metrics$02$i = $293;
                   } else {
                    label = 119;
                    break L140;
                   }
                  }
                  HEAP32[$error$i12>>2] = $$lcssa373;
                  $295 = ((($metrics$02$i$lcssa)) + 12|0);
                  HEAP32[$295>>2] = 0;
                  label = 120;
                 }
                } while(0);
                if ((label|0) == 119) {
                 HEAP32[$error>>2] = 0;
                }
                else if ((label|0) == 120) {
                 $296 = HEAP32[$226>>2]|0;
                 _ft_mem_free($185,$296);
                 HEAP32[$226>>2] = 0;
                 $$pre$i18 = HEAP32[$error$i12>>2]|0;
                 HEAP32[$error>>2] = $$pre$i18;
                 $297 = ($$pre$i18|0)==(0);
                 if (!($297)) {
                  break L1;
                 }
                }
                $298 = HEAP32[$0>>2]|0;
                $299 = HEAP32[$14>>2]|0;
                $300 = HEAP32[$9>>2]|0;
                $301 = ($300|0)==(0);
                L163: do {
                 if ($301) {
                  $$0$i$ph$i25 = 3;
                  label = 126;
                 } else {
                  $i$02$i$i23 = 0;
                  while(1) {
                   $302 = (($299) + ($i$02$i$i23<<4)|0);
                   $303 = HEAP32[$302>>2]|0;
                   $304 = ($303|0)==(8);
                   if ($304) {
                    $i$02$i$i23$lcssa = $i$02$i$i23;
                    break;
                   }
                   $312 = (($i$02$i$i23) + 1)|0;
                   $313 = ($312>>>0)<($300>>>0);
                   if ($313) {
                    $i$02$i$i23 = $312;
                   } else {
                    $$0$i$ph$i25 = 3;
                    label = 126;
                    break L163;
                   }
                  }
                  $305 = HEAP32[$192>>2]|0;
                  $306 = (((($299) + ($i$02$i$i23$lcssa<<4)|0)) + 12|0);
                  $307 = HEAP32[$306>>2]|0;
                  $308 = ($307>>>0)<($305>>>0);
                  if ($308) {
                   $$0$i$ph$i25 = 83;
                   label = 126;
                   break;
                  }
                  $309 = (($307) - ($305))|0;
                  $310 = (_FT_Stream_Skip($stream,$309)|0);
                  $311 = ($310|0)==(0);
                  if (!($311)) {
                   $$0$i$ph$i25 = 83;
                   label = 126;
                   break;
                  }
                  $314 = (((($299) + ($i$02$i$i23$lcssa<<4)|0)) + 8|0);
                  $315 = HEAP32[$314>>2]|0;
                  $316 = (_FT_Stream_EnterFrame($stream,8)|0);
                  HEAP32[$error$i22>>2] = $316;
                  $317 = ($316|0)==(0);
                  if (!($317)) {
                   $$0$i33$ph = $316;
                   break;
                  }
                  $318 = (_FT_Stream_GetULongLE($stream)|0);
                  $319 = $318 & 4;
                  $320 = ($319|0)==(0);
                  if ($320) {
                   $322 = (_FT_Stream_GetULongLE($stream)|0);
                   $nbitmaps$0$i = $322;
                  } else {
                   $321 = (_FT_Stream_GetULong($stream)|0);
                   $nbitmaps$0$i = $321;
                  }
                  _FT_Stream_ExitFrame($stream);
                  $323 = ($318>>>0)<(256);
                  if (!($323)) {
                   $$0$i33$ph = 3;
                   break;
                  }
                  $324 = HEAP32[$219>>2]|0;
                  $325 = ($nbitmaps$0$i|0)==($324|0);
                  if (!($325)) {
                   $$0$i33$ph = 3;
                   break;
                  }
                  $326 = (_ft_mem_realloc($298,4,0,$nbitmaps$0$i,0,$error$i22)|0);
                  $327 = HEAP32[$error$i22>>2]|0;
                  $328 = ($327|0)==(0);
                  if (!($328)) {
                   $$0$i33$ph = $327;
                   break;
                  }
                  $329 = ($nbitmaps$0$i|0)==(0);
                  if (!($329)) {
                   if ($320) {
                    $i$07$us$i = 0;
                    while(1) {
                     $330 = (_FT_Stream_ReadULongLE($stream,$error$i22)|0);
                     $331 = (($326) + ($i$07$us$i<<2)|0);
                     HEAP32[$331>>2] = $330;
                     $332 = (($i$07$us$i) + 1)|0;
                     $exitcond14$i = ($332|0)==($nbitmaps$0$i|0);
                     if ($exitcond14$i) {
                      break;
                     } else {
                      $i$07$us$i = $332;
                     }
                    }
                   } else {
                    $i$07$i = 0;
                    while(1) {
                     $333 = (_FT_Stream_ReadULong($stream,$error$i22)|0);
                     $334 = (($326) + ($i$07$i<<2)|0);
                     HEAP32[$334>>2] = $333;
                     $335 = (($i$07$i) + 1)|0;
                     $exitcond15$i = ($335|0)==($nbitmaps$0$i|0);
                     if ($exitcond15$i) {
                      break;
                     } else {
                      $i$07$i = $335;
                     }
                    }
                   }
                  }
                  $336 = HEAP32[$error$i22>>2]|0;
                  $337 = ($336|0)==(0);
                  do {
                   if ($337) {
                    if ($320) {
                     (_FT_Stream_ReadULongLE($stream,$error$i22)|0);
                     $340 = HEAP32[$error$i22>>2]|0;
                     $341 = ($340|0)==(0);
                     if (!($341)) {
                      break;
                     }
                     (_FT_Stream_ReadULongLE($stream,$error$i22)|0);
                     $342 = HEAP32[$error$i22>>2]|0;
                     $343 = ($342|0)==(0);
                     if (!($343)) {
                      break;
                     }
                     (_FT_Stream_ReadULongLE($stream,$error$i22)|0);
                     $356 = HEAP32[$error$i22>>2]|0;
                     $357 = ($356|0)==(0);
                     if (!($357)) {
                      break;
                     }
                     (_FT_Stream_ReadULongLE($stream,$error$i22)|0);
                     $358 = HEAP32[$error$i22>>2]|0;
                     $359 = ($358|0)==(0);
                     if (!($359)) {
                      break;
                     }
                    } else {
                     (_FT_Stream_ReadULong($stream,$error$i22)|0);
                     $338 = HEAP32[$error$i22>>2]|0;
                     $339 = ($338|0)==(0);
                     if (!($339)) {
                      break;
                     }
                     (_FT_Stream_ReadULong($stream,$error$i22)|0);
                     $344 = HEAP32[$error$i22>>2]|0;
                     $345 = ($344|0)==(0);
                     if (!($345)) {
                      break;
                     }
                     (_FT_Stream_ReadULong($stream,$error$i22)|0);
                     $360 = HEAP32[$error$i22>>2]|0;
                     $361 = ($360|0)==(0);
                     if (!($361)) {
                      break;
                     }
                     (_FT_Stream_ReadULong($stream,$error$i22)|0);
                     $362 = HEAP32[$error$i22>>2]|0;
                     $363 = ($362|0)==(0);
                     if (!($363)) {
                      break;
                     }
                    }
                    if (!($329)) {
                     $i$25$i = 0;
                     while(1) {
                      $346 = (($326) + ($i$25$i<<2)|0);
                      $347 = HEAP32[$346>>2]|0;
                      $348 = ($347|0)<(0);
                      $349 = ($347>>>0)>($315>>>0);
                      $or$cond$i30 = $348 | $349;
                      if (!($or$cond$i30)) {
                       $350 = HEAP32[$192>>2]|0;
                       $351 = (($350) + ($347))|0;
                       $352 = HEAP32[$226>>2]|0;
                       $353 = (((($352) + ($i$25$i<<4)|0)) + 12|0);
                       HEAP32[$353>>2] = $351;
                      }
                      $354 = (($i$25$i) + 1)|0;
                      $exitcond$i = ($354|0)==($nbitmaps$0$i|0);
                      if ($exitcond$i) {
                       break;
                      } else {
                       $i$25$i = $354;
                      }
                     }
                    }
                    $355 = ((($face)) + 308|0);
                    HEAP32[$355>>2] = $318;
                   }
                  } while(0);
                  _ft_mem_free($298,$326);
                  $364 = HEAP32[$error$i22>>2]|0;
                  HEAP32[$error>>2] = $364;
                  $365 = ($364|0)==(0);
                  if (!($365)) {
                   break L1;
                  }
                  $366 = HEAP32[$0>>2]|0;
                  $367 = HEAP32[$14>>2]|0;
                  $368 = HEAP32[$9>>2]|0;
                  $369 = ($368|0)==(0);
                  L207: do {
                   if ($369) {
                    $$0$i$ph$i37 = 3;
                    label = 160;
                   } else {
                    $i$02$i$i35 = 0;
                    while(1) {
                     $370 = (($367) + ($i$02$i$i35<<4)|0);
                     $371 = HEAP32[$370>>2]|0;
                     $372 = ($371|0)==(32);
                     if ($372) {
                      $i$02$i$i35$lcssa = $i$02$i$i35;
                      break;
                     }
                     $380 = (($i$02$i$i35) + 1)|0;
                     $381 = ($380>>>0)<($368>>>0);
                     if ($381) {
                      $i$02$i$i35 = $380;
                     } else {
                      $$0$i$ph$i37 = 3;
                      label = 160;
                      break L207;
                     }
                    }
                    $373 = HEAP32[$192>>2]|0;
                    $374 = (((($367) + ($i$02$i$i35$lcssa<<4)|0)) + 12|0);
                    $375 = HEAP32[$374>>2]|0;
                    $376 = ($375>>>0)<($373>>>0);
                    if ($376) {
                     $$0$i$ph$i37 = 83;
                     label = 160;
                     break;
                    }
                    $377 = (($375) - ($373))|0;
                    $378 = (_FT_Stream_Skip($stream,$377)|0);
                    $379 = ($378|0)==(0);
                    if (!($379)) {
                     $$0$i$ph$i37 = 83;
                     label = 160;
                     break;
                    }
                    $382 = (_FT_Stream_EnterFrame($stream,14)|0);
                    HEAP32[$error$i34>>2] = $382;
                    $383 = ($382|0)==(0);
                    if (!($383)) {
                     $$0$i44$ph = $382;
                     break;
                    }
                    $384 = (_FT_Stream_GetULongLE($stream)|0);
                    $385 = $384 & 4;
                    $386 = ($385|0)==(0);
                    if ($386) {
                     $396 = (_FT_Stream_GetUShortLE($stream)|0);
                     $397 = $396 << 16 >> 16;
                     $398 = (_FT_Stream_GetUShortLE($stream)|0);
                     $399 = $398 << 16 >> 16;
                     $400 = (_FT_Stream_GetUShortLE($stream)|0);
                     $401 = $400 << 16 >> 16;
                     $402 = (_FT_Stream_GetUShortLE($stream)|0);
                     $403 = $402 << 16 >> 16;
                     $404 = (_FT_Stream_GetUShortLE($stream)|0);
                     $$sink$i = $404;$firstCol$0$i = $397;$firstRow$0$i = $401;$lastCol$0$i = $399;$lastRow$0$i = $403;
                    } else {
                     $387 = (_FT_Stream_GetUShort($stream)|0);
                     $388 = $387 << 16 >> 16;
                     $389 = (_FT_Stream_GetUShort($stream)|0);
                     $390 = $389 << 16 >> 16;
                     $391 = (_FT_Stream_GetUShort($stream)|0);
                     $392 = $391 << 16 >> 16;
                     $393 = (_FT_Stream_GetUShort($stream)|0);
                     $394 = $393 << 16 >> 16;
                     $395 = (_FT_Stream_GetUShort($stream)|0);
                     $$sink$i = $395;$firstCol$0$i = $388;$firstRow$0$i = $392;$lastCol$0$i = $390;$lastRow$0$i = $394;
                    }
                    $405 = ((($face)) + 304|0);
                    HEAP16[$405>>1] = $$sink$i;
                    _FT_Stream_ExitFrame($stream);
                    $406 = ($384>>>0)<(256);
                    if (!($406)) {
                     $$0$i44$ph = 3;
                     break;
                    }
                    $407 = ($firstCol$0$i|0)<(0);
                    if ($407) {
                     $$0$i44$ph = 8;
                     break;
                    }
                    $408 = ($lastCol$0$i|0)<($firstCol$0$i|0);
                    $409 = ($lastCol$0$i|0)>(255);
                    $or$cond$i39 = $408 | $409;
                    $410 = ($firstRow$0$i|0)<(0);
                    $or$cond3$i40 = $or$cond$i39 | $410;
                    if ($or$cond3$i40) {
                     $$0$i44$ph = 8;
                     break;
                    }
                    $411 = ($lastRow$0$i|0)<($firstRow$0$i|0);
                    $412 = ($lastRow$0$i|0)>(255);
                    $or$cond5$i = $411 | $412;
                    if ($or$cond5$i) {
                     $$0$i44$ph = 8;
                     break;
                    }
                    $413 = (1 - ($firstCol$0$i))|0;
                    $414 = (($413) + ($lastCol$0$i))|0;
                    $415 = (1 - ($firstRow$0$i))|0;
                    $416 = (($415) + ($lastRow$0$i))|0;
                    $417 = Math_imul($416, $414)|0;
                    $418 = (_ft_mem_realloc($366,8,0,$417,0,$error$i34)|0);
                    $419 = HEAP32[$error$i34>>2]|0;
                    $420 = ($419|0)==(0);
                    if (!($420)) {
                     $$0$i44$ph = 64;
                     break;
                    }
                    $421 = $417 << 1;
                    $422 = (_FT_Stream_EnterFrame($stream,$421)|0);
                    HEAP32[$error$i34>>2] = $422;
                    $423 = ($422|0)==(0);
                    do {
                     if ($423) {
                      if ($386) {
                       $i$010$us24$i = $firstRow$0$i;$k$012$us23$i = 0;
                       while(1) {
                        $424 = $i$010$us24$i << 8;
                        $j$08$us$us38$i = $firstCol$0$i;$k$19$us$us37$i = $k$012$us23$i;
                        while(1) {
                         $425 = (_FT_Stream_GetUShortLE($stream)|0);
                         $426 = ($425<<16>>16)>(-1);
                         if ($426) {
                          $427 = (($j$08$us$us38$i) + ($424))|0;
                          $428 = (($418) + ($k$19$us$us37$i<<3)|0);
                          HEAP32[$428>>2] = $427;
                          $429 = (((($418) + ($k$19$us$us37$i<<3)|0)) + 4|0);
                          HEAP16[$429>>1] = $425;
                          $430 = (($k$19$us$us37$i) + 1)|0;
                          $k$2$us$us40$i = $430;
                         } else {
                          $k$2$us$us40$i = $k$19$us$us37$i;
                         }
                         $431 = (($j$08$us$us38$i) + 1)|0;
                         $432 = ($j$08$us$us38$i|0)<($lastCol$0$i|0);
                         if ($432) {
                          $j$08$us$us38$i = $431;$k$19$us$us37$i = $k$2$us$us40$i;
                         } else {
                          $k$2$us$us40$i$lcssa = $k$2$us$us40$i;
                          break;
                         }
                        }
                        $433 = (($i$010$us24$i) + 1)|0;
                        $434 = ($i$010$us24$i|0)<($lastRow$0$i|0);
                        if ($434) {
                         $i$010$us24$i = $433;$k$012$us23$i = $k$2$us$us40$i$lcssa;
                        } else {
                         $k$0$lcssa$i = $k$2$us$us40$i$lcssa;
                         break;
                        }
                       }
                      } else {
                       $i$010$i41 = $firstRow$0$i;$k$012$i = 0;
                       while(1) {
                        $435 = $i$010$i41 << 8;
                        $j$08$i = $firstCol$0$i;$k$19$i = $k$012$i;
                        while(1) {
                         $436 = (_FT_Stream_GetUShort($stream)|0);
                         $437 = ($436<<16>>16)>(-1);
                         if ($437) {
                          $438 = (($j$08$i) + ($435))|0;
                          $439 = (($418) + ($k$19$i<<3)|0);
                          HEAP32[$439>>2] = $438;
                          $440 = (((($418) + ($k$19$i<<3)|0)) + 4|0);
                          HEAP16[$440>>1] = $436;
                          $441 = (($k$19$i) + 1)|0;
                          $k$2$i = $441;
                         } else {
                          $k$2$i = $k$19$i;
                         }
                         $442 = (($j$08$i) + 1)|0;
                         $443 = ($j$08$i|0)<($lastCol$0$i|0);
                         if ($443) {
                          $j$08$i = $442;$k$19$i = $k$2$i;
                         } else {
                          $k$2$i$lcssa = $k$2$i;
                          break;
                         }
                        }
                        $444 = (($i$010$i41) + 1)|0;
                        $445 = ($i$010$i41|0)<($lastRow$0$i|0);
                        if ($445) {
                         $i$010$i41 = $444;$k$012$i = $k$2$i$lcssa;
                        } else {
                         $k$0$lcssa$i = $k$2$i$lcssa;
                         break;
                        }
                       }
                      }
                      _FT_Stream_ExitFrame($stream);
                      $446 = (_ft_mem_realloc($366,8,$417,$k$0$lcssa$i,$418,$error$i34)|0);
                      $447 = HEAP32[$error$i34>>2]|0;
                      $448 = ($447|0)==(0);
                      if (!($448)) {
                       $encoding$0$i = $446;
                       label = 185;
                       break;
                      }
                      $449 = ((($face)) + 296|0);
                      HEAP32[$449>>2] = $k$0$lcssa$i;
                      $450 = ((($face)) + 300|0);
                      HEAP32[$450>>2] = $446;
                      HEAP32[$error>>2] = 0;
                     } else {
                      $encoding$0$i = $418;
                      label = 185;
                     }
                    } while(0);
                    if ((label|0) == 185) {
                     _ft_mem_free($366,$encoding$0$i);
                     $451 = HEAP32[$error$i34>>2]|0;
                     HEAP32[$error>>2] = $451;
                     $452 = ($451|0)==(0);
                     if (!($452)) {
                      break L1;
                     }
                    }
                    if ($746) {
                     $453 = (_pcf_get_accel($stream,$face,256)|0);
                     HEAP32[$error>>2] = $453;
                     $454 = ($453|0)==(0);
                     if (!($454)) {
                      break L1;
                     }
                    }
                    HEAP32[$face>>2] = 1;
                    $455 = ((($face)) + 4|0);
                    HEAP32[$455>>2] = 0;
                    $456 = ((($face)) + 8|0);
                    $457 = HEAP32[$456>>2]|0;
                    $458 = $457 | 146;
                    HEAP32[$456>>2] = $458;
                    $459 = ((($face)) + 199|0);
                    $460 = HEAP8[$459>>0]|0;
                    $461 = ($460<<24>>24)==(0);
                    if (!($461)) {
                     $462 = $457 | 150;
                     HEAP32[$456>>2] = $462;
                    }
                    HEAP32[$error$i45>>2] = 0;
                    $463 = HEAP32[$0>>2]|0;
                    $464 = ((($face)) + 12|0);
                    HEAP32[$464>>2] = 0;
                    $465 = ((($face)) + 284|0);
                    $466 = HEAP32[$465>>2]|0;
                    $467 = ((($face)) + 280|0);
                    $468 = HEAP32[$467>>2]|0;
                    $469 = ($468|0)>(0);
                    if ($469) {
                     $i$02$i$i46 = 0;
                     while(1) {
                      $470 = (($466) + (($i$02$i$i46*12)|0)|0);
                      $471 = HEAP32[$470>>2]|0;
                      $472 = (_strcmp($471,532415)|0);
                      $473 = (($i$02$i$i46) + 1)|0;
                      $474 = ($473|0)<($468|0);
                      $475 = ($472|0)!=(0);
                      $or$cond$i$i = $474 & $475;
                      if ($or$cond$i$i) {
                       $i$02$i$i46 = $473;
                      } else {
                       break;
                      }
                     }
                     $476 = (($466) + (($473*12)|0)|0);
                     $477 = ((($476)) + -12|0);
                     L258: do {
                      if ($475) {
                       $502 = 2;$strings$i$sroa$8$0 = 0;
                      } else {
                       $478 = ((($477)) + 4|0);
                       $479 = HEAP8[$478>>0]|0;
                       $480 = ($479<<24>>24)==(0);
                       if ($480) {
                        $502 = 2;$strings$i$sroa$8$0 = 0;
                        break;
                       }
                       $481 = ((($477)) + 8|0);
                       $482 = HEAP32[$481>>2]|0;
                       $483 = HEAP8[$482>>0]|0;
                       switch ($483<<24>>24) {
                       case 105: case 73: case 111: case 79:  {
                        break;
                       }
                       default: {
                        $502 = 2;$strings$i$sroa$8$0 = 0;
                        break L258;
                       }
                       }
                       HEAP32[$464>>2] = 1;
                       $484 = HEAP8[$482>>0]|0;
                       $485 = $484 | 32;
                       $486 = ($485<<24>>24)==(111);
                       $487 = $486 ? 532421 : 594736;
                       $502 = 3;$strings$i$sroa$8$0 = $487;
                      }
                     } while(0);
                     $i$02$i5$i = 0;
                     while(1) {
                      $488 = (($466) + (($i$02$i5$i*12)|0)|0);
                      $489 = HEAP32[$488>>2]|0;
                      $490 = (_strcmp($489,532429)|0);
                      $491 = (($i$02$i5$i) + 1)|0;
                      $492 = ($491|0)<($468|0);
                      $493 = ($490|0)!=(0);
                      $or$cond$i6$i = $492 & $493;
                      if ($or$cond$i6$i) {
                       $i$02$i5$i = $491;
                      } else {
                       break;
                      }
                     }
                     $494 = (($466) + (($491*12)|0)|0);
                     $495 = ((($494)) + -12|0);
                     L266: do {
                      if ($493) {
                       $535 = 0;
                      } else {
                       $496 = ((($495)) + 4|0);
                       $497 = HEAP8[$496>>0]|0;
                       $498 = ($497<<24>>24)==(0);
                       if ($498) {
                        $535 = 0;
                        break;
                       }
                       $499 = ((($495)) + 8|0);
                       $500 = HEAP32[$499>>2]|0;
                       $501 = HEAP8[$500>>0]|0;
                       switch ($501<<24>>24) {
                       case 98: case 66:  {
                        break;
                       }
                       default: {
                        $535 = 0;
                        break L266;
                       }
                       }
                       HEAP32[$464>>2] = $502;
                       $535 = 594731;
                      }
                     } while(0);
                     $i$02$i11$i = 0;
                     while(1) {
                      $503 = (($466) + (($i$02$i11$i*12)|0)|0);
                      $504 = HEAP32[$503>>2]|0;
                      $505 = (_strcmp($504,532441)|0);
                      $506 = (($i$02$i11$i) + 1)|0;
                      $507 = ($506|0)<($468|0);
                      $508 = ($505|0)!=(0);
                      $or$cond$i12$i = $507 & $508;
                      if ($or$cond$i12$i) {
                       $i$02$i11$i = $506;
                      } else {
                       break;
                      }
                     }
                     $509 = (($466) + (($506*12)|0)|0);
                     $510 = ((($509)) + -12|0);
                     L274: do {
                      if ($508) {
                       $strings$i$sroa$10$0 = 0;
                      } else {
                       $511 = ((($510)) + 4|0);
                       $512 = HEAP8[$511>>0]|0;
                       $513 = ($512<<24>>24)==(0);
                       if ($513) {
                        $strings$i$sroa$10$0 = 0;
                        break;
                       }
                       $514 = ((($510)) + 8|0);
                       $515 = HEAP32[$514>>2]|0;
                       $516 = HEAP8[$515>>0]|0;
                       switch ($516<<24>>24) {
                       case 110: case 78: case 0:  {
                        $strings$i$sroa$10$0 = 0;
                        break L274;
                        break;
                       }
                       default: {
                       }
                       }
                       $strings$i$sroa$10$0 = $515;
                      }
                     } while(0);
                     $i$02$i17$i = 0;
                     while(1) {
                      $517 = (($466) + (($i$02$i17$i*12)|0)|0);
                      $518 = HEAP32[$517>>2]|0;
                      $519 = (_strcmp($518,532455)|0);
                      $520 = (($i$02$i17$i) + 1)|0;
                      $521 = ($520|0)<($468|0);
                      $522 = ($519|0)!=(0);
                      $or$cond$i18$i = $521 & $522;
                      if ($or$cond$i18$i) {
                       $i$02$i17$i = $520;
                      } else {
                       break;
                      }
                     }
                     $523 = (($466) + (($520*12)|0)|0);
                     $524 = ((($523)) + -12|0);
                     L282: do {
                      if ($522) {
                       $751 = 0;$752 = 0;$len$1$i = 0;
                      } else {
                       $525 = ((($524)) + 4|0);
                       $526 = HEAP8[$525>>0]|0;
                       $527 = ($526<<24>>24)==(0);
                       if ($527) {
                        $751 = 0;$752 = 0;$len$1$i = 0;
                        break;
                       }
                       $528 = ((($524)) + 8|0);
                       $529 = HEAP32[$528>>2]|0;
                       $530 = HEAP8[$529>>0]|0;
                       switch ($530<<24>>24) {
                       case 110: case 78: case 0:  {
                        $751 = 0;$752 = 0;$len$1$i = 0;
                        break L282;
                        break;
                       }
                       default: {
                       }
                       }
                       $531 = ($529|0)==(0|0);
                       if ($531) {
                        $751 = 0;$752 = 0;$len$1$i = 0;
                        break;
                       }
                       $532 = (_strlen($529)|0);
                       $533 = (($532) + 1)|0;
                       $751 = $529;$752 = $532;$len$1$i = $533;
                      }
                     } while(0);
                     $534 = ($535|0)==(0|0);
                     if ($534) {
                      $753 = 1;$754 = 0;$755 = 0;$len$1$1$i = $len$1$i;
                     } else {
                      $571 = (_strlen($535)|0);
                      $572 = (($len$1$i) + 1)|0;
                      $573 = (($572) + ($571))|0;
                      $753 = 0;$754 = $535;$755 = $571;$len$1$1$i = $573;
                     }
                     $574 = ($strings$i$sroa$8$0|0)==(0|0);
                     if ($574) {
                      $758 = 0;$len$1$2$i = $len$1$1$i;$lengths$i$sroa$10$0 = 0;
                     } else {
                      $575 = (_strlen($strings$i$sroa$8$0)|0);
                      $576 = (($len$1$1$i) + 1)|0;
                      $577 = (($576) + ($575))|0;
                      $758 = $strings$i$sroa$8$0;$len$1$2$i = $577;$lengths$i$sroa$10$0 = $575;
                     }
                     $578 = ($strings$i$sroa$10$0|0)==(0|0);
                     if ($578) {
                      $len$1$3$i = $len$1$2$i;$lengths$i$sroa$13$0 = 0;
                     } else {
                      $579 = (_strlen($strings$i$sroa$10$0)|0);
                      $580 = (($len$1$2$i) + 1)|0;
                      $581 = (($580) + ($579))|0;
                      $len$1$3$i = $581;$lengths$i$sroa$13$0 = $579;
                     }
                     $582 = ($len$1$3$i|0)==(0);
                     if ($582) {
                      $747 = $755;$748 = $754;$749 = $758;$750 = $753;$lengths$i$sroa$10$0108110 = $lengths$i$sroa$10$0;$lengths$i$sroa$13$0116 = $lengths$i$sroa$13$0;$strings$i$sroa$10$2102105106114 = $strings$i$sroa$10$0;$strings$i$sroa$8$2103104107112 = $strings$i$sroa$8$0;
                      label = 216;
                     } else {
                      $541 = $751;$542 = $752;$552 = $754;$553 = $755;$756 = $758;$757 = $753;$len$2$i = $len$1$3$i;$lengths$i$sroa$10$0108109 = $lengths$i$sroa$10$0;$lengths$i$sroa$13$0115 = $lengths$i$sroa$13$0;$strings$i$sroa$10$2102105106113 = $strings$i$sroa$10$0;$strings$i$sroa$8$2103104107111 = $strings$i$sroa$8$0;
                     }
                    } else {
                     $747 = 0;$748 = 0;$749 = 0;$750 = 1;$lengths$i$sroa$10$0108110 = 0;$lengths$i$sroa$13$0116 = 0;$strings$i$sroa$10$2102105106114 = 0;$strings$i$sroa$8$2103104107112 = 0;
                     label = 216;
                    }
                    if ((label|0) == 216) {
                     $541 = 594711;$542 = 7;$552 = $748;$553 = $747;$756 = $749;$757 = $750;$len$2$i = 8;$lengths$i$sroa$10$0108109 = $lengths$i$sroa$10$0108110;$lengths$i$sroa$13$0115 = $lengths$i$sroa$13$0116;$strings$i$sroa$10$2102105106113 = $strings$i$sroa$10$2102105106114;$strings$i$sroa$8$2103104107111 = $strings$i$sroa$8$2103104107112;
                    }
                    $536 = (_ft_mem_alloc($463,$len$2$i,$error$i45)|0);
                    $537 = ((($face)) + 24|0);
                    HEAP32[$537>>2] = $536;
                    $538 = HEAP32[$error$i45>>2]|0;
                    $539 = ($538|0)==(0);
                    if (!($539)) {
                     HEAP32[$error>>2] = $538;
                     break L1;
                    }
                    $540 = ($541|0)==(0|0);
                    if ($540) {
                     $s$2$i = $536;
                    } else {
                     _memcpy(($536|0),($541|0),($542|0))|0;
                     $543 = ($542|0)==(0);
                     if (!($543)) {
                      $mm$037$i = 0;
                      while(1) {
                       $544 = (($536) + ($mm$037$i)|0);
                       $545 = HEAP8[$544>>0]|0;
                       $546 = ($545<<24>>24)==(32);
                       if ($546) {
                        HEAP8[$544>>0] = 45;
                       }
                       $547 = (($mm$037$i) + 1)|0;
                       $exitcond$i50 = ($547|0)==($542|0);
                       if ($exitcond$i50) {
                        break;
                       } else {
                        $mm$037$i = $547;
                       }
                      }
                     }
                     $548 = (($536) + ($542)|0);
                     $s$2$i = $548;
                    }
                    if ($757) {
                     $556 = $756;$s$2$1$i = $s$2$i;
                    } else {
                     $549 = HEAP32[$537>>2]|0;
                     $550 = ($s$2$i|0)==($549|0);
                     if ($550) {
                      $s$1$1$i = $s$2$i;
                     } else {
                      $551 = ((($s$2$i)) + 1|0);
                      HEAP8[$s$2$i>>0] = 32;
                      $s$1$1$i = $551;
                     }
                     _memcpy(($s$1$1$i|0),($552|0),($553|0))|0;
                     $554 = (($s$1$1$i) + ($553)|0);
                     $556 = $strings$i$sroa$8$2103104107111;$s$2$1$i = $554;
                    }
                    $555 = ($556|0)==(0|0);
                    if ($555) {
                     $s$2$2$i = $s$2$1$i;
                    } else {
                     $557 = HEAP32[$537>>2]|0;
                     $558 = ($s$2$1$i|0)==($557|0);
                     if ($558) {
                      $s$1$2$i = $s$2$1$i;
                     } else {
                      $559 = ((($s$2$1$i)) + 1|0);
                      HEAP8[$s$2$1$i>>0] = 32;
                      $s$1$2$i = $559;
                     }
                     _memcpy(($s$1$2$i|0),($556|0),($lengths$i$sroa$10$0108109|0))|0;
                     $560 = (($s$1$2$i) + ($lengths$i$sroa$10$0108109)|0);
                     $s$2$2$i = $560;
                    }
                    $561 = ($strings$i$sroa$10$2102105106113|0)==(0|0);
                    if ($561) {
                     $s$2$3$i = $s$2$2$i;
                    } else {
                     $562 = HEAP32[$537>>2]|0;
                     $563 = ($s$2$2$i|0)==($562|0);
                     if ($563) {
                      $s$1$3$i = $s$2$2$i;
                     } else {
                      $564 = ((($s$2$2$i)) + 1|0);
                      HEAP8[$s$2$2$i>>0] = 32;
                      $s$1$3$i = $564;
                     }
                     _memcpy(($s$1$3$i|0),($strings$i$sroa$10$2102105106113|0),($lengths$i$sroa$13$0115|0))|0;
                     $565 = ($lengths$i$sroa$13$0115|0)==(0);
                     if ($565) {
                      $lengths$i$sroa$13$0117 = 0;
                     } else {
                      $mm$037$3$i = 0;
                      while(1) {
                       $566 = (($s$1$3$i) + ($mm$037$3$i)|0);
                       $567 = HEAP8[$566>>0]|0;
                       $568 = ($567<<24>>24)==(32);
                       if ($568) {
                        HEAP8[$566>>0] = 45;
                       }
                       $569 = (($mm$037$3$i) + 1)|0;
                       $exitcond$3$i = ($569|0)==($lengths$i$sroa$13$0115|0);
                       if ($exitcond$3$i) {
                        $lengths$i$sroa$13$0117 = $lengths$i$sroa$13$0115;
                        break;
                       } else {
                        $mm$037$3$i = $569;
                       }
                      }
                     }
                     $570 = (($s$1$3$i) + ($lengths$i$sroa$13$0117)|0);
                     $s$2$3$i = $570;
                    }
                    HEAP8[$s$2$3$i>>0] = 0;
                    $583 = HEAP32[$error$i45>>2]|0;
                    HEAP32[$error>>2] = $583;
                    $584 = ($583|0)==(0);
                    if (!($584)) {
                     break L1;
                    }
                    $585 = HEAP32[$465>>2]|0;
                    $586 = HEAP32[$467>>2]|0;
                    $587 = ($586|0)>(0);
                    do {
                     if ($587) {
                      $i$02$i = 0;
                      while(1) {
                       $588 = (($585) + (($i$02$i*12)|0)|0);
                       $589 = HEAP32[$588>>2]|0;
                       $590 = (_strcmp($589,532470)|0);
                       $591 = (($i$02$i) + 1)|0;
                       $592 = ($591|0)<($586|0);
                       $593 = ($590|0)!=(0);
                       $or$cond$i53 = $592 & $593;
                       if ($or$cond$i53) {
                        $i$02$i = $591;
                       } else {
                        break;
                       }
                      }
                      $594 = (($585) + (($591*12)|0)|0);
                      $595 = ((($594)) + -12|0);
                      if ($593) {
                       label = 253;
                       break;
                      }
                      $596 = ((($595)) + 4|0);
                      $597 = HEAP8[$596>>0]|0;
                      $598 = ($597<<24>>24)==(0);
                      if ($598) {
                       label = 253;
                       break;
                      }
                      $599 = ((($595)) + 8|0);
                      $600 = HEAP32[$599>>2]|0;
                      $601 = (_ft_mem_strdup($1,$600,$error)|0);
                      $602 = ((($face)) + 20|0);
                      HEAP32[$602>>2] = $601;
                      $603 = HEAP32[$error>>2]|0;
                      $604 = ($603|0)==(0);
                      if (!($604)) {
                       break L1;
                      }
                     } else {
                      label = 253;
                     }
                    } while(0);
                    if ((label|0) == 253) {
                     $605 = ((($face)) + 20|0);
                     HEAP32[$605>>2] = 0;
                    }
                    $606 = HEAP32[$219>>2]|0;
                    $607 = (($606) + 1)|0;
                    $608 = ((($face)) + 16|0);
                    HEAP32[$608>>2] = $607;
                    $609 = ((($face)) + 28|0);
                    HEAP32[$609>>2] = 1;
                    $610 = (_ft_mem_realloc($1,16,0,1,0,$error)|0);
                    $611 = ((($face)) + 32|0);
                    HEAP32[$611>>2] = $610;
                    $612 = HEAP32[$error>>2]|0;
                    $613 = ($612|0)==(0);
                    if (!($613)) {
                     break L1;
                    }
                    ;HEAP32[$610>>2]=0|0;HEAP32[$610+4>>2]=0|0;HEAP32[$610+8>>2]=0|0;HEAP32[$610+12>>2]=0|0;
                    $614 = ((($face)) + 204|0);
                    $615 = HEAP32[$614>>2]|0;
                    $616 = ((($face)) + 208|0);
                    $617 = HEAP32[$616>>2]|0;
                    $618 = (($617) + ($615))|0;
                    $619 = $618&65535;
                    HEAP16[$610>>1] = $619;
                    $620 = HEAP32[$465>>2]|0;
                    $621 = HEAP32[$467>>2]|0;
                    $622 = ($621|0)>(0);
                    do {
                     if ($622) {
                      $i$02$i55 = 0;
                      while(1) {
                       $623 = (($620) + (($i$02$i55*12)|0)|0);
                       $624 = HEAP32[$623>>2]|0;
                       $625 = (_strcmp($624,532482)|0);
                       $626 = (($i$02$i55) + 1)|0;
                       $627 = ($626|0)<($621|0);
                       $628 = ($625|0)!=(0);
                       $or$cond$i56 = $627 & $628;
                       if ($or$cond$i56) {
                        $i$02$i55 = $626;
                       } else {
                        break;
                       }
                      }
                      if ($628) {
                       label = 259;
                       break;
                      }
                      $629 = (($620) + (($626*12)|0)|0);
                      $630 = ((($629)) + -12|0);
                      $631 = ((($630)) + 8|0);
                      $632 = HEAP32[$631>>2]|0;
                      $633 = (($632) + 5)|0;
                      $634 = (($633|0) / 10)&-1;
                      $635 = $634&65535;
                      $636 = ((($610)) + 2|0);
                      HEAP16[$636>>1] = $635;
                     } else {
                      label = 259;
                     }
                    } while(0);
                    if ((label|0) == 259) {
                     $sext220 = $618 << 16;
                     $637 = $sext220 >> 15;
                     $638 = (($637|0) / 3)&-1;
                     $639 = $638&65535;
                     $640 = ((($610)) + 2|0);
                     HEAP16[$640>>1] = $639;
                    }
                    do {
                     if ($622) {
                      $i$02$i61 = 0;
                      while(1) {
                       $642 = (($620) + (($i$02$i61*12)|0)|0);
                       $643 = HEAP32[$642>>2]|0;
                       $644 = (_strcmp($643,532496)|0);
                       $645 = (($i$02$i61) + 1)|0;
                       $646 = ($645|0)<($621|0);
                       $647 = ($644|0)!=(0);
                       $or$cond$i62 = $646 & $647;
                       if ($or$cond$i62) {
                        $i$02$i61 = $645;
                       } else {
                        break;
                       }
                      }
                      if ($647) {
                       $692 = 0;
                      } else {
                       $648 = (($620) + (($645*12)|0)|0);
                       $649 = ((($648)) + -12|0);
                       $650 = ((($649)) + 8|0);
                       $651 = HEAP32[$650>>2]|0;
                       $652 = ($651*460800)|0;
                       $653 = (($652) + 36135)|0;
                       $654 = (($653|0) / 72270)&-1;
                       $655 = ((($610)) + 4|0);
                       HEAP32[$655>>2] = $654;
                       $692 = $654;
                      }
                      $i$02$i67 = 0;
                      while(1) {
                       $656 = (($620) + (($i$02$i67*12)|0)|0);
                       $657 = HEAP32[$656>>2]|0;
                       $658 = (_strcmp($657,532507)|0);
                       $659 = (($i$02$i67) + 1)|0;
                       $660 = ($659|0)<($621|0);
                       $661 = ($658|0)!=(0);
                       $or$cond$i68 = $660 & $661;
                       if ($or$cond$i68) {
                        $i$02$i67 = $659;
                       } else {
                        break;
                       }
                      }
                      if ($661) {
                       $691 = 0;
                      } else {
                       $662 = (($620) + (($659*12)|0)|0);
                       $663 = ((($662)) + -12|0);
                       $664 = ((($663)) + 8|0);
                       $665 = HEAP32[$664>>2]|0;
                       $sext = $665 << 16;
                       $666 = $sext >> 10;
                       $667 = ((($610)) + 12|0);
                       HEAP32[$667>>2] = $666;
                       $691 = $666;
                      }
                      $i$02$i73 = 0;
                      while(1) {
                       $668 = (($620) + (($i$02$i73*12)|0)|0);
                       $669 = HEAP32[$668>>2]|0;
                       $670 = (_strcmp($669,532518)|0);
                       $671 = (($i$02$i73) + 1)|0;
                       $672 = ($671|0)<($621|0);
                       $673 = ($670|0)!=(0);
                       $or$cond$i74 = $672 & $673;
                       if ($or$cond$i74) {
                        $i$02$i73 = $671;
                       } else {
                        break;
                       }
                      }
                      if ($673) {
                       $resolution_x$0$ph$ph = 0;
                      } else {
                       $674 = (($620) + (($671*12)|0)|0);
                       $675 = ((($674)) + -12|0);
                       $676 = ((($675)) + 8|0);
                       $677 = HEAP32[$676>>2]|0;
                       $phitmp = $677 << 16;
                       $phitmp3 = $phitmp >> 16;
                       $resolution_x$0$ph$ph = $phitmp3;
                      }
                      $i$02$i79 = 0;
                      while(1) {
                       $678 = (($620) + (($i$02$i79*12)|0)|0);
                       $679 = HEAP32[$678>>2]|0;
                       $680 = (_strcmp($679,532531)|0);
                       $681 = (($i$02$i79) + 1)|0;
                       $682 = ($681|0)<($621|0);
                       $683 = ($680|0)!=(0);
                       $or$cond$i80 = $682 & $683;
                       if ($or$cond$i80) {
                        $i$02$i79 = $681;
                       } else {
                        break;
                       }
                      }
                      if ($683) {
                       $resolution_y$0 = 0;
                      } else {
                       $684 = (($620) + (($681*12)|0)|0);
                       $685 = ((($684)) + -12|0);
                       $686 = ((($685)) + 8|0);
                       $687 = HEAP32[$686>>2]|0;
                       $688 = $687&65535;
                       $resolution_y$0 = $688;
                      }
                      $689 = ((($610)) + 12|0);
                      $690 = ($691|0)==(0);
                      if ($690) {
                       HEAP32[$689>>2] = $692;
                       $693 = ($resolution_y$0<<16>>16)==(0);
                       if ($693) {
                        $705 = $692;
                        label = 282;
                        break;
                       }
                       $694 = $resolution_y$0 << 16 >> 16;
                       $695 = Math_imul($692, $694)|0;
                       $696 = (($695|0) / 72)&-1;
                       HEAP32[$689>>2] = $696;
                       $701 = $696;
                      } else {
                       $701 = $691;
                      }
                      $697 = ($resolution_x$0$ph$ph|0)==(0);
                      $698 = ($resolution_y$0<<16>>16)==(0);
                      $or$cond4 = $697 | $698;
                      if ($or$cond4) {
                       $705 = $701;
                       label = 282;
                       break;
                      }
                      $699 = $resolution_y$0 << 16 >> 16;
                      $700 = Math_imul($701, $resolution_x$0$ph$ph)|0;
                      $702 = (($700|0) / ($699|0))&-1;
                      $703 = ((($610)) + 8|0);
                      HEAP32[$703>>2] = $702;
                     } else {
                      $641 = ((($610)) + 12|0);
                      HEAP32[$641>>2] = 0;
                      $705 = 0;
                      label = 282;
                     }
                    } while(0);
                    if ((label|0) == 282) {
                     $704 = ((($610)) + 8|0);
                     HEAP32[$704>>2] = $705;
                    }
                    if ($622) {
                     $i$02$i85 = 0;
                     while(1) {
                      $706 = (($620) + (($i$02$i85*12)|0)|0);
                      $707 = HEAP32[$706>>2]|0;
                      $708 = (_strcmp($707,593036)|0);
                      $709 = (($i$02$i85) + 1)|0;
                      $710 = ($709|0)<($621|0);
                      $711 = ($708|0)!=(0);
                      $or$cond$i86 = $710 & $711;
                      if ($or$cond$i86) {
                       $i$02$i85 = $709;
                      } else {
                       break;
                      }
                     }
                     $712 = (($620) + (($709*12)|0)|0);
                     $713 = ((($712)) + -12|0);
                     if ($711) {
                      label = 286;
                     } else {
                      $723 = $713;
                     }
                    } else {
                     label = 286;
                    }
                    if ((label|0) == 286) {
                     $723 = 0;
                    }
                    if ($622) {
                     $i$02$i91 = 0;
                     while(1) {
                      $714 = (($620) + (($i$02$i91*12)|0)|0);
                      $715 = HEAP32[$714>>2]|0;
                      $716 = (_strcmp($715,593053)|0);
                      $717 = (($i$02$i91) + 1)|0;
                      $718 = ($717|0)<($621|0);
                      $719 = ($716|0)!=(0);
                      $or$cond$i92 = $718 & $719;
                      if ($or$cond$i92) {
                       $i$02$i91 = $717;
                      } else {
                       break;
                      }
                     }
                     $720 = (($620) + (($717*12)|0)|0);
                     $721 = ((($720)) + -12|0);
                     if ($719) {
                      label = 290;
                     } else {
                      $728 = $721;
                     }
                    } else {
                     label = 290;
                    }
                    if ((label|0) == 290) {
                     $728 = 0;
                    }
                    $722 = ($723|0)==(0|0);
                    if ($722) {
                     $742 = 0;
                     STACKTOP = sp;return ($742|0);
                    }
                    $724 = ((($723)) + 4|0);
                    $725 = HEAP8[$724>>0]|0;
                    $726 = ($725<<24>>24)!=(0);
                    $727 = ($728|0)!=(0|0);
                    $or$cond = $727 & $726;
                    if (!($or$cond)) {
                     $742 = 0;
                     STACKTOP = sp;return ($742|0);
                    }
                    $729 = ((($728)) + 4|0);
                    $730 = HEAP8[$729>>0]|0;
                    $731 = ($730<<24>>24)==(0);
                    if ($731) {
                     $742 = 0;
                     STACKTOP = sp;return ($742|0);
                    }
                    $732 = ((($728)) + 8|0);
                    $733 = HEAP32[$732>>2]|0;
                    $734 = (_ft_mem_strdup($1,$733,$error)|0);
                    $735 = ((($face)) + 176|0);
                    HEAP32[$735>>2] = $734;
                    $736 = HEAP32[$error>>2]|0;
                    $737 = ($736|0)==(0);
                    if (!($737)) {
                     break L1;
                    }
                    $738 = ((($723)) + 8|0);
                    $739 = HEAP32[$738>>2]|0;
                    $740 = (_ft_mem_strdup($1,$739,$error)|0);
                    $741 = ((($face)) + 180|0);
                    HEAP32[$741>>2] = $740;
                    $$pr$pre = HEAP32[$error>>2]|0;
                    $phitmp221 = ($$pr$pre|0)==(0);
                    if ($phitmp221) {
                     $742 = 0;
                    } else {
                     break L1;
                    }
                    STACKTOP = sp;return ($742|0);
                   }
                  } while(0);
                  if ((label|0) == 160) {
                   HEAP32[$error$i34>>2] = $$0$i$ph$i37;
                   $$0$i44$ph = $$0$i$ph$i37;
                  }
                  HEAP32[$error>>2] = $$0$i44$ph;
                  break L1;
                 }
                } while(0);
                if ((label|0) == 126) {
                 HEAP32[$error$i22>>2] = $$0$i$ph$i25;
                 $$0$i33$ph = $$0$i$ph$i25;
                }
                HEAP32[$error>>2] = $$0$i33$ph;
                break L1;
               } else {
                $$0$i20$ph = 64;
               }
              }
             } else {
              $$0$i20$ph = 3;
             }
            } else {
             $$0$i20$ph = $205;
            }
           } else {
            $$0$i$ph$i15 = 83;
            label = 89;
           }
          }
         }
        } while(0);
        if ((label|0) == 89) {
         HEAP32[$error$i12>>2] = $$0$i$ph$i15;
         $$0$i20$ph = $$0$i$ph$i15;
        }
        HEAP32[$error>>2] = $$0$i20$ph;
       }
      } else {
       $$01$i$ph = 64;
       label = 31;
      }
     }
    } else {
     $$01$i$ph = 3;
     label = 31;
    }
   } else {
    $$01$i$ph = 1;
    label = 31;
   }
  } else {
   $$01$i$ph = 1;
   label = 31;
  }
 } while(0);
 if ((label|0) == 31) {
  HEAP32[$error>>2] = $$01$i$ph;
 }
 HEAP32[$error>>2] = 3;
 $742 = 3;
 STACKTOP = sp;return ($742|0);
}
function _pcf_get_accel($stream,$face,$type) {
 $stream = $stream|0;
 $face = $face|0;
 $type = $type|0;
 var $$0$i$ph = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $i$02$i = 0, $i$02$i$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($face)) + 192|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 188|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0);
 L1: do {
  if ($4) {
   $$0$i$ph = 3;
  } else {
   $i$02$i = 0;
   while(1) {
    $5 = (($1) + ($i$02$i<<4)|0);
    $6 = HEAP32[$5>>2]|0;
    $7 = ($6|0)==($type|0);
    if ($7) {
     $i$02$i$lcssa = $i$02$i;
     break;
    }
    $16 = (($i$02$i) + 1)|0;
    $17 = ($16>>>0)<($3>>>0);
    if ($17) {
     $i$02$i = $16;
    } else {
     $$0$i$ph = 3;
     break L1;
    }
   }
   $8 = ((($stream)) + 8|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (((($1) + ($i$02$i$lcssa<<4)|0)) + 12|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = ($11>>>0)<($9>>>0);
   if ($12) {
    $$0$i$ph = 83;
   } else {
    $13 = (($11) - ($9))|0;
    $14 = (_FT_Stream_Skip($stream,$13)|0);
    $15 = ($14|0)==(0);
    if ($15) {
     HEAP32[$error>>2] = 0;
     $18 = (_FT_Stream_ReadULongLE($stream,$error)|0);
     $19 = HEAP32[$error>>2]|0;
     $20 = ($19|0)==(0);
     if (!($20)) {
      $44 = $19;
      STACKTOP = sp;return ($44|0);
     }
     $21 = $18 & -256;
     switch ($21|0) {
     case 256: case 0:  {
      break;
     }
     default: {
      $44 = 0;
      STACKTOP = sp;return ($44|0);
     }
     }
     $22 = $18 & 4;
     $23 = ($22|0)==(0);
     $24 = ((($face)) + 196|0);
     if ($23) {
      $27 = (_FT_Stream_ReadFields($stream,171980,$24)|0);
      HEAP32[$error>>2] = $27;
      $28 = ($27|0)==(0);
      if (!($28)) {
       $44 = $27;
       STACKTOP = sp;return ($44|0);
      }
     } else {
      $25 = (_FT_Stream_ReadFields($stream,171928,$24)|0);
      HEAP32[$error>>2] = $25;
      $26 = ($25|0)==(0);
      if (!($26)) {
       $44 = $25;
       STACKTOP = sp;return ($44|0);
      }
     }
     $29 = ((($face)) + 216|0);
     $30 = ($22|0)!=(0);
     $31 = $30 ? 172032 : 172064;
     $32 = (_FT_Stream_ReadFields($stream,$31,$29)|0);
     HEAP32[$error>>2] = $32;
     $33 = ($32|0)==(0);
     if (!($33)) {
      $44 = $32;
      STACKTOP = sp;return ($44|0);
     }
     $34 = ((($face)) + 232|0);
     $35 = (_FT_Stream_ReadFields($stream,$31,$34)|0);
     HEAP32[$error>>2] = $35;
     $36 = ($35|0)==(0);
     if (!($36)) {
      $44 = $35;
      STACKTOP = sp;return ($44|0);
     }
     $37 = ($21|0)==(256);
     $38 = ((($face)) + 248|0);
     if (!($37)) {
      ;HEAP32[$38>>2]=HEAP32[$29>>2]|0;HEAP32[$38+4>>2]=HEAP32[$29+4>>2]|0;HEAP32[$38+8>>2]=HEAP32[$29+8>>2]|0;HEAP32[$38+12>>2]=HEAP32[$29+12>>2]|0;
      $43 = ((($face)) + 264|0);
      ;HEAP32[$43>>2]=HEAP32[$34>>2]|0;HEAP32[$43+4>>2]=HEAP32[$34+4>>2]|0;HEAP32[$43+8>>2]=HEAP32[$34+8>>2]|0;HEAP32[$43+12>>2]=HEAP32[$34+12>>2]|0;
      $44 = 0;
      STACKTOP = sp;return ($44|0);
     }
     $39 = (_FT_Stream_ReadFields($stream,$31,$38)|0);
     HEAP32[$error>>2] = $39;
     $40 = ($39|0)==(0);
     if (!($40)) {
      $44 = $39;
      STACKTOP = sp;return ($44|0);
     }
     $41 = ((($face)) + 264|0);
     $42 = (_FT_Stream_ReadFields($stream,$31,$41)|0);
     HEAP32[$error>>2] = $42;
     $44 = $42;
     STACKTOP = sp;return ($44|0);
    } else {
     $$0$i$ph = 83;
    }
   }
  }
 } while(0);
 HEAP32[$error>>2] = $$0$i$ph;
 $44 = $$0$i$ph;
 STACKTOP = sp;return ($44|0);
}
function _PCF_Face_Done($pcfface) {
 $pcfface = $pcfface|0;
 var $$pre = 0, $$pre2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($pcfface|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($pcfface)) + 100|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($pcfface)) + 300|0);
 $4 = HEAP32[$3>>2]|0;
 _ft_mem_free($2,$4);
 HEAP32[$3>>2] = 0;
 $5 = ((($pcfface)) + 292|0);
 $6 = HEAP32[$5>>2]|0;
 _ft_mem_free($2,$6);
 HEAP32[$5>>2] = 0;
 $7 = ((($pcfface)) + 284|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 if (!($9)) {
  $10 = ((($pcfface)) + 280|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($11|0)>(0);
  if ($12) {
   $14 = $8;$i$01 = 0;
   while(1) {
    $13 = (($14) + (($i$01*12)|0)|0);
    $15 = ($13|0)==(0|0);
    if (!($15)) {
     $16 = HEAP32[$13>>2]|0;
     _ft_mem_free($2,$16);
     HEAP32[$13>>2] = 0;
     $17 = (((($14) + (($i$01*12)|0)|0)) + 4|0);
     $18 = HEAP8[$17>>0]|0;
     $19 = ($18<<24>>24)==(0);
     if (!($19)) {
      $20 = (((($14) + (($i$01*12)|0)|0)) + 8|0);
      $21 = HEAP32[$20>>2]|0;
      _ft_mem_free($2,$21);
      HEAP32[$20>>2] = 0;
     }
    }
    $22 = (($i$01) + 1)|0;
    $23 = HEAP32[$10>>2]|0;
    $24 = ($22|0)<($23|0);
    if (!($24)) {
     break;
    }
    $$pre = HEAP32[$7>>2]|0;
    $14 = $$pre;$i$01 = $22;
   }
   $$pre2 = HEAP32[$7>>2]|0;
   $25 = $$pre2;
  } else {
   $25 = $8;
  }
  _ft_mem_free($2,$25);
  HEAP32[$7>>2] = 0;
 }
 $26 = ((($pcfface)) + 192|0);
 $27 = HEAP32[$26>>2]|0;
 _ft_mem_free($2,$27);
 HEAP32[$26>>2] = 0;
 $28 = ((($pcfface)) + 20|0);
 $29 = HEAP32[$28>>2]|0;
 _ft_mem_free($2,$29);
 HEAP32[$28>>2] = 0;
 $30 = ((($pcfface)) + 24|0);
 $31 = HEAP32[$30>>2]|0;
 _ft_mem_free($2,$31);
 HEAP32[$30>>2] = 0;
 $32 = ((($pcfface)) + 32|0);
 $33 = HEAP32[$32>>2]|0;
 _ft_mem_free($2,$33);
 HEAP32[$32>>2] = 0;
 $34 = ((($pcfface)) + 176|0);
 $35 = HEAP32[$34>>2]|0;
 _ft_mem_free($2,$35);
 HEAP32[$34>>2] = 0;
 $36 = ((($pcfface)) + 180|0);
 $37 = HEAP32[$36>>2]|0;
 _ft_mem_free($2,$37);
 HEAP32[$36>>2] = 0;
 $38 = ((($pcfface)) + 104|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ((($pcfface)) + 132|0);
 $41 = ($39|0)==($40|0);
 if (!($41)) {
  return;
 }
 _FT_Stream_Close($39);
 $42 = ((($pcfface)) + 172|0);
 $43 = HEAP32[$42>>2]|0;
 HEAP32[$38>>2] = $43;
 return;
}
function _pcf_cmap_init($pcfcmap,$init_data) {
 $pcfcmap = $pcfcmap|0;
 $init_data = $init_data|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$pcfcmap>>2]|0;
 $1 = ((($0)) + 296|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($pcfcmap)) + 16|0);
 HEAP32[$3>>2] = $2;
 $4 = ((($0)) + 300|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($pcfcmap)) + 20|0);
 HEAP32[$6>>2] = $5;
 return 0;
}
function _pcf_cmap_done($pcfcmap) {
 $pcfcmap = $pcfcmap|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pcfcmap)) + 20|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($pcfcmap)) + 16|0);
 HEAP32[$1>>2] = 0;
 return;
}
function _pcf_cmap_char_index($pcfcmap,$charcode) {
 $pcfcmap = $pcfcmap|0;
 $charcode = $charcode|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $max$01 = 0, $max$1 = 0;
 var $min$02 = 0, $min$1 = 0, $result$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pcfcmap)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($pcfcmap)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0);
 if ($4) {
  $result$2 = 0;
  return ($result$2|0);
 } else {
  $max$01 = $3;$min$02 = 0;
 }
 while(1) {
  $5 = (($min$02) + ($max$01))|0;
  $6 = $5 >>> 1;
  $7 = (($1) + ($6<<3)|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8|0)==($charcode|0);
  if ($9) {
   $$lcssa = $6;
   break;
  }
  $10 = ($8>>>0)>($charcode>>>0);
  $11 = (($6) + 1)|0;
  $max$1 = $10 ? $6 : $max$01;
  $min$1 = $10 ? $min$02 : $11;
  $12 = ($min$1>>>0)<($max$1>>>0);
  if ($12) {
   $max$01 = $max$1;$min$02 = $min$1;
  } else {
   $result$2 = 0;
   label = 5;
   break;
  }
 }
 if ((label|0) == 5) {
  return ($result$2|0);
 }
 $13 = (((($1) + ($$lcssa<<3)|0)) + 4|0);
 $14 = HEAP16[$13>>1]|0;
 $15 = $14&65535;
 $16 = (($15) + 1)|0;
 $result$2 = $16;
 return ($result$2|0);
}
function _pcf_cmap_char_next($pcfcmap,$acharcode) {
 $pcfcmap = $pcfcmap|0;
 $acharcode = $acharcode|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $charcode$0 = 0, $max$04 = 0, $max$1 = 0, $min$0$lcssa = 0, $min$05 = 0, $min$1 = 0, $result$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pcfcmap)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$acharcode>>2]|0;
 $3 = (($2) + 1)|0;
 $4 = ((($pcfcmap)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0);
 L1: do {
  if ($6) {
   $min$0$lcssa = 0;
  } else {
   $max$04 = $5;$min$05 = 0;
   while(1) {
    $7 = (($min$05) + ($max$04))|0;
    $8 = $7 >>> 1;
    $9 = (($1) + ($8<<3)|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = ($3|0)==($10|0);
    if ($11) {
     $$lcssa = $8;
     break;
    }
    $12 = ($3>>>0)<($10>>>0);
    $13 = (($8) + 1)|0;
    $max$1 = $12 ? $8 : $max$04;
    $min$1 = $12 ? $min$05 : $13;
    $14 = ($min$1>>>0)<($max$1>>>0);
    if ($14) {
     $max$04 = $max$1;$min$05 = $min$1;
    } else {
     $min$0$lcssa = $min$1;
     break L1;
    }
   }
   $15 = (((($1) + ($$lcssa<<3)|0)) + 4|0);
   $16 = HEAP16[$15>>1]|0;
   $17 = $16&65535;
   $18 = (($17) + 1)|0;
   $charcode$0 = $3;$result$2 = $18;
   HEAP32[$acharcode>>2] = $charcode$0;
   return ($result$2|0);
  }
 } while(0);
 $19 = ($min$0$lcssa>>>0)<($5>>>0);
 if (!($19)) {
  $charcode$0 = 0;$result$2 = 0;
  HEAP32[$acharcode>>2] = $charcode$0;
  return ($result$2|0);
 }
 $20 = (($1) + ($min$0$lcssa<<3)|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = (((($1) + ($min$0$lcssa<<3)|0)) + 4|0);
 $23 = HEAP16[$22>>1]|0;
 $24 = $23&65535;
 $25 = (($24) + 1)|0;
 $charcode$0 = $21;$result$2 = $25;
 HEAP32[$acharcode>>2] = $charcode$0;
 return ($result$2|0);
}
function _PCF_Glyph_Load($slot,$size,$glyph_index,$load_flags) {
 $slot = $slot|0;
 $size = $size|0;
 $glyph_index = $glyph_index|0;
 $load_flags = $load_flags|0;
 var $$0 = 0, $$012$i = 0, $$012$i5 = 0, $$012$i8 = 0, $$03$i = 0, $$03$i4 = 0, $$03$i7 = 0, $$lobit = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $glyph_index$ = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$size>>2]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  $$0 = 35;
  return ($$0|0);
 }
 $2 = ((($0)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3>>>0)>($glyph_index>>>0);
 if (!($4)) {
  $$0 = 6;
  return ($$0|0);
 }
 $5 = ((($0)) + 104|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($glyph_index|0)==(0);
 $8 = (($glyph_index) + -1)|0;
 $glyph_index$ = $7 ? 0 : $8;
 $9 = ((($0)) + 292|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (((($10) + ($glyph_index$<<4)|0)) + 6|0);
 $12 = HEAP16[$11>>1]|0;
 $13 = $12 << 16 >> 16;
 $14 = (((($10) + ($glyph_index$<<4)|0)) + 8|0);
 $15 = HEAP16[$14>>1]|0;
 $16 = $15 << 16 >> 16;
 $17 = (($16) + ($13))|0;
 $18 = ((($slot)) + 76|0);
 HEAP32[$18>>2] = $17;
 $19 = (((($10) + ($glyph_index$<<4)|0)) + 2|0);
 $20 = HEAP16[$19>>1]|0;
 $21 = $20 << 16 >> 16;
 $22 = (($10) + ($glyph_index$<<4)|0);
 $23 = HEAP16[$22>>1]|0;
 $24 = $23 << 16 >> 16;
 $25 = (($21) - ($24))|0;
 $26 = ((($slot)) + 80|0);
 HEAP32[$26>>2] = $25;
 $27 = ((($slot)) + 92|0);
 HEAP16[$27>>1] = 1;
 $28 = ((($slot)) + 94|0);
 HEAP8[$28>>0] = 1;
 $29 = ((($0)) + 308|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = $30 & 3;
 $32 = 1 << $31;
 switch ($32|0) {
 case 1:  {
  $33 = (($25) + 7)|0;
  $34 = $33 >>> 3;
  $35 = ((($slot)) + 84|0);
  HEAP32[$35>>2] = $34;
  $49 = $34;
  break;
 }
 case 2:  {
  $36 = (($25) + 15)|0;
  $37 = $36 >>> 4;
  $38 = $37 << 1;
  $39 = ((($slot)) + 84|0);
  HEAP32[$39>>2] = $38;
  $49 = $38;
  break;
 }
 case 4:  {
  $40 = (($25) + 31)|0;
  $41 = $40 >>> 5;
  $42 = $41 << 2;
  $43 = ((($slot)) + 84|0);
  HEAP32[$43>>2] = $42;
  $49 = $42;
  break;
 }
 case 8:  {
  $44 = (($25) + 63)|0;
  $45 = $44 >>> 6;
  $46 = $45 << 3;
  $47 = ((($slot)) + 84|0);
  HEAP32[$47>>2] = $46;
  $49 = $46;
  break;
 }
 default: {
  $$0 = 3;
  return ($$0|0);
 }
 }
 $48 = Math_imul($17, $49)|0;
 $50 = (_ft_glyphslot_alloc_bitmap($slot,$48)|0);
 $51 = ($50|0)==(0);
 if (!($51)) {
  $$0 = $50;
  return ($$0|0);
 }
 $52 = (((($10) + ($glyph_index$<<4)|0)) + 12|0);
 $53 = HEAP32[$52>>2]|0;
 $54 = (_FT_Stream_Seek($6,$53)|0);
 $55 = ($54|0)==(0);
 if (!($55)) {
  $$0 = $54;
  return ($$0|0);
 }
 $56 = ((($slot)) + 88|0);
 $57 = HEAP32[$56>>2]|0;
 $58 = (_FT_Stream_Read($6,$57,$48)|0);
 $59 = ($58|0)==(0);
 if (!($59)) {
  $$0 = $58;
  return ($$0|0);
 }
 $60 = HEAP32[$29>>2]|0;
 $$lobit = $60 & 8;
 $61 = ($$lobit|0)!=(0);
 $62 = ($48|0)==(0);
 $or$cond = $61 | $62;
 if ($or$cond) {
  $84 = $60;
 } else {
  $63 = HEAP32[$56>>2]|0;
  $$012$i = $48;$$03$i = $63;
  while(1) {
   $64 = HEAP8[$$03$i>>0]|0;
   $65 = $64&255;
   $66 = $65 >>> 1;
   $67 = $66 & 85;
   $68 = $65 << 1;
   $69 = $68 & 170;
   $70 = $67 | $69;
   $71 = $70 >>> 2;
   $72 = $71 & 51;
   $73 = $70 << 2;
   $74 = $73 & 204;
   $75 = $72 | $74;
   $76 = $75 >>> 4;
   $77 = $75 << 4;
   $78 = $76 | $77;
   $79 = $78&255;
   HEAP8[$$03$i>>0] = $79;
   $80 = (($$012$i) + -1)|0;
   $81 = ((($$03$i)) + 1|0);
   $82 = ($80|0)==(0);
   if ($82) {
    break;
   } else {
    $$012$i = $80;$$03$i = $81;
   }
  }
  $$pre = HEAP32[$29>>2]|0;
  $84 = $$pre;
 }
 $83 = $84 >>> 2;
 $85 = $84 >>> 3;
 $86 = $83 ^ $85;
 $87 = $86 & 1;
 $88 = ($87|0)==(0);
 L29: do {
  if (!($88)) {
   $89 = $84 >>> 4;
   $90 = $89 & 3;
   $91 = 1 << $90;
   switch ($91|0) {
   case 4:  {
    $100 = ($48>>>0)>(3);
    if (!($100)) {
     break L29;
    }
    $101 = HEAP32[$56>>2]|0;
    $$012$i5 = $48;$$03$i4 = $101;
    while(1) {
     $102 = HEAP8[$$03$i4>>0]|0;
     $103 = ((($$03$i4)) + 3|0);
     $104 = HEAP8[$103>>0]|0;
     HEAP8[$$03$i4>>0] = $104;
     HEAP8[$103>>0] = $102;
     $105 = ((($$03$i4)) + 1|0);
     $106 = HEAP8[$105>>0]|0;
     $107 = ((($$03$i4)) + 2|0);
     $108 = HEAP8[$107>>0]|0;
     HEAP8[$105>>0] = $108;
     HEAP8[$107>>0] = $106;
     $109 = (($$012$i5) + -4)|0;
     $110 = ((($$03$i4)) + 4|0);
     $111 = ($109>>>0)>(3);
     if ($111) {
      $$012$i5 = $109;$$03$i4 = $110;
     } else {
      break;
     }
    }
    break;
   }
   case 2:  {
    $92 = ($48>>>0)>(1);
    if (!($92)) {
     break L29;
    }
    $93 = HEAP32[$56>>2]|0;
    $$012$i8 = $48;$$03$i7 = $93;
    while(1) {
     $94 = HEAP8[$$03$i7>>0]|0;
     $95 = ((($$03$i7)) + 1|0);
     $96 = HEAP8[$95>>0]|0;
     HEAP8[$$03$i7>>0] = $96;
     HEAP8[$95>>0] = $94;
     $97 = (($$012$i8) + -2)|0;
     $98 = ((($$03$i7)) + 2|0);
     $99 = ($97>>>0)>(1);
     if ($99) {
      $$012$i8 = $97;$$03$i7 = $98;
     } else {
      break;
     }
    }
    break;
   }
   default: {
    break L29;
   }
   }
  }
 } while(0);
 $112 = ((($slot)) + 72|0);
 HEAP32[$112>>2] = 1651078259;
 $113 = HEAP16[$22>>1]|0;
 $114 = $113 << 16 >> 16;
 $115 = ((($slot)) + 100|0);
 HEAP32[$115>>2] = $114;
 $116 = HEAP16[$11>>1]|0;
 $117 = $116 << 16 >> 16;
 $118 = ((($slot)) + 104|0);
 HEAP32[$118>>2] = $117;
 $119 = (((($10) + ($glyph_index$<<4)|0)) + 4|0);
 $120 = HEAP16[$119>>1]|0;
 $121 = $120 << 16 >> 16;
 $122 = $121 << 6;
 $123 = ((($slot)) + 24|0);
 $124 = ((($slot)) + 40|0);
 HEAP32[$124>>2] = $122;
 $125 = $114 << 6;
 $126 = ((($slot)) + 32|0);
 HEAP32[$126>>2] = $125;
 $127 = $117 << 6;
 $128 = ((($slot)) + 36|0);
 HEAP32[$128>>2] = $127;
 $129 = HEAP16[$19>>1]|0;
 $130 = $129 << 16 >> 16;
 $131 = (($130) - ($114))|0;
 $132 = $131 << 6;
 HEAP32[$123>>2] = $132;
 $133 = HEAP32[$18>>2]|0;
 $134 = $133 << 6;
 $135 = ((($slot)) + 28|0);
 HEAP32[$135>>2] = $134;
 $136 = ((($0)) + 204|0);
 $137 = HEAP32[$136>>2]|0;
 $138 = ((($0)) + 208|0);
 $139 = HEAP32[$138>>2]|0;
 $140 = (($139) + ($137))|0;
 $141 = $140 << 6;
 _ft_synthesize_vertical_metrics($123,$141);
 $$0 = 0;
 return ($$0|0);
}
function _PCF_Size_Request($size,$req) {
 $size = $size|0;
 $req = $req|0;
 var $$0 = 0, $$pre = 0, $$pre$phi4Z2D = 0, $$pre$phi6Z2D = 0, $$pre3 = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$size>>2]|0;
 $1 = ((($0)) + 32|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($req)) + 16|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0);
 $6 = ((($req)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 if ($5) {
  $12 = $7;
 } else {
  $8 = Math_imul($7, $4)|0;
  $9 = (($8) + 36)|0;
  $10 = (($9|0) / 72)&-1;
  $12 = $10;
 }
 $11 = (($12) + 32)|0;
 $13 = $11 >> 6;
 $14 = HEAP32[$req>>2]|0;
 L4: do {
  switch ($14|0) {
  case 0:  {
   $15 = ((($2)) + 12|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = (($16) + 32)|0;
   $18 = $17 >> 6;
   $19 = ($13|0)==($18|0);
   if ($19) {
    $$pre = ((($0)) + 196|0);
    $$pre3 = ((($$pre)) + 8|0);
    $$pre5 = ((($$pre)) + 12|0);
    $$pre$phi4Z2D = $$pre3;$$pre$phi6Z2D = $$pre5;
    break L4;
   } else {
    $$0 = 23;
    return ($$0|0);
   }
   break;
  }
  case 1:  {
   $20 = ((($0)) + 196|0);
   $21 = ((($20)) + 8|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = ((($20)) + 12|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = (($24) + ($22))|0;
   $26 = ($13|0)==($25|0);
   if ($26) {
    $$pre$phi4Z2D = $21;$$pre$phi6Z2D = $23;
   } else {
    $$0 = 23;
    return ($$0|0);
   }
   break;
  }
  default: {
   $$0 = 7;
   return ($$0|0);
  }
  }
 } while(0);
 $27 = ((($0)) + 196|0);
 _FT_Select_Metrics($0,0);
 $28 = HEAP32[$$pre$phi4Z2D>>2]|0;
 $29 = $28 << 6;
 $30 = ((($size)) + 24|0);
 HEAP32[$30>>2] = $29;
 $31 = HEAP32[$$pre$phi6Z2D>>2]|0;
 $32 = (0 - ($31))|0;
 $33 = $32 << 6;
 $34 = ((($size)) + 28|0);
 HEAP32[$34>>2] = $33;
 $35 = ((($27)) + 40|0);
 $36 = HEAP16[$35>>1]|0;
 $37 = $36 << 16 >> 16;
 $38 = $37 << 6;
 $39 = ((($size)) + 36|0);
 HEAP32[$39>>2] = $38;
 $$0 = 0;
 return ($$0|0);
}
function _PCF_Size_Select($size,$strike_index) {
 $size = $size|0;
 $strike_index = $strike_index|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$size>>2]|0;
 $1 = ((($0)) + 196|0);
 _FT_Select_Metrics($0,$strike_index);
 $2 = ((($1)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 6;
 $5 = ((($size)) + 24|0);
 HEAP32[$5>>2] = $4;
 $6 = ((($1)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (0 - ($7))|0;
 $9 = $8 << 6;
 $10 = ((($size)) + 28|0);
 HEAP32[$10>>2] = $9;
 $11 = ((($1)) + 40|0);
 $12 = HEAP16[$11>>1]|0;
 $13 = $12 << 16 >> 16;
 $14 = $13 << 6;
 $15 = ((($size)) + 36|0);
 HEAP32[$15>>2] = $14;
 return 0;
}
function _pfr_cmap_init($cmap,$pointer) {
 $cmap = $cmap|0;
 $pointer = $pointer|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$0 = 0, $n$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$cmap>>2]|0;
 $1 = ((($0)) + 408|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($cmap)) + 16|0);
 HEAP32[$3>>2] = $2;
 $4 = ((($0)) + 416|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($cmap)) + 20|0);
 HEAP32[$6>>2] = $5;
 $7 = ($2>>>0)>(1);
 $8 = $5;
 if (!($7)) {
  $error$0 = 0;
  return ($error$0|0);
 }
 $$pre = HEAP32[$8>>2]|0;
 $14 = $$pre;$n$02 = 1;
 while(1) {
  $11 = (($8) + ($n$02<<4)|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ($14>>>0)<($12>>>0);
  $10 = (($n$02) + 1)|0;
  if (!($13)) {
   $error$0 = 8;
   label = 5;
   break;
  }
  $9 = ($10>>>0)<($2>>>0);
  if ($9) {
   $14 = $12;$n$02 = $10;
  } else {
   $error$0 = 0;
   label = 5;
   break;
  }
 }
 if ((label|0) == 5) {
  return ($error$0|0);
 }
 return (0)|0;
}
function _pfr_cmap_done($cmap) {
 $cmap = $cmap|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 20|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($cmap)) + 16|0);
 HEAP32[$1>>2] = 0;
 return;
}
function _pfr_cmap_char_index($cmap,$char_code) {
 $cmap = $cmap|0;
 $char_code = $char_code|0;
 var $$2 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $max$02 = 0, $max$1 = 0, $min$03 = 0;
 var $min$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  $$2 = 0;
  return ($$2|0);
 }
 $3 = ((($cmap)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 $max$02 = $1;$min$03 = 0;
 while(1) {
  $5 = (($max$02) - ($min$03))|0;
  $6 = $5 >>> 1;
  $7 = (($6) + ($min$03))|0;
  $8 = (($4) + ($7<<4)|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)==($char_code|0);
  if ($10) {
   $$lcssa = $7;
   break;
  }
  $11 = ($9>>>0)<($char_code>>>0);
  $12 = (($7) + 1)|0;
  $max$1 = $11 ? $max$02 : $7;
  $min$1 = $11 ? $12 : $min$03;
  $13 = ($max$1>>>0)>($min$1>>>0);
  if ($13) {
   $max$02 = $max$1;$min$03 = $min$1;
  } else {
   $$2 = 0;
   label = 6;
   break;
  }
 }
 if ((label|0) == 6) {
  return ($$2|0);
 }
 $14 = (($$lcssa) + 1)|0;
 $$2 = $14;
 return ($$2|0);
}
function _pfr_cmap_char_next($cmap,$pchar_code) {
 $cmap = $cmap|0;
 $pchar_code = $pchar_code|0;
 var $$lcssa = 0, $$lcssa$lcssa = 0, $$lcssa31 = 0, $$min$0$us = 0, $$min$0$us$lcssa = 0, $$pre = 0, $$sink$us = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $char_code$2$ph = 0, $max$0$$us = 0, $max$011$us = 0, $min$010$us = 0, $or$cond = 0, $result$2$ph = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 $3 = ((($cmap)) + 20|0);
 do {
  if ($2) {
   $char_code$2$ph = 0;$result$2$ph = 0;
  } else {
   $4 = HEAP32[$pchar_code>>2]|0;
   $$pre = HEAP32[$3>>2]|0;
   $$sink$us = $4;
   L3: while(1) {
    $5 = (($$sink$us) + 1)|0;
    $max$011$us = $1;$min$010$us = 0;
    while(1) {
     $6 = (($max$011$us) - ($min$010$us))|0;
     $7 = $6 >>> 1;
     $8 = (($7) + ($min$010$us))|0;
     $9 = (($$pre) + ($8<<4)|0);
     $10 = HEAP32[$9>>2]|0;
     $11 = ($10|0)==($5|0);
     if ($11) {
      $$lcssa = $8;
      break;
     }
     $12 = ($10>>>0)<($5>>>0);
     $13 = (($8) + 1)|0;
     $$min$0$us = $12 ? $13 : $min$010$us;
     $max$0$$us = $12 ? $max$011$us : $8;
     $14 = ($max$0$$us>>>0)>($$min$0$us>>>0);
     if ($14) {
      $max$011$us = $max$0$$us;$min$010$us = $$min$0$us;
     } else {
      $$min$0$us$lcssa = $$min$0$us;
      break L3;
     }
    }
    $15 = ($$lcssa|0)==(0);
    if ($15) {
     $$sink$us = $5;
    } else {
     $$lcssa$lcssa = $$lcssa;$$lcssa31 = $5;
     label = 7;
     break;
    }
   }
   if ((label|0) == 7) {
    $16 = (($$lcssa$lcssa) + 1)|0;
    $char_code$2$ph = $$lcssa31;$result$2$ph = $16;
    break;
   }
   $17 = ($$min$0$us$lcssa>>>0)>=($1>>>0);
   $18 = ($$min$0$us$lcssa|0)==(0);
   $or$cond = $18 | $17;
   if ($or$cond) {
    $char_code$2$ph = 0;$result$2$ph = 0;
   } else {
    $19 = HEAP32[$3>>2]|0;
    $20 = (($$min$0$us$lcssa) + 1)|0;
    $21 = (($19) + ($$min$0$us$lcssa<<4)|0);
    $22 = HEAP32[$21>>2]|0;
    $char_code$2$ph = $22;$result$2$ph = $20;
   }
  }
 } while(0);
 HEAP32[$pchar_code>>2] = $char_code$2$ph;
 return ($result$2$ph|0);
}
function _pfr_get_service($module,$service_id) {
 $module = $module|0;
 $service_id = $service_id|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_ft_service_list_lookup(53472,$service_id)|0);
 return ($0|0);
}
function _pfr_get_metrics($pfrface,$anoutline_resolution,$ametrics_resolution,$ametrics_x_scale,$ametrics_y_scale) {
 $pfrface = $pfrface|0;
 $anoutline_resolution = $anoutline_resolution|0;
 $ametrics_resolution = $ametrics_resolution|0;
 $ametrics_x_scale = $ametrics_x_scale|0;
 $ametrics_y_scale = $ametrics_y_scale|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $x_scale$0 = 0, $y_scale$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pfrface)) + 288|0);
 $1 = ((($pfrface)) + 88|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($anoutline_resolution|0)==(0|0);
 if (!($3)) {
  $4 = ((($0)) + 12|0);
  $5 = HEAP32[$4>>2]|0;
  HEAP32[$anoutline_resolution>>2] = $5;
 }
 $6 = ($ametrics_resolution|0)==(0|0);
 if (!($6)) {
  $7 = ((($0)) + 16|0);
  $8 = HEAP32[$7>>2]|0;
  HEAP32[$ametrics_resolution>>2] = $8;
 }
 $9 = ($2|0)==(0|0);
 if ($9) {
  $x_scale$0 = 65536;$y_scale$0 = 65536;
 } else {
  $10 = ((($2)) + 12|0);
  $11 = HEAP16[$10>>1]|0;
  $12 = $11&65535;
  $13 = $12 << 6;
  $14 = ((($0)) + 16|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = (_FT_DivFix($13,$15)|0);
  $17 = ((($2)) + 14|0);
  $18 = HEAP16[$17>>1]|0;
  $19 = $18&65535;
  $20 = $19 << 6;
  $21 = HEAP32[$14>>2]|0;
  $22 = (_FT_DivFix($20,$21)|0);
  $x_scale$0 = $16;$y_scale$0 = $22;
 }
 $23 = ($ametrics_x_scale|0)==(0|0);
 if (!($23)) {
  HEAP32[$ametrics_x_scale>>2] = $x_scale$0;
 }
 $24 = ($ametrics_y_scale|0)==(0|0);
 if ($24) {
  return 0;
 }
 HEAP32[$ametrics_y_scale>>2] = $y_scale$0;
 return 0;
}
function _pfr_face_get_kerning($pfrface,$glyph1,$glyph2,$kerning) {
 $pfrface = $pfrface|0;
 $glyph1 = $glyph1|0;
 $glyph2 = $glyph2|0;
 $kerning = $kerning|0;
 var $$01 = 0, $$03 = 0, $$base$0 = 0, $$base$0$us = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $base$0$lcssa = 0, $base$0$ph = 0, $base$08 = 0;
 var $base$08$us = 0, $cpair$0 = 0, $cpair$2 = 0, $glyph1$ = 0, $item$0 = 0, $item$010 = 0, $item$010$lcssa = 0, $item$09 = 0, $or$cond = 0, $p$0 = 0, $p$1 = 0, $p$3 = 0, $p$4 = 0, $probe$07 = 0, $probe$07$us = 0, $sext = 0, $value$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pfrface)) + 288|0);
 HEAP32[$kerning>>2] = 0;
 $1 = ((($kerning)) + 4|0);
 HEAP32[$1>>2] = 0;
 $2 = ($glyph1|0)==(0);
 $3 = (($glyph1) + -1)|0;
 $glyph1$ = $2 ? 0 : $3;
 $4 = ($glyph2|0)==(0);
 $5 = (($glyph2) + -1)|0;
 $$03 = $4 ? 0 : $5;
 $6 = ((($0)) + 120|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($glyph1$>>>0)>($7>>>0);
 $9 = ($$03>>>0)>($7>>>0);
 $or$cond = $8 | $9;
 if ($or$cond) {
  $$01 = 0;
  return ($$01|0);
 }
 $10 = ((($0)) + 128|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + ($glyph1$<<4)|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11) + ($$03<<4)|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = $13 << 16;
 $17 = $15 & 65535;
 $18 = $17 | $16;
 $19 = ((($0)) + 136|0);
 $20 = ((($pfrface)) + 104|0);
 $21 = HEAP32[$20>>2]|0;
 $item$09 = HEAP32[$19>>2]|0;
 $22 = ($item$09|0)==(0|0);
 if ($22) {
  $$01 = 0;
  return ($$01|0);
 } else {
  $item$010 = $item$09;
 }
 while(1) {
  $23 = ((($item$010)) + 16|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = ($18>>>0)<($24>>>0);
  if (!($25)) {
   $26 = ((($item$010)) + 20|0);
   $27 = HEAP32[$26>>2]|0;
   $28 = ($18>>>0)>($27>>>0);
   if (!($28)) {
    $item$010$lcssa = $item$010;
    break;
   }
  }
  $item$0 = HEAP32[$item$010>>2]|0;
  $29 = ($item$0|0)==(0|0);
  if ($29) {
   $$01 = 0;
   label = 29;
   break;
  } else {
   $item$010 = $item$0;
  }
 }
 if ((label|0) == 29) {
  return ($$01|0);
 }
 $30 = ((($item$010$lcssa)) + 12|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = (_FT_Stream_Seek($21,$31)|0);
 $33 = ($32|0)==(0);
 if (!($33)) {
  $$01 = $32;
  return ($$01|0);
 }
 $34 = ((($item$010$lcssa)) + 4|0);
 $35 = HEAP8[$34>>0]|0;
 $36 = $35&255;
 $37 = ((($item$010$lcssa)) + 8|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = Math_imul($36, $38)|0;
 $40 = (_FT_Stream_EnterFrame($21,$39)|0);
 $41 = ($40|0)==(0);
 if (!($41)) {
  $$01 = $40;
  return ($$01|0);
 }
 $42 = HEAP32[$34>>2]|0;
 $43 = $42 & 255;
 $44 = HEAP32[$37>>2]|0;
 $45 = (Math_clz32(($43|0))|0);
 $46 = (31 - ($45))|0;
 $47 = 1 << $46;
 $48 = $44 << $46;
 $49 = ((($21)) + 32|0);
 $50 = HEAP32[$49>>2]|0;
 $51 = $42 >>> 8;
 $52 = $51 & 1;
 $53 = $52&255;
 $54 = $51 & 2;
 $55 = $54&255;
 $56 = ($43|0)==($47|0);
 if ($56) {
  $base$0$ph = $50;
  label = 15;
 } else {
  $57 = (($43) - ($47))|0;
  $58 = Math_imul($57, $44)|0;
  $59 = (($50) + ($58)|0);
  $60 = ($53<<24>>24)==(0);
  if ($60) {
   $80 = ((($59)) + 2|0);
   $81 = ((($80)) + -2|0);
   $82 = HEAP8[$81>>0]|0;
   $83 = $82&255;
   $84 = $83 << 16;
   $85 = ((($80)) + -1|0);
   $86 = HEAP8[$85>>0]|0;
   $87 = $86&255;
   $88 = $84 | $87;
   $cpair$0 = $88;$p$0 = $80;
  } else {
   $61 = ((($59)) + 4|0);
   $62 = ((($61)) + -4|0);
   $63 = HEAP8[$62>>0]|0;
   $64 = $63&255;
   $65 = $64 << 24;
   $66 = ((($62)) + 1|0);
   $67 = HEAP8[$66>>0]|0;
   $68 = $67&255;
   $69 = $68 << 16;
   $70 = $69 | $65;
   $71 = ((($62)) + 2|0);
   $72 = HEAP8[$71>>0]|0;
   $73 = $72&255;
   $74 = $73 << 8;
   $75 = $70 | $74;
   $76 = ((($62)) + 3|0);
   $77 = HEAP8[$76>>0]|0;
   $78 = $77&255;
   $79 = $75 | $78;
   $cpair$0 = $79;$p$0 = $61;
  }
  $89 = ($cpair$0|0)==($18|0);
  if ($89) {
   $p$4 = $p$0;
   label = 25;
  } else {
   $90 = ($cpair$0>>>0)<($18>>>0);
   if ($90) {
    $91 = ($55<<24>>24)==(0);
    $92 = ((($p$0)) + 2|0);
    $93 = ((($p$0)) + 1|0);
    $p$1 = $91 ? $93 : $92;
    $base$0$ph = $p$1;
    label = 15;
   } else {
    $base$0$ph = $50;
    label = 15;
   }
  }
 }
 L28: do {
  if ((label|0) == 15) {
   $94 = ($48>>>0)>($44>>>0);
   if ($94) {
    $95 = ($53<<24>>24)==(0);
    if ($95) {
     $base$08$us = $base$0$ph;$probe$07$us = $48;
     while(1) {
      $96 = $probe$07$us >>> 1;
      $97 = (($base$08$us) + ($96)|0);
      $98 = ((($97)) + 2|0);
      $99 = ((($98)) + -2|0);
      $100 = HEAP8[$99>>0]|0;
      $101 = $100&255;
      $102 = $101 << 16;
      $103 = ((($98)) + -1|0);
      $104 = HEAP8[$103>>0]|0;
      $105 = $104&255;
      $106 = $102 | $105;
      $107 = ($106|0)==($18|0);
      if ($107) {
       $p$4 = $98;
       label = 25;
       break L28;
      }
      $108 = ($106>>>0)<($18>>>0);
      $$base$0$us = $108 ? $97 : $base$08$us;
      $109 = ($96>>>0)>($44>>>0);
      if ($109) {
       $base$08$us = $$base$0$us;$probe$07$us = $96;
      } else {
       $base$0$lcssa = $$base$0$us;
       break;
      }
     }
    } else {
     $base$08 = $base$0$ph;$probe$07 = $48;
     while(1) {
      $110 = $probe$07 >>> 1;
      $111 = (($base$08) + ($110)|0);
      $112 = ((($111)) + 4|0);
      $113 = ((($112)) + -4|0);
      $114 = HEAP8[$113>>0]|0;
      $115 = $114&255;
      $116 = $115 << 24;
      $117 = ((($113)) + 1|0);
      $118 = HEAP8[$117>>0]|0;
      $119 = $118&255;
      $120 = $119 << 16;
      $121 = $120 | $116;
      $122 = ((($113)) + 2|0);
      $123 = HEAP8[$122>>0]|0;
      $124 = $123&255;
      $125 = $124 << 8;
      $126 = $121 | $125;
      $127 = ((($113)) + 3|0);
      $128 = HEAP8[$127>>0]|0;
      $129 = $128&255;
      $130 = $126 | $129;
      $131 = ($130|0)==($18|0);
      if ($131) {
       $p$4 = $112;
       label = 25;
       break L28;
      }
      $132 = ($130>>>0)<($18>>>0);
      $$base$0 = $132 ? $111 : $base$08;
      $133 = ($110>>>0)>($44>>>0);
      if ($133) {
       $base$08 = $$base$0;$probe$07 = $110;
      } else {
       $base$0$lcssa = $$base$0;
       break;
      }
     }
    }
   } else {
    $base$0$lcssa = $base$0$ph;
   }
   $134 = ($53<<24>>24)==(0);
   if ($134) {
    $153 = ((($base$0$lcssa)) + 2|0);
    $154 = HEAP8[$base$0$lcssa>>0]|0;
    $155 = $154&255;
    $156 = $155 << 16;
    $157 = ((($base$0$lcssa)) + 1|0);
    $158 = HEAP8[$157>>0]|0;
    $159 = $158&255;
    $160 = $156 | $159;
    $cpair$2 = $160;$p$3 = $153;
   } else {
    $135 = ((($base$0$lcssa)) + 4|0);
    $136 = HEAP8[$base$0$lcssa>>0]|0;
    $137 = $136&255;
    $138 = $137 << 24;
    $139 = ((($base$0$lcssa)) + 1|0);
    $140 = HEAP8[$139>>0]|0;
    $141 = $140&255;
    $142 = $141 << 16;
    $143 = $142 | $138;
    $144 = ((($base$0$lcssa)) + 2|0);
    $145 = HEAP8[$144>>0]|0;
    $146 = $145&255;
    $147 = $146 << 8;
    $148 = $143 | $147;
    $149 = ((($base$0$lcssa)) + 3|0);
    $150 = HEAP8[$149>>0]|0;
    $151 = $150&255;
    $152 = $148 | $151;
    $cpair$2 = $152;$p$3 = $135;
   }
   $161 = ($cpair$2|0)==($18|0);
   if ($161) {
    $p$4 = $p$3;
    label = 25;
   }
  }
 } while(0);
 if ((label|0) == 25) {
  $162 = ($55<<24>>24)==(0);
  $163 = HEAP8[$p$4>>0]|0;
  $164 = $163&255;
  if ($162) {
   $value$0 = $164;
  } else {
   $165 = $164 << 8;
   $166 = ((($p$4)) + 1|0);
   $167 = HEAP8[$166>>0]|0;
   $168 = $167&255;
   $169 = $168 | $165;
   $sext = $169 << 16;
   $170 = $sext >> 16;
   $value$0 = $170;
  }
  $171 = $42 >> 16;
  $172 = (($171) + ($value$0))|0;
  HEAP32[$kerning>>2] = $172;
 }
 _FT_Stream_ExitFrame($21);
 $$01 = 0;
 return ($$01|0);
}
function _pfr_get_advance($pfrface,$gindex,$anadvance) {
 $pfrface = $pfrface|0;
 $gindex = $gindex|0;
 $anadvance = $anadvance|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$anadvance>>2] = 0;
 $0 = ($gindex|0)==(0);
 if ($0) {
  $error$1 = 6;
  return ($error$1|0);
 }
 $1 = (($gindex) + -1)|0;
 $2 = ($pfrface|0)==(0|0);
 if ($2) {
  $error$1 = 6;
  return ($error$1|0);
 }
 $3 = ((($pfrface)) + 288|0);
 $4 = ((($3)) + 120|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($1>>>0)<($5>>>0);
 if (!($6)) {
  $error$1 = 6;
  return ($error$1|0);
 }
 $7 = ((($3)) + 128|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (((($8) + ($1<<4)|0)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 HEAP32[$anadvance>>2] = $10;
 $error$1 = 0;
 return ($error$1|0);
}
function _pfr_face_init($stream,$pfrface,$face_index,$num_params,$params) {
 $stream = $stream|0;
 $pfrface = $pfrface|0;
 $face_index = $face_index|0;
 $num_params = $num_params|0;
 $params = $params|0;
 var $$ = 0, $$$$i = 0, $$0$i$ph = 0, $$05$i$i = 0, $$05$i14$i = 0, $$3$i = 0, $$Size$1$i = 0, $$Size$3$i = 0, $$lcssa = 0, $$lcssa66$i = 0, $$lcssa78 = 0, $$lcssa79 = 0, $$lcssa80 = 0, $$lobit$i = 0, $$lobit37$i = 0, $$lobit38$i = 0, $$max$0 = 0, $$max$0$lcssa = 0, $$ph = 0, $$pre = 0;
 var $$pre$i = 0, $$pre$i$i = 0, $$pre$i20$i = 0, $$pre$i8 = 0, $$pre49 = 0, $$pre50 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0;
 var $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0;
 var $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0;
 var $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0;
 var $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0;
 var $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0;
 var $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0;
 var $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0;
 var $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0;
 var $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0;
 var $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0;
 var $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0;
 var $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0;
 var $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0;
 var $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0;
 var $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0;
 var $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0;
 var $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0;
 var $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0;
 var $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0;
 var $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0;
 var $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0;
 var $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0;
 var $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0;
 var $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0;
 var $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0;
 var $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0;
 var $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0;
 var $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0;
 var $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0;
 var $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0;
 var $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0;
 var $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0;
 var $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0;
 var $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0;
 var $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0;
 var $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0;
 var $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $Size$1$i = 0, $Size$3$i = 0, $charmap = 0;
 var $count3$025 = 0, $div = 0, $error = 0, $error$0$i$ph = 0, $error$4$i$ph$i = 0, $error$i = 0, $error$i$i = 0, $error$i13$i = 0, $error$i4 = 0, $error$i5 = 0, $exitcond = 0, $exitcond$i = 0, $exitcond$i$us = 0, $exitcond69$i = 0, $exitcond69$i$us = 0, $exitcond70$i = 0, $gchar$026 = 0, $local$1$i = 0, $local$2$i = 0, $max$0$lcssa = 0;
 var $max$024 = 0, $n$01$i$i = 0, $n$01$i16$i = 0, $n$027 = 0, $n$053$i = 0, $n3$051$i = 0, $n3$051$i$us = 0, $n3$051$us$i = 0, $n3$051$us$i$us = 0, $nn$0$lcssa = 0, $nn$032 = 0, $num_aux$056$i = 0, $num_items$05$i$i = 0, $num_items$05$us$i$i$i = 0, $or$cond$i = 0, $or$cond$i$i = 0, $or$cond$i17$i = 0, $p$0$i = 0, $p$04$i$i = 0, $p$04$us$i$i$i = 0;
 var $p$1$i = 0, $p$1$i$i = 0, $p$1$i$i$i = 0, $p$2$i = 0, $p$3$i = 0, $p$4$i = 0, $p$sroa$0$0$in$i = 0, $p$sroa$0$1$in$i = 0, $p$sroa$0$10$in$i = 0, $p$sroa$0$10$in$i$us = 0, $p$sroa$0$10$in$us$i = 0, $p$sroa$0$10$in$us$i$us = 0, $p$sroa$0$11$i = 0, $p$sroa$0$11$i$us = 0, $p$sroa$0$11$in$i = 0, $p$sroa$0$11$in$i$us = 0, $p$sroa$0$11$in$us$i = 0, $p$sroa$0$11$in$us$i$us = 0, $p$sroa$0$11$us$i = 0, $p$sroa$0$11$us$i$us = 0;
 var $p$sroa$0$3$i$in = 0, $p$sroa$0$4$in$lcssa$i = 0, $p$sroa$0$4$in52$i = 0, $p$sroa$0$650$i = 0, $p$sroa$0$650$i$us = 0, $p$sroa$0$650$us$i = 0, $p$sroa$0$650$us$i$us = 0, $p$sroa$0$9$v$i = 0, $p$sroa$0$9$v$i$us = 0, $p$sroa$0$9$v$us$i = 0, $p$sroa$0$9$v$us$i$us = 0, $phitmp = 0, $q$055$i = 0, $scevgep$i = 0, $sext = 0, $sext$i = 0, $sext$i6 = 0, $sext10$i = 0, $sext2$i = 0, $sext2$i7 = 0;
 var $sext3$i = 0, $sext4$i = 0, $sext5$i = 0, $sext51 = 0, $sext52 = 0, $sext6$i$us = 0, $sext6$us$i$us = 0, $sext7$i = 0, $sext8$i = 0, $sext9$i = 0, $size$028 = 0, $storemerge = 0, $storemerge53 = 0, $strike$029 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i13$i = sp + 32|0;
 $error$i$i = sp + 28|0;
 $error$i5 = sp + 24|0;
 $error$i4 = sp + 20|0;
 $error$i = sp + 16|0;
 $error = sp + 12|0;
 $charmap = sp;
 $0 = ((($pfrface)) + 132|0);
 $1 = (_FT_Stream_Seek($stream,0)|0);
 $2 = ($1|0)==(0);
 if ($2) {
  $3 = (_FT_Stream_ReadFields($stream,172124,$0)|0);
  $4 = ($3|0)==(0);
  if ($4) {
   $5 = ((($pfrface)) + 204|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = $6 << 16;
   $8 = ((($pfrface)) + 168|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (($9) + ($7))|0;
   HEAP32[$8>>2] = $10;
   HEAP32[$error>>2] = 0;
   $11 = HEAP32[$0>>2]|0;
   $12 = ($11|0)==(1346785840);
   if ($12) {
    $13 = ((($pfrface)) + 136|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = ($14>>>0)>(4);
    if (!($15)) {
     $16 = ((($pfrface)) + 144|0);
     $17 = HEAP32[$16>>2]|0;
     $18 = ($17>>>0)<(58);
     if (!($18)) {
      $19 = ((($pfrface)) + 140|0);
      $20 = HEAP32[$19>>2]|0;
      $21 = ($20|0)==(3338);
      if ($21) {
       $22 = ((($pfrface)) + 152|0);
       $23 = HEAP32[$22>>2]|0;
       $24 = (_FT_Stream_Seek($stream,$23)|0);
       HEAP32[$error$i>>2] = $24;
       $25 = ($24|0)==(0);
       if (!($25)) {
        HEAP32[$error>>2] = $24;
        $890 = $24;
        STACKTOP = sp;return ($890|0);
       }
       $26 = (_FT_Stream_ReadUShort($stream,$error$i)|0);
       $27 = HEAP32[$error$i>>2]|0;
       HEAP32[$error>>2] = $27;
       $28 = ($27|0)==(0);
       if (!($28)) {
        $890 = $27;
        STACKTOP = sp;return ($890|0);
       }
       $29 = $26&65535;
       HEAP32[$pfrface>>2] = $29;
       $30 = ($face_index|0)<(0);
       if ($30) {
        $890 = 0;
        STACKTOP = sp;return ($890|0);
       }
       $31 = ($29|0)>($face_index|0);
       if (!($31)) {
        HEAP32[$error>>2] = 6;
        $890 = 6;
        STACKTOP = sp;return ($890|0);
       }
       $32 = ((($pfrface)) + 240|0);
       $33 = HEAP32[$22>>2]|0;
       $34 = HEAP32[$5>>2]|0;
       $35 = ($34|0)==(0);
       $36 = (_FT_Stream_Seek($stream,$33)|0);
       HEAP32[$error$i4>>2] = $36;
       $37 = ($36|0)==(0);
       do {
        if ($37) {
         $38 = (_FT_Stream_ReadUShort($stream,$error$i4)|0);
         $39 = HEAP32[$error$i4>>2]|0;
         $40 = ($39|0)==(0);
         if ($40) {
          $41 = $38&65535;
          $42 = ($41>>>0)>($face_index>>>0);
          if ($42) {
           $43 = ($face_index*5)|0;
           $44 = (_FT_Stream_Skip($stream,$43)|0);
           HEAP32[$error$i4>>2] = $44;
           $45 = ($44|0)==(0);
           if ($45) {
            $46 = (_FT_Stream_ReadUShort($stream,$error$i4)|0);
            $47 = $46&65535;
            $48 = HEAP32[$error$i4>>2]|0;
            $49 = ($48|0)==(0);
            if ($49) {
             $50 = (_FT_Stream_ReadUOffset($stream,$error$i4)|0);
             $51 = HEAP32[$error$i4>>2]|0;
             $52 = ($51|0)==(0);
             if ($52) {
              HEAP32[$32>>2] = $47;
              $53 = ((($pfrface)) + 244|0);
              HEAP32[$53>>2] = $50;
              $54 = (_FT_Stream_Seek($stream,$50)|0);
              HEAP32[$error$i4>>2] = $54;
              $55 = ($54|0)==(0);
              if ($55) {
               $56 = (_FT_Stream_EnterFrame($stream,$47)|0);
               HEAP32[$error$i4>>2] = $56;
               $57 = ($56|0)==(0);
               if (!($57)) {
                $$0$i$ph = $56;
                break;
               }
               $58 = ((($stream)) + 32|0);
               $59 = HEAP32[$58>>2]|0;
               $60 = (($59) + ($47)|0);
               $61 = ($46&65535)<(13);
               L32: do {
                if ($61) {
                 label = 53;
                } else {
                 $62 = ((($59)) + 3|0);
                 $63 = HEAP8[$59>>0]|0;
                 $64 = $63&255;
                 $65 = $64 << 16;
                 $66 = ((($59)) + 1|0);
                 $67 = HEAP8[$66>>0]|0;
                 $68 = $67&255;
                 $69 = $68 << 8;
                 $70 = $69 | $65;
                 $71 = ((($59)) + 2|0);
                 $72 = HEAP8[$71>>0]|0;
                 $73 = $72&255;
                 $74 = $70 | $73;
                 $75 = ((($32)) + 8|0);
                 HEAP32[$75>>2] = $74;
                 $76 = ((($59)) + 6|0);
                 $77 = HEAP8[$62>>0]|0;
                 $78 = $77&255;
                 $79 = $78 << 16;
                 $80 = ((($59)) + 4|0);
                 $81 = HEAP8[$80>>0]|0;
                 $82 = $81&255;
                 $83 = $82 << 8;
                 $84 = $83 | $79;
                 $85 = ((($59)) + 5|0);
                 $86 = HEAP8[$85>>0]|0;
                 $87 = $86&255;
                 $88 = $84 | $87;
                 $89 = ((($32)) + 12|0);
                 HEAP32[$89>>2] = $88;
                 $90 = ((($59)) + 9|0);
                 $91 = HEAP8[$76>>0]|0;
                 $92 = $91&255;
                 $93 = $92 << 16;
                 $94 = ((($59)) + 7|0);
                 $95 = HEAP8[$94>>0]|0;
                 $96 = $95&255;
                 $97 = $96 << 8;
                 $98 = $97 | $93;
                 $99 = ((($59)) + 8|0);
                 $100 = HEAP8[$99>>0]|0;
                 $101 = $100&255;
                 $102 = $98 | $101;
                 $103 = ((($32)) + 16|0);
                 HEAP32[$103>>2] = $102;
                 $104 = ((($59)) + 12|0);
                 $105 = HEAP8[$90>>0]|0;
                 $106 = $105&255;
                 $107 = $106 << 16;
                 $108 = ((($59)) + 10|0);
                 $109 = HEAP8[$108>>0]|0;
                 $110 = $109&255;
                 $111 = $110 << 8;
                 $112 = $111 | $107;
                 $113 = ((($59)) + 11|0);
                 $114 = HEAP8[$113>>0]|0;
                 $115 = $114&255;
                 $116 = $112 | $115;
                 $117 = ((($32)) + 20|0);
                 HEAP32[$117>>2] = $116;
                 $118 = ((($59)) + 13|0);
                 $119 = HEAP8[$104>>0]|0;
                 $120 = $119&255;
                 $121 = $120 & 4;
                 $122 = ($121|0)!=(0);
                 if ($122) {
                  $123 = $120 >>> 3;
                  $124 = $123 & 1;
                  $125 = (($124) + 1)|0;
                  $126 = $120 & 3;
                  $127 = ($126|0)==(0);
                  $128 = $124 | 4;
                  $$$$i = $127 ? $128 : $125;
                  $local$1$i = $$$$i;
                 } else {
                  $local$1$i = 0;
                 }
                 $129 = $120 & 16;
                 $130 = ($129|0)!=(0);
                 if ($130) {
                  $131 = $120 >>> 5;
                  $132 = $131 & 1;
                  $133 = (($132) + 1)|0;
                  $$3$i = (($133) + ($local$1$i))|0;
                  $local$2$i = $$3$i;
                 } else {
                  $local$2$i = $local$1$i;
                 }
                 $134 = (($118) + ($local$2$i)|0);
                 $135 = ($134>>>0)>($60>>>0);
                 if ($135) {
                  label = 53;
                  break;
                 }
                 do {
                  if ($122) {
                   $136 = $120 & 8;
                   $137 = ($136|0)==(0);
                   if ($137) {
                    $147 = ((($59)) + 14|0);
                    $148 = HEAP8[$118>>0]|0;
                    $149 = $148&255;
                    $151 = $149;$p$0$i = $147;
                   } else {
                    $138 = ((($59)) + 15|0);
                    $139 = HEAP8[$118>>0]|0;
                    $140 = $139&255;
                    $141 = $140 << 8;
                    $142 = ((($59)) + 14|0);
                    $143 = HEAP8[$142>>0]|0;
                    $144 = $143&255;
                    $145 = $141 | $144;
                    $sext2$i = $145 << 16;
                    $146 = $sext2$i >> 16;
                    $151 = $146;$p$0$i = $138;
                   }
                   $150 = ((($32)) + 28|0);
                   HEAP32[$150>>2] = $151;
                   $152 = $120 & 3;
                   $153 = ($152|0)==(0);
                   if (!($153)) {
                    $p$1$i = $p$0$i;
                    break;
                   }
                   $154 = ((($p$0$i)) + 3|0);
                   $155 = HEAP8[$p$0$i>>0]|0;
                   $156 = $155&255;
                   $157 = $156 << 16;
                   $158 = ((($p$0$i)) + 1|0);
                   $159 = HEAP8[$158>>0]|0;
                   $160 = $159&255;
                   $161 = $160 << 8;
                   $162 = $161 | $157;
                   $163 = ((($p$0$i)) + 2|0);
                   $164 = HEAP8[$163>>0]|0;
                   $165 = $164&255;
                   $166 = $162 | $165;
                   $167 = ((($32)) + 36|0);
                   HEAP32[$167>>2] = $166;
                   $p$1$i = $154;
                  } else {
                   $p$1$i = $118;
                  }
                 } while(0);
                 if ($130) {
                  $168 = $120 & 32;
                  $169 = ($168|0)==(0);
                  if ($169) {
                   $179 = ((($p$1$i)) + 1|0);
                   $180 = HEAP8[$p$1$i>>0]|0;
                   $181 = $180&255;
                   $183 = $181;$p$2$i = $179;
                  } else {
                   $170 = ((($p$1$i)) + 2|0);
                   $171 = HEAP8[$p$1$i>>0]|0;
                   $172 = $171&255;
                   $173 = $172 << 8;
                   $174 = ((($p$1$i)) + 1|0);
                   $175 = HEAP8[$174>>0]|0;
                   $176 = $175&255;
                   $177 = $173 | $176;
                   $sext$i = $177 << 16;
                   $178 = $sext$i >> 16;
                   $183 = $178;$p$2$i = $170;
                  }
                  $182 = ((($32)) + 32|0);
                  HEAP32[$182>>2] = $183;
                  $p$3$i = $p$2$i;
                 } else {
                  $p$3$i = $p$1$i;
                 }
                 $184 = $120 & 64;
                 $185 = ($184|0)==(0);
                 L56: do {
                  if ($185) {
                   $p$4$i = $p$3$i;
                  } else {
                   $186 = ((($p$3$i)) + 1|0);
                   $187 = ($186>>>0)>($60>>>0);
                   L58: do {
                    if (!($187)) {
                     $188 = HEAP8[$p$3$i>>0]|0;
                     $189 = ($188<<24>>24)==(0);
                     if ($189) {
                      $p$1$i$i$i = $186;
                     } else {
                      $190 = $188&255;
                      $num_items$05$us$i$i$i = $190;$p$04$us$i$i$i = $186;
                      while(1) {
                       $191 = ((($p$04$us$i$i$i)) + 2|0);
                       $192 = ($191>>>0)>($60>>>0);
                       if ($192) {
                        break L58;
                       }
                       $193 = HEAP8[$p$04$us$i$i$i>>0]|0;
                       $194 = $193&255;
                       $195 = (($191) + ($194)|0);
                       $196 = ($195>>>0)>($60>>>0);
                       if ($196) {
                        break L58;
                       }
                       $197 = (($num_items$05$us$i$i$i) + -1)|0;
                       $198 = ($197|0)==(0);
                       if ($198) {
                        $p$1$i$i$i = $195;
                        break;
                       } else {
                        $num_items$05$us$i$i$i = $197;$p$04$us$i$i$i = $195;
                       }
                      }
                     }
                     HEAP32[$error$i4>>2] = 0;
                     $p$4$i = $p$1$i$i$i;
                     break L56;
                    }
                   } while(0);
                   HEAP32[$error$i4>>2] = 8;
                   break L32;
                  }
                 } while(0);
                 $199 = ((($p$4$i)) + 5|0);
                 $200 = ($199>>>0)>($60>>>0);
                 if ($200) {
                  label = 53;
                  break;
                 }
                 $201 = ((($p$4$i)) + 2|0);
                 $202 = HEAP8[$p$4$i>>0]|0;
                 $203 = $202&255;
                 $204 = $203 << 8;
                 $205 = ((($p$4$i)) + 1|0);
                 $206 = HEAP8[$205>>0]|0;
                 $207 = $206&255;
                 $208 = $204 | $207;
                 $209 = ((($32)) + 40|0);
                 HEAP32[$209>>2] = $208;
                 $210 = HEAP8[$201>>0]|0;
                 $211 = $210&255;
                 $212 = $211 << 16;
                 $213 = ((($p$4$i)) + 3|0);
                 $214 = HEAP8[$213>>0]|0;
                 $215 = $214&255;
                 $216 = $215 << 8;
                 $217 = $216 | $212;
                 $218 = ((($p$4$i)) + 4|0);
                 $219 = HEAP8[$218>>0]|0;
                 $220 = $219&255;
                 $221 = $217 | $220;
                 $222 = ((($32)) + 44|0);
                 HEAP32[$222>>2] = $221;
                 if ($35) {
                  break;
                 }
                 $223 = ((($p$4$i)) + 6|0);
                 $224 = ($223>>>0)>($60>>>0);
                 if ($224) {
                  label = 53;
                  break;
                 }
                 $225 = HEAP8[$199>>0]|0;
                 $226 = $225&255;
                 $227 = $226 << 16;
                 $228 = $227 | $208;
                 HEAP32[$209>>2] = $228;
                }
               } while(0);
               if ((label|0) == 53) {
                HEAP32[$error$i4>>2] = 8;
               }
               _FT_Stream_ExitFrame($stream);
               $$pre$i = HEAP32[$error$i4>>2]|0;
               HEAP32[$error>>2] = $$pre$i;
               $229 = ($$pre$i|0)==(0);
               if (!($229)) {
                $890 = $$pre$i;
                STACKTOP = sp;return ($890|0);
               }
               $230 = ((($pfrface)) + 288|0);
               $231 = ((($32)) + 44|0);
               $232 = HEAP32[$231>>2]|0;
               $233 = ((($32)) + 40|0);
               $234 = HEAP32[$233>>2]|0;
               $235 = ((($stream)) + 28|0);
               $236 = HEAP32[$235>>2]|0;
               HEAP32[$230>>2] = $236;
               $237 = ((($230)) + 4|0);
               HEAP32[$237>>2] = $232;
               $238 = ((($230)) + 136|0);
               HEAP32[$238>>2] = 0;
               $239 = ((($230)) + 140|0);
               HEAP32[$239>>2] = $238;
               $240 = (_FT_Stream_Seek($stream,$232)|0);
               HEAP32[$error$i5>>2] = $240;
               $241 = ($240|0)==(0);
               L78: do {
                if ($241) {
                 $242 = (_FT_Stream_EnterFrame($stream,$234)|0);
                 HEAP32[$error$i5>>2] = $242;
                 $243 = ($242|0)==(0);
                 if (!($243)) {
                  $$ph = $242;
                  break;
                 }
                 $244 = HEAP32[$58>>2]|0;
                 $245 = ((($230)) + 148|0);
                 HEAP32[$245>>2] = $244;
                 $246 = $244;
                 $247 = (($246) + ($234)|0);
                 $248 = ($234|0)<(15);
                 L81: do {
                  if ($248) {
                   label = 147;
                  } else {
                   $249 = ((($246)) + 2|0);
                   $250 = HEAP8[$246>>0]|0;
                   $251 = $250&255;
                   $252 = $251 << 8;
                   $253 = ((($246)) + 1|0);
                   $254 = HEAP8[$253>>0]|0;
                   $255 = $254&255;
                   $256 = $252 | $255;
                   $257 = ((($230)) + 8|0);
                   HEAP32[$257>>2] = $256;
                   $258 = ((($246)) + 4|0);
                   $259 = HEAP8[$249>>0]|0;
                   $260 = $259&255;
                   $261 = $260 << 8;
                   $262 = ((($246)) + 3|0);
                   $263 = HEAP8[$262>>0]|0;
                   $264 = $263&255;
                   $265 = $261 | $264;
                   $266 = ((($230)) + 12|0);
                   HEAP32[$266>>2] = $265;
                   $267 = ((($246)) + 6|0);
                   $268 = HEAP8[$258>>0]|0;
                   $269 = $268&255;
                   $270 = $269 << 8;
                   $271 = ((($246)) + 5|0);
                   $272 = HEAP8[$271>>0]|0;
                   $273 = $272&255;
                   $274 = $270 | $273;
                   $275 = ((($230)) + 16|0);
                   HEAP32[$275>>2] = $274;
                   $276 = ((($246)) + 8|0);
                   $277 = HEAP8[$267>>0]|0;
                   $278 = $277&255;
                   $279 = $278 << 8;
                   $280 = ((($246)) + 7|0);
                   $281 = HEAP8[$280>>0]|0;
                   $282 = $281&255;
                   $283 = $279 | $282;
                   $sext$i6 = $283 << 16;
                   $284 = $sext$i6 >> 16;
                   $285 = ((($230)) + 20|0);
                   HEAP32[$285>>2] = $284;
                   $286 = ((($246)) + 10|0);
                   $287 = HEAP8[$276>>0]|0;
                   $288 = $287&255;
                   $289 = $288 << 8;
                   $290 = ((($246)) + 9|0);
                   $291 = HEAP8[$290>>0]|0;
                   $292 = $291&255;
                   $293 = $289 | $292;
                   $sext2$i7 = $293 << 16;
                   $294 = $sext2$i7 >> 16;
                   $295 = ((($230)) + 24|0);
                   HEAP32[$295>>2] = $294;
                   $296 = ((($246)) + 12|0);
                   $297 = HEAP8[$286>>0]|0;
                   $298 = $297&255;
                   $299 = $298 << 8;
                   $300 = ((($246)) + 11|0);
                   $301 = HEAP8[$300>>0]|0;
                   $302 = $301&255;
                   $303 = $299 | $302;
                   $sext3$i = $303 << 16;
                   $304 = $sext3$i >> 16;
                   $305 = ((($230)) + 28|0);
                   HEAP32[$305>>2] = $304;
                   $306 = ((($246)) + 14|0);
                   $307 = HEAP8[$296>>0]|0;
                   $308 = $307&255;
                   $309 = $308 << 8;
                   $310 = ((($246)) + 13|0);
                   $311 = HEAP8[$310>>0]|0;
                   $312 = $311&255;
                   $313 = $309 | $312;
                   $sext4$i = $313 << 16;
                   $314 = $sext4$i >> 16;
                   $315 = ((($230)) + 32|0);
                   HEAP32[$315>>2] = $314;
                   $316 = ((($246)) + 15|0);
                   $317 = HEAP8[$306>>0]|0;
                   $318 = $317&255;
                   $319 = ((($230)) + 36|0);
                   HEAP32[$319>>2] = $318;
                   $320 = $318 & 4;
                   $321 = ($320|0)!=(0);
                   if ($321) {
                    $p$sroa$0$0$in$i = $316;
                   } else {
                    $322 = ($234|0)<(17);
                    if ($322) {
                     label = 147;
                     break;
                    }
                    $323 = ((($246)) + 17|0);
                    $324 = HEAP8[$316>>0]|0;
                    $325 = $324&255;
                    $326 = $325 << 8;
                    $327 = ((($246)) + 16|0);
                    $328 = HEAP8[$327>>0]|0;
                    $329 = $328&255;
                    $330 = $326 | $329;
                    $sext5$i = $330 << 16;
                    $331 = $sext5$i >> 16;
                    $332 = ((($230)) + 40|0);
                    HEAP32[$332>>2] = $331;
                    $p$sroa$0$0$in$i = $323;
                   }
                   $333 = $318 & 128;
                   $334 = ($333|0)==(0);
                   L87: do {
                    if ($334) {
                     $p$sroa$0$1$in$i = $p$sroa$0$0$in$i;
                    } else {
                     $335 = ((($p$sroa$0$0$in$i)) + 1|0);
                     $336 = ($335>>>0)>($247>>>0);
                     L89: do {
                      if ($336) {
                       $error$4$i$ph$i = 8;
                      } else {
                       $337 = HEAP8[$p$sroa$0$0$in$i>>0]|0;
                       $338 = ($337<<24>>24)==(0);
                       if ($338) {
                        $p$1$i$i = $335;
                       } else {
                        $339 = $337&255;
                        $num_items$05$i$i = $339;$p$04$i$i = $335;
                        while(1) {
                         $340 = ((($p$04$i$i)) + 2|0);
                         $341 = ($340>>>0)>($247>>>0);
                         if ($341) {
                          $error$4$i$ph$i = 8;
                          break L89;
                         }
                         $342 = ((($p$04$i$i)) + 1|0);
                         $343 = HEAP8[$p$04$i$i>>0]|0;
                         $344 = $343&255;
                         $345 = HEAP8[$342>>0]|0;
                         $346 = (($340) + ($344)|0);
                         $347 = ($346>>>0)>($247>>>0);
                         if ($347) {
                          $error$4$i$ph$i = 8;
                          break L89;
                         }
                         $348 = ($345<<24>>24)==(1);
                         do {
                          if ($348) {
                           $$lcssa66$i = 4362;
                           label = 70;
                          } else {
                           $349 = ($345<<24>>24)==(2);
                           if ($349) {
                            $$lcssa66$i = 4363;
                            label = 70;
                            break;
                           }
                           $755 = ($345<<24>>24)==(3);
                           if ($755) {
                            $$lcssa66$i = 4364;
                            label = 70;
                            break;
                           }
                           $756 = ($345<<24>>24)==(4);
                           if ($756) {
                            $$lcssa66$i = 4365;
                            label = 70;
                           }
                          }
                         } while(0);
                         if ((label|0) == 70) {
                          label = 0;
                          $350 = (FUNCTION_TABLE_iiii[$$lcssa66$i & 8191]($340,$346,$230)|0);
                          $351 = ($350|0)==(0);
                          if (!($351)) {
                           $error$4$i$ph$i = $350;
                           break L89;
                          }
                         }
                         $352 = (($num_items$05$i$i) + -1)|0;
                         $353 = ($352|0)==(0);
                         if ($353) {
                          $p$1$i$i = $346;
                          break;
                         } else {
                          $num_items$05$i$i = $352;$p$04$i$i = $346;
                         }
                        }
                       }
                       HEAP32[$error$i5>>2] = 0;
                       $p$sroa$0$1$in$i = $p$1$i$i;
                       break L87;
                      }
                     } while(0);
                     HEAP32[$error$i5>>2] = $error$4$i$ph$i;
                     break L81;
                    }
                   } while(0);
                   $354 = ((($p$sroa$0$1$in$i)) + 3|0);
                   $355 = ($354>>>0)>($247>>>0);
                   if ($355) {
                    label = 147;
                    break;
                   }
                   $356 = HEAP8[$p$sroa$0$1$in$i>>0]|0;
                   $357 = $356&255;
                   $358 = $357 << 16;
                   $359 = ((($p$sroa$0$1$in$i)) + 1|0);
                   $360 = HEAP8[$359>>0]|0;
                   $361 = $360&255;
                   $362 = $361 << 8;
                   $363 = $362 | $358;
                   $364 = ((($p$sroa$0$1$in$i)) + 2|0);
                   $365 = HEAP8[$364>>0]|0;
                   $366 = $365&255;
                   $367 = $363 | $366;
                   $368 = ($367|0)==(0);
                   L108: do {
                    if ($368) {
                     $p$sroa$0$3$i$in = $354;
                    } else {
                     $369 = (($354) + ($367)|0);
                     $370 = ($369>>>0)>($247>>>0);
                     if ($370) {
                      label = 147;
                      break L81;
                     }
                     $371 = ((($230)) + 84|0);
                     $372 = ((($230)) + 44|0);
                     $373 = ((($230)) + 48|0);
                     $374 = ((($230)) + 52|0);
                     $375 = ((($230)) + 88|0);
                     $num_aux$056$i = $367;$q$055$i = $354;
                     L111: while(1) {
                      $376 = ((($q$055$i)) + 4|0);
                      $377 = ($376>>>0)>($369>>>0);
                      if ($377) {
                       $p$sroa$0$3$i$in = $369;
                       break L108;
                      }
                      $378 = ((($q$055$i)) + 2|0);
                      $379 = HEAP8[$q$055$i>>0]|0;
                      $380 = $379&255;
                      $381 = $380 << 8;
                      $382 = ((($q$055$i)) + 1|0);
                      $383 = HEAP8[$382>>0]|0;
                      $384 = $383&255;
                      $385 = $381 | $384;
                      $386 = ($385>>>0)<(4);
                      $387 = ($num_aux$056$i>>>0)<($385>>>0);
                      $or$cond$i = $386 | $387;
                      if ($or$cond$i) {
                       $p$sroa$0$3$i$in = $369;
                       break L108;
                      }
                      $388 = (($378) + ($385)|0);
                      $389 = ((($388)) + -2|0);
                      $390 = HEAP8[$378>>0]|0;
                      $391 = $390&255;
                      $392 = $391 << 8;
                      $393 = ((($q$055$i)) + 3|0);
                      $394 = HEAP8[$393>>0]|0;
                      $395 = $394&255;
                      $396 = $392 | $395;
                      L115: do {
                       switch ($396|0) {
                       case 1:  {
                        $397 = (($385) + -4)|0;
                        HEAP32[$error$i$i>>2] = 0;
                        $398 = ($397|0)==(0);
                        L117: do {
                         if (!($398)) {
                          $399 = (($385) + -5)|0;
                          $400 = (($376) + ($399)|0);
                          $401 = HEAP8[$400>>0]|0;
                          $402 = ($401<<24>>24)==(0);
                          if ($402) {
                           $403 = ($399|0)==(0);
                           if ($403) {
                            break;
                           } else {
                            $$05$i$i = $399;
                           }
                          } else {
                           $$05$i$i = $397;
                          }
                          $n$01$i$i = 0;
                          while(1) {
                           $406 = (($376) + ($n$01$i$i)|0);
                           $407 = HEAP8[$406>>0]|0;
                           $408 = ($407&255)<(32);
                           $409 = ($407<<24>>24)<(0);
                           $or$cond$i$i = $408 | $409;
                           $405 = (($n$01$i$i) + 1)|0;
                           if ($or$cond$i$i) {
                            break L117;
                           }
                           $404 = ($405>>>0)<($$05$i$i>>>0);
                           if ($404) {
                            $n$01$i$i = $405;
                           } else {
                            break;
                           }
                          }
                          $410 = (($$05$i$i) + 1)|0;
                          $411 = (_ft_mem_alloc($236,$410,$error$i$i)|0);
                          $412 = HEAP32[$error$i$i>>2]|0;
                          $413 = ($412|0)==(0);
                          if (!($413)) {
                           $$lcssa79 = $411;$$lcssa80 = $412;
                           label = 88;
                           break L111;
                          }
                          _memcpy(($411|0),($376|0),($$05$i$i|0))|0;
                          $414 = (($411) + ($$05$i$i)|0);
                          HEAP8[$414>>0] = 0;
                          $$pre$i$i = HEAP32[$error$i$i>>2]|0;
                          HEAP32[$371>>2] = $411;
                          HEAP32[$error$i5>>2] = $$pre$i$i;
                          $415 = ($$pre$i$i|0)==(0);
                          if ($415) {
                           break L115;
                          } else {
                           $$ph = $$pre$i$i;
                           break L78;
                          }
                         }
                        } while(0);
                        HEAP32[$371>>2] = 0;
                        HEAP32[$error$i5>>2] = 0;
                        break;
                       }
                       case 2:  {
                        $416 = ((($q$055$i)) + 36|0);
                        $417 = ($416>>>0)>($389>>>0);
                        if ($417) {
                         break L115;
                        }
                        $418 = ((($q$055$i)) + 14|0);
                        $419 = HEAP8[$418>>0]|0;
                        $420 = $419&255;
                        $421 = $420 << 8;
                        $422 = ((($q$055$i)) + 15|0);
                        $423 = HEAP8[$422>>0]|0;
                        $424 = $423&255;
                        $425 = $421 | $424;
                        $sext8$i = $425 << 16;
                        $426 = $sext8$i >> 16;
                        HEAP32[$372>>2] = $426;
                        $427 = ((($q$055$i)) + 16|0);
                        $428 = HEAP8[$427>>0]|0;
                        $429 = $428&255;
                        $430 = $429 << 8;
                        $431 = ((($q$055$i)) + 17|0);
                        $432 = HEAP8[$431>>0]|0;
                        $433 = $432&255;
                        $434 = $430 | $433;
                        $sext9$i = $434 << 16;
                        $435 = $sext9$i >> 16;
                        HEAP32[$373>>2] = $435;
                        $436 = ((($q$055$i)) + 18|0);
                        $437 = HEAP8[$436>>0]|0;
                        $438 = $437&255;
                        $439 = $438 << 8;
                        $440 = ((($q$055$i)) + 19|0);
                        $441 = HEAP8[$440>>0]|0;
                        $442 = $441&255;
                        $443 = $439 | $442;
                        $sext10$i = $443 << 16;
                        $444 = $sext10$i >> 16;
                        HEAP32[$374>>2] = $444;
                        break;
                       }
                       case 3:  {
                        $445 = (($385) + -4)|0;
                        HEAP32[$error$i13$i>>2] = 0;
                        $446 = ($445|0)==(0);
                        L131: do {
                         if (!($446)) {
                          $447 = (($385) + -5)|0;
                          $448 = (($376) + ($447)|0);
                          $449 = HEAP8[$448>>0]|0;
                          $450 = ($449<<24>>24)==(0);
                          if ($450) {
                           $451 = ($447|0)==(0);
                           if ($451) {
                            break;
                           } else {
                            $$05$i14$i = $447;
                           }
                          } else {
                           $$05$i14$i = $445;
                          }
                          $n$01$i16$i = 0;
                          while(1) {
                           $454 = (($376) + ($n$01$i16$i)|0);
                           $455 = HEAP8[$454>>0]|0;
                           $456 = ($455&255)<(32);
                           $457 = ($455<<24>>24)<(0);
                           $or$cond$i17$i = $456 | $457;
                           $453 = (($n$01$i16$i) + 1)|0;
                           if ($or$cond$i17$i) {
                            break L131;
                           }
                           $452 = ($453>>>0)<($$05$i14$i>>>0);
                           if ($452) {
                            $n$01$i16$i = $453;
                           } else {
                            break;
                           }
                          }
                          $458 = (($$05$i14$i) + 1)|0;
                          $459 = (_ft_mem_alloc($236,$458,$error$i13$i)|0);
                          $460 = HEAP32[$error$i13$i>>2]|0;
                          $461 = ($460|0)==(0);
                          if (!($461)) {
                           $$lcssa = $459;$$lcssa78 = $460;
                           label = 100;
                           break L111;
                          }
                          _memcpy(($459|0),($376|0),($$05$i14$i|0))|0;
                          $462 = (($459) + ($$05$i14$i)|0);
                          HEAP8[$462>>0] = 0;
                          $$pre$i20$i = HEAP32[$error$i13$i>>2]|0;
                          HEAP32[$375>>2] = $459;
                          HEAP32[$error$i5>>2] = $$pre$i20$i;
                          $463 = ($$pre$i20$i|0)==(0);
                          if ($463) {
                           break L115;
                          } else {
                           $$ph = $$pre$i20$i;
                           break L78;
                          }
                         }
                        } while(0);
                        HEAP32[$375>>2] = 0;
                        HEAP32[$error$i5>>2] = 0;
                        break;
                       }
                       default: {
                       }
                       }
                      } while(0);
                      $464 = (($num_aux$056$i) - ($385))|0;
                      $465 = ($num_aux$056$i|0)==($385|0);
                      if ($465) {
                       $p$sroa$0$3$i$in = $369;
                       break L108;
                      } else {
                       $num_aux$056$i = $464;$q$055$i = $389;
                      }
                     }
                     if ((label|0) == 88) {
                      HEAP32[$371>>2] = $$lcssa79;
                      HEAP32[$error$i5>>2] = $$lcssa80;
                      $$ph = $$lcssa80;
                      break L78;
                     }
                     else if ((label|0) == 100) {
                      HEAP32[$375>>2] = $$lcssa;
                      HEAP32[$error$i5>>2] = $$lcssa78;
                      $$ph = $$lcssa78;
                      break L78;
                     }
                    }
                   } while(0);
                   $466 = ((($p$sroa$0$3$i$in)) + 1|0);
                   $467 = ($466>>>0)>($247>>>0);
                   if ($467) {
                    label = 147;
                    break;
                   }
                   $468 = HEAP8[$p$sroa$0$3$i$in>>0]|0;
                   $469 = $468&255;
                   $470 = ((($230)) + 104|0);
                   HEAP32[$470>>2] = $469;
                   $471 = $469 << 1;
                   $472 = (($466) + ($471)|0);
                   $473 = ($472>>>0)>($247>>>0);
                   if ($473) {
                    label = 147;
                    break;
                   }
                   $474 = (_ft_mem_realloc($236,4,0,$469,0,$error$i5)|0);
                   $475 = ((($230)) + 108|0);
                   HEAP32[$475>>2] = $474;
                   $476 = HEAP32[$error$i5>>2]|0;
                   $477 = ($476|0)==(0);
                   if (!($477)) {
                    break;
                   }
                   $478 = ($468<<24>>24)==(0);
                   if ($478) {
                    $p$sroa$0$4$in$lcssa$i = $466;
                   } else {
                    $n$053$i = 0;$p$sroa$0$4$in52$i = $466;
                    while(1) {
                     $479 = ((($p$sroa$0$4$in52$i)) + 2|0);
                     $480 = HEAP8[$p$sroa$0$4$in52$i>>0]|0;
                     $481 = $480&255;
                     $482 = $481 << 8;
                     $483 = ((($p$sroa$0$4$in52$i)) + 1|0);
                     $484 = HEAP8[$483>>0]|0;
                     $485 = $484&255;
                     $486 = $482 | $485;
                     $sext7$i = $486 << 16;
                     $487 = $sext7$i >> 16;
                     $488 = (($474) + ($n$053$i<<2)|0);
                     HEAP32[$488>>2] = $487;
                     $489 = (($n$053$i) + 1)|0;
                     $exitcond70$i = ($489|0)==($469|0);
                     if ($exitcond70$i) {
                      break;
                     } else {
                      $n$053$i = $489;$p$sroa$0$4$in52$i = $479;
                     }
                    }
                    $490 = $471 | 1;
                    $scevgep$i = (($p$sroa$0$3$i$in) + ($490)|0);
                    $p$sroa$0$4$in$lcssa$i = $scevgep$i;
                   }
                   $491 = ((($p$sroa$0$4$in$lcssa$i)) + 8|0);
                   $492 = ($491>>>0)>($247>>>0);
                   if ($492) {
                    label = 147;
                    break;
                   }
                   $493 = ((($p$sroa$0$4$in$lcssa$i)) + 1|0);
                   $494 = HEAP8[$p$sroa$0$4$in$lcssa$i>>0]|0;
                   $495 = $494&255;
                   $496 = ((($230)) + 112|0);
                   HEAP32[$496>>2] = $495;
                   $497 = ((($p$sroa$0$4$in$lcssa$i)) + 2|0);
                   $498 = HEAP8[$493>>0]|0;
                   $499 = $498&255;
                   $500 = ((($230)) + 116|0);
                   HEAP32[$500>>2] = $499;
                   $501 = ((($p$sroa$0$4$in$lcssa$i)) + 4|0);
                   $502 = HEAP8[$497>>0]|0;
                   $503 = $502&255;
                   $504 = $503 << 8;
                   $505 = ((($p$sroa$0$4$in$lcssa$i)) + 3|0);
                   $506 = HEAP8[$505>>0]|0;
                   $507 = $506&255;
                   $508 = $504 | $507;
                   $509 = ((($230)) + 68|0);
                   HEAP32[$509>>2] = $508;
                   $510 = ((($p$sroa$0$4$in$lcssa$i)) + 6|0);
                   $511 = HEAP8[$501>>0]|0;
                   $512 = $511&255;
                   $513 = $512 << 8;
                   $514 = ((($p$sroa$0$4$in$lcssa$i)) + 5|0);
                   $515 = HEAP8[$514>>0]|0;
                   $516 = $515&255;
                   $517 = $513 | $516;
                   $518 = ((($230)) + 56|0);
                   HEAP32[$518>>2] = $517;
                   $519 = $491;
                   $520 = HEAP8[$510>>0]|0;
                   $521 = $520&255;
                   $522 = $521 << 8;
                   $523 = ((($p$sroa$0$4$in$lcssa$i)) + 7|0);
                   $524 = HEAP8[$523>>0]|0;
                   $525 = $524&255;
                   $526 = $522 | $525;
                   $527 = ((($230)) + 120|0);
                   HEAP32[$527>>2] = $526;
                   $528 = HEAP32[$58>>2]|0;
                   $529 = (($519) + ($232))|0;
                   $530 = (($529) - ($528))|0;
                   $531 = ((($230)) + 124|0);
                   HEAP32[$531>>2] = $530;
                   $532 = (_ft_mem_realloc($236,16,0,$526,0,$error$i5)|0);
                   $533 = ((($230)) + 128|0);
                   HEAP32[$533>>2] = $532;
                   $534 = HEAP32[$error$i5>>2]|0;
                   $535 = ($534|0)==(0);
                   if (!($535)) {
                    break;
                   }
                   $536 = $318 & 2;
                   $537 = ($536|0)==(0);
                   $538 = $318 >>> 1;
                   $539 = $538 & 3;
                   $Size$1$i = $539 | 4;
                   $540 = $318 & 8;
                   $541 = ($540|0)!=(0);
                   $$lobit$i = $540 >>> 3;
                   $542 = $318 & 16;
                   $543 = ($542|0)==(0);
                   $$lobit37$i = $542 >>> 4;
                   $544 = $318 & 32;
                   $545 = ($544|0)==(0);
                   $$lobit38$i = $544 >>> 5;
                   $$Size$1$i = (($$lobit37$i) + ($$lobit$i))|0;
                   $Size$3$i = (($$Size$1$i) + ($$lobit38$i))|0;
                   $$Size$3$i = (($Size$3$i) + ($Size$1$i))|0;
                   $546 = Math_imul($526, $$Size$3$i)|0;
                   $547 = (($491) + ($546)|0);
                   $548 = ($547>>>0)>($247>>>0);
                   if ($548) {
                    label = 147;
                    break;
                   }
                   $549 = ($526|0)==(0);
                   if ($549) {
                    break;
                   }
                   $550 = ((($230)) + 40|0);
                   if ($537) {
                    if ($321) {
                     $n3$051$us$i$us = 0;$p$sroa$0$650$us$i$us = $519;
                     while(1) {
                      $609 = $p$sroa$0$650$us$i$us;
                      $610 = ((($609)) + 1|0);
                      $611 = HEAP8[$609>>0]|0;
                      $612 = $611&255;
                      $613 = (($532) + ($n3$051$us$i$us<<4)|0);
                      HEAP32[$613>>2] = $612;
                      $614 = ((($609)) + 3|0);
                      $615 = HEAP8[$610>>0]|0;
                      $616 = $615&255;
                      $617 = $616 << 8;
                      $618 = ((($609)) + 2|0);
                      $619 = HEAP8[$618>>0]|0;
                      $620 = $619&255;
                      $621 = $617 | $620;
                      $sext6$us$i$us = $621 << 16;
                      $622 = $sext6$us$i$us >> 16;
                      $623 = (((($532) + ($n3$051$us$i$us<<4)|0)) + 4|0);
                      HEAP32[$623>>2] = $622;
                      $624 = ((($609)) + 4|0);
                      $p$sroa$0$9$v$us$i$us = $541 ? $624 : $614;
                      if ($543) {
                       $633 = ((($p$sroa$0$9$v$us$i$us)) + 1|0);
                       $634 = HEAP8[$p$sroa$0$9$v$us$i$us>>0]|0;
                       $635 = $634&255;
                       $637 = $635;$p$sroa$0$10$in$us$i$us = $633;
                      } else {
                       $625 = ((($p$sroa$0$9$v$us$i$us)) + 2|0);
                       $626 = HEAP8[$p$sroa$0$9$v$us$i$us>>0]|0;
                       $627 = $626&255;
                       $628 = $627 << 8;
                       $629 = ((($p$sroa$0$9$v$us$i$us)) + 1|0);
                       $630 = HEAP8[$629>>0]|0;
                       $631 = $630&255;
                       $632 = $628 | $631;
                       $637 = $632;$p$sroa$0$10$in$us$i$us = $625;
                      }
                      $636 = (((($532) + ($n3$051$us$i$us<<4)|0)) + 8|0);
                      HEAP32[$636>>2] = $637;
                      if ($545) {
                       $651 = ((($p$sroa$0$10$in$us$i$us)) + 2|0);
                       $652 = HEAP8[$p$sroa$0$10$in$us$i$us>>0]|0;
                       $653 = $652&255;
                       $654 = $653 << 8;
                       $655 = ((($p$sroa$0$10$in$us$i$us)) + 1|0);
                       $656 = HEAP8[$655>>0]|0;
                       $657 = $656&255;
                       $658 = $654 | $657;
                       $660 = $658;$p$sroa$0$11$in$us$i$us = $651;
                      } else {
                       $638 = ((($p$sroa$0$10$in$us$i$us)) + 3|0);
                       $639 = HEAP8[$p$sroa$0$10$in$us$i$us>>0]|0;
                       $640 = $639&255;
                       $641 = $640 << 16;
                       $642 = ((($p$sroa$0$10$in$us$i$us)) + 1|0);
                       $643 = HEAP8[$642>>0]|0;
                       $644 = $643&255;
                       $645 = $644 << 8;
                       $646 = $645 | $641;
                       $647 = ((($p$sroa$0$10$in$us$i$us)) + 2|0);
                       $648 = HEAP8[$647>>0]|0;
                       $649 = $648&255;
                       $650 = $646 | $649;
                       $660 = $650;$p$sroa$0$11$in$us$i$us = $638;
                      }
                      $p$sroa$0$11$us$i$us = $p$sroa$0$11$in$us$i$us;
                      $659 = (((($532) + ($n3$051$us$i$us<<4)|0)) + 12|0);
                      HEAP32[$659>>2] = $660;
                      $661 = (($n3$051$us$i$us) + 1)|0;
                      $exitcond$i$us = ($661|0)==($526|0);
                      if ($exitcond$i$us) {
                       break L81;
                      } else {
                       $n3$051$us$i$us = $661;$p$sroa$0$650$us$i$us = $p$sroa$0$11$us$i$us;
                      }
                     }
                    }
                    $$pre49 = HEAP32[$550>>2]|0;
                    $n3$051$us$i = 0;$p$sroa$0$650$us$i = $519;
                    while(1) {
                     $662 = $p$sroa$0$650$us$i;
                     $663 = ((($662)) + 1|0);
                     $664 = HEAP8[$662>>0]|0;
                     $665 = $664&255;
                     $666 = (($532) + ($n3$051$us$i<<4)|0);
                     HEAP32[$666>>2] = $665;
                     $667 = (((($532) + ($n3$051$us$i<<4)|0)) + 4|0);
                     HEAP32[$667>>2] = $$pre49;
                     $668 = ((($662)) + 2|0);
                     $p$sroa$0$9$v$us$i = $541 ? $668 : $663;
                     if ($543) {
                      $677 = ((($p$sroa$0$9$v$us$i)) + 1|0);
                      $678 = HEAP8[$p$sroa$0$9$v$us$i>>0]|0;
                      $679 = $678&255;
                      $681 = $679;$p$sroa$0$10$in$us$i = $677;
                     } else {
                      $669 = ((($p$sroa$0$9$v$us$i)) + 2|0);
                      $670 = HEAP8[$p$sroa$0$9$v$us$i>>0]|0;
                      $671 = $670&255;
                      $672 = $671 << 8;
                      $673 = ((($p$sroa$0$9$v$us$i)) + 1|0);
                      $674 = HEAP8[$673>>0]|0;
                      $675 = $674&255;
                      $676 = $672 | $675;
                      $681 = $676;$p$sroa$0$10$in$us$i = $669;
                     }
                     $680 = (((($532) + ($n3$051$us$i<<4)|0)) + 8|0);
                     HEAP32[$680>>2] = $681;
                     if ($545) {
                      $695 = ((($p$sroa$0$10$in$us$i)) + 2|0);
                      $696 = HEAP8[$p$sroa$0$10$in$us$i>>0]|0;
                      $697 = $696&255;
                      $698 = $697 << 8;
                      $699 = ((($p$sroa$0$10$in$us$i)) + 1|0);
                      $700 = HEAP8[$699>>0]|0;
                      $701 = $700&255;
                      $702 = $698 | $701;
                      $704 = $702;$p$sroa$0$11$in$us$i = $695;
                     } else {
                      $682 = ((($p$sroa$0$10$in$us$i)) + 3|0);
                      $683 = HEAP8[$p$sroa$0$10$in$us$i>>0]|0;
                      $684 = $683&255;
                      $685 = $684 << 16;
                      $686 = ((($p$sroa$0$10$in$us$i)) + 1|0);
                      $687 = HEAP8[$686>>0]|0;
                      $688 = $687&255;
                      $689 = $688 << 8;
                      $690 = $689 | $685;
                      $691 = ((($p$sroa$0$10$in$us$i)) + 2|0);
                      $692 = HEAP8[$691>>0]|0;
                      $693 = $692&255;
                      $694 = $690 | $693;
                      $704 = $694;$p$sroa$0$11$in$us$i = $682;
                     }
                     $p$sroa$0$11$us$i = $p$sroa$0$11$in$us$i;
                     $703 = (((($532) + ($n3$051$us$i<<4)|0)) + 12|0);
                     HEAP32[$703>>2] = $704;
                     $705 = (($n3$051$us$i) + 1)|0;
                     $exitcond$i = ($705|0)==($526|0);
                     if ($exitcond$i) {
                      break;
                     } else {
                      $n3$051$us$i = $705;$p$sroa$0$650$us$i = $p$sroa$0$11$us$i;
                     }
                    }
                   } else {
                    if ($321) {
                     $n3$051$i$us = 0;$p$sroa$0$650$i$us = $519;
                     while(1) {
                      $551 = $p$sroa$0$650$i$us;
                      $552 = ((($551)) + 2|0);
                      $553 = HEAP8[$551>>0]|0;
                      $554 = $553&255;
                      $555 = $554 << 8;
                      $556 = ((($551)) + 1|0);
                      $557 = HEAP8[$556>>0]|0;
                      $558 = $557&255;
                      $559 = $555 | $558;
                      $560 = (($532) + ($n3$051$i$us<<4)|0);
                      HEAP32[$560>>2] = $559;
                      $561 = ((($551)) + 4|0);
                      $562 = HEAP8[$552>>0]|0;
                      $563 = $562&255;
                      $564 = $563 << 8;
                      $565 = ((($551)) + 3|0);
                      $566 = HEAP8[$565>>0]|0;
                      $567 = $566&255;
                      $568 = $564 | $567;
                      $sext6$i$us = $568 << 16;
                      $569 = $sext6$i$us >> 16;
                      $570 = (((($532) + ($n3$051$i$us<<4)|0)) + 4|0);
                      HEAP32[$570>>2] = $569;
                      $571 = ((($551)) + 5|0);
                      $p$sroa$0$9$v$i$us = $541 ? $571 : $561;
                      if ($543) {
                       $580 = ((($p$sroa$0$9$v$i$us)) + 1|0);
                       $581 = HEAP8[$p$sroa$0$9$v$i$us>>0]|0;
                       $582 = $581&255;
                       $584 = $582;$p$sroa$0$10$in$i$us = $580;
                      } else {
                       $572 = ((($p$sroa$0$9$v$i$us)) + 2|0);
                       $573 = HEAP8[$p$sroa$0$9$v$i$us>>0]|0;
                       $574 = $573&255;
                       $575 = $574 << 8;
                       $576 = ((($p$sroa$0$9$v$i$us)) + 1|0);
                       $577 = HEAP8[$576>>0]|0;
                       $578 = $577&255;
                       $579 = $575 | $578;
                       $584 = $579;$p$sroa$0$10$in$i$us = $572;
                      }
                      $583 = (((($532) + ($n3$051$i$us<<4)|0)) + 8|0);
                      HEAP32[$583>>2] = $584;
                      if ($545) {
                       $598 = ((($p$sroa$0$10$in$i$us)) + 2|0);
                       $599 = HEAP8[$p$sroa$0$10$in$i$us>>0]|0;
                       $600 = $599&255;
                       $601 = $600 << 8;
                       $602 = ((($p$sroa$0$10$in$i$us)) + 1|0);
                       $603 = HEAP8[$602>>0]|0;
                       $604 = $603&255;
                       $605 = $601 | $604;
                       $607 = $605;$p$sroa$0$11$in$i$us = $598;
                      } else {
                       $585 = ((($p$sroa$0$10$in$i$us)) + 3|0);
                       $586 = HEAP8[$p$sroa$0$10$in$i$us>>0]|0;
                       $587 = $586&255;
                       $588 = $587 << 16;
                       $589 = ((($p$sroa$0$10$in$i$us)) + 1|0);
                       $590 = HEAP8[$589>>0]|0;
                       $591 = $590&255;
                       $592 = $591 << 8;
                       $593 = $592 | $588;
                       $594 = ((($p$sroa$0$10$in$i$us)) + 2|0);
                       $595 = HEAP8[$594>>0]|0;
                       $596 = $595&255;
                       $597 = $593 | $596;
                       $607 = $597;$p$sroa$0$11$in$i$us = $585;
                      }
                      $p$sroa$0$11$i$us = $p$sroa$0$11$in$i$us;
                      $606 = (((($532) + ($n3$051$i$us<<4)|0)) + 12|0);
                      HEAP32[$606>>2] = $607;
                      $608 = (($n3$051$i$us) + 1)|0;
                      $exitcond69$i$us = ($608|0)==($526|0);
                      if ($exitcond69$i$us) {
                       break L81;
                      } else {
                       $n3$051$i$us = $608;$p$sroa$0$650$i$us = $p$sroa$0$11$i$us;
                      }
                     }
                    }
                    $$pre = HEAP32[$550>>2]|0;
                    $n3$051$i = 0;$p$sroa$0$650$i = $519;
                    while(1) {
                     $706 = $p$sroa$0$650$i;
                     $707 = ((($706)) + 2|0);
                     $708 = HEAP8[$706>>0]|0;
                     $709 = $708&255;
                     $710 = $709 << 8;
                     $711 = ((($706)) + 1|0);
                     $712 = HEAP8[$711>>0]|0;
                     $713 = $712&255;
                     $714 = $710 | $713;
                     $715 = (($532) + ($n3$051$i<<4)|0);
                     HEAP32[$715>>2] = $714;
                     $716 = (((($532) + ($n3$051$i<<4)|0)) + 4|0);
                     HEAP32[$716>>2] = $$pre;
                     $717 = ((($706)) + 3|0);
                     $p$sroa$0$9$v$i = $541 ? $717 : $707;
                     if ($543) {
                      $726 = ((($p$sroa$0$9$v$i)) + 1|0);
                      $727 = HEAP8[$p$sroa$0$9$v$i>>0]|0;
                      $728 = $727&255;
                      $730 = $728;$p$sroa$0$10$in$i = $726;
                     } else {
                      $718 = ((($p$sroa$0$9$v$i)) + 2|0);
                      $719 = HEAP8[$p$sroa$0$9$v$i>>0]|0;
                      $720 = $719&255;
                      $721 = $720 << 8;
                      $722 = ((($p$sroa$0$9$v$i)) + 1|0);
                      $723 = HEAP8[$722>>0]|0;
                      $724 = $723&255;
                      $725 = $721 | $724;
                      $730 = $725;$p$sroa$0$10$in$i = $718;
                     }
                     $729 = (((($532) + ($n3$051$i<<4)|0)) + 8|0);
                     HEAP32[$729>>2] = $730;
                     if ($545) {
                      $744 = ((($p$sroa$0$10$in$i)) + 2|0);
                      $745 = HEAP8[$p$sroa$0$10$in$i>>0]|0;
                      $746 = $745&255;
                      $747 = $746 << 8;
                      $748 = ((($p$sroa$0$10$in$i)) + 1|0);
                      $749 = HEAP8[$748>>0]|0;
                      $750 = $749&255;
                      $751 = $747 | $750;
                      $753 = $751;$p$sroa$0$11$in$i = $744;
                     } else {
                      $731 = ((($p$sroa$0$10$in$i)) + 3|0);
                      $732 = HEAP8[$p$sroa$0$10$in$i>>0]|0;
                      $733 = $732&255;
                      $734 = $733 << 16;
                      $735 = ((($p$sroa$0$10$in$i)) + 1|0);
                      $736 = HEAP8[$735>>0]|0;
                      $737 = $736&255;
                      $738 = $737 << 8;
                      $739 = $738 | $734;
                      $740 = ((($p$sroa$0$10$in$i)) + 2|0);
                      $741 = HEAP8[$740>>0]|0;
                      $742 = $741&255;
                      $743 = $739 | $742;
                      $753 = $743;$p$sroa$0$11$in$i = $731;
                     }
                     $p$sroa$0$11$i = $p$sroa$0$11$in$i;
                     $752 = (((($532) + ($n3$051$i<<4)|0)) + 12|0);
                     HEAP32[$752>>2] = $753;
                     $754 = (($n3$051$i) + 1)|0;
                     $exitcond69$i = ($754|0)==($526|0);
                     if ($exitcond69$i) {
                      break;
                     } else {
                      $n3$051$i = $754;$p$sroa$0$650$i = $p$sroa$0$11$i;
                     }
                    }
                   }
                  }
                 } while(0);
                 if ((label|0) == 147) {
                  HEAP32[$error$i5>>2] = 8;
                 }
                 _FT_Stream_ExitFrame($stream);
                 $757 = (_FT_Stream_Pos($stream)|0);
                 $758 = ((($230)) + 144|0);
                 HEAP32[$758>>2] = $757;
                 HEAP32[$245>>2] = 0;
                 $$pre$i8 = HEAP32[$error$i5>>2]|0;
                 HEAP32[$error>>2] = $$pre$i8;
                 $759 = ($$pre$i8|0)==(0);
                 if (!($759)) {
                  $890 = $$pre$i8;
                  STACKTOP = sp;return ($890|0);
                 }
                 $760 = ((($pfrface)) + 4|0);
                 HEAP32[$760>>2] = $face_index;
                 $761 = ((($230)) + 120|0);
                 $762 = HEAP32[$761>>2]|0;
                 $763 = (($762) + 1)|0;
                 $764 = ((($pfrface)) + 16|0);
                 HEAP32[$764>>2] = $763;
                 $765 = ((($pfrface)) + 8|0);
                 $766 = HEAP32[$765>>2]|0;
                 $767 = $766 | 1;
                 HEAP32[$765>>2] = $767;
                 $768 = ($762|0)==(0);
                 L212: do {
                  if ($768) {
                   $nn$0$lcssa = 0;
                  } else {
                   $769 = ((($230)) + 128|0);
                   $770 = HEAP32[$769>>2]|0;
                   $nn$032 = 0;
                   while(1) {
                    $771 = (((($770) + ($nn$032<<4)|0)) + 12|0);
                    $772 = HEAP32[$771>>2]|0;
                    $773 = ($772|0)==(0);
                    if (!($773)) {
                     $nn$0$lcssa = $nn$032;
                     break L212;
                    }
                    $774 = (($nn$032) + 1)|0;
                    $775 = ($774>>>0)<($762>>>0);
                    if ($775) {
                     $nn$032 = $774;
                    } else {
                     $nn$0$lcssa = $774;
                     break;
                    }
                   }
                  }
                 } while(0);
                 $776 = ($nn$0$lcssa|0)==($762|0);
                 do {
                  if ($776) {
                   $777 = ((($230)) + 92|0);
                   $778 = HEAP32[$777>>2]|0;
                   $779 = ($778|0)==(0);
                   if (!($779)) {
                    HEAP32[$765>>2] = 0;
                    $785 = 0;
                    break;
                   }
                   HEAP32[$error>>2] = 3;
                   $890 = 3;
                   STACKTOP = sp;return ($890|0);
                  } else {
                   $785 = $767;
                  }
                 } while(0);
                 $780 = ((($230)) + 36|0);
                 $781 = HEAP32[$780>>2]|0;
                 $782 = $781 & 4;
                 $783 = ($782|0)==(0);
                 if ($783) {
                  $784 = $785 | 4;
                  HEAP32[$765>>2] = $784;
                  $789 = $784;
                 } else {
                  $789 = $785;
                 }
                 $786 = $781 << 4;
                 $787 = $786 & 16;
                 $788 = (($787) + 16)|0;
                 $storemerge = $788 | $789;
                 HEAP32[$765>>2] = $storemerge;
                 $790 = ((($230)) + 92|0);
                 $791 = HEAP32[$790>>2]|0;
                 $792 = ($791|0)==(0);
                 if ($792) {
                  $798 = $storemerge;
                 } else {
                  $793 = $storemerge | 2;
                  HEAP32[$765>>2] = $793;
                  $798 = $793;
                 }
                 $794 = ((($230)) + 132|0);
                 $795 = HEAP32[$794>>2]|0;
                 $796 = ($795|0)==(0);
                 if (!($796)) {
                  $797 = $798 | 64;
                  HEAP32[$765>>2] = $797;
                 }
                 $799 = ((($230)) + 84|0);
                 $800 = HEAP32[$799>>2]|0;
                 $801 = ((($pfrface)) + 20|0);
                 HEAP32[$801>>2] = $800;
                 $802 = ($800|0)==(0|0);
                 if ($802) {
                  $803 = ((($230)) + 80|0);
                  $804 = HEAP32[$803>>2]|0;
                  HEAP32[$801>>2] = $804;
                 }
                 $805 = ((($230)) + 88|0);
                 $806 = HEAP32[$805>>2]|0;
                 $807 = ((($pfrface)) + 24|0);
                 HEAP32[$807>>2] = $806;
                 $808 = ((($pfrface)) + 28|0);
                 HEAP32[$808>>2] = 0;
                 $809 = ((($pfrface)) + 32|0);
                 HEAP32[$809>>2] = 0;
                 $810 = ((($pfrface)) + 52|0);
                 $811 = ((($230)) + 20|0);
                 ;HEAP32[$810>>2]=HEAP32[$811>>2]|0;HEAP32[$810+4>>2]=HEAP32[$811+4>>2]|0;HEAP32[$810+8>>2]=HEAP32[$811+8>>2]|0;HEAP32[$810+12>>2]=HEAP32[$811+12>>2]|0;
                 $812 = ((($230)) + 12|0);
                 $813 = HEAP32[$812>>2]|0;
                 $814 = $813&65535;
                 $815 = ((($pfrface)) + 68|0);
                 HEAP16[$815>>1] = $814;
                 $816 = ((($230)) + 32|0);
                 $817 = HEAP32[$816>>2]|0;
                 $818 = $817&65535;
                 $819 = ((($pfrface)) + 70|0);
                 HEAP16[$819>>1] = $818;
                 $820 = ((($230)) + 24|0);
                 $821 = HEAP32[$820>>2]|0;
                 $822 = $821&65535;
                 $823 = ((($pfrface)) + 72|0);
                 HEAP16[$823>>1] = $822;
                 $824 = $813 & 65535;
                 $825 = ($824*12)|0;
                 $826 = (($825>>>0) / 10)&-1;
                 $827 = ((($pfrface)) + 74|0);
                 $sext = $826 << 16;
                 $828 = $sext >> 16;
                 $sext51 = $817 << 16;
                 $829 = $sext51 >> 16;
                 $sext52 = $821 << 16;
                 $830 = $sext52 >> 16;
                 $831 = (($829) - ($830))|0;
                 $832 = ($828|0)<($831|0);
                 $$ = $832 ? $831 : $826;
                 $storemerge53 = $$&65535;
                 HEAP16[$827>>1] = $storemerge53;
                 if ($792) {
                  $857 = $781;
                 } else {
                  $833 = ((($pfrface)) + 104|0);
                  $834 = HEAP32[$833>>2]|0;
                  $835 = ((($834)) + 28|0);
                  $836 = HEAP32[$835>>2]|0;
                  $837 = (_ft_mem_realloc($836,16,0,$791,0,$error)|0);
                  HEAP32[$809>>2] = $837;
                  $838 = HEAP32[$error>>2]|0;
                  $839 = ($838|0)==(0);
                  if (!($839)) {
                   $890 = $838;
                   STACKTOP = sp;return ($890|0);
                  }
                  $840 = ((($230)) + 100|0);
                  $841 = HEAP32[$840>>2]|0;
                  $n$027 = 0;$size$028 = $837;$strike$029 = $841;
                  while(1) {
                   $842 = ((($strike$029)) + 4|0);
                   $843 = HEAP32[$842>>2]|0;
                   $844 = $843&65535;
                   HEAP16[$size$028>>1] = $844;
                   $845 = HEAP32[$strike$029>>2]|0;
                   $846 = $845&65535;
                   $847 = ((($size$028)) + 2|0);
                   HEAP16[$847>>1] = $846;
                   $848 = $843 << 6;
                   $849 = ((($size$028)) + 4|0);
                   HEAP32[$849>>2] = $848;
                   $850 = $845 << 6;
                   $851 = ((($size$028)) + 8|0);
                   HEAP32[$851>>2] = $850;
                   $852 = ((($size$028)) + 12|0);
                   HEAP32[$852>>2] = $848;
                   $853 = (($n$027) + 1)|0;
                   $854 = ((($size$028)) + 16|0);
                   $855 = ((($strike$029)) + 36|0);
                   $exitcond = ($853|0)==($791|0);
                   if ($exitcond) {
                    break;
                   } else {
                    $n$027 = $853;$size$028 = $854;$strike$029 = $855;
                   }
                  }
                  HEAP32[$808>>2] = $791;
                  $$pre50 = HEAP32[$780>>2]|0;
                  $857 = $$pre50;
                 }
                 $856 = $857 & 4;
                 $858 = ($856|0)==(0);
                 if ($858) {
                  $859 = ((($230)) + 40|0);
                  $860 = HEAP32[$859>>2]|0;
                  $861 = $860&65535;
                  $862 = ((($pfrface)) + 76|0);
                  HEAP16[$862>>1] = $861;
                 } else {
                  $863 = HEAP32[$761>>2]|0;
                  $864 = ($863|0)==(0);
                  if ($864) {
                   $max$0$lcssa = 0;
                  } else {
                   $865 = ((($230)) + 128|0);
                   $866 = HEAP32[$865>>2]|0;
                   $count3$025 = $863;$gchar$026 = $866;$max$024 = 0;
                   while(1) {
                    $867 = ((($gchar$026)) + 4|0);
                    $868 = HEAP32[$867>>2]|0;
                    $869 = ($max$024|0)<($868|0);
                    $$max$0 = $869 ? $868 : $max$024;
                    $870 = (($count3$025) + -1)|0;
                    $871 = ((($gchar$026)) + 16|0);
                    $872 = ($870|0)==(0);
                    if ($872) {
                     $$max$0$lcssa = $$max$0;
                     break;
                    } else {
                     $count3$025 = $870;$gchar$026 = $871;$max$024 = $$max$0;
                    }
                   }
                   $phitmp = $$max$0$lcssa&65535;
                   $max$0$lcssa = $phitmp;
                  }
                  $873 = ((($pfrface)) + 76|0);
                  HEAP16[$873>>1] = $max$0$lcssa;
                 }
                 $874 = HEAP16[$827>>1]|0;
                 $875 = ((($pfrface)) + 78|0);
                 HEAP16[$875>>1] = $874;
                 $876 = HEAP16[$815>>1]|0;
                 $877 = $876&65535;
                 $878 = (($877|0) / -10)&-1;
                 $879 = $878&65535;
                 $880 = ((($pfrface)) + 80|0);
                 HEAP16[$880>>1] = $879;
                 $div = (($876&65535) / 30)&-1;
                 $881 = ((($pfrface)) + 82|0);
                 HEAP16[$881>>1] = $div;
                 HEAP32[$charmap>>2] = $pfrface;
                 $882 = ((($charmap)) + 8|0);
                 HEAP16[$882>>1] = 3;
                 $883 = ((($charmap)) + 10|0);
                 HEAP16[$883>>1] = 1;
                 $884 = ((($charmap)) + 4|0);
                 HEAP32[$884>>2] = 1970170211;
                 $885 = (_FT_CMap_New(53336,0,$charmap,0)|0);
                 HEAP32[$error>>2] = $885;
                 $886 = HEAP32[$794>>2]|0;
                 $887 = ($886|0)==(0);
                 if ($887) {
                  $890 = $885;
                  STACKTOP = sp;return ($890|0);
                 }
                 $888 = HEAP32[$765>>2]|0;
                 $889 = $888 | 64;
                 HEAP32[$765>>2] = $889;
                 $890 = $885;
                 STACKTOP = sp;return ($890|0);
                } else {
                 $$ph = $240;
                }
               } while(0);
               HEAP32[$error>>2] = $$ph;
               $890 = $$ph;
               STACKTOP = sp;return ($890|0);
              } else {
               $$0$i$ph = $54;
              }
             } else {
              $$0$i$ph = $51;
             }
            } else {
             $$0$i$ph = $48;
            }
           } else {
            $$0$i$ph = $44;
           }
          } else {
           $$0$i$ph = 6;
          }
         } else {
          $$0$i$ph = $39;
         }
        } else {
         $$0$i$ph = $36;
        }
       } while(0);
       HEAP32[$error>>2] = $$0$i$ph;
       $890 = $$0$i$ph;
       STACKTOP = sp;return ($890|0);
      }
     }
    }
   }
   HEAP32[$error>>2] = 2;
   $890 = 2;
   STACKTOP = sp;return ($890|0);
  } else {
   $error$0$i$ph = $3;
  }
 } else {
  $error$0$i$ph = $1;
 }
 HEAP32[$error>>2] = $error$0$i$ph;
 $890 = $error$0$i$ph;
 STACKTOP = sp;return ($890|0);
}
function _pfr_extra_item_load_bitmap_info($p,$limit,$phy_font) {
 $p = $p|0;
 $limit = $limit|0;
 $phy_font = $phy_font|0;
 var $$028 = 0, $$028$us = 0, $$2 = 0, $$2$us = 0, $$3 = 0, $$3$us = 0, $$4 = 0, $$4$us = 0, $$5 = 0, $$5$us = 0, $$lobit = 0, $$lobit3 = 0, $$lobit4 = 0, $$lobit5 = 0, $$pre = 0, $$size1$1 = 0, $$size1$3 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0;
 var $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0;
 var $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0;
 var $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0;
 var $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0;
 var $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $error = 0, $exitcond = 0, $exitcond10 = 0, $n$07 = 0, $n$07$us = 0, $size1$1 = 0, $size1$3 = 0, $strike$06 = 0, $strike$06$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = HEAP32[$phy_font>>2]|0;
 HEAP32[$error>>2] = 0;
 $1 = ((($p)) + 5|0);
 $2 = ($1>>>0)>($limit>>>0);
 if (!($2)) {
  $3 = ((($p)) + 3|0);
  $4 = ((($p)) + 4|0);
  $5 = HEAP8[$3>>0]|0;
  $6 = $5&255;
  $7 = HEAP8[$4>>0]|0;
  $8 = $7&255;
  $9 = ((($phy_font)) + 92|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = (($10) + ($8))|0;
  $12 = ((($phy_font)) + 96|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ($11>>>0)>($13>>>0);
  do {
   if ($14) {
    $15 = (($11) + 3)|0;
    $16 = $15 & -4;
    $17 = ((($phy_font)) + 100|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = (_ft_mem_realloc($0,36,$10,$16,$18,$error)|0);
    HEAP32[$17>>2] = $19;
    $20 = HEAP32[$error>>2]|0;
    $21 = ($20|0)==(0);
    if ($21) {
     HEAP32[$12>>2] = $16;
     break;
    } else {
     $203 = $20;
     STACKTOP = sp;return ($203|0);
    }
   }
  } while(0);
  $22 = $6 & 1;
  $23 = ($22|0)==(0);
  $24 = $22 | 8;
  $25 = $6 & 2;
  $26 = ($25|0)==(0);
  $$lobit = $25 >>> 1;
  $size1$1 = (($24) + ($$lobit))|0;
  $27 = $6 & 4;
  $28 = ($27|0)==(0);
  $$lobit3 = $27 >>> 2;
  $$size1$1 = (($size1$1) + ($$lobit3))|0;
  $29 = $6 & 8;
  $30 = ($29|0)==(0);
  $$lobit4 = $29 >>> 3;
  $size1$3 = (($$size1$1) + ($$lobit4))|0;
  $31 = $6 & 16;
  $32 = ($31|0)==(0);
  $$lobit5 = $31 >>> 4;
  $$size1$3 = (($size1$3) + ($$lobit5))|0;
  $33 = Math_imul($$size1$3, $8)|0;
  $34 = (($1) + ($33)|0);
  $35 = ($34>>>0)>($limit>>>0);
  if (!($35)) {
   $36 = ((($phy_font)) + 100|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = HEAP32[$9>>2]|0;
   $39 = (($37) + (($38*36)|0)|0);
   $40 = ($7<<24>>24)==(0);
   if (!($40)) {
    if ($23) {
     $$028$us = $1;$n$07$us = 0;$strike$06$us = $39;
     while(1) {
      $41 = ((($$028$us)) + 1|0);
      $42 = HEAP8[$$028$us>>0]|0;
      $43 = $42&255;
      HEAP32[$strike$06$us>>2] = $43;
      if ($26) {
       $52 = ((($$028$us)) + 2|0);
       $53 = HEAP8[$41>>0]|0;
       $54 = $53&255;
       $$2$us = $52;$56 = $54;
      } else {
       $44 = ((($$028$us)) + 3|0);
       $45 = HEAP8[$41>>0]|0;
       $46 = $45&255;
       $47 = $46 << 8;
       $48 = ((($$028$us)) + 2|0);
       $49 = HEAP8[$48>>0]|0;
       $50 = $49&255;
       $51 = $47 | $50;
       $$2$us = $44;$56 = $51;
      }
      $55 = ((($strike$06$us)) + 4|0);
      HEAP32[$55>>2] = $56;
      $57 = HEAP8[$$2$us>>0]|0;
      $58 = $57&255;
      $59 = ((($strike$06$us)) + 8|0);
      HEAP32[$59>>2] = $58;
      $60 = ((($$2$us)) + 1|0);
      $61 = HEAP8[$60>>0]|0;
      $62 = ((($$2$us)) + 2|0);
      $63 = HEAP8[$62>>0]|0;
      if ($28) {
       $74 = ((($$2$us)) + 3|0);
       $75 = $61&255;
       $76 = $75 << 8;
       $77 = $63&255;
       $78 = $76 | $77;
       $$3$us = $74;$80 = $78;
      } else {
       $64 = ((($$2$us)) + 4|0);
       $65 = $61&255;
       $66 = $65 << 16;
       $67 = $63&255;
       $68 = $67 << 8;
       $69 = $68 | $66;
       $70 = ((($$2$us)) + 3|0);
       $71 = HEAP8[$70>>0]|0;
       $72 = $71&255;
       $73 = $69 | $72;
       $$3$us = $64;$80 = $73;
      }
      $79 = ((($strike$06$us)) + 20|0);
      HEAP32[$79>>2] = $80;
      if ($30) {
       $94 = ((($$3$us)) + 2|0);
       $95 = HEAP8[$$3$us>>0]|0;
       $96 = $95&255;
       $97 = $96 << 8;
       $98 = ((($$3$us)) + 1|0);
       $99 = HEAP8[$98>>0]|0;
       $100 = $99&255;
       $101 = $97 | $100;
       $$4$us = $94;$103 = $101;
      } else {
       $81 = ((($$3$us)) + 3|0);
       $82 = HEAP8[$$3$us>>0]|0;
       $83 = $82&255;
       $84 = $83 << 16;
       $85 = ((($$3$us)) + 1|0);
       $86 = HEAP8[$85>>0]|0;
       $87 = $86&255;
       $88 = $87 << 8;
       $89 = $88 | $84;
       $90 = ((($$3$us)) + 2|0);
       $91 = HEAP8[$90>>0]|0;
       $92 = $91&255;
       $93 = $89 | $92;
       $$4$us = $81;$103 = $93;
      }
      $102 = ((($strike$06$us)) + 24|0);
      HEAP32[$102>>2] = $103;
      if ($32) {
       $112 = ((($$4$us)) + 1|0);
       $113 = HEAP8[$$4$us>>0]|0;
       $114 = $113&255;
       $$5$us = $112;$116 = $114;
      } else {
       $104 = ((($$4$us)) + 2|0);
       $105 = HEAP8[$$4$us>>0]|0;
       $106 = $105&255;
       $107 = $106 << 8;
       $108 = ((($$4$us)) + 1|0);
       $109 = HEAP8[$108>>0]|0;
       $110 = $109&255;
       $111 = $107 | $110;
       $$5$us = $104;$116 = $111;
      }
      $115 = ((($strike$06$us)) + 28|0);
      HEAP32[$115>>2] = $116;
      $117 = (($n$07$us) + 1)|0;
      $118 = ((($strike$06$us)) + 36|0);
      $exitcond = ($117|0)==($8|0);
      if ($exitcond) {
       break;
      } else {
       $$028$us = $$5$us;$n$07$us = $117;$strike$06$us = $118;
      }
     }
    } else {
     $$028 = $1;$n$07 = 0;$strike$06 = $39;
     while(1) {
      $119 = ((($$028)) + 2|0);
      $120 = HEAP8[$$028>>0]|0;
      $121 = $120&255;
      $122 = $121 << 8;
      $123 = ((($$028)) + 1|0);
      $124 = HEAP8[$123>>0]|0;
      $125 = $124&255;
      $126 = $122 | $125;
      HEAP32[$strike$06>>2] = $126;
      if ($26) {
       $135 = ((($$028)) + 3|0);
       $136 = HEAP8[$119>>0]|0;
       $137 = $136&255;
       $$2 = $135;$139 = $137;
      } else {
       $127 = ((($$028)) + 4|0);
       $128 = HEAP8[$119>>0]|0;
       $129 = $128&255;
       $130 = $129 << 8;
       $131 = ((($$028)) + 3|0);
       $132 = HEAP8[$131>>0]|0;
       $133 = $132&255;
       $134 = $130 | $133;
       $$2 = $127;$139 = $134;
      }
      $138 = ((($strike$06)) + 4|0);
      HEAP32[$138>>2] = $139;
      $140 = HEAP8[$$2>>0]|0;
      $141 = $140&255;
      $142 = ((($strike$06)) + 8|0);
      HEAP32[$142>>2] = $141;
      $143 = ((($$2)) + 1|0);
      $144 = HEAP8[$143>>0]|0;
      $145 = ((($$2)) + 2|0);
      $146 = HEAP8[$145>>0]|0;
      if ($28) {
       $157 = ((($$2)) + 3|0);
       $158 = $144&255;
       $159 = $158 << 8;
       $160 = $146&255;
       $161 = $159 | $160;
       $$3 = $157;$163 = $161;
      } else {
       $147 = ((($$2)) + 4|0);
       $148 = $144&255;
       $149 = $148 << 16;
       $150 = $146&255;
       $151 = $150 << 8;
       $152 = $151 | $149;
       $153 = ((($$2)) + 3|0);
       $154 = HEAP8[$153>>0]|0;
       $155 = $154&255;
       $156 = $152 | $155;
       $$3 = $147;$163 = $156;
      }
      $162 = ((($strike$06)) + 20|0);
      HEAP32[$162>>2] = $163;
      if ($30) {
       $177 = ((($$3)) + 2|0);
       $178 = HEAP8[$$3>>0]|0;
       $179 = $178&255;
       $180 = $179 << 8;
       $181 = ((($$3)) + 1|0);
       $182 = HEAP8[$181>>0]|0;
       $183 = $182&255;
       $184 = $180 | $183;
       $$4 = $177;$186 = $184;
      } else {
       $164 = ((($$3)) + 3|0);
       $165 = HEAP8[$$3>>0]|0;
       $166 = $165&255;
       $167 = $166 << 16;
       $168 = ((($$3)) + 1|0);
       $169 = HEAP8[$168>>0]|0;
       $170 = $169&255;
       $171 = $170 << 8;
       $172 = $171 | $167;
       $173 = ((($$3)) + 2|0);
       $174 = HEAP8[$173>>0]|0;
       $175 = $174&255;
       $176 = $172 | $175;
       $$4 = $164;$186 = $176;
      }
      $185 = ((($strike$06)) + 24|0);
      HEAP32[$185>>2] = $186;
      if ($32) {
       $195 = ((($$4)) + 1|0);
       $196 = HEAP8[$$4>>0]|0;
       $197 = $196&255;
       $$5 = $195;$199 = $197;
      } else {
       $187 = ((($$4)) + 2|0);
       $188 = HEAP8[$$4>>0]|0;
       $189 = $188&255;
       $190 = $189 << 8;
       $191 = ((($$4)) + 1|0);
       $192 = HEAP8[$191>>0]|0;
       $193 = $192&255;
       $194 = $190 | $193;
       $$5 = $187;$199 = $194;
      }
      $198 = ((($strike$06)) + 28|0);
      HEAP32[$198>>2] = $199;
      $200 = (($n$07) + 1)|0;
      $201 = ((($strike$06)) + 36|0);
      $exitcond10 = ($200|0)==($8|0);
      if ($exitcond10) {
       break;
      } else {
       $$028 = $$5;$n$07 = $200;$strike$06 = $201;
      }
     }
    }
   }
   $202 = (($38) + ($8))|0;
   HEAP32[$9>>2] = $202;
   $$pre = HEAP32[$error>>2]|0;
   $203 = $$pre;
   STACKTOP = sp;return ($203|0);
  }
 }
 HEAP32[$error>>2] = 8;
 $203 = 8;
 STACKTOP = sp;return ($203|0);
}
function _pfr_extra_item_load_font_id($p,$limit,$phy_font) {
 $p = $p|0;
 $limit = $limit|0;
 $phy_font = $phy_font|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 HEAP32[$error>>2] = 0;
 $0 = $limit;
 $1 = $p;
 $2 = (($0) - ($1))|0;
 $3 = ((($phy_font)) + 80|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if (!($5)) {
  $13 = 0;
  STACKTOP = sp;return ($13|0);
 }
 $6 = HEAP32[$phy_font>>2]|0;
 $7 = (($2) + 1)|0;
 $8 = (_ft_mem_alloc($6,$7,$error)|0);
 HEAP32[$3>>2] = $8;
 $9 = HEAP32[$error>>2]|0;
 $10 = ($9|0)==(0);
 if (!($10)) {
  $13 = $9;
  STACKTOP = sp;return ($13|0);
 }
 _memcpy(($8|0),($p|0),($2|0))|0;
 $11 = HEAP32[$3>>2]|0;
 $12 = (($11) + ($2)|0);
 HEAP8[$12>>0] = 0;
 $$pre = HEAP32[$error>>2]|0;
 $13 = $$pre;
 STACKTOP = sp;return ($13|0);
}
function _pfr_extra_item_load_stem_snaps($p,$limit,$phy_font) {
 $p = $p|0;
 $limit = $limit|0;
 $phy_font = $phy_font|0;
 var $$03 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $count$01 = 0, $error = 0, $sext = 0, $snaps$02 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 HEAP32[$error>>2] = 0;
 $0 = HEAP32[$phy_font>>2]|0;
 $1 = ((($phy_font)) + 76|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
  $32 = 0;
  STACKTOP = sp;return ($32|0);
 }
 $4 = ((($p)) + 1|0);
 $5 = ($4>>>0)>($limit>>>0);
 if (!($5)) {
  $6 = HEAP8[$p>>0]|0;
  $7 = $6&255;
  $8 = $7 & 15;
  $9 = $7 >>> 4;
  $10 = (($8) + ($9))|0;
  $11 = $10 << 1;
  $12 = (($4) + ($11)|0);
  $13 = ($12>>>0)>($limit>>>0);
  if (!($13)) {
   $14 = (_ft_mem_realloc($0,4,0,$10,0,$error)|0);
   $15 = HEAP32[$error>>2]|0;
   $16 = ($15|0)==(0);
   if (!($16)) {
    $32 = $15;
    STACKTOP = sp;return ($32|0);
   }
   HEAP32[$1>>2] = $14;
   $17 = (($14) + ($8<<2)|0);
   $18 = ((($phy_font)) + 64|0);
   HEAP32[$18>>2] = $17;
   $19 = ($10|0)==(0);
   if ($19) {
    $32 = 0;
    STACKTOP = sp;return ($32|0);
   } else {
    $$03 = $4;$count$01 = $10;$snaps$02 = $14;
   }
   while(1) {
    $20 = ((($$03)) + 2|0);
    $21 = HEAP8[$$03>>0]|0;
    $22 = $21&255;
    $23 = $22 << 8;
    $24 = ((($$03)) + 1|0);
    $25 = HEAP8[$24>>0]|0;
    $26 = $25&255;
    $27 = $23 | $26;
    $sext = $27 << 16;
    $28 = $sext >> 16;
    HEAP32[$snaps$02>>2] = $28;
    $29 = (($count$01) + -1)|0;
    $30 = ((($snaps$02)) + 4|0);
    $31 = ($29|0)==(0);
    if ($31) {
     break;
    } else {
     $$03 = $20;$count$01 = $29;$snaps$02 = $30;
    }
   }
   $$pre = HEAP32[$error>>2]|0;
   $32 = $$pre;
   STACKTOP = sp;return ($32|0);
  }
 }
 HEAP32[$error>>2] = 8;
 $32 = 8;
 STACKTOP = sp;return ($32|0);
}
function _pfr_extra_item_load_kerning_pairs($p,$limit,$phy_font) {
 $p = $p|0;
 $limit = $limit|0;
 $phy_font = $phy_font|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $error = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 HEAP32[$error>>2] = 0;
 $0 = HEAP32[$phy_font>>2]|0;
 $1 = (_ft_mem_alloc($0,24,$error)|0);
 $2 = HEAP32[$error>>2]|0;
 $3 = ($2|0)==(0);
 if (!($3)) {
  $109 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($109|0);
 }
 $4 = ((($p)) + 4|0);
 $5 = ($4>>>0)>($limit>>>0);
 if (!($5)) {
  $6 = HEAP8[$p>>0]|0;
  $7 = ((($1)) + 4|0);
  HEAP8[$7>>0] = $6;
  $8 = ((($p)) + 3|0);
  $9 = ((($p)) + 1|0);
  $10 = HEAP8[$9>>0]|0;
  $11 = $10&255;
  $12 = $11 << 8;
  $13 = ((($p)) + 2|0);
  $14 = HEAP8[$13>>0]|0;
  $15 = $14&255;
  $16 = $12 | $15;
  $17 = $16&65535;
  $18 = ((($1)) + 6|0);
  HEAP16[$18>>1] = $17;
  $19 = HEAP8[$8>>0]|0;
  $20 = ((($1)) + 5|0);
  HEAP8[$20>>0] = $19;
  $21 = ((($phy_font)) + 4|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = ((($phy_font)) + 148|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = $4;
  $26 = (($22) + ($25))|0;
  $27 = (($26) - ($24))|0;
  $28 = ((($1)) + 12|0);
  HEAP32[$28>>2] = $27;
  $29 = ((($1)) + 8|0);
  $30 = $19 & 1;
  $31 = $30&255;
  $32 = $31 << 1;
  $33 = (($32) + 3)|0;
  $34 = $19 & 2;
  $35 = ($34<<24>>24)==(0);
  $36 = $32 | 4;
  $$ = $35 ? $33 : $36;
  HEAP32[$29>>2] = $$;
  $37 = $6&255;
  $38 = Math_imul($$, $37)|0;
  $39 = (($4) + ($38)|0);
  $40 = ($39>>>0)>($limit>>>0);
  if (!($40)) {
   $41 = ($6<<24>>24)==(0);
   if ($41) {
    _ft_mem_free($0,$1);
    $109 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($109|0);
   }
   $42 = ($30<<24>>24)==(0);
   if ($42) {
    $85 = ((($p)) + 5|0);
    $86 = HEAP8[$4>>0]|0;
    $87 = $86&255;
    $88 = HEAP8[$85>>0]|0;
    $89 = $87 << 16;
    $90 = $88&255;
    $91 = $89 | $90;
    $92 = ((($1)) + 16|0);
    HEAP32[$92>>2] = $91;
    $93 = (($37) + -1)|0;
    $94 = Math_imul($93, $$)|0;
    $95 = (($4) + ($94)|0);
    $96 = ((($95)) + 1|0);
    $97 = HEAP8[$95>>0]|0;
    $98 = $97&255;
    $99 = HEAP8[$96>>0]|0;
    $100 = $98 << 16;
    $101 = $99&255;
    $102 = $100 | $101;
    $103 = ((($1)) + 20|0);
    HEAP32[$103>>2] = $102;
   } else {
    $43 = HEAP8[$4>>0]|0;
    $44 = $43&255;
    $45 = $44 << 8;
    $46 = ((($p)) + 5|0);
    $47 = HEAP8[$46>>0]|0;
    $48 = $47&255;
    $49 = $45 | $48;
    $50 = ((($p)) + 6|0);
    $51 = HEAP8[$50>>0]|0;
    $52 = $51&255;
    $53 = $52 << 8;
    $54 = ((($p)) + 7|0);
    $55 = HEAP8[$54>>0]|0;
    $56 = $55&255;
    $57 = $53 | $56;
    $58 = $49 << 16;
    $59 = $57 | $58;
    $60 = ((($1)) + 16|0);
    HEAP32[$60>>2] = $59;
    $61 = (($37) + -1)|0;
    $62 = Math_imul($61, $$)|0;
    $63 = (($4) + ($62)|0);
    $64 = ((($63)) + 2|0);
    $65 = ((($64)) + -2|0);
    $66 = HEAP8[$65>>0]|0;
    $67 = $66&255;
    $68 = $67 << 8;
    $69 = ((($65)) + 1|0);
    $70 = HEAP8[$69>>0]|0;
    $71 = $70&255;
    $72 = $68 | $71;
    $73 = ((($64)) + 2|0);
    $74 = ((($73)) + -2|0);
    $75 = HEAP8[$74>>0]|0;
    $76 = $75&255;
    $77 = $76 << 8;
    $78 = ((($74)) + 1|0);
    $79 = HEAP8[$78>>0]|0;
    $80 = $79&255;
    $81 = $77 | $80;
    $82 = $72 << 16;
    $83 = $81 | $82;
    $84 = ((($1)) + 20|0);
    HEAP32[$84>>2] = $83;
   }
   HEAP32[$1>>2] = 0;
   $104 = ((($phy_font)) + 140|0);
   $105 = HEAP32[$104>>2]|0;
   HEAP32[$105>>2] = $1;
   HEAP32[$104>>2] = $1;
   $106 = ((($phy_font)) + 132|0);
   $107 = HEAP32[$106>>2]|0;
   $108 = (($107) + ($37))|0;
   HEAP32[$106>>2] = $108;
   $109 = HEAP32[$error>>2]|0;
   STACKTOP = sp;return ($109|0);
  }
 }
 _ft_mem_free($0,$1);
 HEAP32[$error>>2] = 8;
 $109 = HEAP32[$error>>2]|0;
 STACKTOP = sp;return ($109|0);
}
function _pfr_face_done($pfrface) {
 $pfrface = $pfrface|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $item$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($pfrface|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($pfrface)) + 96|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($pfrface)) + 20|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($pfrface)) + 24|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($pfrface)) + 288|0);
 $8 = ((($pfrface)) + 100|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($7)) + 80|0);
 $11 = HEAP32[$10>>2]|0;
 _ft_mem_free($9,$11);
 HEAP32[$10>>2] = 0;
 $12 = ((($7)) + 84|0);
 $13 = HEAP32[$12>>2]|0;
 _ft_mem_free($9,$13);
 HEAP32[$12>>2] = 0;
 $14 = ((($7)) + 88|0);
 $15 = HEAP32[$14>>2]|0;
 _ft_mem_free($9,$15);
 HEAP32[$14>>2] = 0;
 $16 = ((($7)) + 76|0);
 $17 = HEAP32[$16>>2]|0;
 _ft_mem_free($9,$17);
 HEAP32[$16>>2] = 0;
 $18 = ((($7)) + 72|0);
 HEAP32[$18>>2] = 0;
 $19 = ((($7)) + 64|0);
 HEAP32[$19>>2] = 0;
 $20 = ((($7)) + 60|0);
 HEAP32[$20>>2] = 0;
 $21 = ((($7)) + 100|0);
 $22 = HEAP32[$21>>2]|0;
 _ft_mem_free($9,$22);
 HEAP32[$21>>2] = 0;
 $23 = ((($7)) + 92|0);
 HEAP32[$23>>2] = 0;
 $24 = ((($7)) + 96|0);
 HEAP32[$24>>2] = 0;
 $25 = ((($7)) + 128|0);
 $26 = HEAP32[$25>>2]|0;
 _ft_mem_free($9,$26);
 HEAP32[$25>>2] = 0;
 $27 = ((($7)) + 120|0);
 HEAP32[$27>>2] = 0;
 $28 = ((($7)) + 124|0);
 HEAP32[$28>>2] = 0;
 $29 = ((($7)) + 108|0);
 $30 = HEAP32[$29>>2]|0;
 _ft_mem_free($9,$30);
 HEAP32[$29>>2] = 0;
 $31 = ((($7)) + 104|0);
 HEAP32[$31>>2] = 0;
 $32 = ((($7)) + 136|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ($33|0)==(0|0);
 if (!($34)) {
  $item$01$i = $33;
  while(1) {
   $35 = HEAP32[$item$01$i>>2]|0;
   _ft_mem_free($9,$item$01$i);
   $36 = ($35|0)==(0|0);
   if ($36) {
    break;
   } else {
    $item$01$i = $35;
   }
  }
 }
 HEAP32[$32>>2] = 0;
 $37 = ((($7)) + 140|0);
 HEAP32[$37>>2] = 0;
 $38 = ((($7)) + 132|0);
 HEAP32[$38>>2] = 0;
 $39 = ((($pfrface)) + 32|0);
 $40 = HEAP32[$39>>2]|0;
 _ft_mem_free($4,$40);
 HEAP32[$39>>2] = 0;
 return;
}
function _pfr_slot_init($pfrslot) {
 $pfrslot = $pfrslot|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($pfrslot)) + 156|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($pfrslot)) + 160|0);
 dest=$3; stop=dest+36|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $4 = ((($pfrslot)) + 188|0);
 HEAP32[$4>>2] = $2;
 $5 = ((($pfrslot)) + 192|0);
 HEAP8[$5>>0] = 0;
 _FT_GlyphLoader_Rewind($2);
 return 0;
}
function _pfr_slot_done($pfrslot) {
 $pfrslot = $pfrslot|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pfrslot)) + 188|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($pfrslot)) + 168|0);
 $4 = HEAP32[$3>>2]|0;
 _ft_mem_free($2,$4);
 HEAP32[$3>>2] = 0;
 $5 = ((($pfrslot)) + 172|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($pfrslot)) + 164|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($pfrslot)) + 184|0);
 $8 = HEAP32[$7>>2]|0;
 _ft_mem_free($2,$8);
 $9 = ((($pfrslot)) + 176|0);
 ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;HEAP32[$9+12>>2]=0|0;HEAP8[$9+16>>0]=0|0;
 return;
}
function _pfr_slot_load($pfrslot,$pfrsize,$gindex,$load_flags) {
 $pfrslot = $pfrslot|0;
 $pfrsize = $pfrsize|0;
 $gindex = $gindex|0;
 $load_flags = $load_flags|0;
 var $$$i$i$i = 0, $$01$i$i$i = 0, $$02$i$i$i = 0, $$0218$i$i$i = 0, $$1$i$i$i = 0, $$1$i11$i$i = 0, $$1$ph$i$i$i = 0, $$2$i$i$i = 0, $$3$i$i$i = 0, $$char_len$1$i$i = 0, $$lcssa52 = 0, $$left$0$i$i = 0, $$left$0$us$i$i = 0, $$lobit$i = 0, $$lobit$i$i = 0, $$lobit15$i = 0, $$lobit3$i$i = 0, $$old1$i$i$i = 0, $$pre = 0, $$pre$i = 0;
 var $$pre21 = 0, $$pre22 = 0, $$pre35$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0;
 var $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0;
 var $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0;
 var $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0;
 var $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0;
 var $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0;
 var $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0;
 var $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0;
 var $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0;
 var $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0;
 var $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0;
 var $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0;
 var $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0;
 var $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0;
 var $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $advance$0 = 0, $advance$0$i = 0, $advance$0$i$i = 0, $buff$0$lcssa$i$i = 0, $buff$2$i$i = 0, $c$0$$i$i$i = 0, $c$0$$i12$i$i = 0, $c$0$$i4$i$i = 0, $c$017$i$i$i = 0;
 var $c$07$i$i$i = 0, $c$09$i$i$i = 0, $c$2$i$i$i = 0, $c$2$i$i$i$lcssa = 0, $c$2$i16$i$i = 0, $c$2$i16$i$i$lcssa = 0, $c$2$i8$i$i = 0, $c$2$i8$i$i$lcssa = 0, $cbox = 0, $char_len$0$$i = 0, $char_len$1$i$i = 0, $char_len$2$i = 0, $count$010$i$i$i = 0, $count$03$i$i$i = 0, $count$1$i$i$i = 0, $count$1$i3$i$i = 0, $count$1$i3$i$i$lcssa = 0, $count$1$ph$i$i$i = 0, $count$2$i$i$i = 0, $count$3$i$i$i = 0;
 var $counts$sroa$5$011$i$i$i = 0, $counts$sroa$5$1$i$i$i = 0, $counts$sroa$5$1$i$i$i$lcssa = 0, $counts$sroa$5$1$ph$i$i$i = 0, $counts$sroa$5$3$i$i$i = 0, $cur$015$i$i$i = 0, $cur$04$i$i$i = 0, $cur$07$i$i$i = 0, $cur$1$i$i$i = 0, $cur$1$i$i$i$lcssa = 0, $cur$1$i14$i$i = 0, $cur$1$i14$i$i$lcssa = 0, $cur$1$i6$i$i = 0, $cur$1$i6$i$i$lcssa = 0, $error$0 = 0, $gindex$ = 0, $left$014$i$i$i = 0, $left$03$i$i$i = 0, $left$06$i$i = 0, $left$06$i$i$i = 0;
 var $left$06$us$i$i = 0, $left$1$i$i$i = 0, $left$1$i13$i$i = 0, $left$1$i5$i$i = 0, $mask$016$i$i$i = 0, $mask$05$i$i$i = 0, $mask$08$i$i$i = 0, $mask$1$i$i$i = 0, $mask$1$i$i$i$lcssa = 0, $mask$1$i15$i$i = 0, $mask$1$i15$i$i$lcssa = 0, $mask$1$i7$i$i = 0, $mask$1$i7$i$i$lcssa = 0, $n$011 = 0, $n$012$i$i$i = 0, $n$026$i = 0, $n$04$i$i$i = 0, $n$11$i$i$i = 0, $p$0$i$i = 0, $p$1$i$i = 0;
 var $p$2$i$i = 0, $phase$02$i$i$i = 0, $phase$09$i$i$i = 0, $phase$1$i$i$i = 0, $phase$1$i2$i$i = 0, $phase$1$i2$i$i$lcssa = 0, $phase$1$ph$i$i$i = 0, $phase$2$i$i$i = 0, $phase$3$i$i$i = 0, $reload$013$i$i$i = 0, $reload$05$i$i$i = 0, $right$0$$i$i = 0, $right$0$$us$i$i = 0, $right$05$i$i = 0, $right$05$us$i$i = 0, $sext$i$i = 0, $sext1$i$i = 0, $sext2$i$i = 0, $storemerge$i$i = 0, $storemerge4$i$i = 0;
 var $strike$025$i = 0, $strike$025$i$lcssa = 0, $val$06$i$i$i = 0, $val$1$i$i$i = 0, $vec$012 = 0, $writer$sroa$0$0$i$i = 0, $writer$sroa$0$1$i$i = 0, $writer$sroa$0$2$i$i = 0, $writer$sroa$0$3$i$i = 0, $writer$sroa$0$4$i$i = 0, $writer$sroa$0$5$i$i = 0, $writer$sroa$0$6$i$i = 0, $writer$sroa$13$0$i$i = 0, $xpos$0$i$i = 0, $xsize$0$i$i = 0, $ypos$0$i$i = 0, $ysize$0$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $cbox = sp;
 $0 = ((($pfrslot)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($pfrslot)) + 108|0);
 $3 = ($gindex|0)==(0);
 $4 = (($gindex) + -1)|0;
 $gindex$ = $3 ? 0 : $4;
 $5 = ($1|0)==(0|0);
 if ($5) {
  $error$0 = 6;
  STACKTOP = sp;return ($error$0|0);
 }
 $6 = ((($1)) + 408|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($gindex$>>>0)<($7>>>0);
 if (!($8)) {
  $error$0 = 6;
  STACKTOP = sp;return ($error$0|0);
 }
 $9 = $load_flags & 9;
 $10 = ($9|0)==(0);
 L7: do {
  if ($10) {
   $11 = ((($1)) + 104|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ((($1)) + 416|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ((($1)) + 380|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = ($16|0)==(0);
   if (!($17)) {
    $18 = ((($1)) + 388|0);
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($pfrsize)) + 12|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = $21 & 65535;
    $23 = $21 >>> 16;
    $n$026$i = 0;$strike$025$i = $19;
    while(1) {
     $24 = HEAP32[$strike$025$i>>2]|0;
     $25 = ($24|0)==($22|0);
     if ($25) {
      $26 = ((($strike$025$i)) + 4|0);
      $27 = HEAP32[$26>>2]|0;
      $28 = ($27|0)==($23|0);
      if ($28) {
       $strike$025$i$lcssa = $strike$025$i;
       break;
      }
     }
     $29 = ((($strike$025$i)) + 36|0);
     $30 = (($n$026$i) + 1)|0;
     $31 = ($30>>>0)<($16>>>0);
     if ($31) {
      $n$026$i = $30;$strike$025$i = $29;
     } else {
      break L7;
     }
    }
    $32 = ((($strike$025$i$lcssa)) + 8|0);
    $33 = HEAP32[$32>>2]|0;
    $34 = ((($1)) + 432|0);
    $35 = HEAP32[$34>>2]|0;
    $36 = ((($strike$025$i$lcssa)) + 24|0);
    $37 = HEAP32[$36>>2]|0;
    $38 = (($37) + ($35))|0;
    $39 = (_FT_Stream_Seek($12,$38)|0);
    $40 = ($39|0)==(0);
    if ($40) {
     $41 = $33 & 1;
     $42 = $41 | 4;
     $43 = $33 >>> 1;
     $$lobit$i = $43 & 1;
     $char_len$0$$i = (($42) + ($$lobit$i))|0;
     $44 = $33 >>> 2;
     $$lobit15$i = $44 & 1;
     $char_len$2$i = (($char_len$0$$i) + ($$lobit15$i))|0;
     $45 = ((($strike$025$i$lcssa)) + 28|0);
     $46 = HEAP32[$45>>2]|0;
     $47 = Math_imul($46, $char_len$2$i)|0;
     $48 = (_FT_Stream_EnterFrame($12,$47)|0);
     $49 = ($48|0)==(0);
     if ($49) {
      $50 = ((($12)) + 32|0);
      $51 = HEAP32[$50>>2]|0;
      $52 = ((($12)) + 36|0);
      $53 = HEAP32[$52>>2]|0;
      $54 = HEAP32[$45>>2]|0;
      $55 = HEAP32[$32>>2]|0;
      $56 = (($14) + ($gindex$<<4)|0);
      $57 = HEAP32[$56>>2]|0;
      $58 = $55 & 1;
      $59 = $58 | 4;
      $60 = $55 & 2;
      $61 = ($60|0)==(0);
      $$lobit$i$i = $60 >>> 1;
      $char_len$1$i$i = (($59) + ($$lobit$i$i))|0;
      $62 = $55 & 4;
      $63 = ($62|0)==(0);
      $$lobit3$i$i = $62 >>> 2;
      $$char_len$1$i$i = (($char_len$1$i$i) + ($$lobit3$i$i))|0;
      $64 = ($54|0)==(0);
      L18: do {
       if (!($64)) {
        $65 = ($58|0)==(0);
        L20: do {
         if ($65) {
          $left$06$us$i$i = 0;$right$05$us$i$i = $54;
          while(1) {
           $66 = (($right$05$us$i$i) + ($left$06$us$i$i))|0;
           $67 = $66 >>> 1;
           $68 = Math_imul($67, $$char_len$1$i$i)|0;
           $69 = (($51) + ($68)|0);
           $70 = (($69) + ($$char_len$1$i$i)|0);
           $71 = ($70>>>0)>($53>>>0);
           if ($71) {
            break L18;
           }
           $72 = HEAP8[$69>>0]|0;
           $73 = $72&255;
           $74 = ($73|0)==($57|0);
           if ($74) {
            $$lcssa52 = $69;
            break;
           }
           $75 = ($73>>>0)<($57>>>0);
           $right$0$$us$i$i = $75 ? $right$05$us$i$i : $67;
           $$left$0$us$i$i = $75 ? $67 : $left$06$us$i$i;
           $76 = ($$left$0$us$i$i>>>0)<($right$0$$us$i$i>>>0);
           if ($76) {
            $left$06$us$i$i = $$left$0$us$i$i;$right$05$us$i$i = $right$0$$us$i$i;
           } else {
            break L18;
           }
          }
          $95 = ((($$lcssa52)) + 1|0);
          $buff$0$lcssa$i$i = $95;
         } else {
          $left$06$i$i = 0;$right$05$i$i = $54;
          while(1) {
           $77 = (($right$05$i$i) + ($left$06$i$i))|0;
           $78 = $77 >>> 1;
           $79 = Math_imul($78, $$char_len$1$i$i)|0;
           $80 = (($51) + ($79)|0);
           $81 = (($80) + ($$char_len$1$i$i)|0);
           $82 = ($81>>>0)>($53>>>0);
           if ($82) {
            break L18;
           }
           $83 = ((($80)) + 2|0);
           $84 = ((($83)) + -2|0);
           $85 = HEAP8[$84>>0]|0;
           $86 = $85&255;
           $87 = $86 << 8;
           $88 = ((($84)) + 1|0);
           $89 = HEAP8[$88>>0]|0;
           $90 = $89&255;
           $91 = $87 | $90;
           $92 = ($91|0)==($57|0);
           if ($92) {
            $buff$0$lcssa$i$i = $83;
            break L20;
           }
           $93 = ($91>>>0)<($57>>>0);
           $right$0$$i$i = $93 ? $right$05$i$i : $78;
           $$left$0$i$i = $93 ? $78 : $left$06$i$i;
           $94 = ($$left$0$i$i>>>0)<($right$0$$i$i>>>0);
           if ($94) {
            $left$06$i$i = $$left$0$i$i;$right$05$i$i = $right$0$$i$i;
           } else {
            break L18;
           }
          }
         }
        } while(0);
        if ($61) {
         $104 = ((($buff$0$lcssa$i$i)) + 1|0);
         $105 = HEAP8[$buff$0$lcssa$i$i>>0]|0;
         $106 = $105&255;
         $buff$2$i$i = $104;$storemerge$i$i = $106;
        } else {
         $96 = ((($buff$0$lcssa$i$i)) + 2|0);
         $97 = HEAP8[$buff$0$lcssa$i$i>>0]|0;
         $98 = $97&255;
         $99 = $98 << 8;
         $100 = ((($buff$0$lcssa$i$i)) + 1|0);
         $101 = HEAP8[$100>>0]|0;
         $102 = $101&255;
         $103 = $99 | $102;
         $buff$2$i$i = $96;$storemerge$i$i = $103;
        }
        $107 = HEAP8[$buff$2$i$i>>0]|0;
        $108 = $107&255;
        $109 = ((($buff$2$i$i)) + 1|0);
        $110 = HEAP8[$109>>0]|0;
        if ($63) {
         $119 = $108 << 8;
         $120 = $110&255;
         $121 = $120 | $119;
         $storemerge4$i$i = $121;
        } else {
         $111 = $108 << 16;
         $112 = $110&255;
         $113 = $112 << 8;
         $114 = $113 | $111;
         $115 = ((($buff$2$i$i)) + 2|0);
         $116 = HEAP8[$115>>0]|0;
         $117 = $116&255;
         $118 = $114 | $117;
         $storemerge4$i$i = $118;
        }
        _FT_Stream_ExitFrame($12);
        $122 = ($storemerge$i$i|0)==(0);
        if ($122) {
         break L7;
        }
        $123 = (((($14) + ($gindex$<<4)|0)) + 4|0);
        $124 = HEAP32[$123>>2]|0;
        $125 = ((($1)) + 304|0);
        $126 = HEAP32[$125>>2]|0;
        $127 = ((($1)) + 300|0);
        $128 = HEAP32[$127>>2]|0;
        $129 = ($126|0)==($128|0);
        if ($129) {
         $135 = $124;$136 = $126;$advance$0$i = $124;
        } else {
         $130 = (_FT_MulDiv($124,$128,$126)|0);
         $$pre$i = HEAP32[$123>>2]|0;
         $$pre35$i = HEAP32[$125>>2]|0;
         $135 = $$pre$i;$136 = $$pre35$i;$advance$0$i = $130;
        }
        $131 = ((($pfrslot)) + 56|0);
        HEAP32[$131>>2] = $advance$0$i;
        $132 = HEAP16[$20>>1]|0;
        $133 = $132&65535;
        $134 = $133 << 8;
        $137 = (_FT_MulDiv($134,$135,$136)|0);
        $138 = ((($1)) + 188|0);
        $139 = HEAP32[$138>>2]|0;
        $140 = (($139) + ($storemerge4$i$i))|0;
        $141 = (_FT_Stream_Seek($12,$140)|0);
        $142 = ($141|0)==(0);
        if (!($142)) {
         break L7;
        }
        $143 = (_FT_Stream_EnterFrame($12,$storemerge$i$i)|0);
        $144 = ($143|0)==(0);
        if (!($144)) {
         break L7;
        }
        $145 = HEAP32[$50>>2]|0;
        $146 = HEAP32[$52>>2]|0;
        $147 = ((($145)) + 1|0);
        $148 = ($147>>>0)>($146>>>0);
        L45: do {
         if (!($148)) {
          $149 = HEAP8[$145>>0]|0;
          $150 = $149&255;
          $151 = $150 & 3;
          switch ($151|0) {
          case 0:  {
           $152 = ((($145)) + 2|0);
           $153 = ($152>>>0)>($146>>>0);
           if ($153) {
            break L45;
           }
           $154 = HEAP8[$147>>0]|0;
           $155 = $154 << 24 >> 24;
           $156 = $155 >> 4;
           $157 = $155 << 28;
           $158 = $157 >> 28;
           $p$0$i$i = $152;$xpos$0$i$i = $156;$ypos$0$i$i = $158;
           break;
          }
          case 1:  {
           $159 = ((($145)) + 3|0);
           $160 = ($159>>>0)>($146>>>0);
           if ($160) {
            break L45;
           }
           $161 = ((($145)) + 2|0);
           $162 = HEAP8[$147>>0]|0;
           $163 = $162 << 24 >> 24;
           $164 = HEAP8[$161>>0]|0;
           $165 = $164 << 24 >> 24;
           $p$0$i$i = $159;$xpos$0$i$i = $163;$ypos$0$i$i = $165;
           break;
          }
          case 2:  {
           $166 = ((($145)) + 5|0);
           $167 = ($166>>>0)>($146>>>0);
           if ($167) {
            break L45;
           }
           $168 = HEAP8[$147>>0]|0;
           $169 = $168&255;
           $170 = $169 << 8;
           $171 = ((($145)) + 2|0);
           $172 = HEAP8[$171>>0]|0;
           $173 = $172&255;
           $174 = $170 | $173;
           $sext1$i$i = $174 << 16;
           $175 = $sext1$i$i >> 16;
           $176 = ((($145)) + 3|0);
           $177 = HEAP8[$176>>0]|0;
           $178 = $177&255;
           $179 = $178 << 8;
           $180 = ((($145)) + 4|0);
           $181 = HEAP8[$180>>0]|0;
           $182 = $181&255;
           $183 = $179 | $182;
           $sext2$i$i = $183 << 16;
           $184 = $sext2$i$i >> 16;
           $p$0$i$i = $166;$xpos$0$i$i = $175;$ypos$0$i$i = $184;
           break;
          }
          case 3:  {
           $185 = ((($145)) + 7|0);
           $186 = ($185>>>0)>($146>>>0);
           if ($186) {
            break L45;
           }
           $187 = HEAP8[$147>>0]|0;
           $188 = $187&255;
           $189 = $188 << 16;
           $190 = ((($145)) + 2|0);
           $191 = HEAP8[$190>>0]|0;
           $192 = $191&255;
           $193 = $192 << 8;
           $194 = $193 | $189;
           $195 = ((($145)) + 3|0);
           $196 = HEAP8[$195>>0]|0;
           $197 = $196&255;
           $198 = $194 | $197;
           $199 = ((($145)) + 4|0);
           $200 = HEAP8[$199>>0]|0;
           $201 = $200&255;
           $202 = $201 << 16;
           $203 = ((($145)) + 5|0);
           $204 = HEAP8[$203>>0]|0;
           $205 = $204&255;
           $206 = $205 << 8;
           $207 = $206 | $202;
           $208 = ((($145)) + 6|0);
           $209 = HEAP8[$208>>0]|0;
           $210 = $209&255;
           $211 = $207 | $210;
           $p$0$i$i = $185;$xpos$0$i$i = $198;$ypos$0$i$i = $211;
           break;
          }
          default: {
           // unreachable;
          }
          }
          $212 = ($149&255) >>> 2;
          $213 = $212&255;
          $214 = $213 & 3;
          switch ($214|0) {
          case 0:  {
           $p$1$i$i = $p$0$i$i;$xsize$0$i$i = 0;$ysize$0$i$i = 0;
           break;
          }
          case 1:  {
           $215 = ((($p$0$i$i)) + 1|0);
           $216 = ($215>>>0)>($146>>>0);
           if ($216) {
            break L45;
           }
           $217 = HEAP8[$p$0$i$i>>0]|0;
           $218 = $217&255;
           $219 = $218 >>> 4;
           $220 = $218 & 15;
           $p$1$i$i = $215;$xsize$0$i$i = $219;$ysize$0$i$i = $220;
           break;
          }
          case 2:  {
           $221 = ((($p$0$i$i)) + 2|0);
           $222 = ($221>>>0)>($146>>>0);
           if ($222) {
            break L45;
           }
           $223 = ((($p$0$i$i)) + 1|0);
           $224 = HEAP8[$p$0$i$i>>0]|0;
           $225 = $224&255;
           $226 = HEAP8[$223>>0]|0;
           $227 = $226&255;
           $p$1$i$i = $221;$xsize$0$i$i = $225;$ysize$0$i$i = $227;
           break;
          }
          case 3:  {
           $228 = ((($p$0$i$i)) + 4|0);
           $229 = ($228>>>0)>($146>>>0);
           if ($229) {
            break L45;
           }
           $230 = HEAP8[$p$0$i$i>>0]|0;
           $231 = $230&255;
           $232 = $231 << 8;
           $233 = ((($p$0$i$i)) + 1|0);
           $234 = HEAP8[$233>>0]|0;
           $235 = $234&255;
           $236 = $232 | $235;
           $237 = ((($p$0$i$i)) + 2|0);
           $238 = HEAP8[$237>>0]|0;
           $239 = $238&255;
           $240 = $239 << 8;
           $241 = ((($p$0$i$i)) + 3|0);
           $242 = HEAP8[$241>>0]|0;
           $243 = $242&255;
           $244 = $240 | $243;
           $p$1$i$i = $228;$xsize$0$i$i = $236;$ysize$0$i$i = $244;
           break;
          }
          default: {
           // unreachable;
          }
          }
          $245 = ($149&255) >>> 4;
          $246 = $245&255;
          $247 = $246 & 3;
          switch ($247|0) {
          case 0:  {
           $advance$0$i$i = $137;$p$2$i$i = $p$1$i$i;
           break;
          }
          case 1:  {
           $248 = ((($p$1$i$i)) + 1|0);
           $249 = ($248>>>0)>($146>>>0);
           if ($249) {
            break L45;
           }
           $250 = HEAP8[$p$1$i$i>>0]|0;
           $251 = $250 << 24 >> 24;
           $252 = $251 << 8;
           $advance$0$i$i = $252;$p$2$i$i = $248;
           break;
          }
          case 2:  {
           $253 = ((($p$1$i$i)) + 2|0);
           $254 = ($253>>>0)>($146>>>0);
           if ($254) {
            break L45;
           }
           $255 = HEAP8[$p$1$i$i>>0]|0;
           $256 = $255&255;
           $257 = $256 << 8;
           $258 = ((($p$1$i$i)) + 1|0);
           $259 = HEAP8[$258>>0]|0;
           $260 = $259&255;
           $261 = $257 | $260;
           $sext$i$i = $261 << 16;
           $262 = $sext$i$i >> 16;
           $advance$0$i$i = $262;$p$2$i$i = $253;
           break;
          }
          case 3:  {
           $263 = ((($p$1$i$i)) + 3|0);
           $264 = ($263>>>0)>($146>>>0);
           if ($264) {
            break L45;
           }
           $265 = HEAP8[$p$1$i$i>>0]|0;
           $266 = $265&255;
           $267 = $266 << 16;
           $268 = ((($p$1$i$i)) + 1|0);
           $269 = HEAP8[$268>>0]|0;
           $270 = $269&255;
           $271 = $270 << 8;
           $272 = $271 | $267;
           $273 = ((($p$1$i$i)) + 2|0);
           $274 = HEAP8[$273>>0]|0;
           $275 = $274&255;
           $276 = $272 | $275;
           $advance$0$i$i = $276;$p$2$i$i = $263;
           break;
          }
          default: {
           // unreachable;
          }
          }
          $277 = $246 >>> 2;
          $278 = $p$2$i$i;
          $279 = (2147483647 - ($ysize$0$i$i))|0;
          $280 = ($ypos$0$i$i|0)>($279|0);
          if (!($280)) {
           $281 = ((($pfrslot)) + 72|0);
           HEAP32[$281>>2] = 1651078259;
           $282 = ((($pfrslot)) + 80|0);
           HEAP32[$282>>2] = $xsize$0$i$i;
           $283 = ((($pfrslot)) + 76|0);
           HEAP32[$283>>2] = $ysize$0$i$i;
           $284 = (($xsize$0$i$i) + 7)|0;
           $285 = $284 >> 3;
           $286 = ((($pfrslot)) + 84|0);
           HEAP32[$286>>2] = $285;
           $287 = ((($pfrslot)) + 94|0);
           HEAP8[$287>>0] = 1;
           $288 = $xsize$0$i$i << 6;
           $289 = ((($pfrslot)) + 24|0);
           HEAP32[$289>>2] = $288;
           $290 = $ysize$0$i$i << 6;
           $291 = ((($pfrslot)) + 28|0);
           HEAP32[$291>>2] = $290;
           $292 = $xpos$0$i$i << 6;
           $293 = ((($pfrslot)) + 32|0);
           HEAP32[$293>>2] = $292;
           $294 = $ypos$0$i$i << 6;
           $295 = ((($pfrslot)) + 36|0);
           HEAP32[$295>>2] = $294;
           $296 = $advance$0$i$i >> 2;
           $297 = (($296) + 32)|0;
           $298 = $297 & -64;
           $299 = ((($pfrslot)) + 40|0);
           HEAP32[$299>>2] = $298;
           $300 = (0 - ($288))|0;
           $301 = $300 >> 1;
           $302 = ((($pfrslot)) + 44|0);
           HEAP32[$302>>2] = $301;
           $303 = ((($pfrslot)) + 48|0);
           HEAP32[$303>>2] = 0;
           $304 = ((($pfrsize)) + 32|0);
           $305 = HEAP32[$304>>2]|0;
           $306 = ((($pfrslot)) + 52|0);
           HEAP32[$306>>2] = $305;
           $307 = ((($pfrslot)) + 100|0);
           HEAP32[$307>>2] = $xpos$0$i$i;
           $308 = (($ysize$0$i$i) + ($ypos$0$i$i))|0;
           $309 = ((($pfrslot)) + 104|0);
           HEAP32[$309>>2] = $308;
           $310 = Math_imul($285, $ysize$0$i$i)|0;
           $311 = (_ft_glyphslot_alloc_bitmap($pfrslot,$310)|0);
           $312 = ($311|0)==(0);
           if ($312) {
            $313 = HEAP32[$52>>2]|0;
            $314 = ((($1)) + 208|0);
            $315 = HEAP32[$314>>2]|0;
            $316 = $315 & 2;
            $317 = HEAP32[$283>>2]|0;
            $318 = ($317|0)==(0);
            L78: do {
             if (!($318)) {
              $319 = HEAP32[$282>>2]|0;
              $320 = ($319|0)==(0);
              if ($320) {
               break;
              }
              $321 = ((($pfrslot)) + 88|0);
              $322 = HEAP32[$321>>2]|0;
              $323 = HEAP32[$286>>2]|0;
              $324 = Math_imul($319, $317)|0;
              $325 = ($316|0)==(0);
              if ($325) {
               $326 = $322;
               $327 = (($317) + -1)|0;
               $328 = Math_imul($323, $327)|0;
               $329 = (($326) + ($328)|0);
               $330 = $329;
               $331 = (0 - ($323))|0;
               $writer$sroa$0$0$i$i = $330;$writer$sroa$13$0$i$i = $331;
              } else {
               $writer$sroa$0$0$i$i = $322;$writer$sroa$13$0$i$i = $323;
              }
              switch ($277|0) {
              case 0:  {
               $332 = $313;
               $333 = (($332) - ($278))|0;
               $334 = $333 << 3;
               $335 = ($334>>>0)>($324>>>0);
               $$$i$i$i = $335 ? $324 : $334;
               $336 = ($$$i$i$i|0)==(0);
               if ($336) {
                break L78;
               }
               $337 = $writer$sroa$0$0$i$i;
               $$02$i$i$i = $p$2$i$i;$c$07$i$i$i = 0;$cur$04$i$i$i = $337;$left$03$i$i$i = $319;$mask$05$i$i$i = 128;$n$11$i$i$i = $$$i$i$i;$val$06$i$i$i = 0;$writer$sroa$0$1$i$i = $writer$sroa$0$0$i$i;
               while(1) {
                $338 = $n$11$i$i$i ^ $$$i$i$i;
                $339 = $338 & 7;
                $340 = ($339|0)==(0);
                if ($340) {
                 $341 = ((($$02$i$i$i)) + 1|0);
                 $342 = HEAP8[$$02$i$i$i>>0]|0;
                 $343 = $342&255;
                 $$1$i11$i$i = $341;$val$1$i$i$i = $343;
                } else {
                 $$1$i11$i$i = $$02$i$i$i;$val$1$i$i$i = $val$06$i$i$i;
                }
                $344 = $val$1$i$i$i & 128;
                $345 = ($344|0)==(0);
                $346 = $345 ? 0 : $mask$05$i$i$i;
                $c$0$$i12$i$i = $346 | $c$07$i$i$i;
                $347 = $val$1$i$i$i << 1;
                $348 = $mask$05$i$i$i >>> 1;
                $349 = (($left$03$i$i$i) + -1)|0;
                $350 = ($349|0)==(0);
                do {
                 if ($350) {
                  $351 = $c$0$$i12$i$i&255;
                  HEAP8[$cur$04$i$i$i>>0] = $351;
                  $352 = $writer$sroa$0$1$i$i;
                  $353 = (($352) + ($writer$sroa$13$0$i$i)|0);
                  $354 = $353;
                  $c$2$i16$i$i = 0;$cur$1$i14$i$i = $353;$left$1$i13$i$i = $319;$mask$1$i15$i$i = 128;$writer$sroa$0$2$i$i = $354;
                 } else {
                  $355 = ($348|0)==(0);
                  if (!($355)) {
                   $c$2$i16$i$i = $c$0$$i12$i$i;$cur$1$i14$i$i = $cur$04$i$i$i;$left$1$i13$i$i = $349;$mask$1$i15$i$i = $348;$writer$sroa$0$2$i$i = $writer$sroa$0$1$i$i;
                   break;
                  }
                  $356 = $c$0$$i12$i$i&255;
                  HEAP8[$cur$04$i$i$i>>0] = $356;
                  $357 = ((($cur$04$i$i$i)) + 1|0);
                  $c$2$i16$i$i = 0;$cur$1$i14$i$i = $357;$left$1$i13$i$i = $349;$mask$1$i15$i$i = 128;$writer$sroa$0$2$i$i = $writer$sroa$0$1$i$i;
                 }
                } while(0);
                $358 = (($n$11$i$i$i) + -1)|0;
                $359 = ($358|0)==(0);
                if ($359) {
                 $c$2$i16$i$i$lcssa = $c$2$i16$i$i;$cur$1$i14$i$i$lcssa = $cur$1$i14$i$i;$mask$1$i15$i$i$lcssa = $mask$1$i15$i$i;
                 break;
                } else {
                 $$02$i$i$i = $$1$i11$i$i;$c$07$i$i$i = $c$2$i16$i$i;$cur$04$i$i$i = $cur$1$i14$i$i;$left$03$i$i$i = $left$1$i13$i$i;$mask$05$i$i$i = $mask$1$i15$i$i;$n$11$i$i$i = $358;$val$06$i$i$i = $347;$writer$sroa$0$1$i$i = $writer$sroa$0$2$i$i;
                }
               }
               $360 = ($mask$1$i15$i$i$lcssa|0)==(128);
               if ($360) {
                break L78;
               }
               $361 = $c$2$i16$i$i$lcssa&255;
               HEAP8[$cur$1$i14$i$i$lcssa>>0] = $361;
               break L78;
               break;
              }
              case 1:  {
               $362 = ($324|0)==(0);
               if ($362) {
                break L78;
               }
               $363 = $writer$sroa$0$0$i$i;
               $$0218$i$i$i = $p$2$i$i;$c$017$i$i$i = 0;$count$010$i$i$i = 0;$counts$sroa$5$011$i$i$i = 0;$cur$015$i$i$i = $363;$left$014$i$i$i = $319;$mask$016$i$i$i = 128;$n$012$i$i$i = $324;$phase$09$i$i$i = 1;$reload$013$i$i$i = 1;$writer$sroa$0$3$i$i = $writer$sroa$0$0$i$i;
               while(1) {
                $364 = ($reload$013$i$i$i|0)==(0);
                L103: do {
                 if ($364) {
                  $$3$i$i$i = $$0218$i$i$i;$count$3$i$i$i = $count$010$i$i$i;$counts$sroa$5$3$i$i$i = $counts$sroa$5$011$i$i$i;$phase$3$i$i$i = $phase$09$i$i$i;
                 } else {
                  $$1$ph$i$i$i = $$0218$i$i$i;$count$1$ph$i$i$i = $count$010$i$i$i;$counts$sroa$5$1$ph$i$i$i = $counts$sroa$5$011$i$i$i;$phase$1$ph$i$i$i = $phase$09$i$i$i;
                  while(1) {
                   $count$1$i3$i$i = $count$1$ph$i$i$i;$counts$sroa$5$1$i$i$i = $counts$sroa$5$1$ph$i$i$i;$phase$1$i2$i$i = $phase$1$ph$i$i$i;
                   while(1) {
                    $365 = ($phase$1$i2$i$i|0)==(0);
                    if (!($365)) {
                     $count$1$i3$i$i$lcssa = $count$1$i3$i$i;$counts$sroa$5$1$i$i$i$lcssa = $counts$sroa$5$1$i$i$i;$phase$1$i2$i$i$lcssa = $phase$1$i2$i$i;
                     break;
                    }
                    $$old1$i$i$i = ($counts$sroa$5$1$i$i$i|0)==(0);
                    if ($$old1$i$i$i) {
                     $count$1$i3$i$i = 0;$counts$sroa$5$1$i$i$i = 0;$phase$1$i2$i$i = 1;
                    } else {
                     $$3$i$i$i = $$1$ph$i$i$i;$count$3$i$i$i = $counts$sroa$5$1$i$i$i;$counts$sroa$5$3$i$i$i = $counts$sroa$5$1$i$i$i;$phase$3$i$i$i = 1;
                     break L103;
                    }
                   }
                   $366 = ($$1$ph$i$i$i>>>0)<($313>>>0);
                   if (!($366)) {
                    $$3$i$i$i = $$1$ph$i$i$i;$count$3$i$i$i = $count$1$i3$i$i$lcssa;$counts$sroa$5$3$i$i$i = $counts$sroa$5$1$i$i$i$lcssa;$phase$3$i$i$i = $phase$1$i2$i$i$lcssa;
                    break L103;
                   }
                   $367 = ((($$1$ph$i$i$i)) + 1|0);
                   $368 = HEAP8[$$1$ph$i$i$i>>0]|0;
                   $369 = $368&255;
                   $370 = $369 >>> 4;
                   $371 = $369 & 15;
                   $372 = ($370|0)==(0);
                   if ($372) {
                    $$1$ph$i$i$i = $367;$count$1$ph$i$i$i = 0;$counts$sroa$5$1$ph$i$i$i = $371;$phase$1$ph$i$i$i = 0;
                   } else {
                    $$3$i$i$i = $367;$count$3$i$i$i = $370;$counts$sroa$5$3$i$i$i = $371;$phase$3$i$i$i = 0;
                    break;
                   }
                  }
                 }
                } while(0);
                $373 = ($phase$3$i$i$i|0)==(0);
                $374 = $373 ? 0 : $mask$016$i$i$i;
                $c$0$$i4$i$i = $374 | $c$017$i$i$i;
                $375 = $mask$016$i$i$i >>> 1;
                $376 = (($left$014$i$i$i) + -1)|0;
                $377 = ($376|0)==(0);
                do {
                 if ($377) {
                  $378 = $c$0$$i4$i$i&255;
                  HEAP8[$cur$015$i$i$i>>0] = $378;
                  $379 = $writer$sroa$0$3$i$i;
                  $380 = (($379) + ($writer$sroa$13$0$i$i)|0);
                  $381 = $380;
                  $c$2$i8$i$i = 0;$cur$1$i6$i$i = $380;$left$1$i5$i$i = $319;$mask$1$i7$i$i = 128;$writer$sroa$0$4$i$i = $381;
                 } else {
                  $382 = ($375|0)==(0);
                  if (!($382)) {
                   $c$2$i8$i$i = $c$0$$i4$i$i;$cur$1$i6$i$i = $cur$015$i$i$i;$left$1$i5$i$i = $376;$mask$1$i7$i$i = $375;$writer$sroa$0$4$i$i = $writer$sroa$0$3$i$i;
                   break;
                  }
                  $383 = $c$0$$i4$i$i&255;
                  HEAP8[$cur$015$i$i$i>>0] = $383;
                  $384 = ((($cur$015$i$i$i)) + 1|0);
                  $c$2$i8$i$i = 0;$cur$1$i6$i$i = $384;$left$1$i5$i$i = $376;$mask$1$i7$i$i = 128;$writer$sroa$0$4$i$i = $writer$sroa$0$3$i$i;
                 }
                } while(0);
                $385 = (($count$3$i$i$i) + -1)|0;
                $386 = ($count$3$i$i$i|0)<(2);
                $387 = $386&1;
                $388 = (($n$012$i$i$i) + -1)|0;
                $389 = ($388|0)==(0);
                if ($389) {
                 $c$2$i8$i$i$lcssa = $c$2$i8$i$i;$cur$1$i6$i$i$lcssa = $cur$1$i6$i$i;$mask$1$i7$i$i$lcssa = $mask$1$i7$i$i;
                 break;
                } else {
                 $$0218$i$i$i = $$3$i$i$i;$c$017$i$i$i = $c$2$i8$i$i;$count$010$i$i$i = $385;$counts$sroa$5$011$i$i$i = $counts$sroa$5$3$i$i$i;$cur$015$i$i$i = $cur$1$i6$i$i;$left$014$i$i$i = $left$1$i5$i$i;$mask$016$i$i$i = $mask$1$i7$i$i;$n$012$i$i$i = $388;$phase$09$i$i$i = $phase$3$i$i$i;$reload$013$i$i$i = $387;$writer$sroa$0$3$i$i = $writer$sroa$0$4$i$i;
                }
               }
               $390 = ($mask$1$i7$i$i$lcssa|0)==(128);
               if ($390) {
                break L78;
               }
               $391 = $c$2$i8$i$i$lcssa&255;
               HEAP8[$cur$1$i6$i$i$lcssa>>0] = $391;
               break L78;
               break;
              }
              case 2:  {
               $392 = ($324|0)==(0);
               if ($392) {
                break L78;
               }
               $393 = $writer$sroa$0$0$i$i;
               $$01$i$i$i = $p$2$i$i;$c$09$i$i$i = 0;$count$03$i$i$i = 0;$cur$07$i$i$i = $393;$left$06$i$i$i = $319;$mask$08$i$i$i = 128;$n$04$i$i$i = $324;$phase$02$i$i$i = 1;$reload$05$i$i$i = 1;$writer$sroa$0$5$i$i = $writer$sroa$0$0$i$i;
               while(1) {
                $394 = ($reload$05$i$i$i|0)==(0);
                L123: do {
                 if ($394) {
                  $$2$i$i$i = $$01$i$i$i;$count$2$i$i$i = $count$03$i$i$i;$phase$2$i$i$i = $phase$02$i$i$i;
                 } else {
                  $$1$i$i$i = $$01$i$i$i;$count$1$i$i$i = $count$03$i$i$i;$phase$1$i$i$i = $phase$02$i$i$i;
                  while(1) {
                   $395 = ($$1$i$i$i>>>0)<($313>>>0);
                   if (!($395)) {
                    $$2$i$i$i = $$1$i$i$i;$count$2$i$i$i = $count$1$i$i$i;$phase$2$i$i$i = $phase$1$i$i$i;
                    break L123;
                   }
                   $396 = ((($$1$i$i$i)) + 1|0);
                   $397 = HEAP8[$$1$i$i$i>>0]|0;
                   $398 = $397&255;
                   $399 = $phase$1$i$i$i ^ 1;
                   $400 = ($397<<24>>24)==(0);
                   if ($400) {
                    $$1$i$i$i = $396;$count$1$i$i$i = $398;$phase$1$i$i$i = $399;
                   } else {
                    $$2$i$i$i = $396;$count$2$i$i$i = $398;$phase$2$i$i$i = $399;
                    break;
                   }
                  }
                 }
                } while(0);
                $401 = ($phase$2$i$i$i|0)==(0);
                $402 = $401 ? 0 : $mask$08$i$i$i;
                $c$0$$i$i$i = $402 | $c$09$i$i$i;
                $403 = $mask$08$i$i$i >>> 1;
                $404 = (($left$06$i$i$i) + -1)|0;
                $405 = ($404|0)==(0);
                do {
                 if ($405) {
                  $406 = $c$0$$i$i$i&255;
                  HEAP8[$cur$07$i$i$i>>0] = $406;
                  $407 = $writer$sroa$0$5$i$i;
                  $408 = (($407) + ($writer$sroa$13$0$i$i)|0);
                  $409 = $408;
                  $c$2$i$i$i = 0;$cur$1$i$i$i = $408;$left$1$i$i$i = $319;$mask$1$i$i$i = 128;$writer$sroa$0$6$i$i = $409;
                 } else {
                  $410 = ($403|0)==(0);
                  if (!($410)) {
                   $c$2$i$i$i = $c$0$$i$i$i;$cur$1$i$i$i = $cur$07$i$i$i;$left$1$i$i$i = $404;$mask$1$i$i$i = $403;$writer$sroa$0$6$i$i = $writer$sroa$0$5$i$i;
                   break;
                  }
                  $411 = $c$0$$i$i$i&255;
                  HEAP8[$cur$07$i$i$i>>0] = $411;
                  $412 = ((($cur$07$i$i$i)) + 1|0);
                  $c$2$i$i$i = 0;$cur$1$i$i$i = $412;$left$1$i$i$i = $404;$mask$1$i$i$i = 128;$writer$sroa$0$6$i$i = $writer$sroa$0$5$i$i;
                 }
                } while(0);
                $413 = (($count$2$i$i$i) + -1)|0;
                $414 = ($count$2$i$i$i|0)<(2);
                $415 = $414&1;
                $416 = (($n$04$i$i$i) + -1)|0;
                $417 = ($416|0)==(0);
                if ($417) {
                 $c$2$i$i$i$lcssa = $c$2$i$i$i;$cur$1$i$i$i$lcssa = $cur$1$i$i$i;$mask$1$i$i$i$lcssa = $mask$1$i$i$i;
                 break;
                } else {
                 $$01$i$i$i = $$2$i$i$i;$c$09$i$i$i = $c$2$i$i$i;$count$03$i$i$i = $413;$cur$07$i$i$i = $cur$1$i$i$i;$left$06$i$i$i = $left$1$i$i$i;$mask$08$i$i$i = $mask$1$i$i$i;$n$04$i$i$i = $416;$phase$02$i$i$i = $phase$2$i$i$i;$reload$05$i$i$i = $415;$writer$sroa$0$5$i$i = $writer$sroa$0$6$i$i;
                }
               }
               $418 = ($mask$1$i$i$i$lcssa|0)==(128);
               if ($418) {
                break L78;
               }
               $419 = $c$2$i$i$i$lcssa&255;
               HEAP8[$cur$1$i$i$i$lcssa>>0] = $419;
               break L78;
               break;
              }
              default: {
               break L45;
              }
              }
             }
            } while(0);
            _FT_Stream_ExitFrame($12);
            $error$0 = 0;
            STACKTOP = sp;return ($error$0|0);
           }
          }
         }
        } while(0);
        _FT_Stream_ExitFrame($12);
        break L7;
       }
      } while(0);
      _FT_Stream_ExitFrame($12);
     }
    }
   }
  }
 } while(0);
 $420 = $load_flags & 16384;
 $421 = ($420|0)==(0);
 if (!($421)) {
  $error$0 = 6;
  STACKTOP = sp;return ($error$0|0);
 }
 $422 = ((($1)) + 416|0);
 $423 = HEAP32[$422>>2]|0;
 $424 = ((($pfrslot)) + 72|0);
 HEAP32[$424>>2] = 1869968492;
 $425 = ((($pfrslot)) + 110|0);
 HEAP16[$425>>1] = 0;
 HEAP16[$2>>1] = 0;
 $426 = ((($1)) + 188|0);
 $427 = HEAP32[$426>>2]|0;
 $428 = ((($pfrslot)) + 160|0);
 $429 = ((($1)) + 104|0);
 $430 = HEAP32[$429>>2]|0;
 $431 = (((($423) + ($gindex$<<4)|0)) + 12|0);
 $432 = HEAP32[$431>>2]|0;
 $433 = (((($423) + ($gindex$<<4)|0)) + 8|0);
 $434 = HEAP32[$433>>2]|0;
 $435 = ((($pfrslot)) + 188|0);
 $436 = HEAP32[$435>>2]|0;
 _FT_GlyphLoader_Rewind($436);
 $437 = ((($pfrslot)) + 176|0);
 HEAP32[$437>>2] = 0;
 $438 = (_pfr_glyph_load_rec($428,$430,$427,$432,$434)|0);
 $439 = ($438|0)==(0);
 if (!($439)) {
  $error$0 = $438;
  STACKTOP = sp;return ($error$0|0);
 }
 $440 = $load_flags & 1;
 $441 = HEAP32[$435>>2]|0;
 $442 = ((($441)) + 20|0);
 ;HEAP32[$2>>2]=HEAP32[$442>>2]|0;HEAP32[$2+4>>2]=HEAP32[$442+4>>2]|0;HEAP32[$2+8>>2]=HEAP32[$442+8>>2]|0;HEAP32[$2+12>>2]=HEAP32[$442+12>>2]|0;HEAP32[$2+16>>2]=HEAP32[$442+16>>2]|0;
 $443 = ((($pfrslot)) + 124|0);
 $444 = HEAP32[$443>>2]|0;
 $445 = $444 & -6;
 $446 = $445 | 4;
 HEAP32[$443>>2] = $446;
 $447 = ($pfrsize|0)==(0|0);
 if (!($447)) {
  $448 = ((($pfrsize)) + 14|0);
  $449 = HEAP16[$448>>1]|0;
  $450 = ($449&65535)<(24);
  if ($450) {
   $451 = $445 | 260;
   HEAP32[$443>>2] = $451;
  }
 }
 $452 = ((($pfrslot)) + 40|0);
 HEAP32[$452>>2] = 0;
 $453 = ((($pfrslot)) + 52|0);
 HEAP32[$453>>2] = 0;
 $454 = (((($423) + ($gindex$<<4)|0)) + 4|0);
 $455 = HEAP32[$454>>2]|0;
 $456 = ((($1)) + 304|0);
 $457 = HEAP32[$456>>2]|0;
 $458 = ((($1)) + 300|0);
 $459 = HEAP32[$458>>2]|0;
 $460 = ($457|0)==($459|0);
 if ($460) {
  $advance$0 = $455;
 } else {
  $461 = (_FT_MulDiv($455,$459,$457)|0);
  $advance$0 = $461;
 }
 $462 = ((($1)) + 324|0);
 $463 = HEAP32[$462>>2]|0;
 $464 = $463 & 1;
 $465 = ($464|0)==(0);
 if ($465) {
  HEAP32[$452>>2] = $advance$0;
  $$pre21 = HEAP32[$453>>2]|0;
  $467 = $advance$0;$469 = $$pre21;
 } else {
  HEAP32[$453>>2] = $advance$0;
  $$pre = HEAP32[$452>>2]|0;
  $467 = $$pre;$469 = $advance$0;
 }
 $466 = ((($pfrslot)) + 56|0);
 HEAP32[$466>>2] = $467;
 $468 = ((($pfrslot)) + 60|0);
 HEAP32[$468>>2] = $469;
 $470 = ((($pfrslot)) + 44|0);
 HEAP32[$470>>2] = 0;
 $471 = ((($pfrslot)) + 48|0);
 HEAP32[$471>>2] = 0;
 $472 = ($440|0)==(0);
 if ($472) {
  $473 = ((($pfrsize)) + 16|0);
  $474 = HEAP32[$473>>2]|0;
  $475 = ((($pfrsize)) + 20|0);
  $476 = HEAP32[$475>>2]|0;
  $477 = HEAP16[$425>>1]|0;
  $478 = ($477<<16>>16)>(0);
  if ($478) {
   $479 = ((($pfrslot)) + 112|0);
   $480 = HEAP32[$479>>2]|0;
   $n$011 = 0;$vec$012 = $480;
   while(1) {
    $481 = HEAP32[$vec$012>>2]|0;
    $482 = (_FT_MulFix($481,$474)|0);
    HEAP32[$vec$012>>2] = $482;
    $483 = ((($vec$012)) + 4|0);
    $484 = HEAP32[$483>>2]|0;
    $485 = (_FT_MulFix($484,$476)|0);
    HEAP32[$483>>2] = $485;
    $486 = (($n$011) + 1)|0;
    $487 = ((($vec$012)) + 8|0);
    $488 = HEAP16[$425>>1]|0;
    $489 = $488 << 16 >> 16;
    $490 = ($486|0)<($489|0);
    if ($490) {
     $n$011 = $486;$vec$012 = $487;
    } else {
     break;
    }
   }
   $$pre22 = HEAP32[$452>>2]|0;
   $491 = $$pre22;
  } else {
   $491 = $467;
  }
  $492 = (_FT_MulFix($491,$474)|0);
  HEAP32[$452>>2] = $492;
  $493 = HEAP32[$453>>2]|0;
  $494 = (_FT_MulFix($493,$476)|0);
  HEAP32[$453>>2] = $494;
 }
 _FT_Outline_Get_CBox($2,$cbox);
 $495 = ((($cbox)) + 8|0);
 $496 = HEAP32[$495>>2]|0;
 $497 = HEAP32[$cbox>>2]|0;
 $498 = (($496) - ($497))|0;
 $499 = ((($pfrslot)) + 24|0);
 HEAP32[$499>>2] = $498;
 $500 = ((($cbox)) + 12|0);
 $501 = HEAP32[$500>>2]|0;
 $502 = ((($cbox)) + 4|0);
 $503 = HEAP32[$502>>2]|0;
 $504 = (($501) - ($503))|0;
 $505 = ((($pfrslot)) + 28|0);
 HEAP32[$505>>2] = $504;
 $506 = ((($pfrslot)) + 32|0);
 HEAP32[$506>>2] = $497;
 $507 = ((($pfrslot)) + 36|0);
 HEAP32[$507>>2] = $503;
 $error$0 = 0;
 STACKTOP = sp;return ($error$0|0);
}
function _pfr_glyph_load_rec($glyph,$stream,$gps_offset,$offset,$size) {
 $glyph = $glyph|0;
 $stream = $stream|0;
 $gps_offset = $gps_offset|0;
 $offset = $offset|0;
 $size = $size|0;
 var $$01 = 0, $$021$i = 0, $$05$i = 0, $$10$i = 0, $$10$i$lcssa = 0, $$122$i = 0, $$1610$i = 0, $$2$i = 0, $$223$lcssa$i = 0, $$22342$i = 0, $$3$i = 0, $$3$i5 = 0, $$4$i = 0, $$4$i6 = 0, $$5$i = 0, $$5$i15 = 0, $$6$i = 0, $$6$i16 = 0, $$7$i = 0, $$7$lcssa68$i = 0;
 var $$7$lcssa69$i = 0, $$737$i = 0, $$8$i = 0, $$9$i = 0, $$lcssa = 0, $$lcssa57 = 0, $$phi$trans$insert$i = 0, $$phi$trans$insert$i3 = 0, $$phi$trans$insert51$i = 0, $$phi$trans$insert53$i = 0, $$phi$trans$insert55$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i$i = 0, $$pre$i17$i = 0, $$pre$i4 = 0, $$pre$i9$i = 0, $$pre$phi$iZ2D = 0, $$pre13$i = 0;
 var $$pre52$i = 0, $$pre54$i = 0, $$pre56$i = 0, $$pre57$i = 0, $$pre57$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0;
 var $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0;
 var $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0;
 var $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0;
 var $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0;
 var $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0;
 var $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0;
 var $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0;
 var $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0;
 var $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0;
 var $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0;
 var $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0;
 var $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0;
 var $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0;
 var $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0;
 var $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0;
 var $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0;
 var $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0;
 var $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0;
 var $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0;
 var $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0;
 var $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0;
 var $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0;
 var $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0;
 var $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0;
 var $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0;
 var $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0;
 var $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $args_count$140$i = 0, $args_count$2$i = 0, $args_format$139$i = 0, $args_format$2$i = 0, $cur$038$i = 0, $error$0$i$i = 0, $error$0$i11$i = 0, $error$0$i20$i = 0, $error$i = 0, $error$i2 = 0, $exitcond = 0, $exitcond41 = 0;
 var $first$0$i$i$i = 0, $first$0$i$i13$i = 0, $glyph$idx$val$i = 0, $glyph$idx$val$pre$i$i = 0, $glyph$idx$val4$i$i = 0, $glyph$idx1$val$pre$i$i = 0, $glyph$idx6$val$i = 0, $glyph$idx7$val$i = 0, $glyph$idx8$val$i = 0, $i$011$i = 0, $i$026 = 0, $i$043$i = 0, $i$128 = 0, $last$1$i$i$i = 0, $last$1$i$i14$i = 0, $mask$044$i = 0, $mask$1$i = 0, $n$031 = 0, $n$041$i = 0, $num_items$05$us$i$i$i = 0;
 var $num_items$05$us$i$i$i8 = 0, $or$cond$i = 0, $p$04$us$i$i$i = 0, $p$04$us$i$i$i9 = 0, $p$1$i$i$i = 0, $p$1$i$i$i13 = 0, $phitmp$i$i = 0, $pos$i = 0, $sext$i = 0, $sext$i$i$i = 0, $sext$i$i15$i = 0, $sext$i19 = 0, $sext1$i$i$i = 0, $sext1$i$i12$i = 0, $sext2$i = 0, $sext3$i = 0, $sext4$i = 0, $sext4$i18 = 0, $sext5$i = 0, $sext59$i = 0;
 var $sext60$i = 0, $subglyph$012$i = 0, $vec$027 = 0, $vec1$029 = 0, $x$045$i = 0, $x$1$i = 0, $x_count$0$i = 0, $x_count$1$i = 0, $x_pos$0$i = 0, $y_count$0$i = 0, $y_pos$0$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i2 = sp + 36|0;
 $pos$i = sp;
 $error$i = sp + 32|0;
 $0 = (($offset) + ($gps_offset))|0;
 $1 = (_FT_Stream_Seek($stream,$0)|0);
 $2 = ($1|0)==(0);
 if (!($2)) {
  $$01 = $1;
  STACKTOP = sp;return ($$01|0);
 }
 $3 = (_FT_Stream_EnterFrame($stream,$size)|0);
 $4 = ($3|0)==(0);
 if (!($4)) {
  $$01 = $3;
  STACKTOP = sp;return ($$01|0);
 }
 $5 = ((($stream)) + 32|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + ($size)|0);
 $8 = ($size|0)==(0);
 if (!($8)) {
  $9 = HEAP8[$6>>0]|0;
  $10 = ($9<<24>>24)<(0);
  if ($10) {
   $11 = ((($glyph)) + 28|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ((($glyph)) + 16|0);
   $14 = HEAP32[$13>>2]|0;
   HEAP32[$error$i>>2] = 0;
   $15 = HEAP32[$12>>2]|0;
   $16 = ((($6)) + 1|0);
   $17 = ($size|0)<(1);
   L10: do {
    if ($17) {
     label = 52;
    } else {
     $18 = HEAP8[$6>>0]|0;
     $19 = $18&255;
     $20 = $19 & 128;
     $21 = ($20|0)==(0);
     if ($21) {
      label = 52;
     } else {
      $22 = $19 & 63;
      $23 = $19 & 8;
      $24 = ($23|0)==(0);
      L13: do {
       if ($24) {
        $$05$i = $16;
       } else {
        $25 = ((($6)) + 2|0);
        $26 = ($size|0)<(2);
        L15: do {
         if (!($26)) {
          $27 = HEAP8[$16>>0]|0;
          $28 = ($27<<24>>24)==(0);
          if ($28) {
           $p$1$i$i$i = $25;
          } else {
           $29 = $27&255;
           $num_items$05$us$i$i$i = $29;$p$04$us$i$i$i = $25;
           while(1) {
            $30 = ((($p$04$us$i$i$i)) + 2|0);
            $31 = ($30>>>0)>($7>>>0);
            if ($31) {
             break L15;
            }
            $32 = HEAP8[$p$04$us$i$i$i>>0]|0;
            $33 = $32&255;
            $34 = (($30) + ($33)|0);
            $35 = ($34>>>0)>($7>>>0);
            if ($35) {
             break L15;
            }
            $36 = (($num_items$05$us$i$i$i) + -1)|0;
            $37 = ($36|0)==(0);
            if ($37) {
             $p$1$i$i$i = $34;
             break;
            } else {
             $num_items$05$us$i$i$i = $36;$p$04$us$i$i$i = $34;
            }
           }
          }
          HEAP32[$error$i>>2] = 0;
          $$05$i = $p$1$i$i$i;
          break L13;
         }
        } while(0);
        HEAP32[$error$i>>2] = 8;
        $165 = 8;
        break L10;
       }
      } while(0);
      $38 = (($14) + ($22))|0;
      $39 = ((($glyph)) + 20|0);
      $40 = HEAP32[$39>>2]|0;
      $41 = ($38>>>0)>($40>>>0);
      if ($41) {
       $42 = (($38) + 3)|0;
       $43 = $42 & -4;
       $44 = ($43>>>0)>(64);
       if ($44) {
        HEAP32[$error$i>>2] = 8;
        $165 = 8;
        break;
       }
       $45 = ((($glyph)) + 24|0);
       $46 = HEAP32[$45>>2]|0;
       $47 = (_ft_mem_realloc($15,24,$40,$43,$46,$error$i)|0);
       HEAP32[$45>>2] = $47;
       $48 = HEAP32[$error$i>>2]|0;
       $49 = ($48|0)==(0);
       if (!($49)) {
        $165 = $48;
        break;
       }
       HEAP32[$39>>2] = $43;
       $52 = $47;
      } else {
       $$phi$trans$insert$i = ((($glyph)) + 24|0);
       $$pre$i = HEAP32[$$phi$trans$insert$i>>2]|0;
       $52 = $$pre$i;
      }
      $50 = ($22|0)==(0);
      if ($50) {
       $165 = 0;
      } else {
       $51 = (($52) + (($14*24)|0)|0);
       $$1610$i = $$05$i;$i$011$i = 0;$subglyph$012$i = $51;
       while(1) {
        $53 = ((($$1610$i)) + 1|0);
        $54 = ($53>>>0)>($7>>>0);
        if ($54) {
         label = 52;
         break L10;
        }
        $55 = HEAP8[$$1610$i>>0]|0;
        $56 = $55&255;
        HEAP32[$subglyph$012$i>>2] = 65536;
        $57 = $56 & 16;
        $58 = ($57|0)==(0);
        if ($58) {
         $$2$i = $53;
        } else {
         $59 = ((($$1610$i)) + 3|0);
         $60 = ($59>>>0)>($7>>>0);
         if ($60) {
          label = 52;
          break L10;
         }
         $61 = HEAP8[$53>>0]|0;
         $62 = $61&255;
         $63 = $62 << 8;
         $64 = ((($$1610$i)) + 2|0);
         $65 = HEAP8[$64>>0]|0;
         $66 = $65&255;
         $67 = $63 | $66;
         $sext4$i = $67 << 16;
         $68 = $sext4$i >> 12;
         HEAP32[$subglyph$012$i>>2] = $68;
         $$2$i = $59;
        }
        $69 = ((($subglyph$012$i)) + 4|0);
        HEAP32[$69>>2] = 65536;
        $70 = $56 & 32;
        $71 = ($70|0)==(0);
        if ($71) {
         $$3$i = $$2$i;
        } else {
         $72 = ((($$2$i)) + 2|0);
         $73 = ($72>>>0)>($7>>>0);
         if ($73) {
          label = 52;
          break L10;
         }
         $74 = HEAP8[$$2$i>>0]|0;
         $75 = $74&255;
         $76 = $75 << 8;
         $77 = ((($$2$i)) + 1|0);
         $78 = HEAP8[$77>>0]|0;
         $79 = $78&255;
         $80 = $76 | $79;
         $sext3$i = $80 << 16;
         $81 = $sext3$i >> 12;
         HEAP32[$69>>2] = $81;
         $$3$i = $72;
        }
        $82 = $56 & 3;
        switch ($82|0) {
        case 1:  {
         $83 = ((($$3$i)) + 2|0);
         $84 = ($83>>>0)>($7>>>0);
         if ($84) {
          label = 52;
          break L10;
         }
         $85 = HEAP8[$$3$i>>0]|0;
         $86 = $85&255;
         $87 = $86 << 8;
         $88 = ((($$3$i)) + 1|0);
         $89 = HEAP8[$88>>0]|0;
         $90 = $89&255;
         $91 = $87 | $90;
         $sext2$i = $91 << 16;
         $92 = $sext2$i >> 16;
         $$4$i = $83;$x_pos$0$i = $92;
         break;
        }
        case 2:  {
         $93 = ((($$3$i)) + 1|0);
         $94 = ($93>>>0)>($7>>>0);
         if ($94) {
          label = 52;
          break L10;
         }
         $95 = HEAP8[$$3$i>>0]|0;
         $96 = $95 << 24 >> 24;
         $$4$i = $93;$x_pos$0$i = $96;
         break;
        }
        default: {
         $$4$i = $$3$i;$x_pos$0$i = 0;
        }
        }
        $97 = $56 >>> 2;
        $98 = $97 & 3;
        switch ($98|0) {
        case 1:  {
         $99 = ((($$4$i)) + 2|0);
         $100 = ($99>>>0)>($7>>>0);
         if ($100) {
          label = 52;
          break L10;
         }
         $101 = HEAP8[$$4$i>>0]|0;
         $102 = $101&255;
         $103 = $102 << 8;
         $104 = ((($$4$i)) + 1|0);
         $105 = HEAP8[$104>>0]|0;
         $106 = $105&255;
         $107 = $103 | $106;
         $sext$i = $107 << 16;
         $108 = $sext$i >> 16;
         $$5$i = $99;$y_pos$0$i = $108;
         break;
        }
        case 2:  {
         $109 = ((($$4$i)) + 1|0);
         $110 = ($109>>>0)>($7>>>0);
         if ($110) {
          label = 52;
          break L10;
         }
         $111 = HEAP8[$$4$i>>0]|0;
         $112 = $111 << 24 >> 24;
         $$5$i = $109;$y_pos$0$i = $112;
         break;
        }
        default: {
         $$5$i = $$4$i;$y_pos$0$i = 0;
        }
        }
        $113 = ((($subglyph$012$i)) + 8|0);
        HEAP32[$113>>2] = $x_pos$0$i;
        $114 = ((($subglyph$012$i)) + 12|0);
        HEAP32[$114>>2] = $y_pos$0$i;
        $115 = $56 & 64;
        $116 = ($115|0)==(0);
        if ($116) {
         $127 = ((($$5$i)) + 1|0);
         $128 = ($127>>>0)>($7>>>0);
         if ($128) {
          label = 52;
          break L10;
         }
         $129 = HEAP8[$$5$i>>0]|0;
         $130 = $129&255;
         $131 = ((($subglyph$012$i)) + 20|0);
         HEAP32[$131>>2] = $130;
         $$6$i = $127;
        } else {
         $117 = ((($$5$i)) + 2|0);
         $118 = ($117>>>0)>($7>>>0);
         if ($118) {
          label = 52;
          break L10;
         }
         $119 = HEAP8[$$5$i>>0]|0;
         $120 = $119&255;
         $121 = $120 << 8;
         $122 = ((($$5$i)) + 1|0);
         $123 = HEAP8[$122>>0]|0;
         $124 = $123&255;
         $125 = $121 | $124;
         $126 = ((($subglyph$012$i)) + 20|0);
         HEAP32[$126>>2] = $125;
         $$6$i = $117;
        }
        $132 = $56 & 128;
        $133 = ($132|0)==(0);
        if ($133) {
         $149 = ((($$6$i)) + 2|0);
         $150 = ($149>>>0)>($7>>>0);
         if ($150) {
          label = 52;
          break L10;
         }
         $151 = HEAP8[$$6$i>>0]|0;
         $152 = $151&255;
         $153 = $152 << 8;
         $154 = ((($$6$i)) + 1|0);
         $155 = HEAP8[$154>>0]|0;
         $156 = $155&255;
         $157 = $153 | $156;
         $158 = ((($subglyph$012$i)) + 16|0);
         HEAP32[$158>>2] = $157;
         $$7$i = $149;
        } else {
         $134 = ((($$6$i)) + 3|0);
         $135 = ($134>>>0)>($7>>>0);
         if ($135) {
          label = 52;
          break L10;
         }
         $136 = HEAP8[$$6$i>>0]|0;
         $137 = $136&255;
         $138 = $137 << 16;
         $139 = ((($$6$i)) + 1|0);
         $140 = HEAP8[$139>>0]|0;
         $141 = $140&255;
         $142 = $141 << 8;
         $143 = $142 | $138;
         $144 = ((($$6$i)) + 2|0);
         $145 = HEAP8[$144>>0]|0;
         $146 = $145&255;
         $147 = $143 | $146;
         $148 = ((($subglyph$012$i)) + 16|0);
         HEAP32[$148>>2] = $147;
         $$7$i = $134;
        }
        $159 = HEAP32[$13>>2]|0;
        $160 = (($159) + 1)|0;
        HEAP32[$13>>2] = $160;
        $161 = (($i$011$i) + 1)|0;
        $162 = ((($subglyph$012$i)) + 24|0);
        $163 = ($161>>>0)<($22>>>0);
        if ($163) {
         $$1610$i = $$7$i;$i$011$i = $161;$subglyph$012$i = $162;
        } else {
         break;
        }
       }
       $$pre13$i = HEAP32[$error$i>>2]|0;
       $165 = $$pre13$i;
      }
     }
    }
   } while(0);
   if ((label|0) == 52) {
    HEAP32[$error$i>>2] = 8;
    $165 = 8;
   }
   _FT_Stream_ExitFrame($stream);
   $164 = ($165|0)==(0);
   if (!($164)) {
    $$01 = $165;
    STACKTOP = sp;return ($$01|0);
   }
   $166 = HEAP32[$13>>2]|0;
   $167 = (($166) - ($14))|0;
   $168 = ((($glyph)) + 24|0);
   $169 = ($166|0)==($14|0);
   if ($169) {
    $$01 = 0;
    STACKTOP = sp;return ($$01|0);
   }
   $170 = ((($12)) + 22|0);
   $171 = ((($12)) + 24|0);
   $n$031 = 0;
   while(1) {
    $172 = HEAP32[$168>>2]|0;
    $173 = (($172) + (($14*24)|0)|0);
    $174 = (($173) + (($n$031*24)|0)|0);
    $175 = HEAP16[$170>>1]|0;
    $176 = $175 << 16 >> 16;
    $177 = ((($174)) + 16|0);
    $178 = HEAP32[$177>>2]|0;
    $179 = ((($174)) + 20|0);
    $180 = HEAP32[$179>>2]|0;
    $181 = (_pfr_glyph_load_rec($glyph,$stream,$gps_offset,$178,$180)|0);
    $182 = ($181|0)==(0);
    if (!($182)) {
     $$01 = $181;
     label = 185;
     break;
    }
    $183 = HEAP32[$168>>2]|0;
    $184 = (($183) + (($14*24)|0)|0);
    $185 = (($184) + (($n$031*24)|0)|0);
    $186 = HEAP16[$170>>1]|0;
    $187 = $186 << 16 >> 16;
    $188 = (($187) - ($176))|0;
    $189 = HEAP32[$185>>2]|0;
    $190 = ($189|0)==(65536);
    if ($190) {
     $191 = ((($185)) + 4|0);
     $192 = HEAP32[$191>>2]|0;
     $193 = ($192|0)==(65536);
     if ($193) {
      $213 = ($188|0)>(0);
      if ($213) {
       $214 = HEAP32[$171>>2]|0;
       $215 = (($214) + ($176<<3)|0);
       $216 = ((($185)) + 8|0);
       $217 = HEAP32[$216>>2]|0;
       $218 = ((($185)) + 12|0);
       $219 = HEAP32[$218>>2]|0;
       $i$128 = 0;$vec1$029 = $215;
       while(1) {
        $220 = HEAP32[$vec1$029>>2]|0;
        $221 = (($220) + ($217))|0;
        HEAP32[$vec1$029>>2] = $221;
        $222 = ((($vec1$029)) + 4|0);
        $223 = HEAP32[$222>>2]|0;
        $224 = (($223) + ($219))|0;
        HEAP32[$222>>2] = $224;
        $225 = (($i$128) + 1)|0;
        $226 = ((($vec1$029)) + 8|0);
        $exitcond41 = ($225|0)==($188|0);
        if ($exitcond41) {
         break;
        } else {
         $i$128 = $225;$vec1$029 = $226;
        }
       }
      }
     } else {
      label = 60;
     }
    } else {
     label = 60;
    }
    L89: do {
     if ((label|0) == 60) {
      label = 0;
      $194 = ($188|0)>(0);
      if ($194) {
       $195 = HEAP32[$171>>2]|0;
       $196 = (($195) + ($176<<3)|0);
       $197 = ((($185)) + 8|0);
       $198 = ((($185)) + 4|0);
       $199 = ((($185)) + 12|0);
       $201 = $189;$i$026 = 0;$vec$027 = $196;
       while(1) {
        $200 = HEAP32[$vec$027>>2]|0;
        $202 = (_FT_MulFix($200,$201)|0);
        $203 = HEAP32[$197>>2]|0;
        $204 = (($203) + ($202))|0;
        HEAP32[$vec$027>>2] = $204;
        $205 = ((($vec$027)) + 4|0);
        $206 = HEAP32[$205>>2]|0;
        $207 = HEAP32[$198>>2]|0;
        $208 = (_FT_MulFix($206,$207)|0);
        $209 = HEAP32[$199>>2]|0;
        $210 = (($209) + ($208))|0;
        HEAP32[$205>>2] = $210;
        $211 = (($i$026) + 1)|0;
        $exitcond = ($211|0)==($188|0);
        if ($exitcond) {
         break L89;
        }
        $212 = ((($vec$027)) + 8|0);
        $$pre = HEAP32[$185>>2]|0;
        $201 = $$pre;$i$026 = $211;$vec$027 = $212;
       }
      }
     }
    } while(0);
    $227 = (($n$031) + 1)|0;
    $228 = ($227>>>0)<($167>>>0);
    if ($228) {
     $n$031 = $227;
    } else {
     $$01 = 0;
     label = 185;
     break;
    }
   }
   if ((label|0) == 185) {
    STACKTOP = sp;return ($$01|0);
   }
  }
 }
 HEAP32[$error$i2>>2] = 0;
 $229 = ((($glyph)) + 28|0);
 $230 = HEAP32[$229>>2]|0;
 $231 = HEAP32[$230>>2]|0;
 $232 = ((($6)) + 1|0);
 $233 = ($size|0)<(1);
 L100: do {
  if ($233) {
   label = 183;
  } else {
   $234 = HEAP8[$6>>0]|0;
   $235 = $234&255;
   $236 = $235 & 128;
   $237 = ($236|0)==(0);
   if ($237) {
    $238 = $235 & 4;
    $239 = ($238|0)==(0);
    if ($239) {
     $246 = $235 & 2;
     $247 = ($246|0)==(0);
     if ($247) {
      $$021$i = $232;$x_count$0$i = 0;
     } else {
      $248 = ($size|0)<(2);
      if ($248) {
       label = 183;
       break;
      }
      $249 = ((($6)) + 2|0);
      $250 = HEAP8[$232>>0]|0;
      $251 = $250&255;
      $$021$i = $249;$x_count$0$i = $251;
     }
     $252 = $235 & 1;
     $253 = ($252|0)==(0);
     if ($253) {
      $$122$i = $$021$i;$x_count$1$i = $x_count$0$i;$y_count$0$i = 0;
     } else {
      $254 = ((($$021$i)) + 1|0);
      $255 = ($254>>>0)>($7>>>0);
      if ($255) {
       label = 183;
       break;
      }
      $256 = HEAP8[$$021$i>>0]|0;
      $257 = $256&255;
      $$122$i = $254;$x_count$1$i = $x_count$0$i;$y_count$0$i = $257;
     }
    } else {
     $240 = ($size|0)<(2);
     if ($240) {
      label = 183;
      break;
     }
     $241 = ((($6)) + 2|0);
     $242 = HEAP8[$232>>0]|0;
     $243 = $242&255;
     $244 = $243 & 15;
     $245 = $243 >>> 4;
     $$122$i = $241;$x_count$1$i = $244;$y_count$0$i = $245;
    }
    $258 = (($x_count$1$i) + ($y_count$0$i))|0;
    $259 = ((($glyph)) + 4|0);
    $260 = HEAP32[$259>>2]|0;
    $261 = ($258>>>0)>($260>>>0);
    if ($261) {
     $262 = (($258) + 7)|0;
     $263 = $262 & -8;
     $264 = ((($glyph)) + 8|0);
     $265 = HEAP32[$264>>2]|0;
     $266 = (_ft_mem_realloc($231,4,$260,$263,$265,$error$i2)|0);
     HEAP32[$264>>2] = $266;
     $267 = HEAP32[$error$i2>>2]|0;
     $268 = ($267|0)==(0);
     if (!($268)) {
      $678 = $267;
      break;
     }
     HEAP32[$259>>2] = $263;
     $$pre$phi$iZ2D = $264;$270 = $266;
    } else {
     $$phi$trans$insert$i3 = ((($glyph)) + 8|0);
     $$pre$i4 = HEAP32[$$phi$trans$insert$i3>>2]|0;
     $$pre$phi$iZ2D = $$phi$trans$insert$i3;$270 = $$pre$i4;
    }
    $269 = (($270) + ($x_count$1$i<<2)|0);
    $271 = ((($glyph)) + 12|0);
    HEAP32[$271>>2] = $269;
    $272 = ($258|0)==(0);
    if ($272) {
     $$223$lcssa$i = $$122$i;
    } else {
     $$22342$i = $$122$i;$i$043$i = 0;$mask$044$i = 0;$x$045$i = 0;
     while(1) {
      $273 = $i$043$i & 7;
      $274 = ($273|0)==(0);
      if ($274) {
       $275 = ((($$22342$i)) + 1|0);
       $276 = ($275>>>0)>($7>>>0);
       if ($276) {
        label = 183;
        break L100;
       }
       $277 = HEAP8[$$22342$i>>0]|0;
       $278 = $277&255;
       $$3$i5 = $275;$mask$1$i = $278;
      } else {
       $$3$i5 = $$22342$i;$mask$1$i = $mask$044$i;
      }
      $279 = $mask$1$i & 1;
      $280 = ($279|0)==(0);
      if ($280) {
       $291 = ((($$3$i5)) + 1|0);
       $292 = ($291>>>0)>($7>>>0);
       if ($292) {
        label = 183;
        break L100;
       }
       $293 = HEAP8[$$3$i5>>0]|0;
       $294 = $293&255;
       $295 = (($294) + ($x$045$i))|0;
       $$4$i6 = $291;$x$1$i = $295;
      } else {
       $281 = ((($$3$i5)) + 2|0);
       $282 = ($281>>>0)>($7>>>0);
       if ($282) {
        label = 183;
        break L100;
       }
       $283 = HEAP8[$$3$i5>>0]|0;
       $284 = $283&255;
       $285 = $284 << 8;
       $286 = ((($$3$i5)) + 1|0);
       $287 = HEAP8[$286>>0]|0;
       $288 = $287&255;
       $289 = $285 | $288;
       $sext5$i = $289 << 16;
       $290 = $sext5$i >> 16;
       $$4$i6 = $281;$x$1$i = $290;
      }
      $296 = (($270) + ($i$043$i<<2)|0);
      HEAP32[$296>>2] = $x$1$i;
      $297 = $mask$1$i >>> 1;
      $298 = (($i$043$i) + 1)|0;
      $299 = ($298>>>0)<($258>>>0);
      if ($299) {
       $$22342$i = $$4$i6;$i$043$i = $298;$mask$044$i = $297;$x$045$i = $x$1$i;
      } else {
       $$223$lcssa$i = $$4$i6;
       break;
      }
     }
    }
    $300 = $235 & 8;
    $301 = ($300|0)==(0);
    L133: do {
     if ($301) {
      $$5$i15 = $$223$lcssa$i;
     } else {
      $302 = ((($$223$lcssa$i)) + 1|0);
      $303 = ($302>>>0)>($7>>>0);
      L135: do {
       if (!($303)) {
        $304 = HEAP8[$$223$lcssa$i>>0]|0;
        $305 = ($304<<24>>24)==(0);
        if ($305) {
         $p$1$i$i$i13 = $302;
        } else {
         $306 = $304&255;
         $num_items$05$us$i$i$i8 = $306;$p$04$us$i$i$i9 = $302;
         while(1) {
          $307 = ((($p$04$us$i$i$i9)) + 2|0);
          $308 = ($307>>>0)>($7>>>0);
          if ($308) {
           break L135;
          }
          $309 = HEAP8[$p$04$us$i$i$i9>>0]|0;
          $310 = $309&255;
          $311 = (($307) + ($310)|0);
          $312 = ($311>>>0)>($7>>>0);
          if ($312) {
           break L135;
          }
          $313 = (($num_items$05$us$i$i$i8) + -1)|0;
          $314 = ($313|0)==(0);
          if ($314) {
           $p$1$i$i$i13 = $311;
           break;
          } else {
           $num_items$05$us$i$i$i8 = $313;$p$04$us$i$i$i9 = $311;
          }
         }
        }
        HEAP32[$error$i2>>2] = 0;
        $$5$i15 = $p$1$i$i$i13;
        break L133;
       }
      } while(0);
      HEAP32[$error$i2>>2] = 8;
      $678 = 8;
      break L100;
     }
    } while(0);
    $315 = ((($glyph)) + 32|0);
    HEAP8[$315>>0] = 0;
    $316 = ((($pos$i)) + 4|0);
    HEAP32[$316>>2] = 0;
    HEAP32[$pos$i>>2] = 0;
    $317 = ((($pos$i)) + 24|0);
    $318 = $pos$i;
    $319 = $318;
    $320 = HEAP32[$319>>2]|0;
    $321 = (($318) + 4)|0;
    $322 = $321;
    $323 = HEAP32[$322>>2]|0;
    $324 = $317;
    $325 = $324;
    HEAP32[$325>>2] = $320;
    $326 = (($324) + 4)|0;
    $327 = $326;
    HEAP32[$327>>2] = $323;
    $328 = ((($pos$i)) + 8|0);
    $329 = ((($pos$i)) + 16|0);
    $$6$i16 = $$5$i15;$340 = $323;$352 = $320;
    L146: while(1) {
     $330 = ((($$6$i16)) + 1|0);
     $331 = ($330>>>0)>($7>>>0);
     if ($331) {
      label = 180;
      break;
     }
     $332 = HEAP8[$$6$i16>>0]|0;
     $333 = $332&255;
     $334 = $333 & 15;
     $335 = $333 >>> 4;
     switch ($335|0) {
     case 0:  {
      label = 136;
      break L146;
      break;
     }
     case 5: case 4: case 1:  {
      $$737$i = $330;$390 = $352;$418 = $340;$args_count$140$i = 1;$args_format$139$i = $334;$cur$038$i = $pos$i;$n$041$i = 0;
      label = 112;
      break;
     }
     case 2:  {
      $336 = ($334>>>0)<($x_count$1$i>>>0);
      if (!($336)) {
       label = 180;
       break L146;
      }
      $337 = HEAP32[$$pre$phi$iZ2D>>2]|0;
      $338 = (($337) + ($334<<2)|0);
      $339 = HEAP32[$338>>2]|0;
      HEAP32[$pos$i>>2] = $339;
      HEAP32[$316>>2] = $340;
      $341 = $pos$i;
      $342 = $341;
      $343 = HEAP32[$342>>2]|0;
      $344 = (($341) + 4)|0;
      $345 = $344;
      $346 = HEAP32[$345>>2]|0;
      $347 = $317;
      $348 = $347;
      HEAP32[$348>>2] = $343;
      $349 = (($347) + 4)|0;
      $350 = $349;
      HEAP32[$350>>2] = $346;
      $$7$lcssa69$i = $330;$679 = $343;$680 = $346;
      label = 147;
      break;
     }
     case 3:  {
      $351 = ($334>>>0)<($y_count$0$i>>>0);
      if (!($351)) {
       label = 180;
       break L146;
      }
      HEAP32[$pos$i>>2] = $352;
      $353 = HEAP32[$271>>2]|0;
      $354 = (($353) + ($334<<2)|0);
      $355 = HEAP32[$354>>2]|0;
      HEAP32[$316>>2] = $355;
      $356 = $pos$i;
      $357 = $356;
      $358 = HEAP32[$357>>2]|0;
      $359 = (($356) + 4)|0;
      $360 = $359;
      $361 = HEAP32[$360>>2]|0;
      $362 = $317;
      $363 = $362;
      HEAP32[$363>>2] = $358;
      $364 = (($362) + 4)|0;
      $365 = $364;
      HEAP32[$365>>2] = $361;
      $$7$lcssa69$i = $330;$679 = $358;$680 = $361;
      label = 147;
      break;
     }
     case 6:  {
      $$737$i = $330;$390 = $352;$418 = $340;$args_count$140$i = 3;$args_format$139$i = 2958;$cur$038$i = $pos$i;$n$041$i = 0;
      label = 112;
      break;
     }
     case 7:  {
      $$737$i = $330;$390 = $352;$418 = $340;$args_count$140$i = 3;$args_format$139$i = 3627;$cur$038$i = $pos$i;$n$041$i = 0;
      label = 112;
      break;
     }
     default: {
      $$737$i = $330;$390 = $352;$418 = $340;$args_count$140$i = 4;$args_format$139$i = $334;$cur$038$i = $pos$i;$n$041$i = 0;
      label = 112;
     }
     }
     L157: do {
      if ((label|0) == 112) {
       while(1) {
        label = 0;
        $366 = $args_format$139$i & 3;
        switch ($366|0) {
        case 0:  {
         $367 = ((($$737$i)) + 1|0);
         $368 = ($367>>>0)>($7>>>0);
         if ($368) {
          label = 182;
          break L146;
         }
         $369 = HEAP8[$$737$i>>0]|0;
         $370 = $369&255;
         $371 = ($370>>>0)<($x_count$1$i>>>0);
         if (!($371)) {
          label = 182;
          break L146;
         }
         $372 = HEAP32[$$pre$phi$iZ2D>>2]|0;
         $373 = (($372) + ($370<<2)|0);
         $374 = HEAP32[$373>>2]|0;
         HEAP32[$cur$038$i>>2] = $374;
         $$8$i = $367;
         break;
        }
        case 1:  {
         $375 = ((($$737$i)) + 2|0);
         $376 = ($375>>>0)>($7>>>0);
         if ($376) {
          label = 182;
          break L146;
         }
         $377 = HEAP8[$$737$i>>0]|0;
         $378 = $377&255;
         $379 = $378 << 8;
         $380 = ((($$737$i)) + 1|0);
         $381 = HEAP8[$380>>0]|0;
         $382 = $381&255;
         $383 = $379 | $382;
         $sext4$i18 = $383 << 16;
         $384 = $sext4$i18 >> 16;
         HEAP32[$cur$038$i>>2] = $384;
         $$8$i = $375;
         break;
        }
        case 2:  {
         $385 = ((($$737$i)) + 1|0);
         $386 = ($385>>>0)>($7>>>0);
         if ($386) {
          label = 182;
          break L146;
         }
         $387 = HEAP8[$$737$i>>0]|0;
         $388 = $387 << 24 >> 24;
         $389 = (($388) + ($390))|0;
         HEAP32[$cur$038$i>>2] = $389;
         $$8$i = $385;
         break;
        }
        default: {
         HEAP32[$cur$038$i>>2] = $390;
         $$8$i = $$737$i;
        }
        }
        $391 = $args_format$139$i >>> 2;
        $392 = $391 & 3;
        switch ($392|0) {
        case 0:  {
         $393 = ((($$8$i)) + 1|0);
         $394 = ($393>>>0)>($7>>>0);
         if ($394) {
          label = 182;
          break L146;
         }
         $395 = HEAP8[$$8$i>>0]|0;
         $396 = $395&255;
         $397 = ($396>>>0)<($y_count$0$i>>>0);
         if (!($397)) {
          label = 182;
          break L146;
         }
         $398 = HEAP32[$271>>2]|0;
         $399 = (($398) + ($396<<2)|0);
         $400 = HEAP32[$399>>2]|0;
         $401 = ((($cur$038$i)) + 4|0);
         HEAP32[$401>>2] = $400;
         $$9$i = $393;
         break;
        }
        case 1:  {
         $402 = ((($$8$i)) + 2|0);
         $403 = ($402>>>0)>($7>>>0);
         if ($403) {
          label = 182;
          break L146;
         }
         $404 = HEAP8[$$8$i>>0]|0;
         $405 = $404&255;
         $406 = $405 << 8;
         $407 = ((($$8$i)) + 1|0);
         $408 = HEAP8[$407>>0]|0;
         $409 = $408&255;
         $410 = $406 | $409;
         $sext$i19 = $410 << 16;
         $411 = $sext$i19 >> 16;
         $412 = ((($cur$038$i)) + 4|0);
         HEAP32[$412>>2] = $411;
         $$9$i = $402;
         break;
        }
        case 2:  {
         $413 = ((($$8$i)) + 1|0);
         $414 = ($413>>>0)>($7>>>0);
         if ($414) {
          label = 182;
          break L146;
         }
         $415 = HEAP8[$$8$i>>0]|0;
         $416 = $415 << 24 >> 24;
         $417 = (($416) + ($418))|0;
         $419 = ((($cur$038$i)) + 4|0);
         HEAP32[$419>>2] = $417;
         $$9$i = $413;
         break;
        }
        default: {
         $420 = ((($cur$038$i)) + 4|0);
         HEAP32[$420>>2] = $418;
         $$9$i = $$8$i;
        }
        }
        $421 = ($n$041$i|0)==(0);
        $422 = ($args_count$140$i|0)==(4);
        $or$cond$i = $421 & $422;
        if ($or$cond$i) {
         $423 = ((($$9$i)) + 1|0);
         $424 = ($423>>>0)>($7>>>0);
         if ($424) {
          label = 182;
          break L146;
         }
         $425 = HEAP8[$$9$i>>0]|0;
         $426 = $425&255;
         $$10$i = $423;$args_count$2$i = 3;$args_format$2$i = $426;
        } else {
         $427 = $args_format$139$i >>> 4;
         $$10$i = $$9$i;$args_count$2$i = $args_count$140$i;$args_format$2$i = $427;
        }
        $428 = $cur$038$i;
        $429 = $428;
        $430 = HEAP32[$429>>2]|0;
        $431 = (($428) + 4)|0;
        $432 = $431;
        $433 = HEAP32[$432>>2]|0;
        $434 = $317;
        $435 = $434;
        HEAP32[$435>>2] = $430;
        $436 = (($434) + 4)|0;
        $437 = $436;
        HEAP32[$437>>2] = $433;
        $438 = ((($cur$038$i)) + 8|0);
        $439 = (($n$041$i) + 1)|0;
        $440 = ($439>>>0)<($args_count$2$i>>>0);
        if ($440) {
         $$737$i = $$10$i;$390 = $430;$418 = $433;$args_count$140$i = $args_count$2$i;$args_format$139$i = $args_format$2$i;$cur$038$i = $438;$n$041$i = $439;
         label = 112;
        } else {
         $$10$i$lcssa = $$10$i;$$lcssa = $430;$$lcssa57 = $433;
         break;
        }
       }
       switch ($335|0) {
       case 0:  {
        label = 136;
        break L146;
        break;
       }
       case 3: case 2: case 1:  {
        $$7$lcssa69$i = $$10$i$lcssa;$679 = $$lcssa;$680 = $$lcssa57;
        label = 147;
        break L157;
        break;
       }
       case 5: case 4:  {
        break;
       }
       default: {
        $glyph$idx7$val$i = HEAP32[$229>>2]|0;
        $glyph$idx8$val$i = HEAP8[$315>>0]|0;
        $616 = ($glyph$idx8$val$i<<24>>24)==(0);
        do {
         if ($616) {
          $error$0$i$i = 8;
         } else {
          $617 = ((($glyph$idx7$val$i)) + 22|0);
          $618 = HEAP16[$617>>1]|0;
          $619 = $618 << 16 >> 16;
          $620 = ((($glyph$idx7$val$i)) + 58|0);
          $621 = HEAP16[$620>>1]|0;
          $622 = $621 << 16 >> 16;
          $623 = (($619) + 3)|0;
          $624 = (($623) + ($622))|0;
          $625 = ((($glyph$idx7$val$i)) + 4|0);
          $626 = HEAP32[$625>>2]|0;
          $627 = ($624>>>0)>($626>>>0);
          if ($627) {
           $628 = (_FT_GlyphLoader_CheckPoints($glyph$idx7$val$i,3,0)|0);
           $629 = ($628|0)==(0);
           if (!($629)) {
            $error$0$i$i = $628;
            break;
           }
           $$pre$i$i = HEAP16[$620>>1]|0;
           $633 = $$pre$i$i;
          } else {
           $633 = $621;
          }
          $630 = ((($glyph$idx7$val$i)) + 60|0);
          $631 = HEAP32[$630>>2]|0;
          $632 = $633 << 16 >> 16;
          $634 = (($631) + ($632<<3)|0);
          $635 = ((($glyph$idx7$val$i)) + 64|0);
          $636 = HEAP32[$635>>2]|0;
          $637 = (($636) + ($632)|0);
          $638 = $pos$i;
          $639 = $638;
          $640 = HEAP32[$639>>2]|0;
          $641 = (($638) + 4)|0;
          $642 = $641;
          $643 = HEAP32[$642>>2]|0;
          $644 = $634;
          $645 = $644;
          HEAP32[$645>>2] = $640;
          $646 = (($644) + 4)|0;
          $647 = $646;
          HEAP32[$647>>2] = $643;
          $648 = ((($634)) + 8|0);
          $649 = $328;
          $650 = $649;
          $651 = HEAP32[$650>>2]|0;
          $652 = (($649) + 4)|0;
          $653 = $652;
          $654 = HEAP32[$653>>2]|0;
          $655 = $648;
          $656 = $655;
          HEAP32[$656>>2] = $651;
          $657 = (($655) + 4)|0;
          $658 = $657;
          HEAP32[$658>>2] = $654;
          $659 = ((($634)) + 16|0);
          $660 = $329;
          $661 = $660;
          $662 = HEAP32[$661>>2]|0;
          $663 = (($660) + 4)|0;
          $664 = $663;
          $665 = HEAP32[$664>>2]|0;
          $666 = $659;
          $667 = $666;
          HEAP32[$667>>2] = $662;
          $668 = (($666) + 4)|0;
          $669 = $668;
          HEAP32[$669>>2] = $665;
          HEAP8[$637>>0] = 2;
          $670 = ((($637)) + 1|0);
          HEAP8[$670>>0] = 2;
          $671 = ((($637)) + 2|0);
          HEAP8[$671>>0] = 1;
          $672 = HEAP16[$620>>1]|0;
          $673 = $672&65535;
          $674 = (($673) + 3)|0;
          $675 = $674&65535;
          HEAP16[$620>>1] = $675;
          $error$0$i$i = 0;
         }
        } while(0);
        HEAP32[$error$i2>>2] = $error$0$i$i;
        $$7$lcssa68$i = $$10$i$lcssa;$677 = $error$0$i$i;$681 = $$lcssa;$682 = $$lcssa57;
        break L157;
       }
       }
       $517 = HEAP32[$229>>2]|0;
       $518 = HEAP8[$315>>0]|0;
       $519 = ($518<<24>>24)==(0);
       $$pre$i9$i = ((($517)) + 58|0);
       if (!($519)) {
        $520 = HEAP16[$$pre$i9$i>>1]|0;
        $521 = $520 << 16 >> 16;
        $522 = (($521) + -1)|0;
        $523 = ((($517)) + 56|0);
        $524 = HEAP32[$523>>2]|0;
        $525 = $524&65535;
        $526 = ($525<<16>>16)>(0);
        $527 = $524 >>> 16;
        $528 = $527&65535;
        if ($526) {
         $sext1$i$i$i = $524 << 16;
         $529 = $sext1$i$i$i >> 16;
         $530 = (($529) + -1)|0;
         $531 = ((($517)) + 68|0);
         $532 = HEAP32[$531>>2]|0;
         $533 = (($532) + ($530<<1)|0);
         $534 = HEAP16[$533>>1]|0;
         $535 = $534 << 16 >> 16;
         $first$0$i$i$i = $535;
        } else {
         $first$0$i$i$i = 0;
        }
        $536 = ($522|0)>($first$0$i$i$i|0);
        if ($536) {
         $537 = ((($517)) + 60|0);
         $538 = HEAP32[$537>>2]|0;
         $539 = (($538) + ($first$0$i$i$i<<3)|0);
         $540 = HEAP32[$539>>2]|0;
         $541 = (($538) + ($522<<3)|0);
         $542 = HEAP32[$541>>2]|0;
         $543 = ($540|0)==($542|0);
         if ($543) {
          $544 = (((($538) + ($first$0$i$i$i<<3)|0)) + 4|0);
          $545 = HEAP32[$544>>2]|0;
          $546 = (((($538) + ($522<<3)|0)) + 4|0);
          $547 = HEAP32[$546>>2]|0;
          $548 = ($545|0)==($547|0);
          if ($548) {
           $549 = (($528) + -1)<<16>>16;
           HEAP16[$$pre$i9$i>>1] = $549;
           $550 = (($521) + -2)|0;
           $last$1$i$i$i = $550;
          } else {
           $last$1$i$i$i = $522;
          }
         } else {
          $last$1$i$i$i = $522;
         }
        } else {
         $last$1$i$i$i = $522;
        }
        $551 = ($last$1$i$i$i|0)<($first$0$i$i$i|0);
        if (!($551)) {
         $552 = $last$1$i$i$i&65535;
         $553 = (($525) + 1)<<16>>16;
         HEAP16[$523>>1] = $553;
         $sext$i$i$i = $524 << 16;
         $554 = $sext$i$i$i >> 16;
         $555 = ((($517)) + 68|0);
         $556 = HEAP32[$555>>2]|0;
         $557 = (($556) + ($554<<1)|0);
         HEAP16[$557>>1] = $552;
        }
        HEAP8[$315>>0] = 0;
       }
       HEAP8[$315>>0] = 1;
       $558 = ((($517)) + 22|0);
       $559 = HEAP16[$558>>1]|0;
       $560 = $559 << 16 >> 16;
       $561 = HEAP16[$$pre$i9$i>>1]|0;
       $562 = $561 << 16 >> 16;
       $563 = (($560) + 1)|0;
       $564 = (($563) + ($562))|0;
       $565 = ((($517)) + 4|0);
       $566 = HEAP32[$565>>2]|0;
       $567 = ($564>>>0)>($566>>>0);
       if ($567) {
        label = 165;
       } else {
        $568 = ((($517)) + 20|0);
        $569 = HEAP32[$568>>2]|0;
        $sext59$i = $569 << 16;
        $570 = $sext59$i >> 16;
        $571 = ((($517)) + 56|0);
        $572 = HEAP32[$571>>2]|0;
        $sext60$i = $572 << 16;
        $573 = $sext60$i >> 16;
        $574 = (($570) + 1)|0;
        $575 = (($574) + ($573))|0;
        $576 = ((($517)) + 8|0);
        $577 = HEAP32[$576>>2]|0;
        $578 = ($575>>>0)>($577>>>0);
        $579 = $569 >>> 16;
        $580 = $579&65535;
        $581 = $572 >>> 16;
        $582 = $581&65535;
        if ($578) {
         label = 165;
        } else {
         $586 = $580;$589 = $582;$593 = $566;$glyph$idx$val4$i$i = $517;
         label = 168;
        }
       }
       if ((label|0) == 165) {
        label = 0;
        $583 = (_FT_GlyphLoader_CheckPoints($517,1,1)|0);
        $584 = ($583|0)==(0);
        if ($584) {
         $glyph$idx$val$pre$i$i = HEAP32[$229>>2]|0;
         $glyph$idx1$val$pre$i$i = HEAP8[$315>>0]|0;
         $phitmp$i$i = ($glyph$idx1$val$pre$i$i<<24>>24)==(0);
         if ($phitmp$i$i) {
          $error$0$i11$i = 8;
         } else {
          $$phi$trans$insert51$i = ((($glyph$idx$val$pre$i$i)) + 22|0);
          $$pre52$i = HEAP16[$$phi$trans$insert51$i>>1]|0;
          $$phi$trans$insert53$i = ((($glyph$idx$val$pre$i$i)) + 58|0);
          $$pre54$i = HEAP16[$$phi$trans$insert53$i>>1]|0;
          $$phi$trans$insert55$i = ((($glyph$idx$val$pre$i$i)) + 4|0);
          $$pre56$i = HEAP32[$$phi$trans$insert55$i>>2]|0;
          $586 = $$pre52$i;$589 = $$pre54$i;$593 = $$pre56$i;$glyph$idx$val4$i$i = $glyph$idx$val$pre$i$i;
          label = 168;
         }
        } else {
         $error$0$i11$i = $583;
        }
       }
       do {
        if ((label|0) == 168) {
         label = 0;
         $585 = $586 << 16 >> 16;
         $587 = ((($glyph$idx$val4$i$i)) + 58|0);
         $588 = $589 << 16 >> 16;
         $590 = (($588) + 1)|0;
         $591 = (($590) + ($585))|0;
         $592 = ($591>>>0)>($593>>>0);
         if ($592) {
          $594 = (_FT_GlyphLoader_CheckPoints($glyph$idx$val4$i$i,1,0)|0);
          $595 = ($594|0)==(0);
          if (!($595)) {
           $error$0$i11$i = $594;
           break;
          }
          $$pre$i$i$i = HEAP16[$587>>1]|0;
          $597 = $$pre$i$i$i;
         } else {
          $597 = $589;
         }
         $596 = $597 << 16 >> 16;
         $598 = ((($glyph$idx$val4$i$i)) + 60|0);
         $599 = HEAP32[$598>>2]|0;
         $600 = (($599) + ($596<<3)|0);
         $601 = $pos$i;
         $602 = $601;
         $603 = HEAP32[$602>>2]|0;
         $604 = (($601) + 4)|0;
         $605 = $604;
         $606 = HEAP32[$605>>2]|0;
         $607 = $600;
         $608 = $607;
         HEAP32[$608>>2] = $603;
         $609 = (($607) + 4)|0;
         $610 = $609;
         HEAP32[$610>>2] = $606;
         $611 = ((($glyph$idx$val4$i$i)) + 64|0);
         $612 = HEAP32[$611>>2]|0;
         $613 = (($612) + ($596)|0);
         HEAP8[$613>>0] = 1;
         $614 = HEAP16[$587>>1]|0;
         $615 = (($614) + 1)<<16>>16;
         HEAP16[$587>>1] = $615;
         $error$0$i11$i = 0;
        }
       } while(0);
       HEAP32[$error$i2>>2] = $error$0$i11$i;
       $$7$lcssa68$i = $$10$i$lcssa;$677 = $error$0$i11$i;$681 = $$lcssa;$682 = $$lcssa57;
      }
     } while(0);
     if ((label|0) == 147) {
      label = 0;
      $glyph$idx$val$i = HEAP32[$229>>2]|0;
      $glyph$idx6$val$i = HEAP8[$315>>0]|0;
      $483 = ($glyph$idx6$val$i<<24>>24)==(0);
      do {
       if ($483) {
        $error$0$i20$i = 8;
       } else {
        $484 = ((($glyph$idx$val$i)) + 22|0);
        $485 = HEAP16[$484>>1]|0;
        $486 = $485 << 16 >> 16;
        $487 = ((($glyph$idx$val$i)) + 58|0);
        $488 = HEAP16[$487>>1]|0;
        $489 = $488 << 16 >> 16;
        $490 = (($486) + 1)|0;
        $491 = (($490) + ($489))|0;
        $492 = ((($glyph$idx$val$i)) + 4|0);
        $493 = HEAP32[$492>>2]|0;
        $494 = ($491>>>0)>($493>>>0);
        if ($494) {
         $495 = (_FT_GlyphLoader_CheckPoints($glyph$idx$val$i,1,0)|0);
         $496 = ($495|0)==(0);
         if (!($496)) {
          $error$0$i20$i = $495;
          break;
         }
         $$pre$i17$i = HEAP16[$487>>1]|0;
         $498 = $$pre$i17$i;
        } else {
         $498 = $488;
        }
        $497 = $498 << 16 >> 16;
        $499 = ((($glyph$idx$val$i)) + 60|0);
        $500 = HEAP32[$499>>2]|0;
        $501 = (($500) + ($497<<3)|0);
        $502 = $pos$i;
        $503 = $502;
        $504 = HEAP32[$503>>2]|0;
        $505 = (($502) + 4)|0;
        $506 = $505;
        $507 = HEAP32[$506>>2]|0;
        $508 = $501;
        $509 = $508;
        HEAP32[$509>>2] = $504;
        $510 = (($508) + 4)|0;
        $511 = $510;
        HEAP32[$511>>2] = $507;
        $512 = ((($glyph$idx$val$i)) + 64|0);
        $513 = HEAP32[$512>>2]|0;
        $514 = (($513) + ($497)|0);
        HEAP8[$514>>0] = 1;
        $515 = HEAP16[$487>>1]|0;
        $516 = (($515) + 1)<<16>>16;
        HEAP16[$487>>1] = $516;
        $error$0$i20$i = 0;
       }
      } while(0);
      HEAP32[$error$i2>>2] = $error$0$i20$i;
      $$7$lcssa68$i = $$7$lcssa69$i;$677 = $error$0$i20$i;$681 = $679;$682 = $680;
     }
     $676 = ($677|0)==(0);
     if ($676) {
      $$6$i16 = $$7$lcssa68$i;$340 = $682;$352 = $681;
     } else {
      $$pre57$i = $677;
      break;
     }
    }
    if ((label|0) == 136) {
     $441 = HEAP32[$229>>2]|0;
     $442 = HEAP8[$315>>0]|0;
     $443 = ($442<<24>>24)==(0);
     if (!($443)) {
      $444 = ((($441)) + 58|0);
      $445 = HEAP16[$444>>1]|0;
      $446 = $445 << 16 >> 16;
      $447 = (($446) + -1)|0;
      $448 = ((($441)) + 56|0);
      $449 = HEAP32[$448>>2]|0;
      $450 = $449&65535;
      $451 = ($450<<16>>16)>(0);
      $452 = $449 >>> 16;
      $453 = $452&65535;
      if ($451) {
       $sext1$i$i12$i = $449 << 16;
       $454 = $sext1$i$i12$i >> 16;
       $455 = (($454) + -1)|0;
       $456 = ((($441)) + 68|0);
       $457 = HEAP32[$456>>2]|0;
       $458 = (($457) + ($455<<1)|0);
       $459 = HEAP16[$458>>1]|0;
       $460 = $459 << 16 >> 16;
       $first$0$i$i13$i = $460;
      } else {
       $first$0$i$i13$i = 0;
      }
      $461 = ($447|0)>($first$0$i$i13$i|0);
      if ($461) {
       $462 = ((($441)) + 60|0);
       $463 = HEAP32[$462>>2]|0;
       $464 = (($463) + ($first$0$i$i13$i<<3)|0);
       $465 = HEAP32[$464>>2]|0;
       $466 = (($463) + ($447<<3)|0);
       $467 = HEAP32[$466>>2]|0;
       $468 = ($465|0)==($467|0);
       if ($468) {
        $469 = (((($463) + ($first$0$i$i13$i<<3)|0)) + 4|0);
        $470 = HEAP32[$469>>2]|0;
        $471 = (((($463) + ($447<<3)|0)) + 4|0);
        $472 = HEAP32[$471>>2]|0;
        $473 = ($470|0)==($472|0);
        if ($473) {
         $474 = (($453) + -1)<<16>>16;
         HEAP16[$444>>1] = $474;
         $475 = (($446) + -2)|0;
         $last$1$i$i14$i = $475;
        } else {
         $last$1$i$i14$i = $447;
        }
       } else {
        $last$1$i$i14$i = $447;
       }
      } else {
       $last$1$i$i14$i = $447;
      }
      $476 = ($last$1$i$i14$i|0)<($first$0$i$i13$i|0);
      if (!($476)) {
       $477 = $last$1$i$i14$i&65535;
       $478 = (($450) + 1)<<16>>16;
       HEAP16[$448>>1] = $478;
       $sext$i$i15$i = $449 << 16;
       $479 = $sext$i$i15$i >> 16;
       $480 = ((($441)) + 68|0);
       $481 = HEAP32[$480>>2]|0;
       $482 = (($481) + ($479<<1)|0);
       HEAP16[$482>>1] = $477;
      }
      HEAP8[$315>>0] = 0;
     }
     _FT_GlyphLoader_Add($441);
     $$pre57$pre$i = HEAP32[$error$i2>>2]|0;
     $$pre57$i = $$pre57$pre$i;
    }
    else if ((label|0) == 180) {
     label = 183;
     break;
    }
    else if ((label|0) == 182) {
     label = 183;
     break;
    }
    $678 = $$pre57$i;
   } else {
    label = 183;
   }
  }
 } while(0);
 if ((label|0) == 183) {
  HEAP32[$error$i2>>2] = 8;
  $678 = 8;
 }
 _FT_Stream_ExitFrame($stream);
 $$01 = $678;
 STACKTOP = sp;return ($$01|0);
}
function _pfr_get_kerning($pfrface,$left,$right,$avector) {
 $pfrface = $pfrface|0;
 $left = $left|0;
 $right = $right|0;
 $avector = $avector|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($pfrface)) + 288|0);
 (_pfr_face_get_kerning($pfrface,$left,$right,$avector)|0);
 $1 = ((($0)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($0)) + 16|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2|0)==($4|0);
 if ($5) {
  return 0;
 }
 $6 = HEAP32[$avector>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = (_FT_MulDiv($6,$2,$4)|0);
  HEAP32[$avector>>2] = $8;
 }
 $9 = ((($avector)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10|0)==(0);
 if ($11) {
  return 0;
 }
 $12 = HEAP32[$1>>2]|0;
 $13 = HEAP32[$3>>2]|0;
 $14 = (_FT_MulDiv($10,$12,$13)|0);
 HEAP32[$9>>2] = $14;
 return 0;
}
function _ps_table_new($table,$count,$memory) {
 $table = $table|0;
 $count = $count|0;
 $memory = $memory|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($table)) + 32|0);
 HEAP32[$0>>2] = $memory;
 $1 = (_ft_mem_realloc($memory,4,0,$count,0,$error)|0);
 $2 = ((($table)) + 24|0);
 HEAP32[$2>>2] = $1;
 $3 = HEAP32[$error>>2]|0;
 $4 = ($3|0)==(0);
 if ($4) {
  $5 = (_ft_mem_realloc($memory,4,0,$count,0,$error)|0);
  $6 = ((($table)) + 28|0);
  HEAP32[$6>>2] = $5;
  $7 = HEAP32[$error>>2]|0;
  $8 = ($7|0)==(0);
  if ($8) {
   $9 = ((($table)) + 16|0);
   HEAP32[$9>>2] = $count;
   $10 = ((($table)) + 12|0);
   HEAP32[$10>>2] = -559038737;
   $11 = ((($table)) + 20|0);
   HEAP32[$11>>2] = 0;
   HEAP32[$table>>2] = 0;
   $12 = ((($table)) + 8|0);
   HEAP32[$12>>2] = 0;
   $13 = ((($table)) + 4|0);
   HEAP32[$13>>2] = 0;
   $14 = ((($table)) + 36|0);
   ;HEAP32[$14>>2]=HEAP32[53508>>2]|0;HEAP32[$14+4>>2]=HEAP32[53508+4>>2]|0;HEAP32[$14+8>>2]=HEAP32[53508+8>>2]|0;HEAP32[$14+12>>2]=HEAP32[53508+12>>2]|0;
   $16 = 0;
   STACKTOP = sp;return ($16|0);
  }
 }
 $15 = HEAP32[$2>>2]|0;
 _ft_mem_free($memory,$15);
 HEAP32[$2>>2] = 0;
 $$pre = HEAP32[$error>>2]|0;
 $16 = $$pre;
 STACKTOP = sp;return ($16|0);
}
function _ps_table_done($table) {
 $table = $table|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $offset$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($table)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$table>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  STACKTOP = sp;return;
 }
 $4 = ((($table)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (_ft_mem_alloc($1,$5,$error)|0);
 HEAP32[$table>>2] = $6;
 $7 = HEAP32[$error>>2]|0;
 $8 = ($7|0)==(0);
 if (!($8)) {
  STACKTOP = sp;return;
 }
 $9 = HEAP32[$4>>2]|0;
 _memcpy(($6|0),($2|0),($9|0))|0;
 $10 = HEAP32[$table>>2]|0;
 $11 = $2;
 $12 = (($10) - ($11))|0;
 $13 = ((($table)) + 24|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($table)) + 16|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = (($14) + ($16<<2)|0);
 $18 = ($16|0)>(0);
 if ($18) {
  $offset$01$i = $14;
  while(1) {
   $19 = HEAP32[$offset$01$i>>2]|0;
   $20 = ($19|0)==(0|0);
   if (!($20)) {
    $21 = (($19) + ($12)|0);
    HEAP32[$offset$01$i>>2] = $21;
   }
   $22 = ((($offset$01$i)) + 4|0);
   $23 = ($22>>>0)<($17>>>0);
   if ($23) {
    $offset$01$i = $22;
   } else {
    break;
   }
  }
 }
 $24 = HEAP32[$4>>2]|0;
 $25 = ((($table)) + 8|0);
 HEAP32[$25>>2] = $24;
 _ft_mem_free($1,$2);
 STACKTOP = sp;return;
}
function _ps_table_add($table,$idx,$object,$length) {
 $table = $table|0;
 $idx = $idx|0;
 $object = $object|0;
 $length = $length|0;
 var $$ = 0, $$1 = 0, $$2 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $error$i = 0, $new_size$06 = 0, $offset$01$i$i = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp;
 $0 = ($idx|0)<(0);
 if ($0) {
  $$1 = 6;
  STACKTOP = sp;return ($$1|0);
 }
 $1 = ((($table)) + 16|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)>($idx|0);
 if (!($3)) {
  $$1 = 6;
  STACKTOP = sp;return ($$1|0);
 }
 $4 = ((($table)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($5) + ($length))|0;
 $7 = ((($table)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($6>>>0)>($8>>>0);
 if ($9) {
  $10 = HEAP32[$table>>2]|0;
  $11 = $object;
  $12 = (($11) - ($10))|0;
  $13 = ($12|0)>(-1);
  $14 = ($12>>>0)<($8>>>0);
  $or$cond = $13 & $14;
  $$ = $or$cond ? $12 : -1;
  $15 = $10;
  $new_size$06 = $8;
  while(1) {
   $16 = $new_size$06 >>> 2;
   $17 = (($new_size$06) + 1024)|0;
   $18 = (($17) + ($16))|0;
   $19 = $18 & -1024;
   $20 = ($19>>>0)<($6>>>0);
   if ($20) {
    $new_size$06 = $19;
   } else {
    $$lcssa = $19;
    break;
   }
  }
  $21 = ((($table)) + 32|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = (_ft_mem_alloc($22,$$lcssa,$error$i)|0);
  HEAP32[$table>>2] = $23;
  $24 = HEAP32[$error$i>>2]|0;
  $25 = ($24|0)==(0);
  if (!($25)) {
   HEAP32[$table>>2] = $15;
   $$1 = $24;
   STACKTOP = sp;return ($$1|0);
  }
  $26 = ($10|0)==(0);
  if (!($26)) {
   $27 = HEAP32[$7>>2]|0;
   _memcpy(($23|0),($15|0),($27|0))|0;
   $28 = HEAP32[$table>>2]|0;
   $29 = (($28) - ($10))|0;
   $30 = ((($table)) + 24|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = HEAP32[$1>>2]|0;
   $33 = (($31) + ($32<<2)|0);
   $34 = ($32|0)>(0);
   if ($34) {
    $offset$01$i$i = $31;
    while(1) {
     $35 = HEAP32[$offset$01$i$i>>2]|0;
     $36 = ($35|0)==(0|0);
     if (!($36)) {
      $37 = (($35) + ($29)|0);
      HEAP32[$offset$01$i$i>>2] = $37;
     }
     $38 = ((($offset$01$i$i)) + 4|0);
     $39 = ($38>>>0)<($33>>>0);
     if ($39) {
      $offset$01$i$i = $38;
     } else {
      break;
     }
    }
   }
   _ft_mem_free($22,$15);
  }
  HEAP32[$7>>2] = $$lcssa;
  $40 = ($$|0)>(-1);
  if ($40) {
   $41 = HEAP32[$table>>2]|0;
   $42 = (($41) + ($$)|0);
   $$2 = $42;
  } else {
   $$2 = $object;
  }
 } else {
  $$2 = $object;
 }
 $43 = HEAP32[$table>>2]|0;
 $44 = HEAP32[$4>>2]|0;
 $45 = (($43) + ($44)|0);
 $46 = ((($table)) + 24|0);
 $47 = HEAP32[$46>>2]|0;
 $48 = (($47) + ($idx<<2)|0);
 HEAP32[$48>>2] = $45;
 $49 = ((($table)) + 28|0);
 $50 = HEAP32[$49>>2]|0;
 $51 = (($50) + ($idx<<2)|0);
 HEAP32[$51>>2] = $length;
 $52 = HEAP32[$table>>2]|0;
 $53 = HEAP32[$4>>2]|0;
 $54 = (($52) + ($53)|0);
 _memcpy(($54|0),($$2|0),($length|0))|0;
 $55 = HEAP32[$4>>2]|0;
 $56 = (($55) + ($length))|0;
 HEAP32[$4>>2] = $56;
 $$1 = 0;
 STACKTOP = sp;return ($$1|0);
}
function _ps_table_release($table) {
 $table = $table|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($table)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($table)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(-559038737);
 if (!($4)) {
  return;
 }
 $5 = HEAP32[$table>>2]|0;
 _ft_mem_free($1,$5);
 HEAP32[$table>>2] = 0;
 $6 = ((($table)) + 24|0);
 $7 = HEAP32[$6>>2]|0;
 _ft_mem_free($1,$7);
 HEAP32[$6>>2] = 0;
 $8 = ((($table)) + 28|0);
 $9 = HEAP32[$8>>2]|0;
 _ft_mem_free($1,$9);
 HEAP32[$8>>2] = 0;
 HEAP32[$2>>2] = 0;
 return;
}
function _ps_parser_init($parser,$base,$limit,$memory) {
 $parser = $parser|0;
 $base = $base|0;
 $limit = $limit|0;
 $memory = $memory|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($parser)) + 12|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($parser)) + 4|0);
 HEAP32[$1>>2] = $base;
 $2 = ((($parser)) + 8|0);
 HEAP32[$2>>2] = $limit;
 HEAP32[$parser>>2] = $base;
 $3 = ((($parser)) + 16|0);
 HEAP32[$3>>2] = $memory;
 $4 = ((($parser)) + 20|0);
 dest=$4; src=53524; stop=dest+52|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 return;
}
function _ps_parser_done($parser) {
 $parser = $parser|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _ps_parser_skip_spaces($parser) {
 $parser = $parser|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cur$0$lcssa$i$i = 0, $cur$01$i$i = 0, $cur$sroa$0$0$lcssa$i = 0, $cur$sroa$0$01$i = 0;
 var $cur$sroa$0$1$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($parser)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$parser>>2]|0;
 $3 = $2;
 $4 = ($3>>>0)<($1>>>0);
 L1: do {
  if ($4) {
   $6 = $3;$cur$sroa$0$01$i = $2;
   while(1) {
    $5 = HEAP8[$6>>0]|0;
    switch ($5<<24>>24) {
    case 0: case 12: case 9: case 10: case 13: case 32:  {
     $cur$sroa$0$1$i = $cur$sroa$0$01$i;
     break;
    }
    case 37:  {
     $7 = 37;$cur$01$i$i = $6;
     L5: while(1) {
      switch ($7<<24>>24) {
      case 10: case 13:  {
       $cur$0$lcssa$i$i = $cur$01$i$i;
       break L5;
       break;
      }
      default: {
      }
      }
      $8 = ((($cur$01$i$i)) + 1|0);
      $9 = ($8>>>0)<($1>>>0);
      if (!($9)) {
       $cur$0$lcssa$i$i = $8;
       break;
      }
      $$pre$i = HEAP8[$8>>0]|0;
      $7 = $$pre$i;$cur$01$i$i = $8;
     }
     $10 = $cur$0$lcssa$i$i;
     $cur$sroa$0$1$i = $10;
     break;
    }
    default: {
     $cur$sroa$0$0$lcssa$i = $cur$sroa$0$01$i;
     break L1;
    }
    }
    $11 = $cur$sroa$0$1$i;
    $12 = ((($11)) + 1|0);
    $13 = $12;
    $14 = ($12>>>0)<($1>>>0);
    if ($14) {
     $6 = $12;$cur$sroa$0$01$i = $13;
    } else {
     $cur$sroa$0$0$lcssa$i = $13;
     break L1;
    }
   }
  } else {
   $cur$sroa$0$0$lcssa$i = $2;
  }
 } while(0);
 HEAP32[$parser>>2] = $cur$sroa$0$0$lcssa$i;
 return;
}
function _ps_parser_skip_PS_token($parser) {
 $parser = $parser|0;
 var $$be114 = 0, $$lcssa177 = 0, $$lcssa178 = 0, $$lcssa181 = 0, $$lcssa185 = 0, $$lcssa188 = 0, $$lcssa189 = 0, $$off$i = 0, $$off$i23 = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i16 = 0, $$pre$i42 = 0, $$pre$i5 = 0, $$pre$i9 = 0, $$pre128 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0;
 var $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0;
 var $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0;
 var $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cur$0$be$i = 0, $cur$0$be$i36 = 0, $cur$0$lcssa$i$i = 0, $cur$0$lcssa$i$i$i = 0, $cur$0$lcssa$i$i$i18 = 0, $cur$0$lcssa$i$ph$i = 0, $cur$0$ph16$i = 0, $cur$0$ph16$i31 = 0, $cur$01$i$i = 0, $cur$01$i$i$i = 0, $cur$01$i$i$i14 = 0, $cur$01$i$i7 = 0, $cur$09$i = 0, $cur$09$i34 = 0, $cur$16$i104 = 0, $cur$16$i3899 = 0;
 var $cur$4$i = 0, $cur$4$i47 = 0, $cur$i$sroa$0$0 = 0, $cur$i$sroa$0$1 = 0, $cur$sroa$0$0$lcssa$i = 0, $cur$sroa$0$0$lcssa$i$i = 0, $cur$sroa$0$0$lcssa$i$i21 = 0, $cur$sroa$0$0$ph = 0, $cur$sroa$0$01$i = 0, $cur$sroa$0$01$i$i = 0, $cur$sroa$0$01$i$i$be = 0, $cur$sroa$0$01$i$i12 = 0, $cur$sroa$0$066 = 0, $cur$sroa$0$1 = 0, $cur$sroa$0$1$i = 0, $cur$sroa$0$1$i$i = 0, $cur$sroa$0$1$i$i20 = 0, $cur$sroa$0$1$i2 = 0, $cur$sroa$0$1$i26 = 0, $embed$0$ph$be$i = 0;
 var $embed$0$ph$be$i44 = 0, $embed$0$ph15$i = 0, $embed$0$ph15$i32 = 0, $embed$05$i = 0, $embed$05$i$lcssa = 0, $embed$1$i = 0, $embed$2$i = 0, $error$0 = 0, $error$0$$i = 0, $error$03$i = 0, $error$1 = 0, $error$1$i = 0, $error$2$i = 0, $error$2$i46 = 0, $i$05$i105 = 0, $i$05$i39100 = 0, $or$cond = 0, $or$cond$i = 0, $or$cond$i41 = 0, $or$cond111 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$parser>>2]|0;
 $1 = ((($parser)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $0;
 $4 = ($3>>>0)<($2>>>0);
 L1: do {
  if ($4) {
   $6 = $3;$cur$sroa$0$01$i = $0;
   while(1) {
    $5 = HEAP8[$6>>0]|0;
    switch ($5<<24>>24) {
    case 0: case 12: case 9: case 10: case 13: case 32:  {
     $cur$sroa$0$1$i = $cur$sroa$0$01$i;
     break;
    }
    case 37:  {
     $7 = 37;$cur$01$i$i = $6;
     L5: while(1) {
      switch ($7<<24>>24) {
      case 10: case 13:  {
       $cur$0$lcssa$i$i = $cur$01$i$i;
       break L5;
       break;
      }
      default: {
      }
      }
      $8 = ((($cur$01$i$i)) + 1|0);
      $9 = ($8>>>0)<($2>>>0);
      if (!($9)) {
       $cur$0$lcssa$i$i = $8;
       break;
      }
      $$pre$i = HEAP8[$8>>0]|0;
      $7 = $$pre$i;$cur$01$i$i = $8;
     }
     $10 = $cur$0$lcssa$i$i;
     $cur$sroa$0$1$i = $10;
     break;
    }
    default: {
     $cur$sroa$0$0$lcssa$i = $cur$sroa$0$01$i;
     break L1;
    }
    }
    $11 = $cur$sroa$0$1$i;
    $12 = ((($11)) + 1|0);
    $13 = $12;
    $14 = ($12>>>0)<($2>>>0);
    if ($14) {
     $6 = $12;$cur$sroa$0$01$i = $13;
    } else {
     $cur$sroa$0$0$lcssa$i = $13;
     break L1;
    }
   }
  } else {
   $cur$sroa$0$0$lcssa$i = $0;
  }
 } while(0);
 $15 = $cur$sroa$0$0$lcssa$i;
 $16 = ($15>>>0)<($2>>>0);
 L12: do {
  if ($16) {
   $17 = HEAP8[$15>>0]|0;
   switch ($17<<24>>24) {
   case 93: case 91:  {
    $18 = ((($15)) + 1|0);
    $19 = $18;
    $cur$sroa$0$1 = $19;$error$0 = 0;
    break L12;
    break;
   }
   case 123:  {
    $164 = $15;$21 = 123;$cur$i$sroa$0$0 = $cur$sroa$0$0$lcssa$i;$embed$05$i = 0;
    L16: while(1) {
     $20 = $21&255;
     switch ($20|0) {
     case 123:  {
      $22 = (($embed$05$i) + 1)|0;
      $cur$i$sroa$0$1 = $cur$i$sroa$0$0;$embed$1$i = $22;$error$1$i = 0;
      break;
     }
     case 125:  {
      $23 = (($embed$05$i) + -1)|0;
      $24 = ($23|0)==(0);
      if ($24) {
       $$lcssa181 = $164;
       label = 14;
       break L16;
      } else {
       $cur$i$sroa$0$1 = $cur$i$sroa$0$0;$embed$1$i = $23;$error$1$i = 0;
      }
      break;
     }
     case 40:  {
      $27 = $cur$i$sroa$0$0;
      $28 = ($27>>>0)<($2>>>0);
      L22: do {
       if ($28) {
        $cur$0$ph16$i31 = $27;$embed$0$ph15$i32 = 0;
        while(1) {
         $cur$09$i34 = $cur$0$ph16$i31;
         L25: while(1) {
          $29 = HEAP8[$cur$09$i34>>0]|0;
          $30 = ((($cur$09$i34)) + 1|0);
          L27: do {
           switch ($29<<24>>24) {
           case 40:  {
            $$lcssa177 = $30;
            label = 25;
            break L25;
            break;
           }
           case 41:  {
            $$lcssa178 = $30;
            label = 27;
            break L25;
            break;
           }
           case 92:  {
            $31 = ($30|0)==($2|0);
            if ($31) {
             $cur$4$i47 = $2;$error$2$i46 = 3;
             break L22;
            }
            $32 = HEAP8[$30>>0]|0;
            $33 = $32&255;
            switch ($33|0) {
            case 41: case 40: case 92: case 102: case 98: case 116: case 114: case 110:  {
             $37 = ((($cur$09$i34)) + 2|0);
             $cur$0$be$i36 = $37;
             break L27;
             break;
            }
            default: {
            }
            }
            $34 = ($30>>>0)<($2>>>0);
            $35 = $32 & -8;
            $36 = ($35<<24>>24)==(48);
            $or$cond = $34 & $36;
            if ($or$cond) {
             $cur$16$i3899 = $30;$i$05$i39100 = 0;
             while(1) {
              $39 = ((($cur$16$i3899)) + 1|0);
              $40 = (($i$05$i39100) + 1)|0;
              $41 = ($40>>>0)<(3);
              $42 = ($39>>>0)<($2>>>0);
              $or$cond$i41 = $42 & $41;
              if (!($or$cond$i41)) {
               $cur$0$be$i36 = $39;
               break L27;
              }
              $$pre$i42 = HEAP8[$39>>0]|0;
              $43 = $$pre$i42 & -8;
              $44 = ($43<<24>>24)==(48);
              if ($44) {
               $cur$16$i3899 = $39;$i$05$i39100 = $40;
              } else {
               $cur$0$be$i36 = $39;
               break;
              }
             }
            } else {
             $cur$0$be$i36 = $30;
            }
            break;
           }
           default: {
            $cur$0$be$i36 = $30;
           }
           }
          } while(0);
          $38 = ($cur$0$be$i36>>>0)<($2>>>0);
          if ($38) {
           $cur$09$i34 = $cur$0$be$i36;
          } else {
           $cur$4$i47 = $cur$0$be$i36;$error$2$i46 = 3;
           break L22;
          }
         }
         if ((label|0) == 25) {
          label = 0;
          $45 = (($embed$0$ph15$i32) + 1)|0;
          $47 = $$lcssa177;$embed$0$ph$be$i44 = $45;
         }
         else if ((label|0) == 27) {
          label = 0;
          $48 = (($embed$0$ph15$i32) + -1)|0;
          $49 = ($48|0)==(0);
          if ($49) {
           $cur$4$i47 = $$lcssa178;$error$2$i46 = 0;
           break L22;
          } else {
           $47 = $$lcssa178;$embed$0$ph$be$i44 = $48;
          }
         }
         $46 = ($47>>>0)<($2>>>0);
         if ($46) {
          $cur$0$ph16$i31 = $47;$embed$0$ph15$i32 = $embed$0$ph$be$i44;
         } else {
          $cur$4$i47 = $47;$error$2$i46 = 3;
          break;
         }
        }
       } else {
        $cur$4$i47 = $27;$error$2$i46 = 3;
       }
      } while(0);
      $50 = $cur$4$i47;
      $cur$i$sroa$0$1 = $50;$embed$1$i = $embed$05$i;$error$1$i = $error$2$i46;
      break;
     }
     case 60:  {
      $51 = $cur$i$sroa$0$0;
      $52 = ((($51)) + 1|0);
      $53 = $52;
      $54 = ($52>>>0)<($2>>>0);
      L43: do {
       if ($54) {
        $56 = $52;$cur$sroa$0$01$i$i12 = $53;
        while(1) {
         $55 = HEAP8[$56>>0]|0;
         switch ($55<<24>>24) {
         case 0: case 12: case 9: case 10: case 13: case 32:  {
          $cur$sroa$0$1$i$i20 = $cur$sroa$0$01$i$i12;
          label = 35;
          break;
         }
         case 37:  {
          $57 = 37;$cur$01$i$i$i14 = $56;
          L47: while(1) {
           switch ($57<<24>>24) {
           case 10: case 13:  {
            $cur$0$lcssa$i$i$i18 = $cur$01$i$i$i14;
            break L47;
            break;
           }
           default: {
           }
           }
           $58 = ((($cur$01$i$i$i14)) + 1|0);
           $59 = ($58>>>0)<($2>>>0);
           if (!($59)) {
            $cur$0$lcssa$i$i$i18 = $58;
            break;
           }
           $$pre$i$i16 = HEAP8[$58>>0]|0;
           $57 = $$pre$i$i16;$cur$01$i$i$i14 = $58;
          }
          $60 = $cur$0$lcssa$i$i$i18;
          $cur$sroa$0$1$i$i20 = $60;
          label = 35;
          break;
         }
         default: {
          $cur$sroa$0$0$lcssa$i$i21 = $cur$sroa$0$01$i$i12;
         }
         }
         if ((label|0) == 35) {
          label = 0;
          $61 = $cur$sroa$0$1$i$i20;
          $62 = ((($61)) + 1|0);
          $63 = $62;
          $64 = ($62>>>0)<($2>>>0);
          if ($64) {
           $56 = $62;$cur$sroa$0$01$i$i12 = $63;
           continue;
          } else {
           $cur$sroa$0$0$lcssa$i$i21 = $63;
          }
         }
         $65 = $cur$sroa$0$0$lcssa$i$i21;
         $66 = ($65>>>0)<($2>>>0);
         if (!($66)) {
          $cur$sroa$0$1$i26 = $cur$sroa$0$0$lcssa$i$i21;
          break L43;
         }
         $67 = HEAP8[$65>>0]|0;
         $$off$i23 = (($67) + -48)<<24>>24;
         $68 = ($$off$i23&255)<(10);
         if (!($68)) {
          switch ($67<<24>>24) {
          case 65: case 66: case 67: case 68: case 69: case 70: case 97: case 98: case 99: case 100: case 101: case 102:  {
           break;
          }
          default: {
           $cur$sroa$0$1$i26 = $cur$sroa$0$0$lcssa$i$i21;
           break L43;
          }
          }
         }
         $69 = ((($65)) + 1|0);
         $70 = $69;
         $71 = ($69>>>0)<($2>>>0);
         if ($71) {
          $56 = $69;$cur$sroa$0$01$i$i12 = $70;
         } else {
          $cur$sroa$0$1$i26 = $70;
          break L43;
         }
        }
       } else {
        $cur$sroa$0$1$i26 = $53;
       }
      } while(0);
      $72 = $cur$sroa$0$1$i26;
      $73 = ($72>>>0)<($2>>>0);
      if ($73) {
       $74 = HEAP8[$72>>0]|0;
       $75 = ($74<<24>>24)==(62);
       if (!($75)) {
        $$lcssa185 = $72;$embed$05$i$lcssa = $embed$05$i;
        label = 42;
        break L16;
       }
      }
      $78 = ((($72)) + 1|0);
      $79 = $78;
      $cur$i$sroa$0$1 = $79;$embed$1$i = $embed$05$i;$error$1$i = 0;
      break;
     }
     case 37:  {
      $80 = $21;$cur$01$i$i7 = $164;
      L63: while(1) {
       switch ($80<<24>>24) {
       case 10: case 13:  {
        $cur$0$lcssa$i$ph$i = $cur$01$i$i7;
        break L63;
        break;
       }
       default: {
       }
       }
       $81 = ((($cur$01$i$i7)) + 1|0);
       $82 = ($81>>>0)<($2>>>0);
       if (!($82)) {
        $cur$0$lcssa$i$ph$i = $81;
        break;
       }
       $$pre$i9 = HEAP8[$81>>0]|0;
       $80 = $$pre$i9;$cur$01$i$i7 = $81;
      }
      $83 = $cur$0$lcssa$i$ph$i;
      $cur$i$sroa$0$1 = $83;$embed$1$i = $embed$05$i;$error$1$i = 0;
      break;
     }
     default: {
      $cur$i$sroa$0$1 = $cur$i$sroa$0$0;$embed$1$i = $embed$05$i;$error$1$i = 0;
     }
     }
     $84 = $cur$i$sroa$0$1;
     $85 = ((($84)) + 1|0);
     $86 = $85;
     $87 = ($85>>>0)<($2>>>0);
     $88 = ($error$1$i|0)==(0);
     $89 = $88 & $87;
     if (!($89)) {
      $165 = $86;$embed$2$i = $embed$1$i;$error$03$i = $error$1$i;
      break;
     }
     $$pre128 = HEAP8[$85>>0]|0;
     $164 = $85;$21 = $$pre128;$cur$i$sroa$0$0 = $86;$embed$05$i = $embed$1$i;
    }
    if ((label|0) == 14) {
     $25 = ((($$lcssa181)) + 1|0);
     $26 = $25;
     $165 = $26;$embed$2$i = 0;$error$03$i = 0;
    }
    else if ((label|0) == 42) {
     $76 = ((($$lcssa185)) + 1|0);
     $77 = $76;
     $165 = $77;$embed$2$i = $embed$05$i$lcssa;$error$03$i = 3;
    }
    $90 = ($embed$2$i|0)==(0);
    $error$0$$i = $90 ? $error$03$i : 3;
    $cur$sroa$0$1 = $165;$error$0 = $error$0$$i;
    break L12;
    break;
   }
   case 40:  {
    $cur$0$ph16$i = $15;$embed$0$ph15$i = 0;
    L74: while(1) {
     $cur$09$i = $cur$0$ph16$i;
     L76: while(1) {
      $91 = HEAP8[$cur$09$i>>0]|0;
      $92 = ((($cur$09$i)) + 1|0);
      L78: do {
       switch ($91<<24>>24) {
       case 40:  {
        $$lcssa188 = $92;
        label = 60;
        break L76;
        break;
       }
       case 41:  {
        $$lcssa189 = $92;
        label = 62;
        break L76;
        break;
       }
       case 92:  {
        $93 = ($92|0)==($2|0);
        if ($93) {
         $cur$4$i = $2;$error$2$i = 3;
         break L74;
        }
        $94 = HEAP8[$92>>0]|0;
        $95 = $94&255;
        switch ($95|0) {
        case 41: case 40: case 92: case 102: case 98: case 116: case 114: case 110:  {
         $99 = ((($cur$09$i)) + 2|0);
         $cur$0$be$i = $99;
         break L78;
         break;
        }
        default: {
        }
        }
        $96 = ($92>>>0)<($2>>>0);
        $97 = $94 & -8;
        $98 = ($97<<24>>24)==(48);
        $or$cond111 = $96 & $98;
        if ($or$cond111) {
         $cur$16$i104 = $92;$i$05$i105 = 0;
         while(1) {
          $101 = ((($cur$16$i104)) + 1|0);
          $102 = (($i$05$i105) + 1)|0;
          $103 = ($102>>>0)<(3);
          $104 = ($101>>>0)<($2>>>0);
          $or$cond$i = $104 & $103;
          if (!($or$cond$i)) {
           $cur$0$be$i = $101;
           break L78;
          }
          $$pre$i5 = HEAP8[$101>>0]|0;
          $105 = $$pre$i5 & -8;
          $106 = ($105<<24>>24)==(48);
          if ($106) {
           $cur$16$i104 = $101;$i$05$i105 = $102;
          } else {
           $cur$0$be$i = $101;
           break;
          }
         }
        } else {
         $cur$0$be$i = $92;
        }
        break;
       }
       default: {
        $cur$0$be$i = $92;
       }
       }
      } while(0);
      $100 = ($cur$0$be$i>>>0)<($2>>>0);
      if ($100) {
       $cur$09$i = $cur$0$be$i;
      } else {
       $cur$4$i = $cur$0$be$i;$error$2$i = 3;
       break L74;
      }
     }
     if ((label|0) == 60) {
      label = 0;
      $107 = (($embed$0$ph15$i) + 1)|0;
      $109 = $$lcssa188;$embed$0$ph$be$i = $107;
     }
     else if ((label|0) == 62) {
      label = 0;
      $110 = (($embed$0$ph15$i) + -1)|0;
      $111 = ($110|0)==(0);
      if ($111) {
       $cur$4$i = $$lcssa189;$error$2$i = 0;
       break;
      } else {
       $109 = $$lcssa189;$embed$0$ph$be$i = $110;
      }
     }
     $108 = ($109>>>0)<($2>>>0);
     if ($108) {
      $cur$0$ph16$i = $109;$embed$0$ph15$i = $embed$0$ph$be$i;
     } else {
      $cur$4$i = $109;$error$2$i = 3;
      break;
     }
    }
    $112 = $cur$4$i;
    $cur$sroa$0$1 = $112;$error$0 = $error$2$i;
    break L12;
    break;
   }
   case 60:  {
    $113 = ((($15)) + 1|0);
    $114 = ($113>>>0)<($2>>>0);
    L94: do {
     if ($114) {
      $116 = HEAP8[$113>>0]|0;
      $117 = ($116<<24>>24)==(60);
      if ($117) {
       $119 = ((($15)) + 2|0);
       $120 = $119;
       $cur$sroa$0$1 = $120;$error$0 = 0;
       break L12;
      }
      $118 = $113;
      $121 = $116;$166 = $113;$cur$sroa$0$01$i$i = $118;
      while(1) {
       switch ($121<<24>>24) {
       case 0: case 12: case 9: case 10: case 13: case 32:  {
        $cur$sroa$0$1$i$i = $cur$sroa$0$01$i$i;
        label = 74;
        break;
       }
       case 37:  {
        $122 = 37;$cur$01$i$i$i = $166;
        L102: while(1) {
         switch ($122<<24>>24) {
         case 10: case 13:  {
          $cur$0$lcssa$i$i$i = $cur$01$i$i$i;
          break L102;
          break;
         }
         default: {
         }
         }
         $123 = ((($cur$01$i$i$i)) + 1|0);
         $124 = ($123>>>0)<($2>>>0);
         if (!($124)) {
          $cur$0$lcssa$i$i$i = $123;
          break;
         }
         $$pre$i$i = HEAP8[$123>>0]|0;
         $122 = $$pre$i$i;$cur$01$i$i$i = $123;
        }
        $125 = $cur$0$lcssa$i$i$i;
        $cur$sroa$0$1$i$i = $125;
        label = 74;
        break;
       }
       default: {
        $cur$sroa$0$0$lcssa$i$i = $cur$sroa$0$01$i$i;
        label = 76;
       }
       }
       if ((label|0) == 74) {
        label = 0;
        $126 = $cur$sroa$0$1$i$i;
        $127 = ((($126)) + 1|0);
        $128 = $127;
        $129 = ($127>>>0)<($2>>>0);
        if ($129) {
         $$be114 = $127;$cur$sroa$0$01$i$i$be = $128;
        } else {
         $cur$sroa$0$0$lcssa$i$i = $128;
         label = 76;
        }
       }
       if ((label|0) == 76) {
        label = 0;
        $130 = $cur$sroa$0$0$lcssa$i$i;
        $131 = ($130>>>0)<($2>>>0);
        if (!($131)) {
         $cur$sroa$0$1$i2 = $cur$sroa$0$0$lcssa$i$i;
         break L94;
        }
        $132 = HEAP8[$130>>0]|0;
        $$off$i = (($132) + -48)<<24>>24;
        $133 = ($$off$i&255)<(10);
        if (!($133)) {
         switch ($132<<24>>24) {
         case 65: case 66: case 67: case 68: case 69: case 70: case 97: case 98: case 99: case 100: case 101: case 102:  {
          break;
         }
         default: {
          $cur$sroa$0$1$i2 = $cur$sroa$0$0$lcssa$i$i;
          break L94;
         }
         }
        }
        $134 = ((($130)) + 1|0);
        $135 = $134;
        $136 = ($134>>>0)<($2>>>0);
        if ($136) {
         $$be114 = $134;$cur$sroa$0$01$i$i$be = $135;
        } else {
         $cur$sroa$0$1$i2 = $135;
         break L94;
        }
       }
       $$pre = HEAP8[$$be114>>0]|0;
       $121 = $$pre;$166 = $$be114;$cur$sroa$0$01$i$i = $cur$sroa$0$01$i$i$be;
      }
     } else {
      $115 = $113;
      $cur$sroa$0$1$i2 = $115;
     }
    } while(0);
    $137 = $cur$sroa$0$1$i2;
    $138 = ($137>>>0)<($2>>>0);
    if ($138) {
     $139 = HEAP8[$137>>0]|0;
     $140 = ($139<<24>>24)==(62);
     if (!($140)) {
      $cur$sroa$0$1 = $cur$sroa$0$1$i2;$error$0 = 3;
      break L12;
     }
    }
    $141 = ((($137)) + 1|0);
    $142 = $141;
    $cur$sroa$0$1 = $142;$error$0 = 0;
    break L12;
    break;
   }
   case 62:  {
    $143 = ((($15)) + 1|0);
    $144 = $143;
    $145 = ($143>>>0)<($2>>>0);
    if (!($145)) {
     $cur$sroa$0$1 = $144;$error$0 = 3;
     break L12;
    }
    $146 = HEAP8[$143>>0]|0;
    $147 = ($146<<24>>24)==(62);
    if (!($147)) {
     $cur$sroa$0$1 = $144;$error$0 = 3;
     break L12;
    }
    $148 = ((($15)) + 2|0);
    $149 = $148;
    $cur$sroa$0$1 = $149;$error$0 = 0;
    break L12;
    break;
   }
   case 47:  {
    $150 = ((($15)) + 1|0);
    $151 = $150;
    $cur$sroa$0$0$ph = $151;
    break;
   }
   default: {
    $cur$sroa$0$0$ph = $cur$sroa$0$0$lcssa$i;
   }
   }
   $152 = $cur$sroa$0$0$ph;
   $153 = ($152>>>0)<($2>>>0);
   if ($153) {
    $155 = $152;$cur$sroa$0$066 = $cur$sroa$0$0$ph;
    while(1) {
     $154 = HEAP8[$155>>0]|0;
     switch ($154<<24>>24) {
     case 37: case 125: case 123: case 93: case 91: case 62: case 60: case 41: case 40: case 47: case 0: case 12: case 9: case 10: case 13: case 32:  {
      $cur$sroa$0$1 = $cur$sroa$0$066;$error$0 = 0;
      break L12;
      break;
     }
     default: {
     }
     }
     $156 = ((($155)) + 1|0);
     $157 = $156;
     $158 = ($156>>>0)<($2>>>0);
     if ($158) {
      $155 = $156;$cur$sroa$0$066 = $157;
     } else {
      $cur$sroa$0$1 = $157;$error$0 = 0;
      break L12;
     }
    }
   } else {
    $cur$sroa$0$1 = $cur$sroa$0$0$ph;$error$0 = 0;
   }
  } else {
   $cur$sroa$0$1 = $cur$sroa$0$0$lcssa$i;$error$0 = 0;
  }
 } while(0);
 $159 = $cur$sroa$0$1;
 $160 = ($159>>>0)<($2>>>0);
 $161 = ($159|0)==($3|0);
 $162 = $160 & $161;
 $error$1 = $162 ? 3 : $error$0;
 $163 = ((($parser)) + 12|0);
 HEAP32[$163>>2] = $error$1;
 HEAP32[$parser>>2] = $cur$sroa$0$1;
 return;
}
function _ps_parser_to_int($parser) {
 $parser = $parser|0;
 var $$0$i = 0, $$cast = 0, $$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $9 = 0, $base$off$i = 0, $cur$0$lcssa$i$i$i = 0, $cur$01$i$i$i = 0, $cur$sroa$0$0$lcssa$i$i = 0, $cur$sroa$0$01$i$i = 0, $cur$sroa$0$1$i$i = 0, $have_overflow$0$lcssa$i = 0, $have_overflow$0$lcssa$i14 = 0, $have_overflow$07$i = 0, $have_overflow$07$i7 = 0, $have_overflow$2$ph$i = 0, $have_overflow$2$ph$i10 = 0, $num$0$i = 0, $num$0$lcssa$i = 0, $num$0$lcssa$i13 = 0, $num$08$i = 0, $num$08$i6 = 0;
 var $num$2$ph$i = 0, $num$2$ph$i11 = 0, $num$3$$i = 0, $num$3$$i15 = 0, $num$5$i = 0, $num$5$i16 = 0, $or$cond$i = 0, $or$cond4$i = 0, $or$cond44$i = 0, $or$cond44$i9 = 0, $p$0$i = 0, $p$0$i3 = 0, $p$1$lcssa$i = 0, $p$1$lcssa$i12 = 0, $p$19$i = 0, $p$19$i5 = 0, $sext$i = 0, $sign$0$i = 0, $sign$0$i2 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($parser)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$parser>>2]|0;
 $3 = $2;
 $4 = ($3>>>0)<($1>>>0);
 L1: do {
  if ($4) {
   $6 = $3;$cur$sroa$0$01$i$i = $2;
   while(1) {
    $5 = HEAP8[$6>>0]|0;
    switch ($5<<24>>24) {
    case 0: case 12: case 9: case 10: case 13: case 32:  {
     $cur$sroa$0$1$i$i = $cur$sroa$0$01$i$i;
     break;
    }
    case 37:  {
     $7 = 37;$cur$01$i$i$i = $6;
     L5: while(1) {
      switch ($7<<24>>24) {
      case 10: case 13:  {
       $cur$0$lcssa$i$i$i = $cur$01$i$i$i;
       break L5;
       break;
      }
      default: {
      }
      }
      $8 = ((($cur$01$i$i$i)) + 1|0);
      $9 = ($8>>>0)<($1>>>0);
      if (!($9)) {
       $cur$0$lcssa$i$i$i = $8;
       break;
      }
      $$pre$i$i = HEAP8[$8>>0]|0;
      $7 = $$pre$i$i;$cur$01$i$i$i = $8;
     }
     $10 = $cur$0$lcssa$i$i$i;
     $cur$sroa$0$1$i$i = $10;
     break;
    }
    default: {
     $cur$sroa$0$0$lcssa$i$i = $cur$sroa$0$01$i$i;
     break L1;
    }
    }
    $11 = $cur$sroa$0$1$i$i;
    $12 = ((($11)) + 1|0);
    $13 = $12;
    $14 = ($12>>>0)<($1>>>0);
    if ($14) {
     $6 = $12;$cur$sroa$0$01$i$i = $13;
    } else {
     $cur$sroa$0$0$lcssa$i$i = $13;
     break L1;
    }
   }
  } else {
   $cur$sroa$0$0$lcssa$i$i = $2;
  }
 } while(0);
 HEAP32[$parser>>2] = $cur$sroa$0$0$lcssa$i$i;
 $$cast = $cur$sroa$0$0$lcssa$i$i;
 $15 = ($$cast>>>0)<($1>>>0);
 if (!($15)) {
  $$0$i = 0;
  return ($$0$i|0);
 }
 $16 = HEAP8[$$cast>>0]|0;
 switch ($16<<24>>24) {
 case 43: case 45:  {
  $17 = ($16<<24>>24)==(45);
  $18 = $17&1;
  $19 = ((($$cast)) + 1|0);
  $20 = ($19|0)==($1|0);
  if ($20) {
   $$0$i = 0;
   return ($$0$i|0);
  } else {
   $p$0$i3 = $19;$sign$0$i2 = $18;
  }
  break;
 }
 default: {
  $p$0$i3 = $$cast;$sign$0$i2 = 0;
 }
 }
 $21 = ($p$0$i3>>>0)<($1>>>0);
 L19: do {
  if ($21) {
   $have_overflow$07$i7 = 0;$num$08$i6 = 0;$p$19$i5 = $p$0$i3;
   while(1) {
    $22 = HEAP8[$p$19$i5>>0]|0;
    $23 = $22&255;
    switch ($22<<24>>24) {
    case 0: case 12: case 9: case 10: case 13: case 32:  {
     $have_overflow$0$lcssa$i14 = $have_overflow$07$i7;$num$0$lcssa$i13 = $num$08$i6;$p$1$lcssa$i12 = $p$19$i5;
     break L19;
     break;
    }
    default: {
    }
    }
    $24 = ($22<<24>>24)<(0);
    if ($24) {
     $have_overflow$0$lcssa$i14 = $have_overflow$07$i7;$num$0$lcssa$i13 = $num$08$i6;$p$1$lcssa$i12 = $p$19$i5;
     break L19;
    }
    $25 = $23 & 127;
    $26 = (532577 + ($25)|0);
    $27 = HEAP8[$26>>0]|0;
    $28 = $27 << 24 >> 24;
    $29 = (($25) + -48)|0;
    $30 = ($29>>>0)<(10);
    if (!($30)) {
     $have_overflow$0$lcssa$i14 = $have_overflow$07$i7;$num$0$lcssa$i13 = $num$08$i6;$p$1$lcssa$i12 = $p$19$i5;
     break L19;
    }
    $31 = ($num$08$i6|0)>(214748364);
    if ($31) {
     $have_overflow$2$ph$i10 = 1;$num$2$ph$i11 = $num$08$i6;
    } else {
     $32 = ($num$08$i6|0)==(214748364);
     $33 = ($27<<24>>24)>(7);
     $or$cond44$i9 = $32 & $33;
     if ($or$cond44$i9) {
      $have_overflow$2$ph$i10 = 1;$num$2$ph$i11 = 214748364;
     } else {
      $34 = ($num$08$i6*10)|0;
      $35 = (($28) + ($34))|0;
      $have_overflow$2$ph$i10 = $have_overflow$07$i7;$num$2$ph$i11 = $35;
     }
    }
    $36 = ((($p$19$i5)) + 1|0);
    $37 = ($36>>>0)<($1>>>0);
    if ($37) {
     $have_overflow$07$i7 = $have_overflow$2$ph$i10;$num$08$i6 = $num$2$ph$i11;$p$19$i5 = $36;
    } else {
     $have_overflow$0$lcssa$i14 = $have_overflow$2$ph$i10;$num$0$lcssa$i13 = $num$2$ph$i11;$p$1$lcssa$i12 = $36;
     break L19;
    }
   }
  } else {
   $have_overflow$0$lcssa$i14 = 0;$num$0$lcssa$i13 = 0;$p$1$lcssa$i12 = $p$0$i3;
  }
 } while(0);
 $38 = ($have_overflow$0$lcssa$i14<<24>>24)==(0);
 $num$3$$i15 = $38 ? $num$0$lcssa$i13 : 2147483647;
 $39 = ($sign$0$i2<<24>>24)==(0);
 $40 = (0 - ($num$3$$i15))|0;
 $num$5$i16 = $39 ? $num$3$$i15 : $40;
 $41 = ($p$1$lcssa$i12|0)==($$cast|0);
 $42 = $p$1$lcssa$i12;
 if ($41) {
  $$0$i = 0;
  return ($$0$i|0);
 }
 $43 = ($p$1$lcssa$i12>>>0)<($1>>>0);
 if ($43) {
  $44 = HEAP8[$p$1$lcssa$i12>>0]|0;
  $45 = ($44<<24>>24)==(35);
  if ($45) {
   $46 = ((($p$1$lcssa$i12)) + 1|0);
   $47 = ($46>>>0)>=($1>>>0);
   $base$off$i = (($num$5$i16) + -2)|0;
   $48 = ($base$off$i>>>0)>(34);
   $or$cond4$i = $47 | $48;
   if ($or$cond4$i) {
    $$0$i = 0;
    return ($$0$i|0);
   }
   $49 = HEAP8[$46>>0]|0;
   switch ($49<<24>>24) {
   case 43: case 45:  {
    $50 = ($49<<24>>24)==(45);
    $51 = $50&1;
    $52 = ((($p$1$lcssa$i12)) + 2|0);
    $53 = ($52|0)==($1|0);
    if ($53) {
     $$0$i = 0;
     return ($$0$i|0);
    } else {
     $p$0$i = $52;$sign$0$i = $51;
    }
    break;
   }
   default: {
    $p$0$i = $46;$sign$0$i = 0;
   }
   }
   $54 = (2147483647 / ($num$5$i16|0))&-1;
   $55 = ($p$0$i>>>0)<($1>>>0);
   L43: do {
    if ($55) {
     $56 = (2147483647 % ($num$5$i16|0))&-1;
     $sext$i = $56 << 24;
     $57 = $sext$i >> 24;
     $have_overflow$07$i = 0;$num$08$i = 0;$p$19$i = $p$0$i;
     while(1) {
      $58 = HEAP8[$p$19$i>>0]|0;
      $59 = $58&255;
      switch ($58<<24>>24) {
      case 0: case 12: case 9: case 10: case 13: case 32:  {
       $have_overflow$0$lcssa$i = $have_overflow$07$i;$num$0$lcssa$i = $num$08$i;$p$1$lcssa$i = $p$19$i;
       break L43;
       break;
      }
      default: {
      }
      }
      $60 = ($58<<24>>24)<(0);
      if ($60) {
       $have_overflow$0$lcssa$i = $have_overflow$07$i;$num$0$lcssa$i = $num$08$i;$p$1$lcssa$i = $p$19$i;
       break L43;
      }
      $61 = $59 & 127;
      $62 = (532577 + ($61)|0);
      $63 = HEAP8[$62>>0]|0;
      $64 = $63 << 24 >> 24;
      $65 = ($63<<24>>24)>(-1);
      $66 = ($64|0)<($num$5$i16|0);
      $or$cond$i = $65 & $66;
      if (!($or$cond$i)) {
       $have_overflow$0$lcssa$i = $have_overflow$07$i;$num$0$lcssa$i = $num$08$i;$p$1$lcssa$i = $p$19$i;
       break L43;
      }
      $67 = ($num$08$i|0)>($54|0);
      if ($67) {
       $have_overflow$2$ph$i = 1;$num$2$ph$i = $num$08$i;
      } else {
       $68 = ($num$08$i|0)==($54|0);
       $69 = ($64|0)>($57|0);
       $or$cond44$i = $68 & $69;
       if ($or$cond44$i) {
        $have_overflow$2$ph$i = 1;$num$2$ph$i = $54;
       } else {
        $70 = Math_imul($num$08$i, $num$5$i16)|0;
        $71 = (($64) + ($70))|0;
        $have_overflow$2$ph$i = $have_overflow$07$i;$num$2$ph$i = $71;
       }
      }
      $72 = ((($p$19$i)) + 1|0);
      $73 = ($72>>>0)<($1>>>0);
      if ($73) {
       $have_overflow$07$i = $have_overflow$2$ph$i;$num$08$i = $num$2$ph$i;$p$19$i = $72;
      } else {
       $have_overflow$0$lcssa$i = $have_overflow$2$ph$i;$num$0$lcssa$i = $num$2$ph$i;$p$1$lcssa$i = $72;
       break L43;
      }
     }
    } else {
     $have_overflow$0$lcssa$i = 0;$num$0$lcssa$i = 0;$p$1$lcssa$i = $p$0$i;
    }
   } while(0);
   $74 = ($have_overflow$0$lcssa$i<<24>>24)==(0);
   $num$3$$i = $74 ? $num$0$lcssa$i : 2147483647;
   $75 = ($sign$0$i<<24>>24)==(0);
   $76 = (0 - ($num$3$$i))|0;
   $num$5$i = $75 ? $num$3$$i : $76;
   $77 = ($p$1$lcssa$i|0)==($46|0);
   $78 = $p$1$lcssa$i;
   if ($77) {
    $$0$i = 0;
    return ($$0$i|0);
   } else {
    $79 = $78;$num$0$i = $num$5$i;
   }
  } else {
   $79 = $42;$num$0$i = $num$5$i16;
  }
 } else {
  $79 = $42;$num$0$i = $num$5$i16;
 }
 HEAP32[$parser>>2] = $79;
 $$0$i = $num$0$i;
 return ($$0$i|0);
}
function _ps_parser_to_fixed($parser,$power_ten) {
 $parser = $parser|0;
 $power_ten = $power_ten|0;
 var $$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cur$0$lcssa$i$i$i = 0, $cur$01$i$i$i = 0, $cur$sroa$0$0$lcssa$i$i = 0;
 var $cur$sroa$0$01$i$i = 0, $cur$sroa$0$1$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($parser)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$parser>>2]|0;
 $3 = $2;
 $4 = ($3>>>0)<($1>>>0);
 L1: do {
  if ($4) {
   $6 = $3;$cur$sroa$0$01$i$i = $2;
   while(1) {
    $5 = HEAP8[$6>>0]|0;
    switch ($5<<24>>24) {
    case 0: case 12: case 9: case 10: case 13: case 32:  {
     $cur$sroa$0$1$i$i = $cur$sroa$0$01$i$i;
     break;
    }
    case 37:  {
     $7 = 37;$cur$01$i$i$i = $6;
     L5: while(1) {
      switch ($7<<24>>24) {
      case 10: case 13:  {
       $cur$0$lcssa$i$i$i = $cur$01$i$i$i;
       break L5;
       break;
      }
      default: {
      }
      }
      $8 = ((($cur$01$i$i$i)) + 1|0);
      $9 = ($8>>>0)<($1>>>0);
      if (!($9)) {
       $cur$0$lcssa$i$i$i = $8;
       break;
      }
      $$pre$i$i = HEAP8[$8>>0]|0;
      $7 = $$pre$i$i;$cur$01$i$i$i = $8;
     }
     $10 = $cur$0$lcssa$i$i$i;
     $cur$sroa$0$1$i$i = $10;
     break;
    }
    default: {
     $cur$sroa$0$0$lcssa$i$i = $cur$sroa$0$01$i$i;
     break L1;
    }
    }
    $11 = $cur$sroa$0$1$i$i;
    $12 = ((($11)) + 1|0);
    $13 = $12;
    $14 = ($12>>>0)<($1>>>0);
    if ($14) {
     $6 = $12;$cur$sroa$0$01$i$i = $13;
    } else {
     $cur$sroa$0$0$lcssa$i$i = $13;
     break L1;
    }
   }
  } else {
   $cur$sroa$0$0$lcssa$i$i = $2;
  }
 } while(0);
 HEAP32[$parser>>2] = $cur$sroa$0$0$lcssa$i$i;
 $15 = (_PS_Conv_ToFixed($parser,$1,$power_ten)|0);
 return ($15|0);
}
function _PS_Conv_ToFixed($cursor,$limit,$power_ten) {
 $cursor = $cursor|0;
 $limit = $limit|0;
 $power_ten = $power_ten|0;
 var $$ = 0, $$0$i = 0, $$0$i36 = 0, $$08$ph$lcssa = 0, $$08$ph$ph228 = 0, $$08$ph$us$lcssa = 0, $$08$ph$us217 = 0, $$08$ph$us217$lcssa264 = 0, $$08$ph$us217$lcssa264$lcssa = 0, $$1 = 0, $$13 = 0, $$14 = 0, $$3 = 0, $$3$ = 0, $$6 = 0, $$7$lcssa = 0, $$7106 = 0, $$8100 = 0, $$lcssa270 = 0, $$lcssa270$lcssa = 0;
 var $$lcssa286 = 0, $$pn = 0, $$pn$ph$ph225 = 0, $$pn$ph$us$lcssa = 0, $$pn$ph$us215 = 0, $$pn$us116 = 0, $$pn$us129 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0;
 var $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0;
 var $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $base$off$i = 0, $base$off$i38 = 0, $decimal$0$ph$lcssa = 0, $decimal$0$ph$ph226 = 0, $decimal$0$ph$us$lcssa = 0, $decimal$0$ph$us216 = 0, $decimal$3 = 0, $decimal$4$lcssa = 0, $decimal$4104 = 0;
 var $decimal$5 = 0, $decimal$6$lcssa = 0, $decimal$698 = 0, $decimal$7 = 0, $divider$0$ph$ph212 = 0, $divider$0$ph$ph227 = 0, $divider$0$ph$ph227279 = 0, $divider$3 = 0, $divider$4$lcssa = 0, $divider$4105 = 0, $divider$5 = 0, $divider$6$lcssa = 0, $divider$699 = 0, $divider$7 = 0, $have_overflow$0 = 0, $have_overflow$0$lcssa$i = 0, $have_overflow$0$lcssa$i28 = 0, $have_overflow$0$lcssa$i53 = 0, $have_overflow$0$lcssa$i71 = 0, $have_overflow$07$i = 0;
 var $have_overflow$07$i21 = 0, $have_overflow$07$i46 = 0, $have_overflow$07$i64 = 0, $have_overflow$2$ph$i = 0, $have_overflow$2$ph$i24 = 0, $have_overflow$2$ph$i49 = 0, $have_overflow$2$ph$i67 = 0, $have_overflow$3 = 0, $have_underflow$2 = 0, $integral$0 = 0, $integral$1$lcssa = 0, $integral$1103 = 0, $integral$2$lcssa = 0, $integral$2$off = 0, $integral$297 = 0, $integral$3 = 0, $integral$3$ = 0, $num$0$i = 0, $num$0$i35 = 0, $num$0$lcssa$i = 0;
 var $num$0$lcssa$i27 = 0, $num$0$lcssa$i52 = 0, $num$0$lcssa$i70 = 0, $num$08$i = 0, $num$08$i20 = 0, $num$08$i45 = 0, $num$08$i63 = 0, $num$2$ph$i = 0, $num$2$ph$i25 = 0, $num$2$ph$i50 = 0, $num$2$ph$i68 = 0, $num$3$$i = 0, $num$3$$i29 = 0, $num$3$$i54 = 0, $num$3$$i72 = 0, $num$5$i = 0, $num$5$i30 = 0, $num$5$i55 = 0, $num$5$i73 = 0, $or$cond$i = 0;
 var $or$cond$i47 = 0, $or$cond3$us = 0, $or$cond4$i = 0, $or$cond4$i39 = 0, $or$cond44$i = 0, $or$cond44$i23 = 0, $or$cond44$i48 = 0, $or$cond44$i66 = 0, $or$cond7 = 0, $p$0$i = 0, $p$0$i17 = 0, $p$0$i41 = 0, $p$0$i60 = 0, $p$1$lcssa$i = 0, $p$1$lcssa$i26 = 0, $p$1$lcssa$i51 = 0, $p$1$lcssa$i69 = 0, $p$19$i = 0, $p$19$i19 = 0, $p$19$i44 = 0;
 var $p$19$i62 = 0, $p$sroa$0$0 = 0, $p$sroa$0$0$in = 0, $p$sroa$0$1 = 0, $p$sroa$0$2 = 0, $p$sroa$0$3 = 0, $p$sroa$0$4 = 0, $p$sroa$0$5 = 0, $sext$i = 0, $sext$i42 = 0, $sign$0 = 0, $sign$0$i = 0, $sign$0$i16 = 0, $sign$0$i40 = 0, $sign$0$i59 = 0, $storemerge = 0, $storemerge$lcssa = 0, $storemerge$us$us = 0, $storemerge$us$us$lcssa268 = 0, $storemerge$us$us$lcssa268$lcssa = 0;
 var $storemerge$us117 = 0, $storemerge$us130 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$cursor>>2]|0;
 $1 = ($0>>>0)<($limit>>>0);
 if (!($1)) {
  $$1 = 0;
  return ($$1|0);
 }
 $2 = HEAP8[$0>>0]|0;
 L4: do {
  switch ($2<<24>>24) {
  case 43: case 45:  {
   $3 = ((($0)) + 1|0);
   $4 = ($3|0)==($limit|0);
   if ($4) {
    $$1 = 0;
    return ($$1|0);
   } else {
    $5 = ($2<<24>>24)==(45);
    $6 = $5&1;
    $$pre = HEAP8[$3>>0]|0;
    $8 = $$pre;$p$sroa$0$0$in = $3;$sign$0 = $6;
    break L4;
   }
   break;
  }
  default: {
   $8 = $2;$p$sroa$0$0$in = $0;$sign$0 = 0;
  }
  }
 } while(0);
 $p$sroa$0$0 = $p$sroa$0$0$in;
 $7 = ($8<<24>>24)==(46);
 do {
  if ($7) {
   $have_overflow$0 = 0;$integral$0 = 0;$p$sroa$0$2 = $p$sroa$0$0;
  } else {
   $9 = ($p$sroa$0$0$in>>>0)<($limit>>>0);
   L12: do {
    if ($9) {
     switch ($8<<24>>24) {
     case 43: case 45:  {
      $10 = ($8<<24>>24)==(45);
      $11 = $10&1;
      $12 = ((($p$sroa$0$0$in)) + 1|0);
      $13 = ($12|0)==($limit|0);
      if ($13) {
       $$0$i = 0;$p$sroa$0$1 = $p$sroa$0$0;
       break L12;
      } else {
       $p$0$i17 = $12;$sign$0$i16 = $11;
      }
      break;
     }
     default: {
      $p$0$i17 = $p$sroa$0$0$in;$sign$0$i16 = 0;
     }
     }
     $14 = ($p$0$i17>>>0)<($limit>>>0);
     L17: do {
      if ($14) {
       $have_overflow$07$i21 = 0;$num$08$i20 = 0;$p$19$i19 = $p$0$i17;
       while(1) {
        $15 = HEAP8[$p$19$i19>>0]|0;
        $16 = $15&255;
        switch ($15<<24>>24) {
        case 0: case 12: case 9: case 10: case 13: case 32:  {
         $have_overflow$0$lcssa$i28 = $have_overflow$07$i21;$num$0$lcssa$i27 = $num$08$i20;$p$1$lcssa$i26 = $p$19$i19;
         break L17;
         break;
        }
        default: {
        }
        }
        $17 = ($15<<24>>24)<(0);
        if ($17) {
         $have_overflow$0$lcssa$i28 = $have_overflow$07$i21;$num$0$lcssa$i27 = $num$08$i20;$p$1$lcssa$i26 = $p$19$i19;
         break L17;
        }
        $18 = $16 & 127;
        $19 = (532577 + ($18)|0);
        $20 = HEAP8[$19>>0]|0;
        $21 = $20 << 24 >> 24;
        $22 = (($18) + -48)|0;
        $23 = ($22>>>0)<(10);
        if (!($23)) {
         $have_overflow$0$lcssa$i28 = $have_overflow$07$i21;$num$0$lcssa$i27 = $num$08$i20;$p$1$lcssa$i26 = $p$19$i19;
         break L17;
        }
        $24 = ($num$08$i20|0)>(214748364);
        if ($24) {
         $have_overflow$2$ph$i24 = 1;$num$2$ph$i25 = $num$08$i20;
        } else {
         $25 = ($num$08$i20|0)==(214748364);
         $26 = ($20<<24>>24)>(7);
         $or$cond44$i23 = $25 & $26;
         if ($or$cond44$i23) {
          $have_overflow$2$ph$i24 = 1;$num$2$ph$i25 = 214748364;
         } else {
          $27 = ($num$08$i20*10)|0;
          $28 = (($21) + ($27))|0;
          $have_overflow$2$ph$i24 = $have_overflow$07$i21;$num$2$ph$i25 = $28;
         }
        }
        $29 = ((($p$19$i19)) + 1|0);
        $30 = ($29>>>0)<($limit>>>0);
        if ($30) {
         $have_overflow$07$i21 = $have_overflow$2$ph$i24;$num$08$i20 = $num$2$ph$i25;$p$19$i19 = $29;
        } else {
         $have_overflow$0$lcssa$i28 = $have_overflow$2$ph$i24;$num$0$lcssa$i27 = $num$2$ph$i25;$p$1$lcssa$i26 = $29;
         break L17;
        }
       }
      } else {
       $have_overflow$0$lcssa$i28 = 0;$num$0$lcssa$i27 = 0;$p$1$lcssa$i26 = $p$0$i17;
      }
     } while(0);
     $31 = ($have_overflow$0$lcssa$i28<<24>>24)==(0);
     $num$3$$i29 = $31 ? $num$0$lcssa$i27 : 2147483647;
     $32 = ($sign$0$i16<<24>>24)==(0);
     $33 = (0 - ($num$3$$i29))|0;
     $num$5$i30 = $32 ? $num$3$$i29 : $33;
     $34 = ($p$1$lcssa$i26|0)==($p$sroa$0$0$in|0);
     $35 = $p$1$lcssa$i26;
     if ($34) {
      $$0$i = 0;$p$sroa$0$1 = $p$sroa$0$0;
     } else {
      $36 = ($p$1$lcssa$i26>>>0)<($limit>>>0);
      if ($36) {
       $37 = HEAP8[$p$1$lcssa$i26>>0]|0;
       $38 = ($37<<24>>24)==(35);
       if ($38) {
        $39 = ((($p$1$lcssa$i26)) + 1|0);
        $40 = ($39>>>0)>=($limit>>>0);
        $base$off$i = (($num$5$i30) + -2)|0;
        $41 = ($base$off$i>>>0)>(34);
        $or$cond4$i = $40 | $41;
        if ($or$cond4$i) {
         $$0$i = 0;$p$sroa$0$1 = $p$sroa$0$0;
         break;
        }
        $42 = HEAP8[$39>>0]|0;
        switch ($42<<24>>24) {
        case 43: case 45:  {
         $43 = ($42<<24>>24)==(45);
         $44 = $43&1;
         $45 = ((($p$1$lcssa$i26)) + 2|0);
         $46 = ($45|0)==($limit|0);
         if ($46) {
          $$0$i = 0;$p$sroa$0$1 = $p$sroa$0$0;
          break L12;
         } else {
          $p$0$i = $45;$sign$0$i = $44;
         }
         break;
        }
        default: {
         $p$0$i = $39;$sign$0$i = 0;
        }
        }
        $47 = (2147483647 / ($num$5$i30|0))&-1;
        $48 = ($p$0$i>>>0)<($limit>>>0);
        L36: do {
         if ($48) {
          $49 = (2147483647 % ($num$5$i30|0))&-1;
          $sext$i = $49 << 24;
          $50 = $sext$i >> 24;
          $have_overflow$07$i = 0;$num$08$i = 0;$p$19$i = $p$0$i;
          while(1) {
           $51 = HEAP8[$p$19$i>>0]|0;
           $52 = $51&255;
           switch ($51<<24>>24) {
           case 0: case 12: case 9: case 10: case 13: case 32:  {
            $have_overflow$0$lcssa$i = $have_overflow$07$i;$num$0$lcssa$i = $num$08$i;$p$1$lcssa$i = $p$19$i;
            break L36;
            break;
           }
           default: {
           }
           }
           $53 = ($51<<24>>24)<(0);
           if ($53) {
            $have_overflow$0$lcssa$i = $have_overflow$07$i;$num$0$lcssa$i = $num$08$i;$p$1$lcssa$i = $p$19$i;
            break L36;
           }
           $54 = $52 & 127;
           $55 = (532577 + ($54)|0);
           $56 = HEAP8[$55>>0]|0;
           $57 = $56 << 24 >> 24;
           $58 = ($56<<24>>24)>(-1);
           $59 = ($57|0)<($num$5$i30|0);
           $or$cond$i = $58 & $59;
           if (!($or$cond$i)) {
            $have_overflow$0$lcssa$i = $have_overflow$07$i;$num$0$lcssa$i = $num$08$i;$p$1$lcssa$i = $p$19$i;
            break L36;
           }
           $60 = ($num$08$i|0)>($47|0);
           if ($60) {
            $have_overflow$2$ph$i = 1;$num$2$ph$i = $num$08$i;
           } else {
            $61 = ($num$08$i|0)==($47|0);
            $62 = ($57|0)>($50|0);
            $or$cond44$i = $61 & $62;
            if ($or$cond44$i) {
             $have_overflow$2$ph$i = 1;$num$2$ph$i = $47;
            } else {
             $63 = Math_imul($num$08$i, $num$5$i30)|0;
             $64 = (($57) + ($63))|0;
             $have_overflow$2$ph$i = $have_overflow$07$i;$num$2$ph$i = $64;
            }
           }
           $65 = ((($p$19$i)) + 1|0);
           $66 = ($65>>>0)<($limit>>>0);
           if ($66) {
            $have_overflow$07$i = $have_overflow$2$ph$i;$num$08$i = $num$2$ph$i;$p$19$i = $65;
           } else {
            $have_overflow$0$lcssa$i = $have_overflow$2$ph$i;$num$0$lcssa$i = $num$2$ph$i;$p$1$lcssa$i = $65;
            break L36;
           }
          }
         } else {
          $have_overflow$0$lcssa$i = 0;$num$0$lcssa$i = 0;$p$1$lcssa$i = $p$0$i;
         }
        } while(0);
        $67 = ($have_overflow$0$lcssa$i<<24>>24)==(0);
        $num$3$$i = $67 ? $num$0$lcssa$i : 2147483647;
        $68 = ($sign$0$i<<24>>24)==(0);
        $69 = (0 - ($num$3$$i))|0;
        $num$5$i = $68 ? $num$3$$i : $69;
        $70 = ($p$1$lcssa$i|0)==($39|0);
        $71 = $p$1$lcssa$i;
        if ($70) {
         $$0$i = 0;$p$sroa$0$1 = $p$sroa$0$0;
         break;
        } else {
         $221 = $71;$num$0$i = $num$5$i;
        }
       } else {
        $221 = $35;$num$0$i = $num$5$i30;
       }
      } else {
       $221 = $35;$num$0$i = $num$5$i30;
      }
      $$0$i = $num$0$i;$p$sroa$0$1 = $221;
     }
    } else {
     $$0$i = 0;$p$sroa$0$1 = $p$sroa$0$0;
    }
   } while(0);
   $72 = $p$sroa$0$1;
   $73 = ($72|0)==($p$sroa$0$0$in|0);
   if ($73) {
    $$1 = 0;
    return ($$1|0);
   } else {
    $74 = ($$0$i|0)>(32767);
    $75 = $$0$i << 16;
    $$ = $74 ? $$0$i : $75;
    $$13 = $74&1;
    $have_overflow$0 = $$13;$integral$0 = $$;$p$sroa$0$2 = $p$sroa$0$1;
    break;
   }
  }
 } while(0);
 $76 = $p$sroa$0$2;
 $77 = ($76>>>0)<($limit>>>0);
 if ($77) {
  $78 = HEAP8[$76>>0]|0;
  $79 = ($78<<24>>24)==(46);
  if ($79) {
   $80 = ($integral$0|0)==(0);
   $$08$ph$ph228 = $power_ten;$$pn$ph$ph225 = $76;$decimal$0$ph$ph226 = 0;$divider$0$ph$ph227 = 1;
   L57: while(1) {
    $81 = ($decimal$0$ph$ph226|0)<(214748364);
    if ($81) {
     $$08$ph$us217 = $$08$ph$ph228;$$pn$ph$us215 = $$pn$ph$ph225;$decimal$0$ph$us216 = $decimal$0$ph$ph226;
    } else {
     $$08$ph$us$lcssa = $$08$ph$ph228;$$pn$ph$us$lcssa = $$pn$ph$ph225;$decimal$0$ph$us$lcssa = $decimal$0$ph$ph226;$divider$0$ph$ph227279 = $divider$0$ph$ph227;
     label = 40;
     break;
    }
    while(1) {
     $storemerge$us$us = ((($$pn$ph$us215)) + 1|0);
     $98 = ($storemerge$us$us>>>0)<($limit>>>0);
     if (!($98)) {
      $$08$ph$lcssa = $$08$ph$us217;$decimal$0$ph$lcssa = $decimal$0$ph$us216;$divider$0$ph$ph212 = $divider$0$ph$ph227;$storemerge$lcssa = $storemerge$us$us;
      break L57;
     }
     $92 = HEAP8[$storemerge$us$us>>0]|0;
     $93 = $92&255;
     switch ($92<<24>>24) {
     case 0: case 12: case 9: case 10: case 13: case 32:  {
      $$08$ph$lcssa = $$08$ph$us217;$decimal$0$ph$lcssa = $decimal$0$ph$us216;$divider$0$ph$ph212 = $divider$0$ph$ph227;$storemerge$lcssa = $storemerge$us$us;
      break L57;
      break;
     }
     default: {
     }
     }
     $94 = ($92<<24>>24)<(0);
     if ($94) {
      $$08$ph$lcssa = $$08$ph$us217;$decimal$0$ph$lcssa = $decimal$0$ph$us216;$divider$0$ph$ph212 = $divider$0$ph$ph227;$storemerge$lcssa = $storemerge$us$us;
      break L57;
     }
     $95 = $93 & 127;
     $96 = (($95) + -48)|0;
     $97 = ($96>>>0)>(9);
     if ($97) {
      $$08$ph$lcssa = $$08$ph$us217;$decimal$0$ph$lcssa = $decimal$0$ph$us216;$divider$0$ph$ph212 = $divider$0$ph$ph227;$storemerge$lcssa = $storemerge$us$us;
      break L57;
     }
     $99 = (532577 + ($95)|0);
     $100 = HEAP8[$99>>0]|0;
     $101 = $100 << 24 >> 24;
     $102 = ($decimal$0$ph$us216*10)|0;
     $90 = (($101) + ($102))|0;
     $103 = ($$08$ph$us217|0)>(0);
     $or$cond3$us = $80 & $103;
     if (!($or$cond3$us)) {
      $$08$ph$us217$lcssa264 = $$08$ph$us217;$$lcssa270 = $90;$storemerge$us$us$lcssa268 = $storemerge$us$us;
      break;
     }
     $88 = (($$08$ph$us217) + -1)|0;
     $89 = ($90|0)<(214748364);
     if ($89) {
      $$08$ph$us217 = $88;$$pn$ph$us215 = $storemerge$us$us;$decimal$0$ph$us216 = $90;
     } else {
      $$08$ph$us$lcssa = $88;$$pn$ph$us$lcssa = $storemerge$us$us;$decimal$0$ph$us$lcssa = $90;$divider$0$ph$ph227279 = $divider$0$ph$ph227;
      label = 40;
      break L57;
     }
    }
    $118 = ($divider$0$ph$ph227*10)|0;
    $119 = ($118|0)<(214748364);
    if ($119) {
     $$08$ph$ph228 = $$08$ph$us217$lcssa264;$$pn$ph$ph225 = $storemerge$us$us$lcssa268;$decimal$0$ph$ph226 = $$lcssa270;$divider$0$ph$ph227 = $118;
    } else {
     $$08$ph$us217$lcssa264$lcssa = $$08$ph$us217$lcssa264;$$lcssa270$lcssa = $$lcssa270;$$lcssa286 = $118;$storemerge$us$us$lcssa268$lcssa = $storemerge$us$us$lcssa268;
     label = 51;
     break;
    }
   }
   L67: do {
    if ((label|0) == 40) {
     $$pn$us116 = $$pn$ph$us$lcssa;
     while(1) {
      $storemerge$us117 = ((($$pn$us116)) + 1|0);
      $91 = ($storemerge$us117>>>0)<($limit>>>0);
      if (!($91)) {
       $$08$ph$lcssa = $$08$ph$us$lcssa;$decimal$0$ph$lcssa = $decimal$0$ph$us$lcssa;$divider$0$ph$ph212 = $divider$0$ph$ph227279;$storemerge$lcssa = $storemerge$us117;
       break L67;
      }
      $82 = HEAP8[$storemerge$us117>>0]|0;
      $83 = $82&255;
      switch ($82<<24>>24) {
      case 0: case 12: case 9: case 10: case 13: case 32:  {
       $$08$ph$lcssa = $$08$ph$us$lcssa;$decimal$0$ph$lcssa = $decimal$0$ph$us$lcssa;$divider$0$ph$ph212 = $divider$0$ph$ph227279;$storemerge$lcssa = $storemerge$us117;
       break L67;
       break;
      }
      default: {
      }
      }
      $84 = ($82<<24>>24)<(0);
      if ($84) {
       $$08$ph$lcssa = $$08$ph$us$lcssa;$decimal$0$ph$lcssa = $decimal$0$ph$us$lcssa;$divider$0$ph$ph212 = $divider$0$ph$ph227279;$storemerge$lcssa = $storemerge$us117;
       break L67;
      }
      $85 = $83 & 127;
      $86 = (($85) + -48)|0;
      $87 = ($86>>>0)>(9);
      if ($87) {
       $$08$ph$lcssa = $$08$ph$us$lcssa;$decimal$0$ph$lcssa = $decimal$0$ph$us$lcssa;$divider$0$ph$ph212 = $divider$0$ph$ph227279;$storemerge$lcssa = $storemerge$us117;
       break L67;
      } else {
       $$pn$us116 = $storemerge$us117;
      }
     }
    }
    else if ((label|0) == 51) {
     if ($80) {
      $$pn$us129 = $storemerge$us$us$lcssa268$lcssa;
      while(1) {
       $storemerge$us130 = ((($$pn$us129)) + 1|0);
       $110 = ($storemerge$us130>>>0)<($limit>>>0);
       if (!($110)) {
        $$08$ph$lcssa = $$08$ph$us217$lcssa264$lcssa;$decimal$0$ph$lcssa = $$lcssa270$lcssa;$divider$0$ph$ph212 = $$lcssa286;$storemerge$lcssa = $storemerge$us130;
        break L67;
       }
       $104 = HEAP8[$storemerge$us130>>0]|0;
       $105 = $104&255;
       switch ($104<<24>>24) {
       case 0: case 12: case 9: case 10: case 13: case 32:  {
        $$08$ph$lcssa = $$08$ph$us217$lcssa264$lcssa;$decimal$0$ph$lcssa = $$lcssa270$lcssa;$divider$0$ph$ph212 = $$lcssa286;$storemerge$lcssa = $storemerge$us130;
        break L67;
        break;
       }
       default: {
       }
       }
       $106 = ($104<<24>>24)<(0);
       if ($106) {
        $$08$ph$lcssa = $$08$ph$us217$lcssa264$lcssa;$decimal$0$ph$lcssa = $$lcssa270$lcssa;$divider$0$ph$ph212 = $$lcssa286;$storemerge$lcssa = $storemerge$us130;
        break L67;
       }
       $107 = $105 & 127;
       $108 = (($107) + -48)|0;
       $109 = ($108>>>0)>(9);
       if ($109) {
        $$08$ph$lcssa = $$08$ph$us217$lcssa264$lcssa;$decimal$0$ph$lcssa = $$lcssa270$lcssa;$divider$0$ph$ph212 = $$lcssa286;$storemerge$lcssa = $storemerge$us130;
        break L67;
       } else {
        $$pn$us129 = $storemerge$us130;
       }
      }
     } else {
      $$pn = $storemerge$us$us$lcssa268$lcssa;
      while(1) {
       $storemerge = ((($$pn)) + 1|0);
       $111 = ($storemerge>>>0)<($limit>>>0);
       if (!($111)) {
        $$08$ph$lcssa = $$08$ph$us217$lcssa264$lcssa;$decimal$0$ph$lcssa = $$lcssa270$lcssa;$divider$0$ph$ph212 = $$lcssa286;$storemerge$lcssa = $storemerge;
        break L67;
       }
       $112 = HEAP8[$storemerge>>0]|0;
       $113 = $112&255;
       switch ($112<<24>>24) {
       case 0: case 12: case 9: case 10: case 13: case 32:  {
        $$08$ph$lcssa = $$08$ph$us217$lcssa264$lcssa;$decimal$0$ph$lcssa = $$lcssa270$lcssa;$divider$0$ph$ph212 = $$lcssa286;$storemerge$lcssa = $storemerge;
        break L67;
        break;
       }
       default: {
       }
       }
       $114 = ($112<<24>>24)<(0);
       if ($114) {
        $$08$ph$lcssa = $$08$ph$us217$lcssa264$lcssa;$decimal$0$ph$lcssa = $$lcssa270$lcssa;$divider$0$ph$ph212 = $$lcssa286;$storemerge$lcssa = $storemerge;
        break L67;
       }
       $115 = $113 & 127;
       $116 = (($115) + -48)|0;
       $117 = ($116>>>0)>(9);
       if ($117) {
        $$08$ph$lcssa = $$08$ph$us217$lcssa264$lcssa;$decimal$0$ph$lcssa = $$lcssa270$lcssa;$divider$0$ph$ph212 = $$lcssa286;$storemerge$lcssa = $storemerge;
        break L67;
       } else {
        $$pn = $storemerge;
       }
      }
     }
    }
   } while(0);
   $120 = $storemerge$lcssa;
   $$3 = $$08$ph$lcssa;$decimal$3 = $decimal$0$ph$lcssa;$divider$3 = $divider$0$ph$ph212;$p$sroa$0$3 = $120;
  } else {
   $$3 = $power_ten;$decimal$3 = 0;$divider$3 = 1;$p$sroa$0$3 = $p$sroa$0$2;
  }
 } else {
  $$3 = $power_ten;$decimal$3 = 0;$divider$3 = 1;$p$sroa$0$3 = $p$sroa$0$2;
 }
 $121 = $p$sroa$0$3;
 $122 = ((($121)) + 1|0);
 $123 = ($122>>>0)<($limit>>>0);
 L88: do {
  if ($123) {
   $124 = HEAP8[$121>>0]|0;
   switch ($124<<24>>24) {
   case 69: case 101:  {
    break;
   }
   default: {
    $$6 = $$3;$have_overflow$3 = $have_overflow$0;$have_underflow$2 = 0;$p$sroa$0$5 = $p$sroa$0$3;
    break L88;
   }
   }
   $125 = $122;
   $126 = HEAP8[$122>>0]|0;
   switch ($126<<24>>24) {
   case 43: case 45:  {
    $127 = ($126<<24>>24)==(45);
    $128 = $127&1;
    $129 = ((($121)) + 2|0);
    $130 = ($129|0)==($limit|0);
    if ($130) {
     $$0$i36 = 0;$p$sroa$0$4 = $125;
    } else {
     $p$0$i60 = $129;$sign$0$i59 = $128;
     label = 66;
    }
    break;
   }
   default: {
    $p$0$i60 = $122;$sign$0$i59 = 0;
    label = 66;
   }
   }
   L93: do {
    if ((label|0) == 66) {
     $131 = ($p$0$i60>>>0)<($limit>>>0);
     L95: do {
      if ($131) {
       $have_overflow$07$i64 = 0;$num$08$i63 = 0;$p$19$i62 = $p$0$i60;
       while(1) {
        $132 = HEAP8[$p$19$i62>>0]|0;
        $133 = $132&255;
        switch ($132<<24>>24) {
        case 0: case 12: case 9: case 10: case 13: case 32:  {
         $have_overflow$0$lcssa$i71 = $have_overflow$07$i64;$num$0$lcssa$i70 = $num$08$i63;$p$1$lcssa$i69 = $p$19$i62;
         break L95;
         break;
        }
        default: {
        }
        }
        $134 = ($132<<24>>24)<(0);
        if ($134) {
         $have_overflow$0$lcssa$i71 = $have_overflow$07$i64;$num$0$lcssa$i70 = $num$08$i63;$p$1$lcssa$i69 = $p$19$i62;
         break L95;
        }
        $135 = $133 & 127;
        $136 = (532577 + ($135)|0);
        $137 = HEAP8[$136>>0]|0;
        $138 = $137 << 24 >> 24;
        $139 = (($135) + -48)|0;
        $140 = ($139>>>0)<(10);
        if (!($140)) {
         $have_overflow$0$lcssa$i71 = $have_overflow$07$i64;$num$0$lcssa$i70 = $num$08$i63;$p$1$lcssa$i69 = $p$19$i62;
         break L95;
        }
        $141 = ($num$08$i63|0)>(214748364);
        if ($141) {
         $have_overflow$2$ph$i67 = 1;$num$2$ph$i68 = $num$08$i63;
        } else {
         $142 = ($num$08$i63|0)==(214748364);
         $143 = ($137<<24>>24)>(7);
         $or$cond44$i66 = $142 & $143;
         if ($or$cond44$i66) {
          $have_overflow$2$ph$i67 = 1;$num$2$ph$i68 = 214748364;
         } else {
          $144 = ($num$08$i63*10)|0;
          $145 = (($138) + ($144))|0;
          $have_overflow$2$ph$i67 = $have_overflow$07$i64;$num$2$ph$i68 = $145;
         }
        }
        $146 = ((($p$19$i62)) + 1|0);
        $147 = ($146>>>0)<($limit>>>0);
        if ($147) {
         $have_overflow$07$i64 = $have_overflow$2$ph$i67;$num$08$i63 = $num$2$ph$i68;$p$19$i62 = $146;
        } else {
         $have_overflow$0$lcssa$i71 = $have_overflow$2$ph$i67;$num$0$lcssa$i70 = $num$2$ph$i68;$p$1$lcssa$i69 = $146;
         break L95;
        }
       }
      } else {
       $have_overflow$0$lcssa$i71 = 0;$num$0$lcssa$i70 = 0;$p$1$lcssa$i69 = $p$0$i60;
      }
     } while(0);
     $148 = ($have_overflow$0$lcssa$i71<<24>>24)==(0);
     $num$3$$i72 = $148 ? $num$0$lcssa$i70 : 2147483647;
     $149 = ($sign$0$i59<<24>>24)==(0);
     $150 = (0 - ($num$3$$i72))|0;
     $num$5$i73 = $149 ? $num$3$$i72 : $150;
     $151 = ($p$1$lcssa$i69|0)==($122|0);
     $152 = $p$1$lcssa$i69;
     if ($151) {
      $$0$i36 = 0;$p$sroa$0$4 = $125;
     } else {
      $153 = ($p$1$lcssa$i69>>>0)<($limit>>>0);
      if ($153) {
       $154 = HEAP8[$p$1$lcssa$i69>>0]|0;
       $155 = ($154<<24>>24)==(35);
       if ($155) {
        $156 = ((($p$1$lcssa$i69)) + 1|0);
        $157 = ($156>>>0)>=($limit>>>0);
        $base$off$i38 = (($num$5$i73) + -2)|0;
        $158 = ($base$off$i38>>>0)>(34);
        $or$cond4$i39 = $157 | $158;
        if ($or$cond4$i39) {
         $$0$i36 = 0;$p$sroa$0$4 = $125;
         break;
        }
        $159 = HEAP8[$156>>0]|0;
        switch ($159<<24>>24) {
        case 43: case 45:  {
         $160 = ($159<<24>>24)==(45);
         $161 = $160&1;
         $162 = ((($p$1$lcssa$i69)) + 2|0);
         $163 = ($162|0)==($limit|0);
         if ($163) {
          $$0$i36 = 0;$p$sroa$0$4 = $125;
          break L93;
         } else {
          $p$0$i41 = $162;$sign$0$i40 = $161;
         }
         break;
        }
        default: {
         $p$0$i41 = $156;$sign$0$i40 = 0;
        }
        }
        $164 = (2147483647 / ($num$5$i73|0))&-1;
        $165 = ($p$0$i41>>>0)<($limit>>>0);
        L114: do {
         if ($165) {
          $166 = (2147483647 % ($num$5$i73|0))&-1;
          $sext$i42 = $166 << 24;
          $167 = $sext$i42 >> 24;
          $have_overflow$07$i46 = 0;$num$08$i45 = 0;$p$19$i44 = $p$0$i41;
          while(1) {
           $168 = HEAP8[$p$19$i44>>0]|0;
           $169 = $168&255;
           switch ($168<<24>>24) {
           case 0: case 12: case 9: case 10: case 13: case 32:  {
            $have_overflow$0$lcssa$i53 = $have_overflow$07$i46;$num$0$lcssa$i52 = $num$08$i45;$p$1$lcssa$i51 = $p$19$i44;
            break L114;
            break;
           }
           default: {
           }
           }
           $170 = ($168<<24>>24)<(0);
           if ($170) {
            $have_overflow$0$lcssa$i53 = $have_overflow$07$i46;$num$0$lcssa$i52 = $num$08$i45;$p$1$lcssa$i51 = $p$19$i44;
            break L114;
           }
           $171 = $169 & 127;
           $172 = (532577 + ($171)|0);
           $173 = HEAP8[$172>>0]|0;
           $174 = $173 << 24 >> 24;
           $175 = ($173<<24>>24)>(-1);
           $176 = ($174|0)<($num$5$i73|0);
           $or$cond$i47 = $175 & $176;
           if (!($or$cond$i47)) {
            $have_overflow$0$lcssa$i53 = $have_overflow$07$i46;$num$0$lcssa$i52 = $num$08$i45;$p$1$lcssa$i51 = $p$19$i44;
            break L114;
           }
           $177 = ($num$08$i45|0)>($164|0);
           if ($177) {
            $have_overflow$2$ph$i49 = 1;$num$2$ph$i50 = $num$08$i45;
           } else {
            $178 = ($num$08$i45|0)==($164|0);
            $179 = ($174|0)>($167|0);
            $or$cond44$i48 = $178 & $179;
            if ($or$cond44$i48) {
             $have_overflow$2$ph$i49 = 1;$num$2$ph$i50 = $164;
            } else {
             $180 = Math_imul($num$08$i45, $num$5$i73)|0;
             $181 = (($174) + ($180))|0;
             $have_overflow$2$ph$i49 = $have_overflow$07$i46;$num$2$ph$i50 = $181;
            }
           }
           $182 = ((($p$19$i44)) + 1|0);
           $183 = ($182>>>0)<($limit>>>0);
           if ($183) {
            $have_overflow$07$i46 = $have_overflow$2$ph$i49;$num$08$i45 = $num$2$ph$i50;$p$19$i44 = $182;
           } else {
            $have_overflow$0$lcssa$i53 = $have_overflow$2$ph$i49;$num$0$lcssa$i52 = $num$2$ph$i50;$p$1$lcssa$i51 = $182;
            break L114;
           }
          }
         } else {
          $have_overflow$0$lcssa$i53 = 0;$num$0$lcssa$i52 = 0;$p$1$lcssa$i51 = $p$0$i41;
         }
        } while(0);
        $184 = ($have_overflow$0$lcssa$i53<<24>>24)==(0);
        $num$3$$i54 = $184 ? $num$0$lcssa$i52 : 2147483647;
        $185 = ($sign$0$i40<<24>>24)==(0);
        $186 = (0 - ($num$3$$i54))|0;
        $num$5$i55 = $185 ? $num$3$$i54 : $186;
        $187 = ($p$1$lcssa$i51|0)==($156|0);
        $188 = $p$1$lcssa$i51;
        if ($187) {
         $$0$i36 = 0;$p$sroa$0$4 = $125;
         break;
        } else {
         $222 = $188;$num$0$i35 = $num$5$i55;
        }
       } else {
        $222 = $152;$num$0$i35 = $num$5$i73;
       }
      } else {
       $222 = $152;$num$0$i35 = $num$5$i73;
      }
      $$0$i36 = $num$0$i35;$p$sroa$0$4 = $222;
     }
    }
   } while(0);
   $189 = $p$sroa$0$4;
   $190 = ($122|0)==($189|0);
   if ($190) {
    $$1 = 0;
    return ($$1|0);
   }
   $191 = ($$0$i36|0)>(1000);
   if ($191) {
    $$6 = $$3;$have_overflow$3 = 1;$have_underflow$2 = 0;$p$sroa$0$5 = $p$sroa$0$4;
   } else {
    $192 = ($$0$i36|0)<(-1000);
    $$14 = $192&1;
    $193 = $192 ? 0 : $$0$i36;
    $$3$ = (($193) + ($$3))|0;
    $$6 = $$3$;$have_overflow$3 = $have_overflow$0;$have_underflow$2 = $$14;$p$sroa$0$5 = $p$sroa$0$4;
   }
  } else {
   $$6 = $$3;$have_overflow$3 = $have_overflow$0;$have_underflow$2 = 0;$p$sroa$0$5 = $p$sroa$0$3;
  }
 } while(0);
 HEAP32[$cursor>>2] = $p$sroa$0$5;
 $194 = $decimal$3 | $integral$0;
 $195 = ($194|0)==(0);
 if ($195) {
  $$1 = 0;
  return ($$1|0);
 }
 $196 = ($have_overflow$3<<24>>24)==(0);
 L136: do {
  if ($196) {
   $197 = ($have_underflow$2<<24>>24)==(0);
   if (!($197)) {
    $$1 = 0;
    return ($$1|0);
   }
   $198 = ($$6|0)>(0);
   if ($198) {
    $$7106 = $$6;$decimal$4104 = $decimal$3;$divider$4105 = $divider$3;$integral$1103 = $integral$0;
    while(1) {
     $200 = ($integral$1103|0)>(214748363);
     if ($200) {
      $integral$3 = 2147483647;
      break L136;
     }
     $201 = ($integral$1103*10)|0;
     $202 = ($decimal$4104|0)>(214748363);
     if ($202) {
      $203 = ($divider$4105|0)==(1);
      if ($203) {
       $integral$3 = 2147483647;
       break L136;
      }
      $204 = (($divider$4105|0) / 10)&-1;
      $decimal$5 = $decimal$4104;$divider$5 = $204;
     } else {
      $205 = ($decimal$4104*10)|0;
      $decimal$5 = $205;$divider$5 = $divider$4105;
     }
     $206 = (($$7106) + -1)|0;
     $207 = ($$7106|0)>(1);
     if ($207) {
      $$7106 = $206;$decimal$4104 = $decimal$5;$divider$4105 = $divider$5;$integral$1103 = $201;
     } else {
      $$7$lcssa = $206;$decimal$4$lcssa = $decimal$5;$divider$4$lcssa = $divider$5;$integral$1$lcssa = $201;
      break;
     }
    }
   } else {
    $$7$lcssa = $$6;$decimal$4$lcssa = $decimal$3;$divider$4$lcssa = $divider$3;$integral$1$lcssa = $integral$0;
   }
   $199 = ($$7$lcssa|0)<(0);
   L151: do {
    if ($199) {
     $$8100 = $$7$lcssa;$decimal$698 = $decimal$4$lcssa;$divider$699 = $divider$4$lcssa;$integral$297 = $integral$1$lcssa;
     while(1) {
      $208 = ($divider$699|0)<(214748364);
      $209 = (($decimal$698|0) / 10)&-1;
      $decimal$7 = $208 ? $decimal$698 : $209;
      $integral$2$off = (($integral$297) + 9)|0;
      $210 = ($integral$2$off>>>0)>(18);
      $211 = ($decimal$7|0)!=(0);
      $or$cond7 = $210 | $211;
      if (!($or$cond7)) {
       $$1 = 0;
       break;
      }
      $212 = ($divider$699*10)|0;
      $divider$7 = $208 ? $212 : $divider$699;
      $213 = (($integral$297|0) / 10)&-1;
      $214 = (($$8100) + 1)|0;
      $215 = ($$8100|0)<(-1);
      if ($215) {
       $$8100 = $214;$decimal$698 = $decimal$7;$divider$699 = $divider$7;$integral$297 = $213;
      } else {
       $decimal$6$lcssa = $decimal$7;$divider$6$lcssa = $divider$7;$integral$2$lcssa = $213;
       break L151;
      }
     }
     return ($$1|0);
    } else {
     $decimal$6$lcssa = $decimal$4$lcssa;$divider$6$lcssa = $divider$4$lcssa;$integral$2$lcssa = $integral$1$lcssa;
    }
   } while(0);
   $216 = ($decimal$6$lcssa|0)==(0);
   if ($216) {
    $integral$3 = $integral$2$lcssa;
   } else {
    $217 = (_FT_DivFix($decimal$6$lcssa,$divider$6$lcssa)|0);
    $218 = (($217) + ($integral$2$lcssa))|0;
    $integral$3 = $218;
   }
  } else {
   $integral$3 = 2147483647;
  }
 } while(0);
 $219 = ($sign$0<<24>>24)==(0);
 $220 = (0 - ($integral$3))|0;
 $integral$3$ = $219 ? $integral$3 : $220;
 $$1 = $integral$3$;
 return ($$1|0);
}
function _ps_parser_to_bytes($parser,$bytes,$max_bytes,$pnum_bytes,$delimiters) {
 $parser = $parser|0;
 $bytes = $bytes|0;
 $max_bytes = $max_bytes|0;
 $pnum_bytes = $pnum_bytes|0;
 $delimiters = $delimiters|0;
 var $$$i = 0, $$$i1 = 0, $$010$i15 = 0, $$cast = 0, $$pre = 0, $$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $9 = 0, $cur$0$lcssa$i$i$i = 0, $cur$01$i$i$i = 0, $cur$sroa$0$0$in = 0, $cur$sroa$0$0$lcssa$i$i = 0, $cur$sroa$0$01$i$i = 0, $cur$sroa$0$1$i$i = 0, $cur$sroa$0$2 = 0, $cur$sroa$0$2$in = 0, $error$0 = 0;
 var $pad$0$lcssa$i = 0, $pad$0$lcssa$i8 = 0, $pad$016$i = 0, $pad$016$i2 = 0, $pad$2$ph$i = 0, $pad$2$ph$i7 = 0, $r$0$lcssa$i = 0, $r$0$lcssa$i10 = 0, $r$0$lcssa27$i = 0, $r$0$lcssa27$i12 = 0, $r$014$i = 0, $r$014$i4 = 0, $w$0$lcssa$i = 0, $w$0$lcssa$i9 = 0, $w$015$i = 0, $w$015$i3 = 0, $w$2$ph$i = 0, $w$2$ph$i6 = 0, $w$4$i = 0, $w$4$i13 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($parser)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$parser>>2]|0;
 $3 = $2;
 $4 = ($3>>>0)<($1>>>0);
 L1: do {
  if ($4) {
   $6 = $3;$cur$sroa$0$01$i$i = $2;
   while(1) {
    $5 = HEAP8[$6>>0]|0;
    switch ($5<<24>>24) {
    case 0: case 12: case 9: case 10: case 13: case 32:  {
     $cur$sroa$0$1$i$i = $cur$sroa$0$01$i$i;
     break;
    }
    case 37:  {
     $7 = 37;$cur$01$i$i$i = $6;
     L5: while(1) {
      switch ($7<<24>>24) {
      case 10: case 13:  {
       $cur$0$lcssa$i$i$i = $cur$01$i$i$i;
       break L5;
       break;
      }
      default: {
      }
      }
      $8 = ((($cur$01$i$i$i)) + 1|0);
      $9 = ($8>>>0)<($1>>>0);
      if (!($9)) {
       $cur$0$lcssa$i$i$i = $8;
       break;
      }
      $$pre$i$i = HEAP8[$8>>0]|0;
      $7 = $$pre$i$i;$cur$01$i$i$i = $8;
     }
     $10 = $cur$0$lcssa$i$i$i;
     $cur$sroa$0$1$i$i = $10;
     break;
    }
    default: {
     $cur$sroa$0$0$lcssa$i$i = $cur$sroa$0$01$i$i;
     break L1;
    }
    }
    $11 = $cur$sroa$0$1$i$i;
    $12 = ((($11)) + 1|0);
    $13 = $12;
    $14 = ($12>>>0)<($1>>>0);
    if ($14) {
     $6 = $12;$cur$sroa$0$01$i$i = $13;
    } else {
     $cur$sroa$0$0$lcssa$i$i = $13;
     break L1;
    }
   }
  } else {
   $cur$sroa$0$0$lcssa$i$i = $2;
  }
 } while(0);
 HEAP32[$parser>>2] = $cur$sroa$0$0$lcssa$i$i;
 $$cast = $cur$sroa$0$0$lcssa$i$i;
 $15 = ($$cast>>>0)<($1>>>0);
 if (!($15)) {
  $error$0 = 0;
  return ($error$0|0);
 }
 $16 = ($delimiters<<24>>24)==(0);
 if ($16) {
  $56 = $max_bytes << 1;
  $57 = $1;
  $58 = (($57) - ($cur$sroa$0$0$lcssa$i$i))|0;
  $59 = ($56>>>0)>($58>>>0);
  $$$i = $59 ? $58 : $56;
  $60 = ($$$i|0)==(0);
  if ($60) {
   $r$0$lcssa27$i = 0;$w$4$i = 0;
  } else {
   $pad$016$i = 1;$r$014$i = 0;$w$015$i = 0;
   L18: while(1) {
    $61 = (($$cast) + ($r$014$i)|0);
    $62 = HEAP8[$61>>0]|0;
    $63 = $62&255;
    switch ($62<<24>>24) {
    case 0: case 9: case 10: case 12: case 13: case 32:  {
     $pad$2$ph$i = $pad$016$i;$w$2$ph$i = $w$015$i;
     break;
    }
    default: {
     $64 = ($62<<24>>24)<(0);
     if ($64) {
      $pad$0$lcssa$i = $pad$016$i;$r$0$lcssa$i = $r$014$i;$w$0$lcssa$i = $w$015$i;
      break L18;
     }
     $65 = $63 & 127;
     $66 = (532577 + ($65)|0);
     $67 = HEAP8[$66>>0]|0;
     $68 = ($67&255)>(15);
     if ($68) {
      $pad$0$lcssa$i = $pad$016$i;$r$0$lcssa$i = $r$014$i;$w$0$lcssa$i = $w$015$i;
      break L18;
     }
     $69 = $67 << 24 >> 24;
     $70 = $pad$016$i << 4;
     $71 = $69 | $70;
     $72 = $71 & 256;
     $73 = ($72|0)==(0);
     if ($73) {
      $pad$2$ph$i = $71;$w$2$ph$i = $w$015$i;
     } else {
      $74 = $71&255;
      $75 = (($w$015$i) + 1)|0;
      $76 = (($bytes) + ($w$015$i)|0);
      HEAP8[$76>>0] = $74;
      $pad$2$ph$i = 1;$w$2$ph$i = $75;
     }
    }
    }
    $77 = (($r$014$i) + 1)|0;
    $78 = ($77>>>0)<($$$i>>>0);
    if ($78) {
     $pad$016$i = $pad$2$ph$i;$r$014$i = $77;$w$015$i = $w$2$ph$i;
    } else {
     $pad$0$lcssa$i = $pad$2$ph$i;$r$0$lcssa$i = $77;$w$0$lcssa$i = $w$2$ph$i;
     break;
    }
   }
   $79 = ($pad$0$lcssa$i|0)==(1);
   if ($79) {
    $r$0$lcssa27$i = $r$0$lcssa$i;$w$4$i = $w$0$lcssa$i;
   } else {
    $80 = $pad$0$lcssa$i << 4;
    $81 = $80&255;
    $82 = (($w$0$lcssa$i) + 1)|0;
    $83 = (($bytes) + ($w$0$lcssa$i)|0);
    HEAP8[$83>>0] = $81;
    $r$0$lcssa27$i = $r$0$lcssa$i;$w$4$i = $82;
   }
  }
  $84 = (($$cast) + ($r$0$lcssa27$i)|0);
  HEAP32[$pnum_bytes>>2] = $w$4$i;
  $cur$sroa$0$2$in = $84;
 } else {
  $17 = HEAP8[$$cast>>0]|0;
  $18 = ($17<<24>>24)==(60);
  if (!($18)) {
   $error$0 = 3;
   return ($error$0|0);
  }
  $19 = ((($$cast)) + 1|0);
  $20 = $max_bytes << 1;
  $21 = ($19>>>0)<($1>>>0);
  if ($21) {
   $22 = $1;
   $23 = $19;
   $24 = (($22) - ($23))|0;
   $25 = ($20>>>0)>($24>>>0);
   $$$i1 = $25 ? $24 : $20;
   $26 = ($$$i1|0)==(0);
   if ($26) {
    $r$0$lcssa27$i12 = 0;$w$4$i13 = 0;
   } else {
    $pad$016$i2 = 1;$r$014$i4 = 0;$w$015$i3 = 0;
    L36: while(1) {
     $27 = (($19) + ($r$014$i4)|0);
     $28 = HEAP8[$27>>0]|0;
     $29 = $28&255;
     switch ($28<<24>>24) {
     case 0: case 9: case 10: case 12: case 13: case 32:  {
      $pad$2$ph$i7 = $pad$016$i2;$w$2$ph$i6 = $w$015$i3;
      break;
     }
     default: {
      $30 = ($28<<24>>24)<(0);
      if ($30) {
       $pad$0$lcssa$i8 = $pad$016$i2;$r$0$lcssa$i10 = $r$014$i4;$w$0$lcssa$i9 = $w$015$i3;
       break L36;
      }
      $31 = $29 & 127;
      $32 = (532577 + ($31)|0);
      $33 = HEAP8[$32>>0]|0;
      $34 = ($33&255)>(15);
      if ($34) {
       $pad$0$lcssa$i8 = $pad$016$i2;$r$0$lcssa$i10 = $r$014$i4;$w$0$lcssa$i9 = $w$015$i3;
       break L36;
      }
      $35 = $33 << 24 >> 24;
      $36 = $pad$016$i2 << 4;
      $37 = $35 | $36;
      $38 = $37 & 256;
      $39 = ($38|0)==(0);
      if ($39) {
       $pad$2$ph$i7 = $37;$w$2$ph$i6 = $w$015$i3;
      } else {
       $40 = $37&255;
       $41 = (($w$015$i3) + 1)|0;
       $42 = (($bytes) + ($w$015$i3)|0);
       HEAP8[$42>>0] = $40;
       $pad$2$ph$i7 = 1;$w$2$ph$i6 = $41;
      }
     }
     }
     $43 = (($r$014$i4) + 1)|0;
     $44 = ($43>>>0)<($$$i1>>>0);
     if ($44) {
      $pad$016$i2 = $pad$2$ph$i7;$r$014$i4 = $43;$w$015$i3 = $w$2$ph$i6;
     } else {
      $pad$0$lcssa$i8 = $pad$2$ph$i7;$r$0$lcssa$i10 = $43;$w$0$lcssa$i9 = $w$2$ph$i6;
      break;
     }
    }
    $45 = ($pad$0$lcssa$i8|0)==(1);
    if ($45) {
     $r$0$lcssa27$i12 = $r$0$lcssa$i10;$w$4$i13 = $w$0$lcssa$i9;
    } else {
     $46 = $pad$0$lcssa$i8 << 4;
     $47 = $46&255;
     $48 = (($w$0$lcssa$i9) + 1)|0;
     $49 = (($bytes) + ($w$0$lcssa$i9)|0);
     HEAP8[$49>>0] = $47;
     $r$0$lcssa27$i12 = $r$0$lcssa$i10;$w$4$i13 = $48;
    }
   }
   $50 = (($19) + ($r$0$lcssa27$i12)|0);
   $$pre = HEAP32[$0>>2]|0;
   $$010$i15 = $w$4$i13;$52 = $$pre;$cur$sroa$0$0$in = $50;
  } else {
   $$010$i15 = 0;$52 = $1;$cur$sroa$0$0$in = $19;
  }
  HEAP32[$pnum_bytes>>2] = $$010$i15;
  $51 = ($cur$sroa$0$0$in>>>0)<($52>>>0);
  if ($51) {
   $53 = HEAP8[$cur$sroa$0$0$in>>0]|0;
   $54 = ($53<<24>>24)==(62);
   if (!($54)) {
    $error$0 = 3;
    return ($error$0|0);
   }
  }
  $55 = ((($cur$sroa$0$0$in)) + 1|0);
  $cur$sroa$0$2$in = $55;
 }
 $cur$sroa$0$2 = $cur$sroa$0$2$in;
 HEAP32[$parser>>2] = $cur$sroa$0$2;
 $error$0 = 0;
 return ($error$0|0);
}
function _ps_parser_to_coord_array($parser,$max_coords,$coords) {
 $parser = $parser|0;
 $max_coords = $max_coords|0;
 $coords = $coords|0;
 var $$ = 0, $$$i = 0, $$7 = 0, $$cast = 0, $$cast$lcssa6$i = 0, $$cast$us$i = 0, $$cast$us$i$us = 0, $$cast$us19$i = 0, $$cast$us19$i$us = 0, $$pre$i$i = 0, $$pre$i$us$i = 0, $$pre$i$us$i$us = 0, $$pre$i$us15$i = 0, $$pre$i$us15$i$us = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $count$0$lcssa4$i = 0, $count$0$us$i$us26 = 0, $count$0$us7$i$us32 = 0, $count$2$i = 0, $cur$0$lcssa$i$i$i = 0, $cur$0$lcssa$i$i$us$i = 0, $cur$0$lcssa$i$i$us$i$us = 0, $cur$0$lcssa$i$i$us13$i = 0, $cur$0$lcssa$i$i$us13$i$us = 0, $cur$01$i$i$i = 0;
 var $cur$01$i$i$us$i = 0, $cur$01$i$i$us$i$us = 0, $cur$01$i$i$us11$i = 0, $cur$01$i$i$us11$i$us = 0, $cur$i = 0, $cur$sroa$0$0$lcssa$i$i = 0, $cur$sroa$0$0$lcssa$i$us$i = 0, $cur$sroa$0$0$lcssa$i$us$i$us = 0, $cur$sroa$0$0$lcssa$i$us18$i = 0, $cur$sroa$0$0$lcssa$i$us18$i$us = 0, $cur$sroa$0$01$i$i = 0, $cur$sroa$0$01$i$us$i = 0, $cur$sroa$0$01$i$us$i$us = 0, $cur$sroa$0$01$i$us9$i = 0, $cur$sroa$0$01$i$us9$i$us = 0, $cur$sroa$0$1$i$i = 0, $cur$sroa$0$1$i$us$i = 0, $cur$sroa$0$1$i$us$i$us = 0, $cur$sroa$0$1$i$us16$i = 0, $cur$sroa$0$1$i$us16$i$us = 0;
 var $ender$0$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $cur$i = sp;
 $0 = ((($parser)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$parser>>2]|0;
 $3 = $2;
 $4 = ($3>>>0)<($1>>>0);
 L1: do {
  if ($4) {
   $6 = $3;$cur$sroa$0$01$i$i = $2;
   while(1) {
    $5 = HEAP8[$6>>0]|0;
    switch ($5<<24>>24) {
    case 0: case 12: case 9: case 10: case 13: case 32:  {
     $cur$sroa$0$1$i$i = $cur$sroa$0$01$i$i;
     break;
    }
    case 37:  {
     $7 = 37;$cur$01$i$i$i = $6;
     L5: while(1) {
      switch ($7<<24>>24) {
      case 10: case 13:  {
       $cur$0$lcssa$i$i$i = $cur$01$i$i$i;
       break L5;
       break;
      }
      default: {
      }
      }
      $8 = ((($cur$01$i$i$i)) + 1|0);
      $9 = ($8>>>0)<($1>>>0);
      if (!($9)) {
       $cur$0$lcssa$i$i$i = $8;
       break;
      }
      $$pre$i$i = HEAP8[$8>>0]|0;
      $7 = $$pre$i$i;$cur$01$i$i$i = $8;
     }
     $10 = $cur$0$lcssa$i$i$i;
     $cur$sroa$0$1$i$i = $10;
     break;
    }
    default: {
     $cur$sroa$0$0$lcssa$i$i = $cur$sroa$0$01$i$i;
     break L1;
    }
    }
    $11 = $cur$sroa$0$1$i$i;
    $12 = ((($11)) + 1|0);
    $13 = $12;
    $14 = ($12>>>0)<($1>>>0);
    if ($14) {
     $6 = $12;$cur$sroa$0$01$i$i = $13;
    } else {
     $cur$sroa$0$0$lcssa$i$i = $13;
     break L1;
    }
   }
  } else {
   $cur$sroa$0$0$lcssa$i$i = $2;
  }
 } while(0);
 HEAP32[$parser>>2] = $cur$sroa$0$0$lcssa$i$i;
 $$cast = $cur$sroa$0$0$lcssa$i$i;
 HEAP32[$cur$i>>2] = $$cast;
 $15 = ($$cast>>>0)<($1>>>0);
 L12: do {
  if ($15) {
   $16 = HEAP8[$$cast>>0]|0;
   $17 = ($16<<24>>24)==(91);
   $18 = ($16<<24>>24)==(123);
   $$$i = $18 ? 125 : 0;
   $ender$0$i = $17 ? 93 : $$$i;
   $19 = ($ender$0$i<<24>>24)!=(0);
   if ($19) {
    $20 = ((($$cast)) + 1|0);
    HEAP32[$cur$i>>2] = $20;
    $21 = $20;
    $24 = $20;$73 = $21;
   } else {
    $24 = $$cast;$73 = $cur$sroa$0$0$lcssa$i$i;
   }
   $22 = ($coords|0)==(0|0);
   $23 = ($24>>>0)<($1>>>0);
   L17: do {
    if ($22) {
     if (!($19)) {
      if (!($23)) {
       $115 = $73;$count$2$i = 0;
       break L12;
      }
      $72 = $73;
      $74 = ($72>>>0)<($1>>>0);
      L22: do {
       if ($74) {
        $76 = $72;$cur$sroa$0$01$i$us$i = $73;
        while(1) {
         $75 = HEAP8[$76>>0]|0;
         switch ($75<<24>>24) {
         case 0: case 12: case 9: case 10: case 13: case 32:  {
          $cur$sroa$0$1$i$us$i = $cur$sroa$0$01$i$us$i;
          break;
         }
         case 37:  {
          $77 = 37;$cur$01$i$i$us$i = $76;
          L26: while(1) {
           switch ($77<<24>>24) {
           case 10: case 13:  {
            $cur$0$lcssa$i$i$us$i = $cur$01$i$i$us$i;
            break L26;
            break;
           }
           default: {
           }
           }
           $78 = ((($cur$01$i$i$us$i)) + 1|0);
           $79 = ($78>>>0)<($1>>>0);
           if (!($79)) {
            $cur$0$lcssa$i$i$us$i = $78;
            break;
           }
           $$pre$i$us$i = HEAP8[$78>>0]|0;
           $77 = $$pre$i$us$i;$cur$01$i$i$us$i = $78;
          }
          $80 = $cur$0$lcssa$i$i$us$i;
          $cur$sroa$0$1$i$us$i = $80;
          break;
         }
         default: {
          $cur$sroa$0$0$lcssa$i$us$i = $cur$sroa$0$01$i$us$i;
          break L22;
         }
         }
         $81 = $cur$sroa$0$1$i$us$i;
         $82 = ((($81)) + 1|0);
         $83 = $82;
         $84 = ($82>>>0)<($1>>>0);
         if ($84) {
          $76 = $82;$cur$sroa$0$01$i$us$i = $83;
         } else {
          $cur$sroa$0$0$lcssa$i$us$i = $83;
          break L22;
         }
        }
       } else {
        $cur$sroa$0$0$lcssa$i$us$i = $73;
       }
      } while(0);
      HEAP32[$cur$i>>2] = $cur$sroa$0$0$lcssa$i$us$i;
      $$cast$us$i = $cur$sroa$0$0$lcssa$i$us$i;
      $85 = ($$cast$us$i>>>0)<($1>>>0);
      if (!($85)) {
       $115 = $cur$sroa$0$0$lcssa$i$us$i;$count$2$i = 0;
       break L12;
      }
      $86 = HEAP8[$$cast$us$i>>0]|0;
      $87 = ($86<<24>>24)==(0);
      if ($87) {
       $$cast$lcssa6$i = $$cast$us$i;$count$0$lcssa4$i = 0;
       break;
      }
      (_PS_Conv_ToFixed($cur$i,$1,0)|0);
      $88 = HEAP32[$cur$i>>2]|0;
      $89 = ($$cast$us$i|0)==($88|0);
      $90 = $88;
      $$ = $89 ? -1 : 1;
      $115 = $90;$count$2$i = $$;
      break L12;
     }
     if ($23) {
      $52 = $73;$count$0$us$i$us26 = 0;
     } else {
      $115 = $73;$count$2$i = 0;
      break L12;
     }
     while(1) {
      $51 = $52;
      $53 = ($51>>>0)<($1>>>0);
      L38: do {
       if ($53) {
        $55 = $51;$cur$sroa$0$01$i$us$i$us = $52;
        while(1) {
         $54 = HEAP8[$55>>0]|0;
         switch ($54<<24>>24) {
         case 0: case 12: case 9: case 10: case 13: case 32:  {
          $cur$sroa$0$1$i$us$i$us = $cur$sroa$0$01$i$us$i$us;
          break;
         }
         case 37:  {
          $56 = 37;$cur$01$i$i$us$i$us = $55;
          L42: while(1) {
           switch ($56<<24>>24) {
           case 10: case 13:  {
            $cur$0$lcssa$i$i$us$i$us = $cur$01$i$i$us$i$us;
            break L42;
            break;
           }
           default: {
           }
           }
           $57 = ((($cur$01$i$i$us$i$us)) + 1|0);
           $58 = ($57>>>0)<($1>>>0);
           if (!($58)) {
            $cur$0$lcssa$i$i$us$i$us = $57;
            break;
           }
           $$pre$i$us$i$us = HEAP8[$57>>0]|0;
           $56 = $$pre$i$us$i$us;$cur$01$i$i$us$i$us = $57;
          }
          $59 = $cur$0$lcssa$i$i$us$i$us;
          $cur$sroa$0$1$i$us$i$us = $59;
          break;
         }
         default: {
          $cur$sroa$0$0$lcssa$i$us$i$us = $cur$sroa$0$01$i$us$i$us;
          break L38;
         }
         }
         $60 = $cur$sroa$0$1$i$us$i$us;
         $61 = ((($60)) + 1|0);
         $62 = $61;
         $63 = ($61>>>0)<($1>>>0);
         if ($63) {
          $55 = $61;$cur$sroa$0$01$i$us$i$us = $62;
         } else {
          $cur$sroa$0$0$lcssa$i$us$i$us = $62;
          break L38;
         }
        }
       } else {
        $cur$sroa$0$0$lcssa$i$us$i$us = $52;
       }
      } while(0);
      HEAP32[$cur$i>>2] = $cur$sroa$0$0$lcssa$i$us$i$us;
      $$cast$us$i$us = $cur$sroa$0$0$lcssa$i$us$i$us;
      $64 = ($$cast$us$i$us>>>0)<($1>>>0);
      if (!($64)) {
       $115 = $cur$sroa$0$0$lcssa$i$us$i$us;$count$2$i = $count$0$us$i$us26;
       break L12;
      }
      $65 = HEAP8[$$cast$us$i$us>>0]|0;
      $66 = ($65<<24>>24)==($ender$0$i<<24>>24);
      if ($66) {
       $$cast$lcssa6$i = $$cast$us$i$us;$count$0$lcssa4$i = $count$0$us$i$us26;
       break L17;
      }
      (_PS_Conv_ToFixed($cur$i,$1,0)|0);
      $67 = HEAP32[$cur$i>>2]|0;
      $68 = ($$cast$us$i$us|0)==($67|0);
      $69 = $67;
      if ($68) {
       $115 = $69;$count$2$i = -1;
       break L12;
      }
      $70 = (($count$0$us$i$us26) + 1)|0;
      $71 = ($67>>>0)<($1>>>0);
      if ($71) {
       $52 = $69;$count$0$us$i$us26 = $70;
      } else {
       $115 = $69;$count$2$i = $70;
       break L12;
      }
     }
    } else {
     if (!($19)) {
      if (!($23)) {
       $115 = $73;$count$2$i = 0;
       break L12;
      }
      $91 = $73;
      $92 = ($91>>>0)<($1>>>0);
      L56: do {
       if ($92) {
        $94 = $91;$cur$sroa$0$01$i$us9$i = $73;
        while(1) {
         $93 = HEAP8[$94>>0]|0;
         switch ($93<<24>>24) {
         case 0: case 12: case 9: case 10: case 13: case 32:  {
          $cur$sroa$0$1$i$us16$i = $cur$sroa$0$01$i$us9$i;
          break;
         }
         case 37:  {
          $95 = 37;$cur$01$i$i$us11$i = $94;
          L60: while(1) {
           switch ($95<<24>>24) {
           case 10: case 13:  {
            $cur$0$lcssa$i$i$us13$i = $cur$01$i$i$us11$i;
            break L60;
            break;
           }
           default: {
           }
           }
           $96 = ((($cur$01$i$i$us11$i)) + 1|0);
           $97 = ($96>>>0)<($1>>>0);
           if (!($97)) {
            $cur$0$lcssa$i$i$us13$i = $96;
            break;
           }
           $$pre$i$us15$i = HEAP8[$96>>0]|0;
           $95 = $$pre$i$us15$i;$cur$01$i$i$us11$i = $96;
          }
          $98 = $cur$0$lcssa$i$i$us13$i;
          $cur$sroa$0$1$i$us16$i = $98;
          break;
         }
         default: {
          $cur$sroa$0$0$lcssa$i$us18$i = $cur$sroa$0$01$i$us9$i;
          break L56;
         }
         }
         $99 = $cur$sroa$0$1$i$us16$i;
         $100 = ((($99)) + 1|0);
         $101 = $100;
         $102 = ($100>>>0)<($1>>>0);
         if ($102) {
          $94 = $100;$cur$sroa$0$01$i$us9$i = $101;
         } else {
          $cur$sroa$0$0$lcssa$i$us18$i = $101;
          break L56;
         }
        }
       } else {
        $cur$sroa$0$0$lcssa$i$us18$i = $73;
       }
      } while(0);
      HEAP32[$cur$i>>2] = $cur$sroa$0$0$lcssa$i$us18$i;
      $$cast$us19$i = $cur$sroa$0$0$lcssa$i$us18$i;
      $103 = ($$cast$us19$i>>>0)<($1>>>0);
      if (!($103)) {
       $115 = $cur$sroa$0$0$lcssa$i$us18$i;$count$2$i = 0;
       break L12;
      }
      $104 = HEAP8[$$cast$us19$i>>0]|0;
      $105 = ($104<<24>>24)==(0);
      if ($105) {
       $$cast$lcssa6$i = $$cast$us19$i;$count$0$lcssa4$i = 0;
       break;
      }
      $106 = ($max_coords|0)>(0);
      if (!($106)) {
       $115 = $cur$sroa$0$0$lcssa$i$us18$i;$count$2$i = 0;
       break L12;
      }
      $107 = (_PS_Conv_ToFixed($cur$i,$1,0)|0);
      $108 = $107 >>> 16;
      $109 = $108&65535;
      HEAP16[$coords>>1] = $109;
      $110 = HEAP32[$cur$i>>2]|0;
      $111 = ($$cast$us19$i|0)==($110|0);
      $112 = $110;
      $$7 = $111 ? -1 : 1;
      $115 = $112;$count$2$i = $$7;
      break L12;
     }
     if ($23) {
      $26 = $73;$count$0$us7$i$us32 = 0;
     } else {
      $115 = $73;$count$2$i = 0;
      break L12;
     }
     while(1) {
      $25 = $26;
      $27 = ($25>>>0)<($1>>>0);
      L73: do {
       if ($27) {
        $29 = $25;$cur$sroa$0$01$i$us9$i$us = $26;
        while(1) {
         $28 = HEAP8[$29>>0]|0;
         switch ($28<<24>>24) {
         case 0: case 12: case 9: case 10: case 13: case 32:  {
          $cur$sroa$0$1$i$us16$i$us = $cur$sroa$0$01$i$us9$i$us;
          break;
         }
         case 37:  {
          $30 = 37;$cur$01$i$i$us11$i$us = $29;
          L77: while(1) {
           switch ($30<<24>>24) {
           case 10: case 13:  {
            $cur$0$lcssa$i$i$us13$i$us = $cur$01$i$i$us11$i$us;
            break L77;
            break;
           }
           default: {
           }
           }
           $31 = ((($cur$01$i$i$us11$i$us)) + 1|0);
           $32 = ($31>>>0)<($1>>>0);
           if (!($32)) {
            $cur$0$lcssa$i$i$us13$i$us = $31;
            break;
           }
           $$pre$i$us15$i$us = HEAP8[$31>>0]|0;
           $30 = $$pre$i$us15$i$us;$cur$01$i$i$us11$i$us = $31;
          }
          $33 = $cur$0$lcssa$i$i$us13$i$us;
          $cur$sroa$0$1$i$us16$i$us = $33;
          break;
         }
         default: {
          $cur$sroa$0$0$lcssa$i$us18$i$us = $cur$sroa$0$01$i$us9$i$us;
          break L73;
         }
         }
         $34 = $cur$sroa$0$1$i$us16$i$us;
         $35 = ((($34)) + 1|0);
         $36 = $35;
         $37 = ($35>>>0)<($1>>>0);
         if ($37) {
          $29 = $35;$cur$sroa$0$01$i$us9$i$us = $36;
         } else {
          $cur$sroa$0$0$lcssa$i$us18$i$us = $36;
          break L73;
         }
        }
       } else {
        $cur$sroa$0$0$lcssa$i$us18$i$us = $26;
       }
      } while(0);
      HEAP32[$cur$i>>2] = $cur$sroa$0$0$lcssa$i$us18$i$us;
      $$cast$us19$i$us = $cur$sroa$0$0$lcssa$i$us18$i$us;
      $38 = ($$cast$us19$i$us>>>0)<($1>>>0);
      if (!($38)) {
       $115 = $cur$sroa$0$0$lcssa$i$us18$i$us;$count$2$i = $count$0$us7$i$us32;
       break L12;
      }
      $39 = HEAP8[$$cast$us19$i$us>>0]|0;
      $40 = ($39<<24>>24)==($ender$0$i<<24>>24);
      if ($40) {
       $$cast$lcssa6$i = $$cast$us19$i$us;$count$0$lcssa4$i = $count$0$us7$i$us32;
       break L17;
      }
      $41 = ($count$0$us7$i$us32|0)<($max_coords|0);
      if (!($41)) {
       $115 = $cur$sroa$0$0$lcssa$i$us18$i$us;$count$2$i = $count$0$us7$i$us32;
       break L12;
      }
      $42 = (_PS_Conv_ToFixed($cur$i,$1,0)|0);
      $43 = $42 >>> 16;
      $44 = $43&65535;
      $45 = (($coords) + ($count$0$us7$i$us32<<1)|0);
      HEAP16[$45>>1] = $44;
      $46 = HEAP32[$cur$i>>2]|0;
      $47 = ($$cast$us19$i$us|0)==($46|0);
      $48 = $46;
      if ($47) {
       $115 = $48;$count$2$i = -1;
       break L12;
      }
      $49 = (($count$0$us7$i$us32) + 1)|0;
      $50 = ($46>>>0)<($1>>>0);
      if ($50) {
       $26 = $48;$count$0$us7$i$us32 = $49;
      } else {
       $115 = $48;$count$2$i = $49;
       break L12;
      }
     }
    }
   } while(0);
   $113 = ((($$cast$lcssa6$i)) + 1|0);
   HEAP32[$cur$i>>2] = $113;
   $114 = $113;
   $115 = $114;$count$2$i = $count$0$lcssa4$i;
  } else {
   $115 = $cur$sroa$0$0$lcssa$i$i;$count$2$i = 0;
  }
 } while(0);
 HEAP32[$parser>>2] = $115;
 STACKTOP = sp;return ($count$2$i|0);
}
function _ps_parser_to_fixed_array($parser,$max_values,$values,$power_ten) {
 $parser = $parser|0;
 $max_values = $max_values|0;
 $values = $values|0;
 $power_ten = $power_ten|0;
 var $$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cur$0$lcssa$i$i$i = 0, $cur$01$i$i$i = 0, $cur$sroa$0$0$lcssa$i$i = 0;
 var $cur$sroa$0$01$i$i = 0, $cur$sroa$0$1$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($parser)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$parser>>2]|0;
 $3 = $2;
 $4 = ($3>>>0)<($1>>>0);
 L1: do {
  if ($4) {
   $6 = $3;$cur$sroa$0$01$i$i = $2;
   while(1) {
    $5 = HEAP8[$6>>0]|0;
    switch ($5<<24>>24) {
    case 0: case 12: case 9: case 10: case 13: case 32:  {
     $cur$sroa$0$1$i$i = $cur$sroa$0$01$i$i;
     break;
    }
    case 37:  {
     $7 = 37;$cur$01$i$i$i = $6;
     L5: while(1) {
      switch ($7<<24>>24) {
      case 10: case 13:  {
       $cur$0$lcssa$i$i$i = $cur$01$i$i$i;
       break L5;
       break;
      }
      default: {
      }
      }
      $8 = ((($cur$01$i$i$i)) + 1|0);
      $9 = ($8>>>0)<($1>>>0);
      if (!($9)) {
       $cur$0$lcssa$i$i$i = $8;
       break;
      }
      $$pre$i$i = HEAP8[$8>>0]|0;
      $7 = $$pre$i$i;$cur$01$i$i$i = $8;
     }
     $10 = $cur$0$lcssa$i$i$i;
     $cur$sroa$0$1$i$i = $10;
     break;
    }
    default: {
     $cur$sroa$0$0$lcssa$i$i = $cur$sroa$0$01$i$i;
     break L1;
    }
    }
    $11 = $cur$sroa$0$1$i$i;
    $12 = ((($11)) + 1|0);
    $13 = $12;
    $14 = ($12>>>0)<($1>>>0);
    if ($14) {
     $6 = $12;$cur$sroa$0$01$i$i = $13;
    } else {
     $cur$sroa$0$0$lcssa$i$i = $13;
     break L1;
    }
   }
  } else {
   $cur$sroa$0$0$lcssa$i$i = $2;
  }
 } while(0);
 HEAP32[$parser>>2] = $cur$sroa$0$0$lcssa$i$i;
 $15 = (_ps_tofixedarray($parser,$1,$max_values,$values,$power_ten)|0);
 return ($15|0);
}
function _ps_tofixedarray($acur,$limit,$max_values,$values,$power_ten) {
 $acur = $acur|0;
 $limit = $limit|0;
 $max_values = $max_values|0;
 $values = $values|0;
 $power_ten = $power_ten|0;
 var $$ = 0, $$lcssa7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $count$0$lcssa4 = 0, $count$0$us = 0, $count$0$us8 = 0, $count$2 = 0, $cur = 0, $ender$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $cur = sp;
 $0 = HEAP32[$acur>>2]|0;
 HEAP32[$cur>>2] = $0;
 $1 = ($0>>>0)<($limit>>>0);
 $2 = $0;
 L1: do {
  if ($1) {
   $3 = HEAP8[$0>>0]|0;
   $4 = ($3<<24>>24)==(91);
   $5 = ($3<<24>>24)==(123);
   $$ = $5 ? 125 : 0;
   $ender$0 = $4 ? 93 : $$;
   $6 = ($ender$0<<24>>24)!=(0);
   if ($6) {
    $7 = ((($0)) + 1|0);
    HEAP32[$cur>>2] = $7;
    $8 = $7;
    $38 = $8;$39 = $7;
   } else {
    $38 = $2;$39 = $0;
   }
   $9 = ($values|0)==(0|0);
   L6: do {
    if ($9) {
     $11 = $39;$40 = $38;$count$0$us = 0;
     while(1) {
      $10 = ($11>>>0)<($limit>>>0);
      if (!($10)) {
       $37 = $40;$count$2 = $count$0$us;
       break L1;
      }
      _skip_spaces($cur,$limit);
      $12 = HEAP32[$cur>>2]|0;
      $13 = ($12>>>0)<($limit>>>0);
      $14 = $12;
      if (!($13)) {
       $37 = $14;$count$2 = $count$0$us;
       break L1;
      }
      $15 = HEAP8[$12>>0]|0;
      $16 = ($15<<24>>24)==($ender$0<<24>>24);
      if ($16) {
       $$lcssa7 = $12;$count$0$lcssa4 = $count$0$us;
       break L6;
      }
      (_PS_Conv_ToFixed($cur,$limit,$power_ten)|0);
      $17 = HEAP32[$cur>>2]|0;
      $18 = ($12|0)==($17|0);
      $19 = $17;
      if ($18) {
       $37 = $19;$count$2 = -1;
       break L1;
      }
      $20 = (($count$0$us) + 1)|0;
      if ($6) {
       $11 = $17;$40 = $19;$count$0$us = $20;
      } else {
       $37 = $19;$count$2 = $20;
       break L1;
      }
     }
    } else {
     $22 = $39;$41 = $38;$count$0$us8 = 0;
     while(1) {
      $21 = ($22>>>0)<($limit>>>0);
      if (!($21)) {
       $37 = $41;$count$2 = $count$0$us8;
       break L1;
      }
      _skip_spaces($cur,$limit);
      $23 = HEAP32[$cur>>2]|0;
      $24 = ($23>>>0)<($limit>>>0);
      $25 = $23;
      if (!($24)) {
       $37 = $25;$count$2 = $count$0$us8;
       break L1;
      }
      $26 = HEAP8[$23>>0]|0;
      $27 = ($26<<24>>24)==($ender$0<<24>>24);
      if ($27) {
       $$lcssa7 = $23;$count$0$lcssa4 = $count$0$us8;
       break L6;
      }
      $28 = ($count$0$us8|0)<($max_values|0);
      if (!($28)) {
       $37 = $25;$count$2 = $count$0$us8;
       break L1;
      }
      $29 = (_PS_Conv_ToFixed($cur,$limit,$power_ten)|0);
      $30 = (($values) + ($count$0$us8<<2)|0);
      HEAP32[$30>>2] = $29;
      $31 = HEAP32[$cur>>2]|0;
      $32 = ($23|0)==($31|0);
      $33 = $31;
      if ($32) {
       $37 = $33;$count$2 = -1;
       break L1;
      }
      $34 = (($count$0$us8) + 1)|0;
      if ($6) {
       $22 = $31;$41 = $33;$count$0$us8 = $34;
      } else {
       $37 = $33;$count$2 = $34;
       break L1;
      }
     }
    }
   } while(0);
   $35 = ((($$lcssa7)) + 1|0);
   HEAP32[$cur>>2] = $35;
   $36 = $35;
   $37 = $36;$count$2 = $count$0$lcssa4;
  } else {
   $37 = $2;$count$2 = 0;
  }
 } while(0);
 HEAP32[$acur>>2] = $37;
 STACKTOP = sp;return ($count$2|0);
}
function _skip_spaces($acur,$limit) {
 $acur = $acur|0;
 $limit = $limit|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cur$0$lcssa$i = 0, $cur$01$i = 0, $cur$sroa$0$0$lcssa = 0, $cur$sroa$0$01 = 0, $cur$sroa$0$1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$acur>>2]|0;
 $1 = $0;
 $2 = ($1>>>0)<($limit>>>0);
 L1: do {
  if ($2) {
   $4 = $1;$cur$sroa$0$01 = $0;
   while(1) {
    $3 = HEAP8[$4>>0]|0;
    switch ($3<<24>>24) {
    case 0: case 12: case 9: case 10: case 13: case 32:  {
     $cur$sroa$0$1 = $cur$sroa$0$01;
     break;
    }
    case 37:  {
     $5 = 37;$cur$01$i = $4;
     L5: while(1) {
      switch ($5<<24>>24) {
      case 10: case 13:  {
       $cur$0$lcssa$i = $cur$01$i;
       break L5;
       break;
      }
      default: {
      }
      }
      $6 = ((($cur$01$i)) + 1|0);
      $7 = ($6>>>0)<($limit>>>0);
      if (!($7)) {
       $cur$0$lcssa$i = $6;
       break;
      }
      $$pre = HEAP8[$6>>0]|0;
      $5 = $$pre;$cur$01$i = $6;
     }
     $8 = $cur$0$lcssa$i;
     $cur$sroa$0$1 = $8;
     break;
    }
    default: {
     $cur$sroa$0$0$lcssa = $cur$sroa$0$01;
     break L1;
    }
    }
    $9 = $cur$sroa$0$1;
    $10 = ((($9)) + 1|0);
    $11 = $10;
    $12 = ($10>>>0)<($limit>>>0);
    if ($12) {
     $4 = $10;$cur$sroa$0$01 = $11;
    } else {
     $cur$sroa$0$0$lcssa = $11;
     break L1;
    }
   }
  } else {
   $cur$sroa$0$0$lcssa = $0;
  }
 } while(0);
 HEAP32[$acur>>2] = $cur$sroa$0$0$lcssa;
 return;
}
function _ps_parser_to_token($parser,$token) {
 $parser = $parser|0;
 $token = $token|0;
 var $$cast = 0, $$cast59 = 0, $$lcssa170 = 0, $$lcssa171 = 0, $$lcssa171$lcssa = 0, $$lcssa175 = 0, $$lcssa176 = 0, $$lcssa179 = 0, $$lcssa183 = 0, $$lcssa186 = 0, $$off$i = 0, $$pr = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i19 = 0, $$pre$i$i48 = 0, $$pre$i$i5 = 0, $$pre$i14 = 0, $$pre$i37 = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cur$0$be$i = 0, $cur$0$be$i31 = 0, $cur$0$lcssa$i$i$i = 0, $cur$0$lcssa$i$i$i21 = 0, $cur$0$lcssa$i$i$i50 = 0, $cur$0$lcssa$i$i$i7 = 0, $cur$0$lcssa$i$ph$i = 0, $cur$0$ph16$i = 0, $cur$0$ph16$i26 = 0, $cur$01$i$i = 0, $cur$01$i$i$i = 0, $cur$01$i$i$i17 = 0, $cur$01$i$i$i3 = 0;
 var $cur$01$i$i$i46 = 0, $cur$09$i = 0, $cur$09$i29 = 0, $cur$16$i3381 = 0, $cur$16$i78 = 0, $cur$4$i$ph = 0, $cur$4$i42 = 0, $cur$i$sroa$0$0 = 0, $cur$i$sroa$0$1 = 0, $cur$sroa$0$0$lcssa$i$i = 0, $cur$sroa$0$0$lcssa$i$i24 = 0, $cur$sroa$0$01$i$i = 0, $cur$sroa$0$01$i$i1 = 0, $cur$sroa$0$01$i$i15 = 0, $cur$sroa$0$01$i$i44 = 0, $cur$sroa$0$1 = 0, $cur$sroa$0$1$i = 0, $cur$sroa$0$1$i$i = 0, $cur$sroa$0$1$i$i23 = 0, $cur$sroa$0$1$i$i52 = 0;
 var $cur$sroa$0$1$i$i9 = 0, $cur$sroa$0$1$ph = 0, $cur$sroa$0$158 = 0, $embed$0$ph$be$i = 0, $embed$0$ph$be$i39 = 0, $embed$0$ph15$i = 0, $embed$0$ph15$i27 = 0, $embed$05$i = 0, $embed$05$i$lcssa = 0, $embed$087 = 0, $embed$1 = 0, $embed$1$i = 0, $embed$2$i = 0, $error$03$i = 0, $error$1$i = 0, $error$2$i41 = 0, $i$05$i3482 = 0, $i$05$i79 = 0, $or$cond = 0, $or$cond$i = 0;
 var $or$cond$i36 = 0, $or$cond93 = 0, $storemerge$be = 0, $storemerge$ph62 = 0, $storemerge86 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($token)) + 8|0);
 HEAP32[$0>>2] = 0;
 HEAP32[$token>>2] = 0;
 $1 = ((($token)) + 4|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($parser)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$parser>>2]|0;
 $5 = $4;
 $6 = ($5>>>0)<($3>>>0);
 L1: do {
  if ($6) {
   $8 = $5;$cur$sroa$0$01$i$i = $4;
   while(1) {
    $7 = HEAP8[$8>>0]|0;
    switch ($7<<24>>24) {
    case 0: case 12: case 9: case 10: case 13: case 32:  {
     $cur$sroa$0$1$i$i = $cur$sroa$0$01$i$i;
     break;
    }
    case 37:  {
     $9 = 37;$cur$01$i$i$i = $8;
     L5: while(1) {
      switch ($9<<24>>24) {
      case 10: case 13:  {
       $cur$0$lcssa$i$i$i = $cur$01$i$i$i;
       break L5;
       break;
      }
      default: {
      }
      }
      $10 = ((($cur$01$i$i$i)) + 1|0);
      $11 = ($10>>>0)<($3>>>0);
      if (!($11)) {
       $cur$0$lcssa$i$i$i = $10;
       break;
      }
      $$pre$i$i = HEAP8[$10>>0]|0;
      $9 = $$pre$i$i;$cur$01$i$i$i = $10;
     }
     $12 = $cur$0$lcssa$i$i$i;
     $cur$sroa$0$1$i$i = $12;
     break;
    }
    default: {
     $cur$sroa$0$0$lcssa$i$i = $cur$sroa$0$01$i$i;
     break L1;
    }
    }
    $13 = $cur$sroa$0$1$i$i;
    $14 = ((($13)) + 1|0);
    $15 = $14;
    $16 = ($14>>>0)<($3>>>0);
    if ($16) {
     $8 = $14;$cur$sroa$0$01$i$i = $15;
    } else {
     $cur$sroa$0$0$lcssa$i$i = $15;
     break L1;
    }
   }
  } else {
   $cur$sroa$0$0$lcssa$i$i = $4;
  }
 } while(0);
 HEAP32[$parser>>2] = $cur$sroa$0$0$lcssa$i$i;
 $$cast59 = $cur$sroa$0$0$lcssa$i$i;
 $17 = ($$cast59>>>0)<($3>>>0);
 if (!($17)) {
  return;
 }
 $18 = HEAP8[$$cast59>>0]|0;
 $19 = $18&255;
 L15: do {
  switch ($19|0) {
  case 40:  {
   HEAP32[$0>>2] = 2;
   HEAP32[$token>>2] = $cur$sroa$0$0$lcssa$i$i;
   $cur$0$ph16$i = $$cast59;$embed$0$ph15$i = 0;
   L19: while(1) {
    $cur$09$i = $cur$0$ph16$i;
    L21: while(1) {
     $20 = HEAP8[$cur$09$i>>0]|0;
     $21 = ((($cur$09$i)) + 1|0);
     L23: do {
      switch ($20<<24>>24) {
      case 40:  {
       $$lcssa170 = $21;
       label = 20;
       break L21;
       break;
      }
      case 41:  {
       $$lcssa171 = $21;
       label = 22;
       break L21;
       break;
      }
      case 92:  {
       $22 = ($21|0)==($3|0);
       if ($22) {
        $cur$4$i$ph = $3;
        label = 23;
        break L19;
       }
       $23 = HEAP8[$21>>0]|0;
       $24 = $23&255;
       switch ($24|0) {
       case 41: case 40: case 92: case 102: case 98: case 116: case 114: case 110:  {
        $28 = ((($cur$09$i)) + 2|0);
        $cur$0$be$i = $28;
        break L23;
        break;
       }
       default: {
       }
       }
       $25 = ($21>>>0)<($3>>>0);
       $26 = $23 & -8;
       $27 = ($26<<24>>24)==(48);
       $or$cond = $25 & $27;
       if ($or$cond) {
        $cur$16$i78 = $21;$i$05$i79 = 0;
        while(1) {
         $30 = ((($cur$16$i78)) + 1|0);
         $31 = (($i$05$i79) + 1)|0;
         $32 = ($31>>>0)<(3);
         $33 = ($30>>>0)<($3>>>0);
         $or$cond$i = $33 & $32;
         if (!($or$cond$i)) {
          $cur$0$be$i = $30;
          break L23;
         }
         $$pre$i = HEAP8[$30>>0]|0;
         $34 = $$pre$i & -8;
         $35 = ($34<<24>>24)==(48);
         if ($35) {
          $cur$16$i78 = $30;$i$05$i79 = $31;
         } else {
          $cur$0$be$i = $30;
          break;
         }
        }
       } else {
        $cur$0$be$i = $21;
       }
       break;
      }
      default: {
       $cur$0$be$i = $21;
      }
      }
     } while(0);
     $29 = ($cur$0$be$i>>>0)<($3>>>0);
     if ($29) {
      $cur$09$i = $cur$0$be$i;
     } else {
      $cur$4$i$ph = $cur$0$be$i;
      label = 23;
      break L19;
     }
    }
    if ((label|0) == 20) {
     label = 0;
     $36 = (($embed$0$ph15$i) + 1)|0;
     $38 = $$lcssa170;$embed$0$ph$be$i = $36;
    }
    else if ((label|0) == 22) {
     label = 0;
     $39 = (($embed$0$ph15$i) + -1)|0;
     $40 = ($39|0)==(0);
     if ($40) {
      $$lcssa171$lcssa = $$lcssa171;
      label = 24;
      break;
     } else {
      $38 = $$lcssa171;$embed$0$ph$be$i = $39;
     }
    }
    $37 = ($38>>>0)<($3>>>0);
    if ($37) {
     $cur$0$ph16$i = $38;$embed$0$ph15$i = $embed$0$ph$be$i;
    } else {
     $cur$4$i$ph = $38;
     label = 23;
     break;
    }
   }
   if ((label|0) == 23) {
    $41 = $cur$4$i$ph;
    $cur$sroa$0$1$ph = $41;
    label = 90;
    break L15;
   }
   else if ((label|0) == 24) {
    $42 = $$lcssa171$lcssa;
    HEAP32[$1>>2] = $42;
    $cur$sroa$0$158 = $42;
    break L15;
   }
   break;
  }
  case 123:  {
   HEAP32[$0>>2] = 3;
   HEAP32[$token>>2] = $cur$sroa$0$0$lcssa$i$i;
   $44 = $$cast59;$cur$i$sroa$0$0 = $cur$sroa$0$0$lcssa$i$i;$embed$05$i = 0;
   L41: while(1) {
    $43 = HEAP8[$44>>0]|0;
    $45 = $43&255;
    switch ($45|0) {
    case 123:  {
     $46 = (($embed$05$i) + 1)|0;
     $cur$i$sroa$0$1 = $cur$i$sroa$0$0;$embed$1$i = $46;$error$1$i = 0;
     break;
    }
    case 125:  {
     $47 = (($embed$05$i) + -1)|0;
     $48 = ($47|0)==(0);
     if ($48) {
      $$lcssa179 = $44;
      label = 29;
      break L41;
     } else {
      $cur$i$sroa$0$1 = $cur$i$sroa$0$0;$embed$1$i = $47;$error$1$i = 0;
     }
     break;
    }
    case 40:  {
     $51 = $cur$i$sroa$0$0;
     $52 = ($51>>>0)<($3>>>0);
     L47: do {
      if ($52) {
       $cur$0$ph16$i26 = $51;$embed$0$ph15$i27 = 0;
       while(1) {
        $cur$09$i29 = $cur$0$ph16$i26;
        L50: while(1) {
         $53 = HEAP8[$cur$09$i29>>0]|0;
         $54 = ((($cur$09$i29)) + 1|0);
         L52: do {
          switch ($53<<24>>24) {
          case 40:  {
           $$lcssa175 = $54;
           label = 40;
           break L50;
           break;
          }
          case 41:  {
           $$lcssa176 = $54;
           label = 42;
           break L50;
           break;
          }
          case 92:  {
           $55 = ($54|0)==($3|0);
           if ($55) {
            $cur$4$i42 = $3;$error$2$i41 = 3;
            break L47;
           }
           $56 = HEAP8[$54>>0]|0;
           $57 = $56&255;
           switch ($57|0) {
           case 41: case 40: case 92: case 102: case 98: case 116: case 114: case 110:  {
            $61 = ((($cur$09$i29)) + 2|0);
            $cur$0$be$i31 = $61;
            break L52;
            break;
           }
           default: {
           }
           }
           $58 = ($54>>>0)<($3>>>0);
           $59 = $56 & -8;
           $60 = ($59<<24>>24)==(48);
           $or$cond93 = $58 & $60;
           if ($or$cond93) {
            $cur$16$i3381 = $54;$i$05$i3482 = 0;
            while(1) {
             $63 = ((($cur$16$i3381)) + 1|0);
             $64 = (($i$05$i3482) + 1)|0;
             $65 = ($64>>>0)<(3);
             $66 = ($63>>>0)<($3>>>0);
             $or$cond$i36 = $66 & $65;
             if (!($or$cond$i36)) {
              $cur$0$be$i31 = $63;
              break L52;
             }
             $$pre$i37 = HEAP8[$63>>0]|0;
             $67 = $$pre$i37 & -8;
             $68 = ($67<<24>>24)==(48);
             if ($68) {
              $cur$16$i3381 = $63;$i$05$i3482 = $64;
             } else {
              $cur$0$be$i31 = $63;
              break;
             }
            }
           } else {
            $cur$0$be$i31 = $54;
           }
           break;
          }
          default: {
           $cur$0$be$i31 = $54;
          }
          }
         } while(0);
         $62 = ($cur$0$be$i31>>>0)<($3>>>0);
         if ($62) {
          $cur$09$i29 = $cur$0$be$i31;
         } else {
          $cur$4$i42 = $cur$0$be$i31;$error$2$i41 = 3;
          break L47;
         }
        }
        if ((label|0) == 40) {
         label = 0;
         $69 = (($embed$0$ph15$i27) + 1)|0;
         $71 = $$lcssa175;$embed$0$ph$be$i39 = $69;
        }
        else if ((label|0) == 42) {
         label = 0;
         $72 = (($embed$0$ph15$i27) + -1)|0;
         $73 = ($72|0)==(0);
         if ($73) {
          $cur$4$i42 = $$lcssa176;$error$2$i41 = 0;
          break L47;
         } else {
          $71 = $$lcssa176;$embed$0$ph$be$i39 = $72;
         }
        }
        $70 = ($71>>>0)<($3>>>0);
        if ($70) {
         $cur$0$ph16$i26 = $71;$embed$0$ph15$i27 = $embed$0$ph$be$i39;
        } else {
         $cur$4$i42 = $71;$error$2$i41 = 3;
         break;
        }
       }
      } else {
       $cur$4$i42 = $51;$error$2$i41 = 3;
      }
     } while(0);
     $74 = $cur$4$i42;
     $cur$i$sroa$0$1 = $74;$embed$1$i = $embed$05$i;$error$1$i = $error$2$i41;
     break;
    }
    case 60:  {
     $75 = $cur$i$sroa$0$0;
     $76 = ((($75)) + 1|0);
     $77 = $76;
     $78 = ($76>>>0)<($3>>>0);
     L68: do {
      if ($78) {
       $80 = $76;$cur$sroa$0$01$i$i15 = $77;
       while(1) {
        $79 = HEAP8[$80>>0]|0;
        switch ($79<<24>>24) {
        case 0: case 12: case 9: case 10: case 13: case 32:  {
         $cur$sroa$0$1$i$i23 = $cur$sroa$0$01$i$i15;
         label = 50;
         break;
        }
        case 37:  {
         $81 = 37;$cur$01$i$i$i17 = $80;
         L72: while(1) {
          switch ($81<<24>>24) {
          case 10: case 13:  {
           $cur$0$lcssa$i$i$i21 = $cur$01$i$i$i17;
           break L72;
           break;
          }
          default: {
          }
          }
          $82 = ((($cur$01$i$i$i17)) + 1|0);
          $83 = ($82>>>0)<($3>>>0);
          if (!($83)) {
           $cur$0$lcssa$i$i$i21 = $82;
           break;
          }
          $$pre$i$i19 = HEAP8[$82>>0]|0;
          $81 = $$pre$i$i19;$cur$01$i$i$i17 = $82;
         }
         $84 = $cur$0$lcssa$i$i$i21;
         $cur$sroa$0$1$i$i23 = $84;
         label = 50;
         break;
        }
        default: {
         $cur$sroa$0$0$lcssa$i$i24 = $cur$sroa$0$01$i$i15;
        }
        }
        if ((label|0) == 50) {
         label = 0;
         $85 = $cur$sroa$0$1$i$i23;
         $86 = ((($85)) + 1|0);
         $87 = $86;
         $88 = ($86>>>0)<($3>>>0);
         if ($88) {
          $80 = $86;$cur$sroa$0$01$i$i15 = $87;
          continue;
         } else {
          $cur$sroa$0$0$lcssa$i$i24 = $87;
         }
        }
        $89 = $cur$sroa$0$0$lcssa$i$i24;
        $90 = ($89>>>0)<($3>>>0);
        if (!($90)) {
         $cur$sroa$0$1$i = $cur$sroa$0$0$lcssa$i$i24;
         break L68;
        }
        $91 = HEAP8[$89>>0]|0;
        $$off$i = (($91) + -48)<<24>>24;
        $92 = ($$off$i&255)<(10);
        if (!($92)) {
         switch ($91<<24>>24) {
         case 65: case 66: case 67: case 68: case 69: case 70: case 97: case 98: case 99: case 100: case 101: case 102:  {
          break;
         }
         default: {
          $cur$sroa$0$1$i = $cur$sroa$0$0$lcssa$i$i24;
          break L68;
         }
         }
        }
        $93 = ((($89)) + 1|0);
        $94 = $93;
        $95 = ($93>>>0)<($3>>>0);
        if ($95) {
         $80 = $93;$cur$sroa$0$01$i$i15 = $94;
        } else {
         $cur$sroa$0$1$i = $94;
         break L68;
        }
       }
      } else {
       $cur$sroa$0$1$i = $77;
      }
     } while(0);
     $96 = $cur$sroa$0$1$i;
     $97 = ($96>>>0)<($3>>>0);
     if ($97) {
      $98 = HEAP8[$96>>0]|0;
      $99 = ($98<<24>>24)==(62);
      if (!($99)) {
       $$lcssa183 = $96;$embed$05$i$lcssa = $embed$05$i;
       label = 57;
       break L41;
      }
     }
     $102 = ((($96)) + 1|0);
     $103 = $102;
     $cur$i$sroa$0$1 = $103;$embed$1$i = $embed$05$i;$error$1$i = 0;
     break;
    }
    case 37:  {
     $104 = $43;$cur$01$i$i = $44;
     L88: while(1) {
      switch ($104<<24>>24) {
      case 10: case 13:  {
       $cur$0$lcssa$i$ph$i = $cur$01$i$i;
       break L88;
       break;
      }
      default: {
      }
      }
      $105 = ((($cur$01$i$i)) + 1|0);
      $106 = ($105>>>0)<($3>>>0);
      if (!($106)) {
       $cur$0$lcssa$i$ph$i = $105;
       break;
      }
      $$pre$i14 = HEAP8[$105>>0]|0;
      $104 = $$pre$i14;$cur$01$i$i = $105;
     }
     $107 = $cur$0$lcssa$i$ph$i;
     $cur$i$sroa$0$1 = $107;$embed$1$i = $embed$05$i;$error$1$i = 0;
     break;
    }
    default: {
     $cur$i$sroa$0$1 = $cur$i$sroa$0$0;$embed$1$i = $embed$05$i;$error$1$i = 0;
    }
    }
    $108 = $cur$i$sroa$0$1;
    $109 = ((($108)) + 1|0);
    $110 = $109;
    $111 = ($109>>>0)<($3>>>0);
    $112 = ($error$1$i|0)==(0);
    $113 = $112 & $111;
    if ($113) {
     $44 = $109;$cur$i$sroa$0$0 = $110;$embed$05$i = $embed$1$i;
    } else {
     $116 = $110;$embed$2$i = $embed$1$i;$error$03$i = $error$1$i;
     break;
    }
   }
   if ((label|0) == 29) {
    $49 = ((($$lcssa179)) + 1|0);
    $50 = $49;
    $116 = $50;$embed$2$i = 0;$error$03$i = 0;
   }
   else if ((label|0) == 57) {
    $100 = ((($$lcssa183)) + 1|0);
    $101 = $100;
    $116 = $101;$embed$2$i = $embed$05$i$lcssa;$error$03$i = 3;
   }
   $114 = $embed$2$i | $error$03$i;
   $115 = ($114|0)==(0);
   if ($115) {
    HEAP32[$1>>2] = $116;
    $117 = $116;
    $167 = $117;$cur$sroa$0$1 = $116;
    label = 91;
   } else {
    $cur$sroa$0$1$ph = $116;
    label = 90;
   }
   break;
  }
  case 91:  {
   HEAP32[$0>>2] = 3;
   $118 = ((($$cast59)) + 1|0);
   HEAP32[$token>>2] = $$cast59;
   $$cast = $118;
   HEAP32[$parser>>2] = $$cast;
   $119 = ($118>>>0)<($3>>>0);
   L100: do {
    if ($119) {
     $121 = $118;$cur$sroa$0$01$i$i44 = $$cast;
     while(1) {
      $120 = HEAP8[$121>>0]|0;
      switch ($120<<24>>24) {
      case 0: case 12: case 9: case 10: case 13: case 32:  {
       $cur$sroa$0$1$i$i52 = $cur$sroa$0$01$i$i44;
       break;
      }
      case 37:  {
       $122 = 37;$cur$01$i$i$i46 = $121;
       L104: while(1) {
        switch ($122<<24>>24) {
        case 10: case 13:  {
         $cur$0$lcssa$i$i$i50 = $cur$01$i$i$i46;
         break L104;
         break;
        }
        default: {
        }
        }
        $123 = ((($cur$01$i$i$i46)) + 1|0);
        $124 = ($123>>>0)<($3>>>0);
        if (!($124)) {
         $cur$0$lcssa$i$i$i50 = $123;
         break;
        }
        $$pre$i$i48 = HEAP8[$123>>0]|0;
        $122 = $$pre$i$i48;$cur$01$i$i$i46 = $123;
       }
       $125 = $cur$0$lcssa$i$i$i50;
       $cur$sroa$0$1$i$i52 = $125;
       break;
      }
      default: {
       $storemerge$ph62 = $cur$sroa$0$01$i$i44;
       break L100;
      }
      }
      $126 = $cur$sroa$0$1$i$i52;
      $127 = ((($126)) + 1|0);
      $128 = $127;
      $129 = ($127>>>0)<($3>>>0);
      if ($129) {
       $121 = $127;$cur$sroa$0$01$i$i44 = $128;
      } else {
       $storemerge$ph62 = $128;
       break L100;
      }
     }
    } else {
     $storemerge$ph62 = $$cast;
    }
   } while(0);
   HEAP32[$parser>>2] = $storemerge$ph62;
   $130 = $storemerge$ph62;
   $131 = ($130>>>0)<($3>>>0);
   if ($131) {
    $132 = ((($parser)) + 12|0);
    $136 = $130;$embed$087 = 1;$storemerge86 = $storemerge$ph62;
    L112: while(1) {
     $133 = HEAP32[$132>>2]|0;
     $134 = ($133|0)==(0);
     if (!($134)) {
      $cur$sroa$0$1$ph = $storemerge86;
      label = 90;
      break L15;
     }
     $135 = HEAP8[$136>>0]|0;
     switch ($135<<24>>24) {
     case 91:  {
      $137 = (($embed$087) + 1)|0;
      $embed$1 = $137;
      break;
     }
     case 93:  {
      $138 = (($embed$087) + -1)|0;
      $139 = ($embed$087|0)<(2);
      if ($139) {
       $$lcssa186 = $136;
       break L112;
      } else {
       $embed$1 = $138;
      }
      break;
     }
     default: {
      $embed$1 = $embed$087;
     }
     }
     HEAP32[$parser>>2] = $storemerge86;
     _ps_parser_skip_PS_token($parser);
     $142 = HEAP32[$2>>2]|0;
     $143 = HEAP32[$parser>>2]|0;
     $144 = $143;
     $145 = ($144>>>0)<($142>>>0);
     L119: do {
      if ($145) {
       $149 = $144;$cur$sroa$0$01$i$i1 = $143;
       while(1) {
        $148 = HEAP8[$149>>0]|0;
        switch ($148<<24>>24) {
        case 0: case 12: case 9: case 10: case 13: case 32:  {
         $cur$sroa$0$1$i$i9 = $cur$sroa$0$01$i$i1;
         break;
        }
        case 37:  {
         $150 = 37;$cur$01$i$i$i3 = $149;
         L123: while(1) {
          switch ($150<<24>>24) {
          case 10: case 13:  {
           $cur$0$lcssa$i$i$i7 = $cur$01$i$i$i3;
           break L123;
           break;
          }
          default: {
          }
          }
          $151 = ((($cur$01$i$i$i3)) + 1|0);
          $152 = ($151>>>0)<($142>>>0);
          if (!($152)) {
           $cur$0$lcssa$i$i$i7 = $151;
           break;
          }
          $$pre$i$i5 = HEAP8[$151>>0]|0;
          $150 = $$pre$i$i5;$cur$01$i$i$i3 = $151;
         }
         $153 = $cur$0$lcssa$i$i$i7;
         $cur$sroa$0$1$i$i9 = $153;
         break;
        }
        default: {
         $storemerge$be = $cur$sroa$0$01$i$i1;
         break L119;
        }
        }
        $154 = $cur$sroa$0$1$i$i9;
        $155 = ((($154)) + 1|0);
        $156 = $155;
        $157 = ($155>>>0)<($142>>>0);
        if ($157) {
         $149 = $155;$cur$sroa$0$01$i$i1 = $156;
        } else {
         $storemerge$be = $156;
         break L119;
        }
       }
      } else {
       $storemerge$be = $143;
      }
     } while(0);
     HEAP32[$parser>>2] = $storemerge$be;
     $146 = $storemerge$be;
     $147 = ($146>>>0)<($3>>>0);
     if ($147) {
      $136 = $146;$embed$087 = $embed$1;$storemerge86 = $storemerge$be;
     } else {
      $cur$sroa$0$1$ph = $storemerge$be;
      label = 90;
      break L15;
     }
    }
    $140 = ((($$lcssa186)) + 1|0);
    $141 = $140;
    HEAP32[$1>>2] = $140;
    $cur$sroa$0$158 = $141;
   } else {
    $cur$sroa$0$1$ph = $storemerge$ph62;
    label = 90;
   }
   break;
  }
  default: {
   HEAP32[$token>>2] = $$cast59;
   $158 = HEAP8[$$cast59>>0]|0;
   $159 = ($158<<24>>24)==(47);
   $160 = $159 ? 4 : 1;
   HEAP32[$0>>2] = $160;
   _ps_parser_skip_PS_token($parser);
   $161 = HEAP32[$parser>>2]|0;
   $162 = ((($parser)) + 12|0);
   $163 = HEAP32[$162>>2]|0;
   $164 = ($163|0)==(0);
   if ($164) {
    HEAP32[$1>>2] = $161;
    $165 = $161;
    $167 = $165;$cur$sroa$0$1 = $161;
    label = 91;
   } else {
    $cur$sroa$0$1$ph = $161;
    label = 90;
   }
  }
  }
 } while(0);
 if ((label|0) == 90) {
  $$pr = HEAP32[$1>>2]|0;
  $167 = $$pr;$cur$sroa$0$1 = $cur$sroa$0$1$ph;
  label = 91;
 }
 if ((label|0) == 91) {
  $166 = ($167|0)==(0|0);
  if ($166) {
   HEAP32[$token>>2] = 0;
   HEAP32[$0>>2] = 0;
   $cur$sroa$0$158 = $cur$sroa$0$1;
  } else {
   $cur$sroa$0$158 = $cur$sroa$0$1;
  }
 }
 HEAP32[$parser>>2] = $cur$sroa$0$158;
 return;
}
function _ps_parser_to_token_array($parser,$tokens,$max_tokens,$pnum_tokens) {
 $parser = $parser|0;
 $tokens = $tokens|0;
 $max_tokens = $max_tokens|0;
 $pnum_tokens = $pnum_tokens|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cur$0$lcssa1 = 0, $cur$02 = 0, $cur$03 = 0, $cur$03$us = 0, $master = 0, $token = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $master = sp + 12|0;
 $token = sp;
 HEAP32[$pnum_tokens>>2] = -1;
 _ps_parser_to_token($parser,$master);
 $0 = ((($master)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(3);
 if (!($2)) {
  STACKTOP = sp;return;
 }
 $3 = HEAP32[$parser>>2]|0;
 $4 = ((($parser)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($tokens) + (($max_tokens*12)|0)|0);
 $7 = HEAP32[$master>>2]|0;
 $8 = ((($7)) + 1|0);
 HEAP32[$parser>>2] = $8;
 $9 = ((($master)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + -1|0);
 HEAP32[$4>>2] = $11;
 $12 = ($8>>>0)<($11>>>0);
 L4: do {
  if ($12) {
   $13 = ((($token)) + 8|0);
   $14 = ($tokens|0)==(0|0);
   L6: do {
    if ($14) {
     $cur$03 = 0;
     while(1) {
      _ps_parser_to_token($parser,$token);
      $22 = HEAP32[$13>>2]|0;
      $23 = ($22|0)==(0);
      if ($23) {
       $cur$0$lcssa1 = $cur$03;
       break L6;
      }
      $24 = ((($cur$03)) + 12|0);
      $25 = HEAP32[$parser>>2]|0;
      $26 = HEAP32[$4>>2]|0;
      $27 = ($25>>>0)<($26>>>0);
      if ($27) {
       $cur$03 = $24;
      } else {
       $cur$02 = $24;
       break L4;
      }
     }
    } else {
     $cur$03$us = $tokens;
     while(1) {
      _ps_parser_to_token($parser,$token);
      $15 = HEAP32[$13>>2]|0;
      $16 = ($15|0)==(0);
      if ($16) {
       $cur$0$lcssa1 = $cur$03$us;
       break L6;
      }
      $17 = ($cur$03$us>>>0)<($6>>>0);
      if ($17) {
       ;HEAP32[$cur$03$us>>2]=HEAP32[$token>>2]|0;HEAP32[$cur$03$us+4>>2]=HEAP32[$token+4>>2]|0;HEAP32[$cur$03$us+8>>2]=HEAP32[$token+8>>2]|0;
      }
      $18 = ((($cur$03$us)) + 12|0);
      $19 = HEAP32[$parser>>2]|0;
      $20 = HEAP32[$4>>2]|0;
      $21 = ($19>>>0)<($20>>>0);
      if ($21) {
       $cur$03$us = $18;
      } else {
       $cur$02 = $18;
       break L4;
      }
     }
    }
   } while(0);
   $cur$02 = $cur$0$lcssa1;
  } else {
   $cur$02 = $tokens;
  }
 } while(0);
 $28 = $cur$02;
 $29 = $tokens;
 $30 = (($28) - ($29))|0;
 $31 = (($30|0) / 12)&-1;
 HEAP32[$pnum_tokens>>2] = $31;
 HEAP32[$parser>>2] = $3;
 HEAP32[$4>>2] = $5;
 STACKTOP = sp;return;
}
function _ps_parser_load_field($parser,$field,$objects,$max_objects,$pflags) {
 $parser = $parser|0;
 $field = $field|0;
 $objects = $objects|0;
 $max_objects = $max_objects|0;
 $pflags = $pflags|0;
 var $$$i = 0, $$$i17 = 0, $$cast = 0, $$cast$lcssa = 0, $$cast$lcssa193 = 0, $$cast$us = 0, $$ph = 0, $$ph106 = 0, $$pre = 0, $$pre$i = 0, $$pre$i10 = 0, $$pre$i10$1 = 0, $$pre$i10$2 = 0, $$pre$i10$3 = 0, $$pre$i42 = 0, $$pre$i42$us = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0;
 var $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0;
 var $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0;
 var $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0;
 var $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0;
 var $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $base$off$i = 0, $count$0 = 0, $count$0$us8$i$lcssa69 = 0, $count$0$us8$i$us148 = 0, $count$1$ph102 = 0, $count$188 = 0, $count$2$i = 0, $cur = 0, $cur$0$i = 0, $cur$0$lcssa$i$i = 0, $cur$0$lcssa$i$i12 = 0, $cur$0$lcssa$i$i12$1 = 0;
 var $cur$0$lcssa$i$i12$2 = 0, $cur$0$lcssa$i$i12$3 = 0, $cur$0$lcssa$i$i44 = 0, $cur$0$lcssa$i$i44$us = 0, $cur$01$i$i = 0, $cur$01$i$i40 = 0, $cur$01$i$i40$us = 0, $cur$01$i$i8 = 0, $cur$01$i$i8$1 = 0, $cur$01$i$i8$2 = 0, $cur$01$i$i8$3 = 0, $cur$i = 0, $cur$sroa$0$0$lcssa$i = 0, $cur$sroa$0$0$lcssa$i15 = 0, $cur$sroa$0$0$lcssa$i15$1 = 0, $cur$sroa$0$0$lcssa$i15$2 = 0, $cur$sroa$0$0$lcssa$i15$3 = 0, $cur$sroa$0$0$lcssa$i47 = 0, $cur$sroa$0$0$lcssa$i47$us = 0, $cur$sroa$0$0$lcssa$i47$us$lcssa179 = 0;
 var $cur$sroa$0$01$i = 0, $cur$sroa$0$01$i38 = 0, $cur$sroa$0$01$i38$us = 0, $cur$sroa$0$01$i6 = 0, $cur$sroa$0$01$i6$1 = 0, $cur$sroa$0$01$i6$2 = 0, $cur$sroa$0$01$i6$3 = 0, $cur$sroa$0$1$i = 0, $cur$sroa$0$1$i14 = 0, $cur$sroa$0$1$i14$1 = 0, $cur$sroa$0$1$i14$2 = 0, $cur$sroa$0$1$i14$3 = 0, $cur$sroa$0$1$i46 = 0, $cur$sroa$0$1$i46$us = 0, $ender$0$i = 0, $error = 0, $exitcond = 0, $have_overflow$0$lcssa$i = 0, $have_overflow$0$lcssa$i32 = 0, $have_overflow$07$i = 0;
 var $have_overflow$07$i25 = 0, $have_overflow$2$ph$i = 0, $have_overflow$2$ph$i28 = 0, $i$186 = 0, $idx$0$ph101 = 0, $idx$087 = 0, $len$0 = 0, $limit$0$ph103 = 0, $num$0$i = 0, $num$0$lcssa$i = 0, $num$0$lcssa$i31 = 0, $num$08$i = 0, $num$08$i24 = 0, $num$2$ph$i = 0, $num$2$ph$i29 = 0, $num$3$$i = 0, $num$3$$i33 = 0, $num$5$i = 0, $num$5$i34 = 0, $or$cond = 0;
 var $or$cond$1 = 0, $or$cond$2 = 0, $or$cond$3 = 0, $or$cond$i = 0, $or$cond4$i = 0, $or$cond44$i = 0, $or$cond44$i27 = 0, $p$0$i = 0, $p$0$i21 = 0, $p$1$lcssa$i = 0, $p$1$lcssa$i30 = 0, $p$19$i = 0, $p$19$i23 = 0, $result$0$i = 0, $sext$i = 0, $sign$0$i = 0, $sign$0$i20 = 0, $temp = 0, $token = 0, $token2 = 0;
 var $type$1 = 0, $type$2$ph100 = 0, $val$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $cur$i = sp + 48|0;
 $token = sp + 36|0;
 $cur = sp + 32|0;
 $error = sp + 28|0;
 $token2 = sp + 16|0;
 $temp = sp;
 _ps_parser_to_token($parser,$token);
 $0 = ((($token)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 L1: do {
  if (!($2)) {
   $3 = HEAP32[$token>>2]|0;
   HEAP32[$cur>>2] = $3;
   $4 = ((($token)) + 4|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = ((($field)) + 8|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(7);
   $9 = $3;
   if ($8) {
    $10 = HEAP32[$parser>>2]|0;
    $11 = ((($parser)) + 8|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = ((($9)) + 1|0);
    HEAP32[$parser>>2] = $13;
    $14 = ((($5)) + -1|0);
    HEAP32[$11>>2] = $14;
    _ps_parser_to_token($parser,$token2);
    HEAP32[$parser>>2] = $10;
    HEAP32[$11>>2] = $12;
    $15 = ((($token2)) + 8|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ($16|0)==(3);
    if ($17) {
     $count$0 = 1;$type$1 = 8;
     label = 5;
    } else {
     $331 = $3;$count$1$ph102 = 1;$idx$0$ph101 = 0;$limit$0$ph103 = $5;$type$2$ph100 = 7;
     label = 7;
    }
   } else {
    $18 = ($1|0)==(3);
    if ($18) {
     $count$0 = $max_objects;$type$1 = $7;
     label = 5;
    } else {
     $331 = $3;$count$1$ph102 = 1;$idx$0$ph101 = 0;$limit$0$ph103 = $5;$type$2$ph100 = $7;
     label = 7;
    }
   }
   if ((label|0) == 5) {
    $19 = ($max_objects|0)==(0);
    if ($19) {
     break;
    }
    $20 = ((($9)) + 1|0);
    HEAP32[$cur>>2] = $20;
    $21 = ((($5)) + -1|0);
    $22 = $20;
    $23 = ($count$0|0)==(0);
    if (!($23)) {
     $331 = $22;$count$1$ph102 = $count$0;$idx$0$ph101 = 1;$limit$0$ph103 = $21;$type$2$ph100 = $type$1;
     label = 7;
    }
   }
   L9: do {
    if ((label|0) == 7) {
     $24 = ((($field)) + 16|0);
     $25 = ((($field)) + 20|0);
     $26 = ((($parser)) + 16|0);
     $27 = $limit$0$ph103;
     $28 = ((($temp)) + 4|0);
     $29 = ((($temp)) + 8|0);
     $30 = ((($temp)) + 12|0);
     $31 = $max_objects << 2;
     $32 = ($max_objects|0)==(0);
     $33 = $max_objects << 1;
     $34 = ($max_objects*3)|0;
     $35 = $max_objects << 1;
     $41 = $331;$count$188 = $count$1$ph102;$idx$087 = $idx$0$ph101;
     L11: while(1) {
      $36 = (($objects) + ($idx$087<<2)|0);
      $37 = HEAP32[$36>>2]|0;
      $38 = HEAP32[$24>>2]|0;
      $39 = (($37) + ($38)|0);
      $40 = $41;
      $42 = ($40>>>0)<($limit$0$ph103>>>0);
      L13: do {
       if ($42) {
        $44 = $40;$cur$sroa$0$01$i = $41;
        while(1) {
         $43 = HEAP8[$44>>0]|0;
         switch ($43<<24>>24) {
         case 0: case 12: case 9: case 10: case 13: case 32:  {
          $cur$sroa$0$1$i = $cur$sroa$0$01$i;
          break;
         }
         case 37:  {
          $45 = 37;$cur$01$i$i = $44;
          L17: while(1) {
           switch ($45<<24>>24) {
           case 10: case 13:  {
            $cur$0$lcssa$i$i = $cur$01$i$i;
            break L17;
            break;
           }
           default: {
           }
           }
           $46 = ((($cur$01$i$i)) + 1|0);
           $47 = ($46>>>0)<($limit$0$ph103>>>0);
           if (!($47)) {
            $cur$0$lcssa$i$i = $46;
            break;
           }
           $$pre$i = HEAP8[$46>>0]|0;
           $45 = $$pre$i;$cur$01$i$i = $46;
          }
          $48 = $cur$0$lcssa$i$i;
          $cur$sroa$0$1$i = $48;
          break;
         }
         default: {
          $cur$sroa$0$0$lcssa$i = $cur$sroa$0$01$i;
          break L13;
         }
         }
         $49 = $cur$sroa$0$1$i;
         $50 = ((($49)) + 1|0);
         $51 = $50;
         $52 = ($50>>>0)<($limit$0$ph103>>>0);
         if ($52) {
          $44 = $50;$cur$sroa$0$01$i = $51;
         } else {
          $cur$sroa$0$0$lcssa$i = $51;
          break L13;
         }
        }
       } else {
        $cur$sroa$0$0$lcssa$i = $41;
       }
      } while(0);
      HEAP32[$cur>>2] = $cur$sroa$0$0$lcssa$i;
      $53 = $cur$sroa$0$0$lcssa$i;
      L24: do {
       switch ($type$2$ph100|0) {
       case 1:  {
        $54 = ((($53)) + 3|0);
        $55 = ($54>>>0)<($limit$0$ph103>>>0);
        if ($55) {
         $56 = HEAP8[$53>>0]|0;
         $57 = ($56<<24>>24)==(116);
         if ($57) {
          $58 = ((($53)) + 1|0);
          $59 = HEAP8[$58>>0]|0;
          $60 = ($59<<24>>24)==(114);
          if ($60) {
           $61 = ((($53)) + 2|0);
           $62 = HEAP8[$61>>0]|0;
           $63 = ($62<<24>>24)==(117);
           if ($63) {
            $64 = HEAP8[$54>>0]|0;
            $65 = ($64<<24>>24)==(101);
            if ($65) {
             $66 = ((($53)) + 5|0);
             $cur$0$i = $66;$result$0$i = 1;
            } else {
             label = 22;
            }
           } else {
            label = 22;
           }
          } else {
           label = 22;
          }
         } else {
          label = 22;
         }
        } else {
         label = 22;
        }
        if ((label|0) == 22) {
         label = 0;
         $67 = ((($53)) + 4|0);
         $68 = ($67>>>0)<($limit$0$ph103>>>0);
         if ($68) {
          $69 = HEAP8[$53>>0]|0;
          $70 = ($69<<24>>24)==(102);
          if ($70) {
           $71 = ((($53)) + 1|0);
           $72 = HEAP8[$71>>0]|0;
           $73 = ($72<<24>>24)==(97);
           if ($73) {
            $74 = ((($53)) + 2|0);
            $75 = HEAP8[$74>>0]|0;
            $76 = ($75<<24>>24)==(108);
            if ($76) {
             $77 = HEAP8[$54>>0]|0;
             $78 = ($77<<24>>24)==(115);
             if ($78) {
              $79 = HEAP8[$67>>0]|0;
              $80 = ($79<<24>>24)==(101);
              $81 = ((($53)) + 6|0);
              $$$i = $80 ? $81 : $53;
              $cur$0$i = $$$i;$result$0$i = 0;
             } else {
              $cur$0$i = $53;$result$0$i = 0;
             }
            } else {
             $cur$0$i = $53;$result$0$i = 0;
            }
           } else {
            $cur$0$i = $53;$result$0$i = 0;
           }
          } else {
           $cur$0$i = $53;$result$0$i = 0;
          }
         } else {
          $cur$0$i = $53;$result$0$i = 0;
         }
        }
        HEAP32[$cur>>2] = $cur$0$i;
        $val$0 = $result$0$i;
        label = 59;
        break;
       }
       case 3:  {
        $82 = (_PS_Conv_ToFixed($cur,$limit$0$ph103,0)|0);
        $val$0 = $82;
        label = 59;
        break;
       }
       case 4:  {
        $83 = (_PS_Conv_ToFixed($cur,$limit$0$ph103,3)|0);
        $val$0 = $83;
        label = 59;
        break;
       }
       case 2:  {
        $84 = ($53>>>0)<($limit$0$ph103>>>0);
        if ($84) {
         $85 = HEAP8[$53>>0]|0;
         switch ($85<<24>>24) {
         case 43: case 45:  {
          $86 = ($85<<24>>24)==(45);
          $87 = $86&1;
          $88 = ((($53)) + 1|0);
          $89 = ($88|0)==($limit$0$ph103|0);
          if ($89) {
           $val$0 = 0;
           label = 59;
           break L24;
          } else {
           $p$0$i21 = $88;$sign$0$i20 = $87;
          }
          break;
         }
         default: {
          $p$0$i21 = $53;$sign$0$i20 = 0;
         }
         }
         $90 = ($p$0$i21>>>0)<($limit$0$ph103>>>0);
         L47: do {
          if ($90) {
           $have_overflow$07$i25 = 0;$num$08$i24 = 0;$p$19$i23 = $p$0$i21;
           while(1) {
            $91 = HEAP8[$p$19$i23>>0]|0;
            $92 = $91&255;
            switch ($91<<24>>24) {
            case 0: case 12: case 9: case 10: case 13: case 32:  {
             $have_overflow$0$lcssa$i32 = $have_overflow$07$i25;$num$0$lcssa$i31 = $num$08$i24;$p$1$lcssa$i30 = $p$19$i23;
             break L47;
             break;
            }
            default: {
            }
            }
            $93 = ($91<<24>>24)<(0);
            if ($93) {
             $have_overflow$0$lcssa$i32 = $have_overflow$07$i25;$num$0$lcssa$i31 = $num$08$i24;$p$1$lcssa$i30 = $p$19$i23;
             break L47;
            }
            $94 = $92 & 127;
            $95 = (532577 + ($94)|0);
            $96 = HEAP8[$95>>0]|0;
            $97 = $96 << 24 >> 24;
            $98 = (($94) + -48)|0;
            $99 = ($98>>>0)<(10);
            if (!($99)) {
             $have_overflow$0$lcssa$i32 = $have_overflow$07$i25;$num$0$lcssa$i31 = $num$08$i24;$p$1$lcssa$i30 = $p$19$i23;
             break L47;
            }
            $100 = ($num$08$i24|0)>(214748364);
            if ($100) {
             $have_overflow$2$ph$i28 = 1;$num$2$ph$i29 = $num$08$i24;
            } else {
             $101 = ($num$08$i24|0)==(214748364);
             $102 = ($96<<24>>24)>(7);
             $or$cond44$i27 = $101 & $102;
             if ($or$cond44$i27) {
              $have_overflow$2$ph$i28 = 1;$num$2$ph$i29 = 214748364;
             } else {
              $103 = ($num$08$i24*10)|0;
              $104 = (($97) + ($103))|0;
              $have_overflow$2$ph$i28 = $have_overflow$07$i25;$num$2$ph$i29 = $104;
             }
            }
            $105 = ((($p$19$i23)) + 1|0);
            $106 = ($105>>>0)<($limit$0$ph103>>>0);
            if ($106) {
             $have_overflow$07$i25 = $have_overflow$2$ph$i28;$num$08$i24 = $num$2$ph$i29;$p$19$i23 = $105;
            } else {
             $have_overflow$0$lcssa$i32 = $have_overflow$2$ph$i28;$num$0$lcssa$i31 = $num$2$ph$i29;$p$1$lcssa$i30 = $105;
             break L47;
            }
           }
          } else {
           $have_overflow$0$lcssa$i32 = 0;$num$0$lcssa$i31 = 0;$p$1$lcssa$i30 = $p$0$i21;
          }
         } while(0);
         $107 = ($have_overflow$0$lcssa$i32<<24>>24)==(0);
         $num$3$$i33 = $107 ? $num$0$lcssa$i31 : 2147483647;
         $108 = ($sign$0$i20<<24>>24)==(0);
         $109 = (0 - ($num$3$$i33))|0;
         $num$5$i34 = $108 ? $num$3$$i33 : $109;
         $110 = ($p$1$lcssa$i30|0)==($53|0);
         $111 = $p$1$lcssa$i30;
         if ($110) {
          $val$0 = 0;
          label = 59;
         } else {
          $112 = ($p$1$lcssa$i30>>>0)<($limit$0$ph103>>>0);
          if ($112) {
           $113 = HEAP8[$p$1$lcssa$i30>>0]|0;
           $114 = ($113<<24>>24)==(35);
           if ($114) {
            $115 = ((($p$1$lcssa$i30)) + 1|0);
            $116 = ($115>>>0)>=($limit$0$ph103>>>0);
            $base$off$i = (($num$5$i34) + -2)|0;
            $117 = ($base$off$i>>>0)>(34);
            $or$cond4$i = $116 | $117;
            if ($or$cond4$i) {
             $val$0 = 0;
             label = 59;
             break L24;
            }
            $118 = HEAP8[$115>>0]|0;
            switch ($118<<24>>24) {
            case 43: case 45:  {
             $119 = ($118<<24>>24)==(45);
             $120 = $119&1;
             $121 = ((($p$1$lcssa$i30)) + 2|0);
             $122 = ($121|0)==($limit$0$ph103|0);
             if ($122) {
              $val$0 = 0;
              label = 59;
              break L24;
             } else {
              $p$0$i = $121;$sign$0$i = $120;
             }
             break;
            }
            default: {
             $p$0$i = $115;$sign$0$i = 0;
            }
            }
            $123 = (2147483647 / ($num$5$i34|0))&-1;
            $124 = ($p$0$i>>>0)<($limit$0$ph103>>>0);
            L66: do {
             if ($124) {
              $125 = (2147483647 % ($num$5$i34|0))&-1;
              $sext$i = $125 << 24;
              $126 = $sext$i >> 24;
              $have_overflow$07$i = 0;$num$08$i = 0;$p$19$i = $p$0$i;
              while(1) {
               $127 = HEAP8[$p$19$i>>0]|0;
               $128 = $127&255;
               switch ($127<<24>>24) {
               case 0: case 12: case 9: case 10: case 13: case 32:  {
                $have_overflow$0$lcssa$i = $have_overflow$07$i;$num$0$lcssa$i = $num$08$i;$p$1$lcssa$i = $p$19$i;
                break L66;
                break;
               }
               default: {
               }
               }
               $129 = ($127<<24>>24)<(0);
               if ($129) {
                $have_overflow$0$lcssa$i = $have_overflow$07$i;$num$0$lcssa$i = $num$08$i;$p$1$lcssa$i = $p$19$i;
                break L66;
               }
               $130 = $128 & 127;
               $131 = (532577 + ($130)|0);
               $132 = HEAP8[$131>>0]|0;
               $133 = $132 << 24 >> 24;
               $134 = ($132<<24>>24)>(-1);
               $135 = ($133|0)<($num$5$i34|0);
               $or$cond$i = $134 & $135;
               if (!($or$cond$i)) {
                $have_overflow$0$lcssa$i = $have_overflow$07$i;$num$0$lcssa$i = $num$08$i;$p$1$lcssa$i = $p$19$i;
                break L66;
               }
               $136 = ($num$08$i|0)>($123|0);
               do {
                if ($136) {
                 $have_overflow$2$ph$i = 1;$num$2$ph$i = $num$08$i;
                } else {
                 $137 = ($num$08$i|0)==($123|0);
                 $138 = ($133|0)>($126|0);
                 $or$cond44$i = $137 & $138;
                 if ($or$cond44$i) {
                  $have_overflow$2$ph$i = 1;$num$2$ph$i = $123;
                  break;
                 }
                 $139 = Math_imul($num$08$i, $num$5$i34)|0;
                 $140 = (($133) + ($139))|0;
                 $have_overflow$2$ph$i = $have_overflow$07$i;$num$2$ph$i = $140;
                }
               } while(0);
               $141 = ((($p$19$i)) + 1|0);
               $142 = ($141>>>0)<($limit$0$ph103>>>0);
               if ($142) {
                $have_overflow$07$i = $have_overflow$2$ph$i;$num$08$i = $num$2$ph$i;$p$19$i = $141;
               } else {
                $have_overflow$0$lcssa$i = $have_overflow$2$ph$i;$num$0$lcssa$i = $num$2$ph$i;$p$1$lcssa$i = $141;
                break L66;
               }
              }
             } else {
              $have_overflow$0$lcssa$i = 0;$num$0$lcssa$i = 0;$p$1$lcssa$i = $p$0$i;
             }
            } while(0);
            $143 = ($have_overflow$0$lcssa$i<<24>>24)==(0);
            $num$3$$i = $143 ? $num$0$lcssa$i : 2147483647;
            $144 = ($sign$0$i<<24>>24)==(0);
            $145 = (0 - ($num$3$$i))|0;
            $num$5$i = $144 ? $num$3$$i : $145;
            $146 = ($p$1$lcssa$i|0)==($115|0);
            $147 = $p$1$lcssa$i;
            if ($146) {
             $val$0 = 0;
             label = 59;
             break L24;
            } else {
             $148 = $147;$num$0$i = $num$5$i;
            }
           } else {
            $148 = $111;$num$0$i = $num$5$i34;
           }
          } else {
           $148 = $111;$num$0$i = $num$5$i34;
          }
          HEAP32[$cur>>2] = $148;
          $val$0 = $num$0$i;
          label = 59;
         }
        } else {
         $val$0 = 0;
         label = 59;
        }
        break;
       }
       case 6: case 5:  {
        $153 = HEAP32[$26>>2]|0;
        $154 = (($27) - ($cur$sroa$0$0$lcssa$i))|0;
        $155 = ($53>>>0)<($limit$0$ph103>>>0);
        if ($155) {
         switch ($1|0) {
         case 4:  {
          $156 = (($154) + -1)|0;
          $157 = ((($53)) + 1|0);
          HEAP32[$cur>>2] = $157;
          $166 = $157;$len$0 = $156;
          break;
         }
         case 2:  {
          $158 = ((($53)) + 1|0);
          HEAP32[$cur>>2] = $158;
          $159 = (($154) + -2)|0;
          $166 = $158;$len$0 = $159;
          break;
         }
         default: {
          label = 68;
          break L11;
         }
         }
         $160 = HEAP32[$39>>2]|0;
         $161 = ($160|0)==(0|0);
         if (!($161)) {
          _ft_mem_free($153,$160);
          HEAP32[$39>>2] = 0;
         }
         $162 = (($len$0) + 1)|0;
         $163 = (_ft_mem_alloc($153,$162,$error)|0);
         $164 = HEAP32[$error>>2]|0;
         $165 = ($164|0)==(0);
         if (!($165)) {
          $283 = $164;
          label = 123;
          break L11;
         }
         _memcpy(($163|0),($166|0),($len$0|0))|0;
         $167 = (($163) + ($len$0)|0);
         HEAP8[$167>>0] = 0;
         HEAP32[$39>>2] = $163;
        }
        break;
       }
       case 7:  {
        HEAP32[$cur$i>>2] = $53;
        $168 = ($53>>>0)<($limit$0$ph103>>>0);
        if (!($168)) {
         $$ph = $cur$sroa$0$0$lcssa$i;
         label = 101;
         break L11;
        }
        $169 = HEAP8[$53>>0]|0;
        $170 = ($169<<24>>24)==(91);
        $171 = ($169<<24>>24)==(123);
        $$$i17 = $171 ? 125 : 0;
        $ender$0$i = $170 ? 93 : $$$i17;
        $172 = ($ender$0$i<<24>>24)==(0);
        L91: do {
         if ($172) {
          $200 = ($53>>>0)<($limit$0$ph103>>>0);
          if (!($200)) {
           $$ph106 = $cur$sroa$0$0$lcssa$i;
           label = 102;
           break L11;
          }
          $201 = $cur$sroa$0$0$lcssa$i;
          $202 = ($201>>>0)<($limit$0$ph103>>>0);
          L94: do {
           if ($202) {
            $204 = $201;$cur$sroa$0$01$i38 = $cur$sroa$0$0$lcssa$i;
            while(1) {
             $203 = HEAP8[$204>>0]|0;
             switch ($203<<24>>24) {
             case 0: case 12: case 9: case 10: case 13: case 32:  {
              $cur$sroa$0$1$i46 = $cur$sroa$0$01$i38;
              break;
             }
             case 37:  {
              $205 = 37;$cur$01$i$i40 = $204;
              L98: while(1) {
               switch ($205<<24>>24) {
               case 10: case 13:  {
                $cur$0$lcssa$i$i44 = $cur$01$i$i40;
                break L98;
                break;
               }
               default: {
               }
               }
               $206 = ((($cur$01$i$i40)) + 1|0);
               $207 = ($206>>>0)<($limit$0$ph103>>>0);
               if (!($207)) {
                $cur$0$lcssa$i$i44 = $206;
                break;
               }
               $$pre$i42 = HEAP8[$206>>0]|0;
               $205 = $$pre$i42;$cur$01$i$i40 = $206;
              }
              $208 = $cur$0$lcssa$i$i44;
              $cur$sroa$0$1$i46 = $208;
              break;
             }
             default: {
              $cur$sroa$0$0$lcssa$i47 = $cur$sroa$0$01$i38;
              break L94;
             }
             }
             $209 = $cur$sroa$0$1$i46;
             $210 = ((($209)) + 1|0);
             $211 = $210;
             $212 = ($210>>>0)<($limit$0$ph103>>>0);
             if ($212) {
              $204 = $210;$cur$sroa$0$01$i38 = $211;
             } else {
              $cur$sroa$0$0$lcssa$i47 = $211;
              break L94;
             }
            }
           } else {
            $cur$sroa$0$0$lcssa$i47 = $cur$sroa$0$0$lcssa$i;
           }
          } while(0);
          HEAP32[$cur$i>>2] = $cur$sroa$0$0$lcssa$i47;
          $$cast = $cur$sroa$0$0$lcssa$i47;
          $213 = ($$cast>>>0)<($limit$0$ph103>>>0);
          if (!($213)) {
           $$ph106 = $cur$sroa$0$0$lcssa$i47;
           label = 102;
           break L11;
          }
          $214 = HEAP8[$$cast>>0]|0;
          $215 = ($214<<24>>24)==(0);
          if ($215) {
           $$cast$lcssa = $$cast;$count$0$us8$i$lcssa69 = 0;
           label = 100;
          } else {
           $$cast$lcssa193 = $$cast;
           label = 99;
           break L11;
          }
         } else {
          $173 = ((($53)) + 1|0);
          HEAP32[$cur$i>>2] = $173;
          $174 = $173;
          $175 = ($173>>>0)<($limit$0$ph103>>>0);
          if ($175) {
           $177 = $174;$count$0$us8$i$us148 = 0;
           while(1) {
            $176 = $177;
            $178 = ($176>>>0)<($limit$0$ph103>>>0);
            L109: do {
             if ($178) {
              $180 = $176;$cur$sroa$0$01$i38$us = $177;
              while(1) {
               $179 = HEAP8[$180>>0]|0;
               switch ($179<<24>>24) {
               case 0: case 12: case 9: case 10: case 13: case 32:  {
                $cur$sroa$0$1$i46$us = $cur$sroa$0$01$i38$us;
                break;
               }
               case 37:  {
                $181 = 37;$cur$01$i$i40$us = $180;
                L113: while(1) {
                 switch ($181<<24>>24) {
                 case 10: case 13:  {
                  $cur$0$lcssa$i$i44$us = $cur$01$i$i40$us;
                  break L113;
                  break;
                 }
                 default: {
                 }
                 }
                 $182 = ((($cur$01$i$i40$us)) + 1|0);
                 $183 = ($182>>>0)<($limit$0$ph103>>>0);
                 if (!($183)) {
                  $cur$0$lcssa$i$i44$us = $182;
                  break;
                 }
                 $$pre$i42$us = HEAP8[$182>>0]|0;
                 $181 = $$pre$i42$us;$cur$01$i$i40$us = $182;
                }
                $184 = $cur$0$lcssa$i$i44$us;
                $cur$sroa$0$1$i46$us = $184;
                break;
               }
               default: {
                $cur$sroa$0$0$lcssa$i47$us = $cur$sroa$0$01$i38$us;
                break L109;
               }
               }
               $185 = $cur$sroa$0$1$i46$us;
               $186 = ((($185)) + 1|0);
               $187 = $186;
               $188 = ($186>>>0)<($limit$0$ph103>>>0);
               if ($188) {
                $180 = $186;$cur$sroa$0$01$i38$us = $187;
               } else {
                $cur$sroa$0$0$lcssa$i47$us = $187;
                break L109;
               }
              }
             } else {
              $cur$sroa$0$0$lcssa$i47$us = $177;
             }
            } while(0);
            HEAP32[$cur$i>>2] = $cur$sroa$0$0$lcssa$i47$us;
            $$cast$us = $cur$sroa$0$0$lcssa$i47$us;
            $189 = ($$cast$us>>>0)<($limit$0$ph103>>>0);
            if (!($189)) {
             $222 = $cur$sroa$0$0$lcssa$i47$us;$count$2$i = $count$0$us8$i$us148;
             label = 103;
             break L91;
            }
            $190 = HEAP8[$$cast$us>>0]|0;
            $191 = ($190<<24>>24)==($ender$0$i<<24>>24);
            if ($191) {
             $$cast$lcssa = $$cast$us;$count$0$us8$i$lcssa69 = $count$0$us8$i$us148;
             label = 100;
             break L91;
            }
            $192 = ($count$0$us8$i$us148|0)<(4);
            if (!($192)) {
             $cur$sroa$0$0$lcssa$i47$us$lcssa179 = $cur$sroa$0$0$lcssa$i47$us;
             break;
            }
            $193 = (_PS_Conv_ToFixed($cur$i,$limit$0$ph103,0)|0);
            $194 = (($temp) + ($count$0$us8$i$us148<<2)|0);
            HEAP32[$194>>2] = $193;
            $195 = HEAP32[$cur$i>>2]|0;
            $196 = ($$cast$us|0)==($195|0);
            $197 = $195;
            if ($196) {
             $$ph = $197;
             label = 101;
             break L11;
            }
            $198 = (($count$0$us8$i$us148) + 1)|0;
            $199 = ($195>>>0)<($limit$0$ph103>>>0);
            if ($199) {
             $177 = $197;$count$0$us8$i$us148 = $198;
            } else {
             $222 = $197;$count$2$i = $198;
             label = 103;
             break L91;
            }
           }
           HEAP32[$cur>>2] = $cur$sroa$0$0$lcssa$i47$us$lcssa179;
          } else {
           $222 = $174;$count$2$i = 0;
           label = 103;
          }
         }
        } while(0);
        if ((label|0) == 100) {
         label = 0;
         $220 = ((($$cast$lcssa)) + 1|0);
         HEAP32[$cur$i>>2] = $220;
         $221 = $220;
         $222 = $221;$count$2$i = $count$0$us8$i$lcssa69;
         label = 103;
        }
        if ((label|0) == 103) {
         label = 0;
         HEAP32[$cur>>2] = $222;
         $223 = ($count$2$i|0)<(4);
         if ($223) {
          break L11;
         }
        }
        $224 = HEAP32[$temp>>2]|0;
        $225 = (_FT_RoundFix($224)|0);
        HEAP32[$39>>2] = $225;
        $226 = HEAP32[$28>>2]|0;
        $227 = (_FT_RoundFix($226)|0);
        $228 = ((($39)) + 4|0);
        HEAP32[$228>>2] = $227;
        $229 = HEAP32[$29>>2]|0;
        $230 = (_FT_RoundFix($229)|0);
        $231 = ((($39)) + 8|0);
        HEAP32[$231>>2] = $230;
        $232 = HEAP32[$30>>2]|0;
        $233 = (_FT_RoundFix($232)|0);
        $234 = ((($39)) + 12|0);
        HEAP32[$234>>2] = $233;
        break;
       }
       case 8:  {
        $235 = HEAP32[$26>>2]|0;
        $236 = (_ft_mem_realloc($235,4,0,$31,0,$error)|0);
        $237 = HEAP32[$error>>2]|0;
        $238 = ($237|0)==(0);
        if (!($238)) {
         $283 = $237;
         label = 123;
         break L11;
        }
        $239 = (_ps_tofixedarray($cur,$limit$0$ph103,$max_objects,$236,0)|0);
        $240 = ($239|0)<(0);
        $241 = ($239>>>0)<($max_objects>>>0);
        $or$cond = $240 | $241;
        if ($or$cond) {
         label = 108;
         break L11;
        }
        $242 = HEAP32[$cur>>2]|0;
        $243 = $242;
        $244 = ($243>>>0)<($limit$0$ph103>>>0);
        L133: do {
         if ($244) {
          $246 = $243;$cur$sroa$0$01$i6 = $242;
          while(1) {
           $245 = HEAP8[$246>>0]|0;
           switch ($245<<24>>24) {
           case 0: case 12: case 9: case 10: case 13: case 32:  {
            $cur$sroa$0$1$i14 = $cur$sroa$0$01$i6;
            break;
           }
           case 37:  {
            $247 = 37;$cur$01$i$i8 = $246;
            L137: while(1) {
             switch ($247<<24>>24) {
             case 10: case 13:  {
              $cur$0$lcssa$i$i12 = $cur$01$i$i8;
              break L137;
              break;
             }
             default: {
             }
             }
             $248 = ((($cur$01$i$i8)) + 1|0);
             $249 = ($248>>>0)<($limit$0$ph103>>>0);
             if (!($249)) {
              $cur$0$lcssa$i$i12 = $248;
              break;
             }
             $$pre$i10 = HEAP8[$248>>0]|0;
             $247 = $$pre$i10;$cur$01$i$i8 = $248;
            }
            $250 = $cur$0$lcssa$i$i12;
            $cur$sroa$0$1$i14 = $250;
            break;
           }
           default: {
            $cur$sroa$0$0$lcssa$i15 = $cur$sroa$0$01$i6;
            break L133;
           }
           }
           $251 = $cur$sroa$0$1$i14;
           $252 = ((($251)) + 1|0);
           $253 = $252;
           $254 = ($252>>>0)<($limit$0$ph103>>>0);
           if ($254) {
            $246 = $252;$cur$sroa$0$01$i6 = $253;
           } else {
            $cur$sroa$0$0$lcssa$i15 = $253;
            break L133;
           }
          }
         } else {
          $cur$sroa$0$0$lcssa$i15 = $242;
         }
        } while(0);
        HEAP32[$cur>>2] = $cur$sroa$0$0$lcssa$i15;
        $255 = (($236) + ($max_objects<<2)|0);
        $256 = (_ps_tofixedarray($cur,$limit$0$ph103,$max_objects,$255,0)|0);
        $257 = ($256|0)<(0);
        $258 = ($256>>>0)<($max_objects>>>0);
        $or$cond$1 = $257 | $258;
        if ($or$cond$1) {
         label = 108;
         break L11;
        }
        $284 = HEAP32[$cur>>2]|0;
        $285 = $284;
        $286 = ($285>>>0)<($limit$0$ph103>>>0);
        L145: do {
         if ($286) {
          $288 = $285;$cur$sroa$0$01$i6$1 = $284;
          while(1) {
           $287 = HEAP8[$288>>0]|0;
           switch ($287<<24>>24) {
           case 0: case 12: case 9: case 10: case 13: case 32:  {
            $cur$sroa$0$1$i14$1 = $cur$sroa$0$01$i6$1;
            break;
           }
           case 37:  {
            $289 = 37;$cur$01$i$i8$1 = $288;
            L149: while(1) {
             switch ($289<<24>>24) {
             case 10: case 13:  {
              $cur$0$lcssa$i$i12$1 = $cur$01$i$i8$1;
              break L149;
              break;
             }
             default: {
             }
             }
             $290 = ((($cur$01$i$i8$1)) + 1|0);
             $291 = ($290>>>0)<($limit$0$ph103>>>0);
             if (!($291)) {
              $cur$0$lcssa$i$i12$1 = $290;
              break;
             }
             $$pre$i10$1 = HEAP8[$290>>0]|0;
             $289 = $$pre$i10$1;$cur$01$i$i8$1 = $290;
            }
            $292 = $cur$0$lcssa$i$i12$1;
            $cur$sroa$0$1$i14$1 = $292;
            break;
           }
           default: {
            $cur$sroa$0$0$lcssa$i15$1 = $cur$sroa$0$01$i6$1;
            break L145;
           }
           }
           $293 = $cur$sroa$0$1$i14$1;
           $294 = ((($293)) + 1|0);
           $295 = $294;
           $296 = ($294>>>0)<($limit$0$ph103>>>0);
           if ($296) {
            $288 = $294;$cur$sroa$0$01$i6$1 = $295;
           } else {
            $cur$sroa$0$0$lcssa$i15$1 = $295;
            break L145;
           }
          }
         } else {
          $cur$sroa$0$0$lcssa$i15$1 = $284;
         }
        } while(0);
        HEAP32[$cur>>2] = $cur$sroa$0$0$lcssa$i15$1;
        $297 = (($236) + ($35<<2)|0);
        $298 = (_ps_tofixedarray($cur,$limit$0$ph103,$max_objects,$297,0)|0);
        $299 = ($298|0)<(0);
        $300 = ($298>>>0)<($max_objects>>>0);
        $or$cond$2 = $299 | $300;
        if ($or$cond$2) {
         label = 108;
         break L11;
        }
        $301 = HEAP32[$cur>>2]|0;
        $302 = $301;
        $303 = ($302>>>0)<($limit$0$ph103>>>0);
        L157: do {
         if ($303) {
          $305 = $302;$cur$sroa$0$01$i6$2 = $301;
          while(1) {
           $304 = HEAP8[$305>>0]|0;
           switch ($304<<24>>24) {
           case 0: case 12: case 9: case 10: case 13: case 32:  {
            $cur$sroa$0$1$i14$2 = $cur$sroa$0$01$i6$2;
            break;
           }
           case 37:  {
            $306 = 37;$cur$01$i$i8$2 = $305;
            L161: while(1) {
             switch ($306<<24>>24) {
             case 10: case 13:  {
              $cur$0$lcssa$i$i12$2 = $cur$01$i$i8$2;
              break L161;
              break;
             }
             default: {
             }
             }
             $307 = ((($cur$01$i$i8$2)) + 1|0);
             $308 = ($307>>>0)<($limit$0$ph103>>>0);
             if (!($308)) {
              $cur$0$lcssa$i$i12$2 = $307;
              break;
             }
             $$pre$i10$2 = HEAP8[$307>>0]|0;
             $306 = $$pre$i10$2;$cur$01$i$i8$2 = $307;
            }
            $309 = $cur$0$lcssa$i$i12$2;
            $cur$sroa$0$1$i14$2 = $309;
            break;
           }
           default: {
            $cur$sroa$0$0$lcssa$i15$2 = $cur$sroa$0$01$i6$2;
            break L157;
           }
           }
           $310 = $cur$sroa$0$1$i14$2;
           $311 = ((($310)) + 1|0);
           $312 = $311;
           $313 = ($311>>>0)<($limit$0$ph103>>>0);
           if ($313) {
            $305 = $311;$cur$sroa$0$01$i6$2 = $312;
           } else {
            $cur$sroa$0$0$lcssa$i15$2 = $312;
            break L157;
           }
          }
         } else {
          $cur$sroa$0$0$lcssa$i15$2 = $301;
         }
        } while(0);
        HEAP32[$cur>>2] = $cur$sroa$0$0$lcssa$i15$2;
        $314 = (($236) + ($34<<2)|0);
        $315 = (_ps_tofixedarray($cur,$limit$0$ph103,$max_objects,$314,0)|0);
        $316 = ($315|0)<(0);
        $317 = ($315>>>0)<($max_objects>>>0);
        $or$cond$3 = $316 | $317;
        if ($or$cond$3) {
         label = 108;
         break L11;
        }
        $318 = HEAP32[$cur>>2]|0;
        $319 = $318;
        $320 = ($319>>>0)<($limit$0$ph103>>>0);
        L169: do {
         if ($320) {
          $322 = $319;$cur$sroa$0$01$i6$3 = $318;
          while(1) {
           $321 = HEAP8[$322>>0]|0;
           switch ($321<<24>>24) {
           case 0: case 12: case 9: case 10: case 13: case 32:  {
            $cur$sroa$0$1$i14$3 = $cur$sroa$0$01$i6$3;
            break;
           }
           case 37:  {
            $323 = 37;$cur$01$i$i8$3 = $322;
            L173: while(1) {
             switch ($323<<24>>24) {
             case 10: case 13:  {
              $cur$0$lcssa$i$i12$3 = $cur$01$i$i8$3;
              break L173;
              break;
             }
             default: {
             }
             }
             $324 = ((($cur$01$i$i8$3)) + 1|0);
             $325 = ($324>>>0)<($limit$0$ph103>>>0);
             if (!($325)) {
              $cur$0$lcssa$i$i12$3 = $324;
              break;
             }
             $$pre$i10$3 = HEAP8[$324>>0]|0;
             $323 = $$pre$i10$3;$cur$01$i$i8$3 = $324;
            }
            $326 = $cur$0$lcssa$i$i12$3;
            $cur$sroa$0$1$i14$3 = $326;
            break;
           }
           default: {
            $cur$sroa$0$0$lcssa$i15$3 = $cur$sroa$0$01$i6$3;
            break L169;
           }
           }
           $327 = $cur$sroa$0$1$i14$3;
           $328 = ((($327)) + 1|0);
           $329 = $328;
           $330 = ($328>>>0)<($limit$0$ph103>>>0);
           if ($330) {
            $322 = $328;$cur$sroa$0$01$i6$3 = $329;
           } else {
            $cur$sroa$0$0$lcssa$i15$3 = $329;
            break L169;
           }
          }
         } else {
          $cur$sroa$0$0$lcssa$i15$3 = $318;
         }
        } while(0);
        HEAP32[$cur>>2] = $cur$sroa$0$0$lcssa$i15$3;
        if (!($32)) {
         $i$186 = 0;
         while(1) {
          $259 = (($objects) + ($i$186<<2)|0);
          $260 = HEAP32[$259>>2]|0;
          $261 = (($236) + ($i$186<<2)|0);
          $262 = HEAP32[$261>>2]|0;
          $263 = (_FT_RoundFix($262)|0);
          HEAP32[$260>>2] = $263;
          $264 = (($i$186) + ($max_objects))|0;
          $265 = (($236) + ($264<<2)|0);
          $266 = HEAP32[$265>>2]|0;
          $267 = (_FT_RoundFix($266)|0);
          $268 = ((($260)) + 4|0);
          HEAP32[$268>>2] = $267;
          $269 = (($i$186) + ($33))|0;
          $270 = (($236) + ($269<<2)|0);
          $271 = HEAP32[$270>>2]|0;
          $272 = (_FT_RoundFix($271)|0);
          $273 = ((($260)) + 8|0);
          HEAP32[$273>>2] = $272;
          $274 = (($i$186) + ($34))|0;
          $275 = (($236) + ($274<<2)|0);
          $276 = HEAP32[$275>>2]|0;
          $277 = (_FT_RoundFix($276)|0);
          $278 = ((($260)) + 12|0);
          HEAP32[$278>>2] = $277;
          $279 = (($i$186) + 1)|0;
          $exitcond = ($279|0)==($max_objects|0);
          if ($exitcond) {
           break;
          } else {
           $i$186 = $279;
          }
         }
        }
        _ft_mem_free($235,$236);
        break;
       }
       default: {
        break L1;
       }
       }
      } while(0);
      L184: do {
       if ((label|0) == 59) {
        label = 0;
        $149 = HEAP8[$25>>0]|0;
        $150 = $149&255;
        switch ($150|0) {
        case 1:  {
         $151 = $val$0&255;
         HEAP8[$39>>0] = $151;
         break L184;
         break;
        }
        case 2:  {
         $152 = $val$0&65535;
         HEAP16[$39>>1] = $152;
         break L184;
         break;
        }
        case 4:  {
         HEAP32[$39>>2] = $val$0;
         break L184;
         break;
        }
        default: {
         HEAP32[$39>>2] = $val$0;
         break L184;
        }
        }
       }
      } while(0);
      $280 = (($count$188) + -1)|0;
      $281 = ($280|0)==(0);
      if ($281) {
       break L9;
      }
      $282 = (($idx$087) + 1)|0;
      $$pre = HEAP32[$cur>>2]|0;
      $41 = $$pre;$count$188 = $280;$idx$087 = $282;
     }
     if ((label|0) == 68) {
      HEAP32[$error>>2] = 3;
      $283 = 3;
      STACKTOP = sp;return ($283|0);
     }
     else if ((label|0) == 99) {
      $216 = (_PS_Conv_ToFixed($cur$i,$limit$0$ph103,0)|0);
      HEAP32[$temp>>2] = $216;
      $217 = HEAP32[$cur$i>>2]|0;
      $218 = ($$cast$lcssa193|0)==($217|0);
      $219 = $217;
      if ($218) {
       $$ph = $219;
       label = 101;
      } else {
       $$ph106 = $219;
       label = 102;
      }
     }
     else if ((label|0) == 108) {
      HEAP32[$error>>2] = 3;
      $283 = 3;
      STACKTOP = sp;return ($283|0);
     }
     else if ((label|0) == 123) {
      STACKTOP = sp;return ($283|0);
     }
     if ((label|0) == 101) {
      HEAP32[$cur>>2] = $$ph;
     }
     else if ((label|0) == 102) {
      HEAP32[$cur>>2] = $$ph106;
     }
     HEAP32[$error>>2] = 3;
     $283 = 3;
     STACKTOP = sp;return ($283|0);
    }
   } while(0);
   HEAP32[$error>>2] = 0;
   $283 = 0;
   STACKTOP = sp;return ($283|0);
  }
 } while(0);
 HEAP32[$error>>2] = 3;
 $283 = 3;
 STACKTOP = sp;return ($283|0);
}
function _ps_parser_load_field_table($parser,$field,$objects,$max_objects,$pflags) {
 $parser = $parser|0;
 $field = $field|0;
 $objects = $objects|0;
 $max_objects = $max_objects|0;
 $pflags = $pflags|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cur$02$i = 0, $cur$03$us$i = 0, $cur$03$us$i$lcssa = 0, $elements = 0;
 var $error$1 = 0, $error$2 = 0, $fieldrec = 0, $master$i = 0, $num_elements$22 = 0, $token$03 = 0, $token$i = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 448|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $master$i = sp + 436|0;
 $token$i = sp + 424|0;
 $elements = sp + 40|0;
 $fieldrec = sp;
 dest=$fieldrec; src=$field; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 $0 = ((($fieldrec)) + 8|0);
 HEAP32[$0>>2] = 2;
 $1 = ((($field)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 switch ($2|0) {
 case 7: case 10:  {
  HEAP32[$0>>2] = 3;
  break;
 }
 default: {
 }
 }
 _ps_parser_to_token($parser,$master$i);
 $3 = ((($master$i)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(3);
 if (!($5)) {
  $error$2 = 162;
  STACKTOP = sp;return ($error$2|0);
 }
 $6 = HEAP32[$parser>>2]|0;
 $7 = ((($parser)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($elements)) + 384|0);
 $10 = HEAP32[$master$i>>2]|0;
 $11 = ((($10)) + 1|0);
 HEAP32[$parser>>2] = $11;
 $12 = ((($master$i)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($13)) + -1|0);
 HEAP32[$7>>2] = $14;
 $15 = ($11>>>0)<($14>>>0);
 L8: do {
  if ($15) {
   $16 = ((($token$i)) + 8|0);
   $cur$03$us$i = $elements;
   while(1) {
    _ps_parser_to_token($parser,$token$i);
    $17 = HEAP32[$16>>2]|0;
    $18 = ($17|0)==(0);
    if ($18) {
     $cur$03$us$i$lcssa = $cur$03$us$i;
     break;
    }
    $19 = ($cur$03$us$i>>>0)<($9>>>0);
    if ($19) {
     ;HEAP32[$cur$03$us$i>>2]=HEAP32[$token$i>>2]|0;HEAP32[$cur$03$us$i+4>>2]=HEAP32[$token$i+4>>2]|0;HEAP32[$cur$03$us$i+8>>2]=HEAP32[$token$i+8>>2]|0;
    }
    $20 = ((($cur$03$us$i)) + 12|0);
    $21 = HEAP32[$parser>>2]|0;
    $22 = HEAP32[$7>>2]|0;
    $23 = ($21>>>0)<($22>>>0);
    if ($23) {
     $cur$03$us$i = $20;
    } else {
     $cur$02$i = $20;
     break L8;
    }
   }
   $cur$02$i = $cur$03$us$i$lcssa;
  } else {
   $cur$02$i = $elements;
  }
 } while(0);
 $24 = $cur$02$i;
 $25 = $elements;
 $26 = (($24) - ($25))|0;
 $27 = (($26|0) / 12)&-1;
 HEAP32[$parser>>2] = $6;
 HEAP32[$7>>2] = $8;
 $28 = ($26|0)<(0);
 if ($28) {
  $error$2 = 162;
  STACKTOP = sp;return ($error$2|0);
 }
 $29 = ((($field)) + 24|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ($27>>>0)>($30>>>0);
 $$ = $31 ? $30 : $27;
 $32 = HEAP32[$1>>2]|0;
 $33 = ($32|0)==(7);
 if (!($33)) {
  $34 = ((($field)) + 28|0);
  $35 = HEAP32[$34>>2]|0;
  $36 = ($35|0)==(0);
  if (!($36)) {
   $37 = $$&255;
   $38 = HEAP32[$objects>>2]|0;
   $39 = (($38) + ($35)|0);
   HEAP8[$39>>0] = $37;
  }
 }
 $40 = ($$|0)>(0);
 L25: do {
  if ($40) {
   $41 = ((($fieldrec)) + 20|0);
   $42 = ((($fieldrec)) + 16|0);
   $num_elements$22 = $$;$token$03 = $elements;
   while(1) {
    $43 = HEAP32[$token$03>>2]|0;
    HEAP32[$parser>>2] = $43;
    $44 = ((($token$03)) + 4|0);
    $45 = HEAP32[$44>>2]|0;
    HEAP32[$7>>2] = $45;
    $46 = (_ps_parser_load_field($parser,$fieldrec,$objects,$max_objects,0)|0);
    $47 = ($46|0)==(0);
    if (!($47)) {
     $error$1 = $46;
     break L25;
    }
    $48 = HEAP8[$41>>0]|0;
    $49 = $48&255;
    $50 = HEAP32[$42>>2]|0;
    $51 = (($50) + ($49))|0;
    HEAP32[$42>>2] = $51;
    $52 = (($num_elements$22) + -1)|0;
    $53 = ((($token$03)) + 12|0);
    $54 = ($num_elements$22|0)>(1);
    if ($54) {
     $num_elements$22 = $52;$token$03 = $53;
    } else {
     $error$1 = 0;
     break;
    }
   }
  } else {
   $error$1 = 0;
  }
 } while(0);
 HEAP32[$parser>>2] = $6;
 HEAP32[$7>>2] = $8;
 $error$2 = $error$1;
 STACKTOP = sp;return ($error$2|0);
}
function _t1_builder_init($builder,$face,$size,$glyph,$hinting) {
 $builder = $builder|0;
 $face = $face|0;
 $size = $size|0;
 $glyph = $glyph|0;
 $hinting = $hinting|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($builder)) + 64|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($builder)) + 68|0);
 HEAP8[$1>>0] = 1;
 $2 = ((($builder)) + 4|0);
 HEAP32[$2>>2] = $face;
 $3 = ((($builder)) + 8|0);
 HEAP32[$3>>2] = $glyph;
 $4 = ((($face)) + 100|0);
 $5 = HEAP32[$4>>2]|0;
 HEAP32[$builder>>2] = $5;
 $6 = ($glyph|0)==(0|0);
 if (!($6)) {
  $7 = ((($glyph)) + 156|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($builder)) + 12|0);
  HEAP32[$10>>2] = $9;
  $11 = ((($9)) + 20|0);
  $12 = ((($builder)) + 16|0);
  HEAP32[$12>>2] = $11;
  $13 = ((($9)) + 56|0);
  $14 = ((($builder)) + 20|0);
  HEAP32[$14>>2] = $13;
  _FT_GlyphLoader_Rewind($9);
  $15 = ((($size)) + 40|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ((($builder)) + 76|0);
  HEAP32[$17>>2] = $16;
  $18 = ((($builder)) + 72|0);
  HEAP32[$18>>2] = 0;
  $19 = ($hinting<<24>>24)==(0);
  if (!($19)) {
   $20 = HEAP32[$7>>2]|0;
   $21 = ((($20)) + 36|0);
   $22 = HEAP32[$21>>2]|0;
   HEAP32[$18>>2] = $22;
  }
 }
 $23 = ((($builder)) + 24|0);
 $24 = ((($builder)) + 80|0);
 ;HEAP32[$23>>2]=0|0;HEAP32[$23+4>>2]=0|0;HEAP32[$23+8>>2]=0|0;HEAP32[$23+12>>2]=0|0;HEAP32[$23+16>>2]=0|0;HEAP32[$23+20>>2]=0|0;
 ;HEAP32[$24>>2]=HEAP32[53576>>2]|0;HEAP32[$24+4>>2]=HEAP32[53576+4>>2]|0;HEAP32[$24+8>>2]=HEAP32[53576+8>>2]|0;HEAP32[$24+12>>2]=HEAP32[53576+12>>2]|0;HEAP32[$24+16>>2]=HEAP32[53576+16>>2]|0;HEAP32[$24+20>>2]=HEAP32[53576+20>>2]|0;HEAP32[$24+24>>2]=HEAP32[53576+24>>2]|0;HEAP32[$24+28>>2]=HEAP32[53576+28>>2]|0;
 return;
}
function _t1_builder_done($builder) {
 $builder = $builder|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($builder)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 $3 = ((($1)) + 108|0);
 $4 = ((($builder)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 ;HEAP32[$3>>2]=HEAP32[$5>>2]|0;HEAP32[$3+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$3+8>>2]=HEAP32[$5+8>>2]|0;HEAP32[$3+12>>2]=HEAP32[$5+12>>2]|0;HEAP32[$3+16>>2]=HEAP32[$5+16>>2]|0;
 return;
}
function _t1_builder_check_points($builder,$count) {
 $builder = $builder|0;
 $count = $count|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($count|0)==(0);
 if ($0) {
  $15 = 0;
  return ($15|0);
 }
 $1 = ((($builder)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 22|0);
 $4 = HEAP16[$3>>1]|0;
 $5 = $4 << 16 >> 16;
 $6 = ((($2)) + 58|0);
 $7 = HEAP16[$6>>1]|0;
 $8 = $7 << 16 >> 16;
 $9 = (($5) + ($count))|0;
 $10 = (($9) + ($8))|0;
 $11 = ((($2)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($10>>>0)>($12>>>0);
 if (!($13)) {
  $15 = 0;
  return ($15|0);
 }
 $14 = (_FT_GlyphLoader_CheckPoints($2,$count,0)|0);
 $15 = $14;
 return ($15|0);
}
function _t1_builder_add_point($builder,$x,$y,$flag) {
 $builder = $builder|0;
 $x = $x|0;
 $y = $y|0;
 $flag = $flag|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($builder)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($builder)) + 68|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = ($3<<24>>24)==(0);
 if ($4) {
  $$pre = ((($1)) + 2|0);
  $$pre$phiZ2D = $$pre;
  $21 = HEAP16[$$pre$phiZ2D>>1]|0;
  $22 = (($21) + 1)<<16>>16;
  HEAP16[$$pre$phiZ2D>>1] = $22;
  return;
 } else {
  $5 = ((($1)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ((($1)) + 2|0);
  $8 = HEAP16[$7>>1]|0;
  $9 = $8 << 16 >> 16;
  $10 = ((($1)) + 8|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + ($9)|0);
  $13 = (_FT_RoundFix($x)|0);
  $14 = $13 >> 16;
  $15 = (($6) + ($9<<3)|0);
  HEAP32[$15>>2] = $14;
  $16 = (_FT_RoundFix($y)|0);
  $17 = $16 >> 16;
  $18 = (((($6) + ($9<<3)|0)) + 4|0);
  HEAP32[$18>>2] = $17;
  $19 = ($flag<<24>>24)!=(0);
  $20 = $19 ? 1 : 2;
  HEAP8[$12>>0] = $20;
  $$pre$phiZ2D = $7;
  $21 = HEAP16[$$pre$phiZ2D>>1]|0;
  $22 = (($21) + 1)<<16>>16;
  HEAP16[$$pre$phiZ2D>>1] = $22;
  return;
 }
}
function _t1_builder_add_point1($builder,$x,$y) {
 $builder = $builder|0;
 $x = $x|0;
 $y = $y|0;
 var $$pre$i = 0, $$pre$phi$iZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($builder)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 22|0);
 $3 = HEAP16[$2>>1]|0;
 $4 = $3 << 16 >> 16;
 $5 = ((($1)) + 58|0);
 $6 = HEAP16[$5>>1]|0;
 $7 = $6 << 16 >> 16;
 $8 = (($4) + 1)|0;
 $9 = (($8) + ($7))|0;
 $10 = ((($1)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($9>>>0)>($11>>>0);
 if ($12) {
  $13 = (_FT_GlyphLoader_CheckPoints($1,1,0)|0);
  $14 = ($13|0)==(0);
  if (!($14)) {
   $36 = $13;
   return ($36|0);
  }
 }
 $15 = ((($builder)) + 20|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($builder)) + 68|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = ($18<<24>>24)==(0);
 if ($19) {
  $$pre$i = ((($16)) + 2|0);
  $$pre$phi$iZ2D = $$pre$i;
 } else {
  $20 = ((($16)) + 4|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = ((($16)) + 2|0);
  $23 = HEAP16[$22>>1]|0;
  $24 = $23 << 16 >> 16;
  $25 = ((($16)) + 8|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = (($26) + ($24)|0);
  $28 = (_FT_RoundFix($x)|0);
  $29 = $28 >> 16;
  $30 = (($21) + ($24<<3)|0);
  HEAP32[$30>>2] = $29;
  $31 = (_FT_RoundFix($y)|0);
  $32 = $31 >> 16;
  $33 = (((($21) + ($24<<3)|0)) + 4|0);
  HEAP32[$33>>2] = $32;
  HEAP8[$27>>0] = 1;
  $$pre$phi$iZ2D = $22;
 }
 $34 = HEAP16[$$pre$phi$iZ2D>>1]|0;
 $35 = (($34) + 1)<<16>>16;
 HEAP16[$$pre$phi$iZ2D>>1] = $35;
 $36 = 0;
 return ($36|0);
}
function _t1_builder_add_contour($builder) {
 $builder = $builder|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $sext = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($builder)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = 3;
  return ($$0|0);
 }
 $3 = ((($builder)) + 68|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)==(0);
 if ($5) {
  $6 = HEAP16[$1>>1]|0;
  $7 = (($6) + 1)<<16>>16;
  HEAP16[$1>>1] = $7;
  $$0 = 0;
  return ($$0|0);
 }
 $8 = ((($builder)) + 12|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 20|0);
 $11 = HEAP16[$10>>1]|0;
 $12 = $11 << 16 >> 16;
 $13 = ((($9)) + 56|0);
 $14 = HEAP16[$13>>1]|0;
 $15 = $14 << 16 >> 16;
 $16 = (($12) + 1)|0;
 $17 = (($16) + ($15))|0;
 $18 = ((($9)) + 8|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($17>>>0)>($19>>>0);
 if ($20) {
  $21 = (_FT_GlyphLoader_CheckPoints($9,0,1)|0);
  $22 = ($21|0)==(0);
  if (!($22)) {
   $$0 = $21;
   return ($$0|0);
  }
 }
 $23 = HEAP32[$1>>2]|0;
 $24 = $23&65535;
 $25 = ($24<<16>>16)>(0);
 if ($25) {
  $26 = $23 >>> 16;
  $sext = $23 << 16;
  $27 = $sext >> 16;
  $28 = (($26) + 65535)|0;
  $29 = $28&65535;
  $30 = (($27) + -1)|0;
  $31 = ((($1)) + 12|0);
  $32 = HEAP32[$31>>2]|0;
  $33 = (($32) + ($30<<1)|0);
  HEAP16[$33>>1] = $29;
  $$pre = HEAP16[$1>>1]|0;
  $35 = $$pre;
 } else {
  $35 = $24;
 }
 $34 = (($35) + 1)<<16>>16;
 HEAP16[$1>>1] = $34;
 $$0 = 0;
 return ($$0|0);
}
function _t1_builder_start_point($builder,$x,$y) {
 $builder = $builder|0;
 $x = $x|0;
 $y = $y|0;
 var $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$phi$i$iZ2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $error$0 = 0, $sext$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($builder)) + 64|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(3);
 if ($2) {
  $error$0 = 0;
  return ($error$0|0);
 }
 HEAP32[$0>>2] = 3;
 $3 = ((($builder)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if ($5) {
  $error$0 = 3;
  return ($error$0|0);
 }
 $6 = ((($builder)) + 68|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = ($7<<24>>24)==(0);
 if ($8) {
  $9 = HEAP16[$4>>1]|0;
  $10 = (($9) + 1)<<16>>16;
  HEAP16[$4>>1] = $10;
  $$pre = ((($builder)) + 12|0);
  $$pre$phiZ2D = $$pre;
 } else {
  $11 = ((($builder)) + 12|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ((($12)) + 20|0);
  $14 = HEAP16[$13>>1]|0;
  $15 = $14 << 16 >> 16;
  $16 = ((($12)) + 56|0);
  $17 = HEAP16[$16>>1]|0;
  $18 = $17 << 16 >> 16;
  $19 = (($15) + 1)|0;
  $20 = (($19) + ($18))|0;
  $21 = ((($12)) + 8|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = ($20>>>0)>($22>>>0);
  if ($23) {
   $24 = (_FT_GlyphLoader_CheckPoints($12,0,1)|0);
   $25 = ($24|0)==(0);
   if (!($25)) {
    $error$0 = $24;
    return ($error$0|0);
   }
  }
  $26 = HEAP32[$4>>2]|0;
  $27 = $26&65535;
  $28 = ($27<<16>>16)>(0);
  if ($28) {
   $29 = $26 >>> 16;
   $sext$i = $26 << 16;
   $30 = $sext$i >> 16;
   $31 = (($29) + 65535)|0;
   $32 = $31&65535;
   $33 = (($30) + -1)|0;
   $34 = ((($4)) + 12|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = (($35) + ($33<<1)|0);
   HEAP16[$36>>1] = $32;
   $$pre$i = HEAP16[$4>>1]|0;
   $38 = $$pre$i;
  } else {
   $38 = $27;
  }
  $37 = (($38) + 1)<<16>>16;
  HEAP16[$4>>1] = $37;
  $$pre$phiZ2D = $11;
 }
 $39 = HEAP32[$$pre$phiZ2D>>2]|0;
 $40 = ((($39)) + 22|0);
 $41 = HEAP16[$40>>1]|0;
 $42 = $41 << 16 >> 16;
 $43 = ((($39)) + 58|0);
 $44 = HEAP16[$43>>1]|0;
 $45 = $44 << 16 >> 16;
 $46 = (($42) + 1)|0;
 $47 = (($46) + ($45))|0;
 $48 = ((($39)) + 4|0);
 $49 = HEAP32[$48>>2]|0;
 $50 = ($47>>>0)>($49>>>0);
 if ($50) {
  $51 = (_FT_GlyphLoader_CheckPoints($39,1,0)|0);
  $52 = ($51|0)==(0);
  if (!($52)) {
   $error$0 = $51;
   return ($error$0|0);
  }
 }
 $53 = HEAP32[$3>>2]|0;
 $54 = HEAP8[$6>>0]|0;
 $55 = ($54<<24>>24)==(0);
 if ($55) {
  $$pre$i$i = ((($53)) + 2|0);
  $$pre$phi$i$iZ2D = $$pre$i$i;
 } else {
  $56 = ((($53)) + 4|0);
  $57 = HEAP32[$56>>2]|0;
  $58 = ((($53)) + 2|0);
  $59 = HEAP16[$58>>1]|0;
  $60 = $59 << 16 >> 16;
  $61 = ((($53)) + 8|0);
  $62 = HEAP32[$61>>2]|0;
  $63 = (($62) + ($60)|0);
  $64 = (_FT_RoundFix($x)|0);
  $65 = $64 >> 16;
  $66 = (($57) + ($60<<3)|0);
  HEAP32[$66>>2] = $65;
  $67 = (_FT_RoundFix($y)|0);
  $68 = $67 >> 16;
  $69 = (((($57) + ($60<<3)|0)) + 4|0);
  HEAP32[$69>>2] = $68;
  HEAP8[$63>>0] = 1;
  $$pre$phi$i$iZ2D = $58;
 }
 $70 = HEAP16[$$pre$phi$i$iZ2D>>1]|0;
 $71 = (($70) + 1)<<16>>16;
 HEAP16[$$pre$phi$i$iZ2D>>1] = $71;
 $error$0 = 0;
 return ($error$0|0);
}
function _t1_builder_close_contour($builder) {
 $builder = $builder|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $sext = 0, $sext2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($builder)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 $3 = HEAP32[$1>>2]|0;
 $4 = $3&65535;
 $5 = ($4<<16>>16)<(2);
 $6 = $3 >>> 16;
 $7 = $6&65535;
 if ($5) {
  $28 = 0;
 } else {
  $sext = $3 << 16;
  $8 = $sext >> 16;
  $9 = (($8) + -2)|0;
  $10 = ((($1)) + 12|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + ($9<<1)|0);
  $13 = HEAP16[$12>>1]|0;
  $14 = $13 << 16 >> 16;
  $15 = (($14) + 1)|0;
  $28 = $15;
 }
 $16 = ((($1)) + 2|0);
 $17 = $3 >> 16;
 $18 = ($7<<16>>16)>(1);
 if ($18) {
  $19 = ((($1)) + 4|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = (($20) + ($17<<3)|0);
  $22 = ((($21)) + -8|0);
  $23 = ((($1)) + 8|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = (($24) + ($17)|0);
  $26 = ((($25)) + -1|0);
  $27 = (($20) + ($28<<3)|0);
  $29 = HEAP32[$27>>2]|0;
  $30 = HEAP32[$22>>2]|0;
  $31 = ($29|0)==($30|0);
  if ($31) {
   $32 = (((($20) + ($28<<3)|0)) + 4|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ((($22)) + 4|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = ($33|0)==($35|0);
   if ($36) {
    $37 = HEAP8[$26>>0]|0;
    $38 = ($37<<24>>24)==(1);
    if ($38) {
     $39 = (($7) + -1)<<16>>16;
     HEAP16[$16>>1] = $39;
     $43 = $39;
    } else {
     $43 = $7;
    }
   } else {
    $43 = $7;
   }
  } else {
   $43 = $7;
  }
 } else {
  $43 = $7;
 }
 $sext2 = $3 << 16;
 $40 = $sext2 >> 16;
 $41 = ($4<<16>>16)>(0);
 if (!($41)) {
  return;
 }
 $42 = $43 << 16 >> 16;
 $44 = (($42) + -1)|0;
 $45 = ($28|0)==($44|0);
 if ($45) {
  $46 = (($4) + -1)<<16>>16;
  HEAP16[$1>>1] = $46;
  $47 = (($43) + -1)<<16>>16;
  HEAP16[$16>>1] = $47;
  return;
 } else {
  $48 = $44&65535;
  $49 = (($40) + -1)|0;
  $50 = ((($1)) + 12|0);
  $51 = HEAP32[$50>>2]|0;
  $52 = (($51) + ($49<<1)|0);
  HEAP16[$52>>1] = $48;
  return;
 }
}
function _t1_decoder_init($decoder,$face,$size,$slot,$glyph_names,$blend,$hinting,$hint_mode,$parse_callback) {
 $decoder = $decoder|0;
 $face = $face|0;
 $size = $size|0;
 $slot = $slot|0;
 $glyph_names = $glyph_names|0;
 $blend = $blend|0;
 $hinting = $hinting|0;
 $hint_mode = $hint_mode|0;
 $parse_callback = $parse_callback|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 _memset(($decoder|0),0,1500)|0;
 $0 = ((($face)) + 96|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_ft_module_get_service($1,594361)|0);
 $3 = ($2|0)==(0|0);
 if ($3) {
  $$1 = 7;
  return ($$1|0);
 }
 $4 = ((($decoder)) + 1348|0);
 HEAP32[$4>>2] = $2;
 $5 = ((($decoder)) + 64|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($decoder)) + 68|0);
 HEAP8[$6>>0] = 1;
 $7 = ((($decoder)) + 4|0);
 HEAP32[$7>>2] = $face;
 $8 = ((($decoder)) + 8|0);
 HEAP32[$8>>2] = $slot;
 $9 = ((($face)) + 100|0);
 $10 = HEAP32[$9>>2]|0;
 HEAP32[$decoder>>2] = $10;
 $11 = ($slot|0)==(0|0);
 if (!($11)) {
  $12 = ((($slot)) + 156|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = HEAP32[$13>>2]|0;
  $15 = ((($decoder)) + 12|0);
  HEAP32[$15>>2] = $14;
  $16 = ((($14)) + 20|0);
  $17 = ((($decoder)) + 16|0);
  HEAP32[$17>>2] = $16;
  $18 = ((($14)) + 56|0);
  $19 = ((($decoder)) + 20|0);
  HEAP32[$19>>2] = $18;
  _FT_GlyphLoader_Rewind($14);
  $20 = ((($size)) + 40|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = ((($decoder)) + 76|0);
  HEAP32[$22>>2] = $21;
  $23 = ((($decoder)) + 72|0);
  HEAP32[$23>>2] = 0;
  $24 = ($hinting<<24>>24)==(0);
  if (!($24)) {
   $25 = HEAP32[$12>>2]|0;
   $26 = ((($25)) + 36|0);
   $27 = HEAP32[$26>>2]|0;
   HEAP32[$23>>2] = $27;
  }
 }
 $28 = ((($decoder)) + 24|0);
 $29 = ((($decoder)) + 80|0);
 ;HEAP32[$28>>2]=0|0;HEAP32[$28+4>>2]=0|0;HEAP32[$28+8>>2]=0|0;HEAP32[$28+12>>2]=0|0;HEAP32[$28+16>>2]=0|0;HEAP32[$28+20>>2]=0|0;
 ;HEAP32[$29>>2]=HEAP32[53576>>2]|0;HEAP32[$29+4>>2]=HEAP32[53576+4>>2]|0;HEAP32[$29+8>>2]=HEAP32[53576+8>>2]|0;HEAP32[$29+12>>2]=HEAP32[53576+12>>2]|0;HEAP32[$29+16>>2]=HEAP32[53576+16>>2]|0;HEAP32[$29+20>>2]=HEAP32[53576+20>>2]|0;HEAP32[$29+24>>2]=HEAP32[53576+24>>2]|0;HEAP32[$29+28>>2]=HEAP32[53576+28>>2]|0;
 $30 = ((($face)) + 16|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = ((($decoder)) + 1352|0);
 HEAP32[$32>>2] = $31;
 $33 = ((($decoder)) + 1356|0);
 HEAP32[$33>>2] = $glyph_names;
 $34 = ((($decoder)) + 1468|0);
 HEAP32[$34>>2] = $hint_mode;
 $35 = ((($decoder)) + 1464|0);
 HEAP32[$35>>2] = $blend;
 $36 = ((($decoder)) + 1472|0);
 HEAP32[$36>>2] = $parse_callback;
 $37 = ((($decoder)) + 1476|0);
 ;HEAP32[$37>>2]=HEAP32[53608>>2]|0;HEAP32[$37+4>>2]=HEAP32[53608+4>>2]|0;HEAP32[$37+8>>2]=HEAP32[53608+8>>2]|0;
 $$1 = 0;
 return ($$1|0);
}
function _t1_decoder_done($decoder) {
 $decoder = $decoder|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($decoder)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 $3 = ((($1)) + 108|0);
 $4 = ((($decoder)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 ;HEAP32[$3>>2]=HEAP32[$5>>2]|0;HEAP32[$3+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$3+8>>2]=HEAP32[$5+8>>2]|0;HEAP32[$3+12>>2]=HEAP32[$5+12>>2]|0;HEAP32[$3+16>>2]=HEAP32[$5+16>>2]|0;
 return;
}
function _t1_decoder_parse_charstrings($decoder,$charstring_base,$charstring_len) {
 $decoder = $decoder|0;
 $charstring_base = $charstring_base|0;
 $charstring_len = $charstring_len|0;
 var $$ = 0, $$2$i$i = 0, $$26 = 0, $$31 = 0, $$32 = 0, $$4 = 0, $$cast = 0, $$lcssa266 = 0, $$lcssa266$lcssa = 0, $$lcssa266$lcssa$lcssa270 = 0, $$lcssa266268 = 0, $$lcssa271 = 0, $$lcssa271$lcssa = 0, $$lcssa282 = 0, $$off = 0, $$off25 = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i39 = 0;
 var $$pre$i43 = 0, $$pre$i47 = 0, $$pre$i52 = 0, $$pre$i56 = 0, $$pre$i60 = 0, $$pre$i65 = 0, $$pre$i69 = 0, $$pre$i73 = 0, $$pre$phi$i$iZ2D = 0, $$pre$phi$i41Z2D = 0, $$pre$phi$i45Z2D = 0, $$pre$phi$i49Z2D = 0, $$pre$phi$i54Z2D = 0, $$pre$phi$i58Z2D = 0, $$pre$phi$i62Z2D = 0, $$pre$phi$i67Z2D = 0, $$pre$phi$i71Z2D = 0, $$pre$phi$i75Z2D = 0, $$pre$phi$iZ2D = 0, $$pre191 = 0;
 var $$pre192 = 0, $$pre193 = 0, $$pre194 = 0, $$pre195 = 0, $$pre196 = 0, $$pre197 = 0, $$pre198 = 0, $$pre199 = 0, $$pre200 = 0, $$pre201 = 0, $$pre202 = 0, $$value$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0;
 var $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0;
 var $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0;
 var $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0;
 var $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0;
 var $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0;
 var $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0;
 var $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0;
 var $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0;
 var $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0;
 var $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0;
 var $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0;
 var $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0;
 var $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0;
 var $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0;
 var $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0;
 var $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0;
 var $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0;
 var $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0;
 var $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0;
 var $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0;
 var $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0;
 var $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0;
 var $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0;
 var $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0;
 var $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0;
 var $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0;
 var $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0;
 var $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0;
 var $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0;
 var $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0;
 var $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0;
 var $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0;
 var $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0;
 var $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0;
 var $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0;
 var $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0;
 var $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0;
 var $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0;
 var $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0;
 var $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0;
 var $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0;
 var $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0;
 var $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0;
 var $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0;
 var $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0;
 var $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0;
 var $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0;
 var $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0;
 var $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0;
 var $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0;
 var $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0;
 var $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $achar_index$0$i = 0, $bchar_index$0$i = 0, $delta$0153 = 0;
 var $delta$1$lcssa = 0, $delta$1147 = 0, $exitcond = 0, $ip$0$ph$ph170 = 0, $ip$0$ph160 = 0, $ip$0133 = 0, $ip$1 = 0, $ip$2 = 0, $ip$2$lcssa = 0, $ip$2$lcssa$lcssa = 0, $ip$4 = 0, $known_othersubr_result_cnt$0$ph$ph165 = 0, $known_othersubr_result_cnt$0$ph159 = 0, $known_othersubr_result_cnt$0$ph159$lcssa279 = 0, $known_othersubr_result_cnt$2 = 0, $known_othersubr_result_cnt$4 = 0, $large_int$0$ = 0, $large_int$0$ph$ph163 = 0, $large_int$0$ph157 = 0, $large_int$0131 = 0;
 var $large_int$1 = 0, $large_int$1$ = 0, $large_int$1$$lcssa = 0, $large_int$1$$lcssa$lcssa = 0, $large_int$3 = 0, $limit$0$ph$ph169 = 0, $limit$2 = 0, $mm$0146 = 0, $n$04$i$i = 0, $n$04$i4$i = 0, $nn$0152 = 0, $op$0 = 0, $op$0$lcssa = 0, $op$0$lcssa$lcssa = 0, $or$cond12 = 0, $or$cond14 = 0, $or$cond16 = 0, $or$cond6 = 0, $or$cond8 = 0, $orig_x$0$ph$ph166 = 0;
 var $orig_x$1 = 0, $seed = 0, $sext$i = 0, $sext$i35 = 0, $sext2$i = 0, $sext2$i36 = 0, $storemerge = 0, $switch34 = 0, $tmp$0$lcssa = 0, $tmp$0148 = 0, $top$0 = 0, $top$1 = 0, $unknown_othersubr_result_cnt$0$ph$ph164 = 0, $unknown_othersubr_result_cnt$0$ph158 = 0, $unknown_othersubr_result_cnt$0132 = 0, $unknown_othersubr_result_cnt$1 = 0, $unknown_othersubr_result_cnt$1$lcssa = 0, $unknown_othersubr_result_cnt$1$lcssa$lcssa = 0, $unknown_othersubr_result_cnt$2 = 0, $unknown_othersubr_result_cnt$4 = 0;
 var $value$0 = 0, $value$1 = 0, $values$0154 = 0, $x$0$ph$ph168 = 0, $x$1 = 0, $x$2 = 0, $y$0$ph$ph167 = 0, $y$1 = 0, $y$2 = 0, $zone$0$ph$ph171 = 0, $zone$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 8|0;
 $1 = sp + 4|0;
 $seed = sp;
 HEAP32[$0>>2] = $decoder;
 HEAP32[$1>>2] = $charstring_base;
 $2 = $seed;
 $3 = $0;
 $4 = $1;
 $5 = $4 ^ $3;
 $6 = $5 ^ $2;
 $7 = $6 >>> 10;
 $8 = $7 ^ $6;
 $9 = $6 >> 20;
 $10 = $8 ^ $9;
 $11 = $10 & 65535;
 $12 = ($11|0)==(0);
 $$ = $12 ? 29572 : $11;
 HEAP32[$seed>>2] = $$;
 $13 = ((($decoder)) + 112|0);
 $14 = ((($decoder)) + 1136|0);
 HEAP32[$14>>2] = $13;
 $15 = ((($decoder)) + 1140|0);
 $16 = ((($decoder)) + 1344|0);
 HEAP32[$16>>2] = $15;
 $17 = ((($decoder)) + 64|0);
 HEAP32[$17>>2] = 0;
 $18 = ((($decoder)) + 72|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ((($decoder)) + 1488|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ($21|0)==(0|0);
 $23 = $charstring_base;
 if ($22) {
  $30 = $23;$32 = $charstring_base;
 } else {
  $24 = ((($decoder)) + 1492|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = ($25|0)==(0);
  if ($26) {
   $30 = $23;$32 = $charstring_base;
  } else {
   $27 = $25 << 2;
   _memset(($21|0),0,($27|0))|0;
   $$pre = HEAP32[$1>>2]|0;
   $28 = $$pre;
   $30 = $$pre;$32 = $28;
  }
 }
 $29 = ((($decoder)) + 1144|0);
 HEAP32[$29>>2] = $30;
 $31 = (($32) + ($charstring_len)|0);
 $33 = ((($decoder)) + 1148|0);
 HEAP32[$33>>2] = $31;
 $$cast = $30;
 HEAP32[$15>>2] = $$cast;
 $34 = ((($decoder)) + 24|0);
 $35 = HEAP32[$34>>2]|0;
 $36 = ((($decoder)) + 28|0);
 $37 = HEAP32[$36>>2]|0;
 $38 = ($19|0)!=(0|0);
 if ($38) {
  $39 = ((($19)) + 4|0);
  $40 = HEAP32[$39>>2]|0;
  $41 = HEAP32[$19>>2]|0;
  FUNCTION_TABLE_vi[$40 & 8191]($41);
 }
 $42 = ($$cast>>>0)<($31>>>0);
 if (!($42)) {
  $$4 = 0;
  STACKTOP = sp;return ($$4|0);
 }
 $43 = ((($decoder)) + 12|0);
 $44 = ((($decoder)) + 20|0);
 $45 = ((($decoder)) + 68|0);
 $46 = ((($19)) + 20|0);
 $47 = ((($decoder)) + 32|0);
 $48 = ((($decoder)) + 40|0);
 $49 = ((($decoder)) + 44|0);
 $50 = ((($decoder)) + 70|0);
 $51 = ((($decoder)) + 36|0);
 $52 = ((($19)) + 12|0);
 $53 = ((($19)) + 16|0);
 $ip$0$ph$ph170 = $$cast;$known_othersubr_result_cnt$0$ph$ph165 = 0;$large_int$0$ph$ph163 = 0;$limit$0$ph$ph169 = $31;$orig_x$0$ph$ph166 = $35;$unknown_othersubr_result_cnt$0$ph$ph164 = 0;$x$0$ph$ph168 = $35;$y$0$ph$ph167 = $37;$zone$0$ph$ph171 = $15;
 L11: while(1) {
  $ip$0$ph160 = $ip$0$ph$ph170;$known_othersubr_result_cnt$0$ph159 = $known_othersubr_result_cnt$0$ph$ph165;$large_int$0$ph157 = $large_int$0$ph$ph163;$unknown_othersubr_result_cnt$0$ph158 = $unknown_othersubr_result_cnt$0$ph$ph164;
  while(1) {
   $ip$0133 = $ip$0$ph160;$large_int$0131 = $large_int$0$ph157;$unknown_othersubr_result_cnt$0132 = $unknown_othersubr_result_cnt$0$ph158;
   while(1) {
    $54 = HEAP32[$0>>2]|0;
    $55 = ((($54)) + 1136|0);
    $56 = HEAP32[$55>>2]|0;
    $57 = ((($ip$0133)) + 1|0);
    $58 = HEAP8[$ip$0133>>0]|0;
    $59 = $58&255;
    L17: do {
     switch ($59|0) {
     case 1:  {
      $ip$2 = $57;$large_int$1 = $large_int$0131;$op$0 = 16;$value$1 = 0;
      break;
     }
     case 3:  {
      $ip$2 = $57;$large_int$1 = $large_int$0131;$op$0 = 18;$value$1 = 0;
      break;
     }
     case 4:  {
      $ip$2 = $57;$large_int$1 = $large_int$0131;$op$0 = 14;$value$1 = 0;
      break;
     }
     case 5:  {
      $ip$2 = $57;$large_int$1 = $large_int$0131;$op$0 = 9;$value$1 = 0;
      break;
     }
     case 6:  {
      $ip$2 = $57;$large_int$1 = $large_int$0131;$op$0 = 6;$value$1 = 0;
      break;
     }
     case 7:  {
      $ip$2 = $57;$large_int$1 = $large_int$0131;$op$0 = 13;$value$1 = 0;
      break;
     }
     case 8:  {
      $ip$2 = $57;$large_int$1 = $large_int$0131;$op$0 = 11;$value$1 = 0;
      break;
     }
     case 9:  {
      $ip$2 = $57;$large_int$1 = $large_int$0131;$op$0 = 5;$value$1 = 0;
      break;
     }
     case 10:  {
      $ip$2 = $57;$large_int$1 = $large_int$0131;$op$0 = 22;$value$1 = 0;
      break;
     }
     case 11:  {
      $ip$2 = $57;$large_int$1 = $large_int$0131;$op$0 = 24;$value$1 = 0;
      break;
     }
     case 13:  {
      $ip$2 = $57;$large_int$1 = $large_int$0131;$op$0 = 2;$value$1 = 0;
      break;
     }
     case 14:  {
      $ip$2 = $57;$large_int$1 = $large_int$0131;$op$0 = 1;$value$1 = 0;
      break;
     }
     case 15:  {
      $ip$2 = $57;$large_int$1 = $large_int$0131;$op$0 = 26;$value$1 = 0;
      break;
     }
     case 21:  {
      $ip$2 = $57;$large_int$1 = $large_int$0131;$op$0 = 10;$value$1 = 0;
      break;
     }
     case 22:  {
      $ip$2 = $57;$large_int$1 = $large_int$0131;$op$0 = 7;$value$1 = 0;
      break;
     }
     case 30:  {
      $ip$2 = $57;$large_int$1 = $large_int$0131;$op$0 = 12;$value$1 = 0;
      break;
     }
     case 31:  {
      $ip$2 = $57;$large_int$1 = $large_int$0131;$op$0 = 8;$value$1 = 0;
      break;
     }
     case 12:  {
      $60 = ($57>>>0)>($limit$0$ph$ph169>>>0);
      if ($60) {
       $$4 = 160;
       label = 257;
       break L11;
      }
      $61 = ((($ip$0133)) + 2|0);
      $62 = HEAP8[$57>>0]|0;
      $63 = $62&255;
      switch ($63|0) {
      case 0:  {
       $ip$2 = $61;$large_int$1 = $large_int$0131;$op$0 = 15;$value$1 = 0;
       break L17;
       break;
      }
      case 1:  {
       $ip$2 = $61;$large_int$1 = $large_int$0131;$op$0 = 19;$value$1 = 0;
       break L17;
       break;
      }
      case 2:  {
       $ip$2 = $61;$large_int$1 = $large_int$0131;$op$0 = 17;$value$1 = 0;
       break L17;
       break;
      }
      case 6:  {
       $ip$2 = $61;$large_int$1 = $large_int$0131;$op$0 = 3;$value$1 = 0;
       break L17;
       break;
      }
      case 7:  {
       $ip$2 = $61;$large_int$1 = $large_int$0131;$op$0 = 4;$value$1 = 0;
       break L17;
       break;
      }
      case 12:  {
       $ip$2 = $61;$large_int$1 = $large_int$0131;$op$0 = 20;$value$1 = 0;
       break L17;
       break;
      }
      case 16:  {
       $ip$2 = $61;$large_int$1 = $large_int$0131;$op$0 = 21;$value$1 = 0;
       break L17;
       break;
      }
      case 17:  {
       $ip$2 = $61;$large_int$1 = $large_int$0131;$op$0 = 23;$value$1 = 0;
       break L17;
       break;
      }
      case 33:  {
       $ip$2 = $61;$large_int$1 = $large_int$0131;$op$0 = 25;$value$1 = 0;
       break L17;
       break;
      }
      default: {
       $$4 = 160;
       label = 257;
       break L11;
      }
      }
      break;
     }
     case 255:  {
      $64 = ((($ip$0133)) + 5|0);
      $65 = ($64>>>0)>($limit$0$ph$ph169>>>0);
      if ($65) {
       $$4 = 160;
       label = 257;
       break L11;
      }
      $66 = HEAP8[$57>>0]|0;
      $67 = $66&255;
      $68 = $67 << 24;
      $69 = ((($ip$0133)) + 2|0);
      $70 = HEAP8[$69>>0]|0;
      $71 = $70&255;
      $72 = $71 << 16;
      $73 = $72 | $68;
      $74 = ((($ip$0133)) + 3|0);
      $75 = HEAP8[$74>>0]|0;
      $76 = $75&255;
      $77 = $76 << 8;
      $78 = $73 | $77;
      $79 = ((($ip$0133)) + 4|0);
      $80 = HEAP8[$79>>0]|0;
      $81 = $80&255;
      $82 = $78 | $81;
      $$off = (($82) + 32000)|0;
      $83 = ($$off>>>0)>(64000);
      $84 = ($large_int$0131<<24>>24)!=(0);
      if ($83) {
       $large_int$0$ = $84 ? $large_int$0131 : 1;
       $ip$2 = $64;$large_int$1 = $large_int$0$;$op$0 = 0;$value$1 = $82;
       break L17;
      } else {
       $85 = $82 << 16;
       $$26 = $84 ? $82 : $85;
       $ip$2 = $64;$large_int$1 = $large_int$0131;$op$0 = 0;$value$1 = $$26;
       break L17;
      }
      break;
     }
     default: {
      $86 = ($58&255)>(31);
      if (!($86)) {
       $$4 = 160;
       label = 257;
       break L11;
      }
      $87 = ($58&255)<(247);
      do {
       if ($87) {
        $88 = (($59) + -139)|0;
        $ip$1 = $57;$value$0 = $88;
       } else {
        $89 = ((($ip$0133)) + 2|0);
        $90 = ($89>>>0)>($limit$0$ph$ph169>>>0);
        if ($90) {
         $$4 = 160;
         label = 257;
         break L11;
        }
        $91 = ($58&255)<(251);
        $92 = $59 << 8;
        if ($91) {
         $93 = (($92) + -63232)|0;
         $94 = HEAP8[$57>>0]|0;
         $95 = $94&255;
         $96 = $95 | $93;
         $97 = (($96) + 108)|0;
         $ip$1 = $89;$value$0 = $97;
         break;
        } else {
         $98 = (($92) + -64256)|0;
         $99 = HEAP8[$57>>0]|0;
         $100 = $99&255;
         $101 = $100 | $98;
         $102 = (-108 - ($101))|0;
         $ip$1 = $89;$value$0 = $102;
         break;
        }
       }
      } while(0);
      $103 = ($large_int$0131<<24>>24)==(0);
      $104 = $value$0 << 16;
      $$value$0 = $103 ? $104 : $value$0;
      $ip$2 = $ip$1;$large_int$1 = $large_int$0131;$op$0 = 0;$value$1 = $$value$0;
     }
     }
    } while(0);
    $105 = ($unknown_othersubr_result_cnt$0132|0)>(0);
    L61: do {
     if ($105) {
      switch ($op$0|0) {
      case 23: case 0: case 24: case 22:  {
       $unknown_othersubr_result_cnt$1 = $unknown_othersubr_result_cnt$0132;
       break L61;
       break;
      }
      default: {
      }
      }
      $unknown_othersubr_result_cnt$1 = 0;
     } else {
      $unknown_othersubr_result_cnt$1 = $unknown_othersubr_result_cnt$0132;
     }
    } while(0);
    $106 = ($large_int$1<<24>>24)==(0);
    $107 = ($op$0|0)==(0);
    $or$cond6 = $107 | $106;
    $108 = ($op$0|0)==(20);
    $or$cond8 = $108 | $or$cond6;
    $large_int$1$ = $or$cond8 ? $large_int$1 : 0;
    if (!($107)) {
     $$lcssa266 = $54;$$lcssa271 = $56;$ip$2$lcssa = $ip$2;$large_int$1$$lcssa = $large_int$1$;$op$0$lcssa = $op$0;$unknown_othersubr_result_cnt$1$lcssa = $unknown_othersubr_result_cnt$1;
     break;
    }
    $109 = ((($54)) + 112|0);
    $110 = $56;
    $111 = $109;
    $112 = (($110) - ($111))|0;
    $113 = ($112|0)>(1020);
    if ($113) {
     $$4 = 160;
     label = 257;
     break L11;
    }
    $114 = ((($56)) + 4|0);
    HEAP32[$56>>2] = $value$1;
    HEAP32[$55>>2] = $114;
    $115 = ($ip$2>>>0)<($limit$0$ph$ph169>>>0);
    if ($115) {
     $ip$0133 = $ip$2;$large_int$0131 = $large_int$1$;$unknown_othersubr_result_cnt$0132 = $unknown_othersubr_result_cnt$1;
    } else {
     $$4 = 0;
     label = 257;
     break L11;
    }
   }
   $116 = ($op$0$lcssa|0)==(21);
   if (!($116)) {
    $$lcssa266$lcssa = $$lcssa266;$$lcssa271$lcssa = $$lcssa271;$ip$2$lcssa$lcssa = $ip$2$lcssa;$known_othersubr_result_cnt$0$ph159$lcssa279 = $known_othersubr_result_cnt$0$ph159;$large_int$1$$lcssa$lcssa = $large_int$1$$lcssa;$op$0$lcssa$lcssa = $op$0$lcssa;$unknown_othersubr_result_cnt$1$lcssa$lcssa = $unknown_othersubr_result_cnt$1$lcssa;
    break;
   }
   $117 = ((($$lcssa266)) + 112|0);
   $118 = $$lcssa271;
   $119 = $117;
   $120 = (($118) - ($119))|0;
   $121 = ($120|0)<(8);
   if ($121) {
    $$4 = 161;
    label = 257;
    break L11;
   }
   $122 = ((($$lcssa271)) + -8|0);
   $123 = ((($$lcssa271)) + -4|0);
   $124 = HEAP32[$123>>2]|0;
   $125 = $124 >> 16;
   $126 = HEAP32[$122>>2]|0;
   $127 = $126 >> 16;
   $128 = $122;
   $129 = (($128) - ($119))|0;
   $130 = $129 >> 2;
   $131 = ($127|0)>($130|0);
   if ($131) {
    $$4 = 161;
    label = 257;
    break L11;
   }
   $132 = (0 - ($127))|0;
   $133 = (($122) + ($132<<2)|0);
   do {
    switch ($125|0) {
    case 0:  {
     $134 = ($127|0)==(3);
     if (!($134)) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $135 = ((($$lcssa266)) + 1400|0);
     $136 = HEAP32[$135>>2]|0;
     $137 = ($136|0)==(0);
     if ($137) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $138 = ((($$lcssa266)) + 1404|0);
     $139 = HEAP32[$138>>2]|0;
     $140 = ($139|0)==(7);
     if (!($140)) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     HEAP32[$133>>2] = $x$0$ph$ph168;
     $141 = ((($133)) + 4|0);
     HEAP32[$141>>2] = $y$0$ph$ph167;
     $known_othersubr_result_cnt$2 = 2;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
     break;
    }
    case 1:  {
     $142 = ($127|0)==(0);
     if (!($142)) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $143 = ((($$lcssa266)) + 1400|0);
     HEAP32[$143>>2] = 1;
     $144 = ((($$lcssa266)) + 1404|0);
     HEAP32[$144>>2] = 0;
     $145 = (_t1_builder_start_point($decoder,$x$0$ph$ph168,$y$0$ph$ph167)|0);
     $146 = ($145|0)==(0);
     if (!($146)) {
      $$4 = $145;
      label = 257;
      break L11;
     }
     $147 = HEAP32[$43>>2]|0;
     $148 = ((($147)) + 22|0);
     $149 = HEAP16[$148>>1]|0;
     $150 = $149 << 16 >> 16;
     $151 = ((($147)) + 58|0);
     $152 = HEAP16[$151>>1]|0;
     $153 = $152 << 16 >> 16;
     $154 = (($150) + 6)|0;
     $155 = (($154) + ($153))|0;
     $156 = ((($147)) + 4|0);
     $157 = HEAP32[$156>>2]|0;
     $158 = ($155>>>0)>($157>>>0);
     if ($158) {
      $159 = (_FT_GlyphLoader_CheckPoints($147,6,0)|0);
      $160 = ($159|0)==(0);
      if ($160) {
       $known_othersubr_result_cnt$2 = 0;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
      } else {
       $$4 = $159;
       label = 257;
       break L11;
      }
     } else {
      $known_othersubr_result_cnt$2 = 0;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
     }
     break;
    }
    case 2:  {
     $161 = ($127|0)==(0);
     if (!($161)) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $162 = ((($$lcssa266)) + 1400|0);
     $163 = HEAP32[$162>>2]|0;
     $164 = ($163|0)==(0);
     if ($164) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $165 = ((($$lcssa266)) + 1404|0);
     $166 = HEAP32[$165>>2]|0;
     $167 = (($166) + 1)|0;
     HEAP32[$165>>2] = $167;
     $$off25 = (($166) + -1)|0;
     $168 = ($$off25>>>0)<(6);
     if ($168) {
      $169 = HEAP32[$44>>2]|0;
      $170 = HEAP8[$45>>0]|0;
      $171 = ($170<<24>>24)==(0);
      if ($171) {
       $$pre$i = ((($169)) + 2|0);
       $$pre$phi$iZ2D = $$pre$i;
      } else {
       $172 = ($166|0)==(3);
       $173 = ($166|0)==(6);
       $174 = $172 | $173;
       $175 = ((($169)) + 4|0);
       $176 = HEAP32[$175>>2]|0;
       $177 = ((($169)) + 2|0);
       $178 = HEAP16[$177>>1]|0;
       $179 = $178 << 16 >> 16;
       $180 = ((($169)) + 8|0);
       $181 = HEAP32[$180>>2]|0;
       $182 = (($181) + ($179)|0);
       $183 = (_FT_RoundFix($x$0$ph$ph168)|0);
       $184 = $183 >> 16;
       $185 = (($176) + ($179<<3)|0);
       HEAP32[$185>>2] = $184;
       $186 = (_FT_RoundFix($y$0$ph$ph167)|0);
       $187 = $186 >> 16;
       $188 = (((($176) + ($179<<3)|0)) + 4|0);
       HEAP32[$188>>2] = $187;
       $189 = $174 ? 1 : 2;
       HEAP8[$182>>0] = $189;
       $$pre$phi$iZ2D = $177;
      }
      $190 = HEAP16[$$pre$phi$iZ2D>>1]|0;
      $191 = (($190) + 1)<<16>>16;
      HEAP16[$$pre$phi$iZ2D>>1] = $191;
      $known_othersubr_result_cnt$2 = 0;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
     } else {
      $known_othersubr_result_cnt$2 = 0;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
     }
     break;
    }
    case 3:  {
     $192 = ($127|0)==(1);
     if (!($192)) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     if ($38) {
      $193 = HEAP32[$46>>2]|0;
      $194 = HEAP32[$19>>2]|0;
      $195 = HEAP32[$44>>2]|0;
      $196 = ((($195)) + 2|0);
      $197 = HEAP16[$196>>1]|0;
      $198 = $197 << 16 >> 16;
      FUNCTION_TABLE_vii[$193 & 8191]($194,$198);
      $known_othersubr_result_cnt$2 = 1;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
     } else {
      $known_othersubr_result_cnt$2 = 1;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
     }
     break;
    }
    case 13: case 12:  {
     $known_othersubr_result_cnt$2 = 0;$top$0 = $117;$unknown_othersubr_result_cnt$2 = 0;
     break;
    }
    case 18: case 17: case 16: case 15: case 14:  {
     $199 = ((($$lcssa266)) + 1464|0);
     $200 = HEAP32[$199>>2]|0;
     $201 = ($200|0)==(0|0);
     if ($201) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $202 = (($125) + -13)|0;
     $203 = ($125|0)==(18);
     $204 = $203&1;
     $205 = (($202) + ($204))|0;
     $206 = HEAP32[$200>>2]|0;
     $207 = Math_imul($206, $205)|0;
     $208 = ($127|0)==($207|0);
     if (!($208)) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $209 = ($205|0)==(0);
     if ($209) {
      $known_othersubr_result_cnt$2 = 0;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
     } else {
      $210 = (($133) + ($205<<2)|0);
      $211 = ((($200)) + 136|0);
      $214 = $206;$delta$0153 = $210;$nn$0152 = 0;$values$0154 = $133;
      while(1) {
       $212 = HEAP32[$values$0154>>2]|0;
       $213 = ($214>>>0)>(1);
       if ($213) {
        $delta$1147 = $delta$0153;$mm$0146 = 1;$tmp$0148 = $212;
        while(1) {
         $215 = ((($delta$1147)) + 4|0);
         $216 = HEAP32[$delta$1147>>2]|0;
         $217 = HEAP32[$211>>2]|0;
         $218 = (($217) + ($mm$0146<<2)|0);
         $219 = HEAP32[$218>>2]|0;
         $220 = (_FT_MulFix($216,$219)|0);
         $221 = (($220) + ($tmp$0148))|0;
         $222 = (($mm$0146) + 1)|0;
         $223 = HEAP32[$200>>2]|0;
         $224 = ($222>>>0)<($223>>>0);
         if ($224) {
          $delta$1147 = $215;$mm$0146 = $222;$tmp$0148 = $221;
         } else {
          $1043 = $223;$delta$1$lcssa = $215;$tmp$0$lcssa = $221;
          break;
         }
        }
       } else {
        $1043 = $214;$delta$1$lcssa = $delta$0153;$tmp$0$lcssa = $212;
       }
       $225 = ((($values$0154)) + 4|0);
       HEAP32[$values$0154>>2] = $tmp$0$lcssa;
       $226 = (($nn$0152) + 1)|0;
       $exitcond = ($226|0)==($205|0);
       if ($exitcond) {
        $known_othersubr_result_cnt$2 = $205;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
        break;
       } else {
        $214 = $1043;$delta$0153 = $delta$1$lcssa;$nn$0152 = $226;$values$0154 = $225;
       }
      }
     }
     break;
    }
    case 19:  {
     $227 = ((($$lcssa266)) + 1464|0);
     $228 = HEAP32[$227>>2]|0;
     $229 = ($127|0)!=(1);
     $230 = ($228|0)==(0|0);
     $or$cond12 = $229 | $230;
     if ($or$cond12) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $231 = HEAP32[$133>>2]|0;
     $232 = $231 >> 16;
     $233 = ($232|0)<(0);
     if ($233) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $234 = HEAP32[$228>>2]|0;
     $235 = (($234) + ($232))|0;
     $236 = ((($$lcssa266)) + 1492|0);
     $237 = HEAP32[$236>>2]|0;
     $238 = ($235>>>0)>($237>>>0);
     if ($238) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $239 = ((($$lcssa266)) + 1488|0);
     $240 = HEAP32[$239>>2]|0;
     $241 = (($240) + ($232<<2)|0);
     $242 = ((($228)) + 136|0);
     $243 = HEAP32[$242>>2]|0;
     $244 = $234 << 2;
     _memcpy(($241|0),($243|0),($244|0))|0;
     $known_othersubr_result_cnt$2 = 0;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
     break;
    }
    case 20:  {
     $245 = ($127|0)==(2);
     if (!($245)) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $246 = ((($133)) + 4|0);
     $247 = HEAP32[$246>>2]|0;
     $248 = HEAP32[$133>>2]|0;
     $249 = (($248) + ($247))|0;
     HEAP32[$133>>2] = $249;
     $known_othersubr_result_cnt$2 = 1;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
     break;
    }
    case 21:  {
     $250 = ($127|0)==(2);
     if (!($250)) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $251 = ((($133)) + 4|0);
     $252 = HEAP32[$251>>2]|0;
     $253 = HEAP32[$133>>2]|0;
     $254 = (($253) - ($252))|0;
     HEAP32[$133>>2] = $254;
     $known_othersubr_result_cnt$2 = 1;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
     break;
    }
    case 22:  {
     $255 = ($127|0)==(2);
     if (!($255)) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $256 = HEAP32[$133>>2]|0;
     $257 = ((($133)) + 4|0);
     $258 = HEAP32[$257>>2]|0;
     $259 = (_FT_MulFix($256,$258)|0);
     HEAP32[$133>>2] = $259;
     $known_othersubr_result_cnt$2 = 1;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
     break;
    }
    case 23:  {
     $260 = ($127|0)==(2);
     if (!($260)) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $261 = ((($133)) + 4|0);
     $262 = HEAP32[$261>>2]|0;
     $263 = ($262|0)==(0);
     if ($263) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $264 = HEAP32[$133>>2]|0;
     $265 = (_FT_DivFix($264,$262)|0);
     HEAP32[$133>>2] = $265;
     $known_othersubr_result_cnt$2 = 1;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
     break;
    }
    case 24:  {
     $266 = ((($$lcssa266)) + 1464|0);
     $267 = HEAP32[$266>>2]|0;
     $268 = ($127|0)!=(2);
     $269 = ($267|0)==(0|0);
     $or$cond14 = $268 | $269;
     if ($or$cond14) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $270 = ((($133)) + 4|0);
     $271 = HEAP32[$270>>2]|0;
     $272 = $271 >> 16;
     $273 = ($272|0)<(0);
     if ($273) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $274 = ((($$lcssa266)) + 1492|0);
     $275 = HEAP32[$274>>2]|0;
     $276 = ($272>>>0)<($275>>>0);
     if (!($276)) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $277 = HEAP32[$133>>2]|0;
     $278 = ((($$lcssa266)) + 1488|0);
     $279 = HEAP32[$278>>2]|0;
     $280 = (($279) + ($272<<2)|0);
     HEAP32[$280>>2] = $277;
     $known_othersubr_result_cnt$2 = 0;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
     break;
    }
    case 25:  {
     $281 = ((($$lcssa266)) + 1464|0);
     $282 = HEAP32[$281>>2]|0;
     $283 = ($127|0)!=(1);
     $284 = ($282|0)==(0|0);
     $or$cond16 = $283 | $284;
     if ($or$cond16) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $285 = HEAP32[$133>>2]|0;
     $286 = $285 >> 16;
     $287 = ($286|0)<(0);
     if ($287) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $288 = ((($$lcssa266)) + 1492|0);
     $289 = HEAP32[$288>>2]|0;
     $290 = ($286>>>0)<($289>>>0);
     if (!($290)) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $291 = ((($$lcssa266)) + 1488|0);
     $292 = HEAP32[$291>>2]|0;
     $293 = (($292) + ($286<<2)|0);
     $294 = HEAP32[$293>>2]|0;
     HEAP32[$133>>2] = $294;
     $known_othersubr_result_cnt$2 = 1;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
     break;
    }
    case 27:  {
     $295 = ($127|0)==(4);
     if (!($295)) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $296 = ((($133)) + 8|0);
     $297 = HEAP32[$296>>2]|0;
     $298 = ((($133)) + 12|0);
     $299 = HEAP32[$298>>2]|0;
     $300 = ($297|0)>($299|0);
     if ($300) {
      $301 = ((($133)) + 4|0);
      $302 = HEAP32[$301>>2]|0;
      HEAP32[$133>>2] = $302;
      $known_othersubr_result_cnt$2 = 1;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
     } else {
      $known_othersubr_result_cnt$2 = 1;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
     }
     break;
    }
    case 28:  {
     $303 = ($127|0)==(0);
     if (!($303)) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     $304 = HEAP32[$seed>>2]|0;
     $305 = ($304|0)>(32767);
     $306 = $305&1;
     $$31 = (($306) + ($304))|0;
     HEAP32[$133>>2] = $$31;
     $307 = HEAP32[$seed>>2]|0;
     $308 = (65536 - ($307))|0;
     $309 = (_FT_MulFix($307,$308)|0);
     $310 = ($309|0)==(0);
     $311 = (($309) + 10355)|0;
     $storemerge = $310 ? $311 : $309;
     HEAP32[$seed>>2] = $storemerge;
     $known_othersubr_result_cnt$2 = 1;$top$0 = $133;$unknown_othersubr_result_cnt$2 = 0;
     break;
    }
    default: {
     $312 = $127 | $125;
     $313 = ($312|0)>(-1);
     if ($313) {
      $known_othersubr_result_cnt$2 = 0;$top$0 = $133;$unknown_othersubr_result_cnt$2 = $127;
     } else {
      $$4 = 160;
      label = 257;
      break L11;
     }
    }
    }
   } while(0);
   $314 = (($top$0) + ($known_othersubr_result_cnt$2<<2)|0);
   $315 = HEAP32[$0>>2]|0;
   $316 = ((($315)) + 1136|0);
   HEAP32[$316>>2] = $314;
   $317 = ($ip$2$lcssa>>>0)<($limit$0$ph$ph169>>>0);
   if ($317) {
    $ip$0$ph160 = $ip$2$lcssa;$known_othersubr_result_cnt$0$ph159 = $known_othersubr_result_cnt$2;$large_int$0$ph157 = $large_int$1$$lcssa;$unknown_othersubr_result_cnt$0$ph158 = $unknown_othersubr_result_cnt$2;
   } else {
    $$4 = 0;
    label = 257;
    break L11;
   }
  }
  $318 = (53620 + ($op$0$lcssa$lcssa<<2)|0);
  $319 = HEAP32[$318>>2]|0;
  $320 = ((($$lcssa266$lcssa)) + 112|0);
  $321 = $$lcssa271$lcssa;
  $322 = $320;
  $323 = (($321) - ($322))|0;
  $324 = $323 >> 2;
  $325 = ($324|0)<($319|0);
  if ($325) {
   $$4 = 161;
   label = 257;
   break;
  }
  $326 = (0 - ($319))|0;
  $327 = (($$lcssa271$lcssa) + ($326<<2)|0);
  L131: do {
   switch ($op$0$lcssa$lcssa|0) {
   case 1:  {
    label = 114;
    break L11;
    break;
   }
   case 3:  {
    $$lcssa266$lcssa$lcssa270 = $$lcssa266$lcssa;$$lcssa282 = $327;
    label = 131;
    break L11;
    break;
   }
   case 2:  {
    HEAP32[$17>>2] = 1;
    $400 = HEAP32[$327>>2]|0;
    $401 = HEAP32[$47>>2]|0;
    $402 = (($401) + ($400))|0;
    HEAP32[$47>>2] = $402;
    $403 = ((($327)) + 4|0);
    $404 = HEAP32[$403>>2]|0;
    HEAP32[$48>>2] = $404;
    HEAP32[$49>>2] = 0;
    $405 = HEAP32[$34>>2]|0;
    $406 = HEAP32[$327>>2]|0;
    $407 = (($406) + ($405))|0;
    $408 = HEAP32[$36>>2]|0;
    $409 = HEAP8[$50>>0]|0;
    $410 = ($409<<24>>24)==(0);
    if ($410) {
     $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $407;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $407;$y$2 = $408;$zone$2 = $zone$0$ph$ph171;
    } else {
     $$4 = 0;
     label = 257;
     break L11;
    }
    break;
   }
   case 4:  {
    HEAP32[$17>>2] = 1;
    $556 = HEAP32[$327>>2]|0;
    $557 = HEAP32[$47>>2]|0;
    $558 = (($557) + ($556))|0;
    HEAP32[$47>>2] = $558;
    $559 = ((($327)) + 4|0);
    $560 = HEAP32[$559>>2]|0;
    $561 = HEAP32[$51>>2]|0;
    $562 = (($561) + ($560))|0;
    HEAP32[$51>>2] = $562;
    $563 = ((($327)) + 8|0);
    $564 = HEAP32[$563>>2]|0;
    HEAP32[$48>>2] = $564;
    $565 = ((($327)) + 12|0);
    $566 = HEAP32[$565>>2]|0;
    HEAP32[$49>>2] = $566;
    $567 = HEAP32[$34>>2]|0;
    $568 = HEAP32[$327>>2]|0;
    $569 = (($568) + ($567))|0;
    $570 = HEAP32[$36>>2]|0;
    $571 = HEAP32[$559>>2]|0;
    $572 = (($571) + ($570))|0;
    $573 = HEAP8[$50>>0]|0;
    $574 = ($573<<24>>24)==(0);
    if ($574) {
     $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $569;$y$2 = $572;$zone$2 = $zone$0$ph$ph171;
    } else {
     $$4 = 0;
     label = 257;
     break L11;
    }
    break;
   }
   case 5:  {
    $575 = HEAP32[$17>>2]|0;
    $576 = $575 & -2;
    $switch34 = ($576|0)==(2);
    do {
     if ($switch34) {
      $577 = HEAP32[$44>>2]|0;
      $578 = ($577|0)==(0|0);
      if (!($578)) {
       $579 = HEAP32[$577>>2]|0;
       $580 = $579&65535;
       $581 = ($580<<16>>16)<(2);
       $582 = $579 >>> 16;
       $583 = $582&65535;
       if ($581) {
        $604 = 0;
       } else {
        $sext$i35 = $579 << 16;
        $584 = $sext$i35 >> 16;
        $585 = (($584) + -2)|0;
        $586 = ((($577)) + 12|0);
        $587 = HEAP32[$586>>2]|0;
        $588 = (($587) + ($585<<1)|0);
        $589 = HEAP16[$588>>1]|0;
        $590 = $589 << 16 >> 16;
        $591 = (($590) + 1)|0;
        $604 = $591;
       }
       $592 = ((($577)) + 2|0);
       $593 = $579 >> 16;
       $594 = ($583<<16>>16)>(1);
       if ($594) {
        $595 = ((($577)) + 4|0);
        $596 = HEAP32[$595>>2]|0;
        $597 = (($596) + ($593<<3)|0);
        $598 = ((($597)) + -8|0);
        $599 = ((($577)) + 8|0);
        $600 = HEAP32[$599>>2]|0;
        $601 = (($600) + ($593)|0);
        $602 = ((($601)) + -1|0);
        $603 = (($596) + ($604<<3)|0);
        $605 = HEAP32[$603>>2]|0;
        $606 = HEAP32[$598>>2]|0;
        $607 = ($605|0)==($606|0);
        if ($607) {
         $608 = (((($596) + ($604<<3)|0)) + 4|0);
         $609 = HEAP32[$608>>2]|0;
         $610 = ((($598)) + 4|0);
         $611 = HEAP32[$610>>2]|0;
         $612 = ($609|0)==($611|0);
         if ($612) {
          $613 = HEAP8[$602>>0]|0;
          $614 = ($613<<24>>24)==(1);
          if ($614) {
           $615 = (($583) + -1)<<16>>16;
           HEAP16[$592>>1] = $615;
           $619 = $615;
          } else {
           $619 = $583;
          }
         } else {
          $619 = $583;
         }
        } else {
         $619 = $583;
        }
       } else {
        $619 = $583;
       }
       $sext2$i36 = $579 << 16;
       $616 = $sext2$i36 >> 16;
       $617 = ($580<<16>>16)>(0);
       if ($617) {
        $618 = $619 << 16 >> 16;
        $620 = (($618) + -1)|0;
        $621 = ($604|0)==($620|0);
        if ($621) {
         $622 = (($580) + -1)<<16>>16;
         HEAP16[$577>>1] = $622;
         $623 = (($619) + -1)<<16>>16;
         HEAP16[$592>>1] = $623;
         break;
        } else {
         $624 = $620&65535;
         $625 = (($616) + -1)|0;
         $626 = ((($577)) + 12|0);
         $627 = HEAP32[$626>>2]|0;
         $628 = (($627) + ($625<<1)|0);
         HEAP16[$628>>1] = $624;
         break;
        }
       }
      }
     }
    } while(0);
    HEAP32[$17>>2] = 1;
    $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $x$0$ph$ph168;$y$2 = $y$0$ph$ph167;$zone$2 = $zone$0$ph$ph171;
    break;
   }
   case 6:  {
    $629 = (_t1_builder_start_point($decoder,$x$0$ph$ph168,$y$0$ph$ph167)|0);
    $630 = ($629|0)==(0);
    if (!($630)) {
     $$4 = $629;
     label = 257;
     break L11;
    }
    $631 = HEAP32[$327>>2]|0;
    $632 = (($631) + ($x$0$ph$ph168))|0;
    $x$1 = $632;$y$1 = $y$0$ph$ph167;
    label = 192;
    break;
   }
   case 7:  {
    $633 = HEAP32[$327>>2]|0;
    $634 = (($633) + ($x$0$ph$ph168))|0;
    $635 = ((($$lcssa266$lcssa)) + 1400|0);
    $636 = HEAP32[$635>>2]|0;
    $637 = ($636|0)==(0);
    if ($637) {
     $638 = HEAP32[$17>>2]|0;
     $639 = ($638|0)==(0);
     if ($639) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     HEAP32[$17>>2] = 2;
     $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $634;$y$2 = $y$0$ph$ph167;$zone$2 = $zone$0$ph$ph171;
    } else {
     $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $634;$y$2 = $y$0$ph$ph167;$zone$2 = $zone$0$ph$ph171;
    }
    break;
   }
   case 8:  {
    $640 = (_t1_builder_start_point($decoder,$x$0$ph$ph168,$y$0$ph$ph167)|0);
    $641 = ($640|0)==(0);
    if (!($641)) {
     $$4 = $640;
     label = 257;
     break L11;
    }
    $642 = HEAP32[$43>>2]|0;
    $643 = ((($642)) + 22|0);
    $644 = HEAP16[$643>>1]|0;
    $645 = $644 << 16 >> 16;
    $646 = ((($642)) + 58|0);
    $647 = HEAP16[$646>>1]|0;
    $648 = $647 << 16 >> 16;
    $649 = (($645) + 3)|0;
    $650 = (($649) + ($648))|0;
    $651 = ((($642)) + 4|0);
    $652 = HEAP32[$651>>2]|0;
    $653 = ($650>>>0)>($652>>>0);
    if ($653) {
     $654 = (_FT_GlyphLoader_CheckPoints($642,3,0)|0);
     $655 = ($654|0)==(0);
     if (!($655)) {
      $$4 = $654;
      label = 257;
      break L11;
     }
    }
    $656 = HEAP32[$327>>2]|0;
    $657 = (($656) + ($x$0$ph$ph168))|0;
    $658 = HEAP32[$44>>2]|0;
    $659 = HEAP8[$45>>0]|0;
    $660 = ($659<<24>>24)==(0);
    if ($660) {
     $$pre$i39 = ((($658)) + 2|0);
     $$pre$phi$i41Z2D = $$pre$i39;$684 = 0;$685 = $658;
    } else {
     $661 = ((($658)) + 4|0);
     $662 = HEAP32[$661>>2]|0;
     $663 = ((($658)) + 2|0);
     $664 = HEAP16[$663>>1]|0;
     $665 = $664 << 16 >> 16;
     $666 = ((($658)) + 8|0);
     $667 = HEAP32[$666>>2]|0;
     $668 = (($667) + ($665)|0);
     $669 = (_FT_RoundFix($657)|0);
     $670 = $669 >> 16;
     $671 = (($662) + ($665<<3)|0);
     HEAP32[$671>>2] = $670;
     $672 = (_FT_RoundFix($y$0$ph$ph167)|0);
     $673 = $672 >> 16;
     $674 = (((($662) + ($665<<3)|0)) + 4|0);
     HEAP32[$674>>2] = $673;
     HEAP8[$668>>0] = 2;
     $$pre199 = HEAP32[$44>>2]|0;
     $$pre200 = HEAP8[$45>>0]|0;
     $$pre$phi$i41Z2D = $663;$684 = $$pre200;$685 = $$pre199;
    }
    $675 = HEAP16[$$pre$phi$i41Z2D>>1]|0;
    $676 = (($675) + 1)<<16>>16;
    HEAP16[$$pre$phi$i41Z2D>>1] = $676;
    $677 = ((($327)) + 4|0);
    $678 = HEAP32[$677>>2]|0;
    $679 = (($678) + ($657))|0;
    $680 = ((($327)) + 8|0);
    $681 = HEAP32[$680>>2]|0;
    $682 = (($681) + ($y$0$ph$ph167))|0;
    $683 = ($684<<24>>24)==(0);
    if ($683) {
     $$pre$i43 = ((($685)) + 2|0);
     $$pre$phi$i45Z2D = $$pre$i43;$706 = 0;$707 = $685;
    } else {
     $686 = ((($685)) + 4|0);
     $687 = HEAP32[$686>>2]|0;
     $688 = ((($685)) + 2|0);
     $689 = HEAP16[$688>>1]|0;
     $690 = $689 << 16 >> 16;
     $691 = ((($685)) + 8|0);
     $692 = HEAP32[$691>>2]|0;
     $693 = (($692) + ($690)|0);
     $694 = (_FT_RoundFix($679)|0);
     $695 = $694 >> 16;
     $696 = (($687) + ($690<<3)|0);
     HEAP32[$696>>2] = $695;
     $697 = (_FT_RoundFix($682)|0);
     $698 = $697 >> 16;
     $699 = (((($687) + ($690<<3)|0)) + 4|0);
     HEAP32[$699>>2] = $698;
     HEAP8[$693>>0] = 2;
     $$pre201 = HEAP32[$44>>2]|0;
     $$pre202 = HEAP8[$45>>0]|0;
     $$pre$phi$i45Z2D = $688;$706 = $$pre202;$707 = $$pre201;
    }
    $700 = HEAP16[$$pre$phi$i45Z2D>>1]|0;
    $701 = (($700) + 1)<<16>>16;
    HEAP16[$$pre$phi$i45Z2D>>1] = $701;
    $702 = ((($327)) + 12|0);
    $703 = HEAP32[$702>>2]|0;
    $704 = (($703) + ($682))|0;
    $705 = ($706<<24>>24)==(0);
    if ($705) {
     $$pre$i47 = ((($707)) + 2|0);
     $$pre$phi$i49Z2D = $$pre$i47;
    } else {
     $708 = ((($707)) + 4|0);
     $709 = HEAP32[$708>>2]|0;
     $710 = ((($707)) + 2|0);
     $711 = HEAP16[$710>>1]|0;
     $712 = $711 << 16 >> 16;
     $713 = ((($707)) + 8|0);
     $714 = HEAP32[$713>>2]|0;
     $715 = (($714) + ($712)|0);
     $716 = (_FT_RoundFix($679)|0);
     $717 = $716 >> 16;
     $718 = (($709) + ($712<<3)|0);
     HEAP32[$718>>2] = $717;
     $719 = (_FT_RoundFix($704)|0);
     $720 = $719 >> 16;
     $721 = (((($709) + ($712<<3)|0)) + 4|0);
     HEAP32[$721>>2] = $720;
     HEAP8[$715>>0] = 1;
     $$pre$phi$i49Z2D = $710;
    }
    $722 = HEAP16[$$pre$phi$i49Z2D>>1]|0;
    $723 = (($722) + 1)<<16>>16;
    HEAP16[$$pre$phi$i49Z2D>>1] = $723;
    $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $679;$y$2 = $704;$zone$2 = $zone$0$ph$ph171;
    break;
   }
   case 9:  {
    $724 = (_t1_builder_start_point($decoder,$x$0$ph$ph168,$y$0$ph$ph167)|0);
    $725 = ($724|0)==(0);
    if (!($725)) {
     $$4 = $724;
     label = 257;
     break L11;
    }
    $726 = HEAP32[$327>>2]|0;
    $727 = (($726) + ($x$0$ph$ph168))|0;
    $728 = ((($327)) + 4|0);
    $729 = HEAP32[$728>>2]|0;
    $730 = (($729) + ($y$0$ph$ph167))|0;
    $x$1 = $727;$y$1 = $730;
    label = 192;
    break;
   }
   case 10:  {
    $764 = HEAP32[$327>>2]|0;
    $765 = (($764) + ($x$0$ph$ph168))|0;
    $766 = ((($327)) + 4|0);
    $767 = HEAP32[$766>>2]|0;
    $768 = (($767) + ($y$0$ph$ph167))|0;
    $769 = ((($$lcssa266$lcssa)) + 1400|0);
    $770 = HEAP32[$769>>2]|0;
    $771 = ($770|0)==(0);
    if ($771) {
     $772 = HEAP32[$17>>2]|0;
     $773 = ($772|0)==(0);
     if ($773) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     HEAP32[$17>>2] = 2;
     $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $765;$y$2 = $768;$zone$2 = $zone$0$ph$ph171;
    } else {
     $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $765;$y$2 = $768;$zone$2 = $zone$0$ph$ph171;
    }
    break;
   }
   case 11:  {
    $774 = (_t1_builder_start_point($decoder,$x$0$ph$ph168,$y$0$ph$ph167)|0);
    $775 = ($774|0)==(0);
    if (!($775)) {
     $$4 = $774;
     label = 257;
     break L11;
    }
    $776 = HEAP32[$43>>2]|0;
    $777 = ((($776)) + 22|0);
    $778 = HEAP16[$777>>1]|0;
    $779 = $778 << 16 >> 16;
    $780 = ((($776)) + 58|0);
    $781 = HEAP16[$780>>1]|0;
    $782 = $781 << 16 >> 16;
    $783 = (($779) + 3)|0;
    $784 = (($783) + ($782))|0;
    $785 = ((($776)) + 4|0);
    $786 = HEAP32[$785>>2]|0;
    $787 = ($784>>>0)>($786>>>0);
    if ($787) {
     $788 = (_FT_GlyphLoader_CheckPoints($776,3,0)|0);
     $789 = ($788|0)==(0);
     if (!($789)) {
      $$4 = $788;
      label = 257;
      break L11;
     }
    }
    $790 = HEAP32[$327>>2]|0;
    $791 = (($790) + ($x$0$ph$ph168))|0;
    $792 = ((($327)) + 4|0);
    $793 = HEAP32[$792>>2]|0;
    $794 = (($793) + ($y$0$ph$ph167))|0;
    $795 = HEAP32[$44>>2]|0;
    $796 = HEAP8[$45>>0]|0;
    $797 = ($796<<24>>24)==(0);
    if ($797) {
     $$pre$i52 = ((($795)) + 2|0);
     $$pre$phi$i54Z2D = $$pre$i52;$821 = 0;$822 = $795;
    } else {
     $798 = ((($795)) + 4|0);
     $799 = HEAP32[$798>>2]|0;
     $800 = ((($795)) + 2|0);
     $801 = HEAP16[$800>>1]|0;
     $802 = $801 << 16 >> 16;
     $803 = ((($795)) + 8|0);
     $804 = HEAP32[$803>>2]|0;
     $805 = (($804) + ($802)|0);
     $806 = (_FT_RoundFix($791)|0);
     $807 = $806 >> 16;
     $808 = (($799) + ($802<<3)|0);
     HEAP32[$808>>2] = $807;
     $809 = (_FT_RoundFix($794)|0);
     $810 = $809 >> 16;
     $811 = (((($799) + ($802<<3)|0)) + 4|0);
     HEAP32[$811>>2] = $810;
     HEAP8[$805>>0] = 2;
     $$pre195 = HEAP32[$44>>2]|0;
     $$pre196 = HEAP8[$45>>0]|0;
     $$pre$phi$i54Z2D = $800;$821 = $$pre196;$822 = $$pre195;
    }
    $812 = HEAP16[$$pre$phi$i54Z2D>>1]|0;
    $813 = (($812) + 1)<<16>>16;
    HEAP16[$$pre$phi$i54Z2D>>1] = $813;
    $814 = ((($327)) + 8|0);
    $815 = HEAP32[$814>>2]|0;
    $816 = (($815) + ($791))|0;
    $817 = ((($327)) + 12|0);
    $818 = HEAP32[$817>>2]|0;
    $819 = (($818) + ($794))|0;
    $820 = ($821<<24>>24)==(0);
    if ($820) {
     $$pre$i56 = ((($822)) + 2|0);
     $$pre$phi$i58Z2D = $$pre$i56;$846 = 0;$847 = $822;
    } else {
     $823 = ((($822)) + 4|0);
     $824 = HEAP32[$823>>2]|0;
     $825 = ((($822)) + 2|0);
     $826 = HEAP16[$825>>1]|0;
     $827 = $826 << 16 >> 16;
     $828 = ((($822)) + 8|0);
     $829 = HEAP32[$828>>2]|0;
     $830 = (($829) + ($827)|0);
     $831 = (_FT_RoundFix($816)|0);
     $832 = $831 >> 16;
     $833 = (($824) + ($827<<3)|0);
     HEAP32[$833>>2] = $832;
     $834 = (_FT_RoundFix($819)|0);
     $835 = $834 >> 16;
     $836 = (((($824) + ($827<<3)|0)) + 4|0);
     HEAP32[$836>>2] = $835;
     HEAP8[$830>>0] = 2;
     $$pre197 = HEAP32[$44>>2]|0;
     $$pre198 = HEAP8[$45>>0]|0;
     $$pre$phi$i58Z2D = $825;$846 = $$pre198;$847 = $$pre197;
    }
    $837 = HEAP16[$$pre$phi$i58Z2D>>1]|0;
    $838 = (($837) + 1)<<16>>16;
    HEAP16[$$pre$phi$i58Z2D>>1] = $838;
    $839 = ((($327)) + 16|0);
    $840 = HEAP32[$839>>2]|0;
    $841 = (($840) + ($816))|0;
    $842 = ((($327)) + 20|0);
    $843 = HEAP32[$842>>2]|0;
    $844 = (($843) + ($819))|0;
    $845 = ($846<<24>>24)==(0);
    if ($845) {
     $$pre$i60 = ((($847)) + 2|0);
     $$pre$phi$i62Z2D = $$pre$i60;
    } else {
     $848 = ((($847)) + 4|0);
     $849 = HEAP32[$848>>2]|0;
     $850 = ((($847)) + 2|0);
     $851 = HEAP16[$850>>1]|0;
     $852 = $851 << 16 >> 16;
     $853 = ((($847)) + 8|0);
     $854 = HEAP32[$853>>2]|0;
     $855 = (($854) + ($852)|0);
     $856 = (_FT_RoundFix($841)|0);
     $857 = $856 >> 16;
     $858 = (($849) + ($852<<3)|0);
     HEAP32[$858>>2] = $857;
     $859 = (_FT_RoundFix($844)|0);
     $860 = $859 >> 16;
     $861 = (((($849) + ($852<<3)|0)) + 4|0);
     HEAP32[$861>>2] = $860;
     HEAP8[$855>>0] = 1;
     $$pre$phi$i62Z2D = $850;
    }
    $862 = HEAP16[$$pre$phi$i62Z2D>>1]|0;
    $863 = (($862) + 1)<<16>>16;
    HEAP16[$$pre$phi$i62Z2D>>1] = $863;
    $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $841;$y$2 = $844;$zone$2 = $zone$0$ph$ph171;
    break;
   }
   case 12:  {
    $864 = (_t1_builder_start_point($decoder,$x$0$ph$ph168,$y$0$ph$ph167)|0);
    $865 = ($864|0)==(0);
    if (!($865)) {
     $$4 = $864;
     label = 257;
     break L11;
    }
    $866 = HEAP32[$43>>2]|0;
    $867 = ((($866)) + 22|0);
    $868 = HEAP16[$867>>1]|0;
    $869 = $868 << 16 >> 16;
    $870 = ((($866)) + 58|0);
    $871 = HEAP16[$870>>1]|0;
    $872 = $871 << 16 >> 16;
    $873 = (($869) + 3)|0;
    $874 = (($873) + ($872))|0;
    $875 = ((($866)) + 4|0);
    $876 = HEAP32[$875>>2]|0;
    $877 = ($874>>>0)>($876>>>0);
    if ($877) {
     $878 = (_FT_GlyphLoader_CheckPoints($866,3,0)|0);
     $879 = ($878|0)==(0);
     if (!($879)) {
      $$4 = $878;
      label = 257;
      break L11;
     }
    }
    $880 = HEAP32[$327>>2]|0;
    $881 = (($880) + ($y$0$ph$ph167))|0;
    $882 = HEAP32[$44>>2]|0;
    $883 = HEAP8[$45>>0]|0;
    $884 = ($883<<24>>24)==(0);
    if ($884) {
     $$pre$i65 = ((($882)) + 2|0);
     $$pre$phi$i67Z2D = $$pre$i65;$908 = 0;$909 = $882;
    } else {
     $885 = ((($882)) + 4|0);
     $886 = HEAP32[$885>>2]|0;
     $887 = ((($882)) + 2|0);
     $888 = HEAP16[$887>>1]|0;
     $889 = $888 << 16 >> 16;
     $890 = ((($882)) + 8|0);
     $891 = HEAP32[$890>>2]|0;
     $892 = (($891) + ($889)|0);
     $893 = (_FT_RoundFix($x$0$ph$ph168)|0);
     $894 = $893 >> 16;
     $895 = (($886) + ($889<<3)|0);
     HEAP32[$895>>2] = $894;
     $896 = (_FT_RoundFix($881)|0);
     $897 = $896 >> 16;
     $898 = (((($886) + ($889<<3)|0)) + 4|0);
     HEAP32[$898>>2] = $897;
     HEAP8[$892>>0] = 2;
     $$pre191 = HEAP32[$44>>2]|0;
     $$pre192 = HEAP8[$45>>0]|0;
     $$pre$phi$i67Z2D = $887;$908 = $$pre192;$909 = $$pre191;
    }
    $899 = HEAP16[$$pre$phi$i67Z2D>>1]|0;
    $900 = (($899) + 1)<<16>>16;
    HEAP16[$$pre$phi$i67Z2D>>1] = $900;
    $901 = ((($327)) + 4|0);
    $902 = HEAP32[$901>>2]|0;
    $903 = (($902) + ($x$0$ph$ph168))|0;
    $904 = ((($327)) + 8|0);
    $905 = HEAP32[$904>>2]|0;
    $906 = (($905) + ($881))|0;
    $907 = ($908<<24>>24)==(0);
    if ($907) {
     $$pre$i69 = ((($909)) + 2|0);
     $$pre$phi$i71Z2D = $$pre$i69;$930 = 0;$931 = $909;
    } else {
     $910 = ((($909)) + 4|0);
     $911 = HEAP32[$910>>2]|0;
     $912 = ((($909)) + 2|0);
     $913 = HEAP16[$912>>1]|0;
     $914 = $913 << 16 >> 16;
     $915 = ((($909)) + 8|0);
     $916 = HEAP32[$915>>2]|0;
     $917 = (($916) + ($914)|0);
     $918 = (_FT_RoundFix($903)|0);
     $919 = $918 >> 16;
     $920 = (($911) + ($914<<3)|0);
     HEAP32[$920>>2] = $919;
     $921 = (_FT_RoundFix($906)|0);
     $922 = $921 >> 16;
     $923 = (((($911) + ($914<<3)|0)) + 4|0);
     HEAP32[$923>>2] = $922;
     HEAP8[$917>>0] = 2;
     $$pre193 = HEAP32[$44>>2]|0;
     $$pre194 = HEAP8[$45>>0]|0;
     $$pre$phi$i71Z2D = $912;$930 = $$pre194;$931 = $$pre193;
    }
    $924 = HEAP16[$$pre$phi$i71Z2D>>1]|0;
    $925 = (($924) + 1)<<16>>16;
    HEAP16[$$pre$phi$i71Z2D>>1] = $925;
    $926 = ((($327)) + 12|0);
    $927 = HEAP32[$926>>2]|0;
    $928 = (($927) + ($903))|0;
    $929 = ($930<<24>>24)==(0);
    if ($929) {
     $$pre$i73 = ((($931)) + 2|0);
     $$pre$phi$i75Z2D = $$pre$i73;
    } else {
     $932 = ((($931)) + 4|0);
     $933 = HEAP32[$932>>2]|0;
     $934 = ((($931)) + 2|0);
     $935 = HEAP16[$934>>1]|0;
     $936 = $935 << 16 >> 16;
     $937 = ((($931)) + 8|0);
     $938 = HEAP32[$937>>2]|0;
     $939 = (($938) + ($936)|0);
     $940 = (_FT_RoundFix($928)|0);
     $941 = $940 >> 16;
     $942 = (($933) + ($936<<3)|0);
     HEAP32[$942>>2] = $941;
     $943 = (_FT_RoundFix($906)|0);
     $944 = $943 >> 16;
     $945 = (((($933) + ($936<<3)|0)) + 4|0);
     HEAP32[$945>>2] = $944;
     HEAP8[$939>>0] = 1;
     $$pre$phi$i75Z2D = $934;
    }
    $946 = HEAP16[$$pre$phi$i75Z2D>>1]|0;
    $947 = (($946) + 1)<<16>>16;
    HEAP16[$$pre$phi$i75Z2D>>1] = $947;
    $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $928;$y$2 = $906;$zone$2 = $zone$0$ph$ph171;
    break;
   }
   case 13:  {
    $948 = (_t1_builder_start_point($decoder,$x$0$ph$ph168,$y$0$ph$ph167)|0);
    $949 = ($948|0)==(0);
    if (!($949)) {
     $$4 = $948;
     label = 257;
     break L11;
    }
    $950 = HEAP32[$327>>2]|0;
    $951 = (($950) + ($y$0$ph$ph167))|0;
    $x$1 = $x$0$ph$ph168;$y$1 = $951;
    label = 192;
    break;
   }
   case 14:  {
    $952 = HEAP32[$327>>2]|0;
    $953 = (($952) + ($y$0$ph$ph167))|0;
    $954 = ((($$lcssa266$lcssa)) + 1400|0);
    $955 = HEAP32[$954>>2]|0;
    $956 = ($955|0)==(0);
    if ($956) {
     $957 = HEAP32[$17>>2]|0;
     $958 = ($957|0)==(0);
     if ($958) {
      $$4 = 160;
      label = 257;
      break L11;
     }
     HEAP32[$17>>2] = 2;
     $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $x$0$ph$ph168;$y$2 = $953;$zone$2 = $zone$0$ph$ph171;
    } else {
     $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $x$0$ph$ph168;$y$2 = $953;$zone$2 = $zone$0$ph$ph171;
    }
    break;
   }
   case 20:  {
    $959 = HEAP32[$327>>2]|0;
    $960 = ((($327)) + 4|0);
    $961 = HEAP32[$960>>2]|0;
    $962 = (_FT_DivFix($959,$961)|0);
    HEAP32[$327>>2] = $962;
    $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = 0;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $960;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $x$0$ph$ph168;$y$2 = $y$0$ph$ph167;$zone$2 = $zone$0$ph$ph171;
    break;
   }
   case 22:  {
    $963 = HEAP32[$327>>2]|0;
    $964 = $963 >> 16;
    $965 = ($964|0)<(0);
    if ($965) {
     $$4 = 160;
     label = 257;
     break L11;
    }
    $966 = ((($$lcssa266$lcssa)) + 1364|0);
    $967 = HEAP32[$966>>2]|0;
    $968 = ($964|0)<($967|0);
    if (!($968)) {
     $$4 = 160;
     label = 257;
     break L11;
    }
    $969 = ((($$lcssa266$lcssa)) + 1140|0);
    $970 = $zone$0$ph$ph171;
    $971 = $969;
    $972 = (($970) - ($971))|0;
    $973 = ($972|0)>(180);
    if ($973) {
     $$4 = 160;
     label = 257;
     break L11;
    }
    HEAP32[$zone$0$ph$ph171>>2] = $ip$2$lcssa$lcssa;
    $974 = ((($zone$0$ph$ph171)) + 12|0);
    $975 = HEAP32[$0>>2]|0;
    $976 = ((($975)) + 1368|0);
    $977 = HEAP32[$976>>2]|0;
    $978 = (($977) + ($964<<2)|0);
    $979 = HEAP32[$978>>2]|0;
    $980 = ((($zone$0$ph$ph171)) + 16|0);
    HEAP32[$980>>2] = $979;
    $981 = HEAP32[$0>>2]|0;
    $982 = ((($981)) + 1372|0);
    $983 = HEAP32[$982>>2]|0;
    $984 = ($983|0)==(0|0);
    if ($984) {
     $989 = ((($981)) + 1360|0);
     $990 = HEAP32[$989>>2]|0;
     $991 = ($990|0)<(0);
     $$32 = $991 ? 0 : $990;
     $992 = (($979) + ($$32)|0);
     HEAP32[$980>>2] = $992;
     $993 = (($964) + 1)|0;
     $994 = HEAP32[$0>>2]|0;
     $995 = ((($994)) + 1368|0);
     $996 = HEAP32[$995>>2]|0;
     $997 = (($996) + ($993<<2)|0);
     $998 = HEAP32[$997>>2]|0;
     $999 = ((($zone$0$ph$ph171)) + 20|0);
     HEAP32[$999>>2] = $998;
     $1000 = $998;
     $1002 = $992;$1044 = $1000;
    } else {
     $985 = (($983) + ($964<<2)|0);
     $986 = HEAP32[$985>>2]|0;
     $987 = (($979) + ($986)|0);
     $988 = ((($zone$0$ph$ph171)) + 20|0);
     HEAP32[$988>>2] = $987;
     $1002 = $979;$1044 = $987;
    }
    $1001 = $1002;
    HEAP32[$974>>2] = $1001;
    $1003 = ($1002|0)==(0|0);
    if ($1003) {
     $$4 = 160;
     label = 257;
     break L11;
    }
    $1004 = HEAP32[$0>>2]|0;
    $1005 = ((($1004)) + 1344|0);
    HEAP32[$1005>>2] = $974;
    $ip$4 = $1002;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $1044;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $x$0$ph$ph168;$y$2 = $y$0$ph$ph167;$zone$2 = $974;
    break;
   }
   case 23:  {
    $1006 = ($known_othersubr_result_cnt$0$ph159$lcssa279|0)>(0);
    if ($1006) {
     $1007 = (($known_othersubr_result_cnt$0$ph159$lcssa279) + -1)|0;
     $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $1007;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $x$0$ph$ph168;$y$2 = $y$0$ph$ph167;$zone$2 = $zone$0$ph$ph171;
     break L131;
    }
    $1008 = ($unknown_othersubr_result_cnt$1$lcssa$lcssa|0)==(0);
    if ($1008) {
     $$4 = 160;
     label = 257;
     break L11;
    }
    $1009 = (($unknown_othersubr_result_cnt$1$lcssa$lcssa) + -1)|0;
    $1010 = ((($327)) + 4|0);
    $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $1010;$unknown_othersubr_result_cnt$4 = $1009;$x$2 = $x$0$ph$ph168;$y$2 = $y$0$ph$ph167;$zone$2 = $zone$0$ph$ph171;
    break;
   }
   case 24:  {
    $1011 = ((($$lcssa266$lcssa)) + 1140|0);
    $1012 = ($zone$0$ph$ph171>>>0)>($1011>>>0);
    if (!($1012)) {
     $$4 = 160;
     label = 257;
     break L11;
    }
    $1013 = ((($zone$0$ph$ph171)) + -12|0);
    $1014 = HEAP32[$1013>>2]|0;
    $1015 = ((($zone$0$ph$ph171)) + -4|0);
    $1016 = HEAP32[$1015>>2]|0;
    $1017 = ((($$lcssa266$lcssa)) + 1344|0);
    HEAP32[$1017>>2] = $1013;
    $ip$4 = $1014;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $1016;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $x$0$ph$ph168;$y$2 = $y$0$ph$ph167;$zone$2 = $1013;
    break;
   }
   case 26: case 15:  {
    $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $x$0$ph$ph168;$y$2 = $y$0$ph$ph167;$zone$2 = $zone$0$ph$ph171;
    break;
   }
   case 16:  {
    if ($38) {
     $1018 = HEAP32[$52>>2]|0;
     $1019 = HEAP32[$19>>2]|0;
     FUNCTION_TABLE_viii[$1018 & 8191]($1019,1,$327);
     $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $x$0$ph$ph168;$y$2 = $y$0$ph$ph167;$zone$2 = $zone$0$ph$ph171;
    } else {
     $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $x$0$ph$ph168;$y$2 = $y$0$ph$ph167;$zone$2 = $zone$0$ph$ph171;
    }
    break;
   }
   case 17:  {
    if ($38) {
     $1020 = HEAP32[$53>>2]|0;
     $1021 = HEAP32[$19>>2]|0;
     FUNCTION_TABLE_viii[$1020 & 8191]($1021,1,$327);
     $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $x$0$ph$ph168;$y$2 = $y$0$ph$ph167;$zone$2 = $zone$0$ph$ph171;
    } else {
     $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $x$0$ph$ph168;$y$2 = $y$0$ph$ph167;$zone$2 = $zone$0$ph$ph171;
    }
    break;
   }
   case 18:  {
    if ($38) {
     $1022 = HEAP32[$327>>2]|0;
     $1023 = (($1022) + ($orig_x$0$ph$ph166))|0;
     HEAP32[$327>>2] = $1023;
     $1024 = HEAP32[$52>>2]|0;
     $1025 = HEAP32[$19>>2]|0;
     FUNCTION_TABLE_viii[$1024 & 8191]($1025,0,$327);
     $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $x$0$ph$ph168;$y$2 = $y$0$ph$ph167;$zone$2 = $zone$0$ph$ph171;
    } else {
     $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $x$0$ph$ph168;$y$2 = $y$0$ph$ph167;$zone$2 = $zone$0$ph$ph171;
    }
    break;
   }
   case 19:  {
    if ($38) {
     $1026 = HEAP32[$327>>2]|0;
     $1027 = (($1026) + ($orig_x$0$ph$ph166))|0;
     HEAP32[$327>>2] = $1027;
     $1028 = ((($327)) + 8|0);
     $1029 = HEAP32[$1028>>2]|0;
     $1030 = (($1029) + ($orig_x$0$ph$ph166))|0;
     HEAP32[$1028>>2] = $1030;
     $1031 = ((($327)) + 16|0);
     $1032 = HEAP32[$1031>>2]|0;
     $1033 = (($1032) + ($orig_x$0$ph$ph166))|0;
     HEAP32[$1031>>2] = $1033;
     $1034 = HEAP32[$53>>2]|0;
     $1035 = HEAP32[$19>>2]|0;
     FUNCTION_TABLE_viii[$1034 & 8191]($1035,0,$327);
     $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $x$0$ph$ph168;$y$2 = $y$0$ph$ph167;$zone$2 = $zone$0$ph$ph171;
    } else {
     $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $x$0$ph$ph168;$y$2 = $y$0$ph$ph167;$zone$2 = $zone$0$ph$ph171;
    }
    break;
   }
   case 25:  {
    $1036 = HEAP32[$327>>2]|0;
    $1037 = ((($327)) + 4|0);
    $1038 = HEAP32[$1037>>2]|0;
    $1039 = ((($$lcssa266$lcssa)) + 1400|0);
    HEAP32[$1039>>2] = 0;
    $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $1036;$y$2 = $1038;$zone$2 = $zone$0$ph$ph171;
    break;
   }
   default: {
    $$4 = 160;
    label = 257;
    break L11;
   }
   }
  } while(0);
  if ((label|0) == 192) {
   label = 0;
   $731 = HEAP32[$43>>2]|0;
   $732 = ((($731)) + 22|0);
   $733 = HEAP16[$732>>1]|0;
   $734 = $733 << 16 >> 16;
   $735 = ((($731)) + 58|0);
   $736 = HEAP16[$735>>1]|0;
   $737 = $736 << 16 >> 16;
   $738 = (($734) + 1)|0;
   $739 = (($738) + ($737))|0;
   $740 = ((($731)) + 4|0);
   $741 = HEAP32[$740>>2]|0;
   $742 = ($739>>>0)>($741>>>0);
   if ($742) {
    $743 = (_FT_GlyphLoader_CheckPoints($731,1,0)|0);
    $744 = ($743|0)==(0);
    if (!($744)) {
     $$4 = $743;
     label = 257;
     break;
    }
   }
   $745 = HEAP32[$44>>2]|0;
   $746 = HEAP8[$45>>0]|0;
   $747 = ($746<<24>>24)==(0);
   if ($747) {
    $$pre$i$i = ((($745)) + 2|0);
    $$pre$phi$i$iZ2D = $$pre$i$i;
   } else {
    $748 = ((($745)) + 4|0);
    $749 = HEAP32[$748>>2]|0;
    $750 = ((($745)) + 2|0);
    $751 = HEAP16[$750>>1]|0;
    $752 = $751 << 16 >> 16;
    $753 = ((($745)) + 8|0);
    $754 = HEAP32[$753>>2]|0;
    $755 = (($754) + ($752)|0);
    $756 = (_FT_RoundFix($x$1)|0);
    $757 = $756 >> 16;
    $758 = (($749) + ($752<<3)|0);
    HEAP32[$758>>2] = $757;
    $759 = (_FT_RoundFix($y$1)|0);
    $760 = $759 >> 16;
    $761 = (((($749) + ($752<<3)|0)) + 4|0);
    HEAP32[$761>>2] = $760;
    HEAP8[$755>>0] = 1;
    $$pre$phi$i$iZ2D = $750;
   }
   $762 = HEAP16[$$pre$phi$i$iZ2D>>1]|0;
   $763 = (($762) + 1)<<16>>16;
   HEAP16[$$pre$phi$i$iZ2D>>1] = $763;
   $ip$4 = $ip$2$lcssa$lcssa;$known_othersubr_result_cnt$4 = $known_othersubr_result_cnt$0$ph159$lcssa279;$large_int$3 = $large_int$1$$lcssa$lcssa;$limit$2 = $limit$0$ph$ph169;$orig_x$1 = $orig_x$0$ph$ph166;$top$1 = $327;$unknown_othersubr_result_cnt$4 = $unknown_othersubr_result_cnt$1$lcssa$lcssa;$x$2 = $x$1;$y$2 = $y$1;$zone$2 = $zone$0$ph$ph171;
  }
  $1040 = HEAP32[$0>>2]|0;
  $1041 = ((($1040)) + 1136|0);
  HEAP32[$1041>>2] = $top$1;
  $1042 = ($ip$4>>>0)<($limit$2>>>0);
  if ($1042) {
   $ip$0$ph$ph170 = $ip$4;$known_othersubr_result_cnt$0$ph$ph165 = $known_othersubr_result_cnt$4;$large_int$0$ph$ph163 = $large_int$3;$limit$0$ph$ph169 = $limit$2;$orig_x$0$ph$ph166 = $orig_x$1;$unknown_othersubr_result_cnt$0$ph$ph164 = $unknown_othersubr_result_cnt$4;$x$0$ph$ph168 = $x$2;$y$0$ph$ph167 = $y$2;$zone$0$ph$ph171 = $zone$2;
  } else {
   $$4 = 0;
   label = 257;
   break;
  }
 }
 if ((label|0) == 114) {
  $328 = HEAP32[$44>>2]|0;
  $329 = ($328|0)==(0|0);
  do {
   if (!($329)) {
    $330 = HEAP32[$328>>2]|0;
    $331 = $330&65535;
    $332 = ($331<<16>>16)<(2);
    $333 = $330 >>> 16;
    $334 = $333&65535;
    if ($332) {
     $355 = 0;
    } else {
     $sext$i = $330 << 16;
     $335 = $sext$i >> 16;
     $336 = (($335) + -2)|0;
     $337 = ((($328)) + 12|0);
     $338 = HEAP32[$337>>2]|0;
     $339 = (($338) + ($336<<1)|0);
     $340 = HEAP16[$339>>1]|0;
     $341 = $340 << 16 >> 16;
     $342 = (($341) + 1)|0;
     $355 = $342;
    }
    $343 = ((($328)) + 2|0);
    $344 = $330 >> 16;
    $345 = ($334<<16>>16)>(1);
    if ($345) {
     $346 = ((($328)) + 4|0);
     $347 = HEAP32[$346>>2]|0;
     $348 = (($347) + ($344<<3)|0);
     $349 = ((($348)) + -8|0);
     $350 = ((($328)) + 8|0);
     $351 = HEAP32[$350>>2]|0;
     $352 = (($351) + ($344)|0);
     $353 = ((($352)) + -1|0);
     $354 = (($347) + ($355<<3)|0);
     $356 = HEAP32[$354>>2]|0;
     $357 = HEAP32[$349>>2]|0;
     $358 = ($356|0)==($357|0);
     if ($358) {
      $359 = (((($347) + ($355<<3)|0)) + 4|0);
      $360 = HEAP32[$359>>2]|0;
      $361 = ((($349)) + 4|0);
      $362 = HEAP32[$361>>2]|0;
      $363 = ($360|0)==($362|0);
      if ($363) {
       $364 = HEAP8[$353>>0]|0;
       $365 = ($364<<24>>24)==(1);
       if ($365) {
        $366 = (($334) + -1)<<16>>16;
        HEAP16[$343>>1] = $366;
        $370 = $366;
       } else {
        $370 = $334;
       }
      } else {
       $370 = $334;
      }
     } else {
      $370 = $334;
     }
    } else {
     $370 = $334;
    }
    $sext2$i = $330 << 16;
    $367 = $sext2$i >> 16;
    $368 = ($331<<16>>16)>(0);
    if ($368) {
     $369 = $370 << 16 >> 16;
     $371 = (($369) + -1)|0;
     $372 = ($355|0)==($371|0);
     if ($372) {
      $373 = (($331) + -1)<<16>>16;
      HEAP16[$328>>1] = $373;
      $374 = (($370) + -1)<<16>>16;
      HEAP16[$343>>1] = $374;
      break;
     } else {
      $375 = $371&65535;
      $376 = (($367) + -1)|0;
      $377 = ((($328)) + 12|0);
      $378 = HEAP32[$377>>2]|0;
      $379 = (($378) + ($376<<1)|0);
      HEAP16[$379>>1] = $375;
      break;
     }
    }
   }
  } while(0);
  if ($38) {
   $380 = ((($19)) + 8|0);
   $381 = HEAP32[$380>>2]|0;
   $382 = HEAP32[$19>>2]|0;
   $383 = ((($328)) + 2|0);
   $384 = HEAP16[$383>>1]|0;
   $385 = $384 << 16 >> 16;
   $386 = (FUNCTION_TABLE_iii[$381 & 8191]($382,$385)|0);
   $387 = ($386|0)==(0);
   if (!($387)) {
    $$4 = 160;
    STACKTOP = sp;return ($$4|0);
   }
   $388 = ((($19)) + 24|0);
   $389 = HEAP32[$388>>2]|0;
   $390 = HEAP32[$19>>2]|0;
   $391 = HEAP32[$44>>2]|0;
   $392 = ((($decoder)) + 76|0);
   $393 = HEAP32[$392>>2]|0;
   $394 = HEAP32[$0>>2]|0;
   $395 = ((($394)) + 1468|0);
   $396 = HEAP32[$395>>2]|0;
   $397 = (FUNCTION_TABLE_iiiii[$389 & 8191]($390,$391,$393,$396)|0);
   $398 = ($397|0)==(0);
   if (!($398)) {
    $$4 = $397;
    STACKTOP = sp;return ($$4|0);
   }
  }
  $399 = HEAP32[$43>>2]|0;
  _FT_GlyphLoader_Add($399);
  $$4 = 0;
  STACKTOP = sp;return ($$4|0);
 }
 else if ((label|0) == 131) {
  $411 = HEAP32[$$lcssa282>>2]|0;
  $412 = ((($$lcssa282)) + 4|0);
  $413 = HEAP32[$412>>2]|0;
  $414 = ((($$lcssa282)) + 8|0);
  $415 = HEAP32[$414>>2]|0;
  $416 = ((($$lcssa282)) + 12|0);
  $417 = HEAP32[$416>>2]|0;
  $418 = $417 >> 16;
  $419 = ((($$lcssa282)) + 16|0);
  $420 = HEAP32[$419>>2]|0;
  $421 = $420 >> 16;
  $422 = ((($$lcssa266$lcssa$lcssa270)) + 4|0);
  $423 = HEAP32[$422>>2]|0;
  $424 = ((($$lcssa266$lcssa$lcssa270)) + 1496|0);
  $425 = HEAP8[$424>>0]|0;
  $426 = ($425<<24>>24)==(0);
  if (!($426)) {
   $$4 = 160;
   STACKTOP = sp;return ($$4|0);
  }
  $427 = ((($$lcssa266$lcssa$lcssa270)) + 70|0);
  $428 = HEAP8[$427>>0]|0;
  $429 = ($428<<24>>24)==(0);
  if (!($429)) {
   $$4 = 160;
   STACKTOP = sp;return ($$4|0);
  }
  $430 = ((($$lcssa266$lcssa$lcssa270)) + 32|0);
  $431 = HEAP32[$430>>2]|0;
  $432 = (($431) + ($413))|0;
  $433 = ((($$lcssa266$lcssa$lcssa270)) + 1356|0);
  $434 = HEAP32[$433>>2]|0;
  $435 = ($434|0)==(0|0);
  $436 = ((($423)) + 128|0);
  $437 = HEAP32[$436>>2]|0;
  $438 = ((($437)) + 48|0);
  $439 = HEAP32[$438>>2]|0;
  $440 = ($439|0)==(0|0);
  L288: do {
   if ($435) {
    if ($440) {
     $$4 = 160;
     STACKTOP = sp;return ($$4|0);
    } else {
     $$lcssa266268 = $$lcssa266$lcssa$lcssa270;$achar_index$0$i = $421;$bchar_index$0$i = $418;
    }
   } else {
    if ($440) {
     $441 = ((($$lcssa266$lcssa$lcssa270)) + 1348|0);
     $442 = HEAP32[$441>>2]|0;
     $443 = ($418>>>0)>(255);
     L293: do {
      if ($443) {
       $$2$i$i = -1;
      } else {
       $444 = ((($442)) + 20|0);
       $445 = HEAP32[$444>>2]|0;
       $446 = ((($442)) + 24|0);
       $447 = HEAP32[$446>>2]|0;
       $448 = (($447) + ($418<<1)|0);
       $449 = HEAP16[$448>>1]|0;
       $450 = $449&65535;
       $451 = (FUNCTION_TABLE_ii[$445 & 8191]($450)|0);
       $452 = ((($$lcssa266$lcssa$lcssa270)) + 1352|0);
       $453 = HEAP32[$452>>2]|0;
       $454 = ($453|0)==(0);
       if ($454) {
        $$2$i$i = -1;
       } else {
        $455 = HEAP32[$433>>2]|0;
        $n$04$i$i = 0;
        while(1) {
         $456 = (($455) + ($n$04$i$i<<2)|0);
         $457 = HEAP32[$456>>2]|0;
         $458 = ($457|0)==(0|0);
         if (!($458)) {
          $459 = HEAP8[$457>>0]|0;
          $460 = HEAP8[$451>>0]|0;
          $461 = ($459<<24>>24)==($460<<24>>24);
          if ($461) {
           $462 = (_strcmp($457,$451)|0);
           $463 = ($462|0)==(0);
           if ($463) {
            $$2$i$i = $n$04$i$i;
            break L293;
           }
          }
         }
         $464 = (($n$04$i$i) + 1)|0;
         $465 = ($464>>>0)<($453>>>0);
         if ($465) {
          $n$04$i$i = $464;
         } else {
          $$2$i$i = -1;
          break;
         }
        }
       }
      }
     } while(0);
     $466 = HEAP32[$441>>2]|0;
     $467 = ($421>>>0)>(255);
     if ($467) {
      $$lcssa266268 = $$lcssa266$lcssa$lcssa270;$achar_index$0$i = -1;$bchar_index$0$i = $$2$i$i;
     } else {
      $468 = ((($466)) + 20|0);
      $469 = HEAP32[$468>>2]|0;
      $470 = ((($466)) + 24|0);
      $471 = HEAP32[$470>>2]|0;
      $472 = (($471) + ($421<<1)|0);
      $473 = HEAP16[$472>>1]|0;
      $474 = $473&65535;
      $475 = (FUNCTION_TABLE_ii[$469 & 8191]($474)|0);
      $476 = ((($$lcssa266$lcssa$lcssa270)) + 1352|0);
      $477 = HEAP32[$476>>2]|0;
      $478 = ($477|0)==(0);
      if ($478) {
       $$lcssa266268 = $$lcssa266$lcssa$lcssa270;$achar_index$0$i = -1;$bchar_index$0$i = $$2$i$i;
      } else {
       $479 = HEAP32[$433>>2]|0;
       $n$04$i4$i = 0;
       while(1) {
        $480 = (($479) + ($n$04$i4$i<<2)|0);
        $481 = HEAP32[$480>>2]|0;
        $482 = ($481|0)==(0|0);
        if (!($482)) {
         $483 = HEAP8[$481>>0]|0;
         $484 = HEAP8[$475>>0]|0;
         $485 = ($483<<24>>24)==($484<<24>>24);
         if ($485) {
          $486 = (_strcmp($481,$475)|0);
          $487 = ($486|0)==(0);
          if ($487) {
           $$lcssa266268 = $$lcssa266$lcssa$lcssa270;$achar_index$0$i = $n$04$i4$i;$bchar_index$0$i = $$2$i$i;
           break L288;
          }
         }
        }
        $488 = (($n$04$i4$i) + 1)|0;
        $489 = ($488>>>0)<($477>>>0);
        if ($489) {
         $n$04$i4$i = $488;
        } else {
         $$lcssa266268 = $$lcssa266$lcssa$lcssa270;$achar_index$0$i = -1;$bchar_index$0$i = $$2$i$i;
         break;
        }
       }
      }
     }
    } else {
     $$lcssa266268 = $$lcssa266$lcssa$lcssa270;$achar_index$0$i = $421;$bchar_index$0$i = $418;
    }
   }
  } while(0);
  $490 = $bchar_index$0$i | $achar_index$0$i;
  $491 = ($490|0)<(0);
  if ($491) {
   $$4 = 160;
   STACKTOP = sp;return ($$4|0);
  }
  $492 = ((($$lcssa266268)) + 69|0);
  $493 = HEAP8[$492>>0]|0;
  $494 = ($493<<24>>24)==(0);
  if (!($494)) {
   $495 = ((($$lcssa266268)) + 8|0);
   $496 = HEAP32[$495>>2]|0;
   $497 = ((($496)) + 156|0);
   $498 = HEAP32[$497>>2]|0;
   $499 = HEAP32[$498>>2]|0;
   $500 = (_FT_GlyphLoader_CheckSubGlyphs($499,2)|0);
   $501 = ($500|0)==(0);
   if (!($501)) {
    $$4 = $500;
    STACKTOP = sp;return ($$4|0);
   }
   $502 = ((($499)) + 88|0);
   $503 = HEAP32[$502>>2]|0;
   HEAP32[$503>>2] = $bchar_index$0$i;
   $504 = ((($503)) + 4|0);
   HEAP16[$504>>1] = 514;
   $505 = ((($503)) + 8|0);
   HEAP32[$505>>2] = 0;
   $506 = ((($503)) + 12|0);
   HEAP32[$506>>2] = 0;
   $507 = ((($503)) + 32|0);
   HEAP32[$507>>2] = $achar_index$0$i;
   $508 = ((($503)) + 36|0);
   HEAP16[$508>>1] = 2;
   $509 = (($432) - ($411))|0;
   $510 = (_FT_RoundFix($509)|0);
   $511 = $510 >> 16;
   $512 = ((($503)) + 40|0);
   HEAP32[$512>>2] = $511;
   $513 = (_FT_RoundFix($415)|0);
   $514 = $513 >> 16;
   $515 = ((($503)) + 44|0);
   HEAP32[$515>>2] = $514;
   $516 = ((($496)) + 128|0);
   HEAP32[$516>>2] = 2;
   $517 = ((($499)) + 52|0);
   $518 = HEAP32[$517>>2]|0;
   $519 = ((($496)) + 132|0);
   HEAP32[$519>>2] = $518;
   $520 = ((($496)) + 72|0);
   HEAP32[$520>>2] = 1668246896;
   $521 = ((($499)) + 84|0);
   HEAP32[$521>>2] = 2;
   $$4 = 0;
   STACKTOP = sp;return ($$4|0);
  }
  $522 = ((($$lcssa266268)) + 12|0);
  $523 = HEAP32[$522>>2]|0;
  _FT_GlyphLoader_Prepare($523);
  HEAP8[$424>>0] = 1;
  $524 = ((($$lcssa266268)) + 1472|0);
  $525 = HEAP32[$524>>2]|0;
  $526 = (FUNCTION_TABLE_iii[$525 & 8191]($$lcssa266268,$bchar_index$0$i)|0);
  HEAP8[$424>>0] = 0;
  $527 = ($526|0)==(0);
  if (!($527)) {
   $$4 = $526;
   STACKTOP = sp;return ($$4|0);
  }
  $528 = $430;
  $529 = $528;
  $530 = HEAP32[$529>>2]|0;
  $531 = (($528) + 4)|0;
  $532 = $531;
  $533 = HEAP32[$532>>2]|0;
  $534 = ((($$lcssa266268)) + 40|0);
  $535 = $534;
  $536 = $535;
  $537 = HEAP32[$536>>2]|0;
  $538 = (($535) + 4)|0;
  $539 = $538;
  $540 = HEAP32[$539>>2]|0;
  HEAP32[$430>>2] = 0;
  $541 = ((($$lcssa266268)) + 36|0);
  HEAP32[$541>>2] = 0;
  $542 = (($432) - ($411))|0;
  $543 = ((($$lcssa266268)) + 24|0);
  HEAP32[$543>>2] = $542;
  $544 = ((($$lcssa266268)) + 28|0);
  HEAP32[$544>>2] = $415;
  HEAP8[$424>>0] = 1;
  $545 = HEAP32[$524>>2]|0;
  $546 = (FUNCTION_TABLE_iii[$545 & 8191]($$lcssa266268,$achar_index$0$i)|0);
  HEAP8[$424>>0] = 0;
  $547 = ($546|0)==(0);
  if (!($547)) {
   $$4 = $546;
   STACKTOP = sp;return ($$4|0);
  }
  $548 = $430;
  $549 = $548;
  HEAP32[$549>>2] = $530;
  $550 = (($548) + 4)|0;
  $551 = $550;
  HEAP32[$551>>2] = $533;
  $552 = $534;
  $553 = $552;
  HEAP32[$553>>2] = $537;
  $554 = (($552) + 4)|0;
  $555 = $554;
  HEAP32[$555>>2] = $540;
  HEAP32[$543>>2] = 0;
  HEAP32[$544>>2] = 0;
  $$4 = 0;
  STACKTOP = sp;return ($$4|0);
 }
 else if ((label|0) == 257) {
  STACKTOP = sp;return ($$4|0);
 }
 return (0)|0;
}
function _afm_parser_init($parser,$memory,$base,$limit) {
 $parser = $parser|0;
 $memory = $memory|0;
 $base = $base|0;
 $limit = $limit|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $error = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = (_ft_mem_alloc($memory,16,$error)|0);
 $1 = HEAP32[$error>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ((($0)) + 4|0);
 HEAP32[$3>>2] = $base;
 HEAP32[$0>>2] = $base;
 $4 = ((($0)) + 8|0);
 HEAP32[$4>>2] = $limit;
 $5 = ((($0)) + 12|0);
 HEAP32[$5>>2] = 2;
 HEAP32[$parser>>2] = $memory;
 $6 = ((($parser)) + 4|0);
 HEAP32[$6>>2] = $0;
 $7 = ((($parser)) + 8|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($parser)) + 12|0);
 HEAP32[$8>>2] = 0;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _afm_parser_done($parser) {
 $parser = $parser|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$parser>>2]|0;
 $1 = ((($parser)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 _ft_mem_free($0,$2);
 HEAP32[$1>>2] = 0;
 return;
}
function _afm_parser_parse($parser) {
 $parser = $parser|0;
 var $$0396$i = 0, $$4 = 0, $$lcssa1488 = 0, $$lcssa1489 = 0, $$lcssa1491 = 0, $$lcssa1493 = 0, $$lcssa1494 = 0, $$lcssa1496 = 0, $$lcssa1497 = 0, $$lcssa1500 = 0, $$lcssa1503 = 0, $$lcssa1510 = 0, $$lcssa1511 = 0, $$lcssa1513 = 0, $$lcssa1515 = 0, $$lcssa1516 = 0, $$lcssa1518 = 0, $$lcssa1519 = 0, $$lcssa1522 = 0, $$lcssa1525 = 0;
 var $$lcssa1545 = 0, $$lcssa1546 = 0, $$lcssa1548 = 0, $$lcssa1550 = 0, $$lcssa1551 = 0, $$lcssa1553 = 0, $$lcssa1554 = 0, $$lcssa1557 = 0, $$lcssa1561 = 0, $$lcssa1578 = 0, $$lcssa1584 = 0, $$lcssa1604 = 0, $$lcssa1607 = 0, $$lcssa1608 = 0, $$lcssa1610 = 0, $$lcssa1611 = 0, $$lcssa1614 = 0, $$lcssa1629 = 0, $$lcssa1636 = 0, $$lcssa1637 = 0;
 var $$lcssa1639 = 0, $$lcssa1641 = 0, $$lcssa1642 = 0, $$lcssa1645 = 0, $$lcssa1646 = 0, $$lcssa1648 = 0, $$lcssa1650 = 0, $$lcssa1651 = 0, $$lcssa1653 = 0, $$lcssa1654 = 0, $$lcssa1657 = 0, $$lcssa1660 = 0, $$lcssa1700 = 0, $$lcssa1726 = 0, $$lcssa1727 = 0, $$lcssa1729 = 0, $$lcssa1730 = 0, $$lcssa1733 = 0, $$lcssa1736 = 0, $$lcssa1740 = 0;
 var $$old1$i = 0, $$old1$i$i = 0, $$old1$i$i$i = 0, $$old1$i$i49$i = 0, $$old1$i$i86 = 0, $$phi$trans$insert = 0, $$phi$trans$insert$i11$i = 0, $$phi$trans$insert$i11$i$lcssa1537 = 0, $$phi$trans$insert$i11$i1543 = 0, $$phi$trans$insert$i11$i87 = 0, $$phi$trans$insert$i14$i = 0, $$phi$trans$insert$i14$i$i = 0, $$phi$trans$insert$i16 = 0, $$phi$trans$insert$i18$i$i = 0, $$phi$trans$insert$i27$i = 0, $$phi$trans$insert$i54 = 0, $$phi$trans$insert$i62 = 0, $$phi$trans$insert$i62$lcssa1716 = 0, $$phi$trans$insert$i62$lcssa1719 = 0, $$phi$trans$insert$i621723 = 0;
 var $$phi$trans$insert$i621724 = 0, $$phi$trans$insert100$i$i = 0, $$phi$trans$insert149$i = 0, $$phi$trans$insert620 = 0, $$phi$trans$insert626 = 0, $$phi$trans$insert629$phi$trans$insert = 0, $$phi$trans$insert8$i$phi$trans$insert = 0, $$phi$trans$insert8$i$phi$trans$insert$i = 0, $$phi$trans$insert8$i$phi$trans$insert$i30$i = 0, $$phi$trans$insert8$i45$phi$trans$insert = 0, $$pr = 0, $$pr$i$i = 0, $$pr$i$i$lcssa = 0, $$pr$lcssa = 0, $$pr112 = 0, $$pr112$lcssa = 0, $$pr112$lcssa$lcssa = 0, $$pr50$i = 0, $$pr50$i$lcssa = 0, $$pre = 0;
 var $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i$i = 0, $$pre$i$i$ph = 0, $$pre$i$i$pre$pre = 0, $$pre$i$i$pre637 = 0, $$pre$i$i$pre637$lcssa = 0, $$pre$i$i$pre640 = 0, $$pre$i$i$pre640$lcssa = 0, $$pre$i$i$pre640$lcssa$lcssa1560 = 0, $$pre$i$i25$i = 0, $$pre$i$i65 = 0, $$pre$i$pre$i = 0, $$pre$i$pre$i$i = 0, $$pre$i11$pre$i$i = 0, $$pre$i12$i = 0, $$pre$i12$i88 = 0, $$pre$i13$pre = 0, $$pre$i14$i = 0, $$pre$i14$pre165$i = 0;
 var $$pre$i14$pre165$i$lcssa = 0, $$pre$i15$i = 0, $$pre$i15$i$i = 0, $$pre$i15$pre$i$i = 0, $$pre$i17 = 0, $$pre$i19$i$i = 0, $$pre$i20 = 0, $$pre$i20$pre = 0, $$pre$i28$i = 0, $$pre$i38$pre$i = 0, $$pre$i44$pre = 0, $$pre$i55 = 0, $$pre$i67 = 0, $$pre$i7$pre$i = 0, $$pre$i8$pre$i = 0, $$pre$phi$i10$iZ2D = 0, $$pre$phi$i18Z2D = 0, $$pre$phi$i22$iZ2D = 0, $$pre$phiZ2D = 0, $$pre103$i$i = 0;
 var $$pre158$i = 0, $$pre311$i = 0, $$pre313$i = 0, $$pre6$i = 0, $$pre6$i$i = 0, $$pre6$i$i$i = 0, $$pre6$i$i50$i = 0, $$pre6$i$i89 = 0, $$pre627 = 0, $$pre634 = 0, $$pre643 = 0, $$pre644 = 0, $$pre7$pre$i$pre = 0, $$pre7$pre$i$pre$i = 0, $$pre7$pre$i$pre$i$i = 0, $$pre7$pre$i$pre$i31$i = 0, $$pre7$pre$i$pre$i68 = 0, $$pre7$pre$i40$pre$i = 0, $$pre7$pre$i46$pre = 0, $$pre88$i$i = 0;
 var $$sink$i$i = 0, $$sink93$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0;
 var $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0;
 var $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0;
 var $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0;
 var $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0;
 var $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0;
 var $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0;
 var $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0;
 var $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0;
 var $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0;
 var $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0;
 var $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0;
 var $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $error$0 = 0, $error$0$lcssa1682 = 0, $error$0$lcssa1686 = 0, $error$0$lcssa1696 = 0, $error$01698 = 0, $error$3 = 0, $error$4$ph = 0;
 var $error$5 = 0, $error$i$i = 0, $error$i18$i = 0, $n$0$i$i = 0, $n$0$i$i$lcssa1570 = 0, $n$0$ph$i$i = 0, $n$0$ph$i$i$lcssa1534 = 0, $n$04$i = 0, $n$04$i$i = 0, $n$04$i$i$i = 0, $n$04$i$i47$i = 0, $n$04$i$i84 = 0, $n$1$i195$i = 0, $n$1$i195$i$lcssa = 0, $n$1$i377 = 0, $n$1$i377$lcssa = 0, $n$1$i52$i$i = 0, $n$1$i52$i$i$lcssa = 0, $n$1$i55$i = 0, $n$1$i60$i$i = 0;
 var $n$1$i60$i$i$lcssa = 0, $n$1$i93$i = 0, $n$1$i93$i$lcssa = 0, $n$2$i = 0, $n$2$i$i = 0, $n$2$i$i$i = 0, $n$2$i$i52$i = 0, $n$2$i$i91 = 0, $or$cond$i$i = 0, $parser$idx = 0, $parser$idx$val = 0, $parser$idx$val$i$i = 0, $parser$idx$val$i$i$lcssa1575 = 0, $parser$idx$val$i$ph = 0, $parser$idx$val$i$pre$pre = 0, $parser$idx$val$pre$i = 0, $parser$idx$val$pre$i$i = 0, $parser$idx6$val = 0, $parser$idx6$val$lcssa1668 = 0, $parser$idx6$val$pre = 0;
 var $phitmp$i = 0, $phitmp$i$i = 0, $phitmp$i$i$i = 0, $phitmp$i$i26$i = 0, $phitmp$i$i66 = 0, $phitmp$i13$i = 0, $phitmp$i15$i = 0, $phitmp$i21 = 0, $shared_vals = 0, $shared_vals$i$i = 0, $shared_vals$i19$i = 0, $val$i = 0, $val$i$i$i = 0, $val$i$i17$i = 0, $val$i58 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 160|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $val$i58 = sp + 144|0;
 $val$i = sp + 136|0;
 $val$i$i17$i = sp + 128|0;
 $error$i18$i = sp + 120|0;
 $shared_vals$i19$i = sp + 80|0;
 $val$i$i$i = sp + 72|0;
 $error$i$i = sp + 64|0;
 $shared_vals$i$i = sp + 32|0;
 $shared_vals = sp;
 $0 = HEAP32[$parser>>2]|0;
 $1 = ((($parser)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $$4 = 6;
  STACKTOP = sp;return ($$4|0);
 }
 $parser$idx = ((($parser)) + 4|0);
 $parser$idx$val = HEAP32[$parser$idx>>2]|0;
 $4 = ((($parser$idx$val)) + 12|0);
 $$pre$i = HEAP32[$4>>2]|0;
 $phitmp$i = ($$pre$i|0)>(1);
 L4: do {
  if ($phitmp$i) {
   $$phi$trans$insert = ((($parser$idx$val)) + 8|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0;
   $19 = $$pre;
  } else {
   $5 = ($$pre$i|0)>(0);
   $$pre$i13$pre = HEAP32[$parser$idx$val>>2]|0;
   $$phi$trans$insert8$i$phi$trans$insert = ((($parser$idx$val)) + 8|0);
   $$pre7$pre$i$pre = HEAP32[$$phi$trans$insert8$i$phi$trans$insert>>2]|0;
   L7: do {
    if ($5) {
     $13 = $$pre$i13$pre;
    } else {
     $6 = ($$pre$i13$pre>>>0)<($$pre7$pre$i$pre>>>0);
     L9: do {
      if ($6) {
       $8 = $$pre$i13$pre;
       L10: while(1) {
        $7 = ((($8)) + 1|0);
        HEAP32[$parser$idx$val>>2] = $7;
        $9 = HEAP8[$8>>0]|0;
        $10 = $9&255;
        switch ($10|0) {
        case 26:  {
         break L9;
         break;
        }
        case 10: case 13:  {
         label = 8;
         break L10;
         break;
        }
        case 59:  {
         $$lcssa1740 = $7;
         label = 9;
         break L10;
         break;
        }
        case 9: case 32:  {
         break;
        }
        default: {
         $13 = $7;
         break L7;
        }
        }
        $11 = ($7>>>0)<($$pre7$pre$i$pre>>>0);
        if ($11) {
         $8 = $7;
        } else {
         break L9;
        }
       }
       if ((label|0) == 8) {
        HEAP32[$4>>2] = 2;
        $19 = $$pre7$pre$i$pre;
        break L4;
       }
       else if ((label|0) == 9) {
        HEAP32[$4>>2] = 1;
        $13 = $$lcssa1740;
        break L7;
       }
      }
     } while(0);
     HEAP32[$4>>2] = 3;
     $19 = $$pre7$pre$i$pre;
     break L4;
    }
   } while(0);
   L17: while(1) {
    $12 = ($13>>>0)<($$pre7$pre$i$pre>>>0);
    if (!($12)) {
     label = 14;
     break;
    }
    $14 = ((($13)) + 1|0);
    HEAP32[$parser$idx$val>>2] = $14;
    $15 = HEAP8[$13>>0]|0;
    $16 = $15&255;
    switch ($16|0) {
    case 26:  {
     label = 14;
     break L17;
     break;
    }
    case 10: case 13:  {
     label = 13;
     break L17;
     break;
    }
    default: {
     $13 = $14;
    }
    }
   }
   if ((label|0) == 13) {
    HEAP32[$4>>2] = 2;
    $19 = $$pre7$pre$i$pre;
    break;
   }
   else if ((label|0) == 14) {
    HEAP32[$4>>2] = 3;
    $19 = $$pre7$pre$i$pre;
    break;
   }
  }
 } while(0);
 HEAP32[$4>>2] = 0;
 $17 = HEAP32[$parser$idx$val>>2]|0;
 $18 = ($17>>>0)<($19>>>0);
 L24: do {
  if ($18) {
   $21 = $17;
   L25: while(1) {
    $20 = ((($21)) + 1|0);
    HEAP32[$parser$idx$val>>2] = $20;
    $22 = HEAP8[$21>>0]|0;
    $23 = $22&255;
    switch ($23|0) {
    case 26:  {
     break L24;
     break;
    }
    case 59:  {
     label = 20;
     break L25;
     break;
    }
    case 9: case 32:  {
     $25 = ($20>>>0)<($19>>>0);
     if ($25) {
      $21 = $20;
      continue L25;
     } else {
      break L24;
     }
     break;
    }
    case 10: case 13:  {
     HEAP32[$4>>2] = 2;
     $35 = $20;
     break;
    }
    default: {
     $24 = $20;
     $27 = $20;$429 = $24;
     L29: while(1) {
      $26 = ($27>>>0)<($19>>>0);
      if (!($26)) {
       $430 = $27;$431 = $429;
       label = 26;
       break;
      }
      $28 = ((($27)) + 1|0);
      HEAP32[$parser$idx$val>>2] = $28;
      $29 = HEAP8[$27>>0]|0;
      $30 = $29&255;
      $31 = $28;
      switch ($30|0) {
      case 26:  {
       $430 = $28;$431 = $31;
       label = 26;
       break L29;
       break;
      }
      case 9: case 32:  {
       $$pr = 0;$432 = $28;$433 = $31;
       break L29;
       break;
      }
      case 10: case 13:  {
       $$lcssa1726 = $28;$$lcssa1729 = $31;
       label = 24;
       break L29;
       break;
      }
      case 59:  {
       $$lcssa1727 = $28;$$lcssa1730 = $31;
       label = 25;
       break L29;
       break;
      }
      default: {
       $27 = $28;$429 = $31;
      }
      }
     }
     if ((label|0) == 24) {
      label = 0;
      HEAP32[$4>>2] = 2;
      $$pr = 2;$432 = $$lcssa1726;$433 = $$lcssa1729;
     }
     else if ((label|0) == 25) {
      label = 0;
      HEAP32[$4>>2] = 1;
      $$pr = 1;$432 = $$lcssa1727;$433 = $$lcssa1730;
     }
     else if ((label|0) == 26) {
      label = 0;
      HEAP32[$4>>2] = 3;
      $$pr = 3;$432 = $430;$433 = $431;
     }
     $32 = ($21|0)==(0|0);
     if (!($32)) {
      $$lcssa1733 = $21;$$lcssa1736 = $433;$$pr$lcssa = $$pr;
      label = 30;
      break L25;
     }
     $33 = ($$pr|0)==(2);
     if ($33) {
      $35 = $432;
     } else {
      $$4 = 2;
      label = 285;
      break L25;
     }
    }
    }
    HEAP32[$4>>2] = 0;
    $34 = ($35>>>0)<($19>>>0);
    if ($34) {
     $21 = $35;
    } else {
     break L24;
    }
   }
   if ((label|0) == 20) {
    HEAP32[$4>>2] = 1;
    $$4 = 2;
    STACKTOP = sp;return ($$4|0);
   }
   else if ((label|0) == 30) {
    $36 = $$lcssa1733;
    $37 = $36 ^ -1;
    $38 = (($$lcssa1736) + ($37))|0;
    $39 = ($38|0)==(16);
    if (!($39)) {
     $$4 = 2;
     STACKTOP = sp;return ($$4|0);
    }
    $40 = (_strncmp($$lcssa1733,532705,16)|0);
    $41 = ($40|0)==(0);
    if (!($41)) {
     $$4 = 2;
     STACKTOP = sp;return ($$4|0);
    }
    $42 = ((($val$i)) + 4|0);
    $43 = ((($shared_vals)) + 4|0);
    $44 = ((($shared_vals)) + 8|0);
    $45 = ((($shared_vals)) + 16|0);
    $46 = ((($shared_vals)) + 24|0);
    $47 = ((($2)) + 4|0);
    $48 = ((($shared_vals)) + 12|0);
    $49 = ((($2)) + 8|0);
    $50 = ((($shared_vals)) + 20|0);
    $51 = ((($2)) + 12|0);
    $52 = ((($shared_vals)) + 28|0);
    $53 = ((($2)) + 16|0);
    $54 = ((($2)) + 20|0);
    $55 = ((($2)) + 24|0);
    $56 = ((($val$i58)) + 4|0);
    $$pre$i20 = $$pr$lcssa;$error$0 = 160;$parser$idx6$val = $parser$idx$val;
    L52: while(1) {
     $57 = ((($parser$idx6$val)) + 12|0);
     $phitmp$i21 = ($$pre$i20|0)>(1);
     L54: do {
      if ($phitmp$i21) {
       $$phi$trans$insert626 = ((($parser$idx6$val)) + 8|0);
       $$pre627 = HEAP32[$$phi$trans$insert626>>2]|0;
       $72 = $$pre627;
      } else {
       $58 = ($$pre$i20|0)>(0);
       $$pre$i44$pre = HEAP32[$parser$idx6$val>>2]|0;
       $$phi$trans$insert8$i45$phi$trans$insert = ((($parser$idx6$val)) + 8|0);
       $$pre7$pre$i46$pre = HEAP32[$$phi$trans$insert8$i45$phi$trans$insert>>2]|0;
       L57: do {
        if ($58) {
         $66 = $$pre$i44$pre;
        } else {
         $59 = ($$pre$i44$pre>>>0)<($$pre7$pre$i46$pre>>>0);
         L59: do {
          if ($59) {
           $61 = $$pre$i44$pre;
           L60: while(1) {
            $60 = ((($61)) + 1|0);
            HEAP32[$parser$idx6$val>>2] = $60;
            $62 = HEAP8[$61>>0]|0;
            $63 = $62&255;
            switch ($63|0) {
            case 26:  {
             break L59;
             break;
            }
            case 10: case 13:  {
             label = 39;
             break L60;
             break;
            }
            case 59:  {
             $$lcssa1604 = $60;
             label = 40;
             break L60;
             break;
            }
            case 9: case 32:  {
             break;
            }
            default: {
             $66 = $60;
             break L57;
            }
            }
            $64 = ($60>>>0)<($$pre7$pre$i46$pre>>>0);
            if ($64) {
             $61 = $60;
            } else {
             break L59;
            }
           }
           if ((label|0) == 39) {
            label = 0;
            HEAP32[$57>>2] = 2;
            $72 = $$pre7$pre$i46$pre;
            break L54;
           }
           else if ((label|0) == 40) {
            label = 0;
            HEAP32[$57>>2] = 1;
            $66 = $$lcssa1604;
            break L57;
           }
          }
         } while(0);
         HEAP32[$57>>2] = 3;
         $72 = $$pre7$pre$i46$pre;
         break L54;
        }
       } while(0);
       L67: while(1) {
        $65 = ($66>>>0)<($$pre7$pre$i46$pre>>>0);
        if (!($65)) {
         label = 45;
         break;
        }
        $67 = ((($66)) + 1|0);
        HEAP32[$parser$idx6$val>>2] = $67;
        $68 = HEAP8[$66>>0]|0;
        $69 = $68&255;
        switch ($69|0) {
        case 26:  {
         label = 45;
         break L67;
         break;
        }
        case 10: case 13:  {
         label = 44;
         break L67;
         break;
        }
        default: {
         $66 = $67;
        }
        }
       }
       if ((label|0) == 44) {
        label = 0;
        HEAP32[$57>>2] = 2;
        $72 = $$pre7$pre$i46$pre;
        break;
       }
       else if ((label|0) == 45) {
        label = 0;
        HEAP32[$57>>2] = 3;
        $72 = $$pre7$pre$i46$pre;
        break;
       }
      }
     } while(0);
     HEAP32[$57>>2] = 0;
     $70 = HEAP32[$parser$idx6$val>>2]|0;
     $71 = ($70>>>0)<($72>>>0);
     if ($71) {
      $74 = $70;
     } else {
      $79 = $57;$error$01698 = $error$0;
      label = 52;
      break;
     }
     L74: while(1) {
      $73 = ((($74)) + 1|0);
      HEAP32[$parser$idx6$val>>2] = $73;
      $75 = HEAP8[$74>>0]|0;
      $76 = $75&255;
      switch ($76|0) {
      case 59:  {
       $$lcssa1700 = $57;$error$0$lcssa1682 = $error$0;
       label = 51;
       break L52;
       break;
      }
      case 26:  {
       $79 = $57;$error$01698 = $error$0;
       label = 52;
       break L52;
       break;
      }
      case 9: case 32:  {
       $78 = ($73>>>0)<($72>>>0);
       if ($78) {
        $74 = $73;
        continue L74;
       } else {
        $79 = $57;$error$01698 = $error$0;
        label = 52;
        break L52;
       }
       break;
      }
      case 10: case 13:  {
       HEAP32[$57>>2] = 2;
       $89 = $73;
       break;
      }
      default: {
       $77 = $73;
       $434 = $77;$81 = $73;
       L78: while(1) {
        $80 = ($81>>>0)<($72>>>0);
        if (!($80)) {
         $435 = $81;$436 = $434;
         label = 57;
         break;
        }
        $82 = ((($81)) + 1|0);
        HEAP32[$parser$idx6$val>>2] = $82;
        $83 = HEAP8[$81>>0]|0;
        $84 = $83&255;
        $85 = $82;
        switch ($84|0) {
        case 26:  {
         $435 = $82;$436 = $85;
         label = 57;
         break L78;
         break;
        }
        case 9: case 32:  {
         $$pr112 = 0;$437 = $82;$438 = $85;
         break L78;
         break;
        }
        case 10: case 13:  {
         $$lcssa1607 = $82;$$lcssa1610 = $85;
         label = 55;
         break L78;
         break;
        }
        case 59:  {
         $$lcssa1608 = $82;$$lcssa1611 = $85;
         label = 56;
         break L78;
         break;
        }
        default: {
         $434 = $85;$81 = $82;
        }
        }
       }
       if ((label|0) == 55) {
        label = 0;
        HEAP32[$57>>2] = 2;
        $$pr112 = 2;$437 = $$lcssa1607;$438 = $$lcssa1610;
       }
       else if ((label|0) == 56) {
        label = 0;
        HEAP32[$57>>2] = 1;
        $$pr112 = 1;$437 = $$lcssa1608;$438 = $$lcssa1611;
       }
       else if ((label|0) == 57) {
        label = 0;
        HEAP32[$57>>2] = 3;
        $$pr112 = 3;$437 = $435;$438 = $436;
       }
       $86 = ($74|0)==(0|0);
       if (!($86)) {
        $$lcssa1614 = $74;$$lcssa1629 = $438;$$pr112$lcssa = $$pr112;
        break L74;
       }
       $87 = ($$pr112|0)==(2);
       if ($87) {
        $89 = $437;
       } else {
        $error$5 = $error$0;
        break L52;
       }
      }
      }
      HEAP32[$57>>2] = 0;
      $88 = ($89>>>0)<($72>>>0);
      if ($88) {
       $74 = $89;
      } else {
       $79 = $57;$error$01698 = $error$0;
       label = 52;
       break L52;
      }
     }
     $90 = $$lcssa1614;
     $91 = $90 ^ -1;
     $92 = (($$lcssa1629) + ($91))|0;
     $93 = HEAP8[$$lcssa1614>>0]|0;
     $n$04$i = 0;
     L91: while(1) {
      $94 = (53740 + ($n$04$i<<2)|0);
      $95 = HEAP32[$94>>2]|0;
      $96 = HEAP8[$95>>0]|0;
      $97 = ($96<<24>>24)==($93<<24>>24);
      L93: do {
       if ($97) {
        $98 = $95;$n$1$i377 = $n$04$i;
        while(1) {
         $99 = (_strncmp($98,$$lcssa1614,$92)|0);
         $100 = ($99|0)==(0);
         if ($100) {
          $n$1$i377$lcssa = $n$1$i377;
          label = 67;
          break L91;
         }
         $101 = (($n$1$i377) + 1)|0;
         $$old1$i = ($101|0)<(74);
         if (!($$old1$i)) {
          $n$2$i = $101;
          break L93;
         }
         $$phi$trans$insert$i54 = (53740 + ($101<<2)|0);
         $$pre$i55 = HEAP32[$$phi$trans$insert$i54>>2]|0;
         $$pre6$i = HEAP8[$$pre$i55>>0]|0;
         $102 = ($$pre6$i<<24>>24)==($93<<24>>24);
         if ($102) {
          $98 = $$pre$i55;$n$1$i377 = $101;
         } else {
          $error$3 = $error$0;
          break L91;
         }
        }
       } else {
        $n$2$i = $n$04$i;
       }
      } while(0);
      $103 = (($n$2$i) + 1)|0;
      $104 = ($103|0)<(74);
      if ($104) {
       $n$04$i = $103;
      } else {
       $error$3 = $error$0;
       break;
      }
     }
     L99: do {
      if ((label|0) == 67) {
       label = 0;
       switch ($n$1$i377$lcssa|0) {
       case 49:  {
        $$pr112$lcssa$lcssa = $$pr112$lcssa;$parser$idx6$val$lcssa1668 = $parser$idx6$val;
        label = 145;
        break L52;
        break;
       }
       case 20:  {
        label = 282;
        break L52;
        break;
       }
       case 40:  {
        HEAP32[$val$i>>2] = 3;
        $105 = (_afm_parser_read_vals($parser,$val$i,1)|0);
        $106 = ($105|0)==(1);
        if (!($106)) {
         $error$0$lcssa1696 = $error$0;
         label = 69;
         break L52;
        }
        $107 = HEAP32[$42>>2]|0;
        switch ($107|0) {
        case 0: case 2:  {
         $error$3 = $error$0;
         break L99;
         break;
        }
        default: {
         $error$4$ph = 7;
         label = 281;
         break L52;
        }
        }
        break;
       }
       case 30:  {
        HEAP32[$shared_vals>>2] = 4;
        $108 = (_afm_parser_read_vals($parser,$shared_vals,1)|0);
        $109 = ($108|0)==(1);
        if (!($109)) {
         $error$4$ph = $error$0;
         label = 281;
         break L52;
        }
        $110 = HEAP8[$43>>0]|0;
        HEAP8[$2>>0] = $110;
        $error$3 = $error$0;
        break L99;
        break;
       }
       case 26:  {
        HEAP32[$shared_vals>>2] = 2;
        HEAP32[$44>>2] = 2;
        HEAP32[$45>>2] = 2;
        HEAP32[$46>>2] = 2;
        $111 = (_afm_parser_read_vals($parser,$shared_vals,4)|0);
        $112 = ($111|0)==(4);
        if (!($112)) {
         $error$4$ph = $error$0;
         label = 281;
         break L52;
        }
        $113 = HEAP32[$43>>2]|0;
        HEAP32[$47>>2] = $113;
        $114 = HEAP32[$48>>2]|0;
        HEAP32[$49>>2] = $114;
        $115 = HEAP32[$50>>2]|0;
        HEAP32[$51>>2] = $115;
        $116 = HEAP32[$52>>2]|0;
        HEAP32[$53>>2] = $116;
        $error$3 = $error$0;
        break L99;
        break;
       }
       case 0:  {
        HEAP32[$shared_vals>>2] = 2;
        $117 = (_afm_parser_read_vals($parser,$shared_vals,1)|0);
        $118 = ($117|0)==(1);
        if (!($118)) {
         $error$4$ph = $error$0;
         label = 281;
         break L52;
        }
        $119 = HEAP32[$43>>2]|0;
        HEAP32[$54>>2] = $119;
        $error$3 = $error$0;
        break L99;
        break;
       }
       case 14:  {
        HEAP32[$shared_vals>>2] = 2;
        $120 = (_afm_parser_read_vals($parser,$shared_vals,1)|0);
        $121 = ($120|0)==(1);
        if (!($121)) {
         $error$4$ph = $error$0;
         label = 281;
         break L52;
        }
        $122 = HEAP32[$43>>2]|0;
        HEAP32[$55>>2] = $122;
        $error$3 = $error$0;
        break L99;
        break;
       }
       case 45:  {
        HEAP32[$val$i58>>2] = 3;
        $123 = (_afm_parser_read_vals($parser,$val$i58,1)|0);
        $124 = ($123|0)==(1);
        if (!($124)) {
         $error$0$lcssa1686 = $error$0;
         label = 283;
         break L52;
        }
        $125 = HEAP32[$56>>2]|0;
        $126 = ($125|0)>(0);
        $parser$idx$val$pre$i = HEAP32[$parser$idx>>2]|0;
        $$phi$trans$insert$i62 = ((($parser$idx$val$pre$i)) + 12|0);
        $$pre$i$pre$i = HEAP32[$$phi$trans$insert$i62>>2]|0;
        $$phi$trans$insert149$i = ((($parser$idx$val$pre$i)) + 8|0);
        if ($126) {
         $$0396$i = $125;$$pre$i$i65 = $$pre$i$pre$i;
         while(1) {
          $128 = (($$0396$i) + -1)|0;
          $phitmp$i$i66 = ($$pre$i$i65|0)>(1);
          L117: do {
           if ($phitmp$i$i66) {
            $$pre$i67 = HEAP32[$$phi$trans$insert149$i>>2]|0;
            $$pre643 = HEAP32[$parser$idx$val$pre$i>>2]|0;
            $142 = $$pre643;$143 = $$pre$i67;
           } else {
            $129 = ($$pre$i$i65|0)>(0);
            $$pre$i8$pre$i = HEAP32[$parser$idx$val$pre$i>>2]|0;
            $$pre7$pre$i$pre$i68 = HEAP32[$$phi$trans$insert149$i>>2]|0;
            L120: do {
             if ($129) {
              $137 = $$pre$i8$pre$i;
             } else {
              $130 = ($$pre$i8$pre$i>>>0)<($$pre7$pre$i$pre$i68>>>0);
              L122: do {
               if ($130) {
                $132 = $$pre$i8$pre$i;
                L123: while(1) {
                 $131 = ((($132)) + 1|0);
                 HEAP32[$parser$idx$val$pre$i>>2] = $131;
                 $133 = HEAP8[$132>>0]|0;
                 $134 = $133&255;
                 switch ($134|0) {
                 case 26:  {
                  $439 = $131;
                  break L122;
                  break;
                 }
                 case 10: case 13:  {
                  $$lcssa1636 = $131;
                  label = 88;
                  break L123;
                  break;
                 }
                 case 59:  {
                  $$lcssa1637 = $131;
                  label = 89;
                  break L123;
                  break;
                 }
                 case 9: case 32:  {
                  break;
                 }
                 default: {
                  $137 = $131;
                  break L120;
                 }
                 }
                 $135 = ($131>>>0)<($$pre7$pre$i$pre$i68>>>0);
                 if ($135) {
                  $132 = $131;
                 } else {
                  $439 = $131;
                  break L122;
                 }
                }
                if ((label|0) == 88) {
                 label = 0;
                 HEAP32[$$phi$trans$insert$i62>>2] = 2;
                 $142 = $$lcssa1636;$143 = $$pre7$pre$i$pre$i68;
                 break L117;
                }
                else if ((label|0) == 89) {
                 label = 0;
                 HEAP32[$$phi$trans$insert$i62>>2] = 1;
                 $137 = $$lcssa1637;
                 break L120;
                }
               } else {
                $439 = $$pre$i8$pre$i;
               }
              } while(0);
              HEAP32[$$phi$trans$insert$i62>>2] = 3;
              $142 = $439;$143 = $$pre7$pre$i$pre$i68;
              break L117;
             }
            } while(0);
            L130: while(1) {
             $136 = ($137>>>0)<($$pre7$pre$i$pre$i68>>>0);
             if (!($136)) {
              $440 = $137;
              label = 94;
              break;
             }
             $138 = ((($137)) + 1|0);
             HEAP32[$parser$idx$val$pre$i>>2] = $138;
             $139 = HEAP8[$137>>0]|0;
             $140 = $139&255;
             switch ($140|0) {
             case 26:  {
              $440 = $138;
              label = 94;
              break L130;
              break;
             }
             case 10: case 13:  {
              $$lcssa1639 = $138;
              label = 93;
              break L130;
              break;
             }
             default: {
              $137 = $138;
             }
             }
            }
            if ((label|0) == 93) {
             label = 0;
             HEAP32[$$phi$trans$insert$i62>>2] = 2;
             $142 = $$lcssa1639;$143 = $$pre7$pre$i$pre$i68;
             break;
            }
            else if ((label|0) == 94) {
             label = 0;
             HEAP32[$$phi$trans$insert$i62>>2] = 3;
             $142 = $440;$143 = $$pre7$pre$i$pre$i68;
             break;
            }
           }
          } while(0);
          HEAP32[$$phi$trans$insert$i62>>2] = 0;
          $141 = ($142>>>0)<($143>>>0);
          if ($141) {
           $145 = $142;
          } else {
           $$phi$trans$insert$i621723 = $$phi$trans$insert$i62;
           label = 100;
           break L52;
          }
          L137: while(1) {
           $144 = ((($145)) + 1|0);
           HEAP32[$parser$idx$val$pre$i>>2] = $144;
           $146 = HEAP8[$145>>0]|0;
           $147 = $146&255;
           switch ($147|0) {
           case 59:  {
            $$phi$trans$insert$i62$lcssa1716 = $$phi$trans$insert$i62;
            label = 99;
            break L52;
            break;
           }
           case 26:  {
            $$phi$trans$insert$i621723 = $$phi$trans$insert$i62;
            label = 100;
            break L52;
            break;
           }
           case 9: case 32:  {
            $148 = ($144>>>0)<($143>>>0);
            if ($148) {
             $145 = $144;
             continue L137;
            } else {
             $$phi$trans$insert$i621723 = $$phi$trans$insert$i62;
             label = 100;
             break L52;
            }
            break;
           }
           case 10: case 13:  {
            HEAP32[$$phi$trans$insert$i62>>2] = 2;
            $157 = $144;
            break;
           }
           default: {
            $150 = $144;
            L140: while(1) {
             $149 = ($150>>>0)<($143>>>0);
             if (!($149)) {
              $441 = $150;
              label = 105;
              break;
             }
             $151 = ((($150)) + 1|0);
             HEAP32[$parser$idx$val$pre$i>>2] = $151;
             $152 = HEAP8[$150>>0]|0;
             $153 = $152&255;
             switch ($153|0) {
             case 26:  {
              $441 = $151;
              label = 105;
              break L140;
              break;
             }
             case 9: case 32:  {
              $$pre$i14$pre165$i = 0;$442 = $151;
              break L140;
              break;
             }
             case 10: case 13:  {
              $$lcssa1641 = $151;
              label = 103;
              break L140;
              break;
             }
             case 59:  {
              $$lcssa1642 = $151;
              label = 104;
              break L140;
              break;
             }
             default: {
              $150 = $151;
             }
             }
            }
            if ((label|0) == 103) {
             label = 0;
             HEAP32[$$phi$trans$insert$i62>>2] = 2;
             $$pre$i14$pre165$i = 2;$442 = $$lcssa1641;
            }
            else if ((label|0) == 104) {
             label = 0;
             HEAP32[$$phi$trans$insert$i62>>2] = 1;
             $$pre$i14$pre165$i = 1;$442 = $$lcssa1642;
            }
            else if ((label|0) == 105) {
             label = 0;
             HEAP32[$$phi$trans$insert$i62>>2] = 3;
             $$pre$i14$pre165$i = 3;$442 = $441;
            }
            $154 = ($145|0)==(0|0);
            if (!($154)) {
             $$pre$i14$pre165$i$lcssa = $$pre$i14$pre165$i;
             break L137;
            }
            $155 = ($$pre$i14$pre165$i|0)==(2);
            if ($155) {
             $157 = $442;
            } else {
             label = 144;
             break L52;
            }
           }
           }
           HEAP32[$$phi$trans$insert$i62>>2] = 0;
           $156 = ($157>>>0)<($143>>>0);
           if ($156) {
            $145 = $157;
           } else {
            $$phi$trans$insert$i621723 = $$phi$trans$insert$i62;
            label = 100;
            break L52;
           }
          }
          $127 = ($$0396$i|0)>(1);
          if ($127) {
           $$0396$i = $128;$$pre$i$i65 = $$pre$i14$pre165$i$lcssa;
          } else {
           $$pre$i14$i = $$pre$i14$pre165$i$lcssa;
           break;
          }
         }
        } else {
         $$pre$i14$i = $$pre$i$pre$i;
        }
        L153: while(1) {
         $phitmp$i15$i = ($$pre$i14$i|0)>(1);
         L155: do {
          if ($phitmp$i15$i) {
           $$pre158$i = HEAP32[$$phi$trans$insert149$i>>2]|0;
           $$pre644 = HEAP32[$parser$idx$val$pre$i>>2]|0;
           $171 = $$pre644;$172 = $$pre158$i;
          } else {
           $158 = ($$pre$i14$i|0)>(0);
           $$pre$i38$pre$i = HEAP32[$parser$idx$val$pre$i>>2]|0;
           $$pre7$pre$i40$pre$i = HEAP32[$$phi$trans$insert149$i>>2]|0;
           L158: do {
            if ($158) {
             $166 = $$pre$i38$pre$i;
            } else {
             $159 = ($$pre$i38$pre$i>>>0)<($$pre7$pre$i40$pre$i>>>0);
             L160: do {
              if ($159) {
               $161 = $$pre$i38$pre$i;
               L161: while(1) {
                $160 = ((($161)) + 1|0);
                HEAP32[$parser$idx$val$pre$i>>2] = $160;
                $162 = HEAP8[$161>>0]|0;
                $163 = $162&255;
                switch ($163|0) {
                case 26:  {
                 $443 = $160;
                 break L160;
                 break;
                }
                case 10: case 13:  {
                 $$lcssa1645 = $160;
                 label = 115;
                 break L161;
                 break;
                }
                case 59:  {
                 $$lcssa1646 = $160;
                 label = 116;
                 break L161;
                 break;
                }
                case 9: case 32:  {
                 break;
                }
                default: {
                 $166 = $160;
                 break L158;
                }
                }
                $164 = ($160>>>0)<($$pre7$pre$i40$pre$i>>>0);
                if ($164) {
                 $161 = $160;
                } else {
                 $443 = $160;
                 break L160;
                }
               }
               if ((label|0) == 115) {
                label = 0;
                HEAP32[$$phi$trans$insert$i62>>2] = 2;
                $171 = $$lcssa1645;$172 = $$pre7$pre$i40$pre$i;
                break L155;
               }
               else if ((label|0) == 116) {
                label = 0;
                HEAP32[$$phi$trans$insert$i62>>2] = 1;
                $166 = $$lcssa1646;
                break L158;
               }
              } else {
               $443 = $$pre$i38$pre$i;
              }
             } while(0);
             HEAP32[$$phi$trans$insert$i62>>2] = 3;
             $171 = $443;$172 = $$pre7$pre$i40$pre$i;
             break L155;
            }
           } while(0);
           L168: while(1) {
            $165 = ($166>>>0)<($$pre7$pre$i40$pre$i>>>0);
            if (!($165)) {
             $444 = $166;
             label = 121;
             break;
            }
            $167 = ((($166)) + 1|0);
            HEAP32[$parser$idx$val$pre$i>>2] = $167;
            $168 = HEAP8[$166>>0]|0;
            $169 = $168&255;
            switch ($169|0) {
            case 26:  {
             $444 = $167;
             label = 121;
             break L168;
             break;
            }
            case 10: case 13:  {
             $$lcssa1648 = $167;
             label = 120;
             break L168;
             break;
            }
            default: {
             $166 = $167;
            }
            }
           }
           if ((label|0) == 120) {
            label = 0;
            HEAP32[$$phi$trans$insert$i62>>2] = 2;
            $171 = $$lcssa1648;$172 = $$pre7$pre$i40$pre$i;
            break;
           }
           else if ((label|0) == 121) {
            label = 0;
            HEAP32[$$phi$trans$insert$i62>>2] = 3;
            $171 = $444;$172 = $$pre7$pre$i40$pre$i;
            break;
           }
          }
         } while(0);
         HEAP32[$$phi$trans$insert$i62>>2] = 0;
         $170 = ($171>>>0)<($172>>>0);
         if ($170) {
          $174 = $171;
         } else {
          $$phi$trans$insert$i621724 = $$phi$trans$insert$i62;
          label = 128;
          break L52;
         }
         L175: while(1) {
          $173 = ((($174)) + 1|0);
          HEAP32[$parser$idx$val$pre$i>>2] = $173;
          $175 = HEAP8[$174>>0]|0;
          $176 = $175&255;
          switch ($176|0) {
          case 59:  {
           $$phi$trans$insert$i62$lcssa1719 = $$phi$trans$insert$i62;
           label = 127;
           break L52;
           break;
          }
          case 26:  {
           $$phi$trans$insert$i621724 = $$phi$trans$insert$i62;
           label = 128;
           break L52;
           break;
          }
          case 9: case 32:  {
           $178 = ($173>>>0)<($172>>>0);
           if ($178) {
            $174 = $173;
            continue L175;
           } else {
            $$phi$trans$insert$i621724 = $$phi$trans$insert$i62;
            label = 128;
            break L52;
           }
           break;
          }
          case 10: case 13:  {
           HEAP32[$$phi$trans$insert$i62>>2] = 2;
           $188 = $173;
           break;
          }
          default: {
           $177 = $173;
           $180 = $173;$445 = $177;
           L179: while(1) {
            $179 = ($180>>>0)<($172>>>0);
            if (!($179)) {
             $446 = $180;$447 = $445;
             label = 133;
             break;
            }
            $181 = ((($180)) + 1|0);
            HEAP32[$parser$idx$val$pre$i>>2] = $181;
            $182 = HEAP8[$180>>0]|0;
            $183 = $182&255;
            $184 = $181;
            switch ($183|0) {
            case 26:  {
             $446 = $181;$447 = $184;
             label = 133;
             break L179;
             break;
            }
            case 9: case 32:  {
             $$pr50$i = 0;$448 = $181;$449 = $184;
             break L179;
             break;
            }
            case 10: case 13:  {
             $$lcssa1650 = $181;$$lcssa1653 = $184;
             label = 131;
             break L179;
             break;
            }
            case 59:  {
             $$lcssa1651 = $181;$$lcssa1654 = $184;
             label = 132;
             break L179;
             break;
            }
            default: {
             $180 = $181;$445 = $184;
            }
            }
           }
           if ((label|0) == 131) {
            label = 0;
            HEAP32[$$phi$trans$insert$i62>>2] = 2;
            $$pr50$i = 2;$448 = $$lcssa1650;$449 = $$lcssa1653;
           }
           else if ((label|0) == 132) {
            label = 0;
            HEAP32[$$phi$trans$insert$i62>>2] = 1;
            $$pr50$i = 1;$448 = $$lcssa1651;$449 = $$lcssa1654;
           }
           else if ((label|0) == 133) {
            label = 0;
            HEAP32[$$phi$trans$insert$i62>>2] = 3;
            $$pr50$i = 3;$448 = $446;$449 = $447;
           }
           $185 = ($174|0)==(0|0);
           if (!($185)) {
            $$lcssa1657 = $174;$$lcssa1660 = $449;$$pr50$i$lcssa = $$pr50$i;
            break L175;
           }
           $186 = ($$pr50$i|0)==(2);
           if ($186) {
            $188 = $448;
           } else {
            label = 144;
            break L52;
           }
          }
          }
          HEAP32[$$phi$trans$insert$i62>>2] = 0;
          $187 = ($188>>>0)<($172>>>0);
          if ($187) {
           $174 = $188;
          } else {
           $$phi$trans$insert$i621724 = $$phi$trans$insert$i62;
           label = 128;
           break L52;
          }
         }
         $189 = $$lcssa1657;
         $190 = $189 ^ -1;
         $191 = (($$lcssa1660) + ($190))|0;
         $192 = HEAP8[$$lcssa1657>>0]|0;
         $n$04$i$i84 = 0;
         L192: while(1) {
          $193 = (53740 + ($n$04$i$i84<<2)|0);
          $194 = HEAP32[$193>>2]|0;
          $195 = HEAP8[$194>>0]|0;
          $196 = ($195<<24>>24)==($192<<24>>24);
          L194: do {
           if ($196) {
            $197 = $194;$n$1$i93$i = $n$04$i$i84;
            while(1) {
             $198 = (_strncmp($197,$$lcssa1657,$191)|0);
             $199 = ($198|0)==(0);
             if ($199) {
              $n$1$i93$i$lcssa = $n$1$i93$i;
              break L192;
             }
             $200 = (($n$1$i93$i) + 1)|0;
             $$old1$i$i86 = ($200|0)<(74);
             if (!($$old1$i$i86)) {
              $n$2$i$i91 = $200;
              break L194;
             }
             $$phi$trans$insert$i11$i87 = (53740 + ($200<<2)|0);
             $$pre$i12$i88 = HEAP32[$$phi$trans$insert$i11$i87>>2]|0;
             $$pre6$i$i89 = HEAP8[$$pre$i12$i88>>0]|0;
             $201 = ($$pre6$i$i89<<24>>24)==($192<<24>>24);
             if ($201) {
              $197 = $$pre$i12$i88;$n$1$i93$i = $200;
             } else {
              $$pre$i14$i = $$pr50$i$lcssa;
              continue L153;
             }
            }
           } else {
            $n$2$i$i91 = $n$04$i$i84;
           }
          } while(0);
          $202 = (($n$2$i$i91) + 1)|0;
          $203 = ($202|0)<(74);
          if ($203) {
           $n$04$i$i84 = $202;
          } else {
           $$pre$i14$i = $$pr50$i$lcssa;
           continue L153;
          }
         }
         switch ($n$1$i93$i$lcssa|0) {
         case 17: case 20:  {
          $error$3 = 0;
          break L99;
          break;
         }
         default: {
          $$pre$i14$i = $$pr50$i$lcssa;
         }
         }
        }
        break;
       }
       default: {
        $error$3 = $error$0;
        break L99;
       }
       }
      }
     } while(0);
     $parser$idx6$val$pre = HEAP32[$parser$idx>>2]|0;
     $$phi$trans$insert620 = ((($parser$idx6$val$pre)) + 12|0);
     $$pre$i20$pre = HEAP32[$$phi$trans$insert620>>2]|0;
     $$pre$i20 = $$pre$i20$pre;$error$0 = $error$3;$parser$idx6$val = $parser$idx6$val$pre;
    }
    L202: do {
     if ((label|0) == 51) {
      HEAP32[$$lcssa1700>>2] = 1;
      $error$5 = $error$0$lcssa1682;
     }
     else if ((label|0) == 52) {
      HEAP32[$79>>2] = 3;
      $error$5 = $error$01698;
     }
     else if ((label|0) == 69) {
      $error$4$ph = $error$0$lcssa1696;
      label = 281;
     }
     else if ((label|0) == 99) {
      HEAP32[$$phi$trans$insert$i62$lcssa1716>>2] = 1;
      label = 144;
     }
     else if ((label|0) == 100) {
      HEAP32[$$phi$trans$insert$i621723>>2] = 3;
      label = 144;
     }
     else if ((label|0) == 127) {
      HEAP32[$$phi$trans$insert$i62$lcssa1719>>2] = 1;
      label = 144;
     }
     else if ((label|0) == 128) {
      HEAP32[$$phi$trans$insert$i621724>>2] = 3;
      label = 144;
     }
     else if ((label|0) == 145) {
      $204 = ((($val$i$i17$i)) + 4|0);
      $205 = ((($shared_vals$i19$i)) + 8|0);
      $206 = ((($shared_vals$i19$i)) + 16|0);
      $207 = ((($shared_vals$i19$i)) + 24|0);
      $208 = ((($shared_vals$i19$i)) + 32|0);
      $209 = ((($shared_vals$i19$i)) + 4|0);
      $210 = ((($shared_vals$i19$i)) + 12|0);
      $211 = ((($shared_vals$i19$i)) + 20|0);
      $212 = ((($shared_vals$i19$i)) + 28|0);
      $213 = ((($shared_vals$i19$i)) + 36|0);
      $214 = ((($val$i$i$i)) + 4|0);
      $215 = ((($shared_vals$i$i)) + 8|0);
      $216 = ((($shared_vals$i$i)) + 16|0);
      $217 = ((($shared_vals$i$i)) + 24|0);
      $218 = ((($shared_vals$i$i)) + 4|0);
      $219 = ((($shared_vals$i$i)) + 12|0);
      $220 = ((($shared_vals$i$i)) + 20|0);
      $221 = ((($shared_vals$i$i)) + 28|0);
      $$pre$i$i$ph = $$pr112$lcssa$lcssa;$parser$idx$val$i$ph = $parser$idx6$val$lcssa1668;
      L204: while(1) {
       $222 = ((($parser$idx$val$i$ph)) + 12|0);
       $$phi$trans$insert$i16 = ((($parser$idx$val$i$ph)) + 8|0);
       $$phi$trans$insert8$i$phi$trans$insert$i = ((($parser$idx$val$i$ph)) + 8|0);
       $$pre$i$i = $$pre$i$i$ph;
       L206: while(1) {
        $phitmp$i$i = ($$pre$i$i|0)>(1);
        L208: do {
         if ($phitmp$i$i) {
          $$pre$i17 = HEAP32[$$phi$trans$insert$i16>>2]|0;
          $$pre634 = HEAP32[$parser$idx$val$i$ph>>2]|0;
          $$pre$phiZ2D = $parser$idx$val$i$ph;$236 = $$pre634;$237 = $$pre$i17;
         } else {
          $223 = ($$pre$i$i|0)>(0);
          $$pre$i7$pre$i = HEAP32[$parser$idx$val$i$ph>>2]|0;
          $$pre7$pre$i$pre$i = HEAP32[$$phi$trans$insert8$i$phi$trans$insert$i>>2]|0;
          L211: do {
           if ($223) {
            $231 = $$pre$i7$pre$i;
           } else {
            $224 = ($$pre$i7$pre$i>>>0)<($$pre7$pre$i$pre$i>>>0);
            L213: do {
             if ($224) {
              $226 = $$pre$i7$pre$i;
              L214: while(1) {
               $225 = ((($226)) + 1|0);
               HEAP32[$parser$idx$val$i$ph>>2] = $225;
               $227 = HEAP8[$226>>0]|0;
               $228 = $227&255;
               switch ($228|0) {
               case 26:  {
                $450 = $225;
                break L213;
                break;
               }
               case 10: case 13:  {
                $$lcssa1488 = $225;
                label = 153;
                break L214;
                break;
               }
               case 59:  {
                $$lcssa1489 = $225;
                label = 154;
                break L214;
                break;
               }
               case 9: case 32:  {
                break;
               }
               default: {
                $231 = $225;
                break L211;
               }
               }
               $229 = ($225>>>0)<($$pre7$pre$i$pre$i>>>0);
               if ($229) {
                $226 = $225;
               } else {
                $450 = $225;
                break L213;
               }
              }
              if ((label|0) == 153) {
               label = 0;
               HEAP32[$222>>2] = 2;
               $$pre$phiZ2D = $parser$idx$val$i$ph;$236 = $$lcssa1488;$237 = $$pre7$pre$i$pre$i;
               break L208;
              }
              else if ((label|0) == 154) {
               label = 0;
               HEAP32[$222>>2] = 1;
               $231 = $$lcssa1489;
               break L211;
              }
             } else {
              $450 = $$pre$i7$pre$i;
             }
            } while(0);
            HEAP32[$222>>2] = 3;
            $$pre$phiZ2D = $parser$idx$val$i$ph;$236 = $450;$237 = $$pre7$pre$i$pre$i;
            break L208;
           }
          } while(0);
          L221: while(1) {
           $230 = ($231>>>0)<($$pre7$pre$i$pre$i>>>0);
           if (!($230)) {
            $451 = $231;
            label = 159;
            break;
           }
           $232 = ((($231)) + 1|0);
           HEAP32[$parser$idx$val$i$ph>>2] = $232;
           $233 = HEAP8[$231>>0]|0;
           $234 = $233&255;
           switch ($234|0) {
           case 26:  {
            $451 = $232;
            label = 159;
            break L221;
            break;
           }
           case 10: case 13:  {
            $$lcssa1491 = $232;
            label = 158;
            break L221;
            break;
           }
           default: {
            $231 = $232;
           }
           }
          }
          if ((label|0) == 158) {
           label = 0;
           HEAP32[$222>>2] = 2;
           $$pre$phiZ2D = $parser$idx$val$i$ph;$236 = $$lcssa1491;$237 = $$pre7$pre$i$pre$i;
           break;
          }
          else if ((label|0) == 159) {
           label = 0;
           HEAP32[$222>>2] = 3;
           $$pre$phiZ2D = $parser$idx$val$i$ph;$236 = $451;$237 = $$pre7$pre$i$pre$i;
           break;
          }
         }
        } while(0);
        HEAP32[$222>>2] = 0;
        $235 = ($236>>>0)<($237>>>0);
        if ($235) {
         $239 = $236;
        } else {
         $244 = $222;
         label = 166;
         break L204;
        }
        L228: while(1) {
         $238 = ((($239)) + 1|0);
         HEAP32[$$pre$phiZ2D>>2] = $238;
         $240 = HEAP8[$239>>0]|0;
         $241 = $240&255;
         switch ($241|0) {
         case 59:  {
          $$lcssa1584 = $222;
          label = 165;
          break L204;
          break;
         }
         case 26:  {
          $244 = $222;
          label = 166;
          break L204;
          break;
         }
         case 9: case 32:  {
          $243 = ($238>>>0)<($237>>>0);
          if ($243) {
           $239 = $238;
           continue L228;
          } else {
           $244 = $222;
           label = 166;
           break L204;
          }
          break;
         }
         case 10: case 13:  {
          HEAP32[$222>>2] = 2;
          $254 = $238;
          break;
         }
         default: {
          $242 = $238;
          $246 = $238;$452 = $242;
          L232: while(1) {
           $245 = ($246>>>0)<($237>>>0);
           if (!($245)) {
            $453 = $246;$454 = $452;
            label = 171;
            break;
           }
           $247 = ((($246)) + 1|0);
           HEAP32[$$pre$phiZ2D>>2] = $247;
           $248 = HEAP8[$246>>0]|0;
           $249 = $248&255;
           $250 = $247;
           switch ($249|0) {
           case 26:  {
            $453 = $247;$454 = $250;
            label = 171;
            break L232;
            break;
           }
           case 9: case 32:  {
            $$pre$i$i$pre637 = 0;$455 = $247;$456 = $250;
            break L232;
            break;
           }
           case 10: case 13:  {
            $$lcssa1493 = $247;$$lcssa1496 = $250;
            label = 169;
            break L232;
            break;
           }
           case 59:  {
            $$lcssa1494 = $247;$$lcssa1497 = $250;
            label = 170;
            break L232;
            break;
           }
           default: {
            $246 = $247;$452 = $250;
           }
           }
          }
          if ((label|0) == 169) {
           label = 0;
           HEAP32[$222>>2] = 2;
           $$pre$i$i$pre637 = 2;$455 = $$lcssa1493;$456 = $$lcssa1496;
          }
          else if ((label|0) == 170) {
           label = 0;
           HEAP32[$222>>2] = 1;
           $$pre$i$i$pre637 = 1;$455 = $$lcssa1494;$456 = $$lcssa1497;
          }
          else if ((label|0) == 171) {
           label = 0;
           HEAP32[$222>>2] = 3;
           $$pre$i$i$pre637 = 3;$455 = $453;$456 = $454;
          }
          $251 = ($239|0)==(0|0);
          if (!($251)) {
           $$lcssa1500 = $239;$$lcssa1503 = $456;$$pre$i$i$pre637$lcssa = $$pre$i$i$pre637;
           break L228;
          }
          $252 = ($$pre$i$i$pre637|0)==(2);
          if ($252) {
           $254 = $455;
          } else {
           $error$4$ph = 160;
           label = 281;
           break L202;
          }
         }
         }
         HEAP32[$222>>2] = 0;
         $253 = ($254>>>0)<($237>>>0);
         if ($253) {
          $239 = $254;
         } else {
          $244 = $222;
          label = 166;
          break L204;
         }
        }
        $255 = $$lcssa1500;
        $256 = $255 ^ -1;
        $257 = (($$lcssa1503) + ($256))|0;
        $258 = HEAP8[$$lcssa1500>>0]|0;
        $n$04$i$i = 0;
        while(1) {
         $259 = (53740 + ($n$04$i$i<<2)|0);
         $260 = HEAP32[$259>>2]|0;
         $261 = HEAP8[$260>>0]|0;
         $262 = ($261<<24>>24)==($258<<24>>24);
         L247: do {
          if ($262) {
           $263 = $260;$n$1$i195$i = $n$04$i$i;
           while(1) {
            $264 = (_strncmp($263,$$lcssa1500,$257)|0);
            $265 = ($264|0)==(0);
            if ($265) {
             $n$1$i195$i$lcssa = $n$1$i195$i;
             break L206;
            }
            $266 = (($n$1$i195$i) + 1)|0;
            $$old1$i$i = ($266|0)<(74);
            if (!($$old1$i$i)) {
             $n$2$i$i = $266;
             break L247;
            }
            $$phi$trans$insert$i14$i = (53740 + ($266<<2)|0);
            $$pre$i15$i = HEAP32[$$phi$trans$insert$i14$i>>2]|0;
            $$pre6$i$i = HEAP8[$$pre$i15$i>>0]|0;
            $267 = ($$pre6$i$i<<24>>24)==($258<<24>>24);
            if ($267) {
             $263 = $$pre$i15$i;$n$1$i195$i = $266;
            } else {
             $$pre$i$i = $$pre$i$i$pre637$lcssa;
             continue L206;
            }
           }
          } else {
           $n$2$i$i = $n$04$i$i;
          }
         } while(0);
         $268 = (($n$2$i$i) + 1)|0;
         $269 = ($268|0)<(74);
         if ($269) {
          $n$04$i$i = $268;
         } else {
          $$pre$i$i = $$pre$i$i$pre637$lcssa;
          continue L206;
         }
        }
       }
       switch ($n$1$i195$i$lcssa|0) {
       case 20: case 21:  {
        label = 282;
        break L202;
        break;
       }
       case 53:  {
        $270 = HEAP32[$1>>2]|0;
        HEAP32[$val$i$i17$i>>2] = 3;
        $271 = (_afm_parser_read_vals($parser,$val$i$i17$i,1)|0);
        $272 = ($271|0)==(1);
        if (!($272)) {
         label = 183;
         break L204;
        }
        $273 = HEAP32[$204>>2]|0;
        $274 = ($273|0)<(0);
        if ($274) {
         $error$4$ph = 160;
         label = 281;
         break L202;
        }
        $275 = ((($270)) + 32|0);
        HEAP32[$275>>2] = $273;
        $276 = ($273|0)==(0);
        if ($276) {
         $$pre88$i$i = ((($270)) + 28|0);
         $$pre$phi$i22$iZ2D = $$pre88$i$i;
        } else {
         $277 = HEAP32[$parser>>2]|0;
         $278 = (_ft_mem_realloc($277,20,0,$273,0,$error$i18$i)|0);
         $279 = ((($270)) + 28|0);
         HEAP32[$279>>2] = $278;
         $280 = HEAP32[$error$i18$i>>2]|0;
         $281 = ($280|0)==(0);
         if ($281) {
          $$pre$phi$i22$iZ2D = $279;
         } else {
          $error$4$ph = $280;
          label = 281;
          break L202;
         }
        }
        $n$0$i$i = -1;
        L262: while(1) {
         $parser$idx$val$i$i = HEAP32[$parser$idx>>2]|0;
         $282 = ((($parser$idx$val$i$i)) + 12|0);
         $$pre$i$i25$i = HEAP32[$282>>2]|0;
         $phitmp$i$i26$i = ($$pre$i$i25$i|0)>(1);
         L264: do {
          if ($phitmp$i$i26$i) {
           $$phi$trans$insert$i27$i = ((($parser$idx$val$i$i)) + 8|0);
           $$pre$i28$i = HEAP32[$$phi$trans$insert$i27$i>>2]|0;
           $$pre313$i = HEAP32[$parser$idx$val$i$i>>2]|0;
           $$pre$phi$i18Z2D = $parser$idx$val$i$i;$296 = $$pre313$i;$297 = $$pre$i28$i;
          } else {
           $283 = ($$pre$i$i25$i|0)>(0);
           $$pre$i11$pre$i$i = HEAP32[$parser$idx$val$i$i>>2]|0;
           $$phi$trans$insert8$i$phi$trans$insert$i30$i = ((($parser$idx$val$i$i)) + 8|0);
           $$pre7$pre$i$pre$i31$i = HEAP32[$$phi$trans$insert8$i$phi$trans$insert$i30$i>>2]|0;
           L267: do {
            if ($283) {
             $291 = $$pre$i11$pre$i$i;
            } else {
             $284 = ($$pre$i11$pre$i$i>>>0)<($$pre7$pre$i$pre$i31$i>>>0);
             L269: do {
              if ($284) {
               $286 = $$pre$i11$pre$i$i;
               L270: while(1) {
                $285 = ((($286)) + 1|0);
                HEAP32[$parser$idx$val$i$i>>2] = $285;
                $287 = HEAP8[$286>>0]|0;
                $288 = $287&255;
                switch ($288|0) {
                case 26:  {
                 $457 = $285;
                 break L269;
                 break;
                }
                case 10: case 13:  {
                 $$lcssa1545 = $285;
                 label = 195;
                 break L270;
                 break;
                }
                case 59:  {
                 $$lcssa1546 = $285;
                 label = 196;
                 break L270;
                 break;
                }
                case 9: case 32:  {
                 break;
                }
                default: {
                 $291 = $285;
                 break L267;
                }
                }
                $289 = ($285>>>0)<($$pre7$pre$i$pre$i31$i>>>0);
                if ($289) {
                 $286 = $285;
                } else {
                 $457 = $285;
                 break L269;
                }
               }
               if ((label|0) == 195) {
                label = 0;
                HEAP32[$282>>2] = 2;
                $$pre$phi$i18Z2D = $parser$idx$val$i$i;$296 = $$lcssa1545;$297 = $$pre7$pre$i$pre$i31$i;
                break L264;
               }
               else if ((label|0) == 196) {
                label = 0;
                HEAP32[$282>>2] = 1;
                $291 = $$lcssa1546;
                break L267;
               }
              } else {
               $457 = $$pre$i11$pre$i$i;
              }
             } while(0);
             HEAP32[$282>>2] = 3;
             $$pre$phi$i18Z2D = $parser$idx$val$i$i;$296 = $457;$297 = $$pre7$pre$i$pre$i31$i;
             break L264;
            }
           } while(0);
           L277: while(1) {
            $290 = ($291>>>0)<($$pre7$pre$i$pre$i31$i>>>0);
            if (!($290)) {
             $458 = $291;
             label = 201;
             break;
            }
            $292 = ((($291)) + 1|0);
            HEAP32[$parser$idx$val$i$i>>2] = $292;
            $293 = HEAP8[$291>>0]|0;
            $294 = $293&255;
            switch ($294|0) {
            case 26:  {
             $458 = $292;
             label = 201;
             break L277;
             break;
            }
            case 10: case 13:  {
             $$lcssa1548 = $292;
             label = 200;
             break L277;
             break;
            }
            default: {
             $291 = $292;
            }
            }
           }
           if ((label|0) == 200) {
            label = 0;
            HEAP32[$282>>2] = 2;
            $$pre$phi$i18Z2D = $parser$idx$val$i$i;$296 = $$lcssa1548;$297 = $$pre7$pre$i$pre$i31$i;
            break;
           }
           else if ((label|0) == 201) {
            label = 0;
            HEAP32[$282>>2] = 3;
            $$pre$phi$i18Z2D = $parser$idx$val$i$i;$296 = $458;$297 = $$pre7$pre$i$pre$i31$i;
            break;
           }
          }
         } while(0);
         HEAP32[$282>>2] = 0;
         $295 = ($296>>>0)<($297>>>0);
         if ($295) {
          $299 = $296;
         } else {
          $304 = $282;
          label = 208;
          break L204;
         }
         L284: while(1) {
          $298 = ((($299)) + 1|0);
          HEAP32[$$pre$phi$i18Z2D>>2] = $298;
          $300 = HEAP8[$299>>0]|0;
          $301 = $300&255;
          switch ($301|0) {
          case 59:  {
           $$lcssa1578 = $282;
           label = 207;
           break L204;
           break;
          }
          case 26:  {
           $304 = $282;
           label = 208;
           break L204;
           break;
          }
          case 9: case 32:  {
           $303 = ($298>>>0)<($297>>>0);
           if ($303) {
            $299 = $298;
            continue L284;
           } else {
            $304 = $282;
            label = 208;
            break L204;
           }
           break;
          }
          case 10: case 13:  {
           HEAP32[$282>>2] = 2;
           $314 = $298;
           break;
          }
          default: {
           $302 = $298;
           $306 = $298;$459 = $302;
           L288: while(1) {
            $305 = ($306>>>0)<($297>>>0);
            if (!($305)) {
             $460 = $306;$461 = $459;
             label = 213;
             break;
            }
            $307 = ((($306)) + 1|0);
            HEAP32[$$pre$phi$i18Z2D>>2] = $307;
            $308 = HEAP8[$306>>0]|0;
            $309 = $308&255;
            $310 = $307;
            switch ($309|0) {
            case 26:  {
             $460 = $307;$461 = $310;
             label = 213;
             break L288;
             break;
            }
            case 9: case 32:  {
             $$pre$i$i$pre640 = 0;$462 = $307;$463 = $310;
             break L288;
             break;
            }
            case 10: case 13:  {
             $$lcssa1550 = $307;$$lcssa1553 = $310;
             label = 211;
             break L288;
             break;
            }
            case 59:  {
             $$lcssa1551 = $307;$$lcssa1554 = $310;
             label = 212;
             break L288;
             break;
            }
            default: {
             $306 = $307;$459 = $310;
            }
            }
           }
           if ((label|0) == 211) {
            label = 0;
            HEAP32[$282>>2] = 2;
            $$pre$i$i$pre640 = 2;$462 = $$lcssa1550;$463 = $$lcssa1553;
           }
           else if ((label|0) == 212) {
            label = 0;
            HEAP32[$282>>2] = 1;
            $$pre$i$i$pre640 = 1;$462 = $$lcssa1551;$463 = $$lcssa1554;
           }
           else if ((label|0) == 213) {
            label = 0;
            HEAP32[$282>>2] = 3;
            $$pre$i$i$pre640 = 3;$462 = $460;$463 = $461;
           }
           $311 = ($299|0)==(0|0);
           if (!($311)) {
            $$lcssa1557 = $299;$$lcssa1561 = $463;$$pre$i$i$pre640$lcssa = $$pre$i$i$pre640;
            break L284;
           }
           $312 = ($$pre$i$i$pre640|0)==(2);
           if ($312) {
            $314 = $462;
           } else {
            $error$4$ph = 160;
            label = 281;
            break L202;
           }
          }
          }
          HEAP32[$282>>2] = 0;
          $313 = ($314>>>0)<($297>>>0);
          if ($313) {
           $299 = $314;
          } else {
           $304 = $282;
           label = 208;
           break L204;
          }
         }
         $315 = $$lcssa1557;
         $316 = $315 ^ -1;
         $317 = (($$lcssa1561) + ($316))|0;
         $318 = HEAP8[$$lcssa1557>>0]|0;
         $n$04$i$i47$i = 0;
         L301: while(1) {
          $319 = (53740 + ($n$04$i$i47$i<<2)|0);
          $320 = HEAP32[$319>>2]|0;
          $321 = HEAP8[$320>>0]|0;
          $322 = ($321<<24>>24)==($318<<24>>24);
          L303: do {
           if ($322) {
            $323 = $320;$n$1$i52$i$i = $n$04$i$i47$i;
            while(1) {
             $324 = (_strncmp($323,$$lcssa1557,$317)|0);
             $325 = ($324|0)==(0);
             if ($325) {
              $n$1$i52$i$i$lcssa = $n$1$i52$i$i;
              label = 223;
              break L301;
             }
             $326 = (($n$1$i52$i$i) + 1)|0;
             $$old1$i$i49$i = ($326|0)<(74);
             if (!($$old1$i$i49$i)) {
              $n$2$i$i52$i = $326;
              break L303;
             }
             $$phi$trans$insert$i14$i$i = (53740 + ($326<<2)|0);
             $$pre$i15$i$i = HEAP32[$$phi$trans$insert$i14$i$i>>2]|0;
             $$pre6$i$i50$i = HEAP8[$$pre$i15$i$i>>0]|0;
             $327 = ($$pre6$i$i50$i<<24>>24)==($318<<24>>24);
             if ($327) {
              $323 = $$pre$i15$i$i;$n$1$i52$i$i = $326;
             } else {
              $n$1$i55$i = $n$0$i$i;
              break L301;
             }
            }
           } else {
            $n$2$i$i52$i = $n$04$i$i47$i;
           }
          } while(0);
          $328 = (($n$2$i$i52$i) + 1)|0;
          $329 = ($328|0)<(74);
          if ($329) {
           $n$04$i$i47$i = $328;
          } else {
           $n$1$i55$i = $n$0$i$i;
           break;
          }
         }
         if ((label|0) == 223) {
          label = 0;
          switch ($n$1$i52$i$i$lcssa|0) {
          case 20: case 21: case 23:  {
           $$pre$i$i$pre640$lcssa$lcssa1560 = $$pre$i$i$pre640$lcssa;$n$0$i$i$lcssa1570 = $n$0$i$i;$parser$idx$val$i$i$lcssa1575 = $parser$idx$val$i$i;
           break L262;
           break;
          }
          case 56:  {
           break;
          }
          default: {
           label = 227;
           break L204;
          }
          }
          $330 = (($n$0$i$i) + 1)|0;
          $331 = HEAP32[$275>>2]|0;
          $332 = ($330|0)<($331|0);
          if (!($332)) {
           label = 227;
           break L204;
          }
          $333 = HEAP32[$$pre$phi$i22$iZ2D>>2]|0;
          HEAP32[$shared_vals$i19$i>>2] = 3;
          HEAP32[$205>>2] = 2;
          HEAP32[$206>>2] = 2;
          HEAP32[$207>>2] = 2;
          HEAP32[$208>>2] = 2;
          $334 = (_afm_parser_read_vals($parser,$shared_vals$i19$i,5)|0);
          $335 = ($334|0)==(5);
          if (!($335)) {
           label = 227;
           break L204;
          }
          $336 = HEAP32[$209>>2]|0;
          $337 = (($333) + (($330*20)|0)|0);
          HEAP32[$337>>2] = $336;
          $338 = HEAP32[$210>>2]|0;
          $339 = (((($333) + (($330*20)|0)|0)) + 4|0);
          HEAP32[$339>>2] = $338;
          $340 = HEAP32[$211>>2]|0;
          $341 = (((($333) + (($330*20)|0)|0)) + 8|0);
          HEAP32[$341>>2] = $340;
          $342 = HEAP32[$212>>2]|0;
          $343 = (((($333) + (($330*20)|0)|0)) + 12|0);
          HEAP32[$343>>2] = $342;
          $344 = HEAP32[$213>>2]|0;
          $345 = (((($333) + (($330*20)|0)|0)) + 16|0);
          HEAP32[$345>>2] = $344;
          $n$1$i55$i = $330;
         }
         $n$0$i$i = $n$1$i55$i;
        }
        $346 = (($n$0$i$i$lcssa1570) + 1)|0;
        HEAP32[$275>>2] = $346;
        $$pre$i$i$ph = $$pre$i$i$pre640$lcssa$lcssa1560;$parser$idx$val$i$ph = $parser$idx$val$i$i$lcssa1575;
        continue L204;
        break;
       }
       case 51: case 50:  {
        break;
       }
       default: {
        $error$4$ph = 160;
        label = 281;
        break L202;
       }
       }
       $347 = HEAP32[$1>>2]|0;
       HEAP32[$val$i$i$i>>2] = 3;
       $348 = (_afm_parser_read_vals($parser,$val$i$i$i,1)|0);
       $349 = ($348|0)==(1);
       if (!($349)) {
        label = 231;
        break;
       }
       $350 = HEAP32[$214>>2]|0;
       $351 = ($350|0)<(0);
       if ($351) {
        $error$4$ph = 160;
        label = 281;
        break L202;
       }
       $352 = ((($347)) + 40|0);
       HEAP32[$352>>2] = $350;
       $353 = ($350|0)==(0);
       if ($353) {
        $$pre103$i$i = ((($347)) + 36|0);
        $$pre$phi$i10$iZ2D = $$pre103$i$i;
       } else {
        $354 = HEAP32[$parser>>2]|0;
        $355 = (_ft_mem_realloc($354,16,0,$350,0,$error$i$i)|0);
        $356 = ((($347)) + 36|0);
        HEAP32[$356>>2] = $355;
        $357 = HEAP32[$error$i$i>>2]|0;
        $358 = ($357|0)==(0);
        if ($358) {
         $$pre$phi$i10$iZ2D = $356;
        } else {
         $error$4$ph = $357;
         label = 281;
         break L202;
        }
       }
       $n$0$ph$i$i = 0;
       L323: while(1) {
        $parser$idx$val$pre$i$i = HEAP32[$parser$idx>>2]|0;
        $$phi$trans$insert$i11$i = ((($parser$idx$val$pre$i$i)) + 12|0);
        $$pre$i$pre$i$i = HEAP32[$$phi$trans$insert$i11$i>>2]|0;
        $$phi$trans$insert100$i$i = ((($parser$idx$val$pre$i$i)) + 8|0);
        $$pre$i$i$i = $$pre$i$pre$i$i;
        L325: while(1) {
         $phitmp$i$i$i = ($$pre$i$i$i|0)>(1);
         L327: do {
          if ($phitmp$i$i$i) {
           $$pre$i12$i = HEAP32[$$phi$trans$insert100$i$i>>2]|0;
           $$pre311$i = HEAP32[$parser$idx$val$pre$i$i>>2]|0;
           $372 = $$pre311$i;$373 = $$pre$i12$i;
          } else {
           $359 = ($$pre$i$i$i|0)>(0);
           $$pre$i15$pre$i$i = HEAP32[$parser$idx$val$pre$i$i>>2]|0;
           $$pre7$pre$i$pre$i$i = HEAP32[$$phi$trans$insert100$i$i>>2]|0;
           L330: do {
            if ($359) {
             $367 = $$pre$i15$pre$i$i;
            } else {
             $360 = ($$pre$i15$pre$i$i>>>0)<($$pre7$pre$i$pre$i$i>>>0);
             L332: do {
              if ($360) {
               $362 = $$pre$i15$pre$i$i;
               L333: while(1) {
                $361 = ((($362)) + 1|0);
                HEAP32[$parser$idx$val$pre$i$i>>2] = $361;
                $363 = HEAP8[$362>>0]|0;
                $364 = $363&255;
                switch ($364|0) {
                case 26:  {
                 $464 = $361;
                 break L332;
                 break;
                }
                case 10: case 13:  {
                 $$lcssa1510 = $361;
                 label = 244;
                 break L333;
                 break;
                }
                case 59:  {
                 $$lcssa1511 = $361;
                 label = 245;
                 break L333;
                 break;
                }
                case 9: case 32:  {
                 break;
                }
                default: {
                 $367 = $361;
                 break L330;
                }
                }
                $365 = ($361>>>0)<($$pre7$pre$i$pre$i$i>>>0);
                if ($365) {
                 $362 = $361;
                } else {
                 $464 = $361;
                 break L332;
                }
               }
               if ((label|0) == 244) {
                label = 0;
                HEAP32[$$phi$trans$insert$i11$i>>2] = 2;
                $372 = $$lcssa1510;$373 = $$pre7$pre$i$pre$i$i;
                break L327;
               }
               else if ((label|0) == 245) {
                label = 0;
                HEAP32[$$phi$trans$insert$i11$i>>2] = 1;
                $367 = $$lcssa1511;
                break L330;
               }
              } else {
               $464 = $$pre$i15$pre$i$i;
              }
             } while(0);
             HEAP32[$$phi$trans$insert$i11$i>>2] = 3;
             $372 = $464;$373 = $$pre7$pre$i$pre$i$i;
             break L327;
            }
           } while(0);
           L340: while(1) {
            $366 = ($367>>>0)<($$pre7$pre$i$pre$i$i>>>0);
            if (!($366)) {
             $465 = $367;
             label = 250;
             break;
            }
            $368 = ((($367)) + 1|0);
            HEAP32[$parser$idx$val$pre$i$i>>2] = $368;
            $369 = HEAP8[$367>>0]|0;
            $370 = $369&255;
            switch ($370|0) {
            case 26:  {
             $465 = $368;
             label = 250;
             break L340;
             break;
            }
            case 10: case 13:  {
             $$lcssa1513 = $368;
             label = 249;
             break L340;
             break;
            }
            default: {
             $367 = $368;
            }
            }
           }
           if ((label|0) == 249) {
            label = 0;
            HEAP32[$$phi$trans$insert$i11$i>>2] = 2;
            $372 = $$lcssa1513;$373 = $$pre7$pre$i$pre$i$i;
            break;
           }
           else if ((label|0) == 250) {
            label = 0;
            HEAP32[$$phi$trans$insert$i11$i>>2] = 3;
            $372 = $465;$373 = $$pre7$pre$i$pre$i$i;
            break;
           }
          }
         } while(0);
         HEAP32[$$phi$trans$insert$i11$i>>2] = 0;
         $371 = ($372>>>0)<($373>>>0);
         if ($371) {
          $375 = $372;
         } else {
          $$phi$trans$insert$i11$i1543 = $$phi$trans$insert$i11$i;
          label = 257;
          break L204;
         }
         L347: while(1) {
          $374 = ((($375)) + 1|0);
          HEAP32[$parser$idx$val$pre$i$i>>2] = $374;
          $376 = HEAP8[$375>>0]|0;
          $377 = $376&255;
          switch ($377|0) {
          case 59:  {
           $$phi$trans$insert$i11$i$lcssa1537 = $$phi$trans$insert$i11$i;
           label = 256;
           break L204;
           break;
          }
          case 26:  {
           $$phi$trans$insert$i11$i1543 = $$phi$trans$insert$i11$i;
           label = 257;
           break L204;
           break;
          }
          case 9: case 32:  {
           $379 = ($374>>>0)<($373>>>0);
           if ($379) {
            $375 = $374;
            continue L347;
           } else {
            $$phi$trans$insert$i11$i1543 = $$phi$trans$insert$i11$i;
            label = 257;
            break L204;
           }
           break;
          }
          case 10: case 13:  {
           HEAP32[$$phi$trans$insert$i11$i>>2] = 2;
           $389 = $374;
           break;
          }
          default: {
           $378 = $374;
           $381 = $374;$466 = $378;
           L351: while(1) {
            $380 = ($381>>>0)<($373>>>0);
            if (!($380)) {
             $467 = $381;$468 = $466;
             label = 262;
             break;
            }
            $382 = ((($381)) + 1|0);
            HEAP32[$parser$idx$val$pre$i$i>>2] = $382;
            $383 = HEAP8[$381>>0]|0;
            $384 = $383&255;
            $385 = $382;
            switch ($384|0) {
            case 26:  {
             $467 = $382;$468 = $385;
             label = 262;
             break L351;
             break;
            }
            case 9: case 32:  {
             $$pr$i$i = 0;$469 = $382;$470 = $385;
             break L351;
             break;
            }
            case 10: case 13:  {
             $$lcssa1515 = $382;$$lcssa1518 = $385;
             label = 260;
             break L351;
             break;
            }
            case 59:  {
             $$lcssa1516 = $382;$$lcssa1519 = $385;
             label = 261;
             break L351;
             break;
            }
            default: {
             $381 = $382;$466 = $385;
            }
            }
           }
           if ((label|0) == 260) {
            label = 0;
            HEAP32[$$phi$trans$insert$i11$i>>2] = 2;
            $$pr$i$i = 2;$469 = $$lcssa1515;$470 = $$lcssa1518;
           }
           else if ((label|0) == 261) {
            label = 0;
            HEAP32[$$phi$trans$insert$i11$i>>2] = 1;
            $$pr$i$i = 1;$469 = $$lcssa1516;$470 = $$lcssa1519;
           }
           else if ((label|0) == 262) {
            label = 0;
            HEAP32[$$phi$trans$insert$i11$i>>2] = 3;
            $$pr$i$i = 3;$469 = $467;$470 = $468;
           }
           $386 = ($375|0)==(0|0);
           if (!($386)) {
            $$lcssa1522 = $375;$$lcssa1525 = $470;$$pr$i$i$lcssa = $$pr$i$i;
            break L347;
           }
           $387 = ($$pr$i$i|0)==(2);
           if ($387) {
            $389 = $469;
           } else {
            $error$4$ph = 160;
            label = 281;
            break L202;
           }
          }
          }
          HEAP32[$$phi$trans$insert$i11$i>>2] = 0;
          $388 = ($389>>>0)<($373>>>0);
          if ($388) {
           $375 = $389;
          } else {
           $$phi$trans$insert$i11$i1543 = $$phi$trans$insert$i11$i;
           label = 257;
           break L204;
          }
         }
         $390 = $$lcssa1522;
         $391 = $390 ^ -1;
         $392 = (($$lcssa1525) + ($391))|0;
         $393 = HEAP8[$$lcssa1522>>0]|0;
         $n$04$i$i$i = 0;
         while(1) {
          $394 = (53740 + ($n$04$i$i$i<<2)|0);
          $395 = HEAP32[$394>>2]|0;
          $396 = HEAP8[$395>>0]|0;
          $397 = ($396<<24>>24)==($393<<24>>24);
          L366: do {
           if ($397) {
            $398 = $395;$n$1$i60$i$i = $n$04$i$i$i;
            while(1) {
             $399 = (_strncmp($398,$$lcssa1522,$392)|0);
             $400 = ($399|0)==(0);
             if ($400) {
              $n$1$i60$i$i$lcssa = $n$1$i60$i$i;
              break L325;
             }
             $401 = (($n$1$i60$i$i) + 1)|0;
             $$old1$i$i$i = ($401|0)<(74);
             if (!($$old1$i$i$i)) {
              $n$2$i$i$i = $401;
              break L366;
             }
             $$phi$trans$insert$i18$i$i = (53740 + ($401<<2)|0);
             $$pre$i19$i$i = HEAP32[$$phi$trans$insert$i18$i$i>>2]|0;
             $$pre6$i$i$i = HEAP8[$$pre$i19$i$i>>0]|0;
             $402 = ($$pre6$i$i$i<<24>>24)==($393<<24>>24);
             if ($402) {
              $398 = $$pre$i19$i$i;$n$1$i60$i$i = $401;
             } else {
              $$pre$i$i$i = $$pr$i$i$lcssa;
              continue L325;
             }
            }
           } else {
            $n$2$i$i$i = $n$04$i$i$i;
           }
          } while(0);
          $403 = (($n$2$i$i$i) + 1)|0;
          $404 = ($403|0)<(74);
          if ($404) {
           $n$04$i$i$i = $403;
          } else {
           $$pre$i$i$i = $$pr$i$i$lcssa;
           continue L325;
          }
         }
        }
        switch ($n$1$i60$i$i$lcssa|0) {
        case 20: case 21: case 22:  {
         $n$0$ph$i$i$lcssa1534 = $n$0$ph$i$i;
         break L323;
         break;
        }
        case 37: case 36: case 34:  {
         break;
        }
        default: {
         $error$4$ph = 160;
         label = 281;
         break L202;
        }
        }
        $405 = HEAP32[$352>>2]|0;
        $406 = ($n$0$ph$i$i|0)<($405|0);
        if (!($406)) {
         label = 278;
         break L204;
        }
        $407 = HEAP32[$$pre$phi$i10$iZ2D>>2]|0;
        HEAP32[$shared_vals$i$i>>2] = 5;
        HEAP32[$215>>2] = 5;
        HEAP32[$216>>2] = 3;
        HEAP32[$217>>2] = 3;
        $408 = (_afm_parser_read_vals($parser,$shared_vals$i$i,4)|0);
        $409 = ($408|0)<(3);
        if ($409) {
         label = 278;
         break L204;
        }
        $410 = HEAP32[$218>>2]|0;
        $411 = (($407) + ($n$0$ph$i$i<<4)|0);
        HEAP32[$411>>2] = $410;
        $412 = HEAP32[$219>>2]|0;
        $413 = (((($407) + ($n$0$ph$i$i<<4)|0)) + 4|0);
        HEAP32[$413>>2] = $412;
        $414 = ($n$1$i60$i$i$lcssa|0)==(37);
        $415 = HEAP32[$220>>2]|0;
        if ($414) {
         $$sink$i$i = $415;$$sink93$i$i = 0;
        } else {
         $416 = ($n$1$i60$i$i$lcssa|0)==(34);
         $417 = ($408|0)==(4);
         $or$cond$i$i = $416 & $417;
         $418 = HEAP32[$221>>2]|0;
         $419 = $or$cond$i$i ? $418 : 0;
         $$sink$i$i = $419;$$sink93$i$i = $415;
        }
        $420 = (((($407) + ($n$0$ph$i$i<<4)|0)) + 8|0);
        HEAP32[$420>>2] = $$sink93$i$i;
        $421 = (((($407) + ($n$0$ph$i$i<<4)|0)) + 12|0);
        HEAP32[$421>>2] = $$sink$i$i;
        $phitmp$i13$i = (($n$0$ph$i$i) + 1)|0;
        $n$0$ph$i$i = $phitmp$i13$i;
       }
       HEAP32[$352>>2] = $n$0$ph$i$i$lcssa1534;
       $422 = HEAP32[$$pre$phi$i10$iZ2D>>2]|0;
       _qsort($422,$n$0$ph$i$i$lcssa1534,16,4366);
       $parser$idx$val$i$pre$pre = HEAP32[$parser$idx>>2]|0;
       $$phi$trans$insert629$phi$trans$insert = ((($parser$idx$val$i$pre$pre)) + 12|0);
       $$pre$i$i$pre$pre = HEAP32[$$phi$trans$insert629$phi$trans$insert>>2]|0;
       $$pre$i$i$ph = $$pre$i$i$pre$pre;$parser$idx$val$i$ph = $parser$idx$val$i$pre$pre;
      }
      switch (label|0) {
       case 165: {
        HEAP32[$$lcssa1584>>2] = 1;
        $error$4$ph = 160;
        label = 281;
        break L202;
        break;
       }
       case 166: {
        HEAP32[$244>>2] = 3;
        $error$4$ph = 160;
        label = 281;
        break L202;
        break;
       }
       case 183: {
        $error$4$ph = 160;
        label = 281;
        break L202;
        break;
       }
       case 207: {
        HEAP32[$$lcssa1578>>2] = 1;
        $error$4$ph = 160;
        label = 281;
        break L202;
        break;
       }
       case 208: {
        HEAP32[$304>>2] = 3;
        $error$4$ph = 160;
        label = 281;
        break L202;
        break;
       }
       case 227: {
        $error$4$ph = 160;
        label = 281;
        break L202;
        break;
       }
       case 231: {
        $error$4$ph = 160;
        label = 281;
        break L202;
        break;
       }
       case 256: {
        HEAP32[$$phi$trans$insert$i11$i$lcssa1537>>2] = 1;
        $error$4$ph = 160;
        label = 281;
        break L202;
        break;
       }
       case 257: {
        HEAP32[$$phi$trans$insert$i11$i1543>>2] = 3;
        $error$4$ph = 160;
        label = 281;
        break L202;
        break;
       }
       case 278: {
        $error$4$ph = 160;
        label = 281;
        break L202;
        break;
       }
      }
     }
     else if ((label|0) == 283) {
      $error$5 = $error$0$lcssa1686;
     }
    } while(0);
    if ((label|0) == 144) {
     $$4 = 160;
     STACKTOP = sp;return ($$4|0);
    }
    else if ((label|0) == 281) {
     $error$5 = $error$4$ph;
    }
    else if ((label|0) == 282) {
     $$4 = 0;
     STACKTOP = sp;return ($$4|0);
    }
    $423 = ((($2)) + 28|0);
    $424 = HEAP32[$423>>2]|0;
    _ft_mem_free($0,$424);
    HEAP32[$423>>2] = 0;
    $425 = ((($2)) + 32|0);
    HEAP32[$425>>2] = 0;
    $426 = ((($2)) + 36|0);
    $427 = HEAP32[$426>>2]|0;
    _ft_mem_free($0,$427);
    HEAP32[$426>>2] = 0;
    $428 = ((($2)) + 40|0);
    HEAP32[$428>>2] = 0;
    HEAP8[$2>>0] = 0;
    $$4 = $error$5;
    STACKTOP = sp;return ($$4|0);
   }
   else if ((label|0) == 285) {
    STACKTOP = sp;return ($$4|0);
   }
  }
 } while(0);
 HEAP32[$4>>2] = 3;
 $$4 = 2;
 STACKTOP = sp;return ($$4|0);
}
function _afm_parser_read_vals($parser,$vals,$n) {
 $parser = $parser|0;
 $vals = $vals|0;
 $n = $n|0;
 var $$0 = 0, $$0$i9 = 0, $$in = 0, $$lcssa = 0, $$lcssa185 = 0, $$lcssa190 = 0, $$lcssa191 = 0, $$lcssa194 = 0, $$lcssa196 = 0, $$phi$trans$insert8$i = 0, $$pre$i = 0, $$pre7$i = 0, $$pre7$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $base$off$i = 0, $error = 0;
 var $have_overflow$0$lcssa$i = 0, $have_overflow$0$lcssa$i23 = 0, $have_overflow$07$i = 0, $have_overflow$07$i16 = 0, $have_overflow$2$ph$i = 0, $have_overflow$2$ph$i19 = 0, $i$066 = 0, $i$066$lcssa198 = 0, $i$066$lcssa199 = 0, $i$066$lcssa203 = 0, $i$066205 = 0, $i$066206 = 0, $i$066207 = 0, $n$off = 0, $num$0$i = 0, $num$0$lcssa$i = 0, $num$0$lcssa$i22 = 0, $num$08$i = 0, $num$08$i15 = 0, $num$2$ph$i = 0;
 var $num$2$ph$i20 = 0, $num$3$$i = 0, $num$3$$i24 = 0, $num$5$i = 0, $num$5$i25 = 0, $or$cond$i = 0, $or$cond4$i = 0, $or$cond44$i = 0, $or$cond44$i18 = 0, $p$0$i = 0, $p$0$i12 = 0, $p$1$lcssa$i = 0, $p$1$lcssa$i21 = 0, $p$19$i = 0, $p$19$i14 = 0, $sext$i = 0, $sign$0$i = 0, $sign$0$i11 = 0, $storemerge = 0, $str = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $str = sp + 4|0;
 $error = sp;
 $0 = ((($parser)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $n$off = (($n) + -1)|0;
 $2 = ($n$off>>>0)<(5);
 if (!($2)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ((($1)) + 12|0);
 $$phi$trans$insert8$i = ((($1)) + 8|0);
 $4 = ((($parser)) + 12|0);
 $5 = ((($parser)) + 16|0);
 $i$066 = 0;
 L4: while(1) {
  $6 = (($vals) + ($i$066<<3)|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)==(0);
  $9 = HEAP32[$3>>2]|0;
  $10 = ($9|0)>(0);
  if ($8) {
   L8: do {
    if ($10) {
     label = 11;
    } else {
     $11 = HEAP32[$1>>2]|0;
     $12 = HEAP32[$$phi$trans$insert8$i>>2]|0;
     $13 = ($11>>>0)<($12>>>0);
     if ($13) {
      $15 = $11;
     } else {
      $i$066207 = $i$066;
      label = 10;
      break L4;
     }
     L10: while(1) {
      $14 = ((($15)) + 1|0);
      HEAP32[$1>>2] = $14;
      $16 = HEAP8[$15>>0]|0;
      $17 = $16&255;
      switch ($17|0) {
      case 10: case 13:  {
       $i$066$lcssa203 = $i$066;
       label = 8;
       break L4;
       break;
      }
      case 26:  {
       $i$066207 = $i$066;
       label = 10;
       break L4;
       break;
      }
      case 59:  {
       $$lcssa194 = $14;
       break L10;
       break;
      }
      case 9: case 32:  {
       break;
      }
      default: {
       label = 11;
       break L8;
      }
      }
      $18 = ($14>>>0)<($12>>>0);
      if ($18) {
       $15 = $14;
      } else {
       $i$066207 = $i$066;
       label = 10;
       break L4;
      }
     }
     HEAP32[$3>>2] = 1;
     $$in = $$lcssa194;$$pre7$i = $12;
    }
   } while(0);
   if ((label|0) == 11) {
    label = 0;
    $19 = ($9|0)>(1);
    if ($19) {
     $i$066205 = $i$066;
     break;
    }
    $$pre$i = HEAP32[$1>>2]|0;
    $$pre7$pre$i = HEAP32[$$phi$trans$insert8$i>>2]|0;
    $$in = $$pre$i;$$pre7$i = $$pre7$pre$i;
   }
   $20 = $$in;
   $21 = ((($$in)) + -1|0);
   $140 = $20;$23 = $$in;
   L18: while(1) {
    $22 = ($23>>>0)<($$pre7$i>>>0);
    if (!($22)) {
     $141 = $140;$storemerge = 3;
     break;
    }
    $24 = ((($23)) + 1|0);
    HEAP32[$1>>2] = $24;
    $25 = HEAP8[$23>>0]|0;
    $26 = $25&255;
    $27 = $24;
    switch ($26|0) {
    case 26:  {
     $141 = $27;$storemerge = 3;
     break L18;
     break;
    }
    case 10: case 13:  {
     $$lcssa196 = $27;
     label = 30;
     break L18;
     break;
    }
    default: {
     $140 = $27;$23 = $24;
    }
    }
   }
   if ((label|0) == 30) {
    label = 0;
    $141 = $$lcssa196;$storemerge = 2;
   }
   HEAP32[$3>>2] = $storemerge;
   HEAP32[$str>>2] = $21;
   $45 = $21;$47 = $141;
  } else {
   if ($10) {
    $i$066205 = $i$066;
    break;
   }
   $28 = HEAP32[$1>>2]|0;
   $29 = HEAP32[$$phi$trans$insert8$i>>2]|0;
   $30 = ($28>>>0)<($29>>>0);
   if ($30) {
    $32 = $28;
   } else {
    $i$066206 = $i$066;
    label = 23;
    break;
   }
   L26: while(1) {
    $31 = ((($32)) + 1|0);
    HEAP32[$1>>2] = $31;
    $33 = HEAP8[$32>>0]|0;
    $34 = $33&255;
    switch ($34|0) {
    case 10: case 13:  {
     $i$066$lcssa198 = $i$066;
     label = 21;
     break L4;
     break;
    }
    case 59:  {
     $i$066$lcssa199 = $i$066;
     label = 22;
     break L4;
     break;
    }
    case 26:  {
     $i$066206 = $i$066;
     label = 23;
     break L4;
     break;
    }
    case 9: case 32:  {
     break;
    }
    default: {
     $$lcssa = $32;$$lcssa185 = $31;
     break L26;
    }
    }
    $36 = ($31>>>0)<($29>>>0);
    if ($36) {
     $32 = $31;
    } else {
     $i$066206 = $i$066;
     label = 23;
     break L4;
    }
   }
   $35 = $$lcssa185;
   $142 = $35;$38 = $$lcssa185;
   L30: while(1) {
    $37 = ($38>>>0)<($29>>>0);
    if (!($37)) {
     $143 = $142;
     label = 28;
     break;
    }
    $39 = ((($38)) + 1|0);
    HEAP32[$1>>2] = $39;
    $40 = HEAP8[$38>>0]|0;
    $41 = $40&255;
    $42 = $39;
    switch ($41|0) {
    case 26:  {
     $143 = $42;
     label = 28;
     break L30;
     break;
    }
    case 9: case 32:  {
     $144 = $42;
     break L30;
     break;
    }
    case 10: case 13:  {
     $$lcssa190 = $42;
     label = 26;
     break L30;
     break;
    }
    case 59:  {
     $$lcssa191 = $42;
     label = 27;
     break L30;
     break;
    }
    default: {
     $142 = $42;$38 = $39;
    }
    }
   }
   if ((label|0) == 26) {
    label = 0;
    HEAP32[$3>>2] = 2;
    $144 = $$lcssa190;
   }
   else if ((label|0) == 27) {
    label = 0;
    HEAP32[$3>>2] = 1;
    $144 = $$lcssa191;
   }
   else if ((label|0) == 28) {
    label = 0;
    HEAP32[$3>>2] = 3;
    $144 = $143;
   }
   HEAP32[$str>>2] = $$lcssa;
   $43 = ($$lcssa|0)==(0|0);
   if ($43) {
    $$0 = $i$066;
    label = 74;
    break;
   } else {
    $45 = $$lcssa;$47 = $144;
   }
  }
  $44 = $45;
  $46 = (($47) - ($44))|0;
  $48 = (($46) + -1)|0;
  L39: do {
   switch ($7|0) {
   case 1: case 0:  {
    $49 = HEAP32[$parser>>2]|0;
    $50 = (_ft_mem_qalloc($49,$46,$error)|0);
    $51 = (((($vals) + ($i$066<<3)|0)) + 4|0);
    HEAP32[$51>>2] = $50;
    $52 = HEAP32[$error>>2]|0;
    $53 = ($52|0)==(0);
    if ($53) {
     _memcpy(($50|0),($45|0),($48|0))|0;
     $54 = HEAP32[$51>>2]|0;
     $55 = (($54) + ($48)|0);
     HEAP8[$55>>0] = 0;
    }
    break;
   }
   case 2:  {
    $56 = (($45) + ($48)|0);
    $57 = (_PS_Conv_ToFixed($str,$56,0)|0);
    $58 = (((($vals) + ($i$066<<3)|0)) + 4|0);
    HEAP32[$58>>2] = $57;
    break;
   }
   case 3:  {
    $59 = (($45) + ($48)|0);
    $60 = ($46|0)<(2);
    L46: do {
     if ($60) {
      $$0$i9 = 0;
     } else {
      $61 = HEAP8[$45>>0]|0;
      switch ($61<<24>>24) {
      case 43: case 45:  {
       $62 = ($61<<24>>24)==(45);
       $63 = $62&1;
       $64 = ((($45)) + 1|0);
       $65 = ($48|0)==(1);
       if ($65) {
        $$0$i9 = 0;
        break L46;
       } else {
        $p$0$i12 = $64;$sign$0$i11 = $63;
       }
       break;
      }
      default: {
       $p$0$i12 = $45;$sign$0$i11 = 0;
      }
      }
      $66 = ($p$0$i12>>>0)<($59>>>0);
      L51: do {
       if ($66) {
        $have_overflow$07$i16 = 0;$num$08$i15 = 0;$p$19$i14 = $p$0$i12;
        while(1) {
         $67 = HEAP8[$p$19$i14>>0]|0;
         $68 = $67&255;
         switch ($67<<24>>24) {
         case 0: case 12: case 9: case 10: case 13: case 32:  {
          $have_overflow$0$lcssa$i23 = $have_overflow$07$i16;$num$0$lcssa$i22 = $num$08$i15;$p$1$lcssa$i21 = $p$19$i14;
          break L51;
          break;
         }
         default: {
         }
         }
         $69 = ($67<<24>>24)<(0);
         if ($69) {
          $have_overflow$0$lcssa$i23 = $have_overflow$07$i16;$num$0$lcssa$i22 = $num$08$i15;$p$1$lcssa$i21 = $p$19$i14;
          break L51;
         }
         $70 = $68 & 127;
         $71 = (532577 + ($70)|0);
         $72 = HEAP8[$71>>0]|0;
         $73 = $72 << 24 >> 24;
         $74 = (($70) + -48)|0;
         $75 = ($74>>>0)<(10);
         if (!($75)) {
          $have_overflow$0$lcssa$i23 = $have_overflow$07$i16;$num$0$lcssa$i22 = $num$08$i15;$p$1$lcssa$i21 = $p$19$i14;
          break L51;
         }
         $76 = ($num$08$i15|0)>(214748364);
         if ($76) {
          $have_overflow$2$ph$i19 = 1;$num$2$ph$i20 = $num$08$i15;
         } else {
          $77 = ($num$08$i15|0)==(214748364);
          $78 = ($72<<24>>24)>(7);
          $or$cond44$i18 = $77 & $78;
          if ($or$cond44$i18) {
           $have_overflow$2$ph$i19 = 1;$num$2$ph$i20 = 214748364;
          } else {
           $79 = ($num$08$i15*10)|0;
           $80 = (($73) + ($79))|0;
           $have_overflow$2$ph$i19 = $have_overflow$07$i16;$num$2$ph$i20 = $80;
          }
         }
         $81 = ((($p$19$i14)) + 1|0);
         $82 = ($81>>>0)<($59>>>0);
         if ($82) {
          $have_overflow$07$i16 = $have_overflow$2$ph$i19;$num$08$i15 = $num$2$ph$i20;$p$19$i14 = $81;
         } else {
          $have_overflow$0$lcssa$i23 = $have_overflow$2$ph$i19;$num$0$lcssa$i22 = $num$2$ph$i20;$p$1$lcssa$i21 = $81;
          break L51;
         }
        }
       } else {
        $have_overflow$0$lcssa$i23 = 0;$num$0$lcssa$i22 = 0;$p$1$lcssa$i21 = $p$0$i12;
       }
      } while(0);
      $83 = ($have_overflow$0$lcssa$i23<<24>>24)==(0);
      $num$3$$i24 = $83 ? $num$0$lcssa$i22 : 2147483647;
      $84 = ($sign$0$i11<<24>>24)==(0);
      $85 = (0 - ($num$3$$i24))|0;
      $num$5$i25 = $84 ? $num$3$$i24 : $85;
      $86 = ($p$1$lcssa$i21|0)==($45|0);
      $87 = $p$1$lcssa$i21;
      if ($86) {
       $$0$i9 = 0;
      } else {
       $88 = ($p$1$lcssa$i21>>>0)<($59>>>0);
       if ($88) {
        $89 = HEAP8[$p$1$lcssa$i21>>0]|0;
        $90 = ($89<<24>>24)==(35);
        if ($90) {
         $91 = ((($p$1$lcssa$i21)) + 1|0);
         $92 = ($91>>>0)>=($59>>>0);
         $base$off$i = (($num$5$i25) + -2)|0;
         $93 = ($base$off$i>>>0)>(34);
         $or$cond4$i = $92 | $93;
         if ($or$cond4$i) {
          $$0$i9 = 0;
          break;
         }
         $94 = HEAP8[$91>>0]|0;
         switch ($94<<24>>24) {
         case 43: case 45:  {
          $95 = ($94<<24>>24)==(45);
          $96 = $95&1;
          $97 = ((($p$1$lcssa$i21)) + 2|0);
          $98 = ($97|0)==($59|0);
          if ($98) {
           $$0$i9 = 0;
           break L46;
          } else {
           $p$0$i = $97;$sign$0$i = $96;
          }
          break;
         }
         default: {
          $p$0$i = $91;$sign$0$i = 0;
         }
         }
         $99 = (2147483647 / ($num$5$i25|0))&-1;
         $100 = ($p$0$i>>>0)<($59>>>0);
         L70: do {
          if ($100) {
           $101 = (2147483647 % ($num$5$i25|0))&-1;
           $sext$i = $101 << 24;
           $102 = $sext$i >> 24;
           $have_overflow$07$i = 0;$num$08$i = 0;$p$19$i = $p$0$i;
           while(1) {
            $103 = HEAP8[$p$19$i>>0]|0;
            $104 = $103&255;
            switch ($103<<24>>24) {
            case 0: case 12: case 9: case 10: case 13: case 32:  {
             $have_overflow$0$lcssa$i = $have_overflow$07$i;$num$0$lcssa$i = $num$08$i;$p$1$lcssa$i = $p$19$i;
             break L70;
             break;
            }
            default: {
            }
            }
            $105 = ($103<<24>>24)<(0);
            if ($105) {
             $have_overflow$0$lcssa$i = $have_overflow$07$i;$num$0$lcssa$i = $num$08$i;$p$1$lcssa$i = $p$19$i;
             break L70;
            }
            $106 = $104 & 127;
            $107 = (532577 + ($106)|0);
            $108 = HEAP8[$107>>0]|0;
            $109 = $108 << 24 >> 24;
            $110 = ($108<<24>>24)>(-1);
            $111 = ($109|0)<($num$5$i25|0);
            $or$cond$i = $110 & $111;
            if (!($or$cond$i)) {
             $have_overflow$0$lcssa$i = $have_overflow$07$i;$num$0$lcssa$i = $num$08$i;$p$1$lcssa$i = $p$19$i;
             break L70;
            }
            $112 = ($num$08$i|0)>($99|0);
            if ($112) {
             $have_overflow$2$ph$i = 1;$num$2$ph$i = $num$08$i;
            } else {
             $113 = ($num$08$i|0)==($99|0);
             $114 = ($109|0)>($102|0);
             $or$cond44$i = $113 & $114;
             if ($or$cond44$i) {
              $have_overflow$2$ph$i = 1;$num$2$ph$i = $99;
             } else {
              $115 = Math_imul($num$08$i, $num$5$i25)|0;
              $116 = (($109) + ($115))|0;
              $have_overflow$2$ph$i = $have_overflow$07$i;$num$2$ph$i = $116;
             }
            }
            $117 = ((($p$19$i)) + 1|0);
            $118 = ($117>>>0)<($59>>>0);
            if ($118) {
             $have_overflow$07$i = $have_overflow$2$ph$i;$num$08$i = $num$2$ph$i;$p$19$i = $117;
            } else {
             $have_overflow$0$lcssa$i = $have_overflow$2$ph$i;$num$0$lcssa$i = $num$2$ph$i;$p$1$lcssa$i = $117;
             break L70;
            }
           }
          } else {
           $have_overflow$0$lcssa$i = 0;$num$0$lcssa$i = 0;$p$1$lcssa$i = $p$0$i;
          }
         } while(0);
         $119 = ($have_overflow$0$lcssa$i<<24>>24)==(0);
         $num$3$$i = $119 ? $num$0$lcssa$i : 2147483647;
         $120 = ($sign$0$i<<24>>24)==(0);
         $121 = (0 - ($num$3$$i))|0;
         $num$5$i = $120 ? $num$3$$i : $121;
         $122 = ($p$1$lcssa$i|0)==($91|0);
         $123 = $p$1$lcssa$i;
         if ($122) {
          $$0$i9 = 0;
          break;
         } else {
          $124 = $123;$num$0$i = $num$5$i;
         }
        } else {
         $124 = $87;$num$0$i = $num$5$i25;
        }
       } else {
        $124 = $87;$num$0$i = $num$5$i25;
       }
       HEAP32[$str>>2] = $124;
       $$0$i9 = $num$0$i;
      }
     }
    } while(0);
    $125 = (((($vals) + ($i$066<<3)|0)) + 4|0);
    HEAP32[$125>>2] = $$0$i9;
    break;
   }
   case 4:  {
    $126 = ($48|0)==(4);
    if ($126) {
     $127 = (_strncmp($45,622520,4)|0);
     $128 = ($127|0)==(0);
     $130 = $128;
    } else {
     $130 = 0;
    }
    $129 = $130&1;
    $131 = (((($vals) + ($i$066<<3)|0)) + 4|0);
    HEAP8[$131>>0] = $129;
    break;
   }
   case 5:  {
    $132 = HEAP32[$4>>2]|0;
    $133 = ($132|0)==(0|0);
    if ($133) {
     $137 = (((($vals) + ($i$066<<3)|0)) + 4|0);
     HEAP32[$137>>2] = 0;
     break L39;
    } else {
     $134 = HEAP32[$5>>2]|0;
     $135 = (FUNCTION_TABLE_iiii[$132 & 8191]($45,$48,$134)|0);
     $136 = (((($vals) + ($i$066<<3)|0)) + 4|0);
     HEAP32[$136>>2] = $135;
     break L39;
    }
    break;
   }
   default: {
   }
   }
  } while(0);
  $138 = (($i$066) + 1)|0;
  $139 = ($138|0)<($n|0);
  if ($139) {
   $i$066 = $138;
  } else {
   $$0 = $138;
   label = 74;
   break;
  }
 }
 if ((label|0) == 8) {
  HEAP32[$3>>2] = 2;
  $i$066205 = $i$066$lcssa203;
 }
 else if ((label|0) == 10) {
  HEAP32[$3>>2] = 3;
  $i$066205 = $i$066207;
 }
 else if ((label|0) == 21) {
  HEAP32[$3>>2] = 2;
  $i$066205 = $i$066$lcssa198;
 }
 else if ((label|0) == 22) {
  HEAP32[$3>>2] = 1;
  $i$066205 = $i$066$lcssa199;
 }
 else if ((label|0) == 23) {
  HEAP32[$3>>2] = 3;
  $i$066205 = $i$066206;
 }
 else if ((label|0) == 74) {
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$str>>2] = 0;
 $$0 = $i$066205;
 STACKTOP = sp;return ($$0|0);
}
function _afm_compare_kern_pairs($a,$b) {
 $a = $a|0;
 $b = $b|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$a>>2]|0;
 $1 = $0 << 16;
 $2 = ((($a)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $1 | $3;
 $5 = HEAP32[$b>>2]|0;
 $6 = $5 << 16;
 $7 = ((($b)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $6 | $8;
 $10 = ($4>>>0)>($9>>>0);
 $11 = ($4>>>0)<($9>>>0);
 $$ = $11 << 31 >> 31;
 $$0 = $10 ? 1 : $$;
 return ($$0|0);
}
function _t1_cmap_standard_init($cmap,$pointer) {
 $cmap = $cmap|0;
 $pointer = $pointer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$cmap>>2]|0;
 $1 = ((($0)) + 484|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($0)) + 416|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($cmap)) + 24|0);
 HEAP32[$5>>2] = $4;
 $6 = ((($0)) + 420|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($cmap)) + 28|0);
 HEAP32[$8>>2] = $7;
 $9 = ((($2)) + 20|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($cmap)) + 20|0);
 HEAP32[$11>>2] = $10;
 $12 = ((($2)) + 24|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($cmap)) + 16|0);
 HEAP32[$14>>2] = $13;
 return 0;
}
function _t1_cmap_std_done($cmap) {
 $cmap = $cmap|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;HEAP32[$0+12>>2]=0|0;
 return;
}
function _t1_cmap_std_char_index($cmap,$char_code) {
 $cmap = $cmap|0;
 $char_code = $char_code|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $n$02 = 0, $result$3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($char_code>>>0)<(256);
 if (!($0)) {
  $result$3 = 0;
  return ($result$3|0);
 }
 $1 = ((($cmap)) + 16|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($2) + ($char_code<<1)|0);
 $4 = HEAP16[$3>>1]|0;
 $5 = $4&65535;
 $6 = ((($cmap)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (FUNCTION_TABLE_ii[$7 & 8191]($5)|0);
 $9 = ((($cmap)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10|0)==(0);
 if ($11) {
  $result$3 = 0;
  return ($result$3|0);
 }
 $12 = ((($cmap)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $n$02 = 0;
 while(1) {
  $14 = (($13) + ($n$02<<2)|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ($15|0)==(0|0);
  if (!($16)) {
   $17 = HEAP8[$15>>0]|0;
   $18 = HEAP8[$8>>0]|0;
   $19 = ($17<<24>>24)==($18<<24>>24);
   if ($19) {
    $20 = (_strcmp($15,$8)|0);
    $21 = ($20|0)==(0);
    if ($21) {
     $result$3 = $n$02;
     label = 8;
     break;
    }
   }
  }
  $22 = (($n$02) + 1)|0;
  $23 = ($22>>>0)<($10>>>0);
  if ($23) {
   $n$02 = $22;
  } else {
   $result$3 = 0;
   label = 8;
   break;
  }
 }
 if ((label|0) == 8) {
  return ($result$3|0);
 }
 return (0)|0;
}
function _t1_cmap_std_char_next($cmap,$pchar_code) {
 $cmap = $cmap|0;
 $pchar_code = $pchar_code|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $char_code$0 = 0, $char_code$02 = 0, $char_code$03 = 0, $char_code$1 = 0, $n$02$i = 0, $n$02$i$lcssa = 0, $result$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$pchar_code>>2]|0;
 $char_code$02 = (($0) + 1)|0;
 $1 = ($char_code$02>>>0)<(256);
 if (!($1)) {
  $char_code$1 = 0;$result$1 = 0;
  HEAP32[$pchar_code>>2] = $char_code$1;
  return ($result$1|0);
 }
 $2 = ((($cmap)) + 16|0);
 $3 = ((($cmap)) + 20|0);
 $4 = ((($cmap)) + 24|0);
 $5 = ((($cmap)) + 28|0);
 $char_code$03 = $char_code$02;
 L4: while(1) {
  $6 = HEAP32[$2>>2]|0;
  $7 = (($6) + ($char_code$03<<1)|0);
  $8 = HEAP16[$7>>1]|0;
  $9 = $8&65535;
  $10 = HEAP32[$3>>2]|0;
  $11 = (FUNCTION_TABLE_ii[$10 & 8191]($9)|0);
  $12 = HEAP32[$4>>2]|0;
  $13 = ($12|0)==(0);
  L6: do {
   if (!($13)) {
    $14 = HEAP32[$5>>2]|0;
    $n$02$i = 0;
    while(1) {
     $15 = (($14) + ($n$02$i<<2)|0);
     $16 = HEAP32[$15>>2]|0;
     $17 = ($16|0)==(0|0);
     if (!($17)) {
      $18 = HEAP8[$16>>0]|0;
      $19 = HEAP8[$11>>0]|0;
      $20 = ($18<<24>>24)==($19<<24>>24);
      if ($20) {
       $21 = (_strcmp($16,$11)|0);
       $22 = ($21|0)==(0);
       if ($22) {
        $n$02$i$lcssa = $n$02$i;
        break;
       }
      }
     }
     $23 = (($n$02$i) + 1)|0;
     $24 = ($23>>>0)<($12>>>0);
     if ($24) {
      $n$02$i = $23;
     } else {
      break L6;
     }
    }
    $25 = ($n$02$i$lcssa|0)==(0);
    if (!($25)) {
     $char_code$1 = $char_code$03;$result$1 = $n$02$i$lcssa;
     label = 11;
     break L4;
    }
   }
  } while(0);
  $char_code$0 = (($char_code$03) + 1)|0;
  $26 = ($char_code$0>>>0)<(256);
  if ($26) {
   $char_code$03 = $char_code$0;
  } else {
   $char_code$1 = 0;$result$1 = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 11) {
  HEAP32[$pchar_code>>2] = $char_code$1;
  return ($result$1|0);
 }
 return (0)|0;
}
function _t1_cmap_expert_init($cmap,$pointer) {
 $cmap = $cmap|0;
 $pointer = $pointer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$cmap>>2]|0;
 $1 = ((($0)) + 484|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($0)) + 416|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($cmap)) + 24|0);
 HEAP32[$5>>2] = $4;
 $6 = ((($0)) + 420|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($cmap)) + 28|0);
 HEAP32[$8>>2] = $7;
 $9 = ((($2)) + 20|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($cmap)) + 20|0);
 HEAP32[$11>>2] = $10;
 $12 = ((($2)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($cmap)) + 16|0);
 HEAP32[$14>>2] = $13;
 return 0;
}
function _t1_cmap_custom_init($cmap,$pointer) {
 $cmap = $cmap|0;
 $pointer = $pointer|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$cmap>>2]|0;
 $1 = ((($0)) + 376|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($cmap)) + 16|0);
 HEAP32[$3>>2] = $2;
 $4 = ((($0)) + 380|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($5) - ($2))|0;
 $7 = ((($cmap)) + 20|0);
 HEAP32[$7>>2] = $6;
 $8 = ((($0)) + 384|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($cmap)) + 24|0);
 HEAP32[$10>>2] = $9;
 return 0;
}
function _t1_cmap_custom_done($cmap) {
 $cmap = $cmap|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 24|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($cmap)) + 16|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($cmap)) + 20|0);
 HEAP32[$2>>2] = 0;
 return;
}
function _t1_cmap_custom_char_index($cmap,$char_code) {
 $cmap = $cmap|0;
 $char_code = $char_code|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1>>>0)>($char_code>>>0);
 if ($2) {
  $result$0 = 0;
  return ($result$0|0);
 }
 $3 = ((($cmap)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + ($1))|0;
 $6 = ($5>>>0)>($char_code>>>0);
 if (!($6)) {
  $result$0 = 0;
  return ($result$0|0);
 }
 $7 = ((($cmap)) + 24|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (($8) + ($char_code<<1)|0);
 $10 = HEAP16[$9>>1]|0;
 $11 = $10&65535;
 $result$0 = $11;
 return ($result$0|0);
}
function _t1_cmap_custom_char_next($cmap,$pchar_code) {
 $cmap = $cmap|0;
 $pchar_code = $pchar_code|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $char_code$01 = 0, $char_code$1 = 0, $result$1 = 0;
 var $result$1$shrunk = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$pchar_code>>2]|0;
 $1 = (($0) + 1)|0;
 $2 = ((($cmap)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1>>>0)<($3>>>0);
 $$ = $4 ? $3 : $1;
 $5 = ((($cmap)) + 20|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + ($3))|0;
 $8 = ($$>>>0)<($7>>>0);
 L1: do {
  if ($8) {
   $9 = ((($cmap)) + 24|0);
   $10 = HEAP32[$9>>2]|0;
   $char_code$01 = $$;
   while(1) {
    $11 = (($10) + ($char_code$01<<1)|0);
    $12 = HEAP16[$11>>1]|0;
    $13 = ($12<<16>>16)==(0);
    if (!($13)) {
     $char_code$1 = $char_code$01;$result$1$shrunk = $12;
     break L1;
    }
    $14 = (($char_code$01) + 1)|0;
    $15 = ($14>>>0)<($7>>>0);
    if ($15) {
     $char_code$01 = $14;
    } else {
     $char_code$1 = 0;$result$1$shrunk = 0;
     break;
    }
   }
  } else {
   $char_code$1 = 0;$result$1$shrunk = 0;
  }
 } while(0);
 $result$1 = $result$1$shrunk&65535;
 HEAP32[$pchar_code>>2] = $char_code$1;
 return ($result$1|0);
}
function _t1_cmap_unicode_init($unicodes,$pointer) {
 $unicodes = $unicodes|0;
 $pointer = $pointer|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$unicodes>>2]|0;
 $1 = ((($0)) + 100|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($0)) + 484|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($0)) + 416|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (FUNCTION_TABLE_iiiiiii[$6 & 8191]($2,$unicodes,$8,4367,0,$0)|0);
 return ($9|0);
}
function _psaux_get_glyph_name($face,$idx) {
 $face = $face|0;
 $idx = $idx|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 420|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + ($idx<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 return ($3|0);
}
function _t1_cmap_unicode_done($unicodes) {
 $unicodes = $unicodes|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$unicodes>>2]|0;
 $1 = ((($0)) + 100|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($unicodes)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 _ft_mem_free($2,$4);
 HEAP32[$3>>2] = 0;
 $5 = ((($unicodes)) + 16|0);
 HEAP32[$5>>2] = 0;
 return;
}
function _t1_cmap_unicode_char_index($unicodes,$char_code) {
 $unicodes = $unicodes|0;
 $char_code = $char_code|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$unicodes>>2]|0;
 $1 = ((($0)) + 484|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (FUNCTION_TABLE_iii[$4 & 8191]($unicodes,$char_code)|0);
 return ($5|0);
}
function _t1_cmap_unicode_char_next($unicodes,$pchar_code) {
 $unicodes = $unicodes|0;
 $pchar_code = $pchar_code|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$unicodes>>2]|0;
 $1 = ((($0)) + 484|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (FUNCTION_TABLE_iii[$4 & 8191]($unicodes,$pchar_code)|0);
 return ($5|0);
}
function _t1_decrypt($buffer,$length,$seed) {
 $buffer = $buffer|0;
 $length = $length|0;
 $seed = $seed|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $r$01$i = 0, $s$02$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($length|0)>(0);
 if (!($0)) {
  return;
 }
 $1 = $seed&65535;
 $r$01$i = 0;$s$02$i = $1;
 while(1) {
  $2 = (($buffer) + ($r$01$i)|0);
  $3 = HEAP8[$2>>0]|0;
  $4 = $3&255;
  $5 = $s$02$i >>> 8;
  $6 = $4 ^ $5;
  $7 = (($4) + ($s$02$i))|0;
  $8 = ($7*52845)|0;
  $9 = (($8) + 22719)|0;
  $10 = $9 & 65535;
  $11 = $6&255;
  HEAP8[$2>>0] = $11;
  $12 = (($r$01$i) + 1)|0;
  $exitcond = ($12|0)==($length|0);
  if ($exitcond) {
   break;
  } else {
   $r$01$i = $12;$s$02$i = $10;
  }
 }
 return;
}
function _pshinter_get_globals_funcs($module) {
 $module = $module|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($module)) + 100|0);
 return ($0|0);
}
function _pshinter_get_t1_funcs($module) {
 $module = $module|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($module)) + 112|0);
 return ($0|0);
}
function _pshinter_get_t2_funcs($module) {
 $module = $module|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($module)) + 140|0);
 return ($0|0);
}
function _ps_hinter_init($module) {
 $module = $module|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($module)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($module)) + 12|0);
 dest=$2; stop=dest+88|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 HEAP32[$2>>2] = $1;
 $3 = ((($module)) + 100|0);
 HEAP32[$3>>2] = 4368;
 $4 = ((($module)) + 104|0);
 HEAP32[$4>>2] = 4369;
 $5 = ((($module)) + 108|0);
 HEAP32[$5>>2] = 4370;
 $6 = ((($module)) + 112|0);
 dest=$6; stop=dest+20|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0));
 $7 = ((($module)) + 116|0);
 HEAP32[$7>>2] = 4371;
 $8 = ((($module)) + 120|0);
 HEAP32[$8>>2] = 4372;
 $9 = ((($module)) + 124|0);
 HEAP32[$9>>2] = 4373;
 $10 = ((($module)) + 128|0);
 HEAP32[$10>>2] = 4374;
 $11 = ((($module)) + 132|0);
 HEAP32[$11>>2] = 4375;
 $12 = ((($module)) + 136|0);
 HEAP32[$12>>2] = 4376;
 HEAP32[$6>>2] = $2;
 $13 = ((($module)) + 140|0);
 ;HEAP32[$13>>2]=0|0;HEAP32[$13+4>>2]=0|0;HEAP32[$13+8>>2]=0|0;HEAP32[$13+12>>2]=0|0;HEAP32[$13+16>>2]=0|0;
 $14 = ((($module)) + 144|0);
 HEAP32[$14>>2] = 4377;
 $15 = ((($module)) + 148|0);
 HEAP32[$15>>2] = 4372;
 $16 = ((($module)) + 152|0);
 HEAP32[$16>>2] = 4378;
 $17 = ((($module)) + 156|0);
 HEAP32[$17>>2] = 4379;
 $18 = ((($module)) + 160|0);
 HEAP32[$18>>2] = 4380;
 $19 = ((($module)) + 164|0);
 HEAP32[$19>>2] = 4376;
 HEAP32[$13>>2] = $2;
 return 0;
}
function _psh_globals_new($memory,$priv,$aglobals) {
 $memory = $memory|0;
 $priv = $priv|0;
 $aglobals = $aglobals|0;
 var $$ = 0, $$$0$i = 0, $$$0$i11 = 0, $$$0$i18 = 0, $$$0$i4 = 0, $$0$lcssa$i = 0, $$0$lcssa$i13 = 0, $$0$lcssa$i20 = 0, $$0$lcssa$i6 = 0, $$02$i = 0, $$02$i1 = 0, $$02$i15 = 0, $$02$i8 = 0, $$mask = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $count$01$i = 0, $count$01$i16 = 0;
 var $count$01$i2 = 0, $count$01$i9 = 0, $count$027 = 0, $count$124 = 0, $error = 0, $read$026 = 0, $read$122 = 0, $sext$i = 0, $sext$i10 = 0, $sext$i17 = 0, $sext$i3 = 0, $write$025 = 0, $write2$023 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = (_ft_mem_alloc($memory,2496,$error)|0);
 $1 = HEAP32[$error>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  $142 = $1;
  HEAP32[$aglobals>>2] = $0;
  STACKTOP = sp;return ($142|0);
 }
 HEAP32[$0>>2] = $memory;
 $3 = ((($0)) + 4|0);
 $4 = ((($0)) + 208|0);
 $5 = ((($0)) + 212|0);
 $6 = ((($priv)) + 120|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 & 65535;
 HEAP32[$5>>2] = $8;
 $9 = ((($priv)) + 124|0);
 $10 = HEAP16[$9>>1]|0;
 $11 = $10&255;
 $$mask = $10 & 255;
 $12 = $$mask&65535;
 $13 = ($11<<24>>24)==(0);
 $14 = $7 >>> 16;
 $15 = ($10&65535) >>> 8;
 if (!($13)) {
  $16 = ((($priv)) + 128|0);
  $17 = ((($0)) + 224|0);
  $count$027 = $12;$read$026 = $16;$write$025 = $17;
  while(1) {
   $18 = HEAP16[$read$026>>1]|0;
   $19 = $18 << 16 >> 16;
   HEAP32[$write$025>>2] = $19;
   $20 = ((($write$025)) + 12|0);
   $21 = ((($read$026)) + 2|0);
   $22 = (($count$027) + -1)|0;
   $23 = ($22|0)==(0);
   if ($23) {
    break;
   } else {
    $count$027 = $22;$read$026 = $21;$write$025 = $20;
   }
  }
 }
 $24 = (($12) + 1)|0;
 HEAP32[$4>>2] = $24;
 $25 = ((($0)) + 8|0);
 HEAP32[$25>>2] = $14;
 $26 = $15&65535;
 $27 = ($15<<16>>16)==(0);
 if (!($27)) {
  $28 = ((($priv)) + 154|0);
  $29 = ((($0)) + 20|0);
  $count$124 = $26;$read$122 = $28;$write2$023 = $29;
  while(1) {
   $30 = HEAP16[$read$122>>1]|0;
   $31 = $30 << 16 >> 16;
   HEAP32[$write2$023>>2] = $31;
   $32 = ((($write2$023)) + 12|0);
   $33 = ((($read$122)) + 2|0);
   $34 = (($count$124) + -1)|0;
   $35 = ($34|0)==(0);
   if ($35) {
    break;
   } else {
    $count$124 = $34;$read$122 = $33;$write2$023 = $32;
   }
  }
 }
 $36 = (($26) + 1)|0;
 HEAP32[$3>>2] = $36;
 $37 = ((($0)) + 412|0);
 $38 = ((($priv)) + 8|0);
 $39 = HEAP16[$38>>1]|0;
 $40 = $39&65535;
 $41 = $40 & 255;
 $42 = ((($priv)) + 12|0);
 $43 = ($39&65535) >>> 8;
 $44 = $43&65535;
 $45 = ((($priv)) + 40|0);
 $46 = ((($priv)) + 116|0);
 $47 = HEAP32[$46>>2]|0;
 _psh_blues_set_zones($37,$41,$42,$44,$45,$47,0);
 $48 = ((($priv)) + 10|0);
 $49 = HEAP16[$48>>1]|0;
 $50 = $49&65535;
 $51 = $50 & 255;
 $52 = ((($priv)) + 60|0);
 $53 = ($49&65535) >>> 8;
 $54 = $53&65535;
 $55 = ((($priv)) + 88|0);
 $56 = HEAP32[$46>>2]|0;
 _psh_blues_set_zones($37,$51,$52,$54,$55,$56,1);
 $57 = HEAP32[$38>>2]|0;
 $58 = $57&65535;
 $59 = $57&255;
 $60 = $57 & 255;
 $61 = ($59<<24>>24)==(0);
 $62 = ($58&65535) >>> 8;
 $63 = $57 >>> 16;
 $64 = $63&255;
 $65 = $57 >>> 24;
 if ($61) {
  $$0$lcssa$i = 1;
 } else {
  $$02$i = 1;$count$01$i = 0;
  while(1) {
   $66 = $count$01$i | 1;
   $67 = (((($priv)) + 12|0) + ($66<<1)|0);
   $68 = HEAP16[$67>>1]|0;
   $69 = $68&65535;
   $70 = (((($priv)) + 12|0) + ($count$01$i<<1)|0);
   $71 = HEAP16[$70>>1]|0;
   $72 = $71&65535;
   $73 = (($69) - ($72))|0;
   $sext$i = $73 << 16;
   $74 = $sext$i >> 16;
   $75 = $$02$i << 16 >> 16;
   $76 = ($74|0)>($75|0);
   $77 = $73&65535;
   $$$0$i = $76 ? $77 : $$02$i;
   $78 = (($count$01$i) + 2)|0;
   $79 = ($78>>>0)<($60>>>0);
   if ($79) {
    $$02$i = $$$0$i;$count$01$i = $78;
   } else {
    $$0$lcssa$i = $$$0$i;
    break;
   }
  }
 }
 $80 = $62&65535;
 $81 = ($62<<16>>16)==(0);
 if ($81) {
  $$0$lcssa$i20 = $$0$lcssa$i;
 } else {
  $$02$i15 = $$0$lcssa$i;$count$01$i16 = 0;
  while(1) {
   $82 = $count$01$i16 | 1;
   $83 = (((($priv)) + 40|0) + ($82<<1)|0);
   $84 = HEAP16[$83>>1]|0;
   $85 = $84&65535;
   $86 = (((($priv)) + 40|0) + ($count$01$i16<<1)|0);
   $87 = HEAP16[$86>>1]|0;
   $88 = $87&65535;
   $89 = (($85) - ($88))|0;
   $sext$i17 = $89 << 16;
   $90 = $sext$i17 >> 16;
   $91 = $$02$i15 << 16 >> 16;
   $92 = ($90|0)>($91|0);
   $93 = $89&65535;
   $$$0$i18 = $92 ? $93 : $$02$i15;
   $94 = (($count$01$i16) + 2)|0;
   $95 = ($94>>>0)<($80>>>0);
   if ($95) {
    $$02$i15 = $$$0$i18;$count$01$i16 = $94;
   } else {
    $$0$lcssa$i20 = $$$0$i18;
    break;
   }
  }
 }
 $96 = $63 & 255;
 $97 = ($64<<24>>24)==(0);
 if ($97) {
  $$0$lcssa$i13 = $$0$lcssa$i20;
 } else {
  $$02$i8 = $$0$lcssa$i20;$count$01$i9 = 0;
  while(1) {
   $98 = $count$01$i9 | 1;
   $99 = (((($priv)) + 60|0) + ($98<<1)|0);
   $100 = HEAP16[$99>>1]|0;
   $101 = $100&65535;
   $102 = (((($priv)) + 60|0) + ($count$01$i9<<1)|0);
   $103 = HEAP16[$102>>1]|0;
   $104 = $103&65535;
   $105 = (($101) - ($104))|0;
   $sext$i10 = $105 << 16;
   $106 = $sext$i10 >> 16;
   $107 = $$02$i8 << 16 >> 16;
   $108 = ($106|0)>($107|0);
   $109 = $105&65535;
   $$$0$i11 = $108 ? $109 : $$02$i8;
   $110 = (($count$01$i9) + 2)|0;
   $111 = ($110>>>0)<($96>>>0);
   if ($111) {
    $$02$i8 = $$$0$i11;$count$01$i9 = $110;
   } else {
    $$0$lcssa$i13 = $$$0$i11;
    break;
   }
  }
 }
 $112 = ($65|0)==(0);
 if ($112) {
  $$0$lcssa$i6 = $$0$lcssa$i13;
 } else {
  $$02$i1 = $$0$lcssa$i13;$count$01$i2 = 0;
  while(1) {
   $113 = $count$01$i2 | 1;
   $114 = (((($priv)) + 88|0) + ($113<<1)|0);
   $115 = HEAP16[$114>>1]|0;
   $116 = $115&65535;
   $117 = (((($priv)) + 88|0) + ($count$01$i2<<1)|0);
   $118 = HEAP16[$117>>1]|0;
   $119 = $118&65535;
   $120 = (($116) - ($119))|0;
   $sext$i3 = $120 << 16;
   $121 = $sext$i3 >> 16;
   $122 = $$02$i1 << 16 >> 16;
   $123 = ($121|0)>($122|0);
   $124 = $120&65535;
   $$$0$i4 = $123 ? $124 : $$02$i1;
   $125 = (($count$01$i2) + 2)|0;
   $126 = ($125>>>0)<($65>>>0);
   if ($126) {
    $$02$i1 = $$$0$i4;$count$01$i2 = $125;
   } else {
    $$0$lcssa$i6 = $$$0$i4;
    break;
   }
  }
 }
 $127 = $$0$lcssa$i6 << 16 >> 16;
 $128 = (_FT_DivFix(1000,$127)|0);
 $129 = ((($priv)) + 108|0);
 $130 = HEAP32[$129>>2]|0;
 $131 = ($130|0)<($128|0);
 $$ = $131 ? $130 : $128;
 $132 = ((($0)) + 2476|0);
 HEAP32[$132>>2] = $$;
 $133 = ((($priv)) + 112|0);
 $134 = HEAP32[$133>>2]|0;
 $135 = ((($0)) + 2480|0);
 HEAP32[$135>>2] = $134;
 $136 = HEAP32[$46>>2]|0;
 $137 = ((($0)) + 2488|0);
 HEAP32[$137>>2] = $136;
 $138 = ((($0)) + 200|0);
 HEAP32[$138>>2] = 0;
 $139 = ((($0)) + 204|0);
 HEAP32[$139>>2] = 0;
 $140 = ((($0)) + 404|0);
 HEAP32[$140>>2] = 0;
 $141 = ((($0)) + 408|0);
 HEAP32[$141>>2] = 0;
 $$pre = HEAP32[$error>>2]|0;
 $142 = $$pre;
 HEAP32[$aglobals>>2] = $0;
 STACKTOP = sp;return ($142|0);
}
function _psh_blues_set_zones($target,$count,$blues,$count_others,$other_blues,$fuzz,$family) {
 $target = $target|0;
 $count = $count|0;
 $blues = $blues|0;
 $count_others = $count_others|0;
 $other_blues = $other_blues|0;
 $fuzz = $fuzz|0;
 $family = $family|0;
 var $$0115$i = 0, $$0115$i2 = 0, $$016$i = 0, $$016$i1 = 0, $$054 = 0, $$150 = 0, $$3 = 0, $$3$1 = 0, $$338 = 0, $$338$1 = 0, $$341 = 0, $$341$1 = 0, $$phi$trans$insert73 = 0, $$phi$trans$insert78 = 0, $$pre = 0, $$pre74 = 0, $$pre79 = 0, $$sink = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $bot_table$0 = 0, $bot_table$pn$i = 0, $count$0$i = 0, $count$12$i = 0, $count$12$i$lcssa = 0, $count$12$i16 = 0, $count$12$i16$lcssa = 0, $count$28$i = 0, $count$28$i18 = 0, $count_bot$0$lcssa$i = 0, $count_bot$0$lcssa$i24 = 0, $count_bot$012$i = 0, $count_bot$012$i4 = 0, $count_bot$1$i = 0, $count_bot$1$i23 = 0, $count_top$0$lcssa$i = 0, $count_top$010$i = 0;
 var $count_top$1$i = 0, $delta$0$i = 0, $first$014$i = 0, $reference$0$i = 0, $scevgep$184 = 0, $scevgep83 = 0, $storemerge = 0, $top$0 = 0, $top$0$1 = 0, $top$0$1$lcssa = 0, $top$0$i = 0, $top$0$lcssa = 0, $top$0$lcssa$1 = 0, $top$0$lcssa121 = 0, $top$039 = 0, $top$039$1 = 0, $top$042 = 0, $top$042$1 = 0, $top_table$0 = 0, $uglygep = 0;
 var $uglygep$1 = 0, $zone$0$lcssa23$i = 0, $zone$0$lcssa23$i20 = 0, $zone$03$i = 0, $zone$03$i$lcssa = 0, $zone$03$i$lcssa126 = 0, $zone$03$i15 = 0, $zone$03$i15$lcssa = 0, $zone$03$i15$lcssa122 = 0, $zone$053 = 0, $zone1$049 = 0, $zone4$048 = 0, $zone4$048$1 = 0, $zone4$1$lcssa = 0, $zone4$1$lcssa$1 = 0, $zone4$140 = 0, $zone4$140$1 = 0, $zones$0$i = 0, $zones$0$i13 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ($family|0)==(0);
 $1 = ((($target)) + 1032|0);
 $2 = ((($target)) + 1548|0);
 $3 = ((($target)) + 516|0);
 $bot_table$0 = $0 ? $3 : $2;
 $top_table$0 = $0 ? $target : $1;
 HEAP32[$top_table$0>>2] = 0;
 HEAP32[$bot_table$0>>2] = 0;
 $4 = ($count>>>0)>(1);
 if ($4) {
  $$0115$i = $blues;$$016$i = $count;$count_bot$012$i = 0;$count_top$010$i = 0;$first$014$i = 0;
  while(1) {
   if ($first$014$i) {
    $11 = HEAP16[$$0115$i>>1]|0;
    $12 = $11 << 16 >> 16;
    $13 = ((($$0115$i)) + 2|0);
    $14 = HEAP16[$13>>1]|0;
    $15 = $14 << 16 >> 16;
    $16 = (($15) - ($12))|0;
    $bot_table$pn$i = $top_table$0;$count$0$i = $count_top$010$i;$delta$0$i = $16;$reference$0$i = $12;$top$0$i = 1;
   } else {
    $5 = ((($$0115$i)) + 2|0);
    $6 = HEAP16[$5>>1]|0;
    $7 = $6 << 16 >> 16;
    $8 = HEAP16[$$0115$i>>1]|0;
    $9 = $8 << 16 >> 16;
    $10 = (($9) - ($7))|0;
    $bot_table$pn$i = $bot_table$0;$count$0$i = $count_bot$012$i;$delta$0$i = $10;$reference$0$i = $7;$top$0$i = 0;
   }
   $zones$0$i = ((($bot_table$pn$i)) + 4|0);
   $17 = ($count$0$i|0)==(0);
   L8: do {
    if ($17) {
     $zone$0$lcssa23$i = $zones$0$i;
     label = 16;
    } else {
     $count$12$i = $count$0$i;$zone$03$i = $zones$0$i;
     while(1) {
      $18 = HEAP32[$zone$03$i>>2]|0;
      $19 = ($reference$0$i|0)<($18|0);
      if ($19) {
       $count$12$i$lcssa = $count$12$i;$zone$03$i$lcssa = $zone$03$i;
       label = 7;
       break;
      }
      $20 = ($reference$0$i|0)==($18|0);
      if ($20) {
       $zone$03$i$lcssa126 = $zone$03$i;
       break;
      }
      $26 = (($count$12$i) + -1)|0;
      $27 = ((($zone$03$i)) + 32|0);
      $28 = ($26|0)==(0);
      if ($28) {
       $zone$0$lcssa23$i = $27;
       label = 16;
       break L8;
      } else {
       $count$12$i = $26;$zone$03$i = $27;
      }
     }
     if ((label|0) == 7) {
      label = 0;
      $count$28$i = $count$12$i$lcssa;
      while(1) {
       $29 = (($zone$03$i$lcssa) + ($count$28$i<<5)|0);
       $30 = (($count$28$i) + -1)|0;
       $31 = (($zone$03$i$lcssa) + ($30<<5)|0);
       ;HEAP32[$29>>2]=HEAP32[$31>>2]|0;HEAP32[$29+4>>2]=HEAP32[$31+4>>2]|0;HEAP32[$29+8>>2]=HEAP32[$31+8>>2]|0;HEAP32[$29+12>>2]=HEAP32[$31+12>>2]|0;HEAP32[$29+16>>2]=HEAP32[$31+16>>2]|0;HEAP32[$29+20>>2]=HEAP32[$31+20>>2]|0;HEAP32[$29+24>>2]=HEAP32[$31+24>>2]|0;HEAP32[$29+28>>2]=HEAP32[$31+28>>2]|0;
       $32 = ($30|0)==(0);
       if ($32) {
        $zone$0$lcssa23$i = $zone$03$i$lcssa;
        label = 16;
        break L8;
       } else {
        $count$28$i = $30;
       }
      }
     }
     $21 = ((($zone$03$i$lcssa126)) + 4|0);
     $22 = HEAP32[$21>>2]|0;
     $23 = ($delta$0$i|0)<(0);
     if ($23) {
      $24 = ($delta$0$i|0)<($22|0);
      if (!($24)) {
       $count_bot$1$i = $count_bot$012$i;$count_top$1$i = $count_top$010$i;
       break;
      }
      HEAP32[$21>>2] = $delta$0$i;
      $count_bot$1$i = $count_bot$012$i;$count_top$1$i = $count_top$010$i;
      break;
     } else {
      $25 = ($delta$0$i|0)>($22|0);
      if (!($25)) {
       $count_bot$1$i = $count_bot$012$i;$count_top$1$i = $count_top$010$i;
       break;
      }
      HEAP32[$21>>2] = $delta$0$i;
      $count_bot$1$i = $count_bot$012$i;$count_top$1$i = $count_top$010$i;
      break;
     }
    }
   } while(0);
   do {
    if ((label|0) == 16) {
     label = 0;
     HEAP32[$zone$0$lcssa23$i>>2] = $reference$0$i;
     $33 = ((($zone$0$lcssa23$i)) + 4|0);
     HEAP32[$33>>2] = $delta$0$i;
     $34 = ($top$0$i<<24>>24)==(0);
     if ($34) {
      $36 = (($count_bot$012$i) + 1)|0;
      $count_bot$1$i = $36;$count_top$1$i = $count_top$010$i;
      break;
     } else {
      $35 = (($count_top$010$i) + 1)|0;
      $count_bot$1$i = $count_bot$012$i;$count_top$1$i = $35;
      break;
     }
    }
   } while(0);
   $37 = ((($$0115$i)) + 4|0);
   $38 = (($$016$i) + -2)|0;
   $39 = ($38>>>0)>(1);
   if ($39) {
    $$0115$i = $37;$$016$i = $38;$count_bot$012$i = $count_bot$1$i;$count_top$010$i = $count_top$1$i;$first$014$i = 1;
   } else {
    $count_bot$0$lcssa$i = $count_bot$1$i;$count_top$0$lcssa$i = $count_top$1$i;
    break;
   }
  }
 } else {
  $count_bot$0$lcssa$i = 0;$count_top$0$lcssa$i = 0;
 }
 HEAP32[$top_table$0>>2] = $count_top$0$lcssa$i;
 HEAP32[$bot_table$0>>2] = $count_bot$0$lcssa$i;
 $40 = ($count_others>>>0)>(1);
 if ($40) {
  $zones$0$i13 = ((($bot_table$0)) + 4|0);
  $$0115$i2 = $other_blues;$$016$i1 = $count_others;$count_bot$012$i4 = $count_bot$0$lcssa$i;
  while(1) {
   $41 = ((($$0115$i2)) + 2|0);
   $42 = HEAP16[$41>>1]|0;
   $43 = $42 << 16 >> 16;
   $44 = HEAP16[$$0115$i2>>1]|0;
   $45 = $44 << 16 >> 16;
   $46 = (($45) - ($43))|0;
   $47 = ($count_bot$012$i4|0)==(0);
   L34: do {
    if ($47) {
     $zone$0$lcssa23$i20 = $zones$0$i13;
     label = 33;
    } else {
     $count$12$i16 = $count_bot$012$i4;$zone$03$i15 = $zones$0$i13;
     while(1) {
      $48 = HEAP32[$zone$03$i15>>2]|0;
      $49 = ($43|0)<($48|0);
      if ($49) {
       $count$12$i16$lcssa = $count$12$i16;$zone$03$i15$lcssa = $zone$03$i15;
       label = 24;
       break;
      }
      $50 = ($43|0)==($48|0);
      if ($50) {
       $zone$03$i15$lcssa122 = $zone$03$i15;
       break;
      }
      $56 = (($count$12$i16) + -1)|0;
      $57 = ((($zone$03$i15)) + 32|0);
      $58 = ($56|0)==(0);
      if ($58) {
       $zone$0$lcssa23$i20 = $57;
       label = 33;
       break L34;
      } else {
       $count$12$i16 = $56;$zone$03$i15 = $57;
      }
     }
     if ((label|0) == 24) {
      label = 0;
      $count$28$i18 = $count$12$i16$lcssa;
      while(1) {
       $59 = (($zone$03$i15$lcssa) + ($count$28$i18<<5)|0);
       $60 = (($count$28$i18) + -1)|0;
       $61 = (($zone$03$i15$lcssa) + ($60<<5)|0);
       ;HEAP32[$59>>2]=HEAP32[$61>>2]|0;HEAP32[$59+4>>2]=HEAP32[$61+4>>2]|0;HEAP32[$59+8>>2]=HEAP32[$61+8>>2]|0;HEAP32[$59+12>>2]=HEAP32[$61+12>>2]|0;HEAP32[$59+16>>2]=HEAP32[$61+16>>2]|0;HEAP32[$59+20>>2]=HEAP32[$61+20>>2]|0;HEAP32[$59+24>>2]=HEAP32[$61+24>>2]|0;HEAP32[$59+28>>2]=HEAP32[$61+28>>2]|0;
       $62 = ($60|0)==(0);
       if ($62) {
        $zone$0$lcssa23$i20 = $zone$03$i15$lcssa;
        label = 33;
        break L34;
       } else {
        $count$28$i18 = $60;
       }
      }
     }
     $51 = ((($zone$03$i15$lcssa122)) + 4|0);
     $52 = HEAP32[$51>>2]|0;
     $53 = ($46|0)<(0);
     if ($53) {
      $54 = ($46|0)<($52|0);
      if (!($54)) {
       $count_bot$1$i23 = $count_bot$012$i4;
       break;
      }
      HEAP32[$51>>2] = $46;
      $count_bot$1$i23 = $count_bot$012$i4;
      break;
     } else {
      $55 = ($46|0)>($52|0);
      if (!($55)) {
       $count_bot$1$i23 = $count_bot$012$i4;
       break;
      }
      HEAP32[$51>>2] = $46;
      $count_bot$1$i23 = $count_bot$012$i4;
      break;
     }
    }
   } while(0);
   if ((label|0) == 33) {
    label = 0;
    HEAP32[$zone$0$lcssa23$i20>>2] = $43;
    $63 = ((($zone$0$lcssa23$i20)) + 4|0);
    HEAP32[$63>>2] = $46;
    $64 = (($count_bot$012$i4) + 1)|0;
    $count_bot$1$i23 = $64;
   }
   $65 = ((($$0115$i2)) + 4|0);
   $66 = (($$016$i1) + -2)|0;
   $67 = ($66>>>0)>(1);
   if ($67) {
    $$0115$i2 = $65;$$016$i1 = $66;$count_bot$012$i4 = $count_bot$1$i23;
   } else {
    $count_bot$0$lcssa$i24 = $count_bot$1$i23;
    break;
   }
  }
 } else {
  $count_bot$0$lcssa$i24 = $count_bot$0$lcssa$i;
 }
 HEAP32[$top_table$0>>2] = $count_top$0$lcssa$i;
 HEAP32[$bot_table$0>>2] = $count_bot$0$lcssa$i24;
 $68 = ($count_top$0$lcssa$i|0)==(0);
 if (!($68)) {
  $69 = ((($top_table$0)) + 4|0);
  $$054 = $count_top$0$lcssa$i;$zone$053 = $69;
  while(1) {
   $70 = ($$054>>>0)>(1);
   if ($70) {
    $71 = ((($zone$053)) + 32|0);
    $72 = HEAP32[$71>>2]|0;
    $73 = HEAP32[$zone$053>>2]|0;
    $74 = (($72) - ($73))|0;
    $75 = ((($zone$053)) + 4|0);
    $76 = HEAP32[$75>>2]|0;
    $77 = ($76|0)>($74|0);
    if ($77) {
     HEAP32[$75>>2] = $74;
     $79 = $73;$81 = $74;
    } else {
     $79 = $73;$81 = $76;
    }
   } else {
    $$pre = HEAP32[$zone$053>>2]|0;
    $$phi$trans$insert73 = ((($zone$053)) + 4|0);
    $$pre74 = HEAP32[$$phi$trans$insert73>>2]|0;
    $79 = $$pre;$81 = $$pre74;
   }
   $78 = ((($zone$053)) + 12|0);
   HEAP32[$78>>2] = $79;
   $80 = (($81) + ($79))|0;
   $82 = ((($zone$053)) + 8|0);
   HEAP32[$82>>2] = $80;
   $83 = (($$054) + -1)|0;
   $84 = ((($zone$053)) + 32|0);
   $85 = ($83|0)==(0);
   if ($85) {
    break;
   } else {
    $$054 = $83;$zone$053 = $84;
   }
  }
 }
 $86 = ($count_bot$0$lcssa$i24|0)==(0);
 if (!($86)) {
  $88 = ((($bot_table$0)) + 4|0);
  $$150 = $count_bot$0$lcssa$i24;$zone1$049 = $88;
  while(1) {
   $89 = ($$150>>>0)>(1);
   $90 = HEAP32[$zone1$049>>2]|0;
   if ($89) {
    $91 = ((($zone1$049)) + 32|0);
    $92 = HEAP32[$91>>2]|0;
    $93 = (($90) - ($92))|0;
    $94 = ((($zone1$049)) + 4|0);
    $95 = HEAP32[$94>>2]|0;
    $96 = ($95|0)<($93|0);
    if ($96) {
     HEAP32[$94>>2] = $93;
     $99 = $93;
    } else {
     $99 = $95;
    }
   } else {
    $$phi$trans$insert78 = ((($zone1$049)) + 4|0);
    $$pre79 = HEAP32[$$phi$trans$insert78>>2]|0;
    $99 = $$pre79;
   }
   $97 = ((($zone1$049)) + 8|0);
   HEAP32[$97>>2] = $90;
   $98 = (($99) + ($90))|0;
   $100 = ((($zone1$049)) + 12|0);
   HEAP32[$100>>2] = $98;
   $101 = (($$150) + -1)|0;
   $102 = ((($zone1$049)) + 32|0);
   $103 = ($101|0)==(0);
   if ($103) {
    break;
   } else {
    $$150 = $101;$zone1$049 = $102;
   }
  }
 }
 $87 = $fuzz << 1;
 $zone4$048 = ((($top_table$0)) + 4|0);
 if (!($68)) {
  $104 = ((($top_table$0)) + 16|0);
  $105 = HEAP32[$104>>2]|0;
  $106 = (($105) - ($fuzz))|0;
  HEAP32[$104>>2] = $106;
  $107 = ((($top_table$0)) + 12|0);
  $$338 = (($count_top$0$lcssa$i) + -1)|0;
  $top$039 = HEAP32[$107>>2]|0;
  $108 = ($$338|0)==(0);
  if ($108) {
   $top$0$lcssa = $top$039;$zone4$1$lcssa = $zone4$048;
  } else {
   $uglygep = ((($top_table$0)) + -28|0);
   $109 = $count_top$0$lcssa$i << 3;
   $scevgep83 = (($uglygep) + ($109<<2)|0);
   $$341 = $$338;$top$042 = $top$039;$zone4$140 = $zone4$048;
   while(1) {
    $110 = ((($zone4$140)) + 32|0);
    $111 = ((($zone4$140)) + 44|0);
    $112 = HEAP32[$111>>2]|0;
    $113 = (($112) - ($top$042))|0;
    $114 = ($113|0)<($87|0);
    if ($114) {
     $115 = (($113|0) / 2)&-1;
     $116 = (($115) + ($top$042))|0;
     HEAP32[$111>>2] = $116;
     $117 = ((($zone4$140)) + 8|0);
     HEAP32[$117>>2] = $116;
    } else {
     $118 = (($top$042) + ($fuzz))|0;
     $119 = ((($zone4$140)) + 8|0);
     HEAP32[$119>>2] = $118;
     $120 = (($112) - ($fuzz))|0;
     HEAP32[$111>>2] = $120;
    }
    $121 = ((($zone4$140)) + 40|0);
    $$3 = (($$341) + -1)|0;
    $top$0 = HEAP32[$121>>2]|0;
    $122 = ($$3|0)==(0);
    if ($122) {
     $top$0$lcssa121 = $top$0;
     break;
    } else {
     $$341 = $$3;$top$042 = $top$0;$zone4$140 = $110;
    }
   }
   $top$0$lcssa = $top$0$lcssa121;$zone4$1$lcssa = $scevgep83;
  }
  $123 = (($top$0$lcssa) + ($fuzz))|0;
  $124 = ((($zone4$1$lcssa)) + 8|0);
  HEAP32[$124>>2] = $123;
 }
 $zone4$048$1 = ((($bot_table$0)) + 4|0);
 if ($86) {
  return;
 }
 $125 = ((($bot_table$0)) + 16|0);
 $126 = HEAP32[$125>>2]|0;
 $127 = (($126) - ($fuzz))|0;
 HEAP32[$125>>2] = $127;
 $128 = ((($bot_table$0)) + 12|0);
 $$338$1 = (($count_bot$0$lcssa$i24) + -1)|0;
 $top$039$1 = HEAP32[$128>>2]|0;
 $129 = ($$338$1|0)==(0);
 if ($129) {
  $top$0$lcssa$1 = $top$039$1;$zone4$1$lcssa$1 = $zone4$048$1;
 } else {
  $uglygep$1 = ((($bot_table$0)) + -28|0);
  $130 = $count_bot$0$lcssa$i24 << 3;
  $scevgep$184 = (($uglygep$1) + ($130<<2)|0);
  $$341$1 = $$338$1;$top$042$1 = $top$039$1;$zone4$140$1 = $zone4$048$1;
  while(1) {
   $131 = ((($zone4$140$1)) + 32|0);
   $132 = ((($zone4$140$1)) + 44|0);
   $133 = HEAP32[$132>>2]|0;
   $134 = (($133) - ($top$042$1))|0;
   $135 = ($134|0)<($87|0);
   if ($135) {
    $138 = (($134|0) / 2)&-1;
    $139 = (($138) + ($top$042$1))|0;
    $$sink = $139;$storemerge = $139;
   } else {
    $136 = (($top$042$1) + ($fuzz))|0;
    $137 = (($133) - ($fuzz))|0;
    $$sink = $136;$storemerge = $137;
   }
   HEAP32[$132>>2] = $storemerge;
   $140 = ((($zone4$140$1)) + 8|0);
   HEAP32[$140>>2] = $$sink;
   $141 = ((($zone4$140$1)) + 40|0);
   $$3$1 = (($$341$1) + -1)|0;
   $top$0$1 = HEAP32[$141>>2]|0;
   $142 = ($$3$1|0)==(0);
   if ($142) {
    $top$0$1$lcssa = $top$0$1;
    break;
   } else {
    $$341$1 = $$3$1;$top$042$1 = $top$0$1;$zone4$140$1 = $131;
   }
  }
  $top$0$lcssa$1 = $top$0$1$lcssa;$zone4$1$lcssa$1 = $scevgep$184;
 }
 $143 = (($top$0$lcssa$1) + ($fuzz))|0;
 $144 = ((($zone4$1$lcssa$1)) + 8|0);
 HEAP32[$144>>2] = $143;
 return;
}
function _psh_globals_set_scale($globals,$x_scale,$y_scale,$x_delta,$y_delta) {
 $globals = $globals|0;
 $x_scale = $x_scale|0;
 $y_scale = $y_scale|0;
 $x_delta = $x_delta|0;
 $y_delta = $y_delta|0;
 var $$$1$i = 0, $$$i = 0, $$$i11 = 0, $$$i5 = 0, $$1$i = 0, $$1$i6 = 0, $$sink$i = 0, $$sink$in$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $count$0$i = 0, $count$0$i7 = 0, $count$02$i = 0, $count$02$i1 = 0, $count$04$i = 0, $count$04$i3 = 0, $count$09$i = 0, $count1$06$1$i = 0, $count1$06$i = 0, $count2$04$1$i = 0, $count2$04$i = 0, $exitcond21$i = 0, $num$014$i = 0, $table$0$i = 0, $threshold$0$lcssa$i = 0, $threshold$015$i = 0, $width$03$i = 0, $width$03$i4 = 0;
 var $zone$010$i = 0, $zone1$05$1$i = 0, $zone1$05$i = 0, $zone2$03$1$i = 0, $zone2$03$1$i$lcssa = 0, $zone2$03$i = 0, $zone2$03$i$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($globals)) + 200|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==($x_scale|0);
 $3 = ((($globals)) + 204|0);
 if ($2) {
  $4 = HEAP32[$3>>2]|0;
  $5 = ($4|0)==($x_delta|0);
  if (!($5)) {
   label = 3;
  }
 } else {
  label = 3;
 }
 if ((label|0) == 3) {
  HEAP32[$0>>2] = $x_scale;
  HEAP32[$3>>2] = $x_delta;
  $6 = ((($globals)) + 4|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)==(0);
  if (!($8)) {
   $9 = ((($globals)) + 8|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = (_FT_MulFix($10,$x_scale)|0);
   $12 = ((($globals)) + 12|0);
   HEAP32[$12>>2] = $11;
   $13 = (($11) + 32)|0;
   $14 = $13 & -64;
   $15 = ((($globals)) + 16|0);
   HEAP32[$15>>2] = $14;
   $count$02$i = (($7) + -1)|0;
   $16 = ($count$02$i|0)==(0);
   if (!($16)) {
    $17 = ((($globals)) + 20|0);
    $count$04$i = $count$02$i;$width$03$i = $17;
    while(1) {
     $18 = HEAP32[$width$03$i>>2]|0;
     $19 = (_FT_MulFix($18,$x_scale)|0);
     $20 = HEAP32[$12>>2]|0;
     $21 = (($19) - ($20))|0;
     $22 = ($21|0)<(0);
     $23 = (0 - ($21))|0;
     $$$i = $22 ? $23 : $21;
     $24 = ($$$i|0)<(128);
     $$1$i = $24 ? $20 : $19;
     $25 = ((($width$03$i)) + 4|0);
     HEAP32[$25>>2] = $$1$i;
     $26 = (($$1$i) + 32)|0;
     $27 = $26 & -64;
     $28 = ((($width$03$i)) + 8|0);
     HEAP32[$28>>2] = $27;
     $29 = ((($width$03$i)) + 12|0);
     $count$0$i = (($count$04$i) + -1)|0;
     $30 = ($count$0$i|0)==(0);
     if ($30) {
      break;
     } else {
      $count$04$i = $count$0$i;$width$03$i = $29;
     }
    }
   }
  }
 }
 $31 = ((($globals)) + 404|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ($32|0)==($y_scale|0);
 $34 = ((($globals)) + 408|0);
 if ($33) {
  $35 = HEAP32[$34>>2]|0;
  $36 = ($35|0)==($y_delta|0);
  if ($36) {
   return;
  }
 }
 HEAP32[$31>>2] = $y_scale;
 HEAP32[$34>>2] = $y_delta;
 $37 = ((($globals)) + 208|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = ($38|0)==(0);
 if (!($39)) {
  $40 = ((($globals)) + 212|0);
  $41 = HEAP32[$40>>2]|0;
  $42 = (_FT_MulFix($41,$y_scale)|0);
  $43 = ((($globals)) + 216|0);
  HEAP32[$43>>2] = $42;
  $44 = (($42) + 32)|0;
  $45 = $44 & -64;
  $46 = ((($globals)) + 220|0);
  HEAP32[$46>>2] = $45;
  $count$02$i1 = (($38) + -1)|0;
  $47 = ($count$02$i1|0)==(0);
  if (!($47)) {
   $48 = ((($globals)) + 224|0);
   $count$04$i3 = $count$02$i1;$width$03$i4 = $48;
   while(1) {
    $49 = HEAP32[$width$03$i4>>2]|0;
    $50 = (_FT_MulFix($49,$y_scale)|0);
    $51 = HEAP32[$43>>2]|0;
    $52 = (($50) - ($51))|0;
    $53 = ($52|0)<(0);
    $54 = (0 - ($52))|0;
    $$$i5 = $53 ? $54 : $52;
    $55 = ($$$i5|0)<(128);
    $$1$i6 = $55 ? $51 : $50;
    $56 = ((($width$03$i4)) + 4|0);
    HEAP32[$56>>2] = $$1$i6;
    $57 = (($$1$i6) + 32)|0;
    $58 = $57 & -64;
    $59 = ((($width$03$i4)) + 8|0);
    HEAP32[$59>>2] = $58;
    $60 = ((($width$03$i4)) + 12|0);
    $count$0$i7 = (($count$04$i3) + -1)|0;
    $61 = ($count$0$i7|0)==(0);
    if ($61) {
     break;
    } else {
     $count$04$i3 = $count$0$i7;$width$03$i4 = $60;
    }
   }
  }
 }
 $62 = ((($globals)) + 412|0);
 $63 = ($y_scale|0)>(34359737);
 $64 = ((($globals)) + 2476|0);
 $65 = HEAP32[$64>>2]|0;
 if ($63) {
  $66 = $65 << 3;
  $67 = (($66|0) / 125)&-1;
  $68 = ($67|0)>($y_scale|0);
  $$sink$in$i = $68;
 } else {
  $69 = ($y_scale*125)|0;
  $70 = $65 << 3;
  $71 = ($69|0)<($70|0);
  $$sink$in$i = $71;
 }
 $$sink$i = $$sink$in$i&1;
 $72 = ((($globals)) + 2492|0);
 HEAP8[$72>>0] = $$sink$i;
 $73 = ((($globals)) + 2480|0);
 $74 = HEAP32[$73>>2]|0;
 $75 = ($74|0)>(0);
 L24: do {
  if ($75) {
   $threshold$015$i = $74;
   while(1) {
    $76 = (_FT_MulFix($threshold$015$i,$y_scale)|0);
    $77 = ($76|0)>(32);
    if (!($77)) {
     $threshold$0$lcssa$i = $threshold$015$i;
     break L24;
    }
    $78 = (($threshold$015$i) + -1)|0;
    $79 = ($threshold$015$i|0)>(1);
    if ($79) {
     $threshold$015$i = $78;
    } else {
     $threshold$0$lcssa$i = $78;
     break;
    }
   }
  } else {
   $threshold$0$lcssa$i = $74;
  }
 } while(0);
 $80 = ((($globals)) + 2484|0);
 HEAP32[$80>>2] = $threshold$0$lcssa$i;
 $81 = ((($globals)) + 1960|0);
 $82 = ((($globals)) + 928|0);
 $83 = ((($globals)) + 1444|0);
 $num$014$i = 0;
 while(1) {
  switch ($num$014$i|0) {
  case 0:  {
   $table$0$i = $62;
   break;
  }
  case 1:  {
   $table$0$i = $82;
   break;
  }
  case 2:  {
   $table$0$i = $83;
   break;
  }
  default: {
   $table$0$i = $81;
  }
  }
  $86 = HEAP32[$table$0$i>>2]|0;
  $87 = ($86|0)==(0);
  if (!($87)) {
   $88 = ((($table$0$i)) + 4|0);
   $count$09$i = $86;$zone$010$i = $88;
   while(1) {
    $89 = ((($zone$010$i)) + 8|0);
    $90 = HEAP32[$89>>2]|0;
    $91 = (_FT_MulFix($90,$y_scale)|0);
    $92 = (($91) + ($y_delta))|0;
    $93 = ((($zone$010$i)) + 28|0);
    HEAP32[$93>>2] = $92;
    $94 = ((($zone$010$i)) + 12|0);
    $95 = HEAP32[$94>>2]|0;
    $96 = (_FT_MulFix($95,$y_scale)|0);
    $97 = (($96) + ($y_delta))|0;
    $98 = ((($zone$010$i)) + 24|0);
    HEAP32[$98>>2] = $97;
    $99 = HEAP32[$zone$010$i>>2]|0;
    $100 = (_FT_MulFix($99,$y_scale)|0);
    $101 = (($100) + ($y_delta))|0;
    $102 = ((($zone$010$i)) + 16|0);
    HEAP32[$102>>2] = $101;
    $103 = ((($zone$010$i)) + 4|0);
    $104 = HEAP32[$103>>2]|0;
    $105 = (_FT_MulFix($104,$y_scale)|0);
    $106 = ((($zone$010$i)) + 20|0);
    HEAP32[$106>>2] = $105;
    $107 = HEAP32[$102>>2]|0;
    $108 = (($107) + 32)|0;
    $109 = $108 & -64;
    HEAP32[$102>>2] = $109;
    $110 = (($count$09$i) + -1)|0;
    $111 = ((($zone$010$i)) + 32|0);
    $112 = ($110|0)==(0);
    if ($112) {
     break;
    } else {
     $count$09$i = $110;$zone$010$i = $111;
    }
   }
  }
  $113 = (($num$014$i) + 1)|0;
  $exitcond21$i = ($113|0)==(4);
  if ($exitcond21$i) {
   break;
  } else {
   $num$014$i = $113;
  }
 }
 $84 = HEAP32[$62>>2]|0;
 $85 = ($84|0)==(0);
 if (!($85)) {
  $114 = ((($globals)) + 416|0);
  $115 = ((($globals)) + 1448|0);
  $count1$06$i = $84;$zone1$05$i = $114;
  while(1) {
   $116 = HEAP32[$83>>2]|0;
   $117 = ($116|0)==(0);
   L46: do {
    if (!($117)) {
     $count2$04$i = $116;$zone2$03$i = $115;
     while(1) {
      $118 = HEAP32[$zone1$05$i>>2]|0;
      $119 = HEAP32[$zone2$03$i>>2]|0;
      $120 = (($118) - ($119))|0;
      $121 = ($120|0)<(0);
      $122 = (0 - ($120))|0;
      $$$i11 = $121 ? $122 : $120;
      $123 = (_FT_MulFix($$$i11,$y_scale)|0);
      $124 = ($123|0)<(64);
      if ($124) {
       $zone2$03$i$lcssa = $zone2$03$i;
       break;
      }
      $137 = (($count2$04$i) + -1)|0;
      $138 = ((($zone2$03$i)) + 32|0);
      $139 = ($137|0)==(0);
      if ($139) {
       break L46;
      } else {
       $count2$04$i = $137;$zone2$03$i = $138;
      }
     }
     $125 = ((($zone2$03$i$lcssa)) + 28|0);
     $126 = HEAP32[$125>>2]|0;
     $127 = ((($zone1$05$i)) + 28|0);
     HEAP32[$127>>2] = $126;
     $128 = ((($zone2$03$i$lcssa)) + 24|0);
     $129 = HEAP32[$128>>2]|0;
     $130 = ((($zone1$05$i)) + 24|0);
     HEAP32[$130>>2] = $129;
     $131 = ((($zone2$03$i$lcssa)) + 16|0);
     $132 = HEAP32[$131>>2]|0;
     $133 = ((($zone1$05$i)) + 16|0);
     HEAP32[$133>>2] = $132;
     $134 = ((($zone2$03$i$lcssa)) + 20|0);
     $135 = HEAP32[$134>>2]|0;
     $136 = ((($zone1$05$i)) + 20|0);
     HEAP32[$136>>2] = $135;
    }
   } while(0);
   $140 = (($count1$06$i) + -1)|0;
   $141 = ((($zone1$05$i)) + 32|0);
   $142 = ($140|0)==(0);
   if ($142) {
    break;
   } else {
    $count1$06$i = $140;$zone1$05$i = $141;
   }
  }
 }
 $143 = HEAP32[$82>>2]|0;
 $144 = ($143|0)==(0);
 if ($144) {
  return;
 }
 $145 = ((($globals)) + 932|0);
 $146 = ((($globals)) + 1964|0);
 $count1$06$1$i = $143;$zone1$05$1$i = $145;
 while(1) {
  $147 = HEAP32[$81>>2]|0;
  $148 = ($147|0)==(0);
  L59: do {
   if (!($148)) {
    $count2$04$1$i = $147;$zone2$03$1$i = $146;
    while(1) {
     $149 = HEAP32[$zone1$05$1$i>>2]|0;
     $150 = HEAP32[$zone2$03$1$i>>2]|0;
     $151 = (($149) - ($150))|0;
     $152 = ($151|0)<(0);
     $153 = (0 - ($151))|0;
     $$$1$i = $152 ? $153 : $151;
     $154 = (_FT_MulFix($$$1$i,$y_scale)|0);
     $155 = ($154|0)<(64);
     if ($155) {
      $zone2$03$1$i$lcssa = $zone2$03$1$i;
      break;
     }
     $156 = (($count2$04$1$i) + -1)|0;
     $157 = ((($zone2$03$1$i)) + 32|0);
     $158 = ($156|0)==(0);
     if ($158) {
      break L59;
     } else {
      $count2$04$1$i = $156;$zone2$03$1$i = $157;
     }
    }
    $159 = ((($zone2$03$1$i$lcssa)) + 28|0);
    $160 = HEAP32[$159>>2]|0;
    $161 = ((($zone1$05$1$i)) + 28|0);
    HEAP32[$161>>2] = $160;
    $162 = ((($zone2$03$1$i$lcssa)) + 24|0);
    $163 = HEAP32[$162>>2]|0;
    $164 = ((($zone1$05$1$i)) + 24|0);
    HEAP32[$164>>2] = $163;
    $165 = ((($zone2$03$1$i$lcssa)) + 16|0);
    $166 = HEAP32[$165>>2]|0;
    $167 = ((($zone1$05$1$i)) + 16|0);
    HEAP32[$167>>2] = $166;
    $168 = ((($zone2$03$1$i$lcssa)) + 20|0);
    $169 = HEAP32[$168>>2]|0;
    $170 = ((($zone1$05$1$i)) + 20|0);
    HEAP32[$170>>2] = $169;
   }
  } while(0);
  $171 = (($count1$06$1$i) + -1)|0;
  $172 = ((($zone1$05$1$i)) + 32|0);
  $173 = ($171|0)==(0);
  if ($173) {
   break;
  } else {
   $count1$06$1$i = $171;$zone1$05$1$i = $172;
  }
 }
 return;
}
function _psh_globals_destroy($globals) {
 $globals = $globals|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($globals|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = HEAP32[$globals>>2]|0;
 $2 = ((($globals)) + 4|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($globals)) + 208|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($globals)) + 412|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($globals)) + 928|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($globals)) + 1444|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($globals)) + 1960|0);
 HEAP32[$7>>2] = 0;
 _ft_mem_free($1,$globals);
 return;
}
function _t1_hints_open($hints) {
 $hints = $hints|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($hints)) + 4|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($hints)) + 12|0);
 HEAP32[$1>>2] = 1;
 $2 = ((($hints)) + 16|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($hints)) + 28|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($hints)) + 40|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($hints)) + 52|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($hints)) + 64|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($hints)) + 76|0);
 HEAP32[$7>>2] = 0;
 return;
}
function _ps_hints_close($hints,$end_point) {
 $hints = $hints|0;
 $end_point = $end_point|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $error$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($hints)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  $error$1 = $1;
  return ($error$1|0);
 }
 $3 = HEAP32[$hints>>2]|0;
 $4 = ((($hints)) + 16|0);
 $5 = (_ps_dimension_end($4,$end_point,$3)|0);
 $6 = ($5|0)==(0);
 if (!($6)) {
  $error$1 = $5;
  return ($error$1|0);
 }
 $7 = ((($hints)) + 52|0);
 $8 = (_ps_dimension_end($7,$end_point,$3)|0);
 $error$1 = $8;
 return ($error$1|0);
}
function _ps_dimension_end($dim,$end_point,$memory) {
 $dim = $dim|0;
 $end_point = $end_point|0;
 $memory = $memory|0;
 var $$lcssa41 = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$phi$iZ2D = 0, $$pre$phi12$i$iZ2D = 0, $$pre10$i$i = 0, $$pre11$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $count$0$lcssa$i$i = 0, $count$06$i$i = 0, $dummy$i$i = 0, $error$2$i = 0, $error$i$i$i = 0, $exitcond$i$i = 0, $index1$014$i = 0, $index1$014$in$i = 0;
 var $index1$index2$i$i = 0, $index2$011$i = 0, $index2$011$i40 = 0, $index2$index1$i$i = 0, $or$cond$i = 0, $p1$0$lcssa$i$i = 0, $p1$05$i$i = 0, $p2$0$lcssa$i$i = 0, $p2$04$i$i = 0, $pos$08$i$i = 0, $pos$15$i$i = 0, $read$06$i$i = 0, $table$idx$i = 0, $table$idx$val$i = 0, $write$07$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i$i$i = sp + 16|0;
 $dummy$i$i = sp;
 $0 = ((($dim)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  $3 = ((($dim)) + 20|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = (($4) + ($1<<4)|0);
  $6 = ((($5)) + -16|0);
  $7 = ((($6)) + 12|0);
  HEAP32[$7>>2] = $end_point;
 }
 $8 = ((($dim)) + 24|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)>(1);
 if (!($10)) {
  $error$2$i = 0;
  STACKTOP = sp;return ($error$2$i|0);
 }
 $table$idx$i = ((($dim)) + 32|0);
 $44 = $9;$index1$014$in$i = $9;
 L7: while(1) {
  $index1$014$i = (($index1$014$in$i) + -1)|0;
  $11 = (($index1$014$in$i) + -2)|0;
  $12 = ($11|0)>(-1);
  L9: do {
   if ($12) {
    $table$idx$val$i = HEAP32[$table$idx$i>>2]|0;
    $13 = (((($table$idx$val$i) + ($index1$014$i<<4)|0)) + 8|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = (($table$idx$val$i) + ($index1$014$i<<4)|0);
    $16 = HEAP32[$15>>2]|0;
    $index2$011$i = $11;
    L11: while(1) {
     $17 = (((($table$idx$val$i) + ($index2$011$i<<4)|0)) + 8|0);
     $18 = HEAP32[$17>>2]|0;
     $19 = (($table$idx$val$i) + ($index2$011$i<<4)|0);
     $20 = HEAP32[$19>>2]|0;
     $21 = ($16>>>0)<($20>>>0);
     $22 = $21 ? $16 : $20;
     $23 = ($22>>>0)>(7);
     if ($23) {
      $count$06$i$i = $22;$p1$05$i$i = $14;$p2$04$i$i = $18;
      while(1) {
       $24 = HEAP8[$p1$05$i$i>>0]|0;
       $25 = HEAP8[$p2$04$i$i>>0]|0;
       $26 = $25 & $24;
       $27 = ($26<<24>>24)==(0);
       if (!($27)) {
        $index2$011$i40 = $index2$011$i;
        break L11;
       }
       $28 = ((($p1$05$i$i)) + 1|0);
       $29 = ((($p2$04$i$i)) + 1|0);
       $30 = (($count$06$i$i) + -8)|0;
       $31 = ($30>>>0)>(7);
       if ($31) {
        $count$06$i$i = $30;$p1$05$i$i = $28;$p2$04$i$i = $29;
       } else {
        $count$0$lcssa$i$i = $30;$p1$0$lcssa$i$i = $28;$p2$0$lcssa$i$i = $29;
        break;
       }
      }
     } else {
      $count$0$lcssa$i$i = $22;$p1$0$lcssa$i$i = $14;$p2$0$lcssa$i$i = $18;
     }
     $32 = ($count$0$lcssa$i$i|0)==(0);
     if (!($32)) {
      $33 = HEAP8[$p1$0$lcssa$i$i>>0]|0;
      $34 = HEAP8[$p2$0$lcssa$i$i>>0]|0;
      $35 = $34 & $33;
      $36 = $35&255;
      $37 = 255 >>> $count$0$lcssa$i$i;
      $38 = $37 ^ -1;
      $39 = $36 & $38;
      $40 = ($39|0)==(0);
      if (!($40)) {
       $index2$011$i40 = $index2$011$i;
       break;
      }
     }
     $102 = (($index2$011$i) + -1)|0;
     $103 = ($index2$011$i|0)>(0);
     if ($103) {
      $index2$011$i = $102;
     } else {
      $104 = $44;
      break L9;
     }
    }
    $41 = ($index2$011$i40>>>0)>($index1$014$i>>>0);
    $index1$index2$i$i = $41 ? $index2$011$i40 : $index1$014$i;
    $index2$index1$i$i = $41 ? $index1$014$i : $index2$011$i40;
    $42 = ($index2$index1$i$i>>>0)<($index1$index2$i$i>>>0);
    $43 = ($index1$index2$i$i>>>0)<($44>>>0);
    $or$cond$i = $42 & $43;
    if ($or$cond$i) {
     $46 = (($table$idx$val$i) + ($index1$index2$i$i<<4)|0);
     $47 = (($table$idx$val$i) + ($index2$index1$i$i<<4)|0);
     $48 = HEAP32[$47>>2]|0;
     $49 = HEAP32[$46>>2]|0;
     $50 = ($49|0)==(0);
     if (!($50)) {
      $51 = ($49>>>0)>($48>>>0);
      if ($51) {
       $52 = (((($table$idx$val$i) + ($index2$index1$i$i<<4)|0)) + 4|0);
       $53 = HEAP32[$52>>2]|0;
       $54 = (($53) + 7)|0;
       $55 = $54 >>> 3;
       $56 = (($49) + 7)|0;
       $57 = $56 >>> 3;
       HEAP32[$error$i$i$i>>2] = 0;
       $58 = ($57>>>0)>($55>>>0);
       if ($58) {
        $59 = (($57) + 7)|0;
        $60 = $59 & 1073741816;
        $61 = (((($table$idx$val$i) + ($index2$index1$i$i<<4)|0)) + 8|0);
        $62 = HEAP32[$61>>2]|0;
        $63 = (_ft_mem_realloc($memory,1,$55,$60,$62,$error$i$i$i)|0);
        HEAP32[$61>>2] = $63;
        $64 = HEAP32[$error$i$i$i>>2]|0;
        $65 = ($64|0)==(0);
        if (!($65)) {
         $$lcssa41 = $64;
         break L7;
        }
        $66 = $60 << 3;
        HEAP32[$52>>2] = $66;
        $$pre$phi$iZ2D = $61;
       } else {
        $$pre$i = (((($table$idx$val$i) + ($index2$index1$i$i<<4)|0)) + 8|0);
        $$pre$phi$iZ2D = $$pre$i;
       }
       $pos$08$i$i = $48;
       while(1) {
        $67 = HEAP32[$47>>2]|0;
        $68 = ($67>>>0)>($pos$08$i$i>>>0);
        if ($68) {
         $69 = HEAP32[$$pre$phi$iZ2D>>2]|0;
         $70 = $pos$08$i$i >>> 3;
         $71 = (($69) + ($70)|0);
         $72 = HEAP8[$71>>0]|0;
         $73 = $72&255;
         $74 = $pos$08$i$i & 7;
         $75 = 128 >>> $74;
         $76 = $75 ^ 255;
         $77 = $73 & $76;
         $78 = $77&255;
         HEAP8[$71>>0] = $78;
        }
        $79 = (($pos$08$i$i) + 1)|0;
        $exitcond$i$i = ($79|0)==($49|0);
        if ($exitcond$i$i) {
         $$pre$phi12$i$iZ2D = $57;
         break;
        } else {
         $pos$08$i$i = $79;
        }
       }
      } else {
       $$pre10$i$i = (($49) + 7)|0;
       $$pre11$i$i = $$pre10$i$i >>> 3;
       $$pre$phi12$i$iZ2D = $$pre11$i$i;
      }
      $80 = ($$pre$phi12$i$iZ2D|0)==(0);
      if (!($80)) {
       $81 = (((($table$idx$val$i) + ($index2$index1$i$i<<4)|0)) + 8|0);
       $82 = HEAP32[$81>>2]|0;
       $83 = (((($table$idx$val$i) + ($index1$index2$i$i<<4)|0)) + 8|0);
       $84 = HEAP32[$83>>2]|0;
       $pos$15$i$i = $$pre$phi12$i$iZ2D;$read$06$i$i = $84;$write$07$i$i = $82;
       while(1) {
        $85 = HEAP8[$write$07$i$i>>0]|0;
        $86 = HEAP8[$read$06$i$i>>0]|0;
        $87 = $86 | $85;
        HEAP8[$write$07$i$i>>0] = $87;
        $88 = ((($write$07$i$i)) + 1|0);
        $89 = ((($read$06$i$i)) + 1|0);
        $90 = (($pos$15$i$i) + -1)|0;
        $91 = ($90|0)==(0);
        if ($91) {
         break;
        } else {
         $pos$15$i$i = $90;$read$06$i$i = $89;$write$07$i$i = $88;
        }
       }
      }
     }
     HEAP32[$46>>2] = 0;
     $92 = (((($table$idx$val$i) + ($index1$index2$i$i<<4)|0)) + 12|0);
     HEAP32[$92>>2] = 0;
     $93 = HEAP32[$8>>2]|0;
     $94 = $index1$index2$i$i ^ -1;
     $95 = (($93) + ($94))|0;
     $96 = ($95|0)>(0);
     if ($96) {
      ;HEAP32[$dummy$i$i>>2]=HEAP32[$46>>2]|0;HEAP32[$dummy$i$i+4>>2]=HEAP32[$46+4>>2]|0;HEAP32[$dummy$i$i+8>>2]=HEAP32[$46+8>>2]|0;HEAP32[$dummy$i$i+12>>2]=HEAP32[$46+12>>2]|0;
      $97 = ((($46)) + 16|0);
      $98 = $95 << 4;
      _memmove(($46|0),($97|0),($98|0))|0;
      $99 = (($46) + ($95<<4)|0);
      ;HEAP32[$99>>2]=HEAP32[$dummy$i$i>>2]|0;HEAP32[$99+4>>2]=HEAP32[$dummy$i$i+4>>2]|0;HEAP32[$99+8>>2]=HEAP32[$dummy$i$i+8>>2]|0;HEAP32[$99+12>>2]=HEAP32[$dummy$i$i+12>>2]|0;
      $$pre$i$i = HEAP32[$8>>2]|0;
      $101 = $$pre$i$i;
     } else {
      $101 = $93;
     }
     $100 = (($101) + -1)|0;
     HEAP32[$8>>2] = $100;
     $104 = $100;
    } else {
     $104 = $44;
    }
   } else {
    $104 = $44;
   }
  } while(0);
  $45 = ($index1$014$i|0)>(1);
  if ($45) {
   $44 = $104;$index1$014$in$i = $index1$014$i;
  } else {
   $error$2$i = 0;
   label = 33;
   break;
  }
 }
 if ((label|0) == 33) {
  STACKTOP = sp;return ($error$2$i|0);
 }
 $error$2$i = $$lcssa41;
 STACKTOP = sp;return ($error$2$i|0);
}
function _t1_hints_stem($hints,$dimension,$coords) {
 $hints = $hints|0;
 $dimension = $dimension|0;
 $coords = $coords|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$coords>>2]|0;
 $1 = (_FT_RoundFix($0)|0);
 $2 = ((($coords)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_FT_RoundFix($3)|0);
 $5 = ((($hints)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  return;
 }
 $8 = $4 >> 16;
 $9 = $1 >> 16;
 $10 = ($dimension>>>0)>(1);
 $11 = ($dimension|0)!=(0);
 $12 = $11&1;
 $$0$i = $10 ? $12 : $dimension;
 $13 = (((($hints)) + 16|0) + (($$0$i*36)|0)|0);
 $14 = HEAP32[$hints>>2]|0;
 $15 = (_ps_dimension_add_t1stem($13,$9,$8,$14,0)|0);
 $16 = ($15|0)==(0);
 if ($16) {
  return;
 }
 HEAP32[$5>>2] = $15;
 return;
}
function _ps_dimension_add_t1stem($dim,$pos,$len,$memory,$aindex) {
 $dim = $dim|0;
 $pos = $pos|0;
 $len = $len|0;
 $memory = $memory|0;
 $aindex = $aindex|0;
 var $$ = 0, $$03 = 0, $$1 = 0, $$phi$trans$insert$i$i = 0, $$pn = 0, $$pos = 0, $$pre = 0, $$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $9 = 0, $error$0 = 0, $error$i$i = 0, $error$i$i$i = 0, $error$i$i5 = 0, $flags$1 = 0, $hint$sroa$0$014 = 0, $idx$0$lcssa = 0, $idx$013 = 0, $idx$015 = 0, $mask$0$i$sink$i$ph19 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i$i5 = sp + 8|0;
 $error$i$i$i = sp + 4|0;
 $error$i$i = sp;
 $0 = ($len|0)<(0);
 if ($0) {
  $1 = ($len|0)==(-21);
  $2 = $1 ? -21 : 0;
  $$pos = (($2) + ($pos))|0;
  $$ = $1 ? 3 : 1;
  $$03 = 0;$$1 = $$pos;$flags$1 = $$;
 } else {
  $$03 = $len;$$1 = $pos;$flags$1 = 0;
 }
 $3 = ($aindex|0)!=(0|0);
 if ($3) {
  HEAP32[$aindex>>2] = -1;
 }
 $4 = HEAP32[$dim>>2]|0;
 $5 = ((($dim)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($4|0)==(0);
 $8 = $6;
 $9 = $6;
 L7: do {
  if ($7) {
   $idx$0$lcssa = 0;
   label = 9;
  } else {
   $hint$sroa$0$014 = $6;$idx$015 = 0;
   while(1) {
    $10 = $hint$sroa$0$014;
    $11 = HEAP32[$10>>2]|0;
    $12 = ($11|0)==($$1|0);
    if ($12) {
     $13 = ((($10)) + 4|0);
     $14 = HEAP32[$13>>2]|0;
     $15 = ($14|0)==($$03|0);
     if ($15) {
      $idx$013 = $idx$015;
      break L7;
     }
    }
    $16 = (($idx$015) + 1)|0;
    $17 = ((($10)) + 12|0);
    $18 = $17;
    $19 = ($16>>>0)<($4>>>0);
    if ($19) {
     $hint$sroa$0$014 = $18;$idx$015 = $16;
    } else {
     $idx$0$lcssa = $16;
     label = 9;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 9) {
  $20 = (($4) + 1)|0;
  $21 = ((($dim)) + 4|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = ($20>>>0)<($22>>>0);
  if ($23) {
   $31 = $9;
  } else {
   HEAP32[$error$i$i>>2] = 0;
   $24 = ($22>>>0)<($20>>>0);
   do {
    if ($24) {
     $25 = (($4) + 8)|0;
     $26 = $25 & -8;
     $27 = (_ft_mem_realloc($memory,12,$22,$26,$8,$error$i$i)|0);
     HEAP32[$5>>2] = $27;
     $28 = HEAP32[$error$i$i>>2]|0;
     $29 = ($28|0)==(0);
     if ($29) {
      HEAP32[$21>>2] = $26;
      $79 = $27;
      break;
     }
     $error$0 = $28;
     STACKTOP = sp;return ($error$0|0);
    } else {
     $79 = $9;
    }
   } while(0);
   $31 = $79;
  }
  $30 = (($31) + (($20*12)|0)|0);
  $32 = ((($30)) + -12|0);
  $33 = ((($32)) + 4|0);
  $34 = ((($32)) + 8|0);
  HEAP32[$dim>>2] = $20;
  HEAP32[$32>>2] = $$1;
  HEAP32[$33>>2] = $$03;
  HEAP32[$34>>2] = $flags$1;
  $idx$013 = $idx$0$lcssa;
 }
 $35 = ((($dim)) + 12|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = ($36|0)==(0);
 if ($37) {
  $38 = ((($dim)) + 16|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = ($39|0)==(0);
  do {
   if ($40) {
    HEAP32[$error$i$i$i>>2] = 0;
    $41 = ((($dim)) + 20|0);
    $42 = HEAP32[$41>>2]|0;
    $43 = (_ft_mem_realloc($memory,16,0,8,$42,$error$i$i$i)|0);
    HEAP32[$41>>2] = $43;
    $44 = HEAP32[$error$i$i$i>>2]|0;
    $45 = ($44|0)==(0);
    if ($45) {
     HEAP32[$38>>2] = 8;
     $46 = $43;
     break;
    }
    $error$0 = $44;
    STACKTOP = sp;return ($error$0|0);
   } else {
    $$phi$trans$insert$i$i = ((($dim)) + 20|0);
    $$pre$i$i = HEAP32[$$phi$trans$insert$i$i>>2]|0;
    $46 = $$pre$i$i;
   }
  } while(0);
  HEAP32[$46>>2] = 0;
  $47 = ((($46)) + 12|0);
  HEAP32[$47>>2] = 0;
  HEAP32[$35>>2] = 1;
  $$pn = $46;
  label = 24;
 } else {
  $48 = ((($dim)) + 20|0);
  $49 = HEAP32[$48>>2]|0;
  $50 = (($49) + ($36<<4)|0);
  $51 = ((($50)) + -16|0);
  $$pre = HEAP32[$51>>2]|0;
  $52 = ($$pre>>>0)>($idx$013>>>0);
  if ($52) {
   $mask$0$i$sink$i$ph19 = $51;
  } else {
   $$pn = $51;
   label = 24;
  }
 }
 if ((label|0) == 24) {
  $53 = (($idx$013) + 1)|0;
  $54 = ((($$pn)) + 4|0);
  $55 = HEAP32[$54>>2]|0;
  $56 = (($55) + 7)|0;
  $57 = $56 >>> 3;
  $58 = (($idx$013) + 8)|0;
  $59 = $58 >>> 3;
  HEAP32[$error$i$i5>>2] = 0;
  $60 = ($59>>>0)>($57>>>0);
  do {
   if ($60) {
    $61 = (($59) + 7)|0;
    $62 = $61 & 1073741816;
    $63 = ((($$pn)) + 8|0);
    $64 = HEAP32[$63>>2]|0;
    $65 = (_ft_mem_realloc($memory,1,$57,$62,$64,$error$i$i5)|0);
    HEAP32[$63>>2] = $65;
    $66 = HEAP32[$error$i$i5>>2]|0;
    $67 = ($66|0)==(0);
    if ($67) {
     $68 = $62 << 3;
     HEAP32[$54>>2] = $68;
     break;
    }
    $error$0 = $66;
    STACKTOP = sp;return ($error$0|0);
   }
  } while(0);
  HEAP32[$$pn>>2] = $53;
  $mask$0$i$sink$i$ph19 = $$pn;
 }
 $69 = ((($mask$0$i$sink$i$ph19)) + 8|0);
 $70 = HEAP32[$69>>2]|0;
 $71 = $idx$013 >>> 3;
 $72 = (($70) + ($71)|0);
 $73 = HEAP8[$72>>0]|0;
 $74 = $73&255;
 $75 = $idx$013 & 7;
 $76 = 128 >>> $75;
 $77 = $74 | $76;
 $78 = $77&255;
 HEAP8[$72>>0] = $78;
 if (!($3)) {
  $error$0 = 0;
  STACKTOP = sp;return ($error$0|0);
 }
 HEAP32[$aindex>>2] = $idx$013;
 $error$0 = 0;
 STACKTOP = sp;return ($error$0|0);
}
function _ps_hints_t1stem3($hints,$dimension,$stems) {
 $hints = $hints|0;
 $dimension = $dimension|0;
 $stems = $stems|0;
 var $$01 = 0, $$old1$i = 0, $$phi$trans$insert$i$i = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$phi$iZ2D = 0, $$pre$phi24$iZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0;
 var $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cond$i = 0;
 var $cond18$i = 0, $count$020$i = 0, $counter$sroa$0$019$i = 0, $counter$sroa$0$1$i = 0, $error$0$ph = 0, $error$i$i$i = 0, $error$i$i10$i = 0, $error$i$i14$i = 0, $error$i$i4$i = 0, $idx = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i$i14$i = sp + 24|0;
 $error$i$i10$i = sp + 20|0;
 $error$i$i4$i = sp + 16|0;
 $error$i$i$i = sp + 12|0;
 $idx = sp;
 $0 = ((($hints)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  STACKTOP = sp;return;
 }
 $3 = HEAP32[$hints>>2]|0;
 $4 = ($dimension>>>0)>(1);
 $5 = ($dimension|0)!=(0);
 $6 = $5&1;
 $$01 = $4 ? $6 : $dimension;
 $7 = (((($hints)) + 16|0) + (($$01*36)|0)|0);
 $8 = ((($hints)) + 12|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)==(1);
 L4: do {
  if ($10) {
   $11 = HEAP32[$stems>>2]|0;
   $12 = (_FT_RoundFix($11)|0);
   $13 = $12 >> 16;
   $14 = ((($stems)) + 4|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = (_FT_RoundFix($15)|0);
   $17 = $16 >> 16;
   $18 = (_ps_dimension_add_t1stem($7,$13,$17,$3,$idx)|0);
   $19 = ($18|0)==(0);
   if ($19) {
    $20 = ((($stems)) + 8|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = (_FT_RoundFix($21)|0);
    $23 = $22 >> 16;
    $24 = ((($stems)) + 12|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = (_FT_RoundFix($25)|0);
    $27 = $26 >> 16;
    $28 = ((($idx)) + 4|0);
    $29 = (_ps_dimension_add_t1stem($7,$23,$27,$3,$28)|0);
    $30 = ($29|0)==(0);
    if ($30) {
     $181 = ((($stems)) + 16|0);
     $182 = HEAP32[$181>>2]|0;
     $183 = (_FT_RoundFix($182)|0);
     $184 = $183 >> 16;
     $185 = ((($stems)) + 20|0);
     $186 = HEAP32[$185>>2]|0;
     $187 = (_FT_RoundFix($186)|0);
     $188 = $187 >> 16;
     $189 = ((($idx)) + 8|0);
     $190 = (_ps_dimension_add_t1stem($7,$184,$188,$3,$189)|0);
     $191 = ($190|0)==(0);
     if ($191) {
      $34 = HEAP32[$idx>>2]|0;
      $38 = HEAP32[$28>>2]|0;
      $42 = HEAP32[$189>>2]|0;
      $90 = (((((($hints)) + 16|0) + (($$01*36)|0)|0)) + 24|0);
      $75 = HEAP32[$90>>2]|0;
      $cond18$i = ($75|0)==(0);
      L9: do {
       if ($cond18$i) {
        label = 13;
       } else {
        $31 = (((((($hints)) + 16|0) + (($$01*36)|0)|0)) + 32|0);
        $32 = HEAP32[$31>>2]|0;
        $33 = $34 >> 3;
        $35 = $34 & 7;
        $36 = 128 >>> $35;
        $37 = $38 >> 3;
        $39 = $38 & 7;
        $40 = 128 >>> $39;
        $41 = $42 >> 3;
        $43 = $42 & 7;
        $44 = 128 >>> $43;
        $count$020$i = $75;$counter$sroa$0$019$i = $32;
        while(1) {
         $45 = $counter$sroa$0$019$i;
         $46 = HEAP32[$45>>2]|0;
         $47 = ($46>>>0)>($34>>>0);
         if ($47) {
          $48 = ((($45)) + 8|0);
          $49 = HEAP32[$48>>2]|0;
          $50 = (($49) + ($33)|0);
          $51 = HEAP8[$50>>0]|0;
          $52 = $51&255;
          $53 = $52 & $36;
          $54 = ($53|0)==(0);
          if (!($54)) {
           $counter$sroa$0$1$i = $counter$sroa$0$019$i;
           break L9;
          }
         }
         $55 = ($46>>>0)>($38>>>0);
         if ($55) {
          $56 = ((($45)) + 8|0);
          $57 = HEAP32[$56>>2]|0;
          $58 = (($57) + ($37)|0);
          $59 = HEAP8[$58>>0]|0;
          $60 = $59&255;
          $61 = $60 & $40;
          $62 = ($61|0)==(0);
          if (!($62)) {
           $counter$sroa$0$1$i = $counter$sroa$0$019$i;
           break L9;
          }
         }
         $63 = ($46>>>0)>($42>>>0);
         if ($63) {
          $64 = ((($45)) + 8|0);
          $65 = HEAP32[$64>>2]|0;
          $66 = (($65) + ($41)|0);
          $67 = HEAP8[$66>>0]|0;
          $68 = $67&255;
          $69 = $68 & $44;
          $70 = ($69|0)==(0);
          if (!($70)) {
           $counter$sroa$0$1$i = $counter$sroa$0$019$i;
           break L9;
          }
         }
         $71 = (($count$020$i) + -1)|0;
         $72 = ((($45)) + 16|0);
         $73 = $72;
         $cond$i = ($71|0)==(0);
         if ($cond$i) {
          label = 13;
          break;
         } else {
          $count$020$i = $71;$counter$sroa$0$019$i = $73;
         }
        }
       }
      } while(0);
      if ((label|0) == 13) {
       $74 = (($75) + 1)|0;
       $76 = (((((($hints)) + 16|0) + (($$01*36)|0)|0)) + 28|0);
       $77 = HEAP32[$76>>2]|0;
       $78 = ($74>>>0)>($77>>>0);
       do {
        if ($78) {
         HEAP32[$error$i$i14$i>>2] = 0;
         $79 = (($75) + 8)|0;
         $80 = $79 & -8;
         $81 = (((((($hints)) + 16|0) + (($$01*36)|0)|0)) + 32|0);
         $82 = HEAP32[$81>>2]|0;
         $83 = (_ft_mem_realloc($3,16,$77,$80,$82,$error$i$i14$i)|0);
         HEAP32[$81>>2] = $83;
         $84 = HEAP32[$error$i$i14$i>>2]|0;
         $85 = ($84|0)==(0);
         if ($85) {
          HEAP32[$76>>2] = $80;
          $87 = $83;
          break;
         } else {
          $error$0$ph = $84;
          break L4;
         }
        } else {
         $$phi$trans$insert$i$i = (((((($hints)) + 16|0) + (($$01*36)|0)|0)) + 32|0);
         $$pre$i$i = HEAP32[$$phi$trans$insert$i$i>>2]|0;
         $87 = $$pre$i$i;
        }
       } while(0);
       $86 = (($87) + ($74<<4)|0);
       $88 = ((($86)) + -16|0);
       HEAP32[$88>>2] = 0;
       $89 = ((($88)) + 12|0);
       HEAP32[$89>>2] = 0;
       HEAP32[$90>>2] = $74;
       $91 = $88;
       $counter$sroa$0$1$i = $91;
      }
      $92 = ($34|0)>(-1);
      if ($92) {
       $93 = $counter$sroa$0$1$i;
       $94 = HEAP32[$93>>2]|0;
       $95 = ($94>>>0)>($34>>>0);
       if (!($95)) {
        $96 = (($34) + 1)|0;
        $97 = ((($93)) + 4|0);
        $98 = HEAP32[$97>>2]|0;
        $99 = (($98) + 7)|0;
        $100 = $99 >>> 3;
        $101 = (($34) + 8)|0;
        $102 = $101 >>> 3;
        HEAP32[$error$i$i10$i>>2] = 0;
        $103 = ($102>>>0)>($100>>>0);
        do {
         if ($103) {
          $104 = (($102) + 7)|0;
          $105 = $104 & 1073741816;
          $106 = ((($93)) + 8|0);
          $107 = HEAP32[$106>>2]|0;
          $108 = (_ft_mem_realloc($3,1,$100,$105,$107,$error$i$i10$i)|0);
          HEAP32[$106>>2] = $108;
          $109 = HEAP32[$error$i$i10$i>>2]|0;
          $110 = ($109|0)==(0);
          if ($110) {
           $111 = $105 << 3;
           HEAP32[$97>>2] = $111;
           break;
          } else {
           $error$0$ph = $109;
           break L4;
          }
         }
        } while(0);
        HEAP32[$93>>2] = $96;
       }
       $112 = ((($93)) + 8|0);
       $113 = HEAP32[$112>>2]|0;
       $114 = $34 >>> 3;
       $115 = (($113) + ($114)|0);
       $116 = HEAP8[$115>>0]|0;
       $117 = $116&255;
       $118 = $34 & 7;
       $119 = 128 >>> $118;
       $120 = $117 | $119;
       $121 = $120&255;
       HEAP8[$115>>0] = $121;
      }
      $122 = ($38|0)>(-1);
      if ($122) {
       $123 = $counter$sroa$0$1$i;
       $124 = HEAP32[$123>>2]|0;
       $125 = ($124>>>0)>($38>>>0);
       if (!($125)) {
        $126 = (($38) + 1)|0;
        $127 = ((($123)) + 4|0);
        $128 = HEAP32[$127>>2]|0;
        $129 = (($128) + 7)|0;
        $130 = $129 >>> 3;
        $131 = (($38) + 8)|0;
        $132 = $131 >>> 3;
        HEAP32[$error$i$i4$i>>2] = 0;
        $133 = ($132>>>0)>($130>>>0);
        do {
         if ($133) {
          $134 = (($132) + 7)|0;
          $135 = $134 & 1073741816;
          $136 = ((($123)) + 8|0);
          $137 = HEAP32[$136>>2]|0;
          $138 = (_ft_mem_realloc($3,1,$130,$135,$137,$error$i$i4$i)|0);
          HEAP32[$136>>2] = $138;
          $139 = HEAP32[$error$i$i4$i>>2]|0;
          $140 = ($139|0)==(0);
          if ($140) {
           $141 = $135 << 3;
           HEAP32[$127>>2] = $141;
           break;
          } else {
           $error$0$ph = $139;
           break L4;
          }
         }
        } while(0);
        HEAP32[$123>>2] = $126;
       }
       $142 = ((($123)) + 8|0);
       $143 = HEAP32[$142>>2]|0;
       $144 = $38 >>> 3;
       $145 = (($143) + ($144)|0);
       $146 = HEAP8[$145>>0]|0;
       $147 = $146&255;
       $148 = $38 & 7;
       $149 = 128 >>> $148;
       $150 = $147 | $149;
       $151 = $150&255;
       HEAP8[$145>>0] = $151;
       $152 = ($42|0)>(-1);
       if ($152) {
        $$pre$phi$iZ2D = $123;$$pre$phi24$iZ2D = $123;
        label = 37;
       }
      } else {
       $$old1$i = ($42|0)>(-1);
       if ($$old1$i) {
        $$pre$i = $counter$sroa$0$1$i;
        $$pre$phi$iZ2D = $$pre$i;$$pre$phi24$iZ2D = $$pre$i;
        label = 37;
       }
      }
      if ((label|0) == 37) {
       $153 = HEAP32[$$pre$phi24$iZ2D>>2]|0;
       $154 = ($153>>>0)>($42>>>0);
       if (!($154)) {
        $155 = (($42) + 1)|0;
        $156 = ((($$pre$phi$iZ2D)) + 4|0);
        $157 = HEAP32[$156>>2]|0;
        $158 = (($157) + 7)|0;
        $159 = $158 >>> 3;
        $160 = (($42) + 8)|0;
        $161 = $160 >>> 3;
        HEAP32[$error$i$i$i>>2] = 0;
        $162 = ($161>>>0)>($159>>>0);
        do {
         if ($162) {
          $163 = (($161) + 7)|0;
          $164 = $163 & 1073741816;
          $165 = ((($$pre$phi$iZ2D)) + 8|0);
          $166 = HEAP32[$165>>2]|0;
          $167 = (_ft_mem_realloc($3,1,$159,$164,$166,$error$i$i$i)|0);
          HEAP32[$165>>2] = $167;
          $168 = HEAP32[$error$i$i$i>>2]|0;
          $169 = ($168|0)==(0);
          if ($169) {
           $170 = $164 << 3;
           HEAP32[$156>>2] = $170;
           break;
          } else {
           $error$0$ph = $168;
           break L4;
          }
         }
        } while(0);
        HEAP32[$$pre$phi24$iZ2D>>2] = $155;
       }
       $171 = ((($$pre$phi$iZ2D)) + 8|0);
       $172 = HEAP32[$171>>2]|0;
       $173 = $42 >>> 3;
       $174 = (($172) + ($173)|0);
       $175 = HEAP8[$174>>0]|0;
       $176 = $175&255;
       $177 = $42 & 7;
       $178 = 128 >>> $177;
       $179 = $176 | $178;
       $180 = $179&255;
       HEAP8[$174>>0] = $180;
      }
      STACKTOP = sp;return;
     } else {
      $error$0$ph = $190;
     }
    } else {
     $error$0$ph = $29;
    }
   } else {
    $error$0$ph = $18;
   }
  } else {
   $error$0$ph = 6;
  }
 } while(0);
 HEAP32[$0>>2] = $error$0$ph;
 STACKTOP = sp;return;
}
function _ps_hints_t1reset($hints,$end_point) {
 $hints = $hints|0;
 $end_point = $end_point|0;
 var $$phi$trans$insert$i$i = 0, $$phi$trans$insert$i$i3 = 0, $$pre$i$i = 0, $$pre$i$i4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$0$ph = 0, $error$i$i$i = 0, $error$i$i$i1 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i$i$i1 = sp + 4|0;
 $error$i$i$i = sp;
 $0 = ((($hints)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  STACKTOP = sp;return;
 }
 $3 = HEAP32[$hints>>2]|0;
 $4 = ((($hints)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(1);
 L4: do {
  if ($6) {
   $7 = ((($hints)) + 28|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ($8|0)==(0);
   if (!($9)) {
    $10 = ((($hints)) + 36|0);
    $11 = HEAP32[$10>>2]|0;
    $12 = (($11) + ($8<<4)|0);
    $13 = ((($12)) + -16|0);
    $14 = ((($13)) + 12|0);
    HEAP32[$14>>2] = $end_point;
   }
   $15 = (($8) + 1)|0;
   $16 = ((($hints)) + 32|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ($15>>>0)>($17>>>0);
   do {
    if ($18) {
     HEAP32[$error$i$i$i>>2] = 0;
     $19 = (($8) + 8)|0;
     $20 = $19 & -8;
     $21 = ((($hints)) + 36|0);
     $22 = HEAP32[$21>>2]|0;
     $23 = (_ft_mem_realloc($3,16,$17,$20,$22,$error$i$i$i)|0);
     HEAP32[$21>>2] = $23;
     $24 = HEAP32[$error$i$i$i>>2]|0;
     $25 = ($24|0)==(0);
     if ($25) {
      HEAP32[$16>>2] = $20;
      $27 = $23;
      break;
     } else {
      $error$0$ph = $24;
      break L4;
     }
    } else {
     $$phi$trans$insert$i$i = ((($hints)) + 36|0);
     $$pre$i$i = HEAP32[$$phi$trans$insert$i$i>>2]|0;
     $27 = $$pre$i$i;
    }
   } while(0);
   $26 = (($27) + ($15<<4)|0);
   $28 = ((($26)) + -16|0);
   HEAP32[$28>>2] = 0;
   $29 = ((($28)) + 12|0);
   HEAP32[$29>>2] = 0;
   HEAP32[$7>>2] = $15;
   $30 = ((($hints)) + 64|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = ($31|0)==(0);
   if (!($32)) {
    $33 = ((($hints)) + 72|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = (($34) + ($31<<4)|0);
    $36 = ((($35)) + -16|0);
    $37 = ((($36)) + 12|0);
    HEAP32[$37>>2] = $end_point;
   }
   $38 = (($31) + 1)|0;
   $39 = ((($hints)) + 68|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = ($38>>>0)>($40>>>0);
   do {
    if ($41) {
     HEAP32[$error$i$i$i1>>2] = 0;
     $42 = (($31) + 8)|0;
     $43 = $42 & -8;
     $44 = ((($hints)) + 72|0);
     $45 = HEAP32[$44>>2]|0;
     $46 = (_ft_mem_realloc($3,16,$40,$43,$45,$error$i$i$i1)|0);
     HEAP32[$44>>2] = $46;
     $47 = HEAP32[$error$i$i$i1>>2]|0;
     $48 = ($47|0)==(0);
     if ($48) {
      HEAP32[$39>>2] = $43;
      $50 = $46;
      break;
     } else {
      $error$0$ph = $47;
      break L4;
     }
    } else {
     $$phi$trans$insert$i$i3 = ((($hints)) + 72|0);
     $$pre$i$i4 = HEAP32[$$phi$trans$insert$i$i3>>2]|0;
     $50 = $$pre$i$i4;
    }
   } while(0);
   $49 = (($50) + ($38<<4)|0);
   $51 = ((($49)) + -16|0);
   HEAP32[$51>>2] = 0;
   $52 = ((($51)) + 12|0);
   HEAP32[$52>>2] = 0;
   HEAP32[$30>>2] = $38;
   STACKTOP = sp;return;
  } else {
   $error$0$ph = 6;
  }
 } while(0);
 HEAP32[$0>>2] = $error$0$ph;
 STACKTOP = sp;return;
}
function _ps_hints_apply($ps_hints,$outline,$globals,$hint_mode) {
 $ps_hints = $ps_hints|0;
 $outline = $outline|0;
 $globals = $globals|0;
 $hint_mode = $hint_mode|0;
 var $$$i = 0, $$0 = 0, $$1$i = 0, $$2$i = 0, $$finished$1$i$i = 0, $$finished$1$i$i$lcssa = 0, $$idx$val = 0, $$idx$val$i = 0, $$idx3 = 0, $$idx3$i = 0, $$idx3$val = 0, $$idx3$val$i = 0, $$idx5$i$pre$phiZZZ2D = 0, $$idx5$val$i = 0, $$idx6$i$pre$phiZZZ2D = 0, $$idx6$val$i = 0, $$idx7$i = 0, $$idx7$val$i = 0, $$idx8$i = 0, $$idx8$val$i = 0;
 var $$lcssa = 0, $$lcssa244 = 0, $$lcssa245 = 0, $$lcssa246 = 0, $$lcssa250 = 0, $$lcssa251 = 0, $$lcssa252 = 0, $$lcssa253 = 0, $$lcssa255 = 0, $$lcssa256 = 0, $$lcssa257 = 0, $$lcssa258 = 0, $$lcssa259 = 0, $$lcssa261 = 0, $$lcssa263 = 0, $$lcssa265 = 0, $$lcssa266 = 0, $$lcssa267 = 0, $$lcssa268 = 0, $$lcssa269 = 0;
 var $$lcssa270 = 0, $$lobit$i = 0, $$ph = 0, $$phi$trans$insert = 0, $$pn$i = 0, $$pn$i23 = 0, $$pr = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i17 = 0, $$pre$i40 = 0, $$pre$i47 = 0, $$pre$phi129Z2D = 0, $$pre$phi131Z2D = 0, $$pre$phi133Z2D = 0, $$pre$phiZ2D = 0, $$pre125 = 0, $$pre126 = 0, $$pre127$pre$phiZZ2D = 0;
 var $$pre128$pre$phiZZ2D = 0, $$pre132 = 0, $$pre135 = 0, $$pre136 = 0, $$pre137 = 0, $$pre4$i = 0, $$pre41$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0;
 var $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0;
 var $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0;
 var $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0;
 var $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0;
 var $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0;
 var $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0;
 var $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0;
 var $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0;
 var $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0;
 var $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0;
 var $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0;
 var $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0;
 var $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0;
 var $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0;
 var $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0;
 var $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0;
 var $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0;
 var $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0;
 var $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0;
 var $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0;
 var $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0;
 var $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0;
 var $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0;
 var $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0;
 var $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0;
 var $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0;
 var $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0;
 var $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0;
 var $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0;
 var $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0;
 var $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0;
 var $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0;
 var $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0;
 var $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0;
 var $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0;
 var $812 = 0, $813 = 0, $814 = 0, $815 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $after$0$i = 0, $after$0$i$i = 0, $after$0$i$i$lcssa = 0, $after3$0$i = 0, $before$0$i = 0, $before$017$i$i = 0, $before$017$i$i$lcssa = 0, $before$017$i$i$phi = 0, $before$1$i = 0, $before2$0$i = 0, $blue_count$01$i = 0, $blue_count$16$i = 0, $contour$014$i = 0, $contour$029$i = 0, $count$02$i = 0, $count$02$i25 = 0, $count$025$i = 0;
 var $count$03$i = 0, $count$03$us$i = 0, $count$03$us$i$i = 0, $count3$09$i = 0, $cur_a$0$i = 0, $cur_ab$0$i = 0, $cur_c$0$i = 0, $delta$015$i = 0, $delta$1$i = 0, $delta$2$i = 0, $dimension$088 = 0, $end$0$i$i = 0, $end$1$i$i = 0, $error$i = 0, $error$i28 = 0, $exitcond = 0, $finished$1$i$i = 0, $first$012$i = 0, $first$03$i = 0, $first$031$i = 0;
 var $first$2$i = 0, $first$2$i$lcssa = 0, $first$3$be$i = 0, $first$37$i = 0, $fit_count$05$i = 0, $fit_count$1$i = 0, $fit_count$1$i$lcssa = 0, $glyph$idx1$val$pre$i = 0, $glyph$idx2$val$idx$val$pre$i = 0, $glyph_count$015$i = 0, $glyphrec = 0, $glyphrec$idx$val = 0, $glyphrec$idx1$val = 0, $glyphrec$idx10$val = 0, $glyphrec$idx11$val = 0, $glyphrec$idx2 = 0, $glyphrec$idx2$val = 0, $glyphrec$idx2$val$idx = 0, $glyphrec$idx2$val$idx$val = 0, $glyphrec$idx5$val = 0;
 var $glyphrec$idx7$val = 0, $glyphrec$idx7$val$pre = 0, $glyphrec$idx8$val = 0, $glyphrec$idx9$val = 0, $hint$01$i = 0, $in_x$0$i$i = 0, $in_y$0$i$i = 0, $insert$0$lcssa$i = 0, $insert$011$i = 0, $mask$0$idx$val$i = 0, $mask$0$idx4$i = 0, $mask$0$idx4$val$i = 0, $mask$015$i = 0, $n$018$i = 0, $n$018$i$i = 0, $n$02$i = 0, $n$02$us$i = 0, $n$030$i = 0, $n$1$i = 0, $n$2$i = 0;
 var $n3$022$i = 0, $next$0$i = 0, $nn$0$lcssa30$i = 0, $nn$03$i = 0, $nn$03$i$lcssa = 0, $nn$1$i = 0, $nn$1$lcssa$i = 0, $num_contours$013$i = 0, $num_masks$013$i = 0, $num_strongs$0$$i = 0, $num_strongs$0$$i$lcssa = 0, $num_strongs$023$i = 0, $num_strongs$218$i = 0, $num_strongs$3$i = 0, $num_strongs$3$i$lcssa = 0, $or$cond = 0, $or$cond$i = 0, $or$cond40$i = 0, $org_a$0$i = 0, $org_ab$0$i = 0;
 var $orient_prev$0$i$i = 0, $point$0$first$0$i = 0, $point$0$i = 0, $point$01$i = 0, $point$01$i45 = 0, $point$01$us$i = 0, $point$011$i = 0, $point$02$i = 0, $point$02$us$i = 0, $point$02$us$i$i = 0, $point$024$i = 0, $point$024$i29 = 0, $point$04$i = 0, $point$1$i = 0, $point$110$i = 0, $point$119$i = 0, $point$2$i = 0, $point$26$i = 0, $point2$023$i = 0, $point4$010$i = 0;
 var $rescale$0 = 0, $result$0$i$i = 0, $result$0$i3$i = 0, $scale_ab$0$i = 0, $start$0$i$i = 0, $start$1$i$i = 0, $strongs$0$i = 0, $strongs_0$i = 0, $vec$01$i = 0, $vec$01$us$i = 0, $vec$01$us$i$i = 0, $x_scale$0 = 0, $zone$02$i = 0, $zone$18$i = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 208|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $strongs_0$i = sp + 136|0;
 $error$i28 = sp + 132|0;
 $error$i = sp + 128|0;
 $glyphrec = sp;
 $0 = ((($outline)) + 2|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = ($1<<16>>16)==(0);
 if ($2) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = HEAP32[$outline>>2]|0;
 $4 = $3&65535;
 $5 = ($4<<16>>16)==(0);
 if ($5) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 dest=$glyphrec; stop=dest+128|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $6 = HEAP32[$globals>>2]|0;
 $7 = ((($glyphrec)) + 16|0);
 HEAP32[$7>>2] = $6;
 $8 = $3 >> 16;
 $9 = (_ft_mem_realloc($6,40,0,$8,0,$error$i)|0);
 $10 = ((($glyphrec)) + 8|0);
 HEAP32[$10>>2] = $9;
 $11 = HEAP32[$error$i>>2]|0;
 $12 = ($11|0)==(0);
 if ($12) {
  $13 = HEAP16[$outline>>1]|0;
  $14 = $13 << 16 >> 16;
  $15 = (_ft_mem_realloc($6,8,0,$14,0,$error$i)|0);
  $16 = ((($glyphrec)) + 12|0);
  HEAP32[$16>>2] = $15;
  $17 = HEAP32[$error$i>>2]|0;
  $18 = ($17|0)==(0);
  if ($18) {
   $19 = HEAP16[$0>>1]|0;
   $20 = $19 << 16 >> 16;
   HEAP32[$glyphrec>>2] = $20;
   $21 = HEAP16[$outline>>1]|0;
   $22 = $21 << 16 >> 16;
   $23 = ((($glyphrec)) + 4|0);
   HEAP32[$23>>2] = $22;
   $24 = HEAP32[$10>>2]|0;
   $25 = ($21<<16>>16)==(0);
   if (!($25)) {
    $26 = ((($outline)) + 12|0);
    $27 = HEAP32[$26>>2]|0;
    $contour$029$i = $15;$first$031$i = 0;$n$030$i = 0;
    while(1) {
     $28 = (($27) + ($n$030$i<<1)|0);
     $29 = HEAP16[$28>>1]|0;
     $30 = $29 << 16 >> 16;
     $31 = (($30) + 1)|0;
     $32 = (($31) - ($first$031$i))|0;
     $33 = (($24) + (($first$031$i*40)|0)|0);
     HEAP32[$contour$029$i>>2] = $33;
     $34 = ((($contour$029$i)) + 4|0);
     HEAP32[$34>>2] = $32;
     $35 = ($31|0)==($first$031$i|0);
     if (!($35)) {
      $36 = (($24) + (($31*40)|0)|0);
      $37 = ((($36)) + -40|0);
      HEAP32[$33>>2] = $37;
      $38 = (((($24) + (($first$031$i*40)|0)|0)) + 8|0);
      HEAP32[$38>>2] = $contour$029$i;
      $39 = ($32>>>0)>(1);
      if ($39) {
       $count$025$i = $32;$point$024$i = $33;
       while(1) {
        $40 = ((($point$024$i)) + 40|0);
        $41 = ((($point$024$i)) + 4|0);
        HEAP32[$41>>2] = $40;
        HEAP32[$40>>2] = $point$024$i;
        $42 = ((($point$024$i)) + 48|0);
        HEAP32[$42>>2] = $contour$029$i;
        $43 = (($count$025$i) + -1)|0;
        $44 = ($43>>>0)>(1);
        if ($44) {
         $count$025$i = $43;$point$024$i = $40;
        } else {
         $46 = $30;
         break;
        }
       }
      } else {
       $46 = $first$031$i;
      }
      $45 = (((($24) + (($46*40)|0)|0)) + 4|0);
      HEAP32[$45>>2] = $33;
     }
     $47 = ((($contour$029$i)) + 8|0);
     $48 = (($n$030$i) + 1)|0;
     $49 = ($48>>>0)<($22>>>0);
     if ($49) {
      $contour$029$i = $47;$first$031$i = $31;$n$030$i = $48;
     } else {
      break;
     }
    }
   }
   $50 = ((($outline)) + 4|0);
   $51 = HEAP32[$50>>2]|0;
   $52 = ($19<<16>>16)==(0);
   if ($52) {
    $53 = ((($glyphrec)) + 20|0);
    HEAP32[$53>>2] = $outline;
    $54 = ((($glyphrec)) + 24|0);
    HEAP32[$54>>2] = $globals;
   } else {
    $55 = $24;
    $56 = ((($outline)) + 8|0);
    $n3$022$i = 0;$point2$023$i = $24;
    while(1) {
     $57 = HEAP32[$point2$023$i>>2]|0;
     $58 = (($57) - ($55))|0;
     $59 = (($58|0) / 40)&-1;
     $60 = ((($point2$023$i)) + 4|0);
     $61 = HEAP32[$60>>2]|0;
     $62 = (($61) - ($55))|0;
     $63 = (($62|0) / 40)&-1;
     $64 = HEAP32[$56>>2]|0;
     $65 = (($64) + ($n3$022$i)|0);
     $66 = HEAP8[$65>>0]|0;
     $67 = $66 & 1;
     $68 = ($67<<24>>24)==(0);
     if ($68) {
      $69 = ((($point2$023$i)) + 12|0);
      HEAP32[$69>>2] = 1;
     }
     $70 = (($51) + ($n3$022$i<<3)|0);
     $71 = HEAP32[$70>>2]|0;
     $72 = (($51) + ($59<<3)|0);
     $73 = HEAP32[$72>>2]|0;
     $74 = (($71) - ($73))|0;
     $75 = (((($51) + ($n3$022$i<<3)|0)) + 4|0);
     $76 = HEAP32[$75>>2]|0;
     $77 = (((($51) + ($59<<3)|0)) + 4|0);
     $78 = HEAP32[$77>>2]|0;
     $79 = (($76) - ($78))|0;
     $80 = ($74|0)<(0);
     $81 = (0 - ($74))|0;
     $82 = $80 ? $81 : $74;
     $83 = ($79|0)<(0);
     $84 = (0 - ($79))|0;
     $85 = $83 ? $84 : $79;
     $86 = ($85*12)|0;
     $87 = ($86|0)<($82|0);
     if ($87) {
      $88 = $74 >> 31;
      $89 = $88 & -4;
      $90 = $89 | 2;
      $result$0$i$i = $90;
     } else {
      $91 = ($82*12)|0;
      $92 = ($91|0)<($85|0);
      if ($92) {
       $93 = $79 >> 31;
       $94 = $93 & 2;
       $95 = (($94) + -1)|0;
       $result$0$i$i = $95;
      } else {
       $result$0$i$i = 4;
      }
     }
     $96 = $result$0$i$i&255;
     $97 = ((($point2$023$i)) + 20|0);
     HEAP8[$97>>0] = $96;
     $98 = (($51) + ($63<<3)|0);
     $99 = HEAP32[$98>>2]|0;
     $100 = (($99) - ($71))|0;
     $101 = (((($51) + ($63<<3)|0)) + 4|0);
     $102 = HEAP32[$101>>2]|0;
     $103 = (($102) - ($76))|0;
     $104 = ($100|0)<(0);
     $105 = (0 - ($100))|0;
     $106 = $104 ? $105 : $100;
     $107 = ($103|0)<(0);
     $108 = (0 - ($103))|0;
     $109 = $107 ? $108 : $103;
     $110 = ($109*12)|0;
     $111 = ($110|0)<($106|0);
     if ($111) {
      $112 = $100 >> 31;
      $113 = $112 & -4;
      $114 = $113 | 2;
      $result$0$i3$i = $114;
     } else {
      $115 = ($106*12)|0;
      $116 = ($115|0)<($109|0);
      if ($116) {
       $117 = $103 >> 31;
       $118 = $117 & 2;
       $119 = (($118) + -1)|0;
       $result$0$i3$i = $119;
      } else {
       $result$0$i3$i = 4;
      }
     }
     $120 = $result$0$i3$i&255;
     $121 = ((($point2$023$i)) + 21|0);
     HEAP8[$121>>0] = $120;
     $122 = ((($point2$023$i)) + 12|0);
     $123 = HEAP32[$122>>2]|0;
     $124 = $123 & 1;
     $125 = ($124|0)==(0);
     do {
      if ($125) {
       $127 = ($result$0$i$i|0)==($result$0$i3$i|0);
       if ($127) {
        $128 = ($result$0$i$i|0)==(4);
        if ($128) {
         $129 = (_ft_corner_is_flat($74,$79,$100,$103)|0);
         $130 = ($129|0)==(0);
         if ($130) {
          break;
         }
         $$pre$i = HEAP32[$122>>2]|0;
         $132 = $$pre$i;
        } else {
         $132 = $123;
        }
        $131 = $132 | 2;
        HEAP32[$122>>2] = $131;
       }
      } else {
       $126 = $123 | 2;
       HEAP32[$122>>2] = $126;
      }
     } while(0);
     $133 = (($n3$022$i) + 1)|0;
     $134 = ((($point2$023$i)) + 40|0);
     $135 = HEAP32[$glyphrec>>2]|0;
     $136 = ($133>>>0)<($135>>>0);
     if ($136) {
      $n3$022$i = $133;$point2$023$i = $134;
     } else {
      $$lcssa270 = $135;
      break;
     }
    }
    $glyph$idx1$val$pre$i = HEAP32[$10>>2]|0;
    $glyph$idx2$val$idx$val$pre$i = HEAP32[$50>>2]|0;
    $137 = ((($glyphrec)) + 20|0);
    HEAP32[$137>>2] = $outline;
    $138 = ((($glyphrec)) + 24|0);
    HEAP32[$138>>2] = $globals;
    $139 = ($$lcssa270|0)==(0);
    if (!($139)) {
     $count$03$us$i$i = $$lcssa270;$point$02$us$i$i = $glyph$idx1$val$pre$i;$vec$01$us$i$i = $glyph$idx2$val$idx$val$pre$i;
     while(1) {
      $140 = ((($point$02$us$i$i)) + 16|0);
      HEAP32[$140>>2] = 0;
      $141 = ((($point$02$us$i$i)) + 24|0);
      HEAP32[$141>>2] = 0;
      $142 = HEAP32[$vec$01$us$i$i>>2]|0;
      $143 = ((($point$02$us$i$i)) + 28|0);
      HEAP32[$143>>2] = $142;
      $144 = ((($vec$01$us$i$i)) + 4|0);
      $145 = HEAP32[$144>>2]|0;
      $146 = ((($point$02$us$i$i)) + 32|0);
      HEAP32[$146>>2] = $145;
      $147 = (($count$03$us$i$i) + -1)|0;
      $148 = ((($point$02$us$i$i)) + 40|0);
      $149 = ((($vec$01$us$i$i)) + 8|0);
      $150 = ($147|0)==(0);
      if ($150) {
       break;
      } else {
       $count$03$us$i$i = $147;$point$02$us$i$i = $148;$vec$01$us$i$i = $149;
      }
     }
    }
   }
   $151 = HEAP32[$23>>2]|0;
   $152 = ($151|0)==(0);
   if (!($152)) {
    $n$018$i$i = 0;
    while(1) {
     $153 = HEAP32[$16>>2]|0;
     $154 = (((($153) + ($n$018$i$i<<3)|0)) + 4|0);
     $155 = HEAP32[$154>>2]|0;
     $156 = ($155>>>0)<(4);
     L56: do {
      if (!($156)) {
       $157 = (($153) + ($n$018$i$i<<3)|0);
       $158 = HEAP32[$157>>2]|0;
       $159 = ((($158)) + 28|0);
       $160 = ((($158)) + 32|0);
       $end$0$i$i = $158;
       while(1) {
        $161 = ((($end$0$i$i)) + 4|0);
        $162 = HEAP32[$161>>2]|0;
        $163 = ($162|0)==($158|0);
        if ($163) {
         break L56;
        }
        $164 = ((($162)) + 28|0);
        $165 = HEAP32[$164>>2]|0;
        $166 = HEAP32[$159>>2]|0;
        $167 = ((($162)) + 32|0);
        $168 = HEAP32[$167>>2]|0;
        $169 = HEAP32[$160>>2]|0;
        $170 = ($165|0)==($166|0);
        $171 = ($168|0)==($169|0);
        $172 = $170 & $171;
        if ($172) {
         $end$0$i$i = $162;
        } else {
         $$lcssa255 = $162;$$lcssa256 = $165;$$lcssa257 = $166;$$lcssa258 = $168;$$lcssa259 = $169;
         break;
        }
       }
       $173 = (($$lcssa256) - ($$lcssa257))|0;
       $174 = (($$lcssa258) - ($$lcssa259))|0;
       $175 = HEAP32[$158>>2]|0;
       $176 = ($175|0)==($158|0);
       if (!($176)) {
        $178 = $175;$181 = $$lcssa257;$185 = $$lcssa259;$before$017$i$i = $158;
        while(1) {
         $177 = ((($178)) + 28|0);
         $179 = HEAP32[$177>>2]|0;
         $180 = (($181) - ($179))|0;
         $182 = ((($178)) + 32|0);
         $183 = HEAP32[$182>>2]|0;
         $184 = (($185) - ($183))|0;
         $186 = ($181|0)==($179|0);
         $187 = ($185|0)==($183|0);
         $188 = $186 & $187;
         if (!($188)) {
          $191 = (_ft_corner_orientation($173,$174,$180,$184)|0);
          $192 = ($191|0)==(0);
          if (!($192)) {
           $$lcssa261 = $180;$$lcssa263 = $184;$$lcssa265 = $191;$before$017$i$i$lcssa = $before$017$i$i;
           break;
          }
         }
         $189 = HEAP32[$178>>2]|0;
         $190 = ($189|0)==($158|0);
         if ($190) {
          break L56;
         }
         $$pre$i$i = HEAP32[$177>>2]|0;
         $$pre41$i$i = HEAP32[$182>>2]|0;
         $before$017$i$i$phi = $178;$178 = $189;$181 = $$pre$i$i;$185 = $$pre41$i$i;$before$017$i$i = $before$017$i$i$phi;
        }
        $end$1$i$i = $$lcssa255;$in_x$0$i$i = $$lcssa261;$in_y$0$i$i = $$lcssa263;$orient_prev$0$i$i = $$lcssa265;$start$0$i$i = $before$017$i$i$lcssa;
        while(1) {
         $after$0$i$i = $end$1$i$i;$finished$1$i$i = 0;
         while(1) {
          $193 = ((($after$0$i$i)) + 4|0);
          $194 = HEAP32[$193>>2]|0;
          $195 = ($194|0)==($before$017$i$i$lcssa|0);
          $$finished$1$i$i = $195 ? 1 : $finished$1$i$i;
          $196 = ((($194)) + 28|0);
          $197 = HEAP32[$196>>2]|0;
          $198 = ((($after$0$i$i)) + 28|0);
          $199 = HEAP32[$198>>2]|0;
          $200 = (($197) - ($199))|0;
          $201 = ((($194)) + 32|0);
          $202 = HEAP32[$201>>2]|0;
          $203 = ((($after$0$i$i)) + 32|0);
          $204 = HEAP32[$203>>2]|0;
          $205 = (($202) - ($204))|0;
          $206 = ($197|0)==($199|0);
          $207 = ($202|0)==($204|0);
          $208 = $206 & $207;
          if ($208) {
           $after$0$i$i = $194;$finished$1$i$i = $$finished$1$i$i;
           continue;
          }
          $209 = (_ft_corner_orientation($in_x$0$i$i,$in_y$0$i$i,$200,$205)|0);
          $210 = ($209|0)==(0);
          if ($210) {
           $after$0$i$i = $194;$finished$1$i$i = $$finished$1$i$i;
          } else {
           $$finished$1$i$i$lcssa = $$finished$1$i$i;$$lcssa266 = $194;$$lcssa267 = $200;$$lcssa268 = $205;$$lcssa269 = $209;$after$0$i$i$lcssa = $after$0$i$i;
           break;
          }
         }
         $211 = $$lcssa269 ^ $orient_prev$0$i$i;
         $212 = ($211|0)<(0);
         if ($212) {
          $start$1$i$i = $start$0$i$i;
          while(1) {
           $213 = ((($start$1$i$i)) + 12|0);
           $214 = HEAP32[$213>>2]|0;
           $215 = $214 | 4;
           HEAP32[$213>>2] = $215;
           $216 = ((($start$1$i$i)) + 4|0);
           $217 = HEAP32[$216>>2]|0;
           $218 = ($217|0)==($after$0$i$i$lcssa|0);
           if ($218) {
            break;
           } else {
            $start$1$i$i = $217;
           }
          }
          $219 = ((($after$0$i$i$lcssa)) + 12|0);
          $220 = HEAP32[$219>>2]|0;
          $221 = $220 | 4;
          HEAP32[$219>>2] = $221;
         }
         $222 = ($$finished$1$i$i$lcssa|0)==(0);
         if ($222) {
          $end$1$i$i = $$lcssa266;$in_x$0$i$i = $$lcssa267;$in_y$0$i$i = $$lcssa268;$orient_prev$0$i$i = $$lcssa269;$start$0$i$i = $after$0$i$i$lcssa;
         } else {
          break;
         }
        }
       }
      }
     } while(0);
     $223 = (($n$018$i$i) + 1)|0;
     $224 = HEAP32[$23>>2]|0;
     $225 = ($223>>>0)<($224>>>0);
     if ($225) {
      $n$018$i$i = $223;
     } else {
      break;
     }
    }
   }
   $226 = ((($glyphrec)) + 28|0);
   $227 = ((($ps_hints)) + 16|0);
   $228 = ((($ps_hints)) + 28|0);
   $229 = (_psh_hint_table_init($226,$227,$228,$6)|0);
   HEAP32[$error$i>>2] = $229;
   $230 = ($229|0)==(0);
   if ($230) {
    $231 = ((($glyphrec)) + 68|0);
    $232 = ((($ps_hints)) + 52|0);
    $233 = ((($ps_hints)) + 64|0);
    $234 = (_psh_hint_table_init($231,$232,$233,$6)|0);
    $235 = ($234|0)==(0);
    if ($235) {
     $236 = ((($glyphrec)) + 24|0);
     $237 = HEAP32[$236>>2]|0;
     $238 = ((($237)) + 200|0);
     $239 = HEAP32[$238>>2]|0;
     $240 = ((($237)) + 404|0);
     $241 = HEAP32[$240>>2]|0;
     $242 = ((($globals)) + 416|0);
     $243 = HEAP32[$242>>2]|0;
     $244 = (_FT_MulFix($243,$241)|0);
     $245 = (($244) + 32)|0;
     $246 = $245 & -64;
     $247 = ($246|0)==(0);
     $248 = ($244|0)==($246|0);
     $or$cond = $247 | $248;
     if ($or$cond) {
      $rescale$0 = 1;
     } else {
      $249 = (_FT_MulDiv($241,$246,$244)|0);
      $250 = ($246|0)<($244|0);
      if ($250) {
       $251 = (($239|0) / 50)&-1;
       $252 = (($239) - ($251))|0;
       $x_scale$0 = $252;
      } else {
       $x_scale$0 = $239;
      }
      $253 = HEAP32[$236>>2]|0;
      _psh_globals_set_scale($253,$x_scale$0,$249,0,0);
      $rescale$0 = 0;
     }
     $254 = ((($glyphrec)) + 120|0);
     HEAP8[$254>>0] = 1;
     $255 = ((($glyphrec)) + 121|0);
     HEAP8[$255>>0] = 1;
     $256 = ($hint_mode|0)==(2);
     $257 = $hint_mode | 1;
     $258 = ($257|0)==(3);
     $259 = $258&1;
     $260 = ((($glyphrec)) + 122|0);
     HEAP8[$260>>0] = $259;
     $261 = ($hint_mode|0)==(4);
     $262 = $256 | $261;
     $263 = $262&1;
     $264 = ((($glyphrec)) + 123|0);
     HEAP8[$264>>0] = $263;
     $265 = ($hint_mode|0)!=(1);
     $266 = $265&1;
     $267 = ((($glyphrec)) + 124|0);
     HEAP8[$267>>0] = $266;
     $glyphrec$idx2 = ((($glyphrec)) + 20|0);
     $268 = ((($globals)) + 412|0);
     $269 = ((($globals)) + 416|0);
     $270 = ((($globals)) + 928|0);
     $271 = ((($globals)) + 2488|0);
     $272 = ((($globals)) + 2492|0);
     $273 = ((($globals)) + 2484|0);
     $dimension$088 = 0;
     while(1) {
      $glyphrec$idx$val = HEAP32[$glyphrec>>2]|0;
      $glyphrec$idx1$val = HEAP32[$10>>2]|0;
      $glyphrec$idx2$val = HEAP32[$glyphrec$idx2>>2]|0;
      $glyphrec$idx2$val$idx = ((($glyphrec$idx2$val)) + 4|0);
      $glyphrec$idx2$val$idx$val = HEAP32[$glyphrec$idx2$val$idx>>2]|0;
      $274 = ($glyphrec$idx$val|0)==(0);
      if (!($274)) {
       $275 = ($dimension$088|0)==(0);
       if ($275) {
        $count$03$us$i = $glyphrec$idx$val;$point$02$us$i = $glyphrec$idx1$val;$vec$01$us$i = $glyphrec$idx2$val$idx$val;
        while(1) {
         $276 = ((($point$02$us$i)) + 16|0);
         HEAP32[$276>>2] = 0;
         $277 = ((($point$02$us$i)) + 24|0);
         HEAP32[$277>>2] = 0;
         $278 = HEAP32[$vec$01$us$i>>2]|0;
         $279 = ((($point$02$us$i)) + 28|0);
         HEAP32[$279>>2] = $278;
         $280 = ((($vec$01$us$i)) + 4|0);
         $281 = HEAP32[$280>>2]|0;
         $282 = ((($point$02$us$i)) + 32|0);
         HEAP32[$282>>2] = $281;
         $283 = (($count$03$us$i) + -1)|0;
         $284 = ((($point$02$us$i)) + 40|0);
         $285 = ((($vec$01$us$i)) + 8|0);
         $286 = ($283|0)==(0);
         if ($286) {
          break;
         } else {
          $count$03$us$i = $283;$point$02$us$i = $284;$vec$01$us$i = $285;
         }
        }
       } else {
        $count$03$i = $glyphrec$idx$val;$point$02$i = $glyphrec$idx1$val;$vec$01$i = $glyphrec$idx2$val$idx$val;
        while(1) {
         $287 = ((($point$02$i)) + 16|0);
         HEAP32[$287>>2] = 0;
         $288 = ((($point$02$i)) + 24|0);
         HEAP32[$288>>2] = 0;
         $289 = ((($vec$01$i)) + 4|0);
         $290 = HEAP32[$289>>2]|0;
         $291 = ((($point$02$i)) + 28|0);
         HEAP32[$291>>2] = $290;
         $292 = HEAP32[$vec$01$i>>2]|0;
         $293 = ((($point$02$i)) + 32|0);
         HEAP32[$293>>2] = $292;
         $294 = (($count$03$i) + -1)|0;
         $295 = ((($point$02$i)) + 40|0);
         $296 = ((($vec$01$i)) + 8|0);
         $297 = ($294|0)==(0);
         if ($297) {
          break;
         } else {
          $count$03$i = $294;$point$02$i = $295;$vec$01$i = $296;
         }
        }
       }
      }
      $298 = HEAP32[$23>>2]|0;
      $299 = ($298|0)==(0);
      L100: do {
       if ($299) {
        $n$1$i = 0;
       } else {
        $300 = HEAP32[$16>>2]|0;
        $n$018$i = 0;
        while(1) {
         $301 = (($300) + ($n$018$i<<3)|0);
         $302 = HEAP32[$301>>2]|0;
         $303 = (((($300) + ($n$018$i<<3)|0)) + 4|0);
         $304 = HEAP32[$303>>2]|0;
         $305 = ($304|0)==(0);
         L104: do {
          if (!($305)) {
           $306 = ((($302)) + 28|0);
           $before$0$i = $302;
           while(1) {
            $307 = HEAP32[$before$0$i>>2]|0;
            $308 = ($307|0)==($302|0);
            if ($308) {
             $n$2$i = $n$018$i;
             label = 95;
             break L100;
            }
            $309 = ((($307)) + 28|0);
            $310 = HEAP32[$309>>2]|0;
            $311 = HEAP32[$306>>2]|0;
            $312 = ($310|0)==($311|0);
            if ($312) {
             $before$0$i = $307;
            } else {
             $$lcssa244 = $307;
             break;
            }
           }
           $313 = ((($$lcssa244)) + 4|0);
           $314 = HEAP32[$313>>2]|0;
           $before$1$i = $$lcssa244;$point$0$i = $314;
           while(1) {
            $315 = ((($point$0$i)) + 28|0);
            $after$0$i = $point$0$i;
            while(1) {
             $316 = ((($after$0$i)) + 4|0);
             $317 = HEAP32[$316>>2]|0;
             $318 = ($317|0)==($314|0);
             if ($318) {
              break L104;
             }
             $319 = ((($317)) + 28|0);
             $320 = HEAP32[$319>>2]|0;
             $321 = HEAP32[$315>>2]|0;
             $322 = ($320|0)==($321|0);
             if ($322) {
              $after$0$i = $317;
             } else {
              $$lcssa = $320;$$lcssa245 = $317;$$lcssa246 = $321;
              break;
             }
            }
            $323 = ((($before$1$i)) + 28|0);
            $324 = HEAP32[$323>>2]|0;
            $325 = ($324|0)<($$lcssa246|0);
            if ($325) {
             $326 = ($$lcssa|0)<($$lcssa246|0);
             if ($326) {
              $point$1$i = $point$0$i;
              label = 80;
             }
            } else {
             $327 = ($$lcssa|0)>($$lcssa246|0);
             if ($327) {
              $point$1$i = $point$0$i;
              label = 80;
             }
            }
            if ((label|0) == 80) {
             while(1) {
              label = 0;
              $328 = ((($point$1$i)) + 16|0);
              $329 = HEAP32[$328>>2]|0;
              $330 = $329 | 64;
              HEAP32[$328>>2] = $330;
              $331 = ((($point$1$i)) + 4|0);
              $332 = HEAP32[$331>>2]|0;
              $333 = ($332|0)==($$lcssa245|0);
              if ($333) {
               break;
              } else {
               $point$1$i = $332;
               label = 80;
              }
             }
            }
            $334 = HEAP32[$$lcssa245>>2]|0;
            $before$1$i = $334;$point$0$i = $$lcssa245;
           }
          }
         } while(0);
         $335 = (($n$018$i) + 1)|0;
         $336 = ($335>>>0)<($298>>>0);
         if ($336) {
          $n$018$i = $335;
         } else {
          $n$1$i = 0;
          break;
         }
        }
       }
      } while(0);
      L124: while(1) {
       if ((label|0) == 95) {
        label = 0;
        $362 = (($n$2$i) + 1)|0;
        $n$1$i = $362;
       }
       $337 = ($n$1$i>>>0)<($glyphrec$idx$val>>>0);
       if (!($337)) {
        break;
       }
       $338 = (($glyphrec$idx1$val) + (($n$1$i*40)|0)|0);
       $339 = (((($glyphrec$idx1$val) + (($n$1$i*40)|0)|0)) + 16|0);
       $340 = HEAP32[$339>>2]|0;
       $341 = $340 & 64;
       $342 = ($341|0)==(0);
       if ($342) {
        $n$2$i = $n$1$i;
        label = 95;
        continue;
       }
       $343 = (((($glyphrec$idx1$val) + (($n$1$i*40)|0)|0)) + 32|0);
       $before2$0$i = $338;
       while(1) {
        $344 = HEAP32[$before2$0$i>>2]|0;
        $345 = ($344|0)==($338|0);
        if ($345) {
         $n$2$i = $n$1$i;
         label = 95;
         continue L124;
        }
        $346 = ((($344)) + 32|0);
        $347 = HEAP32[$346>>2]|0;
        $348 = HEAP32[$343>>2]|0;
        $349 = ($347|0)==($348|0);
        if ($349) {
         $before2$0$i = $344;
        } else {
         $$lcssa251 = $347;$$lcssa252 = $348;
         break;
        }
       }
       $after3$0$i = $338;
       while(1) {
        $350 = ((($after3$0$i)) + 4|0);
        $351 = HEAP32[$350>>2]|0;
        $352 = ($351|0)==($338|0);
        if ($352) {
         $n$2$i = $n$1$i;
         label = 95;
         continue L124;
        }
        $353 = ((($351)) + 32|0);
        $354 = HEAP32[$353>>2]|0;
        $355 = ($354|0)==($$lcssa252|0);
        if ($355) {
         $after3$0$i = $351;
        } else {
         $$lcssa253 = $354;
         break;
        }
       }
       $356 = ($$lcssa251|0)<($$lcssa252|0);
       $357 = ($$lcssa253|0)>($$lcssa252|0);
       $or$cond$i = $356 & $357;
       if ($or$cond$i) {
        $358 = $340 | 128;
        HEAP32[$339>>2] = $358;
        $n$2$i = $n$1$i;
        label = 95;
        continue;
       }
       $359 = ($$lcssa251|0)>($$lcssa252|0);
       $360 = ($$lcssa253|0)<($$lcssa252|0);
       $or$cond40$i = $359 & $360;
       if (!($or$cond40$i)) {
        $n$2$i = $n$1$i;
        label = 95;
        continue;
       }
       $361 = $340 | 256;
       HEAP32[$339>>2] = $361;
       $n$2$i = $n$1$i;
       label = 95;
      }
      $363 = (((($glyphrec)) + 28|0) + (($dimension$088*40)|0)|0);
      $364 = HEAP32[$236>>2]|0;
      $$idx$val = HEAP32[$363>>2]|0;
      $365 = ($$idx$val|0)==(0);
      if ($365) {
       $377 = $364;
      } else {
       $$idx3 = (((((($glyphrec)) + 28|0) + (($dimension$088*40)|0)|0)) + 8|0);
       $$idx3$val = HEAP32[$$idx3>>2]|0;
       $count$02$i = $$idx$val;$hint$01$i = $$idx3$val;
       while(1) {
        _psh_hint_align($hint$01$i,$364,$dimension$088,$glyphrec);
        $366 = (($count$02$i) + -1)|0;
        $367 = ((($hint$01$i)) + 28|0);
        $368 = ($366|0)==(0);
        if ($368) {
         break;
        } else {
         $count$02$i = $366;$hint$01$i = $367;
        }
       }
       $$pre = HEAP32[$236>>2]|0;
       $377 = $$pre;
      }
      $369 = (((((($glyphrec)) + 28|0) + (($dimension$088*40)|0)|0)) + 32|0);
      $370 = HEAP32[$369>>2]|0;
      $371 = ((($370)) + 8|0);
      $372 = HEAP32[$371>>2]|0;
      $373 = HEAP32[$370>>2]|0;
      $374 = ($dimension$088|0)==(0);
      $375 = $374 ? 1 : 2;
      $376 = (((((($377)) + 4|0) + (($dimension$088*204)|0)|0)) + 196|0);
      $378 = HEAP32[$376>>2]|0;
      $379 = (_FT_DivFix(32,$378)|0);
      $380 = ($379|0)>(30);
      $$$i = $380 ? 30 : $379;
      $381 = ($373>>>0)>(1);
      L149: do {
       if ($381) {
        $382 = HEAP32[$glyphrec>>2]|0;
        $383 = ($382|0)==(0);
        if ($383) {
         $808 = 1;$809 = 0;
        } else {
         $384 = ((($372)) + 12|0);
         $385 = HEAP32[$384>>2]|0;
         $386 = ($385>>>0)>($382>>>0);
         $$1$i = $386 ? $382 : $385;
         $$idx7$i = (((((($glyphrec)) + 28|0) + (($dimension$088*40)|0)|0)) + 4|0);
         $$idx8$i = (((((($glyphrec)) + 28|0) + (($dimension$088*40)|0)|0)) + 12|0);
         $$pn$i = $372;$390 = $382;$first$012$i = $$1$i;$num_masks$013$i = $373;
         while(1) {
          $mask$015$i = ((($$pn$i)) + 16|0);
          $387 = ((($$pn$i)) + 28|0);
          $388 = HEAP32[$387>>2]|0;
          $389 = ($388>>>0)<($390>>>0);
          $$2$i = $389 ? $388 : $390;
          $391 = ($$2$i>>>0)>($first$012$i>>>0);
          if ($391) {
           $392 = (($$2$i) - ($first$012$i))|0;
           $393 = HEAP32[$10>>2]|0;
           $394 = (($393) + (($first$012$i*40)|0)|0);
           $mask$0$idx$val$i = HEAP32[$mask$015$i>>2]|0;
           $mask$0$idx4$i = ((($$pn$i)) + 24|0);
           $mask$0$idx4$val$i = HEAP32[$mask$0$idx4$i>>2]|0;
           _psh_hint_table_activate_mask($363,$mask$0$idx$val$i,$mask$0$idx4$val$i);
           $$idx7$val$i = HEAP32[$$idx7$i>>2]|0;
           $$idx8$val$i = HEAP32[$$idx8$i>>2]|0;
           _psh_hint_table_find_strong_points($$idx7$val$i,$$idx8$val$i,$394,$392,$$$i,$375);
          }
          $395 = (($num_masks$013$i) + -1)|0;
          $396 = ($395>>>0)>(1);
          if (!($396)) {
           $$idx5$i$pre$phiZZZ2D = $$idx7$i;$$idx6$i$pre$phiZZZ2D = $$idx8$i;
           label = 109;
           break L149;
          }
          $$pre$i17 = HEAP32[$glyphrec>>2]|0;
          $$pn$i = $mask$015$i;$390 = $$pre$i17;$first$012$i = $$2$i;$num_masks$013$i = $395;
         }
        }
       } else {
        $397 = ($373|0)==(1);
        if ($397) {
         $$pre135 = (((((($glyphrec)) + 28|0) + (($dimension$088*40)|0)|0)) + 4|0);
         $$pre136 = (((((($glyphrec)) + 28|0) + (($dimension$088*40)|0)|0)) + 12|0);
         $$idx5$i$pre$phiZZZ2D = $$pre135;$$idx6$i$pre$phiZZZ2D = $$pre136;
         label = 109;
        } else {
         label = 110;
        }
       }
      } while(0);
      if ((label|0) == 109) {
       label = 0;
       $398 = HEAP32[$glyphrec>>2]|0;
       $399 = HEAP32[$10>>2]|0;
       $400 = HEAP32[$369>>2]|0;
       $401 = ((($400)) + 8|0);
       $402 = HEAP32[$401>>2]|0;
       $$idx$val$i = HEAP32[$402>>2]|0;
       $$idx3$i = ((($402)) + 8|0);
       $$idx3$val$i = HEAP32[$$idx3$i>>2]|0;
       _psh_hint_table_activate_mask($363,$$idx$val$i,$$idx3$val$i);
       $$idx5$val$i = HEAP32[$$idx5$i$pre$phiZZZ2D>>2]|0;
       $$idx6$val$i = HEAP32[$$idx6$i$pre$phiZZZ2D>>2]|0;
       _psh_hint_table_find_strong_points($$idx5$val$i,$$idx6$val$i,$399,$398,$$$i,$375);
       label = 110;
      }
      if ((label|0) == 110) {
       label = 0;
       $$pr = HEAP32[$glyphrec>>2]|0;
       $403 = ($$pr|0)==(0);
       if ($403) {
        $808 = 1;$809 = 0;
       } else {
        $404 = HEAP32[$10>>2]|0;
        $count3$09$i = $$pr;$point4$010$i = $404;
        while(1) {
         $405 = ((($point4$010$i)) + 24|0);
         $406 = HEAP32[$405>>2]|0;
         $407 = ($406|0)==(0|0);
         do {
          if (!($407)) {
           $408 = ((($point4$010$i)) + 16|0);
           $409 = HEAP32[$408>>2]|0;
           $410 = $409 & 16;
           $411 = ($410|0)==(0);
           if (!($411)) {
            break;
           }
           $412 = $409 | 16;
           HEAP32[$408>>2] = $412;
          }
         } while(0);
         $413 = (($count3$09$i) + -1)|0;
         $414 = ((($point4$010$i)) + 40|0);
         $415 = ($413|0)==(0);
         if ($415) {
          $808 = $403;$809 = $$pr;
          break;
         } else {
          $count3$09$i = $413;$point4$010$i = $414;
         }
        }
       }
      }
      $416 = ($dimension$088|0)==(1);
      L172: do {
       if ($416) {
        if ($808) {
         $535 = 0;
        } else {
         $glyphrec$idx5$val = HEAP32[$10>>2]|0;
         $glyph_count$015$i = $809;$point$011$i = $glyphrec$idx5$val;
         while(1) {
          $417 = ((($point$011$i)) + 20|0);
          $418 = HEAP16[$417>>1]|0;
          $419 = $418&255;
          switch ($419<<24>>24) {
          case -2: case 2:  {
           label = 122;
           break;
          }
          default: {
           $420 = ($418&65535) >>> 8;
           $421 = $420&255;
           switch ($421<<24>>24) {
           case -2: case 2:  {
            label = 122;
            break;
           }
           default: {
           }
           }
          }
          }
          L179: do {
           if ((label|0) == 122) {
            label = 0;
            $422 = ((($point$011$i)) + 16|0);
            $423 = HEAP32[$422>>2]|0;
            $424 = $423 & 16;
            $425 = ($424|0)==(0);
            if (!($425)) {
             break;
            }
            $426 = ((($point$011$i)) + 28|0);
            $427 = HEAP32[$426>>2]|0;
            $428 = HEAP32[$268>>2]|0;
            $429 = ($428|0)==(0);
            L182: do {
             if ($429) {
              $810 = $423;
             } else {
              $430 = HEAP32[$271>>2]|0;
              $431 = (0 - ($430))|0;
              $432 = ((($point$011$i)) + 36|0);
              $448 = $423;$blue_count$01$i = $428;$zone$02$i = $269;
              while(1) {
               $433 = ((($zone$02$i)) + 12|0);
               $434 = HEAP32[$433>>2]|0;
               $435 = (($427) - ($434))|0;
               $436 = ($435|0)<($431|0);
               if ($436) {
                $810 = $448;
                break L182;
               }
               $437 = ((($zone$02$i)) + 8|0);
               $438 = HEAP32[$437>>2]|0;
               $439 = (($438) + ($430))|0;
               $440 = ($427|0)>($439|0);
               do {
                if ($440) {
                 $811 = $448;
                } else {
                 $441 = HEAP8[$272>>0]|0;
                 $442 = ($441<<24>>24)==(0);
                 if ($442) {
                  $443 = HEAP32[$273>>2]|0;
                  $444 = ($435|0)>($443|0);
                  if ($444) {
                   $811 = $448;
                   break;
                  }
                 }
                 $445 = ((($zone$02$i)) + 24|0);
                 $446 = HEAP32[$445>>2]|0;
                 HEAP32[$432>>2] = $446;
                 $447 = $448 | 48;
                 HEAP32[$422>>2] = $447;
                 $811 = $447;
                }
               } while(0);
               $449 = (($blue_count$01$i) + -1)|0;
               $450 = ((($zone$02$i)) + 32|0);
               $451 = ($449|0)==(0);
               if ($451) {
                $810 = $811;
                break;
               } else {
                $448 = $811;$blue_count$01$i = $449;$zone$02$i = $450;
               }
              }
             }
            } while(0);
            $452 = HEAP32[$270>>2]|0;
            $453 = ($452|0)==(0);
            if ($453) {
             break;
            }
            $454 = (((($globals)) + 932|0) + ($452<<5)|0);
            $455 = HEAP32[$271>>2]|0;
            $456 = (0 - ($455))|0;
            $457 = ((($point$011$i)) + 36|0);
            $$pn$i23 = $454;$473 = $810;$blue_count$16$i = $452;
            while(1) {
             $zone$18$i = ((($$pn$i23)) + -32|0);
             $458 = ((($$pn$i23)) + -24|0);
             $459 = HEAP32[$458>>2]|0;
             $460 = (($459) - ($427))|0;
             $461 = ($460|0)<($456|0);
             if ($461) {
              break L179;
             }
             $462 = ((($$pn$i23)) + -20|0);
             $463 = HEAP32[$462>>2]|0;
             $464 = (($463) - ($455))|0;
             $465 = ($427|0)<($464|0);
             do {
              if ($465) {
               $812 = $473;
              } else {
               $466 = HEAP8[$272>>0]|0;
               $467 = ($466<<24>>24)==(0);
               if ($467) {
                $468 = HEAP32[$273>>2]|0;
                $469 = ($460|0)<($468|0);
                if (!($469)) {
                 $812 = $473;
                 break;
                }
               }
               $470 = ((($$pn$i23)) + -4|0);
               $471 = HEAP32[$470>>2]|0;
               HEAP32[$457>>2] = $471;
               $472 = $473 | 48;
               HEAP32[$422>>2] = $472;
               $812 = $472;
              }
             } while(0);
             $474 = (($blue_count$16$i) + -1)|0;
             $475 = ($474|0)==(0);
             if ($475) {
              break;
             } else {
              $$pn$i23 = $zone$18$i;$473 = $812;$blue_count$16$i = $474;
             }
            }
           }
          } while(0);
          $476 = (($glyph_count$015$i) + -1)|0;
          $477 = ((($point$011$i)) + 40|0);
          $478 = ($476|0)==(0);
          if ($478) {
           $glyphrec$idx7$val = $glyphrec$idx5$val;
           label = 140;
           break L172;
          } else {
           $glyph_count$015$i = $476;$point$011$i = $477;
          }
         }
        }
       } else {
        $glyphrec$idx7$val$pre = HEAP32[$10>>2]|0;
        $glyphrec$idx7$val = $glyphrec$idx7$val$pre;
        label = 140;
       }
      } while(0);
      if ((label|0) == 140) {
       label = 0;
       $glyphrec$idx8$val = HEAP32[$236>>2]|0;
       $479 = (((((($glyphrec$idx8$val)) + 4|0) + (($dimension$088*204)|0)|0)) + 196|0);
       $480 = HEAP32[$479>>2]|0;
       if ($808) {
        $535 = 0;
       } else {
        $count$02$i25 = $809;$point$01$i = $glyphrec$idx7$val;
        while(1) {
         $481 = ((($point$01$i)) + 24|0);
         $482 = HEAP32[$481>>2]|0;
         $483 = ($482|0)==(0|0);
         if (!($483)) {
          $484 = ((($point$01$i)) + 16|0);
          $485 = HEAP32[$484>>2]|0;
          $486 = $485 & 512;
          $487 = ($486|0)==(0);
          do {
           if ($487) {
            $491 = $485 & 1024;
            $492 = ($491|0)==(0);
            if (!($492)) {
             $493 = ((($482)) + 8|0);
             $494 = HEAP32[$493>>2]|0;
             $495 = ((($482)) + 12|0);
             $496 = HEAP32[$495>>2]|0;
             $497 = (($496) + ($494))|0;
             $498 = ((($point$01$i)) + 36|0);
             HEAP32[$498>>2] = $497;
             break;
            }
            $499 = ((($point$01$i)) + 28|0);
            $500 = HEAP32[$499>>2]|0;
            $501 = HEAP32[$482>>2]|0;
            $502 = (($500) - ($501))|0;
            $503 = ($502|0)<(1);
            if ($503) {
             $504 = ((($482)) + 8|0);
             $505 = HEAP32[$504>>2]|0;
             $506 = (_FT_MulFix($502,$480)|0);
             $507 = (($506) + ($505))|0;
             $508 = ((($point$01$i)) + 36|0);
             HEAP32[$508>>2] = $507;
             break;
            }
            $509 = ((($482)) + 4|0);
            $510 = HEAP32[$509>>2]|0;
            $511 = ($502|0)<($510|0);
            $512 = ((($482)) + 8|0);
            $513 = HEAP32[$512>>2]|0;
            $514 = ((($482)) + 12|0);
            $515 = HEAP32[$514>>2]|0;
            if ($511) {
             $521 = (_FT_MulDiv($502,$515,$510)|0);
             $522 = (($521) + ($513))|0;
             $523 = ((($point$01$i)) + 36|0);
             HEAP32[$523>>2] = $522;
             break;
            } else {
             $516 = (($515) + ($513))|0;
             $517 = (($502) - ($510))|0;
             $518 = (_FT_MulFix($517,$480)|0);
             $519 = (($516) + ($518))|0;
             $520 = ((($point$01$i)) + 36|0);
             HEAP32[$520>>2] = $519;
             break;
            }
           } else {
            $488 = ((($482)) + 8|0);
            $489 = HEAP32[$488>>2]|0;
            $490 = ((($point$01$i)) + 36|0);
            HEAP32[$490>>2] = $489;
           }
          } while(0);
          $524 = HEAP32[$484>>2]|0;
          $525 = $524 | 32;
          HEAP32[$484>>2] = $525;
         }
         $526 = (($count$02$i25) + -1)|0;
         $527 = ((($point$01$i)) + 40|0);
         $528 = ($526|0)==(0);
         if ($528) {
          break;
         } else {
          $count$02$i25 = $526;$point$01$i = $527;
         }
        }
        $$pre125 = HEAP32[$glyphrec>>2]|0;
        $535 = $$pre125;
       }
      }
      $529 = HEAP32[$236>>2]|0;
      $530 = (((((($529)) + 4|0) + (($dimension$088*204)|0)|0)) + 196|0);
      $531 = HEAP32[$530>>2]|0;
      $532 = HEAP32[$7>>2]|0;
      $533 = HEAP32[$10>>2]|0;
      $534 = (($533) + (($535*40)|0)|0);
      $536 = ($535|0)>(0);
      do {
       if ($536) {
        $num_strongs$023$i = 0;$point$024$i29 = $533;
        while(1) {
         $537 = ((($point$024$i29)) + 16|0);
         $538 = HEAP32[$537>>2]|0;
         $539 = $538 >>> 4;
         $$lobit$i = $539 & 1;
         $num_strongs$0$$i = (($$lobit$i) + ($num_strongs$023$i))|0;
         $540 = ((($point$024$i29)) + 40|0);
         $541 = ($540>>>0)<($534>>>0);
         if ($541) {
          $num_strongs$023$i = $num_strongs$0$$i;$point$024$i29 = $540;
         } else {
          $num_strongs$0$$i$lcssa = $num_strongs$0$$i;
          break;
         }
        }
        $542 = ($num_strongs$0$$i$lcssa|0)==(0);
        if (!($542)) {
         $543 = ($num_strongs$0$$i$lcssa>>>0)<(17);
         if ($543) {
          $strongs$0$i = $strongs_0$i;
         } else {
          $544 = (_ft_mem_realloc($532,4,0,$num_strongs$0$$i$lcssa,0,$error$i28)|0);
          $545 = HEAP32[$error$i28>>2]|0;
          $546 = ($545|0)==(0);
          if ($546) {
           $strongs$0$i = $544;
          } else {
           break;
          }
         }
         $num_strongs$218$i = 0;$point$119$i = $533;
         while(1) {
          $550 = ((($point$119$i)) + 16|0);
          $551 = HEAP32[$550>>2]|0;
          $552 = $551 & 16;
          $553 = ($552|0)==(0);
          if ($553) {
           $num_strongs$3$i = $num_strongs$218$i;
          } else {
           $554 = (($strongs$0$i) + ($num_strongs$218$i<<2)|0);
           $555 = ($num_strongs$218$i|0)>(0);
           L240: do {
            if ($555) {
             $556 = ((($point$119$i)) + 28|0);
             $557 = HEAP32[$556>>2]|0;
             $insert$011$i = $554;
             while(1) {
              $558 = ((($insert$011$i)) + -4|0);
              $559 = HEAP32[$558>>2]|0;
              $560 = ((($559)) + 28|0);
              $561 = HEAP32[$560>>2]|0;
              $562 = ($561|0)>($557|0);
              if (!($562)) {
               $insert$0$lcssa$i = $insert$011$i;
               break L240;
              }
              HEAP32[$insert$011$i>>2] = $559;
              $563 = ($558>>>0)>($strongs$0$i>>>0);
              if ($563) {
               $insert$011$i = $558;
              } else {
               $insert$0$lcssa$i = $558;
               break;
              }
             }
            } else {
             $insert$0$lcssa$i = $554;
            }
           } while(0);
           HEAP32[$insert$0$lcssa$i>>2] = $point$119$i;
           $564 = (($num_strongs$218$i) + 1)|0;
           $num_strongs$3$i = $564;
          }
          $565 = ((($point$119$i)) + 40|0);
          $566 = ($565>>>0)<($534>>>0);
          if ($566) {
           $num_strongs$218$i = $num_strongs$3$i;$point$119$i = $565;
          } else {
           $num_strongs$3$i$lcssa = $num_strongs$3$i;
           break;
          }
         }
         $547 = ($num_strongs$3$i$lcssa|0)==(0);
         $548 = (($num_strongs$3$i$lcssa) + -1)|0;
         $549 = (($strongs$0$i) + ($548<<2)|0);
         $point$26$i = $533;
         while(1) {
          $567 = ((($point$26$i)) + 16|0);
          $568 = HEAP32[$567>>2]|0;
          $569 = $568 & 16;
          $570 = ($569|0)==(0);
          do {
           if ($570) {
            $571 = ((($point$26$i)) + 12|0);
            $572 = HEAP32[$571>>2]|0;
            $573 = $572 & 2;
            $574 = ($573|0)==(0);
            if (!($574)) {
             $575 = ((($point$26$i)) + 20|0);
             $576 = HEAP16[$575>>1]|0;
             $577 = $576&255;
             $578 = ($577<<24>>24)==(4);
             if ($578) {
              break;
             }
             $579 = ($576&65535) >>> 8;
             $580 = $579&255;
             $581 = ($577<<24>>24)==($580<<24>>24);
             if (!($581)) {
              break;
             }
             $582 = $568 & 64;
             $583 = $572 & 4;
             $584 = $583 | $582;
             $585 = ($584|0)==(0);
             if ($585) {
              break;
             }
             $586 = $572 & -3;
             HEAP32[$571>>2] = $586;
            }
            $$phi$trans$insert = ((($point$26$i)) + 28|0);
            $$pre126 = HEAP32[$$phi$trans$insert>>2]|0;
            do {
             if ($547) {
              label = 178;
             } else {
              $nn$03$i = 0;
              while(1) {
               $587 = (($strongs$0$i) + ($nn$03$i<<2)|0);
               $588 = HEAP32[$587>>2]|0;
               $589 = ((($588)) + 28|0);
               $590 = HEAP32[$589>>2]|0;
               $591 = ($590|0)>($$pre126|0);
               if ($591) {
                $nn$03$i$lcssa = $nn$03$i;
                label = 177;
                break;
               }
               $592 = (($nn$03$i) + 1)|0;
               $593 = ($592>>>0)<($num_strongs$3$i$lcssa>>>0);
               if ($593) {
                $nn$03$i = $592;
               } else {
                $nn$0$lcssa30$i = $592;
                break;
               }
              }
              if ((label|0) == 177) {
               label = 0;
               $594 = ($nn$03$i$lcssa|0)==(0);
               if ($594) {
                label = 178;
                break;
               } else {
                $nn$0$lcssa30$i = $nn$03$i$lcssa;
               }
              }
              $604 = (($nn$0$lcssa30$i) + -1)|0;
              $605 = (($strongs$0$i) + ($604<<2)|0);
              $606 = HEAP32[$605>>2]|0;
              $nn$1$i = $num_strongs$3$i$lcssa;
              while(1) {
               $607 = ($nn$1$i|0)==(0);
               if ($607) {
                $nn$1$lcssa$i = 0;
                break;
               }
               $608 = (($nn$1$i) + -1)|0;
               $609 = (($strongs$0$i) + ($608<<2)|0);
               $610 = HEAP32[$609>>2]|0;
               $611 = ((($610)) + 28|0);
               $612 = HEAP32[$611>>2]|0;
               $613 = ($612|0)<($$pre126|0);
               if ($613) {
                $nn$1$lcssa$i = $nn$1$i;
                break;
               } else {
                $nn$1$i = $608;
               }
              }
              $614 = ($nn$1$lcssa$i|0)==($num_strongs$3$i$lcssa|0);
              if ($614) {
               $615 = HEAP32[$549>>2]|0;
               $616 = ((($615)) + 36|0);
               $617 = HEAP32[$616>>2]|0;
               $618 = ((($615)) + 28|0);
               $619 = HEAP32[$618>>2]|0;
               $620 = (($$pre126) - ($619))|0;
               $621 = (_FT_MulFix($620,$531)|0);
               $622 = (($621) + ($617))|0;
               $623 = ((($point$26$i)) + 36|0);
               HEAP32[$623>>2] = $622;
               break;
              }
              $624 = (($strongs$0$i) + ($nn$1$lcssa$i<<2)|0);
              $625 = HEAP32[$624>>2]|0;
              $626 = ((($606)) + 28|0);
              $627 = HEAP32[$626>>2]|0;
              $628 = ($$pre126|0)==($627|0);
              if ($628) {
               $629 = ((($606)) + 36|0);
               $630 = HEAP32[$629>>2]|0;
               $631 = ((($point$26$i)) + 36|0);
               HEAP32[$631>>2] = $630;
               break;
              }
              $632 = ((($625)) + 28|0);
              $633 = HEAP32[$632>>2]|0;
              $634 = ($$pre126|0)==($633|0);
              $635 = ((($625)) + 36|0);
              $636 = HEAP32[$635>>2]|0;
              if ($634) {
               $637 = ((($point$26$i)) + 36|0);
               HEAP32[$637>>2] = $636;
               break;
              } else {
               $638 = ((($606)) + 36|0);
               $639 = HEAP32[$638>>2]|0;
               $640 = (($$pre126) - ($627))|0;
               $641 = (($636) - ($639))|0;
               $642 = (($633) - ($627))|0;
               $643 = (_FT_MulDiv($640,$641,$642)|0);
               $644 = (($643) + ($639))|0;
               $645 = ((($point$26$i)) + 36|0);
               HEAP32[$645>>2] = $644;
               break;
              }
             }
            } while(0);
            if ((label|0) == 178) {
             label = 0;
             $595 = HEAP32[$strongs$0$i>>2]|0;
             $596 = ((($595)) + 36|0);
             $597 = HEAP32[$596>>2]|0;
             $598 = ((($595)) + 28|0);
             $599 = HEAP32[$598>>2]|0;
             $600 = (($$pre126) - ($599))|0;
             $601 = (_FT_MulFix($600,$531)|0);
             $602 = (($601) + ($597))|0;
             $603 = ((($point$26$i)) + 36|0);
             HEAP32[$603>>2] = $602;
            }
            $646 = HEAP32[$567>>2]|0;
            $647 = $646 | 32;
            HEAP32[$567>>2] = $647;
           }
          } while(0);
          $648 = ((($point$26$i)) + 40|0);
          $649 = ($648>>>0)<($534>>>0);
          if ($649) {
           $point$26$i = $648;
          } else {
           break;
          }
         }
         $650 = ($strongs$0$i|0)==($strongs_0$i|0);
         if ($650) {
          break;
         }
         _ft_mem_free($532,$strongs$0$i);
        }
       }
      } while(0);
      $glyphrec$idx9$val = HEAP32[$23>>2]|0;
      $glyphrec$idx11$val = HEAP32[$236>>2]|0;
      $651 = (((((($glyphrec$idx11$val)) + 4|0) + (($dimension$088*204)|0)|0)) + 196|0);
      $652 = HEAP32[$651>>2]|0;
      $653 = ($glyphrec$idx9$val|0)==(0);
      if (!($653)) {
       $glyphrec$idx10$val = HEAP32[$16>>2]|0;
       $654 = (((((($glyphrec$idx11$val)) + 4|0) + (($dimension$088*204)|0)|0)) + 200|0);
       $655 = HEAP32[$654>>2]|0;
       $contour$014$i = $glyphrec$idx10$val;$delta$015$i = $655;$num_contours$013$i = $glyphrec$idx9$val;
       while(1) {
        $656 = HEAP32[$contour$014$i>>2]|0;
        $657 = ((($contour$014$i)) + 4|0);
        $658 = HEAP32[$657>>2]|0;
        $659 = (($656) + (($658*40)|0)|0);
        $660 = ($658|0)>(0);
        L289: do {
         if ($660) {
          $first$03$i = 0;$fit_count$05$i = 0;$point$04$i = $656;
          while(1) {
           $661 = ((($point$04$i)) + 16|0);
           $662 = HEAP32[$661>>2]|0;
           $663 = $662 & 32;
           $664 = ($663|0)==(0);
           if ($664) {
            $first$2$i = $first$03$i;$fit_count$1$i = $fit_count$05$i;
           } else {
            $665 = ($first$03$i|0)==(0|0);
            $point$0$first$0$i = $665 ? $point$04$i : $first$03$i;
            $666 = (($fit_count$05$i) + 1)|0;
            $first$2$i = $point$0$first$0$i;$fit_count$1$i = $666;
           }
           $667 = ((($point$04$i)) + 40|0);
           $668 = ($667>>>0)<($659>>>0);
           if ($668) {
            $first$03$i = $first$2$i;$fit_count$05$i = $fit_count$1$i;$point$04$i = $667;
           } else {
            $first$2$i$lcssa = $first$2$i;$fit_count$1$i$lcssa = $fit_count$1$i;
            break;
           }
          }
          $669 = ($fit_count$1$i$lcssa>>>0)<(2);
          if ($669) {
           $673 = ($fit_count$1$i$lcssa|0)==(1);
           if ($673) {
            $674 = ((($first$2$i$lcssa)) + 36|0);
            $675 = HEAP32[$674>>2]|0;
            $676 = ((($first$2$i$lcssa)) + 28|0);
            $677 = HEAP32[$676>>2]|0;
            $678 = (_FT_MulFix($677,$652)|0);
            $679 = (($675) - ($678))|0;
            $delta$1$i = $679;
           } else {
            $delta$1$i = $delta$015$i;
           }
           $point$110$i = $656;
           while(1) {
            $680 = ($point$110$i|0)==($first$2$i$lcssa|0);
            if (!($680)) {
             $681 = ((($point$110$i)) + 28|0);
             $682 = HEAP32[$681>>2]|0;
             $683 = (_FT_MulFix($682,$652)|0);
             $684 = (($683) + ($delta$1$i))|0;
             $685 = ((($point$110$i)) + 36|0);
             HEAP32[$685>>2] = $684;
            }
            $686 = ((($point$110$i)) + 40|0);
            $687 = ($686>>>0)<($659>>>0);
            if ($687) {
             $point$110$i = $686;
            } else {
             $delta$2$i = $delta$1$i;
             break L289;
            }
           }
          }
          $670 = ((($first$2$i$lcssa)) + 4|0);
          $671 = HEAP32[$670>>2]|0;
          $672 = ($671|0)==($first$2$i$lcssa|0);
          if ($672) {
           $delta$2$i = $delta$015$i;
           break;
          } else {
           $689 = $671;$714 = $670;$first$37$i = $first$2$i$lcssa;
          }
          while(1) {
           $688 = ((($689)) + 16|0);
           $690 = HEAP32[$688>>2]|0;
           $691 = $690 & 32;
           $692 = ($691|0)==(0);
           if ($692) {
            $next$0$i = $689;
            while(1) {
             $693 = ((($next$0$i)) + 4|0);
             $694 = HEAP32[$693>>2]|0;
             $695 = ((($694)) + 16|0);
             $696 = HEAP32[$695>>2]|0;
             $697 = $696 & 32;
             $698 = ($697|0)==(0);
             if ($698) {
              $next$0$i = $694;
             } else {
              $$lcssa250 = $694;
              break;
             }
            }
            $699 = ((($first$37$i)) + 28|0);
            $700 = HEAP32[$699>>2]|0;
            $701 = ((($$lcssa250)) + 28|0);
            $702 = HEAP32[$701>>2]|0;
            $703 = ($700|0)>($702|0);
            $704 = ((($$lcssa250)) + 36|0);
            $705 = HEAP32[$704>>2]|0;
            $706 = ((($first$37$i)) + 36|0);
            $707 = HEAP32[$706>>2]|0;
            if ($703) {
             $710 = (($700) - ($702))|0;
             $711 = (($707) - ($705))|0;
             $cur_a$0$i = $705;$cur_ab$0$i = $711;$org_a$0$i = $702;$org_ab$0$i = $710;
            } else {
             $708 = (($702) - ($700))|0;
             $709 = (($705) - ($707))|0;
             $cur_a$0$i = $707;$cur_ab$0$i = $709;$org_a$0$i = $700;$org_ab$0$i = $708;
            }
            $712 = ($org_ab$0$i|0)>(0);
            if ($712) {
             $713 = (_FT_DivFix($cur_ab$0$i,$org_ab$0$i)|0);
             $$pre$i40 = HEAP32[$714>>2]|0;
             $813 = $$pre$i40;$scale_ab$0$i = $713;
            } else {
             $813 = $689;$scale_ab$0$i = 65536;
            }
            $715 = (($cur_ab$0$i) + ($cur_a$0$i))|0;
            $point$2$i = $813;
            while(1) {
             $716 = ((($point$2$i)) + 28|0);
             $717 = HEAP32[$716>>2]|0;
             $718 = (($717) - ($org_a$0$i))|0;
             $719 = ($718|0)<(1);
             do {
              if ($719) {
               $720 = (_FT_MulFix($718,$652)|0);
               $721 = (($720) + ($cur_a$0$i))|0;
               $cur_c$0$i = $721;
              } else {
               $722 = ($718|0)<($org_ab$0$i|0);
               if ($722) {
                $726 = (_FT_MulFix($718,$scale_ab$0$i)|0);
                $727 = (($726) + ($cur_a$0$i))|0;
                $cur_c$0$i = $727;
                break;
               } else {
                $723 = (($718) - ($org_ab$0$i))|0;
                $724 = (_FT_MulFix($723,$652)|0);
                $725 = (($715) + ($724))|0;
                $cur_c$0$i = $725;
                break;
               }
              }
             } while(0);
             $728 = ((($point$2$i)) + 36|0);
             HEAP32[$728>>2] = $cur_c$0$i;
             $729 = ((($point$2$i)) + 4|0);
             $730 = HEAP32[$729>>2]|0;
             $731 = ($730|0)==($$lcssa250|0);
             if ($731) {
              break;
             } else {
              $point$2$i = $730;
             }
            }
            $732 = ($$lcssa250|0)==($first$2$i$lcssa|0);
            if ($732) {
             $delta$2$i = $delta$015$i;
             break L289;
            } else {
             $first$3$be$i = $$lcssa250;
            }
           } else {
            $first$3$be$i = $689;
           }
           $733 = ((($first$3$be$i)) + 4|0);
           $734 = HEAP32[$733>>2]|0;
           $735 = ($734|0)==($first$2$i$lcssa|0);
           if ($735) {
            $delta$2$i = $delta$015$i;
            break;
           } else {
            $689 = $734;$714 = $733;$first$37$i = $first$3$be$i;
           }
          }
         } else {
          $delta$2$i = $delta$015$i;
         }
        } while(0);
        $736 = (($num_contours$013$i) + -1)|0;
        $737 = ((($contour$014$i)) + 8|0);
        $738 = ($736|0)==(0);
        if ($738) {
         break;
        } else {
         $contour$014$i = $737;$delta$015$i = $delta$2$i;$num_contours$013$i = $736;
        }
       }
      }
      $739 = HEAP32[$10>>2]|0;
      $740 = HEAP32[$glyphrec$idx2>>2]|0;
      $741 = ((($740)) + 4|0);
      $742 = HEAP32[$741>>2]|0;
      $743 = ((($740)) + 8|0);
      $744 = HEAP32[$743>>2]|0;
      $745 = HEAP32[$glyphrec>>2]|0;
      $746 = ($745|0)==(0);
      if (!($746)) {
       $747 = $374 ? 32 : 64;
       if ($374) {
        $814 = $745;$n$02$us$i = 0;$point$01$us$i = $739;
        while(1) {
         $748 = ((($point$01$us$i)) + 36|0);
         $749 = HEAP32[$748>>2]|0;
         $750 = (($742) + ($n$02$us$i<<3)|0);
         HEAP32[$750>>2] = $749;
         $751 = ((($point$01$us$i)) + 16|0);
         $752 = HEAP32[$751>>2]|0;
         $753 = $752 & 16;
         $754 = ($753|0)==(0);
         if ($754) {
          $763 = $814;
         } else {
          $755 = (($744) + ($n$02$us$i)|0);
          $756 = HEAP8[$755>>0]|0;
          $757 = $756&255;
          $758 = $757 | $747;
          $759 = $758&255;
          HEAP8[$755>>0] = $759;
          $$pre4$i = HEAP32[$glyphrec>>2]|0;
          $763 = $$pre4$i;
         }
         $760 = ((($point$01$us$i)) + 40|0);
         $761 = (($n$02$us$i) + 1)|0;
         $762 = ($761>>>0)<($763>>>0);
         if ($762) {
          $814 = $763;$n$02$us$i = $761;$point$01$us$i = $760;
         } else {
          break;
         }
        }
       } else {
        $815 = $745;$n$02$i = 0;$point$01$i45 = $739;
        while(1) {
         $764 = ((($point$01$i45)) + 36|0);
         $765 = HEAP32[$764>>2]|0;
         $766 = (((($742) + ($n$02$i<<3)|0)) + 4|0);
         HEAP32[$766>>2] = $765;
         $767 = ((($point$01$i45)) + 16|0);
         $768 = HEAP32[$767>>2]|0;
         $769 = $768 & 16;
         $770 = ($769|0)==(0);
         if ($770) {
          $779 = $815;
         } else {
          $771 = (($744) + ($n$02$i)|0);
          $772 = HEAP8[$771>>0]|0;
          $773 = $772&255;
          $774 = $773 | $747;
          $775 = $774&255;
          HEAP8[$771>>0] = $775;
          $$pre$i47 = HEAP32[$glyphrec>>2]|0;
          $779 = $$pre$i47;
         }
         $776 = ((($point$01$i45)) + 40|0);
         $777 = (($n$02$i) + 1)|0;
         $778 = ($777>>>0)<($779>>>0);
         if ($778) {
          $815 = $779;$n$02$i = $777;$point$01$i45 = $776;
         } else {
          break;
         }
        }
       }
      }
      if (!($rescale$0)) {
       $780 = HEAP32[$236>>2]|0;
       _psh_globals_set_scale($780,$239,$241,0,0);
      }
      $781 = (($dimension$088) + 1)|0;
      $exitcond = ($781|0)==(2);
      if ($exitcond) {
       $$pre$phi129Z2D = $16;$$pre$phi131Z2D = $glyphrec;$$pre$phi133Z2D = $23;$$pre$phiZ2D = $16;$807 = 0;
       break;
      } else {
       $dimension$088 = $781;
      }
     }
    } else {
     $$pre$phi129Z2D = $16;$$pre$phi131Z2D = $glyphrec;$$pre$phi133Z2D = $23;$$pre$phiZ2D = $16;$807 = $234;
    }
   } else {
    $$ph = $229;$$pre127$pre$phiZZ2D = $16;$$pre128$pre$phiZZ2D = $16;
    label = 56;
   }
  } else {
   $$ph = $17;$$pre127$pre$phiZZ2D = $16;$$pre128$pre$phiZZ2D = $16;
   label = 56;
  }
 } else {
  $$pre137 = ((($glyphrec)) + 12|0);
  $$ph = $11;$$pre127$pre$phiZZ2D = $$pre137;$$pre128$pre$phiZZ2D = $$pre137;
  label = 56;
 }
 if ((label|0) == 56) {
  $$pre132 = ((($glyphrec)) + 4|0);
  $$pre$phi129Z2D = $$pre128$pre$phiZZ2D;$$pre$phi131Z2D = $glyphrec;$$pre$phi133Z2D = $$pre132;$$pre$phiZ2D = $$pre127$pre$phiZZ2D;$807 = $$ph;
 }
 $782 = HEAP32[$7>>2]|0;
 $783 = ((($glyphrec)) + 92|0);
 $784 = HEAP32[$783>>2]|0;
 _ft_mem_free($782,$784);
 HEAP32[$783>>2] = 0;
 $785 = ((($glyphrec)) + 88|0);
 HEAP32[$785>>2] = 0;
 $786 = ((($glyphrec)) + 96|0);
 HEAP32[$786>>2] = 0;
 $787 = ((($glyphrec)) + 80|0);
 $788 = HEAP32[$787>>2]|0;
 _ft_mem_free($782,$788);
 HEAP32[$787>>2] = 0;
 $789 = ((($glyphrec)) + 76|0);
 $790 = HEAP32[$789>>2]|0;
 _ft_mem_free($782,$790);
 HEAP32[$789>>2] = 0;
 $791 = ((($glyphrec)) + 72|0);
 HEAP32[$791>>2] = 0;
 $792 = ((($glyphrec)) + 68|0);
 HEAP32[$792>>2] = 0;
 $793 = ((($glyphrec)) + 84|0);
 HEAP32[$793>>2] = 0;
 $794 = ((($glyphrec)) + 52|0);
 $795 = HEAP32[$794>>2]|0;
 _ft_mem_free($782,$795);
 HEAP32[$794>>2] = 0;
 $796 = ((($glyphrec)) + 48|0);
 HEAP32[$796>>2] = 0;
 $797 = ((($glyphrec)) + 56|0);
 HEAP32[$797>>2] = 0;
 $798 = ((($glyphrec)) + 40|0);
 $799 = HEAP32[$798>>2]|0;
 _ft_mem_free($782,$799);
 HEAP32[$798>>2] = 0;
 $800 = ((($glyphrec)) + 36|0);
 $801 = HEAP32[$800>>2]|0;
 _ft_mem_free($782,$801);
 HEAP32[$800>>2] = 0;
 $802 = ((($glyphrec)) + 32|0);
 HEAP32[$802>>2] = 0;
 $803 = ((($glyphrec)) + 28|0);
 HEAP32[$803>>2] = 0;
 $804 = ((($glyphrec)) + 44|0);
 HEAP32[$804>>2] = 0;
 $805 = HEAP32[$10>>2]|0;
 _ft_mem_free($782,$805);
 HEAP32[$10>>2] = 0;
 $806 = HEAP32[$$pre$phi129Z2D>>2]|0;
 _ft_mem_free($782,$806);
 HEAP32[$$pre$phiZ2D>>2] = 0;
 HEAP32[$$pre$phi131Z2D>>2] = 0;
 HEAP32[$$pre$phi133Z2D>>2] = 0;
 HEAP32[$7>>2] = 0;
 $$0 = $807;
 STACKTOP = sp;return ($$0|0);
}
function _psh_hint_table_init($table,$hints,$hint_masks,$memory) {
 $table = $table|0;
 $hints = $hints|0;
 $hint_masks = $hint_masks|0;
 $memory = $memory|0;
 var $$idx$val$i = 0, $$idx$val$i$i = 0, $$idx1$i = 0, $$idx1$i$i = 0, $$idx1$val$i = 0, $$idx1$val$i$i = 0, $$lcssa = 0, $$lcssa34 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $count$011 = 0, $count$03$i = 0, $count$03$i$i = 0, $count$18 = 0, $cursor$03$i = 0, $cursor$1$i = 0, $error = 0, $exitcond = 0, $exitcond$i = 0, $idx$04$i = 0, $idx$07 = 0, $mask$0$idx$val = 0, $mask$0$idx1 = 0, $mask$0$idx1$val = 0, $mask$05$i = 0, $mask$09 = 0, $mask$1$i = 0;
 var $or$cond = 0, $read$013 = 0, $sorted$02$i = 0, $sorted$02$i$i = 0, $val$02$i = 0, $val$1$i = 0, $write$012 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = HEAP32[$hints>>2]|0;
 $1 = $0 << 1;
 $2 = (_ft_mem_realloc($memory,4,0,$1,0,$error)|0);
 $3 = ((($table)) + 12|0);
 HEAP32[$3>>2] = $2;
 $4 = HEAP32[$error>>2]|0;
 $5 = ($4|0)==(0);
 if (!($5)) {
  $110 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($110|0);
 }
 $6 = (_ft_mem_realloc($memory,28,0,$0,0,$error)|0);
 $7 = ((($table)) + 8|0);
 HEAP32[$7>>2] = $6;
 $8 = HEAP32[$error>>2]|0;
 $9 = ($8|0)==(0);
 if (!($9)) {
  $110 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($110|0);
 }
 $10 = $1 | 1;
 $11 = (_ft_mem_realloc($memory,16,0,$10,0,$error)|0);
 $12 = ((($table)) + 24|0);
 HEAP32[$12>>2] = $11;
 $13 = HEAP32[$error>>2]|0;
 $14 = ($13|0)==(0);
 if (!($14)) {
  $110 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($110|0);
 }
 HEAP32[$table>>2] = $0;
 $15 = HEAP32[$3>>2]|0;
 $16 = (($15) + ($0<<2)|0);
 $17 = ((($table)) + 16|0);
 HEAP32[$17>>2] = $16;
 $18 = ((($table)) + 4|0);
 HEAP32[$18>>2] = 0;
 $19 = ((($table)) + 20|0);
 HEAP32[$19>>2] = 0;
 $20 = ((($table)) + 28|0);
 HEAP32[$20>>2] = 0;
 $21 = ($0|0)==(0);
 if (!($21)) {
  $22 = ((($hints)) + 8|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = HEAP32[$7>>2]|0;
  $count$011 = $0;$read$013 = $23;$write$012 = $24;
  while(1) {
   $25 = HEAP32[$read$013>>2]|0;
   HEAP32[$write$012>>2] = $25;
   $26 = ((($read$013)) + 4|0);
   $27 = HEAP32[$26>>2]|0;
   $28 = ((($write$012)) + 4|0);
   HEAP32[$28>>2] = $27;
   $29 = ((($read$013)) + 8|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = ((($write$012)) + 16|0);
   HEAP32[$31>>2] = $30;
   $32 = (($count$011) + -1)|0;
   $33 = ((($write$012)) + 28|0);
   $34 = ((($read$013)) + 12|0);
   $35 = ($32|0)==(0);
   if ($35) {
    break;
   } else {
    $count$011 = $32;$read$013 = $34;$write$012 = $33;
   }
  }
 }
 $36 = ($hint_masks|0)==(0|0);
 if ($36) {
  $81 = 0;
 } else {
  $37 = ((($hint_masks)) + 8|0);
  $38 = HEAP32[$37>>2]|0;
  $39 = HEAP32[$hint_masks>>2]|0;
  $40 = ((($table)) + 32|0);
  HEAP32[$40>>2] = $hint_masks;
  $41 = ($39|0)==(0);
  if ($41) {
   $81 = 0;
  } else {
   $111 = 0;$count$18 = $39;$mask$09 = $38;
   while(1) {
    $mask$0$idx$val = HEAP32[$mask$09>>2]|0;
    $42 = ($mask$0$idx$val|0)==(0);
    if ($42) {
     $112 = $111;
    } else {
     $mask$0$idx1 = ((($mask$09)) + 8|0);
     $mask$0$idx1$val = HEAP32[$mask$0$idx1>>2]|0;
     $60 = $111;$cursor$03$i = $mask$0$idx1$val;$idx$04$i = 0;$mask$05$i = 0;$val$02$i = 0;
     while(1) {
      $43 = ($mask$05$i|0)==(0);
      if ($43) {
       $44 = ((($cursor$03$i)) + 1|0);
       $45 = HEAP8[$cursor$03$i>>0]|0;
       $46 = $45&255;
       $cursor$1$i = $44;$mask$1$i = 128;$val$1$i = $46;
      } else {
       $cursor$1$i = $cursor$03$i;$mask$1$i = $mask$05$i;$val$1$i = $val$02$i;
      }
      $47 = $mask$1$i & $val$1$i;
      $48 = ($47|0)==(0);
      if ($48) {
       $113 = $60;
      } else {
       $49 = HEAP32[$7>>2]|0;
       $50 = (($49) + (($idx$04$i*28)|0)|0);
       $51 = ($0>>>0)>($idx$04$i>>>0);
       if ($51) {
        $52 = (((($49) + (($idx$04$i*28)|0)|0)) + 16|0);
        $53 = HEAP32[$52>>2]|0;
        $54 = $53 & 4;
        $55 = ($54|0)==(0);
        if ($55) {
         $56 = $53 | 4;
         HEAP32[$52>>2] = $56;
         $57 = HEAP32[$17>>2]|0;
         $58 = (((($49) + (($idx$04$i*28)|0)|0)) + 20|0);
         HEAP32[$58>>2] = 0;
         $59 = ($60|0)==(0);
         L30: do {
          if (!($59)) {
           $$idx$val$i$i = HEAP32[$50>>2]|0;
           $$idx1$i$i = (((($49) + (($idx$04$i*28)|0)|0)) + 4|0);
           $$idx1$val$i$i = HEAP32[$$idx1$i$i>>2]|0;
           $61 = (($$idx1$val$i$i) + ($$idx$val$i$i))|0;
           $count$03$i$i = $60;$sorted$02$i$i = $57;
           while(1) {
            $62 = HEAP32[$sorted$02$i$i>>2]|0;
            $63 = HEAP32[$62>>2]|0;
            $64 = ($61|0)<($63|0);
            if (!($64)) {
             $65 = ((($62)) + 4|0);
             $66 = HEAP32[$65>>2]|0;
             $67 = (($66) + ($63))|0;
             $68 = ($67|0)<($$idx$val$i$i|0);
             if (!($68)) {
              $$lcssa34 = $62;
              break;
             }
            }
            $69 = (($count$03$i$i) + -1)|0;
            $70 = ((($sorted$02$i$i)) + 4|0);
            $71 = ($69|0)==(0);
            if ($71) {
             break L30;
            } else {
             $count$03$i$i = $69;$sorted$02$i$i = $70;
            }
           }
           HEAP32[$58>>2] = $$lcssa34;
          }
         } while(0);
         $72 = ($60>>>0)<($0>>>0);
         if ($72) {
          $73 = (($60) + 1)|0;
          HEAP32[$18>>2] = $73;
          $74 = (($57) + ($60<<2)|0);
          HEAP32[$74>>2] = $50;
          $113 = $73;
         } else {
          $113 = $60;
         }
        } else {
         $113 = $60;
        }
       } else {
        $113 = $60;
       }
      }
      $75 = $mask$1$i >> 1;
      $76 = (($idx$04$i) + 1)|0;
      $exitcond$i = ($76|0)==($mask$0$idx$val|0);
      if ($exitcond$i) {
       $112 = $113;
       break;
      } else {
       $60 = $113;$cursor$03$i = $cursor$1$i;$idx$04$i = $76;$mask$05$i = $75;$val$02$i = $val$1$i;
      }
     }
    }
    $77 = (($count$18) + -1)|0;
    $78 = ((($mask$09)) + 16|0);
    $79 = ($77|0)==(0);
    if ($79) {
     $81 = $112;
     break;
    } else {
     $111 = $112;$count$18 = $77;$mask$09 = $78;
    }
   }
  }
 }
 $80 = ($81|0)!=($0|0);
 $82 = ($0|0)!=(0);
 $or$cond = $80 & $82;
 if ($or$cond) {
  $94 = $81;$idx$07 = 0;
 } else {
  $110 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($110|0);
 }
 while(1) {
  $83 = HEAP32[$7>>2]|0;
  $84 = (($83) + (($idx$07*28)|0)|0);
  $85 = ($0>>>0)>($idx$07>>>0);
  if ($85) {
   $86 = (((($83) + (($idx$07*28)|0)|0)) + 16|0);
   $87 = HEAP32[$86>>2]|0;
   $88 = $87 & 4;
   $89 = ($88|0)==(0);
   if ($89) {
    $90 = $87 | 4;
    HEAP32[$86>>2] = $90;
    $91 = HEAP32[$17>>2]|0;
    $92 = (((($83) + (($idx$07*28)|0)|0)) + 20|0);
    HEAP32[$92>>2] = 0;
    $93 = ($94|0)==(0);
    L50: do {
     if (!($93)) {
      $$idx$val$i = HEAP32[$84>>2]|0;
      $$idx1$i = (((($83) + (($idx$07*28)|0)|0)) + 4|0);
      $$idx1$val$i = HEAP32[$$idx1$i>>2]|0;
      $95 = (($$idx1$val$i) + ($$idx$val$i))|0;
      $count$03$i = $94;$sorted$02$i = $91;
      while(1) {
       $96 = HEAP32[$sorted$02$i>>2]|0;
       $97 = HEAP32[$96>>2]|0;
       $98 = ($95|0)<($97|0);
       if (!($98)) {
        $99 = ((($96)) + 4|0);
        $100 = HEAP32[$99>>2]|0;
        $101 = (($100) + ($97))|0;
        $102 = ($101|0)<($$idx$val$i|0);
        if (!($102)) {
         $$lcssa = $96;
         break;
        }
       }
       $103 = (($count$03$i) + -1)|0;
       $104 = ((($sorted$02$i)) + 4|0);
       $105 = ($103|0)==(0);
       if ($105) {
        break L50;
       } else {
        $count$03$i = $103;$sorted$02$i = $104;
       }
      }
      HEAP32[$92>>2] = $$lcssa;
     }
    } while(0);
    $106 = ($94>>>0)<($0>>>0);
    if ($106) {
     $107 = (($94) + 1)|0;
     HEAP32[$18>>2] = $107;
     $108 = (($91) + ($94<<2)|0);
     HEAP32[$108>>2] = $84;
     $114 = $107;
    } else {
     $114 = $94;
    }
   } else {
    $114 = $94;
   }
  } else {
   $114 = $94;
  }
  $109 = (($idx$07) + 1)|0;
  $exitcond = ($109|0)==($0|0);
  if ($exitcond) {
   break;
  } else {
   $94 = $114;$idx$07 = $109;
  }
 }
 $110 = HEAP32[$error>>2]|0;
 STACKTOP = sp;return ($110|0);
}
function _psh_hint_align($hint,$globals,$dimension,$glyph) {
 $hint = $hint|0;
 $globals = $globals|0;
 $dimension = $dimension|0;
 $glyph = $glyph|0;
 var $$ = 0, $$$0$i = 0, $$$i = 0, $$0$i = 0, $$0468$i = 0, $$2 = 0, $$3$i = 0, $$lcssa45 = 0, $$lcssa47 = 0, $$not = 0, $$pn14$i = 0, $$pn14$i$lcssa44 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $align$sroa$0$0 = 0, $align$sroa$0$2 = 0, $align$sroa$0$29 = 0, $align$sroa$14$1 = 0, $align$sroa$9$0 = 0, $align$sroa$9$161114 = 0, $align$sroa$9$17 = 0, $count$010$i = 0, $count$16$i = 0, $len$1 = 0, $len$113 = 0, $len$115 = 0;
 var $len$117 = 0, $len$118 = 0, $pos$0 = 0, $pos$2 = 0, $pos$3 = 0, $right_disp$0 = 0, $storemerge = 0, $storemerge$c = 0, $storemerge$in = 0, $storemerge$in$c = 0, $sum = 0, $zone$011$i = 0, $zone$011$i$lcssa46 = 0, $zone$18$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (((((($globals)) + 4|0) + (($dimension*204)|0)|0)) + 196|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($hint)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 8;
 $5 = ($4|0)==(0);
 if (!($5)) {
  return;
 }
 $6 = (((((($globals)) + 4|0) + (($dimension*204)|0)|0)) + 200|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = HEAP32[$hint>>2]|0;
 $9 = (_FT_MulFix($8,$1)|0);
 $10 = (($9) + ($7))|0;
 $11 = ((($hint)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (_FT_MulFix($12,$1)|0);
 L4: do {
  switch ($dimension|0) {
  case 0:  {
   $14 = ((($glyph)) + 120|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = $15&255;
   $17 = ($16<<24>>24)==(0);
   if ($17) {
    label = 7;
   } else {
    $36 = $15 & 16711680;
    $$not = ($36|0)!=(0);
    $37 = ((($hint)) + 12|0);
    HEAP32[$37>>2] = $13;
    $115 = $37;$194 = $$not;$align$sroa$0$29 = 0;$align$sroa$9$17 = 0;
    label = 28;
   }
   break;
  }
  case 1:  {
   $19 = ((($glyph)) + 121|0);
   $20 = HEAP8[$19>>0]|0;
   $21 = ($20<<24>>24)==(0);
   if ($21) {
    label = 7;
   } else {
    $22 = ((($glyph)) + 123|0);
    $23 = HEAP8[$22>>0]|0;
    $24 = ($23<<24>>24)!=(0);
    $25 = ((($hint)) + 12|0);
    HEAP32[$25>>2] = $13;
    $26 = HEAP32[$hint>>2]|0;
    $27 = HEAP32[$11>>2]|0;
    $28 = (($27) + ($26))|0;
    $29 = ((($globals)) + 2492|0);
    $30 = HEAP8[$29>>0]|0;
    $31 = ((($globals)) + 412|0);
    $32 = HEAP32[$31>>2]|0;
    $33 = ($32|0)==(0);
    L10: do {
     if ($33) {
      $195 = 2;$align$sroa$0$0 = 0;$align$sroa$9$0 = 0;
     } else {
      $38 = ((($globals)) + 416|0);
      $39 = ((($globals)) + 2488|0);
      $40 = HEAP32[$39>>2]|0;
      $41 = (0 - ($40))|0;
      $count$010$i = $32;$zone$011$i = $38;
      while(1) {
       $42 = ((($zone$011$i)) + 12|0);
       $43 = HEAP32[$42>>2]|0;
       $44 = (($28) - ($43))|0;
       $45 = ($44|0)<($41|0);
       if ($45) {
        $195 = 2;$align$sroa$0$0 = 0;$align$sroa$9$0 = 0;
        break L10;
       }
       $46 = ((($zone$011$i)) + 8|0);
       $47 = HEAP32[$46>>2]|0;
       $48 = (($47) + ($40))|0;
       $49 = ($48|0)<($28|0);
       if (!($49)) {
        $$lcssa47 = $44;$zone$011$i$lcssa46 = $zone$011$i;
        break;
       }
       $56 = (($count$010$i) + -1)|0;
       $57 = ((($zone$011$i)) + 32|0);
       $58 = ($56|0)==(0);
       if ($58) {
        $195 = 2;$align$sroa$0$0 = 0;$align$sroa$9$0 = 0;
        break L10;
       } else {
        $count$010$i = $56;$zone$011$i = $57;
       }
      }
      $50 = ($30<<24>>24)==(0);
      if ($50) {
       $51 = ((($globals)) + 2484|0);
       $52 = HEAP32[$51>>2]|0;
       $53 = ($$lcssa47|0)>($52|0);
       if ($53) {
        $195 = 2;$align$sroa$0$0 = 0;$align$sroa$9$0 = 0;
        break;
       }
      }
      $54 = ((($zone$011$i$lcssa46)) + 16|0);
      $55 = HEAP32[$54>>2]|0;
      $195 = 3;$align$sroa$0$0 = 1;$align$sroa$9$0 = $55;
     }
    } while(0);
    $59 = ((($globals)) + 928|0);
    $60 = HEAP32[$59>>2]|0;
    $61 = ($60|0)==(0);
    L21: do {
     if ($61) {
      $align$sroa$0$2 = $align$sroa$0$0;$align$sroa$14$1 = 0;
     } else {
      $62 = (((($globals)) + 932|0) + ($60<<5)|0);
      $63 = ((($globals)) + 2488|0);
      $64 = HEAP32[$63>>2]|0;
      $65 = (0 - ($64))|0;
      $$pn14$i = $62;$count$16$i = $60;
      while(1) {
       $68 = ((($$pn14$i)) + -24|0);
       $69 = HEAP32[$68>>2]|0;
       $70 = (($69) - ($26))|0;
       $71 = ($70|0)<($65|0);
       if ($71) {
        $align$sroa$0$2 = $align$sroa$0$0;$align$sroa$14$1 = 0;
        break L21;
       }
       $72 = ((($$pn14$i)) + -20|0);
       $73 = HEAP32[$72>>2]|0;
       $74 = (($73) - ($64))|0;
       $75 = ($74|0)>($26|0);
       $67 = (($count$16$i) + -1)|0;
       if (!($75)) {
        $$lcssa45 = $70;$$pn14$i$lcssa44 = $$pn14$i;
        break;
       }
       $zone$18$i = ((($$pn14$i)) + -32|0);
       $66 = ($67|0)==(0);
       if ($66) {
        $align$sroa$0$2 = $align$sroa$0$0;$align$sroa$14$1 = 0;
        break L21;
       } else {
        $$pn14$i = $zone$18$i;$count$16$i = $67;
       }
      }
      $76 = ($30<<24>>24)==(0);
      if ($76) {
       $77 = ((($globals)) + 2484|0);
       $78 = HEAP32[$77>>2]|0;
       $79 = ($$lcssa45|0)<($78|0);
       if (!($79)) {
        $align$sroa$0$2 = $align$sroa$0$0;$align$sroa$14$1 = 0;
        break;
       }
      }
      $80 = ((($$pn14$i$lcssa44)) + -16|0);
      $81 = HEAP32[$80>>2]|0;
      $align$sroa$0$2 = $195;$align$sroa$14$1 = $81;
     }
    } while(0);
    switch ($align$sroa$0$2|0) {
    case 1:  {
     $172 = (($align$sroa$9$0) - ($13))|0;
     $173 = ((($hint)) + 8|0);
     HEAP32[$173>>2] = $172;
     if (!($24)) {
      break L4;
     }
     $174 = ($13|0)<(64);
     $175 = (($13) + 32)|0;
     $176 = $175 & -64;
     $len$113 = $174 ? 64 : $176;
     $181 = $173;$182 = $25;$align$sroa$9$161114 = $align$sroa$9$0;$len$115 = $len$113;
     label = 50;
     break L4;
     break;
    }
    case 2:  {
     $82 = ((($hint)) + 8|0);
     HEAP32[$82>>2] = $align$sroa$14$1;
     if (!($24)) {
      break L4;
     }
     $83 = ($13|0)<(64);
     $84 = (($13) + 32)|0;
     $85 = $84 & -64;
     $len$117 = $83 ? 64 : $85;
     $183 = $25;$len$118 = $len$117;
     label = 51;
     break L4;
     break;
    }
    case 3:  {
     $86 = ((($hint)) + 8|0);
     HEAP32[$86>>2] = $align$sroa$14$1;
     $87 = (($align$sroa$9$0) - ($align$sroa$14$1))|0;
     HEAP32[$25>>2] = $87;
     break L4;
     break;
    }
    default: {
     $115 = $25;$194 = $24;$align$sroa$0$29 = $align$sroa$0$2;$align$sroa$9$17 = $align$sroa$9$0;
     label = 28;
     break L4;
    }
    }
   }
   break;
  }
  default: {
   $18 = ((($hint)) + 12|0);
   HEAP32[$18>>2] = $13;
   $115 = $18;$194 = 0;$align$sroa$0$29 = 0;$align$sroa$9$17 = 0;
   label = 28;
  }
  }
 } while(0);
 if ((label|0) == 7) {
  $34 = ((($hint)) + 8|0);
  HEAP32[$34>>2] = $10;
  $35 = ((($hint)) + 12|0);
  HEAP32[$35>>2] = $13;
  $storemerge$in$c = HEAP32[$2>>2]|0;
  $storemerge$c = $storemerge$in$c | 8;
  HEAP32[$2>>2] = $storemerge$c;
  return;
 }
 L41: do {
  if ((label|0) == 28) {
   $88 = ((($hint)) + 20|0);
   $89 = HEAP32[$88>>2]|0;
   $90 = ($89|0)==(0|0);
   if ($90) {
    $pos$0 = $10;
   } else {
    $91 = ((($89)) + 16|0);
    $92 = HEAP32[$91>>2]|0;
    $93 = $92 & 8;
    $94 = ($93|0)==(0);
    if ($94) {
     _psh_hint_align($89,$globals,$dimension,$glyph);
    }
    $95 = HEAP32[$89>>2]|0;
    $96 = ((($89)) + 4|0);
    $97 = HEAP32[$96>>2]|0;
    $98 = $97 >> 1;
    $99 = ((($89)) + 8|0);
    $100 = HEAP32[$99>>2]|0;
    $101 = ((($89)) + 12|0);
    $102 = HEAP32[$101>>2]|0;
    $103 = $102 >> 1;
    $104 = HEAP32[$hint>>2]|0;
    $105 = HEAP32[$11>>2]|0;
    $106 = $105 >> 1;
    $sum = (($98) + ($95))|0;
    $107 = (($104) - ($sum))|0;
    $108 = (($107) + ($106))|0;
    $109 = (_FT_MulFix($108,$1)|0);
    $110 = $13 >> 1;
    $111 = (($100) - ($110))|0;
    $112 = (($111) + ($103))|0;
    $113 = (($112) + ($109))|0;
    $pos$0 = $113;
   }
   $114 = ((($hint)) + 8|0);
   HEAP32[$114>>2] = $pos$0;
   HEAP32[$115>>2] = $13;
   $116 = ((($glyph)) + 124|0);
   $117 = HEAP8[$116>>0]|0;
   $118 = ($117<<24>>24)==(0);
   do {
    if ($118) {
     $160 = $13;$pos$2 = $pos$0;
    } else {
     $119 = ($13|0)<(65);
     if ($119) {
      $120 = ($13|0)>(31);
      if ($120) {
       $121 = $13 >> 1;
       $122 = (($pos$0) + ($121))|0;
       $123 = $122 & -64;
       $160 = 64;$pos$2 = $123;
       break;
      }
      $124 = ($13|0)>(0);
      $125 = (($pos$0) + 32)|0;
      $126 = $125 & -64;
      if (!($124)) {
       $160 = $13;$pos$2 = $126;
       break;
      }
      $127 = (($pos$0) + ($13))|0;
      $128 = (($127) + 32)|0;
      $129 = $128 & -64;
      $130 = (($126) - ($pos$0))|0;
      $131 = (($129) - ($127))|0;
      $132 = ($130|0)<(0);
      $133 = (0 - ($130))|0;
      $$ = $132 ? $133 : $130;
      $134 = ($131|0)<(0);
      $135 = (0 - ($131))|0;
      $right_disp$0 = $134 ? $135 : $131;
      $136 = ($$|0)>($right_disp$0|0);
      $$2 = $136 ? $129 : $126;
      $160 = $13;$pos$2 = $$2;
      break;
     }
     $137 = (((((($globals)) + 4|0) + (($dimension*204)|0)|0)) + 8|0);
     $138 = HEAP32[$137>>2]|0;
     $139 = (($13) - ($138))|0;
     $140 = ($139|0)<(0);
     $141 = (0 - ($139))|0;
     $$3$i = $140 ? $141 : $139;
     $142 = ($$3$i|0)<(40);
     if ($142) {
      $143 = ($138|0)<(48);
      if ($143) {
       $$0468$i = 48;$151 = 48;$153 = 0;
      } else {
       $$0$i = $138;
       label = 40;
      }
     } else {
      $$0$i = $13;
      label = 40;
     }
     if ((label|0) == 40) {
      $144 = ($$0$i|0)<(192);
      if (!($144)) {
       $154 = (($$0$i) + 32)|0;
       $155 = $154 & -64;
       $160 = $155;$pos$2 = $pos$0;
       break;
      }
      $145 = $$0$i & 63;
      $146 = $$0$i & -64;
      $147 = ($145>>>0)<(10);
      if ($147) {
       $160 = $$0$i;$pos$2 = $pos$0;
       break;
      }
      $148 = ($145>>>0)<(32);
      if ($148) {
       $149 = $146 | 10;
       $160 = $149;$pos$2 = $pos$0;
       break;
      } else {
       $$0468$i = $$0$i;$151 = $145;$153 = $146;
      }
     }
     $150 = ($151>>>0)<(54);
     $152 = $153 | 54;
     $$$0$i = $150 ? $152 : $$0468$i;
     $160 = $$$0$i;$pos$2 = $pos$0;
    }
   } while(0);
   $156 = (($pos$2) + 32)|0;
   $157 = $156 & -64;
   $158 = (($157) - ($pos$2))|0;
   $159 = (($156) + ($160))|0;
   $161 = $159 & -64;
   $162 = (($161) - ($pos$2))|0;
   $163 = (($162) - ($160))|0;
   $164 = ($158|0)<(0);
   $165 = (0 - ($158))|0;
   $166 = $164 ? $165 : $158;
   $167 = ($163|0)<(0);
   $168 = (0 - ($163))|0;
   $169 = $167 ? $168 : $163;
   $170 = ($166|0)>($169|0);
   $$$i = $170 ? $163 : $158;
   $171 = (($$$i) + ($pos$2))|0;
   HEAP32[$114>>2] = $171;
   HEAP32[$115>>2] = $160;
   if ($194) {
    $177 = ($160|0)<(64);
    $178 = (($160) + 32)|0;
    $179 = $178 & -64;
    $len$1 = $177 ? 64 : $179;
    switch ($align$sroa$0$29|0) {
    case 3:  {
     break L41;
     break;
    }
    case 1:  {
     $181 = $114;$182 = $115;$align$sroa$9$161114 = $align$sroa$9$17;$len$115 = $len$1;
     label = 50;
     break L41;
     break;
    }
    case 2:  {
     $183 = $115;$len$118 = $len$1;
     label = 51;
     break L41;
     break;
    }
    default: {
    }
    }
    $184 = HEAP32[$114>>2]|0;
    $185 = $len$1 & 64;
    $186 = ($185|0)==(0);
    $187 = $len$1 >> 1;
    $188 = (($184) + ($187))|0;
    $189 = $188 & -64;
    $190 = $189 | 32;
    $191 = (($188) + 32)|0;
    $192 = $191 & -64;
    $pos$3 = $186 ? $192 : $190;
    $193 = (($pos$3) - ($187))|0;
    HEAP32[$114>>2] = $193;
    HEAP32[$115>>2] = $len$1;
   }
  }
 } while(0);
 if ((label|0) == 50) {
  $180 = (($align$sroa$9$161114) - ($len$115))|0;
  HEAP32[$181>>2] = $180;
  HEAP32[$182>>2] = $len$115;
 }
 else if ((label|0) == 51) {
  HEAP32[$183>>2] = $len$118;
 }
 $storemerge$in = HEAP32[$2>>2]|0;
 $storemerge = $storemerge$in | 8;
 HEAP32[$2>>2] = $storemerge;
 return;
}
function _psh_hint_table_activate_mask($table,$hint_mask$0$0$val,$hint_mask$0$2$val) {
 $table = $table|0;
 $hint_mask$0$0$val = $hint_mask$0$0$val|0;
 $hint_mask$0$2$val = $hint_mask$0$2$val|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $count$01$i = 0, $count$09 = 0, $count$3 = 0, $count$3$lcssa = 0, $cursor$011 = 0, $cursor$1 = 0, $exitcond = 0, $exitcond15 = 0, $hint$02$i = 0, $i1$05 = 0, $i2$03 = 0, $i2$03$in = 0, $idx$010 = 0, $mask$013 = 0, $mask$1 = 0, $table$idx$val = 0, $table$idx1 = 0;
 var $table$idx1$val = 0, $val$012 = 0, $val$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $table$idx$val = HEAP32[$table>>2]|0;
 $table$idx1 = ((($table)) + 8|0);
 $0 = ($table$idx$val|0)==(0);
 if (!($0)) {
  $table$idx1$val = HEAP32[$table$idx1>>2]|0;
  $count$01$i = $table$idx$val;$hint$02$i = $table$idx1$val;
  while(1) {
   $1 = ((($hint$02$i)) + 16|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = $2 & -5;
   HEAP32[$1>>2] = $3;
   $4 = ((($hint$02$i)) + 24|0);
   HEAP32[$4>>2] = -1;
   $5 = (($count$01$i) + -1)|0;
   $6 = ((($hint$02$i)) + 28|0);
   $7 = ($5|0)==(0);
   if ($7) {
    break;
   } else {
    $count$01$i = $5;$hint$02$i = $6;
   }
  }
 }
 $8 = ($hint_mask$0$0$val|0)==(0);
 if ($8) {
  $9 = ((($table)) + 4|0);
  HEAP32[$9>>2] = 0;
  return;
 }
 $10 = ((($table)) + 12|0);
 $count$09 = 0;$cursor$011 = $hint_mask$0$2$val;$idx$010 = 0;$mask$013 = 0;$val$012 = 0;
 while(1) {
  $11 = ($mask$013|0)==(0);
  if ($11) {
   $12 = ((($cursor$011)) + 1|0);
   $13 = HEAP8[$cursor$011>>0]|0;
   $14 = $13&255;
   $cursor$1 = $12;$mask$1 = 128;$val$1 = $14;
  } else {
   $cursor$1 = $cursor$011;$mask$1 = $mask$013;$val$1 = $val$012;
  }
  $15 = $mask$1 & $val$1;
  $16 = ($15|0)==(0);
  if ($16) {
   $count$3 = $count$09;
  } else {
   $17 = HEAP32[$table$idx1>>2]|0;
   $18 = (($17) + (($idx$010*28)|0)|0);
   $19 = (((($17) + (($idx$010*28)|0)|0)) + 16|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = $20 & 4;
   $22 = ($21|0)==(0);
   if ($22) {
    $23 = $20 | 4;
    HEAP32[$19>>2] = $23;
    $24 = ($count$09>>>0)<($table$idx$val>>>0);
    if ($24) {
     $25 = (($count$09) + 1)|0;
     $26 = HEAP32[$10>>2]|0;
     $27 = (($26) + ($count$09<<2)|0);
     HEAP32[$27>>2] = $18;
     $count$3 = $25;
    } else {
     $count$3 = $count$09;
    }
   } else {
    $count$3 = $count$09;
   }
  }
  $28 = $mask$1 >> 1;
  $29 = (($idx$010) + 1)|0;
  $exitcond15 = ($29|0)==($hint_mask$0$0$val|0);
  if ($exitcond15) {
   $count$3$lcssa = $count$3;
   break;
  } else {
   $count$09 = $count$3;$cursor$011 = $cursor$1;$idx$010 = $29;$mask$013 = $28;$val$012 = $val$1;
  }
 }
 $30 = ((($table)) + 4|0);
 HEAP32[$30>>2] = $count$3$lcssa;
 $31 = HEAP32[$10>>2]|0;
 $32 = ($count$3$lcssa|0)>(1);
 if ($32) {
  $i1$05 = 1;
 } else {
  return;
 }
 while(1) {
  $33 = (($31) + ($i1$05<<2)|0);
  $34 = HEAP32[$33>>2]|0;
  $35 = HEAP32[$34>>2]|0;
  $i2$03$in = $i1$05;
  while(1) {
   $i2$03 = (($i2$03$in) + -1)|0;
   $36 = (($31) + ($i2$03<<2)|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = HEAP32[$37>>2]|0;
   $39 = ($38|0)<($35|0);
   if ($39) {
    break;
   }
   $40 = (($31) + ($i2$03$in<<2)|0);
   HEAP32[$40>>2] = $37;
   HEAP32[$36>>2] = $34;
   $41 = ($i2$03$in|0)>(1);
   if ($41) {
    $i2$03$in = $i2$03;
   } else {
    break;
   }
  }
  $42 = (($i1$05) + 1)|0;
  $exitcond = ($42|0)==($count$3$lcssa|0);
  if ($exitcond) {
   break;
  } else {
   $i1$05 = $42;
  }
 }
 return;
}
function _psh_hint_table_find_strong_points($table$0$1$val,$table$0$3$val,$point,$count,$threshold,$major_dir) {
 $table$0$1$val = $table$0$1$val|0;
 $table$0$3$val = $table$0$3$val|0;
 $point = $point|0;
 $count = $count|0;
 $threshold = $threshold|0;
 $major_dir = $major_dir|0;
 var $$ = 0, $$024 = 0, $$0623 = 0, $$13 = 0, $$lcssa = 0, $$lcssa90 = 0, $$lcssa92 = 0, $$lcssa94 = 0, $$lcssa96 = 0, $$not = 0, $$not35 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $brmerge = 0, $brmerge34 = 0, $brmerge37 = 0, $nn$015 = 0, $nn1$014 = 0, $nn5$017 = 0, $nn5$119 = 0, $nn5$221 = 0, $or$cond = 0, $or$cond11 = 0, $or$cond15 = 0, $or$cond18 = 0, $or$cond31 = 0, $or$cond9 = 0, $point_dir$0 = 0;
 var $sext = 0, $sext50 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($count|0)==(0);
 if ($0) {
  return;
 }
 $1 = (0 - ($major_dir))|0;
 $2 = ($table$0$1$val|0)==(0);
 $3 = ($major_dir|0)==(2);
 $$ = $3 ? 128 : 256;
 $$13 = $3 ? 256 : 128;
 $$024 = $point;$$0623 = $count;
 while(1) {
  $4 = ((($$024)) + 28|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ((($$024)) + 16|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = $7 & 16;
  $9 = ($8|0)==(0);
  L6: do {
   if ($9) {
    $10 = ((($$024)) + 20|0);
    $11 = HEAP16[$10>>1]|0;
    $12 = $11&65535;
    $sext = $12 << 24;
    $13 = $sext >> 24;
    $14 = ($13|0)==($major_dir|0);
    $15 = ($13|0)==($1|0);
    $or$cond31 = $14 | $15;
    if ($or$cond31) {
     $point_dir$0 = $13;
     label = 6;
    } else {
     $16 = ($11&65535) >>> 8;
     $17 = $16&65535;
     $sext50 = $17 << 24;
     $18 = $sext50 >> 24;
     $19 = ($18|0)==($major_dir|0);
     $20 = ($18|0)==($1|0);
     $or$cond = $19 | $20;
     if ($or$cond) {
      $point_dir$0 = $18;
      label = 6;
     }
    }
    if ((label|0) == 6) {
     label = 0;
     $21 = ($point_dir$0|0)==(0);
     if (!($21)) {
      $22 = ($point_dir$0|0)==($major_dir|0);
      if ($22) {
       if ($2) {
        break;
       } else {
        $nn$015 = 0;
       }
       while(1) {
        $25 = (($table$0$3$val) + ($nn$015<<2)|0);
        $26 = HEAP32[$25>>2]|0;
        $27 = HEAP32[$26>>2]|0;
        $28 = (($5) - ($27))|0;
        $29 = ($28|0)<($threshold|0);
        $30 = (0 - ($28))|0;
        $31 = ($30|0)<($threshold|0);
        $or$cond9 = $29 & $31;
        $24 = (($nn$015) + 1)|0;
        if ($or$cond9) {
         $$lcssa90 = $26;
         break;
        }
        $23 = ($24>>>0)<($table$0$1$val>>>0);
        if ($23) {
         $nn$015 = $24;
        } else {
         break L6;
        }
       }
       $32 = $7 | 528;
       HEAP32[$6>>2] = $32;
       $33 = ((($$024)) + 24|0);
       HEAP32[$33>>2] = $$lcssa90;
       break;
      } else {
       $$not = ($point_dir$0|0)!=($1|0);
       $brmerge = $$not | $2;
       if ($brmerge) {
        break;
       } else {
        $nn1$014 = 0;
       }
       while(1) {
        $36 = (($table$0$3$val) + ($nn1$014<<2)|0);
        $37 = HEAP32[$36>>2]|0;
        $38 = HEAP32[$37>>2]|0;
        $39 = (($5) - ($38))|0;
        $40 = ((($37)) + 4|0);
        $41 = HEAP32[$40>>2]|0;
        $42 = (($39) - ($41))|0;
        $43 = ($42|0)<($threshold|0);
        $44 = (0 - ($42))|0;
        $45 = ($44|0)<($threshold|0);
        $or$cond11 = $43 & $45;
        $35 = (($nn1$014) + 1)|0;
        if ($or$cond11) {
         $$lcssa = $37;
         break;
        }
        $34 = ($35>>>0)<($table$0$1$val>>>0);
        if ($34) {
         $nn1$014 = $35;
        } else {
         break L6;
        }
       }
       $46 = $7 | 1040;
       HEAP32[$6>>2] = $46;
       $47 = ((($$024)) + 24|0);
       HEAP32[$47>>2] = $$lcssa;
       break;
      }
     }
    }
    $48 = $7 & 64;
    $49 = ($48|0)==(0);
    if (!($49)) {
     $50 = $7 & $$;
     $51 = ($50|0)==(0);
     L26: do {
      if ($51) {
       $63 = $7 & $$13;
       $64 = ($63|0)==(0);
       $brmerge34 = $64 | $2;
       if (!($brmerge34)) {
        $nn5$119 = 0;
        while(1) {
         $67 = (($table$0$3$val) + ($nn5$119<<2)|0);
         $68 = HEAP32[$67>>2]|0;
         $69 = HEAP32[$68>>2]|0;
         $70 = (($5) - ($69))|0;
         $71 = ((($68)) + 4|0);
         $72 = HEAP32[$71>>2]|0;
         $73 = (($70) - ($72))|0;
         $74 = ($73|0)<($threshold|0);
         $75 = (0 - ($73))|0;
         $76 = ($75|0)<($threshold|0);
         $or$cond18 = $74 & $76;
         $66 = (($nn5$119) + 1)|0;
         if ($or$cond18) {
          $$lcssa94 = $68;
          break;
         }
         $65 = ($66>>>0)<($table$0$1$val>>>0);
         if ($65) {
          $nn5$119 = $66;
         } else {
          break L26;
         }
        }
        $77 = ((($$024)) + 24|0);
        HEAP32[$77>>2] = $$lcssa94;
        $78 = $7 | 1040;
        HEAP32[$6>>2] = $78;
       }
      } else {
       if ($2) {
        break L6;
       } else {
        $nn5$017 = 0;
       }
       while(1) {
        $54 = (($table$0$3$val) + ($nn5$017<<2)|0);
        $55 = HEAP32[$54>>2]|0;
        $56 = HEAP32[$55>>2]|0;
        $57 = (($5) - ($56))|0;
        $58 = ($57|0)<($threshold|0);
        $59 = (0 - ($57))|0;
        $60 = ($59|0)<($threshold|0);
        $or$cond15 = $58 & $60;
        $53 = (($nn5$017) + 1)|0;
        if ($or$cond15) {
         $$lcssa92 = $55;
         break;
        }
        $52 = ($53>>>0)<($table$0$1$val>>>0);
        if ($52) {
         $nn5$017 = $53;
        } else {
         break L26;
        }
       }
       $61 = ((($$024)) + 24|0);
       HEAP32[$61>>2] = $$lcssa92;
       $62 = $7 | 528;
       HEAP32[$6>>2] = $62;
      }
     } while(0);
     $79 = ((($$024)) + 24|0);
     $80 = HEAP32[$79>>2]|0;
     $$not35 = ($80|0)!=(0|0);
     $brmerge37 = $$not35 | $2;
     if (!($brmerge37)) {
      $nn5$221 = 0;
      while(1) {
       $81 = (($table$0$3$val) + ($nn5$221<<2)|0);
       $82 = HEAP32[$81>>2]|0;
       $83 = HEAP32[$82>>2]|0;
       $84 = ($5|0)<($83|0);
       if (!($84)) {
        $85 = ((($82)) + 4|0);
        $86 = HEAP32[$85>>2]|0;
        $87 = (($86) + ($83))|0;
        $88 = ($5|0)>($87|0);
        if (!($88)) {
         $$lcssa96 = $82;
         break;
        }
       }
       $89 = (($nn5$221) + 1)|0;
       $90 = ($89>>>0)<($table$0$1$val>>>0);
       if ($90) {
        $nn5$221 = $89;
       } else {
        break L6;
       }
      }
      HEAP32[$79>>2] = $$lcssa96;
     }
    }
   }
  } while(0);
  $91 = (($$0623) + -1)|0;
  $92 = ((($$024)) + 40|0);
  $93 = ($91|0)==(0);
  if ($93) {
   break;
  } else {
   $$024 = $92;$$0623 = $91;
  }
 }
 return;
}
function _t2_hints_open($hints) {
 $hints = $hints|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($hints)) + 4|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($hints)) + 12|0);
 HEAP32[$1>>2] = 2;
 $2 = ((($hints)) + 16|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($hints)) + 28|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($hints)) + 40|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($hints)) + 52|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($hints)) + 64|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($hints)) + 76|0);
 HEAP32[$7>>2] = 0;
 return;
}
function _t2_hints_stems($hints,$dimension,$count,$coords) {
 $hints = $hints|0;
 $dimension = $dimension|0;
 $count = $count|0;
 $coords = $coords|0;
 var $$ = 0, $$0$i = 0, $$015$i = 0, $$034$i = 0, $$lcssa = 0, $$lcssa24 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $n$04 = 0, $n$16 = 0, $or$cond = 0, $stems = 0, $total$09 = 0, $y$08 = 0, $y$1$lcssa1517 = 0, $y$13 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $stems = sp;
 $0 = ($count|0)>(0);
 if (!($0)) {
  STACKTOP = sp;return;
 }
 $1 = ((($hints)) + 4|0);
 $2 = ($dimension>>>0)>(1);
 $3 = ($dimension|0)!=(0);
 $4 = $3&1;
 $$0$i = $2 ? $4 : $dimension;
 $5 = (((($hints)) + 16|0) + (($$0$i*36)|0)|0);
 $total$09 = $count;$y$08 = 0;
 while(1) {
  $6 = ($total$09|0)>(16);
  $$ = $6 ? 16 : $total$09;
  $7 = $$ << 1;
  $8 = ($$|0)>(0);
  L6: do {
   if ($8) {
    $n$04 = 0;$y$13 = $y$08;
    while(1) {
     $9 = (($coords) + ($n$04<<2)|0);
     $10 = HEAP32[$9>>2]|0;
     $11 = (($10) + ($y$13))|0;
     $12 = (_FT_RoundFix($11)|0);
     $13 = $12 >> 16;
     $14 = (($stems) + ($n$04<<2)|0);
     HEAP32[$14>>2] = $13;
     $15 = (($n$04) + 1)|0;
     $16 = ($15|0)<($7|0);
     if ($16) {
      $n$04 = $15;$y$13 = $11;
     } else {
      $$lcssa = $11;
      break;
     }
    }
    if ($8) {
     $n$16 = 0;
     while(1) {
      $17 = $n$16 | 1;
      $18 = (($stems) + ($17<<2)|0);
      $19 = HEAP32[$18>>2]|0;
      $20 = (($stems) + ($n$16<<2)|0);
      $21 = HEAP32[$20>>2]|0;
      $22 = (($19) - ($21))|0;
      HEAP32[$18>>2] = $22;
      $23 = (($n$16) + 2)|0;
      $24 = ($23|0)<($7|0);
      if ($24) {
       $n$16 = $23;
      } else {
       break;
      }
     }
     $25 = HEAP32[$1>>2]|0;
     $26 = ($25|0)==(0);
     $or$cond = $26 & $8;
     if ($or$cond) {
      $$015$i = $$;$$034$i = $stems;
      while(1) {
       $27 = HEAP32[$hints>>2]|0;
       $28 = HEAP32[$$034$i>>2]|0;
       $29 = ((($$034$i)) + 4|0);
       $30 = HEAP32[$29>>2]|0;
       $31 = (_ps_dimension_add_t1stem($5,$28,$30,$27,0)|0);
       $32 = ($31|0)==(0);
       if (!($32)) {
        $$lcssa24 = $31;
        break;
       }
       $33 = (($$015$i) + -1)|0;
       $34 = ((($$034$i)) + 8|0);
       $35 = ($$015$i|0)>(1);
       if ($35) {
        $$015$i = $33;$$034$i = $34;
       } else {
        $y$1$lcssa1517 = $$lcssa;
        break L6;
       }
      }
      HEAP32[$1>>2] = $$lcssa24;
      $y$1$lcssa1517 = $$lcssa;
     } else {
      $y$1$lcssa1517 = $$lcssa;
     }
    } else {
     $y$1$lcssa1517 = $$lcssa;
    }
   } else {
    $y$1$lcssa1517 = $y$08;
   }
  } while(0);
  $36 = (($total$09) - ($$))|0;
  $37 = ($36|0)>(0);
  if ($37) {
   $total$09 = $36;$y$08 = $y$1$lcssa1517;
  } else {
   break;
  }
 }
 STACKTOP = sp;return;
}
function _ps_hints_t2mask($hints,$end_point,$bit_count,$bytes) {
 $hints = $hints|0;
 $end_point = $end_point|0;
 $bit_count = $bit_count|0;
 $bytes = $bytes|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$0$ph = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($hints)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  return;
 }
 $3 = ((($hints)) + 16|0);
 $4 = HEAP32[$hints>>2]|0;
 $5 = HEAP32[$3>>2]|0;
 $6 = ((($hints)) + 52|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (($7) + ($5))|0;
 $9 = ($8|0)==($bit_count|0);
 if (!($9)) {
  return;
 }
 $10 = (_ps_dimension_set_mask_bits($3,$bytes,$7,$5,$end_point,$4)|0);
 $11 = ($10|0)==(0);
 if ($11) {
  $12 = (_ps_dimension_set_mask_bits($6,$bytes,0,$7,$end_point,$4)|0);
  $13 = ($12|0)==(0);
  if ($13) {
   return;
  } else {
   $error$0$ph = $12;
  }
 } else {
  $error$0$ph = $10;
 }
 HEAP32[$0>>2] = $error$0$ph;
 return;
}
function _ps_dimension_set_mask_bits($dim,$source,$source_pos,$source_bits,$end_point,$memory) {
 $dim = $dim|0;
 $source = $source|0;
 $source_pos = $source_pos|0;
 $source_bits = $source_bits|0;
 $end_point = $end_point|0;
 $memory = $memory|0;
 var $$$i = 0, $$02$i = 0, $$1$i = 0, $$phi$trans$insert$i$i = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i$i = 0, $$pre$phi$iZ2D = 0, $$write$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $8 = 0, $9 = 0, $error$0 = 0, $error$i$i = 0, $error$i$i$i = 0, $error$i$i$i$i = 0, $mask$0$i$sink$i$ph$i = 0, $read$03$i = 0, $read$1$i = 0, $rmask$04$i = 0, $rmask$1$i = 0, $wmask$06$i = 0, $write$05$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i$i = sp + 8|0;
 $error$i$i$i$i = sp + 4|0;
 $error$i$i$i = sp;
 $0 = ((($dim)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  $3 = ((($dim)) + 20|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = (($4) + ($1<<4)|0);
  $6 = ((($5)) + -16|0);
  $7 = ((($6)) + 12|0);
  HEAP32[$7>>2] = $end_point;
 }
 $8 = (($1) + 1)|0;
 $9 = ((($dim)) + 16|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($8>>>0)>($10>>>0);
 do {
  if ($11) {
   HEAP32[$error$i$i$i>>2] = 0;
   $12 = (($1) + 8)|0;
   $13 = $12 & -8;
   $14 = ((($dim)) + 20|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = (_ft_mem_realloc($memory,16,$10,$13,$15,$error$i$i$i)|0);
   HEAP32[$14>>2] = $16;
   $17 = HEAP32[$error$i$i$i>>2]|0;
   $18 = ($17|0)==(0);
   if ($18) {
    HEAP32[$9>>2] = $13;
    $$pre$i$i$i = $16;$20 = $16;$25 = $13;$27 = $16;
    break;
   }
   $error$0 = $17;
   STACKTOP = sp;return ($error$0|0);
  } else {
   $$phi$trans$insert$i$i = ((($dim)) + 20|0);
   $$pre$i$i = HEAP32[$$phi$trans$insert$i$i>>2]|0;
   $$pre$i$i$i = $$pre$i$i;$20 = $$pre$i$i;$25 = $10;$27 = $$pre$i$i;
  }
 } while(0);
 $19 = (($20) + ($8<<4)|0);
 $21 = ((($19)) + -16|0);
 HEAP32[$21>>2] = 0;
 $22 = ((($21)) + 12|0);
 HEAP32[$22>>2] = 0;
 HEAP32[$0>>2] = $8;
 $23 = ($8|0)==(0);
 if ($23) {
  $24 = ($25|0)==(0);
  do {
   if ($24) {
    HEAP32[$error$i$i$i$i>>2] = 0;
    $26 = ((($dim)) + 20|0);
    $28 = (_ft_mem_realloc($memory,16,0,8,$27,$error$i$i$i$i)|0);
    HEAP32[$26>>2] = $28;
    $29 = HEAP32[$error$i$i$i$i>>2]|0;
    $30 = ($29|0)==(0);
    if ($30) {
     HEAP32[$9>>2] = 8;
     $31 = $28;
     break;
    }
    $error$0 = $29;
    STACKTOP = sp;return ($error$0|0);
   } else {
    $31 = $$pre$i$i$i;
   }
  } while(0);
  HEAP32[$31>>2] = 0;
  $32 = ((($31)) + 12|0);
  HEAP32[$32>>2] = 0;
  HEAP32[$0>>2] = 1;
  $mask$0$i$sink$i$ph$i = $31;
 } else {
  $33 = (($$pre$i$i$i) + ($8<<4)|0);
  $34 = ((($33)) + -16|0);
  $mask$0$i$sink$i$ph$i = $34;
 }
 $35 = ((($mask$0$i$sink$i$ph$i)) + 4|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = (($36) + 7)|0;
 $38 = $37 >>> 3;
 $39 = (($source_bits) + 7)|0;
 $40 = $39 >>> 3;
 HEAP32[$error$i$i>>2] = 0;
 $41 = ($40>>>0)>($38>>>0);
 do {
  if ($41) {
   $42 = (($40) + 7)|0;
   $43 = $42 & 1073741816;
   $44 = ((($mask$0$i$sink$i$ph$i)) + 8|0);
   $45 = HEAP32[$44>>2]|0;
   $46 = (_ft_mem_realloc($memory,1,$38,$43,$45,$error$i$i)|0);
   HEAP32[$44>>2] = $46;
   $47 = HEAP32[$error$i$i>>2]|0;
   $48 = ($47|0)==(0);
   if ($48) {
    $49 = $43 << 3;
    HEAP32[$35>>2] = $49;
    $$pre$phi$iZ2D = $44;
    break;
   }
   $error$0 = $47;
   STACKTOP = sp;return ($error$0|0);
  } else {
   $$pre$i = ((($mask$0$i$sink$i$ph$i)) + 8|0);
   $$pre$phi$iZ2D = $$pre$i;
  }
 } while(0);
 HEAP32[$mask$0$i$sink$i$ph$i>>2] = $source_bits;
 $50 = ($source_bits|0)==(0);
 if ($50) {
  $error$0 = 0;
  STACKTOP = sp;return ($error$0|0);
 }
 $51 = HEAP32[$$pre$phi$iZ2D>>2]|0;
 $52 = $source_pos & 7;
 $53 = 128 >>> $52;
 $54 = $source_pos >>> 3;
 $55 = (($source) + ($54)|0);
 $$02$i = $source_bits;$read$03$i = $55;$rmask$04$i = $53;$wmask$06$i = 128;$write$05$i = $51;
 while(1) {
  $56 = HEAP8[$write$05$i>>0]|0;
  $57 = $56&255;
  $58 = $wmask$06$i ^ 255;
  $59 = $57 & $58;
  $60 = HEAP8[$read$03$i>>0]|0;
  $61 = $60&255;
  $62 = $61 & $rmask$04$i;
  $63 = ($62|0)==(0);
  $64 = $63 ? 0 : $wmask$06$i;
  $$$i = $64 | $59;
  $65 = $$$i&255;
  HEAP8[$write$05$i>>0] = $65;
  $66 = $rmask$04$i >> 1;
  $67 = ($66|0)==(0);
  $68 = ((($read$03$i)) + 1|0);
  $read$1$i = $67 ? $68 : $read$03$i;
  $rmask$1$i = $67 ? 128 : $66;
  $69 = $wmask$06$i >> 1;
  $70 = ($69|0)==(0);
  $71 = ((($write$05$i)) + 1|0);
  $$write$0$i = $70 ? $71 : $write$05$i;
  $$1$i = $70 ? 128 : $69;
  $72 = (($$02$i) + -1)|0;
  $73 = ($72|0)==(0);
  if ($73) {
   $error$0 = 0;
   break;
  } else {
   $$02$i = $72;$read$03$i = $read$1$i;$rmask$04$i = $rmask$1$i;$wmask$06$i = $$1$i;$write$05$i = $$write$0$i;
  }
 }
 STACKTOP = sp;return ($error$0|0);
}
function _ps_hints_t2counter($hints,$bit_count,$bytes) {
 $hints = $hints|0;
 $bit_count = $bit_count|0;
 $bytes = $bytes|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$0$ph = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($hints)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  return;
 }
 $3 = ((($hints)) + 16|0);
 $4 = HEAP32[$hints>>2]|0;
 $5 = HEAP32[$3>>2]|0;
 $6 = ((($hints)) + 52|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (($7) + ($5))|0;
 $9 = ($8|0)==($bit_count|0);
 if (!($9)) {
  return;
 }
 $10 = (_ps_dimension_set_mask_bits($3,$bytes,0,$5,0,$4)|0);
 $11 = ($10|0)==(0);
 if ($11) {
  $12 = (_ps_dimension_set_mask_bits($6,$bytes,$5,$7,0,$4)|0);
  $13 = ($12|0)==(0);
  if ($13) {
   return;
  } else {
   $error$0$ph = $12;
  }
 } else {
  $error$0$ph = $10;
 }
 HEAP32[$0>>2] = $error$0$ph;
 return;
}
function _ps_hinter_done($module) {
 $module = $module|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($module)) + 112|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($module)) + 140|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($module)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($module)) + 28|0);
 _ps_dimension_done($4,$3);
 $5 = ((($module)) + 64|0);
 _ps_dimension_done($5,$3);
 $6 = ((($module)) + 16|0);
 HEAP32[$6>>2] = 0;
 HEAP32[$2>>2] = 0;
 return;
}
function _ps_dimension_done($dimension,$memory) {
 $dimension = $dimension|0;
 $memory = $memory|0;
 var $$pre$i = 0, $$pre$i6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $count$01$i = 0, $count$01$i3 = 0, $mask$02$i = 0, $mask$02$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($dimension)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($dimension)) + 32|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==(0);
 if ($4) {
  $10 = $3;
 } else {
  $count$01$i = $1;$mask$02$i = $3;
  while(1) {
   $5 = ((($mask$02$i)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   _ft_mem_free($memory,$6);
   ;HEAP32[$mask$02$i>>2]=0|0;HEAP32[$mask$02$i+4>>2]=0|0;HEAP32[$mask$02$i+8>>2]=0|0;HEAP32[$mask$02$i+12>>2]=0|0;
   $7 = (($count$01$i) + -1)|0;
   $8 = ((($mask$02$i)) + 16|0);
   $9 = ($7|0)==(0);
   if ($9) {
    break;
   } else {
    $count$01$i = $7;$mask$02$i = $8;
   }
  }
  $$pre$i = HEAP32[$2>>2]|0;
  $10 = $$pre$i;
 }
 _ft_mem_free($memory,$10);
 HEAP32[$2>>2] = 0;
 $11 = ((($dimension)) + 24|0);
 HEAP32[$11>>2] = 0;
 HEAP32[$0>>2] = 0;
 $12 = ((($dimension)) + 16|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($dimension)) + 20|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($13|0)==(0);
 if ($16) {
  $22 = $15;
 } else {
  $count$01$i3 = $13;$mask$02$i2 = $15;
  while(1) {
   $17 = ((($mask$02$i2)) + 8|0);
   $18 = HEAP32[$17>>2]|0;
   _ft_mem_free($memory,$18);
   ;HEAP32[$mask$02$i2>>2]=0|0;HEAP32[$mask$02$i2+4>>2]=0|0;HEAP32[$mask$02$i2+8>>2]=0|0;HEAP32[$mask$02$i2+12>>2]=0|0;
   $19 = (($count$01$i3) + -1)|0;
   $20 = ((($mask$02$i2)) + 16|0);
   $21 = ($19|0)==(0);
   if ($21) {
    break;
   } else {
    $count$01$i3 = $19;$mask$02$i2 = $20;
   }
  }
  $$pre$i6 = HEAP32[$14>>2]|0;
  $22 = $$pre$i6;
 }
 _ft_mem_free($memory,$22);
 HEAP32[$14>>2] = 0;
 $23 = ((($dimension)) + 12|0);
 HEAP32[$23>>2] = 0;
 HEAP32[$12>>2] = 0;
 $24 = ((($dimension)) + 8|0);
 $25 = HEAP32[$24>>2]|0;
 _ft_mem_free($memory,$25);
 HEAP32[$24>>2] = 0;
 HEAP32[$dimension>>2] = 0;
 $26 = ((($dimension)) + 4|0);
 HEAP32[$26>>2] = 0;
 return;
}
function _ps_unicode_value($glyph_name) {
 $glyph_name = $glyph_name|0;
 var $$ = 0, $$$1 = 0, $$$2 = 0, $$$3 = 0, $$4 = 0, $$8 = 0, $$8$1 = 0, $$8$2 = 0, $$8$3 = 0, $$8$4 = 0, $$8$5 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $p4$0 = 0, $p4$0$1 = 0, $p4$0$2 = 0, $p4$0$3 = 0, $p4$0$4 = 0, $p4$0$5 = 0, $p4$030 = 0, $p9$0 = 0, $p9$0$lcssa51 = 0, $p9$042 = 0, $value$111 = 0, $value$111$1 = 0, $value$111$2 = 0;
 var $value$111$3 = 0, $value3$118 = 0, $value3$118$1 = 0, $value3$118$2 = 0, $value3$118$3 = 0, $value3$118$4 = 0, $value3$118$5 = 0, $value3$220 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$glyph_name>>0]|0;
 $1 = ($0<<24>>24)==(117);
 L1: do {
  if ($1) {
   $2 = ((($glyph_name)) + 1|0);
   $3 = HEAP8[$2>>0]|0;
   $4 = ($3<<24>>24)==(110);
   L3: do {
    if ($4) {
     $5 = ((($glyph_name)) + 2|0);
     $6 = HEAP8[$5>>0]|0;
     $7 = ($6<<24>>24)==(105);
     if ($7) {
      $8 = ((($glyph_name)) + 3|0);
      $9 = HEAP8[$8>>0]|0;
      $10 = $9&255;
      $11 = (($10) + -48)|0;
      $12 = ($11>>>0)>(9);
      if ($12) {
       $13 = (($10) + -65)|0;
       $14 = ($13>>>0)>(5);
       $15 = (($10) + -55)|0;
       $$ = $14 ? 16 : $15;
       $16 = ($$>>>0)>(15);
       if ($16) {
        break;
       } else {
        $value$111 = $$;
       }
      } else {
       $value$111 = $11;
      }
      $17 = ((($glyph_name)) + 4|0);
      $18 = HEAP8[$17>>0]|0;
      $19 = $18&255;
      $20 = (($19) + -48)|0;
      $21 = ($20>>>0)>(9);
      if ($21) {
       $103 = (($19) + -65)|0;
       $104 = ($103>>>0)>(5);
       $105 = (($19) + -55)|0;
       $$$1 = $104 ? 16 : $105;
       $106 = ($$$1>>>0)>(15);
       $107 = $value$111 << 4;
       $108 = (($$$1) + ($107))|0;
       if ($106) {
        break;
       } else {
        $value$111$1 = $108;
       }
      } else {
       $101 = $value$111 << 4;
       $102 = (($20) + ($101))|0;
       $value$111$1 = $102;
      }
      $109 = ((($glyph_name)) + 5|0);
      $110 = HEAP8[$109>>0]|0;
      $111 = $110&255;
      $112 = (($111) + -48)|0;
      $113 = ($112>>>0)>(9);
      if ($113) {
       $116 = (($111) + -65)|0;
       $117 = ($116>>>0)>(5);
       $118 = (($111) + -55)|0;
       $$$2 = $117 ? 16 : $118;
       $119 = ($$$2>>>0)>(15);
       $120 = $value$111$1 << 4;
       $121 = (($$$2) + ($120))|0;
       if ($119) {
        break;
       } else {
        $value$111$2 = $121;
       }
      } else {
       $114 = $value$111$1 << 4;
       $115 = (($112) + ($114))|0;
       $value$111$2 = $115;
      }
      $122 = ((($glyph_name)) + 6|0);
      $123 = HEAP8[$122>>0]|0;
      $124 = $123&255;
      $125 = (($124) + -48)|0;
      $126 = ($125>>>0)>(9);
      if ($126) {
       $129 = (($124) + -65)|0;
       $130 = ($129>>>0)>(5);
       $131 = (($124) + -55)|0;
       $$$3 = $130 ? 16 : $131;
       $132 = ($$$3>>>0)>(15);
       $133 = $value$111$2 << 4;
       $134 = (($$$3) + ($133))|0;
       if ($132) {
        break;
       } else {
        $value$111$3 = $134;
       }
      } else {
       $127 = $value$111$2 << 4;
       $128 = (($125) + ($127))|0;
       $value$111$3 = $128;
      }
      $22 = ((($glyph_name)) + 7|0);
      $23 = HEAP8[$22>>0]|0;
      switch ($23<<24>>24) {
      case 46:  {
       break;
      }
      case 0:  {
       $$4 = $value$111$3;
       return ($$4|0);
       break;
      }
      default: {
       break L3;
      }
      }
      $24 = $value$111$3 | -2147483648;
      $$4 = $24;
      return ($$4|0);
     }
    }
   } while(0);
   $25 = $3&255;
   $26 = (($25) + -48)|0;
   $27 = ($26>>>0)>(9);
   if ($27) {
    $28 = (($25) + -65)|0;
    $29 = ($28>>>0)>(5);
    $30 = (($25) + -55)|0;
    $$8 = $29 ? 16 : $30;
    $31 = ($$8>>>0)>(15);
    if ($31) {
     $38 = $0;$p9$0 = $glyph_name;
     break;
    } else {
     $value3$118 = $$8;
    }
   } else {
    $value3$118 = $26;
   }
   $p4$0 = ((($glyph_name)) + 2|0);
   $32 = HEAP8[$p4$0>>0]|0;
   $33 = $32&255;
   $34 = (($33) + -48)|0;
   $35 = ($34>>>0)>(9);
   if ($35) {
    $47 = (($33) + -65)|0;
    $48 = ($47>>>0)>(5);
    $49 = (($33) + -55)|0;
    $$8$1 = $48 ? 16 : $49;
    $50 = ($$8$1>>>0)>(15);
    $51 = $value3$118 << 4;
    $52 = (($$8$1) + ($51))|0;
    if ($50) {
     $38 = $0;$p9$0 = $glyph_name;
     break;
    } else {
     $value3$118$1 = $52;
    }
   } else {
    $45 = $value3$118 << 4;
    $46 = (($34) + ($45))|0;
    $value3$118$1 = $46;
   }
   $p4$0$1 = ((($glyph_name)) + 3|0);
   $53 = HEAP8[$p4$0$1>>0]|0;
   $54 = $53&255;
   $55 = (($54) + -48)|0;
   $56 = ($55>>>0)>(9);
   if ($56) {
    $59 = (($54) + -65)|0;
    $60 = ($59>>>0)>(5);
    $61 = (($54) + -55)|0;
    $$8$2 = $60 ? 16 : $61;
    $62 = ($$8$2>>>0)>(15);
    $63 = $value3$118$1 << 4;
    $64 = (($$8$2) + ($63))|0;
    if ($62) {
     $38 = $0;$p9$0 = $glyph_name;
     break;
    } else {
     $value3$118$2 = $64;
    }
   } else {
    $57 = $value3$118$1 << 4;
    $58 = (($55) + ($57))|0;
    $value3$118$2 = $58;
   }
   $p4$0$2 = ((($glyph_name)) + 4|0);
   $65 = HEAP8[$p4$0$2>>0]|0;
   $66 = $65&255;
   $67 = (($66) + -48)|0;
   $68 = ($67>>>0)>(9);
   if ($68) {
    $71 = (($66) + -65)|0;
    $72 = ($71>>>0)>(5);
    $73 = (($66) + -55)|0;
    $$8$3 = $72 ? 16 : $73;
    $74 = ($$8$3>>>0)>(15);
    $75 = $value3$118$2 << 4;
    $76 = (($$8$3) + ($75))|0;
    if ($74) {
     $38 = $0;$p9$0 = $glyph_name;
     break;
    } else {
     $value3$118$3 = $76;
    }
   } else {
    $69 = $value3$118$2 << 4;
    $70 = (($67) + ($69))|0;
    $value3$118$3 = $70;
   }
   $p4$0$3 = ((($glyph_name)) + 5|0);
   $77 = HEAP8[$p4$0$3>>0]|0;
   $78 = $77&255;
   $79 = (($78) + -48)|0;
   $80 = ($79>>>0)>(9);
   if ($80) {
    $83 = (($78) + -65)|0;
    $84 = ($83>>>0)>(5);
    $85 = (($78) + -55)|0;
    $$8$4 = $84 ? 16 : $85;
    $86 = ($$8$4>>>0)>(15);
    $87 = $value3$118$3 << 4;
    $88 = (($$8$4) + ($87))|0;
    if ($86) {
     $p4$030 = $p4$0$3;$value3$220 = $value3$118$3;
    } else {
     $value3$118$4 = $88;
     label = 32;
    }
   } else {
    $81 = $value3$118$3 << 4;
    $82 = (($79) + ($81))|0;
    $value3$118$4 = $82;
    label = 32;
   }
   do {
    if ((label|0) == 32) {
     $p4$0$4 = ((($glyph_name)) + 6|0);
     $89 = HEAP8[$p4$0$4>>0]|0;
     $90 = $89&255;
     $91 = (($90) + -48)|0;
     $92 = ($91>>>0)>(9);
     if ($92) {
      $95 = (($90) + -65)|0;
      $96 = ($95>>>0)>(5);
      $97 = (($90) + -55)|0;
      $$8$5 = $96 ? 16 : $97;
      $98 = ($$8$5>>>0)>(15);
      $99 = $value3$118$4 << 4;
      $100 = (($$8$5) + ($99))|0;
      if ($98) {
       $p4$030 = $p4$0$4;$value3$220 = $value3$118$4;
       break;
      } else {
       $value3$118$5 = $100;
      }
     } else {
      $93 = $value3$118$4 << 4;
      $94 = (($91) + ($93))|0;
      $value3$118$5 = $94;
     }
     $p4$0$5 = ((($glyph_name)) + 7|0);
     $p4$030 = $p4$0$5;$value3$220 = $value3$118$5;
    }
   } while(0);
   $36 = HEAP8[$p4$030>>0]|0;
   switch ($36<<24>>24) {
   case 46:  {
    break;
   }
   case 0:  {
    $$4 = $value3$220;
    return ($$4|0);
    break;
   }
   default: {
    $38 = $0;$p9$0 = $glyph_name;
    break L1;
   }
   }
   $37 = $value3$220 | -2147483648;
   $$4 = $37;
   return ($$4|0);
  } else {
   $38 = $0;$p9$0 = $glyph_name;
  }
 } while(0);
 L55: while(1) {
  switch ($38<<24>>24) {
  case 0:  {
   $p9$042 = $p9$0;
   break L55;
   break;
  }
  case 46:  {
   $39 = ($p9$0>>>0)>($glyph_name>>>0);
   if ($39) {
    $p9$0$lcssa51 = $p9$0;
    label = 17;
    break L55;
   }
   break;
  }
  default: {
  }
  }
  $40 = ((($p9$0)) + 1|0);
  $$pre = HEAP8[$40>>0]|0;
  $38 = $$pre;$p9$0 = $40;
 }
 if ((label|0) == 17) {
  $41 = ($p9$0$lcssa51|0)==(0|0);
  if ($41) {
   $p9$042 = 0;
  } else {
   $43 = (_ft_get_adobe_glyph_index($glyph_name,$p9$0$lcssa51)|0);
   $44 = $43 | -2147483648;
   $$4 = $44;
   return ($$4|0);
  }
 }
 $42 = (_ft_get_adobe_glyph_index($glyph_name,$p9$042)|0);
 $$4 = $42;
 return ($$4|0);
}
function _ft_get_adobe_glyph_index($name,$limit) {
 $name = $name|0;
 $limit = $limit|0;
 var $$0 = 0, $$0115 = 0, $$lcssa58 = 0, $$lcssa60 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $8 = 0, $9 = 0, $count$012 = 0, $max$018 = 0, $max$1 = 0, $min$017 = 0, $min$1 = 0, $or$cond = 0;
 var $p$2$be = 0, $p$2$lcssa = 0, $p$214 = 0, $p$411 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($name|0)!=(0|0);
 $1 = ($name>>>0)<($limit>>>0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  $$0 = 0;
  return ($$0|0);
 }
 $2 = ((($name)) + 1|0);
 $3 = HEAP8[$name>>0]|0;
 $4 = $3 << 24 >> 24;
 $max$018 = 52;$min$017 = 0;
 while(1) {
  $5 = (($max$018) + ($min$017))|0;
  $6 = $5 >> 1;
  $7 = $6 << 1;
  $8 = ((533237) + ($7)|0);
  $9 = HEAP8[$8>>0]|0;
  $10 = $9&255;
  $11 = $10 << 8;
  $12 = ((($8)) + 1|0);
  $13 = HEAP8[$12>>0]|0;
  $14 = $13&255;
  $15 = $11 | $14;
  $16 = (533235 + ($15)|0);
  $17 = HEAP8[$16>>0]|0;
  $18 = $17&255;
  $19 = $18 & 127;
  $20 = ($19|0)==($4|0);
  if ($20) {
   $$lcssa58 = $16;$$lcssa60 = $17;
   break;
  }
  $22 = ($19|0)<($4|0);
  $23 = (($6) + 1)|0;
  $min$1 = $22 ? $23 : $min$017;
  $max$1 = $22 ? $max$018 : $6;
  $24 = ($min$1|0)<($max$1|0);
  if ($24) {
   $max$018 = $max$1;$min$017 = $min$1;
  } else {
   $$0 = 0;
   label = 16;
   break;
  }
 }
 if ((label|0) == 16) {
  return ($$0|0);
 }
 $21 = ($2>>>0)<($limit>>>0);
 L10: do {
  if ($21) {
   $$0115 = $2;$42 = $$lcssa60;$p$214 = $$lcssa58;
   L11: while(1) {
    $38 = ((($$0115)) + 1|0);
    $39 = HEAP8[$$0115>>0]|0;
    $40 = $39 << 24 >> 24;
    $41 = ($42<<24>>24)<(0);
    $43 = ((($p$214)) + 1|0);
    $44 = HEAP8[$43>>0]|0;
    $45 = $44&255;
    $46 = $45 & 127;
    L13: do {
     if ($41) {
      $47 = ($40|0)==($46|0);
      if ($47) {
       $69 = $44;$p$2$be = $43;
      } else {
       $$0 = 0;
       label = 16;
       break L11;
      }
     } else {
      $49 = ($46|0)==(0);
      if ($49) {
       $$0 = 0;
       label = 16;
       break L11;
      }
      $50 = $45 >>> 6;
      $51 = $50 | 1;
      $52 = (($p$214) + ($51)|0);
      $53 = ((($52)) + 1|0);
      $count$012 = $46;$p$411 = $53;
      while(1) {
       $54 = HEAP8[$p$411>>0]|0;
       $55 = $54&255;
       $56 = $55 << 8;
       $57 = ((($p$411)) + 1|0);
       $58 = HEAP8[$57>>0]|0;
       $59 = $58&255;
       $60 = $56 | $59;
       $61 = (533235 + ($60)|0);
       $62 = HEAP8[$61>>0]|0;
       $63 = $62&255;
       $64 = $63 & 127;
       $65 = ($40|0)==($64|0);
       if ($65) {
        $69 = $62;$p$2$be = $61;
        break L13;
       }
       $66 = (($count$012) + -1)|0;
       $67 = ((($p$411)) + 2|0);
       $68 = ($count$012|0)>(1);
       if ($68) {
        $count$012 = $66;$p$411 = $67;
       } else {
        $$0 = 0;
        label = 16;
        break L11;
       }
      }
     }
    } while(0);
    $48 = ($38>>>0)<($limit>>>0);
    if ($48) {
     $$0115 = $38;$42 = $69;$p$214 = $p$2$be;
    } else {
     $26 = $69;$p$2$lcssa = $p$2$be;
     break L10;
    }
   }
   if ((label|0) == 16) {
    return ($$0|0);
   }
  } else {
   $26 = $$lcssa60;$p$2$lcssa = $$lcssa58;
  }
 } while(0);
 $25 = ($26<<24>>24)>(-1);
 if (!($25)) {
  $$0 = 0;
  return ($$0|0);
 }
 $27 = ((($p$2$lcssa)) + 1|0);
 $28 = HEAP8[$27>>0]|0;
 $29 = ($28<<24>>24)<(0);
 if (!($29)) {
  $$0 = 0;
  return ($$0|0);
 }
 $30 = ((($p$2$lcssa)) + 2|0);
 $31 = HEAP8[$30>>0]|0;
 $32 = $31&255;
 $33 = $32 << 8;
 $34 = ((($p$2$lcssa)) + 3|0);
 $35 = HEAP8[$34>>0]|0;
 $36 = $35&255;
 $37 = $33 | $36;
 $$0 = $37;
 return ($$0|0);
}
function _ps_unicodes_init($memory,$table,$num_glyphs,$get_glyph_name,$free_glyph_name,$glyph_data) {
 $memory = $memory|0;
 $table = $table|0;
 $num_glyphs = $num_glyphs|0;
 $get_glyph_name = $get_glyph_name|0;
 $free_glyph_name = $free_glyph_name|0;
 $glyph_data = $glyph_data|0;
 var $$pre = 0, $$pre18 = 0, $$pre18$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $error = 0, $exitcond8 = 0, $extra_glyph_list_states = 0, $extra_glyphs = 0, $map$04 = 0, $map$1 = 0, $map$2 = 0;
 var $map$2$lcssa = 0, $map$4 = 0, $map$4$1 = 0, $map$4$2 = 0, $map$4$3 = 0, $map$4$4 = 0, $map$4$5 = 0, $map$4$6 = 0, $map$4$7 = 0, $map$4$8 = 0, $map$4$9 = 0, $n$02$lcssa$i = 0, $n$02$lcssa$i1 = 0, $n$05 = 0, $phitmp = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp + 80|0;
 $extra_glyph_list_states = sp + 40|0;
 $extra_glyphs = sp;
 dest=$extra_glyph_list_states; stop=dest+40|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $0 = ((($table)) + 16|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($table)) + 20|0);
 HEAP32[$1>>2] = 0;
 $2 = (($num_glyphs) + 10)|0;
 $3 = (_ft_mem_realloc($memory,8,0,$2,0,$error)|0);
 HEAP32[$1>>2] = $3;
 $4 = HEAP32[$error>>2]|0;
 $5 = ($4|0)==(0);
 if (!($5)) {
  $65 = $4;
  STACKTOP = sp;return ($65|0);
 }
 $6 = ($num_glyphs|0)==(0);
 if ($6) {
  $map$4 = $3;
 } else {
  $7 = ($free_glyph_name|0)==(0|0);
  $map$04 = $3;$n$05 = 0;
  while(1) {
   $8 = (FUNCTION_TABLE_iii[$get_glyph_name & 8191]($glyph_data,$n$05)|0);
   $9 = ($8|0)==(0|0);
   if ($9) {
    $map$2 = $map$04;
   } else {
    $10 = (_strcmp(589232,$8)|0);
    $11 = ($10|0)==(0);
    if ($11) {
     $n$02$lcssa$i = 0;
     label = 6;
    } else {
     $16 = (_strcmp((589238),$8)|0);
     $17 = ($16|0)==(0);
     if ($17) {
      $n$02$lcssa$i = 1;
      label = 6;
     } else {
      $18 = (_strcmp((589244),$8)|0);
      $19 = ($18|0)==(0);
      if ($19) {
       $n$02$lcssa$i = 2;
       label = 6;
      } else {
       $20 = (_strcmp((589253),$8)|0);
       $21 = ($20|0)==(0);
       if ($21) {
        $n$02$lcssa$i = 3;
        label = 6;
       } else {
        $22 = (_strcmp((589260),$8)|0);
        $23 = ($22|0)==(0);
        if ($23) {
         $n$02$lcssa$i = 4;
         label = 6;
        } else {
         $24 = (_strcmp((589267),$8)|0);
         $25 = ($24|0)==(0);
         if ($25) {
          $n$02$lcssa$i = 5;
          label = 6;
         } else {
          $26 = (_strcmp((589270),$8)|0);
          $27 = ($26|0)==(0);
          if ($27) {
           $n$02$lcssa$i = 6;
           label = 6;
          } else {
           $28 = (_strcmp((589285),$8)|0);
           $29 = ($28|0)==(0);
           if ($29) {
            $n$02$lcssa$i = 7;
            label = 6;
           } else {
            $30 = (_strcmp((589291),$8)|0);
            $31 = ($30|0)==(0);
            if ($31) {
             $n$02$lcssa$i = 8;
             label = 6;
            } else {
             $32 = (_strcmp((589304),$8)|0);
             $33 = ($32|0)==(0);
             if ($33) {
              $n$02$lcssa$i = 9;
              label = 6;
             }
            }
           }
          }
         }
        }
       }
      }
     }
    }
    if ((label|0) == 6) {
     label = 0;
     $12 = (($extra_glyph_list_states) + ($n$02$lcssa$i<<2)|0);
     $13 = HEAP32[$12>>2]|0;
     $14 = ($13|0)==(0);
     if ($14) {
      HEAP32[$12>>2] = 1;
      $15 = (($extra_glyphs) + ($n$02$lcssa$i<<2)|0);
      HEAP32[$15>>2] = $n$05;
     }
    }
    $34 = (_ps_unicode_value($8)|0);
    $35 = $34 & 2147483647;
    $36 = ($35|0)==(0);
    if ($36) {
     $map$1 = $map$04;
    } else {
     $37 = ($34|0)==(916);
     if ($37) {
      $n$02$lcssa$i1 = 0;
      label = 19;
     } else {
      $39 = ($34|0)==(937);
      if ($39) {
       $n$02$lcssa$i1 = 1;
       label = 19;
      } else {
       $40 = ($34|0)==(8725);
       if ($40) {
        $n$02$lcssa$i1 = 2;
        label = 19;
       } else {
        $41 = ($34|0)==(173);
        if ($41) {
         $n$02$lcssa$i1 = 3;
         label = 19;
        } else {
         $42 = ($34|0)==(713);
         if ($42) {
          $n$02$lcssa$i1 = 4;
          label = 19;
         } else {
          $43 = ($34|0)==(956);
          if ($43) {
           $n$02$lcssa$i1 = 5;
           label = 19;
          } else {
           $44 = ($34|0)==(8729);
           if ($44) {
            $n$02$lcssa$i1 = 6;
            label = 19;
           } else {
            $45 = ($34|0)==(160);
            if ($45) {
             $n$02$lcssa$i1 = 7;
             label = 19;
            } else {
             $46 = ($34|0)==(538);
             if ($46) {
              $n$02$lcssa$i1 = 8;
              label = 19;
             } else {
              $47 = ($34|0)==(539);
              if ($47) {
               $n$02$lcssa$i1 = 9;
               label = 19;
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
     if ((label|0) == 19) {
      label = 0;
      $38 = (($extra_glyph_list_states) + ($n$02$lcssa$i1<<2)|0);
      HEAP32[$38>>2] = 2;
     }
     HEAP32[$map$04>>2] = $34;
     $48 = ((($map$04)) + 4|0);
     HEAP32[$48>>2] = $n$05;
     $49 = ((($map$04)) + 8|0);
     $map$1 = $49;
    }
    if ($7) {
     $map$2 = $map$1;
    } else {
     FUNCTION_TABLE_vii[$free_glyph_name & 8191]($glyph_data,$8);
     $map$2 = $map$1;
    }
   }
   $50 = (($n$05) + 1)|0;
   $exitcond8 = ($50|0)==($num_glyphs|0);
   if ($exitcond8) {
    $map$2$lcssa = $map$2;
    break;
   } else {
    $map$04 = $map$2;$n$05 = $50;
   }
  }
  $$pre = HEAP32[$extra_glyph_list_states>>2]|0;
  $phitmp = ($$pre|0)==(1);
  if ($phitmp) {
   HEAP32[$map$2$lcssa>>2] = 916;
   $51 = HEAP32[$extra_glyphs>>2]|0;
   $52 = ((($map$2$lcssa)) + 4|0);
   HEAP32[$52>>2] = $51;
   $53 = ((($map$2$lcssa)) + 8|0);
   $map$4 = $53;
  } else {
   $map$4 = $map$2$lcssa;
  }
 }
 $54 = ((($extra_glyph_list_states)) + 4|0);
 $55 = HEAP32[$54>>2]|0;
 $56 = ($55|0)==(1);
 if ($56) {
  HEAP32[$map$4>>2] = 937;
  $66 = ((($extra_glyphs)) + 4|0);
  $67 = HEAP32[$66>>2]|0;
  $68 = ((($map$4)) + 4|0);
  HEAP32[$68>>2] = $67;
  $69 = ((($map$4)) + 8|0);
  $map$4$1 = $69;
 } else {
  $map$4$1 = $map$4;
 }
 $70 = ((($extra_glyph_list_states)) + 8|0);
 $71 = HEAP32[$70>>2]|0;
 $72 = ($71|0)==(1);
 if ($72) {
  HEAP32[$map$4$1>>2] = 8725;
  $73 = ((($extra_glyphs)) + 8|0);
  $74 = HEAP32[$73>>2]|0;
  $75 = ((($map$4$1)) + 4|0);
  HEAP32[$75>>2] = $74;
  $76 = ((($map$4$1)) + 8|0);
  $map$4$2 = $76;
 } else {
  $map$4$2 = $map$4$1;
 }
 $77 = ((($extra_glyph_list_states)) + 12|0);
 $78 = HEAP32[$77>>2]|0;
 $79 = ($78|0)==(1);
 if ($79) {
  HEAP32[$map$4$2>>2] = 173;
  $80 = ((($extra_glyphs)) + 12|0);
  $81 = HEAP32[$80>>2]|0;
  $82 = ((($map$4$2)) + 4|0);
  HEAP32[$82>>2] = $81;
  $83 = ((($map$4$2)) + 8|0);
  $map$4$3 = $83;
 } else {
  $map$4$3 = $map$4$2;
 }
 $84 = ((($extra_glyph_list_states)) + 16|0);
 $85 = HEAP32[$84>>2]|0;
 $86 = ($85|0)==(1);
 if ($86) {
  HEAP32[$map$4$3>>2] = 713;
  $87 = ((($extra_glyphs)) + 16|0);
  $88 = HEAP32[$87>>2]|0;
  $89 = ((($map$4$3)) + 4|0);
  HEAP32[$89>>2] = $88;
  $90 = ((($map$4$3)) + 8|0);
  $map$4$4 = $90;
 } else {
  $map$4$4 = $map$4$3;
 }
 $91 = ((($extra_glyph_list_states)) + 20|0);
 $92 = HEAP32[$91>>2]|0;
 $93 = ($92|0)==(1);
 if ($93) {
  HEAP32[$map$4$4>>2] = 956;
  $94 = ((($extra_glyphs)) + 20|0);
  $95 = HEAP32[$94>>2]|0;
  $96 = ((($map$4$4)) + 4|0);
  HEAP32[$96>>2] = $95;
  $97 = ((($map$4$4)) + 8|0);
  $map$4$5 = $97;
 } else {
  $map$4$5 = $map$4$4;
 }
 $98 = ((($extra_glyph_list_states)) + 24|0);
 $99 = HEAP32[$98>>2]|0;
 $100 = ($99|0)==(1);
 if ($100) {
  HEAP32[$map$4$5>>2] = 8729;
  $101 = ((($extra_glyphs)) + 24|0);
  $102 = HEAP32[$101>>2]|0;
  $103 = ((($map$4$5)) + 4|0);
  HEAP32[$103>>2] = $102;
  $104 = ((($map$4$5)) + 8|0);
  $map$4$6 = $104;
 } else {
  $map$4$6 = $map$4$5;
 }
 $105 = ((($extra_glyph_list_states)) + 28|0);
 $106 = HEAP32[$105>>2]|0;
 $107 = ($106|0)==(1);
 if ($107) {
  HEAP32[$map$4$6>>2] = 160;
  $108 = ((($extra_glyphs)) + 28|0);
  $109 = HEAP32[$108>>2]|0;
  $110 = ((($map$4$6)) + 4|0);
  HEAP32[$110>>2] = $109;
  $111 = ((($map$4$6)) + 8|0);
  $map$4$7 = $111;
 } else {
  $map$4$7 = $map$4$6;
 }
 $112 = ((($extra_glyph_list_states)) + 32|0);
 $113 = HEAP32[$112>>2]|0;
 $114 = ($113|0)==(1);
 if ($114) {
  HEAP32[$map$4$7>>2] = 538;
  $115 = ((($extra_glyphs)) + 32|0);
  $116 = HEAP32[$115>>2]|0;
  $117 = ((($map$4$7)) + 4|0);
  HEAP32[$117>>2] = $116;
  $118 = ((($map$4$7)) + 8|0);
  $map$4$8 = $118;
 } else {
  $map$4$8 = $map$4$7;
 }
 $119 = ((($extra_glyph_list_states)) + 36|0);
 $120 = HEAP32[$119>>2]|0;
 $121 = ($120|0)==(1);
 if ($121) {
  HEAP32[$map$4$8>>2] = 539;
  $122 = ((($extra_glyphs)) + 36|0);
  $123 = HEAP32[$122>>2]|0;
  $124 = ((($map$4$8)) + 4|0);
  HEAP32[$124>>2] = $123;
  $125 = ((($map$4$8)) + 8|0);
  $map$4$9 = $125;
 } else {
  $map$4$9 = $map$4$8;
 }
 $57 = HEAP32[$1>>2]|0;
 $126 = $map$4$9;
 $127 = $57;
 $128 = (($126) - ($127))|0;
 $62 = $128 >> 3;
 $129 = ($62|0)==(0);
 if ($129) {
  _ft_mem_free($memory,$57);
  HEAP32[$1>>2] = 0;
  $58 = HEAP32[$error>>2]|0;
  $59 = ($58|0)==(0);
  if ($59) {
   HEAP32[$error>>2] = 163;
   $$pre18 = 163;
  } else {
   $$pre18 = $58;
  }
 } else {
  $60 = $num_glyphs >>> 1;
  $61 = ($62>>>0)<($60>>>0);
  if ($61) {
   $63 = (_ft_mem_realloc($memory,8,$num_glyphs,$62,$57,$error)|0);
   HEAP32[$1>>2] = $63;
   HEAP32[$error>>2] = 0;
   $64 = $63;
  } else {
   $64 = $57;
  }
  _qsort($64,$62,8,4381);
  $$pre18$pre = HEAP32[$error>>2]|0;
  $$pre18 = $$pre18$pre;
 }
 HEAP32[$0>>2] = $62;
 $65 = $$pre18;
 STACKTOP = sp;return ($65|0);
}
function _compare_uni_maps($a,$b) {
 $a = $a|0;
 $b = $b|0;
 var $$ = 0, $$0 = 0, $$1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$a>>2]|0;
 $1 = $0 & 2147483647;
 $2 = HEAP32[$b>>2]|0;
 $3 = $2 & 2147483647;
 $4 = ($1|0)==($3|0);
 if ($4) {
  $5 = ($0>>>0)>($2>>>0);
  if ($5) {
   $$0 = 1;
   return ($$0|0);
  }
  $6 = ($0>>>0)<($2>>>0);
  $$ = $6 << 31 >> 31;
  $$0 = $$;
  return ($$0|0);
 } else {
  $7 = ($1>>>0)>($3>>>0);
  if ($7) {
   $$0 = 1;
   return ($$0|0);
  }
  $8 = ($1>>>0)<($3>>>0);
  $$1 = $8 << 31 >> 31;
  $$0 = $$1;
  return ($$0|0);
 }
 return (0)|0;
}
function _ps_unicodes_char_index($table,$unicode) {
 $table = $table|0;
 $unicode = $unicode|0;
 var $$01 = 0, $$result$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $max$02 = 0, $max$1 = 0, $min$03 = 0, $min$1 = 0, $result$04 = 0, $result$3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($table)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($table)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3<<3)|0);
 $5 = ((($4)) + -8|0);
 $6 = ($1>>>0)>($5>>>0);
 if ($6) {
  $$01 = 0;
  return ($$01|0);
 } else {
  $max$02 = $5;$min$03 = $1;$result$04 = 0;
 }
 while(1) {
  $7 = $max$02;
  $8 = $min$03;
  $9 = (($7) - ($8))|0;
  $10 = $9 >> 4;
  $11 = (($min$03) + ($10<<3)|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ($12|0)==($unicode|0);
  if ($13) {
   $result$3 = $11;
   break;
  }
  $14 = $12 & 2147483647;
  $15 = ($14|0)==($unicode|0);
  $$result$0 = $15 ? $11 : $result$04;
  $16 = ($min$03|0)==($max$02|0);
  if ($16) {
   $result$3 = $$result$0;
   break;
  }
  $17 = ($14>>>0)<($unicode>>>0);
  $18 = ((($11)) + 8|0);
  $19 = ((($11)) + -8|0);
  $max$1 = $17 ? $max$02 : $19;
  $min$1 = $17 ? $18 : $min$03;
  $20 = ($min$1>>>0)>($max$1>>>0);
  if ($20) {
   $result$3 = $$result$0;
   break;
  } else {
   $max$02 = $max$1;$min$03 = $min$1;$result$04 = $$result$0;
  }
 }
 $21 = ($result$3|0)==(0|0);
 if ($21) {
  $$01 = 0;
  return ($$01|0);
 }
 $22 = ((($result$3)) + 4|0);
 $23 = HEAP32[$22>>2]|0;
 $$01 = $23;
 return ($$01|0);
}
function _ps_unicodes_char_next($table,$unicode) {
 $table = $table|0;
 $unicode = $unicode|0;
 var $$lcssa = 0, $$min$0 = 0, $$min$0$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $char_code$1 = 0, $max$0$ = 0, $max$06 = 0, $min$0$lcssa10 = 0, $min$05 = 0, $result$07 = 0;
 var $result$1 = 0, $result$1$lcssa = 0, $result$3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$unicode>>2]|0;
 $1 = (($0) + 1)|0;
 $2 = ((($table)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0);
 do {
  if ($4) {
   $min$0$lcssa10 = 0;
  } else {
   $5 = ((($table)) + 20|0);
   $6 = HEAP32[$5>>2]|0;
   $max$06 = $3;$min$05 = 0;$result$07 = 0;
   while(1) {
    $7 = (($max$06) - ($min$05))|0;
    $8 = $7 >>> 1;
    $9 = (($8) + ($min$05))|0;
    $10 = (($6) + ($9<<3)|0);
    $11 = HEAP32[$10>>2]|0;
    $12 = ($11|0)==($1|0);
    if ($12) {
     $$lcssa = $9;
     label = 4;
     break;
    }
    $15 = $11 & 2147483647;
    $16 = ($15|0)==($1|0);
    if ($16) {
     $17 = (((($6) + ($9<<3)|0)) + 4|0);
     $18 = HEAP32[$17>>2]|0;
     $result$1 = $18;
    } else {
     $result$1 = $result$07;
    }
    $19 = ($15>>>0)<($1>>>0);
    $20 = (($9) + 1)|0;
    $$min$0 = $19 ? $20 : $min$05;
    $max$0$ = $19 ? $max$06 : $9;
    $21 = ($max$0$>>>0)>($$min$0>>>0);
    if ($21) {
     $max$06 = $max$0$;$min$05 = $$min$0;$result$07 = $result$1;
    } else {
     $$min$0$lcssa = $$min$0;$result$1$lcssa = $result$1;
     label = 8;
     break;
    }
   }
   if ((label|0) == 4) {
    $13 = (((($6) + ($$lcssa<<3)|0)) + 4|0);
    $14 = HEAP32[$13>>2]|0;
    $char_code$1 = $1;$result$3 = $14;
    HEAP32[$unicode>>2] = $char_code$1;
    return ($result$3|0);
   }
   else if ((label|0) == 8) {
    $22 = ($result$1$lcssa|0)==(0);
    if ($22) {
     $min$0$lcssa10 = $$min$0$lcssa;
     break;
    } else {
     $char_code$1 = $1;$result$3 = $result$1$lcssa;
    }
    HEAP32[$unicode>>2] = $char_code$1;
    return ($result$3|0);
   }
  }
 } while(0);
 $23 = ($min$0$lcssa10>>>0)<($3>>>0);
 if (!($23)) {
  $char_code$1 = 0;$result$3 = 0;
  HEAP32[$unicode>>2] = $char_code$1;
  return ($result$3|0);
 }
 $24 = ((($table)) + 20|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = (((($25) + ($min$0$lcssa10<<3)|0)) + 4|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = (($25) + ($min$0$lcssa10<<3)|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = $29 & 2147483647;
 $char_code$1 = $30;$result$3 = $27;
 HEAP32[$unicode>>2] = $char_code$1;
 return ($result$3|0);
}
function _ps_get_macintosh_name($name_index) {
 $name_index = $name_index|0;
 var $$name_index = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($name_index>>>0)>(257);
 $$name_index = $0 ? 0 : $name_index;
 $1 = (172240 + ($$name_index<<1)|0);
 $2 = HEAP16[$1>>1]|0;
 $3 = $2 << 16 >> 16;
 $4 = (589317 + ($3)|0);
 return ($4|0);
}
function _ps_get_standard_strings($sid) {
 $sid = $sid|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($sid>>>0)>(390);
 if ($0) {
  $$0 = 0;
  return ($$0|0);
 }
 $1 = (172756 + ($sid<<1)|0);
 $2 = HEAP16[$1>>1]|0;
 $3 = $2 << 16 >> 16;
 $4 = (589317 + ($3)|0);
 $$0 = $4;
 return ($$0|0);
}
function _psnames_get_service($module,$service_id) {
 $module = $module|0;
 $service_id = $service_id|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_ft_service_list_lookup(54392,$service_id)|0);
 return ($0|0);
}
function _ft_black_new($memory,$araster) {
 $memory = $memory|0;
 $araster = $araster|0;
 var $0 = 0, $1 = 0, $2 = 0, $error = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 HEAP32[$araster>>2] = 0;
 $0 = (_ft_mem_alloc($memory,4,$error)|0);
 $1 = HEAP32[$error>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  STACKTOP = sp;return ($1|0);
 }
 HEAP32[$0>>2] = $memory;
 HEAP32[$araster>>2] = $0;
 STACKTOP = sp;return ($1|0);
}
function _ft_black_reset($raster,$pool_base,$pool_size) {
 $raster = $raster|0;
 $pool_base = $pool_base|0;
 $pool_size = $pool_size|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _ft_black_set_mode($raster,$mode,$palette) {
 $raster = $raster|0;
 $mode = $mode|0;
 $palette = $palette|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _ft_black_render($raster,$params) {
 $raster = $raster|0;
 $params = $params|0;
 var $$$i$i = 0, $$0 = 0, $$3$i$i = 0, $$4$i$i = 0, $$lobit$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $buffer = 0, $worker = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 17408|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $worker = sp + 16384|0;
 $buffer = sp;
 $0 = ((($params)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$params>>2]|0;
 $3 = ($raster|0)==(0|0);
 if ($3) {
  $$0 = 96;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ($1|0)==(0|0);
 if ($4) {
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = ((($1)) + 2|0);
 $6 = HEAP16[$5>>1]|0;
 $7 = $6 << 16 >> 16;
 $8 = ($6<<16>>16)==(0);
 if ($8) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $9 = HEAP16[$1>>1]|0;
 $10 = $9 << 16 >> 16;
 $11 = ($9<<16>>16)<(1);
 if ($11) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $12 = ((($1)) + 12|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13|0)==(0|0);
 if ($14) {
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 $15 = ((($1)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ($16|0)==(0|0);
 if ($17) {
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 $18 = (($10) + -1)|0;
 $19 = (($13) + ($18<<1)|0);
 $20 = HEAP16[$19>>1]|0;
 $21 = $20 << 16 >> 16;
 $22 = (($21) + 1)|0;
 $23 = ($7|0)==($22|0);
 if (!($23)) {
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 $24 = ((($params)) + 8|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = $25 & 3;
 $27 = ($26|0)==(0);
 if (!($27)) {
  $$0 = 19;
  STACKTOP = sp;return ($$0|0);
 }
 $28 = ($2|0)==(0|0);
 if ($28) {
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 $29 = ((($2)) + 4|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ($30|0)==(0);
 if ($31) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $32 = HEAP32[$2>>2]|0;
 $33 = ($32|0)==(0);
 if ($33) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $34 = ((($2)) + 12|0);
 $35 = HEAP32[$34>>2]|0;
 $36 = ($35|0)==(0|0);
 if ($36) {
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 $37 = ((($worker)) + 128|0);
 ;HEAP32[$37>>2]=HEAP32[$1>>2]|0;HEAP32[$37+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$37+8>>2]=HEAP32[$1+8>>2]|0;HEAP32[$37+12>>2]=HEAP32[$1+12>>2]|0;HEAP32[$37+16>>2]=HEAP32[$1+16>>2]|0;
 $38 = ((($worker)) + 104|0);
 ;HEAP32[$38>>2]=HEAP32[$2>>2]|0;HEAP32[$38+4>>2]=HEAP32[$2+4>>2]|0;HEAP32[$38+8>>2]=HEAP32[$2+8>>2]|0;HEAP32[$38+12>>2]=HEAP32[$2+12>>2]|0;HEAP32[$38+16>>2]=HEAP32[$2+16>>2]|0;HEAP32[$38+20>>2]=HEAP32[$2+20>>2]|0;
 $39 = ((($worker)) + 28|0);
 HEAP32[$39>>2] = $buffer;
 $40 = ((($buffer)) + 16384|0);
 $41 = ((($worker)) + 32|0);
 HEAP32[$41>>2] = $40;
 $42 = ((($worker)) + 144|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = $43 & 256;
 $45 = ($44|0)==(0);
 $$$i$i = $45 ? 6 : 12;
 $$3$i$i = $45 ? 32 : 256;
 $$4$i$i = $45 ? 2 : 30;
 HEAP32[$worker>>2] = $$$i$i;
 $46 = ((($worker)) + 16|0);
 HEAP32[$46>>2] = $$3$i$i;
 $47 = ((($worker)) + 20|0);
 HEAP32[$47>>2] = $$4$i$i;
 $48 = 1 << $$$i$i;
 $49 = ((($worker)) + 4|0);
 HEAP32[$49>>2] = $48;
 $50 = $48 >>> 1;
 $51 = ((($worker)) + 8|0);
 HEAP32[$51>>2] = $50;
 $52 = (($$$i$i) + -6)|0;
 $53 = ((($worker)) + 12|0);
 HEAP32[$53>>2] = $52;
 $54 = ((($worker)) + 24|0);
 HEAP32[$54>>2] = $52;
 $55 = $43 & 8;
 $56 = ($55|0)==(0);
 if ($56) {
  $58 = $43 & 16;
  $59 = ((($worker)) + 176|0);
  $60 = $58&255;
  $61 = ($60&255) >>> 2;
  HEAP8[$59>>0] = $61;
  $62 = $43 & 32;
  $63 = ($62|0)==(0);
  if ($63) {
   $64 = $61&255;
   $65 = $64 | 1;
   $66 = $65&255;
   HEAP8[$59>>0] = $66;
  }
 } else {
  $57 = ((($worker)) + 176|0);
  HEAP8[$57>>0] = 2;
 }
 $67 = $43 >>> 9;
 $$lobit$i = $67 & 1;
 $68 = $$lobit$i ^ 1;
 $69 = $68&255;
 $70 = ((($worker)) + 177|0);
 HEAP8[$70>>0] = $69;
 $71 = ((($worker)) + 160|0);
 HEAP32[$71>>2] = 4382;
 $72 = ((($worker)) + 164|0);
 HEAP32[$72>>2] = 4383;
 $73 = ((($worker)) + 168|0);
 HEAP32[$73>>2] = 4384;
 $74 = ((($worker)) + 172|0);
 HEAP32[$74>>2] = 4385;
 $75 = ((($worker)) + 1020|0);
 HEAP32[$75>>2] = 0;
 $76 = ((($worker)) + 956|0);
 HEAP16[$76>>1] = 0;
 $77 = ((($worker)) + 104|0);
 $78 = HEAP32[$77>>2]|0;
 $79 = (($78) + 65535)|0;
 $80 = $79&65535;
 $81 = ((($worker)) + 958|0);
 HEAP16[$81>>1] = $80;
 $82 = ((($worker)) + 108|0);
 $83 = HEAP32[$82>>2]|0;
 $84 = $83&65535;
 $85 = ((($worker)) + 56|0);
 HEAP16[$85>>1] = $84;
 $86 = ((($worker)) + 116|0);
 $87 = HEAP32[$86>>2]|0;
 $88 = ((($worker)) + 60|0);
 HEAP32[$88>>2] = $87;
 $89 = (_Render_Single_Pass($worker,0)|0);
 $90 = ($89|0)==(0);
 if (!($90)) {
  $$0 = $89;
  STACKTOP = sp;return ($$0|0);
 }
 $91 = HEAP8[$70>>0]|0;
 $92 = ($91<<24>>24)==(0);
 if (!($92)) {
  $93 = ((($worker)) + 176|0);
  $94 = HEAP8[$93>>0]|0;
  $95 = ($94<<24>>24)==(2);
  if (!($95)) {
   HEAP32[$71>>2] = 4386;
   HEAP32[$72>>2] = 4387;
   HEAP32[$73>>2] = 4388;
   HEAP32[$74>>2] = 4389;
   HEAP32[$75>>2] = 0;
   HEAP16[$76>>1] = 0;
   $96 = HEAP32[$82>>2]|0;
   $97 = (($96) + 65535)|0;
   $98 = $97&65535;
   HEAP16[$81>>1] = $98;
   $99 = (_Render_Single_Pass($worker,1)|0);
   $100 = ($99|0)==(0);
   if (!($100)) {
    $$0 = $99;
    STACKTOP = sp;return ($$0|0);
   }
  }
 }
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _Vertical_Sweep_Init($worker,$min,$max) {
 $worker = $worker|0;
 $min = $min|0;
 $max = $max|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($worker)) + 112|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (0 - ($1))|0;
 $3 = $2&65535;
 $4 = ((($worker)) + 156|0);
 HEAP16[$4>>1] = $3;
 $5 = HEAP16[$min>>1]|0;
 $6 = $5 << 16 >> 16;
 $7 = Math_imul($6, $2)|0;
 $8 = ((($worker)) + 148|0);
 HEAP32[$8>>2] = $7;
 $9 = ($1|0)>(0);
 if (!($9)) {
  return;
 }
 $10 = ((($worker)) + 104|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + -1)|0;
 $13 = Math_imul($12, $1)|0;
 $14 = (($13) + ($7))|0;
 HEAP32[$8>>2] = $14;
 return;
}
function _Vertical_Sweep_Span($worker,$y,$x1,$x2,$left,$right) {
 $worker = $worker|0;
 $y = $y|0;
 $x1 = $x1|0;
 $x2 = $x2|0;
 $left = $left|0;
 $right = $right|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $e2$0 = 0, $e2$1 = 0, $scevgep = 0, $scevgep3 = 0, $target$0$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($left)) + 12|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = ((($worker)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (0 - ($3))|0;
 $5 = (($x1) + -1)|0;
 $6 = (($5) + ($3))|0;
 $7 = $6 & $4;
 $8 = HEAP32[$worker>>2]|0;
 $9 = $7 >> $8;
 $10 = $1 & 7;
 $11 = ($10<<16>>16)==(2);
 if ($11) {
  label = 3;
 } else {
  $12 = (($x2) - ($x1))|0;
  $13 = (($12) - ($3))|0;
  $14 = ((($worker)) + 20|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ($13|0)>($15|0);
  if ($16) {
   label = 3;
  } else {
   $e2$0 = $9;
  }
 }
 if ((label|0) == 3) {
  $17 = $4 & $x2;
  $18 = $17 >> $8;
  $e2$0 = $18;
 }
 $19 = ($e2$0|0)>(-1);
 if (!($19)) {
  return;
 }
 $20 = ((($worker)) + 56|0);
 $21 = HEAP16[$20>>1]|0;
 $22 = $21&65535;
 $23 = ($9|0)<($22|0);
 if (!($23)) {
  return;
 }
 $24 = ($9|0)<(0);
 $$ = $24 ? 0 : $9;
 $25 = ($e2$0|0)<($22|0);
 $26 = (($22) + -1)|0;
 $e2$1 = $25 ? $e2$0 : $26;
 $27 = $$ << 13;
 $28 = $27 >> 16;
 $29 = $e2$1 << 13;
 $30 = $29 >> 16;
 $31 = $$ & 7;
 $32 = 255 >>> $31;
 $33 = $e2$1 & 7;
 $34 = 127 >>> $33;
 $35 = $34 ^ 255;
 $36 = ((($worker)) + 60|0);
 $37 = HEAP32[$36>>2]|0;
 $38 = ((($worker)) + 148|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = (($37) + ($39)|0);
 $41 = (($40) + ($28)|0);
 $42 = (($30) - ($28))|0;
 $43 = ($42|0)>(0);
 if (!($43)) {
  $60 = $35 & $32;
  $61 = HEAP8[$41>>0]|0;
  $62 = $61&255;
  $63 = $62 | $60;
  $64 = $63&255;
  HEAP8[$41>>0] = $64;
  return;
 }
 $44 = HEAP8[$41>>0]|0;
 $45 = $44&255;
 $46 = $45 | $32;
 $47 = $46&255;
 HEAP8[$41>>0] = $47;
 $48 = ($42|0)>(1);
 if ($48) {
  $49 = (($39) + ($30))|0;
  $50 = (($49) + -1)|0;
  $51 = (($39) + ($28))|0;
  $52 = (($51) + 1)|0;
  $scevgep3 = (($37) + ($52)|0);
  $53 = (($30) + -1)|0;
  $54 = (($53) - ($28))|0;
  _memset(($scevgep3|0),-1,($54|0))|0;
  $scevgep = (($37) + ($50)|0);
  $target$0$lcssa = $scevgep;
 } else {
  $target$0$lcssa = $41;
 }
 $55 = ((($target$0$lcssa)) + 1|0);
 $56 = HEAP8[$55>>0]|0;
 $57 = $56&255;
 $58 = $57 | $35;
 $59 = $58&255;
 HEAP8[$55>>0] = $59;
 return;
}
function _Vertical_Sweep_Drop($worker,$y,$x1,$x2,$left,$right) {
 $worker = $worker|0;
 $y = $y|0;
 $x1 = $x1|0;
 $x2 = $x2|0;
 $left = $left|0;
 $right = $right|0;
 var $$pre = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $pxl$0 = 0, $pxl$0$ = 0, $pxl$1 = 0, $pxl$3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($worker)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($x1) + -1)|0;
 $3 = (($2) + ($1))|0;
 $4 = (0 - ($1))|0;
 $5 = $3 & $4;
 $6 = $4 & $x2;
 $7 = ($5|0)>($6|0);
 if ($7) {
  $8 = ((($left)) + 12|0);
  $9 = HEAP16[$8>>1]|0;
  $10 = $9&65535;
  $11 = $10 & 7;
  $12 = (($6) + ($1))|0;
  $13 = ($5|0)==($12|0);
  if (!($13)) {
   return;
  }
  switch ($11|0) {
  case 0:  {
   $pxl$0 = $6;
   break;
  }
  case 4:  {
   $14 = (($2) + ($x2))|0;
   $15 = (($14|0) / 2)&-1;
   $16 = ((($worker)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = (($17) + ($15))|0;
   $19 = $18 & $4;
   $pxl$0 = $19;
   break;
  }
  case 5: case 1:  {
   $20 = ((($left)) + 28|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = ($21|0)==($right|0);
   if ($22) {
    $23 = ((($left)) + 16|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = ($24|0)<(1);
    if ($25) {
     $26 = $9 & 16;
     $27 = ($26<<16>>16)==(0);
     if ($27) {
      return;
     }
     $28 = (($x2) - ($x1))|0;
     $29 = ((($worker)) + 8|0);
     $30 = HEAP32[$29>>2]|0;
     $31 = ($28|0)<($30|0);
     if ($31) {
      return;
     }
    }
   }
   $32 = ((($right)) + 28|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ($33|0)==($left|0);
   if ($34) {
    $35 = ((($left)) + 20|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = $y << 16 >> 16;
    $38 = ($36|0)==($37|0);
    if ($38) {
     $39 = $9 & 32;
     $40 = ($39<<16>>16)==(0);
     if ($40) {
      return;
     }
     $41 = (($x2) - ($x1))|0;
     $42 = ((($worker)) + 8|0);
     $43 = HEAP32[$42>>2]|0;
     $44 = ($41|0)<($43|0);
     if ($44) {
      return;
     }
    }
   }
   $45 = ($11|0)==(1);
   if ($45) {
    $pxl$0 = $6;
   } else {
    $46 = (($2) + ($x2))|0;
    $47 = (($46|0) / 2)&-1;
    $48 = ((($worker)) + 8|0);
    $49 = HEAP32[$48>>2]|0;
    $50 = (($49) + ($47))|0;
    $51 = $50 & $4;
    $pxl$0 = $51;
   }
   break;
  }
  default: {
   return;
  }
  }
  $52 = ($pxl$0|0)<(0);
  $$pre = HEAP32[$worker>>2]|0;
  if ($52) {
   $pxl$1 = $5;
  } else {
   $53 = $pxl$0 >> $$pre;
   $54 = ((($worker)) + 56|0);
   $55 = HEAP16[$54>>1]|0;
   $56 = $55&65535;
   $57 = ($53|0)<($56|0);
   $pxl$0$ = $57 ? $pxl$0 : $6;
   $pxl$1 = $pxl$0$;
  }
  $58 = ($pxl$1|0)==($5|0);
  $59 = $58 ? $6 : $5;
  $60 = $59 >> $$pre;
  $61 = $60 & 7;
  $62 = ($60|0)>(-1);
  if ($62) {
   $63 = ((($worker)) + 56|0);
   $64 = HEAP16[$63>>1]|0;
   $65 = $64&65535;
   $66 = ($60|0)<($65|0);
   if ($66) {
    $67 = ((($worker)) + 148|0);
    $68 = HEAP32[$67>>2]|0;
    $69 = $60 << 13;
    $70 = $69 >> 16;
    $71 = (($68) + ($70))|0;
    $72 = ((($worker)) + 60|0);
    $73 = HEAP32[$72>>2]|0;
    $74 = (($73) + ($71)|0);
    $75 = HEAP8[$74>>0]|0;
    $76 = $75&255;
    $77 = 128 >>> $61;
    $78 = $76 & $77;
    $79 = ($78|0)==(0);
    if ($79) {
     $81 = $$pre;$pxl$3 = $pxl$1;
    } else {
     return;
    }
   } else {
    $81 = $$pre;$pxl$3 = $pxl$1;
   }
  } else {
   $81 = $$pre;$pxl$3 = $pxl$1;
  }
 } else {
  $$pre5 = HEAP32[$worker>>2]|0;
  $81 = $$pre5;$pxl$3 = $5;
 }
 $80 = $pxl$3 >> $81;
 $82 = ($80|0)>(-1);
 if (!($82)) {
  return;
 }
 $83 = ((($worker)) + 56|0);
 $84 = HEAP16[$83>>1]|0;
 $85 = $84&65535;
 $86 = ($80|0)<($85|0);
 if (!($86)) {
  return;
 }
 $87 = $80 & 7;
 $88 = 128 >>> $87;
 $89 = ((($worker)) + 148|0);
 $90 = HEAP32[$89>>2]|0;
 $91 = $80 << 13;
 $92 = $91 >> 16;
 $93 = (($90) + ($92))|0;
 $94 = ((($worker)) + 60|0);
 $95 = HEAP32[$94>>2]|0;
 $96 = (($95) + ($93)|0);
 $97 = HEAP8[$96>>0]|0;
 $98 = $97&255;
 $99 = $98 | $88;
 $100 = $99&255;
 HEAP8[$96>>0] = $100;
 return;
}
function _Vertical_Sweep_Step($worker) {
 $worker = $worker|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($worker)) + 156|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = $1 << 16 >> 16;
 $3 = ((($worker)) + 148|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + ($2))|0;
 HEAP32[$3>>2] = $5;
 return;
}
function _Render_Single_Pass($worker,$flipped) {
 $worker = $worker|0;
 $flipped = $flipped|0;
 var $$$i = 0, $$$i$i = 0, $$0 = 0, $$1$i$i = 0, $$1$i$i$i = 0, $$1$i13$i$i = 0, $$1$i1344$i$i = 0, $$1$i43$i$i = 0, $$1$us$i$i = 0, $$10$v_control$sroa$0$0$i$i = 0, $$10$v_control$sroa$0$0$us$i$i = 0, $$11$v_control$sroa$11$0$i$i = 0, $$11$v_control$sroa$11$0$us$i$i = 0, $$12$i$i = 0, $$12$us$i$i = 0, $$13$i$i = 0, $$13$us$i$i = 0, $$14$i$i = 0, $$15$i$i = 0, $$16$i$i = 0;
 var $$17$i$i = 0, $$18$i$i = 0, $$19$i$i = 0, $$2$i$i = 0, $$20$i$i = 0, $$21$i$i = 0, $$3$i$i = 0, $$4$i = 0, $$4$i$i = 0, $$5$i$i = 0, $$6$i$i = 0, $$7$i$i = 0, $$lcssa = 0, $$lcssa345 = 0, $$lcssa346 = 0, $$lcssa349 = 0, $$lcssa352 = 0, $$lcssa353 = 0, $$lcssa356 = 0, $$lcssa360 = 0;
 var $$lcssa362 = 0, $$not$i = 0, $$old1$i$i$i = 0, $$old1$i27$i$i = 0, $$old1$i8$i$i = 0, $$old1$i829$i$i = 0, $$phi$trans$insert$i$i$i = 0, $$phi$trans$insert$i10$i$i = 0, $$pr = 0, $$pr$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i$i = 0, $$pre$i1 = 0, $$pre$i11$i$i = 0, $$pre111$i = 0, $$pre112$i = 0, $$pre113$i = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0;
 var $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0;
 var $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0;
 var $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0;
 var $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0;
 var $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0;
 var $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0;
 var $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0;
 var $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0;
 var $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0;
 var $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0;
 var $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0;
 var $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0;
 var $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0;
 var $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0;
 var $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0;
 var $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0;
 var $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0;
 var $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0;
 var $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0;
 var $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0;
 var $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0;
 var $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0;
 var $658 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $P$0134$i = 0, $P$1$i = 0;
 var $P$1130$i = 0, $P$2123$i = 0, $P$3108$i = 0, $P$4107$i = 0, $P_Left$0$i = 0, $P_Left$0113$i = 0, $P_Left$1$i = 0, $P_Left$1118$i = 0, $P_Right$0114$i = 0, $P_Right$0114$i$pre = 0, $P_Right$0114$i142 = 0, $P_Right$0114$i143 = 0, $P_Right$0114$i144 = 0, $P_Right$0114$i146 = 0, $P_Right$0114$i147 = 0, $P_Right$0114$i148 = 0, $P_Right$1119$i = 0, $P_Right$1119$in$i = 0, $bottom$0$i$i = 0, $current$0$i$i = 0;
 var $current$0$i11$i = 0, $current$0$i14$i = 0, $current$0$i33$i = 0, $current$0$i36$i = 0, $current$0$i40$i = 0, $current$0$i60$i = 0, $current$0$i82$i = 0, $current$0$i9$i = 0, $current$0$i91$i = 0, $current$0$in$lcssa$i$i = 0, $current$0$in$lcssa$i84$i = 0, $current$0$in$lcssa$i93$i = 0, $current$0$in2$i$i = 0, $current$0$in2$i80$i = 0, $current$0$in2$i89$i = 0, $current$0$lcssa$i$i = 0, $current$0$lcssa$i83$i = 0, $current$0$lcssa$i92$i = 0, $current$03$i$i = 0, $current$03$i79$i = 0;
 var $current$03$i88$i = 0, $current$08$i$i = 0, $current$08$i13$i = 0, $current$08$i39$i = 0, $current$08$i59$i = 0, $current$1$be$i$i = 0, $current$1$be$i27$i = 0, $current$1$be$i53$i = 0, $current$1$be$i73$i = 0, $current$14$i$i = 0, $current$14$i23$i = 0, $current$14$i49$i = 0, $current$14$i69$i = 0, $draw_left$i = 0, $draw_right$i = 0, $dropouts$0112$i = 0, $dropouts$2$i = 0, $dropouts$2$i$lcssa = 0, $i$060$i = 0, $limit$0$i$i = 0;
 var $limit$1$i$i = 0, $max_Y$i = 0, $min_Y$i = 0, $n$0$i$i$i = 0, $n$0$i$i$i$lcssa348 = 0, $n$0$i4$i$i = 0, $n$0$i4$i$i$lcssa351 = 0, $n$0$in$i$i$i = 0, $n$0$in$i$i$i$lcssa347 = 0, $n$0$in$i3$i$i = 0, $n$0$in$i3$i$i$lcssa350 = 0, $n$032$i$i = 0, $n$2$i$i$i = 0, $n$2$i14$i$i = 0, $next$0$i$i = 0, $next$0$i29$i = 0, $next$0$i55$i = 0, $next$0$i75$i = 0, $next$0$in$i$i = 0, $next$0$in$i28$i = 0;
 var $next$0$in$i54$i = 0, $next$0$in$i74$i = 0, $next$0$in1$i$i = 0, $next$0$in1$i18$i = 0, $next$0$in1$i44$i = 0, $next$0$in1$i64$i = 0, $next$0$in5$i$i = 0, $next$0$in5$i22$i = 0, $next$0$in5$i48$i = 0, $next$0$in5$i68$i = 0, $next$02$i$i = 0, $next$02$i19$i = 0, $next$02$i45$i = 0, $next$02$i65$i = 0, $next$06$i$i = 0, $next$06$i21$i = 0, $next$06$i47$i = 0, $next$06$i67$i = 0, $not$$i = 0, $o$0$in$i = 0;
 var $old$0$be$i$i = 0, $old$0$be$i26$i = 0, $old$0$be$i52$i = 0, $old$0$be$i72$i = 0, $old$0$i$i = 0, $old$0$i$i$lcssa361 = 0, $old$0$i35$i = 0, $old$0$i35$i$lcssa355 = 0, $old$0$i8$i = 0, $old$0$i8$i$lcssa359 = 0, $old$03$i$i = 0, $old$03$i24$i = 0, $old$03$i50$i = 0, $old$03$i70$i = 0, $or$cond = 0, $or$cond$i = 0, $or$cond$i$i = 0, $or$cond5$i = 0, $or$cond7$i = 0, $p$033$i$i = 0;
 var $point$0$i$i = 0, $point$1$be$i$i = 0, $point$143$i$i = 0, $point$2$i$i = 0, $point$2$us$i$i = 0, $sext$i = 0, $sext1$i = 0, $sext2$i = 0, $sext3$i = 0, $start$061$i = 0, $storemerge = 0, $tags$0$i$i = 0, $tags$1$be$i$i = 0, $tags$142$i$i = 0, $tags$2$i$i = 0, $tags$2$us$i$i = 0, $top$0$i$i = 0, $v_control$sroa$0$0$i$i = 0, $v_control$sroa$0$0$lcssa$i$i = 0, $v_control$sroa$0$0$us$i$i = 0;
 var $v_control$sroa$0$1$i$i = 0, $v_control$sroa$0$1$us$i$i = 0, $v_control$sroa$11$0$i$i = 0, $v_control$sroa$11$0$lcssa$i$i = 0, $v_control$sroa$11$0$us$i$i = 0, $v_control$sroa$11$1$i$i = 0, $v_control$sroa$11$1$us$i$i = 0, $v_start$sroa$0$1$i$i = 0, $v_start$sroa$0$2$i$i = 0, $v_start$sroa$13$1$i$i = 0, $v_start$sroa$13$2$i$i = 0, $waiting$i = 0, $y$0$i = 0, $y$0$ph$i = 0, $y$1$lcssa$i = 0, $y$1121$i = 0, $y$2127$i = 0, $y_height$0$i = 0, $y_height$0$ph$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $min_Y$i = sp + 14|0;
 $max_Y$i = sp + 12|0;
 $waiting$i = sp + 8|0;
 $draw_left$i = sp + 4|0;
 $draw_right$i = sp;
 $0 = ((($worker)) + 1020|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(-1);
 if (!($2)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ((($worker)) + 4|0);
 $4 = ((($worker)) + 80|0);
 $5 = ((($worker)) + 76|0);
 $6 = ((($worker)) + 28|0);
 $7 = ((($worker)) + 40|0);
 $8 = ((($worker)) + 44|0);
 $9 = ((($worker)) + 92|0);
 $10 = ((($worker)) + 87|0);
 $11 = ((($worker)) + 86|0);
 $12 = ((($worker)) + 32|0);
 $13 = ((($worker)) + 36|0);
 $14 = ((($worker)) + 48|0);
 $15 = ((($worker)) + 88|0);
 $16 = ((($worker)) + 84|0);
 $17 = ((($worker)) + 128|0);
 $18 = ((($worker)) + 100|0);
 $19 = ((($worker)) + 96|0);
 $20 = ((($worker)) + 140|0);
 $21 = ((($worker)) + 132|0);
 $22 = ((($worker)) + 24|0);
 $23 = ((($worker)) + 8|0);
 $24 = ($flipped<<24>>24)!=(0);
 $25 = ((($worker)) + 136|0);
 $26 = ((($worker)) + 68|0);
 $27 = ((($worker)) + 72|0);
 $28 = ((($worker)) + 176|0);
 $29 = ((($worker)) + 160|0);
 $30 = ((($worker)) + 172|0);
 $31 = ((($worker)) + 168|0);
 $32 = ((($worker)) + 164|0);
 $34 = $1;
 L4: while(1) {
  $33 = (((((($worker)) + 956|0) + ($34<<2)|0)) + 2|0);
  $35 = HEAP16[$33>>1]|0;
  $36 = $35 << 16 >> 16;
  $37 = HEAP32[$3>>2]|0;
  $38 = Math_imul($36, $37)|0;
  HEAP32[$4>>2] = $38;
  $39 = (((($worker)) + 956|0) + ($34<<2)|0);
  $40 = HEAP16[$39>>1]|0;
  $41 = $40 << 16 >> 16;
  $42 = Math_imul($41, $37)|0;
  HEAP32[$5>>2] = $42;
  $43 = HEAP32[$6>>2]|0;
  HEAP32[$7>>2] = $43;
  HEAP32[$8>>2] = 0;
  HEAP32[$9>>2] = 0;
  HEAP8[$10>>0] = 0;
  HEAP8[$11>>0] = 0;
  $44 = HEAP32[$12>>2]|0;
  $45 = ((($44)) + -32|0);
  HEAP32[$13>>2] = $45;
  HEAP32[$14>>2] = 0;
  $46 = $43;
  HEAP32[$15>>2] = $46;
  $47 = ((($46)) + 8|0);
  HEAP32[$47>>2] = $43;
  HEAP16[$16>>1] = 0;
  $48 = HEAP16[$17>>1]|0;
  $49 = ($48<<16>>16)>(0);
  L6: do {
   if ($49) {
    $$pre$i = HEAP32[$20>>2]|0;
    $$pre111$i = HEAP32[$23>>2]|0;
    $51 = $$pre$i;$62 = $$pre111$i;$i$060$i = 0;$start$061$i = 0;
    while(1) {
     HEAP32[$18>>2] = 0;
     HEAP32[$19>>2] = 0;
     $50 = (($51) + ($i$060$i<<1)|0);
     $52 = HEAP16[$50>>1]|0;
     $53 = HEAP32[$21>>2]|0;
     $54 = $52&65535;
     $55 = (($53) + ($54<<3)|0);
     $56 = $start$061$i & 65535;
     $57 = (($53) + ($56<<3)|0);
     $58 = HEAP32[$57>>2]|0;
     $59 = HEAP32[$22>>2]|0;
     $60 = $58 << $59;
     $61 = (($60) - ($62))|0;
     $63 = (((($53) + ($56<<3)|0)) + 4|0);
     $64 = HEAP32[$63>>2]|0;
     $65 = $64 << $59;
     $66 = (($65) - ($62))|0;
     $67 = HEAP32[$55>>2]|0;
     $68 = $67 << $59;
     $69 = (($68) - ($62))|0;
     $70 = (((($53) + ($54<<3)|0)) + 4|0);
     $71 = HEAP32[$70>>2]|0;
     $72 = $71 << $59;
     $73 = (($72) - ($62))|0;
     $$$i$i = $24 ? $66 : $61;
     $$3$i$i = $24 ? $61 : $66;
     $$4$i$i = $24 ? $73 : $69;
     $$5$i$i = $24 ? $69 : $73;
     $74 = HEAP32[$25>>2]|0;
     $75 = (($74) + ($56)|0);
     $76 = HEAP8[$75>>0]|0;
     $77 = $76 & 4;
     $78 = ($77<<24>>24)==(0);
     if ($78) {
      $81 = $76;
     } else {
      $79 = ($76&255) >>> 5;
      HEAP8[$28>>0] = $79;
      $$pre$i$i = HEAP8[$75>>0]|0;
      $81 = $$pre$i$i;
     }
     $80 = $81&255;
     $82 = $80 & 3;
     switch ($82|0) {
     case 2:  {
      label = 35;
      break L4;
      break;
     }
     case 0:  {
      $83 = (($74) + ($54)|0);
      $84 = HEAP8[$83>>0]|0;
      $85 = $84 & 3;
      $86 = ($85<<24>>24)==(1);
      if ($86) {
       $87 = ((($55)) + -8|0);
       $limit$0$i$i = $87;$v_start$sroa$0$1$i$i = $$4$i$i;$v_start$sroa$13$1$i$i = $$5$i$i;
      } else {
       $88 = (($$4$i$i) + ($$$i$i))|0;
       $89 = (($88|0) / 2)&-1;
       $90 = (($$5$i$i) + ($$3$i$i))|0;
       $91 = (($90|0) / 2)&-1;
       $limit$0$i$i = $55;$v_start$sroa$0$1$i$i = $89;$v_start$sroa$13$1$i$i = $91;
      }
      $92 = ((($57)) + -8|0);
      $93 = ((($75)) + -1|0);
      $limit$1$i$i = $limit$0$i$i;$point$0$i$i = $92;$tags$0$i$i = $93;$v_start$sroa$0$2$i$i = $v_start$sroa$0$1$i$i;$v_start$sroa$13$2$i$i = $v_start$sroa$13$1$i$i;
      break;
     }
     default: {
      $limit$1$i$i = $55;$point$0$i$i = $57;$tags$0$i$i = $75;$v_start$sroa$0$2$i$i = $$$i$i;$v_start$sroa$13$2$i$i = $$3$i$i;
     }
     }
     HEAP32[$26>>2] = $v_start$sroa$0$2$i$i;
     HEAP32[$27>>2] = $v_start$sroa$13$2$i$i;
     $94 = ($point$0$i$i>>>0)<($limit$1$i$i>>>0);
     L20: do {
      if ($94) {
       $point$143$i$i = $point$0$i$i;$tags$142$i$i = $tags$0$i$i;
       L21: while(1) {
        $95 = ((($point$143$i$i)) + 8|0);
        $96 = ((($tags$142$i$i)) + 1|0);
        $97 = HEAP8[$96>>0]|0;
        $98 = $97&255;
        $99 = $98 & 3;
        L23: do {
         switch ($99|0) {
         case 1:  {
          $100 = HEAP32[$95>>2]|0;
          $101 = HEAP32[$22>>2]|0;
          $102 = $100 << $101;
          $103 = HEAP32[$23>>2]|0;
          $104 = (($102) - ($103))|0;
          $105 = ((($point$143$i$i)) + 12|0);
          $106 = HEAP32[$105>>2]|0;
          $107 = $106 << $101;
          $108 = (($107) - ($103))|0;
          $$6$i$i = $24 ? $108 : $104;
          $$7$i$i = $24 ? $104 : $108;
          $109 = (_Line_To($worker,$$6$i$i,$$7$i$i)|0);
          $110 = ($109<<24>>24)==(0);
          if ($110) {
           $point$1$be$i$i = $95;$tags$1$be$i$i = $96;
          } else {
           label = 91;
           break L6;
          }
          break;
         }
         case 0:  {
          $111 = HEAP32[$95>>2]|0;
          $112 = HEAP32[$22>>2]|0;
          $113 = $111 << $112;
          $114 = HEAP32[$23>>2]|0;
          $115 = (($113) - ($114))|0;
          $116 = ((($point$143$i$i)) + 12|0);
          $117 = HEAP32[$116>>2]|0;
          $118 = $117 << $112;
          $119 = (($118) - ($114))|0;
          if ($24) {
           $point$2$us$i$i = $95;$tags$2$us$i$i = $96;$v_control$sroa$0$0$us$i$i = $119;$v_control$sroa$11$0$us$i$i = $115;
           while(1) {
            $120 = ($point$2$us$i$i>>>0)<($limit$1$i$i>>>0);
            if (!($120)) {
             $v_control$sroa$0$0$lcssa$i$i = $v_control$sroa$0$0$us$i$i;$v_control$sroa$11$0$lcssa$i$i = $v_control$sroa$11$0$us$i$i;
             break L21;
            }
            $121 = ((($point$2$us$i$i)) + 8|0);
            $122 = ((($tags$2$us$i$i)) + 1|0);
            $123 = HEAP8[$122>>0]|0;
            $124 = $123&255;
            $125 = $124 & 3;
            $126 = HEAP32[$121>>2]|0;
            $127 = HEAP32[$22>>2]|0;
            $128 = $126 << $127;
            $129 = HEAP32[$23>>2]|0;
            $130 = (($128) - ($129))|0;
            $131 = ((($point$2$us$i$i)) + 12|0);
            $132 = HEAP32[$131>>2]|0;
            $133 = $132 << $127;
            $134 = (($133) - ($129))|0;
            switch ($125|0) {
            case 1:  {
             $141 = (_Conic_To($worker,$v_control$sroa$0$0$us$i$i,$v_control$sroa$11$0$us$i$i,$134,$130)|0);
             $142 = ($141<<24>>24)==(0);
             $$12$us$i$i = $142 ? 7 : 12;
             $$1$us$i$i = $$12$us$i$i;$v_control$sroa$0$1$us$i$i = $v_control$sroa$0$0$us$i$i;$v_control$sroa$11$1$us$i$i = $v_control$sroa$11$0$us$i$i;
             break;
            }
            case 0:  {
             $135 = (($134) + ($v_control$sroa$0$0$us$i$i))|0;
             $136 = (($135|0) / 2)&-1;
             $137 = (($130) + ($v_control$sroa$11$0$us$i$i))|0;
             $138 = (($137|0) / 2)&-1;
             $139 = (_Conic_To($worker,$v_control$sroa$0$0$us$i$i,$v_control$sroa$11$0$us$i$i,$136,$138)|0);
             $140 = ($139<<24>>24)==(0);
             $$13$us$i$i = $140 ? 15 : 12;
             $$10$v_control$sroa$0$0$us$i$i = $140 ? $134 : $v_control$sroa$0$0$us$i$i;
             $$11$v_control$sroa$11$0$us$i$i = $140 ? $130 : $v_control$sroa$11$0$us$i$i;
             $$1$us$i$i = $$13$us$i$i;$v_control$sroa$0$1$us$i$i = $$10$v_control$sroa$0$0$us$i$i;$v_control$sroa$11$1$us$i$i = $$11$v_control$sroa$11$0$us$i$i;
             break;
            }
            default: {
             label = 35;
             break L4;
            }
            }
            switch ($$1$us$i$i|0) {
            case 6:  {
             label = 35;
             break L4;
             break;
            }
            case 12:  {
             label = 91;
             break L6;
             break;
            }
            case 7:  {
             $point$1$be$i$i = $121;$tags$1$be$i$i = $122;
             break L23;
             break;
            }
            case 15:  {
             $point$2$us$i$i = $121;$tags$2$us$i$i = $122;$v_control$sroa$0$0$us$i$i = $v_control$sroa$0$1$us$i$i;$v_control$sroa$11$0$us$i$i = $v_control$sroa$11$1$us$i$i;
             break;
            }
            default: {
             break L20;
            }
            }
           }
          } else {
           $point$2$i$i = $95;$tags$2$i$i = $96;$v_control$sroa$0$0$i$i = $115;$v_control$sroa$11$0$i$i = $119;
           while(1) {
            $143 = ($point$2$i$i>>>0)<($limit$1$i$i>>>0);
            if (!($143)) {
             $v_control$sroa$0$0$lcssa$i$i = $v_control$sroa$0$0$i$i;$v_control$sroa$11$0$lcssa$i$i = $v_control$sroa$11$0$i$i;
             break L21;
            }
            $144 = ((($point$2$i$i)) + 8|0);
            $145 = ((($tags$2$i$i)) + 1|0);
            $146 = HEAP8[$145>>0]|0;
            $147 = $146&255;
            $148 = $147 & 3;
            $149 = HEAP32[$144>>2]|0;
            $150 = HEAP32[$22>>2]|0;
            $151 = $149 << $150;
            $152 = HEAP32[$23>>2]|0;
            $153 = (($151) - ($152))|0;
            $154 = ((($point$2$i$i)) + 12|0);
            $155 = HEAP32[$154>>2]|0;
            $156 = $155 << $150;
            $157 = (($156) - ($152))|0;
            switch ($148|0) {
            case 1:  {
             $158 = (_Conic_To($worker,$v_control$sroa$0$0$i$i,$v_control$sroa$11$0$i$i,$153,$157)|0);
             $159 = ($158<<24>>24)==(0);
             $$12$i$i = $159 ? 7 : 12;
             $$1$i$i = $$12$i$i;$v_control$sroa$0$1$i$i = $v_control$sroa$0$0$i$i;$v_control$sroa$11$1$i$i = $v_control$sroa$11$0$i$i;
             break;
            }
            case 0:  {
             $160 = (($153) + ($v_control$sroa$0$0$i$i))|0;
             $161 = (($160|0) / 2)&-1;
             $162 = (($157) + ($v_control$sroa$11$0$i$i))|0;
             $163 = (($162|0) / 2)&-1;
             $164 = (_Conic_To($worker,$v_control$sroa$0$0$i$i,$v_control$sroa$11$0$i$i,$161,$163)|0);
             $165 = ($164<<24>>24)==(0);
             $$13$i$i = $165 ? 15 : 12;
             $$10$v_control$sroa$0$0$i$i = $165 ? $153 : $v_control$sroa$0$0$i$i;
             $$11$v_control$sroa$11$0$i$i = $165 ? $157 : $v_control$sroa$11$0$i$i;
             $$1$i$i = $$13$i$i;$v_control$sroa$0$1$i$i = $$10$v_control$sroa$0$0$i$i;$v_control$sroa$11$1$i$i = $$11$v_control$sroa$11$0$i$i;
             break;
            }
            default: {
             label = 35;
             break L4;
            }
            }
            switch ($$1$i$i|0) {
            case 6:  {
             label = 35;
             break L4;
             break;
            }
            case 12:  {
             label = 91;
             break L6;
             break;
            }
            case 7:  {
             $point$1$be$i$i = $144;$tags$1$be$i$i = $145;
             break L23;
             break;
            }
            case 15:  {
             $point$2$i$i = $144;$tags$2$i$i = $145;$v_control$sroa$0$0$i$i = $v_control$sroa$0$1$i$i;$v_control$sroa$11$0$i$i = $v_control$sroa$11$1$i$i;
             break;
            }
            default: {
             break L20;
            }
            }
           }
          }
          break;
         }
         default: {
          $168 = ((($point$143$i$i)) + 16|0);
          $169 = ($168>>>0)>($limit$1$i$i>>>0);
          if ($169) {
           label = 35;
           break L4;
          }
          $170 = ((($tags$142$i$i)) + 2|0);
          $171 = HEAP8[$170>>0]|0;
          $172 = $171 & 3;
          $173 = ($172<<24>>24)==(2);
          if (!($173)) {
           label = 35;
           break L4;
          }
          $174 = ((($point$143$i$i)) + 24|0);
          $175 = ((($tags$142$i$i)) + 3|0);
          $176 = ((($point$143$i$i)) + 8|0);
          $177 = HEAP32[$176>>2]|0;
          $178 = HEAP32[$22>>2]|0;
          $179 = $177 << $178;
          $180 = HEAP32[$23>>2]|0;
          $181 = (($179) - ($180))|0;
          $182 = ((($point$143$i$i)) + 12|0);
          $183 = HEAP32[$182>>2]|0;
          $184 = $183 << $178;
          $185 = (($184) - ($180))|0;
          $186 = ((($point$143$i$i)) + 16|0);
          $187 = HEAP32[$186>>2]|0;
          $188 = $187 << $178;
          $189 = (($188) - ($180))|0;
          $190 = ((($point$143$i$i)) + 20|0);
          $191 = HEAP32[$190>>2]|0;
          $192 = $191 << $178;
          $193 = (($192) - ($180))|0;
          $$14$i$i = $24 ? $185 : $181;
          $$15$i$i = $24 ? $181 : $185;
          $$16$i$i = $24 ? $193 : $189;
          $$17$i$i = $24 ? $189 : $193;
          $194 = ($174>>>0)>($limit$1$i$i>>>0);
          if ($194) {
           $204 = (_Cubic_To($worker,$$14$i$i,$$15$i$i,$$16$i$i,$$17$i$i,$v_start$sroa$0$2$i$i,$v_start$sroa$13$2$i$i)|0);
           $205 = ($204<<24>>24)==(0);
           $$20$i$i = $205 ? 18 : 12;
           $$2$i$i = $$20$i$i;
          } else {
           $195 = HEAP32[$174>>2]|0;
           $196 = $195 << $178;
           $197 = (($196) - ($180))|0;
           $198 = ((($point$143$i$i)) + 28|0);
           $199 = HEAP32[$198>>2]|0;
           $200 = $199 << $178;
           $201 = (($200) - ($180))|0;
           $$18$i$i = $24 ? $201 : $197;
           $$19$i$i = $24 ? $197 : $201;
           $202 = (_Cubic_To($worker,$$14$i$i,$$15$i$i,$$16$i$i,$$17$i$i,$$18$i$i,$$19$i$i)|0);
           $203 = ($202<<24>>24)==(0);
           $$21$i$i = $203 ? 7 : 12;
           $$2$i$i = $$21$i$i;
          }
          switch ($$2$i$i|0) {
          case 6:  {
           label = 35;
           break L4;
           break;
          }
          case 12:  {
           label = 91;
           break L6;
           break;
          }
          case 7:  {
           $point$1$be$i$i = $174;$tags$1$be$i$i = $175;
           break;
          }
          default: {
           break L20;
          }
          }
         }
         }
        } while(0);
        $206 = ($point$1$be$i$i>>>0)<($limit$1$i$i>>>0);
        if ($206) {
         $point$143$i$i = $point$1$be$i$i;$tags$142$i$i = $tags$1$be$i$i;
        } else {
         label = 34;
         break L20;
        }
       }
       $166 = (_Conic_To($worker,$v_control$sroa$0$0$lcssa$i$i,$v_control$sroa$11$0$lcssa$i$i,$v_start$sroa$0$2$i$i,$v_start$sroa$13$2$i$i)|0);
       $167 = ($166<<24>>24)==(0);
       if (!($167)) {
        label = 91;
        break L6;
       }
      } else {
       label = 34;
      }
     } while(0);
     if ((label|0) == 34) {
      label = 0;
      $207 = (_Line_To($worker,$v_start$sroa$0$2$i$i,$v_start$sroa$13$2$i$i)|0);
      $208 = ($207<<24>>24)==(0);
      if (!($208)) {
       label = 91;
       break L6;
      }
     }
     $209 = HEAP32[$20>>2]|0;
     $210 = (($209) + ($i$060$i<<1)|0);
     $211 = HEAP16[$210>>1]|0;
     $212 = $211&65535;
     $213 = (($212) + 1)|0;
     $214 = HEAP32[$27>>2]|0;
     $215 = HEAP32[$3>>2]|0;
     $216 = (($215) + -1)|0;
     $217 = $216 & $214;
     $218 = ($217|0)==(0);
     if ($218) {
      $219 = HEAP32[$5>>2]|0;
      $220 = ($214|0)<($219|0);
      if (!($220)) {
       $221 = HEAP32[$4>>2]|0;
       $222 = ($214|0)>($221|0);
       if (!($222)) {
        $223 = HEAP32[$19>>2]|0;
        $224 = ($223|0)==(0|0);
        if (!($224)) {
         $225 = ((($223)) + 12|0);
         $226 = HEAP16[$225>>1]|0;
         $227 = HEAP32[$15>>2]|0;
         $228 = ((($227)) + 12|0);
         $229 = HEAP16[$228>>1]|0;
         $230 = $229 ^ $226;
         $231 = $230 & 8;
         $232 = ($231<<16>>16)==(0);
         if ($232) {
          $233 = HEAP32[$7>>2]|0;
          $234 = ((($233)) + -4|0);
          HEAP32[$7>>2] = $234;
         }
        }
       }
      }
     }
     $235 = HEAP32[$15>>2]|0;
     $236 = HEAP32[$7>>2]|0;
     $237 = ((($235)) + 8|0);
     $238 = HEAP32[$237>>2]|0;
     $239 = ($236|0)==($238|0);
     $240 = $236;
     $241 = $238;
     if ($239) {
      label = 45;
     } else {
      $242 = ((($235)) + 12|0);
      $243 = HEAP16[$242>>1]|0;
      $244 = $243 & 8;
      $245 = ($244<<16>>16)==(0);
      if ($245) {
       label = 45;
      } else {
       $246 = HEAP32[$23>>2]|0;
       $247 = ($217|0)>=($246|0);
       $653 = $246;$o$0$in$i = $247;
      }
     }
     if ((label|0) == 45) {
      label = 0;
      $248 = (($214) + -1)|0;
      $249 = (($248) + ($215))|0;
      $250 = (0 - ($215))|0;
      $251 = $249 & $250;
      $252 = (($251) - ($214))|0;
      $253 = HEAP32[$23>>2]|0;
      $254 = ($252|0)>=($253|0);
      $653 = $253;$o$0$in$i = $254;
     }
     $255 = (($240) - ($241))|0;
     $256 = $255 >> 2;
     $257 = ($255|0)<(0);
     if ($257) {
      label = 47;
      break L4;
     }
     $258 = ($255|0)>(0);
     if ($258) {
      $259 = ((($235)) + 16|0);
      HEAP32[$259>>2] = $256;
      do {
       if ($o$0$in$i) {
        $260 = ((($235)) + 12|0);
        $261 = HEAP16[$260>>1]|0;
        $262 = $261&65535;
        $263 = $262 & 8;
        $264 = ($263|0)==(0);
        if ($264) {
         $267 = $262 | 32;
         $268 = $267&65535;
         HEAP16[$260>>1] = $268;
         break;
        } else {
         $265 = $262 | 16;
         $266 = $265&65535;
         HEAP16[$260>>1] = $266;
         break;
        }
       }
      } while(0);
      HEAP32[$15>>2] = $240;
      $269 = ((($236)) + 32|0);
      HEAP32[$7>>2] = $269;
      $270 = ((($236)) + 16|0);
      HEAP32[$270>>2] = 0;
      $271 = $269;
      $272 = ((($236)) + 8|0);
      HEAP32[$272>>2] = $271;
      $273 = ((($235)) + 28|0);
      HEAP32[$273>>2] = $240;
      $274 = HEAP16[$16>>1]|0;
      $275 = (($274) + 1)<<16>>16;
      HEAP16[$16>>1] = $275;
      $278 = $269;
     } else {
      $278 = $236;
     }
     $276 = HEAP32[$13>>2]|0;
     $277 = ($278>>>0)<($276>>>0);
     if (!($277)) {
      label = 55;
      break;
     }
     HEAP8[$10>>0] = 0;
     $279 = HEAP32[$19>>2]|0;
     $280 = ($279|0)==(0|0);
     if (!($280)) {
      $281 = ((($235)) + 28|0);
      HEAP32[$281>>2] = $279;
     }
     $282 = (($i$060$i) + 1)|0;
     $283 = HEAP16[$17>>1]|0;
     $284 = $283 << 16 >> 16;
     $285 = ($282|0)<($284|0);
     if ($285) {
      $51 = $209;$62 = $653;$i$060$i = $282;$start$061$i = $213;
     } else {
      $$lcssa345 = $278;$$lcssa346 = $276;
      break;
     }
    }
    if ((label|0) == 55) {
     label = 0;
     HEAP32[$8>>2] = 98;
     label = 92;
     break;
    }
    $$pre112$i = HEAP16[$16>>1]|0;
    $$pre113$i = HEAP32[$9>>2]|0;
    $286 = ($$pre112$i&65535)>(1);
    $287 = ($$pre113$i|0)!=(0|0);
    $or$cond$i$i = $286 & $287;
    if ($or$cond$i$i) {
     $288 = ($$pre112$i<<16>>16)==(0);
     if ($288) {
      $355 = $$lcssa345;$356 = $$lcssa346;$382 = $$pre113$i;
      label = 90;
     } else {
      $289 = HEAP32[$12>>2]|0;
      $326 = $$lcssa346;$328 = $$lcssa345;$n$032$i$i = $$pre112$i;$p$033$i$i = $$pre113$i;
      while(1) {
       $290 = ($n$032$i$i&65535)>(1);
       $291 = ((($p$033$i$i)) + 16|0);
       $292 = HEAP32[$291>>2]|0;
       if ($290) {
        $293 = ((($p$033$i$i)) + 8|0);
        $294 = HEAP32[$293>>2]|0;
        $295 = (($294) + ($292<<2)|0);
        $296 = ((($p$033$i$i)) + 4|0);
        HEAP32[$296>>2] = $295;
        $654 = $295;
       } else {
        $297 = ((($p$033$i$i)) + 4|0);
        HEAP32[$297>>2] = 0;
        $654 = 0;
       }
       $298 = ((($p$033$i$i)) + 12|0);
       $299 = HEAP16[$298>>1]|0;
       $300 = $299 & 8;
       $301 = ($300<<16>>16)==(0);
       $302 = ((($p$033$i$i)) + 20|0);
       $303 = HEAP32[$302>>2]|0;
       if ($301) {
        $306 = (($303) - ($292))|0;
        $307 = (($306) + 1)|0;
        HEAP32[$302>>2] = $307;
        $308 = (($292) + -1)|0;
        $309 = ((($p$033$i$i)) + 8|0);
        $310 = HEAP32[$309>>2]|0;
        $311 = (($310) + ($308<<2)|0);
        HEAP32[$309>>2] = $311;
        $bottom$0$i$i = $307;$top$0$i$i = $303;
       } else {
        $304 = (($292) + -1)|0;
        $305 = (($304) + ($303))|0;
        $bottom$0$i$i = $303;$top$0$i$i = $305;
       }
       $312 = HEAP32[$14>>2]|0;
       $313 = (0 - ($312))|0;
       $314 = (($289) + ($313<<2)|0);
       $n$0$in$i$i$i = $312;
       while(1) {
        $n$0$i$i$i = (($n$0$in$i$i$i) + -1)|0;
        $315 = ($n$0$in$i$i$i|0)>(0);
        if (!($315)) {
         $$1$i$i$i = $bottom$0$i$i;$n$2$i$i$i = $n$0$i$i$i;
         label = 74;
         break;
        }
        $316 = (($314) + ($n$0$i$i$i<<2)|0);
        $317 = HEAP32[$316>>2]|0;
        $318 = ($317|0)>($bottom$0$i$i|0);
        if ($318) {
         $n$0$in$i$i$i = $n$0$i$i$i;
        } else {
         $$lcssa = $316;$$lcssa349 = $317;$n$0$i$i$i$lcssa348 = $n$0$i$i$i;$n$0$in$i$i$i$lcssa347 = $n$0$in$i$i$i;
         label = 71;
         break;
        }
       }
       if ((label|0) == 71) {
        label = 0;
        $319 = ($$lcssa349|0)<($bottom$0$i$i|0);
        if ($319) {
         HEAP32[$$lcssa>>2] = $bottom$0$i$i;
         $320 = (($n$0$in$i$i$i$lcssa347) + -2)|0;
         $$old1$i27$i$i = ($n$0$in$i$i$i$lcssa347|0)>(1);
         if ($$old1$i27$i$i) {
          $321 = $320;$322 = $$lcssa349;
          while(1) {
           $$phi$trans$insert$i$i$i = (($314) + ($321<<2)|0);
           $$pre$i$i$i = HEAP32[$$phi$trans$insert$i$i$i>>2]|0;
           HEAP32[$$phi$trans$insert$i$i$i>>2] = $322;
           $323 = (($321) + -1)|0;
           $$old1$i$i$i = ($321|0)>(0);
           if ($$old1$i$i$i) {
            $321 = $323;$322 = $$pre$i$i$i;
           } else {
            $$1$i43$i$i = $$pre$i$i$i;
            label = 75;
            break;
           }
          }
         } else {
          $$1$i$i$i = $$lcssa349;$n$2$i$i$i = $320;
          label = 74;
         }
        } else {
         $$1$i$i$i = $bottom$0$i$i;$n$2$i$i$i = $n$0$i$i$i$lcssa348;
         label = 74;
        }
       }
       if ((label|0) == 74) {
        label = 0;
        $324 = ($n$2$i$i$i|0)<(0);
        if ($324) {
         $$1$i43$i$i = $$1$i$i$i;
         label = 75;
        } else {
         $335 = $312;$348 = $326;
        }
       }
       if ((label|0) == 75) {
        label = 0;
        $325 = ((($326)) + -4|0);
        HEAP32[$13>>2] = $325;
        $327 = ($325>>>0)>($328>>>0);
        if (!($327)) {
         label = 77;
         break;
        }
        $329 = (($312) + 1)|0;
        HEAP32[$14>>2] = $329;
        $330 = $312 ^ -1;
        $331 = (($289) + ($330<<2)|0);
        HEAP32[$331>>2] = $$1$i43$i$i;
        $335 = $329;$348 = $325;
       }
       $332 = (($top$0$i$i) + 1)|0;
       $333 = HEAP32[$12>>2]|0;
       $334 = (0 - ($335))|0;
       $336 = (($333) + ($334<<2)|0);
       $n$0$in$i3$i$i = $335;
       while(1) {
        $n$0$i4$i$i = (($n$0$in$i3$i$i) + -1)|0;
        $337 = ($n$0$in$i3$i$i|0)>(0);
        if (!($337)) {
         $$1$i13$i$i = $332;$n$2$i14$i$i = $n$0$i4$i$i;
         label = 84;
         break;
        }
        $338 = (($336) + ($n$0$i4$i$i<<2)|0);
        $339 = HEAP32[$338>>2]|0;
        $340 = ($339|0)>($332|0);
        if ($340) {
         $n$0$in$i3$i$i = $n$0$i4$i$i;
        } else {
         $$lcssa352 = $338;$$lcssa353 = $339;$n$0$i4$i$i$lcssa351 = $n$0$i4$i$i;$n$0$in$i3$i$i$lcssa350 = $n$0$in$i3$i$i;
         label = 81;
         break;
        }
       }
       if ((label|0) == 81) {
        label = 0;
        $341 = ($$lcssa353|0)>($top$0$i$i|0);
        if ($341) {
         $$1$i13$i$i = $332;$n$2$i14$i$i = $n$0$i4$i$i$lcssa351;
         label = 84;
        } else {
         HEAP32[$$lcssa352>>2] = $332;
         $342 = (($n$0$in$i3$i$i$lcssa350) + -2)|0;
         $$old1$i829$i$i = ($n$0$in$i3$i$i$lcssa350|0)>(1);
         if ($$old1$i829$i$i) {
          $343 = $342;$344 = $$lcssa353;
          while(1) {
           $$phi$trans$insert$i10$i$i = (($336) + ($343<<2)|0);
           $$pre$i11$i$i = HEAP32[$$phi$trans$insert$i10$i$i>>2]|0;
           HEAP32[$$phi$trans$insert$i10$i$i>>2] = $344;
           $345 = (($343) + -1)|0;
           $$old1$i8$i$i = ($343|0)>(0);
           if ($$old1$i8$i$i) {
            $343 = $345;$344 = $$pre$i11$i$i;
           } else {
            $$1$i1344$i$i = $$pre$i11$i$i;
            label = 85;
            break;
           }
          }
         } else {
          $$1$i13$i$i = $$lcssa353;$n$2$i14$i$i = $342;
          label = 84;
         }
        }
       }
       if ((label|0) == 84) {
        label = 0;
        $346 = ($n$2$i14$i$i|0)<(0);
        if ($346) {
         $$1$i1344$i$i = $$1$i13$i$i;
         label = 85;
        } else {
         $655 = $328;$656 = $348;
        }
       }
       if ((label|0) == 85) {
        label = 0;
        $347 = ((($348)) + -4|0);
        HEAP32[$13>>2] = $347;
        $349 = ($347>>>0)>($$lcssa345>>>0);
        if (!($349)) {
         label = 87;
         break;
        }
        $350 = (($335) + 1)|0;
        HEAP32[$14>>2] = $350;
        $351 = $335 ^ -1;
        $352 = (($333) + ($351<<2)|0);
        HEAP32[$352>>2] = $$1$i1344$i$i;
        $655 = $$lcssa345;$656 = $347;
       }
       $353 = (($n$032$i$i) + -1)<<16>>16;
       $354 = ($353<<16>>16)==(0);
       if ($354) {
        $355 = $$lcssa345;$356 = $656;$382 = $$pre113$i;
        label = 90;
        break L6;
       } else {
        $326 = $656;$328 = $655;$n$032$i$i = $353;$p$033$i$i = $654;
       }
      }
      if ((label|0) == 77) {
       label = 0;
       HEAP32[$8>>2] = 98;
       label = 92;
       break;
      }
      else if ((label|0) == 87) {
       label = 0;
       HEAP32[$8>>2] = 98;
       label = 92;
       break;
      }
     }
    } else {
     $651 = $$lcssa346;$652 = $$lcssa345;
     label = 89;
    }
   } else {
    $651 = $45;$652 = $46;
    label = 89;
   }
  } while(0);
  if ((label|0) == 89) {
   label = 0;
   HEAP32[$9>>2] = 0;
   $355 = $652;$356 = $651;$382 = 0;
   label = 90;
  }
  if ((label|0) == 90) {
   label = 0;
   $not$$i = ($355>>>0)<($356>>>0);
   if ($not$$i) {
    $381 = ($382|0)==(0|0);
    if (!($381)) {
     HEAP32[$waiting$i>>2] = 0;
     HEAP32[$draw_left$i>>2] = 0;
     HEAP32[$draw_right$i>>2] = 0;
     $383 = HEAP32[$5>>2]|0;
     $384 = HEAP32[$worker>>2]|0;
     $385 = $383 >> $384;
     $386 = $385&65535;
     HEAP16[$max_Y$i>>1] = $386;
     $387 = HEAP32[$4>>2]|0;
     $388 = $387 >> $384;
     $389 = $388&65535;
     HEAP16[$min_Y$i>>1] = $389;
     $400 = $389;$405 = $386;$P$0134$i = $382;
     while(1) {
      $390 = ((($P$0134$i)) + 4|0);
      $391 = HEAP32[$390>>2]|0;
      $392 = ((($P$0134$i)) + 20|0);
      $393 = HEAP32[$392>>2]|0;
      $394 = ((($P$0134$i)) + 16|0);
      $395 = HEAP32[$394>>2]|0;
      $396 = (($393) + 65535)|0;
      $397 = (($396) + ($395))|0;
      $398 = $397&65535;
      $399 = $400 << 16 >> 16;
      $sext2$i = $393 << 16;
      $401 = $sext2$i >> 16;
      $402 = ($399|0)>($401|0);
      if ($402) {
       $403 = $393&65535;
       HEAP16[$min_Y$i>>1] = $403;
       $657 = $403;
      } else {
       $657 = $400;
      }
      $404 = $405 << 16 >> 16;
      $sext3$i = $397 << 16;
      $406 = $sext3$i >> 16;
      $407 = ($404|0)<($406|0);
      if ($407) {
       HEAP16[$max_Y$i>>1] = $398;
       $658 = $398;
      } else {
       $658 = $405;
      }
      HEAP32[$P$0134$i>>2] = 0;
      $408 = HEAP32[$waiting$i>>2]|0;
      $409 = ($408|0)==(0|0);
      L141: do {
       if ($409) {
        $current$0$in$lcssa$i$i = $waiting$i;$current$0$lcssa$i$i = 0;
       } else {
        $current$0$in2$i$i = $waiting$i;$current$03$i$i = $408;
        while(1) {
         $410 = HEAP32[$current$03$i$i>>2]|0;
         $411 = ($410|0)>(0);
         if ($411) {
          $current$0$in$lcssa$i$i = $current$0$in2$i$i;$current$0$lcssa$i$i = $current$03$i$i;
          break L141;
         }
         $412 = ((($current$03$i$i)) + 4|0);
         $current$0$i33$i = HEAP32[$412>>2]|0;
         $413 = ($current$0$i33$i|0)==(0|0);
         if ($413) {
          $current$0$in$lcssa$i$i = $412;$current$0$lcssa$i$i = 0;
          break;
         } else {
          $current$0$in2$i$i = $412;$current$03$i$i = $current$0$i33$i;
         }
        }
       }
      } while(0);
      HEAP32[$390>>2] = $current$0$lcssa$i$i;
      HEAP32[$current$0$in$lcssa$i$i>>2] = $P$0134$i;
      $414 = ($391|0)==(0|0);
      if ($414) {
       break;
      } else {
       $400 = $657;$405 = $658;$P$0134$i = $391;
      }
     }
     $415 = HEAP32[$14>>2]|0;
     $416 = ($415|0)==(0);
     if ($416) {
      label = 106;
      break;
     }
     $417 = HEAP32[$29>>2]|0;
     FUNCTION_TABLE_viii[$417 & 8191]($worker,$min_Y$i,$max_Y$i);
     $418 = HEAP32[$waiting$i>>2]|0;
     $419 = ($418|0)==(0|0);
     $$pre$i1 = HEAP16[$min_Y$i>>1]|0;
     if (!($419)) {
      $420 = $$pre$i1 << 16 >> 16;
      $P$1130$i = $418;
      while(1) {
       $421 = ((($P$1130$i)) + 20|0);
       $422 = HEAP32[$421>>2]|0;
       $423 = (($422) - ($420))|0;
       $424 = ((($P$1130$i)) + 24|0);
       HEAP32[$424>>2] = $423;
       $425 = ((($P$1130$i)) + 4|0);
       $P$1$i = HEAP32[$425>>2]|0;
       $426 = ($P$1$i|0)==(0|0);
       if ($426) {
        break;
       } else {
        $P$1130$i = $P$1$i;
       }
      }
     }
     $427 = HEAP32[$14>>2]|0;
     $428 = ($427|0)>(0);
     if ($428) {
      $429 = (0 - ($427))|0;
      $430 = HEAP32[$12>>2]|0;
      $431 = (($430) + ($429<<2)|0);
      $432 = HEAP32[$431>>2]|0;
      $433 = $$pre$i1 << 16 >> 16;
      $434 = ($432|0)==($433|0);
      if ($434) {
       $435 = (($427) + -1)|0;
       HEAP32[$14>>2] = $435;
       $437 = $435;$y$0$i = $$pre$i1;$y_height$0$i = 0;
      } else {
       $y$0$ph$i = $$pre$i1;$y_height$0$ph$i = 0;
       label = 113;
      }
     } else {
      $437 = $427;$y$0$i = $$pre$i1;$y_height$0$i = 0;
     }
     L156: while(1) {
      if ((label|0) == 113) {
       label = 0;
       $$pr$i = HEAP32[$14>>2]|0;
       $437 = $$pr$i;$y$0$i = $y$0$ph$i;$y_height$0$i = $y_height$0$ph$i;
      }
      $436 = ($437|0)>(0);
      if (!($436)) {
       break;
      }
      $440 = HEAP32[$waiting$i>>2]|0;
      $441 = ($440|0)==(0|0);
      if (!($441)) {
       $sext1$i = $y_height$0$i << 16;
       $442 = $sext1$i >> 16;
       $P$2123$i = $440;
       while(1) {
        $443 = ((($P$2123$i)) + 4|0);
        $444 = HEAP32[$443>>2]|0;
        $445 = ((($P$2123$i)) + 24|0);
        $446 = HEAP32[$445>>2]|0;
        $447 = (($446) - ($442))|0;
        HEAP32[$445>>2] = $447;
        $448 = ($446|0)==($442|0);
        do {
         if ($448) {
          $old$0$i35$i = $waiting$i;
          while(1) {
           $current$0$i36$i = HEAP32[$old$0$i35$i>>2]|0;
           $449 = ($current$0$i36$i|0)==(0|0);
           if ($449) {
            break;
           }
           $450 = ($current$0$i36$i|0)==($P$2123$i|0);
           $451 = ((($current$0$i36$i)) + 4|0);
           if ($450) {
            $$lcssa356 = $451;$old$0$i35$i$lcssa355 = $old$0$i35$i;
            label = 121;
            break;
           } else {
            $old$0$i35$i = $451;
           }
          }
          if ((label|0) == 121) {
           label = 0;
           $452 = HEAP32[$$lcssa356>>2]|0;
           HEAP32[$old$0$i35$i$lcssa355>>2] = $452;
          }
          $453 = ((($P$2123$i)) + 12|0);
          $454 = HEAP16[$453>>1]|0;
          $455 = $454 & 8;
          $456 = ($455<<16>>16)==(0);
          $457 = HEAP32[$P$2123$i>>2]|0;
          if ($456) {
           $465 = HEAP32[$draw_right$i>>2]|0;
           $466 = ($465|0)==(0|0);
           L174: do {
            if ($466) {
             $current$0$in$lcssa$i84$i = $draw_right$i;$current$0$lcssa$i83$i = 0;
            } else {
             $current$0$in2$i80$i = $draw_right$i;$current$03$i79$i = $465;
             while(1) {
              $467 = HEAP32[$current$03$i79$i>>2]|0;
              $468 = ($457|0)<($467|0);
              if ($468) {
               $current$0$in$lcssa$i84$i = $current$0$in2$i80$i;$current$0$lcssa$i83$i = $current$03$i79$i;
               break L174;
              }
              $469 = ((($current$03$i79$i)) + 4|0);
              $current$0$i82$i = HEAP32[$469>>2]|0;
              $470 = ($current$0$i82$i|0)==(0|0);
              if ($470) {
               $current$0$in$lcssa$i84$i = $469;$current$0$lcssa$i83$i = 0;
               break;
              } else {
               $current$0$in2$i80$i = $469;$current$03$i79$i = $current$0$i82$i;
              }
             }
            }
           } while(0);
           HEAP32[$443>>2] = $current$0$lcssa$i83$i;
           HEAP32[$current$0$in$lcssa$i84$i>>2] = $P$2123$i;
           break;
          } else {
           $458 = HEAP32[$draw_left$i>>2]|0;
           $459 = ($458|0)==(0|0);
           L180: do {
            if ($459) {
             $current$0$in$lcssa$i93$i = $draw_left$i;$current$0$lcssa$i92$i = 0;
            } else {
             $current$0$in2$i89$i = $draw_left$i;$current$03$i88$i = $458;
             while(1) {
              $460 = HEAP32[$current$03$i88$i>>2]|0;
              $461 = ($457|0)<($460|0);
              if ($461) {
               $current$0$in$lcssa$i93$i = $current$0$in2$i89$i;$current$0$lcssa$i92$i = $current$03$i88$i;
               break L180;
              }
              $462 = ((($current$03$i88$i)) + 4|0);
              $current$0$i91$i = HEAP32[$462>>2]|0;
              $463 = ($current$0$i91$i|0)==(0|0);
              if ($463) {
               $current$0$in$lcssa$i93$i = $462;$current$0$lcssa$i92$i = 0;
               break;
              } else {
               $current$0$in2$i89$i = $462;$current$03$i88$i = $current$0$i91$i;
              }
             }
            }
           } while(0);
           HEAP32[$443>>2] = $current$0$lcssa$i92$i;
           HEAP32[$current$0$in$lcssa$i93$i>>2] = $P$2123$i;
           break;
          }
         }
        } while(0);
        $464 = ($444|0)==(0|0);
        if ($464) {
         break;
        } else {
         $P$2123$i = $444;
        }
       }
      }
      $471 = HEAP32[$draw_left$i>>2]|0;
      $472 = ($471|0)==(0|0);
      if (!($472)) {
       $current$08$i59$i = $471;
       while(1) {
        $473 = ((($current$08$i59$i)) + 8|0);
        $474 = HEAP32[$473>>2]|0;
        $475 = HEAP32[$474>>2]|0;
        HEAP32[$current$08$i59$i>>2] = $475;
        $476 = ((($current$08$i59$i)) + 12|0);
        $477 = HEAP16[$476>>1]|0;
        $478 = $477 & 8;
        $479 = $478&65535;
        $480 = $479 >>> 2;
        $481 = (($480) + -1)|0;
        $482 = (($474) + ($481<<2)|0);
        HEAP32[$473>>2] = $482;
        $483 = ((($current$08$i59$i)) + 16|0);
        $484 = HEAP32[$483>>2]|0;
        $485 = (($484) + -1)|0;
        HEAP32[$483>>2] = $485;
        $486 = ((($current$08$i59$i)) + 4|0);
        $current$0$i60$i = HEAP32[$486>>2]|0;
        $487 = ($current$0$i60$i|0)==(0|0);
        if ($487) {
         break;
        } else {
         $current$08$i59$i = $current$0$i60$i;
        }
       }
       $next$0$in1$i64$i = ((($471)) + 4|0);
       $next$02$i65$i = HEAP32[$next$0$in1$i64$i>>2]|0;
       $488 = ($next$02$i65$i|0)==(0|0);
       if (!($488)) {
        $current$14$i69$i = $471;$next$0$in5$i68$i = $next$0$in1$i64$i;$next$06$i67$i = $next$02$i65$i;$old$03$i70$i = $draw_left$i;
        while(1) {
         $489 = HEAP32[$current$14$i69$i>>2]|0;
         $490 = HEAP32[$next$06$i67$i>>2]|0;
         $491 = ($489|0)>($490|0);
         if ($491) {
          HEAP32[$old$03$i70$i>>2] = $next$06$i67$i;
          $492 = ((($next$06$i67$i)) + 4|0);
          $493 = HEAP32[$492>>2]|0;
          HEAP32[$next$0$in5$i68$i>>2] = $493;
          HEAP32[$492>>2] = $current$14$i69$i;
          $494 = HEAP32[$draw_left$i>>2]|0;
          $current$1$be$i73$i = $494;$old$0$be$i72$i = $draw_left$i;
         } else {
          $current$1$be$i73$i = $next$06$i67$i;$old$0$be$i72$i = $next$0$in5$i68$i;
         }
         $next$0$in$i74$i = ((($current$1$be$i73$i)) + 4|0);
         $next$0$i75$i = HEAP32[$next$0$in$i74$i>>2]|0;
         $495 = ($next$0$i75$i|0)==(0|0);
         if ($495) {
          break;
         } else {
          $current$14$i69$i = $current$1$be$i73$i;$next$0$in5$i68$i = $next$0$in$i74$i;$next$06$i67$i = $next$0$i75$i;$old$03$i70$i = $old$0$be$i72$i;
         }
        }
       }
      }
      $496 = HEAP32[$draw_right$i>>2]|0;
      $497 = ($496|0)==(0|0);
      if ($497) {
       $P_Right$0114$i148 = 0;
      } else {
       $current$08$i39$i = $496;
       while(1) {
        $498 = ((($current$08$i39$i)) + 8|0);
        $499 = HEAP32[$498>>2]|0;
        $500 = HEAP32[$499>>2]|0;
        HEAP32[$current$08$i39$i>>2] = $500;
        $501 = ((($current$08$i39$i)) + 12|0);
        $502 = HEAP16[$501>>1]|0;
        $503 = $502 & 8;
        $504 = $503&65535;
        $505 = $504 >>> 2;
        $506 = (($505) + -1)|0;
        $507 = (($499) + ($506<<2)|0);
        HEAP32[$498>>2] = $507;
        $508 = ((($current$08$i39$i)) + 16|0);
        $509 = HEAP32[$508>>2]|0;
        $510 = (($509) + -1)|0;
        HEAP32[$508>>2] = $510;
        $511 = ((($current$08$i39$i)) + 4|0);
        $current$0$i40$i = HEAP32[$511>>2]|0;
        $512 = ($current$0$i40$i|0)==(0|0);
        if ($512) {
         break;
        } else {
         $current$08$i39$i = $current$0$i40$i;
        }
       }
       $next$0$in1$i44$i = ((($496)) + 4|0);
       $next$02$i45$i = HEAP32[$next$0$in1$i44$i>>2]|0;
       $513 = ($next$02$i45$i|0)==(0|0);
       if ($513) {
        $P_Right$0114$i148 = $496;
       } else {
        $P_Right$0114$i147 = $496;$current$14$i49$i = $496;$next$0$in5$i48$i = $next$0$in1$i44$i;$next$06$i47$i = $next$02$i45$i;$old$03$i50$i = $draw_right$i;
        while(1) {
         $514 = HEAP32[$current$14$i49$i>>2]|0;
         $515 = HEAP32[$next$06$i47$i>>2]|0;
         $516 = ($514|0)>($515|0);
         if ($516) {
          HEAP32[$old$03$i50$i>>2] = $next$06$i47$i;
          $517 = ((($next$06$i47$i)) + 4|0);
          $518 = HEAP32[$517>>2]|0;
          HEAP32[$next$0$in5$i48$i>>2] = $518;
          HEAP32[$517>>2] = $current$14$i49$i;
          $519 = HEAP32[$draw_right$i>>2]|0;
          $P_Right$0114$i146 = $519;$current$1$be$i53$i = $519;$old$0$be$i52$i = $draw_right$i;
         } else {
          $P_Right$0114$i146 = $P_Right$0114$i147;$current$1$be$i53$i = $next$06$i47$i;$old$0$be$i52$i = $next$0$in5$i48$i;
         }
         $next$0$in$i54$i = ((($current$1$be$i53$i)) + 4|0);
         $next$0$i55$i = HEAP32[$next$0$in$i54$i>>2]|0;
         $520 = ($next$0$i55$i|0)==(0|0);
         if ($520) {
          $P_Right$0114$i148 = $P_Right$0114$i146;
          break;
         } else {
          $P_Right$0114$i147 = $P_Right$0114$i146;$current$14$i49$i = $current$1$be$i53$i;$next$0$in5$i48$i = $next$0$in$i54$i;$next$06$i47$i = $next$0$i55$i;$old$03$i50$i = $old$0$be$i52$i;
         }
        }
       }
      }
      $521 = (($437) + -1)|0;
      HEAP32[$14>>2] = $521;
      $522 = (0 - ($437))|0;
      $523 = HEAP32[$12>>2]|0;
      $524 = (($523) + ($522<<2)|0);
      $525 = HEAP32[$524>>2]|0;
      $sext$i = $525 << 16;
      $526 = $sext$i >> 16;
      $527 = $y$0$i&65535;
      $528 = (($526) - ($527))|0;
      $529 = $y$0$i << 16 >> 16;
      $530 = ($529|0)<($526|0);
      L207: do {
       if ($530) {
        $P_Right$0114$i144 = $P_Right$0114$i148;$y$1121$i = $y$0$i;
        L208: while(1) {
         $531 = HEAP32[$draw_left$i>>2]|0;
         $532 = ($531|0)==(0|0);
         do {
          if (!($532)) {
           $P_Left$0113$i = $531;$P_Right$0114$i = $P_Right$0114$i144;$dropouts$0112$i = 0;
           while(1) {
            $533 = HEAP32[$P_Left$0113$i>>2]|0;
            $534 = HEAP32[$P_Right$0114$i>>2]|0;
            $535 = ($533|0)>($534|0);
            $$$i = $535 ? $534 : $533;
            $$4$i = $535 ? $533 : $534;
            $536 = HEAP32[$3>>2]|0;
            $537 = (0 - ($536))|0;
            $538 = $$$i & $537;
            $539 = (($536) + -1)|0;
            $540 = (($539) + ($$4$i))|0;
            $541 = $540 & $537;
            $542 = (($$4$i) - ($$$i))|0;
            $543 = ($542|0)>($536|0);
            $544 = ($538|0)==($$$i|0);
            $or$cond$i = $543 | $544;
            $545 = ($541|0)==($$4$i|0);
            $or$cond5$i = $or$cond$i | $545;
            do {
             if ($or$cond5$i) {
              label = 151;
             } else {
              $546 = ($538|0)>($541|0);
              $547 = (($538) + ($536))|0;
              $548 = ($541|0)==($547|0);
              $or$cond7$i = $546 | $548;
              if (!($or$cond7$i)) {
               label = 151;
               break;
              }
              $549 = ((($P_Left$0113$i)) + 12|0);
              $550 = HEAP16[$549>>1]|0;
              $551 = $550 & 7;
              $552 = ($551<<16>>16)==(2);
              if ($552) {
               $dropouts$2$i = $dropouts$0112$i;
               break;
              }
              HEAP32[$P_Left$0113$i>>2] = $$$i;
              HEAP32[$P_Right$0114$i>>2] = $$4$i;
              $553 = ((($P_Left$0113$i)) + 24|0);
              HEAP32[$553>>2] = 1;
              $554 = (($dropouts$0112$i) + 1)<<16>>16;
              $dropouts$2$i = $554;
             }
            } while(0);
            if ((label|0) == 151) {
             label = 0;
             $555 = HEAP32[$32>>2]|0;
             FUNCTION_TABLE_viiiiii[$555 & 8191]($worker,$y$1121$i,$$$i,$$4$i,$P_Left$0113$i,$P_Right$0114$i);
             $dropouts$2$i = $dropouts$0112$i;
            }
            $556 = ((($P_Left$0113$i)) + 4|0);
            $P_Left$0$i = HEAP32[$556>>2]|0;
            $557 = ($P_Left$0$i|0)==(0|0);
            if ($557) {
             $dropouts$2$i$lcssa = $dropouts$2$i;
             break;
            }
            $558 = ((($P_Right$0114$i)) + 4|0);
            $P_Right$0114$i$pre = HEAP32[$558>>2]|0;
            $P_Left$0113$i = $P_Left$0$i;$P_Right$0114$i = $P_Right$0114$i$pre;$dropouts$0112$i = $dropouts$2$i;
           }
           $$not$i = ($dropouts$2$i$lcssa<<16>>16)<(1);
           if ($$not$i) {
            break;
           } else {
            $P_Left$1118$i = $531;$P_Right$1119$in$i = $draw_right$i;
           }
           while(1) {
            $P_Right$1119$i = HEAP32[$P_Right$1119$in$i>>2]|0;
            $639 = ((($P_Left$1118$i)) + 24|0);
            $640 = HEAP32[$639>>2]|0;
            $641 = ($640|0)==(0);
            if (!($641)) {
             HEAP32[$639>>2] = 0;
             $642 = HEAP32[$31>>2]|0;
             $643 = HEAP32[$P_Left$1118$i>>2]|0;
             $644 = HEAP32[$P_Right$1119$i>>2]|0;
             FUNCTION_TABLE_viiiiii[$642 & 8191]($worker,$y$1121$i,$643,$644,$P_Left$1118$i,$P_Right$1119$i);
            }
            $645 = ((($P_Left$1118$i)) + 4|0);
            $646 = ((($P_Right$1119$i)) + 4|0);
            $P_Left$1$i = HEAP32[$645>>2]|0;
            $647 = ($P_Left$1$i|0)==(0|0);
            if ($647) {
             break;
            } else {
             $P_Left$1118$i = $P_Left$1$i;$P_Right$1119$in$i = $646;
            }
           }
          }
         } while(0);
         $559 = HEAP32[$30>>2]|0;
         FUNCTION_TABLE_vi[$559 & 8191]($worker);
         $560 = (($y$1121$i) + 1)<<16>>16;
         $561 = $560 << 16 >> 16;
         $562 = ($561|0)<($526|0);
         if (!($562)) {
          $612 = $531;$y$1$lcssa$i = $560;
          break L207;
         }
         do {
          if (!($532)) {
           $current$08$i13$i = $531;
           while(1) {
            $563 = ((($current$08$i13$i)) + 8|0);
            $564 = HEAP32[$563>>2]|0;
            $565 = HEAP32[$564>>2]|0;
            HEAP32[$current$08$i13$i>>2] = $565;
            $566 = ((($current$08$i13$i)) + 12|0);
            $567 = HEAP16[$566>>1]|0;
            $568 = $567 & 8;
            $569 = $568&65535;
            $570 = $569 >>> 2;
            $571 = (($570) + -1)|0;
            $572 = (($564) + ($571<<2)|0);
            HEAP32[$563>>2] = $572;
            $573 = ((($current$08$i13$i)) + 16|0);
            $574 = HEAP32[$573>>2]|0;
            $575 = (($574) + -1)|0;
            HEAP32[$573>>2] = $575;
            $576 = ((($current$08$i13$i)) + 4|0);
            $current$0$i14$i = HEAP32[$576>>2]|0;
            $577 = ($current$0$i14$i|0)==(0|0);
            if ($577) {
             break;
            } else {
             $current$08$i13$i = $current$0$i14$i;
            }
           }
           $next$0$in1$i18$i = ((($531)) + 4|0);
           $next$02$i19$i = HEAP32[$next$0$in1$i18$i>>2]|0;
           $578 = ($next$02$i19$i|0)==(0|0);
           if ($578) {
            break;
           } else {
            $current$14$i23$i = $531;$next$0$in5$i22$i = $next$0$in1$i18$i;$next$06$i21$i = $next$02$i19$i;$old$03$i24$i = $draw_left$i;
           }
           while(1) {
            $579 = HEAP32[$current$14$i23$i>>2]|0;
            $580 = HEAP32[$next$06$i21$i>>2]|0;
            $581 = ($579|0)>($580|0);
            if ($581) {
             HEAP32[$old$03$i24$i>>2] = $next$06$i21$i;
             $582 = ((($next$06$i21$i)) + 4|0);
             $583 = HEAP32[$582>>2]|0;
             HEAP32[$next$0$in5$i22$i>>2] = $583;
             HEAP32[$582>>2] = $current$14$i23$i;
             $584 = HEAP32[$draw_left$i>>2]|0;
             $current$1$be$i27$i = $584;$old$0$be$i26$i = $draw_left$i;
            } else {
             $current$1$be$i27$i = $next$06$i21$i;$old$0$be$i26$i = $next$0$in5$i22$i;
            }
            $next$0$in$i28$i = ((($current$1$be$i27$i)) + 4|0);
            $next$0$i29$i = HEAP32[$next$0$in$i28$i>>2]|0;
            $585 = ($next$0$i29$i|0)==(0|0);
            if ($585) {
             break;
            } else {
             $current$14$i23$i = $current$1$be$i27$i;$next$0$in5$i22$i = $next$0$in$i28$i;$next$06$i21$i = $next$0$i29$i;$old$03$i24$i = $old$0$be$i26$i;
            }
           }
          }
         } while(0);
         $586 = HEAP32[$draw_right$i>>2]|0;
         $587 = ($586|0)==(0|0);
         if ($587) {
          $P_Right$0114$i144 = 0;$y$1121$i = $560;
          continue;
         } else {
          $current$08$i$i = $586;
         }
         while(1) {
          $588 = ((($current$08$i$i)) + 8|0);
          $589 = HEAP32[$588>>2]|0;
          $590 = HEAP32[$589>>2]|0;
          HEAP32[$current$08$i$i>>2] = $590;
          $591 = ((($current$08$i$i)) + 12|0);
          $592 = HEAP16[$591>>1]|0;
          $593 = $592 & 8;
          $594 = $593&65535;
          $595 = $594 >>> 2;
          $596 = (($595) + -1)|0;
          $597 = (($589) + ($596<<2)|0);
          HEAP32[$588>>2] = $597;
          $598 = ((($current$08$i$i)) + 16|0);
          $599 = HEAP32[$598>>2]|0;
          $600 = (($599) + -1)|0;
          HEAP32[$598>>2] = $600;
          $601 = ((($current$08$i$i)) + 4|0);
          $current$0$i11$i = HEAP32[$601>>2]|0;
          $602 = ($current$0$i11$i|0)==(0|0);
          if ($602) {
           break;
          } else {
           $current$08$i$i = $current$0$i11$i;
          }
         }
         $next$0$in1$i$i = ((($586)) + 4|0);
         $next$02$i$i = HEAP32[$next$0$in1$i$i>>2]|0;
         $603 = ($next$02$i$i|0)==(0|0);
         if ($603) {
          $P_Right$0114$i144 = $586;$y$1121$i = $560;
          continue;
         } else {
          $P_Right$0114$i143 = $586;$current$14$i$i = $586;$next$0$in5$i$i = $next$0$in1$i$i;$next$06$i$i = $next$02$i$i;$old$03$i$i = $draw_right$i;
         }
         while(1) {
          $604 = HEAP32[$current$14$i$i>>2]|0;
          $605 = HEAP32[$next$06$i$i>>2]|0;
          $606 = ($604|0)>($605|0);
          if ($606) {
           HEAP32[$old$03$i$i>>2] = $next$06$i$i;
           $607 = ((($next$06$i$i)) + 4|0);
           $608 = HEAP32[$607>>2]|0;
           HEAP32[$next$0$in5$i$i>>2] = $608;
           HEAP32[$607>>2] = $current$14$i$i;
           $609 = HEAP32[$draw_right$i>>2]|0;
           $P_Right$0114$i142 = $609;$current$1$be$i$i = $609;$old$0$be$i$i = $draw_right$i;
          } else {
           $P_Right$0114$i142 = $P_Right$0114$i143;$current$1$be$i$i = $next$06$i$i;$old$0$be$i$i = $next$0$in5$i$i;
          }
          $next$0$in$i$i = ((($current$1$be$i$i)) + 4|0);
          $next$0$i$i = HEAP32[$next$0$in$i$i>>2]|0;
          $610 = ($next$0$i$i|0)==(0|0);
          if ($610) {
           $P_Right$0114$i144 = $P_Right$0114$i142;$y$1121$i = $560;
           continue L208;
          } else {
           $P_Right$0114$i143 = $P_Right$0114$i142;$current$14$i$i = $current$1$be$i$i;$next$0$in5$i$i = $next$0$in$i$i;$next$06$i$i = $next$0$i$i;$old$03$i$i = $old$0$be$i$i;
          }
         }
        }
       } else {
        $$pre = HEAP32[$draw_left$i>>2]|0;
        $612 = $$pre;$y$1$lcssa$i = $y$0$i;
       }
      } while(0);
      $611 = ($612|0)==(0|0);
      if (!($611)) {
       $P$3108$i = $612;
       while(1) {
        $613 = ((($P$3108$i)) + 4|0);
        $614 = HEAP32[$613>>2]|0;
        $615 = ((($P$3108$i)) + 16|0);
        $616 = HEAP32[$615>>2]|0;
        $617 = ($616|0)==(0);
        L252: do {
         if ($617) {
          $old$0$i8$i = $draw_left$i;
          while(1) {
           $current$0$i9$i = HEAP32[$old$0$i8$i>>2]|0;
           $619 = ($current$0$i9$i|0)==(0|0);
           if ($619) {
            break L252;
           }
           $620 = ($current$0$i9$i|0)==($P$3108$i|0);
           $621 = ((($current$0$i9$i)) + 4|0);
           if ($620) {
            $$lcssa360 = $621;$old$0$i8$i$lcssa359 = $old$0$i8$i;
            break;
           } else {
            $old$0$i8$i = $621;
           }
          }
          $622 = HEAP32[$$lcssa360>>2]|0;
          HEAP32[$old$0$i8$i$lcssa359>>2] = $622;
         }
        } while(0);
        $618 = ($614|0)==(0|0);
        if ($618) {
         break;
        } else {
         $P$3108$i = $614;
        }
       }
      }
      $623 = HEAP32[$draw_right$i>>2]|0;
      $624 = ($623|0)==(0|0);
      if ($624) {
       $y$0$ph$i = $y$1$lcssa$i;$y_height$0$ph$i = $528;
       label = 113;
       continue;
      } else {
       $P$4107$i = $623;
      }
      while(1) {
       $625 = ((($P$4107$i)) + 4|0);
       $626 = HEAP32[$625>>2]|0;
       $627 = ((($P$4107$i)) + 16|0);
       $628 = HEAP32[$627>>2]|0;
       $629 = ($628|0)==(0);
       L261: do {
        if ($629) {
         $old$0$i$i = $draw_right$i;
         while(1) {
          $current$0$i$i = HEAP32[$old$0$i$i>>2]|0;
          $631 = ($current$0$i$i|0)==(0|0);
          if ($631) {
           break L261;
          }
          $632 = ($current$0$i$i|0)==($P$4107$i|0);
          $633 = ((($current$0$i$i)) + 4|0);
          if ($632) {
           $$lcssa362 = $633;$old$0$i$i$lcssa361 = $old$0$i$i;
           break;
          } else {
           $old$0$i$i = $633;
          }
         }
         $634 = HEAP32[$$lcssa362>>2]|0;
         HEAP32[$old$0$i$i$lcssa361>>2] = $634;
        }
       } while(0);
       $630 = ($626|0)==(0|0);
       if ($630) {
        $y$0$ph$i = $y$1$lcssa$i;$y_height$0$ph$i = $528;
        label = 113;
        continue L156;
       } else {
        $P$4107$i = $626;
       }
      }
     }
     $438 = HEAP16[$max_Y$i>>1]|0;
     $439 = ($y$0$i<<16>>16)>($438<<16>>16);
     if (!($439)) {
      $y$2127$i = $y$0$i;
      while(1) {
       $635 = HEAP32[$30>>2]|0;
       FUNCTION_TABLE_vi[$635 & 8191]($worker);
       $636 = (($y$2127$i) + 1)<<16>>16;
       $637 = HEAP16[$max_Y$i>>1]|0;
       $638 = ($636<<16>>16)>($637<<16>>16);
       if ($638) {
        break;
       } else {
        $y$2127$i = $636;
       }
      }
     }
    }
    $648 = HEAP32[$0>>2]|0;
    $649 = (($648) + -1)|0;
    $storemerge = $649;
   } else {
    label = 91;
   }
  }
  if ((label|0) == 91) {
   label = 0;
   $$pr = HEAP32[$8>>2]|0;
   $357 = ($$pr|0)==(98);
   if ($357) {
    label = 92;
   } else {
    $$0 = 1;
    label = 187;
    break;
   }
  }
  if ((label|0) == 92) {
   label = 0;
   HEAP32[$8>>2] = 0;
   $358 = HEAP32[$0>>2]|0;
   $359 = (((($worker)) + 956|0) + ($358<<2)|0);
   $360 = HEAP16[$359>>1]|0;
   $361 = (((((($worker)) + 956|0) + ($358<<2)|0)) + 2|0);
   $362 = HEAP16[$361>>1]|0;
   $363 = $360 << 16 >> 16;
   $364 = $362 << 16 >> 16;
   $365 = (($364) + ($363))|0;
   $366 = (($365|0) / 2)&-1;
   $367 = ($358|0)>(6);
   $368 = ($366|0)<($363|0);
   $or$cond = $367 | $368;
   if ($or$cond) {
    label = 93;
    break;
   }
   $369 = $366&65535;
   $370 = (($358) + 1)|0;
   $371 = (((($worker)) + 956|0) + ($370<<2)|0);
   HEAP16[$371>>1] = $369;
   $372 = HEAP32[$0>>2]|0;
   $373 = (($372) + 1)|0;
   $374 = (((((($worker)) + 956|0) + ($373<<2)|0)) + 2|0);
   HEAP16[$374>>1] = $362;
   $375 = (($366) + 65535)|0;
   $376 = $375&65535;
   $377 = HEAP32[$0>>2]|0;
   $378 = (((((($worker)) + 956|0) + ($377<<2)|0)) + 2|0);
   HEAP16[$378>>1] = $376;
   $379 = HEAP32[$0>>2]|0;
   $380 = (($379) + 1)|0;
   $storemerge = $380;
  }
  HEAP32[$0>>2] = $storemerge;
  $650 = ($storemerge|0)>(-1);
  if ($650) {
   $34 = $storemerge;
  } else {
   $$0 = 0;
   label = 187;
   break;
  }
 }
 if ((label|0) == 35) {
  HEAP32[$8>>2] = 20;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 else if ((label|0) == 47) {
  HEAP32[$8>>2] = 99;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 else if ((label|0) == 93) {
  HEAP32[$0>>2] = 0;
  HEAP32[$8>>2] = 20;
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 else if ((label|0) == 106) {
  HEAP32[$8>>2] = 20;
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 else if ((label|0) == 187) {
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _Line_To($worker,$x,$y) {
 $worker = $worker|0;
 $x = $x|0;
 $y = $y|0;
 var $$ = 0, $$$v = 0, $$0 = 0, $$28 = 0, $$28$v = 0, $$cast$i = 0, $$cast$i12 = 0, $$cast$i21 = 0, $$cast$i5 = 0, $$cast1$i = 0, $$cast1$i16 = 0, $$in$i = 0, $$in$i13 = 0, $$in$i22 = 0, $$in$i6 = 0, $$not = 0, $$not2 = 0, $$pre = 0, $$pre$i = 0, $$pre$i10 = 0;
 var $$pre$i19 = 0, $$pre$i3 = 0, $$pre$phi33Z2D = 0, $$pre$phiZ2D = 0, $$pre32 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0;
 var $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0;
 var $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0;
 var $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0;
 var $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $storemerge = 0, $storemerge25 = 0, $storemerge26 = 0, $storemerge27 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($worker)) + 100|0);
 $1 = HEAP32[$0>>2]|0;
 L1: do {
  switch ($1|0) {
  case 0:  {
   $2 = ((($worker)) + 72|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = ($3|0)<($y|0);
   if ($4) {
    $5 = ((($worker)) + 4|0);
    $6 = HEAP32[$5>>2]|0;
    $7 = (($3) + -1)|0;
    $8 = (($7) + ($6))|0;
    $9 = (0 - ($6))|0;
    $10 = $8 & $9;
    $11 = (($10) - ($3))|0;
    $12 = ((($worker)) + 8|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ($11|0)<($13|0);
    $15 = ((($worker)) + 92|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ($16|0)==(0|0);
    $18 = ((($worker)) + 40|0);
    if ($17) {
     $19 = HEAP32[$18>>2]|0;
     $20 = ((($worker)) + 88|0);
     HEAP32[$20>>2] = $19;
     HEAP32[$15>>2] = $19;
     $$cast$i = $19;
     $21 = ((($$cast$i)) + 32|0);
     HEAP32[$18>>2] = $21;
     $$in$i = $21;
    } else {
     $$pre$i = HEAP32[$18>>2]|0;
     $$in$i = $$pre$i;
    }
    $22 = ((($worker)) + 36|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = ($$in$i>>>0)<($23>>>0);
    if (!($24)) {
     $43 = ((($worker)) + 44|0);
     HEAP32[$43>>2] = 98;
     $$0 = 1;
     return ($$0|0);
    }
    $25 = $$in$i;
    $26 = ((($worker)) + 88|0);
    $27 = HEAP32[$26>>2]|0;
    $28 = ((($27)) + 12|0);
    $29 = ((($27)) + 20|0);
    HEAP32[$29>>2] = 0;
    $30 = ((($27)) + 16|0);
    HEAP32[$30>>2] = 0;
    $31 = ((($27)) + 8|0);
    HEAP32[$31>>2] = $25;
    $32 = ((($27)) + 4|0);
    HEAP32[$32>>2] = 0;
    $33 = ((($27)) + 28|0);
    HEAP32[$33>>2] = 0;
    $34 = ((($worker)) + 176|0);
    $35 = HEAP8[$34>>0]|0;
    $36 = $35&255;
    $$$v = $14 ? 8 : 40;
    $$ = $36 | $$$v;
    $storemerge27 = $$&65535;
    HEAP16[$28>>1] = $storemerge27;
    $37 = ((($worker)) + 96|0);
    $38 = HEAP32[$37>>2]|0;
    $39 = ($38|0)==(0|0);
    if ($39) {
     $40 = $27;
     HEAP32[$37>>2] = $40;
    }
    HEAP32[$0>>2] = 1;
    $41 = ((($worker)) + 86|0);
    HEAP8[$41>>0] = 1;
    $42 = ((($worker)) + 87|0);
    HEAP8[$42>>0] = 0;
    label = 61;
    break L1;
   }
   $44 = ($3|0)>($y|0);
   if ($44) {
    $45 = ((($worker)) + 4|0);
    $46 = HEAP32[$45>>2]|0;
    $$not2 = (($46) + -1)|0;
    $47 = $$not2 & $3;
    $48 = ((($worker)) + 8|0);
    $49 = HEAP32[$48>>2]|0;
    $50 = ($47|0)<($49|0);
    $51 = ((($worker)) + 92|0);
    $52 = HEAP32[$51>>2]|0;
    $53 = ($52|0)==(0|0);
    $54 = ((($worker)) + 40|0);
    if ($53) {
     $55 = HEAP32[$54>>2]|0;
     $56 = ((($worker)) + 88|0);
     HEAP32[$56>>2] = $55;
     HEAP32[$51>>2] = $55;
     $$cast$i5 = $55;
     $57 = ((($$cast$i5)) + 32|0);
     HEAP32[$54>>2] = $57;
     $$in$i6 = $57;
    } else {
     $$pre$i3 = HEAP32[$54>>2]|0;
     $$in$i6 = $$pre$i3;
    }
    $58 = ((($worker)) + 36|0);
    $59 = HEAP32[$58>>2]|0;
    $60 = ($$in$i6>>>0)<($59>>>0);
    if (!($60)) {
     $82 = ((($worker)) + 44|0);
     HEAP32[$82>>2] = 98;
     $$0 = 1;
     return ($$0|0);
    }
    $61 = $$in$i6;
    $62 = ((($worker)) + 88|0);
    $63 = HEAP32[$62>>2]|0;
    $64 = ((($63)) + 12|0);
    $65 = ((($63)) + 20|0);
    HEAP32[$65>>2] = 0;
    $66 = ((($63)) + 16|0);
    HEAP32[$66>>2] = 0;
    $67 = ((($63)) + 8|0);
    HEAP32[$67>>2] = $61;
    $68 = ((($63)) + 4|0);
    HEAP32[$68>>2] = 0;
    $69 = ((($63)) + 28|0);
    HEAP32[$69>>2] = 0;
    $70 = ((($worker)) + 176|0);
    $71 = HEAP8[$70>>0]|0;
    $72 = $71&255;
    $73 = $63;
    if ($50) {
     $storemerge26 = $72;
    } else {
     $74 = $71&255;
     $75 = $74 | 16;
     $76 = $75&65535;
     $storemerge26 = $76;
    }
    HEAP16[$64>>1] = $storemerge26;
    $77 = ((($worker)) + 96|0);
    $78 = HEAP32[$77>>2]|0;
    $79 = ($78|0)==(0|0);
    if ($79) {
     HEAP32[$77>>2] = $73;
    }
    HEAP32[$0>>2] = 2;
    $80 = ((($worker)) + 86|0);
    HEAP8[$80>>0] = 1;
    $81 = ((($worker)) + 87|0);
    HEAP8[$81>>0] = 0;
    label = 62;
   } else {
    label = 60;
   }
   break;
  }
  case 1:  {
   $83 = ((($worker)) + 72|0);
   $84 = HEAP32[$83>>2]|0;
   $85 = ($84|0)>($y|0);
   if ($85) {
    $86 = ((($worker)) + 4|0);
    $87 = HEAP32[$86>>2]|0;
    $$not = (($87) + -1)|0;
    $88 = $$not & $84;
    $89 = ((($worker)) + 8|0);
    $90 = HEAP32[$89>>2]|0;
    $91 = ($88|0)<($90|0);
    $92 = ((($worker)) + 40|0);
    $93 = HEAP32[$92>>2]|0;
    $94 = ((($worker)) + 88|0);
    $95 = HEAP32[$94>>2]|0;
    $96 = ((($95)) + 8|0);
    $97 = HEAP32[$96>>2]|0;
    $98 = (($93) - ($97))|0;
    $99 = $98 >> 2;
    $100 = ($98|0)<(0);
    $101 = $93;
    if ($100) {
     $102 = ((($worker)) + 44|0);
     HEAP32[$102>>2] = 99;
     $$0 = 1;
     return ($$0|0);
    }
    $103 = ($98|0)>(0);
    if ($103) {
     $104 = ((($95)) + 16|0);
     HEAP32[$104>>2] = $99;
     do {
      if (!($91)) {
       $105 = ((($95)) + 12|0);
       $106 = HEAP16[$105>>1]|0;
       $107 = $106&65535;
       $108 = $107 & 8;
       $109 = ($108|0)==(0);
       if ($109) {
        $112 = $107 | 32;
        $113 = $112&65535;
        HEAP16[$105>>1] = $113;
        break;
       } else {
        $110 = $107 | 16;
        $111 = $110&65535;
        HEAP16[$105>>1] = $111;
        break;
       }
      }
     } while(0);
     HEAP32[$94>>2] = $93;
     $114 = ((($101)) + 32|0);
     HEAP32[$92>>2] = $114;
     $$cast1$i = $93;
     $115 = ((($$cast1$i)) + 16|0);
     HEAP32[$115>>2] = 0;
     $116 = $114;
     $117 = ((($$cast1$i)) + 8|0);
     HEAP32[$117>>2] = $116;
     $118 = ((($95)) + 28|0);
     HEAP32[$118>>2] = $93;
     $119 = ((($worker)) + 84|0);
     $120 = HEAP16[$119>>1]|0;
     $121 = (($120) + 1)<<16>>16;
     HEAP16[$119>>1] = $121;
     $$pre$i19 = $114;$130 = $116;$260 = $$cast1$i;
    } else {
     $$pre$i19 = $101;$130 = $93;$260 = $95;
    }
    $122 = ((($worker)) + 36|0);
    $123 = HEAP32[$122>>2]|0;
    $124 = ($$pre$i19>>>0)<($123>>>0);
    if (!($124)) {
     $125 = ((($worker)) + 44|0);
     HEAP32[$125>>2] = 98;
     $$0 = 1;
     return ($$0|0);
    }
    $126 = ((($worker)) + 87|0);
    HEAP8[$126>>0] = 0;
    $127 = ((($worker)) + 92|0);
    $128 = HEAP32[$127>>2]|0;
    $129 = ($128|0)==(0|0);
    if ($129) {
     HEAP32[$94>>2] = $130;
     HEAP32[$127>>2] = $130;
     $$cast$i21 = $130;
     $131 = ((($$cast$i21)) + 32|0);
     HEAP32[$92>>2] = $131;
     $132 = $130;
     $$in$i22 = $131;$136 = $132;
    } else {
     $$in$i22 = $$pre$i19;$136 = $260;
    }
    $133 = ($$in$i22>>>0)<($123>>>0);
    if (!($133)) {
     $153 = ((($worker)) + 44|0);
     HEAP32[$153>>2] = 98;
     $$0 = 1;
     return ($$0|0);
    }
    $134 = $$in$i22;
    $135 = ((($136)) + 12|0);
    $137 = ((($136)) + 20|0);
    HEAP32[$137>>2] = 0;
    $138 = ((($136)) + 16|0);
    HEAP32[$138>>2] = 0;
    $139 = ((($136)) + 8|0);
    HEAP32[$139>>2] = $134;
    $140 = ((($136)) + 4|0);
    HEAP32[$140>>2] = 0;
    $141 = ((($136)) + 28|0);
    HEAP32[$141>>2] = 0;
    $142 = ((($worker)) + 176|0);
    $143 = HEAP8[$142>>0]|0;
    $144 = $143&255;
    $145 = $136;
    if ($91) {
     $storemerge25 = $144;
    } else {
     $146 = $143&255;
     $147 = $146 | 16;
     $148 = $147&65535;
     $storemerge25 = $148;
    }
    HEAP16[$135>>1] = $storemerge25;
    $149 = ((($worker)) + 96|0);
    $150 = HEAP32[$149>>2]|0;
    $151 = ($150|0)==(0|0);
    if ($151) {
     HEAP32[$149>>2] = $145;
    }
    HEAP32[$0>>2] = 2;
    $152 = ((($worker)) + 86|0);
    HEAP8[$152>>0] = 1;
    HEAP8[$126>>0] = 0;
    label = 62;
   } else {
    label = 61;
   }
   break;
  }
  case 2:  {
   $154 = ((($worker)) + 72|0);
   $155 = HEAP32[$154>>2]|0;
   $156 = ($155|0)<($y|0);
   if ($156) {
    $157 = ((($worker)) + 4|0);
    $158 = HEAP32[$157>>2]|0;
    $159 = (($155) + -1)|0;
    $160 = (($159) + ($158))|0;
    $161 = (0 - ($158))|0;
    $162 = $160 & $161;
    $163 = (($162) - ($155))|0;
    $164 = ((($worker)) + 8|0);
    $165 = HEAP32[$164>>2]|0;
    $166 = ($163|0)<($165|0);
    $167 = ((($worker)) + 40|0);
    $168 = HEAP32[$167>>2]|0;
    $169 = ((($worker)) + 88|0);
    $170 = HEAP32[$169>>2]|0;
    $171 = ((($170)) + 8|0);
    $172 = HEAP32[$171>>2]|0;
    $173 = (($168) - ($172))|0;
    $174 = $173 >> 2;
    $175 = ($173|0)<(0);
    $176 = $168;
    if ($175) {
     $177 = ((($worker)) + 44|0);
     HEAP32[$177>>2] = 99;
     $$0 = 1;
     return ($$0|0);
    }
    $178 = ($173|0)>(0);
    if ($178) {
     $179 = ((($170)) + 16|0);
     HEAP32[$179>>2] = $174;
     do {
      if (!($166)) {
       $180 = ((($170)) + 12|0);
       $181 = HEAP16[$180>>1]|0;
       $182 = $181&65535;
       $183 = $182 & 8;
       $184 = ($183|0)==(0);
       if ($184) {
        $187 = $182 | 32;
        $188 = $187&65535;
        HEAP16[$180>>1] = $188;
        break;
       } else {
        $185 = $182 | 16;
        $186 = $185&65535;
        HEAP16[$180>>1] = $186;
        break;
       }
      }
     } while(0);
     HEAP32[$169>>2] = $168;
     $189 = ((($176)) + 32|0);
     HEAP32[$167>>2] = $189;
     $$cast1$i16 = $168;
     $190 = ((($$cast1$i16)) + 16|0);
     HEAP32[$190>>2] = 0;
     $191 = $189;
     $192 = ((($$cast1$i16)) + 8|0);
     HEAP32[$192>>2] = $191;
     $193 = ((($170)) + 28|0);
     HEAP32[$193>>2] = $168;
     $194 = ((($worker)) + 84|0);
     $195 = HEAP16[$194>>1]|0;
     $196 = (($195) + 1)<<16>>16;
     HEAP16[$194>>1] = $196;
     $$pre$i10 = $189;$205 = $191;$261 = $$cast1$i16;
    } else {
     $$pre$i10 = $176;$205 = $168;$261 = $170;
    }
    $197 = ((($worker)) + 36|0);
    $198 = HEAP32[$197>>2]|0;
    $199 = ($$pre$i10>>>0)<($198>>>0);
    if (!($199)) {
     $200 = ((($worker)) + 44|0);
     HEAP32[$200>>2] = 98;
     $$0 = 1;
     return ($$0|0);
    }
    $201 = ((($worker)) + 87|0);
    HEAP8[$201>>0] = 0;
    $202 = ((($worker)) + 92|0);
    $203 = HEAP32[$202>>2]|0;
    $204 = ($203|0)==(0|0);
    if ($204) {
     HEAP32[$169>>2] = $205;
     HEAP32[$202>>2] = $205;
     $$cast$i12 = $205;
     $206 = ((($$cast$i12)) + 32|0);
     HEAP32[$167>>2] = $206;
     $207 = $205;
     $$in$i13 = $206;$211 = $207;
    } else {
     $$in$i13 = $$pre$i10;$211 = $261;
    }
    $208 = ($$in$i13>>>0)<($198>>>0);
    if (!($208)) {
     $225 = ((($worker)) + 44|0);
     HEAP32[$225>>2] = 98;
     $$0 = 1;
     return ($$0|0);
    }
    $209 = $$in$i13;
    $210 = ((($211)) + 12|0);
    $212 = ((($211)) + 20|0);
    HEAP32[$212>>2] = 0;
    $213 = ((($211)) + 16|0);
    HEAP32[$213>>2] = 0;
    $214 = ((($211)) + 8|0);
    HEAP32[$214>>2] = $209;
    $215 = ((($211)) + 4|0);
    HEAP32[$215>>2] = 0;
    $216 = ((($211)) + 28|0);
    HEAP32[$216>>2] = 0;
    $217 = ((($worker)) + 176|0);
    $218 = HEAP8[$217>>0]|0;
    $219 = $218&255;
    $$28$v = $166 ? 8 : 40;
    $$28 = $219 | $$28$v;
    $storemerge = $$28&65535;
    HEAP16[$210>>1] = $storemerge;
    $220 = ((($worker)) + 96|0);
    $221 = HEAP32[$220>>2]|0;
    $222 = ($221|0)==(0|0);
    if ($222) {
     $223 = $211;
     HEAP32[$220>>2] = $223;
    }
    HEAP32[$0>>2] = 1;
    $224 = ((($worker)) + 86|0);
    HEAP8[$224>>0] = 1;
    HEAP8[$201>>0] = 0;
    label = 61;
   } else {
    label = 62;
   }
   break;
  }
  default: {
   label = 60;
  }
  }
 } while(0);
 if ((label|0) == 60) {
  $$pre = ((($worker)) + 68|0);
  $$pre32 = ((($worker)) + 72|0);
  $$pre$phi33Z2D = $$pre32;$$pre$phiZ2D = $$pre;
 }
 else if ((label|0) == 61) {
  $226 = ((($worker)) + 68|0);
  $227 = HEAP32[$226>>2]|0;
  $228 = ((($worker)) + 72|0);
  $229 = HEAP32[$228>>2]|0;
  $230 = ((($worker)) + 76|0);
  $231 = HEAP32[$230>>2]|0;
  $232 = ((($worker)) + 80|0);
  $233 = HEAP32[$232>>2]|0;
  $234 = (_Line_Up($worker,$227,$229,$x,$y,$231,$233)|0);
  $235 = ($234<<24>>24)==(0);
  if ($235) {
   $$pre$phi33Z2D = $228;$$pre$phiZ2D = $226;
  } else {
   $$0 = 1;
   return ($$0|0);
  }
 }
 else if ((label|0) == 62) {
  $236 = ((($worker)) + 68|0);
  $237 = HEAP32[$236>>2]|0;
  $238 = ((($worker)) + 72|0);
  $239 = HEAP32[$238>>2]|0;
  $240 = ((($worker)) + 76|0);
  $241 = HEAP32[$240>>2]|0;
  $242 = ((($worker)) + 80|0);
  $243 = HEAP32[$242>>2]|0;
  $244 = ((($worker)) + 86|0);
  $245 = HEAP8[$244>>0]|0;
  $246 = (0 - ($239))|0;
  $247 = (0 - ($y))|0;
  $248 = (0 - ($243))|0;
  $249 = (0 - ($241))|0;
  $250 = (_Line_Up($worker,$237,$246,$x,$247,$248,$249)|0);
  $251 = ($245<<24>>24)==(0);
  if (!($251)) {
   $252 = HEAP8[$244>>0]|0;
   $253 = ($252<<24>>24)==(0);
   if ($253) {
    $254 = ((($worker)) + 88|0);
    $255 = HEAP32[$254>>2]|0;
    $256 = ((($255)) + 20|0);
    $257 = HEAP32[$256>>2]|0;
    $258 = (0 - ($257))|0;
    HEAP32[$256>>2] = $258;
   }
  }
  $259 = ($250<<24>>24)==(0);
  if ($259) {
   $$pre$phi33Z2D = $238;$$pre$phiZ2D = $236;
  } else {
   $$0 = 1;
   return ($$0|0);
  }
 }
 HEAP32[$$pre$phiZ2D>>2] = $x;
 HEAP32[$$pre$phi33Z2D>>2] = $y;
 $$0 = 0;
 return ($$0|0);
}
function _Line_Up($worker,$x1,$y1,$x2,$y2,$miny,$maxy) {
 $worker = $worker|0;
 $x1 = $x1|0;
 $y1 = $y1|0;
 $x2 = $x2|0;
 $y2 = $y2|0;
 $miny = $miny|0;
 $maxy = $maxy|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$23 = 0, $$3 = 0, $$pn = 0, $$pre = 0, $$pre$phi7Z2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $8 = 0, $9 = 0, $Ax$05 = 0, $Ax$1 = 0, $Dx$0 = 0, $Ix$0 = 0, $Rx$0 = 0, $e1$0 = 0, $e1$1 = 0, $e2$0 = 0, $f1$0 = 0, $f2$0 = 0, $maxy$pn = 0, $or$cond = 0, $or$cond2 = 0, $size$04 = 0, $top$0$lcssa = 0, $top$06 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($x2) - ($x1))|0;
 $1 = (($y2) - ($y1))|0;
 $2 = ($1|0)<(1);
 $3 = ($y2|0)<($miny|0);
 $or$cond = $2 | $3;
 $4 = ($y1|0)>($maxy|0);
 $or$cond2 = $or$cond | $4;
 if ($or$cond2) {
  $$0 = 0;
  return ($$0|0);
 }
 $5 = ($miny|0)>($y1|0);
 if ($5) {
  $6 = (($miny) - ($y1))|0;
  $7 = (_FT_MulDiv($0,$6,$1)|0);
  $8 = (($7) + ($x1))|0;
  $9 = HEAP32[$worker>>2]|0;
  $10 = $miny >> $9;
  $$01 = $8;$22 = $9;$e1$0 = $10;$f1$0 = 0;
 } else {
  $11 = HEAP32[$worker>>2]|0;
  $12 = $y1 >> $11;
  $13 = ((($worker)) + 4|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = (($14) + -1)|0;
  $16 = $15 & $y1;
  $$01 = $x1;$22 = $11;$e1$0 = $12;$f1$0 = $16;
 }
 $17 = ($y2|0)>($maxy|0);
 if ($17) {
  $f2$0 = 0;$maxy$pn = $maxy;
 } else {
  $18 = ((($worker)) + 4|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = (($19) + -1)|0;
  $21 = $20 & $y2;
  $f2$0 = $21;$maxy$pn = $y2;
 }
 $e2$0 = $maxy$pn >> $22;
 $23 = ($f1$0|0)>(0);
 do {
  if ($23) {
   $24 = ($e1$0|0)==($e2$0|0);
   if ($24) {
    $$0 = 0;
    return ($$0|0);
   } else {
    $25 = ((($worker)) + 4|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = (($26) - ($f1$0))|0;
    $28 = (_FT_MulDiv($0,$27,$1)|0);
    $29 = (($28) + ($$01))|0;
    $30 = (($e1$0) + 1)|0;
    $$pre = ((($worker)) + 87|0);
    $$1 = $29;$$pre$phi7Z2D = $$pre;$e1$1 = $30;
    break;
   }
  } else {
   $31 = ((($worker)) + 87|0);
   $32 = HEAP8[$31>>0]|0;
   $33 = ($32<<24>>24)==(0);
   if ($33) {
    $$1 = $$01;$$pre$phi7Z2D = $31;$e1$1 = $e1$0;
   } else {
    $34 = ((($worker)) + 40|0);
    $35 = HEAP32[$34>>2]|0;
    $36 = ((($35)) + -4|0);
    HEAP32[$34>>2] = $36;
    HEAP8[$31>>0] = 0;
    $$1 = $$01;$$pre$phi7Z2D = $31;$e1$1 = $e1$0;
   }
  }
 } while(0);
 $37 = ($f2$0|0)==(0);
 $38 = $37&1;
 HEAP8[$$pre$phi7Z2D>>0] = $38;
 $39 = ((($worker)) + 86|0);
 $40 = HEAP8[$39>>0]|0;
 $41 = ($40<<24>>24)==(0);
 if (!($41)) {
  $42 = ((($worker)) + 88|0);
  $43 = HEAP32[$42>>2]|0;
  $44 = ((($43)) + 20|0);
  HEAP32[$44>>2] = $e1$1;
  HEAP8[$39>>0] = 0;
 }
 $45 = (($e2$0) - ($e1$1))|0;
 $46 = (($45) + 1)|0;
 $47 = ((($worker)) + 40|0);
 $48 = HEAP32[$47>>2]|0;
 $49 = (($48) + ($46<<2)|0);
 $50 = ((($worker)) + 36|0);
 $51 = HEAP32[$50>>2]|0;
 $52 = ($49>>>0)<($51>>>0);
 if (!($52)) {
  $53 = ((($worker)) + 44|0);
  HEAP32[$53>>2] = 98;
  $$0 = 1;
  return ($$0|0);
 }
 $54 = ($0|0)>(0);
 $55 = ((($worker)) + 4|0);
 $56 = HEAP32[$55>>2]|0;
 if ($54) {
  $57 = (_FT_MulDiv_No_Round($56,$0,$1)|0);
  $58 = HEAP32[$55>>2]|0;
  $59 = Math_imul($58, $0)|0;
  $$pn = $59;$Dx$0 = 1;$Ix$0 = $57;
 } else {
  $60 = (0 - ($0))|0;
  $61 = (_FT_MulDiv_No_Round($56,$60,$1)|0);
  $62 = (0 - ($61))|0;
  $63 = HEAP32[$55>>2]|0;
  $64 = Math_imul($63, $60)|0;
  $$pn = $64;$Dx$0 = -1;$Ix$0 = $62;
 }
 $Rx$0 = (($$pn|0) % ($1|0))&-1;
 $65 = HEAP32[$47>>2]|0;
 $66 = ($45|0)>(-1);
 if ($66) {
  $67 = (0 - ($1))|0;
  $$23 = $$1;$Ax$05 = $67;$size$04 = $46;$top$06 = $65;
  while(1) {
   $68 = ((($top$06)) + 4|0);
   HEAP32[$top$06>>2] = $$23;
   $69 = (($$23) + ($Ix$0))|0;
   $70 = (($Ax$05) + ($Rx$0))|0;
   $71 = ($70|0)>(-1);
   $72 = $71 ? $Dx$0 : 0;
   $$3 = (($69) + ($72))|0;
   $73 = $71 ? $1 : 0;
   $Ax$1 = (($70) - ($73))|0;
   $74 = (($size$04) + -1)|0;
   $75 = ($size$04|0)>(1);
   if ($75) {
    $$23 = $$3;$Ax$05 = $Ax$1;$size$04 = $74;$top$06 = $68;
   } else {
    $top$0$lcssa = $68;
    break;
   }
  }
 } else {
  $top$0$lcssa = $65;
 }
 HEAP32[$47>>2] = $top$0$lcssa;
 $$0 = 0;
 return ($$0|0);
}
function _Conic_To($worker,$cx,$cy,$x,$y) {
 $worker = $worker|0;
 $cx = $cx|0;
 $cy = $cy|0;
 $x = $x|0;
 $y = $y|0;
 var $$ = 0, $$02 = 0, $$3 = 0, $$cast$i = 0, $$cast1$i = 0, $$in$i = 0, $$lcssa56 = 0, $$lcssa60 = 0, $$not = 0, $$phi$trans$insert = 0, $$phi$trans$insert23 = 0, $$phi$trans$insert25 = 0, $$pre = 0, $$pre$i = 0, $$pre24 = 0, $$pre26 = 0, $$pre28 = 0, $$sink = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($worker)) + 180|0);
 $1 = ((($worker)) + 52|0);
 HEAP32[$1>>2] = $0;
 $2 = ((($worker)) + 68|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($worker)) + 196|0);
 HEAP32[$4>>2] = $3;
 $5 = ((($worker)) + 72|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($worker)) + 200|0);
 HEAP32[$7>>2] = $6;
 $8 = ((($worker)) + 188|0);
 HEAP32[$8>>2] = $cx;
 $9 = ((($worker)) + 192|0);
 HEAP32[$9>>2] = $cy;
 HEAP32[$0>>2] = $x;
 $10 = ((($worker)) + 184|0);
 HEAP32[$10>>2] = $y;
 $11 = ((($worker)) + 100|0);
 $12 = ((($worker)) + 76|0);
 $13 = ((($worker)) + 80|0);
 $14 = ((($worker)) + 86|0);
 $15 = ((($worker)) + 88|0);
 $16 = ((($worker)) + 4|0);
 $17 = ((($worker)) + 8|0);
 $18 = ((($worker)) + 92|0);
 $19 = ((($worker)) + 40|0);
 $20 = ((($worker)) + 36|0);
 $21 = ((($worker)) + 176|0);
 $22 = ((($worker)) + 96|0);
 $23 = ((($worker)) + 87|0);
 $24 = ((($worker)) + 84|0);
 $26 = $0;$30 = $6;$31 = $y;$33 = $cy;$44 = $x;
 L1: while(1) {
  $25 = ((($26)) + 20|0);
  $27 = ((($26)) + 12|0);
  $28 = ((($26)) + 4|0);
  $29 = ($30|0)>($31|0);
  $$ = $29 ? $30 : $31;
  $$3 = $29 ? $31 : $30;
  $32 = ($33|0)<($$3|0);
  $34 = ($33|0)>($$|0);
  $or$cond = $32 | $34;
  do {
   if ($or$cond) {
    $35 = ((($26)) + 16|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ((($26)) + 32|0);
    HEAP32[$37>>2] = $36;
    $38 = ((($26)) + 8|0);
    $39 = HEAP32[$38>>2]|0;
    $40 = (($39) + ($36))|0;
    $41 = (($40|0) / 2)&-1;
    $42 = ((($26)) + 24|0);
    HEAP32[$42>>2] = $41;
    $43 = (($44) + ($39))|0;
    $45 = (($43|0) / 2)&-1;
    HEAP32[$38>>2] = $45;
    $46 = (($45) + ($41))|0;
    $47 = (($46|0) / 2)&-1;
    HEAP32[$35>>2] = $47;
    $48 = ((($26)) + 36|0);
    HEAP32[$48>>2] = $30;
    $49 = (($33) + ($30))|0;
    $50 = (($49|0) / 2)&-1;
    $51 = ((($26)) + 28|0);
    HEAP32[$51>>2] = $50;
    $52 = (($31) + ($33))|0;
    $53 = (($52|0) / 2)&-1;
    HEAP32[$27>>2] = $53;
    $54 = (($53) + ($50))|0;
    $55 = (($54|0) / 2)&-1;
    HEAP32[$25>>2] = $55;
    $56 = ((($26)) + 16|0);
    HEAP32[$1>>2] = $56;
   } else {
    $57 = ($30|0)==($31|0);
    if ($57) {
     $58 = ((($26)) + -16|0);
     HEAP32[$1>>2] = $58;
     break;
    }
    $59 = ($30|0)<($31|0);
    $60 = $59 ? 1 : 2;
    $61 = HEAP32[$11>>2]|0;
    $62 = ($61|0)==($60|0);
    if (!($62)) {
     $63 = HEAP32[$16>>2]|0;
     if ($59) {
      $64 = (($30) + -1)|0;
      $65 = (($64) + ($63))|0;
      $66 = (0 - ($63))|0;
      $67 = $65 & $66;
      $68 = (($67) - ($30))|0;
      $69 = HEAP32[$17>>2]|0;
      $70 = ($68|0)>=($69|0);
      $$sink = $70;
     } else {
      $$not = (($63) + -1)|0;
      $71 = $$not & $30;
      $72 = HEAP32[$17>>2]|0;
      $73 = ($71|0)>=($72|0);
      $$sink = $73;
     }
     $74 = ($61|0)==(0);
     if (!($74)) {
      $75 = HEAP32[$19>>2]|0;
      $76 = HEAP32[$15>>2]|0;
      $77 = ((($76)) + 8|0);
      $78 = HEAP32[$77>>2]|0;
      $79 = (($75) - ($78))|0;
      $80 = $79 >> 2;
      $81 = ($79|0)<(0);
      $82 = $75;
      if ($81) {
       label = 12;
       break L1;
      }
      $84 = ($79|0)>(0);
      if ($84) {
       $85 = ((($76)) + 16|0);
       HEAP32[$85>>2] = $80;
       do {
        if ($$sink) {
         $86 = ((($76)) + 12|0);
         $87 = HEAP16[$86>>1]|0;
         $88 = $87&65535;
         $89 = $88 & 8;
         $90 = ($89|0)==(0);
         if ($90) {
          $93 = $88 | 32;
          $94 = $93&65535;
          HEAP16[$86>>1] = $94;
          break;
         } else {
          $91 = $88 | 16;
          $92 = $91&65535;
          HEAP16[$86>>1] = $92;
          break;
         }
        }
       } while(0);
       HEAP32[$15>>2] = $75;
       $95 = ((($82)) + 32|0);
       HEAP32[$19>>2] = $95;
       $$cast1$i = $75;
       $96 = ((($$cast1$i)) + 16|0);
       HEAP32[$96>>2] = 0;
       $97 = $95;
       $98 = ((($$cast1$i)) + 8|0);
       HEAP32[$98>>2] = $97;
       $99 = ((($76)) + 28|0);
       HEAP32[$99>>2] = $75;
       $100 = HEAP16[$24>>1]|0;
       $101 = (($100) + 1)<<16>>16;
       HEAP16[$24>>1] = $101;
       $104 = $95;
      } else {
       $104 = $82;
      }
      $102 = HEAP32[$20>>2]|0;
      $103 = ($104>>>0)<($102>>>0);
      if (!($103)) {
       label = 20;
       break L1;
      }
      HEAP8[$23>>0] = 0;
     }
     $106 = HEAP32[$18>>2]|0;
     $107 = ($106|0)==(0|0);
     if ($107) {
      $108 = HEAP32[$19>>2]|0;
      HEAP32[$15>>2] = $108;
      HEAP32[$18>>2] = $108;
      $$cast$i = $108;
      $109 = ((($$cast$i)) + 32|0);
      HEAP32[$19>>2] = $109;
      $$in$i = $109;
     } else {
      $$pre$i = HEAP32[$19>>2]|0;
      $$in$i = $$pre$i;
     }
     $110 = HEAP32[$20>>2]|0;
     $111 = ($$in$i>>>0)<($110>>>0);
     if (!($111)) {
      label = 26;
      break L1;
     }
     $113 = $$in$i;
     $114 = HEAP32[$15>>2]|0;
     $115 = ((($114)) + 12|0);
     $116 = ((($114)) + 20|0);
     HEAP32[$116>>2] = 0;
     $117 = ((($114)) + 16|0);
     HEAP32[$117>>2] = 0;
     $118 = ((($114)) + 8|0);
     HEAP32[$118>>2] = $113;
     $119 = ((($114)) + 4|0);
     HEAP32[$119>>2] = 0;
     $120 = ((($114)) + 28|0);
     HEAP32[$120>>2] = 0;
     $121 = HEAP8[$21>>0]|0;
     $122 = $121&255;
     HEAP16[$115>>1] = $122;
     $123 = $114;
     if ($59) {
      $124 = $121&255;
      $125 = $124 | 8;
      $126 = $125&65535;
      HEAP16[$115>>1] = $126;
      if ($$sink) {
       $127 = $124 | 40;
       $128 = $127&65535;
       HEAP16[$115>>1] = $128;
      }
     } else {
      if ($$sink) {
       $129 = $121&255;
       $130 = $129 | 16;
       $131 = $130&65535;
       HEAP16[$115>>1] = $131;
      }
     }
     $132 = HEAP32[$22>>2]|0;
     $133 = ($132|0)==(0|0);
     if ($133) {
      HEAP32[$22>>2] = $123;
     }
     HEAP32[$11>>2] = $60;
     HEAP8[$14>>0] = 1;
     HEAP8[$23>>0] = 0;
    }
    $134 = HEAP32[$12>>2]|0;
    $135 = HEAP32[$13>>2]|0;
    if ($59) {
     $136 = (_Bezier_Up($worker,2,4390,$134,$135)|0);
     $137 = ($136<<24>>24)==(0);
     if ($137) {
      break;
     } else {
      $$02 = 1;
      label = 44;
      break L1;
     }
    }
    $138 = (0 - ($31))|0;
    HEAP32[$28>>2] = $138;
    $139 = (0 - ($33))|0;
    HEAP32[$27>>2] = $139;
    $140 = (0 - ($30))|0;
    HEAP32[$25>>2] = $140;
    $141 = HEAP8[$14>>0]|0;
    $142 = (0 - ($135))|0;
    $143 = (0 - ($134))|0;
    $144 = (_Bezier_Up($worker,2,4390,$142,$143)|0);
    $145 = ($141<<24>>24)==(0);
    if (!($145)) {
     $146 = HEAP8[$14>>0]|0;
     $147 = ($146<<24>>24)==(0);
     if ($147) {
      $148 = HEAP32[$15>>2]|0;
      $149 = ((($148)) + 20|0);
      $150 = HEAP32[$149>>2]|0;
      $151 = (0 - ($150))|0;
      HEAP32[$149>>2] = $151;
     }
    }
    $152 = HEAP32[$28>>2]|0;
    $153 = (0 - ($152))|0;
    HEAP32[$28>>2] = $153;
    $154 = ($144<<24>>24)==(0);
    if (!($154)) {
     $$02 = 1;
     label = 44;
     break L1;
    }
   }
  } while(0);
  $155 = HEAP32[$1>>2]|0;
  $156 = ($155>>>0)<($0>>>0);
  if ($156) {
   $$lcssa56 = $44;$$lcssa60 = $31;
   label = 43;
   break;
  }
  $$phi$trans$insert = ((($155)) + 20|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $$phi$trans$insert23 = ((($155)) + 12|0);
  $$pre24 = HEAP32[$$phi$trans$insert23>>2]|0;
  $$phi$trans$insert25 = ((($155)) + 4|0);
  $$pre26 = HEAP32[$$phi$trans$insert25>>2]|0;
  $$pre28 = HEAP32[$155>>2]|0;
  $26 = $155;$30 = $$pre;$31 = $$pre26;$33 = $$pre24;$44 = $$pre28;
 }
 if ((label|0) == 12) {
  $83 = ((($worker)) + 44|0);
  HEAP32[$83>>2] = 99;
  $$02 = 1;
  return ($$02|0);
 }
 else if ((label|0) == 20) {
  $105 = ((($worker)) + 44|0);
  HEAP32[$105>>2] = 98;
  $$02 = 1;
  return ($$02|0);
 }
 else if ((label|0) == 26) {
  $112 = ((($worker)) + 44|0);
  HEAP32[$112>>2] = 98;
  $$02 = 1;
  return ($$02|0);
 }
 else if ((label|0) == 43) {
  HEAP32[$2>>2] = $$lcssa56;
  HEAP32[$5>>2] = $$lcssa60;
  $$02 = 0;
  return ($$02|0);
 }
 else if ((label|0) == 44) {
  return ($$02|0);
 }
 return (0)|0;
}
function _Split_Conic($base) {
 $base = $base|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($base)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($base)) + 32|0);
 HEAP32[$2>>2] = $1;
 $3 = ((($base)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + ($1))|0;
 $6 = (($5|0) / 2)&-1;
 $7 = ((($base)) + 24|0);
 HEAP32[$7>>2] = $6;
 $8 = HEAP32[$base>>2]|0;
 $9 = (($8) + ($4))|0;
 $10 = (($9|0) / 2)&-1;
 HEAP32[$3>>2] = $10;
 $11 = (($10) + ($6))|0;
 $12 = (($11|0) / 2)&-1;
 HEAP32[$0>>2] = $12;
 $13 = ((($base)) + 20|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($base)) + 36|0);
 HEAP32[$15>>2] = $14;
 $16 = ((($base)) + 12|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = (($17) + ($14))|0;
 $19 = (($18|0) / 2)&-1;
 $20 = ((($base)) + 28|0);
 HEAP32[$20>>2] = $19;
 $21 = ((($base)) + 4|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = (($22) + ($17))|0;
 $24 = (($23|0) / 2)&-1;
 HEAP32[$16>>2] = $24;
 $25 = (($24) + ($19))|0;
 $26 = (($25|0) / 2)&-1;
 HEAP32[$13>>2] = $26;
 return;
}
function _Bezier_Up($worker,$degree,$splitter,$miny,$maxy) {
 $worker = $worker|0;
 $degree = $degree|0;
 $splitter = $splitter|0;
 $miny = $miny|0;
 $maxy = $maxy|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa69 = 0, $$lcssa71 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $9 = 0, $arc$0$ph$be = 0, $arc$0$ph14 = 0, $arc$011 = 0, $arc$011$lcssa = 0, $arc$011$lcssa64 = 0, $arc$01166 = 0, $e$0 = 0, $e$1$ph$be = 0, $e$1$ph13 = 0, $e0$0 = 0, $maxy$ = 0, $or$cond = 0, $or$cond110 = 0, $sext$mask = 0;
 var $top$0 = 0, $top$1 = 0, $top$2$ph$be = 0, $top$2$ph15 = 0, $top$4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($worker)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (((($1) + ($degree<<3)|0)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($worker)) + 40|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($5|0)<($miny|0);
 $9 = ($3|0)>($maxy|0);
 $or$cond = $9 | $8;
 L1: do {
  if ($or$cond) {
   $top$4 = $7;
  } else {
   $10 = ((($worker)) + 4|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = (0 - ($11))|0;
   $13 = $5 & $12;
   $14 = ($13|0)>($maxy|0);
   $maxy$ = $14 ? $maxy : $13;
   $15 = ($3|0)<($miny|0);
   if ($15) {
    $e$0 = $miny;$e0$0 = $miny;$top$1 = $7;
   } else {
    $16 = (($3) + -1)|0;
    $17 = (($16) + ($11))|0;
    $18 = $17 & $12;
    $19 = (($11) + 65535)|0;
    $20 = $3 & 65535;
    $sext$mask = $20 & $19;
    $21 = ($sext$mask|0)==(0);
    if ($21) {
     $22 = ((($worker)) + 87|0);
     $23 = HEAP8[$22>>0]|0;
     $24 = ($23<<24>>24)==(0);
     if ($24) {
      $top$0 = $7;
     } else {
      $25 = ((($7)) + -4|0);
      HEAP8[$22>>0] = 0;
      $top$0 = $25;
     }
     $26 = (($1) + ($degree<<3)|0);
     $27 = HEAP32[$26>>2]|0;
     $28 = ((($top$0)) + 4|0);
     HEAP32[$top$0>>2] = $27;
     $29 = (($11) + ($18))|0;
     $e$0 = $29;$e0$0 = $18;$top$1 = $28;
    } else {
     $e$0 = $18;$e0$0 = $18;$top$1 = $7;
    }
   }
   $30 = ((($worker)) + 86|0);
   $31 = HEAP8[$30>>0]|0;
   $32 = ($31<<24>>24)==(0);
   if (!($32)) {
    $33 = HEAP32[$worker>>2]|0;
    $34 = $e0$0 >> $33;
    $35 = ((($worker)) + 88|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ((($36)) + 20|0);
    HEAP32[$37>>2] = $34;
    HEAP8[$30>>0] = 0;
   }
   $38 = ($maxy$|0)<($e$0|0);
   if ($38) {
    $top$4 = $top$1;
   } else {
    $39 = (($maxy$) - ($e$0))|0;
    $40 = HEAP32[$worker>>2]|0;
    $41 = $39 >> $40;
    $42 = (($top$1) + ($41<<2)|0);
    $43 = ((($42)) + 4|0);
    $44 = ((($worker)) + 36|0);
    $45 = HEAP32[$44>>2]|0;
    $46 = ($43>>>0)<($45>>>0);
    if (!($46)) {
     HEAP32[$6>>2] = $top$1;
     $50 = ((($worker)) + 44|0);
     HEAP32[$50>>2] = 98;
     $$0 = 1;
     return ($$0|0);
    }
    $47 = ((($worker)) + 87|0);
    $48 = ((($worker)) + 16|0);
    $49 = (0 - ($degree))|0;
    $arc$0$ph14 = $1;$e$1$ph13 = $e$0;$top$2$ph15 = $top$1;
    while(1) {
     $arc$011 = $arc$0$ph14;
     while(1) {
      HEAP8[$47>>0] = 0;
      $51 = ((($arc$011)) + 4|0);
      $52 = HEAP32[$51>>2]|0;
      $53 = ($52|0)>($e$1$ph13|0);
      if (!($53)) {
       $$lcssa = $52;$arc$011$lcssa = $arc$011;
       label = 19;
       break;
      }
      $54 = (($arc$011) + ($degree<<3)|0);
      $55 = (((($arc$011) + ($degree<<3)|0)) + 4|0);
      $56 = HEAP32[$55>>2]|0;
      $57 = (($52) - ($56))|0;
      $58 = HEAP32[$48>>2]|0;
      $59 = ($57|0)<($58|0);
      if ($59) {
       $$lcssa69 = $56;$$lcssa71 = $57;$arc$011$lcssa64 = $arc$011;
       label = 16;
       break;
      }
      FUNCTION_TABLE_vi[$splitter & 8191]($arc$011);
      $60 = ($54>>>0)<($1>>>0);
      if ($60) {
       $top$4 = $top$2$ph15;
       break L1;
      } else {
       $arc$011 = $54;
      }
     }
     if ((label|0) == 16) {
      label = 0;
      $61 = (($arc$011$lcssa64) + ($degree<<3)|0);
      $62 = HEAP32[$61>>2]|0;
      $63 = HEAP32[$arc$011$lcssa64>>2]|0;
      $64 = (($63) - ($62))|0;
      $65 = (($e$1$ph13) - ($$lcssa69))|0;
      $66 = Math_imul($64, $65)|0;
      $67 = (($66|0) / ($$lcssa71|0))&-1;
      $68 = (($67) + ($62))|0;
      $69 = ((($top$2$ph15)) + 4|0);
      HEAP32[$top$2$ph15>>2] = $68;
      $70 = HEAP32[$10>>2]|0;
      $71 = (($70) + ($e$1$ph13))|0;
      $arc$01166 = $arc$011$lcssa64;$e$1$ph$be = $71;$top$2$ph$be = $69;
     }
     else if ((label|0) == 19) {
      label = 0;
      $74 = ($$lcssa|0)==($e$1$ph13|0);
      if ($74) {
       HEAP8[$47>>0] = 1;
       $75 = HEAP32[$arc$011$lcssa>>2]|0;
       $76 = ((($top$2$ph15)) + 4|0);
       HEAP32[$top$2$ph15>>2] = $75;
       $77 = HEAP32[$10>>2]|0;
       $78 = (($77) + ($e$1$ph13))|0;
       $arc$01166 = $arc$011$lcssa;$e$1$ph$be = $78;$top$2$ph$be = $76;
      } else {
       $arc$01166 = $arc$011$lcssa;$e$1$ph$be = $e$1$ph13;$top$2$ph$be = $top$2$ph15;
      }
     }
     $arc$0$ph$be = (($arc$01166) + ($49<<3)|0);
     $72 = ($arc$0$ph$be>>>0)<($1>>>0);
     $73 = ($e$1$ph$be|0)>($maxy$|0);
     $or$cond110 = $73 | $72;
     if ($or$cond110) {
      $top$4 = $top$2$ph$be;
      break;
     } else {
      $arc$0$ph14 = $arc$0$ph$be;$e$1$ph13 = $e$1$ph$be;$top$2$ph15 = $top$2$ph$be;
     }
    }
   }
  }
 } while(0);
 HEAP32[$6>>2] = $top$4;
 $79 = HEAP32[$0>>2]|0;
 $80 = (0 - ($degree))|0;
 $81 = (($79) + ($80<<3)|0);
 HEAP32[$0>>2] = $81;
 $$0 = 0;
 return ($$0|0);
}
function _Cubic_To($worker,$cx1,$cy1,$cx2,$cy2,$x,$y) {
 $worker = $worker|0;
 $cx1 = $cx1|0;
 $cy1 = $cy1|0;
 $cx2 = $cx2|0;
 $cy2 = $cy2|0;
 $x = $x|0;
 $y = $y|0;
 var $$ = 0, $$02 = 0, $$3 = 0, $$cast$i = 0, $$cast1$i = 0, $$in$i = 0, $$lcssa58 = 0, $$lcssa62 = 0, $$not = 0, $$phi$trans$insert = 0, $$phi$trans$insert23 = 0, $$phi$trans$insert25 = 0, $$phi$trans$insert27 = 0, $$pre = 0, $$pre$i = 0, $$pre24 = 0, $$pre26 = 0, $$pre28 = 0, $$pre30 = 0, $$sink = 0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, $ymax2$0 = 0, $ymin2$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($worker)) + 180|0);
 $1 = ((($worker)) + 52|0);
 HEAP32[$1>>2] = $0;
 $2 = ((($worker)) + 68|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($worker)) + 204|0);
 HEAP32[$4>>2] = $3;
 $5 = ((($worker)) + 72|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($worker)) + 208|0);
 HEAP32[$7>>2] = $6;
 $8 = ((($worker)) + 196|0);
 HEAP32[$8>>2] = $cx1;
 $9 = ((($worker)) + 200|0);
 HEAP32[$9>>2] = $cy1;
 $10 = ((($worker)) + 188|0);
 HEAP32[$10>>2] = $cx2;
 $11 = ((($worker)) + 192|0);
 HEAP32[$11>>2] = $cy2;
 HEAP32[$0>>2] = $x;
 $12 = ((($worker)) + 184|0);
 HEAP32[$12>>2] = $y;
 $13 = ((($worker)) + 100|0);
 $14 = ((($worker)) + 76|0);
 $15 = ((($worker)) + 80|0);
 $16 = ((($worker)) + 86|0);
 $17 = ((($worker)) + 88|0);
 $18 = ((($worker)) + 4|0);
 $19 = ((($worker)) + 8|0);
 $20 = ((($worker)) + 92|0);
 $21 = ((($worker)) + 40|0);
 $22 = ((($worker)) + 36|0);
 $23 = ((($worker)) + 176|0);
 $24 = ((($worker)) + 96|0);
 $25 = ((($worker)) + 87|0);
 $26 = ((($worker)) + 84|0);
 $28 = $0;$33 = $6;$34 = $y;$36 = $cy1;$37 = $cy2;$49 = $x;
 L1: while(1) {
  $27 = ((($28)) + 28|0);
  $29 = ((($28)) + 20|0);
  $30 = ((($28)) + 12|0);
  $31 = ((($28)) + 4|0);
  $32 = ($33|0)<=($34|0);
  $$ = $32 ? $34 : $33;
  $$3 = $32 ? $33 : $34;
  $35 = ($36|0)>($37|0);
  $ymin2$0 = $35 ? $37 : $36;
  $ymax2$0 = $35 ? $36 : $37;
  $38 = ($ymin2$0|0)<($$3|0);
  $39 = ($ymax2$0|0)>($$|0);
  $or$cond = $38 | $39;
  do {
   if ($or$cond) {
    $40 = ((($28)) + 24|0);
    $41 = HEAP32[$40>>2]|0;
    $42 = ((($28)) + 48|0);
    HEAP32[$42>>2] = $41;
    $43 = ((($28)) + 8|0);
    $44 = HEAP32[$43>>2]|0;
    $45 = ((($28)) + 16|0);
    $46 = HEAP32[$45>>2]|0;
    $47 = (($44) + 1)|0;
    $48 = (($49) + ($47))|0;
    $50 = $48 >> 1;
    HEAP32[$43>>2] = $50;
    $51 = (($41) + 1)|0;
    $52 = (($51) + ($46))|0;
    $53 = $52 >> 1;
    $54 = ((($28)) + 40|0);
    HEAP32[$54>>2] = $53;
    $55 = (($47) + ($46))|0;
    $56 = $55 >> 1;
    $57 = (($56) + 1)|0;
    $58 = (($57) + ($50))|0;
    $59 = $58 >> 1;
    HEAP32[$45>>2] = $59;
    $60 = (($57) + ($53))|0;
    $61 = $60 >> 1;
    $62 = ((($28)) + 32|0);
    HEAP32[$62>>2] = $61;
    $63 = (($59) + 1)|0;
    $64 = (($63) + ($61))|0;
    $65 = $64 >> 1;
    HEAP32[$40>>2] = $65;
    $66 = ((($28)) + 52|0);
    HEAP32[$66>>2] = $33;
    $67 = (($37) + 1)|0;
    $68 = (($34) + ($67))|0;
    $69 = $68 >> 1;
    HEAP32[$30>>2] = $69;
    $70 = (($33) + 1)|0;
    $71 = (($70) + ($36))|0;
    $72 = $71 >> 1;
    $73 = ((($28)) + 44|0);
    HEAP32[$73>>2] = $72;
    $74 = (($67) + ($36))|0;
    $75 = $74 >> 1;
    $76 = (($75) + 1)|0;
    $77 = (($76) + ($69))|0;
    $78 = $77 >> 1;
    HEAP32[$29>>2] = $78;
    $79 = (($76) + ($72))|0;
    $80 = $79 >> 1;
    $81 = ((($28)) + 36|0);
    HEAP32[$81>>2] = $80;
    $82 = (($78) + 1)|0;
    $83 = (($82) + ($80))|0;
    $84 = $83 >> 1;
    HEAP32[$27>>2] = $84;
    $85 = ((($28)) + 24|0);
    HEAP32[$1>>2] = $85;
   } else {
    $86 = ($33|0)==($34|0);
    if ($86) {
     $87 = ((($28)) + -24|0);
     HEAP32[$1>>2] = $87;
     break;
    }
    $88 = $32 ? 1 : 2;
    $89 = HEAP32[$13>>2]|0;
    $90 = ($89|0)==($88|0);
    if (!($90)) {
     $91 = HEAP32[$18>>2]|0;
     if ($32) {
      $92 = (($33) + -1)|0;
      $93 = (($92) + ($91))|0;
      $94 = (0 - ($91))|0;
      $95 = $93 & $94;
      $96 = (($95) - ($33))|0;
      $97 = HEAP32[$19>>2]|0;
      $98 = ($96|0)>=($97|0);
      $$sink = $98;
     } else {
      $$not = (($91) + -1)|0;
      $99 = $$not & $33;
      $100 = HEAP32[$19>>2]|0;
      $101 = ($99|0)>=($100|0);
      $$sink = $101;
     }
     $102 = ($89|0)==(0);
     if (!($102)) {
      $103 = HEAP32[$21>>2]|0;
      $104 = HEAP32[$17>>2]|0;
      $105 = ((($104)) + 8|0);
      $106 = HEAP32[$105>>2]|0;
      $107 = (($103) - ($106))|0;
      $108 = $107 >> 2;
      $109 = ($107|0)<(0);
      $110 = $103;
      if ($109) {
       label = 12;
       break L1;
      }
      $112 = ($107|0)>(0);
      if ($112) {
       $113 = ((($104)) + 16|0);
       HEAP32[$113>>2] = $108;
       do {
        if ($$sink) {
         $114 = ((($104)) + 12|0);
         $115 = HEAP16[$114>>1]|0;
         $116 = $115&65535;
         $117 = $116 & 8;
         $118 = ($117|0)==(0);
         if ($118) {
          $121 = $116 | 32;
          $122 = $121&65535;
          HEAP16[$114>>1] = $122;
          break;
         } else {
          $119 = $116 | 16;
          $120 = $119&65535;
          HEAP16[$114>>1] = $120;
          break;
         }
        }
       } while(0);
       HEAP32[$17>>2] = $103;
       $123 = ((($110)) + 32|0);
       HEAP32[$21>>2] = $123;
       $$cast1$i = $103;
       $124 = ((($$cast1$i)) + 16|0);
       HEAP32[$124>>2] = 0;
       $125 = $123;
       $126 = ((($$cast1$i)) + 8|0);
       HEAP32[$126>>2] = $125;
       $127 = ((($104)) + 28|0);
       HEAP32[$127>>2] = $103;
       $128 = HEAP16[$26>>1]|0;
       $129 = (($128) + 1)<<16>>16;
       HEAP16[$26>>1] = $129;
       $132 = $123;
      } else {
       $132 = $110;
      }
      $130 = HEAP32[$22>>2]|0;
      $131 = ($132>>>0)<($130>>>0);
      if (!($131)) {
       label = 20;
       break L1;
      }
      HEAP8[$25>>0] = 0;
     }
     $134 = HEAP32[$20>>2]|0;
     $135 = ($134|0)==(0|0);
     if ($135) {
      $136 = HEAP32[$21>>2]|0;
      HEAP32[$17>>2] = $136;
      HEAP32[$20>>2] = $136;
      $$cast$i = $136;
      $137 = ((($$cast$i)) + 32|0);
      HEAP32[$21>>2] = $137;
      $$in$i = $137;
     } else {
      $$pre$i = HEAP32[$21>>2]|0;
      $$in$i = $$pre$i;
     }
     $138 = HEAP32[$22>>2]|0;
     $139 = ($$in$i>>>0)<($138>>>0);
     if (!($139)) {
      label = 26;
      break L1;
     }
     $141 = $$in$i;
     $142 = HEAP32[$17>>2]|0;
     $143 = ((($142)) + 12|0);
     $144 = ((($142)) + 20|0);
     HEAP32[$144>>2] = 0;
     $145 = ((($142)) + 16|0);
     HEAP32[$145>>2] = 0;
     $146 = ((($142)) + 8|0);
     HEAP32[$146>>2] = $141;
     $147 = ((($142)) + 4|0);
     HEAP32[$147>>2] = 0;
     $148 = ((($142)) + 28|0);
     HEAP32[$148>>2] = 0;
     $149 = HEAP8[$23>>0]|0;
     $150 = $149&255;
     HEAP16[$143>>1] = $150;
     $151 = $142;
     if ($32) {
      $152 = $149&255;
      $153 = $152 | 8;
      $154 = $153&65535;
      HEAP16[$143>>1] = $154;
      if ($$sink) {
       $155 = $152 | 40;
       $156 = $155&65535;
       HEAP16[$143>>1] = $156;
      }
     } else {
      if ($$sink) {
       $157 = $149&255;
       $158 = $157 | 16;
       $159 = $158&65535;
       HEAP16[$143>>1] = $159;
      }
     }
     $160 = HEAP32[$24>>2]|0;
     $161 = ($160|0)==(0|0);
     if ($161) {
      HEAP32[$24>>2] = $151;
     }
     HEAP32[$13>>2] = $88;
     HEAP8[$16>>0] = 1;
     HEAP8[$25>>0] = 0;
    }
    $162 = HEAP32[$14>>2]|0;
    $163 = HEAP32[$15>>2]|0;
    if ($32) {
     $164 = (_Bezier_Up($worker,3,4391,$162,$163)|0);
     $165 = ($164<<24>>24)==(0);
     if ($165) {
      break;
     } else {
      $$02 = 1;
      label = 44;
      break L1;
     }
    }
    $166 = (0 - ($34))|0;
    HEAP32[$31>>2] = $166;
    $167 = (0 - ($37))|0;
    HEAP32[$30>>2] = $167;
    $168 = (0 - ($36))|0;
    HEAP32[$29>>2] = $168;
    $169 = (0 - ($33))|0;
    HEAP32[$27>>2] = $169;
    $170 = HEAP8[$16>>0]|0;
    $171 = (0 - ($163))|0;
    $172 = (0 - ($162))|0;
    $173 = (_Bezier_Up($worker,3,4391,$171,$172)|0);
    $174 = ($170<<24>>24)==(0);
    if (!($174)) {
     $175 = HEAP8[$16>>0]|0;
     $176 = ($175<<24>>24)==(0);
     if ($176) {
      $177 = HEAP32[$17>>2]|0;
      $178 = ((($177)) + 20|0);
      $179 = HEAP32[$178>>2]|0;
      $180 = (0 - ($179))|0;
      HEAP32[$178>>2] = $180;
     }
    }
    $181 = HEAP32[$31>>2]|0;
    $182 = (0 - ($181))|0;
    HEAP32[$31>>2] = $182;
    $183 = ($173<<24>>24)==(0);
    if (!($183)) {
     $$02 = 1;
     label = 44;
     break L1;
    }
   }
  } while(0);
  $184 = HEAP32[$1>>2]|0;
  $185 = ($184>>>0)<($0>>>0);
  if ($185) {
   $$lcssa58 = $49;$$lcssa62 = $34;
   label = 43;
   break;
  }
  $$phi$trans$insert = ((($184)) + 28|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $$phi$trans$insert23 = ((($184)) + 20|0);
  $$pre24 = HEAP32[$$phi$trans$insert23>>2]|0;
  $$phi$trans$insert25 = ((($184)) + 12|0);
  $$pre26 = HEAP32[$$phi$trans$insert25>>2]|0;
  $$phi$trans$insert27 = ((($184)) + 4|0);
  $$pre28 = HEAP32[$$phi$trans$insert27>>2]|0;
  $$pre30 = HEAP32[$184>>2]|0;
  $28 = $184;$33 = $$pre;$34 = $$pre28;$36 = $$pre24;$37 = $$pre26;$49 = $$pre30;
 }
 if ((label|0) == 12) {
  $111 = ((($worker)) + 44|0);
  HEAP32[$111>>2] = 99;
  $$02 = 1;
  return ($$02|0);
 }
 else if ((label|0) == 20) {
  $133 = ((($worker)) + 44|0);
  HEAP32[$133>>2] = 98;
  $$02 = 1;
  return ($$02|0);
 }
 else if ((label|0) == 26) {
  $140 = ((($worker)) + 44|0);
  HEAP32[$140>>2] = 98;
  $$02 = 1;
  return ($$02|0);
 }
 else if ((label|0) == 43) {
  HEAP32[$2>>2] = $$lcssa58;
  HEAP32[$5>>2] = $$lcssa62;
  $$02 = 0;
  return ($$02|0);
 }
 else if ((label|0) == 44) {
  return ($$02|0);
 }
 return (0)|0;
}
function _Split_Cubic($base) {
 $base = $base|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($base)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($base)) + 48|0);
 HEAP32[$2>>2] = $1;
 $3 = ((($base)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($base)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = HEAP32[$base>>2]|0;
 $8 = (($4) + 1)|0;
 $9 = (($8) + ($7))|0;
 $10 = $9 >> 1;
 HEAP32[$3>>2] = $10;
 $11 = (($6) + 1)|0;
 $12 = (($11) + ($1))|0;
 $13 = $12 >> 1;
 $14 = ((($base)) + 40|0);
 HEAP32[$14>>2] = $13;
 $15 = (($8) + ($6))|0;
 $16 = $15 >> 1;
 $17 = (($16) + 1)|0;
 $18 = (($17) + ($10))|0;
 $19 = $18 >> 1;
 HEAP32[$5>>2] = $19;
 $20 = (($17) + ($13))|0;
 $21 = $20 >> 1;
 $22 = ((($base)) + 32|0);
 HEAP32[$22>>2] = $21;
 $23 = (($19) + 1)|0;
 $24 = (($23) + ($21))|0;
 $25 = $24 >> 1;
 HEAP32[$0>>2] = $25;
 $26 = ((($base)) + 28|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ((($base)) + 52|0);
 HEAP32[$28>>2] = $27;
 $29 = ((($base)) + 12|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ((($base)) + 20|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ((($base)) + 4|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = (($30) + 1)|0;
 $36 = (($35) + ($34))|0;
 $37 = $36 >> 1;
 HEAP32[$29>>2] = $37;
 $38 = (($32) + 1)|0;
 $39 = (($38) + ($27))|0;
 $40 = $39 >> 1;
 $41 = ((($base)) + 44|0);
 HEAP32[$41>>2] = $40;
 $42 = (($35) + ($32))|0;
 $43 = $42 >> 1;
 $44 = (($43) + 1)|0;
 $45 = (($44) + ($37))|0;
 $46 = $45 >> 1;
 HEAP32[$31>>2] = $46;
 $47 = (($44) + ($40))|0;
 $48 = $47 >> 1;
 $49 = ((($base)) + 36|0);
 HEAP32[$49>>2] = $48;
 $50 = (($46) + 1)|0;
 $51 = (($50) + ($48))|0;
 $52 = $51 >> 1;
 HEAP32[$26>>2] = $52;
 return;
}
function _Horizontal_Sweep_Init($worker,$min,$max) {
 $worker = $worker|0;
 $min = $min|0;
 $max = $max|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _Horizontal_Sweep_Span($worker,$y,$x1,$x2,$left,$right) {
 $worker = $worker|0;
 $y = $y|0;
 $x1 = $x1|0;
 $x2 = $x2|0;
 $left = $left|0;
 $right = $right|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($x2) - ($x1))|0;
 $1 = ((($worker)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($0|0)<($2|0);
 if (!($3)) {
  return;
 }
 $4 = (($x1) + -1)|0;
 $5 = (($4) + ($2))|0;
 $6 = (0 - ($2))|0;
 $7 = $5 & $6;
 $8 = $6 & $x2;
 $9 = ($7|0)==($8|0);
 if (!($9)) {
  return;
 }
 $10 = HEAP32[$worker>>2]|0;
 $11 = $7 >> $10;
 $12 = ($11|0)>(-1);
 if (!($12)) {
  return;
 }
 $13 = ((($worker)) + 104|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ($11>>>0)<($14>>>0);
 if (!($15)) {
  return;
 }
 $16 = ((($worker)) + 60|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $y << 16 >> 16;
 $19 = $18 >> 3;
 $20 = (($17) + ($19)|0);
 $21 = $18 & 7;
 $22 = 128 >>> $21;
 $23 = ((($worker)) + 112|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = (0 - ($11))|0;
 $26 = Math_imul($24, $25)|0;
 $27 = (($20) + ($26)|0);
 $28 = ($24|0)>(0);
 $29 = (($14) + -1)|0;
 $30 = Math_imul($24, $29)|0;
 $31 = (($27) + ($30)|0);
 $p$0 = $28 ? $31 : $27;
 $32 = HEAP8[$p$0>>0]|0;
 $33 = $32&255;
 $34 = $33 | $22;
 $35 = $34&255;
 HEAP8[$p$0>>0] = $35;
 return;
}
function _Horizontal_Sweep_Drop($worker,$y,$x1,$x2,$left,$right) {
 $worker = $worker|0;
 $y = $y|0;
 $x1 = $x1|0;
 $x2 = $x2|0;
 $left = $left|0;
 $right = $right|0;
 var $$pre = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $12 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bits$0 = 0, $bits$1 = 0, $pxl$0 = 0, $pxl$0$ = 0, $pxl$1 = 0, $pxl$3 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($worker)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($x1) + -1)|0;
 $3 = (($2) + ($1))|0;
 $4 = (0 - ($1))|0;
 $5 = $3 & $4;
 $6 = $4 & $x2;
 $7 = ($5|0)>($6|0);
 if ($7) {
  $8 = ((($left)) + 12|0);
  $9 = HEAP16[$8>>1]|0;
  $10 = $9&65535;
  $11 = $10 & 7;
  $12 = (($6) + ($1))|0;
  $13 = ($5|0)==($12|0);
  if (!($13)) {
   return;
  }
  switch ($11|0) {
  case 0:  {
   $pxl$0 = $6;
   break;
  }
  case 4:  {
   $14 = (($2) + ($x2))|0;
   $15 = (($14|0) / 2)&-1;
   $16 = ((($worker)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = (($17) + ($15))|0;
   $19 = $18 & $4;
   $pxl$0 = $19;
   break;
  }
  case 5: case 1:  {
   $20 = ((($left)) + 28|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = ($21|0)==($right|0);
   if ($22) {
    $23 = ((($left)) + 16|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = ($24|0)<(1);
    if ($25) {
     $26 = $9 & 16;
     $27 = ($26<<16>>16)==(0);
     if ($27) {
      return;
     }
     $28 = (($x2) - ($x1))|0;
     $29 = ((($worker)) + 8|0);
     $30 = HEAP32[$29>>2]|0;
     $31 = ($28|0)<($30|0);
     if ($31) {
      return;
     }
    }
   }
   $32 = ((($right)) + 28|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ($33|0)==($left|0);
   if ($34) {
    $35 = ((($left)) + 20|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = $y << 16 >> 16;
    $38 = ($36|0)==($37|0);
    if ($38) {
     $39 = $9 & 32;
     $40 = ($39<<16>>16)==(0);
     if ($40) {
      return;
     }
     $41 = (($x2) - ($x1))|0;
     $42 = ((($worker)) + 8|0);
     $43 = HEAP32[$42>>2]|0;
     $44 = ($41|0)<($43|0);
     if ($44) {
      return;
     }
    }
   }
   $45 = ($11|0)==(1);
   if ($45) {
    $pxl$0 = $6;
   } else {
    $46 = (($2) + ($x2))|0;
    $47 = (($46|0) / 2)&-1;
    $48 = ((($worker)) + 8|0);
    $49 = HEAP32[$48>>2]|0;
    $50 = (($49) + ($47))|0;
    $51 = $50 & $4;
    $pxl$0 = $51;
   }
   break;
  }
  default: {
   return;
  }
  }
  $52 = ($pxl$0|0)<(0);
  $$pre = HEAP32[$worker>>2]|0;
  if ($52) {
   $pxl$1 = $5;
  } else {
   $53 = $pxl$0 >> $$pre;
   $54 = ((($worker)) + 104|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = ($53>>>0)<($55>>>0);
   $pxl$0$ = $56 ? $pxl$0 : $6;
   $pxl$1 = $pxl$0$;
  }
  $57 = ($pxl$1|0)==($5|0);
  $58 = $57 ? $6 : $5;
  $59 = $58 >> $$pre;
  $60 = ((($worker)) + 60|0);
  $61 = HEAP32[$60>>2]|0;
  $62 = $y << 16 >> 16;
  $63 = $62 >> 3;
  $64 = (($61) + ($63)|0);
  $65 = $62 & 7;
  $66 = 128 >>> $65;
  $67 = ((($worker)) + 104|0);
  $68 = ((($worker)) + 112|0);
  $69 = HEAP32[$68>>2]|0;
  $70 = (0 - ($69))|0;
  $71 = Math_imul($59, $70)|0;
  $72 = (($64) + ($71)|0);
  $73 = ($69|0)>(0);
  if ($73) {
   $74 = HEAP32[$67>>2]|0;
   $75 = (($74) + -1)|0;
   $76 = Math_imul($75, $69)|0;
   $77 = (($72) + ($76)|0);
   $bits$0 = $77;
  } else {
   $bits$0 = $72;
  }
  $78 = ($59|0)>(-1);
  if ($78) {
   $79 = HEAP32[$67>>2]|0;
   $80 = ($59>>>0)<($79>>>0);
   if ($80) {
    $81 = HEAP8[$bits$0>>0]|0;
    $82 = $81&255;
    $83 = $82 & $66;
    $84 = ($83|0)==(0);
    if ($84) {
     $86 = $$pre;$pxl$3 = $pxl$1;
    } else {
     return;
    }
   } else {
    $86 = $$pre;$pxl$3 = $pxl$1;
   }
  } else {
   $86 = $$pre;$pxl$3 = $pxl$1;
  }
 } else {
  $$pre3 = HEAP32[$worker>>2]|0;
  $86 = $$pre3;$pxl$3 = $5;
 }
 $85 = $pxl$3 >> $86;
 $87 = ($85|0)>(-1);
 if (!($87)) {
  return;
 }
 $88 = ((($worker)) + 104|0);
 $89 = HEAP32[$88>>2]|0;
 $90 = ($85>>>0)<($89>>>0);
 if (!($90)) {
  return;
 }
 $91 = ((($worker)) + 60|0);
 $92 = HEAP32[$91>>2]|0;
 $93 = $y << 16 >> 16;
 $94 = $93 >> 3;
 $95 = (($92) + ($94)|0);
 $96 = $93 & 7;
 $97 = 128 >>> $96;
 $98 = ((($worker)) + 112|0);
 $99 = HEAP32[$98>>2]|0;
 $100 = (0 - ($85))|0;
 $101 = Math_imul($99, $100)|0;
 $102 = (($95) + ($101)|0);
 $103 = ($99|0)>(0);
 $104 = (($89) + -1)|0;
 $105 = Math_imul($99, $104)|0;
 $106 = (($102) + ($105)|0);
 $bits$1 = $103 ? $106 : $102;
 $107 = HEAP8[$bits$1>>0]|0;
 $108 = $107&255;
 $109 = $108 | $97;
 $110 = $109&255;
 HEAP8[$bits$1>>0] = $110;
 return;
}
function _Horizontal_Sweep_Step($worker) {
 $worker = $worker|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _ft_black_done($raster) {
 $raster = $raster|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$raster>>2]|0;
 _ft_mem_free($0,$raster);
 return;
}
function _ft_raster1_init($render) {
 $render = $render|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($render)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($render)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 56|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($render)) + 52|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($1)) + 164|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($1)) + 168|0);
 $13 = HEAP32[$12>>2]|0;
 FUNCTION_TABLE_viii[$7 & 8191]($9,$11,$13);
 return 0;
}
function _ft_raster1_render($render,$slot,$mode,$origin) {
 $render = $render|0;
 $slot = $slot|0;
 $mode = $mode|0;
 $origin = $origin|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $9 = 0, $cbox$sroa$0$0 = 0, $cbox$sroa$19$0 = 0, $cbox$sroa$9$0 = 0, $cbox0 = 0, $error = 0, $height$0 = 0, $params = 0, $width$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp + 64|0;
 $cbox0 = sp + 48|0;
 $params = sp;
 $0 = ((($slot)) + 72|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($render)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==($3|0);
 if (!($4)) {
  HEAP32[$error>>2] = 6;
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = ($mode|0)==(2);
 if (!($5)) {
  $$0 = 19;
  STACKTOP = sp;return ($$0|0);
 }
 $6 = ((($slot)) + 108|0);
 $7 = ($origin|0)==(0|0);
 if (!($7)) {
  $8 = HEAP32[$origin>>2]|0;
  $9 = ((($origin)) + 4|0);
  $10 = HEAP32[$9>>2]|0;
  _FT_Outline_Translate($6,$8,$10);
 }
 _FT_Outline_Get_CBox($6,$cbox0);
 $11 = HEAP32[$cbox0>>2]|0;
 $12 = (($11) + 32)|0;
 $13 = $12 & -64;
 $14 = ((($cbox0)) + 4|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = (($15) + 32)|0;
 $17 = $16 & -64;
 $18 = ((($cbox0)) + 8|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($19) + 32)|0;
 $21 = $20 & -64;
 $22 = ((($cbox0)) + 12|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = (($23) + 32)|0;
 $25 = $24 & -64;
 $26 = (($21) - ($13))|0;
 $27 = $26 >> 6;
 $28 = ($27|0)==(0);
 if ($28) {
  $29 = $11 & -64;
  $30 = (($19) + 63)|0;
  $31 = $30 & -64;
  $32 = (($31) - ($29))|0;
  $33 = $32 >> 6;
  $cbox$sroa$0$0 = $29;$width$0 = $33;
 } else {
  $cbox$sroa$0$0 = $13;$width$0 = $27;
 }
 $34 = (($25) - ($17))|0;
 $35 = $34 >> 6;
 $36 = ($35|0)==(0);
 if ($36) {
  $37 = $15 & -64;
  $38 = (($23) + 63)|0;
  $39 = $38 & -64;
  $40 = (($39) - ($37))|0;
  $41 = $40 >> 6;
  $cbox$sroa$19$0 = $39;$cbox$sroa$9$0 = $37;$height$0 = $41;
 } else {
  $cbox$sroa$19$0 = $25;$cbox$sroa$9$0 = $17;$height$0 = $35;
 }
 $42 = $height$0 | $width$0;
 $43 = ($42>>>0)>(65535);
 if ($43) {
  HEAP32[$error>>2] = 6;
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $44 = ((($slot)) + 76|0);
 $45 = ((($render)) + 8|0);
 $46 = HEAP32[$45>>2]|0;
 $47 = ((($slot)) + 156|0);
 $48 = HEAP32[$47>>2]|0;
 $49 = ((($48)) + 4|0);
 $50 = HEAP32[$49>>2]|0;
 $51 = $50 & 1;
 $52 = ($51|0)==(0);
 $$pre = ((($slot)) + 88|0);
 if (!($52)) {
  $53 = HEAP32[$$pre>>2]|0;
  _ft_mem_free($46,$53);
  HEAP32[$$pre>>2] = 0;
  $54 = HEAP32[$47>>2]|0;
  $55 = ((($54)) + 4|0);
  $56 = HEAP32[$55>>2]|0;
  $57 = $56 & -2;
  HEAP32[$55>>2] = $57;
 }
 $58 = (($width$0) + 15)|0;
 $59 = $58 >>> 4;
 $60 = $59 << 1;
 $61 = ((($slot)) + 94|0);
 HEAP8[$61>>0] = 1;
 $62 = ((($slot)) + 80|0);
 HEAP32[$62>>2] = $width$0;
 HEAP32[$44>>2] = $height$0;
 $63 = ((($slot)) + 84|0);
 HEAP32[$63>>2] = $60;
 $64 = (_ft_mem_realloc($46,$height$0,0,$60,0,$error)|0);
 HEAP32[$$pre>>2] = $64;
 $65 = HEAP32[$error>>2]|0;
 $66 = ($65|0)==(0);
 if (!($66)) {
  $$0 = $65;
  STACKTOP = sp;return ($$0|0);
 }
 $67 = HEAP32[$47>>2]|0;
 $68 = ((($67)) + 4|0);
 $69 = HEAP32[$68>>2]|0;
 $70 = $69 | 1;
 HEAP32[$68>>2] = $70;
 $71 = (0 - ($cbox$sroa$0$0))|0;
 $72 = (0 - ($cbox$sroa$9$0))|0;
 _FT_Outline_Translate($6,$71,$72);
 HEAP32[$params>>2] = $44;
 $73 = ((($params)) + 4|0);
 HEAP32[$73>>2] = $6;
 $74 = ((($params)) + 8|0);
 HEAP32[$74>>2] = 0;
 $75 = ((($render)) + 56|0);
 $76 = HEAP32[$75>>2]|0;
 $77 = ((($render)) + 52|0);
 $78 = HEAP32[$77>>2]|0;
 $79 = (FUNCTION_TABLE_iii[$76 & 8191]($78,$params)|0);
 HEAP32[$error>>2] = $79;
 _FT_Outline_Translate($6,$cbox$sroa$0$0,$cbox$sroa$9$0);
 $80 = HEAP32[$error>>2]|0;
 $81 = ($80|0)==(0);
 if (!($81)) {
  $$0 = $80;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$0>>2] = 1651078259;
 $82 = $cbox$sroa$0$0 >> 6;
 $83 = ((($slot)) + 100|0);
 HEAP32[$83>>2] = $82;
 $84 = $cbox$sroa$19$0 >> 6;
 $85 = ((($slot)) + 104|0);
 HEAP32[$85>>2] = $84;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _ft_raster1_transform($render,$slot,$matrix,$delta) {
 $render = $render|0;
 $slot = $slot|0;
 $matrix = $matrix|0;
 $delta = $delta|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($slot)) + 72|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($render)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==($3|0);
 if (!($4)) {
  $error$0 = 6;
  return ($error$0|0);
 }
 $5 = ($matrix|0)==(0|0);
 if (!($5)) {
  $6 = ((($slot)) + 108|0);
  _FT_Outline_Transform($6,$matrix);
 }
 $7 = ($delta|0)==(0|0);
 if ($7) {
  $error$0 = 0;
  return ($error$0|0);
 }
 $8 = ((($slot)) + 108|0);
 $9 = HEAP32[$delta>>2]|0;
 $10 = ((($delta)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 _FT_Outline_Translate($8,$9,$11);
 $error$0 = 0;
 return ($error$0|0);
}
function _ft_raster1_get_cbox($render,$slot,$cbox) {
 $render = $render|0;
 $slot = $slot|0;
 $cbox = $cbox|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$cbox>>2]=0|0;HEAP32[$cbox+4>>2]=0|0;HEAP32[$cbox+8>>2]=0|0;HEAP32[$cbox+12>>2]=0|0;
 $0 = ((($slot)) + 72|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($render)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==($3|0);
 if (!($4)) {
  return;
 }
 $5 = ((($slot)) + 108|0);
 _FT_Outline_Get_CBox($5,$cbox);
 return;
}
function _ft_raster1_set_mode($render,$mode_tag,$data) {
 $render = $render|0;
 $mode_tag = $mode_tag|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($render)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 56|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($render)) + 52|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (FUNCTION_TABLE_iiii[$5 & 8191]($7,$mode_tag,$data)|0);
 return ($8|0);
}
function _tt_cmap_init($cmap,$table) {
 $cmap = $cmap|0;
 $table = $table|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 HEAP32[$0>>2] = $table;
 return 0;
}
function _tt_cmap0_char_index($cmap,$char_code) {
 $cmap = $cmap|0;
 $char_code = $char_code|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($char_code>>>0)<(256);
 if (!($0)) {
  $7 = 0;
  return ($7|0);
 }
 $1 = ((($cmap)) + 16|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($char_code) + 6)|0;
 $4 = (($2) + ($3)|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = $6;
 return ($7|0);
}
function _tt_cmap0_char_next($cmap,$pchar_code) {
 $cmap = $cmap|0;
 $pchar_code = $pchar_code|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $charcode$0 = 0, $gindex$0 = 0, $gindex$1 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$pchar_code>>2]|0;
 $3 = ((($1)) + 6|0);
 $charcode$0 = $2;$gindex$0 = 0;
 while(1) {
  $4 = (($charcode$0) + 1)|0;
  $5 = ($4>>>0)<(256);
  if (!($5)) {
   $gindex$1 = $gindex$0;$result$0 = 0;
   break;
  }
  $6 = (($3) + ($4)|0);
  $7 = HEAP8[$6>>0]|0;
  $8 = $7&255;
  $9 = ($7<<24>>24)==(0);
  if ($9) {
   $charcode$0 = $4;$gindex$0 = $8;
  } else {
   $gindex$1 = $8;$result$0 = $4;
   break;
  }
 }
 HEAP32[$pchar_code>>2] = $result$0;
 return ($gindex$1|0);
}
function _tt_cmap0_validate($table,$valid) {
 $table = $table|0;
 $valid = $valid|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $n$02 = 0, $or$cond = 0, $p$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($table)) + 4|0);
 $1 = ((($valid)) + 160|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($0>>>0)>($2>>>0);
 if ($3) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $4 = ((($table)) + 2|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = $6 << 8;
 $8 = ((($table)) + 3|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 $11 = $7 | $10;
 $12 = (($table) + ($11)|0);
 $13 = HEAP32[$1>>2]|0;
 $14 = ($12>>>0)>($13>>>0);
 $15 = ($11>>>0)<(262);
 $or$cond = $15 | $14;
 if ($or$cond) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $16 = ((($valid)) + 164|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ($17|0)==(0);
 if ($18) {
  return 0;
 }
 $19 = ((($table)) + 6|0);
 $20 = ((($valid)) + 172|0);
 $n$02 = 0;$p$03 = $19;
 while(1) {
  $21 = ((($p$03)) + 1|0);
  $22 = HEAP8[$p$03>>0]|0;
  $23 = $22&255;
  $24 = HEAP32[$20>>2]|0;
  $25 = ($23>>>0)<($24>>>0);
  if (!($25)) {
   label = 8;
   break;
  }
  $26 = (($n$02) + 1)|0;
  $exitcond = ($26|0)==(256);
  if ($exitcond) {
   label = 10;
   break;
  } else {
   $n$02 = $26;$p$03 = $21;
  }
 }
 if ((label|0) == 8) {
  _ft_validator_error($valid,16);
  // unreachable;
 }
 else if ((label|0) == 10) {
  return 0;
 }
 return (0)|0;
}
function _tt_cmap0_get_info($cmap,$cmap_info) {
 $cmap = $cmap|0;
 $cmap_info = $cmap_info|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 4|0);
 $3 = ((($cmap_info)) + 4|0);
 HEAP32[$3>>2] = 0;
 $4 = HEAP8[$2>>0]|0;
 $5 = $4&255;
 $6 = $5 << 8;
 $7 = ((($1)) + 5|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = $8&255;
 $10 = $6 | $9;
 HEAP32[$cmap_info>>2] = $10;
 return 0;
}
function _tt_cmap2_char_index($cmap,$char_code) {
 $cmap = $cmap|0;
 $char_code = $char_code|0;
 var $$masked$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $8 = 0, $9 = 0, $or$cond = 0, $result$1 = 0;
 var $sub$0$i$ph = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($char_code>>>0)<(65536);
 if (!($2)) {
  $result$1 = 0;
  return ($result$1|0);
 }
 $3 = $char_code >>> 8;
 $4 = ((($1)) + 6|0);
 $5 = ((($1)) + 518|0);
 $6 = ($3|0)==(0);
 if ($6) {
  $7 = $char_code << 1;
  $8 = $7 & 510;
  $9 = (($4) + ($8)|0);
  $10 = HEAP8[$9>>0]|0;
  $11 = $10&255;
  $12 = $11 << 8;
  $13 = ((($9)) + 1|0);
  $14 = HEAP8[$13>>0]|0;
  $15 = $14&255;
  $16 = $12 | $15;
  $17 = ($16|0)==(0);
  if ($17) {
   $sub$0$i$ph = $5;
  } else {
   $result$1 = 0;
   return ($result$1|0);
  }
 } else {
  $18 = $3 << 1;
  $19 = (($4) + ($18)|0);
  $20 = HEAP8[$19>>0]|0;
  $21 = $20&255;
  $22 = $21 << 8;
  $23 = ((($19)) + 1|0);
  $24 = HEAP8[$23>>0]|0;
  $25 = $24&255;
  $$masked$i = $25 & 248;
  $26 = $$masked$i | $22;
  $27 = (($5) + ($26)|0);
  $28 = ($26|0)==(0);
  if ($28) {
   $result$1 = 0;
   return ($result$1|0);
  } else {
   $sub$0$i$ph = $27;
  }
 }
 $29 = $char_code & 255;
 $30 = HEAP8[$sub$0$i$ph>>0]|0;
 $31 = $30&255;
 $32 = $31 << 8;
 $33 = ((($sub$0$i$ph)) + 1|0);
 $34 = HEAP8[$33>>0]|0;
 $35 = $34&255;
 $36 = $32 | $35;
 $37 = ((($sub$0$i$ph)) + 2|0);
 $38 = HEAP8[$37>>0]|0;
 $39 = $38&255;
 $40 = $39 << 8;
 $41 = ((($sub$0$i$ph)) + 3|0);
 $42 = HEAP8[$41>>0]|0;
 $43 = $42&255;
 $44 = $40 | $43;
 $45 = ((($sub$0$i$ph)) + 6|0);
 $46 = ((($sub$0$i$ph)) + 4|0);
 $47 = HEAP8[$46>>0]|0;
 $48 = $47&255;
 $49 = $48 << 8;
 $50 = ((($sub$0$i$ph)) + 5|0);
 $51 = HEAP8[$50>>0]|0;
 $52 = $51&255;
 $53 = $49 | $52;
 $54 = HEAP8[$45>>0]|0;
 $55 = $54&255;
 $56 = $55 << 8;
 $57 = ((($sub$0$i$ph)) + 7|0);
 $58 = HEAP8[$57>>0]|0;
 $59 = $58&255;
 $60 = $56 | $59;
 $61 = (($29) - ($36))|0;
 $62 = ($61>>>0)<($44>>>0);
 $63 = ($60|0)!=(0);
 $or$cond = $62 & $63;
 if (!($or$cond)) {
  $result$1 = 0;
  return ($result$1|0);
 }
 $64 = $61 << 1;
 $65 = (($60) + ($64))|0;
 $66 = (($45) + ($65)|0);
 $67 = HEAP8[$66>>0]|0;
 $68 = $67&255;
 $69 = $68 << 8;
 $70 = ((($66)) + 1|0);
 $71 = HEAP8[$70>>0]|0;
 $72 = $71&255;
 $73 = $69 | $72;
 $74 = ($73|0)==(0);
 if ($74) {
  $result$1 = 0;
  return ($result$1|0);
 }
 $75 = (($73) + ($53))|0;
 $76 = $75 & 65535;
 $result$1 = $76;
 return ($result$1|0);
}
function _tt_cmap2_char_next($cmap,$pcharcode) {
 $cmap = $cmap|0;
 $pcharcode = $pcharcode|0;
 var $$masked$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $char_lo$0 = 0, $charcode$010 = 0, $charcode$18 = 0, $charcode$3 = 0, $gindex$5 = 0, $p$07 = 0, $pos$0 = 0, $pos$19 = 0, $result$3 = 0;
 var $sub$0$i$ph = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$pcharcode>>2]|0;
 $3 = (($2) + 1)|0;
 $4 = ($3>>>0)<(65536);
 if (!($4)) {
  $gindex$5 = 0;$result$3 = 0;
  HEAP32[$pcharcode>>2] = $result$3;
  return ($gindex$5|0);
 }
 $5 = ((($1)) + 6|0);
 $6 = ((($1)) + 518|0);
 $charcode$010 = $3;
 L4: while(1) {
  $7 = $charcode$010 >>> 8;
  $8 = ($7|0)==(0);
  if ($8) {
   $9 = $charcode$010 << 1;
   $10 = $9 & 510;
   $11 = (($5) + ($10)|0);
   $12 = HEAP8[$11>>0]|0;
   $13 = $12&255;
   $14 = $13 << 8;
   $15 = ((($11)) + 1|0);
   $16 = HEAP8[$15>>0]|0;
   $17 = $16&255;
   $18 = $14 | $17;
   $19 = ($18|0)==(0);
   if ($19) {
    $sub$0$i$ph = $6;
    label = 6;
   } else {
    $charcode$3 = $charcode$010;
   }
  } else {
   $20 = $7 << 1;
   $21 = (($5) + ($20)|0);
   $22 = HEAP8[$21>>0]|0;
   $23 = $22&255;
   $24 = $23 << 8;
   $25 = ((($21)) + 1|0);
   $26 = HEAP8[$25>>0]|0;
   $27 = $26&255;
   $$masked$i = $27 & 248;
   $28 = $$masked$i | $24;
   $29 = (($6) + ($28)|0);
   $30 = ($28|0)==(0);
   if ($30) {
    $charcode$3 = $charcode$010;
   } else {
    $sub$0$i$ph = $29;
    label = 6;
   }
  }
  if ((label|0) == 6) {
   label = 0;
   $31 = HEAP8[$sub$0$i$ph>>0]|0;
   $32 = $31&255;
   $33 = $32 << 8;
   $34 = ((($sub$0$i$ph)) + 1|0);
   $35 = HEAP8[$34>>0]|0;
   $36 = $35&255;
   $37 = $33 | $36;
   $38 = ((($sub$0$i$ph)) + 2|0);
   $39 = HEAP8[$38>>0]|0;
   $40 = $39&255;
   $41 = $40 << 8;
   $42 = ((($sub$0$i$ph)) + 3|0);
   $43 = HEAP8[$42>>0]|0;
   $44 = $43&255;
   $45 = $41 | $44;
   $46 = ((($sub$0$i$ph)) + 6|0);
   $47 = ((($sub$0$i$ph)) + 4|0);
   $48 = HEAP8[$47>>0]|0;
   $49 = $48&255;
   $50 = $49 << 8;
   $51 = ((($sub$0$i$ph)) + 5|0);
   $52 = HEAP8[$51>>0]|0;
   $53 = $52&255;
   $54 = $50 | $53;
   $55 = HEAP8[$46>>0]|0;
   $56 = $55&255;
   $57 = $56 << 8;
   $58 = ((($sub$0$i$ph)) + 7|0);
   $59 = HEAP8[$58>>0]|0;
   $60 = $59&255;
   $61 = $57 | $60;
   $62 = $charcode$010 & 255;
   $63 = ($61|0)==(0);
   if ($63) {
    $charcode$3 = $charcode$010;
   } else {
    $64 = ($62>>>0)<($37>>>0);
    $65 = (($62) - ($37))|0;
    $char_lo$0 = $64 ? $37 : $62;
    $pos$0 = $64 ? 0 : $65;
    $66 = $charcode$010 & -256;
    $67 = (($char_lo$0) + ($66))|0;
    $68 = ($pos$0>>>0)<($45>>>0);
    if ($68) {
     $69 = $pos$0 << 1;
     $70 = (($61) + ($69))|0;
     $71 = (($46) + ($70)|0);
     $charcode$18 = $67;$p$07 = $71;$pos$19 = $pos$0;
     while(1) {
      $72 = ((($p$07)) + 2|0);
      $73 = HEAP8[$p$07>>0]|0;
      $74 = $73&255;
      $75 = $74 << 8;
      $76 = ((($p$07)) + 1|0);
      $77 = HEAP8[$76>>0]|0;
      $78 = $77&255;
      $79 = $75 | $78;
      $80 = ($79|0)==(0);
      if (!($80)) {
       $81 = (($79) + ($54))|0;
       $82 = $81 & 65535;
       $83 = ($82|0)==(0);
       if (!($83)) {
        $gindex$5 = $82;$result$3 = $charcode$18;
        label = 13;
        break L4;
       }
      }
      $84 = (($pos$19) + 1)|0;
      $85 = (($charcode$18) + 1)|0;
      $86 = ($84>>>0)<($45>>>0);
      if ($86) {
       $charcode$18 = $85;$p$07 = $72;$pos$19 = $84;
      } else {
       $charcode$3 = $85;
       break;
      }
     }
    } else {
     $charcode$3 = $67;
    }
   }
  }
  $87 = (($charcode$3) + 256)|0;
  $88 = $87 & -256;
  $89 = ($88>>>0)<(65536);
  if ($89) {
   $charcode$010 = $88;
  } else {
   $gindex$5 = 0;$result$3 = 0;
   label = 13;
   break;
  }
 }
 if ((label|0) == 13) {
  HEAP32[$pcharcode>>2] = $result$3;
  return ($gindex$5|0);
 }
 return (0)|0;
}
function _tt_cmap2_validate($table,$valid) {
 $table = $table|0;
 $valid = $valid|0;
 var $$max_subs$0 = 0, $$max_subs$0$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $max_subs$09 = 0, $n$010 = 0, $n$17 = 0, $or$cond = 0, $or$cond3 = 0, $or$cond5 = 0, $p$011 = 0, $p$18 = 0, $p$1817 = 0, $p$26 = 0, $p$5 = 0;
 var $scevgep15 = 0, $scevgep16 = 0, $scevgep18 = 0, $scevgep1819 = 0, $scevgep20 = 0, $scevgep21 = 0, $umax = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($table)) + 4|0);
 $1 = ((($valid)) + 160|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($0>>>0)>($2>>>0);
 if ($3) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $4 = ((($table)) + 2|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = $6 << 8;
 $8 = ((($table)) + 3|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 $11 = $7 | $10;
 $12 = (($table) + ($11)|0);
 $13 = HEAP32[$1>>2]|0;
 $14 = ($12>>>0)>($13>>>0);
 $15 = ($11>>>0)<(518);
 $or$cond = $15 | $14;
 if ($or$cond) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $16 = ((($table)) + 6|0);
 $17 = ((($valid)) + 164|0);
 $max_subs$09 = 0;$n$010 = 0;$p$011 = $16;
 while(1) {
  $18 = ((($p$011)) + 2|0);
  $19 = HEAP8[$p$011>>0]|0;
  $20 = $19&255;
  $21 = $20 << 8;
  $22 = ((($p$011)) + 1|0);
  $23 = HEAP8[$22>>0]|0;
  $24 = $23&255;
  $25 = $21 | $24;
  $26 = HEAP32[$17>>2]|0;
  $27 = ($26>>>0)<(2);
  $28 = $24 & 7;
  $29 = ($28|0)==(0);
  $or$cond3 = $27 | $29;
  if (!($or$cond3)) {
   label = 7;
   break;
  }
  $30 = $25 >>> 3;
  $31 = ($30>>>0)>($max_subs$09>>>0);
  $$max_subs$0 = $31 ? $30 : $max_subs$09;
  $32 = (($n$010) + 1)|0;
  $exitcond = ($32|0)==(256);
  if ($exitcond) {
   $$max_subs$0$lcssa = $$max_subs$0;
   break;
  } else {
   $max_subs$09 = $$max_subs$0;$n$010 = $32;$p$011 = $18;
  }
 }
 if ((label|0) == 7) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $scevgep21 = ((($table)) + 518|0);
 $33 = $$max_subs$0$lcssa << 3;
 $34 = (($33) + 8)|0;
 $35 = (($scevgep21) + ($34)|0);
 $36 = HEAP32[$1>>2]|0;
 $37 = ($35>>>0)>($36>>>0);
 if ($37) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $38 = ((($valid)) + 172|0);
 $n$17 = 0;$p$18 = $scevgep21;
 L16: while(1) {
  $p$1817 = $p$18;
  $39 = HEAP8[$p$18>>0]|0;
  $40 = $39&255;
  $41 = $40 << 8;
  $42 = ((($p$18)) + 1|0);
  $43 = HEAP8[$42>>0]|0;
  $44 = $43&255;
  $45 = $41 | $44;
  $46 = ((($p$18)) + 2|0);
  $47 = HEAP8[$46>>0]|0;
  $48 = $47&255;
  $49 = $48 << 8;
  $50 = ((($p$18)) + 3|0);
  $51 = HEAP8[$50>>0]|0;
  $52 = $51&255;
  $53 = $49 | $52;
  $54 = ((($p$18)) + 4|0);
  $55 = HEAP8[$54>>0]|0;
  $56 = $55&255;
  $57 = $56 << 8;
  $58 = ((($p$18)) + 5|0);
  $59 = HEAP8[$58>>0]|0;
  $60 = $59&255;
  $61 = $57 | $60;
  $62 = ((($p$18)) + 8|0);
  $63 = ((($p$18)) + 6|0);
  $64 = HEAP8[$63>>0]|0;
  $65 = $64&255;
  $66 = $65 << 8;
  $67 = ((($p$18)) + 7|0);
  $68 = HEAP8[$67>>0]|0;
  $69 = $68&255;
  $70 = $66 | $69;
  $71 = ($53|0)==(0);
  if ($71) {
   $p$5 = $62;
  } else {
   $72 = HEAP32[$17>>2]|0;
   $73 = ($72>>>0)>(1);
   if ($73) {
    $74 = ($45>>>0)>(255);
    $75 = (($53) + ($45))|0;
    $76 = ($75>>>0)>(256);
    $or$cond5 = $74 | $76;
    if ($or$cond5) {
     label = 15;
     break;
    }
   }
   $77 = ($70|0)==(0);
   if ($77) {
    $p$5 = $62;
   } else {
    $78 = (($63) + ($70)|0);
    $79 = ($78>>>0)<($35>>>0);
    if ($79) {
     label = 19;
     break;
    }
    $80 = $53 << 1;
    $81 = (($78) + ($80)|0);
    $82 = ($81>>>0)>($12>>>0);
    if ($82) {
     label = 19;
     break;
    }
    $83 = HEAP32[$17>>2]|0;
    $84 = ($83|0)==(0);
    if ($84) {
     $p$5 = $62;
    } else {
     $85 = $53 << 1;
     $86 = (($62) + ($85)|0);
     $scevgep15 = ((($p$18)) + 10|0);
     $87 = ($86>>>0)>($scevgep15>>>0);
     $umax = $87 ? $86 : $scevgep15;
     $scevgep16 = ((($umax)) + -9|0);
     $88 = (0 - ($p$1817))|0;
     $scevgep18 = (($scevgep16) + ($88)|0);
     $scevgep1819 = $scevgep18;
     $89 = $scevgep1819 & -2;
     $p$26 = $62;
     while(1) {
      $90 = ((($p$26)) + 2|0);
      $91 = HEAP8[$p$26>>0]|0;
      $92 = $91&255;
      $93 = $92 << 8;
      $94 = ((($p$26)) + 1|0);
      $95 = HEAP8[$94>>0]|0;
      $96 = $95&255;
      $97 = $93 | $96;
      $98 = ($97|0)==(0);
      if (!($98)) {
       $99 = (($97) + ($61))|0;
       $100 = $99 & 65535;
       $101 = HEAP32[$38>>2]|0;
       $102 = ($100>>>0)<($101>>>0);
       if (!($102)) {
        label = 24;
        break L16;
       }
      }
      $103 = ($90>>>0)<($86>>>0);
      if ($103) {
       $p$26 = $90;
      } else {
       break;
      }
     }
     $scevgep20 = (($scevgep15) + ($89)|0);
     $p$5 = $scevgep20;
    }
   }
  }
  $104 = (($n$17) + 1)|0;
  $105 = ($104>>>0)>($$max_subs$0$lcssa>>>0);
  if ($105) {
   label = 28;
   break;
  } else {
   $n$17 = $104;$p$18 = $p$5;
  }
 }
 if ((label|0) == 15) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 else if ((label|0) == 19) {
  _ft_validator_error($valid,9);
  // unreachable;
 }
 else if ((label|0) == 24) {
  _ft_validator_error($valid,16);
  // unreachable;
 }
 else if ((label|0) == 28) {
  return 0;
 }
 return (0)|0;
}
function _tt_cmap2_get_info($cmap,$cmap_info) {
 $cmap = $cmap|0;
 $cmap_info = $cmap_info|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 4|0);
 $3 = ((($cmap_info)) + 4|0);
 HEAP32[$3>>2] = 2;
 $4 = HEAP8[$2>>0]|0;
 $5 = $4&255;
 $6 = $5 << 8;
 $7 = ((($1)) + 5|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = $8&255;
 $10 = $6 | $9;
 HEAP32[$cmap_info>>2] = $10;
 return 0;
}
function _tt_cmap4_init($cmap,$table) {
 $cmap = $cmap|0;
 $table = $table|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 HEAP32[$0>>2] = $table;
 $1 = ((($table)) + 6|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = $2&255;
 $4 = $3 << 8;
 $5 = ((($table)) + 7|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = $6&255;
 $8 = $4 | $7;
 $9 = $8 >>> 1;
 $10 = ((($cmap)) + 32|0);
 HEAP32[$10>>2] = $9;
 $11 = ((($cmap)) + 24|0);
 HEAP32[$11>>2] = -1;
 $12 = ((($cmap)) + 28|0);
 HEAP32[$12>>2] = 0;
 return 0;
}
function _tt_cmap4_char_index($cmap,$char_code) {
 $cmap = $cmap|0;
 $char_code = $char_code|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 HEAP32[$0>>2] = $char_code;
 $1 = ($char_code>>>0)>(65535);
 do {
  if ($1) {
   $$0 = 0;
  } else {
   $2 = ((($cmap)) + 20|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = $3 & 1;
   $5 = ($4|0)==(0);
   if ($5) {
    $7 = (_tt_cmap4_char_map_binary($cmap,$0,0)|0);
    $$0 = $7;
    break;
   } else {
    $6 = (_tt_cmap4_char_map_linear($cmap,$0,0)|0);
    $$0 = $6;
    break;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _tt_cmap4_char_map_linear($cmap,$pcharcode,$next) {
 $cmap = $cmap|0;
 $pcharcode = $pcharcode|0;
 $next = $next|0;
 var $$ = 0, $$0 = 0, $$10$us = 0, $$11$us = 0, $$lcssa47 = 0, $$lcssa49 = 0, $$masked = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $charcode$019$us = 0;
 var $charcode$019$us$lcssa = 0, $delta$1$us = 0, $delta$1$us$lcssa = 0, $delta$113$us = 0, $gindex$1$us = 0, $gindex$1$us$lcssa = 0, $i$016$us = 0, $offset$1$us = 0, $offset$1$us$lcssa = 0, $or$cond$us = 0, $or$cond3$us = 0, $or$cond5$us = 0, $or$cond7 = 0, $or$cond9$us = 0, $p$017$us = 0, $p$1$us = 0, $q$018$us = 0, $sext$us = 0, $switch$split2D = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 6|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = $4 << 8;
 $6 = ((($1)) + 7|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $$masked = $8 & 254;
 $9 = $$masked | $5;
 $10 = $9 >>> 1;
 $11 = ($10|0)==(0);
 if ($11) {
  $$0 = 0;
  return ($$0|0);
 }
 $12 = HEAP32[$pcharcode>>2]|0;
 $13 = ($next<<24>>24)==(0);
 $14 = $13&1;
 $15 = $14 ^ 1;
 $$ = (($12) + ($15))|0;
 $16 = ($$>>>0)<(65536);
 if (!($16)) {
  $$0 = 0;
  return ($$0|0);
 }
 $17 = ((($1)) + 14|0);
 $18 = ((($1)) + 16|0);
 $19 = (($18) + ($9)|0);
 $20 = (($10) + -1)|0;
 $charcode$019$us = $$;
 while(1) {
  $i$016$us = 0;$p$017$us = $17;$q$018$us = $19;
  L9: while(1) {
   $21 = ((($p$017$us)) + 2|0);
   $22 = HEAP8[$p$017$us>>0]|0;
   $23 = $22&255;
   $24 = $23 << 8;
   $25 = ((($p$017$us)) + 1|0);
   $26 = HEAP8[$25>>0]|0;
   $27 = $26&255;
   $28 = $24 | $27;
   $29 = ((($q$018$us)) + 2|0);
   $30 = HEAP8[$q$018$us>>0]|0;
   $31 = $30&255;
   $32 = $31 << 8;
   $33 = ((($q$018$us)) + 1|0);
   $34 = HEAP8[$33>>0]|0;
   $35 = $34&255;
   $36 = $32 | $35;
   $37 = ($charcode$019$us>>>0)<($36>>>0);
   $38 = ($charcode$019$us>>>0)>($28>>>0);
   $or$cond9$us = $38 | $37;
   if ($or$cond9$us) {
    $p$1$us = $21;
   } else {
    $39 = (($q$018$us) + ($9)|0);
    $40 = HEAP8[$39>>0]|0;
    $41 = $40&255;
    $42 = $41 << 8;
    $43 = ((($39)) + 1|0);
    $44 = HEAP8[$43>>0]|0;
    $45 = $44&255;
    $46 = $42 | $45;
    $sext$us = $46 << 16;
    $47 = $sext$us >> 16;
    $48 = (($39) + ($9)|0);
    $49 = HEAP8[$48>>0]|0;
    $50 = $49&255;
    $51 = $50 << 8;
    $52 = ((($48)) + 1|0);
    $53 = HEAP8[$52>>0]|0;
    $54 = $53&255;
    $55 = $51 | $54;
    $56 = ($i$016$us>>>0)>=($20>>>0);
    $57 = ($36|0)==(65535);
    $or$cond$us = $56 & $57;
    $58 = ($28|0)==(65535);
    $or$cond3$us = $58 & $or$cond$us;
    if ($or$cond3$us) {
     $59 = HEAP32[$cmap>>2]|0;
     $60 = ($55|0)==(0);
     if ($60) {
      $delta$113$us = $47;
      label = 11;
      break;
     }
     $61 = ((($59)) + 500|0);
     $62 = HEAP32[$61>>2]|0;
     $63 = ((($59)) + 504|0);
     $64 = HEAP32[$63>>2]|0;
     $65 = (($62) + ($64)|0);
     $66 = (($48) + ($55)|0);
     $67 = ((($66)) + 2|0);
     $68 = ($67>>>0)>($65>>>0);
     $$10$us = $68 ? 1 : $47;
     $$11$us = $68 ? 0 : $55;
     $delta$1$us = $$10$us;$offset$1$us = $$11$us;
    } else {
     $delta$1$us = $47;$offset$1$us = $55;
    }
    $switch$split2D = ($offset$1$us|0)<(65535);
    if ($switch$split2D) {
     label = 19;
     break;
    }
    switch ($offset$1$us|0) {
    case 65535:  {
     $p$1$us = $48;
     break;
    }
    default: {
     $$lcssa47 = $36;$$lcssa49 = $48;$delta$1$us$lcssa = $delta$1$us;$offset$1$us$lcssa = $offset$1$us;
     label = 9;
     break L9;
    }
    }
   }
   $88 = (($i$016$us) + 1)|0;
   $89 = ($88>>>0)<($10>>>0);
   if ($89) {
    $i$016$us = $88;$p$017$us = $p$1$us;$q$018$us = $29;
   } else {
    $gindex$1$us = 0;
    break;
   }
  }
  if ((label|0) == 19) {
   label = 0;
   switch ($offset$1$us|0) {
   case 0:  {
    $delta$113$us = $delta$1$us;
    label = 11;
    break;
   }
   default: {
    $$lcssa47 = $36;$$lcssa49 = $48;$delta$1$us$lcssa = $delta$1$us;$offset$1$us$lcssa = $offset$1$us;
    label = 9;
   }
   }
  }
  if ((label|0) == 9) {
   label = 0;
   $69 = (($charcode$019$us) - ($$lcssa47))|0;
   $70 = $69 << 1;
   $71 = (($offset$1$us$lcssa) + ($70))|0;
   $72 = (($$lcssa49) + ($71)|0);
   $73 = HEAP8[$72>>0]|0;
   $74 = $73&255;
   $75 = $74 << 8;
   $76 = ((($72)) + 1|0);
   $77 = HEAP8[$76>>0]|0;
   $78 = $77&255;
   $79 = $75 | $78;
   $80 = ($79|0)==(0);
   if ($80) {
    $gindex$1$us = 0;
   } else {
    $81 = (($79) + ($delta$1$us$lcssa))|0;
    $82 = $81 & 65535;
    $gindex$1$us = $82;
   }
  }
  else if ((label|0) == 11) {
   label = 0;
   $83 = (($delta$113$us) + ($charcode$019$us))|0;
   $84 = $83 & 65535;
   $gindex$1$us = $84;
  }
  $85 = ($gindex$1$us|0)!=(0);
  $or$cond5$us = $13 | $85;
  if ($or$cond5$us) {
   $charcode$019$us$lcssa = $charcode$019$us;$gindex$1$us$lcssa = $gindex$1$us;
   break;
  }
  $86 = (($charcode$019$us) + 1)|0;
  $87 = ($86>>>0)<(65536);
  if ($87) {
   $charcode$019$us = $86;
  } else {
   $$0 = 0;
   label = 18;
   break;
  }
 }
 if ((label|0) == 18) {
  return ($$0|0);
 }
 $90 = ($next<<24>>24)!=(0);
 $91 = ($gindex$1$us$lcssa|0)!=(0);
 $or$cond7 = $90 & $91;
 if (!($or$cond7)) {
  $$0 = $gindex$1$us$lcssa;
  return ($$0|0);
 }
 HEAP32[$pcharcode>>2] = $charcode$019$us$lcssa;
 $$0 = $gindex$1$us$lcssa;
 return ($$0|0);
}
function _tt_cmap4_char_map_binary($cmap,$pcharcode,$next) {
 $cmap = $cmap|0;
 $pcharcode = $pcharcode|0;
 $next = $next|0;
 var $$ = 0, $$1 = 0, $$12 = 0, $$13 = 0, $$14 = 0, $$lcssa = 0, $$lcssa$lcssa = 0, $$lcssa117 = 0, $$lcssa117$lcssa = 0, $$lcssa118 = 0, $$lcssa118$lcssa = 0, $$lcssa119 = 0, $$lcssa119$lcssa = 0, $$masked = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $delta$1 = 0;
 var $delta$12 = 0, $delta$2$lcssa = 0, $delta$2$lcssa$ = 0, $delta$263 = 0, $delta$6$lcssa = 0, $delta$642 = 0, $delta$9 = 0, $end$0 = 0, $end$0$ph = 0, $end$1$lcssa = 0, $end$1$lcssa$$lcssa99$lcssa = 0, $end$11 = 0, $end$12 = 0, $end$165 = 0, $end$5$lcssa = 0, $end$544 = 0, $end$8 = 0, $gindex$0 = 0, $gindex$128 = 0, $i$0$lcssa = 0;
 var $i$062 = 0, $i$1$lcssa = 0, $i$141 = 0, $i$2 = 0, $max$0 = 0, $max$0$lcssa115 = 0, $max$0$ph = 0, $mid$0 = 0, $mid$0$ph = 0, $mid$10 = 0, $mid$12 = 0, $mid$13 = 0, $mid$14 = 0, $mid$2$ = 0, $mid$2$lcssa = 0, $mid$260 = 0, $mid$6$i$1 = 0, $mid$6$lcssa = 0, $mid$640 = 0, $min$0$ph = 0;
 var $offset$1 = 0, $offset$12 = 0, $offset$2$lcssa = 0, $offset$2$lcssa$ = 0, $offset$264 = 0, $offset$6$lcssa = 0, $offset$643 = 0, $offset$9 = 0, $or$cond = 0, $or$cond3 = 0, $p$0$lcssa = 0, $p$061 = 0, $p$10 = 0, $p$6 = 0, $p$7 = 0, $sext = 0, $sext11 = 0, $sext8 = 0, $sext9 = 0, $start$0$lcssa = 0;
 var $start$0$lcssa$$lcssa101$lcssa = 0, $start$066 = 0, $start$10 = 0, $start$4$lcssa = 0, $start$445 = 0, $start$7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 6|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = $4 << 8;
 $6 = ((($1)) + 7|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $$masked = $8 & 254;
 $9 = $$masked | $5;
 $10 = ($9|0)==(0);
 if ($10) {
  $$1 = 0;
  return ($$1|0);
 }
 $11 = HEAP32[$pcharcode>>2]|0;
 $12 = $9 >>> 1;
 $13 = ($next<<24>>24)!=(0);
 $14 = $13&1;
 $$ = (($11) + ($14))|0;
 $15 = (($9) + 2)|0;
 $16 = ((($1)) + 14|0);
 $end$0$ph = 65535;$max$0$ph = $12;$mid$0$ph = $12;$min$0$ph = 0;
 L4: while(1) {
  $end$0 = $end$0$ph;$max$0 = $max$0$ph;$mid$0 = $mid$0$ph;
  while(1) {
   $17 = ($min$0$ph>>>0)<($max$0>>>0);
   if (!($17)) {
    $end$12 = $end$0;$gindex$0 = 0;$mid$13 = $mid$0;
    break L4;
   }
   $18 = (($max$0) + ($min$0$ph))|0;
   $19 = $18 >>> 1;
   $20 = $19 << 1;
   $21 = (($16) + ($20)|0);
   $22 = HEAP8[$21>>0]|0;
   $23 = $22&255;
   $24 = $23 << 8;
   $25 = ((($21)) + 1|0);
   $26 = HEAP8[$25>>0]|0;
   $27 = $26&255;
   $28 = $24 | $27;
   $29 = (($21) + ($15)|0);
   $30 = HEAP8[$29>>0]|0;
   $31 = $30&255;
   $32 = $31 << 8;
   $33 = ((($29)) + 1|0);
   $34 = HEAP8[$33>>0]|0;
   $35 = $34&255;
   $36 = $32 | $35;
   $37 = ($$>>>0)<($36>>>0);
   if ($37) {
    $end$0 = $28;$max$0 = $19;$mid$0 = $19;
   } else {
    $$lcssa = $19;$$lcssa117 = $28;$$lcssa118 = $29;$$lcssa119 = $36;$max$0$lcssa115 = $max$0;
    break;
   }
  }
  $38 = ($$>>>0)>($$lcssa117>>>0);
  if (!($38)) {
   $$lcssa$lcssa = $$lcssa;$$lcssa117$lcssa = $$lcssa117;$$lcssa118$lcssa = $$lcssa118;$$lcssa119$lcssa = $$lcssa119;
   label = 8;
   break;
  }
  $39 = (($$lcssa) + 1)|0;
  $end$0$ph = $$lcssa117;$max$0$ph = $max$0$lcssa115;$mid$0$ph = $$lcssa;$min$0$ph = $39;
 }
 do {
  if ((label|0) == 8) {
   $40 = (($$lcssa118$lcssa) + ($9)|0);
   $41 = HEAP8[$40>>0]|0;
   $42 = $41&255;
   $43 = $42 << 8;
   $44 = ((($40)) + 1|0);
   $45 = HEAP8[$44>>0]|0;
   $46 = $45&255;
   $47 = $43 | $46;
   $sext = $47 << 16;
   $48 = $sext >> 16;
   $49 = (($40) + ($9)|0);
   $50 = HEAP8[$49>>0]|0;
   $51 = $50&255;
   $52 = $51 << 8;
   $53 = ((($49)) + 1|0);
   $54 = HEAP8[$53>>0]|0;
   $55 = $54&255;
   $56 = $52 | $55;
   $57 = (($12) + -1)|0;
   $58 = ($$lcssa$lcssa>>>0)>=($57>>>0);
   $59 = ($$lcssa119$lcssa|0)==(65535);
   $or$cond = $58 & $59;
   $60 = ($$lcssa117$lcssa|0)==(65535);
   $or$cond3 = $60 & $or$cond;
   if ($or$cond3) {
    $61 = HEAP32[$cmap>>2]|0;
    $62 = ($56|0)==(0);
    if ($62) {
     $delta$1 = $48;$offset$1 = 0;
    } else {
     $63 = ((($61)) + 500|0);
     $64 = HEAP32[$63>>2]|0;
     $65 = ((($61)) + 504|0);
     $66 = HEAP32[$65>>2]|0;
     $67 = (($64) + ($66)|0);
     $68 = (($49) + ($56)|0);
     $69 = ((($68)) + 2|0);
     $70 = ($69>>>0)>($67>>>0);
     $$12 = $70 ? 1 : $48;
     $$13 = $70 ? 0 : $56;
     $delta$1 = $$12;$offset$1 = $$13;
    }
   } else {
    $delta$1 = $48;$offset$1 = $56;
   }
   $71 = ((($cmap)) + 20|0);
   $72 = HEAP32[$71>>2]|0;
   $73 = $72 & 2;
   $74 = ($73|0)==(0);
   $75 = ($offset$1|0)==(65535);
   if ($74) {
    if ($75) {
     $end$12 = $$lcssa117$lcssa;$gindex$0 = 0;$mid$13 = $$lcssa$lcssa;
     break;
    } else {
     $delta$12 = $delta$1;$end$11 = $$lcssa117$lcssa;$mid$12 = $$lcssa$lcssa;$offset$12 = $offset$1;$p$10 = $49;$start$10 = $$lcssa119$lcssa;
    }
   } else {
    $76 = (($$lcssa$lcssa) + 1)|0;
    $$14 = $75 ? $76 : $$lcssa$lcssa;
    $77 = ($$lcssa$lcssa|0)==(0);
    L20: do {
     if ($77) {
      $delta$2$lcssa = $delta$1;$end$1$lcssa = $$lcssa117$lcssa;$i$0$lcssa = 0;$mid$2$lcssa = $$14;$offset$2$lcssa = $offset$1;$p$0$lcssa = $49;$start$0$lcssa = $$lcssa119$lcssa;
     } else {
      $delta$263 = $delta$1;$end$165 = $$lcssa117$lcssa;$i$062 = $$lcssa$lcssa;$mid$260 = $$14;$offset$264 = $offset$1;$p$061 = $49;$start$066 = $$lcssa119$lcssa;
      while(1) {
       $78 = (($i$062) + -1)|0;
       $79 = $78 << 1;
       $80 = (($16) + ($79)|0);
       $81 = HEAP8[$80>>0]|0;
       $82 = $81&255;
       $83 = $82 << 8;
       $84 = ((($80)) + 1|0);
       $85 = HEAP8[$84>>0]|0;
       $86 = $85&255;
       $87 = $83 | $86;
       $88 = ($$>>>0)>($87>>>0);
       if ($88) {
        $delta$2$lcssa = $delta$263;$end$1$lcssa = $end$165;$i$0$lcssa = $i$062;$mid$2$lcssa = $mid$260;$offset$2$lcssa = $offset$264;$p$0$lcssa = $p$061;$start$0$lcssa = $start$066;
        break L20;
       }
       $89 = (($80) + ($15)|0);
       $90 = HEAP8[$89>>0]|0;
       $91 = $90&255;
       $92 = $91 << 8;
       $93 = ((($89)) + 1|0);
       $94 = HEAP8[$93>>0]|0;
       $95 = $94&255;
       $96 = $92 | $95;
       $97 = (($89) + ($9)|0);
       $98 = HEAP8[$97>>0]|0;
       $99 = $98&255;
       $100 = $99 << 8;
       $101 = ((($97)) + 1|0);
       $102 = HEAP8[$101>>0]|0;
       $103 = $102&255;
       $104 = $100 | $103;
       $sext11 = $104 << 16;
       $105 = $sext11 >> 16;
       $106 = (($97) + ($9)|0);
       $107 = HEAP8[$106>>0]|0;
       $108 = $107&255;
       $109 = $108 << 8;
       $110 = ((($106)) + 1|0);
       $111 = HEAP8[$110>>0]|0;
       $112 = $111&255;
       $113 = $109 | $112;
       $114 = ($113|0)==(65535);
       $mid$2$ = $114 ? $mid$260 : $78;
       $115 = ($78|0)==(0);
       if ($115) {
        $delta$2$lcssa = $105;$end$1$lcssa = $87;$i$0$lcssa = 0;$mid$2$lcssa = $mid$2$;$offset$2$lcssa = $113;$p$0$lcssa = $106;$start$0$lcssa = $96;
        break;
       } else {
        $delta$263 = $105;$end$165 = $87;$i$062 = $78;$mid$260 = $mid$2$;$offset$264 = $113;$p$061 = $106;$start$066 = $96;
       }
      }
     }
    } while(0);
    $116 = ($mid$2$lcssa|0)==($76|0);
    if ($116) {
     $117 = ($i$0$lcssa|0)==($$lcssa$lcssa|0);
     $end$1$lcssa$$lcssa99$lcssa = $117 ? $end$1$lcssa : $$lcssa117$lcssa;
     $118 = ($76>>>0)<($12>>>0);
     if (!($118)) {
      $end$12 = $end$1$lcssa$$lcssa99$lcssa;$gindex$0 = 0;$mid$13 = $$lcssa$lcssa;
      break;
     }
     $start$0$lcssa$$lcssa101$lcssa = $117 ? $start$0$lcssa : $$lcssa119$lcssa;
     $offset$2$lcssa$ = $117 ? $offset$2$lcssa : $56;
     $delta$2$lcssa$ = $117 ? $delta$2$lcssa : $48;
     $delta$642 = $delta$2$lcssa$;$end$544 = $end$1$lcssa$$lcssa99$lcssa;$i$141 = $76;$mid$640 = $$lcssa$lcssa;$offset$643 = $offset$2$lcssa$;$start$445 = $start$0$lcssa$$lcssa101$lcssa;
     while(1) {
      $119 = $i$141 << 1;
      $120 = (($16) + ($119)|0);
      $121 = (($120) + ($15)|0);
      $122 = HEAP8[$121>>0]|0;
      $123 = $122&255;
      $124 = $123 << 8;
      $125 = ((($121)) + 1|0);
      $126 = HEAP8[$125>>0]|0;
      $127 = $126&255;
      $128 = $124 | $127;
      $129 = ($$>>>0)<($128>>>0);
      if ($129) {
       $delta$6$lcssa = $delta$642;$end$5$lcssa = $end$544;$i$1$lcssa = $i$141;$mid$6$lcssa = $mid$640;$offset$6$lcssa = $offset$643;$p$6 = $121;$start$4$lcssa = $start$445;
       break;
      }
      $130 = HEAP8[$120>>0]|0;
      $131 = $130&255;
      $132 = $131 << 8;
      $133 = ((($120)) + 1|0);
      $134 = HEAP8[$133>>0]|0;
      $135 = $134&255;
      $136 = $132 | $135;
      $137 = (($121) + ($9)|0);
      $138 = HEAP8[$137>>0]|0;
      $139 = $138&255;
      $140 = $139 << 8;
      $141 = ((($137)) + 1|0);
      $142 = HEAP8[$141>>0]|0;
      $143 = $142&255;
      $144 = $140 | $143;
      $sext9 = $144 << 16;
      $145 = $sext9 >> 16;
      $146 = (($137) + ($9)|0);
      $147 = HEAP8[$146>>0]|0;
      $148 = $147&255;
      $149 = $148 << 8;
      $150 = ((($146)) + 1|0);
      $151 = HEAP8[$150>>0]|0;
      $152 = $151&255;
      $153 = $149 | $152;
      $154 = ($153|0)==(65535);
      $mid$6$i$1 = $154 ? $mid$640 : $i$141;
      $155 = (($i$141) + 1)|0;
      $156 = ($155>>>0)<($12>>>0);
      if ($156) {
       $delta$642 = $145;$end$544 = $136;$i$141 = $155;$mid$640 = $mid$6$i$1;$offset$643 = $153;$start$445 = $128;
      } else {
       $delta$6$lcssa = $145;$end$5$lcssa = $136;$i$1$lcssa = $155;$mid$6$lcssa = $mid$6$i$1;$offset$6$lcssa = $153;$p$6 = $146;$start$4$lcssa = $128;
       break;
      }
     }
     $157 = (($i$1$lcssa) + -1)|0;
     $158 = ($mid$6$lcssa|0)==($$lcssa$lcssa|0);
     if ($158) {
      $end$12 = $end$5$lcssa;$gindex$0 = 0;$mid$13 = $157;
      break;
     } else {
      $delta$9 = $delta$6$lcssa;$end$8 = $end$5$lcssa;$i$2 = $157;$mid$10 = $mid$6$lcssa;$offset$9 = $offset$6$lcssa;$p$7 = $p$6;$start$7 = $start$4$lcssa;
     }
    } else {
     $delta$9 = $delta$2$lcssa;$end$8 = $end$1$lcssa;$i$2 = $i$0$lcssa;$mid$10 = $mid$2$lcssa;$offset$9 = $offset$2$lcssa;$p$7 = $p$0$lcssa;$start$7 = $start$0$lcssa;
    }
    $159 = ($mid$10|0)==($i$2|0);
    if ($159) {
     $delta$12 = $delta$9;$end$11 = $end$8;$mid$12 = $mid$10;$offset$12 = $offset$9;$p$10 = $p$7;$start$10 = $start$7;
    } else {
     $160 = $mid$10 << 1;
     $161 = (($16) + ($160)|0);
     $162 = HEAP8[$161>>0]|0;
     $163 = $162&255;
     $164 = $163 << 8;
     $165 = ((($161)) + 1|0);
     $166 = HEAP8[$165>>0]|0;
     $167 = $166&255;
     $168 = $164 | $167;
     $169 = (($161) + ($15)|0);
     $170 = HEAP8[$169>>0]|0;
     $171 = $170&255;
     $172 = $171 << 8;
     $173 = ((($169)) + 1|0);
     $174 = HEAP8[$173>>0]|0;
     $175 = $174&255;
     $176 = $172 | $175;
     $177 = (($169) + ($9)|0);
     $178 = HEAP8[$177>>0]|0;
     $179 = $178&255;
     $180 = $179 << 8;
     $181 = ((($177)) + 1|0);
     $182 = HEAP8[$181>>0]|0;
     $183 = $182&255;
     $184 = $180 | $183;
     $sext8 = $184 << 16;
     $185 = $sext8 >> 16;
     $186 = (($177) + ($9)|0);
     $187 = HEAP8[$186>>0]|0;
     $188 = $187&255;
     $189 = $188 << 8;
     $190 = ((($186)) + 1|0);
     $191 = HEAP8[$190>>0]|0;
     $192 = $191&255;
     $193 = $189 | $192;
     $delta$12 = $185;$end$11 = $168;$mid$12 = $mid$10;$offset$12 = $193;$p$10 = $186;$start$10 = $176;
    }
   }
   $194 = ($offset$12|0)==(0);
   if ($194) {
    $209 = (($delta$12) + ($$))|0;
    $210 = $209 & 65535;
    $end$12 = $end$11;$gindex$0 = $210;$mid$13 = $mid$12;
    break;
   }
   $195 = (($$) - ($start$10))|0;
   $196 = $195 << 1;
   $197 = (($196) + ($offset$12))|0;
   $198 = (($p$10) + ($197)|0);
   $199 = HEAP8[$198>>0]|0;
   $200 = $199&255;
   $201 = $200 << 8;
   $202 = ((($198)) + 1|0);
   $203 = HEAP8[$202>>0]|0;
   $204 = $203&255;
   $205 = $201 | $204;
   $206 = ($205|0)==(0);
   if ($206) {
    $end$12 = $end$11;$gindex$0 = 0;$mid$13 = $mid$12;
   } else {
    $207 = (($205) + ($delta$12))|0;
    $208 = $207 & 65535;
    $end$12 = $end$11;$gindex$0 = $208;$mid$13 = $mid$12;
   }
  }
 } while(0);
 if (!($13)) {
  $$1 = $gindex$0;
  return ($$1|0);
 }
 $211 = ($$>>>0)>($end$12>>>0);
 if ($211) {
  $212 = (($mid$13) + 1)|0;
  $213 = ($212|0)==($12|0);
  if ($213) {
   $$1 = 0;
   return ($$1|0);
  } else {
   $mid$14 = $212;
  }
 } else {
  $mid$14 = $mid$13;
 }
 $214 = (_tt_cmap4_set_range($cmap,$mid$14)|0);
 $215 = ($214|0)==(0);
 if (!($215)) {
  $216 = ($gindex$0|0)==(0);
  if ($216) {
   $$1 = 0;
   return ($$1|0);
  }
  HEAP32[$pcharcode>>2] = $$;
  $$1 = $gindex$0;
  return ($$1|0);
 }
 $217 = ((($cmap)) + 24|0);
 HEAP32[$217>>2] = $$;
 $218 = ($gindex$0|0)==(0);
 do {
  if ($218) {
   HEAP32[$217>>2] = $$;
   _tt_cmap4_next($cmap);
   $220 = ((($cmap)) + 28|0);
   $221 = HEAP32[$220>>2]|0;
   $222 = ($221|0)==(0);
   if ($222) {
    $$1 = 0;
    return ($$1|0);
   } else {
    $$pre = HEAP32[$217>>2]|0;
    $223 = $$pre;$gindex$128 = $221;
    break;
   }
  } else {
   $219 = ((($cmap)) + 28|0);
   HEAP32[$219>>2] = $gindex$0;
   $223 = $$;$gindex$128 = $gindex$0;
  }
 } while(0);
 HEAP32[$pcharcode>>2] = $223;
 $$1 = $gindex$128;
 return ($$1|0);
}
function _tt_cmap4_set_range($cmap,$range_index) {
 $cmap = $cmap|0;
 $range_index = $range_index|0;
 var $$0211 = 0, $$0211$lcssa33 = 0, $$021135 = 0, $$2 = 0, $$lcssa36 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $7 = 0, $8 = 0, $9 = 0, $offset$14 = 0, $or$cond = 0, $or$cond22 = 0, $sext = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1>>>0)>($range_index>>>0);
 if (!($2)) {
  $$2 = -1;
  return ($$2|0);
 }
 $3 = ((($cmap)) + 16|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 14|0);
 $6 = ((($cmap)) + 44|0);
 $7 = $1 << 1;
 $8 = (($7) + 2)|0;
 $9 = ((($cmap)) + 40|0);
 $10 = ((($cmap)) + 48|0);
 $11 = (($1) + -1)|0;
 $$0211 = $range_index;
 while(1) {
  $12 = $$0211 << 1;
  $13 = (($5) + ($12)|0);
  $14 = HEAP8[$13>>0]|0;
  $15 = $14&255;
  $16 = $15 << 8;
  $17 = ((($13)) + 1|0);
  $18 = HEAP8[$17>>0]|0;
  $19 = $18&255;
  $20 = $16 | $19;
  HEAP32[$6>>2] = $20;
  $21 = (($13) + ($8)|0);
  $22 = HEAP8[$21>>0]|0;
  $23 = $22&255;
  $24 = $23 << 8;
  $25 = ((($21)) + 1|0);
  $26 = HEAP8[$25>>0]|0;
  $27 = $26&255;
  $28 = $24 | $27;
  HEAP32[$9>>2] = $28;
  $29 = (($21) + ($7)|0);
  $30 = HEAP8[$29>>0]|0;
  $31 = $30&255;
  $32 = $31 << 8;
  $33 = ((($29)) + 1|0);
  $34 = HEAP8[$33>>0]|0;
  $35 = $34&255;
  $36 = $32 | $35;
  $sext = $36 << 16;
  $37 = $sext >> 16;
  HEAP32[$10>>2] = $37;
  $38 = (($29) + ($7)|0);
  $39 = HEAP8[$38>>0]|0;
  $40 = $39&255;
  $41 = $40 << 8;
  $42 = ((($38)) + 1|0);
  $43 = HEAP8[$42>>0]|0;
  $44 = $43&255;
  $45 = $41 | $44;
  $46 = ($$0211>>>0)>=($11>>>0);
  $47 = ($28|0)==(65535);
  $or$cond = $46 & $47;
  $48 = ($20|0)==(65535);
  $or$cond22 = $or$cond & $48;
  if ($or$cond22) {
   $49 = HEAP32[$cmap>>2]|0;
   $50 = ($45|0)==(0);
   if ($50) {
    $$021135 = $$0211;$64 = $38;$offset$14 = 0;
    break;
   }
   $51 = ((($49)) + 500|0);
   $52 = HEAP32[$51>>2]|0;
   $53 = ((($49)) + 504|0);
   $54 = HEAP32[$53>>2]|0;
   $55 = (($52) + ($54)|0);
   $56 = (($38) + ($45)|0);
   $57 = ((($56)) + 2|0);
   $58 = ($57>>>0)>($55>>>0);
   if ($58) {
    $$0211$lcssa33 = $$0211;$$lcssa36 = $38;
    label = 6;
    break;
   }
  }
  $59 = ($45|0)==(65535);
  if (!($59)) {
   $$021135 = $$0211;$64 = $38;$offset$14 = $45;
   break;
  }
  $60 = (($$0211) + 1)|0;
  $61 = ($60>>>0)<($1>>>0);
  if ($61) {
   $$0211 = $60;
  } else {
   $$2 = -1;
   label = 10;
   break;
  }
 }
 if ((label|0) == 6) {
  HEAP32[$10>>2] = 1;
  $$021135 = $$0211$lcssa33;$64 = $$lcssa36;$offset$14 = 0;
 }
 else if ((label|0) == 10) {
  return ($$2|0);
 }
 $62 = ($offset$14|0)!=(0);
 $63 = (($64) + ($offset$14)|0);
 $65 = $62 ? $63 : 0;
 $66 = ((($cmap)) + 52|0);
 HEAP32[$66>>2] = $65;
 $67 = ((($cmap)) + 36|0);
 HEAP32[$67>>2] = $$021135;
 $$2 = 0;
 return ($$2|0);
}
function _tt_cmap4_next($cmap) {
 $cmap = $cmap|0;
 var $$ = 0, $$charcode$4 = 0, $$lcssa = 0, $$lcssa47 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $charcode$0 = 0, $charcode$1 = 0, $charcode$1$lcssa = 0, $charcode$3 = 0, $charcode$3$lcssa = 0, $charcode$4 = 0, $p$0 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1>>>0)>(65534);
 L1: do {
  if (!($2)) {
   $3 = (($1) + 1)|0;
   $4 = ((($cmap)) + 40|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = ($3>>>0)<($5>>>0);
   $$ = $6 ? $5 : $3;
   $7 = ((($cmap)) + 52|0);
   $8 = ((($cmap)) + 44|0);
   $9 = ((($cmap)) + 48|0);
   $10 = ((($cmap)) + 36|0);
   $17 = $5;$charcode$0 = $$;
   L3: while(1) {
    $11 = HEAP32[$7>>2]|0;
    $12 = HEAP32[$8>>2]|0;
    $13 = HEAP32[$9>>2]|0;
    $14 = ($charcode$0>>>0)>($12>>>0);
    L5: do {
     if ($14) {
      $charcode$4 = $charcode$0;
     } else {
      $15 = ($11|0)==(0|0);
      if ($15) {
       $charcode$3 = $charcode$0;
       while(1) {
        $35 = (($charcode$3) + ($13))|0;
        $36 = $35 & 65535;
        $37 = ($36|0)==(0);
        if (!($37)) {
         $$lcssa47 = $36;$charcode$3$lcssa = $charcode$3;
         label = 11;
         break L3;
        }
        $39 = (($charcode$3) + 1)|0;
        $40 = ($39>>>0)>($12>>>0);
        if ($40) {
         $charcode$4 = $39;
         break L5;
        } else {
         $charcode$3 = $39;
        }
       }
      }
      $16 = (($charcode$0) - ($17))|0;
      $18 = $16 << 1;
      $19 = (($11) + ($18)|0);
      $charcode$1 = $charcode$0;$p$0 = $19;
      while(1) {
       $20 = ((($p$0)) + 2|0);
       $21 = HEAP8[$p$0>>0]|0;
       $22 = $21&255;
       $23 = $22 << 8;
       $24 = ((($p$0)) + 1|0);
       $25 = HEAP8[$24>>0]|0;
       $26 = $25&255;
       $27 = $23 | $26;
       $28 = ($27|0)==(0);
       if (!($28)) {
        $29 = (($27) + ($13))|0;
        $30 = $29 & 65535;
        $31 = ($30|0)==(0);
        if (!($31)) {
         $$lcssa = $30;$charcode$1$lcssa = $charcode$1;
         label = 9;
         break L3;
        }
       }
       $32 = (($charcode$1) + 1)|0;
       $33 = ($32>>>0)>($12>>>0);
       if ($33) {
        $charcode$4 = $32;
        break;
       } else {
        $charcode$1 = $32;$p$0 = $20;
       }
      }
     }
    } while(0);
    $41 = HEAP32[$10>>2]|0;
    $42 = (($41) + 1)|0;
    $43 = (_tt_cmap4_set_range($cmap,$42)|0);
    $44 = ($43|0)<(0);
    if ($44) {
     break L1;
    }
    $45 = HEAP32[$4>>2]|0;
    $46 = ($charcode$4>>>0)<($45>>>0);
    $$charcode$4 = $46 ? $45 : $charcode$4;
    $17 = $45;$charcode$0 = $$charcode$4;
   }
   if ((label|0) == 9) {
    HEAP32[$0>>2] = $charcode$1$lcssa;
    $34 = ((($cmap)) + 28|0);
    HEAP32[$34>>2] = $$lcssa;
    return;
   }
   else if ((label|0) == 11) {
    HEAP32[$0>>2] = $charcode$3$lcssa;
    $38 = ((($cmap)) + 28|0);
    HEAP32[$38>>2] = $$lcssa47;
    return;
   }
  }
 } while(0);
 HEAP32[$0>>2] = -1;
 $47 = ((($cmap)) + 28|0);
 HEAP32[$47>>2] = 0;
 return;
}
function _tt_cmap4_char_next($cmap,$pchar_code) {
 $cmap = $cmap|0;
 $pchar_code = $pchar_code|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$pchar_code>>2]|0;
 $1 = ($0>>>0)>(65534);
 if ($1) {
  $$0 = 0;
  return ($$0|0);
 }
 $2 = ((($cmap)) + 20|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 1;
 $5 = ($4|0)==(0);
 if (!($5)) {
  $6 = (_tt_cmap4_char_map_linear($cmap,$pchar_code,1)|0);
  $$0 = $6;
  return ($$0|0);
 }
 $7 = ((($cmap)) + 24|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($0|0)==($8|0);
 if (!($9)) {
  $14 = (_tt_cmap4_char_map_binary($cmap,$pchar_code,1)|0);
  $$0 = $14;
  return ($$0|0);
 }
 _tt_cmap4_next($cmap);
 $10 = ((($cmap)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0);
 if ($12) {
  $$0 = 0;
  return ($$0|0);
 }
 $13 = HEAP32[$7>>2]|0;
 HEAP32[$pchar_code>>2] = $13;
 $$0 = $11;
 return ($$0|0);
}
function _tt_cmap4_validate($table,$valid) {
 $table = $table|0;
 $valid = $valid|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $error$0$lcssa = 0, $error$024 = 0, $error$1 = 0, $exitcond = 0, $exitcond34 = 0, $i$023 = 0, $last_end$027 = 0, $last_start$026 = 0, $length$0 = 0, $n$025 = 0, $notlhs = 0, $notrhs = 0, $or$cond = 0, $or$cond11 = 0, $or$cond13 = 0, $or$cond15$not = 0, $or$cond17 = 0, $or$cond19 = 0;
 var $or$cond20 = 0, $or$cond33 = 0, $or$cond5 = 0, $or$cond7 = 0, $or$cond9 = 0, $p$022 = 0, $p_delta$030 = 0, $p_end$029 = 0, $p_offset$031 = 0, $p_start$028 = 0, $switch$split2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($table)) + 4|0);
 $1 = ((($valid)) + 160|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($0>>>0)>($2>>>0);
 if ($3) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $4 = ((($table)) + 2|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = $6 << 8;
 $8 = ((($table)) + 3|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 $11 = $7 | $10;
 $12 = (($table) + ($11)|0);
 $13 = HEAP32[$1>>2]|0;
 $14 = ($12>>>0)>($13>>>0);
 do {
  if ($14) {
   $15 = ((($valid)) + 164|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = ($16|0)==(0);
   if ($17) {
    $18 = HEAP32[$1>>2]|0;
    $19 = $18;
    $20 = $table;
    $21 = (($19) - ($20))|0;
    $length$0 = $21;
    break;
   } else {
    _ft_validator_error($valid,8);
    // unreachable;
   }
  } else {
   $length$0 = $11;
  }
 } while(0);
 $22 = ($length$0>>>0)<(16);
 if ($22) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $23 = ((($table)) + 6|0);
 $24 = HEAP8[$23>>0]|0;
 $25 = $24&255;
 $26 = $25 << 8;
 $27 = ((($table)) + 7|0);
 $28 = HEAP8[$27>>0]|0;
 $29 = $28&255;
 $30 = $26 | $29;
 $31 = ((($valid)) + 164|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ($32>>>0)<(2);
 $34 = $29 & 1;
 $35 = ($34|0)==(0);
 $or$cond13 = $33 | $35;
 if (!($or$cond13)) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $36 = $30 >>> 1;
 $37 = $36 << 1;
 $38 = $36 << 3;
 $39 = (($38) + 16)|0;
 $40 = ($length$0>>>0)<($39>>>0);
 if ($40) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $41 = HEAP32[$31>>2]|0;
 $42 = ($41>>>0)>(1);
 if ($42) {
  $43 = ((($table)) + 8|0);
  $44 = HEAP8[$43>>0]|0;
  $45 = $44&255;
  $46 = $45 << 8;
  $47 = ((($table)) + 9|0);
  $48 = HEAP8[$47>>0]|0;
  $49 = $48&255;
  $50 = $46 | $49;
  $51 = ((($table)) + 10|0);
  $52 = HEAP8[$51>>0]|0;
  $53 = $52&255;
  $54 = $53 << 8;
  $55 = ((($table)) + 11|0);
  $56 = HEAP8[$55>>0]|0;
  $57 = $56&255;
  $58 = $54 | $57;
  $59 = ((($table)) + 12|0);
  $60 = HEAP8[$59>>0]|0;
  $61 = $60&255;
  $62 = $61 << 8;
  $63 = ((($table)) + 13|0);
  $64 = HEAP8[$63>>0]|0;
  $65 = $64&255;
  $66 = $62 | $65;
  $67 = $64 | $48;
  $68 = $67 & 1;
  $69 = ($68<<24>>24)==(0);
  if (!($69)) {
   _ft_validator_error($valid,8);
   // unreachable;
  }
  $70 = $50 >>> 1;
  $71 = $66 >>> 1;
  $72 = $70 << 1;
  $notlhs = ($70>>>0)<=($36>>>0);
  $notrhs = ($72>>>0)>=($36>>>0);
  $or$cond15$not = $notlhs & $notrhs;
  $73 = (($71) + ($70))|0;
  $74 = ($73|0)==($36|0);
  $or$cond17 = $or$cond15$not & $74;
  $75 = 1 << $58;
  $76 = ($70|0)==($75|0);
  $or$cond19 = $76 & $or$cond17;
  if (!($or$cond19)) {
   _ft_validator_error($valid,8);
   // unreachable;
  }
 }
 $77 = ((($table)) + 14|0);
 $78 = ((($table)) + 16|0);
 $79 = (($78) + ($37)|0);
 $80 = (($79) + ($37)|0);
 $81 = (($80) + ($37)|0);
 $82 = (($81) + ($37)|0);
 $83 = HEAP32[$31>>2]|0;
 $84 = ($83>>>0)>(1);
 if ($84) {
  $85 = (($37) + -2)|0;
  $86 = (($77) + ($85)|0);
  $87 = HEAP8[$86>>0]|0;
  $88 = $87&255;
  $89 = $88 << 8;
  $90 = ((($86)) + 1|0);
  $91 = HEAP8[$90>>0]|0;
  $92 = $91&255;
  $93 = $89 | $92;
  $94 = ($93|0)==(65535);
  if (!($94)) {
   _ft_validator_error($valid,8);
   // unreachable;
  }
 }
 $95 = ($36|0)==(0);
 if ($95) {
  $error$0$lcssa = 0;
  return ($error$0$lcssa|0);
 }
 $96 = (($36) + -1)|0;
 $97 = ((($valid)) + 172|0);
 $98 = (($table) + ($length$0)|0);
 $error$024 = 0;$last_end$027 = 0;$last_start$026 = 0;$n$025 = 0;$p_delta$030 = $80;$p_end$029 = $77;$p_offset$031 = $81;$p_start$028 = $79;
 L33: while(1) {
  $99 = ((($p_start$028)) + 2|0);
  $100 = HEAP8[$p_start$028>>0]|0;
  $101 = $100&255;
  $102 = $101 << 8;
  $103 = ((($p_start$028)) + 1|0);
  $104 = HEAP8[$103>>0]|0;
  $105 = $104&255;
  $106 = $102 | $105;
  $107 = ((($p_end$029)) + 2|0);
  $108 = HEAP8[$p_end$029>>0]|0;
  $109 = $108&255;
  $110 = $109 << 8;
  $111 = ((($p_end$029)) + 1|0);
  $112 = HEAP8[$111>>0]|0;
  $113 = $112&255;
  $114 = $110 | $113;
  $115 = ((($p_delta$030)) + 2|0);
  $116 = HEAP8[$p_delta$030>>0]|0;
  $117 = $116&255;
  $118 = $117 << 8;
  $119 = ((($p_delta$030)) + 1|0);
  $120 = HEAP8[$119>>0]|0;
  $121 = $120&255;
  $122 = $118 | $121;
  $123 = ((($p_offset$031)) + 2|0);
  $124 = HEAP8[$p_offset$031>>0]|0;
  $125 = $124&255;
  $126 = $125 << 8;
  $127 = ((($p_offset$031)) + 1|0);
  $128 = HEAP8[$127>>0]|0;
  $129 = $128&255;
  $130 = $126 | $129;
  $131 = ($114>>>0)<($106>>>0);
  if ($131) {
   label = 24;
   break;
  }
  $132 = ($106>>>0)<=($last_end$027>>>0);
  $133 = ($n$025|0)!=(0);
  $or$cond = $133 & $132;
  do {
   if ($or$cond) {
    $134 = HEAP32[$31>>2]|0;
    $135 = ($134|0)==(0);
    if (!($135)) {
     label = 27;
     break L33;
    }
    $136 = ($last_start$026>>>0)>($106>>>0);
    $137 = ($last_end$027>>>0)>($114>>>0);
    $or$cond20 = $136 | $137;
    if ($or$cond20) {
     $138 = $error$024 | 1;
     $error$1 = $138;
     break;
    } else {
     $139 = $error$024 | 2;
     $error$1 = $139;
     break;
    }
   } else {
    $error$1 = $error$024;
   }
  } while(0);
  $switch$split2D = ($130|0)<(65535);
  L43: do {
   if ($switch$split2D) {
    switch ($130|0) {
    case 0:  {
     break;
    }
    default: {
     label = 32;
    }
    }
   } else {
    switch ($130|0) {
    case 65535:  {
     break;
    }
    default: {
     label = 32;
     break L43;
    }
    }
    $176 = HEAP32[$31>>2]|0;
    $177 = ($176>>>0)>(1);
    if ($177) {
     label = 47;
     break L33;
    }
    $178 = ($n$025|0)==($96|0);
    $179 = ($106|0)==(65535);
    $or$cond9 = $178 & $179;
    $180 = ($114|0)==(65535);
    $or$cond11 = $or$cond9 & $180;
    if (!($or$cond11)) {
     label = 47;
     break L33;
    }
   }
  } while(0);
  if ((label|0) == 32) {
   label = 0;
   $140 = (($p_offset$031) + ($130)|0);
   $141 = HEAP32[$31>>2]|0;
   $142 = ($141|0)==(0);
   if ($142) {
    $149 = ($n$025|0)==($96|0);
    $150 = ($106|0)==(65535);
    $or$cond5 = $149 & $150;
    $151 = ($114|0)==(65535);
    $or$cond7 = $or$cond5 & $151;
    if (!($or$cond7)) {
     $152 = ($140>>>0)<($82>>>0);
     if ($152) {
      label = 39;
      break;
     }
     $153 = (($114) - ($106))|0;
     $154 = $153 << 1;
     $155 = (($154) + 2)|0;
     $156 = (($140) + ($155)|0);
     $157 = HEAP32[$1>>2]|0;
     $158 = ($156>>>0)>($157>>>0);
     if ($158) {
      label = 39;
      break;
     }
    }
   } else {
    $143 = ($140>>>0)<($82>>>0);
    if ($143) {
     label = 35;
     break;
    }
    $144 = (($114) - ($106))|0;
    $145 = $144 << 1;
    $146 = (($145) + 2)|0;
    $147 = (($140) + ($146)|0);
    $148 = ($147>>>0)>($98>>>0);
    if ($148) {
     label = 35;
     break;
    }
   }
   $159 = HEAP32[$31>>2]|0;
   $160 = ($159|0)!=(0);
   $161 = ($114>>>0)>($106>>>0);
   $or$cond33 = $160 & $161;
   if ($or$cond33) {
    $i$023 = $106;$p$022 = $140;
    while(1) {
     $162 = ((($p$022)) + 2|0);
     $163 = HEAP8[$p$022>>0]|0;
     $164 = $163&255;
     $165 = $164 << 8;
     $166 = ((($p$022)) + 1|0);
     $167 = HEAP8[$166>>0]|0;
     $168 = $167&255;
     $169 = $165 | $168;
     $170 = ($169|0)==(0);
     if (!($170)) {
      $171 = (($169) + ($122))|0;
      $172 = $171 & 65535;
      $173 = HEAP32[$97>>2]|0;
      $174 = ($172>>>0)<($173>>>0);
      if (!($174)) {
       label = 43;
       break L33;
      }
     }
     $175 = (($i$023) + 1)|0;
     $exitcond = ($175|0)==($114|0);
     if ($exitcond) {
      break;
     } else {
      $i$023 = $175;$p$022 = $162;
     }
    }
   }
  }
  $181 = (($n$025) + 1)|0;
  $exitcond34 = ($181|0)==($36|0);
  if ($exitcond34) {
   $error$0$lcssa = $error$1;
   label = 49;
   break;
  } else {
   $error$024 = $error$1;$last_end$027 = $114;$last_start$026 = $106;$n$025 = $181;$p_delta$030 = $115;$p_end$029 = $107;$p_offset$031 = $123;$p_start$028 = $99;
  }
 }
 if ((label|0) == 24) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 else if ((label|0) == 27) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 else if ((label|0) == 35) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 else if ((label|0) == 39) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 else if ((label|0) == 43) {
  _ft_validator_error($valid,16);
  // unreachable;
 }
 else if ((label|0) == 47) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 else if ((label|0) == 49) {
  return ($error$0$lcssa|0);
 }
 return (0)|0;
}
function _tt_cmap4_get_info($cmap,$cmap_info) {
 $cmap = $cmap|0;
 $cmap_info = $cmap_info|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 4|0);
 $3 = ((($cmap_info)) + 4|0);
 HEAP32[$3>>2] = 4;
 $4 = HEAP8[$2>>0]|0;
 $5 = $4&255;
 $6 = $5 << 8;
 $7 = ((($1)) + 5|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = $8&255;
 $10 = $6 | $9;
 HEAP32[$cmap_info>>2] = $10;
 return 0;
}
function _tt_cmap6_char_index($cmap,$char_code) {
 $cmap = $cmap|0;
 $char_code = $char_code|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 6|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = $4 << 8;
 $6 = ((($1)) + 7|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $9 = $5 | $8;
 $10 = ((($1)) + 8|0);
 $11 = HEAP8[$10>>0]|0;
 $12 = $11&255;
 $13 = $12 << 8;
 $14 = ((($1)) + 9|0);
 $15 = HEAP8[$14>>0]|0;
 $16 = $15&255;
 $17 = $13 | $16;
 $18 = (($char_code) - ($9))|0;
 $19 = ($18>>>0)<($17>>>0);
 if (!($19)) {
  $result$0 = 0;
  return ($result$0|0);
 }
 $20 = ((($1)) + 10|0);
 $21 = $18 << 1;
 $22 = (($20) + ($21)|0);
 $23 = HEAP8[$22>>0]|0;
 $24 = $23&255;
 $25 = $24 << 8;
 $26 = ((($22)) + 1|0);
 $27 = HEAP8[$26>>0]|0;
 $28 = $27&255;
 $29 = $25 | $28;
 $result$0 = $29;
 return ($result$0|0);
}
function _tt_cmap6_char_next($cmap,$pchar_code) {
 $cmap = $cmap|0;
 $pchar_code = $pchar_code|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $char_code$12 = 0, $gindex$1 = 0, $idx$03 = 0, $p$01 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$pchar_code>>2]|0;
 $3 = (($2) + 1)|0;
 $4 = ((($1)) + 6|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = $6 << 8;
 $8 = ((($1)) + 7|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 $11 = $7 | $10;
 $12 = ((($1)) + 8|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = $13&255;
 $15 = $14 << 8;
 $16 = ((($1)) + 9|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = $17&255;
 $19 = $15 | $18;
 $20 = ($3>>>0)>(65535);
 if ($20) {
  $gindex$1 = 0;$result$0 = 0;
  HEAP32[$pchar_code>>2] = $result$0;
  return ($gindex$1|0);
 }
 $21 = ($3>>>0)<($11>>>0);
 $$ = $21 ? $11 : $3;
 $22 = (($$) - ($11))|0;
 $23 = ($22>>>0)<($19>>>0);
 if (!($23)) {
  $gindex$1 = 0;$result$0 = 0;
  HEAP32[$pchar_code>>2] = $result$0;
  return ($gindex$1|0);
 }
 $24 = ((($1)) + 10|0);
 $25 = $22 << 1;
 $26 = (($24) + ($25)|0);
 $char_code$12 = $$;$idx$03 = $22;$p$01 = $26;
 while(1) {
  $27 = HEAP8[$p$01>>0]|0;
  $28 = $27&255;
  $29 = $28 << 8;
  $30 = ((($p$01)) + 1|0);
  $31 = HEAP8[$30>>0]|0;
  $32 = $31&255;
  $33 = $29 | $32;
  $34 = ($33|0)==(0);
  if (!($34)) {
   $gindex$1 = $33;$result$0 = $char_code$12;
   label = 6;
   break;
  }
  $35 = ((($p$01)) + 2|0);
  $36 = (($char_code$12) + 1)|0;
  $37 = (($idx$03) + 1)|0;
  $38 = ($37>>>0)<($19>>>0);
  if ($38) {
   $char_code$12 = $36;$idx$03 = $37;$p$01 = $35;
  } else {
   $gindex$1 = 0;$result$0 = 0;
   label = 6;
   break;
  }
 }
 if ((label|0) == 6) {
  HEAP32[$pchar_code>>2] = $result$0;
  return ($gindex$1|0);
 }
 return (0)|0;
}
function _tt_cmap6_validate($table,$valid) {
 $table = $table|0;
 $valid = $valid|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $count$01 = 0, $or$cond = 0, $p$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($table)) + 10|0);
 $1 = ((($valid)) + 160|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($0>>>0)>($2>>>0);
 if ($3) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $4 = ((($table)) + 2|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = $6 << 8;
 $8 = ((($table)) + 3|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 $11 = $7 | $10;
 $12 = ((($table)) + 8|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = $13&255;
 $15 = $14 << 8;
 $16 = ((($table)) + 9|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = $17&255;
 $19 = $15 | $18;
 $20 = (($table) + ($11)|0);
 $21 = HEAP32[$1>>2]|0;
 $22 = ($20>>>0)>($21>>>0);
 if ($22) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $23 = $19 << 1;
 $24 = (($23) + 10)|0;
 $25 = ($11>>>0)<($24>>>0);
 if ($25) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $26 = ((($valid)) + 164|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ($27|0)==(0);
 $29 = ($19|0)==(0);
 $or$cond = $28 | $29;
 if ($or$cond) {
  return 0;
 }
 $30 = ((($valid)) + 172|0);
 $count$01 = $19;$p$02 = $0;
 while(1) {
  $31 = ((($p$02)) + 2|0);
  $32 = HEAP8[$p$02>>0]|0;
  $33 = $32&255;
  $34 = $33 << 8;
  $35 = ((($p$02)) + 1|0);
  $36 = HEAP8[$35>>0]|0;
  $37 = $36&255;
  $38 = $34 | $37;
  $39 = HEAP32[$30>>2]|0;
  $40 = ($38>>>0)<($39>>>0);
  if (!($40)) {
   label = 9;
   break;
  }
  $41 = (($count$01) + -1)|0;
  $42 = ($41|0)==(0);
  if ($42) {
   label = 11;
   break;
  } else {
   $count$01 = $41;$p$02 = $31;
  }
 }
 if ((label|0) == 9) {
  _ft_validator_error($valid,16);
  // unreachable;
 }
 else if ((label|0) == 11) {
  return 0;
 }
 return (0)|0;
}
function _tt_cmap6_get_info($cmap,$cmap_info) {
 $cmap = $cmap|0;
 $cmap_info = $cmap_info|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 4|0);
 $3 = ((($cmap_info)) + 4|0);
 HEAP32[$3>>2] = 6;
 $4 = HEAP8[$2>>0]|0;
 $5 = $4&255;
 $6 = $5 << 8;
 $7 = ((($1)) + 5|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = $8&255;
 $10 = $6 | $9;
 HEAP32[$cmap_info>>2] = $10;
 return 0;
}
function _tt_cmap8_char_index($cmap,$char_code) {
 $cmap = $cmap|0;
 $char_code = $char_code|0;
 var $$lcssa34 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $9 = 0, $num_groups$02 = 0, $p$03 = 0, $p$03$lcssa33 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8204|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = $4 << 24;
 $6 = ((($1)) + 8205|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $9 = $8 << 16;
 $10 = $9 | $5;
 $11 = ((($1)) + 8206|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = $12&255;
 $14 = $13 << 8;
 $15 = $10 | $14;
 $16 = ((($1)) + 8207|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = $17&255;
 $19 = $15 | $18;
 $20 = ($19|0)==(0);
 if ($20) {
  $result$0 = 0;
  return ($result$0|0);
 }
 $21 = ((($1)) + 8208|0);
 $num_groups$02 = $19;$p$03 = $21;
 while(1) {
  $25 = HEAP8[$p$03>>0]|0;
  $26 = $25&255;
  $27 = $26 << 24;
  $28 = ((($p$03)) + 1|0);
  $29 = HEAP8[$28>>0]|0;
  $30 = $29&255;
  $31 = $30 << 16;
  $32 = $31 | $27;
  $33 = ((($p$03)) + 2|0);
  $34 = HEAP8[$33>>0]|0;
  $35 = $34&255;
  $36 = $35 << 8;
  $37 = $32 | $36;
  $38 = ((($p$03)) + 3|0);
  $39 = HEAP8[$38>>0]|0;
  $40 = $39&255;
  $41 = $37 | $40;
  $42 = ($41>>>0)>($char_code>>>0);
  if ($42) {
   $result$0 = 0;
   label = 7;
   break;
  }
  $43 = ((($p$03)) + 4|0);
  $44 = HEAP8[$43>>0]|0;
  $45 = $44&255;
  $46 = $45 << 24;
  $47 = ((($p$03)) + 5|0);
  $48 = HEAP8[$47>>0]|0;
  $49 = $48&255;
  $50 = $49 << 16;
  $51 = $50 | $46;
  $52 = ((($p$03)) + 6|0);
  $53 = HEAP8[$52>>0]|0;
  $54 = $53&255;
  $55 = $54 << 8;
  $56 = $51 | $55;
  $57 = ((($p$03)) + 7|0);
  $58 = HEAP8[$57>>0]|0;
  $59 = $58&255;
  $60 = $56 | $59;
  $61 = ($60>>>0)<($char_code>>>0);
  $24 = (($num_groups$02) + -1)|0;
  if (!($61)) {
   $$lcssa34 = $41;$p$03$lcssa33 = $p$03;
   label = 6;
   break;
  }
  $22 = ((($p$03)) + 12|0);
  $23 = ($24|0)==(0);
  if ($23) {
   $result$0 = 0;
   label = 7;
   break;
  } else {
   $num_groups$02 = $24;$p$03 = $22;
  }
 }
 if ((label|0) == 6) {
  $62 = ((($p$03$lcssa33)) + 8|0);
  $63 = HEAP8[$62>>0]|0;
  $64 = $63&255;
  $65 = $64 << 24;
  $66 = ((($p$03$lcssa33)) + 9|0);
  $67 = HEAP8[$66>>0]|0;
  $68 = $67&255;
  $69 = $68 << 16;
  $70 = $69 | $65;
  $71 = ((($p$03$lcssa33)) + 10|0);
  $72 = HEAP8[$71>>0]|0;
  $73 = $72&255;
  $74 = $73 << 8;
  $75 = $70 | $74;
  $76 = ((($p$03$lcssa33)) + 11|0);
  $77 = HEAP8[$76>>0]|0;
  $78 = $77&255;
  $79 = $75 | $78;
  $80 = (($char_code) - ($$lcssa34))|0;
  $81 = (($80) + ($79))|0;
  $result$0 = $81;
  return ($result$0|0);
 }
 else if ((label|0) == 7) {
  return ($result$0|0);
 }
 return (0)|0;
}
function _tt_cmap8_char_next($cmap,$pchar_code) {
 $cmap = $cmap|0;
 $pchar_code = $pchar_code|0;
 var $$char_code$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $9 = 0, $char_code$03 = 0, $gindex$2 = 0, $num_groups$02 = 0, $p$01 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8204|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = $4 << 24;
 $6 = ((($1)) + 8205|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $9 = $8 << 16;
 $10 = $9 | $5;
 $11 = ((($1)) + 8206|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = $12&255;
 $14 = $13 << 8;
 $15 = $10 | $14;
 $16 = ((($1)) + 8207|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = $17&255;
 $19 = $15 | $18;
 $20 = ($19|0)==(0);
 if ($20) {
  $gindex$2 = 0;$result$0 = 0;
  HEAP32[$pchar_code>>2] = $result$0;
  return ($gindex$2|0);
 }
 $21 = ((($1)) + 8208|0);
 $22 = HEAP32[$pchar_code>>2]|0;
 $23 = (($22) + 1)|0;
 $char_code$03 = $23;$num_groups$02 = $19;$p$01 = $21;
 while(1) {
  $24 = HEAP8[$p$01>>0]|0;
  $25 = $24&255;
  $26 = $25 << 24;
  $27 = ((($p$01)) + 1|0);
  $28 = HEAP8[$27>>0]|0;
  $29 = $28&255;
  $30 = $29 << 16;
  $31 = $30 | $26;
  $32 = ((($p$01)) + 2|0);
  $33 = HEAP8[$32>>0]|0;
  $34 = $33&255;
  $35 = $34 << 8;
  $36 = $31 | $35;
  $37 = ((($p$01)) + 3|0);
  $38 = HEAP8[$37>>0]|0;
  $39 = $38&255;
  $40 = $36 | $39;
  $41 = ((($p$01)) + 4|0);
  $42 = HEAP8[$41>>0]|0;
  $43 = $42&255;
  $44 = $43 << 24;
  $45 = ((($p$01)) + 5|0);
  $46 = HEAP8[$45>>0]|0;
  $47 = $46&255;
  $48 = $47 << 16;
  $49 = $48 | $44;
  $50 = ((($p$01)) + 6|0);
  $51 = HEAP8[$50>>0]|0;
  $52 = $51&255;
  $53 = $52 << 8;
  $54 = $49 | $53;
  $55 = ((($p$01)) + 7|0);
  $56 = HEAP8[$55>>0]|0;
  $57 = $56&255;
  $58 = $54 | $57;
  $59 = ((($p$01)) + 12|0);
  $60 = ($char_code$03>>>0)<($40>>>0);
  $$char_code$0 = $60 ? $40 : $char_code$03;
  $61 = ($$char_code$0>>>0)>($58>>>0);
  if (!($61)) {
   $62 = ((($p$01)) + 8|0);
   $63 = HEAP8[$62>>0]|0;
   $64 = $63&255;
   $65 = $64 << 24;
   $66 = ((($p$01)) + 9|0);
   $67 = HEAP8[$66>>0]|0;
   $68 = $67&255;
   $69 = $68 << 16;
   $70 = $69 | $65;
   $71 = ((($p$01)) + 10|0);
   $72 = HEAP8[$71>>0]|0;
   $73 = $72&255;
   $74 = $73 << 8;
   $75 = $70 | $74;
   $76 = ((($p$01)) + 11|0);
   $77 = HEAP8[$76>>0]|0;
   $78 = $77&255;
   $79 = $75 | $78;
   $80 = (($$char_code$0) - ($40))|0;
   $81 = (($79) + ($80))|0;
   $82 = ($81|0)==(0);
   if (!($82)) {
    $gindex$2 = $81;$result$0 = $$char_code$0;
    label = 6;
    break;
   }
  }
  $83 = (($num_groups$02) + -1)|0;
  $84 = ($83|0)==(0);
  if ($84) {
   $gindex$2 = 0;$result$0 = 0;
   label = 6;
   break;
  } else {
   $char_code$03 = $$char_code$0;$num_groups$02 = $83;$p$01 = $59;
  }
 }
 if ((label|0) == 6) {
  HEAP32[$pchar_code>>2] = $result$0;
  return ($gindex$2|0);
 }
 return (0)|0;
}
function _tt_cmap8_validate($table,$valid) {
 $table = $table|0;
 $valid = $valid|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $count$09 = 0;
 var $count$111 = 0, $exitcond = 0, $last$013 = 0, $n$014 = 0, $or$cond = 0, $or$cond2 = 0, $or$cond4 = 0, $p$015 = 0, $start$08 = 0, $start$110 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($table)) + 8208|0);
 $1 = ((($valid)) + 160|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($0>>>0)>($2>>>0);
 if ($3) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $4 = ((($table)) + 4|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = $6 << 24;
 $8 = ((($table)) + 5|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 $11 = $10 << 16;
 $12 = $11 | $7;
 $13 = ((($table)) + 6|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = $14&255;
 $16 = $15 << 8;
 $17 = $12 | $16;
 $18 = ((($table)) + 7|0);
 $19 = HEAP8[$18>>0]|0;
 $20 = $19&255;
 $21 = $17 | $20;
 $22 = HEAP32[$1>>2]|0;
 $23 = $22;
 $24 = $table;
 $25 = (($23) - ($24))|0;
 $26 = ($21>>>0)>($25>>>0);
 $27 = ($21>>>0)<(8208);
 $or$cond = $26 | $27;
 if ($or$cond) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $28 = ((($table)) + 12|0);
 $29 = ((($table)) + 8204|0);
 $30 = HEAP8[$29>>0]|0;
 $31 = $30&255;
 $32 = $31 << 24;
 $33 = ((($table)) + 8205|0);
 $34 = HEAP8[$33>>0]|0;
 $35 = $34&255;
 $36 = $35 << 16;
 $37 = $36 | $32;
 $38 = ((($table)) + 8206|0);
 $39 = HEAP8[$38>>0]|0;
 $40 = $39&255;
 $41 = $40 << 8;
 $42 = $37 | $41;
 $43 = ((($table)) + 8207|0);
 $44 = HEAP8[$43>>0]|0;
 $45 = $44&255;
 $46 = $42 | $45;
 $47 = HEAP32[$1>>2]|0;
 $48 = $47;
 $49 = $0;
 $50 = (($48) - ($49))|0;
 $51 = (($50>>>0) / 12)&-1;
 $52 = ($46>>>0)>($51>>>0);
 if ($52) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $53 = ($46|0)==(0);
 if ($53) {
  return 0;
 }
 $54 = ((($valid)) + 164|0);
 $55 = ((($valid)) + 172|0);
 $last$013 = 0;$n$014 = 0;$p$015 = $0;
 L13: while(1) {
  $56 = HEAP8[$p$015>>0]|0;
  $57 = $56&255;
  $58 = $57 << 24;
  $59 = ((($p$015)) + 1|0);
  $60 = HEAP8[$59>>0]|0;
  $61 = $60&255;
  $62 = $61 << 16;
  $63 = $62 | $58;
  $64 = ((($p$015)) + 2|0);
  $65 = HEAP8[$64>>0]|0;
  $66 = $65&255;
  $67 = $66 << 8;
  $68 = $63 | $67;
  $69 = ((($p$015)) + 3|0);
  $70 = HEAP8[$69>>0]|0;
  $71 = $70&255;
  $72 = $68 | $71;
  $73 = ((($p$015)) + 4|0);
  $74 = HEAP8[$73>>0]|0;
  $75 = $74&255;
  $76 = $75 << 24;
  $77 = ((($p$015)) + 5|0);
  $78 = HEAP8[$77>>0]|0;
  $79 = $78&255;
  $80 = $79 << 16;
  $81 = $80 | $76;
  $82 = ((($p$015)) + 6|0);
  $83 = HEAP8[$82>>0]|0;
  $84 = $83&255;
  $85 = $84 << 8;
  $86 = $81 | $85;
  $87 = ((($p$015)) + 7|0);
  $88 = HEAP8[$87>>0]|0;
  $89 = $88&255;
  $90 = $86 | $89;
  $91 = ((($p$015)) + 12|0);
  $92 = ((($p$015)) + 8|0);
  $93 = HEAP8[$92>>0]|0;
  $94 = $93&255;
  $95 = $94 << 24;
  $96 = ((($p$015)) + 9|0);
  $97 = HEAP8[$96>>0]|0;
  $98 = $97&255;
  $99 = $98 << 16;
  $100 = $99 | $95;
  $101 = ((($p$015)) + 10|0);
  $102 = HEAP8[$101>>0]|0;
  $103 = $102&255;
  $104 = $103 << 8;
  $105 = $100 | $104;
  $106 = ((($p$015)) + 11|0);
  $107 = HEAP8[$106>>0]|0;
  $108 = $107&255;
  $109 = $105 | $108;
  $110 = ($90>>>0)<($72>>>0);
  if ($110) {
   label = 10;
   break;
  }
  $111 = ($n$014|0)==(0);
  $112 = ($72>>>0)>($last$013>>>0);
  $or$cond2 = $111 | $112;
  if (!($or$cond2)) {
   label = 12;
   break;
  }
  $113 = HEAP32[$54>>2]|0;
  $114 = ($113|0)==(0);
  do {
   if (!($114)) {
    $115 = (($90) - ($72))|0;
    $116 = HEAP32[$55>>2]|0;
    $117 = ($116>>>0)>=($115>>>0);
    $118 = (($116) - ($115))|0;
    $119 = ($109>>>0)<($118>>>0);
    $or$cond4 = $117 & $119;
    if (!($or$cond4)) {
     label = 15;
     break L13;
    }
    $120 = (($115) + 1)|0;
    $121 = ($63|0)==(0);
    if ($121) {
     $144 = ($81|0)==(0);
     if (!($144)) {
      label = 24;
      break L13;
     }
     $145 = ($120|0)==(0);
     if ($145) {
      break;
     } else {
      $count$111 = $120;$start$110 = $72;
     }
     while(1) {
      $146 = $start$110 >>> 3;
      $147 = $146 & 8191;
      $148 = (($28) + ($147)|0);
      $149 = HEAP8[$148>>0]|0;
      $150 = $149&255;
      $151 = $start$110 & 7;
      $152 = 128 >>> $151;
      $153 = $150 & $152;
      $154 = ($153|0)==(0);
      if (!($154)) {
       label = 27;
       break L13;
      }
      $155 = (($count$111) + -1)|0;
      $156 = (($start$110) + 1)|0;
      $157 = ($155|0)==(0);
      if ($157) {
       break;
      } else {
       $count$111 = $155;$start$110 = $156;
      }
     }
    } else {
     $122 = ($120|0)==(0);
     if ($122) {
      break;
     } else {
      $count$09 = $120;$start$08 = $72;
     }
     while(1) {
      $123 = $start$08 >>> 16;
      $124 = $start$08 >>> 19;
      $125 = (($28) + ($124)|0);
      $126 = HEAP8[$125>>0]|0;
      $127 = $126&255;
      $128 = $123 & 7;
      $129 = 128 >>> $128;
      $130 = $127 & $129;
      $131 = ($130|0)==(0);
      if ($131) {
       label = 19;
       break L13;
      }
      $132 = $start$08 >>> 3;
      $133 = $132 & 8191;
      $134 = (($28) + ($133)|0);
      $135 = HEAP8[$134>>0]|0;
      $136 = $135&255;
      $137 = $start$08 & 7;
      $138 = 128 >>> $137;
      $139 = $136 & $138;
      $140 = ($139|0)==(0);
      if ($140) {
       label = 21;
       break L13;
      }
      $141 = (($count$09) + -1)|0;
      $142 = (($start$08) + 1)|0;
      $143 = ($141|0)==(0);
      if ($143) {
       break;
      } else {
       $count$09 = $141;$start$08 = $142;
      }
     }
    }
   }
  } while(0);
  $158 = (($n$014) + 1)|0;
  $exitcond = ($158|0)==($46|0);
  if ($exitcond) {
   label = 30;
   break;
  } else {
   $last$013 = $90;$n$014 = $158;$p$015 = $91;
  }
 }
 if ((label|0) == 10) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 else if ((label|0) == 12) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 else if ((label|0) == 15) {
  _ft_validator_error($valid,16);
  // unreachable;
 }
 else if ((label|0) == 19) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 else if ((label|0) == 21) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 else if ((label|0) == 24) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 else if ((label|0) == 27) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 else if ((label|0) == 30) {
  return 0;
 }
 return (0)|0;
}
function _tt_cmap8_get_info($cmap,$cmap_info) {
 $cmap = $cmap|0;
 $cmap_info = $cmap_info|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8|0);
 $3 = ((($cmap_info)) + 4|0);
 HEAP32[$3>>2] = 8;
 $4 = HEAP8[$2>>0]|0;
 $5 = $4&255;
 $6 = $5 << 24;
 $7 = ((($1)) + 9|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = $8&255;
 $10 = $9 << 16;
 $11 = $10 | $6;
 $12 = ((($1)) + 10|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = $13&255;
 $15 = $14 << 8;
 $16 = $11 | $15;
 $17 = ((($1)) + 11|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = $18&255;
 $20 = $16 | $19;
 HEAP32[$cmap_info>>2] = $20;
 return 0;
}
function _tt_cmap10_char_index($cmap,$char_code) {
 $cmap = $cmap|0;
 $char_code = $char_code|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 12|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = $4 << 24;
 $6 = ((($1)) + 13|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $9 = $8 << 16;
 $10 = $9 | $5;
 $11 = ((($1)) + 14|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = $12&255;
 $14 = $13 << 8;
 $15 = $10 | $14;
 $16 = ((($1)) + 15|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = $17&255;
 $19 = $15 | $18;
 $20 = ((($1)) + 16|0);
 $21 = HEAP8[$20>>0]|0;
 $22 = $21&255;
 $23 = $22 << 24;
 $24 = ((($1)) + 17|0);
 $25 = HEAP8[$24>>0]|0;
 $26 = $25&255;
 $27 = $26 << 16;
 $28 = $27 | $23;
 $29 = ((($1)) + 18|0);
 $30 = HEAP8[$29>>0]|0;
 $31 = $30&255;
 $32 = $31 << 8;
 $33 = $28 | $32;
 $34 = ((($1)) + 19|0);
 $35 = HEAP8[$34>>0]|0;
 $36 = $35&255;
 $37 = $33 | $36;
 $38 = (($char_code) - ($19))|0;
 $39 = ($38>>>0)<($37>>>0);
 if (!($39)) {
  $result$0 = 0;
  return ($result$0|0);
 }
 $40 = ((($1)) + 20|0);
 $41 = $38 << 1;
 $42 = (($40) + ($41)|0);
 $43 = HEAP8[$42>>0]|0;
 $44 = $43&255;
 $45 = $44 << 8;
 $46 = ((($42)) + 1|0);
 $47 = HEAP8[$46>>0]|0;
 $48 = $47&255;
 $49 = $45 | $48;
 $result$0 = $49;
 return ($result$0|0);
}
function _tt_cmap10_char_next($cmap,$pchar_code) {
 $cmap = $cmap|0;
 $pchar_code = $pchar_code|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $char_code$1$lcssa = 0;
 var $char_code$12 = 0, $gindex$1 = 0, $idx$03 = 0, $p$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$pchar_code>>2]|0;
 $3 = (($2) + 1)|0;
 $4 = ((($1)) + 12|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = $6 << 24;
 $8 = ((($1)) + 13|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 $11 = $10 << 16;
 $12 = $11 | $7;
 $13 = ((($1)) + 14|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = $14&255;
 $16 = $15 << 8;
 $17 = $12 | $16;
 $18 = ((($1)) + 15|0);
 $19 = HEAP8[$18>>0]|0;
 $20 = $19&255;
 $21 = $17 | $20;
 $22 = ((($1)) + 16|0);
 $23 = HEAP8[$22>>0]|0;
 $24 = $23&255;
 $25 = $24 << 24;
 $26 = ((($1)) + 17|0);
 $27 = HEAP8[$26>>0]|0;
 $28 = $27&255;
 $29 = $28 << 16;
 $30 = $29 | $25;
 $31 = ((($1)) + 18|0);
 $32 = HEAP8[$31>>0]|0;
 $33 = $32&255;
 $34 = $33 << 8;
 $35 = $30 | $34;
 $36 = ((($1)) + 19|0);
 $37 = HEAP8[$36>>0]|0;
 $38 = $37&255;
 $39 = $35 | $38;
 $40 = ($3>>>0)<($21>>>0);
 $$ = $40 ? $21 : $3;
 $41 = (($$) - ($21))|0;
 $42 = ($41>>>0)<($39>>>0);
 if (!($42)) {
  $char_code$1$lcssa = $$;$gindex$1 = 0;
  HEAP32[$pchar_code>>2] = $char_code$1$lcssa;
  return ($gindex$1|0);
 }
 $43 = ((($1)) + 20|0);
 $44 = $41 << 1;
 $45 = (($43) + ($44)|0);
 $char_code$12 = $$;$idx$03 = $41;$p$01 = $45;
 while(1) {
  $46 = HEAP8[$p$01>>0]|0;
  $47 = $46&255;
  $48 = $47 << 8;
  $49 = ((($p$01)) + 1|0);
  $50 = HEAP8[$49>>0]|0;
  $51 = $50&255;
  $52 = $48 | $51;
  $53 = ($52|0)==(0);
  if (!($53)) {
   $char_code$1$lcssa = $char_code$12;$gindex$1 = $52;
   label = 5;
   break;
  }
  $54 = ((($p$01)) + 2|0);
  $55 = (($char_code$12) + 1)|0;
  $56 = (($idx$03) + 1)|0;
  $57 = ($56>>>0)<($39>>>0);
  if ($57) {
   $char_code$12 = $55;$idx$03 = $56;$p$01 = $54;
  } else {
   $char_code$1$lcssa = $55;$gindex$1 = 0;
   label = 5;
   break;
  }
 }
 if ((label|0) == 5) {
  HEAP32[$pchar_code>>2] = $char_code$1$lcssa;
  return ($gindex$1|0);
 }
 return (0)|0;
}
function _tt_cmap10_validate($table,$valid) {
 $table = $table|0;
 $valid = $valid|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $7 = 0, $8 = 0, $9 = 0, $count$02 = 0, $or$cond = 0, $or$cond4 = 0, $p$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($table)) + 20|0);
 $1 = ((($valid)) + 160|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($0>>>0)>($2>>>0);
 if ($3) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $4 = ((($table)) + 4|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = $6 << 24;
 $8 = ((($table)) + 5|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 $11 = $10 << 16;
 $12 = $11 | $7;
 $13 = ((($table)) + 6|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = $14&255;
 $16 = $15 << 8;
 $17 = $12 | $16;
 $18 = ((($table)) + 7|0);
 $19 = HEAP8[$18>>0]|0;
 $20 = $19&255;
 $21 = $17 | $20;
 $22 = ((($table)) + 16|0);
 $23 = HEAP8[$22>>0]|0;
 $24 = $23&255;
 $25 = $24 << 24;
 $26 = ((($table)) + 17|0);
 $27 = HEAP8[$26>>0]|0;
 $28 = $27&255;
 $29 = $28 << 16;
 $30 = $29 | $25;
 $31 = ((($table)) + 18|0);
 $32 = HEAP8[$31>>0]|0;
 $33 = $32&255;
 $34 = $33 << 8;
 $35 = $30 | $34;
 $36 = ((($table)) + 19|0);
 $37 = HEAP8[$36>>0]|0;
 $38 = $37&255;
 $39 = $35 | $38;
 $40 = HEAP32[$1>>2]|0;
 $41 = $40;
 $42 = $table;
 $43 = (($41) - ($42))|0;
 $44 = ($21>>>0)>($43>>>0);
 $45 = ($21>>>0)<(20);
 $or$cond = $45 | $44;
 if ($or$cond) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $46 = (($21) + -20)|0;
 $47 = $46 >>> 1;
 $48 = ($47>>>0)<($39>>>0);
 if ($48) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $49 = ((($valid)) + 164|0);
 $50 = HEAP32[$49>>2]|0;
 $51 = ($50|0)==(0);
 $52 = ($39|0)==(0);
 $or$cond4 = $51 | $52;
 if ($or$cond4) {
  return 0;
 }
 $53 = ((($valid)) + 172|0);
 $count$02 = $39;$p$03 = $0;
 while(1) {
  $54 = ((($p$03)) + 2|0);
  $55 = HEAP8[$p$03>>0]|0;
  $56 = $55&255;
  $57 = $56 << 8;
  $58 = ((($p$03)) + 1|0);
  $59 = HEAP8[$58>>0]|0;
  $60 = $59&255;
  $61 = $57 | $60;
  $62 = HEAP32[$53>>2]|0;
  $63 = ($61>>>0)<($62>>>0);
  if (!($63)) {
   label = 9;
   break;
  }
  $64 = (($count$02) + -1)|0;
  $65 = ($64|0)==(0);
  if ($65) {
   label = 11;
   break;
  } else {
   $count$02 = $64;$p$03 = $54;
  }
 }
 if ((label|0) == 9) {
  _ft_validator_error($valid,16);
  // unreachable;
 }
 else if ((label|0) == 11) {
  return 0;
 }
 return (0)|0;
}
function _tt_cmap10_get_info($cmap,$cmap_info) {
 $cmap = $cmap|0;
 $cmap_info = $cmap_info|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8|0);
 $3 = ((($cmap_info)) + 4|0);
 HEAP32[$3>>2] = 10;
 $4 = HEAP8[$2>>0]|0;
 $5 = $4&255;
 $6 = $5 << 24;
 $7 = ((($1)) + 9|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = $8&255;
 $10 = $9 << 16;
 $11 = $10 | $6;
 $12 = ((($1)) + 10|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = $13&255;
 $15 = $14 << 8;
 $16 = $11 | $15;
 $17 = ((($1)) + 11|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = $18&255;
 $20 = $16 | $19;
 HEAP32[$cmap_info>>2] = $20;
 return 0;
}
function _tt_cmap12_init($cmap,$table) {
 $cmap = $cmap|0;
 $table = $table|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 HEAP32[$0>>2] = $table;
 $1 = ((($table)) + 12|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = $2&255;
 $4 = $3 << 24;
 $5 = ((($table)) + 13|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = $6&255;
 $8 = $7 << 16;
 $9 = $8 | $4;
 $10 = ((($table)) + 14|0);
 $11 = HEAP8[$10>>0]|0;
 $12 = $11&255;
 $13 = $12 << 8;
 $14 = $9 | $13;
 $15 = ((($table)) + 15|0);
 $16 = HEAP8[$15>>0]|0;
 $17 = $16&255;
 $18 = $14 | $17;
 $19 = ((($cmap)) + 40|0);
 HEAP32[$19>>2] = $18;
 $20 = ((($cmap)) + 24|0);
 HEAP8[$20>>0] = 0;
 return 0;
}
function _tt_cmap12_char_index($cmap,$char_code) {
 $cmap = $cmap|0;
 $char_code = $char_code|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 HEAP32[$0>>2] = $char_code;
 $1 = (_tt_cmap12_char_map_binary($cmap,$0,0)|0);
 STACKTOP = sp;return ($1|0);
}
function _tt_cmap12_char_map_binary($cmap,$pchar_code,$next) {
 $cmap = $cmap|0;
 $pchar_code = $pchar_code|0;
 $next = $next|0;
 var $$ = 0, $$1 = 0, $$lcssa = 0, $$lcssa84 = 0, $$lcssa87 = 0, $$lcssa90 = 0, $$lcssa92 = 0, $$lcssa93 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $11 = 0;
 var $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $gindex$0 = 0, $gindex$13 = 0, $max$0$ph22 = 0, $mid$2 = 0;
 var $min$0$ph23 = 0, $min$018 = 0, $min$018$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 12|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = $4 << 24;
 $6 = ((($1)) + 13|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $9 = $8 << 16;
 $10 = $9 | $5;
 $11 = ((($1)) + 14|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = $12&255;
 $14 = $13 << 8;
 $15 = $10 | $14;
 $16 = ((($1)) + 15|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = $17&255;
 $19 = $15 | $18;
 $20 = ($19|0)==(0);
 if ($20) {
  $$1 = 0;
  return ($$1|0);
 }
 $21 = HEAP32[$pchar_code>>2]|0;
 $22 = ($next<<24>>24)!=(0);
 $23 = $22&1;
 $$ = (($21) + ($23))|0;
 $24 = ((($1)) + 16|0);
 $max$0$ph22 = $19;$min$0$ph23 = 0;
 L4: while(1) {
  $min$018 = $min$0$ph23;
  while(1) {
   $26 = (($min$018) + ($max$0$ph22))|0;
   $27 = $26 >>> 1;
   $28 = ($27*12)|0;
   $29 = (($24) + ($28)|0);
   $30 = ((($29)) + 4|0);
   $31 = ((($30)) + -4|0);
   $32 = HEAP8[$31>>0]|0;
   $33 = $32&255;
   $34 = $33 << 24;
   $35 = ((($31)) + 1|0);
   $36 = HEAP8[$35>>0]|0;
   $37 = $36&255;
   $38 = $37 << 16;
   $39 = $38 | $34;
   $40 = ((($31)) + 2|0);
   $41 = HEAP8[$40>>0]|0;
   $42 = $41&255;
   $43 = $42 << 8;
   $44 = $39 | $43;
   $45 = ((($31)) + 3|0);
   $46 = HEAP8[$45>>0]|0;
   $47 = $46&255;
   $48 = $44 | $47;
   $49 = ((($30)) + 4|0);
   $50 = ((($49)) + -4|0);
   $51 = HEAP8[$50>>0]|0;
   $52 = $51&255;
   $53 = $52 << 24;
   $54 = ((($50)) + 1|0);
   $55 = HEAP8[$54>>0]|0;
   $56 = $55&255;
   $57 = $56 << 16;
   $58 = $57 | $53;
   $59 = ((($50)) + 2|0);
   $60 = HEAP8[$59>>0]|0;
   $61 = $60&255;
   $62 = $61 << 8;
   $63 = $58 | $62;
   $64 = ((($50)) + 3|0);
   $65 = HEAP8[$64>>0]|0;
   $66 = $65&255;
   $67 = $63 | $66;
   $68 = ($$>>>0)<($48>>>0);
   if ($68) {
    $$lcssa = $27;$$lcssa92 = $67;$min$018$lcssa = $min$018;
    break;
   }
   $69 = ($$>>>0)>($67>>>0);
   if (!($69)) {
    $$lcssa84 = $27;$$lcssa87 = $48;$$lcssa90 = $49;$$lcssa93 = $67;
    label = 8;
    break L4;
   }
   $70 = (($27) + 1)|0;
   $71 = ($70>>>0)<($max$0$ph22>>>0);
   if ($71) {
    $min$018 = $70;
   } else {
    $92 = $67;$94 = $27;$gindex$0 = 0;
    break L4;
   }
  }
  $25 = ($min$018$lcssa>>>0)<($$lcssa>>>0);
  if ($25) {
   $max$0$ph22 = $$lcssa;$min$0$ph23 = $min$018$lcssa;
  } else {
   $92 = $$lcssa92;$94 = $$lcssa;$gindex$0 = 0;
   break;
  }
 }
 if ((label|0) == 8) {
  $72 = HEAP8[$$lcssa90>>0]|0;
  $73 = $72&255;
  $74 = $73 << 24;
  $75 = ((($$lcssa90)) + 1|0);
  $76 = HEAP8[$75>>0]|0;
  $77 = $76&255;
  $78 = $77 << 16;
  $79 = $78 | $74;
  $80 = ((($$lcssa90)) + 2|0);
  $81 = HEAP8[$80>>0]|0;
  $82 = $81&255;
  $83 = $82 << 8;
  $84 = $79 | $83;
  $85 = ((($$lcssa90)) + 3|0);
  $86 = HEAP8[$85>>0]|0;
  $87 = $86&255;
  $88 = $84 | $87;
  $89 = (($$) - ($$lcssa87))|0;
  $90 = (($89) + ($88))|0;
  $92 = $$lcssa93;$94 = $$lcssa84;$gindex$0 = $90;
 }
 if (!($22)) {
  $$1 = $gindex$0;
  return ($$1|0);
 }
 $91 = ($$>>>0)>($92>>>0);
 if ($91) {
  $93 = (($94) + 1)|0;
  $95 = ($93|0)==($19|0);
  if ($95) {
   $$1 = 0;
   return ($$1|0);
  } else {
   $mid$2 = $93;
  }
 } else {
  $mid$2 = $94;
 }
 $96 = ((($cmap)) + 24|0);
 HEAP8[$96>>0] = 1;
 $97 = ((($cmap)) + 28|0);
 HEAP32[$97>>2] = $$;
 $98 = ((($cmap)) + 36|0);
 HEAP32[$98>>2] = $mid$2;
 $99 = ($gindex$0|0)==(0);
 do {
  if ($99) {
   _tt_cmap12_next($cmap);
   $100 = HEAP8[$96>>0]|0;
   $101 = ($100<<24>>24)==(0);
   if ($101) {
    $$1 = 0;
    return ($$1|0);
   }
   $103 = ((($cmap)) + 32|0);
   $104 = HEAP32[$103>>2]|0;
   $105 = ($104|0)==(0);
   if ($105) {
    $$1 = 0;
    return ($$1|0);
   } else {
    $$pre = HEAP32[$97>>2]|0;
    $106 = $$pre;$gindex$13 = $104;
    break;
   }
  } else {
   $102 = ((($cmap)) + 32|0);
   HEAP32[$102>>2] = $gindex$0;
   $106 = $$;$gindex$13 = $gindex$0;
  }
 } while(0);
 HEAP32[$pchar_code>>2] = $106;
 $$1 = $gindex$13;
 return ($$1|0);
}
function _tt_cmap12_next($cmap) {
 $cmap = $cmap|0;
 var $$char_code$0 = 0, $$lcssa = 0, $$lcssa38 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $9 = 0, $char_code$07 = 0, $char_code$1$lcssa = 0, $char_code$14 = 0, $char_code$14$lcssa = 0, $n$05 = 0, $n$05$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(-1);
 L1: do {
  if (!($2)) {
   $3 = ((($cmap)) + 36|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ((($cmap)) + 40|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   if ($7) {
    $8 = (($1) + 1)|0;
    $9 = ((($cmap)) + 16|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = ((($10)) + 16|0);
    $char_code$07 = $8;$n$05 = $4;
    L4: while(1) {
     $12 = ($n$05*12)|0;
     $13 = (($11) + ($12)|0);
     $14 = ((($13)) + 4|0);
     $15 = ((($14)) + -4|0);
     $16 = HEAP8[$15>>0]|0;
     $17 = $16&255;
     $18 = $17 << 24;
     $19 = ((($15)) + 1|0);
     $20 = HEAP8[$19>>0]|0;
     $21 = $20&255;
     $22 = $21 << 16;
     $23 = $22 | $18;
     $24 = ((($15)) + 2|0);
     $25 = HEAP8[$24>>0]|0;
     $26 = $25&255;
     $27 = $26 << 8;
     $28 = $23 | $27;
     $29 = ((($15)) + 3|0);
     $30 = HEAP8[$29>>0]|0;
     $31 = $30&255;
     $32 = $28 | $31;
     $33 = ((($14)) + 4|0);
     $34 = ((($33)) + -4|0);
     $35 = HEAP8[$34>>0]|0;
     $36 = $35&255;
     $37 = $36 << 24;
     $38 = ((($34)) + 1|0);
     $39 = HEAP8[$38>>0]|0;
     $40 = $39&255;
     $41 = $40 << 16;
     $42 = $41 | $37;
     $43 = ((($34)) + 2|0);
     $44 = HEAP8[$43>>0]|0;
     $45 = $44&255;
     $46 = $45 << 8;
     $47 = $42 | $46;
     $48 = ((($34)) + 3|0);
     $49 = HEAP8[$48>>0]|0;
     $50 = $49&255;
     $51 = $47 | $50;
     $52 = HEAP8[$33>>0]|0;
     $53 = $52&255;
     $54 = $53 << 24;
     $55 = ((($33)) + 1|0);
     $56 = HEAP8[$55>>0]|0;
     $57 = $56&255;
     $58 = $57 << 16;
     $59 = $58 | $54;
     $60 = ((($33)) + 2|0);
     $61 = HEAP8[$60>>0]|0;
     $62 = $61&255;
     $63 = $62 << 8;
     $64 = $59 | $63;
     $65 = ((($33)) + 3|0);
     $66 = HEAP8[$65>>0]|0;
     $67 = $66&255;
     $68 = $64 | $67;
     $69 = ($char_code$07>>>0)<($32>>>0);
     $$char_code$0 = $69 ? $32 : $char_code$07;
     $70 = ($$char_code$0>>>0)>($51>>>0);
     if ($70) {
      $char_code$1$lcssa = $$char_code$0;
     } else {
      $char_code$14 = $$char_code$0;
      while(1) {
       $71 = (($char_code$14) + ($68))|0;
       $72 = ($71|0)==($32|0);
       if (!($72)) {
        $$lcssa = $71;$$lcssa38 = $32;$char_code$14$lcssa = $char_code$14;$n$05$lcssa = $n$05;
        break L4;
       }
       $75 = (($char_code$14) + 1)|0;
       $76 = ($75>>>0)>($51>>>0);
       if ($76) {
        $char_code$1$lcssa = $75;
        break;
       } else {
        $char_code$14 = $75;
       }
      }
     }
     $77 = (($n$05) + 1)|0;
     $78 = ($77>>>0)<($6>>>0);
     if ($78) {
      $char_code$07 = $char_code$1$lcssa;$n$05 = $77;
     } else {
      break L1;
     }
    }
    $73 = (($$lcssa) - ($$lcssa38))|0;
    HEAP32[$0>>2] = $char_code$14$lcssa;
    $74 = ((($cmap)) + 32|0);
    HEAP32[$74>>2] = $73;
    HEAP32[$3>>2] = $n$05$lcssa;
    return;
   }
  }
 } while(0);
 $79 = ((($cmap)) + 24|0);
 HEAP8[$79>>0] = 0;
 return;
}
function _tt_cmap12_char_next($cmap,$pchar_code) {
 $cmap = $cmap|0;
 $pchar_code = $pchar_code|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(-1);
 if ($2) {
  $$0 = 0;
  return ($$0|0);
 }
 $3 = ((($cmap)) + 24|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)==(0);
 if (!($5)) {
  $6 = HEAP32[$pchar_code>>2]|0;
  $7 = ($1|0)==($6|0);
  if ($7) {
   _tt_cmap12_next($cmap);
   $8 = HEAP8[$3>>0]|0;
   $9 = ($8<<24>>24)==(0);
   if ($9) {
    $$0 = 0;
    return ($$0|0);
   }
   $10 = ((($cmap)) + 32|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = ($11|0)==(0);
   if ($12) {
    $$0 = 0;
    return ($$0|0);
   }
   $13 = HEAP32[$0>>2]|0;
   HEAP32[$pchar_code>>2] = $13;
   $$0 = $11;
   return ($$0|0);
  }
 }
 $14 = (_tt_cmap12_char_map_binary($cmap,$pchar_code,1)|0);
 $$0 = $14;
 return ($$0|0);
}
function _tt_cmap12_validate($table,$valid) {
 $table = $table|0;
 $valid = $valid|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $last$06 = 0, $n$05 = 0;
 var $or$cond = 0, $or$cond2 = 0, $or$cond4 = 0, $p$07 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($table)) + 16|0);
 $1 = ((($valid)) + 160|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($0>>>0)>($2>>>0);
 if ($3) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $4 = ((($table)) + 4|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = $6 << 24;
 $8 = ((($table)) + 5|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 $11 = $10 << 16;
 $12 = $11 | $7;
 $13 = ((($table)) + 6|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = $14&255;
 $16 = $15 << 8;
 $17 = $12 | $16;
 $18 = ((($table)) + 7|0);
 $19 = HEAP8[$18>>0]|0;
 $20 = $19&255;
 $21 = $17 | $20;
 $22 = ((($table)) + 12|0);
 $23 = HEAP8[$22>>0]|0;
 $24 = $23&255;
 $25 = $24 << 24;
 $26 = ((($table)) + 13|0);
 $27 = HEAP8[$26>>0]|0;
 $28 = $27&255;
 $29 = $28 << 16;
 $30 = $29 | $25;
 $31 = ((($table)) + 14|0);
 $32 = HEAP8[$31>>0]|0;
 $33 = $32&255;
 $34 = $33 << 8;
 $35 = $30 | $34;
 $36 = ((($table)) + 15|0);
 $37 = HEAP8[$36>>0]|0;
 $38 = $37&255;
 $39 = $35 | $38;
 $40 = HEAP32[$1>>2]|0;
 $41 = $40;
 $42 = $table;
 $43 = (($41) - ($42))|0;
 $44 = ($21>>>0)>($43>>>0);
 $45 = ($21>>>0)<(16);
 $or$cond = $45 | $44;
 if ($or$cond) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $46 = (($21) + -16)|0;
 $47 = (($46>>>0) / 12)&-1;
 $48 = ($47>>>0)<($39>>>0);
 if ($48) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $49 = ($39|0)==(0);
 if ($49) {
  return 0;
 }
 $50 = ((($valid)) + 164|0);
 $51 = ((($valid)) + 172|0);
 $last$06 = 0;$n$05 = 0;$p$07 = $0;
 while(1) {
  $52 = HEAP8[$p$07>>0]|0;
  $53 = $52&255;
  $54 = $53 << 24;
  $55 = ((($p$07)) + 1|0);
  $56 = HEAP8[$55>>0]|0;
  $57 = $56&255;
  $58 = $57 << 16;
  $59 = $58 | $54;
  $60 = ((($p$07)) + 2|0);
  $61 = HEAP8[$60>>0]|0;
  $62 = $61&255;
  $63 = $62 << 8;
  $64 = $59 | $63;
  $65 = ((($p$07)) + 3|0);
  $66 = HEAP8[$65>>0]|0;
  $67 = $66&255;
  $68 = $64 | $67;
  $69 = ((($p$07)) + 4|0);
  $70 = HEAP8[$69>>0]|0;
  $71 = $70&255;
  $72 = $71 << 24;
  $73 = ((($p$07)) + 5|0);
  $74 = HEAP8[$73>>0]|0;
  $75 = $74&255;
  $76 = $75 << 16;
  $77 = $76 | $72;
  $78 = ((($p$07)) + 6|0);
  $79 = HEAP8[$78>>0]|0;
  $80 = $79&255;
  $81 = $80 << 8;
  $82 = $77 | $81;
  $83 = ((($p$07)) + 7|0);
  $84 = HEAP8[$83>>0]|0;
  $85 = $84&255;
  $86 = $82 | $85;
  $87 = ((($p$07)) + 12|0);
  $88 = ((($p$07)) + 8|0);
  $89 = HEAP8[$88>>0]|0;
  $90 = $89&255;
  $91 = $90 << 24;
  $92 = ((($p$07)) + 9|0);
  $93 = HEAP8[$92>>0]|0;
  $94 = $93&255;
  $95 = $94 << 16;
  $96 = $95 | $91;
  $97 = ((($p$07)) + 10|0);
  $98 = HEAP8[$97>>0]|0;
  $99 = $98&255;
  $100 = $99 << 8;
  $101 = $96 | $100;
  $102 = ((($p$07)) + 11|0);
  $103 = HEAP8[$102>>0]|0;
  $104 = $103&255;
  $105 = $101 | $104;
  $106 = ($86>>>0)<($68>>>0);
  if ($106) {
   label = 9;
   break;
  }
  $107 = ($n$05|0)==(0);
  $108 = ($68>>>0)>($last$06>>>0);
  $or$cond2 = $107 | $108;
  if (!($or$cond2)) {
   label = 11;
   break;
  }
  $109 = HEAP32[$50>>2]|0;
  $110 = ($109|0)==(0);
  if (!($110)) {
   $111 = (($86) - ($68))|0;
   $112 = HEAP32[$51>>2]|0;
   $113 = ($112>>>0)>=($111>>>0);
   $114 = (($112) - ($111))|0;
   $115 = ($105>>>0)<($114>>>0);
   $or$cond4 = $113 & $115;
   if (!($or$cond4)) {
    label = 14;
    break;
   }
  }
  $116 = (($n$05) + 1)|0;
  $exitcond = ($116|0)==($39|0);
  if ($exitcond) {
   label = 16;
   break;
  } else {
   $last$06 = $86;$n$05 = $116;$p$07 = $87;
  }
 }
 if ((label|0) == 9) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 else if ((label|0) == 11) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 else if ((label|0) == 14) {
  _ft_validator_error($valid,16);
  // unreachable;
 }
 else if ((label|0) == 16) {
  return 0;
 }
 return (0)|0;
}
function _tt_cmap12_get_info($cmap,$cmap_info) {
 $cmap = $cmap|0;
 $cmap_info = $cmap_info|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8|0);
 $3 = ((($cmap_info)) + 4|0);
 HEAP32[$3>>2] = 12;
 $4 = HEAP8[$2>>0]|0;
 $5 = $4&255;
 $6 = $5 << 24;
 $7 = ((($1)) + 9|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = $8&255;
 $10 = $9 << 16;
 $11 = $10 | $6;
 $12 = ((($1)) + 10|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = $13&255;
 $15 = $14 << 8;
 $16 = $11 | $15;
 $17 = ((($1)) + 11|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = $18&255;
 $20 = $16 | $19;
 HEAP32[$cmap_info>>2] = $20;
 return 0;
}
function _tt_cmap13_init($cmap,$table) {
 $cmap = $cmap|0;
 $table = $table|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 HEAP32[$0>>2] = $table;
 $1 = ((($table)) + 12|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = $2&255;
 $4 = $3 << 24;
 $5 = ((($table)) + 13|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = $6&255;
 $8 = $7 << 16;
 $9 = $8 | $4;
 $10 = ((($table)) + 14|0);
 $11 = HEAP8[$10>>0]|0;
 $12 = $11&255;
 $13 = $12 << 8;
 $14 = $9 | $13;
 $15 = ((($table)) + 15|0);
 $16 = HEAP8[$15>>0]|0;
 $17 = $16&255;
 $18 = $14 | $17;
 $19 = ((($cmap)) + 40|0);
 HEAP32[$19>>2] = $18;
 $20 = ((($cmap)) + 24|0);
 HEAP8[$20>>0] = 0;
 return 0;
}
function _tt_cmap13_char_index($cmap,$char_code) {
 $cmap = $cmap|0;
 $char_code = $char_code|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 HEAP32[$0>>2] = $char_code;
 $1 = (_tt_cmap13_char_map_binary($cmap,$0,0)|0);
 STACKTOP = sp;return ($1|0);
}
function _tt_cmap13_char_map_binary($cmap,$pchar_code,$next) {
 $cmap = $cmap|0;
 $pchar_code = $pchar_code|0;
 $next = $next|0;
 var $$ = 0, $$1 = 0, $$lcssa = 0, $$lcssa70 = 0, $$lcssa73 = 0, $$lcssa75 = 0, $$lcssa76 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $gindex$0 = 0, $gindex$13 = 0, $max$0$ph20 = 0, $mid$2 = 0, $min$0$ph21 = 0, $min$016 = 0, $min$016$lcssa = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 12|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = $4 << 24;
 $6 = ((($1)) + 13|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $9 = $8 << 16;
 $10 = $9 | $5;
 $11 = ((($1)) + 14|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = $12&255;
 $14 = $13 << 8;
 $15 = $10 | $14;
 $16 = ((($1)) + 15|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = $17&255;
 $19 = $15 | $18;
 $20 = ($19|0)==(0);
 if ($20) {
  $$1 = 0;
  return ($$1|0);
 }
 $21 = HEAP32[$pchar_code>>2]|0;
 $22 = ($next<<24>>24)!=(0);
 $23 = $22&1;
 $$ = (($21) + ($23))|0;
 $24 = ((($1)) + 16|0);
 $max$0$ph20 = $19;$min$0$ph21 = 0;
 L4: while(1) {
  $min$016 = $min$0$ph21;
  while(1) {
   $26 = (($min$016) + ($max$0$ph20))|0;
   $27 = $26 >>> 1;
   $28 = ($27*12)|0;
   $29 = (($24) + ($28)|0);
   $30 = ((($29)) + 4|0);
   $31 = ((($30)) + -4|0);
   $32 = HEAP8[$31>>0]|0;
   $33 = $32&255;
   $34 = $33 << 24;
   $35 = ((($31)) + 1|0);
   $36 = HEAP8[$35>>0]|0;
   $37 = $36&255;
   $38 = $37 << 16;
   $39 = $38 | $34;
   $40 = ((($31)) + 2|0);
   $41 = HEAP8[$40>>0]|0;
   $42 = $41&255;
   $43 = $42 << 8;
   $44 = $39 | $43;
   $45 = ((($31)) + 3|0);
   $46 = HEAP8[$45>>0]|0;
   $47 = $46&255;
   $48 = $44 | $47;
   $49 = ((($30)) + 4|0);
   $50 = ((($49)) + -4|0);
   $51 = HEAP8[$50>>0]|0;
   $52 = $51&255;
   $53 = $52 << 24;
   $54 = ((($50)) + 1|0);
   $55 = HEAP8[$54>>0]|0;
   $56 = $55&255;
   $57 = $56 << 16;
   $58 = $57 | $53;
   $59 = ((($50)) + 2|0);
   $60 = HEAP8[$59>>0]|0;
   $61 = $60&255;
   $62 = $61 << 8;
   $63 = $58 | $62;
   $64 = ((($50)) + 3|0);
   $65 = HEAP8[$64>>0]|0;
   $66 = $65&255;
   $67 = $63 | $66;
   $68 = ($$>>>0)<($48>>>0);
   if ($68) {
    $$lcssa = $27;$$lcssa75 = $67;$min$016$lcssa = $min$016;
    break;
   }
   $69 = ($$>>>0)>($67>>>0);
   if (!($69)) {
    $$lcssa70 = $27;$$lcssa73 = $49;$$lcssa76 = $67;
    label = 8;
    break L4;
   }
   $70 = (($27) + 1)|0;
   $71 = ($70>>>0)<($max$0$ph20>>>0);
   if ($71) {
    $min$016 = $70;
   } else {
    $90 = $67;$92 = $27;$gindex$0 = 0;
    break L4;
   }
  }
  $25 = ($min$016$lcssa>>>0)<($$lcssa>>>0);
  if ($25) {
   $max$0$ph20 = $$lcssa;$min$0$ph21 = $min$016$lcssa;
  } else {
   $90 = $$lcssa75;$92 = $$lcssa;$gindex$0 = 0;
   break;
  }
 }
 if ((label|0) == 8) {
  $72 = HEAP8[$$lcssa73>>0]|0;
  $73 = $72&255;
  $74 = $73 << 24;
  $75 = ((($$lcssa73)) + 1|0);
  $76 = HEAP8[$75>>0]|0;
  $77 = $76&255;
  $78 = $77 << 16;
  $79 = $78 | $74;
  $80 = ((($$lcssa73)) + 2|0);
  $81 = HEAP8[$80>>0]|0;
  $82 = $81&255;
  $83 = $82 << 8;
  $84 = $79 | $83;
  $85 = ((($$lcssa73)) + 3|0);
  $86 = HEAP8[$85>>0]|0;
  $87 = $86&255;
  $88 = $84 | $87;
  $90 = $$lcssa76;$92 = $$lcssa70;$gindex$0 = $88;
 }
 if (!($22)) {
  $$1 = $gindex$0;
  return ($$1|0);
 }
 $89 = ($$>>>0)>($90>>>0);
 if ($89) {
  $91 = (($92) + 1)|0;
  $93 = ($91|0)==($19|0);
  if ($93) {
   $$1 = 0;
   return ($$1|0);
  } else {
   $mid$2 = $91;
  }
 } else {
  $mid$2 = $92;
 }
 $94 = ((($cmap)) + 24|0);
 HEAP8[$94>>0] = 1;
 $95 = ((($cmap)) + 28|0);
 HEAP32[$95>>2] = $$;
 $96 = ((($cmap)) + 36|0);
 HEAP32[$96>>2] = $mid$2;
 $97 = ($gindex$0|0)==(0);
 do {
  if ($97) {
   _tt_cmap13_next($cmap);
   $98 = HEAP8[$94>>0]|0;
   $99 = ($98<<24>>24)==(0);
   if ($99) {
    $$1 = 0;
    return ($$1|0);
   }
   $101 = ((($cmap)) + 32|0);
   $102 = HEAP32[$101>>2]|0;
   $103 = ($102|0)==(0);
   if ($103) {
    $$1 = 0;
    return ($$1|0);
   } else {
    $$pre = HEAP32[$95>>2]|0;
    $104 = $$pre;$gindex$13 = $102;
    break;
   }
  } else {
   $100 = ((($cmap)) + 32|0);
   HEAP32[$100>>2] = $gindex$0;
   $104 = $$;$gindex$13 = $gindex$0;
  }
 } while(0);
 HEAP32[$pchar_code>>2] = $104;
 $$1 = $gindex$13;
 return ($$1|0);
}
function _tt_cmap13_next($cmap) {
 $cmap = $cmap|0;
 var $$char_code$0 = 0, $$char_code$0$lcssa = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $8 = 0, $9 = 0, $char_code$03 = 0;
 var $n$02 = 0, $n$02$lcssa = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(-1);
 L1: do {
  if (!($2)) {
   $3 = ((($cmap)) + 36|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ((($cmap)) + 40|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($4>>>0)<($6>>>0);
   if ($7) {
    $8 = (($1) + 1)|0;
    $9 = ((($cmap)) + 16|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = ((($10)) + 16|0);
    $char_code$03 = $8;$n$02 = $4;
    while(1) {
     $12 = ($n$02*12)|0;
     $13 = (($11) + ($12)|0);
     $14 = ((($13)) + 4|0);
     $15 = ((($14)) + -4|0);
     $16 = HEAP8[$15>>0]|0;
     $17 = $16&255;
     $18 = $17 << 24;
     $19 = ((($15)) + 1|0);
     $20 = HEAP8[$19>>0]|0;
     $21 = $20&255;
     $22 = $21 << 16;
     $23 = $22 | $18;
     $24 = ((($15)) + 2|0);
     $25 = HEAP8[$24>>0]|0;
     $26 = $25&255;
     $27 = $26 << 8;
     $28 = $23 | $27;
     $29 = ((($15)) + 3|0);
     $30 = HEAP8[$29>>0]|0;
     $31 = $30&255;
     $32 = $28 | $31;
     $33 = ((($14)) + 4|0);
     $34 = ((($33)) + -4|0);
     $35 = HEAP8[$34>>0]|0;
     $36 = $35&255;
     $37 = $36 << 24;
     $38 = ((($34)) + 1|0);
     $39 = HEAP8[$38>>0]|0;
     $40 = $39&255;
     $41 = $40 << 16;
     $42 = $41 | $37;
     $43 = ((($34)) + 2|0);
     $44 = HEAP8[$43>>0]|0;
     $45 = $44&255;
     $46 = $45 << 8;
     $47 = $42 | $46;
     $48 = ((($34)) + 3|0);
     $49 = HEAP8[$48>>0]|0;
     $50 = $49&255;
     $51 = $47 | $50;
     $52 = HEAP8[$33>>0]|0;
     $53 = $52&255;
     $54 = $53 << 24;
     $55 = ((($33)) + 1|0);
     $56 = HEAP8[$55>>0]|0;
     $57 = $56&255;
     $58 = $57 << 16;
     $59 = $58 | $54;
     $60 = ((($33)) + 2|0);
     $61 = HEAP8[$60>>0]|0;
     $62 = $61&255;
     $63 = $62 << 8;
     $64 = $59 | $63;
     $65 = ((($33)) + 3|0);
     $66 = HEAP8[$65>>0]|0;
     $67 = $66&255;
     $68 = $64 | $67;
     $69 = ($char_code$03>>>0)<($32>>>0);
     $$char_code$0 = $69 ? $32 : $char_code$03;
     $70 = ($$char_code$0>>>0)>($51>>>0);
     $71 = ($68|0)==(0);
     $or$cond = $70 | $71;
     if (!($or$cond)) {
      $$char_code$0$lcssa = $$char_code$0;$$lcssa = $68;$n$02$lcssa = $n$02;
      break;
     }
     $73 = (($n$02) + 1)|0;
     $74 = ($73>>>0)<($6>>>0);
     if ($74) {
      $char_code$03 = $$char_code$0;$n$02 = $73;
     } else {
      break L1;
     }
    }
    HEAP32[$0>>2] = $$char_code$0$lcssa;
    $72 = ((($cmap)) + 32|0);
    HEAP32[$72>>2] = $$lcssa;
    HEAP32[$3>>2] = $n$02$lcssa;
    return;
   }
  }
 } while(0);
 $75 = ((($cmap)) + 24|0);
 HEAP8[$75>>0] = 0;
 return;
}
function _tt_cmap13_char_next($cmap,$pchar_code) {
 $cmap = $cmap|0;
 $pchar_code = $pchar_code|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(-1);
 if ($2) {
  $$0 = 0;
  return ($$0|0);
 }
 $3 = ((($cmap)) + 24|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)==(0);
 if (!($5)) {
  $6 = HEAP32[$pchar_code>>2]|0;
  $7 = ($1|0)==($6|0);
  if ($7) {
   _tt_cmap13_next($cmap);
   $8 = HEAP8[$3>>0]|0;
   $9 = ($8<<24>>24)==(0);
   if ($9) {
    $$0 = 0;
    return ($$0|0);
   }
   $10 = ((($cmap)) + 32|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = ($11|0)==(0);
   if ($12) {
    $$0 = 0;
    return ($$0|0);
   }
   $13 = HEAP32[$0>>2]|0;
   HEAP32[$pchar_code>>2] = $13;
   $$0 = $11;
   return ($$0|0);
  }
 }
 $14 = (_tt_cmap13_char_map_binary($cmap,$pchar_code,1)|0);
 $$0 = $14;
 return ($$0|0);
}
function _tt_cmap13_validate($table,$valid) {
 $table = $table|0;
 $valid = $valid|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $last$05 = 0, $n$03 = 0, $or$cond = 0, $or$cond2 = 0, $p$04 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($table)) + 16|0);
 $1 = ((($valid)) + 160|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($0>>>0)>($2>>>0);
 if ($3) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $4 = ((($table)) + 4|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = $6 << 24;
 $8 = ((($table)) + 5|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 $11 = $10 << 16;
 $12 = $11 | $7;
 $13 = ((($table)) + 6|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = $14&255;
 $16 = $15 << 8;
 $17 = $12 | $16;
 $18 = ((($table)) + 7|0);
 $19 = HEAP8[$18>>0]|0;
 $20 = $19&255;
 $21 = $17 | $20;
 $22 = ((($table)) + 12|0);
 $23 = HEAP8[$22>>0]|0;
 $24 = $23&255;
 $25 = $24 << 24;
 $26 = ((($table)) + 13|0);
 $27 = HEAP8[$26>>0]|0;
 $28 = $27&255;
 $29 = $28 << 16;
 $30 = $29 | $25;
 $31 = ((($table)) + 14|0);
 $32 = HEAP8[$31>>0]|0;
 $33 = $32&255;
 $34 = $33 << 8;
 $35 = $30 | $34;
 $36 = ((($table)) + 15|0);
 $37 = HEAP8[$36>>0]|0;
 $38 = $37&255;
 $39 = $35 | $38;
 $40 = HEAP32[$1>>2]|0;
 $41 = $40;
 $42 = $table;
 $43 = (($41) - ($42))|0;
 $44 = ($21>>>0)>($43>>>0);
 $45 = ($21>>>0)<(16);
 $or$cond = $45 | $44;
 if ($or$cond) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $46 = (($21) + -16)|0;
 $47 = (($46>>>0) / 12)&-1;
 $48 = ($47>>>0)<($39>>>0);
 if ($48) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $49 = ($39|0)==(0);
 if ($49) {
  return 0;
 }
 $50 = ((($valid)) + 164|0);
 $51 = ((($valid)) + 172|0);
 $last$05 = 0;$n$03 = 0;$p$04 = $0;
 while(1) {
  $52 = HEAP8[$p$04>>0]|0;
  $53 = $52&255;
  $54 = $53 << 24;
  $55 = ((($p$04)) + 1|0);
  $56 = HEAP8[$55>>0]|0;
  $57 = $56&255;
  $58 = $57 << 16;
  $59 = $58 | $54;
  $60 = ((($p$04)) + 2|0);
  $61 = HEAP8[$60>>0]|0;
  $62 = $61&255;
  $63 = $62 << 8;
  $64 = $59 | $63;
  $65 = ((($p$04)) + 3|0);
  $66 = HEAP8[$65>>0]|0;
  $67 = $66&255;
  $68 = $64 | $67;
  $69 = ((($p$04)) + 4|0);
  $70 = HEAP8[$69>>0]|0;
  $71 = $70&255;
  $72 = $71 << 24;
  $73 = ((($p$04)) + 5|0);
  $74 = HEAP8[$73>>0]|0;
  $75 = $74&255;
  $76 = $75 << 16;
  $77 = $76 | $72;
  $78 = ((($p$04)) + 6|0);
  $79 = HEAP8[$78>>0]|0;
  $80 = $79&255;
  $81 = $80 << 8;
  $82 = $77 | $81;
  $83 = ((($p$04)) + 7|0);
  $84 = HEAP8[$83>>0]|0;
  $85 = $84&255;
  $86 = $82 | $85;
  $87 = ((($p$04)) + 12|0);
  $88 = ((($p$04)) + 8|0);
  $89 = HEAP8[$88>>0]|0;
  $90 = $89&255;
  $91 = $90 << 24;
  $92 = ((($p$04)) + 9|0);
  $93 = HEAP8[$92>>0]|0;
  $94 = $93&255;
  $95 = $94 << 16;
  $96 = $95 | $91;
  $97 = ((($p$04)) + 10|0);
  $98 = HEAP8[$97>>0]|0;
  $99 = $98&255;
  $100 = $99 << 8;
  $101 = $96 | $100;
  $102 = ((($p$04)) + 11|0);
  $103 = HEAP8[$102>>0]|0;
  $104 = $103&255;
  $105 = $101 | $104;
  $106 = ($68>>>0)>($86>>>0);
  if ($106) {
   label = 9;
   break;
  }
  $107 = ($n$03|0)==(0);
  $108 = ($68>>>0)>($last$05>>>0);
  $or$cond2 = $107 | $108;
  if (!($or$cond2)) {
   label = 11;
   break;
  }
  $109 = HEAP32[$50>>2]|0;
  $110 = ($109|0)==(0);
  if (!($110)) {
   $111 = HEAP32[$51>>2]|0;
   $112 = ($105>>>0)<($111>>>0);
   if (!($112)) {
    label = 14;
    break;
   }
  }
  $113 = (($n$03) + 1)|0;
  $exitcond = ($113|0)==($39|0);
  if ($exitcond) {
   label = 16;
   break;
  } else {
   $last$05 = $86;$n$03 = $113;$p$04 = $87;
  }
 }
 if ((label|0) == 9) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 else if ((label|0) == 11) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 else if ((label|0) == 14) {
  _ft_validator_error($valid,16);
  // unreachable;
 }
 else if ((label|0) == 16) {
  return 0;
 }
 return (0)|0;
}
function _tt_cmap13_get_info($cmap,$cmap_info) {
 $cmap = $cmap|0;
 $cmap_info = $cmap_info|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8|0);
 $3 = ((($cmap_info)) + 4|0);
 HEAP32[$3>>2] = 13;
 $4 = HEAP8[$2>>0]|0;
 $5 = $4&255;
 $6 = $5 << 24;
 $7 = ((($1)) + 9|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = $8&255;
 $10 = $9 << 16;
 $11 = $10 | $6;
 $12 = ((($1)) + 10|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = $13&255;
 $15 = $14 << 8;
 $16 = $11 | $15;
 $17 = ((($1)) + 11|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = $18&255;
 $20 = $16 | $19;
 HEAP32[$cmap_info>>2] = $20;
 return 0;
}
function _tt_cmap14_init($cmap,$table) {
 $cmap = $cmap|0;
 $table = $table|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 HEAP32[$0>>2] = $table;
 $1 = ((($table)) + 6|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = $2&255;
 $4 = $3 << 24;
 $5 = ((($table)) + 7|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = $6&255;
 $8 = $7 << 16;
 $9 = $8 | $4;
 $10 = ((($table)) + 8|0);
 $11 = HEAP8[$10>>0]|0;
 $12 = $11&255;
 $13 = $12 << 8;
 $14 = $9 | $13;
 $15 = ((($table)) + 9|0);
 $16 = HEAP8[$15>>0]|0;
 $17 = $16&255;
 $18 = $14 | $17;
 $19 = ((($cmap)) + 24|0);
 HEAP32[$19>>2] = $18;
 $20 = ((($cmap)) + 28|0);
 HEAP32[$20>>2] = 0;
 $21 = ((($cmap)) + 32|0);
 HEAP32[$21>>2] = 0;
 return 0;
}
function _tt_cmap14_done($cmap) {
 $cmap = $cmap|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 36|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($cmap)) + 28|0);
 HEAP32[$2>>2] = 0;
 $3 = ($1|0)==(0|0);
 if ($3) {
  return;
 }
 $4 = ((($cmap)) + 32|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if ($6) {
  return;
 }
 _ft_mem_free($1,$5);
 HEAP32[$4>>2] = 0;
 return;
}
function _tt_cmap14_char_index($cmap,$char_code) {
 $cmap = $cmap|0;
 $char_code = $char_code|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _tt_cmap14_char_next($cmap,$pchar_code) {
 $cmap = $cmap|0;
 $pchar_code = $pchar_code|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$pchar_code>>2] = 0;
 return 0;
}
function _tt_cmap14_char_var_index($cmap,$ucmap,$charcode,$variantSelector) {
 $cmap = $cmap|0;
 $ucmap = $ucmap|0;
 $charcode = $charcode|0;
 $variantSelector = $variantSelector|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa110 = 0, $$lcssa113 = 0, $$lcssa114 = 0, $$lcssa115 = 0, $$lcssa121 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $max$0$i = 0, $max$0$i$lcssa112 = 0, $max$0$ph$i = 0, $max$0$ph$i4 = 0, $max$0$ph8$i = 0, $min$0$i = 0, $min$0$ph$i = 0, $min$0$ph$i3 = 0, $min$0$ph9$i = 0, $min$07$i = 0, $min$07$i$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 6|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = $4 << 24;
 $6 = ((($1)) + 7|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $9 = $8 << 16;
 $10 = $9 | $5;
 $11 = ((($1)) + 8|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = $12&255;
 $14 = $13 << 8;
 $15 = $10 | $14;
 $16 = ((($1)) + 9|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = $17&255;
 $19 = $15 | $18;
 $20 = ((($1)) + 10|0);
 $max$0$ph$i = $19;$min$0$ph$i = 0;
 L1: while(1) {
  $min$0$i = $min$0$ph$i;
  while(1) {
   $21 = ($min$0$i>>>0)<($max$0$ph$i>>>0);
   if (!($21)) {
    $$0 = 0;
    label = 21;
    break L1;
   }
   $22 = (($min$0$i) + ($max$0$ph$i))|0;
   $23 = $22 >>> 1;
   $24 = ($23*11)|0;
   $25 = (($20) + ($24)|0);
   $26 = ((($25)) + 3|0);
   $27 = ((($26)) + -3|0);
   $28 = HEAP8[$27>>0]|0;
   $29 = $28&255;
   $30 = $29 << 16;
   $31 = ((($27)) + 1|0);
   $32 = HEAP8[$31>>0]|0;
   $33 = $32&255;
   $34 = $33 << 8;
   $35 = $34 | $30;
   $36 = ((($27)) + 2|0);
   $37 = HEAP8[$36>>0]|0;
   $38 = $37&255;
   $39 = $35 | $38;
   $40 = ($39>>>0)>($variantSelector>>>0);
   if ($40) {
    $max$0$ph$i = $23;$min$0$ph$i = $min$0$i;
    continue L1;
   }
   $41 = ($39>>>0)<($variantSelector>>>0);
   $42 = (($23) + 1)|0;
   if ($41) {
    $min$0$i = $42;
   } else {
    $$lcssa121 = $26;
    break L1;
   }
  }
 }
 if ((label|0) == 21) {
  return ($$0|0);
 }
 $43 = ((($$lcssa121)) + 4|0);
 $44 = HEAP8[$$lcssa121>>0]|0;
 $45 = $44&255;
 $46 = $45 << 24;
 $47 = ((($$lcssa121)) + 1|0);
 $48 = HEAP8[$47>>0]|0;
 $49 = $48&255;
 $50 = $49 << 16;
 $51 = $50 | $46;
 $52 = ((($$lcssa121)) + 2|0);
 $53 = HEAP8[$52>>0]|0;
 $54 = $53&255;
 $55 = $54 << 8;
 $56 = $51 | $55;
 $57 = ((($$lcssa121)) + 3|0);
 $58 = HEAP8[$57>>0]|0;
 $59 = $58&255;
 $60 = $56 | $59;
 $61 = HEAP8[$43>>0]|0;
 $62 = $61&255;
 $63 = $62 << 24;
 $64 = ((($$lcssa121)) + 5|0);
 $65 = HEAP8[$64>>0]|0;
 $66 = $65&255;
 $67 = $66 << 16;
 $68 = $67 | $63;
 $69 = ((($$lcssa121)) + 6|0);
 $70 = HEAP8[$69>>0]|0;
 $71 = $70&255;
 $72 = $71 << 8;
 $73 = $68 | $72;
 $74 = ((($$lcssa121)) + 7|0);
 $75 = HEAP8[$74>>0]|0;
 $76 = $75&255;
 $77 = $73 | $76;
 $78 = ($60|0)==(0);
 L10: do {
  if (!($78)) {
   $79 = (($1) + ($60)|0);
   $80 = HEAP8[$79>>0]|0;
   $81 = $80&255;
   $82 = $81 << 24;
   $83 = ((($79)) + 1|0);
   $84 = HEAP8[$83>>0]|0;
   $85 = $84&255;
   $86 = $85 << 16;
   $87 = $86 | $82;
   $88 = ((($79)) + 2|0);
   $89 = HEAP8[$88>>0]|0;
   $90 = $89&255;
   $91 = $90 << 8;
   $92 = $87 | $91;
   $93 = ((($79)) + 3|0);
   $94 = HEAP8[$93>>0]|0;
   $95 = $94&255;
   $96 = $92 | $95;
   $97 = ((($79)) + 4|0);
   $max$0$ph$i4 = $96;$min$0$ph$i3 = 0;
   while(1) {
    $max$0$i = $max$0$ph$i4;
    while(1) {
     $98 = ($min$0$ph$i3>>>0)<($max$0$i>>>0);
     if (!($98)) {
      break L10;
     }
     $99 = (($max$0$i) + ($min$0$ph$i3))|0;
     $100 = $99 >>> 1;
     $101 = $100 << 2;
     $102 = (($97) + ($101)|0);
     $103 = ((($102)) + 3|0);
     $104 = ((($103)) + -3|0);
     $105 = HEAP8[$104>>0]|0;
     $106 = $105&255;
     $107 = $106 << 16;
     $108 = ((($104)) + 1|0);
     $109 = HEAP8[$108>>0]|0;
     $110 = $109&255;
     $111 = $110 << 8;
     $112 = $111 | $107;
     $113 = ((($104)) + 2|0);
     $114 = HEAP8[$113>>0]|0;
     $115 = $114&255;
     $116 = $112 | $115;
     $117 = ($116>>>0)>($charcode>>>0);
     if ($117) {
      $max$0$i = $100;
     } else {
      $$lcssa113 = $100;$$lcssa114 = $103;$$lcssa115 = $116;$max$0$i$lcssa112 = $max$0$i;
      break;
     }
    }
    $118 = HEAP8[$$lcssa114>>0]|0;
    $119 = $118&255;
    $120 = (($119) + ($$lcssa115))|0;
    $121 = ($120>>>0)<($charcode>>>0);
    $122 = (($$lcssa113) + 1)|0;
    if ($121) {
     $max$0$ph$i4 = $max$0$i$lcssa112;$min$0$ph$i3 = $122;
    } else {
     break;
    }
   }
   $123 = ((($ucmap)) + 12|0);
   $124 = HEAP32[$123>>2]|0;
   $125 = ((($124)) + 12|0);
   $126 = HEAP32[$125>>2]|0;
   $127 = (FUNCTION_TABLE_iii[$126 & 8191]($ucmap,$charcode)|0);
   $$0 = $127;
   return ($$0|0);
  }
 } while(0);
 $128 = ($77|0)==(0);
 if ($128) {
  $$0 = 0;
  return ($$0|0);
 }
 $129 = (($1) + ($77)|0);
 $130 = HEAP8[$129>>0]|0;
 $131 = $130&255;
 $132 = $131 << 24;
 $133 = ((($129)) + 1|0);
 $134 = HEAP8[$133>>0]|0;
 $135 = $134&255;
 $136 = $135 << 16;
 $137 = $136 | $132;
 $138 = ((($129)) + 2|0);
 $139 = HEAP8[$138>>0]|0;
 $140 = $139&255;
 $141 = $140 << 8;
 $142 = $137 | $141;
 $143 = ((($129)) + 3|0);
 $144 = HEAP8[$143>>0]|0;
 $145 = $144&255;
 $146 = $142 | $145;
 $147 = ((($129)) + 4|0);
 $148 = ($146|0)==(0);
 if ($148) {
  $$0 = 0;
  return ($$0|0);
 } else {
  $max$0$ph8$i = $146;$min$0$ph9$i = 0;
 }
 L26: while(1) {
  $min$07$i = $min$0$ph9$i;
  while(1) {
   $150 = (($min$07$i) + ($max$0$ph8$i))|0;
   $151 = $150 >>> 1;
   $152 = ($151*5)|0;
   $153 = (($147) + ($152)|0);
   $154 = ((($153)) + 3|0);
   $155 = ((($154)) + -3|0);
   $156 = HEAP8[$155>>0]|0;
   $157 = $156&255;
   $158 = $157 << 16;
   $159 = ((($155)) + 1|0);
   $160 = HEAP8[$159>>0]|0;
   $161 = $160&255;
   $162 = $161 << 8;
   $163 = $162 | $158;
   $164 = ((($155)) + 2|0);
   $165 = HEAP8[$164>>0]|0;
   $166 = $165&255;
   $167 = $163 | $166;
   $168 = ($167>>>0)>($charcode>>>0);
   if ($168) {
    $$lcssa = $151;$min$07$i$lcssa = $min$07$i;
    break;
   }
   $169 = ($167>>>0)<($charcode>>>0);
   if (!($169)) {
    $$lcssa110 = $154;
    label = 20;
    break L26;
   }
   $170 = (($151) + 1)|0;
   $171 = ($170>>>0)<($max$0$ph8$i>>>0);
   if ($171) {
    $min$07$i = $170;
   } else {
    $$0 = 0;
    label = 21;
    break L26;
   }
  }
  $149 = ($min$07$i$lcssa>>>0)<($$lcssa>>>0);
  if ($149) {
   $max$0$ph8$i = $$lcssa;$min$0$ph9$i = $min$07$i$lcssa;
  } else {
   $$0 = 0;
   label = 21;
   break;
  }
 }
 if ((label|0) == 20) {
  $172 = HEAP8[$$lcssa110>>0]|0;
  $173 = $172&255;
  $174 = $173 << 8;
  $175 = ((($$lcssa110)) + 1|0);
  $176 = HEAP8[$175>>0]|0;
  $177 = $176&255;
  $178 = $174 | $177;
  $$0 = $178;
  return ($$0|0);
 }
 else if ((label|0) == 21) {
  return ($$0|0);
 }
 return (0)|0;
}
function _tt_cmap14_char_var_isdefault($cmap,$charcode,$variantSelector) {
 $cmap = $cmap|0;
 $charcode = $charcode|0;
 $variantSelector = $variantSelector|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa110 = 0, $$lcssa113 = 0, $$lcssa114 = 0, $$lcssa115 = 0, $$lcssa121 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $max$0$i = 0, $max$0$i$lcssa112 = 0, $max$0$ph$i = 0, $max$0$ph$i4 = 0, $max$0$ph8$i = 0, $min$0$i = 0, $min$0$ph$i = 0, $min$0$ph$i3 = 0, $min$0$ph9$i = 0, $min$07$i = 0, $min$07$i$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 6|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = $4 << 24;
 $6 = ((($1)) + 7|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $9 = $8 << 16;
 $10 = $9 | $5;
 $11 = ((($1)) + 8|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = $12&255;
 $14 = $13 << 8;
 $15 = $10 | $14;
 $16 = ((($1)) + 9|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = $17&255;
 $19 = $15 | $18;
 $20 = ((($1)) + 10|0);
 $max$0$ph$i = $19;$min$0$ph$i = 0;
 L1: while(1) {
  $min$0$i = $min$0$ph$i;
  while(1) {
   $21 = ($min$0$i>>>0)<($max$0$ph$i>>>0);
   if (!($21)) {
    $$0 = -1;
    label = 21;
    break L1;
   }
   $22 = (($min$0$i) + ($max$0$ph$i))|0;
   $23 = $22 >>> 1;
   $24 = ($23*11)|0;
   $25 = (($20) + ($24)|0);
   $26 = ((($25)) + 3|0);
   $27 = ((($26)) + -3|0);
   $28 = HEAP8[$27>>0]|0;
   $29 = $28&255;
   $30 = $29 << 16;
   $31 = ((($27)) + 1|0);
   $32 = HEAP8[$31>>0]|0;
   $33 = $32&255;
   $34 = $33 << 8;
   $35 = $34 | $30;
   $36 = ((($27)) + 2|0);
   $37 = HEAP8[$36>>0]|0;
   $38 = $37&255;
   $39 = $35 | $38;
   $40 = ($39>>>0)>($variantSelector>>>0);
   if ($40) {
    $max$0$ph$i = $23;$min$0$ph$i = $min$0$i;
    continue L1;
   }
   $41 = ($39>>>0)<($variantSelector>>>0);
   $42 = (($23) + 1)|0;
   if ($41) {
    $min$0$i = $42;
   } else {
    $$lcssa121 = $26;
    break L1;
   }
  }
 }
 if ((label|0) == 21) {
  return ($$0|0);
 }
 $43 = HEAP8[$$lcssa121>>0]|0;
 $44 = $43&255;
 $45 = $44 << 24;
 $46 = ((($$lcssa121)) + 1|0);
 $47 = HEAP8[$46>>0]|0;
 $48 = $47&255;
 $49 = $48 << 16;
 $50 = $49 | $45;
 $51 = ((($$lcssa121)) + 2|0);
 $52 = HEAP8[$51>>0]|0;
 $53 = $52&255;
 $54 = $53 << 8;
 $55 = $50 | $54;
 $56 = ((($$lcssa121)) + 3|0);
 $57 = HEAP8[$56>>0]|0;
 $58 = $57&255;
 $59 = $55 | $58;
 $60 = ((($$lcssa121)) + 4|0);
 $61 = HEAP8[$60>>0]|0;
 $62 = $61&255;
 $63 = $62 << 24;
 $64 = ((($$lcssa121)) + 5|0);
 $65 = HEAP8[$64>>0]|0;
 $66 = $65&255;
 $67 = $66 << 16;
 $68 = $67 | $63;
 $69 = ((($$lcssa121)) + 6|0);
 $70 = HEAP8[$69>>0]|0;
 $71 = $70&255;
 $72 = $71 << 8;
 $73 = $68 | $72;
 $74 = ((($$lcssa121)) + 7|0);
 $75 = HEAP8[$74>>0]|0;
 $76 = $75&255;
 $77 = $73 | $76;
 $78 = ($59|0)==(0);
 L10: do {
  if (!($78)) {
   $79 = (($1) + ($59)|0);
   $80 = HEAP8[$79>>0]|0;
   $81 = $80&255;
   $82 = $81 << 24;
   $83 = ((($79)) + 1|0);
   $84 = HEAP8[$83>>0]|0;
   $85 = $84&255;
   $86 = $85 << 16;
   $87 = $86 | $82;
   $88 = ((($79)) + 2|0);
   $89 = HEAP8[$88>>0]|0;
   $90 = $89&255;
   $91 = $90 << 8;
   $92 = $87 | $91;
   $93 = ((($79)) + 3|0);
   $94 = HEAP8[$93>>0]|0;
   $95 = $94&255;
   $96 = $92 | $95;
   $97 = ((($79)) + 4|0);
   $max$0$ph$i4 = $96;$min$0$ph$i3 = 0;
   while(1) {
    $max$0$i = $max$0$ph$i4;
    while(1) {
     $98 = ($min$0$ph$i3>>>0)<($max$0$i>>>0);
     if (!($98)) {
      break L10;
     }
     $99 = (($max$0$i) + ($min$0$ph$i3))|0;
     $100 = $99 >>> 1;
     $101 = $100 << 2;
     $102 = (($97) + ($101)|0);
     $103 = ((($102)) + 3|0);
     $104 = ((($103)) + -3|0);
     $105 = HEAP8[$104>>0]|0;
     $106 = $105&255;
     $107 = $106 << 16;
     $108 = ((($104)) + 1|0);
     $109 = HEAP8[$108>>0]|0;
     $110 = $109&255;
     $111 = $110 << 8;
     $112 = $111 | $107;
     $113 = ((($104)) + 2|0);
     $114 = HEAP8[$113>>0]|0;
     $115 = $114&255;
     $116 = $112 | $115;
     $117 = ($116>>>0)>($charcode>>>0);
     if ($117) {
      $max$0$i = $100;
     } else {
      $$lcssa113 = $100;$$lcssa114 = $103;$$lcssa115 = $116;$max$0$i$lcssa112 = $max$0$i;
      break;
     }
    }
    $118 = HEAP8[$$lcssa114>>0]|0;
    $119 = $118&255;
    $120 = (($119) + ($$lcssa115))|0;
    $121 = ($120>>>0)<($charcode>>>0);
    $122 = (($$lcssa113) + 1)|0;
    if ($121) {
     $max$0$ph$i4 = $max$0$i$lcssa112;$min$0$ph$i3 = $122;
    } else {
     $$0 = 1;
     break;
    }
   }
   return ($$0|0);
  }
 } while(0);
 $123 = ($77|0)==(0);
 L20: do {
  if (!($123)) {
   $124 = (($1) + ($77)|0);
   $125 = HEAP8[$124>>0]|0;
   $126 = $125&255;
   $127 = $126 << 24;
   $128 = ((($124)) + 1|0);
   $129 = HEAP8[$128>>0]|0;
   $130 = $129&255;
   $131 = $130 << 16;
   $132 = $131 | $127;
   $133 = ((($124)) + 2|0);
   $134 = HEAP8[$133>>0]|0;
   $135 = $134&255;
   $136 = $135 << 8;
   $137 = $132 | $136;
   $138 = ((($124)) + 3|0);
   $139 = HEAP8[$138>>0]|0;
   $140 = $139&255;
   $141 = $137 | $140;
   $142 = ((($124)) + 4|0);
   $143 = ($141|0)==(0);
   if (!($143)) {
    $max$0$ph8$i = $141;$min$0$ph9$i = 0;
    L22: while(1) {
     $min$07$i = $min$0$ph9$i;
     while(1) {
      $145 = (($min$07$i) + ($max$0$ph8$i))|0;
      $146 = $145 >>> 1;
      $147 = ($146*5)|0;
      $148 = (($142) + ($147)|0);
      $149 = ((($148)) + 3|0);
      $150 = ((($149)) + -3|0);
      $151 = HEAP8[$150>>0]|0;
      $152 = $151&255;
      $153 = $152 << 16;
      $154 = ((($150)) + 1|0);
      $155 = HEAP8[$154>>0]|0;
      $156 = $155&255;
      $157 = $156 << 8;
      $158 = $157 | $153;
      $159 = ((($150)) + 2|0);
      $160 = HEAP8[$159>>0]|0;
      $161 = $160&255;
      $162 = $158 | $161;
      $163 = ($162>>>0)>($charcode>>>0);
      if ($163) {
       $$lcssa = $146;$min$07$i$lcssa = $min$07$i;
       break;
      }
      $164 = ($162>>>0)<($charcode>>>0);
      if (!($164)) {
       $$lcssa110 = $149;
       break L22;
      }
      $165 = (($146) + 1)|0;
      $166 = ($165>>>0)<($max$0$ph8$i>>>0);
      if ($166) {
       $min$07$i = $165;
      } else {
       break L20;
      }
     }
     $144 = ($min$07$i$lcssa>>>0)<($$lcssa>>>0);
     if ($144) {
      $max$0$ph8$i = $$lcssa;$min$0$ph9$i = $min$07$i$lcssa;
     } else {
      break L20;
     }
    }
    $167 = HEAP8[$$lcssa110>>0]|0;
    $168 = $167&255;
    $169 = $168 << 8;
    $170 = ((($$lcssa110)) + 1|0);
    $171 = HEAP8[$170>>0]|0;
    $172 = $171&255;
    $173 = $169 | $172;
    $174 = ($173|0)==(0);
    if (!($174)) {
     $$0 = 0;
     return ($$0|0);
    }
   }
  }
 } while(0);
 $$0 = -1;
 return ($$0|0);
}
function _tt_cmap14_variants($cmap,$memory) {
 $cmap = $cmap|0;
 $memory = $memory|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$i = 0, $exitcond = 0, $i$0$lcssa = 0, $i$02 = 0;
 var $p$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp;
 $0 = ((($cmap)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($cmap)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + 1)|0;
 $5 = ((($cmap)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 HEAP32[$error$i>>2] = 0;
 $7 = ($6>>>0)<($4>>>0);
 do {
  if ($7) {
   $8 = ((($cmap)) + 36|0);
   HEAP32[$8>>2] = $memory;
   $9 = ((($cmap)) + 32|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = (_ft_mem_realloc($memory,4,$6,$4,$10,$error$i)|0);
   HEAP32[$9>>2] = $11;
   $12 = HEAP32[$error$i>>2]|0;
   $13 = ($12|0)==(0);
   if ($13) {
    HEAP32[$5>>2] = $4;
    $29 = $11;
    break;
   }
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  } else {
   $$phi$trans$insert = ((($cmap)) + 32|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0;
   $29 = $$pre;
  }
 } while(0);
 $14 = ($1|0)==(0);
 if ($14) {
  $i$0$lcssa = 0;
 } else {
  $15 = ((($3)) + 10|0);
  $i$02 = 0;$p$01 = $15;
  while(1) {
   $16 = HEAP8[$p$01>>0]|0;
   $17 = $16&255;
   $18 = $17 << 16;
   $19 = ((($p$01)) + 1|0);
   $20 = HEAP8[$19>>0]|0;
   $21 = $20&255;
   $22 = $21 << 8;
   $23 = $22 | $18;
   $24 = ((($p$01)) + 2|0);
   $25 = HEAP8[$24>>0]|0;
   $26 = $25&255;
   $27 = $23 | $26;
   $28 = (($29) + ($i$02<<2)|0);
   HEAP32[$28>>2] = $27;
   $30 = ((($p$01)) + 11|0);
   $31 = (($i$02) + 1)|0;
   $exitcond = ($31|0)==($1|0);
   if ($exitcond) {
    $i$0$lcssa = $1;
    break;
   } else {
    $i$02 = $31;$p$01 = $30;
   }
  }
 }
 $32 = (($29) + ($i$0$lcssa<<2)|0);
 HEAP32[$32>>2] = 0;
 $$0 = $29;
 STACKTOP = sp;return ($$0|0);
}
function _tt_cmap14_char_variants($cmap,$memory,$charCode) {
 $cmap = $cmap|0;
 $memory = $memory|0;
 $charCode = $charCode|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa72 = 0, $$lcssa73 = 0, $$lcssa76 = 0, $$lcssa80 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $count$018 = 0, $error$i = 0, $max$0$i = 0, $max$0$i$lcssa71 = 0;
 var $max$0$ph$i = 0, $max$0$ph8$i = 0, $min$0$ph$i = 0, $min$0$ph9$i = 0, $min$07$i = 0, $min$07$i$lcssa = 0, $p$017 = 0, $q$0$lcssa = 0, $q$015 = 0, $q$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp;
 $0 = ((($cmap)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($cmap)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + 1)|0;
 $5 = ((($cmap)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 HEAP32[$error$i>>2] = 0;
 $7 = ($6>>>0)<($4>>>0);
 do {
  if ($7) {
   $8 = ((($cmap)) + 36|0);
   HEAP32[$8>>2] = $memory;
   $9 = ((($cmap)) + 32|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = (_ft_mem_realloc($memory,4,$6,$4,$10,$error$i)|0);
   HEAP32[$9>>2] = $11;
   $12 = HEAP32[$error$i>>2]|0;
   $13 = ($12|0)==(0);
   if ($13) {
    HEAP32[$5>>2] = $4;
    $167 = $11;
    break;
   }
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  } else {
   $$phi$trans$insert = ((($cmap)) + 32|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0;
   $167 = $$pre;
  }
 } while(0);
 $14 = ($1|0)==(0);
 if ($14) {
  $q$0$lcssa = $167;
 } else {
  $15 = ((($3)) + 10|0);
  $count$018 = $1;$p$017 = $15;$q$015 = $167;
  while(1) {
   $16 = HEAP8[$p$017>>0]|0;
   $17 = $16&255;
   $18 = $17 << 16;
   $19 = ((($p$017)) + 1|0);
   $20 = HEAP8[$19>>0]|0;
   $21 = $20&255;
   $22 = $21 << 8;
   $23 = $22 | $18;
   $24 = ((($p$017)) + 2|0);
   $25 = HEAP8[$24>>0]|0;
   $26 = $25&255;
   $27 = $23 | $26;
   $28 = ((($p$017)) + 3|0);
   $29 = HEAP8[$28>>0]|0;
   $30 = $29&255;
   $31 = $30 << 24;
   $32 = ((($p$017)) + 4|0);
   $33 = HEAP8[$32>>0]|0;
   $34 = $33&255;
   $35 = $34 << 16;
   $36 = $35 | $31;
   $37 = ((($p$017)) + 5|0);
   $38 = HEAP8[$37>>0]|0;
   $39 = $38&255;
   $40 = $39 << 8;
   $41 = $36 | $40;
   $42 = ((($p$017)) + 6|0);
   $43 = HEAP8[$42>>0]|0;
   $44 = $43&255;
   $45 = $41 | $44;
   $46 = ((($p$017)) + 11|0);
   $47 = ((($p$017)) + 7|0);
   $48 = HEAP8[$47>>0]|0;
   $49 = $48&255;
   $50 = $49 << 24;
   $51 = ((($p$017)) + 8|0);
   $52 = HEAP8[$51>>0]|0;
   $53 = $52&255;
   $54 = $53 << 16;
   $55 = $54 | $50;
   $56 = ((($p$017)) + 9|0);
   $57 = HEAP8[$56>>0]|0;
   $58 = $57&255;
   $59 = $58 << 8;
   $60 = $55 | $59;
   $61 = ((($p$017)) + 10|0);
   $62 = HEAP8[$61>>0]|0;
   $63 = $62&255;
   $64 = $60 | $63;
   $65 = ($45|0)==(0);
   L13: do {
    if ($65) {
     label = 14;
    } else {
     $66 = HEAP32[$2>>2]|0;
     $67 = (($66) + ($45)|0);
     $68 = HEAP8[$67>>0]|0;
     $69 = $68&255;
     $70 = $69 << 24;
     $71 = ((($67)) + 1|0);
     $72 = HEAP8[$71>>0]|0;
     $73 = $72&255;
     $74 = $73 << 16;
     $75 = $74 | $70;
     $76 = ((($67)) + 2|0);
     $77 = HEAP8[$76>>0]|0;
     $78 = $77&255;
     $79 = $78 << 8;
     $80 = $75 | $79;
     $81 = ((($67)) + 3|0);
     $82 = HEAP8[$81>>0]|0;
     $83 = $82&255;
     $84 = $80 | $83;
     $85 = ((($67)) + 4|0);
     $max$0$ph$i = $84;$min$0$ph$i = 0;
     while(1) {
      $max$0$i = $max$0$ph$i;
      while(1) {
       $86 = ($min$0$ph$i>>>0)<($max$0$i>>>0);
       if (!($86)) {
        label = 14;
        break L13;
       }
       $87 = (($max$0$i) + ($min$0$ph$i))|0;
       $88 = $87 >>> 1;
       $89 = $88 << 2;
       $90 = (($85) + ($89)|0);
       $91 = ((($90)) + 3|0);
       $92 = ((($91)) + -3|0);
       $93 = HEAP8[$92>>0]|0;
       $94 = $93&255;
       $95 = $94 << 16;
       $96 = ((($92)) + 1|0);
       $97 = HEAP8[$96>>0]|0;
       $98 = $97&255;
       $99 = $98 << 8;
       $100 = $99 | $95;
       $101 = ((($92)) + 2|0);
       $102 = HEAP8[$101>>0]|0;
       $103 = $102&255;
       $104 = $100 | $103;
       $105 = ($104>>>0)>($charCode>>>0);
       if ($105) {
        $max$0$i = $88;
       } else {
        $$lcssa = $88;$$lcssa72 = $91;$$lcssa73 = $104;$max$0$i$lcssa71 = $max$0$i;
        break;
       }
      }
      $106 = HEAP8[$$lcssa72>>0]|0;
      $107 = $106&255;
      $108 = (($107) + ($$lcssa73))|0;
      $109 = ($108>>>0)<($charCode>>>0);
      $110 = (($$lcssa) + 1)|0;
      if ($109) {
       $max$0$ph$i = $max$0$i$lcssa71;$min$0$ph$i = $110;
      } else {
       label = 22;
       break;
      }
     }
    }
   } while(0);
   L21: do {
    if ((label|0) == 14) {
     label = 0;
     $111 = ($64|0)==(0);
     if ($111) {
      $q$1 = $q$015;
     } else {
      $112 = HEAP32[$2>>2]|0;
      $113 = (($112) + ($64)|0);
      $114 = HEAP8[$113>>0]|0;
      $115 = $114&255;
      $116 = $115 << 24;
      $117 = ((($113)) + 1|0);
      $118 = HEAP8[$117>>0]|0;
      $119 = $118&255;
      $120 = $119 << 16;
      $121 = $120 | $116;
      $122 = ((($113)) + 2|0);
      $123 = HEAP8[$122>>0]|0;
      $124 = $123&255;
      $125 = $124 << 8;
      $126 = $121 | $125;
      $127 = ((($113)) + 3|0);
      $128 = HEAP8[$127>>0]|0;
      $129 = $128&255;
      $130 = $126 | $129;
      $131 = ((($113)) + 4|0);
      $132 = ($130|0)==(0);
      if ($132) {
       $q$1 = $q$015;
      } else {
       $max$0$ph8$i = $130;$min$0$ph9$i = 0;
       L24: while(1) {
        $min$07$i = $min$0$ph9$i;
        while(1) {
         $134 = (($min$07$i) + ($max$0$ph8$i))|0;
         $135 = $134 >>> 1;
         $136 = ($135*5)|0;
         $137 = (($131) + ($136)|0);
         $138 = ((($137)) + 3|0);
         $139 = ((($138)) + -3|0);
         $140 = HEAP8[$139>>0]|0;
         $141 = $140&255;
         $142 = $141 << 16;
         $143 = ((($139)) + 1|0);
         $144 = HEAP8[$143>>0]|0;
         $145 = $144&255;
         $146 = $145 << 8;
         $147 = $146 | $142;
         $148 = ((($139)) + 2|0);
         $149 = HEAP8[$148>>0]|0;
         $150 = $149&255;
         $151 = $147 | $150;
         $152 = ($151>>>0)>($charCode>>>0);
         if ($152) {
          $$lcssa76 = $135;$min$07$i$lcssa = $min$07$i;
          break;
         }
         $153 = ($151>>>0)<($charCode>>>0);
         if (!($153)) {
          $$lcssa80 = $138;
          break L24;
         }
         $154 = (($135) + 1)|0;
         $155 = ($154>>>0)<($max$0$ph8$i>>>0);
         if ($155) {
          $min$07$i = $154;
         } else {
          $q$1 = $q$015;
          break L21;
         }
        }
        $133 = ($min$07$i$lcssa>>>0)<($$lcssa76>>>0);
        if ($133) {
         $max$0$ph8$i = $$lcssa76;$min$0$ph9$i = $min$07$i$lcssa;
        } else {
         $q$1 = $q$015;
         break L21;
        }
       }
       $156 = HEAP8[$$lcssa80>>0]|0;
       $157 = $156&255;
       $158 = $157 << 8;
       $159 = ((($$lcssa80)) + 1|0);
       $160 = HEAP8[$159>>0]|0;
       $161 = $160&255;
       $162 = $158 | $161;
       $163 = ($162|0)==(0);
       if ($163) {
        $q$1 = $q$015;
       } else {
        label = 22;
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 22) {
    label = 0;
    HEAP32[$q$015>>2] = $27;
    $164 = ((($q$015)) + 4|0);
    $q$1 = $164;
   }
   $165 = (($count$018) + -1)|0;
   $166 = ($165|0)==(0);
   if ($166) {
    $q$0$lcssa = $q$1;
    break;
   } else {
    $count$018 = $165;$p$017 = $46;$q$015 = $q$1;
   }
  }
 }
 HEAP32[$q$0$lcssa>>2] = 0;
 $$0 = $167;
 STACKTOP = sp;return ($$0|0);
}
function _tt_cmap14_variant_chars($cmap,$memory,$variantSelector) {
 $cmap = $cmap|0;
 $memory = $memory|0;
 $variantSelector = $variantSelector|0;
 var $$01$i = 0, $$012$i = 0, $$012$i11 = 0, $$1 = 0, $$lcssa195 = 0, $$lcssa196 = 0, $$lcssa218 = 0, $$phi$trans$insert = 0, $$phi$trans$insert105 = 0, $$phi$trans$insert109 = 0, $$pre = 0, $$pre107 = 0, $$pre111 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0;
 var $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0;
 var $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0;
 var $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0;
 var $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0;
 var $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0;
 var $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0;
 var $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0;
 var $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0;
 var $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0;
 var $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0;
 var $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $dcnt$0$lcssa = 0, $dcnt$0$ph = 0, $dcnt$032 = 0, $dcnt$066 = 0, $di$0$lcssa = 0, $di$0$ph = 0, $di$068 = 0, $di$1 = 0, $di$253 = 0, $dp$0$lcssa = 0, $dp$0$ph = 0, $dp$034 = 0, $dp$067 = 0, $dp$152 = 0, $duni$0$lcssa = 0, $duni$0$ph = 0, $duni$030 = 0, $duni$065 = 0;
 var $error$i = 0, $error$i18 = 0, $error$i3 = 0, $exitcond = 0, $exitcond$i = 0, $exitcond$i12 = 0, $exitcond100 = 0, $exitcond101 = 0, $i$0$lcssa = 0, $i$0$lcssa$i = 0, $i$0$ph = 0, $i$03$i = 0, $i$03$i10 = 0, $i$064 = 0, $i$162 = 0, $i$2 = 0, $i$3 = 0, $i$4 = 0, $i$456 = 0, $i$459 = 0;
 var $i$554 = 0, $i$651 = 0, $i$749 = 0, $i$8 = 0, $k$063 = 0, $k$155 = 0, $k$250 = 0, $max$0$ph$i = 0, $min$0$i = 0, $min$0$ph$i = 0, $ni$0$ph = 0, $ni$1 = 0, $ni$258 = 0, $numRanges$02$i = 0, $nuni$0$ph = 0, $nuni$0$ph209 = 0, $p$0$ph = 0, $p$0$ph207 = 0, $p$157 = 0, $tot$0$lcssa$i = 0;
 var $tot$03$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i18 = sp + 8|0;
 $error$i3 = sp + 4|0;
 $error$i = sp;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 6|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = $4 << 24;
 $6 = ((($1)) + 7|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $9 = $8 << 16;
 $10 = $9 | $5;
 $11 = ((($1)) + 8|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = $12&255;
 $14 = $13 << 8;
 $15 = $10 | $14;
 $16 = ((($1)) + 9|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = $17&255;
 $19 = $15 | $18;
 $20 = ((($1)) + 10|0);
 $max$0$ph$i = $19;$min$0$ph$i = 0;
 L1: while(1) {
  $min$0$i = $min$0$ph$i;
  while(1) {
   $21 = ($min$0$i>>>0)<($max$0$ph$i>>>0);
   if (!($21)) {
    $$1 = 0;
    label = 60;
    break L1;
   }
   $22 = (($min$0$i) + ($max$0$ph$i))|0;
   $23 = $22 >>> 1;
   $24 = ($23*11)|0;
   $25 = (($20) + ($24)|0);
   $26 = ((($25)) + 3|0);
   $27 = ((($26)) + -3|0);
   $28 = HEAP8[$27>>0]|0;
   $29 = $28&255;
   $30 = $29 << 16;
   $31 = ((($27)) + 1|0);
   $32 = HEAP8[$31>>0]|0;
   $33 = $32&255;
   $34 = $33 << 8;
   $35 = $34 | $30;
   $36 = ((($27)) + 2|0);
   $37 = HEAP8[$36>>0]|0;
   $38 = $37&255;
   $39 = $35 | $38;
   $40 = ($39>>>0)>($variantSelector>>>0);
   if ($40) {
    $max$0$ph$i = $23;$min$0$ph$i = $min$0$i;
    continue L1;
   }
   $41 = ($39>>>0)<($variantSelector>>>0);
   $42 = (($23) + 1)|0;
   if ($41) {
    $min$0$i = $42;
   } else {
    $$lcssa218 = $26;
    break L1;
   }
  }
 }
 if ((label|0) == 60) {
  STACKTOP = sp;return ($$1|0);
 }
 $43 = HEAP8[$$lcssa218>>0]|0;
 $44 = $43&255;
 $45 = $44 << 24;
 $46 = ((($$lcssa218)) + 1|0);
 $47 = HEAP8[$46>>0]|0;
 $48 = $47&255;
 $49 = $48 << 16;
 $50 = $49 | $45;
 $51 = ((($$lcssa218)) + 2|0);
 $52 = HEAP8[$51>>0]|0;
 $53 = $52&255;
 $54 = $53 << 8;
 $55 = $50 | $54;
 $56 = ((($$lcssa218)) + 3|0);
 $57 = HEAP8[$56>>0]|0;
 $58 = $57&255;
 $59 = $55 | $58;
 $60 = ((($$lcssa218)) + 4|0);
 $61 = HEAP8[$60>>0]|0;
 $62 = $61&255;
 $63 = $62 << 24;
 $64 = ((($$lcssa218)) + 5|0);
 $65 = HEAP8[$64>>0]|0;
 $66 = $65&255;
 $67 = $66 << 16;
 $68 = $67 | $63;
 $69 = ((($$lcssa218)) + 6|0);
 $70 = HEAP8[$69>>0]|0;
 $71 = $70&255;
 $72 = $71 << 8;
 $73 = $68 | $72;
 $74 = ((($$lcssa218)) + 7|0);
 $75 = HEAP8[$74>>0]|0;
 $76 = $75&255;
 $77 = $73 | $76;
 $78 = ($77|0)==(0);
 $79 = $77 | $59;
 $80 = ($79|0)==(0);
 if ($80) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $81 = ($59|0)==(0);
 if ($81) {
  $82 = (($1) + ($77)|0);
  $83 = HEAP8[$82>>0]|0;
  $84 = $83&255;
  $85 = $84 << 24;
  $86 = ((($82)) + 1|0);
  $87 = HEAP8[$86>>0]|0;
  $88 = $87&255;
  $89 = $88 << 16;
  $90 = $89 | $85;
  $91 = ((($82)) + 2|0);
  $92 = HEAP8[$91>>0]|0;
  $93 = $92&255;
  $94 = $93 << 8;
  $95 = $90 | $94;
  $96 = ((($82)) + 3|0);
  $97 = HEAP8[$96>>0]|0;
  $98 = $97&255;
  $99 = $95 | $98;
  $100 = (($99) + 1)|0;
  $101 = ((($cmap)) + 28|0);
  $102 = HEAP32[$101>>2]|0;
  HEAP32[$error$i3>>2] = 0;
  $103 = ($102>>>0)<($100>>>0);
  do {
   if ($103) {
    $104 = ((($cmap)) + 36|0);
    HEAP32[$104>>2] = $memory;
    $105 = ((($cmap)) + 32|0);
    $106 = HEAP32[$105>>2]|0;
    $107 = (_ft_mem_realloc($memory,4,$102,$100,$106,$error$i3)|0);
    HEAP32[$105>>2] = $107;
    $108 = HEAP32[$error$i3>>2]|0;
    $109 = ($108|0)==(0);
    if ($109) {
     HEAP32[$101>>2] = $100;
     $125 = $107;
     break;
    }
    $$1 = 0;
    STACKTOP = sp;return ($$1|0);
   } else {
    $$phi$trans$insert109 = ((($cmap)) + 32|0);
    $$pre111 = HEAP32[$$phi$trans$insert109>>2]|0;
    $125 = $$pre111;
   }
  } while(0);
  $110 = ($99|0)==(0);
  if ($110) {
   $i$0$lcssa$i = 0;
  } else {
   $111 = ((($82)) + 4|0);
   $$012$i = $111;$i$03$i = 0;
   while(1) {
    $112 = HEAP8[$$012$i>>0]|0;
    $113 = $112&255;
    $114 = $113 << 16;
    $115 = ((($$012$i)) + 1|0);
    $116 = HEAP8[$115>>0]|0;
    $117 = $116&255;
    $118 = $117 << 8;
    $119 = $118 | $114;
    $120 = ((($$012$i)) + 2|0);
    $121 = HEAP8[$120>>0]|0;
    $122 = $121&255;
    $123 = $119 | $122;
    $124 = (($125) + ($i$03$i<<2)|0);
    HEAP32[$124>>2] = $123;
    $126 = ((($$012$i)) + 5|0);
    $127 = (($i$03$i) + 1)|0;
    $exitcond$i = ($127|0)==($99|0);
    if ($exitcond$i) {
     $i$0$lcssa$i = $99;
     break;
    } else {
     $$012$i = $126;$i$03$i = $127;
    }
   }
  }
  $128 = (($125) + ($i$0$lcssa$i<<2)|0);
  HEAP32[$128>>2] = 0;
  $$1 = $125;
  STACKTOP = sp;return ($$1|0);
 }
 if ($78) {
  $129 = (($1) + ($59)|0);
  $130 = (_tt_cmap14_get_def_chars($cmap,$129,$memory)|0);
  $$1 = $130;
  STACKTOP = sp;return ($$1|0);
 }
 $131 = (($1) + ($77)|0);
 $132 = (($1) + ($59)|0);
 $133 = ((($131)) + 4|0);
 $134 = ((($133)) + -4|0);
 $135 = HEAP8[$134>>0]|0;
 $136 = $135&255;
 $137 = $136 << 24;
 $138 = ((($134)) + 1|0);
 $139 = HEAP8[$138>>0]|0;
 $140 = $139&255;
 $141 = $140 << 16;
 $142 = $141 | $137;
 $143 = ((($134)) + 2|0);
 $144 = HEAP8[$143>>0]|0;
 $145 = $144&255;
 $146 = $145 << 8;
 $147 = $142 | $146;
 $148 = ((($134)) + 3|0);
 $149 = HEAP8[$148>>0]|0;
 $150 = $149&255;
 $151 = $147 | $150;
 $152 = HEAP8[$132>>0]|0;
 $153 = $152&255;
 $154 = $153 << 24;
 $155 = ((($132)) + 1|0);
 $156 = HEAP8[$155>>0]|0;
 $157 = $156&255;
 $158 = $157 << 16;
 $159 = $158 | $154;
 $160 = ((($132)) + 2|0);
 $161 = HEAP8[$160>>0]|0;
 $162 = $161&255;
 $163 = $162 << 8;
 $164 = $159 | $163;
 $165 = ((($132)) + 3|0);
 $166 = HEAP8[$165>>0]|0;
 $167 = $166&255;
 $168 = $164 | $167;
 $169 = ($168|0)==(0);
 if ($169) {
  $tot$0$lcssa$i = 0;
 } else {
  $170 = ((($132)) + 7|0);
  $$01$i = $170;$numRanges$02$i = $168;$tot$03$i = 0;
  while(1) {
   $171 = HEAP8[$$01$i>>0]|0;
   $172 = $171&255;
   $173 = (($tot$03$i) + 1)|0;
   $174 = (($173) + ($172))|0;
   $175 = ((($$01$i)) + 4|0);
   $176 = (($numRanges$02$i) + -1)|0;
   $177 = ($176|0)==(0);
   if ($177) {
    $tot$0$lcssa$i = $174;
    break;
   } else {
    $$01$i = $175;$numRanges$02$i = $176;$tot$03$i = $174;
   }
  }
 }
 $178 = ((($132)) + 4|0);
 $179 = ($151|0)==(0);
 if ($179) {
  $180 = (_tt_cmap14_get_def_chars($cmap,$132,$memory)|0);
  $$1 = $180;
  STACKTOP = sp;return ($$1|0);
 }
 $181 = ($tot$0$lcssa$i|0)==(0);
 $182 = (($151) + 1)|0;
 if ($181) {
  $183 = ((($cmap)) + 28|0);
  $184 = HEAP32[$183>>2]|0;
  HEAP32[$error$i18>>2] = 0;
  $185 = ($184>>>0)<($182>>>0);
  do {
   if ($185) {
    $186 = ((($cmap)) + 36|0);
    HEAP32[$186>>2] = $memory;
    $187 = ((($cmap)) + 32|0);
    $188 = HEAP32[$187>>2]|0;
    $189 = (_ft_mem_realloc($memory,4,$184,$182,$188,$error$i18)|0);
    HEAP32[$187>>2] = $189;
    $190 = HEAP32[$error$i18>>2]|0;
    $191 = ($190|0)==(0);
    if ($191) {
     HEAP32[$183>>2] = $182;
     $205 = $189;
     break;
    }
    $$1 = 0;
    STACKTOP = sp;return ($$1|0);
   } else {
    $$phi$trans$insert105 = ((($cmap)) + 32|0);
    $$pre107 = HEAP32[$$phi$trans$insert105>>2]|0;
    $205 = $$pre107;
   }
  } while(0);
  $$012$i11 = $133;$i$03$i10 = 0;
  while(1) {
   $192 = HEAP8[$$012$i11>>0]|0;
   $193 = $192&255;
   $194 = $193 << 16;
   $195 = ((($$012$i11)) + 1|0);
   $196 = HEAP8[$195>>0]|0;
   $197 = $196&255;
   $198 = $197 << 8;
   $199 = $198 | $194;
   $200 = ((($$012$i11)) + 2|0);
   $201 = HEAP8[$200>>0]|0;
   $202 = $201&255;
   $203 = $199 | $202;
   $204 = (($205) + ($i$03$i10<<2)|0);
   HEAP32[$204>>2] = $203;
   $206 = ((($$012$i11)) + 5|0);
   $207 = (($i$03$i10) + 1)|0;
   $exitcond$i12 = ($207|0)==($151|0);
   if ($exitcond$i12) {
    break;
   } else {
    $$012$i11 = $206;$i$03$i10 = $207;
   }
  }
  $208 = (($205) + ($151<<2)|0);
  HEAP32[$208>>2] = 0;
  $$1 = $205;
  STACKTOP = sp;return ($$1|0);
 }
 $209 = (($182) + ($tot$0$lcssa$i))|0;
 $210 = ((($cmap)) + 28|0);
 $211 = HEAP32[$210>>2]|0;
 HEAP32[$error$i>>2] = 0;
 $212 = ($211>>>0)<($209>>>0);
 do {
  if ($212) {
   $213 = ((($cmap)) + 36|0);
   HEAP32[$213>>2] = $memory;
   $214 = ((($cmap)) + 32|0);
   $215 = HEAP32[$214>>2]|0;
   $216 = (_ft_mem_realloc($memory,4,$211,$209,$215,$error$i)|0);
   HEAP32[$214>>2] = $216;
   $217 = HEAP32[$error$i>>2]|0;
   $218 = ($217|0)==(0);
   if ($218) {
    HEAP32[$210>>2] = $209;
    $256 = $216;
    break;
   }
   $$1 = 0;
   STACKTOP = sp;return ($$1|0);
  } else {
   $$phi$trans$insert = ((($cmap)) + 32|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0;
   $256 = $$pre;
  }
 } while(0);
 $219 = ((($178)) + 3|0);
 $220 = ((($219)) + -3|0);
 $221 = HEAP8[$220>>0]|0;
 $222 = $221&255;
 $223 = $222 << 16;
 $224 = ((($220)) + 1|0);
 $225 = HEAP8[$224>>0]|0;
 $226 = $225&255;
 $227 = $226 << 8;
 $228 = $227 | $223;
 $229 = ((($220)) + 2|0);
 $230 = HEAP8[$229>>0]|0;
 $231 = $230&255;
 $232 = $228 | $231;
 $233 = ((($219)) + 1|0);
 $234 = HEAP8[$219>>0]|0;
 $235 = $234&255;
 $236 = ((($133)) + 3|0);
 $237 = ((($236)) + -3|0);
 $238 = HEAP8[$237>>0]|0;
 $239 = $238&255;
 $240 = $239 << 16;
 $241 = ((($237)) + 1|0);
 $242 = HEAP8[$241>>0]|0;
 $243 = $242&255;
 $244 = $243 << 8;
 $245 = $244 | $240;
 $246 = ((($237)) + 2|0);
 $247 = HEAP8[$246>>0]|0;
 $248 = $247&255;
 $249 = $245 | $248;
 $250 = ((($236)) + 2|0);
 $dcnt$0$ph = $235;$di$0$ph = 1;$dp$0$ph = $233;$duni$0$ph = $232;$i$0$ph = 0;$ni$0$ph = 1;$nuni$0$ph = $249;$p$0$ph = $250;
 L66: while(1) {
  $251 = (($dcnt$0$ph) + ($duni$0$ph))|0;
  $252 = ($nuni$0$ph>>>0)>($251>>>0);
  if ($252) {
   $dcnt$066 = $dcnt$0$ph;$di$068 = $di$0$ph;$dp$067 = $dp$0$ph;$duni$065 = $duni$0$ph;$i$064 = $i$0$ph;
   while(1) {
    $i$162 = $i$064;$k$063 = 0;
    while(1) {
     $253 = (($k$063) + ($duni$065))|0;
     $254 = (($i$162) + 1)|0;
     $255 = (($256) + ($i$162<<2)|0);
     HEAP32[$255>>2] = $253;
     $257 = (($k$063) + 1)|0;
     $258 = ($257>>>0)>($dcnt$066>>>0);
     if ($258) {
      $$lcssa196 = $254;
      break;
     } else {
      $i$162 = $254;$k$063 = $257;
     }
    }
    $259 = (($di$068) + 1)|0;
    $260 = ($259>>>0)>($168>>>0);
    if ($260) {
     $dcnt$032 = $dcnt$066;$di$1 = $259;$dp$034 = $dp$067;$duni$030 = $duni$065;$i$3 = $$lcssa196;$ni$1 = $ni$0$ph;$nuni$0$ph209 = $nuni$0$ph;$p$0$ph207 = $p$0$ph;
     break L66;
    }
    $261 = ((($dp$067)) + 3|0);
    $262 = HEAP8[$dp$067>>0]|0;
    $263 = $262&255;
    $264 = $263 << 16;
    $265 = ((($dp$067)) + 1|0);
    $266 = HEAP8[$265>>0]|0;
    $267 = $266&255;
    $268 = $267 << 8;
    $269 = $268 | $264;
    $270 = ((($dp$067)) + 2|0);
    $271 = HEAP8[$270>>0]|0;
    $272 = $271&255;
    $273 = $269 | $272;
    $274 = ((($dp$067)) + 4|0);
    $275 = HEAP8[$261>>0]|0;
    $276 = $275&255;
    $277 = (($276) + ($273))|0;
    $278 = ($nuni$0$ph>>>0)>($277>>>0);
    if ($278) {
     $dcnt$066 = $276;$di$068 = $259;$dp$067 = $274;$duni$065 = $273;$i$064 = $$lcssa196;
    } else {
     $dcnt$0$lcssa = $276;$di$0$lcssa = $259;$dp$0$lcssa = $274;$duni$0$lcssa = $273;$i$0$lcssa = $$lcssa196;
     break;
    }
   }
  } else {
   $dcnt$0$lcssa = $dcnt$0$ph;$di$0$lcssa = $di$0$ph;$dp$0$lcssa = $dp$0$ph;$duni$0$lcssa = $duni$0$ph;$i$0$lcssa = $i$0$ph;
  }
  $279 = ($nuni$0$ph>>>0)<($duni$0$lcssa>>>0);
  if ($279) {
   $280 = (($i$0$lcssa) + 1)|0;
   $281 = (($256) + ($i$0$lcssa<<2)|0);
   HEAP32[$281>>2] = $nuni$0$ph;
   $i$2 = $280;
  } else {
   $i$2 = $i$0$lcssa;
  }
  $282 = (($ni$0$ph) + 1)|0;
  $283 = ($282>>>0)>($151>>>0);
  if ($283) {
   $dcnt$032 = $dcnt$0$lcssa;$di$1 = $di$0$lcssa;$dp$034 = $dp$0$lcssa;$duni$030 = $duni$0$lcssa;$i$3 = $i$2;$ni$1 = $282;$nuni$0$ph209 = $nuni$0$ph;$p$0$ph207 = $p$0$ph;
   break;
  }
  $284 = HEAP8[$p$0$ph>>0]|0;
  $285 = $284&255;
  $286 = $285 << 16;
  $287 = ((($p$0$ph)) + 1|0);
  $288 = HEAP8[$287>>0]|0;
  $289 = $288&255;
  $290 = $289 << 8;
  $291 = $290 | $286;
  $292 = ((($p$0$ph)) + 2|0);
  $293 = HEAP8[$292>>0]|0;
  $294 = $293&255;
  $295 = $291 | $294;
  $296 = ((($p$0$ph)) + 5|0);
  $dcnt$0$ph = $dcnt$0$lcssa;$di$0$ph = $di$0$lcssa;$dp$0$ph = $dp$0$lcssa;$duni$0$ph = $duni$0$lcssa;$i$0$ph = $i$2;$ni$0$ph = $282;$nuni$0$ph = $295;$p$0$ph = $296;
 }
 $297 = ($ni$1>>>0)>($151>>>0);
 if ($297) {
  $317 = ($di$1>>>0)>($168>>>0);
  if ($317) {
   $i$8 = $i$3;
  } else {
   $i$554 = $i$3;$k$155 = 0;
   while(1) {
    $319 = (($k$155) + ($duni$030))|0;
    $320 = (($i$554) + 1)|0;
    $321 = (($256) + ($i$554<<2)|0);
    HEAP32[$321>>2] = $319;
    $322 = (($k$155) + 1)|0;
    $323 = ($322>>>0)>($dcnt$032>>>0);
    if ($323) {
     $$lcssa195 = $320;
     break;
    } else {
     $i$554 = $320;$k$155 = $322;
    }
   }
   $318 = ($di$1>>>0)<($168>>>0);
   if ($318) {
    $di$253 = $di$1;$dp$152 = $dp$034;$i$651 = $$lcssa195;
    while(1) {
     $324 = ((($dp$152)) + 3|0);
     $325 = HEAP8[$dp$152>>0]|0;
     $326 = $325&255;
     $327 = $326 << 16;
     $328 = ((($dp$152)) + 1|0);
     $329 = HEAP8[$328>>0]|0;
     $330 = $329&255;
     $331 = $330 << 8;
     $332 = $331 | $327;
     $333 = ((($dp$152)) + 2|0);
     $334 = HEAP8[$333>>0]|0;
     $335 = $334&255;
     $336 = $332 | $335;
     $337 = HEAP8[$324>>0]|0;
     $338 = $337&255;
     $i$749 = $i$651;$k$250 = 0;
     while(1) {
      $339 = (($k$250) + ($336))|0;
      $340 = (($i$749) + 1)|0;
      $341 = (($256) + ($i$749<<2)|0);
      HEAP32[$341>>2] = $339;
      $342 = (($k$250) + 1)|0;
      $exitcond = ($k$250|0)==($338|0);
      if ($exitcond) {
       break;
      } else {
       $i$749 = $340;$k$250 = $342;
      }
     }
     $343 = ((($dp$152)) + 4|0);
     $344 = (($i$651) + 1)|0;
     $345 = (($344) + ($338))|0;
     $346 = (($di$253) + 1)|0;
     $exitcond100 = ($346|0)==($168|0);
     if ($exitcond100) {
      $i$8 = $345;
      break;
     } else {
      $di$253 = $346;$dp$152 = $343;$i$651 = $345;
     }
    }
   } else {
    $i$8 = $$lcssa195;
   }
  }
 } else {
  $298 = (($256) + ($i$3<<2)|0);
  HEAP32[$298>>2] = $nuni$0$ph209;
  $i$456 = (($i$3) + 1)|0;
  $299 = ($ni$1>>>0)<($151>>>0);
  if ($299) {
   $300 = (($i$3) + ($151))|0;
   $301 = (($300) + 1)|0;
   $i$459 = $i$456;$ni$258 = $ni$1;$p$157 = $p$0$ph207;
   while(1) {
    $302 = HEAP8[$p$157>>0]|0;
    $303 = $302&255;
    $304 = $303 << 16;
    $305 = ((($p$157)) + 1|0);
    $306 = HEAP8[$305>>0]|0;
    $307 = $306&255;
    $308 = $307 << 8;
    $309 = $308 | $304;
    $310 = ((($p$157)) + 2|0);
    $311 = HEAP8[$310>>0]|0;
    $312 = $311&255;
    $313 = $309 | $312;
    $314 = (($256) + ($i$459<<2)|0);
    HEAP32[$314>>2] = $313;
    $315 = ((($p$157)) + 5|0);
    $316 = (($ni$258) + 1)|0;
    $i$4 = (($i$459) + 1)|0;
    $exitcond101 = ($316|0)==($151|0);
    if ($exitcond101) {
     break;
    } else {
     $i$459 = $i$4;$ni$258 = $316;$p$157 = $315;
    }
   }
   $347 = (($301) - ($ni$1))|0;
   $i$8 = $347;
  } else {
   $i$8 = $i$456;
  }
 }
 $348 = (($256) + ($i$8<<2)|0);
 HEAP32[$348>>2] = 0;
 $$1 = $256;
 STACKTOP = sp;return ($$1|0);
}
function _tt_cmap14_get_def_chars($cmap,$p,$memory) {
 $cmap = $cmap|0;
 $p = $p|0;
 $memory = $memory|0;
 var $$0 = 0, $$016 = 0, $$016$pn = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cnt$0 = 0, $numRanges$04 = 0, $q$0$lcssa = 0, $q$05 = 0, $q$1 = 0, $scevgep = 0, $scevgep7 = 0, $uni$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_tt_cmap14_def_char_count($p)|0);
 $1 = HEAP8[$p>>0]|0;
 $2 = ((($p)) + 1|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = ((($p)) + 2|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ((($p)) + 3|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = (($0) + 1)|0;
 $9 = (_tt_cmap14_ensure($cmap,$8,$memory)|0);
 $10 = ($9|0)==(0);
 if (!($10)) {
  $$0 = 0;
  return ($$0|0);
 }
 $11 = $1&255;
 $12 = $11 << 24;
 $13 = $3&255;
 $14 = $13 << 16;
 $15 = $14 | $12;
 $16 = $5&255;
 $17 = $16 << 8;
 $18 = $15 | $17;
 $19 = $7&255;
 $20 = $18 | $19;
 $21 = ((($cmap)) + 32|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ($20|0)==(0);
 if ($23) {
  $q$0$lcssa = $22;
 } else {
  $$016$pn = $p;$numRanges$04 = $20;$q$05 = $22;
  while(1) {
   $$016 = ((($$016$pn)) + 4|0);
   $24 = ((($$016$pn)) + 7|0);
   $25 = HEAP8[$$016>>0]|0;
   $26 = $25&255;
   $27 = $26 << 16;
   $28 = ((($$016$pn)) + 5|0);
   $29 = HEAP8[$28>>0]|0;
   $30 = $29&255;
   $31 = $30 << 8;
   $32 = $31 | $27;
   $33 = ((($$016$pn)) + 6|0);
   $34 = HEAP8[$33>>0]|0;
   $35 = $34&255;
   $36 = $32 | $35;
   $37 = HEAP8[$24>>0]|0;
   $38 = $37&255;
   $39 = (($38) + 1)|0;
   $cnt$0 = $39;$q$1 = $q$05;$uni$0 = $36;
   while(1) {
    HEAP32[$q$1>>2] = $uni$0;
    $40 = (($uni$0) + 1)|0;
    $41 = ((($q$1)) + 4|0);
    $42 = (($cnt$0) + -1)|0;
    $43 = ($42|0)==(0);
    if ($43) {
     break;
    } else {
     $cnt$0 = $42;$q$1 = $41;$uni$0 = $40;
    }
   }
   $scevgep = ((($q$05)) + 4|0);
   $scevgep7 = (($scevgep) + ($38<<2)|0);
   $44 = (($numRanges$04) + -1)|0;
   $45 = ($44|0)==(0);
   if ($45) {
    $q$0$lcssa = $scevgep7;
    break;
   } else {
    $$016$pn = $$016;$numRanges$04 = $44;$q$05 = $scevgep7;
   }
  }
 }
 HEAP32[$q$0$lcssa>>2] = 0;
 $$0 = $22;
 return ($$0|0);
}
function _tt_cmap14_def_char_count($p) {
 $p = $p|0;
 var $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $numRanges$02 = 0, $tot$0$lcssa = 0, $tot$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$p>>0]|0;
 $1 = $0&255;
 $2 = $1 << 24;
 $3 = ((($p)) + 1|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = $4&255;
 $6 = $5 << 16;
 $7 = $6 | $2;
 $8 = ((($p)) + 2|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 $11 = $10 << 8;
 $12 = $7 | $11;
 $13 = ((($p)) + 3|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = $14&255;
 $16 = $12 | $15;
 $17 = ($16|0)==(0);
 if ($17) {
  $tot$0$lcssa = 0;
  return ($tot$0$lcssa|0);
 }
 $18 = ((($p)) + 7|0);
 $$01 = $18;$numRanges$02 = $16;$tot$03 = 0;
 while(1) {
  $19 = HEAP8[$$01>>0]|0;
  $20 = $19&255;
  $21 = (($tot$03) + 1)|0;
  $22 = (($21) + ($20))|0;
  $23 = ((($$01)) + 4|0);
  $24 = (($numRanges$02) + -1)|0;
  $25 = ($24|0)==(0);
  if ($25) {
   $tot$0$lcssa = $22;
   break;
  } else {
   $$01 = $23;$numRanges$02 = $24;$tot$03 = $22;
  }
 }
 return ($tot$0$lcssa|0);
}
function _tt_cmap14_ensure($cmap,$num_results,$memory) {
 $cmap = $cmap|0;
 $num_results = $num_results|0;
 $memory = $memory|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $error = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($cmap)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$error>>2] = 0;
 $2 = ($1>>>0)<($num_results>>>0);
 if (!($2)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ((($cmap)) + 36|0);
 HEAP32[$3>>2] = $memory;
 $4 = ((($cmap)) + 32|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (_ft_mem_realloc($memory,4,$1,$num_results,$5,$error)|0);
 HEAP32[$4>>2] = $6;
 $7 = HEAP32[$error>>2]|0;
 $8 = ($7|0)==(0);
 if (!($8)) {
  $$0 = $7;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$0>>2] = $num_results;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _tt_cmap14_validate($table,$valid) {
 $table = $table|0;
 $valid = $valid|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $defp$08 = 0, $exitcond = 0, $exitcond17 = 0, $exitcond18 = 0, $i$07 = 0, $i1$010 = 0, $lastBase$06 = 0, $lastUni$011 = 0;
 var $lastVarSel$013 = 0, $n$014 = 0, $ndp$09 = 0, $or$cond = 0, $or$cond2 = 0, $p$015 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($table)) + 10|0);
 $1 = ((($valid)) + 160|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($0>>>0)>($2>>>0);
 if ($3) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $4 = ((($table)) + 2|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = $6 << 24;
 $8 = ((($table)) + 3|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 $11 = $10 << 16;
 $12 = $11 | $7;
 $13 = ((($table)) + 4|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = $14&255;
 $16 = $15 << 8;
 $17 = $12 | $16;
 $18 = ((($table)) + 5|0);
 $19 = HEAP8[$18>>0]|0;
 $20 = $19&255;
 $21 = $17 | $20;
 $22 = ((($table)) + 6|0);
 $23 = HEAP8[$22>>0]|0;
 $24 = $23&255;
 $25 = $24 << 24;
 $26 = ((($table)) + 7|0);
 $27 = HEAP8[$26>>0]|0;
 $28 = $27&255;
 $29 = $28 << 16;
 $30 = $29 | $25;
 $31 = ((($table)) + 8|0);
 $32 = HEAP8[$31>>0]|0;
 $33 = $32&255;
 $34 = $33 << 8;
 $35 = $30 | $34;
 $36 = ((($table)) + 9|0);
 $37 = HEAP8[$36>>0]|0;
 $38 = $37&255;
 $39 = $35 | $38;
 $40 = HEAP32[$1>>2]|0;
 $41 = $40;
 $42 = $table;
 $43 = (($41) - ($42))|0;
 $44 = ($21>>>0)>($43>>>0);
 $45 = ($21>>>0)<(10);
 $or$cond = $45 | $44;
 if ($or$cond) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $46 = (($21) + -10)|0;
 $47 = (($46>>>0) / 11)&-1;
 $48 = ($47>>>0)<($39>>>0);
 if ($48) {
  _ft_validator_error($valid,8);
  // unreachable;
 }
 $49 = ($39|0)==(0);
 if ($49) {
  return 0;
 }
 $50 = ((($valid)) + 164|0);
 $51 = ((($valid)) + 172|0);
 $lastVarSel$013 = 1;$n$014 = 0;$p$015 = $0;
 L13: while(1) {
  $52 = HEAP8[$p$015>>0]|0;
  $53 = $52&255;
  $54 = $53 << 16;
  $55 = ((($p$015)) + 1|0);
  $56 = HEAP8[$55>>0]|0;
  $57 = $56&255;
  $58 = $57 << 8;
  $59 = $58 | $54;
  $60 = ((($p$015)) + 2|0);
  $61 = HEAP8[$60>>0]|0;
  $62 = $61&255;
  $63 = $59 | $62;
  $64 = ((($p$015)) + 3|0);
  $65 = HEAP8[$64>>0]|0;
  $66 = $65&255;
  $67 = $66 << 24;
  $68 = ((($p$015)) + 4|0);
  $69 = HEAP8[$68>>0]|0;
  $70 = $69&255;
  $71 = $70 << 16;
  $72 = $71 | $67;
  $73 = ((($p$015)) + 5|0);
  $74 = HEAP8[$73>>0]|0;
  $75 = $74&255;
  $76 = $75 << 8;
  $77 = $72 | $76;
  $78 = ((($p$015)) + 6|0);
  $79 = HEAP8[$78>>0]|0;
  $80 = $79&255;
  $81 = $77 | $80;
  $82 = ((($p$015)) + 11|0);
  $83 = ((($p$015)) + 7|0);
  $84 = HEAP8[$83>>0]|0;
  $85 = $84&255;
  $86 = $85 << 24;
  $87 = ((($p$015)) + 8|0);
  $88 = HEAP8[$87>>0]|0;
  $89 = $88&255;
  $90 = $89 << 16;
  $91 = $90 | $86;
  $92 = ((($p$015)) + 9|0);
  $93 = HEAP8[$92>>0]|0;
  $94 = $93&255;
  $95 = $94 << 8;
  $96 = $91 | $95;
  $97 = ((($p$015)) + 10|0);
  $98 = HEAP8[$97>>0]|0;
  $99 = $98&255;
  $100 = $96 | $99;
  $101 = ($81>>>0)<($21>>>0);
  $102 = ($100>>>0)<($21>>>0);
  $or$cond2 = $101 & $102;
  if (!($or$cond2)) {
   label = 9;
   break;
  }
  $103 = ($63>>>0)<($lastVarSel$013>>>0);
  if ($103) {
   label = 11;
   break;
  }
  $104 = (($63) + 1)|0;
  $105 = ($81|0)==(0);
  if (!($105)) {
   $106 = (($table) + ($81)|0);
   $107 = ((($106)) + 4|0);
   $108 = ((($107)) + -4|0);
   $109 = HEAP8[$108>>0]|0;
   $110 = $109&255;
   $111 = $110 << 24;
   $112 = ((($108)) + 1|0);
   $113 = HEAP8[$112>>0]|0;
   $114 = $113&255;
   $115 = $114 << 16;
   $116 = $115 | $111;
   $117 = ((($108)) + 2|0);
   $118 = HEAP8[$117>>0]|0;
   $119 = $118&255;
   $120 = $119 << 8;
   $121 = $116 | $120;
   $122 = ((($108)) + 3|0);
   $123 = HEAP8[$122>>0]|0;
   $124 = $123&255;
   $125 = $121 | $124;
   $126 = HEAP32[$1>>2]|0;
   $127 = $126;
   $128 = $107;
   $129 = (($127) - ($128))|0;
   $130 = $129 >>> 2;
   $131 = ($125>>>0)>($130>>>0);
   if ($131) {
    label = 14;
    break;
   }
   $132 = ($125|0)==(0);
   if (!($132)) {
    $defp$08 = $107;$i$07 = 0;$lastBase$06 = 0;
    while(1) {
     $133 = ((($defp$08)) + 3|0);
     $134 = HEAP8[$defp$08>>0]|0;
     $135 = $134&255;
     $136 = $135 << 16;
     $137 = ((($defp$08)) + 1|0);
     $138 = HEAP8[$137>>0]|0;
     $139 = $138&255;
     $140 = $139 << 8;
     $141 = $140 | $136;
     $142 = ((($defp$08)) + 2|0);
     $143 = HEAP8[$142>>0]|0;
     $144 = $143&255;
     $145 = $141 | $144;
     $146 = ((($defp$08)) + 4|0);
     $147 = HEAP8[$133>>0]|0;
     $148 = $147&255;
     $149 = (($145) + ($148))|0;
     $150 = ($149>>>0)>(1114111);
     if ($150) {
      label = 17;
      break L13;
     }
     $151 = ($145>>>0)<($lastBase$06>>>0);
     if ($151) {
      label = 19;
      break L13;
     }
     $152 = (($149) + 1)|0;
     $153 = (($i$07) + 1)|0;
     $exitcond = ($153|0)==($125|0);
     if ($exitcond) {
      break;
     } else {
      $defp$08 = $146;$i$07 = $153;$lastBase$06 = $152;
     }
    }
   }
  }
  $154 = ($100|0)==(0);
  if (!($154)) {
   $155 = (($table) + ($100)|0);
   $156 = ((($155)) + 4|0);
   $157 = ((($156)) + -4|0);
   $158 = HEAP8[$157>>0]|0;
   $159 = $158&255;
   $160 = $159 << 24;
   $161 = ((($157)) + 1|0);
   $162 = HEAP8[$161>>0]|0;
   $163 = $162&255;
   $164 = $163 << 16;
   $165 = $164 | $160;
   $166 = ((($157)) + 2|0);
   $167 = HEAP8[$166>>0]|0;
   $168 = $167&255;
   $169 = $168 << 8;
   $170 = $165 | $169;
   $171 = ((($157)) + 3|0);
   $172 = HEAP8[$171>>0]|0;
   $173 = $172&255;
   $174 = $170 | $173;
   $175 = HEAP32[$1>>2]|0;
   $176 = $175;
   $177 = $156;
   $178 = (($176) - ($177))|0;
   $179 = $178 >>> 2;
   $180 = ($174>>>0)>($179>>>0);
   if ($180) {
    label = 23;
    break;
   }
   $181 = ($174|0)==(0);
   if (!($181)) {
    $i1$010 = 0;$lastUni$011 = 0;$ndp$09 = $156;
    while(1) {
     $182 = HEAP8[$ndp$09>>0]|0;
     $183 = $182&255;
     $184 = $183 << 16;
     $185 = ((($ndp$09)) + 1|0);
     $186 = HEAP8[$185>>0]|0;
     $187 = $186&255;
     $188 = $187 << 8;
     $189 = $188 | $184;
     $190 = ((($ndp$09)) + 2|0);
     $191 = HEAP8[$190>>0]|0;
     $192 = $191&255;
     $193 = $189 | $192;
     $194 = ((($ndp$09)) + 5|0);
     $195 = ((($ndp$09)) + 3|0);
     $196 = HEAP8[$195>>0]|0;
     $197 = $196&255;
     $198 = $197 << 8;
     $199 = ((($ndp$09)) + 4|0);
     $200 = HEAP8[$199>>0]|0;
     $201 = $200&255;
     $202 = $198 | $201;
     $203 = ($193>>>0)>(1114111);
     if ($203) {
      label = 26;
      break L13;
     }
     $204 = ($193>>>0)<($lastUni$011>>>0);
     if ($204) {
      label = 28;
      break L13;
     }
     $205 = (($193) + 1)|0;
     $206 = HEAP32[$50>>2]|0;
     $207 = ($206|0)==(0);
     if (!($207)) {
      $208 = HEAP32[$51>>2]|0;
      $209 = ($202>>>0)<($208>>>0);
      if (!($209)) {
       label = 31;
       break L13;
      }
     }
     $210 = (($i1$010) + 1)|0;
     $exitcond17 = ($210|0)==($174|0);
     if ($exitcond17) {
      break;
     } else {
      $i1$010 = $210;$lastUni$011 = $205;$ndp$09 = $194;
     }
    }
   }
  }
  $211 = (($n$014) + 1)|0;
  $exitcond18 = ($211|0)==($39|0);
  if ($exitcond18) {
   label = 34;
   break;
  } else {
   $lastVarSel$013 = $104;$n$014 = $211;$p$015 = $82;
  }
 }
 switch (label|0) {
  case 9: {
   _ft_validator_error($valid,8);
   // unreachable;
   break;
  }
  case 11: {
   _ft_validator_error($valid,8);
   // unreachable;
   break;
  }
  case 14: {
   _ft_validator_error($valid,8);
   // unreachable;
   break;
  }
  case 17: {
   _ft_validator_error($valid,8);
   // unreachable;
   break;
  }
  case 19: {
   _ft_validator_error($valid,8);
   // unreachable;
   break;
  }
  case 23: {
   _ft_validator_error($valid,8);
   // unreachable;
   break;
  }
  case 26: {
   _ft_validator_error($valid,8);
   // unreachable;
   break;
  }
  case 28: {
   _ft_validator_error($valid,8);
   // unreachable;
   break;
  }
  case 31: {
   _ft_validator_error($valid,16);
   // unreachable;
   break;
  }
  case 34: {
   return 0;
   break;
  }
 }
 return (0)|0;
}
function _tt_cmap14_get_info($cmap,$cmap_info) {
 $cmap = $cmap|0;
 $cmap_info = $cmap_info|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap_info)) + 4|0);
 HEAP32[$0>>2] = 14;
 HEAP32[$cmap_info>>2] = -1;
 return 0;
}
function _tt_face_goto_table($face,$tag,$stream,$length) {
 $face = $face|0;
 $tag = $tag|0;
 $stream = $stream|0;
 $length = $length|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $entry$01$i = 0, $entry$01$i$lcssa = 0, $error$0 = 0, $face$idx = 0, $face$idx$val = 0;
 var $face$idx1 = 0, $face$idx1$val = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $face$idx = ((($face)) + 152|0);
 $face$idx$val = HEAP16[$face$idx>>1]|0;
 $face$idx1 = ((($face)) + 156|0);
 $face$idx1$val = HEAP32[$face$idx1>>2]|0;
 $0 = $face$idx$val&65535;
 $1 = (($face$idx1$val) + ($0<<4)|0);
 $2 = ($face$idx$val<<16>>16)==(0);
 if ($2) {
  $error$0 = 142;
  return ($error$0|0);
 } else {
  $entry$01$i = $face$idx1$val;
 }
 while(1) {
  $3 = HEAP32[$entry$01$i>>2]|0;
  $4 = ($3|0)==($tag|0);
  if ($4) {
   $5 = ((($entry$01$i)) + 12|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($6|0)==(0);
   if (!($7)) {
    $$lcssa = $6;$entry$01$i$lcssa = $entry$01$i;
    break;
   }
  }
  $8 = ((($entry$01$i)) + 16|0);
  $9 = ($8>>>0)<($1>>>0);
  if ($9) {
   $entry$01$i = $8;
  } else {
   $error$0 = 142;
   label = 8;
   break;
  }
 }
 if ((label|0) == 8) {
  return ($error$0|0);
 }
 $10 = ($length|0)==(0|0);
 if (!($10)) {
  HEAP32[$length>>2] = $$lcssa;
 }
 $11 = ((($entry$01$i$lcssa)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (_FT_Stream_Seek($stream,$12)|0);
 $error$0 = $13;
 return ($error$0|0);
}
function _sfnt_init_face($stream,$face,$face_index,$num_params,$params) {
 $stream = $stream|0;
 $face = $face|0;
 $face_index = $face_index|0;
 $num_params = $num_params|0;
 $params = $params|0;
 var $$0 = 0, $$0$i$ph$i = 0, $$01126$i = 0, $$01126$i$lcssa62 = 0, $$1$i$ph = 0, $$face_index = 0, $$lcssa57 = 0, $$lcssa60 = 0, $$lobit$i$i = 0, $$not$i = 0, $$pre$i = 0, $$pre46$i$i = 0, $$pre47$i$i = 0, $$pre48$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0;
 var $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0;
 var $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0;
 var $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0;
 var $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0;
 var $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0;
 var $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0;
 var $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0;
 var $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0;
 var $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0;
 var $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0;
 var $374 = 0, $375 = 0, $376 = 0, $377 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $entrySelector$0$lcssa$i$i = 0, $entrySelector$036$i$i = 0, $error$i = 0, $error$i$i = 0, $indices$0$i$i = 0, $n$025$i = 0, $nn$032$i$i = 0, $nn$126$i$i = 0;
 var $nn$222$i$i = 0, $old_tag$033$i$i = 0, $or$cond$i$i = 0, $or$cond36$i = 0, $or$cond38$i = 0, $or$cond4$i$i = 0, $or$cond6$i$i = 0, $or$cond7$i$i = 0, $or$cond9$i$i = 0, $output_len$i$i = 0, $sfnt$0 = 0, $sfnt_header$021$i$i = 0, $sfnt_offset$0$lcssa$i$i = 0, $sfnt_offset$025$i$i = 0, $sfnt_offset$220$i$i = 0, $sfnt_stream$0$i$i = 0, $switch$split102D = 0, $switch$split12D = 0, $switch$split132D = 0, $switch$split162D = 0;
 var $switch$split2D = 0, $switch$split42D = 0, $switch$split72D = 0, $tables$0$i$i = 0, $woff$i$i = 0, $woff_offset$0$lcssa$i$i = 0, $woff_offset$027$i$i = 0, $woff_offset$2$i$i = 0, $woff_offset$3$i$i = 0, $x$037$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i$i = sp + 52|0;
 $woff$i$i = sp + 8|0;
 $output_len$i$i = sp + 4|0;
 $error$i = sp;
 $0 = ((($face)) + 532|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 $3 = ((($face)) + 96|0);
 do {
  if ($2) {
   $4 = HEAP32[$3>>2]|0;
   $5 = ((($4)) + 4|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = (_FT_Get_Module_Interface($6,593168)|0);
   $8 = ($7|0)==(0|0);
   if ($8) {
    $$0 = 11;
    STACKTOP = sp;return ($$0|0);
   } else {
    HEAP32[$0>>2] = $7;
    $9 = HEAP32[$7>>2]|0;
    $10 = ((($face)) + 508|0);
    HEAP32[$10>>2] = $9;
    $sfnt$0 = $7;
    break;
   }
  } else {
   $sfnt$0 = $1;
  }
 } while(0);
 $11 = HEAP32[$3>>2]|0;
 $12 = (_ft_module_get_service($11,594361)|0);
 $13 = ((($face)) + 536|0);
 HEAP32[$13>>2] = $12;
 $14 = ((($stream)) + 28|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($face)) + 132|0);
 HEAP32[$16>>2] = 0;
 $17 = ((($face)) + 136|0);
 HEAP32[$17>>2] = 0;
 $18 = ((($face)) + 140|0);
 HEAP32[$18>>2] = 0;
 $19 = (_FT_Stream_Pos($stream)|0);
 $20 = (_FT_Stream_ReadULong($stream,$error$i)|0);
 $21 = HEAP32[$error$i>>2]|0;
 $22 = ($21|0)==(0);
 L7: do {
  if ($22) {
   $23 = ((($woff$i$i)) + 4|0);
   $24 = ((($woff$i$i)) + 8|0);
   $25 = ((($woff$i$i)) + 12|0);
   $26 = ((($woff$i$i)) + 16|0);
   $27 = ((($woff$i$i)) + 24|0);
   $28 = ((($woff$i$i)) + 28|0);
   $29 = ((($woff$i$i)) + 32|0);
   $30 = ((($woff$i$i)) + 36|0);
   $31 = ((($woff$i$i)) + 40|0);
   $32 = ((($face)) + 104|0);
   $33 = ((($face)) + 8|0);
   $$01126$i = $stream;$34 = $20;$35 = $19;
   L9: while(1) {
    $switch$split2D = ($34|0)<(1953658213);
    if ($switch$split2D) {
     label = 85;
     break;
    }
    $switch$split72D = ($34|0)<(1954115633);
    if ($switch$split72D) {
     label = 91;
     break;
    }
    $switch$split162D = ($34|0)<(2001684038);
    if ($switch$split162D) {
     label = 97;
     break;
    }
    switch ($34|0) {
    case 2001684038:  {
     break;
    }
    default: {
     $$1$i$ph = 2;
     break L7;
    }
    }
    $36 = (_FT_Stream_Seek($$01126$i,$35)|0);
    HEAP32[$error$i>>2] = $36;
    $37 = ($36|0)==(0);
    if (!($37)) {
     $$1$i$ph = $36;
     break L7;
    }
    $38 = ((($$01126$i)) + 28|0);
    $39 = HEAP32[$38>>2]|0;
    HEAP32[$error$i$i>>2] = 0;
    $40 = (_FT_Stream_ReadFields($$01126$i,174050,$woff$i$i)|0);
    HEAP32[$error$i$i>>2] = $40;
    $41 = ($40|0)==(0);
    if (!($41)) {
     $$0$i$ph$i = $40;
     label = 65;
     break;
    }
    $42 = HEAP32[$23>>2]|0;
    $switch$split12D = ($42|0)<(2001684038);
    if ($switch$split12D) {
     switch ($42|0) {
     case 1953784678:  {
      $$0$i$ph$i = 8;
      label = 65;
      break L9;
      break;
     }
     default: {
     }
     }
    } else {
     switch ($42|0) {
     case 2001684038:  {
      $$0$i$ph$i = 8;
      label = 65;
      break L9;
      break;
     }
     default: {
     }
     }
    }
    $43 = HEAP32[$24>>2]|0;
    $44 = ((($$01126$i)) + 4|0);
    $45 = HEAP32[$44>>2]|0;
    $46 = ($43|0)==($45|0);
    if (!($46)) {
     $$0$i$ph$i = 8;
     label = 65;
     break;
    }
    $47 = HEAP16[$25>>1]|0;
    $48 = $47&65535;
    $49 = ($47<<16>>16)==(0);
    if ($49) {
     $$0$i$ph$i = 8;
     label = 65;
     break;
    }
    $50 = ($48*20)|0;
    $51 = (($50) + 44)|0;
    $52 = ($51>>>0)<($43>>>0);
    if (!($52)) {
     $$0$i$ph$i = 8;
     label = 65;
     break;
    }
    $53 = $48 << 4;
    $54 = $53 | 12;
    $55 = HEAP32[$26>>2]|0;
    $56 = ($54>>>0)<($55>>>0);
    $57 = $55 & 3;
    $58 = ($57|0)==(0);
    $or$cond$i$i = $56 & $58;
    if (!($or$cond$i$i)) {
     $$0$i$ph$i = 8;
     label = 65;
     break;
    }
    $59 = HEAP32[$27>>2]|0;
    $60 = ($59|0)==(0);
    $61 = HEAP32[$28>>2]|0;
    if ($60) {
     $62 = HEAP32[$29>>2]|0;
     $63 = $62 | $61;
     $64 = ($63|0)==(0);
     if (!($64)) {
      $$0$i$ph$i = 8;
      label = 65;
      break;
     }
    } else {
     $$not$i = ($61|0)!=(0);
     $65 = HEAP32[$29>>2]|0;
     $66 = ($65|0)==(0);
     $or$cond36$i = $$not$i & $66;
     if ($or$cond36$i) {
      $$0$i$ph$i = 8;
      label = 65;
      break;
     }
    }
    $67 = HEAP32[$30>>2]|0;
    $68 = ($67|0)!=(0);
    $69 = HEAP32[$31>>2]|0;
    $70 = ($69|0)==(0);
    $or$cond38$i = $68 | $70;
    if (!($or$cond38$i)) {
     $$0$i$ph$i = 8;
     label = 65;
     break;
    }
    $71 = (_ft_mem_alloc($39,$55,$error$i$i)|0);
    $72 = HEAP32[$error$i$i>>2]|0;
    $73 = ($72|0)==(0);
    L30: do {
     if ($73) {
      $74 = (_ft_mem_alloc($39,40,$error$i$i)|0);
      $75 = HEAP32[$error$i$i>>2]|0;
      $76 = ($75|0)==(0);
      if ($76) {
       $77 = HEAP16[$25>>1]|0;
       $78 = $77&65535;
       $79 = ($77<<16>>16)==(0);
       if ($79) {
        $entrySelector$0$lcssa$i$i = 0;
       } else {
        $entrySelector$036$i$i = 0;$x$037$i$i = $78;
        while(1) {
         $80 = $x$037$i$i >>> 1;
         $81 = (($entrySelector$036$i$i) + 1)|0;
         $82 = ($80|0)==(0);
         if ($82) {
          $entrySelector$0$lcssa$i$i = $81;
          break;
         } else {
          $entrySelector$036$i$i = $81;$x$037$i$i = $80;
         }
        }
       }
       $83 = (($entrySelector$0$lcssa$i$i) + -1)|0;
       $84 = 16 << $83;
       $85 = $78 << 4;
       $86 = (($85) - ($84))|0;
       $87 = HEAP32[$23>>2]|0;
       $88 = $87 >>> 24;
       $89 = $88&255;
       $90 = ((($71)) + 1|0);
       HEAP8[$71>>0] = $89;
       $91 = HEAP32[$23>>2]|0;
       $92 = $91 >>> 16;
       $93 = $92&255;
       $94 = ((($71)) + 2|0);
       HEAP8[$90>>0] = $93;
       $95 = HEAP32[$23>>2]|0;
       $96 = $95 >>> 8;
       $97 = $96&255;
       $98 = ((($71)) + 3|0);
       HEAP8[$94>>0] = $97;
       $99 = HEAP32[$23>>2]|0;
       $100 = $99&255;
       $101 = ((($71)) + 4|0);
       HEAP8[$98>>0] = $100;
       $102 = HEAP16[$25>>1]|0;
       $103 = ($102&65535) >>> 8;
       $104 = $103&255;
       $105 = ((($71)) + 5|0);
       HEAP8[$101>>0] = $104;
       $106 = HEAP16[$25>>1]|0;
       $107 = $106&255;
       $108 = ((($71)) + 6|0);
       HEAP8[$105>>0] = $107;
       $109 = $84 >>> 8;
       $110 = $109&255;
       $111 = ((($71)) + 7|0);
       HEAP8[$108>>0] = $110;
       $112 = $84&255;
       $113 = ((($71)) + 8|0);
       HEAP8[$111>>0] = $112;
       $114 = $83 >>> 8;
       $115 = $114&255;
       $116 = ((($71)) + 9|0);
       HEAP8[$113>>0] = $115;
       $117 = $83&255;
       $118 = ((($71)) + 10|0);
       HEAP8[$116>>0] = $117;
       $119 = $86 >>> 8;
       $120 = $119&255;
       $121 = ((($71)) + 11|0);
       HEAP8[$118>>0] = $120;
       $122 = $86&255;
       $123 = ((($71)) + 12|0);
       HEAP8[$121>>0] = $122;
       $124 = HEAP16[$25>>1]|0;
       $125 = $124&65535;
       $126 = (_ft_mem_realloc($39,24,0,$125,0,$error$i$i)|0);
       $127 = HEAP32[$error$i$i>>2]|0;
       $128 = ($127|0)==(0);
       if (!($128)) {
        $indices$0$i$i = 0;$sfnt_stream$0$i$i = $74;$tables$0$i$i = $126;
        break;
       }
       $129 = HEAP16[$25>>1]|0;
       $130 = $129&65535;
       $131 = (_ft_mem_realloc($39,4,0,$130,0,$error$i$i)|0);
       $132 = HEAP32[$error$i$i>>2]|0;
       $133 = ($132|0)==(0);
       if (!($133)) {
        $indices$0$i$i = $131;$sfnt_stream$0$i$i = $74;$tables$0$i$i = $126;
        break;
       }
       $134 = HEAP16[$25>>1]|0;
       $135 = $134&65535;
       $136 = ($135*20)|0;
       $137 = (_FT_Stream_EnterFrame($$01126$i,$136)|0);
       HEAP32[$error$i$i>>2] = $137;
       $138 = ($137|0)==(0);
       if (!($138)) {
        $indices$0$i$i = $131;$sfnt_stream$0$i$i = $74;$tables$0$i$i = $126;
        break;
       }
       $139 = HEAP16[$25>>1]|0;
       $140 = ($139<<16>>16)==(0);
       L40: do {
        if (!($140)) {
         $nn$032$i$i = 0;$old_tag$033$i$i = 0;
         while(1) {
          $141 = (($126) + (($nn$032$i$i*24)|0)|0);
          $142 = (_FT_Stream_GetULong($$01126$i)|0);
          HEAP32[$141>>2] = $142;
          $143 = (_FT_Stream_GetULong($$01126$i)|0);
          $144 = (((($126) + (($nn$032$i$i*24)|0)|0)) + 4|0);
          HEAP32[$144>>2] = $143;
          $145 = (_FT_Stream_GetULong($$01126$i)|0);
          $146 = (((($126) + (($nn$032$i$i*24)|0)|0)) + 8|0);
          HEAP32[$146>>2] = $145;
          $147 = (_FT_Stream_GetULong($$01126$i)|0);
          $148 = (((($126) + (($nn$032$i$i*24)|0)|0)) + 12|0);
          HEAP32[$148>>2] = $147;
          $149 = (_FT_Stream_GetULong($$01126$i)|0);
          $150 = (((($126) + (($nn$032$i$i*24)|0)|0)) + 16|0);
          HEAP32[$150>>2] = $149;
          $151 = HEAP32[$141>>2]|0;
          $152 = ($151>>>0)>($old_tag$033$i$i>>>0);
          if (!($152)) {
           break;
          }
          $153 = (($131) + ($nn$032$i$i<<2)|0);
          HEAP32[$153>>2] = $141;
          $154 = (($nn$032$i$i) + 1)|0;
          $155 = HEAP16[$25>>1]|0;
          $156 = $155&65535;
          $157 = ($154|0)<($156|0);
          if ($157) {
           $nn$032$i$i = $154;$old_tag$033$i$i = $151;
          } else {
           break L40;
          }
         }
         _FT_Stream_ExitFrame($$01126$i);
         HEAP32[$error$i$i>>2] = 8;
         $indices$0$i$i = $131;$sfnt_stream$0$i$i = $74;$tables$0$i$i = $126;
         break L30;
        }
       } while(0);
       _FT_Stream_ExitFrame($$01126$i);
       $158 = HEAP16[$25>>1]|0;
       $159 = $158&65535;
       _qsort($131,$159,4,4392);
       $160 = HEAP16[$25>>1]|0;
       $161 = $160&65535;
       $162 = ($161*20)|0;
       $163 = (($162) + 44)|0;
       $164 = $161 << 4;
       $165 = $164 | 12;
       $166 = ($160<<16>>16)==(0);
       L46: do {
        if ($166) {
         $sfnt_offset$0$lcssa$i$i = $165;$woff_offset$0$lcssa$i$i = $163;
        } else {
         $167 = HEAP32[$24>>2]|0;
         $168 = HEAP32[$26>>2]|0;
         $nn$126$i$i = 0;$sfnt_offset$025$i$i = $165;$woff_offset$027$i$i = $163;
         while(1) {
          $169 = (($131) + ($nn$126$i$i<<2)|0);
          $170 = HEAP32[$169>>2]|0;
          $171 = ((($170)) + 4|0);
          $172 = HEAP32[$171>>2]|0;
          $173 = ($172|0)==($woff_offset$027$i$i|0);
          if (!($173)) {
           break;
          }
          $174 = ((($170)) + 8|0);
          $175 = HEAP32[$174>>2]|0;
          $176 = ($167>>>0)<($175>>>0);
          $177 = (($167) - ($175))|0;
          $178 = ($woff_offset$027$i$i>>>0)>($177>>>0);
          $or$cond4$i$i = $176 | $178;
          if ($or$cond4$i$i) {
           break;
          }
          $179 = ((($170)) + 12|0);
          $180 = HEAP32[$179>>2]|0;
          $181 = ($168>>>0)<($180>>>0);
          $182 = (($168) - ($180))|0;
          $183 = ($sfnt_offset$025$i$i>>>0)>($182>>>0);
          $or$cond6$i$i = $181 | $183;
          $184 = ($175>>>0)>($180>>>0);
          $or$cond7$i$i = $184 | $or$cond6$i$i;
          if ($or$cond7$i$i) {
           break;
          }
          $185 = ((($170)) + 20|0);
          HEAP32[$185>>2] = $sfnt_offset$025$i$i;
          $186 = (($175) + 3)|0;
          $187 = $186 & -4;
          $188 = (($187) + ($woff_offset$027$i$i))|0;
          $189 = (($180) + 3)|0;
          $190 = $189 & -4;
          $191 = (($190) + ($sfnt_offset$025$i$i))|0;
          $192 = (($nn$126$i$i) + 1)|0;
          $193 = ($192|0)<($161|0);
          if ($193) {
           $nn$126$i$i = $192;$sfnt_offset$025$i$i = $191;$woff_offset$027$i$i = $188;
          } else {
           $sfnt_offset$0$lcssa$i$i = $191;$woff_offset$0$lcssa$i$i = $188;
           break L46;
          }
         }
         HEAP32[$error$i$i>>2] = 8;
         $indices$0$i$i = $131;$sfnt_stream$0$i$i = $74;$tables$0$i$i = $126;
         break L30;
        }
       } while(0);
       $194 = HEAP32[$27>>2]|0;
       $195 = ($194|0)==(0);
       do {
        if ($195) {
         $woff_offset$2$i$i = $woff_offset$0$lcssa$i$i;
        } else {
         $196 = ($194|0)==($woff_offset$0$lcssa$i$i|0);
         if ($196) {
          $197 = HEAP32[$28>>2]|0;
          $198 = (($197) + ($woff_offset$0$lcssa$i$i))|0;
          $199 = HEAP32[$24>>2]|0;
          $200 = ($198>>>0)>($199>>>0);
          if (!($200)) {
           $woff_offset$2$i$i = $198;
           break;
          }
         }
         HEAP32[$error$i$i>>2] = 8;
         $indices$0$i$i = $131;$sfnt_stream$0$i$i = $74;$tables$0$i$i = $126;
         break L30;
        }
       } while(0);
       $201 = HEAP32[$30>>2]|0;
       $202 = ($201|0)==(0);
       do {
        if ($202) {
         $$pre46$i$i = HEAP32[$24>>2]|0;
         $213 = $$pre46$i$i;$woff_offset$3$i$i = $woff_offset$2$i$i;
        } else {
         $203 = (($woff_offset$2$i$i) + 3)|0;
         $204 = $203 & -4;
         $205 = ($201|0)==($204|0);
         if ($205) {
          $206 = HEAP32[$31>>2]|0;
          $207 = (($206) + ($201))|0;
          $208 = HEAP32[$24>>2]|0;
          $209 = ($207>>>0)>($208>>>0);
          if (!($209)) {
           $213 = $208;$woff_offset$3$i$i = $207;
           break;
          }
         }
         HEAP32[$error$i$i>>2] = 8;
         $indices$0$i$i = $131;$sfnt_stream$0$i$i = $74;$tables$0$i$i = $126;
         break L30;
        }
       } while(0);
       $210 = HEAP32[$26>>2]|0;
       $211 = ($sfnt_offset$0$lcssa$i$i|0)==($210|0);
       $212 = ($woff_offset$3$i$i|0)==($213|0);
       $or$cond9$i$i = $212 & $211;
       if (!($or$cond9$i$i)) {
        HEAP32[$error$i$i>>2] = 8;
        $indices$0$i$i = $131;$sfnt_stream$0$i$i = $74;$tables$0$i$i = $126;
        break;
       }
       do {
        if ($166) {
         $315 = $sfnt_offset$0$lcssa$i$i;
        } else {
         $214 = ((($$01126$i)) + 32|0);
         $nn$222$i$i = 0;$sfnt_header$021$i$i = $123;
         while(1) {
          $215 = (($126) + (($nn$222$i$i*24)|0)|0);
          $216 = HEAP32[$215>>2]|0;
          $217 = $216 >>> 24;
          $218 = $217&255;
          $219 = ((($sfnt_header$021$i$i)) + 1|0);
          HEAP8[$sfnt_header$021$i$i>>0] = $218;
          $220 = HEAP32[$215>>2]|0;
          $221 = $220 >>> 16;
          $222 = $221&255;
          $223 = ((($sfnt_header$021$i$i)) + 2|0);
          HEAP8[$219>>0] = $222;
          $224 = HEAP32[$215>>2]|0;
          $225 = $224 >>> 8;
          $226 = $225&255;
          $227 = ((($sfnt_header$021$i$i)) + 3|0);
          HEAP8[$223>>0] = $226;
          $228 = HEAP32[$215>>2]|0;
          $229 = $228&255;
          $230 = ((($sfnt_header$021$i$i)) + 4|0);
          HEAP8[$227>>0] = $229;
          $231 = (((($126) + (($nn$222$i$i*24)|0)|0)) + 16|0);
          $232 = HEAP32[$231>>2]|0;
          $233 = $232 >>> 24;
          $234 = $233&255;
          $235 = ((($sfnt_header$021$i$i)) + 5|0);
          HEAP8[$230>>0] = $234;
          $236 = HEAP32[$231>>2]|0;
          $237 = $236 >>> 16;
          $238 = $237&255;
          $239 = ((($sfnt_header$021$i$i)) + 6|0);
          HEAP8[$235>>0] = $238;
          $240 = HEAP32[$231>>2]|0;
          $241 = $240 >>> 8;
          $242 = $241&255;
          $243 = ((($sfnt_header$021$i$i)) + 7|0);
          HEAP8[$239>>0] = $242;
          $244 = HEAP32[$231>>2]|0;
          $245 = $244&255;
          $246 = ((($sfnt_header$021$i$i)) + 8|0);
          HEAP8[$243>>0] = $245;
          $247 = (((($126) + (($nn$222$i$i*24)|0)|0)) + 20|0);
          $248 = HEAP32[$247>>2]|0;
          $249 = $248 >>> 24;
          $250 = $249&255;
          $251 = ((($sfnt_header$021$i$i)) + 9|0);
          HEAP8[$246>>0] = $250;
          $252 = HEAP32[$247>>2]|0;
          $253 = $252 >>> 16;
          $254 = $253&255;
          $255 = ((($sfnt_header$021$i$i)) + 10|0);
          HEAP8[$251>>0] = $254;
          $256 = HEAP32[$247>>2]|0;
          $257 = $256 >>> 8;
          $258 = $257&255;
          $259 = ((($sfnt_header$021$i$i)) + 11|0);
          HEAP8[$255>>0] = $258;
          $260 = HEAP32[$247>>2]|0;
          $261 = $260&255;
          $262 = ((($sfnt_header$021$i$i)) + 12|0);
          HEAP8[$259>>0] = $261;
          $263 = (((($126) + (($nn$222$i$i*24)|0)|0)) + 12|0);
          $264 = HEAP32[$263>>2]|0;
          $265 = $264 >>> 24;
          $266 = $265&255;
          $267 = ((($sfnt_header$021$i$i)) + 13|0);
          HEAP8[$262>>0] = $266;
          $268 = HEAP32[$263>>2]|0;
          $269 = $268 >>> 16;
          $270 = $269&255;
          $271 = ((($sfnt_header$021$i$i)) + 14|0);
          HEAP8[$267>>0] = $270;
          $272 = HEAP32[$263>>2]|0;
          $273 = $272 >>> 8;
          $274 = $273&255;
          $275 = ((($sfnt_header$021$i$i)) + 15|0);
          HEAP8[$271>>0] = $274;
          $276 = HEAP32[$263>>2]|0;
          $277 = $276&255;
          $278 = ((($sfnt_header$021$i$i)) + 16|0);
          HEAP8[$275>>0] = $277;
          $279 = (((($126) + (($nn$222$i$i*24)|0)|0)) + 4|0);
          $280 = HEAP32[$279>>2]|0;
          $281 = (_FT_Stream_Seek($$01126$i,$280)|0);
          HEAP32[$error$i$i>>2] = $281;
          $282 = ($281|0)==(0);
          if (!($282)) {
           $indices$0$i$i = $131;$sfnt_stream$0$i$i = $74;$tables$0$i$i = $126;
           break L30;
          }
          $283 = (((($126) + (($nn$222$i$i*24)|0)|0)) + 8|0);
          $284 = HEAP32[$283>>2]|0;
          $285 = (_FT_Stream_EnterFrame($$01126$i,$284)|0);
          HEAP32[$error$i$i>>2] = $285;
          $286 = ($285|0)==(0);
          if (!($286)) {
           $indices$0$i$i = $131;$sfnt_stream$0$i$i = $74;$tables$0$i$i = $126;
           break L30;
          }
          $287 = HEAP32[$283>>2]|0;
          $288 = HEAP32[$263>>2]|0;
          $289 = ($287|0)==($288|0);
          if ($289) {
           $290 = HEAP32[$247>>2]|0;
           $291 = (($71) + ($290)|0);
           $292 = HEAP32[$214>>2]|0;
           _memcpy(($291|0),($292|0),($287|0))|0;
          } else {
           HEAP32[$output_len$i$i>>2] = $288;
           $293 = HEAP32[$247>>2]|0;
           $294 = (($71) + ($293)|0);
           $295 = HEAP32[$214>>2]|0;
           $296 = HEAP32[$283>>2]|0;
           $297 = (_FT_Gzip_Uncompress($39,$294,$output_len$i$i,$295,$296)|0);
           HEAP32[$error$i$i>>2] = $297;
           $298 = ($297|0)==(0);
           if (!($298)) {
            break;
           }
           $299 = HEAP32[$output_len$i$i>>2]|0;
           $300 = HEAP32[$263>>2]|0;
           $301 = ($299|0)==($300|0);
           if (!($301)) {
            label = 55;
            break;
           }
          }
          _FT_Stream_ExitFrame($$01126$i);
          $302 = HEAP32[$247>>2]|0;
          $303 = HEAP32[$263>>2]|0;
          $304 = (($303) + ($302))|0;
          $305 = $304 & 3;
          $306 = ($305|0)==(0);
          if (!($306)) {
           $sfnt_offset$220$i$i = $304;
           while(1) {
            $307 = (($71) + ($sfnt_offset$220$i$i)|0);
            HEAP8[$307>>0] = 0;
            $308 = (($sfnt_offset$220$i$i) + 1)|0;
            $309 = $308 & 3;
            $310 = ($309|0)==(0);
            if ($310) {
             break;
            } else {
             $sfnt_offset$220$i$i = $308;
            }
           }
          }
          $311 = (($nn$222$i$i) + 1)|0;
          $312 = HEAP16[$25>>1]|0;
          $313 = $312&65535;
          $314 = ($311|0)<($313|0);
          if ($314) {
           $nn$222$i$i = $311;$sfnt_header$021$i$i = $278;
          } else {
           label = 61;
           break;
          }
         }
         if ((label|0) == 55) {
          label = 0;
          HEAP32[$error$i$i>>2] = 8;
         }
         else if ((label|0) == 61) {
          label = 0;
          $$pre47$i$i = HEAP32[$26>>2]|0;
          $315 = $$pre47$i$i;
          break;
         }
         $indices$0$i$i = $131;$sfnt_stream$0$i$i = $74;$tables$0$i$i = $126;
         break L30;
        }
       } while(0);
       _FT_Stream_OpenMemory($74,$71,$315);
       $316 = HEAP32[$38>>2]|0;
       $317 = ((($74)) + 28|0);
       HEAP32[$317>>2] = $316;
       $318 = ((($74)) + 24|0);
       HEAP32[$318>>2] = 4393;
       $319 = HEAP32[$32>>2]|0;
       $320 = HEAP32[$33>>2]|0;
       $321 = $320 >>> 10;
       $$lobit$i$i = $321 & 1;
       _FT_Stream_Free($319,$$lobit$i$i);
       HEAP32[$32>>2] = $74;
       $322 = HEAP32[$33>>2]|0;
       $323 = $322 & -1025;
       HEAP32[$33>>2] = $323;
       $indices$0$i$i = $131;$sfnt_stream$0$i$i = $74;$tables$0$i$i = $126;
      } else {
       $indices$0$i$i = 0;$sfnt_stream$0$i$i = $74;$tables$0$i$i = 0;
      }
     } else {
      $indices$0$i$i = 0;$sfnt_stream$0$i$i = 0;$tables$0$i$i = 0;
     }
    } while(0);
    _ft_mem_free($39,$tables$0$i$i);
    _ft_mem_free($39,$indices$0$i$i);
    $324 = HEAP32[$error$i$i>>2]|0;
    $325 = ($324|0)==(0);
    if ($325) {
     HEAP32[$error$i>>2] = 0;
    } else {
     _ft_mem_free($39,$71);
     _FT_Stream_Close($sfnt_stream$0$i$i);
     _ft_mem_free($39,$sfnt_stream$0$i$i);
     $$pre48$i$i = HEAP32[$error$i$i>>2]|0;
     HEAP32[$error$i>>2] = $$pre48$i$i;
     $326 = ($$pre48$i$i|0)==(0);
     if (!($326)) {
      $$1$i$ph = $$pre48$i$i;
      break L7;
     }
    }
    $327 = HEAP32[$32>>2]|0;
    $328 = (_FT_Stream_Pos($327)|0);
    $329 = (_FT_Stream_ReadULong($327,$error$i)|0);
    $330 = HEAP32[$error$i>>2]|0;
    $331 = ($330|0)==(0);
    if ($331) {
     $$01126$i = $327;$34 = $329;$35 = $328;
    } else {
     $$1$i$ph = $330;
     break L7;
    }
   }
   L97: do {
    if ((label|0) == 65) {
     HEAP32[$error$i>>2] = $$0$i$ph$i;
     $$1$i$ph = $$0$i$ph$i;
     break L7;
    }
    else if ((label|0) == 85) {
     $switch$split42D = ($34|0)<(131072);
     if ($switch$split42D) {
      switch ($34|0) {
      case 65536:  {
       $$01126$i$lcssa62 = $$01126$i;$$lcssa57 = $34;$$lcssa60 = $35;
       break L97;
       break;
      }
      default: {
       $$1$i$ph = 2;
       break L7;
      }
      }
     }
     $switch$split102D = ($34|0)<(1330926671);
     if ($switch$split102D) {
      switch ($34|0) {
      case 131072:  {
       $$01126$i$lcssa62 = $$01126$i;$$lcssa57 = $34;$$lcssa60 = $35;
       break L97;
       break;
      }
      default: {
       $$1$i$ph = 2;
       break L7;
      }
      }
     } else {
      switch ($34|0) {
      case 1330926671:  {
       $$01126$i$lcssa62 = $$01126$i;$$lcssa57 = $34;$$lcssa60 = $35;
       break L97;
       break;
      }
      default: {
       $$1$i$ph = 2;
       break L7;
      }
      }
     }
    }
    else if ((label|0) == 91) {
     $switch$split132D = ($34|0)<(1953784678);
     if ($switch$split132D) {
      switch ($34|0) {
      case 1953658213:  {
       $$01126$i$lcssa62 = $$01126$i;$$lcssa57 = $34;$$lcssa60 = $35;
       break L97;
       break;
      }
      default: {
       $$1$i$ph = 2;
       break L7;
      }
      }
     } else {
      switch ($34|0) {
      case 1953784678:  {
       $$01126$i$lcssa62 = $$01126$i;$$lcssa57 = $34;$$lcssa60 = $35;
       break L97;
       break;
      }
      default: {
       $$1$i$ph = 2;
       break L7;
      }
      }
     }
    }
    else if ((label|0) == 97) {
     switch ($34|0) {
     case 1954115633:  {
      $$01126$i$lcssa62 = $$01126$i;$$lcssa57 = $34;$$lcssa60 = $35;
      break;
     }
     default: {
      $$1$i$ph = 2;
      break L7;
     }
     }
    }
   } while(0);
   HEAP32[$16>>2] = 1953784678;
   $332 = ($$lcssa57|0)==(1953784678);
   if ($332) {
    $333 = (_FT_Stream_ReadFields($$01126$i$lcssa62,174110,$16)|0);
    HEAP32[$error$i>>2] = $333;
    $334 = ($333|0)==(0);
    if (!($334)) {
     $$1$i$ph = $333;
     break;
    }
    $335 = HEAP32[$18>>2]|0;
    $336 = ($335|0)==(0);
    if ($336) {
     $$1$i$ph = 8;
     break;
    }
    $337 = ((($$01126$i$lcssa62)) + 4|0);
    $338 = HEAP32[$337>>2]|0;
    $339 = $338 >>> 5;
    $340 = ($335>>>0)>($339>>>0);
    if ($340) {
     $$1$i$ph = 10;
     break;
    }
    $341 = (_ft_mem_realloc($15,4,0,$335,0,$error$i)|0);
    $342 = ((($face)) + 144|0);
    HEAP32[$342>>2] = $341;
    $343 = HEAP32[$error$i>>2]|0;
    $344 = ($343|0)==(0);
    if (!($344)) {
     $$1$i$ph = $343;
     break;
    }
    $345 = HEAP32[$18>>2]|0;
    $346 = $345 << 2;
    $347 = (_FT_Stream_EnterFrame($$01126$i$lcssa62,$346)|0);
    HEAP32[$error$i>>2] = $347;
    $348 = ($347|0)==(0);
    if (!($348)) {
     $$1$i$ph = $347;
     break;
    }
    $349 = HEAP32[$18>>2]|0;
    $350 = ($349|0)>(0);
    if ($350) {
     $n$025$i = 0;
     while(1) {
      $351 = (_FT_Stream_GetULong($$01126$i$lcssa62)|0);
      $352 = HEAP32[$342>>2]|0;
      $353 = (($352) + ($n$025$i<<2)|0);
      HEAP32[$353>>2] = $351;
      $354 = (($n$025$i) + 1)|0;
      $355 = HEAP32[$18>>2]|0;
      $356 = ($354|0)<($355|0);
      if ($356) {
       $n$025$i = $354;
      } else {
       break;
      }
     }
    }
    _FT_Stream_ExitFrame($$01126$i$lcssa62);
    $$pre$i = HEAP32[$error$i>>2]|0;
    $361 = ($$pre$i|0)==(0);
    if (!($361)) {
     $$0 = $$pre$i;
     STACKTOP = sp;return ($$0|0);
    }
   } else {
    HEAP32[$17>>2] = 65536;
    HEAP32[$18>>2] = 1;
    $357 = (_ft_mem_alloc($15,4,$error$i)|0);
    $358 = ((($face)) + 144|0);
    HEAP32[$358>>2] = $357;
    $359 = HEAP32[$error$i>>2]|0;
    $360 = ($359|0)==(0);
    if (!($360)) {
     $$1$i$ph = $359;
     break;
    }
    HEAP32[$357>>2] = $$lcssa60;
   }
   $362 = HEAP32[$32>>2]|0;
   $363 = ($face_index|0)<(0);
   $$face_index = $363 ? 0 : $face_index;
   $364 = HEAP32[$18>>2]|0;
   $365 = ($$face_index|0)<($364|0);
   if (!($365)) {
    $$0 = 6;
    STACKTOP = sp;return ($$0|0);
   }
   $366 = ((($face)) + 144|0);
   $367 = HEAP32[$366>>2]|0;
   $368 = (($367) + ($$face_index<<2)|0);
   $369 = HEAP32[$368>>2]|0;
   $370 = (_FT_Stream_Seek($362,$369)|0);
   $371 = ($370|0)==(0);
   if (!($371)) {
    $$0 = $370;
    STACKTOP = sp;return ($$0|0);
   }
   $372 = ((($sfnt$0)) + 88|0);
   $373 = HEAP32[$372>>2]|0;
   $374 = (FUNCTION_TABLE_iii[$373 & 8191]($face,$362)|0);
   $375 = ($374|0)==(0);
   if (!($375)) {
    $$0 = $374;
    STACKTOP = sp;return ($$0|0);
   }
   $376 = HEAP32[$18>>2]|0;
   HEAP32[$face>>2] = $376;
   $377 = ((($face)) + 4|0);
   HEAP32[$377>>2] = $$face_index;
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  } else {
   $$1$i$ph = $21;
  }
 } while(0);
 $$0 = $$1$i$ph;
 STACKTOP = sp;return ($$0|0);
}
function _compare_offsets($a,$b) {
 $a = $a|0;
 $b = $b|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$a>>2]|0;
 $1 = HEAP32[$b>>2]|0;
 $2 = ((($0)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($3>>>0)>($5>>>0);
 $7 = ($3>>>0)<($5>>>0);
 $$ = $7 << 31 >> 31;
 $$0 = $6 ? 1 : $$;
 return ($$0|0);
}
function _sfnt_stream_close($stream) {
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$stream>>2]|0;
 _ft_mem_free($1,$2);
 $3 = ((($stream)) + 4|0);
 HEAP32[$3>>2] = 0;
 HEAP32[$stream>>2] = 0;
 $4 = ((($stream)) + 24|0);
 HEAP32[$4>>2] = 0;
 return;
}
function _sfnt_load_face($stream,$face,$face_index,$num_params,$params) {
 $stream = $stream|0;
 $face = $face|0;
 $face_index = $face_index|0;
 $num_params = $num_params|0;
 $params = $params|0;
 var $$ = 0, $$0$i26 = 0, $$0$i28 = 0, $$0$i35$ph = 0, $$14 = 0, $$62 = 0, $$9 = 0, $$flags$0 = 0, $$ignore_preferred_subfamily$0 = 0, $$in = 0, $$not = 0, $$pr49 = 0, $$pr51 = 0, $$pre = 0, $$pre82 = 0, $$pre83 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0;
 var $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0;
 var $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0;
 var $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0;
 var $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0;
 var $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0;
 var $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0;
 var $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0;
 var $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0;
 var $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0;
 var $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $avgwidth$0 = 0, $brmerge = 0, $cur$02$lcssa$i = 0, $div = 0, $em_size$0 = 0, $entry$01$i = 0, $entry$01$i24 = 0, $entry$01$i29 = 0;
 var $entry$01$i33 = 0, $entry$01$i37 = 0, $error = 0, $exitcond = 0, $face$idx18 = 0, $face$idx18$val = 0, $face$idx19 = 0, $face$idx19$val = 0, $face$idx22 = 0, $face$idx22$val = 0, $face$idx23 = 0, $face$idx23$val = 0, $flags$2 = 0, $flags$2$ = 0, $flags$4 = 0, $flags$4$ = 0, $flags$4$$ = 0, $flags$6 = 0, $flags$9 = 0, $has_outline$145 = 0;
 var $has_outline$2 = 0, $i$071 = 0, $i1$066100 = 0, $ignore_preferred_family$0$lcssa = 0, $ignore_preferred_family$073 = 0, $ignore_preferred_family$1 = 0, $ignore_preferred_subfamily$0$lcssa = 0, $ignore_preferred_subfamily$072 = 0, $ignore_preferred_subfamily$1 = 0, $m$067 = 0, $metrics = 0, $or$cond = 0, $or$cond5 = 0, $or$cond8 = 0, $sext = 0, $sext87 = 0, $sext90 = 0, $switch = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp + 28|0;
 $metrics = sp;
 $0 = ((($face)) + 532|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($num_params|0)>(0);
 if ($2) {
  $i$071 = 0;$ignore_preferred_family$073 = 0;$ignore_preferred_subfamily$072 = 0;
  while(1) {
   $3 = (($params) + ($i$071<<3)|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==(1768386662);
   $6 = ($4|0)==(1768386675);
   $$ignore_preferred_subfamily$0 = $6 ? 1 : $ignore_preferred_subfamily$072;
   $ignore_preferred_subfamily$1 = $5 ? $ignore_preferred_subfamily$072 : $$ignore_preferred_subfamily$0;
   $ignore_preferred_family$1 = $5 ? 1 : $ignore_preferred_family$073;
   $7 = (($i$071) + 1)|0;
   $exitcond = ($7|0)==($num_params|0);
   if ($exitcond) {
    $ignore_preferred_family$0$lcssa = $ignore_preferred_family$1;$ignore_preferred_subfamily$0$lcssa = $ignore_preferred_subfamily$1;
    break;
   } else {
    $i$071 = $7;$ignore_preferred_family$073 = $ignore_preferred_family$1;$ignore_preferred_subfamily$072 = $ignore_preferred_subfamily$1;
   }
  }
 } else {
  $ignore_preferred_family$0$lcssa = 0;$ignore_preferred_subfamily$0$lcssa = 0;
 }
 $8 = ((($face)) + 128|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 48|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0|0);
 L5: do {
  if ($12) {
   $face$idx22 = ((($face)) + 152|0);
   $face$idx22$val = HEAP16[$face$idx22>>1]|0;
   $face$idx23 = ((($face)) + 156|0);
   $face$idx23$val = HEAP32[$face$idx23>>2]|0;
   $13 = $face$idx22$val&65535;
   $14 = (($face$idx23$val) + ($13<<4)|0);
   $15 = ($face$idx22$val<<16>>16)==(0);
   L7: do {
    if ($15) {
     $$0$i26 = 0;
    } else {
     $entry$01$i = $face$idx23$val;
     while(1) {
      $16 = HEAP32[$entry$01$i>>2]|0;
      $17 = ($16|0)==(1735162214);
      if ($17) {
       $18 = ((($entry$01$i)) + 12|0);
       $19 = HEAP32[$18>>2]|0;
       $20 = ($19|0)==(0);
       if (!($20)) {
        $32 = 1;
        break L5;
       }
      }
      $21 = ((($entry$01$i)) + 16|0);
      $22 = ($21>>>0)<($14>>>0);
      if ($22) {
       $entry$01$i = $21;
      } else {
       break;
      }
     }
     if ($15) {
      $$0$i26 = 0;
     } else {
      $entry$01$i24 = $face$idx23$val;
      while(1) {
       $23 = HEAP32[$entry$01$i24>>2]|0;
       $24 = ($23|0)==(1128678944);
       if ($24) {
        $25 = ((($entry$01$i24)) + 12|0);
        $26 = HEAP32[$25>>2]|0;
        $27 = ($26|0)==(0);
        if (!($27)) {
         $$0$i26 = $entry$01$i24;
         break L7;
        }
       }
       $28 = ((($entry$01$i24)) + 16|0);
       $29 = ($28>>>0)<($14>>>0);
       if ($29) {
        $entry$01$i24 = $28;
       } else {
        $$0$i26 = 0;
        break;
       }
      }
     }
    }
   } while(0);
   $30 = ($$0$i26|0)!=(0|0);
   $32 = $30;
  } else {
   $32 = 1;
  }
 } while(0);
 $31 = $32&1;
 $33 = ((($face)) + 508|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = (FUNCTION_TABLE_iiiii[$34 & 8191]($face,1935829368,$stream,0)|0);
 $36 = ($35|0)==(0);
 $$ = $36 ? 0 : $31;
 $37 = ($$<<24>>24)==(0);
 if ($37) {
  $38 = ((($1)) + 68|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = ($39|0)==(0|0);
  if ($40) {
   $413 = 0;
   label = 16;
  } else {
   $41 = (FUNCTION_TABLE_iii[$39 & 8191]($face,$stream)|0);
   HEAP32[$error>>2] = $41;
   $42 = ($41|0)==(0);
   $$not = $42 ^ 1;
   $brmerge = $36 | $$not;
   if ($brmerge) {
    $413 = $42;
    label = 16;
   } else {
    $414 = $42;
   }
  }
 } else {
  $413 = 0;
  label = 16;
 }
 if ((label|0) == 16) {
  $43 = ((($1)) + 24|0);
  $44 = HEAP32[$43>>2]|0;
  $45 = (FUNCTION_TABLE_iii[$44 & 8191]($face,$stream)|0);
  HEAP32[$error>>2] = $45;
  $46 = ($45|0)==(0);
  if ($46) {
   $414 = $413;
  } else {
   $412 = $45;
   STACKTOP = sp;return ($412|0);
  }
 }
 $47 = ((($face)) + 178|0);
 $48 = HEAP16[$47>>1]|0;
 $49 = ($48<<16>>16)==(0);
 if ($49) {
  HEAP32[$error>>2] = 8;
  $412 = 8;
  STACKTOP = sp;return ($412|0);
 }
 $50 = ((($1)) + 36|0);
 $51 = HEAP32[$50>>2]|0;
 (FUNCTION_TABLE_iii[$51 & 8191]($face,$stream)|0);
 $52 = ((($1)) + 32|0);
 $53 = HEAP32[$52>>2]|0;
 (FUNCTION_TABLE_iii[$53 & 8191]($face,$stream)|0);
 $54 = ((($1)) + 48|0);
 $55 = HEAP32[$54>>2]|0;
 (FUNCTION_TABLE_iii[$55 & 8191]($face,$stream)|0);
 $56 = ((($1)) + 44|0);
 $57 = HEAP32[$56>>2]|0;
 $58 = (FUNCTION_TABLE_iii[$57 & 8191]($face,$stream)|0);
 HEAP32[$error>>2] = $58;
 if ($414) {
  $has_outline$2 = $$;
 } else {
  $59 = ((($1)) + 28|0);
  $60 = HEAP32[$59>>2]|0;
  $61 = (FUNCTION_TABLE_iiii[$60 & 8191]($face,$stream,0)|0);
  HEAP32[$error>>2] = $61;
  $62 = ($61|0)==(0);
  do {
   if ($62) {
    $63 = ((($1)) + 92|0);
    $64 = HEAP32[$63>>2]|0;
    $65 = (FUNCTION_TABLE_iiii[$64 & 8191]($face,$stream,0)|0);
    HEAP32[$error>>2] = $65;
    $66 = $65 & 255;
    $67 = ($66|0)==(142);
    if (!($67)) {
     $91 = ($65|0)==(0);
     if ($91) {
      $has_outline$145 = $$;
      break;
     } else {
      $412 = $65;
     }
     STACKTOP = sp;return ($412|0);
    }
    HEAP32[$error>>2] = 147;
    $68 = HEAP32[$8>>2]|0;
    $69 = ((($68)) + 48|0);
    $70 = HEAP32[$69>>2]|0;
    $71 = ($70|0)==(0|0);
    if ($71) {
     $412 = 147;
     STACKTOP = sp;return ($412|0);
    }
    $72 = HEAP32[$70>>2]|0;
    $73 = ((($72)) + 8|0);
    $74 = HEAP32[$73>>2]|0;
    $75 = ($74|0)==(0|0);
    if ($75) {
     $412 = 147;
     STACKTOP = sp;return ($412|0);
    } else {
     $76 = ((($face)) + 250|0);
     HEAP16[$76>>1] = 0;
     HEAP32[$error>>2] = 0;
     $has_outline$145 = $$;
     break;
    }
   } else {
    $77 = $61 & 255;
    $78 = ($77|0)==(142);
    if (!($78)) {
     $412 = $61;
     STACKTOP = sp;return ($412|0);
    }
    $79 = ((($face)) + 148|0);
    $80 = HEAP32[$79>>2]|0;
    $81 = ($80|0)==(1953658213);
    if ($81) {
     HEAP32[$error>>2] = 0;
     $has_outline$145 = 0;
     break;
    }
    HEAP32[$error>>2] = 143;
    $82 = HEAP32[$8>>2]|0;
    $83 = ((($82)) + 48|0);
    $84 = HEAP32[$83>>2]|0;
    $85 = ($84|0)==(0|0);
    if ($85) {
     $412 = 143;
     STACKTOP = sp;return ($412|0);
    }
    $86 = HEAP32[$84>>2]|0;
    $87 = ((($86)) + 8|0);
    $88 = HEAP32[$87>>2]|0;
    $89 = ($88|0)==(0|0);
    if ($89) {
     $412 = 143;
     STACKTOP = sp;return ($412|0);
    } else {
     $90 = ((($face)) + 250|0);
     HEAP16[$90>>1] = 0;
     HEAP32[$error>>2] = 0;
     $has_outline$145 = $$;
     break;
    }
   }
  } while(0);
  $92 = HEAP32[$59>>2]|0;
  $93 = (FUNCTION_TABLE_iiii[$92 & 8191]($face,$stream,1)|0);
  HEAP32[$error>>2] = $93;
  $94 = ($93|0)==(0);
  if ($94) {
   $95 = ((($1)) + 92|0);
   $96 = HEAP32[$95>>2]|0;
   $97 = (FUNCTION_TABLE_iiii[$96 & 8191]($face,$stream,1)|0);
   HEAP32[$error>>2] = $97;
   $98 = ($97|0)==(0);
   if ($98) {
    $99 = ((($face)) + 292|0);
    HEAP8[$99>>0] = 1;
   } else {
    $101 = $97;
    label = 35;
   }
  } else {
   $101 = $93;
   label = 35;
  }
  if ((label|0) == 35) {
   $100 = ($101|0)==(0);
   $102 = $101 & 255;
   $103 = ($102|0)==(142);
   $or$cond8 = $100 | $103;
   if (!($or$cond8)) {
    $412 = $101;
    STACKTOP = sp;return ($412|0);
   }
  }
  $104 = ((($1)) + 40|0);
  $105 = HEAP32[$104>>2]|0;
  $106 = (FUNCTION_TABLE_iii[$105 & 8191]($face,$stream)|0);
  HEAP32[$error>>2] = $106;
  $107 = ($106|0)==(0);
  if ($107) {
   $has_outline$2 = $has_outline$145;
  } else {
   $108 = ((($face)) + 364|0);
   HEAP16[$108>>1] = -1;
   $has_outline$2 = $has_outline$145;
  }
 }
 $109 = ((($1)) + 96|0);
 $110 = HEAP32[$109>>2]|0;
 $111 = ($110|0)==(0|0);
 do {
  if (!($111)) {
   $112 = (FUNCTION_TABLE_iii[$110 & 8191]($face,$stream)|0);
   HEAP32[$error>>2] = $112;
   $113 = ($112|0)==(0);
   if (!($113)) {
    $114 = $112 & 255;
    $115 = ($114|0)==(142);
    if ($115) {
     HEAP32[$error>>2] = 0;
     break;
    } else {
     $412 = $112;
     STACKTOP = sp;return ($412|0);
    }
   }
  }
 } while(0);
 $116 = ((($1)) + 64|0);
 $117 = HEAP32[$116>>2]|0;
 $118 = (FUNCTION_TABLE_iii[$117 & 8191]($face,$stream)|0);
 HEAP32[$error>>2] = $118;
 $119 = ($118|0)==(0);
 do {
  if (!($119)) {
   $120 = $118 & 255;
   $121 = ($120|0)==(142);
   if ($121) {
    $122 = ((($face)) + 548|0);
    HEAP32[$122>>2] = 0;
    break;
   } else {
    $412 = $118;
    STACKTOP = sp;return ($412|0);
   }
  }
 } while(0);
 $123 = ((($1)) + 60|0);
 $124 = HEAP32[$123>>2]|0;
 (FUNCTION_TABLE_iii[$124 & 8191]($face,$stream)|0);
 $125 = ((($1)) + 56|0);
 $126 = HEAP32[$125>>2]|0;
 $127 = (FUNCTION_TABLE_iii[$126 & 8191]($face,$stream)|0);
 HEAP32[$error>>2] = $127;
 $128 = ((($face)) + 264|0);
 $129 = HEAP16[$128>>1]|0;
 $130 = $129&65535;
 $131 = ((($face)) + 16|0);
 HEAP32[$131>>2] = $130;
 $132 = ((($face)) + 20|0);
 HEAP32[$132>>2] = 0;
 $133 = ((($face)) + 24|0);
 HEAP32[$133>>2] = 0;
 $134 = ((($face)) + 364|0);
 $135 = HEAP16[$134>>1]|0;
 $136 = ($135<<16>>16)==(-1);
 if ($136) {
  label = 55;
 } else {
  $137 = ((($face)) + 428|0);
  $138 = HEAP16[$137>>1]|0;
  $139 = $138 & 256;
  $140 = ($139<<16>>16)==(0);
  if ($140) {
   label = 55;
  } else {
   $141 = ($ignore_preferred_family$0$lcssa<<24>>24)==(0);
   do {
    if ($141) {
     $142 = (_tt_face_get_name($face,16,$132)|0);
     HEAP32[$error>>2] = $142;
     $143 = ($142|0)==(0);
     if ($143) {
      $$pre = HEAP32[$132>>2]|0;
      $144 = ($$pre|0)==(0|0);
      if ($144) {
       label = 50;
       break;
      } else {
       break;
      }
     } else {
      $412 = $142;
      STACKTOP = sp;return ($412|0);
     }
    } else {
     label = 50;
    }
   } while(0);
   if ((label|0) == 50) {
    $145 = (_tt_face_get_name($face,1,$132)|0);
    HEAP32[$error>>2] = $145;
    $146 = ($145|0)==(0);
    if (!($146)) {
     $412 = $145;
     STACKTOP = sp;return ($412|0);
    }
   }
   $147 = ($ignore_preferred_subfamily$0$lcssa<<24>>24)==(0);
   if ($147) {
    $148 = (_tt_face_get_name($face,17,$133)|0);
    HEAP32[$error>>2] = $148;
    $149 = ($148|0)==(0);
    if (!($149)) {
     $412 = $148;
     STACKTOP = sp;return ($412|0);
    }
   }
   $150 = HEAP32[$133>>2]|0;
   $151 = ($150|0)==(0|0);
   if ($151) {
    $152 = (_tt_face_get_name($face,2,$133)|0);
    HEAP32[$error>>2] = $152;
    $153 = ($152|0)==(0);
    if (!($153)) {
     $412 = $152;
     STACKTOP = sp;return ($412|0);
    }
   }
  }
 }
 if ((label|0) == 55) {
  $154 = (_tt_face_get_name($face,21,$132)|0);
  HEAP32[$error>>2] = $154;
  $155 = ($154|0)==(0);
  if (!($155)) {
   $412 = $154;
   STACKTOP = sp;return ($412|0);
  }
  $156 = HEAP32[$132>>2]|0;
  $157 = ($156|0)!=(0|0);
  $158 = ($ignore_preferred_family$0$lcssa<<24>>24)!=(0);
  $or$cond = $158 | $157;
  do {
   if ($or$cond) {
    $162 = $156;
   } else {
    $159 = (_tt_face_get_name($face,16,$132)|0);
    HEAP32[$error>>2] = $159;
    $160 = ($159|0)==(0);
    if ($160) {
     $$pr49 = HEAP32[$132>>2]|0;
     $162 = $$pr49;
     break;
    } else {
     $412 = $159;
     STACKTOP = sp;return ($412|0);
    }
   }
  } while(0);
  $161 = ($162|0)==(0|0);
  if ($161) {
   $163 = (_tt_face_get_name($face,1,$132)|0);
   HEAP32[$error>>2] = $163;
   $164 = ($163|0)==(0);
   if (!($164)) {
    $412 = $163;
    STACKTOP = sp;return ($412|0);
   }
  }
  $165 = (_tt_face_get_name($face,22,$133)|0);
  HEAP32[$error>>2] = $165;
  $166 = ($165|0)==(0);
  if (!($166)) {
   $412 = $165;
   STACKTOP = sp;return ($412|0);
  }
  $167 = HEAP32[$133>>2]|0;
  $168 = ($167|0)!=(0|0);
  $169 = ($ignore_preferred_subfamily$0$lcssa<<24>>24)!=(0);
  $or$cond5 = $169 | $168;
  do {
   if ($or$cond5) {
    $173 = $167;
   } else {
    $170 = (_tt_face_get_name($face,17,$133)|0);
    HEAP32[$error>>2] = $170;
    $171 = ($170|0)==(0);
    if ($171) {
     $$pr51 = HEAP32[$133>>2]|0;
     $173 = $$pr51;
     break;
    } else {
     $412 = $170;
     STACKTOP = sp;return ($412|0);
    }
   }
  } while(0);
  $172 = ($173|0)==(0|0);
  if ($172) {
   $174 = (_tt_face_get_name($face,2,$133)|0);
   HEAP32[$error>>2] = $174;
   $175 = ($174|0)==(0);
   if (!($175)) {
    $412 = $174;
    STACKTOP = sp;return ($412|0);
   }
  }
 }
 $176 = ((($face)) + 8|0);
 $177 = HEAP32[$176>>2]|0;
 $178 = ((($face)) + 732|0);
 $179 = HEAP32[$178>>2]|0;
 $180 = $179 & -2;
 $switch = ($180|0)==(2);
 $181 = $177 | 16384;
 $$14 = $switch ? $181 : $177;
 $182 = ($has_outline$2<<24>>24)==(1);
 $183 = $182&1;
 $$flags$0 = $$14 | $183;
 $184 = $$flags$0 | 24;
 $185 = ($58|0)==(0);
 if ($185) {
  $186 = ((($face)) + 468|0);
  $187 = HEAP32[$186>>2]|0;
  $188 = ($187|0)==(196608);
  $189 = $$flags$0 | 536;
  $$9 = $188 ? $184 : $189;
  $flags$2 = $$9;
 } else {
  $flags$2 = $184;
 }
 $190 = ((($face)) + 480|0);
 $191 = HEAP32[$190>>2]|0;
 $192 = ($191|0)==(0);
 $193 = $flags$2 | 4;
 $flags$2$ = $192 ? $flags$2 : $193;
 $194 = ((($face)) + 292|0);
 $195 = HEAP8[$194>>0]|0;
 $196 = ($195<<24>>24)==(0);
 $197 = $flags$2$ | 32;
 $flags$4 = $196 ? $flags$2$ : $197;
 $198 = ((($face)) + 752|0);
 $199 = HEAP32[$198>>2]|0;
 $200 = ($199|0)==(0);
 $201 = $flags$4 | 64;
 $flags$4$ = $200 ? $flags$4 : $201;
 $face$idx18 = ((($face)) + 152|0);
 $face$idx18$val = HEAP16[$face$idx18>>1]|0;
 $face$idx19 = ((($face)) + 156|0);
 $face$idx19$val = HEAP32[$face$idx19>>2]|0;
 $202 = $face$idx18$val&65535;
 $203 = (($face$idx19$val) + ($202<<4)|0);
 $204 = ($face$idx18$val<<16>>16)==(0);
 L130: do {
  if ($204) {
   $flags$6 = $flags$4$;
  } else {
   $entry$01$i29 = $face$idx19$val;
   while(1) {
    $205 = HEAP32[$entry$01$i29>>2]|0;
    $206 = ($205|0)==(1735162214);
    if ($206) {
     $207 = ((($entry$01$i29)) + 12|0);
     $208 = HEAP32[$207>>2]|0;
     $209 = ($208|0)==(0);
     if (!($209)) {
      $entry$01$i37 = $face$idx19$val;
      break;
     }
    }
    $210 = ((($entry$01$i29)) + 16|0);
    $211 = ($210>>>0)<($203>>>0);
    if ($211) {
     $entry$01$i29 = $210;
    } else {
     $flags$6 = $flags$4$;
     break L130;
    }
   }
   while(1) {
    $212 = HEAP32[$entry$01$i37>>2]|0;
    $213 = ($212|0)==(1719034226);
    if ($213) {
     $214 = ((($entry$01$i37)) + 12|0);
     $215 = HEAP32[$214>>2]|0;
     $216 = ($215|0)==(0);
     if (!($216)) {
      $entry$01$i33 = $face$idx19$val;
      break;
     }
    }
    $217 = ((($entry$01$i37)) + 16|0);
    $218 = ($217>>>0)<($203>>>0);
    if ($218) {
     $entry$01$i37 = $217;
    } else {
     $flags$6 = $flags$4$;
     break L130;
    }
   }
   while(1) {
    $219 = HEAP32[$entry$01$i33>>2]|0;
    $220 = ($219|0)==(1735811442);
    if ($220) {
     $221 = ((($entry$01$i33)) + 12|0);
     $222 = HEAP32[$221>>2]|0;
     $223 = ($222|0)==(0);
     if (!($223)) {
      $$0$i35$ph = $entry$01$i33;
      break;
     }
    }
    $224 = ((($entry$01$i33)) + 16|0);
    $225 = ($224>>>0)<($203>>>0);
    if ($225) {
     $entry$01$i33 = $224;
    } else {
     $$0$i35$ph = 0;
     break;
    }
   }
   $226 = ($$0$i35$ph|0)==(0|0);
   $227 = $flags$4$ | 256;
   $flags$4$$ = $226 ? $flags$4$ : $227;
   $flags$6 = $flags$4$$;
  }
 } while(0);
 HEAP32[$176>>2] = $flags$6;
 if ($182) {
  $228 = HEAP16[$134>>1]|0;
  $229 = ($228<<16>>16)==(-1);
  if ($229) {
   label = 83;
  } else {
   $230 = ((($face)) + 428|0);
   $231 = HEAP16[$230>>1]|0;
   $232 = $231&65535;
   $233 = $232 & 512;
   $234 = ($233|0)==(0);
   $235 = $232 & 1;
   $$62 = $234 ? $235 : 1;
   $236 = ($231&65535) >>> 4;
   $237 = $236 & 2;
   $238 = $237&65535;
   $239 = $238 | $$62;
   $flags$9 = $239;
  }
 } else {
  label = 83;
 }
 if ((label|0) == 83) {
  $240 = ((($face)) + 204|0);
  $241 = HEAP16[$240>>1]|0;
  $242 = $241 & 1;
  $243 = $242&65535;
  $244 = $243 << 1;
  $245 = ($241&65535) >>> 1;
  $246 = $245 & 1;
  $247 = $246&65535;
  $248 = $247 | $244;
  $flags$9 = $248;
 }
 $249 = ((($face)) + 12|0);
 HEAP32[$249>>2] = $flags$9;
 _tt_face_build_cmaps($face);
 $250 = ((($face)) + 36|0);
 $251 = HEAP32[$250>>2]|0;
 $252 = ($251|0)>(0);
 if ($252) {
  $253 = ((($face)) + 40|0);
  $254 = HEAP32[$253>>2]|0;
  $m$067 = 0;
  while(1) {
   $255 = (($254) + ($m$067<<2)|0);
   $256 = HEAP32[$255>>2]|0;
   $257 = ((($256)) + 8|0);
   $258 = HEAP32[$257>>2]|0;
   $259 = $258&65535;
   $260 = $258 >>> 16;
   $261 = $260&65535;
   $262 = ($259<<16>>16)==(2);
   L158: do {
    if ($262) {
     $cur$02$lcssa$i = 55152;
     label = 87;
    } else {
     $265 = $258 & 65535;
     switch ($265|0) {
     case 0:  {
      $cur$02$lcssa$i = (55164);
      label = 87;
      break L158;
      break;
     }
     case 1:  {
      $266 = ($261<<16>>16)==(0);
      if ($266) {
       $cur$02$lcssa$i = (55176);
       label = 87;
       break L158;
      } else {
       $$0$i28 = 0;
       break L158;
      }
      break;
     }
     case 3:  {
      break;
     }
     default: {
      $$0$i28 = 0;
      break L158;
     }
     }
     $267 = ($261<<16>>16)==(0);
     if ($267) {
      $cur$02$lcssa$i = (55188);
      label = 87;
     } else {
      $268 = ($261<<16>>16)==(10);
      if ($268) {
       $cur$02$lcssa$i = (55200);
       label = 87;
      } else {
       $269 = ($261<<16>>16)==(1);
       if ($269) {
        $cur$02$lcssa$i = (55212);
        label = 87;
       } else {
        $270 = ($261<<16>>16)==(2);
        if ($270) {
         $cur$02$lcssa$i = (55224);
         label = 87;
        } else {
         $271 = ($261<<16>>16)==(3);
         if ($271) {
          $cur$02$lcssa$i = (55236);
          label = 87;
          break;
         }
         $272 = ($261<<16>>16)==(4);
         if ($272) {
          $cur$02$lcssa$i = (55248);
          label = 87;
          break;
         }
         $273 = ($261<<16>>16)==(5);
         if ($273) {
          $cur$02$lcssa$i = (55260);
          label = 87;
          break;
         }
         $274 = ($261<<16>>16)==(6);
         if ($274) {
          $cur$02$lcssa$i = (55272);
          label = 87;
         } else {
          $$0$i28 = 0;
         }
        }
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 87) {
    label = 0;
    $263 = ((($cur$02$lcssa$i)) + 8|0);
    $264 = HEAP32[$263>>2]|0;
    $$0$i28 = $264;
   }
   $275 = ((($256)) + 4|0);
   HEAP32[$275>>2] = $$0$i28;
   $276 = (($m$067) + 1)|0;
   $277 = ($276|0)<($251|0);
   if ($277) {
    $m$067 = $276;
   } else {
    break;
   }
  }
 }
 $278 = ((($face)) + 736|0);
 $279 = HEAP32[$278>>2]|0;
 $280 = ($279|0)==(0);
 L174: do {
  if ($280) {
   $$pre83 = HEAP32[$176>>2]|0;
   $330 = $$pre83;
  } else {
   $281 = ((($face)) + 104|0);
   $282 = HEAP32[$281>>2]|0;
   $283 = ((($282)) + 28|0);
   $284 = HEAP32[$283>>2]|0;
   $285 = HEAP16[$47>>1]|0;
   $286 = ((($face)) + 366|0);
   $287 = HEAP16[$286>>1]|0;
   $288 = ($285<<16>>16)==(0);
   if ($288) {
    label = 103;
   } else {
    $289 = HEAP16[$134>>1]|0;
    $290 = ($289<<16>>16)==(-1);
    if ($290) {
     label = 103;
    } else {
     $avgwidth$0 = $287;$em_size$0 = $285;
    }
   }
   if ((label|0) == 103) {
    $avgwidth$0 = 1;$em_size$0 = 1;
   }
   $291 = (_ft_mem_realloc($284,16,0,$279,0,$error)|0);
   $292 = ((($face)) + 32|0);
   HEAP32[$292>>2] = $291;
   $293 = HEAP32[$error>>2]|0;
   $294 = ($293|0)==(0);
   L182: do {
    if ($294) {
     $295 = ((($1)) + 108|0);
     $296 = ((($metrics)) + 20|0);
     $297 = $avgwidth$0 << 16 >> 16;
     $298 = $em_size$0&65535;
     $div = ($em_size$0&65535) >>> 1;
     $299 = $div&65535;
     $300 = HEAP32[$295>>2]|0;
     $301 = (FUNCTION_TABLE_iiii[$300 & 8191]($face,0,$metrics)|0);
     HEAP32[$error>>2] = $301;
     $302 = ($301|0)==(0);
     if ($302) {
      $307 = $291;$i1$066100 = 0;
      while(1) {
       $303 = HEAP32[$296>>2]|0;
       $304 = $303 >>> 6;
       $305 = $304&65535;
       $306 = (($307) + ($i1$066100<<4)|0);
       HEAP16[$306>>1] = $305;
       $308 = HEAP32[$metrics>>2]|0;
       $309 = $308 & 65535;
       $310 = Math_imul($309, $297)|0;
       $311 = (($310) + ($299))|0;
       $312 = (($311|0) / ($298|0))&-1;
       $313 = $312&65535;
       $314 = (((($307) + ($i1$066100<<4)|0)) + 2|0);
       HEAP16[$314>>1] = $313;
       $315 = $309 << 6;
       $316 = (((($307) + ($i1$066100<<4)|0)) + 8|0);
       HEAP32[$316>>2] = $315;
       $317 = $308 >>> 16;
       $318 = $317 << 6;
       $319 = (((($307) + ($i1$066100<<4)|0)) + 12|0);
       HEAP32[$319>>2] = $318;
       $320 = (((($307) + ($i1$066100<<4)|0)) + 4|0);
       HEAP32[$320>>2] = $318;
       $321 = (($i1$066100) + 1)|0;
       $322 = ($321>>>0)<($279>>>0);
       if (!($322)) {
        break;
       }
       $$pre82 = HEAP32[$292>>2]|0;
       $323 = HEAP32[$295>>2]|0;
       $324 = (FUNCTION_TABLE_iiii[$323 & 8191]($face,$321,$metrics)|0);
       HEAP32[$error>>2] = $324;
       $325 = ($324|0)==(0);
       if ($325) {
        $307 = $$pre82;$i1$066100 = $321;
       } else {
        $415 = $324;
        break L182;
       }
      }
      $326 = HEAP32[$176>>2]|0;
      $327 = $326 | 2;
      HEAP32[$176>>2] = $327;
      $328 = ((($face)) + 28|0);
      HEAP32[$328>>2] = $279;
      $330 = $327;
      break L174;
     } else {
      $415 = $301;
     }
    } else {
     $415 = $293;
    }
   } while(0);
   $412 = $415;
   STACKTOP = sp;return ($412|0);
  }
 } while(0);
 $329 = $330 & 3;
 $331 = ($329|0)==(0);
 if ($331) {
  $332 = $330 | 1;
  HEAP32[$176>>2] = $332;
  $334 = $332;
 } else {
  $334 = $330;
 }
 $333 = $334 & 1;
 $335 = ($333|0)==(0);
 if ($335) {
  $412 = 0;
  STACKTOP = sp;return ($412|0);
 }
 $336 = ((($face)) + 196|0);
 $337 = HEAP32[$336>>2]|0;
 $sext = $337 << 16;
 $338 = $sext >> 16;
 $339 = ((($face)) + 52|0);
 HEAP32[$339>>2] = $338;
 $340 = $337 >> 16;
 $341 = ((($face)) + 56|0);
 HEAP32[$341>>2] = $340;
 $342 = ((($face)) + 200|0);
 $343 = HEAP32[$342>>2]|0;
 $sext87 = $343 << 16;
 $344 = $sext87 >> 16;
 $345 = ((($face)) + 60|0);
 HEAP32[$345>>2] = $344;
 $346 = $343 >> 16;
 $347 = ((($face)) + 64|0);
 HEAP32[$347>>2] = $346;
 $348 = HEAP16[$47>>1]|0;
 $349 = ((($face)) + 68|0);
 HEAP16[$349>>1] = $348;
 $350 = ((($face)) + 220|0);
 $351 = HEAP32[$350>>2]|0;
 $352 = $351&65535;
 $353 = ((($face)) + 70|0);
 HEAP16[$353>>1] = $352;
 $354 = $351 >>> 16;
 $355 = $354&65535;
 $356 = ((($face)) + 72|0);
 HEAP16[$356>>1] = $355;
 $357 = (($351) - ($354))|0;
 $358 = ((($face)) + 224|0);
 $359 = HEAP32[$358>>2]|0;
 $360 = (($357) + ($359))|0;
 $361 = $360&65535;
 $362 = ((($face)) + 74|0);
 HEAP16[$362>>1] = $361;
 $363 = $359 >>> 16;
 $364 = $363&65535;
 $365 = $351 | $354;
 $366 = $365&65535;
 $367 = ($366<<16>>16)==(0);
 do {
  if ($367) {
   $368 = HEAP16[$134>>1]|0;
   $369 = ($368<<16>>16)==(-1);
   if ($369) {
    $397 = $360;
   } else {
    $370 = ((($face)) + 434|0);
    $371 = HEAP16[$370>>1]|0;
    $372 = ((($face)) + 436|0);
    $373 = HEAP32[$372>>2]|0;
    $374 = $373&65535;
    $375 = $371 | $374;
    $376 = ($375<<16>>16)==(0);
    if ($376) {
     $382 = ((($face)) + 440|0);
     $383 = HEAP32[$382>>2]|0;
     $384 = $383&65535;
     HEAP16[$353>>1] = $384;
     $385 = $383 >>> 16;
     $386 = (0 - ($385))|0;
     $387 = $386&65535;
     HEAP16[$356>>1] = $387;
     $388 = $383 & 65535;
     $389 = (($388) + ($385))|0;
     $390 = $389&65535;
     HEAP16[$362>>1] = $390;
     $397 = $389;
     break;
    } else {
     $$in = $373 >>> 16;
     HEAP16[$353>>1] = $371;
     HEAP16[$356>>1] = $374;
     $377 = $371&65535;
     $378 = $373 & 65535;
     $379 = (($377) - ($378))|0;
     $380 = (($379) + ($$in))|0;
     $381 = $380&65535;
     HEAP16[$362>>1] = $381;
     $397 = $380;
     break;
    }
   }
  } else {
   $397 = $360;
  }
 } while(0);
 $391 = ((($face)) + 76|0);
 HEAP16[$391>>1] = $364;
 $392 = HEAP8[$194>>0]|0;
 $393 = ($392<<24>>24)==(0);
 if ($393) {
  $sext90 = $397 << 16;
  $398 = $sext90 >> 16;
  $400 = $398;
 } else {
  $394 = ((($face)) + 306|0);
  $395 = HEAP16[$394>>1]|0;
  $396 = $395&65535;
  $400 = $396;
 }
 $399 = $400&65535;
 $401 = ((($face)) + 78|0);
 HEAP16[$401>>1] = $399;
 $402 = ((($face)) + 476|0);
 $403 = HEAP32[$402>>2]|0;
 $404 = $403 >>> 16;
 $405 = $404&65535;
 $406 = $403 >> 16;
 $407 = (($406|0) / 2)&-1;
 $408 = (($403) - ($407))|0;
 $409 = $408&65535;
 $410 = ((($face)) + 80|0);
 HEAP16[$410>>1] = $409;
 $411 = ((($face)) + 82|0);
 HEAP16[$411>>1] = $405;
 $412 = 0;
 STACKTOP = sp;return ($412|0);
}
function _tt_face_get_name($face,$nameid,$name) {
 $face = $face|0;
 $nameid = $nameid|0;
 $name = $name|0;
 var $$found_apple_roman$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $convert$0 = 0, $error = 0, $found_apple_english$0$found_apple_roman$0 = 0, $found_apple_english$011 = 0, $found_apple_english$1 = 0, $found_apple_english$1$lcssa = 0, $found_apple_roman$010 = 0;
 var $found_apple_roman$1 = 0, $found_apple_roman$1$lcssa = 0, $found_unicode$013 = 0, $found_unicode$1 = 0, $found_unicode$1$lcssa = 0, $found_win$012 = 0, $found_win$1 = 0, $found_win$1$lcssa = 0, $indvars$iv = 0, $indvars$iv$next = 0, $is_english$014 = 0, $is_english$1 = 0, $is_english$1$lcssa = 0, $or$cond = 0, $rec$08 = 0, $rec$1 = 0, $result$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($face)) + 100|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$error>>2] = 0;
 $2 = ((($face)) + 356|0);
 $3 = ((($face)) + 340|0);
 $4 = HEAP16[$3>>1]|0;
 $5 = ($4<<16>>16)==(0);
 if ($5) {
  $result$1 = 0;
  HEAP32[$name>>2] = $result$1;
  $71 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($71|0);
 }
 $6 = HEAP32[$2>>2]|0;
 $7 = $4&65535;
 $found_apple_english$011 = -1;$found_apple_roman$010 = -1;$found_unicode$013 = -1;$found_win$012 = -1;$indvars$iv = 0;$is_english$014 = 0;$rec$08 = $6;
 while(1) {
  $8 = ((($rec$08)) + 6|0);
  $9 = HEAP16[$8>>1]|0;
  $10 = ($9<<16>>16)==($nameid<<16>>16);
  L6: do {
   if ($10) {
    $11 = ((($rec$08)) + 8|0);
    $12 = HEAP16[$11>>1]|0;
    $13 = ($12<<16>>16)==(0);
    if ($13) {
     $found_apple_english$1 = $found_apple_english$011;$found_apple_roman$1 = $found_apple_roman$010;$found_unicode$1 = $found_unicode$013;$found_win$1 = $found_win$012;$is_english$1 = $is_english$014;
    } else {
     $14 = HEAP32[$rec$08>>2]|0;
     $15 = $14 & 65535;
     $16 = $14 >>> 16;
     switch ($15|0) {
     case 2: case 0:  {
      $found_apple_english$1 = $found_apple_english$011;$found_apple_roman$1 = $found_apple_roman$010;$found_unicode$1 = $indvars$iv;$found_win$1 = $found_win$012;$is_english$1 = $is_english$014;
      break L6;
      break;
     }
     case 1:  {
      $17 = ((($rec$08)) + 4|0);
      $18 = HEAP16[$17>>1]|0;
      $19 = ($18<<16>>16)==(0);
      if ($19) {
       $found_apple_english$1 = $indvars$iv;$found_apple_roman$1 = $found_apple_roman$010;$found_unicode$1 = $found_unicode$013;$found_win$1 = $found_win$012;$is_english$1 = $is_english$014;
       break L6;
      }
      $20 = ($16|0)==(0);
      $$found_apple_roman$0 = $20 ? $indvars$iv : $found_apple_roman$010;
      $found_apple_english$1 = $found_apple_english$011;$found_apple_roman$1 = $$found_apple_roman$0;$found_unicode$1 = $found_unicode$013;$found_win$1 = $found_win$012;$is_english$1 = $is_english$014;
      break L6;
      break;
     }
     case 3:  {
      $21 = ($found_win$012|0)==(-1);
      if (!($21)) {
       $22 = ((($rec$08)) + 4|0);
       $23 = HEAP16[$22>>1]|0;
       $24 = $23 & 1023;
       $25 = ($24<<16>>16)==(9);
       if (!($25)) {
        $found_apple_english$1 = $found_apple_english$011;$found_apple_roman$1 = $found_apple_roman$010;$found_unicode$1 = $found_unicode$013;$found_win$1 = $found_win$012;$is_english$1 = $is_english$014;
        break L6;
       }
      }
      switch ($16|0) {
      case 10: case 1: case 0:  {
       break;
      }
      default: {
       $found_apple_english$1 = $found_apple_english$011;$found_apple_roman$1 = $found_apple_roman$010;$found_unicode$1 = $found_unicode$013;$found_win$1 = $found_win$012;$is_english$1 = $is_english$014;
       break L6;
      }
      }
      $26 = ((($rec$08)) + 4|0);
      $27 = HEAP16[$26>>1]|0;
      $28 = $27 & 1023;
      $29 = ($28<<16>>16)==(9);
      $30 = $29&1;
      $found_apple_english$1 = $found_apple_english$011;$found_apple_roman$1 = $found_apple_roman$010;$found_unicode$1 = $found_unicode$013;$found_win$1 = $indvars$iv;$is_english$1 = $30;
      break L6;
      break;
     }
     default: {
      $found_apple_english$1 = $found_apple_english$011;$found_apple_roman$1 = $found_apple_roman$010;$found_unicode$1 = $found_unicode$013;$found_win$1 = $found_win$012;$is_english$1 = $is_english$014;
      break L6;
     }
     }
    }
   } else {
    $found_apple_english$1 = $found_apple_english$011;$found_apple_roman$1 = $found_apple_roman$010;$found_unicode$1 = $found_unicode$013;$found_win$1 = $found_win$012;$is_english$1 = $is_english$014;
   }
  } while(0);
  $indvars$iv$next = (($indvars$iv) + 1)|0;
  $31 = ((($rec$08)) + 20|0);
  $32 = ($indvars$iv$next>>>0)<($7>>>0);
  if ($32) {
   $found_apple_english$011 = $found_apple_english$1;$found_apple_roman$010 = $found_apple_roman$1;$found_unicode$013 = $found_unicode$1;$found_win$012 = $found_win$1;$indvars$iv = $indvars$iv$next;$is_english$014 = $is_english$1;$rec$08 = $31;
  } else {
   $found_apple_english$1$lcssa = $found_apple_english$1;$found_apple_roman$1$lcssa = $found_apple_roman$1;$found_unicode$1$lcssa = $found_unicode$1;$found_win$1$lcssa = $found_win$1;$is_english$1$lcssa = $is_english$1;
   break;
  }
 }
 $33 = ($found_apple_english$1$lcssa|0)>(-1);
 $found_apple_english$0$found_apple_roman$0 = $33 ? $found_apple_english$1$lcssa : $found_apple_roman$1$lcssa;
 $34 = ($found_win$1$lcssa|0)>(-1);
 L20: do {
  if ($34) {
   $35 = ($found_apple_english$0$found_apple_roman$0|0)<(0);
   $36 = ($is_english$1$lcssa<<24>>24)!=(0);
   $or$cond = $35 | $36;
   if ($or$cond) {
    $37 = HEAP32[$2>>2]|0;
    $38 = (($37) + (($found_win$1$lcssa*20)|0)|0);
    $39 = (((($37) + (($found_win$1$lcssa*20)|0)|0)) + 2|0);
    $40 = HEAP16[$39>>1]|0;
    $41 = $40&65535;
    switch ($41|0) {
    case 10: case 0: case 1:  {
     $convert$0 = 4394;$rec$1 = $38;
     break L20;
     break;
    }
    default: {
     $result$1 = 0;
    }
    }
    HEAP32[$name>>2] = $result$1;
    $71 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($71|0);
   } else {
    label = 17;
   }
  } else {
   label = 17;
  }
 } while(0);
 do {
  if ((label|0) == 17) {
   $42 = ($found_apple_english$0$found_apple_roman$0|0)>(-1);
   if ($42) {
    $43 = HEAP32[$2>>2]|0;
    $44 = (($43) + (($found_apple_english$0$found_apple_roman$0*20)|0)|0);
    $convert$0 = 4395;$rec$1 = $44;
    break;
   }
   $45 = ($found_unicode$1$lcssa|0)>(-1);
   if ($45) {
    $46 = HEAP32[$2>>2]|0;
    $47 = (($46) + (($found_unicode$1$lcssa*20)|0)|0);
    $convert$0 = 4394;$rec$1 = $47;
    break;
   } else {
    $result$1 = 0;
    HEAP32[$name>>2] = $result$1;
    $71 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($71|0);
   }
  }
 } while(0);
 $48 = ($rec$1|0)==(0|0);
 if ($48) {
  $result$1 = 0;
  HEAP32[$name>>2] = $result$1;
  $71 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($71|0);
 }
 $49 = ((($rec$1)) + 16|0);
 $50 = HEAP32[$49>>2]|0;
 $51 = ($50|0)==(0|0);
 do {
  if ($51) {
   $52 = ((($face)) + 360|0);
   $53 = HEAP32[$52>>2]|0;
   $54 = ((($rec$1)) + 8|0);
   $55 = HEAP16[$54>>1]|0;
   $56 = $55&65535;
   $57 = (_ft_mem_realloc($1,1,0,$56,0,$error)|0);
   HEAP32[$49>>2] = $57;
   $58 = HEAP32[$error>>2]|0;
   $59 = ($58|0)==(0);
   if ($59) {
    $60 = ((($rec$1)) + 12|0);
    $61 = HEAP32[$60>>2]|0;
    $62 = (_FT_Stream_Seek($53,$61)|0);
    HEAP32[$error>>2] = $62;
    $63 = ($62|0)==(0);
    if ($63) {
     $64 = HEAP32[$49>>2]|0;
     $65 = HEAP16[$54>>1]|0;
     $66 = $65&65535;
     $67 = (_FT_Stream_Read($53,$64,$66)|0);
     HEAP32[$error>>2] = $67;
     $68 = ($67|0)==(0);
     if ($68) {
      break;
     }
    }
   }
   $69 = HEAP32[$49>>2]|0;
   _ft_mem_free($1,$69);
   HEAP32[$49>>2] = 0;
   HEAP16[$54>>1] = 0;
   $result$1 = 0;
   HEAP32[$name>>2] = $result$1;
   $71 = HEAP32[$error>>2]|0;
   STACKTOP = sp;return ($71|0);
  }
 } while(0);
 $70 = (FUNCTION_TABLE_iii[$convert$0 & 8191]($rec$1,$1)|0);
 $result$1 = $70;
 HEAP32[$name>>2] = $result$1;
 $71 = HEAP32[$error>>2]|0;
 STACKTOP = sp;return ($71|0);
}
function _tt_name_entry_ascii_from_other($entry,$memory) {
 $entry = $entry|0;
 $memory = $memory|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $error = 0, $n$0$lcssa = 0, $n$02 = 0, $or$cond = 0, $read$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($entry)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($entry)) + 8|0);
 $3 = HEAP16[$2>>1]|0;
 $4 = $3&65535;
 $5 = (($4) + 1)|0;
 $6 = (_ft_mem_realloc($memory,1,0,$5,0,$error)|0);
 $7 = HEAP32[$error>>2]|0;
 $8 = ($7|0)==(0);
 if (!($8)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $9 = ($3<<16>>16)==(0);
 L4: do {
  if ($9) {
   $n$0$lcssa = 0;
  } else {
   $n$02 = 0;$read$03 = $1;
   while(1) {
    $10 = HEAP8[$read$03>>0]|0;
    $11 = ($10<<24>>24)==(0);
    if ($11) {
     $n$0$lcssa = $n$02;
     break L4;
    }
    $12 = ((($read$03)) + 1|0);
    $13 = ($10&255)<(32);
    $14 = ($10<<24>>24)<(0);
    $or$cond = $13 | $14;
    $15 = $or$cond ? 63 : $10;
    $16 = (($6) + ($n$02)|0);
    HEAP8[$16>>0] = $15;
    $17 = (($n$02) + 1)|0;
    $18 = ($17>>>0)<($4>>>0);
    if ($18) {
     $n$02 = $17;$read$03 = $12;
    } else {
     $n$0$lcssa = $17;
     break;
    }
   }
  }
 } while(0);
 $19 = (($6) + ($n$0$lcssa)|0);
 HEAP8[$19>>0] = 0;
 $$0 = $6;
 STACKTOP = sp;return ($$0|0);
}
function _tt_name_entry_ascii_from_utf16($entry,$memory) {
 $entry = $entry|0;
 $memory = $memory|0;
 var $$0 = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $div = 0, $error = 0, $n$0$lcssa = 0, $n$02 = 0, $read$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($entry)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($entry)) + 8|0);
 $3 = HEAP16[$2>>1]|0;
 $div = ($3&65535) >>> 1;
 $4 = $div&65535;
 $5 = (($4) + 1)|0;
 $6 = (_ft_mem_realloc($memory,1,0,$5,0,$error)|0);
 $7 = HEAP32[$error>>2]|0;
 $8 = ($7|0)==(0);
 if (!($8)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $9 = ($div<<16>>16)==(0);
 L4: do {
  if ($9) {
   $n$0$lcssa = 0;
  } else {
   $n$02 = 0;$read$03 = $1;
   while(1) {
    $10 = HEAP8[$read$03>>0]|0;
    $11 = $10&255;
    $12 = $11 << 8;
    $13 = ((($read$03)) + 1|0);
    $14 = HEAP8[$13>>0]|0;
    $15 = $14&255;
    $16 = $12 | $15;
    $17 = ($16|0)==(0);
    if ($17) {
     $n$0$lcssa = $n$02;
     break L4;
    }
    $18 = ((($read$03)) + 2|0);
    $$off = (($16) + -32)|0;
    $19 = ($$off>>>0)>(95);
    $20 = $19 ? 63 : $14;
    $21 = (($6) + ($n$02)|0);
    HEAP8[$21>>0] = $20;
    $22 = (($n$02) + 1)|0;
    $23 = ($22>>>0)<($4>>>0);
    if ($23) {
     $n$02 = $22;$read$03 = $18;
    } else {
     $n$0$lcssa = $22;
     break;
    }
   }
  }
 } while(0);
 $24 = (($6) + ($n$0$lcssa)|0);
 HEAP8[$24>>0] = 0;
 $$0 = $6;
 STACKTOP = sp;return ($$0|0);
}
function _tt_face_build_cmaps($face) {
 $face = $face|0;
 var $$reg2mem82$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $charmap = 0, $clazz$sroa$0 = 0, $cmap = 0, $cond = 0, $cond84 = 0, $cond85 = 0, $cond86 = 0, $error = 0, $format = 0, $num_cmaps = 0, $p = 0, $pclazz = 0, $ttcmap = 0, $valid = 0, _setjmpTable = 0, _setjmpTableSize = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 _setjmpTableSize = 4;_setjmpTable = _malloc(40) | 0;HEAP32[_setjmpTable>>2]=0;
 $num_cmaps = sp + 216|0;
 $p = sp + 212|0;
 $charmap = sp + 200|0;
 $cmap = sp + 196|0;
 $format = sp + 192|0;
 $pclazz = sp + 188|0;
 $clazz$sroa$0 = sp + 184|0;
 $valid = sp + 8|0;
 $error = sp + 4|0;
 $ttcmap = sp;
 $0 = ((($face)) + 500|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 504|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 HEAP32[$p>>2] = $1;
 $5 = HEAP32[$p>>2]|0;
 $6 = ($5|0)==(0|0);
 if ($6) {
  _free(_setjmpTable|0);
  STACKTOP = sp;return;
 }
 $7 = HEAP32[$p>>2]|0;
 $8 = ((($7)) + 4|0);
 $9 = ($8>>>0)>($4>>>0);
 if ($9) {
  _free(_setjmpTable|0);
  STACKTOP = sp;return;
 }
 $10 = HEAP32[$p>>2]|0;
 $11 = ((($10)) + 2|0);
 HEAP32[$p>>2] = $11;
 $12 = HEAP32[$p>>2]|0;
 $13 = ((($12)) + -2|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = $14&255;
 $16 = $15 << 8;
 $17 = HEAP32[$p>>2]|0;
 $18 = ((($17)) + -1|0);
 $19 = HEAP8[$18>>0]|0;
 $20 = $19&255;
 $21 = $20 | $16;
 $22 = ($21|0)==(0);
 if (!($22)) {
  _free(_setjmpTable|0);
  STACKTOP = sp;return;
 }
 $23 = HEAP32[$p>>2]|0;
 $24 = ((($23)) + 2|0);
 HEAP32[$p>>2] = $24;
 $25 = HEAP32[$p>>2]|0;
 $26 = ((($25)) + -2|0);
 $27 = HEAP8[$26>>0]|0;
 $28 = $27&255;
 $29 = $28 << 8;
 $30 = HEAP32[$p>>2]|0;
 $31 = ((($30)) + -1|0);
 $32 = HEAP8[$31>>0]|0;
 $33 = $32&255;
 $34 = $33 | $29;
 HEAP32[$num_cmaps>>2] = $34;
 $35 = HEAP32[$num_cmaps>>2]|0;
 $36 = ($35|0)==(0);
 if ($36) {
  _free(_setjmpTable|0);
  STACKTOP = sp;return;
 }
 $37 = ((($charmap)) + 8|0);
 $38 = ((($charmap)) + 10|0);
 $39 = ((($charmap)) + 4|0);
 $40 = ((($face)) + 264|0);
 $41 = ((($valid)) + 172|0);
 $42 = ((($valid)) + 168|0);
 while(1) {
  $43 = HEAP32[$p>>2]|0;
  $44 = ((($43)) + 8|0);
  $45 = ($44>>>0)>($4>>>0);
  if ($45) {
   label = 26;
   break;
  }
  $46 = HEAP32[$p>>2]|0;
  $47 = ((($46)) + 2|0);
  HEAP32[$p>>2] = $47;
  $48 = HEAP32[$p>>2]|0;
  $49 = ((($48)) + -2|0);
  $50 = HEAP8[$49>>0]|0;
  $51 = $50&255;
  $52 = $51 << 8;
  $53 = HEAP32[$p>>2]|0;
  $54 = ((($53)) + -1|0);
  $55 = HEAP8[$54>>0]|0;
  $56 = $55&255;
  $57 = $56 | $52;
  $58 = $57&65535;
  HEAP16[$37>>1] = $58;
  $59 = HEAP32[$p>>2]|0;
  $60 = ((($59)) + 2|0);
  HEAP32[$p>>2] = $60;
  $61 = HEAP32[$p>>2]|0;
  $62 = ((($61)) + -2|0);
  $63 = HEAP8[$62>>0]|0;
  $64 = $63&255;
  $65 = $64 << 8;
  $66 = HEAP32[$p>>2]|0;
  $67 = ((($66)) + -1|0);
  $68 = HEAP8[$67>>0]|0;
  $69 = $68&255;
  $70 = $69 | $65;
  $71 = $70&65535;
  HEAP16[$38>>1] = $71;
  HEAP32[$charmap>>2] = $face;
  HEAP32[$39>>2] = 0;
  $72 = HEAP32[$p>>2]|0;
  $73 = ((($72)) + 4|0);
  HEAP32[$p>>2] = $73;
  $74 = HEAP32[$p>>2]|0;
  $75 = ((($74)) + -4|0);
  $76 = HEAP8[$75>>0]|0;
  $77 = $76&255;
  $78 = $77 << 24;
  $79 = HEAP32[$p>>2]|0;
  $80 = ((($79)) + -3|0);
  $81 = HEAP8[$80>>0]|0;
  $82 = $81&255;
  $83 = $82 << 16;
  $84 = $83 | $78;
  $85 = HEAP32[$p>>2]|0;
  $86 = ((($85)) + -2|0);
  $87 = HEAP8[$86>>0]|0;
  $88 = $87&255;
  $89 = $88 << 8;
  $90 = $84 | $89;
  $91 = HEAP32[$p>>2]|0;
  $92 = ((($91)) + -1|0);
  $93 = HEAP8[$92>>0]|0;
  $94 = $93&255;
  $95 = $90 | $94;
  $96 = ($95|0)==(0);
  if (!($96)) {
   $97 = HEAP32[$2>>2]|0;
   $98 = (($97) + -2)|0;
   $99 = ($95>>>0)>($98>>>0);
   if (!($99)) {
    $100 = (($1) + ($95)|0);
    HEAP32[$cmap>>2] = $100;
    $101 = HEAP32[$cmap>>2]|0;
    $102 = HEAP8[$101>>0]|0;
    $103 = $102&255;
    $104 = $103 << 8;
    $105 = HEAP32[$cmap>>2]|0;
    $106 = ((($105)) + 1|0);
    $107 = HEAP8[$106>>0]|0;
    $108 = $107&255;
    $109 = $108 | $104;
    HEAP32[$format>>2] = $109;
    HEAP32[$pclazz>>2] = 55112;
    $110 = HEAP32[$pclazz>>2]|0;
    $111 = HEAP32[$110>>2]|0;
    $112 = ($111|0)==(0|0);
    L19: do {
     if (!($112)) {
      while(1) {
       $113 = HEAP32[$pclazz>>2]|0;
       $114 = HEAP32[$113>>2]|0;
       HEAP32[$clazz$sroa$0>>2] = $114;
       $115 = HEAP32[$clazz$sroa$0>>2]|0;
       $116 = $115;
       $117 = ((($116)) + 40|0);
       $118 = HEAP32[$117>>2]|0;
       $119 = HEAP32[$format>>2]|0;
       $120 = ($118|0)==($119|0);
       if ($120) {
        break;
       }
       $153 = HEAP32[$pclazz>>2]|0;
       $154 = ((($153)) + 4|0);
       HEAP32[$pclazz>>2] = $154;
       $155 = HEAP32[$pclazz>>2]|0;
       $156 = HEAP32[$155>>2]|0;
       $157 = ($156|0)==(0|0);
       if ($157) {
        break L19;
       }
      }
      HEAP32[$error>>2] = 0;
      $121 = HEAP32[$cmap>>2]|0;
      __THREW__ = 0;
      invoke_viiii(4396,($valid|0),($121|0),($4|0),0);
      $122 = __THREW__; __THREW__ = 0;
      if ((($122|0) != 0) & ((threwValue|0) != 0)) { $123 = _testSetjmp(HEAP32[$122>>2]|0, _setjmpTable|0, _setjmpTableSize|0)|0; if (($123|0) == 0) { _longjmp($122|0, threwValue|0); } tempRet0 = (threwValue); } else { $123 = -1; };
      $124 = tempRet0;
      $cond = ($123|0)==(1);
      if ($cond) {
       $$reg2mem82$0 = $124;
      } else {
       $125 = HEAP16[$40>>1]|0;
       $126 = $125&65535;
       HEAP32[$41>>2] = $126;
       _setjmpTable = _saveSetjmp($valid,1,_setjmpTable|0,_setjmpTableSize|0)|0;_setjmpTableSize = tempRet0;
       __THREW__ = 0;
       $127 = __THREW__; __THREW__ = 0;
       if ((($127|0) != 0) & ((threwValue|0) != 0)) { $128 = _testSetjmp(HEAP32[$127>>2]|0, _setjmpTable|0, _setjmpTableSize|0)|0; if (($128|0) == 0) { _longjmp($127|0, threwValue|0); } tempRet0 = (threwValue); } else { $128 = -1; };
       $129 = tempRet0;
       $cond84 = ($128|0)==(1);
       if ($cond84) {
        $$reg2mem82$0 = $129;
       } else {
        $$reg2mem82$0 = 0;
       }
      }
      while(1) {
       $130 = ($$reg2mem82$0|0)==(0);
       if ($130) {
        $131 = HEAP32[$clazz$sroa$0>>2]|0;
        $132 = $131;
        $133 = ((($132)) + 44|0);
        $134 = HEAP32[$133>>2]|0;
        $135 = HEAP32[$cmap>>2]|0;
        __THREW__ = 0;
        $136 = (invoke_iii($134|0,($135|0),($valid|0))|0);
        $137 = __THREW__; __THREW__ = 0;
        if ((($137|0) != 0) & ((threwValue|0) != 0)) { $138 = _testSetjmp(HEAP32[$137>>2]|0, _setjmpTable|0, _setjmpTableSize|0)|0; if (($138|0) == 0) { _longjmp($137|0, threwValue|0); } tempRet0 = (threwValue); } else { $138 = -1; };
        $139 = tempRet0;
        $cond86 = ($138|0)==(1);
        if ($cond86) {
         $$reg2mem82$0 = $139;
         continue;
        }
        HEAP32[$error>>2] = $136;
       }
       $140 = HEAP32[$42>>2]|0;
       $141 = ($140|0)==(0);
       if (!($141)) {
        break;
       }
       $142 = HEAP32[$clazz$sroa$0>>2]|0;
       $143 = $142;
       $144 = HEAP32[$cmap>>2]|0;
       __THREW__ = 0;
       $145 = (invoke_iiiii(4397,($143|0),($144|0),($charmap|0),($ttcmap|0))|0);
       $146 = __THREW__; __THREW__ = 0;
       if ((($146|0) != 0) & ((threwValue|0) != 0)) { $147 = _testSetjmp(HEAP32[$146>>2]|0, _setjmpTable|0, _setjmpTableSize|0)|0; if (($147|0) == 0) { _longjmp($146|0, threwValue|0); } tempRet0 = (threwValue); } else { $147 = -1; };
       $148 = tempRet0;
       $cond85 = ($147|0)==(1);
       if ($cond85) {
        $$reg2mem82$0 = $148;
       } else {
        label = 19;
        break;
       }
      }
      if ((label|0) == 19) {
       label = 0;
       $149 = ($145|0)==(0);
       if ($149) {
        $150 = HEAP32[$error>>2]|0;
        $151 = HEAP32[$ttcmap>>2]|0;
        $152 = ((($151)) + 20|0);
        HEAP32[$152>>2] = $150;
       }
      }
     }
    } while(0);
    $158 = HEAP32[$pclazz>>2]|0;
   }
  }
  $159 = HEAP32[$num_cmaps>>2]|0;
  $160 = (($159) + -1)|0;
  HEAP32[$num_cmaps>>2] = $160;
  $161 = HEAP32[$num_cmaps>>2]|0;
  $162 = ($161|0)==(0);
  if ($162) {
   label = 26;
   break;
  }
 }
 if ((label|0) == 26) {
  _free(_setjmpTable|0);
  STACKTOP = sp;return;
 }
}
function _sfnt_done_face($face) {
 $face = $face|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($face|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($face)) + 100|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($face)) + 532|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)!=(0|0);
 if ($5) {
  $6 = ((($4)) + 80|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)==(0|0);
  if (!($8)) {
   FUNCTION_TABLE_vi[$7 & 8191]($face);
  }
  $9 = ((($4)) + 100|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ($10|0)==(0|0);
  if (!($11)) {
   FUNCTION_TABLE_vi[$10 & 8191]($face);
  }
 }
 $12 = ((($face)) + 780|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = ($13<<24>>24)==(0);
 if (!($14)) {
  $15 = ((($face)) + 760|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ($16|0)==(0|0);
  if (!($17)) {
   $18 = ((($face)) + 104|0);
   $19 = HEAP32[$18>>2]|0;
   _FT_Stream_ReleaseFrame($19,$15);
  }
  $20 = ((($face)) + 764|0);
  HEAP32[$20>>2] = 0;
  $21 = ((($face)) + 768|0);
  HEAP32[$21>>2] = 0;
  $22 = ((($face)) + 772|0);
  HEAP32[$22>>2] = 0;
 }
 $23 = ((($face)) + 104|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($face)) + 740|0);
 _FT_Stream_ReleaseFrame($24,$25);
 $26 = ((($face)) + 744|0);
 ;HEAP32[$26>>2]=0|0;HEAP32[$26+4>>2]=0|0;HEAP32[$26+8>>2]=0|0;HEAP32[$26+12>>2]=0|0;
 $27 = ((($face)) + 144|0);
 $28 = HEAP32[$27>>2]|0;
 _ft_mem_free($2,$28);
 HEAP32[$27>>2] = 0;
 $29 = ((($face)) + 140|0);
 HEAP32[$29>>2] = 0;
 $30 = ((($face)) + 156|0);
 $31 = HEAP32[$30>>2]|0;
 _ft_mem_free($2,$31);
 HEAP32[$30>>2] = 0;
 $32 = ((($face)) + 152|0);
 HEAP16[$32>>1] = 0;
 $33 = HEAP32[$23>>2]|0;
 $34 = ((($face)) + 500|0);
 _FT_Stream_ReleaseFrame($33,$34);
 $35 = ((($face)) + 504|0);
 HEAP32[$35>>2] = 0;
 $36 = HEAP32[$23>>2]|0;
 $37 = ((($face)) + 680|0);
 _FT_Stream_ReleaseFrame($36,$37);
 $38 = ((($face)) + 688|0);
 _FT_Stream_ReleaseFrame($36,$38);
 $39 = ((($face)) + 684|0);
 HEAP32[$39>>2] = 0;
 $40 = ((($face)) + 692|0);
 HEAP32[$40>>2] = 0;
 $41 = ((($face)) + 292|0);
 $42 = HEAP8[$41>>0]|0;
 $43 = ($42<<24>>24)==(0);
 if (!($43)) {
  $44 = ((($face)) + 332|0);
  $45 = HEAP32[$44>>2]|0;
  _ft_mem_free($2,$45);
  HEAP32[$44>>2] = 0;
  $46 = ((($face)) + 336|0);
  $47 = HEAP32[$46>>2]|0;
  _ft_mem_free($2,$47);
  HEAP32[$46>>2] = 0;
  HEAP8[$41>>0] = 0;
 }
 $48 = ((($face)) + 544|0);
 $49 = HEAP32[$48>>2]|0;
 _ft_mem_free($2,$49);
 HEAP32[$48>>2] = 0;
 $50 = ((($face)) + 542|0);
 HEAP16[$50>>1] = 0;
 if ($5) {
  $51 = ((($4)) + 52|0);
  $52 = HEAP32[$51>>2]|0;
  FUNCTION_TABLE_vi[$52 & 8191]($face);
 }
 $53 = ((($face)) + 20|0);
 $54 = HEAP32[$53>>2]|0;
 _ft_mem_free($2,$54);
 HEAP32[$53>>2] = 0;
 $55 = ((($face)) + 24|0);
 $56 = HEAP32[$55>>2]|0;
 _ft_mem_free($2,$56);
 HEAP32[$55>>2] = 0;
 $57 = ((($face)) + 32|0);
 $58 = HEAP32[$57>>2]|0;
 _ft_mem_free($2,$58);
 HEAP32[$57>>2] = 0;
 $59 = ((($face)) + 28|0);
 HEAP32[$59>>2] = 0;
 $60 = ((($face)) + 664|0);
 $61 = HEAP32[$60>>2]|0;
 _ft_mem_free($2,$61);
 HEAP32[$60>>2] = 0;
 HEAP32[$3>>2] = 0;
 return;
}
function _sfnt_get_interface($module,$module_interface) {
 $module = $module|0;
 $module_interface = $module_interface|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_ft_service_list_lookup(55284,$module_interface)|0);
 return ($0|0);
}
function _tt_face_load_any($face,$tag,$offset,$buffer,$length) {
 $face = $face|0;
 $tag = $tag|0;
 $offset = $offset|0;
 $buffer = $buffer|0;
 $length = $length|0;
 var $$0 = 0, $$01 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $entry$01$i = 0, $entry$01$i$lcssa = 0, $face$idx = 0, $face$idx$val = 0, $face$idx2 = 0, $face$idx2$val = 0, $size$0 = 0, $size$0$pre = 0, $size$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($tag|0)==(0);
 do {
  if ($0) {
   $14 = ((($face)) + 104|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ((($15)) + 4|0);
   $size$0$pre = HEAP32[$16>>2]|0;
   $$01 = $offset;$size$0 = $size$0$pre;
  } else {
   $face$idx = ((($face)) + 152|0);
   $face$idx$val = HEAP16[$face$idx>>1]|0;
   $face$idx2 = ((($face)) + 156|0);
   $face$idx2$val = HEAP32[$face$idx2>>2]|0;
   $1 = $face$idx$val&65535;
   $2 = (($face$idx2$val) + ($1<<4)|0);
   $3 = ($face$idx$val<<16>>16)==(0);
   if ($3) {
    $$0 = 142;
    return ($$0|0);
   } else {
    $entry$01$i = $face$idx2$val;
   }
   while(1) {
    $4 = HEAP32[$entry$01$i>>2]|0;
    $5 = ($4|0)==($tag|0);
    if ($5) {
     $6 = ((($entry$01$i)) + 12|0);
     $7 = HEAP32[$6>>2]|0;
     $8 = ($7|0)==(0);
     if (!($8)) {
      $$lcssa = $7;$entry$01$i$lcssa = $entry$01$i;
      label = 6;
      break;
     }
    }
    $9 = ((($entry$01$i)) + 16|0);
    $10 = ($9>>>0)<($2>>>0);
    if ($10) {
     $entry$01$i = $9;
    } else {
     $$0 = 142;
     label = 12;
     break;
    }
   }
   if ((label|0) == 6) {
    $11 = ((($entry$01$i$lcssa)) + 8|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = (($12) + ($offset))|0;
    $$01 = $13;$size$0 = $$lcssa;
    break;
   }
   else if ((label|0) == 12) {
    return ($$0|0);
   }
  }
 } while(0);
 $17 = ($length|0)==(0|0);
 if ($17) {
  $size$1 = $size$0;
 } else {
  $18 = HEAP32[$length>>2]|0;
  $19 = ($18|0)==(0);
  if ($19) {
   HEAP32[$length>>2] = $size$0;
   $$0 = 0;
   return ($$0|0);
  } else {
   $size$1 = $18;
  }
 }
 $20 = ((($face)) + 104|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = (_FT_Stream_ReadAt($21,$$01,$buffer,$size$1)|0);
 $$0 = $22;
 return ($$0|0);
}
function _get_sfnt_table($face,$tag) {
 $face = $face|0;
 $tag = $tag|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $table$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 switch ($tag|0) {
 case 0:  {
  $0 = ((($face)) + 160|0);
  $table$0 = $0;
  break;
 }
 case 3:  {
  $1 = ((($face)) + 216|0);
  $table$0 = $1;
  break;
 }
 case 4:  {
  $2 = ((($face)) + 292|0);
  $3 = HEAP8[$2>>0]|0;
  $4 = ($3<<24>>24)!=(0);
  $5 = ((($face)) + 296|0);
  $6 = $4 ? $5 : 0;
  $table$0 = $6;
  break;
 }
 case 2:  {
  $7 = ((($face)) + 364|0);
  $8 = HEAP16[$7>>1]|0;
  $9 = ($8<<16>>16)==(-1);
  $10 = $9 ? 0 : $7;
  $table$0 = $10;
  break;
 }
 case 5:  {
  $11 = ((($face)) + 468|0);
  $table$0 = $11;
  break;
 }
 case 1:  {
  $12 = ((($face)) + 260|0);
  $table$0 = $12;
  break;
 }
 case 6:  {
  $13 = ((($face)) + 548|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ($14|0)!=(0);
  $16 = $15 ? $13 : 0;
  $table$0 = $16;
  break;
 }
 default: {
  $table$0 = 0;
 }
 }
 return ($table$0|0);
}
function _sfnt_table_info($face,$idx,$tag,$offset,$length) {
 $face = $face|0;
 $idx = $idx|0;
 $tag = $tag|0;
 $offset = $offset|0;
 $length = $length|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($offset|0)!=(0|0);
 $1 = ($length|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  $$0 = 6;
  return ($$0|0);
 }
 $2 = ($tag|0)==(0|0);
 $3 = ((($face)) + 152|0);
 $4 = HEAP16[$3>>1]|0;
 $5 = $4&65535;
 do {
  if ($2) {
   $storemerge = $5;
  } else {
   $6 = ($5>>>0)>($idx>>>0);
   if ($6) {
    $7 = ((($face)) + 156|0);
    $8 = HEAP32[$7>>2]|0;
    $9 = (($8) + ($idx<<4)|0);
    $10 = HEAP32[$9>>2]|0;
    HEAP32[$tag>>2] = $10;
    $11 = (((($8) + ($idx<<4)|0)) + 8|0);
    $12 = HEAP32[$11>>2]|0;
    HEAP32[$offset>>2] = $12;
    $13 = (((($8) + ($idx<<4)|0)) + 12|0);
    $14 = HEAP32[$13>>2]|0;
    $storemerge = $14;
    break;
   } else {
    $$0 = 142;
    return ($$0|0);
   }
  }
 } while(0);
 HEAP32[$length>>2] = $storemerge;
 $$0 = 0;
 return ($$0|0);
}
function _sfnt_get_ps_name($face) {
 $face = $face|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $div = 0, $error = 0, $error5 = 0, $found_apple$011 = 0, $found_apple$1 = 0, $found_apple$1$lcssa = 0, $found_win$012 = 0, $found_win$2 = 0;
 var $found_win$2$lcssa = 0, $len$08 = 0, $n$0$found_apple$0 = 0, $n$0$found_win$0 = 0, $n$013 = 0, $p$010 = 0, $r$0$lcssa = 0, $r$09 = 0, $r$1 = 0, $result$2 = 0, $result$5$ph = 0, $result$6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp + 4|0;
 $error5 = sp;
 $0 = ((($face)) + 664|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ((($face)) + 340|0);
 $4 = HEAP16[$3>>1]|0;
 $5 = ($4<<16>>16)==(0);
 do {
  if ($5) {
   $result$6 = 0;
  } else {
   $6 = ((($face)) + 356|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = $4&65535;
   $found_apple$011 = -1;$found_win$012 = -1;$n$013 = 0;
   while(1) {
    $9 = (((($7) + (($n$013*20)|0)|0)) + 6|0);
    $10 = HEAP16[$9>>1]|0;
    $11 = ($10<<16>>16)==(6);
    L8: do {
     if ($11) {
      $12 = (((($7) + (($n$013*20)|0)|0)) + 8|0);
      $13 = HEAP16[$12>>1]|0;
      $14 = ($13<<16>>16)==(0);
      if ($14) {
       $found_apple$1 = $found_apple$011;$found_win$2 = $found_win$012;
      } else {
       $15 = (($7) + (($n$013*20)|0)|0);
       $16 = HEAP16[$15>>1]|0;
       switch ($16<<16>>16) {
       case 3:  {
        $17 = (((($7) + (($n$013*20)|0)|0)) + 2|0);
        $18 = HEAP16[$17>>1]|0;
        $19 = ($18<<16>>16)==(1);
        if (!($19)) {
         $found_apple$1 = $found_apple$011;$found_win$2 = $found_win$012;
         break L8;
        }
        $20 = (((($7) + (($n$013*20)|0)|0)) + 4|0);
        $21 = HEAP16[$20>>1]|0;
        $22 = ($21<<16>>16)==(1033);
        $n$0$found_win$0 = $22 ? $n$013 : $found_win$012;
        $found_apple$1 = $found_apple$011;$found_win$2 = $n$0$found_win$0;
        break L8;
        break;
       }
       case 1:  {
        $23 = (((($7) + (($n$013*20)|0)|0)) + 2|0);
        $24 = HEAP16[$23>>1]|0;
        $25 = ($24<<16>>16)==(0);
        if (!($25)) {
         $found_apple$1 = $found_apple$011;$found_win$2 = $found_win$012;
         break L8;
        }
        $26 = (((($7) + (($n$013*20)|0)|0)) + 4|0);
        $27 = HEAP16[$26>>1]|0;
        $28 = ($27<<16>>16)==(0);
        $n$0$found_apple$0 = $28 ? $n$013 : $found_apple$011;
        $found_apple$1 = $n$0$found_apple$0;$found_win$2 = $found_win$012;
        break L8;
        break;
       }
       default: {
        $found_apple$1 = $found_apple$011;$found_win$2 = $found_win$012;
        break L8;
       }
       }
      }
     } else {
      $found_apple$1 = $found_apple$011;$found_win$2 = $found_win$012;
     }
    } while(0);
    $29 = (($n$013) + 1)|0;
    $30 = ($29|0)<($8|0);
    if ($30) {
     $found_apple$011 = $found_apple$1;$found_win$012 = $found_win$2;$n$013 = $29;
    } else {
     $found_apple$1$lcssa = $found_apple$1;$found_win$2$lcssa = $found_win$2;
     break;
    }
   }
   $31 = ($found_win$2$lcssa|0)==(-1);
   if ($31) {
    $68 = ($found_apple$1$lcssa|0)==(-1);
    if ($68) {
     $result$6 = 0;
     break;
    }
    $69 = ((($face)) + 100|0);
    $70 = HEAP32[$69>>2]|0;
    $71 = ((($face)) + 356|0);
    $72 = HEAP32[$71>>2]|0;
    $73 = (((($72) + (($found_apple$1$lcssa*20)|0)|0)) + 8|0);
    $74 = HEAP16[$73>>1]|0;
    $75 = $74&65535;
    HEAP32[$error5>>2] = 0;
    $76 = (($75) + 1)|0;
    $77 = (_ft_mem_alloc($70,$76,$error5)|0);
    $78 = HEAP32[$error5>>2]|0;
    $79 = ($78|0)==(0);
    do {
     if ($79) {
      $80 = ((($face)) + 360|0);
      $81 = HEAP32[$80>>2]|0;
      $82 = (((($72) + (($found_apple$1$lcssa*20)|0)|0)) + 12|0);
      $83 = HEAP32[$82>>2]|0;
      $84 = (_FT_Stream_Seek($81,$83)|0);
      HEAP32[$error5>>2] = $84;
      $85 = ($84|0)==(0);
      if ($85) {
       $86 = (_FT_Stream_Read($81,$77,$75)|0);
       HEAP32[$error5>>2] = $86;
       $87 = ($86|0)==(0);
       if ($87) {
        $88 = (($77) + ($75)|0);
        HEAP8[$88>>0] = 0;
        $result$5$ph = $77;
        break;
       }
      }
      HEAP32[$82>>2] = 0;
      HEAP16[$73>>1] = 0;
      $89 = (((($72) + (($found_apple$1$lcssa*20)|0)|0)) + 16|0);
      $90 = HEAP32[$89>>2]|0;
      _ft_mem_free($70,$90);
      HEAP32[$89>>2] = 0;
      _ft_mem_free($70,$77);
      $result$5$ph = 0;
     } else {
      $result$5$ph = $77;
     }
    } while(0);
    $result$6 = $result$5$ph;
    break;
   }
   $32 = ((($face)) + 100|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ((($face)) + 356|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = (((($35) + (($found_win$2$lcssa*20)|0)|0)) + 8|0);
   $37 = HEAP16[$36>>1]|0;
   $div = ($37&65535) >>> 1;
   $38 = $div&65535;
   HEAP32[$error>>2] = 0;
   $39 = $37&65535;
   $40 = (($39) + 1)|0;
   $41 = (_ft_mem_alloc($33,$40,$error)|0);
   $42 = HEAP32[$error>>2]|0;
   $43 = ($42|0)==(0);
   do {
    if ($43) {
     $44 = ((($face)) + 360|0);
     $45 = HEAP32[$44>>2]|0;
     $46 = (((($35) + (($found_win$2$lcssa*20)|0)|0)) + 12|0);
     $47 = HEAP32[$46>>2]|0;
     $48 = (_FT_Stream_Seek($45,$47)|0);
     HEAP32[$error>>2] = $48;
     $49 = ($48|0)==(0);
     if ($49) {
      $50 = HEAP16[$36>>1]|0;
      $51 = $50&65535;
      $52 = (_FT_Stream_EnterFrame($45,$51)|0);
      HEAP32[$error>>2] = $52;
      $53 = ($52|0)==(0);
      if ($53) {
       $54 = ($div<<16>>16)==(0);
       if ($54) {
        $r$0$lcssa = $41;
       } else {
        $55 = ((($45)) + 32|0);
        $56 = HEAP32[$55>>2]|0;
        $len$08 = $38;$p$010 = $56;$r$09 = $41;
        while(1) {
         $57 = HEAP8[$p$010>>0]|0;
         $58 = ($57<<24>>24)==(0);
         if ($58) {
          $59 = ((($p$010)) + 1|0);
          $60 = HEAP8[$59>>0]|0;
          $61 = ($60<<24>>24)>(31);
          if ($61) {
           $62 = ((($r$09)) + 1|0);
           HEAP8[$r$09>>0] = $60;
           $r$1 = $62;
          } else {
           $r$1 = $r$09;
          }
         } else {
          $r$1 = $r$09;
         }
         $63 = (($len$08) + -1)|0;
         $64 = ((($p$010)) + 2|0);
         $65 = ($63|0)==(0);
         if ($65) {
          $r$0$lcssa = $r$1;
          break;
         } else {
          $len$08 = $63;$p$010 = $64;$r$09 = $r$1;
         }
        }
       }
       HEAP8[$r$0$lcssa>>0] = 0;
       _FT_Stream_ExitFrame($45);
       $result$2 = $41;
       break;
      }
     }
     _ft_mem_free($33,$41);
     HEAP16[$36>>1] = 0;
     HEAP32[$46>>2] = 0;
     $66 = (((($35) + (($found_win$2$lcssa*20)|0)|0)) + 16|0);
     $67 = HEAP32[$66>>2]|0;
     _ft_mem_free($33,$67);
     HEAP32[$66>>2] = 0;
     $result$2 = 0;
    } else {
     $result$2 = $41;
    }
   } while(0);
   $result$6 = $result$2;
  }
 } while(0);
 HEAP32[$0>>2] = $result$6;
 $$0 = $result$6;
 STACKTOP = sp;return ($$0|0);
}
function _sfnt_get_glyph_name($face,$glyph_index,$buffer,$buffer_max) {
 $face = $face|0;
 $glyph_index = $glyph_index|0;
 $buffer = $buffer|0;
 $buffer_max = $buffer_max|0;
 var $0 = 0, $1 = 0, $2 = 0, $gname = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $gname = sp;
 $0 = (_tt_face_get_ps_name($face,$glyph_index,$gname)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  STACKTOP = sp;return ($0|0);
 }
 $2 = HEAP32[$gname>>2]|0;
 (_ft_mem_strcpyn($buffer,$2,$buffer_max)|0);
 STACKTOP = sp;return ($0|0);
}
function _tt_face_get_ps_name($face,$idx,$PSname) {
 $face = $face|0;
 $idx = $idx|0;
 $PSname = $PSname|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $switch$split12D = 0, $switch$split2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($face|0)==(0|0);
 if ($0) {
  $$0 = 35;
  return ($$0|0);
 }
 $1 = ((($face)) + 264|0);
 $2 = HEAP16[$1>>1]|0;
 $3 = $2&65535;
 $4 = ($3>>>0)>($idx>>>0);
 if (!($4)) {
  $$0 = 16;
  return ($$0|0);
 }
 $5 = ((($face)) + 536|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0|0);
 if ($7) {
  $$0 = 7;
  return ($$0|0);
 }
 $8 = ((($face)) + 612|0);
 $9 = ((($6)) + 16|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (FUNCTION_TABLE_ii[$10 & 8191](0)|0);
 HEAP32[$PSname>>2] = $11;
 $12 = ((($face)) + 468|0);
 $13 = HEAP32[$12>>2]|0;
 $switch$split2D = ($13|0)<(131072);
 if ($switch$split2D) {
  switch ($13|0) {
  case 65536:  {
   break;
  }
  default: {
   $$0 = 0;
   return ($$0|0);
  }
  }
  $14 = ($idx>>>0)<(258);
  if (!($14)) {
   $$0 = 0;
   return ($$0|0);
  }
  $15 = HEAP32[$9>>2]|0;
  $16 = (FUNCTION_TABLE_ii[$15 & 8191]($idx)|0);
  HEAP32[$PSname>>2] = $16;
  $$0 = 0;
  return ($$0|0);
 }
 $switch$split12D = ($13|0)<(163840);
 if (!($switch$split12D)) {
  switch ($13|0) {
  case 163840:  {
   break;
  }
  default: {
   $$0 = 0;
   return ($$0|0);
  }
  }
  $38 = HEAP8[$8>>0]|0;
  $39 = ($38<<24>>24)==(0);
  if ($39) {
   $40 = (_load_post_names($face)|0);
   $41 = ($40|0)==(0);
   if (!($41)) {
    $$0 = 0;
    return ($$0|0);
   }
  }
  $42 = ((($face)) + 616|0);
  $43 = HEAP16[$42>>1]|0;
  $44 = $43&65535;
  $45 = ($44>>>0)>($idx>>>0);
  if (!($45)) {
   $$0 = 0;
   return ($$0|0);
  }
  $46 = HEAP32[$9>>2]|0;
  $47 = ((($face)) + 620|0);
  $48 = HEAP32[$47>>2]|0;
  $49 = (($48) + ($idx)|0);
  $50 = HEAP8[$49>>0]|0;
  $51 = $50 << 24 >> 24;
  $52 = (($51) + ($idx))|0;
  $53 = (FUNCTION_TABLE_ii[$46 & 8191]($52)|0);
  HEAP32[$PSname>>2] = $53;
  $$0 = 0;
  return ($$0|0);
 }
 switch ($13|0) {
 case 131072:  {
  break;
 }
 default: {
  $$0 = 0;
  return ($$0|0);
 }
 }
 $17 = HEAP8[$8>>0]|0;
 $18 = ($17<<24>>24)==(0);
 if ($18) {
  $19 = (_load_post_names($face)|0);
  $20 = ($19|0)==(0);
  if (!($20)) {
   $$0 = 0;
   return ($$0|0);
  }
 }
 $21 = ((($face)) + 616|0);
 $22 = HEAP16[$21>>1]|0;
 $23 = $22&65535;
 $24 = ($23>>>0)>($idx>>>0);
 if (!($24)) {
  $$0 = 0;
  return ($$0|0);
 }
 $25 = ((($face)) + 620|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = (($26) + ($idx<<1)|0);
 $28 = HEAP16[$27>>1]|0;
 $29 = $28&65535;
 $30 = ($28&65535)<(258);
 if ($30) {
  $31 = HEAP32[$9>>2]|0;
  $32 = (FUNCTION_TABLE_ii[$31 & 8191]($29)|0);
  HEAP32[$PSname>>2] = $32;
  $$0 = 0;
  return ($$0|0);
 } else {
  $33 = (($29) + -258)|0;
  $34 = ((($face)) + 624|0);
  $35 = HEAP32[$34>>2]|0;
  $36 = (($35) + ($33<<2)|0);
  $37 = HEAP32[$36>>2]|0;
  HEAP32[$PSname>>2] = $37;
  $$0 = 0;
  return ($$0|0);
 }
 return (0)|0;
}
function _load_post_names($face) {
 $face = $face|0;
 var $$0$i = 0, $$0$i5 = 0, $$num_names$0$i = 0, $$pre$i = 0, $$pre$i4 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $12 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $error$0 = 0, $error$1 = 0, $error$i = 0, $error$i1 = 0, $exitcond = 0;
 var $exitcond36$i = 0, $exitcond37$i = 0, $indvars$iv$i = 0, $indvars$iv$next$i = 0, $indvars$iv$next33$i = 0, $indvars$iv$next35$i = 0, $indvars$iv32$i = 0, $indvars$iv34$i = 0, $indvars$iv34$i$lcssa = 0, $len$0$i = 0, $lftr$wideiv = 0, $n$024$i = 0, $n$03$i = 0, $n1$021$i = 0, $name_strings$0$i = 0, $num_names$0$lcssa$i = 0, $num_names$022$i = 0, $num_names$1$i = 0, $or$cond$i = 0, $post_len = 0;
 var $switch$split2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i1 = sp + 8|0;
 $error$i = sp + 4|0;
 $post_len = sp;
 $0 = ((($face)) + 104|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 508|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiiii[$3 & 8191]($face,1886352244,$1,$post_len)|0);
 $5 = ($4|0)==(0);
 if (!($5)) {
  $error$1 = $4;
  STACKTOP = sp;return ($error$1|0);
 }
 $6 = (_FT_Stream_Pos($1)|0);
 $7 = HEAP32[$post_len>>2]|0;
 $8 = (($7) + ($6))|0;
 $9 = ((($face)) + 468|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (_FT_Stream_Skip($1,32)|0);
 $12 = ($11|0)==(0);
 if (!($12)) {
  $error$1 = $11;
  STACKTOP = sp;return ($error$1|0);
 }
 $switch$split2D = ($10|0)<(163840);
 L7: do {
  if ($switch$split2D) {
   switch ($10|0) {
   case 131072:  {
    break;
   }
   default: {
    $error$0 = 3;
    break L7;
   }
   }
   $13 = ((($1)) + 28|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = (_FT_Stream_ReadUShort($1,$error$i)|0);
   $16 = $15&65535;
   $17 = HEAP32[$error$i>>2]|0;
   $18 = ($17|0)==(0);
   L10: do {
    if ($18) {
     $19 = ((($face)) + 264|0);
     $20 = HEAP16[$19>>1]|0;
     $21 = ($15&65535)>($20&65535);
     if ($21) {
      HEAP32[$error$i>>2] = 3;
      $$0$i = 3;
      break;
     }
     $22 = (_ft_mem_realloc($14,2,0,$16,0,$error$i)|0);
     $23 = HEAP32[$error$i>>2]|0;
     $24 = ($23|0)==(0);
     L15: do {
      if ($24) {
       $25 = $16 << 1;
       $26 = (_FT_Stream_EnterFrame($1,$25)|0);
       HEAP32[$error$i>>2] = $26;
       $27 = ($26|0)==(0);
       if ($27) {
        $28 = ($15<<16>>16)==(0);
        if ($28) {
         _FT_Stream_ExitFrame($1);
         $num_names$0$lcssa$i = 0;
        } else {
         $n$024$i = 0;
         while(1) {
          $29 = (_FT_Stream_GetUShort($1)|0);
          $30 = (($22) + ($n$024$i<<1)|0);
          HEAP16[$30>>1] = $29;
          $31 = (($n$024$i) + 1)|0;
          $exitcond37$i = ($31|0)==($16|0);
          if ($exitcond37$i) {
           break;
          } else {
           $n$024$i = $31;
          }
         }
         _FT_Stream_ExitFrame($1);
         $n1$021$i = 0;$num_names$022$i = 0;
         while(1) {
          $32 = (($22) + ($n1$021$i<<1)|0);
          $33 = HEAP16[$32>>1]|0;
          $34 = ($33&65535)>(257);
          if ($34) {
           $35 = $33&65535;
           $36 = (($35) + -257)|0;
           $37 = $num_names$022$i&65535;
           $38 = ($36|0)>($37|0);
           $39 = $36&65535;
           $$num_names$0$i = $38 ? $39 : $num_names$022$i;
           $num_names$1$i = $$num_names$0$i;
          } else {
           $num_names$1$i = $num_names$022$i;
          }
          $40 = (($n1$021$i) + 1)|0;
          $exitcond36$i = ($40|0)==($16|0);
          if ($exitcond36$i) {
           $num_names$0$lcssa$i = $num_names$1$i;
           break;
          } else {
           $n1$021$i = $40;$num_names$022$i = $num_names$1$i;
          }
         }
        }
        $41 = $num_names$0$lcssa$i&65535;
        $42 = (_ft_mem_realloc($14,4,0,$41,0,$error$i)|0);
        $43 = HEAP32[$error$i>>2]|0;
        $44 = ($43|0)==(0);
        if ($44) {
         $45 = ($num_names$0$lcssa$i<<16>>16)==(0);
         L30: do {
          if (!($45)) {
           $indvars$iv34$i = 0;
           while(1) {
            $46 = (_FT_Stream_Pos($1)|0);
            $47 = ($46>>>0)<($8>>>0);
            if (!($47)) {
             $indvars$iv34$i$lcssa = $indvars$iv34$i;
             label = 19;
             break;
            }
            $51 = (_FT_Stream_ReadChar($1,$error$i)|0);
            $52 = $51&255;
            $53 = HEAP32[$error$i>>2]|0;
            $54 = ($53|0)==(0);
            if (!($54)) {
             $indvars$iv$i = 0;
             break;
            }
            $55 = ($8>>>0)<($52>>>0);
            if ($55) {
             label = 24;
            } else {
             $56 = (_FT_Stream_Pos($1)|0);
             $57 = (($8) - ($52))|0;
             $58 = ($56>>>0)>($57>>>0);
             if ($58) {
              label = 24;
             } else {
              $len$0$i = $52;
             }
            }
            if ((label|0) == 24) {
             label = 0;
             $59 = (_FT_Stream_Pos($1)|0);
             $60 = (($8) - ($59))|0;
             $61 = ($60|0)<(0);
             $62 = $61 ? 0 : $60;
             $len$0$i = $62;
            }
            $63 = (($len$0$i) + 1)|0;
            $64 = (_ft_mem_realloc($14,1,0,$63,0,$error$i)|0);
            $65 = (($42) + ($indvars$iv34$i<<2)|0);
            HEAP32[$65>>2] = $64;
            $66 = HEAP32[$error$i>>2]|0;
            $67 = ($66|0)==(0);
            if (!($67)) {
             $indvars$iv$i = 0;
             break;
            }
            $68 = (_FT_Stream_Read($1,$64,$len$0$i)|0);
            HEAP32[$error$i>>2] = $68;
            $69 = ($68|0)==(0);
            if (!($69)) {
             $indvars$iv$i = 0;
             break;
            }
            $70 = HEAP32[$65>>2]|0;
            $71 = (($70) + ($len$0$i)|0);
            HEAP8[$71>>0] = 0;
            $indvars$iv$next35$i = (($indvars$iv34$i) + 1)|0;
            $72 = ($indvars$iv$next35$i>>>0)<($41>>>0);
            if ($72) {
             $indvars$iv34$i = $indvars$iv$next35$i;
            } else {
             break L30;
            }
           }
           L42: do {
            if ((label|0) == 19) {
             $48 = $indvars$iv34$i$lcssa&65535;
             $49 = ($48&65535)<($num_names$0$lcssa$i&65535);
             if (!($49)) {
              break L30;
             }
             $50 = $indvars$iv34$i$lcssa & 65535;
             $indvars$iv32$i = $50;
             while(1) {
              $73 = (_ft_mem_realloc($14,1,0,1,0,$error$i)|0);
              $74 = (($42) + ($indvars$iv32$i<<2)|0);
              HEAP32[$74>>2] = $73;
              $75 = HEAP32[$error$i>>2]|0;
              $76 = ($75|0)==(0);
              if (!($76)) {
               $indvars$iv$i = 0;
               break L42;
              }
              HEAP8[$73>>0] = 0;
              $indvars$iv$next33$i = (($indvars$iv32$i) + 1)|0;
              $77 = ($indvars$iv$next33$i>>>0)<($41>>>0);
              if ($77) {
               $indvars$iv32$i = $indvars$iv$next33$i;
              } else {
               break L30;
              }
             }
            }
           } while(0);
           while(1) {
            $82 = (($42) + ($indvars$iv$i<<2)|0);
            $83 = HEAP32[$82>>2]|0;
            _ft_mem_free($14,$83);
            HEAP32[$82>>2] = 0;
            $indvars$iv$next$i = (($indvars$iv$i) + 1)|0;
            $lftr$wideiv = $indvars$iv$next$i&65535;
            $exitcond = ($lftr$wideiv<<16>>16)==($num_names$0$lcssa$i<<16>>16);
            if ($exitcond) {
             $name_strings$0$i = $42;
             break L15;
            } else {
             $indvars$iv$i = $indvars$iv$next$i;
            }
           }
          }
         } while(0);
         $78 = ((($face)) + 616|0);
         HEAP16[$78>>1] = $15;
         $79 = ((($face)) + 618|0);
         HEAP16[$79>>1] = $num_names$0$lcssa$i;
         $80 = ((($face)) + 620|0);
         HEAP32[$80>>2] = $22;
         $81 = ((($face)) + 624|0);
         HEAP32[$81>>2] = $42;
         $$0$i = 0;
         break L10;
        } else {
         $name_strings$0$i = $42;
        }
       } else {
        $name_strings$0$i = 0;
       }
      } else {
       $name_strings$0$i = 0;
      }
     } while(0);
     _ft_mem_free($14,$name_strings$0$i);
     _ft_mem_free($14,$22);
     $$pre$i = HEAP32[$error$i>>2]|0;
     $$0$i = $$pre$i;
    } else {
     $$0$i = $17;
    }
   } while(0);
   $error$0 = $$0$i;
  } else {
   switch ($10|0) {
   case 163840:  {
    break;
   }
   default: {
    $error$0 = 3;
    break L7;
   }
   }
   $84 = ((($1)) + 28|0);
   $85 = HEAP32[$84>>2]|0;
   $86 = (_FT_Stream_ReadUShort($1,$error$i1)|0);
   $87 = $86&65535;
   $88 = HEAP32[$error$i1>>2]|0;
   $89 = ($88|0)==(0);
   L55: do {
    if ($89) {
     $90 = ((($face)) + 264|0);
     $91 = HEAP16[$90>>1]|0;
     $92 = ($86&65535)>($91&65535);
     $93 = ($86&65535)>(258);
     $or$cond$i = $93 | $92;
     if ($or$cond$i) {
      HEAP32[$error$i1>>2] = 3;
      $$0$i5 = 3;
      break;
     }
     $94 = (_ft_mem_realloc($85,1,0,$87,0,$error$i1)|0);
     $95 = HEAP32[$error$i1>>2]|0;
     $96 = ($95|0)==(0);
     L60: do {
      if ($96) {
       $97 = (_FT_Stream_Read($1,$94,$87)|0);
       HEAP32[$error$i1>>2] = $97;
       $98 = ($97|0)==(0);
       if ($98) {
        $99 = ($86<<16>>16)==(0);
        L63: do {
         if (!($99)) {
          $n$03$i = 0;
          while(1) {
           $102 = (($94) + ($n$03$i)|0);
           $103 = HEAP8[$102>>0]|0;
           $104 = $103 << 24 >> 24;
           $105 = (($104) + ($n$03$i))|0;
           $106 = ($105>>>0)>($87>>>0);
           $101 = (($n$03$i) + 1)|0;
           if ($106) {
            break;
           }
           $100 = ($101|0)<($87|0);
           if ($100) {
            $n$03$i = $101;
           } else {
            break L63;
           }
          }
          HEAP32[$error$i1>>2] = 3;
          break L60;
         }
        } while(0);
        $107 = ((($face)) + 616|0);
        HEAP16[$107>>1] = $86;
        $108 = ((($face)) + 620|0);
        HEAP32[$108>>2] = $94;
        $$0$i5 = 0;
        break L55;
       }
      }
     } while(0);
     _ft_mem_free($85,$94);
     $$pre$i4 = HEAP32[$error$i1>>2]|0;
     $$0$i5 = $$pre$i4;
    } else {
     $$0$i5 = $88;
    }
   } while(0);
   $error$0 = $$0$i5;
  }
 } while(0);
 $109 = ((($face)) + 612|0);
 HEAP8[$109>>0] = 1;
 $error$1 = $error$0;
 STACKTOP = sp;return ($error$1|0);
}
function _sfnt_get_name_index($face,$glyph_name) {
 $face = $face|0;
 $glyph_name = $glyph_name|0;
 var $$2 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $gname = 0, $i$05 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $gname = sp;
 $0 = ((($face)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(0);
 if (!($2)) {
  $$2 = 0;
  STACKTOP = sp;return ($$2|0);
 }
 $i$05 = 0;
 while(1) {
  $3 = (_tt_face_get_ps_name($face,$i$05,$gname)|0);
  $4 = ($3|0)==(0);
  if ($4) {
   $5 = HEAP32[$gname>>2]|0;
   $6 = (_strcmp($glyph_name,$5)|0);
   $7 = ($6|0)==(0);
   if ($7) {
    $$2 = $i$05;
    label = 7;
    break;
   }
  } else {
  }
  $8 = (($i$05) + 1)|0;
  $9 = ($8>>>0)<($1>>>0);
  if ($9) {
   $i$05 = $8;
  } else {
   $$2 = 0;
   label = 7;
   break;
  }
 }
 if ((label|0) == 7) {
  STACKTOP = sp;return ($$2|0);
 }
 return (0)|0;
}
function _sfnt_get_charset_id($face,$acharset_encoding,$acharset_registry) {
 $face = $face|0;
 $acharset_encoding = $acharset_encoding|0;
 $acharset_registry = $acharset_registry|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $encoding = 0, $error$0 = 0, $registry = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $encoding = sp + 8|0;
 $registry = sp;
 $0 = (_tt_face_find_bdf_prop($face,593036,$registry)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  $2 = (_tt_face_find_bdf_prop($face,593053,$encoding)|0);
  $3 = ($2|0)==(0);
  if ($3) {
   $4 = HEAP32[$registry>>2]|0;
   $5 = ($4|0)==(1);
   if ($5) {
    $6 = HEAP32[$encoding>>2]|0;
    $7 = ($6|0)==(1);
    if ($7) {
     $8 = ((($encoding)) + 4|0);
     $9 = HEAP32[$8>>2]|0;
     HEAP32[$acharset_encoding>>2] = $9;
     $10 = ((($registry)) + 4|0);
     $11 = HEAP32[$10>>2]|0;
     HEAP32[$acharset_registry>>2] = $11;
     $error$0 = 0;
    } else {
     $error$0 = 6;
    }
   } else {
    $error$0 = 6;
   }
  } else {
   $error$0 = $2;
  }
 } else {
  $error$0 = $0;
 }
 STACKTOP = sp;return ($error$0|0);
}
function _tt_face_find_bdf_prop($face,$property_name,$aprop) {
 $face = $face|0;
 $property_name = $property_name|0;
 $aprop = $aprop|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa73 = 0, $$lcssa75 = 0, $$lcssa77 = 0, $$lcssa81 = 0, $$old2 = 0, $$phi$trans$insert = 0, $$phi$trans$insert36 = 0, $$pre = 0, $$pre37 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $count$0 = 0, $count$010$i = 0, $count$224 = 0;
 var $entry$01$i$i$i = 0, $entry$01$i$i$i$lcssa = 0, $face$idx$i$i = 0, $face$idx$val$i$i = 0, $face$idx1$i$i = 0, $face$idx1$val$i$i = 0, $or$cond = 0, $or$cond$i = 0, $or$cond3 = 0, $or$cond3$i = 0, $or$cond7$i = 0, $p$0 = 0, $p$09$i = 0, $p$125 = 0, $strike$0 = 0, $strike$0$lcssa = 0, $strike$0$lcssa$i = 0, $strike$011$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 88|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$aprop>>2] = 0;
 $2 = ((($face)) + 780|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = ($3<<24>>24)==(0);
 do {
  if ($4) {
   $5 = ((($face)) + 104|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ((($face)) + 760|0);
   ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;HEAP32[$7+12>>2]=0|0;HEAP32[$7+16>>2]=0|0;HEAP32[$7+20>>2]=0|0;
   $face$idx$i$i = ((($face)) + 152|0);
   $face$idx$val$i$i = HEAP16[$face$idx$i$i>>1]|0;
   $face$idx1$i$i = ((($face)) + 156|0);
   $face$idx1$val$i$i = HEAP32[$face$idx1$i$i>>2]|0;
   $8 = $face$idx$val$i$i&65535;
   $9 = (($face$idx1$val$i$i) + ($8<<4)|0);
   $10 = ($face$idx$val$i$i<<16>>16)==(0);
   if ($10) {
    $$0 = 8;
    return ($$0|0);
   } else {
    $entry$01$i$i$i = $face$idx1$val$i$i;
   }
   while(1) {
    $11 = HEAP32[$entry$01$i$i$i>>2]|0;
    $12 = ($11|0)==(1111770656);
    if ($12) {
     $13 = ((($entry$01$i$i$i)) + 12|0);
     $14 = HEAP32[$13>>2]|0;
     $15 = ($14|0)==(0);
     if (!($15)) {
      $$lcssa81 = $14;$entry$01$i$i$i$lcssa = $entry$01$i$i$i;
      break;
     }
    }
    $16 = ((($entry$01$i$i$i)) + 16|0);
    $17 = ($16>>>0)<($9>>>0);
    if ($17) {
     $entry$01$i$i$i = $16;
    } else {
     $$0 = 8;
     label = 34;
     break;
    }
   }
   if ((label|0) == 34) {
    return ($$0|0);
   }
   $18 = ((($entry$01$i$i$i$lcssa)) + 8|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = (_FT_Stream_Seek($6,$19)|0);
   $21 = ($20|0)!=(0);
   $22 = ($$lcssa81>>>0)<(8);
   $or$cond$i = $22 | $21;
   if ($or$cond$i) {
    $$0 = 8;
    return ($$0|0);
   }
   $23 = (_FT_Stream_ExtractFrame($6,$$lcssa81,$7)|0);
   $24 = ($23|0)==(0);
   if (!($24)) {
    $$0 = 8;
    return ($$0|0);
   }
   $25 = HEAP32[$7>>2]|0;
   $26 = (($25) + ($$lcssa81)|0);
   $27 = ((($face)) + 764|0);
   HEAP32[$27>>2] = $26;
   $28 = HEAP8[$25>>0]|0;
   $29 = $28&255;
   $30 = $29 << 8;
   $31 = ((($25)) + 1|0);
   $32 = HEAP8[$31>>0]|0;
   $33 = $32&255;
   $34 = $30 | $33;
   $35 = ((($25)) + 2|0);
   $36 = HEAP8[$35>>0]|0;
   $37 = $36&255;
   $38 = $37 << 8;
   $39 = ((($25)) + 3|0);
   $40 = HEAP8[$39>>0]|0;
   $41 = $40&255;
   $42 = $38 | $41;
   $43 = ((($25)) + 4|0);
   $44 = HEAP8[$43>>0]|0;
   $45 = $44&255;
   $46 = $45 << 24;
   $47 = ((($25)) + 5|0);
   $48 = HEAP8[$47>>0]|0;
   $49 = $48&255;
   $50 = $49 << 16;
   $51 = $50 | $46;
   $52 = ((($25)) + 6|0);
   $53 = HEAP8[$52>>0]|0;
   $54 = $53&255;
   $55 = $54 << 8;
   $56 = $51 | $55;
   $57 = ((($25)) + 7|0);
   $58 = HEAP8[$57>>0]|0;
   $59 = $58&255;
   $60 = $56 | $59;
   $61 = ($34|0)!=(1);
   $62 = ($60>>>0)<(8);
   $or$cond3$i = $61 | $62;
   if (!($or$cond3$i)) {
    $63 = (($60) + -8)|0;
    $64 = $63 >>> 2;
    $65 = ($64>>>0)<($42>>>0);
    $66 = (($60) + 1)|0;
    $67 = ($66>>>0)>($$lcssa81>>>0);
    $or$cond7$i = $67 | $65;
    if (!($or$cond7$i)) {
     $68 = ((($face)) + 776|0);
     HEAP32[$68>>2] = $42;
     $69 = (($25) + ($60)|0);
     $70 = ((($face)) + 768|0);
     HEAP32[$70>>2] = $69;
     $71 = (($$lcssa81) - ($60))|0;
     $72 = ((($face)) + 772|0);
     HEAP32[$72>>2] = $71;
     $73 = ((($25)) + 8|0);
     $74 = $42 << 2;
     $75 = (($73) + ($74)|0);
     $76 = ($42|0)==(0);
     if ($76) {
      $strike$0$lcssa$i = $75;
     } else {
      $count$010$i = $42;$p$09$i = $73;$strike$011$i = $75;
      while(1) {
       $77 = ((($p$09$i)) + 2|0);
       $78 = HEAP8[$77>>0]|0;
       $79 = $78&255;
       $80 = $79 << 8;
       $81 = ((($p$09$i)) + 3|0);
       $82 = HEAP8[$81>>0]|0;
       $83 = $82&255;
       $84 = $80 | $83;
       $85 = ($84*10)|0;
       $86 = (($strike$011$i) + ($85)|0);
       $87 = ((($p$09$i)) + 4|0);
       $88 = (($count$010$i) + -1)|0;
       $89 = ($88|0)==(0);
       if ($89) {
        $strike$0$lcssa$i = $86;
        break;
       } else {
        $count$010$i = $88;$p$09$i = $87;$strike$011$i = $86;
       }
      }
     }
     $90 = ($strike$0$lcssa$i>>>0)>($69>>>0);
     if (!($90)) {
      HEAP8[$2>>0] = 1;
      $92 = $25;$98 = $42;
      break;
     }
    }
   }
   _FT_Stream_ReleaseFrame($6,$7);
   ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;HEAP32[$7+12>>2]=0|0;HEAP32[$7+16>>2]=0|0;HEAP32[$7+20>>2]=0|0;
   $$0 = 8;
   return ($$0|0);
  } else {
   $$phi$trans$insert = ((($face)) + 776|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0;
   $$phi$trans$insert36 = ((($face)) + 760|0);
   $$pre37 = HEAP32[$$phi$trans$insert36>>2]|0;
   $92 = $$pre37;$98 = $$pre;
  }
 } while(0);
 $91 = ((($92)) + 8|0);
 $93 = ($1|0)==(0|0);
 $94 = ($property_name|0)==(0|0);
 $or$cond = $94 | $93;
 if ($or$cond) {
  $$0 = 6;
  return ($$0|0);
 }
 $95 = (_strlen($property_name)|0);
 $96 = ($95|0)!=(0);
 $97 = ($98|0)!=(0);
 $or$cond3 = $97 & $96;
 if (!($or$cond3)) {
  $$0 = 6;
  return ($$0|0);
 }
 $99 = $98 << 2;
 $100 = (($91) + ($99)|0);
 $101 = ((($1)) + 14|0);
 $102 = HEAP16[$101>>1]|0;
 $103 = $102&65535;
 $count$0 = $98;$p$0 = $91;$strike$0 = $100;
 while(1) {
  $104 = HEAP8[$p$0>>0]|0;
  $105 = $104&255;
  $106 = $105 << 8;
  $107 = ((($p$0)) + 1|0);
  $108 = HEAP8[$107>>0]|0;
  $109 = $108&255;
  $110 = $106 | $109;
  $111 = ((($p$0)) + 2|0);
  $112 = HEAP8[$111>>0]|0;
  $113 = $112&255;
  $114 = $113 << 8;
  $115 = ((($p$0)) + 3|0);
  $116 = HEAP8[$115>>0]|0;
  $117 = $116&255;
  $118 = $114 | $117;
  $119 = ($110|0)==($103|0);
  if ($119) {
   $$lcssa77 = $118;$strike$0$lcssa = $strike$0;
   break;
  }
  $123 = ($118*10)|0;
  $124 = (($strike$0) + ($123)|0);
  $125 = ((($p$0)) + 4|0);
  $126 = (($count$0) + -1)|0;
  $$old2 = ($126|0)==(0);
  if ($$old2) {
   $$0 = 6;
   label = 34;
   break;
  } else {
   $count$0 = $126;$p$0 = $125;$strike$0 = $124;
  }
 }
 if ((label|0) == 34) {
  return ($$0|0);
 }
 $120 = ($$lcssa77|0)==(0);
 if ($120) {
  $$0 = 6;
  return ($$0|0);
 }
 $121 = ((($face)) + 772|0);
 $122 = ((($face)) + 768|0);
 $count$224 = $$lcssa77;$p$125 = $strike$0$lcssa;
 L46: while(1) {
  $127 = ((($p$125)) + 5|0);
  $128 = HEAP8[$127>>0]|0;
  $129 = $128&255;
  $130 = $129 & 16;
  $131 = ($130|0)==(0);
  L48: do {
   if (!($131)) {
    $132 = HEAP8[$p$125>>0]|0;
    $133 = $132&255;
    $134 = $133 << 24;
    $135 = ((($p$125)) + 1|0);
    $136 = HEAP8[$135>>0]|0;
    $137 = $136&255;
    $138 = $137 << 16;
    $139 = $138 | $134;
    $140 = ((($p$125)) + 2|0);
    $141 = HEAP8[$140>>0]|0;
    $142 = $141&255;
    $143 = $142 << 8;
    $144 = $139 | $143;
    $145 = ((($p$125)) + 3|0);
    $146 = HEAP8[$145>>0]|0;
    $147 = $146&255;
    $148 = $144 | $147;
    $149 = ((($p$125)) + 6|0);
    $150 = HEAP8[$149>>0]|0;
    $151 = $150&255;
    $152 = $151 << 24;
    $153 = ((($p$125)) + 7|0);
    $154 = HEAP8[$153>>0]|0;
    $155 = $154&255;
    $156 = $155 << 16;
    $157 = $156 | $152;
    $158 = ((($p$125)) + 8|0);
    $159 = HEAP8[$158>>0]|0;
    $160 = $159&255;
    $161 = $160 << 8;
    $162 = $157 | $161;
    $163 = ((($p$125)) + 9|0);
    $164 = HEAP8[$163>>0]|0;
    $165 = $164&255;
    $166 = $162 | $165;
    $167 = HEAP32[$121>>2]|0;
    $168 = ($167>>>0)>($148>>>0);
    if ($168) {
     $169 = (($167) - ($148))|0;
     $170 = ($95>>>0)<($169>>>0);
     if ($170) {
      $171 = HEAP32[$122>>2]|0;
      $172 = (($171) + ($148)|0);
      $173 = (_strncmp($property_name,$172,$169)|0);
      $174 = ($173|0)==(0);
      if ($174) {
       $175 = $129 & 15;
       switch ($175|0) {
       case 2:  {
        $$lcssa = $166;
        label = 31;
        break L46;
        break;
       }
       case 3:  {
        $$lcssa73 = $166;
        label = 32;
        break L46;
        break;
       }
       case 1: case 0:  {
        break;
       }
       default: {
        break L48;
       }
       }
       $176 = ($166>>>0)<($167>>>0);
       if ($176) {
        $177 = (($171) + ($166)|0);
        $178 = (_memchr($177,0,$167)|0);
        $179 = ($178|0)==(0|0);
        if (!($179)) {
         $$lcssa75 = $177;
         label = 30;
         break L46;
        }
       }
      }
     }
    }
   }
  } while(0);
  $183 = ((($p$125)) + 10|0);
  $184 = (($count$224) + -1)|0;
  $185 = ($184|0)==(0);
  if ($185) {
   $$0 = 6;
   label = 34;
   break;
  } else {
   $count$224 = $184;$p$125 = $183;
  }
 }
 if ((label|0) == 30) {
  HEAP32[$aprop>>2] = 1;
  $180 = ((($aprop)) + 4|0);
  HEAP32[$180>>2] = $$lcssa75;
  $$0 = 0;
  return ($$0|0);
 }
 else if ((label|0) == 31) {
  HEAP32[$aprop>>2] = 2;
  $181 = ((($aprop)) + 4|0);
  HEAP32[$181>>2] = $$lcssa;
  $$0 = 0;
  return ($$0|0);
 }
 else if ((label|0) == 32) {
  HEAP32[$aprop>>2] = 3;
  $182 = ((($aprop)) + 4|0);
  HEAP32[$182>>2] = $$lcssa73;
  $$0 = 0;
  return ($$0|0);
 }
 else if ((label|0) == 34) {
  return ($$0|0);
 }
 return (0)|0;
}
function _tt_get_cmap_info($charmap,$cmap_info) {
 $charmap = $charmap|0;
 $cmap_info = $cmap_info|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($charmap)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 48|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iii[$3 & 8191]($charmap,$cmap_info)|0);
 return ($4|0);
}
function _tt_face_load_head($face,$stream) {
 $face = $face|0;
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $error$0$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 508|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (FUNCTION_TABLE_iiiii[$1 & 8191]($face,1751474532,$stream,0)|0);
 $3 = ($2|0)==(0);
 if (!($3)) {
  $error$0$i = $2;
  return ($error$0$i|0);
 }
 $4 = ((($face)) + 160|0);
 $5 = (_FT_Stream_ReadFields($stream,174126,$4)|0);
 $error$0$i = $5;
 return ($error$0$i|0);
}
function _tt_face_load_hhea($face,$stream,$vertical) {
 $face = $face|0;
 $stream = $stream|0;
 $vertical = $vertical|0;
 var $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $header$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($vertical<<24>>24)==(0);
 $1 = ((($face)) + 508|0);
 $2 = HEAP32[$1>>2]|0;
 do {
  if ($0) {
   $6 = (FUNCTION_TABLE_iiiii[$2 & 8191]($face,1751672161,$stream,0)|0);
   $7 = ($6|0)==(0);
   if ($7) {
    $8 = ((($face)) + 216|0);
    $header$1 = $8;
    break;
   } else {
    $$01 = $6;
    return ($$01|0);
   }
  } else {
   $3 = (FUNCTION_TABLE_iiiii[$2 & 8191]($face,1986553185,$stream,0)|0);
   $4 = ($3|0)==(0);
   $5 = ((($face)) + 296|0);
   if ($4) {
    $header$1 = $5;
   } else {
    $$01 = $3;
    return ($$01|0);
   }
  }
 } while(0);
 $9 = (_FT_Stream_ReadFields($stream,174210,$header$1)|0);
 $10 = ($9|0)==(0);
 if (!($10)) {
  $$01 = $9;
  return ($$01|0);
 }
 $11 = ((($header$1)) + 36|0);
 HEAP32[$11>>2] = 0;
 $12 = ((($header$1)) + 40|0);
 HEAP32[$12>>2] = 0;
 $$01 = 0;
 return ($$01|0);
}
function _tt_face_load_cmap($face,$stream) {
 $face = $face|0;
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $error$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 508|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 504|0);
 $3 = (FUNCTION_TABLE_iiiii[$1 & 8191]($face,1668112752,$stream,$2)|0);
 $4 = ($3|0)==(0);
 if (!($4)) {
  $error$0 = $3;
  return ($error$0|0);
 }
 $5 = HEAP32[$2>>2]|0;
 $6 = ((($face)) + 500|0);
 $7 = (_FT_Stream_ExtractFrame($stream,$5,$6)|0);
 $8 = ($7|0)==(0);
 if ($8) {
  $error$0 = 0;
  return ($error$0|0);
 }
 HEAP32[$2>>2] = 0;
 $error$0 = $7;
 return ($error$0|0);
}
function _tt_face_load_maxp($face,$stream) {
 $face = $face|0;
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $error$0 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 260|0);
 $1 = ((($face)) + 508|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (FUNCTION_TABLE_iiiii[$2 & 8191]($face,1835104368,$stream,0)|0);
 $4 = ($3|0)==(0);
 if (!($4)) {
  $error$0 = $3;
  return ($error$0|0);
 }
 $5 = (_FT_Stream_ReadFields($stream,174286,$0)|0);
 $6 = ($5|0)==(0);
 if (!($6)) {
  $error$0 = $5;
  return ($error$0|0);
 }
 $7 = ((($face)) + 266|0);
 $8 = ((($face)) + 276|0);
 $9 = ((($face)) + 280|0);
 $10 = ((($face)) + 290|0);
 dest=$7; stop=dest+26|0; do { HEAP16[dest>>1]=0|0; dest=dest+2|0; } while ((dest|0) < (stop|0));
 $11 = HEAP32[$0>>2]|0;
 $12 = ($11|0)>(65535);
 if (!($12)) {
  $error$0 = 0;
  return ($error$0|0);
 }
 $13 = (_FT_Stream_ReadFields($stream,174302,$0)|0);
 $14 = ($13|0)==(0);
 if (!($14)) {
  $error$0 = $13;
  return ($error$0|0);
 }
 $15 = HEAP16[$9>>1]|0;
 $16 = ($15&65535)<(64);
 if ($16) {
  HEAP16[$9>>1] = 64;
 }
 $17 = HEAP16[$8>>1]|0;
 $18 = ($17&65535)>(65531);
 if ($18) {
  HEAP16[$8>>1] = -5;
 }
 $19 = HEAP16[$10>>1]|0;
 $20 = ($19&65535)>(100);
 if (!($20)) {
  $error$0 = 0;
  return ($error$0|0);
 }
 HEAP16[$10>>1] = 100;
 $error$0 = 0;
 return ($error$0|0);
}
function _tt_face_load_os2($face,$stream) {
 $face = $face|0;
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $error$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 508|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (FUNCTION_TABLE_iiiii[$1 & 8191]($face,1330851634,$stream,0)|0);
 $3 = ($2|0)==(0);
 if (!($3)) {
  $error$0 = $2;
  return ($error$0|0);
 }
 $4 = ((($face)) + 364|0);
 $5 = (_FT_Stream_ReadFields($stream,174362,$4)|0);
 $6 = ($5|0)==(0);
 if (!($6)) {
  $error$0 = $5;
  return ($error$0|0);
 }
 $7 = ((($face)) + 444|0);
 $8 = ((($face)) + 464|0);
 ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;HEAP32[$7+12>>2]=0|0;HEAP32[$7+16>>2]=0|0;
 HEAP16[$8>>1] = -1;
 $9 = HEAP16[$4>>1]|0;
 $10 = ($9<<16>>16)==(0);
 if ($10) {
  $error$0 = 0;
  return ($error$0|0);
 }
 $11 = (_FT_Stream_ReadFields($stream,174538,$4)|0);
 $12 = ($11|0)==(0);
 if (!($12)) {
  $error$0 = $11;
  return ($error$0|0);
 }
 $13 = HEAP16[$4>>1]|0;
 $14 = ($13&65535)>(1);
 if (!($14)) {
  $error$0 = 0;
  return ($error$0|0);
 }
 $15 = (_FT_Stream_ReadFields($stream,174554,$4)|0);
 $16 = ($15|0)==(0);
 if (!($16)) {
  $error$0 = $15;
  return ($error$0|0);
 }
 $17 = HEAP16[$4>>1]|0;
 $18 = ($17&65535)>(4);
 if (!($18)) {
  $error$0 = 0;
  return ($error$0|0);
 }
 $19 = (_FT_Stream_ReadFields($stream,174582,$4)|0);
 $error$0 = $19;
 return ($error$0|0);
}
function _tt_face_load_post($face,$stream) {
 $face = $face|0;
 $stream = $stream|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 508|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (FUNCTION_TABLE_iiiii[$1 & 8191]($face,1886352244,$stream,0)|0);
 $3 = ($2|0)==(0);
 if (!($3)) {
  $$0 = $2;
  return ($$0|0);
 }
 $4 = ((($face)) + 468|0);
 $5 = (_FT_Stream_ReadFields($stream,174598,$4)|0);
 $$0 = $5;
 return ($$0|0);
}
function _tt_face_load_name($face,$stream) {
 $face = $face|0;
 $stream = $stream|0;
 var $$pre = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $count$01 = 0, $entry$0$lcssa = 0, $entry$02 = 0, $entry$1 = 0;
 var $entry$1$lcssa = 0, $error = 0, $table_len = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp + 4|0;
 $table_len = sp;
 $0 = ((($stream)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 360|0);
 HEAP32[$2>>2] = $stream;
 $3 = ((($face)) + 508|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (FUNCTION_TABLE_iiiii[$4 & 8191]($face,1851878757,$stream,$table_len)|0);
 HEAP32[$error>>2] = $5;
 $6 = ($5|0)==(0);
 if (!($6)) {
  $53 = $5;
  STACKTOP = sp;return ($53|0);
 }
 $7 = ((($face)) + 344|0);
 $8 = (_FT_Stream_Pos($stream)|0);
 $9 = (_FT_Stream_ReadFields($stream,174642,$7)|0);
 HEAP32[$error>>2] = $9;
 $10 = ($9|0)==(0);
 if (!($10)) {
  $53 = $9;
  STACKTOP = sp;return ($53|0);
 }
 $11 = (($8) + 6)|0;
 $12 = ((($face)) + 348|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13*12)|0;
 $15 = (($11) + ($14))|0;
 $16 = HEAP32[$table_len>>2]|0;
 $17 = (($16) + ($8))|0;
 $18 = ($15>>>0)>($17>>>0);
 if ($18) {
  HEAP32[$error>>2] = 145;
  $53 = 145;
  STACKTOP = sp;return ($53|0);
 }
 HEAP32[$12>>2] = 0;
 $19 = (_ft_mem_realloc($1,20,0,$13,0,$error)|0);
 $20 = ((($face)) + 356|0);
 HEAP32[$20>>2] = $19;
 $21 = HEAP32[$error>>2]|0;
 $22 = ($21|0)==(0);
 if (!($22)) {
  $53 = $21;
  STACKTOP = sp;return ($53|0);
 }
 $23 = (_FT_Stream_EnterFrame($stream,$14)|0);
 HEAP32[$error>>2] = $23;
 $24 = ($23|0)==(0);
 if (!($24)) {
  $53 = $23;
  STACKTOP = sp;return ($53|0);
 }
 $25 = HEAP32[$20>>2]|0;
 $26 = ($13|0)==(0);
 if ($26) {
  $27 = $25;
  $48 = $27;$entry$0$lcssa = $25;
 } else {
  $28 = ((($face)) + 352|0);
  $count$01 = $13;$entry$02 = $25;
  while(1) {
   $29 = (_FT_Stream_ReadFields($stream,174662,$entry$02)|0);
   HEAP32[$error>>2] = $29;
   $30 = ($29|0)==(0);
   do {
    if ($30) {
     $31 = ((($entry$02)) + 8|0);
     $32 = HEAP16[$31>>1]|0;
     $33 = ($32<<16>>16)==(0);
     if ($33) {
      $entry$1 = $entry$02;
     } else {
      $34 = HEAP32[$28>>2]|0;
      $35 = (($34) + ($8))|0;
      $36 = ((($entry$02)) + 12|0);
      $37 = HEAP32[$36>>2]|0;
      $38 = (($35) + ($37))|0;
      HEAP32[$36>>2] = $38;
      $39 = ($38>>>0)<($15>>>0);
      if (!($39)) {
       $40 = $32&65535;
       $41 = (($40) + ($38))|0;
       $42 = ($41>>>0)>($17>>>0);
       if (!($42)) {
        $43 = ((($entry$02)) + 20|0);
        $entry$1 = $43;
        break;
       }
      }
      HEAP32[$36>>2] = 0;
      HEAP16[$31>>1] = 0;
      $entry$1 = $entry$02;
     }
    } else {
     $entry$1 = $entry$02;
    }
   } while(0);
   $44 = (($count$01) + -1)|0;
   $45 = ($44|0)==(0);
   if ($45) {
    $entry$1$lcssa = $entry$1;
    break;
   } else {
    $count$01 = $44;$entry$02 = $entry$1;
   }
  }
  $$pre = HEAP32[$20>>2]|0;
  $48 = $$pre;$entry$0$lcssa = $entry$1$lcssa;
 }
 $46 = $entry$0$lcssa;
 $47 = (($46) - ($48))|0;
 $49 = (($47|0) / 20)&-1;
 HEAP32[$12>>2] = $49;
 _FT_Stream_ExitFrame($stream);
 $50 = HEAP32[$12>>2]|0;
 $51 = $50&65535;
 $52 = ((($face)) + 340|0);
 HEAP16[$52>>1] = $51;
 $$pre3 = HEAP32[$error>>2]|0;
 $53 = $$pre3;
 STACKTOP = sp;return ($53|0);
}
function _tt_face_free_name($face) {
 $face = $face|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $count$02 = 0, $entry$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 96|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($face)) + 356|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($face)) + 348|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($5|0)==(0|0);
 if ($8) {
  HEAP32[$6>>2] = 0;
  $17 = ((($face)) + 344|0);
  HEAP16[$17>>1] = 0;
  $18 = ((($face)) + 352|0);
  HEAP32[$18>>2] = 0;
  return;
 }
 $9 = ($7|0)==(0);
 if ($9) {
  $16 = $5;
 } else {
  $count$02 = $7;$entry$01 = $5;
  while(1) {
   $10 = ((($entry$01)) + 16|0);
   $11 = HEAP32[$10>>2]|0;
   _ft_mem_free($3,$11);
   HEAP32[$10>>2] = 0;
   $12 = ((($entry$01)) + 8|0);
   HEAP16[$12>>1] = 0;
   $13 = (($count$02) + -1)|0;
   $14 = ((($entry$01)) + 20|0);
   $15 = ($13|0)==(0);
   if ($15) {
    break;
   } else {
    $count$02 = $13;$entry$01 = $14;
   }
  }
  $$pre = HEAP32[$4>>2]|0;
  $16 = $$pre;
 }
 _ft_mem_free($3,$16);
 HEAP32[$4>>2] = 0;
 HEAP32[$6>>2] = 0;
 $17 = ((($face)) + 344|0);
 HEAP16[$17>>1] = 0;
 $18 = ((($face)) + 352|0);
 HEAP32[$18>>2] = 0;
 return;
}
function _tt_face_load_kern($face,$stream) {
 $face = $face|0;
 $stream = $stream|0;
 var $$ = 0, $$2 = 0, $$masked = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $avail$0$lcssa = 0, $avail$07 = 0, $avail$2$ph = 0, $cond = 0, $count$0 = 0, $count$0$in = 0, $error$0 = 0, $nn$0$lcssa = 0, $nn$09 = 0, $num_pairs$0 = 0;
 var $old_pair$0 = 0, $or$cond = 0, $ordered$0$lcssa = 0, $ordered$08 = 0, $ordered$3$ph = 0, $p$010 = 0, $p$1 = 0, $table_size = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $table_size = sp;
 $0 = ((($face)) + 508|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (FUNCTION_TABLE_iiiii[$1 & 8191]($face,1801810542,$stream,$table_size)|0);
 $3 = ($2|0)==(0);
 if (!($3)) {
  $error$0 = $2;
  STACKTOP = sp;return ($error$0|0);
 }
 $4 = HEAP32[$table_size>>2]|0;
 $5 = ($4>>>0)<(4);
 if ($5) {
  $error$0 = 142;
  STACKTOP = sp;return ($error$0|0);
 }
 $6 = ((($face)) + 740|0);
 $7 = (_FT_Stream_ExtractFrame($stream,$4,$6)|0);
 $8 = ($7|0)==(0);
 if (!($8)) {
  $error$0 = $7;
  STACKTOP = sp;return ($error$0|0);
 }
 $9 = HEAP32[$table_size>>2]|0;
 $10 = ((($face)) + 744|0);
 HEAP32[$10>>2] = $9;
 $11 = HEAP32[$6>>2]|0;
 $12 = (($11) + ($9)|0);
 $13 = ((($11)) + 2|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = $14&255;
 $16 = $15 << 8;
 $17 = ((($11)) + 3|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = $18&255;
 $20 = $16 | $19;
 $21 = ($20>>>0)>(32);
 $$ = $21 ? 32 : $20;
 $22 = ($$|0)==(0);
 L10: do {
  if ($22) {
   $avail$0$lcssa = 0;$nn$0$lcssa = 0;$ordered$0$lcssa = 0;
  } else {
   $23 = ((($11)) + 4|0);
   $avail$07 = 0;$nn$09 = 0;$ordered$08 = 0;$p$010 = $23;
   while(1) {
    $24 = 1 << $nn$09;
    $25 = ((($p$010)) + 6|0);
    $26 = ($25>>>0)>($12>>>0);
    if ($26) {
     $avail$0$lcssa = $avail$07;$nn$0$lcssa = $nn$09;$ordered$0$lcssa = $ordered$08;
     break L10;
    }
    $27 = ((($p$010)) + 2|0);
    $28 = HEAP8[$27>>0]|0;
    $29 = $28&255;
    $30 = $29 << 8;
    $31 = ((($p$010)) + 3|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    $34 = $30 | $33;
    $35 = ($34>>>0)<(15);
    if ($35) {
     $avail$0$lcssa = $avail$07;$nn$0$lcssa = $nn$09;$ordered$0$lcssa = $ordered$08;
     break L10;
    }
    $36 = ((($p$010)) + 5|0);
    $37 = HEAP8[$36>>0]|0;
    $38 = $37&255;
    $39 = ((($p$010)) + 4|0);
    $40 = HEAP8[$39>>0]|0;
    $41 = $40&255;
    $42 = $41 << 8;
    $43 = (($p$010) + ($34)|0);
    $44 = ($43>>>0)>($12>>>0);
    $$2 = $44 ? $12 : $43;
    $$masked = $38 & 247;
    $45 = $42 | $$masked;
    $46 = ($45|0)!=(1);
    $47 = ((($p$010)) + 14|0);
    $48 = ($47>>>0)>($12>>>0);
    $or$cond = $48 | $46;
    L16: do {
     if ($or$cond) {
      $avail$2$ph = $avail$07;$ordered$3$ph = $ordered$08;
     } else {
      $49 = HEAP8[$25>>0]|0;
      $50 = $49&255;
      $51 = $50 << 8;
      $52 = ((($p$010)) + 7|0);
      $53 = HEAP8[$52>>0]|0;
      $54 = $53&255;
      $55 = $51 | $54;
      $56 = $$2;
      $57 = $47;
      $58 = (($56) - ($57))|0;
      $59 = ($55*6)|0;
      $60 = ($58|0)<($59|0);
      $61 = (($58|0) / 6)&-1;
      $num_pairs$0 = $60 ? $61 : $55;
      $62 = $24 | $avail$07;
      $63 = ($num_pairs$0|0)==(0);
      if ($63) {
       $avail$2$ph = $62;$ordered$3$ph = $ordered$08;
      } else {
       $64 = HEAP8[$47>>0]|0;
       $65 = $64&255;
       $66 = $65 << 24;
       $67 = ((($p$010)) + 15|0);
       $68 = HEAP8[$67>>0]|0;
       $69 = $68&255;
       $70 = $69 << 16;
       $71 = $70 | $66;
       $72 = ((($p$010)) + 16|0);
       $73 = HEAP8[$72>>0]|0;
       $74 = $73&255;
       $75 = $74 << 8;
       $76 = $71 | $75;
       $77 = ((($p$010)) + 17|0);
       $78 = HEAP8[$77>>0]|0;
       $79 = $78&255;
       $80 = $76 | $79;
       $81 = ((($p$010)) + 20|0);
       $count$0$in = $num_pairs$0;$old_pair$0 = $80;$p$1 = $81;
       while(1) {
        $count$0 = (($count$0$in) + -1)|0;
        $cond = ($count$0|0)==(0);
        if ($cond) {
         break;
        }
        $82 = HEAP8[$p$1>>0]|0;
        $83 = $82&255;
        $84 = $83 << 24;
        $85 = ((($p$1)) + 1|0);
        $86 = HEAP8[$85>>0]|0;
        $87 = $86&255;
        $88 = $87 << 16;
        $89 = $88 | $84;
        $90 = ((($p$1)) + 2|0);
        $91 = HEAP8[$90>>0]|0;
        $92 = $91&255;
        $93 = $92 << 8;
        $94 = $89 | $93;
        $95 = ((($p$1)) + 3|0);
        $96 = HEAP8[$95>>0]|0;
        $97 = $96&255;
        $98 = $94 | $97;
        $99 = ($98>>>0)>($old_pair$0>>>0);
        $100 = ((($p$1)) + 6|0);
        if ($99) {
         $count$0$in = $count$0;$old_pair$0 = $98;$p$1 = $100;
        } else {
         $avail$2$ph = $62;$ordered$3$ph = $ordered$08;
         break L16;
        }
       }
       $101 = $24 | $ordered$08;
       $avail$2$ph = $62;$ordered$3$ph = $101;
      }
     }
    } while(0);
    $102 = (($nn$09) + 1)|0;
    $103 = ($102>>>0)<($$>>>0);
    if ($103) {
     $avail$07 = $avail$2$ph;$nn$09 = $102;$ordered$08 = $ordered$3$ph;$p$010 = $$2;
    } else {
     $avail$0$lcssa = $avail$2$ph;$nn$0$lcssa = $102;$ordered$0$lcssa = $ordered$3$ph;
     break;
    }
   }
  }
 } while(0);
 $104 = ((($face)) + 748|0);
 HEAP32[$104>>2] = $nn$0$lcssa;
 $105 = ((($face)) + 752|0);
 HEAP32[$105>>2] = $avail$0$lcssa;
 $106 = ((($face)) + 756|0);
 HEAP32[$106>>2] = $ordered$0$lcssa;
 $error$0 = 0;
 STACKTOP = sp;return ($error$0|0);
}
function _tt_face_load_gasp($face,$stream) {
 $face = $face|0;
 $stream = $stream|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $exitcond = 0, $j$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($stream)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 508|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiiii[$3 & 8191]($face,1734439792,$stream,0)|0);
 HEAP32[$error>>2] = $4;
 $5 = ($4|0)==(0);
 if (!($5)) {
  $30 = $4;
  STACKTOP = sp;return ($30|0);
 }
 $6 = (_FT_Stream_EnterFrame($stream,4)|0);
 HEAP32[$error>>2] = $6;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $30 = $6;
  STACKTOP = sp;return ($30|0);
 }
 $8 = (_FT_Stream_GetUShort($stream)|0);
 $9 = ((($face)) + 540|0);
 HEAP16[$9>>1] = $8;
 $10 = (_FT_Stream_GetUShort($stream)|0);
 $11 = ((($face)) + 542|0);
 HEAP16[$11>>1] = $10;
 _FT_Stream_ExitFrame($stream);
 $12 = HEAP32[$9>>2]|0;
 $13 = $12&65535;
 $14 = ($13&65535)>(1);
 $15 = $12 >>> 16;
 if ($14) {
  HEAP16[$11>>1] = 0;
  HEAP32[$error>>2] = 8;
  $30 = 8;
  STACKTOP = sp;return ($30|0);
 }
 $16 = (_ft_mem_realloc($1,4,0,$15,0,$error)|0);
 $17 = ((($face)) + 544|0);
 HEAP32[$17>>2] = $16;
 $18 = HEAP32[$error>>2]|0;
 $19 = ($18|0)==(0);
 if (!($19)) {
  $30 = $18;
  STACKTOP = sp;return ($30|0);
 }
 $20 = $15 << 2;
 $21 = (_FT_Stream_EnterFrame($stream,$20)|0);
 HEAP32[$error>>2] = $21;
 $22 = ($21|0)==(0);
 if (!($22)) {
  $30 = $21;
  STACKTOP = sp;return ($30|0);
 }
 $23 = HEAP32[$17>>2]|0;
 $24 = ($15|0)==(0);
 if (!($24)) {
  $j$01 = 0;
  while(1) {
   $25 = (_FT_Stream_GetUShort($stream)|0);
   $26 = (($23) + ($j$01<<2)|0);
   HEAP16[$26>>1] = $25;
   $27 = (_FT_Stream_GetUShort($stream)|0);
   $28 = (((($23) + ($j$01<<2)|0)) + 2|0);
   HEAP16[$28>>1] = $27;
   $29 = (($j$01) + 1)|0;
   $exitcond = ($29|0)==($15|0);
   if ($exitcond) {
    break;
   } else {
    $j$01 = $29;
   }
  }
 }
 _FT_Stream_ExitFrame($stream);
 $$pre = HEAP32[$error>>2]|0;
 $30 = $$pre;
 STACKTOP = sp;return ($30|0);
}
function _tt_face_load_pclt($face,$stream) {
 $face = $face|0;
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $error$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 508|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (FUNCTION_TABLE_iiiii[$1 & 8191]($face,1346587732,$stream,0)|0);
 $3 = ($2|0)==(0);
 if (!($3)) {
  $error$0 = $2;
  return ($error$0|0);
 }
 $4 = ((($face)) + 548|0);
 $5 = (_FT_Stream_ReadFields($stream,174690,$4)|0);
 $error$0 = $5;
 return ($error$0|0);
}
function _tt_face_load_bhed($face,$stream) {
 $face = $face|0;
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $error$0$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 508|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (FUNCTION_TABLE_iiiii[$1 & 8191]($face,1651008868,$stream,0)|0);
 $3 = ($2|0)==(0);
 if (!($3)) {
  $error$0$i = $2;
  return ($error$0$i|0);
 }
 $4 = ((($face)) + 160|0);
 $5 = (_FT_Stream_ReadFields($stream,174126,$4)|0);
 $error$0$i = $5;
 return ($error$0$i|0);
}
function _tt_face_load_sbit_image($face,$strike_index,$glyph_index,$load_flags,$stream,$map,$metrics) {
 $face = $face|0;
 $strike_index = $strike_index|0;
 $glyph_index = $glyph_index|0;
 $load_flags = $load_flags|0;
 $stream = $stream|0;
 $map = $map|0;
 $metrics = $metrics|0;
 var $$0$i = 0, $$01$i$ph = 0, $$012$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $decoder = 0, $ebdt_size$i = 0, $error$2 = 0, $error$2$i = 0, $new_map = 0, $recurse_depth$03$i = 0, $switch$split12D = 0, $switch$split2D = 0, $switch$split42D = 0, $table_size$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $table_size$i = sp + 72|0;
 $ebdt_size$i = sp + 68|0;
 $decoder = sp + 24|0;
 $new_map = sp;
 $0 = ((($face)) + 732|0);
 $1 = HEAP32[$0>>2]|0;
 switch ($1|0) {
 case 2: case 1:  {
  $2 = ((($face)) + 104|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($face)) + 508|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = (FUNCTION_TABLE_iiiii[$5 & 8191]($face,1128416340,$3,$ebdt_size$i)|0);
  $7 = ($6|0)==(0);
  if ($7) {
   label = 5;
  } else {
   $8 = HEAP32[$4>>2]|0;
   $9 = (FUNCTION_TABLE_iiiii[$8 & 8191]($face,1161970772,$3,$ebdt_size$i)|0);
   $10 = ($9|0)==(0);
   if ($10) {
    label = 5;
   } else {
    $11 = HEAP32[$4>>2]|0;
    $12 = (FUNCTION_TABLE_iiiii[$11 & 8191]($face,1650745716,$3,$ebdt_size$i)|0);
    $13 = ($12|0)==(0);
    if ($13) {
     label = 5;
    } else {
     $$01$i$ph = $12;
    }
   }
  }
  if ((label|0) == 5) {
   HEAP32[$decoder>>2] = $face;
   $14 = ((($decoder)) + 4|0);
   HEAP32[$14>>2] = $3;
   $15 = ((($face)) + 84|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = ((($16)) + 76|0);
   $18 = ((($decoder)) + 8|0);
   HEAP32[$18>>2] = $17;
   $19 = ((($decoder)) + 12|0);
   HEAP32[$19>>2] = $metrics;
   $20 = ((($decoder)) + 16|0);
   HEAP8[$20>>0] = 0;
   $21 = ((($decoder)) + 17|0);
   HEAP8[$21>>0] = 0;
   $22 = (_FT_Stream_Pos($3)|0);
   $23 = ((($decoder)) + 20|0);
   HEAP32[$23>>2] = $22;
   $24 = HEAP32[$ebdt_size$i>>2]|0;
   $25 = ((($decoder)) + 24|0);
   HEAP32[$25>>2] = $24;
   $26 = ((($face)) + 724|0);
   $27 = HEAP32[$26>>2]|0;
   $28 = ((($decoder)) + 36|0);
   HEAP32[$28>>2] = $27;
   $29 = $27;
   $30 = ((($face)) + 728|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = (($29) + ($31)|0);
   $33 = ((($decoder)) + 40|0);
   HEAP32[$33>>2] = $32;
   $34 = ($strike_index*48)|0;
   $35 = $34 | 8;
   $36 = (($35) + 47)|0;
   $37 = ($36>>>0)>($31>>>0);
   if ($37) {
    $$01$i$ph = 3;
   } else {
    $38 = ((($29)) + 8|0);
    $39 = (($38) + ($34)|0);
    $40 = ((($39)) + 4|0);
    $41 = ((($40)) + -4|0);
    $42 = HEAP8[$41>>0]|0;
    $43 = $42&255;
    $44 = $43 << 24;
    $45 = ((($41)) + 1|0);
    $46 = HEAP8[$45>>0]|0;
    $47 = $46&255;
    $48 = $47 << 16;
    $49 = $48 | $44;
    $50 = ((($41)) + 2|0);
    $51 = HEAP8[$50>>0]|0;
    $52 = $51&255;
    $53 = $52 << 8;
    $54 = $49 | $53;
    $55 = ((($41)) + 3|0);
    $56 = HEAP8[$55>>0]|0;
    $57 = $56&255;
    $58 = $54 | $57;
    $59 = ((($decoder)) + 28|0);
    HEAP32[$59>>2] = $58;
    $60 = ((($40)) + 4|0);
    $61 = ((($60)) + 4|0);
    $62 = ((($61)) + -4|0);
    $63 = HEAP8[$62>>0]|0;
    $64 = $63&255;
    $65 = $64 << 24;
    $66 = ((($62)) + 1|0);
    $67 = HEAP8[$66>>0]|0;
    $68 = $67&255;
    $69 = $68 << 16;
    $70 = $69 | $65;
    $71 = ((($62)) + 2|0);
    $72 = HEAP8[$71>>0]|0;
    $73 = $72&255;
    $74 = $73 << 8;
    $75 = $70 | $74;
    $76 = ((($62)) + 3|0);
    $77 = HEAP8[$76>>0]|0;
    $78 = $77&255;
    $79 = $75 | $78;
    $80 = ((($decoder)) + 32|0);
    HEAP32[$80>>2] = $79;
    $81 = ((($61)) + 34|0);
    $82 = HEAP8[$81>>0]|0;
    $83 = ((($decoder)) + 18|0);
    HEAP8[$83>>0] = $82;
    $84 = ($31>>>0)<($58>>>0);
    if ($84) {
     $$01$i$ph = 3;
    } else {
     $85 = (($31) - ($58))|0;
     $86 = $85 >>> 3;
     $87 = ($79>>>0)>($86>>>0);
     if ($87) {
      $$01$i$ph = 3;
     } else {
      $155 = (_tt_sbit_decoder_load_image($decoder,$glyph_index,0,0)|0);
      $156 = $load_flags & 1048576;
      $157 = $155 | $156;
      $158 = ($157|0)==(0);
      if (!($158)) {
       $error$2 = $155;
       STACKTOP = sp;return ($error$2|0);
      }
      $159 = ((($map)) + 18|0);
      $160 = HEAP8[$159>>0]|0;
      $161 = ($160<<24>>24)==(7);
      if (!($161)) {
       $error$2 = $155;
       STACKTOP = sp;return ($error$2|0);
      }
      $162 = HEAP32[$15>>2]|0;
      $163 = HEAP32[$162>>2]|0;
      _FT_Bitmap_Init($new_map);
      $164 = (_FT_Bitmap_Convert($163,$map,$new_map,1)|0);
      $165 = ($164|0)==(0);
      if ($165) {
       $166 = ((($new_map)) + 18|0);
       $167 = HEAP8[$166>>0]|0;
       HEAP8[$159>>0] = $167;
       $168 = ((($new_map)) + 8|0);
       $169 = HEAP32[$168>>2]|0;
       $170 = ((($map)) + 8|0);
       HEAP32[$170>>2] = $169;
       $171 = ((($new_map)) + 16|0);
       $172 = HEAP16[$171>>1]|0;
       $173 = ((($map)) + 16|0);
       HEAP16[$173>>1] = $172;
       $174 = HEAP32[$15>>2]|0;
       $175 = ((($new_map)) + 12|0);
       $176 = HEAP32[$175>>2]|0;
       _ft_glyphslot_set_bitmap($174,$176);
       $177 = HEAP32[$15>>2]|0;
       $178 = ((($177)) + 156|0);
       $179 = HEAP32[$178>>2]|0;
       $180 = ((($179)) + 4|0);
       $181 = HEAP32[$180>>2]|0;
       $182 = $181 | 1;
       HEAP32[$180>>2] = $182;
      } else {
       (_FT_Bitmap_Done($163,$new_map)|0);
      }
      $error$2 = $164;
      STACKTOP = sp;return ($error$2|0);
     }
    }
   }
  }
  $error$2 = $$01$i$ph;
  STACKTOP = sp;return ($error$2|0);
  break;
 }
 case 3:  {
  $88 = ((($metrics)) + 2|0);
  HEAP16[$88>>1] = 0;
  HEAP16[$metrics>>1] = 0;
  $89 = ((($face)) + 724|0);
  $90 = HEAP32[$89>>2]|0;
  $91 = ((($90)) + 8|0);
  $92 = $strike_index << 2;
  $93 = (($91) + ($92)|0);
  $94 = ((($93)) + 4|0);
  $95 = ((($94)) + -4|0);
  $96 = HEAP8[$95>>0]|0;
  $97 = $96&255;
  $98 = $97 << 24;
  $99 = ((($95)) + 1|0);
  $100 = HEAP8[$99>>0]|0;
  $101 = $100&255;
  $102 = $101 << 16;
  $103 = $102 | $98;
  $104 = ((($95)) + 2|0);
  $105 = HEAP8[$104>>0]|0;
  $106 = $105&255;
  $107 = $106 << 8;
  $108 = $103 | $107;
  $109 = ((($95)) + 3|0);
  $110 = HEAP8[$109>>0]|0;
  $111 = $110&255;
  $112 = $108 | $111;
  $113 = ((($face)) + 508|0);
  $114 = HEAP32[$113>>2]|0;
  $115 = (FUNCTION_TABLE_iiiii[$114 & 8191]($face,1935829368,$stream,$table_size$i)|0);
  $116 = ($115|0)==(0);
  L25: do {
   if ($116) {
    $117 = (_FT_Stream_Pos($stream)|0);
    $118 = ((($face)) + 16|0);
    $119 = HEAP32[$118>>2]|0;
    $120 = ($119>>>0)<($glyph_index>>>0);
    if ($120) {
     $$0$i = 6;
    } else {
     $121 = (($117) + ($112))|0;
     $122 = (($121) + 4)|0;
     $$012$i = $glyph_index;$recurse_depth$03$i = 0;
     L28: while(1) {
      $123 = HEAP32[$table_size$i>>2]|0;
      $124 = ($123>>>0)>($112>>>0);
      if (!($124)) {
       $$0$i = 3;
       break L25;
      }
      $125 = (($123) - ($112))|0;
      $126 = $$012$i << 2;
      $127 = (($126) + 12)|0;
      $128 = ($125>>>0)<($127>>>0);
      if ($128) {
       $$0$i = 3;
       break L25;
      }
      $129 = (($122) + ($126))|0;
      $130 = (_FT_Stream_Seek($stream,$129)|0);
      $131 = ($130|0)==(0);
      if (!($131)) {
       $$0$i = $130;
       break L25;
      }
      $132 = (_FT_Stream_EnterFrame($stream,8)|0);
      $133 = ($132|0)==(0);
      if (!($133)) {
       $$0$i = $132;
       break L25;
      }
      $134 = (_FT_Stream_GetULong($stream)|0);
      $135 = (_FT_Stream_GetULong($stream)|0);
      _FT_Stream_ExitFrame($stream);
      $136 = ($135|0)==($134|0);
      if ($136) {
       $$0$i = 6;
       break L25;
      }
      $137 = ($135>>>0)<($134>>>0);
      if ($137) {
       $$0$i = 3;
       break L25;
      }
      $138 = (($135) - ($134))|0;
      $139 = ($138>>>0)<(8);
      if ($139) {
       $$0$i = 3;
       break L25;
      }
      $140 = HEAP32[$table_size$i>>2]|0;
      $141 = (($140) - ($112))|0;
      $142 = ($141>>>0)<($135>>>0);
      if ($142) {
       $$0$i = 3;
       break L25;
      }
      $143 = (($134) + ($121))|0;
      $144 = (_FT_Stream_Seek($stream,$143)|0);
      $145 = ($144|0)==(0);
      if (!($145)) {
       $$0$i = $144;
       break L25;
      }
      $146 = (_FT_Stream_EnterFrame($stream,$138)|0);
      $147 = ($146|0)==(0);
      if (!($147)) {
       $$0$i = $146;
       break L25;
      }
      (_FT_Stream_GetUShort($stream)|0);
      (_FT_Stream_GetUShort($stream)|0);
      $148 = (_FT_Stream_GetULong($stream)|0);
      $switch$split2D = ($148|0)<(1919378028);
      if (!($switch$split2D)) {
       label = 36;
       break;
      }
      $switch$split12D = ($148|0)<(1785751328);
      if (!($switch$split12D)) {
       label = 38;
       break;
      }
      switch ($148|0) {
      case 1685418085:  {
       break;
      }
      default: {
       $error$2$i = 7;
       break L28;
      }
      }
      $149 = ($recurse_depth$03$i|0)<(4);
      if (!($149)) {
       $error$2$i = 3;
       break;
      }
      $150 = (_FT_Stream_GetUShort($stream)|0);
      $151 = $150&65535;
      _FT_Stream_ExitFrame($stream);
      $152 = (($recurse_depth$03$i) + 1)|0;
      $153 = HEAP32[$118>>2]|0;
      $154 = ($151>>>0)>($153>>>0);
      if ($154) {
       $$0$i = 6;
       break L25;
      } else {
       $$012$i = $151;$recurse_depth$03$i = $152;
      }
     }
     L44: do {
      if ((label|0) == 36) {
       $switch$split42D = ($148|0)<(1953064550);
       if ($switch$split42D) {
        switch ($148|0) {
        case 1919378028:  {
         label = 25;
         break L44;
         break;
        }
        default: {
         $error$2$i = 7;
         break L44;
        }
        }
       } else {
        switch ($148|0) {
        case 1953064550:  {
         label = 25;
         break L44;
         break;
        }
        default: {
         $error$2$i = 7;
         break L44;
        }
        }
       }
      }
      else if ((label|0) == 38) {
       switch ($148|0) {
       case 1785751328:  {
        label = 25;
        break;
       }
       default: {
        $error$2$i = 7;
       }
       }
      }
     } while(0);
     if ((label|0) == 25) {
      $error$2$i = 2;
     }
     _FT_Stream_ExitFrame($stream);
     $$0$i = $error$2$i;
    }
   } else {
    $$0$i = $115;
   }
  } while(0);
  $error$2 = $$0$i;
  STACKTOP = sp;return ($error$2|0);
  break;
 }
 default: {
  $error$2 = 2;
  STACKTOP = sp;return ($error$2|0);
 }
 }
 return (0)|0;
}
function _tt_sbit_decoder_load_image($decoder,$glyph_index,$x_pos,$y_pos) {
 $decoder = $decoder|0;
 $glyph_index = $glyph_index|0;
 $x_pos = $x_pos|0;
 $y_pos = $y_pos|0;
 var $$0 = 0, $$01$i = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0;
 var $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0;
 var $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0;
 var $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0;
 var $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0;
 var $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0;
 var $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0;
 var $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0;
 var $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0;
 var $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0;
 var $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0;
 var $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0;
 var $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0;
 var $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0;
 var $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0;
 var $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $data$i = 0, $error$2$i = 0, $image_end$6 = 0, $image_start$6 = 0, $loader$1$i = 0, $mm$033 = 0, $mm4$035 = 0, $mm4$035$lcssa = 0, $not$ = 0, $num_ranges$038 = 0, $or$cond = 0, $or$cond40 = 0, $p$037 = 0, $p$037$lcssa = 0, $p$132 = 0, $p$132$lcssa = 0, $p$434 = 0, $p$sroa$0$2$ph$i = 0, $p$sroa$0$2$ph$in$i = 0, $p$sroa$0$3$i = 0;
 var $p$sroa$0$4$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $data$i = sp;
 $0 = ((($decoder)) + 36|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($decoder)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($decoder)) + 40|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($decoder)) + 32|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0);
 if ($9) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 } else {
  $num_ranges$038 = $8;$p$037 = $4;
 }
 while(1) {
  $10 = HEAP8[$p$037>>0]|0;
  $11 = $10&255;
  $12 = $11 << 8;
  $13 = ((($p$037)) + 1|0);
  $14 = HEAP8[$13>>0]|0;
  $15 = $14&255;
  $16 = $12 | $15;
  $17 = ($16>>>0)>($glyph_index>>>0);
  if (!($17)) {
   $18 = ((($p$037)) + 2|0);
   $19 = HEAP8[$18>>0]|0;
   $20 = $19&255;
   $21 = $20 << 8;
   $22 = ((($p$037)) + 3|0);
   $23 = HEAP8[$22>>0]|0;
   $24 = $23&255;
   $25 = $21 | $24;
   $26 = ($25>>>0)<($glyph_index>>>0);
   if (!($26)) {
    $$lcssa = $16;$p$037$lcssa = $p$037;
    break;
   }
  }
  $27 = ((($p$037)) + 8|0);
  $28 = (($num_ranges$038) + -1)|0;
  $29 = ($28|0)==(0);
  if ($29) {
   $$0 = 6;
   label = 62;
   break;
  } else {
   $num_ranges$038 = $28;$p$037 = $27;
  }
 }
 if ((label|0) == 62) {
  STACKTOP = sp;return ($$0|0);
 }
 $30 = ((($p$037$lcssa)) + 4|0);
 $31 = HEAP8[$30>>0]|0;
 $32 = $31&255;
 $33 = $32 << 24;
 $34 = ((($p$037$lcssa)) + 5|0);
 $35 = HEAP8[$34>>0]|0;
 $36 = $35&255;
 $37 = $36 << 16;
 $38 = $37 | $33;
 $39 = ((($p$037$lcssa)) + 6|0);
 $40 = HEAP8[$39>>0]|0;
 $41 = $40&255;
 $42 = $41 << 8;
 $43 = $38 | $42;
 $44 = ((($p$037$lcssa)) + 7|0);
 $45 = HEAP8[$44>>0]|0;
 $46 = $45&255;
 $47 = $43 | $46;
 $48 = $6;
 $49 = $4;
 $50 = (($48) - ($49))|0;
 $51 = ($47>>>0)>($50>>>0);
 if ($51) {
  $$0 = 8;
  STACKTOP = sp;return ($$0|0);
 }
 $52 = (($4) + ($47)|0);
 $53 = ((($52)) + 8|0);
 $54 = ($53>>>0)>($6>>>0);
 if ($54) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $55 = ((($52)) + 2|0);
 $56 = ((($55)) + -2|0);
 $57 = HEAP8[$56>>0]|0;
 $58 = $57&255;
 $59 = $58 << 8;
 $60 = ((($56)) + 1|0);
 $61 = HEAP8[$60>>0]|0;
 $62 = $61&255;
 $63 = $59 | $62;
 $64 = ((($55)) + 2|0);
 $65 = ((($64)) + -2|0);
 $66 = HEAP8[$65>>0]|0;
 $67 = $66&255;
 $68 = $67 << 8;
 $69 = ((($65)) + 1|0);
 $70 = HEAP8[$69>>0]|0;
 $71 = $70&255;
 $72 = $68 | $71;
 $73 = ((($64)) + 4|0);
 $74 = ((($73)) + -4|0);
 $75 = HEAP8[$74>>0]|0;
 $76 = $75&255;
 $77 = $76 << 24;
 $78 = ((($74)) + 1|0);
 $79 = HEAP8[$78>>0]|0;
 $80 = $79&255;
 $81 = $80 << 16;
 $82 = $81 | $77;
 $83 = ((($74)) + 2|0);
 $84 = HEAP8[$83>>0]|0;
 $85 = $84&255;
 $86 = $85 << 8;
 $87 = $82 | $86;
 $88 = ((($74)) + 3|0);
 $89 = HEAP8[$88>>0]|0;
 $90 = $89&255;
 $91 = $87 | $90;
 L17: do {
  switch ($63|0) {
  case 1:  {
   $92 = (($glyph_index) - ($$lcssa))|0;
   $93 = $92 << 2;
   $94 = (($73) + ($93)|0);
   $95 = ((($94)) + 8|0);
   $96 = ($95>>>0)>($6>>>0);
   if ($96) {
    $$0 = 6;
    STACKTOP = sp;return ($$0|0);
   }
   $97 = ((($94)) + 4|0);
   $98 = ((($97)) + -4|0);
   $99 = HEAP8[$98>>0]|0;
   $100 = $99&255;
   $101 = $100 << 24;
   $102 = ((($98)) + 1|0);
   $103 = HEAP8[$102>>0]|0;
   $104 = $103&255;
   $105 = $104 << 16;
   $106 = $105 | $101;
   $107 = ((($98)) + 2|0);
   $108 = HEAP8[$107>>0]|0;
   $109 = $108&255;
   $110 = $109 << 8;
   $111 = $106 | $110;
   $112 = ((($98)) + 3|0);
   $113 = HEAP8[$112>>0]|0;
   $114 = $113&255;
   $115 = $111 | $114;
   $116 = ((($97)) + 4|0);
   $117 = ((($116)) + -4|0);
   $118 = HEAP8[$117>>0]|0;
   $119 = $118&255;
   $120 = $119 << 24;
   $121 = ((($117)) + 1|0);
   $122 = HEAP8[$121>>0]|0;
   $123 = $122&255;
   $124 = $123 << 16;
   $125 = $124 | $120;
   $126 = ((($117)) + 2|0);
   $127 = HEAP8[$126>>0]|0;
   $128 = $127&255;
   $129 = $128 << 8;
   $130 = $125 | $129;
   $131 = ((($117)) + 3|0);
   $132 = HEAP8[$131>>0]|0;
   $133 = $132&255;
   $134 = $130 | $133;
   $135 = ($115|0)==($134|0);
   if ($135) {
    $$0 = 6;
    STACKTOP = sp;return ($$0|0);
   } else {
    $image_end$6 = $134;$image_start$6 = $115;
   }
   break;
  }
  case 2:  {
   $136 = ((($73)) + 12|0);
   $137 = ($136>>>0)>($6>>>0);
   if ($137) {
    $$0 = 6;
    STACKTOP = sp;return ($$0|0);
   }
   $138 = ((($73)) + 4|0);
   $139 = ((($138)) + -4|0);
   $140 = HEAP8[$139>>0]|0;
   $141 = $140&255;
   $142 = $141 << 24;
   $143 = ((($139)) + 1|0);
   $144 = HEAP8[$143>>0]|0;
   $145 = $144&255;
   $146 = $145 << 16;
   $147 = $146 | $142;
   $148 = ((($139)) + 2|0);
   $149 = HEAP8[$148>>0]|0;
   $150 = $149&255;
   $151 = $150 << 8;
   $152 = $147 | $151;
   $153 = ((($139)) + 3|0);
   $154 = HEAP8[$153>>0]|0;
   $155 = $154&255;
   $156 = $152 | $155;
   $157 = ((($decoder)) + 12|0);
   $158 = HEAP32[$157>>2]|0;
   $159 = ((($138)) + 5|0);
   $160 = ($159>>>0)>($6>>>0);
   if ($160) {
    $$0 = 6;
    STACKTOP = sp;return ($$0|0);
   }
   $161 = HEAP8[$138>>0]|0;
   $162 = $161&255;
   HEAP16[$158>>1] = $162;
   $163 = ((($138)) + 1|0);
   $164 = HEAP8[$163>>0]|0;
   $165 = $164&255;
   $166 = ((($158)) + 2|0);
   HEAP16[$166>>1] = $165;
   $167 = ((($138)) + 2|0);
   $168 = HEAP8[$167>>0]|0;
   $169 = $168 << 24 >> 24;
   $170 = ((($158)) + 4|0);
   HEAP16[$170>>1] = $169;
   $171 = ((($138)) + 3|0);
   $172 = HEAP8[$171>>0]|0;
   $173 = $172 << 24 >> 24;
   $174 = ((($158)) + 6|0);
   HEAP16[$174>>1] = $173;
   $175 = ((($138)) + 4|0);
   $176 = HEAP8[$175>>0]|0;
   $177 = $176&255;
   $178 = ((($158)) + 8|0);
   HEAP16[$178>>1] = $177;
   $179 = ((($138)) + 8|0);
   $180 = ($179>>>0)>($6>>>0);
   if ($180) {
    $$0 = 6;
    STACKTOP = sp;return ($$0|0);
   } else {
    $181 = HEAP8[$159>>0]|0;
    $182 = $181 << 24 >> 24;
    $183 = ((($158)) + 10|0);
    HEAP16[$183>>1] = $182;
    $184 = ((($138)) + 6|0);
    $185 = HEAP8[$184>>0]|0;
    $186 = $185 << 24 >> 24;
    $187 = ((($158)) + 12|0);
    HEAP16[$187>>1] = $186;
    $188 = ((($138)) + 7|0);
    $189 = HEAP8[$188>>0]|0;
    $190 = $189&255;
    $191 = ((($158)) + 14|0);
    HEAP16[$191>>1] = $190;
    $192 = ((($decoder)) + 16|0);
    HEAP8[$192>>0] = 1;
    $193 = (($glyph_index) - ($$lcssa))|0;
    $194 = Math_imul($156, $193)|0;
    $195 = (($194) + ($156))|0;
    $image_end$6 = $195;$image_start$6 = $194;
    break L17;
   }
   break;
  }
  case 3:  {
   $196 = (($glyph_index) - ($$lcssa))|0;
   $197 = $196 << 1;
   $198 = (($73) + ($197)|0);
   $199 = ((($198)) + 4|0);
   $200 = ($199>>>0)>($6>>>0);
   if ($200) {
    $$0 = 6;
    STACKTOP = sp;return ($$0|0);
   }
   $201 = ((($198)) + 2|0);
   $202 = ((($201)) + -2|0);
   $203 = HEAP8[$202>>0]|0;
   $204 = $203&255;
   $205 = $204 << 8;
   $206 = ((($202)) + 1|0);
   $207 = HEAP8[$206>>0]|0;
   $208 = $207&255;
   $209 = $205 | $208;
   $210 = ((($201)) + 2|0);
   $211 = ((($210)) + -2|0);
   $212 = HEAP8[$211>>0]|0;
   $213 = $212&255;
   $214 = $213 << 8;
   $215 = ((($211)) + 1|0);
   $216 = HEAP8[$215>>0]|0;
   $217 = $216&255;
   $218 = $214 | $217;
   $219 = ($209|0)==($218|0);
   if ($219) {
    $$0 = 6;
    STACKTOP = sp;return ($$0|0);
   } else {
    $image_end$6 = $218;$image_start$6 = $209;
   }
   break;
  }
  case 4:  {
   $220 = ((($73)) + 4|0);
   $221 = ($220>>>0)>($6>>>0);
   if ($221) {
    $$0 = 6;
    STACKTOP = sp;return ($$0|0);
   }
   $222 = ((($220)) + -4|0);
   $223 = HEAP8[$222>>0]|0;
   $224 = $223&255;
   $225 = $224 << 24;
   $226 = ((($222)) + 1|0);
   $227 = HEAP8[$226>>0]|0;
   $228 = $227&255;
   $229 = $228 << 16;
   $230 = $229 | $225;
   $231 = ((($222)) + 2|0);
   $232 = HEAP8[$231>>0]|0;
   $233 = $232&255;
   $234 = $233 << 8;
   $235 = $230 | $234;
   $236 = ((($222)) + 3|0);
   $237 = HEAP8[$236>>0]|0;
   $238 = $237&255;
   $239 = $235 | $238;
   $240 = ((($220)) + 4|0);
   $241 = ($240>>>0)>($6>>>0);
   if ($241) {
    $$0 = 6;
    STACKTOP = sp;return ($$0|0);
   }
   $242 = $220;
   $243 = (($48) - ($242))|0;
   $244 = $243 >> 2;
   $245 = (($244) + -1)|0;
   $246 = ($239>>>0)<=($245>>>0);
   $247 = ($239|0)!=(0);
   $or$cond = $246 & $247;
   if ($or$cond) {
    $mm$033 = 0;$p$132 = $220;
   } else {
    $$0 = 6;
    STACKTOP = sp;return ($$0|0);
   }
   while(1) {
    $248 = HEAP8[$p$132>>0]|0;
    $249 = $248&255;
    $250 = $249 << 8;
    $251 = ((($p$132)) + 1|0);
    $252 = HEAP8[$251>>0]|0;
    $253 = $252&255;
    $254 = $250 | $253;
    $255 = ($254|0)==($glyph_index|0);
    if ($255) {
     $p$132$lcssa = $p$132;
     label = 20;
     break;
    }
    $272 = ((($p$132)) + 4|0);
    $273 = (($mm$033) + 1)|0;
    $274 = ($273>>>0)<($239>>>0);
    if ($274) {
     $mm$033 = $273;$p$132 = $272;
    } else {
     $$0 = 6;
     label = 62;
     break;
    }
   }
   if ((label|0) == 20) {
    $256 = ((($p$132$lcssa)) + 2|0);
    $257 = HEAP8[$256>>0]|0;
    $258 = $257&255;
    $259 = $258 << 8;
    $260 = ((($p$132$lcssa)) + 3|0);
    $261 = HEAP8[$260>>0]|0;
    $262 = $261&255;
    $263 = $259 | $262;
    $264 = ((($p$132$lcssa)) + 6|0);
    $265 = HEAP8[$264>>0]|0;
    $266 = $265&255;
    $267 = $266 << 8;
    $268 = ((($p$132$lcssa)) + 7|0);
    $269 = HEAP8[$268>>0]|0;
    $270 = $269&255;
    $271 = $267 | $270;
    $image_end$6 = $271;$image_start$6 = $263;
    break L17;
   }
   else if ((label|0) == 62) {
    STACKTOP = sp;return ($$0|0);
   }
   break;
  }
  case 19: case 5:  {
   $275 = ((($73)) + 16|0);
   $276 = ($275>>>0)>($6>>>0);
   if ($276) {
    $$0 = 6;
    STACKTOP = sp;return ($$0|0);
   }
   $277 = ((($73)) + 4|0);
   $278 = HEAP8[$73>>0]|0;
   $279 = $278&255;
   $280 = $279 << 24;
   $281 = ((($73)) + 1|0);
   $282 = HEAP8[$281>>0]|0;
   $283 = $282&255;
   $284 = $283 << 16;
   $285 = $284 | $280;
   $286 = ((($73)) + 2|0);
   $287 = HEAP8[$286>>0]|0;
   $288 = $287&255;
   $289 = $288 << 8;
   $290 = $285 | $289;
   $291 = ((($73)) + 3|0);
   $292 = HEAP8[$291>>0]|0;
   $293 = $292&255;
   $294 = $290 | $293;
   $295 = ((($decoder)) + 12|0);
   $296 = HEAP32[$295>>2]|0;
   $297 = ((($277)) + 5|0);
   $298 = ($297>>>0)>($6>>>0);
   if ($298) {
    $$0 = 6;
    STACKTOP = sp;return ($$0|0);
   }
   $299 = HEAP8[$277>>0]|0;
   $300 = $299&255;
   HEAP16[$296>>1] = $300;
   $301 = ((($277)) + 1|0);
   $302 = HEAP8[$301>>0]|0;
   $303 = $302&255;
   $304 = ((($296)) + 2|0);
   HEAP16[$304>>1] = $303;
   $305 = ((($277)) + 2|0);
   $306 = HEAP8[$305>>0]|0;
   $307 = $306 << 24 >> 24;
   $308 = ((($296)) + 4|0);
   HEAP16[$308>>1] = $307;
   $309 = ((($277)) + 3|0);
   $310 = HEAP8[$309>>0]|0;
   $311 = $310 << 24 >> 24;
   $312 = ((($296)) + 6|0);
   HEAP16[$312>>1] = $311;
   $313 = ((($277)) + 4|0);
   $314 = HEAP8[$313>>0]|0;
   $315 = $314&255;
   $316 = ((($296)) + 8|0);
   HEAP16[$316>>1] = $315;
   $317 = ((($277)) + 8|0);
   $318 = ($317>>>0)>($6>>>0);
   if ($318) {
    $$0 = 6;
    STACKTOP = sp;return ($$0|0);
   }
   $319 = HEAP8[$297>>0]|0;
   $320 = $319 << 24 >> 24;
   $321 = ((($296)) + 10|0);
   HEAP16[$321>>1] = $320;
   $322 = ((($277)) + 6|0);
   $323 = HEAP8[$322>>0]|0;
   $324 = $323 << 24 >> 24;
   $325 = ((($296)) + 12|0);
   HEAP16[$325>>1] = $324;
   $326 = ((($277)) + 7|0);
   $327 = HEAP8[$326>>0]|0;
   $328 = $327&255;
   $329 = ((($296)) + 14|0);
   HEAP16[$329>>1] = $328;
   $330 = ((($decoder)) + 16|0);
   HEAP8[$330>>0] = 1;
   $331 = ((($317)) + 4|0);
   $332 = HEAP8[$317>>0]|0;
   $333 = $332&255;
   $334 = $333 << 24;
   $335 = ((($317)) + 1|0);
   $336 = HEAP8[$335>>0]|0;
   $337 = $336&255;
   $338 = $337 << 16;
   $339 = $338 | $334;
   $340 = ((($317)) + 2|0);
   $341 = HEAP8[$340>>0]|0;
   $342 = $341&255;
   $343 = $342 << 8;
   $344 = $339 | $343;
   $345 = ((($317)) + 3|0);
   $346 = HEAP8[$345>>0]|0;
   $347 = $346&255;
   $348 = $344 | $347;
   $349 = $331;
   $350 = (($48) - ($349))|0;
   $351 = $350 >> 1;
   $352 = ($348>>>0)<=($351>>>0);
   $353 = ($348|0)!=(0);
   $or$cond40 = $352 & $353;
   if ($or$cond40) {
    $mm4$035 = 0;$p$434 = $331;
   } else {
    $$0 = 6;
    STACKTOP = sp;return ($$0|0);
   }
   while(1) {
    $354 = HEAP8[$p$434>>0]|0;
    $355 = $354&255;
    $356 = $355 << 8;
    $357 = ((($p$434)) + 1|0);
    $358 = HEAP8[$357>>0]|0;
    $359 = $358&255;
    $360 = $356 | $359;
    $not$ = ($360|0)==($glyph_index|0);
    if ($not$) {
     $mm4$035$lcssa = $mm4$035;
     label = 28;
     break;
    }
    $361 = ((($p$434)) + 2|0);
    $362 = (($mm4$035) + 1)|0;
    $363 = ($362>>>0)<($348>>>0);
    if ($363) {
     $mm4$035 = $362;$p$434 = $361;
    } else {
     $$0 = 6;
     label = 62;
     break;
    }
   }
   if ((label|0) == 28) {
    $364 = Math_imul($mm4$035$lcssa, $294)|0;
    $365 = (($364) + ($294))|0;
    $image_end$6 = $365;$image_start$6 = $364;
    break L17;
   }
   else if ((label|0) == 62) {
    STACKTOP = sp;return ($$0|0);
   }
   break;
  }
  default: {
   $$0 = 6;
   STACKTOP = sp;return ($$0|0);
  }
  }
 } while(0);
 $366 = ($image_end$6>>>0)<($image_start$6>>>0);
 if ($366) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $367 = (($image_end$6) - ($image_start$6))|0;
 $368 = ((($decoder)) + 4|0);
 $369 = HEAP32[$368>>2]|0;
 $370 = (($image_end$6) + ($91))|0;
 $371 = ((($decoder)) + 24|0);
 $372 = HEAP32[$371>>2]|0;
 $373 = ($370>>>0)>($372>>>0);
 if ($373) {
  $$01$i = 6;
 } else {
  $374 = (($image_start$6) + ($91))|0;
  $375 = ((($decoder)) + 20|0);
  $376 = HEAP32[$375>>2]|0;
  $377 = (($374) + ($376))|0;
  $378 = (_FT_Stream_Seek($369,$377)|0);
  $379 = ($378|0)==(0);
  if ($379) {
   $380 = (_FT_Stream_ExtractFrame($369,$367,$data$i)|0);
   $381 = ($380|0)==(0);
   if ($381) {
    $382 = HEAP32[$data$i>>2]|0;
    $383 = (($382) + ($367)|0);
    switch ($72|0) {
    case 17: case 8: case 2: case 1:  {
     $384 = ((($decoder)) + 12|0);
     $385 = HEAP32[$384>>2]|0;
     $386 = ($367|0)<(5);
     if ($386) {
      $error$2$i = 6;
     } else {
      $387 = ((($382)) + 5|0);
      $388 = HEAP8[$382>>0]|0;
      $389 = $388&255;
      HEAP16[$385>>1] = $389;
      $390 = ((($382)) + 1|0);
      $391 = HEAP8[$390>>0]|0;
      $392 = $391&255;
      $393 = ((($385)) + 2|0);
      HEAP16[$393>>1] = $392;
      $394 = ((($382)) + 2|0);
      $395 = HEAP8[$394>>0]|0;
      $396 = $395 << 24 >> 24;
      $397 = ((($385)) + 4|0);
      HEAP16[$397>>1] = $396;
      $398 = ((($382)) + 3|0);
      $399 = HEAP8[$398>>0]|0;
      $400 = $399 << 24 >> 24;
      $401 = ((($385)) + 6|0);
      HEAP16[$401>>1] = $400;
      $402 = ((($382)) + 4|0);
      $403 = HEAP8[$402>>0]|0;
      $404 = $403&255;
      $405 = ((($385)) + 8|0);
      HEAP16[$405>>1] = $404;
      $406 = ((($385)) + 10|0);
      HEAP16[$406>>1] = 0;
      $407 = ((($385)) + 12|0);
      HEAP16[$407>>1] = 0;
      $408 = ((($385)) + 14|0);
      HEAP16[$408>>1] = 0;
      $409 = ((($decoder)) + 16|0);
      HEAP8[$409>>0] = 1;
      $p$sroa$0$2$ph$in$i = $387;
      label = 39;
     }
     break;
    }
    case 18: case 9: case 7: case 6:  {
     $410 = ((($decoder)) + 12|0);
     $411 = HEAP32[$410>>2]|0;
     $412 = ((($382)) + 5|0);
     $413 = ($367|0)<(5);
     if ($413) {
      $error$2$i = 6;
     } else {
      $414 = HEAP8[$382>>0]|0;
      $415 = $414&255;
      HEAP16[$411>>1] = $415;
      $416 = ((($382)) + 1|0);
      $417 = HEAP8[$416>>0]|0;
      $418 = $417&255;
      $419 = ((($411)) + 2|0);
      HEAP16[$419>>1] = $418;
      $420 = ((($382)) + 2|0);
      $421 = HEAP8[$420>>0]|0;
      $422 = $421 << 24 >> 24;
      $423 = ((($411)) + 4|0);
      HEAP16[$423>>1] = $422;
      $424 = ((($382)) + 3|0);
      $425 = HEAP8[$424>>0]|0;
      $426 = $425 << 24 >> 24;
      $427 = ((($411)) + 6|0);
      HEAP16[$427>>1] = $426;
      $428 = ((($382)) + 4|0);
      $429 = HEAP8[$428>>0]|0;
      $430 = $429&255;
      $431 = ((($411)) + 8|0);
      HEAP16[$431>>1] = $430;
      $432 = ($367|0)<(8);
      if ($432) {
       $error$2$i = 6;
      } else {
       $433 = ((($382)) + 8|0);
       $434 = HEAP8[$412>>0]|0;
       $435 = $434 << 24 >> 24;
       $436 = ((($411)) + 10|0);
       HEAP16[$436>>1] = $435;
       $437 = ((($382)) + 6|0);
       $438 = HEAP8[$437>>0]|0;
       $439 = $438 << 24 >> 24;
       $440 = ((($411)) + 12|0);
       HEAP16[$440>>1] = $439;
       $441 = ((($382)) + 7|0);
       $442 = HEAP8[$441>>0]|0;
       $443 = $442&255;
       $444 = ((($411)) + 14|0);
       HEAP16[$444>>1] = $443;
       $445 = ((($decoder)) + 16|0);
       HEAP8[$445>>0] = 1;
       $p$sroa$0$2$ph$in$i = $433;
       label = 39;
      }
     }
     break;
    }
    default: {
     $p$sroa$0$2$ph$in$i = $382;
     label = 39;
    }
    }
    L86: do {
     if ((label|0) == 39) {
      $p$sroa$0$2$ph$i = $p$sroa$0$2$ph$in$i;
      L88: do {
       switch ($72|0) {
       case 19: case 18: case 17:  {
        $error$2$i = 7;
        break L86;
        break;
       }
       case 6: case 1:  {
        $loader$1$i = 4398;$p$sroa$0$4$i = $p$sroa$0$2$ph$i;
        break;
       }
       case 7: case 2:  {
        $446 = ((($decoder)) + 12|0);
        $447 = HEAP32[$446>>2]|0;
        $448 = ((($447)) + 2|0);
        $449 = HEAP16[$448>>1]|0;
        $450 = $449&65535;
        $451 = HEAP16[$447>>1]|0;
        $452 = $451&65535;
        $453 = Math_imul($452, $450)|0;
        $454 = (($453) + 7)|0;
        $455 = $454 >>> 3;
        $456 = (($450) + 7)|0;
        $457 = $456 >>> 3;
        $458 = Math_imul($457, $452)|0;
        $459 = ($455>>>0)<($458>>>0);
        if ($459) {
         $460 = $383;
         $461 = (($460) - ($p$sroa$0$2$ph$i))|0;
         $462 = ($458|0)==($461|0);
         if ($462) {
          $loader$1$i = 4398;$p$sroa$0$4$i = $p$sroa$0$2$ph$i;
          break L88;
         }
        }
        $loader$1$i = 4399;$p$sroa$0$4$i = $p$sroa$0$2$ph$i;
        break;
       }
       case 5:  {
        $loader$1$i = 4399;$p$sroa$0$4$i = $p$sroa$0$2$ph$i;
        break;
       }
       case 8:  {
        $463 = ((($p$sroa$0$2$ph$in$i)) + 1|0);
        $464 = ($463>>>0)>($383>>>0);
        if ($464) {
         $error$2$i = 0;
         break L86;
        }
        $465 = $463;
        $p$sroa$0$3$i = $465;
        label = 46;
        break;
       }
       case 9:  {
        $p$sroa$0$3$i = $p$sroa$0$2$ph$i;
        label = 46;
        break;
       }
       default: {
        $error$2$i = 8;
        break L86;
       }
       }
      } while(0);
      if ((label|0) == 46) {
       $loader$1$i = 4400;$p$sroa$0$4$i = $p$sroa$0$3$i;
      }
      $466 = ((($decoder)) + 17|0);
      $467 = HEAP8[$466>>0]|0;
      $468 = ($467<<24>>24)==(0);
      if ($468) {
       $469 = ((($decoder)) + 8|0);
       $470 = HEAP32[$469>>2]|0;
       $471 = ((($decoder)) + 16|0);
       $472 = HEAP32[$471>>2]|0;
       $473 = $472&255;
       $474 = ($473<<24>>24)==(0);
       if ($474) {
        $error$2$i = 6;
        break;
       }
       $475 = $472 >>> 16;
       $476 = ((($decoder)) + 12|0);
       $477 = HEAP32[$476>>2]|0;
       $478 = ((($477)) + 2|0);
       $479 = HEAP16[$478>>1]|0;
       $480 = $479&65535;
       $481 = HEAP16[$477>>1]|0;
       $482 = $481&65535;
       $483 = ((($470)) + 4|0);
       HEAP32[$483>>2] = $480;
       HEAP32[$470>>2] = $482;
       $484 = $475 & 255;
       switch ($484|0) {
       case 1:  {
        $485 = ((($470)) + 18|0);
        HEAP8[$485>>0] = 1;
        $486 = (($480) + 7)|0;
        $487 = $486 >>> 3;
        $488 = ((($470)) + 8|0);
        HEAP32[$488>>2] = $487;
        $489 = ((($470)) + 16|0);
        HEAP16[$489>>1] = 2;
        $508 = $487;
        break;
       }
       case 2:  {
        $490 = ((($470)) + 18|0);
        HEAP8[$490>>0] = 3;
        $491 = (($480) + 3)|0;
        $492 = $491 >>> 2;
        $493 = ((($470)) + 8|0);
        HEAP32[$493>>2] = $492;
        $494 = ((($470)) + 16|0);
        HEAP16[$494>>1] = 4;
        $508 = $492;
        break;
       }
       case 4:  {
        $495 = ((($470)) + 18|0);
        HEAP8[$495>>0] = 4;
        $496 = (($480) + 1)|0;
        $497 = $496 >>> 1;
        $498 = ((($470)) + 8|0);
        HEAP32[$498>>2] = $497;
        $499 = ((($470)) + 16|0);
        HEAP16[$499>>1] = 16;
        $508 = $497;
        break;
       }
       case 8:  {
        $500 = ((($470)) + 18|0);
        HEAP8[$500>>0] = 2;
        $501 = ((($470)) + 8|0);
        HEAP32[$501>>2] = $480;
        $502 = ((($470)) + 16|0);
        HEAP16[$502>>1] = 256;
        $508 = $480;
        break;
       }
       case 32:  {
        $503 = ((($470)) + 18|0);
        HEAP8[$503>>0] = 7;
        $504 = $480 << 2;
        $505 = ((($470)) + 8|0);
        HEAP32[$505>>2] = $504;
        $506 = ((($470)) + 16|0);
        HEAP16[$506>>1] = 256;
        $508 = $504;
        break;
       }
       default: {
        $error$2$i = 3;
        break L86;
       }
       }
       $507 = Math_imul($508, $482)|0;
       $509 = ($507|0)==(0);
       if (!($509)) {
        $510 = HEAP32[$decoder>>2]|0;
        $511 = ((($510)) + 84|0);
        $512 = HEAP32[$511>>2]|0;
        $513 = (_ft_glyphslot_alloc_bitmap($512,$507)|0);
        $514 = ($513|0)==(0);
        if (!($514)) {
         $error$2$i = $513;
         break;
        }
        HEAP8[$466>>0] = 1;
       }
      }
      $515 = $p$sroa$0$4$i;
      $516 = (FUNCTION_TABLE_iiiiii[$loader$1$i & 8191]($decoder,$515,$383,$x_pos,$y_pos)|0);
      $error$2$i = $516;
     }
    } while(0);
    _FT_Stream_ReleaseFrame($369,$data$i);
    $$01$i = $error$2$i;
   } else {
    $$01$i = $380;
   }
  } else {
   $$01$i = $378;
  }
 }
 $$0 = $$01$i;
 STACKTOP = sp;return ($$0|0);
}
function _tt_sbit_decoder_load_compound($decoder,$p,$limit,$x_pos,$y_pos) {
 $decoder = $decoder|0;
 $p = $p|0;
 $limit = $limit|0;
 $x_pos = $x_pos|0;
 $y_pos = $y_pos|0;
 var $$016 = 0, $$lcssa = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $8 = 0, $9 = 0, $error$1 = 0, $error$2 = 0, $nn$05 = 0, $sext = 0, $sext2 = 0, $sext3 = 0, $sext4 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($decoder)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 4|0);
 $3 = HEAP16[$2>>1]|0;
 $4 = ((($1)) + 6|0);
 $5 = HEAP16[$4>>1]|0;
 $6 = ((($1)) + 8|0);
 $7 = HEAP16[$6>>1]|0;
 $8 = ((($1)) + 10|0);
 $9 = HEAP16[$8>>1]|0;
 $10 = ((($1)) + 12|0);
 $11 = HEAP16[$10>>1]|0;
 $12 = ((($1)) + 14|0);
 $13 = HEAP16[$12>>1]|0;
 $14 = ((($p)) + 2|0);
 $15 = ($14>>>0)>($limit>>>0);
 if ($15) {
  $error$2 = 3;
  return ($error$2|0);
 }
 $16 = HEAP8[$p>>0]|0;
 $17 = $16&255;
 $18 = $17 << 8;
 $19 = ((($p)) + 1|0);
 $20 = HEAP8[$19>>0]|0;
 $21 = $20&255;
 $22 = $18 | $21;
 $23 = $22 << 2;
 $24 = (($14) + ($23)|0);
 $25 = ($24>>>0)>($limit>>>0);
 if ($25) {
  $error$2 = 3;
  return ($error$2|0);
 }
 $26 = ($22|0)==(0);
 if ($26) {
  $49 = $1;$error$1 = 0;
 } else {
  $$016 = $14;$nn$05 = 0;
  while(1) {
   $27 = ((($$016)) + 2|0);
   $28 = HEAP8[$$016>>0]|0;
   $29 = $28&255;
   $30 = $29 << 8;
   $31 = ((($$016)) + 1|0);
   $32 = HEAP8[$31>>0]|0;
   $33 = $32&255;
   $34 = $30 | $33;
   $35 = ((($$016)) + 3|0);
   $36 = HEAP8[$27>>0]|0;
   $37 = HEAP8[$35>>0]|0;
   $38 = $36&255;
   $39 = (($38) + ($x_pos))|0;
   $40 = $37&255;
   $41 = (($40) + ($y_pos))|0;
   $42 = (_tt_sbit_decoder_load_image($decoder,$34,$39,$41)|0);
   $43 = ($42|0)==(0);
   if (!($43)) {
    $$lcssa = $42;
    break;
   }
   $44 = ((($$016)) + 4|0);
   $45 = (($nn$05) + 1)|0;
   $46 = ($45>>>0)<($22>>>0);
   if ($46) {
    $$016 = $44;$nn$05 = $45;
   } else {
    $$lcssa = 0;
    break;
   }
  }
  $$pre = HEAP32[$0>>2]|0;
  $49 = $$pre;$error$1 = $$lcssa;
 }
 $sext = ($3 << 8)&65535;
 $47 = ($sext<<16>>16) >> 8;
 $48 = ((($49)) + 4|0);
 HEAP16[$48>>1] = $47;
 $sext2 = ($5 << 8)&65535;
 $50 = ($sext2<<16>>16) >> 8;
 $51 = ((($49)) + 6|0);
 HEAP16[$51>>1] = $50;
 $52 = $7 & 255;
 $53 = ((($49)) + 8|0);
 HEAP16[$53>>1] = $52;
 $sext3 = ($9 << 8)&65535;
 $54 = ($sext3<<16>>16) >> 8;
 $55 = ((($49)) + 10|0);
 HEAP16[$55>>1] = $54;
 $sext4 = ($11 << 8)&65535;
 $56 = ($sext4<<16>>16) >> 8;
 $57 = ((($49)) + 12|0);
 HEAP16[$57>>1] = $56;
 $58 = $13 & 255;
 $59 = ((($49)) + 14|0);
 HEAP16[$59>>1] = $58;
 $60 = ((($decoder)) + 8|0);
 $61 = HEAP32[$60>>2]|0;
 $62 = ((($61)) + 4|0);
 $63 = HEAP32[$62>>2]|0;
 $64 = $63&65535;
 $65 = $64 & 255;
 $66 = ((($49)) + 2|0);
 HEAP16[$66>>1] = $65;
 $67 = HEAP32[$61>>2]|0;
 $68 = $67&65535;
 $69 = $68 & 255;
 HEAP16[$49>>1] = $69;
 $error$2 = $error$1;
 return ($error$2|0);
}
function _tt_sbit_decoder_load_bit_aligned($decoder,$p,$limit,$x_pos,$y_pos) {
 $decoder = $decoder|0;
 $p = $p|0;
 $limit = $limit|0;
 $x_pos = $x_pos|0;
 $y_pos = $y_pos|0;
 var $$015 = 0, $$015$us = 0, $$1 = 0, $$2 = 0, $$3$lcssa = 0, $$3$lcssa$us = 0, $$35 = 0, $$35$us = 0, $$4 = 0, $$4$us = 0, $$5 = 0, $$5$us = 0, $$lcssa = 0, $$lcssa46 = 0, $$lcssa47 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $error$0 = 0, $h$013 = 0, $h$013$us = 0, $indvars$iv = 0, $indvars$iv31 = 0, $line$012 = 0, $line$012$us = 0, $nbits$014 = 0, $nbits$014$us = 0, $nbits$1 = 0, $nbits$3 = 0, $nbits$3$us = 0, $or$cond = 0, $or$cond3 = 0, $pwrite$0$lcssa = 0, $pwrite$0$lcssa$us = 0, $pwrite$07 = 0, $pwrite$07$us = 0;
 var $rval$016 = 0, $rval$016$us = 0, $rval$1 = 0, $rval$2 = 0, $rval$3$lcssa = 0, $rval$3$lcssa$in = 0, $rval$3$lcssa$us = 0, $rval$36 = 0, $rval$36$us = 0, $rval$4 = 0, $rval$4$us = 0, $rval$5 = 0, $rval$5$us = 0, $scevgep = 0, $scevgep2836 = 0, $scevgep29 = 0, $scevgep30 = 0, $scevgep3335 = 0, $scevgep34 = 0, $w$0$lcssa = 0;
 var $w$0$lcssa$us = 0, $w$08 = 0, $w$08$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($decoder)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($decoder)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($7)) + 2|0);
 $9 = HEAP16[$8>>1]|0;
 $10 = $9&65535;
 $11 = HEAP16[$7>>1]|0;
 $12 = $11&65535;
 $13 = ((($decoder)) + 18|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = $14&255;
 $16 = Math_imul($15, $10)|0;
 $17 = ($x_pos|0)<(0);
 if ($17) {
  $error$0 = 3;
  return ($error$0|0);
 }
 $18 = HEAP32[$1>>2]|0;
 $19 = ((($1)) + 4|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($10) + ($x_pos))|0;
 $22 = ($21>>>0)>($20>>>0);
 $23 = ($y_pos|0)<(0);
 $or$cond = $23 | $22;
 $24 = (($12) + ($y_pos))|0;
 $25 = ($24>>>0)>($18>>>0);
 $or$cond3 = $25 | $or$cond;
 if ($or$cond3) {
  $error$0 = 3;
  return ($error$0|0);
 }
 $26 = Math_imul($16, $12)|0;
 $27 = (($26) + 7)|0;
 $28 = $27 >>> 3;
 $29 = (($p) + ($28)|0);
 $30 = ($29>>>0)>($limit>>>0);
 if ($30) {
  $error$0 = 3;
  return ($error$0|0);
 }
 $31 = Math_imul($3, $y_pos)|0;
 $32 = $x_pos >> 3;
 $33 = (($31) + ($32))|0;
 $34 = (($5) + ($33)|0);
 $35 = $x_pos & 7;
 $36 = ($11<<16>>16)==(0);
 if ($36) {
  $error$0 = 0;
  return ($error$0|0);
 }
 $37 = ($35|0)==(0);
 $38 = (8 - ($35))|0;
 $39 = ($16>>>0)<($38>>>0);
 $40 = $39 ? $16 : $38;
 $41 = 255 << $40;
 $42 = $41 ^ -1;
 $43 = (($38) - ($40))|0;
 $44 = $42 << $43;
 $45 = (($16) - ($40))|0;
 $46 = (8 - ($40))|0;
 if ($37) {
  $60 = (($16) + -8)|0;
  $61 = $60 >>> 3;
  $62 = $61 << 3;
  $63 = (($60) - ($62))|0;
  $64 = (($32) + ($61))|0;
  $65 = (($64) + ($31))|0;
  $66 = (($65) + 1)|0;
  $scevgep = (($5) + ($66)|0);
  $67 = (($61) + 1)|0;
  $68 = ($16>>>0)>(7);
  $$015$us = $p;$h$013$us = $12;$indvars$iv = $scevgep;$line$012$us = $34;$nbits$014$us = 0;$rval$016$us = 0;
  while(1) {
   if ($68) {
    $$35$us = $$015$us;$pwrite$07$us = $line$012$us;$rval$36$us = $rval$016$us;$w$08$us = $16;
    while(1) {
     $102 = ((($$35$us)) + 1|0);
     $103 = HEAP8[$$35$us>>0]|0;
     $104 = $103&255;
     $105 = $rval$36$us&65535;
     $106 = $104 | $105;
     $107 = $106 >>> $nbits$014$us;
     $108 = ((($pwrite$07$us)) + 1|0);
     $109 = HEAP8[$pwrite$07$us>>0]|0;
     $110 = $109&255;
     $111 = $107 | $110;
     $112 = $111&255;
     HEAP8[$pwrite$07$us>>0] = $112;
     $113 = $106 << 8;
     $114 = $113&65535;
     $115 = (($w$08$us) + -8)|0;
     $116 = ($115|0)>(7);
     if ($116) {
      $$35$us = $102;$pwrite$07$us = $108;$rval$36$us = $114;$w$08$us = $115;
     } else {
      $$lcssa = $114;
      break;
     }
    }
    $scevgep29 = (($$015$us) + ($67)|0);
    $$3$lcssa$us = $scevgep29;$pwrite$0$lcssa$us = $indvars$iv;$rval$3$lcssa$us = $$lcssa;$w$0$lcssa$us = $63;
   } else {
    $$3$lcssa$us = $$015$us;$pwrite$0$lcssa$us = $line$012$us;$rval$3$lcssa$us = $rval$016$us;$w$0$lcssa$us = $16;
   }
   $69 = ($w$0$lcssa$us|0)>(0);
   do {
    if ($69) {
     $70 = ($nbits$014$us|0)<($w$0$lcssa$us|0);
     if (!($70)) {
      $71 = $rval$3$lcssa$us&65535;
      $72 = $71 >>> $nbits$014$us;
      $73 = 65280 >>> $w$0$lcssa$us;
      $74 = $73 & $72;
      $75 = HEAP8[$pwrite$0$lcssa$us>>0]|0;
      $76 = $75&255;
      $77 = $76 | $74;
      $78 = $77&255;
      HEAP8[$pwrite$0$lcssa$us>>0] = $78;
      $79 = (($nbits$014$us) - ($w$0$lcssa$us))|0;
      $$5$us = $$3$lcssa$us;$nbits$3$us = $79;$rval$5$us = $rval$3$lcssa$us;
      break;
     }
     $80 = ($$3$lcssa$us>>>0)<($limit>>>0);
     if ($80) {
      $81 = ((($$3$lcssa$us)) + 1|0);
      $82 = HEAP8[$$3$lcssa$us>>0]|0;
      $83 = $82&255;
      $84 = $rval$3$lcssa$us&65535;
      $85 = $83 | $84;
      $86 = $85&65535;
      $$4$us = $81;$rval$4$us = $86;
     } else {
      $$4$us = $$3$lcssa$us;$rval$4$us = $rval$3$lcssa$us;
     }
     $87 = $rval$4$us&65535;
     $88 = $87 >>> $nbits$014$us;
     $89 = 65280 >>> $w$0$lcssa$us;
     $90 = $88 & $89;
     $91 = HEAP8[$pwrite$0$lcssa$us>>0]|0;
     $92 = $91&255;
     $93 = $90 | $92;
     $94 = $93&255;
     HEAP8[$pwrite$0$lcssa$us>>0] = $94;
     $95 = (($nbits$014$us) + 8)|0;
     $96 = (($95) - ($w$0$lcssa$us))|0;
     $97 = $87 << 8;
     $98 = $97&65535;
     $$5$us = $$4$us;$nbits$3$us = $96;$rval$5$us = $98;
    } else {
     $$5$us = $$3$lcssa$us;$nbits$3$us = $nbits$014$us;$rval$5$us = $rval$3$lcssa$us;
    }
   } while(0);
   $99 = (($h$013$us) + -1)|0;
   $100 = (($line$012$us) + ($3)|0);
   $101 = ($h$013$us|0)>(1);
   $scevgep2836 = (($indvars$iv) + ($3)|0);
   if ($101) {
    $$015$us = $$5$us;$h$013$us = $99;$indvars$iv = $scevgep2836;$line$012$us = $100;$nbits$014$us = $nbits$3$us;$rval$016$us = $rval$5$us;
   } else {
    $error$0 = 0;
    break;
   }
  }
  return ($error$0|0);
 }
 $47 = $x_pos & 7;
 $48 = (8 - ($47))|0;
 $49 = ($16>>>0)>($48>>>0);
 $50 = $49 ? $48 : $16;
 $51 = $50 ^ -1;
 $52 = (($16) + ($51))|0;
 $53 = (($52) + -7)|0;
 $54 = $53 >>> 3;
 $55 = (($32) + ($54))|0;
 $56 = (($55) + ($31))|0;
 $57 = (($56) + 2)|0;
 $scevgep30 = (($5) + ($57)|0);
 $58 = (($54) + 1)|0;
 $59 = ($45|0)>(7);
 $$015 = $p;$h$013 = $12;$indvars$iv31 = $scevgep30;$line$012 = $34;$nbits$014 = 0;$rval$016 = 0;
 while(1) {
  $117 = ($h$013|0)==($12|0);
  do {
   if ($117) {
    $118 = ((($$015)) + 1|0);
    $119 = HEAP8[$$015>>0]|0;
    $120 = $119&255;
    $$2 = $118;$nbits$1 = $35;$rval$2 = $120;
   } else {
    $121 = ($nbits$014|0)<($40|0);
    if (!($121)) {
     $130 = ($rval$016&65535) >>> 8;
     $131 = (($nbits$014) - ($40))|0;
     $$2 = $$015;$nbits$1 = $131;$rval$2 = $130;
     break;
    }
    $122 = ($$015>>>0)<($limit>>>0);
    if ($122) {
     $123 = ((($$015)) + 1|0);
     $124 = HEAP8[$$015>>0]|0;
     $125 = $124&255;
     $126 = $rval$016&65535;
     $127 = $125 | $126;
     $128 = $127&65535;
     $$1 = $123;$rval$1 = $128;
    } else {
     $$1 = $$015;$rval$1 = $rval$016;
    }
    $129 = (($46) + ($nbits$014))|0;
    $$2 = $$1;$nbits$1 = $129;$rval$2 = $rval$1;
   }
  } while(0);
  $132 = $rval$2&65535;
  $133 = $132 >>> $nbits$1;
  $134 = $133 & $44;
  $135 = ((($line$012)) + 1|0);
  $136 = HEAP8[$line$012>>0]|0;
  $137 = $136&255;
  $138 = $134 | $137;
  $139 = $138&255;
  HEAP8[$line$012>>0] = $139;
  $140 = $132 << 8;
  if ($59) {
   $$35 = $$2;$pwrite$07 = $135;$rval$36 = $140;$w$08 = $45;
   while(1) {
    $141 = ((($$35)) + 1|0);
    $142 = HEAP8[$$35>>0]|0;
    $143 = $142&255;
    $144 = $rval$36 & 65280;
    $145 = $143 | $144;
    $146 = $145 >>> $nbits$1;
    $147 = ((($pwrite$07)) + 1|0);
    $148 = HEAP8[$pwrite$07>>0]|0;
    $149 = $148&255;
    $150 = $146 | $149;
    $151 = $150&255;
    HEAP8[$pwrite$07>>0] = $151;
    $152 = $145 << 8;
    $153 = (($w$08) + -8)|0;
    $154 = ($153|0)>(7);
    if ($154) {
     $$35 = $141;$pwrite$07 = $147;$rval$36 = $152;$w$08 = $153;
    } else {
     $$lcssa46 = $152;$$lcssa47 = $153;
     break;
    }
   }
   $scevgep34 = (($$2) + ($58)|0);
   $$3$lcssa = $scevgep34;$pwrite$0$lcssa = $indvars$iv31;$rval$3$lcssa$in = $$lcssa46;$w$0$lcssa = $$lcssa47;
  } else {
   $$3$lcssa = $$2;$pwrite$0$lcssa = $135;$rval$3$lcssa$in = $140;$w$0$lcssa = $45;
  }
  $rval$3$lcssa = $rval$3$lcssa$in&65535;
  $155 = ($w$0$lcssa|0)>(0);
  do {
   if ($155) {
    $156 = ($nbits$1|0)<($w$0$lcssa|0);
    if (!($156)) {
     $175 = $rval$3$lcssa$in & 65280;
     $176 = $175 >>> $nbits$1;
     $177 = 65280 >>> $w$0$lcssa;
     $178 = $177 & $176;
     $179 = HEAP8[$pwrite$0$lcssa>>0]|0;
     $180 = $179&255;
     $181 = $180 | $178;
     $182 = $181&255;
     HEAP8[$pwrite$0$lcssa>>0] = $182;
     $183 = (($nbits$1) - ($w$0$lcssa))|0;
     $$5 = $$3$lcssa;$nbits$3 = $183;$rval$5 = $rval$3$lcssa;
     break;
    }
    $157 = ($$3$lcssa>>>0)<($limit>>>0);
    if ($157) {
     $158 = ((($$3$lcssa)) + 1|0);
     $159 = HEAP8[$$3$lcssa>>0]|0;
     $160 = $159&255;
     $161 = $rval$3$lcssa$in & 65280;
     $162 = $160 | $161;
     $$4 = $158;$rval$4 = $162;
    } else {
     $$4 = $$3$lcssa;$rval$4 = $rval$3$lcssa$in;
    }
    $163 = $rval$4 & 65535;
    $164 = $163 >>> $nbits$1;
    $165 = 65280 >>> $w$0$lcssa;
    $166 = $164 & $165;
    $167 = HEAP8[$pwrite$0$lcssa>>0]|0;
    $168 = $167&255;
    $169 = $166 | $168;
    $170 = $169&255;
    HEAP8[$pwrite$0$lcssa>>0] = $170;
    $171 = (($nbits$1) + 8)|0;
    $172 = (($171) - ($w$0$lcssa))|0;
    $173 = $163 << 8;
    $174 = $173&65535;
    $$5 = $$4;$nbits$3 = $172;$rval$5 = $174;
   } else {
    $$5 = $$3$lcssa;$nbits$3 = $nbits$1;$rval$5 = $rval$3$lcssa;
   }
  } while(0);
  $184 = (($h$013) + -1)|0;
  $185 = (($line$012) + ($3)|0);
  $186 = ($h$013|0)>(1);
  $scevgep3335 = (($indvars$iv31) + ($3)|0);
  if ($186) {
   $$015 = $$5;$h$013 = $184;$indvars$iv31 = $scevgep3335;$line$012 = $185;$nbits$014 = $nbits$3;$rval$016 = $rval$5;
  } else {
   $error$0 = 0;
   break;
  }
 }
 return ($error$0|0);
}
function _tt_sbit_decoder_load_byte_aligned($decoder,$p,$limit,$x_pos,$y_pos) {
 $decoder = $decoder|0;
 $p = $p|0;
 $limit = $limit|0;
 $x_pos = $x_pos|0;
 $y_pos = $y_pos|0;
 var $$015 = 0, $$015$us = 0, $$18$us = 0, $$2 = 0, $$2$us = 0, $$328 = 0, $$328$us = 0, $$416$us = 0, $$5 = 0, $$5$us = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $error$0 = 0, $h$014 = 0, $h$014$us = 0, $h$127 = 0, $h$127$us = 0, $indvars$iv = 0, $indvars$iv43 = 0, $indvars$iv49 = 0, $line$013 = 0, $line$013$us = 0, $line$126 = 0, $line$126$us = 0, $or$cond = 0, $or$cond3 = 0, $pwrite$09$us = 0, $pwrite1$017$us = 0, $scevgep = 0, $scevgep3954 = 0;
 var $scevgep40 = 0, $scevgep41 = 0, $scevgep42 = 0, $scevgep4552 = 0, $scevgep46 = 0, $scevgep47 = 0, $scevgep48 = 0, $scevgep5153 = 0, $w$010$us = 0, $w2$018$us = 0, $wval$019$us = 0, $wval$1 = 0, $wval$1$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($decoder)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($decoder)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($7)) + 2|0);
 $9 = HEAP16[$8>>1]|0;
 $10 = $9&65535;
 $11 = HEAP16[$7>>1]|0;
 $12 = $11&65535;
 $13 = ((($decoder)) + 18|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = $14&255;
 $16 = Math_imul($15, $10)|0;
 $17 = ($x_pos|0)<(0);
 if ($17) {
  $error$0 = 3;
  return ($error$0|0);
 }
 $18 = HEAP32[$1>>2]|0;
 $19 = ((($1)) + 4|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($10) + ($x_pos))|0;
 $22 = ($21>>>0)>($20>>>0);
 $23 = ($y_pos|0)<(0);
 $or$cond = $23 | $22;
 $24 = (($12) + ($y_pos))|0;
 $25 = ($24>>>0)>($18>>>0);
 $or$cond3 = $25 | $or$cond;
 if ($or$cond3) {
  $error$0 = 3;
  return ($error$0|0);
 }
 $26 = (($16) + 7)|0;
 $27 = $26 >>> 3;
 $28 = Math_imul($27, $12)|0;
 $29 = (($p) + ($28)|0);
 $30 = ($29>>>0)>($limit>>>0);
 if ($30) {
  $error$0 = 3;
  return ($error$0|0);
 }
 $31 = Math_imul($3, $y_pos)|0;
 $32 = $x_pos >> 3;
 $33 = (($31) + ($32))|0;
 $34 = (($5) + ($33)|0);
 $35 = $x_pos & 7;
 $36 = ($35|0)==(0);
 $37 = ($11<<16>>16)!=(0);
 if ($36) {
  if (!($37)) {
   $error$0 = 0;
   return ($error$0|0);
  }
  $89 = ($16>>>0)>(7);
  if (!($89)) {
   $90 = ($16|0)==(0);
   $91 = 65280 >>> $16;
   $$015 = $p;$h$014 = $12;$line$013 = $34;
   while(1) {
    if ($90) {
     $$2 = $$015;
    } else {
     $121 = HEAP8[$line$013>>0]|0;
     $122 = $121&255;
     $123 = ((($$015)) + 1|0);
     $124 = HEAP8[$$015>>0]|0;
     $125 = $124&255;
     $126 = $125 & $91;
     $127 = $126 | $122;
     $128 = $127&255;
     HEAP8[$line$013>>0] = $128;
     $$2 = $123;
    }
    $129 = (($h$014) + -1)|0;
    $130 = (($line$013) + ($3)|0);
    $131 = ($h$014|0)>(1);
    if ($131) {
     $$015 = $$2;$h$014 = $129;$line$013 = $130;
    } else {
     $error$0 = 0;
     break;
    }
   }
   return ($error$0|0);
  }
  $92 = (($16) + -8)|0;
  $93 = $92 >>> 3;
  $94 = $93 << 3;
  $95 = (($92) - ($94))|0;
  $96 = (($32) + ($93))|0;
  $97 = (($96) + ($31))|0;
  $98 = (($97) + 1)|0;
  $scevgep = (($5) + ($98)|0);
  $99 = (($93) + 1)|0;
  $100 = (($16) - ($94))|0;
  $101 = ($100|0)>(8);
  $102 = 65280 >>> $95;
  $$015$us = $p;$h$014$us = $12;$indvars$iv = $scevgep;$line$013$us = $34;
  while(1) {
   $scevgep40 = (($$015$us) + ($99)|0);
   $$18$us = $$015$us;$pwrite$09$us = $line$013$us;$w$010$us = $16;
   while(1) {
    $114 = HEAP8[$pwrite$09$us>>0]|0;
    $115 = ((($$18$us)) + 1|0);
    $116 = HEAP8[$$18$us>>0]|0;
    $117 = $116 | $114;
    HEAP8[$pwrite$09$us>>0] = $117;
    $118 = ((($pwrite$09$us)) + 1|0);
    $119 = (($w$010$us) + -8)|0;
    $120 = ($119|0)>(7);
    if ($120) {
     $$18$us = $115;$pwrite$09$us = $118;$w$010$us = $119;
    } else {
     break;
    }
   }
   if ($101) {
    $scevgep41 = (($$015$us) + ($93)|0);
    $103 = HEAP8[$indvars$iv>>0]|0;
    $104 = $103&255;
    $105 = ((($scevgep41)) + 2|0);
    $106 = HEAP8[$scevgep40>>0]|0;
    $107 = $106&255;
    $108 = $107 & $102;
    $109 = $108 | $104;
    $110 = $109&255;
    HEAP8[$indvars$iv>>0] = $110;
    $$2$us = $105;
   } else {
    $$2$us = $scevgep40;
   }
   $111 = (($h$014$us) + -1)|0;
   $112 = (($line$013$us) + ($3)|0);
   $113 = ($h$014$us|0)>(1);
   $scevgep3954 = (($indvars$iv) + ($3)|0);
   if ($113) {
    $$015$us = $$2$us;$h$014$us = $111;$indvars$iv = $scevgep3954;$line$013$us = $112;
   } else {
    $error$0 = 0;
    break;
   }
  }
  return ($error$0|0);
 }
 if (!($37)) {
  $error$0 = 0;
  return ($error$0|0);
 }
 $38 = ($16>>>0)>(7);
 if (!($38)) {
  $39 = ($16|0)==(0);
  $40 = (($16) + ($35))|0;
  $41 = ($40>>>0)>(8);
  $42 = 65280 >>> $16;
  $$328 = $p;$h$127 = $12;$line$126 = $34;
  while(1) {
   if ($39) {
    $$5 = $$328;$wval$1 = 0;
   } else {
    $132 = ((($$328)) + 1|0);
    $133 = HEAP8[$$328>>0]|0;
    $134 = $133&255;
    $135 = $134 & $42;
    $$5 = $132;$wval$1 = $135;
   }
   $136 = HEAP8[$line$126>>0]|0;
   $137 = $136&255;
   $138 = $wval$1 >>> $35;
   $139 = $137 | $138;
   $140 = $139&255;
   HEAP8[$line$126>>0] = $140;
   if ($41) {
    $141 = ((($line$126)) + 1|0);
    $142 = $wval$1 << 8;
    $143 = HEAP8[$141>>0]|0;
    $144 = $143&255;
    $145 = $142 >>> $35;
    $146 = $144 | $145;
    $147 = $146&255;
    HEAP8[$141>>0] = $147;
   }
   $148 = (($h$127) + -1)|0;
   $149 = (($line$126) + ($3)|0);
   $150 = ($h$127|0)>(1);
   if ($150) {
    $$328 = $$5;$h$127 = $148;$line$126 = $149;
   } else {
    $error$0 = 0;
    break;
   }
  }
  return ($error$0|0);
 }
 $43 = (($16) + -8)|0;
 $44 = $43 >>> 3;
 $45 = $44 << 3;
 $46 = (($43) - ($45))|0;
 $47 = (($32) + ($44))|0;
 $48 = (($47) + ($31))|0;
 $49 = (($48) + 1)|0;
 $scevgep42 = (($5) + ($49)|0);
 $50 = (($44) + 1)|0;
 $scevgep48 = (($5) + ($48)|0);
 $51 = (($16) - ($45))|0;
 $52 = ($51|0)>(8);
 $53 = 65280 >>> $46;
 $54 = (($46) + ($35))|0;
 $55 = ($54|0)>(8);
 $$328$us = $p;$h$127$us = $12;$indvars$iv43 = $scevgep42;$indvars$iv49 = $scevgep48;$line$126$us = $34;
 while(1) {
  $scevgep46 = (($$328$us) + ($50)|0);
  $$416$us = $$328$us;$pwrite1$017$us = $line$126$us;$w2$018$us = $16;$wval$019$us = 0;
  while(1) {
   $76 = ((($$416$us)) + 1|0);
   $77 = HEAP8[$$416$us>>0]|0;
   $78 = $77&255;
   $79 = $78 | $wval$019$us;
   $80 = HEAP8[$pwrite1$017$us>>0]|0;
   $81 = $80&255;
   $82 = $79 >>> $35;
   $83 = $82 | $81;
   $84 = $83&255;
   HEAP8[$pwrite1$017$us>>0] = $84;
   $85 = ((($pwrite1$017$us)) + 1|0);
   $86 = $79 << 8;
   $87 = (($w2$018$us) + -8)|0;
   $88 = ($87|0)>(7);
   if ($88) {
    $$416$us = $76;$pwrite1$017$us = $85;$w2$018$us = $87;$wval$019$us = $86;
   } else {
    $$lcssa = $86;
    break;
   }
  }
  if ($52) {
   $scevgep47 = (($$328$us) + ($44)|0);
   $56 = ((($scevgep47)) + 2|0);
   $57 = HEAP8[$scevgep46>>0]|0;
   $58 = $57&255;
   $59 = $58 & $53;
   $60 = $59 | $$lcssa;
   $$5$us = $56;$wval$1$us = $60;
  } else {
   $$5$us = $scevgep46;$wval$1$us = $$lcssa;
  }
  $61 = HEAP8[$indvars$iv43>>0]|0;
  $62 = $61&255;
  $63 = $wval$1$us >>> $35;
  $64 = $62 | $63;
  $65 = $64&255;
  HEAP8[$indvars$iv43>>0] = $65;
  if ($55) {
   $66 = ((($indvars$iv49)) + 2|0);
   $67 = $wval$1$us << 8;
   $68 = HEAP8[$66>>0]|0;
   $69 = $68&255;
   $70 = $67 >>> $35;
   $71 = $69 | $70;
   $72 = $71&255;
   HEAP8[$66>>0] = $72;
  }
  $73 = (($h$127$us) + -1)|0;
  $74 = (($line$126$us) + ($3)|0);
  $75 = ($h$127$us|0)>(1);
  $scevgep4552 = (($indvars$iv43) + ($3)|0);
  $scevgep5153 = (($indvars$iv49) + ($3)|0);
  if ($75) {
   $$328$us = $$5$us;$h$127$us = $73;$indvars$iv43 = $scevgep4552;$indvars$iv49 = $scevgep5153;$line$126$us = $74;
  } else {
   $error$0 = 0;
   break;
  }
 }
 return ($error$0|0);
}
function _tt_face_free_ps_names($face) {
 $face = $face|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $n$02 = 0, $switch$split2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 100|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 612|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = ($3<<24>>24)==(0);
 if ($4) {
  HEAP8[$2>>0] = 0;
  return;
 }
 $5 = ((($face)) + 468|0);
 $6 = HEAP32[$5>>2]|0;
 $switch$split2D = ($6|0)<(163840);
 if (!($switch$split2D)) {
  switch ($6|0) {
  case 163840:  {
   break;
  }
  default: {
   HEAP8[$2>>0] = 0;
   return;
  }
  }
  $25 = ((($face)) + 620|0);
  $26 = HEAP32[$25>>2]|0;
  _ft_mem_free($1,$26);
  HEAP32[$25>>2] = 0;
  $27 = ((($face)) + 616|0);
  HEAP16[$27>>1] = 0;
  HEAP8[$2>>0] = 0;
  return;
 }
 switch ($6|0) {
 case 131072:  {
  break;
 }
 default: {
  HEAP8[$2>>0] = 0;
  return;
 }
 }
 $7 = ((($face)) + 620|0);
 $8 = HEAP32[$7>>2]|0;
 _ft_mem_free($1,$8);
 HEAP32[$7>>2] = 0;
 $9 = ((($face)) + 616|0);
 HEAP16[$9>>1] = 0;
 $10 = ((($face)) + 618|0);
 $11 = HEAP16[$10>>1]|0;
 $12 = ($11<<16>>16)==(0);
 $13 = ((($face)) + 624|0);
 if (!($12)) {
  $16 = 0;$n$02 = 0;
  while(1) {
   $14 = HEAP32[$13>>2]|0;
   $15 = (($14) + ($16<<2)|0);
   $17 = HEAP32[$15>>2]|0;
   _ft_mem_free($1,$17);
   $18 = HEAP32[$13>>2]|0;
   $19 = (($18) + ($16<<2)|0);
   HEAP32[$19>>2] = 0;
   $20 = (($n$02) + 1)<<16>>16;
   $21 = $20&65535;
   $22 = HEAP16[$10>>1]|0;
   $23 = ($20&65535)<($22&65535);
   if ($23) {
    $16 = $21;$n$02 = $20;
   } else {
    break;
   }
  }
 }
 $24 = HEAP32[$13>>2]|0;
 _ft_mem_free($1,$24);
 HEAP32[$13>>2] = 0;
 HEAP16[$10>>1] = 0;
 HEAP8[$2>>0] = 0;
 return;
}
function _tt_face_get_kerning($face,$left_glyph,$right_glyph) {
 $face = $face|0;
 $left_glyph = $left_glyph|0;
 $right_glyph = $right_glyph|0;
 var $$ = 0, $$lcssa = 0, $$sink = 0, $$value$7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cond = 0, $count$031 = 0, $count2$026 = 0, $mask$030 = 0, $max$024 = 0, $max$1 = 0, $min$023 = 0, $min$1 = 0, $num_pairs$0 = 0, $or$cond = 0, $or$cond28 = 0, $or$cond8 = 0;
 var $p$029 = 0, $p$125 = 0, $p$125$lcssa = 0, $result$0$lcssa = 0, $result$032 = 0, $result$1 = 0, $sext5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 740|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 744|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + ($3)|0);
 $5 = ((($face)) + 748|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 $8 = ($3|0)<(10);
 $or$cond28 = $8 | $7;
 if ($or$cond28) {
  $result$0$lcssa = 0;
  return ($result$0$lcssa|0);
 }
 $9 = ((($1)) + 4|0);
 $10 = ((($1)) + 10|0);
 $11 = ((($face)) + 752|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = $left_glyph << 16;
 $14 = $13 | $right_glyph;
 $15 = ((($face)) + 756|0);
 $35 = $10;$count$031 = $6;$mask$030 = 1;$p$029 = $9;$result$032 = 0;
 while(1) {
  $16 = ((($p$029)) + 2|0);
  $17 = HEAP8[$16>>0]|0;
  $18 = $17&255;
  $19 = $18 << 8;
  $20 = ((($p$029)) + 3|0);
  $21 = HEAP8[$20>>0]|0;
  $22 = $21&255;
  $23 = $19 | $22;
  $24 = ((($p$029)) + 5|0);
  $25 = HEAP8[$24>>0]|0;
  $26 = (($p$029) + ($23)|0);
  $27 = ($26>>>0)>($4>>>0);
  $$ = $27 ? $4 : $26;
  $28 = $12 & $mask$030;
  $29 = ($28|0)==(0);
  $30 = ((($p$029)) + 14|0);
  $31 = ($30>>>0)>($$>>>0);
  $or$cond8 = $29 | $31;
  L6: do {
   if ($or$cond8) {
    $result$1 = $result$032;
   } else {
    $32 = ((($p$029)) + 4|0);
    $33 = HEAP8[$32>>0]|0;
    $34 = HEAP8[$35>>0]|0;
    $36 = $34&255;
    $37 = $36 << 8;
    $38 = ((($p$029)) + 7|0);
    $39 = HEAP8[$38>>0]|0;
    $40 = $39&255;
    $41 = $37 | $40;
    $42 = $$;
    $43 = $30;
    $44 = (($42) - ($43))|0;
    $45 = ($41*6)|0;
    $46 = ($44|0)<($45|0);
    $47 = (($44|0) / 6)&-1;
    $num_pairs$0 = $46 ? $47 : $41;
    $cond = ($33<<24>>24)==(0);
    if ($cond) {
     $48 = HEAP32[$15>>2]|0;
     $49 = $48 & $mask$030;
     $50 = ($49|0)==(0);
     $51 = ($num_pairs$0|0)==(0);
     if ($50) {
      if ($51) {
       $result$1 = $result$032;
       break;
      } else {
       $count2$026 = $num_pairs$0;$p$125 = $30;
      }
      while(1) {
       $86 = HEAP8[$p$125>>0]|0;
       $87 = $86&255;
       $88 = $87 << 24;
       $89 = ((($p$125)) + 1|0);
       $90 = HEAP8[$89>>0]|0;
       $91 = $90&255;
       $92 = $91 << 16;
       $93 = $92 | $88;
       $94 = ((($p$125)) + 2|0);
       $95 = HEAP8[$94>>0]|0;
       $96 = $95&255;
       $97 = $96 << 8;
       $98 = $93 | $97;
       $99 = ((($p$125)) + 3|0);
       $100 = HEAP8[$99>>0]|0;
       $101 = $100&255;
       $102 = $98 | $101;
       $103 = ($102|0)==($14|0);
       if ($103) {
        $p$125$lcssa = $p$125;
        break;
       }
       $104 = ((($p$125)) + 6|0);
       $105 = (($count2$026) + -1)|0;
       $106 = ($105|0)==(0);
       if ($106) {
        $result$1 = $result$032;
        break L6;
       } else {
        $count2$026 = $105;$p$125 = $104;
       }
      }
      $107 = ((($p$125$lcssa)) + 4|0);
      $108 = HEAP8[$107>>0]|0;
      $109 = $108&255;
      $110 = $109 << 8;
      $111 = ((($p$125$lcssa)) + 5|0);
      $112 = HEAP8[$111>>0]|0;
      $113 = $112&255;
      $114 = $110 | $113;
      $$sink = $114;
     } else {
      if ($51) {
       $result$1 = $result$032;
       break;
      } else {
       $max$024 = $num_pairs$0;$min$023 = 0;
      }
      while(1) {
       $52 = (($max$024) + ($min$023))|0;
       $53 = $52 >>> 1;
       $54 = ($53*6)|0;
       $55 = (($30) + ($54)|0);
       $56 = ((($55)) + 4|0);
       $57 = ((($56)) + -4|0);
       $58 = HEAP8[$57>>0]|0;
       $59 = $58&255;
       $60 = $59 << 24;
       $61 = ((($57)) + 1|0);
       $62 = HEAP8[$61>>0]|0;
       $63 = $62&255;
       $64 = $63 << 16;
       $65 = $64 | $60;
       $66 = ((($57)) + 2|0);
       $67 = HEAP8[$66>>0]|0;
       $68 = $67&255;
       $69 = $68 << 8;
       $70 = $65 | $69;
       $71 = ((($57)) + 3|0);
       $72 = HEAP8[$71>>0]|0;
       $73 = $72&255;
       $74 = $70 | $73;
       $75 = ($74|0)==($14|0);
       if ($75) {
        $$lcssa = $56;
        break;
       }
       $76 = ($74>>>0)<($14>>>0);
       $77 = (($53) + 1)|0;
       $min$1 = $76 ? $77 : $min$023;
       $max$1 = $76 ? $max$024 : $53;
       $78 = ($min$1>>>0)<($max$1>>>0);
       if ($78) {
        $max$024 = $max$1;$min$023 = $min$1;
       } else {
        $result$1 = $result$032;
        break L6;
       }
      }
      $79 = HEAP8[$$lcssa>>0]|0;
      $80 = $79&255;
      $81 = $80 << 8;
      $82 = ((($$lcssa)) + 1|0);
      $83 = HEAP8[$82>>0]|0;
      $84 = $83&255;
      $85 = $81 | $84;
      $$sink = $85;
     }
     $sext5 = $$sink << 16;
     $115 = $sext5 >> 16;
     $116 = $25 & 8;
     $117 = ($116<<24>>24)==(0);
     $118 = $117 ? $result$032 : 0;
     $$value$7 = (($115) + ($118))|0;
     $result$1 = $$value$7;
    } else {
     $result$1 = $result$032;
    }
   }
  } while(0);
  $119 = (($count$031) + -1)|0;
  $120 = $mask$030 << 1;
  $121 = ($119|0)==(0);
  $122 = ((($$)) + 6|0);
  $123 = ($122>>>0)>($4>>>0);
  $or$cond = $123 | $121;
  if ($or$cond) {
   $result$0$lcssa = $result$1;
   break;
  } else {
   $35 = $122;$count$031 = $119;$mask$030 = $120;$p$029 = $$;$result$032 = $result$1;
  }
 }
 return ($result$0$lcssa|0);
}
function _tt_face_load_font_dir($face,$stream) {
 $face = $face|0;
 $stream = $stream|0;
 var $$has_meta$0$i = 0, $$ph = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phiZ2D = 0, $$pre36 = 0, $$pre37 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $entry$017 = 0, $entry$1 = 0, $error = 0, $error$i = 0, $has_head$020$i = 0, $has_head$039$i = 0, $has_head$039$i$lcssa = 0, $has_head$2$ph$i = 0, $has_meta$028$i = 0, $has_meta$041$i = 0, $has_meta$041$i$lcssa = 0, $has_meta$2$ph$i = 0, $has_sing$024$i = 0, $has_sing$040$i = 0, $has_sing$040$i$lcssa = 0, $has_sing$2$ph$i = 0, $nn$018 = 0;
 var $nn$042$i = 0, $nn$042$i$lcssa = 0, $or$cond$i = 0, $sfnt = 0, $switch$split12D = 0, $switch$split2D = 0, $switch$split42D = 0, $switch$split72D = 0, $table$i = 0, $valid_entries$016$i = 0, $valid_entries$038$i = 0, $valid_entries$038$i$lcssa = 0, $valid_entries$1$i = 0, $valid_entries$2$ph$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp + 40|0;
 $table$i = sp + 24|0;
 $sfnt = sp + 8|0;
 $error = sp;
 $0 = ((($stream)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_FT_Stream_Pos($stream)|0);
 $3 = ((($sfnt)) + 12|0);
 HEAP32[$3>>2] = $2;
 $4 = (_FT_Stream_ReadULong($stream,$error)|0);
 HEAP32[$sfnt>>2] = $4;
 $5 = HEAP32[$error>>2]|0;
 $6 = ($5|0)==(0);
 if (!($6)) {
  $94 = $5;
  STACKTOP = sp;return ($94|0);
 }
 $7 = (_FT_Stream_ReadFields($stream,174758,$sfnt)|0);
 HEAP32[$error>>2] = $7;
 $8 = ($7|0)==(0);
 if (!($8)) {
  $94 = $7;
  STACKTOP = sp;return ($94|0);
 }
 $9 = HEAP32[$sfnt>>2]|0;
 $10 = ($9|0)==(1330926671);
 L7: do {
  if ($10) {
   $$phi$trans$insert = ((($sfnt)) + 4|0);
   $$pre = HEAP16[$$phi$trans$insert>>1]|0;
   $$pre$phiZ2D = $$phi$trans$insert;$53 = $$pre;$55 = 1330926671;
  } else {
   $11 = HEAP32[$3>>2]|0;
   $12 = (($11) + 12)|0;
   $13 = (_FT_Stream_Seek($stream,$12)|0);
   HEAP32[$error$i>>2] = $13;
   $14 = ($13|0)==(0);
   do {
    if ($14) {
     $15 = ((($sfnt)) + 4|0);
     $16 = HEAP16[$15>>1]|0;
     $17 = ($16<<16>>16)==(0);
     if ($17) {
      HEAP16[$15>>1] = 0;
     } else {
      $18 = ((($table$i)) + 8|0);
      $19 = ((($stream)) + 4|0);
      $20 = ((($table$i)) + 12|0);
      $21 = (($11) + 28)|0;
      $39 = 0;$has_head$039$i = 0;$has_meta$041$i = 0;$has_sing$040$i = 0;$nn$042$i = 0;$valid_entries$038$i = 0;
      L15: while(1) {
       $22 = (_FT_Stream_ReadFields($stream,174782,$table$i)|0);
       HEAP32[$error$i>>2] = $22;
       $23 = ($22|0)==(0);
       if (!($23)) {
        $has_head$039$i$lcssa = $has_head$039$i;$has_meta$041$i$lcssa = $has_meta$041$i;$has_sing$040$i$lcssa = $has_sing$040$i;$nn$042$i$lcssa = $nn$042$i;$valid_entries$038$i$lcssa = $valid_entries$038$i;
        label = 10;
        break;
       }
       $25 = HEAP32[$18>>2]|0;
       $26 = HEAP32[$19>>2]|0;
       $27 = ($26>>>0)<($25>>>0);
       L18: do {
        if ($27) {
         $has_head$2$ph$i = $has_head$039$i;$has_meta$2$ph$i = $has_meta$041$i;$has_sing$2$ph$i = $has_sing$040$i;$valid_entries$2$ph$i = $valid_entries$038$i;
        } else {
         $28 = HEAP32[$20>>2]|0;
         $29 = (($26) - ($25))|0;
         $30 = ($28>>>0)>($29>>>0);
         $31 = HEAP32[$table$i>>2]|0;
         L20: do {
          if ($30) {
           $switch$split2D = ($31|0)<(1986884728);
           if ($switch$split2D) {
            switch ($31|0) {
            case 1752003704:  {
             break L20;
             break;
            }
            default: {
             $has_head$2$ph$i = $has_head$039$i;$has_meta$2$ph$i = $has_meta$041$i;$has_sing$2$ph$i = $has_sing$040$i;$valid_entries$2$ph$i = $valid_entries$038$i;
             break L18;
            }
            }
           } else {
            switch ($31|0) {
            case 1986884728:  {
             break L20;
             break;
            }
            default: {
             $has_head$2$ph$i = $has_head$039$i;$has_meta$2$ph$i = $has_meta$041$i;$has_sing$2$ph$i = $has_sing$040$i;$valid_entries$2$ph$i = $valid_entries$038$i;
             break L18;
            }
            }
           }
          }
         } while(0);
         $valid_entries$1$i = (($valid_entries$038$i) + 1)<<16>>16;
         $switch$split12D = ($31|0)<(1651008868);
         L26: do {
          if ($switch$split12D) {
           switch ($31|0) {
           case 1397313095:  {
            $has_head$2$ph$i = $has_head$039$i;$has_meta$2$ph$i = $has_meta$041$i;$has_sing$2$ph$i = 1;$valid_entries$2$ph$i = $valid_entries$1$i;
            break L18;
            break;
           }
           default: {
           }
           }
          } else {
           $switch$split72D = ($31|0)<(1751474532);
           if ($switch$split72D) {
            switch ($31|0) {
            case 1651008868:  {
             break;
            }
            default: {
             break L26;
            }
            }
           } else {
            switch ($31|0) {
            case 1751474532:  {
             break;
            }
            default: {
             break L26;
            }
            }
           }
           $32 = ($28>>>0)<(54);
           if ($32) {
            label = 16;
            break L15;
           }
           $33 = (($25) + 12)|0;
           $34 = (_FT_Stream_Seek($stream,$33)|0);
           HEAP32[$error$i>>2] = $34;
           $35 = ($34|0)==(0);
           if (!($35)) {
            $95 = $34;
            label = 21;
            break L15;
           }
           (_FT_Stream_ReadULong($stream,$error$i)|0);
           $36 = HEAP32[$error$i>>2]|0;
           $37 = ($36|0)==(0);
           if (!($37)) {
            $95 = $36;
            label = 21;
            break L15;
           }
           $38 = $39 << 4;
           $40 = (($21) + ($38))|0;
           $41 = (_FT_Stream_Seek($stream,$40)|0);
           HEAP32[$error$i>>2] = $41;
           $42 = ($41|0)==(0);
           if ($42) {
            $has_head$2$ph$i = 1;$has_meta$2$ph$i = $has_meta$041$i;$has_sing$2$ph$i = $has_sing$040$i;$valid_entries$2$ph$i = $valid_entries$1$i;
            break L18;
           } else {
            $95 = $41;
            label = 21;
            break L15;
           }
          }
         } while(0);
         $43 = ($31|0)==(1296389185);
         $$has_meta$0$i = $43 ? 1 : $has_meta$041$i;
         $has_head$2$ph$i = $has_head$039$i;$has_meta$2$ph$i = $$has_meta$0$i;$has_sing$2$ph$i = $has_sing$040$i;$valid_entries$2$ph$i = $valid_entries$1$i;
        }
       } while(0);
       $44 = (($nn$042$i) + 1)<<16>>16;
       $45 = $44&65535;
       $46 = HEAP16[$15>>1]|0;
       $47 = ($44&65535)<($46&65535);
       if ($47) {
        $39 = $45;$has_head$039$i = $has_head$2$ph$i;$has_meta$041$i = $has_meta$2$ph$i;$has_sing$040$i = $has_sing$2$ph$i;$nn$042$i = $44;$valid_entries$038$i = $valid_entries$2$ph$i;
       } else {
        $has_head$020$i = $has_head$2$ph$i;$has_meta$028$i = $has_meta$2$ph$i;$has_sing$024$i = $has_sing$2$ph$i;$valid_entries$016$i = $valid_entries$2$ph$i;
        break;
       }
      }
      if ((label|0) == 10) {
       $24 = (($nn$042$i$lcssa) + -1)<<16>>16;
       HEAP16[$15>>1] = $24;
       $has_head$020$i = $has_head$039$i$lcssa;$has_meta$028$i = $has_meta$041$i$lcssa;$has_sing$024$i = $has_sing$040$i$lcssa;$valid_entries$016$i = $valid_entries$038$i$lcssa;
      }
      else if ((label|0) == 16) {
       HEAP32[$error$i>>2] = 142;
       $95 = 142;
       label = 21;
      }
      if ((label|0) == 21) {
       $$ph = $95;
       break;
      }
      HEAP16[$15>>1] = $valid_entries$016$i;
      $48 = ($valid_entries$016$i<<16>>16)==(0);
      if (!($48)) {
       $49 = ($has_head$020$i|0)==(0);
       if ($49) {
        $50 = ($has_sing$024$i|0)!=(0);
        $51 = ($has_meta$028$i|0)!=(0);
        $or$cond$i = $51 & $50;
        if (!($or$cond$i)) {
         HEAP32[$error$i>>2] = 142;
         $$ph = 142;
         break;
        }
       }
       HEAP32[$error>>2] = 0;
       $$pre36 = HEAP32[$sfnt>>2]|0;
       $$pre$phiZ2D = $15;$53 = $valid_entries$016$i;$55 = $$pre36;
       break L7;
      }
     }
     HEAP32[$error$i>>2] = 2;
     $$ph = 2;
    } else {
     $$ph = $13;
    }
   } while(0);
   HEAP32[$error>>2] = $$ph;
   $94 = $$ph;
   STACKTOP = sp;return ($94|0);
  }
 } while(0);
 $52 = ((($face)) + 152|0);
 HEAP16[$52>>1] = $53;
 $54 = ((($face)) + 148|0);
 HEAP32[$54>>2] = $55;
 $56 = $53&65535;
 $57 = (_ft_mem_realloc($1,16,0,$56,0,$error)|0);
 $58 = ((($face)) + 156|0);
 HEAP32[$58>>2] = $57;
 $59 = HEAP32[$error>>2]|0;
 $60 = ($59|0)==(0);
 if (!($60)) {
  $94 = $59;
  STACKTOP = sp;return ($94|0);
 }
 $61 = HEAP32[$3>>2]|0;
 $62 = (($61) + 12)|0;
 $63 = (_FT_Stream_Seek($stream,$62)|0);
 HEAP32[$error>>2] = $63;
 $64 = ($63|0)==(0);
 if (!($64)) {
  $94 = $63;
  STACKTOP = sp;return ($94|0);
 }
 $65 = HEAP16[$52>>1]|0;
 $66 = $65&65535;
 $67 = $66 << 4;
 $68 = (_FT_Stream_EnterFrame($stream,$67)|0);
 HEAP32[$error>>2] = $68;
 $69 = ($68|0)==(0);
 if (!($69)) {
  $94 = $68;
  STACKTOP = sp;return ($94|0);
 }
 $70 = HEAP16[$$pre$phiZ2D>>1]|0;
 $71 = ($70<<16>>16)==(0);
 if (!($71)) {
  $72 = HEAP32[$58>>2]|0;
  $73 = ((($stream)) + 4|0);
  $entry$017 = $72;$nn$018 = 0;
  while(1) {
   $74 = (_FT_Stream_GetULong($stream)|0);
   HEAP32[$entry$017>>2] = $74;
   $75 = (_FT_Stream_GetULong($stream)|0);
   $76 = ((($entry$017)) + 4|0);
   HEAP32[$76>>2] = $75;
   $77 = (_FT_Stream_GetULong($stream)|0);
   $78 = ((($entry$017)) + 8|0);
   HEAP32[$78>>2] = $77;
   $79 = (_FT_Stream_GetULong($stream)|0);
   $80 = ((($entry$017)) + 12|0);
   HEAP32[$80>>2] = $79;
   $81 = HEAP32[$78>>2]|0;
   $82 = HEAP32[$73>>2]|0;
   $83 = ($82>>>0)<($81>>>0);
   L66: do {
    if ($83) {
     $entry$1 = $entry$017;
    } else {
     $84 = (($82) - ($81))|0;
     $85 = ($79>>>0)>($84>>>0);
     if (!($85)) {
      $89 = ((($entry$017)) + 16|0);
      $entry$1 = $89;
      break;
     }
     $86 = HEAP32[$entry$017>>2]|0;
     $switch$split42D = ($86|0)<(1986884728);
     if ($switch$split42D) {
      switch ($86|0) {
      case 1752003704:  {
       break;
      }
      default: {
       $entry$1 = $entry$017;
       break L66;
      }
      }
     } else {
      switch ($86|0) {
      case 1986884728:  {
       break;
      }
      default: {
       $entry$1 = $entry$017;
       break L66;
      }
      }
     }
     $87 = $84 & -4;
     HEAP32[$80>>2] = $87;
     $88 = ((($entry$017)) + 16|0);
     $entry$1 = $88;
    }
   } while(0);
   $90 = (($nn$018) + 1)|0;
   $91 = HEAP16[$$pre$phiZ2D>>1]|0;
   $92 = $91&65535;
   $93 = ($90|0)<($92|0);
   if ($93) {
    $entry$017 = $entry$1;$nn$018 = $90;
   } else {
    break;
   }
  }
 }
 _FT_Stream_ExitFrame($stream);
 $$pre37 = HEAP32[$error>>2]|0;
 $94 = $$pre37;
 STACKTOP = sp;return ($94|0);
}
function _tt_face_load_hmtx($face,$stream,$vertical) {
 $face = $face|0;
 $stream = $stream|0;
 $vertical = $vertical|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ptable_offset$0 = 0, $ptable_size$0 = 0, $table_size = 0, $tag$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $table_size = sp;
 $0 = ($vertical<<24>>24)==(0);
 $tag$0 = $0 ? 1752003704 : 1986884728;
 $1 = ((($face)) + 508|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (FUNCTION_TABLE_iiiii[$2 & 8191]($face,$tag$0,$stream,$table_size)|0);
 $4 = ($3|0)==(0);
 if (!($4)) {
  STACKTOP = sp;return ($3|0);
 }
 $5 = ((($face)) + 684|0);
 $6 = ((($face)) + 692|0);
 $ptable_size$0 = $0 ? $5 : $6;
 $7 = ((($face)) + 784|0);
 $8 = ((($face)) + 788|0);
 $ptable_offset$0 = $0 ? $7 : $8;
 $9 = HEAP32[$table_size>>2]|0;
 HEAP32[$ptable_size$0>>2] = $9;
 $10 = (_FT_Stream_Pos($stream)|0);
 HEAP32[$ptable_offset$0>>2] = $10;
 STACKTOP = sp;return ($3|0);
}
function _tt_face_load_sbit($face,$stream) {
 $face = $face|0;
 $stream = $stream|0;
 var $$0 = 0, $$old1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $9 = 0, $count$0 = 0, $count3$0 = 0, $error$6 = 0, $or$cond = 0, $table_size = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $table_size = sp;
 $0 = ((($face)) + 724|0);
 $1 = ((($face)) + 728|0);
 $2 = ((($face)) + 732|0);
 $3 = ((($face)) + 736|0);
 $4 = ((($face)) + 508|0);
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;HEAP32[$0+12>>2]=0|0;
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiiii[$5 & 8191]($face,1128418371,$stream,$table_size)|0);
 $7 = ($6|0)==(0);
 do {
  if ($7) {
   HEAP32[$2>>2] = 2;
   $19 = 2;
   label = 8;
  } else {
   $8 = HEAP32[$4>>2]|0;
   $9 = (FUNCTION_TABLE_iiiii[$8 & 8191]($face,1161972803,$stream,$table_size)|0);
   $10 = ($9|0)==(0);
   if (!($10)) {
    $11 = HEAP32[$4>>2]|0;
    $12 = (FUNCTION_TABLE_iiiii[$11 & 8191]($face,1651273571,$stream,$table_size)|0);
    $13 = ($12|0)==(0);
    if (!($13)) {
     $14 = HEAP32[$4>>2]|0;
     $15 = (FUNCTION_TABLE_iiiii[$14 & 8191]($face,1935829368,$stream,$table_size)|0);
     $16 = ($15|0)==(0);
     if (!($16)) {
      $error$6 = $15;
      break;
     }
     HEAP32[$2>>2] = 3;
     $19 = 3;
     label = 8;
     break;
    }
   }
   HEAP32[$2>>2] = 1;
   $19 = 1;
   label = 8;
  }
 } while(0);
 L9: do {
  if ((label|0) == 8) {
   $17 = HEAP32[$table_size>>2]|0;
   $18 = ($17>>>0)<(8);
   if ($18) {
    $error$6 = 3;
   } else {
    switch ($19|0) {
    case 2: case 1:  {
     $20 = (_FT_Stream_ExtractFrame($stream,$17,$0)|0);
     $21 = ($20|0)==(0);
     if (!($21)) {
      $error$6 = $20;
      break L9;
     }
     $22 = HEAP32[$table_size>>2]|0;
     HEAP32[$1>>2] = $22;
     $23 = HEAP32[$0>>2]|0;
     $24 = HEAP8[$23>>0]|0;
     $25 = $24&255;
     $26 = $25 << 24;
     $27 = ((($23)) + 1|0);
     $28 = HEAP8[$27>>0]|0;
     $29 = $28&255;
     $30 = $29 << 16;
     $31 = $30 | $26;
     $32 = ((($23)) + 4|0);
     $33 = HEAP8[$32>>0]|0;
     $34 = $33&255;
     $35 = $34 << 24;
     $36 = ((($23)) + 5|0);
     $37 = HEAP8[$36>>0]|0;
     $38 = $37&255;
     $39 = $38 << 16;
     $40 = $39 | $35;
     $41 = ((($23)) + 6|0);
     $42 = HEAP8[$41>>0]|0;
     $43 = $42&255;
     $44 = $43 << 8;
     $45 = $40 | $44;
     $46 = ((($23)) + 7|0);
     $47 = HEAP8[$46>>0]|0;
     $48 = $47&255;
     $49 = $45 | $48;
     $50 = ($31|0)==(131072);
     if (!($50)) {
      $error$6 = 2;
      break L9;
     }
     $51 = ($49>>>0)>(65535);
     if ($51) {
      $error$6 = 3;
      break L9;
     }
     $52 = ($49*48)|0;
     $53 = $52 | 8;
     $54 = ($53>>>0)>($22>>>0);
     if ($54) {
      $55 = (($22) + -8)|0;
      $56 = (($55>>>0) / 48)&-1;
      $count$0 = $56;
     } else {
      $count$0 = $49;
     }
     HEAP32[$3>>2] = $count$0;
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
     break;
    }
    case 3:  {
     $57 = (_FT_Stream_EnterFrame($stream,8)|0);
     $58 = ($57|0)==(0);
     if (!($58)) {
      $error$6 = $57;
      break L9;
     }
     $59 = (_FT_Stream_GetUShort($stream)|0);
     $60 = (_FT_Stream_GetUShort($stream)|0);
     $61 = (_FT_Stream_GetULong($stream)|0);
     _FT_Stream_ExitFrame($stream);
     $62 = ($59<<16>>16)==(0);
     if ($62) {
      $error$6 = 2;
      break L9;
     }
     $63 = ($60<<16>>16)==(1);
     if ($63) {
      $$old1 = ($61>>>0)>(65535);
      if ($$old1) {
       $error$6 = 3;
       break L9;
      }
     } else {
      $64 = ($60<<16>>16)!=(3);
      $65 = ($61>>>0)>(65535);
      $or$cond = $64 | $65;
      if ($or$cond) {
       $error$6 = 3;
       break L9;
      }
     }
     $66 = $61 << 2;
     $67 = (($66) + 8)|0;
     $68 = HEAP32[$table_size>>2]|0;
     $69 = ($67>>>0)>($68>>>0);
     $70 = (($68) + -8)|0;
     $71 = $70 >>> 2;
     $count3$0 = $69 ? $71 : $61;
     $72 = (_FT_Stream_Pos($stream)|0);
     $73 = (($72) + -8)|0;
     $74 = (_FT_Stream_Seek($stream,$73)|0);
     $75 = ($74|0)==(0);
     if (!($75)) {
      $error$6 = $74;
      break L9;
     }
     $76 = $count3$0 << 2;
     $77 = (($76) + 8)|0;
     HEAP32[$1>>2] = $77;
     $78 = (_FT_Stream_ExtractFrame($stream,$77,$0)|0);
     $79 = ($78|0)==(0);
     if (!($79)) {
      $error$6 = $78;
      break L9;
     }
     HEAP32[$3>>2] = $count3$0;
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
     break;
    }
    default: {
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
    }
   }
  }
 } while(0);
 $80 = HEAP32[$0>>2]|0;
 $81 = ($80|0)==(0|0);
 if (!($81)) {
  _FT_Stream_ReleaseFrame($stream,$0);
 }
 HEAP32[$1>>2] = 0;
 HEAP32[$2>>2] = 0;
 $$0 = $error$6;
 STACKTOP = sp;return ($$0|0);
}
function _tt_face_free_sbit($face) {
 $face = $face|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 104|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 724|0);
 _FT_Stream_ReleaseFrame($1,$2);
 $3 = ((($face)) + 728|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($face)) + 732|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($face)) + 736|0);
 HEAP32[$5>>2] = 0;
 return;
}
function _tt_face_set_sbit_strike($face,$req,$astrike_index) {
 $face = $face|0;
 $req = $req|0;
 $astrike_index = $astrike_index|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_FT_Match_Size($face,$req,0,$astrike_index)|0);
 return ($0|0);
}
function _tt_face_load_strike_metrics($face,$strike_index,$metrics) {
 $face = $face|0;
 $strike_index = $strike_index|0;
 $metrics = $metrics|0;
 var $$0 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $sext = 0, $sext2 = 0, $table_size = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $table_size = sp;
 $0 = ((($face)) + 736|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1>>>0)>($strike_index>>>0);
 if (!($2)) {
  $$1 = 6;
  STACKTOP = sp;return ($$1|0);
 }
 $3 = ((($face)) + 732|0);
 $4 = HEAP32[$3>>2]|0;
 switch ($4|0) {
 case 2: case 1:  {
  $5 = ((($face)) + 724|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ((($6)) + 8|0);
  $8 = ($strike_index*48)|0;
  $9 = (($7) + ($8)|0);
  $10 = ((($9)) + 44|0);
  $11 = HEAP8[$10>>0]|0;
  $12 = $11&255;
  HEAP16[$metrics>>1] = $12;
  $13 = ((($9)) + 45|0);
  $14 = HEAP8[$13>>0]|0;
  $15 = $14&255;
  $16 = ((($metrics)) + 2|0);
  HEAP16[$16>>1] = $15;
  $17 = ((($9)) + 16|0);
  $18 = HEAP8[$17>>0]|0;
  $19 = $18 << 24 >> 24;
  $20 = $19 << 6;
  $21 = ((($metrics)) + 12|0);
  HEAP32[$21>>2] = $20;
  $22 = ((($9)) + 17|0);
  $23 = HEAP8[$22>>0]|0;
  $24 = $23 << 24 >> 24;
  $25 = $24 << 6;
  $26 = ((($metrics)) + 16|0);
  HEAP32[$26>>2] = $25;
  $27 = (($20) - ($25))|0;
  $28 = ((($metrics)) + 20|0);
  HEAP32[$28>>2] = $27;
  $29 = ((($9)) + 22|0);
  $30 = HEAP8[$29>>0]|0;
  $31 = $30 << 24 >> 24;
  $32 = ((($9)) + 18|0);
  $33 = HEAP8[$32>>0]|0;
  $34 = $33&255;
  $35 = (($34) + ($31))|0;
  $36 = ((($9)) + 23|0);
  $37 = HEAP8[$36>>0]|0;
  $38 = $37 << 24 >> 24;
  $39 = (($35) + ($38))|0;
  $40 = $39 << 6;
  $41 = ((($metrics)) + 24|0);
  HEAP32[$41>>2] = $40;
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
  break;
 }
 case 3:  {
  $42 = ((($face)) + 104|0);
  $43 = HEAP32[$42>>2]|0;
  $44 = ((($face)) + 724|0);
  $45 = HEAP32[$44>>2]|0;
  $46 = ((($45)) + 8|0);
  $47 = $strike_index << 2;
  $48 = (($46) + ($47)|0);
  $49 = ((($48)) + 4|0);
  $50 = ((($49)) + -4|0);
  $51 = HEAP8[$50>>0]|0;
  $52 = $51&255;
  $53 = $52 << 24;
  $54 = ((($50)) + 1|0);
  $55 = HEAP8[$54>>0]|0;
  $56 = $55&255;
  $57 = $56 << 16;
  $58 = $57 | $53;
  $59 = ((($50)) + 2|0);
  $60 = HEAP8[$59>>0]|0;
  $61 = $60&255;
  $62 = $61 << 8;
  $63 = $58 | $62;
  $64 = ((($50)) + 3|0);
  $65 = HEAP8[$64>>0]|0;
  $66 = $65&255;
  $67 = $63 | $66;
  $68 = ((($face)) + 508|0);
  $69 = HEAP32[$68>>2]|0;
  $70 = (FUNCTION_TABLE_iiiii[$69 & 8191]($face,1935829368,$43,$table_size)|0);
  $71 = ($70|0)==(0);
  if ($71) {
   $72 = (($67) + 4)|0;
   $73 = HEAP32[$table_size>>2]|0;
   $74 = ($72>>>0)>($73>>>0);
   if ($74) {
    $$0 = 3;
   } else {
    $75 = (_FT_Stream_Pos($43)|0);
    $76 = (($75) + ($67))|0;
    $77 = (_FT_Stream_Seek($43,$76)|0);
    $78 = ($77|0)==(0);
    if ($78) {
     $79 = (_FT_Stream_EnterFrame($43,4)|0);
     $80 = ($79|0)==(0);
     if ($80) {
      $81 = (_FT_Stream_GetUShort($43)|0);
      (_FT_Stream_GetUShort($43)|0);
      _FT_Stream_ExitFrame($43);
      $82 = ((($face)) + 178|0);
      $83 = HEAP16[$82>>1]|0;
      $84 = $83&65535;
      HEAP16[$metrics>>1] = $81;
      $85 = ((($metrics)) + 2|0);
      HEAP16[$85>>1] = $81;
      $86 = $81&65535;
      $87 = ((($face)) + 220|0);
      $88 = HEAP32[$87>>2]|0;
      $sext = $88 << 16;
      $89 = $sext >> 16;
      $90 = $86 << 6;
      $91 = Math_imul($90, $89)|0;
      $92 = (($91|0) / ($84|0))&-1;
      $93 = ((($metrics)) + 12|0);
      HEAP32[$93>>2] = $92;
      $94 = $88 >> 16;
      $95 = Math_imul($90, $94)|0;
      $96 = (($95|0) / ($84|0))&-1;
      $97 = ((($metrics)) + 16|0);
      HEAP32[$97>>2] = $96;
      $98 = (($89) - ($94))|0;
      $99 = ((($face)) + 224|0);
      $100 = HEAP32[$99>>2]|0;
      $sext2 = $100 << 16;
      $101 = $sext2 >> 16;
      $102 = (($98) + ($101))|0;
      $103 = Math_imul($90, $102)|0;
      $104 = (($103|0) / ($84|0))&-1;
      $105 = ((($metrics)) + 20|0);
      HEAP32[$105>>2] = $104;
      $106 = $100 >>> 16;
      $107 = Math_imul($90, $106)|0;
      $108 = (($107|0) / ($84|0))&-1;
      $109 = ((($metrics)) + 24|0);
      HEAP32[$109>>2] = $108;
      $$0 = 0;
     } else {
      $$0 = $79;
     }
    } else {
     $$0 = $77;
    }
   }
  } else {
   $$0 = $70;
  }
  $$1 = $$0;
  STACKTOP = sp;return ($$1|0);
  break;
 }
 default: {
  $$1 = 2;
  STACKTOP = sp;return ($$1|0);
 }
 }
 return (0)|0;
}
function _tt_face_get_metrics($face,$vertical,$gindex,$abearing,$aadvance) {
 $face = $face|0;
 $vertical = $vertical|0;
 $gindex = $gindex|0;
 $abearing = $abearing|0;
 $aadvance = $aadvance|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $header$0 = 0, $table_pos$0 = 0, $table_pos$0$in = 0, $table_size$0 = 0, $table_size$0$in = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($face)) + 104|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($vertical<<24>>24)==(0);
 $3 = ((($face)) + 296|0);
 $4 = ((($face)) + 788|0);
 $5 = ((($face)) + 692|0);
 $6 = ((($face)) + 216|0);
 $7 = ((($face)) + 784|0);
 $8 = ((($face)) + 684|0);
 $table_size$0$in = $2 ? $8 : $5;
 $table_pos$0$in = $2 ? $7 : $4;
 $header$0 = $2 ? $6 : $3;
 $table_pos$0 = HEAP32[$table_pos$0$in>>2]|0;
 $table_size$0 = HEAP32[$table_size$0$in>>2]|0;
 $9 = (($table_size$0) + ($table_pos$0))|0;
 $10 = ((($header$0)) + 34|0);
 $11 = HEAP16[$10>>1]|0;
 $12 = $11&65535;
 $13 = ($11<<16>>16)==(0);
 do {
  if (!($13)) {
   $14 = ($12>>>0)>($gindex>>>0);
   if ($14) {
    $15 = $gindex << 2;
    $16 = (($table_pos$0) + ($15))|0;
    $17 = (($16) + 4)|0;
    $18 = ($17>>>0)>($9>>>0);
    if ($18) {
     break;
    }
    $19 = (_FT_Stream_Seek($1,$16)|0);
    HEAP32[$error>>2] = $19;
    $20 = ($19|0)==(0);
    if (!($20)) {
     break;
    }
    $21 = (_FT_Stream_ReadUShort($1,$error)|0);
    HEAP16[$aadvance>>1] = $21;
    $22 = HEAP32[$error>>2]|0;
    $23 = ($22|0)==(0);
    if (!($23)) {
     break;
    }
    $24 = (_FT_Stream_ReadUShort($1,$error)|0);
    HEAP16[$abearing>>1] = $24;
    $25 = HEAP32[$error>>2]|0;
    $26 = ($25|0)==(0);
    if (!($26)) {
     break;
    }
    STACKTOP = sp;return;
   }
   $27 = $12 << 2;
   $28 = (($table_pos$0) + -4)|0;
   $29 = (($28) + ($27))|0;
   $30 = (($27) + ($table_pos$0))|0;
   $31 = ($30>>>0)>($9>>>0);
   if (!($31)) {
    $32 = (_FT_Stream_Seek($1,$29)|0);
    HEAP32[$error>>2] = $32;
    $33 = ($32|0)==(0);
    if ($33) {
     $34 = (_FT_Stream_ReadUShort($1,$error)|0);
     HEAP16[$aadvance>>1] = $34;
     $35 = HEAP32[$error>>2]|0;
     $36 = ($35|0)==(0);
     if ($36) {
      $37 = (($gindex) - ($12))|0;
      $38 = $37 << 1;
      $39 = (($38) + 4)|0;
      $40 = (($39) + ($29))|0;
      $41 = (($40) + 2)|0;
      $42 = ($41>>>0)>($9>>>0);
      if ($42) {
       HEAP16[$abearing>>1] = 0;
       STACKTOP = sp;return;
      }
      $43 = (_FT_Stream_Seek($1,$40)|0);
      HEAP32[$error>>2] = $43;
      $44 = ($43|0)==(0);
      if (!($44)) {
       STACKTOP = sp;return;
      }
      $45 = (_FT_Stream_ReadUShort($1,$error)|0);
      HEAP16[$abearing>>1] = $45;
      STACKTOP = sp;return;
     }
    }
   }
  }
 } while(0);
 HEAP16[$abearing>>1] = 0;
 HEAP16[$aadvance>>1] = 0;
 STACKTOP = sp;return;
}
function _gray_raster_new($memory,$araster) {
 $memory = $memory|0;
 $araster = $araster|0;
 var $0 = 0, $1 = 0, $2 = 0, $error = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 HEAP32[$araster>>2] = 0;
 $0 = (_ft_mem_alloc($memory,4,$error)|0);
 $1 = HEAP32[$error>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  STACKTOP = sp;return ($1|0);
 }
 HEAP32[$0>>2] = $memory;
 HEAP32[$araster>>2] = $0;
 STACKTOP = sp;return ($1|0);
}
function _gray_raster_reset($raster,$pool_base,$pool_size) {
 $raster = $raster|0;
 $pool_base = $pool_base|0;
 $pool_size = $pool_size|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _gray_raster_set_mode($raster,$mode,$args) {
 $raster = $raster|0;
 $mode = $mode|0;
 $args = $args|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _gray_raster_render($raster,$params) {
 $raster = $raster|0;
 $params = $params|0;
 var $$0 = 0, $$3$i = 0, $$lcssa = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$pn2$i$i = 0, $$pn2$i$i$phi = 0, $$pre$i = 0, $$pre$phi$iZ2D = 0, $$pre29$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0;
 var $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0;
 var $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0;
 var $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0;
 var $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0;
 var $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0;
 var $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $band$i = 0, $bands$i = 0, $buffer = 0, $cell$0$i$i = 0, $cell$01$i$i = 0, $cell$04$i$i = 0, $cell_start$0$i = 0, $cover$02$i$i = 0, $exitcond = 0, $max$i = 0, $max_y$i = 0, $min$i = 0, $n$i = 0, $num_bands$i = 0, $or$cond$i$i = 0;
 var $vec$0$i$i = 0, $vec$01$i$i = 0, $vec$03$i$i = 0, $worker = 0, $x$03$i$i = 0, $yindex$06$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 18160|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $bands$i = sp + 17840|0;
 $band$i = sp + 17832|0;
 $n$i = sp + 17828|0;
 $num_bands$i = sp + 17824|0;
 $min$i = sp + 17820|0;
 $max$i = sp + 17816|0;
 $max_y$i = sp + 17812|0;
 $worker = sp + 16384|0;
 $buffer = sp;
 $0 = ((($params)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$params>>2]|0;
 $3 = ($raster|0)==(0|0);
 if ($3) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ($1|0)==(0|0);
 if ($4) {
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = ((($1)) + 2|0);
 $6 = HEAP16[$5>>1]|0;
 $7 = $6 << 16 >> 16;
 $8 = ($6<<16>>16)==(0);
 if ($8) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $9 = HEAP16[$1>>1]|0;
 $10 = $9 << 16 >> 16;
 $11 = ($9<<16>>16)<(1);
 if ($11) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $12 = ((($1)) + 12|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13|0)==(0|0);
 if ($14) {
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 $15 = ((($1)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ($16|0)==(0|0);
 if ($17) {
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 $18 = (($10) + -1)|0;
 $19 = (($13) + ($18<<1)|0);
 $20 = HEAP16[$19>>1]|0;
 $21 = $20 << 16 >> 16;
 $22 = (($21) + 1)|0;
 $23 = ($7|0)==($22|0);
 if (!($23)) {
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 $24 = ((($params)) + 8|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = $25 & 2;
 $27 = ($26|0)==(0);
 if ($27) {
  $28 = ($2|0)==(0|0);
  if ($28) {
   $$0 = 6;
   STACKTOP = sp;return ($$0|0);
  }
  $29 = ((($2)) + 4|0);
  $30 = HEAP32[$29>>2]|0;
  $31 = ($30|0)==(0);
  if ($31) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
  $32 = HEAP32[$2>>2]|0;
  $33 = ($32|0)==(0);
  if ($33) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
  $34 = ((($2)) + 12|0);
  $35 = HEAP32[$34>>2]|0;
  $36 = ($35|0)==(0|0);
  if ($36) {
   $$0 = 6;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $37 = $25 & 1;
 $38 = ($37|0)==(0);
 if ($38) {
  $$0 = 19;
  STACKTOP = sp;return ($$0|0);
 }
 do {
  if ($27) {
   $39 = ((($worker)) + 1180|0);
   HEAP32[$39>>2] = 0;
   $40 = ((($worker)) + 1184|0);
   HEAP32[$40>>2] = 0;
   $41 = ((($2)) + 4|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = ((($worker)) + 1188|0);
   HEAP32[$43>>2] = $42;
   $44 = HEAP32[$2>>2]|0;
   $45 = ((($worker)) + 1192|0);
   HEAP32[$45>>2] = $44;
  } else {
   $46 = $25 & 4;
   $47 = ($46|0)==(0);
   $48 = ((($worker)) + 1180|0);
   if ($47) {
    HEAP32[$48>>2] = -32768;
    $50 = ((($worker)) + 1184|0);
    HEAP32[$50>>2] = -32768;
    $51 = ((($worker)) + 1188|0);
    HEAP32[$51>>2] = 32767;
    $52 = ((($worker)) + 1192|0);
    HEAP32[$52>>2] = 32767;
    break;
   } else {
    $49 = ((($params)) + 32|0);
    ;HEAP32[$48>>2]=HEAP32[$49>>2]|0;HEAP32[$48+4>>2]=HEAP32[$49+4>>2]|0;HEAP32[$48+8>>2]=HEAP32[$49+8>>2]|0;HEAP32[$48+12>>2]=HEAP32[$49+12>>2]|0;
    break;
   }
  }
 } while(0);
 $53 = ((($worker)) + 1412|0);
 HEAP32[$53>>2] = $buffer;
 $54 = ((($worker)) + 1416|0);
 HEAP32[$54>>2] = 16384;
 $55 = ((($worker)) + 1420|0);
 HEAP32[$55>>2] = $buffer;
 $56 = ((($worker)) + 200|0);
 HEAP32[$56>>2] = 0;
 $57 = ((($worker)) + 204|0);
 HEAP32[$57>>2] = 0;
 $58 = ((($worker)) + 208|0);
 HEAP32[$58>>2] = 0;
 $59 = ((($worker)) + 188|0);
 HEAP32[$59>>2] = 0;
 $60 = ((($worker)) + 192|0);
 HEAP32[$60>>2] = 0;
 $61 = ((($worker)) + 196|0);
 HEAP32[$61>>2] = 1;
 $62 = ((($worker)) + 1136|0);
 ;HEAP32[$62>>2]=HEAP32[$1>>2]|0;HEAP32[$62+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$62+8>>2]=HEAP32[$1+8>>2]|0;HEAP32[$62+12>>2]=HEAP32[$1+12>>2]|0;HEAP32[$62+16>>2]=HEAP32[$1+16>>2]|0;
 HEAP32[$58>>2] = 0;
 HEAP32[$61>>2] = 1;
 $63 = ((($worker)) + 1404|0);
 HEAP32[$63>>2] = 128;
 $64 = ((($worker)) + 1388|0);
 HEAP32[$64>>2] = 0;
 $65 = ((($worker)) + 1400|0);
 HEAP32[$65>>2] = 0;
 if ($27) {
  $72 = ((($worker)) + 1156|0);
  ;HEAP32[$72>>2]=HEAP32[$2>>2]|0;HEAP32[$72+4>>2]=HEAP32[$2+4>>2]|0;HEAP32[$72+8>>2]=HEAP32[$2+8>>2]|0;HEAP32[$72+12>>2]=HEAP32[$2+12>>2]|0;HEAP32[$72+16>>2]=HEAP32[$2+16>>2]|0;HEAP32[$72+20>>2]=HEAP32[$2+20>>2]|0;
  $73 = ((($worker)) + 1392|0);
  HEAP32[$73>>2] = 4401;
  $74 = ((($worker)) + 1396|0);
  HEAP32[$74>>2] = $worker;
 } else {
  $66 = ((($params)) + 12|0);
  $67 = HEAP32[$66>>2]|0;
  $68 = ((($worker)) + 1392|0);
  HEAP32[$68>>2] = $67;
  $69 = ((($params)) + 28|0);
  $70 = HEAP32[$69>>2]|0;
  $71 = ((($worker)) + 1396|0);
  HEAP32[$71>>2] = $70;
 }
 $75 = ((($worker)) + 1140|0);
 $76 = HEAP32[$75>>2]|0;
 $77 = ((($worker)) + 1138|0);
 $78 = HEAP16[$77>>1]|0;
 $79 = $78 << 16 >> 16;
 $80 = (($76) + ($79<<3)|0);
 $81 = ($78<<16>>16)<(1);
 if ($81) {
  $82 = ((($worker)) + 164|0);
  ;HEAP32[$82>>2]=0|0;HEAP32[$82+4>>2]=0|0;HEAP32[$82+8>>2]=0|0;HEAP32[$82+12>>2]=0|0;
  $$pre29$i = ((($worker)) + 168|0);
  $$pre$phi$iZ2D = $$pre29$i;$116 = 0;$121 = 0;$126 = 0;$131 = 0;
 } else {
  $83 = HEAP32[$76>>2]|0;
  $84 = ((($worker)) + 168|0);
  HEAP32[$84>>2] = $83;
  $85 = ((($worker)) + 164|0);
  HEAP32[$85>>2] = $83;
  $86 = ((($76)) + 4|0);
  $87 = HEAP32[$86>>2]|0;
  $88 = ((($worker)) + 176|0);
  HEAP32[$88>>2] = $87;
  $89 = ((($worker)) + 172|0);
  HEAP32[$89>>2] = $87;
  $90 = ($78<<16>>16)>(1);
  if ($90) {
   $vec$01$i$i = ((($76)) + 8|0);
   $$pn2$i$i = $76;$101 = $87;$95 = $83;$97 = $83;$99 = $87;$vec$03$i$i = $vec$01$i$i;
   while(1) {
    $91 = HEAP32[$vec$03$i$i>>2]|0;
    $92 = ((($$pn2$i$i)) + 12|0);
    $93 = HEAP32[$92>>2]|0;
    $94 = ($91|0)<($95|0);
    if ($94) {
     HEAP32[$85>>2] = $91;
     $291 = $91;
    } else {
     $291 = $95;
    }
    $96 = ($91|0)>($97|0);
    if ($96) {
     HEAP32[$84>>2] = $91;
     $292 = $91;
    } else {
     $292 = $97;
    }
    $98 = ($93|0)<($99|0);
    if ($98) {
     HEAP32[$89>>2] = $93;
     $293 = $93;
    } else {
     $293 = $99;
    }
    $100 = ($93|0)>($101|0);
    if ($100) {
     HEAP32[$88>>2] = $93;
     $294 = $93;
    } else {
     $294 = $101;
    }
    $vec$0$i$i = ((($vec$03$i$i)) + 8|0);
    $102 = ($vec$0$i$i>>>0)<($80>>>0);
    if ($102) {
     $$pn2$i$i$phi = $vec$03$i$i;$101 = $294;$95 = $291;$97 = $292;$99 = $293;$vec$03$i$i = $vec$0$i$i;$$pn2$i$i = $$pn2$i$i$phi;
    } else {
     $104 = $291;$106 = $293;$108 = $292;$111 = $294;
     break;
    }
   }
  } else {
   $104 = $83;$106 = $87;$108 = $83;$111 = $87;
  }
  $103 = $104 >> 6;
  HEAP32[$85>>2] = $103;
  $105 = $106 >> 6;
  HEAP32[$89>>2] = $105;
  $107 = (($108) + 63)|0;
  $109 = $107 >> 6;
  HEAP32[$84>>2] = $109;
  $110 = (($111) + 63)|0;
  $112 = $110 >> 6;
  HEAP32[$88>>2] = $112;
  $$pre$phi$iZ2D = $84;$116 = $109;$121 = $103;$126 = $112;$131 = $105;
 }
 $113 = ((($worker)) + 1180|0);
 $114 = HEAP32[$113>>2]|0;
 $115 = ($116|0)>($114|0);
 L70: do {
  if ($115) {
   $117 = ((($worker)) + 164|0);
   $118 = ((($worker)) + 1188|0);
   $119 = HEAP32[$118>>2]|0;
   $120 = ($121|0)<($119|0);
   if ($120) {
    $122 = ((($worker)) + 176|0);
    $123 = ((($worker)) + 1184|0);
    $124 = HEAP32[$123>>2]|0;
    $125 = ($126|0)>($124|0);
    if ($125) {
     $127 = ((($worker)) + 172|0);
     $128 = ((($worker)) + 1192|0);
     $129 = HEAP32[$128>>2]|0;
     $130 = ($131|0)<($129|0);
     if ($130) {
      $132 = ($121|0)<($114|0);
      if ($132) {
       HEAP32[$117>>2] = $114;
       $138 = $114;
      } else {
       $138 = $121;
      }
      $133 = ($131|0)<($124|0);
      if ($133) {
       HEAP32[$127>>2] = $124;
       $142 = $124;
      } else {
       $142 = $131;
      }
      $134 = ($116|0)>($119|0);
      if ($134) {
       HEAP32[$$pre$phi$iZ2D>>2] = $119;
       $137 = $119;
      } else {
       $137 = $116;
      }
      $135 = ($126|0)>($129|0);
      if ($135) {
       HEAP32[$122>>2] = $129;
       $141 = $129;
      } else {
       $141 = $126;
      }
      $136 = (($137) - ($138))|0;
      $139 = ((($worker)) + 180|0);
      HEAP32[$139>>2] = $136;
      $140 = (($141) - ($142))|0;
      $143 = ((($worker)) + 184|0);
      HEAP32[$143>>2] = $140;
      $144 = HEAP32[$63>>2]|0;
      $145 = (($140|0) / ($144|0))&-1;
      HEAP32[$num_bands$i>>2] = $145;
      $146 = HEAP32[$num_bands$i>>2]|0;
      $147 = ($146|0)==(0);
      if ($147) {
       HEAP32[$num_bands$i>>2] = 1;
      }
      $148 = HEAP32[$num_bands$i>>2]|0;
      $149 = ($148|0)>(38);
      if ($149) {
       HEAP32[$num_bands$i>>2] = 39;
      }
      $150 = ((($worker)) + 1408|0);
      HEAP32[$150>>2] = 0;
      $151 = HEAP32[$127>>2]|0;
      HEAP32[$min$i>>2] = $151;
      $152 = HEAP32[$122>>2]|0;
      HEAP32[$max_y$i>>2] = $152;
      HEAP32[$n$i>>2] = 0;
      $153 = HEAP32[$n$i>>2]|0;
      $154 = HEAP32[$num_bands$i>>2]|0;
      $155 = ($153|0)<($154|0);
      if ($155) {
       $156 = ((($bands$i)) + 4|0);
       $157 = ((($worker)) + 1424|0);
       $158 = ((($worker)) + 1392|0);
       $159 = ((($worker)) + 1196|0);
       $160 = ((($worker)) + 1396|0);
       while(1) {
        $161 = HEAP32[$min$i>>2]|0;
        $162 = HEAP32[$63>>2]|0;
        $163 = (($162) + ($161))|0;
        HEAP32[$max$i>>2] = $163;
        $164 = HEAP32[$n$i>>2]|0;
        $165 = HEAP32[$num_bands$i>>2]|0;
        $166 = (($165) + -1)|0;
        $167 = ($164|0)==($166|0);
        if ($167) {
         label = 56;
        } else {
         $168 = HEAP32[$max$i>>2]|0;
         $169 = HEAP32[$max_y$i>>2]|0;
         $170 = ($168|0)>($169|0);
         if ($170) {
          label = 56;
         }
        }
        if ((label|0) == 56) {
         label = 0;
         $171 = HEAP32[$max_y$i>>2]|0;
         HEAP32[$max$i>>2] = $171;
        }
        $172 = HEAP32[$min$i>>2]|0;
        HEAP32[$bands$i>>2] = $172;
        $173 = HEAP32[$max$i>>2]|0;
        HEAP32[$156>>2] = $173;
        HEAP32[$band$i>>2] = $bands$i;
        $174 = HEAP32[$band$i>>2]|0;
        $175 = ($174>>>0)<($bands$i>>>0);
        if (!($175)) {
         while(1) {
          $176 = HEAP32[$53>>2]|0;
          HEAP32[$55>>2] = $176;
          $177 = HEAP32[$band$i>>2]|0;
          $178 = ((($177)) + 4|0);
          $179 = HEAP32[$178>>2]|0;
          $180 = HEAP32[$band$i>>2]|0;
          $181 = HEAP32[$180>>2]|0;
          $182 = (($179) - ($181))|0;
          HEAP32[$157>>2] = $182;
          $183 = $182 << 2;
          $184 = (($183|0) % 16)&-1;
          $185 = ($184|0)>(0);
          $186 = (16 - ($184))|0;
          $187 = $185 ? $186 : 0;
          $cell_start$0$i = (($187) + ($183))|0;
          $188 = HEAP32[$54>>2]|0;
          $189 = (($188|0) % 16)&-1;
          $190 = (($188) - ($189))|0;
          $191 = $176;
          $192 = (($191) + ($cell_start$0$i)|0);
          HEAP32[$56>>2] = $192;
          $193 = ($190|0)>($cell_start$0$i|0);
          $194 = $176;
          L105: do {
           if ($193) {
            $195 = (($190) - ($cell_start$0$i))|0;
            $196 = $195 >> 4;
            HEAP32[$57>>2] = $196;
            $197 = ($195|0)<(32);
            if ($197) {
             label = 80;
             break;
            }
            $198 = ($182|0)>(0);
            do {
             if ($198) {
              HEAP32[$194>>2] = 0;
              $199 = ($182|0)>(1);
              if ($199) {
               $201 = 1;
              } else {
               break;
              }
              while(1) {
               $$pre$i = HEAP32[$55>>2]|0;
               $200 = (($$pre$i) + ($201<<2)|0);
               HEAP32[$200>>2] = 0;
               $202 = (($201) + 1)|0;
               $exitcond = ($202|0)==($182|0);
               if ($exitcond) {
                break;
               } else {
                $201 = $202;
               }
              }
             }
            } while(0);
            HEAP32[$58>>2] = 0;
            HEAP32[$61>>2] = 1;
            $203 = HEAP32[$band$i>>2]|0;
            $204 = HEAP32[$203>>2]|0;
            HEAP32[$127>>2] = $204;
            $205 = HEAP32[$band$i>>2]|0;
            $206 = ((($205)) + 4|0);
            $207 = HEAP32[$206>>2]|0;
            HEAP32[$122>>2] = $207;
            $208 = HEAP32[$band$i>>2]|0;
            $209 = ((($208)) + 4|0);
            $210 = HEAP32[$209>>2]|0;
            $211 = HEAP32[$band$i>>2]|0;
            $212 = HEAP32[$211>>2]|0;
            $213 = (($210) - ($212))|0;
            HEAP32[$143>>2] = $213;
            $214 = (_gray_convert_glyph_inner($worker)|0);
            switch ($214|0) {
            case 64:  {
             label = 80;
             break L105;
             break;
            }
            case 0:  {
             break;
            }
            default: {
             $$3$i = 1;
             break L70;
            }
            }
            $215 = HEAP32[$58>>2]|0;
            $216 = ($215|0)==(0);
            do {
             if (!($216)) {
              HEAP32[$64>>2] = 0;
              $217 = HEAP32[$157>>2]|0;
              $218 = ($217|0)>(0);
              if ($218) {
               $yindex$06$i$i = 0;
               while(1) {
                $219 = HEAP32[$55>>2]|0;
                $220 = (($219) + ($yindex$06$i$i<<2)|0);
                $cell$01$i$i = HEAP32[$220>>2]|0;
                $221 = ($cell$01$i$i|0)==(0|0);
                do {
                 if (!($221)) {
                  $cell$04$i$i = $cell$01$i$i;$cover$02$i$i = 0;$x$03$i$i = 0;
                  while(1) {
                   $222 = HEAP32[$cell$04$i$i>>2]|0;
                   $223 = ($222|0)>($x$03$i$i|0);
                   $224 = ($cover$02$i$i|0)!=(0);
                   $or$cond$i$i = $224 & $223;
                   if ($or$cond$i$i) {
                    $225 = $cover$02$i$i << 9;
                    $226 = (($222) - ($x$03$i$i))|0;
                    _gray_hline($worker,$x$03$i$i,$yindex$06$i$i,$225,$226);
                   }
                   $227 = ((($cell$04$i$i)) + 4|0);
                   $228 = HEAP32[$227>>2]|0;
                   $229 = (($228) + ($cover$02$i$i))|0;
                   $230 = $229 << 9;
                   $231 = ((($cell$04$i$i)) + 8|0);
                   $232 = HEAP32[$231>>2]|0;
                   $233 = (($230) - ($232))|0;
                   $234 = ($230|0)==($232|0);
                   do {
                    if (!($234)) {
                     $235 = HEAP32[$cell$04$i$i>>2]|0;
                     $236 = ($235|0)>(-1);
                     if (!($236)) {
                      break;
                     }
                     _gray_hline($worker,$235,$yindex$06$i$i,$233,1);
                    }
                   } while(0);
                   $237 = HEAP32[$cell$04$i$i>>2]|0;
                   $238 = (($237) + 1)|0;
                   $239 = ((($cell$04$i$i)) + 12|0);
                   $cell$0$i$i = HEAP32[$239>>2]|0;
                   $240 = ($cell$0$i$i|0)==(0|0);
                   if ($240) {
                    $$lcssa = $229;$$lcssa31 = $230;$$lcssa32 = $238;
                    break;
                   } else {
                    $cell$04$i$i = $cell$0$i$i;$cover$02$i$i = $229;$x$03$i$i = $238;
                   }
                  }
                  $241 = ($$lcssa|0)==(0);
                  if ($241) {
                   break;
                  }
                  $242 = HEAP32[$139>>2]|0;
                  $243 = (($242) - ($$lcssa32))|0;
                  _gray_hline($worker,$$lcssa32,$yindex$06$i$i,$$lcssa31,$243);
                 }
                } while(0);
                $244 = (($yindex$06$i$i) + 1)|0;
                $245 = HEAP32[$157>>2]|0;
                $246 = ($244|0)<($245|0);
                if ($246) {
                 $yindex$06$i$i = $244;
                } else {
                 break;
                }
               }
              }
              $247 = HEAP32[$158>>2]|0;
              $248 = ($247|0)==(0|0);
              if ($248) {
               break;
              }
              $249 = HEAP32[$64>>2]|0;
              $250 = ($249|0)>(0);
              if (!($250)) {
               break;
              }
              $251 = HEAP32[$65>>2]|0;
              $252 = HEAP32[$160>>2]|0;
              FUNCTION_TABLE_viiii[$247 & 8191]($251,$249,$159,$252);
             }
            } while(0);
            $253 = HEAP32[$band$i>>2]|0;
            $254 = ((($253)) + -8|0);
            HEAP32[$band$i>>2] = $254;
           } else {
            label = 80;
           }
          } while(0);
          if ((label|0) == 80) {
           label = 0;
           $255 = HEAP32[$band$i>>2]|0;
           $256 = HEAP32[$255>>2]|0;
           $257 = HEAP32[$band$i>>2]|0;
           $258 = ((($257)) + 4|0);
           $259 = HEAP32[$258>>2]|0;
           $260 = (($259) - ($256))|0;
           $261 = $260 >> 1;
           $262 = (($261) + ($256))|0;
           $263 = ($261|0)==(0);
           if ($263) {
            $$3$i = 1;
            break L70;
           }
           $264 = (($256) - ($259))|0;
           $265 = HEAP32[$63>>2]|0;
           $266 = ($264|0)<($265|0);
           if (!($266)) {
            $267 = HEAP32[$150>>2]|0;
            $268 = (($267) + 1)|0;
            HEAP32[$150>>2] = $268;
           }
           $269 = HEAP32[$band$i>>2]|0;
           $270 = ((($269)) + 8|0);
           HEAP32[$270>>2] = $256;
           $271 = HEAP32[$band$i>>2]|0;
           $272 = ((($271)) + 12|0);
           HEAP32[$272>>2] = $262;
           $273 = HEAP32[$band$i>>2]|0;
           HEAP32[$273>>2] = $262;
           $274 = HEAP32[$band$i>>2]|0;
           $275 = ((($274)) + 4|0);
           HEAP32[$275>>2] = $259;
           $276 = HEAP32[$band$i>>2]|0;
           $277 = ((($276)) + 8|0);
           HEAP32[$band$i>>2] = $277;
          }
          $278 = HEAP32[$band$i>>2]|0;
          $279 = ($278>>>0)<($bands$i>>>0);
          if ($279) {
           break;
          }
         }
        }
        $280 = HEAP32[$n$i>>2]|0;
        $281 = (($280) + 1)|0;
        HEAP32[$n$i>>2] = $281;
        $282 = HEAP32[$max$i>>2]|0;
        HEAP32[$min$i>>2] = $282;
        $283 = HEAP32[$n$i>>2]|0;
        $284 = HEAP32[$num_bands$i>>2]|0;
        $285 = ($283|0)<($284|0);
        if (!($285)) {
         break;
        }
       }
      }
      $286 = HEAP32[$150>>2]|0;
      $287 = ($286|0)>(8);
      if (!($287)) {
       $$3$i = 0;
       break;
      }
      $288 = HEAP32[$63>>2]|0;
      $289 = ($288|0)>(16);
      if (!($289)) {
       $$3$i = 0;
       break;
      }
      $290 = (($288|0) / 2)&-1;
      HEAP32[$63>>2] = $290;
      $$3$i = 0;
     } else {
      $$3$i = 0;
     }
    } else {
     $$3$i = 0;
    }
   } else {
    $$3$i = 0;
   }
  } else {
   $$3$i = 0;
  }
 } while(0);
 $$0 = $$3$i;
 STACKTOP = sp;return ($$0|0);
}
function _gray_render_span($y,$count,$spans,$worker) {
 $y = $y|0;
 $count = $count|0;
 $spans = $spans|0;
 $worker = $worker|0;
 var $$012 = 0, $$03 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p$0$ph = 0, $q$0 = 0, $q$1 = 0, $q$2 = 0, $q$3 = 0;
 var $q$4 = 0, $q$5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($worker)) + 1168|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($worker)) + 1164|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (0 - ($y))|0;
 $5 = Math_imul($3, $4)|0;
 $6 = (($1) + ($5)|0);
 $7 = ($3|0)>(-1);
 if ($7) {
  $8 = ((($worker)) + 1156|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = (($9) + -1)|0;
  $11 = Math_imul($10, $3)|0;
  $12 = (($6) + ($11)|0);
  $p$0$ph = $12;
 } else {
  $p$0$ph = $6;
 }
 $13 = ($count|0)>(0);
 if ($13) {
  $$012 = $spans;$$03 = $count;
 } else {
  return;
 }
 while(1) {
  $14 = ((($$012)) + 4|0);
  $15 = HEAP8[$14>>0]|0;
  $16 = ($15<<24>>24)==(0);
  L8: do {
   if (!($16)) {
    $17 = ((($$012)) + 2|0);
    $18 = HEAP16[$17>>1]|0;
    $19 = $18&65535;
    $20 = ($18&65535)>(7);
    $21 = HEAP16[$$012>>1]|0;
    $22 = $21 << 16 >> 16;
    $23 = (($p$0$ph) + ($22)|0);
    if ($20) {
     _memset(($23|0),($15|0),($19|0))|0;
     break;
    }
    switch ($19|0) {
    case 7:  {
     $24 = ((($23)) + 1|0);
     HEAP8[$23>>0] = $15;
     $q$0 = $24;
     label = 9;
     break;
    }
    case 6:  {
     $q$0 = $23;
     label = 9;
     break;
    }
    case 5:  {
     $q$1 = $23;
     label = 10;
     break;
    }
    case 4:  {
     $q$2 = $23;
     label = 11;
     break;
    }
    case 3:  {
     $q$3 = $23;
     label = 12;
     break;
    }
    case 2:  {
     $q$4 = $23;
     label = 13;
     break;
    }
    case 1:  {
     $q$5 = $23;
     break;
    }
    default: {
     break L8;
    }
    }
    if ((label|0) == 9) {
     label = 0;
     $25 = ((($q$0)) + 1|0);
     HEAP8[$q$0>>0] = $15;
     $q$1 = $25;
     label = 10;
    }
    if ((label|0) == 10) {
     label = 0;
     $26 = ((($q$1)) + 1|0);
     HEAP8[$q$1>>0] = $15;
     $q$2 = $26;
     label = 11;
    }
    if ((label|0) == 11) {
     label = 0;
     $27 = ((($q$2)) + 1|0);
     HEAP8[$q$2>>0] = $15;
     $q$3 = $27;
     label = 12;
    }
    if ((label|0) == 12) {
     label = 0;
     $28 = ((($q$3)) + 1|0);
     HEAP8[$q$3>>0] = $15;
     $q$4 = $28;
     label = 13;
    }
    if ((label|0) == 13) {
     label = 0;
     $29 = ((($q$4)) + 1|0);
     HEAP8[$q$4>>0] = $15;
     $q$5 = $29;
    }
    HEAP8[$q$5>>0] = $15;
   }
  } while(0);
  $30 = (($$03) + -1)|0;
  $31 = ((($$012)) + 6|0);
  $32 = ($$03|0)>(1);
  if ($32) {
   $$012 = $31;$$03 = $30;
  } else {
   break;
  }
 }
 return;
}
function _gray_convert_glyph_inner($worker) {
 $worker = $worker|0;
 var $$$i$i = 0, $$reg2mem64$0 = 0, $$reg2mem66$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $cell$0$i$i$reg2mem$0 = 0, $cond = 0, $cond68 = 0, $error = 0, $pcell$0$lcssa$i$i$reg2mem60$0 = 0, $pcell$03$i$i$reg2mem62$0 = 0, _setjmpTable = 0, _setjmpTableSize = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 _setjmpTableSize = 4;_setjmpTable = _malloc(40) | 0;HEAP32[_setjmpTable>>2]=0;
 $error = sp;
 HEAP32[$error>>2] = 0;
 _setjmpTable = _saveSetjmp($worker,1,_setjmpTable|0,_setjmpTableSize|0)|0;_setjmpTableSize = tempRet0;
 __THREW__ = 0;
 $0 = __THREW__; __THREW__ = 0;
 if ((($0|0) != 0) & ((threwValue|0) != 0)) { $1 = _testSetjmp(HEAP32[$0>>2]|0, _setjmpTable|0, _setjmpTableSize|0)|0; if (($1|0) == 0) { _longjmp($0|0, threwValue|0); } tempRet0 = (threwValue); } else { $1 = -1; };
 $2 = tempRet0;
 $cond = ($1|0)==(1);
 if ($cond) {
  $$reg2mem66$0 = $2;
 } else {
  $$reg2mem66$0 = 0;
 }
 L3: while(1) {
  $3 = ($$reg2mem66$0|0)==(0);
  if (!($3)) {
   label = 15;
   break;
  }
  $4 = ((($worker)) + 1136|0);
  __THREW__ = 0;
  $5 = (invoke_iiii(4402,($4|0),(55392|0),($worker|0))|0);
  $6 = __THREW__; __THREW__ = 0;
  if ((($6|0) != 0) & ((threwValue|0) != 0)) { $7 = _testSetjmp(HEAP32[$6>>2]|0, _setjmpTable|0, _setjmpTableSize|0)|0; if (($7|0) == 0) { _longjmp($6|0, threwValue|0); } tempRet0 = (threwValue); } else { $7 = -1; };
  $8 = tempRet0;
  $cond68 = ($7|0)==(1);
  if ($cond68) {
   $$reg2mem66$0 = $8;
   continue;
  }
  HEAP32[$error>>2] = $5;
  $9 = ((($worker)) + 196|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ($10|0)==(0);
  if (!($11)) {
   label = 16;
   break;
  }
  $12 = ((($worker)) + 188|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ((($worker)) + 192|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = $15 | $13;
  $17 = ($16|0)==(0);
  if ($17) {
   label = 16;
   break;
  }
  $18 = ((($worker)) + 156|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = ((($worker)) + 180|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = ($19|0)>($21|0);
  $$$i$i = $22 ? $21 : $19;
  $23 = ((($worker)) + 160|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = ((($worker)) + 1420|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = (($26) + ($24<<2)|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = ($28|0)==(0|0);
  L9: do {
   if ($29) {
    $pcell$0$lcssa$i$i$reg2mem60$0 = $27;
   } else {
    $$reg2mem64$0 = $28;$pcell$03$i$i$reg2mem62$0 = $27;
    while(1) {
     $30 = HEAP32[$$reg2mem64$0>>2]|0;
     $31 = ($30|0)>($$$i$i|0);
     if ($31) {
      $pcell$0$lcssa$i$i$reg2mem60$0 = $pcell$03$i$i$reg2mem62$0;
      break L9;
     }
     $32 = ($30|0)==($$$i$i|0);
     if ($32) {
      $cell$0$i$i$reg2mem$0 = $$reg2mem64$0;
      break L3;
     }
     $33 = ((($$reg2mem64$0)) + 12|0);
     $34 = HEAP32[$33>>2]|0;
     $35 = ($34|0)==(0|0);
     if ($35) {
      $pcell$0$lcssa$i$i$reg2mem60$0 = $33;
      break;
     } else {
      $$reg2mem64$0 = $34;$pcell$03$i$i$reg2mem62$0 = $33;
     }
    }
   }
  } while(0);
  $36 = ((($worker)) + 208|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = ((($worker)) + 204|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = ($37|0)<($39|0);
  if ($40) {
   label = 13;
   break;
  }
  __THREW__ = 0;
  invoke_vii(4403,($worker|0),1);
  $41 = __THREW__; __THREW__ = 0;
  if ((($41|0) != 0) & ((threwValue|0) != 0)) { $42 = _testSetjmp(HEAP32[$41>>2]|0, _setjmpTable|0, _setjmpTableSize|0)|0; if (($42|0) == 0) { _longjmp($41|0, threwValue|0); } tempRet0 = (threwValue); } else { $42 = -1; };
  $43 = tempRet0;
  $$reg2mem66$0 = $43;
 }
 if ((label|0) == 13) {
  $44 = ((($worker)) + 200|0);
  $45 = HEAP32[$44>>2]|0;
  $46 = (($37) + 1)|0;
  HEAP32[$36>>2] = $46;
  $47 = (($45) + ($37<<4)|0);
  HEAP32[$47>>2] = $$$i$i;
  $48 = (((($45) + ($37<<4)|0)) + 8|0);
  HEAP32[$48>>2] = 0;
  $49 = (((($45) + ($37<<4)|0)) + 4|0);
  HEAP32[$49>>2] = 0;
  $50 = HEAP32[$pcell$0$lcssa$i$i$reg2mem60$0>>2]|0;
  $51 = (((($45) + ($37<<4)|0)) + 12|0);
  HEAP32[$51>>2] = $50;
  HEAP32[$pcell$0$lcssa$i$i$reg2mem60$0>>2] = $47;
  $cell$0$i$i$reg2mem$0 = $47;
 }
 else if ((label|0) == 15) {
  HEAP32[$error>>2] = 64;
  $58 = HEAP32[$error>>2]|0;
  _free(_setjmpTable|0);
  STACKTOP = sp;return ($58|0);
 }
 else if ((label|0) == 16) {
  $58 = HEAP32[$error>>2]|0;
  _free(_setjmpTable|0);
  STACKTOP = sp;return ($58|0);
 }
 $52 = ((($cell$0$i$i$reg2mem$0)) + 8|0);
 $53 = HEAP32[$52>>2]|0;
 $54 = (($53) + ($13))|0;
 HEAP32[$52>>2] = $54;
 $55 = ((($cell$0$i$i$reg2mem$0)) + 4|0);
 $56 = HEAP32[$55>>2]|0;
 $57 = (($56) + ($15))|0;
 HEAP32[$55>>2] = $57;
 $58 = HEAP32[$error>>2]|0;
 _free(_setjmpTable|0);
 STACKTOP = sp;return ($58|0);
}
function _gray_move_to($to,$worker) {
 $to = $to|0;
 $worker = $worker|0;
 var $$$i$i = 0, $$$i$i1 = 0, $$1$i = 0, $$ex$i = 0, $$ex$i$i = 0, $$pre$phi5Z2D = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $9 = 0, $cell$0$i$i = 0, $pcell$0$lcssa$i$i = 0, $pcell$03$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($worker)) + 196|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 $3 = ((($worker)) + 188|0);
 if ($2) {
  $4 = HEAP32[$3>>2]|0;
  $5 = ((($worker)) + 192|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = $6 | $4;
  $8 = ($7|0)==(0);
  if ($8) {
   $$pre$phi5Z2D = $5;
  } else {
   $9 = ((($worker)) + 156|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ((($worker)) + 180|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($10|0)>($12|0);
   $$$i$i = $13 ? $12 : $10;
   $14 = ((($worker)) + 160|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ((($worker)) + 1420|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = (($17) + ($15<<2)|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ($19|0)==(0|0);
   L4: do {
    if ($20) {
     $pcell$0$lcssa$i$i = $18;
     label = 8;
    } else {
     $22 = $19;$pcell$03$i$i = $18;
     while(1) {
      $21 = HEAP32[$22>>2]|0;
      $23 = ($21|0)>($$$i$i|0);
      if ($23) {
       $pcell$0$lcssa$i$i = $pcell$03$i$i;
       label = 8;
       break L4;
      }
      $24 = ($21|0)==($$$i$i|0);
      if ($24) {
       $cell$0$i$i = $22;
       break L4;
      }
      $25 = ((($22)) + 12|0);
      $26 = HEAP32[$25>>2]|0;
      $27 = ($26|0)==(0|0);
      if ($27) {
       $pcell$0$lcssa$i$i = $25;
       label = 8;
       break;
      } else {
       $22 = $26;$pcell$03$i$i = $25;
      }
     }
    }
   } while(0);
   do {
    if ((label|0) == 8) {
     $28 = ((($worker)) + 208|0);
     $29 = HEAP32[$28>>2]|0;
     $30 = ((($worker)) + 204|0);
     $31 = HEAP32[$30>>2]|0;
     $32 = ($29|0)<($31|0);
     if ($32) {
      $33 = ((($worker)) + 200|0);
      $34 = HEAP32[$33>>2]|0;
      $35 = (($29) + 1)|0;
      HEAP32[$28>>2] = $35;
      $36 = (($34) + ($29<<4)|0);
      HEAP32[$36>>2] = $$$i$i;
      $37 = (((($34) + ($29<<4)|0)) + 8|0);
      HEAP32[$37>>2] = 0;
      $38 = (((($34) + ($29<<4)|0)) + 4|0);
      HEAP32[$38>>2] = 0;
      $39 = HEAP32[$pcell$0$lcssa$i$i>>2]|0;
      $40 = (((($34) + ($29<<4)|0)) + 12|0);
      HEAP32[$40>>2] = $39;
      HEAP32[$pcell$0$lcssa$i$i>>2] = $36;
      $cell$0$i$i = $36;
      break;
     } else {
      _longjmp(($worker|0),1);
      // unreachable;
     }
    }
   } while(0);
   $41 = ((($cell$0$i$i)) + 8|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = (($42) + ($4))|0;
   HEAP32[$41>>2] = $43;
   $44 = ((($cell$0$i$i)) + 4|0);
   $45 = HEAP32[$44>>2]|0;
   $46 = (($45) + ($6))|0;
   HEAP32[$44>>2] = $46;
   $$pre$phi5Z2D = $5;
  }
 } else {
  $$pre4 = ((($worker)) + 192|0);
  $$pre$phi5Z2D = $$pre4;
 }
 $47 = HEAP32[$to>>2]|0;
 $48 = $47 << 2;
 $49 = ((($to)) + 4|0);
 $50 = HEAP32[$49>>2]|0;
 $51 = $50 << 2;
 $52 = $48 >> 8;
 $53 = $51 >> 8;
 $54 = ((($worker)) + 168|0);
 $55 = HEAP32[$54>>2]|0;
 $56 = ($55|0)<($52|0);
 $$ex$i = $56 ? $55 : $52;
 $57 = ((($worker)) + 164|0);
 $58 = HEAP32[$57>>2]|0;
 $59 = ($$ex$i|0)<($58|0);
 $60 = (($58) + -1)|0;
 $$1$i = $59 ? $60 : $$ex$i;
 HEAP32[$3>>2] = 0;
 HEAP32[$$pre$phi5Z2D>>2] = 0;
 $61 = (($$1$i) - ($58))|0;
 $62 = ((($worker)) + 156|0);
 HEAP32[$62>>2] = $61;
 $63 = ((($worker)) + 172|0);
 $64 = HEAP32[$63>>2]|0;
 $65 = (($53) - ($64))|0;
 $66 = ((($worker)) + 160|0);
 HEAP32[$66>>2] = $65;
 $67 = $53 << 8;
 $68 = ((($worker)) + 228|0);
 HEAP32[$68>>2] = $67;
 HEAP32[$0>>2] = 0;
 $69 = ($55|0)<($$1$i|0);
 $$ex$i$i = $69 ? $55 : $$1$i;
 $70 = (($$ex$i$i) - ($58))|0;
 $71 = ($70|0)>(-1);
 $$$i$i1 = $71 ? $70 : -1;
 $72 = ($$$i$i1|0)==($61|0);
 if (!($72)) {
  HEAP32[$3>>2] = 0;
  HEAP32[$$pre$phi5Z2D>>2] = 0;
  HEAP32[$62>>2] = $$$i$i1;
  HEAP32[$66>>2] = $65;
 }
 $73 = ((($worker)) + 184|0);
 $74 = HEAP32[$73>>2]|0;
 $75 = ($65>>>0)<($74>>>0);
 if (!($75)) {
  $80 = 1;
  $79 = $80&1;
  HEAP32[$0>>2] = $79;
  $81 = ((($worker)) + 220|0);
  HEAP32[$81>>2] = $48;
  $82 = ((($worker)) + 224|0);
  HEAP32[$82>>2] = $51;
  return 0;
 }
 $76 = ((($worker)) + 180|0);
 $77 = HEAP32[$76>>2]|0;
 $78 = ($$$i$i1|0)>=($77|0);
 $80 = $78;
 $79 = $80&1;
 HEAP32[$0>>2] = $79;
 $81 = ((($worker)) + 220|0);
 HEAP32[$81>>2] = $48;
 $82 = ((($worker)) + 224|0);
 HEAP32[$82>>2] = $51;
 return 0;
}
function _gray_line_to($to,$worker) {
 $to = $to|0;
 $worker = $worker|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$to>>2]|0;
 $1 = $0 << 2;
 $2 = ((($to)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 2;
 _gray_render_line($worker,$1,$4);
 return 0;
}
function _gray_render_line($worker,$to_x,$to_y) {
 $worker = $worker|0;
 $to_x = $to_x|0;
 $to_y = $to_y|0;
 var $$ = 0, $$$i = 0, $$$i$i = 0, $$$i$i11 = 0, $$$i$i22 = 0, $$$i$i33 = 0, $$$i20 = 0, $$$i31 = 0, $$$i9 = 0, $$1 = 0, $$4 = 0, $$7 = 0, $$ex$i = 0, $$ex$i19 = 0, $$ex$i30 = 0, $$ex$i8 = 0, $$lcssa = 0, $$lcssa128 = 0, $$lcssa130 = 0, $$lcssa131 = 0;
 var $$lobit = 0, $$lobit41 = 0, $$lobit42 = 0, $$lobit42$not = 0, $$pre$phi80Z2D = 0, $$pre79 = 0, $$promoted = 0, $$promoted57 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0;
 var $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0;
 var $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0;
 var $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0;
 var $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0;
 var $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0;
 var $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0;
 var $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0;
 var $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cell$0$i$i = 0, $cell$0$i$i16 = 0, $cell$0$i$i27 = 0, $cell$0$i$i38 = 0, $delta$0 = 0, $delta$1 = 0, $ey1$049 = 0, $ey1$152 = 0, $lift$0 = 0;
 var $mod$0 = 0, $mod$151 = 0, $mod$2 = 0, $or$cond = 0, $or$cond96 = 0, $or$cond97 = 0, $p$0 = 0, $pcell$0$lcssa$i$i = 0, $pcell$0$lcssa$i$i14 = 0, $pcell$0$lcssa$i$i25 = 0, $pcell$0$lcssa$i$i36 = 0, $pcell$03$i$i = 0, $pcell$03$i$i12 = 0, $pcell$03$i$i23 = 0, $pcell$03$i$i34 = 0, $rem$0 = 0, $x$050 = 0, $x$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($worker)) + 228|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 >> 8;
 $3 = $to_y >> 8;
 $4 = ((($worker)) + 224|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($5) - ($1))|0;
 $7 = $3 << 8;
 $8 = (($to_y) - ($7))|0;
 $9 = ((($worker)) + 220|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($to_x) - ($10))|0;
 $12 = (($to_y) - ($5))|0;
 $13 = ($2|0)>($3|0);
 $$ = $13 ? $3 : $2;
 $14 = ((($worker)) + 176|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($$|0)<($15|0);
 if (!($16)) {
  HEAP32[$9>>2] = $to_x;
  HEAP32[$4>>2] = $to_y;
  HEAP32[$0>>2] = $7;
  return;
 }
 $$1 = $13 ? $2 : $3;
 $17 = ((($worker)) + 172|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ($$1|0)<($18|0);
 if ($19) {
  HEAP32[$9>>2] = $to_x;
  HEAP32[$4>>2] = $to_y;
  HEAP32[$0>>2] = $7;
  return;
 }
 $20 = ($2|0)==($3|0);
 if ($20) {
  _gray_render_scanline($worker,$2,$10,$6,$to_x,$8);
  HEAP32[$9>>2] = $to_x;
  HEAP32[$4>>2] = $to_y;
  HEAP32[$0>>2] = $7;
  return;
 }
 $21 = ($10|0)==($to_x|0);
 if ($21) {
  $22 = $to_x >> 8;
  $23 = $22 << 8;
  $24 = (($to_x) - ($23))|0;
  $25 = $24 << 1;
  $26 = $12 >> 31;
  $27 = (($26) + 256)|0;
  $28 = $27 & -256;
  $29 = $26 | 1;
  $30 = (($28) - ($6))|0;
  $31 = Math_imul($25, $30)|0;
  $32 = ((($worker)) + 188|0);
  $33 = HEAP32[$32>>2]|0;
  $34 = (($33) + ($31))|0;
  HEAP32[$32>>2] = $34;
  $35 = ((($worker)) + 192|0);
  $36 = HEAP32[$35>>2]|0;
  $37 = (($36) + ($30))|0;
  HEAP32[$35>>2] = $37;
  $38 = (($29) + ($2))|0;
  $39 = (($38) - ($18))|0;
  $40 = ((($worker)) + 168|0);
  $41 = HEAP32[$40>>2]|0;
  $42 = ($41|0)<($22|0);
  $$ex$i = $42 ? $41 : $22;
  $43 = ((($worker)) + 164|0);
  $44 = HEAP32[$43>>2]|0;
  $45 = (($$ex$i) - ($44))|0;
  $46 = ($45|0)>(-1);
  $$$i = $46 ? $45 : -1;
  $47 = ((($worker)) + 156|0);
  $48 = HEAP32[$47>>2]|0;
  $49 = ($$$i|0)==($48|0);
  if ($49) {
   $50 = ((($worker)) + 160|0);
   $51 = HEAP32[$50>>2]|0;
   $52 = ($39|0)==($51|0);
   if ($52) {
    $$promoted = $34;$$promoted57 = $37;
   } else {
    label = 8;
   }
  } else {
   label = 8;
  }
  if ((label|0) == 8) {
   $53 = ((($worker)) + 196|0);
   $54 = HEAP32[$53>>2]|0;
   $55 = ($54|0)!=(0);
   $56 = $37 | $34;
   $57 = ($56|0)==(0);
   $or$cond96 = $55 | $57;
   if (!($or$cond96)) {
    $58 = ((($worker)) + 180|0);
    $59 = HEAP32[$58>>2]|0;
    $60 = ($48|0)>($59|0);
    $$$i$i = $60 ? $59 : $48;
    $61 = ((($worker)) + 160|0);
    $62 = HEAP32[$61>>2]|0;
    $63 = ((($worker)) + 1420|0);
    $64 = HEAP32[$63>>2]|0;
    $65 = (($64) + ($62<<2)|0);
    $66 = HEAP32[$65>>2]|0;
    $67 = ($66|0)==(0|0);
    L19: do {
     if ($67) {
      $pcell$0$lcssa$i$i = $65;
      label = 13;
     } else {
      $69 = $66;$pcell$03$i$i = $65;
      while(1) {
       $68 = HEAP32[$69>>2]|0;
       $70 = ($68|0)>($$$i$i|0);
       if ($70) {
        $pcell$0$lcssa$i$i = $pcell$03$i$i;
        label = 13;
        break L19;
       }
       $71 = ($68|0)==($$$i$i|0);
       if ($71) {
        $cell$0$i$i = $69;
        break L19;
       }
       $72 = ((($69)) + 12|0);
       $73 = HEAP32[$72>>2]|0;
       $74 = ($73|0)==(0|0);
       if ($74) {
        $pcell$0$lcssa$i$i = $72;
        label = 13;
        break;
       } else {
        $69 = $73;$pcell$03$i$i = $72;
       }
      }
     }
    } while(0);
    do {
     if ((label|0) == 13) {
      $75 = ((($worker)) + 208|0);
      $76 = HEAP32[$75>>2]|0;
      $77 = ((($worker)) + 204|0);
      $78 = HEAP32[$77>>2]|0;
      $79 = ($76|0)<($78|0);
      if ($79) {
       $80 = ((($worker)) + 200|0);
       $81 = HEAP32[$80>>2]|0;
       $82 = (($76) + 1)|0;
       HEAP32[$75>>2] = $82;
       $83 = (($81) + ($76<<4)|0);
       HEAP32[$83>>2] = $$$i$i;
       $84 = (((($81) + ($76<<4)|0)) + 8|0);
       HEAP32[$84>>2] = 0;
       $85 = (((($81) + ($76<<4)|0)) + 4|0);
       HEAP32[$85>>2] = 0;
       $86 = HEAP32[$pcell$0$lcssa$i$i>>2]|0;
       $87 = (((($81) + ($76<<4)|0)) + 12|0);
       HEAP32[$87>>2] = $86;
       HEAP32[$pcell$0$lcssa$i$i>>2] = $83;
       $cell$0$i$i = $83;
       break;
      } else {
       _longjmp(($worker|0),1);
       // unreachable;
      }
     }
    } while(0);
    $88 = ((($cell$0$i$i)) + 8|0);
    $89 = HEAP32[$88>>2]|0;
    $90 = (($89) + ($34))|0;
    HEAP32[$88>>2] = $90;
    $91 = ((($cell$0$i$i)) + 4|0);
    $92 = HEAP32[$91>>2]|0;
    $93 = (($92) + ($37))|0;
    HEAP32[$91>>2] = $93;
   }
   HEAP32[$32>>2] = 0;
   HEAP32[$35>>2] = 0;
   HEAP32[$47>>2] = $$$i;
   $94 = ((($worker)) + 160|0);
   HEAP32[$94>>2] = $39;
   $$promoted = 0;$$promoted57 = 0;
  }
  $95 = ((($worker)) + 184|0);
  $96 = HEAP32[$95>>2]|0;
  $97 = ($39>>>0)<($96>>>0);
  if ($97) {
   $98 = ((($worker)) + 180|0);
   $99 = HEAP32[$98>>2]|0;
   $100 = ($$$i|0)>=($99|0);
   $102 = $100;
  } else {
   $102 = 1;
  }
  $101 = $102&1;
  $103 = ((($worker)) + 196|0);
  HEAP32[$103>>2] = $101;
  $104 = $28 << 1;
  $105 = (($104) + -256)|0;
  $106 = Math_imul($105, $25)|0;
  $107 = ($38|0)==($3|0);
  do {
   if ($107) {
    $175 = $$promoted;$177 = $$promoted57;
   } else {
    $108 = HEAP32[$17>>2]|0;
    $109 = HEAP32[$40>>2]|0;
    $110 = ($109|0)<($22|0);
    $$ex$i8 = $110 ? $109 : $22;
    $111 = HEAP32[$43>>2]|0;
    $112 = (($$ex$i8) - ($111))|0;
    $113 = ($112|0)>(-1);
    $$$i9 = $113 ? $112 : -1;
    $114 = ((($worker)) + 160|0);
    $115 = ((($worker)) + 180|0);
    $116 = ((($worker)) + 1420|0);
    $117 = ((($worker)) + 208|0);
    $118 = ((($worker)) + 204|0);
    $119 = ((($worker)) + 200|0);
    $121 = $$promoted;$123 = $$promoted57;$127 = $$$i;$129 = $39;$131 = $101;$136 = $$$i;$139 = $39;$ey1$049 = $38;
    while(1) {
     $120 = (($121) + ($106))|0;
     $122 = (($123) + ($105))|0;
     $124 = (($ey1$049) + ($29))|0;
     $125 = (($124) - ($108))|0;
     $126 = ($$$i9|0)==($127|0);
     $128 = ($125|0)==($129|0);
     $or$cond97 = $126 & $128;
     if ($or$cond97) {
      $339 = $139;$340 = $136;$341 = $122;$342 = $120;
     } else {
      $130 = ($131|0)!=(0);
      $132 = $122 | $120;
      $133 = ($132|0)==(0);
      $or$cond = $130 | $133;
      if (!($or$cond)) {
       $134 = HEAP32[$115>>2]|0;
       $135 = ($136|0)>($134|0);
       $$$i$i11 = $135 ? $134 : $136;
       $137 = HEAP32[$116>>2]|0;
       $138 = (($137) + ($139<<2)|0);
       $140 = HEAP32[$138>>2]|0;
       $141 = ($140|0)==(0|0);
       L43: do {
        if ($141) {
         $pcell$0$lcssa$i$i14 = $138;
         label = 28;
        } else {
         $143 = $140;$pcell$03$i$i12 = $138;
         while(1) {
          $142 = HEAP32[$143>>2]|0;
          $144 = ($142|0)>($$$i$i11|0);
          if ($144) {
           $pcell$0$lcssa$i$i14 = $pcell$03$i$i12;
           label = 28;
           break L43;
          }
          $145 = ($142|0)==($$$i$i11|0);
          if ($145) {
           $cell$0$i$i16 = $143;
           break L43;
          }
          $146 = ((($143)) + 12|0);
          $147 = HEAP32[$146>>2]|0;
          $148 = ($147|0)==(0|0);
          if ($148) {
           $pcell$0$lcssa$i$i14 = $146;
           label = 28;
           break;
          } else {
           $143 = $147;$pcell$03$i$i12 = $146;
          }
         }
        }
       } while(0);
       if ((label|0) == 28) {
        label = 0;
        $149 = HEAP32[$117>>2]|0;
        $150 = HEAP32[$118>>2]|0;
        $151 = ($149|0)<($150|0);
        if (!($151)) {
         $$lcssa = $120;$$lcssa128 = $122;
         label = 29;
         break;
        }
        $152 = HEAP32[$119>>2]|0;
        $153 = (($149) + 1)|0;
        HEAP32[$117>>2] = $153;
        $154 = (($152) + ($149<<4)|0);
        HEAP32[$154>>2] = $$$i$i11;
        $155 = (((($152) + ($149<<4)|0)) + 8|0);
        HEAP32[$155>>2] = 0;
        $156 = (((($152) + ($149<<4)|0)) + 4|0);
        HEAP32[$156>>2] = 0;
        $157 = HEAP32[$pcell$0$lcssa$i$i14>>2]|0;
        $158 = (((($152) + ($149<<4)|0)) + 12|0);
        HEAP32[$158>>2] = $157;
        HEAP32[$pcell$0$lcssa$i$i14>>2] = $154;
        $cell$0$i$i16 = $154;
       }
       $159 = ((($cell$0$i$i16)) + 8|0);
       $160 = HEAP32[$159>>2]|0;
       $161 = (($160) + ($120))|0;
       HEAP32[$159>>2] = $161;
       $162 = ((($cell$0$i$i16)) + 4|0);
       $163 = HEAP32[$162>>2]|0;
       $164 = (($163) + ($122))|0;
       HEAP32[$162>>2] = $164;
      }
      HEAP32[$47>>2] = $$$i9;
      HEAP32[$114>>2] = $125;
      $339 = $125;$340 = $$$i9;$341 = 0;$342 = 0;
     }
     $165 = ($125>>>0)<($96>>>0);
     if ($165) {
      $166 = HEAP32[$115>>2]|0;
      $167 = ($$$i9|0)>=($166|0);
      $169 = $167;
     } else {
      $169 = 1;
     }
     $168 = $169&1;
     HEAP32[$103>>2] = $168;
     $170 = ($124|0)==($3|0);
     if ($170) {
      $$lcssa130 = $341;$$lcssa131 = $342;
      label = 36;
      break;
     } else {
      $121 = $342;$123 = $341;$127 = $$$i9;$129 = $125;$131 = $168;$136 = $340;$139 = $339;$ey1$049 = $124;
     }
    }
    if ((label|0) == 29) {
     HEAP32[$32>>2] = $$lcssa;
     HEAP32[$35>>2] = $$lcssa128;
     _longjmp(($worker|0),1);
     // unreachable;
    }
    else if ((label|0) == 36) {
     HEAP32[$32>>2] = $$lcssa131;
     HEAP32[$35>>2] = $$lcssa130;
     $175 = $$lcssa131;$177 = $$lcssa130;
     break;
    }
   }
  } while(0);
  $171 = (($8) + -256)|0;
  $172 = (($171) + ($28))|0;
  $173 = Math_imul($25, $172)|0;
  $174 = (($175) + ($173))|0;
  HEAP32[$32>>2] = $174;
  $176 = (($177) + ($172))|0;
  HEAP32[$35>>2] = $176;
  HEAP32[$9>>2] = $to_x;
  HEAP32[$4>>2] = $to_y;
  HEAP32[$0>>2] = $7;
  return;
 }
 $178 = (256 - ($6))|0;
 $179 = ($12|0)<(0);
 $180 = (0 - ($12))|0;
 $$4 = $179 ? $6 : $178;
 $181 = $12 >> 31;
 $182 = (($181) + 256)|0;
 $183 = $182 & -256;
 $184 = $181 | 1;
 $$7 = $179 ? $180 : $12;
 $p$0 = Math_imul($$4, $11)|0;
 $185 = (($p$0|0) / ($$7|0))&-1;
 $186 = (($p$0|0) % ($$7|0))&-1;
 $187 = ($186|0)<(0);
 $$lobit = $186 >> 31;
 $188 = $187 ? $$7 : 0;
 $delta$0 = (($185) + ($10))|0;
 $189 = (($delta$0) + ($$lobit))|0;
 _gray_render_scanline($worker,$2,$10,$6,$189,$183);
 $190 = (($184) + ($2))|0;
 $191 = $189 >> 8;
 $192 = HEAP32[$17>>2]|0;
 $193 = (($190) - ($192))|0;
 $194 = ((($worker)) + 168|0);
 $195 = HEAP32[$194>>2]|0;
 $196 = ($195|0)<($191|0);
 $$ex$i19 = $196 ? $195 : $191;
 $197 = ((($worker)) + 164|0);
 $198 = HEAP32[$197>>2]|0;
 $199 = (($$ex$i19) - ($198))|0;
 $200 = ($199|0)>(-1);
 $$$i20 = $200 ? $199 : -1;
 $201 = ((($worker)) + 156|0);
 $202 = HEAP32[$201>>2]|0;
 $203 = ($$$i20|0)==($202|0);
 if ($203) {
  $204 = ((($worker)) + 160|0);
  $205 = HEAP32[$204>>2]|0;
  $206 = ($193|0)==($205|0);
  if (!($206)) {
   label = 40;
  }
 } else {
  label = 40;
 }
 if ((label|0) == 40) {
  $207 = ((($worker)) + 196|0);
  $208 = HEAP32[$207>>2]|0;
  $209 = ($208|0)==(0);
  $210 = ((($worker)) + 188|0);
  if ($209) {
   $211 = HEAP32[$210>>2]|0;
   $212 = ((($worker)) + 192|0);
   $213 = HEAP32[$212>>2]|0;
   $214 = $213 | $211;
   $215 = ($214|0)==(0);
   if ($215) {
    $$pre$phi80Z2D = $212;
   } else {
    $216 = ((($worker)) + 180|0);
    $217 = HEAP32[$216>>2]|0;
    $218 = ($202|0)>($217|0);
    $$$i$i22 = $218 ? $217 : $202;
    $219 = ((($worker)) + 160|0);
    $220 = HEAP32[$219>>2]|0;
    $221 = ((($worker)) + 1420|0);
    $222 = HEAP32[$221>>2]|0;
    $223 = (($222) + ($220<<2)|0);
    $224 = HEAP32[$223>>2]|0;
    $225 = ($224|0)==(0|0);
    L70: do {
     if ($225) {
      $pcell$0$lcssa$i$i25 = $223;
      label = 47;
     } else {
      $227 = $224;$pcell$03$i$i23 = $223;
      while(1) {
       $226 = HEAP32[$227>>2]|0;
       $228 = ($226|0)>($$$i$i22|0);
       if ($228) {
        $pcell$0$lcssa$i$i25 = $pcell$03$i$i23;
        label = 47;
        break L70;
       }
       $229 = ($226|0)==($$$i$i22|0);
       if ($229) {
        $cell$0$i$i27 = $227;
        break L70;
       }
       $230 = ((($227)) + 12|0);
       $231 = HEAP32[$230>>2]|0;
       $232 = ($231|0)==(0|0);
       if ($232) {
        $pcell$0$lcssa$i$i25 = $230;
        label = 47;
        break;
       } else {
        $227 = $231;$pcell$03$i$i23 = $230;
       }
      }
     }
    } while(0);
    do {
     if ((label|0) == 47) {
      $233 = ((($worker)) + 208|0);
      $234 = HEAP32[$233>>2]|0;
      $235 = ((($worker)) + 204|0);
      $236 = HEAP32[$235>>2]|0;
      $237 = ($234|0)<($236|0);
      if ($237) {
       $238 = ((($worker)) + 200|0);
       $239 = HEAP32[$238>>2]|0;
       $240 = (($234) + 1)|0;
       HEAP32[$233>>2] = $240;
       $241 = (($239) + ($234<<4)|0);
       HEAP32[$241>>2] = $$$i$i22;
       $242 = (((($239) + ($234<<4)|0)) + 8|0);
       HEAP32[$242>>2] = 0;
       $243 = (((($239) + ($234<<4)|0)) + 4|0);
       HEAP32[$243>>2] = 0;
       $244 = HEAP32[$pcell$0$lcssa$i$i25>>2]|0;
       $245 = (((($239) + ($234<<4)|0)) + 12|0);
       HEAP32[$245>>2] = $244;
       HEAP32[$pcell$0$lcssa$i$i25>>2] = $241;
       $cell$0$i$i27 = $241;
       break;
      } else {
       _longjmp(($worker|0),1);
       // unreachable;
      }
     }
    } while(0);
    $246 = ((($cell$0$i$i27)) + 8|0);
    $247 = HEAP32[$246>>2]|0;
    $248 = (($247) + ($211))|0;
    HEAP32[$246>>2] = $248;
    $249 = ((($cell$0$i$i27)) + 4|0);
    $250 = HEAP32[$249>>2]|0;
    $251 = (($250) + ($213))|0;
    HEAP32[$249>>2] = $251;
    $$pre$phi80Z2D = $212;
   }
  } else {
   $$pre79 = ((($worker)) + 192|0);
   $$pre$phi80Z2D = $$pre79;
  }
  HEAP32[$210>>2] = 0;
  HEAP32[$$pre$phi80Z2D>>2] = 0;
  HEAP32[$201>>2] = $$$i20;
  $252 = ((($worker)) + 160|0);
  HEAP32[$252>>2] = $193;
 }
 $253 = ((($worker)) + 184|0);
 $254 = HEAP32[$253>>2]|0;
 $255 = ($193>>>0)<($254>>>0);
 if ($255) {
  $256 = ((($worker)) + 180|0);
  $257 = HEAP32[$256>>2]|0;
  $258 = ($$$i20|0)>=($257|0);
  $260 = $258;
 } else {
  $260 = 1;
 }
 $259 = $260&1;
 $261 = ((($worker)) + 196|0);
 HEAP32[$261>>2] = $259;
 $262 = ($190|0)==($3|0);
 L87: do {
  if ($262) {
   $x$1 = $189;
  } else {
   $263 = $11 << 8;
   $264 = (($263|0) / ($$7|0))&-1;
   $265 = (($263|0) % ($$7|0))&-1;
   $266 = ($265|0)<(0);
   $267 = $266 ? $$7 : 0;
   $rem$0 = (($267) + ($265))|0;
   $$lobit41 = $265 >> 31;
   $lift$0 = (($$lobit41) + ($264))|0;
   $mod$0 = (($186) - ($$7))|0;
   $268 = (($mod$0) + ($188))|0;
   $269 = (256 - ($183))|0;
   $270 = ((($worker)) + 160|0);
   $271 = ((($worker)) + 180|0);
   $272 = ((($worker)) + 188|0);
   $273 = ((($worker)) + 192|0);
   $274 = ((($worker)) + 1420|0);
   $275 = ((($worker)) + 208|0);
   $276 = ((($worker)) + 204|0);
   $277 = ((($worker)) + 200|0);
   $ey1$152 = $190;$mod$151 = $268;$x$050 = $189;
   while(1) {
    $278 = (($rem$0) + ($mod$151))|0;
    $279 = ($278|0)>(-1);
    $$lobit42 = $278 >>> 31;
    $$lobit42$not = $$lobit42 ^ 1;
    $280 = $279 ? $$7 : 0;
    $mod$2 = (($278) - ($280))|0;
    $delta$1 = (($lift$0) + ($x$050))|0;
    $281 = (($delta$1) + ($$lobit42$not))|0;
    _gray_render_scanline($worker,$ey1$152,$x$050,$269,$281,$183);
    $282 = (($ey1$152) + ($184))|0;
    $283 = $281 >> 8;
    $284 = HEAP32[$17>>2]|0;
    $285 = (($282) - ($284))|0;
    $286 = HEAP32[$194>>2]|0;
    $287 = ($286|0)<($283|0);
    $$ex$i30 = $287 ? $286 : $283;
    $288 = HEAP32[$197>>2]|0;
    $289 = (($$ex$i30) - ($288))|0;
    $290 = ($289|0)>(-1);
    $$$i31 = $290 ? $289 : -1;
    $291 = HEAP32[$201>>2]|0;
    $292 = ($$$i31|0)==($291|0);
    if ($292) {
     $293 = HEAP32[$270>>2]|0;
     $294 = ($285|0)==($293|0);
     if (!($294)) {
      label = 58;
     }
    } else {
     label = 58;
    }
    if ((label|0) == 58) {
     label = 0;
     $295 = HEAP32[$261>>2]|0;
     $296 = ($295|0)==(0);
     if ($296) {
      $297 = HEAP32[$272>>2]|0;
      $298 = HEAP32[$273>>2]|0;
      $299 = $298 | $297;
      $300 = ($299|0)==(0);
      if (!($300)) {
       $301 = HEAP32[$271>>2]|0;
       $302 = ($291|0)>($301|0);
       $$$i$i33 = $302 ? $301 : $291;
       $303 = HEAP32[$270>>2]|0;
       $304 = HEAP32[$274>>2]|0;
       $305 = (($304) + ($303<<2)|0);
       $306 = HEAP32[$305>>2]|0;
       $307 = ($306|0)==(0|0);
       L98: do {
        if ($307) {
         $pcell$0$lcssa$i$i36 = $305;
         label = 64;
        } else {
         $309 = $306;$pcell$03$i$i34 = $305;
         while(1) {
          $308 = HEAP32[$309>>2]|0;
          $310 = ($308|0)>($$$i$i33|0);
          if ($310) {
           $pcell$0$lcssa$i$i36 = $pcell$03$i$i34;
           label = 64;
           break L98;
          }
          $311 = ($308|0)==($$$i$i33|0);
          if ($311) {
           $cell$0$i$i38 = $309;
           break L98;
          }
          $312 = ((($309)) + 12|0);
          $313 = HEAP32[$312>>2]|0;
          $314 = ($313|0)==(0|0);
          if ($314) {
           $pcell$0$lcssa$i$i36 = $312;
           label = 64;
           break;
          } else {
           $309 = $313;$pcell$03$i$i34 = $312;
          }
         }
        }
       } while(0);
       if ((label|0) == 64) {
        label = 0;
        $315 = HEAP32[$275>>2]|0;
        $316 = HEAP32[$276>>2]|0;
        $317 = ($315|0)<($316|0);
        if (!($317)) {
         break;
        }
        $318 = HEAP32[$277>>2]|0;
        $319 = (($315) + 1)|0;
        HEAP32[$275>>2] = $319;
        $320 = (($318) + ($315<<4)|0);
        HEAP32[$320>>2] = $$$i$i33;
        $321 = (((($318) + ($315<<4)|0)) + 8|0);
        HEAP32[$321>>2] = 0;
        $322 = (((($318) + ($315<<4)|0)) + 4|0);
        HEAP32[$322>>2] = 0;
        $323 = HEAP32[$pcell$0$lcssa$i$i36>>2]|0;
        $324 = (((($318) + ($315<<4)|0)) + 12|0);
        HEAP32[$324>>2] = $323;
        HEAP32[$pcell$0$lcssa$i$i36>>2] = $320;
        $cell$0$i$i38 = $320;
       }
       $325 = ((($cell$0$i$i38)) + 8|0);
       $326 = HEAP32[$325>>2]|0;
       $327 = (($326) + ($297))|0;
       HEAP32[$325>>2] = $327;
       $328 = ((($cell$0$i$i38)) + 4|0);
       $329 = HEAP32[$328>>2]|0;
       $330 = (($329) + ($298))|0;
       HEAP32[$328>>2] = $330;
      }
     }
     HEAP32[$272>>2] = 0;
     HEAP32[$273>>2] = 0;
     HEAP32[$201>>2] = $$$i31;
     HEAP32[$270>>2] = $285;
    }
    $331 = HEAP32[$253>>2]|0;
    $332 = ($285>>>0)<($331>>>0);
    if ($332) {
     $333 = HEAP32[$271>>2]|0;
     $334 = ($$$i31|0)>=($333|0);
     $336 = $334;
    } else {
     $336 = 1;
    }
    $335 = $336&1;
    HEAP32[$261>>2] = $335;
    $337 = ($282|0)==($3|0);
    if ($337) {
     $x$1 = $281;
     break L87;
    } else {
     $ey1$152 = $282;$mod$151 = $mod$2;$x$050 = $281;
    }
   }
   _longjmp(($worker|0),1);
   // unreachable;
  }
 } while(0);
 $338 = (256 - ($183))|0;
 _gray_render_scanline($worker,$3,$x$1,$338,$to_x,$8);
 HEAP32[$9>>2] = $to_x;
 HEAP32[$4>>2] = $to_y;
 HEAP32[$0>>2] = $7;
 return;
}
function _gray_render_scanline($worker,$ey,$x1,$y1,$x2,$y2) {
 $worker = $worker|0;
 $ey = $ey|0;
 $x1 = $x1|0;
 $y1 = $y1|0;
 $x2 = $x2|0;
 $y2 = $y2|0;
 var $$ = 0, $$$i = 0, $$$i$i = 0, $$$i$i19 = 0, $$$i$i8 = 0, $$$i17 = 0, $$$i6 = 0, $$035 = 0, $$1 = 0, $$3 = 0, $$ex$i = 0, $$ex$i16 = 0, $$ex$i5 = 0, $$lcssa = 0, $$lcssa100 = 0, $$lcssa101 = 0, $$lcssa102 = 0, $$lcssa97 = 0, $$lobit = 0, $$lobit27 = 0;
 var $$lobit28 = 0, $$lobit28$not = 0, $$pre$phi57Z2D = 0, $$pre56 = 0, $$promoted = 0, $$promoted38 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0;
 var $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0;
 var $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cell$0$i$i = 0, $cell$0$i$i13 = 0, $cell$0$i$i24 = 0, $delta$0 = 0, $delta$1 = 0, $ex1$034 = 0, $lift$0 = 0;
 var $mod$0 = 0, $mod$136 = 0, $mod$2 = 0, $or$cond = 0, $or$cond70 = 0, $or$cond71 = 0, $p$0 = 0, $pcell$0$lcssa$i$i = 0, $pcell$0$lcssa$i$i11 = 0, $pcell$0$lcssa$i$i22 = 0, $pcell$03$i$i = 0, $pcell$03$i$i20 = 0, $pcell$03$i$i9 = 0, $rem$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($x2) - ($x1))|0;
 $1 = $x1 >> 8;
 $2 = $x2 >> 8;
 $3 = $1 << 8;
 $4 = (($x1) - ($3))|0;
 $5 = $2 << 8;
 $6 = (($x2) - ($5))|0;
 $7 = ($y2|0)==($y1|0);
 if ($7) {
  $8 = ((($worker)) + 172|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = (($ey) - ($9))|0;
  $11 = ((($worker)) + 168|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ($12|0)<($2|0);
  $$ex$i = $13 ? $12 : $2;
  $14 = ((($worker)) + 164|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = (($$ex$i) - ($15))|0;
  $17 = ($16|0)>(-1);
  $$$i = $17 ? $16 : -1;
  $18 = ((($worker)) + 156|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = ($$$i|0)==($19|0);
  if ($20) {
   $21 = ((($worker)) + 160|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = ($10|0)==($22|0);
   if (!($23)) {
    label = 4;
   }
  } else {
   label = 4;
  }
  if ((label|0) == 4) {
   $24 = ((($worker)) + 196|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = ($25|0)==(0);
   $27 = ((($worker)) + 188|0);
   if ($26) {
    $28 = HEAP32[$27>>2]|0;
    $29 = ((($worker)) + 192|0);
    $30 = HEAP32[$29>>2]|0;
    $31 = $30 | $28;
    $32 = ($31|0)==(0);
    if ($32) {
     $$pre$phi57Z2D = $29;
    } else {
     $33 = ((($worker)) + 180|0);
     $34 = HEAP32[$33>>2]|0;
     $35 = ($19|0)>($34|0);
     $$$i$i = $35 ? $34 : $19;
     $36 = ((($worker)) + 160|0);
     $37 = HEAP32[$36>>2]|0;
     $38 = ((($worker)) + 1420|0);
     $39 = HEAP32[$38>>2]|0;
     $40 = (($39) + ($37<<2)|0);
     $41 = HEAP32[$40>>2]|0;
     $42 = ($41|0)==(0|0);
     L10: do {
      if ($42) {
       $pcell$0$lcssa$i$i = $40;
       label = 11;
      } else {
       $44 = $41;$pcell$03$i$i = $40;
       while(1) {
        $43 = HEAP32[$44>>2]|0;
        $45 = ($43|0)>($$$i$i|0);
        if ($45) {
         $pcell$0$lcssa$i$i = $pcell$03$i$i;
         label = 11;
         break L10;
        }
        $46 = ($43|0)==($$$i$i|0);
        if ($46) {
         $cell$0$i$i = $44;
         break L10;
        }
        $47 = ((($44)) + 12|0);
        $48 = HEAP32[$47>>2]|0;
        $49 = ($48|0)==(0|0);
        if ($49) {
         $pcell$0$lcssa$i$i = $47;
         label = 11;
         break;
        } else {
         $44 = $48;$pcell$03$i$i = $47;
        }
       }
      }
     } while(0);
     do {
      if ((label|0) == 11) {
       $50 = ((($worker)) + 208|0);
       $51 = HEAP32[$50>>2]|0;
       $52 = ((($worker)) + 204|0);
       $53 = HEAP32[$52>>2]|0;
       $54 = ($51|0)<($53|0);
       if ($54) {
        $55 = ((($worker)) + 200|0);
        $56 = HEAP32[$55>>2]|0;
        $57 = (($51) + 1)|0;
        HEAP32[$50>>2] = $57;
        $58 = (($56) + ($51<<4)|0);
        HEAP32[$58>>2] = $$$i$i;
        $59 = (((($56) + ($51<<4)|0)) + 8|0);
        HEAP32[$59>>2] = 0;
        $60 = (((($56) + ($51<<4)|0)) + 4|0);
        HEAP32[$60>>2] = 0;
        $61 = HEAP32[$pcell$0$lcssa$i$i>>2]|0;
        $62 = (((($56) + ($51<<4)|0)) + 12|0);
        HEAP32[$62>>2] = $61;
        HEAP32[$pcell$0$lcssa$i$i>>2] = $58;
        $cell$0$i$i = $58;
        break;
       } else {
        _longjmp(($worker|0),1);
        // unreachable;
       }
      }
     } while(0);
     $63 = ((($cell$0$i$i)) + 8|0);
     $64 = HEAP32[$63>>2]|0;
     $65 = (($64) + ($28))|0;
     HEAP32[$63>>2] = $65;
     $66 = ((($cell$0$i$i)) + 4|0);
     $67 = HEAP32[$66>>2]|0;
     $68 = (($67) + ($30))|0;
     HEAP32[$66>>2] = $68;
     $$pre$phi57Z2D = $29;
    }
   } else {
    $$pre56 = ((($worker)) + 192|0);
    $$pre$phi57Z2D = $$pre56;
   }
   HEAP32[$27>>2] = 0;
   HEAP32[$$pre$phi57Z2D>>2] = 0;
   HEAP32[$18>>2] = $$$i;
   $69 = ((($worker)) + 160|0);
   HEAP32[$69>>2] = $10;
  }
  $70 = ((($worker)) + 184|0);
  $71 = HEAP32[$70>>2]|0;
  $72 = ($10>>>0)<($71>>>0);
  if ($72) {
   $73 = ((($worker)) + 180|0);
   $74 = HEAP32[$73>>2]|0;
   $75 = ($$$i|0)>=($74|0);
   $77 = $75;
  } else {
   $77 = 1;
  }
  $76 = $77&1;
  $78 = ((($worker)) + 196|0);
  HEAP32[$78>>2] = $76;
  return;
 }
 $79 = ($1|0)==($2|0);
 if ($79) {
  $80 = (($y2) - ($y1))|0;
  $81 = (($6) + ($4))|0;
  $82 = Math_imul($81, $80)|0;
  $83 = ((($worker)) + 188|0);
  $84 = HEAP32[$83>>2]|0;
  $85 = (($84) + ($82))|0;
  HEAP32[$83>>2] = $85;
  $86 = ((($worker)) + 192|0);
  $87 = HEAP32[$86>>2]|0;
  $88 = (($87) + ($80))|0;
  HEAP32[$86>>2] = $88;
  return;
 }
 $89 = (256 - ($4))|0;
 $90 = (($y2) - ($y1))|0;
 $91 = ($0|0)<(0);
 $92 = (0 - ($0))|0;
 $$ = $91 ? $4 : $89;
 $93 = $0 >> 31;
 $94 = (($93) + 256)|0;
 $95 = $94 & -256;
 $$3 = $91 ? $92 : $0;
 $96 = $93 | 1;
 $p$0 = Math_imul($$, $90)|0;
 $97 = (($p$0|0) / ($$3|0))&-1;
 $98 = (($p$0|0) % ($$3|0))&-1;
 $99 = ($98|0)<(0);
 $$lobit = $98 >> 31;
 $delta$0 = (($$lobit) + ($97))|0;
 $100 = $99 ? $$3 : 0;
 $101 = (($95) + ($4))|0;
 $102 = Math_imul($delta$0, $101)|0;
 $103 = ((($worker)) + 188|0);
 $104 = HEAP32[$103>>2]|0;
 $105 = (($102) + ($104))|0;
 HEAP32[$103>>2] = $105;
 $106 = ((($worker)) + 192|0);
 $107 = HEAP32[$106>>2]|0;
 $108 = (($107) + ($delta$0))|0;
 HEAP32[$106>>2] = $108;
 $109 = (($96) + ($1))|0;
 $110 = ((($worker)) + 172|0);
 $111 = HEAP32[$110>>2]|0;
 $112 = (($ey) - ($111))|0;
 $113 = ((($worker)) + 168|0);
 $114 = HEAP32[$113>>2]|0;
 $115 = ($114|0)<($109|0);
 $$ex$i5 = $115 ? $114 : $109;
 $116 = ((($worker)) + 164|0);
 $117 = HEAP32[$116>>2]|0;
 $118 = (($$ex$i5) - ($117))|0;
 $119 = ($118|0)>(-1);
 $$$i6 = $119 ? $118 : -1;
 $120 = ((($worker)) + 156|0);
 $121 = HEAP32[$120>>2]|0;
 $122 = ($$$i6|0)==($121|0);
 if ($122) {
  $123 = ((($worker)) + 160|0);
  $124 = HEAP32[$123>>2]|0;
  $125 = ($112|0)==($124|0);
  if ($125) {
   $$promoted = $105;$$promoted38 = $108;
  } else {
   label = 23;
  }
 } else {
  label = 23;
 }
 if ((label|0) == 23) {
  $126 = ((($worker)) + 196|0);
  $127 = HEAP32[$126>>2]|0;
  $128 = ($127|0)!=(0);
  $129 = $108 | $105;
  $130 = ($129|0)==(0);
  $or$cond70 = $128 | $130;
  if (!($or$cond70)) {
   $131 = ((($worker)) + 180|0);
   $132 = HEAP32[$131>>2]|0;
   $133 = ($121|0)>($132|0);
   $$$i$i8 = $133 ? $132 : $121;
   $134 = ((($worker)) + 160|0);
   $135 = HEAP32[$134>>2]|0;
   $136 = ((($worker)) + 1420|0);
   $137 = HEAP32[$136>>2]|0;
   $138 = (($137) + ($135<<2)|0);
   $139 = HEAP32[$138>>2]|0;
   $140 = ($139|0)==(0|0);
   L39: do {
    if ($140) {
     $pcell$0$lcssa$i$i11 = $138;
     label = 28;
    } else {
     $142 = $139;$pcell$03$i$i9 = $138;
     while(1) {
      $141 = HEAP32[$142>>2]|0;
      $143 = ($141|0)>($$$i$i8|0);
      if ($143) {
       $pcell$0$lcssa$i$i11 = $pcell$03$i$i9;
       label = 28;
       break L39;
      }
      $144 = ($141|0)==($$$i$i8|0);
      if ($144) {
       $cell$0$i$i13 = $142;
       break L39;
      }
      $145 = ((($142)) + 12|0);
      $146 = HEAP32[$145>>2]|0;
      $147 = ($146|0)==(0|0);
      if ($147) {
       $pcell$0$lcssa$i$i11 = $145;
       label = 28;
       break;
      } else {
       $142 = $146;$pcell$03$i$i9 = $145;
      }
     }
    }
   } while(0);
   do {
    if ((label|0) == 28) {
     $148 = ((($worker)) + 208|0);
     $149 = HEAP32[$148>>2]|0;
     $150 = ((($worker)) + 204|0);
     $151 = HEAP32[$150>>2]|0;
     $152 = ($149|0)<($151|0);
     if ($152) {
      $153 = ((($worker)) + 200|0);
      $154 = HEAP32[$153>>2]|0;
      $155 = (($149) + 1)|0;
      HEAP32[$148>>2] = $155;
      $156 = (($154) + ($149<<4)|0);
      HEAP32[$156>>2] = $$$i$i8;
      $157 = (((($154) + ($149<<4)|0)) + 8|0);
      HEAP32[$157>>2] = 0;
      $158 = (((($154) + ($149<<4)|0)) + 4|0);
      HEAP32[$158>>2] = 0;
      $159 = HEAP32[$pcell$0$lcssa$i$i11>>2]|0;
      $160 = (((($154) + ($149<<4)|0)) + 12|0);
      HEAP32[$160>>2] = $159;
      HEAP32[$pcell$0$lcssa$i$i11>>2] = $156;
      $cell$0$i$i13 = $156;
      break;
     } else {
      _longjmp(($worker|0),1);
      // unreachable;
     }
    }
   } while(0);
   $161 = ((($cell$0$i$i13)) + 8|0);
   $162 = HEAP32[$161>>2]|0;
   $163 = (($162) + ($105))|0;
   HEAP32[$161>>2] = $163;
   $164 = ((($cell$0$i$i13)) + 4|0);
   $165 = HEAP32[$164>>2]|0;
   $166 = (($165) + ($108))|0;
   HEAP32[$164>>2] = $166;
  }
  HEAP32[$103>>2] = 0;
  HEAP32[$106>>2] = 0;
  HEAP32[$120>>2] = $$$i6;
  $167 = ((($worker)) + 160|0);
  HEAP32[$167>>2] = $112;
  $$promoted = 0;$$promoted38 = 0;
 }
 $168 = ((($worker)) + 184|0);
 $169 = HEAP32[$168>>2]|0;
 $170 = ($112>>>0)<($169>>>0);
 if ($170) {
  $171 = ((($worker)) + 180|0);
  $172 = HEAP32[$171>>2]|0;
  $173 = ($$$i6|0)>=($172|0);
  $175 = $173;
 } else {
  $175 = 1;
 }
 $174 = $175&1;
 $176 = ((($worker)) + 196|0);
 HEAP32[$176>>2] = $174;
 $177 = (($delta$0) + ($y1))|0;
 $178 = ($109|0)==($2|0);
 do {
  if ($178) {
   $$1 = $177;$260 = $$promoted;$262 = $$promoted38;
  } else {
   $179 = (($delta$0) + ($y2))|0;
   $180 = (($179) - ($177))|0;
   $181 = $180 << 8;
   $182 = (($181|0) / ($$3|0))&-1;
   $183 = (($181|0) % ($$3|0))&-1;
   $184 = ($183|0)<(0);
   $$lobit27 = $183 >> 31;
   $lift$0 = (($$lobit27) + ($182))|0;
   $185 = $184 ? $$3 : 0;
   $rem$0 = (($185) + ($183))|0;
   $mod$0 = (($98) - ($$3))|0;
   $186 = (($mod$0) + ($100))|0;
   $187 = HEAP32[$110>>2]|0;
   $188 = (($ey) - ($187))|0;
   $189 = HEAP32[$113>>2]|0;
   $190 = HEAP32[$116>>2]|0;
   $191 = ((($worker)) + 160|0);
   $192 = ((($worker)) + 180|0);
   $193 = ((($worker)) + 1420|0);
   $194 = ((($worker)) + 208|0);
   $195 = ((($worker)) + 204|0);
   $196 = ((($worker)) + 200|0);
   $197 = ($188>>>0)<($169>>>0);
   $$035 = $177;$203 = $$promoted;$205 = $$promoted38;$212 = $$$i6;$214 = $112;$216 = $174;$221 = $$$i6;$224 = $112;$ex1$034 = $109;$mod$136 = $186;
   while(1) {
    $198 = (($rem$0) + ($mod$136))|0;
    $199 = ($198|0)>(-1);
    $$lobit28 = $198 >>> 31;
    $$lobit28$not = $$lobit28 ^ 1;
    $delta$1 = (($lift$0) + ($$lobit28$not))|0;
    $200 = $199 ? $$3 : 0;
    $mod$2 = (($198) - ($200))|0;
    $201 = $delta$1 << 8;
    $202 = (($203) + ($201))|0;
    $204 = (($205) + ($delta$1))|0;
    $206 = (($delta$1) + ($$035))|0;
    $207 = (($ex1$034) + ($96))|0;
    $208 = ($189|0)<($207|0);
    $$ex$i16 = $208 ? $189 : $207;
    $209 = (($$ex$i16) - ($190))|0;
    $210 = ($209|0)>(-1);
    $$$i17 = $210 ? $209 : -1;
    $211 = ($$$i17|0)==($212|0);
    $213 = ($188|0)==($214|0);
    $or$cond71 = $211 & $213;
    if ($or$cond71) {
     $263 = $224;$264 = $221;$265 = $204;$266 = $202;
    } else {
     $215 = ($216|0)!=(0);
     $217 = $204 | $202;
     $218 = ($217|0)==(0);
     $or$cond = $215 | $218;
     if (!($or$cond)) {
      $219 = HEAP32[$192>>2]|0;
      $220 = ($221|0)>($219|0);
      $$$i$i19 = $220 ? $219 : $221;
      $222 = HEAP32[$193>>2]|0;
      $223 = (($222) + ($224<<2)|0);
      $225 = HEAP32[$223>>2]|0;
      $226 = ($225|0)==(0|0);
      L63: do {
       if ($226) {
        $pcell$0$lcssa$i$i22 = $223;
        label = 43;
       } else {
        $228 = $225;$pcell$03$i$i20 = $223;
        while(1) {
         $227 = HEAP32[$228>>2]|0;
         $229 = ($227|0)>($$$i$i19|0);
         if ($229) {
          $pcell$0$lcssa$i$i22 = $pcell$03$i$i20;
          label = 43;
          break L63;
         }
         $230 = ($227|0)==($$$i$i19|0);
         if ($230) {
          $cell$0$i$i24 = $228;
          break L63;
         }
         $231 = ((($228)) + 12|0);
         $232 = HEAP32[$231>>2]|0;
         $233 = ($232|0)==(0|0);
         if ($233) {
          $pcell$0$lcssa$i$i22 = $231;
          label = 43;
          break;
         } else {
          $228 = $232;$pcell$03$i$i20 = $231;
         }
        }
       }
      } while(0);
      if ((label|0) == 43) {
       label = 0;
       $234 = HEAP32[$194>>2]|0;
       $235 = HEAP32[$195>>2]|0;
       $236 = ($234|0)<($235|0);
       if (!($236)) {
        $$lcssa = $202;$$lcssa97 = $204;
        label = 44;
        break;
       }
       $237 = HEAP32[$196>>2]|0;
       $238 = (($234) + 1)|0;
       HEAP32[$194>>2] = $238;
       $239 = (($237) + ($234<<4)|0);
       HEAP32[$239>>2] = $$$i$i19;
       $240 = (((($237) + ($234<<4)|0)) + 8|0);
       HEAP32[$240>>2] = 0;
       $241 = (((($237) + ($234<<4)|0)) + 4|0);
       HEAP32[$241>>2] = 0;
       $242 = HEAP32[$pcell$0$lcssa$i$i22>>2]|0;
       $243 = (((($237) + ($234<<4)|0)) + 12|0);
       HEAP32[$243>>2] = $242;
       HEAP32[$pcell$0$lcssa$i$i22>>2] = $239;
       $cell$0$i$i24 = $239;
      }
      $244 = ((($cell$0$i$i24)) + 8|0);
      $245 = HEAP32[$244>>2]|0;
      $246 = (($245) + ($202))|0;
      HEAP32[$244>>2] = $246;
      $247 = ((($cell$0$i$i24)) + 4|0);
      $248 = HEAP32[$247>>2]|0;
      $249 = (($248) + ($204))|0;
      HEAP32[$247>>2] = $249;
     }
     HEAP32[$120>>2] = $$$i17;
     HEAP32[$191>>2] = $188;
     $263 = $188;$264 = $$$i17;$265 = 0;$266 = 0;
    }
    if ($197) {
     $250 = HEAP32[$192>>2]|0;
     $251 = ($$$i17|0)>=($250|0);
     $253 = $251;
    } else {
     $253 = 1;
    }
    $252 = $253&1;
    HEAP32[$176>>2] = $252;
    $254 = ($207|0)==($2|0);
    if ($254) {
     $$lcssa100 = $206;$$lcssa101 = $265;$$lcssa102 = $266;
     label = 51;
     break;
    } else {
     $$035 = $206;$203 = $266;$205 = $265;$212 = $$$i17;$214 = $188;$216 = $252;$221 = $264;$224 = $263;$ex1$034 = $207;$mod$136 = $mod$2;
    }
   }
   if ((label|0) == 44) {
    HEAP32[$103>>2] = $$lcssa;
    HEAP32[$106>>2] = $$lcssa97;
    _longjmp(($worker|0),1);
    // unreachable;
   }
   else if ((label|0) == 51) {
    HEAP32[$103>>2] = $$lcssa102;
    HEAP32[$106>>2] = $$lcssa101;
    $$1 = $$lcssa100;$260 = $$lcssa102;$262 = $$lcssa101;
    break;
   }
  }
 } while(0);
 $255 = (($y2) - ($$1))|0;
 $256 = (($6) + 256)|0;
 $257 = (($256) - ($95))|0;
 $258 = Math_imul($255, $257)|0;
 $259 = (($258) + ($260))|0;
 HEAP32[$103>>2] = $259;
 $261 = (($262) + ($255))|0;
 HEAP32[$106>>2] = $261;
 return;
}
function _gray_conic_to($control,$to,$worker) {
 $control = $control|0;
 $to = $to|0;
 $worker = $worker|0;
 var $$$1$i = 0, $$$i = 0, $$1$i = 0, $$lcssa = 0, $$phi$trans$insert3$i = 0, $$pre$i = 0, $$pre4$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $arc$0$i = 0, $arc$1$i = 0, $arc$2$i = 0, $dx$1$i = 0, $level$0$i = 0, $max$0$i = 0;
 var $max$1$i = 0, $to$idx$val = 0, $to$idx1 = 0, $to$idx1$val = 0, $top$0$i = 0, $top$1$i = 0, $top$2$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $to$idx$val = HEAP32[$to>>2]|0;
 $to$idx1 = ((($to)) + 4|0);
 $to$idx1$val = HEAP32[$to$idx1>>2]|0;
 $0 = ((($worker)) + 1008|0);
 $1 = ((($worker)) + 232|0);
 $2 = $to$idx$val << 2;
 HEAP32[$1>>2] = $2;
 $3 = $to$idx1$val << 2;
 $4 = ((($worker)) + 236|0);
 HEAP32[$4>>2] = $3;
 $5 = HEAP32[$control>>2]|0;
 $6 = $5 << 2;
 $7 = ((($worker)) + 240|0);
 HEAP32[$7>>2] = $6;
 $8 = ((($control)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 2;
 $11 = ((($worker)) + 244|0);
 HEAP32[$11>>2] = $10;
 $12 = ((($worker)) + 220|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($worker)) + 248|0);
 HEAP32[$14>>2] = $13;
 $15 = ((($worker)) + 224|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($worker)) + 252|0);
 HEAP32[$17>>2] = $16;
 $18 = $5 << 3;
 $19 = (($2) - ($18))|0;
 $20 = (($19) + ($13))|0;
 $21 = ($20|0)<(0);
 $22 = (0 - ($20))|0;
 $23 = $21 ? $22 : $20;
 $24 = $9 << 3;
 $25 = (($3) - ($24))|0;
 $26 = (($25) + ($16))|0;
 $27 = ($26|0)<(0);
 $28 = (0 - ($26))|0;
 $29 = $27 ? $28 : $26;
 $30 = ($23|0)<($29|0);
 $$$i = $30 ? $29 : $23;
 $31 = ($$$i|0)<(64);
 if ($31) {
  $83 = $2;$arc$1$i = $1;$top$1$i = 0;
  label = 9;
 } else {
  $32 = ($10|0)<($3|0);
  $$1$i = $32 ? $10 : $3;
  $33 = ($10|0)>($3|0);
  $max$0$i = $33 ? $10 : $3;
  $34 = ($16|0)<($$1$i|0);
  $$$1$i = $34 ? $16 : $$1$i;
  $35 = $$$1$i >> 8;
  $36 = ((($worker)) + 176|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = ($35|0)<($37|0);
  if ($38) {
   $39 = ($16|0)>($max$0$i|0);
   $max$1$i = $39 ? $16 : $max$0$i;
   $40 = $max$1$i >> 8;
   $41 = ((($worker)) + 172|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = ($40|0)<($42|0);
   if ($43) {
    $83 = $2;$arc$1$i = $1;$top$1$i = 0;
    label = 9;
   } else {
    $dx$1$i = $$$i;$level$0$i = 0;
    while(1) {
     $44 = $dx$1$i >> 2;
     $45 = (($level$0$i) + 1)|0;
     $46 = ($44|0)>(64);
     if ($46) {
      $dx$1$i = $44;$level$0$i = $45;
     } else {
      $$lcssa = $45;
      break;
     }
    }
    HEAP32[$0>>2] = $$lcssa;
    $48 = $$lcssa;$arc$0$i = $1;$top$0$i = 0;
    label = 6;
   }
  } else {
   $83 = $2;$arc$1$i = $1;$top$1$i = 0;
   label = 9;
  }
 }
 L7: while(1) {
  do {
   if ((label|0) == 6) {
    label = 0;
    $47 = ($48|0)>(0);
    if ($47) {
     $49 = (((($worker)) + 1008|0) + ($top$0$i<<2)|0);
     $50 = ((($arc$0$i)) + 16|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = ((($arc$0$i)) + 32|0);
     HEAP32[$52>>2] = $51;
     $53 = ((($arc$0$i)) + 8|0);
     $54 = HEAP32[$53>>2]|0;
     $55 = (($54) + ($51))|0;
     $56 = (($55|0) / 2)&-1;
     $57 = ((($arc$0$i)) + 24|0);
     HEAP32[$57>>2] = $56;
     $58 = HEAP32[$arc$0$i>>2]|0;
     $59 = (($58) + ($54))|0;
     $60 = (($59|0) / 2)&-1;
     HEAP32[$53>>2] = $60;
     $61 = (($60) + ($56))|0;
     $62 = (($61|0) / 2)&-1;
     HEAP32[$50>>2] = $62;
     $63 = ((($arc$0$i)) + 20|0);
     $64 = HEAP32[$63>>2]|0;
     $65 = ((($arc$0$i)) + 36|0);
     HEAP32[$65>>2] = $64;
     $66 = ((($arc$0$i)) + 12|0);
     $67 = HEAP32[$66>>2]|0;
     $68 = (($67) + ($64))|0;
     $69 = (($68|0) / 2)&-1;
     $70 = ((($arc$0$i)) + 28|0);
     HEAP32[$70>>2] = $69;
     $71 = ((($arc$0$i)) + 4|0);
     $72 = HEAP32[$71>>2]|0;
     $73 = (($72) + ($67))|0;
     $74 = (($73|0) / 2)&-1;
     HEAP32[$66>>2] = $74;
     $75 = (($74) + ($69))|0;
     $76 = (($75|0) / 2)&-1;
     HEAP32[$63>>2] = $76;
     $77 = ((($arc$0$i)) + 16|0);
     $78 = (($top$0$i) + 1)|0;
     $79 = (($48) + -1)|0;
     HEAP32[$49>>2] = $79;
     $80 = (((($worker)) + 1008|0) + ($78<<2)|0);
     HEAP32[$80>>2] = $79;
     $arc$2$i = $77;$top$2$i = $78;
     break;
    } else {
     $$pre$i = HEAP32[$arc$0$i>>2]|0;
     $83 = $$pre$i;$arc$1$i = $arc$0$i;$top$1$i = $top$0$i;
     label = 9;
     continue L7;
    }
   }
   else if ((label|0) == 9) {
    label = 0;
    $81 = ((($arc$1$i)) + 4|0);
    $82 = HEAP32[$81>>2]|0;
    _gray_render_line($worker,$83,$82);
    $84 = (($top$1$i) + -1)|0;
    $85 = ((($arc$1$i)) + -16|0);
    $arc$2$i = $85;$top$2$i = $84;
   }
  } while(0);
  $86 = ($top$2$i|0)>(-1);
  if (!($86)) {
   break;
  }
  $$phi$trans$insert3$i = (((($worker)) + 1008|0) + ($top$2$i<<2)|0);
  $$pre4$i = HEAP32[$$phi$trans$insert3$i>>2]|0;
  $48 = $$pre4$i;$arc$0$i = $arc$2$i;$top$0$i = $top$2$i;
  label = 6;
 }
 return 0;
}
function _gray_cubic_to($control1,$control2,$to,$worker) {
 $control1 = $control1|0;
 $control2 = $control2|0;
 $to = $to|0;
 $worker = $worker|0;
 var $$$$$i = 0, $$$$i = 0, $$$i = 0, $$phi$trans$insert3$i = 0, $$phi$trans$insert5$i = 0, $$pre$i = 0, $$pre$phi$iZ2D = 0, $$pre$phi7$iZ2D = 0, $$pre2$pre$i = 0, $$pre4$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $arc$0$i = 0, $arc$1$i = 0, $max$0$i = 0, $max$1$i = 0, $max$2$i = 0, $to$idx$val = 0, $to$idx1 = 0;
 var $to$idx1$val = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $to$idx$val = HEAP32[$to>>2]|0;
 $to$idx1 = ((($to)) + 4|0);
 $to$idx1$val = HEAP32[$to$idx1>>2]|0;
 $0 = ((($worker)) + 232|0);
 $1 = $to$idx$val << 2;
 HEAP32[$0>>2] = $1;
 $2 = $to$idx1$val << 2;
 $3 = ((($worker)) + 236|0);
 HEAP32[$3>>2] = $2;
 $4 = HEAP32[$control2>>2]|0;
 $5 = $4 << 2;
 $6 = ((($worker)) + 240|0);
 HEAP32[$6>>2] = $5;
 $7 = ((($control2)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $8 << 2;
 $10 = ((($worker)) + 244|0);
 HEAP32[$10>>2] = $9;
 $11 = HEAP32[$control1>>2]|0;
 $12 = $11 << 2;
 $13 = ((($worker)) + 248|0);
 HEAP32[$13>>2] = $12;
 $14 = ((($control1)) + 4|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = $15 << 2;
 $17 = ((($worker)) + 252|0);
 HEAP32[$17>>2] = $16;
 $18 = ((($worker)) + 220|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ((($worker)) + 256|0);
 HEAP32[$20>>2] = $19;
 $21 = ((($worker)) + 224|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ((($worker)) + 260|0);
 HEAP32[$23>>2] = $22;
 $24 = ($9|0)<($2|0);
 $$$i = $24 ? $9 : $2;
 $25 = ($9|0)>($2|0);
 $max$0$i = $25 ? $9 : $2;
 $26 = ($16|0)<($$$i|0);
 $$$$i = $26 ? $16 : $$$i;
 $27 = ($16|0)>($max$0$i|0);
 $max$1$i = $27 ? $16 : $max$0$i;
 $28 = ($22|0)<($$$$i|0);
 $$$$$i = $28 ? $22 : $$$$i;
 $29 = $$$$$i >> 8;
 $30 = ((($worker)) + 176|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = ($29|0)<($31|0);
 if ($32) {
  $33 = ($22|0)>($max$1$i|0);
  $max$2$i = $33 ? $22 : $max$1$i;
  $34 = $max$2$i >> 8;
  $35 = ((($worker)) + 172|0);
  $36 = HEAP32[$35>>2]|0;
  $37 = ($34|0)<($36|0);
  if ($37) {
   $140 = $1;$arc$1$i = $0;
   label = 13;
  } else {
   $41 = $1;$arc$0$i = $0;
  }
 } else {
  $140 = $1;$arc$1$i = $0;
  label = 13;
 }
 while(1) {
  if ((label|0) == 13) {
   label = 0;
   $138 = ((($arc$1$i)) + 4|0);
   $139 = HEAP32[$138>>2]|0;
   _gray_render_line($worker,$140,$139);
   $141 = ($arc$1$i|0)==($0|0);
   if ($141) {
    break;
   }
   $142 = ((($arc$1$i)) + -24|0);
   $$pre2$pre$i = HEAP32[$142>>2]|0;
   $41 = $$pre2$pre$i;$arc$0$i = $142;
  }
  $38 = ((($arc$0$i)) + 24|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = (($39) - ($41))|0;
  $42 = ((($arc$0$i)) + 28|0);
  $43 = HEAP32[$42>>2]|0;
  $44 = ((($arc$0$i)) + 4|0);
  $45 = HEAP32[$44>>2]|0;
  $46 = (($43) - ($45))|0;
  $47 = ($40|0)<(0);
  $48 = (0 - ($40))|0;
  $49 = $47 ? $48 : $40;
  $50 = ($46|0)<(0);
  $51 = (0 - ($46))|0;
  $52 = $50 ? $51 : $46;
  $53 = ($49|0)>($52|0);
  if ($53) {
   $54 = ($52*3)|0;
   $55 = $54 >> 3;
   $56 = (($55) + ($49))|0;
   $61 = $56;
  } else {
   $57 = ($49*3)|0;
   $58 = $57 >> 3;
   $59 = (($52) + ($58))|0;
   $61 = $59;
  }
  $60 = ($61|0)>(32767);
  if ($60) {
   $$phi$trans$insert3$i = ((($arc$0$i)) + 8|0);
   $$pre4$i = HEAP32[$$phi$trans$insert3$i>>2]|0;
   $$phi$trans$insert5$i = ((($arc$0$i)) + 12|0);
   $$pre$i = HEAP32[$$phi$trans$insert5$i>>2]|0;
   $$pre$phi$iZ2D = $$phi$trans$insert3$i;$$pre$phi7$iZ2D = $$phi$trans$insert5$i;$106 = $$pre4$i;$124 = $$pre$i;
  } else {
   $62 = ($61*42)|0;
   $63 = ((($arc$0$i)) + 8|0);
   $64 = HEAP32[$63>>2]|0;
   $65 = (($64) - ($41))|0;
   $66 = ((($arc$0$i)) + 12|0);
   $67 = HEAP32[$66>>2]|0;
   $68 = (($67) - ($45))|0;
   $69 = Math_imul($65, $46)|0;
   $70 = Math_imul($68, $40)|0;
   $71 = (($69) - ($70))|0;
   $72 = ($71|0)<(0);
   $73 = (0 - ($71))|0;
   $74 = $72 ? $73 : $71;
   $75 = ($74|0)>($62|0);
   if ($75) {
    $$pre$phi$iZ2D = $63;$$pre$phi7$iZ2D = $66;$106 = $64;$124 = $67;
   } else {
    $76 = ((($arc$0$i)) + 16|0);
    $77 = HEAP32[$76>>2]|0;
    $78 = (($77) - ($41))|0;
    $79 = ((($arc$0$i)) + 20|0);
    $80 = HEAP32[$79>>2]|0;
    $81 = (($80) - ($45))|0;
    $82 = Math_imul($78, $46)|0;
    $83 = Math_imul($81, $40)|0;
    $84 = (($82) - ($83))|0;
    $85 = ($84|0)<(0);
    $86 = (0 - ($84))|0;
    $87 = $85 ? $86 : $84;
    $88 = ($87|0)>($62|0);
    if ($88) {
     $$pre$phi$iZ2D = $63;$$pre$phi7$iZ2D = $66;$106 = $64;$124 = $67;
    } else {
     $89 = (($65) - ($40))|0;
     $90 = Math_imul($89, $65)|0;
     $91 = (($68) - ($46))|0;
     $92 = Math_imul($91, $68)|0;
     $93 = (($92) + ($90))|0;
     $94 = ($93|0)>(0);
     if ($94) {
      $$pre$phi$iZ2D = $63;$$pre$phi7$iZ2D = $66;$106 = $64;$124 = $67;
     } else {
      $95 = (($78) - ($40))|0;
      $96 = Math_imul($95, $78)|0;
      $97 = (($81) - ($46))|0;
      $98 = Math_imul($97, $81)|0;
      $99 = (($98) + ($96))|0;
      $100 = ($99|0)>(0);
      if ($100) {
       $$pre$phi$iZ2D = $63;$$pre$phi7$iZ2D = $66;$106 = $64;$124 = $67;
      } else {
       $140 = $41;$arc$1$i = $arc$0$i;
       label = 13;
       continue;
      }
     }
    }
   }
  }
  $101 = ((($arc$0$i)) + 24|0);
  $102 = ((($arc$0$i)) + 48|0);
  HEAP32[$102>>2] = $39;
  $103 = ((($arc$0$i)) + 16|0);
  $104 = HEAP32[$103>>2]|0;
  $105 = (($106) + ($41))|0;
  $107 = (($105|0) / 2)&-1;
  HEAP32[$$pre$phi$iZ2D>>2] = $107;
  $108 = (($104) + ($39))|0;
  $109 = (($108|0) / 2)&-1;
  $110 = ((($arc$0$i)) + 40|0);
  HEAP32[$110>>2] = $109;
  $111 = (($104) + ($106))|0;
  $112 = (($111|0) / 2)&-1;
  $113 = (($112) + ($107))|0;
  $114 = (($113|0) / 2)&-1;
  HEAP32[$103>>2] = $114;
  $115 = (($109) + ($112))|0;
  $116 = (($115|0) / 2)&-1;
  $117 = ((($arc$0$i)) + 32|0);
  HEAP32[$117>>2] = $116;
  $118 = (($114) + ($116))|0;
  $119 = (($118|0) / 2)&-1;
  HEAP32[$101>>2] = $119;
  $120 = ((($arc$0$i)) + 52|0);
  HEAP32[$120>>2] = $43;
  $121 = ((($arc$0$i)) + 20|0);
  $122 = HEAP32[$121>>2]|0;
  $123 = (($124) + ($45))|0;
  $125 = (($123|0) / 2)&-1;
  HEAP32[$$pre$phi7$iZ2D>>2] = $125;
  $126 = (($122) + ($43))|0;
  $127 = (($126|0) / 2)&-1;
  $128 = ((($arc$0$i)) + 44|0);
  HEAP32[$128>>2] = $127;
  $129 = (($122) + ($124))|0;
  $130 = (($129|0) / 2)&-1;
  $131 = (($130) + ($125))|0;
  $132 = (($131|0) / 2)&-1;
  HEAP32[$121>>2] = $132;
  $133 = (($127) + ($130))|0;
  $134 = (($133|0) / 2)&-1;
  $135 = ((($arc$0$i)) + 36|0);
  HEAP32[$135>>2] = $134;
  $136 = (($132) + ($134))|0;
  $137 = (($136|0) / 2)&-1;
  HEAP32[$42>>2] = $137;
  $41 = $119;$arc$0$i = $38;
 }
 return 0;
}
function _gray_hline($worker,$x,$y,$area,$acount) {
 $worker = $worker|0;
 $x = $x|0;
 $y = $y|0;
 $area = $area|0;
 $acount = $acount|0;
 var $$ = 0, $$$5 = 0, $$12 = 0, $$5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $coverage$1 = 0, $or$cond = 0, $or$cond13 = 0, $or$cond3 = 0, $span$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $area >> 9;
 $1 = ($0|0)<(0);
 $2 = (0 - ($0))|0;
 $$5 = $1 ? $2 : $0;
 $3 = ((($worker)) + 1152|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 & 2;
 $6 = ($5|0)==(0);
 do {
  if ($6) {
   $11 = ($$5|0)>(255);
   $$$5 = $11 ? 255 : $$5;
   $coverage$1 = $$$5;
  } else {
   $7 = $$5 & 511;
   $8 = ($7>>>0)>(256);
   if ($8) {
    $9 = (512 - ($7))|0;
    $coverage$1 = $9;
    break;
   } else {
    $10 = ($7|0)==(256);
    $$12 = $10 ? 255 : $7;
    $coverage$1 = $$12;
    break;
   }
  }
 } while(0);
 $12 = ((($worker)) + 172|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($13) + ($y))|0;
 $15 = ((($worker)) + 164|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = (($16) + ($x))|0;
 $18 = ($17|0)<(32767);
 $$ = $18 ? $17 : 32767;
 $19 = ($coverage$1|0)==(0);
 if ($19) {
  return;
 }
 $20 = ((($worker)) + 1388|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ((($worker)) + 1196|0);
 $23 = (((($worker)) + 1196|0) + (($21*6)|0)|0);
 $24 = ($21|0)>(0);
 $25 = ((($worker)) + 1400|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ($26|0)==($14|0);
 $or$cond13 = $24 & $27;
 if ($or$cond13) {
  $28 = ((($23)) + -6|0);
  $29 = HEAP16[$28>>1]|0;
  $30 = $29 << 16 >> 16;
  $31 = ((($23)) + -4|0);
  $32 = HEAP16[$31>>1]|0;
  $33 = $32&65535;
  $34 = (($33) + ($30))|0;
  $35 = ($34|0)==($$|0);
  if ($35) {
   $36 = ((($23)) + -2|0);
   $37 = HEAP8[$36>>0]|0;
   $38 = $37&255;
   $39 = ($38|0)==($coverage$1|0);
   if ($39) {
    $40 = (($33) + ($acount))|0;
    $41 = $40&65535;
    HEAP16[$31>>1] = $41;
    return;
   } else {
    $43 = $14;
   }
  } else {
   $43 = $14;
  }
 } else {
  $43 = $26;
 }
 $42 = ($43|0)!=($14|0);
 $44 = ($21|0)>(31);
 $or$cond = $44 | $42;
 if ($or$cond) {
  $45 = ((($worker)) + 1392|0);
  $46 = HEAP32[$45>>2]|0;
  $47 = ($46|0)!=(0|0);
  $or$cond3 = $24 & $47;
  if ($or$cond3) {
   $48 = ((($worker)) + 1396|0);
   $49 = HEAP32[$48>>2]|0;
   FUNCTION_TABLE_viiii[$46 & 8191]($43,$21,$22,$49);
  }
  HEAP32[$20>>2] = 0;
  HEAP32[$25>>2] = $14;
  $span$0 = $22;
 } else {
  $span$0 = $23;
 }
 $50 = $$&65535;
 HEAP16[$span$0>>1] = $50;
 $51 = $acount&65535;
 $52 = ((($span$0)) + 2|0);
 HEAP16[$52>>1] = $51;
 $53 = $coverage$1&255;
 $54 = ((($span$0)) + 4|0);
 HEAP8[$54>>0] = $53;
 $55 = HEAP32[$20>>2]|0;
 $56 = (($55) + 1)|0;
 HEAP32[$20>>2] = $56;
 return;
}
function _gray_raster_done($raster) {
 $raster = $raster|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$raster>>2]|0;
 _ft_mem_free($0,$raster);
 return;
}
function _ft_smooth_init($render) {
 $render = $render|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($render)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($render)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 56|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($render)) + 52|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($1)) + 164|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($1)) + 168|0);
 $13 = HEAP32[$12>>2]|0;
 FUNCTION_TABLE_viii[$7 & 8191]($9,$11,$13);
 return 0;
}
function _ft_smooth_render($render,$slot,$mode,$origin) {
 $render = $render|0;
 $slot = $slot|0;
 $mode = $mode|0;
 $origin = $origin|0;
 var $$mode = 0, $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mode|0)==(1);
 $$mode = $0 ? 0 : $mode;
 $1 = (_ft_smooth_render_generic($render,$slot,$$mode,$origin,0)|0);
 return ($1|0);
}
function _ft_smooth_render_generic($render,$slot,$mode,$origin,$required_mode) {
 $render = $render|0;
 $slot = $slot|0;
 $mode = $mode|0;
 $origin = $origin|0;
 $required_mode = $required_mode|0;
 var $$ = 0, $$not = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $cbox = 0, $end$017 = 0, $error = 0, $have_buffer$0 = 0, $have_outline_shifted$0 = 0, $hh$020 = 0, $hh1$015 = 0, $line$019 = 0, $or$cond = 0, $or$cond25 = 0, $or$cond7 = 0, $or$cond9 = 0, $params = 0, $pitch$0 = 0, $read$013 = 0, $width$0 = 0, $write$014 = 0, $x_shift$0 = 0, $xx$016 = 0, $y_shift$0 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp + 64|0;
 $cbox = sp + 48|0;
 $params = sp;
 $0 = ((($slot)) + 108|0);
 $1 = ((($slot)) + 76|0);
 $2 = ((($render)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($mode|0)==(3);
 $5 = ($mode|0)==(4);
 $6 = ((($slot)) + 72|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($render)) + 16|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($7|0)==($9|0);
 if (!($10)) {
  HEAP32[$error>>2] = 6;
  $116 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($116|0);
 }
 $11 = ($mode|0)==($required_mode|0);
 if (!($11)) {
  HEAP32[$error>>2] = 19;
  $116 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($116|0);
 }
 $12 = ($origin|0)==(0|0);
 if ($12) {
  $x_shift$0 = 0;$y_shift$0 = 0;
 } else {
  $13 = HEAP32[$origin>>2]|0;
  $14 = ((($origin)) + 4|0);
  $15 = HEAP32[$14>>2]|0;
  $x_shift$0 = $13;$y_shift$0 = $15;
 }
 _FT_Outline_Get_CBox($0,$cbox);
 $16 = HEAP32[$cbox>>2]|0;
 $17 = (($16) + ($x_shift$0))|0;
 $18 = $17 & -64;
 HEAP32[$cbox>>2] = $18;
 $19 = ((($cbox)) + 4|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($20) + ($y_shift$0))|0;
 $22 = $21 & -64;
 HEAP32[$19>>2] = $22;
 $23 = ((($cbox)) + 8|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = (($x_shift$0) + 63)|0;
 $26 = (($25) + ($24))|0;
 $27 = $26 & -64;
 HEAP32[$23>>2] = $27;
 $28 = ((($cbox)) + 12|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = (($y_shift$0) + 63)|0;
 $31 = (($30) + ($29))|0;
 $32 = $31 & -64;
 HEAP32[$28>>2] = $32;
 $33 = (($x_shift$0) - ($18))|0;
 $34 = (($y_shift$0) - ($22))|0;
 $35 = $17 >> 6;
 $36 = $31 >> 6;
 $37 = (($27) - ($18))|0;
 $38 = $37 >>> 6;
 $39 = (($32) - ($22))|0;
 $40 = $39 >>> 6;
 if ($4) {
  $41 = ($38*3)|0;
  $42 = (($41) + 3)|0;
  $43 = $42 & 536870908;
  $pitch$0 = $43;$width$0 = $41;
 } else {
  $pitch$0 = $38;$width$0 = $38;
 }
 $44 = ($40*3)|0;
 $$ = $5 ? $44 : $40;
 $45 = ($width$0|0)>(32767);
 $46 = ($$>>>0)>(32767);
 $or$cond7 = $46 | $45;
 if ($or$cond7) {
  HEAP32[$error>>2] = 98;
  $116 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($116|0);
 }
 $47 = ((($slot)) + 156|0);
 $48 = HEAP32[$47>>2]|0;
 $49 = ((($48)) + 4|0);
 $50 = HEAP32[$49>>2]|0;
 $51 = $50 & 1;
 $52 = ($51|0)==(0);
 $$pre = ((($slot)) + 88|0);
 if (!($52)) {
  $53 = HEAP32[$$pre>>2]|0;
  _ft_mem_free($3,$53);
  HEAP32[$$pre>>2] = 0;
  $54 = HEAP32[$47>>2]|0;
  $55 = ((($54)) + 4|0);
  $56 = HEAP32[$55>>2]|0;
  $57 = $56 & -2;
  HEAP32[$55>>2] = $57;
 }
 $58 = Math_imul($pitch$0, $$)|0;
 $59 = (_ft_mem_alloc($3,$58,$error)|0);
 HEAP32[$$pre>>2] = $59;
 $60 = HEAP32[$error>>2]|0;
 $61 = ($60|0)==(0);
 if (!($61)) {
  $116 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($116|0);
 }
 $62 = HEAP32[$47>>2]|0;
 $63 = ((($62)) + 4|0);
 $64 = HEAP32[$63>>2]|0;
 $65 = $64 | 1;
 HEAP32[$63>>2] = $65;
 HEAP32[$6>>2] = 1651078259;
 $66 = ((($slot)) + 100|0);
 HEAP32[$66>>2] = $35;
 $67 = ((($slot)) + 104|0);
 HEAP32[$67>>2] = $36;
 $68 = ((($slot)) + 94|0);
 HEAP8[$68>>0] = 2;
 $69 = ((($slot)) + 92|0);
 HEAP16[$69>>1] = 256;
 $70 = ((($slot)) + 80|0);
 HEAP32[$70>>2] = $width$0;
 HEAP32[$1>>2] = $$;
 $71 = ((($slot)) + 84|0);
 HEAP32[$71>>2] = $pitch$0;
 $72 = ($x_shift$0|0)!=($18|0);
 $73 = ($y_shift$0|0)!=($22|0);
 $or$cond9 = $72 | $73;
 if ($or$cond9) {
  _FT_Outline_Translate($0,$33,$34);
  $have_outline_shifted$0 = 1;
 } else {
  $have_outline_shifted$0 = 0;
 }
 HEAP32[$params>>2] = $1;
 $74 = ((($params)) + 4|0);
 HEAP32[$74>>2] = $0;
 $75 = ((($params)) + 8|0);
 HEAP32[$75>>2] = 1;
 $76 = ((($render)) + 56|0);
 $77 = HEAP32[$76>>2]|0;
 $78 = ((($render)) + 52|0);
 $79 = HEAP32[$78>>2]|0;
 $80 = (FUNCTION_TABLE_iii[$77 & 8191]($79,$params)|0);
 HEAP32[$error>>2] = $80;
 $81 = ($80|0)==(0);
 if ($81) {
  $$not = $4 ^ 1;
  $82 = ($40|0)==(0);
  $or$cond = $82 | $$not;
  $83 = ($38|0)==(0);
  $or$cond25 = $or$cond | $83;
  if (!($or$cond25)) {
   $84 = HEAP32[$$pre>>2]|0;
   $hh$020 = $40;$line$019 = $84;
   while(1) {
    $85 = (($line$019) + ($width$0)|0);
    $end$017 = $85;$xx$016 = $38;
    while(1) {
     $86 = (($xx$016) + -1)|0;
     $87 = (($line$019) + ($86)|0);
     $88 = HEAP8[$87>>0]|0;
     $89 = ((($end$017)) + -3|0);
     HEAP8[$89>>0] = $88;
     $90 = ((($end$017)) + -2|0);
     HEAP8[$90>>0] = $88;
     $91 = ((($end$017)) + -1|0);
     HEAP8[$91>>0] = $88;
     $92 = ($86|0)==(0);
     if ($92) {
      break;
     } else {
      $end$017 = $89;$xx$016 = $86;
     }
    }
    $93 = (($hh$020) + -1)|0;
    $94 = (($line$019) + ($pitch$0)|0);
    $95 = ($93|0)==(0);
    if ($95) {
     break;
    } else {
     $hh$020 = $93;$line$019 = $94;
    }
   }
  }
  if ($5) {
   $96 = HEAP32[$$pre>>2]|0;
   $97 = ($40|0)==(0);
   if (!($97)) {
    $98 = (($$) - ($40))|0;
    $99 = Math_imul($pitch$0, $98)|0;
    $100 = (($96) + ($99)|0);
    $hh1$015 = $40;$read$013 = $100;$write$014 = $96;
    while(1) {
     _memcpy(($write$014|0),($read$013|0),($pitch$0|0))|0;
     $101 = (($write$014) + ($pitch$0)|0);
     _memcpy(($101|0),($read$013|0),($pitch$0|0))|0;
     $102 = (($101) + ($pitch$0)|0);
     _memcpy(($102|0),($read$013|0),($pitch$0|0))|0;
     $103 = (($102) + ($pitch$0)|0);
     $104 = (($read$013) + ($pitch$0)|0);
     $105 = (($hh1$015) + -1)|0;
     $106 = ($105|0)==(0);
     if ($106) {
      break;
     } else {
      $hh1$015 = $105;$read$013 = $104;$write$014 = $103;
     }
    }
   }
  }
  HEAP32[$error>>2] = 0;
  $have_buffer$0 = 0;
 } else {
  $have_buffer$0 = 1;
 }
 $107 = ($have_outline_shifted$0<<24>>24)==(0);
 if (!($107)) {
  $108 = (0 - ($33))|0;
  $109 = (0 - ($34))|0;
  _FT_Outline_Translate($0,$108,$109);
 }
 $110 = ($have_buffer$0<<24>>24)==(0);
 if ($110) {
  $116 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($116|0);
 }
 $111 = HEAP32[$$pre>>2]|0;
 _ft_mem_free($3,$111);
 HEAP32[$$pre>>2] = 0;
 $112 = HEAP32[$47>>2]|0;
 $113 = ((($112)) + 4|0);
 $114 = HEAP32[$113>>2]|0;
 $115 = $114 & -2;
 HEAP32[$113>>2] = $115;
 $116 = HEAP32[$error>>2]|0;
 STACKTOP = sp;return ($116|0);
}
function _ft_smooth_transform($render,$slot,$matrix,$delta) {
 $render = $render|0;
 $slot = $slot|0;
 $matrix = $matrix|0;
 $delta = $delta|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($slot)) + 72|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($render)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==($3|0);
 if (!($4)) {
  $error$0 = 6;
  return ($error$0|0);
 }
 $5 = ($matrix|0)==(0|0);
 if (!($5)) {
  $6 = ((($slot)) + 108|0);
  _FT_Outline_Transform($6,$matrix);
 }
 $7 = ($delta|0)==(0|0);
 if ($7) {
  $error$0 = 0;
  return ($error$0|0);
 }
 $8 = ((($slot)) + 108|0);
 $9 = HEAP32[$delta>>2]|0;
 $10 = ((($delta)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 _FT_Outline_Translate($8,$9,$11);
 $error$0 = 0;
 return ($error$0|0);
}
function _ft_smooth_get_cbox($render,$slot,$cbox) {
 $render = $render|0;
 $slot = $slot|0;
 $cbox = $cbox|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$cbox>>2]=0|0;HEAP32[$cbox+4>>2]=0|0;HEAP32[$cbox+8>>2]=0|0;HEAP32[$cbox+12>>2]=0|0;
 $0 = ((($slot)) + 72|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($render)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==($3|0);
 if (!($4)) {
  return;
 }
 $5 = ((($slot)) + 108|0);
 _FT_Outline_Get_CBox($5,$cbox);
 return;
}
function _ft_smooth_set_mode($render,$mode_tag,$data) {
 $render = $render|0;
 $mode_tag = $mode_tag|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($render)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 56|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($render)) + 52|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (FUNCTION_TABLE_iiii[$5 & 8191]($7,$mode_tag,$data)|0);
 return ($8|0);
}
function _ft_smooth_render_lcd($render,$slot,$mode,$origin) {
 $render = $render|0;
 $slot = $slot|0;
 $mode = $mode|0;
 $origin = $origin|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_ft_smooth_render_generic($render,$slot,$mode,$origin,3)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  return ($0|0);
 }
 $2 = ((($slot)) + 94|0);
 HEAP8[$2>>0] = 5;
 return ($0|0);
}
function _ft_smooth_render_lcd_v($render,$slot,$mode,$origin) {
 $render = $render|0;
 $slot = $slot|0;
 $mode = $mode|0;
 $origin = $origin|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_ft_smooth_render_generic($render,$slot,$mode,$origin,4)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  return ($0|0);
 }
 $2 = ((($slot)) + 94|0);
 HEAP8[$2>>0] = 6;
 return ($0|0);
}
function _tt_driver_init($ttdriver) {
 $ttdriver = $ttdriver|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($ttdriver)) + 64|0);
 HEAP32[$0>>2] = 35;
 return 0;
}
function _tt_driver_done($ttdriver) {
 $ttdriver = $ttdriver|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _tt_get_interface($driver,$tt_interface) {
 $driver = $driver|0;
 $tt_interface = $tt_interface|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_ft_service_list_lookup(55692,$tt_interface)|0);
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  $$0 = $0;
  return ($$0|0);
 }
 $2 = ($driver|0)==(0|0);
 if ($2) {
  $$0 = 0;
  return ($$0|0);
 }
 $3 = ((($driver)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if ($5) {
  $$0 = 0;
  return ($$0|0);
 }
 $6 = (_FT_Get_Module($4,593168)|0);
 $7 = ($6|0)==(0|0);
 if ($7) {
  $$0 = 0;
  return ($$0|0);
 }
 $8 = HEAP32[$6>>2]|0;
 $9 = ((($8)) + 20|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10|0)==(0|0);
 if ($11) {
  $$0 = 0;
  return ($$0|0);
 }
 $12 = ((($10)) + 16|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (FUNCTION_TABLE_iii[$13 & 8191]($driver,$tt_interface)|0);
 $$0 = $14;
 return ($$0|0);
}
function _TT_Set_MM_Blend($face,$num_coords,$coords) {
 $face = $face|0;
 $num_coords = $num_coords|0;
 $coords = $coords|0;
 var $$lcssa = 0, $$num_coords = 0, $$off = 0, $$pr = 0, $$pre = 0, $$pre$i = 0, $$pre10$i = 0, $$pre15 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $error = 0, $error$i = 0, $gvar_head$i = 0, $i$012 = 0, $i$09$i = 0, $i$1$lcssa = 0, $i$15 = 0, $i$18$i = 0, $i$22 = 0, $i$25$i = 0, $j$04$i = 0;
 var $manageCvt$0$ph = 0, $manageCvt$1 = 0, $table_len$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp + 28|0;
 $table_len$i = sp + 24|0;
 $gvar_head$i = sp + 4|0;
 $error = sp;
 HEAP32[$error>>2] = 0;
 $0 = ((($face)) + 100|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 672|0);
 HEAP8[$2>>0] = 0;
 $3 = ((($face)) + 676|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 do {
  if ($5) {
   $6 = (_TT_Get_MM_Var($face,0)|0);
   HEAP32[$error>>2] = $6;
   $7 = ($6|0)==(0);
   if ($7) {
    $$pre = HEAP32[$3>>2]|0;
    $9 = $$pre;
    break;
   } else {
    $158 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($158|0);
   }
  } else {
   $9 = $4;
  }
 } while(0);
 $8 = ((($9)) + 8|0);
 $10 = HEAP32[$8>>2]|0;
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11>>>0)<($num_coords>>>0);
 $$num_coords = $12 ? $11 : $num_coords;
 $13 = ($$num_coords|0)==(0);
 L7: do {
  if (!($13)) {
   $i$012 = 0;
   while(1) {
    $16 = (($coords) + ($i$012<<2)|0);
    $17 = HEAP32[$16>>2]|0;
    $$off = (($17) + 65536)|0;
    $18 = ($$off>>>0)>(131072);
    $15 = (($i$012) + 1)|0;
    if ($18) {
     break;
    }
    $14 = ($15>>>0)<($$num_coords>>>0);
    if ($14) {
     $i$012 = $15;
    } else {
     break L7;
    }
   }
   HEAP32[$error>>2] = 6;
   $158 = HEAP32[$error>>2]|0;
   STACKTOP = sp;return ($158|0);
  }
 } while(0);
 $19 = ((($9)) + 36|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ($20|0)==(0|0);
 if ($21) {
  $22 = ((($face)) + 104|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = ((($23)) + 28|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = ((($face)) + 508|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = (FUNCTION_TABLE_iiiii[$27 & 8191]($face,1735811442,$23,$table_len$i)|0);
  HEAP32[$error$i>>2] = $28;
  $29 = ($28|0)==(0);
  do {
   if ($29) {
    $30 = (_FT_Stream_Pos($23)|0);
    $31 = (_FT_Stream_ReadFields($23,174874,$gvar_head$i)|0);
    HEAP32[$error$i>>2] = $31;
    $32 = ($31|0)==(0);
    if ($32) {
     $33 = ((($gvar_head$i)) + 6|0);
     $34 = HEAP16[$33>>1]|0;
     $35 = $34&65535;
     $36 = ((($9)) + 24|0);
     HEAP32[$36>>2] = $35;
     $37 = ((($gvar_head$i)) + 12|0);
     $38 = HEAP16[$37>>1]|0;
     $39 = $38&65535;
     $40 = ((($9)) + 32|0);
     HEAP32[$40>>2] = $39;
     $41 = ((($gvar_head$i)) + 16|0);
     $42 = HEAP32[$41>>2]|0;
     $43 = (($42) + ($30))|0;
     $44 = HEAP32[$gvar_head$i>>2]|0;
     $45 = ($44|0)==(65536);
     if (!($45)) {
      HEAP32[$error$i>>2] = 8;
      $123 = 8;
      break;
     }
     $46 = ((($gvar_head$i)) + 4|0);
     $47 = HEAP16[$46>>1]|0;
     $48 = HEAP32[$8>>2]|0;
     $49 = HEAP32[$48>>2]|0;
     $50 = $49&65535;
     $51 = ($47<<16>>16)==($50<<16>>16);
     if (!($51)) {
      HEAP32[$error$i>>2] = 8;
      $123 = 8;
      break;
     }
     $52 = (($39) + 1)|0;
     $53 = (_ft_mem_realloc($25,4,0,$52,0,$error$i)|0);
     HEAP32[$19>>2] = $53;
     $54 = HEAP32[$error$i>>2]|0;
     $55 = ($54|0)==(0);
     if ($55) {
      $56 = ((($gvar_head$i)) + 14|0);
      $57 = HEAP16[$56>>1]|0;
      $58 = $57 & 1;
      $59 = ($58<<16>>16)==(0);
      $60 = HEAP32[$40>>2]|0;
      $61 = (($60) + 1)|0;
      if ($59) {
       $72 = $61 << 1;
       $73 = (_FT_Stream_EnterFrame($23,$72)|0);
       HEAP32[$error$i>>2] = $73;
       $74 = ($73|0)==(0);
       if ($74) {
        $i$18$i = 0;
       } else {
        $123 = $73;
        break;
       }
       while(1) {
        $75 = (_FT_Stream_GetUShort($23)|0);
        $76 = $75&65535;
        $77 = $76 << 1;
        $78 = (($77) + ($43))|0;
        $79 = HEAP32[$19>>2]|0;
        $80 = (($79) + ($i$18$i<<2)|0);
        HEAP32[$80>>2] = $78;
        $81 = (($i$18$i) + 1)|0;
        $82 = HEAP32[$40>>2]|0;
        $83 = ($81>>>0)>($82>>>0);
        if ($83) {
         break;
        } else {
         $i$18$i = $81;
        }
       }
       _FT_Stream_ExitFrame($23);
      } else {
       $62 = $61 << 2;
       $63 = (_FT_Stream_EnterFrame($23,$62)|0);
       HEAP32[$error$i>>2] = $63;
       $64 = ($63|0)==(0);
       if ($64) {
        $i$09$i = 0;
       } else {
        $123 = $63;
        break;
       }
       while(1) {
        $65 = (_FT_Stream_GetULong($23)|0);
        $66 = (($65) + ($43))|0;
        $67 = HEAP32[$19>>2]|0;
        $68 = (($67) + ($i$09$i<<2)|0);
        HEAP32[$68>>2] = $66;
        $69 = (($i$09$i) + 1)|0;
        $70 = HEAP32[$40>>2]|0;
        $71 = ($69>>>0)>($70>>>0);
        if ($71) {
         break;
        } else {
         $i$09$i = $69;
        }
       }
       _FT_Stream_ExitFrame($23);
      }
      $84 = HEAP32[$36>>2]|0;
      $85 = ($84|0)==(0);
      if (!($85)) {
       $86 = HEAP16[$46>>1]|0;
       $87 = $86&65535;
       $88 = Math_imul($87, $84)|0;
       $89 = (_ft_mem_realloc($25,4,0,$88,0,$error$i)|0);
       $90 = ((($9)) + 28|0);
       HEAP32[$90>>2] = $89;
       $91 = HEAP32[$error$i>>2]|0;
       $92 = ($91|0)==(0);
       if (!($92)) {
        $123 = $91;
        break;
       }
       $93 = ((($gvar_head$i)) + 8|0);
       $94 = HEAP32[$93>>2]|0;
       $95 = (($94) + ($30))|0;
       $96 = (_FT_Stream_Seek($23,$95)|0);
       HEAP32[$error$i>>2] = $96;
       $97 = ($96|0)==(0);
       if (!($97)) {
        $123 = $96;
        break;
       }
       $98 = HEAP32[$36>>2]|0;
       $99 = HEAP16[$46>>1]|0;
       $100 = $99&65535;
       $101 = $98 << 1;
       $102 = Math_imul($101, $100)|0;
       $103 = (_FT_Stream_EnterFrame($23,$102)|0);
       HEAP32[$error$i>>2] = $103;
       $104 = ($103|0)==(0);
       if (!($104)) {
        $123 = $103;
        break;
       }
       $105 = HEAP32[$36>>2]|0;
       $106 = ($105|0)==(0);
       if (!($106)) {
        $$pre$i = HEAP16[$46>>1]|0;
        $108 = $$pre$i;$159 = $105;$i$25$i = 0;
        while(1) {
         $107 = ($108<<16>>16)==(0);
         if ($107) {
          $122 = $159;$160 = 0;
         } else {
          $j$04$i = 0;
          while(1) {
           $109 = (_FT_Stream_GetUShort($23)|0);
           $110 = $109 << 16 >> 16;
           $111 = $110 << 2;
           $112 = HEAP16[$46>>1]|0;
           $113 = $112&65535;
           $114 = Math_imul($113, $i$25$i)|0;
           $115 = (($114) + ($j$04$i))|0;
           $116 = HEAP32[$90>>2]|0;
           $117 = (($116) + ($115<<2)|0);
           HEAP32[$117>>2] = $111;
           $118 = (($j$04$i) + 1)|0;
           $119 = ($118>>>0)<($113>>>0);
           if ($119) {
            $j$04$i = $118;
           } else {
            $$lcssa = $112;
            break;
           }
          }
          $$pre10$i = HEAP32[$36>>2]|0;
          $122 = $$pre10$i;$160 = $$lcssa;
         }
         $120 = (($i$25$i) + 1)|0;
         $121 = ($120>>>0)<($122>>>0);
         if ($121) {
          $108 = $160;$159 = $122;$i$25$i = $120;
         } else {
          break;
         }
        }
       }
       _FT_Stream_ExitFrame($23);
      }
      $$pr = HEAP32[$error$i>>2]|0;
      $123 = $$pr;
     } else {
      $123 = $54;
     }
    } else {
     $123 = $31;
    }
   } else {
    $123 = $28;
   }
  } while(0);
  HEAP32[$error>>2] = $123;
  $124 = ($123|0)==(0);
  if (!($124)) {
   $158 = HEAP32[$error>>2]|0;
   STACKTOP = sp;return ($158|0);
  }
 }
 $125 = ((($9)) + 4|0);
 $126 = HEAP32[$125>>2]|0;
 $127 = ($126|0)==(0|0);
 L55: do {
  if ($127) {
   $128 = HEAP32[$10>>2]|0;
   $129 = (_ft_mem_realloc($1,4,0,$128,0,$error)|0);
   HEAP32[$125>>2] = $129;
   $130 = HEAP32[$error>>2]|0;
   $131 = ($130|0)==(0);
   if ($131) {
    $$pre15 = HEAP32[$10>>2]|0;
    $146 = $$pre15;$148 = $129;$manageCvt$1 = 1;
    break;
   } else {
    $158 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($158|0);
   }
  } else {
   L61: do {
    if ($13) {
     $i$1$lcssa = 0;$manageCvt$0$ph = 0;
    } else {
     $i$15 = 0;
     while(1) {
      $132 = (($126) + ($i$15<<2)|0);
      $133 = HEAP32[$132>>2]|0;
      $134 = (($coords) + ($i$15<<2)|0);
      $135 = HEAP32[$134>>2]|0;
      $136 = ($133|0)==($135|0);
      if (!($136)) {
       $i$1$lcssa = $i$15;$manageCvt$0$ph = 2;
       break L61;
      }
      $139 = (($i$15) + 1)|0;
      $140 = ($139>>>0)<($$num_coords>>>0);
      if ($140) {
       $i$15 = $139;
      } else {
       $i$1$lcssa = $139;$manageCvt$0$ph = 0;
       break;
      }
     }
    }
   } while(0);
   $137 = HEAP32[$10>>2]|0;
   $138 = ($i$1$lcssa>>>0)<($137>>>0);
   if ($138) {
    $i$22 = $i$1$lcssa;
    while(1) {
     $143 = (($126) + ($i$22<<2)|0);
     $144 = HEAP32[$143>>2]|0;
     $145 = ($144|0)==(0);
     $142 = (($i$22) + 1)|0;
     if (!($145)) {
      $146 = $137;$148 = $126;$manageCvt$1 = 2;
      break L55;
     }
     $141 = ($142>>>0)<($137>>>0);
     if ($141) {
      $i$22 = $142;
     } else {
      $146 = $137;$148 = $126;$manageCvt$1 = $manageCvt$0$ph;
      break;
     }
    }
   } else {
    $146 = $137;$148 = $126;$manageCvt$1 = $manageCvt$0$ph;
   }
  }
 } while(0);
 HEAP32[$9>>2] = $146;
 $147 = $$num_coords << 2;
 _memcpy(($148|0),($coords|0),($147|0))|0;
 HEAP8[$2>>0] = 1;
 $149 = ((($face)) + 648|0);
 $150 = HEAP32[$149>>2]|0;
 $151 = ($150|0)==(0|0);
 if ($151) {
  $158 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($158|0);
 }
 switch ($manageCvt$1|0) {
 case 2:  {
  _ft_mem_free($1,$150);
  HEAP32[$149>>2] = 0;
  $152 = ((($face)) + 104|0);
  $153 = HEAP32[$152>>2]|0;
  $154 = (_tt_face_load_cvt($face,$153)|0);
  HEAP32[$error>>2] = $154;
  $158 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($158|0);
  break;
 }
 case 1:  {
  $155 = ((($face)) + 104|0);
  $156 = HEAP32[$155>>2]|0;
  $157 = (_tt_face_vary_cvt($face,$156)|0);
  HEAP32[$error>>2] = $157;
  $158 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($158|0);
  break;
 }
 default: {
  $158 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($158|0);
 }
 }
 return (0)|0;
}
function _TT_Get_MM_Var($face,$master) {
 $face = $face|0;
 $master = $master|0;
 var $$lcssa = 0, $$pre = 0, $$pre40 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $a$022 = 0, $a$26 = 0, $axis_rec = 0, $error = 0, $fvar_head = 0, $i$031 = 0, $i$126 = 0, $i$221 = 0, $i$31746 = 0, $j$013 = 0, $n$09 = 0, $n$17 = 0, $next_coords$0$lcssa = 0;
 var $next_coords$032 = 0, $next_coords$1$lcssa = 0, $next_coords$18 = 0, $next_name$027 = 0, $next_name$15 = 0, $ns$01845 = 0, $switch$split12D = 0, $switch$split2D = 0, $switch$split42D = 0, $table_len = 0, $tmp3 = 0, $tmp4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $table_len = sp + 44|0;
 $error = sp + 40|0;
 $fvar_head = sp + 24|0;
 $axis_rec = sp;
 $0 = ((($face)) + 104|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 100|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$error>>2] = 0;
 $4 = ((($face)) + 676|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 L1: do {
  if ($6) {
   $7 = ((($face)) + 508|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = (FUNCTION_TABLE_iiiii[$8 & 8191]($face,1735811442,$1,$table_len)|0);
   HEAP32[$error>>2] = $9;
   $10 = ($9|0)==(0);
   if (!($10)) {
    $203 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($203|0);
   }
   $11 = HEAP32[$7>>2]|0;
   $12 = (FUNCTION_TABLE_iiiii[$11 & 8191]($face,1719034226,$1,$table_len)|0);
   HEAP32[$error>>2] = $12;
   $13 = ($12|0)==(0);
   if (!($13)) {
    $203 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($203|0);
   }
   $14 = (_FT_Stream_Pos($1)|0);
   $15 = (_FT_Stream_ReadFields($1,174806,$fvar_head)|0);
   HEAP32[$error>>2] = $15;
   $16 = ($15|0)==(0);
   if (!($16)) {
    $203 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($203|0);
   }
   $17 = HEAP32[$fvar_head>>2]|0;
   $18 = ($17|0)==(65536);
   if ($18) {
    $19 = ((($fvar_head)) + 10|0);
    $20 = HEAP16[$19>>1]|0;
    $21 = ($20<<16>>16)==(20);
    if ($21) {
     $22 = ((($fvar_head)) + 8|0);
     $23 = HEAP16[$22>>1]|0;
     $24 = $23&65535;
     $25 = ($23&65535)>(16382);
     if (!($25)) {
      $26 = ((($fvar_head)) + 14|0);
      $27 = HEAP16[$26>>1]|0;
      $28 = $27&65535;
      $29 = $24 << 2;
      $30 = (($29) + 4)|0;
      $31 = ($28|0)==($30|0);
      if ($31) {
       $32 = ((($fvar_head)) + 12|0);
       $33 = HEAP16[$32>>1]|0;
       $34 = ($33&65535)>(32511);
       if (!($34)) {
        $35 = $33&65535;
        $36 = ((($fvar_head)) + 4|0);
        $37 = HEAP16[$36>>1]|0;
        $38 = $37&65535;
        $39 = ($24*20)|0;
        $40 = Math_imul($35, $28)|0;
        $41 = (($40) + ($39))|0;
        $42 = (($41) + ($38))|0;
        $43 = HEAP32[$table_len>>2]|0;
        $44 = ($42>>>0)>($43>>>0);
        if (!($44)) {
         $45 = (_ft_mem_alloc($3,40,$error)|0);
         HEAP32[$4>>2] = $45;
         $46 = HEAP32[$error>>2]|0;
         $47 = ($46|0)==(0);
         if (!($47)) {
          $203 = HEAP32[$error>>2]|0;
          STACKTOP = sp;return ($203|0);
         }
         $48 = HEAP16[$22>>1]|0;
         $49 = $48&65535;
         $50 = HEAP16[$32>>1]|0;
         $51 = $50&65535;
         $52 = $51 << 3;
         $53 = $51 << 2;
         $tmp3 = (($53) + 29)|0;
         $tmp4 = Math_imul($tmp3, $49)|0;
         $54 = (($52) + 20)|0;
         $55 = (($54) + ($tmp4))|0;
         $56 = ((($45)) + 12|0);
         HEAP32[$56>>2] = $55;
         $57 = (_ft_mem_alloc($3,$55,$error)|0);
         $58 = HEAP32[$error>>2]|0;
         $59 = ($58|0)==(0);
         if (!($59)) {
          $203 = HEAP32[$error>>2]|0;
          STACKTOP = sp;return ($203|0);
         }
         $60 = HEAP32[$4>>2]|0;
         $61 = ((($60)) + 8|0);
         HEAP32[$61>>2] = $57;
         $62 = HEAP16[$22>>1]|0;
         $63 = $62&65535;
         HEAP32[$57>>2] = $63;
         $64 = ((($57)) + 4|0);
         HEAP32[$64>>2] = -1;
         $65 = HEAP16[$32>>1]|0;
         $66 = $65&65535;
         $67 = ((($57)) + 8|0);
         HEAP32[$67>>2] = $66;
         $68 = ((($57)) + 20|0);
         $69 = ((($57)) + 12|0);
         HEAP32[$69>>2] = $68;
         $70 = (($68) + (($63*24)|0)|0);
         $71 = ((($57)) + 16|0);
         HEAP32[$71>>2] = $70;
         $72 = (($70) + ($66<<3)|0);
         $73 = ($65<<16>>16)==(0);
         if ($73) {
          $next_coords$0$lcssa = $72;
         } else {
          $74 = HEAP16[$22>>1]|0;
          $75 = $74&65535;
          $76 = HEAP16[$32>>1]|0;
          $77 = $76&65535;
          $i$031 = 0;$next_coords$032 = $72;
          while(1) {
           $78 = (($70) + ($i$031<<3)|0);
           HEAP32[$78>>2] = $next_coords$032;
           $79 = (($next_coords$032) + ($75<<2)|0);
           $80 = (($i$031) + 1)|0;
           $81 = ($80|0)<($77|0);
           if ($81) {
            $i$031 = $80;$next_coords$032 = $79;
           } else {
            $next_coords$0$lcssa = $79;
            break;
           }
          }
         }
         $82 = HEAP16[$22>>1]|0;
         $83 = ($82<<16>>16)==(0);
         if (!($83)) {
          $84 = HEAP32[$69>>2]|0;
          $85 = $82&65535;
          $i$126 = 0;$next_name$027 = $next_coords$0$lcssa;
          while(1) {
           $86 = (($84) + (($i$126*24)|0)|0);
           HEAP32[$86>>2] = $next_name$027;
           $87 = ((($next_name$027)) + 5|0);
           $88 = (($i$126) + 1)|0;
           $89 = ($88|0)<($85|0);
           if ($89) {
            $i$126 = $88;$next_name$027 = $87;
           } else {
            break;
           }
          }
         }
         $90 = HEAP16[$36>>1]|0;
         $91 = $90&65535;
         $92 = (($91) + ($14))|0;
         $93 = (_FT_Stream_Seek($1,$92)|0);
         HEAP32[$error>>2] = $93;
         $94 = ($93|0)==(0);
         if (!($94)) {
          $203 = HEAP32[$error>>2]|0;
          STACKTOP = sp;return ($203|0);
         }
         $95 = HEAP16[$22>>1]|0;
         $96 = ($95<<16>>16)==(0);
         L38: do {
          if ($96) {
           $143 = 0;
          } else {
           $97 = HEAP32[$69>>2]|0;
           $98 = ((($axis_rec)) + 4|0);
           $99 = ((($axis_rec)) + 8|0);
           $100 = ((($axis_rec)) + 12|0);
           $101 = ((($axis_rec)) + 18|0);
           $a$022 = $97;$i$221 = 0;
           while(1) {
            $102 = (_FT_Stream_ReadFields($1,174842,$axis_rec)|0);
            HEAP32[$error>>2] = $102;
            $103 = ($102|0)==(0);
            if (!($103)) {
             break;
            }
            $104 = HEAP32[$axis_rec>>2]|0;
            $105 = ((($a$022)) + 16|0);
            HEAP32[$105>>2] = $104;
            $106 = HEAP32[$98>>2]|0;
            $107 = ((($a$022)) + 4|0);
            HEAP32[$107>>2] = $106;
            $108 = HEAP32[$99>>2]|0;
            $109 = ((($a$022)) + 8|0);
            HEAP32[$109>>2] = $108;
            $110 = HEAP32[$100>>2]|0;
            $111 = ((($a$022)) + 12|0);
            HEAP32[$111>>2] = $110;
            $112 = HEAP16[$101>>1]|0;
            $113 = $112&65535;
            $114 = ((($a$022)) + 20|0);
            HEAP32[$114>>2] = $113;
            $115 = $104 >>> 24;
            $116 = $115&255;
            $117 = HEAP32[$a$022>>2]|0;
            HEAP8[$117>>0] = $116;
            $118 = HEAP32[$105>>2]|0;
            $119 = $118 >>> 16;
            $120 = $119&255;
            $121 = HEAP32[$a$022>>2]|0;
            $122 = ((($121)) + 1|0);
            HEAP8[$122>>0] = $120;
            $123 = HEAP32[$105>>2]|0;
            $124 = $123 >>> 8;
            $125 = $124&255;
            $126 = HEAP32[$a$022>>2]|0;
            $127 = ((($126)) + 2|0);
            HEAP8[$127>>0] = $125;
            $128 = HEAP32[$105>>2]|0;
            $129 = $128&255;
            $130 = HEAP32[$a$022>>2]|0;
            $131 = ((($130)) + 3|0);
            HEAP8[$131>>0] = $129;
            $132 = HEAP32[$a$022>>2]|0;
            $133 = ((($132)) + 4|0);
            HEAP8[$133>>0] = 0;
            $134 = ((($a$022)) + 24|0);
            $135 = (($i$221) + 1)|0;
            $136 = HEAP16[$22>>1]|0;
            $137 = $136&65535;
            $138 = ($135|0)<($137|0);
            if ($138) {
             $a$022 = $134;$i$221 = $135;
            } else {
             $143 = $136;
             break L38;
            }
           }
           $203 = HEAP32[$error>>2]|0;
           STACKTOP = sp;return ($203|0);
          }
         } while(0);
         $139 = HEAP16[$32>>1]|0;
         $140 = ($139<<16>>16)==(0);
         if ($140) {
          break;
         }
         $141 = HEAP32[$71>>2]|0;
         $142 = $143&65535;
         $144 = $142 << 2;
         $145 = (($144) + 4)|0;
         $146 = (_FT_Stream_EnterFrame($1,$145)|0);
         HEAP32[$error>>2] = $146;
         $147 = ($146|0)==(0);
         if ($147) {
          $i$31746 = 0;$ns$01845 = $141;
         } else {
          $203 = HEAP32[$error>>2]|0;
          STACKTOP = sp;return ($203|0);
         }
         while(1) {
          $148 = (_FT_Stream_GetUShort($1)|0);
          $149 = $148&65535;
          $150 = ((($ns$01845)) + 4|0);
          HEAP32[$150>>2] = $149;
          (_FT_Stream_GetUShort($1)|0);
          $151 = HEAP16[$22>>1]|0;
          $152 = ($151<<16>>16)==(0);
          if (!($152)) {
           $j$013 = 0;
           while(1) {
            $153 = (_FT_Stream_GetULong($1)|0);
            $154 = HEAP32[$ns$01845>>2]|0;
            $155 = (($154) + ($j$013<<2)|0);
            HEAP32[$155>>2] = $153;
            $156 = (($j$013) + 1)|0;
            $157 = HEAP16[$22>>1]|0;
            $158 = $157&65535;
            $159 = ($156|0)<($158|0);
            if ($159) {
             $j$013 = $156;
            } else {
             break;
            }
           }
          }
          _FT_Stream_ExitFrame($1);
          $160 = (($i$31746) + 1)|0;
          $161 = HEAP16[$32>>1]|0;
          $162 = $161&65535;
          $163 = ($160|0)<($162|0);
          if (!($163)) {
           break L1;
          }
          $164 = ((($ns$01845)) + 8|0);
          $$pre = HEAP16[$22>>1]|0;
          $165 = $$pre&65535;
          $166 = $165 << 2;
          $167 = (($166) + 4)|0;
          $168 = (_FT_Stream_EnterFrame($1,$167)|0);
          HEAP32[$error>>2] = $168;
          $169 = ($168|0)==(0);
          if ($169) {
           $i$31746 = $160;$ns$01845 = $164;
          } else {
           break;
          }
         }
         $203 = HEAP32[$error>>2]|0;
         STACKTOP = sp;return ($203|0);
        }
       }
      }
     }
    }
   }
   HEAP32[$error>>2] = 8;
   $203 = HEAP32[$error>>2]|0;
   STACKTOP = sp;return ($203|0);
  }
 } while(0);
 $170 = ($master|0)==(0|0);
 if ($170) {
  $203 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($203|0);
 }
 $171 = HEAP32[$4>>2]|0;
 $172 = ((($171)) + 12|0);
 $173 = HEAP32[$172>>2]|0;
 $174 = (_ft_mem_alloc($3,$173,$error)|0);
 $175 = HEAP32[$error>>2]|0;
 $176 = ($175|0)==(0);
 if (!($176)) {
  $203 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($203|0);
 }
 $177 = HEAP32[$4>>2]|0;
 $178 = ((($177)) + 8|0);
 $179 = HEAP32[$178>>2]|0;
 $180 = ((($177)) + 12|0);
 $181 = HEAP32[$180>>2]|0;
 _memcpy(($174|0),($179|0),($181|0))|0;
 $182 = ((($174)) + 20|0);
 $183 = ((($174)) + 12|0);
 HEAP32[$183>>2] = $182;
 $184 = HEAP32[$174>>2]|0;
 $185 = (($182) + (($184*24)|0)|0);
 $186 = ((($174)) + 16|0);
 HEAP32[$186>>2] = $185;
 $187 = ((($174)) + 8|0);
 $188 = HEAP32[$187>>2]|0;
 $189 = (($185) + ($188<<3)|0);
 $190 = ($188|0)==(0);
 if ($190) {
  $196 = $184;$next_coords$1$lcssa = $189;
 } else {
  $n$09 = 0;$next_coords$18 = $189;
  while(1) {
   $191 = (($185) + ($n$09<<3)|0);
   HEAP32[$191>>2] = $next_coords$18;
   $192 = (($next_coords$18) + ($184<<2)|0);
   $193 = (($n$09) + 1)|0;
   $194 = ($193>>>0)<($188>>>0);
   if ($194) {
    $n$09 = $193;$next_coords$18 = $192;
   } else {
    $$lcssa = $192;
    break;
   }
  }
  $$pre40 = HEAP32[$174>>2]|0;
  $196 = $$pre40;$next_coords$1$lcssa = $$lcssa;
 }
 $195 = ($196|0)==(0);
 if (!($195)) {
  $a$26 = $182;$n$17 = 0;$next_name$15 = $next_coords$1$lcssa;
  while(1) {
   HEAP32[$a$26>>2] = $next_name$15;
   $197 = ((($a$26)) + 16|0);
   $198 = HEAP32[$197>>2]|0;
   $switch$split2D = ($198|0)<(2003072104);
   L76: do {
    if ($switch$split2D) {
     $switch$split12D = ($198|0)<(1936486004);
     if ($switch$split12D) {
      switch ($198|0) {
      case 1869640570:  {
       break;
      }
      default: {
       break L76;
      }
      }
      HEAP32[$a$26>>2] = 593906;
      break;
     } else {
      switch ($198|0) {
      case 1936486004:  {
       break;
      }
      default: {
       break L76;
      }
      }
      HEAP32[$a$26>>2] = 593118;
      break;
     }
    } else {
     $switch$split42D = ($198|0)<(2003265652);
     if ($switch$split42D) {
      switch ($198|0) {
      case 2003072104:  {
       break;
      }
      default: {
       break L76;
      }
      }
      HEAP32[$a$26>>2] = 593900;
      break;
     } else {
      switch ($198|0) {
      case 2003265652:  {
       break;
      }
      default: {
       break L76;
      }
      }
      HEAP32[$a$26>>2] = 594457;
      break;
     }
    }
   } while(0);
   $199 = ((($next_name$15)) + 5|0);
   $200 = ((($a$26)) + 24|0);
   $201 = (($n$17) + 1)|0;
   $202 = ($201>>>0)<($196>>>0);
   if ($202) {
    $a$26 = $200;$n$17 = $201;$next_name$15 = $199;
   } else {
    break;
   }
  }
 }
 HEAP32[$master>>2] = $174;
 $203 = HEAP32[$error>>2]|0;
 STACKTOP = sp;return ($203|0);
}
function _tt_face_load_cvt($face,$stream) {
 $face = $face|0;
 $stream = $stream|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cur$01 = 0, $error = 0, $table_len = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp + 4|0;
 $table_len = sp;
 $0 = ((($stream)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 508|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiiii[$3 & 8191]($face,1668707360,$stream,$table_len)|0);
 HEAP32[$error>>2] = $4;
 $5 = ($4|0)==(0);
 if (!($5)) {
  $6 = ((($face)) + 644|0);
  HEAP32[$6>>2] = 0;
  $7 = ((($face)) + 648|0);
  HEAP32[$7>>2] = 0;
  HEAP32[$error>>2] = 0;
  $30 = 0;
  STACKTOP = sp;return ($30|0);
 }
 $8 = HEAP32[$table_len>>2]|0;
 $9 = $8 >>> 1;
 $10 = ((($face)) + 644|0);
 HEAP32[$10>>2] = $9;
 $11 = (_ft_mem_realloc($1,2,0,$9,0,$error)|0);
 $12 = ((($face)) + 648|0);
 HEAP32[$12>>2] = $11;
 $13 = HEAP32[$error>>2]|0;
 $14 = ($13|0)==(0);
 if (!($14)) {
  $30 = $13;
  STACKTOP = sp;return ($30|0);
 }
 $15 = HEAP32[$10>>2]|0;
 $16 = $15 << 1;
 $17 = (_FT_Stream_EnterFrame($stream,$16)|0);
 HEAP32[$error>>2] = $17;
 $18 = ($17|0)==(0);
 if (!($18)) {
  $30 = $17;
  STACKTOP = sp;return ($30|0);
 }
 $19 = HEAP32[$12>>2]|0;
 $20 = HEAP32[$10>>2]|0;
 $21 = (($19) + ($20<<1)|0);
 $22 = ($20|0)>(0);
 if ($22) {
  $cur$01 = $19;
  while(1) {
   $23 = (_FT_Stream_GetUShort($stream)|0);
   HEAP16[$cur$01>>1] = $23;
   $24 = ((($cur$01)) + 2|0);
   $25 = ($24>>>0)<($21>>>0);
   if ($25) {
    $cur$01 = $24;
   } else {
    break;
   }
  }
 }
 _FT_Stream_ExitFrame($stream);
 $26 = ((($face)) + 672|0);
 $27 = HEAP8[$26>>0]|0;
 $28 = ($27<<24>>24)==(0);
 if ($28) {
  $$pre = HEAP32[$error>>2]|0;
  $30 = $$pre;
  STACKTOP = sp;return ($30|0);
 } else {
  $29 = (_tt_face_vary_cvt($face,$stream)|0);
  HEAP32[$error>>2] = $29;
  $30 = $29;
  STACKTOP = sp;return ($30|0);
 }
 return (0)|0;
}
function _tt_face_vary_cvt($face,$stream) {
 $face = $face|0;
 $stream = $stream|0;
 var $$lcssa38 = 0, $$mask = 0, $$pre = 0, $$pre29 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $error = 0, $exitcond = 0, $i$024 = 0, $im_end_coords$0 = 0, $im_end_coords$1 = 0, $im_start_coords$0 = 0, $im_start_coords$1 = 0, $j$013 = 0, $j$122 = 0, $j$214 = 0, $j$316 = 0, $j$420 = 0, $j$518 = 0, $offsetToData$025 = 0, $offsetToData$1 = 0;
 var $or$cond = 0, $or$cond3 = 0, $or$cond33 = 0, $point_count = 0, $table_len = 0, $tuple_coords$0 = 0, $tuple_coords$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp + 8|0;
 $table_len = sp + 4|0;
 $point_count = sp;
 $0 = ((($stream)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 676|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  HEAP32[$error>>2] = 0;
  $im_end_coords$1 = 0;$im_start_coords$1 = 0;$tuple_coords$1 = 0;
  _ft_mem_free($1,$tuple_coords$1);
  _ft_mem_free($1,$im_start_coords$1);
  _ft_mem_free($1,$im_end_coords$1);
  $140 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($140|0);
 }
 $5 = ((($face)) + 648|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0|0);
 if ($7) {
  HEAP32[$error>>2] = 0;
  $im_end_coords$1 = 0;$im_start_coords$1 = 0;$tuple_coords$1 = 0;
  _ft_mem_free($1,$tuple_coords$1);
  _ft_mem_free($1,$im_start_coords$1);
  _ft_mem_free($1,$im_end_coords$1);
  $140 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($140|0);
 }
 $8 = ((($face)) + 508|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (FUNCTION_TABLE_iiiii[$9 & 8191]($face,1668702578,$stream,$table_len)|0);
 HEAP32[$error>>2] = $10;
 $11 = ($10|0)==(0);
 if (!($11)) {
  HEAP32[$error>>2] = 0;
  $im_end_coords$1 = 0;$im_start_coords$1 = 0;$tuple_coords$1 = 0;
  _ft_mem_free($1,$tuple_coords$1);
  _ft_mem_free($1,$im_start_coords$1);
  _ft_mem_free($1,$im_end_coords$1);
  $140 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($140|0);
 }
 $12 = HEAP32[$table_len>>2]|0;
 $13 = (_FT_Stream_EnterFrame($stream,$12)|0);
 HEAP32[$error>>2] = $13;
 $14 = ($13|0)==(0);
 if (!($14)) {
  HEAP32[$error>>2] = 0;
  $im_end_coords$1 = 0;$im_start_coords$1 = 0;$tuple_coords$1 = 0;
  _ft_mem_free($1,$tuple_coords$1);
  _ft_mem_free($1,$im_start_coords$1);
  _ft_mem_free($1,$im_end_coords$1);
  $140 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($140|0);
 }
 $15 = ((($stream)) + 32|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = HEAP32[$stream>>2]|0;
 $18 = (($16) - ($17))|0;
 $19 = (_FT_Stream_GetULong($stream)|0);
 $20 = ($19|0)==(65536);
 if ($20) {
  $21 = HEAP32[$3>>2]|0;
  $22 = (_ft_mem_realloc($1,4,0,$21,0,$error)|0);
  $23 = HEAP32[$error>>2]|0;
  $24 = ($23|0)==(0);
  if ($24) {
   $25 = HEAP32[$3>>2]|0;
   $26 = (_ft_mem_realloc($1,4,0,$25,0,$error)|0);
   $27 = HEAP32[$error>>2]|0;
   $28 = ($27|0)==(0);
   if ($28) {
    $29 = HEAP32[$3>>2]|0;
    $30 = (_ft_mem_realloc($1,4,0,$29,0,$error)|0);
    $31 = HEAP32[$error>>2]|0;
    $32 = ($31|0)==(0);
    if ($32) {
     $33 = (_FT_Stream_GetUShort($stream)|0);
     $34 = (_FT_Stream_GetUShort($stream)|0);
     $35 = $33 & 4095;
     $36 = ($35<<16>>16)==(0);
     if ($36) {
      $im_end_coords$0 = $30;$im_start_coords$0 = $26;$tuple_coords$0 = $22;
     } else {
      $37 = $34&65535;
      $38 = (($18) + ($37))|0;
      $39 = ((($face)) + 644|0);
      $40 = $33&65535;
      $41 = $40 & 4095;
      $i$024 = 0;$offsetToData$025 = $38;
      while(1) {
       $42 = (_FT_Stream_GetUShort($stream)|0);
       $43 = $42&65535;
       $44 = (_FT_Stream_GetUShort($stream)|0);
       $45 = $44&65535;
       $46 = $45 & 32768;
       $47 = ($46|0)==(0);
       if ($47) {
        $57 = $45 & 16384;
        $58 = ($57|0)==(0);
        if (!($58)) {
         $59 = HEAP32[$3>>2]|0;
         $$mask = $59 & 2147483647;
         $60 = ($$mask|0)==(0);
         if (!($60)) {
          $j$122 = 0;
          while(1) {
           (_FT_Stream_GetUShort($stream)|0);
           $61 = (($j$122) + 1)|0;
           $62 = HEAP32[$3>>2]|0;
           $63 = $62 << 1;
           $64 = ($61>>>0)<($63>>>0);
           if ($64) {
            $j$122 = $61;
           } else {
            break;
           }
          }
         }
        }
       } else {
        $48 = HEAP32[$3>>2]|0;
        $49 = ($48|0)==(0);
        if ($49) {
         $68 = 0;
        } else {
         $j$013 = 0;
         while(1) {
          $50 = (_FT_Stream_GetUShort($stream)|0);
          $51 = $50 << 16 >> 16;
          $52 = $51 << 2;
          $53 = (($22) + ($j$013<<2)|0);
          HEAP32[$53>>2] = $52;
          $54 = (($j$013) + 1)|0;
          $55 = HEAP32[$3>>2]|0;
          $56 = ($54>>>0)<($55>>>0);
          if ($56) {
           $j$013 = $54;
          } else {
           $68 = $55;
           break;
          }
         }
        }
        $65 = $45 & 16384;
        $66 = ($65|0)==(0);
        $67 = ($68|0)==(0);
        $or$cond33 = $66 | $67;
        if (!($or$cond33)) {
         $j$214 = 0;
         while(1) {
          $70 = (_FT_Stream_GetUShort($stream)|0);
          $71 = $70 << 16 >> 16;
          $72 = $71 << 2;
          $73 = (($26) + ($j$214<<2)|0);
          HEAP32[$73>>2] = $72;
          $74 = (($j$214) + 1)|0;
          $75 = HEAP32[$3>>2]|0;
          $76 = ($74>>>0)<($75>>>0);
          if ($76) {
           $j$214 = $74;
          } else {
           $$lcssa38 = $75;
           break;
          }
         }
         $69 = ($$lcssa38|0)==(0);
         if (!($69)) {
          $j$316 = 0;
          while(1) {
           $77 = (_FT_Stream_GetUShort($stream)|0);
           $78 = $77 << 16 >> 16;
           $79 = $78 << 2;
           $80 = (($30) + ($j$316<<2)|0);
           HEAP32[$80>>2] = $79;
           $81 = (($j$316) + 1)|0;
           $82 = HEAP32[$3>>2]|0;
           $83 = ($81>>>0)<($82>>>0);
           if ($83) {
            $j$316 = $81;
           } else {
            break;
           }
          }
         }
        }
        $84 = (_ft_var_apply_tuple($3,$44,$22,$26,$30)|0);
        $85 = ($84|0)==(0);
        $86 = $45 & 8192;
        $87 = ($86|0)==(0);
        $or$cond3 = $87 | $85;
        if (!($or$cond3)) {
         $88 = HEAP32[$15>>2]|0;
         $89 = HEAP32[$stream>>2]|0;
         $90 = $89;
         $91 = (($88) - ($90))|0;
         $92 = (($89) + ($offsetToData$025)|0);
         HEAP32[$15>>2] = $92;
         $93 = (_ft_var_readpackedpoints($stream,$point_count)|0);
         $94 = HEAP32[$point_count>>2]|0;
         $95 = ($94|0)==(0);
         if ($95) {
          $96 = HEAP32[$39>>2]|0;
          $97 = $96;
         } else {
          $97 = $94;
         }
         $98 = (_ft_var_readpackeddeltas($stream,$97)|0);
         $99 = ($93|0)==(0|0);
         $100 = ($98|0)==(0|0);
         $or$cond = $99 | $100;
         do {
          if ($or$cond) {
           label = 37;
          } else {
           $101 = ($93|0)==((-1)|0);
           if ($101) {
            $102 = HEAP32[$39>>2]|0;
            $103 = ($102|0)==(0);
            if ($103) {
             break;
            }
            $$pre29 = HEAP32[$5>>2]|0;
            $105 = $$pre29;$j$420 = 0;
            while(1) {
             $104 = (($105) + ($j$420<<1)|0);
             $106 = HEAP16[$104>>1]|0;
             $107 = $106&65535;
             $108 = (($98) + ($j$420<<1)|0);
             $109 = HEAP16[$108>>1]|0;
             $110 = $109 << 16 >> 16;
             $111 = (_FT_MulFix($110,$84)|0);
             $112 = (($111) + ($107))|0;
             $113 = $112&65535;
             $114 = HEAP32[$5>>2]|0;
             $115 = (($114) + ($j$420<<1)|0);
             HEAP16[$115>>1] = $113;
             $116 = (($j$420) + 1)|0;
             $117 = HEAP32[$39>>2]|0;
             $118 = ($116>>>0)<($117>>>0);
             if ($118) {
              $105 = $114;$j$420 = $116;
             } else {
              label = 37;
              break;
             }
            }
           } else {
            if ($95) {
             label = 38;
             break;
            }
            $$pre = HEAP32[$5>>2]|0;
            $123 = $$pre;$j$518 = 0;
            while(1) {
             $119 = (($93) + ($j$518<<1)|0);
             $120 = HEAP16[$119>>1]|0;
             $121 = $120&65535;
             $122 = (($123) + ($121<<1)|0);
             $124 = HEAP16[$122>>1]|0;
             $125 = $124&65535;
             $126 = (($98) + ($j$518<<1)|0);
             $127 = HEAP16[$126>>1]|0;
             $128 = $127 << 16 >> 16;
             $129 = (_FT_MulFix($128,$84)|0);
             $130 = (($129) + ($125))|0;
             $131 = $130&65535;
             $132 = HEAP32[$5>>2]|0;
             $133 = (($132) + ($121<<1)|0);
             HEAP16[$133>>1] = $131;
             $134 = (($j$518) + 1)|0;
             $135 = ($134>>>0)<($94>>>0);
             if ($135) {
              $123 = $132;$j$518 = $134;
             } else {
              label = 37;
              break;
             }
            }
           }
          }
         } while(0);
         if ((label|0) == 37) {
          label = 0;
          $136 = ($93|0)==((-1)|0);
          if (!($136)) {
           label = 38;
          }
         }
         if ((label|0) == 38) {
          label = 0;
          _ft_mem_free($1,$93);
         }
         _ft_mem_free($1,$98);
         $137 = HEAP32[$stream>>2]|0;
         $138 = (($137) + ($91)|0);
         HEAP32[$15>>2] = $138;
        }
       }
       $offsetToData$1 = (($43) + ($offsetToData$025))|0;
       $139 = (($i$024) + 1)|0;
       $exitcond = ($139|0)==($41|0);
       if ($exitcond) {
        $im_end_coords$0 = $30;$im_start_coords$0 = $26;$tuple_coords$0 = $22;
        break;
       } else {
        $i$024 = $139;$offsetToData$025 = $offsetToData$1;
       }
      }
     }
    } else {
     $im_end_coords$0 = $30;$im_start_coords$0 = $26;$tuple_coords$0 = $22;
    }
   } else {
    $im_end_coords$0 = 0;$im_start_coords$0 = $26;$tuple_coords$0 = $22;
   }
  } else {
   $im_end_coords$0 = 0;$im_start_coords$0 = 0;$tuple_coords$0 = $22;
  }
 } else {
  HEAP32[$error>>2] = 0;
  $im_end_coords$0 = 0;$im_start_coords$0 = 0;$tuple_coords$0 = 0;
 }
 _FT_Stream_ExitFrame($stream);
 $im_end_coords$1 = $im_end_coords$0;$im_start_coords$1 = $im_start_coords$0;$tuple_coords$1 = $tuple_coords$0;
 _ft_mem_free($1,$tuple_coords$1);
 _ft_mem_free($1,$im_start_coords$1);
 _ft_mem_free($1,$im_end_coords$1);
 $140 = HEAP32[$error>>2]|0;
 STACKTOP = sp;return ($140|0);
}
function _ft_var_apply_tuple($blend,$tupleIndex,$tuple_coords,$im_start_coords,$im_end_coords) {
 $blend = $blend|0;
 $tupleIndex = $tupleIndex|0;
 $tuple_coords = $tuple_coords|0;
 $im_start_coords = $im_start_coords|0;
 $im_end_coords = $im_end_coords|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $apply$03 = 0, $apply$03$us = 0, $apply$1 = 0, $apply$1$us = 0, $apply$2 = 0, $i$02 = 0, $i$02$us = 0, $or$cond = 0, $or$cond$us = 0;
 var $or$cond1 = 0, $or$cond1$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($blend)) + 4|0);
 $1 = HEAP32[$blend>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  $apply$2 = 65536;
  return ($apply$2|0);
 }
 $3 = $tupleIndex & 16384;
 $4 = ($3<<16>>16)==(0);
 if ($4) {
  $49 = $1;$apply$03$us = 65536;$i$02$us = 0;
  while(1) {
   $5 = (($tuple_coords) + ($i$02$us<<2)|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($6|0)==(0);
   if ($7) {
    $21 = $49;$apply$1$us = $apply$03$us;
   } else {
    $8 = HEAP32[$0>>2]|0;
    $9 = (($8) + ($i$02$us<<2)|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = ($10|0)==(0);
    if ($11) {
     $apply$2 = 0;
     label = 19;
     break;
    }
    $12 = ($10|0)<(0);
    $13 = ($6|0)>(0);
    $or$cond$us = $13 & $12;
    if ($or$cond$us) {
     $apply$2 = 0;
     label = 19;
     break;
    }
    $14 = ($10|0)>(0);
    $15 = ($6|0)<(0);
    $or$cond1$us = $15 & $14;
    if ($or$cond1$us) {
     $apply$2 = 0;
     label = 19;
     break;
    }
    $16 = (0 - ($10))|0;
    $17 = $14 ? $10 : $16;
    $18 = (_FT_MulFix($apply$03$us,$17)|0);
    $$pre = HEAP32[$blend>>2]|0;
    $21 = $$pre;$apply$1$us = $18;
   }
   $19 = (($i$02$us) + 1)|0;
   $20 = ($19>>>0)<($21>>>0);
   if ($20) {
    $49 = $21;$apply$03$us = $apply$1$us;$i$02$us = $19;
   } else {
    $apply$2 = $apply$1$us;
    label = 19;
    break;
   }
  }
  if ((label|0) == 19) {
   return ($apply$2|0);
  }
 } else {
  $apply$03 = 65536;$i$02 = 0;
 }
 L18: while(1) {
  $22 = (($tuple_coords) + ($i$02<<2)|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = ($23|0)==(0);
  do {
   if ($24) {
    $apply$1 = $apply$03;
   } else {
    $25 = HEAP32[$0>>2]|0;
    $26 = (($25) + ($i$02<<2)|0);
    $27 = HEAP32[$26>>2]|0;
    $28 = ($27|0)==(0);
    if ($28) {
     $apply$2 = 0;
     label = 19;
     break L18;
    }
    $29 = ($27|0)<(0);
    $30 = ($23|0)>(0);
    $or$cond = $30 & $29;
    if ($or$cond) {
     $apply$2 = 0;
     label = 19;
     break L18;
    }
    $31 = ($27|0)>(0);
    $32 = ($23|0)<(0);
    $or$cond1 = $32 & $31;
    if ($or$cond1) {
     $apply$2 = 0;
     label = 19;
     break L18;
    }
    $33 = (($im_start_coords) + ($i$02<<2)|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = ($27|0)<($34|0);
    if ($35) {
     $apply$2 = 0;
     label = 19;
     break L18;
    }
    $36 = (($im_end_coords) + ($i$02<<2)|0);
    $37 = HEAP32[$36>>2]|0;
    $38 = ($37|0)<($27|0);
    if ($38) {
     $apply$2 = 0;
     label = 19;
     break L18;
    }
    $39 = ($27|0)<($23|0);
    if ($39) {
     $40 = (($27) - ($34))|0;
     $41 = (($23) - ($34))|0;
     $42 = (_FT_MulDiv($apply$03,$40,$41)|0);
     $apply$1 = $42;
     break;
    } else {
     $43 = (($37) - ($27))|0;
     $44 = (($37) - ($23))|0;
     $45 = (_FT_MulDiv($apply$03,$43,$44)|0);
     $apply$1 = $45;
     break;
    }
   }
  } while(0);
  $46 = (($i$02) + 1)|0;
  $47 = HEAP32[$blend>>2]|0;
  $48 = ($46>>>0)<($47>>>0);
  if ($48) {
   $apply$03 = $apply$1;$i$02 = $46;
  } else {
   $apply$2 = $apply$1;
   label = 19;
   break;
  }
 }
 if ((label|0) == 19) {
  return ($apply$2|0);
 }
 return (0)|0;
}
function _ft_var_readpackedpoints($stream,$point_cnt) {
 $stream = $stream|0;
 $point_cnt = $point_cnt|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $exitcond = 0, $exitcond21 = 0, $first$09 = 0, $first$112 = 0;
 var $i$0$be = 0, $i$015 = 0, $i$17 = 0, $i$210 = 0, $j$08 = 0, $j$111 = 0, $n$0 = 0, $or$cond = 0, $or$cond3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($stream)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$error>>2] = 0;
 HEAP32[$point_cnt>>2] = 0;
 $2 = (_FT_Stream_GetChar($stream)|0);
 $3 = $2&255;
 $4 = ($2<<24>>24)==(0);
 if ($4) {
  $$0 = (-1);
  STACKTOP = sp;return ($$0|0);
 }
 $5 = $3 & 128;
 $6 = ($5|0)==(0);
 if ($6) {
  $n$0 = $3;
 } else {
  $7 = $3 << 8;
  $8 = $7 & 32512;
  $9 = (_FT_Stream_GetChar($stream)|0);
  $10 = $9&255;
  $11 = $10 | $8;
  $n$0 = $11;
 }
 $12 = (_ft_mem_realloc($1,2,0,$n$0,0,$error)|0);
 $13 = HEAP32[$error>>2]|0;
 $14 = ($13|0)==(0);
 if (!($14)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$point_cnt>>2] = $n$0;
 $15 = ($n$0|0)==(0);
 if ($15) {
  $$0 = $12;
  STACKTOP = sp;return ($$0|0);
 } else {
  $i$015 = 0;
 }
 while(1) {
  $20 = (_FT_Stream_GetChar($stream)|0);
  $21 = $20&255;
  $22 = $21 & 128;
  $23 = ($22|0)==(0);
  if ($23) {
   $39 = (_FT_Stream_GetChar($stream)|0);
   $40 = $39&255;
   $41 = (($i$015) + 1)|0;
   $42 = (($12) + ($i$015<<1)|0);
   HEAP16[$42>>1] = $40;
   $43 = ($20<<24>>24)==(0);
   $44 = (($21) + ($41))|0;
   $45 = ($44>>>0)>($n$0>>>0);
   $or$cond3 = $43 | $45;
   if ($or$cond3) {
    $$0 = $12;
    label = 14;
    break;
   } else {
    $first$112 = $40;$i$210 = $41;$j$111 = 0;
   }
   while(1) {
    $46 = (_FT_Stream_GetChar($stream)|0);
    $47 = $46&255;
    $48 = $first$112&65535;
    $49 = (($47) + ($48))|0;
    $50 = $49&65535;
    $51 = (($i$210) + 1)|0;
    $52 = (($12) + ($i$210<<1)|0);
    HEAP16[$52>>1] = $50;
    $53 = (($j$111) + 1)|0;
    $exitcond21 = ($53|0)==($21|0);
    if ($exitcond21) {
     $i$0$be = $44;
     break;
    } else {
     $first$112 = $50;$i$210 = $51;$j$111 = $53;
    }
   }
  } else {
   $24 = $21 & 127;
   $25 = (_FT_Stream_GetUShort($stream)|0);
   $17 = (($i$015) + 1)|0;
   $26 = (($12) + ($i$015<<1)|0);
   HEAP16[$26>>1] = $25;
   $27 = ($24|0)==(0);
   $28 = (($24) + ($17))|0;
   $29 = ($28>>>0)>($n$0>>>0);
   $or$cond = $27 | $29;
   if ($or$cond) {
    $$0 = $12;
    label = 14;
    break;
   }
   $30 = $20&255;
   $18 = $30 & 127;
   $first$09 = $25;$i$17 = $17;$j$08 = 0;
   while(1) {
    $31 = (_FT_Stream_GetUShort($stream)|0);
    $32 = $31&65535;
    $33 = $first$09&65535;
    $34 = (($32) + ($33))|0;
    $35 = $34&65535;
    $36 = (($i$17) + 1)|0;
    $37 = (($12) + ($i$17<<1)|0);
    HEAP16[$37>>1] = $35;
    $38 = (($j$08) + 1)|0;
    $exitcond = ($38|0)==($18|0);
    if ($exitcond) {
     break;
    } else {
     $first$09 = $35;$i$17 = $36;$j$08 = $38;
    }
   }
   $16 = (($17) + ($18))|0;
   $i$0$be = $16;
  }
  $19 = ($i$0$be>>>0)<($n$0>>>0);
  if ($19) {
   $i$015 = $i$0$be;
  } else {
   $$0 = $12;
   label = 14;
   break;
  }
 }
 if ((label|0) == 14) {
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _ft_var_readpackeddeltas($stream,$delta_cnt) {
 $stream = $stream|0;
 $delta_cnt = $delta_cnt|0;
 var $$0 = 0, $$not = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0;
 var $i$0 = 0, $i$111 = 0, $i$215 = 0, $i$321 = 0, $i$4 = 0, $j$012 = 0, $j$116 = 0, $j$222 = 0, $j$3 = 0, $or$cond = 0, $or$cond3 = 0, $or$cond4 = 0, $scevgep = 0, $umax = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($stream)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$error>>2] = 0;
 $2 = (_ft_mem_realloc($1,2,0,$delta_cnt,0,$error)|0);
 $3 = HEAP32[$error>>2]|0;
 $4 = ($3|0)==(0);
 if ($4) {
  $i$0 = 0;
 } else {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 while(1) {
  $5 = ($i$0>>>0)<($delta_cnt>>>0);
  if (!($5)) {
   $$0 = $2;
   label = 11;
   break;
  }
  $6 = (_FT_Stream_GetChar($stream)|0);
  $7 = $6&255;
  $8 = $7 & 63;
  $9 = $7 & 128;
  $10 = ($9|0)==(0);
  if ($10) {
   $21 = $7 & 64;
   $22 = ($21|0)==(0);
   if ($22) {
    $i$321 = $i$0;$j$222 = 0;
    while(1) {
     $29 = (_FT_Stream_GetChar($stream)|0);
     $30 = $29 << 24 >> 24;
     $31 = (($i$321) + 1)|0;
     $32 = (($2) + ($i$321<<1)|0);
     HEAP16[$32>>1] = $30;
     $33 = (($j$222) + 1)|0;
     $34 = ($33>>>0)<=($8>>>0);
     $35 = ($31>>>0)<($delta_cnt>>>0);
     $or$cond4 = $35 & $34;
     if ($or$cond4) {
      $i$321 = $31;$j$222 = $33;
     } else {
      $i$4 = $31;$j$3 = $33;
      break;
     }
    }
   } else {
    $i$215 = $i$0;$j$116 = 0;
    while(1) {
     $23 = (_FT_Stream_GetUShort($stream)|0);
     $24 = (($i$215) + 1)|0;
     $25 = (($2) + ($i$215<<1)|0);
     HEAP16[$25>>1] = $23;
     $26 = (($j$116) + 1)|0;
     $27 = ($26>>>0)<=($8>>>0);
     $28 = ($24>>>0)<($delta_cnt>>>0);
     $or$cond3 = $28 & $27;
     if ($or$cond3) {
      $i$215 = $24;$j$116 = $26;
     } else {
      $i$4 = $24;$j$3 = $26;
      break;
     }
    }
   }
  } else {
   $11 = $i$0 << 1;
   $scevgep = (($2) + ($11)|0);
   $12 = (($i$0) - ($delta_cnt))|0;
   $13 = $6&255;
   $$not = $13 | -64;
   $14 = $$not ^ 63;
   $15 = ($12>>>0)>($14>>>0);
   $umax = $15 ? $12 : $14;
   $16 = Math_imul($umax, -2)|0;
   _memset(($scevgep|0),0,($16|0))|0;
   $i$111 = $i$0;$j$012 = 0;
   while(1) {
    $17 = (($i$111) + 1)|0;
    $18 = (($j$012) + 1)|0;
    $19 = ($18>>>0)<=($8>>>0);
    $20 = ($17>>>0)<($delta_cnt>>>0);
    $or$cond = $20 & $19;
    if ($or$cond) {
     $i$111 = $17;$j$012 = $18;
    } else {
     $i$4 = $17;$j$3 = $18;
     break;
    }
   }
  }
  $36 = ($j$3>>>0)>($8>>>0);
  if ($36) {
   $i$0 = $i$4;
  } else {
   break;
  }
 }
 if ((label|0) == 11) {
  STACKTOP = sp;return ($$0|0);
 }
 _ft_mem_free($1,$2);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _TT_Set_Var_Design($face,$num_coords,$coords) {
 $face = $face|0;
 $num_coords = $num_coords|0;
 $coords = $coords|0;
 var $$lcssa = 0, $$lcssa50 = 0, $$num_coords = 0, $$pre = 0, $$pre$i = 0, $$pre13$i = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $a$018 = 0, $av$014 = 0;
 var $error = 0, $error$i = 0, $i$0$lcssa = 0, $i$017 = 0, $i$09$i = 0, $i$09$i$lcssa = 0, $i$213 = 0, $j$012 = 0, $j$012$lcssa = 0, $j$04$i = 0, $j$04$in$i = 0, $j$15$i = 0, $normalized$0 = 0, $or$cond$i = 0, $scevgep = 0, $segment$08$i = 0, $table_len$i = 0, $umax = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp + 8|0;
 $table_len$i = sp + 4|0;
 $error = sp;
 HEAP32[$error>>2] = 0;
 $0 = ((($face)) + 100|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 676|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 do {
  if ($4) {
   $5 = (_TT_Get_MM_Var($face,0)|0);
   HEAP32[$error>>2] = $5;
   $6 = ($5|0)==(0);
   if ($6) {
    $$pre = HEAP32[$2>>2]|0;
    $8 = $$pre;
    break;
   } else {
    $normalized$0 = 0;
    _ft_mem_free($1,$normalized$0);
    $148 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($148|0);
   }
  } else {
   $8 = $3;
  }
 } while(0);
 $7 = ((($8)) + 8|0);
 $9 = HEAP32[$7>>2]|0;
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10>>>0)<($num_coords>>>0);
 $$num_coords = $11 ? $10 : $num_coords;
 $12 = (_ft_mem_realloc($1,4,0,$10,0,$error)|0);
 $13 = HEAP32[$error>>2]|0;
 $14 = ($13|0)==(0);
 if (!($14)) {
  $normalized$0 = $12;
  _ft_mem_free($1,$normalized$0);
  $148 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($148|0);
 }
 $15 = ($$num_coords|0)==(0);
 L10: do {
  if ($15) {
   $i$0$lcssa = 0;
  } else {
   $16 = ((($9)) + 12|0);
   $17 = HEAP32[$16>>2]|0;
   $a$018 = $17;$i$017 = 0;
   while(1) {
    $25 = (($coords) + ($i$017<<2)|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($a$018)) + 12|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = ($26|0)>($28|0);
    if ($29) {
     break;
    }
    $30 = ((($a$018)) + 4|0);
    $31 = HEAP32[$30>>2]|0;
    $32 = ($26|0)<($31|0);
    if ($32) {
     break;
    }
    $33 = ((($a$018)) + 8|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = ($26|0)<($34|0);
    do {
     if ($35) {
      $36 = (($26) - ($34))|0;
      $37 = (($31) - ($34))|0;
      $38 = (_FT_DivFix($36,$37)|0);
      $39 = (0 - ($38))|0;
      $40 = (($12) + ($i$017<<2)|0);
      HEAP32[$40>>2] = $39;
     } else {
      $41 = ($28|0)==($34|0);
      if ($41) {
       $42 = (($12) + ($i$017<<2)|0);
       HEAP32[$42>>2] = 0;
       break;
      } else {
       $43 = (($26) - ($34))|0;
       $44 = (($28) - ($34))|0;
       $45 = (_FT_DivFix($43,$44)|0);
       $46 = (($12) + ($i$017<<2)|0);
       HEAP32[$46>>2] = $45;
       break;
      }
     }
    } while(0);
    $47 = (($i$017) + 1)|0;
    $48 = ((($a$018)) + 24|0);
    $49 = ($47>>>0)<($$num_coords>>>0);
    if ($49) {
     $a$018 = $48;$i$017 = $47;
    } else {
     $i$0$lcssa = $47;
     break L10;
    }
   }
   HEAP32[$error>>2] = 6;
   $normalized$0 = $12;
   _ft_mem_free($1,$normalized$0);
   $148 = HEAP32[$error>>2]|0;
   STACKTOP = sp;return ($148|0);
  }
 } while(0);
 $18 = HEAP32[$9>>2]|0;
 $19 = ($i$0$lcssa>>>0)<($18>>>0);
 if ($19) {
  $20 = $i$0$lcssa << 2;
  $scevgep = (($12) + ($20)|0);
  $21 = (($i$0$lcssa) + 1)|0;
  $22 = ($18>>>0)>($21>>>0);
  $umax = $22 ? $18 : $21;
  $23 = $umax << 2;
  $24 = (($23) - ($20))|0;
  _memset(($scevgep|0),0,($24|0))|0;
 }
 $50 = ((($8)) + 16|0);
 $51 = HEAP8[$50>>0]|0;
 $52 = ($51<<24>>24)==(0);
 if ($52) {
  $53 = ((($face)) + 104|0);
  $54 = HEAP32[$53>>2]|0;
  $55 = ((($54)) + 28|0);
  $56 = HEAP32[$55>>2]|0;
  $57 = HEAP32[$2>>2]|0;
  $58 = ((($57)) + 16|0);
  HEAP8[$58>>0] = 1;
  $59 = ((($face)) + 508|0);
  $60 = HEAP32[$59>>2]|0;
  $61 = (FUNCTION_TABLE_iiiii[$60 & 8191]($face,1635148146,$54,$table_len$i)|0);
  HEAP32[$error$i>>2] = $61;
  $62 = ($61|0)==(0);
  if ($62) {
   $63 = HEAP32[$table_len$i>>2]|0;
   $64 = (_FT_Stream_EnterFrame($54,$63)|0);
   HEAP32[$error$i>>2] = $64;
   $65 = ($64|0)==(0);
   if ($65) {
    $66 = (_FT_Stream_GetULong($54)|0);
    $67 = (_FT_Stream_GetULong($54)|0);
    $68 = ($66|0)==(65536);
    L34: do {
     if ($68) {
      $69 = ((($57)) + 8|0);
      $70 = HEAP32[$69>>2]|0;
      $71 = HEAP32[$70>>2]|0;
      $72 = ($67|0)==($71|0);
      if ($72) {
       $73 = (_ft_mem_realloc($56,8,0,$67,0,$error$i)|0);
       $74 = ((($57)) + 20|0);
       HEAP32[$74>>2] = $73;
       $75 = HEAP32[$error$i>>2]|0;
       $76 = ($75|0)==(0);
       $77 = ($67|0)>(0);
       $or$cond$i = $77 & $76;
       if ($or$cond$i) {
        $i$09$i = 0;$segment$08$i = $73;
        while(1) {
         $78 = (_FT_Stream_GetUShort($54)|0);
         HEAP16[$segment$08$i>>1] = $78;
         $79 = $78&65535;
         $80 = (_ft_mem_realloc($56,8,0,$79,0,$error$i)|0);
         $81 = ((($segment$08$i)) + 4|0);
         HEAP32[$81>>2] = $80;
         $82 = HEAP32[$error$i>>2]|0;
         $83 = ($82|0)==(0);
         if (!($83)) {
          $i$09$i$lcssa = $i$09$i;
          break;
         }
         $84 = HEAP16[$segment$08$i>>1]|0;
         $85 = ($84<<16>>16)==(0);
         if (!($85)) {
          $j$15$i = 0;
          while(1) {
           $94 = (_FT_Stream_GetUShort($54)|0);
           $95 = $94 << 16 >> 16;
           $96 = $95 << 2;
           $97 = HEAP32[$81>>2]|0;
           $98 = (($97) + ($j$15$i<<3)|0);
           HEAP32[$98>>2] = $96;
           $99 = (_FT_Stream_GetUShort($54)|0);
           $100 = $99 << 16 >> 16;
           $101 = $100 << 2;
           $102 = HEAP32[$81>>2]|0;
           $103 = (((($102) + ($j$15$i<<3)|0)) + 4|0);
           HEAP32[$103>>2] = $101;
           $104 = (($j$15$i) + 1)|0;
           $105 = HEAP16[$segment$08$i>>1]|0;
           $106 = $105&65535;
           $107 = ($104|0)<($106|0);
           if ($107) {
            $j$15$i = $104;
           } else {
            break;
           }
          }
         }
         $108 = (($i$09$i) + 1)|0;
         $109 = ((($segment$08$i)) + 8|0);
         $110 = ($108|0)<($67|0);
         if ($110) {
          $i$09$i = $108;$segment$08$i = $109;
         } else {
          break L34;
         }
        }
        $86 = ($i$09$i$lcssa|0)>(0);
        if ($86) {
         $$pre$i = HEAP32[$74>>2]|0;
         $88 = $$pre$i;$j$04$in$i = $i$09$i$lcssa;
         while(1) {
          $j$04$i = (($j$04$in$i) + -1)|0;
          $87 = (((($88) + ($j$04$i<<3)|0)) + 4|0);
          $89 = HEAP32[$87>>2]|0;
          _ft_mem_free($56,$89);
          $90 = HEAP32[$74>>2]|0;
          $91 = (((($90) + ($j$04$i<<3)|0)) + 4|0);
          HEAP32[$91>>2] = 0;
          $92 = ($j$04$in$i|0)>(1);
          if ($92) {
           $88 = $90;$j$04$in$i = $j$04$i;
          } else {
           $$lcssa50 = $90;
           break;
          }
         }
         $93 = $$lcssa50;
        } else {
         $$pre13$i = HEAP32[$74>>2]|0;
         $93 = $$pre13$i;
        }
        _ft_mem_free($56,$93);
        HEAP32[$74>>2] = 0;
       }
      }
     }
    } while(0);
    _FT_Stream_ExitFrame($54);
   }
  }
 }
 $111 = ((($8)) + 20|0);
 $112 = HEAP32[$111>>2]|0;
 $113 = ($112|0)==(0|0);
 $$pre28 = HEAP32[$9>>2]|0;
 if ($113) {
  $146 = $$pre28;
 } else {
  $114 = ($$pre28|0)==(0);
  if ($114) {
   $146 = 0;
  } else {
   $149 = $$pre28;$av$014 = $112;$i$213 = 0;
   while(1) {
    $115 = HEAP16[$av$014>>1]|0;
    $116 = ($115&65535)>(1);
    L60: do {
     if ($116) {
      $117 = (($12) + ($i$213<<2)|0);
      $118 = HEAP32[$117>>2]|0;
      $119 = ((($av$014)) + 4|0);
      $120 = HEAP32[$119>>2]|0;
      $121 = $115&65535;
      $j$012 = 1;
      while(1) {
       $122 = (($120) + ($j$012<<3)|0);
       $123 = HEAP32[$122>>2]|0;
       $124 = ($118|0)<($123|0);
       if ($124) {
        $$lcssa = $123;$j$012$lcssa = $j$012;
        break;
       }
       $140 = (($j$012) + 1)|0;
       $141 = ($140>>>0)<($121>>>0);
       if ($141) {
        $j$012 = $140;
       } else {
        $145 = $149;
        break L60;
       }
      }
      $125 = (($j$012$lcssa) + -1)|0;
      $126 = (($120) + ($125<<3)|0);
      $127 = HEAP32[$126>>2]|0;
      $128 = (($118) - ($127))|0;
      $129 = (((($120) + ($j$012$lcssa<<3)|0)) + 4|0);
      $130 = HEAP32[$129>>2]|0;
      $131 = (((($120) + ($125<<3)|0)) + 4|0);
      $132 = HEAP32[$131>>2]|0;
      $133 = (($130) - ($132))|0;
      $134 = (($$lcssa) - ($127))|0;
      $135 = (_FT_MulDiv($128,$133,$134)|0);
      $136 = HEAP32[$119>>2]|0;
      $137 = (((($136) + ($125<<3)|0)) + 4|0);
      $138 = HEAP32[$137>>2]|0;
      $139 = (($138) + ($135))|0;
      HEAP32[$117>>2] = $139;
      $$pre27 = HEAP32[$9>>2]|0;
      $145 = $$pre27;
     } else {
      $145 = $149;
     }
    } while(0);
    $142 = (($i$213) + 1)|0;
    $143 = ((($av$014)) + 8|0);
    $144 = ($142>>>0)<($145>>>0);
    if ($144) {
     $149 = $145;$av$014 = $143;$i$213 = $142;
    } else {
     $146 = $145;
     break;
    }
   }
  }
 }
 $147 = (_TT_Set_MM_Blend($face,$146,$12)|0);
 HEAP32[$error>>2] = $147;
 $normalized$0 = $12;
 _ft_mem_free($1,$normalized$0);
 $148 = HEAP32[$error>>2]|0;
 STACKTOP = sp;return ($148|0);
}
function _tt_face_get_location($face,$gindex,$asize) {
 $face = $face|0;
 $gindex = $gindex|0;
 $asize = $asize|0;
 var $$$pos2$1 = 0, $$0 = 0, $$pos2$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $9 = 0, $pos1$0 = 0, $pos1$02 = 0, $pos2$0 = 0, $pos2$1 = 0, $pos2$13 = 0, $storemerge = 0, $storemerge1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 696|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1>>>0)>($gindex>>>0);
 if ($2) {
  $5 = ((($face)) + 210|0);
  $6 = HEAP16[$5>>1]|0;
  $7 = ($6<<16>>16)==(0);
  $8 = ((($face)) + 700|0);
  $9 = HEAP32[$8>>2]|0;
  if ($7) {
   $53 = $gindex << 1;
   $54 = (($9) + ($53)|0);
   $55 = $1 << 1;
   $56 = (($9) + ($55)|0);
   $57 = ((($54)) + 2|0);
   $58 = ((($57)) + -2|0);
   $59 = HEAP8[$58>>0]|0;
   $60 = $59&255;
   $61 = $60 << 8;
   $62 = ((($58)) + 1|0);
   $63 = HEAP8[$62>>0]|0;
   $64 = $63&255;
   $65 = $61 | $64;
   $66 = ((($57)) + 2|0);
   $67 = ($66>>>0)>($56>>>0);
   if ($67) {
    $pos2$0 = $65;
   } else {
    $68 = ((($66)) + -2|0);
    $69 = HEAP8[$68>>0]|0;
    $70 = $69&255;
    $71 = $70 << 8;
    $72 = ((($68)) + 1|0);
    $73 = HEAP8[$72>>0]|0;
    $74 = $73&255;
    $75 = $71 | $74;
    $pos2$0 = $75;
   }
   $76 = $65 << 1;
   $77 = $pos2$0 << 1;
   $pos1$0 = $76;$pos2$1 = $77;
  } else {
   $10 = $gindex << 2;
   $11 = (($9) + ($10)|0);
   $12 = $1 << 2;
   $13 = (($9) + ($12)|0);
   $14 = ((($11)) + 4|0);
   $15 = ((($14)) + -4|0);
   $16 = HEAP8[$15>>0]|0;
   $17 = $16&255;
   $18 = $17 << 24;
   $19 = ((($15)) + 1|0);
   $20 = HEAP8[$19>>0]|0;
   $21 = $20&255;
   $22 = $21 << 16;
   $23 = $22 | $18;
   $24 = ((($15)) + 2|0);
   $25 = HEAP8[$24>>0]|0;
   $26 = $25&255;
   $27 = $26 << 8;
   $28 = $23 | $27;
   $29 = ((($15)) + 3|0);
   $30 = HEAP8[$29>>0]|0;
   $31 = $30&255;
   $32 = $28 | $31;
   $33 = ((($14)) + 4|0);
   $34 = ($33>>>0)>($13>>>0);
   if ($34) {
    $pos1$0 = $32;$pos2$1 = $32;
   } else {
    $35 = ((($33)) + -4|0);
    $36 = HEAP8[$35>>0]|0;
    $37 = $36&255;
    $38 = $37 << 24;
    $39 = ((($35)) + 1|0);
    $40 = HEAP8[$39>>0]|0;
    $41 = $40&255;
    $42 = $41 << 16;
    $43 = $42 | $38;
    $44 = ((($35)) + 2|0);
    $45 = HEAP8[$44>>0]|0;
    $46 = $45&255;
    $47 = $46 << 8;
    $48 = $43 | $47;
    $49 = ((($35)) + 3|0);
    $50 = HEAP8[$49>>0]|0;
    $51 = $50&255;
    $52 = $48 | $51;
    $pos1$0 = $32;$pos2$1 = $52;
   }
  }
  $78 = ((($face)) + 668|0);
  $79 = HEAP32[$78>>2]|0;
  $80 = ($pos1$0>>>0)>($79>>>0);
  if ($80) {
   $$0 = 0;$storemerge1 = 0;
   HEAP32[$asize>>2] = $storemerge1;
   return ($$0|0);
  } else {
   $82 = $79;$pos1$02 = $pos1$0;$pos2$13 = $pos2$1;
  }
 } else {
  $3 = ((($face)) + 668|0);
  $4 = HEAP32[$3>>2]|0;
  $82 = $4;$pos1$02 = 0;$pos2$13 = 0;
 }
 $81 = ($pos2$13>>>0)>($82>>>0);
 $$pos2$1 = $81 ? $82 : $pos2$13;
 $83 = ($$pos2$1>>>0)<($pos1$02>>>0);
 $$$pos2$1 = $83 ? $82 : $$pos2$1;
 $storemerge = (($$$pos2$1) - ($pos1$02))|0;
 $$0 = $pos1$02;$storemerge1 = $storemerge;
 HEAP32[$asize>>2] = $storemerge1;
 return ($$0|0);
}
function _tt_property_set($module,$property_name,$value) {
 $module = $module|0;
 $property_name = $property_name|0;
 $value = $value|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strcmp($property_name,593148)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  $$0 = 12;
  return ($$0|0);
 }
 $2 = HEAP32[$value>>2]|0;
 $3 = ($2|0)==(35);
 if (!($3)) {
  $$0 = 7;
  return ($$0|0);
 }
 $4 = ((($module)) + 64|0);
 HEAP32[$4>>2] = 35;
 $$0 = 0;
 return ($$0|0);
}
function _tt_property_get($module,$property_name,$value) {
 $module = $module|0;
 $property_name = $property_name|0;
 $value = $value|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strcmp($property_name,593148)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  $$0 = 12;
  return ($$0|0);
 }
 $2 = ((($module)) + 64|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$value>>2] = $3;
 $$0 = 0;
 return ($$0|0);
}
function _tt_face_init($stream,$ttface,$face_index,$num_params,$params) {
 $stream = $stream|0;
 $ttface = $ttface|0;
 $face_index = $face_index|0;
 $num_params = $num_params|0;
 $params = $params|0;
 var $$$i = 0, $$0 = 0, $$01$lcssa$i$i$i$i = 0, $$016$i$i$i$i = 0, $$13$i$i$i$i = 0, $$lcssa = 0, $$phi$trans$insert32$i$i = 0, $$pr = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i6 = 0, $$pre$phi$iZ2D = 0, $$pre33$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0;
 var $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0;
 var $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0;
 var $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0;
 var $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0;
 var $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0;
 var $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0;
 var $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0;
 var $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0;
 var $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0;
 var $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0;
 var $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $asize$i = 0, $buf$i = 0, $checksum$0$lcssa$i$i$i$i = 0, $checksum$06$i$i = 0, $checksum$07$i$i$i$i = 0, $checksum$1$i$i = 0, $checksum$1$lcssa$i$i$i$i = 0, $checksum$14$i$i$i$i = 0, $checksum$2$i$i = 0, $count$06$i = 0, $count$1$i = 0, $count$1$i$lcssa = 0, $dist$0$lcssa$i = 0;
 var $dist$03$i = 0, $dist$1$i = 0, $dist$1$i$lcssa = 0, $dist$2$i = 0, $entry$0$lcssa$i = 0, $entry$02$i = 0, $error$0$i$ph = 0, $error$0$i10 = 0, $error$3 = 0, $error$4 = 0, $error$i = 0, $glyph_index$05$i = 0, $glyph_index$1$i = 0, $glyph_index$1$i$lcssa = 0, $has_cvt$09$i$i = 0, $has_cvt$1$i$i = 0, $has_cvt$2$i$i = 0, $has_cvt$2$i$i$lcssa = 0, $i$010$i$i = 0, $i$04$i = 0;
 var $i$05$i$i$i$i = 0, $j$07$i$i = 0, $j$15$i$i = 0, $k$0$i$i = 0, $nn$0$lcssa$i = 0, $nn$08$i = 0, $not$$i$i = 0, $num_matched_ids$i$i = 0, $or$cond$i = 0, $or$cond$i13 = 0, $or$cond$i2 = 0, $or$cond1$i = 0, $or$cond3$i = 0, $p$09$i = 0, $phitmp$i = 0, $record_size$0$off$i = 0, $scevgep5$i = 0, $scevgep8$i = 0, $shift$0$i = 0, $switch$split12D = 0;
 var $switch$split2D = 0, $switch$split42D = 0, $switch$split72D = 0, $table_len$i = 0, $table_len$i7 = 0, $table_len$i9 = 0, $table_size$i = 0, $uglygep$i = 0, $uglygep7$i = 0, $umax$i = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $asize$i = sp + 88|0;
 $buf$i = sp + 96|0;
 $table_len$i9 = sp + 84|0;
 $table_len$i7 = sp + 80|0;
 $table_len$i = sp + 76|0;
 $error$i = sp + 72|0;
 $table_size$i = sp + 68|0;
 $num_matched_ids$i$i = sp;
 $0 = ((($ttface)) + 96|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_FT_Get_Module_Interface($3,593168)|0);
 $5 = ($4|0)==(0|0);
 if ($5) {
  $$0 = 11;
  STACKTOP = sp;return ($$0|0);
 }
 $6 = (_FT_Stream_Seek($stream,0)|0);
 $7 = ($6|0)==(0);
 if (!($7)) {
  $$0 = $6;
  STACKTOP = sp;return ($$0|0);
 }
 $8 = ((($4)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (FUNCTION_TABLE_iiiiii[$9 & 8191]($stream,$ttface,$face_index,$num_params,$params)|0);
 $11 = ((($ttface)) + 104|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($10|0)==(0);
 if (!($13)) {
  $$0 = $10;
  STACKTOP = sp;return ($$0|0);
 }
 $14 = ((($ttface)) + 148|0);
 $15 = HEAP32[$14>>2]|0;
 $switch$split2D = ($15|0)<(131072);
 L10: do {
  if ($switch$split2D) {
   switch ($15|0) {
   case 65536:  {
    break L10;
    break;
   }
   default: {
    $$0 = 2;
   }
   }
   STACKTOP = sp;return ($$0|0);
  } else {
   $switch$split42D = ($15|0)<(1953658213);
   if ($switch$split42D) {
    switch ($15|0) {
    case 131072:  {
     break L10;
     break;
    }
    default: {
     $$0 = 2;
    }
    }
    STACKTOP = sp;return ($$0|0);
   } else {
    switch ($15|0) {
    case 1953658213:  {
     break L10;
     break;
    }
    default: {
     $$0 = 2;
    }
    }
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 $16 = ((($ttface)) + 8|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 | 2048;
 HEAP32[$16>>2] = $18;
 $19 = ($face_index|0)<(0);
 if ($19) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $20 = ((($4)) + 8|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = (FUNCTION_TABLE_iiiiii[$21 & 8191]($12,$ttface,$face_index,$num_params,$params)|0);
 $23 = ($22|0)==(0);
 if (!($23)) {
  $$0 = $22;
  STACKTOP = sp;return ($$0|0);
 }
 $24 = ((($ttface)) + 20|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = ($25|0)==(0|0);
 if ($26) {
  label = 17;
 } else {
  $27 = (_strstr($25,593173)|0);
  $28 = ($27|0)==(0|0);
  if ($28) {
   $29 = (_strstr($25,(593193))|0);
   $30 = ($29|0)==(0|0);
   if ($30) {
    $31 = (_strstr($25,(593213))|0);
    $32 = ($31|0)==(0|0);
    if ($32) {
     $33 = (_strstr($25,(593233))|0);
     $34 = ($33|0)==(0|0);
     if ($34) {
      $35 = (_strstr($25,(593253))|0);
      $36 = ($35|0)==(0|0);
      if ($36) {
       $37 = (_strstr($25,(593273))|0);
       $38 = ($37|0)==(0|0);
       if ($38) {
        $39 = (_strstr($25,(593293))|0);
        $40 = ($39|0)==(0|0);
        if ($40) {
         $41 = (_strstr($25,(593313))|0);
         $42 = ($41|0)==(0|0);
         if ($42) {
          $43 = (_strstr($25,(593333))|0);
          $not$$i$i = ($43|0)==(0|0);
          if ($not$$i$i) {
           label = 17;
          } else {
           label = 65;
          }
         } else {
          label = 65;
         }
        } else {
         label = 65;
        }
       } else {
        label = 65;
       }
      } else {
       label = 65;
      }
     } else {
      label = 65;
     }
    } else {
     label = 65;
    }
   } else {
    label = 65;
   }
  } else {
   label = 65;
  }
 }
 L36: do {
  if ((label|0) == 17) {
   dest=$num_matched_ids$i$i; stop=dest+68|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
   $44 = ((($ttface)) + 152|0);
   $45 = HEAP16[$44>>1]|0;
   $46 = ($45<<16>>16)==(0);
   L38: do {
    if ($46) {
     $j$15$i$i = 0;
     label = 46;
    } else {
     $47 = ((($ttface)) + 156|0);
     $48 = ((($ttface)) + 508|0);
     $353 = $45;$57 = 0;$has_cvt$09$i$i = 0;$i$010$i$i = 0;
     while(1) {
      $55 = HEAP32[$47>>2]|0;
      $56 = (($55) + ($57<<4)|0);
      $58 = HEAP32[$56>>2]|0;
      $switch$split12D = ($58|0)<(1718642541);
      L42: do {
       if ($switch$split12D) {
        switch ($58|0) {
        case 1668707360:  {
         $has_cvt$1$i$i = 1;$k$0$i$i = 0;
         label = 25;
         break;
        }
        default: {
         $107 = $353;$has_cvt$2$i$i = $has_cvt$09$i$i;
        }
        }
       } else {
        $switch$split72D = ($58|0)<(1886545264);
        if ($switch$split72D) {
         switch ($58|0) {
         case 1718642541:  {
          break;
         }
         default: {
          $107 = $353;$has_cvt$2$i$i = $has_cvt$09$i$i;
          break L42;
         }
         }
         $has_cvt$1$i$i = $has_cvt$09$i$i;$k$0$i$i = 1;
         label = 25;
         break;
        } else {
         switch ($58|0) {
         case 1886545264:  {
          break;
         }
         default: {
          $107 = $353;$has_cvt$2$i$i = $has_cvt$09$i$i;
          break L42;
         }
         }
         $has_cvt$1$i$i = $has_cvt$09$i$i;$k$0$i$i = 2;
         label = 25;
         break;
        }
       }
      } while(0);
      if ((label|0) == 25) {
       label = 0;
       $60 = $55;$checksum$06$i$i = 0;$j$07$i$i = 0;
       while(1) {
        $59 = (((($60) + ($57<<4)|0)) + 12|0);
        $61 = HEAP32[$59>>2]|0;
        $62 = ((((55776 + (($j$07$i$i*24)|0)|0) + ($k$0$i$i<<3)|0)) + 4|0);
        $63 = HEAP32[$62>>2]|0;
        $64 = ($61|0)==($63|0);
        if ($64) {
         $65 = ($checksum$06$i$i|0)==(0);
         do {
          if ($65) {
           $66 = HEAP32[$48>>2]|0;
           $67 = ($66|0)==(0|0);
           if ($67) {
            $checksum$1$i$i = 0;
           } else {
            $68 = (($60) + ($57<<4)|0);
            $69 = HEAP32[$68>>2]|0;
            $70 = HEAP32[$11>>2]|0;
            $71 = (FUNCTION_TABLE_iiiii[$66 & 8191]($ttface,$69,$70,0)|0);
            $72 = ($71|0)==(0);
            if ($72) {
             $73 = HEAP32[$11>>2]|0;
             $74 = HEAP32[$47>>2]|0;
             $75 = (((($74) + ($57<<4)|0)) + 12|0);
             $76 = HEAP32[$75>>2]|0;
             $77 = (_FT_Stream_EnterFrame($73,$76)|0);
             $78 = ($77|0)==(0);
             if (!($78)) {
              $checksum$1$i$i = 0;
              break;
             }
             $79 = ($76>>>0)>(3);
             if ($79) {
              $$016$i$i$i$i = $76;$checksum$07$i$i$i$i = 0;
              while(1) {
               $82 = (_FT_Stream_GetULong($73)|0);
               $83 = (($82) + ($checksum$07$i$i$i$i))|0;
               $84 = (($$016$i$i$i$i) + -4)|0;
               $85 = ($84>>>0)>(3);
               if ($85) {
                $$016$i$i$i$i = $84;$checksum$07$i$i$i$i = $83;
               } else {
                $$lcssa = $83;
                break;
               }
              }
              $80 = $76 & 3;
              $$01$lcssa$i$i$i$i = $80;$checksum$0$lcssa$i$i$i$i = $$lcssa;
             } else {
              $$01$lcssa$i$i$i$i = $76;$checksum$0$lcssa$i$i$i$i = 0;
             }
             $81 = ($$01$lcssa$i$i$i$i|0)==(0);
             if ($81) {
              $checksum$1$lcssa$i$i$i$i = $checksum$0$lcssa$i$i$i$i;
             } else {
              $$13$i$i$i$i = $$01$lcssa$i$i$i$i;$checksum$14$i$i$i$i = $checksum$0$lcssa$i$i$i$i;$i$05$i$i$i$i = 3;
              while(1) {
               $86 = (_FT_Stream_GetChar($73)|0);
               $87 = $86&255;
               $88 = $i$05$i$i$i$i << 3;
               $89 = $87 << $88;
               $90 = (($89) + ($checksum$14$i$i$i$i))|0;
               $91 = (($$13$i$i$i$i) + -1)|0;
               $92 = (($i$05$i$i$i$i) + -1)|0;
               $93 = ($91|0)==(0);
               if ($93) {
                $checksum$1$lcssa$i$i$i$i = $90;
                break;
               } else {
                $$13$i$i$i$i = $91;$checksum$14$i$i$i$i = $90;$i$05$i$i$i$i = $92;
               }
              }
             }
             _FT_Stream_ExitFrame($73);
             $checksum$1$i$i = $checksum$1$lcssa$i$i$i$i;
            } else {
             $checksum$1$i$i = 0;
            }
           }
          } else {
           $checksum$1$i$i = $checksum$06$i$i;
          }
         } while(0);
         $94 = ((55776 + (($j$07$i$i*24)|0)|0) + ($k$0$i$i<<3)|0);
         $95 = HEAP32[$94>>2]|0;
         $96 = ($95|0)==($checksum$1$i$i|0);
         $97 = (($num_matched_ids$i$i) + ($j$07$i$i<<2)|0);
         $98 = HEAP32[$97>>2]|0;
         if ($96) {
          $99 = (($98) + 1)|0;
          HEAP32[$97>>2] = $99;
          $101 = $99;
         } else {
          $101 = $98;
         }
         $100 = ($101|0)==(3);
         if ($100) {
          break L38;
         } else {
          $checksum$2$i$i = $checksum$1$i$i;
         }
        } else {
         $checksum$2$i$i = $checksum$06$i$i;
        }
        $102 = (($j$07$i$i) + 1)|0;
        $103 = ($102|0)<(17);
        if (!($103)) {
         break;
        }
        $$pre$i$i = HEAP32[$47>>2]|0;
        $60 = $$pre$i$i;$checksum$06$i$i = $checksum$2$i$i;$j$07$i$i = $102;
       }
       $$pre$i = HEAP16[$44>>1]|0;
       $107 = $$pre$i;$has_cvt$2$i$i = $has_cvt$1$i$i;
      }
      $104 = (($i$010$i$i) + 1)<<16>>16;
      $105 = $104&65535;
      $106 = ($104&65535)<($107&65535);
      if ($106) {
       $353 = $107;$57 = $105;$has_cvt$09$i$i = $has_cvt$2$i$i;$i$010$i$i = $104;
      } else {
       $has_cvt$2$i$i$lcssa = $has_cvt$2$i$i;
       break;
      }
     }
     $49 = ($has_cvt$2$i$i$lcssa<<24>>24)==(0);
     if ($49) {
      $j$15$i$i = 0;
      label = 46;
     } else {
      $50 = HEAP32[$num_matched_ids$i$i>>2]|0;
      $51 = ($50|0)==(3);
      if (!($51)) {
       $52 = ((($num_matched_ids$i$i)) + 4|0);
       $53 = HEAP32[$52>>2]|0;
       $54 = ($53|0)==(3);
       if (!($54)) {
        $114 = ((($num_matched_ids$i$i)) + 8|0);
        $115 = HEAP32[$114>>2]|0;
        $116 = ($115|0)==(3);
        if (!($116)) {
         $117 = ((($num_matched_ids$i$i)) + 12|0);
         $118 = HEAP32[$117>>2]|0;
         $119 = ($118|0)==(3);
         if (!($119)) {
          $120 = ((($num_matched_ids$i$i)) + 16|0);
          $121 = HEAP32[$120>>2]|0;
          $122 = ($121|0)==(3);
          if (!($122)) {
           $123 = ((($num_matched_ids$i$i)) + 20|0);
           $124 = HEAP32[$123>>2]|0;
           $125 = ($124|0)==(3);
           if (!($125)) {
            $126 = ((($num_matched_ids$i$i)) + 24|0);
            $127 = HEAP32[$126>>2]|0;
            $128 = ($127|0)==(3);
            if (!($128)) {
             $129 = ((($num_matched_ids$i$i)) + 28|0);
             $130 = HEAP32[$129>>2]|0;
             $131 = ($130|0)==(3);
             if (!($131)) {
              $132 = ((($num_matched_ids$i$i)) + 32|0);
              $133 = HEAP32[$132>>2]|0;
              $134 = ($133|0)==(3);
              if (!($134)) {
               $135 = ((($num_matched_ids$i$i)) + 36|0);
               $136 = HEAP32[$135>>2]|0;
               $137 = ($136|0)==(3);
               if ($137) {
                break;
               }
               $138 = ((($num_matched_ids$i$i)) + 40|0);
               $139 = HEAP32[$138>>2]|0;
               $140 = ($139|0)==(3);
               if ($140) {
                break;
               }
               $141 = ((($num_matched_ids$i$i)) + 44|0);
               $142 = HEAP32[$141>>2]|0;
               $143 = ($142|0)==(3);
               if ($143) {
                break;
               }
               $144 = ((($num_matched_ids$i$i)) + 48|0);
               $145 = HEAP32[$144>>2]|0;
               $146 = ($145|0)==(3);
               if ($146) {
                break;
               }
               $147 = ((($num_matched_ids$i$i)) + 52|0);
               $148 = HEAP32[$147>>2]|0;
               $149 = ($148|0)==(3);
               if ($149) {
                break;
               }
               $150 = ((($num_matched_ids$i$i)) + 56|0);
               $151 = HEAP32[$150>>2]|0;
               $152 = ($151|0)==(3);
               if ($152) {
                break;
               }
               $153 = ((($num_matched_ids$i$i)) + 60|0);
               $154 = HEAP32[$153>>2]|0;
               $155 = ($154|0)==(3);
               if ($155) {
                break;
               }
               $156 = ((($num_matched_ids$i$i)) + 64|0);
               $157 = HEAP32[$156>>2]|0;
               $phitmp$i = ($157|0)==(3);
               if ($phitmp$i) {
                label = 65;
                break L36;
               } else {
                break L36;
               }
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    }
   } while(0);
   L96: do {
    if ((label|0) == 46) {
     while(1) {
      label = 0;
      $110 = ($j$15$i$i>>>0)<(5);
      $$phi$trans$insert32$i$i = (($num_matched_ids$i$i) + ($j$15$i$i<<2)|0);
      $$pre33$i$i = HEAP32[$$phi$trans$insert32$i$i>>2]|0;
      if ($110) {
       $113 = $$pre33$i$i;
      } else {
       $111 = (($$pre33$i$i) + 1)|0;
       HEAP32[$$phi$trans$insert32$i$i>>2] = $111;
       $113 = $111;
      }
      $112 = ($113|0)==(3);
      $109 = (($j$15$i$i) + 1)|0;
      if ($112) {
       break L96;
      }
      $108 = ($109|0)<(17);
      if ($108) {
       $j$15$i$i = $109;
       label = 46;
      } else {
       break;
      }
     }
     break L36;
    }
   } while(0);
   label = 65;
  }
 } while(0);
 if ((label|0) == 65) {
  $158 = HEAP32[$16>>2]|0;
  $159 = $158 | 8192;
  HEAP32[$16>>2] = $159;
 }
 $160 = ((($12)) + 28|0);
 $161 = HEAP32[$160>>2]|0;
 $162 = ((($ttface)) + 508|0);
 $163 = HEAP32[$162>>2]|0;
 $164 = (FUNCTION_TABLE_iiiii[$163 & 8191]($ttface,1751412088,$12,$table_size$i)|0);
 HEAP32[$error$i>>2] = $164;
 $165 = ($164|0)!=(0);
 $166 = HEAP32[$table_size$i>>2]|0;
 $167 = ($166>>>0)<(8);
 $or$cond$i = $165 | $167;
 if ($or$cond$i) {
 } else {
  $168 = ((($ttface)) + 704|0);
  $169 = (_FT_Stream_ExtractFrame($12,$166,$168)|0);
  HEAP32[$error$i>>2] = $169;
  $170 = ($169|0)==(0);
  if (!($170)) {
   $$0 = $169;
   STACKTOP = sp;return ($$0|0);
  }
  $171 = HEAP32[$168>>2]|0;
  $172 = HEAP32[$table_size$i>>2]|0;
  $173 = (($171) + ($172)|0);
  $174 = HEAP8[$171>>0]|0;
  $175 = $174&255;
  $176 = $175 << 8;
  $177 = ((($171)) + 1|0);
  $178 = HEAP8[$177>>0]|0;
  $179 = $178&255;
  $180 = $176 | $179;
  $181 = ((($171)) + 2|0);
  $182 = HEAP8[$181>>0]|0;
  $183 = $182&255;
  $184 = $183 << 8;
  $185 = ((($171)) + 3|0);
  $186 = HEAP8[$185>>0]|0;
  $187 = $186&255;
  $188 = $184 | $187;
  $189 = ((($171)) + 4|0);
  $190 = HEAP8[$189>>0]|0;
  $191 = $190&255;
  $192 = $191 << 24;
  $193 = ((($171)) + 5|0);
  $194 = HEAP8[$193>>0]|0;
  $195 = $194&255;
  $196 = $195 << 16;
  $197 = $196 | $192;
  $198 = ((($171)) + 6|0);
  $199 = HEAP8[$198>>0]|0;
  $200 = $199&255;
  $201 = $200 << 8;
  $202 = $197 | $201;
  $203 = ((($171)) + 7|0);
  $204 = HEAP8[$203>>0]|0;
  $205 = $204&255;
  $206 = $202 | $205;
  $207 = ($206>>>0)>(4294901759);
  $208 = $206 & 65535;
  $$$i = $207 ? $208 : $206;
  $209 = ($180|0)!=(0);
  $210 = ($188>>>0)>(255);
  $or$cond3$i = $209 | $210;
  $record_size$0$off$i = (($$$i) + -4)|0;
  $211 = ($record_size$0$off$i>>>0)>(65533);
  $212 = $or$cond3$i | $211;
  if ($212) {
   HEAP32[$error$i>>2] = 3;
   label = 78;
  } else {
   $213 = (_ft_mem_realloc($161,1,0,$188,0,$error$i)|0);
   $214 = ((($ttface)) + 720|0);
   HEAP32[$214>>2] = $213;
   $215 = HEAP32[$error$i>>2]|0;
   $216 = ($215|0)==(0);
   if ($216) {
    $217 = ($188|0)==(0);
    L119: do {
     if ($217) {
      $nn$0$lcssa$i = 0;
     } else {
      $218 = ((($171)) + 8|0);
      $nn$08$i = 0;$p$09$i = $218;
      while(1) {
       $219 = (($p$09$i) + ($$$i)|0);
       $220 = ($219>>>0)>($173>>>0);
       if ($220) {
        $nn$0$lcssa$i = $nn$08$i;
        break L119;
       }
       $221 = HEAP8[$p$09$i>>0]|0;
       $222 = HEAP32[$214>>2]|0;
       $223 = (($222) + ($nn$08$i)|0);
       HEAP8[$223>>0] = $221;
       $224 = (($nn$08$i) + 1)|0;
       $225 = ($224>>>0)<($188>>>0);
       if ($225) {
        $nn$08$i = $224;$p$09$i = $219;
       } else {
        $nn$0$lcssa$i = $224;
        break;
       }
      }
     }
    } while(0);
    $226 = ((($ttface)) + 712|0);
    HEAP32[$226>>2] = $nn$0$lcssa$i;
    $227 = HEAP32[$table_size$i>>2]|0;
    $228 = ((($ttface)) + 708|0);
    HEAP32[$228>>2] = $227;
    $229 = ((($ttface)) + 716|0);
    HEAP32[$229>>2] = $$$i;
   } else {
    label = 78;
   }
  }
  if ((label|0) == 78) {
   _FT_Stream_ReleaseFrame($12,$168);
   $230 = ((($ttface)) + 708|0);
   HEAP32[$230>>2] = 0;
  }
  $$pr = HEAP32[$error$i>>2]|0;
  $231 = ($$pr|0)==(0);
  if (!($231)) {
   $$0 = $$pr;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $232 = HEAP32[$16>>2]|0;
 $233 = $232 & 1;
 $234 = ($233|0)==(0);
 L130: do {
  if ($234) {
   $error$4 = 0;
  } else {
   $235 = ((($ttface)) + 128|0);
   $236 = HEAP32[$235>>2]|0;
   $237 = ((($236)) + 48|0);
   $238 = HEAP32[$237>>2]|0;
   $239 = ($238|0)==(0|0);
   L132: do {
    if ($239) {
     $240 = HEAP32[$162>>2]|0;
     $241 = ((($ttface)) + 668|0);
     $242 = (FUNCTION_TABLE_iiiii[$240 & 8191]($ttface,1735162214,$12,$241)|0);
     $243 = $242 & 255;
     $244 = ($243|0)==(142);
     if ($244) {
      HEAP32[$241>>2] = 0;
      label = 85;
     } else {
      $245 = ($242|0)==(0);
      if ($245) {
       label = 85;
      } else {
       $error$0$i$ph = $242;
      }
     }
     do {
      if ((label|0) == 85) {
       $246 = HEAP32[$162>>2]|0;
       $247 = (FUNCTION_TABLE_iiiii[$246 & 8191]($ttface,1819239265,$12,$table_len$i)|0);
       $248 = ($247|0)==(0);
       if ($248) {
        $249 = ((($ttface)) + 210|0);
        $250 = HEAP16[$249>>1]|0;
        $251 = ($250<<16>>16)==(0);
        $252 = HEAP32[$table_len$i>>2]|0;
        if ($251) {
         $256 = ($252>>>0)>(131071);
         if ($256) {
          $error$0$i$ph = 8;
          break;
         }
         $257 = $252 >>> 1;
         $258 = ((($ttface)) + 696|0);
         HEAP32[$258>>2] = $257;
         $$pre$phi$iZ2D = $258;$263 = $257;$shift$0$i = 1;
        } else {
         $253 = ($252>>>0)>(262143);
         if ($253) {
          $error$0$i$ph = 8;
          break;
         }
         $254 = $252 >>> 2;
         $255 = ((($ttface)) + 696|0);
         HEAP32[$255>>2] = $254;
         $$pre$phi$iZ2D = $255;$263 = $254;$shift$0$i = 2;
        }
        $259 = ((($ttface)) + 16|0);
        $260 = HEAP32[$259>>2]|0;
        $261 = (($260) + 1)|0;
        $262 = ($263|0)==($261|0);
        $264 = ($263>>>0)>($260>>>0);
        $or$cond$i2 = $264 | $262;
        do {
         if ($or$cond$i2) {
          $294 = $252;
         } else {
          $265 = $261 << $shift$0$i;
          $266 = ((($ttface)) + 156|0);
          $267 = HEAP32[$266>>2]|0;
          $268 = ((($ttface)) + 152|0);
          $269 = HEAP16[$268>>1]|0;
          $270 = $269&65535;
          $271 = (($267) + ($270<<4)|0);
          $272 = (_FT_Stream_Pos($12)|0);
          $273 = ($269<<16>>16)==(0);
          if ($273) {
           $dist$0$lcssa$i = 2147483647;$entry$0$lcssa$i = $267;
          } else {
           $274 = $267;
           $scevgep5$i = ((($267)) + 16|0);
           $275 = ($271>>>0)>($scevgep5$i>>>0);
           $umax$i = $275 ? $271 : $scevgep5$i;
           $276 = $274 ^ -1;
           $uglygep$i = (($umax$i) + ($276)|0);
           $uglygep7$i = $uglygep$i;
           $277 = $uglygep7$i >>> 4;
           $dist$03$i = 2147483647;$entry$02$i = $267;
           while(1) {
            $278 = ((($entry$02$i)) + 8|0);
            $279 = HEAP32[$278>>2]|0;
            $280 = (($279) - ($272))|0;
            $281 = ($280|0)>(0);
            $282 = ($280|0)<($dist$03$i|0);
            $or$cond1$i = $281 & $282;
            $dist$1$i = $or$cond1$i ? $280 : $dist$03$i;
            $283 = ((($entry$02$i)) + 16|0);
            $284 = ($283>>>0)<($271>>>0);
            if ($284) {
             $dist$03$i = $dist$1$i;$entry$02$i = $283;
            } else {
             $dist$1$i$lcssa = $dist$1$i;
             break;
            }
           }
           $285 = (($277) + 1)|0;
           $scevgep8$i = (($267) + ($285<<4)|0);
           $dist$0$lcssa$i = $dist$1$i$lcssa;$entry$0$lcssa$i = $scevgep8$i;
          }
          $286 = ($entry$0$lcssa$i|0)==($271|0);
          if ($286) {
           $287 = ((($12)) + 4|0);
           $288 = HEAP32[$287>>2]|0;
           $289 = (($288) - ($272))|0;
           $dist$2$i = $289;
          } else {
           $dist$2$i = $dist$0$lcssa$i;
          }
          $290 = ($265>>>0)>($dist$2$i>>>0);
          if ($290) {
           $$pre$i6 = HEAP32[$table_len$i>>2]|0;
           $294 = $$pre$i6;
           break;
          } else {
           $291 = HEAP32[$259>>2]|0;
           $292 = (($291) + 1)|0;
           HEAP32[$$pre$phi$iZ2D>>2] = $292;
           HEAP32[$table_len$i>>2] = $265;
           $294 = $265;
           break;
          }
         }
        } while(0);
        $293 = ((($ttface)) + 700|0);
        $295 = (_FT_Stream_ExtractFrame($12,$294,$293)|0);
        $296 = ($295|0)==(0);
        if ($296) {
         label = 103;
         break L132;
        } else {
         $error$3 = $295;
         break L132;
        }
       } else {
        $error$0$i$ph = 144;
       }
      }
     } while(0);
     $error$3 = $error$0$i$ph;
    } else {
     label = 103;
    }
   } while(0);
   do {
    if ((label|0) == 103) {
     $297 = (_tt_face_load_cvt($ttface,$12)|0);
     $298 = ($297|0)==(0);
     if ($298) {
      $299 = HEAP32[$162>>2]|0;
      $300 = (FUNCTION_TABLE_iiiii[$299 & 8191]($ttface,1718642541,$12,$table_len$i7)|0);
      $301 = ($300|0)==(0);
      if ($301) {
       $304 = HEAP32[$table_len$i7>>2]|0;
       $305 = ((($ttface)) + 628|0);
       HEAP32[$305>>2] = $304;
       $306 = ((($ttface)) + 632|0);
       $307 = (_FT_Stream_ExtractFrame($12,$304,$306)|0);
       $308 = ($307|0)==(0);
       if (!($308)) {
        $error$3 = $307;
        break;
       }
      } else {
       $302 = ((($ttface)) + 632|0);
       HEAP32[$302>>2] = 0;
       $303 = ((($ttface)) + 628|0);
       HEAP32[$303>>2] = 0;
      }
      $309 = HEAP32[$162>>2]|0;
      $310 = (FUNCTION_TABLE_iiiii[$309 & 8191]($ttface,1886545264,$12,$table_len$i9)|0);
      $311 = ($310|0)==(0);
      if ($311) {
       $314 = HEAP32[$table_len$i9>>2]|0;
       $315 = ((($ttface)) + 636|0);
       HEAP32[$315>>2] = $314;
       $316 = ((($ttface)) + 640|0);
       $317 = (_FT_Stream_ExtractFrame($12,$314,$316)|0);
       $error$0$i10 = $317;
      } else {
       $312 = ((($ttface)) + 640|0);
       HEAP32[$312>>2] = 0;
       $313 = ((($ttface)) + 636|0);
       HEAP32[$313>>2] = 0;
       $error$0$i10 = 0;
      }
      $error$3 = $error$0$i10;
     } else {
      $error$3 = $297;
     }
    }
   } while(0);
   $318 = HEAP32[$235>>2]|0;
   $319 = ((($318)) + 48|0);
   $320 = HEAP32[$319>>2]|0;
   $321 = ($320|0)==(0|0);
   if ($321) {
    $322 = ((($ttface)) + 28|0);
    $323 = HEAP32[$322>>2]|0;
    $324 = ($323|0)==(0);
    if ($324) {
     $error$4 = $error$3;
    } else {
     $325 = ((($ttface)) + 700|0);
     $326 = HEAP32[$325>>2]|0;
     $327 = ($326|0)==(0|0);
     if ($327) {
      $error$4 = $error$3;
     } else {
      $328 = ((($ttface)) + 696|0);
      $329 = HEAP32[$328>>2]|0;
      $330 = ($329|0)==(0);
      L177: do {
       if (!($330)) {
        $count$06$i = 0;$glyph_index$05$i = 0;$i$04$i = 0;
        while(1) {
         (_tt_face_get_location($ttface,$i$04$i,$asize$i)|0);
         $331 = HEAP32[$asize$i>>2]|0;
         $332 = ($331|0)==(0);
         if ($332) {
          $count$1$i = $count$06$i;$glyph_index$1$i = $glyph_index$05$i;
         } else {
          $333 = (($count$06$i) + 1)|0;
          $334 = ($333>>>0)>(1);
          if ($334) {
           break L177;
          } else {
           $count$1$i = $333;$glyph_index$1$i = $i$04$i;
          }
         }
         $335 = (($i$04$i) + 1)|0;
         $336 = HEAP32[$328>>2]|0;
         $337 = ($335>>>0)<($336>>>0);
         if ($337) {
          $count$06$i = $count$1$i;$glyph_index$05$i = $glyph_index$1$i;$i$04$i = $335;
         } else {
          $count$1$i$lcssa = $count$1$i;$glyph_index$1$i$lcssa = $glyph_index$1$i;
          break;
         }
        }
        $338 = ($count$1$i$lcssa|0)==(1);
        if ($338) {
         $339 = ($glyph_index$1$i$lcssa|0)==(0);
         do {
          if ($339) {
          } else {
           $340 = (_FT_Get_Glyph_Name($ttface,$glyph_index$1$i$lcssa,$buf$i,8)|0);
           $341 = ($340|0)==(0);
           $342 = HEAP8[$buf$i>>0]|0;
           $343 = ($342<<24>>24)==(46);
           $or$cond$i13 = $341 & $343;
           if ($or$cond$i13) {
            $344 = (_strncmp($buf$i,594602,8)|0);
            $345 = ($344|0)==(0);
            if ($345) {
             break;
            } else {
             $error$4 = $error$3;
             break L130;
            }
           } else {
            $error$4 = $error$3;
            break L130;
           }
          }
         } while(0);
         $346 = HEAP32[$16>>2]|0;
         $347 = $346 & -2;
         HEAP32[$16>>2] = $347;
         $error$4 = $error$3;
         break L130;
        }
       }
      } while(0);
      $error$4 = $error$3;
     }
    }
   } else {
    $error$4 = $error$3;
   }
  }
 } while(0);
 $348 = ((($ttface)) + 512|0);
 HEAP32[$348>>2] = 4404;
 $349 = ((($ttface)) + 520|0);
 HEAP32[$349>>2] = 4405;
 $350 = ((($ttface)) + 524|0);
 HEAP32[$350>>2] = 4406;
 $351 = ((($ttface)) + 528|0);
 HEAP32[$351>>2] = 4407;
 $352 = ((($ttface)) + 516|0);
 HEAP32[$352>>2] = 4408;
 $$0 = $error$4;
 STACKTOP = sp;return ($$0|0);
}
function _TT_Access_Glyph_Frame($loader,$glyph_index,$offset,$byte_count) {
 $loader = $loader|0;
 $glyph_index = $glyph_index|0;
 $offset = $offset|0;
 $byte_count = $byte_count|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($loader)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_FT_Stream_Seek($1,$offset)|0);
 $3 = ($2|0)==(0);
 if (!($3)) {
  $$0 = $2;
  return ($$0|0);
 }
 $4 = (_FT_Stream_EnterFrame($1,$byte_count)|0);
 $5 = ($4|0)==(0);
 if (!($5)) {
  $$0 = $4;
  return ($$0|0);
 }
 $6 = ((($1)) + 32|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($loader)) + 200|0);
 HEAP32[$8>>2] = $7;
 $9 = ((($1)) + 36|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($loader)) + 204|0);
 HEAP32[$11>>2] = $10;
 $$0 = 0;
 return ($$0|0);
}
function _TT_Load_Glyph_Header($loader) {
 $loader = $loader|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $sext = 0, $sext1 = 0, $sext2 = 0, $sext3 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($loader)) + 200|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($loader)) + 204|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 10|0);
 $5 = ($4>>>0)>($3>>>0);
 if ($5) {
  $$0 = 20;
  return ($$0|0);
 }
 $6 = HEAP8[$1>>0]|0;
 $7 = $6&255;
 $8 = $7 << 8;
 $9 = ((($1)) + 1|0);
 $10 = HEAP8[$9>>0]|0;
 $11 = $10&255;
 $12 = $8 | $11;
 $13 = $12&65535;
 $14 = ((($loader)) + 32|0);
 HEAP16[$14>>1] = $13;
 $15 = ((($1)) + 2|0);
 $16 = HEAP8[$15>>0]|0;
 $17 = $16&255;
 $18 = $17 << 8;
 $19 = ((($1)) + 3|0);
 $20 = HEAP8[$19>>0]|0;
 $21 = $20&255;
 $22 = $18 | $21;
 $sext = $22 << 16;
 $23 = $sext >> 16;
 $24 = ((($loader)) + 36|0);
 HEAP32[$24>>2] = $23;
 $25 = ((($1)) + 4|0);
 $26 = HEAP8[$25>>0]|0;
 $27 = $26&255;
 $28 = $27 << 8;
 $29 = ((($1)) + 5|0);
 $30 = HEAP8[$29>>0]|0;
 $31 = $30&255;
 $32 = $28 | $31;
 $sext1 = $32 << 16;
 $33 = $sext1 >> 16;
 $34 = ((($loader)) + 40|0);
 HEAP32[$34>>2] = $33;
 $35 = ((($1)) + 6|0);
 $36 = HEAP8[$35>>0]|0;
 $37 = $36&255;
 $38 = $37 << 8;
 $39 = ((($1)) + 7|0);
 $40 = HEAP8[$39>>0]|0;
 $41 = $40&255;
 $42 = $38 | $41;
 $sext2 = $42 << 16;
 $43 = $sext2 >> 16;
 $44 = ((($loader)) + 44|0);
 HEAP32[$44>>2] = $43;
 $45 = ((($1)) + 8|0);
 $46 = HEAP8[$45>>0]|0;
 $47 = $46&255;
 $48 = $47 << 8;
 $49 = ((($1)) + 9|0);
 $50 = HEAP8[$49>>0]|0;
 $51 = $50&255;
 $52 = $48 | $51;
 $sext3 = $52 << 16;
 $53 = $sext3 >> 16;
 $54 = ((($loader)) + 48|0);
 HEAP32[$54>>2] = $53;
 HEAP32[$0>>2] = $4;
 $$0 = 0;
 return ($$0|0);
}
function _TT_Load_Simple_Glyph($load) {
 $load = $load|0;
 var $$ = 0, $$0 = 0, $$21 = 0, $$7 = 0, $$pn = 0, $$pn$lcssa = 0, $$pn23 = 0, $$pn23$in = 0, $$pn4 = 0, $$pn4$lcssa = 0, $$pre = 0, $$pre64 = 0, $$pre65 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0;
 var $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0;
 var $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0;
 var $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cont$0 = 0, $error$i = 0, $flag$0$be = 0;
 var $flag$050 = 0, $flag$239 = 0, $flag$335 = 0, $n_points$010 = 0, $n_points$01114 = 0, $p$0 = 0, $p$0$lcssa = 0, $p$1$be = 0, $p$1$lcssa = 0, $p$151 = 0, $p$242 = 0, $p$3 = 0, $p$3$lcssa = 0, $p$5$lcssa = 0, $p$538 = 0, $p$6 = 0, $prev_cont$0 = 0, $scevgep = 0, $scevgep62 = 0, $sext$mask = 0;
 var $sext2 = 0, $sext3 = 0, $sext5 = 0, $sext6 = 0, $vec$040 = 0, $vec$136 = 0, $x$041 = 0, $x$237 = 0, $y$0 = 0, $y1$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp;
 $0 = ((($load)) + 200|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($load)) + 204|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($load)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($load)) + 32|0);
 $7 = HEAP16[$6>>1]|0;
 $8 = $7 << 16 >> 16;
 $9 = ($7<<16>>16)==(0);
 if (!($9)) {
  $10 = ((($5)) + 20|0);
  $11 = HEAP16[$10>>1]|0;
  $12 = $11 << 16 >> 16;
  $13 = ((($5)) + 56|0);
  $14 = HEAP16[$13>>1]|0;
  $15 = $14 << 16 >> 16;
  $16 = (($12) + ($8))|0;
  $17 = (($16) + ($15))|0;
  $18 = ((($5)) + 8|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = ($17>>>0)>($19>>>0);
  if ($20) {
   $21 = (_FT_GlyphLoader_CheckPoints($5,0,$8)|0);
   $22 = ($21|0)==(0);
   if (!($22)) {
    $$0 = $21;
    STACKTOP = sp;return ($$0|0);
   }
  }
  $23 = ($7<<16>>16)>(4094);
  if ($23) {
   $$0 = 20;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $$pn23$in = ((($5)) + 68|0);
 $$pn23 = HEAP32[$$pn23$in>>2]|0;
 $24 = (($$pn23) + ($8<<1)|0);
 $25 = $8 << 1;
 $26 = (($25) + 2)|0;
 $27 = (($1) + ($26)|0);
 $28 = ($27>>>0)>($3>>>0);
 if ($28) {
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 $29 = HEAP8[$1>>0]|0;
 $30 = $29&255;
 $31 = $30 << 8;
 $32 = ((($1)) + 1|0);
 $33 = HEAP8[$32>>0]|0;
 $34 = $33&255;
 $35 = $31 | $34;
 $36 = ($7<<16>>16)>(0);
 if ($36) {
  $37 = $35&65535;
  HEAP16[$$pn23>>1] = $37;
 }
 $sext$mask = $31 & 32768;
 $38 = ($sext$mask|0)==(0);
 if ($38) {
  $$pn = $$pn23;$$pn4 = $1;$prev_cont$0 = $35;
 } else {
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 while(1) {
  $p$0 = ((($$pn4)) + 2|0);
  $cont$0 = ((($$pn)) + 2|0);
  $39 = ($cont$0>>>0)<($24>>>0);
  if (!($39)) {
   $$pn$lcssa = $$pn;$$pn4$lcssa = $$pn4;$p$0$lcssa = $p$0;
   break;
  }
  $40 = HEAP8[$p$0>>0]|0;
  $41 = $40&255;
  $42 = $41 << 8;
  $43 = ((($$pn4)) + 3|0);
  $44 = HEAP8[$43>>0]|0;
  $45 = $44&255;
  $46 = $42 | $45;
  $47 = $46&65535;
  HEAP16[$cont$0>>1] = $47;
  $sext5 = $46 << 16;
  $sext6 = $prev_cont$0 << 16;
  $48 = ($sext5|0)>($sext6|0);
  if ($48) {
   $$pn = $cont$0;$$pn4 = $p$0;$prev_cont$0 = $46;
  } else {
   $$0 = 20;
   label = 51;
   break;
  }
 }
 if ((label|0) == 51) {
  STACKTOP = sp;return ($$0|0);
 }
 do {
  if ($36) {
   $49 = HEAP16[$$pn$lcssa>>1]|0;
   $50 = $49 << 16 >> 16;
   $51 = (($50) + 1)|0;
   $52 = ($49<<16>>16)<(-1);
   if ($52) {
    $$0 = 20;
    STACKTOP = sp;return ($$0|0);
   } else {
    $53 = (($50) + 5)|0;
    $54 = ($53|0)==(0);
    if ($54) {
     $n_points$01114 = $51;
     break;
    } else {
     $62 = $53;$n_points$010 = $51;
     label = 14;
     break;
    }
   }
  } else {
   $62 = 4;$n_points$010 = 0;
   label = 14;
  }
 } while(0);
 if ((label|0) == 14) {
  $55 = ((($5)) + 22|0);
  $56 = HEAP16[$55>>1]|0;
  $57 = $56 << 16 >> 16;
  $58 = ((($5)) + 58|0);
  $59 = HEAP16[$58>>1]|0;
  $60 = $59 << 16 >> 16;
  $61 = (($57) + ($62))|0;
  $63 = (($61) + ($60))|0;
  $64 = ((($5)) + 4|0);
  $65 = HEAP32[$64>>2]|0;
  $66 = ($63>>>0)>($65>>>0);
  if ($66) {
   $67 = (_FT_GlyphLoader_CheckPoints($5,$62,0)|0);
   $68 = ($67|0)==(0);
   if ($68) {
    $n_points$01114 = $n_points$010;
   } else {
    $$0 = $67;
    STACKTOP = sp;return ($$0|0);
   }
  } else {
   $n_points$01114 = $n_points$010;
  }
 }
 $69 = ((($load)) + 8|0);
 $70 = HEAP32[$69>>2]|0;
 $71 = ((($70)) + 140|0);
 HEAP32[$71>>2] = 0;
 $72 = ((($70)) + 136|0);
 HEAP32[$72>>2] = 0;
 $73 = ((($$pn4$lcssa)) + 4|0);
 $74 = ($73>>>0)>($3>>>0);
 if ($74) {
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 $75 = HEAP8[$p$0$lcssa>>0]|0;
 $76 = $75&255;
 $77 = $76 << 8;
 $78 = ((($$pn4$lcssa)) + 3|0);
 $79 = HEAP8[$78>>0]|0;
 $80 = $79&255;
 $81 = $77 | $80;
 $82 = $3;
 $83 = $73;
 $84 = (($82) - ($83))|0;
 $85 = ($84|0)<($81|0);
 if ($85) {
  $$0 = 22;
  STACKTOP = sp;return ($$0|0);
 }
 $86 = ((($load)) + 16|0);
 $87 = HEAP32[$86>>2]|0;
 $88 = $87 & 2;
 $89 = ($88|0)==(0);
 if ($89) {
  $90 = ((($load)) + 160|0);
  $91 = HEAP32[$90>>2]|0;
  $92 = ((($91)) + 388|0);
  $93 = HEAP32[$92>>2]|0;
  $94 = ((($91)) + 8|0);
  $95 = HEAP32[$94>>2]|0;
  $96 = ((($91)) + 392|0);
  $97 = ($93>>>0)<($81>>>0);
  if ($97) {
   $100 = HEAP32[$96>>2]|0;
   $101 = (_ft_mem_realloc($95,1,$93,$81,$100,$error$i)|0);
   HEAP32[$96>>2] = $101;
   $102 = HEAP32[$error$i>>2]|0;
   $103 = ($102|0)==(0);
   $$21 = $103 ? $81 : $93;
   $$pre = HEAP32[$90>>2]|0;
   $104 = $$21 & 65535;
   $105 = ((($$pre)) + 388|0);
   HEAP32[$105>>2] = $104;
   $106 = ($102|0)==(0);
   if ($106) {
    $110 = $$pre;
   } else {
    $$0 = $102;
    STACKTOP = sp;return ($$0|0);
   }
  } else {
   $98 = $93 & 65535;
   $99 = ((($91)) + 388|0);
   HEAP32[$99>>2] = $98;
   $110 = $91;
  }
  $107 = HEAP32[$69>>2]|0;
  $108 = ((($107)) + 140|0);
  HEAP32[$108>>2] = $81;
  $109 = ((($110)) + 392|0);
  $111 = HEAP32[$109>>2]|0;
  $112 = ((($107)) + 136|0);
  HEAP32[$112>>2] = $111;
  $113 = $111;
  _memcpy(($113|0),($73|0),($81|0))|0;
 }
 $114 = (($73) + ($81)|0);
 $115 = ((($5)) + 64|0);
 $116 = HEAP32[$115>>2]|0;
 $117 = (($116) + ($n_points$01114)|0);
 $118 = ($n_points$01114|0)>(0);
 L47: do {
  if ($118) {
   $flag$050 = $116;$p$151 = $114;
   while(1) {
    $119 = ((($p$151)) + 1|0);
    $120 = ($119>>>0)>($3>>>0);
    if ($120) {
     $$0 = 20;
     label = 51;
     break;
    }
    $121 = HEAP8[$p$151>>0]|0;
    $122 = ((($flag$050)) + 1|0);
    HEAP8[$flag$050>>0] = $121;
    $123 = $121 & 8;
    $124 = ($123<<24>>24)==(0);
    if ($124) {
     $flag$0$be = $122;$p$1$be = $119;
    } else {
     $126 = ((($p$151)) + 2|0);
     $127 = ($126>>>0)>($3>>>0);
     if ($127) {
      $$0 = 20;
      label = 51;
      break;
     }
     $128 = HEAP8[$119>>0]|0;
     $129 = $128&255;
     $130 = (($122) + ($129)|0);
     $131 = ($130>>>0)>($117>>>0);
     if ($131) {
      $$0 = 20;
      label = 51;
      break;
     }
     $132 = ($128<<24>>24)==(0);
     if ($132) {
      $flag$0$be = $122;$p$1$be = $126;
     } else {
      $133 = (($128) + -1)<<24>>24;
      $134 = $133&255;
      $135 = (($134) + 1)|0;
      _memset(($122|0),($121|0),($135|0))|0;
      $scevgep = ((($flag$050)) + 2|0);
      $scevgep62 = (($scevgep) + ($134)|0);
      $flag$0$be = $scevgep62;$p$1$be = $126;
     }
    }
    $125 = ($flag$0$be>>>0)<($117>>>0);
    if ($125) {
     $flag$050 = $flag$0$be;$p$151 = $p$1$be;
    } else {
     $p$1$lcssa = $p$1$be;
     break L47;
    }
   }
   if ((label|0) == 51) {
    STACKTOP = sp;return ($$0|0);
   }
  } else {
   $p$1$lcssa = $114;
  }
 } while(0);
 $136 = ((($5)) + 60|0);
 $137 = HEAP32[$136>>2]|0;
 $138 = (($137) + ($n_points$01114<<3)|0);
 $139 = ($p$1$lcssa>>>0)>($3>>>0);
 if ($139) {
  $$0 = 20;
  STACKTOP = sp;return ($$0|0);
 }
 L65: do {
  if ($118) {
   $140 = HEAP32[$115>>2]|0;
   $flag$239 = $140;$p$242 = $p$1$lcssa;$vec$040 = $137;$x$041 = 0;
   while(1) {
    $141 = HEAP8[$flag$239>>0]|0;
    $142 = $141&255;
    $143 = $142 & 2;
    $144 = ($143|0)==(0);
    if ($144) {
     $152 = $142 & 16;
     $153 = ($152|0)==(0);
     if ($153) {
      $154 = ((($p$242)) + 2|0);
      $155 = ($154>>>0)>($3>>>0);
      if ($155) {
       $$0 = 20;
       label = 51;
       break;
      }
      $156 = HEAP8[$p$242>>0]|0;
      $157 = $156&255;
      $158 = $157 << 8;
      $159 = ((($p$242)) + 1|0);
      $160 = HEAP8[$159>>0]|0;
      $161 = $160&255;
      $162 = $158 | $161;
      $sext3 = $162 << 16;
      $163 = $sext3 >> 16;
      $p$3 = $154;$y$0 = $163;
     } else {
      $p$3 = $p$242;$y$0 = 0;
     }
    } else {
     $145 = ((($p$242)) + 1|0);
     $146 = ($145>>>0)>($3>>>0);
     if ($146) {
      $$0 = 20;
      label = 51;
      break;
     }
     $147 = HEAP8[$p$242>>0]|0;
     $148 = $147&255;
     $149 = $142 & 16;
     $150 = ($149|0)==(0);
     $151 = (0 - ($148))|0;
     $$ = $150 ? $151 : $148;
     $p$3 = $145;$y$0 = $$;
    }
    $164 = (($y$0) + ($x$041))|0;
    HEAP32[$vec$040>>2] = $164;
    $165 = $142 & 237;
    $166 = $165&255;
    HEAP8[$flag$239>>0] = $166;
    $167 = ((($vec$040)) + 8|0);
    $168 = ((($flag$239)) + 1|0);
    $169 = ($167>>>0)<($138>>>0);
    if ($169) {
     $flag$239 = $168;$p$242 = $p$3;$vec$040 = $167;$x$041 = $164;
    } else {
     $p$3$lcssa = $p$3;
     label = 41;
     break;
    }
   }
   if ((label|0) == 41) {
    $$pre64 = HEAP32[$136>>2]|0;
    $170 = (($$pre64) + ($n_points$01114<<3)|0);
    if (!($118)) {
     $p$5$lcssa = $p$3$lcssa;
     break;
    }
    $$pre65 = HEAP32[$115>>2]|0;
    $flag$335 = $$pre65;$p$538 = $p$3$lcssa;$vec$136 = $$pre64;$x$237 = 0;
    while(1) {
     $171 = HEAP8[$flag$335>>0]|0;
     $172 = $171&255;
     $173 = $172 & 4;
     $174 = ($173|0)==(0);
     if ($174) {
      $182 = $172 & 32;
      $183 = ($182|0)==(0);
      if ($183) {
       $184 = ((($p$538)) + 2|0);
       $185 = ($184>>>0)>($3>>>0);
       if ($185) {
        $$0 = 20;
        label = 51;
        break;
       }
       $186 = HEAP8[$p$538>>0]|0;
       $187 = $186&255;
       $188 = $187 << 8;
       $189 = ((($p$538)) + 1|0);
       $190 = HEAP8[$189>>0]|0;
       $191 = $190&255;
       $192 = $188 | $191;
       $sext2 = $192 << 16;
       $193 = $sext2 >> 16;
       $p$6 = $184;$y1$0 = $193;
      } else {
       $p$6 = $p$538;$y1$0 = 0;
      }
     } else {
      $175 = ((($p$538)) + 1|0);
      $176 = ($175>>>0)>($3>>>0);
      if ($176) {
       $$0 = 20;
       label = 51;
       break;
      }
      $177 = HEAP8[$p$538>>0]|0;
      $178 = $177&255;
      $179 = $172 & 32;
      $180 = ($179|0)==(0);
      $181 = (0 - ($178))|0;
      $$7 = $180 ? $181 : $178;
      $p$6 = $175;$y1$0 = $$7;
     }
     $194 = (($y1$0) + ($x$237))|0;
     $195 = ((($vec$136)) + 4|0);
     HEAP32[$195>>2] = $194;
     $196 = $172 & 1;
     $197 = $196&255;
     HEAP8[$flag$335>>0] = $197;
     $198 = ((($vec$136)) + 8|0);
     $199 = ((($flag$335)) + 1|0);
     $200 = ($198>>>0)<($170>>>0);
     if ($200) {
      $flag$335 = $199;$p$538 = $p$6;$vec$136 = $198;$x$237 = $194;
     } else {
      $p$5$lcssa = $p$6;
      break L65;
     }
    }
    if ((label|0) == 51) {
     STACKTOP = sp;return ($$0|0);
    }
   }
   else if ((label|0) == 51) {
    STACKTOP = sp;return ($$0|0);
   }
  } else {
   $p$5$lcssa = $p$1$lcssa;
  }
 } while(0);
 $201 = $n_points$01114&65535;
 $202 = ((($5)) + 58|0);
 HEAP16[$202>>1] = $201;
 $203 = ((($5)) + 56|0);
 HEAP16[$203>>1] = $7;
 HEAP32[$0>>2] = $p$5$lcssa;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _TT_Load_Composite_Glyph($loader) {
 $loader = $loader|0;
 var $$0 = 0, $$lcssa18 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $count$1 = 0, $num_subglyphs$0 = 0, $p$0 = 0, $p$1 = 0, $p$2 = 0, $p$2$lcssa = 0, $sext = 0, $sext2 = 0, $sext3 = 0, $sext4 = 0, $sext5 = 0, $sext6 = 0, $sext7 = 0, $sext8 = 0, $sext9 = 0, $xx$0 = 0, $xy$0 = 0, $yx$0 = 0, $yy$0 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($loader)) + 200|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($loader)) + 204|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($loader)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 88|0);
 $num_subglyphs$0 = 0;$p$0 = $1;
 while(1) {
  $7 = (($num_subglyphs$0) + 1)|0;
  $8 = (_FT_GlyphLoader_CheckSubGlyphs($5,$7)|0);
  $9 = ($8|0)==(0);
  if (!($9)) {
   $$0 = $8;
   label = 25;
   break;
  }
  $10 = ((($p$0)) + 4|0);
  $11 = ($10>>>0)>($3>>>0);
  if ($11) {
   $$0 = 21;
   label = 25;
   break;
  }
  $12 = HEAP32[$6>>2]|0;
  $13 = (((($12) + ($num_subglyphs$0<<5)|0)) + 12|0);
  HEAP32[$13>>2] = 0;
  $14 = (((($12) + ($num_subglyphs$0<<5)|0)) + 8|0);
  HEAP32[$14>>2] = 0;
  $15 = HEAP8[$p$0>>0]|0;
  $16 = $15&255;
  $17 = $16 << 8;
  $18 = ((($p$0)) + 1|0);
  $19 = HEAP8[$18>>0]|0;
  $20 = $19&255;
  $21 = $17 | $20;
  $22 = $21&65535;
  $23 = (((($12) + ($num_subglyphs$0<<5)|0)) + 4|0);
  HEAP16[$23>>1] = $22;
  $24 = ((($p$0)) + 2|0);
  $25 = HEAP8[$24>>0]|0;
  $26 = $25&255;
  $27 = $26 << 8;
  $28 = ((($p$0)) + 3|0);
  $29 = HEAP8[$28>>0]|0;
  $30 = $29&255;
  $31 = $27 | $30;
  $32 = (($12) + ($num_subglyphs$0<<5)|0);
  HEAP32[$32>>2] = $31;
  $33 = $20 << 1;
  $34 = $33 & 2;
  $35 = (($34) + 2)|0;
  $36 = $20 & 8;
  $37 = ($36|0)==(0);
  do {
   if ($37) {
    $39 = $20 & 64;
    $40 = ($39|0)==(0);
    if ($40) {
     $42 = $20 >>> 4;
     $43 = $42 & 8;
     $44 = $35 | $43;
     $count$1 = $44;
     break;
    } else {
     $41 = (($34) + 6)|0;
     $count$1 = $41;
     break;
    }
   } else {
    $38 = $34 | 4;
    $count$1 = $38;
   }
  } while(0);
  $45 = (($10) + ($count$1)|0);
  $46 = ($45>>>0)>($3>>>0);
  if ($46) {
   $$0 = 21;
   label = 25;
   break;
  }
  $47 = $20 & 2;
  $48 = ($47|0)==(0);
  $49 = $20 & 1;
  $50 = ($49|0)!=(0);
  do {
   if ($48) {
    if ($50) {
     $75 = HEAP8[$10>>0]|0;
     $76 = $75&255;
     $77 = $76 << 8;
     $78 = ((($p$0)) + 5|0);
     $79 = HEAP8[$78>>0]|0;
     $80 = $79&255;
     $81 = $77 | $80;
     HEAP32[$14>>2] = $81;
     $82 = ((($p$0)) + 8|0);
     $83 = ((($p$0)) + 6|0);
     $84 = HEAP8[$83>>0]|0;
     $85 = $84&255;
     $86 = $85 << 8;
     $87 = ((($p$0)) + 7|0);
     $88 = HEAP8[$87>>0]|0;
     $89 = $88&255;
     $90 = $86 | $89;
     HEAP32[$13>>2] = $90;
     $p$1 = $82;
     break;
    } else {
     $91 = ((($p$0)) + 5|0);
     $92 = HEAP8[$10>>0]|0;
     $93 = $92&255;
     HEAP32[$14>>2] = $93;
     $94 = ((($p$0)) + 6|0);
     $95 = HEAP8[$91>>0]|0;
     $96 = $95&255;
     HEAP32[$13>>2] = $96;
     $p$1 = $94;
     break;
    }
   } else {
    if ($50) {
     $51 = HEAP8[$10>>0]|0;
     $52 = $51&255;
     $53 = $52 << 8;
     $54 = ((($p$0)) + 5|0);
     $55 = HEAP8[$54>>0]|0;
     $56 = $55&255;
     $57 = $53 | $56;
     $sext8 = $57 << 16;
     $58 = $sext8 >> 16;
     HEAP32[$14>>2] = $58;
     $59 = ((($p$0)) + 8|0);
     $60 = ((($p$0)) + 6|0);
     $61 = HEAP8[$60>>0]|0;
     $62 = $61&255;
     $63 = $62 << 8;
     $64 = ((($p$0)) + 7|0);
     $65 = HEAP8[$64>>0]|0;
     $66 = $65&255;
     $67 = $63 | $66;
     $sext9 = $67 << 16;
     $68 = $sext9 >> 16;
     HEAP32[$13>>2] = $68;
     $p$1 = $59;
     break;
    } else {
     $69 = ((($p$0)) + 5|0);
     $70 = HEAP8[$10>>0]|0;
     $71 = $70 << 24 >> 24;
     HEAP32[$14>>2] = $71;
     $72 = ((($p$0)) + 6|0);
     $73 = HEAP8[$69>>0]|0;
     $74 = $73 << 24 >> 24;
     HEAP32[$13>>2] = $74;
     $p$1 = $72;
     break;
    }
   }
  } while(0);
  do {
   if ($37) {
    $106 = $20 & 64;
    $107 = ($106|0)==(0);
    if (!($107)) {
     $108 = HEAP8[$p$1>>0]|0;
     $109 = $108&255;
     $110 = $109 << 8;
     $111 = ((($p$1)) + 1|0);
     $112 = HEAP8[$111>>0]|0;
     $113 = $112&255;
     $114 = $110 | $113;
     $sext5 = $114 << 16;
     $115 = $sext5 >> 14;
     $116 = ((($p$1)) + 4|0);
     $117 = ((($p$1)) + 2|0);
     $118 = HEAP8[$117>>0]|0;
     $119 = $118&255;
     $120 = $119 << 8;
     $121 = ((($p$1)) + 3|0);
     $122 = HEAP8[$121>>0]|0;
     $123 = $122&255;
     $124 = $120 | $123;
     $sext6 = $124 << 16;
     $125 = $sext6 >> 14;
     $p$2 = $116;$xx$0 = $115;$xy$0 = 0;$yx$0 = 0;$yy$0 = $125;
     break;
    }
    $126 = $20 & 128;
    $127 = ($126|0)==(0);
    if ($127) {
     $p$2 = $p$1;$xx$0 = 65536;$xy$0 = 0;$yx$0 = 0;$yy$0 = 65536;
    } else {
     $128 = HEAP8[$p$1>>0]|0;
     $129 = $128&255;
     $130 = $129 << 8;
     $131 = ((($p$1)) + 1|0);
     $132 = HEAP8[$131>>0]|0;
     $133 = $132&255;
     $134 = $130 | $133;
     $sext = $134 << 16;
     $135 = $sext >> 14;
     $136 = ((($p$1)) + 2|0);
     $137 = HEAP8[$136>>0]|0;
     $138 = $137&255;
     $139 = $138 << 8;
     $140 = ((($p$1)) + 3|0);
     $141 = HEAP8[$140>>0]|0;
     $142 = $141&255;
     $143 = $139 | $142;
     $sext2 = $143 << 16;
     $144 = $sext2 >> 14;
     $145 = ((($p$1)) + 4|0);
     $146 = HEAP8[$145>>0]|0;
     $147 = $146&255;
     $148 = $147 << 8;
     $149 = ((($p$1)) + 5|0);
     $150 = HEAP8[$149>>0]|0;
     $151 = $150&255;
     $152 = $148 | $151;
     $sext3 = $152 << 16;
     $153 = $sext3 >> 14;
     $154 = ((($p$1)) + 8|0);
     $155 = ((($p$1)) + 6|0);
     $156 = HEAP8[$155>>0]|0;
     $157 = $156&255;
     $158 = $157 << 8;
     $159 = ((($p$1)) + 7|0);
     $160 = HEAP8[$159>>0]|0;
     $161 = $160&255;
     $162 = $158 | $161;
     $sext4 = $162 << 16;
     $163 = $sext4 >> 14;
     $p$2 = $154;$xx$0 = $135;$xy$0 = $153;$yx$0 = $144;$yy$0 = $163;
    }
   } else {
    $97 = ((($p$1)) + 2|0);
    $98 = HEAP8[$p$1>>0]|0;
    $99 = $98&255;
    $100 = $99 << 8;
    $101 = ((($p$1)) + 1|0);
    $102 = HEAP8[$101>>0]|0;
    $103 = $102&255;
    $104 = $100 | $103;
    $sext7 = $104 << 16;
    $105 = $sext7 >> 14;
    $p$2 = $97;$xx$0 = $105;$xy$0 = 0;$yx$0 = 0;$yy$0 = $105;
   }
  } while(0);
  $164 = (((($12) + ($num_subglyphs$0<<5)|0)) + 16|0);
  HEAP32[$164>>2] = $xx$0;
  $165 = (((($12) + ($num_subglyphs$0<<5)|0)) + 20|0);
  HEAP32[$165>>2] = $xy$0;
  $166 = (((($12) + ($num_subglyphs$0<<5)|0)) + 24|0);
  HEAP32[$166>>2] = $yx$0;
  $167 = (((($12) + ($num_subglyphs$0<<5)|0)) + 28|0);
  HEAP32[$167>>2] = $yy$0;
  $168 = $22 & 32;
  $169 = ($168<<16>>16)==(0);
  if ($169) {
   $$lcssa18 = $7;$p$2$lcssa = $p$2;
   label = 24;
   break;
  } else {
   $num_subglyphs$0 = $7;$p$0 = $p$2;
  }
 }
 if ((label|0) == 24) {
  $170 = ((($5)) + 84|0);
  HEAP32[$170>>2] = $$lcssa18;
  $171 = ((($loader)) + 24|0);
  $172 = HEAP32[$171>>2]|0;
  $173 = (_FT_Stream_Pos($172)|0);
  $174 = (($p$2$lcssa) + ($173)|0);
  $175 = $174;
  $176 = $3;
  $177 = (($175) - ($176))|0;
  $178 = ((($loader)) + 168|0);
  HEAP32[$178>>2] = $177;
  HEAP32[$0>>2] = $p$2$lcssa;
  $$0 = 0;
  return ($$0|0);
 }
 else if ((label|0) == 25) {
  return ($$0|0);
 }
 return (0)|0;
}
function _TT_Forget_Glyph_Frame($loader) {
 $loader = $loader|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($loader)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 _FT_Stream_ExitFrame($1);
 return;
}
function _tt_face_done($ttface) {
 $ttface = $ttface|0;
 var $$in$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($ttface|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($ttface)) + 100|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($ttface)) + 104|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($ttface)) + 532|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($ttface)) + 656|0);
 $8 = ((($7)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)==(0|0);
 if (!($10)) {
  $11 = HEAP32[$7>>2]|0;
  FUNCTION_TABLE_vi[$9 & 8191]($11);
 }
 $12 = ($6|0)==(0|0);
 if (!($12)) {
  $13 = ((($6)) + 12|0);
  $14 = HEAP32[$13>>2]|0;
  FUNCTION_TABLE_vi[$14 & 8191]($ttface);
 }
 $15 = HEAP32[$3>>2]|0;
 $16 = ((($ttface)) + 700|0);
 _FT_Stream_ReleaseFrame($15,$16);
 $17 = ((($ttface)) + 696|0);
 HEAP32[$17>>2] = 0;
 $18 = HEAP32[$3>>2]|0;
 $19 = ((($18)) + 28|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ((($ttface)) + 720|0);
 $22 = HEAP32[$21>>2]|0;
 _ft_mem_free($20,$22);
 HEAP32[$21>>2] = 0;
 $23 = ((($ttface)) + 704|0);
 _FT_Stream_ReleaseFrame($18,$23);
 $24 = ((($ttface)) + 648|0);
 $25 = HEAP32[$24>>2]|0;
 _ft_mem_free($2,$25);
 HEAP32[$24>>2] = 0;
 $26 = ((($ttface)) + 644|0);
 HEAP32[$26>>2] = 0;
 $27 = ((($ttface)) + 632|0);
 _FT_Stream_ReleaseFrame($4,$27);
 $28 = ((($ttface)) + 640|0);
 _FT_Stream_ReleaseFrame($4,$28);
 $29 = ((($ttface)) + 628|0);
 HEAP32[$29>>2] = 0;
 $30 = ((($ttface)) + 636|0);
 HEAP32[$30>>2] = 0;
 $31 = ((($ttface)) + 676|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ($32|0)==(0|0);
 if (!($33)) {
  $34 = ((($32)) + 4|0);
  $35 = HEAP32[$34>>2]|0;
  _ft_mem_free($2,$35);
  HEAP32[$34>>2] = 0;
  $36 = ((($32)) + 8|0);
  $37 = HEAP32[$36>>2]|0;
  _ft_mem_free($2,$37);
  HEAP32[$36>>2] = 0;
  $38 = ((($32)) + 20|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = ($39|0)==(0|0);
  if (!($40)) {
   $41 = HEAP32[$32>>2]|0;
   $42 = ($41|0)==(0);
   if ($42) {
    $$in$i = $39;
   } else {
    $44 = $39;$i$01$i = 0;
    while(1) {
     $43 = (((($44) + ($i$01$i<<3)|0)) + 4|0);
     $45 = HEAP32[$43>>2]|0;
     _ft_mem_free($2,$45);
     $46 = HEAP32[$38>>2]|0;
     $47 = (((($46) + ($i$01$i<<3)|0)) + 4|0);
     HEAP32[$47>>2] = 0;
     $48 = (($i$01$i) + 1)|0;
     $49 = HEAP32[$32>>2]|0;
     $50 = ($48>>>0)<($49>>>0);
     if ($50) {
      $44 = $46;$i$01$i = $48;
     } else {
      $$in$i = $46;
      break;
     }
    }
   }
   _ft_mem_free($2,$$in$i);
   HEAP32[$38>>2] = 0;
  }
  $51 = ((($32)) + 28|0);
  $52 = HEAP32[$51>>2]|0;
  _ft_mem_free($2,$52);
  HEAP32[$51>>2] = 0;
  $53 = ((($32)) + 36|0);
  $54 = HEAP32[$53>>2]|0;
  _ft_mem_free($2,$54);
  HEAP32[$53>>2] = 0;
  _ft_mem_free($2,$32);
 }
 HEAP32[$31>>2] = 0;
 return;
}
function _tt_size_init($ttsize) {
 $ttsize = $ttsize|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($ttsize)) + 296|0);
 HEAP32[$0>>2] = -1;
 $1 = ((($ttsize)) + 300|0);
 HEAP32[$1>>2] = -1;
 $2 = ((($ttsize)) + 72|0);
 $3 = ((($2)) + 36|0);
 HEAP8[$3>>0] = 0;
 $4 = ((($ttsize)) + 112|0);
 HEAP32[$4>>2] = -1;
 return 0;
}
function _tt_size_done($ttsize) {
 $ttsize = $ttsize|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _tt_size_done_bytecode($ttsize);
 $0 = ((($ttsize)) + 72|0);
 $1 = ((($0)) + 36|0);
 HEAP8[$1>>0] = 0;
 return;
}
function _tt_size_done_bytecode($ftsize) {
 $ftsize = $ftsize|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$ftsize>>2]|0;
 $1 = ((($0)) + 100|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($ftsize)) + 292|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if (!($5)) {
  $6 = ((($4)) + 8|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ((($4)) + 440|0);
  HEAP16[$8>>1] = 0;
  $9 = ((($4)) + 442|0);
  HEAP16[$9>>1] = 0;
  $10 = ((($4)) + 24|0);
  $11 = HEAP32[$10>>2]|0;
  _ft_mem_free($7,$11);
  HEAP32[$10>>2] = 0;
  $12 = ((($4)) + 20|0);
  HEAP32[$12>>2] = 0;
  $13 = ((($4)) + 436|0);
  $14 = HEAP32[$13>>2]|0;
  _ft_mem_free($7,$14);
  HEAP32[$13>>2] = 0;
  $15 = ((($4)) + 432|0);
  HEAP32[$15>>2] = 0;
  $16 = ((($4)) + 428|0);
  HEAP32[$16>>2] = 0;
  $17 = ((($4)) + 392|0);
  $18 = HEAP32[$17>>2]|0;
  _ft_mem_free($7,$18);
  HEAP32[$17>>2] = 0;
  $19 = ((($4)) + 388|0);
  HEAP32[$19>>2] = 0;
  $20 = ((($4)) + 4|0);
  HEAP32[$20>>2] = 0;
  HEAP32[$4>>2] = 0;
  _ft_mem_free($7,$4);
  HEAP32[$3>>2] = 0;
 }
 $21 = ((($ftsize)) + 244|0);
 $22 = HEAP32[$21>>2]|0;
 _ft_mem_free($2,$22);
 HEAP32[$21>>2] = 0;
 $23 = ((($ftsize)) + 240|0);
 HEAP32[$23>>2] = 0;
 $24 = ((($ftsize)) + 252|0);
 $25 = HEAP32[$24>>2]|0;
 _ft_mem_free($2,$25);
 HEAP32[$24>>2] = 0;
 $26 = ((($ftsize)) + 248|0);
 HEAP16[$26>>1] = 0;
 $27 = ((($ftsize)) + 256|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ($28|0)==(0|0);
 if (!($29)) {
  $30 = ((($27)) + 28|0);
  $31 = HEAP32[$30>>2]|0;
  _ft_mem_free($28,$31);
  HEAP32[$30>>2] = 0;
  $32 = ((($27)) + 24|0);
  $33 = HEAP32[$32>>2]|0;
  _ft_mem_free($28,$33);
  HEAP32[$32>>2] = 0;
  $34 = ((($27)) + 16|0);
  $35 = HEAP32[$34>>2]|0;
  _ft_mem_free($28,$35);
  HEAP32[$34>>2] = 0;
  $36 = ((($27)) + 12|0);
  $37 = HEAP32[$36>>2]|0;
  _ft_mem_free($28,$37);
  HEAP32[$36>>2] = 0;
  $38 = ((($27)) + 20|0);
  $39 = HEAP32[$38>>2]|0;
  _ft_mem_free($28,$39);
  HEAP32[$38>>2] = 0;
  ;HEAP32[$27>>2]=0|0;HEAP32[$27+4>>2]=0|0;HEAP32[$27+8>>2]=0|0;
 }
 $40 = ((($ftsize)) + 124|0);
 $41 = HEAP32[$40>>2]|0;
 _ft_mem_free($2,$41);
 HEAP32[$40>>2] = 0;
 $42 = ((($ftsize)) + 136|0);
 $43 = HEAP32[$42>>2]|0;
 _ft_mem_free($2,$43);
 $44 = ((($ftsize)) + 116|0);
 HEAP32[$44>>2] = 0;
 $45 = ((($ftsize)) + 120|0);
 HEAP32[$45>>2] = 0;
 $46 = ((($ftsize)) + 128|0);
 $47 = ((($ftsize)) + 296|0);
 ;HEAP32[$46>>2]=0|0;HEAP32[$46+4>>2]=0|0;HEAP32[$46+8>>2]=0|0;HEAP32[$46+12>>2]=0|0;HEAP32[$46+16>>2]=0|0;
 HEAP32[$47>>2] = -1;
 $48 = ((($ftsize)) + 300|0);
 HEAP32[$48>>2] = -1;
 return;
}
function _tt_slot_init($slot) {
 $slot = $slot|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($slot)) + 156|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = (_FT_GlyphLoader_CreateExtra($2)|0);
 return ($3|0);
}
function _tt_glyph_load($ttslot,$ttsize,$glyph_index,$load_flags) {
 $ttslot = $ttslot|0;
 $ttsize = $ttsize|0;
 $glyph_index = $glyph_index|0;
 $load_flags = $load_flags|0;
 var $$ = 0, $$$i = 0, $$$i$i$i$i = 0, $$0 = 0, $$0$i = 0, $$0$i$ph$i$i$i$i = 0, $$1 = 0, $$1$i$i = 0, $$2 = 0, $$27$i$i = 0, $$9$i = 0, $$pr$pre$i$i$i = 0, $$pre = 0, $$pre$i$i$i$i = 0, $$pre$i$i$i$i$i = 0, $$pre$phi$i$i$i$i$iZ2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0;
 var $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0;
 var $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0;
 var $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0;
 var $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0;
 var $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0;
 var $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0;
 var $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0;
 var $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0;
 var $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0;
 var $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0;
 var $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0;
 var $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0;
 var $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0;
 var $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0;
 var $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0;
 var $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0;
 var $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0;
 var $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0;
 var $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0;
 var $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0;
 var $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0;
 var $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0;
 var $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0;
 var $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0;
 var $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0;
 var $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0;
 var $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0;
 var $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0;
 var $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0;
 var $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0;
 var $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $advance$0$i$i = 0, $advance$1$i$i = 0, $advance$3$ph$i$i = 0, $advance$4$i$i = 0, $bbox$i$i = 0;
 var $error$0$i = 0, $error$i$i$i$i = 0, $error$i$i$i$i$i = 0, $exitcond$i$i = 0, $i$013$i$i = 0, $i$07$i$i$i = 0, $i$15$i$i$i = 0, $load_flags$ = 0, $loader$i = 0, $metrics$i$i = 0, $metrics$i4$i = 0, $nn$02$i$i$i = 0, $nn$02$i$i$i$lcssa = 0, $or$cond$i = 0, $or$cond10$i = 0, $sext$i$i = 0, $sext12$i$i = 0, $sext4$i$i = 0, $top$0$i$i = 0, $top$2$ph$i$i = 0;
 var $top$3$i$i = 0, $umax$i$i$i = 0, $y_scale$0$i$i = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 272|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $bbox$i$i = sp + 232|0;
 $metrics$i4$i = sp + 216|0;
 $error$i$i$i$i$i = sp + 212|0;
 $error$i$i$i$i = sp + 208|0;
 $metrics$i$i = sp + 248|0;
 $loader$i = sp;
 $0 = ((($ttslot)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($ttslot|0)==(0|0);
 if ($2) {
  $$0 = 37;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ($ttsize|0)==(0|0);
 if ($3) {
  $$0 = 36;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ($1|0)==(0|0);
 if ($4) {
  $$0 = 35;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = ((($1)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6>>>0)>($glyph_index>>>0);
 if (!($7)) {
  $8 = ((($1)) + 128|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($9)) + 48|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($11|0)==(0|0);
  if ($12) {
   $$0 = 6;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $13 = $load_flags & 2;
 $14 = ($13|0)==(0);
 if ($14) {
  $$1 = $load_flags;
 } else {
  $15 = ((($1)) + 8|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = $16 & 8192;
  $18 = ($17|0)==(0);
  $19 = $load_flags & -3;
  $load_flags$ = $18 ? $load_flags : $19;
  $20 = $load_flags$ >>> 14;
  $21 = $20 & 2;
  $22 = $21 | $load_flags$;
  $$1 = $22;
 }
 $23 = $$1 & 1025;
 $24 = ($23|0)==(0);
 if ($24) {
  $$2 = $$1;
 } else {
  $25 = ((($1)) + 8|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = $26 >>> 12;
  $28 = $27 & 2;
  $29 = $28 ^ 11;
  $$ = $29 | $$1;
  $$2 = $$;
 }
 $30 = ((($ttsize)) + 112|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = ($31|0)!=(-1);
 $33 = $$2 & 8;
 $34 = ($33|0)==(0);
 $or$cond$i = $34 & $32;
 do {
  if ($or$cond$i) {
   $35 = ((($1)) + 532|0);
   $36 = HEAP32[$35>>2]|0;
   $37 = ((($1)) + 104|0);
   $38 = HEAP32[$37>>2]|0;
   $39 = ((($36)) + 72|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = ((($ttslot)) + 76|0);
   $42 = (FUNCTION_TABLE_iiiiiiii[$40 & 8191]($1,$31,$glyph_index,$$2,$38,$41,$metrics$i$i)|0);
   $43 = ($42|0)==(0);
   if (!($43)) {
    label = 26;
    break;
   }
   $44 = ((($ttslot)) + 110|0);
   HEAP16[$44>>1] = 0;
   $45 = ((($ttslot)) + 108|0);
   HEAP16[$45>>1] = 0;
   $46 = ((($metrics$i$i)) + 2|0);
   $47 = HEAP16[$46>>1]|0;
   $48 = $47&65535;
   $49 = $48 << 6;
   $50 = ((($ttslot)) + 24|0);
   HEAP32[$50>>2] = $49;
   $51 = HEAP16[$metrics$i$i>>1]|0;
   $52 = $51&65535;
   $53 = $52 << 6;
   $54 = ((($ttslot)) + 28|0);
   HEAP32[$54>>2] = $53;
   $55 = ((($metrics$i$i)) + 4|0);
   $56 = HEAP16[$55>>1]|0;
   $57 = $56 << 16 >> 16;
   $58 = $57 << 6;
   $59 = ((($ttslot)) + 32|0);
   HEAP32[$59>>2] = $58;
   $60 = ((($metrics$i$i)) + 6|0);
   $61 = HEAP16[$60>>1]|0;
   $62 = $61 << 16 >> 16;
   $63 = $62 << 6;
   $64 = ((($ttslot)) + 36|0);
   HEAP32[$64>>2] = $63;
   $65 = ((($metrics$i$i)) + 8|0);
   $66 = HEAP16[$65>>1]|0;
   $67 = $66&65535;
   $68 = $67 << 6;
   $69 = ((($ttslot)) + 40|0);
   HEAP32[$69>>2] = $68;
   $70 = ((($metrics$i$i)) + 10|0);
   $71 = HEAP16[$70>>1]|0;
   $72 = $71 << 16 >> 16;
   $73 = $72 << 6;
   $74 = ((($ttslot)) + 44|0);
   HEAP32[$74>>2] = $73;
   $75 = ((($metrics$i$i)) + 12|0);
   $76 = HEAP16[$75>>1]|0;
   $77 = $76 << 16 >> 16;
   $78 = $77 << 6;
   $79 = ((($ttslot)) + 48|0);
   HEAP32[$79>>2] = $78;
   $80 = ((($metrics$i$i)) + 14|0);
   $81 = HEAP16[$80>>1]|0;
   $82 = $81&65535;
   $83 = $82 << 6;
   $84 = ((($ttslot)) + 52|0);
   HEAP32[$84>>2] = $83;
   $85 = ((($ttslot)) + 72|0);
   HEAP32[$85>>2] = 1651078259;
   $86 = $$2 & 16;
   $87 = ($86|0)==(0);
   $88 = ((($ttslot)) + 100|0);
   $$$i = $87 ? $57 : $72;
   $$9$i = $87 ? $62 : $77;
   HEAP32[$88>>2] = $$$i;
   $89 = ((($ttslot)) + 104|0);
   HEAP32[$89>>2] = $$9$i;
   $90 = HEAP32[$0>>2]|0;
   $91 = ((($90)) + 8|0);
   $92 = HEAP32[$91>>2]|0;
   $93 = $92 & 1;
   $94 = ($93|0)==(0);
   if ($94) {
    $$0$i = 0;
   } else {
    $95 = ((($90)) + 104|0);
    $96 = HEAP32[$95>>2]|0;
    _memset(($loader$i|0),0,208)|0;
    $97 = ((($90)) + 128|0);
    $98 = HEAP32[$97>>2]|0;
    $99 = ((($98)) + 48|0);
    $100 = HEAP32[$99>>2]|0;
    $101 = ($100|0)==(0|0);
    do {
     if ($101) {
      $103 = ((($90)) + 508|0);
      $104 = HEAP32[$103>>2]|0;
      $105 = (FUNCTION_TABLE_iiiii[$104 & 8191]($90,1735162214,$96,0)|0);
      $106 = $105 & 255;
      $107 = ($106|0)==(142);
      if ($107) {
       $108 = ((($loader$i)) + 84|0);
       HEAP32[$108>>2] = 0;
       label = 20;
       break;
      }
      $109 = ($105|0)==(0);
      if ($109) {
       $110 = (_FT_Stream_Pos($96)|0);
       $111 = ((($loader$i)) + 84|0);
       HEAP32[$111>>2] = $110;
       label = 20;
      }
     } else {
      $102 = ((($loader$i)) + 84|0);
      HEAP32[$102>>2] = 0;
      label = 20;
     }
    } while(0);
    if ((label|0) == 20) {
     $112 = ((($loader$i)) + 16|0);
     HEAP32[$112>>2] = $$2;
     HEAP32[$loader$i>>2] = $90;
     $113 = ((($loader$i)) + 4|0);
     HEAP32[$113>>2] = $ttsize;
     $114 = ((($loader$i)) + 8|0);
     HEAP32[$114>>2] = $ttslot;
     $115 = ((($loader$i)) + 24|0);
     HEAP32[$115>>2] = $96;
    }
    (_load_truetype_glyph($loader$i,$glyph_index,0,1)|0);
    $116 = ((($loader$i)) + 60|0);
    $117 = HEAP32[$116>>2]|0;
    $118 = ((($ttslot)) + 56|0);
    HEAP32[$118>>2] = $117;
    $119 = ((($loader$i)) + 180|0);
    $120 = HEAP32[$119>>2]|0;
    $121 = ((($ttslot)) + 60|0);
    HEAP32[$121>>2] = $120;
    $122 = HEAP32[$69>>2]|0;
    $123 = ($122|0)!=(0);
    $124 = ($117|0)==(0);
    $or$cond10$i = $124 | $123;
    if (!($or$cond10$i)) {
     $125 = ((($ttsize)) + 16|0);
     $126 = HEAP32[$125>>2]|0;
     $127 = (_FT_MulFix($117,$126)|0);
     HEAP32[$69>>2] = $127;
    }
    $128 = HEAP32[$84>>2]|0;
    $129 = ($128|0)==(0);
    if ($129) {
     $130 = HEAP32[$121>>2]|0;
     $131 = ($130|0)==(0);
     if ($131) {
      $$0$i = 0;
     } else {
      $132 = ((($ttsize)) + 20|0);
      $133 = HEAP32[$132>>2]|0;
      $134 = (_FT_MulFix($130,$133)|0);
      HEAP32[$84>>2] = $134;
      $$0$i = 0;
     }
    } else {
     $$0$i = 0;
    }
   }
  } else {
   label = 26;
  }
 } while(0);
 L41: do {
  if ((label|0) == 26) {
   $135 = $$2 & 1;
   $136 = ($135|0)!=(0);
   if (!($136)) {
    $137 = ((($ttsize)) + 108|0);
    $138 = HEAP8[$137>>0]|0;
    $139 = ($138<<24>>24)==(0);
    if ($139) {
     $$0$i = 36;
     break;
    }
   }
   $140 = $$2 & 16384;
   $141 = ($140|0)==(0);
   if ($141) {
    $142 = $$2 & 128;
    $143 = $142&255;
    $144 = HEAP32[$0>>2]|0;
    $145 = ((($144)) + 104|0);
    $146 = HEAP32[$145>>2]|0;
    _memset(($loader$i|0),0,208)|0;
    $147 = $$2 & 2;
    $148 = ($147|0)==(0);
    if ($148) {
     $149 = ((($ttsize)) + 296|0);
     $150 = HEAP32[$149>>2]|0;
     $151 = ($150|0)<(0);
     L49: do {
      if ($151) {
       $155 = HEAP32[$ttsize>>2]|0;
       $156 = ((($155)) + 100|0);
       $157 = HEAP32[$156>>2]|0;
       $158 = ((($ttsize)) + 124|0);
       $159 = HEAP32[$158>>2]|0;
       _ft_mem_free($157,$159);
       HEAP32[$158>>2] = 0;
       $160 = ((($ttsize)) + 136|0);
       $161 = HEAP32[$160>>2]|0;
       _ft_mem_free($157,$161);
       HEAP32[$160>>2] = 0;
       $162 = ((($ttsize)) + 244|0);
       $163 = HEAP32[$162>>2]|0;
       _ft_mem_free($157,$163);
       HEAP32[$162>>2] = 0;
       $164 = ((($ttsize)) + 252|0);
       $165 = HEAP32[$164>>2]|0;
       _ft_mem_free($157,$165);
       HEAP32[$164>>2] = 0;
       $166 = ((($ttsize)) + 292|0);
       $167 = HEAP32[$166>>2]|0;
       $168 = ($167|0)==(0|0);
       if (!($168)) {
        $169 = ((($167)) + 8|0);
        $170 = HEAP32[$169>>2]|0;
        $171 = ((($167)) + 440|0);
        HEAP16[$171>>1] = 0;
        $172 = ((($167)) + 442|0);
        HEAP16[$172>>1] = 0;
        $173 = ((($167)) + 24|0);
        $174 = HEAP32[$173>>2]|0;
        _ft_mem_free($170,$174);
        HEAP32[$173>>2] = 0;
        $175 = ((($167)) + 20|0);
        HEAP32[$175>>2] = 0;
        $176 = ((($167)) + 436|0);
        $177 = HEAP32[$176>>2]|0;
        _ft_mem_free($170,$177);
        HEAP32[$176>>2] = 0;
        $178 = ((($167)) + 432|0);
        HEAP32[$178>>2] = 0;
        $179 = ((($167)) + 428|0);
        HEAP32[$179>>2] = 0;
        $180 = ((($167)) + 392|0);
        $181 = HEAP32[$180>>2]|0;
        _ft_mem_free($170,$181);
        HEAP32[$180>>2] = 0;
        $182 = ((($167)) + 388|0);
        HEAP32[$182>>2] = 0;
        $183 = ((($167)) + 4|0);
        HEAP32[$183>>2] = 0;
        HEAP32[$167>>2] = 0;
        _ft_mem_free($170,$167);
       }
       $184 = ((($ttsize)) + 256|0);
       $185 = HEAP32[$184>>2]|0;
       $186 = ($185|0)==(0|0);
       if (!($186)) {
        $187 = ((($184)) + 28|0);
        $188 = HEAP32[$187>>2]|0;
        _ft_mem_free($185,$188);
        HEAP32[$187>>2] = 0;
        $189 = ((($184)) + 24|0);
        $190 = HEAP32[$189>>2]|0;
        _ft_mem_free($185,$190);
        HEAP32[$189>>2] = 0;
        $191 = ((($184)) + 16|0);
        $192 = HEAP32[$191>>2]|0;
        _ft_mem_free($185,$192);
        HEAP32[$191>>2] = 0;
        $193 = ((($184)) + 12|0);
        $194 = HEAP32[$193>>2]|0;
        _ft_mem_free($185,$194);
        HEAP32[$193>>2] = 0;
        $195 = ((($184)) + 20|0);
        $196 = HEAP32[$195>>2]|0;
        _ft_mem_free($185,$196);
        HEAP32[$195>>2] = 0;
        ;HEAP32[$184>>2]=0|0;HEAP32[$184+4>>2]=0|0;HEAP32[$184+8>>2]=0|0;
       }
       HEAP32[$149>>2] = -1;
       $197 = ((($ttsize)) + 300|0);
       HEAP32[$197>>2] = -1;
       $198 = ((($155)) + 96|0);
       $199 = HEAP32[$198>>2]|0;
       $200 = (_TT_New_Context($199)|0);
       HEAP32[$166>>2] = $200;
       $201 = ((($155)) + 280|0);
       $202 = HEAP32[$201>>2]|0;
       $203 = $202 & 65535;
       $204 = ((($ttsize)) + 120|0);
       HEAP32[$204>>2] = $203;
       $205 = $202 >>> 16;
       $206 = ((($ttsize)) + 132|0);
       HEAP32[$206>>2] = $205;
       $207 = ((($ttsize)) + 116|0);
       HEAP32[$207>>2] = 0;
       $208 = ((($ttsize)) + 128|0);
       HEAP32[$208>>2] = 0;
       $209 = ((($ttsize)) + 140|0);
       HEAP32[$209>>2] = 0;
       $210 = ((($ttsize)) + 144|0);
       HEAP32[$210>>2] = 0;
       $211 = ((($155)) + 644|0);
       $212 = HEAP32[$211>>2]|0;
       $213 = ((($ttsize)) + 240|0);
       HEAP32[$213>>2] = $212;
       $214 = ((($155)) + 278|0);
       $215 = HEAP16[$214>>1]|0;
       $216 = ((($ttsize)) + 248|0);
       HEAP16[$216>>1] = $215;
       $217 = ((($ttsize)) + 72|0);
       $218 = ((($217)) + 37|0);
       HEAP8[$218>>0] = 0;
       $219 = ((($217)) + 38|0);
       HEAP8[$219>>0] = 0;
       $220 = ((($217)) + 20|0);
       ;HEAP32[$220>>2]=0|0;HEAP32[$220+4>>2]=0|0;HEAP32[$220+8>>2]=0|0;HEAP32[$220+12>>2]=0|0;
       $221 = (_ft_mem_realloc($157,24,0,$203,0,$error$i$i$i$i)|0);
       HEAP32[$158>>2] = $221;
       $222 = HEAP32[$error$i$i$i$i>>2]|0;
       $223 = ($222|0)==(0);
       L57: do {
        if ($223) {
         $224 = HEAP32[$206>>2]|0;
         $225 = (_ft_mem_realloc($157,24,0,$224,0,$error$i$i$i$i)|0);
         HEAP32[$160>>2] = $225;
         $226 = HEAP32[$error$i$i$i$i>>2]|0;
         $227 = ($226|0)==(0);
         if ($227) {
          $228 = HEAP32[$213>>2]|0;
          $229 = (_ft_mem_realloc($157,4,0,$228,0,$error$i$i$i$i)|0);
          HEAP32[$162>>2] = $229;
          $230 = HEAP32[$error$i$i$i$i>>2]|0;
          $231 = ($230|0)==(0);
          if ($231) {
           $232 = HEAP16[$216>>1]|0;
           $233 = $232&65535;
           $234 = (_ft_mem_realloc($157,4,0,$233,0,$error$i$i$i$i)|0);
           HEAP32[$164>>2] = $234;
           $235 = HEAP32[$error$i$i$i$i>>2]|0;
           $236 = ($235|0)==(0);
           if (!($236)) {
            break;
           }
           $237 = ((($155)) + 276|0);
           $238 = HEAP16[$237>>1]|0;
           $239 = $238&65535;
           $240 = (($239) + 4)|0;
           $241 = $240&65535;
           dest=$184; stop=dest+36|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
           HEAP32[$184>>2] = $157;
           $242 = $240 & 65535;
           $243 = (_ft_mem_realloc($157,8,0,$242,0,$error$i$i$i$i$i)|0);
           $244 = ((($184)) + 12|0);
           HEAP32[$244>>2] = $243;
           $245 = HEAP32[$error$i$i$i$i$i>>2]|0;
           $246 = ($245|0)==(0);
           do {
            if ($246) {
             $247 = (_ft_mem_realloc($157,8,0,$242,0,$error$i$i$i$i$i)|0);
             $248 = ((($184)) + 16|0);
             HEAP32[$248>>2] = $247;
             $249 = HEAP32[$error$i$i$i$i$i>>2]|0;
             $250 = ($249|0)==(0);
             if (!($250)) {
              $265 = $249;
              label = 45;
              break;
             }
             $251 = (_ft_mem_realloc($157,8,0,$242,0,$error$i$i$i$i$i)|0);
             $252 = ((($184)) + 20|0);
             HEAP32[$252>>2] = $251;
             $253 = HEAP32[$error$i$i$i$i$i>>2]|0;
             $254 = ($253|0)==(0);
             if (!($254)) {
              $265 = $253;
              label = 45;
              break;
             }
             $255 = (_ft_mem_realloc($157,1,0,$242,0,$error$i$i$i$i$i)|0);
             $256 = ((($184)) + 24|0);
             HEAP32[$256>>2] = $255;
             $257 = HEAP32[$error$i$i$i$i$i>>2]|0;
             $258 = ($257|0)==(0);
             if (!($258)) {
              $265 = $257;
              label = 45;
              break;
             }
             $259 = (_ft_mem_realloc($157,2,0,0,0,$error$i$i$i$i$i)|0);
             $260 = ((($184)) + 28|0);
             HEAP32[$260>>2] = $259;
             $261 = HEAP32[$error$i$i$i$i$i>>2]|0;
             $262 = ($261|0)==(0);
             if (!($262)) {
              $265 = $261;
              label = 45;
              break;
             }
             $266 = ((($184)) + 4|0);
             HEAP16[$266>>1] = $241;
             $267 = ((($184)) + 6|0);
             HEAP16[$267>>1] = 0;
             HEAP32[$error$i$i$i$i>>2] = 0;
            } else {
             $265 = $245;
             label = 45;
            }
           } while(0);
           do {
            if ((label|0) == 45) {
             $263 = HEAP32[$184>>2]|0;
             $264 = ($263|0)==(0|0);
             if ($264) {
              HEAP32[$error$i$i$i$i>>2] = $265;
              break L57;
             } else {
              $268 = ((($184)) + 28|0);
              $269 = HEAP32[$268>>2]|0;
              _ft_mem_free($263,$269);
              HEAP32[$268>>2] = 0;
              $270 = ((($184)) + 24|0);
              $271 = HEAP32[$270>>2]|0;
              _ft_mem_free($263,$271);
              HEAP32[$270>>2] = 0;
              $272 = ((($184)) + 16|0);
              $273 = HEAP32[$272>>2]|0;
              _ft_mem_free($263,$273);
              HEAP32[$272>>2] = 0;
              $274 = HEAP32[$244>>2]|0;
              _ft_mem_free($263,$274);
              HEAP32[$244>>2] = 0;
              $275 = ((($184)) + 20|0);
              $276 = HEAP32[$275>>2]|0;
              _ft_mem_free($263,$276);
              HEAP32[$275>>2] = 0;
              ;HEAP32[$184>>2]=0|0;HEAP32[$184+4>>2]=0|0;HEAP32[$184+8>>2]=0|0;
              $$pre$i$i$i$i$i = HEAP32[$error$i$i$i$i$i>>2]|0;
              HEAP32[$error$i$i$i$i>>2] = $$pre$i$i$i$i$i;
              $277 = ($$pre$i$i$i$i$i|0)==(0);
              if ($277) {
               break;
              } else {
               break L57;
              }
             }
            }
           } while(0);
           $278 = ((($184)) + 8|0);
           HEAP16[$278>>1] = $241;
           $279 = ((($ttsize)) + 172|0);
           dest=$279; src=56184; stop=dest+68|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
           $280 = HEAP32[$198>>2]|0;
           $281 = ((($280)) + 4|0);
           $282 = HEAP32[$281>>2]|0;
           $283 = ((($282)) + 172|0);
           $284 = HEAP32[$283>>2]|0;
           $285 = ((($155)) + 652|0);
           $286 = ($284|0)==(0|0);
           $$$i$i$i$i = $286 ? 4409 : $284;
           HEAP32[$285>>2] = $$$i$i$i$i;
           $287 = HEAP32[$ttsize>>2]|0;
           $288 = HEAP32[$166>>2]|0;
           $289 = (_TT_Load_Context($288,$287,$ttsize)|0);
           $290 = ($289|0)==(0);
           do {
            if ($290) {
             $291 = ((($288)) + 428|0);
             HEAP32[$291>>2] = 0;
             $292 = ((($288)) + 16|0);
             HEAP32[$292>>2] = 0;
             $293 = ((($288)) + 476|0);
             HEAP32[$293>>2] = 64;
             $294 = ((($288)) + 480|0);
             HEAP32[$294>>2] = 0;
             $295 = ((($288)) + 484|0);
             HEAP32[$295>>2] = 0;
             $296 = ((($288)) + 488|0);
             HEAP8[$296>>0] = 0;
             $297 = ((($288)) + 564|0);
             HEAP32[$297>>2] = 16384;
             $298 = ((($288)) + 561|0);
             HEAP8[$298>>0] = $143;
             $299 = ((($288)) + 216|0);
             $300 = ((($288)) + 252|0);
             HEAP16[$300>>1] = 0;
             $301 = ((($288)) + 260|0);
             HEAP32[$301>>2] = 0;
             $302 = ((($288)) + 256|0);
             ;HEAP32[$299>>2]=0|0;HEAP32[$299+4>>2]=0|0;HEAP32[$299+8>>2]=0|0;
             HEAP32[$302>>2] = 65536;
             $303 = ((($287)) + 632|0);
             $304 = HEAP32[$303>>2]|0;
             $305 = ((($287)) + 628|0);
             $306 = HEAP32[$305>>2]|0;
             $307 = ((($288)) + 444|0);
             HEAP32[$307>>2] = $304;
             $308 = ((($288)) + 448|0);
             HEAP32[$308>>2] = $306;
             $309 = ((($288)) + 452|0);
             $310 = ($306|0)==(0);
             ;HEAP32[$309>>2]=0|0;HEAP32[$309+4>>2]=0|0;HEAP32[$309+8>>2]=0|0;HEAP32[$309+12>>2]=0|0;
             if ($310) {
              HEAP32[$149>>2] = 0;
             } else {
              $311 = ((($288)) + 356|0);
              HEAP32[$311>>2] = $304;
              $312 = ((($288)) + 364|0);
              HEAP32[$312>>2] = $306;
              $313 = ((($288)) + 360|0);
              HEAP32[$313>>2] = 0;
              $314 = ((($288)) + 352|0);
              HEAP32[$314>>2] = 1;
              $315 = ((($287)) + 652|0);
              $316 = HEAP32[$315>>2]|0;
              $317 = (FUNCTION_TABLE_ii[$316 & 8191]($288)|0);
              HEAP32[$149>>2] = $317;
              $318 = ($317|0)==(0);
              if (!($318)) {
               $$0$i$ph$i$i$i$i = $317;
               break;
              }
             }
             $$pre$phi$i$i$i$i$iZ2D = ((($288)) + 444|0);
             $319 = ((($288)) + 396|0);
             $320 = HEAP32[$319>>2]|0;
             HEAP32[$207>>2] = $320;
             $321 = ((($288)) + 408|0);
             $322 = HEAP32[$321>>2]|0;
             HEAP32[$208>>2] = $322;
             $323 = ((($288)) + 420|0);
             $324 = HEAP32[$323>>2]|0;
             HEAP32[$209>>2] = $324;
             $325 = ((($288)) + 424|0);
             $326 = HEAP32[$325>>2]|0;
             HEAP32[$210>>2] = $326;
             $327 = ((($ttsize)) + 148|0);
             $328 = $$pre$phi$i$i$i$i$iZ2D;
             $329 = $328;
             $330 = HEAP32[$329>>2]|0;
             $331 = (($328) + 4)|0;
             $332 = $331;
             $333 = HEAP32[$332>>2]|0;
             $334 = $327;
             $335 = $334;
             HEAP32[$335>>2] = $330;
             $336 = (($334) + 4)|0;
             $337 = $336;
             HEAP32[$337>>2] = $333;
             $338 = ((($ttsize)) + 156|0);
             $339 = ((($288)) + 452|0);
             $340 = $339;
             $341 = $340;
             $342 = HEAP32[$341>>2]|0;
             $343 = (($340) + 4)|0;
             $344 = $343;
             $345 = HEAP32[$344>>2]|0;
             $346 = $338;
             $347 = $346;
             HEAP32[$347>>2] = $342;
             $348 = (($346) + 4)|0;
             $349 = $348;
             HEAP32[$349>>2] = $345;
             $350 = ((($ttsize)) + 164|0);
             $351 = ((($288)) + 460|0);
             $352 = $351;
             $353 = $352;
             $354 = HEAP32[$353>>2]|0;
             $355 = (($352) + 4)|0;
             $356 = $355;
             $357 = HEAP32[$356>>2]|0;
             $358 = $350;
             $359 = $358;
             HEAP32[$359>>2] = $354;
             $360 = (($358) + 4)|0;
             $361 = $360;
             HEAP32[$361>>2] = $357;
             $$pre$phiZ2D = $197;
             label = 59;
             break L49;
            } else {
             $$0$i$ph$i$i$i$i = $289;
            }
           } while(0);
           HEAP32[$error$i$i$i$i>>2] = $$0$i$ph$i$i$i$i;
          }
         }
        }
       } while(0);
       _tt_size_done_bytecode($ttsize);
       $$pre$i$i$i$i = HEAP32[$error$i$i$i$i>>2]|0;
       $362 = ($$pre$i$i$i$i|0)==(0);
       if (!($362)) {
        $$0$i = $$pre$i$i$i$i;
        break L41;
       }
       $$pr$pre$i$i$i = HEAP32[$149>>2]|0;
       $364 = $$pr$pre$i$i$i;
       label = 57;
      } else {
       $152 = ((($ttsize)) + 300|0);
       $153 = HEAP32[$152>>2]|0;
       $154 = ($153|0)<(0);
       if ($154) {
        $364 = $150;
        label = 57;
       } else {
        $409 = ($150|0)==(0);
        if (!($409)) {
         $$0$i = $150;
         break L41;
        }
        $410 = ($153|0)==(0);
        if (!($410)) {
         $$0$i = $153;
         break L41;
        }
       }
      }
     } while(0);
     if ((label|0) == 57) {
      $363 = ($364|0)==(0);
      if ($363) {
       $$pre = ((($ttsize)) + 300|0);
       $$pre$phiZ2D = $$pre;
       label = 59;
      }
     }
     if ((label|0) == 59) {
      $365 = HEAP32[$$pre$phiZ2D>>2]|0;
      $366 = ($365|0)<(0);
      if ($366) {
       $367 = ((($ttsize)) + 240|0);
       $368 = HEAP32[$367>>2]|0;
       $369 = ($368|0)==(0);
       if (!($369)) {
        $370 = HEAP32[$ttsize>>2]|0;
        $371 = ((($370)) + 648|0);
        $372 = ((($ttsize)) + 88|0);
        $373 = ((($ttsize)) + 244|0);
        $i$07$i$i$i = 0;
        while(1) {
         $382 = HEAP32[$371>>2]|0;
         $383 = (($382) + ($i$07$i$i$i<<1)|0);
         $384 = HEAP16[$383>>1]|0;
         $385 = $384 << 16 >> 16;
         $386 = HEAP32[$372>>2]|0;
         $387 = (_FT_MulFix($385,$386)|0);
         $388 = HEAP32[$373>>2]|0;
         $389 = (($388) + ($i$07$i$i$i<<2)|0);
         HEAP32[$389>>2] = $387;
         $390 = (($i$07$i$i$i) + 1)|0;
         $391 = HEAP32[$367>>2]|0;
         $392 = ($390>>>0)<($391>>>0);
         if ($392) {
          $i$07$i$i$i = $390;
         } else {
          break;
         }
        }
       }
       $374 = ((($ttsize)) + 264|0);
       $375 = HEAP16[$374>>1]|0;
       $376 = ($375<<16>>16)==(0);
       if (!($376)) {
        $377 = ((($ttsize)) + 268|0);
        $378 = HEAP32[$377>>2]|0;
        $379 = ((($ttsize)) + 272|0);
        $380 = HEAP32[$379>>2]|0;
        $381 = $375&65535;
        $i$15$i$i$i = 0;
        while(1) {
         $401 = (($378) + ($i$15$i$i$i<<3)|0);
         HEAP32[$401>>2] = 0;
         $402 = (((($378) + ($i$15$i$i$i<<3)|0)) + 4|0);
         HEAP32[$402>>2] = 0;
         $403 = (($380) + ($i$15$i$i$i<<3)|0);
         HEAP32[$403>>2] = 0;
         $404 = (((($380) + ($i$15$i$i$i<<3)|0)) + 4|0);
         HEAP32[$404>>2] = 0;
         $405 = (($i$15$i$i$i) + 1)|0;
         $exitcond$i$i = ($405|0)==($381|0);
         if ($exitcond$i$i) {
          break;
         } else {
          $i$15$i$i$i = $405;
         }
        }
       }
       $393 = ((($ttsize)) + 248|0);
       $394 = HEAP16[$393>>1]|0;
       $395 = ($394<<16>>16)==(0);
       if (!($395)) {
        $396 = ((($ttsize)) + 252|0);
        $397 = HEAP32[$396>>2]|0;
        $398 = $394&65535;
        $399 = ($398>>>0)>(1);
        $umax$i$i$i = $399 ? $398 : 1;
        $400 = $umax$i$i$i << 2;
        _memset(($397|0),0,($400|0))|0;
       }
       $406 = ((($ttsize)) + 172|0);
       dest=$406; src=56184; stop=dest+68|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
       $407 = (_tt_size_run_prep($ttsize,$143)|0);
       $408 = ($407|0)==(0);
       if (!($408)) {
        $$0$i = $407;
        break;
       }
      }
     }
     $411 = ((($ttsize)) + 292|0);
     $412 = HEAP32[$411>>2]|0;
     $413 = ($412|0)==(0|0);
     if ($413) {
      $$0$i = 153;
      break;
     }
     $414 = $$2 & 983040;
     $415 = ($414|0)!=(131072);
     $416 = $415&1;
     $417 = (_TT_Load_Context($412,$144,$ttsize)|0);
     $418 = ($417|0)==(0);
     if (!($418)) {
      $$0$i = $417;
      break;
     }
     $419 = $415&1;
     $420 = ((($412)) + 608|0);
     $421 = HEAP8[$420>>0]|0;
     $422 = $421&255;
     $423 = ($419|0)==($422|0);
     if (!($423)) {
      HEAP8[$420>>0] = $416;
      $424 = ((($ttsize)) + 240|0);
      $425 = HEAP32[$424>>2]|0;
      $426 = ($425|0)==(0);
      if (!($426)) {
       $427 = ((($144)) + 648|0);
       $428 = ((($ttsize)) + 88|0);
       $429 = ((($ttsize)) + 244|0);
       $i$013$i$i = 0;
       while(1) {
        $430 = HEAP32[$427>>2]|0;
        $431 = (($430) + ($i$013$i$i<<1)|0);
        $432 = HEAP16[$431>>1]|0;
        $433 = $432 << 16 >> 16;
        $434 = HEAP32[$428>>2]|0;
        $435 = (_FT_MulFix($433,$434)|0);
        $436 = HEAP32[$429>>2]|0;
        $437 = (($436) + ($i$013$i$i<<2)|0);
        HEAP32[$437>>2] = $435;
        $438 = (($i$013$i$i) + 1)|0;
        $439 = HEAP32[$424>>2]|0;
        $440 = ($438>>>0)<($439>>>0);
        if ($440) {
         $i$013$i$i = $438;
        } else {
         break;
        }
       }
      }
      $441 = (_tt_size_run_prep($ttsize,$143)|0);
      $442 = ($441|0)==(0);
      if (!($442)) {
       $$0$i = $441;
       break;
      }
     }
     $443 = ((($412)) + 336|0);
     $444 = HEAP8[$443>>0]|0;
     $445 = $444 & 1;
     $446 = $445&255;
     $447 = $446 << 1;
     $448 = $447 | $$2;
     $449 = $444 & 2;
     $450 = ($449<<24>>24)==(0);
     if (!($450)) {
      $451 = ((($412)) + 284|0);
      dest=$451; src=56184; stop=dest+68|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
     }
     $452 = ((($412)) + 561|0);
     HEAP8[$452>>0] = $143;
     $453 = ((($loader$i)) + 160|0);
     HEAP32[$453>>2] = $412;
     $454 = ((($412)) + 392|0);
     $455 = HEAP32[$454>>2]|0;
     $456 = ((($loader$i)) + 164|0);
     HEAP32[$456>>2] = $455;
     $$27$i$i = $448;
    } else {
     $$27$i$i = $$2;
    }
    $457 = ((($144)) + 128|0);
    $458 = HEAP32[$457>>2]|0;
    $459 = ((($458)) + 48|0);
    $460 = HEAP32[$459>>2]|0;
    $461 = ($460|0)==(0|0);
    do {
     if ($461) {
      $463 = ((($144)) + 508|0);
      $464 = HEAP32[$463>>2]|0;
      $465 = (FUNCTION_TABLE_iiiii[$464 & 8191]($144,1735162214,$146,0)|0);
      $466 = $465 & 255;
      $467 = ($466|0)==(142);
      if ($467) {
       $468 = ((($loader$i)) + 84|0);
       HEAP32[$468>>2] = 0;
       break;
      }
      $469 = ($465|0)==(0);
      if (!($469)) {
       $$0$i = $465;
       break L41;
      }
      $470 = (_FT_Stream_Pos($146)|0);
      $471 = ((($loader$i)) + 84|0);
      HEAP32[$471>>2] = $470;
     } else {
      $462 = ((($loader$i)) + 84|0);
      HEAP32[$462>>2] = 0;
     }
    } while(0);
    $472 = ((($ttslot)) + 156|0);
    $473 = HEAP32[$472>>2]|0;
    $474 = HEAP32[$473>>2]|0;
    _FT_GlyphLoader_Rewind($474);
    $475 = ((($loader$i)) + 12|0);
    HEAP32[$475>>2] = $474;
    $476 = ((($loader$i)) + 16|0);
    HEAP32[$476>>2] = $$27$i$i;
    HEAP32[$loader$i>>2] = $144;
    $477 = ((($loader$i)) + 4|0);
    HEAP32[$477>>2] = $ttsize;
    $478 = ((($loader$i)) + 8|0);
    HEAP32[$478>>2] = $ttslot;
    $479 = ((($loader$i)) + 24|0);
    HEAP32[$479>>2] = $146;
    $480 = ((($ttslot)) + 72|0);
    HEAP32[$480>>2] = 1869968492;
    $481 = ((($ttslot)) + 128|0);
    HEAP32[$481>>2] = 0;
    $482 = ((($ttslot)) + 108|0);
    $483 = ((($ttslot)) + 124|0);
    HEAP32[$483>>2] = 0;
    $484 = (_load_truetype_glyph($loader$i,$glyph_index,0,0)|0);
    $485 = ($484|0)==(0);
    if ($485) {
     $486 = HEAP32[$480>>2]|0;
     $487 = ($486|0)==(1668246896);
     $488 = HEAP32[$475>>2]|0;
     if ($487) {
      $489 = ((($488)) + 48|0);
      $490 = HEAP32[$489>>2]|0;
      HEAP32[$481>>2] = $490;
      $491 = ((($488)) + 52|0);
      $492 = HEAP32[$491>>2]|0;
      $493 = ((($ttslot)) + 132|0);
      HEAP32[$493>>2] = $492;
     } else {
      $494 = ((($488)) + 20|0);
      ;HEAP32[$482>>2]=HEAP32[$494>>2]|0;HEAP32[$482+4>>2]=HEAP32[$494+4>>2]|0;HEAP32[$482+8>>2]=HEAP32[$494+8>>2]|0;HEAP32[$482+12>>2]=HEAP32[$494+12>>2]|0;HEAP32[$482+16>>2]=HEAP32[$494+16>>2]|0;
      $495 = HEAP32[$483>>2]|0;
      $496 = $495 & -513;
      HEAP32[$483>>2] = $496;
      $497 = ((($loader$i)) + 68|0);
      $498 = HEAP32[$497>>2]|0;
      $499 = ($498|0)==(0);
      if (!($499)) {
       $500 = (0 - ($498))|0;
       _FT_Outline_Translate($482,$500,0);
      }
     }
     L135: do {
      if ($148) {
       $501 = ((($loader$i)) + 160|0);
       $502 = HEAP32[$501>>2]|0;
       $503 = ((($502)) + 337|0);
       $504 = HEAP8[$503>>0]|0;
       $505 = ($504<<24>>24)==(0);
       if ($505) {
        $516 = HEAP32[$483>>2]|0;
        $517 = $516 | 8;
        HEAP32[$483>>2] = $517;
        break;
       }
       $506 = ((($502)) + 340|0);
       $507 = HEAP32[$506>>2]|0;
       switch ($507|0) {
       case 1:  {
        break L135;
        break;
       }
       case 0:  {
        $508 = HEAP32[$483>>2]|0;
        $509 = $508 | 32;
        HEAP32[$483>>2] = $509;
        break L135;
        break;
       }
       case 4:  {
        $510 = HEAP32[$483>>2]|0;
        $511 = $510 | 48;
        HEAP32[$483>>2] = $511;
        break L135;
        break;
       }
       case 5:  {
        $512 = HEAP32[$483>>2]|0;
        $513 = $512 | 16;
        HEAP32[$483>>2] = $513;
        break L135;
        break;
       }
       default: {
        $514 = HEAP32[$483>>2]|0;
        $515 = $514 | 8;
        HEAP32[$483>>2] = $515;
        break L135;
       }
       }
      }
     } while(0);
     $518 = HEAP32[$loader$i>>2]|0;
     $519 = HEAP32[$478>>2]|0;
     $520 = HEAP32[$477>>2]|0;
     $521 = HEAP32[$476>>2]|0;
     $522 = $521 & 1;
     $523 = ($522|0)==(0);
     if ($523) {
      $524 = ((($520)) + 20|0);
      $525 = HEAP32[$524>>2]|0;
      $y_scale$0$i$i = $525;
     } else {
      $y_scale$0$i$i = 65536;
     }
     $526 = ((($519)) + 72|0);
     $527 = HEAP32[$526>>2]|0;
     $528 = ($527|0)==(1668246896);
     if ($528) {
      $530 = ((($loader$i)) + 36|0);
      ;HEAP32[$bbox$i$i>>2]=HEAP32[$530>>2]|0;HEAP32[$bbox$i$i+4>>2]=HEAP32[$530+4>>2]|0;HEAP32[$bbox$i$i+8>>2]=HEAP32[$530+8>>2]|0;HEAP32[$bbox$i$i+12>>2]=HEAP32[$530+12>>2]|0;
     } else {
      $529 = ((($519)) + 108|0);
      _FT_Outline_Get_CBox($529,$bbox$i$i);
     }
     $531 = ((($loader$i)) + 60|0);
     $532 = HEAP32[$531>>2]|0;
     $533 = ((($519)) + 56|0);
     HEAP32[$533>>2] = $532;
     $534 = HEAP32[$bbox$i$i>>2]|0;
     $535 = ((($519)) + 32|0);
     HEAP32[$535>>2] = $534;
     $536 = ((($bbox$i$i)) + 12|0);
     $537 = HEAP32[$536>>2]|0;
     $538 = ((($519)) + 36|0);
     HEAP32[$538>>2] = $537;
     $539 = ((($loader$i)) + 76|0);
     $540 = HEAP32[$539>>2]|0;
     $541 = ((($loader$i)) + 68|0);
     $542 = HEAP32[$541>>2]|0;
     $543 = (($540) - ($542))|0;
     $544 = ((($519)) + 40|0);
     HEAP32[$544>>2] = $543;
     $545 = ((($518)) + 480|0);
     $546 = HEAP32[$545>>2]|0;
     $547 = ($546|0)==(0);
     L153: do {
      if ($547) {
       $548 = HEAP32[$476>>2]|0;
       $549 = $548 & 2;
       $550 = ($549|0)==(0);
       if (!($550)) {
        break;
       }
       $551 = ((($520)) + 12|0);
       $552 = HEAP16[$551>>1]|0;
       $553 = $552&65535;
       $554 = ((($518)) + 716|0);
       $555 = HEAP32[$554>>2]|0;
       $556 = ((($518)) + 704|0);
       $557 = HEAP32[$556>>2]|0;
       $558 = ((($557)) + 8|0);
       $559 = ((($518)) + 712|0);
       $560 = HEAP32[$559>>2]|0;
       $561 = ($560|0)==(0);
       if ($561) {
        break;
       }
       $562 = ((($518)) + 720|0);
       $563 = HEAP32[$562>>2]|0;
       $nn$02$i$i$i = 0;
       while(1) {
        $564 = (($563) + ($nn$02$i$i$i)|0);
        $565 = HEAP8[$564>>0]|0;
        $566 = $565&255;
        $567 = ($566|0)==($553|0);
        if ($567) {
         $nn$02$i$i$i$lcssa = $nn$02$i$i$i;
         break;
        }
        $570 = (($nn$02$i$i$i) + 1)|0;
        $571 = ($570>>>0)<($560>>>0);
        if ($571) {
         $nn$02$i$i$i = $570;
        } else {
         break L153;
        }
       }
       $568 = (($glyph_index) + 2)|0;
       $569 = ($568>>>0)<($555>>>0);
       if (!($569)) {
        break;
       }
       $572 = Math_imul($nn$02$i$i$i$lcssa, $555)|0;
       $573 = (($558) + ($572)|0);
       $574 = (($573) + ($568)|0);
       $575 = HEAP8[$574>>0]|0;
       $576 = $575&255;
       $577 = $576 << 6;
       HEAP32[$544>>2] = $577;
      }
     } while(0);
     $578 = ((($bbox$i$i)) + 8|0);
     $579 = HEAP32[$578>>2]|0;
     $580 = (($579) - ($534))|0;
     $581 = ((($519)) + 24|0);
     HEAP32[$581>>2] = $580;
     $582 = ((($bbox$i$i)) + 4|0);
     $583 = HEAP32[$582>>2]|0;
     $584 = (($537) - ($583))|0;
     $585 = ((($519)) + 28|0);
     HEAP32[$585>>2] = $584;
     $586 = ((($518)) + 292|0);
     $587 = HEAP8[$586>>0]|0;
     $588 = ($587<<24>>24)==(0);
     do {
      if ($588) {
       label = 117;
      } else {
       $589 = ((($518)) + 330|0);
       $590 = HEAP16[$589>>1]|0;
       $591 = ($590<<16>>16)==(0);
       if ($591) {
        label = 117;
        break;
       }
       $592 = ((($loader$i)) + 188|0);
       $593 = HEAP32[$592>>2]|0;
       $594 = (($593) - ($537))|0;
       $595 = (_FT_DivFix($594,$y_scale$0$i$i)|0);
       $sext4$i$i = $595 << 16;
       $596 = $sext4$i$i >> 16;
       $597 = HEAP32[$592>>2]|0;
       $598 = ((($loader$i)) + 196|0);
       $599 = HEAP32[$598>>2]|0;
       $600 = ($597|0)>($599|0);
       if (!($600)) {
        $advance$1$i$i = 0;$top$0$i$i = $596;
        break;
       }
       $601 = (($597) - ($599))|0;
       $602 = (_FT_DivFix($601,$y_scale$0$i$i)|0);
       $603 = $602 & 65535;
       $advance$1$i$i = $603;$top$0$i$i = $596;
      }
     } while(0);
     if ((label|0) == 117) {
      $604 = (_FT_DivFix($584,$y_scale$0$i$i)|0);
      $sext$i$i = $604 << 16;
      $605 = $sext$i$i >> 16;
      $606 = ((($518)) + 364|0);
      $607 = HEAP16[$606>>1]|0;
      $608 = ($607<<16>>16)==(-1);
      if ($608) {
       $616 = ((($518)) + 220|0);
       $617 = HEAP32[$616>>2]|0;
       $sext12$i$i = $617 << 16;
       $618 = $sext12$i$i >> 16;
       $619 = $617 >> 16;
       $620 = (($618) - ($619))|0;
       $advance$0$i$i = $620;
      } else {
       $609 = ((($518)) + 434|0);
       $610 = HEAP16[$609>>1]|0;
       $611 = $610 << 16 >> 16;
       $612 = ((($518)) + 436|0);
       $613 = HEAP16[$612>>1]|0;
       $614 = $613 << 16 >> 16;
       $615 = (($611) - ($614))|0;
       $advance$0$i$i = $615;
      }
      $621 = (($advance$0$i$i) - ($605))|0;
      $622 = (($621|0) / 2)&-1;
      $advance$1$i$i = $advance$0$i$i;$top$0$i$i = $622;
     }
     $623 = ((($518)) + 128|0);
     $624 = HEAP32[$623>>2]|0;
     $625 = ((($624)) + 48|0);
     $626 = HEAP32[$625>>2]|0;
     $627 = ($626|0)==(0|0);
     do {
      if ($627) {
       $advance$3$ph$i$i = $advance$1$i$i;$top$2$ph$i$i = $top$0$i$i;
       label = 126;
      } else {
       $628 = HEAP32[$626>>2]|0;
       $629 = ((($628)) + 8|0);
       $630 = HEAP32[$629>>2]|0;
       $631 = ($630|0)==(0|0);
       if ($631) {
        $advance$3$ph$i$i = $advance$1$i$i;$top$2$ph$i$i = $top$0$i$i;
        label = 126;
        break;
       }
       HEAP32[$metrics$i4$i>>2] = 0;
       $632 = ((($metrics$i4$i)) + 4|0);
       HEAP32[$632>>2] = $top$0$i$i;
       $633 = ((($metrics$i4$i)) + 8|0);
       HEAP32[$633>>2] = $advance$1$i$i;
       $634 = ((($626)) + 4|0);
       $635 = HEAP32[$634>>2]|0;
       $636 = (FUNCTION_TABLE_iiiii[$630 & 8191]($635,$glyph_index,1,$metrics$i4$i)|0);
       $637 = ($636|0)==(0);
       if ($637) {
        $638 = HEAP32[$632>>2]|0;
        $639 = HEAP32[$633>>2]|0;
        $advance$3$ph$i$i = $639;$top$2$ph$i$i = $638;
        label = 126;
        break;
       } else {
        $$1$i$i = $636;
        break;
       }
      }
     } while(0);
     if ((label|0) == 126) {
      $640 = ((($519)) + 60|0);
      HEAP32[$640>>2] = $advance$3$ph$i$i;
      $641 = HEAP32[$476>>2]|0;
      $642 = $641 & 1;
      $643 = ($642|0)==(0);
      if ($643) {
       $644 = (_FT_MulFix($top$2$ph$i$i,$y_scale$0$i$i)|0);
       $645 = (_FT_MulFix($advance$3$ph$i$i,$y_scale$0$i$i)|0);
       $advance$4$i$i = $645;$top$3$i$i = $644;
      } else {
       $advance$4$i$i = $advance$3$ph$i$i;$top$3$i$i = $top$2$ph$i$i;
      }
      $646 = HEAP32[$535>>2]|0;
      $647 = HEAP32[$544>>2]|0;
      $648 = (($647|0) / 2)&-1;
      $649 = (($646) - ($648))|0;
      $650 = ((($519)) + 44|0);
      HEAP32[$650>>2] = $649;
      $651 = ((($519)) + 48|0);
      HEAP32[$651>>2] = $top$3$i$i;
      $652 = ((($519)) + 52|0);
      HEAP32[$652>>2] = $advance$4$i$i;
      $$1$i$i = 0;
     }
     $error$0$i = $$1$i$i;
    } else {
     $error$0$i = $484;
    }
    if ($136) {
     $$0$i = $error$0$i;
    } else {
     $653 = ((($ttsize)) + 14|0);
     $654 = HEAP16[$653>>1]|0;
     $655 = ($654&65535)<(24);
     if ($655) {
      $656 = HEAP32[$483>>2]|0;
      $657 = $656 | 256;
      HEAP32[$483>>2] = $657;
      $$0$i = $error$0$i;
     } else {
      $$0$i = $error$0$i;
     }
    }
   } else {
    $$0$i = 6;
   }
  }
 } while(0);
 $$0 = $$0$i;
 STACKTOP = sp;return ($$0|0);
}
function _load_truetype_glyph($loader,$glyph_index,$recurse_count,$header_only) {
 $loader = $loader|0;
 $glyph_index = $glyph_index|0;
 $recurse_count = $recurse_count|0;
 $header_only = $header_only|0;
 var $$$i = 0, $$0$i$i = 0, $$1$i26$ph = 0, $$idx = 0, $$idx$i = 0, $$idx$val = 0, $$idx$val$i = 0, $$idx11$val = 0, $$idx12 = 0, $$idx12$val = 0, $$idx13 = 0, $$idx13$val = 0, $$idx14$val = 0, $$idx15 = 0, $$idx15$val = 0, $$idx2$i = 0, $$idx2$val$i = 0, $$lcssa48 = 0, $$lcssa5$i = 0, $$lcssa93 = 0;
 var $$lcssa95 = 0, $$not$i = 0, $$phi$trans$insert$i = 0, $$pr = 0, $$pre = 0, $$pre$i = 0, $$pre$i20 = 0, $$pre$i23 = 0, $$pre$phi$iZ2D = 0, $$pre$phiZ2D = 0, $$pre11$i = 0, $$pre13$i = 0, $$pre6$i = 0, $$pre7$i = 0, $$pre7$pre$i = 0, $$pre74 = 0, $$pre75 = 0, $$pre9$i = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0;
 var $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0;
 var $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0;
 var $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0;
 var $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0;
 var $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0;
 var $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0;
 var $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0;
 var $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0;
 var $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0;
 var $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0;
 var $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0;
 var $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0;
 var $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0;
 var $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0;
 var $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0;
 var $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0;
 var $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0;
 var $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0;
 var $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0;
 var $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0;
 var $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0;
 var $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0;
 var $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0;
 var $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0;
 var $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0;
 var $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0;
 var $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0;
 var $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0;
 var $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0;
 var $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0;
 var $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0;
 var $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0;
 var $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0;
 var $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0;
 var $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0;
 var $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0;
 var $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0;
 var $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0;
 var $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0;
 var $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $contours = 0, $error = 0, $error$i = 0;
 var $error$i$i = 0, $exitcond = 0, $exitcond$i = 0, $exitcond$i$i = 0, $exitcond7$i$i = 0, $exitcond73 = 0, $glyph_data = 0, $glyph_data_loaded$0 = 0, $i$0$lcssa = 0, $i$059 = 0, $i$07$i = 0, $i$1$lcssa = 0, $i$152 = 0, $inc_stream = 0, $k$04$i$i = 0, $k$13$i$i = 0, $metrics$i = 0, $metrics$i16 = 0, $n$049 = 0, $num_points$0$lcssa = 0;
 var $offset$0 = 0, $opened_frame$0 = 0, $or$cond = 0, $or$cond$i = 0, $or$cond$i$i = 0, $or$cond10 = 0, $or$cond4 = 0, $or$cond6$i = 0, $or$cond6$i$i = 0, $or$cond8$i = 0, $outline = 0, $outline1 = 0, $points = 0, $sext = 0, $sext$i = 0, $sext$i17 = 0, $sext6 = 0, $sext7 = 0, $sext76 = 0, $subglyph$060 = 0;
 var $subglyph$153 = 0, $subglyph6$0$lcssa = 0, $tags = 0, $tmp$0$i = 0, $vec$05$i = 0, $x$1$i = 0, $x$3$i = 0, $x_scale$0 = 0, $y$1$i = 0, $y$3$i = 0, $y_scale$0 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i$i = sp + 180|0;
 $error$i = sp + 176|0;
 $metrics$i16 = sp + 160|0;
 $metrics$i = sp + 144|0;
 $error = sp + 136|0;
 $inc_stream = sp + 96|0;
 $glyph_data = sp + 88|0;
 $points = sp + 56|0;
 $tags = sp + 48|0;
 $contours = sp;
 $outline = sp + 28|0;
 $outline1 = sp + 8|0;
 HEAP32[$error>>2] = 0;
 $0 = HEAP32[$loader>>2]|0;
 $1 = ((($loader)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($recurse_count>>>0)>(1);
 if ($3) {
  $4 = ((($0)) + 290|0);
  $5 = HEAP16[$4>>1]|0;
  $6 = $5&65535;
  $7 = ($6>>>0)<($recurse_count>>>0);
  if ($7) {
   HEAP32[$error>>2] = 21;
   $918 = HEAP32[$error>>2]|0;
   STACKTOP = sp;return ($918|0);
  }
 }
 $8 = ((($loader)) + 20|0);
 HEAP32[$8>>2] = $glyph_index;
 $9 = ((($loader)) + 16|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 & 1;
 $12 = ($11|0)==(0);
 if ($12) {
  $13 = ((($loader)) + 4|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ((($14)) + 48|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ((($14)) + 52|0);
  $18 = HEAP32[$17>>2]|0;
  $x_scale$0 = $16;$y_scale$0 = $18;
 } else {
  $x_scale$0 = 65536;$y_scale$0 = 65536;
 }
 $19 = ((($0)) + 128|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ((($20)) + 48|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ($22|0)==(0|0);
 do {
  if ($23) {
   $35 = ((($loader)) + 28|0);
   $36 = (_tt_face_get_location($0,$glyph_index,$35)|0);
   $$pre$phiZ2D = $35;$glyph_data_loaded$0 = 0;$offset$0 = $36;
  } else {
   $24 = HEAP32[$22>>2]|0;
   $25 = HEAP32[$24>>2]|0;
   $26 = ((($22)) + 4|0);
   $27 = HEAP32[$26>>2]|0;
   $28 = (FUNCTION_TABLE_iiii[$25 & 8191]($27,$glyph_index,$glyph_data)|0);
   HEAP32[$error>>2] = $28;
   $29 = ($28|0)==(0);
   if ($29) {
    $30 = ((($glyph_data)) + 4|0);
    $31 = HEAP32[$30>>2]|0;
    $32 = ((($loader)) + 28|0);
    HEAP32[$32>>2] = $31;
    dest=$inc_stream; stop=dest+40|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
    $33 = HEAP32[$glyph_data>>2]|0;
    _FT_Stream_OpenMemory($inc_stream,$33,$31);
    $34 = ((($loader)) + 24|0);
    HEAP32[$34>>2] = $inc_stream;
    $$pre$phiZ2D = $32;$glyph_data_loaded$0 = 1;$offset$0 = 0;
    break;
   } else {
    $918 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($918|0);
   }
  }
 } while(0);
 $37 = HEAP32[$$pre$phiZ2D>>2]|0;
 $38 = ($37|0)>(0);
 do {
  if ($38) {
   $39 = ((($loader)) + 84|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = ($40|0)==(0);
   if ($41) {
    $42 = HEAP32[$19>>2]|0;
    $43 = ((($42)) + 48|0);
    $44 = HEAP32[$43>>2]|0;
    $45 = ($44|0)==(0|0);
    if ($45) {
     HEAP32[$error>>2] = 8;
     break;
    }
   }
   $46 = ((($0)) + 512|0);
   $47 = HEAP32[$46>>2]|0;
   $48 = (($40) + ($offset$0))|0;
   $49 = (FUNCTION_TABLE_iiiii[$47 & 8191]($loader,$glyph_index,$48,$37)|0);
   HEAP32[$error>>2] = $49;
   $50 = ($49|0)==(0);
   if ($50) {
    $51 = ((($0)) + 520|0);
    $52 = HEAP32[$51>>2]|0;
    $53 = (FUNCTION_TABLE_ii[$52 & 8191]($loader)|0);
    HEAP32[$error>>2] = $53;
    $54 = ($53|0)==(0);
    if ($54) {
     $55 = (_tt_get_metrics($loader,$glyph_index)|0);
     HEAP32[$error>>2] = $55;
     $56 = ($55|0)!=(0);
     $57 = ($header_only<<24>>24)!=(0);
     $or$cond = $57 | $56;
     if ($or$cond) {
      label = 128;
     } else {
      $$pr = HEAP32[$$pre$phiZ2D>>2]|0;
      $59 = $$pr;$opened_frame$0 = 1;
      label = 18;
     }
    } else {
     label = 128;
    }
   }
  } else {
   $59 = $37;$opened_frame$0 = 0;
   label = 18;
  }
 } while(0);
 L25: do {
  if ((label|0) == 18) {
   $58 = ($59|0)==(0);
   do {
    if ($58) {
     label = 20;
    } else {
     $60 = ((($loader)) + 32|0);
     $61 = HEAP16[$60>>1]|0;
     $62 = ($61<<16>>16)==(0);
     if ($62) {
      label = 20;
     } else {
      $167 = ((($loader)) + 36|0);
      $168 = HEAP32[$167>>2]|0;
      $169 = ((($loader)) + 52|0);
      $170 = HEAP32[$169>>2]|0;
      $171 = (($168) - ($170))|0;
      $172 = ((($loader)) + 68|0);
      HEAP32[$172>>2] = $171;
      $173 = ((($loader)) + 72|0);
      HEAP32[$173>>2] = 0;
      $174 = ((($loader)) + 56|0);
      $175 = HEAP32[$174>>2]|0;
      $176 = (($175) + ($171))|0;
      $177 = ((($loader)) + 76|0);
      HEAP32[$177>>2] = $176;
      $178 = ((($loader)) + 80|0);
      HEAP32[$178>>2] = 0;
      $179 = ((($loader)) + 184|0);
      HEAP32[$179>>2] = 0;
      $180 = ((($loader)) + 48|0);
      $181 = HEAP32[$180>>2]|0;
      $182 = ((($loader)) + 176|0);
      $183 = HEAP32[$182>>2]|0;
      $184 = (($183) + ($181))|0;
      $185 = ((($loader)) + 188|0);
      HEAP32[$185>>2] = $184;
      $186 = ((($loader)) + 192|0);
      HEAP32[$186>>2] = 0;
      $187 = ((($loader)) + 180|0);
      $188 = HEAP32[$187>>2]|0;
      $189 = (($184) - ($188))|0;
      $190 = ((($loader)) + 196|0);
      HEAP32[$190>>2] = $189;
      $191 = HEAP32[$loader>>2]|0;
      $192 = ((($191)) + 128|0);
      $193 = HEAP32[$192>>2]|0;
      $194 = ((($193)) + 48|0);
      $195 = HEAP32[$194>>2]|0;
      $196 = ($195|0)==(0|0);
      if ($196) {
       $217 = $61;
      } else {
       $197 = HEAP32[$195>>2]|0;
       $198 = ((($197)) + 8|0);
       $199 = HEAP32[$198>>2]|0;
       $200 = ($199|0)==(0|0);
       if ($200) {
        $217 = $61;
       } else {
        HEAP32[$metrics$i16>>2] = $170;
        $201 = ((($metrics$i16)) + 4|0);
        HEAP32[$201>>2] = 0;
        $202 = ((($metrics$i16)) + 8|0);
        HEAP32[$202>>2] = $175;
        $203 = ((($metrics$i16)) + 12|0);
        HEAP32[$203>>2] = 0;
        $204 = ((($195)) + 4|0);
        $205 = HEAP32[$204>>2]|0;
        $206 = (FUNCTION_TABLE_iiiii[$199 & 8191]($205,$glyph_index,0,$metrics$i16)|0);
        $207 = ($206|0)==(0);
        if ($207) {
         $208 = HEAP32[$metrics$i16>>2]|0;
         $209 = HEAP32[$202>>2]|0;
         $sext$i17 = $208 << 16;
         $210 = $sext$i17 >> 16;
         HEAP32[$169>>2] = $210;
         $211 = $209 & 65535;
         HEAP32[$174>>2] = $211;
         HEAP32[$182>>2] = 0;
         HEAP32[$187>>2] = 0;
         $212 = ((($loader)) + 64|0);
         $213 = HEAP8[$212>>0]|0;
         $214 = ($213<<24>>24)==(0);
         if ($214) {
          HEAP8[$212>>0] = 1;
          $215 = ((($loader)) + 60|0);
          HEAP32[$215>>2] = $211;
         }
        }
        $$pre = HEAP16[$60>>1]|0;
        $217 = $$pre;
       }
      }
      $216 = ($217<<16>>16)>(0);
      if ($216) {
       $218 = ((($0)) + 524|0);
       $219 = HEAP32[$218>>2]|0;
       $220 = (FUNCTION_TABLE_ii[$219 & 8191]($loader)|0);
       HEAP32[$error>>2] = $220;
       $221 = ($220|0)==(0);
       if (!($221)) {
        break;
       }
       $222 = ((($0)) + 516|0);
       $223 = HEAP32[$222>>2]|0;
       FUNCTION_TABLE_vi[$223 & 8191]($loader);
       $224 = HEAP32[$1>>2]|0;
       $225 = ((($224)) + 56|0);
       $226 = ((($224)) + 58|0);
       $227 = HEAP16[$226>>1]|0;
       $228 = $227 << 16 >> 16;
       $229 = ((($224)) + 60|0);
       $230 = HEAP32[$229>>2]|0;
       $231 = (($230) + ($228<<3)|0);
       $232 = $172;
       $233 = $232;
       $234 = HEAP32[$233>>2]|0;
       $235 = (($232) + 4)|0;
       $236 = $235;
       $237 = HEAP32[$236>>2]|0;
       $238 = $231;
       $239 = $238;
       HEAP32[$239>>2] = $234;
       $240 = (($238) + 4)|0;
       $241 = $240;
       HEAP32[$241>>2] = $237;
       $242 = (($228) + 1)|0;
       $243 = HEAP32[$229>>2]|0;
       $244 = (($243) + ($242<<3)|0);
       $245 = $177;
       $246 = $245;
       $247 = HEAP32[$246>>2]|0;
       $248 = (($245) + 4)|0;
       $249 = $248;
       $250 = HEAP32[$249>>2]|0;
       $251 = $244;
       $252 = $251;
       HEAP32[$252>>2] = $247;
       $253 = (($251) + 4)|0;
       $254 = $253;
       HEAP32[$254>>2] = $250;
       $255 = (($228) + 2)|0;
       $256 = HEAP32[$229>>2]|0;
       $257 = (($256) + ($255<<3)|0);
       $258 = $179;
       $259 = $258;
       $260 = HEAP32[$259>>2]|0;
       $261 = (($258) + 4)|0;
       $262 = $261;
       $263 = HEAP32[$262>>2]|0;
       $264 = $257;
       $265 = $264;
       HEAP32[$265>>2] = $260;
       $266 = (($264) + 4)|0;
       $267 = $266;
       HEAP32[$267>>2] = $263;
       $268 = (($228) + 3)|0;
       $269 = HEAP32[$229>>2]|0;
       $270 = (($269) + ($268<<3)|0);
       $271 = $186;
       $272 = $271;
       $273 = HEAP32[$272>>2]|0;
       $274 = (($271) + 4)|0;
       $275 = $274;
       $276 = HEAP32[$275>>2]|0;
       $277 = $270;
       $278 = $277;
       HEAP32[$278>>2] = $273;
       $279 = (($277) + 4)|0;
       $280 = $279;
       HEAP32[$280>>2] = $276;
       $281 = ((($224)) + 64|0);
       $282 = HEAP32[$281>>2]|0;
       $283 = (($282) + ($228)|0);
       HEAP8[$283>>0] = 0;
       $284 = HEAP32[$281>>2]|0;
       $285 = (($284) + ($242)|0);
       HEAP8[$285>>0] = 0;
       $286 = HEAP32[$281>>2]|0;
       $287 = (($286) + ($255)|0);
       HEAP8[$287>>0] = 0;
       $288 = HEAP32[$281>>2]|0;
       $289 = (($288) + ($268)|0);
       HEAP8[$289>>0] = 0;
       $290 = (($228) + 4)|0;
       $291 = HEAP32[$loader>>2]|0;
       $292 = ((($291)) + 672|0);
       $293 = HEAP8[$292>>0]|0;
       $294 = ($293<<24>>24)==(0);
       if (!($294)) {
        $295 = HEAP32[$8>>2]|0;
        $$idx$i = ((($291)) + 104|0);
        $$idx$val$i = HEAP32[$$idx$i>>2]|0;
        $$idx2$i = ((($291)) + 676|0);
        $$idx2$val$i = HEAP32[$$idx2$i>>2]|0;
        $296 = (_TT_Vary_Apply_Glyph_Deltas($$idx$val$i,$293,$$idx2$val$i,$295,$225,$290)|0);
        $297 = ($296|0)==(0);
        if (!($297)) {
         HEAP32[$error>>2] = $296;
         break L25;
        }
       }
       $298 = HEAP32[$9>>2]|0;
       $299 = $298 & 2;
       $300 = ($299|0)==(0);
       if ($300) {
        $301 = HEAP16[$226>>1]|0;
        $302 = ((($loader)) + 132|0);
        HEAP16[$302>>1] = $301;
        $303 = HEAP16[$225>>1]|0;
        $304 = ((($loader)) + 134|0);
        HEAP16[$304>>1] = $303;
        $305 = ((($224)) + 76|0);
        $306 = HEAP32[$305>>2]|0;
        $307 = ((($loader)) + 136|0);
        HEAP32[$307>>2] = $306;
        $308 = HEAP32[$229>>2]|0;
        $309 = ((($loader)) + 140|0);
        HEAP32[$309>>2] = $308;
        $310 = ((($224)) + 80|0);
        $311 = HEAP32[$310>>2]|0;
        $312 = ((($loader)) + 144|0);
        HEAP32[$312>>2] = $311;
        $313 = HEAP32[$281>>2]|0;
        $314 = ((($loader)) + 148|0);
        HEAP32[$314>>2] = $313;
        $315 = ((($224)) + 68|0);
        $316 = HEAP32[$315>>2]|0;
        $317 = ((($loader)) + 152|0);
        HEAP32[$317>>2] = $316;
        $318 = ((($loader)) + 156|0);
        HEAP16[$318>>1] = 0;
        $319 = $311;
        $320 = $308;
        $321 = $301&65535;
        $322 = $321 << 3;
        $323 = (($322) + 32)|0;
        _memcpy(($319|0),($320|0),($323|0))|0;
        $$pre$i = HEAP32[$9>>2]|0;
        $327 = $$pre$i;
       } else {
        $327 = $298;
       }
       $324 = HEAP32[$229>>2]|0;
       $325 = (($324) + ($290<<3)|0);
       $326 = $327 & 1;
       $328 = ($326|0)==(0);
       if ($328) {
        $329 = ((($loader)) + 4|0);
        $330 = HEAP32[$329>>2]|0;
        $331 = ((($330)) + 48|0);
        $332 = HEAP32[$331>>2]|0;
        $333 = ((($330)) + 52|0);
        $334 = HEAP32[$333>>2]|0;
        $335 = ($227<<16>>16)>(-4);
        if ($335) {
         $vec$05$i = $324;
         while(1) {
          $336 = HEAP32[$vec$05$i>>2]|0;
          $337 = (_FT_MulFix($336,$332)|0);
          HEAP32[$vec$05$i>>2] = $337;
          $338 = ((($vec$05$i)) + 4|0);
          $339 = HEAP32[$338>>2]|0;
          $340 = (_FT_MulFix($339,$334)|0);
          HEAP32[$338>>2] = $340;
          $341 = ((($vec$05$i)) + 8|0);
          $342 = ($341>>>0)<($325>>>0);
          if ($342) {
           $vec$05$i = $341;
          } else {
           break;
          }
         }
         $$pre6$i = HEAP32[$229>>2]|0;
         $$pre7$pre$i = HEAP32[$9>>2]|0;
         $$pre7$i = $$pre7$pre$i;$344 = $$pre6$i;
        } else {
         $$pre7$i = $327;$344 = $324;
        }
        $343 = (($344) + ($228<<3)|0);
        $345 = $343;
        $346 = $345;
        $347 = HEAP32[$346>>2]|0;
        $348 = (($345) + 4)|0;
        $349 = $348;
        $350 = HEAP32[$349>>2]|0;
        $351 = $172;
        $352 = $351;
        HEAP32[$352>>2] = $347;
        $353 = (($351) + 4)|0;
        $354 = $353;
        HEAP32[$354>>2] = $350;
        $355 = HEAP32[$229>>2]|0;
        $356 = (($355) + ($242<<3)|0);
        $357 = $356;
        $358 = $357;
        $359 = HEAP32[$358>>2]|0;
        $360 = (($357) + 4)|0;
        $361 = $360;
        $362 = HEAP32[$361>>2]|0;
        $363 = $177;
        $364 = $363;
        HEAP32[$364>>2] = $359;
        $365 = (($363) + 4)|0;
        $366 = $365;
        HEAP32[$366>>2] = $362;
        $367 = HEAP32[$229>>2]|0;
        $368 = (($367) + ($255<<3)|0);
        $369 = $368;
        $370 = $369;
        $371 = HEAP32[$370>>2]|0;
        $372 = (($369) + 4)|0;
        $373 = $372;
        $374 = HEAP32[$373>>2]|0;
        $375 = $179;
        $376 = $375;
        HEAP32[$376>>2] = $371;
        $377 = (($375) + 4)|0;
        $378 = $377;
        HEAP32[$378>>2] = $374;
        $379 = HEAP32[$229>>2]|0;
        $380 = (($379) + ($268<<3)|0);
        $381 = $380;
        $382 = $381;
        $383 = HEAP32[$382>>2]|0;
        $384 = (($381) + 4)|0;
        $385 = $384;
        $386 = HEAP32[$385>>2]|0;
        $387 = $186;
        $388 = $387;
        HEAP32[$388>>2] = $383;
        $389 = (($387) + 4)|0;
        $390 = $389;
        HEAP32[$390>>2] = $386;
        $392 = $$pre7$i;
       } else {
        $392 = $327;
       }
       $391 = $392 & 2;
       $393 = ($391|0)==(0);
       if ($393) {
        $394 = ((($loader)) + 132|0);
        $395 = HEAP16[$394>>1]|0;
        $396 = $395&65535;
        $397 = (($396) + 4)|0;
        $398 = $397&65535;
        HEAP16[$394>>1] = $398;
        $399 = (_TT_Hint_Glyph($loader,0)|0);
        HEAP32[$error>>2] = $399;
        $400 = ($399|0)==(0);
        if (!($400)) {
         break L25;
        }
       } else {
        HEAP32[$error>>2] = 0;
       }
       _FT_GlyphLoader_Add($2);
       break L25;
      }
      $401 = ($217<<16>>16)==(-1);
      if (!($401)) {
       HEAP32[$error>>2] = 20;
       break;
      }
      $402 = ((($2)) + 22|0);
      $403 = HEAP16[$402>>1]|0;
      $404 = $403 << 16 >> 16;
      $405 = ((($2)) + 20|0);
      $406 = HEAP16[$405>>1]|0;
      $407 = $406 << 16 >> 16;
      $408 = ((($0)) + 528|0);
      $409 = HEAP32[$408>>2]|0;
      $410 = (FUNCTION_TABLE_ii[$409 & 8191]($loader)|0);
      HEAP32[$error>>2] = $410;
      $411 = ($410|0)==(0);
      if ($411) {
       $412 = ((($loader)) + 168|0);
       $413 = HEAP32[$412>>2]|0;
       $414 = ((($0)) + 516|0);
       $415 = HEAP32[$414>>2]|0;
       FUNCTION_TABLE_vi[$415 & 8191]($loader);
       $416 = ((($0)) + 672|0);
       $417 = HEAP8[$416>>0]|0;
       $418 = ($417<<24>>24)==(0);
       if (!($418)) {
        $419 = ((($0)) + 100|0);
        $420 = HEAP32[$419>>2]|0;
        $421 = ((($2)) + 84|0);
        $422 = HEAP32[$421>>2]|0;
        $423 = (($422) + 4)|0;
        $424 = $423&65535;
        $425 = ((($outline1)) + 2|0);
        HEAP16[$425>>1] = $424;
        HEAP16[$outline1>>1] = $424;
        $sext = $423 << 16;
        $426 = $sext >> 16;
        $427 = (_ft_mem_realloc($420,8,0,$426,0,$error)|0);
        $428 = HEAP32[$error>>2]|0;
        $429 = ($428|0)==(0);
        if ($429) {
         $sext76 = $423 << 16;
         $430 = $sext76 >> 16;
         $431 = (_ft_mem_realloc($420,1,0,$430,0,$error)|0);
         $432 = HEAP32[$error>>2]|0;
         $433 = ($432|0)==(0);
         if ($433) {
          $434 = (_ft_mem_realloc($420,2,0,$430,0,$error)|0);
          $435 = HEAP32[$error>>2]|0;
          $436 = ($435|0)==(0);
          if ($436) {
           $437 = ((($2)) + 88|0);
           $438 = ((($2)) + 48|0);
           $439 = ($422|0)==(0);
           if ($439) {
            $i$0$lcssa = 0;
           } else {
            $440 = HEAP32[$437>>2]|0;
            $441 = HEAP32[$438>>2]|0;
            $442 = (($440) + ($441<<5)|0);
            $i$059 = 0;$subglyph$060 = $442;
            while(1) {
             $443 = ((($subglyph$060)) + 8|0);
             $444 = HEAP32[$443>>2]|0;
             $445 = (($427) + ($i$059<<3)|0);
             HEAP32[$445>>2] = $444;
             $446 = ((($subglyph$060)) + 12|0);
             $447 = HEAP32[$446>>2]|0;
             $448 = (((($427) + ($i$059<<3)|0)) + 4|0);
             HEAP32[$448>>2] = $447;
             $449 = (($431) + ($i$059)|0);
             HEAP8[$449>>0] = 1;
             $450 = $i$059&65535;
             $451 = (($434) + ($i$059<<1)|0);
             HEAP16[$451>>1] = $450;
             $452 = (($i$059) + 1)|0;
             $453 = ((($subglyph$060)) + 32|0);
             $exitcond73 = ($452|0)==($422|0);
             if ($exitcond73) {
              $i$0$lcssa = $422;
              break;
             } else {
              $i$059 = $452;$subglyph$060 = $453;
             }
            }
           }
           $454 = HEAP32[$172>>2]|0;
           $455 = (($427) + ($i$0$lcssa<<3)|0);
           HEAP32[$455>>2] = $454;
           $456 = HEAP32[$173>>2]|0;
           $457 = (((($427) + ($i$0$lcssa<<3)|0)) + 4|0);
           HEAP32[$457>>2] = $456;
           $458 = (($431) + ($i$0$lcssa)|0);
           HEAP8[$458>>0] = 1;
           $459 = $i$0$lcssa&65535;
           $460 = (($434) + ($i$0$lcssa<<1)|0);
           HEAP16[$460>>1] = $459;
           $461 = (($i$0$lcssa) + 1)|0;
           $462 = HEAP32[$177>>2]|0;
           $463 = (($427) + ($461<<3)|0);
           HEAP32[$463>>2] = $462;
           $464 = HEAP32[$178>>2]|0;
           $465 = (((($427) + ($461<<3)|0)) + 4|0);
           HEAP32[$465>>2] = $464;
           $466 = (($431) + ($461)|0);
           HEAP8[$466>>0] = 1;
           $467 = $461&65535;
           $468 = (($434) + ($461<<1)|0);
           HEAP16[$468>>1] = $467;
           $469 = (($i$0$lcssa) + 2)|0;
           $470 = HEAP32[$179>>2]|0;
           $471 = (($427) + ($469<<3)|0);
           HEAP32[$471>>2] = $470;
           $472 = HEAP32[$185>>2]|0;
           $473 = (((($427) + ($469<<3)|0)) + 4|0);
           HEAP32[$473>>2] = $472;
           $474 = (($431) + ($469)|0);
           HEAP8[$474>>0] = 1;
           $475 = $469&65535;
           $476 = (($434) + ($469<<1)|0);
           HEAP16[$476>>1] = $475;
           $477 = (($i$0$lcssa) + 3)|0;
           $478 = HEAP32[$186>>2]|0;
           $479 = (($427) + ($477<<3)|0);
           HEAP32[$479>>2] = $478;
           $480 = HEAP32[$190>>2]|0;
           $481 = (((($427) + ($477<<3)|0)) + 4|0);
           HEAP32[$481>>2] = $480;
           $482 = (($431) + ($477)|0);
           HEAP8[$482>>0] = 1;
           $483 = $477&65535;
           $484 = (($434) + ($477<<1)|0);
           HEAP16[$484>>1] = $483;
           $485 = ((($outline1)) + 4|0);
           HEAP32[$485>>2] = $427;
           $486 = ((($outline1)) + 8|0);
           HEAP32[$486>>2] = $431;
           $487 = ((($outline1)) + 12|0);
           HEAP32[$487>>2] = $434;
           $$idx = ((($0)) + 104|0);
           $$idx$val = HEAP32[$$idx>>2]|0;
           $$idx11$val = HEAP8[$416>>0]|0;
           $$idx12 = ((($0)) + 676|0);
           $$idx12$val = HEAP32[$$idx12>>2]|0;
           $488 = (_TT_Vary_Apply_Glyph_Deltas($$idx$val,$$idx11$val,$$idx12$val,$glyph_index,$outline1,$430)|0);
           HEAP32[$error>>2] = $488;
           $489 = ($488|0)==(0);
           if ($489) {
            $490 = HEAP32[$427>>2]|0;
            if ($439) {
             $$lcssa48 = $490;$i$1$lcssa = 0;
            } else {
             $491 = HEAP32[$437>>2]|0;
             $492 = HEAP32[$438>>2]|0;
             $493 = (($491) + ($492<<5)|0);
             $494 = $490;$i$152 = 0;$subglyph$153 = $493;
             while(1) {
              $sext6 = $494 << 16;
              $495 = $sext6 >> 16;
              $496 = ((($subglyph$153)) + 8|0);
              HEAP32[$496>>2] = $495;
              $497 = (((($427) + ($i$152<<3)|0)) + 4|0);
              $498 = HEAP32[$497>>2]|0;
              $sext7 = $498 << 16;
              $499 = $sext7 >> 16;
              $500 = ((($subglyph$153)) + 12|0);
              HEAP32[$500>>2] = $499;
              $501 = (($i$152) + 1)|0;
              $502 = ((($subglyph$153)) + 32|0);
              $503 = (($427) + ($501<<3)|0);
              $504 = HEAP32[$503>>2]|0;
              $exitcond = ($501|0)==($422|0);
              if ($exitcond) {
               $$lcssa48 = $504;$i$1$lcssa = $422;
               break;
              } else {
               $494 = $504;$i$152 = $501;$subglyph$153 = $502;
              }
             }
            }
            HEAP32[$172>>2] = $$lcssa48;
            $505 = (((($427) + ($i$1$lcssa<<3)|0)) + 4|0);
            $506 = HEAP32[$505>>2]|0;
            HEAP32[$173>>2] = $506;
            $507 = (($i$1$lcssa) + 1)|0;
            $508 = (($427) + ($507<<3)|0);
            $509 = HEAP32[$508>>2]|0;
            HEAP32[$177>>2] = $509;
            $510 = (((($427) + ($507<<3)|0)) + 4|0);
            $511 = HEAP32[$510>>2]|0;
            HEAP32[$178>>2] = $511;
            $512 = (($i$1$lcssa) + 2)|0;
            $513 = (($427) + ($512<<3)|0);
            $514 = HEAP32[$513>>2]|0;
            HEAP32[$179>>2] = $514;
            $515 = (((($427) + ($512<<3)|0)) + 4|0);
            $516 = HEAP32[$515>>2]|0;
            HEAP32[$185>>2] = $516;
            $517 = (($i$1$lcssa) + 3)|0;
            $518 = (($427) + ($517<<3)|0);
            $519 = HEAP32[$518>>2]|0;
            HEAP32[$186>>2] = $519;
            $520 = (((($427) + ($517<<3)|0)) + 4|0);
            $521 = HEAP32[$520>>2]|0;
            HEAP32[$190>>2] = $521;
            $522 = $431;$523 = $434;
           } else {
            $522 = $431;$523 = $434;
           }
          } else {
           $522 = 0;$523 = 0;
          }
         } else {
          $522 = 0;$523 = 0;
         }
        } else {
         $522 = 0;$523 = 0;
        }
        _ft_mem_free($420,$427);
        _ft_mem_free($420,$522);
        _ft_mem_free($420,$523);
        $524 = HEAP32[$error>>2]|0;
        $525 = ($524|0)==(0);
        if (!($525)) {
         break L25;
        }
       }
       $526 = HEAP32[$9>>2]|0;
       $527 = $526 & 1;
       $528 = ($527|0)==(0);
       if ($528) {
        $529 = HEAP32[$172>>2]|0;
        $530 = (_FT_MulFix($529,$x_scale$0)|0);
        HEAP32[$172>>2] = $530;
        $531 = HEAP32[$177>>2]|0;
        $532 = (_FT_MulFix($531,$x_scale$0)|0);
        HEAP32[$177>>2] = $532;
        $533 = HEAP32[$179>>2]|0;
        $534 = (_FT_MulFix($533,$x_scale$0)|0);
        HEAP32[$179>>2] = $534;
        $535 = HEAP32[$185>>2]|0;
        $536 = (_FT_MulFix($535,$y_scale$0)|0);
        HEAP32[$185>>2] = $536;
        $537 = HEAP32[$186>>2]|0;
        $538 = (_FT_MulFix($537,$x_scale$0)|0);
        HEAP32[$186>>2] = $538;
        $539 = HEAP32[$190>>2]|0;
        $540 = (_FT_MulFix($539,$y_scale$0)|0);
        HEAP32[$190>>2] = $540;
        $$pre74 = HEAP32[$9>>2]|0;
        $542 = $$pre74;
       } else {
        $542 = $526;
       }
       $541 = $542 & 1024;
       $543 = ($541|0)==(0);
       if (!($543)) {
        _FT_GlyphLoader_Add($2);
        $544 = ((($loader)) + 8|0);
        $545 = HEAP32[$544>>2]|0;
        $546 = ((($545)) + 72|0);
        HEAP32[$546>>2] = 1668246896;
        break L25;
       }
       $547 = ((($2)) + 84|0);
       $548 = HEAP32[$547>>2]|0;
       $549 = ((($2)) + 48|0);
       $550 = HEAP32[$549>>2]|0;
       $551 = ((($loader)) + 24|0);
       $552 = HEAP32[$551>>2]|0;
       $553 = HEAP32[$$pre$phiZ2D>>2]|0;
       _FT_GlyphLoader_Add($2);
       $554 = ($548|0)==(0);
       do {
        if ($554) {
         $num_points$0$lcssa = $404;$subglyph6$0$lcssa = 0;
        } else {
         $555 = ((($2)) + 52|0);
         $556 = (($recurse_count) + 1)|0;
         $557 = ((($loader)) + 4|0);
         $n$049 = 0;
         L92: while(1) {
          $558 = HEAP32[$555>>2]|0;
          $559 = (($558) + ($550<<5)|0);
          $560 = (($559) + ($n$049<<5)|0);
          $561 = $172;
          $562 = $561;
          $563 = HEAP32[$562>>2]|0;
          $564 = (($561) + 4)|0;
          $565 = $564;
          $566 = HEAP32[$565>>2]|0;
          $567 = $177;
          $568 = $567;
          $569 = HEAP32[$568>>2]|0;
          $570 = (($567) + 4)|0;
          $571 = $570;
          $572 = HEAP32[$571>>2]|0;
          $573 = $179;
          $574 = $573;
          $575 = HEAP32[$574>>2]|0;
          $576 = (($573) + 4)|0;
          $577 = $576;
          $578 = HEAP32[$577>>2]|0;
          $579 = $186;
          $580 = $579;
          $581 = HEAP32[$580>>2]|0;
          $582 = (($579) + 4)|0;
          $583 = $582;
          $584 = HEAP32[$583>>2]|0;
          $585 = HEAP16[$402>>1]|0;
          $586 = $585 << 16 >> 16;
          $587 = HEAP32[$560>>2]|0;
          $588 = (_load_truetype_glyph($loader,$587,$556,0)|0);
          HEAP32[$error>>2] = $588;
          $589 = ($588|0)==(0);
          if (!($589)) {
           break L25;
          }
          $590 = HEAP32[$555>>2]|0;
          $591 = (($590) + ($550<<5)|0);
          $592 = (($591) + ($n$049<<5)|0);
          $593 = ((($592)) + 4|0);
          $594 = HEAP16[$593>>1]|0;
          $595 = $594 & 512;
          $596 = ($595<<16>>16)==(0);
          if ($596) {
           $597 = $172;
           $598 = $597;
           HEAP32[$598>>2] = $563;
           $599 = (($597) + 4)|0;
           $600 = $599;
           HEAP32[$600>>2] = $566;
           $601 = $177;
           $602 = $601;
           HEAP32[$602>>2] = $569;
           $603 = (($601) + 4)|0;
           $604 = $603;
           HEAP32[$604>>2] = $572;
           $605 = $179;
           $606 = $605;
           HEAP32[$606>>2] = $575;
           $607 = (($605) + 4)|0;
           $608 = $607;
           HEAP32[$608>>2] = $578;
           $609 = $186;
           $610 = $609;
           HEAP32[$610>>2] = $581;
           $611 = (($609) + 4)|0;
           $612 = $611;
           HEAP32[$612>>2] = $584;
          }
          $613 = HEAP16[$402>>1]|0;
          $614 = ($613<<16>>16)==($585<<16>>16);
          if (!($614)) {
           $615 = HEAP32[$1>>2]|0;
           $616 = ((($615)) + 24|0);
           $617 = HEAP32[$616>>2]|0;
           $618 = ((($615)) + 22|0);
           $619 = HEAP16[$618>>1]|0;
           $620 = $619 << 16 >> 16;
           $621 = HEAP16[$593>>1]|0;
           $622 = $621 & 200;
           $623 = ($622<<16>>16)==(0);
           $$not$i = $623 ^ 1;
           $624 = ($619&65535)>($585&65535);
           $or$cond8$i = $624 & $$not$i;
           if ($or$cond8$i) {
            $625 = ((($592)) + 16|0);
            $i$07$i = $586;
            while(1) {
             $626 = (($617) + ($i$07$i<<3)|0);
             _FT_Vector_Transform($626,$625);
             $627 = (($i$07$i) + 1)|0;
             $exitcond$i = ($627|0)==($620|0);
             if ($exitcond$i) {
              break;
             } else {
              $i$07$i = $627;
             }
            }
            $$pre$i20 = HEAP16[$593>>1]|0;
            $629 = $$pre$i20;
           } else {
            $629 = $621;
           }
           $628 = $629&65535;
           $630 = $628 & 2;
           $631 = ($630|0)==(0);
           $632 = ((($592)) + 8|0);
           $633 = HEAP32[$632>>2]|0;
           $634 = ((($592)) + 12|0);
           $635 = HEAP32[$634>>2]|0;
           do {
            if ($631) {
             $636 = (($633) + ($404))|0;
             $637 = (($635) + ($586))|0;
             $638 = ($636>>>0)<($586>>>0);
             $639 = ($637>>>0)<($620>>>0);
             $or$cond$i = $638 & $639;
             if (!($or$cond$i)) {
              label = 87;
              break L92;
             }
             $640 = HEAP32[$616>>2]|0;
             $641 = (($640) + ($636<<3)|0);
             $642 = HEAP32[$641>>2]|0;
             $643 = (($640) + ($637<<3)|0);
             $644 = HEAP32[$643>>2]|0;
             $645 = (($642) - ($644))|0;
             $646 = (((($640) + ($636<<3)|0)) + 4|0);
             $647 = HEAP32[$646>>2]|0;
             $648 = (((($640) + ($637<<3)|0)) + 4|0);
             $649 = HEAP32[$648>>2]|0;
             $650 = (($647) - ($649))|0;
             $x$3$i = $645;$y$3$i = $650;
             label = 95;
            } else {
             $651 = $635 | $633;
             $652 = ($651|0)==(0);
             if ($652) {
              break;
             }
             $653 = $628 & 2048;
             $654 = ($653|0)==(0);
             $or$cond6$i = $623 | $654;
             if ($or$cond6$i) {
              $x$1$i = $633;$y$1$i = $635;
             } else {
              $655 = ((($592)) + 16|0);
              $656 = HEAP32[$655>>2]|0;
              $657 = ((($592)) + 20|0);
              $658 = HEAP32[$657>>2]|0;
              $659 = (_FT_Hypot($656,$658)|0);
              $660 = ((($592)) + 28|0);
              $661 = HEAP32[$660>>2]|0;
              $662 = ((($592)) + 24|0);
              $663 = HEAP32[$662>>2]|0;
              $664 = (_FT_Hypot($661,$663)|0);
              $665 = (_FT_MulFix($633,$659)|0);
              $666 = (_FT_MulFix($635,$664)|0);
              $x$1$i = $665;$y$1$i = $666;
             }
             $667 = HEAP32[$9>>2]|0;
             $668 = $667 & 1;
             $669 = ($668|0)==(0);
             if (!($669)) {
              $x$3$i = $x$1$i;$y$3$i = $y$1$i;
              label = 95;
              break;
             }
             $670 = HEAP32[$557>>2]|0;
             $671 = ((($670)) + 48|0);
             $672 = HEAP32[$671>>2]|0;
             $673 = ((($670)) + 52|0);
             $674 = HEAP32[$673>>2]|0;
             $675 = (_FT_MulFix($x$1$i,$672)|0);
             $676 = (_FT_MulFix($y$1$i,$674)|0);
             $677 = HEAP16[$593>>1]|0;
             $678 = $677 & 4;
             $679 = ($678<<16>>16)==(0);
             if ($679) {
              $x$3$i = $675;$y$3$i = $676;
              label = 95;
              break;
             }
             $680 = (($675) + 32)|0;
             $681 = $680 & -64;
             $682 = (($676) + 32)|0;
             $683 = $682 & -64;
             $x$3$i = $681;$y$3$i = $683;
             label = 95;
            }
           } while(0);
           do {
            if ((label|0) == 95) {
             label = 0;
             $684 = $x$3$i | $y$3$i;
             $685 = ($684|0)==(0);
             if ($685) {
              break;
             }
             $686 = (($620) - ($586))|0;
             $687 = (($617) + ($586<<3)|0);
             $688 = ($x$3$i|0)!=(0);
             $689 = ($619<<16>>16)!=($585<<16>>16);
             $or$cond$i$i = $689 & $688;
             if ($or$cond$i$i) {
              $k$04$i$i = 0;
              while(1) {
               $690 = (($687) + ($k$04$i$i<<3)|0);
               $691 = HEAP32[$690>>2]|0;
               $692 = (($691) + ($x$3$i))|0;
               HEAP32[$690>>2] = $692;
               $693 = (($k$04$i$i) + 1)|0;
               $exitcond7$i$i = ($693|0)==($686|0);
               if ($exitcond7$i$i) {
                break;
               } else {
                $k$04$i$i = $693;
               }
              }
             }
             $694 = ($y$3$i|0)!=(0);
             $or$cond6$i$i = $689 & $694;
             if ($or$cond6$i$i) {
              $k$13$i$i = 0;
             } else {
              break;
             }
             while(1) {
              $695 = (((($687) + ($k$13$i$i<<3)|0)) + 4|0);
              $696 = HEAP32[$695>>2]|0;
              $697 = (($696) + ($y$3$i))|0;
              HEAP32[$695>>2] = $697;
              $698 = (($k$13$i$i) + 1)|0;
              $exitcond$i$i = ($698|0)==($686|0);
              if ($exitcond$i$i) {
               break;
              } else {
               $k$13$i$i = $698;
              }
             }
            }
           } while(0);
           HEAP32[$error>>2] = 0;
          }
          $699 = (($n$049) + 1)|0;
          $700 = ($699>>>0)<($548>>>0);
          if ($700) {
           $n$049 = $699;
          } else {
           $$lcssa93 = $592;$$lcssa95 = $613;
           label = 102;
           break;
          }
         }
         if ((label|0) == 87) {
          HEAP32[$error>>2] = 21;
          break L25;
         }
         else if ((label|0) == 102) {
          $701 = $$lcssa95 << 16 >> 16;
          $num_points$0$lcssa = $701;$subglyph6$0$lcssa = $$lcssa93;
          break;
         }
        }
       } while(0);
       HEAP32[$551>>2] = $552;
       HEAP32[$$pre$phiZ2D>>2] = $553;
       HEAP32[$412>>2] = $413;
       $702 = HEAP32[$9>>2]|0;
       $703 = $702 & 2;
       $704 = ($703|0)==(0);
       if (!($704)) {
        break L25;
       }
       $705 = ((($subglyph6$0$lcssa)) + 4|0);
       $706 = HEAP16[$705>>1]|0;
       $707 = $706 & 256;
       $708 = ($707<<16>>16)!=(0);
       $709 = ($num_points$0$lcssa>>>0)>($404>>>0);
       $or$cond10 = $709 & $708;
       if (!($or$cond10)) {
        break L25;
       }
       $710 = HEAP32[$1>>2]|0;
       $711 = ((($710)) + 22|0);
       $712 = HEAP16[$711>>1]|0;
       $713 = $712 << 16 >> 16;
       $714 = (($713) + 4)|0;
       $715 = ($714|0)==(0);
       do {
        if ($715) {
         label = 107;
        } else {
         $716 = ((($710)) + 58|0);
         $717 = HEAP16[$716>>1]|0;
         $718 = $717 << 16 >> 16;
         $719 = (($714) + ($713))|0;
         $720 = (($719) + ($718))|0;
         $721 = ((($710)) + 4|0);
         $722 = HEAP32[$721>>2]|0;
         $723 = ($720>>>0)>($722>>>0);
         if (!($723)) {
          label = 107;
          break;
         }
         $724 = (_FT_GlyphLoader_CheckPoints($710,$714,0)|0);
         HEAP32[$error$i>>2] = $724;
         $725 = ($724|0)==(0);
         if (!($725)) {
          $$1$i26$ph = $724;
          break;
         }
         $$pre$i23 = HEAP16[$711>>1]|0;
         $727 = $$pre$i23;
         label = 110;
        }
       } while(0);
       if ((label|0) == 107) {
        HEAP32[$error$i>>2] = 0;
        $727 = $712;
        label = 110;
       }
       L139: do {
        if ((label|0) == 110) {
         $726 = $727 << 16 >> 16;
         $728 = ((($710)) + 24|0);
         $729 = HEAP32[$728>>2]|0;
         $730 = (($729) + ($726<<3)|0);
         $731 = $172;
         $732 = $731;
         $733 = HEAP32[$732>>2]|0;
         $734 = (($731) + 4)|0;
         $735 = $734;
         $736 = HEAP32[$735>>2]|0;
         $737 = $730;
         $738 = $737;
         HEAP32[$738>>2] = $733;
         $739 = (($737) + 4)|0;
         $740 = $739;
         HEAP32[$740>>2] = $736;
         $741 = HEAP16[$711>>1]|0;
         $742 = $741 << 16 >> 16;
         $743 = (($742) + 1)|0;
         $744 = HEAP32[$728>>2]|0;
         $745 = (($744) + ($743<<3)|0);
         $746 = $177;
         $747 = $746;
         $748 = HEAP32[$747>>2]|0;
         $749 = (($746) + 4)|0;
         $750 = $749;
         $751 = HEAP32[$750>>2]|0;
         $752 = $745;
         $753 = $752;
         HEAP32[$753>>2] = $748;
         $754 = (($752) + 4)|0;
         $755 = $754;
         HEAP32[$755>>2] = $751;
         $756 = HEAP16[$711>>1]|0;
         $757 = $756 << 16 >> 16;
         $758 = (($757) + 2)|0;
         $759 = HEAP32[$728>>2]|0;
         $760 = (($759) + ($758<<3)|0);
         $761 = $179;
         $762 = $761;
         $763 = HEAP32[$762>>2]|0;
         $764 = (($761) + 4)|0;
         $765 = $764;
         $766 = HEAP32[$765>>2]|0;
         $767 = $760;
         $768 = $767;
         HEAP32[$768>>2] = $763;
         $769 = (($767) + 4)|0;
         $770 = $769;
         HEAP32[$770>>2] = $766;
         $771 = HEAP16[$711>>1]|0;
         $772 = $771 << 16 >> 16;
         $773 = (($772) + 3)|0;
         $774 = HEAP32[$728>>2]|0;
         $775 = (($774) + ($773<<3)|0);
         $776 = $186;
         $777 = $776;
         $778 = HEAP32[$777>>2]|0;
         $779 = (($776) + 4)|0;
         $780 = $779;
         $781 = HEAP32[$780>>2]|0;
         $782 = $775;
         $783 = $782;
         HEAP32[$783>>2] = $778;
         $784 = (($782) + 4)|0;
         $785 = $784;
         HEAP32[$785>>2] = $781;
         $786 = HEAP16[$711>>1]|0;
         $787 = $786 << 16 >> 16;
         $788 = ((($710)) + 28|0);
         $789 = HEAP32[$788>>2]|0;
         $790 = (($789) + ($787)|0);
         HEAP8[$790>>0] = 0;
         $791 = HEAP16[$711>>1]|0;
         $792 = $791 << 16 >> 16;
         $793 = (($792) + 1)|0;
         $794 = HEAP32[$788>>2]|0;
         $795 = (($794) + ($793)|0);
         HEAP8[$795>>0] = 0;
         $796 = HEAP16[$711>>1]|0;
         $797 = $796 << 16 >> 16;
         $798 = (($797) + 2)|0;
         $799 = HEAP32[$788>>2]|0;
         $800 = (($799) + ($798)|0);
         HEAP8[$800>>0] = 0;
         $801 = HEAP16[$711>>1]|0;
         $802 = $801 << 16 >> 16;
         $803 = (($802) + 3)|0;
         $804 = HEAP32[$788>>2]|0;
         $805 = (($804) + ($803)|0);
         HEAP8[$805>>0] = 0;
         $806 = HEAP32[$551>>2]|0;
         $807 = HEAP32[$412>>2]|0;
         $808 = (_FT_Stream_Seek($806,$807)|0);
         HEAP32[$error$i>>2] = $808;
         $809 = ($808|0)==(0);
         if (!($809)) {
          $$1$i26$ph = $808;
          break;
         }
         $810 = (_FT_Stream_ReadUShort($806,$error$i)|0);
         $811 = HEAP32[$error$i>>2]|0;
         $812 = ($811|0)==(0);
         if (!($812)) {
          $$1$i26$ph = $811;
          break;
         }
         $813 = HEAP32[$loader>>2]|0;
         $814 = ((($813)) + 286|0);
         $815 = HEAP16[$814>>1]|0;
         $816 = $810&65535;
         $817 = ($810&65535)>($815&65535);
         do {
          if ($817) {
           $818 = HEAP32[$$pre$phiZ2D>>2]|0;
           $819 = ($816|0)>($818|0);
           if ($819) {
            $$1$i26$ph = 22;
            break L139;
           }
           $820 = ((($loader)) + 160|0);
           $821 = HEAP32[$820>>2]|0;
           $822 = ((($821)) + 388|0);
           $823 = HEAP32[$822>>2]|0;
           $824 = ((($821)) + 8|0);
           $825 = HEAP32[$824>>2]|0;
           $826 = ((($821)) + 392|0);
           $827 = ($823>>>0)<($816>>>0);
           if ($827) {
            $828 = HEAP32[$826>>2]|0;
            $829 = (_ft_mem_realloc($825,1,$823,$816,$828,$error$i$i)|0);
            HEAP32[$826>>2] = $829;
            $830 = HEAP32[$error$i$i>>2]|0;
            $831 = ($830|0)==(0);
            $$$i = $831 ? $816 : $823;
            $$pre9$i = HEAP32[$820>>2]|0;
            $$0$i$i = $830;$834 = $$pre9$i;$tmp$0$i = $$$i;
           } else {
            $$0$i$i = 0;$834 = $821;$tmp$0$i = $823;
           }
           HEAP32[$error$i>>2] = $$0$i$i;
           $832 = $tmp$0$i & 65535;
           $833 = ((($834)) + 388|0);
           HEAP32[$833>>2] = $832;
           $835 = HEAP32[$error$i>>2]|0;
           $836 = ($835|0)==(0);
           if ($836) {
            $$pre$phi$iZ2D = $820;$839 = $834;
           } else {
            $$1$i26$ph = $835;
            break L139;
           }
          } else {
           $837 = ($810<<16>>16)==(0);
           if ($837) {
            HEAP32[$error>>2] = 0;
            break L25;
           } else {
            $$phi$trans$insert$i = ((($loader)) + 160|0);
            $$pre11$i = HEAP32[$$phi$trans$insert$i>>2]|0;
            $$pre$phi$iZ2D = $$phi$trans$insert$i;$839 = $$pre11$i;
            break;
           }
          }
         } while(0);
         $838 = ((($839)) + 392|0);
         $840 = HEAP32[$838>>2]|0;
         $841 = (_FT_Stream_Read($806,$840,$816)|0);
         HEAP32[$error$i>>2] = $841;
         $842 = ($841|0)==(0);
         if (!($842)) {
          $$1$i26$ph = $841;
          break;
         }
         $843 = HEAP32[$$pre$phi$iZ2D>>2]|0;
         $844 = ((($843)) + 392|0);
         $845 = HEAP32[$844>>2]|0;
         $846 = ((($loader)) + 8|0);
         $847 = HEAP32[$846>>2]|0;
         $848 = ((($847)) + 136|0);
         HEAP32[$848>>2] = $845;
         $849 = ((($847)) + 140|0);
         HEAP32[$849>>2] = $816;
         $850 = HEAP32[$1>>2]|0;
         $851 = ((($850)) + 22|0);
         $852 = HEAP16[$851>>1]|0;
         $853 = $852&65535;
         $854 = (($853) - ($404))|0;
         $855 = $854&65535;
         $856 = ((($loader)) + 132|0);
         HEAP16[$856>>1] = $855;
         $857 = ((($850)) + 20|0);
         $858 = HEAP16[$857>>1]|0;
         $859 = $858&65535;
         $860 = (($859) - ($407))|0;
         $861 = $860&65535;
         $862 = ((($loader)) + 134|0);
         HEAP16[$862>>1] = $861;
         $863 = ((($850)) + 40|0);
         $864 = HEAP32[$863>>2]|0;
         $865 = (($864) + ($404<<3)|0);
         $866 = ((($loader)) + 136|0);
         HEAP32[$866>>2] = $865;
         $867 = ((($850)) + 24|0);
         $868 = HEAP32[$867>>2]|0;
         $869 = (($868) + ($404<<3)|0);
         $870 = ((($loader)) + 140|0);
         HEAP32[$870>>2] = $869;
         $871 = ((($850)) + 44|0);
         $872 = HEAP32[$871>>2]|0;
         $873 = (($872) + ($404<<3)|0);
         $874 = ((($loader)) + 144|0);
         HEAP32[$874>>2] = $873;
         $875 = ((($850)) + 28|0);
         $876 = HEAP32[$875>>2]|0;
         $877 = (($876) + ($404)|0);
         $878 = ((($loader)) + 148|0);
         HEAP32[$878>>2] = $877;
         $879 = ((($850)) + 32|0);
         $880 = HEAP32[$879>>2]|0;
         $881 = (($880) + ($407<<1)|0);
         $882 = ((($loader)) + 152|0);
         HEAP32[$882>>2] = $881;
         $883 = ((($loader)) + 156|0);
         HEAP16[$883>>1] = $403;
         $884 = $854 & 65535;
         $885 = ($855<<16>>16)==(0);
         do {
          if ($885) {
           $$lcssa5$i = $884;
          } else {
           $886 = HEAP8[$877>>0]|0;
           $887 = $886&255;
           $888 = $887 & 231;
           $889 = $888&255;
           HEAP8[$877>>0] = $889;
           $890 = HEAP16[$856>>1]|0;
           $891 = $890&65535;
           $892 = ($890&65535)>(1);
           if ($892) {
            $894 = 1;
           } else {
            $$lcssa5$i = $891;
            break;
           }
           while(1) {
            $$pre13$i = HEAP32[$878>>2]|0;
            $893 = (($$pre13$i) + ($894)|0);
            $895 = HEAP8[$893>>0]|0;
            $896 = $895&255;
            $897 = $896 & 231;
            $898 = $897&255;
            HEAP8[$893>>0] = $898;
            $899 = (($894) + 1)|0;
            $900 = HEAP16[$856>>1]|0;
            $901 = $900&65535;
            $902 = ($899>>>0)<($901>>>0);
            if ($902) {
             $894 = $899;
            } else {
             $$lcssa5$i = $901;
             break;
            }
           }
          }
         } while(0);
         $903 = (($$lcssa5$i) + 4)|0;
         $904 = $903&65535;
         HEAP16[$856>>1] = $904;
         $905 = (_TT_Hint_Glyph($loader,1)|0);
         HEAP32[$error>>2] = $905;
         break L25;
        }
       } while(0);
       HEAP32[$error>>2] = $$1$i26$ph;
       break L25;
      }
     }
    }
   } while(0);
   L161: do {
    if ((label|0) == 20) {
     $63 = ((($loader)) + 36|0);
     ;HEAP32[$63>>2]=0|0;HEAP32[$63+4>>2]=0|0;HEAP32[$63+8>>2]=0|0;HEAP32[$63+12>>2]=0|0;
     $64 = (_tt_get_metrics($loader,$glyph_index)|0);
     HEAP32[$error>>2] = $64;
     $65 = ($64|0)!=(0);
     $66 = ($header_only<<24>>24)!=(0);
     $or$cond4 = $66 | $65;
     if (!($or$cond4)) {
      $67 = ((($loader)) + 48|0);
      $68 = HEAP32[$63>>2]|0;
      $69 = ((($loader)) + 52|0);
      $70 = HEAP32[$69>>2]|0;
      $71 = (($68) - ($70))|0;
      $72 = ((($loader)) + 68|0);
      HEAP32[$72>>2] = $71;
      $73 = ((($loader)) + 72|0);
      HEAP32[$73>>2] = 0;
      $74 = ((($loader)) + 56|0);
      $75 = HEAP32[$74>>2]|0;
      $76 = (($75) + ($71))|0;
      $77 = ((($loader)) + 76|0);
      HEAP32[$77>>2] = $76;
      $78 = ((($loader)) + 80|0);
      HEAP32[$78>>2] = 0;
      $79 = ((($loader)) + 184|0);
      HEAP32[$79>>2] = 0;
      $80 = HEAP32[$67>>2]|0;
      $81 = ((($loader)) + 176|0);
      $82 = HEAP32[$81>>2]|0;
      $83 = (($82) + ($80))|0;
      $84 = ((($loader)) + 188|0);
      HEAP32[$84>>2] = $83;
      $85 = ((($loader)) + 192|0);
      HEAP32[$85>>2] = 0;
      $86 = ((($loader)) + 180|0);
      $87 = HEAP32[$86>>2]|0;
      $88 = (($83) - ($87))|0;
      $89 = ((($loader)) + 196|0);
      HEAP32[$89>>2] = $88;
      $90 = HEAP32[$loader>>2]|0;
      $91 = ((($90)) + 128|0);
      $92 = HEAP32[$91>>2]|0;
      $93 = ((($92)) + 48|0);
      $94 = HEAP32[$93>>2]|0;
      $95 = ($94|0)==(0|0);
      if ($95) {
       $116 = $90;
      } else {
       $96 = HEAP32[$94>>2]|0;
       $97 = ((($96)) + 8|0);
       $98 = HEAP32[$97>>2]|0;
       $99 = ($98|0)==(0|0);
       if ($99) {
        $116 = $90;
       } else {
        HEAP32[$metrics$i>>2] = $70;
        $100 = ((($metrics$i)) + 4|0);
        HEAP32[$100>>2] = 0;
        $101 = ((($metrics$i)) + 8|0);
        HEAP32[$101>>2] = $75;
        $102 = ((($metrics$i)) + 12|0);
        HEAP32[$102>>2] = 0;
        $103 = ((($94)) + 4|0);
        $104 = HEAP32[$103>>2]|0;
        $105 = (FUNCTION_TABLE_iiiii[$98 & 8191]($104,$glyph_index,0,$metrics$i)|0);
        $106 = ($105|0)==(0);
        if ($106) {
         $107 = HEAP32[$metrics$i>>2]|0;
         $108 = HEAP32[$101>>2]|0;
         $sext$i = $107 << 16;
         $109 = $sext$i >> 16;
         HEAP32[$69>>2] = $109;
         $110 = $108 & 65535;
         HEAP32[$74>>2] = $110;
         HEAP32[$81>>2] = 0;
         HEAP32[$86>>2] = 0;
         $111 = ((($loader)) + 64|0);
         $112 = HEAP8[$111>>0]|0;
         $113 = ($112<<24>>24)==(0);
         if ($113) {
          HEAP8[$111>>0] = 1;
          $114 = ((($loader)) + 60|0);
          HEAP32[$114>>2] = $110;
         }
        }
        $$pre75 = HEAP32[$loader>>2]|0;
        $116 = $$pre75;
       }
      }
      $115 = ((($116)) + 672|0);
      $117 = HEAP8[$115>>0]|0;
      $118 = ($117<<24>>24)==(0);
      do {
       if (!($118)) {
        HEAP32[$tags>>2] = 16843009;
        $119 = $contours;
        $120 = $119;
        HEAP32[$120>>2] = 65536;
        $121 = (($119) + 4)|0;
        $122 = $121;
        HEAP32[$122>>2] = 196610;
        $123 = HEAP32[$72>>2]|0;
        HEAP32[$points>>2] = $123;
        $124 = HEAP32[$73>>2]|0;
        $125 = ((($points)) + 4|0);
        HEAP32[$125>>2] = $124;
        $126 = HEAP32[$77>>2]|0;
        $127 = ((($points)) + 8|0);
        HEAP32[$127>>2] = $126;
        $128 = HEAP32[$78>>2]|0;
        $129 = ((($points)) + 12|0);
        HEAP32[$129>>2] = $128;
        $130 = HEAP32[$79>>2]|0;
        $131 = ((($points)) + 16|0);
        HEAP32[$131>>2] = $130;
        $132 = HEAP32[$84>>2]|0;
        $133 = ((($points)) + 20|0);
        HEAP32[$133>>2] = $132;
        $134 = HEAP32[$85>>2]|0;
        $135 = ((($points)) + 24|0);
        HEAP32[$135>>2] = $134;
        $136 = HEAP32[$89>>2]|0;
        $137 = ((($points)) + 28|0);
        HEAP32[$137>>2] = $136;
        $138 = ((($outline)) + 2|0);
        HEAP16[$138>>1] = 4;
        HEAP16[$outline>>1] = 4;
        $139 = ((($outline)) + 4|0);
        HEAP32[$139>>2] = $points;
        $140 = ((($outline)) + 8|0);
        HEAP32[$140>>2] = $tags;
        $141 = ((($outline)) + 12|0);
        HEAP32[$141>>2] = $contours;
        $$idx13 = ((($116)) + 104|0);
        $$idx13$val = HEAP32[$$idx13>>2]|0;
        $$idx14$val = HEAP8[$115>>0]|0;
        $$idx15 = ((($116)) + 676|0);
        $$idx15$val = HEAP32[$$idx15>>2]|0;
        $142 = (_TT_Vary_Apply_Glyph_Deltas($$idx13$val,$$idx14$val,$$idx15$val,$glyph_index,$outline,4)|0);
        HEAP32[$error>>2] = $142;
        $143 = ($142|0)==(0);
        if ($143) {
         $144 = HEAP32[$points>>2]|0;
         HEAP32[$72>>2] = $144;
         $145 = HEAP32[$125>>2]|0;
         HEAP32[$73>>2] = $145;
         $146 = HEAP32[$127>>2]|0;
         HEAP32[$77>>2] = $146;
         $147 = HEAP32[$129>>2]|0;
         HEAP32[$78>>2] = $147;
         $148 = HEAP32[$131>>2]|0;
         HEAP32[$79>>2] = $148;
         $149 = HEAP32[$133>>2]|0;
         HEAP32[$84>>2] = $149;
         $150 = HEAP32[$135>>2]|0;
         HEAP32[$85>>2] = $150;
         $151 = HEAP32[$137>>2]|0;
         HEAP32[$89>>2] = $151;
         break;
        } else {
         break L161;
        }
       }
      } while(0);
      $152 = HEAP32[$9>>2]|0;
      $153 = $152 & 1;
      $154 = ($153|0)==(0);
      if ($154) {
       $155 = HEAP32[$72>>2]|0;
       $156 = (_FT_MulFix($155,$x_scale$0)|0);
       HEAP32[$72>>2] = $156;
       $157 = HEAP32[$77>>2]|0;
       $158 = (_FT_MulFix($157,$x_scale$0)|0);
       HEAP32[$77>>2] = $158;
       $159 = HEAP32[$79>>2]|0;
       $160 = (_FT_MulFix($159,$x_scale$0)|0);
       HEAP32[$79>>2] = $160;
       $161 = HEAP32[$84>>2]|0;
       $162 = (_FT_MulFix($161,$y_scale$0)|0);
       HEAP32[$84>>2] = $162;
       $163 = HEAP32[$85>>2]|0;
       $164 = (_FT_MulFix($163,$x_scale$0)|0);
       HEAP32[$85>>2] = $164;
       $165 = HEAP32[$89>>2]|0;
       $166 = (_FT_MulFix($165,$y_scale$0)|0);
       HEAP32[$89>>2] = $166;
      }
      HEAP32[$error>>2] = 0;
     }
    }
   } while(0);
   $906 = ($opened_frame$0<<24>>24)==(0);
   if (!($906)) {
    label = 128;
   }
  }
 } while(0);
 if ((label|0) == 128) {
  $907 = ((($0)) + 516|0);
  $908 = HEAP32[$907>>2]|0;
  FUNCTION_TABLE_vi[$908 & 8191]($loader);
 }
 $909 = ($glyph_data_loaded$0<<24>>24)==(0);
 if ($909) {
  $918 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($918|0);
 }
 $910 = HEAP32[$19>>2]|0;
 $911 = ((($910)) + 48|0);
 $912 = HEAP32[$911>>2]|0;
 $913 = HEAP32[$912>>2]|0;
 $914 = ((($913)) + 4|0);
 $915 = HEAP32[$914>>2]|0;
 $916 = ((($912)) + 4|0);
 $917 = HEAP32[$916>>2]|0;
 FUNCTION_TABLE_vii[$915 & 8191]($917,$glyph_data);
 $918 = HEAP32[$error>>2]|0;
 STACKTOP = sp;return ($918|0);
}
function _tt_get_metrics($loader,$glyph_index) {
 $loader = $loader|0;
 $glyph_index = $glyph_index|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $advance_height = 0, $advance_width = 0, $left_bearing = 0, $sext$i = 0, $top_bearing = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $left_bearing = sp + 6|0;
 $top_bearing = sp + 4|0;
 $advance_width = sp + 2|0;
 $advance_height = sp;
 $0 = HEAP32[$loader>>2]|0;
 $1 = ((($loader)) + 24|0);
 $2 = HEAP32[$1>>2]|0;
 HEAP16[$left_bearing>>1] = 0;
 HEAP16[$top_bearing>>1] = 0;
 HEAP16[$advance_width>>1] = 0;
 HEAP16[$advance_height>>1] = 0;
 $3 = (_FT_Stream_Pos($2)|0);
 $4 = ((($0)) + 532|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 112|0);
 $7 = HEAP32[$6>>2]|0;
 FUNCTION_TABLE_viiiii[$7 & 8191]($0,0,$glyph_index,$left_bearing,$advance_width);
 $8 = ((($loader)) + 48|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($0)) + 292|0);
 $11 = HEAP8[$10>>0]|0;
 $12 = ($11<<24>>24)==(0);
 do {
  if ($12) {
   $16 = ((($0)) + 364|0);
   $17 = HEAP16[$16>>1]|0;
   $18 = ($17<<16>>16)==(-1);
   if ($18) {
    $34 = ((($0)) + 220|0);
    $35 = HEAP16[$34>>1]|0;
    $36 = $35&65535;
    $37 = (($36) - ($9))|0;
    $38 = $37&65535;
    HEAP16[$top_bearing>>1] = $38;
    $39 = HEAP32[$34>>2]|0;
    $sext$i = $39 << 16;
    $40 = $sext$i >> 16;
    $41 = $39 >> 16;
    $42 = (($40) - ($41))|0;
    $43 = ($42|0)<(0);
    $44 = (0 - ($42))|0;
    $45 = $43 ? $44 : $42;
    $46 = $45&65535;
    HEAP16[$advance_height>>1] = $46;
    break;
   } else {
    $19 = ((($0)) + 434|0);
    $20 = HEAP16[$19>>1]|0;
    $21 = $20&65535;
    $22 = (($21) - ($9))|0;
    $23 = $22&65535;
    HEAP16[$top_bearing>>1] = $23;
    $24 = HEAP16[$19>>1]|0;
    $25 = $24 << 16 >> 16;
    $26 = ((($0)) + 436|0);
    $27 = HEAP16[$26>>1]|0;
    $28 = $27 << 16 >> 16;
    $29 = (($25) - ($28))|0;
    $30 = ($29|0)<(0);
    $31 = (0 - ($29))|0;
    $32 = $30 ? $31 : $29;
    $33 = $32&65535;
    HEAP16[$advance_height>>1] = $33;
    break;
   }
  } else {
   $13 = HEAP32[$4>>2]|0;
   $14 = ((($13)) + 112|0);
   $15 = HEAP32[$14>>2]|0;
   FUNCTION_TABLE_viiiii[$15 & 8191]($0,1,$glyph_index,$top_bearing,$advance_height);
  }
 } while(0);
 $47 = (_FT_Stream_Seek($2,$3)|0);
 $48 = ($47|0)==(0);
 if (!($48)) {
  $$0 = $47;
  STACKTOP = sp;return ($$0|0);
 }
 $49 = HEAP16[$left_bearing>>1]|0;
 $50 = $49 << 16 >> 16;
 $51 = ((($loader)) + 52|0);
 HEAP32[$51>>2] = $50;
 $52 = HEAP16[$advance_width>>1]|0;
 $53 = $52&65535;
 $54 = ((($loader)) + 56|0);
 HEAP32[$54>>2] = $53;
 $55 = HEAP16[$top_bearing>>1]|0;
 $56 = $55 << 16 >> 16;
 $57 = ((($loader)) + 176|0);
 HEAP32[$57>>2] = $56;
 $58 = HEAP16[$advance_height>>1]|0;
 $59 = $58&65535;
 $60 = ((($loader)) + 180|0);
 HEAP32[$60>>2] = $59;
 $61 = ((($loader)) + 64|0);
 $62 = HEAP8[$61>>0]|0;
 $63 = ($62<<24>>24)==(0);
 if (!($63)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP8[$61>>0] = 1;
 $64 = ((($loader)) + 60|0);
 HEAP32[$64>>2] = $53;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _TT_Vary_Apply_Glyph_Deltas($face$0$0$26$val,$face$0$39$val,$face$0$40$val,$glyph_index,$outline,$n_points) {
 $face$0$0$26$val = $face$0$0$26$val|0;
 $face$0$39$val = $face$0$39$val|0;
 $face$0$40$val = $face$0$40$val|0;
 $glyph_index = $glyph_index|0;
 $outline = $outline|0;
 $n_points = $n_points|0;
 var $$0 = 0, $$lcssa = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $contour$0$i = 0, $cur_delta$0$ph$i = 0, $cur_delta$0$ph$i$lcssa = 0, $error = 0, $exitcond = 0, $exitcond$i$i = 0;
 var $exitcond36 = 0, $exitcond6$i$i = 0, $has_delta$0 = 0, $i$026 = 0, $im_end_coords$0 = 0, $im_start_coords$0 = 0, $j$015 = 0, $j$116 = 0, $j$218 = 0, $j$324 = 0, $j$420 = 0, $j$522 = 0, $localpoints$027 = 0, $localpoints$1 = 0, $localpoints$2 = 0, $localpoints$3$ph = 0, $offsetToData$0 = 0, $offsetToData$128 = 0, $offsetToData$2$ph = 0, $or$cond = 0;
 var $or$cond$i$i = 0, $or$cond3 = 0, $or$cond5 = 0, $p$04$i$i = 0, $p$13$i$i = 0, $p$13$in$i$i = 0, $point$0$i = 0, $point$16$i = 0, $point$16$i$lcssa = 0, $point$2$i = 0, $point$2$i$lcssa = 0, $point$2$i$lcssa58 = 0, $point$2$in$i = 0, $point$2$in$i$lcssa57 = 0, $point$3$i = 0, $point_count = 0, $points$0 = 0, $sharedpoints$0 = 0, $spoint_count = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp + 8|0;
 $point_count = sp + 4|0;
 $spoint_count = sp;
 $0 = ((($face$0$0$26$val)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$spoint_count>>2] = 0;
 $2 = ($face$0$39$val<<24>>24)==(0);
 $3 = ($face$0$40$val|0)==(0|0);
 $or$cond = $2 | $3;
 if ($or$cond) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ((($face$0$40$val)) + 32|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5>>>0)>($glyph_index>>>0);
 if (!($6)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $7 = ((($face$0$40$val)) + 36|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (($8) + ($glyph_index<<2)|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($glyph_index) + 1)|0;
 $12 = (($8) + ($11<<2)|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($10|0)==($13|0);
 if ($14) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $15 = (_ft_mem_realloc($1,8,0,$n_points,0,$error)|0);
 $16 = HEAP32[$error>>2]|0;
 $17 = ($16|0)==(0);
 if ($17) {
  $18 = (_ft_mem_realloc($1,1,0,$n_points,0,$error)|0);
  $19 = HEAP32[$error>>2]|0;
  $20 = ($19|0)==(0);
  if ($20) {
   $21 = HEAP32[$7>>2]|0;
   $22 = (($21) + ($glyph_index<<2)|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = (_FT_Stream_Seek($face$0$0$26$val,$23)|0);
   HEAP32[$error>>2] = $24;
   $25 = ($24|0)==(0);
   if ($25) {
    $26 = HEAP32[$7>>2]|0;
    $27 = (($26) + ($11<<2)|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = (($26) + ($glyph_index<<2)|0);
    $30 = HEAP32[$29>>2]|0;
    $31 = (($28) - ($30))|0;
    $32 = (_FT_Stream_EnterFrame($face$0$0$26$val,$31)|0);
    HEAP32[$error>>2] = $32;
    $33 = ($32|0)==(0);
    if ($33) {
     $34 = ((($face$0$0$26$val)) + 32|0);
     $35 = HEAP32[$34>>2]|0;
     $36 = HEAP32[$face$0$0$26$val>>2]|0;
     $37 = (($35) - ($36))|0;
     $38 = HEAP32[$face$0$40$val>>2]|0;
     $39 = (_ft_mem_realloc($1,4,0,$38,0,$error)|0);
     $40 = HEAP32[$error>>2]|0;
     $41 = ($40|0)==(0);
     L15: do {
      if ($41) {
       $42 = HEAP32[$face$0$40$val>>2]|0;
       $43 = (_ft_mem_realloc($1,4,0,$42,0,$error)|0);
       $44 = HEAP32[$error>>2]|0;
       $45 = ($44|0)==(0);
       if ($45) {
        $46 = HEAP32[$face$0$40$val>>2]|0;
        $47 = (_ft_mem_realloc($1,4,0,$46,0,$error)|0);
        $48 = HEAP32[$error>>2]|0;
        $49 = ($48|0)==(0);
        if ($49) {
         $50 = (_FT_Stream_GetUShort($face$0$0$26$val)|0);
         $51 = $50&65535;
         $52 = (_FT_Stream_GetUShort($face$0$0$26$val)|0);
         $53 = $52&65535;
         $54 = (($53) + ($37))|0;
         $55 = $51 & 32768;
         $56 = ($55|0)==(0);
         if ($56) {
          $offsetToData$0 = $54;$sharedpoints$0 = 0;
         } else {
          $57 = HEAP32[$34>>2]|0;
          $58 = HEAP32[$face$0$0$26$val>>2]|0;
          $59 = $58;
          $60 = (($57) - ($59))|0;
          $61 = (($58) + ($54)|0);
          HEAP32[$34>>2] = $61;
          $62 = (_ft_var_readpackedpoints($face$0$0$26$val,$spoint_count)|0);
          $63 = HEAP32[$34>>2]|0;
          $64 = HEAP32[$face$0$0$26$val>>2]|0;
          $65 = $64;
          $66 = (($63) - ($65))|0;
          $67 = (($64) + ($60)|0);
          HEAP32[$34>>2] = $67;
          $offsetToData$0 = $66;$sharedpoints$0 = $62;
         }
         $68 = $51 & 4095;
         $69 = ($68|0)==(0);
         if ($69) {
          $im_end_coords$0 = $47;$im_start_coords$0 = $43;
         } else {
          $70 = ((($face$0$40$val)) + 24|0);
          $71 = ((($face$0$40$val)) + 28|0);
          $72 = HEAP32[$spoint_count>>2]|0;
          $73 = ($n_points|0)==(0);
          $74 = ((($outline)) + 4|0);
          $75 = ((($outline)) + 12|0);
          $i$026 = 0;$localpoints$027 = 0;$offsetToData$128 = $offsetToData$0;
          while(1) {
           $76 = (_FT_Stream_GetUShort($face$0$0$26$val)|0);
           $77 = $76&65535;
           $78 = (_FT_Stream_GetUShort($face$0$0$26$val)|0);
           $79 = $78&65535;
           $80 = $79 & 32768;
           $81 = ($80|0)==(0);
           if ($81) {
            $91 = $79 & 4095;
            $92 = HEAP32[$70>>2]|0;
            $93 = ($91>>>0)<($92>>>0);
            if (!($93)) {
             break;
            }
            $94 = HEAP32[$face$0$40$val>>2]|0;
            $95 = Math_imul($94, $91)|0;
            $96 = HEAP32[$71>>2]|0;
            $97 = (($96) + ($95<<2)|0);
            $98 = $94 << 2;
            _memcpy(($39|0),($97|0),($98|0))|0;
           } else {
            $82 = HEAP32[$face$0$40$val>>2]|0;
            $83 = ($82|0)==(0);
            if (!($83)) {
             $j$015 = 0;
             while(1) {
              $84 = (_FT_Stream_GetUShort($face$0$0$26$val)|0);
              $85 = $84 << 16 >> 16;
              $86 = $85 << 2;
              $87 = (($39) + ($j$015<<2)|0);
              HEAP32[$87>>2] = $86;
              $88 = (($j$015) + 1)|0;
              $89 = HEAP32[$face$0$40$val>>2]|0;
              $90 = ($88>>>0)<($89>>>0);
              if ($90) {
               $j$015 = $88;
              } else {
               break;
              }
             }
            }
           }
           $99 = $79 & 16384;
           $100 = ($99|0)==(0);
           if (!($100)) {
            $101 = HEAP32[$face$0$40$val>>2]|0;
            $102 = ($101|0)==(0);
            if (!($102)) {
             $j$116 = 0;
             while(1) {
              $104 = (_FT_Stream_GetUShort($face$0$0$26$val)|0);
              $105 = $104 << 16 >> 16;
              $106 = $105 << 2;
              $107 = (($43) + ($j$116<<2)|0);
              HEAP32[$107>>2] = $106;
              $108 = (($j$116) + 1)|0;
              $109 = HEAP32[$face$0$40$val>>2]|0;
              $110 = ($108>>>0)<($109>>>0);
              if ($110) {
               $j$116 = $108;
              } else {
               $$lcssa = $109;
               break;
              }
             }
             $103 = ($$lcssa|0)==(0);
             if (!($103)) {
              $j$218 = 0;
              while(1) {
               $111 = (_FT_Stream_GetUShort($face$0$0$26$val)|0);
               $112 = $111 << 16 >> 16;
               $113 = $112 << 2;
               $114 = (($47) + ($j$218<<2)|0);
               HEAP32[$114>>2] = $113;
               $115 = (($j$218) + 1)|0;
               $116 = HEAP32[$face$0$40$val>>2]|0;
               $117 = ($115>>>0)<($116>>>0);
               if ($117) {
                $j$218 = $115;
               } else {
                break;
               }
              }
             }
            }
           }
           $118 = (_ft_var_apply_tuple($face$0$40$val,$78,$39,$43,$47)|0);
           $119 = ($118|0)==(0);
           if ($119) {
            $localpoints$3$ph = $localpoints$027;
           } else {
            $120 = HEAP32[$34>>2]|0;
            $121 = HEAP32[$face$0$0$26$val>>2]|0;
            $122 = $121;
            $123 = (($120) - ($122))|0;
            $124 = $79 & 8192;
            $125 = ($124|0)==(0);
            if ($125) {
             HEAP32[$point_count>>2] = $72;
             $129 = $72;$localpoints$1 = $localpoints$027;$points$0 = $sharedpoints$0;
            } else {
             $126 = (($121) + ($offsetToData$128)|0);
             HEAP32[$34>>2] = $126;
             $127 = (_ft_var_readpackedpoints($face$0$0$26$val,$point_count)|0);
             $$pre = HEAP32[$point_count>>2]|0;
             $129 = $$pre;$localpoints$1 = $127;$points$0 = $127;
            }
            $128 = ($129|0)==(0);
            $130 = $128 ? $n_points : $129;
            $131 = (_ft_var_readpackeddeltas($face$0$0$26$val,$130)|0);
            $132 = (_ft_var_readpackeddeltas($face$0$0$26$val,$130)|0);
            $133 = ($points$0|0)==(0|0);
            $134 = ($132|0)==(0|0);
            $or$cond3 = $133 | $134;
            $135 = ($131|0)==(0|0);
            $or$cond5 = $135 | $or$cond3;
            L46: do {
             if (!($or$cond5)) {
              $136 = ($points$0|0)==((-1)|0);
              if ($136) {
               if ($73) {
                break;
               } else {
                $j$324 = 0;
               }
               while(1) {
                $137 = (($131) + ($j$324<<1)|0);
                $138 = HEAP16[$137>>1]|0;
                $139 = $138 << 16 >> 16;
                $140 = (_FT_MulFix($139,$118)|0);
                $141 = HEAP32[$74>>2]|0;
                $142 = (($141) + ($j$324<<3)|0);
                $143 = HEAP32[$142>>2]|0;
                $144 = (($143) + ($140))|0;
                HEAP32[$142>>2] = $144;
                $145 = (($132) + ($j$324<<1)|0);
                $146 = HEAP16[$145>>1]|0;
                $147 = $146 << 16 >> 16;
                $148 = (_FT_MulFix($147,$118)|0);
                $149 = HEAP32[$74>>2]|0;
                $150 = (((($149) + ($j$324<<3)|0)) + 4|0);
                $151 = HEAP32[$150>>2]|0;
                $152 = (($151) + ($148))|0;
                HEAP32[$150>>2] = $152;
                $153 = (($j$324) + 1)|0;
                $exitcond36 = ($153|0)==($n_points|0);
                if ($exitcond36) {
                 break L46;
                } else {
                 $j$324 = $153;
                }
               }
              }
              if (!($73)) {
               $j$420 = 0;
               while(1) {
                $154 = (($15) + ($j$420<<3)|0);
                $155 = HEAP32[$74>>2]|0;
                $156 = (($155) + ($j$420<<3)|0);
                $157 = $156;
                $158 = $157;
                $159 = HEAP32[$158>>2]|0;
                $160 = (($157) + 4)|0;
                $161 = $160;
                $162 = HEAP32[$161>>2]|0;
                $163 = $154;
                $164 = $163;
                HEAP32[$164>>2] = $159;
                $165 = (($163) + 4)|0;
                $166 = $165;
                HEAP32[$166>>2] = $162;
                $167 = (($18) + ($j$420)|0);
                HEAP8[$167>>0] = 0;
                $168 = (($j$420) + 1)|0;
                $exitcond = ($168|0)==($n_points|0);
                if ($exitcond) {
                 break;
                } else {
                 $j$420 = $168;
                }
               }
              }
              if (!($128)) {
               $j$522 = 0;
               while(1) {
                $169 = (($localpoints$1) + ($j$522<<1)|0);
                $170 = HEAP16[$169>>1]|0;
                $171 = $170&65535;
                $172 = ($171>>>0)<($n_points>>>0);
                if ($172) {
                 $173 = (($18) + ($171)|0);
                 HEAP8[$173>>0] = 1;
                 $174 = (($131) + ($j$522<<1)|0);
                 $175 = HEAP16[$174>>1]|0;
                 $176 = $175 << 16 >> 16;
                 $177 = (_FT_MulFix($176,$118)|0);
                 $178 = HEAP32[$74>>2]|0;
                 $179 = (($178) + ($171<<3)|0);
                 $180 = HEAP32[$179>>2]|0;
                 $181 = (($180) + ($177))|0;
                 HEAP32[$179>>2] = $181;
                 $182 = (($132) + ($j$522<<1)|0);
                 $183 = HEAP16[$182>>1]|0;
                 $184 = $183 << 16 >> 16;
                 $185 = (_FT_MulFix($184,$118)|0);
                 $186 = HEAP32[$74>>2]|0;
                 $187 = (((($186) + ($171<<3)|0)) + 4|0);
                 $188 = HEAP32[$187>>2]|0;
                 $189 = (($188) + ($185))|0;
                 HEAP32[$187>>2] = $189;
                }
                $190 = (($j$522) + 1)|0;
                $191 = ($190>>>0)<($129>>>0);
                if ($191) {
                 $j$522 = $190;
                } else {
                 break;
                }
               }
              }
              $192 = HEAP16[$outline>>1]|0;
              $193 = ($192<<16>>16)==(0);
              if ($193) {
               break;
              }
              $194 = HEAP32[$74>>2]|0;
              $contour$0$i = 0;$point$0$i = 0;
              while(1) {
               $195 = $contour$0$i << 16 >> 16;
               $196 = HEAP32[$75>>2]|0;
               $197 = (($196) + ($195<<1)|0);
               $198 = HEAP16[$197>>1]|0;
               $199 = $198 << 16 >> 16;
               $200 = ($point$0$i>>>0)>($199>>>0);
               L67: do {
                if ($200) {
                 $point$3$i = $point$0$i;
                } else {
                 $point$16$i = $point$0$i;
                 while(1) {
                  $201 = (($18) + ($point$16$i)|0);
                  $202 = HEAP8[$201>>0]|0;
                  $203 = ($202<<24>>24)==(0);
                  if (!($203)) {
                   $point$16$i$lcssa = $point$16$i;
                   break;
                  }
                  $204 = (($point$16$i) + 1)|0;
                  $205 = ($204>>>0)>($199>>>0);
                  if ($205) {
                   $point$3$i = $204;
                   break L67;
                  } else {
                   $point$16$i = $204;
                  }
                 }
                 $cur_delta$0$ph$i = $point$16$i$lcssa;
                 L72: while(1) {
                  $point$2$in$i = $cur_delta$0$ph$i;
                  while(1) {
                   $point$2$i = (($point$2$in$i) + 1)|0;
                   $206 = ($point$2$i>>>0)>($199>>>0);
                   if ($206) {
                    $cur_delta$0$ph$i$lcssa = $cur_delta$0$ph$i;$point$2$i$lcssa = $point$2$i;
                    break L72;
                   }
                   $207 = (($18) + ($point$2$i)|0);
                   $208 = HEAP8[$207>>0]|0;
                   $209 = ($208<<24>>24)==(0);
                   if ($209) {
                    $point$2$in$i = $point$2$i;
                   } else {
                    $point$2$i$lcssa58 = $point$2$i;$point$2$in$i$lcssa57 = $point$2$in$i;
                    break;
                   }
                  }
                  $210 = (($cur_delta$0$ph$i) + 1)|0;
                  _tt_delta_interpolate($210,$point$2$in$i$lcssa57,$cur_delta$0$ph$i,$point$2$i$lcssa58,$15,$194);
                  $cur_delta$0$ph$i = $point$2$i$lcssa58;
                 }
                 $211 = ($cur_delta$0$ph$i$lcssa|0)==($point$16$i$lcssa|0);
                 if (!($211)) {
                  $239 = (($cur_delta$0$ph$i$lcssa) + 1)|0;
                  _tt_delta_interpolate($239,$199,$cur_delta$0$ph$i$lcssa,$point$16$i$lcssa,$15,$194);
                  $240 = ($point$16$i$lcssa|0)==(0);
                  if ($240) {
                   $point$3$i = $point$2$i$lcssa;
                   break;
                  }
                  $241 = (($point$16$i$lcssa) + -1)|0;
                  _tt_delta_interpolate($point$0$i,$241,$cur_delta$0$ph$i$lcssa,$point$16$i$lcssa,$15,$194);
                  $point$3$i = $point$2$i$lcssa;
                  break;
                 }
                 $212 = (($194) + ($point$16$i$lcssa<<3)|0);
                 $213 = HEAP32[$212>>2]|0;
                 $214 = (($15) + ($point$16$i$lcssa<<3)|0);
                 $215 = HEAP32[$214>>2]|0;
                 $216 = (($213) - ($215))|0;
                 $217 = (((($194) + ($point$16$i$lcssa<<3)|0)) + 4|0);
                 $218 = HEAP32[$217>>2]|0;
                 $219 = (((($15) + ($point$16$i$lcssa<<3)|0)) + 4|0);
                 $220 = HEAP32[$219>>2]|0;
                 $221 = (($218) - ($220))|0;
                 $222 = ($213|0)==($215|0);
                 $223 = ($218|0)==($220|0);
                 $or$cond$i$i = $222 & $223;
                 if ($or$cond$i$i) {
                  $point$3$i = $point$2$i$lcssa;
                  break;
                 }
                 $224 = ($point$0$i|0)<($point$16$i$lcssa|0);
                 if ($224) {
                  $p$04$i$i = $point$0$i;
                  while(1) {
                   $226 = (($194) + ($p$04$i$i<<3)|0);
                   $227 = HEAP32[$226>>2]|0;
                   $228 = (($227) + ($216))|0;
                   HEAP32[$226>>2] = $228;
                   $229 = (((($194) + ($p$04$i$i<<3)|0)) + 4|0);
                   $230 = HEAP32[$229>>2]|0;
                   $231 = (($230) + ($221))|0;
                   HEAP32[$229>>2] = $231;
                   $232 = (($p$04$i$i) + 1)|0;
                   $exitcond6$i$i = ($232|0)==($point$16$i$lcssa|0);
                   if ($exitcond6$i$i) {
                    break;
                   } else {
                    $p$04$i$i = $232;
                   }
                  }
                 }
                 $225 = ($point$16$i$lcssa|0)<($199|0);
                 if ($225) {
                  $p$13$in$i$i = $point$16$i$lcssa;
                 } else {
                  $point$3$i = $point$2$i$lcssa;
                  break;
                 }
                 while(1) {
                  $p$13$i$i = (($p$13$in$i$i) + 1)|0;
                  $233 = (($194) + ($p$13$i$i<<3)|0);
                  $234 = HEAP32[$233>>2]|0;
                  $235 = (($234) + ($216))|0;
                  HEAP32[$233>>2] = $235;
                  $236 = (((($194) + ($p$13$i$i<<3)|0)) + 4|0);
                  $237 = HEAP32[$236>>2]|0;
                  $238 = (($237) + ($221))|0;
                  HEAP32[$236>>2] = $238;
                  $exitcond$i$i = ($p$13$i$i|0)==($199|0);
                  if ($exitcond$i$i) {
                   $point$3$i = $point$2$i$lcssa;
                   break;
                  } else {
                   $p$13$in$i$i = $p$13$i$i;
                  }
                 }
                }
               } while(0);
               $242 = (($contour$0$i) + 1)<<16>>16;
               $243 = HEAP16[$outline>>1]|0;
               $244 = ($242<<16>>16)<($243<<16>>16);
               if ($244) {
                $contour$0$i = $242;$point$0$i = $point$3$i;
               } else {
                break;
               }
              }
             }
            } while(0);
            $245 = ($localpoints$1|0)==((-1)|0);
            if ($245) {
             $localpoints$2 = (-1);
            } else {
             _ft_mem_free($1,$localpoints$1);
             $localpoints$2 = 0;
            }
            _ft_mem_free($1,$131);
            _ft_mem_free($1,$132);
            $246 = HEAP32[$face$0$0$26$val>>2]|0;
            $247 = (($246) + ($123)|0);
            HEAP32[$34>>2] = $247;
            $localpoints$3$ph = $localpoints$2;
           }
           $offsetToData$2$ph = (($77) + ($offsetToData$128))|0;
           $248 = (($i$026) + 1)|0;
           $249 = ($248>>>0)<($68>>>0);
           if ($249) {
            $i$026 = $248;$localpoints$027 = $localpoints$3$ph;$offsetToData$128 = $offsetToData$2$ph;
           } else {
            $im_end_coords$0 = $47;$im_start_coords$0 = $43;
            break L15;
           }
          }
          HEAP32[$error>>2] = 8;
          $im_end_coords$0 = $47;$im_start_coords$0 = $43;
         }
        } else {
         $im_end_coords$0 = $47;$im_start_coords$0 = $43;
        }
       } else {
        $im_end_coords$0 = 0;$im_start_coords$0 = $43;
       }
      } else {
       $im_end_coords$0 = 0;$im_start_coords$0 = 0;
      }
     } while(0);
     _ft_mem_free($1,$39);
     _ft_mem_free($1,$im_start_coords$0);
     _ft_mem_free($1,$im_end_coords$0);
     _FT_Stream_ExitFrame($face$0$0$26$val);
     $has_delta$0 = $18;
    } else {
     $has_delta$0 = $18;
    }
   } else {
    $has_delta$0 = $18;
   }
  } else {
   $has_delta$0 = $18;
  }
 } else {
  $has_delta$0 = 0;
 }
 _ft_mem_free($1,$15);
 _ft_mem_free($1,$has_delta$0);
 $250 = HEAP32[$error>>2]|0;
 $$0 = $250;
 STACKTOP = sp;return ($$0|0);
}
function _tt_delta_interpolate($p1,$p2,$ref1,$ref2,$in_points,$out_points) {
 $p1 = $p1|0;
 $p2 = $p2|0;
 $ref1 = $ref1|0;
 $ref2 = $ref2|0;
 $in_points = $in_points|0;
 $out_points = $out_points|0;
 var $$ = 0, $$$1 = 0, $$0$$01 = 0, $$0$$01$1 = 0, $$01$$0 = 0, $$01$$0$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $9 = 0, $or$cond = 0, $or$cond$1 = 0, $out$0 = 0, $out$0$1 = 0, $out$1 = 0, $out$1$1 = 0, $p$09 = 0, $p$09$1 = 0, $p$18 = 0;
 var $p$18$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($p1|0)>($p2|0);
 if ($0) {
  return;
 }
 $1 = (($in_points) + ($ref1<<3)|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($in_points) + ($ref2<<3)|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2|0)>($4|0);
 $$0$$01 = $5 ? $ref1 : $ref2;
 $$01$$0 = $5 ? $ref2 : $ref1;
 $6 = (($in_points) + ($$01$$0<<3)|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (($in_points) + ($$0$$01<<3)|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (($out_points) + ($$01$$0<<3)|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($out_points) + ($$0$$01<<3)|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11) - ($7))|0;
 $15 = (($13) - ($9))|0;
 $16 = ($13|0)==($11|0);
 $17 = ($9|0)==($7|0);
 $or$cond = $17 | $16;
 if ($or$cond) {
  $p$09 = $p1;
  while(1) {
   $18 = (($in_points) + ($p$09<<3)|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ($19|0)>($7|0);
   if ($20) {
    $22 = ($19|0)<($9|0);
    $23 = (($19) + ($15))|0;
    $$ = $22 ? $11 : $23;
    $out$0 = $$;
   } else {
    $21 = (($19) + ($14))|0;
    $out$0 = $21;
   }
   $24 = (($out_points) + ($p$09<<3)|0);
   HEAP32[$24>>2] = $out$0;
   $25 = (($p$09) + 1)|0;
   $26 = ($p$09|0)<($p2|0);
   if ($26) {
    $p$09 = $25;
   } else {
    break;
   }
  }
 } else {
  $27 = (($13) - ($11))|0;
  $28 = (($9) - ($7))|0;
  $29 = (_FT_DivFix($27,$28)|0);
  $p$18 = $p1;
  while(1) {
   $30 = (($in_points) + ($p$18<<3)|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = ($31|0)>($7|0);
   do {
    if ($32) {
     $34 = ($31|0)<($9|0);
     if ($34) {
      $36 = (($31) - ($7))|0;
      $37 = (_FT_MulFix($36,$29)|0);
      $38 = (($37) + ($11))|0;
      $out$1 = $38;
      break;
     } else {
      $35 = (($31) + ($15))|0;
      $out$1 = $35;
      break;
     }
    } else {
     $33 = (($31) + ($14))|0;
     $out$1 = $33;
    }
   } while(0);
   $39 = (($out_points) + ($p$18<<3)|0);
   HEAP32[$39>>2] = $out$1;
   $40 = (($p$18) + 1)|0;
   $41 = ($p$18|0)<($p2|0);
   if ($41) {
    $p$18 = $40;
   } else {
    break;
   }
  }
 }
 $42 = ((($in_points)) + 4|0);
 $43 = ((($out_points)) + 4|0);
 $44 = (($42) + ($$01$$0<<3)|0);
 $45 = HEAP32[$44>>2]|0;
 $46 = (($42) + ($$0$$01<<3)|0);
 $47 = HEAP32[$46>>2]|0;
 $48 = ($45|0)>($47|0);
 $$0$$01$1 = $48 ? $$01$$0 : $$0$$01;
 $$01$$0$1 = $48 ? $$0$$01 : $$01$$0;
 $49 = (($42) + ($$01$$0$1<<3)|0);
 $50 = HEAP32[$49>>2]|0;
 $51 = (($42) + ($$0$$01$1<<3)|0);
 $52 = HEAP32[$51>>2]|0;
 $53 = (($43) + ($$01$$0$1<<3)|0);
 $54 = HEAP32[$53>>2]|0;
 $55 = (($43) + ($$0$$01$1<<3)|0);
 $56 = HEAP32[$55>>2]|0;
 $57 = (($54) - ($50))|0;
 $58 = (($56) - ($52))|0;
 $59 = ($56|0)==($54|0);
 $60 = ($52|0)==($50|0);
 $or$cond$1 = $60 | $59;
 if ($or$cond$1) {
  $p$09$1 = $p1;
  while(1) {
   $76 = (($42) + ($p$09$1<<3)|0);
   $77 = HEAP32[$76>>2]|0;
   $78 = ($77|0)>($50|0);
   if ($78) {
    $80 = ($77|0)<($52|0);
    $81 = (($77) + ($58))|0;
    $$$1 = $80 ? $54 : $81;
    $out$0$1 = $$$1;
   } else {
    $79 = (($77) + ($57))|0;
    $out$0$1 = $79;
   }
   $82 = (($43) + ($p$09$1<<3)|0);
   HEAP32[$82>>2] = $out$0$1;
   $83 = (($p$09$1) + 1)|0;
   $84 = ($p$09$1|0)<($p2|0);
   if ($84) {
    $p$09$1 = $83;
   } else {
    break;
   }
  }
  return;
 }
 $61 = (($56) - ($54))|0;
 $62 = (($52) - ($50))|0;
 $63 = (_FT_DivFix($61,$62)|0);
 $p$18$1 = $p1;
 while(1) {
  $64 = (($42) + ($p$18$1<<3)|0);
  $65 = HEAP32[$64>>2]|0;
  $66 = ($65|0)>($50|0);
  do {
   if ($66) {
    $68 = ($65|0)<($52|0);
    if ($68) {
     $70 = (($65) - ($50))|0;
     $71 = (_FT_MulFix($70,$63)|0);
     $72 = (($71) + ($54))|0;
     $out$1$1 = $72;
     break;
    } else {
     $69 = (($65) + ($58))|0;
     $out$1$1 = $69;
     break;
    }
   } else {
    $67 = (($65) + ($57))|0;
    $out$1$1 = $67;
   }
  } while(0);
  $73 = (($43) + ($p$18$1<<3)|0);
  HEAP32[$73>>2] = $out$1$1;
  $74 = (($p$18$1) + 1)|0;
  $75 = ($p$18$1|0)<($p2|0);
  if ($75) {
   $p$18$1 = $74;
  } else {
   break;
  }
 }
 return;
}
function _TT_Hint_Glyph($loader,$is_composite) {
 $loader = $loader|0;
 $is_composite = $is_composite|0;
 var $$1 = 0, $$pre = 0, $$pre$phi6Z2D = 0, $$pre$phiZ2D = 0, $$pre2 = 0, $$pre3 = 0, $$pre4 = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, dest = 0, label = 0;
 var sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($loader)) + 124|0);
 $1 = ((($loader)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 140|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)>(0);
 if ($5) {
  $6 = ((($loader)) + 136|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ((($loader)) + 140|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($loader)) + 132|0);
  $11 = HEAP16[$10>>1]|0;
  $12 = $11&65535;
  $13 = $12 << 3;
  _memcpy(($7|0),($9|0),($13|0))|0;
 }
 $14 = ((($loader)) + 160|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($15)) + 284|0);
 $17 = ((($loader)) + 4|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($18)) + 172|0);
 dest=$16; src=$19; stop=dest+68|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 $20 = ($is_composite<<24>>24)==(0);
 if ($20) {
  $32 = HEAP32[$17>>2]|0;
  $33 = ((($32)) + 48|0);
  $34 = HEAP32[$33>>2]|0;
  $35 = HEAP32[$14>>2]|0;
  $36 = ((($35)) + 220|0);
  HEAP32[$36>>2] = $34;
  $37 = ((($32)) + 52|0);
  $38 = HEAP32[$37>>2]|0;
  $39 = ((($35)) + 224|0);
  HEAP32[$39>>2] = $38;
  $$pre4 = ((($loader)) + 132|0);
  $$pre5 = ((($loader)) + 140|0);
  $$pre$phi6Z2D = $$pre5;$$pre$phiZ2D = $$pre4;
 } else {
  $21 = HEAP32[$14>>2]|0;
  $22 = ((($21)) + 220|0);
  HEAP32[$22>>2] = 65536;
  $23 = ((($21)) + 224|0);
  HEAP32[$23>>2] = 65536;
  $24 = ((($loader)) + 144|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = ((($loader)) + 140|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = ((($loader)) + 132|0);
  $29 = HEAP16[$28>>1]|0;
  $30 = $29&65535;
  $31 = $30 << 3;
  _memcpy(($25|0),($27|0),($31|0))|0;
  $$pre$phi6Z2D = $26;$$pre$phiZ2D = $28;
 }
 $40 = HEAP16[$$pre$phiZ2D>>1]|0;
 $41 = $40&65535;
 $42 = (($41) + -4)|0;
 $43 = HEAP32[$$pre$phi6Z2D>>2]|0;
 $44 = (($43) + ($42<<3)|0);
 $45 = HEAP32[$44>>2]|0;
 $46 = (($45) + 32)|0;
 $47 = $46 & -64;
 HEAP32[$44>>2] = $47;
 $48 = (($41) + -3)|0;
 $49 = (($43) + ($48<<3)|0);
 $50 = HEAP32[$49>>2]|0;
 $51 = (($50) + 32)|0;
 $52 = $51 & -64;
 HEAP32[$49>>2] = $52;
 $53 = (($41) + -2)|0;
 $54 = (((($43) + ($53<<3)|0)) + 4|0);
 $55 = HEAP32[$54>>2]|0;
 $56 = (($55) + 32)|0;
 $57 = $56 & -64;
 HEAP32[$54>>2] = $57;
 $58 = (($41) + -1)|0;
 $59 = (((($43) + ($58<<3)|0)) + 4|0);
 $60 = HEAP32[$59>>2]|0;
 $61 = (($60) + 32)|0;
 $62 = $61 & -64;
 HEAP32[$59>>2] = $62;
 if ($5) {
  $63 = ((($loader)) + 12|0);
  $64 = HEAP32[$63>>2]|0;
  $65 = ((($64)) + 64|0);
  $66 = HEAP32[$65>>2]|0;
  $67 = HEAP32[$14>>2]|0;
  $68 = ((($67)) + 392|0);
  $69 = HEAP32[$68>>2]|0;
  $70 = ((($67)) + 460|0);
  HEAP32[$70>>2] = $69;
  $71 = ((($67)) + 464|0);
  HEAP32[$71>>2] = $4;
  $72 = ((($67)) + 560|0);
  HEAP8[$72>>0] = $is_composite;
  $73 = ((($67)) + 144|0);
  dest=$73; src=$0; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
  $74 = HEAP32[$14>>2]|0;
  $75 = ((($74)) + 460|0);
  $76 = HEAP32[$75>>2]|0;
  $77 = ((($74)) + 356|0);
  HEAP32[$77>>2] = $76;
  $78 = ((($74)) + 464|0);
  $79 = HEAP32[$78>>2]|0;
  $80 = ((($74)) + 364|0);
  HEAP32[$80>>2] = $79;
  $81 = ((($74)) + 360|0);
  HEAP32[$81>>2] = 0;
  $82 = ((($74)) + 352|0);
  HEAP32[$82>>2] = 3;
  $83 = ((($74)) + 36|0);
  $84 = ((($74)) + 144|0);
  dest=$83; src=$84; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
  $85 = ((($74)) + 72|0);
  dest=$85; src=$84; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
  $86 = ((($74)) + 108|0);
  dest=$86; src=$84; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
  $87 = ((($74)) + 344|0);
  HEAP16[$87>>1] = 1;
  $88 = ((($74)) + 346|0);
  HEAP16[$88>>1] = 1;
  $89 = ((($74)) + 348|0);
  HEAP16[$89>>1] = 1;
  $90 = ((($74)) + 294|0);
  HEAP16[$90>>1] = 16384;
  $91 = ((($74)) + 296|0);
  HEAP16[$91>>1] = 0;
  $92 = ((($74)) + 298|0);
  $93 = HEAPU16[$90>>1]|(HEAPU16[$90+2>>1]<<16);
  HEAP16[$92>>1]=$93&65535;HEAP16[$92+2>>1]=$93>>>16;
  $94 = ((($74)) + 290|0);
  HEAP16[$94>>1]=$93&65535;HEAP16[$94+2>>1]=$93>>>16;
  $95 = ((($74)) + 312|0);
  HEAP32[$95>>2] = 1;
  $96 = ((($74)) + 304|0);
  HEAP32[$96>>2] = 1;
  $97 = ((($74)) + 16|0);
  HEAP32[$97>>2] = 0;
  $98 = ((($74)) + 428|0);
  HEAP32[$98>>2] = 0;
  $99 = HEAP32[$74>>2]|0;
  $100 = ((($99)) + 652|0);
  $101 = HEAP32[$100>>2]|0;
  $102 = (FUNCTION_TABLE_ii[$101 & 8191]($74)|0);
  $103 = ($102|0)==(0);
  $$pre = HEAP32[$14>>2]|0;
  if (!($103)) {
   $104 = ((($$pre)) + 561|0);
   $105 = HEAP8[$104>>0]|0;
   $106 = ($105<<24>>24)==(0);
   if (!($106)) {
    $$1 = $102;
    return ($$1|0);
   }
  }
  $107 = ((($$pre)) + 340|0);
  $108 = HEAP32[$107>>2]|0;
  $109 = $108 << 5;
  $110 = HEAP8[$66>>0]|0;
  $111 = $110&255;
  $112 = $109 | $111;
  $113 = $112 | 4;
  $114 = $113&255;
  HEAP8[$66>>0] = $114;
  $$pre2 = HEAP16[$$pre$phiZ2D>>1]|0;
  $$pre3 = HEAP32[$$pre$phi6Z2D>>2]|0;
  $117 = $$pre2;$120 = $$pre3;
 } else {
  $117 = $40;$120 = $43;
 }
 $115 = ((($loader)) + 68|0);
 $116 = $117&65535;
 $118 = (($116) + -4)|0;
 $119 = (($120) + ($118<<3)|0);
 $121 = $119;
 $122 = $121;
 $123 = HEAP32[$122>>2]|0;
 $124 = (($121) + 4)|0;
 $125 = $124;
 $126 = HEAP32[$125>>2]|0;
 $127 = $115;
 $128 = $127;
 HEAP32[$128>>2] = $123;
 $129 = (($127) + 4)|0;
 $130 = $129;
 HEAP32[$130>>2] = $126;
 $131 = ((($loader)) + 76|0);
 $132 = (($116) + -3)|0;
 $133 = (($120) + ($132<<3)|0);
 $134 = $133;
 $135 = $134;
 $136 = HEAP32[$135>>2]|0;
 $137 = (($134) + 4)|0;
 $138 = $137;
 $139 = HEAP32[$138>>2]|0;
 $140 = $131;
 $141 = $140;
 HEAP32[$141>>2] = $136;
 $142 = (($140) + 4)|0;
 $143 = $142;
 HEAP32[$143>>2] = $139;
 $144 = ((($loader)) + 184|0);
 $145 = (($116) + -2)|0;
 $146 = (($120) + ($145<<3)|0);
 $147 = $146;
 $148 = $147;
 $149 = HEAP32[$148>>2]|0;
 $150 = (($147) + 4)|0;
 $151 = $150;
 $152 = HEAP32[$151>>2]|0;
 $153 = $144;
 $154 = $153;
 HEAP32[$154>>2] = $149;
 $155 = (($153) + 4)|0;
 $156 = $155;
 HEAP32[$156>>2] = $152;
 $157 = ((($loader)) + 192|0);
 $158 = (($116) + -1)|0;
 $159 = (($120) + ($158<<3)|0);
 $160 = $159;
 $161 = $160;
 $162 = HEAP32[$161>>2]|0;
 $163 = (($160) + 4)|0;
 $164 = $163;
 $165 = HEAP32[$164>>2]|0;
 $166 = $157;
 $167 = $166;
 HEAP32[$167>>2] = $162;
 $168 = (($166) + 4)|0;
 $169 = $168;
 HEAP32[$169>>2] = $165;
 $$1 = 0;
 return ($$1|0);
}
function _TT_New_Context($driver) {
 $driver = $driver|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $error$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp + 4|0;
 $error = sp;
 $0 = ($driver|0)==(0|0);
 do {
  if (!($0)) {
   $1 = ((($driver)) + 8|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = (_ft_mem_alloc($2,612,$error)|0);
   $4 = HEAP32[$error>>2]|0;
   $5 = ($4|0)==(0);
   if ($5) {
    $6 = ((($3)) + 8|0);
    HEAP32[$6>>2] = $2;
    $7 = ((($3)) + 432|0);
    HEAP32[$7>>2] = 32;
    $8 = (_ft_mem_realloc($2,16,0,32,0,$error$i)|0);
    $9 = ((($3)) + 436|0);
    HEAP32[$9>>2] = $8;
    $10 = HEAP32[$error$i>>2]|0;
    $11 = ($10|0)==(0);
    if ($11) {
     $12 = ((($3)) + 440|0);
     HEAP16[$12>>1] = 0;
     $13 = ((($3)) + 442|0);
     HEAP16[$13>>1] = 0;
     $14 = ((($3)) + 20|0);
     HEAP32[$14>>2] = 0;
     $15 = ((($3)) + 388|0);
     HEAP32[$15>>2] = 0;
     $16 = ((($3)) + 24|0);
     HEAP32[$16>>2] = 0;
     $17 = ((($3)) + 392|0);
     HEAP32[$17>>2] = 0;
     HEAP32[$3>>2] = 0;
     $18 = ((($3)) + 4|0);
     HEAP32[$18>>2] = 0;
     HEAP32[$error>>2] = 0;
     $$0 = $3;
     STACKTOP = sp;return ($$0|0);
    } else {
     $19 = HEAP32[$6>>2]|0;
     $20 = ((($3)) + 440|0);
     HEAP16[$20>>1] = 0;
     $21 = ((($3)) + 442|0);
     HEAP16[$21>>1] = 0;
     $22 = ((($3)) + 24|0);
     $23 = HEAP32[$22>>2]|0;
     _ft_mem_free($19,$23);
     HEAP32[$22>>2] = 0;
     $24 = ((($3)) + 20|0);
     HEAP32[$24>>2] = 0;
     $25 = HEAP32[$9>>2]|0;
     _ft_mem_free($19,$25);
     HEAP32[$9>>2] = 0;
     HEAP32[$7>>2] = 0;
     $26 = ((($3)) + 428|0);
     HEAP32[$26>>2] = 0;
     $27 = ((($3)) + 392|0);
     $28 = HEAP32[$27>>2]|0;
     _ft_mem_free($19,$28);
     HEAP32[$27>>2] = 0;
     $29 = ((($3)) + 388|0);
     HEAP32[$29>>2] = 0;
     $30 = ((($3)) + 4|0);
     HEAP32[$30>>2] = 0;
     HEAP32[$3>>2] = 0;
     _ft_mem_free($19,$3);
     $31 = HEAP32[$error$i>>2]|0;
     HEAP32[$error>>2] = $31;
     $32 = ($31|0)==(0);
     if ($32) {
      $$0 = $3;
     } else {
      break;
     }
     STACKTOP = sp;return ($$0|0);
    }
   }
  }
 } while(0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _TT_RunIns($exc) {
 $exc = $exc|0;
 var $$$$i = 0, $$$i = 0, $$$i$i = 0, $$$i$i136 = 0, $$$i$i160 = 0, $$$i$i187 = 0, $$$i$i54 = 0, $$$i$i56 = 0, $$$i$i98 = 0, $$$i107 = 0, $$$i164 = 0, $$$i192 = 0, $$$i220 = 0, $$$i3$i = 0, $$$i3$i162 = 0, $$$i5$i = 0, $$$i6$i = 0, $$$i76 = 0, $$$i94 = 0, $$01 = 0;
 var $$01$i$i = 0, $$01$i$i135 = 0, $$01$i$i159 = 0, $$01$i2$i = 0, $$01$i2$i161 = 0, $$01$i4$i = 0, $$02$i$i = 0, $$02$i$i223 = 0, $$02$i$i61 = 0, $$02$i$i64 = 0, $$02$i$i68 = 0, $$02$i7$i = 0, $$03$i$i = 0, $$03$i$i222 = 0, $$03$i$i60 = 0, $$03$i$i63 = 0, $$03$i$i67 = 0, $$03$i6$i = 0, $$1$i$i = 0, $$1$i$i101 = 0;
 var $$1$i$i188 = 0, $$4$i = 0, $$5$i = 0, $$a$i$i = 0, $$a$i$i134 = 0, $$a$i$i158 = 0, $$cvt_dist$2$i = 0, $$distance$0$i = 0, $$distance$0$i104 = 0, $$distance$01$i = 0, $$distance$02$i = 0, $$idx = 0, $$idx$val = 0, $$idx23 = 0, $$idx23$val = 0, $$idx3 = 0, $$idx3$val = 0, $$idx31 = 0, $$idx31$val = 0, $$idx38 = 0;
 var $$idx38$val = 0, $$idx40 = 0, $$idx40$val = 0, $$idx44 = 0, $$idx44$val = 0, $$idx46 = 0, $$idx46$val = 0, $$idx5 = 0, $$idx5$val = 0, $$idx7 = 0, $$idx7$val = 0, $$lcssa = 0, $$lcssa270 = 0, $$lcssa271 = 0, $$lcssa273 = 0, $$lcssa620 = 0, $$lcssa623 = 0, $$lcssa624 = 0, $$lcssa625 = 0, $$lcssa626 = 0;
 var $$off$i = 0, $$op$i = 0, $$op$i$i = 0, $$op$i$i57 = 0, $$pn$i = 0, $$pn$in$i = 0, $$pr$i = 0, $$pr$i141 = 0, $$pr$pr = 0, $$pr$pr$lcssa = 0, $$pr$pre$i = 0, $$pre = 0, $$pre$i100 = 0, $$pre$i110 = 0, $$pre$i127 = 0, $$pre$i144 = 0, $$pre$i149 = 0, $$pre$i154 = 0, $$pre$i169 = 0, $$pre$i174 = 0;
 var $$pre$i226 = 0, $$pre$i51 = 0, $$pre$i53 = 0, $$pre$i69 = 0, $$pre$phi3$i150Z2D = 0, $$pre$phi7$iZ2D = 0, $$pre$phiZ2D = 0, $$pre$pre$i = 0, $$pre1$i = 0, $$pre1$pre$i = 0, $$pre1$pre$i181 = 0, $$pre2$i147 = 0, $$pre2$pre$i = 0, $$pre2$pre$i182 = 0, $$pre4$i = 0, $$pre4$i216 = 0, $$pre438 = 0, $$pre442 = 0, $$pre443 = 0, $$pre445 = 0;
 var $$pre446 = 0, $$pre447 = 0, $$pre449 = 0, $$pre450 = 0, $$pre452 = 0, $$pre453 = 0, $$pre454 = 0, $$pre455 = 0, $$pre456 = 0, $$pre457 = 0, $$pre6$i = 0, $$sink$i$i = 0, $$sink$i$i195 = 0, $$val = 0, $$val11 = 0, $$val12 = 0, $$val13 = 0, $$val14 = 0, $$val15 = 0, $$val16 = 0;
 var $$val17 = 0, $$val18 = 0, $$val19 = 0, $$val2 = 0, $$val21 = 0, $$val22 = 0, $$val24 = 0, $$val25 = 0, $$val26 = 0, $$val27 = 0, $$val28 = 0, $$val29 = 0, $$val30 = 0, $$val32 = 0, $$val33 = 0, $$val34 = 0, $$val35 = 0, $$val36 = 0, $$val37 = 0, $$val39 = 0;
 var $$val4 = 0, $$val41 = 0, $$val42 = 0, $$val43 = 0, $$val45 = 0, $$val47 = 0, $$val6 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0;
 var $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0;
 var $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0;
 var $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0;
 var $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0;
 var $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0;
 var $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0;
 var $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0;
 var $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0;
 var $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0;
 var $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0;
 var $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0;
 var $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0;
 var $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0;
 var $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0;
 var $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0, $1267 = 0, $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0;
 var $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0, $1285 = 0, $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0;
 var $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0, $1302 = 0, $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0;
 var $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0, $1320 = 0, $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0, $1332 = 0;
 var $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0, $1339 = 0, $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0, $1350 = 0;
 var $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0, $1357 = 0, $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0, $1369 = 0;
 var $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0, $1375 = 0, $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0, $1387 = 0;
 var $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0, $1393 = 0, $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0, $1404 = 0;
 var $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0, $1410 = 0, $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0, $1422 = 0;
 var $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0, $1429 = 0, $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0, $1440 = 0;
 var $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0, $1447 = 0, $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0, $1459 = 0;
 var $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0, $1465 = 0, $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0, $1477 = 0;
 var $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0, $1483 = 0, $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0, $1495 = 0;
 var $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0, $1500 = 0, $1501 = 0, $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0, $1511 = 0, $1512 = 0;
 var $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0, $1518 = 0, $1519 = 0, $152 = 0, $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0, $153 = 0, $1530 = 0;
 var $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0, $1536 = 0, $1537 = 0, $1538 = 0, $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0, $1548 = 0, $1549 = 0;
 var $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0, $1554 = 0, $1555 = 0, $1556 = 0, $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $1565 = 0, $1566 = 0, $1567 = 0;
 var $1568 = 0, $1569 = 0, $157 = 0, $1570 = 0, $1571 = 0, $1572 = 0, $1573 = 0, $1574 = 0, $1575 = 0, $1576 = 0, $1577 = 0, $1578 = 0, $1579 = 0, $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0, $1583 = 0, $1584 = 0, $1585 = 0;
 var $1586 = 0, $1587 = 0, $1588 = 0, $1589 = 0, $159 = 0, $1590 = 0, $1591 = 0, $1592 = 0, $1593 = 0, $1594 = 0, $1595 = 0, $1596 = 0, $1597 = 0, $1598 = 0, $1599 = 0, $16 = 0, $160 = 0, $1600 = 0, $1601 = 0, $1602 = 0;
 var $1603 = 0, $1604 = 0, $1605 = 0, $1606 = 0, $1607 = 0, $1608 = 0, $1609 = 0, $161 = 0, $1610 = 0, $1611 = 0, $1612 = 0, $1613 = 0, $1614 = 0, $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0, $1619 = 0, $162 = 0, $1620 = 0;
 var $1621 = 0, $1622 = 0, $1623 = 0, $1624 = 0, $1625 = 0, $1626 = 0, $1627 = 0, $1628 = 0, $1629 = 0, $163 = 0, $1630 = 0, $1631 = 0, $1632 = 0, $1633 = 0, $1634 = 0, $1635 = 0, $1636 = 0, $1637 = 0, $1638 = 0, $1639 = 0;
 var $164 = 0, $1640 = 0, $1641 = 0, $1642 = 0, $1643 = 0, $1644 = 0, $1645 = 0, $1646 = 0, $1647 = 0, $1648 = 0, $1649 = 0, $165 = 0, $1650 = 0, $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0, $1655 = 0, $1656 = 0, $1657 = 0;
 var $1658 = 0, $1659 = 0, $166 = 0, $1660 = 0, $1661 = 0, $1662 = 0, $1663 = 0, $1664 = 0, $1665 = 0, $1666 = 0, $1667 = 0, $1668 = 0, $1669 = 0, $167 = 0, $1670 = 0, $1671 = 0, $1672 = 0, $1673 = 0, $1674 = 0, $1675 = 0;
 var $1676 = 0, $1677 = 0, $1678 = 0, $1679 = 0, $168 = 0, $1680 = 0, $1681 = 0, $1682 = 0, $1683 = 0, $1684 = 0, $1685 = 0, $1686 = 0, $1687 = 0, $1688 = 0, $1689 = 0, $169 = 0, $1690 = 0, $1691 = 0, $1692 = 0, $1693 = 0;
 var $1694 = 0, $1695 = 0, $1696 = 0, $1697 = 0, $1698 = 0, $1699 = 0, $17 = 0, $170 = 0, $1700 = 0, $1701 = 0, $1702 = 0, $1703 = 0, $1704 = 0, $1705 = 0, $1706 = 0, $1707 = 0, $1708 = 0, $1709 = 0, $171 = 0, $1710 = 0;
 var $1711 = 0, $1712 = 0, $1713 = 0, $1714 = 0, $1715 = 0, $1716 = 0, $1717 = 0, $1718 = 0, $1719 = 0, $172 = 0, $1720 = 0, $1721 = 0, $1722 = 0, $1723 = 0, $1724 = 0, $1725 = 0, $1726 = 0, $1727 = 0, $1728 = 0, $1729 = 0;
 var $173 = 0, $1730 = 0, $1731 = 0, $1732 = 0, $1733 = 0, $1734 = 0, $1735 = 0, $1736 = 0, $1737 = 0, $1738 = 0, $1739 = 0, $174 = 0, $1740 = 0, $1741 = 0, $1742 = 0, $1743 = 0, $1744 = 0, $1745 = 0, $1746 = 0, $1747 = 0;
 var $1748 = 0, $1749 = 0, $175 = 0, $1750 = 0, $1751 = 0, $1752 = 0, $1753 = 0, $1754 = 0, $1755 = 0, $1756 = 0, $1757 = 0, $1758 = 0, $1759 = 0, $176 = 0, $1760 = 0, $1761 = 0, $1762 = 0, $1763 = 0, $1764 = 0, $1765 = 0;
 var $1766 = 0, $1767 = 0, $1768 = 0, $1769 = 0, $177 = 0, $1770 = 0, $1771 = 0, $1772 = 0, $1773 = 0, $1774 = 0, $1775 = 0, $1776 = 0, $1777 = 0, $1778 = 0, $1779 = 0, $178 = 0, $1780 = 0, $1781 = 0, $1782 = 0, $1783 = 0;
 var $1784 = 0, $1785 = 0, $1786 = 0, $1787 = 0, $1788 = 0, $1789 = 0, $179 = 0, $1790 = 0, $1791 = 0, $1792 = 0, $1793 = 0, $1794 = 0, $1795 = 0, $1796 = 0, $1797 = 0, $1798 = 0, $1799 = 0, $18 = 0, $180 = 0, $1800 = 0;
 var $1801 = 0, $1802 = 0, $1803 = 0, $1804 = 0, $1805 = 0, $1806 = 0, $1807 = 0, $1808 = 0, $1809 = 0, $181 = 0, $1810 = 0, $1811 = 0, $1812 = 0, $1813 = 0, $1814 = 0, $1815 = 0, $1816 = 0, $1817 = 0, $1818 = 0, $1819 = 0;
 var $182 = 0, $1820 = 0, $1821 = 0, $1822 = 0, $1823 = 0, $1824 = 0, $1825 = 0, $1826 = 0, $1827 = 0, $1828 = 0, $1829 = 0, $183 = 0, $1830 = 0, $1831 = 0, $1832 = 0, $1833 = 0, $1834 = 0, $1835 = 0, $1836 = 0, $1837 = 0;
 var $1838 = 0, $1839 = 0, $184 = 0, $1840 = 0, $1841 = 0, $1842 = 0, $1843 = 0, $1844 = 0, $1845 = 0, $1846 = 0, $1847 = 0, $1848 = 0, $1849 = 0, $185 = 0, $1850 = 0, $1851 = 0, $1852 = 0, $1853 = 0, $1854 = 0, $1855 = 0;
 var $1856 = 0, $1857 = 0, $1858 = 0, $1859 = 0, $186 = 0, $1860 = 0, $1861 = 0, $1862 = 0, $1863 = 0, $1864 = 0, $1865 = 0, $1866 = 0, $1867 = 0, $1868 = 0, $1869 = 0, $187 = 0, $1870 = 0, $1871 = 0, $1872 = 0, $1873 = 0;
 var $1874 = 0, $1875 = 0, $1876 = 0, $1877 = 0, $1878 = 0, $1879 = 0, $188 = 0, $1880 = 0, $1881 = 0, $1882 = 0, $1883 = 0, $1884 = 0, $1885 = 0, $1886 = 0, $1887 = 0, $1888 = 0, $1889 = 0, $189 = 0, $1890 = 0, $1891 = 0;
 var $1892 = 0, $1893 = 0, $1894 = 0, $1895 = 0, $1896 = 0, $1897 = 0, $1898 = 0, $1899 = 0, $19 = 0, $190 = 0, $1900 = 0, $1901 = 0, $1902 = 0, $1903 = 0, $1904 = 0, $1905 = 0, $1906 = 0, $1907 = 0, $1908 = 0, $1909 = 0;
 var $191 = 0, $1910 = 0, $1911 = 0, $1912 = 0, $1913 = 0, $1914 = 0, $1915 = 0, $1916 = 0, $1917 = 0, $1918 = 0, $1919 = 0, $192 = 0, $1920 = 0, $1921 = 0, $1922 = 0, $1923 = 0, $1924 = 0, $1925 = 0, $1926 = 0, $1927 = 0;
 var $1928 = 0, $1929 = 0, $193 = 0, $1930 = 0, $1931 = 0, $1932 = 0, $1933 = 0, $1934 = 0, $1935 = 0, $1936 = 0, $1937 = 0, $1938 = 0, $1939 = 0, $194 = 0, $1940 = 0, $1941 = 0, $1942 = 0, $1943 = 0, $1944 = 0, $1945 = 0;
 var $1946 = 0, $1947 = 0, $1948 = 0, $1949 = 0, $195 = 0, $1950 = 0, $1951 = 0, $1952 = 0, $1953 = 0, $1954 = 0, $1955 = 0, $1956 = 0, $1957 = 0, $1958 = 0, $1959 = 0, $196 = 0, $1960 = 0, $1961 = 0, $1962 = 0, $1963 = 0;
 var $1964 = 0, $1965 = 0, $1966 = 0, $1967 = 0, $1968 = 0, $1969 = 0, $197 = 0, $1970 = 0, $1971 = 0, $1972 = 0, $1973 = 0, $1974 = 0, $1975 = 0, $1976 = 0, $1977 = 0, $1978 = 0, $1979 = 0, $198 = 0, $1980 = 0, $1981 = 0;
 var $1982 = 0, $1983 = 0, $1984 = 0, $1985 = 0, $1986 = 0, $1987 = 0, $1988 = 0, $1989 = 0, $199 = 0, $1990 = 0, $1991 = 0, $1992 = 0, $1993 = 0, $1994 = 0, $1995 = 0, $1996 = 0, $1997 = 0, $1998 = 0, $1999 = 0, $2 = 0;
 var $20 = 0, $200 = 0, $2000 = 0, $2001 = 0, $2002 = 0, $2003 = 0, $2004 = 0, $2005 = 0, $2006 = 0, $2007 = 0, $2008 = 0, $2009 = 0, $201 = 0, $2010 = 0, $2011 = 0, $2012 = 0, $2013 = 0, $2014 = 0, $2015 = 0, $2016 = 0;
 var $2017 = 0, $2018 = 0, $2019 = 0, $202 = 0, $2020 = 0, $2021 = 0, $2022 = 0, $2023 = 0, $2024 = 0, $2025 = 0, $2026 = 0, $2027 = 0, $2028 = 0, $2029 = 0, $203 = 0, $2030 = 0, $2031 = 0, $2032 = 0, $2033 = 0, $2034 = 0;
 var $2035 = 0, $2036 = 0, $2037 = 0, $2038 = 0, $2039 = 0, $204 = 0, $2040 = 0, $2041 = 0, $2042 = 0, $2043 = 0, $2044 = 0, $2045 = 0, $2046 = 0, $2047 = 0, $2048 = 0, $2049 = 0, $205 = 0, $2050 = 0, $2051 = 0, $2052 = 0;
 var $2053 = 0, $2054 = 0, $2055 = 0, $2056 = 0, $2057 = 0, $2058 = 0, $2059 = 0, $206 = 0, $2060 = 0, $2061 = 0, $2062 = 0, $2063 = 0, $2064 = 0, $2065 = 0, $2066 = 0, $2067 = 0, $2068 = 0, $2069 = 0, $207 = 0, $2070 = 0;
 var $2071 = 0, $2072 = 0, $2073 = 0, $2074 = 0, $2075 = 0, $2076 = 0, $2077 = 0, $2078 = 0, $2079 = 0, $208 = 0, $2080 = 0, $2081 = 0, $2082 = 0, $2083 = 0, $2084 = 0, $2085 = 0, $2086 = 0, $2087 = 0, $2088 = 0, $2089 = 0;
 var $209 = 0, $2090 = 0, $2091 = 0, $2092 = 0, $2093 = 0, $2094 = 0, $2095 = 0, $2096 = 0, $2097 = 0, $2098 = 0, $2099 = 0, $21 = 0, $210 = 0, $2100 = 0, $2101 = 0, $2102 = 0, $2103 = 0, $2104 = 0, $2105 = 0, $2106 = 0;
 var $2107 = 0, $2108 = 0, $2109 = 0, $211 = 0, $2110 = 0, $2111 = 0, $2112 = 0, $2113 = 0, $2114 = 0, $2115 = 0, $2116 = 0, $2117 = 0, $2118 = 0, $2119 = 0, $212 = 0, $2120 = 0, $2121 = 0, $2122 = 0, $2123 = 0, $2124 = 0;
 var $2125 = 0, $2126 = 0, $2127 = 0, $2128 = 0, $2129 = 0, $213 = 0, $2130 = 0, $2131 = 0, $2132 = 0, $2133 = 0, $2134 = 0, $2135 = 0, $2136 = 0, $2137 = 0, $2138 = 0, $2139 = 0, $214 = 0, $2140 = 0, $2141 = 0, $2142 = 0;
 var $2143 = 0, $2144 = 0, $2145 = 0, $2146 = 0, $2147 = 0, $2148 = 0, $2149 = 0, $215 = 0, $2150 = 0, $2151 = 0, $2152 = 0, $2153 = 0, $2154 = 0, $2155 = 0, $2156 = 0, $2157 = 0, $2158 = 0, $2159 = 0, $216 = 0, $2160 = 0;
 var $2161 = 0, $2162 = 0, $2163 = 0, $2164 = 0, $2165 = 0, $2166 = 0, $2167 = 0, $2168 = 0, $2169 = 0, $217 = 0, $2170 = 0, $2171 = 0, $2172 = 0, $2173 = 0, $2174 = 0, $2175 = 0, $2176 = 0, $2177 = 0, $2178 = 0, $2179 = 0;
 var $218 = 0, $2180 = 0, $2181 = 0, $2182 = 0, $2183 = 0, $2184 = 0, $2185 = 0, $2186 = 0, $2187 = 0, $2188 = 0, $2189 = 0, $219 = 0, $2190 = 0, $2191 = 0, $2192 = 0, $2193 = 0, $2194 = 0, $2195 = 0, $2196 = 0, $2197 = 0;
 var $2198 = 0, $2199 = 0, $22 = 0, $220 = 0, $2200 = 0, $2201 = 0, $2202 = 0, $2203 = 0, $2204 = 0, $2205 = 0, $2206 = 0, $2207 = 0, $2208 = 0, $2209 = 0, $221 = 0, $2210 = 0, $2211 = 0, $2212 = 0, $2213 = 0, $2214 = 0;
 var $2215 = 0, $2216 = 0, $2217 = 0, $2218 = 0, $2219 = 0, $222 = 0, $2220 = 0, $2221 = 0, $2222 = 0, $2223 = 0, $2224 = 0, $2225 = 0, $2226 = 0, $2227 = 0, $2228 = 0, $2229 = 0, $223 = 0, $2230 = 0, $2231 = 0, $2232 = 0;
 var $2233 = 0, $2234 = 0, $2235 = 0, $2236 = 0, $2237 = 0, $2238 = 0, $2239 = 0, $224 = 0, $2240 = 0, $2241 = 0, $2242 = 0, $2243 = 0, $2244 = 0, $2245 = 0, $2246 = 0, $2247 = 0, $2248 = 0, $2249 = 0, $225 = 0, $2250 = 0;
 var $2251 = 0, $2252 = 0, $2253 = 0, $2254 = 0, $2255 = 0, $2256 = 0, $2257 = 0, $2258 = 0, $2259 = 0, $226 = 0, $2260 = 0, $2261 = 0, $2262 = 0, $2263 = 0, $2264 = 0, $2265 = 0, $2266 = 0, $2267 = 0, $2268 = 0, $2269 = 0;
 var $227 = 0, $2270 = 0, $2271 = 0, $2272 = 0, $2273 = 0, $2274 = 0, $2275 = 0, $2276 = 0, $2277 = 0, $2278 = 0, $2279 = 0, $228 = 0, $2280 = 0, $2281 = 0, $2282 = 0, $2283 = 0, $2284 = 0, $2285 = 0, $2286 = 0, $2287 = 0;
 var $2288 = 0, $2289 = 0, $229 = 0, $2290 = 0, $2291 = 0, $2292 = 0, $2293 = 0, $2294 = 0, $2295 = 0, $2296 = 0, $2297 = 0, $2298 = 0, $2299 = 0, $23 = 0, $230 = 0, $2300 = 0, $2301 = 0, $2302 = 0, $2303 = 0, $2304 = 0;
 var $2305 = 0, $2306 = 0, $2307 = 0, $2308 = 0, $2309 = 0, $231 = 0, $2310 = 0, $2311 = 0, $2312 = 0, $2313 = 0, $2314 = 0, $2315 = 0, $2316 = 0, $2317 = 0, $2318 = 0, $2319 = 0, $232 = 0, $2320 = 0, $2321 = 0, $2322 = 0;
 var $2323 = 0, $2324 = 0, $2325 = 0, $2326 = 0, $2327 = 0, $2328 = 0, $2329 = 0, $233 = 0, $2330 = 0, $2331 = 0, $2332 = 0, $2333 = 0, $2334 = 0, $2335 = 0, $2336 = 0, $2337 = 0, $2338 = 0, $2339 = 0, $234 = 0, $2340 = 0;
 var $2341 = 0, $2342 = 0, $2343 = 0, $2344 = 0, $2345 = 0, $2346 = 0, $2347 = 0, $2348 = 0, $2349 = 0, $235 = 0, $2350 = 0, $2351 = 0, $2352 = 0, $2353 = 0, $2354 = 0, $2355 = 0, $2356 = 0, $2357 = 0, $2358 = 0, $2359 = 0;
 var $236 = 0, $2360 = 0, $2361 = 0, $2362 = 0, $2363 = 0, $2364 = 0, $2365 = 0, $2366 = 0, $2367 = 0, $2368 = 0, $2369 = 0, $237 = 0, $2370 = 0, $2371 = 0, $2372 = 0, $2373 = 0, $2374 = 0, $2375 = 0, $2376 = 0, $2377 = 0;
 var $2378 = 0, $2379 = 0, $238 = 0, $2380 = 0, $2381 = 0, $2382 = 0, $2383 = 0, $2384 = 0, $2385 = 0, $2386 = 0, $2387 = 0, $2388 = 0, $2389 = 0, $239 = 0, $2390 = 0, $2391 = 0, $2392 = 0, $2393 = 0, $2394 = 0, $2395 = 0;
 var $2396 = 0, $2397 = 0, $2398 = 0, $2399 = 0, $24 = 0, $240 = 0, $2400 = 0, $2401 = 0, $2402 = 0, $2403 = 0, $2404 = 0, $2405 = 0, $2406 = 0, $2407 = 0, $2408 = 0, $2409 = 0, $241 = 0, $2410 = 0, $2411 = 0, $2412 = 0;
 var $2413 = 0, $2414 = 0, $2415 = 0, $2416 = 0, $2417 = 0, $2418 = 0, $2419 = 0, $242 = 0, $2420 = 0, $2421 = 0, $2422 = 0, $2423 = 0, $2424 = 0, $2425 = 0, $2426 = 0, $2427 = 0, $2428 = 0, $2429 = 0, $243 = 0, $2430 = 0;
 var $2431 = 0, $2432 = 0, $2433 = 0, $2434 = 0, $2435 = 0, $2436 = 0, $2437 = 0, $2438 = 0, $2439 = 0, $244 = 0, $2440 = 0, $2441 = 0, $2442 = 0, $2443 = 0, $2444 = 0, $2445 = 0, $2446 = 0, $2447 = 0, $2448 = 0, $2449 = 0;
 var $245 = 0, $2450 = 0, $2451 = 0, $2452 = 0, $2453 = 0, $2454 = 0, $2455 = 0, $2456 = 0, $2457 = 0, $2458 = 0, $2459 = 0, $246 = 0, $2460 = 0, $2461 = 0, $2462 = 0, $2463 = 0, $2464 = 0, $2465 = 0, $2466 = 0, $2467 = 0;
 var $2468 = 0, $2469 = 0, $247 = 0, $2470 = 0, $2471 = 0, $2472 = 0, $2473 = 0, $2474 = 0, $2475 = 0, $2476 = 0, $2477 = 0, $2478 = 0, $2479 = 0, $248 = 0, $2480 = 0, $2481 = 0, $2482 = 0, $2483 = 0, $2484 = 0, $2485 = 0;
 var $2486 = 0, $2487 = 0, $2488 = 0, $2489 = 0, $249 = 0, $2490 = 0, $2491 = 0, $2492 = 0, $2493 = 0, $2494 = 0, $2495 = 0, $2496 = 0, $2497 = 0, $2498 = 0, $2499 = 0, $25 = 0, $250 = 0, $2500 = 0, $2501 = 0, $2502 = 0;
 var $2503 = 0, $2504 = 0, $2505 = 0, $2506 = 0, $2507 = 0, $2508 = 0, $2509 = 0, $251 = 0, $2510 = 0, $2511 = 0, $2512 = 0, $2513 = 0, $2514 = 0, $2515 = 0, $2516 = 0, $2517 = 0, $2518 = 0, $2519 = 0, $252 = 0, $2520 = 0;
 var $2521 = 0, $2522 = 0, $2523 = 0, $2524 = 0, $2525 = 0, $2526 = 0, $2527 = 0, $2528 = 0, $2529 = 0, $253 = 0, $2530 = 0, $2531 = 0, $2532 = 0, $2533 = 0, $2534 = 0, $2535 = 0, $2536 = 0, $2537 = 0, $2538 = 0, $2539 = 0;
 var $254 = 0, $2540 = 0, $2541 = 0, $2542 = 0, $2543 = 0, $2544 = 0, $2545 = 0, $2546 = 0, $2547 = 0, $2548 = 0, $2549 = 0, $255 = 0, $2550 = 0, $2551 = 0, $2552 = 0, $2553 = 0, $2554 = 0, $2555 = 0, $2556 = 0, $2557 = 0;
 var $2558 = 0, $2559 = 0, $256 = 0, $2560 = 0, $2561 = 0, $2562 = 0, $2563 = 0, $2564 = 0, $2565 = 0, $2566 = 0, $2567 = 0, $2568 = 0, $2569 = 0, $257 = 0, $2570 = 0, $2571 = 0, $2572 = 0, $2573 = 0, $2574 = 0, $2575 = 0;
 var $2576 = 0, $2577 = 0, $2578 = 0, $2579 = 0, $258 = 0, $2580 = 0, $2581 = 0, $2582 = 0, $2583 = 0, $2584 = 0, $2585 = 0, $2586 = 0, $2587 = 0, $2588 = 0, $2589 = 0, $259 = 0, $2590 = 0, $2591 = 0, $2592 = 0, $2593 = 0;
 var $2594 = 0, $2595 = 0, $2596 = 0, $2597 = 0, $2598 = 0, $2599 = 0, $26 = 0, $260 = 0, $2600 = 0, $2601 = 0, $2602 = 0, $2603 = 0, $2604 = 0, $2605 = 0, $2606 = 0, $2607 = 0, $2608 = 0, $2609 = 0, $261 = 0, $2610 = 0;
 var $2611 = 0, $2612 = 0, $2613 = 0, $2614 = 0, $2615 = 0, $2616 = 0, $2617 = 0, $2618 = 0, $2619 = 0, $262 = 0, $2620 = 0, $2621 = 0, $2622 = 0, $2623 = 0, $2624 = 0, $2625 = 0, $2626 = 0, $2627 = 0, $2628 = 0, $2629 = 0;
 var $263 = 0, $2630 = 0, $2631 = 0, $2632 = 0, $2633 = 0, $2634 = 0, $2635 = 0, $2636 = 0, $2637 = 0, $2638 = 0, $2639 = 0, $264 = 0, $2640 = 0, $2641 = 0, $2642 = 0, $2643 = 0, $2644 = 0, $2645 = 0, $2646 = 0, $2647 = 0;
 var $2648 = 0, $2649 = 0, $265 = 0, $2650 = 0, $2651 = 0, $2652 = 0, $2653 = 0, $2654 = 0, $2655 = 0, $2656 = 0, $2657 = 0, $2658 = 0, $2659 = 0, $266 = 0, $2660 = 0, $2661 = 0, $2662 = 0, $2663 = 0, $2664 = 0, $2665 = 0;
 var $2666 = 0, $2667 = 0, $2668 = 0, $2669 = 0, $267 = 0, $2670 = 0, $2671 = 0, $2672 = 0, $2673 = 0, $2674 = 0, $2675 = 0, $2676 = 0, $2677 = 0, $2678 = 0, $2679 = 0, $268 = 0, $2680 = 0, $2681 = 0, $2682 = 0, $2683 = 0;
 var $2684 = 0, $2685 = 0, $2686 = 0, $2687 = 0, $2688 = 0, $2689 = 0, $269 = 0, $2690 = 0, $2691 = 0, $2692 = 0, $2693 = 0, $2694 = 0, $2695 = 0, $2696 = 0, $2697 = 0, $2698 = 0, $2699 = 0, $27 = 0, $270 = 0, $2700 = 0;
 var $2701 = 0, $2702 = 0, $2703 = 0, $2704 = 0, $2705 = 0, $2706 = 0, $2707 = 0, $2708 = 0, $2709 = 0, $271 = 0, $2710 = 0, $2711 = 0, $2712 = 0, $2713 = 0, $2714 = 0, $2715 = 0, $2716 = 0, $2717 = 0, $2718 = 0, $2719 = 0;
 var $272 = 0, $2720 = 0, $2721 = 0, $2722 = 0, $2723 = 0, $2724 = 0, $2725 = 0, $2726 = 0, $2727 = 0, $2728 = 0, $2729 = 0, $273 = 0, $2730 = 0, $2731 = 0, $2732 = 0, $2733 = 0, $2734 = 0, $2735 = 0, $2736 = 0, $2737 = 0;
 var $2738 = 0, $2739 = 0, $274 = 0, $2740 = 0, $2741 = 0, $2742 = 0, $2743 = 0, $2744 = 0, $2745 = 0, $2746 = 0, $2747 = 0, $2748 = 0, $2749 = 0, $275 = 0, $2750 = 0, $2751 = 0, $2752 = 0, $2753 = 0, $2754 = 0, $2755 = 0;
 var $2756 = 0, $2757 = 0, $2758 = 0, $2759 = 0, $276 = 0, $2760 = 0, $2761 = 0, $2762 = 0, $2763 = 0, $2764 = 0, $2765 = 0, $2766 = 0, $2767 = 0, $2768 = 0, $2769 = 0, $277 = 0, $2770 = 0, $2771 = 0, $2772 = 0, $2773 = 0;
 var $2774 = 0, $2775 = 0, $2776 = 0, $2777 = 0, $2778 = 0, $2779 = 0, $278 = 0, $2780 = 0, $2781 = 0, $2782 = 0, $2783 = 0, $2784 = 0, $2785 = 0, $2786 = 0, $2787 = 0, $2788 = 0, $2789 = 0, $279 = 0, $2790 = 0, $2791 = 0;
 var $2792 = 0, $2793 = 0, $2794 = 0, $2795 = 0, $2796 = 0, $2797 = 0, $2798 = 0, $2799 = 0, $28 = 0, $280 = 0, $2800 = 0, $2801 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0;
 var $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0;
 var $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0;
 var $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0;
 var $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0;
 var $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0;
 var $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0;
 var $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0;
 var $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0;
 var $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0;
 var $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0;
 var $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0;
 var $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0;
 var $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0;
 var $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0;
 var $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0;
 var $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0;
 var $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0;
 var $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0;
 var $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0;
 var $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0;
 var $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0;
 var $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0;
 var $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0;
 var $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0;
 var $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0;
 var $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0;
 var $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0;
 var $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0;
 var $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0;
 var $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0;
 var $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0;
 var $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0;
 var $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0;
 var $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0;
 var $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0;
 var $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0;
 var $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0;
 var $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0;
 var $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0;
 var $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $A$1$i = 0, $A$1$i$i = 0, $A$1$i$i59 = 0, $A$3$i = 0, $B$0$i = 0, $B$0$i$i = 0, $B$0$i$i58 = 0, $B$1$i = 0, $C$0$i = 0, $Current_Ppem$Current_Ppem_Stretched = 0, $D$1$i = 0;
 var $I$01$i = 0, $I$01$i235 = 0, $K$0$$i = 0, $K$0$i = 0, $K$01$i = 0, $K$01$i171 = 0, $K$01$i50 = 0, $K$1$$i = 0, $K$1$i = 0, $K$2$i = 0, $Move_CVT$Move_CVT_Stretched = 0, $R$0$i = 0, $Read_CVT$Read_CVT_Stretched = 0, $V$i = 0, $V$idx$val14$i = 0, $V$idx$val14$in$i = 0, $V$idx2$val17$i = 0, $V$idx2$val17$in$i = 0, $Write_CVT$Write_CVT_Stretched = 0, $aRange$off$i = 0;
 var $aRange$off$i$i = 0, $aRange$off$i$i200 = 0, $aRange$off$i$i206 = 0, $aRange$off$i$i239 = 0, $aRange$off$i$i72 = 0, $aRange$off$i$i78 = 0, $aRange$off$i$i85 = 0, $aRange$off$i$i92 = 0, $args$val$i = 0, $args$val$i196 = 0, $brmerge$i = 0, $contour$0$i = 0, $cur_range$0$ph$i = 0, $cur_touched$0$ph$i = 0, $cur_touched$0$ph$i$lcssa = 0, $cvt_dist$0$i = 0, $cvt_dist$1$i = 0, $cvt_dist$2$i = 0, $cvt_dist$3$i = 0, $def$0$lcssa$i = 0;
 var $def$0$lcssa$i213 = 0, $def$0$lcssa$i82 = 0, $def$01$i = 0, $def$01$i212 = 0, $def$01$i80 = 0, $def$02$i = 0, $def$02$i$lcssa = 0, $def$02$i199 = 0, $def$02$i199$lcssa = 0, $def$02$i205 = 0, $def$02$i205$lcssa = 0, $def$02$i238 = 0, $def$02$i238$lcssa = 0, $def$02$i71 = 0, $def$02$i71$lcssa = 0, $def$0371 = 0, $def$0371$lcssa = 0, $def$1$i = 0, $def$1$i84 = 0, $distance$0$i = 0;
 var $distance$0$i102 = 0, $distance$0$i106 = 0, $distance$1$i = 0, $distance$1$i105 = 0, $distance$1$i165 = 0, $dx$i = 0, $dx$i114 = 0, $dx$i123 = 0, $dy$i = 0, $dy$i115 = 0, $dy$i124 = 0, $exc$idx$val = 0, $exc$idx10$val = 0, $exc$idx20$val = 0, $exc$idx8$val = 0, $exc$idx9$val = 0, $exitcond = 0, $exitcond$i = 0, $exitcond$i$i = 0, $exitcond431 = 0;
 var $exitcond433 = 0, $exitcond437 = 0, $exitcond551 = 0, $i$0362 = 0, $i$04$i$i = 0, $i$1$i$i = 0, $i$12$i$i = 0, $i$13$i$i = 0, $indvars$iv$i = 0, $indvars$iv$i128 = 0, $indvars$iv$next$i = 0, $indvars$iv$next$i129 = 0, $indvars$iv$next5$i = 0, $indvars$iv$next5$i$us = 0, $indvars$iv$next5$i$us380 = 0, $indvars$iv$next5$i552 = 0, $indvars$iv4$i$us = 0, $indvars$iv4$i$us378 = 0, $ins_counter$0 = 0, $ins_counter$1 = 0;
 var $k$02$i = 0, $lftr$wideiv = 0, $lftr$wideiv$i = 0, $lftr$wideiv430 = 0, $lftr$wideiv432 = 0, $lftr$wideiv436 = 0, $limit$0$i = 0, $limit$0$i125 = 0, $mask$0$i = 0, $mask$1$i = 0, $nIfs$0$ph$i = 0, $nIfs$0$ph$ph$i$be = 0, $nIfs$0$ph$ph$i368 = 0, $nIfs$1$i = 0, $new_dist$0$i = 0, $old_range$0$i = 0, $old_range$1$ph$i = 0, $or$cond$i = 0, $or$cond$i$i = 0, $or$cond$i$i55 = 0;
 var $or$cond$i118 = 0, $or$cond$i140 = 0, $or$cond$i179 = 0, $or$cond$i210 = 0, $or$cond$i221 = 0, $or$cond$i228 = 0, $or$cond$i232 = 0, $or$cond$i99 = 0, $or$cond250 = 0, $or$cond251 = 0, $or$cond3$i = 0, $org_dist$0$i = 0, $org_dist$1$i = 0, $org_dist$2$i = 0, $orus_base$0$i = 0, $point$0$i = 0, $point$17$i = 0, $point$17$i$lcssa = 0, $point$2$i = 0, $point$2$i$lcssa = 0;
 var $point$2$i$lcssa617 = 0, $point$2$in$i = 0, $point$2$in$i$lcssa616 = 0, $point$3$i = 0, $rec$0$lcssa$i = 0, $rec$01$i = 0, $refp$i = 0, $refp$i113 = 0, $refp$i122 = 0, $sext$i = 0, $sext$i116 = 0, $sext$i172 = 0, $sext$i62 = 0, $sext$i65 = 0, $sext1$i = 0, $sext1$i66 = 0, $start$0$i = 0, $storemerge$i = 0, $storemerge$i177 = 0, $storemerge$i178 = 0;
 var $val$0$i$i = 0, $zp$i = 0, $zp$i112 = 0, $zp$i121 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 160|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $zp$i121 = sp + 112|0;
 $refp$i122 = sp + 152|0;
 $dx$i123 = sp + 108|0;
 $dy$i124 = sp + 104|0;
 $zp$i112 = sp + 68|0;
 $refp$i113 = sp + 150|0;
 $dx$i114 = sp + 64|0;
 $dy$i115 = sp + 60|0;
 $zp$i = sp + 24|0;
 $refp$i = sp + 148|0;
 $dx$i = sp + 20|0;
 $dy$i = sp + 16|0;
 $V$i = sp;
 $0 = ((($exc)) + 256|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($exc)) + 216|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2&65535;
 $4 = $2 >>> 16;
 $5 = $4&65535;
 $6 = ($3<<16>>16)==($5<<16>>16);
 $7 = ((($exc)) + 592|0);
 $Current_Ppem$Current_Ppem_Stretched = $6 ? 4410 : 4411;
 $Read_CVT$Read_CVT_Stretched = $6 ? 4412 : 4413;
 $Write_CVT$Write_CVT_Stretched = $6 ? 4414 : 4415;
 $Move_CVT$Move_CVT_Stretched = $6 ? 4416 : 4417;
 HEAP32[$7>>2] = $Current_Ppem$Current_Ppem_Stretched;
 $8 = ((($exc)) + 596|0);
 HEAP32[$8>>2] = $Read_CVT$Read_CVT_Stretched;
 $9 = ((($exc)) + 600|0);
 HEAP32[$9>>2] = $Write_CVT$Write_CVT_Stretched;
 $10 = ((($exc)) + 604|0);
 HEAP32[$10>>2] = $Move_CVT$Move_CVT_Stretched;
 _Compute_Funcs($exc);
 $11 = ((($exc)) + 312|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = $12 & 255;
 switch ($13|0) {
 case 5:  {
  $14 = ((($exc)) + 568|0);
  HEAP32[$14>>2] = 4418;
  $$pre$phiZ2D = $14;
  break;
 }
 case 1:  {
  $15 = ((($exc)) + 568|0);
  HEAP32[$15>>2] = 4419;
  $$pre$phiZ2D = $15;
  break;
 }
 case 4:  {
  $16 = ((($exc)) + 568|0);
  HEAP32[$16>>2] = 4420;
  $$pre$phiZ2D = $16;
  break;
 }
 case 3:  {
  $17 = ((($exc)) + 568|0);
  HEAP32[$17>>2] = 4421;
  $$pre$phiZ2D = $17;
  break;
 }
 case 0:  {
  $18 = ((($exc)) + 568|0);
  HEAP32[$18>>2] = 4422;
  $$pre$phiZ2D = $18;
  break;
 }
 case 2:  {
  $19 = ((($exc)) + 568|0);
  HEAP32[$19>>2] = 4423;
  $$pre$phiZ2D = $19;
  break;
 }
 case 6:  {
  $20 = ((($exc)) + 568|0);
  HEAP32[$20>>2] = 4424;
  $$pre$phiZ2D = $20;
  break;
 }
 case 7:  {
  $21 = ((($exc)) + 568|0);
  HEAP32[$21>>2] = 4425;
  $$pre$phiZ2D = $21;
  break;
 }
 default: {
  $$pre457 = ((($exc)) + 568|0);
  $$pre$phiZ2D = $$pre457;
 }
 }
 $22 = ((($exc)) + 360|0);
 $23 = ((($exc)) + 356|0);
 $24 = ((($exc)) + 368|0);
 $25 = ((($exc)) + 372|0);
 $26 = ((($exc)) + 364|0);
 $27 = ((($exc)) + 16|0);
 $28 = ((($exc)) + 28|0);
 $29 = ((($exc)) + 561|0);
 $30 = ((($exc)) + 24|0);
 $31 = ((($exc)) + 32|0);
 $32 = ((($exc)) + 20|0);
 $33 = ((($exc)) + 376|0);
 $34 = ((($exc)) + 12|0);
 $35 = ((($exc)) + 308|0);
 $36 = ((($exc)) + 320|0);
 $37 = ((($exc)) + 72|0);
 $38 = ((($exc)) + 80|0);
 $39 = ((($exc)) + 380|0);
 $40 = ((($exc)) + 284|0);
 $41 = ((($exc)) + 44|0);
 $42 = ((($exc)) + 328|0);
 $43 = ((($exc)) + 324|0);
 $44 = ((($exc)) + 346|0);
 $45 = ((($exc)) + 48|0);
 $46 = ((($exc)) + 298|0);
 $47 = ((($exc)) + 84|0);
 $48 = ((($exc)) + 300|0);
 $49 = ((($exc)) + 88|0);
 $50 = ((($exc)) + 576|0);
 $51 = ((($exc)) + 572|0);
 $52 = ((($exc)) + 52|0);
 $53 = ((($exc)) + 316|0);
 $54 = ((($exc)) + 584|0);
 $55 = ((($exc)) + 344|0);
 $56 = ((($exc)) + 286|0);
 $57 = ((($exc)) + 288|0);
 $58 = ((($exc)) + 92|0);
 $59 = ((($exc)) + 56|0);
 $60 = ((($exc)) + 220|0);
 $61 = ((($exc)) + 224|0);
 $62 = ((($exc)) + 416|0);
 $63 = ((($exc)) + 408|0);
 $64 = ((($exc)) + 428|0);
 $65 = ((($exc)) + 432|0);
 $66 = ((($exc)) + 436|0);
 $67 = ((($exc)) + 352|0);
 $68 = ((($exc)) + 488|0);
 $69 = ((($exc)) + 294|0);
 $70 = ((($exc)) + 296|0);
 $71 = ((($exc)) + 290|0);
 $72 = ((($exc)) + 292|0);
 $73 = ((($exc)) + 294|0);
 $74 = ((($exc)) + 116|0);
 $75 = ((($exc)) + 124|0);
 $76 = ((($exc)) + 290|0);
 $77 = ((($exc)) + 298|0);
 $78 = ((($exc)) + 132|0);
 $79 = ((($exc)) + 284|0);
 $80 = ((($exc)) + 36|0);
 $81 = ((($exc)) + 180|0);
 $82 = ((($exc)) + 144|0);
 $83 = ((($exc)) + 108|0);
 $84 = ((($exc)) + 348|0);
 $85 = ((($exc)) + 304|0);
 $86 = ((($exc)) + 260|0);
 $87 = ((($exc)) + 60|0);
 $88 = ((($exc)) + 420|0);
 $89 = ((($exc)) + 404|0);
 $90 = ((($exc)) + 396|0);
 $91 = ((($exc)) + 400|0);
 $92 = ((($exc)) + 264|0);
 $93 = ((($exc)) + 154|0);
 $94 = ((($exc)) + 156|0);
 $95 = ((($exc)) + 160|0);
 $96 = ((($V$i)) + 4|0);
 $97 = ((($exc)) + 164|0);
 $98 = ((($V$i)) + 8|0);
 $99 = ((($exc)) + 152|0);
 $100 = ((($V$i)) + 12|0);
 $101 = ((($exc)) + 172|0);
 $102 = ((($exc)) + 176|0);
 $103 = ((($exc)) + 168|0);
 $104 = ((($zp$i112)) + 16|0);
 $105 = ((($exc)) + 136|0);
 $106 = ((($exc)) + 140|0);
 $107 = ((($exc)) + 118|0);
 $108 = ((($zp$i121)) + 16|0);
 $109 = ((($exc)) + 120|0);
 $110 = ((($exc)) + 128|0);
 $111 = ((($exc)) + 588|0);
 $112 = ((($exc)) + 468|0);
 $113 = ((($exc)) + 472|0);
 $114 = ((($exc)) + 332|0);
 $115 = ((($exc)) + 334|0);
 $116 = ((($exc)) + 384|0);
 $117 = ((($exc)) + 476|0);
 $118 = ((($exc)) + 480|0);
 $119 = ((($exc)) + 484|0);
 $120 = ((($exc)) + 282|0);
 $121 = ((($exc)) + 337|0);
 $122 = ((($exc)) + 281|0);
 $123 = ((($exc)) + 252|0);
 $124 = ((($exc)) + 608|0);
 $125 = ((($exc)) + 412|0);
 $126 = ((($exc)) + 424|0);
 $127 = ((($exc)) + 340|0);
 $128 = ((($exc)) + 336|0);
 $$pre = HEAP32[$22>>2]|0;
 $131 = $$pre;$ins_counter$0 = 0;
 L12: while(1) {
  $129 = HEAP32[$23>>2]|0;
  $130 = (($129) + ($131)|0);
  $132 = HEAP8[$130>>0]|0;
  HEAP8[$24>>0] = $132;
  $133 = $132&255;
  $134 = (593353 + ($133)|0);
  $135 = HEAP8[$134>>0]|0;
  $136 = $135 << 24 >> 24;
  HEAP32[$25>>2] = $136;
  $137 = $133 | 1;
  $138 = ($137|0)==(65);
  if ($138) {
   $139 = (($131) + 1)|0;
   $140 = HEAP32[$26>>2]|0;
   $141 = ($139|0)<($140|0);
   if (!($141)) {
    label = 844;
    break;
   }
   $142 = (($129) + ($139)|0);
   $143 = HEAP8[$142>>0]|0;
   $144 = $143&255;
   $145 = Math_imul($144, $136)|0;
   $146 = (2 - ($145))|0;
   HEAP32[$25>>2] = $146;
   $$pre$pre$i = $146;$149 = $140;
  } else {
   $$pre438 = HEAP32[$26>>2]|0;
   $$pre$pre$i = $136;$149 = $$pre438;
  }
  $147 = (($$pre$pre$i) + ($131))|0;
  $148 = ($147|0)>($149|0);
  if ($148) {
   label = 844;
   break;
  }
  $150 = HEAP32[$27>>2]|0;
  $151 = (593609 + ($133)|0);
  $152 = HEAP8[$151>>0]|0;
  $153 = $152&255;
  $154 = $153 >>> 4;
  $155 = (($150) - ($154))|0;
  HEAP32[$28>>2] = $155;
  $156 = ($155|0)<(0);
  if ($156) {
   $157 = HEAP8[$29>>0]|0;
   $158 = ($157<<24>>24)==(0);
   if (!($158)) {
    label = 21;
    break;
   }
   $159 = ($152&255)>(15);
   if ($159) {
    $160 = HEAP32[$30>>2]|0;
    $162 = 0;$i$0362 = 0;
    while(1) {
     $161 = (($160) + ($162<<2)|0);
     HEAP32[$161>>2] = 0;
     $163 = (($i$0362) + 1)<<16>>16;
     $164 = $163&65535;
     $165 = ($164>>>0)<($154>>>0);
     if ($165) {
      $162 = $164;$i$0362 = $163;
     } else {
      break;
     }
    }
   }
   HEAP32[$28>>2] = 0;
   $168 = 0;
  } else {
   $168 = $155;
  }
  $166 = $153 & 15;
  $167 = (($166) + ($168))|0;
  HEAP32[$31>>2] = $167;
  $169 = HEAP32[$32>>2]|0;
  $170 = ($167|0)>($169|0);
  if ($170) {
   label = 25;
   break;
  }
  HEAP8[$33>>0] = 1;
  HEAP32[$34>>2] = 0;
  $171 = HEAP32[$30>>2]|0;
  $172 = (($171) + ($168<<2)|0);
  L30: do {
   switch ($133|0) {
   case 79:  {
    label = 470;
    break L12;
    break;
   }
   case 5: case 4: case 3: case 2: case 1: case 0:  {
    $173 = $133 << 14;
    $174 = $173 & 16384;
    $175 = $174&65535;
    $176 = $174 ^ 16384;
    $177 = $176&65535;
    $178 = ($132&255)<(4);
    if ($178) {
     HEAP16[$69>>1] = $175;
     HEAP16[$70>>1] = $177;
     HEAP16[$71>>1] = $175;
     HEAP16[$72>>1] = $177;
    }
    $179 = $133 & 2;
    $180 = ($179|0)==(0);
    if ($180) {
     HEAP16[$46>>1] = $175;
     HEAP16[$48>>1] = $177;
    }
    _Compute_Funcs($exc);
    label = 821;
    break;
   }
   case 7: case 6:  {
    $$val = HEAP32[$172>>2]|0;
    $$idx = ((($172)) + 4|0);
    $$idx$val = HEAP32[$$idx>>2]|0;
    $181 = $$idx$val&65535;
    $182 = $$idx$val & 65535;
    $183 = HEAP16[$74>>1]|0;
    $184 = ($183&65535)>($181&65535);
    if ($184) {
     $185 = $$val&65535;
     $186 = HEAP16[$38>>1]|0;
     $187 = ($186&65535)>($185&65535);
     if ($187) {
      $190 = $$val & 65535;
      $191 = HEAP32[$49>>2]|0;
      $192 = HEAP32[$75>>2]|0;
      $193 = (($191) + ($190<<3)|0);
      $194 = HEAP32[$193>>2]|0;
      $195 = (($192) + ($182<<3)|0);
      $196 = HEAP32[$195>>2]|0;
      $197 = (($194) - ($196))|0;
      $198 = (((($191) + ($190<<3)|0)) + 4|0);
      $199 = HEAP32[$198>>2]|0;
      $200 = (((($192) + ($182<<3)|0)) + 4|0);
      $201 = HEAP32[$200>>2]|0;
      $202 = (($199) - ($201))|0;
      $203 = ($194|0)==($196|0);
      $204 = ($199|0)==($201|0);
      $or$cond$i$i = $203 & $204;
      $$$i$i54 = $or$cond$i$i ? 16384 : $197;
      $$op$i$i = $132 & 1;
      $205 = ($$op$i$i<<24>>24)==(0);
      $206 = $205 | $or$cond$i$i;
      $207 = (0 - ($202))|0;
      $B$0$i$i = $206 ? $202 : $$$i$i54;
      $A$1$i$i = $206 ? $$$i$i54 : $207;
      $208 = ($A$1$i$i|0)<(0);
      $209 = (0 - ($A$1$i$i))|0;
      $210 = $208 ? $209 : $A$1$i$i;
      $211 = ($210|0)<(16384);
      if ($211) {
       $212 = ($B$0$i$i|0)<(0);
       $213 = (0 - ($B$0$i$i))|0;
       $214 = $212 ? $213 : $B$0$i$i;
       $215 = ($214|0)<(16384);
       if ($215) {
        $216 = $B$0$i$i | $A$1$i$i;
        $217 = ($216|0)==(0);
        if (!($217)) {
         $218 = $A$1$i$i << 14;
         $219 = $B$0$i$i << 14;
         $$02$i$i = $218;$$03$i$i = $219;
         label = 40;
        }
       } else {
        $$02$i$i = $A$1$i$i;$$03$i$i = $B$0$i$i;
        label = 40;
       }
      } else {
       $$02$i$i = $A$1$i$i;$$03$i$i = $B$0$i$i;
       label = 40;
      }
      if ((label|0) == 40) {
       label = 0;
       $220 = (_FT_Hypot($$02$i$i,$$03$i$i)|0);
       $221 = $220 << 2;
       $222 = (_FT_DivFix($$02$i$i,$221)|0);
       $223 = $222&65535;
       HEAP16[$73>>1] = $223;
       $224 = (_FT_DivFix($$03$i$i,$221)|0);
       $225 = $224&65535;
       HEAP16[$70>>1] = $225;
      }
      $226 = HEAPU16[$73>>1]|(HEAPU16[$73+2>>1]<<16);
      HEAP16[$76>>1]=$226&65535;HEAP16[$76+2>>1]=$226>>>16;
      _Compute_Funcs($exc);
      label = 821;
      break L30;
     }
    }
    $188 = HEAP8[$29>>0]|0;
    $189 = ($188<<24>>24)==(0);
    if ($189) {
     label = 821;
    } else {
     label = 35;
     break L12;
    }
    break;
   }
   case 9: case 8:  {
    $$val2 = HEAP32[$172>>2]|0;
    $$idx3 = ((($172)) + 4|0);
    $$idx3$val = HEAP32[$$idx3>>2]|0;
    $227 = $$idx3$val&65535;
    $228 = $$idx3$val & 65535;
    $229 = HEAP16[$74>>1]|0;
    $230 = ($229&65535)>($227&65535);
    if ($230) {
     $231 = $$val2&65535;
     $232 = HEAP16[$38>>1]|0;
     $233 = ($232&65535)>($231&65535);
     if ($233) {
      $236 = $$val2 & 65535;
      $237 = HEAP32[$49>>2]|0;
      $238 = HEAP32[$75>>2]|0;
      $239 = (($237) + ($236<<3)|0);
      $240 = HEAP32[$239>>2]|0;
      $241 = (($238) + ($228<<3)|0);
      $242 = HEAP32[$241>>2]|0;
      $243 = (($240) - ($242))|0;
      $244 = (((($237) + ($236<<3)|0)) + 4|0);
      $245 = HEAP32[$244>>2]|0;
      $246 = (((($238) + ($228<<3)|0)) + 4|0);
      $247 = HEAP32[$246>>2]|0;
      $248 = (($245) - ($247))|0;
      $249 = ($240|0)==($242|0);
      $250 = ($245|0)==($247|0);
      $or$cond$i$i55 = $249 & $250;
      $$$i$i56 = $or$cond$i$i55 ? 16384 : $243;
      $$op$i$i57 = $132 & 1;
      $251 = ($$op$i$i57<<24>>24)==(0);
      $252 = $251 | $or$cond$i$i55;
      $253 = (0 - ($248))|0;
      $B$0$i$i58 = $252 ? $248 : $$$i$i56;
      $A$1$i$i59 = $252 ? $$$i$i56 : $253;
      $254 = ($A$1$i$i59|0)<(0);
      $255 = (0 - ($A$1$i$i59))|0;
      $256 = $254 ? $255 : $A$1$i$i59;
      $257 = ($256|0)<(16384);
      if ($257) {
       $258 = ($B$0$i$i58|0)<(0);
       $259 = (0 - ($B$0$i$i58))|0;
       $260 = $258 ? $259 : $B$0$i$i58;
       $261 = ($260|0)<(16384);
       if ($261) {
        $262 = $B$0$i$i58 | $A$1$i$i59;
        $263 = ($262|0)==(0);
        if (!($263)) {
         $264 = $A$1$i$i59 << 14;
         $265 = $B$0$i$i58 << 14;
         $$02$i$i61 = $264;$$03$i$i60 = $265;
         label = 50;
        }
       } else {
        $$02$i$i61 = $A$1$i$i59;$$03$i$i60 = $B$0$i$i58;
        label = 50;
       }
      } else {
       $$02$i$i61 = $A$1$i$i59;$$03$i$i60 = $B$0$i$i58;
       label = 50;
      }
      if ((label|0) == 50) {
       label = 0;
       $266 = (_FT_Hypot($$02$i$i61,$$03$i$i60)|0);
       $267 = $266 << 2;
       $268 = (_FT_DivFix($$02$i$i61,$267)|0);
       $269 = $268&65535;
       HEAP16[$46>>1] = $269;
       $270 = (_FT_DivFix($$03$i$i60,$267)|0);
       $271 = $270&65535;
       HEAP16[$48>>1] = $271;
      }
      _Compute_Funcs($exc);
      label = 821;
      break L30;
     }
    }
    $234 = HEAP8[$29>>0]|0;
    $235 = ($234<<24>>24)==(0);
    if ($235) {
     label = 821;
    } else {
     label = 45;
     break L12;
    }
    break;
   }
   case 10:  {
    $$val4 = HEAP32[$172>>2]|0;
    $$idx5 = ((($172)) + 4|0);
    $$idx5$val = HEAP32[$$idx5>>2]|0;
    $sext$i62 = $$idx5$val << 16;
    $272 = $sext$i62 >> 16;
    $sext1$i = $$val4 << 16;
    $273 = $sext1$i >> 16;
    $274 = ($sext1$i|0)<(0);
    $275 = (0 - ($273))|0;
    $276 = $274 ? $275 : $273;
    $277 = ($276|0)<(16384);
    if ($277) {
     $278 = ($sext$i62|0)<(0);
     $279 = (0 - ($272))|0;
     $280 = $278 ? $279 : $272;
     $281 = ($280|0)<(16384);
     if ($281) {
      $282 = $272 | $273;
      $283 = ($282|0)==(0);
      if (!($283)) {
       $284 = $sext1$i >> 2;
       $285 = $sext$i62 >> 2;
       $$02$i$i64 = $284;$$03$i$i63 = $285;
       label = 56;
      }
     } else {
      $$02$i$i64 = $273;$$03$i$i63 = $272;
      label = 56;
     }
    } else {
     $$02$i$i64 = $273;$$03$i$i63 = $272;
     label = 56;
    }
    if ((label|0) == 56) {
     label = 0;
     $286 = (_FT_Hypot($$02$i$i64,$$03$i$i63)|0);
     $287 = $286 << 2;
     $288 = (_FT_DivFix($$02$i$i64,$287)|0);
     $289 = $288&65535;
     HEAP16[$73>>1] = $289;
     $290 = (_FT_DivFix($$03$i$i63,$287)|0);
     $291 = $290&65535;
     HEAP16[$70>>1] = $291;
    }
    $292 = HEAPU16[$73>>1]|(HEAPU16[$73+2>>1]<<16);
    HEAP16[$76>>1]=$292&65535;HEAP16[$76+2>>1]=$292>>>16;
    _Compute_Funcs($exc);
    label = 821;
    break;
   }
   case 11:  {
    $$val6 = HEAP32[$172>>2]|0;
    $$idx7 = ((($172)) + 4|0);
    $$idx7$val = HEAP32[$$idx7>>2]|0;
    $sext$i65 = $$idx7$val << 16;
    $293 = $sext$i65 >> 16;
    $sext1$i66 = $$val6 << 16;
    $294 = $sext1$i66 >> 16;
    $295 = ($sext1$i66|0)<(0);
    $296 = (0 - ($294))|0;
    $297 = $295 ? $296 : $294;
    $298 = ($297|0)<(16384);
    if ($298) {
     $299 = ($sext$i65|0)<(0);
     $300 = (0 - ($293))|0;
     $301 = $299 ? $300 : $293;
     $302 = ($301|0)<(16384);
     if ($302) {
      $303 = $293 | $294;
      $304 = ($303|0)==(0);
      if (!($304)) {
       $305 = $sext1$i66 >> 2;
       $306 = $sext$i65 >> 2;
       $$02$i$i68 = $305;$$03$i$i67 = $306;
       label = 62;
      }
     } else {
      $$02$i$i68 = $294;$$03$i$i67 = $293;
      label = 62;
     }
    } else {
     $$02$i$i68 = $294;$$03$i$i67 = $293;
     label = 62;
    }
    if ((label|0) == 62) {
     label = 0;
     $307 = (_FT_Hypot($$02$i$i68,$$03$i$i67)|0);
     $308 = $307 << 2;
     $309 = (_FT_DivFix($$02$i$i68,$308)|0);
     $310 = $309&65535;
     HEAP16[$46>>1] = $310;
     $311 = (_FT_DivFix($$03$i$i67,$308)|0);
     $312 = $311&65535;
     HEAP16[$48>>1] = $312;
    }
    _Compute_Funcs($exc);
    label = 821;
    break;
   }
   case 12:  {
    $exc$idx$val = HEAP16[$69>>1]|0;
    $exc$idx8$val = HEAP16[$70>>1]|0;
    $313 = $exc$idx$val << 16 >> 16;
    HEAP32[$172>>2] = $313;
    $314 = $exc$idx8$val << 16 >> 16;
    $315 = ((($172)) + 4|0);
    HEAP32[$315>>2] = $314;
    label = 821;
    break;
   }
   case 13:  {
    $exc$idx9$val = HEAP16[$46>>1]|0;
    $exc$idx10$val = HEAP16[$48>>1]|0;
    $316 = $exc$idx9$val << 16 >> 16;
    HEAP32[$172>>2] = $316;
    $317 = $exc$idx10$val << 16 >> 16;
    $318 = ((($172)) + 4|0);
    HEAP32[$318>>2] = $317;
    label = 821;
    break;
   }
   case 14:  {
    $319 = HEAPU16[$73>>1]|(HEAPU16[$73+2>>1]<<16);
    HEAP16[$77>>1]=$319&65535;HEAP16[$77+2>>1]=$319>>>16;
    _Compute_Funcs($exc);
    label = 821;
    break;
   }
   case 15:  {
    $320 = HEAP32[$172>>2]|0;
    $321 = ((($172)) + 4|0);
    $322 = HEAP32[$321>>2]|0;
    $323 = ((($172)) + 8|0);
    $324 = HEAP32[$323>>2]|0;
    $325 = ((($172)) + 12|0);
    $326 = HEAP32[$325>>2]|0;
    $327 = $326 & 65535;
    $328 = HEAP16[$41>>1]|0;
    $329 = $328&65535;
    $330 = ($327>>>0)<($329>>>0);
    if ($330) {
     $331 = ((($172)) + 16|0);
     $332 = HEAP32[$331>>2]|0;
     $333 = $332 & 65535;
     $334 = ($333>>>0)<($329>>>0);
     if ($334) {
      $335 = $322 & 65535;
      $336 = HEAP16[$38>>1]|0;
      $337 = $336&65535;
      $338 = ($335>>>0)<($337>>>0);
      if ($338) {
       $339 = $324 & 65535;
       $340 = ($339>>>0)<($337>>>0);
       if ($340) {
        $341 = $320 & 65535;
        $342 = HEAP16[$74>>1]|0;
        $343 = $342&65535;
        $344 = ($341>>>0)<($343>>>0);
        if ($344) {
         $347 = HEAP32[$52>>2]|0;
         $348 = (($347) + ($333<<3)|0);
         $349 = HEAP32[$348>>2]|0;
         $350 = (($347) + ($327<<3)|0);
         $351 = HEAP32[$350>>2]|0;
         $352 = (($349) - ($351))|0;
         $353 = (((($347) + ($333<<3)|0)) + 4|0);
         $354 = HEAP32[$353>>2]|0;
         $355 = (((($347) + ($327<<3)|0)) + 4|0);
         $356 = HEAP32[$355>>2]|0;
         $357 = (($354) - ($356))|0;
         $358 = HEAP32[$49>>2]|0;
         $359 = (($358) + ($339<<3)|0);
         $360 = HEAP32[$359>>2]|0;
         $361 = (($358) + ($335<<3)|0);
         $362 = HEAP32[$361>>2]|0;
         $363 = (($360) - ($362))|0;
         $364 = (((($358) + ($339<<3)|0)) + 4|0);
         $365 = HEAP32[$364>>2]|0;
         $366 = (((($358) + ($335<<3)|0)) + 4|0);
         $367 = HEAP32[$366>>2]|0;
         $368 = (($365) - ($367))|0;
         $369 = HEAP32[$78>>2]|0;
         $370 = (($369) + ($341)|0);
         $371 = HEAP8[$370>>0]|0;
         $372 = $371&255;
         $373 = $372 | 24;
         $374 = $373&255;
         HEAP8[$370>>0] = $374;
         $375 = (0 - ($357))|0;
         $376 = (_FT_MulDiv($363,$375,64)|0);
         $377 = (_FT_MulDiv($368,$352,64)|0);
         $378 = (($377) + ($376))|0;
         $379 = (_FT_MulDiv($363,$352,64)|0);
         $380 = (_FT_MulDiv($368,$357,64)|0);
         $381 = (($380) + ($379))|0;
         $382 = ($378|0)<(0);
         $383 = (0 - ($378))|0;
         $384 = $382 ? $383 : $378;
         $385 = ($384*19)|0;
         $386 = ($381|0)<(0);
         $387 = (0 - ($381))|0;
         $388 = $386 ? $387 : $381;
         $389 = ($385|0)>($388|0);
         if ($389) {
          $390 = (($356) - ($367))|0;
          $391 = (($351) - ($362))|0;
          $392 = (_FT_MulDiv($391,$375,64)|0);
          $393 = (_FT_MulDiv($390,$352,64)|0);
          $394 = (($393) + ($392))|0;
          $395 = (_FT_MulDiv($394,$363,$378)|0);
          $396 = (_FT_MulDiv($394,$368,$378)|0);
          $397 = HEAP32[$49>>2]|0;
          $398 = (($397) + ($335<<3)|0);
          $399 = HEAP32[$398>>2]|0;
          $400 = (($399) + ($395))|0;
          $401 = HEAP32[$75>>2]|0;
          $402 = (($401) + ($341<<3)|0);
          HEAP32[$402>>2] = $400;
          $403 = (((($397) + ($335<<3)|0)) + 4|0);
          $404 = HEAP32[$403>>2]|0;
          $405 = (($404) + ($396))|0;
          $406 = (((($401) + ($341<<3)|0)) + 4|0);
          HEAP32[$406>>2] = $405;
          label = 821;
          break L30;
         } else {
          $407 = HEAP32[$49>>2]|0;
          $408 = (($407) + ($335<<3)|0);
          $409 = HEAP32[$408>>2]|0;
          $410 = (($407) + ($339<<3)|0);
          $411 = HEAP32[$410>>2]|0;
          $412 = (($411) + ($409))|0;
          $413 = HEAP32[$52>>2]|0;
          $414 = (($413) + ($327<<3)|0);
          $415 = HEAP32[$414>>2]|0;
          $416 = (($412) + ($415))|0;
          $417 = (($413) + ($333<<3)|0);
          $418 = HEAP32[$417>>2]|0;
          $419 = (($416) + ($418))|0;
          $420 = (($419|0) / 4)&-1;
          $421 = HEAP32[$75>>2]|0;
          $422 = (($421) + ($341<<3)|0);
          HEAP32[$422>>2] = $420;
          $423 = (((($407) + ($335<<3)|0)) + 4|0);
          $424 = HEAP32[$423>>2]|0;
          $425 = (((($407) + ($339<<3)|0)) + 4|0);
          $426 = HEAP32[$425>>2]|0;
          $427 = (($426) + ($424))|0;
          $428 = (((($413) + ($327<<3)|0)) + 4|0);
          $429 = HEAP32[$428>>2]|0;
          $430 = (($427) + ($429))|0;
          $431 = (((($413) + ($333<<3)|0)) + 4|0);
          $432 = HEAP32[$431>>2]|0;
          $433 = (($430) + ($432))|0;
          $434 = (($433|0) / 4)&-1;
          $435 = (((($421) + ($341<<3)|0)) + 4|0);
          HEAP32[$435>>2] = $434;
          label = 821;
          break L30;
         }
        }
       }
      }
     }
    }
    $345 = HEAP8[$29>>0]|0;
    $346 = ($345<<24>>24)==(0);
    if ($346) {
     label = 821;
    } else {
     label = 73;
     break L12;
    }
    break;
   }
   case 16:  {
    $$val11 = HEAP32[$172>>2]|0;
    $436 = $$val11&65535;
    HEAP16[$79>>1] = $436;
    label = 821;
    break;
   }
   case 17:  {
    $$val12 = HEAP32[$172>>2]|0;
    $437 = $$val12&65535;
    HEAP16[$56>>1] = $437;
    label = 821;
    break;
   }
   case 18:  {
    $$val13 = HEAP32[$172>>2]|0;
    $438 = $$val13&65535;
    HEAP16[$57>>1] = $438;
    label = 821;
    break;
   }
   case 19:  {
    $439 = HEAP32[$172>>2]|0;
    switch ($439|0) {
    case 0:  {
     dest=$80; src=$81; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
     break;
    }
    case 1:  {
     dest=$80; src=$82; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
     break;
    }
    default: {
     $440 = HEAP8[$29>>0]|0;
     $441 = ($440<<24>>24)==(0);
     if ($441) {
      label = 821;
      break L30;
     } else {
      label = 84;
      break L12;
     }
    }
    }
    $442 = HEAP32[$172>>2]|0;
    $443 = $442&65535;
    HEAP16[$55>>1] = $443;
    label = 821;
    break;
   }
   case 20:  {
    $444 = HEAP32[$172>>2]|0;
    switch ($444|0) {
    case 0:  {
     dest=$37; src=$81; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
     break;
    }
    case 1:  {
     dest=$37; src=$82; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
     break;
    }
    default: {
     $445 = HEAP8[$29>>0]|0;
     $446 = ($445<<24>>24)==(0);
     if ($446) {
      label = 821;
      break L30;
     } else {
      label = 90;
      break L12;
     }
    }
    }
    $447 = HEAP32[$172>>2]|0;
    $448 = $447&65535;
    HEAP16[$44>>1] = $448;
    label = 821;
    break;
   }
   case 21:  {
    $449 = HEAP32[$172>>2]|0;
    switch ($449|0) {
    case 0:  {
     dest=$83; src=$81; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
     break;
    }
    case 1:  {
     dest=$83; src=$82; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
     break;
    }
    default: {
     $450 = HEAP8[$29>>0]|0;
     $451 = ($450<<24>>24)==(0);
     if ($451) {
      label = 821;
      break L30;
     } else {
      label = 96;
      break L12;
     }
    }
    }
    $452 = HEAP32[$172>>2]|0;
    $453 = $452&65535;
    HEAP16[$84>>1] = $453;
    label = 821;
    break;
   }
   case 22:  {
    $454 = HEAP32[$172>>2]|0;
    switch ($454|0) {
    case 0:  {
     dest=$80; src=$81; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
     break;
    }
    case 1:  {
     dest=$80; src=$82; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
     break;
    }
    default: {
     $455 = HEAP8[$29>>0]|0;
     $456 = ($455<<24>>24)==(0);
     if ($456) {
      label = 821;
      break L30;
     } else {
      label = 102;
      break L12;
     }
    }
    }
    dest=$37; src=$80; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
    dest=$83; src=$80; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
    $457 = HEAP32[$172>>2]|0;
    $458 = $457&65535;
    HEAP16[$55>>1] = $458;
    HEAP16[$44>>1] = $458;
    HEAP16[$84>>1] = $458;
    label = 821;
    break;
   }
   case 23:  {
    $$val14 = HEAP32[$172>>2]|0;
    $459 = ($$val14|0)<(0);
    if ($459) {
     label = 105;
     break L12;
    }
    HEAP32[$85>>2] = $$val14;
    label = 821;
    break;
   }
   case 24:  {
    HEAP32[$11>>2] = 1;
    HEAP32[$$pre$phiZ2D>>2] = 4419;
    label = 821;
    break;
   }
   case 25:  {
    HEAP32[$11>>2] = 0;
    HEAP32[$$pre$phiZ2D>>2] = 4422;
    label = 821;
    break;
   }
   case 26:  {
    $$val15 = HEAP32[$172>>2]|0;
    HEAP32[$35>>2] = $$val15;
    label = 821;
    break;
   }
   case 27:  {
    $$pre1$pre$i = HEAP32[$22>>2]|0;
    $$pre2$pre$i = HEAP32[$26>>2]|0;
    $$pre$i69 = $$pre$pre$i;$$pre1$i = $$pre1$pre$i;$nIfs$0$ph$i = 1;
    while(1) {
     $461 = $$pre$i69;$462 = $$pre1$i;
     L236: while(1) {
      $460 = (($461) + ($462))|0;
      HEAP32[$22>>2] = $460;
      $463 = ($460|0)<($$pre2$pre$i|0);
      if (!($463)) {
       label = 117;
       break L12;
      }
      $464 = (($129) + ($460)|0);
      $465 = HEAP8[$464>>0]|0;
      HEAP8[$24>>0] = $465;
      $466 = $465&255;
      $467 = (593353 + ($466)|0);
      $468 = HEAP8[$467>>0]|0;
      $469 = $468 << 24 >> 24;
      HEAP32[$25>>2] = $469;
      $470 = $466 | 1;
      $471 = ($470|0)==(65);
      if ($471) {
       $472 = (($460) + 1)|0;
       $473 = ($472|0)<($$pre2$pre$i|0);
       if (!($473)) {
        label = 117;
        break L12;
       }
       $474 = (($129) + ($472)|0);
       $475 = HEAP8[$474>>0]|0;
       $476 = $475&255;
       $477 = Math_imul($476, $469)|0;
       $478 = (2 - ($477))|0;
       HEAP32[$25>>2] = $478;
       $480 = $478;
      } else {
       $480 = $469;
      }
      $479 = (($480) + ($460))|0;
      $481 = ($479|0)>($$pre2$pre$i|0);
      if ($481) {
       label = 117;
       break L12;
      }
      switch ($466|0) {
      case 88:  {
       $$lcssa623 = $460;$$lcssa625 = $480;
       label = 119;
       break L236;
       break;
      }
      case 89:  {
       $$lcssa624 = $460;$$lcssa626 = $480;
       label = 120;
       break L236;
       break;
      }
      default: {
       $461 = $480;$462 = $460;
      }
      }
     }
     if ((label|0) == 119) {
      label = 0;
      $482 = (($nIfs$0$ph$i) + 1)|0;
      $2795 = $$lcssa625;$2796 = $$lcssa623;$nIfs$1$i = $482;
     }
     else if ((label|0) == 120) {
      label = 0;
      $483 = (($nIfs$0$ph$i) + -1)|0;
      $2795 = $$lcssa626;$2796 = $$lcssa624;$nIfs$1$i = $483;
     }
     $484 = ($nIfs$1$i|0)==(0);
     if ($484) {
      label = 821;
      break;
     } else {
      $$pre$i69 = $2795;$$pre1$i = $2796;$nIfs$0$ph$i = $nIfs$1$i;
     }
    }
    break;
   }
   case 28:  {
    $$val16 = HEAP32[$172>>2]|0;
    $485 = $$val16 | $168;
    $486 = ($485|0)==(0);
    if ($486) {
     HEAP32[$34>>2] = 132;
    }
    $487 = HEAP32[$22>>2]|0;
    $488 = (($487) + ($$val16))|0;
    HEAP32[$22>>2] = $488;
    $489 = ($488|0)<(0);
    if ($489) {
     label = 127;
    } else {
     $490 = HEAP32[$64>>2]|0;
     $491 = ($490|0)>(0);
     if ($491) {
      $492 = (($490) + -1)|0;
      $493 = HEAP32[$66>>2]|0;
      $494 = (((($493) + ($492<<4)|0)) + 12|0);
      $495 = HEAP32[$494>>2]|0;
      $496 = ((($495)) + 8|0);
      $497 = HEAP32[$496>>2]|0;
      $498 = ($488|0)>($497|0);
      if ($498) {
       label = 127;
      }
     }
    }
    if ((label|0) == 127) {
     label = 0;
     HEAP32[$34>>2] = 132;
    }
    HEAP8[$33>>0] = 0;
    label = 821;
    break;
   }
   case 29:  {
    $$val17 = HEAP32[$172>>2]|0;
    HEAP32[$36>>2] = $$val17;
    label = 821;
    break;
   }
   case 30:  {
    $$val18 = HEAP32[$172>>2]|0;
    HEAP32[$43>>2] = $$val18;
    label = 821;
    break;
   }
   case 31:  {
    $$val19 = HEAP32[$172>>2]|0;
    $499 = HEAP32[$86>>2]|0;
    $500 = (_FT_MulFix($$val19,$499)|0);
    HEAP32[$42>>2] = $500;
    label = 821;
    break;
   }
   case 32:  {
    $501 = HEAP32[$172>>2]|0;
    $502 = ((($172)) + 4|0);
    HEAP32[$502>>2] = $501;
    label = 821;
    break;
   }
   case 127: case 126: case 89: case 33:  {
    label = 837;
    break;
   }
   case 34:  {
    HEAP32[$31>>2] = 0;
    label = 821;
    break;
   }
   case 35:  {
    $503 = HEAP32[$172>>2]|0;
    $504 = ((($172)) + 4|0);
    $505 = HEAP32[$504>>2]|0;
    HEAP32[$172>>2] = $505;
    HEAP32[$504>>2] = $503;
    label = 821;
    break;
   }
   case 36:  {
    $exc$idx20$val = HEAP32[$27>>2]|0;
    HEAP32[$172>>2] = $exc$idx20$val;
    label = 821;
    break;
   }
   case 37:  {
    $506 = HEAP32[$172>>2]|0;
    $507 = ($506|0)<(1);
    $508 = ($168|0)<($506|0);
    $or$cond250 = $507 | $508;
    if ($or$cond250) {
     $509 = HEAP8[$29>>0]|0;
     $510 = ($509<<24>>24)==(0);
     if ($510) {
      $storemerge$i = 0;
     } else {
      HEAP32[$34>>2] = 134;
      $storemerge$i = 0;
     }
    } else {
     $511 = (($168) - ($506))|0;
     $512 = (($171) + ($511<<2)|0);
     $513 = HEAP32[$512>>2]|0;
     $storemerge$i = $513;
    }
    HEAP32[$172>>2] = $storemerge$i;
    label = 821;
    break;
   }
   case 38:  {
    $$val21 = HEAP32[$172>>2]|0;
    $514 = ($$val21|0)<(1);
    $515 = ($168|0)<($$val21|0);
    $or$cond251 = $514 | $515;
    if ($or$cond251) {
     $516 = HEAP8[$29>>0]|0;
     $517 = ($516<<24>>24)==(0);
     if ($517) {
      label = 821;
      break L30;
     } else {
      label = 143;
      break L12;
     }
    } else {
     $518 = (($168) - ($$val21))|0;
     $519 = (($171) + ($518<<2)|0);
     $520 = HEAP32[$519>>2]|0;
     $521 = (($518) + 1)|0;
     $522 = (($171) + ($521<<2)|0);
     $523 = $$val21 << 2;
     $524 = (($523) + -4)|0;
     _memmove(($519|0),($522|0),($524|0))|0;
     $525 = HEAP32[$28>>2]|0;
     $526 = (($525) + -1)|0;
     $527 = HEAP32[$30>>2]|0;
     $528 = (($527) + ($526<<2)|0);
     HEAP32[$528>>2] = $520;
     label = 821;
     break L30;
    }
    break;
   }
   case 39:  {
    $$val22 = HEAP32[$172>>2]|0;
    $$idx23 = ((($172)) + 4|0);
    $$idx23$val = HEAP32[$$idx23>>2]|0;
    $529 = $$val22&65535;
    $530 = $$idx23$val&65535;
    $531 = $$val22 & 65535;
    $532 = HEAP16[$38>>1]|0;
    $533 = $532&65535;
    $534 = ($531>>>0)<($533>>>0);
    if ($534) {
     $535 = $$idx23$val & 65535;
     $536 = HEAP16[$41>>1]|0;
     $537 = $536&65535;
     $538 = ($535>>>0)<($537>>>0);
     if ($538) {
      $541 = HEAP32[$51>>2]|0;
      $542 = HEAP32[$52>>2]|0;
      $543 = (($542) + ($535<<3)|0);
      $544 = HEAP32[$543>>2]|0;
      $545 = HEAP32[$49>>2]|0;
      $546 = (($545) + ($531<<3)|0);
      $547 = HEAP32[$546>>2]|0;
      $548 = (($544) - ($547))|0;
      $549 = (((($542) + ($535<<3)|0)) + 4|0);
      $550 = HEAP32[$549>>2]|0;
      $551 = (((($545) + ($531<<3)|0)) + 4|0);
      $552 = HEAP32[$551>>2]|0;
      $553 = (($550) - ($552))|0;
      $554 = (FUNCTION_TABLE_iiii[$541 & 8191]($exc,$548,$553)|0);
      $555 = (($554|0) / 2)&-1;
      $556 = HEAP32[$54>>2]|0;
      FUNCTION_TABLE_viiii[$556 & 8191]($exc,$37,$529,$555);
      $557 = HEAP32[$54>>2]|0;
      $558 = (0 - ($555))|0;
      FUNCTION_TABLE_viiii[$557 & 8191]($exc,$80,$530,$558);
      label = 821;
      break L30;
     }
    }
    $539 = HEAP8[$29>>0]|0;
    $540 = ($539<<24>>24)==(0);
    if ($540) {
     label = 821;
    } else {
     label = 148;
     break L12;
    }
    break;
   }
   case 40:  {
    $559 = HEAP32[$62>>2]|0;
    $560 = HEAP32[$63>>2]|0;
    $561 = (($559) + (($560*24)|0)|0);
    $562 = ($560|0)>(0);
    L281: do {
     if ($562) {
      $def$02$i71 = $559;
      while(1) {
       $563 = ((($def$02$i71)) + 12|0);
       $564 = HEAP32[$563>>2]|0;
       $565 = $564&255;
       $566 = ($132<<24>>24)==($565<<24>>24);
       if ($566) {
        $567 = ((($def$02$i71)) + 16|0);
        $568 = HEAP8[$567>>0]|0;
        $569 = ($568<<24>>24)==(0);
        if (!($569)) {
         $def$02$i71$lcssa = $def$02$i71;
         break;
        }
       }
       $592 = ((($def$02$i71)) + 24|0);
       $593 = ($592>>>0)<($561>>>0);
       if ($593) {
        $def$02$i71 = $592;
       } else {
        break L281;
       }
      }
      $570 = HEAP32[$64>>2]|0;
      $571 = HEAP32[$65>>2]|0;
      $572 = ($570|0)<($571|0);
      if (!($572)) {
       label = 154;
       break L12;
      }
      $573 = HEAP32[$66>>2]|0;
      $574 = (($570) + 1)|0;
      HEAP32[$64>>2] = $574;
      $575 = HEAP32[$67>>2]|0;
      $576 = (($573) + ($570<<4)|0);
      HEAP32[$576>>2] = $575;
      $577 = HEAP32[$22>>2]|0;
      $578 = (($577) + 1)|0;
      $579 = (((($573) + ($570<<4)|0)) + 4|0);
      HEAP32[$579>>2] = $578;
      $580 = (((($573) + ($570<<4)|0)) + 8|0);
      HEAP32[$580>>2] = 1;
      $581 = (((($573) + ($570<<4)|0)) + 12|0);
      HEAP32[$581>>2] = $def$02$i71$lcssa;
      $582 = HEAP32[$def$02$i71$lcssa>>2]|0;
      $583 = ((($def$02$i71$lcssa)) + 4|0);
      $584 = HEAP32[$583>>2]|0;
      $aRange$off$i$i72 = (($582) + -1)|0;
      $585 = ($aRange$off$i$i72>>>0)>(2);
      do {
       if ($585) {
        HEAP32[$34>>2] = 132;
       } else {
        $586 = (((($exc)) + 444|0) + ($aRange$off$i$i72<<3)|0);
        $587 = HEAP32[$586>>2]|0;
        $588 = ($587|0)==(0|0);
        if ($588) {
         HEAP32[$34>>2] = 138;
         break;
        }
        $589 = (((((($exc)) + 444|0) + ($aRange$off$i$i72<<3)|0)) + 4|0);
        $590 = HEAP32[$589>>2]|0;
        $591 = ($590|0)<($584|0);
        if ($591) {
         HEAP32[$34>>2] = 131;
         break;
        } else {
         HEAP32[$23>>2] = $587;
         HEAP32[$26>>2] = $590;
         HEAP32[$22>>2] = $584;
         HEAP32[$67>>2] = $582;
         break;
        }
       }
      } while(0);
      HEAP8[$33>>0] = 0;
      label = 821;
      break L30;
     }
    } while(0);
    HEAP32[$34>>2] = 128;
    $2736 = $559;$2737 = $560;
    label = 823;
    break;
   }
   case 41:  {
    $$val24 = HEAP32[$172>>2]|0;
    $594 = $$val24 & 65535;
    $595 = HEAP16[$41>>1]|0;
    $596 = $595&65535;
    $597 = ($594>>>0)<($596>>>0);
    if (!($597)) {
     $598 = HEAP8[$29>>0]|0;
     $599 = ($598<<24>>24)==(0);
     if ($599) {
      label = 821;
      break L30;
     } else {
      label = 167;
      break L12;
     }
    }
    $600 = HEAP16[$46>>1]|0;
    $601 = ($600<<16>>16)==(0);
    $$$i76 = $601 ? -1 : -9;
    $602 = HEAP16[$48>>1]|0;
    $603 = ($602<<16>>16)==(0);
    if ($603) {
     $mask$1$i = $$$i76;
    } else {
     $604 = $$$i76&255;
     $605 = $604 & 239;
     $606 = $605&255;
     $mask$1$i = $606;
    }
    $607 = HEAP32[$87>>2]|0;
    $608 = (($607) + ($594)|0);
    $609 = HEAP8[$608>>0]|0;
    $610 = $609 & $mask$1$i;
    HEAP8[$608>>0] = $610;
    label = 821;
    break;
   }
   case 42:  {
    $611 = ((($172)) + 4|0);
    $612 = HEAP32[$611>>2]|0;
    $613 = HEAP32[$88>>2]|0;
    $614 = (($613) + 1)|0;
    $615 = ($612>>>0)<($614>>>0);
    if (!($615)) {
     label = 190;
     break L12;
    }
    $616 = HEAP32[$89>>2]|0;
    $617 = HEAP32[$90>>2]|0;
    $618 = ($614|0)==($617|0);
    if ($618) {
     $619 = (($616) + (($612*24)|0)|0);
     $620 = (((($616) + (($612*24)|0)|0)) + 12|0);
     $621 = HEAP32[$620>>2]|0;
     $622 = ($621|0)==($612|0);
     if ($622) {
      $def$1$i = $619;
     } else {
      label = 174;
     }
    } else {
     label = 174;
    }
    if ((label|0) == 174) {
     label = 0;
     $623 = (($616) + (($617*24)|0)|0);
     $624 = ($617|0)>(0);
     L313: do {
      if ($624) {
       $def$01$i = $616;
       while(1) {
        $625 = ((($def$01$i)) + 12|0);
        $626 = HEAP32[$625>>2]|0;
        $627 = ($626|0)==($612|0);
        if ($627) {
         $def$0$lcssa$i = $def$01$i;
         break L313;
        }
        $628 = ((($def$01$i)) + 24|0);
        $629 = ($628>>>0)<($623>>>0);
        if ($629) {
         $def$01$i = $628;
        } else {
         $def$0$lcssa$i = $628;
         break;
        }
       }
      } else {
       $def$0$lcssa$i = $616;
      }
     } while(0);
     $630 = ($def$0$lcssa$i|0)==($623|0);
     if ($630) {
      label = 190;
      break L12;
     } else {
      $def$1$i = $def$0$lcssa$i;
     }
    }
    $631 = ((($def$1$i)) + 16|0);
    $632 = HEAP8[$631>>0]|0;
    $633 = ($632<<24>>24)==(0);
    if ($633) {
     label = 190;
     break L12;
    }
    $634 = HEAP32[$64>>2]|0;
    $635 = HEAP32[$65>>2]|0;
    $636 = ($634|0)<($635|0);
    if (!($636)) {
     label = 180;
     break L12;
    }
    $637 = HEAP32[$172>>2]|0;
    $638 = ($637|0)>(0);
    if ($638) {
     $639 = HEAP32[$66>>2]|0;
     $640 = HEAP32[$67>>2]|0;
     $641 = (($639) + ($634<<4)|0);
     HEAP32[$641>>2] = $640;
     $642 = HEAP32[$22>>2]|0;
     $643 = (($642) + 1)|0;
     $644 = (((($639) + ($634<<4)|0)) + 4|0);
     HEAP32[$644>>2] = $643;
     $645 = HEAP32[$172>>2]|0;
     $646 = (((($639) + ($634<<4)|0)) + 8|0);
     HEAP32[$646>>2] = $645;
     $647 = (((($639) + ($634<<4)|0)) + 12|0);
     HEAP32[$647>>2] = $def$1$i;
     $648 = (($634) + 1)|0;
     HEAP32[$64>>2] = $648;
     $649 = HEAP32[$def$1$i>>2]|0;
     $650 = ((($def$1$i)) + 4|0);
     $651 = HEAP32[$650>>2]|0;
     $aRange$off$i$i78 = (($649) + -1)|0;
     $652 = ($aRange$off$i$i78>>>0)>(2);
     do {
      if ($652) {
       HEAP32[$34>>2] = 132;
      } else {
       $653 = (((($exc)) + 444|0) + ($aRange$off$i$i78<<3)|0);
       $654 = HEAP32[$653>>2]|0;
       $655 = ($654|0)==(0|0);
       if ($655) {
        HEAP32[$34>>2] = 138;
        break;
       }
       $656 = (((((($exc)) + 444|0) + ($aRange$off$i$i78<<3)|0)) + 4|0);
       $657 = HEAP32[$656>>2]|0;
       $658 = ($657|0)<($651|0);
       if ($658) {
        HEAP32[$34>>2] = 131;
        break;
       } else {
        HEAP32[$23>>2] = $654;
        HEAP32[$26>>2] = $657;
        HEAP32[$22>>2] = $651;
        HEAP32[$67>>2] = $649;
        break;
       }
      }
     } while(0);
     HEAP8[$33>>0] = 0;
     label = 821;
    } else {
     label = 821;
    }
    break;
   }
   case 43:  {
    $$val25 = HEAP32[$172>>2]|0;
    $659 = HEAP32[$88>>2]|0;
    $660 = (($659) + 1)|0;
    $661 = ($660>>>0)>($$val25>>>0);
    if (!($661)) {
     label = 209;
     break L12;
    }
    $662 = HEAP32[$89>>2]|0;
    $663 = HEAP32[$90>>2]|0;
    $664 = ($660|0)==($663|0);
    if ($664) {
     $665 = (($662) + (($$val25*24)|0)|0);
     $666 = (((($662) + (($$val25*24)|0)|0)) + 12|0);
     $667 = HEAP32[$666>>2]|0;
     $668 = ($667|0)==($$val25|0);
     if ($668) {
      $def$1$i84 = $665;
     } else {
      label = 194;
     }
    } else {
     label = 194;
    }
    if ((label|0) == 194) {
     label = 0;
     $669 = (($662) + (($663*24)|0)|0);
     $670 = ($663|0)>(0);
     L338: do {
      if ($670) {
       $def$01$i80 = $662;
       while(1) {
        $671 = ((($def$01$i80)) + 12|0);
        $672 = HEAP32[$671>>2]|0;
        $673 = ($672|0)==($$val25|0);
        if ($673) {
         $def$0$lcssa$i82 = $def$01$i80;
         break L338;
        }
        $674 = ((($def$01$i80)) + 24|0);
        $675 = ($674>>>0)<($669>>>0);
        if ($675) {
         $def$01$i80 = $674;
        } else {
         $def$0$lcssa$i82 = $674;
         break;
        }
       }
      } else {
       $def$0$lcssa$i82 = $662;
      }
     } while(0);
     $676 = ($def$0$lcssa$i82|0)==($669|0);
     if ($676) {
      label = 209;
      break L12;
     } else {
      $def$1$i84 = $def$0$lcssa$i82;
     }
    }
    $677 = ((($def$1$i84)) + 16|0);
    $678 = HEAP8[$677>>0]|0;
    $679 = ($678<<24>>24)==(0);
    if ($679) {
     label = 209;
     break L12;
    }
    $680 = HEAP32[$64>>2]|0;
    $681 = HEAP32[$65>>2]|0;
    $682 = ($680|0)<($681|0);
    if (!($682)) {
     label = 200;
     break L12;
    }
    $683 = HEAP32[$66>>2]|0;
    $684 = HEAP32[$67>>2]|0;
    $685 = (($683) + ($680<<4)|0);
    HEAP32[$685>>2] = $684;
    $686 = HEAP32[$22>>2]|0;
    $687 = (($686) + 1)|0;
    $688 = (((($683) + ($680<<4)|0)) + 4|0);
    HEAP32[$688>>2] = $687;
    $689 = (((($683) + ($680<<4)|0)) + 8|0);
    HEAP32[$689>>2] = 1;
    $690 = (((($683) + ($680<<4)|0)) + 12|0);
    HEAP32[$690>>2] = $def$1$i84;
    $691 = (($680) + 1)|0;
    HEAP32[$64>>2] = $691;
    $692 = HEAP32[$def$1$i84>>2]|0;
    $693 = ((($def$1$i84)) + 4|0);
    $694 = HEAP32[$693>>2]|0;
    $aRange$off$i$i85 = (($692) + -1)|0;
    $695 = ($aRange$off$i$i85>>>0)>(2);
    do {
     if ($695) {
      HEAP32[$34>>2] = 132;
     } else {
      $696 = (((($exc)) + 444|0) + ($aRange$off$i$i85<<3)|0);
      $697 = HEAP32[$696>>2]|0;
      $698 = ($697|0)==(0|0);
      if ($698) {
       HEAP32[$34>>2] = 138;
       break;
      }
      $699 = (((((($exc)) + 444|0) + ($aRange$off$i$i85<<3)|0)) + 4|0);
      $700 = HEAP32[$699>>2]|0;
      $701 = ($700|0)<($694|0);
      if ($701) {
       HEAP32[$34>>2] = 131;
       break;
      } else {
       HEAP32[$23>>2] = $697;
       HEAP32[$26>>2] = $700;
       HEAP32[$22>>2] = $694;
       HEAP32[$67>>2] = $692;
       break;
      }
     }
    } while(0);
    HEAP8[$33>>0] = 0;
    label = 821;
    break;
   }
   case 44:  {
    $$val26 = HEAP32[$172>>2]|0;
    $702 = HEAP32[$89>>2]|0;
    $703 = HEAP32[$90>>2]|0;
    $704 = (($702) + (($703*24)|0)|0);
    $705 = ($703|0)>(0);
    L357: do {
     if ($705) {
      $rec$01$i = $702;
      while(1) {
       $706 = ((($rec$01$i)) + 12|0);
       $707 = HEAP32[$706>>2]|0;
       $708 = ($707|0)==($$val26|0);
       if ($708) {
        $rec$0$lcssa$i = $rec$01$i;
        break L357;
       }
       $709 = ((($rec$01$i)) + 24|0);
       $710 = ($709>>>0)<($704>>>0);
       if ($710) {
        $rec$01$i = $709;
       } else {
        $rec$0$lcssa$i = $709;
        break;
       }
      }
     } else {
      $rec$0$lcssa$i = $702;
     }
    } while(0);
    $711 = ($rec$0$lcssa$i|0)==($704|0);
    if ($711) {
     $712 = HEAP32[$91>>2]|0;
     $713 = ($703>>>0)<($712>>>0);
     if (!($713)) {
      label = 215;
      break L12;
     }
     $714 = (($703) + 1)|0;
     HEAP32[$90>>2] = $714;
    }
    $715 = ($$val26>>>0)>(65535);
    if ($715) {
     label = 218;
     break L12;
    }
    $716 = HEAP32[$67>>2]|0;
    HEAP32[$rec$0$lcssa$i>>2] = $716;
    $717 = $$val26 & 65535;
    $718 = ((($rec$0$lcssa$i)) + 12|0);
    HEAP32[$718>>2] = $717;
    $719 = HEAP32[$22>>2]|0;
    $720 = (($719) + 1)|0;
    $721 = ((($rec$0$lcssa$i)) + 4|0);
    HEAP32[$721>>2] = $720;
    $722 = ((($rec$0$lcssa$i)) + 16|0);
    HEAP8[$722>>0] = 1;
    $723 = ((($rec$0$lcssa$i)) + 17|0);
    HEAP8[$723>>0] = 0;
    $724 = ((($rec$0$lcssa$i)) + 20|0);
    HEAP32[$724>>2] = 0;
    $725 = HEAP32[$88>>2]|0;
    $726 = ($725>>>0)<($$val26>>>0);
    if ($726) {
     HEAP32[$88>>2] = $717;
    }
    $$pre4$i = HEAP32[$26>>2]|0;
    $728 = $$pre$pre$i;$729 = $719;
    L370: while(1) {
     $727 = (($728) + ($729))|0;
     HEAP32[$22>>2] = $727;
     $730 = ($727|0)<($$pre4$i|0);
     if (!($730)) {
      label = 227;
      break L12;
     }
     $731 = (($129) + ($727)|0);
     $732 = HEAP8[$731>>0]|0;
     HEAP8[$24>>0] = $732;
     $733 = $732&255;
     $734 = (593353 + ($733)|0);
     $735 = HEAP8[$734>>0]|0;
     $736 = $735 << 24 >> 24;
     HEAP32[$25>>2] = $736;
     $737 = $733 | 1;
     $738 = ($737|0)==(65);
     if ($738) {
      $739 = (($727) + 1)|0;
      $740 = ($739|0)<($$pre4$i|0);
      if (!($740)) {
       label = 227;
       break L12;
      }
      $741 = (($129) + ($739)|0);
      $742 = HEAP8[$741>>0]|0;
      $743 = $742&255;
      $744 = Math_imul($743, $736)|0;
      $745 = (2 - ($744))|0;
      HEAP32[$25>>2] = $745;
      $747 = $745;
     } else {
      $747 = $736;
     }
     $746 = (($747) + ($727))|0;
     $748 = ($746|0)>($$pre4$i|0);
     if ($748) {
      label = 227;
      break L12;
     }
     switch ($733|0) {
     case 44: case 137:  {
      label = 229;
      break L12;
      break;
     }
     case 45:  {
      $$lcssa620 = $727;
      break L370;
      break;
     }
     default: {
      $728 = $747;$729 = $727;
     }
     }
    }
    $749 = ((($rec$0$lcssa$i)) + 8|0);
    HEAP32[$749>>2] = $$lcssa620;
    label = 821;
    break;
   }
   case 45:  {
    $750 = HEAP32[$64>>2]|0;
    $751 = ($750|0)<(1);
    if ($751) {
     label = 232;
     break L12;
    }
    $752 = (($750) + -1)|0;
    HEAP32[$64>>2] = $752;
    $753 = HEAP32[$66>>2]|0;
    $754 = (((($753) + ($752<<4)|0)) + 8|0);
    $755 = HEAP32[$754>>2]|0;
    $756 = (($755) + -1)|0;
    HEAP32[$754>>2] = $756;
    HEAP8[$33>>0] = 0;
    $757 = ($755|0)>(1);
    if ($757) {
     HEAP32[$64>>2] = $750;
     $758 = (((($753) + ($752<<4)|0)) + 12|0);
     $759 = HEAP32[$758>>2]|0;
     $760 = ((($759)) + 4|0);
     $761 = HEAP32[$760>>2]|0;
     HEAP32[$22>>2] = $761;
     label = 821;
     break L30;
    }
    $762 = (($753) + ($752<<4)|0);
    $763 = HEAP32[$762>>2]|0;
    $764 = (((($753) + ($752<<4)|0)) + 4|0);
    $765 = HEAP32[$764>>2]|0;
    $aRange$off$i$i92 = (($763) + -1)|0;
    $766 = ($aRange$off$i$i92>>>0)>(2);
    if ($766) {
     label = 236;
     break L12;
    }
    $767 = (((($exc)) + 444|0) + ($aRange$off$i$i92<<3)|0);
    $768 = HEAP32[$767>>2]|0;
    $769 = ($768|0)==(0|0);
    if ($769) {
     label = 238;
     break L12;
    }
    $770 = (((((($exc)) + 444|0) + ($aRange$off$i$i92<<3)|0)) + 4|0);
    $771 = HEAP32[$770>>2]|0;
    $772 = ($771|0)<($765|0);
    if ($772) {
     label = 240;
     break L12;
    }
    HEAP32[$23>>2] = $768;
    HEAP32[$26>>2] = $771;
    HEAP32[$22>>2] = $765;
    HEAP32[$67>>2] = $763;
    label = 821;
    break;
   }
   case 47: case 46:  {
    $$val27 = HEAP32[$172>>2]|0;
    $773 = $$val27&65535;
    $774 = $$val27 & 65535;
    $775 = HEAP16[$41>>1]|0;
    $776 = $775&65535;
    $777 = ($774>>>0)<($776>>>0);
    if (!($777)) {
     $778 = HEAP8[$29>>0]|0;
     $779 = ($778<<24>>24)==(0);
     if ($779) {
      label = 821;
      break L30;
     } else {
      label = 244;
      break L12;
     }
    }
    $780 = $132 & 1;
    $781 = ($780<<24>>24)==(0);
    if ($781) {
     $distance$0$i106 = 0;
    } else {
     $782 = HEAP32[$51>>2]|0;
     $783 = HEAP32[$52>>2]|0;
     $784 = (($783) + ($774<<3)|0);
     $785 = HEAP32[$784>>2]|0;
     $786 = (((($783) + ($774<<3)|0)) + 4|0);
     $787 = HEAP32[$786>>2]|0;
     $788 = (FUNCTION_TABLE_iiii[$782 & 8191]($exc,$785,$787)|0);
     $789 = HEAP32[$$pre$phiZ2D>>2]|0;
     $790 = HEAP32[$92>>2]|0;
     $791 = (FUNCTION_TABLE_iiii[$789 & 8191]($exc,$788,$790)|0);
     $792 = (($791) - ($788))|0;
     $distance$0$i106 = $792;
    }
    $793 = HEAP32[$54>>2]|0;
    FUNCTION_TABLE_viiii[$793 & 8191]($exc,$80,$773,$distance$0$i106);
    HEAP16[$79>>1] = $773;
    HEAP16[$56>>1] = $773;
    label = 821;
    break;
   }
   case 49: case 48:  {
    $794 = HEAP16[$93>>1]|0;
    $795 = ($794<<16>>16)==(0);
    L395: do {
     if (!($795)) {
      $796 = $132 & 1;
      $797 = ($796<<24>>24)==(0);
      if ($797) {
       $801 = HEAP32[$94>>2]|0;
       $802 = ((($801)) + 4|0);
       HEAP32[$V$i>>2] = $802;
       $803 = HEAP32[$95>>2]|0;
       $804 = ((($803)) + 4|0);
       HEAP32[$96>>2] = $804;
       $805 = HEAP32[$97>>2]|0;
       $806 = ((($805)) + 4|0);
       HEAP32[$98>>2] = $806;
       $807 = $802;
       $808 = $804;
       $V$idx$val14$in$i = $807;$V$idx2$val17$in$i = $808;$mask$0$i = 16;
      } else {
       $798 = HEAP32[$94>>2]|0;
       HEAP32[$V$i>>2] = $798;
       $799 = HEAP32[$95>>2]|0;
       HEAP32[$96>>2] = $799;
       $800 = HEAP32[$97>>2]|0;
       HEAP32[$98>>2] = $800;
       $V$idx$val14$in$i = $798;$V$idx2$val17$in$i = $799;$mask$0$i = 8;
      }
      $V$idx$val14$i = $V$idx$val14$in$i;
      $V$idx2$val17$i = $V$idx2$val17$in$i;
      $809 = HEAP16[$99>>1]|0;
      $810 = $809&65535;
      HEAP32[$100>>2] = $810;
      $820 = $809;$contour$0$i = 0;$point$0$i = 0;
      while(1) {
       $811 = $contour$0$i << 16 >> 16;
       $812 = HEAP32[$101>>2]|0;
       $813 = (($812) + ($811<<1)|0);
       $814 = HEAP16[$813>>1]|0;
       $815 = $814&65535;
       $816 = HEAP16[$102>>1]|0;
       $817 = $816&65535;
       $818 = (($815) - ($817))|0;
       $819 = $820&65535;
       $821 = ($818>>>0)<($819>>>0);
       $822 = (($819) + -1)|0;
       $$$i107 = $821 ? $818 : $822;
       $823 = ($point$0$i>>>0)>($$$i107>>>0);
       L403: do {
        if ($823) {
         $point$3$i = $point$0$i;
        } else {
         $824 = HEAP32[$103>>2]|0;
         $point$17$i = $point$0$i;
         while(1) {
          $825 = (($824) + ($point$17$i)|0);
          $826 = HEAP8[$825>>0]|0;
          $827 = $826 & $mask$0$i;
          $828 = ($827<<24>>24)==(0);
          if (!($828)) {
           $point$17$i$lcssa = $point$17$i;
           break;
          }
          $829 = (($point$17$i) + 1)|0;
          $830 = ($829>>>0)>($$$i107>>>0);
          if ($830) {
           $point$3$i = $829;
           break L403;
          } else {
           $point$17$i = $829;
          }
         }
         $cur_touched$0$ph$i = $point$17$i$lcssa;
         L409: while(1) {
          $point$2$in$i = $cur_touched$0$ph$i;
          while(1) {
           $point$2$i = (($point$2$in$i) + 1)|0;
           $831 = ($point$2$i>>>0)>($$$i107>>>0);
           if ($831) {
            $cur_touched$0$ph$i$lcssa = $cur_touched$0$ph$i;$point$2$i$lcssa = $point$2$i;
            break L409;
           }
           $832 = HEAP32[$103>>2]|0;
           $833 = (($832) + ($point$2$i)|0);
           $834 = HEAP8[$833>>0]|0;
           $835 = $834 & $mask$0$i;
           $836 = ($835<<24>>24)==(0);
           if ($836) {
            $point$2$in$i = $point$2$i;
           } else {
            $point$2$i$lcssa617 = $point$2$i;$point$2$in$i$lcssa616 = $point$2$in$i;
            break;
           }
          }
          $837 = (($cur_touched$0$ph$i) + 1)|0;
          __iup_worker_interpolate($V$i,$837,$point$2$in$i$lcssa616,$cur_touched$0$ph$i,$point$2$i$lcssa617);
          $cur_touched$0$ph$i = $point$2$i$lcssa617;
         }
         $838 = ($cur_touched$0$ph$i$lcssa|0)==($point$17$i$lcssa|0);
         if (!($838)) {
          $855 = (($cur_touched$0$ph$i$lcssa) + 1)|0;
          $856 = $855 & 65535;
          __iup_worker_interpolate($V$i,$856,$$$i107,$cur_touched$0$ph$i$lcssa,$point$17$i$lcssa);
          $857 = ($point$17$i$lcssa|0)==(0);
          if ($857) {
           $point$3$i = $point$2$i$lcssa;
           break;
          }
          $858 = (($point$17$i$lcssa) + -1)|0;
          __iup_worker_interpolate($V$i,$point$0$i,$858,$cur_touched$0$ph$i$lcssa,$point$17$i$lcssa);
          $point$3$i = $point$2$i$lcssa;
          break;
         }
         $839 = (($V$idx2$val17$i) + ($point$17$i$lcssa<<3)|0);
         $840 = HEAP32[$839>>2]|0;
         $841 = (($V$idx$val14$i) + ($point$17$i$lcssa<<3)|0);
         $842 = HEAP32[$841>>2]|0;
         $843 = (($840) - ($842))|0;
         $844 = ($840|0)==($842|0);
         if ($844) {
          $point$3$i = $point$2$i$lcssa;
         } else {
          $845 = ($point$0$i>>>0)<($point$17$i$lcssa>>>0);
          if ($845) {
           $i$04$i$i = $point$0$i;
           while(1) {
            $847 = (($V$idx2$val17$i) + ($i$04$i$i<<3)|0);
            $848 = HEAP32[$847>>2]|0;
            $849 = (($848) + ($843))|0;
            HEAP32[$847>>2] = $849;
            $850 = (($i$04$i$i) + 1)|0;
            $exitcond$i$i = ($850|0)==($point$17$i$lcssa|0);
            if ($exitcond$i$i) {
             break;
            } else {
             $i$04$i$i = $850;
            }
           }
          }
          $i$12$i$i = (($point$17$i$lcssa) + 1)|0;
          $846 = ($i$12$i$i>>>0)>($$$i107>>>0);
          if ($846) {
           $point$3$i = $point$2$i$lcssa;
          } else {
           $i$13$i$i = $i$12$i$i;
           while(1) {
            $851 = (($V$idx2$val17$i) + ($i$13$i$i<<3)|0);
            $852 = HEAP32[$851>>2]|0;
            $853 = (($852) + ($843))|0;
            HEAP32[$851>>2] = $853;
            $i$1$i$i = (($i$13$i$i) + 1)|0;
            $854 = ($i$1$i$i>>>0)>($$$i107>>>0);
            if ($854) {
             $point$3$i = $point$2$i$lcssa;
             break;
            } else {
             $i$13$i$i = $i$1$i$i;
            }
           }
          }
         }
        }
       } while(0);
       $859 = (($contour$0$i) + 1)<<16>>16;
       $860 = HEAP16[$93>>1]|0;
       $861 = ($859<<16>>16)<($860<<16>>16);
       if (!($861)) {
        break L395;
       }
       $$pre$i110 = HEAP16[$99>>1]|0;
       $820 = $$pre$i110;$contour$0$i = $859;$point$0$i = $point$3$i;
      }
     }
    } while(0);
    label = 821;
    break;
   }
   case 51: case 50:  {
    $862 = HEAP32[$27>>2]|0;
    $863 = HEAP32[$85>>2]|0;
    $864 = ($862|0)<($863|0);
    L431: do {
     if ($864) {
      $865 = HEAP8[$29>>0]|0;
      $866 = ($865<<24>>24)==(0);
      if ($866) {
       label = 287;
      } else {
       HEAP32[$34>>2] = 134;
       label = 287;
      }
     } else {
      $867 = (_Compute_Point_Displacement($exc,$dx$i,$dy$i,$zp$i,$refp$i)|0);
      $868 = ($867<<24>>24)==(0);
      if ($868) {
       $$pr$i = HEAP32[$85>>2]|0;
       $869 = ($$pr$i|0)>(0);
       if ($869) {
        $870 = HEAP32[$dy$i>>2]|0;
        $871 = HEAP32[$dx$i>>2]|0;
        while(1) {
         $872 = HEAP32[$28>>2]|0;
         $873 = (($872) + -1)|0;
         HEAP32[$28>>2] = $873;
         $874 = HEAP32[$30>>2]|0;
         $875 = (($874) + ($873<<2)|0);
         $876 = HEAP32[$875>>2]|0;
         $877 = $876 & 65535;
         $878 = HEAP16[$74>>1]|0;
         $879 = $878&65535;
         $880 = ($877>>>0)<($879>>>0);
         if ($880) {
          $883 = HEAP16[$46>>1]|0;
          $884 = ($883<<16>>16)==(0);
          if (!($884)) {
           $885 = HEAP32[$75>>2]|0;
           $886 = (($885) + ($877<<3)|0);
           $887 = HEAP32[$886>>2]|0;
           $888 = (($887) + ($871))|0;
           HEAP32[$886>>2] = $888;
           $889 = HEAP32[$78>>2]|0;
           $890 = (($889) + ($877)|0);
           $891 = HEAP8[$890>>0]|0;
           $892 = $891&255;
           $893 = $892 | 8;
           $894 = $893&255;
           HEAP8[$890>>0] = $894;
          }
          $895 = HEAP16[$48>>1]|0;
          $896 = ($895<<16>>16)==(0);
          if (!($896)) {
           $897 = HEAP32[$75>>2]|0;
           $898 = (((($897) + ($877<<3)|0)) + 4|0);
           $899 = HEAP32[$898>>2]|0;
           $900 = (($899) + ($870))|0;
           HEAP32[$898>>2] = $900;
           $901 = HEAP32[$78>>2]|0;
           $902 = (($901) + ($877)|0);
           $903 = HEAP8[$902>>0]|0;
           $904 = $903&255;
           $905 = $904 | 16;
           $906 = $905&255;
           HEAP8[$902>>0] = $906;
          }
         } else {
          $881 = HEAP8[$29>>0]|0;
          $882 = ($881<<24>>24)==(0);
          if (!($882)) {
           break;
          }
         }
         $907 = HEAP32[$85>>2]|0;
         $908 = (($907) + -1)|0;
         HEAP32[$85>>2] = $908;
         $909 = ($907|0)>(1);
         if (!($909)) {
          label = 287;
          break L431;
         }
        }
        HEAP32[$34>>2] = 134;
       } else {
        label = 287;
       }
      }
     }
    } while(0);
    if ((label|0) == 287) {
     label = 0;
     HEAP32[$85>>2] = 1;
     $910 = HEAP32[$28>>2]|0;
     HEAP32[$31>>2] = $910;
    }
    label = 821;
    break;
   }
   case 53: case 52:  {
    $$val28 = HEAP32[$172>>2]|0;
    $911 = HEAP16[$84>>1]|0;
    $912 = ($911<<16>>16)==(0);
    if ($912) {
     $916 = 1;
    } else {
     $913 = HEAP16[$107>>1]|0;
     $916 = $913;
    }
    $sext$i116 = $$val28 << 16;
    $914 = $sext$i116 >> 16;
    $915 = $916 << 16 >> 16;
    $917 = ($914>>>0)<($915>>>0);
    if ($917) {
     $920 = (_Compute_Point_Displacement($exc,$dx$i114,$dy$i115,$zp$i112,$refp$i113)|0);
     $921 = ($920<<24>>24)==(0);
     if ($921) {
      $922 = ($914|0)==(0);
      if ($922) {
       $start$0$i = 0;
      } else {
       $923 = (($914) + -1)|0;
       $924 = HEAP32[$105>>2]|0;
       $925 = (($924) + ($923<<1)|0);
       $926 = HEAP16[$925>>1]|0;
       $927 = $926&65535;
       $928 = (($927) + 1)|0;
       $929 = HEAP16[$106>>1]|0;
       $930 = $929&65535;
       $931 = (($928) - ($930))|0;
       $932 = $931&65535;
       $start$0$i = $932;
      }
      $933 = HEAP16[$84>>1]|0;
      $934 = ($933<<16>>16)==(0);
      if ($934) {
       $935 = HEAP16[$74>>1]|0;
       $limit$0$i = $935;
      } else {
       $936 = HEAP32[$105>>2]|0;
       $937 = (($936) + ($914<<1)|0);
       $938 = HEAP16[$937>>1]|0;
       $939 = $938&65535;
       $940 = HEAP16[$106>>1]|0;
       $941 = $940&65535;
       $942 = (($939) + 1)|0;
       $943 = (($942) - ($941))|0;
       $944 = $943&65535;
       $limit$0$i = $944;
      }
      $945 = ($start$0$i&65535)<($limit$0$i&65535);
      if ($945) {
       $946 = HEAP32[$104>>2]|0;
       $947 = HEAP16[$refp$i113>>1]|0;
       $948 = HEAP32[$dy$i115>>2]|0;
       $949 = HEAP32[$dx$i114>>2]|0;
       $950 = $start$0$i&65535;
       $951 = $947&65535;
       $indvars$iv$i = $950;
       while(1) {
        $952 = HEAP32[$75>>2]|0;
        $953 = ($946|0)==($952|0);
        $954 = ($951|0)==($indvars$iv$i|0);
        $or$cond$i118 = $954 & $953;
        if (!($or$cond$i118)) {
         $955 = HEAP16[$46>>1]|0;
         $956 = ($955<<16>>16)==(0);
         if (!($956)) {
          $957 = (($952) + ($indvars$iv$i<<3)|0);
          $958 = HEAP32[$957>>2]|0;
          $959 = (($958) + ($949))|0;
          HEAP32[$957>>2] = $959;
          $960 = HEAP32[$78>>2]|0;
          $961 = (($960) + ($indvars$iv$i)|0);
          $962 = HEAP8[$961>>0]|0;
          $963 = $962&255;
          $964 = $963 | 8;
          $965 = $964&255;
          HEAP8[$961>>0] = $965;
         }
         $966 = HEAP16[$48>>1]|0;
         $967 = ($966<<16>>16)==(0);
         if (!($967)) {
          $968 = HEAP32[$75>>2]|0;
          $969 = (((($968) + ($indvars$iv$i<<3)|0)) + 4|0);
          $970 = HEAP32[$969>>2]|0;
          $971 = (($970) + ($948))|0;
          HEAP32[$969>>2] = $971;
          $972 = HEAP32[$78>>2]|0;
          $973 = (($972) + ($indvars$iv$i)|0);
          $974 = HEAP8[$973>>0]|0;
          $975 = $974&255;
          $976 = $975 | 16;
          $977 = $976&255;
          HEAP8[$973>>0] = $977;
         }
        }
        $indvars$iv$next$i = (($indvars$iv$i) + 1)|0;
        $lftr$wideiv$i = $indvars$iv$next$i&65535;
        $exitcond$i = ($lftr$wideiv$i<<16>>16)==($limit$0$i<<16>>16);
        if ($exitcond$i) {
         break;
        } else {
         $indvars$iv$i = $indvars$iv$next$i;
        }
       }
      }
     }
    } else {
     $918 = HEAP8[$29>>0]|0;
     $919 = ($918<<24>>24)==(0);
     if (!($919)) {
      HEAP32[$34>>2] = 134;
     }
    }
    label = 821;
    break;
   }
   case 55: case 54:  {
    $$val29 = HEAP32[$172>>2]|0;
    $978 = ($$val29>>>0)>(1);
    L479: do {
     if ($978) {
      $979 = HEAP8[$29>>0]|0;
      $980 = ($979<<24>>24)==(0);
      if (!($980)) {
       HEAP32[$34>>2] = 134;
      }
     } else {
      $981 = (_Compute_Point_Displacement($exc,$dx$i123,$dy$i124,$zp$i121,$refp$i122)|0);
      $982 = ($981<<24>>24)==(0);
      if ($982) {
       $983 = HEAP16[$84>>1]|0;
       switch ($983<<16>>16) {
       case 0:  {
        $984 = HEAP16[$74>>1]|0;
        $limit$0$i125 = $984;
        break;
       }
       case 1:  {
        $985 = HEAP16[$107>>1]|0;
        $986 = ($985<<16>>16)>(0);
        if (!($986)) {
         break L479;
        }
        $987 = $985 << 16 >> 16;
        $988 = (($987) + -1)|0;
        $989 = HEAP32[$105>>2]|0;
        $990 = (($989) + ($988<<1)|0);
        $991 = HEAP16[$990>>1]|0;
        $992 = $991&65535;
        $993 = (($992) + 1)|0;
        $994 = $993&65535;
        $limit$0$i125 = $994;
        break;
       }
       default: {
        break L479;
       }
       }
       $995 = ($limit$0$i125<<16>>16)==(0);
       if (!($995)) {
        $996 = HEAP32[$108>>2]|0;
        $997 = HEAP32[$75>>2]|0;
        $998 = ($996|0)==($997|0);
        $999 = HEAP32[$dy$i124>>2]|0;
        $1000 = HEAP32[$dx$i123>>2]|0;
        if ($998) {
         $1015 = HEAP16[$refp$i122>>1]|0;
         $1016 = $1015&65535;
         $indvars$iv$i128 = 0;
         while(1) {
          $1017 = ($1016|0)==($indvars$iv$i128|0);
          if (!($1017)) {
           $1018 = HEAP16[$46>>1]|0;
           $1019 = ($1018<<16>>16)==(0);
           if (!($1019)) {
            $1020 = (($996) + ($indvars$iv$i128<<3)|0);
            $1021 = HEAP32[$1020>>2]|0;
            $1022 = (($1021) + ($1000))|0;
            HEAP32[$1020>>2] = $1022;
           }
           $1023 = HEAP16[$48>>1]|0;
           $1024 = ($1023<<16>>16)==(0);
           if (!($1024)) {
            $1025 = (((($996) + ($indvars$iv$i128<<3)|0)) + 4|0);
            $1026 = HEAP32[$1025>>2]|0;
            $1027 = (($1026) + ($999))|0;
            HEAP32[$1025>>2] = $1027;
           }
          }
          $indvars$iv$next$i129 = (($indvars$iv$i128) + 1)|0;
          $lftr$wideiv436 = $indvars$iv$next$i129&65535;
          $exitcond437 = ($lftr$wideiv436<<16>>16)==($limit$0$i125<<16>>16);
          if ($exitcond437) {
           break L479;
          } else {
           $indvars$iv$i128 = $indvars$iv$next$i129;
          }
         }
        }
        $$pre$i127 = HEAP16[$46>>1]|0;
        $1001 = ($$pre$i127<<16>>16)==(0);
        $1002 = HEAP16[$48>>1]|0;
        $1003 = ($1002<<16>>16)==(0);
        if ($1001) {
         if ($1003) {
          break;
         } else {
          $indvars$iv4$i$us = 0;
         }
         while(1) {
          $1004 = (((($997) + ($indvars$iv4$i$us<<3)|0)) + 4|0);
          $1005 = HEAP32[$1004>>2]|0;
          $1006 = (($1005) + ($999))|0;
          HEAP32[$1004>>2] = $1006;
          $indvars$iv$next5$i$us = (($indvars$iv4$i$us) + 1)|0;
          $lftr$wideiv432 = $indvars$iv$next5$i$us&65535;
          $exitcond433 = ($lftr$wideiv432<<16>>16)==($limit$0$i125<<16>>16);
          if ($exitcond433) {
           break L479;
          } else {
           $indvars$iv4$i$us = $indvars$iv$next5$i$us;
          }
         }
        }
        if ($1003) {
         $indvars$iv4$i$us378 = 0;
         while(1) {
          $1012 = (($997) + ($indvars$iv4$i$us378<<3)|0);
          $1013 = HEAP32[$1012>>2]|0;
          $1014 = (($1013) + ($1000))|0;
          HEAP32[$1012>>2] = $1014;
          $indvars$iv$next5$i$us380 = (($indvars$iv4$i$us378) + 1)|0;
          $lftr$wideiv430 = $indvars$iv$next5$i$us380&65535;
          $exitcond431 = ($lftr$wideiv430<<16>>16)==($limit$0$i125<<16>>16);
          if ($exitcond431) {
           break L479;
          } else {
           $indvars$iv4$i$us378 = $indvars$iv$next5$i$us380;
          }
         }
        }
        $1007 = HEAP32[$997>>2]|0;
        $1008 = (($1007) + ($1000))|0;
        HEAP32[$997>>2] = $1008;
        $1009 = ((($997)) + 4|0);
        $1010 = HEAP32[$1009>>2]|0;
        $1011 = (($1010) + ($999))|0;
        HEAP32[$1009>>2] = $1011;
        $exitcond551 = ($limit$0$i125<<16>>16)==(1);
        if (!($exitcond551)) {
         $$pre452 = HEAP32[$75>>2]|0;
         $indvars$iv$next5$i552 = 1;
         while(1) {
          $1028 = (($997) + ($indvars$iv$next5$i552<<3)|0);
          $1029 = HEAP32[$1028>>2]|0;
          $1030 = (($1029) + ($1000))|0;
          HEAP32[$1028>>2] = $1030;
          $1031 = (((($$pre452) + ($indvars$iv$next5$i552<<3)|0)) + 4|0);
          $1032 = HEAP32[$1031>>2]|0;
          $1033 = (($1032) + ($999))|0;
          HEAP32[$1031>>2] = $1033;
          $indvars$iv$next5$i = (($indvars$iv$next5$i552) + 1)|0;
          $lftr$wideiv = $indvars$iv$next5$i&65535;
          $exitcond = ($lftr$wideiv<<16>>16)==($limit$0$i125<<16>>16);
          if ($exitcond) {
           break;
          } else {
           $indvars$iv$next5$i552 = $indvars$iv$next5$i;
          }
         }
        }
       }
      }
     }
    } while(0);
    label = 821;
    break;
   }
   case 56:  {
    $1034 = HEAP32[$27>>2]|0;
    $1035 = HEAP32[$85>>2]|0;
    $1036 = ($1034|0)>($1035|0);
    L516: do {
     if ($1036) {
      $1039 = HEAP32[$172>>2]|0;
      $1040 = HEAP16[$46>>1]|0;
      $1041 = $1040 << 16 >> 16;
      $1042 = $1041 ^ $1039;
      $1043 = ($1039|0)<(0);
      $1044 = (0 - ($1039))|0;
      $$a$i$i134 = $1043 ? $1044 : $1039;
      $1045 = ($1040<<16>>16)<(0);
      $1046 = (0 - ($1041))|0;
      $$01$i$i135 = $1045 ? $1046 : $1041;
      $1047 = $$a$i$i134 >>> 16;
      $1048 = $$a$i$i134 & 65535;
      $1049 = Math_imul($1048, $$01$i$i135)|0;
      $1050 = Math_imul($1047, $$01$i$i135)|0;
      $1051 = $1050 >>> 16;
      $1052 = $1050 << 16;
      $1053 = $1052 | 8192;
      $1054 = (($1053) + ($1049))|0;
      $1055 = ($1054>>>0)<($1053>>>0);
      $1056 = $1055&1;
      $$$i$i136 = (($1056) + ($1051))|0;
      $1057 = $1054 >>> 14;
      $1058 = $$$i$i136 << 18;
      $1059 = $1058 | $1057;
      $1060 = ($1042|0)>(-1);
      $1061 = (0 - ($1059))|0;
      $1062 = $1060 ? $1059 : $1061;
      $1063 = HEAP16[$48>>1]|0;
      $1064 = $1063 << 16 >> 16;
      $1065 = $1064 ^ $1039;
      $1066 = ($1063<<16>>16)<(0);
      $1067 = (0 - ($1064))|0;
      $$01$i2$i = $1066 ? $1067 : $1064;
      $1068 = Math_imul($$01$i2$i, $1048)|0;
      $1069 = Math_imul($$01$i2$i, $1047)|0;
      $1070 = $1069 >>> 16;
      $1071 = $1069 << 16;
      $1072 = $1071 | 8192;
      $1073 = (($1072) + ($1068))|0;
      $1074 = ($1073>>>0)<($1072>>>0);
      $1075 = $1074&1;
      $$$i3$i = (($1075) + ($1070))|0;
      $1076 = $1073 >>> 14;
      $1077 = $$$i3$i << 18;
      $1078 = $1077 | $1076;
      $1079 = ($1065|0)>(-1);
      $1080 = (0 - ($1078))|0;
      $1081 = $1079 ? $1078 : $1080;
      $1082 = ($1035|0)>(0);
      if ($1082) {
       $1084 = $168;$1086 = $171;
       while(1) {
        $1083 = (($1084) + -1)|0;
        HEAP32[$28>>2] = $1083;
        $1085 = (($1086) + ($1083<<2)|0);
        $1087 = HEAP32[$1085>>2]|0;
        $1088 = $1087 & 65535;
        $1089 = HEAP16[$74>>1]|0;
        $1090 = $1089&65535;
        $1091 = ($1088>>>0)<($1090>>>0);
        if ($1091) {
         $1094 = HEAP16[$46>>1]|0;
         $1095 = ($1094<<16>>16)==(0);
         if (!($1095)) {
          $1096 = HEAP32[$75>>2]|0;
          $1097 = (($1096) + ($1088<<3)|0);
          $1098 = HEAP32[$1097>>2]|0;
          $1099 = (($1098) + ($1062))|0;
          HEAP32[$1097>>2] = $1099;
          $1100 = HEAP32[$78>>2]|0;
          $1101 = (($1100) + ($1088)|0);
          $1102 = HEAP8[$1101>>0]|0;
          $1103 = $1102&255;
          $1104 = $1103 | 8;
          $1105 = $1104&255;
          HEAP8[$1101>>0] = $1105;
         }
         $1106 = HEAP16[$48>>1]|0;
         $1107 = ($1106<<16>>16)==(0);
         if (!($1107)) {
          $1108 = HEAP32[$75>>2]|0;
          $1109 = (((($1108) + ($1088<<3)|0)) + 4|0);
          $1110 = HEAP32[$1109>>2]|0;
          $1111 = (($1110) + ($1081))|0;
          HEAP32[$1109>>2] = $1111;
          $1112 = HEAP32[$78>>2]|0;
          $1113 = (($1112) + ($1088)|0);
          $1114 = HEAP8[$1113>>0]|0;
          $1115 = $1114&255;
          $1116 = $1115 | 16;
          $1117 = $1116&255;
          HEAP8[$1113>>0] = $1117;
         }
        } else {
         $1092 = HEAP8[$29>>0]|0;
         $1093 = ($1092<<24>>24)==(0);
         if (!($1093)) {
          label = 341;
          break L12;
         }
        }
        $1118 = HEAP32[$85>>2]|0;
        $1119 = (($1118) + -1)|0;
        HEAP32[$85>>2] = $1119;
        $1120 = ($1118|0)>(1);
        $$pre449 = HEAP32[$28>>2]|0;
        if (!($1120)) {
         $1121 = $$pre449;
         break L516;
        }
        $$pre450 = HEAP32[$30>>2]|0;
        $1084 = $$pre449;$1086 = $$pre450;
       }
      } else {
       $1121 = $168;
      }
     } else {
      $1037 = HEAP8[$29>>0]|0;
      $1038 = ($1037<<24>>24)==(0);
      if ($1038) {
       $1121 = $168;
      } else {
       HEAP32[$34>>2] = 134;
       $1121 = $168;
      }
     }
    } while(0);
    HEAP32[$85>>2] = 1;
    HEAP32[$31>>2] = $1121;
    label = 821;
    break;
   }
   case 57:  {
    $1122 = HEAP32[$27>>2]|0;
    $1123 = HEAP32[$85>>2]|0;
    $1124 = ($1122|0)<($1123|0);
    do {
     if ($1124) {
      $1125 = HEAP8[$29>>0]|0;
      $1126 = ($1125<<24>>24)==(0);
      if (!($1126)) {
       HEAP32[$34>>2] = 134;
      }
     } else {
      $1127 = HEAP32[$55>>2]|0;
      $1128 = $1127&65535;
      $1129 = ($1128<<16>>16)==(0);
      $1130 = ($1127>>>0)<(65536);
      $or$cond$i140 = $1130 | $1129;
      if ($or$cond$i140) {
       $1139 = 1;
      } else {
       $1131 = HEAP16[$84>>1]|0;
       $1132 = ($1131<<16>>16)==(0);
       $1139 = $1132;
      }
      $1133 = HEAP16[$56>>1]|0;
      $1134 = $1133&65535;
      $1135 = HEAP16[$41>>1]|0;
      $1136 = ($1133&65535)<($1135&65535);
      if (!($1136)) {
       $1137 = HEAP8[$29>>0]|0;
       $1138 = ($1137<<24>>24)==(0);
       if ($1138) {
        break;
       }
       HEAP32[$34>>2] = 134;
       break;
      }
      $$pn$in$i = $1139 ? $45 : $59;
      $$pn$i = HEAP32[$$pn$in$i>>2]|0;
      $orus_base$0$i = (($$pn$i) + ($1134<<3)|0);
      $1140 = HEAP32[$52>>2]|0;
      $1141 = (($1140) + ($1134<<3)|0);
      $1142 = HEAP16[$57>>1]|0;
      $1143 = $1142&65535;
      $1144 = HEAP16[$38>>1]|0;
      $1145 = ($1142&65535)<($1144&65535);
      if ($1145) {
       do {
        if ($1139) {
         $1146 = HEAP32[$50>>2]|0;
         $1147 = HEAP32[$47>>2]|0;
         $1148 = (($1147) + ($1143<<3)|0);
         $1149 = HEAP32[$1148>>2]|0;
         $1150 = HEAP32[$orus_base$0$i>>2]|0;
         $1151 = (($1149) - ($1150))|0;
         $1152 = (((($1147) + ($1143<<3)|0)) + 4|0);
         $1153 = HEAP32[$1152>>2]|0;
         $1154 = (((($$pn$i) + ($1134<<3)|0)) + 4|0);
         $1155 = HEAP32[$1154>>2]|0;
         $1156 = (($1153) - ($1155))|0;
         $1157 = (FUNCTION_TABLE_iiii[$1146 & 8191]($exc,$1151,$1156)|0);
         $old_range$0$i = $1157;
        } else {
         $1158 = HEAP32[$60>>2]|0;
         $1159 = HEAP32[$61>>2]|0;
         $1160 = ($1158|0)==($1159|0);
         if ($1160) {
          $1161 = HEAP32[$50>>2]|0;
          $1162 = HEAP32[$58>>2]|0;
          $1163 = (($1162) + ($1143<<3)|0);
          $1164 = HEAP32[$1163>>2]|0;
          $1165 = HEAP32[$orus_base$0$i>>2]|0;
          $1166 = (($1164) - ($1165))|0;
          $1167 = (((($1162) + ($1143<<3)|0)) + 4|0);
          $1168 = HEAP32[$1167>>2]|0;
          $1169 = (((($$pn$i) + ($1134<<3)|0)) + 4|0);
          $1170 = HEAP32[$1169>>2]|0;
          $1171 = (($1168) - ($1170))|0;
          $1172 = (FUNCTION_TABLE_iiii[$1161 & 8191]($exc,$1166,$1171)|0);
          $old_range$0$i = $1172;
          break;
         } else {
          $1173 = HEAP32[$58>>2]|0;
          $1174 = (($1173) + ($1143<<3)|0);
          $1175 = HEAP32[$1174>>2]|0;
          $1176 = HEAP32[$orus_base$0$i>>2]|0;
          $1177 = (($1175) - ($1176))|0;
          $1178 = (_FT_MulFix($1177,$1158)|0);
          $1179 = HEAP16[$57>>1]|0;
          $1180 = $1179&65535;
          $1181 = HEAP32[$58>>2]|0;
          $1182 = (((($1181) + ($1180<<3)|0)) + 4|0);
          $1183 = HEAP32[$1182>>2]|0;
          $1184 = (((($$pn$i) + ($1134<<3)|0)) + 4|0);
          $1185 = HEAP32[$1184>>2]|0;
          $1186 = (($1183) - ($1185))|0;
          $1187 = HEAP32[$61>>2]|0;
          $1188 = (_FT_MulFix($1186,$1187)|0);
          $1189 = HEAP32[$50>>2]|0;
          $1190 = (FUNCTION_TABLE_iiii[$1189 & 8191]($exc,$1178,$1188)|0);
          $old_range$0$i = $1190;
          break;
         }
        }
       } while(0);
       $1191 = HEAP32[$51>>2]|0;
       $1192 = HEAP16[$57>>1]|0;
       $1193 = $1192&65535;
       $1194 = HEAP32[$49>>2]|0;
       $1195 = (($1194) + ($1193<<3)|0);
       $1196 = HEAP32[$1195>>2]|0;
       $1197 = HEAP32[$1141>>2]|0;
       $1198 = (($1196) - ($1197))|0;
       $1199 = (((($1194) + ($1193<<3)|0)) + 4|0);
       $1200 = HEAP32[$1199>>2]|0;
       $1201 = (((($1140) + ($1134<<3)|0)) + 4|0);
       $1202 = HEAP32[$1201>>2]|0;
       $1203 = (($1200) - ($1202))|0;
       $1204 = (FUNCTION_TABLE_iiii[$1191 & 8191]($exc,$1198,$1203)|0);
       $$pr$pre$i = HEAP32[$85>>2]|0;
       $$pr$i141 = $$pr$pre$i;$cur_range$0$ph$i = $1204;$old_range$1$ph$i = $old_range$0$i;
      } else {
       $$pr$i141 = $1123;$cur_range$0$ph$i = 0;$old_range$1$ph$i = 0;
      }
      $1205 = ($$pr$i141|0)>(0);
      if ($1205) {
       $1206 = (((($$pn$i) + ($1134<<3)|0)) + 4|0);
       $1207 = (((($1140) + ($1134<<3)|0)) + 4|0);
       $1208 = ($old_range$1$ph$i|0)==(0);
       $2797 = $$pr$i141;
       while(1) {
        $1209 = HEAP32[$28>>2]|0;
        $1210 = (($1209) + -1)|0;
        HEAP32[$28>>2] = $1210;
        $1211 = HEAP32[$30>>2]|0;
        $1212 = (($1211) + ($1210<<2)|0);
        $1213 = HEAP32[$1212>>2]|0;
        $1214 = HEAP16[$74>>1]|0;
        $1215 = $1214&65535;
        $1216 = ($1213>>>0)<($1215>>>0);
        if ($1216) {
         do {
          if ($1139) {
           $1219 = HEAP32[$50>>2]|0;
           $1220 = HEAP32[$109>>2]|0;
           $1221 = (($1220) + ($1213<<3)|0);
           $1222 = HEAP32[$1221>>2]|0;
           $1223 = HEAP32[$orus_base$0$i>>2]|0;
           $1224 = (($1222) - ($1223))|0;
           $1225 = (((($1220) + ($1213<<3)|0)) + 4|0);
           $1226 = HEAP32[$1225>>2]|0;
           $1227 = HEAP32[$1206>>2]|0;
           $1228 = (($1226) - ($1227))|0;
           $1229 = (FUNCTION_TABLE_iiii[$1219 & 8191]($exc,$1224,$1228)|0);
           $org_dist$0$i = $1229;
          } else {
           $1230 = HEAP32[$60>>2]|0;
           $1231 = HEAP32[$61>>2]|0;
           $1232 = ($1230|0)==($1231|0);
           if ($1232) {
            $1233 = HEAP32[$50>>2]|0;
            $1234 = HEAP32[$110>>2]|0;
            $1235 = (($1234) + ($1213<<3)|0);
            $1236 = HEAP32[$1235>>2]|0;
            $1237 = HEAP32[$orus_base$0$i>>2]|0;
            $1238 = (($1236) - ($1237))|0;
            $1239 = (((($1234) + ($1213<<3)|0)) + 4|0);
            $1240 = HEAP32[$1239>>2]|0;
            $1241 = HEAP32[$1206>>2]|0;
            $1242 = (($1240) - ($1241))|0;
            $1243 = (FUNCTION_TABLE_iiii[$1233 & 8191]($exc,$1238,$1242)|0);
            $org_dist$0$i = $1243;
            break;
           } else {
            $1244 = HEAP32[$110>>2]|0;
            $1245 = (($1244) + ($1213<<3)|0);
            $1246 = HEAP32[$1245>>2]|0;
            $1247 = HEAP32[$orus_base$0$i>>2]|0;
            $1248 = (($1246) - ($1247))|0;
            $1249 = (_FT_MulFix($1248,$1230)|0);
            $1250 = HEAP32[$110>>2]|0;
            $1251 = (((($1250) + ($1213<<3)|0)) + 4|0);
            $1252 = HEAP32[$1251>>2]|0;
            $1253 = HEAP32[$1206>>2]|0;
            $1254 = (($1252) - ($1253))|0;
            $1255 = HEAP32[$61>>2]|0;
            $1256 = (_FT_MulFix($1254,$1255)|0);
            $1257 = HEAP32[$50>>2]|0;
            $1258 = (FUNCTION_TABLE_iiii[$1257 & 8191]($exc,$1249,$1256)|0);
            $org_dist$0$i = $1258;
            break;
           }
          }
         } while(0);
         $1259 = HEAP32[$51>>2]|0;
         $1260 = HEAP32[$75>>2]|0;
         $1261 = (($1260) + ($1213<<3)|0);
         $1262 = HEAP32[$1261>>2]|0;
         $1263 = HEAP32[$1141>>2]|0;
         $1264 = (($1262) - ($1263))|0;
         $1265 = (((($1260) + ($1213<<3)|0)) + 4|0);
         $1266 = HEAP32[$1265>>2]|0;
         $1267 = HEAP32[$1207>>2]|0;
         $1268 = (($1266) - ($1267))|0;
         $1269 = (FUNCTION_TABLE_iiii[$1259 & 8191]($exc,$1264,$1268)|0);
         $1270 = ($org_dist$0$i|0)==(0);
         $brmerge$i = $1208 | $1270;
         if ($brmerge$i) {
          $new_dist$0$i = $org_dist$0$i;
         } else {
          $1271 = (_FT_MulDiv($org_dist$0$i,$cur_range$0$ph$i,$old_range$1$ph$i)|0);
          $new_dist$0$i = $1271;
         }
         $1272 = HEAP32[$54>>2]|0;
         $1273 = $1213&65535;
         $1274 = (($new_dist$0$i) - ($1269))|0;
         FUNCTION_TABLE_viiii[$1272 & 8191]($exc,$83,$1273,$1274);
         $$pre$i144 = HEAP32[$85>>2]|0;
         $1276 = $$pre$i144;
        } else {
         $1217 = HEAP8[$29>>0]|0;
         $1218 = ($1217<<24>>24)==(0);
         if ($1218) {
          $1276 = $2797;
         } else {
          label = 376;
          break L12;
         }
        }
        $1275 = (($1276) + -1)|0;
        HEAP32[$85>>2] = $1275;
        $1277 = ($1276|0)>(1);
        if ($1277) {
         $2797 = $1275;
        } else {
         break;
        }
       }
      }
     }
    } while(0);
    HEAP32[$85>>2] = 1;
    $1278 = HEAP32[$28>>2]|0;
    HEAP32[$31>>2] = $1278;
    label = 821;
    break;
   }
   case 59: case 58:  {
    $1279 = HEAP32[$172>>2]|0;
    $1280 = $1279&65535;
    $1281 = $1279 & 65535;
    $1282 = HEAP16[$38>>1]|0;
    $1283 = $1282&65535;
    $1284 = ($1281>>>0)<($1283>>>0);
    if ($1284) {
     $1285 = HEAP16[$79>>1]|0;
     $1286 = $1285&65535;
     $1287 = HEAP16[$41>>1]|0;
     $1288 = ($1285&65535)<($1287&65535);
     if ($1288) {
      $1291 = HEAP16[$44>>1]|0;
      $1292 = ($1291<<16>>16)==(0);
      if ($1292) {
       $1293 = HEAP32[$47>>2]|0;
       $1294 = (($1293) + ($1281<<3)|0);
       $1295 = HEAP32[$45>>2]|0;
       $1296 = (($1295) + ($1286<<3)|0);
       $1297 = $1296;
       $1298 = $1297;
       $1299 = HEAP32[$1298>>2]|0;
       $1300 = (($1297) + 4)|0;
       $1301 = $1300;
       $1302 = HEAP32[$1301>>2]|0;
       $1303 = $1294;
       $1304 = $1303;
       HEAP32[$1304>>2] = $1299;
       $1305 = (($1303) + 4)|0;
       $1306 = $1305;
       HEAP32[$1306>>2] = $1302;
       $1307 = HEAP32[$111>>2]|0;
       $1308 = ((($172)) + 4|0);
       $1309 = HEAP32[$1308>>2]|0;
       FUNCTION_TABLE_viiii[$1307 & 8191]($exc,$37,$1280,$1309);
       $1310 = HEAP32[$49>>2]|0;
       $1311 = (($1310) + ($1281<<3)|0);
       $1312 = HEAP32[$47>>2]|0;
       $1313 = (($1312) + ($1281<<3)|0);
       $1314 = $1313;
       $1315 = $1314;
       $1316 = HEAP32[$1315>>2]|0;
       $1317 = (($1314) + 4)|0;
       $1318 = $1317;
       $1319 = HEAP32[$1318>>2]|0;
       $1320 = $1311;
       $1321 = $1320;
       HEAP32[$1321>>2] = $1316;
       $1322 = (($1320) + 4)|0;
       $1323 = $1322;
       HEAP32[$1323>>2] = $1319;
       $$pre$i149 = HEAP16[$79>>1]|0;
       $$pre$phi3$i150Z2D = $1308;$1330 = $$pre$i149;
      } else {
       $$pre2$i147 = ((($172)) + 4|0);
       $$pre$phi3$i150Z2D = $$pre2$i147;$1330 = $1285;
      }
      $1324 = HEAP32[$51>>2]|0;
      $1325 = HEAP32[$49>>2]|0;
      $1326 = (($1325) + ($1281<<3)|0);
      $1327 = HEAP32[$1326>>2]|0;
      $1328 = HEAP32[$52>>2]|0;
      $1329 = $1330&65535;
      $1331 = (($1328) + ($1329<<3)|0);
      $1332 = HEAP32[$1331>>2]|0;
      $1333 = (($1327) - ($1332))|0;
      $1334 = (((($1325) + ($1281<<3)|0)) + 4|0);
      $1335 = HEAP32[$1334>>2]|0;
      $1336 = (((($1328) + ($1329<<3)|0)) + 4|0);
      $1337 = HEAP32[$1336>>2]|0;
      $1338 = (($1335) - ($1337))|0;
      $1339 = (FUNCTION_TABLE_iiii[$1324 & 8191]($exc,$1333,$1338)|0);
      $1340 = HEAP32[$54>>2]|0;
      $1341 = HEAP32[$$pre$phi3$i150Z2D>>2]|0;
      $1342 = (($1341) - ($1339))|0;
      FUNCTION_TABLE_viiii[$1340 & 8191]($exc,$37,$1280,$1342);
      $1343 = HEAP16[$79>>1]|0;
      HEAP16[$56>>1] = $1343;
      HEAP16[$57>>1] = $1280;
      $1344 = HEAP8[$24>>0]|0;
      $1345 = $1344 & 1;
      $1346 = ($1345<<24>>24)==(0);
      if ($1346) {
       label = 821;
       break L30;
      }
      HEAP16[$79>>1] = $1280;
      label = 821;
      break L30;
     }
    }
    $1289 = HEAP8[$29>>0]|0;
    $1290 = ($1289<<24>>24)==(0);
    if ($1290) {
     label = 821;
    } else {
     label = 382;
     break L12;
    }
    break;
   }
   case 60:  {
    $1347 = HEAP32[$27>>2]|0;
    $1348 = HEAP32[$85>>2]|0;
    $1349 = ($1347|0)<($1348|0);
    L583: do {
     if ($1349) {
      label = 390;
     } else {
      $1350 = HEAP16[$79>>1]|0;
      $1351 = HEAP16[$41>>1]|0;
      $1352 = ($1350&65535)<($1351&65535);
      if ($1352) {
       $1355 = ($1348|0)>(0);
       if ($1355) {
        $1357 = $168;$1359 = $171;$2798 = $1348;
        while(1) {
         $1356 = (($1357) + -1)|0;
         HEAP32[$28>>2] = $1356;
         $1358 = (($1359) + ($1356<<2)|0);
         $1360 = HEAP32[$1358>>2]|0;
         $1361 = $1360 & 65535;
         $1362 = HEAP16[$38>>1]|0;
         $1363 = $1362&65535;
         $1364 = ($1361>>>0)<($1363>>>0);
         if ($1364) {
          $1367 = $1360&65535;
          $1368 = HEAP32[$51>>2]|0;
          $1369 = HEAP32[$49>>2]|0;
          $1370 = (($1369) + ($1361<<3)|0);
          $1371 = HEAP32[$1370>>2]|0;
          $1372 = HEAP32[$52>>2]|0;
          $1373 = HEAP16[$79>>1]|0;
          $1374 = $1373&65535;
          $1375 = (($1372) + ($1374<<3)|0);
          $1376 = HEAP32[$1375>>2]|0;
          $1377 = (($1371) - ($1376))|0;
          $1378 = (((($1369) + ($1361<<3)|0)) + 4|0);
          $1379 = HEAP32[$1378>>2]|0;
          $1380 = (((($1372) + ($1374<<3)|0)) + 4|0);
          $1381 = HEAP32[$1380>>2]|0;
          $1382 = (($1379) - ($1381))|0;
          $1383 = (FUNCTION_TABLE_iiii[$1368 & 8191]($exc,$1377,$1382)|0);
          $1384 = HEAP32[$54>>2]|0;
          $1385 = (0 - ($1383))|0;
          FUNCTION_TABLE_viiii[$1384 & 8191]($exc,$37,$1367,$1385);
          $$pre$i154 = HEAP32[$85>>2]|0;
          $1387 = $$pre$i154;
         } else {
          $1365 = HEAP8[$29>>0]|0;
          $1366 = ($1365<<24>>24)==(0);
          if ($1366) {
           $1387 = $2798;
          } else {
           label = 395;
           break L12;
          }
         }
         $1386 = (($1387) + -1)|0;
         HEAP32[$85>>2] = $1386;
         $1388 = ($1387|0)>(1);
         $$pre446 = HEAP32[$28>>2]|0;
         if (!($1388)) {
          $1389 = $$pre446;
          break L583;
         }
         $$pre447 = HEAP32[$30>>2]|0;
         $1357 = $$pre446;$1359 = $$pre447;$2798 = $1386;
        }
       } else {
        $1389 = $168;
       }
      } else {
       label = 390;
      }
     }
    } while(0);
    if ((label|0) == 390) {
     label = 0;
     $1353 = HEAP8[$29>>0]|0;
     $1354 = ($1353<<24>>24)==(0);
     if ($1354) {
      $1389 = $168;
     } else {
      HEAP32[$34>>2] = 134;
      $1389 = $168;
     }
    }
    HEAP32[$85>>2] = 1;
    HEAP32[$31>>2] = $1389;
    label = 821;
    break;
   }
   case 61:  {
    HEAP32[$11>>2] = 2;
    HEAP32[$$pre$phiZ2D>>2] = 4423;
    label = 821;
    break;
   }
   case 63: case 62:  {
    $$val30 = HEAP32[$172>>2]|0;
    $$idx31 = ((($172)) + 4|0);
    $$idx31$val = HEAP32[$$idx31>>2]|0;
    $1390 = HEAP32[$36>>2]|0;
    $1391 = $$val30&65535;
    $1392 = $$val30 & 65535;
    $1393 = HEAP16[$41>>1]|0;
    $1394 = $1393&65535;
    $1395 = ($1392>>>0)<($1394>>>0);
    if ($1395) {
     $1396 = HEAP32[$39>>2]|0;
     $1397 = ($1396>>>0)>($$idx31$val>>>0);
     if ($1397) {
      $1400 = HEAP32[$8>>2]|0;
      $1401 = (FUNCTION_TABLE_iii[$1400 & 8191]($exc,$$idx31$val)|0);
      $1402 = HEAP16[$55>>1]|0;
      $1403 = ($1402<<16>>16)==(0);
      if ($1403) {
       $1404 = HEAP16[$46>>1]|0;
       $1405 = $1404 << 16 >> 16;
       $1406 = $1405 ^ $1401;
       $1407 = ($1401|0)<(0);
       $1408 = (0 - ($1401))|0;
       $$a$i$i158 = $1407 ? $1408 : $1401;
       $1409 = ($1404<<16>>16)<(0);
       $1410 = (0 - ($1405))|0;
       $$01$i$i159 = $1409 ? $1410 : $1405;
       $1411 = $$a$i$i158 >>> 16;
       $1412 = $$a$i$i158 & 65535;
       $1413 = Math_imul($$01$i$i159, $1412)|0;
       $1414 = Math_imul($$01$i$i159, $1411)|0;
       $1415 = $1414 >>> 16;
       $1416 = $1414 << 16;
       $1417 = $1416 | 8192;
       $1418 = (($1417) + ($1413))|0;
       $1419 = ($1418>>>0)<($1417>>>0);
       $1420 = $1419&1;
       $$$i$i160 = (($1420) + ($1415))|0;
       $1421 = $1418 >>> 14;
       $1422 = $$$i$i160 << 18;
       $1423 = $1422 | $1421;
       $1424 = ($1406|0)>(-1);
       $1425 = (0 - ($1423))|0;
       $1426 = $1424 ? $1423 : $1425;
       $1427 = HEAP32[$45>>2]|0;
       $1428 = (($1427) + ($1392<<3)|0);
       HEAP32[$1428>>2] = $1426;
       $1429 = HEAP16[$48>>1]|0;
       $1430 = $1429 << 16 >> 16;
       $1431 = $1430 ^ $1401;
       $1432 = ($1429<<16>>16)<(0);
       $1433 = (0 - ($1430))|0;
       $$01$i2$i161 = $1432 ? $1433 : $1430;
       $1434 = Math_imul($$01$i2$i161, $1412)|0;
       $1435 = Math_imul($$01$i2$i161, $1411)|0;
       $1436 = $1435 >>> 16;
       $1437 = $1435 << 16;
       $1438 = $1437 | 8192;
       $1439 = (($1438) + ($1434))|0;
       $1440 = ($1439>>>0)<($1438>>>0);
       $1441 = $1440&1;
       $$$i3$i162 = (($1441) + ($1436))|0;
       $1442 = $1439 >>> 14;
       $1443 = $$$i3$i162 << 18;
       $1444 = $1443 | $1442;
       $1445 = ($1431|0)>(-1);
       $1446 = (0 - ($1444))|0;
       $1447 = $1445 ? $1444 : $1446;
       $1448 = (((($1427) + ($1392<<3)|0)) + 4|0);
       HEAP32[$1448>>2] = $1447;
       $1449 = HEAP32[$52>>2]|0;
       $1450 = (($1449) + ($1392<<3)|0);
       $1451 = (($1427) + ($1392<<3)|0);
       $1452 = $1451;
       $1453 = $1452;
       $1454 = HEAP32[$1453>>2]|0;
       $1455 = (($1452) + 4)|0;
       $1456 = $1455;
       $1457 = HEAP32[$1456>>2]|0;
       $1458 = $1450;
       $1459 = $1458;
       HEAP32[$1459>>2] = $1454;
       $1460 = (($1458) + 4)|0;
       $1461 = $1460;
       HEAP32[$1461>>2] = $1457;
      }
      $1462 = HEAP32[$51>>2]|0;
      $1463 = HEAP32[$52>>2]|0;
      $1464 = (($1463) + ($1392<<3)|0);
      $1465 = HEAP32[$1464>>2]|0;
      $1466 = (((($1463) + ($1392<<3)|0)) + 4|0);
      $1467 = HEAP32[$1466>>2]|0;
      $1468 = (FUNCTION_TABLE_iiii[$1462 & 8191]($exc,$1465,$1467)|0);
      $1469 = HEAP8[$24>>0]|0;
      $1470 = $1469 & 1;
      $1471 = ($1470<<24>>24)==(0);
      if ($1471) {
       $distance$1$i165 = $1401;
      } else {
       $1472 = (($1401) - ($1468))|0;
       $1473 = ($1472|0)<(0);
       $1474 = (0 - ($1472))|0;
       $1475 = $1473 ? $1474 : $1472;
       $1476 = ($1475|0)>($1390|0);
       $$$i164 = $1476 ? $1468 : $1401;
       $1477 = HEAP32[$$pre$phiZ2D>>2]|0;
       $1478 = HEAP32[$92>>2]|0;
       $1479 = (FUNCTION_TABLE_iiii[$1477 & 8191]($exc,$$$i164,$1478)|0);
       $distance$1$i165 = $1479;
      }
      $1480 = HEAP32[$54>>2]|0;
      $1481 = (($distance$1$i165) - ($1468))|0;
      FUNCTION_TABLE_viiii[$1480 & 8191]($exc,$80,$1391,$1481);
     } else {
      label = 403;
     }
    } else {
     label = 403;
    }
    if ((label|0) == 403) {
     label = 0;
     $1398 = HEAP8[$29>>0]|0;
     $1399 = ($1398<<24>>24)==(0);
     if (!($1399)) {
      HEAP32[$34>>2] = 134;
     }
    }
    HEAP16[$79>>1] = $1391;
    HEAP16[$56>>1] = $1391;
    label = 821;
    break;
   }
   case 64:  {
    $1482 = HEAP32[$22>>2]|0;
    $1483 = (($1482) + 1)|0;
    $1484 = (($129) + ($1483)|0);
    $1485 = HEAP8[$1484>>0]|0;
    $1486 = $1485&255;
    $1487 = (($169) + 1)|0;
    $1488 = HEAP32[$27>>2]|0;
    $1489 = (($1487) - ($1488))|0;
    $1490 = ($1486>>>0)<($1489>>>0);
    if (!($1490)) {
     label = 414;
     break L12;
    }
    $1491 = ($1485<<24>>24)==(0);
    if (!($1491)) {
     $1492 = (($1482) + 2)|0;
     $1493 = (($129) + ($1492)|0);
     $1494 = HEAP8[$1493>>0]|0;
     $1495 = $1494&255;
     HEAP32[$172>>2] = $1495;
     $1496 = ($1485&255)<(2);
     if (!($1496)) {
      $1499 = 2;$1506 = 2;
      while(1) {
       $$pre$i169 = HEAP32[$22>>2]|0;
       $1497 = (($$pre$i169) + 1)|0;
       $1498 = (($1497) + ($1499))|0;
       $1500 = (($129) + ($1498)|0);
       $1501 = HEAP8[$1500>>0]|0;
       $1502 = $1501&255;
       $1503 = (($1499) + -1)|0;
       $1504 = (($172) + ($1503<<2)|0);
       HEAP32[$1504>>2] = $1502;
       $1505 = (($1506) + 1)<<16>>16;
       $1507 = $1505&65535;
       $1508 = ($1507>>>0)>($1486>>>0);
       if ($1508) {
        break;
       } else {
        $1499 = $1507;$1506 = $1505;
       }
      }
     }
    }
    $1509 = HEAP32[$31>>2]|0;
    $1510 = (($1509) + ($1486))|0;
    HEAP32[$31>>2] = $1510;
    label = 821;
    break;
   }
   case 65:  {
    $1511 = HEAP32[$22>>2]|0;
    $1512 = (($1511) + 1)|0;
    $1513 = (($129) + ($1512)|0);
    $1514 = HEAP8[$1513>>0]|0;
    $1515 = $1514&255;
    $1516 = (($169) + 1)|0;
    $1517 = HEAP32[$27>>2]|0;
    $1518 = (($1516) - ($1517))|0;
    $1519 = ($1515>>>0)<($1518>>>0);
    if (!($1519)) {
     label = 418;
     break L12;
    }
    $1520 = (($1511) + 2)|0;
    HEAP32[$22>>2] = $1520;
    $1521 = ($1514<<24>>24)==(0);
    if ($1521) {
     $1540 = $167;
    } else {
     $1523 = $1520;$1535 = 0;$K$01$i171 = 0;
     while(1) {
      $1522 = (($1523) + 2)|0;
      HEAP32[$22>>2] = $1522;
      $1524 = (($129) + ($1523)|0);
      $1525 = HEAP8[$1524>>0]|0;
      $1526 = $1525&255;
      $1527 = $1526 << 8;
      $1528 = (($1523) + 1)|0;
      $1529 = (($129) + ($1528)|0);
      $1530 = HEAP8[$1529>>0]|0;
      $1531 = $1530&255;
      $1532 = $1527 | $1531;
      $sext$i172 = $1532 << 16;
      $1533 = $sext$i172 >> 16;
      $1534 = (($172) + ($1535<<2)|0);
      HEAP32[$1534>>2] = $1533;
      $1536 = (($K$01$i171) + 1)<<16>>16;
      $1537 = $1536&65535;
      $1538 = ($1537>>>0)<($1515>>>0);
      if (!($1538)) {
       break;
      }
      $$pre$i174 = HEAP32[$22>>2]|0;
      $1523 = $$pre$i174;$1535 = $1537;$K$01$i171 = $1536;
     }
     $$pre445 = HEAP32[$31>>2]|0;
     $1540 = $$pre445;
    }
    HEAP8[$33>>0] = 0;
    $1539 = (($1540) + ($1515))|0;
    HEAP32[$31>>2] = $1539;
    label = 821;
    break;
   }
   case 66:  {
    $1541 = HEAP32[$172>>2]|0;
    $1542 = HEAP16[$112>>1]|0;
    $1543 = $1542&65535;
    $1544 = ($1541>>>0)<($1543>>>0);
    if ($1544) {
     $1547 = ((($172)) + 4|0);
     $1548 = HEAP32[$1547>>2]|0;
     $1549 = HEAP32[$113>>2]|0;
     $1550 = (($1549) + ($1541<<2)|0);
     HEAP32[$1550>>2] = $1548;
     label = 821;
     break L30;
    } else {
     $1545 = HEAP8[$29>>0]|0;
     $1546 = ($1545<<24>>24)==(0);
     if ($1546) {
      label = 821;
      break L30;
     } else {
      label = 426;
      break L12;
     }
    }
    break;
   }
   case 67:  {
    $1551 = HEAP32[$172>>2]|0;
    $1552 = HEAP16[$112>>1]|0;
    $1553 = $1552&65535;
    $1554 = ($1551>>>0)<($1553>>>0);
    if ($1554) {
     $1557 = HEAP32[$113>>2]|0;
     $1558 = (($1557) + ($1551<<2)|0);
     $1559 = HEAP32[$1558>>2]|0;
     $storemerge$i177 = $1559;
    } else {
     $1555 = HEAP8[$29>>0]|0;
     $1556 = ($1555<<24>>24)==(0);
     if ($1556) {
      $storemerge$i177 = 0;
     } else {
      label = 430;
      break L12;
     }
    }
    HEAP32[$172>>2] = $storemerge$i177;
    label = 821;
    break;
   }
   case 68:  {
    $1560 = HEAP32[$172>>2]|0;
    $1561 = HEAP32[$39>>2]|0;
    $1562 = ($1560>>>0)<($1561>>>0);
    if ($1562) {
     $1565 = HEAP32[$9>>2]|0;
     $1566 = ((($172)) + 4|0);
     $1567 = HEAP32[$1566>>2]|0;
     FUNCTION_TABLE_viii[$1565 & 8191]($exc,$1560,$1567);
     label = 821;
     break L30;
    } else {
     $1563 = HEAP8[$29>>0]|0;
     $1564 = ($1563<<24>>24)==(0);
     if ($1564) {
      label = 821;
      break L30;
     } else {
      label = 435;
      break L12;
     }
    }
    break;
   }
   case 69:  {
    $1568 = HEAP32[$172>>2]|0;
    $1569 = HEAP32[$39>>2]|0;
    $1570 = ($1568>>>0)<($1569>>>0);
    if ($1570) {
     $1573 = HEAP32[$8>>2]|0;
     $1574 = (FUNCTION_TABLE_iii[$1573 & 8191]($exc,$1568)|0);
     $storemerge$i178 = $1574;
    } else {
     $1571 = HEAP8[$29>>0]|0;
     $1572 = ($1571<<24>>24)==(0);
     if ($1572) {
      $storemerge$i178 = 0;
     } else {
      label = 439;
      break L12;
     }
    }
    HEAP32[$172>>2] = $storemerge$i178;
    label = 821;
    break;
   }
   case 71: case 70:  {
    $1575 = HEAP32[$172>>2]|0;
    $1576 = HEAP16[$74>>1]|0;
    $1577 = $1576&65535;
    $1578 = ($1575>>>0)<($1577>>>0);
    do {
     if ($1578) {
      $1581 = $132 & 1;
      $1582 = ($1581<<24>>24)==(0);
      if ($1582) {
       $1590 = HEAP32[$51>>2]|0;
       $1591 = HEAP32[$75>>2]|0;
       $1592 = (($1591) + ($1575<<3)|0);
       $1593 = HEAP32[$1592>>2]|0;
       $1594 = (((($1591) + ($1575<<3)|0)) + 4|0);
       $1595 = HEAP32[$1594>>2]|0;
       $1596 = (FUNCTION_TABLE_iiii[$1590 & 8191]($exc,$1593,$1595)|0);
       $R$0$i = $1596;
       break;
      } else {
       $1583 = HEAP32[$50>>2]|0;
       $1584 = HEAP32[$109>>2]|0;
       $1585 = (($1584) + ($1575<<3)|0);
       $1586 = HEAP32[$1585>>2]|0;
       $1587 = (((($1584) + ($1575<<3)|0)) + 4|0);
       $1588 = HEAP32[$1587>>2]|0;
       $1589 = (FUNCTION_TABLE_iiii[$1583 & 8191]($exc,$1586,$1588)|0);
       $R$0$i = $1589;
       break;
      }
     } else {
      $1579 = HEAP8[$29>>0]|0;
      $1580 = ($1579<<24>>24)==(0);
      if ($1580) {
       $R$0$i = 0;
      } else {
       HEAP32[$34>>2] = 134;
       $R$0$i = 0;
      }
     }
    } while(0);
    HEAP32[$172>>2] = $R$0$i;
    label = 821;
    break;
   }
   case 72:  {
    $1597 = HEAP32[$172>>2]|0;
    $1598 = $1597 & 65535;
    $1599 = HEAP16[$74>>1]|0;
    $1600 = $1599&65535;
    $1601 = ($1598>>>0)<($1600>>>0);
    if (!($1601)) {
     $1602 = HEAP8[$29>>0]|0;
     $1603 = ($1602<<24>>24)==(0);
     if ($1603) {
      label = 821;
      break L30;
     } else {
      label = 451;
      break L12;
     }
    }
    $1604 = $1597&65535;
    $1605 = HEAP32[$51>>2]|0;
    $1606 = HEAP32[$75>>2]|0;
    $1607 = (($1606) + ($1598<<3)|0);
    $1608 = HEAP32[$1607>>2]|0;
    $1609 = (((($1606) + ($1598<<3)|0)) + 4|0);
    $1610 = HEAP32[$1609>>2]|0;
    $1611 = (FUNCTION_TABLE_iiii[$1605 & 8191]($exc,$1608,$1610)|0);
    $1612 = HEAP32[$54>>2]|0;
    $1613 = ((($172)) + 4|0);
    $1614 = HEAP32[$1613>>2]|0;
    $1615 = (($1614) - ($1611))|0;
    FUNCTION_TABLE_viiii[$1612 & 8191]($exc,$83,$1604,$1615);
    $1616 = HEAP16[$84>>1]|0;
    $1617 = ($1616<<16>>16)==(0);
    if ($1617) {
     $1618 = HEAP32[$109>>2]|0;
     $1619 = (($1618) + ($1598<<3)|0);
     $1620 = HEAP32[$75>>2]|0;
     $1621 = (($1620) + ($1598<<3)|0);
     $1622 = $1621;
     $1623 = $1622;
     $1624 = HEAP32[$1623>>2]|0;
     $1625 = (($1622) + 4)|0;
     $1626 = $1625;
     $1627 = HEAP32[$1626>>2]|0;
     $1628 = $1619;
     $1629 = $1628;
     HEAP32[$1629>>2] = $1624;
     $1630 = (($1628) + 4)|0;
     $1631 = $1630;
     HEAP32[$1631>>2] = $1627;
     label = 821;
    } else {
     label = 821;
    }
    break;
   }
   case 74: case 73:  {
    $1632 = HEAP32[$172>>2]|0;
    $1633 = $1632 & 65535;
    $1634 = HEAP16[$41>>1]|0;
    $1635 = $1634&65535;
    $1636 = ($1633>>>0)<($1635>>>0);
    do {
     if ($1636) {
      $1637 = ((($172)) + 4|0);
      $1638 = HEAP32[$1637>>2]|0;
      $1639 = $1638 & 65535;
      $1640 = HEAP16[$38>>1]|0;
      $1641 = $1640&65535;
      $1642 = ($1639>>>0)<($1641>>>0);
      if ($1642) {
       $1645 = $132 & 1;
       $1646 = ($1645<<24>>24)==(0);
       if (!($1646)) {
        $1647 = HEAP32[$51>>2]|0;
        $1648 = HEAP32[$52>>2]|0;
        $1649 = (($1648) + ($1633<<3)|0);
        $1650 = HEAP32[$1649>>2]|0;
        $1651 = HEAP32[$49>>2]|0;
        $1652 = (($1651) + ($1639<<3)|0);
        $1653 = HEAP32[$1652>>2]|0;
        $1654 = (($1650) - ($1653))|0;
        $1655 = (((($1648) + ($1633<<3)|0)) + 4|0);
        $1656 = HEAP32[$1655>>2]|0;
        $1657 = (((($1651) + ($1639<<3)|0)) + 4|0);
        $1658 = HEAP32[$1657>>2]|0;
        $1659 = (($1656) - ($1658))|0;
        $1660 = (FUNCTION_TABLE_iiii[$1647 & 8191]($exc,$1654,$1659)|0);
        $D$1$i = $1660;
        break;
       }
       $1661 = HEAP32[$55>>2]|0;
       $1662 = $1661&65535;
       $1663 = ($1662<<16>>16)==(0);
       $1664 = ($1661>>>0)<(65536);
       $or$cond$i179 = $1664 | $1663;
       if ($or$cond$i179) {
        $1665 = HEAP32[$45>>2]|0;
        $1666 = HEAP32[$47>>2]|0;
        $1667 = HEAP32[$50>>2]|0;
        $1668 = (($1665) + ($1633<<3)|0);
        $1669 = HEAP32[$1668>>2]|0;
        $1670 = (($1666) + ($1639<<3)|0);
        $1671 = HEAP32[$1670>>2]|0;
        $1672 = (($1669) - ($1671))|0;
        $1673 = (((($1665) + ($1633<<3)|0)) + 4|0);
        $1674 = HEAP32[$1673>>2]|0;
        $1675 = (((($1666) + ($1639<<3)|0)) + 4|0);
        $1676 = HEAP32[$1675>>2]|0;
        $1677 = (($1674) - ($1676))|0;
        $1678 = (FUNCTION_TABLE_iiii[$1667 & 8191]($exc,$1672,$1677)|0);
        $D$1$i = $1678;
        break;
       }
       $1679 = HEAP32[$59>>2]|0;
       $1680 = HEAP32[$58>>2]|0;
       $1681 = HEAP32[$60>>2]|0;
       $1682 = HEAP32[$61>>2]|0;
       $1683 = ($1681|0)==($1682|0);
       $1684 = (($1679) + ($1633<<3)|0);
       $1685 = HEAP32[$1684>>2]|0;
       $1686 = (($1680) + ($1639<<3)|0);
       $1687 = HEAP32[$1686>>2]|0;
       if ($1683) {
        $1688 = HEAP32[$50>>2]|0;
        $1689 = (($1685) - ($1687))|0;
        $1690 = (((($1679) + ($1633<<3)|0)) + 4|0);
        $1691 = HEAP32[$1690>>2]|0;
        $1692 = (((($1680) + ($1639<<3)|0)) + 4|0);
        $1693 = HEAP32[$1692>>2]|0;
        $1694 = (($1691) - ($1693))|0;
        $1695 = (FUNCTION_TABLE_iiii[$1688 & 8191]($exc,$1689,$1694)|0);
        $1696 = HEAP32[$60>>2]|0;
        $1697 = (_FT_MulFix($1695,$1696)|0);
        $D$1$i = $1697;
        break;
       } else {
        $1698 = (($1685) - ($1687))|0;
        $1699 = (_FT_MulFix($1698,$1681)|0);
        $1700 = (((($1679) + ($1633<<3)|0)) + 4|0);
        $1701 = HEAP32[$1700>>2]|0;
        $1702 = (((($1680) + ($1639<<3)|0)) + 4|0);
        $1703 = HEAP32[$1702>>2]|0;
        $1704 = (($1701) - ($1703))|0;
        $1705 = HEAP32[$61>>2]|0;
        $1706 = (_FT_MulFix($1704,$1705)|0);
        $1707 = HEAP32[$50>>2]|0;
        $1708 = (FUNCTION_TABLE_iiii[$1707 & 8191]($exc,$1699,$1706)|0);
        $D$1$i = $1708;
        break;
       }
      } else {
       label = 456;
      }
     } else {
      label = 456;
     }
    } while(0);
    if ((label|0) == 456) {
     label = 0;
     $1643 = HEAP8[$29>>0]|0;
     $1644 = ($1643<<24>>24)==(0);
     if ($1644) {
      $D$1$i = 0;
     } else {
      HEAP32[$34>>2] = 134;
      $D$1$i = 0;
     }
    }
    HEAP32[$172>>2] = $D$1$i;
    label = 821;
    break;
   }
   case 75:  {
    $1709 = HEAP32[$7>>2]|0;
    $1710 = (FUNCTION_TABLE_ii[$1709 & 8191]($exc)|0);
    HEAP32[$172>>2] = $1710;
    label = 821;
    break;
   }
   case 76:  {
    $1711 = HEAP32[$7>>2]|0;
    $1712 = (FUNCTION_TABLE_ii[$1711 & 8191]($exc)|0);
    HEAP32[$172>>2] = $1712;
    label = 821;
    break;
   }
   case 77:  {
    HEAP8[$53>>0] = 1;
    label = 821;
    break;
   }
   case 78:  {
    HEAP8[$53>>0] = 0;
    label = 821;
    break;
   }
   case 80:  {
    $1713 = HEAP32[$172>>2]|0;
    $1714 = ((($172)) + 4|0);
    $1715 = HEAP32[$1714>>2]|0;
    $1716 = ($1713|0)<($1715|0);
    $1717 = $1716&1;
    HEAP32[$172>>2] = $1717;
    label = 821;
    break;
   }
   case 81:  {
    $1718 = HEAP32[$172>>2]|0;
    $1719 = ((($172)) + 4|0);
    $1720 = HEAP32[$1719>>2]|0;
    $1721 = ($1718|0)<=($1720|0);
    $1722 = $1721&1;
    HEAP32[$172>>2] = $1722;
    label = 821;
    break;
   }
   case 82:  {
    $1723 = HEAP32[$172>>2]|0;
    $1724 = ((($172)) + 4|0);
    $1725 = HEAP32[$1724>>2]|0;
    $1726 = ($1723|0)>($1725|0);
    $1727 = $1726&1;
    HEAP32[$172>>2] = $1727;
    label = 821;
    break;
   }
   case 83:  {
    $1728 = HEAP32[$172>>2]|0;
    $1729 = ((($172)) + 4|0);
    $1730 = HEAP32[$1729>>2]|0;
    $1731 = ($1728|0)>=($1730|0);
    $1732 = $1731&1;
    HEAP32[$172>>2] = $1732;
    label = 821;
    break;
   }
   case 84:  {
    $1733 = HEAP32[$172>>2]|0;
    $1734 = ((($172)) + 4|0);
    $1735 = HEAP32[$1734>>2]|0;
    $1736 = ($1733|0)==($1735|0);
    $1737 = $1736&1;
    HEAP32[$172>>2] = $1737;
    label = 821;
    break;
   }
   case 85:  {
    $1738 = HEAP32[$172>>2]|0;
    $1739 = ((($172)) + 4|0);
    $1740 = HEAP32[$1739>>2]|0;
    $1741 = ($1738|0)!=($1740|0);
    $1742 = $1741&1;
    HEAP32[$172>>2] = $1742;
    label = 821;
    break;
   }
   case 86:  {
    $1743 = HEAP32[$$pre$phiZ2D>>2]|0;
    $1744 = HEAP32[$172>>2]|0;
    $1745 = (FUNCTION_TABLE_iiii[$1743 & 8191]($exc,$1744,0)|0);
    $1746 = $1745 & 127;
    $1747 = ($1746|0)==(64);
    $1748 = $1747&1;
    HEAP32[$172>>2] = $1748;
    label = 821;
    break;
   }
   case 87:  {
    $1749 = HEAP32[$$pre$phiZ2D>>2]|0;
    $1750 = HEAP32[$172>>2]|0;
    $1751 = (FUNCTION_TABLE_iiii[$1749 & 8191]($exc,$1750,0)|0);
    $1752 = $1751 & 127;
    $1753 = ($1752|0)==(0);
    $1754 = $1753&1;
    HEAP32[$172>>2] = $1754;
    label = 821;
    break;
   }
   case 88:  {
    $$val32 = HEAP32[$172>>2]|0;
    $1755 = ($$val32|0)==(0);
    if ($1755) {
     $$pre1$pre$i181 = HEAP32[$22>>2]|0;
     $$pre2$pre$i182 = HEAP32[$26>>2]|0;
     $1756 = (($$pre$pre$i) + ($$pre1$pre$i181))|0;
     HEAP32[$22>>2] = $1756;
     $1757 = ($1756|0)<($$pre2$pre$i182|0);
     if ($1757) {
      $2799 = $1756;$2800 = 1;$nIfs$0$ph$ph$i368 = 1;
     } else {
      label = 492;
      break L12;
     }
     while(1) {
      L692: do {
       if ($2800) {
        $1759 = $2799;
        while(1) {
         $1758 = (($129) + ($1759)|0);
         $1760 = HEAP8[$1758>>0]|0;
         HEAP8[$24>>0] = $1760;
         $1761 = $1760&255;
         $1762 = (593353 + ($1761)|0);
         $1763 = HEAP8[$1762>>0]|0;
         $1764 = $1763 << 24 >> 24;
         HEAP32[$25>>2] = $1764;
         $1765 = $1761 | 1;
         $1766 = ($1765|0)==(65);
         if ($1766) {
          $1767 = (($1759) + 1)|0;
          $1768 = ($1767|0)<($$pre2$pre$i182|0);
          if (!($1768)) {
           label = 492;
           break L12;
          }
          $1769 = (($129) + ($1767)|0);
          $1770 = HEAP8[$1769>>0]|0;
          $1771 = $1770&255;
          $1772 = Math_imul($1771, $1764)|0;
          $1773 = (2 - ($1772))|0;
          HEAP32[$25>>2] = $1773;
          $1775 = $1773;
         } else {
          $1775 = $1764;
         }
         $1774 = (($1775) + ($1759))|0;
         $1776 = ($1774|0)>($$pre2$pre$i182|0);
         if ($1776) {
          label = 492;
          break L12;
         }
         switch ($1761|0) {
         case 27:  {
          label = 821;
          break L30;
          break;
         }
         case 88:  {
          $$lcssa = $1775;$$lcssa270 = $1759;
          label = 495;
          break L692;
          break;
         }
         case 89:  {
          $$lcssa271 = $1759;$$lcssa273 = $1775;
          label = 497;
          break L692;
          break;
         }
         default: {
         }
         }
         HEAP32[$22>>2] = $1774;
         $1777 = ($1774|0)<($$pre2$pre$i182|0);
         if ($1777) {
          $1759 = $1774;
         } else {
          label = 492;
          break L12;
         }
        }
       } else {
        $1779 = $2799;
        while(1) {
         $1778 = (($129) + ($1779)|0);
         $1780 = HEAP8[$1778>>0]|0;
         HEAP8[$24>>0] = $1780;
         $1781 = $1780&255;
         $1782 = (593353 + ($1781)|0);
         $1783 = HEAP8[$1782>>0]|0;
         $1784 = $1783 << 24 >> 24;
         HEAP32[$25>>2] = $1784;
         $1785 = $1781 | 1;
         $1786 = ($1785|0)==(65);
         if ($1786) {
          $1787 = (($1779) + 1)|0;
          $1788 = ($1787|0)<($$pre2$pre$i182|0);
          if (!($1788)) {
           label = 492;
           break L12;
          }
          $1789 = (($129) + ($1787)|0);
          $1790 = HEAP8[$1789>>0]|0;
          $1791 = $1790&255;
          $1792 = Math_imul($1791, $1784)|0;
          $1793 = (2 - ($1792))|0;
          HEAP32[$25>>2] = $1793;
          $1795 = $1793;
         } else {
          $1795 = $1784;
         }
         $1794 = (($1795) + ($1779))|0;
         $1796 = ($1794|0)>($$pre2$pre$i182|0);
         if ($1796) {
          label = 492;
          break L12;
         }
         switch ($1781|0) {
         case 88:  {
          $$lcssa = $1795;$$lcssa270 = $1779;
          label = 495;
          break L692;
          break;
         }
         case 89:  {
          $$lcssa271 = $1779;$$lcssa273 = $1795;
          label = 497;
          break L692;
          break;
         }
         default: {
         }
         }
         HEAP32[$22>>2] = $1794;
         $1797 = ($1794|0)<($$pre2$pre$i182|0);
         if ($1797) {
          $1779 = $1794;
         } else {
          label = 492;
          break L12;
         }
        }
       }
      } while(0);
      if ((label|0) == 495) {
       label = 0;
       $1798 = (($nIfs$0$ph$ph$i368) + 1)|0;
       $1801 = $$lcssa;$1802 = $$lcssa270;$nIfs$0$ph$ph$i$be = $1798;
      }
      else if ((label|0) == 497) {
       label = 0;
       $1804 = (($nIfs$0$ph$ph$i368) + -1)|0;
       $1805 = ($1804|0)==(0);
       if ($1805) {
        label = 821;
        break L30;
       } else {
        $1801 = $$lcssa273;$1802 = $$lcssa271;$nIfs$0$ph$ph$i$be = $1804;
       }
      }
      $1799 = ($nIfs$0$ph$ph$i$be|0)==(1);
      $1800 = (($1801) + ($1802))|0;
      HEAP32[$22>>2] = $1800;
      $1803 = ($1800|0)<($$pre2$pre$i182|0);
      if ($1803) {
       $2799 = $1800;$2800 = $1799;$nIfs$0$ph$ph$i368 = $nIfs$0$ph$ph$i$be;
      } else {
       label = 492;
       break L12;
      }
     }
    } else {
     label = 821;
    }
    break;
   }
   case 90:  {
    $1806 = HEAP32[$172>>2]|0;
    $1807 = ($1806|0)==(0);
    if ($1807) {
     $1812 = 0;
    } else {
     $1808 = ((($172)) + 4|0);
     $1809 = HEAP32[$1808>>2]|0;
     $1810 = ($1809|0)!=(0);
     $1812 = $1810;
    }
    $1811 = $1812&1;
    HEAP32[$172>>2] = $1811;
    label = 821;
    break;
   }
   case 91:  {
    $1813 = HEAP32[$172>>2]|0;
    $1814 = ($1813|0)==(0);
    if ($1814) {
     $1815 = ((($172)) + 4|0);
     $1816 = HEAP32[$1815>>2]|0;
     $1817 = ($1816|0)!=(0);
     $1819 = $1817;
    } else {
     $1819 = 1;
    }
    $1818 = $1819&1;
    HEAP32[$172>>2] = $1818;
    label = 821;
    break;
   }
   case 92:  {
    $1820 = HEAP32[$172>>2]|0;
    $1821 = ($1820|0)==(0);
    $1822 = $1821&1;
    HEAP32[$172>>2] = $1822;
    label = 821;
    break;
   }
   case 93:  {
    _Ins_DELTAP($exc,$172);
    label = 821;
    break;
   }
   case 94:  {
    $$val33 = HEAP32[$172>>2]|0;
    $1823 = $$val33&65535;
    HEAP16[$114>>1] = $1823;
    label = 821;
    break;
   }
   case 95:  {
    $$val34 = HEAP32[$172>>2]|0;
    $1824 = ($$val34>>>0)>(6);
    if ($1824) {
     label = 508;
     break L12;
    }
    $1825 = $$val34&65535;
    HEAP16[$115>>1] = $1825;
    label = 821;
    break;
   }
   case 96:  {
    $1826 = ((($172)) + 4|0);
    $1827 = HEAP32[$1826>>2]|0;
    $1828 = HEAP32[$172>>2]|0;
    $1829 = (($1828) + ($1827))|0;
    HEAP32[$172>>2] = $1829;
    label = 821;
    break;
   }
   case 97:  {
    $1830 = ((($172)) + 4|0);
    $1831 = HEAP32[$1830>>2]|0;
    $1832 = HEAP32[$172>>2]|0;
    $1833 = (($1832) - ($1831))|0;
    HEAP32[$172>>2] = $1833;
    label = 821;
    break;
   }
   case 98:  {
    $1834 = ((($172)) + 4|0);
    $1835 = HEAP32[$1834>>2]|0;
    $1836 = ($1835|0)==(0);
    if ($1836) {
     label = 513;
     break L12;
    }
    $1837 = HEAP32[$172>>2]|0;
    $1838 = (_FT_MulDiv_No_Round($1837,64,$1835)|0);
    HEAP32[$172>>2] = $1838;
    label = 821;
    break;
   }
   case 99:  {
    $1839 = HEAP32[$172>>2]|0;
    $1840 = ((($172)) + 4|0);
    $1841 = HEAP32[$1840>>2]|0;
    $1842 = (_FT_MulDiv($1839,$1841,64)|0);
    HEAP32[$172>>2] = $1842;
    label = 821;
    break;
   }
   case 100:  {
    $1843 = HEAP32[$172>>2]|0;
    $1844 = ($1843|0)<(0);
    $1845 = (0 - ($1843))|0;
    $1846 = $1844 ? $1845 : $1843;
    HEAP32[$172>>2] = $1846;
    label = 821;
    break;
   }
   case 101:  {
    $1847 = HEAP32[$172>>2]|0;
    $1848 = (0 - ($1847))|0;
    HEAP32[$172>>2] = $1848;
    label = 821;
    break;
   }
   case 102:  {
    $1849 = HEAP32[$172>>2]|0;
    $1850 = $1849 & -64;
    HEAP32[$172>>2] = $1850;
    label = 821;
    break;
   }
   case 103:  {
    $1851 = HEAP32[$172>>2]|0;
    $1852 = (($1851) + 63)|0;
    $1853 = $1852 & -64;
    HEAP32[$172>>2] = $1853;
    label = 821;
    break;
   }
   case 107: case 106: case 105: case 104:  {
    $1854 = HEAP32[$$pre$phiZ2D>>2]|0;
    $1855 = HEAP32[$172>>2]|0;
    $1856 = (($133) + -104)|0;
    $1857 = (((($exc)) + 264|0) + ($1856<<2)|0);
    $1858 = HEAP32[$1857>>2]|0;
    $1859 = (FUNCTION_TABLE_iiii[$1854 & 8191]($exc,$1855,$1858)|0);
    HEAP32[$172>>2] = $1859;
    label = 821;
    break;
   }
   case 111: case 110: case 109: case 108:  {
    $1860 = HEAP32[$172>>2]|0;
    $1861 = (($133) + -108)|0;
    $1862 = (((($exc)) + 264|0) + ($1861<<2)|0);
    $1863 = HEAP32[$1862>>2]|0;
    $1864 = ($1860|0)>(-1);
    if ($1864) {
     $1865 = (($1863) + ($1860))|0;
     $1866 = ($1865|0)<(0);
     $$$i$i187 = $1866 ? 0 : $1865;
     $val$0$i$i = $$$i$i187;
    } else {
     $1867 = (($1860) - ($1863))|0;
     $1868 = ($1867|0)>(0);
     $$1$i$i188 = $1868 ? 0 : $1867;
     $val$0$i$i = $$1$i$i188;
    }
    HEAP32[$172>>2] = $val$0$i$i;
    label = 821;
    break;
   }
   case 112:  {
    $1869 = HEAP32[$172>>2]|0;
    $1870 = HEAP32[$39>>2]|0;
    $1871 = ($1869>>>0)<($1870>>>0);
    if ($1871) {
     $1874 = ((($172)) + 4|0);
     $1875 = HEAP32[$1874>>2]|0;
     $1876 = HEAP32[$86>>2]|0;
     $1877 = (_FT_MulFix($1875,$1876)|0);
     $1878 = HEAP32[$116>>2]|0;
     $1879 = (($1878) + ($1869<<2)|0);
     HEAP32[$1879>>2] = $1877;
     label = 821;
     break L30;
    } else {
     $1872 = HEAP8[$29>>0]|0;
     $1873 = ($1872<<24>>24)==(0);
     if ($1873) {
      label = 821;
      break L30;
     } else {
      label = 527;
      break L12;
     }
    }
    break;
   }
   case 114: case 113:  {
    _Ins_DELTAP($exc,$172);
    label = 821;
    break;
   }
   case 117: case 116: case 115:  {
    $1880 = HEAP32[$7>>2]|0;
    $1881 = (FUNCTION_TABLE_ii[$1880 & 8191]($exc)|0);
    $1882 = HEAP32[$172>>2]|0;
    $1883 = ($1882|0)==(0);
    L747: do {
     if (!($1883)) {
      $k$02$i = 1;
      while(1) {
       $1884 = HEAP32[$28>>2]|0;
       $1885 = ($1884|0)<(2);
       if ($1885) {
        break;
       }
       $1888 = (($1884) + -2)|0;
       HEAP32[$28>>2] = $1888;
       $1889 = (($1884) + -1)|0;
       $1890 = HEAP32[$30>>2]|0;
       $1891 = (($1890) + ($1889<<2)|0);
       $1892 = HEAP32[$1891>>2]|0;
       $1893 = (($1890) + ($1888<<2)|0);
       $1894 = HEAP32[$1893>>2]|0;
       $1895 = HEAP32[$39>>2]|0;
       $1896 = ($1892>>>0)<($1895>>>0);
       if ($1896) {
        $1899 = $1894 >>> 4;
        $1900 = $1899 & 15;
        $1901 = HEAP8[$24>>0]|0;
        $1902 = $1901&255;
        switch ($1902|0) {
        case 117:  {
         $1904 = $1900 | 32;
         $C$0$i = $1904;
         break;
        }
        case 116:  {
         $1903 = $1900 | 16;
         $C$0$i = $1903;
         break;
        }
        default: {
         $C$0$i = $1900;
        }
        }
        $1905 = HEAP32[$114>>2]|0;
        $1906 = $1905 & 65535;
        $1907 = (($1906) + ($C$0$i))|0;
        $1908 = ($1881|0)==($1907|0);
        if ($1908) {
         $1909 = $1905 >>> 16;
         $1910 = $1894 & 15;
         $1911 = (($1910) + -8)|0;
         $1912 = ($1911|0)>(-1);
         $1913 = (($1910) + -7)|0;
         $$$i192 = $1912 ? $1913 : $1911;
         $1914 = (6 - ($1909))|0;
         $1915 = $$$i192 << $1914;
         $1916 = HEAP32[$10>>2]|0;
         FUNCTION_TABLE_viii[$1916 & 8191]($exc,$1892,$1915);
        }
       } else {
        $1897 = HEAP8[$29>>0]|0;
        $1898 = ($1897<<24>>24)==(0);
        if (!($1898)) {
         label = 537;
         break L12;
        }
       }
       $1917 = (($k$02$i) + 1)|0;
       $1918 = ($1917>>>0)>($1882>>>0);
       if ($1918) {
        break L747;
       } else {
        $k$02$i = $1917;
       }
      }
      $1886 = HEAP8[$29>>0]|0;
      $1887 = ($1886<<24>>24)==(0);
      if (!($1887)) {
       HEAP32[$34>>2] = 129;
      }
      HEAP32[$28>>2] = 0;
     }
    } while(0);
    $1919 = HEAP32[$28>>2]|0;
    HEAP32[$31>>2] = $1919;
    label = 821;
    break;
   }
   case 118:  {
    $$val35 = HEAP32[$172>>2]|0;
    $1920 = $$val35 & 192;
    switch ($1920|0) {
    case 0:  {
     HEAP32[$117>>2] = 8192;
     $1923 = 8192;
     break;
    }
    case 64:  {
     HEAP32[$117>>2] = 16384;
     $1923 = 16384;
     break;
    }
    case 128:  {
     HEAP32[$117>>2] = 32768;
     $1923 = 32768;
     break;
    }
    case 192:  {
     HEAP32[$117>>2] = 16384;
     $1923 = 16384;
     break;
    }
    default: {
     label = 550;
     break L12;
    }
    }
    $1921 = $$val35 & 48;
    switch ($1921|0) {
    case 0:  {
     HEAP32[$118>>2] = 0;
     $1935 = 0;
     break;
    }
    case 16:  {
     $1922 = $1923 >>> 2;
     HEAP32[$118>>2] = $1922;
     $1935 = $1922;
     break;
    }
    case 32:  {
     $1924 = $1923 >>> 1;
     HEAP32[$118>>2] = $1924;
     $1935 = $1924;
     break;
    }
    case 48:  {
     $1925 = ($1923*3)|0;
     $1926 = $1925 >>> 2;
     HEAP32[$118>>2] = $1926;
     $1935 = $1926;
     break;
    }
    default: {
     label = 556;
     break L12;
    }
    }
    $1927 = $$val35 & 15;
    $1928 = ($1927|0)==(0);
    if ($1928) {
     $1929 = (($1923) + -1)|0;
     $$sink$i$i = $1929;
    } else {
     $1930 = (($1927) + -4)|0;
     $1931 = Math_imul($1923, $1930)|0;
     $1932 = (($1931|0) / 8)&-1;
     $$sink$i$i = $1932;
    }
    $1933 = $1923 >>> 8;
    HEAP32[$117>>2] = $1933;
    $1934 = $1935 >>> 8;
    HEAP32[$118>>2] = $1934;
    $1936 = $$sink$i$i >> 8;
    HEAP32[$119>>2] = $1936;
    HEAP32[$11>>2] = 6;
    HEAP32[$$pre$phiZ2D>>2] = 4424;
    label = 821;
    break;
   }
   case 119:  {
    $$val36 = HEAP32[$172>>2]|0;
    $1937 = $$val36 & 192;
    switch ($1937|0) {
    case 0:  {
     HEAP32[$117>>2] = 5792;
     $1940 = 5792;
     break;
    }
    case 64:  {
     HEAP32[$117>>2] = 11585;
     $1940 = 11585;
     break;
    }
    case 128:  {
     HEAP32[$117>>2] = 23170;
     $1940 = 23170;
     break;
    }
    case 192:  {
     HEAP32[$117>>2] = 11585;
     $1940 = 11585;
     break;
    }
    default: {
     label = 566;
     break L12;
    }
    }
    $1938 = $$val36 & 48;
    switch ($1938|0) {
    case 0:  {
     HEAP32[$118>>2] = 0;
     $1952 = 0;
     break;
    }
    case 16:  {
     $1939 = $1940 >>> 2;
     HEAP32[$118>>2] = $1939;
     $1952 = $1939;
     break;
    }
    case 32:  {
     $1941 = $1940 >>> 1;
     HEAP32[$118>>2] = $1941;
     $1952 = $1941;
     break;
    }
    case 48:  {
     $1942 = ($1940*3)|0;
     $1943 = $1942 >>> 2;
     HEAP32[$118>>2] = $1943;
     $1952 = $1943;
     break;
    }
    default: {
     label = 572;
     break L12;
    }
    }
    $1944 = $$val36 & 15;
    $1945 = ($1944|0)==(0);
    if ($1945) {
     $1946 = (($1940) + -1)|0;
     $$sink$i$i195 = $1946;
    } else {
     $1947 = (($1944) + -4)|0;
     $1948 = Math_imul($1940, $1947)|0;
     $1949 = (($1948|0) / 8)&-1;
     $$sink$i$i195 = $1949;
    }
    $1950 = $1940 >>> 8;
    HEAP32[$117>>2] = $1950;
    $1951 = $1952 >>> 8;
    HEAP32[$118>>2] = $1951;
    $1953 = $$sink$i$i195 >> 8;
    HEAP32[$119>>2] = $1953;
    HEAP32[$11>>2] = 7;
    HEAP32[$$pre$phiZ2D>>2] = 4425;
    label = 821;
    break;
   }
   case 120:  {
    $1954 = ((($172)) + 4|0);
    $1955 = HEAP32[$1954>>2]|0;
    $1956 = ($1955|0)==(0);
    if ($1956) {
     label = 821;
    } else {
     $args$val$i = HEAP32[$172>>2]|0;
     $1957 = $args$val$i | $168;
     $1958 = ($1957|0)==(0);
     if ($1958) {
      HEAP32[$34>>2] = 132;
     }
     $1959 = HEAP32[$22>>2]|0;
     $1960 = (($1959) + ($args$val$i))|0;
     HEAP32[$22>>2] = $1960;
     $1961 = ($1960|0)<(0);
     if ($1961) {
      label = 583;
     } else {
      $1962 = HEAP32[$64>>2]|0;
      $1963 = ($1962|0)>(0);
      if ($1963) {
       $1964 = (($1962) + -1)|0;
       $1965 = HEAP32[$66>>2]|0;
       $1966 = (((($1965) + ($1964<<4)|0)) + 12|0);
       $1967 = HEAP32[$1966>>2]|0;
       $1968 = ((($1967)) + 8|0);
       $1969 = HEAP32[$1968>>2]|0;
       $1970 = ($1960|0)>($1969|0);
       if ($1970) {
        label = 583;
       }
      }
     }
     if ((label|0) == 583) {
      label = 0;
      HEAP32[$34>>2] = 132;
     }
     HEAP8[$33>>0] = 0;
     label = 821;
    }
    break;
   }
   case 121:  {
    $1971 = ((($172)) + 4|0);
    $1972 = HEAP32[$1971>>2]|0;
    $1973 = ($1972|0)==(0);
    if ($1973) {
     $args$val$i196 = HEAP32[$172>>2]|0;
     $1974 = $args$val$i196 | $168;
     $1975 = ($1974|0)==(0);
     if ($1975) {
      HEAP32[$34>>2] = 132;
     }
     $1976 = HEAP32[$22>>2]|0;
     $1977 = (($1976) + ($args$val$i196))|0;
     HEAP32[$22>>2] = $1977;
     $1978 = ($1977|0)<(0);
     if ($1978) {
      label = 591;
     } else {
      $1979 = HEAP32[$64>>2]|0;
      $1980 = ($1979|0)>(0);
      if ($1980) {
       $1981 = (($1979) + -1)|0;
       $1982 = HEAP32[$66>>2]|0;
       $1983 = (((($1982) + ($1981<<4)|0)) + 12|0);
       $1984 = HEAP32[$1983>>2]|0;
       $1985 = ((($1984)) + 8|0);
       $1986 = HEAP32[$1985>>2]|0;
       $1987 = ($1977|0)>($1986|0);
       if ($1987) {
        label = 591;
       }
      }
     }
     if ((label|0) == 591) {
      label = 0;
      HEAP32[$34>>2] = 132;
     }
     HEAP8[$33>>0] = 0;
     label = 821;
    } else {
     label = 821;
    }
    break;
   }
   case 122:  {
    HEAP32[$11>>2] = 5;
    HEAP32[$$pre$phiZ2D>>2] = 4418;
    label = 821;
    break;
   }
   case 123:  {
    $1988 = HEAP32[$62>>2]|0;
    $1989 = HEAP32[$63>>2]|0;
    $1990 = (($1988) + (($1989*24)|0)|0);
    $1991 = ($1989|0)>(0);
    L823: do {
     if ($1991) {
      $def$02$i199 = $1988;
      while(1) {
       $1992 = ((($def$02$i199)) + 12|0);
       $1993 = HEAP32[$1992>>2]|0;
       $1994 = $1993&255;
       $1995 = ($132<<24>>24)==($1994<<24>>24);
       if ($1995) {
        $1996 = ((($def$02$i199)) + 16|0);
        $1997 = HEAP8[$1996>>0]|0;
        $1998 = ($1997<<24>>24)==(0);
        if (!($1998)) {
         $def$02$i199$lcssa = $def$02$i199;
         break;
        }
       }
       $2021 = ((($def$02$i199)) + 24|0);
       $2022 = ($2021>>>0)<($1990>>>0);
       if ($2022) {
        $def$02$i199 = $2021;
       } else {
        break L823;
       }
      }
      $1999 = HEAP32[$64>>2]|0;
      $2000 = HEAP32[$65>>2]|0;
      $2001 = ($1999|0)<($2000|0);
      if (!($2001)) {
       label = 598;
       break L12;
      }
      $2002 = HEAP32[$66>>2]|0;
      $2003 = (($1999) + 1)|0;
      HEAP32[$64>>2] = $2003;
      $2004 = HEAP32[$67>>2]|0;
      $2005 = (($2002) + ($1999<<4)|0);
      HEAP32[$2005>>2] = $2004;
      $2006 = HEAP32[$22>>2]|0;
      $2007 = (($2006) + 1)|0;
      $2008 = (((($2002) + ($1999<<4)|0)) + 4|0);
      HEAP32[$2008>>2] = $2007;
      $2009 = (((($2002) + ($1999<<4)|0)) + 8|0);
      HEAP32[$2009>>2] = 1;
      $2010 = (((($2002) + ($1999<<4)|0)) + 12|0);
      HEAP32[$2010>>2] = $def$02$i199$lcssa;
      $2011 = HEAP32[$def$02$i199$lcssa>>2]|0;
      $2012 = ((($def$02$i199$lcssa)) + 4|0);
      $2013 = HEAP32[$2012>>2]|0;
      $aRange$off$i$i200 = (($2011) + -1)|0;
      $2014 = ($aRange$off$i$i200>>>0)>(2);
      do {
       if ($2014) {
        HEAP32[$34>>2] = 132;
       } else {
        $2015 = (((($exc)) + 444|0) + ($aRange$off$i$i200<<3)|0);
        $2016 = HEAP32[$2015>>2]|0;
        $2017 = ($2016|0)==(0|0);
        if ($2017) {
         HEAP32[$34>>2] = 138;
         break;
        }
        $2018 = (((((($exc)) + 444|0) + ($aRange$off$i$i200<<3)|0)) + 4|0);
        $2019 = HEAP32[$2018>>2]|0;
        $2020 = ($2019|0)<($2013|0);
        if ($2020) {
         HEAP32[$34>>2] = 131;
         break;
        } else {
         HEAP32[$23>>2] = $2016;
         HEAP32[$26>>2] = $2019;
         HEAP32[$22>>2] = $2013;
         HEAP32[$67>>2] = $2011;
         break;
        }
       }
      } while(0);
      HEAP8[$33>>0] = 0;
      label = 821;
      break L30;
     }
    } while(0);
    HEAP32[$34>>2] = 128;
    $2736 = $1988;$2737 = $1989;
    label = 823;
    break;
   }
   case 124:  {
    HEAP32[$11>>2] = 4;
    HEAP32[$$pre$phiZ2D>>2] = 4420;
    label = 821;
    break;
   }
   case 125:  {
    HEAP32[$11>>2] = 3;
    HEAP32[$$pre$phiZ2D>>2] = 4421;
    label = 821;
    break;
   }
   case 128:  {
    $2023 = HEAP32[$27>>2]|0;
    $2024 = HEAP32[$85>>2]|0;
    $2025 = ($2023|0)<($2024|0);
    L845: do {
     if ($2025) {
      $2027 = HEAP8[$29>>0]|0;
      $2028 = ($2027<<24>>24)==(0);
      if ($2028) {
       $2049 = $168;
      } else {
       HEAP32[$34>>2] = 129;
       $2049 = $168;
      }
     } else {
      $2026 = ($2024|0)>(0);
      if ($2026) {
       $2030 = $168;$2032 = $171;$2801 = $2024;
       while(1) {
        $2029 = (($2030) + -1)|0;
        HEAP32[$28>>2] = $2029;
        $2031 = (($2032) + ($2029<<2)|0);
        $2033 = HEAP32[$2031>>2]|0;
        $2034 = $2033 & 65535;
        $2035 = HEAP16[$99>>1]|0;
        $2036 = $2035&65535;
        $2037 = ($2034>>>0)<($2036>>>0);
        if ($2037) {
         $2040 = HEAP32[$103>>2]|0;
         $2041 = (($2040) + ($2034)|0);
         $2042 = HEAP8[$2041>>0]|0;
         $2043 = $2042&255;
         $2044 = $2043 ^ 1;
         $2045 = $2044&255;
         HEAP8[$2041>>0] = $2045;
         $$pre$i226 = HEAP32[$85>>2]|0;
         $2047 = $$pre$i226;
        } else {
         $2038 = HEAP8[$29>>0]|0;
         $2039 = ($2038<<24>>24)==(0);
         if ($2039) {
          $2047 = $2801;
         } else {
          label = 617;
          break L12;
         }
        }
        $2046 = (($2047) + -1)|0;
        HEAP32[$85>>2] = $2046;
        $2048 = ($2047|0)>(1);
        $$pre442 = HEAP32[$28>>2]|0;
        if (!($2048)) {
         $2049 = $$pre442;
         break L845;
        }
        $$pre443 = HEAP32[$30>>2]|0;
        $2030 = $$pre442;$2032 = $$pre443;$2801 = $2046;
       }
      } else {
       $2049 = $168;
      }
     }
    } while(0);
    HEAP32[$85>>2] = 1;
    HEAP32[$31>>2] = $2049;
    label = 821;
    break;
   }
   case 129:  {
    $$val37 = HEAP32[$172>>2]|0;
    $$idx38 = ((($172)) + 4|0);
    $$idx38$val = HEAP32[$$idx38>>2]|0;
    $2050 = $$idx38$val & 65535;
    $2051 = HEAP16[$99>>1]|0;
    $2052 = $2051&65535;
    $2053 = ($2050>>>0)<($2052>>>0);
    $2054 = $$val37 & 65535;
    $2055 = ($2054>>>0)<($2052>>>0);
    $or$cond$i228 = $2053 & $2055;
    if (!($or$cond$i228)) {
     $2058 = HEAP8[$29>>0]|0;
     $2059 = ($2058<<24>>24)==(0);
     if ($2059) {
      label = 821;
      break L30;
     } else {
      label = 626;
      break L12;
     }
    }
    $2056 = ($2054>>>0)>($2050>>>0);
    if ($2056) {
     label = 821;
    } else {
     $2057 = $$val37&65535;
     $2062 = $2054;$I$01$i = $2057;
     while(1) {
      $2060 = HEAP32[$103>>2]|0;
      $2061 = (($2060) + ($2062)|0);
      $2063 = HEAP8[$2061>>0]|0;
      $2064 = $2063&255;
      $2065 = $2064 | 1;
      $2066 = $2065&255;
      HEAP8[$2061>>0] = $2066;
      $2067 = (($I$01$i) + 1)<<16>>16;
      $2068 = $2067&65535;
      $2069 = ($2068>>>0)>($2050>>>0);
      if ($2069) {
       label = 821;
       break;
      } else {
       $2062 = $2068;$I$01$i = $2067;
      }
     }
    }
    break;
   }
   case 130:  {
    $$val39 = HEAP32[$172>>2]|0;
    $$idx40 = ((($172)) + 4|0);
    $$idx40$val = HEAP32[$$idx40>>2]|0;
    $2070 = $$idx40$val & 65535;
    $2071 = HEAP16[$99>>1]|0;
    $2072 = $2071&65535;
    $2073 = ($2070>>>0)<($2072>>>0);
    $2074 = $$val39 & 65535;
    $2075 = ($2074>>>0)<($2072>>>0);
    $or$cond$i232 = $2073 & $2075;
    if (!($or$cond$i232)) {
     $2078 = HEAP8[$29>>0]|0;
     $2079 = ($2078<<24>>24)==(0);
     if ($2079) {
      label = 821;
      break L30;
     } else {
      label = 632;
      break L12;
     }
    }
    $2076 = ($2074>>>0)>($2070>>>0);
    if ($2076) {
     label = 821;
    } else {
     $2077 = $$val39&65535;
     $2082 = $2074;$I$01$i235 = $2077;
     while(1) {
      $2080 = HEAP32[$103>>2]|0;
      $2081 = (($2080) + ($2082)|0);
      $2083 = HEAP8[$2081>>0]|0;
      $2084 = $2083&255;
      $2085 = $2084 & 254;
      $2086 = $2085&255;
      HEAP8[$2081>>0] = $2086;
      $2087 = (($I$01$i235) + 1)<<16>>16;
      $2088 = $2087&65535;
      $2089 = ($2088>>>0)>($2070>>>0);
      if ($2089) {
       label = 821;
       break;
      } else {
       $2082 = $2088;$I$01$i235 = $2087;
      }
     }
    }
    break;
   }
   case 132: case 131:  {
    $2090 = HEAP32[$62>>2]|0;
    $2091 = HEAP32[$63>>2]|0;
    $2092 = (($2090) + (($2091*24)|0)|0);
    $2093 = ($2091|0)>(0);
    L872: do {
     if ($2093) {
      $def$02$i238 = $2090;
      while(1) {
       $2094 = ((($def$02$i238)) + 12|0);
       $2095 = HEAP32[$2094>>2]|0;
       $2096 = $2095&255;
       $2097 = ($132<<24>>24)==($2096<<24>>24);
       if ($2097) {
        $2098 = ((($def$02$i238)) + 16|0);
        $2099 = HEAP8[$2098>>0]|0;
        $2100 = ($2099<<24>>24)==(0);
        if (!($2100)) {
         $def$02$i238$lcssa = $def$02$i238;
         break;
        }
       }
       $2123 = ((($def$02$i238)) + 24|0);
       $2124 = ($2123>>>0)<($2092>>>0);
       if ($2124) {
        $def$02$i238 = $2123;
       } else {
        break L872;
       }
      }
      $2101 = HEAP32[$64>>2]|0;
      $2102 = HEAP32[$65>>2]|0;
      $2103 = ($2101|0)<($2102|0);
      if (!($2103)) {
       label = 638;
       break L12;
      }
      $2104 = HEAP32[$66>>2]|0;
      $2105 = (($2101) + 1)|0;
      HEAP32[$64>>2] = $2105;
      $2106 = HEAP32[$67>>2]|0;
      $2107 = (($2104) + ($2101<<4)|0);
      HEAP32[$2107>>2] = $2106;
      $2108 = HEAP32[$22>>2]|0;
      $2109 = (($2108) + 1)|0;
      $2110 = (((($2104) + ($2101<<4)|0)) + 4|0);
      HEAP32[$2110>>2] = $2109;
      $2111 = (((($2104) + ($2101<<4)|0)) + 8|0);
      HEAP32[$2111>>2] = 1;
      $2112 = (((($2104) + ($2101<<4)|0)) + 12|0);
      HEAP32[$2112>>2] = $def$02$i238$lcssa;
      $2113 = HEAP32[$def$02$i238$lcssa>>2]|0;
      $2114 = ((($def$02$i238$lcssa)) + 4|0);
      $2115 = HEAP32[$2114>>2]|0;
      $aRange$off$i$i239 = (($2113) + -1)|0;
      $2116 = ($aRange$off$i$i239>>>0)>(2);
      do {
       if ($2116) {
        HEAP32[$34>>2] = 132;
       } else {
        $2117 = (((($exc)) + 444|0) + ($aRange$off$i$i239<<3)|0);
        $2118 = HEAP32[$2117>>2]|0;
        $2119 = ($2118|0)==(0|0);
        if ($2119) {
         HEAP32[$34>>2] = 138;
         break;
        }
        $2120 = (((((($exc)) + 444|0) + ($aRange$off$i$i239<<3)|0)) + 4|0);
        $2121 = HEAP32[$2120>>2]|0;
        $2122 = ($2121|0)<($2115|0);
        if ($2122) {
         HEAP32[$34>>2] = 131;
         break;
        } else {
         HEAP32[$23>>2] = $2118;
         HEAP32[$26>>2] = $2121;
         HEAP32[$22>>2] = $2115;
         HEAP32[$67>>2] = $2113;
         break;
        }
       }
      } while(0);
      HEAP8[$33>>0] = 0;
      label = 821;
      break L30;
     }
    } while(0);
    HEAP32[$34>>2] = 128;
    $2736 = $2090;$2737 = $2091;
    label = 823;
    break;
   }
   case 133:  {
    $$val41 = HEAP32[$172>>2]|0;
    $2125 = $$val41 & 255;
    switch ($2125|0) {
    case 255:  {
     HEAP8[$121>>0] = 1;
     label = 821;
     break L30;
     break;
    }
    case 0:  {
     HEAP8[$121>>0] = 0;
     label = 821;
     break L30;
     break;
    }
    default: {
     $2126 = $$val41 & 256;
     $2127 = ($2126|0)==(0);
     if (!($2127)) {
      $2128 = HEAP16[$123>>1]|0;
      $2129 = $2128&65535;
      $2130 = ($2129>>>0)>($2125>>>0);
      if (!($2130)) {
       HEAP8[$121>>0] = 1;
      }
     }
     $2131 = $$val41 & 512;
     $2132 = ($2131|0)==(0);
     if (!($2132)) {
      $2133 = HEAP8[$122>>0]|0;
      $2134 = ($2133<<24>>24)==(0);
      if (!($2134)) {
       HEAP8[$121>>0] = 1;
      }
     }
     $2135 = $$val41 & 1024;
     $2136 = ($2135|0)==(0);
     if (!($2136)) {
      $2137 = HEAP8[$120>>0]|0;
      $2138 = ($2137<<24>>24)==(0);
      if (!($2138)) {
       HEAP8[$121>>0] = 1;
      }
     }
     $2139 = $$val41 & 2048;
     $2140 = ($2139|0)==(0);
     if (!($2140)) {
      $2141 = HEAP16[$123>>1]|0;
      $2142 = $2141&65535;
      $2143 = ($2142>>>0)>($2125>>>0);
      if ($2143) {
       HEAP8[$121>>0] = 0;
      }
     }
     $2144 = $$val41 & 4096;
     $2145 = ($2144|0)==(0);
     if (!($2145)) {
      $2146 = HEAP8[$122>>0]|0;
      $2147 = ($2146<<24>>24)==(0);
      if (!($2147)) {
       HEAP8[$121>>0] = 0;
      }
     }
     $2148 = $$val41 & 8192;
     $2149 = ($2148|0)==(0);
     if ($2149) {
      label = 821;
      break L30;
     }
     $2150 = HEAP8[$120>>0]|0;
     $2151 = ($2150<<24>>24)==(0);
     if ($2151) {
      label = 821;
      break L30;
     }
     HEAP8[$121>>0] = 0;
     label = 821;
     break L30;
    }
    }
    break;
   }
   case 135: case 134:  {
    $2152 = HEAP32[$172>>2]|0;
    $2153 = $2152 & 65535;
    $2154 = HEAP16[$38>>1]|0;
    $2155 = $2154&65535;
    $2156 = ($2153>>>0)<($2155>>>0);
    if ($2156) {
     $2157 = ((($172)) + 4|0);
     $2158 = HEAP32[$2157>>2]|0;
     $2159 = $2158 & 65535;
     $2160 = HEAP16[$74>>1]|0;
     $2161 = $2160&65535;
     $2162 = ($2159>>>0)<($2161>>>0);
     if ($2162) {
      $2165 = HEAP32[$47>>2]|0;
      $2166 = HEAP32[$109>>2]|0;
      $2167 = (($2165) + ($2153<<3)|0);
      $2168 = HEAP32[$2167>>2]|0;
      $2169 = (($2166) + ($2159<<3)|0);
      $2170 = HEAP32[$2169>>2]|0;
      $2171 = (($2168) - ($2170))|0;
      $2172 = (((($2165) + ($2153<<3)|0)) + 4|0);
      $2173 = HEAP32[$2172>>2]|0;
      $2174 = (((($2166) + ($2159<<3)|0)) + 4|0);
      $2175 = HEAP32[$2174>>2]|0;
      $2176 = (($2173) - ($2175))|0;
      $2177 = ($2168|0)==($2170|0);
      $2178 = ($2173|0)==($2175|0);
      $or$cond$i221 = $2177 & $2178;
      $$4$i = $or$cond$i221 ? 16384 : $2171;
      $$op$i = $132 & 1;
      $2179 = ($$op$i<<24>>24)==(0);
      $2180 = $2179 | $or$cond$i221;
      $2181 = (0 - ($2176))|0;
      $B$0$i = $2180 ? $2176 : $$4$i;
      $A$1$i = $2180 ? $$4$i : $2181;
      $2182 = ($A$1$i|0)<(0);
      $2183 = (0 - ($A$1$i))|0;
      $2184 = $2182 ? $2183 : $A$1$i;
      $2185 = ($2184|0)<(16384);
      if ($2185) {
       $2186 = ($B$0$i|0)<(0);
       $2187 = (0 - ($B$0$i))|0;
       $2188 = $2186 ? $2187 : $B$0$i;
       $2189 = ($2188|0)<(16384);
       if ($2189) {
        $2190 = $B$0$i | $A$1$i;
        $2191 = ($2190|0)==(0);
        if (!($2191)) {
         $2192 = $A$1$i << 14;
         $2193 = $B$0$i << 14;
         $$02$i$i223 = $2192;$$03$i$i222 = $2193;
         label = 678;
        }
       } else {
        $$02$i$i223 = $A$1$i;$$03$i$i222 = $B$0$i;
        label = 678;
       }
      } else {
       $$02$i$i223 = $A$1$i;$$03$i$i222 = $B$0$i;
       label = 678;
      }
      if ((label|0) == 678) {
       label = 0;
       $2194 = (_FT_Hypot($$02$i$i223,$$03$i$i222)|0);
       $2195 = $2194 << 2;
       $2196 = (_FT_DivFix($$02$i$i223,$2195)|0);
       $2197 = $2196&65535;
       HEAP16[$71>>1] = $2197;
       $2198 = (_FT_DivFix($$03$i$i222,$2195)|0);
       $2199 = $2198&65535;
       HEAP16[$72>>1] = $2199;
      }
      $2200 = HEAP32[$49>>2]|0;
      $2201 = HEAP32[$75>>2]|0;
      $2202 = (($2200) + ($2153<<3)|0);
      $2203 = HEAP32[$2202>>2]|0;
      $2204 = (($2201) + ($2159<<3)|0);
      $2205 = HEAP32[$2204>>2]|0;
      $2206 = (($2203) - ($2205))|0;
      $2207 = (((($2200) + ($2153<<3)|0)) + 4|0);
      $2208 = HEAP32[$2207>>2]|0;
      $2209 = (((($2201) + ($2159<<3)|0)) + 4|0);
      $2210 = HEAP32[$2209>>2]|0;
      $2211 = (($2208) - ($2210))|0;
      $2212 = ($2203|0)==($2205|0);
      $2213 = ($2208|0)==($2210|0);
      $or$cond3$i = $2212 & $2213;
      $$5$i = $or$cond3$i ? 16384 : $2206;
      $2214 = $2180 | $or$cond3$i;
      $2215 = (0 - ($2211))|0;
      $B$1$i = $2214 ? $2211 : $$5$i;
      $A$3$i = $2214 ? $$5$i : $2215;
      $2216 = ($A$3$i|0)<(0);
      $2217 = (0 - ($A$3$i))|0;
      $2218 = $2216 ? $2217 : $A$3$i;
      $2219 = ($2218|0)<(16384);
      if ($2219) {
       $2220 = ($B$1$i|0)<(0);
       $2221 = (0 - ($B$1$i))|0;
       $2222 = $2220 ? $2221 : $B$1$i;
       $2223 = ($2222|0)<(16384);
       if ($2223) {
        $2224 = $B$1$i | $A$3$i;
        $2225 = ($2224|0)==(0);
        if (!($2225)) {
         $2226 = $A$3$i << 14;
         $2227 = $B$1$i << 14;
         $$02$i7$i = $2226;$$03$i6$i = $2227;
         label = 683;
        }
       } else {
        $$02$i7$i = $A$3$i;$$03$i6$i = $B$1$i;
        label = 683;
       }
      } else {
       $$02$i7$i = $A$3$i;$$03$i6$i = $B$1$i;
       label = 683;
      }
      if ((label|0) == 683) {
       label = 0;
       $2228 = (_FT_Hypot($$02$i7$i,$$03$i6$i)|0);
       $2229 = $2228 << 2;
       $2230 = (_FT_DivFix($$02$i7$i,$2229)|0);
       $2231 = $2230&65535;
       HEAP16[$69>>1] = $2231;
       $2232 = (_FT_DivFix($$03$i6$i,$2229)|0);
       $2233 = $2232&65535;
       HEAP16[$70>>1] = $2233;
      }
      _Compute_Funcs($exc);
      label = 821;
      break L30;
     }
    }
    $2163 = HEAP8[$29>>0]|0;
    $2164 = ($2163<<24>>24)==(0);
    if ($2164) {
     label = 821;
    } else {
     label = 673;
     break L12;
    }
    break;
   }
   case 136:  {
    $2234 = HEAP32[$172>>2]|0;
    $2235 = $2234 & 1;
    $2236 = ($2235|0)!=(0);
    $$$i220 = $2236 ? 35 : 0;
    $2237 = $2234 & 2;
    $2238 = ($2237|0)==(0);
    if ($2238) {
     $K$0$i = $$$i220;
    } else {
     $2239 = HEAP8[$122>>0]|0;
     $2240 = ($2239<<24>>24)==(0);
     $2241 = $$$i220 | 128;
     $$$$i = $2240 ? $$$i220 : $2241;
     $K$0$i = $$$$i;
    }
    $2242 = $2234 & 4;
    $2243 = ($2242|0)==(0);
    if ($2243) {
     $K$1$i = $K$0$i;
    } else {
     $2244 = HEAP8[$120>>0]|0;
     $2245 = ($2244<<24>>24)==(0);
     $2246 = $K$0$i | 256;
     $K$0$$i = $2245 ? $K$0$i : $2246;
     $K$1$i = $K$0$$i;
    }
    $2247 = $2234 & 32;
    $2248 = ($2247|0)==(0);
    if ($2248) {
     $K$2$i = $K$1$i;
    } else {
     $2249 = HEAP8[$124>>0]|0;
     $2250 = ($2249<<24>>24)==(0);
     $2251 = $K$1$i | 4096;
     $K$1$$i = $2250 ? $K$1$i : $2251;
     $K$2$i = $K$1$$i;
    }
    HEAP32[$172>>2] = $K$2$i;
    label = 821;
    break;
   }
   case 137:  {
    $2252 = HEAP32[$62>>2]|0;
    $2253 = HEAP32[$63>>2]|0;
    $2254 = (($2252) + (($2253*24)|0)|0);
    $2255 = ($2253|0)>(0);
    L948: do {
     if ($2255) {
      $2256 = HEAP32[$172>>2]|0;
      $def$01$i212 = $2252;
      while(1) {
       $2257 = ((($def$01$i212)) + 12|0);
       $2258 = HEAP32[$2257>>2]|0;
       $2259 = ($2258|0)==($2256|0);
       if ($2259) {
        $def$0$lcssa$i213 = $def$01$i212;
        break L948;
       }
       $2260 = ((($def$01$i212)) + 24|0);
       $2261 = ($2260>>>0)<($2254>>>0);
       if ($2261) {
        $def$01$i212 = $2260;
       } else {
        $def$0$lcssa$i213 = $2260;
        break;
       }
      }
     } else {
      $def$0$lcssa$i213 = $2252;
     }
    } while(0);
    $2262 = ($def$0$lcssa$i213|0)==($2254|0);
    if ($2262) {
     $2263 = HEAP32[$125>>2]|0;
     $2264 = ($2253>>>0)<($2263>>>0);
     if (!($2264)) {
      label = 698;
      break L12;
     }
     $2265 = (($2253) + 1)|0;
     HEAP32[$63>>2] = $2265;
    }
    $2266 = HEAP32[$172>>2]|0;
    $2267 = ($2266>>>0)>(255);
    if ($2267) {
     label = 701;
     break L12;
    }
    $2268 = $2266 & 255;
    $2269 = ((($def$0$lcssa$i213)) + 12|0);
    HEAP32[$2269>>2] = $2268;
    $2270 = HEAP32[$22>>2]|0;
    $2271 = (($2270) + 1)|0;
    $2272 = ((($def$0$lcssa$i213)) + 4|0);
    HEAP32[$2272>>2] = $2271;
    $2273 = HEAP32[$67>>2]|0;
    HEAP32[$def$0$lcssa$i213>>2] = $2273;
    $2274 = ((($def$0$lcssa$i213)) + 16|0);
    HEAP8[$2274>>0] = 1;
    $2275 = HEAP32[$172>>2]|0;
    $2276 = HEAP32[$126>>2]|0;
    $2277 = ($2275>>>0)>($2276>>>0);
    if ($2277) {
     $2278 = $2275 & 255;
     HEAP32[$126>>2] = $2278;
    }
    $$pre4$i216 = HEAP32[$26>>2]|0;
    $2280 = $$pre$pre$i;$2281 = $2270;
    while(1) {
     $2279 = (($2280) + ($2281))|0;
     HEAP32[$22>>2] = $2279;
     $2282 = ($2279|0)<($$pre4$i216|0);
     if (!($2282)) {
      label = 710;
      break L12;
     }
     $2283 = (($129) + ($2279)|0);
     $2284 = HEAP8[$2283>>0]|0;
     HEAP8[$24>>0] = $2284;
     $2285 = $2284&255;
     $2286 = (593353 + ($2285)|0);
     $2287 = HEAP8[$2286>>0]|0;
     $2288 = $2287 << 24 >> 24;
     HEAP32[$25>>2] = $2288;
     $2289 = $2285 | 1;
     $2290 = ($2289|0)==(65);
     if ($2290) {
      $2291 = (($2279) + 1)|0;
      $2292 = ($2291|0)<($$pre4$i216|0);
      if (!($2292)) {
       label = 710;
       break L12;
      }
      $2293 = (($129) + ($2291)|0);
      $2294 = HEAP8[$2293>>0]|0;
      $2295 = $2294&255;
      $2296 = Math_imul($2295, $2288)|0;
      $2297 = (2 - ($2296))|0;
      HEAP32[$25>>2] = $2297;
      $2299 = $2297;
     } else {
      $2299 = $2288;
     }
     $2298 = (($2299) + ($2279))|0;
     $2300 = ($2298|0)>($$pre4$i216|0);
     if ($2300) {
      label = 710;
      break L12;
     }
     switch ($2285|0) {
     case 44: case 137:  {
      label = 712;
      break L12;
      break;
     }
     case 45:  {
      label = 821;
      break L30;
      break;
     }
     default: {
      $2280 = $2299;$2281 = $2279;
     }
     }
    }
    break;
   }
   case 138:  {
    $2301 = ((($172)) + 8|0);
    $2302 = HEAP32[$2301>>2]|0;
    $2303 = ((($172)) + 4|0);
    $2304 = HEAP32[$2303>>2]|0;
    $2305 = HEAP32[$172>>2]|0;
    HEAP32[$2301>>2] = $2305;
    HEAP32[$2303>>2] = $2302;
    HEAP32[$172>>2] = $2304;
    label = 821;
    break;
   }
   case 139:  {
    $2306 = ((($172)) + 4|0);
    $2307 = HEAP32[$2306>>2]|0;
    $2308 = HEAP32[$172>>2]|0;
    $2309 = ($2307|0)>($2308|0);
    if ($2309) {
     HEAP32[$172>>2] = $2307;
     label = 821;
    } else {
     label = 821;
    }
    break;
   }
   case 140:  {
    $2310 = ((($172)) + 4|0);
    $2311 = HEAP32[$2310>>2]|0;
    $2312 = HEAP32[$172>>2]|0;
    $2313 = ($2311|0)<($2312|0);
    if ($2313) {
     HEAP32[$172>>2] = $2311;
     label = 821;
    } else {
     label = 821;
    }
    break;
   }
   case 141:  {
    $$val42 = HEAP32[$172>>2]|0;
    $2314 = ($$val42|0)>(-1);
    if ($2314) {
     HEAP32[$127>>2] = $$val42;
     label = 821;
    } else {
     label = 821;
    }
    break;
   }
   case 142:  {
    $$val43 = HEAP32[$172>>2]|0;
    $$idx44 = ((($172)) + 4|0);
    $$idx44$val = HEAP32[$$idx44>>2]|0;
    $$off$i = (($$idx44$val) + -1)|0;
    $2315 = ($$off$i>>>0)>(2);
    if ($2315) {
     $2316 = HEAP8[$29>>0]|0;
     $2317 = ($2316<<24>>24)==(0);
     if ($2317) {
      label = 821;
      break L30;
     } else {
      label = 722;
      break L12;
     }
    }
    $2318 = 1 << $$off$i;
    $2319 = ($$val43|0)==(0);
    $2320 = ($2318|0)==($$val43|0);
    $or$cond$i210 = $2319 | $2320;
    if ($or$cond$i210) {
     $2323 = $2318 ^ 255;
     $2324 = HEAP8[$128>>0]|0;
     $2325 = $2324&255;
     $2326 = $2325 & $2323;
     $2327 = $2326 | $$val43;
     $2328 = $2327&255;
     HEAP8[$128>>0] = $2328;
     label = 821;
     break L30;
    } else {
     $2321 = HEAP8[$29>>0]|0;
     $2322 = ($2321<<24>>24)==(0);
     if ($2322) {
      label = 821;
      break L30;
     } else {
      label = 725;
      break L12;
     }
    }
    break;
   }
   case 143:  {
    $2329 = HEAP32[$62>>2]|0;
    $2330 = HEAP32[$63>>2]|0;
    $2331 = (($2329) + (($2330*24)|0)|0);
    $2332 = ($2330|0)>(0);
    L985: do {
     if ($2332) {
      $def$02$i205 = $2329;
      while(1) {
       $2333 = ((($def$02$i205)) + 12|0);
       $2334 = HEAP32[$2333>>2]|0;
       $2335 = $2334&255;
       $2336 = ($132<<24>>24)==($2335<<24>>24);
       if ($2336) {
        $2337 = ((($def$02$i205)) + 16|0);
        $2338 = HEAP8[$2337>>0]|0;
        $2339 = ($2338<<24>>24)==(0);
        if (!($2339)) {
         $def$02$i205$lcssa = $def$02$i205;
         break;
        }
       }
       $2362 = ((($def$02$i205)) + 24|0);
       $2363 = ($2362>>>0)<($2331>>>0);
       if ($2363) {
        $def$02$i205 = $2362;
       } else {
        break L985;
       }
      }
      $2340 = HEAP32[$64>>2]|0;
      $2341 = HEAP32[$65>>2]|0;
      $2342 = ($2340|0)<($2341|0);
      if (!($2342)) {
       label = 731;
       break L12;
      }
      $2343 = HEAP32[$66>>2]|0;
      $2344 = (($2340) + 1)|0;
      HEAP32[$64>>2] = $2344;
      $2345 = HEAP32[$67>>2]|0;
      $2346 = (($2343) + ($2340<<4)|0);
      HEAP32[$2346>>2] = $2345;
      $2347 = HEAP32[$22>>2]|0;
      $2348 = (($2347) + 1)|0;
      $2349 = (((($2343) + ($2340<<4)|0)) + 4|0);
      HEAP32[$2349>>2] = $2348;
      $2350 = (((($2343) + ($2340<<4)|0)) + 8|0);
      HEAP32[$2350>>2] = 1;
      $2351 = (((($2343) + ($2340<<4)|0)) + 12|0);
      HEAP32[$2351>>2] = $def$02$i205$lcssa;
      $2352 = HEAP32[$def$02$i205$lcssa>>2]|0;
      $2353 = ((($def$02$i205$lcssa)) + 4|0);
      $2354 = HEAP32[$2353>>2]|0;
      $aRange$off$i$i206 = (($2352) + -1)|0;
      $2355 = ($aRange$off$i$i206>>>0)>(2);
      do {
       if ($2355) {
        HEAP32[$34>>2] = 132;
       } else {
        $2356 = (((($exc)) + 444|0) + ($aRange$off$i$i206<<3)|0);
        $2357 = HEAP32[$2356>>2]|0;
        $2358 = ($2357|0)==(0|0);
        if ($2358) {
         HEAP32[$34>>2] = 138;
         break;
        }
        $2359 = (((((($exc)) + 444|0) + ($aRange$off$i$i206<<3)|0)) + 4|0);
        $2360 = HEAP32[$2359>>2]|0;
        $2361 = ($2360|0)<($2354|0);
        if ($2361) {
         HEAP32[$34>>2] = 131;
         break;
        } else {
         HEAP32[$23>>2] = $2357;
         HEAP32[$26>>2] = $2360;
         HEAP32[$22>>2] = $2354;
         HEAP32[$67>>2] = $2352;
         break;
        }
       }
      } while(0);
      HEAP8[$33>>0] = 0;
      label = 821;
      break L30;
     }
    } while(0);
    HEAP32[$34>>2] = 128;
    $2736 = $2329;$2737 = $2330;
    label = 823;
    break;
   }
   default: {
    $2364 = ($132&255)>(223);
    if ($2364) {
     $$val45 = HEAP32[$172>>2]|0;
     $$idx46 = ((($172)) + 4|0);
     $$idx46$val = HEAP32[$$idx46>>2]|0;
     $2365 = HEAP32[$35>>2]|0;
     $2366 = HEAP32[$36>>2]|0;
     $2367 = $$val45&65535;
     $2368 = (($$idx46$val) + 1)|0;
     $2369 = $$val45 & 65535;
     $2370 = HEAP16[$38>>1]|0;
     $2371 = $2370&65535;
     $2372 = ($2369>>>0)<($2371>>>0);
     if ($2372) {
      $2373 = HEAP32[$39>>2]|0;
      $2374 = (($2373) + 1)|0;
      $2375 = ($2368>>>0)<($2374>>>0);
      if ($2375) {
       $2376 = HEAP16[$40>>1]|0;
       $2377 = HEAP16[$41>>1]|0;
       $2378 = ($2376&65535)<($2377&65535);
       if ($2378) {
        $2381 = ($2368|0)==(0);
        if ($2381) {
         $cvt_dist$0$i = 0;
        } else {
         $2382 = HEAP32[$8>>2]|0;
         $2383 = (FUNCTION_TABLE_iii[$2382 & 8191]($exc,$$idx46$val)|0);
         $cvt_dist$0$i = $2383;
        }
        $2384 = HEAP32[$42>>2]|0;
        $2385 = (($cvt_dist$0$i) - ($2384))|0;
        $2386 = ($2385|0)<(0);
        $2387 = (0 - ($2385))|0;
        $2388 = $2386 ? $2387 : $2385;
        $2389 = HEAP32[$43>>2]|0;
        $2390 = ($2388|0)<($2389|0);
        if ($2390) {
         $2391 = ($cvt_dist$0$i|0)>(-1);
         $2392 = (0 - ($2384))|0;
         $$$i94 = $2391 ? $2384 : $2392;
         $cvt_dist$1$i = $$$i94;
        } else {
         $cvt_dist$1$i = $cvt_dist$0$i;
        }
        $2393 = HEAP16[$44>>1]|0;
        $2394 = ($2393<<16>>16)==(0);
        if ($2394) {
         $2395 = HEAP16[$40>>1]|0;
         $2396 = $2395&65535;
         $2397 = HEAP32[$45>>2]|0;
         $2398 = (($2397) + ($2396<<3)|0);
         $2399 = HEAP32[$2398>>2]|0;
         $2400 = HEAP16[$46>>1]|0;
         $2401 = $2400 << 16 >> 16;
         $2402 = $2401 ^ $cvt_dist$1$i;
         $2403 = ($cvt_dist$1$i|0)<(0);
         $2404 = (0 - ($cvt_dist$1$i))|0;
         $$a$i$i = $2403 ? $2404 : $cvt_dist$1$i;
         $2405 = ($2400<<16>>16)<(0);
         $2406 = (0 - ($2401))|0;
         $$01$i$i = $2405 ? $2406 : $2401;
         $2407 = $$a$i$i >>> 16;
         $2408 = $$a$i$i & 65535;
         $2409 = Math_imul($$01$i$i, $2408)|0;
         $2410 = Math_imul($$01$i$i, $2407)|0;
         $2411 = $2410 >>> 16;
         $2412 = $2410 << 16;
         $2413 = $2412 | 8192;
         $2414 = (($2413) + ($2409))|0;
         $2415 = ($2414>>>0)<($2413>>>0);
         $2416 = $2415&1;
         $$$i$i98 = (($2416) + ($2411))|0;
         $2417 = $2414 >>> 14;
         $2418 = $$$i$i98 << 18;
         $2419 = $2418 | $2417;
         $2420 = ($2402|0)>(-1);
         $2421 = (0 - ($2419))|0;
         $2422 = $2420 ? $2419 : $2421;
         $2423 = (($2422) + ($2399))|0;
         $2424 = HEAP32[$47>>2]|0;
         $2425 = (($2424) + ($2369<<3)|0);
         HEAP32[$2425>>2] = $2423;
         $2426 = (((($2397) + ($2396<<3)|0)) + 4|0);
         $2427 = HEAP32[$2426>>2]|0;
         $2428 = HEAP16[$48>>1]|0;
         $2429 = $2428 << 16 >> 16;
         $2430 = $2429 ^ $cvt_dist$1$i;
         $2431 = ($2428<<16>>16)<(0);
         $2432 = (0 - ($2429))|0;
         $$01$i4$i = $2431 ? $2432 : $2429;
         $2433 = Math_imul($$01$i4$i, $2408)|0;
         $2434 = Math_imul($$01$i4$i, $2407)|0;
         $2435 = $2434 >>> 16;
         $2436 = $2434 << 16;
         $2437 = $2436 | 8192;
         $2438 = (($2437) + ($2433))|0;
         $2439 = ($2438>>>0)<($2437>>>0);
         $2440 = $2439&1;
         $$$i5$i = (($2440) + ($2435))|0;
         $2441 = $2438 >>> 14;
         $2442 = $$$i5$i << 18;
         $2443 = $2442 | $2441;
         $2444 = ($2430|0)>(-1);
         $2445 = (0 - ($2443))|0;
         $2446 = $2444 ? $2443 : $2445;
         $2447 = (($2446) + ($2427))|0;
         $2448 = (((($2424) + ($2369<<3)|0)) + 4|0);
         HEAP32[$2448>>2] = $2447;
         $2449 = HEAP32[$49>>2]|0;
         $2450 = (($2449) + ($2369<<3)|0);
         $2451 = (($2424) + ($2369<<3)|0);
         $2452 = $2451;
         $2453 = $2452;
         $2454 = HEAP32[$2453>>2]|0;
         $2455 = (($2452) + 4)|0;
         $2456 = $2455;
         $2457 = HEAP32[$2456>>2]|0;
         $2458 = $2450;
         $2459 = $2458;
         HEAP32[$2459>>2] = $2454;
         $2460 = (($2458) + 4)|0;
         $2461 = $2460;
         HEAP32[$2461>>2] = $2457;
         $$pre$phi7$iZ2D = $2404;
        } else {
         $$pre6$i = (0 - ($cvt_dist$1$i))|0;
         $$pre$phi7$iZ2D = $$pre6$i;
        }
        $2462 = HEAP32[$50>>2]|0;
        $2463 = HEAP32[$47>>2]|0;
        $2464 = (($2463) + ($2369<<3)|0);
        $2465 = HEAP32[$2464>>2]|0;
        $2466 = HEAP16[$40>>1]|0;
        $2467 = $2466&65535;
        $2468 = HEAP32[$45>>2]|0;
        $2469 = (($2468) + ($2467<<3)|0);
        $2470 = HEAP32[$2469>>2]|0;
        $2471 = (($2465) - ($2470))|0;
        $2472 = (((($2463) + ($2369<<3)|0)) + 4|0);
        $2473 = HEAP32[$2472>>2]|0;
        $2474 = (((($2468) + ($2467<<3)|0)) + 4|0);
        $2475 = HEAP32[$2474>>2]|0;
        $2476 = (($2473) - ($2475))|0;
        $2477 = (FUNCTION_TABLE_iiii[$2462 & 8191]($exc,$2471,$2476)|0);
        $2478 = HEAP32[$51>>2]|0;
        $2479 = HEAP32[$49>>2]|0;
        $2480 = (($2479) + ($2369<<3)|0);
        $2481 = HEAP32[$2480>>2]|0;
        $2482 = HEAP16[$40>>1]|0;
        $2483 = $2482&65535;
        $2484 = HEAP32[$52>>2]|0;
        $2485 = (($2484) + ($2483<<3)|0);
        $2486 = HEAP32[$2485>>2]|0;
        $2487 = (($2481) - ($2486))|0;
        $2488 = (((($2479) + ($2369<<3)|0)) + 4|0);
        $2489 = HEAP32[$2488>>2]|0;
        $2490 = (((($2484) + ($2483<<3)|0)) + 4|0);
        $2491 = HEAP32[$2490>>2]|0;
        $2492 = (($2489) - ($2491))|0;
        $2493 = (FUNCTION_TABLE_iiii[$2478 & 8191]($exc,$2487,$2492)|0);
        $2494 = HEAP8[$53>>0]|0;
        $2495 = ($2494<<24>>24)!=(0);
        $2496 = $2477 ^ $cvt_dist$1$i;
        $2497 = ($2496|0)<(0);
        $or$cond$i99 = $2497 & $2495;
        $cvt_dist$2$i = $or$cond$i99 ? $$pre$phi7$iZ2D : $cvt_dist$1$i;
        $2498 = HEAP8[$24>>0]|0;
        $2499 = $2498&255;
        $2500 = $2499 & 4;
        $2501 = ($2500|0)==(0);
        do {
         if ($2501) {
          $2517 = $2499 & 3;
          $2518 = (((($exc)) + 264|0) + ($2517<<2)|0);
          $2519 = HEAP32[$2518>>2]|0;
          $2520 = ($cvt_dist$2$i|0)>(-1);
          if ($2520) {
           $2521 = (($2519) + ($cvt_dist$2$i))|0;
           $2522 = ($2521|0)<(0);
           $$$i6$i = $2522 ? 0 : $2521;
           $2526 = $2498;$distance$0$i102 = $$$i6$i;
           break;
          } else {
           $2523 = (($cvt_dist$2$i) - ($2519))|0;
           $2524 = ($2523|0)>(0);
           $$1$i$i101 = $2524 ? 0 : $2523;
           $2526 = $2498;$distance$0$i102 = $$1$i$i101;
           break;
          }
         } else {
          $2502 = HEAP32[$55>>2]|0;
          $2503 = $2502&65535;
          $2504 = $2502 >>> 16;
          $2505 = $2504&65535;
          $2506 = ($2503<<16>>16)==($2505<<16>>16);
          if ($2506) {
           $2507 = (($cvt_dist$2$i) - ($2477))|0;
           $2508 = ($2507|0)<(0);
           $2509 = (0 - ($2507))|0;
           $2510 = $2508 ? $2509 : $2507;
           $2511 = ($2510|0)>($2366|0);
           $$cvt_dist$2$i = $2511 ? $2477 : $cvt_dist$2$i;
           $cvt_dist$3$i = $$cvt_dist$2$i;
          } else {
           $cvt_dist$3$i = $cvt_dist$2$i;
          }
          $2512 = HEAP32[$$pre$phiZ2D>>2]|0;
          $2513 = $2499 & 3;
          $2514 = (((($exc)) + 264|0) + ($2513<<2)|0);
          $2515 = HEAP32[$2514>>2]|0;
          $2516 = (FUNCTION_TABLE_iiii[$2512 & 8191]($exc,$cvt_dist$3$i,$2515)|0);
          $$pre$i100 = HEAP8[$24>>0]|0;
          $2526 = $$pre$i100;$distance$0$i102 = $2516;
         }
        } while(0);
        $2525 = $2526 & 8;
        $2527 = ($2525<<24>>24)==(0);
        do {
         if ($2527) {
          $distance$1$i105 = $distance$0$i102;
         } else {
          $2528 = ($2477|0)>(-1);
          if ($2528) {
           $2529 = ($distance$0$i102|0)<($2365|0);
           $$distance$0$i104 = $2529 ? $2365 : $distance$0$i102;
           $distance$1$i105 = $$distance$0$i104;
           break;
          } else {
           $2530 = (0 - ($2365))|0;
           $2531 = ($distance$0$i102|0)>($2530|0);
           $$distance$02$i = $2531 ? $2530 : $distance$0$i102;
           $distance$1$i105 = $$distance$02$i;
           break;
          }
         }
        } while(0);
        $2532 = HEAP32[$54>>2]|0;
        $2533 = (($distance$1$i105) - ($2493))|0;
        FUNCTION_TABLE_viiii[$2532 & 8191]($exc,$37,$2367,$2533);
        $$pre454 = HEAP8[$24>>0]|0;
        $2536 = $$pre454;
       } else {
        label = 746;
       }
      } else {
       label = 746;
      }
     } else {
      label = 746;
     }
     if ((label|0) == 746) {
      label = 0;
      $2379 = HEAP8[$29>>0]|0;
      $2380 = ($2379<<24>>24)==(0);
      if ($2380) {
       $2536 = $132;
      } else {
       HEAP32[$34>>2] = 134;
       $2536 = $132;
      }
     }
     $2534 = HEAP16[$40>>1]|0;
     HEAP16[$56>>1] = $2534;
     $2535 = $2536 & 16;
     $2537 = ($2535<<24>>24)==(0);
     if (!($2537)) {
      HEAP16[$40>>1] = $2367;
     }
     HEAP16[$57>>1] = $2367;
     label = 821;
     break L30;
    }
    $2538 = ($132&255)>(191);
    if ($2538) {
     $$val47 = HEAP32[$172>>2]|0;
     $2539 = HEAP32[$35>>2]|0;
     $2540 = $$val47&65535;
     $2541 = $$val47 & 65535;
     $2542 = HEAP16[$38>>1]|0;
     $2543 = $2542&65535;
     $2544 = ($2541>>>0)<($2543>>>0);
     if ($2544) {
      $2545 = HEAP16[$40>>1]|0;
      $2546 = $2545&65535;
      $2547 = HEAP16[$41>>1]|0;
      $2548 = ($2545&65535)<($2547&65535);
      if ($2548) {
       $2551 = HEAP32[$55>>2]|0;
       $2552 = $2551&65535;
       $2553 = ($2552<<16>>16)==(0);
       $2554 = ($2551>>>0)<(65536);
       $or$cond$i = $2554 | $2553;
       do {
        if ($or$cond$i) {
         $2555 = HEAP32[$47>>2]|0;
         $2556 = HEAP32[$45>>2]|0;
         $2557 = HEAP32[$50>>2]|0;
         $2558 = (($2555) + ($2541<<3)|0);
         $2559 = HEAP32[$2558>>2]|0;
         $2560 = (($2556) + ($2546<<3)|0);
         $2561 = HEAP32[$2560>>2]|0;
         $2562 = (($2559) - ($2561))|0;
         $2563 = (((($2555) + ($2541<<3)|0)) + 4|0);
         $2564 = HEAP32[$2563>>2]|0;
         $2565 = (((($2556) + ($2546<<3)|0)) + 4|0);
         $2566 = HEAP32[$2565>>2]|0;
         $2567 = (($2564) - ($2566))|0;
         $2568 = (FUNCTION_TABLE_iiii[$2557 & 8191]($exc,$2562,$2567)|0);
         $org_dist$1$i = $2568;
        } else {
         $2569 = HEAP32[$58>>2]|0;
         $2570 = HEAP32[$59>>2]|0;
         $2571 = HEAP32[$60>>2]|0;
         $2572 = HEAP32[$61>>2]|0;
         $2573 = ($2571|0)==($2572|0);
         $2574 = (($2569) + ($2541<<3)|0);
         $2575 = HEAP32[$2574>>2]|0;
         $2576 = (($2570) + ($2546<<3)|0);
         $2577 = HEAP32[$2576>>2]|0;
         if ($2573) {
          $2578 = HEAP32[$50>>2]|0;
          $2579 = (($2575) - ($2577))|0;
          $2580 = (((($2569) + ($2541<<3)|0)) + 4|0);
          $2581 = HEAP32[$2580>>2]|0;
          $2582 = (((($2570) + ($2546<<3)|0)) + 4|0);
          $2583 = HEAP32[$2582>>2]|0;
          $2584 = (($2581) - ($2583))|0;
          $2585 = (FUNCTION_TABLE_iiii[$2578 & 8191]($exc,$2579,$2584)|0);
          $2586 = HEAP32[$60>>2]|0;
          $2587 = (_FT_MulFix($2585,$2586)|0);
          $org_dist$1$i = $2587;
          break;
         } else {
          $2588 = (($2575) - ($2577))|0;
          $2589 = (_FT_MulFix($2588,$2571)|0);
          $2590 = (((($2569) + ($2541<<3)|0)) + 4|0);
          $2591 = HEAP32[$2590>>2]|0;
          $2592 = (((($2570) + ($2546<<3)|0)) + 4|0);
          $2593 = HEAP32[$2592>>2]|0;
          $2594 = (($2591) - ($2593))|0;
          $2595 = HEAP32[$61>>2]|0;
          $2596 = (_FT_MulFix($2594,$2595)|0);
          $2597 = HEAP32[$50>>2]|0;
          $2598 = (FUNCTION_TABLE_iiii[$2597 & 8191]($exc,$2589,$2596)|0);
          $org_dist$1$i = $2598;
          break;
         }
        }
       } while(0);
       $2599 = HEAP32[$42>>2]|0;
       $2600 = (($org_dist$1$i) - ($2599))|0;
       $2601 = ($2600|0)<(0);
       $2602 = (0 - ($2600))|0;
       $2603 = $2601 ? $2602 : $2600;
       $2604 = HEAP32[$43>>2]|0;
       $2605 = ($2603|0)<($2604|0);
       if ($2605) {
        $2606 = ($org_dist$1$i|0)>(-1);
        $2607 = (0 - ($2599))|0;
        $$$i = $2606 ? $2599 : $2607;
        $org_dist$2$i = $$$i;
       } else {
        $org_dist$2$i = $org_dist$1$i;
       }
       $2608 = HEAP8[$24>>0]|0;
       $2609 = $2608&255;
       $2610 = $2609 & 4;
       $2611 = ($2610|0)==(0);
       do {
        if ($2611) {
         $2617 = $2609 & 3;
         $2618 = (((($exc)) + 264|0) + ($2617<<2)|0);
         $2619 = HEAP32[$2618>>2]|0;
         $2620 = ($org_dist$2$i|0)>(-1);
         if ($2620) {
          $2621 = (($2619) + ($org_dist$2$i))|0;
          $2622 = ($2621|0)<(0);
          $$$i$i = $2622 ? 0 : $2621;
          $2626 = $2608;$distance$0$i = $$$i$i;
          break;
         } else {
          $2623 = (($org_dist$2$i) - ($2619))|0;
          $2624 = ($2623|0)>(0);
          $$1$i$i = $2624 ? 0 : $2623;
          $2626 = $2608;$distance$0$i = $$1$i$i;
          break;
         }
        } else {
         $2612 = HEAP32[$$pre$phiZ2D>>2]|0;
         $2613 = $2609 & 3;
         $2614 = (((($exc)) + 264|0) + ($2613<<2)|0);
         $2615 = HEAP32[$2614>>2]|0;
         $2616 = (FUNCTION_TABLE_iiii[$2612 & 8191]($exc,$org_dist$2$i,$2615)|0);
         $$pre$i53 = HEAP8[$24>>0]|0;
         $2626 = $$pre$i53;$distance$0$i = $2616;
        }
       } while(0);
       $2625 = $2626 & 8;
       $2627 = ($2625<<24>>24)==(0);
       do {
        if ($2627) {
         $distance$1$i = $distance$0$i;
        } else {
         $2628 = ($org_dist$2$i|0)>(-1);
         if ($2628) {
          $2629 = ($distance$0$i|0)<($2539|0);
          $$distance$0$i = $2629 ? $2539 : $distance$0$i;
          $distance$1$i = $$distance$0$i;
          break;
         } else {
          $2630 = (0 - ($2539))|0;
          $2631 = ($distance$0$i|0)>($2630|0);
          $$distance$01$i = $2631 ? $2630 : $distance$0$i;
          $distance$1$i = $$distance$01$i;
          break;
         }
        }
       } while(0);
       $2632 = HEAP32[$51>>2]|0;
       $2633 = HEAP32[$49>>2]|0;
       $2634 = (($2633) + ($2541<<3)|0);
       $2635 = HEAP32[$2634>>2]|0;
       $2636 = HEAP32[$52>>2]|0;
       $2637 = HEAP16[$40>>1]|0;
       $2638 = $2637&65535;
       $2639 = (($2636) + ($2638<<3)|0);
       $2640 = HEAP32[$2639>>2]|0;
       $2641 = (($2635) - ($2640))|0;
       $2642 = (((($2633) + ($2541<<3)|0)) + 4|0);
       $2643 = HEAP32[$2642>>2]|0;
       $2644 = (((($2636) + ($2638<<3)|0)) + 4|0);
       $2645 = HEAP32[$2644>>2]|0;
       $2646 = (($2643) - ($2645))|0;
       $2647 = (FUNCTION_TABLE_iiii[$2632 & 8191]($exc,$2641,$2646)|0);
       $2648 = HEAP32[$54>>2]|0;
       $2649 = (($distance$1$i) - ($2647))|0;
       FUNCTION_TABLE_viiii[$2648 & 8191]($exc,$37,$2540,$2649);
       $$pre453 = HEAP8[$24>>0]|0;
       $2652 = $$pre453;
      } else {
       label = 773;
      }
     } else {
      label = 773;
     }
     if ((label|0) == 773) {
      label = 0;
      $2549 = HEAP8[$29>>0]|0;
      $2550 = ($2549<<24>>24)==(0);
      if ($2550) {
       $2652 = $132;
      } else {
       HEAP32[$34>>2] = 134;
       $2652 = $132;
      }
     }
     $2650 = HEAP16[$40>>1]|0;
     HEAP16[$56>>1] = $2650;
     HEAP16[$57>>1] = $2540;
     $2651 = $2652 & 16;
     $2653 = ($2651<<24>>24)==(0);
     if ($2653) {
      label = 821;
      break L30;
     }
     HEAP16[$40>>1] = $2540;
     label = 821;
     break L30;
    }
    $2654 = ($132&255)>(183);
    if ($2654) {
     $2655 = (($133) + 65353)|0;
     $2656 = $2655 & 65535;
     $2657 = (($169) + 1)|0;
     $2658 = HEAP32[$27>>2]|0;
     $2659 = (($2657) - ($2658))|0;
     $2660 = ($2656>>>0)<($2659>>>0);
     if (!($2660)) {
      label = 796;
      break L12;
     }
     $2661 = HEAP32[$22>>2]|0;
     $2662 = (($2661) + 1)|0;
     HEAP32[$22>>2] = $2662;
     $2663 = ($2656|0)==(0);
     L109: do {
      if (!($2663)) {
       $2665 = $2662;$2677 = 0;$K$01$i50 = 0;
       while(1) {
        $2664 = (($2665) + 2)|0;
        HEAP32[$22>>2] = $2664;
        $2666 = (($129) + ($2665)|0);
        $2667 = HEAP8[$2666>>0]|0;
        $2668 = $2667&255;
        $2669 = $2668 << 8;
        $2670 = (($2665) + 1)|0;
        $2671 = (($129) + ($2670)|0);
        $2672 = HEAP8[$2671>>0]|0;
        $2673 = $2672&255;
        $2674 = $2669 | $2673;
        $sext$i = $2674 << 16;
        $2675 = $sext$i >> 16;
        $2676 = (($172) + ($2677<<2)|0);
        HEAP32[$2676>>2] = $2675;
        $2678 = (($K$01$i50) + 1)<<16>>16;
        $2679 = $2678&65535;
        $2680 = ($2679>>>0)<($2656>>>0);
        if (!($2680)) {
         break L109;
        }
        $$pre$i51 = HEAP32[$22>>2]|0;
        $2665 = $$pre$i51;$2677 = $2679;$K$01$i50 = $2678;
       }
      }
     } while(0);
     HEAP8[$33>>0] = 0;
     label = 821;
     break L30;
    }
    $2681 = ($132&255)>(175);
    if ($2681) {
     $2682 = (($133) + 65361)|0;
     $2683 = $2682 & 65535;
     $2684 = (($169) + 1)|0;
     $2685 = HEAP32[$27>>2]|0;
     $2686 = (($2684) - ($2685))|0;
     $2687 = ($2683>>>0)<($2686>>>0);
     if (!($2687)) {
      label = 804;
      break L12;
     }
     $2688 = ($2683|0)==(0);
     if ($2688) {
      label = 821;
      break L30;
     } else {
      $2691 = 1;$K$01$i = 1;
     }
     while(1) {
      $2689 = HEAP32[$22>>2]|0;
      $2690 = (($2689) + ($2691))|0;
      $2692 = (($129) + ($2690)|0);
      $2693 = HEAP8[$2692>>0]|0;
      $2694 = $2693&255;
      $2695 = (($2691) + -1)|0;
      $2696 = (($172) + ($2695<<2)|0);
      HEAP32[$2696>>2] = $2694;
      $2697 = (($K$01$i) + 1)<<16>>16;
      $2698 = $2697&65535;
      $2699 = ($2698>>>0)>($2683>>>0);
      if ($2699) {
       label = 821;
       break L30;
      } else {
       $2691 = $2698;$K$01$i = $2697;
      }
     }
    }
    $2700 = HEAP32[$62>>2]|0;
    $2701 = HEAP32[$63>>2]|0;
    $2702 = (($2700) + (($2701*24)|0)|0);
    $2703 = ($2701|0)>(0);
    L121: do {
     if ($2703) {
      $def$02$i = $2700;
      while(1) {
       $2704 = ((($def$02$i)) + 12|0);
       $2705 = HEAP32[$2704>>2]|0;
       $2706 = $2705&255;
       $2707 = ($132<<24>>24)==($2706<<24>>24);
       if ($2707) {
        $2708 = ((($def$02$i)) + 16|0);
        $2709 = HEAP8[$2708>>0]|0;
        $2710 = ($2709<<24>>24)==(0);
        if (!($2710)) {
         $def$02$i$lcssa = $def$02$i;
         break;
        }
       }
       $2733 = ((($def$02$i)) + 24|0);
       $2734 = ($2733>>>0)<($2702>>>0);
       if ($2734) {
        $def$02$i = $2733;
       } else {
        break L121;
       }
      }
      $2711 = HEAP32[$64>>2]|0;
      $2712 = HEAP32[$65>>2]|0;
      $2713 = ($2711|0)<($2712|0);
      if (!($2713)) {
       label = 810;
       break L12;
      }
      $2714 = HEAP32[$66>>2]|0;
      $2715 = (($2711) + 1)|0;
      HEAP32[$64>>2] = $2715;
      $2716 = HEAP32[$67>>2]|0;
      $2717 = (($2714) + ($2711<<4)|0);
      HEAP32[$2717>>2] = $2716;
      $2718 = HEAP32[$22>>2]|0;
      $2719 = (($2718) + 1)|0;
      $2720 = (((($2714) + ($2711<<4)|0)) + 4|0);
      HEAP32[$2720>>2] = $2719;
      $2721 = (((($2714) + ($2711<<4)|0)) + 8|0);
      HEAP32[$2721>>2] = 1;
      $2722 = (((($2714) + ($2711<<4)|0)) + 12|0);
      HEAP32[$2722>>2] = $def$02$i$lcssa;
      $2723 = HEAP32[$def$02$i$lcssa>>2]|0;
      $2724 = ((($def$02$i$lcssa)) + 4|0);
      $2725 = HEAP32[$2724>>2]|0;
      $aRange$off$i$i = (($2723) + -1)|0;
      $2726 = ($aRange$off$i$i>>>0)>(2);
      do {
       if ($2726) {
        HEAP32[$34>>2] = 132;
       } else {
        $2727 = (((($exc)) + 444|0) + ($aRange$off$i$i<<3)|0);
        $2728 = HEAP32[$2727>>2]|0;
        $2729 = ($2728|0)==(0|0);
        if ($2729) {
         HEAP32[$34>>2] = 138;
         break;
        }
        $2730 = (((((($exc)) + 444|0) + ($aRange$off$i$i<<3)|0)) + 4|0);
        $2731 = HEAP32[$2730>>2]|0;
        $2732 = ($2731|0)<($2725|0);
        if ($2732) {
         HEAP32[$34>>2] = 131;
         break;
        } else {
         HEAP32[$23>>2] = $2728;
         HEAP32[$26>>2] = $2731;
         HEAP32[$22>>2] = $2725;
         HEAP32[$67>>2] = $2723;
         break;
        }
       }
      } while(0);
      HEAP8[$33>>0] = 0;
      label = 821;
      break L30;
     }
    } while(0);
    HEAP32[$34>>2] = 128;
    $2736 = $2700;$2737 = $2701;
    label = 823;
   }
   }
  } while(0);
  L1004: do {
   if ((label|0) == 821) {
    label = 0;
    $$pr$pr = HEAP32[$34>>2]|0;
    switch ($$pr$pr|0) {
    case 0:  {
     label = 837;
     break L1004;
     break;
    }
    case 128:  {
     break;
    }
    default: {
     $$pr$pr$lcssa = $$pr$pr;
     label = 845;
     break L12;
    }
    }
    $$pre455 = HEAP32[$62>>2]|0;
    $$pre456 = HEAP32[$63>>2]|0;
    $2736 = $$pre455;$2737 = $$pre456;
    label = 823;
   }
  } while(0);
  if ((label|0) == 823) {
   label = 0;
   $2735 = (($2736) + (($2737*24)|0)|0);
   $2738 = ($2737|0)>(0);
   if ($2738) {
    $def$0371 = $2736;
   } else {
    label = 836;
    break;
   }
   while(1) {
    $2739 = ((($def$0371)) + 16|0);
    $2740 = HEAP8[$2739>>0]|0;
    $2741 = ($2740<<24>>24)==(0);
    if (!($2741)) {
     $2742 = HEAP8[$24>>0]|0;
     $2743 = ((($def$0371)) + 12|0);
     $2744 = HEAP32[$2743>>2]|0;
     $2745 = $2744&255;
     $2746 = ($2742<<24>>24)==($2745<<24>>24);
     if ($2746) {
      $def$0371$lcssa = $def$0371;
      break;
     }
    }
    $2768 = ((($def$0371)) + 24|0);
    $2769 = ($2768>>>0)<($2735>>>0);
    if ($2769) {
     $def$0371 = $2768;
    } else {
     label = 836;
     break L12;
    }
   }
   $2747 = HEAP32[$64>>2]|0;
   $2748 = HEAP32[$65>>2]|0;
   $2749 = ($2747|0)<($2748|0);
   if (!($2749)) {
    label = 827;
    break;
   }
   $2750 = HEAP32[$66>>2]|0;
   $2751 = HEAP32[$67>>2]|0;
   $2752 = (($2750) + ($2747<<4)|0);
   HEAP32[$2752>>2] = $2751;
   $2753 = HEAP32[$22>>2]|0;
   $2754 = (($2753) + 1)|0;
   $2755 = (((($2750) + ($2747<<4)|0)) + 4|0);
   HEAP32[$2755>>2] = $2754;
   $2756 = (((($2750) + ($2747<<4)|0)) + 8|0);
   HEAP32[$2756>>2] = 1;
   $2757 = (((($2750) + ($2747<<4)|0)) + 12|0);
   HEAP32[$2757>>2] = $def$0371$lcssa;
   $2758 = HEAP32[$def$0371$lcssa>>2]|0;
   $2759 = ((($def$0371$lcssa)) + 4|0);
   $2760 = HEAP32[$2759>>2]|0;
   $aRange$off$i = (($2758) + -1)|0;
   $2761 = ($aRange$off$i>>>0)>(2);
   if ($2761) {
    label = 829;
    break;
   }
   $2762 = (((($exc)) + 444|0) + ($aRange$off$i<<3)|0);
   $2763 = HEAP32[$2762>>2]|0;
   $2764 = ($2763|0)==(0|0);
   if ($2764) {
    label = 831;
    break;
   }
   $2765 = (((((($exc)) + 444|0) + ($aRange$off$i<<3)|0)) + 4|0);
   $2766 = HEAP32[$2765>>2]|0;
   $2767 = ($2766|0)<($2760|0);
   if ($2767) {
    label = 834;
    break;
   }
   HEAP32[$23>>2] = $2763;
   HEAP32[$26>>2] = $2766;
   HEAP32[$22>>2] = $2760;
   HEAP32[$67>>2] = $2758;
   $ins_counter$1 = $ins_counter$0;
  }
  else if ((label|0) == 837) {
   label = 0;
   $2770 = HEAP32[$31>>2]|0;
   HEAP32[$27>>2] = $2770;
   $2771 = HEAP8[$33>>0]|0;
   $2772 = ($2771<<24>>24)==(0);
   if (!($2772)) {
    $2773 = HEAP32[$25>>2]|0;
    $2774 = HEAP32[$22>>2]|0;
    $2775 = (($2774) + ($2773))|0;
    HEAP32[$22>>2] = $2775;
   }
   $2776 = (($ins_counter$0) + 1)|0;
   $2777 = ($ins_counter$0|0)>(999999);
   if ($2777) {
    $$01 = 139;
    label = 849;
    break;
   } else {
    $ins_counter$1 = $2776;
   }
  }
  $2778 = HEAP32[$22>>2]|0;
  $2779 = HEAP32[$26>>2]|0;
  $2780 = ($2778|0)<($2779|0);
  if (!($2780)) {
   label = 841;
   break;
  }
  $2783 = HEAP8[$68>>0]|0;
  $2784 = ($2783<<24>>24)==(0);
  if ($2784) {
   $131 = $2778;$ins_counter$0 = $ins_counter$1;
  } else {
   $$01 = 0;
   label = 849;
   break;
  }
 }
 switch (label|0) {
  case 21: {
   HEAP32[$34>>2] = 129;
   $2794 = 129;
   break;
  }
  case 25: {
   HEAP32[$34>>2] = 130;
   $2794 = 130;
   break;
  }
  case 35: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 45: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 73: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 84: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 90: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 96: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 102: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 105: {
   HEAP32[$34>>2] = 132;
   $2794 = 132;
   break;
  }
  case 117: {
   HEAP32[$34>>2] = 131;
   $2794 = 131;
   break;
  }
  case 143: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 148: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 154: {
   HEAP32[$34>>2] = 130;
   $2794 = 130;
   break;
  }
  case 167: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 180: {
   HEAP32[$34>>2] = 130;
   $2794 = 130;
   break;
  }
  case 190: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 200: {
   HEAP32[$34>>2] = 130;
   $2794 = 130;
   break;
  }
  case 209: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 215: {
   HEAP32[$34>>2] = 140;
   $2794 = 140;
   break;
  }
  case 218: {
   HEAP32[$34>>2] = 140;
   $2794 = 140;
   break;
  }
  case 227: {
   HEAP32[$34>>2] = 131;
   $2794 = 131;
   break;
  }
  case 229: {
   HEAP32[$34>>2] = 137;
   $2794 = 137;
   break;
  }
  case 232: {
   HEAP32[$34>>2] = 136;
   $2794 = 136;
   break;
  }
  case 236: {
   HEAP32[$34>>2] = 132;
   $2794 = 132;
   break;
  }
  case 238: {
   HEAP32[$34>>2] = 138;
   $2794 = 138;
   break;
  }
  case 240: {
   HEAP32[$34>>2] = 131;
   $2794 = 131;
   break;
  }
  case 244: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 341: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 376: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 382: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 395: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 414: {
   HEAP32[$34>>2] = 130;
   $2794 = 130;
   break;
  }
  case 418: {
   HEAP32[$34>>2] = 130;
   $2794 = 130;
   break;
  }
  case 426: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 430: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 435: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 439: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 451: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 470: {
   HEAP32[$34>>2] = 135;
   $2794 = 135;
   break;
  }
  case 492: {
   HEAP32[$34>>2] = 131;
   $2794 = 131;
   break;
  }
  case 508: {
   HEAP32[$34>>2] = 132;
   $2794 = 132;
   break;
  }
  case 513: {
   HEAP32[$34>>2] = 133;
   $2794 = 133;
   break;
  }
  case 527: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 537: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 550: {
   // unreachable;
   break;
  }
  case 556: {
   // unreachable;
   break;
  }
  case 566: {
   // unreachable;
   break;
  }
  case 572: {
   // unreachable;
   break;
  }
  case 598: {
   HEAP32[$34>>2] = 130;
   $2794 = 130;
   break;
  }
  case 617: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 626: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 632: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 638: {
   HEAP32[$34>>2] = 130;
   $2794 = 130;
   break;
  }
  case 673: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 698: {
   HEAP32[$34>>2] = 141;
   $2794 = 141;
   break;
  }
  case 701: {
   HEAP32[$34>>2] = 141;
   $2794 = 141;
   break;
  }
  case 710: {
   HEAP32[$34>>2] = 131;
   $2794 = 131;
   break;
  }
  case 712: {
   HEAP32[$34>>2] = 137;
   $2794 = 137;
   break;
  }
  case 722: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 725: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 731: {
   HEAP32[$34>>2] = 130;
   $2794 = 130;
   break;
  }
  case 796: {
   HEAP32[$34>>2] = 130;
   $2794 = 130;
   break;
  }
  case 804: {
   HEAP32[$34>>2] = 130;
   $2794 = 130;
   break;
  }
  case 810: {
   HEAP32[$34>>2] = 130;
   $2794 = 130;
   break;
  }
  case 827: {
   HEAP32[$34>>2] = 134;
   $2794 = 134;
   break;
  }
  case 829: {
   HEAP32[$34>>2] = 132;
   $2794 = 132;
   break;
  }
  case 831: {
   HEAP32[$34>>2] = 138;
   $2794 = 138;
   break;
  }
  case 834: {
   HEAP32[$34>>2] = 131;
   $2794 = 131;
   break;
  }
  case 836: {
   HEAP32[$34>>2] = 128;
   $2794 = 128;
   break;
  }
  case 841: {
   $2781 = HEAP32[$64>>2]|0;
   $2782 = ($2781|0)>(0);
   if ($2782) {
    HEAP32[$34>>2] = 131;
    $2794 = 131;
    break;
   } else {
    $$01 = 0;
    STACKTOP = sp;return ($$01|0);
   }
   break;
  }
  case 844: {
   HEAP32[$34>>2] = 131;
   $2794 = 131;
   break;
  }
  case 845: {
   $2785 = ($$pr$pr$lcssa|0)==(0);
   if ($2785) {
    $$01 = 0;
    STACKTOP = sp;return ($$01|0);
   } else {
    $2794 = $$pr$pr$lcssa;
   }
   break;
  }
  case 849: {
   STACKTOP = sp;return ($$01|0);
   break;
  }
 }
 $2786 = HEAP8[$68>>0]|0;
 $2787 = ($2786<<24>>24)==(0);
 if (!($2787)) {
  $$01 = $2794;
  STACKTOP = sp;return ($$01|0);
 }
 $2788 = HEAP32[$67>>2]|0;
 $2789 = ($2788|0)==(3);
 if (!($2789)) {
  $$01 = $2794;
  STACKTOP = sp;return ($$01|0);
 }
 $2790 = ((($exc)) + 4|0);
 $2791 = HEAP32[$2790>>2]|0;
 $2792 = ((($2791)) + 296|0);
 HEAP32[$2792>>2] = -1;
 $2793 = ((($2791)) + 300|0);
 HEAP32[$2793>>2] = -1;
 $$01 = $2794;
 STACKTOP = sp;return ($$01|0);
}
function _Current_Ppem($exc) {
 $exc = $exc|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($exc)) + 252|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = $1&65535;
 return ($2|0);
}
function _Current_Ppem_Stretched($exc) {
 $exc = $exc|0;
 var $$$i = 0, $$$i3 = 0, $$01$i = 0, $$01$i2 = 0, $$a$i = 0, $$a$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($exc)) + 252|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = $1&65535;
 $3 = ((($exc)) + 244|0);
 $4 = ((($exc)) + 256|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0);
 if (!($6)) {
  $62 = $5;
  $63 = (_FT_MulFix($2,$62)|0);
  return ($63|0);
 }
 $7 = ((($exc)) + 296|0);
 $8 = HEAP16[$7>>1]|0;
 $9 = ($8<<16>>16)==(0);
 if ($9) {
  $10 = HEAP32[$3>>2]|0;
  HEAP32[$4>>2] = $10;
  $62 = $10;
  $63 = (_FT_MulFix($2,$62)|0);
  return ($63|0);
 }
 $11 = ((($exc)) + 294|0);
 $12 = HEAP16[$11>>1]|0;
 $13 = ($12<<16>>16)==(0);
 $14 = ((($exc)) + 248|0);
 $15 = HEAP32[$14>>2]|0;
 if ($13) {
  HEAP32[$4>>2] = $15;
  $62 = $15;
  $63 = (_FT_MulFix($2,$62)|0);
  return ($63|0);
 } else {
  $16 = $12 << 16 >> 16;
  $17 = HEAP32[$3>>2]|0;
  $18 = $17 ^ $16;
  $19 = ($17|0)<(0);
  $20 = (0 - ($17))|0;
  $$a$i = $19 ? $20 : $17;
  $21 = ($12<<16>>16)<(0);
  $22 = (0 - ($16))|0;
  $$01$i = $21 ? $22 : $16;
  $23 = $$a$i >>> 16;
  $24 = $$a$i & 65535;
  $25 = Math_imul($24, $$01$i)|0;
  $26 = Math_imul($23, $$01$i)|0;
  $27 = $26 >>> 16;
  $28 = $26 << 16;
  $29 = $28 | 8192;
  $30 = (($29) + ($25))|0;
  $31 = ($30>>>0)<($29>>>0);
  $32 = $31&1;
  $$$i = (($32) + ($27))|0;
  $33 = $30 >>> 14;
  $34 = $$$i << 18;
  $35 = $34 | $33;
  $36 = ($18|0)>(-1);
  $37 = (0 - ($35))|0;
  $38 = $36 ? $35 : $37;
  $39 = $8 << 16 >> 16;
  $40 = $15 ^ $39;
  $41 = ($15|0)<(0);
  $42 = (0 - ($15))|0;
  $$a$i1 = $41 ? $42 : $15;
  $43 = ($8<<16>>16)<(0);
  $44 = (0 - ($39))|0;
  $$01$i2 = $43 ? $44 : $39;
  $45 = $$a$i1 >>> 16;
  $46 = $$a$i1 & 65535;
  $47 = Math_imul($46, $$01$i2)|0;
  $48 = Math_imul($45, $$01$i2)|0;
  $49 = $48 >>> 16;
  $50 = $48 << 16;
  $51 = $50 | 8192;
  $52 = (($51) + ($47))|0;
  $53 = ($52>>>0)<($51>>>0);
  $54 = $53&1;
  $$$i3 = (($54) + ($49))|0;
  $55 = $52 >>> 14;
  $56 = $$$i3 << 18;
  $57 = $56 | $55;
  $58 = ($40|0)>(-1);
  $59 = (0 - ($57))|0;
  $60 = $58 ? $57 : $59;
  $61 = (_FT_Hypot($38,$60)|0);
  HEAP32[$4>>2] = $61;
  $62 = $61;
  $63 = (_FT_MulFix($2,$62)|0);
  return ($63|0);
 }
 return (0)|0;
}
function _Read_CVT($exc,$idx) {
 $exc = $exc|0;
 $idx = $idx|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($exc)) + 384|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + ($idx<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 return ($3|0);
}
function _Read_CVT_Stretched($exc,$idx) {
 $exc = $exc|0;
 $idx = $idx|0;
 var $$$i = 0, $$$i3 = 0, $$01$i = 0, $$01$i2 = 0, $$a$i = 0, $$a$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($exc)) + 384|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + ($idx<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($exc)) + 244|0);
 $5 = ((($exc)) + 256|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $63 = $6;
  $64 = (_FT_MulFix($3,$63)|0);
  return ($64|0);
 }
 $8 = ((($exc)) + 296|0);
 $9 = HEAP16[$8>>1]|0;
 $10 = ($9<<16>>16)==(0);
 if ($10) {
  $11 = HEAP32[$4>>2]|0;
  HEAP32[$5>>2] = $11;
  $63 = $11;
  $64 = (_FT_MulFix($3,$63)|0);
  return ($64|0);
 }
 $12 = ((($exc)) + 294|0);
 $13 = HEAP16[$12>>1]|0;
 $14 = ($13<<16>>16)==(0);
 $15 = ((($exc)) + 248|0);
 $16 = HEAP32[$15>>2]|0;
 if ($14) {
  HEAP32[$5>>2] = $16;
  $63 = $16;
  $64 = (_FT_MulFix($3,$63)|0);
  return ($64|0);
 } else {
  $17 = $13 << 16 >> 16;
  $18 = HEAP32[$4>>2]|0;
  $19 = $18 ^ $17;
  $20 = ($18|0)<(0);
  $21 = (0 - ($18))|0;
  $$a$i = $20 ? $21 : $18;
  $22 = ($13<<16>>16)<(0);
  $23 = (0 - ($17))|0;
  $$01$i = $22 ? $23 : $17;
  $24 = $$a$i >>> 16;
  $25 = $$a$i & 65535;
  $26 = Math_imul($25, $$01$i)|0;
  $27 = Math_imul($24, $$01$i)|0;
  $28 = $27 >>> 16;
  $29 = $27 << 16;
  $30 = $29 | 8192;
  $31 = (($30) + ($26))|0;
  $32 = ($31>>>0)<($30>>>0);
  $33 = $32&1;
  $$$i = (($33) + ($28))|0;
  $34 = $31 >>> 14;
  $35 = $$$i << 18;
  $36 = $35 | $34;
  $37 = ($19|0)>(-1);
  $38 = (0 - ($36))|0;
  $39 = $37 ? $36 : $38;
  $40 = $9 << 16 >> 16;
  $41 = $16 ^ $40;
  $42 = ($16|0)<(0);
  $43 = (0 - ($16))|0;
  $$a$i1 = $42 ? $43 : $16;
  $44 = ($9<<16>>16)<(0);
  $45 = (0 - ($40))|0;
  $$01$i2 = $44 ? $45 : $40;
  $46 = $$a$i1 >>> 16;
  $47 = $$a$i1 & 65535;
  $48 = Math_imul($47, $$01$i2)|0;
  $49 = Math_imul($46, $$01$i2)|0;
  $50 = $49 >>> 16;
  $51 = $49 << 16;
  $52 = $51 | 8192;
  $53 = (($52) + ($48))|0;
  $54 = ($53>>>0)<($52>>>0);
  $55 = $54&1;
  $$$i3 = (($55) + ($50))|0;
  $56 = $53 >>> 14;
  $57 = $$$i3 << 18;
  $58 = $57 | $56;
  $59 = ($41|0)>(-1);
  $60 = (0 - ($58))|0;
  $61 = $59 ? $58 : $60;
  $62 = (_FT_Hypot($39,$61)|0);
  HEAP32[$5>>2] = $62;
  $63 = $62;
  $64 = (_FT_MulFix($3,$63)|0);
  return ($64|0);
 }
 return (0)|0;
}
function _Write_CVT($exc,$idx,$value) {
 $exc = $exc|0;
 $idx = $idx|0;
 $value = $value|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($exc)) + 384|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + ($idx<<2)|0);
 HEAP32[$2>>2] = $value;
 return;
}
function _Write_CVT_Stretched($exc,$idx,$value) {
 $exc = $exc|0;
 $idx = $idx|0;
 $value = $value|0;
 var $$$i = 0, $$$i3 = 0, $$01$i = 0, $$01$i2 = 0, $$a$i = 0, $$a$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($exc)) + 244|0);
 $1 = ((($exc)) + 256|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0);
 do {
  if ($3) {
   $4 = ((($exc)) + 296|0);
   $5 = HEAP16[$4>>1]|0;
   $6 = ($5<<16>>16)==(0);
   if ($6) {
    $7 = HEAP32[$0>>2]|0;
    HEAP32[$1>>2] = $7;
    $59 = $7;
    break;
   }
   $8 = ((($exc)) + 294|0);
   $9 = HEAP16[$8>>1]|0;
   $10 = ($9<<16>>16)==(0);
   $11 = ((($exc)) + 248|0);
   $12 = HEAP32[$11>>2]|0;
   if ($10) {
    HEAP32[$1>>2] = $12;
    $59 = $12;
    break;
   } else {
    $13 = $9 << 16 >> 16;
    $14 = HEAP32[$0>>2]|0;
    $15 = $14 ^ $13;
    $16 = ($14|0)<(0);
    $17 = (0 - ($14))|0;
    $$a$i = $16 ? $17 : $14;
    $18 = ($9<<16>>16)<(0);
    $19 = (0 - ($13))|0;
    $$01$i = $18 ? $19 : $13;
    $20 = $$a$i >>> 16;
    $21 = $$a$i & 65535;
    $22 = Math_imul($21, $$01$i)|0;
    $23 = Math_imul($20, $$01$i)|0;
    $24 = $23 >>> 16;
    $25 = $23 << 16;
    $26 = $25 | 8192;
    $27 = (($26) + ($22))|0;
    $28 = ($27>>>0)<($26>>>0);
    $29 = $28&1;
    $$$i = (($29) + ($24))|0;
    $30 = $27 >>> 14;
    $31 = $$$i << 18;
    $32 = $31 | $30;
    $33 = ($15|0)>(-1);
    $34 = (0 - ($32))|0;
    $35 = $33 ? $32 : $34;
    $36 = $5 << 16 >> 16;
    $37 = $12 ^ $36;
    $38 = ($12|0)<(0);
    $39 = (0 - ($12))|0;
    $$a$i1 = $38 ? $39 : $12;
    $40 = ($5<<16>>16)<(0);
    $41 = (0 - ($36))|0;
    $$01$i2 = $40 ? $41 : $36;
    $42 = $$a$i1 >>> 16;
    $43 = $$a$i1 & 65535;
    $44 = Math_imul($43, $$01$i2)|0;
    $45 = Math_imul($42, $$01$i2)|0;
    $46 = $45 >>> 16;
    $47 = $45 << 16;
    $48 = $47 | 8192;
    $49 = (($48) + ($44))|0;
    $50 = ($49>>>0)<($48>>>0);
    $51 = $50&1;
    $$$i3 = (($51) + ($46))|0;
    $52 = $49 >>> 14;
    $53 = $$$i3 << 18;
    $54 = $53 | $52;
    $55 = ($37|0)>(-1);
    $56 = (0 - ($54))|0;
    $57 = $55 ? $54 : $56;
    $58 = (_FT_Hypot($35,$57)|0);
    HEAP32[$1>>2] = $58;
    $59 = $58;
    break;
   }
  } else {
   $59 = $2;
  }
 } while(0);
 $60 = (_FT_DivFix($value,$59)|0);
 $61 = ((($exc)) + 384|0);
 $62 = HEAP32[$61>>2]|0;
 $63 = (($62) + ($idx<<2)|0);
 HEAP32[$63>>2] = $60;
 return;
}
function _Move_CVT($exc,$idx,$value) {
 $exc = $exc|0;
 $idx = $idx|0;
 $value = $value|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($exc)) + 384|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + ($idx<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + ($value))|0;
 HEAP32[$2>>2] = $4;
 return;
}
function _Move_CVT_Stretched($exc,$idx,$value) {
 $exc = $exc|0;
 $idx = $idx|0;
 $value = $value|0;
 var $$$i = 0, $$$i3 = 0, $$01$i = 0, $$01$i2 = 0, $$a$i = 0, $$a$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($exc)) + 244|0);
 $1 = ((($exc)) + 256|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0);
 do {
  if ($3) {
   $4 = ((($exc)) + 296|0);
   $5 = HEAP16[$4>>1]|0;
   $6 = ($5<<16>>16)==(0);
   if ($6) {
    $7 = HEAP32[$0>>2]|0;
    HEAP32[$1>>2] = $7;
    $59 = $7;
    break;
   }
   $8 = ((($exc)) + 294|0);
   $9 = HEAP16[$8>>1]|0;
   $10 = ($9<<16>>16)==(0);
   $11 = ((($exc)) + 248|0);
   $12 = HEAP32[$11>>2]|0;
   if ($10) {
    HEAP32[$1>>2] = $12;
    $59 = $12;
    break;
   } else {
    $13 = $9 << 16 >> 16;
    $14 = HEAP32[$0>>2]|0;
    $15 = $14 ^ $13;
    $16 = ($14|0)<(0);
    $17 = (0 - ($14))|0;
    $$a$i = $16 ? $17 : $14;
    $18 = ($9<<16>>16)<(0);
    $19 = (0 - ($13))|0;
    $$01$i = $18 ? $19 : $13;
    $20 = $$a$i >>> 16;
    $21 = $$a$i & 65535;
    $22 = Math_imul($21, $$01$i)|0;
    $23 = Math_imul($20, $$01$i)|0;
    $24 = $23 >>> 16;
    $25 = $23 << 16;
    $26 = $25 | 8192;
    $27 = (($26) + ($22))|0;
    $28 = ($27>>>0)<($26>>>0);
    $29 = $28&1;
    $$$i = (($29) + ($24))|0;
    $30 = $27 >>> 14;
    $31 = $$$i << 18;
    $32 = $31 | $30;
    $33 = ($15|0)>(-1);
    $34 = (0 - ($32))|0;
    $35 = $33 ? $32 : $34;
    $36 = $5 << 16 >> 16;
    $37 = $12 ^ $36;
    $38 = ($12|0)<(0);
    $39 = (0 - ($12))|0;
    $$a$i1 = $38 ? $39 : $12;
    $40 = ($5<<16>>16)<(0);
    $41 = (0 - ($36))|0;
    $$01$i2 = $40 ? $41 : $36;
    $42 = $$a$i1 >>> 16;
    $43 = $$a$i1 & 65535;
    $44 = Math_imul($43, $$01$i2)|0;
    $45 = Math_imul($42, $$01$i2)|0;
    $46 = $45 >>> 16;
    $47 = $45 << 16;
    $48 = $47 | 8192;
    $49 = (($48) + ($44))|0;
    $50 = ($49>>>0)<($48>>>0);
    $51 = $50&1;
    $$$i3 = (($51) + ($46))|0;
    $52 = $49 >>> 14;
    $53 = $$$i3 << 18;
    $54 = $53 | $52;
    $55 = ($37|0)>(-1);
    $56 = (0 - ($54))|0;
    $57 = $55 ? $54 : $56;
    $58 = (_FT_Hypot($35,$57)|0);
    HEAP32[$1>>2] = $58;
    $59 = $58;
    break;
   }
  } else {
   $59 = $2;
  }
 } while(0);
 $60 = (_FT_DivFix($value,$59)|0);
 $61 = ((($exc)) + 384|0);
 $62 = HEAP32[$61>>2]|0;
 $63 = (($62) + ($idx<<2)|0);
 $64 = HEAP32[$63>>2]|0;
 $65 = (($64) + ($60))|0;
 HEAP32[$63>>2] = $65;
 return;
}
function _Compute_Funcs($exc) {
 $exc = $exc|0;
 var $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($exc)) + 298|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = $1 << 16 >> 16;
 $3 = ($1<<16>>16)==(16384);
 do {
  if ($3) {
   $4 = ((($exc)) + 294|0);
   $5 = HEAP16[$4>>1]|0;
   $6 = $5 << 16 >> 16;
   $7 = ((($exc)) + 564|0);
   HEAP32[$7>>2] = $6;
   $26 = $5;$44 = $6;
  } else {
   $8 = ((($exc)) + 300|0);
   $9 = HEAP16[$8>>1]|0;
   $10 = ($9<<16>>16)==(16384);
   $11 = ((($exc)) + 296|0);
   $12 = HEAP16[$11>>1]|0;
   if ($10) {
    $13 = $12 << 16 >> 16;
    $14 = ((($exc)) + 564|0);
    HEAP32[$14>>2] = $13;
    $$phi$trans$insert = ((($exc)) + 294|0);
    $$pre = HEAP16[$$phi$trans$insert>>1]|0;
    $26 = $$pre;$44 = $13;
    break;
   } else {
    $15 = $9 << 16 >> 16;
    $16 = ((($exc)) + 294|0);
    $17 = HEAP16[$16>>1]|0;
    $18 = $17 << 16 >> 16;
    $19 = Math_imul($18, $2)|0;
    $20 = $12 << 16 >> 16;
    $21 = Math_imul($20, $15)|0;
    $22 = (($21) + ($19))|0;
    $23 = $22 >> 14;
    $24 = ((($exc)) + 564|0);
    HEAP32[$24>>2] = $23;
    $26 = $17;$44 = $23;
    break;
   }
  }
 } while(0);
 $25 = ($26<<16>>16)==(16384);
 do {
  if ($25) {
   $27 = ((($exc)) + 572|0);
   HEAP32[$27>>2] = 4426;
  } else {
   $28 = ((($exc)) + 296|0);
   $29 = HEAP16[$28>>1]|0;
   $30 = ($29<<16>>16)==(16384);
   $31 = ((($exc)) + 572|0);
   if ($30) {
    HEAP32[$31>>2] = 4427;
    break;
   } else {
    HEAP32[$31>>2] = 4428;
    break;
   }
  }
 } while(0);
 $32 = ((($exc)) + 290|0);
 $33 = HEAP16[$32>>1]|0;
 $34 = ($33<<16>>16)==(16384);
 do {
  if ($34) {
   $35 = ((($exc)) + 576|0);
   HEAP32[$35>>2] = 4426;
  } else {
   $36 = ((($exc)) + 292|0);
   $37 = HEAP16[$36>>1]|0;
   $38 = ($37<<16>>16)==(16384);
   $39 = ((($exc)) + 576|0);
   if ($38) {
    HEAP32[$39>>2] = 4427;
    break;
   } else {
    HEAP32[$39>>2] = 4429;
    break;
   }
  }
 } while(0);
 $40 = ((($exc)) + 584|0);
 HEAP32[$40>>2] = 4430;
 $41 = ((($exc)) + 588|0);
 HEAP32[$41>>2] = 4431;
 $42 = ((($exc)) + 564|0);
 $43 = ($44|0)==(16384);
 do {
  if ($43) {
   if ($3) {
    HEAP32[$40>>2] = 4432;
    HEAP32[$41>>2] = 4433;
    break;
   }
   $45 = ((($exc)) + 300|0);
   $46 = HEAP16[$45>>1]|0;
   $47 = ($46<<16>>16)==(16384);
   if ($47) {
    HEAP32[$40>>2] = 4434;
    HEAP32[$41>>2] = 4435;
   }
  }
 } while(0);
 $48 = ($44|0)<(0);
 $49 = (0 - ($44))|0;
 $50 = $48 ? $49 : $44;
 $51 = ($50|0)<(1024);
 if (!($51)) {
  $52 = ((($exc)) + 256|0);
  HEAP32[$52>>2] = 0;
  return;
 }
 HEAP32[$42>>2] = 16384;
 $52 = ((($exc)) + 256|0);
 HEAP32[$52>>2] = 0;
 return;
}
function _Project_x($exc,$dx,$dy) {
 $exc = $exc|0;
 $dx = $dx|0;
 $dy = $dy|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ($dx|0);
}
function _Project_y($exc,$dx,$dy) {
 $exc = $exc|0;
 $dx = $dx|0;
 $dy = $dy|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ($dy|0);
}
function _Project($exc,$dx,$dy) {
 $exc = $exc|0;
 $dx = $dx|0;
 $dy = $dy|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($exc)) + 294|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = $1 << 16 >> 16;
 $3 = ((($exc)) + 296|0);
 $4 = HEAP16[$3>>1]|0;
 $5 = $4 << 16 >> 16;
 $6 = $dx & 65535;
 $7 = Math_imul($2, $6)|0;
 $8 = $dx >> 16;
 $9 = Math_imul($2, $8)|0;
 $10 = $9 << 16;
 $11 = (($10) + ($7))|0;
 $12 = $9 >> 16;
 $13 = $7 >> 31;
 $14 = ($11>>>0)<($7>>>0);
 $15 = $14&1;
 $16 = $dy & 65535;
 $17 = Math_imul($5, $16)|0;
 $18 = $dy >> 16;
 $19 = Math_imul($5, $18)|0;
 $20 = $19 << 16;
 $21 = (($20) + ($17))|0;
 $22 = $19 >> 16;
 $23 = $17 >> 31;
 $24 = ($21>>>0)<($17>>>0);
 $25 = $24&1;
 $26 = (($21) + ($11))|0;
 $27 = ($26>>>0)<($11>>>0);
 $28 = $27&1;
 $29 = (($12) + ($13))|0;
 $30 = (($29) + ($23))|0;
 $31 = (($30) + ($22))|0;
 $32 = (($31) + ($15))|0;
 $33 = (($32) + ($25))|0;
 $34 = (($33) + ($28))|0;
 $35 = $34 >> 31;
 $36 = (($35) + ($26))|0;
 $37 = ($36>>>0)<($26>>>0);
 $38 = $37&1;
 $39 = (($36) + 8192)|0;
 $40 = ($36>>>0)>(4294959103);
 $41 = $40&1;
 $42 = (($35) + ($34))|0;
 $43 = (($42) + ($38))|0;
 $44 = (($43) + ($41))|0;
 $45 = $44 << 18;
 $46 = $39 >>> 14;
 $47 = $45 | $46;
 return ($47|0);
}
function _Dual_Project($exc,$dx,$dy) {
 $exc = $exc|0;
 $dx = $dx|0;
 $dy = $dy|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($exc)) + 290|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = $1 << 16 >> 16;
 $3 = ((($exc)) + 292|0);
 $4 = HEAP16[$3>>1]|0;
 $5 = $4 << 16 >> 16;
 $6 = $dx & 65535;
 $7 = Math_imul($2, $6)|0;
 $8 = $dx >> 16;
 $9 = Math_imul($2, $8)|0;
 $10 = $9 << 16;
 $11 = (($10) + ($7))|0;
 $12 = $9 >> 16;
 $13 = $7 >> 31;
 $14 = ($11>>>0)<($7>>>0);
 $15 = $14&1;
 $16 = $dy & 65535;
 $17 = Math_imul($5, $16)|0;
 $18 = $dy >> 16;
 $19 = Math_imul($5, $18)|0;
 $20 = $19 << 16;
 $21 = (($20) + ($17))|0;
 $22 = $19 >> 16;
 $23 = $17 >> 31;
 $24 = ($21>>>0)<($17>>>0);
 $25 = $24&1;
 $26 = (($21) + ($11))|0;
 $27 = ($26>>>0)<($11>>>0);
 $28 = $27&1;
 $29 = (($12) + ($13))|0;
 $30 = (($29) + ($23))|0;
 $31 = (($30) + ($22))|0;
 $32 = (($31) + ($15))|0;
 $33 = (($32) + ($25))|0;
 $34 = (($33) + ($28))|0;
 $35 = $34 >> 31;
 $36 = (($35) + ($26))|0;
 $37 = ($36>>>0)<($26>>>0);
 $38 = $37&1;
 $39 = (($36) + 8192)|0;
 $40 = ($36>>>0)>(4294959103);
 $41 = $40&1;
 $42 = (($35) + ($34))|0;
 $43 = (($42) + ($38))|0;
 $44 = (($43) + ($41))|0;
 $45 = $44 << 18;
 $46 = $39 >>> 14;
 $47 = $45 | $46;
 return ($47|0);
}
function _Direct_Move($exc,$zone,$point,$distance) {
 $exc = $exc|0;
 $zone = $zone|0;
 $point = $point|0;
 $distance = $distance|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($exc)) + 298|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = ($1<<16>>16)==(0);
 if (!($2)) {
  $3 = $1 << 16 >> 16;
  $4 = ((($exc)) + 564|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = (_FT_MulDiv($distance,$3,$5)|0);
  $7 = $point&65535;
  $8 = ((($zone)) + 16|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = (($9) + ($7<<3)|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + ($6))|0;
  HEAP32[$10>>2] = $12;
  $13 = ((($zone)) + 24|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = (($14) + ($7)|0);
  $16 = HEAP8[$15>>0]|0;
  $17 = $16&255;
  $18 = $17 | 8;
  $19 = $18&255;
  HEAP8[$15>>0] = $19;
 }
 $20 = ((($exc)) + 300|0);
 $21 = HEAP16[$20>>1]|0;
 $22 = ($21<<16>>16)==(0);
 if ($22) {
  return;
 }
 $23 = $21 << 16 >> 16;
 $24 = ((($exc)) + 564|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = (_FT_MulDiv($distance,$23,$25)|0);
 $27 = $point&65535;
 $28 = ((($zone)) + 16|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = (((($29) + ($27<<3)|0)) + 4|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = (($31) + ($26))|0;
 HEAP32[$30>>2] = $32;
 $33 = ((($zone)) + 24|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = (($34) + ($27)|0);
 $36 = HEAP8[$35>>0]|0;
 $37 = $36&255;
 $38 = $37 | 16;
 $39 = $38&255;
 HEAP8[$35>>0] = $39;
 return;
}
function _Direct_Move_Orig($exc,$zone,$point,$distance) {
 $exc = $exc|0;
 $zone = $zone|0;
 $point = $point|0;
 $distance = $distance|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($exc)) + 298|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = ($1<<16>>16)==(0);
 if (!($2)) {
  $3 = $1 << 16 >> 16;
  $4 = ((($exc)) + 564|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = (_FT_MulDiv($distance,$3,$5)|0);
  $7 = $point&65535;
  $8 = ((($zone)) + 12|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = (($9) + ($7<<3)|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + ($6))|0;
  HEAP32[$10>>2] = $12;
 }
 $13 = ((($exc)) + 300|0);
 $14 = HEAP16[$13>>1]|0;
 $15 = ($14<<16>>16)==(0);
 if ($15) {
  return;
 }
 $16 = $14 << 16 >> 16;
 $17 = ((($exc)) + 564|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = (_FT_MulDiv($distance,$16,$18)|0);
 $20 = $point&65535;
 $21 = ((($zone)) + 12|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = (((($22) + ($20<<3)|0)) + 4|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = (($24) + ($19))|0;
 HEAP32[$23>>2] = $25;
 return;
}
function _Direct_Move_X($exc,$zone,$point,$distance) {
 $exc = $exc|0;
 $zone = $zone|0;
 $point = $point|0;
 $distance = $distance|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $point&65535;
 $1 = ((($zone)) + 16|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($2) + ($0<<3)|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + ($distance))|0;
 HEAP32[$3>>2] = $5;
 $6 = ((($zone)) + 24|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (($7) + ($0)|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 $11 = $10 | 8;
 $12 = $11&255;
 HEAP8[$8>>0] = $12;
 return;
}
function _Direct_Move_Orig_X($exc,$zone,$point,$distance) {
 $exc = $exc|0;
 $zone = $zone|0;
 $point = $point|0;
 $distance = $distance|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $point&65535;
 $1 = ((($zone)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($2) + ($0<<3)|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + ($distance))|0;
 HEAP32[$3>>2] = $5;
 return;
}
function _Direct_Move_Y($exc,$zone,$point,$distance) {
 $exc = $exc|0;
 $zone = $zone|0;
 $point = $point|0;
 $distance = $distance|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $point&65535;
 $1 = ((($zone)) + 16|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (((($2) + ($0<<3)|0)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + ($distance))|0;
 HEAP32[$3>>2] = $5;
 $6 = ((($zone)) + 24|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (($7) + ($0)|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 $11 = $10 | 16;
 $12 = $11&255;
 HEAP8[$8>>0] = $12;
 return;
}
function _Direct_Move_Orig_Y($exc,$zone,$point,$distance) {
 $exc = $exc|0;
 $zone = $zone|0;
 $point = $point|0;
 $distance = $distance|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $point&65535;
 $1 = ((($zone)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (((($2) + ($0<<3)|0)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + ($distance))|0;
 HEAP32[$3>>2] = $5;
 return;
}
function _Round_None($exc,$distance,$compensation) {
 $exc = $exc|0;
 $distance = $distance|0;
 $compensation = $compensation|0;
 var $$ = 0, $$1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $val$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($distance|0)>(-1);
 if ($0) {
  $1 = (($compensation) + ($distance))|0;
  $2 = ($1|0)<(0);
  $$ = $2 ? 0 : $1;
  $val$0 = $$;
  return ($val$0|0);
 } else {
  $3 = (($distance) - ($compensation))|0;
  $4 = ($3|0)>(0);
  $$1 = $4 ? 0 : $3;
  $val$0 = $$1;
  return ($val$0|0);
 }
 return (0)|0;
}
function _Round_To_Grid($exc,$distance,$compensation) {
 $exc = $exc|0;
 $distance = $distance|0;
 $compensation = $compensation|0;
 var $$ = 0, $$1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $val$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($distance|0)>(-1);
 if ($0) {
  $1 = (($distance) + 32)|0;
  $2 = (($1) + ($compensation))|0;
  $3 = $2 & -64;
  $4 = ($3|0)<(0);
  $$ = $4 ? 0 : $3;
  $val$0 = $$;
  return ($val$0|0);
 } else {
  $5 = (32 - ($distance))|0;
  $6 = (($5) + ($compensation))|0;
  $7 = $6 & -64;
  $8 = (0 - ($7))|0;
  $9 = ($8|0)>(0);
  $$1 = $9 ? 0 : $8;
  $val$0 = $$1;
  return ($val$0|0);
 }
 return (0)|0;
}
function _Round_Up_To_Grid($exc,$distance,$compensation) {
 $exc = $exc|0;
 $distance = $distance|0;
 $compensation = $compensation|0;
 var $$ = 0, $$1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $val$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($distance|0)>(-1);
 if ($0) {
  $1 = (($distance) + 63)|0;
  $2 = (($1) + ($compensation))|0;
  $3 = $2 & -64;
  $4 = ($3|0)<(0);
  $$ = $4 ? 0 : $3;
  $val$0 = $$;
  return ($val$0|0);
 } else {
  $5 = (63 - ($distance))|0;
  $6 = (($5) + ($compensation))|0;
  $7 = $6 & -64;
  $8 = (0 - ($7))|0;
  $9 = ($8|0)>(0);
  $$1 = $9 ? 0 : $8;
  $val$0 = $$1;
  return ($val$0|0);
 }
 return (0)|0;
}
function _Round_Down_To_Grid($exc,$distance,$compensation) {
 $exc = $exc|0;
 $distance = $distance|0;
 $compensation = $compensation|0;
 var $$ = 0, $$1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $val$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($distance|0)>(-1);
 if ($0) {
  $1 = (($compensation) + ($distance))|0;
  $2 = $1 & -64;
  $3 = ($2|0)<(0);
  $$ = $3 ? 0 : $2;
  $val$0 = $$;
  return ($val$0|0);
 } else {
  $4 = (($compensation) - ($distance))|0;
  $5 = $4 & -64;
  $6 = (0 - ($5))|0;
  $7 = ($6|0)>(0);
  $$1 = $7 ? 0 : $6;
  $val$0 = $$1;
  return ($val$0|0);
 }
 return (0)|0;
}
function _Round_To_Half_Grid($exc,$distance,$compensation) {
 $exc = $exc|0;
 $distance = $distance|0;
 $compensation = $compensation|0;
 var $$ = 0, $$1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $val$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($distance|0)>(-1);
 if ($0) {
  $1 = (($compensation) + ($distance))|0;
  $2 = $1 & -64;
  $3 = $2 | 32;
  $4 = ($3|0)<(0);
  $$ = $4 ? 32 : $3;
  $val$0 = $$;
  return ($val$0|0);
 } else {
  $5 = (($compensation) - ($distance))|0;
  $6 = $5 & -64;
  $7 = $6 | 32;
  $8 = (0 - ($7))|0;
  $9 = ($7|0)<(0);
  $$1 = $9 ? -32 : $8;
  $val$0 = $$1;
  return ($val$0|0);
 }
 return (0)|0;
}
function _Round_To_Double_Grid($exc,$distance,$compensation) {
 $exc = $exc|0;
 $distance = $distance|0;
 $compensation = $compensation|0;
 var $$ = 0, $$1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $val$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($distance|0)>(-1);
 if ($0) {
  $1 = (($distance) + 16)|0;
  $2 = (($1) + ($compensation))|0;
  $3 = $2 & -32;
  $4 = ($3|0)<(0);
  $$ = $4 ? 0 : $3;
  $val$0 = $$;
  return ($val$0|0);
 } else {
  $5 = (16 - ($distance))|0;
  $6 = (($5) + ($compensation))|0;
  $7 = $6 & -32;
  $8 = (0 - ($7))|0;
  $9 = ($8|0)>(0);
  $$1 = $9 ? 0 : $8;
  $val$0 = $$1;
  return ($val$0|0);
 }
 return (0)|0;
}
function _Round_Super($exc,$distance,$compensation) {
 $exc = $exc|0;
 $distance = $distance|0;
 $compensation = $compensation|0;
 var $$ = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $sum = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($distance|0)>(-1);
 if ($0) {
  $1 = ((($exc)) + 480|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ((($exc)) + 484|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = (($compensation) + ($distance))|0;
  $6 = (($5) - ($2))|0;
  $7 = (($6) + ($4))|0;
  $8 = ((($exc)) + 476|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = (0 - ($9))|0;
  $11 = $7 & $10;
  $12 = (($11) + ($2))|0;
  $13 = ($12|0)<(0);
  $$ = $13 ? $2 : $12;
  return ($$|0);
 } else {
  $14 = ((($exc)) + 484|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ((($exc)) + 480|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = (0 - ($17))|0;
  $19 = (($compensation) - ($distance))|0;
  $20 = (($19) + ($15))|0;
  $21 = (($20) - ($17))|0;
  $22 = ((($exc)) + 476|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = (0 - ($23))|0;
  $25 = $21 & $24;
  $sum = (($25) + ($17))|0;
  $26 = (0 - ($sum))|0;
  $27 = ($26|0)>(0);
  $$1 = $27 ? $18 : $26;
  return ($$1|0);
 }
 return (0)|0;
}
function _Round_Super_45($exc,$distance,$compensation) {
 $exc = $exc|0;
 $distance = $distance|0;
 $compensation = $compensation|0;
 var $$ = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($distance|0)>(-1);
 if ($0) {
  $1 = ((($exc)) + 480|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ((($exc)) + 484|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = (($compensation) + ($distance))|0;
  $6 = (($5) - ($2))|0;
  $7 = (($6) + ($4))|0;
  $8 = ((($exc)) + 476|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = (($7|0) % ($9|0))&-1;
  $11 = (($7) - ($10))|0;
  $12 = (($11) + ($2))|0;
  $13 = ($12|0)<(0);
  $$ = $13 ? $2 : $12;
  return ($$|0);
 } else {
  $14 = ((($exc)) + 484|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ((($exc)) + 480|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = (0 - ($17))|0;
  $19 = (($compensation) - ($distance))|0;
  $20 = (($19) + ($15))|0;
  $21 = (($20) - ($17))|0;
  $22 = ((($exc)) + 476|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = (($21|0) % ($23|0))&-1;
  $25 = (($24) - ($21))|0;
  $26 = (($25) - ($17))|0;
  $27 = ($26|0)>(0);
  $$1 = $27 ? $18 : $26;
  return ($$1|0);
 }
 return (0)|0;
}
function __iup_worker_interpolate($worker,$p1,$p2,$ref1,$ref2) {
 $worker = $worker|0;
 $p1 = $p1|0;
 $p2 = $p2|0;
 $ref1 = $ref1|0;
 $ref2 = $ref2|0;
 var $$ = 0, $$2 = 0, $$4 = 0, $$pre = 0, $$pre13 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $i$07 = 0, $i$18 = 0, $or$cond = 0, $or$cond3 = 0, $ref1$ref2 = 0, $ref2$ref1 = 0, $scale$09 = 0, $scale$1 = 0, $scale$2 = 0, $scale_valid$010 = 0, $scale_valid$1 = 0, $scale_valid$2 = 0, $x$0 = 0, $x1$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($p1>>>0)>($p2>>>0);
 if ($0) {
  return;
 }
 $1 = ((($worker)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2>>>0)>($ref1>>>0);
 $4 = ($2>>>0)>($ref2>>>0);
 $or$cond = $3 & $4;
 if (!($or$cond)) {
  return;
 }
 $5 = ((($worker)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + ($ref1<<3)|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (($6) + ($ref2<<3)|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($8|0)>($10|0);
 $$ = $11 ? $8 : $10;
 $$2 = $11 ? $10 : $8;
 $ref1$ref2 = $11 ? $ref1 : $ref2;
 $ref2$ref1 = $11 ? $ref2 : $ref1;
 $12 = HEAP32[$worker>>2]|0;
 $13 = (($12) + ($ref2$ref1<<3)|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12) + ($ref1$ref2<<3)|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($worker)) + 4|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = (($18) + ($ref2$ref1<<3)|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($18) + ($ref1$ref2<<3)|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = (($20) - ($14))|0;
 $24 = (($22) - ($16))|0;
 $25 = ($22|0)==($20|0);
 $26 = ($$|0)==($$2|0);
 $or$cond3 = $26 | $25;
 if ($or$cond3) {
  $i$07 = $p1;
  while(1) {
   $29 = (($12) + ($i$07<<3)|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = ($30|0)>($14|0);
   if ($31) {
    $33 = ($30|0)<($16|0);
    $34 = (($30) + ($24))|0;
    $$4 = $33 ? $20 : $34;
    $x$0 = $$4;
   } else {
    $32 = (($30) + ($23))|0;
    $x$0 = $32;
   }
   $35 = (($18) + ($i$07<<3)|0);
   HEAP32[$35>>2] = $x$0;
   $36 = (($i$07) + 1)|0;
   $37 = ($36>>>0)>($p2>>>0);
   if ($37) {
    break;
   } else {
    $i$07 = $36;
   }
  }
  return;
 }
 $27 = (($22) - ($20))|0;
 $28 = (($$) - ($$2))|0;
 $39 = $12;$57 = $18;$i$18 = $p1;$scale$09 = 0;$scale_valid$010 = 0;
 while(1) {
  $38 = (($39) + ($i$18<<3)|0);
  $40 = HEAP32[$38>>2]|0;
  $41 = ($40|0)>($14|0);
  do {
   if ($41) {
    $43 = ($40|0)<($16|0);
    if (!($43)) {
     $44 = (($40) + ($24))|0;
     $54 = $57;$scale$2 = $scale$09;$scale_valid$2 = $scale_valid$010;$x1$0 = $44;
     break;
    }
    $45 = ($scale_valid$010<<24>>24)==(0);
    if ($45) {
     $46 = (_FT_DivFix($27,$28)|0);
     $scale$1 = $46;$scale_valid$1 = 1;
    } else {
     $scale$1 = $scale$09;$scale_valid$1 = $scale_valid$010;
    }
    $47 = HEAP32[$5>>2]|0;
    $48 = (($47) + ($i$18<<3)|0);
    $49 = HEAP32[$48>>2]|0;
    $50 = (($49) - ($$2))|0;
    $51 = (_FT_MulFix($50,$scale$1)|0);
    $52 = (($51) + ($20))|0;
    $$pre13 = HEAP32[$17>>2]|0;
    $54 = $$pre13;$scale$2 = $scale$1;$scale_valid$2 = $scale_valid$1;$x1$0 = $52;
   } else {
    $42 = (($40) + ($23))|0;
    $54 = $57;$scale$2 = $scale$09;$scale_valid$2 = $scale_valid$010;$x1$0 = $42;
   }
  } while(0);
  $53 = (($54) + ($i$18<<3)|0);
  HEAP32[$53>>2] = $x1$0;
  $55 = (($i$18) + 1)|0;
  $56 = ($55>>>0)>($p2>>>0);
  if ($56) {
   break;
  }
  $$pre = HEAP32[$worker>>2]|0;
  $39 = $$pre;$57 = $54;$i$18 = $55;$scale$09 = $scale$2;$scale_valid$010 = $scale_valid$2;
 }
 return;
}
function _Compute_Point_Displacement($exc,$x,$y,$zone,$refp) {
 $exc = $exc|0;
 $x = $x|0;
 $y = $y|0;
 $zone = $zone|0;
 $refp = $refp|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $8 = 0, $9 = 0, $p$0 = 0, $p$0$in = 0, $zp$sroa$10 = 0, $zp$sroa$5$0 = 0;
 var $zp$sroa$6$0 = 0, $zp$sroa$6$0$in = 0, $zp$sroa$6$0$in$in = 0, $zp$sroa$61$0 = 0, $zp$sroa$8$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $zp$sroa$10 = sp;
 $0 = ((($exc)) + 368|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = $1 & 1;
 $3 = ($2<<24>>24)==(0);
 if ($3) {
  $19 = ((($exc)) + 72|0);
  $20 = $19;
  $21 = $20;
  $22 = HEAP32[$21>>2]|0;
  $23 = (($20) + 4)|0;
  $24 = $23;
  $25 = HEAP32[$24>>2]|0;
  $26 = ((($exc)) + 80|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = ((($exc)) + 84|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = ((($exc)) + 88|0);
  $31 = HEAP32[$30>>2]|0;
  $32 = ((($exc)) + 92|0);
  ;HEAP32[$zp$sroa$10>>2]=HEAP32[$32>>2]|0;HEAP32[$zp$sroa$10+4>>2]=HEAP32[$32+4>>2]|0;HEAP32[$zp$sroa$10+8>>2]=HEAP32[$32+8>>2]|0;HEAP32[$zp$sroa$10+12>>2]=HEAP32[$32+12>>2]|0;
  $33 = ((($exc)) + 288|0);
  $42 = $22;$45 = $25;$p$0$in = $33;$zp$sroa$6$0$in$in = $27;$zp$sroa$61$0 = $29;$zp$sroa$8$0 = $31;
 } else {
  $4 = ((($exc)) + 36|0);
  $5 = $4;
  $6 = $5;
  $7 = HEAP32[$6>>2]|0;
  $8 = (($5) + 4)|0;
  $9 = $8;
  $10 = HEAP32[$9>>2]|0;
  $11 = ((($exc)) + 44|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ((($exc)) + 48|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ((($exc)) + 52|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ((($exc)) + 56|0);
  ;HEAP32[$zp$sroa$10>>2]=HEAP32[$17>>2]|0;HEAP32[$zp$sroa$10+4>>2]=HEAP32[$17+4>>2]|0;HEAP32[$zp$sroa$10+8>>2]=HEAP32[$17+8>>2]|0;HEAP32[$zp$sroa$10+12>>2]=HEAP32[$17+12>>2]|0;
  $18 = ((($exc)) + 286|0);
  $42 = $7;$45 = $10;$p$0$in = $18;$zp$sroa$6$0$in$in = $12;$zp$sroa$61$0 = $14;$zp$sroa$8$0 = $16;
 }
 $zp$sroa$5$0 = $zp$sroa$6$0$in$in&65535;
 $p$0 = HEAP16[$p$0$in>>1]|0;
 $34 = $p$0&65535;
 $35 = ($p$0&65535)<($zp$sroa$5$0&65535);
 if ($35) {
  $zp$sroa$6$0$in = $zp$sroa$6$0$in$in >>> 16;
  $zp$sroa$6$0 = $zp$sroa$6$0$in&65535;
  $40 = $zone;
  $41 = $40;
  HEAP32[$41>>2] = $42;
  $43 = (($40) + 4)|0;
  $44 = $43;
  HEAP32[$44>>2] = $45;
  $46 = ((($zone)) + 8|0);
  HEAP16[$46>>1] = $zp$sroa$5$0;
  $47 = ((($zone)) + 10|0);
  HEAP16[$47>>1] = $zp$sroa$6$0;
  $48 = ((($zone)) + 12|0);
  HEAP32[$48>>2] = $zp$sroa$61$0;
  $49 = ((($zone)) + 16|0);
  HEAP32[$49>>2] = $zp$sroa$8$0;
  $50 = ((($zone)) + 20|0);
  ;HEAP32[$50>>2]=HEAP32[$zp$sroa$10>>2]|0;HEAP32[$50+4>>2]=HEAP32[$zp$sroa$10+4>>2]|0;HEAP32[$50+8>>2]=HEAP32[$zp$sroa$10+8>>2]|0;HEAP32[$50+12>>2]=HEAP32[$zp$sroa$10+12>>2]|0;
  HEAP16[$refp>>1] = $p$0;
  $51 = ((($exc)) + 572|0);
  $52 = HEAP32[$51>>2]|0;
  $53 = (($zp$sroa$8$0) + ($34<<3)|0);
  $54 = HEAP32[$53>>2]|0;
  $55 = (($zp$sroa$61$0) + ($34<<3)|0);
  $56 = HEAP32[$55>>2]|0;
  $57 = (($54) - ($56))|0;
  $58 = (((($zp$sroa$8$0) + ($34<<3)|0)) + 4|0);
  $59 = HEAP32[$58>>2]|0;
  $60 = (((($zp$sroa$61$0) + ($34<<3)|0)) + 4|0);
  $61 = HEAP32[$60>>2]|0;
  $62 = (($59) - ($61))|0;
  $63 = (FUNCTION_TABLE_iiii[$52 & 8191]($exc,$57,$62)|0);
  $64 = ((($exc)) + 298|0);
  $65 = HEAP16[$64>>1]|0;
  $66 = $65 << 16 >> 16;
  $67 = ((($exc)) + 564|0);
  $68 = HEAP32[$67>>2]|0;
  $69 = (_FT_MulDiv($63,$66,$68)|0);
  HEAP32[$x>>2] = $69;
  $70 = ((($exc)) + 300|0);
  $71 = HEAP16[$70>>1]|0;
  $72 = $71 << 16 >> 16;
  $73 = HEAP32[$67>>2]|0;
  $74 = (_FT_MulDiv($63,$72,$73)|0);
  HEAP32[$y>>2] = $74;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $36 = ((($exc)) + 561|0);
 $37 = HEAP8[$36>>0]|0;
 $38 = ($37<<24>>24)==(0);
 if (!($38)) {
  $39 = ((($exc)) + 12|0);
  HEAP32[$39>>2] = 134;
 }
 HEAP16[$refp>>1] = 0;
 $$0 = 1;
 STACKTOP = sp;return ($$0|0);
}
function _Ins_DELTAP($exc,$args) {
 $exc = $exc|0;
 $args = $args|0;
 var $$ = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $C$0 = 0, $k$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($exc)) + 592|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (FUNCTION_TABLE_ii[$1 & 8191]($exc)|0);
 $3 = HEAP32[$args>>2]|0;
 $4 = ($3|0)==(0);
 $$pre = ((($exc)) + 28|0);
 if ($4) {
  $51 = HEAP32[$$pre>>2]|0;
  $52 = ((($exc)) + 32|0);
  HEAP32[$52>>2] = $51;
  return;
 }
 $5 = ((($exc)) + 24|0);
 $6 = ((($exc)) + 36|0);
 $7 = ((($exc)) + 44|0);
 $8 = ((($exc)) + 368|0);
 $9 = ((($exc)) + 332|0);
 $10 = ((($exc)) + 584|0);
 $11 = ((($exc)) + 561|0);
 $12 = ((($exc)) + 12|0);
 $k$01 = 1;
 while(1) {
  $13 = HEAP32[$$pre>>2]|0;
  $14 = ($13|0)<(2);
  if ($14) {
   break;
  }
  $17 = (($13) + -2)|0;
  HEAP32[$$pre>>2] = $17;
  $18 = (($13) + -1)|0;
  $19 = HEAP32[$5>>2]|0;
  $20 = (($19) + ($18<<2)|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = $21&65535;
  $23 = (($19) + ($17<<2)|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = $21 & 65535;
  $26 = HEAP16[$7>>1]|0;
  $27 = $26&65535;
  $28 = ($25>>>0)<($27>>>0);
  if ($28) {
   $29 = $24 >>> 4;
   $30 = $29 & 15;
   $31 = HEAP8[$8>>0]|0;
   $32 = $31&255;
   switch ($32|0) {
   case 114:  {
    $34 = $30 | 32;
    $C$0 = $34;
    break;
   }
   case 113:  {
    $33 = $30 | 16;
    $C$0 = $33;
    break;
   }
   default: {
    $C$0 = $30;
   }
   }
   $35 = HEAP32[$9>>2]|0;
   $36 = $35 & 65535;
   $37 = (($36) + ($C$0))|0;
   $38 = ($2|0)==($37|0);
   if ($38) {
    $39 = $35 >>> 16;
    $40 = $24 & 15;
    $41 = (($40) + -8)|0;
    $42 = ($41|0)>(-1);
    $43 = (($40) + -7)|0;
    $$ = $42 ? $43 : $41;
    $44 = (6 - ($39))|0;
    $45 = $$ << $44;
    $46 = HEAP32[$10>>2]|0;
    FUNCTION_TABLE_viiii[$46 & 8191]($exc,$6,$22,$45);
   }
  } else {
   $47 = HEAP8[$11>>0]|0;
   $48 = ($47<<24>>24)==(0);
   if (!($48)) {
    HEAP32[$12>>2] = 134;
   }
  }
  $49 = (($k$01) + 1)|0;
  $50 = ($49>>>0)>($3>>>0);
  if ($50) {
   label = 16;
   break;
  } else {
   $k$01 = $49;
  }
 }
 if ((label|0) == 16) {
  $51 = HEAP32[$$pre>>2]|0;
  $52 = ((($exc)) + 32|0);
  HEAP32[$52>>2] = $51;
  return;
 }
 $15 = HEAP8[$11>>0]|0;
 $16 = ($15<<24>>24)==(0);
 if (!($16)) {
  HEAP32[$12>>2] = 129;
 }
 HEAP32[$$pre>>2] = 0;
 $51 = HEAP32[$$pre>>2]|0;
 $52 = ((($exc)) + 32|0);
 HEAP32[$52>>2] = $51;
 return;
}
function _TT_Load_Context($exec,$face,$size) {
 $exec = $exec|0;
 $face = $face|0;
 $size = $size|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre6 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $error$i = 0, $error$i1 = 0, $tmp$0$ph = 0, $tmp$1$ph = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i1 = sp + 4|0;
 $error$i = sp;
 HEAP32[$exec>>2] = $face;
 $0 = ((($exec)) + 4|0);
 HEAP32[$0>>2] = $size;
 $1 = ($size|0)==(0|0);
 if (!($1)) {
  $2 = ((($size)) + 116|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($exec)) + 396|0);
  HEAP32[$4>>2] = $3;
  $5 = ((($size)) + 120|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ((($exec)) + 400|0);
  HEAP32[$7>>2] = $6;
  $8 = ((($size)) + 128|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($exec)) + 408|0);
  HEAP32[$10>>2] = $9;
  $11 = ((($size)) + 132|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ((($exec)) + 412|0);
  HEAP32[$13>>2] = $12;
  $14 = ((($size)) + 124|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ((($exec)) + 404|0);
  HEAP32[$16>>2] = $15;
  $17 = ((($size)) + 136|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = ((($exec)) + 416|0);
  HEAP32[$19>>2] = $18;
  $20 = ((($exec)) + 244|0);
  $21 = ((($size)) + 72|0);
  dest=$20; src=$21; stop=dest+40|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
  $22 = ((($exec)) + 216|0);
  $23 = ((($size)) + 44|0);
  ;HEAP32[$22>>2]=HEAP32[$23>>2]|0;HEAP32[$22+4>>2]=HEAP32[$23+4>>2]|0;HEAP32[$22+8>>2]=HEAP32[$23+8>>2]|0;HEAP32[$22+12>>2]=HEAP32[$23+12>>2]|0;HEAP32[$22+16>>2]=HEAP32[$23+16>>2]|0;HEAP32[$22+20>>2]=HEAP32[$23+20>>2]|0;HEAP32[$22+24>>2]=HEAP32[$23+24>>2]|0;
  $24 = ((($size)) + 140|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = ((($exec)) + 420|0);
  HEAP32[$26>>2] = $25;
  $27 = ((($size)) + 144|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = ((($exec)) + 424|0);
  HEAP32[$29>>2] = $28;
  $30 = ((($exec)) + 444|0);
  $31 = ((($size)) + 148|0);
  $32 = $31;
  $33 = $32;
  $34 = HEAP32[$33>>2]|0;
  $35 = (($32) + 4)|0;
  $36 = $35;
  $37 = HEAP32[$36>>2]|0;
  $38 = $30;
  $39 = $38;
  HEAP32[$39>>2] = $34;
  $40 = (($38) + 4)|0;
  $41 = $40;
  HEAP32[$41>>2] = $37;
  $42 = ((($exec)) + 452|0);
  $43 = ((($size)) + 156|0);
  $44 = $43;
  $45 = $44;
  $46 = HEAP32[$45>>2]|0;
  $47 = (($44) + 4)|0;
  $48 = $47;
  $49 = HEAP32[$48>>2]|0;
  $50 = $42;
  $51 = $50;
  HEAP32[$51>>2] = $46;
  $52 = (($50) + 4)|0;
  $53 = $52;
  HEAP32[$53>>2] = $49;
  $54 = ((($exec)) + 460|0);
  $55 = ((($size)) + 164|0);
  $56 = $55;
  $57 = $56;
  $58 = HEAP32[$57>>2]|0;
  $59 = (($56) + 4)|0;
  $60 = $59;
  $61 = HEAP32[$60>>2]|0;
  $62 = $54;
  $63 = $62;
  HEAP32[$63>>2] = $58;
  $64 = (($62) + 4)|0;
  $65 = $64;
  HEAP32[$65>>2] = $61;
  $66 = ((($exec)) + 284|0);
  $67 = ((($size)) + 172|0);
  dest=$66; src=$67; stop=dest+68|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
  $68 = ((($size)) + 240|0);
  $69 = HEAP32[$68>>2]|0;
  $70 = ((($exec)) + 380|0);
  HEAP32[$70>>2] = $69;
  $71 = ((($size)) + 244|0);
  $72 = HEAP32[$71>>2]|0;
  $73 = ((($exec)) + 384|0);
  HEAP32[$73>>2] = $72;
  $74 = ((($size)) + 248|0);
  $75 = HEAP16[$74>>1]|0;
  $76 = ((($exec)) + 468|0);
  HEAP16[$76>>1] = $75;
  $77 = ((($size)) + 252|0);
  $78 = HEAP32[$77>>2]|0;
  $79 = ((($exec)) + 472|0);
  HEAP32[$79>>2] = $78;
  $80 = ((($exec)) + 180|0);
  $81 = ((($size)) + 256|0);
  dest=$80; src=$81; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
  $82 = ((($exec)) + 36|0);
  dest=$82; stop=dest+108|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 }
 $83 = ((($exec)) + 20|0);
 $84 = HEAP32[$83>>2]|0;
 $85 = ((($exec)) + 8|0);
 $86 = HEAP32[$85>>2]|0;
 $87 = ((($exec)) + 24|0);
 $88 = ((($face)) + 284|0);
 $89 = HEAP32[$88>>2]|0;
 $90 = $89 & 65535;
 $91 = (($90) + 32)|0;
 $92 = ($84>>>0)<($91>>>0);
 do {
  if ($92) {
   $95 = $84 << 2;
   $96 = $91 << 2;
   $97 = HEAP32[$87>>2]|0;
   $98 = (_ft_mem_realloc($86,1,$95,$96,$97,$error$i)|0);
   HEAP32[$87>>2] = $98;
   $99 = HEAP32[$error$i>>2]|0;
   $100 = ($99|0)==(0);
   if ($100) {
    $$pre = HEAP32[$85>>2]|0;
    $$phi$trans$insert = ((($face)) + 286|0);
    $$pre6 = HEAP16[$$phi$trans$insert>>1]|0;
    $105 = $$pre6;$108 = $$pre;$tmp$0$ph = $91;
    break;
   }
   HEAP32[$83>>2] = $84;
   $$0 = $99;
   STACKTOP = sp;return ($$0|0);
  } else {
   $93 = $89 >>> 16;
   $94 = $93&65535;
   $105 = $94;$108 = $86;$tmp$0$ph = $84;
  }
 } while(0);
 HEAP32[$83>>2] = $tmp$0$ph;
 $101 = ((($exec)) + 388|0);
 $102 = HEAP32[$101>>2]|0;
 $103 = ((($exec)) + 392|0);
 $104 = $105&65535;
 $106 = ($102>>>0)<($104>>>0);
 if ($106) {
  $107 = HEAP32[$103>>2]|0;
  $109 = (_ft_mem_realloc($108,1,$102,$104,$107,$error$i1)|0);
  HEAP32[$103>>2] = $109;
  $110 = HEAP32[$error$i1>>2]|0;
  $111 = ($110|0)==(0);
  if ($111) {
   $tmp$1$ph = $104;
  } else {
   $112 = $102 & 65535;
   HEAP32[$101>>2] = $112;
   $$0 = $110;
   STACKTOP = sp;return ($$0|0);
  }
 } else {
  $tmp$1$ph = $102;
 }
 $113 = $tmp$1$ph & 65535;
 HEAP32[$101>>2] = $113;
 $114 = ((($exec)) + 144|0);
 $115 = ((($exec)) + 152|0);
 HEAP16[$115>>1] = 0;
 $116 = ((($exec)) + 154|0);
 HEAP16[$116>>1] = 0;
 $117 = ((($exec)) + 72|0);
 dest=$117; src=$114; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 $118 = ((($exec)) + 108|0);
 dest=$118; src=$114; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 $119 = ((($exec)) + 36|0);
 dest=$119; src=$114; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 $120 = ((($exec)) + 488|0);
 HEAP8[$120>>0] = 0;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _tt_size_run_prep($size,$pedantic) {
 $size = $size|0;
 $pedantic = $pedantic|0;
 var $$0 = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $error$0 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = HEAP32[$size>>2]|0;
 $1 = ((($size)) + 292|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (_TT_Load_Context($2,$0,$size)|0);
 $4 = ($3|0)==(0);
 if (!($4)) {
  $$0 = $3;
  return ($$0|0);
 }
 $5 = ((($2)) + 428|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($2)) + 16|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($2)) + 488|0);
 HEAP8[$7>>0] = 0;
 $8 = ((($2)) + 561|0);
 HEAP8[$8>>0] = $pedantic;
 $9 = ((($0)) + 640|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($0)) + 636|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($2)) + 452|0);
 HEAP32[$13>>2] = $10;
 $14 = ((($2)) + 456|0);
 HEAP32[$14>>2] = $12;
 $15 = ((($2)) + 460|0);
 HEAP32[$15>>2] = 0;
 $16 = ((($2)) + 464|0);
 HEAP32[$16>>2] = 0;
 $17 = ($12|0)==(0);
 if ($17) {
  $error$0 = 0;
 } else {
  $18 = ((($2)) + 356|0);
  HEAP32[$18>>2] = $10;
  $19 = ((($2)) + 364|0);
  HEAP32[$19>>2] = $12;
  $20 = ((($2)) + 360|0);
  HEAP32[$20>>2] = 0;
  $21 = ((($2)) + 352|0);
  HEAP32[$21>>2] = 2;
  $22 = ((($0)) + 652|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = (FUNCTION_TABLE_ii[$23 & 8191]($2)|0);
  $error$0 = $24;
 }
 $$pre$phiZ2D = ((($2)) + 452|0);
 $25 = ((($size)) + 300|0);
 HEAP32[$25>>2] = $error$0;
 $26 = ((($2)) + 284|0);
 $27 = ((($2)) + 290|0);
 HEAP16[$27>>1] = 16384;
 $28 = ((($2)) + 292|0);
 HEAP16[$28>>1] = 0;
 $29 = ((($2)) + 294|0);
 HEAP16[$29>>1] = 16384;
 $30 = ((($2)) + 296|0);
 HEAP16[$30>>1] = 0;
 $31 = ((($2)) + 298|0);
 HEAP16[$31>>1] = 16384;
 $32 = ((($2)) + 300|0);
 HEAP16[$32>>1] = 0;
 HEAP16[$26>>1] = 0;
 $33 = ((($2)) + 286|0);
 HEAP16[$33>>1] = 0;
 $34 = ((($2)) + 288|0);
 HEAP16[$34>>1] = 0;
 $35 = ((($2)) + 344|0);
 HEAP16[$35>>1] = 1;
 $36 = ((($2)) + 346|0);
 HEAP16[$36>>1] = 1;
 $37 = ((($2)) + 348|0);
 HEAP16[$37>>1] = 1;
 $38 = ((($2)) + 304|0);
 HEAP32[$38>>2] = 1;
 $39 = ((($size)) + 172|0);
 dest=$39; src=$26; stop=dest+68|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 $40 = ((($2)) + 396|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = ((($size)) + 116|0);
 HEAP32[$42>>2] = $41;
 $43 = ((($2)) + 408|0);
 $44 = HEAP32[$43>>2]|0;
 $45 = ((($size)) + 128|0);
 HEAP32[$45>>2] = $44;
 $46 = ((($2)) + 420|0);
 $47 = HEAP32[$46>>2]|0;
 $48 = ((($size)) + 140|0);
 HEAP32[$48>>2] = $47;
 $49 = ((($2)) + 424|0);
 $50 = HEAP32[$49>>2]|0;
 $51 = ((($size)) + 144|0);
 HEAP32[$51>>2] = $50;
 $52 = ((($size)) + 148|0);
 $53 = ((($2)) + 444|0);
 $54 = $53;
 $55 = $54;
 $56 = HEAP32[$55>>2]|0;
 $57 = (($54) + 4)|0;
 $58 = $57;
 $59 = HEAP32[$58>>2]|0;
 $60 = $52;
 $61 = $60;
 HEAP32[$61>>2] = $56;
 $62 = (($60) + 4)|0;
 $63 = $62;
 HEAP32[$63>>2] = $59;
 $64 = ((($size)) + 156|0);
 $65 = $$pre$phiZ2D;
 $66 = $65;
 $67 = HEAP32[$66>>2]|0;
 $68 = (($65) + 4)|0;
 $69 = $68;
 $70 = HEAP32[$69>>2]|0;
 $71 = $64;
 $72 = $71;
 HEAP32[$72>>2] = $67;
 $73 = (($71) + 4)|0;
 $74 = $73;
 HEAP32[$74>>2] = $70;
 $75 = ((($size)) + 164|0);
 $76 = ((($2)) + 460|0);
 $77 = $76;
 $78 = $77;
 $79 = HEAP32[$78>>2]|0;
 $80 = (($77) + 4)|0;
 $81 = $80;
 $82 = HEAP32[$81>>2]|0;
 $83 = $75;
 $84 = $83;
 HEAP32[$84>>2] = $79;
 $85 = (($83) + 4)|0;
 $86 = $85;
 HEAP32[$86>>2] = $82;
 $$0 = $error$0;
 return ($$0|0);
}
function _tt_get_kerning($ttface,$left_glyph,$right_glyph,$kerning) {
 $ttface = $ttface|0;
 $left_glyph = $left_glyph|0;
 $right_glyph = $right_glyph|0;
 $kerning = $kerning|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($ttface)) + 532|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$kerning>>2] = 0;
 $2 = ((($kerning)) + 4|0);
 HEAP32[$2>>2] = 0;
 $3 = ($1|0)==(0|0);
 if ($3) {
  return 0;
 }
 $4 = ((($1)) + 84|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iiii[$5 & 8191]($ttface,$left_glyph,$right_glyph)|0);
 HEAP32[$kerning>>2] = $6;
 return 0;
}
function _tt_get_advances($ttface,$start,$count,$flags,$advances) {
 $ttface = $ttface|0;
 $start = $start|0;
 $count = $count|0;
 $flags = $flags|0;
 $advances = $advances|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ah = 0, $aw = 0, $exitcond = 0, $exitcond7 = 0, $lsb = 0, $nn$04 = 0, $nn$13 = 0, $sext$i = 0, $tsb = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tsb = sp + 6|0;
 $ah = sp + 4|0;
 $lsb = sp + 2|0;
 $aw = sp;
 $0 = $flags & 16;
 $1 = ($0|0)==(0);
 $2 = ($count|0)!=(0);
 if ($1) {
  if (!($2)) {
   STACKTOP = sp;return 0;
  }
  $9 = ((($ttface)) + 532|0);
  $nn$13 = 0;
  while(1) {
   $40 = (($nn$13) + ($start))|0;
   $41 = HEAP32[$9>>2]|0;
   $42 = ((($41)) + 112|0);
   $43 = HEAP32[$42>>2]|0;
   FUNCTION_TABLE_viiiii[$43 & 8191]($ttface,0,$40,$lsb,$aw);
   $44 = HEAP16[$aw>>1]|0;
   $45 = $44&65535;
   $46 = (($advances) + ($nn$13<<2)|0);
   HEAP32[$46>>2] = $45;
   $47 = (($nn$13) + 1)|0;
   $exitcond = ($47|0)==($count|0);
   if ($exitcond) {
    break;
   } else {
    $nn$13 = $47;
   }
  }
  STACKTOP = sp;return 0;
 }
 if (!($2)) {
  STACKTOP = sp;return 0;
 }
 $3 = ((($ttface)) + 292|0);
 $4 = ((($ttface)) + 364|0);
 $5 = ((($ttface)) + 220|0);
 $6 = ((($ttface)) + 434|0);
 $7 = ((($ttface)) + 436|0);
 $8 = ((($ttface)) + 532|0);
 $nn$04 = 0;
 while(1) {
  $10 = HEAP8[$3>>0]|0;
  $11 = ($10<<24>>24)==(0);
  do {
   if ($11) {
    $16 = HEAP16[$4>>1]|0;
    $17 = ($16<<16>>16)==(-1);
    if ($17) {
     $27 = HEAP16[$5>>1]|0;
     HEAP16[$tsb>>1] = $27;
     $28 = HEAP32[$5>>2]|0;
     $sext$i = $28 << 16;
     $29 = $sext$i >> 16;
     $30 = $28 >> 16;
     $31 = (($29) - ($30))|0;
     $32 = ($31|0)<(0);
     $33 = (0 - ($31))|0;
     $34 = $32 ? $33 : $31;
     $35 = $34&65535;
     HEAP16[$ah>>1] = $35;
     $37 = $35;
     break;
    } else {
     $18 = HEAP16[$6>>1]|0;
     HEAP16[$tsb>>1] = $18;
     $19 = $18 << 16 >> 16;
     $20 = HEAP16[$7>>1]|0;
     $21 = $20 << 16 >> 16;
     $22 = (($19) - ($21))|0;
     $23 = ($22|0)<(0);
     $24 = (0 - ($22))|0;
     $25 = $23 ? $24 : $22;
     $26 = $25&65535;
     HEAP16[$ah>>1] = $26;
     $37 = $26;
     break;
    }
   } else {
    $12 = (($nn$04) + ($start))|0;
    $13 = HEAP32[$8>>2]|0;
    $14 = ((($13)) + 112|0);
    $15 = HEAP32[$14>>2]|0;
    FUNCTION_TABLE_viiiii[$15 & 8191]($ttface,1,$12,$tsb,$ah);
    $$pre = HEAP16[$ah>>1]|0;
    $37 = $$pre;
   }
  } while(0);
  $36 = $37&65535;
  $38 = (($advances) + ($nn$04<<2)|0);
  HEAP32[$38>>2] = $36;
  $39 = (($nn$04) + 1)|0;
  $exitcond7 = ($39|0)==($count|0);
  if ($exitcond7) {
   break;
  } else {
   $nn$04 = $39;
  }
 }
 STACKTOP = sp;return 0;
}
function _tt_size_request($size,$req) {
 $size = $size|0;
 $req = $req|0;
 var $$01$ph = 0, $$1 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$0 = 0, $strike_index = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $strike_index = sp;
 $0 = HEAP32[$size>>2]|0;
 $1 = ((($0)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2 & 2;
 $4 = ($3|0)==(0);
 do {
  if ($4) {
   $26 = $0;$error$0 = 0;
  } else {
   $5 = ((($0)) + 532|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ((($6)) + 104|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = (FUNCTION_TABLE_iiii[$8 & 8191]($0,$req,$strike_index)|0);
   $10 = ($9|0)==(0);
   if (!($10)) {
    $25 = ((($size)) + 112|0);
    HEAP32[$25>>2] = -1;
    $$pre = HEAP32[$size>>2]|0;
    $26 = $$pre;$error$0 = $9;
    break;
   }
   $11 = HEAP32[$strike_index>>2]|0;
   $12 = HEAP32[$size>>2]|0;
   $13 = ((($size)) + 112|0);
   HEAP32[$13>>2] = $11;
   $14 = ((($12)) + 8|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = $15 & 1;
   $17 = ($16|0)==(0);
   if ($17) {
    $18 = ((($12)) + 532|0);
    $19 = HEAP32[$18>>2]|0;
    $20 = ((($size)) + 12|0);
    $21 = ((($19)) + 108|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = (FUNCTION_TABLE_iiii[$22 & 8191]($12,$11,$20)|0);
    $24 = ($23|0)==(0);
    if ($24) {
     $$01$ph = 0;
    } else {
     HEAP32[$13>>2] = -1;
     $$01$ph = $23;
    }
   } else {
    _FT_Select_Metrics($12,$11);
    (_tt_size_reset($size)|0);
    $$01$ph = 0;
   }
   $$1 = $$01$ph;
   STACKTOP = sp;return ($$1|0);
  }
 } while(0);
 _FT_Request_Metrics($26,$req);
 $27 = HEAP32[$size>>2]|0;
 $28 = ((($27)) + 8|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = $29 & 1;
 $31 = ($30|0)==(0);
 if ($31) {
  $$1 = $error$0;
  STACKTOP = sp;return ($$1|0);
 }
 $32 = (_tt_size_reset($size)|0);
 $33 = ((($size)) + 12|0);
 $34 = ((($size)) + 44|0);
 ;HEAP32[$33>>2]=HEAP32[$34>>2]|0;HEAP32[$33+4>>2]=HEAP32[$34+4>>2]|0;HEAP32[$33+8>>2]=HEAP32[$34+8>>2]|0;HEAP32[$33+12>>2]=HEAP32[$34+12>>2]|0;HEAP32[$33+16>>2]=HEAP32[$34+16>>2]|0;HEAP32[$33+20>>2]=HEAP32[$34+20>>2]|0;HEAP32[$33+24>>2]=HEAP32[$34+24>>2]|0;
 $$1 = $32;
 STACKTOP = sp;return ($$1|0);
}
function _tt_size_reset($size) {
 $size = $size|0;
 var $$0 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($size)) + 72|0);
 $1 = ((($size)) + 108|0);
 HEAP8[$1>>0] = 0;
 $2 = HEAP32[$size>>2]|0;
 $3 = ((($size)) + 44|0);
 $4 = ((($size)) + 12|0);
 ;HEAP32[$3>>2]=HEAP32[$4>>2]|0;HEAP32[$3+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$3+8>>2]=HEAP32[$4+8>>2]|0;HEAP32[$3+12>>2]=HEAP32[$4+12>>2]|0;HEAP32[$3+16>>2]=HEAP32[$4+16>>2]|0;HEAP32[$3+20>>2]=HEAP32[$4+20>>2]|0;HEAP32[$3+24>>2]=HEAP32[$4+24>>2]|0;
 $5 = HEAP32[$3>>2]|0;
 $6 = $5&65535;
 $7 = ($6<<16>>16)==(0);
 $8 = $5 >>> 16;
 $9 = $8&65535;
 if ($7) {
  $$0 = 151;
  return ($$0|0);
 }
 $10 = ((($size)) + 46|0);
 $11 = ($9<<16>>16)==(0);
 if ($11) {
  $$0 = 151;
  return ($$0|0);
 }
 $12 = ((($2)) + 176|0);
 $13 = HEAP16[$12>>1]|0;
 $14 = $13 & 8;
 $15 = ($14<<16>>16)==(0);
 if ($15) {
  $66 = $6;$67 = $9;$73 = $5;$74 = $8;
 } else {
  $16 = $5 << 6;
  $17 = $16 & 4194240;
  $18 = ((($2)) + 68|0);
  $19 = HEAP16[$18>>1]|0;
  $20 = $19&65535;
  $21 = (_FT_DivFix($17,$20)|0);
  $22 = ((($size)) + 48|0);
  HEAP32[$22>>2] = $21;
  $23 = HEAP16[$10>>1]|0;
  $24 = $23&65535;
  $25 = $24 << 6;
  $26 = HEAP16[$18>>1]|0;
  $27 = $26&65535;
  $28 = (_FT_DivFix($25,$27)|0);
  $29 = ((($size)) + 52|0);
  HEAP32[$29>>2] = $28;
  $30 = ((($2)) + 70|0);
  $31 = HEAP16[$30>>1]|0;
  $32 = $31 << 16 >> 16;
  $33 = (_FT_MulFix($32,$28)|0);
  $34 = (($33) + 32)|0;
  $35 = $34 & -64;
  $36 = ((($size)) + 56|0);
  HEAP32[$36>>2] = $35;
  $37 = ((($2)) + 72|0);
  $38 = HEAP16[$37>>1]|0;
  $39 = $38 << 16 >> 16;
  $40 = HEAP32[$29>>2]|0;
  $41 = (_FT_MulFix($39,$40)|0);
  $42 = (($41) + 32)|0;
  $43 = $42 & -64;
  $44 = ((($size)) + 60|0);
  HEAP32[$44>>2] = $43;
  $45 = ((($2)) + 74|0);
  $46 = HEAP16[$45>>1]|0;
  $47 = $46 << 16 >> 16;
  $48 = HEAP32[$29>>2]|0;
  $49 = (_FT_MulFix($47,$48)|0);
  $50 = (($49) + 32)|0;
  $51 = $50 & -64;
  $52 = ((($size)) + 64|0);
  HEAP32[$52>>2] = $51;
  $53 = ((($2)) + 76|0);
  $54 = HEAP16[$53>>1]|0;
  $55 = $54 << 16 >> 16;
  $56 = HEAP32[$22>>2]|0;
  $57 = (_FT_MulFix($55,$56)|0);
  $58 = (($57) + 32)|0;
  $59 = $58 & -64;
  $60 = ((($size)) + 68|0);
  HEAP32[$60>>2] = $59;
  $61 = HEAP32[$3>>2]|0;
  $62 = $61&65535;
  $63 = $61 >>> 16;
  $64 = $63&65535;
  $66 = $62;$67 = $64;$73 = $61;$74 = $63;
 }
 $65 = ($66&65535)<($67&65535);
 if ($65) {
  $76 = ((($size)) + 52|0);
  $77 = HEAP32[$76>>2]|0;
  $78 = ((($size)) + 88|0);
  HEAP32[$78>>2] = $77;
  $79 = ((($size)) + 80|0);
  HEAP16[$79>>1] = $67;
  $80 = $73 & 65535;
  $81 = $67&65535;
  $82 = (_FT_DivFix($80,$81)|0);
  HEAP32[$0>>2] = $82;
  $$sink = 65536;
 } else {
  $68 = ((($size)) + 48|0);
  $69 = HEAP32[$68>>2]|0;
  $70 = ((($size)) + 88|0);
  HEAP32[$70>>2] = $69;
  $71 = ((($size)) + 80|0);
  HEAP16[$71>>1] = $66;
  HEAP32[$0>>2] = 65536;
  $72 = $73 & 65535;
  $75 = (_FT_DivFix($74,$72)|0);
  $$sink = $75;
 }
 $83 = ((($size)) + 76|0);
 HEAP32[$83>>2] = $$sink;
 $84 = ((($size)) + 300|0);
 HEAP32[$84>>2] = -1;
 HEAP8[$1>>0] = 1;
 $$0 = 0;
 return ($$0|0);
}
function _tt_size_select($size,$strike_index) {
 $size = $size|0;
 $strike_index = $strike_index|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$size>>2]|0;
 $1 = ((($size)) + 112|0);
 HEAP32[$1>>2] = $strike_index;
 $2 = ((($0)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 1;
 $5 = ($4|0)==(0);
 if (!($5)) {
  _FT_Select_Metrics($0,$strike_index);
  (_tt_size_reset($size)|0);
  $error$0 = 0;
  return ($error$0|0);
 }
 $6 = ((($0)) + 532|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($size)) + 12|0);
 $9 = ((($7)) + 108|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (FUNCTION_TABLE_iiii[$10 & 8191]($0,$strike_index,$8)|0);
 $12 = ($11|0)==(0);
 if ($12) {
  $error$0 = 0;
  return ($error$0|0);
 }
 HEAP32[$1>>2] = -1;
 $error$0 = $11;
 return ($error$0|0);
}
function _T1_Driver_Init($driver) {
 $driver = $driver|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _T1_Driver_Done($driver) {
 $driver = $driver|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _Get_Interface($module,$t1_interface) {
 $module = $module|0;
 $t1_interface = $t1_interface|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_ft_service_list_lookup(56348,$t1_interface)|0);
 return ($0|0);
}
function _t1_get_ps_name($face) {
 $face = $face|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 364|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function _t1_get_glyph_name($face,$glyph_index,$buffer,$buffer_max) {
 $face = $face|0;
 $glyph_index = $glyph_index|0;
 $buffer = $buffer|0;
 $buffer_max = $buffer_max|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 420|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + ($glyph_index<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 (_ft_mem_strcpyn($buffer,$3,$buffer_max)|0);
 return 0;
}
function _t1_get_name_index($face,$glyph_name) {
 $face = $face|0;
 $glyph_name = $glyph_name|0;
 var $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 416|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(0);
 if (!($2)) {
  $$2 = 0;
  return ($$2|0);
 }
 $3 = ((($face)) + 420|0);
 $4 = HEAP32[$3>>2]|0;
 $i$02 = 0;
 while(1) {
  $7 = (($4) + ($i$02<<2)|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = (_strcmp($glyph_name,$8)|0);
  $10 = ($9|0)==(0);
  $6 = (($i$02) + 1)|0;
  if ($10) {
   $$2 = $i$02;
   label = 5;
   break;
  }
  $5 = ($6|0)<($1|0);
  if ($5) {
   $i$02 = $6;
  } else {
   $$2 = 0;
   label = 5;
   break;
  }
 }
 if ((label|0) == 5) {
  return ($$2|0);
 }
 return (0)|0;
}
function _t1_ps_get_font_info($face,$afont_info) {
 $face = $face|0;
 $afont_info = $afont_info|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 132|0);
 ;HEAP32[$afont_info>>2]=HEAP32[$0>>2]|0;HEAP32[$afont_info+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$afont_info+8>>2]=HEAP32[$0+8>>2]|0;HEAP32[$afont_info+12>>2]=HEAP32[$0+12>>2]|0;HEAP32[$afont_info+16>>2]=HEAP32[$0+16>>2]|0;HEAP32[$afont_info+20>>2]=HEAP32[$0+20>>2]|0;HEAP32[$afont_info+24>>2]=HEAP32[$0+24>>2]|0;HEAP32[$afont_info+28>>2]=HEAP32[$0+28>>2]|0;
 return 0;
}
function _t1_ps_get_font_extra($face,$afont_extra) {
 $face = $face|0;
 $afont_extra = $afont_extra|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 164|0);
 $1 = HEAP16[$0>>1]|0;
 HEAP16[$afont_extra>>1] = $1;
 return 0;
}
function _t1_ps_has_glyph_names($face) {
 $face = $face|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1;
}
function _t1_ps_get_font_private($face,$afont_private) {
 $face = $face|0;
 $afont_private = $afont_private|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 168|0);
 _memcpy(($afont_private|0),($0|0),196)|0;
 return 0;
}
function _t1_ps_get_font_value($face,$key,$idx,$value,$value_len_) {
 $face = $face|0;
 $key = $key|0;
 $idx = $idx|0;
 $value = $value|0;
 $value_len_ = $value_len_|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, $or$cond1 = 0, $or$cond10 = 0, $or$cond11 = 0, $or$cond12 = 0;
 var $or$cond13 = 0, $or$cond14 = 0, $or$cond15 = 0, $or$cond16 = 0, $or$cond17 = 0, $or$cond18 = 0, $or$cond19 = 0, $or$cond2 = 0, $or$cond20 = 0, $or$cond21 = 0, $or$cond22 = 0, $or$cond23 = 0, $or$cond24 = 0, $or$cond25 = 0, $or$cond26 = 0, $or$cond27 = 0, $or$cond28 = 0, $or$cond29 = 0, $or$cond3 = 0, $or$cond30 = 0;
 var $or$cond31 = 0, $or$cond32 = 0, $or$cond33 = 0, $or$cond34 = 0, $or$cond35 = 0, $or$cond36 = 0, $or$cond37 = 0, $or$cond38 = 0, $or$cond39 = 0, $or$cond4 = 0, $or$cond40 = 0, $or$cond41 = 0, $or$cond42 = 0, $or$cond43 = 0, $or$cond44 = 0, $or$cond45 = 0, $or$cond5 = 0, $or$cond6 = 0, $or$cond7 = 0, $or$cond8 = 0;
 var $or$cond9 = 0, $retval$0 = 0, $val$0 = 0, $val1$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($value_len_|0)<(0);
 $1 = $0 ? 0 : $value_len_;
 $2 = ((($face)) + 132|0);
 do {
  switch ($key|0) {
  case 0:  {
   $3 = ($value|0)==(0|0);
   $4 = ($1|0)==(0);
   $or$cond = $3 | $4;
   if ($or$cond) {
    $retval$0 = 1;
   } else {
    $5 = ((($2)) + 301|0);
    $6 = HEAP8[$5>>0]|0;
    HEAP8[$value>>0] = $6;
    $retval$0 = 1;
   }
   break;
  }
  case 1:  {
   $7 = ($idx>>>0)<(4);
   if ($7) {
    $8 = ($value|0)!=(0|0);
    $9 = ($1>>>0)>(3);
    $or$cond1 = $8 & $9;
    if ($or$cond1) {
     switch ($idx|0) {
     case 0:  {
      $10 = ((($face)) + 436|0);
      $11 = HEAP32[$10>>2]|0;
      $val$0 = $11;
      break;
     }
     case 1:  {
      $12 = ((($face)) + 436|0);
      $13 = ((($12)) + 4|0);
      $14 = HEAP32[$13>>2]|0;
      $val$0 = $14;
      break;
     }
     case 2:  {
      $15 = ((($face)) + 436|0);
      $16 = ((($15)) + 8|0);
      $17 = HEAP32[$16>>2]|0;
      $val$0 = $17;
      break;
     }
     case 3:  {
      $18 = ((($face)) + 436|0);
      $19 = ((($18)) + 12|0);
      $20 = HEAP32[$19>>2]|0;
      $val$0 = $20;
      break;
     }
     default: {
      $val$0 = 0;
     }
     }
     HEAP32[$value>>2] = $val$0;
     $retval$0 = 4;
    } else {
     $retval$0 = 4;
    }
   } else {
    $retval$0 = 0;
   }
   break;
  }
  case 2:  {
   $21 = ($idx>>>0)<(4);
   if ($21) {
    $22 = ($value|0)!=(0|0);
    $23 = ($1>>>0)>(3);
    $or$cond2 = $22 & $23;
    if ($or$cond2) {
     switch ($idx|0) {
     case 0:  {
      $24 = ((($face)) + 460|0);
      $25 = HEAP32[$24>>2]|0;
      $val1$0 = $25;
      break;
     }
     case 1:  {
      $26 = ((($face)) + 460|0);
      $27 = ((($26)) + 4|0);
      $28 = HEAP32[$27>>2]|0;
      $val1$0 = $28;
      break;
     }
     case 2:  {
      $29 = ((($face)) + 460|0);
      $30 = ((($29)) + 8|0);
      $31 = HEAP32[$30>>2]|0;
      $val1$0 = $31;
      break;
     }
     case 3:  {
      $32 = ((($face)) + 460|0);
      $33 = ((($32)) + 12|0);
      $34 = HEAP32[$33>>2]|0;
      $val1$0 = $34;
      break;
     }
     default: {
      $val1$0 = 0;
     }
     }
     HEAP32[$value>>2] = $val1$0;
     $retval$0 = 4;
    } else {
     $retval$0 = 4;
    }
   } else {
    $retval$0 = 0;
   }
   break;
  }
  case 3:  {
   $35 = ($value|0)==(0|0);
   $36 = ($1|0)==(0);
   $or$cond3 = $35 | $36;
   if ($or$cond3) {
    $retval$0 = 1;
   } else {
    $37 = ((($face)) + 432|0);
    $38 = HEAP8[$37>>0]|0;
    HEAP8[$value>>0] = $38;
    $retval$0 = 1;
   }
   break;
  }
  case 4:  {
   $39 = ((($face)) + 364|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = (_strlen($40)|0);
   $42 = (($41) + 1)|0;
   $43 = ($value|0)==(0|0);
   $44 = ($1>>>0)<($42>>>0);
   $or$cond4 = $43 | $44;
   if ($or$cond4) {
    $retval$0 = $42;
   } else {
    _memcpy(($value|0),($40|0),($42|0))|0;
    $retval$0 = $42;
   }
   break;
  }
  case 5:  {
   $45 = ($value|0)!=(0|0);
   $46 = ($1>>>0)>(3);
   $or$cond5 = $45 & $46;
   if ($or$cond5) {
    $47 = ((($face)) + 168|0);
    $48 = HEAP32[$47>>2]|0;
    HEAP32[$value>>2] = $48;
    $retval$0 = 4;
   } else {
    $retval$0 = 4;
   }
   break;
  }
  case 6:  {
   $49 = ($value|0)!=(0|0);
   $50 = ($1>>>0)>(3);
   $or$cond6 = $49 & $50;
   if ($or$cond6) {
    $51 = ((($face)) + 416|0);
    $52 = HEAP32[$51>>2]|0;
    HEAP32[$value>>2] = $52;
    $retval$0 = 4;
   } else {
    $retval$0 = 4;
   }
   break;
  }
  case 7:  {
   $53 = ((($face)) + 416|0);
   $54 = HEAP32[$53>>2]|0;
   $55 = ($54>>>0)>($idx>>>0);
   if ($55) {
    $56 = ((($face)) + 420|0);
    $57 = HEAP32[$56>>2]|0;
    $58 = (($57) + ($idx<<2)|0);
    $59 = HEAP32[$58>>2]|0;
    $60 = (_strlen($59)|0);
    $61 = (($60) + 1)|0;
    $62 = ($value|0)==(0|0);
    $63 = ($1>>>0)<($61>>>0);
    $or$cond7 = $62 | $63;
    if ($or$cond7) {
     $retval$0 = $61;
    } else {
     _memcpy(($value|0),($59|0),($61|0))|0;
     $64 = (($value) + ($60)|0);
     HEAP8[$64>>0] = 0;
     $retval$0 = $61;
    }
   } else {
    $retval$0 = 0;
   }
   break;
  }
  case 8:  {
   $65 = ((($face)) + 416|0);
   $66 = HEAP32[$65>>2]|0;
   $67 = ($66>>>0)>($idx>>>0);
   if ($67) {
    $68 = ((($face)) + 428|0);
    $69 = HEAP32[$68>>2]|0;
    $70 = (($69) + ($idx<<2)|0);
    $71 = HEAP32[$70>>2]|0;
    $72 = (($71) + 1)|0;
    $73 = ($value|0)==(0|0);
    $74 = ($1>>>0)<($72>>>0);
    $or$cond8 = $73 | $74;
    if ($or$cond8) {
     $retval$0 = $72;
    } else {
     $75 = ((($face)) + 424|0);
     $76 = HEAP32[$75>>2]|0;
     $77 = (($76) + ($idx<<2)|0);
     $78 = HEAP32[$77>>2]|0;
     _memcpy(($value|0),($78|0),($71|0))|0;
     $79 = (($value) + ($71)|0);
     HEAP8[$79>>0] = 0;
     $retval$0 = $72;
    }
   } else {
    $retval$0 = 0;
   }
   break;
  }
  case 9:  {
   $80 = ($value|0)!=(0|0);
   $81 = ($1>>>0)>(3);
   $or$cond9 = $80 & $81;
   if ($or$cond9) {
    $82 = ((($face)) + 368|0);
    $83 = HEAP32[$82>>2]|0;
    HEAP32[$value>>2] = $83;
    $retval$0 = 4;
   } else {
    $retval$0 = 4;
   }
   break;
  }
  case 10:  {
   $84 = ((($face)) + 368|0);
   $85 = HEAP32[$84>>2]|0;
   $86 = ($85|0)==(1);
   if ($86) {
    $87 = ((($face)) + 372|0);
    $88 = HEAP32[$87>>2]|0;
    $89 = ($88>>>0)>($idx>>>0);
    if ($89) {
     $90 = ((($87)) + 16|0);
     $91 = HEAP32[$90>>2]|0;
     $92 = (($91) + ($idx<<2)|0);
     $93 = HEAP32[$92>>2]|0;
     $94 = (_strlen($93)|0);
     $95 = (($94) + 1)|0;
     $96 = ($value|0)==(0|0);
     $97 = ($1>>>0)<($95>>>0);
     $or$cond10 = $96 | $97;
     if ($or$cond10) {
      $retval$0 = $95;
     } else {
      _memcpy(($value|0),($93|0),($94|0))|0;
      $98 = (($value) + ($94)|0);
      HEAP8[$98>>0] = 0;
      $retval$0 = $95;
     }
    } else {
     $retval$0 = 0;
    }
   } else {
    $retval$0 = 0;
   }
   break;
  }
  case 11:  {
   $99 = ($value|0)!=(0|0);
   $100 = ($1>>>0)>(3);
   $or$cond11 = $99 & $100;
   if ($or$cond11) {
    $101 = ((($face)) + 404|0);
    $102 = HEAP32[$101>>2]|0;
    HEAP32[$value>>2] = $102;
    $retval$0 = 4;
   } else {
    $retval$0 = 4;
   }
   break;
  }
  case 12:  {
   $103 = ((($face)) + 404|0);
   $104 = HEAP32[$103>>2]|0;
   $105 = ($104>>>0)>($idx>>>0);
   if ($105) {
    $106 = ((($face)) + 412|0);
    $107 = HEAP32[$106>>2]|0;
    $108 = (($107) + ($idx<<2)|0);
    $109 = HEAP32[$108>>2]|0;
    $110 = (($109) + 1)|0;
    $111 = ($value|0)==(0|0);
    $112 = ($1>>>0)<($110>>>0);
    $or$cond12 = $111 | $112;
    if ($or$cond12) {
     $retval$0 = $110;
    } else {
     $113 = ((($face)) + 408|0);
     $114 = HEAP32[$113>>2]|0;
     $115 = (($114) + ($idx<<2)|0);
     $116 = HEAP32[$115>>2]|0;
     _memcpy(($value|0),($116|0),($109|0))|0;
     $117 = (($value) + ($109)|0);
     HEAP8[$117>>0] = 0;
     $retval$0 = $110;
    }
   } else {
    $retval$0 = 0;
   }
   break;
  }
  case 13:  {
   $118 = ($value|0)!=(0|0);
   $119 = ($1>>>0)>(1);
   $or$cond13 = $118 & $119;
   if ($or$cond13) {
    $120 = ((($face)) + 168|0);
    $121 = ((($120)) + 120|0);
    $122 = HEAP16[$121>>1]|0;
    HEAP16[$value>>1] = $122;
    $retval$0 = 2;
   } else {
    $retval$0 = 2;
   }
   break;
  }
  case 14:  {
   $123 = ($value|0)!=(0|0);
   $124 = ($1>>>0)>(1);
   $or$cond14 = $123 & $124;
   if ($or$cond14) {
    $125 = ((($face)) + 168|0);
    $126 = ((($125)) + 122|0);
    $127 = HEAP16[$126>>1]|0;
    HEAP16[$value>>1] = $127;
    $retval$0 = 2;
   } else {
    $retval$0 = 2;
   }
   break;
  }
  case 15:  {
   $128 = ($value|0)==(0|0);
   $129 = ($1|0)==(0);
   $or$cond15 = $128 | $129;
   if ($or$cond15) {
    $retval$0 = 1;
   } else {
    $130 = ((($face)) + 168|0);
    $131 = ((($130)) + 8|0);
    $132 = HEAP8[$131>>0]|0;
    HEAP8[$value>>0] = $132;
    $retval$0 = 1;
   }
   break;
  }
  case 16:  {
   $133 = ((($face)) + 168|0);
   $134 = ((($133)) + 8|0);
   $135 = HEAP8[$134>>0]|0;
   $136 = $135&255;
   $137 = ($136>>>0)>($idx>>>0);
   if ($137) {
    $138 = ($value|0)!=(0|0);
    $139 = ($1>>>0)>(1);
    $or$cond16 = $138 & $139;
    if ($or$cond16) {
     $140 = ((($133)) + 12|0);
     $141 = (($140) + ($idx<<1)|0);
     $142 = HEAP16[$141>>1]|0;
     HEAP16[$value>>1] = $142;
     $retval$0 = 2;
    } else {
     $retval$0 = 2;
    }
   } else {
    $retval$0 = 0;
   }
   break;
  }
  case 24:  {
   $143 = ($value|0)!=(0|0);
   $144 = ($1>>>0)>(3);
   $or$cond17 = $143 & $144;
   if ($or$cond17) {
    $145 = ((($face)) + 168|0);
    $146 = ((($145)) + 108|0);
    $147 = HEAP32[$146>>2]|0;
    HEAP32[$value>>2] = $147;
    $retval$0 = 4;
   } else {
    $retval$0 = 4;
   }
   break;
  }
  case 17:  {
   $148 = ($value|0)!=(0|0);
   $149 = ($1>>>0)>(3);
   $or$cond18 = $148 & $149;
   if ($or$cond18) {
    $150 = ((($face)) + 168|0);
    $151 = ((($150)) + 116|0);
    $152 = HEAP32[$151>>2]|0;
    HEAP32[$value>>2] = $152;
    $retval$0 = 4;
   } else {
    $retval$0 = 4;
   }
   break;
  }
  case 25:  {
   $153 = ($value|0)!=(0|0);
   $154 = ($1>>>0)>(3);
   $or$cond19 = $153 & $154;
   if ($or$cond19) {
    $155 = ((($face)) + 168|0);
    $156 = ((($155)) + 112|0);
    $157 = HEAP32[$156>>2]|0;
    HEAP32[$value>>2] = $157;
    $retval$0 = 4;
   } else {
    $retval$0 = 4;
   }
   break;
  }
  case 18:  {
   $158 = ($value|0)==(0|0);
   $159 = ($1|0)==(0);
   $or$cond20 = $158 | $159;
   if ($or$cond20) {
    $retval$0 = 1;
   } else {
    $160 = ((($face)) + 168|0);
    $161 = ((($160)) + 9|0);
    $162 = HEAP8[$161>>0]|0;
    HEAP8[$value>>0] = $162;
    $retval$0 = 1;
   }
   break;
  }
  case 19:  {
   $163 = ((($face)) + 168|0);
   $164 = ((($163)) + 9|0);
   $165 = HEAP8[$164>>0]|0;
   $166 = $165&255;
   $167 = ($166>>>0)>($idx>>>0);
   if ($167) {
    $168 = ($value|0)!=(0|0);
    $169 = ($1>>>0)>(1);
    $or$cond21 = $168 & $169;
    if ($or$cond21) {
     $170 = ((($163)) + 40|0);
     $171 = (($170) + ($idx<<1)|0);
     $172 = HEAP16[$171>>1]|0;
     HEAP16[$value>>1] = $172;
     $retval$0 = 2;
    } else {
     $retval$0 = 2;
    }
   } else {
    $retval$0 = 0;
   }
   break;
  }
  case 20:  {
   $173 = ($value|0)==(0|0);
   $174 = ($1|0)==(0);
   $or$cond22 = $173 | $174;
   if ($or$cond22) {
    $retval$0 = 1;
   } else {
    $175 = ((($face)) + 168|0);
    $176 = ((($175)) + 10|0);
    $177 = HEAP8[$176>>0]|0;
    HEAP8[$value>>0] = $177;
    $retval$0 = 1;
   }
   break;
  }
  case 21:  {
   $178 = ((($face)) + 168|0);
   $179 = ((($178)) + 10|0);
   $180 = HEAP8[$179>>0]|0;
   $181 = $180&255;
   $182 = ($181>>>0)>($idx>>>0);
   if ($182) {
    $183 = ($value|0)!=(0|0);
    $184 = ($1>>>0)>(1);
    $or$cond23 = $183 & $184;
    if ($or$cond23) {
     $185 = ((($178)) + 60|0);
     $186 = (($185) + ($idx<<1)|0);
     $187 = HEAP16[$186>>1]|0;
     HEAP16[$value>>1] = $187;
     $retval$0 = 2;
    } else {
     $retval$0 = 2;
    }
   } else {
    $retval$0 = 0;
   }
   break;
  }
  case 22:  {
   $188 = ($value|0)==(0|0);
   $189 = ($1|0)==(0);
   $or$cond24 = $188 | $189;
   if ($or$cond24) {
    $retval$0 = 1;
   } else {
    $190 = ((($face)) + 168|0);
    $191 = ((($190)) + 11|0);
    $192 = HEAP8[$191>>0]|0;
    HEAP8[$value>>0] = $192;
    $retval$0 = 1;
   }
   break;
  }
  case 23:  {
   $193 = ((($face)) + 168|0);
   $194 = ((($193)) + 11|0);
   $195 = HEAP8[$194>>0]|0;
   $196 = $195&255;
   $197 = ($196>>>0)>($idx>>>0);
   if ($197) {
    $198 = ($value|0)!=(0|0);
    $199 = ($1>>>0)>(1);
    $or$cond25 = $198 & $199;
    if ($or$cond25) {
     $200 = ((($193)) + 88|0);
     $201 = (($200) + ($idx<<1)|0);
     $202 = HEAP16[$201>>1]|0;
     HEAP16[$value>>1] = $202;
     $retval$0 = 2;
    } else {
     $retval$0 = 2;
    }
   } else {
    $retval$0 = 0;
   }
   break;
  }
  case 26:  {
   $203 = ($value|0)==(0|0);
   $204 = ($1|0)==(0);
   $or$cond26 = $203 | $204;
   if ($or$cond26) {
    $retval$0 = 1;
   } else {
    $205 = ((($face)) + 168|0);
    $206 = ((($205)) + 124|0);
    $207 = HEAP8[$206>>0]|0;
    HEAP8[$value>>0] = $207;
    $retval$0 = 1;
   }
   break;
  }
  case 27:  {
   $208 = ((($face)) + 168|0);
   $209 = ((($208)) + 124|0);
   $210 = HEAP8[$209>>0]|0;
   $211 = $210&255;
   $212 = ($211>>>0)>($idx>>>0);
   if ($212) {
    $213 = ($value|0)!=(0|0);
    $214 = ($1>>>0)>(1);
    $or$cond27 = $213 & $214;
    if ($or$cond27) {
     $215 = ((($208)) + 128|0);
     $216 = (($215) + ($idx<<1)|0);
     $217 = HEAP16[$216>>1]|0;
     HEAP16[$value>>1] = $217;
     $retval$0 = 2;
    } else {
     $retval$0 = 2;
    }
   } else {
    $retval$0 = 0;
   }
   break;
  }
  case 28:  {
   $218 = ($value|0)==(0|0);
   $219 = ($1|0)==(0);
   $or$cond28 = $218 | $219;
   if ($or$cond28) {
    $retval$0 = 1;
   } else {
    $220 = ((($face)) + 168|0);
    $221 = ((($220)) + 125|0);
    $222 = HEAP8[$221>>0]|0;
    HEAP8[$value>>0] = $222;
    $retval$0 = 1;
   }
   break;
  }
  case 29:  {
   $223 = ((($face)) + 168|0);
   $224 = ((($223)) + 125|0);
   $225 = HEAP8[$224>>0]|0;
   $226 = $225&255;
   $227 = ($226>>>0)>($idx>>>0);
   if ($227) {
    $228 = ($value|0)!=(0|0);
    $229 = ($1>>>0)>(1);
    $or$cond29 = $228 & $229;
    if ($or$cond29) {
     $230 = (((($223)) + 154|0) + ($idx<<1)|0);
     $231 = HEAP16[$230>>1]|0;
     HEAP16[$value>>1] = $231;
     $retval$0 = 2;
    } else {
     $retval$0 = 2;
    }
   } else {
    $retval$0 = 0;
   }
   break;
  }
  case 31:  {
   $232 = ($value|0)==(0|0);
   $233 = ($1|0)==(0);
   $or$cond30 = $232 | $233;
   if ($or$cond30) {
    $retval$0 = 1;
   } else {
    $234 = ((($face)) + 168|0);
    $235 = ((($234)) + 127|0);
    $236 = HEAP8[$235>>0]|0;
    HEAP8[$value>>0] = $236;
    $retval$0 = 1;
   }
   break;
  }
  case 30:  {
   $237 = ($value|0)==(0|0);
   $238 = ($1|0)==(0);
   $or$cond31 = $237 | $238;
   if ($or$cond31) {
    $retval$0 = 1;
   } else {
    $239 = ((($face)) + 168|0);
    $240 = ((($239)) + 126|0);
    $241 = HEAP8[$240>>0]|0;
    HEAP8[$value>>0] = $241;
    $retval$0 = 1;
   }
   break;
  }
  case 32:  {
   $242 = ($idx>>>0)<(2);
   if ($242) {
    $243 = ($value|0)!=(0|0);
    $244 = ($1>>>0)>(1);
    $or$cond32 = $243 & $244;
    if ($or$cond32) {
     $245 = ((($face)) + 168|0);
     $246 = ((($245)) + 192|0);
     $247 = (($246) + ($idx<<1)|0);
     $248 = HEAP16[$247>>1]|0;
     HEAP16[$value>>1] = $248;
     $retval$0 = 2;
    } else {
     $retval$0 = 2;
    }
   } else {
    $retval$0 = 0;
   }
   break;
  }
  case 33:  {
   $249 = ($value|0)!=(0|0);
   $250 = ($1>>>0)>(3);
   $or$cond33 = $249 & $250;
   if ($or$cond33) {
    $251 = ((($face)) + 168|0);
    $252 = ((($251)) + 4|0);
    $253 = HEAP32[$252>>2]|0;
    HEAP32[$value>>2] = $253;
    $retval$0 = 4;
   } else {
    $retval$0 = 4;
   }
   break;
  }
  case 34:  {
   $254 = ($value|0)!=(0|0);
   $255 = ($1>>>0)>(3);
   $or$cond34 = $254 & $255;
   if ($or$cond34) {
    $256 = ((($face)) + 168|0);
    $257 = ((($256)) + 188|0);
    $258 = HEAP32[$257>>2]|0;
    HEAP32[$value>>2] = $258;
    $retval$0 = 4;
   } else {
    $retval$0 = 4;
   }
   break;
  }
  case 35:  {
   $259 = ($value|0)!=(0|0);
   $260 = ($1>>>0)>(3);
   $or$cond35 = $259 & $260;
   if ($or$cond35) {
    $261 = ((($face)) + 168|0);
    $262 = ((($261)) + 184|0);
    $263 = HEAP32[$262>>2]|0;
    HEAP32[$value>>2] = $263;
    $retval$0 = 4;
   } else {
    $retval$0 = 4;
   }
   break;
  }
  case 41:  {
   $264 = ($value|0)==(0|0);
   $265 = ($1|0)==(0);
   $or$cond36 = $264 | $265;
   if ($or$cond36) {
    $retval$0 = 1;
   } else {
    $266 = ((($face)) + 156|0);
    $267 = HEAP8[$266>>0]|0;
    HEAP8[$value>>0] = $267;
    $retval$0 = 1;
   }
   break;
  }
  case 42:  {
   $268 = ($value|0)!=(0|0);
   $269 = ($1>>>0)>(1);
   $or$cond37 = $268 & $269;
   if ($or$cond37) {
    $270 = ((($2)) + 26|0);
    $271 = HEAP16[$270>>1]|0;
    HEAP16[$value>>1] = $271;
    $retval$0 = 2;
   } else {
    $retval$0 = 2;
   }
   break;
  }
  case 43:  {
   $272 = ($value|0)!=(0|0);
   $273 = ($1>>>0)>(1);
   $or$cond38 = $272 & $273;
   if ($or$cond38) {
    $274 = ((($face)) + 160|0);
    $275 = HEAP16[$274>>1]|0;
    HEAP16[$value>>1] = $275;
    $retval$0 = 2;
   } else {
    $retval$0 = 2;
   }
   break;
  }
  case 44:  {
   $276 = ($value|0)!=(0|0);
   $277 = ($1>>>0)>(1);
   $or$cond39 = $276 & $277;
   if ($or$cond39) {
    $278 = ((($face)) + 164|0);
    $279 = HEAP16[$278>>1]|0;
    HEAP16[$value>>1] = $279;
    $retval$0 = 2;
   } else {
    $retval$0 = 2;
   }
   break;
  }
  case 36:  {
   $280 = HEAP32[$2>>2]|0;
   $281 = (_strlen($280)|0);
   $282 = (($281) + 1)|0;
   $283 = ($value|0)==(0|0);
   $284 = ($1>>>0)<($282>>>0);
   $or$cond40 = $283 | $284;
   if ($or$cond40) {
    $retval$0 = $282;
   } else {
    _memcpy(($value|0),($280|0),($282|0))|0;
    $retval$0 = $282;
   }
   break;
  }
  case 37:  {
   $285 = ((($face)) + 136|0);
   $286 = HEAP32[$285>>2]|0;
   $287 = (_strlen($286)|0);
   $288 = (($287) + 1)|0;
   $289 = ($value|0)==(0|0);
   $290 = ($1>>>0)<($288>>>0);
   $or$cond41 = $289 | $290;
   if ($or$cond41) {
    $retval$0 = $288;
   } else {
    _memcpy(($value|0),($286|0),($288|0))|0;
    $retval$0 = $288;
   }
   break;
  }
  case 38:  {
   $291 = ((($face)) + 140|0);
   $292 = HEAP32[$291>>2]|0;
   $293 = (_strlen($292)|0);
   $294 = (($293) + 1)|0;
   $295 = ($value|0)==(0|0);
   $296 = ($1>>>0)<($294>>>0);
   $or$cond42 = $295 | $296;
   if ($or$cond42) {
    $retval$0 = $294;
   } else {
    _memcpy(($value|0),($292|0),($294|0))|0;
    $retval$0 = $294;
   }
   break;
  }
  case 39:  {
   $297 = ((($face)) + 144|0);
   $298 = HEAP32[$297>>2]|0;
   $299 = (_strlen($298)|0);
   $300 = (($299) + 1)|0;
   $301 = ($value|0)==(0|0);
   $302 = ($1>>>0)<($300>>>0);
   $or$cond43 = $301 | $302;
   if ($or$cond43) {
    $retval$0 = $300;
   } else {
    _memcpy(($value|0),($298|0),($300|0))|0;
    $retval$0 = $300;
   }
   break;
  }
  case 40:  {
   $303 = ((($face)) + 148|0);
   $304 = HEAP32[$303>>2]|0;
   $305 = (_strlen($304)|0);
   $306 = (($305) + 1)|0;
   $307 = ($value|0)==(0|0);
   $308 = ($1>>>0)<($306>>>0);
   $or$cond44 = $307 | $308;
   if ($or$cond44) {
    $retval$0 = $306;
   } else {
    _memcpy(($value|0),($304|0),($306|0))|0;
    $retval$0 = $306;
   }
   break;
  }
  case 45:  {
   $309 = ($value|0)!=(0|0);
   $310 = ($1>>>0)>(3);
   $or$cond45 = $309 & $310;
   if ($or$cond45) {
    $311 = ((($face)) + 152|0);
    $312 = HEAP32[$311>>2]|0;
    HEAP32[$value>>2] = $312;
    $retval$0 = 4;
   } else {
    $retval$0 = 4;
   }
   break;
  }
  default: {
   $retval$0 = 0;
  }
  }
 } while(0);
 $313 = ($retval$0|0)==(0);
 $314 = $313 ? -1 : $retval$0;
 return ($314|0);
}
function _T1_Get_Track_Kerning($face,$ptsize,$degree,$kerning) {
 $face = $face|0;
 $ptsize = $ptsize|0;
 $degree = $degree|0;
 $kerning = $kerning|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 492|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = 6;
  return ($$0|0);
 }
 $3 = ((($1)) + 32|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0);
 if ($5) {
  $$0 = 0;
  return ($$0|0);
 }
 $6 = ((($1)) + 28|0);
 $32 = $4;$i$02 = 0;
 while(1) {
  $7 = HEAP32[$6>>2]|0;
  $8 = (($7) + (($i$02*20)|0)|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)==($degree|0);
  do {
   if ($10) {
    $11 = (((($7) + (($i$02*20)|0)|0)) + 4|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = ($12|0)>($ptsize|0);
    if ($13) {
     $14 = (((($7) + (($i$02*20)|0)|0)) + 8|0);
     $15 = HEAP32[$14>>2]|0;
     HEAP32[$kerning>>2] = $15;
     $31 = $32;
     break;
    }
    $16 = (((($7) + (($i$02*20)|0)|0)) + 12|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = ($17|0)<($ptsize|0);
    $19 = (((($7) + (($i$02*20)|0)|0)) + 16|0);
    $20 = HEAP32[$19>>2]|0;
    if ($18) {
     HEAP32[$kerning>>2] = $20;
     $31 = $32;
     break;
    } else {
     $21 = (($ptsize) - ($12))|0;
     $22 = (((($7) + (($i$02*20)|0)|0)) + 8|0);
     $23 = HEAP32[$22>>2]|0;
     $24 = (($20) - ($23))|0;
     $25 = (($17) - ($12))|0;
     $26 = (_FT_MulDiv($21,$24,$25)|0);
     $27 = HEAP32[$22>>2]|0;
     $28 = (($27) + ($26))|0;
     HEAP32[$kerning>>2] = $28;
     $$pre = HEAP32[$3>>2]|0;
     $31 = $$pre;
     break;
    }
   } else {
    $31 = $32;
   }
  } while(0);
  $29 = (($i$02) + 1)|0;
  $30 = ($29>>>0)<($31>>>0);
  if ($30) {
   $32 = $31;$i$02 = $29;
  } else {
   $$0 = 0;
   break;
  }
 }
 return ($$0|0);
}
function _T1_Get_Multi_Master($face,$master) {
 $face = $face|0;
 $master = $master|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$0 = 0, $n$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 528|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $error$0 = 6;
  return ($error$0|0);
 }
 $3 = ((($1)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 HEAP32[$master>>2] = $4;
 $5 = HEAP32[$1>>2]|0;
 $6 = ((($master)) + 4|0);
 HEAP32[$6>>2] = $5;
 $7 = ($4|0)==(0);
 if ($7) {
  $error$0 = 0;
  return ($error$0|0);
 } else {
  $n$01 = 0;
 }
 while(1) {
  $8 = (((($master)) + 8|0) + (($n$01*12)|0)|0);
  $9 = (((($1)) + 8|0) + ($n$01<<2)|0);
  $10 = HEAP32[$9>>2]|0;
  HEAP32[$8>>2] = $10;
  $11 = (((((($1)) + 88|0) + (($n$01*12)|0)|0)) + 4|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = HEAP32[$12>>2]|0;
  $14 = (((((($master)) + 8|0) + (($n$01*12)|0)|0)) + 4|0);
  HEAP32[$14>>2] = $13;
  $15 = (((($1)) + 88|0) + (($n$01*12)|0)|0);
  $16 = HEAP8[$15>>0]|0;
  $17 = $16&255;
  $18 = (($17) + -1)|0;
  $19 = (($12) + ($18<<2)|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = (((((($master)) + 8|0) + (($n$01*12)|0)|0)) + 8|0);
  HEAP32[$21>>2] = $20;
  $22 = (($n$01) + 1)|0;
  $23 = ($22>>>0)<($4>>>0);
  if ($23) {
   $n$01 = $22;
  } else {
   $error$0 = 0;
   break;
  }
 }
 return ($error$0|0);
}
function _T1_Set_MM_Design($face,$num_coords,$coords) {
 $face = $face|0;
 $num_coords = $num_coords|0;
 $coords = $coords|0;
 var $$$$$i = 0, $$$$i = 0, $$$i = 0, $$0 = 0, $$lcssa = 0, $$lcssa37 = 0, $$lcssa39 = 0, $$lcssa727 = 0, $$num_coords = 0, $$num_coords$i = 0, $$pre = 0, $$pre$i = 0, $$pre$phiZ2D = 0, $$pre25 = 0, $$pre26 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $after$2 = 0, $before$09 = 0, $before$09$phi = 0, $before$2 = 0, $design$0 = 0, $final_blends = 0, $m$02$i = 0, $n$017 = 0, $n$04$i = 0, $p$08 = 0, $p$08$lcssa = 0, $result$0$lcssa$i = 0, $result$03$i = 0, $the_blend$3 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $final_blends = sp;
 $0 = ((($face)) + 528|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ((($1)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4>>>0)<($num_coords>>>0);
 $$num_coords = $5 ? $4 : $num_coords;
 $6 = ($4|0)==(0);
 if ($6) {
  $$lcssa727 = 0;$61 = $1;
 } else {
  $88 = $4;$n$017 = 0;
  while(1) {
   $7 = (((($1)) + 88|0) + (($n$017*12)|0)|0);
   $8 = (((((($1)) + 88|0) + (($n$017*12)|0)|0)) + 4|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (((((($1)) + 88|0) + (($n$017*12)|0)|0)) + 8|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = ($n$017>>>0)<($$num_coords>>>0);
   if ($12) {
    $13 = (($coords) + ($n$017<<2)|0);
    $14 = HEAP32[$13>>2]|0;
    $$pre = HEAP8[$7>>0]|0;
    $$pre$phiZ2D = $7;$24 = $$pre;$design$0 = $14;
   } else {
    $15 = HEAP8[$7>>0]|0;
    $16 = $15&255;
    $17 = (($16) + -1)|0;
    $18 = (($9) + ($17<<2)|0);
    $19 = HEAP32[$18>>2]|0;
    $20 = HEAP32[$9>>2]|0;
    $21 = (($19) - ($20))|0;
    $22 = (($21|0) / 2)&-1;
    $$pre$phiZ2D = $7;$24 = $15;$design$0 = $22;
   }
   $23 = ($24<<24>>24)==(0);
   do {
    if ($23) {
     label = 12;
    } else {
     $before$09 = -1;$p$08 = 0;
     while(1) {
      $25 = (($9) + ($p$08<<2)|0);
      $26 = HEAP32[$25>>2]|0;
      $27 = ($design$0|0)==($26|0);
      if ($27) {
       $p$08$lcssa = $p$08;
       label = 8;
       break;
      }
      $30 = ($design$0|0)<($26|0);
      if ($30) {
       $after$2 = $p$08;$before$2 = $before$09;
       break;
      }
      $31 = (($p$08) + 1)|0;
      $32 = HEAP8[$$pre$phiZ2D>>0]|0;
      $33 = $32&255;
      $34 = ($31>>>0)<($33>>>0);
      if ($34) {
       $before$09$phi = $p$08;$p$08 = $31;$before$09 = $before$09$phi;
      } else {
       $after$2 = -1;$before$2 = $p$08;
       break;
      }
     }
     if ((label|0) == 8) {
      label = 0;
      $28 = (($11) + ($p$08$lcssa<<2)|0);
      $29 = HEAP32[$28>>2]|0;
      $58 = $88;$the_blend$3 = $29;
      break;
     }
     $35 = ($before$2|0)<(0);
     if ($35) {
      label = 12;
     } else {
      $37 = ($after$2|0)<(0);
      if ($37) {
       $38 = HEAP8[$$pre$phiZ2D>>0]|0;
       $39 = $38&255;
       $40 = (($39) + -1)|0;
       $41 = (($11) + ($40<<2)|0);
       $42 = HEAP32[$41>>2]|0;
       $58 = $88;$the_blend$3 = $42;
       break;
      } else {
       $43 = (($9) + ($before$2<<2)|0);
       $44 = HEAP32[$43>>2]|0;
       $45 = (($design$0) - ($44))|0;
       $46 = (($11) + ($after$2<<2)|0);
       $47 = HEAP32[$46>>2]|0;
       $48 = (($11) + ($before$2<<2)|0);
       $49 = HEAP32[$48>>2]|0;
       $50 = (($47) - ($49))|0;
       $51 = (($9) + ($after$2<<2)|0);
       $52 = HEAP32[$51>>2]|0;
       $53 = (($52) - ($44))|0;
       $54 = (_FT_MulDiv($45,$50,$53)|0);
       $$pre25 = HEAP32[$3>>2]|0;
       $58 = $$pre25;$the_blend$3 = $54;
       break;
      }
     }
    }
   } while(0);
   if ((label|0) == 12) {
    label = 0;
    $36 = HEAP32[$11>>2]|0;
    $58 = $88;$the_blend$3 = $36;
   }
   $55 = (($final_blends) + ($n$017<<2)|0);
   HEAP32[$55>>2] = $the_blend$3;
   $56 = (($n$017) + 1)|0;
   $57 = ($56>>>0)<($58>>>0);
   if ($57) {
    $88 = $58;$n$017 = $56;
   } else {
    $$lcssa39 = $58;
    break;
   }
  }
  $$pre26 = HEAP32[$0>>2]|0;
  $59 = ($$pre26|0)==(0|0);
  if ($59) {
   $$0 = 6;
   STACKTOP = sp;return ($$0|0);
  } else {
   $$lcssa727 = $$lcssa39;$61 = $$pre26;
  }
 }
 $60 = ((($61)) + 4|0);
 $62 = HEAP32[$60>>2]|0;
 $63 = ($62>>>0)<($$lcssa727>>>0);
 $$num_coords$i = $63 ? $62 : $$lcssa727;
 $64 = HEAP32[$61>>2]|0;
 $65 = ($64|0)==(0);
 if ($65) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $66 = ((($61)) + 136|0);
 $68 = $62;$89 = $64;$n$04$i = 0;
 while(1) {
  $67 = ($68|0)==(0);
  if ($67) {
   $87 = $89;$90 = 0;$result$0$lcssa$i = 65536;
  } else {
   $m$02$i = 0;$result$03$i = 65536;
   while(1) {
    $69 = ($m$02$i>>>0)<($$num_coords$i>>>0);
    if ($69) {
     $70 = (($final_blends) + ($m$02$i<<2)|0);
     $71 = HEAP32[$70>>2]|0;
     $73 = $71;
    } else {
     $73 = 32768;
    }
    $72 = ($73|0)<(0);
    $$$i = $72 ? 0 : $73;
    $74 = ($$$i|0)>(65536);
    $$$$i = $74 ? 65536 : $$$i;
    $75 = 1 << $m$02$i;
    $76 = $75 & $n$04$i;
    $77 = ($76|0)==(0);
    $78 = (65536 - ($$$$i))|0;
    $$$$$i = $77 ? $78 : $$$$i;
    $79 = (_FT_MulFix($result$03$i,$$$$$i)|0);
    $80 = (($m$02$i) + 1)|0;
    $81 = HEAP32[$60>>2]|0;
    $82 = ($80>>>0)<($81>>>0);
    if ($82) {
     $m$02$i = $80;$result$03$i = $79;
    } else {
     $$lcssa = $79;$$lcssa37 = $81;
     break;
    }
   }
   $$pre$i = HEAP32[$61>>2]|0;
   $87 = $$pre$i;$90 = $$lcssa37;$result$0$lcssa$i = $$lcssa;
  }
  $83 = HEAP32[$66>>2]|0;
  $84 = (($83) + ($n$04$i<<2)|0);
  HEAP32[$84>>2] = $result$0$lcssa$i;
  $85 = (($n$04$i) + 1)|0;
  $86 = ($85>>>0)<($87>>>0);
  if ($86) {
   $68 = $90;$89 = $87;$n$04$i = $85;
  } else {
   $$0 = 0;
   break;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _T1_Set_MM_Blend($face,$num_coords,$coords) {
 $face = $face|0;
 $num_coords = $num_coords|0;
 $coords = $coords|0;
 var $$ = 0, $$$ = 0, $$$$ = 0, $$0 = 0, $$lcssa = 0, $$lcssa9 = 0, $$num_coords = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $m$02 = 0, $n$04 = 0, $result$0$lcssa = 0, $result$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 528|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = 6;
  return ($$0|0);
 }
 $3 = ((($1)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4>>>0)<($num_coords>>>0);
 $$num_coords = $5 ? $4 : $num_coords;
 $6 = HEAP32[$1>>2]|0;
 $7 = ($6|0)==(0);
 if ($7) {
  $$0 = 0;
  return ($$0|0);
 }
 $8 = ((($1)) + 136|0);
 $10 = $4;$30 = $6;$n$04 = 0;
 while(1) {
  $9 = ($10|0)==(0);
  if ($9) {
   $29 = $30;$31 = 0;$result$0$lcssa = 65536;
  } else {
   $m$02 = 0;$result$03 = 65536;
   while(1) {
    $11 = ($m$02>>>0)<($$num_coords>>>0);
    if ($11) {
     $12 = (($coords) + ($m$02<<2)|0);
     $13 = HEAP32[$12>>2]|0;
     $15 = $13;
    } else {
     $15 = 32768;
    }
    $14 = ($15|0)<(0);
    $$ = $14 ? 0 : $15;
    $16 = ($$|0)>(65536);
    $$$ = $16 ? 65536 : $$;
    $17 = 1 << $m$02;
    $18 = $17 & $n$04;
    $19 = ($18|0)==(0);
    $20 = (65536 - ($$$))|0;
    $$$$ = $19 ? $20 : $$$;
    $21 = (_FT_MulFix($result$03,$$$$)|0);
    $22 = (($m$02) + 1)|0;
    $23 = HEAP32[$3>>2]|0;
    $24 = ($22>>>0)<($23>>>0);
    if ($24) {
     $m$02 = $22;$result$03 = $21;
    } else {
     $$lcssa = $21;$$lcssa9 = $23;
     break;
    }
   }
   $$pre = HEAP32[$1>>2]|0;
   $29 = $$pre;$31 = $$lcssa9;$result$0$lcssa = $$lcssa;
  }
  $25 = HEAP32[$8>>2]|0;
  $26 = (($25) + ($n$04<<2)|0);
  HEAP32[$26>>2] = $result$0$lcssa;
  $27 = (($n$04) + 1)|0;
  $28 = ($27>>>0)<($29>>>0);
  if ($28) {
   $10 = $31;$30 = $29;$n$04 = $27;
  } else {
   $$0 = 0;
   break;
  }
 }
 return ($$0|0);
}
function _T1_Get_MM_Var($face,$master) {
 $face = $face|0;
 $master = $master|0;
 var $$0$i = 0, $$lcssa = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre12 = 0, $$pre13 = 0, $$pre14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0;
 var $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0;
 var $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $axiscoords = 0, $error = 0;
 var $exitcond = 0, $i$08 = 0, $i$15 = 0, $j$05$i = 0, $j$05$i$lcssa = 0, $mmaster = 0, $n$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $mmaster = sp + 24|0;
 $error = sp + 16|0;
 $axiscoords = sp;
 $0 = ((($face)) + 100|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 528|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  HEAP32[$error>>2] = 6;
  $208 = 6;
  STACKTOP = sp;return ($208|0);
 }
 $5 = ((($3)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 HEAP32[$mmaster>>2] = $6;
 $7 = HEAP32[$3>>2]|0;
 $8 = ((($mmaster)) + 4|0);
 HEAP32[$8>>2] = $7;
 $9 = ($6|0)==(0);
 if ($9) {
  $26 = 0;
 } else {
  $n$01$i = 0;
  while(1) {
   $10 = (((($mmaster)) + 8|0) + (($n$01$i*12)|0)|0);
   $11 = (((($3)) + 8|0) + ($n$01$i<<2)|0);
   $12 = HEAP32[$11>>2]|0;
   HEAP32[$10>>2] = $12;
   $13 = (((((($3)) + 88|0) + (($n$01$i*12)|0)|0)) + 4|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = HEAP32[$14>>2]|0;
   $16 = (((((($mmaster)) + 8|0) + (($n$01$i*12)|0)|0)) + 4|0);
   HEAP32[$16>>2] = $15;
   $17 = (((($3)) + 88|0) + (($n$01$i*12)|0)|0);
   $18 = HEAP8[$17>>0]|0;
   $19 = $18&255;
   $20 = (($19) + -1)|0;
   $21 = (($14) + ($20<<2)|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = (((((($mmaster)) + 8|0) + (($n$01$i*12)|0)|0)) + 8|0);
   HEAP32[$23>>2] = $22;
   $24 = (($n$01$i) + 1)|0;
   $exitcond = ($24|0)==($6|0);
   if ($exitcond) {
    break;
   } else {
    $n$01$i = $24;
   }
  }
  $$pre = HEAP32[$mmaster>>2]|0;
  $26 = $$pre;
 }
 HEAP32[$error>>2] = 0;
 $25 = ($26*24)|0;
 $27 = (($25) + 20)|0;
 $28 = (_ft_mem_alloc($1,$27,$error)|0);
 $29 = HEAP32[$error>>2]|0;
 $30 = ($29|0)==(0);
 if (!($30)) {
  $208 = $29;
  STACKTOP = sp;return ($208|0);
 }
 HEAP32[$28>>2] = $26;
 $31 = HEAP32[$8>>2]|0;
 $32 = ((($28)) + 4|0);
 HEAP32[$32>>2] = $31;
 $33 = ((($28)) + 8|0);
 HEAP32[$33>>2] = -1;
 $34 = ((($28)) + 20|0);
 $35 = ((($28)) + 12|0);
 HEAP32[$35>>2] = $34;
 $36 = ((($28)) + 16|0);
 HEAP32[$36>>2] = 0;
 $37 = ($26|0)==(0);
 if (!($37)) {
  $i$08 = 0;
  while(1) {
   $38 = (((($mmaster)) + 8|0) + (($i$08*12)|0)|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = (($34) + (($i$08*24)|0)|0);
   HEAP32[$40>>2] = $39;
   $41 = (((((($mmaster)) + 8|0) + (($i$08*12)|0)|0)) + 4|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = $42 << 16;
   $44 = (((($34) + (($i$08*24)|0)|0)) + 4|0);
   HEAP32[$44>>2] = $43;
   $45 = (((((($mmaster)) + 8|0) + (($i$08*12)|0)|0)) + 8|0);
   $46 = HEAP32[$45>>2]|0;
   $47 = $46 << 16;
   $48 = (((($34) + (($i$08*24)|0)|0)) + 12|0);
   HEAP32[$48>>2] = $47;
   $49 = (($47) + ($43))|0;
   $50 = (($49|0) / 2)&-1;
   $51 = (((($34) + (($i$08*24)|0)|0)) + 8|0);
   HEAP32[$51>>2] = $50;
   $52 = (((($34) + (($i$08*24)|0)|0)) + 20|0);
   HEAP32[$52>>2] = -1;
   $53 = (((($34) + (($i$08*24)|0)|0)) + 16|0);
   HEAP32[$53>>2] = -1;
   $54 = (($34) + (($i$08*24)|0)|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = (_strcmp($55,594457)|0);
   $57 = ($56|0)==(0);
   do {
    if ($57) {
     HEAP32[$53>>2] = 2003265652;
    } else {
     $58 = (_strcmp($55,593900)|0);
     $59 = ($58|0)==(0);
     if ($59) {
      HEAP32[$53>>2] = 2003072104;
      break;
     }
     $60 = (_strcmp($55,593906)|0);
     $61 = ($60|0)==(0);
     if ($61) {
      HEAP32[$53>>2] = 1869640570;
     }
    }
   } while(0);
   $62 = (($i$08) + 1)|0;
   $63 = ($62>>>0)<($26>>>0);
   if ($63) {
    $i$08 = $62;
   } else {
    break;
   }
  }
 }
 $64 = HEAP32[$3>>2]|0;
 $65 = HEAP32[$5>>2]|0;
 $66 = 1 << $65;
 $67 = ($64|0)==($66|0);
 L25: do {
  if ($67) {
   $68 = ((($3)) + 140|0);
   $69 = HEAP32[$68>>2]|0;
   switch ($65|0) {
   case 1:  {
    $70 = ((($69)) + 4|0);
    $71 = HEAP32[$70>>2]|0;
    HEAP32[$axiscoords>>2] = $71;
    $209 = $71;
    break;
   }
   case 2:  {
    $72 = ((($69)) + 12|0);
    $73 = HEAP32[$72>>2]|0;
    $74 = ((($69)) + 4|0);
    $75 = HEAP32[$74>>2]|0;
    $76 = (($75) + ($73))|0;
    HEAP32[$axiscoords>>2] = $76;
    $77 = ((($69)) + 8|0);
    $78 = HEAP32[$77>>2]|0;
    $79 = (($78) + ($73))|0;
    $80 = ((($axiscoords)) + 4|0);
    HEAP32[$80>>2] = $79;
    $209 = $76;
    break;
   }
   case 3:  {
    $81 = ((($69)) + 28|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = ((($69)) + 20|0);
    $84 = HEAP32[$83>>2]|0;
    $85 = (($84) + ($82))|0;
    $86 = ((($69)) + 12|0);
    $87 = HEAP32[$86>>2]|0;
    $88 = (($85) + ($87))|0;
    $89 = ((($69)) + 4|0);
    $90 = HEAP32[$89>>2]|0;
    $91 = (($88) + ($90))|0;
    HEAP32[$axiscoords>>2] = $91;
    $92 = ((($69)) + 24|0);
    $93 = HEAP32[$92>>2]|0;
    $94 = (($93) + ($82))|0;
    $95 = (($94) + ($87))|0;
    $96 = ((($69)) + 8|0);
    $97 = HEAP32[$96>>2]|0;
    $98 = (($95) + ($97))|0;
    $99 = ((($axiscoords)) + 4|0);
    HEAP32[$99>>2] = $98;
    $100 = (($94) + ($84))|0;
    $101 = ((($69)) + 16|0);
    $102 = HEAP32[$101>>2]|0;
    $103 = (($100) + ($102))|0;
    $104 = ((($axiscoords)) + 8|0);
    HEAP32[$104>>2] = $103;
    $209 = $91;
    break;
   }
   default: {
    $105 = ((($69)) + 60|0);
    $106 = HEAP32[$105>>2]|0;
    $107 = ((($69)) + 52|0);
    $108 = HEAP32[$107>>2]|0;
    $109 = (($108) + ($106))|0;
    $110 = ((($69)) + 44|0);
    $111 = HEAP32[$110>>2]|0;
    $112 = (($109) + ($111))|0;
    $113 = ((($69)) + 36|0);
    $114 = HEAP32[$113>>2]|0;
    $115 = (($112) + ($114))|0;
    $116 = ((($69)) + 28|0);
    $117 = HEAP32[$116>>2]|0;
    $118 = (($115) + ($117))|0;
    $119 = ((($69)) + 20|0);
    $120 = HEAP32[$119>>2]|0;
    $121 = (($118) + ($120))|0;
    $122 = ((($69)) + 12|0);
    $123 = HEAP32[$122>>2]|0;
    $124 = (($121) + ($123))|0;
    $125 = ((($69)) + 4|0);
    $126 = HEAP32[$125>>2]|0;
    $127 = (($124) + ($126))|0;
    HEAP32[$axiscoords>>2] = $127;
    $128 = ((($69)) + 56|0);
    $129 = HEAP32[$128>>2]|0;
    $130 = (($129) + ($106))|0;
    $131 = (($130) + ($111))|0;
    $132 = ((($69)) + 40|0);
    $133 = HEAP32[$132>>2]|0;
    $134 = (($131) + ($133))|0;
    $135 = (($134) + ($117))|0;
    $136 = ((($69)) + 24|0);
    $137 = HEAP32[$136>>2]|0;
    $138 = (($135) + ($137))|0;
    $139 = (($138) + ($123))|0;
    $140 = ((($69)) + 8|0);
    $141 = HEAP32[$140>>2]|0;
    $142 = (($139) + ($141))|0;
    $143 = ((($axiscoords)) + 4|0);
    HEAP32[$143>>2] = $142;
    $144 = (($130) + ($108))|0;
    $145 = ((($69)) + 48|0);
    $146 = HEAP32[$145>>2]|0;
    $147 = (($144) + ($146))|0;
    $148 = (($147) + ($117))|0;
    $149 = (($148) + ($137))|0;
    $150 = (($149) + ($120))|0;
    $151 = ((($69)) + 16|0);
    $152 = HEAP32[$151>>2]|0;
    $153 = (($150) + ($152))|0;
    $154 = ((($axiscoords)) + 8|0);
    HEAP32[$154>>2] = $153;
    $155 = (($147) + ($111))|0;
    $156 = (($155) + ($133))|0;
    $157 = (($156) + ($114))|0;
    $158 = ((($69)) + 32|0);
    $159 = HEAP32[$158>>2]|0;
    $160 = (($157) + ($159))|0;
    $161 = ((($axiscoords)) + 12|0);
    HEAP32[$161>>2] = $160;
    $209 = $127;
   }
   }
   $162 = HEAP32[$mmaster>>2]|0;
   $163 = ($162|0)==(0);
   if (!($163)) {
    $168 = $209;$210 = $34;$i$15 = 0;
    while(1) {
     $164 = (((((($3)) + 88|0) + (($i$15*12)|0)|0)) + 8|0);
     $165 = HEAP32[$164>>2]|0;
     $166 = HEAP32[$165>>2]|0;
     $167 = ($166|0)<($168|0);
     L35: do {
      if ($167) {
       $169 = (((($3)) + 88|0) + (($i$15*12)|0)|0);
       $170 = HEAP8[$169>>0]|0;
       $171 = $170&255;
       $172 = ($170&255)>(1);
       L37: do {
        if ($172) {
         $j$05$i = 1;
         while(1) {
          $177 = (($165) + ($j$05$i<<2)|0);
          $178 = HEAP32[$177>>2]|0;
          $179 = ($178|0)<($168|0);
          if (!($179)) {
           $$lcssa = $178;$j$05$i$lcssa = $j$05$i;
           break;
          }
          $196 = (($j$05$i) + 1)|0;
          $197 = ($196|0)<($171|0);
          if ($197) {
           $j$05$i = $196;
          } else {
           break L37;
          }
         }
         $180 = (($j$05$i$lcssa) + -1)|0;
         $181 = (((((($3)) + 88|0) + (($i$15*12)|0)|0)) + 4|0);
         $182 = HEAP32[$181>>2]|0;
         $183 = (($182) + ($180<<2)|0);
         $184 = HEAP32[$183>>2]|0;
         $185 = $184 << 16;
         $186 = (($182) + ($j$05$i$lcssa<<2)|0);
         $187 = HEAP32[$186>>2]|0;
         $188 = (($187) - ($184))|0;
         $189 = (($165) + ($180<<2)|0);
         $190 = HEAP32[$189>>2]|0;
         $191 = (($168) - ($190))|0;
         $192 = (($$lcssa) - ($190))|0;
         $193 = (_FT_DivFix($191,$192)|0);
         $194 = Math_imul($193, $188)|0;
         $195 = (($194) + ($185))|0;
         $$pre13 = HEAP32[$35>>2]|0;
         $$0$i = $195;$205 = $$pre13;
         break L35;
        }
       } while(0);
       $198 = (($171) + -1)|0;
       $199 = (((((($3)) + 88|0) + (($i$15*12)|0)|0)) + 4|0);
       $200 = HEAP32[$199>>2]|0;
       $201 = (($200) + ($198<<2)|0);
       $202 = HEAP32[$201>>2]|0;
       $203 = $202 << 16;
       $$0$i = $203;$205 = $210;
      } else {
       $173 = (((((($3)) + 88|0) + (($i$15*12)|0)|0)) + 4|0);
       $174 = HEAP32[$173>>2]|0;
       $175 = HEAP32[$174>>2]|0;
       $176 = $175 << 16;
       $$0$i = $176;$205 = $210;
      }
     } while(0);
     $204 = (((($205) + (($i$15*24)|0)|0)) + 8|0);
     HEAP32[$204>>2] = $$0$i;
     $206 = (($i$15) + 1)|0;
     $207 = ($206>>>0)<($162>>>0);
     if (!($207)) {
      break L25;
     }
     $$phi$trans$insert = (($axiscoords) + ($206<<2)|0);
     $$pre12 = HEAP32[$$phi$trans$insert>>2]|0;
     $168 = $$pre12;$210 = $205;$i$15 = $206;
    }
   }
  }
 } while(0);
 HEAP32[$master>>2] = $28;
 $$pre14 = HEAP32[$error>>2]|0;
 $208 = $$pre14;
 STACKTOP = sp;return ($208|0);
}
function _T1_Set_Var_Design($face,$num_coords,$coords) {
 $face = $face|0;
 $num_coords = $num_coords|0;
 $coords = $coords|0;
 var $$num_coords = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $lcoords = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lcoords = sp;
 $0 = ($num_coords>>>0)>(4);
 $$num_coords = $0 ? 4 : $num_coords;
 $1 = ($$num_coords|0)==(0);
 if (!($1)) {
  $i$01 = 0;
  while(1) {
   $2 = (($coords) + ($i$01<<2)|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = (_FT_RoundFix($3)|0);
   $5 = $4 >> 16;
   $6 = (($lcoords) + ($i$01<<2)|0);
   HEAP32[$6>>2] = $5;
   $7 = (($i$01) + 1)|0;
   $8 = ($7>>>0)<($$num_coords>>>0);
   if ($8) {
    $i$01 = $7;
   } else {
    break;
   }
  }
 }
 $9 = (_T1_Set_MM_Design($face,$$num_coords,$lcoords)|0);
 STACKTOP = sp;return ($9|0);
}
function _T1_Face_Init($stream,$t1face,$face_index,$num_params,$params) {
 $stream = $stream|0;
 $t1face = $t1face|0;
 $face_index = $face_index|0;
 $num_params = $num_params|0;
 $params = $params|0;
 var $$ = 0, $$02 = 0, $$71 = 0, $$cast = 0, $$cast$i$i = 0, $$cast$i10$i = 0, $$idx$i = 0, $$idx$val$i = 0, $$idx$val$idx$val$i = 0, $$lcssa = 0, $$not$i = 0, $$off$i$i$i = 0, $$off$i$i4$i = 0, $$off$i29$i$i = 0, $$ph$i = 0, $$ph$i$i = 0, $$phi$trans$insert = 0, $$pr12$pr$pre$i = 0, $$pr14$pre$i = 0, $$pre = 0;
 var $$pre$i = 0, $$pre$i$i = 0, $$pre$i7 = 0, $$pre$i8$i = 0, $$pre$phi$i$iZ2D = 0, $$pre$phi77$i$iZ2D = 0, $$pre6$i$i = 0, $$pre74$i$i = 0, $$pre75$i$i = 0, $$pre76$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0;
 var $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0;
 var $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0;
 var $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0;
 var $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0;
 var $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0;
 var $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0;
 var $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0;
 var $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0;
 var $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0;
 var $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0;
 var $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0;
 var $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0;
 var $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0;
 var $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0;
 var $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0;
 var $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0;
 var $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0;
 var $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0;
 var $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0;
 var $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0;
 var $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0;
 var $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $charcode$0$min_char$0$i = 0, $charcode$033$i = 0, $charcode$033$i$us = 0, $charmap = 0;
 var $clazz$0 = 0, $clazz$0$in = 0, $cur$0$i$i = 0, $cur$0$i$i$lcssa = 0, $cur$0$ph$i$i = 0, $cur$155$i$i = 0, $cur$2$lcssa43$i$i = 0, $cur$254$i$i = 0, $cur$254$us$i$i = 0, $decoder$i = 0, $error$2 = 0, $error$i = 0, $error$i$i = 0, $error$i$i$i = 0, $error$i$i2$i = 0, $error$i28$i$i = 0, $error$i3$i = 0, $exitcond = 0, $exitcond51 = 0, $family$0$ph$be = 0;
 var $family$0$ph31 = 0, $family$029 = 0, $family$029$lcssa = 0, $full$0$ph30 = 0, $full$0$ph30$lcssa114 = 0, $glyph_data$i$i = 0, $glyph_index$01$i = 0, $i$040$i = 0, $idx$032$i = 0, $idx$032$i$lcssa = 0, $len$i$i = 0, $limit$0$ph$i$i = 0, $loader$i = 0, $max_advance$0 = 0, $max_advance$1 = 0, $max_advance$2 = 0, $max_char$0$$i = 0, $max_char$0$lcssa$i = 0, $max_char$037$i = 0, $max_char$1$i = 0;
 var $min_char$0$lcssa$i = 0, $min_char$036$i = 0, $min_char$2$i = 0, $not$ = 0, $or$cond = 0, $or$cond$i = 0, $or$cond12 = 0, $or$cond5 = 0, $or$cond5$i$i = 0, $sext = 0, $size$2$i$i = 0, $storemerge = 0, $switch$i$i$i = 0, $switch$i$i5$i = 0, $switch$i30$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1936|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $glyph_data$i$i = sp + 1920|0;
 $decoder$i = sp + 416|0;
 $error$i28$i$i = sp + 412|0;
 $error$i$i2$i = sp + 408|0;
 $error$i3$i = sp + 404|0;
 $len$i$i = sp + 400|0;
 $error$i$i$i = sp + 396|0;
 $error$i$i = sp + 392|0;
 $loader$i = sp + 16|0;
 $error$i = sp + 12|0;
 $charmap = sp;
 $0 = ((($t1face)) + 132|0);
 HEAP32[$t1face>>2] = 1;
 $1 = ((($t1face)) + 96|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (_ft_module_get_service($2,594361)|0);
 $4 = ((($t1face)) + 484|0);
 HEAP32[$4>>2] = $3;
 $5 = HEAP32[$1>>2]|0;
 $6 = ((($5)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (_FT_Get_Module_Interface($7,594378)|0);
 $9 = ((($t1face)) + 488|0);
 HEAP32[$9>>2] = $8;
 $10 = ($8|0)==(0|0);
 if ($10) {
  $$02 = 11;
  STACKTOP = sp;return ($$02|0);
 }
 $11 = HEAP32[$1>>2]|0;
 $12 = ((($11)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (_FT_Get_Module_Interface($13,593918)|0);
 $15 = ((($t1face)) + 548|0);
 HEAP32[$15>>2] = $14;
 $16 = HEAP32[$9>>2]|0;
 _memset(($loader$i|0),0,376)|0;
 $17 = ((($t1face)) + 532|0);
 HEAP32[$17>>2] = -1;
 $18 = ((($t1face)) + 536|0);
 HEAP32[$18>>2] = -1;
 $19 = ((($t1face)) + 540|0);
 HEAP32[$19>>2] = 0;
 $20 = ((($t1face)) + 280|0);
 HEAP32[$20>>2] = 7;
 $21 = ((($t1face)) + 284|0);
 HEAP32[$21>>2] = 1;
 $22 = ((($t1face)) + 172|0);
 HEAP32[$22>>2] = 4;
 $23 = ((($t1face)) + 348|0);
 HEAP32[$23>>2] = 3932;
 $24 = ((($t1face)) + 276|0);
 HEAP32[$24>>2] = 2596864;
 $25 = ((($t1face)) + 104|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ((($t1face)) + 100|0);
 $28 = HEAP32[$27>>2]|0;
 $$idx$i = ((($16)) + 4|0);
 $$idx$val$i = HEAP32[$$idx$i>>2]|0;
 $$idx$val$idx$val$i = HEAP32[$$idx$val$i>>2]|0;
 FUNCTION_TABLE_viiii[$$idx$val$idx$val$i & 8191]($loader$i,0,0,$28);
 $29 = ((($loader$i)) + 72|0);
 HEAP32[$29>>2] = $26;
 $30 = ((($loader$i)) + 80|0);
 $31 = ((($loader$i)) + 76|0);
 $32 = ((($loader$i)) + 92|0);
 $33 = ((($loader$i)) + 93|0);
 ;HEAP32[$31>>2]=0|0;HEAP32[$31+4>>2]=0|0;HEAP32[$31+8>>2]=0|0;HEAP32[$31+12>>2]=0|0;HEAP16[$31+16>>1]=0|0;HEAP8[$31+18>>0]=0|0;
 $34 = (_check_type1_format($26,593927,14)|0);
 HEAP32[$error$i$i>>2] = $34;
 $35 = ($34|0)==(0);
 if ($35) {
  label = 5;
 } else {
  $36 = $34 & 255;
  $37 = ($36|0)==(2);
  if ($37) {
   $38 = (_check_type1_format($26,593942,10)|0);
   HEAP32[$error$i$i>>2] = $38;
   $39 = ($38|0)==(0);
   if ($39) {
    label = 5;
   } else {
    $75 = $38;
    label = 22;
   }
  } else {
   $75 = $34;
   label = 22;
  }
 }
 L7: do {
  if ((label|0) == 5) {
   $40 = (_FT_Stream_Seek($26,0)|0);
   HEAP32[$error$i$i>>2] = $40;
   $41 = ($40|0)==(0);
   if ($41) {
    $42 = (_FT_Stream_ReadUShort($26,$error$i$i$i)|0);
    $43 = HEAP32[$error$i$i$i>>2]|0;
    $44 = ($43|0)==(0);
    do {
     if ($44) {
      $$off$i$i$i = (($42) + 32767)<<16>>16;
      $switch$i$i$i = ($$off$i$i$i&65535)<(2);
      if ($switch$i$i$i) {
       $45 = (_FT_Stream_ReadULongLE($26,$error$i$i$i)|0);
       $46 = HEAP32[$error$i$i$i>>2]|0;
       $47 = ($46|0)==(0);
       if (!($47)) {
        $48 = $46;
        break;
       }
       HEAP32[$error$i$i>>2] = 0;
       $49 = ($42<<16>>16)==(-32767);
       if ($49) {
        HEAP8[$32>>0] = 1;
        $size$2$i$i = $45;
       } else {
        label = 12;
       }
      } else {
       HEAP32[$error$i$i>>2] = 0;
       label = 12;
      }
      if ((label|0) == 12) {
       $50 = (_FT_Stream_Seek($26,0)|0);
       HEAP32[$error$i$i>>2] = $50;
       $51 = ($50|0)==(0);
       if (!($51)) {
        $75 = $50;
        label = 22;
        break L7;
       }
       $52 = ((($26)) + 4|0);
       $53 = HEAP32[$52>>2]|0;
       $size$2$i$i = $53;
      }
      $54 = ((($26)) + 20|0);
      $55 = HEAP32[$54>>2]|0;
      $56 = ($55|0)==(0|0);
      if ($56) {
       $57 = HEAP32[$26>>2]|0;
       $58 = ((($26)) + 8|0);
       $59 = HEAP32[$58>>2]|0;
       $60 = (($57) + ($59)|0);
       HEAP32[$31>>2] = $60;
       HEAP32[$30>>2] = $size$2$i$i;
       HEAP8[$33>>0] = 1;
       $61 = (_FT_Stream_Skip($26,$size$2$i$i)|0);
       HEAP32[$error$i$i>>2] = $61;
       $62 = ($61|0)==(0);
       if (!($62)) {
        $75 = $61;
        label = 22;
        break L7;
       }
       $$pre$i$i = HEAP32[$30>>2]|0;
       $71 = $$pre$i$i;
      } else {
       $63 = (_ft_mem_alloc($28,$size$2$i$i,$error$i$i)|0);
       HEAP32[$31>>2] = $63;
       $64 = HEAP32[$error$i$i>>2]|0;
       $65 = ($64|0)==(0);
       if (!($65)) {
        $75 = $64;
        label = 22;
        break L7;
       }
       $66 = (_FT_Stream_Read($26,$63,$size$2$i$i)|0);
       HEAP32[$error$i$i>>2] = $66;
       $67 = ($66|0)==(0);
       if (!($67)) {
        $75 = $66;
        label = 22;
        break L7;
       }
       HEAP32[$30>>2] = $size$2$i$i;
       $71 = $size$2$i$i;
      }
      $68 = HEAP32[$31>>2]|0;
      $69 = ((($loader$i)) + 4|0);
      HEAP32[$69>>2] = $68;
      $$cast$i$i = $68;
      HEAP32[$loader$i>>2] = $$cast$i$i;
      $70 = (($$cast$i$i) + ($71)|0);
      $72 = ((($loader$i)) + 8|0);
      HEAP32[$72>>2] = $70;
      HEAP32[$error$i>>2] = 0;
      $78 = $$cast$i$i;$79 = $71;
      label = 26;
      break L7;
     } else {
      $48 = $43;
     }
    } while(0);
    HEAP32[$error$i$i>>2] = $48;
    $75 = $48;
    label = 22;
   } else {
    $75 = $40;
    label = 22;
   }
  }
 } while(0);
 do {
  if ((label|0) == 22) {
   $73 = HEAP8[$33>>0]|0;
   $74 = ($73<<24>>24)==(0);
   if (!($74)) {
    HEAP32[$error$i>>2] = $75;
    break;
   }
   $76 = HEAP32[$31>>2]|0;
   _ft_mem_free($28,$76);
   HEAP32[$31>>2] = 0;
   $$pre6$i$i = HEAP32[$error$i$i>>2]|0;
   HEAP32[$error$i>>2] = $$pre6$i$i;
   $77 = ($$pre6$i$i|0)==(0);
   if ($77) {
    $$pre$i = HEAP32[$30>>2]|0;
    $78 = 0;$79 = $$pre$i;
    label = 26;
   }
  }
 } while(0);
 L35: do {
  if ((label|0) == 26) {
   $80 = (_parse_dict($t1face,$loader$i,$78,$79)|0);
   HEAP32[$error$i>>2] = $80;
   $81 = ($80|0)==(0);
   if ($81) {
    $82 = HEAP32[$29>>2]|0;
    $83 = ((($loader$i)) + 16|0);
    $84 = HEAP32[$83>>2]|0;
    HEAP32[$error$i3$i>>2] = 0;
    $85 = HEAP8[$32>>0]|0;
    $86 = ($85<<24>>24)==(0);
    L38: do {
     if ($86) {
      $128 = HEAP32[$31>>2]|0;
      $129 = HEAP32[$30>>2]|0;
      $130 = (($128) + ($129)|0);
      $131 = ((($loader$i)) + 8|0);
      $132 = ((($loader$i)) + 32|0);
      $133 = ((($loader$i)) + 12|0);
      $134 = ((($loader$i)) + 28|0);
      $136 = $128;$584 = $129;$cur$0$ph$i$i = $128;$limit$0$ph$i$i = $130;
      L40: while(1) {
       $135 = $136;
       $cur$0$i$i = $cur$0$ph$i$i;
       while(1) {
        $137 = HEAP8[$cur$0$i$i>>0]|0;
        $138 = ($137<<24>>24)==(101);
        $139 = ((($cur$0$i$i)) + 9|0);
        $140 = ($139>>>0)<($limit$0$ph$i$i>>>0);
        $or$cond5$i$i = $138 & $140;
        $141 = ((($cur$0$i$i)) + 1|0);
        if ($or$cond5$i$i) {
         $142 = HEAP8[$141>>0]|0;
         $143 = ($142<<24>>24)==(101);
         if ($143) {
          $144 = ((($cur$0$i$i)) + 2|0);
          $145 = HEAP8[$144>>0]|0;
          $146 = ($145<<24>>24)==(120);
          if ($146) {
           $147 = ((($cur$0$i$i)) + 3|0);
           $148 = HEAP8[$147>>0]|0;
           $149 = ($148<<24>>24)==(101);
           if ($149) {
            $150 = ((($cur$0$i$i)) + 4|0);
            $151 = HEAP8[$150>>0]|0;
            $152 = ($151<<24>>24)==(99);
            if ($152) {
             $cur$0$i$i$lcssa = $cur$0$i$i;
             break;
            }
           }
          }
         }
        }
        $153 = ($141>>>0)<($limit$0$ph$i$i>>>0);
        if ($153) {
         $cur$0$i$i = $141;
        } else {
         label = 56;
         break L40;
        }
       }
       HEAP32[$loader$i>>2] = $135;
       $154 = ((($cur$0$i$i$lcssa)) + 10|0);
       HEAP32[$131>>2] = $154;
       $155 = ($136>>>0)<($154>>>0);
       if ($155) {
        $cur$155$i$i = $136;
        while(1) {
         $156 = HEAP8[$cur$155$i$i>>0]|0;
         $157 = ($156<<24>>24)==(101);
         if ($157) {
          $158 = (_strncmp($cur$155$i$i,593953,5)|0);
          $159 = ($158|0)==(0);
          if ($159) {
           break L40;
          }
         }
         $160 = HEAP32[$132>>2]|0;
         FUNCTION_TABLE_vi[$160 & 8191]($loader$i);
         $161 = HEAP32[$133>>2]|0;
         $162 = ($161|0)==(0);
         if (!($162)) {
          break;
         }
         $163 = HEAP32[$134>>2]|0;
         FUNCTION_TABLE_vi[$163 & 8191]($loader$i);
         $164 = HEAP32[$loader$i>>2]|0;
         $165 = ($164>>>0)<($154>>>0);
         if ($165) {
          $cur$155$i$i = $164;
         } else {
          break;
         }
        }
        $$pre$i8$i = HEAP32[$31>>2]|0;
        $$pre74$i$i = HEAP32[$30>>2]|0;
        $167 = $$pre$i8$i;$168 = $$pre74$i$i;
       } else {
        $167 = $136;$168 = $584;
       }
       $166 = (($167) + ($168)|0);
       $136 = $167;$584 = $168;$cur$0$ph$i$i = $154;$limit$0$ph$i$i = $166;
      }
      if ((label|0) == 56) {
       HEAP32[$error$i3$i>>2] = 3;
       $$ph$i = 3;
       break;
      }
      $169 = HEAP32[$31>>2]|0;
      $170 = HEAP32[$30>>2]|0;
      $171 = (($169) + ($170)|0);
      HEAP32[$131>>2] = $171;
      $172 = HEAP32[$132>>2]|0;
      FUNCTION_TABLE_vi[$172 & 8191]($loader$i);
      $173 = HEAP32[$loader$i>>2]|0;
      $174 = HEAP32[$131>>2]|0;
      $175 = $174;
      $176 = $173;
      $177 = (($175) - ($176))|0;
      $178 = (_memchr($173,10,$177)|0);
      $179 = ($178|0)==(0|0);
      if ($179) {
       $585 = 1;
      } else {
       $180 = (_memchr($173,13,$177)|0);
       $181 = ($178>>>0)>($180>>>0);
       $585 = $181;
      }
      $182 = ($173>>>0)<($174>>>0);
      L66: do {
       if ($182) {
        L68: do {
         if ($585) {
          $cur$254$us$i$i = $173;
          while(1) {
           $183 = HEAP8[$cur$254$us$i$i>>0]|0;
           switch ($183<<24>>24) {
           case 10: case 13: case 9: case 32:  {
            break;
           }
           default: {
            $cur$2$lcssa43$i$i = $cur$254$us$i$i;
            break L68;
           }
           }
           $184 = ((($cur$254$us$i$i)) + 1|0);
           $185 = ($184>>>0)<($174>>>0);
           if ($185) {
            $cur$254$us$i$i = $184;
           } else {
            break L66;
           }
          }
         } else {
          $cur$254$i$i = $173;
          while(1) {
           $186 = HEAP8[$cur$254$i$i>>0]|0;
           switch ($186<<24>>24) {
           case 10: case 9: case 32:  {
            break;
           }
           default: {
            $cur$2$lcssa43$i$i = $cur$254$i$i;
            break L68;
           }
           }
           $187 = ((($cur$254$i$i)) + 1|0);
           $188 = ($187>>>0)<($174>>>0);
           if ($188) {
            $cur$254$i$i = $187;
           } else {
            break L66;
           }
          }
         }
        } while(0);
        $189 = HEAP32[$30>>2]|0;
        $190 = HEAP32[$31>>2]|0;
        $191 = $cur$2$lcssa43$i$i;
        $192 = (($190) - ($191))|0;
        $193 = (($192) + ($189))|0;
        $194 = HEAP8[$33>>0]|0;
        $195 = ($194<<24>>24)==(0);
        if ($195) {
         $202 = ((($loader$i)) + 94|0);
         HEAP8[$202>>0] = 1;
         $203 = ((($loader$i)) + 84|0);
         HEAP32[$203>>2] = $190;
         $204 = ((($loader$i)) + 88|0);
         HEAP32[$204>>2] = $193;
         HEAP32[$31>>2] = 0;
         HEAP32[$30>>2] = 0;
         $205 = $190;
         $231 = $205;
        } else {
         $196 = (($193) + 1)|0;
         $197 = (_ft_mem_alloc($84,$196,$error$i3$i)|0);
         $198 = ((($loader$i)) + 84|0);
         HEAP32[$198>>2] = $197;
         $199 = HEAP32[$error$i3$i>>2]|0;
         $200 = ($199|0)==(0);
         if (!($200)) {
          $$ph$i = $199;
          break L38;
         }
         $201 = ((($loader$i)) + 88|0);
         HEAP32[$201>>2] = $193;
         $231 = $197;
        }
        $206 = ((($cur$2$lcssa43$i$i)) + 3|0);
        $207 = ($206>>>0)<($174>>>0);
        L81: do {
         if ($207) {
          $208 = HEAP8[$cur$2$lcssa43$i$i>>0]|0;
          $209 = $208&255;
          $210 = (($209) + -48)|0;
          $211 = ($210>>>0)<(10);
          if (!($211)) {
           switch ($209|0) {
           case 65: case 66: case 67: case 68: case 69: case 70: case 97: case 98: case 99: case 100: case 101: case 102:  {
            break;
           }
           default: {
            break L81;
           }
           }
          }
          $212 = ((($cur$2$lcssa43$i$i)) + 1|0);
          $213 = HEAP8[$212>>0]|0;
          $214 = $213&255;
          $215 = (($214) + -48)|0;
          $216 = ($215>>>0)<(10);
          if (!($216)) {
           switch ($214|0) {
           case 65: case 66: case 67: case 68: case 69: case 70: case 97: case 98: case 99: case 100: case 101: case 102:  {
            break;
           }
           default: {
            break L81;
           }
           }
          }
          $217 = ((($cur$2$lcssa43$i$i)) + 2|0);
          $218 = HEAP8[$217>>0]|0;
          $219 = $218&255;
          $220 = (($219) + -48)|0;
          $221 = ($220>>>0)<(10);
          if (!($221)) {
           switch ($219|0) {
           case 65: case 66: case 67: case 68: case 69: case 70: case 97: case 98: case 99: case 100: case 101: case 102:  {
            break;
           }
           default: {
            break L81;
           }
           }
          }
          $222 = HEAP8[$206>>0]|0;
          $223 = $222&255;
          $224 = (($223) + -48)|0;
          $225 = ($224>>>0)<(10);
          if (!($225)) {
           switch ($223|0) {
           case 65: case 66: case 67: case 68: case 69: case 70: case 97: case 98: case 99: case 100: case 101: case 102:  {
            break;
           }
           default: {
            break L81;
           }
           }
          }
          HEAP32[$loader$i>>2] = $cur$2$lcssa43$i$i;
          $226 = HEAP32[$$idx$i>>2]|0;
          $227 = ((($226)) + 24|0);
          $228 = HEAP32[$227>>2]|0;
          $229 = ((($loader$i)) + 84|0);
          $230 = ((($loader$i)) + 88|0);
          (FUNCTION_TABLE_iiiiii[$228 & 8191]($loader$i,$231,$193,$len$i$i,0)|0);
          $232 = HEAP32[$len$i$i>>2]|0;
          HEAP32[$230>>2] = $232;
          $233 = HEAP32[$229>>2]|0;
          $234 = (($233) + ($232)|0);
          HEAP8[$234>>0] = 0;
          $$pre$phi$i$iZ2D = $229;$$pre$phi77$i$iZ2D = $230;
          label = 88;
          break L38;
         }
        } while(0);
        $235 = ((($loader$i)) + 84|0);
        _memmove(($231|0),($cur$2$lcssa43$i$i|0),($193|0))|0;
        $$pre76$i$i = ((($loader$i)) + 88|0);
        $$pre$phi$i$iZ2D = $235;$$pre$phi77$i$iZ2D = $$pre76$i$i;
        label = 88;
        break L38;
       }
      } while(0);
      HEAP32[$error$i3$i>>2] = 3;
      $$ph$i = 3;
     } else {
      $87 = (_FT_Stream_Pos($82)|0);
      $88 = ((($loader$i)) + 88|0);
      HEAP32[$88>>2] = 0;
      while(1) {
       $89 = (_FT_Stream_ReadUShort($82,$error$i$i2$i)|0);
       $90 = HEAP32[$error$i$i2$i>>2]|0;
       $91 = ($90|0)==(0);
       if (!($91)) {
        $95 = $90;
        label = 33;
        break;
       }
       $$off$i$i4$i = (($89) + 32767)<<16>>16;
       $switch$i$i5$i = ($$off$i$i4$i&65535)<(2);
       if (!($switch$i$i5$i)) {
        label = 31;
        break;
       }
       $92 = (_FT_Stream_ReadULongLE($82,$error$i$i2$i)|0);
       $93 = HEAP32[$error$i$i2$i>>2]|0;
       $94 = ($93|0)==(0);
       if (!($94)) {
        $95 = $93;
        label = 33;
        break;
       }
       HEAP32[$error$i3$i>>2] = 0;
       $96 = ($89<<16>>16)==(-32766);
       if (!($96)) {
        break;
       }
       $97 = HEAP32[$88>>2]|0;
       $98 = (($97) + ($92))|0;
       HEAP32[$88>>2] = $98;
       $99 = (_FT_Stream_Skip($82,$92)|0);
       HEAP32[$error$i3$i>>2] = $99;
       $100 = ($99|0)==(0);
       if (!($100)) {
        $$ph$i = $99;
        break L38;
       }
      }
      if ((label|0) == 31) {
       HEAP32[$error$i3$i>>2] = 0;
      }
      else if ((label|0) == 33) {
       HEAP32[$error$i3$i>>2] = $95;
       $$ph$i = $95;
       break;
      }
      $101 = HEAP32[$88>>2]|0;
      $102 = ($101|0)==(0);
      if ($102) {
       HEAP32[$error$i3$i>>2] = 3;
       $$ph$i = 3;
       break;
      }
      $103 = (_FT_Stream_Seek($82,$87)|0);
      HEAP32[$error$i3$i>>2] = $103;
      $104 = ($103|0)==(0);
      if ($104) {
       $105 = HEAP32[$88>>2]|0;
       $106 = (_ft_mem_alloc($84,$105,$error$i3$i)|0);
       $107 = ((($loader$i)) + 84|0);
       HEAP32[$107>>2] = $106;
       $108 = HEAP32[$error$i3$i>>2]|0;
       $109 = ($108|0)==(0);
       if ($109) {
        HEAP32[$88>>2] = 0;
        $110 = (_FT_Stream_ReadUShort($82,$error$i28$i$i)|0);
        $111 = HEAP32[$error$i28$i$i>>2]|0;
        $112 = ($111|0)==(0);
        L113: do {
         if ($112) {
          $113 = $110;
          while(1) {
           $$off$i29$i$i = (($113) + 32767)<<16>>16;
           $switch$i30$i$i = ($$off$i29$i$i&65535)<(2);
           if (!($switch$i30$i$i)) {
            $$ph$i$i = 0;
            label = 43;
            break L113;
           }
           $114 = (_FT_Stream_ReadULongLE($82,$error$i28$i$i)|0);
           $115 = HEAP32[$error$i28$i$i>>2]|0;
           $116 = ($115|0)==(0);
           if (!($116)) {
            $$ph$i$i = $115;
            label = 43;
            break L113;
           }
           HEAP32[$error$i3$i>>2] = 0;
           $117 = ($113<<16>>16)==(-32766);
           if (!($117)) {
            break L113;
           }
           $118 = HEAP32[$107>>2]|0;
           $119 = HEAP32[$88>>2]|0;
           $120 = (($118) + ($119)|0);
           $121 = (_FT_Stream_Read($82,$120,$114)|0);
           HEAP32[$error$i3$i>>2] = $121;
           $122 = ($121|0)==(0);
           if (!($122)) {
            $$ph$i = $121;
            break L38;
           }
           $123 = HEAP32[$88>>2]|0;
           $124 = (($123) + ($114))|0;
           HEAP32[$88>>2] = $124;
           $125 = (_FT_Stream_ReadUShort($82,$error$i28$i$i)|0);
           $126 = HEAP32[$error$i28$i$i>>2]|0;
           $127 = ($126|0)==(0);
           if ($127) {
            $113 = $125;
           } else {
            $$ph$i$i = $126;
            label = 43;
            break;
           }
          }
         } else {
          $$ph$i$i = $111;
          label = 43;
         }
        } while(0);
        if ((label|0) == 43) {
         HEAP32[$error$i3$i>>2] = $$ph$i$i;
        }
        HEAP32[$error$i3$i>>2] = 0;
        $$pre$phi$i$iZ2D = $107;$$pre$phi77$i$iZ2D = $88;
        label = 88;
       } else {
        $$ph$i = $108;
       }
      } else {
       $$ph$i = $103;
      }
     }
    } while(0);
    do {
     if ((label|0) == 88) {
      $236 = ((($16)) + 16|0);
      $237 = HEAP32[$236>>2]|0;
      $238 = HEAP32[$$pre$phi$i$iZ2D>>2]|0;
      $239 = HEAP32[$$pre$phi77$i$iZ2D>>2]|0;
      FUNCTION_TABLE_viii[$237 & 8191]($238,$239,-9871);
      $240 = HEAP32[$$pre$phi77$i$iZ2D>>2]|0;
      $241 = ($240>>>0)<(4);
      if ($241) {
       HEAP32[$error$i3$i>>2] = 3;
       $$ph$i = 3;
       break;
      }
      $242 = HEAP32[$$pre$phi$i$iZ2D>>2]|0;
      HEAP8[$242>>0] = 32;
      $243 = HEAP32[$$pre$phi$i$iZ2D>>2]|0;
      $244 = ((($243)) + 1|0);
      HEAP8[$244>>0] = 32;
      $245 = HEAP32[$$pre$phi$i$iZ2D>>2]|0;
      $246 = ((($245)) + 2|0);
      HEAP8[$246>>0] = 32;
      $247 = HEAP32[$$pre$phi$i$iZ2D>>2]|0;
      $248 = ((($247)) + 3|0);
      HEAP8[$248>>0] = 32;
      $249 = HEAP32[$$pre$phi$i$iZ2D>>2]|0;
      $250 = ((($loader$i)) + 4|0);
      HEAP32[$250>>2] = $249;
      $$cast$i10$i = $249;
      HEAP32[$loader$i>>2] = $$cast$i10$i;
      $251 = HEAP32[$$pre$phi77$i$iZ2D>>2]|0;
      $252 = (($$cast$i10$i) + ($251)|0);
      $253 = ((($loader$i)) + 8|0);
      HEAP32[$253>>2] = $252;
      $$pre75$i$i = HEAP32[$error$i3$i>>2]|0;
      HEAP32[$error$i>>2] = $$pre75$i$i;
      $254 = ($$pre75$i$i|0)==(0);
      if (!($254)) {
       break L35;
      }
      $255 = ((($loader$i)) + 84|0);
      $256 = HEAP32[$255>>2]|0;
      $257 = ((($loader$i)) + 88|0);
      $258 = HEAP32[$257>>2]|0;
      $259 = (_parse_dict($t1face,$loader$i,$256,$258)|0);
      HEAP32[$error$i>>2] = $259;
      $260 = ($259|0)==(0);
      if (!($260)) {
       break L35;
      }
      $261 = ((($t1face)) + 176|0);
      $262 = HEAP8[$261>>0]|0;
      $263 = $262&255;
      $264 = $263 & 254;
      $265 = $264&255;
      HEAP8[$261>>0] = $265;
      $266 = ((($t1face)) + 528|0);
      $267 = HEAP32[$266>>2]|0;
      $268 = ($267|0)==(0|0);
      L130: do {
       if ($268) {
        label = 108;
       } else {
        $269 = ((($267)) + 416|0);
        $270 = HEAP32[$269>>2]|0;
        $271 = ($270|0)==(0);
        if (!($271)) {
         $272 = ((($267)) + 4|0);
         $273 = HEAP32[$272>>2]|0;
         $274 = ($270|0)==($273|0);
         if (!($274)) {
          HEAP32[$269>>2] = 0;
         }
        }
        $275 = HEAP32[$267>>2]|0;
        $276 = ($275|0)==(0);
        if ($276) {
         label = 99;
        } else {
         $277 = ((($267)) + 4|0);
         $278 = HEAP32[$277>>2]|0;
         $279 = ($278|0)==(0);
         if ($279) {
          label = 99;
         } else {
          $284 = $278;$286 = $267;
          label = 101;
         }
        }
        if ((label|0) == 99) {
         _T1_Done_Blend($t1face);
         $$pr12$pr$pre$i = HEAP32[$266>>2]|0;
         $280 = ($$pr12$pr$pre$i|0)==(0|0);
         if ($280) {
          label = 108;
          break;
         }
         $$phi$trans$insert = ((($$pr12$pr$pre$i)) + 4|0);
         $$pre = HEAP32[$$phi$trans$insert>>2]|0;
         $281 = ($$pre|0)==(0);
         if (!($281)) {
          $284 = $$pre;$286 = $$pr12$pr$pre$i;
          label = 101;
         }
        }
        L141: do {
         if ((label|0) == 101) {
          $i$040$i = 0;
          while(1) {
           $285 = (((($286)) + 88|0) + (($i$040$i*12)|0)|0);
           $287 = HEAP8[$285>>0]|0;
           $288 = ($287<<24>>24)==(0);
           $283 = (($i$040$i) + 1)|0;
           if ($288) {
            break;
           }
           $282 = ($283>>>0)<($284>>>0);
           if ($282) {
            $i$040$i = $283;
           } else {
            break L141;
           }
          }
          _T1_Done_Blend($t1face);
          $$pr14$pre$i = HEAP32[$266>>2]|0;
          $289 = ($$pr14$pre$i|0)==(0|0);
          if ($289) {
           label = 108;
           break L130;
          }
         }
        } while(0);
        $290 = HEAP32[$19>>2]|0;
        $291 = ($290|0)==(0);
        if (!($291)) {
         $292 = HEAP32[$27>>2]|0;
         $293 = (_ft_mem_realloc($292,4,0,$290,0,$error$i)|0);
         $294 = ((($t1face)) + 544|0);
         HEAP32[$294>>2] = $293;
         $295 = HEAP32[$error$i>>2]|0;
         $296 = ($295|0)==(0);
         if (!($296)) {
          HEAP32[$19>>2] = 0;
          break L35;
         }
        }
       }
      } while(0);
      if ((label|0) == 108) {
       HEAP32[$19>>2] = 0;
      }
      $297 = ((($loader$i)) + 152|0);
      $298 = HEAP32[$297>>2]|0;
      $299 = ((($t1face)) + 416|0);
      HEAP32[$299>>2] = $298;
      $300 = ((($loader$i)) + 328|0);
      $301 = HEAP32[$300>>2]|0;
      $302 = ($301|0)==(0);
      if (!($302)) {
       $303 = ((($loader$i)) + 316|0);
       HEAP32[$300>>2] = 0;
       $304 = ((($loader$i)) + 312|0);
       $305 = HEAP32[$304>>2]|0;
       $306 = ((($t1face)) + 404|0);
       HEAP32[$306>>2] = $305;
       $307 = HEAP32[$303>>2]|0;
       $308 = ((($t1face)) + 392|0);
       HEAP32[$308>>2] = $307;
       $309 = ((($loader$i)) + 340|0);
       $310 = HEAP32[$309>>2]|0;
       $311 = ((($t1face)) + 408|0);
       HEAP32[$311>>2] = $310;
       $312 = ((($loader$i)) + 344|0);
       $313 = HEAP32[$312>>2]|0;
       $314 = ((($t1face)) + 412|0);
       HEAP32[$314>>2] = $313;
      }
      $315 = ((($t1face)) + 128|0);
      $316 = HEAP32[$315>>2]|0;
      $317 = ((($316)) + 48|0);
      $318 = HEAP32[$317>>2]|0;
      $319 = ($318|0)==(0|0);
      $320 = ((($loader$i)) + 220|0);
      $321 = HEAP32[$320>>2]|0;
      $322 = ($321|0)==(0);
      $or$cond$i = $319 & $322;
      if ($or$cond$i) {
       HEAP32[$error$i>>2] = 3;
      }
      $323 = ((($loader$i)) + 208|0);
      HEAP32[$320>>2] = 0;
      $324 = HEAP32[$323>>2]|0;
      $325 = ((($t1face)) + 396|0);
      HEAP32[$325>>2] = $324;
      $326 = ((($loader$i)) + 232|0);
      $327 = HEAP32[$326>>2]|0;
      $328 = ((($t1face)) + 424|0);
      HEAP32[$328>>2] = $327;
      $329 = ((($loader$i)) + 236|0);
      $330 = HEAP32[$329>>2]|0;
      $331 = ((($t1face)) + 428|0);
      HEAP32[$331>>2] = $330;
      $332 = ((($loader$i)) + 156|0);
      $333 = HEAP32[$332>>2]|0;
      $334 = ((($t1face)) + 400|0);
      HEAP32[$334>>2] = $333;
      $335 = ((($loader$i)) + 180|0);
      $336 = HEAP32[$335>>2]|0;
      $337 = ((($t1face)) + 420|0);
      HEAP32[$337>>2] = $336;
      HEAP32[$332>>2] = 0;
      HEAP32[$335>>2] = 0;
      $338 = ((($t1face)) + 368|0);
      $339 = HEAP32[$338>>2]|0;
      $340 = ($339|0)==(1);
      if (!($340)) {
       break L35;
      }
      $341 = ((($loader$i)) + 116|0);
      $342 = HEAP32[$341>>2]|0;
      $343 = ($342|0)>(0);
      L160: do {
       if ($343) {
        $344 = ((($t1face)) + 384|0);
        $345 = ((($t1face)) + 388|0);
        $346 = ((($loader$i)) + 124|0);
        $$not$i = ($298|0)<(1);
        if ($$not$i) {
         $charcode$033$i$us = 0;
         while(1) {
          $347 = HEAP32[$344>>2]|0;
          $348 = (($347) + ($charcode$033$i$us<<1)|0);
          HEAP16[$348>>1] = 0;
          $349 = HEAP32[$345>>2]|0;
          $350 = (($349) + ($charcode$033$i$us<<2)|0);
          HEAP32[$350>>2] = 594602;
          $351 = (($charcode$033$i$us) + 1)|0;
          $exitcond = ($351|0)==($342|0);
          if ($exitcond) {
           $max_char$0$lcssa$i = 0;$min_char$0$lcssa$i = 0;
           break L160;
          } else {
           $charcode$033$i$us = $351;
          }
         }
        } else {
         $charcode$033$i = 0;$max_char$037$i = 0;$min_char$036$i = 0;
        }
        while(1) {
         $352 = HEAP32[$344>>2]|0;
         $353 = (($352) + ($charcode$033$i<<1)|0);
         HEAP16[$353>>1] = 0;
         $354 = HEAP32[$345>>2]|0;
         $355 = (($354) + ($charcode$033$i<<2)|0);
         HEAP32[$355>>2] = 594602;
         $356 = HEAP32[$346>>2]|0;
         $357 = (($356) + ($charcode$033$i<<2)|0);
         $358 = HEAP32[$357>>2]|0;
         $359 = ($358|0)==(0|0);
         L167: do {
          if ($359) {
           $max_char$1$i = $max_char$037$i;$min_char$2$i = $min_char$036$i;
          } else {
           $360 = HEAP32[$337>>2]|0;
           $idx$032$i = 0;
           while(1) {
            $361 = (($360) + ($idx$032$i<<2)|0);
            $362 = HEAP32[$361>>2]|0;
            $363 = (_strcmp($358,$362)|0);
            $364 = ($363|0)==(0);
            if ($364) {
             $$lcssa = $362;$idx$032$i$lcssa = $idx$032$i;
             break;
            }
            $375 = (($idx$032$i) + 1)|0;
            $376 = ($375|0)<($298|0);
            if ($376) {
             $idx$032$i = $375;
            } else {
             $max_char$1$i = $max_char$037$i;$min_char$2$i = $min_char$036$i;
             break L167;
            }
           }
           $365 = $idx$032$i$lcssa&65535;
           $366 = HEAP32[$344>>2]|0;
           $367 = (($366) + ($charcode$033$i<<1)|0);
           HEAP16[$367>>1] = $365;
           $368 = HEAP32[$345>>2]|0;
           $369 = (($368) + ($charcode$033$i<<2)|0);
           HEAP32[$369>>2] = $$lcssa;
           $370 = (_strcmp(594602,$$lcssa)|0);
           $371 = ($370|0)==(0);
           if ($371) {
            $max_char$1$i = $max_char$037$i;$min_char$2$i = $min_char$036$i;
           } else {
            $372 = ($charcode$033$i|0)<($min_char$036$i|0);
            $charcode$0$min_char$0$i = $372 ? $charcode$033$i : $min_char$036$i;
            $373 = ($charcode$033$i|0)<($max_char$037$i|0);
            $374 = (($charcode$033$i) + 1)|0;
            $max_char$0$$i = $373 ? $max_char$037$i : $374;
            $max_char$1$i = $max_char$0$$i;$min_char$2$i = $charcode$0$min_char$0$i;
           }
          }
         } while(0);
         $377 = (($charcode$033$i) + 1)|0;
         $exitcond51 = ($377|0)==($342|0);
         if ($exitcond51) {
          $max_char$0$lcssa$i = $max_char$1$i;$min_char$0$lcssa$i = $min_char$2$i;
          break;
         } else {
          $charcode$033$i = $377;$max_char$037$i = $max_char$1$i;$min_char$036$i = $min_char$2$i;
         }
        }
       } else {
        $max_char$0$lcssa$i = 0;$min_char$0$lcssa$i = 0;
       }
      } while(0);
      $378 = ((($t1face)) + 376|0);
      HEAP32[$378>>2] = $min_char$0$lcssa$i;
      $379 = ((($t1face)) + 380|0);
      HEAP32[$379>>2] = $max_char$0$lcssa$i;
      $380 = ((($loader$i)) + 96|0);
      $381 = HEAP32[$380>>2]|0;
      $382 = ((($t1face)) + 372|0);
      HEAP32[$382>>2] = $381;
      break L35;
     }
    } while(0);
    HEAP32[$error$i>>2] = $$ph$i;
   }
  }
 } while(0);
 $383 = ((($loader$i)) + 148|0);
 $384 = HEAP32[$383>>2]|0;
 $385 = ($384|0)==(0|0);
 if (!($385)) {
  $386 = ((($loader$i)) + 100|0);
  FUNCTION_TABLE_vi[$384 & 8191]($386);
 }
 $387 = ((($loader$i)) + 256|0);
 $388 = HEAP32[$387>>2]|0;
 $389 = ($388|0)==(0|0);
 if (!($389)) {
  $390 = ((($loader$i)) + 208|0);
  FUNCTION_TABLE_vi[$388 & 8191]($390);
 }
 $391 = ((($loader$i)) + 204|0);
 $392 = HEAP32[$391>>2]|0;
 $393 = ($392|0)==(0|0);
 if (!($393)) {
  $394 = ((($loader$i)) + 156|0);
  FUNCTION_TABLE_vi[$392 & 8191]($394);
 }
 $395 = ((($loader$i)) + 308|0);
 $396 = HEAP32[$395>>2]|0;
 $397 = ($396|0)==(0|0);
 if (!($397)) {
  $398 = ((($loader$i)) + 260|0);
  FUNCTION_TABLE_vi[$396 & 8191]($398);
 }
 $399 = ((($loader$i)) + 364|0);
 $400 = HEAP32[$399>>2]|0;
 $401 = ($400|0)==(0|0);
 if (!($401)) {
  $402 = ((($loader$i)) + 316|0);
  FUNCTION_TABLE_vi[$400 & 8191]($402);
 }
 $403 = ((($loader$i)) + 16|0);
 $404 = HEAP32[$403>>2]|0;
 $405 = ((($loader$i)) + 84|0);
 $406 = HEAP32[$405>>2]|0;
 _ft_mem_free($404,$406);
 HEAP32[$405>>2] = 0;
 $407 = HEAP8[$33>>0]|0;
 $408 = ($407<<24>>24)==(0);
 if ($408) {
  $409 = HEAP32[$31>>2]|0;
  _ft_mem_free($404,$409);
  HEAP32[$31>>2] = 0;
 }
 $410 = ((($loader$i)) + 24|0);
 $411 = HEAP32[$410>>2]|0;
 FUNCTION_TABLE_vi[$411 & 8191]($loader$i);
 $412 = HEAP32[$error$i>>2]|0;
 $413 = ($412|0)!=(0);
 $414 = ($face_index|0)<(0);
 $or$cond = $414 | $413;
 if ($or$cond) {
  $$02 = $412;
  STACKTOP = sp;return ($$02|0);
 }
 $415 = ($face_index|0)>(0);
 if ($415) {
  $$02 = 6;
  STACKTOP = sp;return ($$02|0);
 }
 $416 = ((($t1face)) + 416|0);
 $417 = HEAP32[$416>>2]|0;
 $418 = ((($t1face)) + 16|0);
 HEAP32[$418>>2] = $417;
 $419 = ((($t1face)) + 4|0);
 HEAP32[$419>>2] = 0;
 $420 = ((($t1face)) + 8|0);
 $421 = HEAP32[$420>>2]|0;
 $422 = $421 | 2577;
 HEAP32[$420>>2] = $422;
 $423 = ((($t1face)) + 156|0);
 $424 = HEAP8[$423>>0]|0;
 $425 = ($424<<24>>24)==(0);
 if ($425) {
  $431 = $422;
 } else {
  $426 = $421 | 2581;
  HEAP32[$420>>2] = $426;
  $431 = $426;
 }
 $427 = ((($t1face)) + 528|0);
 $428 = HEAP32[$427>>2]|0;
 $429 = ($428|0)==(0|0);
 if (!($429)) {
  $430 = $431 | 256;
  HEAP32[$420>>2] = $430;
 }
 $432 = ((($t1face)) + 144|0);
 $433 = HEAP32[$432>>2]|0;
 $434 = ((($t1face)) + 20|0);
 HEAP32[$434>>2] = $433;
 $435 = ((($t1face)) + 24|0);
 HEAP32[$435>>2] = 0;
 $436 = ($433|0)==(0);
 L208: do {
  if ($436) {
   $451 = ((($t1face)) + 364|0);
   $452 = HEAP32[$451>>2]|0;
   $453 = ($452|0)==(0|0);
   if ($453) {
    label = 158;
   } else {
    HEAP32[$434>>2] = $452;
    label = 158;
   }
  } else {
   $437 = ((($t1face)) + 140|0);
   $438 = HEAP32[$437>>2]|0;
   $439 = ($438|0)==(0|0);
   if ($439) {
    label = 158;
   } else {
    $440 = HEAP8[$438>>0]|0;
    $441 = ($440<<24>>24)==(0);
    L213: do {
     if (!($441)) {
      $$cast = $433;
      $443 = $440;$family$0$ph31 = $$cast;$full$0$ph30 = $438;
      L215: while(1) {
       $442 = ($443<<24>>24)==(0);
       $family$029 = $family$0$ph31;
       L217: while(1) {
        $444 = HEAP8[$family$029>>0]|0;
        $445 = ($443<<24>>24)==($444<<24>>24);
        if ($445) {
         $family$029$lcssa = $family$029;
         label = 149;
         break;
        }
        switch ($443<<24>>24) {
        case 45: case 32:  {
         $family$0$ph$be = $family$029;
         break L217;
         break;
        }
        default: {
        }
        }
        switch ($444<<24>>24) {
        case 0:  {
         $full$0$ph30$lcssa114 = $full$0$ph30;
         break L215;
         break;
        }
        case 45: case 32:  {
         break;
        }
        default: {
         label = 158;
         break L208;
        }
        }
        $450 = ((($family$029)) + 1|0);
        if ($442) {
         break L213;
        } else {
         $family$029 = $450;
        }
       }
       if ((label|0) == 149) {
        label = 0;
        $446 = ((($family$029$lcssa)) + 1|0);
        $family$0$ph$be = $446;
       }
       $447 = ((($full$0$ph30)) + 1|0);
       $448 = HEAP8[$447>>0]|0;
       $449 = ($448<<24>>24)==(0);
       if ($449) {
        break L213;
       } else {
        $443 = $448;$family$0$ph31 = $family$0$ph$be;$full$0$ph30 = $447;
       }
      }
      HEAP32[$435>>2] = $full$0$ph30$lcssa114;
      $454 = ($full$0$ph30$lcssa114|0)==(0|0);
      if ($454) {
       label = 158;
       break L208;
      } else {
       break L208;
      }
     }
    } while(0);
    HEAP32[$435>>2] = 594711;
   }
  }
 } while(0);
 do {
  if ((label|0) == 158) {
   $455 = ((($t1face)) + 148|0);
   $456 = HEAP32[$455>>2]|0;
   $457 = ($456|0)==(0|0);
   if ($457) {
    HEAP32[$435>>2] = 594711;
    break;
   } else {
    HEAP32[$435>>2] = $456;
    break;
   }
  }
 } while(0);
 $458 = ((($t1face)) + 12|0);
 $459 = ((($t1face)) + 152|0);
 $460 = HEAP32[$459>>2]|0;
 $not$ = ($460|0)!=(0);
 $$ = $not$&1;
 HEAP32[$458>>2] = $$;
 $461 = ((($t1face)) + 148|0);
 $462 = HEAP32[$461>>2]|0;
 $463 = ($462|0)==(0|0);
 do {
  if (!($463)) {
   $464 = (_strcmp($462,594731)|0);
   $465 = ($464|0)==(0);
   if (!($465)) {
    $466 = (_strcmp($462,594283)|0);
    $467 = ($466|0)==(0);
    if (!($467)) {
     break;
    }
   }
   $468 = $$ | 2;
   HEAP32[$458>>2] = $468;
  }
 } while(0);
 $469 = ((($t1face)) + 28|0);
 HEAP32[$469>>2] = 0;
 $470 = ((($t1face)) + 32|0);
 HEAP32[$470>>2] = 0;
 $471 = ((($t1face)) + 460|0);
 $472 = HEAP32[$471>>2]|0;
 $473 = $472 >> 16;
 $474 = ((($t1face)) + 52|0);
 HEAP32[$474>>2] = $473;
 $475 = ((($471)) + 4|0);
 $476 = HEAP32[$475>>2]|0;
 $477 = $476 >> 16;
 $478 = ((($t1face)) + 56|0);
 HEAP32[$478>>2] = $477;
 $479 = ((($471)) + 8|0);
 $480 = HEAP32[$479>>2]|0;
 $481 = (($480) + 65535)|0;
 $482 = $481 >> 16;
 $483 = ((($t1face)) + 60|0);
 HEAP32[$483>>2] = $482;
 $484 = ((($471)) + 12|0);
 $485 = HEAP32[$484>>2]|0;
 $486 = (($485) + 65535)|0;
 $487 = $486 >> 16;
 $488 = ((($t1face)) + 64|0);
 HEAP32[$488>>2] = $487;
 $489 = ((($t1face)) + 68|0);
 $490 = HEAP16[$489>>1]|0;
 $491 = ($490<<16>>16)==(0);
 if ($491) {
  HEAP16[$489>>1] = 1000;
  $497 = 1000;
 } else {
  $497 = $490;
 }
 $492 = $487&65535;
 $493 = ((($t1face)) + 70|0);
 HEAP16[$493>>1] = $492;
 $494 = $477&65535;
 $495 = ((($t1face)) + 72|0);
 HEAP16[$495>>1] = $494;
 $496 = $497&65535;
 $498 = ($496*12)|0;
 $499 = (($498>>>0) / 10)&-1;
 $500 = ((($t1face)) + 74|0);
 $sext = $499 << 16;
 $501 = $sext >> 16;
 $502 = (($487) - ($477))|0;
 $503 = ($501|0)<($502|0);
 $$71 = $503 ? $502 : $499;
 $storemerge = $$71&65535;
 HEAP16[$500>>1] = $storemerge;
 $504 = $482&65535;
 $505 = ((($t1face)) + 76|0);
 HEAP16[$505>>1] = $504;
 $506 = HEAP32[$9>>2]|0;
 $507 = ((($506)) + 12|0);
 $508 = HEAP32[$507>>2]|0;
 $509 = HEAP32[$508>>2]|0;
 $510 = ((($t1face)) + 420|0);
 $511 = HEAP32[$510>>2]|0;
 $512 = (FUNCTION_TABLE_iiiiiiiiii[$509 & 8191]($decoder$i,$t1face,0,0,$511,$428,0,0,4436)|0);
 $513 = ($512|0)==(0);
 if ($513) {
  $514 = ((($decoder$i)) + 70|0);
  HEAP8[$514>>0] = 1;
  $515 = ((($decoder$i)) + 68|0);
  HEAP8[$515>>0] = 0;
  $516 = ((($t1face)) + 404|0);
  $517 = HEAP32[$516>>2]|0;
  $518 = ((($decoder$i)) + 1364|0);
  HEAP32[$518>>2] = $517;
  $519 = ((($t1face)) + 408|0);
  $520 = HEAP32[$519>>2]|0;
  $521 = ((($decoder$i)) + 1368|0);
  HEAP32[$521>>2] = $520;
  $522 = ((($t1face)) + 412|0);
  $523 = HEAP32[$522>>2]|0;
  $524 = ((($decoder$i)) + 1372|0);
  HEAP32[$524>>2] = $523;
  $525 = ((($t1face)) + 544|0);
  $526 = HEAP32[$525>>2]|0;
  $527 = ((($decoder$i)) + 1488|0);
  HEAP32[$527>>2] = $526;
  $528 = HEAP32[$19>>2]|0;
  $529 = ((($decoder$i)) + 1492|0);
  HEAP32[$529>>2] = $528;
  $530 = HEAP32[$416>>2]|0;
  $531 = ($530|0)>(0);
  if ($531) {
   $532 = ((($decoder$i)) + 4|0);
   $533 = ((($decoder$i)) + 40|0);
   $glyph_index$01$i = 0;$max_advance$0 = 0;
   while(1) {
    $534 = (_T1_Parse_Glyph_And_Get_Char_String($decoder$i,$glyph_index$01$i,$glyph_data$i$i)|0);
    $535 = ($534|0)==(0);
    do {
     if ($535) {
      $536 = HEAP32[$532>>2]|0;
      $537 = ((($536)) + 128|0);
      $538 = HEAP32[$537>>2]|0;
      $539 = ((($538)) + 48|0);
      $540 = HEAP32[$539>>2]|0;
      $541 = ($540|0)==(0|0);
      if ($541) {
       break;
      }
      $542 = HEAP32[$540>>2]|0;
      $543 = ((($542)) + 4|0);
      $544 = HEAP32[$543>>2]|0;
      $545 = ((($540)) + 4|0);
      $546 = HEAP32[$545>>2]|0;
      FUNCTION_TABLE_vii[$544 & 8191]($546,$glyph_data$i$i);
     }
    } while(0);
    $547 = ($glyph_index$01$i|0)==(0);
    $$pre$i7 = HEAP32[$533>>2]|0;
    $548 = ($$pre$i7|0)>($max_advance$0|0);
    $or$cond12 = $547 | $548;
    $max_advance$1 = $or$cond12 ? $$pre$i7 : $max_advance$0;
    $549 = (($glyph_index$01$i) + 1)|0;
    $550 = HEAP32[$416>>2]|0;
    $551 = ($549|0)<($550|0);
    if ($551) {
     $glyph_index$01$i = $549;$max_advance$0 = $max_advance$1;
    } else {
     $max_advance$2 = $max_advance$1;
     break;
    }
   }
  } else {
   $max_advance$2 = 0;
  }
  $552 = HEAP32[$507>>2]|0;
  $553 = ((($552)) + 4|0);
  $554 = HEAP32[$553>>2]|0;
  FUNCTION_TABLE_vi[$554 & 8191]($decoder$i);
  $555 = (_FT_RoundFix($max_advance$2)|0);
  $556 = $555 >>> 16;
  $557 = $556&65535;
  HEAP16[$505>>1] = $557;
 } else {
 }
 $558 = HEAP16[$500>>1]|0;
 $559 = ((($t1face)) + 78|0);
 HEAP16[$559>>1] = $558;
 $560 = ((($0)) + 26|0);
 $561 = HEAP16[$560>>1]|0;
 $562 = ((($t1face)) + 80|0);
 HEAP16[$562>>1] = $561;
 $563 = ((($t1face)) + 160|0);
 $564 = HEAP16[$563>>1]|0;
 $565 = ((($t1face)) + 82|0);
 HEAP16[$565>>1] = $564;
 $566 = ($3|0)==(0|0);
 if ($566) {
  $$02 = 0;
  STACKTOP = sp;return ($$02|0);
 }
 $567 = ((($8)) + 20|0);
 $568 = HEAP32[$567>>2]|0;
 HEAP32[$charmap>>2] = $t1face;
 $569 = ((($charmap)) + 8|0);
 HEAP16[$569>>1] = 3;
 $570 = ((($charmap)) + 10|0);
 HEAP16[$570>>1] = 1;
 $571 = ((($charmap)) + 4|0);
 HEAP32[$571>>2] = 1970170211;
 $572 = ((($568)) + 12|0);
 $573 = HEAP32[$572>>2]|0;
 $574 = (_FT_CMap_New($573,0,$charmap,0)|0);
 $575 = ($574|0)==(0);
 $576 = $574 & 255;
 $577 = ($576|0)==(163);
 $or$cond5 = $575 | $577;
 if (!($or$cond5)) {
  $error$2 = $574;
  STACKTOP = sp;return ($error$2|0);
 }
 HEAP16[$569>>1] = 7;
 $578 = ((($t1face)) + 368|0);
 $579 = HEAP32[$578>>2]|0;
 switch ($579|0) {
 case 2:  {
  HEAP32[$571>>2] = 1094995778;
  HEAP16[$570>>1] = 0;
  $clazz$0$in = $568;
  break;
 }
 case 4:  {
  HEAP32[$571>>2] = 1094992453;
  HEAP16[$570>>1] = 1;
  $580 = ((($568)) + 4|0);
  $clazz$0$in = $580;
  break;
 }
 case 1:  {
  HEAP32[$571>>2] = 1094992451;
  HEAP16[$570>>1] = 2;
  $581 = ((($568)) + 8|0);
  $clazz$0$in = $581;
  break;
 }
 case 3:  {
  HEAP32[$571>>2] = 1818326065;
  HEAP16[$570>>1] = 3;
  $clazz$0$in = $572;
  break;
 }
 default: {
  $error$2 = 0;
  STACKTOP = sp;return ($error$2|0);
 }
 }
 $clazz$0 = HEAP32[$clazz$0$in>>2]|0;
 $582 = ($clazz$0|0)==(0|0);
 if ($582) {
  $error$2 = 0;
  STACKTOP = sp;return ($error$2|0);
 }
 $583 = (_FT_CMap_New($clazz$0,0,$charmap,0)|0);
 $error$2 = $583;
 STACKTOP = sp;return ($error$2|0);
}
function _check_type1_format($stream,$header_string,$header_length) {
 $stream = $stream|0;
 $header_string = $header_string|0;
 $header_length = $header_length|0;
 var $$ = 0, $$off$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$1 = 0;
 var $error$i = 0, $switch$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp;
 $0 = (_FT_Stream_Seek($stream,0)|0);
 $1 = ($0|0)==(0);
 L1: do {
  if ($1) {
   $2 = (_FT_Stream_ReadUShort($stream,$error$i)|0);
   $3 = HEAP32[$error$i>>2]|0;
   $4 = ($3|0)==(0);
   do {
    if ($4) {
     $$off$i = (($2) + 32767)<<16>>16;
     $switch$i = ($$off$i&65535)<(2);
     if ($switch$i) {
      (_FT_Stream_ReadULongLE($stream,$error$i)|0);
      $5 = HEAP32[$error$i>>2]|0;
      $6 = ($5|0)==(0);
      if (!($6)) {
       $16 = $5;
       break;
      }
      $7 = ($2<<16>>16)==(-32767);
      if (!($7)) {
       label = 8;
      }
     } else {
      label = 8;
     }
     if ((label|0) == 8) {
      $8 = (_FT_Stream_Seek($stream,0)|0);
      $9 = ($8|0)==(0);
      if (!($9)) {
       $error$1 = $8;
       break L1;
      }
     }
     $10 = (_FT_Stream_EnterFrame($stream,$header_length)|0);
     $11 = ($10|0)==(0);
     if (!($11)) {
      $error$1 = $10;
      break L1;
     }
     $12 = ((($stream)) + 32|0);
     $13 = HEAP32[$12>>2]|0;
     $14 = (_memcmp($13,$header_string,$header_length)|0);
     $15 = ($14|0)==(0);
     $$ = $15 ? 0 : 2;
     _FT_Stream_ExitFrame($stream);
     $error$1 = $$;
     break L1;
    } else {
     $16 = $3;
    }
   } while(0);
   $error$1 = $16;
  } else {
   $error$1 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($error$1|0);
}
function _parse_dict($face,$loader,$base,$size) {
 $face = $face|0;
 $loader = $loader|0;
 $base = $base|0;
 $size = $size|0;
 var $$$i = 0, $$10 = 0, $$lcssa = 0, $$lcssa81 = 0, $$lcssa82 = 0, $$off$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $blend$0$i = 0, $dummy_object$i = 0, $error$0$i = 0;
 var $have_integer$051 = 0, $have_integer$4 = 0, $keyword$050 = 0, $keyword$050$lcssa = 0, $max_objects$0$i = 0, $objects$0$i = 0, $or$cond = 0, $or$cond13 = 0, $or$cond14 = 0, $or$cond16 = 0, $or$cond17 = 0, $or$cond22 = 0, $or$cond23 = 0, $start_binary$053 = 0, $start_binary$1 = 0, $switch$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $dummy_object$i = sp;
 HEAP32[$loader>>2] = $base;
 $0 = (($base) + ($size)|0);
 $1 = ((($loader)) + 8|0);
 HEAP32[$1>>2] = $0;
 $2 = ((($loader)) + 12|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($loader)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 FUNCTION_TABLE_vi[$4 & 8191]($loader);
 $5 = HEAP32[$loader>>2]|0;
 $6 = ($5>>>0)<($0>>>0);
 L1: do {
  if ($6) {
   $7 = ((($loader)) + 372|0);
   $8 = ((($loader)) + 32|0);
   $9 = ((($face)) + 128|0);
   $10 = ((($loader)) + 36|0);
   $11 = ((($face)) + 528|0);
   $12 = ((($face)) + 132|0);
   $13 = ((($loader)) + 68|0);
   $14 = ((($loader)) + 64|0);
   $15 = ((($face)) + 132|0);
   $16 = ((($face)) + 164|0);
   $17 = ((($face)) + 168|0);
   $18 = ((($face)) + 460|0);
   $20 = $5;$have_integer$051 = 0;$start_binary$053 = 0;
   L3: while(1) {
    $19 = HEAP8[$20>>0]|0;
    L5: do {
     switch ($19<<24>>24) {
     case 101:  {
      $21 = ((($20)) + 6|0);
      $22 = ($21|0)==($0|0);
      if (!($22)) {
       $23 = ($21>>>0)<($0>>>0);
       if (!($23)) {
        label = 19;
        break L5;
       }
       $24 = ((($20)) + 5|0);
       $25 = HEAP8[$24>>0]|0;
       switch ($25<<24>>24) {
       case 37: case 125: case 123: case 93: case 91: case 62: case 60: case 41: case 40: case 47: case 0: case 12: case 9: case 10: case 13: case 32:  {
        break;
       }
       default: {
        label = 19;
        break L5;
       }
       }
      }
      $26 = (_strncmp($20,593953,5)|0);
      $27 = ($26|0)==(0);
      if ($27) {
       break L1;
      } else {
       label = 19;
      }
      break;
     }
     case 99:  {
      $28 = ((($20)) + 10|0);
      $29 = ($28|0)==($0|0);
      if (!($29)) {
       $30 = ($28>>>0)<($0>>>0);
       if (!($30)) {
        label = 19;
        break L5;
       }
       $31 = ((($20)) + 9|0);
       $32 = HEAP8[$31>>0]|0;
       switch ($32<<24>>24) {
       case 37: case 125: case 123: case 93: case 91: case 62: case 60: case 41: case 40: case 47: case 0: case 12: case 9: case 10: case 13: case 32:  {
        break;
       }
       default: {
        label = 19;
        break L5;
       }
       }
      }
      $33 = (_strncmp($20,593959,9)|0);
      $34 = ($33|0)==(0);
      if ($34) {
       break L1;
      } else {
       label = 19;
      }
      break;
     }
     case 70:  {
      $35 = ((($20)) + 14|0);
      $36 = ($35|0)==($0|0);
      if (!($36)) {
       $37 = ($35>>>0)<($0>>>0);
       if (!($37)) {
        label = 19;
        break L5;
       }
       $38 = ((($20)) + 13|0);
       $39 = HEAP8[$38>>0]|0;
       switch ($39<<24>>24) {
       case 37: case 125: case 123: case 93: case 91: case 62: case 60: case 41: case 40: case 47: case 0: case 12: case 9: case 10: case 13: case 32:  {
        break;
       }
       default: {
        label = 19;
        break L5;
       }
       }
      }
      $40 = (_strncmp($20,594402,13)|0);
      $41 = ($40|0)==(0);
      if ($41) {
       $42 = HEAP32[$7>>2]|0;
       $43 = $42 & 1;
       $44 = ($43|0)==(0);
       if (!($44)) {
        $45 = $42 | 2;
        HEAP32[$7>>2] = $45;
       }
       $46 = ((($20)) + 13|0);
       HEAP32[$loader>>2] = $46;
       $have_integer$4 = $have_integer$051;$start_binary$1 = $start_binary$053;
      } else {
       label = 19;
      }
      break;
     }
     default: {
      label = 19;
     }
     }
    } while(0);
    L25: do {
     if ((label|0) == 19) {
      label = 0;
      $47 = $19&255;
      $48 = (($47) + -48)|0;
      $49 = ($48>>>0)<(10);
      if ($49) {
       $50 = HEAP32[$8>>2]|0;
       FUNCTION_TABLE_vi[$50 & 8191]($loader);
       $51 = HEAP32[$2>>2]|0;
       $52 = ($51|0)==(0);
       if ($52) {
        $have_integer$4 = 1;$start_binary$1 = $20;
        break;
       } else {
        break L1;
       }
      }
      $53 = ($19<<24>>24)==(82);
      $54 = ((($20)) + 6|0);
      $55 = ($54>>>0)<($0>>>0);
      $or$cond13 = $55 & $53;
      if ($or$cond13) {
       $56 = ((($20)) + 1|0);
       $57 = HEAP8[$56>>0]|0;
       $58 = ($57<<24>>24)!=(68);
       $59 = ($have_integer$051<<24>>24)==(0);
       $or$cond14 = $59 | $58;
       if (!($or$cond14)) {
        HEAP32[$loader>>2] = $start_binary$053;
        $60 = HEAP32[$9>>2]|0;
        $61 = ((($60)) + 48|0);
        $62 = HEAP32[$61>>2]|0;
        $63 = HEAP32[$1>>2]|0;
        $64 = HEAP32[$3>>2]|0;
        FUNCTION_TABLE_vi[$64 & 8191]($loader);
        $65 = HEAP32[$loader>>2]|0;
        $66 = ($65>>>0)<($63>>>0);
        if (!($66)) {
         $$lcssa81 = $62;
         label = 27;
         break L3;
        }
        $67 = HEAP8[$65>>0]|0;
        $68 = $67&255;
        $69 = (($68) + -48)|0;
        $70 = ($69>>>0)<(10);
        if (!($70)) {
         $$lcssa81 = $62;
         label = 27;
         break L3;
        }
        $71 = HEAP32[$10>>2]|0;
        $72 = (FUNCTION_TABLE_ii[$71 & 8191]($loader)|0);
        $73 = HEAP32[$8>>2]|0;
        FUNCTION_TABLE_vi[$73 & 8191]($loader);
        $74 = HEAP32[$loader>>2]|0;
        $75 = ($72|0)>(-1);
        if (!($75)) {
         $$lcssa81 = $62;
         label = 27;
         break L3;
        }
        $76 = ((($74)) + 1|0);
        $77 = $63;
        $78 = $76;
        $79 = (($77) - ($78))|0;
        $80 = ($72|0)<($79|0);
        if (!($80)) {
         $$lcssa81 = $62;
         label = 27;
         break L3;
        }
        $82 = (($72) + 1)|0;
        $83 = (($74) + ($82)|0);
        HEAP32[$loader>>2] = $83;
        $84 = HEAP32[$2>>2]|0;
        $85 = ($84|0)==(0);
        if ($85) {
         $have_integer$4 = 0;$start_binary$1 = $start_binary$053;
         break;
        } else {
         $$10 = 3;
         label = 77;
         break L3;
        }
       }
      }
      $86 = ($19<<24>>24)==(45);
      $or$cond16 = $55 & $86;
      if ($or$cond16) {
       $87 = ((($20)) + 1|0);
       $88 = HEAP8[$87>>0]|0;
       $89 = ($88<<24>>24)!=(124);
       $90 = ($have_integer$051<<24>>24)==(0);
       $or$cond17 = $90 | $89;
       if (!($or$cond17)) {
        HEAP32[$loader>>2] = $start_binary$053;
        $91 = HEAP32[$9>>2]|0;
        $92 = ((($91)) + 48|0);
        $93 = HEAP32[$92>>2]|0;
        $94 = HEAP32[$1>>2]|0;
        $95 = HEAP32[$3>>2]|0;
        FUNCTION_TABLE_vi[$95 & 8191]($loader);
        $96 = HEAP32[$loader>>2]|0;
        $97 = ($96>>>0)<($94>>>0);
        if (!($97)) {
         $$lcssa82 = $93;
         label = 36;
         break L3;
        }
        $98 = HEAP8[$96>>0]|0;
        $99 = $98&255;
        $100 = (($99) + -48)|0;
        $101 = ($100>>>0)<(10);
        if (!($101)) {
         $$lcssa82 = $93;
         label = 36;
         break L3;
        }
        $102 = HEAP32[$10>>2]|0;
        $103 = (FUNCTION_TABLE_ii[$102 & 8191]($loader)|0);
        $104 = HEAP32[$8>>2]|0;
        FUNCTION_TABLE_vi[$104 & 8191]($loader);
        $105 = HEAP32[$loader>>2]|0;
        $106 = ($103|0)>(-1);
        if (!($106)) {
         $$lcssa82 = $93;
         label = 36;
         break L3;
        }
        $107 = ((($105)) + 1|0);
        $108 = $94;
        $109 = $107;
        $110 = (($108) - ($109))|0;
        $111 = ($103|0)<($110|0);
        if (!($111)) {
         $$lcssa82 = $93;
         label = 36;
         break L3;
        }
        $113 = (($103) + 1)|0;
        $114 = (($105) + ($113)|0);
        HEAP32[$loader>>2] = $114;
        $115 = HEAP32[$2>>2]|0;
        $116 = ($115|0)==(0);
        if ($116) {
         $have_integer$4 = 0;$start_binary$1 = $start_binary$053;
         break;
        } else {
         $$10 = 3;
         label = 77;
         break L3;
        }
       }
      }
      $117 = ($19<<24>>24)==(47);
      $118 = ((($20)) + 2|0);
      $119 = ($118>>>0)<($0>>>0);
      $or$cond22 = $119 & $117;
      if (!($or$cond22)) {
       $184 = HEAP32[$8>>2]|0;
       FUNCTION_TABLE_vi[$184 & 8191]($loader);
       $185 = HEAP32[$2>>2]|0;
       $186 = ($185|0)==(0);
       if ($186) {
        $have_integer$4 = 0;$start_binary$1 = $start_binary$053;
        break;
       } else {
        break L1;
       }
      }
      $120 = ((($20)) + 1|0);
      HEAP32[$loader>>2] = $120;
      $121 = HEAP32[$8>>2]|0;
      FUNCTION_TABLE_vi[$121 & 8191]($loader);
      $122 = HEAP32[$2>>2]|0;
      $123 = ($122|0)==(0);
      if (!($123)) {
       break L1;
      }
      $124 = HEAP32[$loader>>2]|0;
      $125 = $124;
      $126 = $120;
      $127 = (($125) - ($126))|0;
      $128 = ($124|0)!=($120|0);
      $129 = ($127>>>0)<(22);
      $or$cond = $128 & $129;
      $130 = ($124>>>0)<($0>>>0);
      $or$cond23 = $130 & $or$cond;
      if ($or$cond23) {
       $131 = HEAP8[$120>>0]|0;
       $133 = 594422;$keyword$050 = 56460;
       while(1) {
        $132 = HEAP8[$133>>0]|0;
        $134 = ($131<<24>>24)==($132<<24>>24);
        if ($134) {
         $135 = (_strlen($133)|0);
         $136 = ($127|0)==($135|0);
         if ($136) {
          $137 = (_memcmp($120,$133,$127)|0);
          $138 = ($137|0)==(0);
          if ($138) {
           $$lcssa = $133;$keyword$050$lcssa = $keyword$050;
           break;
          }
         }
        }
        $181 = ((($keyword$050)) + 36|0);
        $182 = HEAP32[$181>>2]|0;
        $183 = ($182|0)==(0|0);
        if ($183) {
         $have_integer$4 = 0;$start_binary$1 = $start_binary$053;
         break L25;
        } else {
         $133 = $182;$keyword$050 = $181;
        }
       }
       $139 = HEAP32[$7>>2]|0;
       $140 = $139 & 1;
       $141 = (($140) + 1)|0;
       $142 = ((($keyword$050$lcssa)) + 32|0);
       $143 = HEAP32[$142>>2]|0;
       $144 = $141 & $143;
       $145 = ($144|0)==(0);
       if ($145) {
        $have_integer$4 = 0;$start_binary$1 = $start_binary$053;
       } else {
        $146 = $139 & 2;
        $147 = ($146|0)==(0);
        if (!($147)) {
         $148 = (_strcmp($$lcssa,594660)|0);
         $149 = ($148|0)==(0);
         if (!($149)) {
          $have_integer$4 = 0;$start_binary$1 = $start_binary$053;
          break;
         }
        }
        $150 = HEAP32[$11>>2]|0;
        $151 = ($150|0)==(0|0);
        $152 = $150;
        if ($151) {
         $blend$0$i = 0;
        } else {
         $153 = HEAP32[$150>>2]|0;
         $154 = ($153|0)==(0);
         $$$i = $154 ? 0 : $150;
         $blend$0$i = $$$i;
        }
        $155 = ((($keyword$050$lcssa)) + 8|0);
        $156 = HEAP32[$155>>2]|0;
        $157 = ($156|0)==(11);
        do {
         if ($157) {
          $158 = ((($keyword$050$lcssa)) + 12|0);
          $159 = HEAP32[$158>>2]|0;
          FUNCTION_TABLE_vii[$159 & 8191]($face,$loader);
          $160 = HEAP32[$2>>2]|0;
          $error$0$i = $160;
         } else {
          $161 = ((($keyword$050$lcssa)) + 4|0);
          $162 = HEAP32[$161>>2]|0;
          switch ($162|0) {
          case 3:  {
           HEAP32[$dummy_object$i>>2] = $15;
           $163 = ($blend$0$i|0)==(0|0);
           if ($163) {
            $max_objects$0$i = 0;$objects$0$i = $dummy_object$i;
           } else {
            $164 = ((($blend$0$i)) + 144|0);
            $165 = HEAP32[$blend$0$i>>2]|0;
            $max_objects$0$i = $165;$objects$0$i = $164;
           }
           break;
          }
          case 2:  {
           HEAP32[$dummy_object$i>>2] = $16;
           $max_objects$0$i = 0;$objects$0$i = $dummy_object$i;
           break;
          }
          case 4:  {
           HEAP32[$dummy_object$i>>2] = $17;
           $166 = ($blend$0$i|0)==(0|0);
           if ($166) {
            $max_objects$0$i = 0;$objects$0$i = $dummy_object$i;
           } else {
            $167 = ((($blend$0$i)) + 212|0);
            $168 = HEAP32[$blend$0$i>>2]|0;
            $max_objects$0$i = $168;$objects$0$i = $167;
           }
           break;
          }
          case 5:  {
           HEAP32[$dummy_object$i>>2] = $18;
           $169 = ($blend$0$i|0)==(0|0);
           if ($169) {
            $max_objects$0$i = 0;$objects$0$i = $dummy_object$i;
           } else {
            $170 = ((($blend$0$i)) + 284|0);
            $171 = HEAP32[$blend$0$i>>2]|0;
            $max_objects$0$i = $171;$objects$0$i = $170;
           }
           break;
          }
          case 6:  {
           HEAP32[$dummy_object$i>>2] = $loader;
           $max_objects$0$i = 0;$objects$0$i = $dummy_object$i;
           break;
          }
          case 7:  {
           HEAP32[$dummy_object$i>>2] = $face;
           $max_objects$0$i = 0;$objects$0$i = $dummy_object$i;
           break;
          }
          case 8:  {
           HEAP32[$dummy_object$i>>2] = $152;
           $max_objects$0$i = 0;$objects$0$i = $dummy_object$i;
           break;
          }
          default: {
           HEAP32[$dummy_object$i>>2] = $12;
           $max_objects$0$i = 0;$objects$0$i = $dummy_object$i;
          }
          }
          $172 = HEAP32[$objects$0$i>>2]|0;
          $173 = ($172|0)==(0|0);
          if ($173) {
           HEAP32[$2>>2] = 0;
           $have_integer$4 = 0;$start_binary$1 = $start_binary$053;
           break L25;
          }
          $$off$i = (($156) + -9)|0;
          $switch$i = ($$off$i>>>0)<(2);
          if ($switch$i) {
           $174 = HEAP32[$13>>2]|0;
           $175 = (FUNCTION_TABLE_iiiiii[$174 & 8191]($loader,$keyword$050$lcssa,$objects$0$i,$max_objects$0$i,0)|0);
           $error$0$i = $175;
           break;
          } else {
           $176 = HEAP32[$14>>2]|0;
           $177 = (FUNCTION_TABLE_iiiiii[$176 & 8191]($loader,$keyword$050$lcssa,$objects$0$i,$max_objects$0$i,0)|0);
           $error$0$i = $177;
           break;
          }
         }
        } while(0);
        HEAP32[$2>>2] = $error$0$i;
        $178 = ($error$0$i|0)==(0);
        if ($178) {
         $have_integer$4 = 0;$start_binary$1 = $start_binary$053;
        } else {
         $179 = $error$0$i & 255;
         $180 = ($179|0)==(162);
         if (!($180)) {
          $$10 = $error$0$i;
          label = 77;
          break L3;
         }
         HEAP32[$2>>2] = 0;
         $have_integer$4 = 0;$start_binary$1 = $start_binary$053;
        }
       }
      } else {
       $have_integer$4 = 0;$start_binary$1 = $start_binary$053;
      }
     }
    } while(0);
    $187 = HEAP32[$3>>2]|0;
    FUNCTION_TABLE_vi[$187 & 8191]($loader);
    $188 = HEAP32[$loader>>2]|0;
    $189 = ($188>>>0)<($0>>>0);
    if ($189) {
     $20 = $188;$have_integer$051 = $have_integer$4;$start_binary$053 = $start_binary$1;
    } else {
     break L1;
    }
   }
   if ((label|0) == 27) {
    $81 = ($$lcssa81|0)==(0|0);
    if (!($81)) {
     $$10 = 3;
     STACKTOP = sp;return ($$10|0);
    }
    HEAP32[$2>>2] = 3;
    $$10 = 3;
    STACKTOP = sp;return ($$10|0);
   }
   else if ((label|0) == 36) {
    $112 = ($$lcssa82|0)==(0|0);
    if (!($112)) {
     $$10 = 3;
     STACKTOP = sp;return ($$10|0);
    }
    HEAP32[$2>>2] = 3;
    $$10 = 3;
    STACKTOP = sp;return ($$10|0);
   }
   else if ((label|0) == 77) {
    STACKTOP = sp;return ($$10|0);
   }
  }
 } while(0);
 $190 = HEAP32[$2>>2]|0;
 $$10 = $190;
 STACKTOP = sp;return ($$10|0);
}
function _t1_parse_font_matrix($face,$loader) {
 $face = $face|0;
 $loader = $loader|0;
 var $$phi$trans$insert = 0, $$pre = 0, $$pre$phi4Z2D = 0, $$pre$phi6Z2D = 0, $$pre$phiZ2D = 0, $$pre1 = 0, $$pre2 = 0, $$pre3 = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $temp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $temp = sp;
 $0 = ((($loader)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (FUNCTION_TABLE_iiiii[$1 & 8191]($loader,6,$temp,3)|0);
 $3 = ($2|0)<(6);
 if ($3) {
  $4 = ((($loader)) + 12|0);
  HEAP32[$4>>2] = 3;
  STACKTOP = sp;return;
 }
 $5 = ((($temp)) + 12|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)<(0);
 $8 = (0 - ($6))|0;
 $9 = $7 ? $8 : $6;
 $10 = ($9|0)==(0);
 if ($10) {
  $11 = ((($loader)) + 12|0);
  HEAP32[$11>>2] = 3;
  STACKTOP = sp;return;
 }
 $12 = (_FT_DivFix(1000,$9)|0);
 $13 = $12&65535;
 $14 = ((($face)) + 68|0);
 HEAP16[$14>>1] = $13;
 $15 = ($9|0)==(65536);
 if ($15) {
  $$pre = HEAP32[$5>>2]|0;
  $$phi$trans$insert = ((($temp)) + 20|0);
  $$pre1 = HEAP32[$$phi$trans$insert>>2]|0;
  $$pre2 = ((($temp)) + 4|0);
  $$pre3 = ((($temp)) + 8|0);
  $$pre5 = ((($temp)) + 16|0);
  $$pre$phi4Z2D = $$pre3;$$pre$phi6Z2D = $$pre5;$$pre$phiZ2D = $$pre2;$41 = $$pre;$46 = $$pre1;
 } else {
  $16 = HEAP32[$temp>>2]|0;
  $17 = (_FT_DivFix($16,$9)|0);
  HEAP32[$temp>>2] = $17;
  $18 = ((($temp)) + 4|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = (_FT_DivFix($19,$9)|0);
  HEAP32[$18>>2] = $20;
  $21 = ((($temp)) + 8|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = (_FT_DivFix($22,$9)|0);
  HEAP32[$21>>2] = $23;
  $24 = ((($temp)) + 16|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = (_FT_DivFix($25,$9)|0);
  HEAP32[$24>>2] = $26;
  $27 = ((($temp)) + 20|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = (_FT_DivFix($28,$9)|0);
  HEAP32[$27>>2] = $29;
  $30 = HEAP32[$5>>2]|0;
  $31 = $30 >> 31;
  $32 = $31 & -131072;
  $33 = $32 | 65536;
  HEAP32[$5>>2] = $33;
  $$pre$phi4Z2D = $21;$$pre$phi6Z2D = $24;$$pre$phiZ2D = $18;$41 = $33;$46 = $29;
 }
 $34 = HEAP32[$temp>>2]|0;
 $35 = ((($face)) + 436|0);
 HEAP32[$35>>2] = $34;
 $36 = HEAP32[$$pre$phiZ2D>>2]|0;
 $37 = ((($face)) + 444|0);
 HEAP32[$37>>2] = $36;
 $38 = HEAP32[$$pre$phi4Z2D>>2]|0;
 $39 = ((($face)) + 440|0);
 HEAP32[$39>>2] = $38;
 $40 = ((($face)) + 448|0);
 HEAP32[$40>>2] = $41;
 $42 = HEAP32[$$pre$phi6Z2D>>2]|0;
 $43 = $42 >> 16;
 $44 = ((($face)) + 452|0);
 HEAP32[$44>>2] = $43;
 $45 = $46 >> 16;
 $47 = ((($face)) + 456|0);
 HEAP32[$47>>2] = $45;
 STACKTOP = sp;return;
}
function _parse_encoding($face,$loader) {
 $face = $face|0;
 $loader = $loader|0;
 var $$lcssa9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $count$0 = 0, $cur$3$us = 0, $cur$4 = 0, $error = 0;
 var $exitcond = 0, $n$046 = 0, $n$113 = 0, $n$113$us = 0, $n$5$us = 0, $only_immediates$0 = 0, $or$cond = 0, $or$cond4 = 0, $or$cond4$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($loader)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 488|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($loader)) + 28|0);
 $5 = HEAP32[$4>>2]|0;
 FUNCTION_TABLE_vi[$5 & 8191]($loader);
 $6 = HEAP32[$loader>>2]|0;
 $7 = ($6>>>0)<($1>>>0);
 if (!($7)) {
  $8 = ((($loader)) + 12|0);
  HEAP32[$8>>2] = 3;
  STACKTOP = sp;return;
 }
 $9 = HEAP8[$6>>0]|0;
 $10 = $9&255;
 $11 = (($10) + -48)|0;
 $12 = ($11>>>0)<(10);
 $13 = ($9<<24>>24)==(91);
 $or$cond = $13 | $12;
 if (!($or$cond)) {
  $139 = ((($6)) + 17|0);
  $140 = ($139>>>0)<($1>>>0);
  if ($140) {
   $141 = (_strncmp($6,594610,16)|0);
   $142 = ($141|0)==(0);
   if ($142) {
    $143 = ((($face)) + 368|0);
    HEAP32[$143>>2] = 2;
    STACKTOP = sp;return;
   }
  }
  $144 = ((($6)) + 15|0);
  $145 = ($144>>>0)<($1>>>0);
  if ($145) {
   $146 = (_strncmp($6,594627,14)|0);
   $147 = ($146|0)==(0);
   if ($147) {
    $148 = ((($face)) + 368|0);
    HEAP32[$148>>2] = 4;
    STACKTOP = sp;return;
   }
  }
  $149 = ((($6)) + 18|0);
  $150 = ($149>>>0)<($1>>>0);
  if ($150) {
   $151 = (_strncmp($6,594642,17)|0);
   $152 = ($151|0)==(0);
   if ($152) {
    $153 = ((($face)) + 368|0);
    HEAP32[$153>>2] = 3;
    STACKTOP = sp;return;
   }
  }
  $154 = ((($loader)) + 12|0);
  HEAP32[$154>>2] = 162;
  STACKTOP = sp;return;
 }
 $14 = ((($loader)) + 100|0);
 $15 = ((($loader)) + 16|0);
 $16 = HEAP32[$15>>2]|0;
 if ($13) {
  $17 = ((($6)) + 1|0);
  HEAP32[$loader>>2] = $17;
  $count$0 = 256;$only_immediates$0 = 1;
 } else {
  $18 = ((($loader)) + 36|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = (FUNCTION_TABLE_ii[$19 & 8191]($loader)|0);
  $count$0 = $20;$only_immediates$0 = 0;
 }
 $21 = HEAP32[$4>>2]|0;
 FUNCTION_TABLE_vi[$21 & 8191]($loader);
 $22 = HEAP32[$loader>>2]|0;
 $23 = ($22>>>0)<($1>>>0);
 if (!($23)) {
  STACKTOP = sp;return;
 }
 $24 = ((($face)) + 372|0);
 HEAP32[$24>>2] = $count$0;
 $25 = ((($loader)) + 96|0);
 HEAP32[$25>>2] = $count$0;
 $26 = (_ft_mem_realloc($16,2,0,$count$0,0,$error)|0);
 $27 = ((($face)) + 384|0);
 HEAP32[$27>>2] = $26;
 $28 = HEAP32[$error>>2]|0;
 $29 = ($28|0)==(0);
 if ($29) {
  $30 = (_ft_mem_realloc($16,4,0,$count$0,0,$error)|0);
  $31 = ((($face)) + 388|0);
  HEAP32[$31>>2] = $30;
  $32 = HEAP32[$error>>2]|0;
  $33 = ($32|0)==(0);
  if ($33) {
   $34 = HEAP32[$3>>2]|0;
   $35 = HEAP32[$34>>2]|0;
   $36 = (FUNCTION_TABLE_iiii[$35 & 8191]($14,$count$0,$16)|0);
   HEAP32[$error>>2] = $36;
   $37 = ($36|0)==(0);
   if ($37) {
    $38 = ($count$0|0)>(0);
    if ($38) {
     $39 = ((($loader)) + 144|0);
     $n$046 = 0;
     while(1) {
      $42 = HEAP32[$39>>2]|0;
      (FUNCTION_TABLE_iiiii[$42 & 8191]($14,$n$046,594602,8)|0);
      $43 = (($n$046) + 1)|0;
      $exitcond = ($43|0)==($count$0|0);
      if ($exitcond) {
       break;
      } else {
       $n$046 = $43;
      }
     }
    }
    $44 = HEAP32[$4>>2]|0;
    FUNCTION_TABLE_vi[$44 & 8191]($loader);
    $45 = HEAP32[$loader>>2]|0;
    $46 = ($45>>>0)<($1>>>0);
    L40: do {
     if ($46) {
      $47 = ($only_immediates$0<<24>>24)==(0);
      $48 = ((($loader)) + 32|0);
      $49 = ((($loader)) + 12|0);
      $50 = ((($loader)) + 144|0);
      $51 = ((($loader)) + 124|0);
      $52 = ((($loader)) + 36|0);
      L42: do {
       if ($47) {
        $54 = $45;$n$113$us = 0;
        while(1) {
         $53 = HEAP8[$54>>0]|0;
         switch ($53<<24>>24) {
         case 93:  {
          $$lcssa9 = $54;
          break L42;
          break;
         }
         case 100:  {
          $55 = ((($54)) + 3|0);
          $56 = ($55>>>0)<($1>>>0);
          if ($56) {
           $57 = ((($54)) + 1|0);
           $58 = HEAP8[$57>>0]|0;
           $59 = ($58<<24>>24)==(101);
           if ($59) {
            $60 = ((($54)) + 2|0);
            $61 = HEAP8[$60>>0]|0;
            $62 = ($61<<24>>24)==(102);
            if ($62) {
             $63 = HEAP8[$55>>0]|0;
             switch ($63<<24>>24) {
             case 37: case 125: case 123: case 93: case 91: case 62: case 60: case 41: case 40: case 47: case 0: case 12: case 9: case 10: case 13: case 32:  {
              $cur$4 = $55;
              break L40;
              break;
             }
             default: {
             }
             }
            }
           }
          }
          break;
         }
         default: {
         }
         }
         $64 = $53&255;
         $65 = (($64) + -48)|0;
         $66 = ($65>>>0)>(9);
         if ($66) {
          $95 = HEAP32[$48>>2]|0;
          FUNCTION_TABLE_vi[$95 & 8191]($loader);
          $96 = HEAP32[$49>>2]|0;
          $97 = ($96|0)==(0);
          if ($97) {
           $cur$3$us = $54;$n$5$us = $n$113$us;
          } else {
           label = 47;
           break;
          }
         } else {
          $67 = HEAP32[$52>>2]|0;
          $68 = (FUNCTION_TABLE_ii[$67 & 8191]($loader)|0);
          $69 = HEAP32[$4>>2]|0;
          FUNCTION_TABLE_vi[$69 & 8191]($loader);
          $70 = HEAP32[$loader>>2]|0;
          $71 = ($54|0)==($70|0);
          if ($71) {
           label = 38;
           break;
          }
          $72 = ((($70)) + 2|0);
          $73 = ($72>>>0)<($1>>>0);
          if ($73) {
           $74 = HEAP8[$70>>0]|0;
           $75 = ($74<<24>>24)==(47);
           $76 = ($n$113$us|0)<($count$0|0);
           $or$cond4$us = $76 & $75;
           if ($or$cond4$us) {
            $77 = ((($70)) + 1|0);
            HEAP32[$loader>>2] = $77;
            $78 = HEAP32[$48>>2]|0;
            FUNCTION_TABLE_vi[$78 & 8191]($loader);
            $79 = HEAP32[$loader>>2]|0;
            $80 = ($79>>>0)<($1>>>0);
            if (!($80)) {
             label = 47;
             break;
            }
            $81 = HEAP32[$49>>2]|0;
            $82 = ($81|0)==(0);
            if (!($82)) {
             label = 47;
             break;
            }
            $83 = $79;
            $84 = $77;
            $85 = (($83) - ($84))|0;
            $86 = HEAP32[$50>>2]|0;
            $87 = (($85) + 1)|0;
            $88 = (FUNCTION_TABLE_iiiii[$86 & 8191]($14,$68,$77,$87)|0);
            HEAP32[$49>>2] = $88;
            $89 = ($88|0)==(0);
            if (!($89)) {
             label = 47;
             break;
            }
            $90 = HEAP32[$51>>2]|0;
            $91 = (($90) + ($68<<2)|0);
            $92 = HEAP32[$91>>2]|0;
            $93 = (($92) + ($85)|0);
            HEAP8[$93>>0] = 0;
            $94 = (($n$113$us) + 1)|0;
            $cur$3$us = $77;$n$5$us = $94;
           } else {
            $cur$3$us = $70;$n$5$us = $n$113$us;
           }
          } else {
           $cur$3$us = $70;$n$5$us = $n$113$us;
          }
         }
         $98 = HEAP32[$4>>2]|0;
         FUNCTION_TABLE_vi[$98 & 8191]($loader);
         $99 = HEAP32[$loader>>2]|0;
         $100 = ($99>>>0)<($1>>>0);
         if ($100) {
          $54 = $99;$n$113$us = $n$5$us;
         } else {
          $cur$4 = $cur$3$us;
          break L40;
         }
        }
        if ((label|0) == 38) {
         HEAP32[$49>>2] = 2;
         STACKTOP = sp;return;
        }
        else if ((label|0) == 47) {
         STACKTOP = sp;return;
        }
       } else {
        $102 = $45;$n$113 = 0;
        while(1) {
         $101 = HEAP8[$102>>0]|0;
         switch ($101<<24>>24) {
         case 93:  {
          $$lcssa9 = $102;
          break L42;
          break;
         }
         case 100:  {
          $103 = ((($102)) + 3|0);
          $104 = ($103>>>0)<($1>>>0);
          if ($104) {
           $105 = ((($102)) + 1|0);
           $106 = HEAP8[$105>>0]|0;
           $107 = ($106<<24>>24)==(101);
           if ($107) {
            $108 = ((($102)) + 2|0);
            $109 = HEAP8[$108>>0]|0;
            $110 = ($109<<24>>24)==(102);
            if ($110) {
             $111 = HEAP8[$103>>0]|0;
             switch ($111<<24>>24) {
             case 37: case 125: case 123: case 93: case 91: case 62: case 60: case 41: case 40: case 47: case 0: case 12: case 9: case 10: case 13: case 32:  {
              $cur$4 = $103;
              break L40;
              break;
             }
             default: {
             }
             }
            }
           }
          }
          break;
         }
         default: {
         }
         }
         $113 = ((($102)) + 2|0);
         $114 = ($113>>>0)<($1>>>0);
         if (!($114)) {
          label = 44;
          break;
         }
         $115 = ($101<<24>>24)==(47);
         $116 = ($n$113|0)<($count$0|0);
         $or$cond4 = $116 & $115;
         if (!($or$cond4)) {
          label = 44;
          break;
         }
         $117 = ((($102)) + 1|0);
         HEAP32[$loader>>2] = $117;
         $118 = HEAP32[$48>>2]|0;
         FUNCTION_TABLE_vi[$118 & 8191]($loader);
         $119 = HEAP32[$loader>>2]|0;
         $120 = ($119>>>0)<($1>>>0);
         if (!($120)) {
          label = 47;
          break;
         }
         $121 = HEAP32[$49>>2]|0;
         $122 = ($121|0)==(0);
         if (!($122)) {
          label = 47;
          break;
         }
         $123 = $119;
         $124 = $117;
         $125 = (($123) - ($124))|0;
         $126 = HEAP32[$50>>2]|0;
         $127 = (($125) + 1)|0;
         $128 = (FUNCTION_TABLE_iiiii[$126 & 8191]($14,$n$113,$117,$127)|0);
         HEAP32[$49>>2] = $128;
         $129 = ($128|0)==(0);
         if (!($129)) {
          label = 47;
          break;
         }
         $130 = HEAP32[$51>>2]|0;
         $131 = (($130) + ($n$113<<2)|0);
         $132 = HEAP32[$131>>2]|0;
         $133 = (($132) + ($125)|0);
         HEAP8[$133>>0] = 0;
         $134 = (($n$113) + 1)|0;
         $135 = HEAP32[$4>>2]|0;
         FUNCTION_TABLE_vi[$135 & 8191]($loader);
         $136 = HEAP32[$loader>>2]|0;
         $137 = ($136>>>0)<($1>>>0);
         if ($137) {
          $102 = $136;$n$113 = $134;
         } else {
          $cur$4 = $117;
          break L40;
         }
        }
        if ((label|0) == 44) {
         HEAP32[$49>>2] = 2;
         STACKTOP = sp;return;
        }
        else if ((label|0) == 47) {
         STACKTOP = sp;return;
        }
       }
      } while(0);
      $112 = ((($$lcssa9)) + 1|0);
      $cur$4 = $112;
     } else {
      $cur$4 = $6;
     }
    } while(0);
    $138 = ((($face)) + 368|0);
    HEAP32[$138>>2] = 1;
    HEAP32[$loader>>2] = $cur$4;
    STACKTOP = sp;return;
   } else {
    $41 = $36;
   }
  } else {
   $41 = $32;
  }
 } else {
  $41 = $28;
 }
 $40 = ((($loader)) + 12|0);
 HEAP32[$40>>2] = $41;
 STACKTOP = sp;return;
}
function _parse_subrs($face,$loader) {
 $face = $face|0;
 $loader = $loader|0;
 var $$lcssa = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $error = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($loader)) + 316|0);
 $1 = ((($loader)) + 16|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($face)) + 488|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($loader)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 FUNCTION_TABLE_vi[$6 & 8191]($loader);
 $7 = HEAP32[$loader>>2]|0;
 $8 = ((($loader)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($7>>>0)<($9>>>0);
 if ($10) {
  $11 = HEAP8[$7>>0]|0;
  $12 = ($11<<24>>24)==(91);
  if ($12) {
   $13 = ((($loader)) + 32|0);
   $14 = HEAP32[$13>>2]|0;
   FUNCTION_TABLE_vi[$14 & 8191]($loader);
   $15 = HEAP32[$5>>2]|0;
   FUNCTION_TABLE_vi[$15 & 8191]($loader);
   $16 = HEAP32[$loader>>2]|0;
   $17 = HEAP32[$8>>2]|0;
   $18 = ($16>>>0)<($17>>>0);
   if ($18) {
    $19 = HEAP8[$16>>0]|0;
    $20 = ($19<<24>>24)==(93);
    if ($20) {
     STACKTOP = sp;return;
    }
   }
   $21 = ((($loader)) + 12|0);
   HEAP32[$21>>2] = 3;
   STACKTOP = sp;return;
  }
 }
 $22 = ((($loader)) + 36|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = (FUNCTION_TABLE_ii[$23 & 8191]($loader)|0);
 $25 = ((($loader)) + 32|0);
 $26 = HEAP32[$25>>2]|0;
 FUNCTION_TABLE_vi[$26 & 8191]($loader);
 $27 = ((($loader)) + 12|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ($28|0)==(0);
 if (!($29)) {
  STACKTOP = sp;return;
 }
 $30 = HEAP32[$5>>2]|0;
 FUNCTION_TABLE_vi[$30 & 8191]($loader);
 $31 = ((($loader)) + 312|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ($32|0)==(0);
 if ($33) {
  $34 = HEAP32[$4>>2]|0;
  $35 = HEAP32[$34>>2]|0;
  $36 = (FUNCTION_TABLE_iiii[$35 & 8191]($0,$24,$2)|0);
  HEAP32[$error>>2] = $36;
  $37 = ($36|0)==(0);
  if ($37) {
   label = 9;
  } else {
   $114 = $36;
  }
 } else {
  label = 9;
 }
 L15: do {
  if ((label|0) == 9) {
   $38 = HEAP32[$loader>>2]|0;
   $39 = ((($38)) + 4|0);
   $40 = HEAP32[$8>>2]|0;
   $41 = ($39>>>0)<($40>>>0);
   L17: do {
    if ($41) {
     $42 = ((($face)) + 128|0);
     $43 = ((($face)) + 172|0);
     $44 = ((($4)) + 16|0);
     $45 = ((($loader)) + 360|0);
     $46 = $38;
     while(1) {
      $47 = (_strncmp($46,594183,3)|0);
      $48 = ($47|0)==(0);
      if (!($48)) {
       break L17;
      }
      $49 = HEAP32[$25>>2]|0;
      FUNCTION_TABLE_vi[$49 & 8191]($loader);
      $50 = HEAP32[$22>>2]|0;
      $51 = (FUNCTION_TABLE_ii[$50 & 8191]($loader)|0);
      $52 = HEAP32[$42>>2]|0;
      $53 = ((($52)) + 48|0);
      $54 = HEAP32[$53>>2]|0;
      $55 = HEAP32[$8>>2]|0;
      $56 = HEAP32[$5>>2]|0;
      FUNCTION_TABLE_vi[$56 & 8191]($loader);
      $57 = HEAP32[$loader>>2]|0;
      $58 = ($57>>>0)<($55>>>0);
      if (!($58)) {
       $$lcssa = $54;
       label = 16;
       break;
      }
      $59 = HEAP8[$57>>0]|0;
      $60 = $59&255;
      $61 = (($60) + -48)|0;
      $62 = ($61>>>0)<(10);
      if (!($62)) {
       $$lcssa = $54;
       label = 16;
       break;
      }
      $63 = HEAP32[$22>>2]|0;
      $64 = (FUNCTION_TABLE_ii[$63 & 8191]($loader)|0);
      $65 = HEAP32[$25>>2]|0;
      FUNCTION_TABLE_vi[$65 & 8191]($loader);
      $66 = HEAP32[$loader>>2]|0;
      $67 = ((($66)) + 1|0);
      $68 = ($64|0)>(-1);
      if (!($68)) {
       $$lcssa = $54;
       label = 16;
       break;
      }
      $69 = $55;
      $70 = $67;
      $71 = (($69) - ($70))|0;
      $72 = ($64|0)<($71|0);
      if (!($72)) {
       $$lcssa = $54;
       label = 16;
       break;
      }
      $74 = (($64) + 1)|0;
      $75 = (($66) + ($74)|0);
      HEAP32[$loader>>2] = $75;
      $76 = HEAP32[$27>>2]|0;
      $77 = ($76|0)==(0);
      if (!($77)) {
       label = 35;
       break;
      }
      $78 = HEAP32[$25>>2]|0;
      FUNCTION_TABLE_vi[$78 & 8191]($loader);
      $79 = HEAP32[$27>>2]|0;
      $80 = ($79|0)==(0);
      if (!($80)) {
       label = 35;
       break;
      }
      $81 = HEAP32[$5>>2]|0;
      FUNCTION_TABLE_vi[$81 & 8191]($loader);
      $82 = HEAP32[$loader>>2]|0;
      $83 = ((($82)) + 4|0);
      $84 = HEAP32[$8>>2]|0;
      $85 = ($83>>>0)<($84>>>0);
      if ($85) {
       $86 = (_strncmp($82,594187,3)|0);
       $87 = ($86|0)==(0);
       if ($87) {
        $88 = HEAP32[$25>>2]|0;
        FUNCTION_TABLE_vi[$88 & 8191]($loader);
        $89 = HEAP32[$5>>2]|0;
        FUNCTION_TABLE_vi[$89 & 8191]($loader);
       }
      }
      $90 = HEAP32[$31>>2]|0;
      $91 = ($90|0)==(0);
      if ($91) {
       $96 = HEAP32[$43>>2]|0;
       $97 = ($96|0)>(-1);
       if ($97) {
        $98 = ($64>>>0)<($96>>>0);
        if ($98) {
         label = 27;
         break;
        }
        $99 = (_ft_mem_alloc($2,$64,$error)|0);
        $100 = HEAP32[$error>>2]|0;
        $101 = ($100|0)==(0);
        if (!($101)) {
         $114 = $100;
         break L15;
        }
        _memcpy(($99|0),($67|0),($64|0))|0;
        $102 = HEAP32[$44>>2]|0;
        FUNCTION_TABLE_viii[$102 & 8191]($99,$64,4330);
        $103 = HEAP32[$43>>2]|0;
        $104 = (($64) - ($103))|0;
        $105 = HEAP32[$45>>2]|0;
        $106 = (($99) + ($103)|0);
        $107 = (FUNCTION_TABLE_iiiii[$105 & 8191]($0,$51,$106,$104)|0);
        HEAP32[$error>>2] = $107;
        _ft_mem_free($2,$99);
        $$pre = HEAP32[$error>>2]|0;
        $111 = $$pre;
       } else {
        $108 = HEAP32[$45>>2]|0;
        $109 = (FUNCTION_TABLE_iiiii[$108 & 8191]($0,$51,$67,$64)|0);
        HEAP32[$error>>2] = $109;
        $111 = $109;
       }
       $110 = ($111|0)==(0);
       if (!($110)) {
        $114 = $111;
        break L15;
       }
      }
      $92 = HEAP32[$loader>>2]|0;
      $93 = ((($92)) + 4|0);
      $94 = HEAP32[$8>>2]|0;
      $95 = ($93>>>0)<($94>>>0);
      if ($95) {
       $46 = $92;
      } else {
       break L17;
      }
     }
     if ((label|0) == 16) {
      $73 = ($$lcssa|0)==(0|0);
      if (!($73)) {
       STACKTOP = sp;return;
      }
      HEAP32[$27>>2] = 3;
      STACKTOP = sp;return;
     }
     else if ((label|0) == 27) {
      HEAP32[$error>>2] = 3;
      $114 = 3;
      break L15;
     }
     else if ((label|0) == 35) {
      STACKTOP = sp;return;
     }
    }
   } while(0);
   $112 = HEAP32[$31>>2]|0;
   $113 = ($112|0)==(0);
   if (!($113)) {
    STACKTOP = sp;return;
   }
   HEAP32[$31>>2] = $24;
   STACKTOP = sp;return;
  }
 } while(0);
 HEAP32[$27>>2] = $114;
 STACKTOP = sp;return;
}
function _parse_charstrings($face,$loader) {
 $face = $face|0;
 $loader = $loader|0;
 var $$lcssa = 0, $$lcssa32 = 0, $$lcssa33 = 0, $$lcssa64 = 0, $$lcssa67 = 0, $$notdef_found$0 = 0, $$pr28 = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0;
 var $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0;
 var $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0;
 var $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0;
 var $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0;
 var $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0;
 var $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0;
 var $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $cond5 = 0, $error = 0, $n$0$notdef_index$0 = 0, $n$0$ph = 0, $n$0$ph144 = 0, $notdef_found$0$ph = 0, $notdef_found$0$ph425 = 0, $notdef_found$1 = 0, $notdef_glyph = 0, $notdef_index$0$ph = 0, $notdef_index$0$ph412 = 0, $notdef_index$1 = 0, $or$cond4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $notdef_glyph = sp + 4|0;
 $0 = ((($loader)) + 208|0);
 $1 = ((($loader)) + 156|0);
 $2 = ((($loader)) + 260|0);
 $3 = ((($loader)) + 16|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($face)) + 488|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($loader)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($loader)) + 36|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (FUNCTION_TABLE_ii[$10 & 8191]($loader)|0);
 $12 = ($11|0)<(0);
 L1: do {
  if ($12) {
   HEAP32[$error>>2] = 3;
   $$pre = ((($loader)) + 12|0);
   $$pre$phiZ2D = $$pre;$309 = 3;
  } else {
   $13 = ($11|0)==(0);
   if ($13) {
    STACKTOP = sp;return;
   }
   $14 = ((($loader)) + 12|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)==(0);
   if (!($16)) {
    STACKTOP = sp;return;
   }
   $17 = ((($loader)) + 152|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = ($18|0)==(0);
   if ($19) {
    $20 = HEAP32[$6>>2]|0;
    $21 = HEAP32[$20>>2]|0;
    $22 = (($11) + 6)|0;
    $23 = (FUNCTION_TABLE_iiii[$21 & 8191]($0,$22,$4)|0);
    HEAP32[$error>>2] = $23;
    $24 = ($23|0)==(0);
    if (!($24)) {
     $$pre$phiZ2D = $14;$309 = $23;
     break;
    }
    $25 = HEAP32[$6>>2]|0;
    $26 = HEAP32[$25>>2]|0;
    $27 = (FUNCTION_TABLE_iiii[$26 & 8191]($1,$22,$4)|0);
    HEAP32[$error>>2] = $27;
    $28 = ($27|0)==(0);
    if (!($28)) {
     $$pre$phiZ2D = $14;$309 = $27;
     break;
    }
    $29 = HEAP32[$6>>2]|0;
    $30 = HEAP32[$29>>2]|0;
    $31 = (FUNCTION_TABLE_iiii[$30 & 8191]($2,4,$4)|0);
    HEAP32[$error>>2] = $31;
    $32 = ($31|0)==(0);
    if (!($32)) {
     $$pre$phiZ2D = $14;$309 = $31;
     break;
    }
   }
   $33 = ((($loader)) + 28|0);
   $34 = ((($loader)) + 32|0);
   $35 = ((($face)) + 128|0);
   $36 = ((($loader)) + 200|0);
   $37 = ((($loader)) + 180|0);
   $38 = ((($face)) + 172|0);
   $39 = (($11) + 5)|0;
   $40 = ((($6)) + 16|0);
   $41 = ((($loader)) + 252|0);
   $n$0$ph = 0;$notdef_found$0$ph = 0;$notdef_index$0$ph = 0;
   L15: while(1) {
    $42 = HEAP32[$33>>2]|0;
    FUNCTION_TABLE_vi[$42 & 8191]($loader);
    $43 = HEAP32[$loader>>2]|0;
    $44 = ($43>>>0)<($8>>>0);
    if (!($44)) {
     $n$0$ph144 = $n$0$ph;$notdef_found$0$ph425 = $notdef_found$0$ph;$notdef_index$0$ph412 = $notdef_index$0$ph;
     label = 60;
     break;
    }
    $45 = ($n$0$ph|0)==(0);
    L18: do {
     if ($45) {
      $103 = $43;
      while(1) {
       $102 = ((($103)) + 3|0);
       $104 = ($102>>>0)<($8>>>0);
       L21: do {
        if ($104) {
         $105 = HEAP8[$102>>0]|0;
         switch ($105<<24>>24) {
         case 37: case 125: case 123: case 93: case 91: case 62: case 60: case 41: case 40: case 47: case 0: case 12: case 9: case 10: case 13: case 32:  {
          break;
         }
         default: {
          break L21;
         }
         }
         $106 = HEAP8[$103>>0]|0;
         $107 = ($106<<24>>24)==(101);
         if ($107) {
          $108 = ((($103)) + 1|0);
          $109 = HEAP8[$108>>0]|0;
          $110 = ($109<<24>>24)==(110);
          if ($110) {
           $111 = ((($103)) + 2|0);
           $112 = HEAP8[$111>>0]|0;
           $113 = ($112<<24>>24)==(100);
           if ($113) {
            $n$0$ph144 = 0;$notdef_found$0$ph425 = $notdef_found$0$ph;$notdef_index$0$ph412 = $notdef_index$0$ph;
            label = 60;
            break L15;
           }
          }
         }
        }
       } while(0);
       $114 = HEAP32[$34>>2]|0;
       FUNCTION_TABLE_vi[$114 & 8191]($loader);
       $115 = HEAP32[$loader>>2]|0;
       $116 = ($115>>>0)<($8>>>0);
       if (!($116)) {
        label = 36;
        break L15;
       }
       $117 = HEAP32[$14>>2]|0;
       $118 = ($117|0)==(0);
       if (!($118)) {
        label = 79;
        break L15;
       }
       $119 = HEAP8[$103>>0]|0;
       $120 = ($119<<24>>24)==(47);
       if ($120) {
        $121 = ((($103)) + 2|0);
        $122 = ($121>>>0)<($8>>>0);
        if (!($122)) {
         label = 40;
         break L15;
        }
        $123 = HEAP32[$35>>2]|0;
        $124 = ((($123)) + 48|0);
        $125 = HEAP32[$124>>2]|0;
        $126 = HEAP32[$7>>2]|0;
        $127 = HEAP32[$33>>2]|0;
        FUNCTION_TABLE_vi[$127 & 8191]($loader);
        $128 = HEAP32[$loader>>2]|0;
        $129 = ($128>>>0)<($126>>>0);
        if (!($129)) {
         $$lcssa64 = $125;
         label = 45;
         break L15;
        }
        $130 = HEAP8[$128>>0]|0;
        $131 = $130&255;
        $132 = (($131) + -48)|0;
        $133 = ($132>>>0)<(10);
        if (!($133)) {
         $$lcssa64 = $125;
         label = 45;
         break L15;
        }
        $134 = HEAP32[$9>>2]|0;
        $135 = (FUNCTION_TABLE_ii[$134 & 8191]($loader)|0);
        $136 = HEAP32[$34>>2]|0;
        FUNCTION_TABLE_vi[$136 & 8191]($loader);
        $137 = HEAP32[$loader>>2]|0;
        $138 = ((($137)) + 1|0);
        $139 = ($135|0)>(-1);
        if (!($139)) {
         $$lcssa64 = $125;
         label = 45;
         break L15;
        }
        $140 = $126;
        $141 = $138;
        $142 = (($140) - ($141))|0;
        $143 = ($135|0)<($142|0);
        if (!($143)) {
         $$lcssa64 = $125;
         label = 45;
         break L15;
        }
        $145 = (($135) + 1)|0;
        $146 = (($137) + ($145)|0);
        HEAP32[$loader>>2] = $146;
        $147 = HEAP32[$14>>2]|0;
        $148 = ($147|0)==(0);
        if (!($148)) {
         label = 79;
         break L15;
        }
        $149 = HEAP32[$17>>2]|0;
        $150 = ($149|0)==(0);
        if ($150) {
         $$lcssa = $103;$$lcssa32 = $135;$$lcssa33 = $138;$$lcssa67 = $115;
         break L18;
        }
       }
       $151 = HEAP32[$33>>2]|0;
       FUNCTION_TABLE_vi[$151 & 8191]($loader);
       $152 = HEAP32[$loader>>2]|0;
       $153 = ($152>>>0)<($8>>>0);
       if ($153) {
        $103 = $152;
       } else {
        $n$0$ph144 = 0;$notdef_found$0$ph425 = $notdef_found$0$ph;$notdef_index$0$ph412 = $notdef_index$0$ph;
        label = 60;
        break L15;
       }
      }
     } else {
      $47 = $43;
      while(1) {
       $46 = ((($47)) + 3|0);
       $48 = ($46>>>0)<($8>>>0);
       L40: do {
        if ($48) {
         $49 = HEAP8[$46>>0]|0;
         switch ($49<<24>>24) {
         case 37: case 125: case 123: case 93: case 91: case 62: case 60: case 41: case 40: case 47: case 0: case 12: case 9: case 10: case 13: case 32:  {
          break;
         }
         default: {
          break L40;
         }
         }
         $50 = HEAP8[$47>>0]|0;
         switch ($50<<24>>24) {
         case 100:  {
          $51 = ((($47)) + 1|0);
          $52 = HEAP8[$51>>0]|0;
          $53 = ($52<<24>>24)==(101);
          if (!($53)) {
           break L40;
          }
          $54 = ((($47)) + 2|0);
          $55 = HEAP8[$54>>0]|0;
          $56 = ($55<<24>>24)==(102);
          if ($56) {
           $n$0$ph144 = $n$0$ph;$notdef_found$0$ph425 = $notdef_found$0$ph;$notdef_index$0$ph412 = $notdef_index$0$ph;
           label = 60;
           break L15;
          } else {
           break L40;
          }
          break;
         }
         case 101:  {
          $57 = ((($47)) + 1|0);
          $58 = HEAP8[$57>>0]|0;
          $59 = ($58<<24>>24)==(110);
          if (!($59)) {
           break L40;
          }
          $60 = ((($47)) + 2|0);
          $61 = HEAP8[$60>>0]|0;
          $62 = ($61<<24>>24)==(100);
          if ($62) {
           $n$0$ph144 = $n$0$ph;$notdef_found$0$ph425 = $notdef_found$0$ph;$notdef_index$0$ph412 = $notdef_index$0$ph;
           label = 60;
           break L15;
          } else {
           break L40;
          }
          break;
         }
         default: {
          break L40;
         }
         }
        }
       } while(0);
       $63 = HEAP32[$34>>2]|0;
       FUNCTION_TABLE_vi[$63 & 8191]($loader);
       $64 = HEAP32[$loader>>2]|0;
       $65 = ($64>>>0)<($8>>>0);
       if (!($65)) {
        label = 36;
        break L15;
       }
       $66 = HEAP32[$14>>2]|0;
       $67 = ($66|0)==(0);
       if (!($67)) {
        label = 79;
        break L15;
       }
       $68 = HEAP8[$47>>0]|0;
       $69 = ($68<<24>>24)==(47);
       if ($69) {
        $70 = ((($47)) + 2|0);
        $71 = ($70>>>0)<($8>>>0);
        if (!($71)) {
         label = 40;
         break L15;
        }
        $72 = HEAP32[$35>>2]|0;
        $73 = ((($72)) + 48|0);
        $74 = HEAP32[$73>>2]|0;
        $75 = HEAP32[$7>>2]|0;
        $76 = HEAP32[$33>>2]|0;
        FUNCTION_TABLE_vi[$76 & 8191]($loader);
        $77 = HEAP32[$loader>>2]|0;
        $78 = ($77>>>0)<($75>>>0);
        if (!($78)) {
         $$lcssa64 = $74;
         label = 45;
         break L15;
        }
        $79 = HEAP8[$77>>0]|0;
        $80 = $79&255;
        $81 = (($80) + -48)|0;
        $82 = ($81>>>0)<(10);
        if (!($82)) {
         $$lcssa64 = $74;
         label = 45;
         break L15;
        }
        $83 = HEAP32[$9>>2]|0;
        $84 = (FUNCTION_TABLE_ii[$83 & 8191]($loader)|0);
        $85 = HEAP32[$34>>2]|0;
        FUNCTION_TABLE_vi[$85 & 8191]($loader);
        $86 = HEAP32[$loader>>2]|0;
        $87 = ((($86)) + 1|0);
        $88 = ($84|0)>(-1);
        if (!($88)) {
         $$lcssa64 = $74;
         label = 45;
         break L15;
        }
        $89 = $75;
        $90 = $87;
        $91 = (($89) - ($90))|0;
        $92 = ($84|0)<($91|0);
        if (!($92)) {
         $$lcssa64 = $74;
         label = 45;
         break L15;
        }
        $93 = (($84) + 1)|0;
        $94 = (($86) + ($93)|0);
        HEAP32[$loader>>2] = $94;
        $95 = HEAP32[$14>>2]|0;
        $96 = ($95|0)==(0);
        if (!($96)) {
         label = 79;
         break L15;
        }
        $97 = HEAP32[$17>>2]|0;
        $98 = ($97|0)==(0);
        if ($98) {
         $$lcssa = $47;$$lcssa32 = $84;$$lcssa33 = $87;$$lcssa67 = $64;
         break L18;
        }
       }
       $99 = HEAP32[$33>>2]|0;
       FUNCTION_TABLE_vi[$99 & 8191]($loader);
       $100 = HEAP32[$loader>>2]|0;
       $101 = ($100>>>0)<($8>>>0);
       if ($101) {
        $47 = $100;
       } else {
        $n$0$ph144 = $n$0$ph;$notdef_found$0$ph425 = $notdef_found$0$ph;$notdef_index$0$ph412 = $notdef_index$0$ph;
        label = 60;
        break L15;
       }
      }
     }
    } while(0);
    $154 = ((($$lcssa)) + 1|0);
    $155 = $$lcssa67;
    $156 = $154;
    $157 = (($155) - ($156))|0;
    $158 = HEAP32[$36>>2]|0;
    $159 = (($157) + 1)|0;
    $160 = (FUNCTION_TABLE_iiiii[$158 & 8191]($1,$n$0$ph,$154,$159)|0);
    HEAP32[$error>>2] = $160;
    $161 = ($160|0)==(0);
    if (!($161)) {
     $$pre$phiZ2D = $14;$309 = $160;
     break L1;
    }
    $162 = HEAP32[$37>>2]|0;
    $163 = (($162) + ($n$0$ph<<2)|0);
    $164 = HEAP32[$163>>2]|0;
    $165 = (($164) + ($157)|0);
    HEAP8[$165>>0] = 0;
    $166 = HEAP8[$154>>0]|0;
    $167 = ($166<<24>>24)==(46);
    if ($167) {
     $168 = HEAP32[$37>>2]|0;
     $169 = (($168) + ($n$0$ph<<2)|0);
     $170 = HEAP32[$169>>2]|0;
     $171 = (_strcmp(594602,$170)|0);
     $172 = ($171|0)==(0);
     $n$0$notdef_index$0 = $172 ? $n$0$ph : $notdef_index$0$ph;
     $$notdef_found$0 = $172 ? 1 : $notdef_found$0$ph;
     $notdef_found$1 = $$notdef_found$0;$notdef_index$1 = $n$0$notdef_index$0;
    } else {
     $notdef_found$1 = $notdef_found$0$ph;$notdef_index$1 = $notdef_index$0$ph;
    }
    $173 = HEAP32[$38>>2]|0;
    $174 = ($173|0)>(-1);
    $175 = ($n$0$ph|0)<($39|0);
    $or$cond4 = $175 & $174;
    if ($or$cond4) {
     $176 = ($$lcssa32>>>0)>($173>>>0);
     if (!($176)) {
      label = 55;
      break;
     }
     $177 = (_ft_mem_alloc($4,$$lcssa32,$error)|0);
     $178 = HEAP32[$error>>2]|0;
     $179 = ($178|0)==(0);
     if (!($179)) {
      $$pre$phiZ2D = $14;$309 = $178;
      break L1;
     }
     _memcpy(($177|0),($$lcssa33|0),($$lcssa32|0))|0;
     $180 = HEAP32[$40>>2]|0;
     FUNCTION_TABLE_viii[$180 & 8191]($177,$$lcssa32,4330);
     $181 = HEAP32[$38>>2]|0;
     $182 = (($$lcssa32) - ($181))|0;
     $183 = HEAP32[$41>>2]|0;
     $184 = (($177) + ($181)|0);
     $185 = (FUNCTION_TABLE_iiiii[$183 & 8191]($0,$n$0$ph,$184,$182)|0);
     HEAP32[$error>>2] = $185;
     _ft_mem_free($4,$177);
     $$pr28 = HEAP32[$error>>2]|0;
     $189 = $$pr28;
    } else {
     $186 = HEAP32[$41>>2]|0;
     $187 = (FUNCTION_TABLE_iiiii[$186 & 8191]($0,$n$0$ph,$$lcssa33,$$lcssa32)|0);
     HEAP32[$error>>2] = $187;
     $189 = $187;
    }
    $188 = ($189|0)==(0);
    $190 = (($n$0$ph) + 1)|0;
    if ($188) {
     $n$0$ph = $190;$notdef_found$0$ph = $notdef_found$1;$notdef_index$0$ph = $notdef_index$1;
    } else {
     $$pre$phiZ2D = $14;$309 = $189;
     break L1;
    }
   }
   if ((label|0) == 36) {
    HEAP32[$error>>2] = 3;
    $$pre$phiZ2D = $14;$309 = 3;
    break;
   }
   else if ((label|0) == 40) {
    HEAP32[$error>>2] = 3;
    $$pre$phiZ2D = $14;$309 = 3;
    break;
   }
   else if ((label|0) == 45) {
    $144 = ($$lcssa64|0)==(0|0);
    if (!($144)) {
     STACKTOP = sp;return;
    }
    HEAP32[$14>>2] = 3;
    STACKTOP = sp;return;
   }
   else if ((label|0) == 55) {
    HEAP32[$error>>2] = 3;
    $$pre$phiZ2D = $14;$309 = 3;
    break;
   }
   else if ((label|0) == 60) {
    HEAP32[$17>>2] = $n$0$ph144;
    $cond5 = ($notdef_found$0$ph425<<24>>24)==(0);
    if ($cond5) {
     ;HEAP8[$notdef_glyph>>0]=HEAP8[594191>>0]|0;HEAP8[$notdef_glyph+1>>0]=HEAP8[594191+1>>0]|0;HEAP8[$notdef_glyph+2>>0]=HEAP8[594191+2>>0]|0;HEAP8[$notdef_glyph+3>>0]=HEAP8[594191+3>>0]|0;HEAP8[$notdef_glyph+4>>0]=HEAP8[594191+4>>0]|0;
     $265 = ((($loader)) + 304|0);
     $266 = HEAP32[$265>>2]|0;
     $267 = HEAP32[$37>>2]|0;
     $268 = HEAP32[$267>>2]|0;
     $269 = ((($loader)) + 184|0);
     $270 = HEAP32[$269>>2]|0;
     $271 = HEAP32[$270>>2]|0;
     $272 = (FUNCTION_TABLE_iiiii[$266 & 8191]($2,0,$268,$271)|0);
     HEAP32[$error>>2] = $272;
     $273 = ($272|0)==(0);
     if ($273) {
      $274 = HEAP32[$265>>2]|0;
      $275 = ((($loader)) + 232|0);
      $276 = HEAP32[$275>>2]|0;
      $277 = HEAP32[$276>>2]|0;
      $278 = ((($loader)) + 236|0);
      $279 = HEAP32[$278>>2]|0;
      $280 = HEAP32[$279>>2]|0;
      $281 = (FUNCTION_TABLE_iiiii[$274 & 8191]($2,1,$277,$280)|0);
      HEAP32[$error>>2] = $281;
      $282 = ($281|0)==(0);
      if ($282) {
       $283 = HEAP32[$36>>2]|0;
       $284 = (FUNCTION_TABLE_iiiii[$283 & 8191]($1,0,594602,8)|0);
       HEAP32[$error>>2] = $284;
       $285 = ($284|0)==(0);
       if ($285) {
        $286 = HEAP32[$41>>2]|0;
        $287 = (FUNCTION_TABLE_iiiii[$286 & 8191]($0,0,$notdef_glyph,5)|0);
        HEAP32[$error>>2] = $287;
        $288 = ($287|0)==(0);
        if ($288) {
         $289 = HEAP32[$36>>2]|0;
         $290 = ((($loader)) + 284|0);
         $291 = HEAP32[$290>>2]|0;
         $292 = HEAP32[$291>>2]|0;
         $293 = ((($loader)) + 288|0);
         $294 = HEAP32[$293>>2]|0;
         $295 = HEAP32[$294>>2]|0;
         $296 = (FUNCTION_TABLE_iiiii[$289 & 8191]($1,$n$0$ph144,$292,$295)|0);
         HEAP32[$error>>2] = $296;
         $297 = ($296|0)==(0);
         if ($297) {
          $298 = HEAP32[$41>>2]|0;
          $299 = HEAP32[$290>>2]|0;
          $300 = ((($299)) + 4|0);
          $301 = HEAP32[$300>>2]|0;
          $302 = HEAP32[$293>>2]|0;
          $303 = ((($302)) + 4|0);
          $304 = HEAP32[$303>>2]|0;
          $305 = (FUNCTION_TABLE_iiiii[$298 & 8191]($0,$n$0$ph144,$301,$304)|0);
          HEAP32[$error>>2] = $305;
          $306 = ($305|0)==(0);
          if ($306) {
           $307 = HEAP32[$17>>2]|0;
           $308 = (($307) + 1)|0;
           HEAP32[$17>>2] = $308;
           STACKTOP = sp;return;
          } else {
           $310 = $305;
          }
         } else {
          $310 = $296;
         }
        } else {
         $310 = $287;
        }
       } else {
        $310 = $284;
       }
      } else {
       $310 = $281;
      }
     } else {
      $310 = $272;
     }
     $$pre$phiZ2D = $14;$309 = $310;
     break;
    }
    $191 = HEAP32[$37>>2]|0;
    $192 = HEAP32[$191>>2]|0;
    $193 = (_strcmp(594602,$192)|0);
    $194 = ($193|0)==(0);
    if ($194) {
     STACKTOP = sp;return;
    }
    $195 = ((($loader)) + 304|0);
    $196 = HEAP32[$195>>2]|0;
    $197 = ((($loader)) + 184|0);
    $198 = HEAP32[$197>>2]|0;
    $199 = HEAP32[$198>>2]|0;
    $200 = (FUNCTION_TABLE_iiiii[$196 & 8191]($2,0,$192,$199)|0);
    HEAP32[$error>>2] = $200;
    $201 = ($200|0)==(0);
    if (!($201)) {
     $$pre$phiZ2D = $14;$309 = $200;
     break;
    }
    $202 = HEAP32[$195>>2]|0;
    $203 = ((($loader)) + 232|0);
    $204 = HEAP32[$203>>2]|0;
    $205 = HEAP32[$204>>2]|0;
    $206 = ((($loader)) + 236|0);
    $207 = HEAP32[$206>>2]|0;
    $208 = HEAP32[$207>>2]|0;
    $209 = (FUNCTION_TABLE_iiiii[$202 & 8191]($2,1,$205,$208)|0);
    HEAP32[$error>>2] = $209;
    $210 = ($209|0)==(0);
    if (!($210)) {
     $$pre$phiZ2D = $14;$309 = $209;
     break;
    }
    $211 = HEAP32[$195>>2]|0;
    $212 = HEAP32[$37>>2]|0;
    $213 = (($212) + ($notdef_index$0$ph412<<2)|0);
    $214 = HEAP32[$213>>2]|0;
    $215 = HEAP32[$197>>2]|0;
    $216 = (($215) + ($notdef_index$0$ph412<<2)|0);
    $217 = HEAP32[$216>>2]|0;
    $218 = (FUNCTION_TABLE_iiiii[$211 & 8191]($2,2,$214,$217)|0);
    HEAP32[$error>>2] = $218;
    $219 = ($218|0)==(0);
    if (!($219)) {
     $$pre$phiZ2D = $14;$309 = $218;
     break;
    }
    $220 = HEAP32[$195>>2]|0;
    $221 = HEAP32[$203>>2]|0;
    $222 = (($221) + ($notdef_index$0$ph412<<2)|0);
    $223 = HEAP32[$222>>2]|0;
    $224 = HEAP32[$206>>2]|0;
    $225 = (($224) + ($notdef_index$0$ph412<<2)|0);
    $226 = HEAP32[$225>>2]|0;
    $227 = (FUNCTION_TABLE_iiiii[$220 & 8191]($2,3,$223,$226)|0);
    HEAP32[$error>>2] = $227;
    $228 = ($227|0)==(0);
    if (!($228)) {
     $$pre$phiZ2D = $14;$309 = $227;
     break;
    }
    $229 = HEAP32[$36>>2]|0;
    $230 = ((($loader)) + 284|0);
    $231 = HEAP32[$230>>2]|0;
    $232 = HEAP32[$231>>2]|0;
    $233 = ((($loader)) + 288|0);
    $234 = HEAP32[$233>>2]|0;
    $235 = HEAP32[$234>>2]|0;
    $236 = (FUNCTION_TABLE_iiiii[$229 & 8191]($1,$notdef_index$0$ph412,$232,$235)|0);
    HEAP32[$error>>2] = $236;
    $237 = ($236|0)==(0);
    if (!($237)) {
     $$pre$phiZ2D = $14;$309 = $236;
     break;
    }
    $238 = HEAP32[$41>>2]|0;
    $239 = HEAP32[$230>>2]|0;
    $240 = ((($239)) + 4|0);
    $241 = HEAP32[$240>>2]|0;
    $242 = HEAP32[$233>>2]|0;
    $243 = ((($242)) + 4|0);
    $244 = HEAP32[$243>>2]|0;
    $245 = (FUNCTION_TABLE_iiiii[$238 & 8191]($0,$notdef_index$0$ph412,$241,$244)|0);
    HEAP32[$error>>2] = $245;
    $246 = ($245|0)==(0);
    if (!($246)) {
     $$pre$phiZ2D = $14;$309 = $245;
     break;
    }
    $247 = HEAP32[$36>>2]|0;
    $248 = HEAP32[$230>>2]|0;
    $249 = ((($248)) + 8|0);
    $250 = HEAP32[$249>>2]|0;
    $251 = HEAP32[$233>>2]|0;
    $252 = ((($251)) + 8|0);
    $253 = HEAP32[$252>>2]|0;
    $254 = (FUNCTION_TABLE_iiiii[$247 & 8191]($1,0,$250,$253)|0);
    HEAP32[$error>>2] = $254;
    $255 = ($254|0)==(0);
    if (!($255)) {
     $$pre$phiZ2D = $14;$309 = $254;
     break;
    }
    $256 = HEAP32[$41>>2]|0;
    $257 = HEAP32[$230>>2]|0;
    $258 = ((($257)) + 12|0);
    $259 = HEAP32[$258>>2]|0;
    $260 = HEAP32[$233>>2]|0;
    $261 = ((($260)) + 12|0);
    $262 = HEAP32[$261>>2]|0;
    $263 = (FUNCTION_TABLE_iiiii[$256 & 8191]($0,0,$259,$262)|0);
    HEAP32[$error>>2] = $263;
    $264 = ($263|0)==(0);
    if (!($264)) {
     $$pre$phiZ2D = $14;$309 = $263;
     break;
    }
    STACKTOP = sp;return;
   }
   else if ((label|0) == 79) {
    STACKTOP = sp;return;
   }
  }
 } while(0);
 HEAP32[$$pre$phiZ2D>>2] = $309;
 STACKTOP = sp;return;
}
function _parse_private($face,$loader) {
 $face = $face|0;
 $loader = $loader|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($loader)) + 372|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 | 1;
 HEAP32[$0>>2] = $2;
 return;
}
function _parse_blend_design_positions($face,$loader) {
 $face = $face|0;
 $loader = $loader|0;
 var $$off = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $axis$05 = 0, $axis_tokens = 0, $blend$06 = 0, $blend$1 = 0, $design_tokens = 0, $error$2$ph = 0, $error$4 = 0, $n$07 = 0, $n_axis = 0, $num_axis$09 = 0, $num_axis$1 = 0, $num_designs = 0, $or$cond = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $design_tokens = sp + 64|0;
 $num_designs = sp + 56|0;
 $axis_tokens = sp + 8|0;
 $n_axis = sp;
 $0 = ((($loader)) + 60|0);
 $1 = HEAP32[$0>>2]|0;
 FUNCTION_TABLE_viiii[$1 & 8191]($loader,$design_tokens,16,$num_designs);
 $2 = HEAP32[$num_designs>>2]|0;
 $3 = ($2|0)<(0);
 if ($3) {
  $error$4 = 162;
  $43 = ((($loader)) + 12|0);
  HEAP32[$43>>2] = $error$4;
  STACKTOP = sp;return;
 }
 $4 = ($2|0)==(0);
 $5 = ($2|0)>(16);
 $or$cond = $4 | $5;
 if ($or$cond) {
  $error$4 = 3;
  $43 = ((($loader)) + 12|0);
  HEAP32[$43>>2] = $error$4;
  STACKTOP = sp;return;
 }
 $6 = HEAP32[$loader>>2]|0;
 $7 = ((($loader)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($face)) + 528|0);
 $10 = ($2|0)>(0);
 L7: do {
  if ($10) {
   $11 = HEAP32[$9>>2]|0;
   $12 = ((($loader)) + 40|0);
   $blend$06 = $11;$n$07 = 0;$num_axis$09 = 0;
   while(1) {
    $13 = (($design_tokens) + (($n$07*12)|0)|0);
    $14 = HEAP32[$13>>2]|0;
    HEAP32[$loader>>2] = $14;
    $15 = (((($design_tokens) + (($n$07*12)|0)|0)) + 4|0);
    $16 = HEAP32[$15>>2]|0;
    HEAP32[$7>>2] = $16;
    $17 = HEAP32[$0>>2]|0;
    FUNCTION_TABLE_viiii[$17 & 8191]($loader,$axis_tokens,4,$n_axis);
    $18 = ($n$07|0)==(0);
    $19 = HEAP32[$n_axis>>2]|0;
    if ($18) {
     $$off = (($19) + -1)|0;
     $20 = ($$off>>>0)>(3);
     if ($20) {
      $error$2$ph = 3;
      break;
     }
     $21 = HEAP32[$num_designs>>2]|0;
     $22 = (_t1_allocate_blend($face,$21,$19)|0);
     $23 = ($22|0)==(0);
     if (!($23)) {
      $error$2$ph = $22;
      break;
     }
     $24 = HEAP32[$9>>2]|0;
     $$pre = HEAP32[$n_axis>>2]|0;
     $27 = $$pre;$blend$1 = $24;$num_axis$1 = $19;
    } else {
     $25 = ($19|0)==($num_axis$09|0);
     if ($25) {
      $27 = $num_axis$09;$blend$1 = $blend$06;$num_axis$1 = $num_axis$09;
     } else {
      $error$2$ph = 3;
      break;
     }
    }
    $26 = ($27|0)>(0);
    if ($26) {
     $28 = (((($blend$1)) + 24|0) + ($n$07<<2)|0);
     $axis$05 = 0;
     while(1) {
      $29 = (($axis_tokens) + (($axis$05*12)|0)|0);
      $30 = HEAP32[$29>>2]|0;
      HEAP32[$loader>>2] = $30;
      $31 = (((($axis_tokens) + (($axis$05*12)|0)|0)) + 4|0);
      $32 = HEAP32[$31>>2]|0;
      HEAP32[$7>>2] = $32;
      $33 = HEAP32[$12>>2]|0;
      $34 = (FUNCTION_TABLE_iii[$33 & 8191]($loader,0)|0);
      $35 = HEAP32[$28>>2]|0;
      $36 = (($35) + ($axis$05<<2)|0);
      HEAP32[$36>>2] = $34;
      $37 = (($axis$05) + 1)|0;
      $38 = HEAP32[$n_axis>>2]|0;
      $39 = ($37|0)<($38|0);
      if ($39) {
       $axis$05 = $37;
      } else {
       break;
      }
     }
    }
    $40 = (($n$07) + 1)|0;
    $41 = HEAP32[$num_designs>>2]|0;
    $42 = ($40|0)<($41|0);
    if ($42) {
     $blend$06 = $blend$1;$n$07 = $40;$num_axis$09 = $num_axis$1;
    } else {
     break L7;
    }
   }
   $error$4 = $error$2$ph;
   $43 = ((($loader)) + 12|0);
   HEAP32[$43>>2] = $error$4;
   STACKTOP = sp;return;
  }
 } while(0);
 HEAP32[$loader>>2] = $6;
 HEAP32[$7>>2] = $8;
 $error$4 = 0;
 $43 = ((($loader)) + 12|0);
 HEAP32[$43>>2] = $error$4;
 STACKTOP = sp;return;
}
function _t1_allocate_blend($face,$num_designs,$num_axis) {
 $face = $face|0;
 $num_designs = $num_designs|0;
 $num_axis = $num_axis|0;
 var $$phi$trans$insert = 0, $$pre = 0, $$pre10 = 0, $$pre11 = 0, $$pre13 = 0, $$pre14 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0, $blend$0 = 0;
 var $error = 0, $exitcond = 0, $exitcond15 = 0, $nn$07 = 0, $or$cond = 0, $or$cond3 = 0, $or$cond9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($face)) + 100|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$error>>2] = 0;
 $2 = ((($face)) + 528|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 do {
  if ($4) {
   $5 = (_ft_mem_alloc($1,420,$error)|0);
   $6 = HEAP32[$error>>2]|0;
   $7 = ($6|0)==(0);
   if ($7) {
    $8 = ((($5)) + 416|0);
    HEAP32[$8>>2] = 0;
    HEAP32[$2>>2] = $5;
    $blend$0 = $5;
    break;
   } else {
    $72 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($72|0);
   }
  } else {
   $blend$0 = $3;
  }
 } while(0);
 $9 = ($num_designs|0)==(0);
 do {
  if ($9) {
   label = 15;
  } else {
   $10 = HEAP32[$blend$0>>2]|0;
   $11 = ($10|0)==(0);
   if (!($11)) {
    $49 = ($10|0)==($num_designs|0);
    if ($49) {
     label = 15;
     break;
    } else {
     break;
    }
   }
   $12 = (_ft_mem_realloc($1,32,0,$num_designs,0,$error)|0);
   $13 = ((($blend$0)) + 148|0);
   HEAP32[$13>>2] = $12;
   $14 = HEAP32[$error>>2]|0;
   $15 = ($14|0)==(0);
   if (!($15)) {
    $72 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($72|0);
   }
   $16 = (_ft_mem_realloc($1,196,0,$num_designs,0,$error)|0);
   $17 = ((($blend$0)) + 216|0);
   HEAP32[$17>>2] = $16;
   $18 = HEAP32[$error>>2]|0;
   $19 = ($18|0)==(0);
   if (!($19)) {
    $72 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($72|0);
   }
   $20 = (_ft_mem_realloc($1,16,0,$num_designs,0,$error)|0);
   $21 = ((($blend$0)) + 288|0);
   HEAP32[$21>>2] = $20;
   $22 = HEAP32[$error>>2]|0;
   $23 = ($22|0)==(0);
   if (!($23)) {
    $72 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($72|0);
   }
   $24 = $num_designs << 1;
   $25 = (_ft_mem_realloc($1,4,0,$24,0,$error)|0);
   $26 = ((($blend$0)) + 136|0);
   HEAP32[$26>>2] = $25;
   $27 = HEAP32[$error>>2]|0;
   $28 = ($27|0)==(0);
   if (!($28)) {
    $72 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($72|0);
   }
   $29 = (($25) + ($num_designs<<2)|0);
   $30 = ((($blend$0)) + 140|0);
   HEAP32[$30>>2] = $29;
   $31 = ((($face)) + 132|0);
   $32 = ((($blend$0)) + 144|0);
   HEAP32[$32>>2] = $31;
   $33 = ((($face)) + 168|0);
   $34 = ((($blend$0)) + 212|0);
   HEAP32[$34>>2] = $33;
   $35 = ((($face)) + 460|0);
   $36 = ((($blend$0)) + 284|0);
   HEAP32[$36>>2] = $35;
   $37 = ($num_designs>>>0)<(2);
   if (!($37)) {
    $$pre = HEAP32[$13>>2]|0;
    $$pre10 = HEAP32[$17>>2]|0;
    $$pre11 = HEAP32[$21>>2]|0;
    $39 = $$pre;$42 = $$pre10;$45 = $$pre11;$nn$07 = 2;
    while(1) {
     $38 = ((($39)) + 32|0);
     $40 = (((($blend$0)) + 144|0) + ($nn$07<<2)|0);
     HEAP32[$40>>2] = $38;
     $41 = ((($42)) + 196|0);
     $43 = (((($blend$0)) + 212|0) + ($nn$07<<2)|0);
     HEAP32[$43>>2] = $41;
     $44 = ((($45)) + 16|0);
     $46 = (((($blend$0)) + 284|0) + ($nn$07<<2)|0);
     HEAP32[$46>>2] = $44;
     $47 = (($nn$07) + 1)|0;
     $48 = ($47>>>0)>($num_designs>>>0);
     if ($48) {
      break;
     } else {
      $39 = $38;$42 = $41;$45 = $44;$nn$07 = $47;
     }
    }
   }
   HEAP32[$blend$0>>2] = $num_designs;
   label = 15;
  }
 } while(0);
 do {
  if ((label|0) == 15) {
   $50 = ($num_axis|0)==(0);
   $$phi$trans$insert = ((($blend$0)) + 4|0);
   $$pre13 = HEAP32[$$phi$trans$insert>>2]|0;
   if ($50) {
    $56 = $$pre13;
   } else {
    $51 = ($$pre13|0)==(0);
    $52 = ($$pre13|0)==($num_axis|0);
    $or$cond3 = $51 | $52;
    if (!($or$cond3)) {
     break;
    }
    HEAP32[$$phi$trans$insert>>2] = $num_axis;
    $56 = $num_axis;
   }
   $53 = HEAP32[$blend$0>>2]|0;
   $54 = ($53|0)!=(0);
   $55 = ($56|0)!=(0);
   $or$cond = $54 & $55;
   if (!($or$cond)) {
    $72 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($72|0);
   }
   $57 = ((($blend$0)) + 24|0);
   $58 = HEAP32[$57>>2]|0;
   $59 = ($58|0)==(0|0);
   if (!($59)) {
    $72 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($72|0);
   }
   $60 = Math_imul($56, $53)|0;
   $61 = (_ft_mem_realloc($1,4,0,$60,0,$error)|0);
   HEAP32[$57>>2] = $61;
   $62 = HEAP32[$error>>2]|0;
   $63 = ($62|0)==(0);
   $64 = ($53>>>0)>(1);
   $or$cond9 = $63 & $64;
   if (!($or$cond9)) {
    $72 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($72|0);
   }
   $65 = (($61) + ($56<<2)|0);
   $66 = ((($blend$0)) + 28|0);
   HEAP32[$66>>2] = $65;
   $exitcond15 = ($53|0)==(2);
   if ($exitcond15) {
    $72 = HEAP32[$error>>2]|0;
    STACKTOP = sp;return ($72|0);
   } else {
    $68 = 2;
   }
   while(1) {
    $$pre14 = HEAP32[$57>>2]|0;
    $67 = Math_imul($68, $56)|0;
    $69 = (($$pre14) + ($67<<2)|0);
    $70 = (((($blend$0)) + 24|0) + ($68<<2)|0);
    HEAP32[$70>>2] = $69;
    $71 = (($68) + 1)|0;
    $exitcond = ($71|0)==($53|0);
    if ($exitcond) {
     break;
    } else {
     $68 = $71;
    }
   }
   $72 = HEAP32[$error>>2]|0;
   STACKTOP = sp;return ($72|0);
  }
 } while(0);
 HEAP32[$error>>2] = 3;
 $72 = HEAP32[$error>>2]|0;
 STACKTOP = sp;return ($72|0);
}
function _parse_blend_design_map($face,$loader) {
 $face = $face|0;
 $loader = $loader|0;
 var $$lcssa = 0, $$off = 0, $$pre = 0, $$pre$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $axis_tokens = 0, $error = 0, $n$09 = 0, $num_axis = 0, $num_points = 0, $or$cond = 0, $p$08 = 0, $point_tokens = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 320|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp + 304|0;
 $axis_tokens = sp + 256|0;
 $num_axis = sp + 248|0;
 $point_tokens = sp + 8|0;
 $num_points = sp;
 HEAP32[$error>>2] = 0;
 $0 = ((($face)) + 100|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($loader)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 FUNCTION_TABLE_viiii[$3 & 8191]($loader,$axis_tokens,4,$num_axis);
 $4 = HEAP32[$num_axis>>2]|0;
 $5 = ($4|0)<(0);
 if ($5) {
  HEAP32[$error>>2] = 162;
  $58 = 162;
  $57 = ((($loader)) + 12|0);
  HEAP32[$57>>2] = $58;
  STACKTOP = sp;return;
 }
 $6 = ($4|0)==(0);
 $7 = ($4|0)>(4);
 $or$cond = $6 | $7;
 if ($or$cond) {
  HEAP32[$error>>2] = 3;
  $58 = 3;
  $57 = ((($loader)) + 12|0);
  HEAP32[$57>>2] = $58;
  STACKTOP = sp;return;
 }
 $8 = HEAP32[$loader>>2]|0;
 $9 = ((($loader)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (_t1_allocate_blend($face,0,$4)|0);
 HEAP32[$error>>2] = $11;
 $12 = ($11|0)==(0);
 if (!($12)) {
  $58 = $11;
  $57 = ((($loader)) + 12|0);
  HEAP32[$57>>2] = $58;
  STACKTOP = sp;return;
 }
 $13 = ((($face)) + 528|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = HEAP32[$num_axis>>2]|0;
 $16 = ($15|0)>(0);
 do {
  if ($16) {
   $17 = ((($loader)) + 36|0);
   $18 = ((($loader)) + 40|0);
   $n$09 = 0;
   while(1) {
    $19 = (($axis_tokens) + (($n$09*12)|0)|0);
    $20 = HEAP32[$19>>2]|0;
    HEAP32[$loader>>2] = $20;
    $21 = (((($axis_tokens) + (($n$09*12)|0)|0)) + 4|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$9>>2] = $22;
    $23 = HEAP32[$2>>2]|0;
    FUNCTION_TABLE_viiii[$23 & 8191]($loader,$point_tokens,20,$num_points);
    $24 = HEAP32[$num_points>>2]|0;
    $$off = (($24) + -1)|0;
    $25 = ($$off>>>0)>(19);
    if ($25) {
     label = 9;
     break;
    }
    $26 = $24 << 1;
    $27 = (_ft_mem_realloc($1,4,0,$26,0,$error)|0);
    $28 = (((((($14)) + 88|0) + (($n$09*12)|0)|0)) + 4|0);
    HEAP32[$28>>2] = $27;
    $29 = HEAP32[$error>>2]|0;
    $30 = ($29|0)==(0);
    if (!($30)) {
     $$lcssa = $29;
     label = 16;
     break;
    }
    $31 = HEAP32[$num_points>>2]|0;
    $32 = (($27) + ($31<<2)|0);
    $33 = (((((($14)) + 88|0) + (($n$09*12)|0)|0)) + 8|0);
    HEAP32[$33>>2] = $32;
    $34 = $31&255;
    $35 = (((($14)) + 88|0) + (($n$09*12)|0)|0);
    HEAP8[$35>>0] = $34;
    $36 = ($31|0)>(0);
    if ($36) {
     $p$08 = 0;
     while(1) {
      $37 = (($point_tokens) + (($p$08*12)|0)|0);
      $38 = HEAP32[$37>>2]|0;
      $39 = ((($38)) + 1|0);
      HEAP32[$loader>>2] = $39;
      $40 = (((($point_tokens) + (($p$08*12)|0)|0)) + 4|0);
      $41 = HEAP32[$40>>2]|0;
      $42 = ((($41)) + -1|0);
      HEAP32[$9>>2] = $42;
      $43 = HEAP32[$17>>2]|0;
      $44 = (FUNCTION_TABLE_ii[$43 & 8191]($loader)|0);
      $45 = HEAP32[$28>>2]|0;
      $46 = (($45) + ($p$08<<2)|0);
      HEAP32[$46>>2] = $44;
      $47 = HEAP32[$18>>2]|0;
      $48 = (FUNCTION_TABLE_iii[$47 & 8191]($loader,0)|0);
      $49 = HEAP32[$33>>2]|0;
      $50 = (($49) + ($p$08<<2)|0);
      HEAP32[$50>>2] = $48;
      $51 = (($p$08) + 1)|0;
      $52 = HEAP32[$num_points>>2]|0;
      $53 = ($51|0)<($52|0);
      if ($53) {
       $p$08 = $51;
      } else {
       break;
      }
     }
    }
    $54 = (($n$09) + 1)|0;
    $55 = HEAP32[$num_axis>>2]|0;
    $56 = ($54|0)<($55|0);
    if ($56) {
     $n$09 = $54;
    } else {
     label = 14;
     break;
    }
   }
   if ((label|0) == 9) {
    HEAP32[$error>>2] = 3;
    $58 = 3;
    $57 = ((($loader)) + 12|0);
    HEAP32[$57>>2] = $58;
    STACKTOP = sp;return;
   }
   else if ((label|0) == 14) {
    $$pre$pre = HEAP32[$error>>2]|0;
    $$pre = $$pre$pre;
    break;
   }
   else if ((label|0) == 16) {
    $58 = $$lcssa;
    $57 = ((($loader)) + 12|0);
    HEAP32[$57>>2] = $58;
    STACKTOP = sp;return;
   }
  } else {
   $$pre = 0;
  }
 } while(0);
 HEAP32[$loader>>2] = $8;
 HEAP32[$9>>2] = $10;
 $58 = $$pre;
 $57 = ((($loader)) + 12|0);
 HEAP32[$57>>2] = $58;
 STACKTOP = sp;return;
}
function _parse_blend_axis_types($face,$loader) {
 $face = $face|0;
 $loader = $loader|0;
 var $$in = 0, $$pre$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $axis_tokens = 0, $error = 0;
 var $n$03 = 0, $num_axis = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $axis_tokens = sp + 8|0;
 $num_axis = sp + 4|0;
 $error = sp;
 HEAP32[$error>>2] = 0;
 $0 = ((($loader)) + 60|0);
 $1 = HEAP32[$0>>2]|0;
 FUNCTION_TABLE_viiii[$1 & 8191]($loader,$axis_tokens,4,$num_axis);
 $2 = HEAP32[$num_axis>>2]|0;
 $3 = ($2|0)<(0);
 L1: do {
  if ($3) {
   HEAP32[$error>>2] = 162;
   $35 = 162;
  } else {
   $4 = ($2|0)==(0);
   $5 = ($2|0)>(4);
   $or$cond = $4 | $5;
   if ($or$cond) {
    HEAP32[$error>>2] = 3;
    $35 = 3;
    break;
   }
   $6 = (_t1_allocate_blend($face,0,$2)|0);
   HEAP32[$error>>2] = $6;
   $7 = ($6|0)==(0);
   if ($7) {
    $8 = ((($face)) + 528|0);
    $9 = HEAP32[$8>>2]|0;
    $10 = ((($face)) + 100|0);
    $11 = HEAP32[$10>>2]|0;
    $12 = HEAP32[$num_axis>>2]|0;
    $13 = ($12|0)>(0);
    if ($13) {
     $n$03 = 0;
     while(1) {
      $14 = (($axis_tokens) + (($n$03*12)|0)|0);
      $15 = HEAP32[$14>>2]|0;
      $16 = HEAP8[$15>>0]|0;
      $17 = ($16<<24>>24)==(47);
      if ($17) {
       $18 = ((($15)) + 1|0);
       HEAP32[$14>>2] = $18;
       $$in = $18;
      } else {
       $$in = $15;
      }
      $19 = $$in;
      $20 = (((($axis_tokens) + (($n$03*12)|0)|0)) + 4|0);
      $21 = HEAP32[$20>>2]|0;
      $22 = (($21) - ($19))|0;
      $23 = ($21|0)==($19|0);
      if ($23) {
       label = 10;
       break;
      }
      $24 = (($22) + 1)|0;
      $25 = (_ft_mem_alloc($11,$24,$error)|0);
      $26 = (((($9)) + 8|0) + ($n$03<<2)|0);
      HEAP32[$26>>2] = $25;
      $27 = HEAP32[$error>>2]|0;
      $28 = ($27|0)==(0);
      if (!($28)) {
       $35 = $27;
       break L1;
      }
      $29 = HEAP32[$14>>2]|0;
      _memcpy(($25|0),($29|0),($22|0))|0;
      $30 = (($25) + ($22)|0);
      HEAP8[$30>>0] = 0;
      $31 = (($n$03) + 1)|0;
      $32 = HEAP32[$num_axis>>2]|0;
      $33 = ($31|0)<($32|0);
      if ($33) {
       $n$03 = $31;
      } else {
       label = 13;
       break;
      }
     }
     if ((label|0) == 10) {
      HEAP32[$error>>2] = 3;
      $35 = 3;
      break;
     }
     else if ((label|0) == 13) {
      $$pre$pre = HEAP32[$error>>2]|0;
      $35 = $$pre$pre;
      break;
     }
    } else {
     $35 = 0;
    }
   } else {
    $35 = $6;
   }
  }
 } while(0);
 $34 = ((($loader)) + 12|0);
 HEAP32[$34>>2] = $35;
 STACKTOP = sp;return;
}
function _parse_weight_vector($face,$loader) {
 $face = $face|0;
 $loader = $loader|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $blend$0 = 0, $design_tokens = 0;
 var $error$1 = 0, $n$02 = 0, $num_designs = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 208|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $design_tokens = sp + 8|0;
 $num_designs = sp;
 $0 = ((($face)) + 528|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($loader)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 FUNCTION_TABLE_viiii[$3 & 8191]($loader,$design_tokens,16,$num_designs);
 $4 = HEAP32[$num_designs>>2]|0;
 $5 = ($4|0)<(0);
 do {
  if ($5) {
   $error$1 = 162;
  } else {
   $6 = ($4|0)==(0);
   $7 = ($4|0)>(16);
   $or$cond = $6 | $7;
   if ($or$cond) {
    $error$1 = 3;
   } else {
    $8 = ($1|0)==(0|0);
    if ($8) {
     label = 5;
    } else {
     $9 = HEAP32[$1>>2]|0;
     $10 = ($9|0)==(0);
     if ($10) {
      label = 5;
     } else {
      $14 = ($9|0)==($4|0);
      if ($14) {
       $19 = $4;$blend$0 = $1;
      } else {
       $error$1 = 3;
       break;
      }
     }
    }
    if ((label|0) == 5) {
     $11 = (_t1_allocate_blend($face,$4,0)|0);
     $12 = ($11|0)==(0);
     if (!($12)) {
      $error$1 = $11;
      break;
     }
     $13 = HEAP32[$0>>2]|0;
     $$pre = HEAP32[$num_designs>>2]|0;
     $19 = $$pre;$blend$0 = $13;
    }
    $15 = HEAP32[$loader>>2]|0;
    $16 = ((($loader)) + 8|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = ($19|0)>(0);
    if ($18) {
     $20 = ((($loader)) + 40|0);
     $21 = ((($blend$0)) + 136|0);
     $22 = ((($blend$0)) + 140|0);
     $n$02 = 0;
     while(1) {
      $23 = (($design_tokens) + (($n$02*12)|0)|0);
      $24 = HEAP32[$23>>2]|0;
      HEAP32[$loader>>2] = $24;
      $25 = (((($design_tokens) + (($n$02*12)|0)|0)) + 4|0);
      $26 = HEAP32[$25>>2]|0;
      HEAP32[$16>>2] = $26;
      $27 = HEAP32[$20>>2]|0;
      $28 = (FUNCTION_TABLE_iii[$27 & 8191]($loader,0)|0);
      $29 = HEAP32[$21>>2]|0;
      $30 = (($29) + ($n$02<<2)|0);
      HEAP32[$30>>2] = $28;
      $31 = HEAP32[$22>>2]|0;
      $32 = (($31) + ($n$02<<2)|0);
      HEAP32[$32>>2] = $28;
      $33 = (($n$02) + 1)|0;
      $34 = HEAP32[$num_designs>>2]|0;
      $35 = ($33|0)<($34|0);
      if ($35) {
       $n$02 = $33;
      } else {
       break;
      }
     }
    }
    HEAP32[$loader>>2] = $15;
    HEAP32[$16>>2] = $17;
    $error$1 = 0;
   }
  }
 } while(0);
 $36 = ((($loader)) + 12|0);
 HEAP32[$36>>2] = $error$1;
 STACKTOP = sp;return;
}
function _parse_buildchar($face,$loader) {
 $face = $face|0;
 $loader = $loader|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($loader)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (FUNCTION_TABLE_iiiii[$1 & 8191]($loader,0,0,0)|0);
 $3 = ((($face)) + 540|0);
 HEAP32[$3>>2] = $2;
 return;
}
function _T1_Done_Blend($face) {
 $face = $face|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $exitcond13 = 0, $exitcond14 = 0, $n$15 = 0;
 var $n$22 = 0, $n$31 = 0, $scevgep = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 100|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 528|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  return;
 }
 $5 = HEAP32[$3>>2]|0;
 $6 = ((($3)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($3)) + 24|0);
 $9 = HEAP32[$8>>2]|0;
 _ft_mem_free($1,$9);
 HEAP32[$8>>2] = 0;
 $10 = ($5>>>0)>(1);
 if ($10) {
  $scevgep = ((($3)) + 28|0);
  $11 = $5 << 2;
  $12 = (($11) + -4)|0;
  _memset(($scevgep|0),0,($12|0))|0;
 }
 $13 = ((($3)) + 216|0);
 $14 = HEAP32[$13>>2]|0;
 _ft_mem_free($1,$14);
 HEAP32[$13>>2] = 0;
 $15 = ((($3)) + 148|0);
 $16 = HEAP32[$15>>2]|0;
 _ft_mem_free($1,$16);
 HEAP32[$15>>2] = 0;
 $17 = ((($3)) + 288|0);
 $18 = HEAP32[$17>>2]|0;
 _ft_mem_free($1,$18);
 HEAP32[$17>>2] = 0;
 $19 = ($5|0)==(0);
 if (!($19)) {
  $n$15 = 0;
  while(1) {
   $20 = (((($3)) + 212|0) + ($n$15<<2)|0);
   HEAP32[$20>>2] = 0;
   $21 = (((($3)) + 144|0) + ($n$15<<2)|0);
   HEAP32[$21>>2] = 0;
   $22 = (((($3)) + 284|0) + ($n$15<<2)|0);
   HEAP32[$22>>2] = 0;
   $23 = (($n$15) + 1)|0;
   $exitcond14 = ($23|0)==($5|0);
   if ($exitcond14) {
    break;
   } else {
    $n$15 = $23;
   }
  }
 }
 $24 = ((($3)) + 136|0);
 $25 = HEAP32[$24>>2]|0;
 _ft_mem_free($1,$25);
 HEAP32[$24>>2] = 0;
 $26 = ((($3)) + 140|0);
 HEAP32[$26>>2] = 0;
 $27 = ($7|0)==(0);
 if (!($27)) {
  $n$22 = 0;
  while(1) {
   $28 = (((($3)) + 8|0) + ($n$22<<2)|0);
   $29 = HEAP32[$28>>2]|0;
   _ft_mem_free($1,$29);
   HEAP32[$28>>2] = 0;
   $30 = (($n$22) + 1)|0;
   $exitcond13 = ($30|0)==($7|0);
   if ($exitcond13) {
    break;
   } else {
    $n$22 = $30;
   }
  }
  if (!($27)) {
   $n$31 = 0;
   while(1) {
    $31 = (((((($3)) + 88|0) + (($n$31*12)|0)|0)) + 4|0);
    $32 = HEAP32[$31>>2]|0;
    _ft_mem_free($1,$32);
    HEAP32[$31>>2] = 0;
    $33 = (((($3)) + 88|0) + (($n$31*12)|0)|0);
    HEAP8[$33>>0] = 0;
    $34 = (($n$31) + 1)|0;
    $exitcond = ($34|0)==($7|0);
    if ($exitcond) {
     break;
    } else {
     $n$31 = $34;
    }
   }
  }
 }
 $35 = HEAP32[$2>>2]|0;
 _ft_mem_free($1,$35);
 HEAP32[$2>>2] = 0;
 return;
}
function _T1_Parse_Glyph($decoder,$glyph_index) {
 $decoder = $decoder|0;
 $glyph_index = $glyph_index|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $glyph_data = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $glyph_data = sp;
 $0 = (_T1_Parse_Glyph_And_Get_Char_String($decoder,$glyph_index,$glyph_data)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  STACKTOP = sp;return ($0|0);
 }
 $2 = ((($decoder)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 128|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 48|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 if ($8) {
  STACKTOP = sp;return ($0|0);
 }
 $9 = HEAP32[$7>>2]|0;
 $10 = ((($9)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($7)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 FUNCTION_TABLE_vii[$11 & 8191]($13,$glyph_data);
 STACKTOP = sp;return ($0|0);
}
function _T1_Parse_Glyph_And_Get_Char_String($decoder,$glyph_index,$char_string) {
 $decoder = $decoder|0;
 $glyph_index = $glyph_index|0;
 $char_string = $char_string|0;
 var $$phi$trans$insert4 = 0, $$pre = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $8 = 0, $9 = 0, $error$2 = 0, $metrics = 0, $or$cond = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $metrics = sp;
 $0 = ((($decoder)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 128|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 48|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($decoder)) + 1376|0);
 $7 = ((($1)) + 436|0);
 ;HEAP32[$6>>2]=HEAP32[$7>>2]|0;HEAP32[$6+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$7+8>>2]|0;HEAP32[$6+12>>2]=HEAP32[$7+12>>2]|0;
 $8 = ((($decoder)) + 1392|0);
 $9 = ((($1)) + 452|0);
 $10 = $9;
 $11 = $10;
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10) + 4)|0;
 $14 = $13;
 $15 = HEAP32[$14>>2]|0;
 $16 = $8;
 $17 = $16;
 HEAP32[$17>>2] = $12;
 $18 = (($16) + 4)|0;
 $19 = $18;
 HEAP32[$19>>2] = $15;
 $20 = ($5|0)!=(0|0);
 do {
  if ($20) {
   $31 = HEAP32[$5>>2]|0;
   $32 = HEAP32[$31>>2]|0;
   $33 = ((($5)) + 4|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = (FUNCTION_TABLE_iiii[$32 & 8191]($34,$glyph_index,$char_string)|0);
   $36 = ($35|0)==(0);
   if ($36) {
    $$pre = HEAP32[$char_string>>2]|0;
    $$phi$trans$insert4 = ((($char_string)) + 4|0);
    $$pre5 = HEAP32[$$phi$trans$insert4>>2]|0;
    $39 = $$pre;$40 = $$pre5;
    break;
   } else {
    $error$2 = $35;
    STACKTOP = sp;return ($error$2|0);
   }
  } else {
   $21 = ((($1)) + 424|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = (($22) + ($glyph_index<<2)|0);
   $24 = HEAP32[$23>>2]|0;
   HEAP32[$char_string>>2] = $24;
   $25 = ((($1)) + 428|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = (($26) + ($glyph_index<<2)|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = ((($char_string)) + 4|0);
   HEAP32[$29>>2] = $28;
   $30 = $24;
   $39 = $30;$40 = $28;
  }
 } while(0);
 $37 = ((($decoder)) + 1484|0);
 $38 = HEAP32[$37>>2]|0;
 $41 = (FUNCTION_TABLE_iiii[$38 & 8191]($decoder,$39,$40)|0);
 $42 = ($41|0)==(0);
 $or$cond = $20 & $42;
 if (!($or$cond)) {
  $error$2 = $41;
  STACKTOP = sp;return ($error$2|0);
 }
 $43 = HEAP32[$5>>2]|0;
 $44 = ((($43)) + 8|0);
 $45 = HEAP32[$44>>2]|0;
 $46 = ($45|0)==(0|0);
 if ($46) {
  $error$2 = 0;
  STACKTOP = sp;return ($error$2|0);
 }
 $47 = ((($decoder)) + 32|0);
 $48 = HEAP32[$47>>2]|0;
 $49 = (_FT_RoundFix($48)|0);
 $50 = $49 >> 16;
 HEAP32[$metrics>>2] = $50;
 $51 = ((($metrics)) + 4|0);
 HEAP32[$51>>2] = 0;
 $52 = ((($decoder)) + 40|0);
 $53 = HEAP32[$52>>2]|0;
 $54 = (_FT_RoundFix($53)|0);
 $55 = $54 >> 16;
 $56 = ((($metrics)) + 8|0);
 HEAP32[$56>>2] = $55;
 $57 = ((($decoder)) + 44|0);
 $58 = HEAP32[$57>>2]|0;
 $59 = (_FT_RoundFix($58)|0);
 $60 = $59 >> 16;
 $61 = ((($metrics)) + 12|0);
 HEAP32[$61>>2] = $60;
 $62 = HEAP32[$5>>2]|0;
 $63 = ((($62)) + 8|0);
 $64 = HEAP32[$63>>2]|0;
 $65 = ((($5)) + 4|0);
 $66 = HEAP32[$65>>2]|0;
 $67 = (FUNCTION_TABLE_iiiii[$64 & 8191]($66,$glyph_index,0,$metrics)|0);
 $68 = HEAP32[$metrics>>2]|0;
 $69 = $68 << 16;
 HEAP32[$47>>2] = $69;
 $70 = HEAP32[$56>>2]|0;
 $71 = $70 << 16;
 HEAP32[$52>>2] = $71;
 $72 = HEAP32[$61>>2]|0;
 $73 = $72 << 16;
 HEAP32[$57>>2] = $73;
 $error$2 = $67;
 STACKTOP = sp;return ($error$2|0);
}
function _T1_Face_Done($t1face) {
 $t1face = $t1face|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($t1face|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($t1face)) + 100|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($t1face)) + 132|0);
 $4 = ((($t1face)) + 544|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if (!($6)) {
  _ft_mem_free($2,$5);
  HEAP32[$4>>2] = 0;
  $7 = ((($t1face)) + 540|0);
  HEAP32[$7>>2] = 0;
 }
 _T1_Done_Blend($t1face);
 $8 = ((($t1face)) + 528|0);
 HEAP32[$8>>2] = 0;
 $9 = HEAP32[$3>>2]|0;
 _ft_mem_free($2,$9);
 HEAP32[$3>>2] = 0;
 $10 = ((($t1face)) + 136|0);
 $11 = HEAP32[$10>>2]|0;
 _ft_mem_free($2,$11);
 HEAP32[$10>>2] = 0;
 $12 = ((($t1face)) + 140|0);
 $13 = HEAP32[$12>>2]|0;
 _ft_mem_free($2,$13);
 HEAP32[$12>>2] = 0;
 $14 = ((($t1face)) + 144|0);
 $15 = HEAP32[$14>>2]|0;
 _ft_mem_free($2,$15);
 HEAP32[$14>>2] = 0;
 $16 = ((($t1face)) + 148|0);
 $17 = HEAP32[$16>>2]|0;
 _ft_mem_free($2,$17);
 HEAP32[$16>>2] = 0;
 $18 = ((($t1face)) + 428|0);
 $19 = HEAP32[$18>>2]|0;
 _ft_mem_free($2,$19);
 HEAP32[$18>>2] = 0;
 $20 = ((($t1face)) + 424|0);
 $21 = HEAP32[$20>>2]|0;
 _ft_mem_free($2,$21);
 HEAP32[$20>>2] = 0;
 $22 = ((($t1face)) + 420|0);
 $23 = HEAP32[$22>>2]|0;
 _ft_mem_free($2,$23);
 HEAP32[$22>>2] = 0;
 $24 = ((($t1face)) + 408|0);
 $25 = HEAP32[$24>>2]|0;
 _ft_mem_free($2,$25);
 HEAP32[$24>>2] = 0;
 $26 = ((($t1face)) + 412|0);
 $27 = HEAP32[$26>>2]|0;
 _ft_mem_free($2,$27);
 HEAP32[$26>>2] = 0;
 $28 = ((($t1face)) + 392|0);
 $29 = HEAP32[$28>>2]|0;
 _ft_mem_free($2,$29);
 HEAP32[$28>>2] = 0;
 $30 = ((($t1face)) + 396|0);
 $31 = HEAP32[$30>>2]|0;
 _ft_mem_free($2,$31);
 HEAP32[$30>>2] = 0;
 $32 = ((($t1face)) + 400|0);
 $33 = HEAP32[$32>>2]|0;
 _ft_mem_free($2,$33);
 HEAP32[$32>>2] = 0;
 $34 = ((($t1face)) + 372|0);
 $35 = ((($34)) + 12|0);
 $36 = HEAP32[$35>>2]|0;
 _ft_mem_free($2,$36);
 HEAP32[$35>>2] = 0;
 $37 = ((($34)) + 16|0);
 $38 = HEAP32[$37>>2]|0;
 _ft_mem_free($2,$38);
 HEAP32[$37>>2] = 0;
 $39 = ((($t1face)) + 364|0);
 $40 = HEAP32[$39>>2]|0;
 _ft_mem_free($2,$40);
 HEAP32[$39>>2] = 0;
 $41 = ((($t1face)) + 492|0);
 $42 = HEAP32[$41>>2]|0;
 $43 = ($42|0)==(0|0);
 if (!($43)) {
  $44 = ((($42)) + 36|0);
  $45 = HEAP32[$44>>2]|0;
  _ft_mem_free($2,$45);
  HEAP32[$44>>2] = 0;
  $46 = ((($42)) + 40|0);
  HEAP32[$46>>2] = 0;
  $47 = ((($42)) + 28|0);
  $48 = HEAP32[$47>>2]|0;
  _ft_mem_free($2,$48);
  HEAP32[$47>>2] = 0;
  $49 = ((($42)) + 32|0);
  HEAP32[$49>>2] = 0;
  _ft_mem_free($2,$42);
 }
 $50 = ((($t1face)) + 20|0);
 HEAP32[$50>>2] = 0;
 $51 = ((($t1face)) + 24|0);
 HEAP32[$51>>2] = 0;
 return;
}
function _T1_Size_Init($t1size) {
 $t1size = $t1size|0;
 var $$idx$val = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $error$0 = 0, $globals = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $globals = sp;
 $$idx$val = HEAP32[$t1size>>2]|0;
 $0 = ((($$idx$val)) + 548|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($$idx$val)) + 96|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (_FT_Get_Module($5,593918)|0);
 $7 = ($6|0)!=(0|0);
 $8 = ($1|0)!=(0|0);
 $or$cond$i = $8 & $7;
 if (!($or$cond$i)) {
  $error$0 = 0;
  STACKTOP = sp;return ($error$0|0);
 }
 $9 = HEAP32[$1>>2]|0;
 $10 = ($9|0)==(0|0);
 if ($10) {
  $error$0 = 0;
  STACKTOP = sp;return ($error$0|0);
 }
 $11 = (FUNCTION_TABLE_ii[$9 & 8191]($6)|0);
 $12 = ($11|0)==(0|0);
 if ($12) {
  $error$0 = 0;
  STACKTOP = sp;return ($error$0|0);
 }
 $13 = HEAP32[$t1size>>2]|0;
 $14 = HEAP32[$11>>2]|0;
 $15 = ((($13)) + 100|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($13)) + 168|0);
 $18 = (FUNCTION_TABLE_iiii[$14 & 8191]($16,$17,$globals)|0);
 $19 = ($18|0)==(0);
 if ($19) {
  $20 = HEAP32[$globals>>2]|0;
  $21 = ((($t1size)) + 40|0);
  HEAP32[$21>>2] = $20;
 }
 $error$0 = $18;
 STACKTOP = sp;return ($error$0|0);
}
function _T1_Size_Done($t1size) {
 $t1size = $t1size|0;
 var $$idx$val = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($t1size)) + 40|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 $$idx$val = HEAP32[$t1size>>2]|0;
 $3 = ((($$idx$val)) + 548|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($$idx$val)) + 96|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($6)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (_FT_Get_Module($8,593918)|0);
 $10 = ($9|0)!=(0|0);
 $11 = ($4|0)!=(0|0);
 $or$cond$i = $11 & $10;
 if ($or$cond$i) {
  $12 = HEAP32[$4>>2]|0;
  $13 = ($12|0)==(0|0);
  if (!($13)) {
   $14 = (FUNCTION_TABLE_ii[$12 & 8191]($9)|0);
   $15 = ($14|0)==(0|0);
   if (!($15)) {
    $16 = ((($14)) + 8|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = HEAP32[$0>>2]|0;
    FUNCTION_TABLE_vi[$17 & 8191]($18);
   }
  }
 }
 HEAP32[$0>>2] = 0;
 return;
}
function _T1_GlyphSlot_Init($slot) {
 $slot = $slot|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($slot)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 548|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  return 0;
 }
 $5 = ((($1)) + 96|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($6)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (_FT_Get_Module($8,593918)|0);
 $10 = ($9|0)==(0|0);
 if ($10) {
  return 0;
 }
 $11 = ((($3)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (FUNCTION_TABLE_ii[$12 & 8191]($9)|0);
 $14 = ((($slot)) + 156|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($15)) + 36|0);
 HEAP32[$16>>2] = $13;
 return 0;
}
function _T1_GlyphSlot_Done($slot) {
 $slot = $slot|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($slot)) + 156|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 36|0);
 HEAP32[$2>>2] = 0;
 return;
}
function _T1_Load_Glyph($t1glyph,$t1size,$glyph_index,$load_flags) {
 $t1glyph = $t1glyph|0;
 $t1size = $t1size|0;
 $glyph_index = $glyph_index|0;
 $load_flags = $load_flags|0;
 var $$lobit = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $advance = 0, $cbox = 0, $decoder = 0, $error$025 = 0, $font_matrix = 0, $glyph_data = 0, $load_flags$ = 0, $n$06 = 0, $vec$07 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1552|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $decoder = sp + 48|0;
 $font_matrix = sp + 32|0;
 $glyph_data = sp + 24|0;
 $cbox = sp + 8|0;
 $advance = sp;
 $0 = ((($t1glyph)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 488|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($1)) + 16|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7>>>0)>($glyph_index>>>0);
 if (!($8)) {
  $9 = ((($1)) + 128|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ((($10)) + 48|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ($12|0)==(0|0);
  if ($13) {
   $error$025 = 6;
   STACKTOP = sp;return ($error$025|0);
  }
 }
 $14 = $load_flags & 1024;
 $15 = ($14|0)==(0);
 $16 = $load_flags | 3;
 $load_flags$ = $15 ? $load_flags : $16;
 $17 = ($t1size|0)!=(0|0);
 if ($17) {
  $18 = ((($t1size)) + 16|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = ((($t1glyph)) + 172|0);
  HEAP32[$20>>2] = $19;
  $21 = ((($t1size)) + 20|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = ((($t1glyph)) + 176|0);
  HEAP32[$23>>2] = $22;
 } else {
  $24 = ((($t1glyph)) + 172|0);
  HEAP32[$24>>2] = 65536;
  $25 = ((($t1glyph)) + 176|0);
  HEAP32[$25>>2] = 65536;
 }
 $26 = ((($t1glyph)) + 108|0);
 $27 = ((($t1glyph)) + 110|0);
 HEAP16[$27>>1] = 0;
 HEAP16[$26>>1] = 0;
 $28 = $load_flags$ & 1;
 $29 = ($28|0)==(0);
 $30 = $load_flags$ & 3;
 $31 = ($30|0)==(0);
 $32 = $31&1;
 $33 = ((($t1glyph)) + 72|0);
 HEAP32[$33>>2] = 1869968492;
 $34 = HEAP32[$5>>2]|0;
 $35 = ((($1)) + 420|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = ((($1)) + 528|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = $load_flags$ >>> 16;
 $40 = $39 & 15;
 $41 = (FUNCTION_TABLE_iiiiiiiiii[$34 & 8191]($decoder,$1,$t1size,$t1glyph,$36,$38,$32,$40,4436)|0);
 $42 = ($41|0)==(0);
 if (!($42)) {
  $error$025 = $41;
  STACKTOP = sp;return ($error$025|0);
 }
 $43 = $load_flags$ & 1024;
 $$lobit = $43 >>> 10;
 $44 = $$lobit&255;
 $45 = ((($decoder)) + 69|0);
 HEAP8[$45>>0] = $44;
 $46 = ((($1)) + 404|0);
 $47 = HEAP32[$46>>2]|0;
 $48 = ((($decoder)) + 1364|0);
 HEAP32[$48>>2] = $47;
 $49 = ((($1)) + 408|0);
 $50 = HEAP32[$49>>2]|0;
 $51 = ((($decoder)) + 1368|0);
 HEAP32[$51>>2] = $50;
 $52 = ((($1)) + 412|0);
 $53 = HEAP32[$52>>2]|0;
 $54 = ((($decoder)) + 1372|0);
 HEAP32[$54>>2] = $53;
 $55 = ((($1)) + 544|0);
 $56 = HEAP32[$55>>2]|0;
 $57 = ((($decoder)) + 1488|0);
 HEAP32[$57>>2] = $56;
 $58 = ((($1)) + 540|0);
 $59 = HEAP32[$58>>2]|0;
 $60 = ((($decoder)) + 1492|0);
 HEAP32[$60>>2] = $59;
 $61 = (_T1_Parse_Glyph_And_Get_Char_String($decoder,$glyph_index,$glyph_data)|0);
 $62 = ($61|0)==(0);
 if (!($62)) {
  $203 = ((($5)) + 4|0);
  $204 = HEAP32[$203>>2]|0;
  FUNCTION_TABLE_vi[$204 & 8191]($decoder);
  $error$025 = $61;
  STACKTOP = sp;return ($error$025|0);
 }
 $63 = ($43|0)==(0);
 $64 = ((($decoder)) + 1376|0);
 ;HEAP32[$font_matrix>>2]=HEAP32[$64>>2]|0;HEAP32[$font_matrix+4>>2]=HEAP32[$64+4>>2]|0;HEAP32[$font_matrix+8>>2]=HEAP32[$64+8>>2]|0;HEAP32[$font_matrix+12>>2]=HEAP32[$64+12>>2]|0;
 $65 = ((($decoder)) + 1392|0);
 $66 = HEAP32[$65>>2]|0;
 $67 = ((($decoder)) + 1396|0);
 $68 = HEAP32[$67>>2]|0;
 $69 = ((($5)) + 4|0);
 $70 = HEAP32[$69>>2]|0;
 FUNCTION_TABLE_vi[$70 & 8191]($decoder);
 $71 = ((($t1glyph)) + 124|0);
 $72 = HEAP32[$71>>2]|0;
 $73 = $72 & 1;
 $74 = $73 | 4;
 HEAP32[$71>>2] = $74;
 if ($63) {
  $91 = ((($t1glyph)) + 24|0);
  $92 = ((($decoder)) + 40|0);
  $93 = HEAP32[$92>>2]|0;
  $94 = (_FT_RoundFix($93)|0);
  $95 = $94 >> 16;
  $96 = ((($t1glyph)) + 40|0);
  HEAP32[$96>>2] = $95;
  $97 = HEAP32[$92>>2]|0;
  $98 = (_FT_RoundFix($97)|0);
  $99 = $98 >> 16;
  $100 = ((($t1glyph)) + 56|0);
  HEAP32[$100>>2] = $99;
  $101 = ((($t1glyph)) + 156|0);
  $102 = HEAP32[$101>>2]|0;
  $103 = ((($102)) + 8|0);
  HEAP8[$103>>0] = 0;
  $104 = $load_flags$ & 16;
  $105 = ($104|0)!=(0);
  if ($105) {
   $106 = ((($1)) + 472|0);
   $107 = HEAP32[$106>>2]|0;
   $108 = ((($1)) + 464|0);
   $109 = HEAP32[$108>>2]|0;
   $110 = (($107) - ($109))|0;
   $111 = $110 >> 16;
   $112 = ((($t1glyph)) + 52|0);
   HEAP32[$112>>2] = $111;
   $$sink = $111;
  } else {
   $113 = ((($decoder)) + 44|0);
   $114 = HEAP32[$113>>2]|0;
   $115 = (_FT_RoundFix($114)|0);
   $116 = $115 >> 16;
   $117 = ((($t1glyph)) + 52|0);
   HEAP32[$117>>2] = $116;
   $118 = HEAP32[$113>>2]|0;
   $119 = (_FT_RoundFix($118)|0);
   $120 = $119 >> 16;
   $$sink = $120;
  }
  $121 = ((($t1glyph)) + 60|0);
  HEAP32[$121>>2] = $$sink;
  HEAP32[$33>>2] = 1869968492;
  if ($17) {
   $122 = ((($t1size)) + 14|0);
   $123 = HEAP16[$122>>1]|0;
   $124 = ($123&65535)<(24);
   if ($124) {
    $125 = HEAP32[$71>>2]|0;
    $126 = $125 | 256;
    HEAP32[$71>>2] = $126;
   }
  }
  $127 = HEAP32[$font_matrix>>2]|0;
  $128 = ($127|0)==(65536);
  if ($128) {
   $129 = ((($font_matrix)) + 12|0);
   $130 = HEAP32[$129>>2]|0;
   $131 = ($130|0)==(65536);
   if ($131) {
    $132 = ((($font_matrix)) + 4|0);
    $133 = HEAP32[$132>>2]|0;
    $134 = ($133|0)==(0);
    if ($134) {
     $135 = ((($font_matrix)) + 8|0);
     $136 = HEAP32[$135>>2]|0;
     $137 = ($136|0)==(0);
     if (!($137)) {
      label = 20;
     }
    } else {
     label = 20;
    }
   } else {
    label = 20;
   }
  } else {
   label = 20;
  }
  if ((label|0) == 20) {
   _FT_Outline_Transform($26,$font_matrix);
  }
  $138 = $68 | $66;
  $139 = ($138|0)==(0);
  if (!($139)) {
   _FT_Outline_Translate($26,$66,$68);
  }
  $140 = HEAP32[$96>>2]|0;
  HEAP32[$advance>>2] = $140;
  $141 = ((($advance)) + 4|0);
  HEAP32[$141>>2] = 0;
  _FT_Vector_Transform($advance,$font_matrix);
  $142 = HEAP32[$advance>>2]|0;
  $143 = (($142) + ($66))|0;
  HEAP32[$96>>2] = $143;
  HEAP32[$advance>>2] = 0;
  $144 = ((($t1glyph)) + 52|0);
  $145 = HEAP32[$144>>2]|0;
  HEAP32[$141>>2] = $145;
  _FT_Vector_Transform($advance,$font_matrix);
  $146 = HEAP32[$141>>2]|0;
  $147 = (($146) + ($68))|0;
  HEAP32[$144>>2] = $147;
  if ($29) {
   $148 = ((($decoder)) + 16|0);
   $149 = HEAP32[$148>>2]|0;
   $150 = ((($149)) + 4|0);
   $151 = HEAP32[$150>>2]|0;
   $152 = ((($t1glyph)) + 172|0);
   $153 = HEAP32[$152>>2]|0;
   $154 = ((($t1glyph)) + 176|0);
   $155 = HEAP32[$154>>2]|0;
   if ($31) {
    $156 = ((($decoder)) + 72|0);
    $157 = HEAP32[$156>>2]|0;
    $158 = ($157|0)==(0|0);
    if ($158) {
     label = 26;
    }
   } else {
    label = 26;
   }
   if ((label|0) == 26) {
    $159 = ((($149)) + 2|0);
    $160 = HEAP16[$159>>1]|0;
    $161 = ($160<<16>>16)>(0);
    if ($161) {
     $162 = $160 << 16 >> 16;
     $n$06 = $162;$vec$07 = $151;
     while(1) {
      $163 = HEAP32[$vec$07>>2]|0;
      $164 = (_FT_MulFix($163,$153)|0);
      HEAP32[$vec$07>>2] = $164;
      $165 = ((($vec$07)) + 4|0);
      $166 = HEAP32[$165>>2]|0;
      $167 = (_FT_MulFix($166,$155)|0);
      HEAP32[$165>>2] = $167;
      $168 = (($n$06) + -1)|0;
      $169 = ((($vec$07)) + 8|0);
      $170 = ($n$06|0)>(1);
      if ($170) {
       $n$06 = $168;$vec$07 = $169;
      } else {
       break;
      }
     }
    }
   }
   $171 = HEAP32[$96>>2]|0;
   $172 = (_FT_MulFix($171,$153)|0);
   HEAP32[$96>>2] = $172;
   $173 = HEAP32[$144>>2]|0;
   $174 = (_FT_MulFix($173,$155)|0);
   HEAP32[$144>>2] = $174;
  }
  _FT_Outline_Get_CBox($26,$cbox);
  $175 = ((($cbox)) + 8|0);
  $176 = HEAP32[$175>>2]|0;
  $177 = HEAP32[$cbox>>2]|0;
  $178 = (($176) - ($177))|0;
  HEAP32[$91>>2] = $178;
  $179 = ((($cbox)) + 12|0);
  $180 = HEAP32[$179>>2]|0;
  $181 = ((($cbox)) + 4|0);
  $182 = HEAP32[$181>>2]|0;
  $183 = (($180) - ($182))|0;
  $184 = ((($t1glyph)) + 28|0);
  HEAP32[$184>>2] = $183;
  $185 = ((($t1glyph)) + 32|0);
  HEAP32[$185>>2] = $177;
  $186 = ((($t1glyph)) + 36|0);
  HEAP32[$186>>2] = $180;
  if ($105) {
   $187 = HEAP32[$144>>2]|0;
   _ft_synthesize_vertical_metrics($91,$187);
  }
 } else {
  $75 = ((($t1glyph)) + 156|0);
  $76 = HEAP32[$75>>2]|0;
  $77 = ((($decoder)) + 32|0);
  $78 = HEAP32[$77>>2]|0;
  $79 = (_FT_RoundFix($78)|0);
  $80 = $79 >> 16;
  $81 = ((($t1glyph)) + 32|0);
  HEAP32[$81>>2] = $80;
  $82 = ((($decoder)) + 40|0);
  $83 = HEAP32[$82>>2]|0;
  $84 = (_FT_RoundFix($83)|0);
  $85 = $84 >> 16;
  $86 = ((($t1glyph)) + 40|0);
  HEAP32[$86>>2] = $85;
  $87 = ((($76)) + 12|0);
  ;HEAP32[$87>>2]=HEAP32[$font_matrix>>2]|0;HEAP32[$87+4>>2]=HEAP32[$font_matrix+4>>2]|0;HEAP32[$87+8>>2]=HEAP32[$font_matrix+8>>2]|0;HEAP32[$87+12>>2]=HEAP32[$font_matrix+12>>2]|0;
  $88 = ((($76)) + 28|0);
  HEAP32[$88>>2] = $66;
  $89 = ((($76)) + 32|0);
  HEAP32[$89>>2] = $68;
  $90 = ((($76)) + 8|0);
  HEAP8[$90>>0] = 1;
 }
 $188 = HEAP32[$glyph_data>>2]|0;
 $189 = ((($t1glyph)) + 136|0);
 HEAP32[$189>>2] = $188;
 $190 = ((($glyph_data)) + 4|0);
 $191 = HEAP32[$190>>2]|0;
 $192 = ((($t1glyph)) + 140|0);
 HEAP32[$192>>2] = $191;
 $193 = ((($1)) + 128|0);
 $194 = HEAP32[$193>>2]|0;
 $195 = ((($194)) + 48|0);
 $196 = HEAP32[$195>>2]|0;
 $197 = ($196|0)==(0|0);
 if ($197) {
  $error$025 = 0;
  STACKTOP = sp;return ($error$025|0);
 }
 $198 = HEAP32[$196>>2]|0;
 $199 = ((($198)) + 4|0);
 $200 = HEAP32[$199>>2]|0;
 $201 = ((($196)) + 4|0);
 $202 = HEAP32[$201>>2]|0;
 FUNCTION_TABLE_vii[$200 & 8191]($202,$glyph_data);
 HEAP32[$189>>2] = 0;
 HEAP32[$192>>2] = 0;
 $error$025 = 0;
 STACKTOP = sp;return ($error$025|0);
}
function _Get_Kerning($t1face,$left_glyph,$right_glyph,$kerning) {
 $t1face = $t1face|0;
 $left_glyph = $left_glyph|0;
 $right_glyph = $right_glyph|0;
 $kerning = $kerning|0;
 var $$idx = 0, $$idx$val = 0, $$idx1 = 0, $$idx1$val = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $max$02$i = 0, $max$1$i = 0, $min$03$i = 0, $min$03$i$lcssa = 0, $min$1$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 HEAP32[$kerning>>2] = 0;
 $0 = ((($kerning)) + 4|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($t1face)) + 492|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  return 0;
 }
 $$idx = ((($2)) + 36|0);
 $$idx$val = HEAP32[$$idx>>2]|0;
 $$idx1 = ((($2)) + 40|0);
 $$idx1$val = HEAP32[$$idx1>>2]|0;
 $4 = $left_glyph << 16;
 $5 = $4 | $right_glyph;
 $6 = (($$idx$val) + ($$idx1$val<<4)|0);
 $7 = ((($6)) + -16|0);
 $8 = ($7>>>0)<($$idx$val>>>0);
 L4: do {
  if (!($8)) {
   $max$02$i = $7;$min$03$i = $$idx$val;
   while(1) {
    $9 = $max$02$i;
    $10 = $min$03$i;
    $11 = (($9) - ($10))|0;
    $12 = $11 >> 4;
    $13 = (($12|0) / 2)&-1;
    $14 = (($min$03$i) + ($13<<4)|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = $15 << 16;
    $17 = (((($min$03$i) + ($13<<4)|0)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = $16 | $18;
    $20 = ($19|0)==($5|0);
    if ($20) {
     $$lcssa = $13;$min$03$i$lcssa = $min$03$i;
     break;
    }
    $21 = ($19>>>0)<($5>>>0);
    $22 = ((($14)) + 16|0);
    $23 = ((($14)) + -16|0);
    $max$1$i = $21 ? $max$02$i : $23;
    $min$1$i = $21 ? $22 : $min$03$i;
    $24 = ($min$1$i>>>0)>($max$1$i>>>0);
    if ($24) {
     break L4;
    } else {
     $max$02$i = $max$1$i;$min$03$i = $min$1$i;
    }
   }
   $25 = (((($min$03$i$lcssa) + ($$lcssa<<4)|0)) + 8|0);
   $26 = HEAP32[$25>>2]|0;
   HEAP32[$kerning>>2] = $26;
   $27 = (((($min$03$i$lcssa) + ($$lcssa<<4)|0)) + 12|0);
   $28 = HEAP32[$27>>2]|0;
   HEAP32[$0>>2] = $28;
   return 0;
  }
 } while(0);
 HEAP32[$kerning>>2] = 0;
 HEAP32[$0>>2] = 0;
 return 0;
}
function _T1_Read_Metrics($t1_face,$stream) {
 $t1_face = $t1_face|0;
 $stream = $stream|0;
 var $$lcssa = 0, $$pr = 0, $$pr$i = 0, $$pr2$pre$i = 0, $$pre = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $error = 0, $error$i = 0, $fi$0 = 0, $fi$1 = 0, $kp$03$i = 0, $n$05$i = 0, $p$04$i = 0, $parser = 0;
 var $sext$i = 0, $stream$idx$val = 0, $stream$idx2 = 0, $stream$idx2$val = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp + 24|0;
 $parser = sp + 4|0;
 $error = sp;
 $0 = ((($stream)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$error>>2] = 2;
 $2 = ((($t1_face)) + 132|0);
 $3 = (_ft_mem_alloc($1,44,$error)|0);
 $4 = HEAP32[$error>>2]|0;
 $5 = ($4|0)==(0);
 if ($5) {
  $6 = ((($stream)) + 4|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (_FT_Stream_EnterFrame($stream,$7)|0);
  HEAP32[$error>>2] = $8;
  $9 = ($8|0)==(0);
  if ($9) {
   $10 = ((($3)) + 4|0);
   $11 = ((($t1_face)) + 460|0);
   ;HEAP32[$10>>2]=HEAP32[$11>>2]|0;HEAP32[$10+4>>2]=HEAP32[$11+4>>2]|0;HEAP32[$10+8>>2]=HEAP32[$11+8>>2]|0;HEAP32[$10+12>>2]=HEAP32[$11+12>>2]|0;
   $12 = ((($11)) + 12|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = ((($3)) + 20|0);
   HEAP32[$14>>2] = $13;
   $15 = ((($11)) + 4|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = ((($3)) + 24|0);
   HEAP32[$17>>2] = $16;
   $18 = ((($t1_face)) + 488|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ((($19)) + 24|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = ($21|0)==(0|0);
   if ($22) {
    label = 36;
   } else {
    $23 = HEAP32[$21>>2]|0;
    $24 = HEAP32[$0>>2]|0;
    $25 = ((($stream)) + 32|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ((($stream)) + 36|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = (FUNCTION_TABLE_iiiii[$23 & 8191]($parser,$24,$26,$28)|0);
    HEAP32[$error>>2] = $29;
    $30 = ($29|0)==(0);
    if ($30) {
     $31 = ((($parser)) + 8|0);
     HEAP32[$31>>2] = $3;
     $32 = ((($parser)) + 12|0);
     HEAP32[$32>>2] = 4437;
     $33 = ((($parser)) + 16|0);
     HEAP32[$33>>2] = $2;
     $34 = HEAP32[$20>>2]|0;
     $35 = ((($34)) + 8|0);
     $36 = HEAP32[$35>>2]|0;
     $37 = (FUNCTION_TABLE_ii[$36 & 8191]($parser)|0);
     HEAP32[$error>>2] = $37;
     $38 = HEAP32[$20>>2]|0;
     $39 = ((($38)) + 4|0);
     $40 = HEAP32[$39>>2]|0;
     FUNCTION_TABLE_vi[$40 & 8191]($parser);
     $$pre = HEAP32[$error>>2]|0;
     $$pr = $$pre;
    } else {
     $$pr = $29;
    }
    $41 = $$pr & 255;
    $42 = ($41|0)==(2);
    if ($42) {
     $43 = ((($stream)) + 32|0);
     $44 = HEAP32[$43>>2]|0;
     $45 = HEAP32[$6>>2]|0;
     $46 = ($45>>>0)>(6);
     if ($46) {
      $47 = ((($44)) + 1|0);
      $48 = HEAP8[$47>>0]|0;
      $49 = ($48&255)<(4);
      if ($49) {
       $50 = ((($44)) + 2|0);
       $51 = ((($44)) + 5|0);
       $52 = HEAP8[$51>>0]|0;
       $53 = $52&255;
       $54 = $53 << 24;
       $55 = ((($44)) + 4|0);
       $56 = HEAP8[$55>>0]|0;
       $57 = $56&255;
       $58 = $57 << 16;
       $59 = $58 | $54;
       $60 = ((($44)) + 3|0);
       $61 = HEAP8[$60>>0]|0;
       $62 = $61&255;
       $63 = $62 << 8;
       $64 = $59 | $63;
       $65 = HEAP8[$50>>0]|0;
       $66 = $65&255;
       $67 = $64 | $66;
       $68 = ($67|0)==($45|0);
       if ($68) {
        $stream$idx$val = HEAP32[$0>>2]|0;
        $stream$idx2 = ((($stream)) + 36|0);
        $stream$idx2$val = HEAP32[$stream$idx2>>2]|0;
        HEAP32[$error$i>>2] = 0;
        $69 = ((($44)) + 99|0);
        $70 = ((($44)) + 101|0);
        $71 = ($70>>>0)>($stream$idx2$val>>>0);
        L14: do {
         if ($71) {
          HEAP32[$error$i>>2] = 2;
          label = 33;
         } else {
          $72 = ((($44)) + 100|0);
          $73 = HEAP8[$72>>0]|0;
          $74 = $73&255;
          $75 = $74 << 8;
          $76 = HEAP8[$69>>0]|0;
          $77 = $76&255;
          $78 = $75 | $77;
          $79 = (($78) + 18)|0;
          $80 = (($69) + ($79)|0);
          $81 = ((($80)) + 18|0);
          $82 = ($81>>>0)>($stream$idx2$val>>>0);
          if ($82) {
           $181 = 0;
          } else {
           $83 = ((($80)) + 1|0);
           $84 = HEAP8[$83>>0]|0;
           $85 = $84&255;
           $86 = $85 << 8;
           $87 = HEAP8[$80>>0]|0;
           $88 = $87&255;
           $89 = $86 | $88;
           $90 = ($89>>>0)<(18);
           if ($90) {
            $181 = 0;
           } else {
            $91 = ((($80)) + 14|0);
            $92 = ((($91)) + 3|0);
            $93 = HEAP8[$92>>0]|0;
            $94 = $93&255;
            $95 = $94 << 24;
            $96 = ((($91)) + 2|0);
            $97 = HEAP8[$96>>0]|0;
            $98 = $97&255;
            $99 = $98 << 16;
            $100 = $99 | $95;
            $101 = ((($91)) + 1|0);
            $102 = HEAP8[$101>>0]|0;
            $103 = $102&255;
            $104 = $103 << 8;
            $105 = $100 | $104;
            $106 = HEAP8[$91>>0]|0;
            $107 = $106&255;
            $108 = $105 | $107;
            $109 = (($44) + ($108)|0);
            $110 = ($108|0)==(0);
            if ($110) {
             $181 = 0;
            } else {
             $111 = ((($109)) + 2|0);
             $112 = ($111>>>0)>($stream$idx2$val>>>0);
             if ($112) {
              HEAP32[$error$i>>2] = 2;
              label = 33;
              break;
             }
             $113 = ((($109)) + 1|0);
             $114 = HEAP8[$113>>0]|0;
             $115 = $114&255;
             $116 = $115 << 8;
             $117 = HEAP8[$109>>0]|0;
             $118 = $117&255;
             $119 = $116 | $118;
             $120 = ((($3)) + 40|0);
             HEAP32[$120>>2] = $119;
             $121 = $119 << 2;
             $122 = (($111) + ($121)|0);
             $123 = ($122>>>0)>($stream$idx2$val>>>0);
             if ($123) {
              HEAP32[$error$i>>2] = 2;
              label = 33;
              break;
             }
             $124 = ($119|0)==(0);
             if ($124) {
              $181 = 0;
             } else {
              $125 = (_ft_mem_realloc($stream$idx$val,16,0,$119,0,$error$i)|0);
              $126 = ((($3)) + 36|0);
              HEAP32[$126>>2] = $125;
              $127 = HEAP32[$error$i>>2]|0;
              $128 = ($127|0)==(0);
              if ($128) {
               $129 = HEAP32[$120>>2]|0;
               $130 = $129 << 2;
               $131 = (($111) + ($130)|0);
               $132 = ((($t1_face)) + 92|0);
               $133 = HEAP32[$132>>2]|0;
               $134 = ((($t1_face)) + 36|0);
               $135 = HEAP32[$134>>2]|0;
               $136 = ($135|0)>(0);
               L28: do {
                if ($136) {
                 $137 = ((($t1_face)) + 40|0);
                 $138 = HEAP32[$137>>2]|0;
                 $n$05$i = 0;
                 while(1) {
                  $141 = (($138) + ($n$05$i<<2)|0);
                  $142 = HEAP32[$141>>2]|0;
                  $143 = ((($142)) + 8|0);
                  $144 = HEAP16[$143>>1]|0;
                  $145 = ($144<<16>>16)==(7);
                  $140 = (($n$05$i) + 1)|0;
                  if ($145) {
                   $$lcssa = $142;
                   break;
                  }
                  $139 = ($140|0)<($135|0);
                  if ($139) {
                   $n$05$i = $140;
                  } else {
                   break L28;
                  }
                 }
                 $146 = (_FT_Set_Charmap($t1_face,$$lcssa)|0);
                 HEAP32[$error$i>>2] = $146;
                 $147 = ($146|0)==(0);
                 if (!($147)) {
                  label = 33;
                  break L14;
                 }
                }
               } while(0);
               $148 = ($130|0)>(0);
               if ($148) {
                $kp$03$i = $125;$p$04$i = $111;
                while(1) {
                 $149 = HEAP8[$p$04$i>>0]|0;
                 $150 = $149&255;
                 $151 = (_FT_Get_Char_Index($t1_face,$150)|0);
                 HEAP32[$kp$03$i>>2] = $151;
                 $152 = ((($p$04$i)) + 1|0);
                 $153 = HEAP8[$152>>0]|0;
                 $154 = $153&255;
                 $155 = (_FT_Get_Char_Index($t1_face,$154)|0);
                 $156 = ((($kp$03$i)) + 4|0);
                 HEAP32[$156>>2] = $155;
                 $157 = ((($p$04$i)) + 2|0);
                 $158 = ((($p$04$i)) + 3|0);
                 $159 = HEAP8[$158>>0]|0;
                 $160 = $159&255;
                 $161 = $160 << 8;
                 $162 = HEAP8[$157>>0]|0;
                 $163 = $162&255;
                 $164 = $161 | $163;
                 $sext$i = $164 << 16;
                 $165 = $sext$i >> 16;
                 $166 = ((($kp$03$i)) + 8|0);
                 HEAP32[$166>>2] = $165;
                 $167 = ((($kp$03$i)) + 12|0);
                 HEAP32[$167>>2] = 0;
                 $168 = ((($kp$03$i)) + 16|0);
                 $169 = ((($p$04$i)) + 4|0);
                 $170 = ($169>>>0)<($131>>>0);
                 if ($170) {
                  $kp$03$i = $168;$p$04$i = $169;
                 } else {
                  break;
                 }
                }
               }
               $171 = ($133|0)==(0|0);
               if ($171) {
                $$pr$i = HEAP32[$error$i>>2]|0;
                $174 = $$pr$i;
               } else {
                $172 = (_FT_Set_Charmap($t1_face,$133)|0);
                HEAP32[$error$i>>2] = $172;
                $174 = $172;
               }
               $173 = ($174|0)==(0);
               if (!($173)) {
                label = 33;
                break;
               }
               $175 = HEAP32[$126>>2]|0;
               $176 = HEAP32[$120>>2]|0;
               _qsort($175,$176,16,4438);
               $$pr2$pre$i = HEAP32[$error$i>>2]|0;
               $177 = ($$pr2$pre$i|0)==(0);
               if ($177) {
                $181 = 0;
               } else {
                label = 33;
               }
              } else {
               label = 33;
              }
             }
            }
           }
          }
         }
        } while(0);
        if ((label|0) == 33) {
         $178 = ((($3)) + 36|0);
         $179 = HEAP32[$178>>2]|0;
         _ft_mem_free($stream$idx$val,$179);
         HEAP32[$178>>2] = 0;
         $180 = ((($3)) + 40|0);
         HEAP32[$180>>2] = 0;
         $$pre$i = HEAP32[$error$i>>2]|0;
         $181 = $$pre$i;
        }
        HEAP32[$error>>2] = $181;
        $183 = $181;
       } else {
        $183 = $$pr;
       }
      } else {
       $183 = $$pr;
      }
     } else {
      $183 = $$pr;
     }
    } else {
     $183 = $$pr;
    }
    $182 = ($183|0)==(0);
    if ($182) {
     label = 36;
    } else {
     $fi$0 = $3;
    }
   }
   if ((label|0) == 36) {
    ;HEAP32[$11>>2]=HEAP32[$10>>2]|0;HEAP32[$11+4>>2]=HEAP32[$10+4>>2]|0;HEAP32[$11+8>>2]=HEAP32[$10+8>>2]|0;HEAP32[$11+12>>2]=HEAP32[$10+12>>2]|0;
    $184 = HEAP32[$10>>2]|0;
    $185 = $184 >> 16;
    $186 = ((($t1_face)) + 52|0);
    HEAP32[$186>>2] = $185;
    $187 = ((($3)) + 8|0);
    $188 = HEAP32[$187>>2]|0;
    $189 = $188 >> 16;
    $190 = ((($t1_face)) + 56|0);
    HEAP32[$190>>2] = $189;
    $191 = ((($3)) + 12|0);
    $192 = HEAP32[$191>>2]|0;
    $193 = (($192) + 65535)|0;
    $194 = $193 >> 16;
    $195 = ((($t1_face)) + 60|0);
    HEAP32[$195>>2] = $194;
    $196 = ((($3)) + 16|0);
    $197 = HEAP32[$196>>2]|0;
    $198 = (($197) + 65535)|0;
    $199 = $198 >> 16;
    $200 = ((($t1_face)) + 64|0);
    HEAP32[$200>>2] = $199;
    $201 = HEAP32[$14>>2]|0;
    $202 = (($201) + 32768)|0;
    $203 = $202 >>> 16;
    $204 = $203&65535;
    $205 = ((($t1_face)) + 70|0);
    HEAP16[$205>>1] = $204;
    $206 = HEAP32[$17>>2]|0;
    $207 = (($206) + 32768)|0;
    $208 = $207 >>> 16;
    $209 = $208&65535;
    $210 = ((($t1_face)) + 72|0);
    HEAP16[$210>>1] = $209;
    $211 = ((($3)) + 40|0);
    $212 = HEAP32[$211>>2]|0;
    $213 = ($212|0)==(0);
    if ($213) {
     $fi$0 = $3;
    } else {
     $214 = ((($t1_face)) + 8|0);
     $215 = HEAP32[$214>>2]|0;
     $216 = $215 | 64;
     HEAP32[$214>>2] = $216;
     $217 = ((($t1_face)) + 492|0);
     HEAP32[$217>>2] = $3;
     $fi$0 = 0;
    }
   }
   _FT_Stream_ExitFrame($stream);
   $fi$1 = $fi$0;
  } else {
   $fi$1 = $3;
  }
 } else {
  $fi$1 = $3;
 }
 $218 = ($fi$1|0)==(0|0);
 if ($218) {
  $225 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($225|0);
 }
 $219 = ((($fi$1)) + 36|0);
 $220 = HEAP32[$219>>2]|0;
 _ft_mem_free($1,$220);
 HEAP32[$219>>2] = 0;
 $221 = ((($fi$1)) + 40|0);
 HEAP32[$221>>2] = 0;
 $222 = ((($fi$1)) + 28|0);
 $223 = HEAP32[$222>>2]|0;
 _ft_mem_free($1,$223);
 HEAP32[$222>>2] = 0;
 $224 = ((($fi$1)) + 32|0);
 HEAP32[$224>>2] = 0;
 _ft_mem_free($1,$fi$1);
 $225 = HEAP32[$error>>2]|0;
 STACKTOP = sp;return ($225|0);
}
function _t1_get_index($name,$len,$user_data) {
 $name = $name|0;
 $len = $len|0;
 $user_data = $user_data|0;
 var $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $n$03 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($len>>>0)>(65535);
 if ($0) {
  $$2 = 0;
  return ($$2|0);
 }
 $1 = ((($user_data)) + 284|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)>(0);
 if (!($3)) {
  $$2 = 0;
  return ($$2|0);
 }
 $4 = ((($user_data)) + 288|0);
 $5 = HEAP32[$4>>2]|0;
 $n$03 = 0;
 while(1) {
  $6 = (($5) + ($n$03<<2)|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)==(0|0);
  if (!($8)) {
   $9 = HEAP8[$7>>0]|0;
   $10 = HEAP8[$name>>0]|0;
   $11 = ($9<<24>>24)==($10<<24>>24);
   if ($11) {
    $12 = (_strlen($7)|0);
    $13 = ($12|0)==($len|0);
    if ($13) {
     $14 = (_strncmp($7,$name,$len)|0);
     $15 = ($14|0)==(0);
     if ($15) {
      $$2 = $n$03;
      label = 9;
      break;
     }
    }
   }
  }
  $16 = (($n$03) + 1)|0;
  $17 = ($16|0)<($2|0);
  if ($17) {
   $n$03 = $16;
  } else {
   $$2 = 0;
   label = 9;
   break;
  }
 }
 if ((label|0) == 9) {
  return ($$2|0);
 }
 return (0)|0;
}
function _compare_kern_pairs($a,$b) {
 $a = $a|0;
 $b = $b|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$a>>2]|0;
 $1 = $0 << 16;
 $2 = ((($a)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $1 | $3;
 $5 = HEAP32[$b>>2]|0;
 $6 = $5 << 16;
 $7 = ((($b)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $6 | $8;
 $10 = ($4>>>0)>($9>>>0);
 $11 = ($4>>>0)<($9>>>0);
 $$ = $11 << 31 >> 31;
 $$0 = $10 ? 1 : $$;
 return ($$0|0);
}
function _T1_Get_Advances($t1face,$first,$count,$load_flags,$advances) {
 $t1face = $t1face|0;
 $first = $first|0;
 $count = $count|0;
 $load_flags = $load_flags|0;
 $advances = $advances|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $decoder = 0, $exitcond = 0, $glyph_data$i = 0, $nn$12 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1520|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $glyph_data$i = sp + 1504|0;
 $decoder = sp;
 $0 = $load_flags & 16;
 $1 = ($0|0)==(0);
 if (!($1)) {
  $2 = ($count|0)==(0);
  if ($2) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
  $3 = $count << 2;
  _memset(($advances|0),0,($3|0))|0;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ((($t1face)) + 488|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($t1face)) + 420|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($t1face)) + 528|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (FUNCTION_TABLE_iiiiiiiiii[$8 & 8191]($decoder,$t1face,0,0,$10,$12,0,0,4436)|0);
 $14 = ($13|0)==(0);
 if (!($14)) {
  $$0 = $13;
  STACKTOP = sp;return ($$0|0);
 }
 $15 = ((($decoder)) + 70|0);
 HEAP8[$15>>0] = 1;
 $16 = ((($decoder)) + 68|0);
 HEAP8[$16>>0] = 0;
 $17 = ((($t1face)) + 404|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($decoder)) + 1364|0);
 HEAP32[$19>>2] = $18;
 $20 = ((($t1face)) + 408|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ((($decoder)) + 1368|0);
 HEAP32[$22>>2] = $21;
 $23 = ((($t1face)) + 412|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($decoder)) + 1372|0);
 HEAP32[$25>>2] = $24;
 $26 = ((($t1face)) + 544|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ((($decoder)) + 1488|0);
 HEAP32[$28>>2] = $27;
 $29 = ((($t1face)) + 540|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ((($decoder)) + 1492|0);
 HEAP32[$31>>2] = $30;
 $32 = ($count|0)==(0);
 if ($32) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $33 = ((($decoder)) + 4|0);
 $34 = ((($decoder)) + 40|0);
 $nn$12 = 0;
 while(1) {
  $35 = (($nn$12) + ($first))|0;
  $36 = (_T1_Parse_Glyph_And_Get_Char_String($decoder,$35,$glyph_data$i)|0);
  $37 = ($36|0)==(0);
  if ($37) {
   $38 = HEAP32[$33>>2]|0;
   $39 = ((($38)) + 128|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = ((($40)) + 48|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = ($42|0)==(0|0);
   if (!($43)) {
    $44 = HEAP32[$42>>2]|0;
    $45 = ((($44)) + 4|0);
    $46 = HEAP32[$45>>2]|0;
    $47 = ((($42)) + 4|0);
    $48 = HEAP32[$47>>2]|0;
    FUNCTION_TABLE_vii[$46 & 8191]($48,$glyph_data$i);
   }
   $49 = HEAP32[$34>>2]|0;
   $50 = (_FT_RoundFix($49)|0);
   $51 = $50 >> 16;
   $52 = (($advances) + ($nn$12<<2)|0);
   HEAP32[$52>>2] = $51;
  } else {
   $53 = (($advances) + ($nn$12<<2)|0);
   HEAP32[$53>>2] = 0;
  }
  $54 = (($nn$12) + 1)|0;
  $exitcond = ($54|0)==($count|0);
  if ($exitcond) {
   $$0 = 0;
   break;
  } else {
   $nn$12 = $54;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _T1_Size_Request($t1size,$req) {
 $t1size = $t1size|0;
 $req = $req|0;
 var $$idx$val = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$idx$val = HEAP32[$t1size>>2]|0;
 $0 = ((($$idx$val)) + 548|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($$idx$val)) + 96|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (_FT_Get_Module($5,593918)|0);
 $7 = ($6|0)!=(0|0);
 $8 = ($1|0)!=(0|0);
 $or$cond$i = $8 & $7;
 if ($or$cond$i) {
  $9 = HEAP32[$1>>2]|0;
  $10 = ($9|0)==(0|0);
  if (!($10)) {
   $12 = (FUNCTION_TABLE_ii[$9 & 8191]($6)|0);
   $13 = HEAP32[$t1size>>2]|0;
   _FT_Request_Metrics($13,$req);
   $14 = ($12|0)==(0|0);
   if ($14) {
    return 0;
   }
   $15 = ((($12)) + 4|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = ((($t1size)) + 40|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = ((($t1size)) + 16|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = ((($t1size)) + 20|0);
   $22 = HEAP32[$21>>2]|0;
   FUNCTION_TABLE_viiiii[$16 & 8191]($18,$20,$22,0,0);
   return 0;
  }
 }
 $11 = HEAP32[$t1size>>2]|0;
 _FT_Request_Metrics($11,$req);
 return 0;
}
function _T42_Driver_Init($module) {
 $module = $module|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($module)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_FT_Get_Module($1,594296)|0);
 $3 = ($2|0)==(0|0);
 if ($3) {
  $$0 = 11;
  return ($$0|0);
 }
 $4 = HEAP32[$2>>2]|0;
 $5 = ((($module)) + 28|0);
 HEAP32[$5>>2] = $4;
 $$0 = 0;
 return ($$0|0);
}
function _T42_Driver_Done($module) {
 $module = $module|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _T42_Get_Interface($module,$t42_interface) {
 $module = $module|0;
 $t42_interface = $t42_interface|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_ft_service_list_lookup(58248,$t42_interface)|0);
 return ($0|0);
}
function _t42_get_glyph_name($face,$glyph_index,$buffer,$buffer_max) {
 $face = $face|0;
 $glyph_index = $glyph_index|0;
 $buffer = $buffer|0;
 $buffer_max = $buffer_max|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 420|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + ($glyph_index<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 (_ft_mem_strcpyn($buffer,$3,$buffer_max)|0);
 return 0;
}
function _t42_get_name_index($face,$glyph_name) {
 $face = $face|0;
 $glyph_name = $glyph_name|0;
 var $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $i$04 = 0, $i$04$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 416|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(0);
 if (!($2)) {
  $$2 = 0;
  return ($$2|0);
 }
 $3 = ((($face)) + 420|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = HEAP8[$glyph_name>>0]|0;
 $i$04 = 0;
 while(1) {
  $6 = (($4) + ($i$04<<2)|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = HEAP8[$7>>0]|0;
  $9 = ($5<<24>>24)==($8<<24>>24);
  if ($9) {
   $10 = (_strcmp($glyph_name,$7)|0);
   $11 = ($10|0)==(0);
   if ($11) {
    $i$04$lcssa = $i$04;
    break;
   }
  }
  $17 = (($i$04) + 1)|0;
  $18 = ($17|0)<($1|0);
  if ($18) {
   $i$04 = $17;
  } else {
   $$2 = 0;
   label = 7;
   break;
  }
 }
 if ((label|0) == 7) {
  return ($$2|0);
 }
 $12 = ((($face)) + 424|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($13) + ($i$04$lcssa<<2)|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = (_atol($15)|0);
 $$2 = $16;
 return ($$2|0);
}
function _t42_get_ps_font_name($face) {
 $face = $face|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 364|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function _t42_ps_get_font_info($face,$afont_info) {
 $face = $face|0;
 $afont_info = $afont_info|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 132|0);
 ;HEAP32[$afont_info>>2]=HEAP32[$0>>2]|0;HEAP32[$afont_info+4>>2]=HEAP32[$0+4>>2]|0;HEAP32[$afont_info+8>>2]=HEAP32[$0+8>>2]|0;HEAP32[$afont_info+12>>2]=HEAP32[$0+12>>2]|0;HEAP32[$afont_info+16>>2]=HEAP32[$0+16>>2]|0;HEAP32[$afont_info+20>>2]=HEAP32[$0+20>>2]|0;HEAP32[$afont_info+24>>2]=HEAP32[$0+24>>2]|0;HEAP32[$afont_info+28>>2]=HEAP32[$0+28>>2]|0;
 return 0;
}
function _t42_ps_get_font_extra($face,$afont_extra) {
 $face = $face|0;
 $afont_extra = $afont_extra|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 164|0);
 $1 = HEAP16[$0>>1]|0;
 HEAP16[$afont_extra>>1] = $1;
 return 0;
}
function _t42_ps_has_glyph_names($face) {
 $face = $face|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1;
}
function _t42_ps_get_font_private($face,$afont_private) {
 $face = $face|0;
 $afont_private = $afont_private|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 168|0);
 _memcpy(($afont_private|0),($0|0),196)|0;
 return 0;
}
function _T42_Face_Init($stream,$t42face,$face_index,$num_params,$params) {
 $stream = $stream|0;
 $t42face = $t42face|0;
 $face_index = $face_index|0;
 $num_params = $num_params|0;
 $params = $params|0;
 var $$ = 0, $$$i = 0, $$02 = 0, $$cast = 0, $$cast$i$i = 0, $$idx$i = 0, $$idx$val$i = 0, $$idx$val$idx$val$i = 0, $$lcssa = 0, $$lcssa90 = 0, $$not$i = 0, $$off$i$i$i = 0, $$ph$i = 0, $$pr$i = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$phi$iZ2D = 0, $$pre$phi30$iZ2D = 0, $$pre2$i$i = 0, $$pre29$i = 0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0;
 var $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0;
 var $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0;
 var $369 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $args = 0, $charcode$0$min_char$0$i = 0, $charcode$015$i = 0, $charcode$015$i$us = 0, $charmap = 0, $clazz$0 = 0, $clazz$0$in = 0, $cur$037$i$i = 0, $cur2$1$i$i = 0, $dummy_object$i$i$i = 0;
 var $error$0$i$i$i = 0, $error$0$i$i$i$lcssa = 0, $error$1 = 0, $error$i = 0, $error$i$i = 0, $exitcond = 0, $exitcond36 = 0, $family$0$ph$be = 0, $family$0$ph20 = 0, $family$017 = 0, $family$017$lcssa = 0, $full$0$ph19 = 0, $full$0$ph19$lcssa85 = 0, $i$036$i$i = 0, $i$036$i$i$lcssa = 0, $idx$012$i = 0, $idx$012$i$lcssa = 0, $loader$i = 0, $max_char$0$$i = 0, $max_char$0$lcssa$i = 0;
 var $max_char$014$i = 0, $max_char$1$i = 0, $min_char$0$lcssa$i = 0, $min_char$013$i = 0, $min_char$2$i = 0, $not$ = 0, $or$cond = 0, $or$cond$i$i = 0, $or$cond10$i$i = 0, $or$cond12$i$i = 0, $or$cond14$i$i = 0, $or$cond15$i$i = 0, $or$cond4 = 0, $switch$i$i$i = 0, $token$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 384|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $dummy_object$i$i$i = sp + 376|0;
 $token$i$i = sp + 364|0;
 $error$i$i = sp + 360|0;
 $loader$i = sp + 56|0;
 $error$i = sp + 48|0;
 $args = sp + 16|0;
 $charmap = sp;
 $0 = ((($t42face)) + 132|0);
 $1 = ((($t42face)) + 500|0);
 HEAP32[$1>>2] = 0;
 HEAP32[$t42face>>2] = 1;
 $2 = ((($t42face)) + 96|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_ft_module_get_service($3,594361)|0);
 $5 = ((($t42face)) + 484|0);
 HEAP32[$5>>2] = $4;
 $6 = HEAP32[$2>>2]|0;
 $7 = ((($6)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (_FT_Get_Module_Interface($8,594378)|0);
 $10 = ((($t42face)) + 488|0);
 HEAP32[$10>>2] = $9;
 $11 = ($9|0)==(0|0);
 if ($11) {
  $$02 = 11;
  STACKTOP = sp;return ($$02|0);
 }
 $12 = ((($t42face)) + 100|0);
 $13 = HEAP32[$12>>2]|0;
 _memset(($loader$i|0),0,304)|0;
 $14 = (_ft_mem_alloc($13,12,$error$i)|0);
 $15 = ((($t42face)) + 492|0);
 HEAP32[$15>>2] = $14;
 $16 = HEAP32[$error$i>>2]|0;
 $17 = ($16|0)==(0);
 L4: do {
  if ($17) {
   $18 = ((($t42face)) + 496|0);
   HEAP32[$18>>2] = 12;
   $19 = ((($t42face)) + 104|0);
   $20 = HEAP32[$19>>2]|0;
   $$idx$i = ((($9)) + 4|0);
   $$idx$val$i = HEAP32[$$idx$i>>2]|0;
   $$idx$val$idx$val$i = HEAP32[$$idx$val$i>>2]|0;
   FUNCTION_TABLE_viiii[$$idx$val$idx$val$i & 8191]($loader$i,0,0,$13);
   $21 = ((($loader$i)) + 72|0);
   HEAP32[$21>>2] = $20;
   $22 = ((($loader$i)) + 80|0);
   HEAP32[$22>>2] = 0;
   $23 = ((($loader$i)) + 76|0);
   HEAP32[$23>>2] = 0;
   $24 = ((($loader$i)) + 84|0);
   HEAP8[$24>>0] = 0;
   $25 = (_FT_Stream_Seek($20,0)|0);
   HEAP32[$error$i$i>>2] = $25;
   $26 = ($25|0)==(0);
   do {
    if ($26) {
     $27 = (_FT_Stream_EnterFrame($20,17)|0);
     HEAP32[$error$i$i>>2] = $27;
     $28 = ($27|0)==(0);
     if ($28) {
      $29 = ((($20)) + 32|0);
      $30 = HEAP32[$29>>2]|0;
      $31 = (_memcmp($30,594384,17)|0);
      $32 = ($31|0)==(0);
      if (!($32)) {
       HEAP32[$error$i$i>>2] = 2;
       _FT_Stream_ExitFrame($20);
       $58 = 2;
       label = 15;
       break;
      }
      _FT_Stream_ExitFrame($20);
      $33 = (_FT_Stream_Seek($20,0)|0);
      HEAP32[$error$i$i>>2] = $33;
      $34 = ($33|0)==(0);
      if ($34) {
       $35 = ((($20)) + 4|0);
       $36 = HEAP32[$35>>2]|0;
       $37 = ((($20)) + 20|0);
       $38 = HEAP32[$37>>2]|0;
       $39 = ($38|0)==(0|0);
       if ($39) {
        $40 = HEAP32[$20>>2]|0;
        $41 = ((($20)) + 8|0);
        $42 = HEAP32[$41>>2]|0;
        $43 = (($40) + ($42)|0);
        HEAP32[$23>>2] = $43;
        HEAP32[$22>>2] = $36;
        HEAP8[$24>>0] = 1;
        $44 = (_FT_Stream_Skip($20,$36)|0);
        HEAP32[$error$i$i>>2] = $44;
        $45 = ($44|0)==(0);
        if (!($45)) {
         $58 = $44;
         label = 15;
         break;
        }
        $$pre$i$i = HEAP32[$22>>2]|0;
        $54 = $$pre$i$i;
       } else {
        $46 = (_ft_mem_alloc($13,$36,$error$i$i)|0);
        HEAP32[$23>>2] = $46;
        $47 = HEAP32[$error$i$i>>2]|0;
        $48 = ($47|0)==(0);
        if (!($48)) {
         $58 = $47;
         label = 15;
         break;
        }
        $49 = (_FT_Stream_Read($20,$46,$36)|0);
        HEAP32[$error$i$i>>2] = $49;
        $50 = ($49|0)==(0);
        if (!($50)) {
         $58 = $49;
         label = 15;
         break;
        }
        HEAP32[$22>>2] = $36;
        $54 = $36;
       }
       $51 = HEAP32[$23>>2]|0;
       $52 = ((($loader$i)) + 4|0);
       HEAP32[$52>>2] = $51;
       $$cast$i$i = $51;
       HEAP32[$loader$i>>2] = $$cast$i$i;
       $53 = (($$cast$i$i) + ($54)|0);
       $55 = ((($loader$i)) + 8|0);
       HEAP32[$55>>2] = $53;
       HEAP32[$error$i>>2] = 0;
       $$pre$phi$iZ2D = $loader$i;$$pre$phi30$iZ2D = $55;$61 = $$cast$i$i;$63 = $54;
      } else {
       $58 = $33;
       label = 15;
      }
     } else {
      $58 = $27;
      label = 15;
     }
    } else {
     $58 = $25;
     label = 15;
    }
   } while(0);
   if ((label|0) == 15) {
    $56 = HEAP8[$24>>0]|0;
    $57 = ($56<<24>>24)==(0);
    if (!($57)) {
     HEAP32[$error$i>>2] = $58;
     break;
    }
    $59 = HEAP32[$23>>2]|0;
    _ft_mem_free($13,$59);
    HEAP32[$23>>2] = 0;
    $$pre2$i$i = HEAP32[$error$i$i>>2]|0;
    HEAP32[$error$i>>2] = $$pre2$i$i;
    $60 = ($$pre2$i$i|0)==(0);
    if (!($60)) {
     break;
    }
    $$pre$i = HEAP32[$22>>2]|0;
    $$pre29$i = ((($loader$i)) + 8|0);
    $$pre$phi$iZ2D = $loader$i;$$pre$phi30$iZ2D = $$pre29$i;$61 = 0;$63 = $$pre$i;
   }
   HEAP32[$$pre$phi$iZ2D>>2] = $61;
   $62 = (($61) + ($63)|0);
   HEAP32[$$pre$phi30$iZ2D>>2] = $62;
   $64 = ((($loader$i)) + 12|0);
   HEAP32[$64>>2] = 0;
   $65 = ((($loader$i)) + 28|0);
   $66 = HEAP32[$65>>2]|0;
   FUNCTION_TABLE_vi[$66 & 8191]($loader$i);
   $67 = HEAP32[$$pre$phi$iZ2D>>2]|0;
   $68 = ($67>>>0)<($62>>>0);
   L27: do {
    if ($68) {
     $69 = ((($loader$i)) + 32|0);
     $70 = ((($loader$i)) + 56|0);
     $71 = ((($token$i$i)) + 8|0);
     $72 = ((($loader$i)) + 68|0);
     $73 = ((($loader$i)) + 64|0);
     $74 = ((($t42face)) + 164|0);
     $75 = ((($t42face)) + 460|0);
     $77 = $67;
     L29: while(1) {
      $76 = HEAP8[$77>>0]|0;
      $78 = ($76<<24>>24)==(70);
      $79 = ((($77)) + 25|0);
      $80 = ($79>>>0)<($62>>>0);
      $or$cond10$i$i = $80 & $78;
      if ($or$cond10$i$i) {
       $81 = (_strncmp($77,594402,13)|0);
       $82 = ($81|0)==(0);
       if ($82) {
        $83 = HEAP32[$69>>2]|0;
        FUNCTION_TABLE_vi[$83 & 8191]($loader$i);
        $84 = HEAP32[$65>>2]|0;
        FUNCTION_TABLE_vi[$84 & 8191]($loader$i);
        $85 = HEAP32[$$pre$phi$iZ2D>>2]|0;
        $86 = ($85>>>0)<($62>>>0);
        L34: do {
         if ($86) {
          $cur$037$i$i = $85;
          while(1) {
           $87 = HEAP8[$cur$037$i$i>>0]|0;
           $88 = ($87<<24>>24)==(107);
           $89 = ((($cur$037$i$i)) + 5|0);
           $90 = ($89>>>0)<($62>>>0);
           $or$cond12$i$i = $90 & $88;
           if ($or$cond12$i$i) {
            $91 = (_strncmp($cur$037$i$i,594416,5)|0);
            $92 = ($91|0)==(0);
            if ($92) {
             break;
            }
           }
           $93 = HEAP32[$69>>2]|0;
           FUNCTION_TABLE_vi[$93 & 8191]($loader$i);
           $94 = HEAP32[$64>>2]|0;
           $95 = ($94|0)==(0);
           if (!($95)) {
            $$ph$i = $94;
            label = 52;
            break L29;
           }
           $96 = HEAP32[$65>>2]|0;
           FUNCTION_TABLE_vi[$96 & 8191]($loader$i);
           $97 = HEAP32[$$pre$phi$iZ2D>>2]|0;
           $98 = ($97>>>0)<($62>>>0);
           if ($98) {
            $cur$037$i$i = $97;
           } else {
            $cur2$1$i$i = $85;
            break L34;
           }
          }
          $99 = HEAP32[$69>>2]|0;
          FUNCTION_TABLE_vi[$99 & 8191]($loader$i);
          $100 = HEAP32[$70>>2]|0;
          FUNCTION_TABLE_vii[$100 & 8191]($loader$i,$token$i$i);
          $101 = HEAP32[$71>>2]|0;
          $102 = ($101|0)==(3);
          $103 = HEAP32[$$pre$phi$iZ2D>>2]|0;
          $$$i = $102 ? $103 : $85;
          $cur2$1$i$i = $$$i;
         } else {
          $cur2$1$i$i = $85;
         }
        } while(0);
        HEAP32[$$pre$phi$iZ2D>>2] = $cur2$1$i$i;
       } else {
        label = 30;
       }
      } else {
       label = 30;
      }
      L43: do {
       if ((label|0) == 30) {
        label = 0;
        $104 = ($76<<24>>24)==(47);
        $105 = ((($77)) + 2|0);
        $106 = ($105>>>0)<($62>>>0);
        $or$cond14$i$i = $106 & $104;
        if (!($or$cond14$i$i)) {
         $144 = HEAP32[$69>>2]|0;
         FUNCTION_TABLE_vi[$144 & 8191]($loader$i);
         $145 = HEAP32[$64>>2]|0;
         $146 = ($145|0)==(0);
         if ($146) {
          break;
         } else {
          $$ph$i = $145;
          label = 52;
          break L29;
         }
        }
        $107 = ((($77)) + 1|0);
        HEAP32[$$pre$phi$iZ2D>>2] = $107;
        $108 = HEAP32[$69>>2]|0;
        FUNCTION_TABLE_vi[$108 & 8191]($loader$i);
        $109 = HEAP32[$64>>2]|0;
        $110 = ($109|0)==(0);
        if (!($110)) {
         $$ph$i = $109;
         label = 52;
         break L29;
        }
        $111 = HEAP32[$$pre$phi$iZ2D>>2]|0;
        $112 = $111;
        $113 = $107;
        $114 = (($112) - ($113))|0;
        $115 = ($111|0)!=($107|0);
        $116 = ($114>>>0)<(22);
        $or$cond$i$i = $115 & $116;
        $117 = ($111>>>0)<($62>>>0);
        $or$cond15$i$i = $117 & $or$cond$i$i;
        if ($or$cond15$i$i) {
         $i$036$i$i = 0;
         while(1) {
          $118 = (58320 + (($i$036$i$i*36)|0)|0);
          $119 = HEAP32[$118>>2]|0;
          $120 = ($119|0)==(0|0);
          if (!($120)) {
           $121 = HEAP8[$107>>0]|0;
           $122 = HEAP8[$119>>0]|0;
           $123 = ($121<<24>>24)==($122<<24>>24);
           if ($123) {
            $124 = (_strlen($119)|0);
            $125 = ($114|0)==($124|0);
            if ($125) {
             $126 = (_memcmp($107,$119,$114)|0);
             $127 = ($126|0)==(0);
             if ($127) {
              $$lcssa90 = $118;$i$036$i$i$lcssa = $i$036$i$i;
              break;
             }
            }
           }
          }
          $142 = (($i$036$i$i) + 1)|0;
          $143 = ($142|0)<(20);
          if ($143) {
           $i$036$i$i = $142;
          } else {
           break L43;
          }
         }
         $128 = (((58320 + (($i$036$i$i$lcssa*36)|0)|0)) + 8|0);
         $129 = HEAP32[$128>>2]|0;
         $130 = (($i$036$i$i$lcssa) + -15)|0;
         $131 = ($130>>>0)<(4);
         do {
          if ($131) {
           $132 = (((58320 + (($i$036$i$i$lcssa*36)|0)|0)) + 12|0);
           $133 = HEAP32[$132>>2]|0;
           FUNCTION_TABLE_vii[$133 & 8191]($t42face,$loader$i);
           $134 = HEAP32[$64>>2]|0;
           $error$0$i$i$i = $134;
          } else {
           $135 = (((58320 + (($i$036$i$i$lcssa*36)|0)|0)) + 4|0);
           $136 = HEAP32[$135>>2]|0;
           switch ($136|0) {
           case 3:  {
            HEAP32[$dummy_object$i$i$i>>2] = $0;
            break;
           }
           case 2:  {
            HEAP32[$dummy_object$i$i$i>>2] = $74;
            break;
           }
           case 5:  {
            HEAP32[$dummy_object$i$i$i>>2] = $75;
            break;
           }
           default: {
            HEAP32[$dummy_object$i$i$i>>2] = $0;
           }
           }
           $$off$i$i$i = (($129) + -9)|0;
           $switch$i$i$i = ($$off$i$i$i>>>0)<(2);
           if ($switch$i$i$i) {
            $137 = HEAP32[$72>>2]|0;
            $138 = (FUNCTION_TABLE_iiiiii[$137 & 8191]($loader$i,$$lcssa90,$dummy_object$i$i$i,0,0)|0);
            $error$0$i$i$i = $138;
            break;
           } else {
            $139 = HEAP32[$73>>2]|0;
            $140 = (FUNCTION_TABLE_iiiiii[$139 & 8191]($loader$i,$$lcssa90,$dummy_object$i$i$i,0,0)|0);
            $error$0$i$i$i = $140;
            break;
           }
          }
         } while(0);
         HEAP32[$64>>2] = $error$0$i$i$i;
         $141 = ($error$0$i$i$i|0)==(0);
         if (!($141)) {
          $error$0$i$i$i$lcssa = $error$0$i$i$i;
          label = 48;
          break L29;
         }
        }
       }
      } while(0);
      $147 = HEAP32[$65>>2]|0;
      FUNCTION_TABLE_vi[$147 & 8191]($loader$i);
      $148 = HEAP32[$$pre$phi$iZ2D>>2]|0;
      $149 = ($148>>>0)<($62>>>0);
      if ($149) {
       $77 = $148;
      } else {
       break L27;
      }
     }
     if ((label|0) == 48) {
      HEAP32[$error$i>>2] = $error$0$i$i$i$lcssa;
      break L4;
     }
     else if ((label|0) == 52) {
      HEAP32[$error$i>>2] = $$ph$i;
      break L4;
     }
    }
   } while(0);
   $$pr$i = HEAP32[$64>>2]|0;
   HEAP32[$error$i>>2] = $$pr$i;
   $150 = ($$pr$i|0)==(0);
   if ($150) {
    $151 = ((($t42face)) + 433|0);
    $152 = HEAP8[$151>>0]|0;
    $153 = ($152<<24>>24)==(42);
    if (!($153)) {
     HEAP32[$error$i>>2] = 2;
     break;
    }
    $154 = ((($loader$i)) + 144|0);
    $155 = HEAP32[$154>>2]|0;
    $156 = ((($t42face)) + 416|0);
    HEAP32[$156>>2] = $155;
    $157 = ((($loader$i)) + 200|0);
    $158 = ((($loader$i)) + 212|0);
    $159 = HEAP32[$158>>2]|0;
    $160 = ($159|0)==(0);
    if ($160) {
     HEAP32[$error$i>>2] = 3;
    }
    HEAP32[$158>>2] = 0;
    $161 = HEAP32[$157>>2]|0;
    $162 = ((($t42face)) + 396|0);
    HEAP32[$162>>2] = $161;
    $163 = ((($loader$i)) + 224|0);
    $164 = HEAP32[$163>>2]|0;
    $165 = ((($t42face)) + 424|0);
    HEAP32[$165>>2] = $164;
    $166 = ((($loader$i)) + 228|0);
    $167 = HEAP32[$166>>2]|0;
    $168 = ((($t42face)) + 428|0);
    HEAP32[$168>>2] = $167;
    $169 = ((($loader$i)) + 148|0);
    $170 = HEAP32[$169>>2]|0;
    $171 = ((($t42face)) + 400|0);
    HEAP32[$171>>2] = $170;
    $172 = ((($loader$i)) + 172|0);
    $173 = HEAP32[$172>>2]|0;
    $174 = ((($t42face)) + 420|0);
    HEAP32[$174>>2] = $173;
    HEAP32[$169>>2] = 0;
    HEAP32[$172>>2] = 0;
    $175 = ((($t42face)) + 368|0);
    $176 = HEAP32[$175>>2]|0;
    $177 = ($176|0)==(1);
    if ($177) {
     $178 = ((($loader$i)) + 108|0);
     $179 = HEAP32[$178>>2]|0;
     $180 = ($179|0)>(0);
     L83: do {
      if ($180) {
       $181 = ((($t42face)) + 384|0);
       $182 = ((($t42face)) + 388|0);
       $183 = ((($loader$i)) + 116|0);
       $$not$i = ($155|0)<(1);
       if ($$not$i) {
        $charcode$015$i$us = 0;
        while(1) {
         $184 = HEAP32[$181>>2]|0;
         $185 = (($184) + ($charcode$015$i$us<<1)|0);
         HEAP16[$185>>1] = 0;
         $186 = HEAP32[$182>>2]|0;
         $187 = (($186) + ($charcode$015$i$us<<2)|0);
         HEAP32[$187>>2] = 594602;
         $188 = (($charcode$015$i$us) + 1)|0;
         $exitcond = ($188|0)==($179|0);
         if ($exitcond) {
          $max_char$0$lcssa$i = 0;$min_char$0$lcssa$i = 0;
          break L83;
         } else {
          $charcode$015$i$us = $188;
         }
        }
       } else {
        $charcode$015$i = 0;$max_char$014$i = 0;$min_char$013$i = 0;
       }
       while(1) {
        $189 = HEAP32[$181>>2]|0;
        $190 = (($189) + ($charcode$015$i<<1)|0);
        HEAP16[$190>>1] = 0;
        $191 = HEAP32[$182>>2]|0;
        $192 = (($191) + ($charcode$015$i<<2)|0);
        HEAP32[$192>>2] = 594602;
        $193 = HEAP32[$183>>2]|0;
        $194 = (($193) + ($charcode$015$i<<2)|0);
        $195 = HEAP32[$194>>2]|0;
        $196 = ($195|0)==(0|0);
        L90: do {
         if ($196) {
          $max_char$1$i = $max_char$014$i;$min_char$2$i = $min_char$013$i;
         } else {
          $197 = HEAP32[$174>>2]|0;
          $idx$012$i = 0;
          while(1) {
           $198 = (($197) + ($idx$012$i<<2)|0);
           $199 = HEAP32[$198>>2]|0;
           $200 = (_strcmp($195,$199)|0);
           $201 = ($200|0)==(0);
           if ($201) {
            $$lcssa = $199;$idx$012$i$lcssa = $idx$012$i;
            break;
           }
           $212 = (($idx$012$i) + 1)|0;
           $213 = ($212|0)<($155|0);
           if ($213) {
            $idx$012$i = $212;
           } else {
            $max_char$1$i = $max_char$014$i;$min_char$2$i = $min_char$013$i;
            break L90;
           }
          }
          $202 = $idx$012$i$lcssa&65535;
          $203 = HEAP32[$181>>2]|0;
          $204 = (($203) + ($charcode$015$i<<1)|0);
          HEAP16[$204>>1] = $202;
          $205 = HEAP32[$182>>2]|0;
          $206 = (($205) + ($charcode$015$i<<2)|0);
          HEAP32[$206>>2] = $$lcssa;
          $207 = (_strcmp(594602,$$lcssa)|0);
          $208 = ($207|0)==(0);
          if ($208) {
           $max_char$1$i = $max_char$014$i;$min_char$2$i = $min_char$013$i;
          } else {
           $209 = ($charcode$015$i|0)<($min_char$013$i|0);
           $charcode$0$min_char$0$i = $209 ? $charcode$015$i : $min_char$013$i;
           $210 = ($charcode$015$i|0)<($max_char$014$i|0);
           $211 = (($charcode$015$i) + 1)|0;
           $max_char$0$$i = $210 ? $max_char$014$i : $211;
           $max_char$1$i = $max_char$0$$i;$min_char$2$i = $charcode$0$min_char$0$i;
          }
         }
        } while(0);
        $214 = (($charcode$015$i) + 1)|0;
        $exitcond36 = ($214|0)==($179|0);
        if ($exitcond36) {
         $max_char$0$lcssa$i = $max_char$1$i;$min_char$0$lcssa$i = $min_char$2$i;
         break;
        } else {
         $charcode$015$i = $214;$max_char$014$i = $max_char$1$i;$min_char$013$i = $min_char$2$i;
        }
       }
      } else {
       $max_char$0$lcssa$i = 0;$min_char$0$lcssa$i = 0;
      }
     } while(0);
     $215 = ((($t42face)) + 376|0);
     HEAP32[$215>>2] = $min_char$0$lcssa$i;
     $216 = ((($t42face)) + 380|0);
     HEAP32[$216>>2] = $max_char$0$lcssa$i;
     $217 = ((($loader$i)) + 88|0);
     $218 = HEAP32[$217>>2]|0;
     $219 = ((($t42face)) + 372|0);
     HEAP32[$219>>2] = $218;
    }
   }
  }
 } while(0);
 $220 = ((($loader$i)) + 140|0);
 $221 = HEAP32[$220>>2]|0;
 $222 = ($221|0)==(0|0);
 if (!($222)) {
  $223 = ((($loader$i)) + 92|0);
  FUNCTION_TABLE_vi[$221 & 8191]($223);
 }
 $224 = ((($loader$i)) + 248|0);
 $225 = HEAP32[$224>>2]|0;
 $226 = ($225|0)==(0|0);
 if (!($226)) {
  $227 = ((($loader$i)) + 200|0);
  FUNCTION_TABLE_vi[$225 & 8191]($227);
 }
 $228 = ((($loader$i)) + 196|0);
 $229 = HEAP32[$228>>2]|0;
 $230 = ($229|0)==(0|0);
 if (!($230)) {
  $231 = ((($loader$i)) + 148|0);
  FUNCTION_TABLE_vi[$229 & 8191]($231);
 }
 $232 = ((($loader$i)) + 300|0);
 $233 = HEAP32[$232>>2]|0;
 $234 = ($233|0)==(0|0);
 if (!($234)) {
  $235 = ((($loader$i)) + 252|0);
  FUNCTION_TABLE_vi[$233 & 8191]($235);
 }
 $236 = ((($loader$i)) + 84|0);
 $237 = HEAP8[$236>>0]|0;
 $238 = ($237<<24>>24)==(0);
 if ($238) {
  $239 = ((($loader$i)) + 16|0);
  $240 = HEAP32[$239>>2]|0;
  $241 = ((($loader$i)) + 76|0);
  $242 = HEAP32[$241>>2]|0;
  _ft_mem_free($240,$242);
  HEAP32[$241>>2] = 0;
 }
 $243 = ((($loader$i)) + 24|0);
 $244 = HEAP32[$243>>2]|0;
 FUNCTION_TABLE_vi[$244 & 8191]($loader$i);
 $245 = HEAP32[$error$i>>2]|0;
 $246 = ($245|0)!=(0);
 $247 = ($face_index|0)<(0);
 $or$cond = $247 | $246;
 if ($or$cond) {
  $$02 = $245;
  STACKTOP = sp;return ($$02|0);
 }
 $248 = ($face_index|0)>(0);
 if ($248) {
  $$02 = 6;
  STACKTOP = sp;return ($$02|0);
 }
 $249 = ((($t42face)) + 416|0);
 $250 = HEAP32[$249>>2]|0;
 $251 = ((($t42face)) + 16|0);
 HEAP32[$251>>2] = $250;
 $252 = ((($t42face)) + 36|0);
 HEAP32[$252>>2] = 0;
 $253 = ((($t42face)) + 4|0);
 HEAP32[$253>>2] = 0;
 $254 = ((($t42face)) + 8|0);
 $255 = HEAP32[$254>>2]|0;
 $256 = $255 | 529;
 HEAP32[$254>>2] = $256;
 $257 = ((($t42face)) + 156|0);
 $258 = HEAP8[$257>>0]|0;
 $259 = ($258<<24>>24)==(0);
 if ($259) {
  $262 = $256;
 } else {
  $260 = $255 | 533;
  HEAP32[$254>>2] = $260;
  $262 = $260;
 }
 $261 = $262 | 2048;
 HEAP32[$254>>2] = $261;
 $263 = ((($t42face)) + 144|0);
 $264 = HEAP32[$263>>2]|0;
 $265 = ((($t42face)) + 20|0);
 HEAP32[$265>>2] = $264;
 $266 = ((($t42face)) + 24|0);
 HEAP32[$266>>2] = 594711;
 $267 = ($264|0)==(0);
 L124: do {
  if ($267) {
   $282 = ((($t42face)) + 364|0);
   $283 = HEAP32[$282>>2]|0;
   $284 = ($283|0)==(0|0);
   if (!($284)) {
    HEAP32[$265>>2] = $283;
   }
  } else {
   $268 = ((($t42face)) + 140|0);
   $269 = HEAP32[$268>>2]|0;
   $270 = ($269|0)==(0|0);
   if (!($270)) {
    $271 = HEAP8[$269>>0]|0;
    $272 = ($271<<24>>24)==(0);
    if (!($272)) {
     $$cast = $264;
     $274 = $271;$family$0$ph20 = $$cast;$full$0$ph19 = $269;
     L130: while(1) {
      $273 = ($274<<24>>24)==(0);
      $family$017 = $family$0$ph20;
      L132: while(1) {
       $275 = HEAP8[$family$017>>0]|0;
       $276 = ($274<<24>>24)==($275<<24>>24);
       if ($276) {
        $family$017$lcssa = $family$017;
        label = 90;
        break;
       }
       switch ($274<<24>>24) {
       case 45: case 32:  {
        $family$0$ph$be = $family$017;
        break L132;
        break;
       }
       default: {
       }
       }
       switch ($275<<24>>24) {
       case 0:  {
        $full$0$ph19$lcssa85 = $full$0$ph19;
        break L130;
        break;
       }
       case 45: case 32:  {
        break;
       }
       default: {
        break L124;
       }
       }
       $281 = ((($family$017)) + 1|0);
       if ($273) {
        break L124;
       } else {
        $family$017 = $281;
       }
      }
      if ((label|0) == 90) {
       label = 0;
       $277 = ((($family$017$lcssa)) + 1|0);
       $family$0$ph$be = $277;
      }
      $278 = ((($full$0$ph19)) + 1|0);
      $279 = HEAP8[$278>>0]|0;
      $280 = ($279<<24>>24)==(0);
      if ($280) {
       break L124;
      } else {
       $274 = $279;$family$0$ph20 = $family$0$ph$be;$full$0$ph19 = $278;
      }
     }
     HEAP32[$266>>2] = $full$0$ph19$lcssa85;
    }
   }
  }
 } while(0);
 $285 = ((($t42face)) + 28|0);
 HEAP32[$285>>2] = 0;
 $286 = ((($t42face)) + 32|0);
 HEAP32[$286>>2] = 0;
 HEAP32[$args>>2] = 9;
 $287 = HEAP32[$2>>2]|0;
 $288 = ((($287)) + 4|0);
 $289 = HEAP32[$288>>2]|0;
 $290 = (_FT_Get_Module($289,594296)|0);
 $291 = ((($args)) + 20|0);
 HEAP32[$291>>2] = $290;
 $292 = HEAP32[$15>>2]|0;
 $293 = ((($args)) + 4|0);
 HEAP32[$293>>2] = $292;
 $294 = ((($t42face)) + 496|0);
 $295 = HEAP32[$294>>2]|0;
 $296 = ((($args)) + 8|0);
 HEAP32[$296>>2] = $295;
 $297 = ($num_params|0)==(0);
 if (!($297)) {
  HEAP32[$args>>2] = 25;
  $298 = ((($args)) + 24|0);
  HEAP32[$298>>2] = $num_params;
  $299 = ((($args)) + 28|0);
  HEAP32[$299>>2] = $params;
 }
 $300 = HEAP32[$2>>2]|0;
 $301 = ((($300)) + 4|0);
 $302 = HEAP32[$301>>2]|0;
 $303 = (_FT_Open_Face($302,$args,0,$1)|0);
 $304 = ($303|0)==(0);
 if (!($304)) {
  $$02 = $303;
  STACKTOP = sp;return ($$02|0);
 }
 $305 = HEAP32[$1>>2]|0;
 $306 = ((($305)) + 88|0);
 $307 = HEAP32[$306>>2]|0;
 (_FT_Done_Size($307)|0);
 $308 = ((($t42face)) + 52|0);
 $309 = HEAP32[$1>>2]|0;
 $310 = ((($309)) + 52|0);
 ;HEAP32[$308>>2]=HEAP32[$310>>2]|0;HEAP32[$308+4>>2]=HEAP32[$310+4>>2]|0;HEAP32[$308+8>>2]=HEAP32[$310+8>>2]|0;HEAP32[$308+12>>2]=HEAP32[$310+12>>2]|0;
 $311 = ((($309)) + 68|0);
 $312 = HEAP32[$311>>2]|0;
 $313 = $312&65535;
 $314 = ((($t42face)) + 68|0);
 HEAP16[$314>>1] = $313;
 $315 = $312 >>> 16;
 $316 = $315&65535;
 $317 = ((($t42face)) + 70|0);
 HEAP16[$317>>1] = $316;
 $318 = ((($309)) + 72|0);
 $319 = HEAP32[$318>>2]|0;
 $320 = $319&65535;
 $321 = ((($t42face)) + 72|0);
 HEAP16[$321>>1] = $320;
 $322 = $319 >>> 16;
 $323 = $322&65535;
 $324 = ((($t42face)) + 74|0);
 HEAP16[$324>>1] = $323;
 $325 = ((($309)) + 76|0);
 $326 = HEAP32[$325>>2]|0;
 $327 = $326&65535;
 $328 = ((($t42face)) + 76|0);
 HEAP16[$328>>1] = $327;
 $329 = $326 >>> 16;
 $330 = $329&65535;
 $331 = ((($t42face)) + 78|0);
 HEAP16[$331>>1] = $330;
 $332 = ((($0)) + 26|0);
 $333 = HEAP16[$332>>1]|0;
 $334 = ((($t42face)) + 80|0);
 HEAP16[$334>>1] = $333;
 $335 = ((($t42face)) + 160|0);
 $336 = HEAP16[$335>>1]|0;
 $337 = ((($t42face)) + 82|0);
 HEAP16[$337>>1] = $336;
 $338 = ((($t42face)) + 12|0);
 $339 = ((($t42face)) + 152|0);
 $340 = HEAP32[$339>>2]|0;
 $not$ = ($340|0)!=(0);
 $$ = $not$&1;
 HEAP32[$338>>2] = $$;
 $341 = ((($309)) + 12|0);
 $342 = HEAP32[$341>>2]|0;
 $343 = $342 & 2;
 $344 = ($343|0)==(0);
 if (!($344)) {
  $345 = $$ | 2;
  HEAP32[$338>>2] = $345;
 }
 $346 = ((($309)) + 8|0);
 $347 = HEAP32[$346>>2]|0;
 $348 = $347 & 32;
 $349 = ($348|0)==(0);
 if (!($349)) {
  $350 = HEAP32[$254>>2]|0;
  $351 = $350 | 32;
  HEAP32[$254>>2] = $351;
 }
 $352 = ($4|0)==(0|0);
 if ($352) {
  $$02 = 0;
  STACKTOP = sp;return ($$02|0);
 }
 $353 = ((($9)) + 20|0);
 $354 = HEAP32[$353>>2]|0;
 HEAP32[$charmap>>2] = $t42face;
 $355 = ((($charmap)) + 8|0);
 HEAP16[$355>>1] = 3;
 $356 = ((($charmap)) + 10|0);
 HEAP16[$356>>1] = 1;
 $357 = ((($charmap)) + 4|0);
 HEAP32[$357>>2] = 1970170211;
 $358 = ((($354)) + 12|0);
 $359 = HEAP32[$358>>2]|0;
 $360 = (_FT_CMap_New($359,0,$charmap,0)|0);
 $361 = ($360|0)==(0);
 $362 = $360 & 255;
 $363 = ($362|0)==(163);
 $or$cond4 = $361 | $363;
 L157: do {
  if ($or$cond4) {
   HEAP16[$355>>1] = 7;
   $364 = ((($t42face)) + 368|0);
   $365 = HEAP32[$364>>2]|0;
   switch ($365|0) {
   case 2:  {
    HEAP32[$357>>2] = 1094995778;
    HEAP16[$356>>1] = 0;
    $clazz$0$in = $354;
    break;
   }
   case 4:  {
    HEAP32[$357>>2] = 1094992453;
    HEAP16[$356>>1] = 1;
    $366 = ((($354)) + 4|0);
    $clazz$0$in = $366;
    break;
   }
   case 1:  {
    HEAP32[$357>>2] = 1094992451;
    HEAP16[$356>>1] = 2;
    $367 = ((($354)) + 8|0);
    $clazz$0$in = $367;
    break;
   }
   case 3:  {
    HEAP32[$357>>2] = 1818326065;
    HEAP16[$356>>1] = 3;
    $clazz$0$in = $358;
    break;
   }
   default: {
    $error$1 = 0;
    break L157;
   }
   }
   $clazz$0 = HEAP32[$clazz$0$in>>2]|0;
   $368 = ($clazz$0|0)==(0|0);
   if ($368) {
    $error$1 = 0;
   } else {
    $369 = (_FT_CMap_New($clazz$0,0,$charmap,0)|0);
    $error$1 = $369;
   }
  } else {
   $error$1 = $360;
  }
 } while(0);
 $$02 = $error$1;
 STACKTOP = sp;return ($$02|0);
}
function _t42_parse_font_matrix($face,$loader) {
 $face = $face|0;
 $loader = $loader|0;
 var $$phi$trans$insert = 0, $$pre = 0, $$pre$phi4Z2D = 0, $$pre$phi6Z2D = 0, $$pre$phiZ2D = 0, $$pre1 = 0, $$pre2 = 0, $$pre3 = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $temp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $temp = sp;
 $0 = ((($loader)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (FUNCTION_TABLE_iiiii[$1 & 8191]($loader,6,$temp,3)|0);
 $3 = ($2|0)<(6);
 if ($3) {
  $4 = ((($loader)) + 12|0);
  HEAP32[$4>>2] = 3;
  STACKTOP = sp;return;
 }
 $5 = ((($temp)) + 12|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)<(0);
 $8 = (0 - ($6))|0;
 $9 = $7 ? $8 : $6;
 $10 = ($9|0)==(0);
 if ($10) {
  $11 = ((($loader)) + 12|0);
  HEAP32[$11>>2] = 3;
  STACKTOP = sp;return;
 }
 $12 = (_FT_DivFix(1000,$9)|0);
 $13 = $12&65535;
 $14 = ((($face)) + 68|0);
 HEAP16[$14>>1] = $13;
 $15 = ($9|0)==(65536);
 if ($15) {
  $$pre = HEAP32[$5>>2]|0;
  $$phi$trans$insert = ((($temp)) + 20|0);
  $$pre1 = HEAP32[$$phi$trans$insert>>2]|0;
  $$pre2 = ((($temp)) + 4|0);
  $$pre3 = ((($temp)) + 8|0);
  $$pre5 = ((($temp)) + 16|0);
  $$pre$phi4Z2D = $$pre3;$$pre$phi6Z2D = $$pre5;$$pre$phiZ2D = $$pre2;$41 = $$pre;$46 = $$pre1;
 } else {
  $16 = HEAP32[$temp>>2]|0;
  $17 = (_FT_DivFix($16,$9)|0);
  HEAP32[$temp>>2] = $17;
  $18 = ((($temp)) + 4|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = (_FT_DivFix($19,$9)|0);
  HEAP32[$18>>2] = $20;
  $21 = ((($temp)) + 8|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = (_FT_DivFix($22,$9)|0);
  HEAP32[$21>>2] = $23;
  $24 = ((($temp)) + 16|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = (_FT_DivFix($25,$9)|0);
  HEAP32[$24>>2] = $26;
  $27 = ((($temp)) + 20|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = (_FT_DivFix($28,$9)|0);
  HEAP32[$27>>2] = $29;
  $30 = HEAP32[$5>>2]|0;
  $31 = $30 >> 31;
  $32 = $31 & -131072;
  $33 = $32 | 65536;
  HEAP32[$5>>2] = $33;
  $$pre$phi4Z2D = $21;$$pre$phi6Z2D = $24;$$pre$phiZ2D = $18;$41 = $33;$46 = $29;
 }
 $34 = HEAP32[$temp>>2]|0;
 $35 = ((($face)) + 436|0);
 HEAP32[$35>>2] = $34;
 $36 = HEAP32[$$pre$phiZ2D>>2]|0;
 $37 = ((($face)) + 444|0);
 HEAP32[$37>>2] = $36;
 $38 = HEAP32[$$pre$phi4Z2D>>2]|0;
 $39 = ((($face)) + 440|0);
 HEAP32[$39>>2] = $38;
 $40 = ((($face)) + 448|0);
 HEAP32[$40>>2] = $41;
 $42 = HEAP32[$$pre$phi6Z2D>>2]|0;
 $43 = $42 >> 16;
 $44 = ((($face)) + 452|0);
 HEAP32[$44>>2] = $43;
 $45 = $46 >> 16;
 $47 = ((($face)) + 456|0);
 HEAP32[$47>>2] = $45;
 STACKTOP = sp;return;
}
function _t42_parse_encoding($face,$loader) {
 $face = $face|0;
 $loader = $loader|0;
 var $$lcssa9 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $count$0 = 0, $cur$3$us = 0, $cur$4 = 0, $error = 0, $exitcond = 0;
 var $n$025 = 0, $n$112 = 0, $n$112$us = 0, $n$5$us = 0, $only_immediates$0 = 0, $or$cond = 0, $or$cond4 = 0, $or$cond4$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = ((($loader)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 488|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($loader)) + 28|0);
 $5 = HEAP32[$4>>2]|0;
 FUNCTION_TABLE_vi[$5 & 8191]($loader);
 $6 = HEAP32[$loader>>2]|0;
 $7 = ($6>>>0)<($1>>>0);
 if (!($7)) {
  $8 = ((($loader)) + 12|0);
  HEAP32[$8>>2] = 3;
  STACKTOP = sp;return;
 }
 $9 = HEAP8[$6>>0]|0;
 $10 = $9&255;
 $11 = (($10) + -48)|0;
 $12 = ($11>>>0)<(10);
 $13 = ($9<<24>>24)==(91);
 $or$cond = $13 | $12;
 if (!($or$cond)) {
  $137 = ((($6)) + 17|0);
  $138 = ($137>>>0)<($1>>>0);
  if ($138) {
   $139 = (_strncmp($6,594610,16)|0);
   $140 = ($139|0)==(0);
   if ($140) {
    $141 = ((($face)) + 368|0);
    HEAP32[$141>>2] = 2;
    STACKTOP = sp;return;
   }
  }
  $142 = ((($6)) + 15|0);
  $143 = ($142>>>0)<($1>>>0);
  if ($143) {
   $144 = (_strncmp($6,594627,14)|0);
   $145 = ($144|0)==(0);
   if ($145) {
    $146 = ((($face)) + 368|0);
    HEAP32[$146>>2] = 4;
    STACKTOP = sp;return;
   }
  }
  $147 = ((($6)) + 18|0);
  $148 = ($147>>>0)<($1>>>0);
  if ($148) {
   $149 = (_strncmp($6,594642,17)|0);
   $150 = ($149|0)==(0);
   if ($150) {
    $151 = ((($face)) + 368|0);
    HEAP32[$151>>2] = 3;
    STACKTOP = sp;return;
   }
  }
  $152 = ((($loader)) + 12|0);
  HEAP32[$152>>2] = 162;
  STACKTOP = sp;return;
 }
 $14 = ((($loader)) + 92|0);
 $15 = ((($loader)) + 16|0);
 $16 = HEAP32[$15>>2]|0;
 if ($13) {
  $17 = ((($6)) + 1|0);
  HEAP32[$loader>>2] = $17;
  $count$0 = 256;$only_immediates$0 = 1;
 } else {
  $18 = ((($loader)) + 36|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = (FUNCTION_TABLE_ii[$19 & 8191]($loader)|0);
  $count$0 = $20;$only_immediates$0 = 0;
 }
 $21 = HEAP32[$4>>2]|0;
 FUNCTION_TABLE_vi[$21 & 8191]($loader);
 $22 = HEAP32[$loader>>2]|0;
 $23 = ($22>>>0)<($1>>>0);
 if (!($23)) {
  STACKTOP = sp;return;
 }
 $24 = ((($face)) + 372|0);
 HEAP32[$24>>2] = $count$0;
 $25 = ((($loader)) + 88|0);
 HEAP32[$25>>2] = $count$0;
 $26 = (_ft_mem_realloc($16,2,0,$count$0,0,$error)|0);
 $27 = ((($face)) + 384|0);
 HEAP32[$27>>2] = $26;
 $28 = HEAP32[$error>>2]|0;
 $29 = ($28|0)==(0);
 if ($29) {
  $30 = (_ft_mem_realloc($16,4,0,$count$0,0,$error)|0);
  $31 = ((($face)) + 388|0);
  HEAP32[$31>>2] = $30;
  $32 = HEAP32[$error>>2]|0;
  $33 = ($32|0)==(0);
  if ($33) {
   $34 = HEAP32[$3>>2]|0;
   $35 = HEAP32[$34>>2]|0;
   $36 = (FUNCTION_TABLE_iiii[$35 & 8191]($14,$count$0,$16)|0);
   HEAP32[$error>>2] = $36;
   $37 = ($36|0)==(0);
   if ($37) {
    $38 = ($count$0|0)>(0);
    if ($38) {
     $39 = ((($loader)) + 136|0);
     $n$025 = 0;
     while(1) {
      $42 = HEAP32[$39>>2]|0;
      (FUNCTION_TABLE_iiiii[$42 & 8191]($14,$n$025,594602,8)|0);
      $43 = (($n$025) + 1)|0;
      $exitcond = ($43|0)==($count$0|0);
      if ($exitcond) {
       break;
      } else {
       $n$025 = $43;
      }
     }
    }
    $44 = HEAP32[$4>>2]|0;
    FUNCTION_TABLE_vi[$44 & 8191]($loader);
    $45 = HEAP32[$loader>>2]|0;
    $46 = ($45>>>0)<($1>>>0);
    L40: do {
     if ($46) {
      $47 = ($only_immediates$0<<24>>24)==(0);
      $48 = ((($loader)) + 32|0);
      $49 = ((($loader)) + 12|0);
      $50 = ((($loader)) + 136|0);
      $51 = ((($loader)) + 116|0);
      $52 = ((($loader)) + 36|0);
      L42: do {
       if ($47) {
        $54 = $45;$n$112$us = 0;
        while(1) {
         $53 = HEAP8[$54>>0]|0;
         switch ($53<<24>>24) {
         case 93:  {
          $$lcssa9 = $54;
          break L42;
          break;
         }
         case 100:  {
          $55 = ((($54)) + 3|0);
          $56 = ($55>>>0)<($1>>>0);
          if ($56) {
           $57 = ((($54)) + 1|0);
           $58 = HEAP8[$57>>0]|0;
           $59 = ($58<<24>>24)==(101);
           if ($59) {
            $60 = ((($54)) + 2|0);
            $61 = HEAP8[$60>>0]|0;
            $62 = ($61<<24>>24)==(102);
            if ($62) {
             $63 = HEAP8[$55>>0]|0;
             switch ($63<<24>>24) {
             case 0: case 12: case 10: case 13: case 9: case 32:  {
              $cur$4 = $55;
              break L40;
              break;
             }
             default: {
             }
             }
            }
           }
          }
          break;
         }
         default: {
         }
         }
         $64 = $53&255;
         $65 = (($64) + -48)|0;
         $66 = ($65>>>0)>(9);
         if ($66) {
          $93 = HEAP32[$48>>2]|0;
          FUNCTION_TABLE_vi[$93 & 8191]($loader);
          $94 = HEAP32[$49>>2]|0;
          $95 = ($94|0)==(0);
          if ($95) {
           $cur$3$us = $54;$n$5$us = $n$112$us;
          } else {
           label = 45;
           break;
          }
         } else {
          $67 = HEAP32[$52>>2]|0;
          $68 = (FUNCTION_TABLE_ii[$67 & 8191]($loader)|0);
          $69 = HEAP32[$4>>2]|0;
          FUNCTION_TABLE_vi[$69 & 8191]($loader);
          $$pre = HEAP32[$loader>>2]|0;
          $70 = ((($$pre)) + 2|0);
          $71 = ($70>>>0)<($1>>>0);
          if ($71) {
           $72 = HEAP8[$$pre>>0]|0;
           $73 = ($72<<24>>24)==(47);
           $74 = ($n$112$us|0)<($count$0|0);
           $or$cond4$us = $74 & $73;
           if ($or$cond4$us) {
            $75 = ((($$pre)) + 1|0);
            HEAP32[$loader>>2] = $75;
            $76 = HEAP32[$48>>2]|0;
            FUNCTION_TABLE_vi[$76 & 8191]($loader);
            $77 = HEAP32[$loader>>2]|0;
            $78 = ($77>>>0)<($1>>>0);
            if (!($78)) {
             label = 45;
             break;
            }
            $79 = HEAP32[$49>>2]|0;
            $80 = ($79|0)==(0);
            if (!($80)) {
             label = 45;
             break;
            }
            $81 = $77;
            $82 = $75;
            $83 = (($81) - ($82))|0;
            $84 = HEAP32[$50>>2]|0;
            $85 = (($83) + 1)|0;
            $86 = (FUNCTION_TABLE_iiiii[$84 & 8191]($14,$68,$75,$85)|0);
            HEAP32[$49>>2] = $86;
            $87 = ($86|0)==(0);
            if (!($87)) {
             label = 45;
             break;
            }
            $88 = HEAP32[$51>>2]|0;
            $89 = (($88) + ($68<<2)|0);
            $90 = HEAP32[$89>>2]|0;
            $91 = (($90) + ($83)|0);
            HEAP8[$91>>0] = 0;
            $92 = (($n$112$us) + 1)|0;
            $cur$3$us = $75;$n$5$us = $92;
           } else {
            $cur$3$us = $$pre;$n$5$us = $n$112$us;
           }
          } else {
           $cur$3$us = $$pre;$n$5$us = $n$112$us;
          }
         }
         $96 = HEAP32[$4>>2]|0;
         FUNCTION_TABLE_vi[$96 & 8191]($loader);
         $97 = HEAP32[$loader>>2]|0;
         $98 = ($97>>>0)<($1>>>0);
         if ($98) {
          $54 = $97;$n$112$us = $n$5$us;
         } else {
          $cur$4 = $cur$3$us;
          break L40;
         }
        }
        if ((label|0) == 45) {
         STACKTOP = sp;return;
        }
       } else {
        $100 = $45;$n$112 = 0;
        while(1) {
         $99 = HEAP8[$100>>0]|0;
         switch ($99<<24>>24) {
         case 93:  {
          $$lcssa9 = $100;
          break L42;
          break;
         }
         case 100:  {
          $101 = ((($100)) + 3|0);
          $102 = ($101>>>0)<($1>>>0);
          if ($102) {
           $103 = ((($100)) + 1|0);
           $104 = HEAP8[$103>>0]|0;
           $105 = ($104<<24>>24)==(101);
           if ($105) {
            $106 = ((($100)) + 2|0);
            $107 = HEAP8[$106>>0]|0;
            $108 = ($107<<24>>24)==(102);
            if ($108) {
             $109 = HEAP8[$101>>0]|0;
             switch ($109<<24>>24) {
             case 0: case 12: case 10: case 13: case 9: case 32:  {
              $cur$4 = $101;
              break L40;
              break;
             }
             default: {
             }
             }
            }
           }
          }
          break;
         }
         default: {
         }
         }
         $111 = ((($100)) + 2|0);
         $112 = ($111>>>0)<($1>>>0);
         if (!($112)) {
          label = 42;
          break;
         }
         $113 = ($99<<24>>24)==(47);
         $114 = ($n$112|0)<($count$0|0);
         $or$cond4 = $114 & $113;
         if (!($or$cond4)) {
          label = 42;
          break;
         }
         $115 = ((($100)) + 1|0);
         HEAP32[$loader>>2] = $115;
         $116 = HEAP32[$48>>2]|0;
         FUNCTION_TABLE_vi[$116 & 8191]($loader);
         $117 = HEAP32[$loader>>2]|0;
         $118 = ($117>>>0)<($1>>>0);
         if (!($118)) {
          label = 45;
          break;
         }
         $119 = HEAP32[$49>>2]|0;
         $120 = ($119|0)==(0);
         if (!($120)) {
          label = 45;
          break;
         }
         $121 = $117;
         $122 = $115;
         $123 = (($121) - ($122))|0;
         $124 = HEAP32[$50>>2]|0;
         $125 = (($123) + 1)|0;
         $126 = (FUNCTION_TABLE_iiiii[$124 & 8191]($14,$n$112,$115,$125)|0);
         HEAP32[$49>>2] = $126;
         $127 = ($126|0)==(0);
         if (!($127)) {
          label = 45;
          break;
         }
         $128 = HEAP32[$51>>2]|0;
         $129 = (($128) + ($n$112<<2)|0);
         $130 = HEAP32[$129>>2]|0;
         $131 = (($130) + ($123)|0);
         HEAP8[$131>>0] = 0;
         $132 = (($n$112) + 1)|0;
         $133 = HEAP32[$4>>2]|0;
         FUNCTION_TABLE_vi[$133 & 8191]($loader);
         $134 = HEAP32[$loader>>2]|0;
         $135 = ($134>>>0)<($1>>>0);
         if ($135) {
          $100 = $134;$n$112 = $132;
         } else {
          $cur$4 = $115;
          break L40;
         }
        }
        if ((label|0) == 42) {
         HEAP32[$49>>2] = 2;
         STACKTOP = sp;return;
        }
        else if ((label|0) == 45) {
         STACKTOP = sp;return;
        }
       }
      } while(0);
      $110 = ((($$lcssa9)) + 1|0);
      $cur$4 = $110;
     } else {
      $cur$4 = $6;
     }
    } while(0);
    $136 = ((($face)) + 368|0);
    HEAP32[$136>>2] = 1;
    HEAP32[$loader>>2] = $cur$4;
    STACKTOP = sp;return;
   } else {
    $41 = $36;
   }
  } else {
   $41 = $32;
  }
 } else {
  $41 = $28;
 }
 $40 = ((($loader)) + 12|0);
 HEAP32[$40>>2] = $41;
 STACKTOP = sp;return;
}
function _t42_parse_charstrings($face,$loader) {
 $face = $face|0;
 $loader = $loader|0;
 var $$cast = 0, $$lcssa118 = 0, $$lcssa120 = 0, $$notdef_found$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0;
 var $201 = 0, $202 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $count$024 = 0, $count$024$lcssa = 0, $count$1 = 0, $error$1 = 0, $n$0$notdef_index$0 = 0, $n$0$ph = 0, $n$2 = 0, $notdef_found$0$ph = 0, $notdef_found$1 = 0, $notdef_found$3 = 0, $notdef_index$0$ph = 0, $notdef_index$1 = 0;
 var $notdef_index$3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($loader)) + 200|0);
 $1 = ((($loader)) + 148|0);
 $2 = ((($loader)) + 252|0);
 $3 = ((($loader)) + 16|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($face)) + 488|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($loader)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($loader)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 FUNCTION_TABLE_vi[$10 & 8191]($loader);
 $11 = HEAP32[$loader>>2]|0;
 $12 = ($11>>>0)<($8>>>0);
 L1: do {
  if ($12) {
   $13 = HEAP8[$11>>0]|0;
   $14 = $13&255;
   $15 = (($14) + -48)|0;
   $16 = ($15>>>0)<(10);
   L3: do {
    if ($16) {
     $17 = ((($loader)) + 36|0);
     $18 = HEAP32[$17>>2]|0;
     $19 = (FUNCTION_TABLE_ii[$18 & 8191]($loader)|0);
     $20 = ((($loader)) + 144|0);
     HEAP32[$20>>2] = $19;
     $21 = ((($loader)) + 12|0);
     $22 = HEAP32[$21>>2]|0;
     $23 = ($22|0)==(0);
     if (!($23)) {
      return;
     }
     $24 = ($19|0)<(0);
     if ($24) {
      $error$1 = 3;
      break L1;
     }
     $$pre = HEAP32[$loader>>2]|0;
     $45 = $$pre;
    } else {
     $25 = ($13<<24>>24)==(60);
     if (!($25)) {
      $error$1 = 3;
      break L1;
     }
     $26 = ((($loader)) + 32|0);
     $27 = HEAP32[$26>>2]|0;
     FUNCTION_TABLE_vi[$27 & 8191]($loader);
     $28 = ((($loader)) + 12|0);
     $29 = HEAP32[$28>>2]|0;
     $30 = ($29|0)==(0);
     if (!($30)) {
      return;
     }
     $31 = HEAP32[$9>>2]|0;
     FUNCTION_TABLE_vi[$31 & 8191]($loader);
     $32 = HEAP32[$loader>>2]|0;
     $$cast = $32;
     $33 = ($$cast>>>0)<($8>>>0);
     if ($33) {
      $35 = $$cast;$count$024 = 0;
      L14: while(1) {
       $34 = HEAP8[$35>>0]|0;
       switch ($34<<24>>24) {
       case 62:  {
        $count$024$lcssa = $count$024;
        label = 11;
        break L14;
        break;
       }
       case 47:  {
        $36 = (($count$024) + 1)|0;
        $count$1 = $36;
        break;
       }
       default: {
        $count$1 = $count$024;
       }
       }
       $38 = HEAP32[$26>>2]|0;
       FUNCTION_TABLE_vi[$38 & 8191]($loader);
       $39 = HEAP32[$28>>2]|0;
       $40 = ($39|0)==(0);
       if (!($40)) {
        label = 47;
        break;
       }
       $41 = HEAP32[$9>>2]|0;
       FUNCTION_TABLE_vi[$41 & 8191]($loader);
       $42 = HEAP32[$loader>>2]|0;
       $43 = ($42>>>0)<($8>>>0);
       if ($43) {
        $35 = $42;$count$024 = $count$1;
       } else {
        $45 = $42;
        break L3;
       }
      }
      if ((label|0) == 11) {
       $37 = ((($loader)) + 144|0);
       HEAP32[$37>>2] = $count$024$lcssa;
       HEAP32[$loader>>2] = $32;
       $45 = $$cast;
       break;
      }
      else if ((label|0) == 47) {
       return;
      }
     } else {
      $45 = $$cast;
     }
    }
   } while(0);
   $44 = ($45>>>0)<($8>>>0);
   if ($44) {
    $46 = HEAP32[$6>>2]|0;
    $47 = HEAP32[$46>>2]|0;
    $48 = ((($loader)) + 144|0);
    $49 = HEAP32[$48>>2]|0;
    $50 = (FUNCTION_TABLE_iiii[$47 & 8191]($0,$49,$4)|0);
    $51 = ($50|0)==(0);
    if ($51) {
     $52 = HEAP32[$6>>2]|0;
     $53 = HEAP32[$52>>2]|0;
     $54 = HEAP32[$48>>2]|0;
     $55 = (FUNCTION_TABLE_iiii[$53 & 8191]($1,$54,$4)|0);
     $56 = ($55|0)==(0);
     if ($56) {
      $57 = HEAP32[$6>>2]|0;
      $58 = HEAP32[$57>>2]|0;
      $59 = (FUNCTION_TABLE_iiii[$58 & 8191]($2,4,$4)|0);
      $60 = ($59|0)==(0);
      if ($60) {
       $61 = ((($loader)) + 32|0);
       $62 = ((($loader)) + 12|0);
       $63 = ((($loader)) + 192|0);
       $64 = ((($loader)) + 172|0);
       $65 = ((($loader)) + 36|0);
       $66 = ((($loader)) + 244|0);
       $67 = ((($loader)) + 224|0);
       $n$0$ph = 0;$notdef_found$0$ph = 0;$notdef_index$0$ph = 0;
       L28: while(1) {
        while(1) {
         $68 = HEAP32[$9>>2]|0;
         FUNCTION_TABLE_vi[$68 & 8191]($loader);
         $69 = HEAP32[$loader>>2]|0;
         $70 = ($69>>>0)<($8>>>0);
         if (!($70)) {
          $n$2 = $n$0$ph;$notdef_found$3 = $notdef_found$0$ph;$notdef_index$3 = $notdef_index$0$ph;
          break L28;
         }
         $71 = HEAP8[$69>>0]|0;
         switch ($71<<24>>24) {
         case 62:  {
          $n$2 = $n$0$ph;$notdef_found$3 = $notdef_found$0$ph;$notdef_index$3 = $notdef_index$0$ph;
          break L28;
          break;
         }
         case 101:  {
          $72 = ((($69)) + 3|0);
          $73 = ($72>>>0)<($8>>>0);
          if ($73) {
           $74 = ((($69)) + 1|0);
           $75 = HEAP8[$74>>0]|0;
           $76 = ($75<<24>>24)==(110);
           if ($76) {
            $77 = ((($69)) + 2|0);
            $78 = HEAP8[$77>>0]|0;
            $79 = ($78<<24>>24)==(100);
            if ($79) {
             $80 = HEAP8[$72>>0]|0;
             switch ($80<<24>>24) {
             case 0: case 12: case 10: case 13: case 9: case 32:  {
              $n$2 = $n$0$ph;$notdef_found$3 = $notdef_found$0$ph;$notdef_index$3 = $notdef_index$0$ph;
              break L28;
              break;
             }
             default: {
             }
             }
            }
           }
          }
          break;
         }
         default: {
         }
         }
         $81 = HEAP32[$61>>2]|0;
         FUNCTION_TABLE_vi[$81 & 8191]($loader);
         $82 = HEAP32[$loader>>2]|0;
         $83 = ($82>>>0)<($8>>>0);
         if (!($83)) {
          $error$1 = 3;
          break L1;
         }
         $84 = HEAP32[$62>>2]|0;
         $85 = ($84|0)==(0);
         if (!($85)) {
          label = 47;
          break L28;
         }
         $86 = HEAP8[$69>>0]|0;
         $87 = ($86<<24>>24)==(47);
         if ($87) {
          $$lcssa118 = $69;$$lcssa120 = $82;
          break;
         }
        }
        $88 = ((($$lcssa118)) + 2|0);
        $89 = ($88>>>0)<($8>>>0);
        if (!($89)) {
         $error$1 = 3;
         break L1;
        }
        $90 = ((($$lcssa118)) + 1|0);
        $91 = $$lcssa120;
        $92 = $90;
        $93 = (($91) - ($92))|0;
        $94 = HEAP32[$63>>2]|0;
        $95 = (($93) + 1)|0;
        $96 = (FUNCTION_TABLE_iiiii[$94 & 8191]($1,$n$0$ph,$90,$95)|0);
        $97 = ($96|0)==(0);
        if (!($97)) {
         $error$1 = $96;
         break L1;
        }
        $98 = HEAP32[$64>>2]|0;
        $99 = (($98) + ($n$0$ph<<2)|0);
        $100 = HEAP32[$99>>2]|0;
        $101 = (($100) + ($93)|0);
        HEAP8[$101>>0] = 0;
        $102 = HEAP8[$90>>0]|0;
        $103 = ($102<<24>>24)==(46);
        if ($103) {
         $104 = HEAP32[$64>>2]|0;
         $105 = (($104) + ($n$0$ph<<2)|0);
         $106 = HEAP32[$105>>2]|0;
         $107 = (_strcmp(594602,$106)|0);
         $108 = ($107|0)==(0);
         $n$0$notdef_index$0 = $108 ? $n$0$ph : $notdef_index$0$ph;
         $$notdef_found$0 = $108 ? 1 : $notdef_found$0$ph;
         $notdef_found$1 = $$notdef_found$0;$notdef_index$1 = $n$0$notdef_index$0;
        } else {
         $notdef_found$1 = $notdef_found$0$ph;$notdef_index$1 = $notdef_index$0$ph;
        }
        $109 = HEAP32[$9>>2]|0;
        FUNCTION_TABLE_vi[$109 & 8191]($loader);
        $110 = HEAP32[$loader>>2]|0;
        $111 = HEAP32[$65>>2]|0;
        (FUNCTION_TABLE_ii[$111 & 8191]($loader)|0);
        $112 = HEAP32[$loader>>2]|0;
        $113 = ($112>>>0)<($8>>>0);
        if (!($113)) {
         $error$1 = 3;
         break L1;
        }
        $114 = $112;
        $115 = $110;
        $116 = (($114) - ($115))|0;
        $117 = HEAP32[$66>>2]|0;
        $118 = (($116) + 1)|0;
        $119 = (FUNCTION_TABLE_iiiii[$117 & 8191]($0,$n$0$ph,$110,$118)|0);
        $120 = ($119|0)==(0);
        if (!($120)) {
         $error$1 = $119;
         break L1;
        }
        $121 = HEAP32[$67>>2]|0;
        $122 = (($121) + ($n$0$ph<<2)|0);
        $123 = HEAP32[$122>>2]|0;
        $124 = (($123) + ($116)|0);
        HEAP8[$124>>0] = 0;
        $125 = (($n$0$ph) + 1)|0;
        $126 = HEAP32[$48>>2]|0;
        $127 = ($125|0)<($126|0);
        if ($127) {
         $n$0$ph = $125;$notdef_found$0$ph = $notdef_found$1;$notdef_index$0$ph = $notdef_index$1;
        } else {
         $n$2 = $125;$notdef_found$3 = $notdef_found$1;$notdef_index$3 = $notdef_index$1;
         break;
        }
       }
       if ((label|0) == 47) {
        return;
       }
       HEAP32[$48>>2] = $n$2;
       $128 = ($notdef_found$3<<24>>24)==(0);
       if ($128) {
        $error$1 = 3;
       } else {
        $129 = HEAP32[$64>>2]|0;
        $130 = HEAP32[$129>>2]|0;
        $131 = (_strcmp(594602,$130)|0);
        $132 = ($131|0)==(0);
        if ($132) {
         return;
        }
        $133 = ((($loader)) + 296|0);
        $134 = HEAP32[$133>>2]|0;
        $135 = ((($loader)) + 176|0);
        $136 = HEAP32[$135>>2]|0;
        $137 = HEAP32[$136>>2]|0;
        $138 = (FUNCTION_TABLE_iiiii[$134 & 8191]($2,0,$130,$137)|0);
        $139 = ($138|0)==(0);
        if ($139) {
         $140 = HEAP32[$133>>2]|0;
         $141 = HEAP32[$67>>2]|0;
         $142 = HEAP32[$141>>2]|0;
         $143 = ((($loader)) + 228|0);
         $144 = HEAP32[$143>>2]|0;
         $145 = HEAP32[$144>>2]|0;
         $146 = (FUNCTION_TABLE_iiiii[$140 & 8191]($2,1,$142,$145)|0);
         $147 = ($146|0)==(0);
         if ($147) {
          $148 = HEAP32[$133>>2]|0;
          $149 = HEAP32[$64>>2]|0;
          $150 = (($149) + ($notdef_index$3<<2)|0);
          $151 = HEAP32[$150>>2]|0;
          $152 = HEAP32[$135>>2]|0;
          $153 = (($152) + ($notdef_index$3<<2)|0);
          $154 = HEAP32[$153>>2]|0;
          $155 = (FUNCTION_TABLE_iiiii[$148 & 8191]($2,2,$151,$154)|0);
          $156 = ($155|0)==(0);
          if ($156) {
           $157 = HEAP32[$133>>2]|0;
           $158 = HEAP32[$67>>2]|0;
           $159 = (($158) + ($notdef_index$3<<2)|0);
           $160 = HEAP32[$159>>2]|0;
           $161 = HEAP32[$143>>2]|0;
           $162 = (($161) + ($notdef_index$3<<2)|0);
           $163 = HEAP32[$162>>2]|0;
           $164 = (FUNCTION_TABLE_iiiii[$157 & 8191]($2,3,$160,$163)|0);
           $165 = ($164|0)==(0);
           if ($165) {
            $166 = HEAP32[$63>>2]|0;
            $167 = ((($loader)) + 276|0);
            $168 = HEAP32[$167>>2]|0;
            $169 = HEAP32[$168>>2]|0;
            $170 = ((($loader)) + 280|0);
            $171 = HEAP32[$170>>2]|0;
            $172 = HEAP32[$171>>2]|0;
            $173 = (FUNCTION_TABLE_iiiii[$166 & 8191]($1,$notdef_index$3,$169,$172)|0);
            $174 = ($173|0)==(0);
            if ($174) {
             $175 = HEAP32[$66>>2]|0;
             $176 = HEAP32[$167>>2]|0;
             $177 = ((($176)) + 4|0);
             $178 = HEAP32[$177>>2]|0;
             $179 = HEAP32[$170>>2]|0;
             $180 = ((($179)) + 4|0);
             $181 = HEAP32[$180>>2]|0;
             $182 = (FUNCTION_TABLE_iiiii[$175 & 8191]($0,$notdef_index$3,$178,$181)|0);
             $183 = ($182|0)==(0);
             if ($183) {
              $184 = HEAP32[$63>>2]|0;
              $185 = HEAP32[$167>>2]|0;
              $186 = ((($185)) + 8|0);
              $187 = HEAP32[$186>>2]|0;
              $188 = HEAP32[$170>>2]|0;
              $189 = ((($188)) + 8|0);
              $190 = HEAP32[$189>>2]|0;
              $191 = (FUNCTION_TABLE_iiiii[$184 & 8191]($1,0,$187,$190)|0);
              $192 = ($191|0)==(0);
              if ($192) {
               $193 = HEAP32[$66>>2]|0;
               $194 = HEAP32[$167>>2]|0;
               $195 = ((($194)) + 12|0);
               $196 = HEAP32[$195>>2]|0;
               $197 = HEAP32[$170>>2]|0;
               $198 = ((($197)) + 12|0);
               $199 = HEAP32[$198>>2]|0;
               $200 = (FUNCTION_TABLE_iiiii[$193 & 8191]($0,0,$196,$199)|0);
               $201 = ($200|0)==(0);
               if ($201) {
                return;
               } else {
                $error$1 = $200;
               }
              } else {
               $error$1 = $191;
              }
             } else {
              $error$1 = $182;
             }
            } else {
             $error$1 = $173;
            }
           } else {
            $error$1 = $164;
           }
          } else {
           $error$1 = $155;
          }
         } else {
          $error$1 = $146;
         }
        } else {
         $error$1 = $138;
        }
       }
      } else {
       $error$1 = $59;
      }
     } else {
      $error$1 = $55;
     }
    } else {
     $error$1 = $50;
    }
   } else {
    $error$1 = 3;
   }
  } else {
   $error$1 = 3;
  }
 } while(0);
 $202 = ((($loader)) + 12|0);
 HEAP32[$202>>2] = $error$1;
 return;
}
function _t42_parse_sfnts($face,$loader) {
 $face = $face|0;
 $loader = $loader|0;
 var $$lcssa295 = 0, $$lcssa337 = 0, $$pre = 0, $$string_size$2 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $allocated$0$lcssa = 0, $allocated$079 = 0, $allocated$079$lcssa = 0, $allocated$079$lcssa310 = 0, $allocated$079$lcssa313 = 0, $allocated$1 = 0, $allocated$1$lcssa = 0, $allocated$16 = 0, $allocated$16$lcssa = 0, $allocated$16$lcssa338 = 0, $allocated$16$lcssa340 = 0, $allocated$2 = 0, $allocated$3 = 0, $count$081 = 0;
 var $count$169 = 0, $count$2 = 0, $count$2$lcssa = 0, $error = 0, $exitcond = 0, $i$064 = 0, $n$066 = 0, $num_tables$082 = 0, $num_tables$171 = 0, $num_tables$2 = 0, $num_tables$3 = 0, $num_tables$4 = 0, $num_tables$4$lcssa = 0, $old_string_size$077 = 0, $old_string_size$1 = 0, $old_string_size$14 = 0, $real_size = 0, $status$080 = 0, $status$165 = 0, $status$4 = 0;
 var $status$4$lcssa = 0, $string_buf$0$lcssa = 0, $string_buf$078 = 0, $string_buf$078$lcssa = 0, $string_buf$078$lcssa323 = 0, $string_buf$078$lcssa326 = 0, $string_buf$078$lcssa327 = 0, $string_buf$2 = 0, $string_buf$25 = 0, $string_buf$25$lcssa = 0, $string_buf$25$lcssa342 = 0, $string_buf$25$lcssa344 = 0, $string_buf$3 = 0, $string_buf$4 = 0, $string_size$076 = 0, $string_size$2 = 0, $string_size$23 = 0, $string_size$3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp + 4|0;
 $real_size = sp;
 $0 = ((($loader)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($loader)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($loader)) + 28|0);
 $5 = HEAP32[$4>>2]|0;
 FUNCTION_TABLE_vi[$5 & 8191]($loader);
 $6 = HEAP32[$loader>>2]|0;
 $7 = ($6>>>0)<($3>>>0);
 L1: do {
  if ($7) {
   $8 = ((($6)) + 1|0);
   HEAP32[$loader>>2] = $8;
   $9 = HEAP8[$6>>0]|0;
   $10 = ($9<<24>>24)==(91);
   if ($10) {
    $11 = HEAP32[$4>>2]|0;
    FUNCTION_TABLE_vi[$11 & 8191]($loader);
    $12 = HEAP32[$loader>>2]|0;
    $13 = ($12>>>0)<($3>>>0);
    L4: do {
     if ($13) {
      $14 = ((($loader)) + 36|0);
      $15 = ((($loader)) + 32|0);
      $16 = ((($loader)) + 12|0);
      $17 = $3;
      $18 = ((($face)) + 492|0);
      $19 = ((($face)) + 496|0);
      $20 = ((($loader)) + 44|0);
      $22 = $12;$allocated$079 = 0;$count$081 = 0;$num_tables$082 = 0;$old_string_size$077 = 0;$status$080 = 0;$string_buf$078 = 0;$string_size$076 = 0;
      L6: while(1) {
       $21 = HEAP8[$22>>0]|0;
       switch ($21<<24>>24) {
       case 93:  {
        $$lcssa295 = $22;$allocated$079$lcssa = $allocated$079;$string_buf$078$lcssa = $string_buf$078;
        label = 7;
        break L6;
        break;
       }
       case 60:  {
        $24 = HEAP32[$15>>2]|0;
        FUNCTION_TABLE_vi[$24 & 8191]($loader);
        $25 = HEAP32[$16>>2]|0;
        $26 = ($25|0)==(0);
        if (!($26)) {
         $allocated$3 = $allocated$079;$string_buf$4 = $string_buf$078;
         break L1;
        }
        $27 = HEAP32[$loader>>2]|0;
        $28 = $22;
        $29 = (($27) - ($28))|0;
        $30 = (($29) + -1)|0;
        $31 = (($30|0) / 2)&-1;
        $32 = ($29>>>0)>(2);
        if (!($32)) {
         $allocated$079$lcssa310 = $allocated$079;$string_buf$078$lcssa323 = $string_buf$078;
         label = 10;
         break L6;
        }
        $33 = (_ft_mem_realloc($1,1,$old_string_size$077,$31,$string_buf$078,$error)|0);
        $34 = HEAP32[$error>>2]|0;
        $35 = ($34|0)==(0);
        if (!($35)) {
         $140 = $34;$allocated$2 = $allocated$079;$string_buf$3 = $33;
         label = 46;
         break L1;
        }
        HEAP32[$loader>>2] = $22;
        $36 = HEAP32[$20>>2]|0;
        (FUNCTION_TABLE_iiiiii[$36 & 8191]($loader,$33,$31,$real_size,1)|0);
        $37 = HEAP32[$real_size>>2]|0;
        $allocated$1 = 1;$old_string_size$1 = $31;$string_buf$2 = $33;$string_size$2 = $37;
        label = 22;
        break;
       }
       default: {
        $38 = $21&255;
        $39 = (($38) + -48)|0;
        $40 = ($39>>>0)<(10);
        if ($40) {
         $41 = ($allocated$079<<24>>24)==(0);
         if (!($41)) {
          $allocated$079$lcssa313 = $allocated$079;$string_buf$078$lcssa326 = $string_buf$078;
          label = 15;
          break L6;
         }
         $42 = HEAP32[$14>>2]|0;
         $43 = (FUNCTION_TABLE_ii[$42 & 8191]($loader)|0);
         $44 = ($43|0)<(0);
         if ($44) {
          $string_buf$078$lcssa327 = $string_buf$078;
          label = 17;
          break L6;
         }
         $45 = HEAP32[$15>>2]|0;
         FUNCTION_TABLE_vi[$45 & 8191]($loader);
         $46 = HEAP32[$16>>2]|0;
         $47 = ($46|0)==(0);
         if (!($47)) {
          label = 49;
          break L6;
         }
         $48 = HEAP32[$loader>>2]|0;
         $49 = ((($48)) + 1|0);
         $50 = $48;
         $51 = (($17) - ($50))|0;
         $52 = ($51>>>0)<($43>>>0);
         if ($52) {
          $$lcssa337 = $49;
          label = 20;
          break L6;
         }
         $53 = (($43) + 1)|0;
         $54 = (($48) + ($53)|0);
         HEAP32[$loader>>2] = $54;
         $allocated$16 = 0;$old_string_size$14 = $old_string_size$077;$string_buf$25 = $49;$string_size$23 = $43;
        } else {
         $allocated$1 = $allocated$079;$old_string_size$1 = $old_string_size$077;$string_buf$2 = $string_buf$078;$string_size$2 = $string_size$076;
         label = 22;
        }
       }
       }
       if ((label|0) == 22) {
        label = 0;
        $55 = ($string_buf$2|0)==(0|0);
        if ($55) {
         $allocated$1$lcssa = $allocated$1;
         label = 23;
         break;
        } else {
         $allocated$16 = $allocated$1;$old_string_size$14 = $old_string_size$1;$string_buf$25 = $string_buf$2;$string_size$23 = $string_size$2;
        }
       }
       $56 = $string_size$23 & 1;
       $57 = ($56|0)==(0);
       if ($57) {
        $string_size$3 = $string_size$23;
       } else {
        $58 = (($string_size$23) + -1)|0;
        $59 = (($string_buf$25) + ($58)|0);
        $60 = HEAP8[$59>>0]|0;
        $61 = ($60<<24>>24)==(0);
        $$string_size$2 = $61 ? $58 : $string_size$23;
        $string_size$3 = $$string_size$2;
       }
       $62 = ($string_size$3|0)==(0);
       if ($62) {
        $allocated$16$lcssa = $allocated$16;$string_buf$25$lcssa = $string_buf$25;
        label = 27;
        break;
       } else {
        $count$169 = $count$081;$n$066 = 0;$num_tables$171 = $num_tables$082;$status$165 = $status$080;
       }
       while(1) {
        L27: do {
         switch ($status$165|0) {
         case 0:  {
          $63 = ($count$169|0)<(12);
          if ($63) {
           $64 = (($string_buf$25) + ($n$066)|0);
           $65 = HEAP8[$64>>0]|0;
           $66 = (($count$169) + 1)|0;
           $67 = HEAP32[$18>>2]|0;
           $68 = (($67) + ($count$169)|0);
           HEAP8[$68>>0] = $65;
           $count$2 = $66;$num_tables$4 = $num_tables$171;$status$4 = 0;
           break L27;
          }
          $69 = HEAP32[$18>>2]|0;
          $70 = ((($69)) + 4|0);
          $71 = HEAP8[$70>>0]|0;
          $72 = $71&255;
          $73 = $72 << 4;
          $74 = ((($69)) + 5|0);
          $75 = HEAP8[$74>>0]|0;
          $76 = $75&255;
          $77 = (($73) + ($76))|0;
          $78 = $77 << 4;
          $79 = $78 | 12;
          HEAP32[$19>>2] = $79;
          $80 = HEAP32[$loader>>2]|0;
          $81 = (($17) - ($80))|0;
          $82 = ($81|0)<($79|0);
          if ($82) {
           $allocated$16$lcssa338 = $allocated$16;$string_buf$25$lcssa342 = $string_buf$25;
           label = 32;
           break L6;
          }
          $83 = (_ft_mem_realloc($1,1,12,$79,$69,$error)|0);
          HEAP32[$18>>2] = $83;
          $84 = HEAP32[$error>>2]|0;
          $85 = ($84|0)==(0);
          if ($85) {
           $num_tables$2 = $77;
           label = 34;
          } else {
           $140 = $84;$allocated$2 = $allocated$16;$string_buf$3 = $string_buf$25;
           label = 46;
           break L1;
          }
          break;
         }
         case 1:  {
          $num_tables$2 = $num_tables$171;
          label = 34;
          break;
         }
         case 2:  {
          $num_tables$3 = $num_tables$171;
          label = 40;
          break;
         }
         default: {
          $count$2 = $count$169;$num_tables$4 = $num_tables$171;$status$4 = $status$165;
         }
         }
        } while(0);
        do {
         if ((label|0) == 34) {
          label = 0;
          $86 = HEAP32[$19>>2]|0;
          $87 = ($count$169|0)<($86|0);
          if ($87) {
           $90 = (($string_buf$25) + ($n$066)|0);
           $91 = HEAP8[$90>>0]|0;
           $92 = (($count$169) + 1)|0;
           $93 = HEAP32[$18>>2]|0;
           $94 = (($93) + ($count$169)|0);
           HEAP8[$94>>0] = $91;
           $count$2 = $92;$num_tables$4 = $num_tables$2;$status$4 = 1;
           break;
          }
          $88 = ($num_tables$2|0)>(0);
          $$pre = HEAP32[$18>>2]|0;
          if ($88) {
           $89 = ((($$pre)) + 12|0);
           $118 = $86;$i$064 = 0;
           while(1) {
            $95 = $i$064 << 4;
            $96 = (($89) + ($95)|0);
            $97 = ((($96)) + 12|0);
            $98 = HEAP8[$97>>0]|0;
            $99 = $98&255;
            $100 = $99 << 24;
            $101 = ((($97)) + 1|0);
            $102 = HEAP8[$101>>0]|0;
            $103 = $102&255;
            $104 = $103 << 16;
            $105 = $104 | $100;
            $106 = ((($97)) + 2|0);
            $107 = HEAP8[$106>>0]|0;
            $108 = $107&255;
            $109 = $108 << 8;
            $110 = $105 | $109;
            $111 = ((($97)) + 3|0);
            $112 = HEAP8[$111>>0]|0;
            $113 = $112&255;
            $114 = $110 | $113;
            $115 = (($114) + 3)|0;
            $116 = $115 & -4;
            $117 = (($116) + ($118))|0;
            HEAP32[$19>>2] = $117;
            $119 = (($i$064) + 1)|0;
            $exitcond = ($119|0)==($num_tables$2|0);
            if ($exitcond) {
             $123 = $117;
             break;
            } else {
             $118 = $117;$i$064 = $119;
            }
           }
          } else {
           $123 = $86;
          }
          $120 = $num_tables$2 << 4;
          $121 = $120 | 12;
          $122 = (($123) + 1)|0;
          $124 = (_ft_mem_realloc($1,1,$121,$122,$$pre,$error)|0);
          HEAP32[$18>>2] = $124;
          $125 = HEAP32[$error>>2]|0;
          $126 = ($125|0)==(0);
          if ($126) {
           $num_tables$3 = $num_tables$2;
           label = 40;
          } else {
           $140 = $125;$allocated$2 = $allocated$16;$string_buf$3 = $string_buf$25;
           label = 46;
           break L1;
          }
         }
        } while(0);
        if ((label|0) == 40) {
         label = 0;
         $127 = HEAP32[$19>>2]|0;
         $128 = ($count$169|0)<($127|0);
         if (!($128)) {
          $allocated$16$lcssa340 = $allocated$16;$string_buf$25$lcssa344 = $string_buf$25;
          label = 41;
          break L6;
         }
         $129 = (($string_buf$25) + ($n$066)|0);
         $130 = HEAP8[$129>>0]|0;
         $131 = (($count$169) + 1)|0;
         $132 = HEAP32[$18>>2]|0;
         $133 = (($132) + ($count$169)|0);
         HEAP8[$133>>0] = $130;
         $count$2 = $131;$num_tables$4 = $num_tables$3;$status$4 = 2;
        }
        $134 = (($n$066) + 1)|0;
        $135 = ($134>>>0)<($string_size$3>>>0);
        if ($135) {
         $count$169 = $count$2;$n$066 = $134;$num_tables$171 = $num_tables$4;$status$165 = $status$4;
        } else {
         $count$2$lcssa = $count$2;$num_tables$4$lcssa = $num_tables$4;$status$4$lcssa = $status$4;
         break;
        }
       }
       $136 = HEAP32[$4>>2]|0;
       FUNCTION_TABLE_vi[$136 & 8191]($loader);
       $137 = HEAP32[$loader>>2]|0;
       $138 = ($137>>>0)<($3>>>0);
       if ($138) {
        $22 = $137;$allocated$079 = $allocated$16;$count$081 = $count$2$lcssa;$num_tables$082 = $num_tables$4$lcssa;$old_string_size$077 = $old_string_size$14;$status$080 = $status$4$lcssa;$string_buf$078 = $string_buf$25;$string_size$076 = $string_size$3;
       } else {
        $allocated$0$lcssa = $allocated$16;$string_buf$0$lcssa = $string_buf$25;
        break L4;
       }
      }
      switch (label|0) {
       case 7: {
        $23 = ((($$lcssa295)) + 1|0);
        HEAP32[$loader>>2] = $23;
        $allocated$3 = $allocated$079$lcssa;$string_buf$4 = $string_buf$078$lcssa;
        break L1;
        break;
       }
       case 10: {
        HEAP32[$error>>2] = 3;
        $140 = 3;$allocated$2 = $allocated$079$lcssa310;$string_buf$3 = $string_buf$078$lcssa323;
        label = 46;
        break L1;
        break;
       }
       case 15: {
        HEAP32[$error>>2] = 3;
        $140 = 3;$allocated$2 = $allocated$079$lcssa313;$string_buf$3 = $string_buf$078$lcssa326;
        label = 46;
        break L1;
        break;
       }
       case 17: {
        HEAP32[$error>>2] = 3;
        $140 = 3;$allocated$2 = 0;$string_buf$3 = $string_buf$078$lcssa327;
        label = 46;
        break L1;
        break;
       }
       case 20: {
        HEAP32[$error>>2] = 3;
        $140 = 3;$allocated$2 = 0;$string_buf$3 = $$lcssa337;
        label = 46;
        break L1;
        break;
       }
       case 23: {
        HEAP32[$error>>2] = 3;
        $140 = 3;$allocated$2 = $allocated$1$lcssa;$string_buf$3 = 0;
        label = 46;
        break L1;
        break;
       }
       case 27: {
        HEAP32[$error>>2] = 3;
        $140 = 3;$allocated$2 = $allocated$16$lcssa;$string_buf$3 = $string_buf$25$lcssa;
        label = 46;
        break L1;
        break;
       }
       case 32: {
        HEAP32[$error>>2] = 3;
        $140 = 3;$allocated$2 = $allocated$16$lcssa338;$string_buf$3 = $string_buf$25$lcssa342;
        label = 46;
        break L1;
        break;
       }
       case 41: {
        HEAP32[$error>>2] = 3;
        $140 = 3;$allocated$2 = $allocated$16$lcssa340;$string_buf$3 = $string_buf$25$lcssa344;
        label = 46;
        break L1;
        break;
       }
       case 49: {
        STACKTOP = sp;return;
        break;
       }
      }
     } else {
      $allocated$0$lcssa = 0;$string_buf$0$lcssa = 0;
     }
    } while(0);
    HEAP32[$error>>2] = 3;
    $140 = 3;$allocated$2 = $allocated$0$lcssa;$string_buf$3 = $string_buf$0$lcssa;
    label = 46;
   } else {
    label = 3;
   }
  } else {
   label = 3;
  }
 } while(0);
 if ((label|0) == 3) {
  HEAP32[$error>>2] = 3;
  $140 = 3;$allocated$2 = 0;$string_buf$3 = 0;
  label = 46;
 }
 if ((label|0) == 46) {
  $139 = ((($loader)) + 12|0);
  HEAP32[$139>>2] = $140;
  $allocated$3 = $allocated$2;$string_buf$4 = $string_buf$3;
 }
 $141 = ($allocated$3<<24>>24)==(0);
 if ($141) {
  STACKTOP = sp;return;
 }
 _ft_mem_free($1,$string_buf$4);
 STACKTOP = sp;return;
}
function _T42_Face_Done($t42face) {
 $t42face = $t42face|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($t42face|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($t42face)) + 132|0);
 $2 = ((($t42face)) + 100|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($t42face)) + 500|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if (!($6)) {
  (_FT_Done_Face($5)|0);
 }
 $7 = HEAP32[$1>>2]|0;
 _ft_mem_free($3,$7);
 HEAP32[$1>>2] = 0;
 $8 = ((($t42face)) + 136|0);
 $9 = HEAP32[$8>>2]|0;
 _ft_mem_free($3,$9);
 HEAP32[$8>>2] = 0;
 $10 = ((($t42face)) + 140|0);
 $11 = HEAP32[$10>>2]|0;
 _ft_mem_free($3,$11);
 HEAP32[$10>>2] = 0;
 $12 = ((($t42face)) + 144|0);
 $13 = HEAP32[$12>>2]|0;
 _ft_mem_free($3,$13);
 HEAP32[$12>>2] = 0;
 $14 = ((($t42face)) + 148|0);
 $15 = HEAP32[$14>>2]|0;
 _ft_mem_free($3,$15);
 HEAP32[$14>>2] = 0;
 $16 = ((($t42face)) + 428|0);
 $17 = HEAP32[$16>>2]|0;
 _ft_mem_free($3,$17);
 HEAP32[$16>>2] = 0;
 $18 = ((($t42face)) + 424|0);
 $19 = HEAP32[$18>>2]|0;
 _ft_mem_free($3,$19);
 HEAP32[$18>>2] = 0;
 $20 = ((($t42face)) + 420|0);
 $21 = HEAP32[$20>>2]|0;
 _ft_mem_free($3,$21);
 HEAP32[$20>>2] = 0;
 $22 = ((($t42face)) + 396|0);
 $23 = HEAP32[$22>>2]|0;
 _ft_mem_free($3,$23);
 HEAP32[$22>>2] = 0;
 $24 = ((($t42face)) + 400|0);
 $25 = HEAP32[$24>>2]|0;
 _ft_mem_free($3,$25);
 HEAP32[$24>>2] = 0;
 $26 = ((($t42face)) + 372|0);
 $27 = ((($26)) + 12|0);
 $28 = HEAP32[$27>>2]|0;
 _ft_mem_free($3,$28);
 HEAP32[$27>>2] = 0;
 $29 = ((($26)) + 16|0);
 $30 = HEAP32[$29>>2]|0;
 _ft_mem_free($3,$30);
 HEAP32[$29>>2] = 0;
 $31 = ((($t42face)) + 364|0);
 $32 = HEAP32[$31>>2]|0;
 _ft_mem_free($3,$32);
 HEAP32[$31>>2] = 0;
 $33 = ((($t42face)) + 492|0);
 $34 = HEAP32[$33>>2]|0;
 _ft_mem_free($3,$34);
 HEAP32[$33>>2] = 0;
 $35 = ((($t42face)) + 536|0);
 $36 = ((($35)) + 20|0);
 $37 = HEAP32[$36>>2]|0;
 _ft_mem_free($3,$37);
 HEAP32[$36>>2] = 0;
 $38 = ((($35)) + 16|0);
 HEAP32[$38>>2] = 0;
 $39 = ((($t42face)) + 20|0);
 HEAP32[$39>>2] = 0;
 $40 = ((($t42face)) + 24|0);
 HEAP32[$40>>2] = 0;
 return;
}
function _T42_Size_Init($size) {
 $size = $size|0;
 var $$cast = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $ttsize = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ttsize = sp;
 $0 = HEAP32[$size>>2]|0;
 $1 = ((($0)) + 500|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (_FT_New_Size($2,$ttsize)|0);
 $4 = HEAP32[$ttsize>>2]|0;
 $5 = ((($size)) + 44|0);
 HEAP32[$5>>2] = $4;
 $$cast = $4;
 (_FT_Activate_Size($$cast)|0);
 STACKTOP = sp;return ($3|0);
}
function _T42_Size_Done($t42size) {
 $t42size = $t42size|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$t42size>>2]|0;
 $1 = ((($0)) + 500|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 108|0);
 $4 = ((($t42size)) + 44|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (_FT_List_Find($3,$5)|0);
 $7 = ($6|0)==(0|0);
 if ($7) {
  return;
 }
 $8 = HEAP32[$4>>2]|0;
 (_FT_Done_Size($8)|0);
 HEAP32[$4>>2] = 0;
 return;
}
function _T42_GlyphSlot_Init($t42slot) {
 $t42slot = $t42slot|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$0 = 0, $ttslot = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ttslot = sp;
 $0 = ((($t42slot)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 84|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 $5 = ((($1)) + 500|0);
 $6 = HEAP32[$5>>2]|0;
 if ($4) {
  $7 = ((($6)) + 84|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ((($t42slot)) + 160|0);
  HEAP32[$9>>2] = $8;
  $error$0 = 0;
  STACKTOP = sp;return ($error$0|0);
 } else {
  $10 = (_FT_New_GlyphSlot($6,$ttslot)|0);
  $11 = HEAP32[$ttslot>>2]|0;
  $12 = ((($t42slot)) + 160|0);
  HEAP32[$12>>2] = $11;
  $error$0 = $10;
  STACKTOP = sp;return ($error$0|0);
 }
 return (0)|0;
}
function _T42_GlyphSlot_Done($t42slot) {
 $t42slot = $t42slot|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($t42slot)) + 160|0);
 $1 = HEAP32[$0>>2]|0;
 _FT_Done_GlyphSlot($1);
 return;
}
function _T42_GlyphSlot_Load($glyph,$size,$glyph_index,$load_flags) {
 $glyph = $glyph|0;
 $size = $size|0;
 $glyph_index = $glyph_index|0;
 $load_flags = $load_flags|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($glyph)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 96|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 28|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($glyph)) + 160|0);
 $7 = HEAP32[$6>>2]|0;
 _ft_glyphslot_free_bitmap($7);
 $8 = ((($7)) + 24|0);
 $9 = ((($7)) + 152|0);
 HEAP32[$9>>2] = 0;
 $10 = ((($7)) + 72|0);
 dest=$8; stop=dest+40|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 dest=$10; stop=dest+72|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $11 = ((($5)) + 72|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = HEAP32[$6>>2]|0;
 $14 = ((($size)) + 44|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = $load_flags | 8;
 $17 = (FUNCTION_TABLE_iiiii[$12 & 8191]($13,$15,$glyph_index,$16)|0);
 $18 = ($17|0)==(0);
 if (!($18)) {
  return ($17|0);
 }
 $19 = ((($glyph)) + 24|0);
 $20 = HEAP32[$6>>2]|0;
 $21 = ((($20)) + 24|0);
 ;HEAP32[$19>>2]=HEAP32[$21>>2]|0;HEAP32[$19+4>>2]=HEAP32[$21+4>>2]|0;HEAP32[$19+8>>2]=HEAP32[$21+8>>2]|0;HEAP32[$19+12>>2]=HEAP32[$21+12>>2]|0;HEAP32[$19+16>>2]=HEAP32[$21+16>>2]|0;HEAP32[$19+20>>2]=HEAP32[$21+20>>2]|0;HEAP32[$19+24>>2]=HEAP32[$21+24>>2]|0;HEAP32[$19+28>>2]=HEAP32[$21+28>>2]|0;
 $22 = ((($20)) + 56|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = ((($glyph)) + 56|0);
 HEAP32[$24>>2] = $23;
 $25 = ((($20)) + 60|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ((($glyph)) + 60|0);
 HEAP32[$27>>2] = $26;
 $28 = ((($20)) + 72|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = ((($glyph)) + 72|0);
 HEAP32[$30>>2] = $29;
 $31 = ((($glyph)) + 108|0);
 $32 = ((($20)) + 108|0);
 ;HEAP32[$31>>2]=HEAP32[$32>>2]|0;HEAP32[$31+4>>2]=HEAP32[$32+4>>2]|0;HEAP32[$31+8>>2]=HEAP32[$32+8>>2]|0;HEAP32[$31+12>>2]=HEAP32[$32+12>>2]|0;HEAP32[$31+16>>2]=HEAP32[$32+16>>2]|0;
 $33 = ((($glyph)) + 76|0);
 $34 = ((($20)) + 76|0);
 ;HEAP32[$33>>2]=HEAP32[$34>>2]|0;HEAP32[$33+4>>2]=HEAP32[$34+4>>2]|0;HEAP32[$33+8>>2]=HEAP32[$34+8>>2]|0;HEAP32[$33+12>>2]=HEAP32[$34+12>>2]|0;HEAP32[$33+16>>2]=HEAP32[$34+16>>2]|0;HEAP32[$33+20>>2]=HEAP32[$34+20>>2]|0;
 $35 = ((($20)) + 100|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = ((($glyph)) + 100|0);
 HEAP32[$37>>2] = $36;
 $38 = ((($20)) + 104|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ((($glyph)) + 104|0);
 HEAP32[$40>>2] = $39;
 $41 = ((($20)) + 128|0);
 $42 = HEAP32[$41>>2]|0;
 $43 = ((($glyph)) + 128|0);
 HEAP32[$43>>2] = $42;
 $44 = ((($20)) + 132|0);
 $45 = HEAP32[$44>>2]|0;
 $46 = ((($glyph)) + 132|0);
 HEAP32[$46>>2] = $45;
 $47 = ((($20)) + 136|0);
 $48 = HEAP32[$47>>2]|0;
 $49 = ((($glyph)) + 136|0);
 HEAP32[$49>>2] = $48;
 $50 = ((($20)) + 140|0);
 $51 = HEAP32[$50>>2]|0;
 $52 = ((($glyph)) + 140|0);
 HEAP32[$52>>2] = $51;
 return ($17|0);
}
function _T42_Size_Request($t42size,$req) {
 $t42size = $t42size|0;
 $req = $req|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$t42size>>2]|0;
 $1 = ((($t42size)) + 44|0);
 $2 = HEAP32[$1>>2]|0;
 (_FT_Activate_Size($2)|0);
 $3 = ((($0)) + 500|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (_FT_Request_Size($4,$req)|0);
 $6 = ($5|0)==(0);
 if (!($6)) {
  return ($5|0);
 }
 $7 = ((($t42size)) + 12|0);
 $8 = HEAP32[$3>>2]|0;
 $9 = ((($8)) + 88|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + 12|0);
 ;HEAP32[$7>>2]=HEAP32[$11>>2]|0;HEAP32[$7+4>>2]=HEAP32[$11+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$11+8>>2]|0;HEAP32[$7+12>>2]=HEAP32[$11+12>>2]|0;HEAP32[$7+16>>2]=HEAP32[$11+16>>2]|0;HEAP32[$7+20>>2]=HEAP32[$11+20>>2]|0;HEAP32[$7+24>>2]=HEAP32[$11+24>>2]|0;
 return ($5|0);
}
function _T42_Size_Select($t42size,$strike_index) {
 $t42size = $t42size|0;
 $strike_index = $strike_index|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$t42size>>2]|0;
 $1 = ((($t42size)) + 44|0);
 $2 = HEAP32[$1>>2]|0;
 (_FT_Activate_Size($2)|0);
 $3 = ((($0)) + 500|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (_FT_Select_Size($4,$strike_index)|0);
 $6 = ($5|0)==(0);
 if (!($6)) {
  return ($5|0);
 }
 $7 = ((($t42size)) + 12|0);
 $8 = HEAP32[$3>>2]|0;
 $9 = ((($8)) + 88|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + 12|0);
 ;HEAP32[$7>>2]=HEAP32[$11>>2]|0;HEAP32[$7+4>>2]=HEAP32[$11+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$11+8>>2]|0;HEAP32[$7+12>>2]=HEAP32[$11+12>>2]|0;HEAP32[$7+16>>2]=HEAP32[$11+16>>2]|0;HEAP32[$7+20>>2]=HEAP32[$11+20>>2]|0;HEAP32[$7+24>>2]=HEAP32[$11+24>>2]|0;
 return ($5|0);
}
function _winfnt_get_service($module,$service_id) {
 $module = $module|0;
 $service_id = $service_id|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_ft_service_list_lookup(59136,$service_id)|0);
 return ($0|0);
}
function _winfnt_get_header($face,$aheader) {
 $face = $face|0;
 $aheader = $aheader|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 132|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 4|0);
 _memcpy(($aheader|0),($2|0),156)|0;
 return 0;
}
function _FNT_Face_Init($stream,$fntface,$face_index,$num_params,$params) {
 $stream = $stream|0;
 $fntface = $fntface|0;
 $face_index = $face_index|0;
 $num_params = $num_params|0;
 $params = $params|0;
 var $$0$i = 0, $$2$i = 0, $$lcssa = 0, $$pr$i = 0, $$pr48$i = 0, $$pr49$i = 0, $$pre = 0, $$pre$i = 0, $$pre39 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0;
 var $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0;
 var $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0;
 var $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0;
 var $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0;
 var $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0;
 var $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0;
 var $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0;
 var $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0;
 var $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0;
 var $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0;
 var $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0;
 var $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $charmap = 0, $cond$i = 0, $data_entry$i = 0, $dir_entry1$i = 0, $dir_entry2$i = 0, $dir_entry3$i = 0, $error = 0, $error$i = 0, $font_count$1$ph$i = 0, $font_offset$1$ph$i = 0, $i$037$i = 0;
 var $i$126$i = 0, $j$023$i = 0, $k$021$i = 0, $lang_dir$i = 0, $mz_header$i = 0, $name_dir$i = 0, $ne_header$i = 0, $or$cond = 0, $or$cond$i = 0, $pe32_header$i = 0, $pe32_section$i = 0, $root_dir$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 160|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp + 144|0;
 $mz_header$i = sp + 156|0;
 $ne_header$i = sp + 148|0;
 $pe32_header$i = sp + 124|0;
 $pe32_section$i = sp + 104|0;
 $root_dir$i = sp + 88|0;
 $name_dir$i = sp + 72|0;
 $lang_dir$i = sp + 56|0;
 $dir_entry1$i = sp + 48|0;
 $dir_entry2$i = sp + 40|0;
 $dir_entry3$i = sp + 32|0;
 $data_entry$i = sp + 16|0;
 $error = sp + 12|0;
 $charmap = sp;
 $0 = ((($fntface)) + 100|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($fntface)) + 104|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($fntface)) + 132|0);
 HEAP32[$4>>2] = 0;
 $5 = (_FT_Stream_Seek($3,0)|0);
 HEAP32[$error$i>>2] = $5;
 $6 = ($5|0)==(0);
 L1: do {
  if ($6) {
   $7 = (_FT_Stream_ReadFields($3,174910,$mz_header$i)|0);
   HEAP32[$error$i>>2] = $7;
   $8 = ($7|0)==(0);
   if ($8) {
    HEAP32[$error$i>>2] = 2;
    $9 = HEAP16[$mz_header$i>>1]|0;
    $10 = ($9<<16>>16)==(23117);
    L4: do {
     if ($10) {
      $11 = ((($mz_header$i)) + 2|0);
      $12 = HEAP16[$11>>1]|0;
      $13 = $12&65535;
      $14 = (_FT_Stream_Seek($3,$13)|0);
      HEAP32[$error$i>>2] = $14;
      $15 = ($14|0)==(0);
      L6: do {
       if ($15) {
        $16 = (_FT_Stream_ReadFields($3,174930,$ne_header$i)|0);
        HEAP32[$error$i>>2] = $16;
        $17 = ($16|0)==(0);
        if ($17) {
         HEAP32[$error$i>>2] = 2;
         $18 = HEAP16[$ne_header$i>>1]|0;
         L9: do {
          switch ($18<<16>>16) {
          case 17742:  {
           $19 = HEAP16[$11>>1]|0;
           $20 = $19&65535;
           $21 = ((($ne_header$i)) + 2|0);
           $22 = HEAP16[$21>>1]|0;
           $23 = $22&65535;
           $24 = (($23) + ($20))|0;
           $25 = (_FT_Stream_Seek($3,$24)|0);
           HEAP32[$error$i>>2] = $25;
           $26 = ($25|0)==(0);
           do {
            if ($26) {
             $27 = ((($ne_header$i)) + 4|0);
             $28 = HEAP16[$27>>1]|0;
             $29 = $28&65535;
             $30 = HEAP16[$21>>1]|0;
             $31 = $30&65535;
             $32 = (($29) - ($31))|0;
             $33 = (_FT_Stream_EnterFrame($3,$32)|0);
             HEAP32[$error$i>>2] = $33;
             $34 = ($33|0)==(0);
             if ($34) {
              $35 = (_FT_Stream_GetUShortLE($3)|0);
              $36 = (_FT_Stream_GetUShortLE($3)|0);
              $37 = ($36<<16>>16)==(0);
              L14: do {
               if ($37) {
                $font_count$1$ph$i = 0;$font_offset$1$ph$i = 0;
               } else {
                $38 = ((($3)) + 32|0);
                $41 = $36;
                while(1) {
                 $39 = (_FT_Stream_GetUShortLE($3)|0);
                 $40 = ($41<<16>>16)==(-32760);
                 if ($40) {
                  $$lcssa = $39;
                  break;
                 }
                 $49 = $39&65535;
                 $50 = ($49*12)|0;
                 $51 = (($50) + 4)|0;
                 $52 = HEAP32[$38>>2]|0;
                 $53 = (($52) + ($51)|0);
                 HEAP32[$38>>2] = $53;
                 $54 = (_FT_Stream_GetUShortLE($3)|0);
                 $55 = ($54<<16>>16)==(0);
                 if ($55) {
                  $font_count$1$ph$i = 0;$font_offset$1$ph$i = 0;
                  break L14;
                 } else {
                  $41 = $54;
                 }
                }
                $42 = (_FT_Stream_Pos($3)|0);
                $43 = HEAP32[$38>>2]|0;
                $44 = ((($3)) + 36|0);
                $45 = HEAP32[$44>>2]|0;
                $46 = (($42) + 4)|0;
                $47 = (($46) + ($43))|0;
                $48 = (($47) - ($45))|0;
                $font_count$1$ph$i = $$lcssa;$font_offset$1$ph$i = $48;
               }
              } while(0);
              _FT_Stream_ExitFrame($3);
              $56 = ($font_count$1$ph$i<<16>>16)!=(0);
              $57 = ($font_offset$1$ph$i|0)!=(0);
              $or$cond$i = $57 & $56;
              if (!($or$cond$i)) {
               HEAP32[$error$i>>2] = 3;
               break;
              }
              $58 = $font_count$1$ph$i&65535;
              $59 = ($58*118)|0;
              $60 = ((($3)) + 4|0);
              $61 = HEAP32[$60>>2]|0;
              $62 = ($59>>>0)>($61>>>0);
              if ($62) {
               HEAP32[$error$i>>2] = 3;
               break;
              }
              HEAP32[$fntface>>2] = $58;
              $63 = ($58|0)>($face_index|0);
              if (!($63)) {
               HEAP32[$error$i>>2] = 6;
               break;
              }
              $64 = ($face_index|0)<(0);
              if (!($64)) {
               $65 = (_ft_mem_alloc($1,172,$error$i)|0);
               HEAP32[$4>>2] = $65;
               $66 = HEAP32[$error$i>>2]|0;
               $67 = ($66|0)==(0);
               if ($67) {
                $68 = ($face_index*12)|0;
                $69 = (($font_offset$1$ph$i) + ($68))|0;
                $70 = (_FT_Stream_Seek($3,$69)|0);
                HEAP32[$error$i>>2] = $70;
                $71 = ($70|0)==(0);
                do {
                 if ($71) {
                  $72 = (_FT_Stream_EnterFrame($3,12)|0);
                  HEAP32[$error$i>>2] = $72;
                  $73 = ($72|0)==(0);
                  if (!($73)) {
                   break;
                  }
                  $74 = (_FT_Stream_GetUShortLE($3)|0);
                  $75 = $74&65535;
                  $76 = $35&65535;
                  $77 = $75 << $76;
                  $78 = HEAP32[$4>>2]|0;
                  HEAP32[$78>>2] = $77;
                  $79 = (_FT_Stream_GetUShortLE($3)|0);
                  $80 = $79&65535;
                  $81 = $80 << $76;
                  $82 = HEAP32[$4>>2]|0;
                  $83 = ((($82)) + 164|0);
                  HEAP32[$83>>2] = $81;
                  $84 = ((($3)) + 32|0);
                  $85 = HEAP32[$84>>2]|0;
                  $86 = ((($85)) + 8|0);
                  HEAP32[$84>>2] = $86;
                  _FT_Stream_ExitFrame($3);
                  $87 = HEAP32[$4>>2]|0;
                  $88 = (_fnt_font_load($87,$3)|0);
                  HEAP32[$error$i>>2] = $88;
                  $$pr48$i = $88;
                  label = 73;
                  break L9;
                 }
                } while(0);
                break L4;
               }
              }
             }
            }
           } while(0);
           label = 87;
           break L1;
           break;
          }
          case 17744:  {
           $89 = HEAP16[$11>>1]|0;
           $90 = $89&65535;
           $91 = (_FT_Stream_Seek($3,$90)|0);
           HEAP32[$error$i>>2] = $91;
           $92 = ($91|0)==(0);
           L38: do {
            if ($92) {
             $93 = (_FT_Stream_ReadFields($3,175106,$pe32_header$i)|0);
             HEAP32[$error$i>>2] = $93;
             $94 = ($93|0)==(0);
             if ($94) {
              $95 = HEAP32[$pe32_header$i>>2]|0;
              $96 = ($95|0)==(17744);
              if ($96) {
               $97 = ((($pe32_header$i)) + 4|0);
               $98 = HEAP32[$97>>2]|0;
               $99 = $98&65535;
               $100 = ($99<<16>>16)==(332);
               if ($100) {
                $101 = ((($pe32_header$i)) + 8|0);
                $102 = HEAP32[$101>>2]|0;
                $103 = ($102|0)==(17498336);
                if ($103) {
                 HEAP32[$fntface>>2] = 0;
                 $104 = ((($pe32_header$i)) + 6|0);
                 $105 = ($98>>>0)<(65536);
                 L45: do {
                  if (!($105)) {
                   $106 = ((($pe32_header$i)) + 12|0);
                   $107 = ((($pe32_section$i)) + 8|0);
                   $i$037$i = 0;
                   while(1) {
                    $111 = (_FT_Stream_ReadFields($3,175158,$pe32_section$i)|0);
                    HEAP32[$error$i>>2] = $111;
                    $112 = ($111|0)==(0);
                    if (!($112)) {
                     $$2$i = 2;$$pr49$i = $111;
                     break L38;
                    }
                    $113 = HEAP32[$106>>2]|0;
                    $114 = HEAP32[$107>>2]|0;
                    $115 = ($113|0)==($114|0);
                    $110 = (($i$037$i) + 1)<<16>>16;
                    if ($115) {
                     break;
                    }
                    $108 = HEAP16[$104>>1]|0;
                    $109 = ($110&65535)<($108&65535);
                    if ($109) {
                     $i$037$i = $110;
                    } else {
                     break L45;
                    }
                   }
                   $116 = ((($pe32_section$i)) + 16|0);
                   $117 = HEAP32[$116>>2]|0;
                   $118 = (_FT_Stream_Seek($3,$117)|0);
                   HEAP32[$error$i>>2] = $118;
                   $119 = ($118|0)==(0);
                   if (!($119)) {
                    $$2$i = 2;$$pr49$i = $118;
                    break L38;
                   }
                   $120 = (_FT_Stream_ReadFields($3,175190,$root_dir$i)|0);
                   HEAP32[$error$i>>2] = $120;
                   $121 = ($120|0)==(0);
                   if (!($121)) {
                    $$2$i = 2;$$pr49$i = $120;
                    break L38;
                   }
                   $122 = HEAP32[$116>>2]|0;
                   $123 = ((($root_dir$i)) + 12|0);
                   $124 = HEAP32[$123>>2]|0;
                   $125 = $124 & 65535;
                   $126 = $124 >>> 16;
                   $127 = (0 - ($125))|0;
                   $128 = ($126|0)==($127|0);
                   if ($128) {
                    $$2$i = 0;$$pr49$i = 0;
                    break L38;
                   }
                   $129 = (($122) + 16)|0;
                   $130 = ((($dir_entry1$i)) + 4|0);
                   $131 = ((($name_dir$i)) + 12|0);
                   $132 = ((($dir_entry2$i)) + 4|0);
                   $133 = ((($lang_dir$i)) + 12|0);
                   $134 = ((($dir_entry3$i)) + 4|0);
                   $135 = ((($data_entry$i)) + 4|0);
                   $137 = 0;$i$126$i = 0;
                   L55: while(1) {
                    $136 = $137 << 3;
                    $138 = (($129) + ($136))|0;
                    $139 = (_FT_Stream_Seek($3,$138)|0);
                    HEAP32[$error$i>>2] = $139;
                    $140 = ($139|0)==(0);
                    if (!($140)) {
                     $$2$i = 2;$$pr49$i = $139;
                     break L38;
                    }
                    $141 = (_FT_Stream_ReadFields($3,175222,$dir_entry1$i)|0);
                    HEAP32[$error$i>>2] = $141;
                    $142 = ($141|0)==(0);
                    if (!($142)) {
                     $$2$i = 2;$$pr49$i = $141;
                     break L38;
                    }
                    $143 = HEAP32[$130>>2]|0;
                    $144 = ($143|0)<(0);
                    if (!($144)) {
                     label = 44;
                     break;
                    }
                    $145 = $143 & 2147483647;
                    HEAP32[$130>>2] = $145;
                    $146 = HEAP32[$116>>2]|0;
                    $147 = (($146) + ($145))|0;
                    $148 = (_FT_Stream_Seek($3,$147)|0);
                    HEAP32[$error$i>>2] = $148;
                    $149 = ($148|0)==(0);
                    if (!($149)) {
                     $$2$i = 2;$$pr49$i = $148;
                     break L38;
                    }
                    $150 = (_FT_Stream_ReadFields($3,175190,$name_dir$i)|0);
                    HEAP32[$error$i>>2] = $150;
                    $151 = ($150|0)==(0);
                    if (!($151)) {
                     $$2$i = 2;$$pr49$i = $150;
                     break L38;
                    }
                    $152 = HEAP32[$131>>2]|0;
                    $153 = $152 & 65535;
                    $154 = $152 >>> 16;
                    $155 = (0 - ($153))|0;
                    $156 = ($154|0)==($155|0);
                    if (!($156)) {
                     $157 = (($147) + 16)|0;
                     $159 = 0;$j$023$i = 0;
                     while(1) {
                      $158 = $159 << 3;
                      $160 = (($157) + ($158))|0;
                      $161 = (_FT_Stream_Seek($3,$160)|0);
                      HEAP32[$error$i>>2] = $161;
                      $162 = ($161|0)==(0);
                      if (!($162)) {
                       $$2$i = 2;$$pr49$i = $161;
                       break L38;
                      }
                      $163 = (_FT_Stream_ReadFields($3,175222,$dir_entry2$i)|0);
                      HEAP32[$error$i>>2] = $163;
                      $164 = ($163|0)==(0);
                      if (!($164)) {
                       $$2$i = 2;$$pr49$i = $163;
                       break L38;
                      }
                      $165 = HEAP32[$132>>2]|0;
                      $166 = ($165|0)<(0);
                      if (!($166)) {
                       label = 52;
                       break L55;
                      }
                      $167 = $165 & 2147483647;
                      HEAP32[$132>>2] = $167;
                      $168 = HEAP32[$116>>2]|0;
                      $169 = (($168) + ($167))|0;
                      $170 = (_FT_Stream_Seek($3,$169)|0);
                      HEAP32[$error$i>>2] = $170;
                      $171 = ($170|0)==(0);
                      if (!($171)) {
                       $$2$i = 2;$$pr49$i = $170;
                       break L38;
                      }
                      $172 = (_FT_Stream_ReadFields($3,175190,$lang_dir$i)|0);
                      HEAP32[$error$i>>2] = $172;
                      $173 = ($172|0)==(0);
                      if (!($173)) {
                       $$2$i = 2;$$pr49$i = $172;
                       break L38;
                      }
                      $174 = HEAP32[$133>>2]|0;
                      $175 = $174 & 65535;
                      $176 = $174 >>> 16;
                      $177 = (0 - ($175))|0;
                      $178 = ($176|0)==($177|0);
                      if (!($178)) {
                       $179 = (($169) + 16)|0;
                       $181 = 0;$k$021$i = 0;
                       while(1) {
                        $180 = $181 << 3;
                        $182 = (($179) + ($180))|0;
                        $183 = (_FT_Stream_Seek($3,$182)|0);
                        HEAP32[$error$i>>2] = $183;
                        $184 = ($183|0)==(0);
                        if (!($184)) {
                         $$2$i = 2;$$pr49$i = $183;
                         break L38;
                        }
                        $185 = (_FT_Stream_ReadFields($3,175222,$dir_entry3$i)|0);
                        HEAP32[$error$i>>2] = $185;
                        $186 = ($185|0)==(0);
                        if (!($186)) {
                         $$2$i = 2;$$pr49$i = $185;
                         break L38;
                        }
                        $187 = HEAP32[$132>>2]|0;
                        $188 = ($187|0)<(0);
                        if ($188) {
                         label = 60;
                         break L55;
                        }
                        $189 = HEAP32[$dir_entry1$i>>2]|0;
                        $190 = ($189|0)==(8);
                        if ($190) {
                         $191 = HEAP32[$134>>2]|0;
                         $192 = (($191) + ($122))|0;
                         $193 = (_FT_Stream_Seek($3,$192)|0);
                         HEAP32[$error$i>>2] = $193;
                         $194 = ($193|0)==(0);
                         if (!($194)) {
                          $$2$i = 2;$$pr49$i = $193;
                          break L38;
                         }
                         $195 = (_FT_Stream_ReadFields($3,175238,$data_entry$i)|0);
                         HEAP32[$error$i>>2] = $195;
                         $196 = ($195|0)==(0);
                         if (!($196)) {
                          $$2$i = 2;$$pr49$i = $195;
                          break L38;
                         }
                         $197 = HEAP32[$fntface>>2]|0;
                         $198 = ($197|0)==($face_index|0);
                         if ($198) {
                          $199 = (_ft_mem_alloc($1,172,$error$i)|0);
                          HEAP32[$4>>2] = $199;
                          $200 = HEAP32[$error$i>>2]|0;
                          $201 = ($200|0)==(0);
                          if (!($201)) {
                           $$2$i = 2;$$pr49$i = $200;
                           break L38;
                          }
                          $202 = HEAP32[$116>>2]|0;
                          $203 = HEAP32[$data_entry$i>>2]|0;
                          $204 = (($203) + ($202))|0;
                          $205 = HEAP32[$107>>2]|0;
                          $206 = (($204) - ($205))|0;
                          HEAP32[$199>>2] = $206;
                          $207 = HEAP32[$135>>2]|0;
                          $208 = ((($199)) + 164|0);
                          HEAP32[$208>>2] = $207;
                          $209 = (_fnt_font_load($199,$3)|0);
                          HEAP32[$error$i>>2] = $209;
                          $210 = ($209|0)==(0);
                          if (!($210)) {
                           $$2$i = 13;$$pr49$i = $209;
                           break L38;
                          }
                          $$pre$i = HEAP32[$fntface>>2]|0;
                          $212 = $$pre$i;
                         } else {
                          $212 = $197;
                         }
                         $211 = (($212) + 1)|0;
                         HEAP32[$fntface>>2] = $211;
                        }
                        $213 = (($k$021$i) + 1)<<16>>16;
                        $214 = $213&65535;
                        $215 = HEAP32[$133>>2]|0;
                        $216 = $215 & 65535;
                        $217 = $215 >>> 16;
                        $218 = (($217) + ($216))|0;
                        $219 = ($214>>>0)<($218>>>0);
                        if ($219) {
                         $181 = $214;$k$021$i = $213;
                        } else {
                         break;
                        }
                       }
                      }
                      $220 = (($j$023$i) + 1)<<16>>16;
                      $221 = $220&65535;
                      $222 = HEAP32[$131>>2]|0;
                      $223 = $222 & 65535;
                      $224 = $222 >>> 16;
                      $225 = (($224) + ($223))|0;
                      $226 = ($221>>>0)<($225>>>0);
                      if ($226) {
                       $159 = $221;$j$023$i = $220;
                      } else {
                       break;
                      }
                     }
                    }
                    $227 = (($i$126$i) + 1)<<16>>16;
                    $228 = $227&65535;
                    $229 = HEAP32[$123>>2]|0;
                    $230 = $229 & 65535;
                    $231 = $229 >>> 16;
                    $232 = (($231) + ($230))|0;
                    $233 = ($228>>>0)<($232>>>0);
                    if ($233) {
                     $137 = $228;$i$126$i = $227;
                    } else {
                     $$2$i = 0;$$pr49$i = 0;
                     break L38;
                    }
                   }
                   if ((label|0) == 44) {
                    HEAP32[$error$i>>2] = 3;
                    $$2$i = 2;$$pr49$i = 3;
                    break L38;
                   }
                   else if ((label|0) == 52) {
                    HEAP32[$error$i>>2] = 3;
                    $$2$i = 2;$$pr49$i = 3;
                    break L38;
                   }
                   else if ((label|0) == 60) {
                    HEAP32[$error$i>>2] = 3;
                    $$2$i = 2;$$pr49$i = 3;
                    break L38;
                   }
                  }
                 } while(0);
                 HEAP32[$error$i>>2] = 3;
                 $$2$i = 2;$$pr49$i = 3;
                 break;
                }
               }
              }
              HEAP32[$error$i>>2] = 3;
              $$2$i = 2;$$pr49$i = 3;
             } else {
              $$2$i = 2;$$pr49$i = $93;
             }
            } else {
             $$2$i = 2;$$pr49$i = $91;
            }
           } while(0);
           $cond$i = ($$2$i|0)==(0);
           if ($cond$i) {
            $$pr48$i = $$pr49$i;
            label = 73;
           } else {
            switch ($$2$i|0) {
            case 2:  {
             label = 87;
             break L1;
             break;
            }
            case 13:  {
             $$pr$i = $$pr49$i;
             break;
            }
            default: {
             $$0$i = 0;
             break L1;
            }
            }
           }
           break;
          }
          default: {
           $$pr48$i = 2;
           label = 73;
          }
          }
         } while(0);
         do {
          if ((label|0) == 73) {
           $234 = HEAP32[$fntface>>2]|0;
           $235 = ($234|0)==(0);
           if ($235) {
            HEAP32[$error$i>>2] = 3;
            break L6;
           }
           $236 = ($234|0)>($face_index|0);
           if ($236) {
            $$pr$i = $$pr48$i;
            break;
           } else {
            HEAP32[$error$i>>2] = 6;
            break L6;
           }
          }
         } while(0);
         $237 = ($$pr$i|0)==(0);
         if ($237) {
          label = 87;
          break L1;
         } else {
          break L4;
         }
        }
       }
      } while(0);
      label = 87;
      break L1;
     }
    } while(0);
    $238 = HEAP32[$0>>2]|0;
    $239 = HEAP32[$2>>2]|0;
    $240 = HEAP32[$4>>2]|0;
    $241 = ($240|0)==(0|0);
    if ($241) {
     label = 87;
    } else {
     $242 = ((($240)) + 160|0);
     $243 = HEAP32[$242>>2]|0;
     $244 = ($243|0)==(0|0);
     if (!($244)) {
      _FT_Stream_ReleaseFrame($239,$242);
     }
     $245 = ((($240)) + 168|0);
     $246 = HEAP32[$245>>2]|0;
     _ft_mem_free($238,$246);
     HEAP32[$245>>2] = 0;
     _ft_mem_free($238,$240);
     HEAP32[$4>>2] = 0;
     label = 87;
    }
   } else {
    label = 87;
   }
  } else {
   label = 87;
  }
 } while(0);
 if ((label|0) == 87) {
  $247 = HEAP32[$error$i>>2]|0;
  $$0$i = $247;
 }
 HEAP32[$error>>2] = $$0$i;
 $248 = ($$0$i|0)==(0);
 $249 = ($face_index|0)<(0);
 $or$cond = $249 & $248;
 if ($or$cond) {
  $399 = 0;
  STACKTOP = sp;return ($399|0);
 }
 $250 = $$0$i & 255;
 $251 = ($250|0)==(2);
 do {
  if ($251) {
   $252 = (_ft_mem_alloc($1,172,$error)|0);
   HEAP32[$4>>2] = $252;
   $253 = HEAP32[$error>>2]|0;
   $254 = ($253|0)==(0);
   if (!($254)) {
    $399 = $253;
    STACKTOP = sp;return ($399|0);
   }
   HEAP32[$fntface>>2] = 1;
   HEAP32[$252>>2] = 0;
   $255 = ((($stream)) + 4|0);
   $256 = HEAP32[$255>>2]|0;
   $257 = ((($252)) + 164|0);
   HEAP32[$257>>2] = $256;
   $258 = (_fnt_font_load($252,$stream)|0);
   HEAP32[$error>>2] = $258;
   $259 = ($258|0)==(0);
   if ($259) {
    $260 = ($face_index|0)>(0);
    if ($260) {
     HEAP32[$error>>2] = 6;
     break;
    }
    if ($249) {
     $399 = 0;
     STACKTOP = sp;return ($399|0);
    } else {
     label = 96;
    }
   }
  } else {
   $261 = ($$0$i|0)==(0);
   if ($261) {
    label = 96;
   } else {
    $386 = ($fntface|0)==(0|0);
    if ($386) {
     $399 = $$0$i;
     STACKTOP = sp;return ($399|0);
    }
   }
  }
 } while(0);
 do {
  if ((label|0) == 96) {
   $262 = HEAP32[$4>>2]|0;
   $263 = ((($fntface)) + 4|0);
   HEAP32[$263>>2] = $face_index;
   $264 = ((($fntface)) + 8|0);
   $265 = HEAP32[$264>>2]|0;
   $266 = $265 | 18;
   HEAP32[$264>>2] = $266;
   $267 = ((($262)) + 100|0);
   $268 = HEAP32[$267>>2]|0;
   $269 = $268&65535;
   $270 = $268 >>> 16;
   $271 = $270&65535;
   $272 = ($269<<16>>16)==($271<<16>>16);
   if ($272) {
    $273 = $265 | 22;
    HEAP32[$264>>2] = $273;
   }
   $274 = ((($262)) + 86|0);
   $275 = HEAP8[$274>>0]|0;
   $276 = ($275<<24>>24)==(0);
   if (!($276)) {
    $277 = ((($fntface)) + 12|0);
    $278 = HEAP32[$277>>2]|0;
    $279 = $278 | 1;
    HEAP32[$277>>2] = $279;
   }
   $280 = ((($262)) + 90|0);
   $281 = HEAP16[$280>>1]|0;
   $282 = ($281&65535)>(799);
   if ($282) {
    $283 = ((($fntface)) + 12|0);
    $284 = HEAP32[$283>>2]|0;
    $285 = $284 | 2;
    HEAP32[$283>>2] = $285;
   }
   $286 = (_ft_mem_realloc($1,16,0,1,0,$error)|0);
   $287 = ((($fntface)) + 32|0);
   HEAP32[$287>>2] = $286;
   $288 = HEAP32[$error>>2]|0;
   $289 = ($288|0)==(0);
   if ($289) {
    $290 = ((($fntface)) + 28|0);
    HEAP32[$290>>2] = 1;
    $291 = HEAP16[$267>>1]|0;
    $292 = ((($286)) + 2|0);
    HEAP16[$292>>1] = $291;
    $293 = ((($262)) + 96|0);
    $294 = HEAP16[$293>>1]|0;
    $295 = $294&65535;
    $296 = ((($262)) + 84|0);
    $297 = HEAP16[$296>>1]|0;
    $298 = $297&65535;
    $299 = (($298) + ($295))|0;
    $300 = $299&65535;
    HEAP16[$286>>1] = $300;
    $301 = ((($262)) + 74|0);
    $302 = HEAP16[$301>>1]|0;
    $303 = $302&65535;
    $304 = $303 << 6;
    $305 = ((($286)) + 4|0);
    HEAP32[$305>>2] = $304;
    $306 = ((($262)) + 78|0);
    $307 = HEAP16[$306>>1]|0;
    $308 = ($307<<16>>16)!=(0);
    $309 = ((($262)) + 76|0);
    $310 = HEAP16[$309>>1]|0;
    $311 = ($310<<16>>16)!=(0);
    $312 = $310&65535;
    $313 = $311 ? $312 : 72;
    $314 = (_FT_MulDiv($304,$313,72)|0);
    $315 = ((($286)) + 12|0);
    $316 = (($314) + 32)|0;
    $317 = $316 & -64;
    HEAP32[$315>>2] = $317;
    $318 = HEAP16[$293>>1]|0;
    $319 = $318&65535;
    $320 = $319 << 6;
    $321 = ($317|0)>($320|0);
    if ($321) {
     HEAP32[$315>>2] = $320;
     $322 = (_FT_MulDiv($320,72,$313)|0);
     HEAP32[$305>>2] = $322;
     $325 = $322;
    } else {
     $$pre = HEAP32[$305>>2]|0;
     $325 = $$pre;
    }
    $323 = $307&65535;
    $324 = $308 ? $323 : 72;
    $326 = (_FT_MulDiv($325,$324,72)|0);
    $327 = ((($286)) + 8|0);
    $328 = (($326) + 32)|0;
    $329 = $328 & -64;
    HEAP32[$327>>2] = $329;
    $330 = ((($charmap)) + 4|0);
    HEAP32[$330>>2] = 0;
    $331 = ((($charmap)) + 8|0);
    HEAP16[$331>>1] = 0;
    $332 = ((($charmap)) + 10|0);
    HEAP16[$332>>1] = 0;
    HEAP32[$charmap>>2] = $fntface;
    $333 = ((($262)) + 92|0);
    $334 = HEAP8[$333>>0]|0;
    $335 = ($334<<24>>24)==(77);
    if ($335) {
     HEAP32[$330>>2] = 1634889070;
     HEAP16[$331>>1] = 1;
    }
    $336 = (_FT_CMap_New(59164,0,$charmap,0)|0);
    HEAP32[$error>>2] = $336;
    $337 = ($336|0)==(0);
    if (!($337)) {
     break;
    }
    $338 = ((($fntface)) + 36|0);
    $339 = HEAP32[$338>>2]|0;
    $340 = ($339|0)==(0);
    if (!($340)) {
     $341 = ((($fntface)) + 40|0);
     $342 = HEAP32[$341>>2]|0;
     $343 = HEAP32[$342>>2]|0;
     $344 = ((($fntface)) + 92|0);
     HEAP32[$344>>2] = $343;
    }
    $345 = ((($262)) + 105|0);
    $346 = HEAP8[$345>>0]|0;
    $347 = ((($262)) + 104|0);
    $348 = HEAP8[$347>>0]|0;
    $349 = ($346&255)<($348&255);
    if ($349) {
     HEAP32[$error>>2] = 3;
     break;
    }
    $350 = $346&255;
    $351 = $348&255;
    $352 = (($350) + 2)|0;
    $353 = (($352) - ($351))|0;
    $354 = ((($fntface)) + 16|0);
    HEAP32[$354>>2] = $353;
    $355 = ((($262)) + 116|0);
    $356 = HEAP32[$355>>2]|0;
    $357 = ((($262)) + 8|0);
    $358 = HEAP32[$357>>2]|0;
    $359 = ($358>>>0)>($356>>>0);
    if (!($359)) {
     HEAP32[$error>>2] = 3;
     break;
    }
    $360 = (($358) - ($356))|0;
    $361 = (($360) + 1)|0;
    $362 = (_ft_mem_alloc($1,$361,$error)|0);
    $363 = ((($262)) + 168|0);
    HEAP32[$363>>2] = $362;
    $364 = HEAP32[$error>>2]|0;
    $365 = ($364|0)==(0);
    if ($365) {
     $366 = ((($262)) + 160|0);
     $367 = HEAP32[$366>>2]|0;
     $368 = HEAP32[$355>>2]|0;
     $369 = (($367) + ($368)|0);
     _memcpy(($362|0),($369|0),($360|0))|0;
     $370 = HEAP32[$363>>2]|0;
     $371 = (($370) + ($360)|0);
     HEAP8[$371>>0] = 0;
     $372 = HEAP32[$363>>2]|0;
     $373 = (_strlen($372)|0);
     $374 = (($373) + 1)|0;
     $375 = (_ft_mem_realloc($1,1,$360,$374,$372,$error)|0);
     HEAP32[$363>>2] = $375;
     $376 = HEAP32[$error>>2]|0;
     $377 = ($376|0)==(0);
     if ($377) {
      $378 = ((($fntface)) + 20|0);
      HEAP32[$378>>2] = $375;
      $379 = ((($fntface)) + 24|0);
      HEAP32[$379>>2] = 594711;
      $380 = ((($fntface)) + 12|0);
      $381 = HEAP32[$380>>2]|0;
      $382 = $381 & 2;
      $383 = ($382|0)==(0);
      $384 = $381 & 1;
      $385 = ($384|0)!=(0);
      if ($383) {
       if (!($385)) {
        $399 = 0;
        STACKTOP = sp;return ($399|0);
       }
       HEAP32[$379>>2] = 594736;
       $399 = 0;
       STACKTOP = sp;return ($399|0);
      }
      if ($385) {
       HEAP32[$379>>2] = 594719;
       $399 = 0;
       STACKTOP = sp;return ($399|0);
      } else {
       HEAP32[$379>>2] = 594731;
       $399 = 0;
       STACKTOP = sp;return ($399|0);
      }
     }
    }
   }
  }
 } while(0);
 $387 = HEAP32[$0>>2]|0;
 $388 = HEAP32[$2>>2]|0;
 $389 = HEAP32[$4>>2]|0;
 $390 = ($389|0)==(0|0);
 if (!($390)) {
  $391 = ((($389)) + 160|0);
  $392 = HEAP32[$391>>2]|0;
  $393 = ($392|0)==(0|0);
  if (!($393)) {
   _FT_Stream_ReleaseFrame($388,$391);
  }
  $394 = ((($389)) + 168|0);
  $395 = HEAP32[$394>>2]|0;
  _ft_mem_free($387,$395);
  HEAP32[$394>>2] = 0;
  _ft_mem_free($387,$389);
  HEAP32[$4>>2] = 0;
 }
 $396 = ((($fntface)) + 32|0);
 $397 = HEAP32[$396>>2]|0;
 _ft_mem_free($387,$397);
 HEAP32[$396>>2] = 0;
 $398 = ((($fntface)) + 28|0);
 HEAP32[$398>>2] = 0;
 $$pre39 = HEAP32[$error>>2]|0;
 $399 = $$pre39;
 STACKTOP = sp;return ($399|0);
}
function _fnt_font_load($font,$stream) {
 $font = $font|0;
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($font)) + 4|0);
 $1 = HEAP32[$font>>2]|0;
 $2 = (_FT_Stream_Seek($stream,$1)|0);
 $3 = ($2|0)==(0);
 if (!($3)) {
  $error$0 = $2;
  return ($error$0|0);
 }
 $4 = (_FT_Stream_ReadFields($stream,174954,$0)|0);
 $5 = ($4|0)==(0);
 if (!($5)) {
  $error$0 = $4;
  return ($error$0|0);
 }
 $6 = HEAP16[$0>>1]|0;
 switch ($6<<16>>16) {
 case 768: case 512:  {
  break;
 }
 default: {
  $error$0 = 2;
  return ($error$0|0);
 }
 }
 $7 = ($6<<16>>16)==(768);
 $8 = $7 ? 148 : 118;
 $9 = ((($font)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10>>>0)<($8>>>0);
 if ($11) {
  $error$0 = 2;
  return ($error$0|0);
 }
 $12 = ($6<<16>>16)==(512);
 if ($12) {
  $13 = ((($font)) + 132|0);
  ;HEAP32[$13>>2]=0|0;HEAP32[$13+4>>2]=0|0;HEAP32[$13+8>>2]=0|0;
 }
 $14 = ((($font)) + 72|0);
 $15 = HEAP16[$14>>1]|0;
 $16 = $15 & 1;
 $17 = ($16<<16>>16)==(0);
 if (!($17)) {
  $error$0 = 2;
  return ($error$0|0);
 }
 $18 = HEAP32[$font>>2]|0;
 $19 = (_FT_Stream_Seek($stream,$18)|0);
 $20 = ($19|0)==(0);
 if (!($20)) {
  $error$0 = $19;
  return ($error$0|0);
 }
 $21 = HEAP32[$9>>2]|0;
 $22 = ((($font)) + 160|0);
 $23 = (_FT_Stream_ExtractFrame($stream,$21,$22)|0);
 $error$0 = $23;
 return ($error$0|0);
}
function _fnt_cmap_init($cmap,$pointer) {
 $cmap = $cmap|0;
 $pointer = $pointer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$cmap>>2]|0;
 $1 = ((($0)) + 132|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 104|0);
 $4 = HEAP16[$3>>1]|0;
 $5 = $4&65535;
 $6 = $5 & 255;
 $7 = ((($cmap)) + 16|0);
 HEAP32[$7>>2] = $6;
 $8 = ($4&65535) >>> 8;
 $9 = $8&65535;
 $10 = (1 - ($6))|0;
 $11 = (($10) + ($9))|0;
 $12 = ((($cmap)) + 20|0);
 HEAP32[$12>>2] = $11;
 return 0;
}
function _fnt_cmap_char_index($cmap,$char_code) {
 $cmap = $cmap|0;
 $char_code = $char_code|0;
 var $$ = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($cmap)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($char_code) - ($1))|0;
 $3 = ((($cmap)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2>>>0)<($4>>>0);
 $6 = (($2) + 1)|0;
 $$ = $5 ? $6 : 0;
 return ($$|0);
}
function _fnt_cmap_char_next($cmap,$pchar_code) {
 $cmap = $cmap|0;
 $pchar_code = $pchar_code|0;
 var $$ = 0, $$1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $gindex$0 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$pchar_code>>2]|0;
 $1 = (($0) + 1)|0;
 $2 = ((($cmap)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1>>>0)>($3>>>0);
 if (!($4)) {
  $gindex$0 = 1;$result$0 = $3;
  HEAP32[$pchar_code>>2] = $result$0;
  return ($gindex$0|0);
 }
 $5 = (($1) - ($3))|0;
 $6 = ((($cmap)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($5>>>0)<($7>>>0);
 $9 = (($5) + 1)|0;
 $$ = $8 ? $9 : 0;
 $$1 = $8 ? $1 : 0;
 $gindex$0 = $$;$result$0 = $$1;
 HEAP32[$pchar_code>>2] = $result$0;
 return ($gindex$0|0);
}
function _FNT_Face_Done($fntface) {
 $fntface = $fntface|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($fntface|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($fntface)) + 100|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($fntface)) + 104|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($fntface)) + 132|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0|0);
 if (!($7)) {
  $8 = ((($6)) + 160|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)==(0|0);
  if (!($10)) {
   _FT_Stream_ReleaseFrame($4,$8);
  }
  $11 = ((($6)) + 168|0);
  $12 = HEAP32[$11>>2]|0;
  _ft_mem_free($2,$12);
  HEAP32[$11>>2] = 0;
  _ft_mem_free($2,$6);
  HEAP32[$5>>2] = 0;
 }
 $13 = ((($fntface)) + 32|0);
 $14 = HEAP32[$13>>2]|0;
 _ft_mem_free($2,$14);
 HEAP32[$13>>2] = 0;
 $15 = ((($fntface)) + 28|0);
 HEAP32[$15>>2] = 0;
 return;
}
function _FNT_Load_Glyph($slot,$size,$glyph_index,$load_flags) {
 $slot = $slot|0;
 $size = $size|0;
 $glyph_index = $glyph_index|0;
 $load_flags = $load_flags|0;
 var $$02 = 0, $$neg = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $column$08 = 0, $error = 0, $offset$0 = 0, $p$06 = 0, $p$1$lcssa = 0, $p$14 = 0, $pitch$07 = 0, $write$05 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = HEAP32[$size>>2]|0;
 HEAP32[$error>>2] = 0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  HEAP32[$error>>2] = 35;
  $118 = 35;
  STACKTOP = sp;return ($118|0);
 }
 $2 = ((($0)) + 132|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  $5 = ((($0)) + 16|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6>>>0)>($glyph_index>>>0);
  if ($7) {
   $8 = ($glyph_index|0)==(0);
   if ($8) {
    $10 = ((($3)) + 106|0);
    $11 = HEAP8[$10>>0]|0;
    $12 = $11&255;
    $$02 = $12;
   } else {
    $9 = (($glyph_index) + -1)|0;
    $$02 = $9;
   }
   $13 = ((($3)) + 4|0);
   $14 = HEAP16[$13>>1]|0;
   $15 = ($14<<16>>16)==(768);
   $16 = $15 ? 6 : 4;
   $17 = $15 ? 148 : 118;
   $18 = Math_imul($16, $$02)|0;
   $19 = (($18) + ($17))|0;
   $20 = ((($3)) + 8|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = (($21) + -2)|0;
   $$neg = $15 ? -4 : -2;
   $23 = (($22) + ($$neg))|0;
   $24 = ($19>>>0)<($23>>>0);
   if (!($24)) {
    HEAP32[$error>>2] = 3;
    $118 = 3;
    STACKTOP = sp;return ($118|0);
   }
   $25 = ((($3)) + 160|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = (($26) + ($19)|0);
   $28 = ((($27)) + 2|0);
   $29 = ((($28)) + -2|0);
   $30 = ((($29)) + 1|0);
   $31 = HEAP8[$30>>0]|0;
   $32 = $31&255;
   $33 = $32 << 8;
   $34 = HEAP8[$29>>0]|0;
   $35 = $34&255;
   $36 = $33 | $35;
   $37 = ((($slot)) + 80|0);
   HEAP32[$37>>2] = $36;
   if ($15) {
    $38 = ((($28)) + 4|0);
    $39 = ((($38)) + -4|0);
    $40 = ((($39)) + 3|0);
    $41 = HEAP8[$40>>0]|0;
    $42 = $41&255;
    $43 = $42 << 24;
    $44 = ((($39)) + 2|0);
    $45 = HEAP8[$44>>0]|0;
    $46 = $45&255;
    $47 = $46 << 16;
    $48 = $47 | $43;
    $49 = ((($39)) + 1|0);
    $50 = HEAP8[$49>>0]|0;
    $51 = $50&255;
    $52 = $51 << 8;
    $53 = $48 | $52;
    $54 = HEAP8[$39>>0]|0;
    $55 = $54&255;
    $56 = $53 | $55;
    $offset$0 = $56;
   } else {
    $57 = ((($28)) + 2|0);
    $58 = ((($57)) + -2|0);
    $59 = ((($58)) + 1|0);
    $60 = HEAP8[$59>>0]|0;
    $61 = $60&255;
    $62 = $61 << 8;
    $63 = HEAP8[$58>>0]|0;
    $64 = $63&255;
    $65 = $62 | $64;
    $offset$0 = $65;
   }
   $66 = ($offset$0>>>0)<($21>>>0);
   if (!($66)) {
    HEAP32[$error>>2] = 3;
    $118 = 3;
    STACKTOP = sp;return ($118|0);
   }
   $67 = ((($slot)) + 4|0);
   $68 = HEAP32[$67>>2]|0;
   $69 = ((($68)) + 100|0);
   $70 = HEAP32[$69>>2]|0;
   $71 = (($36) + 7)|0;
   $72 = $71 >>> 3;
   $73 = ((($slot)) + 84|0);
   HEAP32[$73>>2] = $72;
   $74 = ((($3)) + 96|0);
   $75 = HEAP16[$74>>1]|0;
   $76 = $75&65535;
   $77 = ((($slot)) + 76|0);
   HEAP32[$77>>2] = $76;
   $78 = ((($slot)) + 94|0);
   HEAP8[$78>>0] = 1;
   $79 = Math_imul($76, $72)|0;
   $80 = (($79) + ($offset$0))|0;
   $81 = ($80>>>0)>($21>>>0);
   if ($81) {
    HEAP32[$error>>2] = 3;
    $118 = 3;
    STACKTOP = sp;return ($118|0);
   }
   $82 = (_ft_mem_realloc($70,$76,0,$72,0,$error)|0);
   $83 = ((($slot)) + 88|0);
   HEAP32[$83>>2] = $82;
   $84 = HEAP32[$error>>2]|0;
   $85 = ($84|0)==(0);
   if (!($85)) {
    $118 = $84;
    STACKTOP = sp;return ($118|0);
   }
   $86 = ($72|0)==(0);
   if (!($86)) {
    $87 = (($26) + ($offset$0)|0);
    $column$08 = $82;$p$06 = $87;$pitch$07 = $72;
    while(1) {
     $88 = HEAP32[$77>>2]|0;
     $89 = (($p$06) + ($88)|0);
     $90 = ($88|0)>(0);
     if ($90) {
      $p$14 = $p$06;$write$05 = $column$08;
      while(1) {
       $91 = HEAP8[$p$14>>0]|0;
       HEAP8[$write$05>>0] = $91;
       $92 = ((($p$14)) + 1|0);
       $93 = HEAP32[$73>>2]|0;
       $94 = (($write$05) + ($93)|0);
       $95 = ($92>>>0)<($89>>>0);
       if ($95) {
        $p$14 = $92;$write$05 = $94;
       } else {
        $p$1$lcssa = $92;
        break;
       }
      }
     } else {
      $p$1$lcssa = $p$06;
     }
     $96 = (($pitch$07) + -1)|0;
     $97 = ((($column$08)) + 1|0);
     $98 = ($96|0)==(0);
     if ($98) {
      break;
     } else {
      $column$08 = $97;$p$06 = $p$1$lcssa;$pitch$07 = $96;
     }
    }
   }
   $99 = ((($slot)) + 156|0);
   $100 = HEAP32[$99>>2]|0;
   $101 = ((($100)) + 4|0);
   HEAP32[$101>>2] = 1;
   $102 = ((($slot)) + 100|0);
   HEAP32[$102>>2] = 0;
   $103 = ((($3)) + 80|0);
   $104 = HEAP16[$103>>1]|0;
   $105 = $104&65535;
   $106 = ((($slot)) + 104|0);
   HEAP32[$106>>2] = $105;
   $107 = ((($slot)) + 72|0);
   HEAP32[$107>>2] = 1651078259;
   $108 = HEAP32[$37>>2]|0;
   $109 = $108 << 6;
   $110 = ((($slot)) + 24|0);
   HEAP32[$110>>2] = $109;
   $111 = HEAP32[$77>>2]|0;
   $112 = $111 << 6;
   $113 = ((($slot)) + 28|0);
   HEAP32[$113>>2] = $112;
   $114 = ((($slot)) + 40|0);
   HEAP32[$114>>2] = $109;
   $115 = ((($slot)) + 32|0);
   HEAP32[$115>>2] = 0;
   $116 = $105 << 6;
   $117 = ((($slot)) + 36|0);
   HEAP32[$117>>2] = $116;
   _ft_synthesize_vertical_metrics($110,$112);
   $$pre = HEAP32[$error>>2]|0;
   $118 = $$pre;
   STACKTOP = sp;return ($118|0);
  }
 }
 HEAP32[$error>>2] = 6;
 $118 = 6;
 STACKTOP = sp;return ($118|0);
}
function _FNT_Size_Request($size,$req) {
 $size = $size|0;
 $req = $req|0;
 var $$0 = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$size>>2]|0;
 $1 = ((($0)) + 132|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($0)) + 32|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($req)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 $8 = ((($req)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 if ($7) {
  $14 = $9;
 } else {
  $10 = Math_imul($9, $6)|0;
  $11 = (($10) + 36)|0;
  $12 = (($11|0) / 72)&-1;
  $14 = $12;
 }
 $13 = (($14) + 32)|0;
 $15 = $13 >> 6;
 $16 = HEAP32[$req>>2]|0;
 L4: do {
  switch ($16|0) {
  case 0:  {
   $17 = ((($4)) + 12|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = (($18) + 32)|0;
   $20 = $19 >> 6;
   $21 = ($15|0)==($20|0);
   if ($21) {
    $$pre = ((($2)) + 96|0);
    $$pre$phiZ2D = $$pre;
    break L4;
   } else {
    $$0 = 23;
    return ($$0|0);
   }
   break;
  }
  case 1:  {
   $22 = ((($2)) + 96|0);
   $23 = HEAP16[$22>>1]|0;
   $24 = $23&65535;
   $25 = ($15|0)==($24|0);
   if ($25) {
    $$pre$phiZ2D = $22;
   } else {
    $$0 = 23;
    return ($$0|0);
   }
   break;
  }
  default: {
   $$0 = 7;
   return ($$0|0);
  }
  }
 } while(0);
 _FT_Select_Metrics($0,0);
 $26 = ((($2)) + 80|0);
 $27 = HEAP16[$26>>1]|0;
 $28 = $27&65535;
 $29 = $28 << 6;
 $30 = ((($size)) + 24|0);
 HEAP32[$30>>2] = $29;
 $31 = HEAP16[$$pre$phiZ2D>>1]|0;
 $32 = $31&65535;
 $33 = (($28) - ($32))|0;
 $34 = $33 << 6;
 $35 = ((($size)) + 28|0);
 HEAP32[$35>>2] = $34;
 $36 = ((($2)) + 102|0);
 $37 = HEAP16[$36>>1]|0;
 $38 = $37&65535;
 $39 = $38 << 6;
 $40 = ((($size)) + 36|0);
 HEAP32[$40>>2] = $39;
 $$0 = 0;
 return ($$0|0);
}
function _FNT_Size_Select($size,$strike_index) {
 $size = $size|0;
 $strike_index = $strike_index|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$size>>2]|0;
 $1 = ((($0)) + 132|0);
 $2 = HEAP32[$1>>2]|0;
 _FT_Select_Metrics($0,0);
 $3 = ((($2)) + 80|0);
 $4 = HEAP16[$3>>1]|0;
 $5 = $4&65535;
 $6 = $5 << 6;
 $7 = ((($size)) + 24|0);
 HEAP32[$7>>2] = $6;
 $8 = ((($2)) + 96|0);
 $9 = HEAP16[$8>>1]|0;
 $10 = $9&65535;
 $11 = (($5) - ($10))|0;
 $12 = $11 << 6;
 $13 = ((($size)) + 28|0);
 HEAP32[$13>>2] = $12;
 $14 = ((($2)) + 102|0);
 $15 = HEAP16[$14>>1]|0;
 $16 = $15&65535;
 $17 = $16 << 6;
 $18 = ((($size)) + 36|0);
 HEAP32[$18>>2] = $17;
 return 0;
}
function _af_dummy_hints_init($hints,$metrics) {
 $hints = $hints|0;
 $metrics = $metrics|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($hints)) + 2988|0);
 HEAP32[$0>>2] = $metrics;
 $1 = ((($metrics)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($hints)) + 2980|0);
 HEAP32[$3>>2] = $2;
 $4 = ((($metrics)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($hints)) + 4|0);
 HEAP32[$6>>2] = $5;
 $7 = ((($metrics)) + 12|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($hints)) + 12|0);
 HEAP32[$9>>2] = $8;
 $10 = ((($metrics)) + 16|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($hints)) + 8|0);
 HEAP32[$12>>2] = $11;
 $13 = ((($metrics)) + 20|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($hints)) + 16|0);
 HEAP32[$15>>2] = $14;
 return 0;
}
function _af_dummy_hints_apply($hints,$outline,$metrics) {
 $hints = $hints|0;
 $outline = $outline|0;
 $metrics = $metrics|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hints$idx = 0;
 var $hints$idx$val = 0, $hints$idx1 = 0, $hints$idx1$val = 0, $outline$idx = 0, $outline$idx$val = 0, $outline$idx2 = 0, $outline$idx2$val = 0, $point$01$i = 0, $tag$03$i = 0, $vec$02$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_af_glyph_hints_reload($hints,$outline)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  return ($0|0);
 }
 $hints$idx = ((($hints)) + 24|0);
 $hints$idx$val = HEAP32[$hints$idx>>2]|0;
 $hints$idx1 = ((($hints)) + 28|0);
 $hints$idx1$val = HEAP32[$hints$idx1>>2]|0;
 $2 = (($hints$idx1$val) + (($hints$idx$val*40)|0)|0);
 $3 = ($hints$idx$val|0)>(0);
 if (!($3)) {
  return ($0|0);
 }
 $outline$idx2 = ((($outline)) + 8|0);
 $outline$idx2$val = HEAP32[$outline$idx2>>2]|0;
 $outline$idx = ((($outline)) + 4|0);
 $outline$idx$val = HEAP32[$outline$idx>>2]|0;
 $point$01$i = $hints$idx1$val;$tag$03$i = $outline$idx2$val;$vec$02$i = $outline$idx$val;
 while(1) {
  $4 = ((($point$01$i)) + 16|0);
  $5 = HEAP32[$4>>2]|0;
  HEAP32[$vec$02$i>>2] = $5;
  $6 = ((($point$01$i)) + 20|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ((($vec$02$i)) + 4|0);
  HEAP32[$8>>2] = $7;
  $9 = HEAP16[$point$01$i>>1]|0;
  $10 = $9&65535;
  $11 = $10 & 1;
  $12 = ($11|0)==(0);
  do {
   if ($12) {
    $13 = $10 & 2;
    $14 = ($13|0)==(0);
    if ($14) {
     HEAP8[$tag$03$i>>0] = 1;
     break;
    } else {
     HEAP8[$tag$03$i>>0] = 2;
     break;
    }
   } else {
    HEAP8[$tag$03$i>>0] = 0;
   }
  } while(0);
  $15 = ((($point$01$i)) + 40|0);
  $16 = ((($vec$02$i)) + 8|0);
  $17 = ((($tag$03$i)) + 1|0);
  $18 = ($15>>>0)<($2>>>0);
  if ($18) {
   $point$01$i = $15;$tag$03$i = $17;$vec$02$i = $16;
  } else {
   break;
  }
 }
 return ($0|0);
}
function _af_glyph_hints_reload($hints,$outline) {
 $hints = $hints|0;
 $outline = $outline|0;
 var $$$i = 0, $$1$i = 0, $$idx24 = 0, $$idx57 = 0, $$in = 0, $$phi$trans$insert = 0, $$phi$trans$insert37$phi$trans$insert = 0, $$pn = 0, $$pn$lcssa = 0, $$pre = 0, $$pre35 = 0, $$pre36 = 0, $$pre38$pre = 0, $$pre38$pre$lcssa = 0, $$pre39 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0;
 var $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0;
 var $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0;
 var $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0;
 var $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0;
 var $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $contour$021 = 0, $contour2$016 = 0, $contour_index$025 = 0, $contour_index$1 = 0, $curr$0 = 0, $curr$1 = 0, $curr$1$in = 0, $curr$1$in12 = 0;
 var $curr$113 = 0, $curr$114 = 0, $curr$2 = 0, $dir$0$i = 0, $dx$$i = 0, $dx$dy$i = 0, $dy$$i = 0, $dy$dx$i = 0, $end$027 = 0, $end$1 = 0, $end1$020 = 0, $error = 0, $idx$019 = 0, $ll$0$i = 0, $next$0 = 0, $out_x$0 = 0, $out_x$1 = 0, $out_y$0 = 0, $out_y$1 = 0, $point$030 = 0;
 var $point$210 = 0, $point$39 = 0, $prev$026 = 0, $prev$1 = 0, $prev4$0 = 0, $prev4$0$in = 0, $sext = 0, $sext43 = 0, $sext45 = 0, $sext46 = 0, $sext48 = 0, $sext50 = 0, $sext51 = 0, $sext54 = 0, $sext55 = 0, $ss$0$i = 0, $tag$028 = 0, $vec$029 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 HEAP32[$error>>2] = 0;
 $0 = ((($hints)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($hints)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($hints)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($hints)) + 16|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = HEAP32[$hints>>2]|0;
 $9 = ((($hints)) + 24|0);
 HEAP32[$9>>2] = 0;
 $10 = ((($hints)) + 36|0);
 HEAP32[$10>>2] = 0;
 $11 = ((($hints)) + 44|0);
 HEAP32[$11>>2] = 0;
 $12 = ((($hints)) + 56|0);
 HEAP32[$12>>2] = 0;
 $13 = ((($hints)) + 1512|0);
 HEAP32[$13>>2] = 0;
 $14 = ((($hints)) + 1524|0);
 HEAP32[$14>>2] = 0;
 $15 = HEAP32[$outline>>2]|0;
 $16 = $15&65535;
 $sext43 = $15 << 16;
 $17 = $sext43 >> 16;
 $18 = ((($hints)) + 32|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($16&65535)<(9);
 $21 = $15 >>> 16;
 $22 = $21&65535;
 do {
  if ($20) {
   $23 = ((($hints)) + 40|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = ($24|0)==(0|0);
   if ($25) {
    $26 = ((($hints)) + 3000|0);
    HEAP32[$23>>2] = $26;
    HEAP32[$18>>2] = 8;
    $40 = $22;
   } else {
    $40 = $22;
   }
  } else {
   $27 = ($17>>>0)>($19>>>0);
   if ($27) {
    $28 = ((($hints)) + 40|0);
    $29 = HEAP32[$28>>2]|0;
    $30 = ((($hints)) + 3000|0);
    $31 = ($29|0)==($30|0);
    if ($31) {
     HEAP32[$28>>2] = 0;
     $34 = 0;
    } else {
     $34 = $29;
    }
    $32 = (($17) + 3)|0;
    $33 = $32 & -4;
    $35 = (_ft_mem_realloc($8,4,$19,$33,$34,$error)|0);
    HEAP32[$28>>2] = $35;
    $36 = HEAP32[$error>>2]|0;
    $37 = ($36|0)==(0);
    if ($37) {
     HEAP32[$18>>2] = $33;
     $$phi$trans$insert = ((($outline)) + 2|0);
     $$pre = HEAP16[$$phi$trans$insert>>1]|0;
     $40 = $$pre;
     break;
    } else {
     $315 = HEAP32[$error>>2]|0;
     STACKTOP = sp;return ($315|0);
    }
   } else {
    $40 = $22;
   }
  }
 } while(0);
 $38 = ((($outline)) + 2|0);
 $39 = $40 << 16 >> 16;
 $41 = (($39) + 2)|0;
 $42 = ((($hints)) + 20|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = ($41>>>0)<(97);
 do {
  if ($44) {
   $45 = ((($hints)) + 28|0);
   $46 = HEAP32[$45>>2]|0;
   $47 = ($46|0)==(0|0);
   if ($47) {
    $48 = ((($hints)) + 3032|0);
    HEAP32[$45>>2] = $48;
    HEAP32[$42>>2] = 96;
   }
  } else {
   $49 = ($41>>>0)>($43>>>0);
   if ($49) {
    $50 = ((($hints)) + 28|0);
    $51 = HEAP32[$50>>2]|0;
    $52 = ((($hints)) + 3032|0);
    $53 = ($51|0)==($52|0);
    if ($53) {
     HEAP32[$50>>2] = 0;
     $56 = 0;
    } else {
     $56 = $51;
    }
    $54 = (($39) + 11)|0;
    $55 = $54 & -8;
    $57 = (_ft_mem_realloc($8,40,$43,$55,$56,$error)|0);
    HEAP32[$50>>2] = $57;
    $58 = HEAP32[$error>>2]|0;
    $59 = ($58|0)==(0);
    if ($59) {
     HEAP32[$42>>2] = $55;
     break;
    } else {
     $315 = HEAP32[$error>>2]|0;
     STACKTOP = sp;return ($315|0);
    }
   }
  }
 } while(0);
 $60 = HEAP16[$38>>1]|0;
 $61 = $60 << 16 >> 16;
 HEAP32[$9>>2] = $61;
 $62 = HEAP16[$outline>>1]|0;
 $63 = $62 << 16 >> 16;
 HEAP32[$10>>2] = $63;
 $64 = ((($hints)) + 68|0);
 HEAP32[$64>>2] = 2;
 $65 = ((($hints)) + 1536|0);
 HEAP32[$65>>2] = -1;
 $66 = (_FT_Outline_Get_Orientation($outline)|0);
 $67 = ($66|0)==(1);
 if ($67) {
  HEAP32[$64>>2] = -2;
  HEAP32[$65>>2] = 1;
 }
 HEAP32[$0>>2] = $1;
 HEAP32[$2>>2] = $3;
 HEAP32[$4>>2] = $5;
 HEAP32[$6>>2] = $7;
 $68 = ((($hints)) + 2992|0);
 HEAP32[$68>>2] = 0;
 $69 = ((($hints)) + 2996|0);
 HEAP32[$69>>2] = 0;
 $70 = ((($hints)) + 28|0);
 $71 = HEAP32[$70>>2]|0;
 $72 = HEAP32[$9>>2]|0;
 $73 = ($72|0)==(0);
 if ($73) {
  $315 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($315|0);
 }
 $74 = (($71) + (($72*40)|0)|0);
 $75 = ((($outline)) + 12|0);
 $76 = HEAP32[$75>>2]|0;
 $77 = HEAP16[$76>>1]|0;
 $78 = $77 << 16 >> 16;
 $79 = (($71) + (($78*40)|0)|0);
 $80 = ($72|0)>(0);
 if ($80) {
  $81 = ((($outline)) + 8|0);
  $82 = HEAP32[$81>>2]|0;
  $83 = ((($outline)) + 4|0);
  $84 = HEAP32[$83>>2]|0;
  $contour_index$025 = 0;$end$027 = $79;$point$030 = $71;$prev$026 = $79;$tag$028 = $82;$vec$029 = $84;
  while(1) {
   $85 = ((($point$030)) + 2|0);
   HEAP8[$85>>0] = 4;
   $86 = ((($point$030)) + 3|0);
   HEAP8[$86>>0] = 4;
   $87 = HEAP32[$vec$029>>2]|0;
   $88 = $87&65535;
   $89 = ((($point$030)) + 12|0);
   HEAP16[$89>>1] = $88;
   $90 = ((($vec$029)) + 4|0);
   $91 = HEAP32[$90>>2]|0;
   $92 = $91&65535;
   $93 = ((($point$030)) + 14|0);
   HEAP16[$93>>1] = $92;
   $94 = (_FT_MulFix($87,$1)|0);
   $95 = (($94) + ($5))|0;
   $96 = ((($point$030)) + 16|0);
   HEAP32[$96>>2] = $95;
   $97 = ((($point$030)) + 4|0);
   HEAP32[$97>>2] = $95;
   $98 = HEAP32[$90>>2]|0;
   $99 = (_FT_MulFix($98,$3)|0);
   $100 = (($99) + ($7))|0;
   $101 = ((($point$030)) + 20|0);
   HEAP32[$101>>2] = $100;
   $102 = ((($point$030)) + 8|0);
   HEAP32[$102>>2] = $100;
   $103 = HEAP8[$tag$028>>0]|0;
   $104 = $103&255;
   $105 = $104 & 3;
   switch ($105|0) {
   case 0:  {
    HEAP16[$point$030>>1] = 1;
    break;
   }
   case 2:  {
    HEAP16[$point$030>>1] = 2;
    break;
   }
   default: {
    HEAP16[$point$030>>1] = 0;
   }
   }
   $106 = ((($point$030)) + 36|0);
   HEAP32[$106>>2] = $prev$026;
   $107 = ((($prev$026)) + 32|0);
   HEAP32[$107>>2] = $point$030;
   $108 = ($point$030|0)==($end$027|0);
   if ($108) {
    $109 = (($contour_index$025) + 1)|0;
    $110 = HEAP16[$outline>>1]|0;
    $111 = $110 << 16 >> 16;
    $112 = ($109|0)<($111|0);
    if ($112) {
     $113 = HEAP32[$75>>2]|0;
     $114 = (($113) + ($109<<1)|0);
     $115 = HEAP16[$114>>1]|0;
     $116 = $115 << 16 >> 16;
     $117 = (($71) + (($116*40)|0)|0);
     $contour_index$1 = $109;$end$1 = $117;$prev$1 = $117;
    } else {
     $contour_index$1 = $109;$end$1 = $point$030;$prev$1 = $point$030;
    }
   } else {
    $contour_index$1 = $contour_index$025;$end$1 = $end$027;$prev$1 = $point$030;
   }
   $118 = ((($point$030)) + 40|0);
   $119 = ((($vec$029)) + 8|0);
   $120 = ((($tag$028)) + 1|0);
   $121 = ($118>>>0)<($74>>>0);
   if ($121) {
    $contour_index$025 = $contour_index$1;$end$027 = $end$1;$point$030 = $118;$prev$026 = $prev$1;$tag$028 = $120;$vec$029 = $119;
   } else {
    break;
   }
  }
  $$pre35 = HEAP32[$75>>2]|0;
  $316 = $$pre35;
 } else {
  $316 = $76;
 }
 $122 = ((($hints)) + 40|0);
 $123 = HEAP32[$122>>2]|0;
 $124 = HEAP32[$10>>2]|0;
 $125 = (($123) + ($124<<2)|0);
 $126 = ($124|0)>(0);
 if ($126) {
  $contour$021 = $123;$end1$020 = $316;$idx$019 = 0;
  while(1) {
   $sext = $idx$019 << 16;
   $127 = $sext >> 16;
   $128 = (($71) + (($127*40)|0)|0);
   HEAP32[$contour$021>>2] = $128;
   $129 = HEAP16[$end1$020>>1]|0;
   $130 = $129 << 16 >> 16;
   $131 = (($130) + 1)|0;
   $132 = ((($contour$021)) + 4|0);
   $133 = ((($end1$020)) + 2|0);
   $134 = ($132>>>0)<($125>>>0);
   if ($134) {
    $contour$021 = $132;$end1$020 = $133;$idx$019 = $131;
   } else {
    break;
   }
  }
  $$pre36 = HEAP32[$122>>2]|0;
  $147 = $$pre36;
 } else {
  $147 = $123;
 }
 $135 = ((($hints)) + 2988|0);
 $136 = HEAP32[$135>>2]|0;
 $137 = ((($136)) + 4|0);
 $138 = HEAP32[$137>>2]|0;
 $139 = ((($138)) + 68|0);
 $140 = HEAP16[$139>>1]|0;
 $141 = $140&65535;
 $142 = ($141*20)|0;
 $143 = $142 >>> 11;
 $144 = $143 << 1;
 $145 = (($144) + -1)|0;
 $146 = (($147) + ($124<<2)|0);
 if ($126) {
  $contour2$016 = $147;
  while(1) {
   $148 = HEAP32[$contour2$016>>2]|0;
   $$pn = $148;
   while(1) {
    $prev4$0$in = ((($$pn)) + 36|0);
    $prev4$0 = HEAP32[$prev4$0$in>>2]|0;
    $149 = ($prev4$0|0)==($148|0);
    $$phi$trans$insert37$phi$trans$insert = ((($$pn)) + 12|0);
    $$pre38$pre = HEAP32[$$phi$trans$insert37$phi$trans$insert>>2]|0;
    if ($149) {
     $$pn$lcssa = $$pn;$$pre38$pre$lcssa = $$pre38$pre;
     break;
    }
    $sext50 = $$pre38$pre << 16;
    $150 = $sext50 >> 16;
    $151 = ((($prev4$0)) + 12|0);
    $152 = HEAP32[$151>>2]|0;
    $sext51 = $152 << 16;
    $153 = $sext51 >> 16;
    $154 = (($150) - ($153))|0;
    $155 = $$pre38$pre >> 16;
    $156 = $152 >> 16;
    $157 = (($155) - ($156))|0;
    $158 = ($154|0)<(0);
    $159 = (0 - ($154))|0;
    $160 = $158 ? $159 : $154;
    $161 = ($157|0)<(0);
    $162 = (0 - ($157))|0;
    $163 = $161 ? $162 : $157;
    $164 = (($163) + ($160))|0;
    $165 = ($164|0)<($145|0);
    if ($165) {
     $$pn = $prev4$0;
    } else {
     $$pn$lcssa = $$pn;$$pre38$pre$lcssa = $$pre38$pre;
     break;
    }
   }
   $166 = $$pn$lcssa;
   $167 = ((($$pn$lcssa)) + 24|0);
   HEAP32[$167>>2] = 0;
   $168 = ((($$pn$lcssa)) + 28|0);
   HEAP32[$168>>2] = 0;
   $$in = $$pre38$pre$lcssa;$curr$0 = $$pn$lcssa;$next$0 = $$pn$lcssa;$out_x$0 = 0;$out_y$0 = 0;
   while(1) {
    $169 = ((($next$0)) + 32|0);
    $170 = HEAP32[$169>>2]|0;
    $171 = ((($170)) + 12|0);
    $172 = HEAP32[$171>>2]|0;
    $sext54 = $172 << 16;
    $173 = $sext54 >> 16;
    $sext55 = $$in << 16;
    $174 = $sext55 >> 16;
    $175 = (($173) - ($174))|0;
    $176 = (($175) + ($out_x$0))|0;
    $177 = $172 >> 16;
    $178 = $$in >> 16;
    $179 = (($177) - ($178))|0;
    $180 = (($179) + ($out_y$0))|0;
    $181 = ($176|0)<(0);
    $182 = (0 - ($176))|0;
    $183 = $181 ? $182 : $176;
    $184 = ($180|0)<(0);
    $185 = (0 - ($180))|0;
    $186 = $184 ? $185 : $180;
    $187 = (($186) + ($183))|0;
    $188 = ($187|0)<($143|0);
    if ($188) {
     $189 = HEAP16[$170>>1]|0;
     $190 = $189&65535;
     $191 = $190 | 16;
     $192 = $191&65535;
     HEAP16[$170>>1] = $192;
     $curr$2 = $curr$0;$out_x$1 = $176;$out_y$1 = $180;
    } else {
     $193 = $170;
     $194 = $curr$0;
     $195 = (($193) - ($194))|0;
     $196 = (($195|0) / 40)&-1;
     $197 = ((($curr$0)) + 24|0);
     HEAP32[$197>>2] = $196;
     $198 = (0 - ($196))|0;
     $199 = ((($170)) + 28|0);
     HEAP32[$199>>2] = $198;
     $200 = ($180|0)<($176|0);
     $201 = ($180|0)>=($182|0);
     if ($200) {
      $dx$$i = $201 ? $176 : $185;
      $dy$dx$i = $201 ? $180 : $176;
      $$1$i = $201 ? 1 : -2;
      $dir$0$i = $$1$i;$ll$0$i = $dx$$i;$ss$0$i = $dy$dx$i;
     } else {
      $dy$$i = $201 ? $180 : $182;
      $dx$dy$i = $201 ? $176 : $180;
      $$$i = $201 ? 2 : -1;
      $dir$0$i = $$$i;$ll$0$i = $dy$$i;$ss$0$i = $dx$dy$i;
     }
     $202 = ($ss$0$i|0)<(0);
     $203 = (0 - ($ss$0$i))|0;
     $204 = $202 ? $203 : $ss$0$i;
     $205 = ($204*14)|0;
     $206 = ($ll$0$i|0)>($205|0);
     $207 = $dir$0$i&255;
     $208 = $206 ? $207 : 4;
     $209 = ((($curr$0)) + 3|0);
     HEAP8[$209>>0] = $208;
     $curr$1$in12 = ((($curr$0)) + 32|0);
     $curr$113 = HEAP32[$curr$1$in12>>2]|0;
     $210 = ($curr$113|0)==($170|0);
     if (!($210)) {
      $curr$114 = $curr$113;
      while(1) {
       $211 = ((($curr$114)) + 2|0);
       HEAP8[$211>>0] = $208;
       $212 = ((($curr$114)) + 3|0);
       HEAP8[$212>>0] = $208;
       $curr$1$in = ((($curr$114)) + 32|0);
       $curr$1 = HEAP32[$curr$1$in>>2]|0;
       $213 = ($curr$1|0)==($170|0);
       if ($213) {
        break;
       } else {
        $curr$114 = $curr$1;
       }
      }
     }
     $214 = ((($170)) + 2|0);
     HEAP8[$214>>0] = $208;
     $215 = (($166) - ($193))|0;
     $216 = (($215|0) / 40)&-1;
     $217 = ((($170)) + 24|0);
     HEAP32[$217>>2] = $216;
     $218 = (0 - ($216))|0;
     HEAP32[$168>>2] = $218;
     $curr$2 = $170;$out_x$1 = 0;$out_y$1 = 0;
    }
    $219 = ($170|0)==($$pn$lcssa|0);
    if ($219) {
     break;
    } else {
     $$in = $172;$curr$0 = $curr$2;$next$0 = $170;$out_x$0 = $out_x$1;$out_y$0 = $out_y$1;
    }
   }
   $220 = ((($contour2$016)) + 4|0);
   $221 = ($220>>>0)<($146>>>0);
   if ($221) {
    $contour2$016 = $220;
   } else {
    break;
   }
  }
 }
 if ($80) {
  $point$210 = $71;
 } else {
  $315 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($315|0);
 }
 while(1) {
  $222 = HEAP32[$point$210>>2]|0;
  $223 = $222 & -65520;
  $224 = ($223|0)==(67371008);
  if ($224) {
   $225 = ((($point$210)) + 24|0);
   $226 = HEAP32[$225>>2]|0;
   $227 = ((($point$210)) + 28|0);
   $228 = HEAP32[$227>>2]|0;
   $229 = ((($point$210)) + 12|0);
   $230 = HEAP32[$229>>2]|0;
   $sext48 = $230 << 16;
   $231 = $sext48 >> 16;
   $232 = (((($point$210) + (($228*40)|0)|0)) + 12|0);
   $233 = HEAP16[$232>>1]|0;
   $234 = $233 << 16 >> 16;
   $235 = (($231) - ($234))|0;
   $236 = $230 >> 16;
   $237 = (((($point$210) + (($226*40)|0)|0)) + 12|0);
   $238 = HEAP16[$237>>1]|0;
   $239 = $238 << 16 >> 16;
   $240 = (($239) - ($231))|0;
   $241 = $240 ^ $235;
   $242 = ($241|0)>(-1);
   if ($242) {
    $243 = (((($point$210) + (($226*40)|0)|0)) + 14|0);
    $244 = HEAP16[$243>>1]|0;
    $245 = $244 << 16 >> 16;
    $246 = (($245) - ($236))|0;
    $247 = (((($point$210) + (($228*40)|0)|0)) + 14|0);
    $248 = HEAP16[$247>>1]|0;
    $249 = $248 << 16 >> 16;
    $250 = (($236) - ($249))|0;
    $251 = $250 ^ $246;
    $252 = ($251|0)>(-1);
    if ($252) {
     $253 = $222 | 16;
     $254 = $253&65535;
     HEAP16[$point$210>>1] = $254;
     $$idx57 = (($226) - ($228))|0;
     $255 = ($$idx57*40)|0;
     $256 = (($255|0) / 40)&-1;
     $257 = (((($point$210) + (($228*40)|0)|0)) + 24|0);
     HEAP32[$257>>2] = $256;
     $258 = (0 - ($256))|0;
     $259 = (((($point$210) + (($226*40)|0)|0)) + 28|0);
     HEAP32[$259>>2] = $258;
    }
   }
  }
  $260 = ((($point$210)) + 40|0);
  $261 = ($260>>>0)<($74>>>0);
  if ($261) {
   $point$210 = $260;
  } else {
   break;
  }
 }
 if ($80) {
  $point$39 = $71;
 } else {
  $315 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($315|0);
 }
 while(1) {
  $262 = HEAP32[$point$39>>2]|0;
  $263 = $262&65535;
  $264 = $262 & 16;
  $265 = ($264|0)==(0);
  $266 = $262 >>> 24;
  $267 = $266&255;
  $268 = $262 >>> 16;
  $269 = $268&255;
  L88: do {
   if ($265) {
    $270 = $262 & 3;
    $271 = ($270|0)==(0);
    do {
     if ($271) {
      $276 = ($267<<24>>24)==($269<<24>>24);
      if (!($276)) {
       $309 = $262 >> 24;
       $sext45 = $268 << 24;
       $310 = $sext45 >> 24;
       $311 = (0 - ($309))|0;
       $312 = ($310|0)==($311|0);
       if ($312) {
        $273 = $263;
        break;
       } else {
        break L88;
       }
      }
      $277 = ($267<<24>>24)==(4);
      if ($277) {
       $278 = ((($point$39)) + 24|0);
       $279 = HEAP32[$278>>2]|0;
       $280 = ((($point$39)) + 28|0);
       $281 = HEAP32[$280>>2]|0;
       $282 = ((($point$39)) + 12|0);
       $283 = HEAP32[$282>>2]|0;
       $sext46 = $283 << 16;
       $284 = $sext46 >> 16;
       $285 = (((($point$39) + (($281*40)|0)|0)) + 12|0);
       $286 = HEAP16[$285>>1]|0;
       $287 = $286 << 16 >> 16;
       $288 = (($284) - ($287))|0;
       $289 = $283 >> 16;
       $290 = (((($point$39) + (($281*40)|0)|0)) + 14|0);
       $291 = HEAP16[$290>>1]|0;
       $292 = $291 << 16 >> 16;
       $293 = (($289) - ($292))|0;
       $294 = (((($point$39) + (($279*40)|0)|0)) + 12|0);
       $295 = HEAP16[$294>>1]|0;
       $296 = $295 << 16 >> 16;
       $297 = (($296) - ($284))|0;
       $298 = (((($point$39) + (($279*40)|0)|0)) + 14|0);
       $299 = HEAP16[$298>>1]|0;
       $300 = $299 << 16 >> 16;
       $301 = (($300) - ($289))|0;
       $302 = (_ft_corner_is_flat($288,$293,$297,$301)|0);
       $303 = ($302|0)==(0);
       if ($303) {
        break L88;
       }
       $$idx24 = (($279) - ($281))|0;
       $304 = ($$idx24*40)|0;
       $305 = (($304|0) / 40)&-1;
       $306 = (((($point$39) + (($281*40)|0)|0)) + 24|0);
       HEAP32[$306>>2] = $305;
       $307 = (0 - ($305))|0;
       $308 = (((($point$39) + (($279*40)|0)|0)) + 28|0);
       HEAP32[$308>>2] = $307;
       $$pre39 = HEAP16[$point$39>>1]|0;
       $273 = $$pre39;
      } else {
       $273 = $263;
      }
     } else {
      $273 = $263;
     }
    } while(0);
    $272 = $273&65535;
    $274 = $272 | 16;
    $275 = $274&65535;
    HEAP16[$point$39>>1] = $275;
   }
  } while(0);
  $313 = ((($point$39)) + 40|0);
  $314 = ($313>>>0)<($74>>>0);
  if ($314) {
   $point$39 = $313;
  } else {
   break;
  }
 }
 $315 = HEAP32[$error>>2]|0;
 STACKTOP = sp;return ($315|0);
}
function _af_latin_metrics_init($metrics,$face) {
 $metrics = $metrics|0;
 $face = $face|0;
 var $$$1$i = 0, $$$i = 0, $$best_contour_last$0$i = 0, $$lcssa = 0, $$lcssa127 = 0, $$lcssa132 = 0, $$lcssa134 = 0, $$lcssa135 = 0, $$phi$trans$insert$i = 0, $$phi$trans$insert$i$i = 0, $$phi$trans$insert$i12$i = 0, $$phi$trans$insert147$i = 0, $$phi$trans$insert155$i = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i13$i = 0, $$pre$phi$iZ2D = 0, $$pre148$i = 0, $$pre151$i = 0, $$pre153$i = 0;
 var $$pre156$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0;
 var $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0;
 var $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0;
 var $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0;
 var $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0;
 var $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0;
 var $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0;
 var $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0;
 var $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0;
 var $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0;
 var $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0;
 var $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0;
 var $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0;
 var $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0;
 var $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $advance$i = 0, $best_contour_first$0$first$0$i = 0, $best_contour_first$080$i = 0, $best_contour_first$2$i = 0, $best_contour_first$2$i$lcssa = 0, $best_contour_last$0$$i = 0, $best_contour_last$079$i = 0, $best_contour_last$2$i = 0, $best_contour_last$2$i$lcssa = 0, $best_on_point_first$1$i = 0, $best_on_point_first$11$i = 0, $best_on_point_first$2$i = 0, $best_on_point_first$3$i = 0, $best_on_point_first$3$ph$i = 0, $best_on_point_first$4$i = 0, $best_on_point_first$5$i = 0, $best_on_point_last$1$i = 0, $best_on_point_last$11$i = 0, $best_on_point_last$2$i = 0;
 var $best_on_point_last$3$i = 0, $best_on_point_last$3$ph$i = 0, $best_on_point_last$4$i = 0, $best_on_point_last$5$i = 0, $best_point$0$$i = 0, $best_point$081$i = 0, $best_point$167$i = 0, $best_point$2$i = 0, $best_point$372$i = 0, $best_point$4$i = 0, $best_point$5$i = 0, $best_point$6$i = 0, $best_point$6$i$lcssa = 0, $best_segment_first$1$ph$i = 0, $best_segment_first$7$i = 0, $best_segment_last$1$i = 0, $best_segment_last$7$i = 0, $best_y$082$i = 0, $best_y$12$i = 0, $best_y$12$i$lcssa = 0;
 var $best_y$14$ph$i = 0, $best_y$168$i = 0, $best_y$2$i = 0, $best_y$373$i = 0, $best_y$4$i = 0, $best_y$5$i = 0, $best_y$6$i = 0, $best_y$6$i$lcssa = 0, $bs$0115$i = 0, $ch$063$i = 0, $ch$063$ph$i = 0, $ch$1$i = 0, $dummy$i = 0, $exitcond$i = 0, $exitcond$i$i = 0, $exitcond$i17$i = 0, $first$078$i = 0, $first1$0$i = 0, $first1$1$i = 0, $first1$1$i$lcssa = 0;
 var $flats$i = 0, $glyph_index$3$i = 0, $hints$i = 0, $hit$0$i = 0, $hit$1$i = 0, $hit$2$i = 0, $i$03$i$i = 0, $i$03$i11$i = 0, $i$05$i = 0, $j$01$i$i = 0, $j$01$i15$i = 0, $last2$0$$i = 0, $last2$0$i = 0, $last2$1$i = 0, $last2$1$i$lcssa = 0, $last2$2$i = 0, $last2$3$i = 0, $last2$3$i$lcssa = 0, $len$062$i = 0, $len$062$ph$i = 0;
 var $next$0$i = 0, $next$1$i = 0, $nn$077$i = 0, $num_flats$0$ph110$i = 0, $num_flats$0$ph45$lcssa61$i = 0, $num_flats$0$ph4599$i = 0, $num_flats$0$ph4599$i$lcssa = 0, $num_rounds$0$ph$lcssa$i = 0, $num_rounds$0$ph109$i = 0, $num_widths$i = 0, $old_advance$07$i = 0, $old_advance$2$ph$i = 0, $or$cond$1$i = 0, $or$cond$i = 0, $or$cond$i3 = 0, $or$cond189$i = 0, $or$cond190$i = 0, $or$cond9$i = 0, $p$0$ph108$i = 0, $p$0$ph4498$i = 0;
 var $p$089$i = 0, $p$164$i = 0, $p$2$i = 0, $p$2$i$lcssa = 0, $p$2$i$lcssa139 = 0, $p_first$0$i = 0, $p_first$2$i = 0, $p_first$3$i = 0, $p_first$3$i$lcssa = 0, $p_first$4$i = 0, $p_first$4$i$lcssa = 0, $p_first$5$i = 0, $p_first$7$i = 0, $p_last$0$i = 0, $p_last$2$i = 0, $p_last$3$i = 0, $p_last$3$i$lcssa = 0, $p_last$4$i = 0, $p_last$4$i$lcssa = 0, $p_last$7$i = 0;
 var $pp$065$i = 0, $pp$170$i = 0, $prev$0$i = 0, $prev$1$i = 0, $prev$2$i = 0, $prev$3$i = 0, $rounds$i = 0, $same_width$2$i = 0, $scevgep146$i = 0, $seg$08$1$i = 0, $seg$08$i = 0, $sext$i = 0, $started$06$i = 0, $started$2$ph$i = 0, $tmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 10480|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $advance$i = sp + 10472|0;
 $flats$i = sp + 10268|0;
 $rounds$i = sp + 10064|0;
 $hints$i = sp + 3192|0;
 $dummy$i = sp + 4|0;
 $num_widths$i = sp;
 $0 = ((($face)) + 92|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 68|0);
 $3 = HEAP16[$2>>1]|0;
 $4 = $3&65535;
 $5 = ((($metrics)) + 40|0);
 HEAP32[$5>>2] = $4;
 $6 = (_FT_Select_Charmap($face,1970170211)|0);
 $7 = ($6|0)==(0);
 if (!($7)) {
  (_FT_Set_Charmap($face,$1)|0);
  STACKTOP = sp;return 0;
 }
 $8 = ((($face)) + 100|0);
 $9 = HEAP32[$8>>2]|0;
 _memset(($hints$i|0),0,3000)|0;
 HEAP32[$hints$i>>2] = $9;
 $10 = ((($metrics)) + 52|0);
 HEAP32[$10>>2] = 0;
 $11 = ((($metrics)) + 1624|0);
 HEAP32[$11>>2] = 0;
 $12 = HEAP32[$metrics>>2]|0;
 $13 = ((($12)) + 8|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (61928 + ($14<<2)|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($16)) + 8|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($metrics)) + 36|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = HEAP32[$20>>2]|0;
 $22 = (_FT_Get_Char_Index($21,$18)|0);
 $23 = ($22|0)==(0);
 if ($23) {
  $24 = ((($16)) + 12|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = ($25|0)==(0);
  if (!($26)) {
   $27 = HEAP32[$19>>2]|0;
   $28 = HEAP32[$27>>2]|0;
   $29 = (_FT_Get_Char_Index($28,$25)|0);
   $30 = ($29|0)==(0);
   if ($30) {
    $31 = ((($16)) + 16|0);
    $32 = HEAP32[$31>>2]|0;
    $33 = ($32|0)==(0);
    if (!($33)) {
     $34 = HEAP32[$19>>2]|0;
     $35 = HEAP32[$34>>2]|0;
     $36 = (_FT_Get_Char_Index($35,$32)|0);
     $37 = ($36|0)==(0);
     if (!($37)) {
      $glyph_index$3$i = $36;
      label = 7;
     }
    }
   } else {
    $glyph_index$3$i = $29;
    label = 7;
   }
  }
 } else {
  $glyph_index$3$i = $22;
  label = 7;
 }
 do {
  if ((label|0) == 7) {
   $38 = (_FT_Load_Glyph($face,$glyph_index$3$i,1)|0);
   $39 = ($38|0)==(0);
   if ($39) {
    $40 = ((($face)) + 84|0);
    $41 = HEAP32[$40>>2]|0;
    $42 = ((($41)) + 110|0);
    $43 = HEAP16[$42>>1]|0;
    $44 = ($43<<16>>16)<(1);
    if (!($44)) {
     _memset(($dummy$i|0),0,3188)|0;
     $45 = HEAP32[$5>>2]|0;
     $46 = ((($dummy$i)) + 40|0);
     HEAP32[$46>>2] = $45;
     $47 = ((($dummy$i)) + 8|0);
     HEAP32[$47>>2] = 65536;
     $48 = ((($dummy$i)) + 12|0);
     HEAP32[$48>>2] = 65536;
     $49 = ((($dummy$i)) + 16|0);
     HEAP32[$49>>2] = 0;
     $50 = ((($dummy$i)) + 20|0);
     HEAP32[$50>>2] = 0;
     $51 = ((($dummy$i)) + 4|0);
     HEAP32[$51>>2] = $face;
     $52 = ((($dummy$i)) + 24|0);
     HEAP32[$52>>2] = 0;
     $53 = ((($dummy$i)) + 28|0);
     HEAP32[$53>>2] = 0;
     $54 = ((($hints$i)) + 2988|0);
     HEAP32[$54>>2] = $dummy$i;
     $55 = ((($hints$i)) + 2980|0);
     HEAP32[$55>>2] = 0;
     $56 = ((($41)) + 108|0);
     $57 = (_af_glyph_hints_reload($hints$i,$56)|0);
     $58 = ($57|0)==(0);
     if ($58) {
      HEAP32[$num_widths$i>>2] = 0;
      $59 = (_af_latin_hints_compute_segments($hints$i,0)|0);
      $60 = ($59|0)==(0);
      if ($60) {
       _af_latin_hints_link_segments($hints$i,0,0,0);
       $61 = ((($hints$i)) + 52|0);
       $62 = HEAP32[$61>>2]|0;
       $63 = ((($hints$i)) + 44|0);
       $64 = HEAP32[$63>>2]|0;
       $65 = (($62) + (($64*48)|0)|0);
       $66 = ($64|0)>(0);
       if ($66) {
        $82 = 0;$seg$08$i = $62;
        while(1) {
         $67 = ((($seg$08$i)) + 20|0);
         $68 = HEAP32[$67>>2]|0;
         $69 = ($68|0)==(0|0);
         if ($69) {
          $476 = $82;
         } else {
          $70 = ((($68)) + 20|0);
          $71 = HEAP32[$70>>2]|0;
          $72 = ($71|0)==($seg$08$i|0);
          $73 = ($68>>>0)>($seg$08$i>>>0);
          $or$cond$i = $73 & $72;
          if ($or$cond$i) {
           $74 = ((($seg$08$i)) + 2|0);
           $75 = HEAP16[$74>>1]|0;
           $76 = $75 << 16 >> 16;
           $77 = ((($68)) + 2|0);
           $78 = HEAP16[$77>>1]|0;
           $79 = $78 << 16 >> 16;
           $80 = (($76) - ($79))|0;
           $81 = ($82>>>0)<(16);
           if ($81) {
            $83 = ($80|0)<(0);
            $84 = (0 - ($80))|0;
            $$$i = $83 ? $84 : $80;
            $85 = (($82) + 1)|0;
            HEAP32[$num_widths$i>>2] = $85;
            $86 = (((($metrics)) + 56|0) + (($82*12)|0)|0);
            HEAP32[$86>>2] = $$$i;
            $476 = $85;
           } else {
            $476 = $82;
           }
          } else {
           $476 = $82;
          }
         }
         $87 = ((($seg$08$i)) + 48|0);
         $88 = ($87>>>0)<($65>>>0);
         if ($88) {
          $82 = $476;$seg$08$i = $87;
         } else {
          break;
         }
        }
       }
       $89 = ((($metrics)) + 56|0);
       $90 = HEAP32[$46>>2]|0;
       $91 = (($90>>>0) / 100)&-1;
       _af_sort_and_quantize_widths($num_widths$i,$89,$91);
       $92 = HEAP32[$num_widths$i>>2]|0;
       HEAP32[$10>>2] = $92;
       HEAP32[$num_widths$i>>2] = 0;
       $93 = (_af_latin_hints_compute_segments($hints$i,1)|0);
       $94 = ($93|0)==(0);
       if ($94) {
        _af_latin_hints_link_segments($hints$i,0,0,1);
        $114 = ((($hints$i)) + 1520|0);
        $115 = HEAP32[$114>>2]|0;
        $116 = ((($hints$i)) + 1512|0);
        $117 = HEAP32[$116>>2]|0;
        $118 = (($115) + (($117*48)|0)|0);
        $119 = ($117|0)>(0);
        if ($119) {
         $135 = 0;$seg$08$1$i = $115;
         while(1) {
          $120 = ((($seg$08$1$i)) + 20|0);
          $121 = HEAP32[$120>>2]|0;
          $122 = ($121|0)==(0|0);
          if ($122) {
           $477 = $135;
          } else {
           $123 = ((($121)) + 20|0);
           $124 = HEAP32[$123>>2]|0;
           $125 = ($124|0)==($seg$08$1$i|0);
           $126 = ($121>>>0)>($seg$08$1$i>>>0);
           $or$cond$1$i = $126 & $125;
           if ($or$cond$1$i) {
            $127 = ((($seg$08$1$i)) + 2|0);
            $128 = HEAP16[$127>>1]|0;
            $129 = $128 << 16 >> 16;
            $130 = ((($121)) + 2|0);
            $131 = HEAP16[$130>>1]|0;
            $132 = $131 << 16 >> 16;
            $133 = (($129) - ($132))|0;
            $134 = ($135>>>0)<(16);
            if ($134) {
             $136 = ($133|0)<(0);
             $137 = (0 - ($133))|0;
             $$$1$i = $136 ? $137 : $133;
             $138 = (($135) + 1)|0;
             HEAP32[$num_widths$i>>2] = $138;
             $139 = (((($metrics)) + 1628|0) + (($135*12)|0)|0);
             HEAP32[$139>>2] = $$$1$i;
             $477 = $138;
            } else {
             $477 = $135;
            }
           } else {
            $477 = $135;
           }
          }
          $140 = ((($seg$08$1$i)) + 48|0);
          $141 = ($140>>>0)<($118>>>0);
          if ($141) {
           $135 = $477;$seg$08$1$i = $140;
          } else {
           break;
          }
         }
        }
        $142 = ((($metrics)) + 1628|0);
        $143 = HEAP32[$46>>2]|0;
        $144 = (($143>>>0) / 100)&-1;
        _af_sort_and_quantize_widths($num_widths$i,$142,$144);
        $145 = HEAP32[$num_widths$i>>2]|0;
        HEAP32[$11>>2] = $145;
        break;
       }
      }
     }
    }
   }
  }
 } while(0);
 $95 = HEAP32[$10>>2]|0;
 $96 = ($95|0)==(0);
 if ($96) {
  $99 = HEAP32[$5>>2]|0;
  $100 = ($99*50)|0;
  $101 = (($100|0) / 2048)&-1;
  $103 = $101;
 } else {
  $97 = ((($metrics)) + 56|0);
  $98 = HEAP32[$97>>2]|0;
  $103 = $98;
 }
 $102 = (($103|0) / 5)&-1;
 $104 = ((($metrics)) + 248|0);
 HEAP32[$104>>2] = $102;
 $105 = ((($metrics)) + 252|0);
 HEAP32[$105>>2] = $103;
 $106 = ((($metrics)) + 256|0);
 HEAP8[$106>>0] = 0;
 $107 = HEAP32[$11>>2]|0;
 $108 = ($107|0)==(0);
 if ($108) {
  $111 = HEAP32[$5>>2]|0;
  $112 = ($111*50)|0;
  $113 = (($112|0) / 2048)&-1;
  $147 = $113;
 } else {
  $109 = ((($metrics)) + 1628|0);
  $110 = HEAP32[$109>>2]|0;
  $147 = $110;
 }
 $146 = (($147|0) / 5)&-1;
 $148 = ((($metrics)) + 1820|0);
 HEAP32[$148>>2] = $146;
 $149 = ((($metrics)) + 1824|0);
 HEAP32[$149>>2] = $147;
 $150 = ((($metrics)) + 1828|0);
 HEAP8[$150>>0] = 0;
 _af_glyph_hints_done($hints$i);
 $151 = HEAP32[$metrics>>2]|0;
 $152 = ((($151)) + 12|0);
 $153 = HEAP32[$152>>2]|0;
 $154 = (59204 + ($153<<3)|0);
 $155 = HEAP32[$154>>2]|0;
 $156 = ($155|0)==(759);
 if (!($156)) {
  $157 = ((($metrics)) + 1832|0);
  $158 = ((($face)) + 84|0);
  $160 = $155;$bs$0115$i = $154;
  while(1) {
   $159 = (594743 + ($160)|0);
   $161 = HEAP8[$159>>0]|0;
   $162 = ($161<<24>>24)==(0);
   L49: do {
    if ($162) {
     $num_flats$0$ph45$lcssa61$i = 0;$num_rounds$0$ph$lcssa$i = 0;
    } else {
     $163 = ((($bs$0115$i)) + 4|0);
     $478 = $161;$num_flats$0$ph110$i = 0;$num_rounds$0$ph109$i = 0;$p$0$ph108$i = $159;
     while(1) {
      $479 = $478;$num_flats$0$ph4599$i = $num_flats$0$ph110$i;$p$0$ph4498$i = $p$0$ph108$i;
      L53: while(1) {
       $166 = $479;$p$089$i = $p$0$ph4498$i;
       L55: while(1) {
        $164 = ((($p$089$i)) + 1|0);
        $165 = $166&255;
        $167 = ($166<<24>>24)<(0);
        if ($167) {
         $168 = ($166&255)<(224);
         do {
          if ($168) {
           $169 = $165 & 31;
           $ch$063$ph$i = $169;$len$062$ph$i = 1;
          } else {
           $170 = ($166&255)<(240);
           if ($170) {
            $171 = $165 & 15;
            $ch$063$ph$i = $171;$len$062$ph$i = 2;
            break;
           } else {
            $172 = $165 & 7;
            $ch$063$ph$i = $172;$len$062$ph$i = 3;
            break;
           }
          }
         } while(0);
         $ch$063$i = $ch$063$ph$i;$len$062$i = $len$062$ph$i;$p$164$i = $164;
         while(1) {
          $173 = $ch$063$i << 6;
          $174 = ((($p$164$i)) + 1|0);
          $175 = HEAP8[$p$164$i>>0]|0;
          $176 = $175&255;
          $177 = $176 & 63;
          $178 = $177 | $173;
          $179 = (($len$062$i) + -1)|0;
          $180 = ($179|0)==(0);
          if ($180) {
           $$lcssa = $178;
           break;
          } else {
           $ch$063$i = $178;$len$062$i = $179;$p$164$i = $174;
          }
         }
         $scevgep146$i = (($164) + ($len$062$ph$i)|0);
         $ch$1$i = $$lcssa;$p$2$i = $scevgep146$i;
        } else {
         $ch$1$i = $165;$p$2$i = $164;
        }
        $181 = HEAP32[$19>>2]|0;
        $182 = HEAP32[$181>>2]|0;
        $183 = (_FT_Get_Char_Index($182,$ch$1$i)|0);
        $184 = ($183|0)==(0);
        L70: do {
         if (!($184)) {
          $185 = (_FT_Load_Glyph($face,$183,1)|0);
          $186 = HEAP32[$158>>2]|0;
          $187 = ((($186)) + 108|0);
          $188 = HEAP32[$187>>2]|0;
          $189 = $188&65535;
          $190 = ((($186)) + 112|0);
          $191 = HEAP32[$190>>2]|0;
          $192 = ((($186)) + 116|0);
          $193 = HEAP32[$192>>2]|0;
          $194 = ((($186)) + 120|0);
          $195 = HEAP32[$194>>2]|0;
          $196 = ($185|0)==(0);
          if ($196) {
           $197 = $188 >>> 16;
           $198 = $197&65535;
           $199 = ($198<<16>>16)<(3);
           if (!($199)) {
            $202 = ($189<<16>>16)>(0);
            if (!($202)) {
             $best_y$14$ph$i = 0;$p$2$i$lcssa = $p$2$i;
             break L55;
            }
            $sext$i = $188 << 16;
            $203 = $sext$i >> 16;
            $best_contour_first$080$i = 0;$best_contour_last$079$i = 0;$best_point$081$i = -1;$best_y$082$i = 0;$first$078$i = 0;$nn$077$i = 0;
            while(1) {
             $204 = (($195) + ($nn$077$i<<1)|0);
             $205 = HEAP16[$204>>1]|0;
             $206 = $205 << 16 >> 16;
             $207 = ($206|0)>($first$078$i|0);
             if ($207) {
              $208 = HEAP16[$163>>1]|0;
              $209 = $208 & 1;
              $210 = ($209<<16>>16)==(0);
              $211 = ($first$078$i|0)>($206|0);
              do {
               if ($210) {
                if ($211) {
                 $best_point$5$i = $best_point$081$i;$best_y$5$i = $best_y$082$i;
                 break;
                } else {
                 $best_point$372$i = $best_point$081$i;$best_y$373$i = $best_y$082$i;$pp$170$i = $first$078$i;
                }
                while(1) {
                 $216 = ($best_point$372$i|0)<(0);
                 $$phi$trans$insert147$i = (((($191) + ($pp$170$i<<3)|0)) + 4|0);
                 $$pre148$i = HEAP32[$$phi$trans$insert147$i>>2]|0;
                 $217 = ($$pre148$i|0)<($best_y$373$i|0);
                 $or$cond190$i = $216 | $217;
                 $best_point$4$i = $or$cond190$i ? $pp$170$i : $best_point$372$i;
                 $best_y$4$i = $or$cond190$i ? $$pre148$i : $best_y$373$i;
                 $218 = (($pp$170$i) + 1)|0;
                 $219 = ($pp$170$i|0)<($206|0);
                 if ($219) {
                  $best_point$372$i = $best_point$4$i;$best_y$373$i = $best_y$4$i;$pp$170$i = $218;
                 } else {
                  $best_point$5$i = $best_point$4$i;$best_y$5$i = $best_y$4$i;
                  break;
                 }
                }
               } else {
                if ($211) {
                 $best_point$5$i = $best_point$081$i;$best_y$5$i = $best_y$082$i;
                 break;
                } else {
                 $best_point$167$i = $best_point$081$i;$best_y$168$i = $best_y$082$i;$pp$065$i = $first$078$i;
                }
                while(1) {
                 $212 = ($best_point$167$i|0)<(0);
                 $$phi$trans$insert$i = (((($191) + ($pp$065$i<<3)|0)) + 4|0);
                 $$pre$i = HEAP32[$$phi$trans$insert$i>>2]|0;
                 $213 = ($$pre$i|0)>($best_y$168$i|0);
                 $or$cond189$i = $212 | $213;
                 $best_point$2$i = $or$cond189$i ? $pp$065$i : $best_point$167$i;
                 $best_y$2$i = $or$cond189$i ? $$pre$i : $best_y$168$i;
                 $214 = (($pp$065$i) + 1)|0;
                 $215 = ($pp$065$i|0)<($206|0);
                 if ($215) {
                  $best_point$167$i = $best_point$2$i;$best_y$168$i = $best_y$2$i;$pp$065$i = $214;
                 } else {
                  $best_point$5$i = $best_point$2$i;$best_y$5$i = $best_y$2$i;
                  break;
                 }
                }
               }
              } while(0);
              $220 = ($best_point$5$i|0)==($best_point$081$i|0);
              $best_contour_last$0$$i = $220 ? $best_contour_last$079$i : $206;
              $best_contour_first$0$first$0$i = $220 ? $best_contour_first$080$i : $first$078$i;
              $best_contour_first$2$i = $best_contour_first$0$first$0$i;$best_contour_last$2$i = $best_contour_last$0$$i;$best_point$6$i = $best_point$5$i;$best_y$6$i = $best_y$5$i;
             } else {
              $best_contour_first$2$i = $best_contour_first$080$i;$best_contour_last$2$i = $best_contour_last$079$i;$best_point$6$i = $best_point$081$i;$best_y$6$i = $best_y$082$i;
             }
             $221 = (($206) + 1)|0;
             $222 = (($nn$077$i) + 1)|0;
             $exitcond$i = ($222|0)==($203|0);
             if ($exitcond$i) {
              $best_contour_first$2$i$lcssa = $best_contour_first$2$i;$best_contour_last$2$i$lcssa = $best_contour_last$2$i;$best_point$6$i$lcssa = $best_point$6$i;$best_y$6$i$lcssa = $best_y$6$i;
              break;
             } else {
              $best_contour_first$080$i = $best_contour_first$2$i;$best_contour_last$079$i = $best_contour_last$2$i;$best_point$081$i = $best_point$6$i;$best_y$082$i = $best_y$6$i;$first$078$i = $221;$nn$077$i = $222;
             }
            }
            $223 = ($best_point$6$i$lcssa|0)>(-1);
            if (!($223)) {
             $best_y$14$ph$i = $best_y$6$i$lcssa;$p$2$i$lcssa = $p$2$i;
             break L55;
            }
            $224 = (($191) + ($best_point$6$i$lcssa<<3)|0);
            $225 = HEAP32[$224>>2]|0;
            $226 = (($193) + ($best_point$6$i$lcssa)|0);
            $227 = HEAP8[$226>>0]|0;
            $228 = $227 & 3;
            $229 = ($228<<24>>24)==(1);
            $best_point$0$$i = $229 ? $best_point$6$i$lcssa : -1;
            $best_on_point_first$1$i = $best_point$0$$i;$best_on_point_last$1$i = $best_point$0$$i;$prev$0$i = $best_point$6$i$lcssa;
            while(1) {
             $230 = ($prev$0$i|0)>($best_contour_first$2$i$lcssa|0);
             $231 = (($prev$0$i) + -1)|0;
             $prev$1$i = $230 ? $231 : $best_contour_last$2$i$lcssa;
             $232 = (((($191) + ($prev$1$i<<3)|0)) + 4|0);
             $233 = HEAP32[$232>>2]|0;
             $234 = (($233) - ($best_y$6$i$lcssa))|0;
             $235 = ($234|0)<(0);
             $236 = (0 - ($234))|0;
             $237 = $235 ? $236 : $234;
             $238 = ($237|0)>(5);
             if ($238) {
              $239 = (($191) + ($prev$1$i<<3)|0);
              $240 = HEAP32[$239>>2]|0;
              $241 = (($240) - ($225))|0;
              $242 = ($241|0)<(0);
              $243 = (0 - ($241))|0;
              $244 = $242 ? $243 : $241;
              $245 = ($237*20)|0;
              $246 = ($244|0)>($245|0);
              if (!($246)) {
               $best_on_point_first$3$ph$i = $best_on_point_first$1$i;$best_on_point_last$3$ph$i = $best_on_point_last$1$i;$best_segment_first$1$ph$i = $prev$0$i;
               break;
              }
             }
             $247 = (($193) + ($prev$1$i)|0);
             $248 = HEAP8[$247>>0]|0;
             $249 = $248 & 3;
             $250 = ($249<<24>>24)==(1);
             $251 = ($best_on_point_last$1$i|0)<(0);
             $best_on_point_first$2$i = $250 ? $prev$1$i : $best_on_point_first$1$i;
             $252 = $251 & $250;
             $best_on_point_last$2$i = $252 ? $prev$1$i : $best_on_point_last$1$i;
             $253 = ($prev$1$i|0)==($best_point$6$i$lcssa|0);
             if ($253) {
              $best_on_point_first$3$ph$i = $best_on_point_first$2$i;$best_on_point_last$3$ph$i = $best_on_point_last$2$i;$best_segment_first$1$ph$i = $best_point$6$i$lcssa;
              break;
             } else {
              $best_on_point_first$1$i = $best_on_point_first$2$i;$best_on_point_last$1$i = $best_on_point_last$2$i;$prev$0$i = $prev$1$i;
             }
            }
            $best_on_point_first$3$i = $best_on_point_first$3$ph$i;$best_on_point_last$3$i = $best_on_point_last$3$ph$i;$next$0$i = $best_point$6$i$lcssa;
            while(1) {
             $254 = ($next$0$i|0)<($best_contour_last$2$i$lcssa|0);
             $255 = (($next$0$i) + 1)|0;
             $next$1$i = $254 ? $255 : $best_contour_first$2$i$lcssa;
             $256 = (($191) + ($next$1$i<<3)|0);
             $257 = (((($191) + ($next$1$i<<3)|0)) + 4|0);
             $258 = HEAP32[$257>>2]|0;
             $259 = (($258) - ($best_y$6$i$lcssa))|0;
             $260 = ($259|0)<(0);
             $261 = (0 - ($259))|0;
             $262 = $260 ? $261 : $259;
             $263 = ($262|0)>(5);
             if ($263) {
              $264 = HEAP32[$256>>2]|0;
              $265 = (($264) - ($225))|0;
              $266 = ($265|0)<(0);
              $267 = (0 - ($265))|0;
              $268 = $266 ? $267 : $265;
              $269 = ($262*20)|0;
              $270 = ($268|0)>($269|0);
              if (!($270)) {
               $$lcssa127 = $256;$best_on_point_first$5$i = $best_on_point_first$3$i;$best_on_point_last$5$i = $best_on_point_last$3$i;$best_segment_last$1$i = $next$0$i;
               break;
              }
             }
             $271 = (($193) + ($next$1$i)|0);
             $272 = HEAP8[$271>>0]|0;
             $273 = $272 & 3;
             $274 = ($273<<24>>24)==(1);
             $275 = ($best_on_point_first$3$i|0)<(0);
             $276 = $275 & $274;
             $best_on_point_first$4$i = $276 ? $next$1$i : $best_on_point_first$3$i;
             $best_on_point_last$4$i = $274 ? $next$1$i : $best_on_point_last$3$i;
             $277 = ($next$1$i|0)==($best_point$6$i$lcssa|0);
             if ($277) {
              $$lcssa127 = $256;$best_on_point_first$5$i = $best_on_point_first$4$i;$best_on_point_last$5$i = $best_on_point_last$4$i;$best_segment_last$1$i = $best_point$6$i$lcssa;
              break;
             } else {
              $best_on_point_first$3$i = $best_on_point_first$4$i;$best_on_point_last$3$i = $best_on_point_last$4$i;$next$0$i = $next$1$i;
             }
            }
            $278 = HEAP16[$163>>1]|0;
            $279 = $278 & 8;
            $280 = ($279<<16>>16)==(0);
            L102: do {
             if ($280) {
              $best_on_point_first$11$i = $best_on_point_first$5$i;$best_on_point_last$11$i = $best_on_point_last$5$i;$best_segment_first$7$i = $best_segment_first$1$ph$i;$best_segment_last$7$i = $best_segment_last$1$i;$best_y$12$i = $best_y$6$i$lcssa;
             } else {
              $281 = HEAP32[$5>>2]|0;
              $282 = (($281>>>0) / 25)&-1;
              $283 = (($191) + ($best_segment_last$1$i<<3)|0);
              $284 = HEAP32[$283>>2]|0;
              $285 = (($191) + ($best_segment_first$1$ph$i<<3)|0);
              $286 = HEAP32[$285>>2]|0;
              $287 = (($284) - ($286))|0;
              $288 = ($287|0)<(0);
              $289 = (0 - ($287))|0;
              $290 = $288 ? $289 : $287;
              $291 = ($290|0)<($282|0);
              if (!($291)) {
               $best_on_point_first$11$i = $best_on_point_first$5$i;$best_on_point_last$11$i = $best_on_point_last$5$i;$best_segment_first$7$i = $best_segment_first$1$ph$i;$best_segment_last$7$i = $best_segment_last$1$i;$best_y$12$i = $best_y$6$i$lcssa;
               break;
              }
              $292 = (2 - ($best_segment_first$1$ph$i))|0;
              $293 = (($292) + ($best_segment_last$1$i))|0;
              $294 = (($best_contour_last$2$i$lcssa) - ($best_contour_first$2$i$lcssa))|0;
              $295 = ($293|0)>($294|0);
              if ($295) {
               $best_on_point_first$11$i = $best_on_point_first$5$i;$best_on_point_last$11$i = $best_on_point_last$5$i;$best_segment_first$7$i = $best_segment_first$1$ph$i;$best_segment_last$7$i = $best_segment_last$1$i;$best_y$12$i = $best_y$6$i$lcssa;
               break;
              } else {
               $prev$2$i = $best_point$6$i$lcssa;
              }
              while(1) {
               $296 = ($prev$2$i|0)>($best_contour_first$2$i$lcssa|0);
               $297 = (($prev$2$i) + -1)|0;
               $prev$3$i = $296 ? $297 : $best_contour_last$2$i$lcssa;
               $298 = (($191) + ($prev$3$i<<3)|0);
               $299 = HEAP32[$298>>2]|0;
               $300 = ($299|0)!=($225|0);
               $301 = ($prev$3$i|0)==($best_point$6$i$lcssa|0);
               $or$cond$i3 = $300 | $301;
               if ($or$cond$i3) {
                break;
               } else {
                $prev$2$i = $prev$3$i;
               }
              }
              $302 = $281 >>> 2;
              if ($301) {
               break L70;
              }
              $303 = HEAP32[$224>>2]|0;
              $304 = ($299|0)<($303|0);
              $first1$0$i = $best_segment_last$1$i;$hit$0$i = 0;$last2$0$i = $best_segment_last$1$i;$p_first$0$i = 0;$p_last$0$i = 0;
              L109: while(1) {
               $305 = ($hit$0$i<<24>>24)==(0);
               if ($305) {
                $306 = (($193) + ($last2$0$i)|0);
                $307 = HEAP8[$306>>0]|0;
                $308 = $307 & 3;
                $309 = ($308<<24>>24)==(1);
                $last2$0$$i = $309 ? $last2$0$i : -1;
                $first1$1$i = $last2$0$i;$hit$1$i = 1;$p_first$2$i = $last2$0$$i;$p_last$2$i = $last2$0$$i;
               } else {
                $first1$1$i = $first1$0$i;$hit$1$i = $hit$0$i;$p_first$2$i = $p_first$0$i;$p_last$2$i = $p_last$0$i;
               }
               $310 = ($last2$0$i|0)<($best_contour_last$2$i$lcssa|0);
               $311 = (($last2$0$i) + 1)|0;
               $last2$1$i = $310 ? $311 : $best_contour_first$2$i$lcssa;
               $312 = (($191) + ($first1$1$i<<3)|0);
               $313 = (((($191) + ($first1$1$i<<3)|0)) + 4|0);
               $314 = HEAP32[$313>>2]|0;
               $315 = (($best_y$6$i$lcssa) - ($314))|0;
               $316 = ($315|0)<(0);
               $317 = (0 - ($315))|0;
               $318 = $316 ? $317 : $315;
               $319 = ($318|0)>($302|0);
               do {
                if ($319) {
                 $hit$2$i = 0;$p_first$7$i = $p_first$2$i;$p_last$7$i = $p_last$2$i;
                } else {
                 $320 = (($191) + ($last2$1$i<<3)|0);
                 $321 = (((($191) + ($last2$1$i<<3)|0)) + 4|0);
                 $322 = HEAP32[$321>>2]|0;
                 $323 = (($322) - ($314))|0;
                 $324 = ($323|0)<(0);
                 $325 = (0 - ($323))|0;
                 $326 = $324 ? $325 : $323;
                 $327 = ($326|0)>(5);
                 if ($327) {
                  $328 = HEAP32[$320>>2]|0;
                  $329 = HEAP32[$312>>2]|0;
                  $330 = (($328) - ($329))|0;
                  $331 = ($330|0)<(0);
                  $332 = (0 - ($330))|0;
                  $333 = $331 ? $332 : $330;
                  $334 = ($326*20)|0;
                  $335 = ($333|0)>($334|0);
                  if ($335) {
                   $343 = $328;$344 = $329;
                  } else {
                   $hit$2$i = 0;$p_first$7$i = $p_first$2$i;$p_last$7$i = $p_last$2$i;
                   break;
                  }
                 } else {
                  $$pre151$i = HEAP32[$312>>2]|0;
                  $$pre153$i = HEAP32[$320>>2]|0;
                  $343 = $$pre153$i;$344 = $$pre151$i;
                 }
                 $336 = (($193) + ($last2$1$i)|0);
                 $337 = HEAP8[$336>>0]|0;
                 $338 = $337 & 3;
                 $339 = ($338<<24>>24)==(1);
                 $340 = ($p_first$2$i|0)<(0);
                 $341 = $340 & $339;
                 $p_first$3$i = $341 ? $last2$1$i : $p_first$2$i;
                 $p_last$3$i = $339 ? $last2$1$i : $p_last$2$i;
                 $342 = ($343|0)>($344|0);
                 $345 = (($343) - ($344))|0;
                 $tmp$i = $304 ^ $342;
                 if ($tmp$i) {
                  $hit$2$i = $hit$1$i;$p_first$7$i = $p_first$3$i;$p_last$7$i = $p_last$3$i;
                  break;
                 }
                 $346 = ($345|0)<(0);
                 $347 = (0 - ($345))|0;
                 $348 = $346 ? $347 : $345;
                 $349 = ($348|0)<($282|0);
                 if ($349) {
                  $hit$2$i = $hit$1$i;$p_first$7$i = $p_first$3$i;$p_last$7$i = $p_last$3$i;
                 } else {
                  $$lcssa132 = $314;$$lcssa134 = $326;$$lcssa135 = $344;$first1$1$i$lcssa = $first1$1$i;$last2$1$i$lcssa = $last2$1$i;$p_first$3$i$lcssa = $p_first$3$i;$p_last$3$i$lcssa = $p_last$3$i;
                  break L109;
                 }
                }
               } while(0);
               $374 = ($last2$1$i|0)==($best_segment_first$1$ph$i|0);
               if ($374) {
                $best_on_point_first$11$i = $best_on_point_first$5$i;$best_on_point_last$11$i = $best_on_point_last$5$i;$best_segment_first$7$i = $best_segment_first$1$ph$i;$best_segment_last$7$i = $best_segment_last$1$i;$best_y$12$i = $best_y$6$i$lcssa;
                break L102;
               } else {
                $first1$0$i = $first1$1$i;$hit$0$i = $hit$2$i;$last2$0$i = $last2$1$i;$p_first$0$i = $p_first$7$i;$p_last$0$i = $p_last$7$i;
               }
              }
              $350 = ($$lcssa134*20)|0;
              $last2$2$i = $last2$1$i$lcssa;$p_first$4$i = $p_first$3$i$lcssa;$p_last$4$i = $p_last$3$i$lcssa;
              while(1) {
               $351 = ($last2$2$i|0)<($best_contour_last$2$i$lcssa|0);
               $352 = (($last2$2$i) + 1)|0;
               $last2$3$i = $351 ? $352 : $best_contour_first$2$i$lcssa;
               $353 = (((($191) + ($last2$3$i<<3)|0)) + 4|0);
               $354 = HEAP32[$353>>2]|0;
               $355 = (($354) - ($$lcssa132))|0;
               $356 = ($355|0)<(0);
               $357 = (0 - ($355))|0;
               $358 = $356 ? $357 : $355;
               $359 = ($358|0)>(5);
               if ($359) {
                $360 = HEAP32[$$lcssa127>>2]|0;
                $361 = (($360) - ($$lcssa135))|0;
                $362 = ($361|0)<(0);
                $363 = (0 - ($361))|0;
                $364 = $362 ? $363 : $361;
                $365 = ($364|0)>($350|0);
                if (!($365)) {
                 $last2$3$i$lcssa = $last2$3$i;$p_first$4$i$lcssa = $p_first$4$i;$p_last$4$i$lcssa = $p_last$4$i;
                 break;
                }
               }
               $368 = (($193) + ($last2$3$i)|0);
               $369 = HEAP8[$368>>0]|0;
               $370 = $369 & 3;
               $371 = ($370<<24>>24)==(1);
               $372 = ($p_first$4$i|0)<(0);
               $or$cond9$i = $372 & $371;
               $p_first$5$i = $or$cond9$i ? $last2$3$i : $p_first$4$i;
               $373 = ($last2$3$i|0)==($best_segment_first$1$ph$i|0);
               if ($373) {
                $best_on_point_first$11$i = $p_first$5$i;$best_on_point_last$11$i = $best_segment_first$1$ph$i;$best_segment_first$7$i = $first1$1$i$lcssa;$best_segment_last$7$i = $best_segment_first$1$ph$i;$best_y$12$i = $$lcssa132;
                break L102;
               } else {
                $last2$2$i = $last2$3$i;$p_first$4$i = $p_first$5$i;$p_last$4$i = $last2$3$i;
               }
              }
              $366 = ($last2$3$i$lcssa|0)>($best_contour_first$2$i$lcssa|0);
              $367 = (($last2$3$i$lcssa) + -1)|0;
              $$best_contour_last$0$i = $366 ? $367 : $best_contour_last$2$i$lcssa;
              $best_on_point_first$11$i = $p_first$4$i$lcssa;$best_on_point_last$11$i = $p_last$4$i$lcssa;$best_segment_first$7$i = $first1$1$i$lcssa;$best_segment_last$7$i = $$best_contour_last$0$i;$best_y$12$i = $$lcssa132;
             }
            } while(0);
            $375 = $best_on_point_last$11$i | $best_on_point_first$11$i;
            $376 = ($375|0)>(-1);
            if ($376) {
             $377 = (($191) + ($best_on_point_last$11$i<<3)|0);
             $378 = HEAP32[$377>>2]|0;
             $379 = (($191) + ($best_on_point_first$11$i<<3)|0);
             $380 = HEAP32[$379>>2]|0;
             $381 = (($378) - ($380))|0;
             $382 = ($381|0)<(0);
             $383 = (0 - ($381))|0;
             $384 = $382 ? $383 : $381;
             $385 = HEAP32[$5>>2]|0;
             $386 = $385 >>> 3;
             $387 = ($384>>>0)>($386>>>0);
             if ($387) {
              $best_y$14$ph$i = $best_y$12$i;$p$2$i$lcssa = $p$2$i;
              break L55;
             }
            }
            $388 = (($193) + ($best_segment_first$7$i)|0);
            $389 = HEAP8[$388>>0]|0;
            $390 = $389 & 3;
            $391 = ($390<<24>>24)==(1);
            if ($391) {
             $392 = (($193) + ($best_segment_last$7$i)|0);
             $393 = HEAP8[$392>>0]|0;
             $394 = $393 & 3;
             $395 = ($394<<24>>24)==(1);
             if ($395) {
              $best_y$14$ph$i = $best_y$12$i;$p$2$i$lcssa = $p$2$i;
              break L55;
             }
            }
            $396 = $278 & 2;
            $397 = ($396<<16>>16)==(0);
            if ($397) {
             $best_y$12$i$lcssa = $best_y$12$i;$num_flats$0$ph4599$i$lcssa = $num_flats$0$ph4599$i;$p$2$i$lcssa139 = $p$2$i;
             break L53;
            }
           }
          }
         }
        } while(0);
        $200 = HEAP8[$p$2$i>>0]|0;
        $201 = ($200<<24>>24)==(0);
        if ($201) {
         $num_flats$0$ph45$lcssa61$i = $num_flats$0$ph4599$i;$num_rounds$0$ph$lcssa$i = $num_rounds$0$ph109$i;
         break L49;
        } else {
         $166 = $200;$p$089$i = $p$2$i;
        }
       }
       $402 = (($num_flats$0$ph4599$i) + 1)|0;
       $403 = (($flats$i) + ($num_flats$0$ph4599$i<<2)|0);
       HEAP32[$403>>2] = $best_y$14$ph$i;
       $404 = HEAP8[$p$2$i$lcssa>>0]|0;
       $405 = ($404<<24>>24)==(0);
       if ($405) {
        $num_flats$0$ph45$lcssa61$i = $402;$num_rounds$0$ph$lcssa$i = $num_rounds$0$ph109$i;
        break L49;
       } else {
        $479 = $404;$num_flats$0$ph4599$i = $402;$p$0$ph4498$i = $p$2$i$lcssa;
       }
      }
      $398 = (($num_rounds$0$ph109$i) + 1)|0;
      $399 = (($rounds$i) + ($num_rounds$0$ph109$i<<2)|0);
      HEAP32[$399>>2] = $best_y$12$i$lcssa;
      $400 = HEAP8[$p$2$i$lcssa139>>0]|0;
      $401 = ($400<<24>>24)==(0);
      if ($401) {
       $num_flats$0$ph45$lcssa61$i = $num_flats$0$ph4599$i$lcssa;$num_rounds$0$ph$lcssa$i = $398;
       break;
      } else {
       $478 = $400;$num_flats$0$ph110$i = $num_flats$0$ph4599$i$lcssa;$num_rounds$0$ph109$i = $398;$p$0$ph108$i = $p$2$i$lcssa139;
      }
     }
    }
   } while(0);
   $406 = ($num_rounds$0$ph$lcssa$i|0)==(0);
   $407 = $num_flats$0$ph45$lcssa61$i | $num_rounds$0$ph$lcssa$i;
   $408 = ($407|0)==(0);
   if (!($408)) {
    $409 = ($num_flats$0$ph45$lcssa61$i|0)==(0);
    $410 = ($num_rounds$0$ph$lcssa$i>>>0)>(1);
    if ($410) {
     $i$03$i$i = 1;
     while(1) {
      $$phi$trans$insert$i$i = (($rounds$i) + ($i$03$i$i<<2)|0);
      $$pre$i$i = HEAP32[$$phi$trans$insert$i$i>>2]|0;
      $j$01$i$i = $i$03$i$i;
      while(1) {
       $411 = (($j$01$i$i) + -1)|0;
       $412 = (($rounds$i) + ($411<<2)|0);
       $413 = HEAP32[$412>>2]|0;
       $414 = ($$pre$i$i|0)<($413|0);
       if (!($414)) {
        break;
       }
       $415 = (($rounds$i) + ($j$01$i$i<<2)|0);
       HEAP32[$415>>2] = $413;
       HEAP32[$412>>2] = $$pre$i$i;
       $416 = ($411|0)==(0);
       if ($416) {
        break;
       } else {
        $j$01$i$i = $411;
       }
      }
      $417 = (($i$03$i$i) + 1)|0;
      $exitcond$i$i = ($417|0)==($num_rounds$0$ph$lcssa$i|0);
      if ($exitcond$i$i) {
       break;
      } else {
       $i$03$i$i = $417;
      }
     }
    }
    $418 = ($num_flats$0$ph45$lcssa61$i>>>0)>(1);
    if ($418) {
     $i$03$i11$i = 1;
     while(1) {
      $$phi$trans$insert$i12$i = (($flats$i) + ($i$03$i11$i<<2)|0);
      $$pre$i13$i = HEAP32[$$phi$trans$insert$i12$i>>2]|0;
      $j$01$i15$i = $i$03$i11$i;
      while(1) {
       $419 = (($j$01$i15$i) + -1)|0;
       $420 = (($flats$i) + ($419<<2)|0);
       $421 = HEAP32[$420>>2]|0;
       $422 = ($$pre$i13$i|0)<($421|0);
       if (!($422)) {
        break;
       }
       $423 = (($flats$i) + ($j$01$i15$i<<2)|0);
       HEAP32[$423>>2] = $421;
       HEAP32[$420>>2] = $$pre$i13$i;
       $424 = ($419|0)==(0);
       if ($424) {
        break;
       } else {
        $j$01$i15$i = $419;
       }
      }
      $425 = (($i$03$i11$i) + 1)|0;
      $exitcond$i17$i = ($425|0)==($num_flats$0$ph45$lcssa61$i|0);
      if ($exitcond$i17$i) {
       break;
      } else {
       $i$03$i11$i = $425;
      }
     }
    }
    $426 = HEAP32[$157>>2]|0;
    $427 = (((($metrics)) + 1836|0) + (($426*28)|0)|0);
    $428 = (((((($metrics)) + 1836|0) + (($426*28)|0)|0)) + 12|0);
    $429 = (($426) + 1)|0;
    HEAP32[$157>>2] = $429;
    do {
     if ($409) {
      $430 = $num_rounds$0$ph$lcssa$i >>> 1;
      $431 = (($rounds$i) + ($430<<2)|0);
      $432 = HEAP32[$431>>2]|0;
      HEAP32[$428>>2] = $432;
      HEAP32[$427>>2] = $432;
      label = 113;
     } else {
      $433 = $num_flats$0$ph45$lcssa61$i >>> 1;
      $434 = (($flats$i) + ($433<<2)|0);
      $435 = HEAP32[$434>>2]|0;
      if ($406) {
       HEAP32[$428>>2] = $435;
       HEAP32[$427>>2] = $435;
       label = 113;
       break;
      }
      HEAP32[$427>>2] = $435;
      $436 = $num_rounds$0$ph$lcssa$i >>> 1;
      $437 = (($rounds$i) + ($436<<2)|0);
      $438 = HEAP32[$437>>2]|0;
      HEAP32[$428>>2] = $438;
      $439 = ($438|0)==($435|0);
      if ($439) {
       label = 113;
      } else {
       $440 = ($438|0)>($435|0);
       $441 = ((($bs$0115$i)) + 4|0);
       $442 = HEAP16[$441>>1]|0;
       $443 = $442 & 1;
       $444 = ($443<<16>>16)!=(0);
       $445 = $440 ^ $444;
       if ($445) {
        $446 = (($438) + ($435))|0;
        $447 = (($446|0) / 2)&-1;
        HEAP32[$428>>2] = $447;
        HEAP32[$427>>2] = $447;
        $$pre$phi$iZ2D = $441;$450 = $442;
       } else {
        $$pre$phi$iZ2D = $441;$450 = $442;
       }
      }
     }
    } while(0);
    if ((label|0) == 113) {
     label = 0;
     $$phi$trans$insert155$i = ((($bs$0115$i)) + 4|0);
     $$pre156$i = HEAP16[$$phi$trans$insert155$i>>1]|0;
     $$pre$phi$iZ2D = $$phi$trans$insert155$i;$450 = $$pre156$i;
    }
    $448 = (((((($metrics)) + 1836|0) + (($426*28)|0)|0)) + 24|0);
    $449 = $450 & 1;
    $451 = $449&65535;
    $452 = $451 << 1;
    HEAP32[$448>>2] = $452;
    $453 = HEAP16[$$pre$phi$iZ2D>>1]|0;
    $454 = $453 & 2;
    $455 = ($454<<16>>16)==(0);
    if ($455) {
     $460 = $452;
    } else {
     $456 = $452 | 4;
     HEAP32[$448>>2] = $456;
     $460 = $456;
    }
    $457 = $453 & 4;
    $458 = ($457<<16>>16)==(0);
    if (!($458)) {
     $459 = $460 | 8;
     HEAP32[$448>>2] = $459;
    }
   }
   $461 = ((($bs$0115$i)) + 8|0);
   $462 = HEAP32[$461>>2]|0;
   $463 = ($462|0)==(759);
   if ($463) {
    break;
   } else {
    $160 = $462;$bs$0115$i = $461;
   }
  }
 }
 $i$05$i = 48;$old_advance$07$i = 0;$started$06$i = 0;
 while(1) {
  $464 = HEAP32[$19>>2]|0;
  $465 = HEAP32[$464>>2]|0;
  $466 = (_FT_Get_Char_Index($465,$i$05$i)|0);
  $467 = ($466|0)==(0);
  if ($467) {
   $old_advance$2$ph$i = $old_advance$07$i;$started$2$ph$i = $started$06$i;
  } else {
   $468 = (_FT_Get_Advance($face,$466,2051,$advance$i)|0);
   $469 = ($468|0)==(0);
   if ($469) {
    $470 = ($started$06$i<<24>>24)==(0);
    $471 = HEAP32[$advance$i>>2]|0;
    if ($470) {
     $old_advance$2$ph$i = $471;$started$2$ph$i = 1;
    } else {
     $472 = ($471|0)==($old_advance$07$i|0);
     if ($472) {
      $old_advance$2$ph$i = $old_advance$07$i;$started$2$ph$i = $started$06$i;
     } else {
      $same_width$2$i = 0;
      break;
     }
    }
   } else {
    $old_advance$2$ph$i = $old_advance$07$i;$started$2$ph$i = $started$06$i;
   }
  }
  $473 = (($i$05$i) + 1)|0;
  $474 = ($473>>>0)<(58);
  if ($474) {
   $i$05$i = $473;$old_advance$07$i = $old_advance$2$ph$i;$started$06$i = $started$2$ph$i;
  } else {
   $same_width$2$i = 1;
   break;
  }
 }
 $475 = ((($metrics)) + 32|0);
 HEAP8[$475>>0] = $same_width$2$i;
 (_FT_Set_Charmap($face,$1)|0);
 STACKTOP = sp;return 0;
}
function _af_latin_hints_compute_segments($hints,$dim) {
 $hints = $hints|0;
 $dim = $dim|0;
 var $$ = 0, $$0 = 0, $$lcssa = 0, $$min_pos$0 = 0, $$ph = 0, $$phi$trans$insert = 0, $$phi$trans$insert38 = 0, $$pre = 0, $$pre$i = 0, $$pre39 = 0, $$pre40 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0;
 var $214 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $contour$026 = 0, $error$i = 0, $max_pos$0 = 0, $max_pos$1 = 0, $max_pos$2 = 0, $max_pos$3 = 0, $max_pos$4 = 0, $min_pos$0 = 0, $min_pos$3 = 0, $min_pos$4 = 0, $new_max$0$i = 0, $on_edge$0 = 0, $on_edge$1 = 0, $on_edge$2 = 0;
 var $or$cond = 0, $or$cond$i = 0, $passed$0 = 0, $passed$1 = 0, $point$028 = 0, $point1$030 = 0, $point3$1 = 0, $point3$2 = 0, $seg0$sroa$5 = 0, $seg0$sroa$6 = 0, $segment$024 = 0, $segment$1 = 0, $segment$2 = 0, $segment$3 = 0, $segment$5$ph = 0, $segment_dir$025 = 0, $segment_dir$1 = 0, $segment_dir$2 = 0, $segment_dir$4$ph = 0, $sext = 0;
 var $storemerge22 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp + 12|0;
 $seg0$sroa$5 = sp + 16|0;
 $seg0$sroa$6 = sp;
 $0 = HEAP32[$hints>>2]|0;
 $1 = ((($hints)) + 40|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($hints)) + 36|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2) + ($4<<2)|0);
 dest=$seg0$sroa$5; stop=dest+31|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0));
 ;HEAP32[$seg0$sroa$6>>2]=0|0;HEAP32[$seg0$sroa$6+4>>2]=0|0;HEAP32[$seg0$sroa$6+8>>2]=0|0;
 $6 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 24|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)<(0);
 $9 = (0 - ($7))|0;
 $10 = $8 ? $9 : $7;
 $11 = (((($hints)) + 44|0) + (($dim*1468)|0)|0);
 HEAP32[$11>>2] = 0;
 $12 = ($dim|0)==(0);
 $13 = ((($hints)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($hints)) + 24|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = (($14) + (($16*40)|0)|0);
 $18 = ($16|0)>(0);
 if ($12) {
  if ($18) {
   $point$028 = $14;
   while(1) {
    $19 = ((($point$028)) + 12|0);
    $20 = HEAP32[$19>>2]|0;
    $sext = $20 << 16;
    $21 = $sext >> 16;
    $22 = ((($point$028)) + 24|0);
    HEAP32[$22>>2] = $21;
    $23 = $20 >> 16;
    $24 = ((($point$028)) + 28|0);
    HEAP32[$24>>2] = $23;
    $25 = ((($point$028)) + 40|0);
    $26 = ($25>>>0)<($17>>>0);
    if ($26) {
     $point$028 = $25;
    } else {
     break;
    }
   }
  }
 } else {
  if ($18) {
   $point1$030 = $14;
   while(1) {
    $31 = ((($point1$030)) + 14|0);
    $32 = HEAP16[$31>>1]|0;
    $33 = $32 << 16 >> 16;
    $34 = ((($point1$030)) + 24|0);
    HEAP32[$34>>2] = $33;
    $35 = ((($point1$030)) + 12|0);
    $36 = HEAP16[$35>>1]|0;
    $37 = $36 << 16 >> 16;
    $38 = ((($point1$030)) + 28|0);
    HEAP32[$38>>2] = $37;
    $39 = ((($point1$030)) + 40|0);
    $40 = ($39>>>0)<($17>>>0);
    if ($40) {
     $point1$030 = $39;
    } else {
     break;
    }
   }
  }
 }
 $27 = ($4|0)>(0);
 if (!($27)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $28 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 8|0);
 $29 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 28|0);
 $30 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 4|0);
 $contour$026 = $2;$segment$024 = 0;$segment_dir$025 = $10;
 L12: while(1) {
  $41 = HEAP32[$contour$026>>2]|0;
  $42 = ((($41)) + 36|0);
  $43 = HEAP32[$42>>2]|0;
  $44 = ($41|0)==($43|0);
  L14: do {
   if ($44) {
    $segment$5$ph = $segment$024;$segment_dir$4$ph = $segment_dir$025;
   } else {
    $45 = ((($43)) + 3|0);
    $46 = HEAP8[$45>>0]|0;
    $47 = $46 << 24 >> 24;
    $48 = ($46<<24>>24)<(0);
    $49 = (0 - ($47))|0;
    $50 = $48 ? $49 : $47;
    $51 = ($50|0)==($10|0);
    L16: do {
     if ($51) {
      $52 = ((($41)) + 3|0);
      $53 = HEAP8[$52>>0]|0;
      $54 = $53 << 24 >> 24;
      $55 = ($53<<24>>24)<(0);
      $56 = (0 - ($54))|0;
      $57 = $55 ? $56 : $54;
      $58 = ($57|0)==($10|0);
      if ($58) {
       $59 = $46 << 24 >> 24;
       $60 = ($46<<24>>24)<(0);
       $61 = (0 - ($59))|0;
       $62 = $60 ? $61 : $59;
       $63 = ($62|0)==($10|0);
       if ($63) {
        $67 = $43;
        while(1) {
         $66 = ($67|0)==($41|0);
         if ($66) {
          $point3$1 = $41;
          break L16;
         }
         $$phi$trans$insert = ((($67)) + 36|0);
         $$pre = HEAP32[$$phi$trans$insert>>2]|0;
         $$phi$trans$insert38 = ((($$pre)) + 3|0);
         $$pre39 = HEAP8[$$phi$trans$insert38>>0]|0;
         $68 = $$pre39 << 24 >> 24;
         $69 = ($$pre39<<24>>24)<(0);
         $70 = (0 - ($68))|0;
         $71 = $69 ? $70 : $68;
         $72 = ($71|0)==($10|0);
         if ($72) {
          $67 = $$pre;
         } else {
          $$lcssa = $$pre;
          break;
         }
        }
       } else {
        $$lcssa = $43;
       }
       $64 = ((($$lcssa)) + 32|0);
       $65 = HEAP32[$64>>2]|0;
       $point3$1 = $65;
      } else {
       $point3$1 = $41;
      }
     } else {
      $point3$1 = $41;
     }
    } while(0);
    $max_pos$0 = -32000;$min_pos$0 = 32000;$on_edge$0 = 0;$passed$0 = 0;$point3$2 = $point3$1;$segment$1 = $segment$024;$segment_dir$1 = $segment_dir$025;
    while(1) {
     $73 = ($on_edge$0|0)==(0);
     if ($73) {
      $max_pos$3 = $max_pos$0;$min_pos$3 = $min_pos$0;$on_edge$1 = 0;$segment$2 = $segment$1;
     } else {
      $74 = ((($point3$2)) + 24|0);
      $75 = HEAP32[$74>>2]|0;
      $76 = ($75|0)<($min_pos$0|0);
      $$min_pos$0 = $76 ? $75 : $min_pos$0;
      $77 = ($75|0)>($max_pos$0|0);
      $max_pos$1 = $77 ? $75 : $max_pos$0;
      $78 = ((($point3$2)) + 3|0);
      $79 = HEAP8[$78>>0]|0;
      $80 = $79 << 24 >> 24;
      $81 = ($80|0)!=($segment_dir$1|0);
      $82 = ($point3$2|0)==($point3$1|0);
      $or$cond = $82 | $81;
      if ($or$cond) {
       $83 = ((($segment$1)) + 44|0);
       HEAP32[$83>>2] = $point3$2;
       $84 = (($$min_pos$0) + ($max_pos$1))|0;
       $85 = $84 >>> 1;
       $86 = $85&65535;
       $87 = ((($segment$1)) + 2|0);
       HEAP16[$87>>1] = $86;
       $88 = ((($segment$1)) + 40|0);
       $89 = HEAP32[$88>>2]|0;
       $90 = HEAP16[$89>>1]|0;
       $91 = HEAP16[$point3$2>>1]|0;
       $92 = $91 | $90;
       $93 = $92 & 3;
       $94 = ($93<<16>>16)==(0);
       if (!($94)) {
        $95 = HEAP8[$segment$1>>0]|0;
        $96 = $95&255;
        $97 = $96 | 1;
        $98 = $97&255;
        HEAP8[$segment$1>>0] = $98;
       }
       $99 = ((($point3$2)) + 28|0);
       $100 = HEAP32[$99>>2]|0;
       $101 = ((($89)) + 28|0);
       $102 = HEAP32[$101>>2]|0;
       $103 = ($102|0)<($100|0);
       $$ = $103 ? $102 : $100;
       $104 = ($102|0)>($100|0);
       $max_pos$2 = $104 ? $102 : $100;
       $105 = $$&65535;
       $106 = ((($segment$1)) + 4|0);
       HEAP16[$106>>1] = $105;
       $107 = $max_pos$2&65535;
       $108 = ((($segment$1)) + 6|0);
       HEAP16[$108>>1] = $107;
       $109 = (($max_pos$2) - ($$))|0;
       $110 = $109&65535;
       $111 = ((($segment$1)) + 8|0);
       HEAP16[$111>>1] = $110;
       $max_pos$3 = $max_pos$2;$min_pos$3 = $$;$on_edge$1 = 0;$segment$2 = 0;
      } else {
       $max_pos$3 = $max_pos$1;$min_pos$3 = $$min_pos$0;$on_edge$1 = $on_edge$0;$segment$2 = $segment$1;
      }
     }
     $112 = ($point3$2|0)==($point3$1|0);
     if ($112) {
      $113 = ($passed$0<<24>>24)==(0);
      if ($113) {
       $passed$1 = 1;
      } else {
       $segment$5$ph = $segment$2;$segment_dir$4$ph = $segment_dir$1;
       break L14;
      }
     } else {
      $passed$1 = $passed$0;
     }
     $114 = ($on_edge$1|0)==(0);
     if ($114) {
      $115 = ((($point3$2)) + 3|0);
      $116 = HEAP8[$115>>0]|0;
      $117 = $116 << 24 >> 24;
      $118 = ($116<<24>>24)<(0);
      $119 = (0 - ($117))|0;
      $120 = $118 ? $119 : $117;
      $121 = ($120|0)==($10|0);
      if ($121) {
       HEAP32[$error$i>>2] = 0;
       $122 = HEAP32[$11>>2]|0;
       $123 = ($122|0)<(18);
       if ($123) {
        $124 = HEAP32[$28>>2]|0;
        $125 = ($124|0)==(0|0);
        if ($125) {
         HEAP32[$28>>2] = $29;
         HEAP32[$30>>2] = 18;
         $144 = $122;
        } else {
         $144 = $122;
        }
       } else {
        $126 = HEAP32[$30>>2]|0;
        $127 = ($122|0)<($126|0);
        if ($127) {
         $144 = $122;
        } else {
         $128 = ($126|0)>(44739241);
         if ($128) {
          label = 30;
          break L12;
         }
         $129 = $126 >> 2;
         $130 = (($126) + 4)|0;
         $131 = (($130) + ($129))|0;
         $132 = ($129|0)<(-4);
         $133 = ($131|0)>(44739242);
         $or$cond$i = $132 | $133;
         $new_max$0$i = $or$cond$i ? 44739242 : $131;
         $134 = HEAP32[$28>>2]|0;
         $135 = ($134|0)==($29|0);
         if ($135) {
          $136 = (_ft_mem_realloc($0,48,0,$new_max$0$i,0,$error$i)|0);
          HEAP32[$28>>2] = $136;
          $137 = HEAP32[$error$i>>2]|0;
          $138 = ($137|0)==(0);
          if (!($138)) {
           $$ph = $137;
           break L12;
          }
          _memcpy(($136|0),($29|0),864)|0;
         } else {
          $139 = (_ft_mem_realloc($0,48,$126,$new_max$0$i,$134,$error$i)|0);
          HEAP32[$28>>2] = $139;
          $140 = HEAP32[$error$i>>2]|0;
          $141 = ($140|0)==(0);
          if (!($141)) {
           $$ph = $140;
           break L12;
          }
         }
         HEAP32[$30>>2] = $new_max$0$i;
         $$pre$i = HEAP32[$11>>2]|0;
         $144 = $$pre$i;
        }
       }
       $142 = HEAP32[$28>>2]|0;
       $143 = (($144) + 1)|0;
       HEAP32[$11>>2] = $143;
       $145 = (($142) + (($144*48)|0)|0);
       HEAP8[$145>>0] = 0;
       $146 = (((($142) + (($144*48)|0)|0)) + 1|0);
       dest=$146; src=$seg0$sroa$5; stop=dest+31|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
       $147 = (((($142) + (($144*48)|0)|0)) + 32|0);
       HEAP32[$147>>2] = 32000;
       $148 = (((($142) + (($144*48)|0)|0)) + 36|0);
       ;HEAP32[$148>>2]=HEAP32[$seg0$sroa$6>>2]|0;HEAP32[$148+4>>2]=HEAP32[$seg0$sroa$6+4>>2]|0;HEAP32[$148+8>>2]=HEAP32[$seg0$sroa$6+8>>2]|0;
       HEAP8[$146>>0] = $116;
       $149 = ((($point3$2)) + 24|0);
       $150 = HEAP32[$149>>2]|0;
       $151 = (((($142) + (($144*48)|0)|0)) + 40|0);
       HEAP32[$151>>2] = $point3$2;
       $152 = (((($142) + (($144*48)|0)|0)) + 44|0);
       HEAP32[$152>>2] = $point3$2;
       $max_pos$4 = $150;$min_pos$4 = $150;$on_edge$2 = 1;$segment$3 = $145;$segment_dir$2 = $117;
      } else {
       $max_pos$4 = $max_pos$3;$min_pos$4 = $min_pos$3;$on_edge$2 = 0;$segment$3 = $segment$2;$segment_dir$2 = $segment_dir$1;
      }
     } else {
      $max_pos$4 = $max_pos$3;$min_pos$4 = $min_pos$3;$on_edge$2 = $on_edge$1;$segment$3 = $segment$2;$segment_dir$2 = $segment_dir$1;
     }
     $153 = ((($point3$2)) + 32|0);
     $154 = HEAP32[$153>>2]|0;
     $max_pos$0 = $max_pos$4;$min_pos$0 = $min_pos$4;$on_edge$0 = $on_edge$2;$passed$0 = $passed$1;$point3$2 = $154;$segment$1 = $segment$3;$segment_dir$1 = $segment_dir$2;
    }
   }
  } while(0);
  $155 = ((($contour$026)) + 4|0);
  $156 = ($155>>>0)<($5>>>0);
  if ($156) {
   $contour$026 = $155;$segment$024 = $segment$5$ph;$segment_dir$025 = $segment_dir$4$ph;
  } else {
   label = 40;
   break;
  }
 }
 if ((label|0) == 30) {
  HEAP32[$error$i>>2] = 64;
  $$ph = 64;
 }
 else if ((label|0) == 40) {
  $$pre40 = HEAP32[$11>>2]|0;
  $157 = HEAP32[$28>>2]|0;
  $158 = (($157) + (($$pre40*48)|0)|0);
  $159 = ($$pre40|0)>(0);
  if ($159) {
   $storemerge22 = $157;
  } else {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
  while(1) {
   $160 = ((($storemerge22)) + 40|0);
   $161 = HEAP32[$160>>2]|0;
   $162 = ((($storemerge22)) + 44|0);
   $163 = HEAP32[$162>>2]|0;
   $164 = ((($161)) + 28|0);
   $165 = HEAP32[$164>>2]|0;
   $166 = ((($163)) + 28|0);
   $167 = HEAP32[$166>>2]|0;
   $168 = ($165|0)<($167|0);
   $169 = ((($161)) + 36|0);
   $170 = HEAP32[$169>>2]|0;
   $171 = ((($170)) + 28|0);
   $172 = HEAP32[$171>>2]|0;
   if ($168) {
    $173 = ($165|0)>($172|0);
    if ($173) {
     $174 = ((($storemerge22)) + 8|0);
     $175 = HEAP16[$174>>1]|0;
     $176 = $175&65535;
     $177 = (($165) - ($172))|0;
     $178 = $177 >>> 1;
     $179 = (($176) + ($178))|0;
     $180 = $179&65535;
     HEAP16[$174>>1] = $180;
    }
    $181 = ((($163)) + 32|0);
    $182 = HEAP32[$181>>2]|0;
    $183 = ((($182)) + 28|0);
    $184 = HEAP32[$183>>2]|0;
    $185 = ($184|0)>($167|0);
    if ($185) {
     $186 = ((($storemerge22)) + 8|0);
     $187 = HEAP16[$186>>1]|0;
     $188 = $187&65535;
     $189 = (($184) - ($167))|0;
     $190 = $189 >>> 1;
     $191 = (($188) + ($190))|0;
     $192 = $191&65535;
     HEAP16[$186>>1] = $192;
    }
   } else {
    $193 = ($172|0)>($165|0);
    if ($193) {
     $194 = ((($storemerge22)) + 8|0);
     $195 = HEAP16[$194>>1]|0;
     $196 = $195&65535;
     $197 = (($172) - ($165))|0;
     $198 = $197 >>> 1;
     $199 = (($196) + ($198))|0;
     $200 = $199&65535;
     HEAP16[$194>>1] = $200;
    }
    $201 = ((($163)) + 32|0);
    $202 = HEAP32[$201>>2]|0;
    $203 = ((($202)) + 28|0);
    $204 = HEAP32[$203>>2]|0;
    $205 = ($167|0)>($204|0);
    if ($205) {
     $206 = ((($storemerge22)) + 8|0);
     $207 = HEAP16[$206>>1]|0;
     $208 = $207&65535;
     $209 = (($167) - ($204))|0;
     $210 = $209 >>> 1;
     $211 = (($208) + ($210))|0;
     $212 = $211&65535;
     HEAP16[$206>>1] = $212;
    }
   }
   $213 = ((($storemerge22)) + 48|0);
   $214 = ($213>>>0)<($158>>>0);
   if ($214) {
    $storemerge22 = $213;
   } else {
    $$0 = 0;
    break;
   }
  }
  STACKTOP = sp;return ($$0|0);
 }
 $$0 = $$ph;
 STACKTOP = sp;return ($$0|0);
}
function _af_latin_hints_link_segments($hints,$width_count,$widths,$dim) {
 $hints = $hints|0;
 $width_count = $width_count|0;
 $widths = $widths|0;
 $dim = $dim|0;
 var $$ = 0, $$1 = 0, $$1$us = 0, $$1$v = 0, $$1$v$us = 0, $$not = 0, $$not11 = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $brmerge = 0, $dist_demerit$1 = 0, $max$0 = 0, $max$0$us = 0, $max$0$v = 0, $max$0$v$us = 0, $max_width$0 = 0, $or$cond = 0, $or$cond$us = 0, $seg1$06 = 0, $seg1$13 = 0, $seg2$04 = 0, $seg2$04$us = 0, $sext = 0, $sext13 = 0, $sext14 = 0, $sext15 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (((($hints)) + 44|0) + (($dim*1468)|0)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + (($3*48)|0)|0);
 $5 = ($width_count|0)==(0);
 if ($5) {
  $max_width$0 = 0;
 } else {
  $6 = (($width_count) + -1)|0;
  $7 = (($widths) + (($6*12)|0)|0);
  $8 = HEAP32[$7>>2]|0;
  $max_width$0 = $8;
 }
 $9 = ((($hints)) + 2988|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + 40|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($12|0) / 256)&-1;
 $$off = (($12) + 255)|0;
 $14 = ($$off>>>0)<(511);
 $$ = $14 ? 1 : $13;
 $15 = ($12*6000)|0;
 $16 = (($15|0) / 2048)&-1;
 $17 = ($3|0)>(0);
 if (!($17)) {
  return;
 }
 $18 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($max_width$0|0)==(0);
 $$not11 = ($3|0)<(1);
 $seg1$06 = $1;
 while(1) {
  $21 = ((($seg1$06)) + 1|0);
  $22 = HEAP8[$21>>0]|0;
  $23 = $22 << 24 >> 24;
  $$not = ($23|0)!=($19|0);
  $brmerge = $$not | $$not11;
  L9: do {
   if (!($brmerge)) {
    $24 = ((($seg1$06)) + 2|0);
    $25 = HEAP16[$24>>1]|0;
    $26 = $25 << 16 >> 16;
    $27 = ((($seg1$06)) + 4|0);
    $28 = ((($seg1$06)) + 32|0);
    $29 = ((($seg1$06)) + 20|0);
    if ($20) {
     $seg2$04$us = $1;
     while(1) {
      $30 = ((($seg2$04$us)) + 2|0);
      $31 = HEAP16[$30>>1]|0;
      $32 = $31 << 16 >> 16;
      $33 = ((($seg2$04$us)) + 1|0);
      $34 = HEAP8[$33>>0]|0;
      $35 = $34 << 24 >> 24;
      $36 = (0 - ($35))|0;
      $37 = ($19|0)==($36|0);
      $38 = ($31<<16>>16)>($25<<16>>16);
      $or$cond$us = $38 & $37;
      if ($or$cond$us) {
       $39 = HEAP32[$27>>2]|0;
       $40 = $39&65535;
       $41 = $39 >>> 16;
       $42 = $41&65535;
       $43 = ((($seg2$04$us)) + 4|0);
       $44 = HEAP32[$43>>2]|0;
       $45 = $44&65535;
       $46 = ($40<<16>>16)<($45<<16>>16);
       $$1$v$us = $46 ? $44 : $39;
       $sext14 = $$1$v$us << 16;
       $$1$us = $sext14 >> 16;
       $47 = $44 >>> 16;
       $48 = $47&65535;
       $49 = ($42<<16>>16)>($48<<16>>16);
       $max$0$v$us = $49 ? $47 : $41;
       $sext15 = $max$0$v$us << 16;
       $max$0$us = $sext15 >> 16;
       $50 = (($max$0$us) - ($$1$us))|0;
       $51 = ($50|0)<($$|0);
       if (!($51)) {
        $52 = (($32) - ($26))|0;
        $53 = (($16|0) / ($50|0))&-1;
        $54 = (($53) + ($52))|0;
        $55 = HEAP32[$28>>2]|0;
        $56 = ($54|0)<($55|0);
        if ($56) {
         HEAP32[$28>>2] = $54;
         HEAP32[$29>>2] = $seg2$04$us;
        }
        $57 = ((($seg2$04$us)) + 32|0);
        $58 = HEAP32[$57>>2]|0;
        $59 = ($54|0)<($58|0);
        if ($59) {
         HEAP32[$57>>2] = $54;
         $60 = ((($seg2$04$us)) + 20|0);
         HEAP32[$60>>2] = $seg1$06;
        }
       }
      }
      $61 = ((($seg2$04$us)) + 48|0);
      $62 = ($61>>>0)<($4>>>0);
      if ($62) {
       $seg2$04$us = $61;
      } else {
       break L9;
      }
     }
    } else {
     $seg2$04 = $1;
    }
    while(1) {
     $63 = ((($seg2$04)) + 2|0);
     $64 = HEAP16[$63>>1]|0;
     $65 = $64 << 16 >> 16;
     $66 = ((($seg2$04)) + 1|0);
     $67 = HEAP8[$66>>0]|0;
     $68 = $67 << 24 >> 24;
     $69 = (0 - ($68))|0;
     $70 = ($19|0)==($69|0);
     $71 = ($64<<16>>16)>($25<<16>>16);
     $or$cond = $71 & $70;
     if ($or$cond) {
      $72 = HEAP32[$27>>2]|0;
      $73 = $72&65535;
      $74 = $72 >>> 16;
      $75 = $74&65535;
      $76 = ((($seg2$04)) + 4|0);
      $77 = HEAP32[$76>>2]|0;
      $78 = $77&65535;
      $79 = ($73<<16>>16)<($78<<16>>16);
      $$1$v = $79 ? $77 : $72;
      $sext = $$1$v << 16;
      $$1 = $sext >> 16;
      $80 = $77 >>> 16;
      $81 = $80&65535;
      $82 = ($75<<16>>16)>($81<<16>>16);
      $max$0$v = $82 ? $80 : $74;
      $sext13 = $max$0$v << 16;
      $max$0 = $sext13 >> 16;
      $83 = (($max$0) - ($$1))|0;
      $84 = ($83|0)<($$|0);
      if (!($84)) {
       $85 = (($65) - ($26))|0;
       $86 = $85 << 10;
       $87 = (($86|0) / ($max_width$0|0))&-1;
       $88 = (($87) + -1024)|0;
       $89 = ($88|0)>(10000);
       if ($89) {
        $dist_demerit$1 = 32000;
       } else {
        $90 = ($87|0)>(1024);
        if ($90) {
         $91 = Math_imul($88, $88)|0;
         $92 = (($91>>>0) / 3000)&-1;
         $dist_demerit$1 = $92;
        } else {
         $dist_demerit$1 = 0;
        }
       }
       $93 = (($16|0) / ($83|0))&-1;
       $94 = (($93) + ($dist_demerit$1))|0;
       $95 = HEAP32[$28>>2]|0;
       $96 = ($94|0)<($95|0);
       if ($96) {
        HEAP32[$28>>2] = $94;
        HEAP32[$29>>2] = $seg2$04;
       }
       $97 = ((($seg2$04)) + 32|0);
       $98 = HEAP32[$97>>2]|0;
       $99 = ($94|0)<($98|0);
       if ($99) {
        HEAP32[$97>>2] = $94;
        $100 = ((($seg2$04)) + 20|0);
        HEAP32[$100>>2] = $seg1$06;
       }
      }
     }
     $101 = ((($seg2$04)) + 48|0);
     $102 = ($101>>>0)<($4>>>0);
     if ($102) {
      $seg2$04 = $101;
     } else {
      break;
     }
    }
   }
  } while(0);
  $103 = ((($seg1$06)) + 48|0);
  $104 = ($103>>>0)<($4>>>0);
  if ($104) {
   $seg1$06 = $103;
  } else {
   break;
  }
 }
 if ($17) {
  $seg1$13 = $1;
 } else {
  return;
 }
 while(1) {
  $105 = ((($seg1$13)) + 20|0);
  $106 = HEAP32[$105>>2]|0;
  $107 = ($106|0)==(0|0);
  if (!($107)) {
   $108 = ((($106)) + 20|0);
   $109 = HEAP32[$108>>2]|0;
   $110 = ($109|0)==($seg1$13|0);
   if (!($110)) {
    HEAP32[$105>>2] = 0;
    $111 = HEAP32[$108>>2]|0;
    $112 = ((($seg1$13)) + 24|0);
    HEAP32[$112>>2] = $111;
   }
  }
  $113 = ((($seg1$13)) + 48|0);
  $114 = ($113>>>0)<($4>>>0);
  if ($114) {
   $seg1$13 = $113;
  } else {
   break;
  }
 }
 return;
}
function _af_sort_and_quantize_widths($count,$table,$threshold) {
 $count = $count|0;
 $table = $table|0;
 $threshold = $threshold|0;
 var $$i$1 = 0, $$lcssa30 = 0, $$pre = 0, $$pre26 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cur_idx$014 = 0, $cur_idx$1 = 0, $cur_idx$2$lcssa = 0, $cur_idx$25 = 0, $cur_idx$3 = 0, $cur_val$015 = 0, $cur_val$1 = 0, $exitcond = 0, $i$022 = 0, $i$113 = 0;
 var $i$2 = 0, $i$3 = 0, $i$44 = 0, $j$018 = 0, $j$1$lcssa = 0, $j$16 = 0, $or$cond = 0, $sum$0$lcssa = 0, $sum$07 = 0, $swap = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $swap = sp;
 $0 = HEAP32[$count>>2]|0;
 $1 = ($0|0)==(1);
 if ($1) {
  STACKTOP = sp;return;
 }
 $2 = ($0>>>0)>(1);
 if ($2) {
  $44 = $0;$i$022 = 1;
  while(1) {
   $3 = ($i$022|0)==(0);
   if ($3) {
    $13 = $44;
   } else {
    $j$018 = $i$022;
    while(1) {
     $4 = (($table) + (($j$018*12)|0)|0);
     $5 = HEAP32[$4>>2]|0;
     $6 = (($j$018) + -1)|0;
     $7 = (($table) + (($6*12)|0)|0);
     $8 = HEAP32[$7>>2]|0;
     $9 = ($5|0)<($8|0);
     if (!($9)) {
      break;
     }
     ;HEAP32[$swap>>2]=HEAP32[$4>>2]|0;HEAP32[$swap+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$swap+8>>2]=HEAP32[$4+8>>2]|0;
     ;HEAP32[$4>>2]=HEAP32[$7>>2]|0;HEAP32[$4+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$4+8>>2]=HEAP32[$7+8>>2]|0;
     ;HEAP32[$7>>2]=HEAP32[$swap>>2]|0;HEAP32[$7+4>>2]=HEAP32[$swap+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$swap+8>>2]|0;
     $10 = ($6|0)==(0);
     if ($10) {
      break;
     } else {
      $j$018 = $6;
     }
    }
    $$pre = HEAP32[$count>>2]|0;
    $13 = $$pre;
   }
   $11 = (($i$022) + 1)|0;
   $12 = ($11>>>0)<($13>>>0);
   if ($12) {
    $44 = $13;$i$022 = $11;
   } else {
    $$lcssa30 = $13;
    break;
   }
  }
  $14 = ($$lcssa30>>>0)>(1);
  if ($14) {
   $15 = HEAP32[$table>>2]|0;
   $16 = (($$lcssa30) + -1)|0;
   $cur_idx$014 = 0;$cur_val$015 = $15;$i$113 = 1;
   while(1) {
    $17 = (($table) + (($i$113*12)|0)|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = (($18) - ($cur_val$015))|0;
    $20 = ($19|0)>($threshold|0);
    $21 = ($i$113|0)==($16|0);
    $or$cond = $21 | $20;
    if ($or$cond) {
     $22 = $21&1;
     $$i$1 = $20 ? 0 : $22;
     $i$2 = (($$i$1) + ($i$113))|0;
     $23 = ($cur_idx$014>>>0)<($i$2>>>0);
     if ($23) {
      $j$16 = $cur_idx$014;$sum$07 = 0;
      while(1) {
       $24 = (($table) + (($j$16*12)|0)|0);
       $25 = HEAP32[$24>>2]|0;
       $26 = (($25) + ($sum$07))|0;
       HEAP32[$24>>2] = 0;
       $27 = (($j$16) + 1)|0;
       $exitcond = ($27|0)==($i$2|0);
       if ($exitcond) {
        $j$1$lcssa = $i$2;$sum$0$lcssa = $26;
        break;
       } else {
        $j$16 = $27;$sum$07 = $26;
       }
      }
     } else {
      $j$1$lcssa = $cur_idx$014;$sum$0$lcssa = 0;
     }
     $28 = (($sum$0$lcssa|0) / ($j$1$lcssa|0))&-1;
     $29 = (($table) + (($cur_idx$014*12)|0)|0);
     HEAP32[$29>>2] = $28;
     $30 = ($i$2>>>0)<($16>>>0);
     if ($30) {
      $31 = (($i$2) + 1)|0;
      $32 = (($table) + (($31*12)|0)|0);
      $33 = HEAP32[$32>>2]|0;
      $cur_idx$1 = $31;$cur_val$1 = $33;$i$3 = $i$2;
     } else {
      $cur_idx$1 = $cur_idx$014;$cur_val$1 = $cur_val$015;$i$3 = $i$2;
     }
    } else {
     $cur_idx$1 = $cur_idx$014;$cur_val$1 = $cur_val$015;$i$3 = $i$113;
    }
    $34 = (($i$3) + 1)|0;
    $35 = ($34>>>0)<($$lcssa30>>>0);
    if ($35) {
     $cur_idx$014 = $cur_idx$1;$cur_val$015 = $cur_val$1;$i$113 = $34;
    } else {
     break;
    }
   }
   if ($14) {
    $45 = $$lcssa30;$cur_idx$25 = 1;$i$44 = 1;
    while(1) {
     $36 = (($table) + (($i$44*12)|0)|0);
     $37 = HEAP32[$36>>2]|0;
     $38 = ($37|0)==(0);
     if ($38) {
      $43 = $45;$cur_idx$3 = $cur_idx$25;
     } else {
      $39 = (($cur_idx$25) + 1)|0;
      $40 = (($table) + (($cur_idx$25*12)|0)|0);
      ;HEAP32[$40>>2]=HEAP32[$36>>2]|0;HEAP32[$40+4>>2]=HEAP32[$36+4>>2]|0;HEAP32[$40+8>>2]=HEAP32[$36+8>>2]|0;
      $$pre26 = HEAP32[$count>>2]|0;
      $43 = $$pre26;$cur_idx$3 = $39;
     }
     $41 = (($i$44) + 1)|0;
     $42 = ($41>>>0)<($43>>>0);
     if ($42) {
      $45 = $43;$cur_idx$25 = $cur_idx$3;$i$44 = $41;
     } else {
      $cur_idx$2$lcssa = $cur_idx$3;
      break;
     }
    }
   } else {
    $cur_idx$2$lcssa = 1;
   }
  } else {
   $cur_idx$2$lcssa = 1;
  }
 } else {
  $cur_idx$2$lcssa = 1;
 }
 HEAP32[$count>>2] = $cur_idx$2$lcssa;
 STACKTOP = sp;return;
}
function _af_glyph_hints_done($hints) {
 $hints = $hints|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ($hints|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = HEAP32[$hints>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 $3 = ((($hints)) + 44|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($hints)) + 48|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($hints)) + 52|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($hints)) + 72|0);
 $8 = ($6|0)==($7|0);
 if (!($8)) {
  _ft_mem_free($1,$6);
  HEAP32[$5>>2] = 0;
 }
 $9 = ((($hints)) + 56|0);
 HEAP32[$9>>2] = 0;
 $10 = ((($hints)) + 60|0);
 HEAP32[$10>>2] = 0;
 $11 = ((($hints)) + 64|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($hints)) + 936|0);
 $14 = ($12|0)==($13|0);
 if (!($14)) {
  _ft_mem_free($1,$12);
  HEAP32[$11>>2] = 0;
 }
 $15 = ((($hints)) + 1512|0);
 HEAP32[$15>>2] = 0;
 $16 = ((($hints)) + 1516|0);
 HEAP32[$16>>2] = 0;
 $17 = ((($hints)) + 1520|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($hints)) + 1540|0);
 $20 = ($18|0)==($19|0);
 if (!($20)) {
  _ft_mem_free($1,$18);
  HEAP32[$17>>2] = 0;
 }
 $31 = ((($hints)) + 1524|0);
 HEAP32[$31>>2] = 0;
 $32 = ((($hints)) + 1528|0);
 HEAP32[$32>>2] = 0;
 $33 = ((($hints)) + 1532|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = ((($hints)) + 2404|0);
 $36 = ($34|0)==($35|0);
 if (!($36)) {
  _ft_mem_free($1,$34);
  HEAP32[$33>>2] = 0;
 }
 $22 = ((($hints)) + 40|0);
 $21 = HEAP32[$22>>2]|0;
 $37 = ((($hints)) + 3000|0);
 $38 = ($21|0)==($37|0);
 if (!($38)) {
  _ft_mem_free($1,$21);
  HEAP32[$22>>2] = 0;
 }
 $23 = ((($hints)) + 32|0);
 HEAP32[$23>>2] = 0;
 $24 = ((($hints)) + 36|0);
 HEAP32[$24>>2] = 0;
 $25 = ((($hints)) + 28|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ((($hints)) + 3032|0);
 $28 = ($26|0)==($27|0);
 if (!($28)) {
  _ft_mem_free($1,$26);
  HEAP32[$25>>2] = 0;
 }
 $29 = ((($hints)) + 20|0);
 HEAP32[$29>>2] = 0;
 $30 = ((($hints)) + 24|0);
 HEAP32[$30>>2] = 0;
 HEAP32[$hints>>2] = 0;
 return;
}
function _af_latin_metrics_scale($metrics,$scaler) {
 $metrics = $metrics|0;
 $scaler = $scaler|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($scaler)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($metrics)) + 4|0);
 $3 = ((($metrics)) + 24|0);
 HEAP32[$3>>2] = $1;
 $4 = HEAP32[$scaler>>2]|0;
 HEAP32[$2>>2] = $4;
 $5 = ((($scaler)) + 24|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($metrics)) + 28|0);
 HEAP32[$7>>2] = $6;
 _af_latin_metrics_scale_dim($metrics,$scaler,0);
 _af_latin_metrics_scale_dim($metrics,$scaler,1);
 return;
}
function _af_latin_metrics_scale_dim($metrics,$scaler,$dim) {
 $metrics = $metrics|0;
 $scaler = $scaler|0;
 $dim = $dim|0;
 var $$ = 0, $$6 = 0, $$7 = 0, $$delta2$1 = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $delta$0 = 0, $delta$0$in = 0, $delta2$1 = 0, $nn$013 = 0, $nn$013$lcssa = 0, $nn$111 = 0, $nn$210 = 0, $or$cond = 0, $or$cond5 = 0, $scale$0 = 0, $scale$0$in = 0, $scale$2 = 0;
 var $threshold$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($dim|0)==(0);
 $1 = ((($scaler)) + 4|0);
 $2 = ((($scaler)) + 12|0);
 $3 = ((($scaler)) + 8|0);
 $4 = ((($scaler)) + 16|0);
 $delta$0$in = $0 ? $2 : $4;
 $scale$0$in = $0 ? $1 : $3;
 $scale$0 = HEAP32[$scale$0$in>>2]|0;
 $delta$0 = HEAP32[$delta$0$in>>2]|0;
 $5 = (((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 1564|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==($scale$0|0);
 $8 = (((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 1568|0);
 if ($7) {
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)==($delta$0|0);
  if ($10) {
   return;
  }
 }
 HEAP32[$5>>2] = $scale$0;
 HEAP32[$8>>2] = $delta$0;
 $11 = ((($metrics)) + 1832|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)==(0);
 L5: do {
  if ($13) {
   $scale$2 = $scale$0;
  } else {
   $nn$013 = 0;
   while(1) {
    $14 = (((((($metrics)) + 1836|0) + (($nn$013*28)|0)|0)) + 24|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = $15 & 8;
    $17 = ($16|0)==(0);
    if (!($17)) {
     $nn$013$lcssa = $nn$013;
     break;
    }
    $18 = (($nn$013) + 1)|0;
    $19 = ($18>>>0)<($12>>>0);
    if ($19) {
     $nn$013 = $18;
    } else {
     $scale$2 = $scale$0;
     break L5;
    }
   }
   $20 = (((((($metrics)) + 1836|0) + (($nn$013$lcssa*28)|0)|0)) + 12|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = HEAP32[$3>>2]|0;
   $23 = (_FT_MulFix($21,$22)|0);
   $24 = ((($metrics)) + 4|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = ((($25)) + 88|0);
   $27 = HEAP32[$26>>2]|0;
   $28 = ((($27)) + 12|0);
   $29 = HEAP16[$28>>1]|0;
   $30 = ((($metrics)) + 36|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = ((($31)) + 12|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ($33|0)==(0);
   if ($34) {
    $threshold$0 = 40;
   } else {
    $35 = $29&65535;
    $36 = ($35>>>0)<=($33>>>0);
    $37 = ($29&65535)>(5);
    $or$cond = $37 & $36;
    $$ = $or$cond ? 52 : 40;
    $threshold$0 = $$;
   }
   $38 = (($threshold$0) + ($23))|0;
   $39 = $38 & -64;
   $40 = ($23|0)!=($39|0);
   $41 = ($dim|0)==(1);
   $or$cond5 = $41 & $40;
   if ($or$cond5) {
    $42 = (_FT_MulDiv($scale$0,$39,$23)|0);
    $scale$2 = $42;
   } else {
    $scale$2 = $scale$0;
   }
  }
 } while(0);
 $43 = (((($metrics)) + 44|0) + (($dim*1572)|0)|0);
 HEAP32[$43>>2] = $scale$2;
 $44 = (((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 4|0);
 HEAP32[$44>>2] = $delta$0;
 if ($0) {
  $45 = ((($metrics)) + 8|0);
  HEAP32[$45>>2] = $scale$2;
  $46 = ((($metrics)) + 16|0);
  HEAP32[$46>>2] = $delta$0;
 } else {
  $47 = ((($metrics)) + 12|0);
  HEAP32[$47>>2] = $scale$2;
  $48 = ((($metrics)) + 20|0);
  HEAP32[$48>>2] = $delta$0;
 }
 $49 = (((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 8|0);
 $50 = HEAP32[$49>>2]|0;
 $51 = ($50|0)==(0);
 if (!($51)) {
  $nn$111 = 0;
  while(1) {
   $52 = ((((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 12|0) + (($nn$111*12)|0)|0);
   $53 = HEAP32[$52>>2]|0;
   $54 = (_FT_MulFix($53,$scale$2)|0);
   $55 = ((((((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 12|0) + (($nn$111*12)|0)|0)) + 4|0);
   HEAP32[$55>>2] = $54;
   $56 = ((((((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 12|0) + (($nn$111*12)|0)|0)) + 8|0);
   HEAP32[$56>>2] = $54;
   $57 = (($nn$111) + 1)|0;
   $58 = HEAP32[$49>>2]|0;
   $59 = ($57>>>0)<($58>>>0);
   if ($59) {
    $nn$111 = $57;
   } else {
    break;
   }
  }
 }
 $60 = (((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 208|0);
 $61 = HEAP32[$60>>2]|0;
 $62 = (_FT_MulFix($61,$scale$2)|0);
 $63 = ($62|0)<(40);
 $64 = $63&1;
 $65 = (((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 212|0);
 HEAP8[$65>>0] = $64;
 $66 = ($dim|0)==(1);
 if (!($66)) {
  return;
 }
 $67 = HEAP32[$11>>2]|0;
 $68 = ($67|0)==(0);
 if ($68) {
  return;
 } else {
  $nn$210 = 0;
 }
 while(1) {
  $69 = (((($metrics)) + 1836|0) + (($nn$210*28)|0)|0);
  $70 = HEAP32[$69>>2]|0;
  $71 = (_FT_MulFix($70,$scale$2)|0);
  $72 = (($71) + ($delta$0))|0;
  $73 = (((((($metrics)) + 1836|0) + (($nn$210*28)|0)|0)) + 4|0);
  HEAP32[$73>>2] = $72;
  $74 = (((((($metrics)) + 1836|0) + (($nn$210*28)|0)|0)) + 8|0);
  HEAP32[$74>>2] = $72;
  $75 = (((((($metrics)) + 1836|0) + (($nn$210*28)|0)|0)) + 12|0);
  $76 = HEAP32[$75>>2]|0;
  $77 = (_FT_MulFix($76,$scale$2)|0);
  $78 = (($77) + ($delta$0))|0;
  $79 = (((((($metrics)) + 1836|0) + (($nn$210*28)|0)|0)) + 16|0);
  HEAP32[$79>>2] = $78;
  $80 = (((((($metrics)) + 1836|0) + (($nn$210*28)|0)|0)) + 20|0);
  HEAP32[$80>>2] = $78;
  $81 = (((((($metrics)) + 1836|0) + (($nn$210*28)|0)|0)) + 24|0);
  $82 = HEAP32[$81>>2]|0;
  $83 = $82 & -2;
  HEAP32[$81>>2] = $83;
  $84 = HEAP32[$69>>2]|0;
  $85 = HEAP32[$75>>2]|0;
  $86 = (($84) - ($85))|0;
  $87 = (_FT_MulFix($86,$scale$2)|0);
  $$off = (($87) + 48)|0;
  $88 = ($$off>>>0)<(97);
  if ($88) {
   $89 = ($87|0)<(0);
   $90 = (0 - ($87))|0;
   $$6 = $89 ? $90 : $87;
   $91 = ($$6|0)<(32);
   $92 = ($$6|0)<(48);
   $$7 = $92 ? 32 : 64;
   $delta2$1 = $91 ? 0 : $$7;
   $93 = (0 - ($delta2$1))|0;
   $$delta2$1 = $89 ? $93 : $delta2$1;
   $94 = HEAP32[$73>>2]|0;
   $95 = (($94) + 32)|0;
   $96 = $95 & -64;
   HEAP32[$74>>2] = $96;
   $97 = (($96) - ($$delta2$1))|0;
   HEAP32[$80>>2] = $97;
   $98 = HEAP32[$81>>2]|0;
   $99 = $98 | 1;
   HEAP32[$81>>2] = $99;
  }
  $100 = (($nn$210) + 1)|0;
  $101 = HEAP32[$11>>2]|0;
  $102 = ($100>>>0)<($101>>>0);
  if ($102) {
   $nn$210 = $100;
  } else {
   break;
  }
 }
 return;
}
function _af_latin_hints_init($hints,$metrics) {
 $hints = $hints|0;
 $metrics = $metrics|0;
 var $$ = 0, $$scaler_flags$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond3 = 0, $other_flags$1 = 0, $other_flags$1$ = 0, $other_flags$3 = 0, $scaler_flags$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($metrics)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($hints)) + 2988|0);
 HEAP32[$2>>2] = $metrics;
 $3 = ((($metrics)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($hints)) + 2980|0);
 HEAP32[$5>>2] = $4;
 $6 = ((($metrics)) + 44|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($hints)) + 4|0);
 HEAP32[$8>>2] = $7;
 $9 = ((($metrics)) + 48|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($hints)) + 8|0);
 HEAP32[$11>>2] = $10;
 $12 = ((($metrics)) + 1616|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($hints)) + 12|0);
 HEAP32[$14>>2] = $13;
 $15 = ((($metrics)) + 1620|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($hints)) + 16|0);
 HEAP32[$17>>2] = $16;
 $18 = ((($metrics)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($19|0)==(2);
 $21 = $19 | 1;
 $22 = ($21|0)==(3);
 $$ = $22&1;
 $23 = ($19|0)==(4);
 $or$cond3 = $20 | $23;
 $24 = $$ | 2;
 $other_flags$1 = $or$cond3 ? $24 : $$;
 $25 = ($19|0)==(1);
 $26 = $other_flags$1 | 4;
 $other_flags$1$ = $25 ? $other_flags$1 : $26;
 $27 = $other_flags$1$ | 8;
 $other_flags$3 = $20 ? $27 : $other_flags$1$;
 if ($25) {
  label = 3;
 } else {
  $28 = ((($1)) + 12|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = $29 & 1;
  $31 = ($30|0)==(0);
  if ($31) {
   $scaler_flags$0 = $4;
  } else {
   label = 3;
  }
 }
 if ((label|0) == 3) {
  $32 = $4 | 1;
  $scaler_flags$0 = $32;
 }
 $33 = ((($metrics)) + 36|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = ((($34)) + 212|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = ((($36)) + 20|0);
 $38 = HEAP8[$37>>0]|0;
 $39 = ($38<<24>>24)==(0);
 $40 = $scaler_flags$0 | 8;
 $$scaler_flags$0 = $39 ? $40 : $scaler_flags$0;
 HEAP32[$5>>2] = $$scaler_flags$0;
 $41 = ((($hints)) + 2984|0);
 HEAP32[$41>>2] = $other_flags$3;
 return 0;
}
function _af_latin_hints_apply($hints,$outline,$metrics) {
 $hints = $hints|0;
 $outline = $outline|0;
 $metrics = $metrics|0;
 var $$$i = 0, $$11$i = 0, $$12$i = 0, $$13$i = 0, $$14$i = 0, $$15$i = 0, $$16$i = 0, $$17$i = 0, $$18$i = 0, $$19$i = 0, $$20$i = 0, $$5$i = 0, $$6$i = 0, $$best_blue$1$i = 0, $$best_blue_is_neutral$1$i = 0, $$best_dist$1$i = 0, $$idx$i = 0, $$idx27$i = 0, $$idx27$val$i = 0, $$idx28$i = 0;
 var $$idx28$val$i = 0, $$idx29$i = 0, $$idx29$val$i = 0, $$idx33$i = 0, $$idx33$val$i = 0, $$lcssa$i = 0, $$lobit4$i = 0, $$not$i = 0, $$not$i7 = 0, $$not9$i = 0, $$phi$trans$insert27 = 0, $$pre$i = 0, $$pre$phi$iZ2D = 0, $$pre$phi67$iZ2D = 0, $$pre$phiZ2D = 0, $$pre28 = 0, $$pre64$i = 0, $$pre65$i = 0, $$pre66$i = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0;
 var $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0;
 var $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0;
 var $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0;
 var $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0;
 var $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0;
 var $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0;
 var $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0;
 var $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0;
 var $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0;
 var $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0;
 var $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0;
 var $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0;
 var $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0;
 var $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $after$0$i = 0, $after$0$i$lcssa50 = 0, $anchor$055$i = 0, $anchor$2$i = 0, $anchor$4$lcssa$i = 0, $anchor$4$ph$i = 0, $anchor$450$i = 0, $anchor$6$i = 0, $anchor$745$i = 0, $anchor$8$i = 0, $anchor$9$i = 0, $bb$010$i = 0, $before$0$i = 0;
 var $before$0$i$lcssa = 0, $best_blue$09$i = 0, $best_blue$1$i = 0, $best_blue$5$i = 0, $best_blue$5$i$lcssa = 0, $best_blue_is_neutral$08$i = 0, $best_blue_is_neutral$1$i = 0, $best_blue_is_neutral$5$i = 0, $best_blue_is_neutral$5$i$lcssa = 0, $best_dist$07$i = 0, $best_dist$1$i = 0, $best_dist$5$i = 0, $blue$037$i = 0, $brmerge$i = 0, $brmerge23$demorgan$i = 0, $cur_pos15$0$i = 0, $delta = 0, $delta1$0$i = 0, $delta2$1$i = 0, $dim$021 = 0;
 var $edge$0$anchor$0$i = 0, $edge$012$i = 0, $edge$056$i = 0, $edge$1$idx25$val$i = 0, $edge$1$idx25$val$pre$i = 0, $edge$151$i = 0, $edge$2$pn$i = 0, $edge$2$pn$i$lcssa = 0, $edge$2$pn8$i = 0, $edge$2$pn8$i$lcssa49 = 0, $edge$246$i = 0, $edge1$0$idx$i = 0, $edge1$0$idx$val$i = 0, $edge1$036$i = 0, $edge19$0$i = 0, $edge2$035$i = 0, $edge210$0$i = 0, $edge3$0$i = 0, $error$2 = 0, $error2$0$i = 0;
 var $exitcond = 0, $has_serifs$0$lcssa$i = 0, $has_serifs$049$i = 0, $has_serifs$1$i = 0, $hints$idx$val = 0, $hints$idx1$val = 0, $hints$idx3 = 0, $hints$idx3$val = 0, $hints$idx4 = 0, $hints$idx4$val = 0, $notlhs$i = 0, $notlhs$i9 = 0, $notrhs$i = 0, $notrhs$i10 = 0, $or$cond = 0, $or$cond$i = 0, $or$cond$i6 = 0, $or$cond$not$i = 0, $or$cond24$i = 0, $or$cond3$i = 0;
 var $or$cond40 = 0, $or$cond42 = 0, $or$cond6$not$i = 0, $or$cond60$i = 0, $or$cond61$i = 0, $outline$idx = 0, $outline$idx$val = 0, $outline$idx2 = 0, $outline$idx2$val = 0, $point$0$i = 0, $point$0$in$i = 0, $point$01$i = 0, $point$03$i = 0, $point2$0$i = 0, $point2$0$in$i = 0, $scale = 0, $seg$06$i = 0, $seg$17$i = 0, $tag$03$i = 0, $vec$02$i = 0;
 var $warper = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $warper = sp + 8|0;
 $scale = sp + 4|0;
 $delta = sp;
 $0 = (_af_glyph_hints_reload($hints,$outline)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  $error$2 = $0;
  STACKTOP = sp;return ($error$2|0);
 }
 $2 = ((($metrics)) + 24|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(1);
 $5 = ((($hints)) + 2980|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $6 & 8;
 $8 = ($7|0)==(0);
 $or$cond = $4 & $8;
 $9 = $6 & 1;
 $10 = ($9|0)==(0);
 $or$cond40 = $or$cond | $10;
 do {
  if ($or$cond40) {
   $11 = ((($metrics)) + 52|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ((($metrics)) + 56|0);
   $14 = (_af_latin_hints_detect_features($hints,$12,$13,0)|0);
   $15 = ($14|0)==(0);
   if ($15) {
    $$phi$trans$insert27 = ((($hints)) + 2980|0);
    $$pre28 = HEAP32[$$phi$trans$insert27>>2]|0;
    $$pre$phiZ2D = $$phi$trans$insert27;$17 = $$pre28;
    break;
   } else {
    $error$2 = $14;
    STACKTOP = sp;return ($error$2|0);
   }
  } else {
   $$pre$phiZ2D = $5;$17 = $6;
  }
 } while(0);
 $16 = $17 & 2;
 $18 = ($16|0)==(0);
 if ($18) {
  $19 = ((($metrics)) + 1624|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = ((($metrics)) + 1628|0);
  $22 = (_af_latin_hints_detect_features($hints,$20,$21,1)|0);
  $23 = ($22|0)==(0);
  if (!($23)) {
   $error$2 = $22;
   STACKTOP = sp;return ($error$2|0);
  }
  $24 = ((($hints)) + 1532|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = ((($hints)) + 1524|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = (($25) + (($27*48)|0)|0);
  $29 = ((($metrics)) + 1616|0);
  $30 = HEAP32[$29>>2]|0;
  $31 = ($27|0)>(0);
  if ($31) {
   $32 = ((($metrics)) + 40|0);
   $33 = ((($metrics)) + 1832|0);
   $34 = ((($hints)) + 1536|0);
   $edge$012$i = $25;
   while(1) {
    $35 = HEAP32[$32>>2]|0;
    $36 = (($35>>>0) / 40)&-1;
    $37 = (_FT_MulFix($36,$30)|0);
    $38 = HEAP32[$33>>2]|0;
    $39 = ($38|0)==(0);
    if (!($39)) {
     $40 = ($37|0)>(32);
     $$$i = $40 ? 32 : $37;
     $41 = ((($edge$012$i)) + 13|0);
     $42 = ((($edge$012$i)) + 12|0);
     $bb$010$i = 0;$best_blue$09$i = 0;$best_blue_is_neutral$08$i = 0;$best_dist$07$i = $$$i;
     while(1) {
      $43 = (((($metrics)) + 1836|0) + (($bb$010$i*28)|0)|0);
      $44 = (((((($metrics)) + 1836|0) + (($bb$010$i*28)|0)|0)) + 24|0);
      $45 = HEAP32[$44>>2]|0;
      $46 = $45 & 1;
      $47 = ($46|0)==(0);
      if ($47) {
       $best_blue$5$i = $best_blue$09$i;$best_blue_is_neutral$5$i = $best_blue_is_neutral$08$i;$best_dist$5$i = $best_dist$07$i;
      } else {
       $48 = $45 >>> 2;
       $$lobit4$i = $48 & 1;
       $49 = $$lobit4$i&255;
       $50 = HEAP8[$41>>0]|0;
       $51 = $50 << 24 >> 24;
       $52 = HEAP32[$34>>2]|0;
       $53 = ($51|0)==($52|0);
       $54 = $45 & 2;
       $55 = ($54|0)!=(0);
       $56 = $55 ^ $53;
       $$not$i = $56 ^ 1;
       $57 = ($$lobit4$i|0)==(0);
       $or$cond$i = $57 & $$not$i;
       if ($or$cond$i) {
        $best_blue$5$i = $best_blue$09$i;$best_blue_is_neutral$5$i = $best_blue_is_neutral$08$i;$best_dist$5$i = $best_dist$07$i;
       } else {
        $58 = HEAP16[$edge$012$i>>1]|0;
        $59 = $58 << 16 >> 16;
        $60 = HEAP32[$43>>2]|0;
        $61 = (($59) - ($60))|0;
        $62 = ($61|0)<(0);
        $63 = (0 - ($61))|0;
        $$5$i = $62 ? $63 : $61;
        $64 = (_FT_MulFix($$5$i,$30)|0);
        $65 = ($64|0)<($best_dist$07$i|0);
        $best_dist$1$i = $65 ? $64 : $best_dist$07$i;
        $best_blue_is_neutral$1$i = $65 ? $49 : $best_blue_is_neutral$08$i;
        $best_blue$1$i = $65 ? $43 : $best_blue$09$i;
        $66 = HEAP8[$42>>0]|0;
        $67 = $66 & 1;
        $notlhs$i = ($67<<24>>24)==(0);
        $notrhs$i = ($64|0)==(0);
        $or$cond$not$i = $notrhs$i | $notlhs$i;
        $68 = ($49<<24>>24)!=(0);
        $or$cond3$i = $68 | $or$cond$not$i;
        if ($or$cond3$i) {
         $best_blue$5$i = $best_blue$1$i;$best_blue_is_neutral$5$i = $best_blue_is_neutral$1$i;$best_dist$5$i = $best_dist$1$i;
        } else {
         $69 = HEAP16[$edge$012$i>>1]|0;
         $70 = $69 << 16 >> 16;
         $71 = HEAP32[$43>>2]|0;
         $72 = ($70|0)<($71|0);
         $73 = $55 ^ $72;
         if ($73) {
          $74 = (((((($metrics)) + 1836|0) + (($bb$010$i*28)|0)|0)) + 12|0);
          $75 = HEAP32[$74>>2]|0;
          $76 = (($70) - ($75))|0;
          $77 = ($76|0)<(0);
          $78 = (0 - ($76))|0;
          $$6$i = $77 ? $78 : $76;
          $79 = (_FT_MulFix($$6$i,$30)|0);
          $80 = ($79|0)<($best_dist$1$i|0);
          $$best_dist$1$i = $80 ? $79 : $best_dist$1$i;
          $$best_blue_is_neutral$1$i = $80 ? 0 : $best_blue_is_neutral$1$i;
          $$best_blue$1$i = $80 ? $74 : $best_blue$1$i;
          $best_blue$5$i = $$best_blue$1$i;$best_blue_is_neutral$5$i = $$best_blue_is_neutral$1$i;$best_dist$5$i = $$best_dist$1$i;
         } else {
          $best_blue$5$i = $best_blue$1$i;$best_blue_is_neutral$5$i = $best_blue_is_neutral$1$i;$best_dist$5$i = $best_dist$1$i;
         }
        }
       }
      }
      $81 = (($bb$010$i) + 1)|0;
      $82 = HEAP32[$33>>2]|0;
      $83 = ($81>>>0)<($82>>>0);
      if ($83) {
       $bb$010$i = $81;$best_blue$09$i = $best_blue$5$i;$best_blue_is_neutral$08$i = $best_blue_is_neutral$5$i;$best_dist$07$i = $best_dist$5$i;
      } else {
       $best_blue$5$i$lcssa = $best_blue$5$i;$best_blue_is_neutral$5$i$lcssa = $best_blue_is_neutral$5$i;
       break;
      }
     }
     $84 = ($best_blue$5$i$lcssa|0)==(0|0);
     if (!($84)) {
      $85 = ((($edge$012$i)) + 20|0);
      HEAP32[$85>>2] = $best_blue$5$i$lcssa;
      $86 = ($best_blue_is_neutral$5$i$lcssa<<24>>24)==(0);
      if (!($86)) {
       $87 = HEAP8[$42>>0]|0;
       $88 = $87&255;
       $89 = $88 | 8;
       $90 = $89&255;
       HEAP8[$42>>0] = $90;
      }
     }
    }
    $91 = ((($edge$012$i)) + 48|0);
    $92 = ($91>>>0)<($28>>>0);
    if ($92) {
     $edge$012$i = $91;
    } else {
     break;
    }
   }
  }
 }
 $hints$idx3 = ((($hints)) + 24|0);
 $hints$idx4 = ((($hints)) + 28|0);
 $dim$021 = 0;
 L33: while(1) {
  $93 = ($dim$021|0)==(0);
  do {
   if ($93) {
    $94 = HEAP32[$2>>2]|0;
    $95 = ($94|0)==(1);
    $96 = HEAP32[$$pre$phiZ2D>>2]|0;
    $97 = $96 & 8;
    $98 = ($97|0)==(0);
    $or$cond42 = $95 & $98;
    if (!($or$cond42)) {
     $111 = $96 & 1;
     $112 = ($111|0)==(0);
     if ($112) {
      label = 30;
      break;
     } else {
      $dim$021 = 1;
      continue L33;
     }
    }
    _af_warper_compute($warper,$hints,0,$scale,$delta);
    $99 = HEAP32[$scale>>2]|0;
    $100 = HEAP32[$delta>>2]|0;
    $hints$idx3$val = HEAP32[$hints$idx3>>2]|0;
    $hints$idx4$val = HEAP32[$hints$idx4>>2]|0;
    $101 = (($hints$idx4$val) + (($hints$idx3$val*40)|0)|0);
    $102 = ($hints$idx3$val|0)>(0);
    if ($102) {
     $point$03$i = $hints$idx4$val;
     while(1) {
      $103 = ((($point$03$i)) + 12|0);
      $104 = HEAP16[$103>>1]|0;
      $105 = $104 << 16 >> 16;
      $106 = (_FT_MulFix($105,$99)|0);
      $107 = (($106) + ($100))|0;
      $108 = ((($point$03$i)) + 16|0);
      HEAP32[$108>>2] = $107;
      $109 = ((($point$03$i)) + 40|0);
      $110 = ($109>>>0)<($101>>>0);
      if ($110) {
       $point$03$i = $109;
      } else {
       break;
      }
     }
    }
   } else {
    $113 = ($dim$021|0)==(1);
    if ($113) {
     $114 = HEAP32[$$pre$phiZ2D>>2]|0;
     $115 = $114 & 2;
     $116 = ($115|0)==(0);
     if ($116) {
      label = 30;
     } else {
      break L33;
     }
    }
   }
  } while(0);
  if ((label|0) == 30) {
   label = 0;
   $117 = (((((($hints)) + 44|0) + (($dim$021*1468)|0)|0)) + 20|0);
   $118 = HEAP32[$117>>2]|0;
   $119 = (((((($hints)) + 44|0) + (($dim$021*1468)|0)|0)) + 12|0);
   $120 = HEAP32[$119>>2]|0;
   $121 = (($118) + (($120*48)|0)|0);
   $122 = ($dim$021|0)==(1);
   $123 = ($120|0)>(0);
   $or$cond60$i = $122 & $123;
   if ($or$cond60$i) {
    $anchor$055$i = 0;$edge$056$i = $118;
    while(1) {
     $124 = ((($edge$056$i)) + 12|0);
     $125 = HEAP8[$124>>0]|0;
     $126 = $125&255;
     $127 = $126 & 4;
     $128 = ($127|0)==(0);
     do {
      if ($128) {
       $129 = ((($edge$056$i)) + 24|0);
       $130 = HEAP32[$129>>2]|0;
       $131 = ((($edge$056$i)) + 20|0);
       $132 = HEAP32[$131>>2]|0;
       $133 = ($132|0)!=(0|0);
       $134 = ($130|0)!=(0|0);
       $or$cond$i6 = $134 & $133;
       do {
        if ($or$cond$i6) {
         $135 = ((($130)) + 20|0);
         $136 = HEAP32[$135>>2]|0;
         $137 = ($136|0)==(0|0);
         if ($137) {
          $149 = $132;
         } else {
          $138 = ((($130)) + 12|0);
          $139 = HEAP8[$138>>0]|0;
          $140 = $139 & 8;
          $141 = ($140<<24>>24)==(0);
          if (!($141)) {
           HEAP32[$135>>2] = 0;
           $142 = $139&255;
           $143 = $142 & 247;
           $144 = $143&255;
           HEAP8[$138>>0] = $144;
           $$pre$i = HEAP32[$131>>2]|0;
           $149 = $$pre$i;
           break;
          }
          $145 = $126 & 8;
          $146 = ($145|0)==(0);
          if ($146) {
           $149 = $132;
          } else {
           HEAP32[$131>>2] = 0;
           $147 = $126 & 247;
           $148 = $147&255;
           HEAP8[$124>>0] = $148;
           $149 = 0;
          }
         }
        } else {
         $149 = $132;
        }
       } while(0);
       $$not$i7 = ($149|0)!=(0|0);
       $$not9$i = $134 ^ 1;
       $brmerge$i = $$not$i7 | $$not9$i;
       if ($brmerge$i) {
        if ($$not$i7) {
         $blue$037$i = $149;$edge1$036$i = $edge$056$i;$edge2$035$i = $130;
        } else {
         $anchor$2$i = $anchor$055$i;
         break;
        }
       } else {
        $150 = ((($130)) + 20|0);
        $151 = HEAP32[$150>>2]|0;
        $152 = ($151|0)==(0|0);
        if ($152) {
         $anchor$2$i = $anchor$055$i;
         break;
        } else {
         $blue$037$i = $151;$edge1$036$i = $130;$edge2$035$i = $edge$056$i;
        }
       }
       $153 = ((($blue$037$i)) + 8|0);
       $154 = HEAP32[$153>>2]|0;
       $155 = ((($edge1$036$i)) + 8|0);
       HEAP32[$155>>2] = $154;
       $156 = ((($edge1$036$i)) + 12|0);
       $157 = HEAP8[$156>>0]|0;
       $158 = $157&255;
       $159 = $158 | 4;
       $160 = $159&255;
       HEAP8[$156>>0] = $160;
       $161 = ($edge2$035$i|0)==(0|0);
       if (!($161)) {
        $162 = ((($edge2$035$i)) + 20|0);
        $163 = HEAP32[$162>>2]|0;
        $164 = ($163|0)==(0|0);
        if ($164) {
         $edge1$0$idx$i = ((($edge1$036$i)) + 4|0);
         $edge1$0$idx$val$i = HEAP32[$edge1$0$idx$i>>2]|0;
         $165 = ((($edge2$035$i)) + 4|0);
         $166 = HEAP32[$165>>2]|0;
         $167 = (($166) - ($edge1$0$idx$val$i))|0;
         $168 = ((($edge2$035$i)) + 12|0);
         $169 = HEAP8[$168>>0]|0;
         $170 = $169&255;
         $171 = (_af_latin_compute_stem_width($hints,1,$167,$159,$170)|0);
         $172 = (($171) + ($154))|0;
         $173 = ((($edge2$035$i)) + 8|0);
         HEAP32[$173>>2] = $172;
         $174 = $170 | 4;
         $175 = $174&255;
         HEAP8[$168>>0] = $175;
        }
       }
       $176 = ($anchor$055$i|0)==(0|0);
       $edge$0$anchor$0$i = $176 ? $edge$056$i : $anchor$055$i;
       $anchor$2$i = $edge$0$anchor$0$i;
      } else {
       $anchor$2$i = $anchor$055$i;
      }
     } while(0);
     $177 = ((($edge$056$i)) + 48|0);
     $178 = ($177>>>0)<($121>>>0);
     if ($178) {
      $anchor$055$i = $anchor$2$i;$edge$056$i = $177;
     } else {
      $anchor$4$ph$i = $anchor$2$i;
      break;
     }
    }
   } else {
    $anchor$4$ph$i = 0;
   }
   if ($123) {
    $anchor$450$i = $anchor$4$ph$i;$edge$151$i = $118;$has_serifs$049$i = 0;
    while(1) {
     $179 = ((($edge$151$i)) + 12|0);
     $180 = HEAP8[$179>>0]|0;
     $181 = $180&255;
     $182 = $181 & 4;
     $183 = ($182|0)==(0);
     do {
      if ($183) {
       $184 = ((($edge$151$i)) + 24|0);
       $185 = HEAP32[$184>>2]|0;
       $186 = ($185|0)==(0|0);
       if ($186) {
        $187 = (($has_serifs$049$i) + 1)|0;
        $anchor$6$i = $anchor$450$i;$has_serifs$1$i = $187;
        break;
       }
       $188 = ((($185)) + 20|0);
       $189 = HEAP32[$188>>2]|0;
       $190 = ($189|0)==(0|0);
       if (!($190)) {
        $$idx27$i = ((($185)) + 4|0);
        $$idx27$val$i = HEAP32[$$idx27$i>>2]|0;
        $$idx28$i = ((($185)) + 8|0);
        $$idx28$val$i = HEAP32[$$idx28$i>>2]|0;
        $$idx29$i = ((($185)) + 12|0);
        $$idx29$val$i = HEAP8[$$idx29$i>>0]|0;
        $191 = ((($edge$151$i)) + 4|0);
        $192 = HEAP32[$191>>2]|0;
        $193 = (($192) - ($$idx27$val$i))|0;
        $194 = $$idx29$val$i&255;
        $195 = (_af_latin_compute_stem_width($hints,$dim$021,$193,$194,$181)|0);
        $196 = (($195) + ($$idx28$val$i))|0;
        $197 = ((($edge$151$i)) + 8|0);
        HEAP32[$197>>2] = $196;
        $198 = $181 | 4;
        $199 = $198&255;
        HEAP8[$179>>0] = $199;
        $anchor$6$i = $anchor$450$i;$has_serifs$1$i = $has_serifs$049$i;
        break;
       }
       $200 = ($anchor$450$i|0)==(0|0);
       if ($200) {
        $201 = ((($185)) + 4|0);
        $202 = HEAP32[$201>>2]|0;
        $203 = ((($edge$151$i)) + 4|0);
        $204 = HEAP32[$203>>2]|0;
        $205 = (($202) - ($204))|0;
        $206 = ((($185)) + 12|0);
        $207 = HEAP8[$206>>0]|0;
        $208 = $207&255;
        $209 = (_af_latin_compute_stem_width($hints,$dim$021,$205,$181,$208)|0);
        $210 = ($209|0)<(65);
        $211 = ($209|0)<(96);
        if ($211) {
         $$12$i = $210 ? 32 : 26;
         $$11$i = $210 ? 32 : 38;
         $212 = $205 >> 1;
         $213 = (($212) + ($204))|0;
         $214 = (($213) + 32)|0;
         $215 = $214 & -64;
         $216 = (($215) - ($$11$i))|0;
         $217 = (($213) - ($216))|0;
         $218 = ($217|0)<(0);
         $219 = (0 - ($217))|0;
         $$13$i = $218 ? $219 : $217;
         $220 = $215 | $$12$i;
         $221 = (($213) - ($220))|0;
         $222 = ($221|0)<(0);
         $223 = (0 - ($221))|0;
         $error2$0$i = $222 ? $223 : $221;
         $224 = ($$13$i|0)<($error2$0$i|0);
         $$14$i = $224 ? $216 : $220;
         $225 = (($209|0) / 2)&-1;
         $226 = (($$14$i) - ($225))|0;
         $227 = ((($edge$151$i)) + 8|0);
         HEAP32[$227>>2] = $226;
         $228 = (($226) + ($209))|0;
         $229 = ((($185)) + 8|0);
         HEAP32[$229>>2] = $228;
         $edge$1$idx25$val$pre$i = HEAP32[$227>>2]|0;
         $$pre$phi$iZ2D = $229;$edge$1$idx25$val$i = $edge$1$idx25$val$pre$i;
        } else {
         $230 = (($204) + 32)|0;
         $231 = $230 & -64;
         $232 = ((($edge$151$i)) + 8|0);
         HEAP32[$232>>2] = $231;
         $$pre65$i = ((($185)) + 8|0);
         $$pre$phi$iZ2D = $$pre65$i;$edge$1$idx25$val$i = $231;
        }
        $233 = $181 | 4;
        $234 = $233&255;
        HEAP8[$179>>0] = $234;
        $235 = HEAP8[$206>>0]|0;
        $236 = $235&255;
        $237 = (_af_latin_compute_stem_width($hints,$dim$021,$205,$233,$236)|0);
        $238 = (($237) + ($edge$1$idx25$val$i))|0;
        HEAP32[$$pre$phi$iZ2D>>2] = $238;
        $anchor$6$i = $edge$151$i;$has_serifs$1$i = $has_serifs$049$i;
        break;
       }
       $239 = ((($anchor$450$i)) + 8|0);
       $240 = HEAP32[$239>>2]|0;
       $241 = ((($edge$151$i)) + 4|0);
       $242 = HEAP32[$241>>2]|0;
       $243 = ((($anchor$450$i)) + 4|0);
       $244 = HEAP32[$243>>2]|0;
       $245 = (($242) - ($244))|0;
       $246 = (($245) + ($240))|0;
       $247 = ((($185)) + 4|0);
       $248 = HEAP32[$247>>2]|0;
       $249 = (($248) - ($242))|0;
       $250 = $249 >> 1;
       $251 = (($250) + ($246))|0;
       $252 = ((($185)) + 12|0);
       $253 = HEAP8[$252>>0]|0;
       $254 = $253&255;
       $255 = (_af_latin_compute_stem_width($hints,$dim$021,$249,$181,$254)|0);
       $256 = $254 & 4;
       $257 = ($256|0)==(0);
       do {
        if ($257) {
         $262 = ($255|0)<(96);
         if ($262) {
          $263 = (($251) + 32)|0;
          $264 = $263 & -64;
          $265 = ($255|0)<(65);
          $$15$i = $265 ? 32 : 38;
          $$16$i = $265 ? 32 : 26;
          $266 = (($264) - ($$15$i))|0;
          $267 = (($251) - ($266))|0;
          $268 = ($267|0)<(0);
          $269 = (0 - ($267))|0;
          $delta1$0$i = $268 ? $269 : $267;
          $270 = $264 | $$16$i;
          $271 = (($251) - ($270))|0;
          $272 = ($271|0)<(0);
          $273 = (0 - ($271))|0;
          $$17$i = $272 ? $273 : $271;
          $274 = ($delta1$0$i|0)<($$17$i|0);
          $cur_pos15$0$i = $274 ? $266 : $270;
          $275 = (($255|0) / 2)&-1;
          $276 = (($cur_pos15$0$i) - ($275))|0;
          $277 = ((($edge$151$i)) + 8|0);
          HEAP32[$277>>2] = $276;
          $278 = (($cur_pos15$0$i) + ($275))|0;
          $279 = ((($185)) + 8|0);
          HEAP32[$279>>2] = $278;
          break;
         } else {
          $280 = (($246) + 32)|0;
          $281 = $280 & -64;
          $282 = $255 >> 1;
          $283 = (($281) + ($282))|0;
          $284 = (($283) - ($251))|0;
          $285 = ($284|0)<(0);
          $286 = (0 - ($284))|0;
          $$18$i = $285 ? $286 : $284;
          $287 = (($249) + 32)|0;
          $288 = (($287) + ($246))|0;
          $289 = $288 & -64;
          $290 = (($289) - ($255))|0;
          $291 = (($282) - ($251))|0;
          $292 = (($290) + ($291))|0;
          $293 = ($292|0)<(0);
          $294 = (0 - ($292))|0;
          $delta2$1$i = $293 ? $294 : $292;
          $295 = ($$18$i|0)<($delta2$1$i|0);
          $296 = $295 ? $281 : $290;
          $297 = ((($edge$151$i)) + 8|0);
          HEAP32[$297>>2] = $296;
          $298 = (($296) + ($255))|0;
          $299 = ((($185)) + 8|0);
          HEAP32[$299>>2] = $298;
          break;
         }
        } else {
         $258 = ((($185)) + 8|0);
         $259 = HEAP32[$258>>2]|0;
         $260 = (($259) - ($255))|0;
         $261 = ((($edge$151$i)) + 8|0);
         HEAP32[$261>>2] = $260;
        }
       } while(0);
       $300 = $181 | 4;
       $301 = $300&255;
       HEAP8[$179>>0] = $301;
       $302 = HEAP8[$252>>0]|0;
       $303 = $302&255;
       $304 = $303 | 4;
       $305 = $304&255;
       HEAP8[$252>>0] = $305;
       $306 = ($edge$151$i>>>0)>($118>>>0);
       if ($306) {
        $307 = ((($edge$151$i)) + 8|0);
        $308 = HEAP32[$307>>2]|0;
        $309 = ((($edge$151$i)) + -40|0);
        $310 = HEAP32[$309>>2]|0;
        $311 = ($308|0)<($310|0);
        if ($311) {
         HEAP32[$307>>2] = $310;
         $anchor$6$i = $anchor$450$i;$has_serifs$1$i = $has_serifs$049$i;
        } else {
         $anchor$6$i = $anchor$450$i;$has_serifs$1$i = $has_serifs$049$i;
        }
       } else {
        $anchor$6$i = $anchor$450$i;$has_serifs$1$i = $has_serifs$049$i;
       }
      } else {
       $anchor$6$i = $anchor$450$i;$has_serifs$1$i = $has_serifs$049$i;
      }
     } while(0);
     $312 = ((($edge$151$i)) + 48|0);
     $313 = ($312>>>0)<($121>>>0);
     if ($313) {
      $anchor$450$i = $anchor$6$i;$edge$151$i = $312;$has_serifs$049$i = $has_serifs$1$i;
     } else {
      $anchor$4$lcssa$i = $anchor$6$i;$has_serifs$0$lcssa$i = $has_serifs$1$i;
      break;
     }
    }
   } else {
    $anchor$4$lcssa$i = $anchor$4$ph$i;$has_serifs$0$lcssa$i = 0;
   }
   $$idx$i = ($120*48)|0;
   L100: do {
    if ($93) {
     $314 = ($$idx$i|0)==(576);
     switch ($$idx$i|0) {
     case 288:  {
      $315 = ((($118)) + 96|0);
      $316 = ((($118)) + 192|0);
      $edge19$0$i = $118;$edge210$0$i = $315;$edge3$0$i = $316;
      break;
     }
     case 576:  {
      $317 = ((($118)) + 48|0);
      $318 = ((($118)) + 240|0);
      $319 = ((($118)) + 432|0);
      $edge19$0$i = $317;$edge210$0$i = $318;$edge3$0$i = $319;
      break;
     }
     default: {
      break L100;
     }
     }
     $320 = ((($edge210$0$i)) + 4|0);
     $321 = HEAP32[$320>>2]|0;
     $322 = ((($edge19$0$i)) + 4|0);
     $323 = HEAP32[$322>>2]|0;
     $324 = (($321) - ($323))|0;
     $325 = ((($edge3$0$i)) + 4|0);
     $326 = HEAP32[$325>>2]|0;
     $327 = (($321) - ($326))|0;
     $328 = (($327) + ($324))|0;
     $329 = ($328|0)<(0);
     $330 = (0 - ($328))|0;
     $$19$i = $329 ? $330 : $328;
     $331 = ($$19$i|0)<(8);
     if ($331) {
      $332 = ((($edge3$0$i)) + 8|0);
      $333 = HEAP32[$332>>2]|0;
      $334 = ((($edge210$0$i)) + 8|0);
      $335 = HEAP32[$334>>2]|0;
      $336 = $335 << 1;
      $337 = ((($edge19$0$i)) + 8|0);
      $338 = HEAP32[$337>>2]|0;
      $339 = (($336) - ($338))|0;
      $340 = (($333) - ($339))|0;
      HEAP32[$332>>2] = $339;
      $341 = ((($edge3$0$i)) + 24|0);
      $342 = HEAP32[$341>>2]|0;
      $343 = ($342|0)==(0|0);
      if (!($343)) {
       $344 = ((($342)) + 8|0);
       $345 = HEAP32[$344>>2]|0;
       $346 = (($345) - ($340))|0;
       HEAP32[$344>>2] = $346;
      }
      if ($314) {
       $347 = ((($118)) + 392|0);
       $348 = HEAP32[$347>>2]|0;
       $349 = (($348) - ($340))|0;
       HEAP32[$347>>2] = $349;
       $350 = ((($118)) + 536|0);
       $351 = HEAP32[$350>>2]|0;
       $352 = (($351) - ($340))|0;
       HEAP32[$350>>2] = $352;
      }
      $353 = ((($edge3$0$i)) + 12|0);
      $354 = HEAP8[$353>>0]|0;
      $355 = $354&255;
      $356 = $355 | 4;
      $357 = $356&255;
      HEAP8[$353>>0] = $357;
      if (!($343)) {
       $358 = ((($342)) + 12|0);
       $359 = HEAP8[$358>>0]|0;
       $360 = $359&255;
       $361 = $360 | 4;
       $362 = $361&255;
       HEAP8[$358>>0] = $362;
      }
     }
    }
   } while(0);
   $notlhs$i9 = ($has_serifs$0$lcssa$i|0)!=(0);
   $notrhs$i10 = ($anchor$4$lcssa$i|0)==(0|0);
   $or$cond6$not$i = $notrhs$i10 | $notlhs$i9;
   $or$cond61$i = $123 & $or$cond6$not$i;
   if ($or$cond61$i) {
    $anchor$745$i = $anchor$4$lcssa$i;$edge$246$i = $118;
    while(1) {
     $363 = ((($edge$246$i)) + 12|0);
     $364 = HEAP8[$363>>0]|0;
     $365 = $364 & 4;
     $366 = ($365<<24>>24)==(0);
     if ($366) {
      $367 = ((($edge$246$i)) + 28|0);
      $368 = HEAP32[$367>>2]|0;
      $369 = ($368|0)==(0|0);
      if ($369) {
       label = 83;
      } else {
       $370 = ((($368)) + 4|0);
       $371 = HEAP32[$370>>2]|0;
       $372 = ((($edge$246$i)) + 4|0);
       $373 = HEAP32[$372>>2]|0;
       $374 = (($371) - ($373))|0;
       $375 = ($374|0)<(0);
       $376 = (0 - ($374))|0;
       $$20$i = $375 ? $376 : $374;
       $377 = ($$20$i|0)<(80);
       if ($377) {
        $$idx33$i = ((($368)) + 8|0);
        $$idx33$val$i = HEAP32[$$idx33$i>>2]|0;
        $378 = (($$idx33$val$i) - ($371))|0;
        $379 = (($378) + ($373))|0;
        $380 = ((($edge$246$i)) + 8|0);
        HEAP32[$380>>2] = $379;
        $429 = $364;$436 = $379;$anchor$8$i = $anchor$745$i;
       } else {
        label = 83;
       }
      }
      do {
       if ((label|0) == 83) {
        label = 0;
        $381 = ($anchor$745$i|0)==(0|0);
        if ($381) {
         $382 = ((($edge$246$i)) + 4|0);
         $383 = HEAP32[$382>>2]|0;
         $384 = (($383) + 32)|0;
         $385 = $384 & -64;
         $386 = ((($edge$246$i)) + 8|0);
         HEAP32[$386>>2] = $385;
         $429 = $364;$436 = $385;$anchor$8$i = $edge$246$i;
         break;
        } else {
         $edge$2$pn$i = $edge$246$i;
        }
        while(1) {
         $before$0$i = ((($edge$2$pn$i)) + -48|0);
         $387 = ($before$0$i>>>0)<($118>>>0);
         if ($387) {
          $$lcssa$i = 0;$before$0$i$lcssa = $before$0$i;$edge$2$pn$i$lcssa = $edge$2$pn$i;
          break;
         }
         $388 = ((($edge$2$pn$i)) + -36|0);
         $389 = HEAP8[$388>>0]|0;
         $390 = $389 & 4;
         $391 = ($390<<24>>24)==(0);
         if ($391) {
          $edge$2$pn$i = $before$0$i;
         } else {
          $$lcssa$i = 1;$before$0$i$lcssa = $before$0$i;$edge$2$pn$i$lcssa = $edge$2$pn$i;
          break;
         }
        }
        $edge$2$pn8$i = $edge$246$i;
        while(1) {
         $after$0$i = ((($edge$2$pn8$i)) + 48|0);
         $392 = ($after$0$i>>>0)<($121>>>0);
         if (!($392)) {
          break;
         }
         $393 = ((($edge$2$pn8$i)) + 60|0);
         $394 = HEAP8[$393>>0]|0;
         $395 = $394 & 4;
         $396 = ($395<<24>>24)==(0);
         if ($396) {
          $edge$2$pn8$i = $after$0$i;
         } else {
          $after$0$i$lcssa50 = $after$0$i;$edge$2$pn8$i$lcssa49 = $edge$2$pn8$i;
          label = 90;
          break;
         }
        }
        if ((label|0) == 90) {
         label = 0;
         $397 = ($before$0$i$lcssa>>>0)<($edge$246$i>>>0);
         $brmerge23$demorgan$i = $397 & $$lcssa$i;
         $398 = ($after$0$i$lcssa50>>>0)>($edge$246$i>>>0);
         $or$cond24$i = $brmerge23$demorgan$i & $398;
         if ($or$cond24$i) {
          $399 = ((($edge$2$pn8$i$lcssa49)) + 52|0);
          $400 = HEAP32[$399>>2]|0;
          $401 = ((($edge$2$pn$i$lcssa)) + -44|0);
          $402 = HEAP32[$401>>2]|0;
          $403 = ($400|0)==($402|0);
          $404 = ((($edge$2$pn$i$lcssa)) + -40|0);
          $405 = HEAP32[$404>>2]|0;
          if ($403) {
           $406 = ((($edge$246$i)) + 8|0);
           HEAP32[$406>>2] = $405;
           $429 = $364;$436 = $405;$anchor$8$i = $anchor$745$i;
           break;
          } else {
           $407 = ((($edge$246$i)) + 4|0);
           $408 = HEAP32[$407>>2]|0;
           $409 = (($408) - ($402))|0;
           $410 = ((($edge$2$pn8$i$lcssa49)) + 56|0);
           $411 = HEAP32[$410>>2]|0;
           $412 = (($411) - ($405))|0;
           $413 = (($400) - ($402))|0;
           $414 = (_FT_MulDiv($409,$412,$413)|0);
           $415 = (($414) + ($405))|0;
           $416 = ((($edge$246$i)) + 8|0);
           HEAP32[$416>>2] = $415;
           $$pre64$i = HEAP8[$363>>0]|0;
           $429 = $$pre64$i;$436 = $415;$anchor$8$i = $anchor$745$i;
           break;
          }
         }
        }
        $417 = ((($anchor$745$i)) + 8|0);
        $418 = HEAP32[$417>>2]|0;
        $419 = ((($edge$246$i)) + 4|0);
        $420 = HEAP32[$419>>2]|0;
        $421 = ((($anchor$745$i)) + 4|0);
        $422 = HEAP32[$421>>2]|0;
        $423 = (($420) + 16)|0;
        $424 = (($423) - ($422))|0;
        $425 = $424 & -32;
        $426 = (($425) + ($418))|0;
        $427 = ((($edge$246$i)) + 8|0);
        HEAP32[$427>>2] = $426;
        $429 = $364;$436 = $426;$anchor$8$i = $anchor$745$i;
       }
      } while(0);
      $428 = $429&255;
      $430 = $428 | 4;
      $431 = $430&255;
      HEAP8[$363>>0] = $431;
      $432 = ($edge$246$i>>>0)>($118>>>0);
      if ($432) {
       $433 = ((($edge$246$i)) + -40|0);
       $434 = HEAP32[$433>>2]|0;
       $435 = ($436|0)<($434|0);
       if ($435) {
        $437 = ((($edge$246$i)) + 8|0);
        HEAP32[$437>>2] = $434;
        $447 = $434;
       } else {
        $447 = $436;
       }
      } else {
       $447 = $436;
      }
      $438 = ((($edge$246$i)) + 48|0);
      $439 = ($438>>>0)<($121>>>0);
      if ($439) {
       $440 = ((($edge$246$i)) + 60|0);
       $441 = HEAP8[$440>>0]|0;
       $442 = $441 & 4;
       $443 = ($442<<24>>24)==(0);
       if ($443) {
        $$pre$phi67$iZ2D = $438;$anchor$9$i = $anchor$8$i;
       } else {
        $444 = ((($edge$246$i)) + 56|0);
        $445 = HEAP32[$444>>2]|0;
        $446 = ($447|0)>($445|0);
        if ($446) {
         $448 = ((($edge$246$i)) + 8|0);
         HEAP32[$448>>2] = $445;
         $$pre$phi67$iZ2D = $438;$anchor$9$i = $anchor$8$i;
        } else {
         $$pre$phi67$iZ2D = $438;$anchor$9$i = $anchor$8$i;
        }
       }
      } else {
       $$pre$phi67$iZ2D = $438;$anchor$9$i = $anchor$8$i;
      }
     } else {
      $$pre66$i = ((($edge$246$i)) + 48|0);
      $$pre$phi67$iZ2D = $$pre66$i;$anchor$9$i = $anchor$745$i;
     }
     $449 = ($$pre$phi67$iZ2D>>>0)<($121>>>0);
     if ($449) {
      $anchor$745$i = $anchor$9$i;$edge$246$i = $$pre$phi67$iZ2D;
     } else {
      break;
     }
    }
   }
   $450 = (((((($hints)) + 44|0) + (($dim$021*1468)|0)|0)) + 8|0);
   $451 = HEAP32[$450>>2]|0;
   $452 = (((($hints)) + 44|0) + (($dim$021*1468)|0)|0);
   $453 = HEAP32[$452>>2]|0;
   $454 = (($451) + (($453*48)|0)|0);
   $455 = ($453|0)>(0);
   if ($93) {
    if ($455) {
     $seg$06$i = $451;
     while(1) {
      $456 = ((($seg$06$i)) + 12|0);
      $457 = HEAP32[$456>>2]|0;
      $458 = ($457|0)==(0|0);
      if (!($458)) {
       $459 = ((($seg$06$i)) + 40|0);
       $460 = ((($seg$06$i)) + 44|0);
       $461 = HEAP32[$460>>2]|0;
       $462 = ((($457)) + 8|0);
       $463 = HEAP32[$462>>2]|0;
       $point$0$in$i = $459;
       while(1) {
        $point$0$i = HEAP32[$point$0$in$i>>2]|0;
        $464 = ((($point$0$i)) + 16|0);
        HEAP32[$464>>2] = $463;
        $465 = HEAP16[$point$0$i>>1]|0;
        $466 = $465&65535;
        $467 = $466 | 4;
        $468 = $467&65535;
        HEAP16[$point$0$i>>1] = $468;
        $469 = ($point$0$i|0)==($461|0);
        $470 = ((($point$0$i)) + 32|0);
        if ($469) {
         break;
        } else {
         $point$0$in$i = $470;
        }
       }
      }
      $471 = ((($seg$06$i)) + 48|0);
      $472 = ($471>>>0)<($454>>>0);
      if ($472) {
       $seg$06$i = $471;
      } else {
       break;
      }
     }
    }
   } else {
    if ($455) {
     $seg$17$i = $451;
     while(1) {
      $473 = ((($seg$17$i)) + 12|0);
      $474 = HEAP32[$473>>2]|0;
      $475 = ($474|0)==(0|0);
      if (!($475)) {
       $476 = ((($seg$17$i)) + 40|0);
       $477 = ((($seg$17$i)) + 44|0);
       $478 = HEAP32[$477>>2]|0;
       $479 = ((($474)) + 8|0);
       $480 = HEAP32[$479>>2]|0;
       $point2$0$in$i = $476;
       while(1) {
        $point2$0$i = HEAP32[$point2$0$in$i>>2]|0;
        $481 = ((($point2$0$i)) + 20|0);
        HEAP32[$481>>2] = $480;
        $482 = HEAP16[$point2$0$i>>1]|0;
        $483 = $482&65535;
        $484 = $483 | 8;
        $485 = $484&65535;
        HEAP16[$point2$0$i>>1] = $485;
        $486 = ($point2$0$i|0)==($478|0);
        $487 = ((($point2$0$i)) + 32|0);
        if ($486) {
         break;
        } else {
         $point2$0$in$i = $487;
        }
       }
      }
      $488 = ((($seg$17$i)) + 48|0);
      $489 = ($488>>>0)<($454>>>0);
      if ($489) {
       $seg$17$i = $488;
      } else {
       break;
      }
     }
    }
   }
   _af_glyph_hints_align_strong_points($hints,$dim$021);
   _af_glyph_hints_align_weak_points($hints,$dim$021);
  }
  $490 = (($dim$021) + 1)|0;
  $exitcond = ($490|0)==(2);
  if ($exitcond) {
   break;
  } else {
   $dim$021 = $490;
  }
 }
 $hints$idx$val = HEAP32[$hints$idx3>>2]|0;
 $hints$idx1$val = HEAP32[$hints$idx4>>2]|0;
 $491 = (($hints$idx1$val) + (($hints$idx$val*40)|0)|0);
 $492 = ($hints$idx$val|0)>(0);
 if (!($492)) {
  $error$2 = 0;
  STACKTOP = sp;return ($error$2|0);
 }
 $outline$idx2 = ((($outline)) + 8|0);
 $outline$idx2$val = HEAP32[$outline$idx2>>2]|0;
 $outline$idx = ((($outline)) + 4|0);
 $outline$idx$val = HEAP32[$outline$idx>>2]|0;
 $point$01$i = $hints$idx1$val;$tag$03$i = $outline$idx2$val;$vec$02$i = $outline$idx$val;
 while(1) {
  $493 = ((($point$01$i)) + 16|0);
  $494 = HEAP32[$493>>2]|0;
  HEAP32[$vec$02$i>>2] = $494;
  $495 = ((($point$01$i)) + 20|0);
  $496 = HEAP32[$495>>2]|0;
  $497 = ((($vec$02$i)) + 4|0);
  HEAP32[$497>>2] = $496;
  $498 = HEAP16[$point$01$i>>1]|0;
  $499 = $498&65535;
  $500 = $499 & 1;
  $501 = ($500|0)==(0);
  do {
   if ($501) {
    $502 = $499 & 2;
    $503 = ($502|0)==(0);
    if ($503) {
     HEAP8[$tag$03$i>>0] = 1;
     break;
    } else {
     HEAP8[$tag$03$i>>0] = 2;
     break;
    }
   } else {
    HEAP8[$tag$03$i>>0] = 0;
   }
  } while(0);
  $504 = ((($point$01$i)) + 40|0);
  $505 = ((($vec$02$i)) + 8|0);
  $506 = ((($tag$03$i)) + 1|0);
  $507 = ($504>>>0)<($491>>>0);
  if ($507) {
   $point$01$i = $504;$tag$03$i = $506;$vec$02$i = $505;
  } else {
   $error$2 = 0;
   break;
  }
 }
 STACKTOP = sp;return ($error$2|0);
}
function _af_latin_hints_detect_features($hints,$width_count,$widths,$dim) {
 $hints = $hints|0;
 $width_count = $width_count|0;
 $widths = $widths|0;
 $dim = $dim|0;
 var $$$i = 0, $$3$i = 0, $$4$i = 0, $$5$i = 0, $$6$i = 0, $$7$i = 0, $$8$i = 0, $$in$i = 0, $$lcssa = 0, $$not$i = 0, $$phi$trans$insert$i = 0, $$pre$i = 0, $$pre$phi$iZ2D = 0, $$pre$phi33$iZ2D = 0, $$pre$phi35$iZ2D = 0, $$pre32$i = 0, $$pre34$i = 0, $0 = 0, $1 = 0, $10 = 0;
 var $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0;
 var $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0;
 var $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $brmerge$i = 0, $edge1$i = 0, $edge2$0$i = 0, $edge2$2$i = 0, $edge4$020$i = 0, $edge4$119$i = 0, $ee$022$i = 0, $ee$022$i$lcssa = 0, $error$0 = 0, $is_round$0$i = 0, $is_round$1$i = 0, $is_round$1$i$lcssa = 0, $is_straight$0$i = 0, $is_straight$1$i = 0;
 var $is_straight$1$i$lcssa = 0, $not$$i = 0, $or$cond$i = 0, $or$cond30$i = 0, $seg$024$i = 0, $seg$1$i = 0, $seg$2$i = 0, $seg2$0$i = 0, $seg_delta$0$i = 0, $segment_length_threshold$0$i = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $edge1$i = sp;
 $0 = (_af_latin_hints_compute_segments($hints,$dim)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  $error$0 = $0;
  STACKTOP = sp;return ($error$0|0);
 }
 _af_latin_hints_link_segments($hints,$width_count,$widths,$dim);
 $2 = (((($hints)) + 44|0) + (($dim*1468)|0)|0);
 $3 = HEAP32[$hints>>2]|0;
 $4 = ((($hints)) + 2988|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = HEAP32[$2>>2]|0;
 $9 = (($7) + (($8*48)|0)|0);
 $10 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 12|0);
 HEAP32[$10>>2] = 0;
 $11 = ($dim|0)==(0);
 $12 = ((($hints)) + 4|0);
 $13 = ((($hints)) + 12|0);
 $$in$i = $11 ? $12 : $13;
 $14 = HEAP32[$$in$i>>2]|0;
 if ($11) {
  $15 = HEAP32[$13>>2]|0;
  $16 = (_FT_DivFix(64,$15)|0);
  $segment_length_threshold$0$i = $16;
 } else {
  $segment_length_threshold$0$i = 0;
 }
 $17 = (((((($5)) + 44|0) + (($dim*1572)|0)|0)) + 204|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = (_FT_MulFix($18,$14)|0);
 $20 = ($19|0)>(16);
 $$$i = $20 ? 16 : $19;
 $21 = (_FT_DivFix($$$i,$14)|0);
 $22 = ($8|0)>(0);
 L7: do {
  if ($22) {
   $23 = ($segment_length_threshold$0$i*3)|0;
   $24 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 20|0);
   $seg$024$i = $7;
   L9: while(1) {
    $25 = ((($seg$024$i)) + 8|0);
    $26 = HEAP16[$25>>1]|0;
    $27 = $26 << 16 >> 16;
    $28 = ($27|0)<($segment_length_threshold$0$i|0);
    L11: do {
     if (!($28)) {
      $29 = ((($seg$024$i)) + 24|0);
      $30 = HEAP32[$29>>2]|0;
      $31 = ($30|0)!=(0|0);
      $32 = $27 << 1;
      $33 = ($32|0)<($23|0);
      $or$cond30$i = $33 & $31;
      if (!($or$cond30$i)) {
       $34 = HEAP32[$10>>2]|0;
       $35 = ($34|0)>(0);
       L14: do {
        if ($35) {
         $36 = HEAP32[$24>>2]|0;
         $37 = ((($seg$024$i)) + 2|0);
         $38 = HEAP16[$37>>1]|0;
         $39 = $38 << 16 >> 16;
         $40 = ((($seg$024$i)) + 1|0);
         $ee$022$i = 0;
         while(1) {
          $41 = (($36) + (($ee$022$i*48)|0)|0);
          $42 = HEAP16[$41>>1]|0;
          $43 = $42 << 16 >> 16;
          $44 = (($39) - ($43))|0;
          $45 = ($44|0)<(0);
          $46 = (0 - ($44))|0;
          $$3$i = $45 ? $46 : $44;
          $47 = ($$3$i|0)<($21|0);
          if ($47) {
           $48 = (((($36) + (($ee$022$i*48)|0)|0)) + 13|0);
           $49 = HEAP8[$48>>0]|0;
           $50 = HEAP8[$40>>0]|0;
           $51 = ($49<<24>>24)==($50<<24>>24);
           if ($51) {
            $ee$022$i$lcssa = $ee$022$i;
            break;
           }
          }
          $52 = (($ee$022$i) + 1)|0;
          $53 = ($52|0)<($34|0);
          if ($53) {
           $ee$022$i = $52;
          } else {
           $$pre$phi$iZ2D = $37;$$pre$phi33$iZ2D = $40;$55 = $38;
           break L14;
          }
         }
         $71 = (((($36) + (($ee$022$i$lcssa*48)|0)|0)) + 40|0);
         $72 = HEAP32[$71>>2]|0;
         $73 = ((($seg$024$i)) + 16|0);
         HEAP32[$73>>2] = $72;
         $74 = (((($36) + (($ee$022$i$lcssa*48)|0)|0)) + 44|0);
         $75 = HEAP32[$74>>2]|0;
         $76 = ((($75)) + 16|0);
         HEAP32[$76>>2] = $seg$024$i;
         HEAP32[$74>>2] = $seg$024$i;
         break L11;
        } else {
         $$phi$trans$insert$i = ((($seg$024$i)) + 2|0);
         $$pre$i = HEAP16[$$phi$trans$insert$i>>1]|0;
         $$pre32$i = ((($seg$024$i)) + 1|0);
         $$pre$phi$iZ2D = $$phi$trans$insert$i;$$pre$phi33$iZ2D = $$pre32$i;$55 = $$pre$i;
        }
       } while(0);
       $54 = $55 << 16 >> 16;
       $56 = HEAP8[$$pre$phi33$iZ2D>>0]|0;
       $57 = $56 << 24 >> 24;
       $58 = (_af_axis_hints_new_edge($2,$54,$57,$3,$edge1$i)|0);
       $59 = ($58|0)==(0);
       if (!($59)) {
        $$lcssa = $58;
        break L9;
       }
       $60 = HEAP32[$edge1$i>>2]|0;
       dest=$60; stop=dest+40|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
       $61 = ((($60)) + 40|0);
       HEAP32[$61>>2] = $seg$024$i;
       $62 = ((($60)) + 44|0);
       HEAP32[$62>>2] = $seg$024$i;
       $63 = HEAP8[$$pre$phi33$iZ2D>>0]|0;
       $64 = ((($60)) + 13|0);
       HEAP8[$64>>0] = $63;
       $65 = HEAP16[$$pre$phi$iZ2D>>1]|0;
       HEAP16[$60>>1] = $65;
       $66 = $65 << 16 >> 16;
       $67 = (_FT_MulFix($66,$14)|0);
       $68 = ((($60)) + 4|0);
       HEAP32[$68>>2] = $67;
       $69 = ((($60)) + 8|0);
       HEAP32[$69>>2] = $67;
       $70 = ((($seg$024$i)) + 16|0);
       HEAP32[$70>>2] = $seg$024$i;
      }
     }
    } while(0);
    $77 = ((($seg$024$i)) + 48|0);
    $78 = ($77>>>0)<($9>>>0);
    if ($78) {
     $seg$024$i = $77;
    } else {
     $$pre$phi35$iZ2D = $24;
     break L7;
    }
   }
   $error$0 = $$lcssa;
   STACKTOP = sp;return ($error$0|0);
  } else {
   $$pre34$i = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 20|0);
   $$pre$phi35$iZ2D = $$pre34$i;
  }
 } while(0);
 $79 = HEAP32[$$pre$phi35$iZ2D>>2]|0;
 $80 = HEAP32[$10>>2]|0;
 $81 = (($79) + (($80*48)|0)|0);
 $82 = ($80|0)>(0);
 if ($82) {
  $edge4$020$i = $79;
 } else {
  $error$0 = 0;
  STACKTOP = sp;return ($error$0|0);
 }
 while(1) {
  $83 = ((($edge4$020$i)) + 40|0);
  $84 = HEAP32[$83>>2]|0;
  $85 = ($84|0)==(0|0);
  if (!($85)) {
   $seg$1$i = $84;
   while(1) {
    $86 = ((($seg$1$i)) + 12|0);
    HEAP32[$86>>2] = $edge4$020$i;
    $87 = ((($seg$1$i)) + 16|0);
    $88 = HEAP32[$87>>2]|0;
    $89 = ($88|0)==($84|0);
    if ($89) {
     break;
    } else {
     $seg$1$i = $88;
    }
   }
  }
  $90 = ((($edge4$020$i)) + 48|0);
  $91 = ($90>>>0)<($81>>>0);
  if ($91) {
   $edge4$020$i = $90;
  } else {
   $edge4$119$i = $79;
   break;
  }
 }
 while(1) {
  $92 = ((($edge4$119$i)) + 40|0);
  $93 = HEAP32[$92>>2]|0;
  $94 = ((($edge4$119$i)) + 24|0);
  $95 = ((($edge4$119$i)) + 28|0);
  $is_round$0$i = 0;$is_straight$0$i = 0;$seg$2$i = $93;
  while(1) {
   $96 = HEAP32[$seg$2$i>>2]|0;
   $97 = $96 & 1;
   $98 = $97 ^ 1;
   $is_round$1$i = (($97) + ($is_round$0$i))|0;
   $is_straight$1$i = (($98) + ($is_straight$0$i))|0;
   $99 = ((($seg$2$i)) + 24|0);
   $100 = HEAP32[$99>>2]|0;
   $101 = ($100|0)==(0|0);
   if ($101) {
    $110 = 0;
   } else {
    $102 = ((($100)) + 12|0);
    $103 = HEAP32[$102>>2]|0;
    $104 = ($103|0)!=($edge4$119$i|0);
    $not$$i = ($103|0)!=(0|0);
    $$4$i = $104 & $not$$i;
    $110 = $$4$i;
   }
   $105 = ((($seg$2$i)) + 20|0);
   $106 = HEAP32[$105>>2]|0;
   $107 = ($106|0)==(0|0);
   if ($107) {
    if ($110) {
     label = 30;
    }
   } else {
    $108 = ((($106)) + 12|0);
    $109 = HEAP32[$108>>2]|0;
    $$not$i = ($109|0)!=(0|0);
    $brmerge$i = $110 | $$not$i;
    if ($brmerge$i) {
     label = 30;
    }
   }
   do {
    if ((label|0) == 30) {
     label = 0;
     $$5$i = $110 ? $95 : $94;
     $$6$i = $110 ? $99 : $105;
     $seg2$0$i = HEAP32[$$6$i>>2]|0;
     $edge2$0$i = HEAP32[$$5$i>>2]|0;
     $111 = ($edge2$0$i|0)==(0|0);
     if ($111) {
      $129 = ((($seg2$0$i)) + 12|0);
      $130 = HEAP32[$129>>2]|0;
      $edge2$2$i = $130;
     } else {
      $112 = HEAP16[$edge4$119$i>>1]|0;
      $113 = $112 << 16 >> 16;
      $114 = HEAP16[$edge2$0$i>>1]|0;
      $115 = $114 << 16 >> 16;
      $116 = (($113) - ($115))|0;
      $117 = ($116|0)<(0);
      $118 = (0 - ($116))|0;
      $$7$i = $117 ? $118 : $116;
      $119 = $96 >> 16;
      $120 = ((($seg2$0$i)) + 2|0);
      $121 = HEAP16[$120>>1]|0;
      $122 = $121 << 16 >> 16;
      $123 = (($119) - ($122))|0;
      $124 = ($123|0)<(0);
      $125 = (0 - ($123))|0;
      $seg_delta$0$i = $124 ? $125 : $123;
      $126 = ($seg_delta$0$i|0)<($$7$i|0);
      if ($126) {
       $127 = ((($seg2$0$i)) + 12|0);
       $128 = HEAP32[$127>>2]|0;
       $edge2$2$i = $128;
      } else {
       $edge2$2$i = $edge2$0$i;
      }
     }
     if ($110) {
      HEAP32[$95>>2] = $edge2$2$i;
      $131 = ((($edge2$2$i)) + 12|0);
      $132 = HEAP8[$131>>0]|0;
      $133 = $132&255;
      $134 = $133 | 2;
      $135 = $134&255;
      HEAP8[$131>>0] = $135;
      break;
     } else {
      HEAP32[$94>>2] = $edge2$2$i;
      break;
     }
    }
   } while(0);
   $136 = ((($seg$2$i)) + 16|0);
   $137 = HEAP32[$136>>2]|0;
   $138 = ($137|0)==($93|0);
   if ($138) {
    $is_round$1$i$lcssa = $is_round$1$i;$is_straight$1$i$lcssa = $is_straight$1$i;
    break;
   } else {
    $is_round$0$i = $is_round$1$i;$is_straight$0$i = $is_straight$1$i;$seg$2$i = $137;
   }
  }
  $139 = ((($edge4$119$i)) + 12|0);
  $140 = ($is_round$1$i$lcssa|0)<(1);
  $141 = ($is_round$1$i$lcssa|0)<($is_straight$1$i$lcssa|0);
  $or$cond$i = $140 | $141;
  $142 = $or$cond$i&1;
  $$8$i = $142 ^ 1;
  HEAP8[$139>>0] = $$8$i;
  $143 = HEAP32[$95>>2]|0;
  $144 = ($143|0)==(0|0);
  if (!($144)) {
   $145 = HEAP32[$94>>2]|0;
   $146 = ($145|0)==(0|0);
   if (!($146)) {
    HEAP32[$95>>2] = 0;
   }
  }
  $147 = ((($edge4$119$i)) + 48|0);
  $148 = ($147>>>0)<($81>>>0);
  if ($148) {
   $edge4$119$i = $147;
  } else {
   $error$0 = 0;
   break;
  }
 }
 STACKTOP = sp;return ($error$0|0);
}
function _af_axis_hints_new_edge($axis,$fpos,$dir,$memory,$anedge) {
 $axis = $axis|0;
 $fpos = $fpos|0;
 $dir = $dir|0;
 $memory = $memory|0;
 $anedge = $anedge|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $$pre10 = 0, $$pre8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $edge$0$lcssa = 0, $edge$0$lcssa$ph = 0, $edge$03 = 0, $edge$1 = 0, $error = 0, $new_max$0 = 0, $or$cond = 0, dest = 0, label = 0;
 var sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 HEAP32[$error>>2] = 0;
 $0 = ((($axis)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)<(12);
 do {
  if ($2) {
   $3 = ((($axis)) + 20|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==(0|0);
   if ($5) {
    $6 = ((($axis)) + 892|0);
    HEAP32[$3>>2] = $6;
    $7 = ((($axis)) + 16|0);
    HEAP32[$7>>2] = 12;
    $$pre$phiZ2D = $3;$31 = $1;
    label = 13;
   } else {
    $$pre$phiZ2D = $3;$31 = $1;
    label = 13;
   }
  } else {
   $8 = ((($axis)) + 16|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = ($1|0)<($9|0);
   if ($10) {
    $$pre10 = ((($axis)) + 20|0);
    $11 = HEAP32[$$pre10>>2]|0;
    $12 = (($11) + (($1*48)|0)|0);
    $42 = $11;$46 = $12;
    label = 14;
    break;
   }
   $13 = ($9|0)>(44739241);
   if ($13) {
    HEAP32[$error>>2] = 64;
    $45 = 64;$edge$1 = 0;
    HEAP32[$anedge>>2] = $edge$1;
    STACKTOP = sp;return ($45|0);
   }
   $14 = $9 >> 2;
   $15 = (($9) + 4)|0;
   $16 = (($15) + ($14))|0;
   $17 = ($14|0)<(-4);
   $18 = ($16|0)>(44739242);
   $or$cond = $17 | $18;
   $new_max$0 = $or$cond ? 44739242 : $16;
   $19 = ((($axis)) + 20|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = ((($axis)) + 892|0);
   $22 = ($20|0)==($21|0);
   do {
    if ($22) {
     $23 = (_ft_mem_realloc($memory,48,0,$new_max$0,0,$error)|0);
     HEAP32[$19>>2] = $23;
     $24 = HEAP32[$error>>2]|0;
     $25 = ($24|0)==(0);
     if ($25) {
      _memcpy(($23|0),($20|0),576)|0;
      break;
     } else {
      $45 = $24;$edge$1 = 0;
      HEAP32[$anedge>>2] = $edge$1;
      STACKTOP = sp;return ($45|0);
     }
    } else {
     $26 = (_ft_mem_realloc($memory,48,$9,$new_max$0,$20,$error)|0);
     HEAP32[$19>>2] = $26;
     $27 = HEAP32[$error>>2]|0;
     $28 = ($27|0)==(0);
     if (!($28)) {
      $45 = $27;$edge$1 = 0;
      HEAP32[$anedge>>2] = $edge$1;
      STACKTOP = sp;return ($45|0);
     }
    }
   } while(0);
   HEAP32[$8>>2] = $new_max$0;
   $$pre = HEAP32[$0>>2]|0;
   $$pre$phiZ2D = $19;$31 = $$pre;
   label = 13;
  }
 } while(0);
 if ((label|0) == 13) {
  $29 = HEAP32[$$pre$phiZ2D>>2]|0;
  $30 = (($29) + (($31*48)|0)|0);
  $32 = ($31|0)>(0);
  if ($32) {
   $42 = $29;$46 = $30;
   label = 14;
  } else {
   $44 = $31;$edge$0$lcssa = $30;
  }
 }
 if ((label|0) == 14) {
  $33 = ((($axis)) + 24|0);
  $edge$03 = $46;
  while(1) {
   $34 = ((($edge$03)) + -48|0);
   $35 = HEAP16[$34>>1]|0;
   $36 = $35 << 16 >> 16;
   $37 = ($36|0)<($fpos|0);
   if ($37) {
    $edge$0$lcssa$ph = $edge$03;
    break;
   }
   $38 = ($36|0)==($fpos|0);
   if ($38) {
    $39 = HEAP32[$33>>2]|0;
    $40 = ($39|0)==($dir|0);
    if ($40) {
     $edge$0$lcssa$ph = $edge$03;
     break;
    }
   }
   dest=$edge$03; src=$34; stop=dest+48|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
   $41 = ($34>>>0)>($42>>>0);
   if ($41) {
    $edge$03 = $34;
   } else {
    $edge$0$lcssa$ph = $34;
    break;
   }
  }
  $$pre8 = HEAP32[$0>>2]|0;
  $44 = $$pre8;$edge$0$lcssa = $edge$0$lcssa$ph;
 }
 $43 = (($44) + 1)|0;
 HEAP32[$0>>2] = $43;
 $45 = 0;$edge$1 = $edge$0$lcssa;
 HEAP32[$anedge>>2] = $edge$1;
 STACKTOP = sp;return ($45|0);
}
function _af_warper_compute($warper,$hints,$dim,$a_scale,$a_delta) {
 $warper = $warper|0;
 $hints = $hints|0;
 $dim = $dim|0;
 $a_scale = $a_scale|0;
 $a_delta = $a_delta|0;
 var $$ = 0, $$$i = 0, $$X1$0 = 0, $$X1$0$lcssa = 0, $$pn = 0, $$pre = 0, $$pre$i = 0, $$pre10 = 0, $$pre11 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $X1$04 = 0, $X2$03 = 0, $X2$1 = 0, $X2$1$lcssa = 0, $base_distort$0 = 0, $base_distort$1 = 0, $exitcond = 0, $exitcond10$i = 0, $idx$04$i = 0, $idx1$02$i = 0, $margin$0 = 0, $nn$05 = 0, $nn$16$i = 0, $or$cond$i = 0, $org_delta$0 = 0, $org_delta$0$in = 0, $org_scale$0 = 0;
 var $org_scale$0$in = 0, $scores$i = 0, $w$02 = 0, $xx1$0 = 0, $xx1max$0$i = 0, $xx2$0 = 0, $y$03$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 272|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $scores$i = sp;
 $0 = ($dim|0)==(1);
 $1 = ((($hints)) + 12|0);
 $2 = ((($hints)) + 16|0);
 $3 = ((($hints)) + 4|0);
 $4 = ((($hints)) + 8|0);
 $org_delta$0$in = $0 ? $2 : $4;
 $org_scale$0$in = $0 ? $1 : $3;
 $org_scale$0 = HEAP32[$org_scale$0$in>>2]|0;
 $org_delta$0 = HEAP32[$org_delta$0$in>>2]|0;
 $5 = ((($warper)) + 44|0);
 HEAP32[$5>>2] = $org_scale$0;
 $6 = ((($warper)) + 48|0);
 HEAP32[$6>>2] = $org_delta$0;
 $7 = ((($warper)) + 52|0);
 HEAP32[$7>>2] = -2147483648;
 $8 = ((($warper)) + 56|0);
 HEAP32[$8>>2] = 0;
 $9 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (((($hints)) + 44|0) + (($dim*1468)|0)|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($hints)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($hints)) + 24|0);
 $16 = HEAP32[$15>>2]|0;
 HEAP32[$a_scale>>2] = $org_scale$0;
 HEAP32[$a_delta>>2] = $org_delta$0;
 $17 = ($12|0)<(1);
 if ($17) {
  STACKTOP = sp;return;
 }
 $18 = ((($14)) + 12|0);
 $19 = HEAP16[$18>>1]|0;
 $20 = $19 << 16 >> 16;
 $21 = ($16|0)>(1);
 if ($21) {
  $X1$04 = $20;$X2$03 = $20;$nn$05 = 1;
 } else {
  STACKTOP = sp;return;
 }
 while(1) {
  $22 = (((($14) + (($nn$05*40)|0)|0)) + 12|0);
  $23 = HEAP16[$22>>1]|0;
  $24 = $23 << 16 >> 16;
  $25 = ($24|0)<($X1$04|0);
  $$X1$0 = $25 ? $24 : $X1$04;
  $26 = ($24|0)>($X2$03|0);
  $X2$1 = $26 ? $24 : $X2$03;
  $27 = (($nn$05) + 1)|0;
  $exitcond = ($27|0)==($16|0);
  if ($exitcond) {
   $$X1$0$lcssa = $$X1$0;$X2$1$lcssa = $X2$1;
   break;
  } else {
   $X1$04 = $$X1$0;$X2$03 = $X2$1;$nn$05 = $27;
  }
 }
 $28 = ($X2$1$lcssa|0)>($$X1$0$lcssa|0);
 if (!($28)) {
  STACKTOP = sp;return;
 }
 $29 = (_FT_MulFix($$X1$0$lcssa,$org_scale$0)|0);
 $30 = (($29) + ($org_delta$0))|0;
 HEAP32[$warper>>2] = $30;
 $31 = (_FT_MulFix($X2$1$lcssa,$org_scale$0)|0);
 $32 = (($31) + ($org_delta$0))|0;
 $33 = ((($warper)) + 4|0);
 HEAP32[$33>>2] = $32;
 $34 = HEAP32[$warper>>2]|0;
 $35 = $34 & -64;
 $36 = ((($warper)) + 8|0);
 HEAP32[$36>>2] = $35;
 $37 = (($32) + 63)|0;
 $38 = $37 & -64;
 $39 = ((($warper)) + 12|0);
 HEAP32[$39>>2] = $38;
 $40 = $34 & -32;
 $41 = ((($warper)) + 16|0);
 HEAP32[$41>>2] = $40;
 $42 = (($40) + 32)|0;
 $43 = ((($warper)) + 20|0);
 HEAP32[$43>>2] = $42;
 $44 = $32 & -32;
 $45 = ((($warper)) + 24|0);
 HEAP32[$45>>2] = $44;
 $46 = (($44) + 32)|0;
 $47 = ((($warper)) + 28|0);
 HEAP32[$47>>2] = $46;
 $48 = ($42|0)>($32|0);
 if ($48) {
  HEAP32[$43>>2] = $32;
  $171 = $32;
 } else {
  $171 = $42;
 }
 $49 = ($44|0)<($34|0);
 if ($49) {
  HEAP32[$45>>2] = $34;
  $172 = $34;
 } else {
  $172 = $44;
 }
 $50 = (($32) - ($34))|0;
 $51 = ((($warper)) + 32|0);
 HEAP32[$51>>2] = $50;
 $52 = ($50|0)<(65);
 if ($52) {
  HEAP32[$43>>2] = $34;
  HEAP32[$45>>2] = $32;
  $54 = $32;$55 = $34;
 } else {
  $54 = $172;$55 = $171;
 }
 $53 = (($54) - ($55))|0;
 $56 = ((($warper)) + 36|0);
 HEAP32[$56>>2] = $53;
 $57 = (($46) - ($40))|0;
 $58 = ((($warper)) + 40|0);
 HEAP32[$58>>2] = $57;
 $59 = ($50|0)<(129);
 $60 = ($50|0)<(97);
 $$ = $60 ? 4 : 8;
 $margin$0 = $59 ? $$ : 16;
 $61 = (($50) - ($margin$0))|0;
 $62 = ($53|0)<($61|0);
 if ($62) {
  HEAP32[$56>>2] = $61;
  $68 = $61;
 } else {
  $68 = $53;
 }
 $63 = (($50) + ($margin$0))|0;
 $64 = ($57|0)>($63|0);
 if ($64) {
  HEAP32[$58>>2] = $63;
  $72 = $63;
 } else {
  $72 = $57;
 }
 $65 = ($50*3)|0;
 $66 = (($65|0) / 4)&-1;
 $67 = ($68|0)<($66|0);
 if ($67) {
  HEAP32[$56>>2] = $66;
  $74 = $66;
 } else {
  $74 = $68;
 }
 $69 = ($50*5)|0;
 $70 = (($69|0) / 4)&-1;
 $71 = ($72|0)>($70|0);
 if ($71) {
  HEAP32[$58>>2] = $70;
  $75 = $70;
 } else {
  $75 = $72;
 }
 $73 = ($74|0)>($75|0);
 L33: do {
  if (!($73)) {
   $76 = (($X2$1$lcssa) - ($$X1$0$lcssa))|0;
   $77 = ($12|0)>(0);
   $79 = $50;$82 = $34;$86 = $32;$w$02 = $74;
   while(1) {
    $78 = ($79|0)>($w$02|0);
    $80 = (($79) - ($w$02))|0;
    $81 = (($80) + ($82))|0;
    if ($78) {
     $88 = HEAP32[$43>>2]|0;
     $89 = ($88|0)<($81|0);
     if ($89) {
      $90 = (($86) - ($81))|0;
      $91 = (($90) + ($88))|0;
      $xx1$0 = $88;$xx2$0 = $91;
     } else {
      $xx1$0 = $81;$xx2$0 = $86;
     }
    } else {
     $83 = HEAP32[$41>>2]|0;
     $84 = ($83|0)>($81|0);
     if ($84) {
      $85 = (($86) - ($81))|0;
      $87 = (($85) + ($83))|0;
      $xx1$0 = $83;$xx2$0 = $87;
     } else {
      $xx1$0 = $81;$xx2$0 = $86;
     }
    }
    $92 = ($xx1$0|0)<($82|0);
    $93 = (($82) - ($xx1$0))|0;
    $94 = (($xx1$0) - ($82))|0;
    $base_distort$0 = $92 ? $93 : $94;
    $95 = ($xx2$0|0)<($86|0);
    $96 = (($86) - ($xx2$0))|0;
    $97 = (($xx2$0) - ($86))|0;
    $$pn = $95 ? $96 : $97;
    $base_distort$1 = (($$pn) + ($base_distort$0))|0;
    $98 = ($base_distort$1*10)|0;
    $99 = (($w$02) - ($79))|0;
    $100 = (_FT_DivFix($99,$76)|0);
    $101 = (($100) + ($org_scale$0))|0;
    $102 = (_FT_MulFix($$X1$0$lcssa,$101)|0);
    $103 = (($xx1$0) - ($102))|0;
    _memset(($scores$i|0),0,260)|0;
    $104 = HEAP32[$36>>2]|0;
    $105 = (($xx1$0) - ($104))|0;
    $106 = HEAP32[$41>>2]|0;
    $107 = (($xx2$0) - ($xx1$0))|0;
    $108 = (($106) + ($107))|0;
    $109 = HEAP32[$45>>2]|0;
    $110 = ($108|0)<($109|0);
    $111 = (($109) - ($107))|0;
    $$$i = $110 ? $111 : $106;
    $112 = HEAP32[$43>>2]|0;
    $113 = (($112) + ($107))|0;
    $114 = HEAP32[$47>>2]|0;
    $115 = ($113|0)>($114|0);
    $116 = (($114) - ($107))|0;
    $xx1max$0$i = $115 ? $116 : $112;
    $117 = (($$$i) - ($104))|0;
    $118 = (($xx1max$0$i) - ($104))|0;
    $119 = ($117|0)<(0);
    if (!($119)) {
     $120 = ($117|0)>($118|0);
     $121 = ($118|0)>(64);
     $or$cond$i = $120 | $121;
     if (!($or$cond$i)) {
      if ($77) {
       $122 = (($103) - ($105))|0;
       $123 = (($122) + ($117))|0;
       $nn$16$i = 0;
       while(1) {
        $124 = (((($10) + (($nn$16$i*48)|0)|0)) + 6|0);
        $125 = HEAP16[$124>>1]|0;
        $126 = $125 << 16 >> 16;
        $127 = (((($10) + (($nn$16$i*48)|0)|0)) + 4|0);
        $128 = HEAP16[$127>>1]|0;
        $129 = $128 << 16 >> 16;
        $130 = (($126) - ($129))|0;
        $131 = (((($10) + (($nn$16$i*48)|0)|0)) + 2|0);
        $132 = HEAP16[$131>>1]|0;
        $133 = $132 << 16 >> 16;
        $134 = (_FT_MulFix($133,$101)|0);
        $135 = (($123) + ($134))|0;
        $idx$04$i = $117;$y$03$i = $135;
        while(1) {
         $136 = $y$03$i & 63;
         $137 = (62020 + ($136<<2)|0);
         $138 = HEAP32[$137>>2]|0;
         $139 = Math_imul($138, $130)|0;
         $140 = (($scores$i) + ($idx$04$i<<2)|0);
         $141 = HEAP32[$140>>2]|0;
         $142 = (($141) + ($139))|0;
         HEAP32[$140>>2] = $142;
         $143 = (($idx$04$i) + 1)|0;
         $144 = (($y$03$i) + 1)|0;
         $145 = ($idx$04$i|0)<($118|0);
         if ($145) {
          $idx$04$i = $143;$y$03$i = $144;
         } else {
          break;
         }
        }
        $146 = (($nn$16$i) + 1)|0;
        $exitcond10$i = ($146|0)==($12|0);
        if ($exitcond10$i) {
         break;
        } else {
         $nn$16$i = $146;
        }
       }
      }
      $$pre$i = HEAP32[$7>>2]|0;
      $152 = $$pre$i;$idx1$02$i = $117;
      while(1) {
       $147 = (($scores$i) + ($idx1$02$i<<2)|0);
       $148 = HEAP32[$147>>2]|0;
       $149 = (($idx1$02$i) - ($105))|0;
       $150 = (($149) + ($98))|0;
       $151 = ($148|0)>($152|0);
       do {
        if ($151) {
         label = 37;
        } else {
         $153 = ($148|0)==($152|0);
         if (!($153)) {
          $173 = $152;
          break;
         }
         $154 = HEAP32[$8>>2]|0;
         $155 = ($150|0)<($154|0);
         if ($155) {
          label = 37;
         } else {
          $173 = $152;
         }
        }
       } while(0);
       if ((label|0) == 37) {
        label = 0;
        HEAP32[$7>>2] = $148;
        HEAP32[$8>>2] = $150;
        HEAP32[$5>>2] = $101;
        $156 = (($149) + ($103))|0;
        HEAP32[$6>>2] = $156;
        $173 = $148;
       }
       $157 = (($idx1$02$i) + 1)|0;
       $158 = ($idx1$02$i|0)<($118|0);
       if ($158) {
        $152 = $173;$idx1$02$i = $157;
       } else {
        break;
       }
      }
     }
    }
    $159 = HEAP32[$58>>2]|0;
    $160 = ($w$02|0)<($159|0);
    if (!($160)) {
     break L33;
    }
    $161 = (($w$02) + 1)|0;
    $$pre = HEAP32[$warper>>2]|0;
    $$pre10 = HEAP32[$33>>2]|0;
    $$pre11 = HEAP32[$51>>2]|0;
    $79 = $$pre11;$82 = $$pre;$86 = $$pre10;$w$02 = $161;
   }
  }
 } while(0);
 $162 = HEAP32[$5>>2]|0;
 $163 = HEAP32[$6>>2]|0;
 $164 = (($162) - ($org_scale$0))|0;
 $165 = (_FT_MulFix($$X1$0$lcssa,$164)|0);
 $166 = (($165) + ($163))|0;
 $167 = ((($hints)) + 2992|0);
 HEAP32[$167>>2] = $166;
 $168 = (_FT_MulFix($X2$1$lcssa,$164)|0);
 $169 = (($168) + ($163))|0;
 $170 = ((($hints)) + 2996|0);
 HEAP32[$170>>2] = $169;
 HEAP32[$a_scale>>2] = $162;
 HEAP32[$a_delta>>2] = $163;
 STACKTOP = sp;return;
}
function _af_latin_compute_stem_width($hints,$dim,$width,$base_flags,$stem_flags) {
 $hints = $hints|0;
 $dim = $dim|0;
 $width = $width|0;
 $base_flags = $base_flags|0;
 $stem_flags = $stem_flags|0;
 var $$ = 0, $$$i = 0, $$$width = 0, $$$width6 = 0, $$0 = 0, $$7 = 0, $$8 = 0, $$dist$1 = 0, $$width = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $9 = 0, $best$02$i = 0, $best$1$i = 0;
 var $dist$1 = 0, $dist$6 = 0, $dist$6$ = 0, $exitcond$i = 0, $n$04$i = 0, $or$cond3 = 0, $reference$0$lcssa5$i = 0, $reference$0$width$i = 0, $reference$0$width1$i = 0, $reference$03$i = 0, $reference$1$i = 0, $reference$1$i$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($hints)) + 2988|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($dim|0)==(1);
 $3 = ((($hints)) + 2984|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 & 4;
 $6 = ($5|0)==(0);
 if ($6) {
  $$0 = $width;
  return ($$0|0);
 }
 $7 = (((((($1)) + 44|0) + (($dim*1572)|0)|0)) + 212|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = ($8<<24>>24)==(0);
 if (!($9)) {
  $$0 = $width;
  return ($$0|0);
 }
 $10 = ($width|0)<(0);
 $11 = (0 - ($width))|0;
 $$width = $10 ? $11 : $width;
 if ($2) {
  $12 = $4 & 2;
  $13 = ($12|0)==(0);
  if ($13) {
   $16 = $stem_flags & 2;
   $17 = ($16|0)!=(0);
   $18 = ($$width|0)<(192);
   $or$cond3 = $18 & $17;
   if ($or$cond3) {
    $dist$6 = $$width;
   } else {
    label = 7;
   }
  } else {
   label = 16;
  }
 } else {
  $14 = $4 & 1;
  $15 = ($14|0)==(0);
  if ($15) {
   label = 7;
  } else {
   label = 16;
  }
 }
 do {
  if ((label|0) == 7) {
   $19 = $base_flags & 1;
   $20 = ($19|0)==(0);
   $21 = ($$width|0)<(80);
   $$$width = $21 ? 64 : $$width;
   $22 = ($$width|0)<(56);
   $$$width6 = $22 ? 56 : $$width;
   $dist$1 = $20 ? $$$width6 : $$$width;
   $23 = (((((($1)) + 44|0) + (($dim*1572)|0)|0)) + 8|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = ($24|0)==(0);
   if ($25) {
    $dist$6 = $dist$1;
   } else {
    $26 = (((((($1)) + 44|0) + (($dim*1572)|0)|0)) + 16|0);
    $27 = HEAP32[$26>>2]|0;
    $28 = (($dist$1) - ($27))|0;
    $29 = ($28|0)<(0);
    $30 = (0 - ($28))|0;
    $$7 = $29 ? $30 : $28;
    $31 = ($$7|0)<(40);
    if ($31) {
     $32 = ($27|0)<(48);
     $$ = $32 ? 48 : $27;
     $dist$6 = $$;
     break;
    }
    $33 = ($dist$1|0)<(192);
    if (!($33)) {
     $41 = (($dist$1) + 32)|0;
     $42 = $41 & -64;
     $dist$6 = $42;
     break;
    }
    $34 = $dist$1 & 63;
    $35 = $dist$1 & -64;
    $36 = ($34>>>0)<(10);
    if ($36) {
     $dist$6 = $dist$1;
    } else {
     $37 = ($34>>>0)<(32);
     if ($37) {
      $38 = $35 | 10;
      $dist$6 = $38;
      break;
     } else {
      $39 = ($34>>>0)<(54);
      $40 = $35 | 54;
      $$dist$1 = $39 ? $40 : $dist$1;
      $dist$6 = $$dist$1;
      break;
     }
    }
   }
  }
  else if ((label|0) == 16) {
   $43 = (((((($1)) + 44|0) + (($dim*1572)|0)|0)) + 8|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = ($44|0)==(0);
   if ($45) {
    $46 = (($$width) + 32)|0;
    $47 = $46 & -64;
    $59 = $47;$reference$0$lcssa5$i = $$width;
    label = 20;
   } else {
    $best$02$i = 98;$n$04$i = 0;$reference$03$i = $$width;
    while(1) {
     $48 = ((((((((($1)) + 44|0) + (($dim*1572)|0)|0)) + 12|0) + (($n$04$i*12)|0)|0)) + 4|0);
     $49 = HEAP32[$48>>2]|0;
     $50 = (($$width) - ($49))|0;
     $51 = ($50|0)<(0);
     $52 = (0 - ($50))|0;
     $$$i = $51 ? $52 : $50;
     $53 = ($$$i|0)<($best$02$i|0);
     $best$1$i = $53 ? $$$i : $best$02$i;
     $reference$1$i = $53 ? $49 : $reference$03$i;
     $54 = (($n$04$i) + 1)|0;
     $exitcond$i = ($54|0)==($44|0);
     if ($exitcond$i) {
      $reference$1$i$lcssa = $reference$1$i;
      break;
     } else {
      $best$02$i = $best$1$i;$n$04$i = $54;$reference$03$i = $reference$1$i;
     }
    }
    $55 = (($reference$1$i$lcssa) + 32)|0;
    $56 = $55 & -64;
    $57 = ($reference$1$i$lcssa|0)>($$width|0);
    if ($57) {
     $61 = (($56) + -48)|0;
     $62 = ($61|0)<($$width|0);
     $reference$0$width1$i = $62 ? $reference$1$i$lcssa : $$width;
     $64 = $reference$0$width1$i;
    } else {
     $59 = $56;$reference$0$lcssa5$i = $reference$1$i$lcssa;
     label = 20;
    }
   }
   if ((label|0) == 20) {
    $58 = $59 | 48;
    $60 = ($58|0)>($$width|0);
    $reference$0$width$i = $60 ? $reference$0$lcssa5$i : $$width;
    $64 = $reference$0$width$i;
   }
   if ($2) {
    $63 = ($64|0)>(63);
    if (!($63)) {
     $dist$6 = 64;
     break;
    }
    $65 = (($64) + 16)|0;
    $66 = $65 & -64;
    $dist$6 = $66;
    break;
   }
   $67 = $4 & 8;
   $68 = ($67|0)==(0);
   if (!($68)) {
    $69 = ($64|0)<(64);
    if ($69) {
     $dist$6 = 64;
     break;
    }
    $70 = (($64) + 32)|0;
    $71 = $70 & -64;
    $dist$6 = $71;
    break;
   }
   $72 = ($64|0)<(48);
   if ($72) {
    $73 = (($64) + 64)|0;
    $74 = $73 >> 1;
    $dist$6 = $74;
    break;
   }
   $75 = ($64|0)<(128);
   if (!($75)) {
    $85 = (($64) + 32)|0;
    $86 = $85 & -64;
    $dist$6 = $86;
    break;
   }
   $76 = (($64) + 22)|0;
   $77 = $76 & -64;
   $78 = (($77) - ($$width))|0;
   $79 = ($78|0)<(0);
   $80 = (0 - ($78))|0;
   $$8 = $79 ? $80 : $78;
   $81 = ($$8|0)>(15);
   if ($81) {
    $82 = ($$width|0)<(48);
    if ($82) {
     $83 = (($$width) + 64)|0;
     $84 = $83 >> 1;
     $dist$6 = $84;
    } else {
     $dist$6 = $$width;
    }
   } else {
    $dist$6 = $77;
   }
  }
 } while(0);
 $87 = ($width|0)>(-1);
 $88 = (0 - ($dist$6))|0;
 $dist$6$ = $87 ? $dist$6 : $88;
 $$0 = $dist$6$;
 return ($$0|0);
}
function _af_glyph_hints_align_strong_points($hints,$dim) {
 $hints = $hints|0;
 $dim = $dim|0;
 var $$ = 0, $$idx = 0, $$lcssa = 0, $$lcssa$lcssa = 0, $$lcssa51 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phi24Z2D = 0, $$pre$phiZ2D = 0, $$pre22 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $max$0 = 0, $max$0$lcssa50 = 0, $max$0$ph = 0, $min$1$ph = 0, $min$2 = 0, $nn$0$lcssa = 0, $nn$01039 = 0, $or$cond = 0;
 var $or$cond43 = 0, $or$cond45 = 0, $ou$0 = 0, $ou$0$in = 0, $point$013 = 0, $u$0 = 0, $u$3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($hints)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($hints)) + 24|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + (($3*40)|0)|0);
 $5 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 20|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 12|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($dim|0)==(0);
 $$ = $9 ? 4 : 8;
 $10 = ($8|0)>(0);
 $11 = ($3|0)>(0);
 $or$cond = $10 & $11;
 if (!($or$cond)) {
  return;
 }
 $12 = (($6) + (($8*48)|0)|0);
 $13 = $$ | 16;
 $14 = ($dim|0)==(1);
 $15 = ((($12)) + -48|0);
 $$idx = ($8*48)|0;
 $16 = (($$idx|0) / 48)&-1;
 $17 = ($$idx|0)<(432);
 $18 = ($$idx|0)>(0);
 $19 = ((($15)) + 8|0);
 $20 = ((($15)) + 4|0);
 $21 = ((($6)) + 8|0);
 $22 = ((($6)) + 4|0);
 $point$013 = $1;
 while(1) {
  $23 = HEAP16[$point$013>>1]|0;
  $24 = $23&65535;
  $25 = $24 & $13;
  $26 = ($25|0)==(0);
  if ($26) {
   if ($14) {
    $27 = ((($point$013)) + 14|0);
    $28 = HEAP16[$27>>1]|0;
    $29 = $28 << 16 >> 16;
    $30 = ((($point$013)) + 8|0);
    $ou$0$in = $30;$u$0 = $29;
   } else {
    $31 = ((($point$013)) + 12|0);
    $32 = HEAP16[$31>>1]|0;
    $33 = $32 << 16 >> 16;
    $34 = ((($point$013)) + 4|0);
    $ou$0$in = $34;$u$0 = $33;
   }
   $ou$0 = HEAP32[$ou$0$in>>2]|0;
   $35 = HEAP16[$6>>1]|0;
   $36 = $35 << 16 >> 16;
   $37 = ($36|0)<($u$0|0);
   L12: do {
    if ($37) {
     $42 = HEAP16[$15>>1]|0;
     $43 = $42 << 16 >> 16;
     $44 = ($u$0|0)<($43|0);
     if (!($44)) {
      $45 = HEAP32[$19>>2]|0;
      $46 = HEAP32[$20>>2]|0;
      $47 = (($45) + ($ou$0))|0;
      $48 = (($47) - ($46))|0;
      $u$3 = $48;
      break;
     }
     L17: do {
      if ($17) {
       $49 = $35 << 16 >> 16;
       $50 = ($49|0)<($u$0|0);
       $or$cond43 = $18 & $50;
       if ($or$cond43) {
        $nn$01039 = 0;
        while(1) {
         $51 = (($nn$01039) + 1)|0;
         $52 = ($51|0)<($16|0);
         $$phi$trans$insert = (($6) + (($51*48)|0)|0);
         $$pre = HEAP16[$$phi$trans$insert>>1]|0;
         $53 = $$pre << 16 >> 16;
         $54 = ($53|0)<($u$0|0);
         $or$cond45 = $52 & $54;
         if ($or$cond45) {
          $nn$01039 = $51;
         } else {
          $56 = $$pre;$nn$0$lcssa = $51;
          break;
         }
        }
       } else {
        $56 = $35;$nn$0$lcssa = 0;
       }
       $55 = $56 << 16 >> 16;
       $57 = ($55|0)==($u$0|0);
       if ($57) {
        $58 = (((($6) + (($nn$0$lcssa*48)|0)|0)) + 8|0);
        $59 = HEAP32[$58>>2]|0;
        $u$3 = $59;
        break L12;
       } else {
        $min$2 = $nn$0$lcssa;
       }
      } else {
       $max$0$ph = $16;$min$1$ph = 0;
       while(1) {
        $max$0 = $max$0$ph;
        while(1) {
         $60 = ($min$1$ph|0)<($max$0|0);
         if (!($60)) {
          $min$2 = $min$1$ph;
          break L17;
         }
         $61 = (($max$0) + ($min$1$ph))|0;
         $62 = $61 >> 1;
         $63 = (($6) + (($62*48)|0)|0);
         $64 = HEAP16[$63>>1]|0;
         $65 = $64 << 16 >> 16;
         $66 = ($u$0|0)<($65|0);
         if ($66) {
          $max$0 = $62;
         } else {
          $$lcssa = $62;$$lcssa51 = $65;$max$0$lcssa50 = $max$0;
          break;
         }
        }
        $67 = ($u$0|0)>($$lcssa51|0);
        if (!($67)) {
         $$lcssa$lcssa = $$lcssa;
         break;
        }
        $68 = (($$lcssa) + 1)|0;
        $max$0$ph = $max$0$lcssa50;$min$1$ph = $68;
       }
       $69 = (((($6) + (($$lcssa$lcssa*48)|0)|0)) + 8|0);
       $70 = HEAP32[$69>>2]|0;
       $u$3 = $70;
       break L12;
      }
     } while(0);
     $71 = (($6) + (($min$2*48)|0)|0);
     $72 = ((($71)) + -48|0);
     $73 = ((($72)) + 16|0);
     $74 = HEAP32[$73>>2]|0;
     $75 = ($74|0)==(0);
     if ($75) {
      $76 = (((($6) + (($min$2*48)|0)|0)) + 8|0);
      $77 = HEAP32[$76>>2]|0;
      $78 = ((($72)) + 8|0);
      $79 = HEAP32[$78>>2]|0;
      $80 = (($77) - ($79))|0;
      $81 = HEAP16[$71>>1]|0;
      $82 = $81 << 16 >> 16;
      $83 = HEAP16[$72>>1]|0;
      $84 = $83 << 16 >> 16;
      $85 = (($82) - ($84))|0;
      $86 = (_FT_DivFix($80,$85)|0);
      HEAP32[$73>>2] = $86;
      $$pre$phi24Z2D = $72;$$pre$phiZ2D = $78;$91 = $86;
     } else {
      $$pre22 = ((($72)) + 8|0);
      $$pre$phi24Z2D = $72;$$pre$phiZ2D = $$pre22;$91 = $74;
     }
     $87 = HEAP32[$$pre$phiZ2D>>2]|0;
     $88 = HEAP16[$$pre$phi24Z2D>>1]|0;
     $89 = $88 << 16 >> 16;
     $90 = (($u$0) - ($89))|0;
     $92 = (_FT_MulFix($90,$91)|0);
     $93 = (($92) + ($87))|0;
     $u$3 = $93;
    } else {
     $38 = HEAP32[$21>>2]|0;
     $39 = HEAP32[$22>>2]|0;
     $40 = (($38) + ($ou$0))|0;
     $41 = (($40) - ($39))|0;
     $u$3 = $41;
    }
   } while(0);
   if ($9) {
    $94 = ((($point$013)) + 16|0);
    HEAP32[$94>>2] = $u$3;
   } else {
    $95 = ((($point$013)) + 20|0);
    HEAP32[$95>>2] = $u$3;
   }
   $96 = HEAP16[$point$013>>1]|0;
   $97 = $96&65535;
   $98 = $97 | $$;
   $99 = $98&65535;
   HEAP16[$point$013>>1] = $99;
  }
  $100 = ((($point$013)) + 40|0);
  $101 = ($100>>>0)<($4>>>0);
  if ($101) {
   $point$013 = $100;
  } else {
   break;
  }
 }
 return;
}
function _af_glyph_hints_align_weak_points($hints,$dim) {
 $hints = $hints|0;
 $dim = $dim|0;
 var $$$i = 0, $$$i20 = 0, $$$i8 = 0, $$not = 0, $$ph = 0, $$ph132 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $contour$053 = 0, $or$cond = 0, $or$cond$i = 0, $or$cond$i17 = 0, $or$cond$i5 = 0, $p$03$i = 0, $p$03$i18 = 0, $p$03$i6 = 0, $p$05$i = 0, $p$1$i = 0, $p$12$i = 0, $p$14$i = 0, $p$14$i1 = 0, $p$14$i11 = 0, $p$14$i23 = 0, $point$056 = 0, $point$158 = 0, $point$244 = 0;
 var $point$244$lcssa = 0, $point$3 = 0, $point$3129 = 0, $point$446 = 0, $point$446$lcssa = 0, $point$541 = 0, $point$642 = 0, $ref$pn3$i = 0, $ref$pn3$i$phi = 0, $ref1$ref2$i = 0, $ref1$ref2$i15 = 0, $ref1$ref2$i3 = 0, $ref2$ref1$i = 0, $ref2$ref1$i16 = 0, $ref2$ref1$i4 = 0, $touch_flag$0$ph = 0, $u$0$i = 0, $u$0$i21 = 0, $u$0$i9 = 0, $u$1$i = 0;
 var $u$1$i13 = 0, $u$1$i25 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($hints)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($hints)) + 24|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + (($3*40)|0)|0);
 $5 = ((($hints)) + 40|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($hints)) + 36|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (($6) + ($8<<2)|0);
 $10 = ($dim|0)==(0);
 $11 = ($3|0)>(0);
 if ($10) {
  if ($11) {
   $point$056 = $1;
   while(1) {
    $13 = ((($point$056)) + 16|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = ((($point$056)) + 24|0);
    HEAP32[$15>>2] = $14;
    $16 = ((($point$056)) + 4|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = ((($point$056)) + 28|0);
    HEAP32[$18>>2] = $17;
    $19 = ((($point$056)) + 40|0);
    $20 = ($19>>>0)<($4>>>0);
    if ($20) {
     $point$056 = $19;
    } else {
     $touch_flag$0$ph = 4;
     break;
    }
   }
  } else {
   $touch_flag$0$ph = 4;
  }
 } else {
  if ($11) {
   $point$158 = $1;
   while(1) {
    $21 = ((($point$158)) + 20|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = ((($point$158)) + 24|0);
    HEAP32[$23>>2] = $22;
    $24 = ((($point$158)) + 8|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = ((($point$158)) + 28|0);
    HEAP32[$26>>2] = $25;
    $27 = ((($point$158)) + 40|0);
    $28 = ($27>>>0)<($4>>>0);
    if ($28) {
     $point$158 = $27;
    } else {
     $touch_flag$0$ph = 8;
     break;
    }
   }
  } else {
   $touch_flag$0$ph = 8;
  }
 }
 $12 = ($8|0)>(0);
 if ($12) {
  $contour$053 = $6;
  while(1) {
   $29 = HEAP32[$contour$053>>2]|0;
   $30 = ((($29)) + 36|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = ($29>>>0)>($31>>>0);
   L12: do {
    if (!($32)) {
     $point$244 = $29;
     while(1) {
      $33 = HEAP16[$point$244>>1]|0;
      $34 = $33&65535;
      $35 = $34 & $touch_flag$0$ph;
      $36 = ($35|0)==(0);
      if (!($36)) {
       $point$244$lcssa = $point$244;
       break;
      }
      $37 = ((($point$244)) + 40|0);
      $38 = ($37>>>0)>($31>>>0);
      if ($38) {
       break L12;
      } else {
       $point$244 = $37;
      }
     }
     $point$3 = $point$244$lcssa;
     L17: while(1) {
      $39 = ($point$3>>>0)<($31>>>0);
      if ($39) {
       $41 = ((($point$3)) + 40|0);
       $42 = HEAP16[$41>>1]|0;
       $43 = $42&65535;
       $44 = $43 & $touch_flag$0$ph;
       $45 = ($44|0)==(0);
       $46 = ((($point$3)) + 40|0);
       if ($45) {
        $$ph = $46;
       } else {
        $point$3 = $46;
        continue;
       }
      } else {
       $40 = ((($point$3)) + 40|0);
       $$ph = $40;
      }
      $47 = ($$ph>>>0)>($31>>>0);
      if ($47) {
       $$ph132 = $$ph;$117 = $39;$118 = $47;$point$3129 = $point$3;
       break;
      } else {
       $point$446 = $$ph;
      }
      while(1) {
       $48 = HEAP16[$point$446>>1]|0;
       $49 = $48&65535;
       $50 = $49 & $touch_flag$0$ph;
       $51 = ($50|0)==(0);
       if (!($51)) {
        $point$446$lcssa = $point$446;
        break;
       }
       $52 = ((($point$446)) + 40|0);
       $53 = ($52>>>0)>($31>>>0);
       if ($53) {
        $$ph132 = $$ph;$117 = $39;$118 = $47;$point$3129 = $point$3;
        break L17;
       } else {
        $point$446 = $52;
       }
      }
      $54 = ((($point$446$lcssa)) + -40|0);
      $55 = ($$ph>>>0)>($54>>>0);
      if ($55) {
       $point$3 = $point$446$lcssa;
       continue;
      }
      $56 = ((($point$3)) + 28|0);
      $57 = HEAP32[$56>>2]|0;
      $58 = ((($point$446$lcssa)) + 28|0);
      $59 = HEAP32[$58>>2]|0;
      $60 = ($57|0)>($59|0);
      $ref1$ref2$i = $60 ? $point$3 : $point$446$lcssa;
      $ref2$ref1$i = $60 ? $point$446$lcssa : $point$3;
      $61 = ((($ref2$ref1$i)) + 28|0);
      $62 = HEAP32[$61>>2]|0;
      $63 = ((($ref1$ref2$i)) + 28|0);
      $64 = HEAP32[$63>>2]|0;
      $65 = ((($ref2$ref1$i)) + 24|0);
      $66 = HEAP32[$65>>2]|0;
      $67 = ((($ref1$ref2$i)) + 24|0);
      $68 = HEAP32[$67>>2]|0;
      $69 = (($66) - ($62))|0;
      $70 = (($68) - ($64))|0;
      $71 = ($68|0)==($66|0);
      $72 = ($64|0)==($62|0);
      $or$cond$i = $72 | $71;
      if ($or$cond$i) {
       $p$03$i = $$ph;
       while(1) {
        $73 = ((($p$03$i)) + 28|0);
        $74 = HEAP32[$73>>2]|0;
        $75 = ($74|0)>($62|0);
        if ($75) {
         $77 = ($74|0)<($64|0);
         $78 = (($74) + ($70))|0;
         $$$i = $77 ? $66 : $78;
         $u$0$i = $$$i;
        } else {
         $76 = (($74) + ($69))|0;
         $u$0$i = $76;
        }
        $79 = ((($p$03$i)) + 24|0);
        HEAP32[$79>>2] = $u$0$i;
        $80 = ((($p$03$i)) + 40|0);
        $81 = ($80>>>0)>($54>>>0);
        if ($81) {
         $point$3 = $point$446$lcssa;
         continue L17;
        } else {
         $p$03$i = $80;
        }
       }
      }
      $82 = (($68) - ($66))|0;
      $83 = (($64) - ($62))|0;
      $84 = (_FT_DivFix($82,$83)|0);
      $p$14$i = $$ph;
      while(1) {
       $85 = ((($p$14$i)) + 28|0);
       $86 = HEAP32[$85>>2]|0;
       $87 = ($86|0)>($62|0);
       do {
        if ($87) {
         $89 = ($86|0)<($64|0);
         if ($89) {
          $91 = (($86) - ($62))|0;
          $92 = (_FT_MulFix($91,$84)|0);
          $93 = (($92) + ($66))|0;
          $u$1$i = $93;
          break;
         } else {
          $90 = (($86) + ($70))|0;
          $u$1$i = $90;
          break;
         }
        } else {
         $88 = (($86) + ($69))|0;
         $u$1$i = $88;
        }
       } while(0);
       $94 = ((($p$14$i)) + 24|0);
       HEAP32[$94>>2] = $u$1$i;
       $95 = ((($p$14$i)) + 40|0);
       $96 = ($95>>>0)>($54>>>0);
       if ($96) {
        $point$3 = $point$446$lcssa;
        continue L17;
       } else {
        $p$14$i = $95;
       }
      }
     }
     $97 = ($point$3129|0)==($point$244$lcssa|0);
     if ($97) {
      $98 = ((($point$244$lcssa)) + 24|0);
      $99 = HEAP32[$98>>2]|0;
      $100 = ((($point$244$lcssa)) + 28|0);
      $101 = HEAP32[$100>>2]|0;
      $102 = (($99) - ($101))|0;
      $103 = ($99|0)==($101|0);
      if ($103) {
       break;
      }
      $104 = ($29>>>0)<($point$244$lcssa>>>0);
      if ($104) {
       $p$05$i = $29;
       while(1) {
        $106 = ((($p$05$i)) + 28|0);
        $107 = HEAP32[$106>>2]|0;
        $108 = (($107) + ($102))|0;
        $109 = ((($p$05$i)) + 24|0);
        HEAP32[$109>>2] = $108;
        $110 = ((($p$05$i)) + 40|0);
        $111 = ($110>>>0)<($point$244$lcssa>>>0);
        if ($111) {
         $p$05$i = $110;
        } else {
         break;
        }
       }
      }
      $p$12$i = ((($point$244$lcssa)) + 40|0);
      $105 = ($p$12$i>>>0)>($31>>>0);
      if ($105) {
       break;
      } else {
       $p$14$i1 = $p$12$i;$ref$pn3$i = $point$244$lcssa;
      }
      while(1) {
       $112 = ((($ref$pn3$i)) + 68|0);
       $113 = HEAP32[$112>>2]|0;
       $114 = (($113) + ($102))|0;
       $115 = ((($ref$pn3$i)) + 64|0);
       HEAP32[$115>>2] = $114;
       $p$1$i = ((($p$14$i1)) + 40|0);
       $116 = ($p$1$i>>>0)>($31>>>0);
       if ($116) {
        break L12;
       } else {
        $ref$pn3$i$phi = $p$14$i1;$p$14$i1 = $p$1$i;$ref$pn3$i = $ref$pn3$i$phi;
       }
      }
     }
     $$not = $117 ^ 1;
     $or$cond = $118 | $$not;
     L56: do {
      if (!($or$cond)) {
       $119 = ((($point$3129)) + 28|0);
       $120 = HEAP32[$119>>2]|0;
       $121 = ((($point$244$lcssa)) + 28|0);
       $122 = HEAP32[$121>>2]|0;
       $123 = ($120|0)>($122|0);
       $ref1$ref2$i3 = $123 ? $point$3129 : $point$244$lcssa;
       $ref2$ref1$i4 = $123 ? $point$244$lcssa : $point$3129;
       $124 = ((($ref2$ref1$i4)) + 28|0);
       $125 = HEAP32[$124>>2]|0;
       $126 = ((($ref1$ref2$i3)) + 28|0);
       $127 = HEAP32[$126>>2]|0;
       $128 = ((($ref2$ref1$i4)) + 24|0);
       $129 = HEAP32[$128>>2]|0;
       $130 = ((($ref1$ref2$i3)) + 24|0);
       $131 = HEAP32[$130>>2]|0;
       $132 = (($129) - ($125))|0;
       $133 = (($131) - ($127))|0;
       $134 = ($131|0)==($129|0);
       $135 = ($127|0)==($125|0);
       $or$cond$i5 = $135 | $134;
       if ($or$cond$i5) {
        $p$03$i6 = $$ph132;
        while(1) {
         $136 = ((($p$03$i6)) + 28|0);
         $137 = HEAP32[$136>>2]|0;
         $138 = ($137|0)>($125|0);
         if ($138) {
          $140 = ($137|0)<($127|0);
          $141 = (($137) + ($133))|0;
          $$$i8 = $140 ? $129 : $141;
          $u$0$i9 = $$$i8;
         } else {
          $139 = (($137) + ($132))|0;
          $u$0$i9 = $139;
         }
         $142 = ((($p$03$i6)) + 24|0);
         HEAP32[$142>>2] = $u$0$i9;
         $143 = ((($p$03$i6)) + 40|0);
         $144 = ($143>>>0)>($31>>>0);
         if ($144) {
          break L56;
         } else {
          $p$03$i6 = $143;
         }
        }
       }
       $145 = (($131) - ($129))|0;
       $146 = (($127) - ($125))|0;
       $147 = (_FT_DivFix($145,$146)|0);
       $p$14$i11 = $$ph132;
       while(1) {
        $148 = ((($p$14$i11)) + 28|0);
        $149 = HEAP32[$148>>2]|0;
        $150 = ($149|0)>($125|0);
        do {
         if ($150) {
          $152 = ($149|0)<($127|0);
          if ($152) {
           $154 = (($149) - ($125))|0;
           $155 = (_FT_MulFix($154,$147)|0);
           $156 = (($155) + ($129))|0;
           $u$1$i13 = $156;
           break;
          } else {
           $153 = (($149) + ($133))|0;
           $u$1$i13 = $153;
           break;
          }
         } else {
          $151 = (($149) + ($132))|0;
          $u$1$i13 = $151;
         }
        } while(0);
        $157 = ((($p$14$i11)) + 24|0);
        HEAP32[$157>>2] = $u$1$i13;
        $158 = ((($p$14$i11)) + 40|0);
        $159 = ($158>>>0)>($31>>>0);
        if ($159) {
         break;
        } else {
         $p$14$i11 = $158;
        }
       }
      }
     } while(0);
     $160 = ($point$244$lcssa>>>0)>($1>>>0);
     if ($160) {
      $161 = ((($point$244$lcssa)) + -40|0);
      $162 = ($29>>>0)>($161>>>0);
      if (!($162)) {
       $163 = ((($point$3129)) + 28|0);
       $164 = HEAP32[$163>>2]|0;
       $165 = ((($point$244$lcssa)) + 28|0);
       $166 = HEAP32[$165>>2]|0;
       $167 = ($164|0)>($166|0);
       $ref1$ref2$i15 = $167 ? $point$3129 : $point$244$lcssa;
       $ref2$ref1$i16 = $167 ? $point$244$lcssa : $point$3129;
       $168 = ((($ref2$ref1$i16)) + 28|0);
       $169 = HEAP32[$168>>2]|0;
       $170 = ((($ref1$ref2$i15)) + 28|0);
       $171 = HEAP32[$170>>2]|0;
       $172 = ((($ref2$ref1$i16)) + 24|0);
       $173 = HEAP32[$172>>2]|0;
       $174 = ((($ref1$ref2$i15)) + 24|0);
       $175 = HEAP32[$174>>2]|0;
       $176 = (($173) - ($169))|0;
       $177 = (($175) - ($171))|0;
       $178 = ($175|0)==($173|0);
       $179 = ($171|0)==($169|0);
       $or$cond$i17 = $179 | $178;
       if ($or$cond$i17) {
        $p$03$i18 = $29;
        while(1) {
         $180 = ((($p$03$i18)) + 28|0);
         $181 = HEAP32[$180>>2]|0;
         $182 = ($181|0)>($169|0);
         if ($182) {
          $184 = ($181|0)<($171|0);
          $185 = (($181) + ($177))|0;
          $$$i20 = $184 ? $173 : $185;
          $u$0$i21 = $$$i20;
         } else {
          $183 = (($181) + ($176))|0;
          $u$0$i21 = $183;
         }
         $186 = ((($p$03$i18)) + 24|0);
         HEAP32[$186>>2] = $u$0$i21;
         $187 = ((($p$03$i18)) + 40|0);
         $188 = ($187>>>0)>($161>>>0);
         if ($188) {
          break L12;
         } else {
          $p$03$i18 = $187;
         }
        }
       }
       $189 = (($175) - ($173))|0;
       $190 = (($171) - ($169))|0;
       $191 = (_FT_DivFix($189,$190)|0);
       $p$14$i23 = $29;
       while(1) {
        $192 = ((($p$14$i23)) + 28|0);
        $193 = HEAP32[$192>>2]|0;
        $194 = ($193|0)>($169|0);
        do {
         if ($194) {
          $196 = ($193|0)<($171|0);
          if ($196) {
           $198 = (($193) - ($169))|0;
           $199 = (_FT_MulFix($198,$191)|0);
           $200 = (($199) + ($173))|0;
           $u$1$i25 = $200;
           break;
          } else {
           $197 = (($193) + ($177))|0;
           $u$1$i25 = $197;
           break;
          }
         } else {
          $195 = (($193) + ($176))|0;
          $u$1$i25 = $195;
         }
        } while(0);
        $201 = ((($p$14$i23)) + 24|0);
        HEAP32[$201>>2] = $u$1$i25;
        $202 = ((($p$14$i23)) + 40|0);
        $203 = ($202>>>0)>($161>>>0);
        if ($203) {
         break;
        } else {
         $p$14$i23 = $202;
        }
       }
      }
     }
    }
   } while(0);
   $204 = ((($contour$053)) + 4|0);
   $205 = ($204>>>0)<($9>>>0);
   if ($205) {
    $contour$053 = $204;
   } else {
    break;
   }
  }
 }
 if ($10) {
  if ($11) {
   $point$541 = $1;
  } else {
   return;
  }
  while(1) {
   $206 = ((($point$541)) + 24|0);
   $207 = HEAP32[$206>>2]|0;
   $208 = ((($point$541)) + 16|0);
   HEAP32[$208>>2] = $207;
   $209 = ((($point$541)) + 40|0);
   $210 = ($209>>>0)<($4>>>0);
   if ($210) {
    $point$541 = $209;
   } else {
    break;
   }
  }
  return;
 } else {
  if ($11) {
   $point$642 = $1;
  } else {
   return;
  }
  while(1) {
   $211 = ((($point$642)) + 24|0);
   $212 = HEAP32[$211>>2]|0;
   $213 = ((($point$642)) + 20|0);
   HEAP32[$213>>2] = $212;
   $214 = ((($point$642)) + 40|0);
   $215 = ($214>>>0)<($4>>>0);
   if ($215) {
    $point$642 = $214;
   } else {
    break;
   }
  }
  return;
 }
}
function _af_cjk_metrics_init($metrics,$face) {
 $metrics = $metrics|0;
 $face = $face|0;
 var $$lcssa = 0, $$lcssa$i = 0, $$lcssa$in$i = 0, $$lcssa120 = 0, $$lcssa30$i = 0, $$lcssa30$in$i = 0, $$lcssa31$i = 0, $$lcssa31$in$i = 0, $$phi$trans$insert$i = 0, $$phi$trans$insert$i$i = 0, $$phi$trans$insert$i6$i = 0, $$phi$trans$insert104$i = 0, $$phi$trans$insert106$i = 0, $$phi$trans$insert108$i = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i7$i = 0, $$pre105$i = 0, $$pre107$i = 0, $$pre109$i = 0;
 var $$pre111$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $axis$0$i = 0, $best_point$065$i = 0, $best_point$10$i = 0, $best_point$144$i = 0, $best_point$2$i = 0;
 var $best_point$339$i = 0, $best_point$4$i = 0, $best_point$558$i = 0, $best_point$6$i = 0, $best_point$751$i = 0, $best_point$8$i = 0, $best_pos$0$lcssa$i = 0, $best_pos$0$lcssa$i$lcssa = 0, $best_pos$064$i = 0, $best_pos$10$i = 0, $best_pos$143$i = 0, $best_pos$2$i = 0, $best_pos$338$i = 0, $best_pos$4$i = 0, $best_pos$557$i = 0, $best_pos$6$i = 0, $best_pos$750$i = 0, $best_pos$8$i = 0, $bs$071$i = 0, $ch$036$i = 0;
 var $ch$036$ph$i = 0, $ch$1$i = 0, $exitcond$i = 0, $exitcond$i$i = 0, $exitcond$i11$i = 0, $fill$0$i$ph15 = 0, $fill$0$i$ph15$lcssa = 0, $fill$0$ph$ph$i28 = 0, $fill$0$ph$ph26$i21 = 0, $fills$i = 0, $first$067$i = 0, $flats$i = 0, $i$03$i$i = 0, $i$03$i5$i = 0, $j$01$i$i = 0, $j$01$i9$i = 0, $len$037$i = 0, $len$037$ph$i = 0, $nn$066$i = 0, $num_fills$0$ph$ph$i30 = 0;
 var $num_fills$0$ph$ph27$i$lcssa = 0, $num_fills$0$ph$ph27$i22 = 0, $num_fills$0$ph$ph27$i22$lcssa = 0, $num_flats$0$ph$ph$i$lcssa = 0, $num_flats$0$ph$ph$i29 = 0, $or$cond$i = 0, $or$cond140$i = 0, $or$cond141$i = 0, $or$cond142$i = 0, $p$0$i$ph14 = 0, $p$0$i10 = 0, $p$0$ph$ph$i27 = 0, $p$0$ph$ph25$i20 = 0, $p$135$i = 0, $p$2$i = 0, $p$2$i$lcssa = 0, $p$2$i$lcssa117 = 0, $p$2$i$lcssa117$lcssa = 0, $pp$045$i = 0, $pp$140$i = 0;
 var $pp$259$i = 0, $pp$352$i = 0, $scevgep101$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 416|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $fills$i = sp + 204|0;
 $flats$i = sp;
 $0 = ((($face)) + 92|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 68|0);
 $3 = HEAP16[$2>>1]|0;
 $4 = $3&65535;
 $5 = ((($metrics)) + 40|0);
 HEAP32[$5>>2] = $4;
 $6 = (_FT_Select_Charmap($face,1970170211)|0);
 $7 = ($6|0)==(0);
 if (!($7)) {
  (_FT_Set_Charmap($face,$1)|0);
  STACKTOP = sp;return 0;
 }
 _af_cjk_metrics_init_widths($metrics,$face);
 $8 = HEAP32[$metrics>>2]|0;
 $9 = ((($8)) + 12|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (59204 + ($10<<3)|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)==(759);
 if (!($13)) {
  $14 = ((($metrics)) + 44|0);
  $15 = ((($metrics)) + 1616|0);
  $16 = ((($face)) + 84|0);
  $17 = ((($metrics)) + 36|0);
  $19 = $12;$bs$071$i = $11;
  while(1) {
   $18 = (594743 + ($19)|0);
   $20 = ((($bs$071$i)) + 4|0);
   $21 = HEAP16[$20>>1]|0;
   $22 = $21 & 2;
   $23 = ($22<<16>>16)==(0);
   $axis$0$i = $23 ? $15 : $14;
   $24 = HEAP8[$18>>0]|0;
   $25 = ($24<<24>>24)==(0);
   L8: do {
    if ($25) {
     $num_fills$0$ph$ph27$i$lcssa = 0;$num_flats$0$ph$ph$i$lcssa = 0;
    } else {
     $151 = $24;$fill$0$ph$ph$i28 = 1;$num_fills$0$ph$ph$i30 = 0;$num_flats$0$ph$ph$i29 = 0;$p$0$ph$ph$i27 = $18;
     while(1) {
      $152 = $151;$fill$0$ph$ph26$i21 = $fill$0$ph$ph$i28;$num_fills$0$ph$ph27$i22 = $num_fills$0$ph$ph$i30;$p$0$ph$ph25$i20 = $p$0$ph$ph$i27;
      while(1) {
       $153 = $152;$fill$0$i$ph15 = $fill$0$ph$ph26$i21;$p$0$i$ph14 = $p$0$ph$ph25$i20;
       L13: while(1) {
        $30 = $153;$p$0$i10 = $p$0$i$ph14;
        while(1) {
         $28 = ((($p$0$i10)) + 1|0);
         $29 = $30&255;
         $31 = ($30<<24>>24)<(0);
         if ($31) {
          $32 = ($30&255)<(224);
          do {
           if ($32) {
            $33 = $29 & 31;
            $ch$036$ph$i = $33;$len$037$ph$i = 1;
           } else {
            $34 = ($30&255)<(240);
            if ($34) {
             $35 = $29 & 15;
             $ch$036$ph$i = $35;$len$037$ph$i = 2;
             break;
            } else {
             $36 = $29 & 7;
             $ch$036$ph$i = $36;$len$037$ph$i = 3;
             break;
            }
           }
          } while(0);
          $ch$036$i = $ch$036$ph$i;$len$037$i = $len$037$ph$i;$p$135$i = $28;
          while(1) {
           $37 = $ch$036$i << 6;
           $38 = ((($p$135$i)) + 1|0);
           $39 = HEAP8[$p$135$i>>0]|0;
           $40 = $39&255;
           $41 = $40 & 63;
           $42 = $41 | $37;
           $43 = (($len$037$i) + -1)|0;
           $44 = ($43|0)==(0);
           if ($44) {
            $$lcssa = $42;
            break;
           } else {
            $ch$036$i = $42;$len$037$i = $43;$p$135$i = $38;
           }
          }
          $scevgep101$i = (($28) + ($len$037$ph$i)|0);
          $ch$1$i = $$lcssa;$p$2$i = $scevgep101$i;
         } else {
          $ch$1$i = $29;$p$2$i = $28;
         }
         $45 = ($ch$1$i|0)==(124);
         if ($45) {
          $p$2$i$lcssa = $p$2$i;
          break;
         }
         $46 = HEAP32[$17>>2]|0;
         $47 = HEAP32[$46>>2]|0;
         $48 = (_FT_Get_Char_Index($47,$ch$1$i)|0);
         $49 = ($48|0)==(0);
         if (!($49)) {
          $50 = (_FT_Load_Glyph($face,$48,1)|0);
          $51 = HEAP32[$16>>2]|0;
          $52 = ($50|0)==(0);
          if ($52) {
           $53 = ((($51)) + 110|0);
           $54 = HEAP16[$53>>1]|0;
           $55 = ($54<<16>>16)<(1);
           if (!($55)) {
            $$lcssa120 = $51;$fill$0$i$ph15$lcssa = $fill$0$i$ph15;$p$2$i$lcssa117 = $p$2$i;
            break L13;
           }
          }
         }
         $56 = HEAP8[$p$2$i>>0]|0;
         $57 = ($56<<24>>24)==(0);
         if ($57) {
          $num_fills$0$ph$ph27$i$lcssa = $num_fills$0$ph$ph27$i22;$num_flats$0$ph$ph$i$lcssa = $num_flats$0$ph$ph$i29;
          break L8;
         } else {
          $30 = $56;$p$0$i10 = $p$2$i;
         }
        }
        $26 = HEAP8[$p$2$i$lcssa>>0]|0;
        $27 = ($26<<24>>24)==(0);
        if ($27) {
         $num_fills$0$ph$ph27$i$lcssa = $num_fills$0$ph$ph27$i22;$num_flats$0$ph$ph$i$lcssa = $num_flats$0$ph$ph$i29;
         break L8;
        } else {
         $153 = $26;$fill$0$i$ph15 = 0;$p$0$i$ph14 = $p$2$i$lcssa;
        }
       }
       $$lcssa$in$i = ((($$lcssa120)) + 108|0);
       $$lcssa30$in$i = ((($$lcssa120)) + 112|0);
       $$lcssa31$in$i = ((($$lcssa120)) + 120|0);
       $$lcssa$i = HEAP16[$$lcssa$in$i>>1]|0;
       $$lcssa30$i = HEAP32[$$lcssa30$in$i>>2]|0;
       $$lcssa31$i = HEAP32[$$lcssa31$in$i>>2]|0;
       $58 = ($$lcssa$i<<16>>16)>(0);
       if ($58) {
        $59 = $$lcssa$i << 16 >> 16;
        $best_point$065$i = -1;$best_pos$064$i = 0;$first$067$i = 0;$nn$066$i = 0;
        while(1) {
         $60 = (($$lcssa31$i) + ($nn$066$i<<1)|0);
         $61 = HEAP16[$60>>1]|0;
         $62 = $61 << 16 >> 16;
         $63 = ($62|0)>($first$067$i|0);
         do {
          if ($63) {
           $64 = HEAP16[$20>>1]|0;
           $65 = $64&65535;
           $66 = $65 & 2;
           $67 = ($66|0)==(0);
           $68 = $65 & 1;
           $69 = ($68|0)!=(0);
           $70 = ($first$067$i|0)>($62|0);
           if ($67) {
            if ($69) {
             if ($70) {
              $best_point$10$i = $best_point$065$i;$best_pos$10$i = $best_pos$064$i;
              break;
             } else {
              $best_point$558$i = $best_point$065$i;$best_pos$557$i = $best_pos$064$i;$pp$259$i = $first$067$i;
             }
             while(1) {
              $79 = ($best_point$558$i|0)<(0);
              $$phi$trans$insert108$i = (((($$lcssa30$i) + ($pp$259$i<<3)|0)) + 4|0);
              $$pre109$i = HEAP32[$$phi$trans$insert108$i>>2]|0;
              $80 = ($$pre109$i|0)>($best_pos$557$i|0);
              $or$cond141$i = $79 | $80;
              $best_pos$6$i = $or$cond141$i ? $$pre109$i : $best_pos$557$i;
              $best_point$6$i = $or$cond141$i ? $pp$259$i : $best_point$558$i;
              $81 = (($pp$259$i) + 1)|0;
              $82 = ($pp$259$i|0)<($62|0);
              if ($82) {
               $best_point$558$i = $best_point$6$i;$best_pos$557$i = $best_pos$6$i;$pp$259$i = $81;
              } else {
               $best_point$10$i = $best_point$6$i;$best_pos$10$i = $best_pos$6$i;
               break;
              }
             }
            } else {
             if ($70) {
              $best_point$10$i = $best_point$065$i;$best_pos$10$i = $best_pos$064$i;
              break;
             } else {
              $best_point$751$i = $best_point$065$i;$best_pos$750$i = $best_pos$064$i;$pp$352$i = $first$067$i;
             }
             while(1) {
              $83 = ($best_point$751$i|0)<(0);
              $$phi$trans$insert106$i = (((($$lcssa30$i) + ($pp$352$i<<3)|0)) + 4|0);
              $$pre107$i = HEAP32[$$phi$trans$insert106$i>>2]|0;
              $84 = ($$pre107$i|0)<($best_pos$750$i|0);
              $or$cond142$i = $83 | $84;
              $best_pos$8$i = $or$cond142$i ? $$pre107$i : $best_pos$750$i;
              $best_point$8$i = $or$cond142$i ? $pp$352$i : $best_point$751$i;
              $85 = (($pp$352$i) + 1)|0;
              $86 = ($pp$352$i|0)<($62|0);
              if ($86) {
               $best_point$751$i = $best_point$8$i;$best_pos$750$i = $best_pos$8$i;$pp$352$i = $85;
              } else {
               $best_point$10$i = $best_point$8$i;$best_pos$10$i = $best_pos$8$i;
               break;
              }
             }
            }
           } else {
            if ($69) {
             if ($70) {
              $best_point$10$i = $best_point$065$i;$best_pos$10$i = $best_pos$064$i;
              break;
             } else {
              $best_point$144$i = $best_point$065$i;$best_pos$143$i = $best_pos$064$i;$pp$045$i = $first$067$i;
             }
             while(1) {
              $71 = ($best_point$144$i|0)<(0);
              $$phi$trans$insert104$i = (($$lcssa30$i) + ($pp$045$i<<3)|0);
              $$pre105$i = HEAP32[$$phi$trans$insert104$i>>2]|0;
              $72 = ($$pre105$i|0)>($best_pos$143$i|0);
              $or$cond$i = $71 | $72;
              $best_pos$2$i = $or$cond$i ? $$pre105$i : $best_pos$143$i;
              $best_point$2$i = $or$cond$i ? $pp$045$i : $best_point$144$i;
              $73 = (($pp$045$i) + 1)|0;
              $74 = ($pp$045$i|0)<($62|0);
              if ($74) {
               $best_point$144$i = $best_point$2$i;$best_pos$143$i = $best_pos$2$i;$pp$045$i = $73;
              } else {
               $best_point$10$i = $best_point$2$i;$best_pos$10$i = $best_pos$2$i;
               break;
              }
             }
            } else {
             if ($70) {
              $best_point$10$i = $best_point$065$i;$best_pos$10$i = $best_pos$064$i;
              break;
             } else {
              $best_point$339$i = $best_point$065$i;$best_pos$338$i = $best_pos$064$i;$pp$140$i = $first$067$i;
             }
             while(1) {
              $75 = ($best_point$339$i|0)<(0);
              $$phi$trans$insert$i = (($$lcssa30$i) + ($pp$140$i<<3)|0);
              $$pre$i = HEAP32[$$phi$trans$insert$i>>2]|0;
              $76 = ($$pre$i|0)<($best_pos$338$i|0);
              $or$cond140$i = $75 | $76;
              $best_pos$4$i = $or$cond140$i ? $$pre$i : $best_pos$338$i;
              $best_point$4$i = $or$cond140$i ? $pp$140$i : $best_point$339$i;
              $77 = (($pp$140$i) + 1)|0;
              $78 = ($pp$140$i|0)<($62|0);
              if ($78) {
               $best_point$339$i = $best_point$4$i;$best_pos$338$i = $best_pos$4$i;$pp$140$i = $77;
              } else {
               $best_point$10$i = $best_point$4$i;$best_pos$10$i = $best_pos$4$i;
               break;
              }
             }
            }
           }
          } else {
           $best_point$10$i = $best_point$065$i;$best_pos$10$i = $best_pos$064$i;
          }
         } while(0);
         $87 = (($62) + 1)|0;
         $88 = (($nn$066$i) + 1)|0;
         $exitcond$i = ($88|0)==($59|0);
         if ($exitcond$i) {
          $best_pos$0$lcssa$i = $best_pos$10$i;
          break;
         } else {
          $best_point$065$i = $best_point$10$i;$best_pos$064$i = $best_pos$10$i;$first$067$i = $87;$nn$066$i = $88;
         }
        }
       } else {
        $best_pos$0$lcssa$i = 0;
       }
       $89 = ($fill$0$i$ph15$lcssa<<24>>24)==(0);
       if ($89) {
        $best_pos$0$lcssa$i$lcssa = $best_pos$0$lcssa$i;$num_fills$0$ph$ph27$i22$lcssa = $num_fills$0$ph$ph27$i22;$p$2$i$lcssa117$lcssa = $p$2$i$lcssa117;
        break;
       }
       $90 = (($num_fills$0$ph$ph27$i22) + 1)|0;
       $91 = (($fills$i) + ($num_fills$0$ph$ph27$i22<<2)|0);
       HEAP32[$91>>2] = $best_pos$0$lcssa$i;
       $92 = HEAP8[$p$2$i$lcssa117>>0]|0;
       $93 = ($92<<24>>24)==(0);
       if ($93) {
        $num_fills$0$ph$ph27$i$lcssa = $90;$num_flats$0$ph$ph$i$lcssa = $num_flats$0$ph$ph$i29;
        break L8;
       } else {
        $152 = $92;$fill$0$ph$ph26$i21 = $fill$0$i$ph15$lcssa;$num_fills$0$ph$ph27$i22 = $90;$p$0$ph$ph25$i20 = $p$2$i$lcssa117;
       }
      }
      $94 = (($num_flats$0$ph$ph$i29) + 1)|0;
      $95 = (($flats$i) + ($num_flats$0$ph$ph$i29<<2)|0);
      HEAP32[$95>>2] = $best_pos$0$lcssa$i$lcssa;
      $96 = HEAP8[$p$2$i$lcssa117$lcssa>>0]|0;
      $97 = ($96<<24>>24)==(0);
      if ($97) {
       $num_fills$0$ph$ph27$i$lcssa = $num_fills$0$ph$ph27$i22$lcssa;$num_flats$0$ph$ph$i$lcssa = $94;
       break;
      } else {
       $151 = $96;$fill$0$ph$ph$i28 = 0;$num_fills$0$ph$ph$i30 = $num_fills$0$ph$ph27$i22$lcssa;$num_flats$0$ph$ph$i29 = $94;$p$0$ph$ph$i27 = $p$2$i$lcssa117$lcssa;
      }
     }
    }
   } while(0);
   $98 = ($num_fills$0$ph$ph27$i$lcssa|0)==(0);
   $99 = $num_fills$0$ph$ph27$i$lcssa | $num_flats$0$ph$ph$i$lcssa;
   $100 = ($99|0)==(0);
   if (!($100)) {
    $101 = ($num_flats$0$ph$ph$i$lcssa|0)==(0);
    $102 = ($num_fills$0$ph$ph27$i$lcssa>>>0)>(1);
    if ($102) {
     $i$03$i$i = 1;
     while(1) {
      $$phi$trans$insert$i$i = (($fills$i) + ($i$03$i$i<<2)|0);
      $$pre$i$i = HEAP32[$$phi$trans$insert$i$i>>2]|0;
      $j$01$i$i = $i$03$i$i;
      while(1) {
       $103 = (($j$01$i$i) + -1)|0;
       $104 = (($fills$i) + ($103<<2)|0);
       $105 = HEAP32[$104>>2]|0;
       $106 = ($$pre$i$i|0)<($105|0);
       if (!($106)) {
        break;
       }
       $107 = (($fills$i) + ($j$01$i$i<<2)|0);
       HEAP32[$107>>2] = $105;
       HEAP32[$104>>2] = $$pre$i$i;
       $108 = ($103|0)==(0);
       if ($108) {
        break;
       } else {
        $j$01$i$i = $103;
       }
      }
      $109 = (($i$03$i$i) + 1)|0;
      $exitcond$i$i = ($109|0)==($num_fills$0$ph$ph27$i$lcssa|0);
      if ($exitcond$i$i) {
       break;
      } else {
       $i$03$i$i = $109;
      }
     }
    }
    $110 = ($num_flats$0$ph$ph$i$lcssa>>>0)>(1);
    if ($110) {
     $i$03$i5$i = 1;
     while(1) {
      $$phi$trans$insert$i6$i = (($flats$i) + ($i$03$i5$i<<2)|0);
      $$pre$i7$i = HEAP32[$$phi$trans$insert$i6$i>>2]|0;
      $j$01$i9$i = $i$03$i5$i;
      while(1) {
       $111 = (($j$01$i9$i) + -1)|0;
       $112 = (($flats$i) + ($111<<2)|0);
       $113 = HEAP32[$112>>2]|0;
       $114 = ($$pre$i7$i|0)<($113|0);
       if (!($114)) {
        break;
       }
       $115 = (($flats$i) + ($j$01$i9$i<<2)|0);
       HEAP32[$115>>2] = $113;
       HEAP32[$112>>2] = $$pre$i7$i;
       $116 = ($111|0)==(0);
       if ($116) {
        break;
       } else {
        $j$01$i9$i = $111;
       }
      }
      $117 = (($i$03$i5$i) + 1)|0;
      $exitcond$i11$i = ($117|0)==($num_flats$0$ph$ph$i$lcssa|0);
      if ($exitcond$i11$i) {
       break;
      } else {
       $i$03$i5$i = $117;
      }
     }
    }
    $118 = ((($axis$0$i)) + 216|0);
    $119 = HEAP32[$118>>2]|0;
    $120 = (((($axis$0$i)) + 220|0) + (($119*28)|0)|0);
    $121 = (((((($axis$0$i)) + 220|0) + (($119*28)|0)|0)) + 12|0);
    $122 = (($119) + 1)|0;
    HEAP32[$118>>2] = $122;
    do {
     if ($101) {
      $123 = $num_fills$0$ph$ph27$i$lcssa >>> 1;
      $124 = (($fills$i) + ($123<<2)|0);
      $125 = HEAP32[$124>>2]|0;
      HEAP32[$121>>2] = $125;
      HEAP32[$120>>2] = $125;
      label = 57;
     } else {
      if ($98) {
       $126 = $num_flats$0$ph$ph$i$lcssa >>> 1;
       $127 = (($flats$i) + ($126<<2)|0);
       $128 = HEAP32[$127>>2]|0;
       HEAP32[$121>>2] = $128;
       HEAP32[$120>>2] = $128;
       label = 57;
       break;
      }
      $129 = $num_fills$0$ph$ph27$i$lcssa >>> 1;
      $130 = (($fills$i) + ($129<<2)|0);
      $131 = HEAP32[$130>>2]|0;
      HEAP32[$120>>2] = $131;
      $132 = $num_flats$0$ph$ph$i$lcssa >>> 1;
      $133 = (($flats$i) + ($132<<2)|0);
      $134 = HEAP32[$133>>2]|0;
      HEAP32[$121>>2] = $134;
      $135 = ($134|0)==($131|0);
      if ($135) {
       label = 57;
      } else {
       $136 = ($134|0)<($131|0);
       $137 = HEAP16[$20>>1]|0;
       $138 = $137 & 1;
       $139 = ($138<<16>>16)!=(0);
       $140 = $136 ^ $139;
       if ($140) {
        $141 = (($134) + ($131))|0;
        $142 = (($141|0) / 2)&-1;
        HEAP32[$121>>2] = $142;
        HEAP32[$120>>2] = $142;
        $145 = $137;
       } else {
        $145 = $137;
       }
      }
     }
    } while(0);
    if ((label|0) == 57) {
     label = 0;
     $$pre111$i = HEAP16[$20>>1]|0;
     $145 = $$pre111$i;
    }
    $143 = (((((($axis$0$i)) + 220|0) + (($119*28)|0)|0)) + 24|0);
    $144 = $145 & 1;
    $146 = $144&65535;
    $147 = $146 << 1;
    HEAP32[$143>>2] = $147;
   }
   $148 = ((($bs$071$i)) + 8|0);
   $149 = HEAP32[$148>>2]|0;
   $150 = ($149|0)==(759);
   if ($150) {
    break;
   } else {
    $19 = $149;$bs$071$i = $148;
   }
  }
 }
 _af_cjk_metrics_check_digits($metrics,$face);
 (_FT_Set_Charmap($face,$1)|0);
 STACKTOP = sp;return 0;
}
function _af_cjk_metrics_init_widths($metrics,$face) {
 $metrics = $metrics|0;
 $face = $face|0;
 var $$ = 0, $$$1 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $brmerge = 0, $brmerge12 = 0, $dummy = 0, $glyph_index$3 = 0, $hints = 0, $num_widths = 0, $or$cond = 0, $or$cond$1 = 0;
 var $seg$08 = 0, $seg$08$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 10064|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $hints = sp + 3192|0;
 $dummy = sp + 4|0;
 $num_widths = sp;
 $0 = ((($face)) + 100|0);
 $1 = HEAP32[$0>>2]|0;
 _memset(($hints|0),0,3000)|0;
 HEAP32[$hints>>2] = $1;
 $2 = ((($metrics)) + 52|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($metrics)) + 1624|0);
 HEAP32[$3>>2] = 0;
 $4 = HEAP32[$metrics>>2]|0;
 $5 = ((($4)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (61928 + ($6<<2)|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($metrics|0)==(0|0);
 if ($9) {
  label = 3;
 } else {
  $10 = ((($8)) + 8|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ((($metrics)) + 36|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = HEAP32[$13>>2]|0;
  $15 = (_FT_Get_Char_Index($14,$11)|0);
  $16 = ($15|0)==(0);
  if ($16) {
   label = 3;
  } else {
   $glyph_index$3 = $15;
   label = 7;
  }
 }
 if ((label|0) == 3) {
  $17 = ((($8)) + 12|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = ($18|0)==(0);
  $brmerge12 = $19 | $9;
  if (!($brmerge12)) {
   $20 = ((($metrics)) + 36|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = HEAP32[$21>>2]|0;
   $23 = (_FT_Get_Char_Index($22,$18)|0);
   $24 = ($23|0)==(0);
   if ($24) {
    $25 = ((($8)) + 16|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ($26|0)==(0);
    $brmerge = $9 | $27;
    if (!($brmerge)) {
     $28 = ((($metrics)) + 36|0);
     $29 = HEAP32[$28>>2]|0;
     $30 = HEAP32[$29>>2]|0;
     $31 = (_FT_Get_Char_Index($30,$26)|0);
     $32 = ($31|0)==(0);
     if (!($32)) {
      $glyph_index$3 = $31;
      label = 7;
     }
    }
   } else {
    $glyph_index$3 = $23;
    label = 7;
   }
  }
 }
 do {
  if ((label|0) == 7) {
   $33 = (_FT_Load_Glyph($face,$glyph_index$3,1)|0);
   $34 = ($33|0)==(0);
   if ($34) {
    $35 = ((($face)) + 84|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ((($36)) + 110|0);
    $38 = HEAP16[$37>>1]|0;
    $39 = ($38<<16>>16)<(1);
    if (!($39)) {
     _memset(($dummy|0),0,3188)|0;
     $40 = ((($metrics)) + 40|0);
     $41 = HEAP32[$40>>2]|0;
     $42 = ((($dummy)) + 40|0);
     HEAP32[$42>>2] = $41;
     $43 = ((($dummy)) + 8|0);
     HEAP32[$43>>2] = 65536;
     $44 = ((($dummy)) + 12|0);
     HEAP32[$44>>2] = 65536;
     $45 = ((($dummy)) + 16|0);
     HEAP32[$45>>2] = 0;
     $46 = ((($dummy)) + 20|0);
     HEAP32[$46>>2] = 0;
     $47 = ((($dummy)) + 4|0);
     HEAP32[$47>>2] = $face;
     $48 = ((($dummy)) + 24|0);
     HEAP32[$48>>2] = 0;
     $49 = ((($dummy)) + 28|0);
     HEAP32[$49>>2] = 0;
     $50 = ((($hints)) + 2988|0);
     HEAP32[$50>>2] = $dummy;
     $51 = ((($hints)) + 2980|0);
     HEAP32[$51>>2] = 0;
     $52 = ((($36)) + 108|0);
     $53 = (_af_glyph_hints_reload($hints,$52)|0);
     $54 = ($53|0)==(0);
     if ($54) {
      HEAP32[$num_widths>>2] = 0;
      $55 = (_af_latin_hints_compute_segments($hints,0)|0);
      $56 = ($55|0)==(0);
      if ($56) {
       _af_latin_hints_link_segments($hints,0,0,0);
       $57 = ((($hints)) + 52|0);
       $58 = HEAP32[$57>>2]|0;
       $59 = ((($hints)) + 44|0);
       $60 = HEAP32[$59>>2]|0;
       $61 = (($58) + (($60*48)|0)|0);
       $62 = ($60|0)>(0);
       if ($62) {
        $78 = 0;$seg$08 = $58;
        while(1) {
         $63 = ((($seg$08)) + 20|0);
         $64 = HEAP32[$63>>2]|0;
         $65 = ($64|0)==(0|0);
         if ($65) {
          $148 = $78;
         } else {
          $66 = ((($64)) + 20|0);
          $67 = HEAP32[$66>>2]|0;
          $68 = ($67|0)==($seg$08|0);
          $69 = ($64>>>0)>($seg$08>>>0);
          $or$cond = $69 & $68;
          if ($or$cond) {
           $70 = ((($seg$08)) + 2|0);
           $71 = HEAP16[$70>>1]|0;
           $72 = $71 << 16 >> 16;
           $73 = ((($64)) + 2|0);
           $74 = HEAP16[$73>>1]|0;
           $75 = $74 << 16 >> 16;
           $76 = (($72) - ($75))|0;
           $77 = ($78>>>0)<(16);
           if ($77) {
            $79 = ($76|0)<(0);
            $80 = (0 - ($76))|0;
            $$ = $79 ? $80 : $76;
            $81 = (($78) + 1)|0;
            HEAP32[$num_widths>>2] = $81;
            $82 = (((($metrics)) + 56|0) + (($78*12)|0)|0);
            HEAP32[$82>>2] = $$;
            $148 = $81;
           } else {
            $148 = $78;
           }
          } else {
           $148 = $78;
          }
         }
         $83 = ((($seg$08)) + 48|0);
         $84 = ($83>>>0)<($61>>>0);
         if ($84) {
          $78 = $148;$seg$08 = $83;
         } else {
          break;
         }
        }
       }
       $85 = ((($metrics)) + 56|0);
       $86 = HEAP32[$42>>2]|0;
       $87 = (($86>>>0) / 100)&-1;
       _af_sort_and_quantize_widths($num_widths,$85,$87);
       $88 = HEAP32[$num_widths>>2]|0;
       HEAP32[$2>>2] = $88;
       HEAP32[$num_widths>>2] = 0;
       $89 = (_af_latin_hints_compute_segments($hints,1)|0);
       $90 = ($89|0)==(0);
       if ($90) {
        _af_latin_hints_link_segments($hints,0,0,1);
        $116 = ((($hints)) + 1520|0);
        $117 = HEAP32[$116>>2]|0;
        $118 = ((($hints)) + 1512|0);
        $119 = HEAP32[$118>>2]|0;
        $120 = (($117) + (($119*48)|0)|0);
        $121 = ($119|0)>(0);
        if ($121) {
         $137 = 0;$seg$08$1 = $117;
         while(1) {
          $122 = ((($seg$08$1)) + 20|0);
          $123 = HEAP32[$122>>2]|0;
          $124 = ($123|0)==(0|0);
          if ($124) {
           $149 = $137;
          } else {
           $125 = ((($123)) + 20|0);
           $126 = HEAP32[$125>>2]|0;
           $127 = ($126|0)==($seg$08$1|0);
           $128 = ($123>>>0)>($seg$08$1>>>0);
           $or$cond$1 = $128 & $127;
           if ($or$cond$1) {
            $129 = ((($seg$08$1)) + 2|0);
            $130 = HEAP16[$129>>1]|0;
            $131 = $130 << 16 >> 16;
            $132 = ((($123)) + 2|0);
            $133 = HEAP16[$132>>1]|0;
            $134 = $133 << 16 >> 16;
            $135 = (($131) - ($134))|0;
            $136 = ($137>>>0)<(16);
            if ($136) {
             $138 = ($135|0)<(0);
             $139 = (0 - ($135))|0;
             $$$1 = $138 ? $139 : $135;
             $140 = (($137) + 1)|0;
             HEAP32[$num_widths>>2] = $140;
             $141 = (((($metrics)) + 1628|0) + (($137*12)|0)|0);
             HEAP32[$141>>2] = $$$1;
             $149 = $140;
            } else {
             $149 = $137;
            }
           } else {
            $149 = $137;
           }
          }
          $142 = ((($seg$08$1)) + 48|0);
          $143 = ($142>>>0)<($120>>>0);
          if ($143) {
           $137 = $149;$seg$08$1 = $142;
          } else {
           break;
          }
         }
        }
        $144 = ((($metrics)) + 1628|0);
        $145 = HEAP32[$42>>2]|0;
        $146 = (($145>>>0) / 100)&-1;
        _af_sort_and_quantize_widths($num_widths,$144,$146);
        $147 = HEAP32[$num_widths>>2]|0;
        HEAP32[$3>>2] = $147;
        break;
       }
      }
     }
    }
   }
  }
 } while(0);
 $91 = ((($metrics)) + 40|0);
 $92 = HEAP32[$2>>2]|0;
 $93 = ($92|0)==(0);
 if ($93) {
  $96 = HEAP32[$91>>2]|0;
  $97 = ($96*50)|0;
  $98 = (($97|0) / 2048)&-1;
  $100 = $98;
 } else {
  $94 = ((($metrics)) + 56|0);
  $95 = HEAP32[$94>>2]|0;
  $100 = $95;
 }
 $99 = (($100|0) / 5)&-1;
 $101 = ((($metrics)) + 248|0);
 HEAP32[$101>>2] = $99;
 $102 = ((($metrics)) + 252|0);
 HEAP32[$102>>2] = $100;
 $103 = ((($metrics)) + 256|0);
 HEAP8[$103>>0] = 0;
 $104 = HEAP32[$3>>2]|0;
 $105 = ($104|0)==(0);
 if ($105) {
  $108 = HEAP32[$91>>2]|0;
  $109 = ($108*50)|0;
  $110 = (($109|0) / 2048)&-1;
  $112 = $110;
  $111 = (($112|0) / 5)&-1;
  $113 = ((($metrics)) + 1820|0);
  HEAP32[$113>>2] = $111;
  $114 = ((($metrics)) + 1824|0);
  HEAP32[$114>>2] = $112;
  $115 = ((($metrics)) + 1828|0);
  HEAP8[$115>>0] = 0;
  _af_glyph_hints_done($hints);
  STACKTOP = sp;return;
 } else {
  $106 = ((($metrics)) + 1628|0);
  $107 = HEAP32[$106>>2]|0;
  $112 = $107;
  $111 = (($112|0) / 5)&-1;
  $113 = ((($metrics)) + 1820|0);
  HEAP32[$113>>2] = $111;
  $114 = ((($metrics)) + 1824|0);
  HEAP32[$114>>2] = $112;
  $115 = ((($metrics)) + 1828|0);
  HEAP8[$115>>0] = 0;
  _af_glyph_hints_done($hints);
  STACKTOP = sp;return;
 }
}
function _af_cjk_metrics_check_digits($metrics,$face) {
 $metrics = $metrics|0;
 $face = $face|0;
 var $$ = 0, $$19 = 0, $$22 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $7 = 0, $8 = 0, $9 = 0, $advance = 0, $i$05 = 0, $old_advance$07 = 0, $old_advance$2$ph = 0, $old_advance$2$ph$us$1 = 0, $old_advance$2$ph$us$2 = 0, $old_advance$2$ph$us$3 = 0, $old_advance$2$ph$us$4 = 0, $old_advance$2$ph$us$5 = 0, $old_advance$2$ph$us$6 = 0, $old_advance$2$ph$us$7 = 0, $old_advance$2$ph$us$8 = 0, $or$cond = 0, $or$cond21 = 0, $same_width$2 = 0, $started$06 = 0;
 var $started$2$ph = 0, $started$2$ph$us$1 = 0, $started$2$ph$us$2 = 0, $started$2$ph$us$3 = 0, $started$2$ph$us$4 = 0, $started$2$ph$us$5 = 0, $started$2$ph$us$6 = 0, $started$2$ph$us$7 = 0, $started$2$ph$us$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $advance = sp;
 $0 = ($metrics|0)==(0|0);
 $1 = ((($metrics)) + 36|0);
 L1: do {
  if ($0) {
   $2 = (_FT_Get_Advance($face,0,2051,$advance)|0);
   $3 = ($2|0)==(0);
   $4 = HEAP32[$advance>>2]|0;
   $$ = $3&1;
   $$19 = $3 ? $4 : 0;
   $5 = (_FT_Get_Advance($face,0,2051,$advance)|0);
   $6 = ($5|0)==(0);
   if ($6) {
    $19 = HEAP32[$advance>>2]|0;
    if ($3) {
     $20 = ($19|0)==($$19|0);
     if ($20) {
      $old_advance$2$ph$us$1 = $$19;$started$2$ph$us$1 = 1;
     } else {
      $same_width$2 = 0;
      break;
     }
    } else {
     $old_advance$2$ph$us$1 = $19;$started$2$ph$us$1 = 1;
    }
   } else {
    $old_advance$2$ph$us$1 = $$19;$started$2$ph$us$1 = $$;
   }
   $21 = (_FT_Get_Advance($face,0,2051,$advance)|0);
   $22 = ($21|0)==(0);
   if ($22) {
    $23 = ($started$2$ph$us$1<<24>>24)==(0);
    $24 = HEAP32[$advance>>2]|0;
    if ($23) {
     $old_advance$2$ph$us$2 = $24;$started$2$ph$us$2 = 1;
    } else {
     $25 = ($24|0)==($old_advance$2$ph$us$1|0);
     if ($25) {
      $old_advance$2$ph$us$2 = $old_advance$2$ph$us$1;$started$2$ph$us$2 = 1;
     } else {
      $same_width$2 = 0;
      break;
     }
    }
   } else {
    $old_advance$2$ph$us$2 = $old_advance$2$ph$us$1;$started$2$ph$us$2 = $started$2$ph$us$1;
   }
   $26 = (_FT_Get_Advance($face,0,2051,$advance)|0);
   $27 = ($26|0)==(0);
   if ($27) {
    $28 = ($started$2$ph$us$2<<24>>24)==(0);
    $29 = HEAP32[$advance>>2]|0;
    if ($28) {
     $old_advance$2$ph$us$3 = $29;$started$2$ph$us$3 = 1;
    } else {
     $30 = ($29|0)==($old_advance$2$ph$us$2|0);
     if ($30) {
      $old_advance$2$ph$us$3 = $old_advance$2$ph$us$2;$started$2$ph$us$3 = 1;
     } else {
      $same_width$2 = 0;
      break;
     }
    }
   } else {
    $old_advance$2$ph$us$3 = $old_advance$2$ph$us$2;$started$2$ph$us$3 = $started$2$ph$us$2;
   }
   $31 = (_FT_Get_Advance($face,0,2051,$advance)|0);
   $32 = ($31|0)==(0);
   if ($32) {
    $33 = ($started$2$ph$us$3<<24>>24)==(0);
    $34 = HEAP32[$advance>>2]|0;
    if ($33) {
     $old_advance$2$ph$us$4 = $34;$started$2$ph$us$4 = 1;
    } else {
     $35 = ($34|0)==($old_advance$2$ph$us$3|0);
     if ($35) {
      $old_advance$2$ph$us$4 = $old_advance$2$ph$us$3;$started$2$ph$us$4 = 1;
     } else {
      $same_width$2 = 0;
      break;
     }
    }
   } else {
    $old_advance$2$ph$us$4 = $old_advance$2$ph$us$3;$started$2$ph$us$4 = $started$2$ph$us$3;
   }
   $36 = (_FT_Get_Advance($face,0,2051,$advance)|0);
   $37 = ($36|0)==(0);
   if ($37) {
    $38 = ($started$2$ph$us$4<<24>>24)==(0);
    $39 = HEAP32[$advance>>2]|0;
    if ($38) {
     $old_advance$2$ph$us$5 = $39;$started$2$ph$us$5 = 1;
    } else {
     $40 = ($39|0)==($old_advance$2$ph$us$4|0);
     if ($40) {
      $old_advance$2$ph$us$5 = $old_advance$2$ph$us$4;$started$2$ph$us$5 = 1;
     } else {
      $same_width$2 = 0;
      break;
     }
    }
   } else {
    $old_advance$2$ph$us$5 = $old_advance$2$ph$us$4;$started$2$ph$us$5 = $started$2$ph$us$4;
   }
   $41 = (_FT_Get_Advance($face,0,2051,$advance)|0);
   $42 = ($41|0)==(0);
   if ($42) {
    $43 = ($started$2$ph$us$5<<24>>24)==(0);
    $44 = HEAP32[$advance>>2]|0;
    if ($43) {
     $old_advance$2$ph$us$6 = $44;$started$2$ph$us$6 = 1;
    } else {
     $45 = ($44|0)==($old_advance$2$ph$us$5|0);
     if ($45) {
      $old_advance$2$ph$us$6 = $old_advance$2$ph$us$5;$started$2$ph$us$6 = 1;
     } else {
      $same_width$2 = 0;
      break;
     }
    }
   } else {
    $old_advance$2$ph$us$6 = $old_advance$2$ph$us$5;$started$2$ph$us$6 = $started$2$ph$us$5;
   }
   $46 = (_FT_Get_Advance($face,0,2051,$advance)|0);
   $47 = ($46|0)==(0);
   if ($47) {
    $48 = ($started$2$ph$us$6<<24>>24)==(0);
    $49 = HEAP32[$advance>>2]|0;
    if ($48) {
     $old_advance$2$ph$us$7 = $49;$started$2$ph$us$7 = 1;
    } else {
     $50 = ($49|0)==($old_advance$2$ph$us$6|0);
     if ($50) {
      $old_advance$2$ph$us$7 = $old_advance$2$ph$us$6;$started$2$ph$us$7 = 1;
     } else {
      $same_width$2 = 0;
      break;
     }
    }
   } else {
    $old_advance$2$ph$us$7 = $old_advance$2$ph$us$6;$started$2$ph$us$7 = $started$2$ph$us$6;
   }
   $51 = (_FT_Get_Advance($face,0,2051,$advance)|0);
   $52 = ($51|0)==(0);
   if ($52) {
    $53 = ($started$2$ph$us$7<<24>>24)==(0);
    $54 = HEAP32[$advance>>2]|0;
    if ($53) {
     $old_advance$2$ph$us$8 = $54;$started$2$ph$us$8 = 1;
    } else {
     $55 = ($54|0)==($old_advance$2$ph$us$7|0);
     if ($55) {
      $old_advance$2$ph$us$8 = $old_advance$2$ph$us$7;$started$2$ph$us$8 = 1;
     } else {
      $same_width$2 = 0;
      break;
     }
    }
   } else {
    $old_advance$2$ph$us$8 = $old_advance$2$ph$us$7;$started$2$ph$us$8 = $started$2$ph$us$7;
   }
   $56 = (_FT_Get_Advance($face,0,2051,$advance)|0);
   $57 = ($56|0)!=(0);
   $58 = ($started$2$ph$us$8<<24>>24)==(0);
   $or$cond = $57 | $58;
   $59 = HEAP32[$advance>>2]|0;
   $60 = ($59|0)==($old_advance$2$ph$us$8|0);
   $or$cond21 = $or$cond | $60;
   $$22 = $or$cond21&1;
   $same_width$2 = $$22;
  } else {
   $i$05 = 48;$old_advance$07 = 0;$started$06 = 0;
   while(1) {
    $7 = HEAP32[$1>>2]|0;
    $8 = HEAP32[$7>>2]|0;
    $9 = (_FT_Get_Char_Index($8,$i$05)|0);
    $10 = ($9|0)==(0);
    if ($10) {
     $old_advance$2$ph = $old_advance$07;$started$2$ph = $started$06;
    } else {
     $11 = (_FT_Get_Advance($face,$9,2051,$advance)|0);
     $12 = ($11|0)==(0);
     if ($12) {
      $13 = ($started$06<<24>>24)==(0);
      $14 = HEAP32[$advance>>2]|0;
      if ($13) {
       $old_advance$2$ph = $14;$started$2$ph = 1;
      } else {
       $15 = ($14|0)==($old_advance$07|0);
       if ($15) {
        $old_advance$2$ph = $old_advance$07;$started$2$ph = $started$06;
       } else {
        $same_width$2 = 0;
        break L1;
       }
      }
     } else {
      $old_advance$2$ph = $old_advance$07;$started$2$ph = $started$06;
     }
    }
    $16 = (($i$05) + 1)|0;
    $17 = ($16>>>0)<(58);
    if ($17) {
     $i$05 = $16;$old_advance$07 = $old_advance$2$ph;$started$06 = $started$2$ph;
    } else {
     $same_width$2 = 1;
     break;
    }
   }
  }
 } while(0);
 $18 = ((($metrics)) + 32|0);
 HEAP8[$18>>0] = $same_width$2;
 STACKTOP = sp;return;
}
function _af_cjk_metrics_scale($metrics,$scaler) {
 $metrics = $metrics|0;
 $scaler = $scaler|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($metrics)) + 4|0);
 ;HEAP32[$0>>2]=HEAP32[$scaler>>2]|0;HEAP32[$0+4>>2]=HEAP32[$scaler+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$scaler+8>>2]|0;HEAP32[$0+12>>2]=HEAP32[$scaler+12>>2]|0;HEAP32[$0+16>>2]=HEAP32[$scaler+16>>2]|0;HEAP32[$0+20>>2]=HEAP32[$scaler+20>>2]|0;HEAP32[$0+24>>2]=HEAP32[$scaler+24>>2]|0;
 _af_cjk_metrics_scale_dim($metrics,$scaler,0);
 _af_cjk_metrics_scale_dim($metrics,$scaler,1);
 return;
}
function _af_cjk_metrics_scale_dim($metrics,$scaler,$dim) {
 $metrics = $metrics|0;
 $scaler = $scaler|0;
 $dim = $dim|0;
 var $$ = 0, $$delta2$1 = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $delta$0 = 0;
 var $delta$0$in = 0, $delta2$1 = 0, $nn$02 = 0, $scale$0 = 0, $scale$0$in = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($dim|0)==(0);
 $1 = ((($scaler)) + 4|0);
 $2 = ((($scaler)) + 12|0);
 $3 = ((($scaler)) + 8|0);
 $4 = ((($scaler)) + 16|0);
 $delta$0$in = $0 ? $2 : $4;
 $scale$0$in = $0 ? $1 : $3;
 $scale$0 = HEAP32[$scale$0$in>>2]|0;
 $delta$0 = HEAP32[$delta$0$in>>2]|0;
 $5 = (((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 1564|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==($scale$0|0);
 $8 = (((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 1568|0);
 if ($7) {
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)==($delta$0|0);
  if ($10) {
   return;
  }
 }
 HEAP32[$5>>2] = $scale$0;
 HEAP32[$8>>2] = $delta$0;
 $11 = (((($metrics)) + 44|0) + (($dim*1572)|0)|0);
 HEAP32[$11>>2] = $scale$0;
 $12 = (((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 4|0);
 HEAP32[$12>>2] = $delta$0;
 $13 = (((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 216|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ($14|0)==(0);
 if ($15) {
  return;
 } else {
  $nn$02 = 0;
 }
 while(1) {
  $16 = ((((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 220|0) + (($nn$02*28)|0)|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = (_FT_MulFix($17,$scale$0)|0);
  $19 = (($18) + ($delta$0))|0;
  $20 = ((((((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 220|0) + (($nn$02*28)|0)|0)) + 4|0);
  HEAP32[$20>>2] = $19;
  $21 = ((((((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 220|0) + (($nn$02*28)|0)|0)) + 8|0);
  HEAP32[$21>>2] = $19;
  $22 = ((((((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 220|0) + (($nn$02*28)|0)|0)) + 12|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = (_FT_MulFix($23,$scale$0)|0);
  $25 = (($24) + ($delta$0))|0;
  $26 = ((((((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 220|0) + (($nn$02*28)|0)|0)) + 16|0);
  HEAP32[$26>>2] = $25;
  $27 = ((((((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 220|0) + (($nn$02*28)|0)|0)) + 20|0);
  HEAP32[$27>>2] = $25;
  $28 = ((((((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 220|0) + (($nn$02*28)|0)|0)) + 24|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = $29 & -2;
  HEAP32[$28>>2] = $30;
  $31 = HEAP32[$16>>2]|0;
  $32 = HEAP32[$22>>2]|0;
  $33 = (($31) - ($32))|0;
  $34 = (_FT_MulFix($33,$scale$0)|0);
  $$off = (($34) + 48)|0;
  $35 = ($$off>>>0)<(97);
  if ($35) {
   $36 = HEAP32[$20>>2]|0;
   $37 = (($36) + 32)|0;
   $38 = $37 & -64;
   HEAP32[$21>>2] = $38;
   $39 = (_FT_DivFix($38,$scale$0)|0);
   $40 = HEAP32[$22>>2]|0;
   $41 = (($39) - ($40))|0;
   $42 = ($41|0)<(0);
   $43 = (0 - ($41))|0;
   $$ = $42 ? $43 : $41;
   $44 = (_FT_MulFix($$,$scale$0)|0);
   $45 = ($44|0)<(32);
   $46 = (($44) + 32)|0;
   $47 = $46 & -64;
   $delta2$1 = $45 ? 0 : $47;
   $48 = (0 - ($delta2$1))|0;
   $$delta2$1 = $42 ? $48 : $delta2$1;
   $49 = HEAP32[$21>>2]|0;
   $50 = (($49) - ($$delta2$1))|0;
   HEAP32[$27>>2] = $50;
   $51 = HEAP32[$28>>2]|0;
   $52 = $51 | 1;
   HEAP32[$28>>2] = $52;
  }
  $53 = (($nn$02) + 1)|0;
  $54 = HEAP32[$13>>2]|0;
  $55 = ($53>>>0)<($54>>>0);
  if ($55) {
   $nn$02 = $53;
  } else {
   break;
  }
 }
 return;
}
function _af_cjk_hints_init($hints,$metrics) {
 $hints = $hints|0;
 $metrics = $metrics|0;
 var $$ = 0, $$4 = 0, $$4$v = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond3 = 0, $other_flags$1 = 0, $other_flags$1$ = 0;
 var $other_flags$3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($hints)) + 2988|0);
 HEAP32[$0>>2] = $metrics;
 $1 = ((($metrics)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($hints)) + 2980|0);
 $4 = ((($metrics)) + 44|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($hints)) + 4|0);
 HEAP32[$6>>2] = $5;
 $7 = ((($metrics)) + 48|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($hints)) + 8|0);
 HEAP32[$9>>2] = $8;
 $10 = ((($metrics)) + 1616|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($hints)) + 12|0);
 HEAP32[$12>>2] = $11;
 $13 = ((($metrics)) + 1620|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($hints)) + 16|0);
 HEAP32[$15>>2] = $14;
 $16 = ((($metrics)) + 24|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ($17|0)==(2);
 $19 = $17 | 1;
 $20 = ($19|0)==(3);
 $$ = $20&1;
 $21 = ($17|0)==(4);
 $or$cond3 = $18 | $21;
 $22 = $$ | 2;
 $other_flags$1 = $or$cond3 ? $22 : $$;
 $23 = ($17|0)==(1);
 $24 = $other_flags$1 | 4;
 $other_flags$1$ = $23 ? $other_flags$1 : $24;
 $25 = $other_flags$1$ | 8;
 $other_flags$3 = $18 ? $25 : $other_flags$1$;
 $26 = ((($metrics)) + 36|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ((($27)) + 212|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = ((($29)) + 20|0);
 $31 = HEAP8[$30>>0]|0;
 $32 = ($31<<24>>24)==(0);
 $$4$v = $32 ? 12 : 4;
 $$4 = $$4$v | $2;
 HEAP32[$3>>2] = $$4;
 $33 = ((($hints)) + 2984|0);
 HEAP32[$33>>2] = $other_flags$3;
 return 0;
}
function _af_cjk_hints_apply($hints,$outline,$metrics) {
 $hints = $hints|0;
 $outline = $outline|0;
 $metrics = $metrics|0;
 var $$$i = 0, $$9$i = 0, $$idx$i = 0, $$idx10$i = 0, $$idx10$us$i = 0, $$idx10$val$i = 0, $$idx10$val$us$i = 0, $$idx11$i = 0, $$idx11$us$i = 0, $$idx11$val$i = 0, $$idx11$val$us$i = 0, $$idx12$i = 0, $$idx12$us$i = 0, $$idx12$val$i = 0, $$idx12$val$us$i = 0, $$idx13$i = 0, $$idx13$us$i = 0, $$idx13$val$i = 0, $$idx13$val$us$i = 0, $$idx15$i = 0;
 var $$idx15$val$i = 0, $$idx16$i = 0, $$idx16$val$i = 0, $$idx17$i = 0, $$idx17$val$i = 0, $$idx18$i = 0, $$idx18$val$i = 0, $$idx19$i = 0, $$idx19$val$i = 0, $$idx20$i = 0, $$idx20$val$i = 0, $$lcssa = 0, $$lcssa29 = 0, $$lcssa56$i = 0, $$lcssa71 = 0, $$lobit$i = 0, $$not$i = 0, $$phi$trans$insert$i = 0, $$phi$trans$insert34 = 0, $$pre$i = 0;
 var $$pre$phiZ2D = 0, $$pre35 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0;
 var $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0;
 var $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0;
 var $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0;
 var $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0;
 var $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0;
 var $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0;
 var $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0;
 var $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0;
 var $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0;
 var $420 = 0, $421 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $after$0$i = 0, $after$0$i$lcssa = 0, $after$0$i70 = 0, $anchor$049$i = 0, $anchor$2$i = 0, $anchor$2$i$lcssa = 0, $anchor$343$i = 0, $anchor$4$i = 0, $before$0$i = 0, $before$0$i$lcssa = 0, $blue$0$i = 0, $brmerge$i = 0, $delta = 0, $delta$042$i = 0, $delta$1$i = 0, $delta$2$i = 0;
 var $dim$023 = 0, $edge$0$anchor$0$i = 0, $edge$02$i = 0, $edge$02$i$us = 0, $edge$050$i = 0, $edge$144$i = 0, $edge$144$us$i = 0, $edge$237$i = 0, $edge$334$i = 0, $edge1$0$i = 0, $edge1$0$idx$i = 0, $edge1$0$idx$val$i = 0, $edge13$0$i = 0, $edge2$0$i = 0, $edge24$0$i = 0, $edge3$0$i = 0, $error$2 = 0, $exitcond = 0, $has_last_stem$040$i = 0, $has_last_stem$040$us$i = 0;
 var $has_last_stem$1$i = 0, $has_last_stem$1$us$i = 0, $hints$idx$val = 0, $hints$idx1$val = 0, $hints$idx3 = 0, $hints$idx3$val = 0, $hints$idx4 = 0, $hints$idx4$val = 0, $last_stem_pos$041$i = 0, $last_stem_pos$041$us$i = 0, $last_stem_pos$1$i = 0, $last_stem_pos$1$us$i = 0, $or$cond = 0, $or$cond$i = 0, $or$cond50 = 0, $or$cond52 = 0, $or$cond6$i = 0, $outline$idx = 0, $outline$idx$val = 0, $outline$idx2 = 0;
 var $outline$idx2$val = 0, $phitmp$i = 0, $point$0$i = 0, $point$0$in$i = 0, $point$0$in$us$us$i = 0, $point$0$us$us$i = 0, $point$01$i = 0, $point$03$i = 0, $point1$0$i$us = 0, $point1$0$in$i$us = 0, $point1$0$in$us$us$i$us = 0, $point1$0$us$us$i$us = 0, $scale = 0, $seg$0$i = 0, $seg$0$us$i = 0, $seg$1$i$us = 0, $seg$1$us$i$us = 0, $skipped$0$lcssa$i = 0, $skipped$039$i = 0, $skipped$039$us$i = 0;
 var $skipped$1$i = 0, $skipped$1$us$i = 0, $skipped$236$i = 0, $skipped$3$i = 0, $skipped$3$i$lcssa = 0, $tag$03$i = 0, $vec$02$i = 0, $warper = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $warper = sp + 8|0;
 $scale = sp + 4|0;
 $delta = sp;
 $0 = (_af_glyph_hints_reload($hints,$outline)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  $error$2 = $0;
  STACKTOP = sp;return ($error$2|0);
 }
 $2 = ((($metrics)) + 24|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(1);
 $5 = ((($hints)) + 2980|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $6 & 8;
 $8 = ($7|0)==(0);
 $or$cond = $4 & $8;
 $9 = $6 & 1;
 $10 = ($9|0)==(0);
 $or$cond50 = $or$cond | $10;
 do {
  if ($or$cond50) {
   $11 = (_af_cjk_hints_detect_features($hints,0)|0);
   $12 = ($11|0)==(0);
   if ($12) {
    _af_cjk_hints_compute_blue_edges($hints,$metrics,0);
    $$phi$trans$insert34 = ((($hints)) + 2980|0);
    $$pre35 = HEAP32[$$phi$trans$insert34>>2]|0;
    $$pre$phiZ2D = $$phi$trans$insert34;$14 = $$pre35;
    break;
   } else {
    $error$2 = $11;
    STACKTOP = sp;return ($error$2|0);
   }
  } else {
   $$pre$phiZ2D = $5;$14 = $6;
  }
 } while(0);
 $13 = $14 & 2;
 $15 = ($13|0)==(0);
 do {
  if ($15) {
   $16 = (_af_cjk_hints_detect_features($hints,1)|0);
   $17 = ($16|0)==(0);
   if ($17) {
    _af_cjk_hints_compute_blue_edges($hints,$metrics,1);
    break;
   } else {
    $error$2 = $16;
    STACKTOP = sp;return ($error$2|0);
   }
  }
 } while(0);
 $hints$idx3 = ((($hints)) + 24|0);
 $hints$idx4 = ((($hints)) + 28|0);
 $18 = ((($hints)) + 2984|0);
 $dim$023 = 0;
 while(1) {
  $19 = ($dim$023|0)==(0);
  if ($19) {
   $20 = HEAP32[$$pre$phiZ2D>>2]|0;
   $21 = $20 & 1;
   $22 = ($21|0)==(0);
   if (!($22)) {
    $dim$023 = 1;
    continue;
   }
   $27 = HEAP32[$2>>2]|0;
   $28 = ($27|0)==(1);
   $29 = $20 & 8;
   $30 = ($29|0)==(0);
   $or$cond52 = $28 & $30;
   if ($or$cond52) {
    _af_warper_compute($warper,$hints,0,$scale,$delta);
    $31 = HEAP32[$scale>>2]|0;
    $32 = HEAP32[$delta>>2]|0;
    $hints$idx3$val = HEAP32[$hints$idx3>>2]|0;
    $hints$idx4$val = HEAP32[$hints$idx4>>2]|0;
    $33 = (($hints$idx4$val) + (($hints$idx3$val*40)|0)|0);
    $34 = ($hints$idx3$val|0)>(0);
    if ($34) {
     $point$03$i = $hints$idx4$val;
     while(1) {
      $35 = ((($point$03$i)) + 12|0);
      $36 = HEAP16[$35>>1]|0;
      $37 = $36 << 16 >> 16;
      $38 = (_FT_MulFix($37,$31)|0);
      $39 = (($38) + ($32))|0;
      $40 = ((($point$03$i)) + 16|0);
      HEAP32[$40>>2] = $39;
      $41 = ((($point$03$i)) + 40|0);
      $42 = ($41>>>0)<($33>>>0);
      if ($42) {
       $point$03$i = $41;
      } else {
       break;
      }
     }
    }
   } else {
    label = 17;
   }
  } else {
   $23 = ($dim$023|0)==(1);
   if ($23) {
    $24 = HEAP32[$$pre$phiZ2D>>2]|0;
    $25 = $24 & 2;
    $26 = ($25|0)==(0);
    if ($26) {
     label = 17;
    } else {
     break;
    }
   }
  }
  if ((label|0) == 17) {
   label = 0;
   $43 = (((((($hints)) + 44|0) + (($dim$023*1468)|0)|0)) + 20|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = (((((($hints)) + 44|0) + (($dim$023*1468)|0)|0)) + 12|0);
   $46 = HEAP32[$45>>2]|0;
   $47 = (($44) + (($46*48)|0)|0);
   $48 = ($46|0)>(0);
   L30: do {
    if ($48) {
     $anchor$049$i = 0;$edge$050$i = $44;
     while(1) {
      $102 = ((($edge$050$i)) + 12|0);
      $103 = HEAP8[$102>>0]|0;
      $104 = $103 & 4;
      $105 = ($104<<24>>24)==(0);
      do {
       if ($105) {
        $106 = ((($edge$050$i)) + 20|0);
        $107 = HEAP32[$106>>2]|0;
        $108 = ((($edge$050$i)) + 24|0);
        $109 = HEAP32[$108>>2]|0;
        $110 = ($107|0)==(0|0);
        if ($110) {
         $111 = ($109|0)==(0|0);
         if ($111) {
          $anchor$2$i = $anchor$049$i;
          break;
         }
         $112 = ((($109)) + 20|0);
         $113 = HEAP32[$112>>2]|0;
         $114 = ($113|0)==(0|0);
         if ($114) {
          $anchor$2$i = $anchor$049$i;
          break;
         }
         $$phi$trans$insert$i = ((($109)) + 12|0);
         $$pre$i = HEAP8[$$phi$trans$insert$i>>0]|0;
         $120 = $$pre$i;$blue$0$i = $113;$edge1$0$i = $109;$edge2$0$i = $edge$050$i;
        } else {
         $120 = $103;$blue$0$i = $107;$edge1$0$i = $edge$050$i;$edge2$0$i = $109;
        }
        $115 = ((($blue$0$i)) + 8|0);
        $116 = HEAP32[$115>>2]|0;
        $117 = ((($edge1$0$i)) + 8|0);
        HEAP32[$117>>2] = $116;
        $118 = ((($edge1$0$i)) + 12|0);
        $119 = $120&255;
        $121 = $119 | 4;
        $122 = $121&255;
        HEAP8[$118>>0] = $122;
        $123 = ($edge2$0$i|0)==(0|0);
        if (!($123)) {
         $124 = ((($edge2$0$i)) + 20|0);
         $125 = HEAP32[$124>>2]|0;
         $126 = ($125|0)==(0|0);
         if ($126) {
          $edge1$0$idx$i = ((($edge1$0$i)) + 4|0);
          $edge1$0$idx$val$i = HEAP32[$edge1$0$idx$i>>2]|0;
          $127 = ((($edge2$0$i)) + 4|0);
          $128 = HEAP32[$127>>2]|0;
          $129 = (($128) - ($edge1$0$idx$val$i))|0;
          $130 = (_af_cjk_compute_stem_width($hints,$dim$023,$129)|0);
          $131 = (($130) + ($116))|0;
          $132 = ((($edge2$0$i)) + 8|0);
          HEAP32[$132>>2] = $131;
          $133 = ((($edge2$0$i)) + 12|0);
          $134 = HEAP8[$133>>0]|0;
          $135 = $134&255;
          $136 = $135 | 4;
          $137 = $136&255;
          HEAP8[$133>>0] = $137;
         }
        }
        $138 = ($anchor$049$i|0)==(0|0);
        $edge$0$anchor$0$i = $138 ? $edge$050$i : $anchor$049$i;
        $anchor$2$i = $edge$0$anchor$0$i;
       } else {
        $anchor$2$i = $anchor$049$i;
       }
      } while(0);
      $139 = ((($edge$050$i)) + 48|0);
      $140 = ($139>>>0)<($47>>>0);
      if ($140) {
       $anchor$049$i = $anchor$2$i;$edge$050$i = $139;
      } else {
       $anchor$2$i$lcssa = $anchor$2$i;
       break;
      }
     }
     $49 = ($dim$023|0)==(1);
     if ($49) {
      $edge$144$us$i = $44;$has_last_stem$040$us$i = 0;$last_stem_pos$041$us$i = 0;$skipped$039$us$i = 0;
      while(1) {
       $50 = ((($edge$144$us$i)) + 12|0);
       $51 = HEAP8[$50>>0]|0;
       $52 = $51 & 4;
       $53 = ($52<<24>>24)==(0);
       L49: do {
        if ($53) {
         $54 = ((($edge$144$us$i)) + 24|0);
         $55 = HEAP32[$54>>2]|0;
         $56 = ($55|0)==(0|0);
         if ($56) {
          $99 = (($skipped$039$us$i) + 1)|0;
          $has_last_stem$1$us$i = $has_last_stem$040$us$i;$last_stem_pos$1$us$i = $last_stem_pos$041$us$i;$skipped$1$us$i = $99;
          break;
         }
         $57 = ($has_last_stem$040$us$i<<24>>24)==(0);
         do {
          if (!($57)) {
           $58 = ((($edge$144$us$i)) + 8|0);
           $59 = HEAP32[$58>>2]|0;
           $60 = (($last_stem_pos$041$us$i) + 64)|0;
           $61 = ($59|0)<($60|0);
           if (!($61)) {
            $62 = ((($55)) + 8|0);
            $63 = HEAP32[$62>>2]|0;
            $64 = ($63|0)<($60|0);
            if (!($64)) {
             break;
            }
           }
           $65 = (($skipped$039$us$i) + 1)|0;
           $has_last_stem$1$us$i = $has_last_stem$040$us$i;$last_stem_pos$1$us$i = $last_stem_pos$041$us$i;$skipped$1$us$i = $65;
           break L49;
          }
         } while(0);
         $66 = ((($55)) + 20|0);
         $67 = HEAP32[$66>>2]|0;
         $68 = ($67|0)==(0|0);
         if (!($68)) {
          $$idx12$us$i = ((($55)) + 4|0);
          $$idx12$val$us$i = HEAP32[$$idx12$us$i>>2]|0;
          $$idx13$us$i = ((($55)) + 8|0);
          $$idx13$val$us$i = HEAP32[$$idx13$us$i>>2]|0;
          $69 = ((($edge$144$us$i)) + 4|0);
          $70 = HEAP32[$69>>2]|0;
          $71 = (($70) - ($$idx12$val$us$i))|0;
          $72 = (_af_cjk_compute_stem_width($hints,1,$71)|0);
          $73 = (($72) + ($$idx13$val$us$i))|0;
          $74 = ((($edge$144$us$i)) + 8|0);
          HEAP32[$74>>2] = $73;
          $75 = $51&255;
          $76 = $75 | 4;
          $77 = $76&255;
          HEAP8[$50>>0] = $77;
          $has_last_stem$1$us$i = $has_last_stem$040$us$i;$last_stem_pos$1$us$i = $last_stem_pos$041$us$i;$skipped$1$us$i = $skipped$039$us$i;
          break;
         }
         $78 = ($55>>>0)<($edge$144$us$i>>>0);
         if ($78) {
          $$idx10$us$i = ((($55)) + 4|0);
          $$idx10$val$us$i = HEAP32[$$idx10$us$i>>2]|0;
          $$idx11$us$i = ((($55)) + 8|0);
          $$idx11$val$us$i = HEAP32[$$idx11$us$i>>2]|0;
          $90 = ((($edge$144$us$i)) + 4|0);
          $91 = HEAP32[$90>>2]|0;
          $92 = (($91) - ($$idx10$val$us$i))|0;
          $93 = (_af_cjk_compute_stem_width($hints,1,$92)|0);
          $94 = (($93) + ($$idx11$val$us$i))|0;
          $95 = ((($edge$144$us$i)) + 8|0);
          HEAP32[$95>>2] = $94;
          $96 = $51&255;
          $97 = $96 | 4;
          $98 = $97&255;
          HEAP8[$50>>0] = $98;
          $has_last_stem$1$us$i = 1;$last_stem_pos$1$us$i = $94;$skipped$1$us$i = $skipped$039$us$i;
          break;
         } else {
          (_af_hint_normal_stem($hints,$edge$144$us$i,$55,0,1)|0);
          $79 = HEAP8[$50>>0]|0;
          $80 = $79&255;
          $81 = $80 | 4;
          $82 = $81&255;
          HEAP8[$50>>0] = $82;
          $83 = ((($55)) + 12|0);
          $84 = HEAP8[$83>>0]|0;
          $85 = $84&255;
          $86 = $85 | 4;
          $87 = $86&255;
          HEAP8[$83>>0] = $87;
          $88 = ((($55)) + 8|0);
          $89 = HEAP32[$88>>2]|0;
          $has_last_stem$1$us$i = 1;$last_stem_pos$1$us$i = $89;$skipped$1$us$i = $skipped$039$us$i;
          break;
         }
        } else {
         $has_last_stem$1$us$i = $has_last_stem$040$us$i;$last_stem_pos$1$us$i = $last_stem_pos$041$us$i;$skipped$1$us$i = $skipped$039$us$i;
        }
       } while(0);
       $100 = ((($edge$144$us$i)) + 48|0);
       $101 = ($100>>>0)<($47>>>0);
       if ($101) {
        $edge$144$us$i = $100;$has_last_stem$040$us$i = $has_last_stem$1$us$i;$last_stem_pos$041$us$i = $last_stem_pos$1$us$i;$skipped$039$us$i = $skipped$1$us$i;
       } else {
        $skipped$0$lcssa$i = $skipped$1$us$i;
        break L30;
       }
      }
     } else {
      $anchor$343$i = $anchor$2$i$lcssa;$delta$042$i = 0;$edge$144$i = $44;$has_last_stem$040$i = 0;$last_stem_pos$041$i = 0;$skipped$039$i = 0;
     }
     while(1) {
      $141 = ((($edge$144$i)) + 12|0);
      $142 = HEAP8[$141>>0]|0;
      $143 = $142 & 4;
      $144 = ($143<<24>>24)==(0);
      L69: do {
       if ($144) {
        $145 = ((($edge$144$i)) + 24|0);
        $146 = HEAP32[$145>>2]|0;
        $147 = ($146|0)==(0|0);
        if ($147) {
         $148 = (($skipped$039$i) + 1)|0;
         $anchor$4$i = $anchor$343$i;$delta$2$i = $delta$042$i;$has_last_stem$1$i = $has_last_stem$040$i;$last_stem_pos$1$i = $last_stem_pos$041$i;$skipped$1$i = $148;
         break;
        }
        $149 = ($has_last_stem$040$i<<24>>24)==(0);
        do {
         if (!($149)) {
          $150 = ((($edge$144$i)) + 8|0);
          $151 = HEAP32[$150>>2]|0;
          $152 = (($last_stem_pos$041$i) + 64)|0;
          $153 = ($151|0)<($152|0);
          if (!($153)) {
           $154 = ((($146)) + 8|0);
           $155 = HEAP32[$154>>2]|0;
           $156 = ($155|0)<($152|0);
           if (!($156)) {
            break;
           }
          }
          $157 = (($skipped$039$i) + 1)|0;
          $anchor$4$i = $anchor$343$i;$delta$2$i = $delta$042$i;$has_last_stem$1$i = $has_last_stem$040$i;$last_stem_pos$1$i = $last_stem_pos$041$i;$skipped$1$i = $157;
          break L69;
         }
        } while(0);
        $158 = ((($146)) + 20|0);
        $159 = HEAP32[$158>>2]|0;
        $160 = ($159|0)==(0|0);
        if (!($160)) {
         $$idx12$i = ((($146)) + 4|0);
         $$idx12$val$i = HEAP32[$$idx12$i>>2]|0;
         $$idx13$i = ((($146)) + 8|0);
         $$idx13$val$i = HEAP32[$$idx13$i>>2]|0;
         $161 = ((($edge$144$i)) + 4|0);
         $162 = HEAP32[$161>>2]|0;
         $163 = (($162) - ($$idx12$val$i))|0;
         $164 = (_af_cjk_compute_stem_width($hints,$dim$023,$163)|0);
         $165 = (($164) + ($$idx13$val$i))|0;
         $166 = ((($edge$144$i)) + 8|0);
         HEAP32[$166>>2] = $165;
         $167 = $142&255;
         $168 = $167 | 4;
         $169 = $168&255;
         HEAP8[$141>>0] = $169;
         $anchor$4$i = $anchor$343$i;$delta$2$i = $delta$042$i;$has_last_stem$1$i = $has_last_stem$040$i;$last_stem_pos$1$i = $last_stem_pos$041$i;$skipped$1$i = $skipped$039$i;
         break;
        }
        $170 = ($146>>>0)<($edge$144$i>>>0);
        if ($170) {
         $$idx10$i = ((($146)) + 4|0);
         $$idx10$val$i = HEAP32[$$idx10$i>>2]|0;
         $$idx11$i = ((($146)) + 8|0);
         $$idx11$val$i = HEAP32[$$idx11$i>>2]|0;
         $171 = ((($edge$144$i)) + 4|0);
         $172 = HEAP32[$171>>2]|0;
         $173 = (($172) - ($$idx10$val$i))|0;
         $174 = (_af_cjk_compute_stem_width($hints,$dim$023,$173)|0);
         $175 = (($174) + ($$idx11$val$i))|0;
         $176 = ((($edge$144$i)) + 8|0);
         HEAP32[$176>>2] = $175;
         $177 = $142&255;
         $178 = $177 | 4;
         $179 = $178&255;
         HEAP8[$141>>0] = $179;
         $anchor$4$i = $anchor$343$i;$delta$2$i = $delta$042$i;$has_last_stem$1$i = 1;$last_stem_pos$1$i = $175;$skipped$1$i = $skipped$039$i;
         break;
        }
        $180 = ($anchor$343$i|0)==(0|0);
        if ($180) {
         $181 = (_af_hint_normal_stem($hints,$edge$144$i,$146,0,0)|0);
         $delta$1$i = $181;
        } else {
         (_af_hint_normal_stem($hints,$edge$144$i,$146,$delta$042$i,$dim$023)|0);
         $delta$1$i = $delta$042$i;
        }
        $182 = HEAP8[$141>>0]|0;
        $183 = $182&255;
        $184 = $183 | 4;
        $185 = $184&255;
        HEAP8[$141>>0] = $185;
        $186 = ((($146)) + 12|0);
        $187 = HEAP8[$186>>0]|0;
        $188 = $187&255;
        $189 = $188 | 4;
        $190 = $189&255;
        HEAP8[$186>>0] = $190;
        $191 = ((($146)) + 8|0);
        $192 = HEAP32[$191>>2]|0;
        $anchor$4$i = $edge$144$i;$delta$2$i = $delta$1$i;$has_last_stem$1$i = 1;$last_stem_pos$1$i = $192;$skipped$1$i = $skipped$039$i;
       } else {
        $anchor$4$i = $anchor$343$i;$delta$2$i = $delta$042$i;$has_last_stem$1$i = $has_last_stem$040$i;$last_stem_pos$1$i = $last_stem_pos$041$i;$skipped$1$i = $skipped$039$i;
       }
      } while(0);
      $193 = ((($edge$144$i)) + 48|0);
      $194 = ($193>>>0)<($47>>>0);
      if ($194) {
       $anchor$343$i = $anchor$4$i;$delta$042$i = $delta$2$i;$edge$144$i = $193;$has_last_stem$040$i = $has_last_stem$1$i;$last_stem_pos$041$i = $last_stem_pos$1$i;$skipped$039$i = $skipped$1$i;
      } else {
       $skipped$0$lcssa$i = $skipped$1$i;
       break;
      }
     }
    } else {
     $skipped$0$lcssa$i = 0;
    }
   } while(0);
   $$idx$i = ($46*48)|0;
   L92: do {
    if ($19) {
     $195 = ($$idx$i|0)==(576);
     switch ($$idx$i|0) {
     case 288:  {
      $196 = ((($44)) + 96|0);
      $197 = ((($44)) + 192|0);
      $edge13$0$i = $44;$edge24$0$i = $196;$edge3$0$i = $197;
      break;
     }
     case 576:  {
      $198 = ((($44)) + 48|0);
      $199 = ((($44)) + 240|0);
      $200 = ((($44)) + 432|0);
      $edge13$0$i = $198;$edge24$0$i = $199;$edge3$0$i = $200;
      break;
     }
     default: {
      break L92;
     }
     }
     $201 = ((($edge24$0$i)) + 4|0);
     $202 = HEAP32[$201>>2]|0;
     $203 = ((($edge13$0$i)) + 4|0);
     $204 = HEAP32[$203>>2]|0;
     $205 = (($202) - ($204))|0;
     $206 = ((($edge3$0$i)) + 4|0);
     $207 = HEAP32[$206>>2]|0;
     $208 = (($202) - ($207))|0;
     $209 = (($208) + ($205))|0;
     $210 = ($209|0)<(0);
     $211 = (0 - ($209))|0;
     $$9$i = $210 ? $211 : $209;
     $212 = ((($edge13$0$i)) + 24|0);
     $213 = HEAP32[$212>>2]|0;
     $214 = ((($edge13$0$i)) + 48|0);
     $215 = ($213|0)==($214|0);
     if ($215) {
      $216 = ((($edge24$0$i)) + 24|0);
      $217 = HEAP32[$216>>2]|0;
      $218 = ((($edge24$0$i)) + 48|0);
      $219 = ($217|0)==($218|0);
      if ($219) {
       $220 = ((($edge3$0$i)) + 24|0);
       $221 = HEAP32[$220>>2]|0;
       $222 = ((($edge3$0$i)) + 48|0);
       $223 = ($221|0)==($222|0);
       $224 = ($$9$i|0)<(8);
       $or$cond6$i = $224 & $223;
       if ($or$cond6$i) {
        $225 = ((($edge3$0$i)) + 8|0);
        $226 = HEAP32[$225>>2]|0;
        $227 = ((($edge24$0$i)) + 8|0);
        $228 = HEAP32[$227>>2]|0;
        $229 = $228 << 1;
        $230 = ((($edge13$0$i)) + 8|0);
        $231 = HEAP32[$230>>2]|0;
        $232 = (($229) - ($231))|0;
        $233 = (($226) - ($232))|0;
        HEAP32[$225>>2] = $232;
        $234 = ($221|0)==(0|0);
        if (!($234)) {
         $235 = ((($221)) + 8|0);
         $236 = HEAP32[$235>>2]|0;
         $237 = (($236) - ($233))|0;
         HEAP32[$235>>2] = $237;
        }
        if ($195) {
         $238 = ((($44)) + 392|0);
         $239 = HEAP32[$238>>2]|0;
         $240 = (($239) - ($233))|0;
         HEAP32[$238>>2] = $240;
         $241 = ((($44)) + 536|0);
         $242 = HEAP32[$241>>2]|0;
         $243 = (($242) - ($233))|0;
         HEAP32[$241>>2] = $243;
        }
        $244 = ((($edge3$0$i)) + 12|0);
        $245 = HEAP8[$244>>0]|0;
        $246 = $245&255;
        $247 = $246 | 4;
        $248 = $247&255;
        HEAP8[$244>>0] = $248;
        if (!($234)) {
         $249 = ((($221)) + 12|0);
         $250 = HEAP8[$249>>0]|0;
         $251 = $250&255;
         $252 = $251 | 4;
         $253 = $252&255;
         HEAP8[$249>>0] = $253;
        }
       }
      }
     }
    }
   } while(0);
   $254 = ($skipped$0$lcssa$i|0)==(0);
   $$not$i = $48 ^ 1;
   $brmerge$i = $254 | $$not$i;
   if (!($brmerge$i)) {
    $edge$237$i = $44;$skipped$236$i = $skipped$0$lcssa$i;
    while(1) {
     $255 = ((($edge$237$i)) + 12|0);
     $256 = HEAP8[$255>>0]|0;
     $257 = $256 & 4;
     $258 = ($257<<24>>24)==(0);
     if ($258) {
      $259 = ((($edge$237$i)) + 28|0);
      $260 = HEAP32[$259>>2]|0;
      $261 = ($260|0)==(0|0);
      if ($261) {
       $skipped$3$i = $skipped$236$i;
      } else {
       $$idx15$i = ((($260)) + 4|0);
       $$idx15$val$i = HEAP32[$$idx15$i>>2]|0;
       $$idx16$i = ((($260)) + 8|0);
       $$idx16$val$i = HEAP32[$$idx16$i>>2]|0;
       $262 = ((($edge$237$i)) + 4|0);
       $263 = HEAP32[$262>>2]|0;
       $264 = (($$idx16$val$i) - ($$idx15$val$i))|0;
       $265 = (($264) + ($263))|0;
       $266 = ((($edge$237$i)) + 8|0);
       HEAP32[$266>>2] = $265;
       $267 = $256&255;
       $268 = $267 | 4;
       $269 = $268&255;
       HEAP8[$255>>0] = $269;
       $270 = (($skipped$236$i) + -1)|0;
       $skipped$3$i = $270;
      }
     } else {
      $skipped$3$i = $skipped$236$i;
     }
     $271 = ((($edge$237$i)) + 48|0);
     $272 = ($271>>>0)<($47>>>0);
     if ($272) {
      $edge$237$i = $271;$skipped$236$i = $skipped$3$i;
     } else {
      $skipped$3$i$lcssa = $skipped$3$i;
      break;
     }
    }
    $273 = ($skipped$3$i$lcssa|0)!=(0);
    $or$cond$i = $48 & $273;
    if ($or$cond$i) {
     $edge$334$i = $44;
     while(1) {
      $274 = ((($edge$334$i)) + 12|0);
      $275 = HEAP8[$274>>0]|0;
      $276 = $275 & 4;
      $277 = ($276<<24>>24)==(0);
      do {
       if ($277) {
        $before$0$i = $edge$334$i;
        while(1) {
         $278 = ((($before$0$i)) + -48|0);
         $279 = ($278>>>0)<($44>>>0);
         if ($279) {
          $$lcssa = $278;$$lcssa29 = 1;$$lcssa56$i = 0;$before$0$i$lcssa = $before$0$i;
          break;
         }
         $280 = ((($before$0$i)) + -36|0);
         $281 = HEAP8[$280>>0]|0;
         $282 = $281 & 4;
         $283 = ($282<<24>>24)==(0);
         if ($283) {
          $before$0$i = $278;
         } else {
          $$lcssa = $278;$$lcssa29 = 0;$$lcssa56$i = 1;$before$0$i$lcssa = $before$0$i;
          break;
         }
        }
        $after$0$i = $edge$334$i;
        while(1) {
         $284 = ((($after$0$i)) + 48|0);
         $285 = ($284>>>0)<($47>>>0);
         if (!($285)) {
          $$lcssa71 = $284;$after$0$i$lcssa = $after$0$i;
          label = 83;
          break;
         }
         $286 = ((($after$0$i)) + 60|0);
         $287 = HEAP8[$286>>0]|0;
         $288 = $287 & 4;
         $289 = ($288<<24>>24)==(0);
         if ($289) {
          $after$0$i = $284;
         } else {
          $301 = $284;$421 = 1;$after$0$i70 = $after$0$i;
          break;
         }
        }
        if ((label|0) == 83) {
         label = 0;
         if ($$lcssa56$i) {
          $301 = $$lcssa71;$421 = 0;$after$0$i70 = $after$0$i$lcssa;
         } else {
          break;
         }
        }
        if ($$lcssa29) {
         $$idx17$i = ((($after$0$i70)) + 52|0);
         $$idx17$val$i = HEAP32[$$idx17$i>>2]|0;
         $$idx18$i = ((($after$0$i70)) + 56|0);
         $$idx18$val$i = HEAP32[$$idx18$i>>2]|0;
         $290 = ((($edge$334$i)) + 4|0);
         $291 = HEAP32[$290>>2]|0;
         $292 = (($$idx18$val$i) - ($$idx17$val$i))|0;
         $293 = (($292) + ($291))|0;
         $294 = ((($edge$334$i)) + 8|0);
         HEAP32[$294>>2] = $293;
         break;
        }
        if (!($421)) {
         $$idx19$i = ((($before$0$i$lcssa)) + -44|0);
         $$idx19$val$i = HEAP32[$$idx19$i>>2]|0;
         $$idx20$i = ((($before$0$i$lcssa)) + -40|0);
         $$idx20$val$i = HEAP32[$$idx20$i>>2]|0;
         $295 = ((($edge$334$i)) + 4|0);
         $296 = HEAP32[$295>>2]|0;
         $297 = (($$idx20$val$i) - ($$idx19$val$i))|0;
         $298 = (($297) + ($296))|0;
         $299 = ((($edge$334$i)) + 8|0);
         HEAP32[$299>>2] = $298;
         break;
        }
        $300 = HEAP16[$301>>1]|0;
        $302 = HEAP16[$$lcssa>>1]|0;
        $303 = $302 << 16 >> 16;
        $304 = ($300<<16>>16)==($302<<16>>16);
        $305 = ((($before$0$i$lcssa)) + -40|0);
        $306 = HEAP32[$305>>2]|0;
        if ($304) {
         $307 = ((($edge$334$i)) + 8|0);
         HEAP32[$307>>2] = $306;
         break;
        } else {
         $308 = $300 << 16 >> 16;
         $309 = HEAP16[$edge$334$i>>1]|0;
         $310 = $309 << 16 >> 16;
         $311 = (($310) - ($303))|0;
         $312 = ((($after$0$i70)) + 56|0);
         $313 = HEAP32[$312>>2]|0;
         $314 = (($313) - ($306))|0;
         $315 = (($308) - ($303))|0;
         $316 = (_FT_MulDiv($311,$314,$315)|0);
         $317 = (($316) + ($306))|0;
         $318 = ((($edge$334$i)) + 8|0);
         HEAP32[$318>>2] = $317;
         break;
        }
       }
      } while(0);
      $319 = ((($edge$334$i)) + 48|0);
      $320 = ($319>>>0)<($47>>>0);
      if ($320) {
       $edge$334$i = $319;
      } else {
       break;
      }
     }
    }
   }
   $321 = HEAP32[$43>>2]|0;
   $322 = HEAP32[$45>>2]|0;
   $323 = (($321) + (($322*48)|0)|0);
   if ($19) {
    $324 = HEAP32[$18>>2]|0;
    $325 = $324 & 1;
    $$$i = $325&255;
    $331 = $$$i;
   } else {
    $326 = ($dim$023|0)==(1);
    if ($326) {
     $327 = HEAP32[$18>>2]|0;
     $328 = $327 >>> 1;
     $$lobit$i = $328 & 1;
     $phitmp$i = $$lobit$i&255;
     $331 = $phitmp$i;
    } else {
     $331 = 0;
    }
   }
   $329 = ($322|0)>(0);
   if ($329) {
    $330 = ($331<<24>>24)==(0);
    if ($330) {
     $edge$02$i$us = $321;
     while(1) {
      $332 = ((($edge$02$i$us)) + 40|0);
      $333 = HEAP32[$332>>2]|0;
      $334 = ((($edge$02$i$us)) + 8|0);
      $335 = HEAP32[$334>>2]|0;
      $336 = ((($edge$02$i$us)) + 4|0);
      $337 = HEAP32[$336>>2]|0;
      $338 = (($335) - ($337))|0;
      if ($19) {
       $seg$1$us$i$us = $333;
       while(1) {
        $354 = ((($seg$1$us$i$us)) + 40|0);
        $355 = ((($seg$1$us$i$us)) + 44|0);
        $356 = HEAP32[$355>>2]|0;
        $point1$0$in$us$us$i$us = $354;
        while(1) {
         $point1$0$us$us$i$us = HEAP32[$point1$0$in$us$us$i$us>>2]|0;
         $357 = ((($point1$0$us$us$i$us)) + 16|0);
         $358 = HEAP32[$357>>2]|0;
         $359 = (($358) + ($338))|0;
         HEAP32[$357>>2] = $359;
         $360 = HEAP16[$point1$0$us$us$i$us>>1]|0;
         $361 = $360&65535;
         $362 = $361 | 4;
         $363 = $362&65535;
         HEAP16[$point1$0$us$us$i$us>>1] = $363;
         $364 = ($point1$0$us$us$i$us|0)==($356|0);
         $365 = ((($point1$0$us$us$i$us)) + 32|0);
         if ($364) {
          break;
         } else {
          $point1$0$in$us$us$i$us = $365;
         }
        }
        $366 = ((($seg$1$us$i$us)) + 16|0);
        $367 = HEAP32[$366>>2]|0;
        $368 = ($367|0)==($333|0);
        if ($368) {
         break;
        } else {
         $seg$1$us$i$us = $367;
        }
       }
      } else {
       $seg$1$i$us = $333;
       while(1) {
        $339 = ((($seg$1$i$us)) + 40|0);
        $340 = ((($seg$1$i$us)) + 44|0);
        $341 = HEAP32[$340>>2]|0;
        $point1$0$in$i$us = $339;
        while(1) {
         $point1$0$i$us = HEAP32[$point1$0$in$i$us>>2]|0;
         $342 = ((($point1$0$i$us)) + 20|0);
         $343 = HEAP32[$342>>2]|0;
         $344 = (($343) + ($338))|0;
         HEAP32[$342>>2] = $344;
         $345 = HEAP16[$point1$0$i$us>>1]|0;
         $346 = $345&65535;
         $347 = $346 | 8;
         $348 = $347&65535;
         HEAP16[$point1$0$i$us>>1] = $348;
         $349 = ($point1$0$i$us|0)==($341|0);
         $350 = ((($point1$0$i$us)) + 32|0);
         if ($349) {
          break;
         } else {
          $point1$0$in$i$us = $350;
         }
        }
        $351 = ((($seg$1$i$us)) + 16|0);
        $352 = HEAP32[$351>>2]|0;
        $353 = ($352|0)==($333|0);
        if ($353) {
         break;
        } else {
         $seg$1$i$us = $352;
        }
       }
      }
      $369 = ((($edge$02$i$us)) + 48|0);
      $370 = ($369>>>0)<($323>>>0);
      if ($370) {
       $edge$02$i$us = $369;
      } else {
       break;
      }
     }
    } else {
     $edge$02$i = $321;
     while(1) {
      $371 = ((($edge$02$i)) + 40|0);
      $372 = HEAP32[$371>>2]|0;
      $373 = ((($edge$02$i)) + 8|0);
      $374 = HEAP32[$373>>2]|0;
      if ($19) {
       $seg$0$us$i = $372;
       while(1) {
        $375 = ((($seg$0$us$i)) + 40|0);
        $376 = ((($seg$0$us$i)) + 44|0);
        $377 = HEAP32[$376>>2]|0;
        $point$0$in$us$us$i = $375;
        while(1) {
         $point$0$us$us$i = HEAP32[$point$0$in$us$us$i>>2]|0;
         $378 = ((($point$0$us$us$i)) + 16|0);
         HEAP32[$378>>2] = $374;
         $379 = HEAP16[$point$0$us$us$i>>1]|0;
         $380 = $379&65535;
         $381 = $380 | 4;
         $382 = $381&65535;
         HEAP16[$point$0$us$us$i>>1] = $382;
         $383 = ($point$0$us$us$i|0)==($377|0);
         $384 = ((($point$0$us$us$i)) + 32|0);
         if ($383) {
          break;
         } else {
          $point$0$in$us$us$i = $384;
         }
        }
        $385 = ((($seg$0$us$i)) + 16|0);
        $386 = HEAP32[$385>>2]|0;
        $387 = ($386|0)==($372|0);
        if ($387) {
         break;
        } else {
         $seg$0$us$i = $386;
        }
       }
      } else {
       $seg$0$i = $372;
       while(1) {
        $388 = ((($seg$0$i)) + 40|0);
        $389 = ((($seg$0$i)) + 44|0);
        $390 = HEAP32[$389>>2]|0;
        $point$0$in$i = $388;
        while(1) {
         $point$0$i = HEAP32[$point$0$in$i>>2]|0;
         $391 = ((($point$0$i)) + 20|0);
         HEAP32[$391>>2] = $374;
         $392 = HEAP16[$point$0$i>>1]|0;
         $393 = $392&65535;
         $394 = $393 | 8;
         $395 = $394&65535;
         HEAP16[$point$0$i>>1] = $395;
         $396 = ($point$0$i|0)==($390|0);
         $397 = ((($point$0$i)) + 32|0);
         if ($396) {
          break;
         } else {
          $point$0$in$i = $397;
         }
        }
        $398 = ((($seg$0$i)) + 16|0);
        $399 = HEAP32[$398>>2]|0;
        $400 = ($399|0)==($372|0);
        if ($400) {
         break;
        } else {
         $seg$0$i = $399;
        }
       }
      }
      $401 = ((($edge$02$i)) + 48|0);
      $402 = ($401>>>0)<($323>>>0);
      if ($402) {
       $edge$02$i = $401;
      } else {
       break;
      }
     }
    }
   }
   _af_glyph_hints_align_strong_points($hints,$dim$023);
   _af_glyph_hints_align_weak_points($hints,$dim$023);
  }
  $403 = (($dim$023) + 1)|0;
  $exitcond = ($403|0)==(2);
  if ($exitcond) {
   break;
  } else {
   $dim$023 = $403;
  }
 }
 $hints$idx$val = HEAP32[$hints$idx3>>2]|0;
 $hints$idx1$val = HEAP32[$hints$idx4>>2]|0;
 $404 = (($hints$idx1$val) + (($hints$idx$val*40)|0)|0);
 $405 = ($hints$idx$val|0)>(0);
 if (!($405)) {
  $error$2 = 0;
  STACKTOP = sp;return ($error$2|0);
 }
 $outline$idx2 = ((($outline)) + 8|0);
 $outline$idx2$val = HEAP32[$outline$idx2>>2]|0;
 $outline$idx = ((($outline)) + 4|0);
 $outline$idx$val = HEAP32[$outline$idx>>2]|0;
 $point$01$i = $hints$idx1$val;$tag$03$i = $outline$idx2$val;$vec$02$i = $outline$idx$val;
 while(1) {
  $406 = ((($point$01$i)) + 16|0);
  $407 = HEAP32[$406>>2]|0;
  HEAP32[$vec$02$i>>2] = $407;
  $408 = ((($point$01$i)) + 20|0);
  $409 = HEAP32[$408>>2]|0;
  $410 = ((($vec$02$i)) + 4|0);
  HEAP32[$410>>2] = $409;
  $411 = HEAP16[$point$01$i>>1]|0;
  $412 = $411&65535;
  $413 = $412 & 1;
  $414 = ($413|0)==(0);
  do {
   if ($414) {
    $415 = $412 & 2;
    $416 = ($415|0)==(0);
    if ($416) {
     HEAP8[$tag$03$i>>0] = 1;
     break;
    } else {
     HEAP8[$tag$03$i>>0] = 2;
     break;
    }
   } else {
    HEAP8[$tag$03$i>>0] = 0;
   }
  } while(0);
  $417 = ((($point$01$i)) + 40|0);
  $418 = ((($vec$02$i)) + 8|0);
  $419 = ((($tag$03$i)) + 1|0);
  $420 = ($417>>>0)<($404>>>0);
  if ($420) {
   $point$01$i = $417;$tag$03$i = $419;$vec$02$i = $418;
  } else {
   $error$2 = 0;
   break;
  }
 }
 STACKTOP = sp;return ($error$2|0);
}
function _af_cjk_hints_detect_features($hints,$dim) {
 $hints = $hints|0;
 $dim = $dim|0;
 var $$$i = 0, $$$us$i = 0, $$$v$us$i = 0, $$12$i = 0, $$7$i = 0, $$8$i = 0, $$9$i = 0, $$in$i = 0, $$lcssa = 0, $$not$i = 0, $$not$i9 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$i = 0, $$pre$i5 = 0, $$pre$phi$iZ2D = 0, $$pre$phi13Z2D = 0, $$pre$phiZ2D = 0, $$pre12 = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0;
 var $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0;
 var $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0;
 var $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0;
 var $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $best$027$i = 0, $best$3$i = 0, $brmerge$i10 = 0, $dist2$0$i = 0, $dist2$1$i = 0, $dist2$214$i = 0, $edge2$0$i = 0, $edge2$2$i = 0, $edge4$i = 0, $edge7$024$i = 0, $edge7$123$i = 0;
 var $edge_distance_threshold$0$i = 0, $ee$026$i = 0, $error$0 = 0, $exitcond = 0, $f0$03$i = 0, $found$028$i = 0, $found$3$i = 0, $found$3$i$lcssa = 0, $is_round$0$i = 0, $is_round$1$i = 0, $is_round$1$i$lcssa = 0, $is_straight$0$i = 0, $is_straight$1$i = 0, $is_straight$1$i$lcssa = 0, $max$0$us$i = 0, $max$0$v$us$i = 0, $or$cond$i = 0, $or$cond$us$i = 0, $or$cond1$us$i = 0, $or$cond10$i = 0;
 var $or$cond3$us$i = 0, $pt$02$i = 0, $seg$011$us$i = 0, $seg$030$i = 0, $seg$05$i = 0, $seg$1$i = 0, $seg$2$i = 0, $seg1$0$i = 0, $seg1$024$us$i = 0, $seg1$117$i = 0, $seg1$210$i = 0, $seg2$0$i = 0, $seg2$021$us$i = 0, $seg2$113$us$i = 0, $sext$i = 0, $sext31$i = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $edge4$i = sp;
 $0 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (((($hints)) + 44|0) + (($dim*1468)|0)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + (($3*48)|0)|0);
 $5 = (_af_latin_hints_compute_segments($hints,$dim)|0);
 $6 = ($5|0)==(0);
 if (!($6)) {
  $error$0 = $5;
  STACKTOP = sp;return ($error$0|0);
 }
 $7 = ($3|0)>(0);
 if ($7) {
  $seg$05$i = $1;
  while(1) {
   $8 = ((($seg$05$i)) + 40|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = ((($seg$05$i)) + 44|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = HEAP16[$9>>1]|0;
   $13 = HEAP8[$seg$05$i>>0]|0;
   $14 = $13&255;
   $15 = $14 & 254;
   $16 = $15&255;
   HEAP8[$seg$05$i>>0] = $16;
   $17 = ($9|0)==($11|0);
   L7: do {
    if (!($17)) {
     $18 = $12&65535;
     $19 = $18 & 3;
     $f0$03$i = $19;$pt$02$i = $9;
     while(1) {
      $20 = ((($pt$02$i)) + 32|0);
      $21 = HEAP32[$20>>2]|0;
      $22 = HEAP16[$21>>1]|0;
      $23 = $22&65535;
      $24 = $23 & 3;
      $25 = $24 | $f0$03$i;
      $26 = ($25|0)==(0);
      if ($26) {
       break L7;
      }
      $27 = ($21|0)==($11|0);
      if ($27) {
       break;
      } else {
       $f0$03$i = $24;$pt$02$i = $21;
      }
     }
     $28 = $14 | 1;
     $29 = $28&255;
     HEAP8[$seg$05$i>>0] = $29;
    }
   } while(0);
   $30 = ((($seg$05$i)) + 48|0);
   $31 = ($30>>>0)<($4>>>0);
   if ($31) {
    $seg$05$i = $30;
   } else {
    break;
   }
  }
 }
 $32 = HEAP32[$0>>2]|0;
 $33 = HEAP32[$2>>2]|0;
 $34 = (($32) + (($33*48)|0)|0);
 $35 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 24|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = ((($hints)) + 2988|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = ((($38)) + 40|0);
 $40 = HEAP32[$39>>2]|0;
 $41 = (($40|0) / 256)&-1;
 $42 = ($dim|0)==(0);
 $43 = ((($hints)) + 4|0);
 $44 = ((($hints)) + 12|0);
 $$in$i = $42 ? $43 : $44;
 $45 = HEAP32[$$in$i>>2]|0;
 $46 = (_FT_DivFix(192,$45)|0);
 $47 = ($33|0)>(0);
 if ($47) {
  $seg1$024$us$i = $32;
  while(1) {
   $48 = ((($seg1$024$us$i)) + 1|0);
   $49 = HEAP8[$48>>0]|0;
   $50 = $49 << 24 >> 24;
   $51 = ($50|0)==($36|0);
   if ($51) {
    $64 = ((($seg1$024$us$i)) + 2|0);
    $69 = ((($seg1$024$us$i)) + 4|0);
    $84 = ((($seg1$024$us$i)) + 32|0);
    $90 = ((($seg1$024$us$i)) + 36|0);
    $92 = ((($seg1$024$us$i)) + 20|0);
    $seg2$021$us$i = $32;
    while(1) {
     $54 = ($seg2$021$us$i|0)==($seg1$024$us$i|0);
     do {
      if (!($54)) {
       $55 = ((($seg2$021$us$i)) + 1|0);
       $56 = HEAP8[$55>>0]|0;
       $57 = $56 << 24 >> 24;
       $58 = (0 - ($57))|0;
       $59 = ($36|0)==($58|0);
       if ($59) {
        $60 = ((($seg2$021$us$i)) + 2|0);
        $61 = HEAP16[$60>>1]|0;
        $62 = $61 << 16 >> 16;
        $63 = HEAP16[$64>>1]|0;
        $65 = $63 << 16 >> 16;
        $66 = (($62) - ($65))|0;
        $67 = ($66|0)<(0);
        if (!($67)) {
         $68 = HEAP32[$69>>2]|0;
         $70 = $68&65535;
         $71 = $68 >>> 16;
         $72 = $71&65535;
         $73 = ((($seg2$021$us$i)) + 4|0);
         $74 = HEAP32[$73>>2]|0;
         $75 = $74&65535;
         $76 = ($70<<16>>16)<($75<<16>>16);
         $$$v$us$i = $76 ? $74 : $68;
         $sext$i = $$$v$us$i << 16;
         $$$us$i = $sext$i >> 16;
         $77 = $74 >>> 16;
         $78 = $77&65535;
         $79 = ($72<<16>>16)>($78<<16>>16);
         $max$0$v$us$i = $79 ? $77 : $71;
         $sext31$i = $max$0$v$us$i << 16;
         $max$0$us$i = $sext31$i >> 16;
         $80 = (($max$0$us$i) - ($$$us$i))|0;
         $81 = ($80|0)<($41|0);
         if (!($81)) {
          $82 = $66 << 3;
          $83 = HEAP32[$84>>2]|0;
          $85 = ($83*9)|0;
          $86 = ($82|0)<($85|0);
          do {
           if ($86) {
            $87 = ($83*7)|0;
            $88 = ($82|0)<($87|0);
            if (!($88)) {
             $89 = HEAP32[$90>>2]|0;
             $91 = ($89|0)<($80|0);
             if (!($91)) {
              break;
             }
            }
            HEAP32[$84>>2] = $66;
            HEAP32[$90>>2] = $80;
            HEAP32[$92>>2] = $seg2$021$us$i;
           }
          } while(0);
          $93 = ((($seg2$021$us$i)) + 32|0);
          $94 = HEAP32[$93>>2]|0;
          $95 = ($94*9)|0;
          $96 = ($82|0)<($95|0);
          if ($96) {
           $97 = ($94*7)|0;
           $98 = ($82|0)<($97|0);
           $$pre$i = ((($seg2$021$us$i)) + 36|0);
           if (!($98)) {
            $99 = HEAP32[$$pre$i>>2]|0;
            $100 = ($99|0)<($80|0);
            if (!($100)) {
             break;
            }
           }
           HEAP32[$93>>2] = $66;
           HEAP32[$$pre$i>>2] = $80;
           $101 = ((($seg2$021$us$i)) + 20|0);
           HEAP32[$101>>2] = $seg1$024$us$i;
          }
         }
        }
       }
      }
     } while(0);
     $102 = ((($seg2$021$us$i)) + 48|0);
     $103 = ($102>>>0)<($34>>>0);
     if ($103) {
      $seg2$021$us$i = $102;
     } else {
      break;
     }
    }
   }
   $52 = ((($seg1$024$us$i)) + 48|0);
   $53 = ($52>>>0)<($34>>>0);
   if ($53) {
    $seg1$024$us$i = $52;
   } else {
    $seg1$117$i = $32;
    break;
   }
  }
  while(1) {
   $104 = ((($seg1$117$i)) + 20|0);
   $105 = HEAP32[$104>>2]|0;
   $106 = ($105|0)==(0|0);
   L41: do {
    if (!($106)) {
     $107 = ((($105)) + 20|0);
     $108 = HEAP32[$107>>2]|0;
     $109 = ($108|0)==($seg1$117$i|0);
     if ($109) {
      $110 = ((($105)) + 2|0);
      $111 = HEAP16[$110>>1]|0;
      $112 = ((($seg1$117$i)) + 2|0);
      $113 = HEAP16[$112>>1]|0;
      $114 = ($111<<16>>16)>($113<<16>>16);
      if ($114) {
       $115 = ((($seg1$117$i)) + 32|0);
       $116 = HEAP32[$115>>2]|0;
       $$not$i = ($116|0)<($46|0);
       if ($$not$i) {
        $117 = ((($seg1$117$i)) + 36|0);
        $118 = $116 << 2;
        $seg2$113$us$i = $32;
        while(1) {
         $119 = ((($seg2$113$us$i)) + 2|0);
         $120 = HEAP16[$119>>1]|0;
         $121 = ($120<<16>>16)>($113<<16>>16);
         $122 = ($seg1$117$i|0)==($seg2$113$us$i|0);
         $or$cond$us$i = $122 | $121;
         if (!($or$cond$us$i)) {
          $123 = ((($seg2$113$us$i)) + 20|0);
          $124 = HEAP32[$123>>2]|0;
          $125 = ($124|0)==(0|0);
          if (!($125)) {
           $126 = ((($124)) + 20|0);
           $127 = HEAP32[$126>>2]|0;
           $128 = ($127|0)==($seg2$113$us$i|0);
           if ($128) {
            $129 = ((($124)) + 2|0);
            $130 = HEAP16[$129>>1]|0;
            $131 = ($130<<16>>16)<($111<<16>>16);
            if (!($131)) {
             $132 = ($113<<16>>16)==($120<<16>>16);
             $133 = ($111<<16>>16)==($130<<16>>16);
             $or$cond1$us$i = $132 & $133;
             if (!($or$cond1$us$i)) {
              $134 = ((($seg2$113$us$i)) + 32|0);
              $135 = HEAP32[$134>>2]|0;
              $136 = ($135|0)>($116|0);
              $137 = ($118|0)>($135|0);
              $or$cond3$us$i = $136 & $137;
              if ($or$cond3$us$i) {
               $138 = HEAP32[$117>>2]|0;
               $139 = ((($seg2$113$us$i)) + 36|0);
               $140 = HEAP32[$139>>2]|0;
               $141 = ($140*3)|0;
               $142 = ($138|0)<($141|0);
               if ($142) {
                break;
               } else {
                $seg$011$us$i = $32;
               }
               while(1) {
                $143 = ((($seg$011$us$i)) + 20|0);
                $144 = HEAP32[$143>>2]|0;
                $145 = ($144|0)==($seg2$113$us$i|0);
                do {
                 if ($145) {
                  HEAP32[$143>>2] = 0;
                  $148 = ((($seg$011$us$i)) + 24|0);
                  HEAP32[$148>>2] = $105;
                 } else {
                  $146 = ($144|0)==($124|0);
                  if (!($146)) {
                   break;
                  }
                  HEAP32[$143>>2] = 0;
                  $147 = ((($seg$011$us$i)) + 24|0);
                  HEAP32[$147>>2] = $seg1$117$i;
                 }
                } while(0);
                $149 = ((($seg$011$us$i)) + 48|0);
                $150 = ($149>>>0)<($34>>>0);
                if ($150) {
                 $seg$011$us$i = $149;
                } else {
                 break;
                }
               }
              }
             }
            }
           }
          }
         }
         $151 = ((($seg2$113$us$i)) + 48|0);
         $152 = ($151>>>0)<($34>>>0);
         if ($152) {
          $seg2$113$us$i = $151;
         } else {
          break L41;
         }
        }
        HEAP32[$107>>2] = 0;
        HEAP32[$104>>2] = 0;
       }
      }
     }
    }
   } while(0);
   $153 = ((($seg1$117$i)) + 48|0);
   $154 = ($153>>>0)<($34>>>0);
   if ($154) {
    $seg1$117$i = $153;
   } else {
    $seg1$210$i = $32;
    break;
   }
  }
  while(1) {
   $155 = ((($seg1$210$i)) + 20|0);
   $156 = HEAP32[$155>>2]|0;
   $157 = ($156|0)==(0|0);
   do {
    if (!($157)) {
     $158 = ((($156)) + 28|0);
     $159 = HEAP32[$158>>2]|0;
     $160 = (($159) + 1)|0;
     HEAP32[$158>>2] = $160;
     $161 = ((($156)) + 20|0);
     $162 = HEAP32[$161>>2]|0;
     $163 = ($162|0)==($seg1$210$i|0);
     if (!($163)) {
      HEAP32[$155>>2] = 0;
      $164 = ((($156)) + 32|0);
      $165 = HEAP32[$164>>2]|0;
      $166 = ($165|0)<($46|0);
      if (!($166)) {
       $167 = ((($seg1$210$i)) + 32|0);
       $168 = HEAP32[$167>>2]|0;
       $169 = $165 << 2;
       $170 = ($168|0)<($169|0);
       if (!($170)) {
        HEAP32[$158>>2] = $159;
        break;
       }
      }
      $171 = HEAP32[$161>>2]|0;
      $172 = ((($seg1$210$i)) + 24|0);
      HEAP32[$172>>2] = $171;
     }
    }
   } while(0);
   $173 = ((($seg1$210$i)) + 48|0);
   $174 = ($173>>>0)<($34>>>0);
   if ($174) {
    $seg1$210$i = $173;
   } else {
    break;
   }
  }
 }
 $175 = (((($hints)) + 44|0) + (($dim*1468)|0)|0);
 $176 = HEAP32[$hints>>2]|0;
 $177 = HEAP32[$37>>2]|0;
 $178 = HEAP32[$0>>2]|0;
 $179 = HEAP32[$175>>2]|0;
 $180 = (($178) + (($179*48)|0)|0);
 $181 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 12|0);
 HEAP32[$181>>2] = 0;
 $182 = HEAP32[$$in$i>>2]|0;
 $183 = (((((($177)) + 44|0) + (($dim*1572)|0)|0)) + 204|0);
 $184 = HEAP32[$183>>2]|0;
 $185 = (_FT_MulFix($184,$182)|0);
 $186 = ($185|0)>(16);
 if ($186) {
  $187 = (_FT_DivFix(16,$182)|0);
  $edge_distance_threshold$0$i = $187;
 } else {
  $188 = HEAP32[$183>>2]|0;
  $edge_distance_threshold$0$i = $188;
 }
 $189 = ($179|0)>(0);
 L80: do {
  if ($189) {
   $190 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 20|0);
   $seg$030$i = $178;
   while(1) {
    $191 = HEAP32[$181>>2]|0;
    $192 = ($191|0)>(0);
    if ($192) {
     $193 = HEAP32[$190>>2]|0;
     $194 = ((($seg$030$i)) + 1|0);
     $195 = HEAP8[$194>>0]|0;
     $196 = ((($seg$030$i)) + 2|0);
     $197 = ((($seg$030$i)) + 20|0);
     $best$027$i = 65535;$ee$026$i = 0;$found$028$i = 0;
     while(1) {
      $198 = (($193) + (($ee$026$i*48)|0)|0);
      $199 = (((($193) + (($ee$026$i*48)|0)|0)) + 13|0);
      $200 = HEAP8[$199>>0]|0;
      $201 = ($200<<24>>24)==($195<<24>>24);
      do {
       if ($201) {
        $202 = HEAP16[$196>>1]|0;
        $203 = $202 << 16 >> 16;
        $204 = HEAP16[$198>>1]|0;
        $205 = $204 << 16 >> 16;
        $206 = (($203) - ($205))|0;
        $207 = ($206|0)<(0);
        $208 = (0 - ($206))|0;
        $$$i = $207 ? $208 : $206;
        $209 = ($$$i|0)<($edge_distance_threshold$0$i|0);
        $210 = ($$$i|0)<($best$027$i|0);
        $or$cond$i = $209 & $210;
        if ($or$cond$i) {
         $211 = HEAP32[$197>>2]|0;
         $212 = ($211|0)==(0|0);
         if (!($212)) {
          $213 = (((($193) + (($ee$026$i*48)|0)|0)) + 40|0);
          $214 = HEAP32[$213>>2]|0;
          $215 = ((($211)) + 2|0);
          $dist2$0$i = 0;$seg1$0$i = $214;
          while(1) {
           $216 = ((($seg1$0$i)) + 20|0);
           $217 = HEAP32[$216>>2]|0;
           $218 = ($217|0)==(0|0);
           if ($218) {
            $dist2$1$i = $dist2$0$i;
           } else {
            $219 = HEAP16[$215>>1]|0;
            $220 = $219 << 16 >> 16;
            $221 = ((($217)) + 2|0);
            $222 = HEAP16[$221>>1]|0;
            $223 = $222 << 16 >> 16;
            $224 = ($219<<16>>16)>($222<<16>>16);
            $225 = (($220) - ($223))|0;
            $226 = (($223) - ($220))|0;
            $227 = $224 ? $225 : $226;
            $228 = ($227|0)<($edge_distance_threshold$0$i|0);
            if ($228) {
             $dist2$1$i = $227;
            } else {
             $dist2$214$i = $227;
             break;
            }
           }
           $229 = ((($seg1$0$i)) + 16|0);
           $230 = HEAP32[$229>>2]|0;
           $231 = ($230|0)==($214|0);
           if ($231) {
            $dist2$214$i = $dist2$1$i;
            break;
           } else {
            $dist2$0$i = $dist2$1$i;$seg1$0$i = $230;
           }
          }
          $232 = ($dist2$214$i|0)<($edge_distance_threshold$0$i|0);
          if (!($232)) {
           $best$3$i = $best$027$i;$found$3$i = $found$028$i;
           break;
          }
         }
         $best$3$i = $$$i;$found$3$i = $198;
        } else {
         $best$3$i = $best$027$i;$found$3$i = $found$028$i;
        }
       } else {
        $best$3$i = $best$027$i;$found$3$i = $found$028$i;
       }
      } while(0);
      $233 = (($ee$026$i) + 1)|0;
      $exitcond = ($233|0)==($191|0);
      if ($exitcond) {
       $found$3$i$lcssa = $found$3$i;
       break;
      } else {
       $best$027$i = $best$3$i;$ee$026$i = $233;$found$028$i = $found$3$i;
      }
     }
     $234 = ($found$3$i$lcssa|0)==(0|0);
     if ($234) {
      $$pre$phi13Z2D = $194;$$pre$phiZ2D = $196;$238 = $195;
      label = 73;
     } else {
      $252 = ((($found$3$i$lcssa)) + 40|0);
      $253 = HEAP32[$252>>2]|0;
      $254 = ((($seg$030$i)) + 16|0);
      HEAP32[$254>>2] = $253;
      $255 = ((($found$3$i$lcssa)) + 44|0);
      $256 = HEAP32[$255>>2]|0;
      $257 = ((($256)) + 16|0);
      HEAP32[$257>>2] = $seg$030$i;
      HEAP32[$255>>2] = $seg$030$i;
     }
    } else {
     $$phi$trans$insert = ((($seg$030$i)) + 1|0);
     $$pre = HEAP8[$$phi$trans$insert>>0]|0;
     $$pre12 = ((($seg$030$i)) + 2|0);
     $$pre$phi13Z2D = $$phi$trans$insert;$$pre$phiZ2D = $$pre12;$238 = $$pre;
     label = 73;
    }
    if ((label|0) == 73) {
     label = 0;
     $235 = HEAP16[$$pre$phiZ2D>>1]|0;
     $236 = $235 << 16 >> 16;
     $237 = $238 << 24 >> 24;
     $239 = (_af_axis_hints_new_edge($175,$236,$237,$176,$edge4$i)|0);
     $240 = ($239|0)==(0);
     if (!($240)) {
      $$lcssa = $239;
      break;
     }
     $241 = HEAP32[$edge4$i>>2]|0;
     dest=$241; stop=dest+40|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
     $242 = ((($241)) + 40|0);
     HEAP32[$242>>2] = $seg$030$i;
     $243 = ((($241)) + 44|0);
     HEAP32[$243>>2] = $seg$030$i;
     $244 = HEAP8[$$pre$phi13Z2D>>0]|0;
     $245 = ((($241)) + 13|0);
     HEAP8[$245>>0] = $244;
     $246 = HEAP16[$$pre$phiZ2D>>1]|0;
     HEAP16[$241>>1] = $246;
     $247 = $246 << 16 >> 16;
     $248 = (_FT_MulFix($247,$182)|0);
     $249 = ((($241)) + 4|0);
     HEAP32[$249>>2] = $248;
     $250 = ((($241)) + 8|0);
     HEAP32[$250>>2] = $248;
     $251 = ((($seg$030$i)) + 16|0);
     HEAP32[$251>>2] = $seg$030$i;
    }
    $258 = ((($seg$030$i)) + 48|0);
    $259 = ($258>>>0)<($180>>>0);
    if ($259) {
     $seg$030$i = $258;
    } else {
     $$pre$phi$iZ2D = $190;
     break L80;
    }
   }
   $error$0 = $$lcssa;
   STACKTOP = sp;return ($error$0|0);
  } else {
   $$pre$i5 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 20|0);
   $$pre$phi$iZ2D = $$pre$i5;
  }
 } while(0);
 $260 = HEAP32[$$pre$phi$iZ2D>>2]|0;
 $261 = HEAP32[$181>>2]|0;
 $262 = (($260) + (($261*48)|0)|0);
 $263 = ($261|0)>(0);
 if ($263) {
  $edge7$024$i = $260;
 } else {
  $error$0 = 0;
  STACKTOP = sp;return ($error$0|0);
 }
 while(1) {
  $264 = ((($edge7$024$i)) + 40|0);
  $265 = HEAP32[$264>>2]|0;
  $266 = ($265|0)==(0|0);
  if (!($266)) {
   $seg$1$i = $265;
   while(1) {
    $267 = ((($seg$1$i)) + 12|0);
    HEAP32[$267>>2] = $edge7$024$i;
    $268 = ((($seg$1$i)) + 16|0);
    $269 = HEAP32[$268>>2]|0;
    $270 = ($269|0)==($265|0);
    if ($270) {
     break;
    } else {
     $seg$1$i = $269;
    }
   }
  }
  $271 = ((($edge7$024$i)) + 48|0);
  $272 = ($271>>>0)<($262>>>0);
  if ($272) {
   $edge7$024$i = $271;
  } else {
   $edge7$123$i = $260;
   break;
  }
 }
 while(1) {
  $273 = ((($edge7$123$i)) + 40|0);
  $274 = HEAP32[$273>>2]|0;
  $275 = ((($edge7$123$i)) + 24|0);
  $276 = ((($edge7$123$i)) + 28|0);
  $is_round$0$i = 0;$is_straight$0$i = 0;$seg$2$i = $274;
  while(1) {
   $277 = HEAP32[$seg$2$i>>2]|0;
   $278 = $277 & 1;
   $279 = $278 ^ 1;
   $is_round$1$i = (($278) + ($is_round$0$i))|0;
   $is_straight$1$i = (($279) + ($is_straight$0$i))|0;
   $280 = ((($seg$2$i)) + 24|0);
   $281 = HEAP32[$280>>2]|0;
   $282 = ($281|0)==(0|0);
   $283 = $277 >>> 16;
   $284 = $283&65535;
   if ($282) {
    $290 = 0;
   } else {
    $285 = ((($281)) + 12|0);
    $286 = HEAP32[$285>>2]|0;
    $287 = ($286|0)!=($edge7$123$i|0);
    $290 = $287;
   }
   $288 = ((($seg$2$i)) + 20|0);
   $289 = HEAP32[$288>>2]|0;
   $$not$i9 = ($289|0)!=(0|0);
   $brmerge$i10 = $290 | $$not$i9;
   do {
    if ($brmerge$i10) {
     $$7$i = $290 ? $276 : $275;
     $$8$i = $290 ? $280 : $288;
     $seg2$0$i = HEAP32[$$8$i>>2]|0;
     $edge2$0$i = HEAP32[$$7$i>>2]|0;
     $291 = ($edge2$0$i|0)==(0|0);
     if ($291) {
      $310 = ((($seg2$0$i)) + 12|0);
      $311 = HEAP32[$310>>2]|0;
      $edge2$2$i = $311;
     } else {
      $292 = HEAP16[$edge7$123$i>>1]|0;
      $293 = $292 << 16 >> 16;
      $294 = HEAP16[$edge2$0$i>>1]|0;
      $295 = $294 << 16 >> 16;
      $296 = (($293) - ($295))|0;
      $297 = ($296|0)<(0);
      $298 = (0 - ($296))|0;
      $$9$i = $297 ? $298 : $296;
      $299 = $277 >> 16;
      $300 = ((($seg2$0$i)) + 2|0);
      $301 = HEAP16[$300>>1]|0;
      $302 = $301 << 16 >> 16;
      $303 = ($284<<16>>16)>($301<<16>>16);
      $304 = (($299) - ($302))|0;
      $305 = (($302) - ($299))|0;
      $306 = $303 ? $304 : $305;
      $307 = ($306|0)<($$9$i|0);
      if ($307) {
       $308 = ((($seg2$0$i)) + 12|0);
       $309 = HEAP32[$308>>2]|0;
       $edge2$2$i = $309;
      } else {
       $edge2$2$i = $edge2$0$i;
      }
     }
     if ($290) {
      HEAP32[$276>>2] = $edge2$2$i;
      $312 = ((($edge2$2$i)) + 12|0);
      $313 = HEAP8[$312>>0]|0;
      $314 = $313&255;
      $315 = $314 | 2;
      $316 = $315&255;
      HEAP8[$312>>0] = $316;
      break;
     } else {
      HEAP32[$275>>2] = $edge2$2$i;
      break;
     }
    }
   } while(0);
   $317 = ((($seg$2$i)) + 16|0);
   $318 = HEAP32[$317>>2]|0;
   $319 = ($318|0)==($274|0);
   if ($319) {
    $is_round$1$i$lcssa = $is_round$1$i;$is_straight$1$i$lcssa = $is_straight$1$i;
    break;
   } else {
    $is_round$0$i = $is_round$1$i;$is_straight$0$i = $is_straight$1$i;$seg$2$i = $318;
   }
  }
  $320 = ((($edge7$123$i)) + 12|0);
  $321 = ($is_round$1$i$lcssa|0)<(1);
  $322 = ($is_round$1$i$lcssa|0)<($is_straight$1$i$lcssa|0);
  $or$cond10$i = $321 | $322;
  $323 = $or$cond10$i&1;
  $$12$i = $323 ^ 1;
  HEAP8[$320>>0] = $$12$i;
  $324 = HEAP32[$276>>2]|0;
  $325 = ($324|0)==(0|0);
  if (!($325)) {
   $326 = HEAP32[$275>>2]|0;
   $327 = ($326|0)==(0|0);
   if (!($327)) {
    HEAP32[$276>>2] = 0;
   }
  }
  $328 = ((($edge7$123$i)) + 48|0);
  $329 = ($328>>>0)<($262>>>0);
  if ($329) {
   $edge7$123$i = $328;
  } else {
   $error$0 = 0;
   break;
  }
 }
 STACKTOP = sp;return ($error$0|0);
}
function _af_cjk_hints_compute_blue_edges($hints,$metrics,$dim) {
 $hints = $hints|0;
 $metrics = $metrics|0;
 $dim = $dim|0;
 var $$ = 0, $$1 = 0, $$lcssa = 0, $$pre = 0, $$pre7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0, $9 = 0, $bb$04 = 0, $best_blue$02 = 0, $best_blue$1 = 0, $best_blue$3 = 0, $best_blue$3$lcssa = 0, $best_dist$03 = 0, $best_dist$1 = 0, $best_dist$3 = 0, $compare$0 = 0, $edge$05 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + (($3*48)|0)|0);
 $5 = (((($metrics)) + 44|0) + (($dim*1572)|0)|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($metrics)) + 40|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (($8>>>0) / 40)&-1;
 $10 = (_FT_MulFix($9,$6)|0);
 $11 = ($10|0)>(32);
 $$ = $11 ? 32 : $10;
 $12 = ($3|0)>(0);
 if (!($12)) {
  return;
 }
 $13 = (((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 216|0);
 $14 = (((((($hints)) + 44|0) + (($dim*1468)|0)|0)) + 24|0);
 $$pre = HEAP32[$13>>2]|0;
 $16 = $$pre;$57 = $$pre;$edge$05 = $1;
 while(1) {
  $15 = ($16|0)==(0);
  if ($15) {
   $58 = 0;$59 = $57;
  } else {
   $17 = ((($edge$05)) + 13|0);
   $60 = $57;$bb$04 = 0;$best_blue$02 = 0;$best_dist$03 = $$;
   while(1) {
    $18 = ((((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 220|0) + (($bb$04*28)|0)|0);
    $19 = ((((((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 220|0) + (($bb$04*28)|0)|0)) + 24|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = $20 & 1;
    $22 = ($21|0)==(0);
    if ($22) {
     $52 = $60;$best_blue$3 = $best_blue$02;$best_dist$3 = $best_dist$03;
    } else {
     $23 = HEAP8[$17>>0]|0;
     $24 = $23 << 24 >> 24;
     $25 = HEAP32[$14>>2]|0;
     $26 = ($24|0)==($25|0);
     $27 = $20 & 2;
     $28 = ($27|0)!=(0);
     $29 = $28 ^ $26;
     if ($29) {
      $30 = HEAP16[$edge$05>>1]|0;
      $31 = $30 << 16 >> 16;
      $32 = HEAP32[$18>>2]|0;
      $33 = (($31) - ($32))|0;
      $34 = ($33|0)<(0);
      $35 = (0 - ($33))|0;
      $36 = $34 ? $35 : $33;
      $37 = ((((((((($metrics)) + 44|0) + (($dim*1572)|0)|0)) + 220|0) + (($bb$04*28)|0)|0)) + 12|0);
      $38 = HEAP32[$37>>2]|0;
      $39 = (($31) - ($38))|0;
      $40 = ($39|0)<(0);
      $41 = (0 - ($39))|0;
      $42 = $40 ? $41 : $39;
      $43 = ($36|0)>($42|0);
      $compare$0 = $43 ? $37 : $18;
      $44 = HEAP32[$compare$0>>2]|0;
      $45 = (($31) - ($44))|0;
      $46 = ($45|0)<(0);
      $47 = (0 - ($45))|0;
      $$1 = $46 ? $47 : $45;
      $48 = (_FT_MulFix($$1,$6)|0);
      $49 = ($48|0)<($best_dist$03|0);
      $best_blue$1 = $49 ? $compare$0 : $best_blue$02;
      $best_dist$1 = $49 ? $48 : $best_dist$03;
      $$pre7 = HEAP32[$13>>2]|0;
      $52 = $$pre7;$best_blue$3 = $best_blue$1;$best_dist$3 = $best_dist$1;
     } else {
      $52 = $60;$best_blue$3 = $best_blue$02;$best_dist$3 = $best_dist$03;
     }
    }
    $50 = (($bb$04) + 1)|0;
    $51 = ($50>>>0)<($52>>>0);
    if ($51) {
     $60 = $52;$bb$04 = $50;$best_blue$02 = $best_blue$3;$best_dist$03 = $best_dist$3;
    } else {
     $$lcssa = $52;$best_blue$3$lcssa = $best_blue$3;
     break;
    }
   }
   $53 = ($best_blue$3$lcssa|0)==(0|0);
   if ($53) {
    $58 = $$lcssa;$59 = $$lcssa;
   } else {
    $54 = ((($edge$05)) + 20|0);
    HEAP32[$54>>2] = $best_blue$3$lcssa;
    $58 = $$lcssa;$59 = $$lcssa;
   }
  }
  $55 = ((($edge$05)) + 48|0);
  $56 = ($55>>>0)<($4>>>0);
  if ($56) {
   $16 = $58;$57 = $59;$edge$05 = $55;
  } else {
   break;
  }
 }
 return;
}
function _af_cjk_compute_stem_width($hints,$dim,$width) {
 $hints = $hints|0;
 $dim = $dim|0;
 $width = $width|0;
 var $$ = 0, $$$i = 0, $$$width = 0, $$0 = 0, $$off = 0, $$width = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $8 = 0;
 var $9 = 0, $best$02$i = 0, $best$1$i = 0, $dist$2 = 0, $dist$2$ = 0, $exitcond$i = 0, $n$04$i = 0, $reference$0$lcssa5$i = 0, $reference$0$width$i = 0, $reference$0$width1$i = 0, $reference$03$i = 0, $reference$1$i = 0, $reference$1$i$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($hints)) + 2988|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($dim|0)==(1);
 $3 = ((($hints)) + 2984|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 & 4;
 $6 = ($5|0)==(0);
 if ($6) {
  $$0 = $width;
  return ($$0|0);
 }
 $7 = ($width|0)<(0);
 $8 = (0 - ($width))|0;
 $$width = $7 ? $8 : $width;
 if ($2) {
  $9 = $4 & 2;
  $10 = ($9|0)==(0);
  if ($10) {
   label = 5;
  } else {
   label = 15;
  }
 } else {
  $11 = $4 & 1;
  $12 = ($11|0)==(0);
  if ($12) {
   label = 5;
  } else {
   label = 15;
  }
 }
 do {
  if ((label|0) == 5) {
   $13 = (((((($1)) + 44|0) + (($dim*1572)|0)|0)) + 8|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ($14|0)==(0);
   if (!($15)) {
    $16 = (((((($1)) + 44|0) + (($dim*1572)|0)|0)) + 16|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = (($$width) - ($17))|0;
    $19 = ($18|0)<(0);
    $20 = (0 - ($18))|0;
    $21 = $19 ? $20 : $18;
    $22 = ($21|0)<(40);
    if ($22) {
     $23 = ($17|0)<(48);
     $$ = $23 ? 48 : $17;
     $dist$2 = $$;
     break;
    }
   }
   $24 = ($$width|0)<(54);
   if ($24) {
    $25 = (54 - ($$width))|0;
    $26 = (($25|0) / 2)&-1;
    $27 = (($26) + ($$width))|0;
    $dist$2 = $27;
    break;
   }
   $28 = ($$width|0)<(192);
   if ($28) {
    $29 = $$width & 63;
    $30 = $$width & -64;
    $31 = ($29>>>0)<(10);
    if ($31) {
     $dist$2 = $$width;
    } else {
     $32 = ($29>>>0)<(22);
     if ($32) {
      $33 = $30 | 10;
      $dist$2 = $33;
      break;
     } else {
      $$off = (($29) + -42)|0;
      $34 = ($$off>>>0)<(12);
      $35 = $30 | 54;
      $$$width = $34 ? $35 : $$width;
      $dist$2 = $$$width;
      break;
     }
    }
   } else {
    $dist$2 = $$width;
   }
  }
  else if ((label|0) == 15) {
   $36 = (((((($1)) + 44|0) + (($dim*1572)|0)|0)) + 8|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = ($37|0)==(0);
   if ($38) {
    $39 = (($$width) + 32)|0;
    $40 = $39 & -64;
    $52 = $40;$reference$0$lcssa5$i = $$width;
    label = 19;
   } else {
    $best$02$i = 98;$n$04$i = 0;$reference$03$i = $$width;
    while(1) {
     $41 = ((((((((($1)) + 44|0) + (($dim*1572)|0)|0)) + 12|0) + (($n$04$i*12)|0)|0)) + 4|0);
     $42 = HEAP32[$41>>2]|0;
     $43 = (($$width) - ($42))|0;
     $44 = ($43|0)<(0);
     $45 = (0 - ($43))|0;
     $$$i = $44 ? $45 : $43;
     $46 = ($$$i|0)<($best$02$i|0);
     $best$1$i = $46 ? $$$i : $best$02$i;
     $reference$1$i = $46 ? $42 : $reference$03$i;
     $47 = (($n$04$i) + 1)|0;
     $exitcond$i = ($47|0)==($37|0);
     if ($exitcond$i) {
      $reference$1$i$lcssa = $reference$1$i;
      break;
     } else {
      $best$02$i = $best$1$i;$n$04$i = $47;$reference$03$i = $reference$1$i;
     }
    }
    $48 = (($reference$1$i$lcssa) + 32)|0;
    $49 = $48 & -64;
    $50 = ($reference$1$i$lcssa|0)>($$width|0);
    if ($50) {
     $54 = (($49) + -48)|0;
     $55 = ($54|0)<($$width|0);
     $reference$0$width1$i = $55 ? $reference$1$i$lcssa : $$width;
     $57 = $reference$0$width1$i;
    } else {
     $52 = $49;$reference$0$lcssa5$i = $reference$1$i$lcssa;
     label = 19;
    }
   }
   if ((label|0) == 19) {
    $51 = $52 | 48;
    $53 = ($51|0)>($$width|0);
    $reference$0$width$i = $53 ? $reference$0$lcssa5$i : $$width;
    $57 = $reference$0$width$i;
   }
   if ($2) {
    $56 = ($57|0)>(63);
    if (!($56)) {
     $dist$2 = 64;
     break;
    }
    $58 = (($57) + 16)|0;
    $59 = $58 & -64;
    $dist$2 = $59;
    break;
   }
   $60 = $4 & 8;
   $61 = ($60|0)==(0);
   if (!($61)) {
    $62 = ($57|0)<(64);
    if ($62) {
     $dist$2 = 64;
     break;
    }
    $63 = (($57) + 32)|0;
    $64 = $63 & -64;
    $dist$2 = $64;
    break;
   }
   $65 = ($57|0)<(48);
   if ($65) {
    $66 = (($57) + 64)|0;
    $67 = $66 >> 1;
    $dist$2 = $67;
    break;
   }
   $68 = ($57|0)<(128);
   if ($68) {
    $69 = (($57) + 22)|0;
    $70 = $69 & -64;
    $dist$2 = $70;
    break;
   } else {
    $71 = (($57) + 32)|0;
    $72 = $71 & -64;
    $dist$2 = $72;
    break;
   }
  }
 } while(0);
 $73 = ($width|0)>(-1);
 $74 = (0 - ($dist$2))|0;
 $dist$2$ = $73 ? $dist$2 : $74;
 $$0 = $dist$2$;
 return ($$0|0);
}
function _af_hint_normal_stem($hints,$edge,$edge2,$anchor,$dim) {
 $hints = $hints|0;
 $edge = $edge|0;
 $edge2 = $edge2|0;
 $anchor = $anchor|0;
 $dim = $dim|0;
 var $$ = 0, $$10 = 0, $$11 = 0, $$2 = 0, $$7 = 0, $$8 = 0, $$delta$0 = 0, $$neg = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, $delta$0 = 0, $delta$1 = 0, $offset$0 = 0, $or$cond = 0, $or$cond3 = 0;
 var $or$cond4 = 0, $or$cond5 = 0, $or$cond6 = 0, $threshold$0 = 0, $u_off2$0 = 0, $u_off2$0$$7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($hints)) + 2984|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 4;
 $3 = ($2|0)==(0);
 do {
  if ($3) {
   $4 = ((($edge)) + 12|0);
   $5 = HEAP8[$4>>0]|0;
   $6 = $5 & 1;
   $7 = ($6<<24>>24)==(0);
   if (!($7)) {
    $8 = ((($edge2)) + 12|0);
    $9 = HEAP8[$8>>0]|0;
    $10 = $9 & 1;
    $11 = ($10<<24>>24)==(0);
    if (!($11)) {
     $12 = ($dim|0)==(1);
     $$ = $12 ? 55 : 49;
     $threshold$0 = $$;
     break;
    }
   }
   $13 = ($dim|0)==(1);
   $$2 = $13 ? 61 : 59;
   $threshold$0 = $$2;
  } else {
   $threshold$0 = 64;
  }
 } while(0);
 $14 = ((($edge2)) + 4|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($edge)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = (($15) - ($17))|0;
 $19 = (_af_cjk_compute_stem_width($hints,$dim,$18)|0);
 $20 = (($17) + ($15))|0;
 $21 = (($20|0) / 2)&-1;
 $22 = (($21) + ($anchor))|0;
 $$neg = (($19|0) / -2)&-1;
 $23 = (($22) + ($$neg))|0;
 $24 = (($23) + ($19))|0;
 $25 = $23 & 63;
 $26 = $24 & 63;
 $27 = (64 - ($25))|0;
 $28 = (64 - ($26))|0;
 $29 = ($25|0)==(0);
 $30 = ($26|0)==(0);
 $or$cond = $29 | $30;
 do {
  if ($or$cond) {
   $delta$0 = 0;
  } else {
   $31 = ($19|0)>($threshold$0|0);
   if (!($31)) {
    $32 = ($26|0)<($19|0);
    if (!($32)) {
     $delta$0 = 0;
     break;
    }
    $33 = ($27>>>0)>($26>>>0);
    $34 = (0 - ($26))|0;
    $$8 = $33 ? $34 : $27;
    $delta$0 = $$8;
    break;
   }
   $35 = ($threshold$0>>>0)<(64);
   if ($35) {
    $36 = ($25>>>0)<($threshold$0>>>0);
    $37 = ($threshold$0>>>0)>($27>>>0);
    $or$cond3 = $36 & $37;
    $38 = ($threshold$0>>>0)>($26>>>0);
    $or$cond4 = $38 & $or$cond3;
    $39 = ($28>>>0)<($threshold$0>>>0);
    $or$cond5 = $39 & $or$cond4;
    if (!($or$cond5)) {
     $delta$0 = 0;
     break;
    }
   }
   $40 = $19 & 63;
   $41 = ($40>>>0)<(32);
   if ($41) {
    $42 = ($27>>>0)>($40>>>0);
    $43 = ($26>>>0)>($40>>>0);
    $or$cond6 = $42 & $43;
    if ($or$cond6) {
     $offset$0 = $40;
    } else {
     $delta$0 = 0;
     break;
    }
   } else {
    $44 = (64 - ($threshold$0))|0;
    $offset$0 = $44;
   }
   $45 = (($threshold$0) - ($27))|0;
   $46 = (($27) - ($offset$0))|0;
   $47 = (($threshold$0) - ($26))|0;
   $48 = (($26) - ($offset$0))|0;
   $49 = ($45|0)>($46|0);
   $50 = (0 - ($45))|0;
   $$7 = $49 ? $46 : $50;
   $51 = ($48|0)>($47|0);
   $52 = (0 - ($48))|0;
   $u_off2$0 = $51 ? $47 : $52;
   $53 = ($$7|0)<(0);
   $54 = (0 - ($$7))|0;
   $55 = $53 ? $54 : $$7;
   $56 = ($u_off2$0|0)<(0);
   $57 = (0 - ($u_off2$0))|0;
   $58 = $56 ? $57 : $u_off2$0;
   $59 = ($55|0)>($58|0);
   $u_off2$0$$7 = $59 ? $u_off2$0 : $$7;
   $delta$0 = $u_off2$0$$7;
  }
 } while(0);
 if ($3) {
  $60 = ($delta$0|0)>(14);
  if ($60) {
   $delta$1 = 14;
  } else {
   $61 = ($delta$0|0)<(-14);
   $$delta$0 = $61 ? -14 : $delta$0;
   $delta$1 = $$delta$0;
  }
 } else {
  $delta$1 = $delta$0;
 }
 $62 = (($delta$1) + ($23))|0;
 $63 = ($15|0)>($17|0);
 $64 = (($62) + ($19))|0;
 $$10 = $63 ? $62 : $64;
 $$11 = $63 ? $64 : $62;
 $65 = ((($edge)) + 8|0);
 HEAP32[$65>>2] = $$10;
 $66 = ((($edge2)) + 8|0);
 HEAP32[$66>>2] = $$11;
 return ($delta$1|0);
}
function _af_indic_metrics_init($metrics,$face) {
 $metrics = $metrics|0;
 $face = $face|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($face)) + 92|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 68|0);
 $3 = HEAP16[$2>>1]|0;
 $4 = $3&65535;
 $5 = ((($metrics)) + 40|0);
 HEAP32[$5>>2] = $4;
 $6 = (_FT_Select_Charmap($face,1970170211)|0);
 $7 = ($6|0)==(0);
 if ($7) {
  _af_cjk_metrics_init_widths($metrics,$face);
  _af_cjk_metrics_check_digits($metrics,$face);
  (_FT_Set_Charmap($face,$1)|0);
  return 0;
 } else {
  HEAP32[$0>>2] = 0;
  (_FT_Set_Charmap($face,$1)|0);
  return 0;
 }
 return (0)|0;
}
function _af_indic_metrics_scale($metrics,$scaler) {
 $metrics = $metrics|0;
 $scaler = $scaler|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($metrics)) + 4|0);
 ;HEAP32[$0>>2]=HEAP32[$scaler>>2]|0;HEAP32[$0+4>>2]=HEAP32[$scaler+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$scaler+8>>2]|0;HEAP32[$0+12>>2]=HEAP32[$scaler+12>>2]|0;HEAP32[$0+16>>2]=HEAP32[$scaler+16>>2]|0;HEAP32[$0+20>>2]=HEAP32[$scaler+20>>2]|0;HEAP32[$0+24>>2]=HEAP32[$scaler+24>>2]|0;
 _af_cjk_metrics_scale_dim($metrics,$scaler,0);
 _af_cjk_metrics_scale_dim($metrics,$scaler,1);
 return;
}
function _af_indic_hints_init($hints,$metrics) {
 $hints = $hints|0;
 $metrics = $metrics|0;
 var $$$i = 0, $$4$i = 0, $$4$v$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond3$i = 0, $other_flags$1$$i = 0, $other_flags$1$i = 0;
 var $other_flags$3$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($hints)) + 2988|0);
 HEAP32[$0>>2] = $metrics;
 $1 = ((($metrics)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($hints)) + 2980|0);
 $4 = ((($metrics)) + 44|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($hints)) + 4|0);
 HEAP32[$6>>2] = $5;
 $7 = ((($metrics)) + 48|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($hints)) + 8|0);
 HEAP32[$9>>2] = $8;
 $10 = ((($metrics)) + 1616|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($hints)) + 12|0);
 HEAP32[$12>>2] = $11;
 $13 = ((($metrics)) + 1620|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($hints)) + 16|0);
 HEAP32[$15>>2] = $14;
 $16 = ((($metrics)) + 24|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ($17|0)==(2);
 $19 = $17 | 1;
 $20 = ($19|0)==(3);
 $$$i = $20&1;
 $21 = ($17|0)==(4);
 $or$cond3$i = $18 | $21;
 $22 = $$$i | 2;
 $other_flags$1$i = $or$cond3$i ? $22 : $$$i;
 $23 = ($17|0)==(1);
 $24 = $other_flags$1$i | 4;
 $other_flags$1$$i = $23 ? $other_flags$1$i : $24;
 $25 = $other_flags$1$$i | 8;
 $other_flags$3$i = $18 ? $25 : $other_flags$1$$i;
 $26 = ((($metrics)) + 36|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ((($27)) + 212|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = ((($29)) + 20|0);
 $31 = HEAP8[$30>>0]|0;
 $32 = ($31<<24>>24)==(0);
 $$4$v$i = $32 ? 12 : 4;
 $$4$i = $$4$v$i | $2;
 HEAP32[$3>>2] = $$4$i;
 $33 = ((($hints)) + 2984|0);
 HEAP32[$33>>2] = $other_flags$3$i;
 return 0;
}
function _af_indic_hints_apply($hints,$outline,$metrics) {
 $hints = $hints|0;
 $outline = $outline|0;
 $metrics = $metrics|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_af_cjk_hints_apply($hints,$outline,$metrics)|0);
 return ($0|0);
}
function _af_autofitter_load_glyph($module,$slot,$size,$glyph_index,$load_flags) {
 $module = $module|0;
 $slot = $slot|0;
 $size = $size|0;
 $glyph_index = $glyph_index|0;
 $load_flags = $load_flags|0;
 var $$$i$i = 0, $$0$i = 0, $$ph$i = 0, $$pre$i$i = 0, $$pre$phi$i$iZ2D = 0, $$pre$phi2Z2D = 0, $$pre$phiZ2D = 0, $$pre1 = 0, $$pre9$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0;
 var $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0;
 var $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0;
 var $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0;
 var $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0;
 var $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0;
 var $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0;
 var $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $bbox$i$i = 0, $cond$i$i = 0, $error$0$i2$i = 0, $error$i$i = 0, $hints = 0, $inverse$i$i = 0, $loader = 0, $metrics$1$i$ph$i = 0, $metrics$1$i4$i = 0, $or$cond$i$i = 0, $or$cond3$i$i = 0, $pp2x_uh$0$i$i = 0, $scaler$i = 0, $vvector$i$i = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 7008|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $inverse$i$i = sp + 6992|0;
 $bbox$i$i = sp + 6976|0;
 $vvector$i$i = sp + 6968|0;
 $error$i$i = sp + 6964|0;
 $scaler$i = sp + 6936|0;
 $hints = sp + 64|0;
 $loader = sp;
 $0 = ((($module)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 _memset(($hints|0),0,3000)|0;
 HEAP32[$hints>>2] = $2;
 dest=$loader; stop=dest+60|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $3 = ((($loader)) + 8|0);
 HEAP32[$3>>2] = $hints;
 $4 = ((($slot)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 88|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 if ($8) {
  $$pre1 = ((($loader)) + 4|0);
  $$0$i = 36;$$pre$phi2Z2D = $$pre1;$$pre$phiZ2D = $loader;
  HEAP32[$$pre$phiZ2D>>2] = 0;
  HEAP32[$$pre$phi2Z2D>>2] = 0;
  HEAP32[$3>>2] = 0;
  _af_glyph_hints_done($hints);
  STACKTOP = sp;return ($$0$i|0);
 }
 ;HEAP32[$scaler$i>>2]=0|0;HEAP32[$scaler$i+4>>2]=0|0;HEAP32[$scaler$i+8>>2]=0|0;HEAP32[$scaler$i+12>>2]=0|0;HEAP32[$scaler$i+16>>2]=0|0;
 HEAP32[$scaler$i>>2] = $5;
 $9 = ((($7)) + 16|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($scaler$i)) + 4|0);
 HEAP32[$11>>2] = $10;
 $12 = ((($scaler$i)) + 12|0);
 HEAP32[$12>>2] = 0;
 $13 = ((($7)) + 20|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($scaler$i)) + 8|0);
 HEAP32[$15>>2] = $14;
 $16 = ((($scaler$i)) + 16|0);
 HEAP32[$16>>2] = 0;
 $17 = $load_flags >>> 16;
 $18 = $17 & 15;
 $19 = ((($scaler$i)) + 20|0);
 HEAP32[$19>>2] = $18;
 $20 = ((($scaler$i)) + 24|0);
 HEAP32[$20>>2] = 0;
 HEAP32[$loader>>2] = $5;
 $21 = ((($5)) + 116|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ((($loader)) + 4|0);
 HEAP32[$23>>2] = $22;
 $24 = ($22|0)==(0|0);
 do {
  if ($24) {
   $25 = (_af_face_globals_new($5,$23,$module)|0);
   $26 = ($25|0)==(0);
   if ($26) {
    $27 = HEAP32[$23>>2]|0;
    HEAP32[$21>>2] = $27;
    $28 = ((($5)) + 120|0);
    HEAP32[$28>>2] = 4439;
    $29 = $27;
    $31 = $29;
    break;
   } else {
    $$0$i = $25;$$pre$phi2Z2D = $23;$$pre$phiZ2D = $loader;
    HEAP32[$$pre$phiZ2D>>2] = 0;
    HEAP32[$$pre$phi2Z2D>>2] = 0;
    HEAP32[$3>>2] = 0;
    _af_glyph_hints_done($hints);
    STACKTOP = sp;return ($$0$i|0);
   }
  } else {
   $31 = $22;
  }
 } while(0);
 HEAP32[$error$i$i>>2] = 0;
 $30 = ((($31)) + 4|0);
 $32 = HEAP32[$30>>2]|0;
 $33 = ($32>>>0)>($glyph_index>>>0);
 L11: do {
  if ($33) {
   $34 = ((($31)) + 8|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = (($35) + ($glyph_index)|0);
   $37 = HEAP8[$36>>0]|0;
   $38 = $37&255;
   $39 = $38 & 127;
   $40 = (62352 + ($39<<2)|0);
   $41 = HEAP32[$40>>2]|0;
   $42 = ((($41)) + 4|0);
   $43 = HEAP32[$42>>2]|0;
   $44 = (62332 + ($43<<2)|0);
   $45 = HEAP32[$44>>2]|0;
   $46 = (((($31)) + 16|0) + ($39<<2)|0);
   $47 = HEAP32[$46>>2]|0;
   $48 = ($47|0)==(0|0);
   do {
    if ($48) {
     $49 = HEAP32[$31>>2]|0;
     $50 = ((($49)) + 100|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = ((($45)) + 4|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = (_ft_mem_alloc($51,$53,$error$i$i)|0);
     $55 = HEAP32[$error$i$i>>2]|0;
     $56 = ($55|0)==(0);
     if (!($56)) {
      $$ph$i = $55;
      break L11;
     }
     HEAP32[$54>>2] = $41;
     $57 = ((($54)) + 36|0);
     HEAP32[$57>>2] = $31;
     $58 = ((($45)) + 8|0);
     $59 = HEAP32[$58>>2]|0;
     $60 = ($59|0)==(0|0);
     if (!($60)) {
      $61 = HEAP32[$31>>2]|0;
      $62 = (FUNCTION_TABLE_iii[$59 & 8191]($54,$61)|0);
      HEAP32[$error$i$i>>2] = $62;
      $63 = ($62|0)==(0);
      if (!($63)) {
       $64 = ((($45)) + 16|0);
       $65 = HEAP32[$64>>2]|0;
       $66 = ($65|0)==(0|0);
       if (!($66)) {
        FUNCTION_TABLE_vi[$65 & 8191]($54);
       }
       _ft_mem_free($51,$54);
       $$pre$i$i = HEAP32[$error$i$i>>2]|0;
       $67 = ($$pre$i$i|0)==(0);
       if ($67) {
        $metrics$1$i4$i = 0;
        break;
       } else {
        $$0$i = $$pre$i$i;$$pre$phi2Z2D = $23;$$pre$phiZ2D = $loader;
       }
       HEAP32[$$pre$phiZ2D>>2] = 0;
       HEAP32[$$pre$phi2Z2D>>2] = 0;
       HEAP32[$3>>2] = 0;
       _af_glyph_hints_done($hints);
       STACKTOP = sp;return ($$0$i|0);
      }
     }
     HEAP32[$46>>2] = $54;
     $metrics$1$i$ph$i = $54;
     label = 15;
    } else {
     $metrics$1$i$ph$i = $47;
     label = 15;
    }
   } while(0);
   if ((label|0) == 15) {
    $metrics$1$i4$i = $metrics$1$i$ph$i;
   }
   $68 = HEAP32[$metrics$1$i4$i>>2]|0;
   $69 = ((($68)) + 4|0);
   $70 = HEAP32[$69>>2]|0;
   $71 = (62332 + ($70<<2)|0);
   $72 = HEAP32[$71>>2]|0;
   $73 = ((($loader)) + 12|0);
   HEAP32[$73>>2] = $metrics$1$i4$i;
   $74 = ((($72)) + 12|0);
   $75 = HEAP32[$74>>2]|0;
   $76 = ($75|0)==(0|0);
   if ($76) {
    $77 = ((($metrics$1$i4$i)) + 4|0);
    ;HEAP32[$77>>2]=HEAP32[$scaler$i>>2]|0;HEAP32[$77+4>>2]=HEAP32[$scaler$i+4>>2]|0;HEAP32[$77+8>>2]=HEAP32[$scaler$i+8>>2]|0;HEAP32[$77+12>>2]=HEAP32[$scaler$i+12>>2]|0;HEAP32[$77+16>>2]=HEAP32[$scaler$i+16>>2]|0;HEAP32[$77+20>>2]=HEAP32[$scaler$i+20>>2]|0;HEAP32[$77+24>>2]=HEAP32[$scaler$i+24>>2]|0;
   } else {
    FUNCTION_TABLE_vii[$75 & 8191]($metrics$1$i4$i,$scaler$i);
   }
   $78 = $load_flags & -10246;
   $79 = ((($72)) + 20|0);
   $80 = HEAP32[$79>>2]|0;
   $81 = ($80|0)==(0|0);
   if (!($81)) {
    $82 = HEAP32[$3>>2]|0;
    $83 = (FUNCTION_TABLE_iii[$80 & 8191]($82,$metrics$1$i4$i)|0);
    $84 = ($83|0)==(0);
    if (!($84)) {
     $$0$i = $83;$$pre$phi2Z2D = $23;$$pre$phiZ2D = $loader;
     HEAP32[$$pre$phiZ2D>>2] = 0;
     HEAP32[$$pre$phi2Z2D>>2] = 0;
     HEAP32[$3>>2] = 0;
     _af_glyph_hints_done($hints);
     STACKTOP = sp;return ($$0$i|0);
    }
   }
   $85 = HEAP32[$loader>>2]|0;
   $86 = HEAP32[$73>>2]|0;
   $87 = HEAP32[$3>>2]|0;
   $88 = ((($85)) + 84|0);
   $89 = HEAP32[$88>>2]|0;
   $90 = ((($89)) + 156|0);
   $91 = HEAP32[$90>>2]|0;
   $92 = HEAP32[$91>>2]|0;
   $93 = $78 | 10241;
   $94 = (_FT_Load_Glyph($85,$glyph_index,$93)|0);
   $95 = ($94|0)==(0);
   if (!($95)) {
    $$0$i = $94;$$pre$phi2Z2D = $23;$$pre$phiZ2D = $loader;
    HEAP32[$$pre$phiZ2D>>2] = 0;
    HEAP32[$$pre$phi2Z2D>>2] = 0;
    HEAP32[$3>>2] = 0;
    _af_glyph_hints_done($hints);
    STACKTOP = sp;return ($$0$i|0);
   }
   $96 = ((($91)) + 8|0);
   $97 = HEAP8[$96>>0]|0;
   $98 = ((($loader)) + 16|0);
   HEAP8[$98>>0] = $97;
   $99 = ($97<<24>>24)==(0);
   if (!($99)) {
    $100 = ((($loader)) + 20|0);
    $101 = ((($91)) + 12|0);
    ;HEAP32[$100>>2]=HEAP32[$101>>2]|0;HEAP32[$100+4>>2]=HEAP32[$101+4>>2]|0;HEAP32[$100+8>>2]=HEAP32[$101+8>>2]|0;HEAP32[$100+12>>2]=HEAP32[$101+12>>2]|0;
    $102 = ((($loader)) + 36|0);
    $103 = ((($91)) + 28|0);
    $104 = $103;
    $105 = $104;
    $106 = HEAP32[$105>>2]|0;
    $107 = (($104) + 4)|0;
    $108 = $107;
    $109 = HEAP32[$108>>2]|0;
    $110 = $102;
    $111 = $110;
    HEAP32[$111>>2] = $106;
    $112 = (($110) + 4)|0;
    $113 = $112;
    HEAP32[$113>>2] = $109;
    ;HEAP32[$inverse$i$i>>2]=HEAP32[$100>>2]|0;HEAP32[$inverse$i$i+4>>2]=HEAP32[$100+4>>2]|0;HEAP32[$inverse$i$i+8>>2]=HEAP32[$100+8>>2]|0;HEAP32[$inverse$i$i+12>>2]=HEAP32[$100+12>>2]|0;
    $114 = (_FT_Matrix_Invert($inverse$i$i)|0);
    $115 = ($114|0)==(0);
    if ($115) {
     _FT_Vector_Transform($102,$inverse$i$i);
    }
   }
   $116 = ((($89)) + 72|0);
   $117 = HEAP32[$116>>2]|0;
   $cond$i$i = ($117|0)==(1869968492);
   do {
    if ($cond$i$i) {
     $118 = HEAP8[$98>>0]|0;
     $119 = ($118<<24>>24)==(0);
     if (!($119)) {
      $120 = ((($89)) + 108|0);
      $121 = ((($loader)) + 36|0);
      $122 = HEAP32[$121>>2]|0;
      $123 = ((($loader)) + 40|0);
      $124 = HEAP32[$123>>2]|0;
      _FT_Outline_Translate($120,$122,$124);
     }
     $125 = ((($87)) + 8|0);
     $126 = HEAP32[$125>>2]|0;
     $127 = ((($loader)) + 44|0);
     HEAP32[$127>>2] = $126;
     $128 = ((($87)) + 16|0);
     $129 = HEAP32[$128>>2]|0;
     $130 = ((($loader)) + 48|0);
     HEAP32[$130>>2] = $129;
     $131 = ((($89)) + 40|0);
     $132 = HEAP32[$131>>2]|0;
     $133 = ((($87)) + 4|0);
     $134 = HEAP32[$133>>2]|0;
     $135 = (_FT_MulFix($132,$134)|0);
     $136 = HEAP32[$125>>2]|0;
     $137 = (($136) + ($135))|0;
     $138 = ((($loader)) + 52|0);
     HEAP32[$138>>2] = $137;
     $139 = HEAP32[$128>>2]|0;
     $140 = ((($loader)) + 56|0);
     HEAP32[$140>>2] = $139;
     $141 = ((($89)) + 110|0);
     $142 = HEAP16[$141>>1]|0;
     $143 = ($142<<16>>16)==(0);
     if ($143) {
      $error$0$i2$i = 0;
     } else {
      $144 = HEAP32[$86>>2]|0;
      $145 = ((($144)) + 4|0);
      $146 = HEAP32[$145>>2]|0;
      $147 = (62332 + ($146<<2)|0);
      $148 = HEAP32[$147>>2]|0;
      $149 = ((($148)) + 24|0);
      $150 = HEAP32[$149>>2]|0;
      $151 = ($150|0)==(0|0);
      if (!($151)) {
       $152 = ((($92)) + 20|0);
       (FUNCTION_TABLE_iiii[$150 & 8191]($87,$152,$86)|0);
      }
      $153 = HEAP32[$19>>2]|0;
      $154 = ($153|0)==(1);
      if ($154) {
       $208 = HEAP32[$127>>2]|0;
       $209 = HEAP32[$138>>2]|0;
       $210 = ((($87)) + 2992|0);
       $211 = HEAP32[$210>>2]|0;
       $212 = (($208) + 32)|0;
       $213 = (($212) + ($211))|0;
       $214 = $213 & -64;
       HEAP32[$127>>2] = $214;
       $215 = ((($87)) + 2996|0);
       $216 = HEAP32[$215>>2]|0;
       $217 = (($209) + 32)|0;
       $218 = (($217) + ($216))|0;
       $219 = $218 & -64;
       HEAP32[$138>>2] = $219;
       $220 = (($214) - ($208))|0;
       $221 = ((($89)) + 144|0);
       HEAP32[$221>>2] = $220;
       $222 = (($219) - ($209))|0;
       $223 = ((($89)) + 148|0);
       HEAP32[$223>>2] = $222;
       $error$0$i2$i = 0;
       break;
      }
      $155 = ((($87)) + 64|0);
      $156 = HEAP32[$155>>2]|0;
      $157 = ((($87)) + 56|0);
      $158 = HEAP32[$157>>2]|0;
      $159 = (($156) + (($158*48)|0)|0);
      $160 = ((($159)) + -48|0);
      $161 = ($158|0)>(1);
      if ($161) {
       $162 = ((($87)) + 2980|0);
       $163 = HEAP32[$162>>2]|0;
       $164 = $163 & 4;
       $165 = ($164|0)==(0);
       if ($165) {
        $166 = HEAP32[$138>>2]|0;
        $167 = ((($160)) + 4|0);
        $168 = HEAP32[$167>>2]|0;
        $169 = (($166) - ($168))|0;
        $170 = ((($156)) + 4|0);
        $171 = HEAP32[$170>>2]|0;
        $172 = ((($156)) + 8|0);
        $173 = HEAP32[$172>>2]|0;
        $174 = (($173) - ($171))|0;
        $175 = ((($160)) + 8|0);
        $176 = HEAP32[$175>>2]|0;
        $177 = (($176) + ($169))|0;
        $178 = ($171|0)<(24);
        $179 = (($174) + -8)|0;
        $$$i$i = $178 ? $179 : $174;
        $180 = ($169|0)<(24);
        $181 = (($177) + 8)|0;
        $pp2x_uh$0$i$i = $180 ? $181 : $177;
        $182 = (($$$i$i) + 32)|0;
        $183 = $182 & -64;
        HEAP32[$127>>2] = $183;
        $184 = (($pp2x_uh$0$i$i) + 32)|0;
        $185 = $184 & -64;
        HEAP32[$138>>2] = $185;
        $186 = ($183|0)>=($173|0);
        $187 = ($171|0)>(0);
        $or$cond$i$i = $187 & $186;
        if ($or$cond$i$i) {
         $188 = (($183) + -64)|0;
         HEAP32[$127>>2] = $188;
         $193 = $188;
        } else {
         $193 = $183;
        }
        $189 = ($185|0)<=($176|0);
        $190 = ($169|0)>(0);
        $or$cond3$i$i = $190 & $189;
        if ($or$cond3$i$i) {
         $191 = (($185) + 64)|0;
         HEAP32[$138>>2] = $191;
         $196 = $191;
        } else {
         $196 = $185;
        }
        $192 = (($193) - ($$$i$i))|0;
        $194 = ((($89)) + 144|0);
        HEAP32[$194>>2] = $192;
        $195 = (($196) - ($pp2x_uh$0$i$i))|0;
        $197 = ((($89)) + 148|0);
        HEAP32[$197>>2] = $195;
        $error$0$i2$i = 0;
        break;
       }
      }
      $198 = HEAP32[$127>>2]|0;
      $199 = HEAP32[$138>>2]|0;
      $200 = (($198) + 32)|0;
      $201 = $200 & -64;
      HEAP32[$127>>2] = $201;
      $202 = (($199) + 32)|0;
      $203 = $202 & -64;
      HEAP32[$138>>2] = $203;
      $204 = (($201) - ($198))|0;
      $205 = ((($89)) + 144|0);
      HEAP32[$205>>2] = $204;
      $206 = (($203) - ($199))|0;
      $207 = ((($89)) + 148|0);
      HEAP32[$207>>2] = $206;
      $error$0$i2$i = 0;
     }
    } else {
     $error$0$i2$i = 7;
    }
   } while(0);
   $224 = ((($89)) + 44|0);
   $225 = HEAP32[$224>>2]|0;
   $226 = ((($89)) + 32|0);
   $227 = HEAP32[$226>>2]|0;
   $228 = (($225) - ($227))|0;
   $229 = ((($89)) + 48|0);
   $230 = HEAP32[$229>>2]|0;
   $231 = ((($89)) + 36|0);
   $232 = HEAP32[$231>>2]|0;
   $233 = (($230) - ($232))|0;
   $234 = ((($vvector$i$i)) + 4|0);
   $235 = ((($86)) + 8|0);
   $236 = HEAP32[$235>>2]|0;
   $237 = (_FT_MulFix($228,$236)|0);
   HEAP32[$vvector$i$i>>2] = $237;
   $238 = ((($86)) + 12|0);
   $239 = HEAP32[$238>>2]|0;
   $240 = (_FT_MulFix($233,$239)|0);
   HEAP32[$234>>2] = $240;
   $241 = HEAP8[$98>>0]|0;
   $242 = ($241<<24>>24)==(0);
   $$pre9$i = ((($92)) + 20|0);
   if (!($242)) {
    $243 = ((($loader)) + 20|0);
    _FT_Outline_Transform($$pre9$i,$243);
    _FT_Vector_Transform($vvector$i$i,$243);
   }
   $244 = ((($loader)) + 44|0);
   $245 = HEAP32[$244>>2]|0;
   $246 = ($245|0)==(0);
   if (!($246)) {
    $247 = (0 - ($245))|0;
    _FT_Outline_Translate($$pre9$i,$247,0);
   }
   _FT_Outline_Get_CBox($$pre9$i,$bbox$i$i);
   $248 = HEAP32[$bbox$i$i>>2]|0;
   $249 = $248 & -64;
   HEAP32[$bbox$i$i>>2] = $249;
   $250 = ((($bbox$i$i)) + 4|0);
   $251 = HEAP32[$250>>2]|0;
   $252 = $251 & -64;
   HEAP32[$250>>2] = $252;
   $253 = ((($bbox$i$i)) + 8|0);
   $254 = HEAP32[$253>>2]|0;
   $255 = (($254) + 63)|0;
   $256 = $255 & -64;
   HEAP32[$253>>2] = $256;
   $257 = ((($bbox$i$i)) + 12|0);
   $258 = HEAP32[$257>>2]|0;
   $259 = (($258) + 63)|0;
   $260 = $259 & -64;
   HEAP32[$257>>2] = $260;
   $261 = (($256) - ($249))|0;
   $262 = ((($89)) + 24|0);
   HEAP32[$262>>2] = $261;
   $263 = (($260) - ($252))|0;
   $264 = ((($89)) + 28|0);
   HEAP32[$264>>2] = $263;
   HEAP32[$226>>2] = $249;
   HEAP32[$231>>2] = $260;
   $265 = HEAP32[$vvector$i$i>>2]|0;
   $266 = (($265) + ($249))|0;
   $267 = $266 & -64;
   HEAP32[$224>>2] = $267;
   $268 = HEAP32[$234>>2]|0;
   $269 = (($268) + ($260))|0;
   $270 = $269 & -64;
   HEAP32[$229>>2] = $270;
   $271 = HEAP32[$19>>2]|0;
   $272 = ($271|0)==(1);
   do {
    if ($272) {
     label = 54;
    } else {
     $273 = ((($89)) + 4|0);
     $274 = HEAP32[$273>>2]|0;
     $275 = ((($274)) + 8|0);
     $276 = HEAP32[$275>>2]|0;
     $277 = $276 & 4;
     $278 = ($277|0)==(0);
     if ($278) {
      $279 = HEAP32[$23>>2]|0;
      $280 = ((($279)) + 4|0);
      $281 = HEAP32[$280>>2]|0;
      $282 = ($281>>>0)>($glyph_index>>>0);
      if (!($282)) {
       label = 54;
       break;
      }
      $283 = ((($279)) + 8|0);
      $284 = HEAP32[$283>>2]|0;
      $285 = (($284) + ($glyph_index)|0);
      $286 = HEAP8[$285>>0]|0;
      $287 = ($286<<24>>24)>(-1);
      if ($287) {
       label = 54;
       break;
      }
      $288 = ((($86)) + 32|0);
      $289 = HEAP8[$288>>0]|0;
      $290 = ($289<<24>>24)==(0);
      if ($290) {
       label = 54;
       break;
      }
     }
     $291 = ((($89)) + 40|0);
     $292 = HEAP32[$291>>2]|0;
     $293 = HEAP32[$235>>2]|0;
     $294 = (_FT_MulFix($292,$293)|0);
     HEAP32[$291>>2] = $294;
     $295 = ((($89)) + 144|0);
     HEAP32[$295>>2] = 0;
     $296 = ((($89)) + 148|0);
     HEAP32[$296>>2] = 0;
     $$pre$phi$i$iZ2D = $291;
    }
   } while(0);
   if ((label|0) == 54) {
    $297 = ((($89)) + 40|0);
    $298 = HEAP32[$297>>2]|0;
    $299 = ($298|0)==(0);
    if ($299) {
     $$pre$phi$i$iZ2D = $297;
    } else {
     $300 = ((($loader)) + 52|0);
     $301 = HEAP32[$300>>2]|0;
     $302 = HEAP32[$244>>2]|0;
     $303 = (($301) - ($302))|0;
     HEAP32[$297>>2] = $303;
     $$pre$phi$i$iZ2D = $297;
    }
   }
   $304 = ((($89)) + 52|0);
   $305 = HEAP32[$304>>2]|0;
   $306 = HEAP32[$238>>2]|0;
   $307 = (_FT_MulFix($305,$306)|0);
   HEAP32[$304>>2] = $307;
   $308 = HEAP32[$$pre$phi$i$iZ2D>>2]|0;
   $309 = (($308) + 32)|0;
   $310 = $309 & -64;
   HEAP32[$$pre$phi$i$iZ2D>>2] = $310;
   $311 = (($307) + 32)|0;
   $312 = $311 & -64;
   HEAP32[$304>>2] = $312;
   HEAP32[$116>>2] = 1869968492;
   $$0$i = $error$0$i2$i;$$pre$phi2Z2D = $23;$$pre$phiZ2D = $loader;
   HEAP32[$$pre$phiZ2D>>2] = 0;
   HEAP32[$$pre$phi2Z2D>>2] = 0;
   HEAP32[$3>>2] = 0;
   _af_glyph_hints_done($hints);
   STACKTOP = sp;return ($$0$i|0);
  } else {
   HEAP32[$error$i$i>>2] = 6;
   $$ph$i = 6;
  }
 } while(0);
 $$0$i = $$ph$i;$$pre$phi2Z2D = $23;$$pre$phiZ2D = $loader;
 HEAP32[$$pre$phiZ2D>>2] = 0;
 HEAP32[$$pre$phi2Z2D>>2] = 0;
 HEAP32[$3>>2] = 0;
 _af_glyph_hints_done($hints);
 STACKTOP = sp;return ($$0$i|0);
}
function _af_face_globals_new($face,$aglobals,$module) {
 $face = $face|0;
 $aglobals = $aglobals|0;
 $module = $module|0;
 var $$pre = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $error = 0, $exitcond$i = 0;
 var $gindex$i = 0, $i$08$i = 0, $nn$07$i = 0, $range$011$i = 0, $ss$015$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $gindex$i = sp + 4|0;
 $error = sp;
 $0 = ((($face)) + 100|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($face)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + 216)|0;
 $5 = (_ft_mem_alloc($1,$4,$error)|0);
 $6 = HEAP32[$error>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $94 = $6;
  HEAP32[$aglobals>>2] = $5;
  STACKTOP = sp;return ($94|0);
 }
 HEAP32[$5>>2] = $face;
 $8 = HEAP32[$2>>2]|0;
 $9 = ((($5)) + 4|0);
 HEAP32[$9>>2] = $8;
 $10 = ((($5)) + 216|0);
 $11 = ((($5)) + 8|0);
 HEAP32[$11>>2] = $10;
 $12 = ((($5)) + 212|0);
 HEAP32[$12>>2] = $module;
 $13 = ((($face)) + 92|0);
 $14 = HEAP32[$13>>2]|0;
 _memset(($10|0),127,($8|0))|0;
 $15 = (_FT_Select_Charmap($face,1970170211)|0);
 $16 = ($15|0)==(0);
 if ($16) {
  $18 = 60892;$ss$015$i = 0;
  while(1) {
   $17 = ((($18)) + 8|0);
   $19 = HEAP32[$17>>2]|0;
   $20 = (61928 + ($19<<2)|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = ((($21)) + 4|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = ($23|0)==(0|0);
   if (!($24)) {
    $25 = ((($18)) + 16|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ($26|0)==(10);
    if ($27) {
     $28 = HEAP32[$23>>2]|0;
     $29 = ($28|0)==(0);
     if (!($29)) {
      $30 = $ss$015$i&255;
      $31 = $28;$range$011$i = $23;
      while(1) {
       $32 = (_FT_Get_Char_Index($face,$31)|0);
       HEAP32[$gindex$i>>2] = $32;
       $33 = ($32|0)==(0);
       if (!($33)) {
        $34 = HEAP32[$9>>2]|0;
        $35 = ($32>>>0)<($34>>>0);
        if ($35) {
         $36 = (($10) + ($32)|0);
         $37 = HEAP8[$36>>0]|0;
         $38 = ($37<<24>>24)==(127);
         if ($38) {
          HEAP8[$36>>0] = $30;
         }
        }
       }
       $39 = ((($range$011$i)) + 4|0);
       $40 = (_FT_Get_Next_Char($face,$31,$gindex$i)|0);
       $41 = HEAP32[$gindex$i>>2]|0;
       $42 = ($41|0)==(0);
       L19: do {
        if (!($42)) {
         $45 = $40;$48 = $41;
         while(1) {
          $43 = HEAP32[$39>>2]|0;
          $44 = ($45>>>0)>($43>>>0);
          if ($44) {
           break L19;
          }
          $46 = HEAP32[$9>>2]|0;
          $47 = ($48>>>0)<($46>>>0);
          if ($47) {
           $49 = (($10) + ($48)|0);
           $50 = HEAP8[$49>>0]|0;
           $51 = ($50<<24>>24)==(127);
           if ($51) {
            HEAP8[$49>>0] = $30;
           }
          }
          $52 = (_FT_Get_Next_Char($face,$45,$gindex$i)|0);
          $53 = HEAP32[$gindex$i>>2]|0;
          $54 = ($53|0)==(0);
          if ($54) {
           break;
          } else {
           $45 = $52;$48 = $53;
          }
         }
        }
       } while(0);
       $55 = ((($range$011$i)) + 8|0);
       $56 = HEAP32[$55>>2]|0;
       $57 = ($56|0)==(0);
       if ($57) {
        break;
       } else {
        $31 = $56;$range$011$i = $55;
       }
      }
     }
    }
   }
   $58 = (($ss$015$i) + 1)|0;
   $59 = (62352 + ($58<<2)|0);
   $60 = HEAP32[$59>>2]|0;
   $61 = ($58|0)==(49);
   if ($61) {
    $i$08$i = 48;
    break;
   } else {
    $18 = $60;$ss$015$i = $58;
   }
  }
  while(1) {
   $62 = (_FT_Get_Char_Index($face,$i$08$i)|0);
   $63 = ($62|0)==(0);
   if (!($63)) {
    $64 = HEAP32[$9>>2]|0;
    $65 = ($62>>>0)<($64>>>0);
    if ($65) {
     $66 = (($10) + ($62)|0);
     $67 = HEAP8[$66>>0]|0;
     $68 = $67&255;
     $69 = $68 | 128;
     $70 = $69&255;
     HEAP8[$66>>0] = $70;
    }
   }
   $71 = (($i$08$i) + 1)|0;
   $exitcond$i = ($71|0)==(58);
   if ($exitcond$i) {
    break;
   } else {
    $i$08$i = $71;
   }
  }
 }
 $72 = HEAP32[$12>>2]|0;
 $73 = ((($72)) + 12|0);
 $74 = HEAP32[$73>>2]|0;
 $75 = ($74|0)==(127);
 if (!($75)) {
  $76 = HEAP32[$9>>2]|0;
  $77 = ($76|0)>(0);
  if ($77) {
   $95 = $76;$nn$07$i = 0;
   while(1) {
    $78 = (($10) + ($nn$07$i)|0);
    $79 = HEAP8[$78>>0]|0;
    $80 = $79&255;
    $81 = $80 & 127;
    $82 = ($81|0)==(127);
    if ($82) {
     $83 = $80 & 128;
     $84 = $83&255;
     HEAP8[$78>>0] = $84;
     $85 = HEAP32[$12>>2]|0;
     $86 = ((($85)) + 12|0);
     $87 = HEAP32[$86>>2]|0;
     $88 = $87 | $83;
     $89 = $88&255;
     HEAP8[$78>>0] = $89;
     $$pre$i = HEAP32[$9>>2]|0;
     $92 = $$pre$i;
    } else {
     $92 = $95;
    }
    $90 = (($nn$07$i) + 1)|0;
    $91 = ($90|0)<($92|0);
    if ($91) {
     $95 = $92;$nn$07$i = $90;
    } else {
     break;
    }
   }
  }
 }
 (_FT_Set_Charmap($face,$14)|0);
 HEAP32[$error>>2] = 0;
 $93 = ((($5)) + 12|0);
 HEAP32[$93>>2] = 0;
 $$pre = HEAP32[$error>>2]|0;
 $94 = $$pre;
 HEAP32[$aglobals>>2] = $5;
 STACKTOP = sp;return ($94|0);
}
function _af_face_globals_free($globals) {
 $globals = $globals|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $exitcond = 0, $nn$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($globals|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = HEAP32[$globals>>2]|0;
 $2 = ((($1)) + 100|0);
 $3 = HEAP32[$2>>2]|0;
 $nn$01 = 0;
 while(1) {
  $4 = (((($globals)) + 16|0) + ($nn$01<<2)|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ($5|0)==(0|0);
  if (!($6)) {
   $7 = (62352 + ($nn$01<<2)|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ((($8)) + 4|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = (62332 + ($10<<2)|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ((($12)) + 16|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ($14|0)==(0|0);
   if ($15) {
    $16 = $5;
   } else {
    FUNCTION_TABLE_vi[$14 & 8191]($5);
    $$pre = HEAP32[$4>>2]|0;
    $16 = $$pre;
   }
   _ft_mem_free($3,$16);
   HEAP32[$4>>2] = 0;
  }
  $17 = (($nn$01) + 1)|0;
  $exitcond = ($17|0)==(49);
  if ($exitcond) {
   break;
  } else {
   $nn$01 = $17;
  }
 }
 $18 = ((($globals)) + 4|0);
 HEAP32[$18>>2] = 0;
 $19 = ((($globals)) + 8|0);
 HEAP32[$19>>2] = 0;
 HEAP32[$globals>>2] = 0;
 _ft_mem_free($3,$globals);
 return;
}
function _af_autofitter_init($ft_module) {
 $ft_module = $ft_module|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($ft_module)) + 12|0);
 HEAP32[$0>>2] = 48;
 $1 = ((($ft_module)) + 16|0);
 HEAP32[$1>>2] = 5;
 $2 = ((($ft_module)) + 20|0);
 HEAP8[$2>>0] = 0;
 return 0;
}
function _af_autofitter_done($ft_module) {
 $ft_module = $ft_module|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _af_get_interface($module,$module_interface) {
 $module = $module|0;
 $module_interface = $module_interface|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_ft_service_list_lookup(62604,$module_interface)|0);
 return ($0|0);
}
function _af_property_set($ft_module,$property_name,$value) {
 $ft_module = $ft_module|0;
 $property_name = $property_name|0;
 $value = $value|0;
 var $$0$i = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $cond = 0, $globals$i = 0, $globals$sroa$0$0$ph = 0, $ss$06 = 0, $ss$06$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $globals$i = sp;
 $0 = (_strcmp($property_name,595524)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  $2 = HEAP32[$value>>2]|0;
  $4 = 60892;$ss$06 = 0;
  while(1) {
   $3 = ((($4)) + 8|0);
   $5 = HEAP32[$3>>2]|0;
   $6 = ($5|0)==($2|0);
   if ($6) {
    $7 = ((($4)) + 16|0);
    $8 = HEAP32[$7>>2]|0;
    $9 = ($8|0)==(10);
    if ($9) {
     $ss$06$lcssa = $ss$06;
     label = 5;
     break;
    }
   }
   $11 = (($ss$06) + 1)|0;
   $12 = (62352 + ($11<<2)|0);
   $13 = HEAP32[$12>>2]|0;
   $cond = ($11|0)==(49);
   if ($cond) {
    $$1 = 6;
    label = 19;
    break;
   } else {
    $4 = $13;$ss$06 = $11;
   }
  }
  if ((label|0) == 5) {
   $10 = ((($ft_module)) + 12|0);
   HEAP32[$10>>2] = $ss$06$lcssa;
   STACKTOP = sp;return 0;
  }
  else if ((label|0) == 19) {
   STACKTOP = sp;return ($$1|0);
  }
 }
 $14 = (_strcmp($property_name,595540)|0);
 $15 = ($14|0)==(0);
 if ($15) {
  $16 = HEAP32[$value>>2]|0;
  $17 = ((($ft_module)) + 16|0);
  HEAP32[$17>>2] = $16;
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $18 = (_strcmp($property_name,595555)|0);
 $19 = ($18|0)==(0);
 if (!($19)) {
  $34 = (_strcmp($property_name,595573)|0);
  $35 = ($34|0)==(0);
  if (!($35)) {
   $$1 = 12;
   STACKTOP = sp;return ($$1|0);
  }
  $36 = HEAP8[$value>>0]|0;
  $37 = ((($ft_module)) + 20|0);
  HEAP8[$37>>0] = $36;
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $20 = HEAP32[$value>>2]|0;
 $21 = ($20|0)==(0|0);
 do {
  if ($21) {
   $$0$i = 35;
  } else {
   $22 = ((($20)) + 116|0);
   $23 = HEAP32[$22>>2]|0;
   HEAP32[$globals$i>>2] = $23;
   $24 = ($23|0)==(0|0);
   if ($24) {
    $26 = (_af_face_globals_new($20,$globals$i,$ft_module)|0);
    $27 = ($26|0)==(0);
    if (!($27)) {
     $$0$i = $26;
     break;
    }
    $28 = HEAP32[$globals$i>>2]|0;
    HEAP32[$22>>2] = $28;
    $29 = ((($20)) + 120|0);
    HEAP32[$29>>2] = 4439;
    $globals$sroa$0$0$ph = $28;
   } else {
    $25 = $23;
    $globals$sroa$0$0$ph = $25;
   }
   $30 = ((($value)) + 4|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = $globals$sroa$0$0$ph;
   $33 = ((($32)) + 12|0);
   HEAP32[$33>>2] = $31;
   $$1 = 0;
   STACKTOP = sp;return ($$1|0);
  }
 } while(0);
 $$1 = $$0$i;
 STACKTOP = sp;return ($$1|0);
}
function _af_property_get($ft_module,$property_name,$value) {
 $ft_module = $ft_module|0;
 $property_name = $property_name|0;
 $value = $value|0;
 var $$0 = 0, $$0$i = 0, $$0$i3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $globals$i = 0, $globals$i1 = 0, $globals$sroa$0$0$ph = 0, $globals3$sroa$0$0$ph = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $globals$i1 = sp + 4|0;
 $globals$i = sp;
 $0 = ((($ft_module)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($ft_module)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($ft_module)) + 20|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = (_strcmp($property_name,595581)|0);
 $7 = ($6|0)==(0);
 if ($7) {
  $8 = HEAP32[$value>>2]|0;
  $9 = ($8|0)==(0|0);
  do {
   if ($9) {
    $$0$i = 35;
   } else {
    $10 = ((($8)) + 116|0);
    $11 = HEAP32[$10>>2]|0;
    HEAP32[$globals$i>>2] = $11;
    $12 = ($11|0)==(0|0);
    if ($12) {
     $14 = (_af_face_globals_new($8,$globals$i,$ft_module)|0);
     $15 = ($14|0)==(0);
     if (!($15)) {
      $$0$i = $14;
      break;
     }
     $16 = HEAP32[$globals$i>>2]|0;
     HEAP32[$10>>2] = $16;
     $17 = ((($8)) + 120|0);
     HEAP32[$17>>2] = 4439;
     $globals$sroa$0$0$ph = $16;
    } else {
     $13 = $11;
     $globals$sroa$0$0$ph = $13;
    }
    $18 = $globals$sroa$0$0$ph;
    $19 = ((($18)) + 8|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = ((($value)) + 4|0);
    HEAP32[$21>>2] = $20;
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
  } while(0);
  $$0 = $$0$i;
  STACKTOP = sp;return ($$0|0);
 }
 $22 = (_strcmp($property_name,595524)|0);
 $23 = ($22|0)==(0);
 if ($23) {
  $24 = (62352 + ($1<<2)|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = ((($25)) + 8|0);
  $27 = HEAP32[$26>>2]|0;
  HEAP32[$value>>2] = $27;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $28 = (_strcmp($property_name,595540)|0);
 $29 = ($28|0)==(0);
 if ($29) {
  HEAP32[$value>>2] = $3;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $30 = (_strcmp($property_name,595555)|0);
 $31 = ($30|0)==(0);
 if (!($31)) {
  $46 = (_strcmp($property_name,595573)|0);
  $47 = ($46|0)==(0);
  if (!($47)) {
   $$0 = 12;
   STACKTOP = sp;return ($$0|0);
  }
  HEAP8[$value>>0] = $5;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $32 = HEAP32[$value>>2]|0;
 $33 = ($32|0)==(0|0);
 do {
  if ($33) {
   $$0$i3 = 35;
  } else {
   $34 = ((($32)) + 116|0);
   $35 = HEAP32[$34>>2]|0;
   HEAP32[$globals$i1>>2] = $35;
   $36 = ($35|0)==(0|0);
   if ($36) {
    $38 = (_af_face_globals_new($32,$globals$i1,$ft_module)|0);
    $39 = ($38|0)==(0);
    if (!($39)) {
     $$0$i3 = $38;
     break;
    }
    $40 = HEAP32[$globals$i1>>2]|0;
    HEAP32[$34>>2] = $40;
    $41 = ((($32)) + 120|0);
    HEAP32[$41>>2] = 4439;
    $globals3$sroa$0$0$ph = $40;
   } else {
    $37 = $35;
    $globals3$sroa$0$0$ph = $37;
   }
   $42 = $globals3$sroa$0$0$ph;
   $43 = ((($42)) + 12|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = ((($value)) + 4|0);
   HEAP32[$45>>2] = $44;
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 $$0 = $$0$i3;
 STACKTOP = sp;return ($$0|0);
}
function _FT_Get_Advance($face,$gindex,$flags,$padvance) {
 $face = $face|0;
 $gindex = $gindex|0;
 $flags = $flags|0;
 $padvance = $padvance|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $scale$0$i = 0, $scale$0$in$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($face|0)==(0|0);
 if ($0) {
  $$1 = 35;
  return ($$1|0);
 }
 $1 = ($padvance|0)==(0|0);
 if ($1) {
  $$1 = 6;
  return ($$1|0);
 }
 $2 = ((($face)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3>>>0)>($gindex>>>0);
 if (!($4)) {
  $$1 = 16;
  return ($$1|0);
 }
 $5 = ((($face)) + 96|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($6)) + 12|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($8)) + 84|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10|0)==(0|0);
 do {
  if (!($11)) {
   $12 = $flags & 3;
   $13 = ($12|0)!=(0);
   $14 = $flags & 983040;
   $15 = ($14|0)==(65536);
   $or$cond = $13 | $15;
   if ($or$cond) {
    $16 = (FUNCTION_TABLE_iiiiii[$10 & 8191]($face,$gindex,1,$flags,$padvance)|0);
    $17 = ($16|0)==(0);
    if (!($17)) {
     $29 = $16 & 255;
     $30 = ($29|0)==(7);
     if ($30) {
      break;
     } else {
      $$1 = $16;
     }
     return ($$1|0);
    }
    $18 = $flags & 1;
    $19 = ($18|0)==(0);
    if (!($19)) {
     $$1 = 0;
     return ($$1|0);
    }
    $20 = ((($face)) + 88|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = ($21|0)==(0|0);
    if ($22) {
     $$1 = 36;
     return ($$1|0);
    }
    $23 = $flags & 16;
    $24 = ($23|0)==(0);
    $25 = ((($21)) + 20|0);
    $26 = ((($21)) + 16|0);
    $scale$0$in$i = $24 ? $26 : $25;
    $scale$0$i = HEAP32[$scale$0$in$i>>2]|0;
    $27 = HEAP32[$padvance>>2]|0;
    $28 = (_FT_MulDiv($27,$scale$0$i,64)|0);
    HEAP32[$padvance>>2] = $28;
    $$1 = 0;
    return ($$1|0);
   }
  }
 } while(0);
 $31 = (_FT_Get_Advances($face,$gindex,1,$flags,$padvance)|0);
 $$1 = $31;
 return ($$1|0);
}
function _FT_Get_Advances($face,$start,$count,$flags,$padvances) {
 $face = $face|0;
 $start = $start|0;
 $count = $count|0;
 $flags = $flags|0;
 $padvances = $padvances|0;
 var $$0 = 0, $$sink = 0, $$sink$us = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i = 0, $nn$01$i = 0, $nn$05 = 0, $nn$05$us = 0, $or$cond = 0, $or$cond1 = 0, $or$cond3 = 0, $scale$0$i = 0, $scale$0$in$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($face|0)==(0|0);
 if ($0) {
  $$0 = 35;
  return ($$0|0);
 }
 $1 = ($padvances|0)==(0|0);
 if ($1) {
  $$0 = 6;
  return ($$0|0);
 }
 $2 = ((($face)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($count) + ($start))|0;
 $5 = ($3>>>0)<=($start>>>0);
 $6 = ($4>>>0)<($start>>>0);
 $or$cond = $6 | $5;
 $7 = ($4>>>0)>($3>>>0);
 $or$cond1 = $7 | $or$cond;
 if ($or$cond1) {
  $$0 = 16;
  return ($$0|0);
 }
 $8 = ($count|0)==(0);
 if ($8) {
  $$0 = 0;
  return ($$0|0);
 }
 $9 = ((($face)) + 96|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + 12|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($12)) + 84|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ($14|0)==(0|0);
 do {
  if (!($15)) {
   $16 = $flags & 3;
   $17 = ($16|0)!=(0);
   $18 = $flags & 983040;
   $19 = ($18|0)==(65536);
   $or$cond3 = $17 | $19;
   if ($or$cond3) {
    $20 = (FUNCTION_TABLE_iiiiii[$14 & 8191]($face,$start,$count,$flags,$padvances)|0);
    $21 = ($20|0)==(0);
    if (!($21)) {
     $35 = $20 & 255;
     $36 = ($35|0)==(7);
     if ($36) {
      break;
     } else {
      $$0 = $20;
     }
     return ($$0|0);
    }
    $22 = $flags & 1;
    $23 = ($22|0)==(0);
    if (!($23)) {
     $$0 = 0;
     return ($$0|0);
    }
    $24 = ((($face)) + 88|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = ($25|0)==(0|0);
    if ($26) {
     $$0 = 36;
     return ($$0|0);
    }
    $27 = $flags & 16;
    $28 = ($27|0)==(0);
    $29 = ((($25)) + 20|0);
    $30 = ((($25)) + 16|0);
    $scale$0$in$i = $28 ? $30 : $29;
    $scale$0$i = HEAP32[$scale$0$in$i>>2]|0;
    $nn$01$i = 0;
    while(1) {
     $31 = (($padvances) + ($nn$01$i<<2)|0);
     $32 = HEAP32[$31>>2]|0;
     $33 = (_FT_MulDiv($32,$scale$0$i,64)|0);
     HEAP32[$31>>2] = $33;
     $34 = (($nn$01$i) + 1)|0;
     $exitcond$i = ($34|0)==($count|0);
     if ($exitcond$i) {
      $$0 = 0;
      break;
     } else {
      $nn$01$i = $34;
     }
    }
    return ($$0|0);
   }
  }
 } while(0);
 $37 = $flags & 536870912;
 $38 = ($37|0)==(0);
 if (!($38)) {
  $$0 = 7;
  return ($$0|0);
 }
 $39 = $flags | 256;
 $40 = $flags & 16;
 $41 = ($40|0)==(0);
 $42 = ((($face)) + 84|0);
 if ($41) {
  $nn$05$us = 0;
  while(1) {
   $43 = (($nn$05$us) + ($start))|0;
   $44 = (_FT_Load_Glyph($face,$43,$39)|0);
   $45 = ($44|0)==(0);
   if (!($45)) {
    $$0 = $44;
    label = 19;
    break;
   }
   $46 = HEAP32[$42>>2]|0;
   $47 = ((($46)) + 64|0);
   $$sink$us = HEAP32[$47>>2]|0;
   $48 = $$sink$us << 10;
   $49 = (($padvances) + ($nn$05$us<<2)|0);
   HEAP32[$49>>2] = $48;
   $50 = (($nn$05$us) + 1)|0;
   $51 = ($50>>>0)<($count>>>0);
   if ($51) {
    $nn$05$us = $50;
   } else {
    $$0 = 0;
    label = 19;
    break;
   }
  }
  if ((label|0) == 19) {
   return ($$0|0);
  }
 } else {
  $nn$05 = 0;
  while(1) {
   $52 = (($nn$05) + ($start))|0;
   $53 = (_FT_Load_Glyph($face,$52,$39)|0);
   $54 = ($53|0)==(0);
   if (!($54)) {
    $$0 = $53;
    label = 19;
    break;
   }
   $55 = HEAP32[$42>>2]|0;
   $56 = ((($55)) + 68|0);
   $$sink = HEAP32[$56>>2]|0;
   $57 = $$sink << 10;
   $58 = (($padvances) + ($nn$05<<2)|0);
   HEAP32[$58>>2] = $57;
   $59 = (($nn$05) + 1)|0;
   $60 = ($59>>>0)<($count>>>0);
   if ($60) {
    $nn$05 = $59;
   } else {
    $$0 = 0;
    label = 19;
    break;
   }
  }
  if ((label|0) == 19) {
   return ($$0|0);
  }
 }
 return (0)|0;
}
function _FT_Bitmap_Init($abitmap) {
 $abitmap = $abitmap|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($abitmap|0)==(0|0);
 if ($0) {
  return;
 }
 ;HEAP32[$abitmap>>2]=0|0;HEAP32[$abitmap+4>>2]=0|0;HEAP32[$abitmap+8>>2]=0|0;HEAP32[$abitmap+12>>2]=0|0;HEAP32[$abitmap+16>>2]=0|0;HEAP32[$abitmap+20>>2]=0|0;
 return;
}
function _FT_Bitmap_Copy($library,$source,$target) {
 $library = $library|0;
 $source = $source|0;
 $target = $target|0;
 var $$ = 0, $$0 = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $i$03 = 0, $or$cond = 0, $s$04 = 0, $t$05 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 HEAP32[$error>>2] = 0;
 $0 = ($library|0)==(0|0);
 if ($0) {
  $$0 = 33;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ($source|0)!=(0|0);
 $2 = ($target|0)!=(0|0);
 $or$cond = $1 & $2;
 if (!($or$cond)) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ($source|0)==($target|0);
 if ($3) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ((($source)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $5 >> 31;
 $7 = $6 | 1;
 $8 = ((($target)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 >> 31;
 $11 = $10 | 1;
 $12 = ((($source)) + 12|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13|0)==(0|0);
 if ($14) {
  ;HEAP32[$target>>2]=HEAP32[$source>>2]|0;HEAP32[$target+4>>2]=HEAP32[$source+4>>2]|0;HEAP32[$target+8>>2]=HEAP32[$source+8>>2]|0;HEAP32[$target+12>>2]=HEAP32[$source+12>>2]|0;HEAP32[$target+16>>2]=HEAP32[$source+16>>2]|0;HEAP32[$target+20>>2]=HEAP32[$source+20>>2]|0;
  $15 = ($7|0)==($11|0);
  if ($15) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
  $16 = HEAP32[$8>>2]|0;
  $17 = (0 - ($16))|0;
  HEAP32[$8>>2] = $17;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $18 = ($5|0)<(0);
 $19 = HEAP32[$library>>2]|0;
 $20 = (0 - ($5))|0;
 $$ = $18 ? $20 : $5;
 $21 = HEAP32[$source>>2]|0;
 $22 = Math_imul($21, $$)|0;
 $23 = ((($target)) + 12|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ($24|0)==(0|0);
 if ($25) {
  $32 = (_ft_mem_qalloc($19,$22,$error)|0);
  HEAP32[$23>>2] = $32;
  $35 = $32;
 } else {
  $26 = ($9|0)<(0);
  $27 = (0 - ($9))|0;
  $$2 = $26 ? $27 : $9;
  $28 = HEAP32[$target>>2]|0;
  $29 = Math_imul($28, $$2)|0;
  $30 = ($29|0)==($22|0);
  if ($30) {
   $35 = $24;
  } else {
   $31 = (_ft_mem_qrealloc($19,1,$29,$22,$24,$error)|0);
   HEAP32[$23>>2] = $31;
   $35 = $31;
  }
 }
 $33 = HEAP32[$error>>2]|0;
 $34 = ($33|0)==(0);
 do {
  if ($34) {
   ;HEAP32[$target>>2]=HEAP32[$source>>2]|0;HEAP32[$target+4>>2]=HEAP32[$source+4>>2]|0;HEAP32[$target+8>>2]=HEAP32[$source+8>>2]|0;HEAP32[$target+12>>2]=HEAP32[$source+12>>2]|0;HEAP32[$target+16>>2]=HEAP32[$source+16>>2]|0;HEAP32[$target+20>>2]=HEAP32[$source+20>>2]|0;
   HEAP32[$23>>2] = $35;
   $36 = ($7|0)==($11|0);
   $37 = HEAP32[$12>>2]|0;
   if ($36) {
    _memcpy(($35|0),($37|0),($22|0))|0;
    break;
   }
   $38 = HEAP32[$target>>2]|0;
   $39 = ($38|0)==(0);
   if (!($39)) {
    $40 = (($38) + -1)|0;
    $41 = Math_imul($40, $$)|0;
    $42 = (($35) + ($41)|0);
    $43 = (0 - ($$))|0;
    $i$03 = $38;$s$04 = $37;$t$05 = $42;
    while(1) {
     _memcpy(($t$05|0),($s$04|0),($$|0))|0;
     $44 = (($s$04) + ($$)|0);
     $45 = (($t$05) + ($43)|0);
     $46 = (($i$03) + -1)|0;
     $47 = ($46|0)==(0);
     if ($47) {
      break;
     } else {
      $i$03 = $46;$s$04 = $44;$t$05 = $45;
     }
    }
   }
  }
 } while(0);
 $48 = HEAP32[$error>>2]|0;
 $$0 = $48;
 STACKTOP = sp;return ($$0|0);
}
function _FT_Bitmap_Convert($library,$source,$target,$alignment) {
 $library = $library|0;
 $source = $source|0;
 $target = $target|0;
 $alignment = $alignment|0;
 var $$ = 0, $$0$i = 0, $$1 = 0, $$3 = 0, $$off = 0, $$pre87 = 0, $$pre88 = 0, $$pre89 = 0, $$pre90 = 0, $$pre91 = 0, $$pre92 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0;
 var $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0;
 var $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $error = 0, $i$019 = 0;
 var $i14$068 = 0, $i2$025 = 0, $i3$043 = 0, $i9$057 = 0, $j$012 = 0, $j$115 = 0, $j12$049 = 0, $j17$062 = 0, $j6$031 = 0, $j6$138 = 0, $or$cond = 0, $pad$0 = 0, $s$0 = 0, $s$121 = 0, $s$226 = 0, $s$344 = 0, $s$456 = 0, $s$567 = 0, $scevgep = 0, $scevgep80 = 0;
 var $scevgep81 = 0, $scevgep82 = 0, $scevgep83 = 0, $scevgep84 = 0, $ss$0$lcssa = 0, $ss$010 = 0, $ss10$0$lcssa = 0, $ss10$047 = 0, $ss15$060 = 0, $ss4$0$lcssa = 0, $ss4$029 = 0, $switch = 0, $t$0 = 0, $t$120 = 0, $t$224 = 0, $t$342 = 0, $t$455 = 0, $t$566 = 0, $tt$0$lcssa = 0, $tt$011 = 0;
 var $tt$114 = 0, $tt11$0$lcssa = 0, $tt11$048 = 0, $tt16$061 = 0, $tt5$0$lcssa = 0, $tt5$030 = 0, $tt5$137 = 0, $val1$016 = 0, $val8$039 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 HEAP32[$error>>2] = 0;
 $0 = ($library|0)==(0|0);
 if ($0) {
  $$1 = 33;
  STACKTOP = sp;return ($$1|0);
 }
 $1 = ($source|0)!=(0|0);
 $2 = ($target|0)!=(0|0);
 $or$cond = $1 & $2;
 if (!($or$cond)) {
  $$1 = 6;
  STACKTOP = sp;return ($$1|0);
 }
 $3 = HEAP32[$library>>2]|0;
 $4 = ((($source)) + 18|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $$off = (($6) + -1)|0;
 $switch = ($$off>>>0)<(7);
 if ($switch) {
  $7 = ((($target)) + 8|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8|0)<(0);
  $10 = (0 - ($8))|0;
  $$ = $9 ? $10 : $8;
  $11 = HEAP32[$target>>2]|0;
  $12 = Math_imul($$, $11)|0;
  $13 = ((($target)) + 18|0);
  HEAP8[$13>>0] = 2;
  $14 = HEAP32[$source>>2]|0;
  HEAP32[$target>>2] = $14;
  $15 = ((($source)) + 4|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ((($target)) + 4|0);
  HEAP32[$17>>2] = $16;
  $18 = ($alignment|0)>(0);
  $19 = HEAP32[$15>>2]|0;
  if ($18) {
   $20 = (($19|0) % ($alignment|0))&-1;
   $21 = ($20|0)==(0);
   $22 = (($alignment) - ($20))|0;
   $$3 = $21 ? 0 : $22;
   $pad$0 = $$3;
  } else {
   $pad$0 = 0;
  }
  $23 = (($19) + ($pad$0))|0;
  $24 = ($23|0)>(0);
  if ($24) {
   $25 = (4294967295 / ($23>>>0))&-1;
   $26 = ($14>>>0)>($25>>>0);
   if ($26) {
    $$1 = 6;
    STACKTOP = sp;return ($$1|0);
   }
  }
  $27 = Math_imul($14, $23)|0;
  $28 = ($27>>>0)>($12>>>0);
  do {
   if ($28) {
    $29 = ((($target)) + 12|0);
    $30 = HEAP32[$29>>2]|0;
    $31 = (_ft_mem_qrealloc($3,1,$12,$27,$30,$error)|0);
    HEAP32[$29>>2] = $31;
    $32 = HEAP32[$error>>2]|0;
    $33 = ($32|0)==(0);
    if ($33) {
     $$pre87 = HEAP32[$7>>2]|0;
     $35 = $$pre87;
     break;
    } else {
     $$1 = $32;
     STACKTOP = sp;return ($$1|0);
    }
   } else {
    $35 = $8;
   }
  } while(0);
  $34 = ($35|0)<(0);
  $36 = (0 - ($23))|0;
  $37 = $34 ? $36 : $23;
  HEAP32[$7>>2] = $37;
 } else {
  HEAP32[$error>>2] = 6;
 }
 $38 = ((($source)) + 12|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ((($target)) + 12|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = ((($source)) + 8|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = ($43|0)<(0);
 if ($44) {
  $45 = HEAP32[$source>>2]|0;
  $46 = (($45) + -1)|0;
  $47 = (0 - ($43))|0;
  $48 = Math_imul($46, $47)|0;
  $49 = (($39) + ($48)|0);
  $s$0 = $49;
 } else {
  $s$0 = $39;
 }
 $50 = ((($target)) + 8|0);
 $51 = HEAP32[$50>>2]|0;
 $52 = ($51|0)<(0);
 if ($52) {
  $53 = HEAP32[$target>>2]|0;
  $54 = (($53) + -1)|0;
  $55 = (0 - ($51))|0;
  $56 = Math_imul($54, $55)|0;
  $57 = (($41) + ($56)|0);
  $t$0 = $57;
 } else {
  $t$0 = $41;
 }
 $58 = HEAP8[$4>>0]|0;
 $59 = $58&255;
 switch ($59|0) {
 case 1:  {
  $60 = ((($target)) + 16|0);
  HEAP16[$60>>1] = 2;
  $61 = HEAP32[$source>>2]|0;
  $62 = ($61|0)==(0);
  if (!($62)) {
   $63 = ((($source)) + 4|0);
   $i$019 = $61;$s$121 = $s$0;$t$120 = $t$0;
   while(1) {
    $64 = HEAP32[$63>>2]|0;
    $65 = $64 >>> 3;
    $66 = ($65|0)==(0);
    if ($66) {
     $103 = $64;$ss$0$lcssa = $s$121;$tt$0$lcssa = $t$120;
    } else {
     $67 = $65 << 3;
     $j$012 = $65;$ss$010 = $s$121;$tt$011 = $t$120;
     while(1) {
      $68 = HEAP8[$ss$010>>0]|0;
      $69 = $68&255;
      $70 = ($68&255) >>> 7;
      HEAP8[$tt$011>>0] = $70;
      $71 = $69 >>> 6;
      $72 = $71 & 1;
      $73 = $72&255;
      $74 = ((($tt$011)) + 1|0);
      HEAP8[$74>>0] = $73;
      $75 = $69 >>> 5;
      $76 = $75 & 1;
      $77 = $76&255;
      $78 = ((($tt$011)) + 2|0);
      HEAP8[$78>>0] = $77;
      $79 = $69 >>> 4;
      $80 = $79 & 1;
      $81 = $80&255;
      $82 = ((($tt$011)) + 3|0);
      HEAP8[$82>>0] = $81;
      $83 = $69 >>> 3;
      $84 = $83 & 1;
      $85 = $84&255;
      $86 = ((($tt$011)) + 4|0);
      HEAP8[$86>>0] = $85;
      $87 = $69 >>> 2;
      $88 = $87 & 1;
      $89 = $88&255;
      $90 = ((($tt$011)) + 5|0);
      HEAP8[$90>>0] = $89;
      $91 = $69 >>> 1;
      $92 = $91 & 1;
      $93 = $92&255;
      $94 = ((($tt$011)) + 6|0);
      HEAP8[$94>>0] = $93;
      $95 = $69 & 1;
      $96 = $95&255;
      $97 = ((($tt$011)) + 7|0);
      HEAP8[$97>>0] = $96;
      $98 = ((($tt$011)) + 8|0);
      $99 = ((($ss$010)) + 1|0);
      $100 = (($j$012) + -1)|0;
      $101 = ($100|0)==(0);
      if ($101) {
       break;
      } else {
       $j$012 = $100;$ss$010 = $99;$tt$011 = $98;
      }
     }
     $scevgep = (($s$121) + ($65)|0);
     $scevgep80 = (($t$120) + ($67)|0);
     $$pre92 = HEAP32[$63>>2]|0;
     $103 = $$pre92;$ss$0$lcssa = $scevgep;$tt$0$lcssa = $scevgep80;
    }
    $102 = $103 & 7;
    $104 = ($102|0)==(0);
    if (!($104)) {
     $105 = HEAP8[$ss$0$lcssa>>0]|0;
     $106 = $105&255;
     $j$115 = $102;$tt$114 = $tt$0$lcssa;$val1$016 = $106;
     while(1) {
      $107 = $val1$016 >>> 7;
      $108 = $107 & 1;
      $109 = $108&255;
      HEAP8[$tt$114>>0] = $109;
      $110 = $val1$016 << 1;
      $111 = ((($tt$114)) + 1|0);
      $112 = (($j$115) + -1)|0;
      $113 = ($112|0)==(0);
      if ($113) {
       break;
      } else {
       $j$115 = $112;$tt$114 = $111;$val1$016 = $110;
      }
     }
    }
    $114 = HEAP32[$42>>2]|0;
    $115 = (($s$121) + ($114)|0);
    $116 = HEAP32[$50>>2]|0;
    $117 = (($t$120) + ($116)|0);
    $118 = (($i$019) + -1)|0;
    $119 = ($118|0)==(0);
    if ($119) {
     break;
    } else {
     $i$019 = $118;$s$121 = $115;$t$120 = $117;
    }
   }
  }
  break;
 }
 case 6: case 5: case 2:  {
  $120 = ((($source)) + 4|0);
  $121 = HEAP32[$120>>2]|0;
  $122 = ((($target)) + 16|0);
  HEAP16[$122>>1] = 256;
  $123 = HEAP32[$source>>2]|0;
  $124 = ($123|0)==(0);
  if (!($124)) {
   $i2$025 = $123;$s$226 = $s$0;$t$224 = $t$0;
   while(1) {
    _memcpy(($t$224|0),($s$226|0),($121|0))|0;
    $125 = HEAP32[$42>>2]|0;
    $126 = (($s$226) + ($125)|0);
    $127 = HEAP32[$50>>2]|0;
    $128 = (($t$224) + ($127)|0);
    $129 = (($i2$025) + -1)|0;
    $130 = ($129|0)==(0);
    if ($130) {
     break;
    } else {
     $i2$025 = $129;$s$226 = $126;$t$224 = $128;
    }
   }
  }
  break;
 }
 case 3:  {
  $131 = ((($target)) + 16|0);
  HEAP16[$131>>1] = 4;
  $132 = HEAP32[$source>>2]|0;
  $133 = ($132|0)==(0);
  if (!($133)) {
   $134 = ((($source)) + 4|0);
   $i3$043 = $132;$s$344 = $s$0;$t$342 = $t$0;
   while(1) {
    $135 = HEAP32[$134>>2]|0;
    $136 = $135 >>> 2;
    $137 = ($136|0)==(0);
    if ($137) {
     $158 = $135;$ss4$0$lcssa = $s$344;$tt5$0$lcssa = $t$342;
    } else {
     $138 = $136 << 2;
     $scevgep81 = (($t$342) + ($138)|0);
     $j6$031 = $136;$ss4$029 = $s$344;$tt5$030 = $t$342;
     while(1) {
      $139 = HEAP8[$ss4$029>>0]|0;
      $140 = $139&255;
      $141 = ($139&255) >>> 6;
      HEAP8[$tt5$030>>0] = $141;
      $142 = $140 >>> 4;
      $143 = $142 & 3;
      $144 = $143&255;
      $145 = ((($tt5$030)) + 1|0);
      HEAP8[$145>>0] = $144;
      $146 = $140 >>> 2;
      $147 = $146 & 3;
      $148 = $147&255;
      $149 = ((($tt5$030)) + 2|0);
      HEAP8[$149>>0] = $148;
      $150 = $140 & 3;
      $151 = $150&255;
      $152 = ((($tt5$030)) + 3|0);
      HEAP8[$152>>0] = $151;
      $153 = ((($ss4$029)) + 1|0);
      $154 = ((($tt5$030)) + 4|0);
      $155 = (($j6$031) + -1)|0;
      $156 = ($155|0)==(0);
      if ($156) {
       break;
      } else {
       $j6$031 = $155;$ss4$029 = $153;$tt5$030 = $154;
      }
     }
     $scevgep82 = (($s$344) + ($136)|0);
     $$pre91 = HEAP32[$134>>2]|0;
     $158 = $$pre91;$ss4$0$lcssa = $scevgep82;$tt5$0$lcssa = $scevgep81;
    }
    $157 = $158 & 3;
    $159 = ($157|0)==(0);
    if (!($159)) {
     $160 = HEAP8[$ss4$0$lcssa>>0]|0;
     $161 = $160&255;
     $j6$138 = $157;$tt5$137 = $tt5$0$lcssa;$val8$039 = $161;
     while(1) {
      $162 = $val8$039 >>> 6;
      $163 = $162 & 3;
      $164 = $163&255;
      HEAP8[$tt5$137>>0] = $164;
      $165 = $val8$039 << 2;
      $166 = ((($tt5$137)) + 1|0);
      $167 = (($j6$138) + -1)|0;
      $168 = ($167|0)==(0);
      if ($168) {
       break;
      } else {
       $j6$138 = $167;$tt5$137 = $166;$val8$039 = $165;
      }
     }
    }
    $169 = HEAP32[$42>>2]|0;
    $170 = (($s$344) + ($169)|0);
    $171 = HEAP32[$50>>2]|0;
    $172 = (($t$342) + ($171)|0);
    $173 = (($i3$043) + -1)|0;
    $174 = ($173|0)==(0);
    if ($174) {
     break;
    } else {
     $i3$043 = $173;$s$344 = $170;$t$342 = $172;
    }
   }
  }
  break;
 }
 case 4:  {
  $175 = ((($target)) + 16|0);
  HEAP16[$175>>1] = 16;
  $176 = HEAP32[$source>>2]|0;
  $177 = ($176|0)==(0);
  if (!($177)) {
   $178 = ((($source)) + 4|0);
   $i9$057 = $176;$s$456 = $s$0;$t$455 = $t$0;
   while(1) {
    $179 = HEAP32[$178>>2]|0;
    $180 = $179 >>> 1;
    $181 = ($180|0)==(0);
    if ($181) {
     $194 = $179;$ss10$0$lcssa = $s$456;$tt11$0$lcssa = $t$455;
    } else {
     $182 = $180 << 1;
     $scevgep83 = (($t$455) + ($182)|0);
     $j12$049 = $180;$ss10$047 = $s$456;$tt11$048 = $t$455;
     while(1) {
      $183 = HEAP8[$ss10$047>>0]|0;
      $184 = $183&255;
      $185 = ($183&255) >>> 4;
      HEAP8[$tt11$048>>0] = $185;
      $186 = $184 & 15;
      $187 = $186&255;
      $188 = ((($tt11$048)) + 1|0);
      HEAP8[$188>>0] = $187;
      $189 = ((($ss10$047)) + 1|0);
      $190 = ((($tt11$048)) + 2|0);
      $191 = (($j12$049) + -1)|0;
      $192 = ($191|0)==(0);
      if ($192) {
       break;
      } else {
       $j12$049 = $191;$ss10$047 = $189;$tt11$048 = $190;
      }
     }
     $scevgep84 = (($s$456) + ($180)|0);
     $$pre90 = HEAP32[$178>>2]|0;
     $194 = $$pre90;$ss10$0$lcssa = $scevgep84;$tt11$0$lcssa = $scevgep83;
    }
    $193 = $194 & 1;
    $195 = ($193|0)==(0);
    if (!($195)) {
     $196 = HEAP8[$ss10$0$lcssa>>0]|0;
     $197 = ($196&255) >>> 4;
     HEAP8[$tt11$0$lcssa>>0] = $197;
    }
    $198 = HEAP32[$42>>2]|0;
    $199 = (($s$456) + ($198)|0);
    $200 = HEAP32[$50>>2]|0;
    $201 = (($t$455) + ($200)|0);
    $202 = (($i9$057) + -1)|0;
    $203 = ($202|0)==(0);
    if ($203) {
     break;
    } else {
     $i9$057 = $202;$s$456 = $199;$t$455 = $201;
    }
   }
  }
  break;
 }
 case 7:  {
  $204 = ((($target)) + 16|0);
  HEAP16[$204>>1] = 256;
  $205 = HEAP32[$source>>2]|0;
  $206 = ($205|0)==(0);
  if (!($206)) {
   $207 = ((($source)) + 4|0);
   $245 = $51;$246 = $43;$i14$068 = $205;$s$567 = $s$0;$t$566 = $t$0;
   while(1) {
    $208 = HEAP32[$207>>2]|0;
    $209 = ($208|0)==(0);
    if ($209) {
     $239 = $246;$241 = $245;
    } else {
     $j17$062 = $208;$ss15$060 = $s$567;$tt16$061 = $t$566;
     while(1) {
      $210 = ((($ss15$060)) + 3|0);
      $211 = HEAP8[$210>>0]|0;
      $212 = $211&255;
      $213 = ($211<<24>>24)==(0);
      if ($213) {
       $$0$i = 0;
      } else {
       $214 = HEAP8[$ss15$060>>0]|0;
       $215 = $214&255;
       $216 = ($215*4732)|0;
       $217 = Math_imul($216, $215)|0;
       $218 = ((($ss15$060)) + 1|0);
       $219 = HEAP8[$218>>0]|0;
       $220 = $219&255;
       $221 = ($220*46871)|0;
       $222 = Math_imul($221, $220)|0;
       $223 = (($222) + ($217))|0;
       $224 = ((($ss15$060)) + 2|0);
       $225 = HEAP8[$224>>0]|0;
       $226 = $225&255;
       $227 = ($226*13933)|0;
       $228 = Math_imul($227, $226)|0;
       $229 = (($223) + ($228))|0;
       $230 = $229 >>> 16;
       $231 = (($230>>>0) / ($212>>>0))&-1;
       $232 = (($212) - ($231))|0;
       $233 = $232&255;
       $$0$i = $233;
      }
      HEAP8[$tt16$061>>0] = $$0$i;
      $234 = ((($ss15$060)) + 4|0);
      $235 = ((($tt16$061)) + 1|0);
      $236 = (($j17$062) + -1)|0;
      $237 = ($236|0)==(0);
      if ($237) {
       break;
      } else {
       $j17$062 = $236;$ss15$060 = $234;$tt16$061 = $235;
      }
     }
     $$pre88 = HEAP32[$42>>2]|0;
     $$pre89 = HEAP32[$50>>2]|0;
     $239 = $$pre88;$241 = $$pre89;
    }
    $238 = (($s$567) + ($239)|0);
    $240 = (($t$566) + ($241)|0);
    $242 = (($i14$068) + -1)|0;
    $243 = ($242|0)==(0);
    if ($243) {
     break;
    } else {
     $245 = $241;$246 = $239;$i14$068 = $242;$s$567 = $238;$t$566 = $240;
    }
   }
  }
  break;
 }
 default: {
 }
 }
 $244 = HEAP32[$error>>2]|0;
 $$1 = $244;
 STACKTOP = sp;return ($$1|0);
}
function _FT_Bitmap_Done($library,$bitmap) {
 $library = $library|0;
 $bitmap = $bitmap|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($library|0)==(0|0);
 if ($0) {
  $$0 = 33;
  return ($$0|0);
 }
 $1 = ($bitmap|0)==(0|0);
 if ($1) {
  $$0 = 6;
  return ($$0|0);
 }
 $2 = HEAP32[$library>>2]|0;
 $3 = ((($bitmap)) + 12|0);
 $4 = HEAP32[$3>>2]|0;
 _ft_mem_free($2,$4);
 ;HEAP32[$bitmap>>2]=0|0;HEAP32[$bitmap+4>>2]=0|0;HEAP32[$bitmap+8>>2]=0|0;HEAP32[$bitmap+12>>2]=0|0;HEAP32[$bitmap+16>>2]=0|0;HEAP32[$bitmap+20>>2]=0|0;
 $$0 = 0;
 return ($$0|0);
}
function _FT_RoundFix($a) {
 $a = $a|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($a|0)>(-1);
 if ($0) {
  $1 = (($a) + 32768)|0;
  $2 = $1 & -65536;
  $6 = $2;
  return ($6|0);
 } else {
  $3 = (32768 - ($a))|0;
  $4 = $3 & -65536;
  $5 = (0 - ($4))|0;
  $6 = $5;
  return ($6|0);
 }
 return (0)|0;
}
function _FT_Hypot($x,$y) {
 $x = $x|0;
 $y = $y|0;
 var $0 = 0, $1 = 0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $v = sp;
 HEAP32[$v>>2] = $x;
 $0 = ((($v)) + 4|0);
 HEAP32[$0>>2] = $y;
 $1 = (_FT_Vector_Length($v)|0);
 STACKTOP = sp;return ($1|0);
}
function _FT_MulDiv($a_,$b_,$c_) {
 $a_ = $a_|0;
 $b_ = $b_|0;
 $c_ = $c_|0;
 var $$0 = 0, $$01$i = 0, $$02 = 0, $$03 = 0, $$a_ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $a$0 = 0, $i$0$i = 0, $or$cond = 0;
 var $q$0$i = 0, $q$1$i = 0, $r$0$i = 0, $r$1$i = 0, $s$1 = 0, $s$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($a_|0)==(0);
 $1 = ($b_|0)==($c_|0);
 $or$cond = $0 | $1;
 if ($or$cond) {
  $$0 = $a_;
  return ($$0|0);
 }
 $2 = ($a_|0)<(0);
 $3 = (0 - ($a_))|0;
 $4 = $a_ >> 31;
 $5 = $4 | 1;
 $$a_ = $2 ? $3 : $a_;
 $6 = ($b_|0)<(0);
 $7 = (0 - ($b_))|0;
 $8 = (0 - ($5))|0;
 $$03 = $6 ? $7 : $b_;
 $s$1 = $6 ? $8 : $5;
 $9 = ($c_|0)<(0);
 $10 = (0 - ($c_))|0;
 $11 = (0 - ($s$1))|0;
 $$02 = $9 ? $10 : $c_;
 $s$2 = $9 ? $11 : $s$1;
 $12 = ($$02|0)==(0);
 do {
  if ($12) {
   $a$0 = 2147483647;
  } else {
   $13 = (($$03) + ($$a_))|0;
   $14 = $$02 >>> 17;
   $15 = (129894 - ($14))|0;
   $16 = ($13>>>0)>($15>>>0);
   if (!($16)) {
    $17 = Math_imul($$03, $$a_)|0;
    $18 = $$02 >>> 1;
    $19 = (($18) + ($17))|0;
    $20 = (($19>>>0) / ($$02>>>0))&-1;
    $a$0 = $20;
    break;
   }
   $21 = $$a_ & 65535;
   $22 = $$a_ >>> 16;
   $23 = $$03 & 65535;
   $24 = $$03 >>> 16;
   $25 = Math_imul($23, $21)|0;
   $26 = Math_imul($24, $21)|0;
   $27 = Math_imul($23, $22)|0;
   $28 = Math_imul($24, $22)|0;
   $29 = (($26) + ($27))|0;
   $30 = ($29>>>0)<($27>>>0);
   $31 = $30&1;
   $32 = $31 << 16;
   $33 = $29 >>> 16;
   $34 = $29 << 16;
   $35 = (($34) + ($25))|0;
   $36 = ($35>>>0)<($34>>>0);
   $37 = $36&1;
   $38 = (($33) + ($28))|0;
   $39 = (($38) + ($32))|0;
   $40 = (($39) + ($37))|0;
   $41 = $$02 >>> 1;
   $42 = (($35) + ($41))|0;
   $43 = ($42>>>0)<($35>>>0);
   $44 = $43&1;
   $45 = (($40) + ($44))|0;
   $46 = ($45|0)==(0);
   if ($46) {
    $47 = (($42>>>0) / ($$02>>>0))&-1;
    $a$0 = $47;
    break;
   }
   $48 = ($45>>>0)<($$02>>>0);
   if ($48) {
    $49 = (Math_clz32(($45|0))|0);
    $50 = $45 << $49;
    $51 = (32 - ($49))|0;
    $52 = $42 >>> $51;
    $53 = $52 | $50;
    $54 = $42 << $49;
    $55 = (($53>>>0) / ($$02>>>0))&-1;
    $56 = Math_imul($55, $$02)|0;
    $57 = (($53) - ($56))|0;
    $$01$i = $54;$i$0$i = $51;$q$0$i = $55;$r$0$i = $57;
    while(1) {
     $58 = $q$0$i << 1;
     $59 = $r$0$i << 1;
     $60 = $$01$i >>> 31;
     $61 = $59 | $60;
     $62 = $$01$i << 1;
     $63 = ($61>>>0)<($$02>>>0);
     $64 = $63 ? 0 : $$02;
     $r$1$i = (($61) - ($64))|0;
     $65 = $63&1;
     $66 = $65 | $58;
     $q$1$i = $66 ^ 1;
     $67 = (($i$0$i) + -1)|0;
     $68 = ($67|0)==(0);
     if ($68) {
      $a$0 = $q$1$i;
      break;
     } else {
      $$01$i = $62;$i$0$i = $67;$q$0$i = $q$1$i;$r$0$i = $r$1$i;
     }
    }
   } else {
    $a$0 = 2147483647;
   }
  }
 } while(0);
 $69 = ($s$2|0)<(0);
 $70 = (0 - ($a$0))|0;
 $71 = $69 ? $70 : $a$0;
 $$0 = $71;
 return ($$0|0);
}
function _FT_MulDiv_No_Round($a_,$b_,$c_) {
 $a_ = $a_|0;
 $b_ = $b_|0;
 $c_ = $c_|0;
 var $$0 = 0, $$01$i = 0, $$02 = 0, $$03 = 0, $$a_ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $7 = 0, $8 = 0, $9 = 0, $a$0 = 0, $i$0$i = 0, $or$cond = 0, $q$0$i = 0, $q$1$i = 0, $r$0$i = 0, $r$1$i = 0, $s$1 = 0, $s$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($a_|0)==(0);
 $1 = ($b_|0)==($c_|0);
 $or$cond = $0 | $1;
 if ($or$cond) {
  $$0 = $a_;
  return ($$0|0);
 }
 $2 = ($a_|0)<(0);
 $3 = (0 - ($a_))|0;
 $4 = $a_ >> 31;
 $5 = $4 | 1;
 $$a_ = $2 ? $3 : $a_;
 $6 = ($b_|0)<(0);
 $7 = (0 - ($b_))|0;
 $8 = (0 - ($5))|0;
 $$03 = $6 ? $7 : $b_;
 $s$1 = $6 ? $8 : $5;
 $9 = ($c_|0)<(0);
 $10 = (0 - ($c_))|0;
 $11 = (0 - ($s$1))|0;
 $$02 = $9 ? $10 : $c_;
 $s$2 = $9 ? $11 : $s$1;
 $12 = ($$02|0)==(0);
 do {
  if ($12) {
   $a$0 = 2147483647;
  } else {
   $13 = (($$03) + ($$a_))|0;
   $14 = ($13>>>0)<(131072);
   if ($14) {
    $15 = Math_imul($$03, $$a_)|0;
    $16 = (($15>>>0) / ($$02>>>0))&-1;
    $a$0 = $16;
    break;
   }
   $17 = $$a_ & 65535;
   $18 = $$a_ >>> 16;
   $19 = $$03 & 65535;
   $20 = $$03 >>> 16;
   $21 = Math_imul($19, $17)|0;
   $22 = Math_imul($20, $17)|0;
   $23 = Math_imul($19, $18)|0;
   $24 = Math_imul($20, $18)|0;
   $25 = (($22) + ($23))|0;
   $26 = ($25>>>0)<($23>>>0);
   $27 = $26&1;
   $28 = $27 << 16;
   $29 = $25 >>> 16;
   $30 = $25 << 16;
   $31 = (($30) + ($21))|0;
   $32 = ($31>>>0)<($30>>>0);
   $33 = $32&1;
   $34 = (($29) + ($24))|0;
   $35 = (($34) + ($28))|0;
   $36 = (($35) + ($33))|0;
   $37 = ($36|0)==(0);
   if ($37) {
    $38 = (($31>>>0) / ($$02>>>0))&-1;
    $a$0 = $38;
    break;
   }
   $39 = ($36>>>0)<($$02>>>0);
   if ($39) {
    $40 = (Math_clz32(($36|0))|0);
    $41 = $36 << $40;
    $42 = (32 - ($40))|0;
    $43 = $31 >>> $42;
    $44 = $43 | $41;
    $45 = $31 << $40;
    $46 = (($44>>>0) / ($$02>>>0))&-1;
    $47 = Math_imul($46, $$02)|0;
    $48 = (($44) - ($47))|0;
    $$01$i = $45;$i$0$i = $42;$q$0$i = $46;$r$0$i = $48;
    while(1) {
     $49 = $q$0$i << 1;
     $50 = $r$0$i << 1;
     $51 = $$01$i >>> 31;
     $52 = $50 | $51;
     $53 = $$01$i << 1;
     $54 = ($52>>>0)<($$02>>>0);
     $55 = $54 ? 0 : $$02;
     $r$1$i = (($52) - ($55))|0;
     $56 = $54&1;
     $57 = $56 | $49;
     $q$1$i = $57 ^ 1;
     $58 = (($i$0$i) + -1)|0;
     $59 = ($58|0)==(0);
     if ($59) {
      $a$0 = $q$1$i;
      break;
     } else {
      $$01$i = $53;$i$0$i = $58;$q$0$i = $q$1$i;$r$0$i = $r$1$i;
     }
    }
   } else {
    $a$0 = 2147483647;
   }
  }
 } while(0);
 $60 = ($s$2|0)<(0);
 $61 = (0 - ($a$0))|0;
 $62 = $60 ? $61 : $a$0;
 $$0 = $62;
 return ($$0|0);
}
function _FT_MulFix($a_,$b_) {
 $a_ = $a_|0;
 $b_ = $b_|0;
 var $$0 = 0, $$03 = 0, $$a_ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $a$0 = 0, $or$cond = 0, $s$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($a_|0)==(0);
 $1 = ($b_|0)==(65536);
 $or$cond = $0 | $1;
 if ($or$cond) {
  $$0 = $a_;
  return ($$0|0);
 }
 $2 = ($a_|0)<(0);
 $3 = (0 - ($a_))|0;
 $4 = $a_ >> 31;
 $5 = $4 | 1;
 $$a_ = $2 ? $3 : $a_;
 $6 = ($b_|0)<(0);
 $7 = (0 - ($b_))|0;
 $8 = (0 - ($5))|0;
 $$03 = $6 ? $7 : $b_;
 $s$1 = $6 ? $8 : $4;
 $9 = $$03 >>> 8;
 $10 = (($9) + ($$a_))|0;
 $11 = ($10>>>0)<(8191);
 if ($11) {
  $12 = Math_imul($$03, $$a_)|0;
  $13 = (($12) + 32768)|0;
  $14 = $13 >>> 16;
  $a$0 = $14;
 } else {
  $15 = $$a_ & 65535;
  $16 = $$a_ >>> 16;
  $17 = Math_imul($16, $$03)|0;
  $18 = $$03 >>> 16;
  $19 = Math_imul($18, $15)|0;
  $20 = (($19) + ($17))|0;
  $21 = $$03 & 65535;
  $22 = Math_imul($21, $15)|0;
  $23 = (($22) + 32768)|0;
  $24 = $23 >>> 16;
  $25 = (($20) + ($24))|0;
  $a$0 = $25;
 }
 $26 = ($s$1|0)<(0);
 $27 = (0 - ($a$0))|0;
 $28 = $26 ? $27 : $a$0;
 $$0 = $28;
 return ($$0|0);
}
function _FT_DivFix($a_,$b_) {
 $a_ = $a_|0;
 $b_ = $b_|0;
 var $$01 = 0, $$01$i = 0, $$a_ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0$i = 0, $q$0 = 0, $q$0$i = 0, $q$1$i = 0, $r$0$i = 0, $r$1$i = 0, $s$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($a_|0)<(0);
 $1 = (0 - ($a_))|0;
 $2 = $a_ >> 31;
 $3 = $2 | 1;
 $$a_ = $0 ? $1 : $a_;
 $4 = ($b_|0)<(0);
 $5 = (0 - ($b_))|0;
 $6 = (0 - ($3))|0;
 $s$1 = $4 ? $6 : $2;
 $$01 = $4 ? $5 : $b_;
 $7 = ($$01|0)==(0);
 do {
  if ($7) {
   $q$0 = 2147483647;
  } else {
   $8 = $$01 >>> 17;
   $9 = $8 ^ 65535;
   $10 = ($$a_>>>0)>($9>>>0);
   if (!($10)) {
    $11 = $$a_ << 16;
    $12 = $$01 >>> 1;
    $13 = (($12) + ($11))|0;
    $14 = (($13>>>0) / ($$01>>>0))&-1;
    $q$0 = $14;
    break;
   }
   $15 = $$a_ >>> 16;
   $16 = $$a_ << 16;
   $17 = $$01 >>> 1;
   $18 = (($17) + ($16))|0;
   $19 = ($18>>>0)<($16>>>0);
   $20 = $19&1;
   $21 = (($20) + ($15))|0;
   $22 = ($21>>>0)<($$01>>>0);
   if ($22) {
    $23 = (Math_clz32(($21|0))|0);
    $24 = $21 << $23;
    $25 = (32 - ($23))|0;
    $26 = $18 >>> $25;
    $27 = $26 | $24;
    $28 = $18 << $23;
    $29 = (($27>>>0) / ($$01>>>0))&-1;
    $30 = Math_imul($29, $$01)|0;
    $31 = (($27) - ($30))|0;
    $$01$i = $28;$i$0$i = $25;$q$0$i = $29;$r$0$i = $31;
    while(1) {
     $32 = $q$0$i << 1;
     $33 = $r$0$i << 1;
     $34 = $$01$i >>> 31;
     $35 = $33 | $34;
     $36 = $$01$i << 1;
     $37 = ($35>>>0)<($$01>>>0);
     $38 = $37 ? 0 : $$01;
     $r$1$i = (($35) - ($38))|0;
     $39 = $37&1;
     $40 = $39 | $32;
     $q$1$i = $40 ^ 1;
     $41 = (($i$0$i) + -1)|0;
     $42 = ($41|0)==(0);
     if ($42) {
      $q$0 = $q$1$i;
      break;
     } else {
      $$01$i = $36;$i$0$i = $41;$q$0$i = $q$1$i;$r$0$i = $r$1$i;
     }
    }
   } else {
    $q$0 = 2147483647;
   }
  }
 } while(0);
 $43 = ($s$1|0)<(0);
 $44 = (0 - ($q$0))|0;
 $45 = $43 ? $44 : $q$0;
 return ($45|0);
}
function _FT_Matrix_Invert($matrix) {
 $matrix = $matrix|0;
 var $$0 = 0, $$0$i = 0, $$0$i41 = 0, $$01$i$i = 0, $$01$i$i17 = 0, $$01$i$i28 = 0, $$01$i$i6 = 0, $$01$i27 = 0, $$03$i = 0, $$03$i38 = 0, $$a_$i = 0, $$a_$i1 = 0, $$a_$i14 = 0, $$a_$i25 = 0, $$a_$i3 = 0, $$a_$i37 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $a$0$i = 0, $a$0$i40 = 0, $i$0$i$i = 0, $i$0$i$i20 = 0, $i$0$i$i31 = 0, $i$0$i$i9 = 0, $or$cond$i = 0, $or$cond$i36 = 0, $q$0$i = 0, $q$0$i$i = 0, $q$0$i$i19 = 0, $q$0$i$i30 = 0, $q$0$i$i8 = 0, $q$0$i12 = 0, $q$0$i23 = 0, $q$0$i34 = 0, $q$1$i$i = 0, $q$1$i$i11 = 0, $q$1$i$i22 = 0;
 var $q$1$i$i33 = 0, $r$0$i$i = 0, $r$0$i$i18 = 0, $r$0$i$i29 = 0, $r$0$i$i7 = 0, $r$1$i$i = 0, $r$1$i$i10 = 0, $r$1$i$i21 = 0, $r$1$i$i32 = 0, $s$1$i = 0, $s$1$i15 = 0, $s$1$i2 = 0, $s$1$i26 = 0, $s$1$i39 = 0, $s$1$i4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($matrix|0)==(0|0);
 if ($0) {
  $$0 = 6;
  return ($$0|0);
 }
 $1 = HEAP32[$matrix>>2]|0;
 $2 = ((($matrix)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==(0);
 $5 = ($3|0)==(65536);
 $or$cond$i = $4 | $5;
 if ($or$cond$i) {
  $$0$i = $1;
 } else {
  $6 = ($1|0)<(0);
  $7 = (0 - ($1))|0;
  $8 = $1 >> 31;
  $9 = $8 | 1;
  $$a_$i = $6 ? $7 : $1;
  $10 = ($3|0)<(0);
  $11 = (0 - ($3))|0;
  $12 = (0 - ($9))|0;
  $$03$i = $10 ? $11 : $3;
  $s$1$i = $10 ? $12 : $8;
  $13 = $$03$i >>> 8;
  $14 = (($13) + ($$a_$i))|0;
  $15 = ($14>>>0)<(8191);
  if ($15) {
   $16 = Math_imul($$03$i, $$a_$i)|0;
   $17 = (($16) + 32768)|0;
   $18 = $17 >>> 16;
   $a$0$i = $18;
  } else {
   $19 = $$a_$i & 65535;
   $20 = $$a_$i >>> 16;
   $21 = Math_imul($20, $$03$i)|0;
   $22 = $$03$i >>> 16;
   $23 = Math_imul($22, $19)|0;
   $24 = (($23) + ($21))|0;
   $25 = $$03$i & 65535;
   $26 = Math_imul($25, $19)|0;
   $27 = (($26) + 32768)|0;
   $28 = $27 >>> 16;
   $29 = (($24) + ($28))|0;
   $a$0$i = $29;
  }
  $30 = ($s$1$i|0)<(0);
  $31 = (0 - ($a$0$i))|0;
  $32 = $30 ? $31 : $a$0$i;
  $$0$i = $32;
 }
 $33 = ((($matrix)) + 4|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = ((($matrix)) + 8|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = ($34|0)==(0);
 $38 = ($36|0)==(65536);
 $or$cond$i36 = $37 | $38;
 if ($or$cond$i36) {
  $$0$i41 = $34;
 } else {
  $39 = ($34|0)<(0);
  $40 = (0 - ($34))|0;
  $41 = $34 >> 31;
  $42 = $41 | 1;
  $$a_$i37 = $39 ? $40 : $34;
  $43 = ($36|0)<(0);
  $44 = (0 - ($36))|0;
  $45 = (0 - ($42))|0;
  $$03$i38 = $43 ? $44 : $36;
  $s$1$i39 = $43 ? $45 : $41;
  $46 = $$03$i38 >>> 8;
  $47 = (($46) + ($$a_$i37))|0;
  $48 = ($47>>>0)<(8191);
  if ($48) {
   $49 = Math_imul($$03$i38, $$a_$i37)|0;
   $50 = (($49) + 32768)|0;
   $51 = $50 >>> 16;
   $a$0$i40 = $51;
  } else {
   $52 = $$a_$i37 & 65535;
   $53 = $$a_$i37 >>> 16;
   $54 = Math_imul($53, $$03$i38)|0;
   $55 = $$03$i38 >>> 16;
   $56 = Math_imul($55, $52)|0;
   $57 = (($56) + ($54))|0;
   $58 = $$03$i38 & 65535;
   $59 = Math_imul($58, $52)|0;
   $60 = (($59) + 32768)|0;
   $61 = $60 >>> 16;
   $62 = (($57) + ($61))|0;
   $a$0$i40 = $62;
  }
  $63 = ($s$1$i39|0)<(0);
  $64 = (0 - ($a$0$i40))|0;
  $65 = $63 ? $64 : $a$0$i40;
  $$0$i41 = $65;
 }
 $66 = (($$0$i) - ($$0$i41))|0;
 $67 = ($$0$i|0)==($$0$i41|0);
 if ($67) {
  $$0 = 6;
  return ($$0|0);
 }
 $68 = ($34|0)<(0);
 $69 = (0 - ($34))|0;
 $70 = $34 >> 31;
 $71 = $70 | 1;
 $$a_$i25 = $68 ? $69 : $34;
 $72 = ($66|0)<(0);
 $73 = (0 - ($66))|0;
 $74 = (0 - ($71))|0;
 $s$1$i26 = $72 ? $74 : $70;
 $$01$i27 = $72 ? $73 : $66;
 $75 = ($$01$i27|0)==(0);
 do {
  if ($75) {
   $q$0$i34 = 2147483647;
  } else {
   $76 = $$01$i27 >>> 17;
   $77 = $76 ^ 65535;
   $78 = ($$a_$i25>>>0)>($77>>>0);
   if (!($78)) {
    $79 = $$a_$i25 << 16;
    $80 = $$01$i27 >>> 1;
    $81 = (($79) + ($80))|0;
    $82 = (($81>>>0) / ($$01$i27>>>0))&-1;
    $q$0$i34 = $82;
    break;
   }
   $83 = $$a_$i25 >>> 16;
   $84 = $$a_$i25 << 16;
   $85 = $$01$i27 >>> 1;
   $86 = (($84) + ($85))|0;
   $87 = ($86>>>0)<($84>>>0);
   $88 = $87&1;
   $89 = (($88) + ($83))|0;
   $90 = ($89>>>0)<($$01$i27>>>0);
   if ($90) {
    $91 = (Math_clz32(($89|0))|0);
    $92 = $89 << $91;
    $93 = (32 - ($91))|0;
    $94 = $86 >>> $93;
    $95 = $94 | $92;
    $96 = $86 << $91;
    $97 = (($95>>>0) / ($$01$i27>>>0))&-1;
    $98 = Math_imul($97, $$01$i27)|0;
    $99 = (($95) - ($98))|0;
    $$01$i$i28 = $96;$i$0$i$i31 = $93;$q$0$i$i30 = $97;$r$0$i$i29 = $99;
    while(1) {
     $100 = $q$0$i$i30 << 1;
     $101 = $r$0$i$i29 << 1;
     $102 = $$01$i$i28 >>> 31;
     $103 = $101 | $102;
     $104 = $$01$i$i28 << 1;
     $105 = ($103>>>0)<($$01$i27>>>0);
     $106 = $105 ? 0 : $$01$i27;
     $r$1$i$i32 = (($103) - ($106))|0;
     $107 = $105&1;
     $108 = $107 | $100;
     $q$1$i$i33 = $108 ^ 1;
     $109 = (($i$0$i$i31) + -1)|0;
     $110 = ($109|0)==(0);
     if ($110) {
      $q$0$i34 = $q$1$i$i33;
      break;
     } else {
      $$01$i$i28 = $104;$i$0$i$i31 = $109;$q$0$i$i30 = $q$1$i$i33;$r$0$i$i29 = $r$1$i$i32;
     }
    }
   } else {
    $q$0$i34 = 2147483647;
   }
  }
 } while(0);
 $111 = ($s$1$i26|0)<(0);
 $112 = (0 - ($q$0$i34))|0;
 $113 = $111 ? $112 : $q$0$i34;
 $114 = (0 - ($113))|0;
 HEAP32[$33>>2] = $114;
 $115 = ($36|0)<(0);
 $116 = (0 - ($36))|0;
 $117 = $36 >> 31;
 $118 = $117 | 1;
 $$a_$i14 = $115 ? $116 : $36;
 $119 = (0 - ($118))|0;
 $s$1$i15 = $72 ? $119 : $117;
 do {
  if ($75) {
   $q$0$i23 = 2147483647;
  } else {
   $120 = $$01$i27 >>> 17;
   $121 = $120 ^ 65535;
   $122 = ($$a_$i14>>>0)>($121>>>0);
   if (!($122)) {
    $123 = $$a_$i14 << 16;
    $124 = $$01$i27 >>> 1;
    $125 = (($123) + ($124))|0;
    $126 = (($125>>>0) / ($$01$i27>>>0))&-1;
    $q$0$i23 = $126;
    break;
   }
   $127 = $$a_$i14 >>> 16;
   $128 = $$a_$i14 << 16;
   $129 = $$01$i27 >>> 1;
   $130 = (($128) + ($129))|0;
   $131 = ($130>>>0)<($128>>>0);
   $132 = $131&1;
   $133 = (($132) + ($127))|0;
   $134 = ($133>>>0)<($$01$i27>>>0);
   if ($134) {
    $135 = (Math_clz32(($133|0))|0);
    $136 = $133 << $135;
    $137 = (32 - ($135))|0;
    $138 = $130 >>> $137;
    $139 = $138 | $136;
    $140 = $130 << $135;
    $141 = (($139>>>0) / ($$01$i27>>>0))&-1;
    $142 = Math_imul($141, $$01$i27)|0;
    $143 = (($139) - ($142))|0;
    $$01$i$i17 = $140;$i$0$i$i20 = $137;$q$0$i$i19 = $141;$r$0$i$i18 = $143;
    while(1) {
     $144 = $q$0$i$i19 << 1;
     $145 = $r$0$i$i18 << 1;
     $146 = $$01$i$i17 >>> 31;
     $147 = $145 | $146;
     $148 = $$01$i$i17 << 1;
     $149 = ($147>>>0)<($$01$i27>>>0);
     $150 = $149 ? 0 : $$01$i27;
     $r$1$i$i21 = (($147) - ($150))|0;
     $151 = $149&1;
     $152 = $151 | $144;
     $q$1$i$i22 = $152 ^ 1;
     $153 = (($i$0$i$i20) + -1)|0;
     $154 = ($153|0)==(0);
     if ($154) {
      $q$0$i23 = $q$1$i$i22;
      break;
     } else {
      $$01$i$i17 = $148;$i$0$i$i20 = $153;$q$0$i$i19 = $q$1$i$i22;$r$0$i$i18 = $r$1$i$i21;
     }
    }
   } else {
    $q$0$i23 = 2147483647;
   }
  }
 } while(0);
 $155 = ($s$1$i15|0)<(0);
 $156 = (0 - ($q$0$i23))|0;
 $157 = $155 ? $156 : $q$0$i23;
 $158 = (0 - ($157))|0;
 HEAP32[$35>>2] = $158;
 $159 = ($3|0)<(0);
 $160 = (0 - ($3))|0;
 $161 = $3 >> 31;
 $162 = $161 | 1;
 $$a_$i3 = $159 ? $160 : $3;
 $163 = (0 - ($162))|0;
 $s$1$i4 = $72 ? $163 : $161;
 do {
  if ($75) {
   $q$0$i12 = 2147483647;
  } else {
   $164 = $$01$i27 >>> 17;
   $165 = $164 ^ 65535;
   $166 = ($$a_$i3>>>0)>($165>>>0);
   if (!($166)) {
    $167 = $$a_$i3 << 16;
    $168 = $$01$i27 >>> 1;
    $169 = (($167) + ($168))|0;
    $170 = (($169>>>0) / ($$01$i27>>>0))&-1;
    $q$0$i12 = $170;
    break;
   }
   $171 = $$a_$i3 >>> 16;
   $172 = $$a_$i3 << 16;
   $173 = $$01$i27 >>> 1;
   $174 = (($172) + ($173))|0;
   $175 = ($174>>>0)<($172>>>0);
   $176 = $175&1;
   $177 = (($176) + ($171))|0;
   $178 = ($177>>>0)<($$01$i27>>>0);
   if ($178) {
    $179 = (Math_clz32(($177|0))|0);
    $180 = $177 << $179;
    $181 = (32 - ($179))|0;
    $182 = $174 >>> $181;
    $183 = $182 | $180;
    $184 = $174 << $179;
    $185 = (($183>>>0) / ($$01$i27>>>0))&-1;
    $186 = Math_imul($185, $$01$i27)|0;
    $187 = (($183) - ($186))|0;
    $$01$i$i6 = $184;$i$0$i$i9 = $181;$q$0$i$i8 = $185;$r$0$i$i7 = $187;
    while(1) {
     $188 = $q$0$i$i8 << 1;
     $189 = $r$0$i$i7 << 1;
     $190 = $$01$i$i6 >>> 31;
     $191 = $189 | $190;
     $192 = $$01$i$i6 << 1;
     $193 = ($191>>>0)<($$01$i27>>>0);
     $194 = $193 ? 0 : $$01$i27;
     $r$1$i$i10 = (($191) - ($194))|0;
     $195 = $193&1;
     $196 = $195 | $188;
     $q$1$i$i11 = $196 ^ 1;
     $197 = (($i$0$i$i9) + -1)|0;
     $198 = ($197|0)==(0);
     if ($198) {
      $q$0$i12 = $q$1$i$i11;
      break;
     } else {
      $$01$i$i6 = $192;$i$0$i$i9 = $197;$q$0$i$i8 = $q$1$i$i11;$r$0$i$i7 = $r$1$i$i10;
     }
    }
   } else {
    $q$0$i12 = 2147483647;
   }
  }
 } while(0);
 $199 = ($s$1$i4|0)<(0);
 $200 = (0 - ($q$0$i12))|0;
 $201 = $199 ? $200 : $q$0$i12;
 HEAP32[$matrix>>2] = $201;
 $202 = ($1|0)<(0);
 $203 = (0 - ($1))|0;
 $204 = $1 >> 31;
 $205 = $204 | 1;
 $$a_$i1 = $202 ? $203 : $1;
 $206 = (0 - ($205))|0;
 $s$1$i2 = $72 ? $206 : $204;
 do {
  if ($75) {
   $q$0$i = 2147483647;
  } else {
   $207 = $$01$i27 >>> 17;
   $208 = $207 ^ 65535;
   $209 = ($$a_$i1>>>0)>($208>>>0);
   if (!($209)) {
    $210 = $$a_$i1 << 16;
    $211 = $$01$i27 >>> 1;
    $212 = (($210) + ($211))|0;
    $213 = (($212>>>0) / ($$01$i27>>>0))&-1;
    $q$0$i = $213;
    break;
   }
   $214 = $$a_$i1 >>> 16;
   $215 = $$a_$i1 << 16;
   $216 = $$01$i27 >>> 1;
   $217 = (($215) + ($216))|0;
   $218 = ($217>>>0)<($215>>>0);
   $219 = $218&1;
   $220 = (($219) + ($214))|0;
   $221 = ($220>>>0)<($$01$i27>>>0);
   if ($221) {
    $222 = (Math_clz32(($220|0))|0);
    $223 = $220 << $222;
    $224 = (32 - ($222))|0;
    $225 = $217 >>> $224;
    $226 = $225 | $223;
    $227 = $217 << $222;
    $228 = (($226>>>0) / ($$01$i27>>>0))&-1;
    $229 = Math_imul($228, $$01$i27)|0;
    $230 = (($226) - ($229))|0;
    $$01$i$i = $227;$i$0$i$i = $224;$q$0$i$i = $228;$r$0$i$i = $230;
    while(1) {
     $231 = $q$0$i$i << 1;
     $232 = $r$0$i$i << 1;
     $233 = $$01$i$i >>> 31;
     $234 = $232 | $233;
     $235 = $$01$i$i << 1;
     $236 = ($234>>>0)<($$01$i27>>>0);
     $237 = $236 ? 0 : $$01$i27;
     $r$1$i$i = (($234) - ($237))|0;
     $238 = $236&1;
     $239 = $238 | $231;
     $q$1$i$i = $239 ^ 1;
     $240 = (($i$0$i$i) + -1)|0;
     $241 = ($240|0)==(0);
     if ($241) {
      $q$0$i = $q$1$i$i;
      break;
     } else {
      $$01$i$i = $235;$i$0$i$i = $240;$q$0$i$i = $q$1$i$i;$r$0$i$i = $r$1$i$i;
     }
    }
   } else {
    $q$0$i = 2147483647;
   }
  }
 } while(0);
 $242 = ($s$1$i2|0)<(0);
 $243 = (0 - ($q$0$i))|0;
 $244 = $242 ? $243 : $q$0$i;
 HEAP32[$2>>2] = $244;
 $$0 = 0;
 return ($$0|0);
}
function _FT_Matrix_Multiply_Scaled($a,$b,$scaling) {
 $a = $a|0;
 $b = $b|0;
 $scaling = $scaling|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $scaling << 16;
 $1 = ($a|0)!=(0|0);
 $2 = ($b|0)!=(0|0);
 $or$cond = $1 & $2;
 if (!($or$cond)) {
  return;
 }
 $3 = HEAP32[$a>>2]|0;
 $4 = HEAP32[$b>>2]|0;
 $5 = (_FT_MulDiv($3,$4,$0)|0);
 $6 = ((($a)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($b)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (_FT_MulDiv($7,$9,$0)|0);
 $11 = (($10) + ($5))|0;
 $12 = ((($b)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (_FT_MulDiv($3,$13,$0)|0);
 $15 = ((($b)) + 12|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = (_FT_MulDiv($7,$16,$0)|0);
 $18 = (($17) + ($14))|0;
 $19 = ((($a)) + 8|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (_FT_MulDiv($20,$4,$0)|0);
 $22 = ((($a)) + 12|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = (_FT_MulDiv($23,$9,$0)|0);
 $25 = (($24) + ($21))|0;
 $26 = (_FT_MulDiv($20,$13,$0)|0);
 $27 = (_FT_MulDiv($23,$16,$0)|0);
 $28 = (($27) + ($26))|0;
 HEAP32[$b>>2] = $11;
 HEAP32[$12>>2] = $18;
 HEAP32[$8>>2] = $25;
 HEAP32[$15>>2] = $28;
 return;
}
function _FT_Vector_Transform_Scaled($vector,$matrix,$scaling) {
 $vector = $vector|0;
 $matrix = $matrix|0;
 $scaling = $scaling|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $scaling << 16;
 $1 = ($vector|0)!=(0|0);
 $2 = ($matrix|0)!=(0|0);
 $or$cond = $1 & $2;
 if (!($or$cond)) {
  return;
 }
 $3 = HEAP32[$vector>>2]|0;
 $4 = HEAP32[$matrix>>2]|0;
 $5 = (_FT_MulDiv($3,$4,$0)|0);
 $6 = ((($vector)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($matrix)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (_FT_MulDiv($7,$9,$0)|0);
 $11 = (($10) + ($5))|0;
 $12 = ((($matrix)) + 8|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (_FT_MulDiv($3,$13,$0)|0);
 $15 = ((($matrix)) + 12|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = (_FT_MulDiv($7,$16,$0)|0);
 $18 = (($17) + ($14))|0;
 HEAP32[$vector>>2] = $11;
 HEAP32[$6>>2] = $18;
 return;
}
function _ft_corner_orientation($in_x,$in_y,$out_x,$out_y) {
 $in_x = $in_x|0;
 $in_y = $in_y|0;
 $out_x = $out_x|0;
 $out_y = $out_y|0;
 var $$ = 0, $$in_x = 0, $$out_x = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $in_y$ = 0;
 var $out_y$ = 0, $result$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($in_y|0)==(0);
 if ($0) {
  $1 = ($in_x|0)>(-1);
  $2 = (0 - ($out_y))|0;
  $out_y$ = $1 ? $out_y : $2;
  return ($out_y$|0);
 }
 $3 = ($in_x|0)==(0);
 if ($3) {
  $4 = ($in_y|0)>(-1);
  $5 = (0 - ($out_x))|0;
  $$out_x = $4 ? $5 : $out_x;
  $result$1 = $$out_x;
  return ($result$1|0);
 }
 $6 = ($out_y|0)==(0);
 if ($6) {
  $7 = ($out_x|0)>(-1);
  $8 = (0 - ($in_y))|0;
  $in_y$ = $7 ? $in_y : $8;
  $result$1 = $in_y$;
  return ($result$1|0);
 }
 $9 = ($out_x|0)==(0);
 if ($9) {
  $10 = ($out_y|0)>(-1);
  $11 = (0 - ($in_x))|0;
  $$in_x = $10 ? $11 : $in_x;
  $result$1 = $$in_x;
  return ($result$1|0);
 }
 $12 = $in_x & 65535;
 $13 = $in_x >>> 16;
 $14 = $out_y & 65535;
 $15 = $out_y >>> 16;
 $16 = Math_imul($14, $12)|0;
 $17 = Math_imul($15, $12)|0;
 $18 = Math_imul($14, $13)|0;
 $19 = Math_imul($15, $13)|0;
 $20 = (($17) + ($18))|0;
 $21 = ($20>>>0)<($18>>>0);
 $22 = $21&1;
 $23 = $22 << 16;
 $24 = $20 >>> 16;
 $25 = $20 << 16;
 $26 = (($25) + ($16))|0;
 $27 = ($26>>>0)<($25>>>0);
 $28 = $27&1;
 $29 = (($24) + ($19))|0;
 $30 = (($29) + ($23))|0;
 $31 = (($30) + ($28))|0;
 $32 = $in_y & 65535;
 $33 = $in_y >>> 16;
 $34 = $out_x & 65535;
 $35 = $out_x >>> 16;
 $36 = Math_imul($34, $32)|0;
 $37 = Math_imul($35, $32)|0;
 $38 = Math_imul($34, $33)|0;
 $39 = Math_imul($35, $33)|0;
 $40 = (($37) + ($38))|0;
 $41 = ($40>>>0)<($38>>>0);
 $42 = $41&1;
 $43 = $42 << 16;
 $44 = $40 >>> 16;
 $45 = $40 << 16;
 $46 = (($45) + ($36))|0;
 $47 = ($46>>>0)<($45>>>0);
 $48 = $47&1;
 $49 = (($44) + ($39))|0;
 $50 = (($49) + ($43))|0;
 $51 = (($50) + ($48))|0;
 $52 = ($31>>>0)>($51>>>0);
 if ($52) {
  $result$1 = 1;
  return ($result$1|0);
 }
 $53 = ($31>>>0)<($51>>>0);
 if ($53) {
  $result$1 = -1;
  return ($result$1|0);
 }
 $54 = ($26>>>0)>($46>>>0);
 if ($54) {
  $result$1 = 1;
  return ($result$1|0);
 }
 $55 = ($26>>>0)<($46>>>0);
 $$ = $55 << 31 >> 31;
 $result$1 = $$;
 return ($result$1|0);
}
function _ft_corner_is_flat($in_x,$in_y,$out_x,$out_y) {
 $in_x = $in_x|0;
 $in_y = $in_y|0;
 $out_x = $out_x|0;
 $out_y = $out_y|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($out_x) + ($in_x))|0;
 $1 = (($out_y) + ($in_y))|0;
 $2 = ($in_x|0)<(0);
 $3 = (0 - ($in_x))|0;
 $4 = $2 ? $3 : $in_x;
 $5 = ($in_y|0)<(0);
 $6 = (0 - ($in_y))|0;
 $7 = $5 ? $6 : $in_y;
 $8 = ($4|0)>($7|0);
 if ($8) {
  $9 = ($7*3)|0;
  $10 = $9 >> 3;
  $11 = (($10) + ($4))|0;
  $43 = $11;
 } else {
  $12 = ($4*3)|0;
  $13 = $12 >> 3;
  $14 = (($13) + ($7))|0;
  $43 = $14;
 }
 $15 = ($out_x|0)<(0);
 $16 = (0 - ($out_x))|0;
 $17 = $15 ? $16 : $out_x;
 $18 = ($out_y|0)<(0);
 $19 = (0 - ($out_y))|0;
 $20 = $18 ? $19 : $out_y;
 $21 = ($17|0)>($20|0);
 if ($21) {
  $22 = ($20*3)|0;
  $23 = $22 >> 3;
  $24 = (($23) + ($17))|0;
  $42 = $24;
 } else {
  $25 = ($17*3)|0;
  $26 = $25 >> 3;
  $27 = (($26) + ($20))|0;
  $42 = $27;
 }
 $28 = ($0|0)<(0);
 $29 = (0 - ($0))|0;
 $30 = $28 ? $29 : $0;
 $31 = ($1|0)<(0);
 $32 = (0 - ($1))|0;
 $33 = $31 ? $32 : $1;
 $34 = ($30|0)>($33|0);
 if ($34) {
  $35 = ($33*3)|0;
  $36 = $35 >> 3;
  $37 = (($36) + ($30))|0;
  $45 = $37;
  $41 = (($42) + ($43))|0;
  $44 = (($41) - ($45))|0;
  $46 = $45 >> 4;
  $47 = ($44|0)<($46|0);
  $48 = $47&1;
  return ($48|0);
 } else {
  $38 = ($30*3)|0;
  $39 = $38 >> 3;
  $40 = (($39) + ($33))|0;
  $45 = $40;
  $41 = (($42) + ($43))|0;
  $44 = (($41) - ($45))|0;
  $46 = $45 >> 4;
  $47 = ($44|0)<($46|0);
  $48 = $47&1;
  return ($48|0);
 }
 return (0)|0;
}
function _FT_GlyphLoader_New($memory,$aloader) {
 $memory = $memory|0;
 $aloader = $aloader|0;
 var $0 = 0, $1 = 0, $2 = 0, $error = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = (_ft_mem_alloc($memory,96,$error)|0);
 $1 = HEAP32[$error>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  STACKTOP = sp;return ($1|0);
 }
 HEAP32[$0>>2] = $memory;
 HEAP32[$aloader>>2] = $0;
 STACKTOP = sp;return ($1|0);
}
function _FT_GlyphLoader_Rewind($loader) {
 $loader = $loader|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($loader)) + 20|0);
 $1 = ((($loader)) + 56|0);
 $2 = ((($loader)) + 22|0);
 HEAP16[$2>>1] = 0;
 HEAP16[$0>>1] = 0;
 $3 = ((($loader)) + 48|0);
 HEAP32[$3>>2] = 0;
 dest=$1; src=$0; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 return;
}
function _FT_GlyphLoader_Done($loader) {
 $loader = $loader|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ($loader|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = HEAP32[$loader>>2]|0;
 $2 = ((($loader)) + 24|0);
 $3 = HEAP32[$2>>2]|0;
 _ft_mem_free($1,$3);
 HEAP32[$2>>2] = 0;
 $4 = ((($loader)) + 28|0);
 $5 = HEAP32[$4>>2]|0;
 _ft_mem_free($1,$5);
 HEAP32[$4>>2] = 0;
 $6 = ((($loader)) + 32|0);
 $7 = HEAP32[$6>>2]|0;
 _ft_mem_free($1,$7);
 HEAP32[$6>>2] = 0;
 $8 = ((($loader)) + 40|0);
 $9 = HEAP32[$8>>2]|0;
 _ft_mem_free($1,$9);
 HEAP32[$8>>2] = 0;
 $10 = ((($loader)) + 52|0);
 $11 = HEAP32[$10>>2]|0;
 _ft_mem_free($1,$11);
 HEAP32[$10>>2] = 0;
 $12 = ((($loader)) + 44|0);
 HEAP32[$12>>2] = 0;
 $13 = ((($loader)) + 4|0);
 HEAP32[$13>>2] = 0;
 $14 = ((($loader)) + 8|0);
 HEAP32[$14>>2] = 0;
 $15 = ((($loader)) + 12|0);
 HEAP32[$15>>2] = 0;
 $16 = ((($loader)) + 20|0);
 $17 = ((($loader)) + 56|0);
 $18 = ((($loader)) + 22|0);
 HEAP16[$18>>1] = 0;
 HEAP16[$16>>1] = 0;
 $19 = ((($loader)) + 48|0);
 HEAP32[$19>>2] = 0;
 dest=$17; src=$16; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 _ft_mem_free($1,$loader);
 return;
}
function _FT_GlyphLoader_CreateExtra($loader) {
 $loader = $loader|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $sext$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = HEAP32[$loader>>2]|0;
 $1 = ((($loader)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2 << 1;
 $4 = (_ft_mem_realloc($0,8,0,$3,0,$error)|0);
 $5 = ((($loader)) + 40|0);
 HEAP32[$5>>2] = $4;
 $6 = HEAP32[$error>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  STACKTOP = sp;return ($6|0);
 }
 $8 = ((($loader)) + 16|0);
 HEAP8[$8>>0] = 1;
 $9 = HEAP32[$1>>2]|0;
 $10 = (($4) + ($9<<3)|0);
 $11 = ((($loader)) + 44|0);
 HEAP32[$11>>2] = $10;
 $12 = ((($loader)) + 24|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($loader)) + 22|0);
 $15 = HEAP16[$14>>1]|0;
 $16 = $15 << 16 >> 16;
 $17 = (($13) + ($16<<3)|0);
 $18 = ((($loader)) + 60|0);
 HEAP32[$18>>2] = $17;
 $19 = ((($loader)) + 28|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($20) + ($16)|0);
 $22 = ((($loader)) + 64|0);
 HEAP32[$22>>2] = $21;
 $23 = ((($loader)) + 32|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($loader)) + 20|0);
 $26 = HEAP32[$25>>2]|0;
 $sext$i = $26 << 16;
 $27 = $sext$i >> 16;
 $28 = (($24) + ($27<<1)|0);
 $29 = ((($loader)) + 68|0);
 HEAP32[$29>>2] = $28;
 $30 = $26 >> 16;
 $31 = (($4) + ($30<<3)|0);
 $32 = ((($loader)) + 76|0);
 HEAP32[$32>>2] = $31;
 $33 = (($10) + ($30<<3)|0);
 $34 = ((($loader)) + 80|0);
 HEAP32[$34>>2] = $33;
 STACKTOP = sp;return ($6|0);
}
function _FT_GlyphLoader_CheckPoints($loader,$n_points,$n_contours) {
 $loader = $loader|0;
 $n_points = $n_points|0;
 $n_contours = $n_contours|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$pr = 0, $$pre = 0, $$pre1 = 0, $$pre3 = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $11 = 0, $12 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $adjust$0 = 0, $error = 0, $sext = 0, $sext$i = 0, dest = 0;
 var label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = HEAP32[$loader>>2]|0;
 HEAP32[$error>>2] = 0;
 $1 = ((($loader)) + 22|0);
 $2 = HEAP16[$1>>1]|0;
 $3 = $2 << 16 >> 16;
 $4 = ((($loader)) + 58|0);
 $5 = HEAP16[$4>>1]|0;
 $6 = $5 << 16 >> 16;
 $7 = (($3) + ($n_points))|0;
 $8 = (($7) + ($6))|0;
 $9 = ((($loader)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($8>>>0)>($10>>>0);
 do {
  if ($11) {
   $12 = (($8) + 7)|0;
   $13 = $12 & -8;
   $14 = ($13>>>0)>(32767);
   if ($14) {
    $$0 = 10;
    STACKTOP = sp;return ($$0|0);
   }
   $15 = ((($loader)) + 24|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = (_ft_mem_realloc($0,8,$10,$13,$16,$error)|0);
   HEAP32[$15>>2] = $17;
   $18 = HEAP32[$error>>2]|0;
   $19 = ($18|0)==(0);
   if ($19) {
    $20 = ((($loader)) + 28|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = (_ft_mem_realloc($0,1,$10,$13,$21,$error)|0);
    HEAP32[$20>>2] = $22;
    $23 = HEAP32[$error>>2]|0;
    $24 = ($23|0)==(0);
    if ($24) {
     $25 = ((($loader)) + 16|0);
     $26 = HEAP8[$25>>0]|0;
     $27 = ($26<<24>>24)==(0);
     if (!($27)) {
      $28 = $10 << 1;
      $29 = $13 << 1;
      $30 = ((($loader)) + 40|0);
      $31 = HEAP32[$30>>2]|0;
      $32 = (_ft_mem_realloc($0,8,$28,$29,$31,$error)|0);
      HEAP32[$30>>2] = $32;
      $33 = HEAP32[$error>>2]|0;
      $34 = ($33|0)==(0);
      if (!($34)) {
       break;
      }
      $35 = (($32) + ($13<<3)|0);
      $36 = (($32) + ($10<<3)|0);
      $37 = $10 << 3;
      _memmove(($35|0),($36|0),($37|0))|0;
      $38 = HEAP32[$30>>2]|0;
      $39 = (($38) + ($13<<3)|0);
      $40 = ((($loader)) + 44|0);
      HEAP32[$40>>2] = $39;
     }
     HEAP32[$9>>2] = $13;
     $adjust$0 = 1;
     label = 9;
    }
   }
  } else {
   $adjust$0 = 0;
   label = 9;
  }
 } while(0);
 do {
  if ((label|0) == 9) {
   $41 = ((($loader)) + 8|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = ((($loader)) + 20|0);
   $44 = HEAP32[$43>>2]|0;
   $sext = $44 << 16;
   $45 = $sext >> 16;
   $46 = ((($loader)) + 56|0);
   $47 = HEAP16[$46>>1]|0;
   $48 = $47 << 16 >> 16;
   $49 = (($45) + ($n_contours))|0;
   $50 = (($49) + ($48))|0;
   $51 = ($50>>>0)>($42>>>0);
   $52 = $44 >>> 16;
   $53 = $52&65535;
   if ($51) {
    $54 = (($50) + 3)|0;
    $55 = $54 & -4;
    $56 = ($55>>>0)>(32767);
    if ($56) {
     $$0 = 10;
     STACKTOP = sp;return ($$0|0);
    }
    $57 = ((($loader)) + 32|0);
    $58 = HEAP32[$57>>2]|0;
    $59 = (_ft_mem_realloc($0,2,$42,$55,$58,$error)|0);
    HEAP32[$57>>2] = $59;
    $60 = HEAP32[$error>>2]|0;
    $61 = ($60|0)==(0);
    if (!($61)) {
     break;
    }
    HEAP32[$41>>2] = $55;
    $$pre = HEAP16[$1>>1]|0;
    $$pre3 = HEAP32[$43>>2]|0;
    $66 = $$pre;$73 = $$pre3;$76 = $59;
    label = 15;
   } else {
    $62 = ($adjust$0<<24>>24)==(0);
    if (!($62)) {
     $$phi$trans$insert = ((($loader)) + 32|0);
     $$pre1 = HEAP32[$$phi$trans$insert>>2]|0;
     $66 = $53;$73 = $44;$76 = $$pre1;
     label = 15;
    }
   }
   if ((label|0) == 15) {
    $63 = ((($loader)) + 24|0);
    $64 = HEAP32[$63>>2]|0;
    $65 = $66 << 16 >> 16;
    $67 = (($64) + ($65<<3)|0);
    $68 = ((($loader)) + 60|0);
    HEAP32[$68>>2] = $67;
    $69 = ((($loader)) + 28|0);
    $70 = HEAP32[$69>>2]|0;
    $71 = (($70) + ($65)|0);
    $72 = ((($loader)) + 64|0);
    HEAP32[$72>>2] = $71;
    $sext$i = $73 << 16;
    $74 = $sext$i >> 16;
    $75 = (($76) + ($74<<1)|0);
    $77 = ((($loader)) + 68|0);
    HEAP32[$77>>2] = $75;
    $78 = ((($loader)) + 16|0);
    $79 = HEAP8[$78>>0]|0;
    $80 = ($79<<24>>24)==(0);
    if (!($80)) {
     $81 = ((($loader)) + 40|0);
     $82 = HEAP32[$81>>2]|0;
     $83 = $73 >> 16;
     $84 = (($82) + ($83<<3)|0);
     $85 = ((($loader)) + 76|0);
     HEAP32[$85>>2] = $84;
     $86 = ((($loader)) + 44|0);
     $87 = HEAP32[$86>>2]|0;
     $88 = (($87) + ($83<<3)|0);
     $89 = ((($loader)) + 80|0);
     HEAP32[$89>>2] = $88;
    }
   }
   $$pr = HEAP32[$error>>2]|0;
   $90 = ($$pr|0)==(0);
   if ($90) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 $91 = HEAP32[$loader>>2]|0;
 $92 = ((($loader)) + 24|0);
 $93 = HEAP32[$92>>2]|0;
 _ft_mem_free($91,$93);
 HEAP32[$92>>2] = 0;
 $94 = ((($loader)) + 28|0);
 $95 = HEAP32[$94>>2]|0;
 _ft_mem_free($91,$95);
 HEAP32[$94>>2] = 0;
 $96 = ((($loader)) + 32|0);
 $97 = HEAP32[$96>>2]|0;
 _ft_mem_free($91,$97);
 HEAP32[$96>>2] = 0;
 $98 = ((($loader)) + 40|0);
 $99 = HEAP32[$98>>2]|0;
 _ft_mem_free($91,$99);
 HEAP32[$98>>2] = 0;
 $100 = ((($loader)) + 52|0);
 $101 = HEAP32[$100>>2]|0;
 _ft_mem_free($91,$101);
 HEAP32[$100>>2] = 0;
 $102 = ((($loader)) + 44|0);
 HEAP32[$102>>2] = 0;
 HEAP32[$9>>2] = 0;
 $103 = ((($loader)) + 8|0);
 HEAP32[$103>>2] = 0;
 $104 = ((($loader)) + 12|0);
 HEAP32[$104>>2] = 0;
 $105 = ((($loader)) + 20|0);
 $106 = ((($loader)) + 56|0);
 HEAP16[$1>>1] = 0;
 HEAP16[$105>>1] = 0;
 $107 = ((($loader)) + 48|0);
 HEAP32[$107>>2] = 0;
 dest=$106; src=$105; stop=dest+36|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 $$pre4 = HEAP32[$error>>2]|0;
 $$0 = $$pre4;
 STACKTOP = sp;return ($$0|0);
}
function _FT_GlyphLoader_CheckSubGlyphs($loader,$n_subs) {
 $loader = $loader|0;
 $n_subs = $n_subs|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $error = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = HEAP32[$loader>>2]|0;
 HEAP32[$error>>2] = 0;
 $1 = ((($loader)) + 48|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($loader)) + 84|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2) + ($n_subs))|0;
 $6 = (($5) + ($4))|0;
 $7 = ((($loader)) + 12|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($6>>>0)>($8>>>0);
 if (!($9)) {
  $20 = 0;
  STACKTOP = sp;return ($20|0);
 }
 $10 = (($6) + 1)|0;
 $11 = $10 & -2;
 $12 = ((($loader)) + 52|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (_ft_mem_realloc($0,32,$8,$11,$13,$error)|0);
 HEAP32[$12>>2] = $14;
 $15 = HEAP32[$error>>2]|0;
 $16 = ($15|0)==(0);
 if (!($16)) {
  $20 = $15;
  STACKTOP = sp;return ($20|0);
 }
 HEAP32[$7>>2] = $11;
 $17 = HEAP32[$1>>2]|0;
 $18 = (($14) + ($17<<5)|0);
 $19 = ((($loader)) + 88|0);
 HEAP32[$19>>2] = $18;
 $20 = 0;
 STACKTOP = sp;return ($20|0);
}
function _FT_GlyphLoader_Prepare($loader) {
 $loader = $loader|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $sext$i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($loader)) + 58|0);
 HEAP16[$0>>1] = 0;
 $1 = ((($loader)) + 56|0);
 HEAP16[$1>>1] = 0;
 $2 = ((($loader)) + 84|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($loader)) + 24|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($loader)) + 22|0);
 $6 = HEAP16[$5>>1]|0;
 $7 = $6 << 16 >> 16;
 $8 = (($4) + ($7<<3)|0);
 $9 = ((($loader)) + 60|0);
 HEAP32[$9>>2] = $8;
 $10 = ((($loader)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + ($7)|0);
 $13 = ((($loader)) + 64|0);
 HEAP32[$13>>2] = $12;
 $14 = ((($loader)) + 32|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($loader)) + 20|0);
 $17 = HEAP32[$16>>2]|0;
 $sext$i = $17 << 16;
 $18 = $sext$i >> 16;
 $19 = (($15) + ($18<<1)|0);
 $20 = ((($loader)) + 68|0);
 HEAP32[$20>>2] = $19;
 $21 = ((($loader)) + 16|0);
 $22 = HEAP8[$21>>0]|0;
 $23 = ($22<<24>>24)==(0);
 if ($23) {
  $33 = ((($loader)) + 52|0);
  $34 = HEAP32[$33>>2]|0;
  $35 = ((($loader)) + 48|0);
  $36 = HEAP32[$35>>2]|0;
  $37 = (($34) + ($36<<5)|0);
  $38 = ((($loader)) + 88|0);
  HEAP32[$38>>2] = $37;
  return;
 }
 $24 = ((($loader)) + 40|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = $17 >> 16;
 $27 = (($25) + ($26<<3)|0);
 $28 = ((($loader)) + 76|0);
 HEAP32[$28>>2] = $27;
 $29 = ((($loader)) + 44|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = (($30) + ($26<<3)|0);
 $32 = ((($loader)) + 80|0);
 HEAP32[$32>>2] = $31;
 $33 = ((($loader)) + 52|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = ((($loader)) + 48|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = (($34) + ($36<<5)|0);
 $38 = ((($loader)) + 88|0);
 HEAP32[$38>>2] = $37;
 return;
}
function _FT_GlyphLoader_Add($loader) {
 $loader = $loader|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $$pre2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $n$01 = 0, $sext = 0, $sext$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($loader|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($loader)) + 56|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2&65535;
 $4 = ((($loader)) + 22|0);
 $5 = HEAP16[$4>>1]|0;
 $6 = $5 << 16 >> 16;
 $7 = ((($loader)) + 58|0);
 $8 = $2 >>> 16;
 $9 = (($8) + ($6))|0;
 $10 = $9&65535;
 HEAP16[$4>>1] = $10;
 $11 = ((($loader)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($2) + ($12))|0;
 $14 = $13&65535;
 HEAP16[$11>>1] = $14;
 $15 = ((($loader)) + 84|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($loader)) + 48|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = (($18) + ($16))|0;
 HEAP32[$17>>2] = $19;
 $20 = ($3<<16>>16)>(0);
 if ($20) {
  $23 = ((($loader)) + 68|0);
  $24 = HEAP32[$23>>2]|0;
  $sext = $2 << 16;
  $25 = $sext >> 16;
  $n$01 = 0;
  while(1) {
   $26 = (($24) + ($n$01<<1)|0);
   $27 = HEAP16[$26>>1]|0;
   $28 = $27&65535;
   $29 = (($28) + ($6))|0;
   $30 = $29&65535;
   HEAP16[$26>>1] = $30;
   $31 = (($n$01) + 1)|0;
   $exitcond = ($31|0)==($25|0);
   if ($exitcond) {
    break;
   } else {
    $n$01 = $31;
   }
  }
  $$pre = HEAP16[$4>>1]|0;
  $$pre$phiZ2D = $23;$35 = $$pre;
 } else {
  $21 = $12 >>> 16;
  $22 = $21&65535;
  $$pre2 = ((($loader)) + 68|0);
  $$pre$phiZ2D = $$pre2;$35 = $22;
 }
 HEAP16[$7>>1] = 0;
 HEAP16[$1>>1] = 0;
 HEAP32[$15>>2] = 0;
 $32 = ((($loader)) + 24|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = $35 << 16 >> 16;
 $36 = (($33) + ($34<<3)|0);
 $37 = ((($loader)) + 60|0);
 HEAP32[$37>>2] = $36;
 $38 = ((($loader)) + 28|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = (($39) + ($34)|0);
 $41 = ((($loader)) + 64|0);
 HEAP32[$41>>2] = $40;
 $42 = ((($loader)) + 32|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = HEAP32[$11>>2]|0;
 $sext$i$i = $44 << 16;
 $45 = $sext$i$i >> 16;
 $46 = (($43) + ($45<<1)|0);
 HEAP32[$$pre$phiZ2D>>2] = $46;
 $47 = ((($loader)) + 16|0);
 $48 = HEAP8[$47>>0]|0;
 $49 = ($48<<24>>24)==(0);
 if (!($49)) {
  $50 = ((($loader)) + 40|0);
  $51 = HEAP32[$50>>2]|0;
  $52 = $44 >> 16;
  $53 = (($51) + ($52<<3)|0);
  $54 = ((($loader)) + 76|0);
  HEAP32[$54>>2] = $53;
  $55 = ((($loader)) + 44|0);
  $56 = HEAP32[$55>>2]|0;
  $57 = (($56) + ($52<<3)|0);
  $58 = ((($loader)) + 80|0);
  HEAP32[$58>>2] = $57;
 }
 $59 = ((($loader)) + 52|0);
 $60 = HEAP32[$59>>2]|0;
 $61 = (($60) + ($19<<5)|0);
 $62 = ((($loader)) + 88|0);
 HEAP32[$62>>2] = $61;
 return;
}
function _FT_Stream_OpenGzip($stream,$source) {
 $stream = $stream|0;
 $source = $source|0;
 var $$$i = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $error$0$i$ph = 0, $error$i = 0, $or$cond = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp + 4|0;
 $error = sp;
 $0 = ($stream|0)!=(0|0);
 $1 = ($source|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  HEAP32[$error>>2] = 40;
  $63 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($63|0);
 }
 $2 = ((($source)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_ft_gzip_check_header($source)|0);
 HEAP32[$error>>2] = $4;
 $5 = ($4|0)==(0);
 if (!($5)) {
  $63 = HEAP32[$error>>2]|0;
  STACKTOP = sp;return ($63|0);
 }
 dest=$stream; stop=dest+40|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $6 = ((($stream)) + 28|0);
 HEAP32[$6>>2] = $3;
 $7 = (_ft_mem_qalloc($3,8276,$error)|0);
 $8 = HEAP32[$error>>2]|0;
 $9 = ($8|0)==(0);
 do {
  if ($9) {
   $10 = ((($7)) + 12|0);
   $11 = ((($7)) + 4|0);
   HEAP32[$11>>2] = $stream;
   HEAP32[$7>>2] = $source;
   $12 = HEAP32[$6>>2]|0;
   $13 = ((($7)) + 8|0);
   HEAP32[$13>>2] = $12;
   $14 = ((($7)) + 8264|0);
   $15 = ((($7)) + 8272|0);
   HEAP32[$15>>2] = $14;
   $16 = ((($7)) + 8268|0);
   HEAP32[$16>>2] = $14;
   HEAP32[$14>>2] = 0;
   $17 = (_ft_gzip_check_header($source)|0);
   $18 = ($17|0)==(0);
   if ($18) {
    $19 = ((($7)) + 4168|0);
    $20 = (_FT_Stream_Pos($source)|0);
    $21 = ((($7)) + 68|0);
    HEAP32[$21>>2] = $20;
    $22 = ((($7)) + 44|0);
    HEAP32[$22>>2] = 4440;
    $23 = ((($7)) + 48|0);
    HEAP32[$23>>2] = 4441;
    $24 = HEAP32[$2>>2]|0;
    $25 = ((($7)) + 52|0);
    HEAP32[$25>>2] = $24;
    $26 = ((($7)) + 16|0);
    HEAP32[$26>>2] = 0;
    HEAP32[$10>>2] = $19;
    $27 = (_inflateInit2__8492($10,-15)|0);
    $28 = ($27|0)==(0);
    if ($28) {
     $29 = HEAP32[$10>>2]|0;
     $30 = ($29|0)==(0|0);
     if ($30) {
      $error$0$i$ph = 3;
     } else {
      HEAP32[$error>>2] = 0;
      $31 = ((($stream)) + 12|0);
      HEAP32[$31>>2] = $7;
      break;
     }
    } else {
     $error$0$i$ph = 3;
    }
   } else {
    $error$0$i$ph = $17;
   }
   HEAP32[$error>>2] = $error$0$i$ph;
   _ft_mem_free($3,$7);
   $63 = HEAP32[$error>>2]|0;
   STACKTOP = sp;return ($63|0);
  }
 } while(0);
 $32 = ((($source)) + 8|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($source)) + 4|0);
 $35 = HEAP32[$34>>2]|0;
 $36 = (($35) + -4)|0;
 $37 = (_FT_Stream_Seek($source,$36)|0);
 $38 = ($37|0)==(0);
 if ($38) {
  $39 = (_FT_Stream_ReadULong($source,$error$i)|0);
  $40 = HEAP32[$error$i>>2]|0;
  $41 = ($40|0)!=(0);
  $$$i = $41 ? 0 : $39;
  (_FT_Stream_Seek($source,$33)|0);
  $$off = (($$$i) + -1)|0;
  $42 = ($$off>>>0)<(40959);
  if ($42) {
   $43 = (_ft_mem_alloc($3,$$$i,$error)|0);
   $44 = HEAP32[$error>>2]|0;
   $45 = ($44|0)==(0);
   do {
    if ($45) {
     $46 = (_ft_gzip_file_io($7,0,$43,$$$i)|0);
     $47 = ($46|0)==($$$i|0);
     if (!($47)) {
      (_ft_gzip_file_io($7,0,0,0)|0);
      _ft_mem_free($3,$43);
      break;
     }
     $48 = ((($7)) + 12|0);
     (_inflateEnd_8493($48)|0);
     $49 = ((($7)) + 44|0);
     HEAP32[$49>>2] = 0;
     $50 = ((($7)) + 48|0);
     HEAP32[$50>>2] = 0;
     $51 = ((($7)) + 52|0);
     HEAP32[$51>>2] = 0;
     $52 = ((($7)) + 24|0);
     HEAP32[$52>>2] = 0;
     $53 = ((($7)) + 28|0);
     HEAP32[$53>>2] = 0;
     ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;HEAP32[$7+12>>2]=0|0;HEAP32[$7+16>>2]=0|0;
     _ft_mem_free($3,$7);
     $54 = ((($stream)) + 12|0);
     HEAP32[$54>>2] = 0;
     $55 = ((($stream)) + 4|0);
     HEAP32[$55>>2] = $$$i;
     $56 = ((($stream)) + 8|0);
     HEAP32[$56>>2] = 0;
     HEAP32[$stream>>2] = $43;
     $57 = ((($stream)) + 20|0);
     HEAP32[$57>>2] = 0;
     $58 = ((($stream)) + 24|0);
     HEAP32[$58>>2] = 4442;
     $63 = HEAP32[$error>>2]|0;
     STACKTOP = sp;return ($63|0);
    }
   } while(0);
   HEAP32[$error>>2] = 0;
  }
 } else {
 }
 $59 = ((($stream)) + 4|0);
 HEAP32[$59>>2] = 2147483647;
 $60 = ((($stream)) + 8|0);
 HEAP32[$60>>2] = 0;
 HEAP32[$stream>>2] = 0;
 $61 = ((($stream)) + 20|0);
 HEAP32[$61>>2] = 4443;
 $62 = ((($stream)) + 24|0);
 HEAP32[$62>>2] = 4442;
 $63 = HEAP32[$error>>2]|0;
 STACKTOP = sp;return ($63|0);
}
function _ft_gzip_check_header($stream) {
 $stream = $stream|0;
 var $$ = 0, $$0 = 0, $$1 = 0, $$2 = 0, $$3 = 0, $$lcssa27 = 0, $$lcssa29 = 0, $$pre = 0, $$pre13 = 0, $$pre14 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $head = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $head = sp + 4|0;
 $0 = (_FT_Stream_Seek($stream,0)|0);
 HEAP32[$error>>2] = $0;
 $1 = ($0|0)==(0);
 if (!($1)) {
  $$0 = $0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = (_FT_Stream_Read($stream,$head,4)|0);
 HEAP32[$error>>2] = $2;
 $3 = ($2|0)==(0);
 if (!($3)) {
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = HEAP8[$head>>0]|0;
 $5 = ($4<<24>>24)==(31);
 if ($5) {
  $6 = ((($head)) + 1|0);
  $7 = HEAP8[$6>>0]|0;
  $8 = ($7<<24>>24)==(-117);
  if ($8) {
   $9 = ((($head)) + 2|0);
   $10 = HEAP8[$9>>0]|0;
   $11 = ($10<<24>>24)==(8);
   if ($11) {
    $12 = ((($head)) + 3|0);
    $13 = HEAP8[$12>>0]|0;
    $14 = ($13&255)>(31);
    if (!($14)) {
     $15 = (_FT_Stream_Skip($stream,6)|0);
     HEAP32[$error>>2] = $15;
     $16 = HEAP8[$12>>0]|0;
     $17 = $16 & 4;
     $18 = ($17<<24>>24)==(0);
     do {
      if ($18) {
       $26 = $16;$43 = $15;
      } else {
       $19 = (_FT_Stream_ReadUShortLE($stream,$error)|0);
       $20 = HEAP32[$error>>2]|0;
       $21 = ($20|0)==(0);
       if (!($21)) {
        $$0 = $20;
        STACKTOP = sp;return ($$0|0);
       }
       $22 = $19&65535;
       $23 = (_FT_Stream_Skip($stream,$22)|0);
       HEAP32[$error>>2] = $23;
       $24 = ($23|0)==(0);
       if ($24) {
        $$pre = HEAP8[$12>>0]|0;
        $26 = $$pre;$43 = 0;
        break;
       } else {
        $$0 = $23;
        STACKTOP = sp;return ($$0|0);
       }
      }
     } while(0);
     $25 = $26 & 8;
     $27 = ($25<<24>>24)==(0);
     do {
      if ($27) {
       $33 = $26;$44 = $43;
      } else {
       L22: while(1) {
        $28 = (_FT_Stream_ReadChar($stream,$error)|0);
        $29 = HEAP32[$error>>2]|0;
        $30 = ($29|0)==(0);
        $31 = ($28<<24>>24)==(0);
        $$ = $31 ? 3 : 0;
        $$1 = $30 ? $$ : 2;
        switch ($$1|0) {
        case 0:  {
         break;
        }
        case 3:  {
         $$lcssa29 = $29;
         label = 14;
         break L22;
         break;
        }
        default: {
         $$0 = $29;
         label = 20;
         break L22;
        }
        }
       }
       if ((label|0) == 14) {
        $$pre13 = HEAP8[$12>>0]|0;
        $33 = $$pre13;$44 = $$lcssa29;
        break;
       }
       else if ((label|0) == 20) {
        STACKTOP = sp;return ($$0|0);
       }
      }
     } while(0);
     $32 = $33 & 16;
     $34 = ($32<<24>>24)==(0);
     do {
      if ($34) {
       $40 = $33;$45 = $44;
      } else {
       L29: while(1) {
        $35 = (_FT_Stream_ReadChar($stream,$error)|0);
        $36 = HEAP32[$error>>2]|0;
        $37 = ($36|0)==(0);
        $38 = ($35<<24>>24)==(0);
        $$3 = $38 ? 5 : 0;
        $$2 = $37 ? $$3 : 2;
        switch ($$2|0) {
        case 0:  {
         break;
        }
        case 5:  {
         $$lcssa27 = $36;
         label = 17;
         break L29;
         break;
        }
        default: {
         $$0 = $36;
         label = 20;
         break L29;
        }
        }
       }
       if ((label|0) == 17) {
        $$pre14 = HEAP8[$12>>0]|0;
        $40 = $$pre14;$45 = $$lcssa27;
        break;
       }
       else if ((label|0) == 20) {
        STACKTOP = sp;return ($$0|0);
       }
      }
     } while(0);
     $39 = $40 & 2;
     $41 = ($39<<24>>24)==(0);
     if ($41) {
      $$0 = $45;
      STACKTOP = sp;return ($$0|0);
     }
     $42 = (_FT_Stream_Skip($stream,2)|0);
     HEAP32[$error>>2] = $42;
     $$0 = $42;
     STACKTOP = sp;return ($$0|0);
    }
   }
  }
 }
 HEAP32[$error>>2] = 3;
 $$0 = 3;
 STACKTOP = sp;return ($$0|0);
}
function _ft_gzip_alloc($memory,$items,$size) {
 $memory = $memory|0;
 $items = $items|0;
 $size = $size|0;
 var $0 = 0, $1 = 0, $error = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error = sp;
 $0 = Math_imul($size, $items)|0;
 $1 = (_ft_mem_alloc($memory,$0,$error)|0);
 STACKTOP = sp;return ($1|0);
}
function _ft_gzip_free($memory,$address) {
 $memory = $memory|0;
 $address = $address|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _ft_mem_free($memory,$address);
 return;
}
function _inflateInit2__8492($z,$w) {
 $z = $z|0;
 $w = $w|0;
 var $$0 = 0, $$pr$i$i = 0, $$w = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $switch$i$i = 0, $w$lobit = 0, $z$idx$val$i$i = 0, $z$idx1$val$i$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ($z|0)==(0|0);
 if ($0) {
  $$0 = -2;
  return ($$0|0);
 }
 $1 = ((($z)) + 24|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($z)) + 32|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  HEAP32[$2>>2] = 4444;
  $5 = ((($z)) + 40|0);
  HEAP32[$5>>2] = 0;
  $11 = 4444;
 } else {
  $11 = $3;
 }
 $6 = ((($z)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 if ($8) {
  HEAP32[$6>>2] = 4445;
 }
 $9 = ((($z)) + 40|0);
 $10 = HEAP32[$9>>2]|0;
 $12 = (FUNCTION_TABLE_iiii[$11 & 8191]($10,1,24)|0);
 $13 = ((($z)) + 28|0);
 HEAP32[$13>>2] = $12;
 $14 = ($12|0)==(0|0);
 if ($14) {
  $$0 = -4;
  return ($$0|0);
 }
 $15 = ((($12)) + 20|0);
 HEAP32[$15>>2] = 0;
 $16 = ((($12)) + 12|0);
 $17 = ($w|0)<(0);
 $18 = (0 - ($w))|0;
 $w$lobit = $w >>> 31;
 $$w = $17 ? $18 : $w;
 HEAP32[$16>>2] = $w$lobit;
 $19 = $$w & -8;
 $20 = ($19|0)==(8);
 if (!($20)) {
  (_inflateEnd_8493($z)|0);
  $$0 = -2;
  return ($$0|0);
 }
 $21 = ((($12)) + 16|0);
 HEAP32[$21>>2] = $$w;
 $22 = ($w|0)<(0);
 $23 = $22 ? 0 : 4446;
 $24 = 1 << $$w;
 $25 = HEAP32[$2>>2]|0;
 $26 = HEAP32[$9>>2]|0;
 $27 = (FUNCTION_TABLE_iiii[$25 & 8191]($26,1,64)|0);
 $28 = ($27|0)==(0|0);
 if ($28) {
  $62 = HEAP32[$13>>2]|0;
  $63 = ((($62)) + 20|0);
  HEAP32[$63>>2] = 0;
 } else {
  $29 = HEAP32[$2>>2]|0;
  $30 = HEAP32[$9>>2]|0;
  $31 = (FUNCTION_TABLE_iiii[$29 & 8191]($30,8,1440)|0);
  $32 = ((($27)) + 36|0);
  HEAP32[$32>>2] = $31;
  $33 = ($31|0)==(0|0);
  do {
   if ($33) {
    $34 = HEAP32[$6>>2]|0;
    $35 = HEAP32[$9>>2]|0;
    FUNCTION_TABLE_vii[$34 & 8191]($35,$27);
   } else {
    $36 = HEAP32[$2>>2]|0;
    $37 = HEAP32[$9>>2]|0;
    $38 = (FUNCTION_TABLE_iiii[$36 & 8191]($37,1,$24)|0);
    $39 = ((($27)) + 40|0);
    HEAP32[$39>>2] = $38;
    $40 = ($38|0)==(0|0);
    $41 = $38;
    if ($40) {
     $42 = HEAP32[$6>>2]|0;
     $43 = HEAP32[$9>>2]|0;
     $44 = HEAP32[$32>>2]|0;
     FUNCTION_TABLE_vii[$42 & 8191]($43,$44);
     $45 = HEAP32[$6>>2]|0;
     $46 = HEAP32[$9>>2]|0;
     FUNCTION_TABLE_vii[$45 & 8191]($46,$27);
     break;
    }
    $47 = (($38) + ($24)|0);
    $48 = ((($27)) + 44|0);
    HEAP32[$48>>2] = $47;
    $49 = ((($27)) + 56|0);
    HEAP32[$49>>2] = $23;
    HEAP32[$27>>2] = 0;
    $50 = ((($27)) + 28|0);
    HEAP32[$50>>2] = 0;
    $51 = ((($27)) + 32|0);
    HEAP32[$51>>2] = 0;
    $52 = ((($27)) + 52|0);
    HEAP32[$52>>2] = $41;
    $53 = ((($27)) + 48|0);
    HEAP32[$53>>2] = $41;
    if (!($22)) {
     $54 = (FUNCTION_TABLE_iiii[$23 & 8191](0,0,0)|0);
     $55 = ((($27)) + 60|0);
     HEAP32[$55>>2] = $54;
     $56 = ((($z)) + 48|0);
     HEAP32[$56>>2] = $54;
    }
    $59 = HEAP32[$13>>2]|0;
    $60 = ((($59)) + 20|0);
    HEAP32[$60>>2] = $27;
    $61 = ($59|0)==(0|0);
    if ($61) {
     $$0 = 0;
     return ($$0|0);
    }
    $64 = ((($z)) + 20|0);
    HEAP32[$64>>2] = 0;
    $65 = ((($z)) + 8|0);
    HEAP32[$65>>2] = 0;
    HEAP32[$1>>2] = 0;
    $66 = ((($59)) + 12|0);
    $67 = HEAP32[$66>>2]|0;
    $68 = ($67|0)!=(0);
    $69 = $68 ? 7 : 0;
    HEAP32[$59>>2] = $69;
    $70 = HEAP32[$27>>2]|0;
    $71 = $70 & -2;
    $switch$i$i = ($71|0)==(4);
    if ($switch$i$i) {
     $72 = HEAP32[$6>>2]|0;
     $73 = HEAP32[$9>>2]|0;
     $74 = ((($27)) + 12|0);
     $75 = HEAP32[$74>>2]|0;
     FUNCTION_TABLE_vii[$72 & 8191]($73,$75);
     $$pr$i$i = HEAP32[$27>>2]|0;
     $77 = $$pr$i$i;
    } else {
     $77 = $70;
    }
    $76 = ($77|0)==(6);
    if ($76) {
     $78 = ((($27)) + 4|0);
     $79 = HEAP32[$78>>2]|0;
     $z$idx$val$i$i = HEAP32[$6>>2]|0;
     $z$idx1$val$i$i = HEAP32[$9>>2]|0;
     FUNCTION_TABLE_vii[$z$idx$val$i$i & 8191]($z$idx1$val$i$i,$79);
    }
    HEAP32[$27>>2] = 0;
    $80 = ((($27)) + 28|0);
    HEAP32[$80>>2] = 0;
    $81 = ((($27)) + 32|0);
    HEAP32[$81>>2] = 0;
    $82 = ((($27)) + 40|0);
    $83 = HEAP32[$82>>2]|0;
    $84 = ((($27)) + 52|0);
    HEAP32[$84>>2] = $83;
    $85 = ((($27)) + 48|0);
    HEAP32[$85>>2] = $83;
    $86 = ((($27)) + 56|0);
    $87 = HEAP32[$86>>2]|0;
    $88 = ($87|0)==(0|0);
    if ($88) {
     $$0 = 0;
     return ($$0|0);
    }
    $89 = (FUNCTION_TABLE_iiii[$87 & 8191](0,0,0)|0);
    $90 = ((($27)) + 60|0);
    HEAP32[$90>>2] = $89;
    $91 = ((($z)) + 48|0);
    HEAP32[$91>>2] = $89;
    $$0 = 0;
    return ($$0|0);
   }
  } while(0);
  $57 = HEAP32[$13>>2]|0;
  $58 = ((($57)) + 20|0);
  HEAP32[$58>>2] = 0;
 }
 (_inflateEnd_8493($z)|0);
 $$0 = -4;
 return ($$0|0);
}
function _zcalloc_9179($opaque,$items,$size) {
 $opaque = $opaque|0;
 $items = $items|0;
 $size = $size|0;
 var $0 = 0, $1 = 0, $error$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i = sp;
 $0 = Math_imul($size, $items)|0;
 $1 = (_ft_mem_alloc($opaque,$0,$error$i)|0);
 STACKTOP = sp;return ($1|0);
}
function _zcfree_9181($opaque,$ptr) {
 $opaque = $opaque|0;
 $ptr = $ptr|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _ft_mem_free($opaque,$ptr);
 return;
}
function _inflateEnd_8493($z) {
 $z = $z|0;
 var $$0 = 0, $$pr$i$i = 0, $$pre = 0, $$pre$i = 0, $$pre$phiZ2D = 0, $$pre1 = 0, $$pre2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $switch$i$i = 0, $z$idx$val$i$i = 0, $z$idx1$i$i = 0, $z$idx1$val$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($z|0)==(0|0);
 if ($0) {
  $$0 = -2;
  return ($$0|0);
 }
 $1 = ((($z)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $$0 = -2;
  return ($$0|0);
 }
 $4 = ((($z)) + 36|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if ($6) {
  $$0 = -2;
  return ($$0|0);
 }
 $7 = ((($2)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 if ($9) {
  $$pre2 = ((($z)) + 40|0);
  $$pre$phiZ2D = $$pre2;$44 = $5;$45 = $2;
 } else {
  $10 = HEAP32[$8>>2]|0;
  $11 = $10 & -2;
  $switch$i$i = ($11|0)==(4);
  if ($switch$i$i) {
   $12 = ((($z)) + 40|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = ((($8)) + 12|0);
   $15 = HEAP32[$14>>2]|0;
   FUNCTION_TABLE_vii[$5 & 8191]($13,$15);
   $$pr$i$i = HEAP32[$8>>2]|0;
   $17 = $$pr$i$i;
  } else {
   $17 = $10;
  }
  $16 = ($17|0)==(6);
  if ($16) {
   $18 = ((($8)) + 4|0);
   $19 = HEAP32[$18>>2]|0;
   $z$idx$val$i$i = HEAP32[$4>>2]|0;
   $z$idx1$i$i = ((($z)) + 40|0);
   $z$idx1$val$i$i = HEAP32[$z$idx1$i$i>>2]|0;
   FUNCTION_TABLE_vii[$z$idx$val$i$i & 8191]($z$idx1$val$i$i,$19);
  }
  HEAP32[$8>>2] = 0;
  $20 = ((($8)) + 28|0);
  HEAP32[$20>>2] = 0;
  $21 = ((($8)) + 32|0);
  HEAP32[$21>>2] = 0;
  $22 = ((($8)) + 40|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = ((($8)) + 52|0);
  HEAP32[$24>>2] = $23;
  $25 = ((($8)) + 48|0);
  HEAP32[$25>>2] = $23;
  $26 = ((($8)) + 56|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = ($27|0)==(0|0);
  $29 = $23;
  if ($28) {
   $36 = $29;
  } else {
   $30 = (FUNCTION_TABLE_iiii[$27 & 8191](0,0,0)|0);
   $31 = ((($8)) + 60|0);
   HEAP32[$31>>2] = $30;
   $32 = ((($z)) + 48|0);
   HEAP32[$32>>2] = $30;
   $$pre$i = HEAP32[$22>>2]|0;
   $36 = $$pre$i;
  }
  $33 = HEAP32[$4>>2]|0;
  $34 = ((($z)) + 40|0);
  $35 = HEAP32[$34>>2]|0;
  FUNCTION_TABLE_vii[$33 & 8191]($35,$36);
  $37 = HEAP32[$4>>2]|0;
  $38 = HEAP32[$34>>2]|0;
  $39 = ((($8)) + 36|0);
  $40 = HEAP32[$39>>2]|0;
  FUNCTION_TABLE_vii[$37 & 8191]($38,$40);
  $41 = HEAP32[$4>>2]|0;
  $42 = HEAP32[$34>>2]|0;
  FUNCTION_TABLE_vii[$41 & 8191]($42,$8);
  $$pre = HEAP32[$4>>2]|0;
  $$pre1 = HEAP32[$1>>2]|0;
  $$pre$phiZ2D = $34;$44 = $$pre;$45 = $$pre1;
 }
 $43 = HEAP32[$$pre$phiZ2D>>2]|0;
 FUNCTION_TABLE_vii[$44 & 8191]($43,$45);
 HEAP32[$1>>2] = 0;
 $$0 = 0;
 return ($$0|0);
}
function _adler32_9183($adler,$buf,$len) {
 $adler = $adler|0;
 $buf = $buf|0;
 $len = $len|0;
 var $$0113 = 0, $$014 = 0, $$02 = 0, $$1$lcssa = 0, $$16 = 0, $$2 = 0, $$3 = 0, $$lcssa = 0, $$lcssa49 = 0, $$lcssa50 = 0, $$lcssa51 = 0, $$lcssa52 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $k$0$lcssa = 0;
 var $k$09 = 0, $k$1 = 0, $s1$0$lcssa = 0, $s1$015 = 0, $s1$1$lcssa = 0, $s1$17 = 0, $s1$2 = 0, $s1$3 = 0, $s2$0$lcssa = 0, $s2$016 = 0, $s2$1$lcssa = 0, $s2$18 = 0, $s2$2 = 0, $s2$3 = 0, $scevgep = 0, $scevgep29 = 0, $scevgep30 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $adler & 65535;
 $1 = $adler >>> 16;
 $2 = ($buf|0)==(0|0);
 if ($2) {
  $$02 = 1;
  return ($$02|0);
 }
 $3 = ($len|0)==(0);
 if ($3) {
  $s1$0$lcssa = $0;$s2$0$lcssa = $1;
 } else {
  $$0113 = $buf;$$014 = $len;$s1$015 = $0;$s2$016 = $1;
  while(1) {
   $4 = ($$014>>>0)<(5552);
   $5 = $4 ? $$014 : 5552;
   $6 = (($$014) - ($5))|0;
   $7 = ($5|0)>(15);
   if ($7) {
    $8 = ($$014>>>0)<(5552);
    $9 = $8 ? $$014 : 5552;
    $10 = (($9) + -16)|0;
    $11 = $10 & -16;
    $scevgep = ((($$0113)) + 16|0);
    $$16 = $$0113;$k$09 = $5;$s1$17 = $s1$015;$s2$18 = $s2$016;
    while(1) {
     $12 = HEAP8[$$16>>0]|0;
     $13 = $12&255;
     $14 = (($13) + ($s1$17))|0;
     $15 = (($14) + ($s2$18))|0;
     $16 = ((($$16)) + 1|0);
     $17 = HEAP8[$16>>0]|0;
     $18 = $17&255;
     $19 = (($14) + ($18))|0;
     $20 = (($15) + ($19))|0;
     $21 = ((($$16)) + 2|0);
     $22 = HEAP8[$21>>0]|0;
     $23 = $22&255;
     $24 = (($19) + ($23))|0;
     $25 = (($20) + ($24))|0;
     $26 = ((($$16)) + 3|0);
     $27 = HEAP8[$26>>0]|0;
     $28 = $27&255;
     $29 = (($24) + ($28))|0;
     $30 = (($25) + ($29))|0;
     $31 = ((($$16)) + 4|0);
     $32 = HEAP8[$31>>0]|0;
     $33 = $32&255;
     $34 = (($29) + ($33))|0;
     $35 = (($30) + ($34))|0;
     $36 = ((($$16)) + 5|0);
     $37 = HEAP8[$36>>0]|0;
     $38 = $37&255;
     $39 = (($34) + ($38))|0;
     $40 = (($35) + ($39))|0;
     $41 = ((($$16)) + 6|0);
     $42 = HEAP8[$41>>0]|0;
     $43 = $42&255;
     $44 = (($39) + ($43))|0;
     $45 = (($40) + ($44))|0;
     $46 = ((($$16)) + 7|0);
     $47 = HEAP8[$46>>0]|0;
     $48 = $47&255;
     $49 = (($44) + ($48))|0;
     $50 = (($45) + ($49))|0;
     $51 = ((($$16)) + 8|0);
     $52 = HEAP8[$51>>0]|0;
     $53 = $52&255;
     $54 = (($49) + ($53))|0;
     $55 = (($50) + ($54))|0;
     $56 = ((($$16)) + 9|0);
     $57 = HEAP8[$56>>0]|0;
     $58 = $57&255;
     $59 = (($54) + ($58))|0;
     $60 = (($55) + ($59))|0;
     $61 = ((($$16)) + 10|0);
     $62 = HEAP8[$61>>0]|0;
     $63 = $62&255;
     $64 = (($59) + ($63))|0;
     $65 = (($60) + ($64))|0;
     $66 = ((($$16)) + 11|0);
     $67 = HEAP8[$66>>0]|0;
     $68 = $67&255;
     $69 = (($64) + ($68))|0;
     $70 = (($65) + ($69))|0;
     $71 = ((($$16)) + 12|0);
     $72 = HEAP8[$71>>0]|0;
     $73 = $72&255;
     $74 = (($69) + ($73))|0;
     $75 = (($70) + ($74))|0;
     $76 = ((($$16)) + 13|0);
     $77 = HEAP8[$76>>0]|0;
     $78 = $77&255;
     $79 = (($74) + ($78))|0;
     $80 = (($75) + ($79))|0;
     $81 = ((($$16)) + 14|0);
     $82 = HEAP8[$81>>0]|0;
     $83 = $82&255;
     $84 = (($79) + ($83))|0;
     $85 = (($80) + ($84))|0;
     $86 = ((($$16)) + 15|0);
     $87 = HEAP8[$86>>0]|0;
     $88 = $87&255;
     $89 = (($84) + ($88))|0;
     $90 = (($85) + ($89))|0;
     $91 = ((($$16)) + 16|0);
     $92 = (($k$09) + -16)|0;
     $93 = ($92|0)>(15);
     if ($93) {
      $$16 = $91;$k$09 = $92;$s1$17 = $89;$s2$18 = $90;
     } else {
      $$lcssa = $89;$$lcssa49 = $90;$$lcssa50 = $92;
      break;
     }
    }
    $scevgep29 = (($scevgep) + ($11)|0);
    $$1$lcssa = $scevgep29;$k$0$lcssa = $$lcssa50;$s1$1$lcssa = $$lcssa;$s2$1$lcssa = $$lcssa49;
   } else {
    $$1$lcssa = $$0113;$k$0$lcssa = $$014;$s1$1$lcssa = $s1$015;$s2$1$lcssa = $s2$016;
   }
   $94 = ($k$0$lcssa|0)==(0);
   if ($94) {
    $$3 = $$1$lcssa;$s1$3 = $s1$1$lcssa;$s2$3 = $s2$1$lcssa;
   } else {
    $$2 = $$1$lcssa;$k$1 = $k$0$lcssa;$s1$2 = $s1$1$lcssa;$s2$2 = $s2$1$lcssa;
    while(1) {
     $95 = ((($$2)) + 1|0);
     $96 = HEAP8[$$2>>0]|0;
     $97 = $96&255;
     $98 = (($97) + ($s1$2))|0;
     $99 = (($98) + ($s2$2))|0;
     $100 = (($k$1) + -1)|0;
     $101 = ($100|0)==(0);
     if ($101) {
      $$lcssa51 = $98;$$lcssa52 = $99;
      break;
     } else {
      $$2 = $95;$k$1 = $100;$s1$2 = $98;$s2$2 = $99;
     }
    }
    $scevgep30 = (($$1$lcssa) + ($k$0$lcssa)|0);
    $$3 = $scevgep30;$s1$3 = $$lcssa51;$s2$3 = $$lcssa52;
   }
   $102 = (($s1$3>>>0) % 65521)&-1;
   $103 = (($s2$3>>>0) % 65521)&-1;
   $104 = ($$014|0)==($5|0);
   if ($104) {
    $s1$0$lcssa = $102;$s2$0$lcssa = $103;
    break;
   } else {
    $$0113 = $$3;$$014 = $6;$s1$015 = $102;$s2$016 = $103;
   }
  }
 }
 $105 = $s2$0$lcssa << 16;
 $106 = $105 | $s1$0$lcssa;
 $$02 = $106;
 return ($$02|0);
}
function _ft_gzip_file_io($zip,$pos,$buffer,$count) {
 $zip = $zip|0;
 $pos = $pos|0;
 $buffer = $buffer|0;
 $count = $count|0;
 var $$$0$i = 0, $$$02 = 0, $$$02$i = 0, $$$0212 = 0, $$$05$i = 0, $$$i$i = 0, $$$i$i$i = 0, $$04$i = 0, $$cast = 0, $$cast$i = 0, $$cast13 = 0, $$cast3$i = 0, $$in = 0, $$old1 = 0, $$phi$trans$insert = 0, $$phi$trans$insert20 = 0, $$pr$i = 0, $$pr$i$i = 0, $$pr$i$i$i = 0, $$pre = 0;
 var $$pre$i = 0, $$pre$phi$i$i$iZ2D = 0, $$pre$phi$i$iZ2D = 0, $$pre$phi25Z2D = 0, $$pre$phi26Z2D = 0, $$pre$phi27Z2D = 0, $$pre$phiZ2D = 0, $$pre22 = 0, $$pre23 = 0, $$pre24 = 0, $$pre8$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $not$ = 0, $result$1 = 0, $size$0$i$i = 0, $size$0$i$i$i = 0, $switch$i$i$i = 0, $z$idx$i$i$i = 0, $z$idx$val$i$i$i = 0, $z$idx1$i$i$i = 0, $z$idx1$val$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($zip)) + 8264|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1>>>0)>($pos>>>0);
 if ($2) {
  $3 = HEAP32[$zip>>2]|0;
  $4 = ((($zip)) + 68|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = (_FT_Stream_Seek($3,$5)|0);
  $7 = ($6|0)==(0);
  if (!($7)) {
   $result$1 = 0;
   return ($result$1|0);
  }
  $8 = ((($zip)) + 40|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)==(0|0);
  if (!($10)) {
   $11 = ((($zip)) + 32|0);
   HEAP32[$11>>2] = 0;
   $12 = ((($zip)) + 20|0);
   HEAP32[$12>>2] = 0;
   $13 = ((($zip)) + 36|0);
   HEAP32[$13>>2] = 0;
   $14 = ((($9)) + 12|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)!=(0);
   $17 = $16 ? 7 : 0;
   HEAP32[$9>>2] = $17;
   $18 = ((($9)) + 20|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = HEAP32[$19>>2]|0;
   $21 = $20 & -2;
   $switch$i$i$i = ($21|0)==(4);
   if ($switch$i$i$i) {
    $22 = ((($zip)) + 48|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = ((($zip)) + 52|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = ((($19)) + 12|0);
    $27 = HEAP32[$26>>2]|0;
    FUNCTION_TABLE_vii[$23 & 8191]($25,$27);
    $$pr$i$i$i = HEAP32[$19>>2]|0;
    $29 = $$pr$i$i$i;
   } else {
    $29 = $20;
   }
   $28 = ($29|0)==(6);
   if ($28) {
    $30 = ((($19)) + 4|0);
    $31 = HEAP32[$30>>2]|0;
    $z$idx$i$i$i = ((($zip)) + 48|0);
    $z$idx$val$i$i$i = HEAP32[$z$idx$i$i$i>>2]|0;
    $z$idx1$i$i$i = ((($zip)) + 52|0);
    $z$idx1$val$i$i$i = HEAP32[$z$idx1$i$i$i>>2]|0;
    FUNCTION_TABLE_vii[$z$idx$val$i$i$i & 8191]($z$idx1$val$i$i$i,$31);
   }
   HEAP32[$19>>2] = 0;
   $32 = ((($19)) + 28|0);
   HEAP32[$32>>2] = 0;
   $33 = ((($19)) + 32|0);
   HEAP32[$33>>2] = 0;
   $34 = ((($19)) + 40|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = ((($19)) + 52|0);
   HEAP32[$36>>2] = $35;
   $37 = ((($19)) + 48|0);
   HEAP32[$37>>2] = $35;
   $38 = ((($19)) + 56|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = ($39|0)==(0|0);
   if (!($40)) {
    $41 = (FUNCTION_TABLE_iiii[$39 & 8191](0,0,0)|0);
    $42 = ((($19)) + 60|0);
    HEAP32[$42>>2] = $41;
    $43 = ((($zip)) + 60|0);
    HEAP32[$43>>2] = $41;
   }
  }
  $44 = ((($zip)) + 16|0);
  HEAP32[$44>>2] = 0;
  $45 = ((($zip)) + 72|0);
  $46 = ((($zip)) + 12|0);
  HEAP32[$46>>2] = $45;
  $47 = ((($zip)) + 28|0);
  HEAP32[$47>>2] = 0;
  $48 = ((($zip)) + 4168|0);
  $49 = ((($zip)) + 24|0);
  HEAP32[$49>>2] = $48;
  $50 = ((($zip)) + 8264|0);
  $51 = ((($zip)) + 8272|0);
  HEAP32[$51>>2] = $50;
  $52 = ((($zip)) + 8268|0);
  HEAP32[$52>>2] = $50;
  HEAP32[$0>>2] = 0;
  $54 = 0;
 } else {
  $54 = $1;
 }
 $53 = ($54>>>0)<($pos>>>0);
 do {
  if ($53) {
   $55 = (($pos) - ($54))|0;
   $56 = ((($zip)) + 8272|0);
   $57 = HEAP32[$56>>2]|0;
   $58 = ((($zip)) + 8268|0);
   $59 = HEAP32[$58>>2]|0;
   $60 = (($57) - ($59))|0;
   $61 = ($60>>>0)<($55>>>0);
   $$$02$i = $61 ? $60 : $55;
   $$cast3$i = $59;
   $62 = (($$cast3$i) + ($$$02$i)|0);
   HEAP32[$58>>2] = $62;
   $63 = (($$$02$i) + ($54))|0;
   HEAP32[$0>>2] = $63;
   $64 = ($$$02$i|0)==($55|0);
   L19: do {
    if ($64) {
     $$in = $62;$171 = $57;
    } else {
     $65 = ((($zip)) + 12|0);
     $66 = ((($zip)) + 4168|0);
     $67 = ((($zip)) + 24|0);
     $68 = ((($zip)) + 28|0);
     $69 = ((($zip)) + 16|0);
     $70 = ((($zip)) + 72|0);
     $71 = ((($zip)) + 12|0);
     $$$05$i = $$$02$i;$$04$i = $55;
     L21: while(1) {
      $72 = (($$04$i) - ($$$05$i))|0;
      HEAP32[$58>>2] = $66;
      HEAP32[$67>>2] = $66;
      HEAP32[$68>>2] = 4096;
      L23: while(1) {
       $74 = HEAP32[$69>>2]|0;
       $75 = ($74|0)==(0);
       if ($75) {
        $76 = HEAP32[$zip>>2]|0;
        $77 = ((($76)) + 20|0);
        $78 = HEAP32[$77>>2]|0;
        $79 = ($78|0)==(0|0);
        if ($79) {
         $84 = ((($76)) + 4|0);
         $85 = HEAP32[$84>>2]|0;
         $86 = ((($76)) + 8|0);
         $87 = HEAP32[$86>>2]|0;
         $88 = (($85) - ($87))|0;
         $89 = ($88>>>0)>(4096);
         $$$i$i$i = $89 ? 4096 : $88;
         $90 = ($$$i$i$i|0)==(0);
         if ($90) {
          $result$1 = 0;
          label = 43;
          break L21;
         }
         $91 = HEAP32[$76>>2]|0;
         $92 = (($91) + ($87)|0);
         _memcpy(($70|0),($92|0),($$$i$i$i|0))|0;
         $$pre$phi$i$i$iZ2D = $86;$size$0$i$i$i = $$$i$i$i;
        } else {
         $80 = ((($76)) + 8|0);
         $81 = HEAP32[$80>>2]|0;
         $82 = (FUNCTION_TABLE_iiiii[$78 & 8191]($76,$81,$70,4096)|0);
         $83 = ($82|0)==(0);
         if ($83) {
          $result$1 = 0;
          label = 43;
          break L21;
         } else {
          $$pre$phi$i$i$iZ2D = $80;$size$0$i$i$i = $82;
         }
        }
        $93 = HEAP32[$$pre$phi$i$i$iZ2D>>2]|0;
        $94 = (($93) + ($size$0$i$i$i))|0;
        HEAP32[$$pre$phi$i$i$iZ2D>>2] = $94;
        HEAP32[$71>>2] = $70;
        HEAP32[$69>>2] = $size$0$i$i$i;
       }
       $95 = (_inflate_8494($65,0)|0);
       switch ($95|0) {
       case 1:  {
        label = 23;
        break L23;
        break;
       }
       case 0:  {
        break;
       }
       default: {
        $result$1 = 0;
        label = 43;
        break L21;
       }
       }
       $$pr$i$i = HEAP32[$68>>2]|0;
       $73 = ($$pr$i$i|0)==(0);
       if ($73) {
        label = 24;
        break;
       }
      }
      if ((label|0) == 23) {
       label = 0;
       $96 = HEAP32[$67>>2]|0;
       HEAP32[$56>>2] = $96;
       $97 = HEAP32[$58>>2]|0;
       $98 = ($96|0)==($97|0);
       $99 = $96;
       $100 = $97;
       if ($98) {
        $result$1 = 0;
        label = 43;
        break;
       } else {
        $102 = $99;$103 = $100;
       }
      }
      else if ((label|0) == 24) {
       label = 0;
       $$pre$i = HEAP32[$56>>2]|0;
       $$pre8$i = HEAP32[$58>>2]|0;
       $102 = $$pre$i;$103 = $$pre8$i;
      }
      $101 = (($102) - ($103))|0;
      $104 = ($101>>>0)<($72>>>0);
      $$$0$i = $104 ? $101 : $72;
      $$cast$i = $103;
      $105 = (($$cast$i) + ($$$0$i)|0);
      HEAP32[$58>>2] = $105;
      $106 = HEAP32[$0>>2]|0;
      $107 = (($$$0$i) + ($106))|0;
      HEAP32[$0>>2] = $107;
      $108 = ($72|0)==($$$0$i|0);
      if ($108) {
       $$in = $105;$171 = $102;
       break L19;
      } else {
       $$$05$i = $$$0$i;$$04$i = $72;
      }
     }
     if ((label|0) == 43) {
      return ($result$1|0);
     }
    }
   } while(0);
   $109 = $$in;
   $110 = ($count|0)==(0);
   if ($110) {
    $result$1 = 0;
    return ($result$1|0);
   } else {
    $$pre$phi25Z2D = $56;$$pre$phi26Z2D = $58;$$pre$phi27Z2D = $58;$$pre$phiZ2D = $56;$112 = $171;$113 = $109;
   }
  } else {
   $$old1 = ($count|0)==(0);
   if ($$old1) {
    $result$1 = 0;
    return ($result$1|0);
   } else {
    $$phi$trans$insert = ((($zip)) + 8272|0);
    $$pre = HEAP32[$$phi$trans$insert>>2]|0;
    $$phi$trans$insert20 = ((($zip)) + 8268|0);
    $$pre22 = HEAP32[$$phi$trans$insert20>>2]|0;
    $$pre$phi25Z2D = $$phi$trans$insert;$$pre$phi26Z2D = $$phi$trans$insert20;$$pre$phi27Z2D = $$phi$trans$insert20;$$pre$phiZ2D = $$phi$trans$insert;$112 = $$pre;$113 = $$pre22;
    break;
   }
  }
 } while(0);
 $111 = (($112) - ($113))|0;
 $114 = ($111>>>0)<($count>>>0);
 $$$0212 = $114 ? $111 : $count;
 $$cast13 = $113;
 _memcpy(($buffer|0),($$cast13|0),($$$0212|0))|0;
 $115 = HEAP32[$$pre$phi26Z2D>>2]|0;
 $116 = (($115) + ($$$0212)|0);
 HEAP32[$$pre$phi26Z2D>>2] = $116;
 $117 = HEAP32[$0>>2]|0;
 $118 = (($117) + ($$$0212))|0;
 HEAP32[$0>>2] = $118;
 $119 = ($$$0212|0)==($count|0);
 if ($119) {
  $result$1 = $count;
  return ($result$1|0);
 }
 $120 = (($count) - ($$$0212))|0;
 $121 = (($buffer) + ($$$0212)|0);
 $122 = ((($zip)) + 12|0);
 $123 = ((($zip)) + 4168|0);
 $124 = ((($zip)) + 24|0);
 $125 = ((($zip)) + 28|0);
 $126 = ((($zip)) + 16|0);
 $127 = ((($zip)) + 72|0);
 $128 = ((($zip)) + 12|0);
 $160 = $120;$161 = $121;$164 = $$$0212;
 L53: while(1) {
  HEAP32[$$pre$phi26Z2D>>2] = $123;
  HEAP32[$124>>2] = $123;
  HEAP32[$125>>2] = 4096;
  L55: while(1) {
   $130 = HEAP32[$126>>2]|0;
   $131 = ($130|0)==(0);
   if ($131) {
    $132 = HEAP32[$zip>>2]|0;
    $133 = ((($132)) + 20|0);
    $134 = HEAP32[$133>>2]|0;
    $135 = ($134|0)==(0|0);
    if ($135) {
     $140 = ((($132)) + 4|0);
     $141 = HEAP32[$140>>2]|0;
     $142 = ((($132)) + 8|0);
     $143 = HEAP32[$142>>2]|0;
     $144 = (($141) - ($143))|0;
     $145 = ($144>>>0)>(4096);
     $$$i$i = $145 ? 4096 : $144;
     $146 = ($$$i$i|0)==(0);
     if ($146) {
      $result$1 = $164;
      label = 43;
      break L53;
     }
     $147 = HEAP32[$132>>2]|0;
     $148 = (($147) + ($143)|0);
     _memcpy(($127|0),($148|0),($$$i$i|0))|0;
     $$pre$phi$i$iZ2D = $142;$size$0$i$i = $$$i$i;
    } else {
     $136 = ((($132)) + 8|0);
     $137 = HEAP32[$136>>2]|0;
     $138 = (FUNCTION_TABLE_iiiii[$134 & 8191]($132,$137,$127,4096)|0);
     $139 = ($138|0)==(0);
     if ($139) {
      $result$1 = $164;
      label = 43;
      break L53;
     } else {
      $$pre$phi$i$iZ2D = $136;$size$0$i$i = $138;
     }
    }
    $149 = HEAP32[$$pre$phi$i$iZ2D>>2]|0;
    $150 = (($149) + ($size$0$i$i))|0;
    HEAP32[$$pre$phi$i$iZ2D>>2] = $150;
    HEAP32[$128>>2] = $127;
    HEAP32[$126>>2] = $size$0$i$i;
   }
   $151 = (_inflate_8494($122,0)|0);
   switch ($151|0) {
   case 1:  {
    label = 40;
    break L55;
    break;
   }
   case 0:  {
    break;
   }
   default: {
    $result$1 = $164;
    label = 43;
    break L53;
   }
   }
   $$pr$i = HEAP32[$125>>2]|0;
   $129 = ($$pr$i|0)==(0);
   if ($129) {
    label = 41;
    break;
   }
  }
  if ((label|0) == 40) {
   label = 0;
   $152 = HEAP32[$124>>2]|0;
   HEAP32[$$pre$phiZ2D>>2] = $152;
   $153 = HEAP32[$$pre$phi26Z2D>>2]|0;
   $not$ = ($152|0)==($153|0);
   $154 = $152;
   $155 = $153;
   if ($not$) {
    $result$1 = $164;
    label = 43;
    break;
   } else {
    $157 = $154;$158 = $155;
   }
  }
  else if ((label|0) == 41) {
   label = 0;
   $$pre23 = HEAP32[$$pre$phi25Z2D>>2]|0;
   $$pre24 = HEAP32[$$pre$phi27Z2D>>2]|0;
   $157 = $$pre23;$158 = $$pre24;
  }
  $156 = (($157) - ($158))|0;
  $159 = ($156>>>0)<($160>>>0);
  $$$02 = $159 ? $156 : $160;
  $$cast = $158;
  _memcpy(($161|0),($$cast|0),($$$02|0))|0;
  $162 = (($161) + ($$$02)|0);
  $163 = (($$$02) + ($164))|0;
  $165 = HEAP32[$$pre$phi26Z2D>>2]|0;
  $166 = (($165) + ($$$02)|0);
  HEAP32[$$pre$phi26Z2D>>2] = $166;
  $167 = HEAP32[$0>>2]|0;
  $168 = (($167) + ($$$02))|0;
  HEAP32[$0>>2] = $168;
  $169 = (($160) - ($$$02))|0;
  $170 = ($160|0)==($$$02|0);
  if ($170) {
   $result$1 = $163;
   label = 43;
   break;
  } else {
   $160 = $169;$161 = $162;$164 = $163;
  }
 }
 if ((label|0) == 43) {
  return ($result$1|0);
 }
 return (0)|0;
}
function _inflate_8494($z,$f) {
 $z = $z|0;
 $f = $f|0;
 var $$0 = 0, $$0$i21$i = 0, $$0$i2133$i = 0, $$0$i24$i = 0, $$0$i24$i$lcssa = 0, $$0$i28$i = 0, $$0$ph = 0, $$0$ph$ph$i = 0, $$01$i$i = 0, $$01$i$i$lcssa2098 = 0, $$04$i = 0, $$04$i$lcssa1961 = 0, $$04$i$lcssa1961$lcssa = 0, $$04$i$lcssa1962 = 0, $$04$i$lcssa1963 = 0, $$04$i$lcssa1964 = 0, $$04$i$lcssa1965 = 0, $$04$i$lcssa1970 = 0, $$04$ph$i = 0, $$04$ph$ph$i = 0;
 var $$04$ph916$i = 0, $$1$lcssa$i$i = 0, $$1182$i$i = 0, $$1182$i$i$lcssa = 0, $$12$i = 0, $$13$i = 0, $$14$i = 0, $$15$lcssa$i = 0, $$2$lcssa$i$i = 0, $$213$i = 0, $$2156$i$i = 0, $$2156$i$i$lcssa = 0, $$3$i$i = 0, $$37$lcssa$i = 0, $$37469$i = 0, $$37469$i$lcssa = 0, $$4$i = 0, $$4$lcssa$i$i = 0, $$4169$i$i = 0, $$4169$i$i$lcssa = 0;
 var $$48$lcssa$i = 0, $$48490$i = 0, $$48490$i$lcssa = 0, $$5$i = 0, $$5$lcssa$i$i = 0, $$5133$i$i = 0, $$5133$i$i$lcssa = 0, $$59$lcssa$i = 0, $$59514$i = 0, $$59514$i$lcssa = 0, $$59514$i$ph = 0, $$6$i = 0, $$6$i$i = 0, $$610$lcssa$i = 0, $$7$be$i = 0, $$7$lcssa$i = 0, $$7$lcssa$i$i = 0, $$7$ph$i = 0, $$7147$i$i = 0, $$7553$i = 0;
 var $$8$lcssa$i = 0, $$8527$i = 0, $$8527$i$lcssa = 0, $$9$lcssa$i = 0, $$9540$i = 0, $$9540$i$lcssa = 0, $$lcssa = 0, $$lcssa1976 = 0, $$lcssa2058 = 0, $$lcssa2059 = 0, $$lcssa2071 = 0, $$lcssa2165 = 0, $$lcssa2166 = 0, $$lcssa2167 = 0, $$lcssa2168 = 0, $$lcssa2169 = 0, $$lcssa2212 = 0, $$lcssa2223 = 0, $$lcssa2224 = 0, $$lcssa2225 = 0;
 var $$lcssa2226 = 0, $$lcssa2227 = 0, $$lcssa2244 = 0, $$lcssa2245 = 0, $$lcssa2246 = 0, $$lcssa2247 = 0, $$lcssa2248 = 0, $$lcssa2254 = 0, $$lcssa2255 = 0, $$ph = 0, $$ph756 = 0, $$ph756$lcssa2228 = 0, $$ph756$lcssa2229 = 0, $$ph756$lcssa2230 = 0, $$ph756$lcssa2231 = 0, $$ph756$lcssa2232 = 0, $$ph756$lcssa2238 = 0, $$ph7562239 = 0, $$ph7562240 = 0, $$ph7562241 = 0;
 var $$ph7562242 = 0, $$pr$i = 0, $$pr$i1 = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$phi$iZ2D = 0, $$pre$phi906$iZ2D = 0, $$pre309$i$i = 0, $$pre310$i$i = 0, $$pre311$i$i = 0, $$pre313$i$i = 0, $$pre315$i$i = 0, $$pre748 = 0, $$pre750 = 0, $$pre751 = 0, $$pre752 = 0, $$pre753 = 0, $$pre754 = 0, $$pre896$i = 0;
 var $$pre898$i = 0, $$pre899$i = 0, $$pre900$i = 0, $$pre901$i = 0, $$pre902$i = 0, $$pre903$i = 0, $$pre904$i = 0, $$v$i = 0, $$v$i$i = 0, $$v195$i$i = 0, $$v196$i$i = 0, $$v197$i$i = 0, $$v563$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0;
 var $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0;
 var $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0;
 var $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0;
 var $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0;
 var $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0;
 var $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0;
 var $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0;
 var $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0;
 var $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0;
 var $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0;
 var $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0;
 var $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0;
 var $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0;
 var $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0;
 var $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0;
 var $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0;
 var $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0;
 var $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0;
 var $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0;
 var $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0;
 var $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0;
 var $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0;
 var $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0;
 var $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0;
 var $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0;
 var $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0;
 var $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0;
 var $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0;
 var $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0;
 var $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0;
 var $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0;
 var $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0;
 var $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0;
 var $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0;
 var $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0;
 var $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0;
 var $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0;
 var $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0;
 var $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0;
 var $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0;
 var $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0;
 var $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0;
 var $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $b$0$i$i = 0, $b$0$i$i$lcssa = 0, $b$0$i$i$lcssa2109 = 0, $b$0$i$i$lcssa2111 = 0, $b$0$i$i$lcssa2112 = 0;
 var $b$0$i$i$phi = 0, $b$0$i$i2120 = 0, $b$0$ph$i = 0, $b$0$ph$i$lcssa = 0, $b$0$ph$i$lcssa1989 = 0, $b$0$ph$i$lcssa1990 = 0, $b$0$ph$i$lcssa1991 = 0, $b$0$ph$i$lcssa1995 = 0, $b$0$ph$i$lcssa1996 = 0, $b$0$ph$i$lcssa1997 = 0, $b$0$ph$i$lcssa2000 = 0, $b$0$ph$ph$i = 0, $b$1$lcssa$i = 0, $b$1$lcssa$i$i = 0, $b$1$lcssa$i$lcssa2025 = 0, $b$1$lcssa$i$lcssa2026 = 0, $b$11$i = 0, $b$1183$i$i = 0, $b$1183$i$i$lcssa = 0, $b$12$i = 0;
 var $b$13$i = 0, $b$2$lcssa$i = 0, $b$2$lcssa$i$i = 0, $b$2$lcssa$i$lcssa = 0, $b$2157$i$i = 0, $b$2157$i$i$lcssa = 0, $b$2465$i = 0, $b$2465$i$lcssa = 0, $b$3$i$i = 0, $b$3$lcssa$i = 0, $b$3$lcssa$i$lcssa = 0, $b$3$lcssa$i$lcssa2174 = 0, $b$3486$i = 0, $b$3486$i$lcssa = 0, $b$4$lcssa$i = 0, $b$4$lcssa$i$i = 0, $b$4$lcssa$i$lcssa = 0, $b$4$lcssa$i$lcssa2204 = 0, $b$4$lcssa$i2207 = 0, $b$4$lcssa$i2208 = 0;
 var $b$4$lcssa$i2209 = 0, $b$4170$i$i = 0, $b$4170$i$i$lcssa = 0, $b$4510$i = 0, $b$4510$i$lcssa = 0, $b$4510$i$ph = 0, $b$5$lcssa$i = 0, $b$5$lcssa$i$i = 0, $b$5134$i$i = 0, $b$5134$i$i$lcssa = 0, $b$6$be$i = 0, $b$6$i$i = 0, $b$6$i$i$lcssa = 0, $b$6$lcssa$i = 0, $b$6$ph$i = 0, $b$6549$i = 0, $b$7$lcssa$i = 0, $b$7523$i = 0, $b$7523$i$lcssa = 0, $b$8$lcssa$i = 0;
 var $b$8536$i = 0, $b$8536$i$lcssa = 0, $bd3$i = 0, $bl2$i = 0, $f$0$lcssa$i$i = 0, $f$0142$i$i = 0, $f$1150$i$i = 0, $f$2$i$i = 0, $hn$i$i = 0, $hn$i19$i = 0, $i$0$i = 0, $j$0$i = 0, $k$0$i$i = 0, $k$0$i$i$lcssa = 0, $k$0$i$i$lcssa2121 = 0, $k$0$i$i$lcssa2123 = 0, $k$0$i$i$lcssa2124 = 0, $k$0$i$i$phi = 0, $k$0$ph$i = 0, $k$0$ph$i$lcssa = 0;
 var $k$0$ph$i$lcssa2004 = 0, $k$0$ph$i$lcssa2005 = 0, $k$0$ph$i$lcssa2006 = 0, $k$0$ph$i$lcssa2010 = 0, $k$0$ph$i$lcssa2011 = 0, $k$0$ph$i$lcssa2012 = 0, $k$0$ph$i$lcssa2015 = 0, $k$0$ph$ph$i = 0, $k$1$lcssa$i = 0, $k$1$lcssa$i$i = 0, $k$1$lcssa$i$lcssa2023 = 0, $k$1$lcssa$i$lcssa2024 = 0, $k$11$i = 0, $k$1184$i$i = 0, $k$1184$i$i$lcssa = 0, $k$12$i = 0, $k$13$i = 0, $k$2$lcssa$i = 0, $k$2$lcssa$i$i = 0, $k$2$lcssa$i$lcssa = 0;
 var $k$2158$i$i = 0, $k$2158$i$i$lcssa = 0, $k$2466$i = 0, $k$2466$i$lcssa = 0, $k$3$i$i = 0, $k$3$lcssa$i = 0, $k$3$lcssa$i$lcssa = 0, $k$3$lcssa$i$lcssa2173 = 0, $k$3487$i = 0, $k$3487$i$lcssa = 0, $k$4$lcssa$i = 0, $k$4$lcssa$i$i = 0, $k$4$lcssa$i$lcssa = 0, $k$4$lcssa$i$lcssa2197 = 0, $k$4$lcssa$i2200 = 0, $k$4$lcssa$i2201 = 0, $k$4$lcssa$i2202 = 0, $k$4171$i$i = 0, $k$4171$i$i$lcssa = 0, $k$4511$i = 0;
 var $k$4511$i$lcssa = 0, $k$4511$i$ph = 0, $k$5$lcssa$i = 0, $k$5$lcssa$i$i = 0, $k$5135$i$i = 0, $k$5135$i$i$lcssa = 0, $k$6$be$i = 0, $k$6$i$i = 0, $k$6$i$i$lcssa = 0, $k$6$lcssa$i = 0, $k$6$ph$i = 0, $k$6550$i = 0, $k$7$i$i = 0, $k$7$lcssa$i = 0, $k$7524$i = 0, $k$7524$i$lcssa = 0, $k$8$i$i = 0, $k$8$lcssa$i = 0, $k$8537$i = 0, $k$8537$i$lcssa = 0;
 var $m$0$i = 0, $m$0$i$i = 0, $m$0$i$i$phi = 0, $m$0$i$lcssa1948 = 0, $m$0$i$lcssa1949 = 0, $m$0$ph$i = 0, $m$0$ph$i$i = 0, $m$0$ph84$i = 0, $m$0$ph84$ph$i = 0, $m$0$ph915$i = 0, $m$1$lcssa$i$i = 0, $m$1149$i$i = 0, $m$2$i = 0, $m$3$i = 0, $m$3$i$i = 0, $m$4$i$i = 0, $m$6$i$i = 0, $m$7$i$i = 0, $n$0$$i = 0, $n$0$i = 0;
 var $n$0$i$i = 0, $n$0$i$i$lcssa = 0, $n$0$i$i$lcssa2143 = 0, $n$0$i$i$lcssa2145 = 0, $n$0$i$i$lcssa2146 = 0, $n$0$i$i$phi = 0, $n$0$i$lcssa = 0, $n$0$i$lcssa1912 = 0, $n$0$i$lcssa1913 = 0, $n$0$i$lcssa1914 = 0, $n$0$i$lcssa1915 = 0, $n$0$i$lcssa1916 = 0, $n$0$i$lcssa1920 = 0, $n$0$i$lcssa1922 = 0, $n$0$ph$i = 0, $n$0$ph$ph$i = 0, $n$0$ph913$i = 0, $n$1$lcssa$i = 0, $n$1$lcssa$i$i = 0, $n$1$lcssa$i$i$lcssa = 0;
 var $n$1$lcssa$i$lcssa2019 = 0, $n$1$lcssa$i$lcssa2020 = 0, $n$11$i = 0, $n$1186$i$i = 0, $n$12$i = 0, $n$13$i = 0, $n$2$lcssa$i = 0, $n$2$lcssa$i$i = 0, $n$2$lcssa$i$lcssa = 0, $n$2160$i$i = 0, $n$2468$i = 0, $n$3$i$i = 0, $n$3$lcssa$i = 0, $n$3$lcssa$i$lcssa = 0, $n$3$lcssa$i$lcssa2171 = 0, $n$3489$i = 0, $n$4$lcssa$i = 0, $n$4$lcssa$i$i = 0, $n$4$lcssa$i$i$lcssa = 0, $n$4$lcssa$i$lcssa = 0;
 var $n$4$lcssa$i$lcssa2183 = 0, $n$4$lcssa$i2186 = 0, $n$4$lcssa$i2187 = 0, $n$4$lcssa$i2188 = 0, $n$4173$i$i = 0, $n$4513$i = 0, $n$4513$i$ph = 0, $n$5$lcssa$i = 0, $n$5$lcssa$i$i = 0, $n$5137$i$i = 0, $n$6$be$i = 0, $n$6$i$i = 0, $n$6$i$i$lcssa = 0, $n$6$lcssa$i = 0, $n$6$ph$i = 0, $n$6552$i = 0, $n$7$i$i = 0, $n$7$lcssa$i = 0, $n$7526$i = 0, $n$8$i$i = 0;
 var $n$8$lcssa$i = 0, $n$8$lcssa$i$lcssa = 0, $n$8539$i = 0, $or$cond$i = 0, $or$cond$i$i = 0, $p$0$i = 0, $p$0$i$i = 0, $p$0$i$i$lcssa = 0, $p$0$i$i$lcssa2132 = 0, $p$0$i$i$lcssa2134 = 0, $p$0$i$i$lcssa2135 = 0, $p$0$i$i$phi = 0, $p$0$i$lcssa = 0, $p$0$i$lcssa1897 = 0, $p$0$i$lcssa1897$lcssa = 0, $p$0$i$lcssa1898 = 0, $p$0$i$lcssa1899 = 0, $p$0$i$lcssa1900 = 0, $p$0$i$lcssa1901 = 0, $p$0$i$lcssa1905 = 0;
 var $p$0$i$lcssa1906 = 0, $p$0$i$lcssa1907 = 0, $p$0$ph$i = 0, $p$0$ph$ph$i = 0, $p$0$ph912$i = 0, $p$1$lcssa$i = 0, $p$1$lcssa$i$i = 0, $p$1$lcssa$i$i$lcssa = 0, $p$1$lcssa$i$lcssa2021 = 0, $p$1$lcssa$i$lcssa2022 = 0, $p$11$i = 0, $p$1185$i$i = 0, $p$1185$i$i$lcssa = 0, $p$12$i = 0, $p$13$i = 0, $p$2$lcssa$i = 0, $p$2$lcssa$i$i = 0, $p$2$lcssa$i$lcssa = 0, $p$2159$i$i = 0, $p$2159$i$i$lcssa = 0;
 var $p$2467$i = 0, $p$2467$i$lcssa = 0, $p$3$i$i = 0, $p$3$lcssa$i = 0, $p$3$lcssa$i$lcssa = 0, $p$3$lcssa$i$lcssa2172 = 0, $p$3488$i = 0, $p$3488$i$lcssa = 0, $p$4$lcssa$i = 0, $p$4$lcssa$i$i = 0, $p$4$lcssa$i$i$lcssa = 0, $p$4$lcssa$i$lcssa = 0, $p$4$lcssa$i$lcssa2190 = 0, $p$4$lcssa$i2193 = 0, $p$4$lcssa$i2194 = 0, $p$4$lcssa$i2195 = 0, $p$4172$i$i = 0, $p$4172$i$i$lcssa = 0, $p$4512$i = 0, $p$4512$i$lcssa = 0;
 var $p$4512$i$ph = 0, $p$5$lcssa$i = 0, $p$5$lcssa$i$i = 0, $p$5136$i$i = 0, $p$5136$i$i$lcssa = 0, $p$6$be$i = 0, $p$6$i$i = 0, $p$6$i$i$lcssa = 0, $p$6$lcssa$i = 0, $p$6$ph$i = 0, $p$6551$i = 0, $p$7$i$i = 0, $p$7$lcssa$i = 0, $p$7525$i = 0, $p$7525$i$lcssa = 0, $p$8$i$i = 0, $p$8$lcssa$i = 0, $p$8$lcssa$i$lcssa = 0, $p$8538$i = 0, $p$8538$i$lcssa = 0;
 var $q$0$i = 0, $q$0$i$i = 0, $q$0$i$i$lcssa = 0, $q$0$i$i$lcssa2154 = 0, $q$0$i$i$lcssa2156 = 0, $q$0$i$i$lcssa2158 = 0, $q$0$i$i$lcssa2160 = 0, $q$0$i$i$lcssa2161 = 0, $q$0$i$i$lcssa2162 = 0, $q$0$i$i$lcssa2163 = 0, $q$0$i$i$lcssa2164 = 0, $q$0$i$i$phi = 0, $q$0$i$lcssa = 0, $q$0$i$lcssa1924 = 0, $q$0$i$lcssa1924$lcssa = 0, $q$0$i$lcssa1924$lcssa1943 = 0, $q$0$i$lcssa1924$lcssa1944 = 0, $q$0$i$lcssa1925 = 0, $q$0$i$lcssa1925$lcssa = 0, $q$0$i$lcssa1925$lcssa1945 = 0;
 var $q$0$i$lcssa1926 = 0, $q$0$i$lcssa1926$lcssa = 0, $q$0$i$lcssa1927 = 0, $q$0$i$lcssa1928 = 0, $q$0$i$lcssa1932 = 0, $q$0$i$lcssa1933 = 0, $q$0$i1937 = 0, $q$0$i1937$lcssa = 0, $q$0$i1937$lcssa2170 = 0, $q$0$i1938 = 0, $q$0$i1939 = 0, $q$0$i1939$lcssa = 0, $q$0$i1939$lcssa2219 = 0, $q$0$i1939$lcssa2220 = 0, $q$0$i1940 = 0, $q$0$i1940$lcssa = 0, $q$0$i1940$lcssa2176 = 0, $q$0$i19402179 = 0, $q$0$i19402180 = 0, $q$0$i19402181 = 0;
 var $q$0$i1941 = 0, $q$0$i1941$lcssa = 0, $q$0$ph$i = 0, $q$0$ph$ph$i = 0, $q$0$ph914$i = 0, $q$1$lcssa$i$i = 0, $q$1148$i$i = 0, $q$127$i = 0, $q$2$i = 0, $q$2$i$lcssa = 0, $q$22$i$i = 0, $q$3$i = 0, $q$3$i$i = 0, $q$3$i$i$lcssa = 0, $q$4$i = 0, $q$4$i$i = 0, $q$5$i = 0, $q$55$i$i = 0, $q$6$i$i = 0, $q$6$i$i$lcssa = 0;
 var $q$7$i$i = 0, $q$8$i$i = 0, $r$0 = 0, $r$0$i20$i = 0, $r$0$lcssa1884 = 0, $r$0$lcssa1885 = 0, $r$0$lcssa1886 = 0, $r$0$lcssa1887 = 0, $r$0$lcssa1889 = 0, $r$0$lcssa1891 = 0, $r$0$lcssa1892 = 0, $r$0$lcssa1893 = 0, $r$0$ph = 0, $r$1 = 0, $r$1$i$i = 0, $r$10 = 0, $r$2 = 0, $r$3 = 0, $r$4 = 0, $r$5 = 0;
 var $r$7 = 0, $r$8 = 0, $r$9 = 0, $storemerge$i = 0, $switch$i = 0, $t$1$i = 0, $td5$i = 0, $tl4$i = 0, $z$idx$i = 0, $z$idx$val$i = 0, $z$idx$val$i3 = 0, $z$idx1$val$i = 0, $z$idx14$val$i = 0, $z$idx15$val$i = 0, $z$idx16$val$i = 0, $z$idx17$i = 0, $z$idx17$val$i = 0, $z$idx18$i = 0, $z$idx18$val$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $hn$i19$i = sp + 20|0;
 $hn$i$i = sp + 16|0;
 $bl2$i = sp + 12|0;
 $bd3$i = sp + 8|0;
 $tl4$i = sp + 4|0;
 $td5$i = sp;
 $0 = ($z|0)==(0|0);
 if ($0) {
  $$0 = -2;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ((($z)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $$0 = -2;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = HEAP32[$z>>2]|0;
 $5 = ($4|0)==(0|0);
 if ($5) {
  $$0 = -2;
  STACKTOP = sp;return ($$0|0);
 }
 $6 = ($f|0)==(4);
 $7 = $6 ? -5 : 0;
 $8 = ((($z)) + 4|0);
 $9 = ((($z)) + 8|0);
 $10 = ((($z)) + 24|0);
 $z$idx17$i = ((($z)) + 32|0);
 $z$idx18$i = ((($z)) + 40|0);
 $z$idx$i = ((($z)) + 36|0);
 $11 = ((($z)) + 48|0);
 $$ph756 = $2;$r$0$ph = -5;
 L10: while(1) {
  $12 = ((($$ph756)) + 4|0);
  $13 = ((($$ph756)) + 16|0);
  $14 = ((($$ph756)) + 4|0);
  $r$0 = $r$0$ph;
  L12: while(1) {
   $15 = HEAP32[$$ph756>>2]|0;
   L14: do {
    switch ($15|0) {
    case 2:  {
     $$lcssa2244 = $$ph756;$$ph756$lcssa2228 = $$ph756;$r$0$lcssa1884 = $r$0;
     label = 22;
     break L10;
     break;
    }
    case 3:  {
     $$lcssa2245 = $$ph756;$$ph756$lcssa2229 = $$ph756;$r$0$lcssa1885 = $r$0;
     label = 25;
     break L10;
     break;
    }
    case 4:  {
     $$lcssa2246 = $$ph756;$$ph756$lcssa2230 = $$ph756;$r$0$lcssa1886 = $r$0;
     label = 28;
     break L10;
     break;
    }
    case 5:  {
     $$lcssa2247 = $$ph756;$$ph756$lcssa2231 = $$ph756;$r$0$lcssa1887 = $r$0;
     label = 31;
     break L10;
     break;
    }
    case 6:  {
     $$lcssa2248 = $$ph756;$$ph756$lcssa2232 = $$ph756;
     label = 34;
     break L10;
     break;
    }
    case 12:  {
     $$0$ph = 1;
     label = 272;
     break L10;
     break;
    }
    case 13:  {
     label = 273;
     break L10;
     break;
    }
    case 7:  {
     $r$0$lcssa1889 = $r$0;
     label = 35;
     break L12;
     break;
    }
    case 8:  {
     $952 = $$ph756;$r$7 = $r$0;
     label = 262;
     break L12;
     break;
    }
    case 9:  {
     $r$0$lcssa1891 = $r$0;
     label = 9;
     break L12;
     break;
    }
    case 10:  {
     $r$0$lcssa1892 = $r$0;
     label = 8;
     break L12;
     break;
    }
    case 11:  {
     $r$0$lcssa1893 = $r$0;
     label = 7;
     break L12;
     break;
    }
    case 0:  {
     $16 = HEAP32[$8>>2]|0;
     $17 = ($16|0)==(0);
     if ($17) {
      $$0$ph = $r$0;
      label = 272;
      break L10;
     }
     $18 = (($16) + -1)|0;
     HEAP32[$8>>2] = $18;
     $19 = HEAP32[$9>>2]|0;
     $20 = (($19) + 1)|0;
     HEAP32[$9>>2] = $20;
     $21 = HEAP32[$z>>2]|0;
     $22 = ((($21)) + 1|0);
     HEAP32[$z>>2] = $22;
     $23 = HEAP8[$21>>0]|0;
     $24 = $23&255;
     HEAP32[$12>>2] = $24;
     $25 = $24 & 15;
     $26 = ($25|0)==(8);
     if (!($26)) {
      HEAP32[$$ph756>>2] = 13;
      HEAP32[$10>>2] = 612432;
      HEAP32[$12>>2] = 5;
      $r$0 = $7;
      continue L12;
     }
     $27 = $24 >>> 4;
     $28 = (($27) + 8)|0;
     $29 = HEAP32[$13>>2]|0;
     $30 = ($28>>>0)>($29>>>0);
     if ($30) {
      HEAP32[$$ph756>>2] = 13;
      HEAP32[$10>>2] = 612459;
      HEAP32[$12>>2] = 5;
      $r$0 = $7;
      continue L12;
     } else {
      HEAP32[$$ph756>>2] = 1;
      $r$1 = $7;
      break L14;
     }
     break;
    }
    case 1:  {
     $r$1 = $r$0;
     break;
    }
    default: {
     $$0 = -2;
     label = 274;
     break L10;
    }
    }
   } while(0);
   $31 = HEAP32[$8>>2]|0;
   $32 = ($31|0)==(0);
   if ($32) {
    $$0$ph = $r$1;
    label = 272;
    break L10;
   }
   $33 = (($31) + -1)|0;
   HEAP32[$8>>2] = $33;
   $34 = HEAP32[$9>>2]|0;
   $35 = (($34) + 1)|0;
   HEAP32[$9>>2] = $35;
   $36 = HEAP32[$z>>2]|0;
   $37 = ((($36)) + 1|0);
   HEAP32[$z>>2] = $37;
   $38 = HEAP8[$36>>0]|0;
   $39 = $38&255;
   $40 = HEAP32[$14>>2]|0;
   $41 = $40 << 8;
   $42 = $41 | $39;
   $43 = (($42>>>0) % 31)&-1;
   $44 = ($43|0)==(0);
   if (!($44)) {
    HEAP32[$$ph756>>2] = 13;
    HEAP32[$10>>2] = 612409;
    HEAP32[$14>>2] = 5;
    $r$0 = $7;
    continue;
   }
   $45 = $39 & 32;
   $46 = ($45|0)==(0);
   if (!($46)) {
    $$lcssa = $33;$$lcssa2254 = $$ph756;$$ph756$lcssa2238 = $$ph756;
    label = 21;
    break L10;
   }
   HEAP32[$$ph756>>2] = 7;
   $r$0 = $7;
  }
  do {
   if ((label|0) == 7) {
    label = 0;
    $$pre750 = HEAP32[$8>>2]|0;
    $982 = $$pre750;$991 = $$ph756;$r$10 = $r$0$lcssa1893;
   }
   else if ((label|0) == 8) {
    label = 0;
    $$pre748 = HEAP32[$8>>2]|0;
    $968 = $$pre748;$978 = $$ph756;$r$9 = $r$0$lcssa1892;
    label = 266;
   }
   else if ((label|0) == 9) {
    label = 0;
    $$pre = HEAP32[$8>>2]|0;
    $954 = $$pre;$964 = $$ph756;$r$8 = $r$0$lcssa1891;
    label = 264;
   }
   else if ((label|0) == 35) {
    label = 0;
    $101 = ((($$ph756)) + 20|0);
    $102 = HEAP32[$101>>2]|0;
    $103 = HEAP32[$z>>2]|0;
    $104 = HEAP32[$8>>2]|0;
    $105 = ((($102)) + 32|0);
    $106 = HEAP32[$105>>2]|0;
    $107 = ((($102)) + 28|0);
    $108 = HEAP32[$107>>2]|0;
    $109 = ((($102)) + 52|0);
    $110 = HEAP32[$109>>2]|0;
    $111 = ((($102)) + 48|0);
    $112 = HEAP32[$111>>2]|0;
    $113 = ($110>>>0)<($112>>>0);
    if ($113) {
     $114 = $112;
     $115 = $110;
     $116 = $115 ^ -1;
     $117 = (($114) + ($116))|0;
     $$pre904$i = ((($102)) + 44|0);
     $$pre$phi$iZ2D = $$pre904$i;$$pre$phi906$iZ2D = $$pre904$i;$m$0$ph$i = $117;
    } else {
     $127 = ((($102)) + 44|0);
     $128 = HEAP32[$127>>2]|0;
     $129 = $110;
     $130 = (($128) - ($129))|0;
     $$pre$phi$iZ2D = $127;$$pre$phi906$iZ2D = $127;$m$0$ph$i = $130;
    }
    $118 = ((($102)) + 40|0);
    $119 = ((($102)) + 4|0);
    $120 = ((($102)) + 24|0);
    $121 = ((($102)) + 4|0);
    $122 = ((($102)) + 8|0);
    $123 = ((($102)) + 20|0);
    $124 = ((($102)) + 12|0);
    $125 = ((($102)) + 36|0);
    $126 = ((($102)) + 16|0);
    $$0$ph$ph$i = 0;$$04$ph$ph$i = $r$0$lcssa1889;$b$0$ph$ph$i = $106;$k$0$ph$ph$i = $108;$m$0$ph84$ph$i = $m$0$ph$i;$n$0$ph$ph$i = $104;$p$0$ph$ph$i = $103;$q$0$ph$ph$i = $110;
    L35: while(1) {
     $$04$ph$i = $$04$ph$ph$i;$b$0$ph$i = $b$0$ph$ph$i;$k$0$ph$i = $k$0$ph$ph$i;$m$0$ph84$i = $m$0$ph84$ph$i;$n$0$ph$i = $n$0$ph$ph$i;$p$0$ph$i = $p$0$ph$ph$i;$q$0$ph$i = $q$0$ph$ph$i;
     L37: while(1) {
      $$04$ph916$i = $$04$ph$i;$m$0$ph915$i = $m$0$ph84$i;$n$0$ph913$i = $n$0$ph$i;$p$0$ph912$i = $p$0$ph$i;$q$0$ph914$i = $q$0$ph$i;
      L39: while(1) {
       $$pr$i = HEAP32[$102>>2]|0;
       $$04$i = $$04$ph916$i;$131 = $$pr$i;$m$0$i = $m$0$ph915$i;$n$0$i = $n$0$ph913$i;$p$0$i = $p$0$ph912$i;$q$0$i = $q$0$ph914$i;
       while(1) {
        switch ($131|0) {
        case 7:  {
         $$14$i = $$04$i;$b$12$i = $b$0$ph$i;$k$12$i = $k$0$ph$i;$n$12$i = $n$0$i;$p$12$i = $p$0$i;$q$4$i = $q$0$i;
         label = 244;
         break L35;
         break;
        }
        case 8:  {
         $b$13$i = $b$0$ph$i;$k$13$i = $k$0$ph$i;$n$13$i = $n$0$i;$p$13$i = $p$0$i;$q$5$i = $q$0$i;
         label = 247;
         break L35;
         break;
        }
        case 9:  {
         $b$0$ph$i$lcssa1995 = $b$0$ph$i;$k$0$ph$i$lcssa2010 = $k$0$ph$i;$n$0$i$lcssa1920 = $n$0$i;$p$0$i$lcssa1905 = $p$0$i;$q$0$i$lcssa1932 = $q$0$i;
         label = 248;
         break L35;
         break;
        }
        case 3:  {
         $$04$i$lcssa1963 = $$04$i;$b$0$ph$i$lcssa1989 = $b$0$ph$i;$k$0$ph$i$lcssa2004 = $k$0$ph$i;$n$0$i$lcssa1914 = $n$0$i;$p$0$i$lcssa1899 = $p$0$i;$q$0$i$lcssa1926 = $q$0$i;
         label = 43;
         break L37;
         break;
        }
        case 4:  {
         $$04$i$lcssa1964 = $$04$i;$b$0$ph$i$lcssa1990 = $b$0$ph$i;$k$0$ph$i$lcssa2005 = $k$0$ph$i;$n$0$i$lcssa1915 = $n$0$i;$p$0$i$lcssa1900 = $p$0$i;$q$0$i$lcssa1927 = $q$0$i;
         label = 91;
         break L37;
         break;
        }
        case 5:  {
         $$04$i$lcssa1965 = $$04$i;$b$0$ph$i$lcssa1991 = $b$0$ph$i;$k$0$ph$i$lcssa2006 = $k$0$ph$i;$n$0$i$lcssa1916 = $n$0$i;$p$0$i$lcssa1901 = $p$0$i;$q$0$i$lcssa1928 = $q$0$i;
         label = 110;
         break L37;
         break;
        }
        case 6:  {
         $$13$i = $$04$i;$$5$i = $$0$ph$ph$i;$b$11$i = $b$0$ph$i;$k$11$i = $k$0$ph$i;$n$11$i = $n$0$i;$p$11$i = $p$0$i;$q$0$i1938 = $q$0$i;
         break L37;
         break;
        }
        case 0:  {
         $$04$i$lcssa1961 = $$04$i;$m$0$i$lcssa1948 = $m$0$i;$n$0$i$lcssa1912 = $n$0$i;$p$0$i$lcssa1897 = $p$0$i;$q$0$i$lcssa1924 = $q$0$i;
         break L39;
         break;
        }
        case 1:  {
         $$04$i$lcssa1962 = $$04$i;$m$0$i$lcssa1949 = $m$0$i;$n$0$i$lcssa1913 = $n$0$i;$p$0$i$lcssa1898 = $p$0$i;$q$0$i$lcssa1925 = $q$0$i;
         label = 44;
         break L39;
         break;
        }
        case 2:  {
         break;
        }
        default: {
         $b$0$ph$i$lcssa = $b$0$ph$i;$k$0$ph$i$lcssa = $k$0$ph$i;$n$0$i$lcssa = $n$0$i;$p$0$i$lcssa = $p$0$i;$q$0$i$lcssa = $q$0$i;
         label = 249;
         break L35;
        }
        }
        $211 = ($n$0$i|0)==(0);
        if ($211) {
         $$04$i$lcssa1970 = $$04$i;$b$0$ph$i$lcssa1996 = $b$0$ph$i;$k$0$ph$i$lcssa2011 = $k$0$ph$i;$p$0$i$lcssa1906 = $p$0$i;$q$0$i$lcssa1933 = $q$0$i;
         label = 66;
         break L35;
        }
        $218 = ($m$0$i|0)==(0);
        do {
         if ($218) {
          $219 = HEAP32[$$pre$phi$iZ2D>>2]|0;
          $220 = ($q$0$i|0)==($219|0);
          if ($220) {
           $221 = HEAP32[$111>>2]|0;
           $222 = HEAP32[$118>>2]|0;
           $223 = ($221|0)==($222|0);
           if ($223) {
            $q$127$i = $q$0$i;
           } else {
            $224 = ($222>>>0)<($221>>>0);
            $225 = $221;
            $226 = $222;
            $227 = (($225) + -1)|0;
            $228 = $q$0$i;
            $$v563$i = $224 ? $227 : $228;
            $229 = (($$v563$i) - ($226))|0;
            $230 = ($$v563$i|0)==($226|0);
            if ($230) {
             $q$127$i = $222;
            } else {
             $m$3$i = $229;$q$3$i = $222;
             break;
            }
           }
          } else {
           $q$127$i = $q$0$i;
          }
          HEAP32[$109>>2] = $q$127$i;
          $231 = (_inflate_flush($102,$z,$$04$i)|0);
          $232 = HEAP32[$109>>2]|0;
          $233 = HEAP32[$111>>2]|0;
          $234 = ($232>>>0)<($233>>>0);
          if ($234) {
           $235 = $233;
           $236 = $232;
           $237 = $236 ^ -1;
           $238 = (($235) + ($237))|0;
           $$pre903$i = HEAP32[$$pre$phi$iZ2D>>2]|0;
           $244 = $$pre903$i;$997 = $238;
          } else {
           $239 = HEAP32[$$pre$phi906$iZ2D>>2]|0;
           $240 = $232;
           $241 = (($239) - ($240))|0;
           $242 = $239;
           $244 = $242;$997 = $241;
          }
          $243 = ($232|0)==($244|0);
          if ($243) {
           $245 = HEAP32[$118>>2]|0;
           $246 = ($233|0)==($245|0);
           if ($246) {
            $m$2$i = $997;$q$2$i = $232;
           } else {
            $247 = ($245>>>0)<($233>>>0);
            $248 = $233;
            $249 = $245;
            $250 = (($248) + -1)|0;
            $251 = $232;
            $$v$i = $247 ? $250 : $251;
            $252 = (($$v$i) - ($249))|0;
            $m$2$i = $252;$q$2$i = $245;
           }
          } else {
           $m$2$i = $997;$q$2$i = $232;
          }
          $253 = ($m$2$i|0)==(0);
          if ($253) {
           $$lcssa1976 = $231;$b$0$ph$i$lcssa1997 = $b$0$ph$i;$k$0$ph$i$lcssa2012 = $k$0$ph$i;$n$0$i$lcssa1922 = $n$0$i;$p$0$i$lcssa1907 = $p$0$i;$q$2$i$lcssa = $q$2$i;
           label = 78;
           break L35;
          } else {
           $m$3$i = $m$2$i;$q$3$i = $q$2$i;
          }
         } else {
          $m$3$i = $m$0$i;$q$3$i = $q$0$i;
         }
        } while(0);
        $260 = HEAP32[$119>>2]|0;
        $261 = ($260>>>0)>($n$0$i>>>0);
        $n$0$$i = $261 ? $n$0$i : $260;
        $262 = ($n$0$$i>>>0)>($m$3$i>>>0);
        $t$1$i = $262 ? $m$3$i : $n$0$$i;
        _memcpy(($q$3$i|0),($p$0$i|0),($t$1$i|0))|0;
        $263 = (($p$0$i) + ($t$1$i)|0);
        $264 = (($n$0$i) - ($t$1$i))|0;
        $265 = (($q$3$i) + ($t$1$i)|0);
        $266 = (($m$3$i) - ($t$1$i))|0;
        $267 = HEAP32[$119>>2]|0;
        $268 = (($267) - ($t$1$i))|0;
        HEAP32[$119>>2] = $268;
        $269 = ($267|0)==($t$1$i|0);
        if (!($269)) {
         $$04$ph916$i = 0;$m$0$ph915$i = $266;$n$0$ph913$i = $264;$p$0$ph912$i = $263;$q$0$ph914$i = $265;
         continue L39;
        }
        $270 = HEAP32[$120>>2]|0;
        $271 = ($270|0)!=(0);
        $272 = $271 ? 7 : 0;
        HEAP32[$102>>2] = $272;
        $$04$i = 0;$131 = $272;$m$0$i = $266;$n$0$i = $264;$p$0$i = $263;$q$0$i = $265;
       }
      }
      if ((label|0) == 44) {
       label = 0;
       $133 = ($k$0$ph$i>>>0)<(32);
       if ($133) {
        $$37469$i = $$04$i$lcssa1962;$b$2465$i = $b$0$ph$i;$k$2466$i = $k$0$ph$i;$n$2468$i = $n$0$i$lcssa1913;$p$2467$i = $p$0$i$lcssa1898;
        while(1) {
         $181 = ($n$2468$i|0)==(0);
         if ($181) {
          $$37469$i$lcssa = $$37469$i;$b$2465$i$lcssa = $b$2465$i;$k$2466$i$lcssa = $k$2466$i;$p$2467$i$lcssa = $p$2467$i;$q$0$i$lcssa1925$lcssa = $q$0$i$lcssa1925;
          label = 58;
          break L35;
         }
         $182 = (($n$2468$i) + -1)|0;
         $183 = ((($p$2467$i)) + 1|0);
         $184 = HEAP8[$p$2467$i>>0]|0;
         $185 = $184&255;
         $186 = $185 << $k$2466$i;
         $187 = $186 | $b$2465$i;
         $188 = (($k$2466$i) + 8)|0;
         $189 = ($188>>>0)<(32);
         if ($189) {
          $$37469$i = 0;$b$2465$i = $187;$k$2466$i = $188;$n$2468$i = $182;$p$2467$i = $183;
         } else {
          $$37$lcssa$i = 0;$b$2$lcssa$i = $187;$k$2$lcssa$i = $188;$n$2$lcssa$i = $182;$p$2$lcssa$i = $183;
          break;
         }
        }
       } else {
        $$37$lcssa$i = $$04$i$lcssa1962;$b$2$lcssa$i = $b$0$ph$i;$k$2$lcssa$i = $k$0$ph$i;$n$2$lcssa$i = $n$0$i$lcssa1913;$p$2$lcssa$i = $p$0$i$lcssa1898;
       }
       $196 = $b$2$lcssa$i >>> 16;
       $197 = $196 ^ 65535;
       $198 = $b$2$lcssa$i & 65535;
       $199 = ($197|0)==($198|0);
       if (!($199)) {
        $b$2$lcssa$i$lcssa = $b$2$lcssa$i;$k$2$lcssa$i$lcssa = $k$2$lcssa$i;$n$2$lcssa$i$lcssa = $n$2$lcssa$i;$p$2$lcssa$i$lcssa = $p$2$lcssa$i;$q$0$i$lcssa1925$lcssa1945 = $q$0$i$lcssa1925;
        label = 60;
        break L35;
       }
       HEAP32[$119>>2] = $197;
       $206 = ($196|0)==(65535);
       if ($206) {
        $207 = HEAP32[$120>>2]|0;
        $208 = ($207|0)!=(0);
        $209 = $208 ? 7 : 0;
        $210 = $209;
       } else {
        $210 = 2;
       }
       HEAP32[$102>>2] = $210;
       $$04$ph$i = $$37$lcssa$i;$b$0$ph$i = 0;$k$0$ph$i = 0;$m$0$ph84$i = $m$0$i$lcssa1949;$n$0$ph$i = $n$2$lcssa$i;$p$0$ph$i = $p$2$lcssa$i;$q$0$ph$i = $q$0$i$lcssa1925;
       continue;
      }
      $134 = ($k$0$ph$i>>>0)<(3);
      if ($134) {
       $135 = ($n$0$i$lcssa1912|0)==(0);
       if ($135) {
        $$04$i$lcssa1961$lcssa = $$04$i$lcssa1961;$b$0$ph$i$lcssa2000 = $b$0$ph$i;$k$0$ph$i$lcssa2015 = $k$0$ph$i;$p$0$i$lcssa1897$lcssa = $p$0$i$lcssa1897;$q$0$i$lcssa1924$lcssa = $q$0$i$lcssa1924;
        label = 47;
        break L35;
       }
       $142 = (($k$0$ph$i) + 8)|0;
       $143 = HEAP8[$p$0$i$lcssa1897>>0]|0;
       $144 = $143&255;
       $145 = $144 << $k$0$ph$i;
       $146 = $145 | $b$0$ph$i;
       $147 = ((($p$0$i$lcssa1897)) + 1|0);
       $148 = (($n$0$i$lcssa1912) + -1)|0;
       $$15$lcssa$i = 0;$b$1$lcssa$i = $146;$k$1$lcssa$i = $142;$n$1$lcssa$i = $148;$p$1$lcssa$i = $147;
      } else {
       $$15$lcssa$i = $$04$i$lcssa1961;$b$1$lcssa$i = $b$0$ph$i;$k$1$lcssa$i = $k$0$ph$i;$n$1$lcssa$i = $n$0$i$lcssa1912;$p$1$lcssa$i = $p$0$i$lcssa1897;
      }
      $149 = $b$1$lcssa$i & 1;
      HEAP32[$120>>2] = $149;
      $150 = $b$1$lcssa$i >>> 1;
      $151 = $150 & 3;
      switch ($151|0) {
      case 3:  {
       $b$1$lcssa$i$lcssa2025 = $b$1$lcssa$i;$k$1$lcssa$i$lcssa2023 = $k$1$lcssa$i;$n$1$lcssa$i$lcssa2019 = $n$1$lcssa$i;$p$1$lcssa$i$lcssa2021 = $p$1$lcssa$i;$q$0$i$lcssa1924$lcssa1943 = $q$0$i$lcssa1924;
       label = 55;
       break L35;
       break;
      }
      case 0:  {
       $152 = $b$1$lcssa$i >>> 3;
       $153 = (($k$1$lcssa$i) + -3)|0;
       $154 = $153 & 7;
       $155 = $152 >>> $154;
       $156 = (($153) - ($154))|0;
       HEAP32[$102>>2] = 1;
       $$04$ph$i = $$15$lcssa$i;$b$0$ph$i = $155;$k$0$ph$i = $156;$m$0$ph84$i = $m$0$i$lcssa1948;$n$0$ph$i = $n$1$lcssa$i;$p$0$ph$i = $p$1$lcssa$i;$q$0$ph$i = $q$0$i$lcssa1924;
       continue L37;
       break;
      }
      case 1:  {
       $z$idx17$val$i = HEAP32[$z$idx17$i>>2]|0;
       $z$idx18$val$i = HEAP32[$z$idx18$i>>2]|0;
       $157 = (FUNCTION_TABLE_iiii[$z$idx17$val$i & 8191]($z$idx18$val$i,1,28)|0);
       $158 = ($157|0)==(0|0);
       if ($158) {
        $b$1$lcssa$i$lcssa2026 = $b$1$lcssa$i;$k$1$lcssa$i$lcssa2024 = $k$1$lcssa$i;$n$1$lcssa$i$lcssa2020 = $n$1$lcssa$i;$p$1$lcssa$i$lcssa2022 = $p$1$lcssa$i;$q$0$i$lcssa1924$lcssa1944 = $q$0$i$lcssa1924;
        label = 53;
        break L35;
       }
       HEAP32[$157>>2] = 0;
       $159 = ((($157)) + 16|0);
       HEAP8[$159>>0] = 9;
       $160 = ((($157)) + 17|0);
       HEAP8[$160>>0] = 5;
       $161 = ((($157)) + 20|0);
       HEAP32[$161>>2] = 62628;
       $162 = ((($157)) + 24|0);
       HEAP32[$162>>2] = 66724;
       HEAP32[$121>>2] = $157;
       $163 = $b$1$lcssa$i >>> 3;
       $164 = (($k$1$lcssa$i) + -3)|0;
       HEAP32[$102>>2] = 6;
       $$04$ph$i = $$15$lcssa$i;$b$0$ph$i = $163;$k$0$ph$i = $164;$m$0$ph84$i = $m$0$i$lcssa1948;$n$0$ph$i = $n$1$lcssa$i;$p$0$ph$i = $p$1$lcssa$i;$q$0$ph$i = $q$0$i$lcssa1924;
       continue L37;
       break;
      }
      case 2:  {
       $171 = $b$1$lcssa$i >>> 3;
       $172 = (($k$1$lcssa$i) + -3)|0;
       HEAP32[$102>>2] = 3;
       $$04$ph$i = $$15$lcssa$i;$b$0$ph$i = $171;$k$0$ph$i = $172;$m$0$ph84$i = $m$0$i$lcssa1948;$n$0$ph$i = $n$1$lcssa$i;$p$0$ph$i = $p$1$lcssa$i;$q$0$ph$i = $q$0$i$lcssa1924;
       continue L37;
       break;
      }
      default: {
       label = 40;
       break L10;
      }
      }
     }
     if ((label|0) == 43) {
      label = 0;
      $132 = ($k$0$ph$i$lcssa2004>>>0)<(14);
      if ($132) {
       $$48490$i = $$04$i$lcssa1963;$b$3486$i = $b$0$ph$i$lcssa1989;$k$3487$i = $k$0$ph$i$lcssa2004;$n$3489$i = $n$0$i$lcssa1914;$p$3488$i = $p$0$i$lcssa1899;
       while(1) {
        $273 = ($n$3489$i|0)==(0);
        if ($273) {
         $$48490$i$lcssa = $$48490$i;$b$3486$i$lcssa = $b$3486$i;$k$3487$i$lcssa = $k$3487$i;$p$3488$i$lcssa = $p$3488$i;$q$0$i$lcssa1926$lcssa = $q$0$i$lcssa1926;
         label = 83;
         break L35;
        }
        $274 = (($n$3489$i) + -1)|0;
        $275 = ((($p$3488$i)) + 1|0);
        $276 = HEAP8[$p$3488$i>>0]|0;
        $277 = $276&255;
        $278 = $277 << $k$3487$i;
        $279 = $278 | $b$3486$i;
        $280 = (($k$3487$i) + 8)|0;
        $281 = ($280>>>0)<(14);
        if ($281) {
         $$48490$i = 0;$b$3486$i = $279;$k$3487$i = $280;$n$3489$i = $274;$p$3488$i = $275;
        } else {
         $$48$lcssa$i = 0;$b$3$lcssa$i = $279;$k$3$lcssa$i = $280;$n$3$lcssa$i = $274;$p$3$lcssa$i = $275;$q$0$i1937 = $q$0$i$lcssa1926;
         break;
        }
       }
      } else {
       $$48$lcssa$i = $$04$i$lcssa1963;$b$3$lcssa$i = $b$0$ph$i$lcssa1989;$k$3$lcssa$i = $k$0$ph$i$lcssa2004;$n$3$lcssa$i = $n$0$i$lcssa1914;$p$3$lcssa$i = $p$0$i$lcssa1899;$q$0$i1937 = $q$0$i$lcssa1926;
      }
      $288 = $b$3$lcssa$i & 16383;
      HEAP32[$119>>2] = $288;
      $289 = $b$3$lcssa$i & 31;
      $290 = ($289>>>0)>(29);
      if ($290) {
       $b$3$lcssa$i$lcssa = $b$3$lcssa$i;$k$3$lcssa$i$lcssa = $k$3$lcssa$i;$n$3$lcssa$i$lcssa = $n$3$lcssa$i;$p$3$lcssa$i$lcssa = $p$3$lcssa$i;$q$0$i1937$lcssa = $q$0$i1937;
       label = 86;
       break;
      }
      $291 = $b$3$lcssa$i >>> 5;
      $292 = $291 & 31;
      $293 = ($292>>>0)>(29);
      if ($293) {
       $b$3$lcssa$i$lcssa = $b$3$lcssa$i;$k$3$lcssa$i$lcssa = $k$3$lcssa$i;$n$3$lcssa$i$lcssa = $n$3$lcssa$i;$p$3$lcssa$i$lcssa = $p$3$lcssa$i;$q$0$i1937$lcssa = $q$0$i1937;
       label = 86;
       break;
      }
      $300 = (($289) + 258)|0;
      $301 = (($300) + ($292))|0;
      $302 = HEAP32[$z$idx17$i>>2]|0;
      $303 = HEAP32[$z$idx18$i>>2]|0;
      $304 = (FUNCTION_TABLE_iiii[$302 & 8191]($303,$301,4)|0);
      HEAP32[$124>>2] = $304;
      $305 = ($304|0)==(0|0);
      if ($305) {
       $b$3$lcssa$i$lcssa2174 = $b$3$lcssa$i;$k$3$lcssa$i$lcssa2173 = $k$3$lcssa$i;$n$3$lcssa$i$lcssa2171 = $n$3$lcssa$i;$p$3$lcssa$i$lcssa2172 = $p$3$lcssa$i;$q$0$i1937$lcssa2170 = $q$0$i1937;
       label = 88;
       break;
      }
      $312 = $b$3$lcssa$i >>> 14;
      $313 = (($k$3$lcssa$i) + -14)|0;
      HEAP32[$122>>2] = 0;
      HEAP32[$102>>2] = 4;
      $$59514$i$ph = $$48$lcssa$i;$$ph = 0;$b$4510$i$ph = $312;$k$4511$i$ph = $313;$n$4513$i$ph = $n$3$lcssa$i;$p$4512$i$ph = $p$3$lcssa$i;$q$0$i1941 = $q$0$i1937;
      label = 90;
     }
     else if ((label|0) == 91) {
      label = 0;
      $$pre$i = HEAP32[$122>>2]|0;
      $314 = HEAP32[$119>>2]|0;
      $315 = $314 >>> 10;
      $316 = (($315) + 4)|0;
      $317 = ($$pre$i>>>0)<($316>>>0);
      if ($317) {
       $$59514$i$ph = $$04$i$lcssa1964;$$ph = $$pre$i;$b$4510$i$ph = $b$0$ph$i$lcssa1990;$k$4511$i$ph = $k$0$ph$i$lcssa2005;$n$4513$i$ph = $n$0$i$lcssa1915;$p$4512$i$ph = $p$0$i$lcssa1900;$q$0$i1941 = $q$0$i$lcssa1927;
       label = 90;
      } else {
       $$59$lcssa$i = $$04$i$lcssa1964;$319 = $$pre$i;$b$4$lcssa$i = $b$0$ph$i$lcssa1990;$k$4$lcssa$i = $k$0$ph$i$lcssa2005;$n$4$lcssa$i = $n$0$i$lcssa1915;$p$4$lcssa$i = $p$0$i$lcssa1900;$q$0$i1940 = $q$0$i$lcssa1927;
       label = 92;
      }
     }
     else if ((label|0) == 110) {
      label = 0;
      $$pre896$i = HEAP32[$122>>2]|0;
      $$7$ph$i = $$04$i$lcssa1965;$387 = $$pre896$i;$b$6$ph$i = $b$0$ph$i$lcssa1991;$k$6$ph$i = $k$0$ph$i$lcssa2006;$n$6$ph$i = $n$0$i$lcssa1916;$p$6$ph$i = $p$0$i$lcssa1901;$q$0$i1939 = $q$0$i$lcssa1928;
      label = 111;
     }
     if ((label|0) == 90) {
      label = 0;
      $$59514$i = $$59514$i$ph;$338 = $$ph;$b$4510$i = $b$4510$i$ph;$k$4511$i = $k$4511$i$ph;$n$4513$i = $n$4513$i$ph;$p$4512$i = $p$4512$i$ph;
      while(1) {
       $321 = ($k$4511$i>>>0)<(3);
       if ($321) {
        $322 = ($n$4513$i|0)==(0);
        if ($322) {
         $$59514$i$lcssa = $$59514$i;$b$4510$i$lcssa = $b$4510$i;$k$4511$i$lcssa = $k$4511$i;$p$4512$i$lcssa = $p$4512$i;$q$0$i1941$lcssa = $q$0$i1941;
         label = 96;
         break L35;
        }
        $329 = (($k$4511$i) + 8)|0;
        $330 = HEAP8[$p$4512$i>>0]|0;
        $331 = $330&255;
        $332 = $331 << $k$4511$i;
        $333 = $332 | $b$4510$i;
        $334 = ((($p$4512$i)) + 1|0);
        $335 = (($n$4513$i) + -1)|0;
        $$610$lcssa$i = 0;$b$5$lcssa$i = $333;$k$5$lcssa$i = $329;$n$5$lcssa$i = $335;$p$5$lcssa$i = $334;
       } else {
        $$610$lcssa$i = $$59514$i;$b$5$lcssa$i = $b$4510$i;$k$5$lcssa$i = $k$4511$i;$n$5$lcssa$i = $n$4513$i;$p$5$lcssa$i = $p$4512$i;
       }
       $336 = $b$5$lcssa$i & 7;
       $337 = (($338) + 1)|0;
       HEAP32[$122>>2] = $337;
       $339 = (66980 + ($338<<2)|0);
       $340 = HEAP32[$339>>2]|0;
       $341 = HEAP32[$124>>2]|0;
       $342 = (($341) + ($340<<2)|0);
       HEAP32[$342>>2] = $336;
       $343 = $b$5$lcssa$i >>> 3;
       $344 = (($k$5$lcssa$i) + -3)|0;
       $345 = HEAP32[$122>>2]|0;
       $346 = HEAP32[$119>>2]|0;
       $347 = $346 >>> 10;
       $348 = (($347) + 4)|0;
       $349 = ($345>>>0)<($348>>>0);
       if ($349) {
        $$59514$i = $$610$lcssa$i;$338 = $345;$b$4510$i = $343;$k$4511$i = $344;$n$4513$i = $n$5$lcssa$i;$p$4512$i = $p$5$lcssa$i;
       } else {
        $$59$lcssa$i = $$610$lcssa$i;$319 = $345;$b$4$lcssa$i = $343;$k$4$lcssa$i = $344;$n$4$lcssa$i = $n$5$lcssa$i;$p$4$lcssa$i = $p$5$lcssa$i;$q$0$i1940 = $q$0$i1941;
        label = 92;
        break;
       }
      }
     }
     if ((label|0) == 92) {
      label = 0;
      $318 = ($319>>>0)<(19);
      if ($318) {
       $320 = HEAP32[$124>>2]|0;
       $351 = $319;
       while(1) {
        $350 = (($351) + 1)|0;
        HEAP32[$122>>2] = $350;
        $352 = (66980 + ($351<<2)|0);
        $353 = HEAP32[$352>>2]|0;
        $354 = (($320) + ($353<<2)|0);
        HEAP32[$354>>2] = 0;
        $355 = HEAP32[$122>>2]|0;
        $356 = ($355>>>0)<(19);
        if ($356) {
         $351 = $355;
        } else {
         break;
        }
       }
      }
      HEAP32[$126>>2] = 7;
      $357 = HEAP32[$124>>2]|0;
      $358 = HEAP32[$125>>2]|0;
      HEAP32[$hn$i$i>>2] = 0;
      $359 = HEAP32[$z$idx17$i>>2]|0;
      $360 = HEAP32[$z$idx18$i>>2]|0;
      $361 = (FUNCTION_TABLE_iiii[$359 & 8191]($360,19,4)|0);
      $362 = ($361|0)==(0|0);
      if ($362) {
       $b$4$lcssa$i$lcssa = $b$4$lcssa$i;$k$4$lcssa$i$lcssa = $k$4$lcssa$i;$n$4$lcssa$i$lcssa = $n$4$lcssa$i;$p$4$lcssa$i$lcssa = $p$4$lcssa$i;$q$0$i1940$lcssa = $q$0$i1940;
       label = 101;
       break;
      }
      $363 = (_huft_build($357,19,19,0,0,$123,$126,$358,$hn$i$i,$361)|0);
      switch ($363|0) {
      case -5:  {
       $$lcssa2212 = $361;$b$4$lcssa$i$lcssa2204 = $b$4$lcssa$i;$k$4$lcssa$i$lcssa2197 = $k$4$lcssa$i;$n$4$lcssa$i$lcssa2183 = $n$4$lcssa$i;$p$4$lcssa$i$lcssa2190 = $p$4$lcssa$i;$q$0$i1940$lcssa2176 = $q$0$i1940;
       label = 104;
       break L35;
       break;
      }
      case -3:  {
       $368 = $361;$b$4$lcssa$i2209 = $b$4$lcssa$i;$k$4$lcssa$i2202 = $k$4$lcssa$i;$n$4$lcssa$i2188 = $n$4$lcssa$i;$p$4$lcssa$i2195 = $p$4$lcssa$i;$q$0$i19402181 = $q$0$i1940;$storemerge$i = 595601;
       label = 105;
       break L35;
       break;
      }
      default: {
      }
      }
      $364 = HEAP32[$126>>2]|0;
      $365 = ($364|0)==(0);
      if ($365) {
       $$lcssa2212 = $361;$b$4$lcssa$i$lcssa2204 = $b$4$lcssa$i;$k$4$lcssa$i$lcssa2197 = $k$4$lcssa$i;$n$4$lcssa$i$lcssa2183 = $n$4$lcssa$i;$p$4$lcssa$i$lcssa2190 = $p$4$lcssa$i;$q$0$i1940$lcssa2176 = $q$0$i1940;
       label = 104;
       break;
      }
      $369 = HEAP32[$z$idx$i>>2]|0;
      $370 = HEAP32[$z$idx18$i>>2]|0;
      FUNCTION_TABLE_vii[$369 & 8191]($370,$361);
      switch ($363|0) {
      case -3:  {
       $b$4$lcssa$i2208 = $b$4$lcssa$i;$k$4$lcssa$i2201 = $k$4$lcssa$i;$n$4$lcssa$i2187 = $n$4$lcssa$i;$p$4$lcssa$i2194 = $p$4$lcssa$i;$q$0$i19402180 = $q$0$i1940;
       label = 107;
       break L35;
       break;
      }
      case 0:  {
       break;
      }
      default: {
       $$0$i28$i = $363;$b$4$lcssa$i2207 = $b$4$lcssa$i;$k$4$lcssa$i2200 = $k$4$lcssa$i;$n$4$lcssa$i2186 = $n$4$lcssa$i;$p$4$lcssa$i2193 = $p$4$lcssa$i;$q$0$i19402179 = $q$0$i1940;
       label = 108;
       break L35;
      }
      }
      HEAP32[$122>>2] = 0;
      HEAP32[$102>>2] = 5;
      $$7$ph$i = $$59$lcssa$i;$387 = 0;$b$6$ph$i = $b$4$lcssa$i;$k$6$ph$i = $k$4$lcssa$i;$n$6$ph$i = $n$4$lcssa$i;$p$6$ph$i = $p$4$lcssa$i;$q$0$i1939 = $q$0$i1940;
      label = 111;
     }
     if ((label|0) == 111) {
      label = 0;
      $380 = HEAP32[$121>>2]|0;
      $381 = $380 & 31;
      $382 = (($381) + 258)|0;
      $383 = $380 >>> 5;
      $384 = $383 & 31;
      $385 = (($382) + ($384))|0;
      $386 = ($387>>>0)<($385>>>0);
      if ($386) {
       $$7553$i = $$7$ph$i;$419 = $387;$461 = $380;$b$6549$i = $b$6$ph$i;$k$6550$i = $k$6$ph$i;$n$6552$i = $n$6$ph$i;$p$6551$i = $p$6$ph$i;
       while(1) {
        $388 = HEAP32[$126>>2]|0;
        $389 = ($k$6550$i>>>0)<($388>>>0);
        if ($389) {
         $$8527$i = $$7553$i;$b$7523$i = $b$6549$i;$k$7524$i = $k$6550$i;$n$7526$i = $n$6552$i;$p$7525$i = $p$6551$i;
         while(1) {
          $390 = ($n$7526$i|0)==(0);
          if ($390) {
           $$8527$i$lcssa = $$8527$i;$b$7523$i$lcssa = $b$7523$i;$k$7524$i$lcssa = $k$7524$i;$p$7525$i$lcssa = $p$7525$i;$q$0$i1939$lcssa = $q$0$i1939;
           label = 116;
           break L35;
          }
          $391 = (($n$7526$i) + -1)|0;
          $392 = ((($p$7525$i)) + 1|0);
          $393 = HEAP8[$p$7525$i>>0]|0;
          $394 = $393&255;
          $395 = $394 << $k$7524$i;
          $396 = $395 | $b$7523$i;
          $397 = (($k$7524$i) + 8)|0;
          $398 = ($397>>>0)<($388>>>0);
          if ($398) {
           $$8527$i = 0;$b$7523$i = $396;$k$7524$i = $397;$n$7526$i = $391;$p$7525$i = $392;
          } else {
           $$8$lcssa$i = 0;$b$7$lcssa$i = $396;$k$7$lcssa$i = $397;$n$7$lcssa$i = $391;$p$7$lcssa$i = $392;
           break;
          }
         }
        } else {
         $$8$lcssa$i = $$7553$i;$b$7$lcssa$i = $b$6549$i;$k$7$lcssa$i = $k$6550$i;$n$7$lcssa$i = $n$6552$i;$p$7$lcssa$i = $p$6551$i;
        }
        $405 = HEAP32[$123>>2]|0;
        $406 = (67056 + ($388<<2)|0);
        $407 = HEAP32[$406>>2]|0;
        $408 = $407 & $b$7$lcssa$i;
        $409 = (($405) + ($408<<3)|0);
        $410 = ((($409)) + 1|0);
        $411 = HEAP8[$410>>0]|0;
        $412 = $411&255;
        $413 = (((($405) + ($408<<3)|0)) + 4|0);
        $414 = HEAP32[$413>>2]|0;
        $415 = ($414>>>0)<(16);
        if ($415) {
         $416 = $b$7$lcssa$i >>> $412;
         $417 = (($k$7$lcssa$i) - ($412))|0;
         $418 = (($419) + 1)|0;
         HEAP32[$122>>2] = $418;
         $420 = HEAP32[$124>>2]|0;
         $421 = (($420) + ($419<<2)|0);
         HEAP32[$421>>2] = $414;
         $$pre899$i = HEAP32[$122>>2]|0;
         $$7$be$i = $$8$lcssa$i;$429 = $$pre899$i;$998 = $420;$b$6$be$i = $416;$k$6$be$i = $417;$n$6$be$i = $n$7$lcssa$i;$p$6$be$i = $p$7$lcssa$i;
        } else {
         $430 = ($414|0)==(18);
         $431 = (($414) + -14)|0;
         $432 = $430 ? 7 : $431;
         $433 = $430 ? 11 : 3;
         $434 = (($432) + ($412))|0;
         $435 = ($k$7$lcssa$i>>>0)<($434>>>0);
         if ($435) {
          $$9540$i = $$8$lcssa$i;$b$8536$i = $b$7$lcssa$i;$k$8537$i = $k$7$lcssa$i;$n$8539$i = $n$7$lcssa$i;$p$8538$i = $p$7$lcssa$i;
          while(1) {
           $436 = ($n$8539$i|0)==(0);
           if ($436) {
            $$9540$i$lcssa = $$9540$i;$b$8536$i$lcssa = $b$8536$i;$k$8537$i$lcssa = $k$8537$i;$p$8538$i$lcssa = $p$8538$i;$q$0$i1939$lcssa2219 = $q$0$i1939;
            label = 123;
            break L35;
           }
           $437 = (($n$8539$i) + -1)|0;
           $438 = ((($p$8538$i)) + 1|0);
           $439 = HEAP8[$p$8538$i>>0]|0;
           $440 = $439&255;
           $441 = $440 << $k$8537$i;
           $442 = $441 | $b$8536$i;
           $443 = (($k$8537$i) + 8)|0;
           $444 = ($443>>>0)<($434>>>0);
           if ($444) {
            $$9540$i = 0;$b$8536$i = $442;$k$8537$i = $443;$n$8539$i = $437;$p$8538$i = $438;
           } else {
            $$9$lcssa$i = 0;$b$8$lcssa$i = $442;$k$8$lcssa$i = $443;$n$8$lcssa$i = $437;$p$8$lcssa$i = $438;
            break;
           }
          }
         } else {
          $$9$lcssa$i = $$8$lcssa$i;$b$8$lcssa$i = $b$7$lcssa$i;$k$8$lcssa$i = $k$7$lcssa$i;$n$8$lcssa$i = $n$7$lcssa$i;$p$8$lcssa$i = $p$7$lcssa$i;
         }
         $451 = $b$8$lcssa$i >>> $412;
         $452 = (($k$8$lcssa$i) - ($412))|0;
         $453 = (67056 + ($432<<2)|0);
         $454 = HEAP32[$453>>2]|0;
         $455 = $454 & $451;
         $456 = (($455) + ($433))|0;
         $457 = $451 >>> $432;
         $458 = (($452) - ($432))|0;
         $459 = (($456) + ($419))|0;
         $460 = $461 & 31;
         $462 = (($460) + 258)|0;
         $463 = $461 >>> 5;
         $464 = $463 & 31;
         $465 = (($462) + ($464))|0;
         $466 = ($459>>>0)>($465>>>0);
         if ($466) {
          $$lcssa2058 = $457;$$lcssa2059 = $458;$n$8$lcssa$i$lcssa = $n$8$lcssa$i;$p$8$lcssa$i$lcssa = $p$8$lcssa$i;$q$0$i1939$lcssa2220 = $q$0$i1939;
          label = 126;
          break L35;
         }
         $467 = ($414|0)==(16);
         $468 = ($419|0)==(0);
         $or$cond$i = $468 & $467;
         if ($or$cond$i) {
          $$lcssa2058 = $457;$$lcssa2059 = $458;$n$8$lcssa$i$lcssa = $n$8$lcssa$i;$p$8$lcssa$i$lcssa = $p$8$lcssa$i;$q$0$i1939$lcssa2220 = $q$0$i1939;
          label = 126;
          break L35;
         }
         if ($467) {
          $478 = (($419) + -1)|0;
          $479 = HEAP32[$124>>2]|0;
          $480 = (($479) + ($478<<2)|0);
          $481 = HEAP32[$480>>2]|0;
          $484 = $479;$485 = $481;
         } else {
          $$pre898$i = HEAP32[$124>>2]|0;
          $484 = $$pre898$i;$485 = 0;
         }
         $i$0$i = $419;$j$0$i = $456;
         while(1) {
          $482 = (($i$0$i) + 1)|0;
          $483 = (($484) + ($i$0$i<<2)|0);
          HEAP32[$483>>2] = $485;
          $486 = (($j$0$i) + -1)|0;
          $487 = ($486|0)==(0);
          if ($487) {
           break;
          } else {
           $i$0$i = $482;$j$0$i = $486;
          }
         }
         $488 = (($433) + ($419))|0;
         $489 = (($488) + ($455))|0;
         HEAP32[$122>>2] = $489;
         $$7$be$i = $$9$lcssa$i;$429 = $489;$998 = $484;$b$6$be$i = $457;$k$6$be$i = $458;$n$6$be$i = $n$8$lcssa$i;$p$6$be$i = $p$8$lcssa$i;
        }
        $422 = HEAP32[$121>>2]|0;
        $423 = $422 & 31;
        $424 = (($423) + 258)|0;
        $425 = $422 >>> 5;
        $426 = $425 & 31;
        $427 = (($424) + ($426))|0;
        $428 = ($429>>>0)<($427>>>0);
        if ($428) {
         $$7553$i = $$7$be$i;$419 = $429;$461 = $422;$b$6549$i = $b$6$be$i;$k$6550$i = $k$6$be$i;$n$6552$i = $n$6$be$i;$p$6551$i = $p$6$be$i;
        } else {
         $$7$lcssa$i = $$7$be$i;$491 = $422;$501 = $998;$b$6$lcssa$i = $b$6$be$i;$k$6$lcssa$i = $k$6$be$i;$n$6$lcssa$i = $n$6$be$i;$p$6$lcssa$i = $p$6$be$i;
         break;
        }
       }
      } else {
       $$pre900$i = HEAP32[$124>>2]|0;
       $$7$lcssa$i = $$7$ph$i;$491 = $380;$501 = $$pre900$i;$b$6$lcssa$i = $b$6$ph$i;$k$6$lcssa$i = $k$6$ph$i;$n$6$lcssa$i = $n$6$ph$i;$p$6$lcssa$i = $p$6$ph$i;
      }
      HEAP32[$123>>2] = 0;
      HEAP32[$bl2$i>>2] = 9;
      HEAP32[$bd3$i>>2] = 6;
      $490 = $491 & 31;
      $492 = (($490) + 257)|0;
      $493 = $491 >>> 5;
      $494 = $493 & 31;
      $495 = (($494) + 1)|0;
      $496 = HEAP32[$125>>2]|0;
      HEAP32[$hn$i19$i>>2] = 0;
      $497 = HEAP32[$z$idx17$i>>2]|0;
      $498 = HEAP32[$z$idx18$i>>2]|0;
      $499 = (FUNCTION_TABLE_iiii[$497 & 8191]($498,288,4)|0);
      $500 = ($499|0)==(0|0);
      L143: do {
       if ($500) {
        $$0$i2133$i = -4;
        label = 150;
       } else {
        $502 = (_huft_build($501,$492,257,67124,67248,$tl4$i,$bl2$i,$496,$hn$i19$i,$499)|0);
        L146: do {
         switch ($502|0) {
         case 0:  {
          $503 = HEAP32[$bl2$i>>2]|0;
          $504 = ($503|0)==(0);
          if ($504) {
           label = 138;
          } else {
           $507 = (($501) + ($492<<2)|0);
           $508 = (_huft_build($507,$495,0,67372,33620,$td5$i,$bd3$i,$496,$hn$i19$i,$499)|0);
           switch ($508|0) {
           case 0:  {
            $509 = HEAP32[$bd3$i>>2]|0;
            $510 = ($509|0)==(0);
            $511 = ($490|0)!=(0);
            $or$cond$i$i = $511 & $510;
            if ($or$cond$i$i) {
             label = 144;
            } else {
             $514 = HEAP32[$z$idx$i>>2]|0;
             $515 = HEAP32[$z$idx18$i>>2]|0;
             FUNCTION_TABLE_vii[$514 & 8191]($515,$499);
             $530 = $503;$533 = $509;
             break L146;
            }
            break;
           }
           case -3:  {
            HEAP32[$10>>2] = 595743;
            $r$1$i$i = -3;
            break;
           }
           case -5:  {
            HEAP32[$10>>2] = 595772;
            $r$1$i$i = -3;
            break;
           }
           case -4:  {
            $r$1$i$i = -4;
            break;
           }
           default: {
            label = 144;
           }
           }
           if ((label|0) == 144) {
            label = 0;
            HEAP32[$10>>2] = 595797;
            $r$1$i$i = -3;
           }
           $512 = HEAP32[$z$idx$i>>2]|0;
           $513 = HEAP32[$z$idx18$i>>2]|0;
           FUNCTION_TABLE_vii[$512 & 8191]($513,$499);
           $$0$i21$i = $r$1$i$i;
           label = 147;
          }
          break;
         }
         case -3:  {
          HEAP32[$10>>2] = 595677;
          $r$0$i20$i = -3;
          label = 139;
          break;
         }
         case -4:  {
          $r$0$i20$i = -4;
          label = 139;
          break;
         }
         default: {
          label = 138;
         }
         }
        } while(0);
        if ((label|0) == 138) {
         label = 0;
         HEAP32[$10>>2] = 595712;
         $r$0$i20$i = -3;
         label = 139;
        }
        if ((label|0) == 139) {
         label = 0;
         $505 = HEAP32[$z$idx$i>>2]|0;
         $506 = HEAP32[$z$idx18$i>>2]|0;
         FUNCTION_TABLE_vii[$505 & 8191]($506,$499);
         $$0$i21$i = $r$0$i20$i;
         label = 147;
        }
        L162: do {
         if ((label|0) == 147) {
          label = 0;
          switch ($$0$i21$i|0) {
          case 0:  {
           $$pre901$i = HEAP32[$bl2$i>>2]|0;
           $$pre902$i = HEAP32[$bd3$i>>2]|0;
           $530 = $$pre901$i;$533 = $$pre902$i;
           break L162;
           break;
          }
          case -3:  {
           $516 = HEAP32[$z$idx$i>>2]|0;
           $517 = HEAP32[$z$idx18$i>>2]|0;
           $518 = HEAP32[$124>>2]|0;
           FUNCTION_TABLE_vii[$516 & 8191]($517,$518);
           HEAP32[$102>>2] = 9;
           $$0$i2133$i = -3;
           label = 150;
           break L143;
           break;
          }
          default: {
           $$0$i2133$i = -4;
           label = 150;
           break L143;
          }
          }
         }
        } while(0);
        $525 = HEAP32[$tl4$i>>2]|0;
        $526 = HEAP32[$td5$i>>2]|0;
        $z$idx15$val$i = HEAP32[$z$idx17$i>>2]|0;
        $z$idx16$val$i = HEAP32[$z$idx18$i>>2]|0;
        $527 = (FUNCTION_TABLE_iiii[$z$idx15$val$i & 8191]($z$idx16$val$i,1,28)|0);
        $528 = ($527|0)==(0|0);
        if ($528) {
         HEAP32[$105>>2] = $b$6$lcssa$i;
         HEAP32[$107>>2] = $k$6$lcssa$i;
         HEAP32[$8>>2] = $n$6$lcssa$i;
         $537 = HEAP32[$z>>2]|0;
         $538 = $p$6$lcssa$i;
         $539 = (($538) - ($537))|0;
         $540 = HEAP32[$9>>2]|0;
         $541 = (($539) + ($540))|0;
         HEAP32[$9>>2] = $541;
         HEAP32[$z>>2] = $p$6$lcssa$i;
         HEAP32[$109>>2] = $q$0$i1939;
         $542 = (_inflate_flush($102,$z,-4)|0);
         $$12$i = -4;$$213$i = 0;$$4$i = $542;
         break;
        } else {
         HEAP32[$527>>2] = 0;
         $529 = $530&255;
         $531 = ((($527)) + 16|0);
         HEAP8[$531>>0] = $529;
         $532 = $533&255;
         $534 = ((($527)) + 17|0);
         HEAP8[$534>>0] = $532;
         $535 = ((($527)) + 20|0);
         HEAP32[$535>>2] = $525;
         $536 = ((($527)) + 24|0);
         HEAP32[$536>>2] = $526;
         HEAP32[$121>>2] = $527;
         $$12$i = $$7$lcssa$i;$$213$i = 1;$$4$i = $$0$ph$ph$i;
         break;
        }
       }
      } while(0);
      if ((label|0) == 150) {
       label = 0;
       HEAP32[$105>>2] = $b$6$lcssa$i;
       HEAP32[$107>>2] = $k$6$lcssa$i;
       HEAP32[$8>>2] = $n$6$lcssa$i;
       $519 = HEAP32[$z>>2]|0;
       $520 = $p$6$lcssa$i;
       $521 = (($520) - ($519))|0;
       $522 = HEAP32[$9>>2]|0;
       $523 = (($521) + ($522))|0;
       HEAP32[$9>>2] = $523;
       HEAP32[$z>>2] = $p$6$lcssa$i;
       HEAP32[$109>>2] = $q$0$i1939;
       $524 = (_inflate_flush($102,$z,$$0$i2133$i)|0);
       $$12$i = $$0$i2133$i;$$213$i = 0;$$4$i = $524;
      }
      if (!($$213$i)) {
       $$6$i = $$4$i;
       break;
      }
      $543 = HEAP32[$z$idx$i>>2]|0;
      $544 = HEAP32[$z$idx18$i>>2]|0;
      $545 = HEAP32[$124>>2]|0;
      FUNCTION_TABLE_vii[$543 & 8191]($544,$545);
      HEAP32[$102>>2] = 6;
      $$13$i = $$12$i;$$5$i = $$4$i;$b$11$i = $b$6$lcssa$i;$k$11$i = $k$6$lcssa$i;$n$11$i = $n$6$lcssa$i;$p$11$i = $p$6$lcssa$i;$q$0$i1938 = $q$0$i1939;
     }
     HEAP32[$105>>2] = $b$11$i;
     HEAP32[$107>>2] = $k$11$i;
     HEAP32[$8>>2] = $n$11$i;
     $546 = HEAP32[$z>>2]|0;
     $547 = $p$11$i;
     $548 = (($547) - ($546))|0;
     $549 = HEAP32[$9>>2]|0;
     $550 = (($548) + ($549))|0;
     HEAP32[$9>>2] = $550;
     HEAP32[$z>>2] = $p$11$i;
     HEAP32[$109>>2] = $q$0$i1938;
     $551 = HEAP32[$121>>2]|0;
     $552 = HEAP32[$111>>2]|0;
     $553 = ($q$0$i1938>>>0)<($552>>>0);
     if ($553) {
      $554 = $552;
      $555 = $q$0$i1938;
      $556 = $555 ^ -1;
      $557 = (($554) + ($556))|0;
      $m$0$ph$i$i = $557;
     } else {
      $558 = HEAP32[$$pre$phi906$iZ2D>>2]|0;
      $559 = $q$0$i1938;
      $560 = (($558) - ($559))|0;
      $m$0$ph$i$i = $560;
     }
     $561 = ((($551)) + 16|0);
     $562 = ((($551)) + 8|0);
     $563 = ((($551)) + 12|0);
     $564 = ((($551)) + 20|0);
     $565 = ((($551)) + 8|0);
     $566 = ((($551)) + 4|0);
     $567 = ((($551)) + 17|0);
     $568 = ((($551)) + 24|0);
     $$pre$i$i = HEAP32[$551>>2]|0;
     $$01$i$i = $$13$i;$569 = $$pre$i$i;$b$0$i$i = $b$11$i;$k$0$i$i = $k$11$i;$m$0$i$i = $m$0$ph$i$i;$n$0$i$i = $n$11$i;$p$0$i$i = $p$11$i;$q$0$i$i = $q$0$i1938;
     L180: while(1) {
      switch ($569|0) {
      case 7:  {
       $$01$i$i$lcssa2098 = $$01$i$i;$b$0$i$i$lcssa2109 = $b$0$i$i;$k$0$i$i$lcssa2121 = $k$0$i$i;$n$0$i$i$lcssa2143 = $n$0$i$i;$p$0$i$i$lcssa2132 = $p$0$i$i;$q$0$i$i$lcssa2154 = $q$0$i$i;
       label = 230;
       break L180;
       break;
      }
      case 8:  {
       $b$0$i$i2120 = $b$0$i$i;$k$8$i$i = $k$0$i$i;$n$8$i$i = $n$0$i$i;$p$8$i$i = $p$0$i$i;$q$8$i$i = $q$0$i$i;
       label = 233;
       break L180;
       break;
      }
      case 9:  {
       $b$0$i$i$lcssa2111 = $b$0$i$i;$k$0$i$i$lcssa2123 = $k$0$i$i;$n$0$i$i$lcssa2145 = $n$0$i$i;$p$0$i$i$lcssa2134 = $p$0$i$i;$q$0$i$i$lcssa2156 = $q$0$i$i;
       label = 234;
       break L180;
       break;
      }
      case 0:  {
       $570 = HEAP8[$561>>0]|0;
       $571 = $570&255;
       HEAP32[$563>>2] = $571;
       $572 = HEAP32[$564>>2]|0;
       HEAP32[$562>>2] = $572;
       HEAP32[$551>>2] = 1;
       $574 = $571;
       label = 165;
       break;
      }
      case 1:  {
       $$pre315$i$i = HEAP32[$563>>2]|0;
       $574 = $$pre315$i$i;
       label = 165;
       break;
      }
      case 2:  {
       $623 = HEAP32[$565>>2]|0;
       $624 = ($k$0$i$i>>>0)<($623>>>0);
       if ($624) {
        $$2156$i$i = $$01$i$i;$b$2157$i$i = $b$0$i$i;$k$2158$i$i = $k$0$i$i;$n$2160$i$i = $n$0$i$i;$p$2159$i$i = $p$0$i$i;
        while(1) {
         $625 = ($n$2160$i$i|0)==(0);
         if ($625) {
          $$2156$i$i$lcssa = $$2156$i$i;$b$2157$i$i$lcssa = $b$2157$i$i;$k$2158$i$i$lcssa = $k$2158$i$i;$p$2159$i$i$lcssa = $p$2159$i$i;$q$0$i$i$lcssa2160 = $q$0$i$i;
          label = 181;
          break L180;
         }
         $626 = (($n$2160$i$i) + -1)|0;
         $627 = ((($p$2159$i$i)) + 1|0);
         $628 = HEAP8[$p$2159$i$i>>0]|0;
         $629 = $628&255;
         $630 = $629 << $k$2158$i$i;
         $631 = $630 | $b$2157$i$i;
         $632 = (($k$2158$i$i) + 8)|0;
         $633 = ($632>>>0)<($623>>>0);
         if ($633) {
          $$2156$i$i = 0;$b$2157$i$i = $631;$k$2158$i$i = $632;$n$2160$i$i = $626;$p$2159$i$i = $627;
         } else {
          $$2$lcssa$i$i = 0;$b$2$lcssa$i$i = $631;$k$2$lcssa$i$i = $632;$n$2$lcssa$i$i = $626;$p$2$lcssa$i$i = $627;
          break;
         }
        }
       } else {
        $$2$lcssa$i$i = $$01$i$i;$b$2$lcssa$i$i = $b$0$i$i;$k$2$lcssa$i$i = $k$0$i$i;$n$2$lcssa$i$i = $n$0$i$i;$p$2$lcssa$i$i = $p$0$i$i;
       }
       $640 = (67056 + ($623<<2)|0);
       $641 = HEAP32[$640>>2]|0;
       $642 = $641 & $b$2$lcssa$i$i;
       $643 = HEAP32[$566>>2]|0;
       $644 = (($643) + ($642))|0;
       HEAP32[$566>>2] = $644;
       $645 = $b$2$lcssa$i$i >>> $623;
       $646 = (($k$2$lcssa$i$i) - ($623))|0;
       $647 = HEAP8[$567>>0]|0;
       $648 = $647&255;
       HEAP32[$563>>2] = $648;
       $649 = HEAP32[$568>>2]|0;
       HEAP32[$565>>2] = $649;
       HEAP32[$551>>2] = 3;
       $$3$i$i = $$2$lcssa$i$i;$651 = $648;$b$3$i$i = $645;$k$3$i$i = $646;$n$3$i$i = $n$2$lcssa$i$i;$p$3$i$i = $p$2$lcssa$i$i;
       label = 183;
       break;
      }
      case 3:  {
       $$pre313$i$i = HEAP32[$563>>2]|0;
       $$3$i$i = $$01$i$i;$651 = $$pre313$i$i;$b$3$i$i = $b$0$i$i;$k$3$i$i = $k$0$i$i;$n$3$i$i = $n$0$i$i;$p$3$i$i = $p$0$i$i;
       label = 183;
       break;
      }
      case 4:  {
       $695 = HEAP32[$565>>2]|0;
       $696 = ($k$0$i$i>>>0)<($695>>>0);
       if ($696) {
        $$5133$i$i = $$01$i$i;$b$5134$i$i = $b$0$i$i;$k$5135$i$i = $k$0$i$i;$n$5137$i$i = $n$0$i$i;$p$5136$i$i = $p$0$i$i;
        while(1) {
         $697 = ($n$5137$i$i|0)==(0);
         if ($697) {
          $$5133$i$i$lcssa = $$5133$i$i;$b$5134$i$i$lcssa = $b$5134$i$i;$k$5135$i$i$lcssa = $k$5135$i$i;$p$5136$i$i$lcssa = $p$5136$i$i;$q$0$i$i$lcssa2158 = $q$0$i$i;
          label = 195;
          break L180;
         }
         $698 = (($n$5137$i$i) + -1)|0;
         $699 = ((($p$5136$i$i)) + 1|0);
         $700 = HEAP8[$p$5136$i$i>>0]|0;
         $701 = $700&255;
         $702 = $701 << $k$5135$i$i;
         $703 = $702 | $b$5134$i$i;
         $704 = (($k$5135$i$i) + 8)|0;
         $705 = ($704>>>0)<($695>>>0);
         if ($705) {
          $$5133$i$i = 0;$b$5134$i$i = $703;$k$5135$i$i = $704;$n$5137$i$i = $698;$p$5136$i$i = $699;
         } else {
          $$5$lcssa$i$i = 0;$b$5$lcssa$i$i = $703;$k$5$lcssa$i$i = $704;$n$5$lcssa$i$i = $698;$p$5$lcssa$i$i = $699;
          break;
         }
        }
       } else {
        $$5$lcssa$i$i = $$01$i$i;$b$5$lcssa$i$i = $b$0$i$i;$k$5$lcssa$i$i = $k$0$i$i;$n$5$lcssa$i$i = $n$0$i$i;$p$5$lcssa$i$i = $p$0$i$i;
       }
       $712 = (67056 + ($695<<2)|0);
       $713 = HEAP32[$712>>2]|0;
       $714 = $713 & $b$5$lcssa$i$i;
       $715 = HEAP32[$563>>2]|0;
       $716 = (($715) + ($714))|0;
       HEAP32[$563>>2] = $716;
       $717 = $b$5$lcssa$i$i >>> $695;
       $718 = (($k$5$lcssa$i$i) - ($695))|0;
       HEAP32[$551>>2] = 5;
       $$6$i$i = $$5$lcssa$i$i;$720 = $716;$b$6$i$i = $717;$k$6$i$i = $718;$n$6$i$i = $n$5$lcssa$i$i;$p$6$i$i = $p$5$lcssa$i$i;
       label = 197;
       break;
      }
      case 5:  {
       $$pre310$i$i = HEAP32[$563>>2]|0;
       $$6$i$i = $$01$i$i;$720 = $$pre310$i$i;$b$6$i$i = $b$0$i$i;$k$6$i$i = $k$0$i$i;$n$6$i$i = $n$0$i$i;$p$6$i$i = $p$0$i$i;
       label = 197;
       break;
      }
      case 6:  {
       $783 = ($m$0$i$i|0)==(0);
       do {
        if ($783) {
         $784 = HEAP32[$$pre$phi$iZ2D>>2]|0;
         $785 = ($q$0$i$i|0)==($784|0);
         if ($785) {
          $786 = HEAP32[$111>>2]|0;
          $787 = HEAP32[$118>>2]|0;
          $788 = ($786|0)==($787|0);
          if ($788) {
           $q$55$i$i = $q$0$i$i;
          } else {
           $789 = ($787>>>0)<($786>>>0);
           $790 = $786;
           $791 = $787;
           $792 = (($790) + -1)|0;
           $793 = $q$0$i$i;
           $$v195$i$i = $789 ? $792 : $793;
           $794 = (($$v195$i$i) - ($791))|0;
           $795 = ($$v195$i$i|0)==($791|0);
           if ($795) {
            $q$55$i$i = $787;
           } else {
            $m$7$i$i = $794;$q$7$i$i = $787;
            break;
           }
          }
         } else {
          $q$55$i$i = $q$0$i$i;
         }
         HEAP32[$109>>2] = $q$55$i$i;
         $796 = (_inflate_flush($102,$z,$$01$i$i)|0);
         $797 = HEAP32[$109>>2]|0;
         $798 = HEAP32[$111>>2]|0;
         $799 = ($797>>>0)<($798>>>0);
         if ($799) {
          $800 = $798;
          $801 = $797;
          $802 = $801 ^ -1;
          $803 = (($800) + ($802))|0;
          $$pre309$i$i = HEAP32[$$pre$phi$iZ2D>>2]|0;
          $1000 = $803;$809 = $$pre309$i$i;
         } else {
          $804 = HEAP32[$$pre$phi906$iZ2D>>2]|0;
          $805 = $797;
          $806 = (($804) - ($805))|0;
          $807 = $804;
          $1000 = $806;$809 = $807;
         }
         $808 = ($797|0)==($809|0);
         if ($808) {
          $810 = HEAP32[$118>>2]|0;
          $811 = ($798|0)==($810|0);
          if ($811) {
           $m$6$i$i = $1000;$q$6$i$i = $797;
          } else {
           $812 = ($810>>>0)<($798>>>0);
           $813 = $798;
           $814 = $810;
           $815 = (($813) + -1)|0;
           $816 = $797;
           $$v$i$i = $812 ? $815 : $816;
           $817 = (($$v$i$i) - ($814))|0;
           $m$6$i$i = $817;$q$6$i$i = $810;
          }
         } else {
          $m$6$i$i = $1000;$q$6$i$i = $797;
         }
         $818 = ($m$6$i$i|0)==(0);
         if ($818) {
          $$lcssa2165 = $796;$b$0$i$i$lcssa2112 = $b$0$i$i;$k$0$i$i$lcssa2124 = $k$0$i$i;$n$0$i$i$lcssa2146 = $n$0$i$i;$p$0$i$i$lcssa2135 = $p$0$i$i;$q$6$i$i$lcssa = $q$6$i$i;
          label = 228;
          break L180;
         } else {
          $m$7$i$i = $m$6$i$i;$q$7$i$i = $q$6$i$i;
         }
        } else {
         $m$7$i$i = $m$0$i$i;$q$7$i$i = $q$0$i$i;
        }
       } while(0);
       $825 = HEAP32[$565>>2]|0;
       $826 = $825&255;
       $827 = ((($q$7$i$i)) + 1|0);
       HEAP8[$q$7$i$i>>0] = $826;
       $828 = (($m$7$i$i) + -1)|0;
       HEAP32[$551>>2] = 0;
       $p$0$i$i$phi = $p$0$i$i;$n$0$i$i$phi = $n$0$i$i;$k$0$i$i$phi = $k$0$i$i;$b$0$i$i$phi = $b$0$i$i;$$01$i$i = 0;$569 = 0;$m$0$i$i = $828;$q$0$i$i = $827;$p$0$i$i = $p$0$i$i$phi;$n$0$i$i = $n$0$i$i$phi;$k$0$i$i = $k$0$i$i$phi;$b$0$i$i = $b$0$i$i$phi;
       continue L180;
       break;
      }
      default: {
       $b$0$i$i$lcssa = $b$0$i$i;$k$0$i$i$lcssa = $k$0$i$i;$n$0$i$i$lcssa = $n$0$i$i;$p$0$i$i$lcssa = $p$0$i$i;$q$0$i$i$lcssa = $q$0$i$i;
       label = 235;
       break L180;
      }
      }
      if ((label|0) == 165) {
       label = 0;
       $573 = ($k$0$i$i>>>0)<($574>>>0);
       if ($573) {
        $$1182$i$i = $$01$i$i;$b$1183$i$i = $b$0$i$i;$k$1184$i$i = $k$0$i$i;$n$1186$i$i = $n$0$i$i;$p$1185$i$i = $p$0$i$i;
        while(1) {
         $575 = ($n$1186$i$i|0)==(0);
         if ($575) {
          $$1182$i$i$lcssa = $$1182$i$i;$b$1183$i$i$lcssa = $b$1183$i$i;$k$1184$i$i$lcssa = $k$1184$i$i;$p$1185$i$i$lcssa = $p$1185$i$i;$q$0$i$i$lcssa2163 = $q$0$i$i;
          label = 168;
          break L180;
         }
         $576 = (($n$1186$i$i) + -1)|0;
         $577 = ((($p$1185$i$i)) + 1|0);
         $578 = HEAP8[$p$1185$i$i>>0]|0;
         $579 = $578&255;
         $580 = $579 << $k$1184$i$i;
         $581 = $580 | $b$1183$i$i;
         $582 = (($k$1184$i$i) + 8)|0;
         $583 = ($582>>>0)<($574>>>0);
         if ($583) {
          $$1182$i$i = 0;$b$1183$i$i = $581;$k$1184$i$i = $582;$n$1186$i$i = $576;$p$1185$i$i = $577;
         } else {
          $$1$lcssa$i$i = 0;$b$1$lcssa$i$i = $581;$k$1$lcssa$i$i = $582;$n$1$lcssa$i$i = $576;$p$1$lcssa$i$i = $577;
          break;
         }
        }
       } else {
        $$1$lcssa$i$i = $$01$i$i;$b$1$lcssa$i$i = $b$0$i$i;$k$1$lcssa$i$i = $k$0$i$i;$n$1$lcssa$i$i = $n$0$i$i;$p$1$lcssa$i$i = $p$0$i$i;
       }
       $590 = HEAP32[$565>>2]|0;
       $591 = (67056 + ($574<<2)|0);
       $592 = HEAP32[$591>>2]|0;
       $593 = $592 & $b$1$lcssa$i$i;
       $594 = (($590) + ($593<<3)|0);
       $595 = ((($594)) + 1|0);
       $596 = HEAP8[$595>>0]|0;
       $597 = $596&255;
       $598 = $b$1$lcssa$i$i >>> $597;
       $599 = (($k$1$lcssa$i$i) - ($597))|0;
       $600 = HEAP8[$594>>0]|0;
       $601 = $600&255;
       $602 = ($600<<24>>24)==(0);
       if ($602) {
        $603 = (((($590) + ($593<<3)|0)) + 4|0);
        $604 = HEAP32[$603>>2]|0;
        HEAP32[$565>>2] = $604;
        HEAP32[$551>>2] = 6;
        $q$0$i$i$phi = $q$0$i$i;$m$0$i$i$phi = $m$0$i$i;$$01$i$i = $$1$lcssa$i$i;$569 = 6;$b$0$i$i = $598;$k$0$i$i = $599;$n$0$i$i = $n$1$lcssa$i$i;$p$0$i$i = $p$1$lcssa$i$i;$q$0$i$i = $q$0$i$i$phi;$m$0$i$i = $m$0$i$i$phi;
        continue;
       }
       $605 = $601 & 16;
       $606 = ($605|0)==(0);
       if (!($606)) {
        $607 = $601 & 15;
        HEAP32[$565>>2] = $607;
        $608 = (((($590) + ($593<<3)|0)) + 4|0);
        $609 = HEAP32[$608>>2]|0;
        HEAP32[$566>>2] = $609;
        HEAP32[$551>>2] = 2;
        $q$0$i$i$phi = $q$0$i$i;$m$0$i$i$phi = $m$0$i$i;$$01$i$i = $$1$lcssa$i$i;$569 = 2;$b$0$i$i = $598;$k$0$i$i = $599;$n$0$i$i = $n$1$lcssa$i$i;$p$0$i$i = $p$1$lcssa$i$i;$q$0$i$i = $q$0$i$i$phi;$m$0$i$i = $m$0$i$i$phi;
        continue;
       }
       $610 = $601 & 64;
       $611 = ($610|0)==(0);
       if ($611) {
        HEAP32[$563>>2] = $601;
        $612 = (((($590) + ($593<<3)|0)) + 4|0);
        $613 = HEAP32[$612>>2]|0;
        $614 = (($594) + ($613<<3)|0);
        HEAP32[$565>>2] = $614;
        $q$0$i$i$phi = $q$0$i$i;$m$0$i$i$phi = $m$0$i$i;$$01$i$i = $$1$lcssa$i$i;$569 = 1;$b$0$i$i = $598;$k$0$i$i = $599;$n$0$i$i = $n$1$lcssa$i$i;$p$0$i$i = $p$1$lcssa$i$i;$q$0$i$i = $q$0$i$i$phi;$m$0$i$i = $m$0$i$i$phi;
        continue;
       }
       $615 = $601 & 32;
       $616 = ($615|0)==(0);
       if ($616) {
        $$lcssa2168 = $598;$$lcssa2169 = $599;$n$1$lcssa$i$i$lcssa = $n$1$lcssa$i$i;$p$1$lcssa$i$i$lcssa = $p$1$lcssa$i$i;$q$0$i$i$lcssa2164 = $q$0$i$i;
        label = 177;
        break;
       }
       HEAP32[$551>>2] = 7;
       $q$0$i$i$phi = $q$0$i$i;$m$0$i$i$phi = $m$0$i$i;$$01$i$i = $$1$lcssa$i$i;$569 = 7;$b$0$i$i = $598;$k$0$i$i = $599;$n$0$i$i = $n$1$lcssa$i$i;$p$0$i$i = $p$1$lcssa$i$i;$q$0$i$i = $q$0$i$i$phi;$m$0$i$i = $m$0$i$i$phi;
       continue;
      }
      else if ((label|0) == 183) {
       label = 0;
       $650 = ($k$3$i$i>>>0)<($651>>>0);
       if ($650) {
        $$4169$i$i = $$3$i$i;$b$4170$i$i = $b$3$i$i;$k$4171$i$i = $k$3$i$i;$n$4173$i$i = $n$3$i$i;$p$4172$i$i = $p$3$i$i;
        while(1) {
         $652 = ($n$4173$i$i|0)==(0);
         if ($652) {
          $$4169$i$i$lcssa = $$4169$i$i;$b$4170$i$i$lcssa = $b$4170$i$i;$k$4171$i$i$lcssa = $k$4171$i$i;$p$4172$i$i$lcssa = $p$4172$i$i;$q$0$i$i$lcssa2161 = $q$0$i$i;
          label = 186;
          break L180;
         }
         $653 = (($n$4173$i$i) + -1)|0;
         $654 = ((($p$4172$i$i)) + 1|0);
         $655 = HEAP8[$p$4172$i$i>>0]|0;
         $656 = $655&255;
         $657 = $656 << $k$4171$i$i;
         $658 = $657 | $b$4170$i$i;
         $659 = (($k$4171$i$i) + 8)|0;
         $660 = ($659>>>0)<($651>>>0);
         if ($660) {
          $$4169$i$i = 0;$b$4170$i$i = $658;$k$4171$i$i = $659;$n$4173$i$i = $653;$p$4172$i$i = $654;
         } else {
          $$4$lcssa$i$i = 0;$b$4$lcssa$i$i = $658;$k$4$lcssa$i$i = $659;$n$4$lcssa$i$i = $653;$p$4$lcssa$i$i = $654;
          break;
         }
        }
       } else {
        $$4$lcssa$i$i = $$3$i$i;$b$4$lcssa$i$i = $b$3$i$i;$k$4$lcssa$i$i = $k$3$i$i;$n$4$lcssa$i$i = $n$3$i$i;$p$4$lcssa$i$i = $p$3$i$i;
       }
       $667 = HEAP32[$565>>2]|0;
       $668 = (67056 + ($651<<2)|0);
       $669 = HEAP32[$668>>2]|0;
       $670 = $669 & $b$4$lcssa$i$i;
       $671 = (($667) + ($670<<3)|0);
       $672 = ((($671)) + 1|0);
       $673 = HEAP8[$672>>0]|0;
       $674 = $673&255;
       $675 = $b$4$lcssa$i$i >>> $674;
       $676 = (($k$4$lcssa$i$i) - ($674))|0;
       $677 = HEAP8[$671>>0]|0;
       $678 = $677&255;
       $679 = $678 & 16;
       $680 = ($679|0)==(0);
       if (!($680)) {
        $681 = $678 & 15;
        HEAP32[$565>>2] = $681;
        $682 = (((($667) + ($670<<3)|0)) + 4|0);
        $683 = HEAP32[$682>>2]|0;
        HEAP32[$563>>2] = $683;
        HEAP32[$551>>2] = 4;
        $q$0$i$i$phi = $q$0$i$i;$m$0$i$i$phi = $m$0$i$i;$$01$i$i = $$4$lcssa$i$i;$569 = 4;$b$0$i$i = $675;$k$0$i$i = $676;$n$0$i$i = $n$4$lcssa$i$i;$p$0$i$i = $p$4$lcssa$i$i;$q$0$i$i = $q$0$i$i$phi;$m$0$i$i = $m$0$i$i$phi;
        continue;
       }
       $684 = $678 & 64;
       $685 = ($684|0)==(0);
       if (!($685)) {
        $$lcssa2166 = $675;$$lcssa2167 = $676;$n$4$lcssa$i$i$lcssa = $n$4$lcssa$i$i;$p$4$lcssa$i$i$lcssa = $p$4$lcssa$i$i;$q$0$i$i$lcssa2162 = $q$0$i$i;
        label = 191;
        break;
       }
       HEAP32[$563>>2] = $678;
       $686 = (((($667) + ($670<<3)|0)) + 4|0);
       $687 = HEAP32[$686>>2]|0;
       $688 = (($671) + ($687<<3)|0);
       HEAP32[$565>>2] = $688;
       $q$0$i$i$phi = $q$0$i$i;$m$0$i$i$phi = $m$0$i$i;$$01$i$i = $$4$lcssa$i$i;$569 = 3;$b$0$i$i = $675;$k$0$i$i = $676;$n$0$i$i = $n$4$lcssa$i$i;$p$0$i$i = $p$4$lcssa$i$i;$q$0$i$i = $q$0$i$i$phi;$m$0$i$i = $m$0$i$i$phi;
       continue;
      }
      else if ((label|0) == 197) {
       label = 0;
       $719 = (0 - ($720))|0;
       $721 = (($q$0$i$i) + ($719)|0);
       $722 = HEAP32[$118>>2]|0;
       $723 = ($721>>>0)<($722>>>0);
       if ($723) {
        $724 = HEAP32[$$pre$phi906$iZ2D>>2]|0;
        $725 = $722;
        $726 = (($724) - ($725))|0;
        $f$0142$i$i = $721;
        while(1) {
         $729 = (($f$0142$i$i) + ($726)|0);
         $730 = ($729>>>0)<($722>>>0);
         if ($730) {
          $f$0142$i$i = $729;
         } else {
          $f$0$lcssa$i$i = $729;
          break;
         }
        }
       } else {
        $f$0$lcssa$i$i = $721;
       }
       $727 = HEAP32[$566>>2]|0;
       $728 = ($727|0)==(0);
       if ($728) {
        $$7$lcssa$i$i = $$6$i$i;$m$1$lcssa$i$i = $m$0$i$i;$q$1$lcssa$i$i = $q$0$i$i;
       } else {
        $$7147$i$i = $$6$i$i;$f$1150$i$i = $f$0$lcssa$i$i;$m$1149$i$i = $m$0$i$i;$q$1148$i$i = $q$0$i$i;
        while(1) {
         $731 = ($m$1149$i$i|0)==(0);
         do {
          if ($731) {
           $732 = HEAP32[$$pre$phi$iZ2D>>2]|0;
           $733 = ($q$1148$i$i|0)==($732|0);
           if ($733) {
            $734 = HEAP32[$111>>2]|0;
            $735 = HEAP32[$118>>2]|0;
            $736 = ($734|0)==($735|0);
            if ($736) {
             $q$22$i$i = $q$1148$i$i;
            } else {
             $737 = ($735>>>0)<($734>>>0);
             $738 = $734;
             $739 = $735;
             $740 = (($738) + -1)|0;
             $741 = $q$1148$i$i;
             $$v197$i$i = $737 ? $740 : $741;
             $742 = (($$v197$i$i) - ($739))|0;
             $743 = ($$v197$i$i|0)==($739|0);
             if ($743) {
              $q$22$i$i = $735;
             } else {
              $m$4$i$i = $742;$q$4$i$i = $735;
              break;
             }
            }
           } else {
            $q$22$i$i = $q$1148$i$i;
           }
           HEAP32[$109>>2] = $q$22$i$i;
           $744 = (_inflate_flush($102,$z,$$7147$i$i)|0);
           $745 = HEAP32[$109>>2]|0;
           $746 = HEAP32[$111>>2]|0;
           $747 = ($745>>>0)<($746>>>0);
           if ($747) {
            $748 = $746;
            $749 = $745;
            $750 = $749 ^ -1;
            $751 = (($748) + ($750))|0;
            $$pre311$i$i = HEAP32[$$pre$phi$iZ2D>>2]|0;
            $757 = $$pre311$i$i;$999 = $751;
           } else {
            $752 = HEAP32[$$pre$phi906$iZ2D>>2]|0;
            $753 = $745;
            $754 = (($752) - ($753))|0;
            $755 = $752;
            $757 = $755;$999 = $754;
           }
           $756 = ($745|0)==($757|0);
           do {
            if ($756) {
             $758 = HEAP32[$118>>2]|0;
             $759 = ($746|0)==($758|0);
             if ($759) {
              $m$3$i$i = $999;$q$3$i$i = $745;
              break;
             }
             $760 = ($758>>>0)<($746>>>0);
             $761 = $746;
             $762 = $758;
             $763 = (($761) + -1)|0;
             $764 = $745;
             $$v196$i$i = $760 ? $763 : $764;
             $765 = (($$v196$i$i) - ($762))|0;
             $m$3$i$i = $765;$q$3$i$i = $758;
            } else {
             $m$3$i$i = $999;$q$3$i$i = $745;
            }
           } while(0);
           $766 = ($m$3$i$i|0)==(0);
           if ($766) {
            $$lcssa2071 = $744;$b$6$i$i$lcssa = $b$6$i$i;$k$6$i$i$lcssa = $k$6$i$i;$n$6$i$i$lcssa = $n$6$i$i;$p$6$i$i$lcssa = $p$6$i$i;$q$3$i$i$lcssa = $q$3$i$i;
            label = 212;
            break L180;
           } else {
            $m$4$i$i = $m$3$i$i;$q$4$i$i = $q$3$i$i;
           }
          } else {
           $m$4$i$i = $m$1149$i$i;$q$4$i$i = $q$1148$i$i;
          }
         } while(0);
         $773 = ((($f$1150$i$i)) + 1|0);
         $774 = HEAP8[$f$1150$i$i>>0]|0;
         $775 = ((($q$4$i$i)) + 1|0);
         HEAP8[$q$4$i$i>>0] = $774;
         $776 = (($m$4$i$i) + -1)|0;
         $777 = HEAP32[$$pre$phi$iZ2D>>2]|0;
         $778 = ($773|0)==($777|0);
         if ($778) {
          $779 = HEAP32[$118>>2]|0;
          $f$2$i$i = $779;
         } else {
          $f$2$i$i = $773;
         }
         $780 = HEAP32[$566>>2]|0;
         $781 = (($780) + -1)|0;
         HEAP32[$566>>2] = $781;
         $782 = ($781|0)==(0);
         if ($782) {
          $$7$lcssa$i$i = 0;$m$1$lcssa$i$i = $776;$q$1$lcssa$i$i = $775;
          break;
         } else {
          $$7147$i$i = 0;$f$1150$i$i = $f$2$i$i;$m$1149$i$i = $776;$q$1148$i$i = $775;
         }
        }
       }
       HEAP32[$551>>2] = 0;
       $$01$i$i = $$7$lcssa$i$i;$569 = 0;$b$0$i$i = $b$6$i$i;$k$0$i$i = $k$6$i$i;$m$0$i$i = $m$1$lcssa$i$i;$n$0$i$i = $n$6$i$i;$p$0$i$i = $p$6$i$i;$q$0$i$i = $q$1$lcssa$i$i;
       continue;
      }
     }
     switch (label|0) {
      case 168: {
       label = 0;
       HEAP32[$105>>2] = $b$1183$i$i$lcssa;
       HEAP32[$107>>2] = $k$1184$i$i$lcssa;
       HEAP32[$8>>2] = 0;
       $584 = HEAP32[$z>>2]|0;
       $585 = $p$1185$i$i$lcssa;
       $586 = (($585) - ($584))|0;
       $587 = HEAP32[$9>>2]|0;
       $588 = (($586) + ($587))|0;
       HEAP32[$9>>2] = $588;
       HEAP32[$z>>2] = $p$1185$i$i$lcssa;
       HEAP32[$109>>2] = $q$0$i$i$lcssa2163;
       $589 = (_inflate_flush($102,$z,$$1182$i$i$lcssa)|0);
       $$0$i24$i = $589;
       break;
      }
      case 177: {
       label = 0;
       HEAP32[$551>>2] = 9;
       HEAP32[$10>>2] = 613751;
       HEAP32[$105>>2] = $$lcssa2168;
       HEAP32[$107>>2] = $$lcssa2169;
       HEAP32[$8>>2] = $n$1$lcssa$i$i$lcssa;
       $617 = HEAP32[$z>>2]|0;
       $618 = $p$1$lcssa$i$i$lcssa;
       $619 = (($618) - ($617))|0;
       $620 = HEAP32[$9>>2]|0;
       $621 = (($619) + ($620))|0;
       HEAP32[$9>>2] = $621;
       HEAP32[$z>>2] = $p$1$lcssa$i$i$lcssa;
       HEAP32[$109>>2] = $q$0$i$i$lcssa2164;
       $622 = (_inflate_flush($102,$z,-3)|0);
       $$0$i24$i = $622;
       break;
      }
      case 181: {
       label = 0;
       HEAP32[$105>>2] = $b$2157$i$i$lcssa;
       HEAP32[$107>>2] = $k$2158$i$i$lcssa;
       HEAP32[$8>>2] = 0;
       $634 = HEAP32[$z>>2]|0;
       $635 = $p$2159$i$i$lcssa;
       $636 = (($635) - ($634))|0;
       $637 = HEAP32[$9>>2]|0;
       $638 = (($636) + ($637))|0;
       HEAP32[$9>>2] = $638;
       HEAP32[$z>>2] = $p$2159$i$i$lcssa;
       HEAP32[$109>>2] = $q$0$i$i$lcssa2160;
       $639 = (_inflate_flush($102,$z,$$2156$i$i$lcssa)|0);
       $$0$i24$i = $639;
       break;
      }
      case 186: {
       label = 0;
       HEAP32[$105>>2] = $b$4170$i$i$lcssa;
       HEAP32[$107>>2] = $k$4171$i$i$lcssa;
       HEAP32[$8>>2] = 0;
       $661 = HEAP32[$z>>2]|0;
       $662 = $p$4172$i$i$lcssa;
       $663 = (($662) - ($661))|0;
       $664 = HEAP32[$9>>2]|0;
       $665 = (($663) + ($664))|0;
       HEAP32[$9>>2] = $665;
       HEAP32[$z>>2] = $p$4172$i$i$lcssa;
       HEAP32[$109>>2] = $q$0$i$i$lcssa2161;
       $666 = (_inflate_flush($102,$z,$$4169$i$i$lcssa)|0);
       $$0$i24$i = $666;
       break;
      }
      case 191: {
       label = 0;
       HEAP32[$551>>2] = 9;
       HEAP32[$10>>2] = 613729;
       HEAP32[$105>>2] = $$lcssa2166;
       HEAP32[$107>>2] = $$lcssa2167;
       HEAP32[$8>>2] = $n$4$lcssa$i$i$lcssa;
       $689 = HEAP32[$z>>2]|0;
       $690 = $p$4$lcssa$i$i$lcssa;
       $691 = (($690) - ($689))|0;
       $692 = HEAP32[$9>>2]|0;
       $693 = (($691) + ($692))|0;
       HEAP32[$9>>2] = $693;
       HEAP32[$z>>2] = $p$4$lcssa$i$i$lcssa;
       HEAP32[$109>>2] = $q$0$i$i$lcssa2162;
       $694 = (_inflate_flush($102,$z,-3)|0);
       $$0$i24$i = $694;
       break;
      }
      case 195: {
       label = 0;
       HEAP32[$105>>2] = $b$5134$i$i$lcssa;
       HEAP32[$107>>2] = $k$5135$i$i$lcssa;
       HEAP32[$8>>2] = 0;
       $706 = HEAP32[$z>>2]|0;
       $707 = $p$5136$i$i$lcssa;
       $708 = (($707) - ($706))|0;
       $709 = HEAP32[$9>>2]|0;
       $710 = (($708) + ($709))|0;
       HEAP32[$9>>2] = $710;
       HEAP32[$z>>2] = $p$5136$i$i$lcssa;
       HEAP32[$109>>2] = $q$0$i$i$lcssa2158;
       $711 = (_inflate_flush($102,$z,$$5133$i$i$lcssa)|0);
       $$0$i24$i = $711;
       break;
      }
      case 212: {
       label = 0;
       HEAP32[$105>>2] = $b$6$i$i$lcssa;
       HEAP32[$107>>2] = $k$6$i$i$lcssa;
       HEAP32[$8>>2] = $n$6$i$i$lcssa;
       $767 = HEAP32[$z>>2]|0;
       $768 = $p$6$i$i$lcssa;
       $769 = (($768) - ($767))|0;
       $770 = HEAP32[$9>>2]|0;
       $771 = (($769) + ($770))|0;
       HEAP32[$9>>2] = $771;
       HEAP32[$z>>2] = $p$6$i$i$lcssa;
       HEAP32[$109>>2] = $q$3$i$i$lcssa;
       $772 = (_inflate_flush($102,$z,$$lcssa2071)|0);
       $$0$i24$i = $772;
       break;
      }
      case 228: {
       label = 0;
       HEAP32[$105>>2] = $b$0$i$i$lcssa2112;
       HEAP32[$107>>2] = $k$0$i$i$lcssa2124;
       HEAP32[$8>>2] = $n$0$i$i$lcssa2146;
       $819 = HEAP32[$z>>2]|0;
       $820 = $p$0$i$i$lcssa2135;
       $821 = (($820) - ($819))|0;
       $822 = HEAP32[$9>>2]|0;
       $823 = (($821) + ($822))|0;
       HEAP32[$9>>2] = $823;
       HEAP32[$z>>2] = $p$0$i$i$lcssa2135;
       HEAP32[$109>>2] = $q$6$i$i$lcssa;
       $824 = (_inflate_flush($102,$z,$$lcssa2165)|0);
       $$0$i24$i = $824;
       break;
      }
      case 230: {
       label = 0;
       $829 = ($k$0$i$i$lcssa2121>>>0)>(7);
       $830 = (($k$0$i$i$lcssa2121) + -8)|0;
       $831 = ((($p$0$i$i$lcssa2132)) + -1|0);
       $k$7$i$i = $829 ? $830 : $k$0$i$i$lcssa2121;
       $p$7$i$i = $829 ? $831 : $p$0$i$i$lcssa2132;
       $832 = $829&1;
       $n$7$i$i = (($n$0$i$i$lcssa2143) + ($832))|0;
       HEAP32[$109>>2] = $q$0$i$i$lcssa2154;
       $833 = (_inflate_flush($102,$z,$$01$i$i$lcssa2098)|0);
       $834 = HEAP32[$109>>2]|0;
       $835 = HEAP32[$111>>2]|0;
       $836 = ($835|0)==($834|0);
       if ($836) {
        HEAP32[$551>>2] = 8;
        $b$0$i$i2120 = $b$0$i$i$lcssa2109;$k$8$i$i = $k$7$i$i;$n$8$i$i = $n$7$i$i;$p$8$i$i = $p$7$i$i;$q$8$i$i = $834;
        label = 233;
        break;
       } else {
        HEAP32[$105>>2] = $b$0$i$i$lcssa2109;
        HEAP32[$107>>2] = $k$7$i$i;
        HEAP32[$8>>2] = $n$7$i$i;
        $837 = HEAP32[$z>>2]|0;
        $838 = $p$7$i$i;
        $839 = (($838) - ($837))|0;
        $840 = HEAP32[$9>>2]|0;
        $841 = (($839) + ($840))|0;
        HEAP32[$9>>2] = $841;
        HEAP32[$z>>2] = $p$7$i$i;
        $842 = (_inflate_flush($102,$z,$833)|0);
        $$0$i24$i = $842;
        break;
       }
       break;
      }
      case 234: {
       label = 0;
       HEAP32[$105>>2] = $b$0$i$i$lcssa2111;
       HEAP32[$107>>2] = $k$0$i$i$lcssa2123;
       HEAP32[$8>>2] = $n$0$i$i$lcssa2145;
       $849 = HEAP32[$z>>2]|0;
       $850 = $p$0$i$i$lcssa2134;
       $851 = (($850) - ($849))|0;
       $852 = HEAP32[$9>>2]|0;
       $853 = (($851) + ($852))|0;
       HEAP32[$9>>2] = $853;
       HEAP32[$z>>2] = $p$0$i$i$lcssa2134;
       HEAP32[$109>>2] = $q$0$i$i$lcssa2156;
       $854 = (_inflate_flush($102,$z,-3)|0);
       $$0$i24$i = $854;
       break;
      }
      case 235: {
       label = 0;
       HEAP32[$105>>2] = $b$0$i$i$lcssa;
       HEAP32[$107>>2] = $k$0$i$i$lcssa;
       HEAP32[$8>>2] = $n$0$i$i$lcssa;
       $855 = HEAP32[$z>>2]|0;
       $856 = $p$0$i$i$lcssa;
       $857 = (($856) - ($855))|0;
       $858 = HEAP32[$9>>2]|0;
       $859 = (($857) + ($858))|0;
       HEAP32[$9>>2] = $859;
       HEAP32[$z>>2] = $p$0$i$i$lcssa;
       HEAP32[$109>>2] = $q$0$i$i$lcssa;
       $860 = (_inflate_flush($102,$z,-2)|0);
       $$0$i24$i = $860;
       break;
      }
     }
     if ((label|0) == 233) {
      label = 0;
      HEAP32[$105>>2] = $b$0$i$i2120;
      HEAP32[$107>>2] = $k$8$i$i;
      HEAP32[$8>>2] = $n$8$i$i;
      $843 = HEAP32[$z>>2]|0;
      $844 = $p$8$i$i;
      $845 = (($844) - ($843))|0;
      $846 = HEAP32[$9>>2]|0;
      $847 = (($845) + ($846))|0;
      HEAP32[$9>>2] = $847;
      HEAP32[$z>>2] = $p$8$i$i;
      HEAP32[$109>>2] = $q$8$i$i;
      $848 = (_inflate_flush($102,$z,1)|0);
      $$0$i24$i = $848;
     }
     $861 = ($$0$i24$i|0)==(1);
     if (!($861)) {
      $$0$i24$i$lcssa = $$0$i24$i;
      label = 237;
      break;
     }
     $863 = HEAP32[$121>>2]|0;
     $z$idx$val$i = HEAP32[$z$idx$i>>2]|0;
     $z$idx14$val$i = HEAP32[$z$idx18$i>>2]|0;
     FUNCTION_TABLE_vii[$z$idx$val$i & 8191]($z$idx14$val$i,$863);
     $864 = HEAP32[$z>>2]|0;
     $865 = HEAP32[$8>>2]|0;
     $866 = HEAP32[$105>>2]|0;
     $867 = HEAP32[$107>>2]|0;
     $868 = HEAP32[$109>>2]|0;
     $869 = HEAP32[$111>>2]|0;
     $870 = ($868>>>0)<($869>>>0);
     if ($870) {
      $871 = $869;
      $872 = $868;
      $873 = $872 ^ -1;
      $874 = (($871) + ($873))|0;
      $1001 = $874;
     } else {
      $875 = HEAP32[$$pre$phi906$iZ2D>>2]|0;
      $876 = $868;
      $877 = (($875) - ($876))|0;
      $1001 = $877;
     }
     $878 = HEAP32[$120>>2]|0;
     $879 = ($878|0)==(0);
     if (!($879)) {
      $$lcssa2223 = $864;$$lcssa2224 = $865;$$lcssa2225 = $866;$$lcssa2226 = $867;$$lcssa2227 = $868;
      label = 243;
      break;
     }
     HEAP32[$102>>2] = 0;
     $$0$ph$ph$i = $$5$i;$$04$ph$ph$i = 0;$b$0$ph$ph$i = $866;$k$0$ph$ph$i = $867;$m$0$ph84$ph$i = $1001;$n$0$ph$ph$i = $865;$p$0$ph$ph$i = $864;$q$0$ph$ph$i = $868;
    }
    switch (label|0) {
     case 47: {
      label = 0;
      HEAP32[$105>>2] = $b$0$ph$i$lcssa2000;
      HEAP32[$107>>2] = $k$0$ph$i$lcssa2015;
      HEAP32[$8>>2] = 0;
      $136 = HEAP32[$z>>2]|0;
      $137 = $p$0$i$lcssa1897$lcssa;
      $138 = (($137) - ($136))|0;
      $139 = HEAP32[$9>>2]|0;
      $140 = (($138) + ($139))|0;
      HEAP32[$9>>2] = $140;
      HEAP32[$z>>2] = $p$0$i$lcssa1897$lcssa;
      HEAP32[$109>>2] = $q$0$i$lcssa1924$lcssa;
      $141 = (_inflate_flush($102,$z,$$04$i$lcssa1961$lcssa)|0);
      $$6$i = $141;
      break;
     }
     case 53: {
      label = 0;
      HEAP32[$121>>2] = 0;
      HEAP32[$105>>2] = $b$1$lcssa$i$lcssa2026;
      HEAP32[$107>>2] = $k$1$lcssa$i$lcssa2024;
      HEAP32[$8>>2] = $n$1$lcssa$i$lcssa2020;
      $165 = HEAP32[$z>>2]|0;
      $166 = $p$1$lcssa$i$lcssa2022;
      $167 = (($166) - ($165))|0;
      $168 = HEAP32[$9>>2]|0;
      $169 = (($167) + ($168))|0;
      HEAP32[$9>>2] = $169;
      HEAP32[$z>>2] = $p$1$lcssa$i$lcssa2022;
      HEAP32[$109>>2] = $q$0$i$lcssa1924$lcssa1944;
      $170 = (_inflate_flush($102,$z,-4)|0);
      $$6$i = $170;
      break;
     }
     case 55: {
      label = 0;
      $173 = $b$1$lcssa$i$lcssa2025 >>> 3;
      $174 = (($k$1$lcssa$i$lcssa2023) + -3)|0;
      HEAP32[$102>>2] = 9;
      HEAP32[$10>>2] = 612524;
      HEAP32[$105>>2] = $173;
      HEAP32[$107>>2] = $174;
      HEAP32[$8>>2] = $n$1$lcssa$i$lcssa2019;
      $175 = HEAP32[$z>>2]|0;
      $176 = $p$1$lcssa$i$lcssa2021;
      $177 = (($176) - ($175))|0;
      $178 = HEAP32[$9>>2]|0;
      $179 = (($177) + ($178))|0;
      HEAP32[$9>>2] = $179;
      HEAP32[$z>>2] = $p$1$lcssa$i$lcssa2021;
      HEAP32[$109>>2] = $q$0$i$lcssa1924$lcssa1943;
      $180 = (_inflate_flush($102,$z,-3)|0);
      $$6$i = $180;
      break;
     }
     case 58: {
      label = 0;
      HEAP32[$105>>2] = $b$2465$i$lcssa;
      HEAP32[$107>>2] = $k$2466$i$lcssa;
      HEAP32[$8>>2] = 0;
      $190 = HEAP32[$z>>2]|0;
      $191 = $p$2467$i$lcssa;
      $192 = (($191) - ($190))|0;
      $193 = HEAP32[$9>>2]|0;
      $194 = (($192) + ($193))|0;
      HEAP32[$9>>2] = $194;
      HEAP32[$z>>2] = $p$2467$i$lcssa;
      HEAP32[$109>>2] = $q$0$i$lcssa1925$lcssa;
      $195 = (_inflate_flush($102,$z,$$37469$i$lcssa)|0);
      $$6$i = $195;
      break;
     }
     case 60: {
      label = 0;
      HEAP32[$102>>2] = 9;
      HEAP32[$10>>2] = 612543;
      HEAP32[$105>>2] = $b$2$lcssa$i$lcssa;
      HEAP32[$107>>2] = $k$2$lcssa$i$lcssa;
      HEAP32[$8>>2] = $n$2$lcssa$i$lcssa;
      $200 = HEAP32[$z>>2]|0;
      $201 = $p$2$lcssa$i$lcssa;
      $202 = (($201) - ($200))|0;
      $203 = HEAP32[$9>>2]|0;
      $204 = (($202) + ($203))|0;
      HEAP32[$9>>2] = $204;
      HEAP32[$z>>2] = $p$2$lcssa$i$lcssa;
      HEAP32[$109>>2] = $q$0$i$lcssa1925$lcssa1945;
      $205 = (_inflate_flush($102,$z,-3)|0);
      $$6$i = $205;
      break;
     }
     case 66: {
      label = 0;
      HEAP32[$105>>2] = $b$0$ph$i$lcssa1996;
      HEAP32[$107>>2] = $k$0$ph$i$lcssa2011;
      HEAP32[$8>>2] = 0;
      $212 = HEAP32[$z>>2]|0;
      $213 = $p$0$i$lcssa1906;
      $214 = (($213) - ($212))|0;
      $215 = HEAP32[$9>>2]|0;
      $216 = (($214) + ($215))|0;
      HEAP32[$9>>2] = $216;
      HEAP32[$z>>2] = $p$0$i$lcssa1906;
      HEAP32[$109>>2] = $q$0$i$lcssa1933;
      $217 = (_inflate_flush($102,$z,$$04$i$lcssa1970)|0);
      $$6$i = $217;
      break;
     }
     case 78: {
      label = 0;
      HEAP32[$105>>2] = $b$0$ph$i$lcssa1997;
      HEAP32[$107>>2] = $k$0$ph$i$lcssa2012;
      HEAP32[$8>>2] = $n$0$i$lcssa1922;
      $254 = HEAP32[$z>>2]|0;
      $255 = $p$0$i$lcssa1907;
      $256 = (($255) - ($254))|0;
      $257 = HEAP32[$9>>2]|0;
      $258 = (($256) + ($257))|0;
      HEAP32[$9>>2] = $258;
      HEAP32[$z>>2] = $p$0$i$lcssa1907;
      HEAP32[$109>>2] = $q$2$i$lcssa;
      $259 = (_inflate_flush($102,$z,$$lcssa1976)|0);
      $$6$i = $259;
      break;
     }
     case 83: {
      label = 0;
      HEAP32[$105>>2] = $b$3486$i$lcssa;
      HEAP32[$107>>2] = $k$3487$i$lcssa;
      HEAP32[$8>>2] = 0;
      $282 = HEAP32[$z>>2]|0;
      $283 = $p$3488$i$lcssa;
      $284 = (($283) - ($282))|0;
      $285 = HEAP32[$9>>2]|0;
      $286 = (($284) + ($285))|0;
      HEAP32[$9>>2] = $286;
      HEAP32[$z>>2] = $p$3488$i$lcssa;
      HEAP32[$109>>2] = $q$0$i$lcssa1926$lcssa;
      $287 = (_inflate_flush($102,$z,$$48490$i$lcssa)|0);
      $$6$i = $287;
      break;
     }
     case 86: {
      label = 0;
      HEAP32[$102>>2] = 9;
      HEAP32[$10>>2] = 612572;
      HEAP32[$105>>2] = $b$3$lcssa$i$lcssa;
      HEAP32[$107>>2] = $k$3$lcssa$i$lcssa;
      HEAP32[$8>>2] = $n$3$lcssa$i$lcssa;
      $294 = HEAP32[$z>>2]|0;
      $295 = $p$3$lcssa$i$lcssa;
      $296 = (($295) - ($294))|0;
      $297 = HEAP32[$9>>2]|0;
      $298 = (($296) + ($297))|0;
      HEAP32[$9>>2] = $298;
      HEAP32[$z>>2] = $p$3$lcssa$i$lcssa;
      HEAP32[$109>>2] = $q$0$i1937$lcssa;
      $299 = (_inflate_flush($102,$z,-3)|0);
      $$6$i = $299;
      break;
     }
     case 88: {
      label = 0;
      HEAP32[$105>>2] = $b$3$lcssa$i$lcssa2174;
      HEAP32[$107>>2] = $k$3$lcssa$i$lcssa2173;
      HEAP32[$8>>2] = $n$3$lcssa$i$lcssa2171;
      $306 = HEAP32[$z>>2]|0;
      $307 = $p$3$lcssa$i$lcssa2172;
      $308 = (($307) - ($306))|0;
      $309 = HEAP32[$9>>2]|0;
      $310 = (($308) + ($309))|0;
      HEAP32[$9>>2] = $310;
      HEAP32[$z>>2] = $p$3$lcssa$i$lcssa2172;
      HEAP32[$109>>2] = $q$0$i1937$lcssa2170;
      $311 = (_inflate_flush($102,$z,-4)|0);
      $$6$i = $311;
      break;
     }
     case 96: {
      label = 0;
      HEAP32[$105>>2] = $b$4510$i$lcssa;
      HEAP32[$107>>2] = $k$4511$i$lcssa;
      HEAP32[$8>>2] = 0;
      $323 = HEAP32[$z>>2]|0;
      $324 = $p$4512$i$lcssa;
      $325 = (($324) - ($323))|0;
      $326 = HEAP32[$9>>2]|0;
      $327 = (($325) + ($326))|0;
      HEAP32[$9>>2] = $327;
      HEAP32[$z>>2] = $p$4512$i$lcssa;
      HEAP32[$109>>2] = $q$0$i1941$lcssa;
      $328 = (_inflate_flush($102,$z,$$59514$i$lcssa)|0);
      $$6$i = $328;
      break;
     }
     case 101: {
      label = 0;
      $$0$i28$i = -4;$b$4$lcssa$i2207 = $b$4$lcssa$i$lcssa;$k$4$lcssa$i2200 = $k$4$lcssa$i$lcssa;$n$4$lcssa$i2186 = $n$4$lcssa$i$lcssa;$p$4$lcssa$i2193 = $p$4$lcssa$i$lcssa;$q$0$i19402179 = $q$0$i1940$lcssa;
      label = 108;
      break;
     }
     case 104: {
      label = 0;
      $368 = $$lcssa2212;$b$4$lcssa$i2209 = $b$4$lcssa$i$lcssa2204;$k$4$lcssa$i2202 = $k$4$lcssa$i$lcssa2197;$n$4$lcssa$i2188 = $n$4$lcssa$i$lcssa2183;$p$4$lcssa$i2195 = $p$4$lcssa$i$lcssa2190;$q$0$i19402181 = $q$0$i1940$lcssa2176;$storemerge$i = 595641;
      label = 105;
      break;
     }
     case 116: {
      label = 0;
      HEAP32[$105>>2] = $b$7523$i$lcssa;
      HEAP32[$107>>2] = $k$7524$i$lcssa;
      HEAP32[$8>>2] = 0;
      $399 = HEAP32[$z>>2]|0;
      $400 = $p$7525$i$lcssa;
      $401 = (($400) - ($399))|0;
      $402 = HEAP32[$9>>2]|0;
      $403 = (($401) + ($402))|0;
      HEAP32[$9>>2] = $403;
      HEAP32[$z>>2] = $p$7525$i$lcssa;
      HEAP32[$109>>2] = $q$0$i1939$lcssa;
      $404 = (_inflate_flush($102,$z,$$8527$i$lcssa)|0);
      $$6$i = $404;
      break;
     }
     case 123: {
      label = 0;
      HEAP32[$105>>2] = $b$8536$i$lcssa;
      HEAP32[$107>>2] = $k$8537$i$lcssa;
      HEAP32[$8>>2] = 0;
      $445 = HEAP32[$z>>2]|0;
      $446 = $p$8538$i$lcssa;
      $447 = (($446) - ($445))|0;
      $448 = HEAP32[$9>>2]|0;
      $449 = (($447) + ($448))|0;
      HEAP32[$9>>2] = $449;
      HEAP32[$z>>2] = $p$8538$i$lcssa;
      HEAP32[$109>>2] = $q$0$i1939$lcssa2219;
      $450 = (_inflate_flush($102,$z,$$9540$i$lcssa)|0);
      $$6$i = $450;
      break;
     }
     case 126: {
      label = 0;
      $469 = HEAP32[$z$idx$i>>2]|0;
      $470 = HEAP32[$z$idx18$i>>2]|0;
      $471 = HEAP32[$124>>2]|0;
      FUNCTION_TABLE_vii[$469 & 8191]($470,$471);
      HEAP32[$102>>2] = 9;
      HEAP32[$10>>2] = 612633;
      HEAP32[$105>>2] = $$lcssa2058;
      HEAP32[$107>>2] = $$lcssa2059;
      HEAP32[$8>>2] = $n$8$lcssa$i$lcssa;
      $472 = HEAP32[$z>>2]|0;
      $473 = $p$8$lcssa$i$lcssa;
      $474 = (($473) - ($472))|0;
      $475 = HEAP32[$9>>2]|0;
      $476 = (($474) + ($475))|0;
      HEAP32[$9>>2] = $476;
      HEAP32[$z>>2] = $p$8$lcssa$i$lcssa;
      HEAP32[$109>>2] = $q$0$i1939$lcssa2220;
      $477 = (_inflate_flush($102,$z,-3)|0);
      $$6$i = $477;
      break;
     }
     case 237: {
      label = 0;
      $862 = (_inflate_flush($102,$z,$$0$i24$i$lcssa)|0);
      $$6$i = $862;
      break;
     }
     case 243: {
      label = 0;
      HEAP32[$102>>2] = 7;
      $$14$i = 0;$b$12$i = $$lcssa2225;$k$12$i = $$lcssa2226;$n$12$i = $$lcssa2224;$p$12$i = $$lcssa2223;$q$4$i = $$lcssa2227;
      label = 244;
      break;
     }
     case 248: {
      label = 0;
      HEAP32[$105>>2] = $b$0$ph$i$lcssa1995;
      HEAP32[$107>>2] = $k$0$ph$i$lcssa2010;
      HEAP32[$8>>2] = $n$0$i$lcssa1920;
      $896 = HEAP32[$z>>2]|0;
      $897 = $p$0$i$lcssa1905;
      $898 = (($897) - ($896))|0;
      $899 = HEAP32[$9>>2]|0;
      $900 = (($898) + ($899))|0;
      HEAP32[$9>>2] = $900;
      HEAP32[$z>>2] = $p$0$i$lcssa1905;
      HEAP32[$109>>2] = $q$0$i$lcssa1932;
      $901 = (_inflate_flush($102,$z,-3)|0);
      $$6$i = $901;
      break;
     }
     case 249: {
      label = 0;
      HEAP32[$105>>2] = $b$0$ph$i$lcssa;
      HEAP32[$107>>2] = $k$0$ph$i$lcssa;
      HEAP32[$8>>2] = $n$0$i$lcssa;
      $902 = HEAP32[$z>>2]|0;
      $903 = $p$0$i$lcssa;
      $904 = (($903) - ($902))|0;
      $905 = HEAP32[$9>>2]|0;
      $906 = (($904) + ($905))|0;
      HEAP32[$9>>2] = $906;
      HEAP32[$z>>2] = $p$0$i$lcssa;
      HEAP32[$109>>2] = $q$0$i$lcssa;
      $907 = (_inflate_flush($102,$z,-2)|0);
      $$6$i = $907;
      break;
     }
    }
    do {
     if ((label|0) == 105) {
      label = 0;
      HEAP32[$10>>2] = $storemerge$i;
      $366 = HEAP32[$z$idx$i>>2]|0;
      $367 = HEAP32[$z$idx18$i>>2]|0;
      FUNCTION_TABLE_vii[$366 & 8191]($367,$368);
      $b$4$lcssa$i2208 = $b$4$lcssa$i2209;$k$4$lcssa$i2201 = $k$4$lcssa$i2202;$n$4$lcssa$i2187 = $n$4$lcssa$i2188;$p$4$lcssa$i2194 = $p$4$lcssa$i2195;$q$0$i19402180 = $q$0$i19402181;
      label = 107;
     }
     else if ((label|0) == 244) {
      label = 0;
      HEAP32[$109>>2] = $q$4$i;
      $880 = (_inflate_flush($102,$z,$$14$i)|0);
      $881 = HEAP32[$109>>2]|0;
      $882 = HEAP32[$111>>2]|0;
      $883 = ($882|0)==($881|0);
      if ($883) {
       HEAP32[$102>>2] = 8;
       $b$13$i = $b$12$i;$k$13$i = $k$12$i;$n$13$i = $n$12$i;$p$13$i = $p$12$i;$q$5$i = $881;
       label = 247;
       break;
      } else {
       HEAP32[$105>>2] = $b$12$i;
       HEAP32[$107>>2] = $k$12$i;
       HEAP32[$8>>2] = $n$12$i;
       $884 = HEAP32[$z>>2]|0;
       $885 = $p$12$i;
       $886 = (($885) - ($884))|0;
       $887 = HEAP32[$9>>2]|0;
       $888 = (($886) + ($887))|0;
       HEAP32[$9>>2] = $888;
       HEAP32[$z>>2] = $p$12$i;
       $889 = (_inflate_flush($102,$z,$880)|0);
       $$6$i = $889;
       break;
      }
     }
    } while(0);
    if ((label|0) == 107) {
     label = 0;
     $371 = HEAP32[$z$idx$i>>2]|0;
     $372 = HEAP32[$z$idx18$i>>2]|0;
     $373 = HEAP32[$124>>2]|0;
     FUNCTION_TABLE_vii[$371 & 8191]($372,$373);
     HEAP32[$102>>2] = 9;
     $$0$i28$i = -3;$b$4$lcssa$i2207 = $b$4$lcssa$i2208;$k$4$lcssa$i2200 = $k$4$lcssa$i2201;$n$4$lcssa$i2186 = $n$4$lcssa$i2187;$p$4$lcssa$i2193 = $p$4$lcssa$i2194;$q$0$i19402179 = $q$0$i19402180;
     label = 108;
    }
    else if ((label|0) == 247) {
     label = 0;
     HEAP32[$105>>2] = $b$13$i;
     HEAP32[$107>>2] = $k$13$i;
     HEAP32[$8>>2] = $n$13$i;
     $890 = HEAP32[$z>>2]|0;
     $891 = $p$13$i;
     $892 = (($891) - ($890))|0;
     $893 = HEAP32[$9>>2]|0;
     $894 = (($892) + ($893))|0;
     HEAP32[$9>>2] = $894;
     HEAP32[$z>>2] = $p$13$i;
     HEAP32[$109>>2] = $q$5$i;
     $895 = (_inflate_flush($102,$z,1)|0);
     $$6$i = $895;
    }
    if ((label|0) == 108) {
     label = 0;
     HEAP32[$105>>2] = $b$4$lcssa$i2207;
     HEAP32[$107>>2] = $k$4$lcssa$i2200;
     HEAP32[$8>>2] = $n$4$lcssa$i2186;
     $374 = HEAP32[$z>>2]|0;
     $375 = $p$4$lcssa$i2193;
     $376 = (($375) - ($374))|0;
     $377 = HEAP32[$9>>2]|0;
     $378 = (($376) + ($377))|0;
     HEAP32[$9>>2] = $378;
     HEAP32[$z>>2] = $p$4$lcssa$i2193;
     HEAP32[$109>>2] = $q$0$i19402179;
     $379 = (_inflate_flush($102,$z,$$0$i28$i)|0);
     $$6$i = $379;
    }
    switch ($$6$i|0) {
    case 0:  {
     $$0 = $7;
     label = 274;
     break L10;
     break;
    }
    case -3:  {
     $908 = HEAP32[$1>>2]|0;
     HEAP32[$908>>2] = 13;
     $909 = ((($908)) + 4|0);
     HEAP32[$909>>2] = 0;
     $$ph756 = $908;$r$0$ph = -3;
     continue L10;
     break;
    }
    default: {
    }
    }
    $910 = ($$6$i|0)==(1);
    if (!($910)) {
     $$0 = $$6$i;
     label = 274;
     break L10;
    }
    $911 = HEAP32[$1>>2]|0;
    $912 = ((($911)) + 20|0);
    $913 = HEAP32[$912>>2]|0;
    $914 = ((($911)) + 4|0);
    $915 = ((($913)) + 60|0);
    $916 = HEAP32[$915>>2]|0;
    HEAP32[$914>>2] = $916;
    $917 = HEAP32[$913>>2]|0;
    $918 = $917 & -2;
    $switch$i = ($918|0)==(4);
    if ($switch$i) {
     $919 = HEAP32[$z$idx$i>>2]|0;
     $920 = HEAP32[$z$idx18$i>>2]|0;
     $921 = ((($913)) + 12|0);
     $922 = HEAP32[$921>>2]|0;
     FUNCTION_TABLE_vii[$919 & 8191]($920,$922);
     $$pr$i1 = HEAP32[$913>>2]|0;
     $924 = $$pr$i1;
    } else {
     $924 = $917;
    }
    $923 = ($924|0)==(6);
    if ($923) {
     $925 = ((($913)) + 4|0);
     $926 = HEAP32[$925>>2]|0;
     $z$idx$val$i3 = HEAP32[$z$idx$i>>2]|0;
     $z$idx1$val$i = HEAP32[$z$idx18$i>>2]|0;
     FUNCTION_TABLE_vii[$z$idx$val$i3 & 8191]($z$idx1$val$i,$926);
    }
    HEAP32[$913>>2] = 0;
    $927 = ((($913)) + 28|0);
    HEAP32[$927>>2] = 0;
    $928 = ((($913)) + 32|0);
    HEAP32[$928>>2] = 0;
    $929 = ((($913)) + 40|0);
    $930 = HEAP32[$929>>2]|0;
    $931 = ((($913)) + 52|0);
    HEAP32[$931>>2] = $930;
    $932 = ((($913)) + 48|0);
    HEAP32[$932>>2] = $930;
    $933 = ((($913)) + 56|0);
    $934 = HEAP32[$933>>2]|0;
    $935 = ($934|0)==(0|0);
    if (!($935)) {
     $936 = (FUNCTION_TABLE_iiii[$934 & 8191](0,0,0)|0);
     HEAP32[$915>>2] = $936;
     HEAP32[$11>>2] = $936;
    }
    $937 = HEAP32[$1>>2]|0;
    $938 = ((($937)) + 12|0);
    $939 = HEAP32[$938>>2]|0;
    $940 = ($939|0)==(0);
    if ($940) {
     HEAP32[$937>>2] = 8;
     $952 = $937;$r$7 = $7;
     label = 262;
     break;
    } else {
     HEAP32[$937>>2] = 12;
     $$ph756 = $937;$r$0$ph = $7;
     continue L10;
    }
   }
  } while(0);
  if ((label|0) == 262) {
   label = 0;
   $941 = HEAP32[$8>>2]|0;
   $942 = ($941|0)==(0);
   if ($942) {
    $$0 = $r$7;
    label = 274;
    break;
   }
   $943 = (($941) + -1)|0;
   HEAP32[$8>>2] = $943;
   $944 = HEAP32[$9>>2]|0;
   $945 = (($944) + 1)|0;
   HEAP32[$9>>2] = $945;
   $946 = HEAP32[$z>>2]|0;
   $947 = ((($946)) + 1|0);
   HEAP32[$z>>2] = $947;
   $948 = HEAP8[$946>>0]|0;
   $949 = $948&255;
   $950 = $949 << 24;
   $951 = ((($952)) + 8|0);
   HEAP32[$951>>2] = $950;
   HEAP32[$952>>2] = 9;
   $954 = $943;$964 = $952;$r$8 = $7;
   label = 264;
  }
  if ((label|0) == 264) {
   label = 0;
   $953 = ($954|0)==(0);
   if ($953) {
    $$0 = $r$8;
    label = 274;
    break;
   }
   $955 = (($954) + -1)|0;
   HEAP32[$8>>2] = $955;
   $956 = HEAP32[$9>>2]|0;
   $957 = (($956) + 1)|0;
   HEAP32[$9>>2] = $957;
   $958 = HEAP32[$z>>2]|0;
   $959 = ((($958)) + 1|0);
   HEAP32[$z>>2] = $959;
   $960 = HEAP8[$958>>0]|0;
   $961 = $960&255;
   $962 = $961 << 16;
   $963 = ((($964)) + 8|0);
   $965 = HEAP32[$963>>2]|0;
   $966 = (($965) + ($962))|0;
   HEAP32[$963>>2] = $966;
   HEAP32[$964>>2] = 10;
   $968 = $955;$978 = $964;$r$9 = $7;
   label = 266;
  }
  if ((label|0) == 266) {
   label = 0;
   $967 = ($968|0)==(0);
   if ($967) {
    $$0 = $r$9;
    label = 274;
    break;
   }
   $969 = (($968) + -1)|0;
   HEAP32[$8>>2] = $969;
   $970 = HEAP32[$9>>2]|0;
   $971 = (($970) + 1)|0;
   HEAP32[$9>>2] = $971;
   $972 = HEAP32[$z>>2]|0;
   $973 = ((($972)) + 1|0);
   HEAP32[$z>>2] = $973;
   $974 = HEAP8[$972>>0]|0;
   $975 = $974&255;
   $976 = $975 << 8;
   $977 = ((($978)) + 8|0);
   $979 = HEAP32[$977>>2]|0;
   $980 = (($979) + ($976))|0;
   HEAP32[$977>>2] = $980;
   HEAP32[$978>>2] = 11;
   $982 = $969;$991 = $978;$r$10 = $7;
  }
  $981 = ($982|0)==(0);
  if ($981) {
   $$0 = $r$10;
   label = 274;
   break;
  }
  $983 = (($982) + -1)|0;
  HEAP32[$8>>2] = $983;
  $984 = HEAP32[$9>>2]|0;
  $985 = (($984) + 1)|0;
  HEAP32[$9>>2] = $985;
  $986 = HEAP32[$z>>2]|0;
  $987 = ((($986)) + 1|0);
  HEAP32[$z>>2] = $987;
  $988 = HEAP8[$986>>0]|0;
  $989 = $988&255;
  $990 = ((($991)) + 8|0);
  $992 = HEAP32[$990>>2]|0;
  $993 = (($992) + ($989))|0;
  HEAP32[$990>>2] = $993;
  $994 = ((($991)) + 4|0);
  $995 = HEAP32[$994>>2]|0;
  $996 = ($995|0)==($993|0);
  if ($996) {
   $$lcssa2255 = $991;
   label = 271;
   break;
  }
  HEAP32[$991>>2] = 13;
  HEAP32[$10>>2] = 612746;
  HEAP32[$994>>2] = 5;
  $$ph756 = $991;$r$0$ph = $7;
 }
 switch (label|0) {
  case 21: {
   HEAP32[$$lcssa2254>>2] = 2;
   $$ph7562242 = $$ph756$lcssa2238;$48 = $$lcssa;$58 = $$lcssa2254;$r$2 = $7;
   label = 23;
   break;
  }
  case 22: {
   $$pre751 = HEAP32[$8>>2]|0;
   $$ph7562242 = $$ph756$lcssa2228;$48 = $$pre751;$58 = $$lcssa2244;$r$2 = $r$0$lcssa1884;
   label = 23;
   break;
  }
  case 25: {
   $$pre752 = HEAP32[$8>>2]|0;
   $$ph7562241 = $$ph756$lcssa2229;$60 = $$pre752;$72 = $$lcssa2245;$r$3 = $r$0$lcssa1885;
   label = 26;
   break;
  }
  case 28: {
   $$pre753 = HEAP32[$8>>2]|0;
   $$ph7562240 = $$ph756$lcssa2230;$74 = $$pre753;$86 = $$lcssa2246;$r$4 = $r$0$lcssa1886;
   label = 29;
   break;
  }
  case 31: {
   $$pre754 = HEAP32[$8>>2]|0;
   $$ph7562239 = $$ph756$lcssa2231;$88 = $$pre754;$99 = $$lcssa2247;$r$5 = $r$0$lcssa1887;
   break;
  }
  case 34: {
   HEAP32[$$lcssa2248>>2] = 13;
   HEAP32[$10>>2] = 613583;
   $100 = ((($$ph756$lcssa2232)) + 4|0);
   HEAP32[$100>>2] = 0;
   $$0 = -2;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 40: {
   // unreachable;
   break;
  }
  case 271: {
   HEAP32[$$lcssa2255>>2] = 12;
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 272: {
   $$0 = $$0$ph;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 273: {
   $$0 = -3;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 274: {
   STACKTOP = sp;return ($$0|0);
   break;
  }
 }
 do {
  if ((label|0) == 23) {
   $47 = ($48|0)==(0);
   if ($47) {
    $$0 = $r$2;
    STACKTOP = sp;return ($$0|0);
   } else {
    $49 = (($48) + -1)|0;
    HEAP32[$8>>2] = $49;
    $50 = HEAP32[$9>>2]|0;
    $51 = (($50) + 1)|0;
    HEAP32[$9>>2] = $51;
    $52 = HEAP32[$z>>2]|0;
    $53 = ((($52)) + 1|0);
    HEAP32[$z>>2] = $53;
    $54 = HEAP8[$52>>0]|0;
    $55 = $54&255;
    $56 = $55 << 24;
    $57 = ((($$ph7562242)) + 8|0);
    HEAP32[$57>>2] = $56;
    HEAP32[$58>>2] = 3;
    $$ph7562241 = $$ph7562242;$60 = $49;$72 = $58;$r$3 = $7;
    label = 26;
    break;
   }
  }
 } while(0);
 do {
  if ((label|0) == 26) {
   $59 = ($60|0)==(0);
   if ($59) {
    $$0 = $r$3;
    STACKTOP = sp;return ($$0|0);
   } else {
    $61 = (($60) + -1)|0;
    HEAP32[$8>>2] = $61;
    $62 = HEAP32[$9>>2]|0;
    $63 = (($62) + 1)|0;
    HEAP32[$9>>2] = $63;
    $64 = HEAP32[$z>>2]|0;
    $65 = ((($64)) + 1|0);
    HEAP32[$z>>2] = $65;
    $66 = HEAP8[$64>>0]|0;
    $67 = $66&255;
    $68 = $67 << 16;
    $69 = ((($$ph7562241)) + 8|0);
    $70 = HEAP32[$69>>2]|0;
    $71 = (($70) + ($68))|0;
    HEAP32[$69>>2] = $71;
    HEAP32[$72>>2] = 4;
    $$ph7562240 = $$ph7562241;$74 = $61;$86 = $72;$r$4 = $7;
    label = 29;
    break;
   }
  }
 } while(0);
 do {
  if ((label|0) == 29) {
   $73 = ($74|0)==(0);
   if ($73) {
    $$0 = $r$4;
    STACKTOP = sp;return ($$0|0);
   } else {
    $75 = (($74) + -1)|0;
    HEAP32[$8>>2] = $75;
    $76 = HEAP32[$9>>2]|0;
    $77 = (($76) + 1)|0;
    HEAP32[$9>>2] = $77;
    $78 = HEAP32[$z>>2]|0;
    $79 = ((($78)) + 1|0);
    HEAP32[$z>>2] = $79;
    $80 = HEAP8[$78>>0]|0;
    $81 = $80&255;
    $82 = $81 << 8;
    $83 = ((($$ph7562240)) + 8|0);
    $84 = HEAP32[$83>>2]|0;
    $85 = (($84) + ($82))|0;
    HEAP32[$83>>2] = $85;
    HEAP32[$86>>2] = 5;
    $$ph7562239 = $$ph7562240;$88 = $75;$99 = $86;$r$5 = $7;
    break;
   }
  }
 } while(0);
 $87 = ($88|0)==(0);
 if ($87) {
  $$0 = $r$5;
  STACKTOP = sp;return ($$0|0);
 }
 $89 = (($88) + -1)|0;
 HEAP32[$8>>2] = $89;
 $90 = HEAP32[$9>>2]|0;
 $91 = (($90) + 1)|0;
 HEAP32[$9>>2] = $91;
 $92 = HEAP32[$z>>2]|0;
 $93 = ((($92)) + 1|0);
 HEAP32[$z>>2] = $93;
 $94 = HEAP8[$92>>0]|0;
 $95 = $94&255;
 $96 = ((($$ph7562239)) + 8|0);
 $97 = HEAP32[$96>>2]|0;
 $98 = (($97) + ($95))|0;
 HEAP32[$96>>2] = $98;
 HEAP32[$11>>2] = $98;
 HEAP32[$99>>2] = 6;
 $$0 = 2;
 STACKTOP = sp;return ($$0|0);
}
function _inflate_flush($s,$z,$r) {
 $s = $s|0;
 $z = $z|0;
 $r = $r|0;
 var $$ = 0, $$0 = 0, $$1 = 0, $$2 = 0, $$4 = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond3 = 0, $p$0 = 0, $q$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($z)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($s)) + 48|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($s)) + 52|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($3>>>0)>($5>>>0);
 if ($6) {
  $7 = ((($s)) + 44|0);
  $8 = HEAP32[$7>>2]|0;
  $10 = $8;
 } else {
  $10 = $5;
 }
 $9 = $10;
 $11 = $3;
 $12 = (($9) - ($11))|0;
 $13 = ((($z)) + 16|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ($12>>>0)>($14>>>0);
 $$ = $15 ? $14 : $12;
 $16 = ($$|0)!=(0);
 $17 = ($r|0)==(-5);
 $or$cond = $17 & $16;
 $$0 = $or$cond ? 0 : $r;
 $18 = (($14) - ($$))|0;
 HEAP32[$13>>2] = $18;
 $19 = ((($z)) + 20|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($$) + ($20))|0;
 HEAP32[$19>>2] = $21;
 $22 = ((($s)) + 56|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = ($23|0)==(0|0);
 if (!($24)) {
  $25 = ((($s)) + 60|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = (FUNCTION_TABLE_iiii[$23 & 8191]($26,$3,$$)|0);
  HEAP32[$25>>2] = $27;
  $28 = ((($z)) + 48|0);
  HEAP32[$28>>2] = $27;
 }
 _memcpy(($1|0),($3|0),($$|0))|0;
 $29 = (($1) + ($$)|0);
 $30 = (($3) + ($$)|0);
 $31 = ((($s)) + 44|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ($30|0)==($32|0);
 if (!($33)) {
  $$2 = $$0;$p$0 = $29;$q$0 = $30;
  HEAP32[$0>>2] = $p$0;
  HEAP32[$2>>2] = $q$0;
  return ($$2|0);
 }
 $34 = ((($s)) + 40|0);
 $35 = HEAP32[$34>>2]|0;
 $36 = HEAP32[$4>>2]|0;
 $37 = ($36|0)==($30|0);
 if ($37) {
  HEAP32[$4>>2] = $35;
  $39 = $35;
  $$pre$phiZ2D = $39;$41 = $39;
 } else {
  $38 = $36;
  $$pre = $35;
  $$pre$phiZ2D = $$pre;$41 = $38;
 }
 $40 = (($41) - ($$pre$phiZ2D))|0;
 $42 = HEAP32[$13>>2]|0;
 $43 = ($40>>>0)>($42>>>0);
 $$4 = $43 ? $42 : $40;
 $44 = ($$4|0)!=(0);
 $45 = ($$0|0)==(-5);
 $or$cond3 = $45 & $44;
 $$1 = $or$cond3 ? 0 : $$0;
 $46 = (($42) - ($$4))|0;
 HEAP32[$13>>2] = $46;
 $47 = HEAP32[$19>>2]|0;
 $48 = (($47) + ($$4))|0;
 HEAP32[$19>>2] = $48;
 $49 = HEAP32[$22>>2]|0;
 $50 = ($49|0)==(0|0);
 if (!($50)) {
  $51 = ((($s)) + 60|0);
  $52 = HEAP32[$51>>2]|0;
  $53 = (FUNCTION_TABLE_iiii[$49 & 8191]($52,$35,$$4)|0);
  HEAP32[$51>>2] = $53;
  $54 = ((($z)) + 48|0);
  HEAP32[$54>>2] = $53;
 }
 _memcpy(($29|0),($35|0),($$4|0))|0;
 $55 = (($29) + ($$4)|0);
 $56 = (($35) + ($$4)|0);
 $$2 = $$1;$p$0 = $55;$q$0 = $56;
 HEAP32[$0>>2] = $p$0;
 HEAP32[$2>>2] = $q$0;
 return ($$2|0);
}
function _huft_build($b,$n,$s,$d,$e,$t,$m,$hp,$hn,$v) {
 $b = $b|0;
 $n = $n|0;
 $s = $s|0;
 $d = $d|0;
 $e = $e|0;
 $t = $t|0;
 $m = $m|0;
 $hp = $hp|0;
 $hn = $hn|0;
 $v = $v|0;
 var $$ = 0, $$0 = 0, $$in = 0, $$lcssa = 0, $$lcssa217 = 0, $$lcssa223 = 0, $$lcssa227 = 0, $$lcssa233 = 0, $$pr$pr = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0;
 var $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0;
 var $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $c = 0, $f$035 = 0, $h$087 = 0;
 var $h$1$lcssa = 0, $h$172 = 0, $h$2$lcssa = 0, $h$2$ph52 = 0, $h$239 = 0, $h$239$lcssa221 = 0, $h$3$lcssa = 0, $h$366 = 0, $i$0 = 0, $i$1$j$0$ = 0, $i$1$j$0$138 = 0, $i$1$j$0$141 = 0, $i$1$j$0$147 = 0, $i$1$j$0$150$ph = 0, $i$1$lcssa = 0, $i$1$lcssa$ph = 0, $i$1$lcssa140 = 0, $i$1$lcssa149$ph = 0, $i$3 = 0, $i$488 = 0;
 var $i$5$lcssa = 0, $i$573 = 0, $i$6 = 0, $j$0$ = 0, $j$0$134 = 0, $j$0$136$ph = 0, $j$0$143 = 0, $j$0$145 = 0, $j$0$lcssa$ph$ph = 0, $j$0$lcssa$ph144 = 0, $j$0$lcssa135$ph = 0, $j$0$lcssa135139 = 0, $j$0$lcssa135148$ph = 0, $j$1104 = 0, $j$299 = 0, $j$4 = 0, $j$4$lcssa225 = 0, $j$560 = 0, $j$6 = 0, $k$089 = 0;
 var $mask$0 = 0, $mask$0$in = 0, $mask$0$in64 = 0, $mask$065 = 0, $or$cond = 0, $p$0 = 0, $p$1100 = 0, $p$2 = 0, $p$391 = 0, $p$4$lcssa = 0, $p$474 = 0, $p$5 = 0, $q$092 = 0, $q$1$lcssa = 0, $q$175 = 0, $q$2$lcssa = 0, $r$sroa$0$0 = 0, $r$sroa$101$093 = 0, $r$sroa$101$1$lcssa = 0, $r$sroa$101$176 = 0;
 var $r$sroa$101$2$ph$lcssa25 = 0, $r$sroa$101$2$ph53 = 0, $r$sroa$101$3 = 0, $u = 0, $w$094 = 0, $w$1$lcssa = 0, $w$177 = 0, $w$2$lcssa = 0, $w$2$ph54 = 0, $w$2$pn$lcssa = 0, $w$2$pn67 = 0, $w$240 = 0, $w$240$lcssa219 = 0, $w$240$phi = 0, $x = 0, $xp$0101 = 0, $xp$136 = 0, $y$0$lcssa = 0, $y$0105 = 0, $y$0105$ph = 0;
 var $z$095 = 0, $z$1$lcssa = 0, $z$178 = 0, $z$2$lcssa = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $c = sp + 128|0;
 $u = sp + 64|0;
 $x = sp;
 $0 = ((($c)) + 4|0);
 $1 = ((($c)) + 8|0);
 $2 = ((($c)) + 12|0);
 $3 = ((($c)) + 16|0);
 $4 = ((($c)) + 20|0);
 $5 = ((($c)) + 24|0);
 $6 = ((($c)) + 28|0);
 $7 = ((($c)) + 32|0);
 $8 = ((($c)) + 36|0);
 $9 = ((($c)) + 40|0);
 $10 = ((($c)) + 44|0);
 $11 = ((($c)) + 48|0);
 $12 = ((($c)) + 52|0);
 $13 = ((($c)) + 56|0);
 $14 = ((($c)) + 60|0);
 dest=$c; stop=dest+64|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $i$0 = $n;$p$0 = $b;
 while(1) {
  $15 = ((($p$0)) + 4|0);
  $16 = HEAP32[$p$0>>2]|0;
  $17 = (($c) + ($16<<2)|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = (($18) + 1)|0;
  HEAP32[$17>>2] = $19;
  $20 = (($i$0) + -1)|0;
  $21 = ($20|0)==(0);
  if ($21) {
   break;
  } else {
   $i$0 = $20;$p$0 = $15;
  }
 }
 $22 = HEAP32[$c>>2]|0;
 $23 = ($22|0)==($n|0);
 if ($23) {
  HEAP32[$t>>2] = 0;
  HEAP32[$m>>2] = 0;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $24 = HEAP32[$m>>2]|0;
 $25 = HEAP32[$0>>2]|0;
 $26 = ($25|0)==(0);
 do {
  if ($26) {
   $27 = HEAP32[$1>>2]|0;
   $28 = ($27|0)==(0);
   if ($28) {
    $199 = HEAP32[$2>>2]|0;
    $200 = ($199|0)==(0);
    if ($200) {
     $201 = HEAP32[$3>>2]|0;
     $202 = ($201|0)==(0);
     if ($202) {
      $203 = HEAP32[$4>>2]|0;
      $204 = ($203|0)==(0);
      if ($204) {
       $205 = HEAP32[$5>>2]|0;
       $206 = ($205|0)==(0);
       if ($206) {
        $207 = HEAP32[$6>>2]|0;
        $208 = ($207|0)==(0);
        if ($208) {
         $209 = HEAP32[$7>>2]|0;
         $210 = ($209|0)==(0);
         if ($210) {
          $211 = HEAP32[$8>>2]|0;
          $212 = ($211|0)==(0);
          if ($212) {
           $213 = HEAP32[$9>>2]|0;
           $214 = ($213|0)==(0);
           if ($214) {
            $215 = HEAP32[$10>>2]|0;
            $216 = ($215|0)==(0);
            if ($216) {
             $217 = HEAP32[$11>>2]|0;
             $218 = ($217|0)==(0);
             if ($218) {
              $219 = HEAP32[$12>>2]|0;
              $220 = ($219|0)==(0);
              if ($220) {
               $221 = HEAP32[$13>>2]|0;
               $222 = ($221|0)==(0);
               if ($222) {
                $223 = HEAP32[$14>>2]|0;
                $224 = ($223|0)==(0);
                if ($224) {
                 $225 = ($24>>>0)<(16);
                 $j$0$143 = $225 ? 16 : $24;
                 $j$0$145 = $j$0$143;$j$0$lcssa$ph144 = 16;
                 label = 9;
                 break;
                } else {
                 $226 = ($24>>>0)<(15);
                 $j$0$134 = $226 ? 15 : $24;
                 $i$1$lcssa$ph = 15;$j$0$136$ph = $j$0$134;$j$0$lcssa135$ph = 15;
                 label = 10;
                 break;
                }
               } else {
                $j$0$lcssa$ph$ph = 14;
                label = 7;
               }
              } else {
               $j$0$lcssa$ph$ph = 13;
               label = 7;
              }
             } else {
              $j$0$lcssa$ph$ph = 12;
              label = 7;
             }
            } else {
             $j$0$lcssa$ph$ph = 11;
             label = 7;
            }
           } else {
            $j$0$lcssa$ph$ph = 10;
            label = 7;
           }
          } else {
           $j$0$lcssa$ph$ph = 9;
           label = 7;
          }
         } else {
          $j$0$lcssa$ph$ph = 8;
          label = 7;
         }
        } else {
         $j$0$lcssa$ph$ph = 7;
         label = 7;
        }
       } else {
        $j$0$lcssa$ph$ph = 6;
        label = 7;
       }
      } else {
       $j$0$lcssa$ph$ph = 5;
       label = 7;
      }
     } else {
      $j$0$lcssa$ph$ph = 4;
      label = 7;
     }
    } else {
     $j$0$lcssa$ph$ph = 3;
     label = 7;
    }
   } else {
    $j$0$lcssa$ph$ph = 2;
    label = 7;
   }
  } else {
   $j$0$lcssa$ph$ph = 1;
   label = 7;
  }
 } while(0);
 if ((label|0) == 7) {
  $$pr$pr = HEAP32[$14>>2]|0;
  $29 = ($24>>>0)<($j$0$lcssa$ph$ph>>>0);
  $j$0$ = $29 ? $j$0$lcssa$ph$ph : $24;
  $30 = ($$pr$pr|0)==(0);
  if ($30) {
   $j$0$145 = $j$0$;$j$0$lcssa$ph144 = $j$0$lcssa$ph$ph;
   label = 9;
  } else {
   $31 = ($j$0$>>>0)>(15);
   $i$1$j$0$147 = $31 ? 15 : $j$0$;
   HEAP32[$m>>2] = $i$1$j$0$147;
   $32 = 1 << $j$0$lcssa$ph$ph;
   $i$1$j$0$150$ph = $i$1$j$0$147;$i$1$lcssa149$ph = 15;$j$0$lcssa135148$ph = $j$0$lcssa$ph$ph;$y$0105$ph = $32;
   label = 12;
  }
 }
 do {
  if ((label|0) == 9) {
   $33 = HEAP32[$13>>2]|0;
   $34 = ($33|0)==(0);
   if ($34) {
    $174 = HEAP32[$12>>2]|0;
    $175 = ($174|0)==(0);
    if ($175) {
     $176 = HEAP32[$11>>2]|0;
     $177 = ($176|0)==(0);
     if ($177) {
      $178 = HEAP32[$10>>2]|0;
      $179 = ($178|0)==(0);
      if ($179) {
       $180 = HEAP32[$9>>2]|0;
       $181 = ($180|0)==(0);
       if ($181) {
        $182 = HEAP32[$8>>2]|0;
        $183 = ($182|0)==(0);
        if ($183) {
         $184 = HEAP32[$7>>2]|0;
         $185 = ($184|0)==(0);
         if ($185) {
          $186 = HEAP32[$6>>2]|0;
          $187 = ($186|0)==(0);
          if ($187) {
           $188 = HEAP32[$5>>2]|0;
           $189 = ($188|0)==(0);
           if ($189) {
            $190 = HEAP32[$4>>2]|0;
            $191 = ($190|0)==(0);
            if ($191) {
             $192 = HEAP32[$3>>2]|0;
             $193 = ($192|0)==(0);
             if ($193) {
              $194 = HEAP32[$2>>2]|0;
              $195 = ($194|0)==(0);
              if ($195) {
               $196 = HEAP32[$1>>2]|0;
               $197 = ($196|0)==(0);
               if ($197) {
                $198 = $26&1;
                $$ = $198 ^ 1;
                $i$1$lcssa$ph = $$;$j$0$136$ph = $j$0$145;$j$0$lcssa135$ph = $j$0$lcssa$ph144;
                label = 10;
                break;
               } else {
                $i$1$lcssa = 2;
               }
              } else {
               $i$1$lcssa = 3;
              }
             } else {
              $i$1$lcssa = 4;
             }
            } else {
             $i$1$lcssa = 5;
            }
           } else {
            $i$1$lcssa = 6;
           }
          } else {
           $i$1$lcssa = 7;
          }
         } else {
          $i$1$lcssa = 8;
         }
        } else {
         $i$1$lcssa = 9;
        }
       } else {
        $i$1$lcssa = 10;
       }
      } else {
       $i$1$lcssa = 11;
      }
     } else {
      $i$1$lcssa = 12;
     }
    } else {
     $i$1$lcssa = 13;
    }
   } else {
    $i$1$lcssa = 14;
   }
   $37 = ($j$0$145>>>0)>($i$1$lcssa>>>0);
   $i$1$j$0$ = $37 ? $i$1$lcssa : $j$0$145;
   HEAP32[$m>>2] = $i$1$j$0$;
   $38 = 1 << $j$0$lcssa$ph144;
   $39 = ($j$0$lcssa$ph144>>>0)<($i$1$lcssa>>>0);
   if ($39) {
    $i$1$j$0$150$ph = $i$1$j$0$;$i$1$lcssa149$ph = $i$1$lcssa;$j$0$lcssa135148$ph = $j$0$lcssa$ph144;$y$0105$ph = $38;
    label = 12;
   } else {
    $i$1$j$0$141 = $i$1$j$0$;$i$1$lcssa140 = $i$1$lcssa;$j$0$lcssa135139 = $j$0$lcssa$ph144;$y$0$lcssa = $38;
   }
  }
 } while(0);
 L46: do {
  if ((label|0) == 10) {
   $35 = ($j$0$136$ph>>>0)>($i$1$lcssa$ph>>>0);
   $i$1$j$0$138 = $35 ? $i$1$lcssa$ph : $j$0$136$ph;
   HEAP32[$m>>2] = $i$1$j$0$138;
   $36 = 1 << $j$0$lcssa135$ph;
   $i$1$j$0$141 = $i$1$j$0$138;$i$1$lcssa140 = $i$1$lcssa$ph;$j$0$lcssa135139 = $j$0$lcssa135$ph;$y$0$lcssa = $36;
  }
  else if ((label|0) == 12) {
   $j$1104 = $j$0$lcssa135148$ph;$y$0105 = $y$0105$ph;
   while(1) {
    $40 = (($c) + ($j$1104<<2)|0);
    $41 = HEAP32[$40>>2]|0;
    $42 = (($y$0105) - ($41))|0;
    $43 = ($42|0)<(0);
    if ($43) {
     $$0 = -3;
     break;
    }
    $44 = (($j$1104) + 1)|0;
    $45 = $42 << 1;
    $46 = ($44>>>0)<($i$1$lcssa149$ph>>>0);
    if ($46) {
     $j$1104 = $44;$y$0105 = $45;
    } else {
     $i$1$j$0$141 = $i$1$j$0$150$ph;$i$1$lcssa140 = $i$1$lcssa149$ph;$j$0$lcssa135139 = $j$0$lcssa135148$ph;$y$0$lcssa = $45;
     break L46;
    }
   }
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 $47 = (($c) + ($i$1$lcssa140<<2)|0);
 $48 = HEAP32[$47>>2]|0;
 $49 = (($y$0$lcssa) - ($48))|0;
 $50 = ($49|0)<(0);
 if ($50) {
  $$0 = -3;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$47>>2] = $y$0$lcssa;
 $51 = ((($x)) + 4|0);
 HEAP32[$51>>2] = 0;
 $52 = (($i$1$lcssa140) + -1)|0;
 $53 = ($52|0)==(0);
 if ($53) {
  $i$3 = 0;$p$2 = $b;
 } else {
  $54 = ((($x)) + 8|0);
  $60 = $52;$j$299 = 0;$p$1100 = $0;$xp$0101 = $54;
  while(1) {
   $55 = ((($p$1100)) + 4|0);
   $56 = HEAP32[$p$1100>>2]|0;
   $57 = (($56) + ($j$299))|0;
   $58 = ((($xp$0101)) + 4|0);
   HEAP32[$xp$0101>>2] = $57;
   $59 = (($60) + -1)|0;
   $61 = ($59|0)==(0);
   if ($61) {
    $i$3 = 0;$p$2 = $b;
    break;
   } else {
    $60 = $59;$j$299 = $57;$p$1100 = $55;$xp$0101 = $58;
   }
  }
 }
 while(1) {
  $62 = ((($p$2)) + 4|0);
  $63 = HEAP32[$p$2>>2]|0;
  $64 = ($63|0)==(0);
  if (!($64)) {
   $65 = (($x) + ($63<<2)|0);
   $66 = HEAP32[$65>>2]|0;
   $67 = (($66) + 1)|0;
   HEAP32[$65>>2] = $67;
   $68 = (($v) + ($66<<2)|0);
   HEAP32[$68>>2] = $i$3;
  }
  $69 = (($i$3) + 1)|0;
  $70 = ($69>>>0)<($n>>>0);
  if ($70) {
   $i$3 = $69;$p$2 = $62;
  } else {
   break;
  }
 }
 $71 = (($x) + ($i$1$lcssa140<<2)|0);
 $72 = HEAP32[$71>>2]|0;
 HEAP32[$x>>2] = 0;
 HEAP32[$u>>2] = 0;
 $73 = ($j$0$lcssa135139>>>0)>($i$1$lcssa140>>>0);
 L67: do {
  if (!($73)) {
   $74 = (0 - ($i$1$j$0$141))|0;
   $75 = (($v) + ($72<<2)|0);
   $76 = $i$1$j$0$141&255;
   $h$087 = -1;$i$488 = 0;$k$089 = $j$0$lcssa135139;$p$391 = $v;$q$092 = 0;$r$sroa$101$093 = 0;$w$094 = $74;$z$095 = 0;
   L69: while(1) {
    $77 = (($c) + ($k$089<<2)|0);
    $78 = HEAP32[$77>>2]|0;
    $79 = ($78|0)==(0);
    if ($79) {
     $h$1$lcssa = $h$087;$i$5$lcssa = $i$488;$p$4$lcssa = $p$391;$q$1$lcssa = $q$092;$r$sroa$101$1$lcssa = $r$sroa$101$093;$w$1$lcssa = $w$094;$z$1$lcssa = $z$095;
    } else {
     $80 = (($k$089) + -1)|0;
     $81 = 1 << $80;
     $$in = $78;$h$172 = $h$087;$i$573 = $i$488;$p$474 = $p$391;$q$175 = $q$092;$r$sroa$101$176 = $r$sroa$101$093;$w$177 = $w$094;$z$178 = $z$095;
     while(1) {
      $83 = (($$in) + -1)|0;
      $84 = (($w$177) + ($i$1$j$0$141))|0;
      $85 = ($k$089|0)>($84|0);
      L75: do {
       if ($85) {
        $227 = $84;$h$2$ph52 = $h$172;$r$sroa$101$2$ph53 = $r$sroa$101$176;$w$2$ph54 = $w$177;
        while(1) {
         $88 = $227;$h$239 = $h$2$ph52;$w$240 = $w$2$ph54;
         while(1) {
          $86 = (($h$239) + 1)|0;
          $87 = (($i$1$lcssa140) - ($88))|0;
          $89 = ($87>>>0)>($i$1$j$0$141>>>0);
          $90 = $89 ? $i$1$j$0$141 : $87;
          $91 = (($k$089) - ($88))|0;
          $92 = 1 << $91;
          $93 = ($92>>>0)>($$in>>>0);
          $94 = ($91>>>0)<($90>>>0);
          $or$cond = $93 & $94;
          L80: do {
           if ($or$cond) {
            $95 = (($91) + 1)|0;
            $96 = ($95>>>0)<($90>>>0);
            if ($96) {
             $97 = (($92) - ($$in))|0;
             $102 = $95;$f$035 = $97;$xp$136 = $77;
             while(1) {
              $99 = $f$035 << 1;
              $104 = ((($xp$136)) + 4|0);
              $100 = HEAP32[$104>>2]|0;
              $105 = ($99>>>0)>($100>>>0);
              if (!($105)) {
               $j$4 = $102;
               break L80;
              }
              $98 = (($99) - ($100))|0;
              $101 = (($102) + 1)|0;
              $103 = ($101>>>0)<($90>>>0);
              if ($103) {
               $102 = $101;$f$035 = $98;$xp$136 = $104;
              } else {
               $j$4 = $101;
               break;
              }
             }
            } else {
             $j$4 = $95;
            }
           } else {
            $j$4 = $91;
           }
          } while(0);
          $106 = 1 << $j$4;
          $107 = HEAP32[$hn>>2]|0;
          $108 = (($107) + ($106))|0;
          $109 = ($108>>>0)>(1440);
          if ($109) {
           $$0 = -3;
           break L69;
          }
          $110 = (($hp) + ($107<<3)|0);
          $111 = (($u) + ($86<<2)|0);
          HEAP32[$111>>2] = $110;
          HEAP32[$hn>>2] = $108;
          $112 = ($86|0)==(0);
          if (!($112)) {
           $$lcssa = $110;$$lcssa217 = $88;$$lcssa223 = $86;$$lcssa227 = $106;$h$239$lcssa221 = $h$239;$j$4$lcssa225 = $j$4;$w$240$lcssa219 = $w$240;
           break;
          }
          HEAP32[$t>>2] = $110;
          $128 = (($88) + ($i$1$j$0$141))|0;
          $129 = ($k$089|0)>($128|0);
          if ($129) {
           $w$240$phi = $88;$88 = $128;$h$239 = 0;$w$240 = $w$240$phi;
          } else {
           $h$2$lcssa = 0;$q$2$lcssa = $110;$r$sroa$101$2$ph$lcssa25 = $r$sroa$101$2$ph53;$w$2$lcssa = $88;$z$2$lcssa = $106;
           break L75;
          }
         }
         $113 = (($x) + ($$lcssa223<<2)|0);
         HEAP32[$113>>2] = $i$573;
         $114 = $j$4$lcssa225&255;
         $115 = $i$573 >>> $w$240$lcssa219;
         $116 = (($u) + ($h$239$lcssa221<<2)|0);
         $117 = HEAP32[$116>>2]|0;
         $118 = $$lcssa;
         $119 = $117;
         $120 = (($118) - ($119))|0;
         $121 = $120 >> 3;
         $122 = (($121) - ($115))|0;
         $123 = (($117) + ($115<<3)|0);
         HEAP8[$123>>0] = $114;
         $124 = ((($123)) + 1|0);
         HEAP8[$124>>0] = $76;
         $125 = (((($117) + ($115<<3)|0)) + 4|0);
         HEAP32[$125>>2] = $122;
         $126 = (($$lcssa217) + ($i$1$j$0$141))|0;
         $127 = ($k$089|0)>($126|0);
         if ($127) {
          $227 = $126;$h$2$ph52 = $$lcssa223;$r$sroa$101$2$ph53 = $122;$w$2$ph54 = $$lcssa217;
         } else {
          $h$2$lcssa = $$lcssa223;$q$2$lcssa = $$lcssa;$r$sroa$101$2$ph$lcssa25 = $122;$w$2$lcssa = $$lcssa217;$z$2$lcssa = $$lcssa227;
          break;
         }
        }
       } else {
        $h$2$lcssa = $h$172;$q$2$lcssa = $q$175;$r$sroa$101$2$ph$lcssa25 = $r$sroa$101$176;$w$2$lcssa = $w$177;$z$2$lcssa = $z$178;
       }
      } while(0);
      $130 = (($k$089) - ($w$2$lcssa))|0;
      $131 = $130&255;
      $132 = ($p$474>>>0)<($75>>>0);
      do {
       if ($132) {
        $133 = HEAP32[$p$474>>2]|0;
        $134 = ($133>>>0)<($s>>>0);
        if ($134) {
         $135 = ($133>>>0)<(256);
         $136 = $135 ? 0 : 96;
         $137 = ((($p$474)) + 4|0);
         $p$5 = $137;$r$sroa$0$0 = $136;$r$sroa$101$3 = $133;
         break;
        } else {
         $138 = (($133) - ($s))|0;
         $139 = (($e) + ($138<<2)|0);
         $140 = HEAP32[$139>>2]|0;
         $141 = (($140) + 80)|0;
         $142 = $141&255;
         $143 = ((($p$474)) + 4|0);
         $144 = (($d) + ($138<<2)|0);
         $145 = HEAP32[$144>>2]|0;
         $p$5 = $143;$r$sroa$0$0 = $142;$r$sroa$101$3 = $145;
         break;
        }
       } else {
        $p$5 = $p$474;$r$sroa$0$0 = -64;$r$sroa$101$3 = $r$sroa$101$2$ph$lcssa25;
       }
      } while(0);
      $146 = 1 << $130;
      $147 = $i$573 >>> $w$2$lcssa;
      $148 = ($147>>>0)<($z$2$lcssa>>>0);
      if ($148) {
       $j$560 = $147;
       while(1) {
        $149 = (($q$2$lcssa) + ($j$560<<3)|0);
        HEAP8[$149>>0] = $r$sroa$0$0;
        $150 = ((($149)) + 1|0);
        HEAP8[$150>>0] = $131;
        $151 = (((($q$2$lcssa) + ($j$560<<3)|0)) + 4|0);
        HEAP32[$151>>2] = $r$sroa$101$3;
        $152 = (($j$560) + ($146))|0;
        $153 = ($152>>>0)<($z$2$lcssa>>>0);
        if ($153) {
         $j$560 = $152;
        } else {
         $i$6 = $i$573;$j$6 = $81;
         break;
        }
       }
      } else {
       $i$6 = $i$573;$j$6 = $81;
      }
      while(1) {
       $154 = $j$6 & $i$6;
       $155 = ($154|0)==(0);
       $156 = $j$6 ^ $i$6;
       $157 = $j$6 >>> 1;
       if ($155) {
        $$lcssa233 = $156;
        break;
       } else {
        $i$6 = $156;$j$6 = $157;
       }
      }
      $mask$0$in64 = 1 << $w$2$lcssa;
      $mask$065 = (($mask$0$in64) + -1)|0;
      $158 = $mask$065 & $$lcssa233;
      $159 = (($x) + ($h$2$lcssa<<2)|0);
      $160 = HEAP32[$159>>2]|0;
      $161 = ($158|0)==($160|0);
      if ($161) {
       $h$3$lcssa = $h$2$lcssa;$w$2$pn$lcssa = $w$2$lcssa;
      } else {
       $h$366 = $h$2$lcssa;$w$2$pn67 = $w$2$lcssa;
       while(1) {
        $162 = (($h$366) + -1)|0;
        $163 = (($w$2$pn67) - ($i$1$j$0$141))|0;
        $mask$0$in = 1 << $163;
        $mask$0 = (($mask$0$in) + -1)|0;
        $164 = $mask$0 & $$lcssa233;
        $165 = (($x) + ($162<<2)|0);
        $166 = HEAP32[$165>>2]|0;
        $167 = ($164|0)==($166|0);
        if ($167) {
         $h$3$lcssa = $162;$w$2$pn$lcssa = $163;
         break;
        } else {
         $h$366 = $162;$w$2$pn67 = $163;
        }
       }
      }
      $82 = ($83|0)==(0);
      if ($82) {
       $h$1$lcssa = $h$3$lcssa;$i$5$lcssa = $$lcssa233;$p$4$lcssa = $p$5;$q$1$lcssa = $q$2$lcssa;$r$sroa$101$1$lcssa = $r$sroa$101$3;$w$1$lcssa = $w$2$pn$lcssa;$z$1$lcssa = $z$2$lcssa;
       break;
      } else {
       $$in = $83;$h$172 = $h$3$lcssa;$i$573 = $$lcssa233;$p$474 = $p$5;$q$175 = $q$2$lcssa;$r$sroa$101$176 = $r$sroa$101$3;$w$177 = $w$2$pn$lcssa;$z$178 = $z$2$lcssa;
      }
     }
    }
    $168 = (($k$089) + 1)|0;
    $169 = ($k$089|0)<($i$1$lcssa140|0);
    if ($169) {
     $h$087 = $h$1$lcssa;$i$488 = $i$5$lcssa;$k$089 = $168;$p$391 = $p$4$lcssa;$q$092 = $q$1$lcssa;$r$sroa$101$093 = $r$sroa$101$1$lcssa;$w$094 = $w$1$lcssa;$z$095 = $z$1$lcssa;
    } else {
     break L67;
    }
   }
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 $170 = ($y$0$lcssa|0)!=($48|0);
 $171 = ($i$1$lcssa140|0)!=(1);
 $172 = $171 & $170;
 $173 = $172 ? -5 : 0;
 $$0 = $173;
 STACKTOP = sp;return ($$0|0);
}
function _ft_gzip_stream_close($stream) {
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 $3 = ((($stream)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($1)) + 12|0);
 (_inflateEnd_8493($5)|0);
 $6 = ((($1)) + 44|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($1)) + 48|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($1)) + 52|0);
 HEAP32[$8>>2] = 0;
 $9 = ((($1)) + 24|0);
 HEAP32[$9>>2] = 0;
 $10 = ((($1)) + 28|0);
 HEAP32[$10>>2] = 0;
 ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0;HEAP32[$1+12>>2]=0|0;HEAP32[$1+16>>2]=0|0;
 _ft_mem_free($4,$1);
 HEAP32[$0>>2] = 0;
 return;
}
function _ft_gzip_stream_io($stream,$pos,$buffer,$count) {
 $stream = $stream|0;
 $pos = $pos|0;
 $buffer = $buffer|0;
 $count = $count|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_ft_gzip_file_io($1,$pos,$buffer,$count)|0);
 return ($2|0);
}
function _FT_Gzip_Uncompress($memory,$output,$output_len,$input,$input_len) {
 $memory = $memory|0;
 $output = $output|0;
 $output_len = $output_len|0;
 $input = $input|0;
 $input_len = $input_len|0;
 var $$ = 0, $$0 = 0, $$pr$i$i$i = 0, $$pr$i$i$i7 = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i$i11 = 0, $$pre$i12 = 0, $$pre1$i = 0, $$pre1$i14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, $or$cond3 = 0, $stream = 0, $switch$i$i$i = 0, $switch$i$i$i6 = 0, $z$idx$val$i$i$i = 0, $z$idx$val$i$i$i8 = 0, $z$idx1$val$i$i$i = 0, $z$idx1$val$i$i$i10 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $stream = sp;
 $0 = ($memory|0)!=(0|0);
 $1 = ($output_len|0)!=(0|0);
 $or$cond = $0 & $1;
 $2 = ($output|0)!=(0|0);
 $or$cond3 = $2 & $or$cond;
 if (!($or$cond3)) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$stream>>2] = $input;
 $3 = ((($stream)) + 4|0);
 HEAP32[$3>>2] = $input_len;
 $4 = ((($stream)) + 12|0);
 HEAP32[$4>>2] = $output;
 $5 = HEAP32[$output_len>>2]|0;
 $6 = ((($stream)) + 16|0);
 HEAP32[$6>>2] = $5;
 $7 = ((($stream)) + 32|0);
 HEAP32[$7>>2] = 4440;
 $8 = ((($stream)) + 36|0);
 HEAP32[$8>>2] = 4441;
 $9 = ((($stream)) + 40|0);
 HEAP32[$9>>2] = $memory;
 $10 = (_inflateInit2__8492($stream,15)|0);
 $11 = ($10|0)==(0);
 if (!($11)) {
  $$0 = 6;
  STACKTOP = sp;return ($$0|0);
 }
 $12 = (_inflate_8494($stream,4)|0);
 $13 = ($12|0)==(1);
 L7: do {
  if ($13) {
   $56 = ((($stream)) + 20|0);
   $57 = HEAP32[$56>>2]|0;
   HEAP32[$output_len>>2] = $57;
   $58 = ((($stream)) + 28|0);
   $59 = HEAP32[$58>>2]|0;
   $60 = ($59|0)==(0|0);
   if (!($60)) {
    $61 = HEAP32[$8>>2]|0;
    $62 = ($61|0)==(0|0);
    if (!($62)) {
     $63 = ((($59)) + 20|0);
     $64 = HEAP32[$63>>2]|0;
     $65 = ($64|0)==(0|0);
     if ($65) {
      $98 = $61;$99 = $59;
     } else {
      $66 = HEAP32[$64>>2]|0;
      $67 = $66 & -2;
      $switch$i$i$i6 = ($67|0)==(4);
      if ($switch$i$i$i6) {
       $68 = HEAP32[$9>>2]|0;
       $69 = ((($64)) + 12|0);
       $70 = HEAP32[$69>>2]|0;
       FUNCTION_TABLE_vii[$61 & 8191]($68,$70);
       $$pr$i$i$i7 = HEAP32[$64>>2]|0;
       $72 = $$pr$i$i$i7;
      } else {
       $72 = $66;
      }
      $71 = ($72|0)==(6);
      if ($71) {
       $73 = ((($64)) + 4|0);
       $74 = HEAP32[$73>>2]|0;
       $z$idx$val$i$i$i8 = HEAP32[$8>>2]|0;
       $z$idx1$val$i$i$i10 = HEAP32[$9>>2]|0;
       FUNCTION_TABLE_vii[$z$idx$val$i$i$i8 & 8191]($z$idx1$val$i$i$i10,$74);
      }
      HEAP32[$64>>2] = 0;
      $75 = ((($64)) + 28|0);
      HEAP32[$75>>2] = 0;
      $76 = ((($64)) + 32|0);
      HEAP32[$76>>2] = 0;
      $77 = ((($64)) + 40|0);
      $78 = HEAP32[$77>>2]|0;
      $79 = ((($64)) + 52|0);
      HEAP32[$79>>2] = $78;
      $80 = ((($64)) + 48|0);
      HEAP32[$80>>2] = $78;
      $81 = ((($64)) + 56|0);
      $82 = HEAP32[$81>>2]|0;
      $83 = ($82|0)==(0|0);
      $84 = $78;
      if ($83) {
       $90 = $84;
      } else {
       $85 = (FUNCTION_TABLE_iiii[$82 & 8191](0,0,0)|0);
       $86 = ((($64)) + 60|0);
       HEAP32[$86>>2] = $85;
       $87 = ((($stream)) + 48|0);
       HEAP32[$87>>2] = $85;
       $$pre$i$i11 = HEAP32[$77>>2]|0;
       $90 = $$pre$i$i11;
      }
      $88 = HEAP32[$8>>2]|0;
      $89 = HEAP32[$9>>2]|0;
      FUNCTION_TABLE_vii[$88 & 8191]($89,$90);
      $91 = HEAP32[$8>>2]|0;
      $92 = HEAP32[$9>>2]|0;
      $93 = ((($64)) + 36|0);
      $94 = HEAP32[$93>>2]|0;
      FUNCTION_TABLE_vii[$91 & 8191]($92,$94);
      $95 = HEAP32[$8>>2]|0;
      $96 = HEAP32[$9>>2]|0;
      FUNCTION_TABLE_vii[$95 & 8191]($96,$64);
      $$pre$i12 = HEAP32[$8>>2]|0;
      $$pre1$i14 = HEAP32[$58>>2]|0;
      $98 = $$pre$i12;$99 = $$pre1$i14;
     }
     $97 = HEAP32[$9>>2]|0;
     FUNCTION_TABLE_vii[$98 & 8191]($97,$99);
     HEAP32[$58>>2] = 0;
    }
   }
  } else {
   $14 = ((($stream)) + 28|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)==(0|0);
   if (!($16)) {
    $17 = HEAP32[$8>>2]|0;
    $18 = ($17|0)==(0|0);
    if (!($18)) {
     $19 = ((($15)) + 20|0);
     $20 = HEAP32[$19>>2]|0;
     $21 = ($20|0)==(0|0);
     if ($21) {
      $54 = $17;$55 = $15;
     } else {
      $22 = HEAP32[$20>>2]|0;
      $23 = $22 & -2;
      $switch$i$i$i = ($23|0)==(4);
      if ($switch$i$i$i) {
       $24 = HEAP32[$9>>2]|0;
       $25 = ((($20)) + 12|0);
       $26 = HEAP32[$25>>2]|0;
       FUNCTION_TABLE_vii[$17 & 8191]($24,$26);
       $$pr$i$i$i = HEAP32[$20>>2]|0;
       $28 = $$pr$i$i$i;
      } else {
       $28 = $22;
      }
      $27 = ($28|0)==(6);
      if ($27) {
       $29 = ((($20)) + 4|0);
       $30 = HEAP32[$29>>2]|0;
       $z$idx$val$i$i$i = HEAP32[$8>>2]|0;
       $z$idx1$val$i$i$i = HEAP32[$9>>2]|0;
       FUNCTION_TABLE_vii[$z$idx$val$i$i$i & 8191]($z$idx1$val$i$i$i,$30);
      }
      HEAP32[$20>>2] = 0;
      $31 = ((($20)) + 28|0);
      HEAP32[$31>>2] = 0;
      $32 = ((($20)) + 32|0);
      HEAP32[$32>>2] = 0;
      $33 = ((($20)) + 40|0);
      $34 = HEAP32[$33>>2]|0;
      $35 = ((($20)) + 52|0);
      HEAP32[$35>>2] = $34;
      $36 = ((($20)) + 48|0);
      HEAP32[$36>>2] = $34;
      $37 = ((($20)) + 56|0);
      $38 = HEAP32[$37>>2]|0;
      $39 = ($38|0)==(0|0);
      $40 = $34;
      if ($39) {
       $46 = $40;
      } else {
       $41 = (FUNCTION_TABLE_iiii[$38 & 8191](0,0,0)|0);
       $42 = ((($20)) + 60|0);
       HEAP32[$42>>2] = $41;
       $43 = ((($stream)) + 48|0);
       HEAP32[$43>>2] = $41;
       $$pre$i$i = HEAP32[$33>>2]|0;
       $46 = $$pre$i$i;
      }
      $44 = HEAP32[$8>>2]|0;
      $45 = HEAP32[$9>>2]|0;
      FUNCTION_TABLE_vii[$44 & 8191]($45,$46);
      $47 = HEAP32[$8>>2]|0;
      $48 = HEAP32[$9>>2]|0;
      $49 = ((($20)) + 36|0);
      $50 = HEAP32[$49>>2]|0;
      FUNCTION_TABLE_vii[$47 & 8191]($48,$50);
      $51 = HEAP32[$8>>2]|0;
      $52 = HEAP32[$9>>2]|0;
      FUNCTION_TABLE_vii[$51 & 8191]($52,$20);
      $$pre$i = HEAP32[$8>>2]|0;
      $$pre1$i = HEAP32[$14>>2]|0;
      $54 = $$pre$i;$55 = $$pre1$i;
     }
     $53 = HEAP32[$9>>2]|0;
     FUNCTION_TABLE_vii[$54 & 8191]($53,$55);
     HEAP32[$14>>2] = 0;
    }
   }
   $100 = ($12|0)==(0);
   $$ = $100 ? -5 : $12;
   switch ($$|0) {
   case -5:  {
    $$0 = 10;
    STACKTOP = sp;return ($$0|0);
    break;
   }
   case -3:  {
    $$0 = 8;
    STACKTOP = sp;return ($$0|0);
    break;
   }
   case -4:  {
    $$0 = 64;
    STACKTOP = sp;return ($$0|0);
    break;
   }
   default: {
    break L7;
   }
   }
  }
 } while(0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _FT_Stream_OpenLZW($stream,$source) {
 $stream = $stream|0;
 $source = $source|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $error = 0, $error$0$i$ph = 0, $error$0$i$ph$i = 0, $head$i = 0;
 var $or$cond = 0, $or$cond3 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $head$i = sp + 4|0;
 $error = sp;
 $0 = ($stream|0)!=(0|0);
 $1 = ($source|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  HEAP32[$error>>2] = 40;
  $53 = 40;
  STACKTOP = sp;return ($53|0);
 }
 $2 = ((($source)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_FT_Stream_Seek($source,0)|0);
 $5 = ($4|0)==(0);
 if ($5) {
  $6 = (_FT_Stream_Read($source,$head$i,2)|0);
  $7 = ($6|0)==(0);
  if ($7) {
   $8 = HEAP8[$head$i>>0]|0;
   $9 = ($8<<24>>24)==(31);
   if ($9) {
    $10 = ((($head$i)) + 1|0);
    $11 = HEAP8[$10>>0]|0;
    $12 = ($11<<24>>24)==(-99);
    if ($12) {
     HEAP32[$error>>2] = 0;
     dest=$stream; stop=dest+40|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
     $13 = ((($stream)) + 28|0);
     HEAP32[$13>>2] = $3;
     $14 = (_ft_mem_alloc($3,4292,$error)|0);
     $15 = HEAP32[$error>>2]|0;
     $16 = ($15|0)==(0);
     do {
      if ($16) {
       $17 = ((($14)) + 4|0);
       HEAP32[$17>>2] = $stream;
       HEAP32[$14>>2] = $source;
       $18 = HEAP32[$13>>2]|0;
       $19 = ((($14)) + 8|0);
       HEAP32[$19>>2] = $18;
       $20 = ((($14)) + 4280|0);
       $21 = ((($14)) + 4288|0);
       HEAP32[$21>>2] = $20;
       $22 = ((($14)) + 4284|0);
       HEAP32[$22>>2] = $20;
       HEAP32[$20>>2] = 0;
       $23 = (_FT_Stream_Seek($source,0)|0);
       $24 = ($23|0)==(0);
       if ($24) {
        $25 = (_FT_Stream_Read($source,$head$i,2)|0);
        $26 = ($25|0)==(0);
        if ($26) {
         $27 = HEAP8[$head$i>>0]|0;
         $28 = ($27<<24>>24)==(31);
         $29 = HEAP8[$10>>0]|0;
         $30 = ($29<<24>>24)==(-99);
         $or$cond3 = $28 & $30;
         if ($or$cond3) {
          $31 = ((($14)) + 12|0);
          _memset(($31|0),0,164)|0;
          $32 = ((($14)) + 176|0);
          HEAP32[$32>>2] = $source;
          $33 = HEAP32[$2>>2]|0;
          $34 = ((($14)) + 180|0);
          HEAP32[$34>>2] = $33;
          $35 = ((($14)) + 88|0);
          HEAP32[$35>>2] = 0;
          $36 = ((($14)) + 92|0);
          HEAP32[$36>>2] = 0;
          $37 = ((($14)) + 96|0);
          HEAP32[$37>>2] = 0;
          $38 = ((($14)) + 112|0);
          $39 = ((($14)) + 100|0);
          HEAP32[$39>>2] = $38;
          $40 = ((($14)) + 108|0);
          HEAP32[$40>>2] = 64;
          $41 = ((($14)) + 16|0);
          HEAP32[$41>>2] = 0;
          $42 = ((($14)) + 36|0);
          HEAP32[$42>>2] = 0;
          $43 = ((($14)) + 40|0);
          HEAP32[$43>>2] = 0;
          $44 = ((($14)) + 44|0);
          HEAP8[$44>>0] = 0;
          $45 = ((($14)) + 48|0);
          HEAP32[$45>>2] = 0;
          $46 = ((($14)) + 104|0);
          HEAP32[$46>>2] = 0;
          $47 = ((($14)) + 64|0);
          HEAP32[$47>>2] = 9;
          HEAP32[$31>>2] = 0;
          HEAP32[$error>>2] = 0;
          $48 = ((($stream)) + 12|0);
          HEAP32[$48>>2] = $14;
          $54 = 0;
          break;
         } else {
          $error$0$i$ph$i = 3;
         }
        } else {
         $error$0$i$ph$i = $25;
        }
       } else {
        $error$0$i$ph$i = $23;
       }
       HEAP32[$error>>2] = $error$0$i$ph$i;
       _ft_mem_free($3,$14);
       $$pre = HEAP32[$error>>2]|0;
       $53 = $$pre;
       STACKTOP = sp;return ($53|0);
      } else {
       $54 = $15;
      }
     } while(0);
     $49 = ((($stream)) + 4|0);
     HEAP32[$49>>2] = 2147483647;
     $50 = ((($stream)) + 8|0);
     HEAP32[$50>>2] = 0;
     HEAP32[$stream>>2] = 0;
     $51 = ((($stream)) + 20|0);
     HEAP32[$51>>2] = 4447;
     $52 = ((($stream)) + 24|0);
     HEAP32[$52>>2] = 4448;
     $53 = $54;
     STACKTOP = sp;return ($53|0);
    } else {
     $error$0$i$ph = 3;
    }
   } else {
    $error$0$i$ph = 3;
   }
  } else {
   $error$0$i$ph = $6;
  }
 } else {
  $error$0$i$ph = $4;
 }
 HEAP32[$error>>2] = $error$0$i$ph;
 $53 = $error$0$i$ph;
 STACKTOP = sp;return ($53|0);
}
function _ft_lzw_stream_io($stream,$pos,$buffer,$count) {
 $stream = $stream|0;
 $pos = $pos|0;
 $buffer = $buffer|0;
 $count = $count|0;
 var $$$02$i = 0, $$02$$i$i = 0, $$02$i = 0, $$02$i$i = 0, $$cast$i = 0, $$cast$i$i = 0, $$count$i$i = 0, $$old1$i = 0, $$old1$i$i = 0, $$pre$i = 0, $$pre$phi$iZ2D = 0, $$pre$phi11$iZ2D = 0, $$pre$phi13$iZ2D = 0, $$pre$phi9$iZ2D = 0, $$pre10$i = 0, $$pre6$i = 0, $$pre7$i = 0, $0 = 0, $1 = 0, $10 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $7 = 0, $8 = 0, $9 = 0, $not$$i = 0, $result$0$i = 0, $result$1$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 4280|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3>>>0)>($pos>>>0);
 do {
  if ($4) {
   $5 = (($3) - ($pos))|0;
   $6 = ((($1)) + 4284|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ((($1)) + 184|0);
   $9 = $7;
   $10 = $8;
   $11 = (($9) - ($10))|0;
   $12 = ($5>>>0)>($11>>>0);
   if (!($12)) {
    $13 = (0 - ($5))|0;
    $14 = (($7) + ($13)|0);
    HEAP32[$6>>2] = $14;
    HEAP32[$2>>2] = $pos;
    label = 12;
    break;
   }
   $15 = HEAP32[$1>>2]|0;
   $16 = (_FT_Stream_Seek($15,0)|0);
   $17 = ($16|0)==(0);
   if ($17) {
    $18 = ((($1)) + 16|0);
    HEAP32[$18>>2] = 0;
    $19 = ((($1)) + 36|0);
    HEAP32[$19>>2] = 0;
    $20 = ((($1)) + 40|0);
    HEAP32[$20>>2] = 0;
    $21 = ((($1)) + 44|0);
    HEAP8[$21>>0] = 0;
    $22 = ((($1)) + 48|0);
    HEAP32[$22>>2] = 0;
    $23 = ((($1)) + 104|0);
    HEAP32[$23>>2] = 0;
    $24 = ((($1)) + 64|0);
    HEAP32[$24>>2] = 9;
    $25 = ((($1)) + 12|0);
    HEAP32[$25>>2] = 0;
    $26 = ((($1)) + 4280|0);
    $27 = ((($1)) + 4288|0);
    HEAP32[$27>>2] = $26;
    HEAP32[$6>>2] = $26;
    HEAP32[$2>>2] = 0;
    $29 = 0;
    label = 6;
    break;
   } else {
    $result$1$i = 0;
    return ($result$1$i|0);
   }
  } else {
   $29 = $3;
   label = 6;
  }
 } while(0);
 if ((label|0) == 6) {
  $28 = ($29>>>0)<($pos>>>0);
  if ($28) {
   $30 = (($pos) - ($29))|0;
   $31 = ((($1)) + 4288|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = ((($1)) + 4284|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = (($32) - ($34))|0;
   $36 = ($35>>>0)<($30>>>0);
   $$count$i$i = $36 ? $35 : $30;
   $$cast$i$i = $34;
   $37 = (($$cast$i$i) + ($$count$i$i)|0);
   HEAP32[$33>>2] = $37;
   $38 = (($$count$i$i) + ($29))|0;
   HEAP32[$2>>2] = $38;
   $$old1$i$i = ($30|0)==($$count$i$i|0);
   L12: do {
    if (!($$old1$i$i)) {
     $39 = (($30) - ($$count$i$i))|0;
     $40 = ((($1)) + 12|0);
     $$02$i$i = $39;
     while(1) {
      $41 = ($$02$i$i>>>0)<(4096);
      $$02$$i$i = $41 ? $$02$i$i : 4096;
      $42 = (_ft_lzwstate_io($40,0,$$02$$i$i)|0);
      $43 = ($42>>>0)<($$02$$i$i>>>0);
      if ($43) {
       $result$1$i = 0;
       break;
      }
      $44 = HEAP32[$2>>2]|0;
      $45 = (($44) + ($$02$$i$i))|0;
      HEAP32[$2>>2] = $45;
      $46 = (($$02$i$i) - ($$02$$i$i))|0;
      $47 = ($$02$i$i|0)==($$02$$i$i|0);
      if ($47) {
       break L12;
      } else {
       $$02$i$i = $46;
      }
     }
     return ($result$1$i|0);
    }
   } while(0);
   $48 = ($count|0)==(0);
   if ($48) {
    $result$1$i = 0;
    return ($result$1$i|0);
   } else {
    $$pre$phi$iZ2D = $31;$$pre$phi11$iZ2D = $33;$$pre$phi13$iZ2D = $33;$$pre$phi9$iZ2D = $31;
   }
  } else {
   label = 12;
  }
 }
 do {
  if ((label|0) == 12) {
   $$old1$i = ($count|0)==(0);
   if ($$old1$i) {
    $result$1$i = 0;
    return ($result$1$i|0);
   } else {
    $$pre7$i = ((($1)) + 4288|0);
    $$pre10$i = ((($1)) + 4284|0);
    $$pre$phi$iZ2D = $$pre7$i;$$pre$phi11$iZ2D = $$pre10$i;$$pre$phi13$iZ2D = $$pre10$i;$$pre$phi9$iZ2D = $$pre7$i;
    break;
   }
  }
 } while(0);
 $49 = ((($1)) + 12|0);
 $50 = ((($1)) + 184|0);
 $$pre$i = HEAP32[$$pre$phi9$iZ2D>>2]|0;
 $$pre6$i = HEAP32[$$pre$phi13$iZ2D>>2]|0;
 $$02$i = $count;$52 = $$pre$i;$53 = $$pre6$i;$result$0$i = 0;
 while(1) {
  $51 = (($52) - ($53))|0;
  $54 = ($51>>>0)<($$02$i>>>0);
  $$$02$i = $54 ? $51 : $$02$i;
  $55 = (($buffer) + ($result$0$i)|0);
  $$cast$i = $53;
  _memcpy(($55|0),($$cast$i|0),($$$02$i|0))|0;
  $56 = (($$$02$i) + ($result$0$i))|0;
  $57 = HEAP32[$$pre$phi11$iZ2D>>2]|0;
  $58 = (($57) + ($$$02$i)|0);
  HEAP32[$$pre$phi11$iZ2D>>2] = $58;
  $59 = HEAP32[$2>>2]|0;
  $60 = (($59) + ($$$02$i))|0;
  HEAP32[$2>>2] = $60;
  $61 = ($$02$i|0)==($$$02$i|0);
  if ($61) {
   $result$1$i = $56;
   label = 17;
   break;
  }
  $62 = (($$02$i) - ($$$02$i))|0;
  HEAP32[$$pre$phi11$iZ2D>>2] = $50;
  $63 = (_ft_lzwstate_io($49,$50,4096)|0);
  $64 = HEAP32[$$pre$phi11$iZ2D>>2]|0;
  $65 = (($64) + ($63)|0);
  HEAP32[$$pre$phi$iZ2D>>2] = $65;
  $not$$i = ($63|0)==(0);
  $66 = $65;
  $67 = $64;
  if ($not$$i) {
   $result$1$i = $56;
   label = 17;
   break;
  } else {
   $$02$i = $62;$52 = $66;$53 = $67;$result$0$i = $56;
  }
 }
 if ((label|0) == 17) {
  return ($result$1$i|0);
 }
 return (0)|0;
}
function _ft_lzwstate_io($state,$buffer,$out_size) {
 $state = $state|0;
 $buffer = $buffer|0;
 $out_size = $out_size|0;
 var $$ = 0, $$lcssa161 = 0, $$lobit = 0, $$phi$trans$insert = 0, $$phi$trans$insert80 = 0, $$pre = 0, $$pre$phi87Z2D = 0, $$pre$phi89Z2D = 0, $$pre$phi91Z2D = 0, $$pre$phi93Z2D = 0, $$pre$phiZ2D = 0, $$pre74 = 0, $$pre75 = 0, $$pre76 = 0, $$pre78 = 0, $$pre79 = 0, $$pre81 = 0, $$pre82 = 0, $$pre83 = 0, $$pre86 = 0;
 var $$pre88 = 0, $$pre90 = 0, $$pre92 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $code$0$ph94 = 0, $code$045 = 0, $code$1 = 0, $error$i = 0, $error$i15 = 0, $error$i3 = 0, $error$i9 = 0, $in_code$0$ph = 0, $in_code$2$ph = 0, $in_code$4$ph = 0, $in_code$5 = 0;
 var $in_code$6 = 0, $max_bits = 0, $new_size$0$i = 0, $new_size$0$i11 = 0, $new_size$0$i16 = 0, $new_size$0$i5 = 0, $old_char$2$ph = 0, $old_char$246 = 0, $old_char$246$lcssa = 0, $old_char$246173 = 0, $old_char$246174 = 0, $old_char$246175 = 0, $old_char$4$ph = 0, $old_char$5 = 0, $old_char$6 = 0, $old_code$2$ph = 0, $old_code$247 = 0, $old_code$247$lcssa = 0, $old_code$247166 = 0, $old_code$247167 = 0;
 var $old_code$247168 = 0, $old_code$247169 = 0, $old_code$247170 = 0, $old_code$3$ph = 0, $old_code$4 = 0, $old_code$5 = 0, $old_size$0$i = 0, $old_size$0$i10 = 0, $old_size$0$i4 = 0, $result$2$ph = 0, $result$3 = 0, $result$3$lcssa = 0, $result$3$ph = 0, $result$3$us = 0, $result$4 = 0, $result$5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $error$i15 = sp + 12|0;
 $error$i9 = sp + 8|0;
 $error$i3 = sp + 4|0;
 $error$i = sp;
 $max_bits = sp + 16|0;
 $0 = ((($state)) + 68|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($state)) + 64|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($state)) + 72|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($out_size|0)==(0);
 if ($6) {
  $in_code$5 = $5;$old_char$5 = $1;$old_code$4 = $3;$result$4 = 0;
  HEAP32[$2>>2] = $old_code$4;
  HEAP32[$0>>2] = $old_char$5;
  HEAP32[$4>>2] = $in_code$5;
  STACKTOP = sp;return ($result$4|0);
 }
 $7 = HEAP32[$state>>2]|0;
 L4: do {
  switch ($7|0) {
  case 0:  {
   $8 = ((($state)) + 164|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (_FT_Stream_Seek($9,2)|0);
   $11 = ($10|0)==(0);
   if ($11) {
    $12 = HEAP32[$8>>2]|0;
    $13 = (_FT_Stream_TryRead($12,$max_bits,1)|0);
    $14 = ($13|0)==(1);
    if ($14) {
     $15 = HEAP8[$max_bits>>0]|0;
     $16 = $15&255;
     $17 = $16 & 31;
     $18 = ((($state)) + 40|0);
     HEAP32[$18>>2] = $17;
     $19 = $16 & 128;
     $20 = ((($state)) + 44|0);
     HEAP32[$20>>2] = $19;
     $21 = 1 << $17;
     $22 = (($21) + -256)|0;
     $23 = ((($state)) + 48|0);
     HEAP32[$23>>2] = $22;
     $24 = ($17>>>0)>(16);
     if ($24) {
      $in_code$0$ph = $5;
     } else {
      $25 = ((($state)) + 52|0);
      HEAP32[$25>>2] = 9;
      $$lobit = $16 >>> 7;
      $26 = ((($state)) + 56|0);
      HEAP32[$26>>2] = $$lobit;
      $27 = ($17>>>0)>(9);
      $28 = (($21) + -255)|0;
      $$ = $27 ? 256 : $28;
      $29 = ((($state)) + 60|0);
      HEAP32[$29>>2] = $$;
      $30 = (_ft_lzwstate_get_code($state)|0);
      $31 = ($30>>>0)>(255);
      if ($31) {
       $in_code$0$ph = 0;
      } else {
       $32 = ($buffer|0)==(0|0);
       if (!($32)) {
        $33 = $30&255;
        HEAP8[$buffer>>0] = $33;
       }
       $34 = ($out_size>>>0)<(2);
       if (!($34)) {
        HEAP32[$state>>2] = 1;
        $in_code$2$ph = 0;$old_char$2$ph = $30;$old_code$2$ph = $30;$result$2$ph = 1;
        label = 12;
        break L4;
       }
       $in_code$5 = 0;$old_char$5 = $30;$old_code$4 = $30;$result$4 = 1;
       HEAP32[$2>>2] = $old_code$4;
       HEAP32[$0>>2] = $old_char$5;
       HEAP32[$4>>2] = $in_code$5;
       STACKTOP = sp;return ($result$4|0);
      }
     }
    } else {
     $in_code$0$ph = $5;
    }
   } else {
    $in_code$0$ph = $5;
   }
   $in_code$6 = $in_code$0$ph;$old_char$6 = $1;$old_code$5 = $3;$result$5 = 0;
   label = 68;
   break;
  }
  case 1:  {
   $in_code$2$ph = $5;$old_char$2$ph = $1;$old_code$2$ph = $3;$result$2$ph = 0;
   label = 12;
   break;
  }
  case 2:  {
   $$pre86 = ((($state)) + 92|0);
   $$pre88 = ((($state)) + 88|0);
   $$pre$phi87Z2D = $$pre86;$$pre$phi89Z2D = $$pre88;$in_code$4$ph = $5;$old_char$4$ph = $1;$old_code$3$ph = $3;$result$3$ph = 0;
   label = 54;
   break;
  }
  default: {
   $in_code$5 = $5;$old_char$5 = $1;$old_code$4 = $3;$result$4 = 0;
   HEAP32[$2>>2] = $old_code$4;
   HEAP32[$0>>2] = $old_char$5;
   HEAP32[$4>>2] = $in_code$5;
   STACKTOP = sp;return ($result$4|0);
  }
  }
 } while(0);
 L21: while(1) {
  if ((label|0) == 12) {
   label = 0;
   $35 = (_ft_lzwstate_get_code($state)|0);
   $36 = ($35|0)<(0);
   if ($36) {
    $in_code$6 = $in_code$2$ph;$old_char$6 = $old_char$2$ph;$old_code$5 = $old_code$2$ph;$result$5 = $result$2$ph;
    label = 68;
    continue;
   }
   $37 = ((($state)) + 44|0);
   $38 = ((($state)) + 56|0);
   $39 = ((($state)) + 32|0);
   $41 = $35;$old_char$246 = $old_char$2$ph;$old_code$247 = $old_code$2$ph;
   while(1) {
    $40 = ($41|0)==(256);
    if (!($40)) {
     $$lcssa161 = $41;$old_char$246$lcssa = $old_char$246;$old_code$247$lcssa = $old_code$247;
     label = 19;
     break;
    }
    $42 = HEAP32[$37>>2]|0;
    $43 = ($42|0)==(0);
    if ($43) {
     $48 = 256;$old_char$246173 = $old_char$246;$old_code$247167 = $old_code$247;
     label = 20;
     break;
    }
    HEAP32[$38>>2] = 0;
    HEAP8[$39>>0] = 1;
    $44 = (_ft_lzwstate_get_code($state)|0);
    $45 = ($44|0)<(0);
    if ($45) {
     $in_code$6 = $in_code$2$ph;$old_char$6 = 0;$old_code$5 = 0;$result$5 = $result$2$ph;
     label = 68;
     continue L21;
    } else {
     $41 = $44;$old_char$246 = 0;$old_code$247 = 0;
    }
   }
   if ((label|0) == 19) {
    label = 0;
    $46 = ($$lcssa161>>>0)>(255);
    if ($46) {
     $48 = $$lcssa161;$old_char$246173 = $old_char$246$lcssa;$old_code$247167 = $old_code$247$lcssa;
     label = 20;
    } else {
     $185 = $$lcssa161;$code$1 = $$lcssa161;$old_code$247166 = $old_code$247$lcssa;
    }
   }
   L32: do {
    if ((label|0) == 20) {
     label = 0;
     $47 = (($48) + -256)|0;
     $49 = HEAP32[$38>>2]|0;
     $50 = ($47>>>0)<($49>>>0);
     if ($50) {
      $code$0$ph94 = $48;
     } else {
      $51 = ($47>>>0)>($49>>>0);
      if ($51) {
       $in_code$6 = $48;$old_char$6 = $old_char$246173;$old_code$5 = $old_code$247167;$result$5 = $result$2$ph;
       label = 68;
       continue L21;
      }
      $52 = ((($state)) + 92|0);
      $53 = HEAP32[$52>>2]|0;
      $54 = ((($state)) + 96|0);
      $55 = HEAP32[$54>>2]|0;
      $56 = ($53>>>0)<($55>>>0);
      do {
       if ($56) {
        $$phi$trans$insert = ((($state)) + 88|0);
        $$pre74 = HEAP32[$$phi$trans$insert>>2]|0;
        $74 = $53;$76 = $$pre74;
       } else {
        $57 = ((($state)) + 168|0);
        $58 = HEAP32[$57>>2]|0;
        $59 = $55 >>> 1;
        $60 = (($55) + 4)|0;
        $61 = (($60) + ($59))|0;
        $62 = ((($state)) + 88|0);
        $63 = HEAP32[$62>>2]|0;
        $64 = ((($state)) + 100|0);
        $65 = ($63|0)==($64|0);
        if ($65) {
         HEAP32[$62>>2] = 0;
         $68 = 0;$old_size$0$i = 0;
        } else {
         $68 = $63;$old_size$0$i = $55;
        }
        $66 = ($61>>>0)>(65536);
        if ($66) {
         $67 = ($old_size$0$i|0)==(65536);
         if ($67) {
          $old_char$246174 = $old_char$246173;$old_code$247169 = $old_code$247167;
         } else {
          $new_size$0$i = 65536;
          label = 28;
         }
        } else {
         $new_size$0$i = $61;
         label = 28;
        }
        if ((label|0) == 28) {
         label = 0;
         $69 = (_ft_mem_realloc($58,1,$old_size$0$i,$new_size$0$i,$68,$error$i)|0);
         HEAP32[$62>>2] = $69;
         $70 = HEAP32[$error$i>>2]|0;
         $71 = ($70|0)==(0);
         if ($71) {
          HEAP32[$54>>2] = $new_size$0$i;
          $$pre = HEAP32[$52>>2]|0;
          $74 = $$pre;$76 = $69;
          break;
         } else {
          $old_char$246174 = $old_char$246173;$old_code$247169 = $old_code$247167;
         }
        }
        $in_code$6 = $48;$old_char$6 = $old_char$246174;$old_code$5 = $old_code$247169;$result$5 = $result$2$ph;
        label = 68;
        continue L21;
       }
      } while(0);
      $72 = $old_char$246173&255;
      $73 = (($74) + 1)|0;
      HEAP32[$52>>2] = $73;
      $75 = (($76) + ($74)|0);
      HEAP8[$75>>0] = $72;
      $77 = ($old_code$247167>>>0)>(255);
      if ($77) {
       $code$0$ph94 = $old_code$247167;
      } else {
       $185 = $48;$code$1 = $old_code$247167;$old_code$247166 = $old_code$247167;
       break;
      }
     }
     $78 = ((($state)) + 76|0);
     $79 = ((($state)) + 92|0);
     $80 = ((($state)) + 96|0);
     $81 = ((($state)) + 80|0);
     $82 = ((($state)) + 88|0);
     $83 = ((($state)) + 168|0);
     $84 = ((($state)) + 100|0);
     $$pre75 = HEAP32[$78>>2]|0;
     $86 = $$pre75;$code$045 = $code$0$ph94;
     while(1) {
      $85 = ($86|0)==(0|0);
      if ($85) {
       $in_code$6 = $48;$old_char$6 = $old_char$246173;$old_code$5 = $old_code$247167;$result$5 = $result$2$ph;
       label = 68;
       continue L21;
      }
      $87 = HEAP32[$79>>2]|0;
      $88 = HEAP32[$80>>2]|0;
      $89 = ($87>>>0)<($88>>>0);
      if ($89) {
       $$pre78 = HEAP32[$82>>2]|0;
       $107 = $87;$109 = $$pre78;
      } else {
       $90 = HEAP32[$83>>2]|0;
       $91 = $88 >>> 1;
       $92 = (($88) + 4)|0;
       $93 = (($92) + ($91))|0;
       $94 = HEAP32[$82>>2]|0;
       $95 = ($94|0)==($84|0);
       if ($95) {
        HEAP32[$82>>2] = 0;
        $98 = 0;$old_size$0$i4 = 0;
       } else {
        $98 = $94;$old_size$0$i4 = $88;
       }
       $96 = ($93>>>0)>(65536);
       if ($96) {
        $97 = ($old_size$0$i4|0)==(65536);
        if ($97) {
         $old_char$246175 = $old_char$246173;$old_code$247170 = $old_code$247167;
         break;
        } else {
         $new_size$0$i5 = 65536;
        }
       } else {
        $new_size$0$i5 = $93;
       }
       $99 = (_ft_mem_realloc($90,1,$old_size$0$i4,$new_size$0$i5,$98,$error$i3)|0);
       HEAP32[$82>>2] = $99;
       $100 = HEAP32[$error$i3>>2]|0;
       $101 = ($100|0)==(0);
       if (!($101)) {
        $old_char$246175 = $old_char$246173;$old_code$247170 = $old_code$247167;
        break;
       }
       HEAP32[$80>>2] = $new_size$0$i5;
       $$pre76 = HEAP32[$79>>2]|0;
       $107 = $$pre76;$109 = $99;
      }
      $102 = (($code$045) + -256)|0;
      $103 = HEAP32[$81>>2]|0;
      $104 = (($103) + ($102)|0);
      $105 = HEAP8[$104>>0]|0;
      $106 = (($107) + 1)|0;
      HEAP32[$79>>2] = $106;
      $108 = (($109) + ($107)|0);
      HEAP8[$108>>0] = $105;
      $110 = HEAP32[$78>>2]|0;
      $111 = (($110) + ($102<<1)|0);
      $112 = HEAP16[$111>>1]|0;
      $113 = $112&65535;
      $114 = ($112&65535)>(255);
      if ($114) {
       $86 = $110;$code$045 = $113;
      } else {
       $185 = $48;$code$1 = $113;$old_code$247166 = $old_code$247167;
       break L32;
      }
     }
     $in_code$6 = $48;$old_char$6 = $old_char$246175;$old_code$5 = $old_code$247170;$result$5 = $result$2$ph;
     label = 68;
     continue L21;
    }
   } while(0);
   $115 = ((($state)) + 92|0);
   $116 = HEAP32[$115>>2]|0;
   $117 = ((($state)) + 96|0);
   $118 = HEAP32[$117>>2]|0;
   $119 = ($116>>>0)<($118>>>0);
   do {
    if ($119) {
     $$phi$trans$insert80 = ((($state)) + 88|0);
     $$pre81 = HEAP32[$$phi$trans$insert80>>2]|0;
     $$pre$phiZ2D = $$phi$trans$insert80;$137 = $116;$139 = $$pre81;
    } else {
     $120 = ((($state)) + 168|0);
     $121 = HEAP32[$120>>2]|0;
     $122 = $118 >>> 1;
     $123 = (($118) + 4)|0;
     $124 = (($123) + ($122))|0;
     $125 = ((($state)) + 88|0);
     $126 = HEAP32[$125>>2]|0;
     $127 = ((($state)) + 100|0);
     $128 = ($126|0)==($127|0);
     if ($128) {
      HEAP32[$125>>2] = 0;
      $131 = 0;$old_size$0$i10 = 0;
     } else {
      $131 = $126;$old_size$0$i10 = $118;
     }
     $129 = ($124>>>0)>(65536);
     if ($129) {
      $130 = ($old_size$0$i10|0)==(65536);
      if ($130) {
       $old_code$247168 = $old_code$247166;
      } else {
       $new_size$0$i11 = 65536;
       label = 50;
      }
     } else {
      $new_size$0$i11 = $124;
      label = 50;
     }
     if ((label|0) == 50) {
      label = 0;
      $132 = (_ft_mem_realloc($121,1,$old_size$0$i10,$new_size$0$i11,$131,$error$i9)|0);
      HEAP32[$125>>2] = $132;
      $133 = HEAP32[$error$i9>>2]|0;
      $134 = ($133|0)==(0);
      if ($134) {
       HEAP32[$117>>2] = $new_size$0$i11;
       $$pre79 = HEAP32[$115>>2]|0;
       $$pre$phiZ2D = $125;$137 = $$pre79;$139 = $132;
       break;
      } else {
       $old_code$247168 = $old_code$247166;
      }
     }
     $in_code$6 = $185;$old_char$6 = $code$1;$old_code$5 = $old_code$247168;$result$5 = $result$2$ph;
     label = 68;
     continue L21;
    }
   } while(0);
   $135 = $code$1&255;
   $136 = (($137) + 1)|0;
   HEAP32[$115>>2] = $136;
   $138 = (($139) + ($137)|0);
   HEAP8[$138>>0] = $135;
   HEAP32[$state>>2] = 2;
   $$pre$phi87Z2D = $115;$$pre$phi89Z2D = $$pre$phiZ2D;$in_code$4$ph = $185;$old_char$4$ph = $code$1;$old_code$3$ph = $old_code$247166;$result$3$ph = $result$2$ph;
   label = 54;
   continue;
  }
  else if ((label|0) == 54) {
   label = 0;
   $140 = ($buffer|0)==(0|0);
   L81: do {
    if ($140) {
     $$pre82 = HEAP32[$$pre$phi87Z2D>>2]|0;
     $142 = $$pre82;$result$3$us = $result$3$ph;
     while(1) {
      $141 = ($142|0)==(0);
      if ($141) {
       $result$3$lcssa = $result$3$us;
       break L81;
      }
      $143 = (($142) + -1)|0;
      HEAP32[$$pre$phi87Z2D>>2] = $143;
      $144 = (($result$3$us) + 1)|0;
      $145 = ($144|0)==($out_size|0);
      if ($145) {
       $in_code$5 = $in_code$4$ph;$old_char$5 = $old_char$4$ph;$old_code$4 = $old_code$3$ph;$result$4 = $out_size;
       label = 69;
       break L21;
      } else {
       $142 = $143;$result$3$us = $144;
      }
     }
    } else {
     $result$3 = $result$3$ph;
     while(1) {
      $146 = HEAP32[$$pre$phi87Z2D>>2]|0;
      $147 = ($146|0)==(0);
      if ($147) {
       $result$3$lcssa = $result$3;
       break L81;
      }
      $148 = (($146) + -1)|0;
      HEAP32[$$pre$phi87Z2D>>2] = $148;
      $149 = HEAP32[$$pre$phi89Z2D>>2]|0;
      $150 = (($149) + ($148)|0);
      $151 = HEAP8[$150>>0]|0;
      $152 = (($buffer) + ($result$3)|0);
      HEAP8[$152>>0] = $151;
      $153 = (($result$3) + 1)|0;
      $154 = ($153|0)==($out_size|0);
      if ($154) {
       $in_code$5 = $in_code$4$ph;$old_char$5 = $old_char$4$ph;$old_code$4 = $old_code$3$ph;$result$4 = $out_size;
       label = 69;
       break L21;
      } else {
       $result$3 = $153;
      }
     }
    }
   } while(0);
   $155 = ((($state)) + 56|0);
   $156 = HEAP32[$155>>2]|0;
   $157 = ((($state)) + 48|0);
   $158 = HEAP32[$157>>2]|0;
   $159 = ($156>>>0)<($158>>>0);
   if ($159) {
    $160 = ((($state)) + 84|0);
    $161 = HEAP32[$160>>2]|0;
    $162 = ($156>>>0)<($161>>>0);
    do {
     if ($162) {
      $$pre90 = ((($state)) + 76|0);
      $$pre92 = ((($state)) + 80|0);
      $$pre$phi91Z2D = $$pre90;$$pre$phi93Z2D = $$pre92;$179 = $156;
     } else {
      $163 = ((($state)) + 168|0);
      $164 = HEAP32[$163>>2]|0;
      $165 = ($161|0)==(0);
      $166 = $161 >>> 2;
      $167 = (($166) + ($161))|0;
      $new_size$0$i16 = $165 ? 512 : $167;
      $168 = ((($state)) + 76|0);
      $169 = HEAP32[$168>>2]|0;
      $170 = (_ft_mem_realloc($164,3,$161,$new_size$0$i16,$169,$error$i15)|0);
      HEAP32[$168>>2] = $170;
      $171 = HEAP32[$error$i15>>2]|0;
      $172 = ($171|0)==(0);
      if ($172) {
       $173 = (($170) + ($new_size$0$i16<<1)|0);
       $174 = ((($state)) + 80|0);
       HEAP32[$174>>2] = $173;
       $175 = (($170) + ($161<<1)|0);
       _memmove(($173|0),($175|0),($161|0))|0;
       HEAP32[$160>>2] = $new_size$0$i16;
       $$pre83 = HEAP32[$155>>2]|0;
       $$pre$phi91Z2D = $168;$$pre$phi93Z2D = $174;$179 = $$pre83;
       break;
      } else {
       $in_code$6 = $in_code$4$ph;$old_char$6 = $old_char$4$ph;$old_code$5 = $old_code$3$ph;$result$5 = $result$3$lcssa;
       label = 68;
       continue L21;
      }
     }
    } while(0);
    $176 = $old_code$3$ph&65535;
    $177 = HEAP32[$$pre$phi91Z2D>>2]|0;
    $178 = (($177) + ($179<<1)|0);
    HEAP16[$178>>1] = $176;
    $180 = $old_char$4$ph&255;
    $181 = HEAP32[$$pre$phi93Z2D>>2]|0;
    $182 = (($181) + ($179)|0);
    HEAP8[$182>>0] = $180;
    $183 = HEAP32[$155>>2]|0;
    $184 = (($183) + 1)|0;
    HEAP32[$155>>2] = $184;
   }
   HEAP32[$state>>2] = 1;
   $in_code$2$ph = $in_code$4$ph;$old_char$2$ph = $old_char$4$ph;$old_code$2$ph = $in_code$4$ph;$result$2$ph = $result$3$lcssa;
   label = 12;
   continue;
  }
  else if ((label|0) == 68) {
   label = 0;
   HEAP32[$state>>2] = 3;
   $in_code$5 = $in_code$6;$old_char$5 = $old_char$6;$old_code$4 = $old_code$5;$result$4 = $result$5;
   label = 69;
   break;
  }
 }
 if ((label|0) == 69) {
  HEAP32[$2>>2] = $old_code$4;
  HEAP32[$0>>2] = $old_char$5;
  HEAP32[$4>>2] = $in_code$5;
  STACKTOP = sp;return ($result$4|0);
 }
 return (0)|0;
}
function _ft_lzwstate_get_code($state) {
 $state = $state|0;
 var $$0 = 0, $$neg1$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $8 = 0, $9 = 0, $num_bits$0 = 0, $num_bits$2 = 0, $num_bits$3 = 0, $offset$0 = 0;
 var $offset$1 = 0, $p$0 = 0, $result$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($state)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($state)) + 24|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($state)) + 32|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ($5<<24>>24)==(0);
 if ($6) {
  $7 = ((($state)) + 28|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($3>>>0)<($8>>>0);
  if ($9) {
   $10 = ((($state)) + 56|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = ((($state)) + 60|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = ($11>>>0)<($13>>>0);
   if ($14) {
    $num_bits$2 = $1;$offset$0 = $3;
   } else {
    label = 4;
   }
  } else {
   label = 4;
  }
 } else {
  label = 4;
 }
 if ((label|0) == 4) {
  $15 = ((($state)) + 56|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ((($state)) + 60|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = ($16>>>0)<($18>>>0);
  if ($19) {
   $num_bits$0 = $1;
  } else {
   $20 = (($1) + 1)|0;
   HEAP32[$0>>2] = $20;
   $21 = ((($state)) + 40|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = ($20>>>0)<($22>>>0);
   if ($23) {
    $24 = 1 << $20;
    $25 = (($24) + -256)|0;
    $29 = $25;
   } else {
    $26 = ((($state)) + 48|0);
    $27 = HEAP32[$26>>2]|0;
    $28 = (($27) + 1)|0;
    $29 = $28;
   }
   HEAP32[$17>>2] = $29;
   $num_bits$0 = $20;
  }
  if ($6) {
   $36 = $num_bits$0;
  } else {
   HEAP32[$0>>2] = 9;
   HEAP32[$17>>2] = 256;
   HEAP8[$4>>0] = 0;
   $36 = 9;
  }
  $30 = ((($state)) + 4|0);
  $31 = HEAP32[$30>>2]|0;
  $32 = ($31|0)==(0);
  if (!($32)) {
   $$0 = -1;
   return ($$0|0);
  }
  $33 = ((($state)) + 164|0);
  $34 = HEAP32[$33>>2]|0;
  $35 = ((($state)) + 8|0);
  $37 = (_FT_Stream_TryRead($34,$35,$36)|0);
  $38 = ((($state)) + 28|0);
  $39 = ((($state)) + 36|0);
  $40 = HEAP32[$39>>2]|0;
  $41 = (($40) + ($37))|0;
  HEAP32[$39>>2] = $41;
  $42 = HEAP32[$0>>2]|0;
  $43 = ($37>>>0)<($42>>>0);
  $44 = $43&1;
  HEAP32[$30>>2] = $44;
  HEAP32[$2>>2] = 0;
  $45 = $37 << 3;
  $$neg1$i = $45 | 1;
  $46 = (($$neg1$i) - ($42))|0;
  HEAP32[$38>>2] = $46;
  $47 = ($37|0)==(0);
  if ($47) {
   $$0 = -1;
   return ($$0|0);
  } else {
   $num_bits$2 = $36;$offset$0 = 0;
  }
 }
 $48 = (($offset$0) + ($num_bits$2))|0;
 HEAP32[$2>>2] = $48;
 $49 = $offset$0 >>> 3;
 $50 = (((($state)) + 8|0) + ($49)|0);
 $51 = $offset$0 & 7;
 $52 = ((($50)) + 1|0);
 $53 = HEAP8[$50>>0]|0;
 $54 = $53&255;
 $55 = $54 >>> $51;
 $56 = (8 - ($51))|0;
 $57 = (($num_bits$2) - ($56))|0;
 $58 = ($57>>>0)>(7);
 if ($58) {
  $59 = ((($50)) + 2|0);
  $60 = HEAP8[$52>>0]|0;
  $61 = $60&255;
  $62 = $61 << $56;
  $63 = $62 | $55;
  $64 = (($56) + 8)|0;
  $65 = (($57) + -8)|0;
  $num_bits$3 = $65;$offset$1 = $64;$p$0 = $59;$result$0 = $63;
 } else {
  $num_bits$3 = $57;$offset$1 = $56;$p$0 = $52;$result$0 = $55;
 }
 $66 = ($num_bits$3|0)==(0);
 if ($66) {
  $$0 = $result$0;
  return ($$0|0);
 }
 $67 = HEAP8[$p$0>>0]|0;
 $68 = $67&255;
 $69 = 1 << $num_bits$3;
 $70 = (($69) + -1)|0;
 $71 = $68 & $70;
 $72 = $71 << $offset$1;
 $73 = $72 | $result$0;
 $$0 = $73;
 return ($$0|0);
}
function _ft_lzw_stream_close($stream) {
 $stream = $stream|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($stream)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 $3 = ((($stream)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($1)) + 180|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($1)) + 16|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($1)) + 36|0);
 HEAP32[$8>>2] = 0;
 $9 = ((($1)) + 40|0);
 HEAP32[$9>>2] = 0;
 $10 = ((($1)) + 44|0);
 HEAP8[$10>>0] = 0;
 $11 = ((($1)) + 48|0);
 HEAP32[$11>>2] = 0;
 $12 = ((($1)) + 104|0);
 HEAP32[$12>>2] = 0;
 $13 = ((($1)) + 64|0);
 HEAP32[$13>>2] = 9;
 $14 = ((($1)) + 12|0);
 HEAP32[$14>>2] = 0;
 $15 = ((($1)) + 100|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($1)) + 112|0);
 $18 = ($16|0)==($17|0);
 if (!($18)) {
  _ft_mem_free($6,$16);
  HEAP32[$15>>2] = 0;
 }
 $19 = ((($1)) + 88|0);
 $20 = HEAP32[$19>>2]|0;
 _ft_mem_free($6,$20);
 _memset(($1|0),0,184)|0;
 _ft_mem_free($4,$1);
 HEAP32[$0>>2] = 0;
 return;
}
function _floor0_unpack($vi,$opb) {
 $vi = $vi|0;
 $opb = $opb|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $j$01 = 0;
 var $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($vi)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_malloc(96)|0);
 $3 = (_oggpack_read($opb,8)|0);
 HEAP32[$2>>2] = $3;
 $4 = (_oggpack_read($opb,16)|0);
 $5 = ((($2)) + 4|0);
 HEAP32[$5>>2] = $4;
 $6 = (_oggpack_read($opb,16)|0);
 $7 = ((($2)) + 8|0);
 HEAP32[$7>>2] = $6;
 $8 = (_oggpack_read($opb,6)|0);
 $9 = ((($2)) + 12|0);
 HEAP32[$9>>2] = $8;
 $10 = (_oggpack_read($opb,8)|0);
 $11 = ((($2)) + 16|0);
 HEAP32[$11>>2] = $10;
 $12 = (_oggpack_read($opb,4)|0);
 $13 = (($12) + 1)|0;
 $14 = ((($2)) + 20|0);
 HEAP32[$14>>2] = $13;
 $15 = ($3|0)<(1);
 L1: do {
  if ($15) {
   $37 = ($2|0)==(0|0);
   if ($37) {
    $$0 = 0;
    return ($$0|0);
   }
  } else {
   $16 = ($4|0)<(1);
   if (!($16)) {
    $17 = ($6|0)<(1);
    $18 = ($12|0)<(0);
    $or$cond = $18 | $17;
    if (!($or$cond)) {
     $19 = ($12|0)>(-1);
     if (!($19)) {
      $$0 = $2;
      return ($$0|0);
     }
     $20 = ((($2)) + 24|0);
     $21 = ((($1)) + 24|0);
     $j$01 = 0;
     while(1) {
      $25 = (_oggpack_read($opb,8)|0);
      $26 = (($20) + ($j$01<<2)|0);
      HEAP32[$26>>2] = $25;
      $27 = ($25|0)<(0);
      if ($27) {
       break L1;
      }
      $28 = HEAP32[$21>>2]|0;
      $29 = ($25|0)<($28|0);
      if (!($29)) {
       break L1;
      }
      $30 = (((($1)) + 1824|0) + ($25<<2)|0);
      $31 = HEAP32[$30>>2]|0;
      $32 = ((($31)) + 12|0);
      $33 = HEAP32[$32>>2]|0;
      $34 = ($33|0)==(0);
      if ($34) {
       break L1;
      }
      $35 = HEAP32[$31>>2]|0;
      $36 = ($35|0)<(1);
      $24 = (($j$01) + 1)|0;
      if ($36) {
       break L1;
      }
      $22 = HEAP32[$14>>2]|0;
      $23 = ($24|0)<($22|0);
      if ($23) {
       $j$01 = $24;
      } else {
       $$0 = $2;
       break;
      }
     }
     return ($$0|0);
    }
   }
  }
 } while(0);
 _free($2);
 $$0 = 0;
 return ($$0|0);
}
function _floor0_look($vd,$i) {
 $vd = $vd|0;
 $i = $i|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_calloc(1,32)|0);
 $1 = HEAP32[$i>>2]|0;
 $2 = ((($0)) + 4|0);
 HEAP32[$2>>2] = $1;
 $3 = ((($i)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 HEAP32[$0>>2] = $4;
 $5 = ((($0)) + 20|0);
 HEAP32[$5>>2] = $i;
 $6 = (_calloc(2,4)|0);
 $7 = ((($0)) + 8|0);
 HEAP32[$7>>2] = $6;
 return ($0|0);
}
function _floor0_free_info($i) {
 $i = $i|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($i|0)==(0|0);
 if (!($0)) {
  _free($i);
 }
 return;
}
function _floor0_free_look($i) {
 $i = $i|0;
 var $$in = 0, $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($i|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($i)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
  $4 = HEAP32[$2>>2]|0;
  $5 = ($4|0)==(0|0);
  if ($5) {
   $$in = $2;
  } else {
   _free($4);
   $$pre = HEAP32[$1>>2]|0;
   $$in = $$pre;
  }
  $6 = ((($$in)) + 4|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)==(0|0);
  if ($8) {
   $9 = $$in;
  } else {
   _free($7);
   $$pre1 = HEAP32[$1>>2]|0;
   $9 = $$pre1;
  }
  _free($9);
 }
 _free($i);
 return;
}
function _floor0_inverse1($vb,$i) {
 $vb = $vb|0;
 $i = $i|0;
 var $$2 = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $j$08 = 0, $j$1$lcssa = 0, $j$15 = 0, $k$04 = 0, $last$09 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($i)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($vb)) + 4|0);
 $3 = ((($1)) + 12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (_oggpack_read($2,$4)|0);
 $6 = ($5|0)>(0);
 if (!($6)) {
  $$2 = 0;
  return ($$2|0);
 }
 $7 = HEAP32[$3>>2]|0;
 $8 = 1 << $7;
 $9 = (($8) + -1)|0;
 $10 = (+($5|0));
 $11 = (+($9|0));
 $12 = $10 / $11;
 $13 = ((($1)) + 16|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (+($14|0));
 $16 = $15 * $12;
 $17 = ((($1)) + 20|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = (_ov_ilog($18)|0);
 $20 = (_oggpack_read($2,$19)|0);
 $21 = ($20|0)==(-1);
 if ($21) {
  $$2 = 0;
  return ($$2|0);
 }
 $22 = HEAP32[$17>>2]|0;
 $23 = ($20|0)<($22|0);
 if (!($23)) {
  $$2 = 0;
  return ($$2|0);
 }
 $24 = ((($vb)) + 64|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = ((($25)) + 4|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ((($27)) + 28|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = ((($29)) + 2848|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = (((($1)) + 24|0) + ($20<<2)|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = (($31) + (($33*56)|0)|0);
 $35 = ((($i)) + 4|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = HEAP32[$34>>2]|0;
 $38 = (($37) + ($36))|0;
 $39 = $38 << 2;
 $40 = (($39) + 4)|0;
 $41 = (__vorbis_block_alloc($vb,$40)|0);
 $42 = HEAP32[$35>>2]|0;
 $43 = (_vorbis_book_decodev_set($34,$41,$2,$42)|0);
 $44 = ($43|0)==(-1);
 if ($44) {
  $$2 = 0;
  return ($$2|0);
 }
 $45 = HEAP32[$35>>2]|0;
 $46 = ($45|0)>(0);
 if ($46) {
  $j$08 = 0;$last$09 = 0.0;
  while(1) {
   $47 = ($j$08|0)<($45|0);
   L16: do {
    if ($47) {
     $48 = HEAP32[$34>>2]|0;
     $j$15 = $j$08;$k$04 = 0;
     while(1) {
      $49 = ($k$04|0)<($48|0);
      if (!($49)) {
       $j$1$lcssa = $j$15;
       break L16;
      }
      $50 = (($41) + ($j$15<<2)|0);
      $51 = +HEAPF32[$50>>2];
      $52 = $last$09 + $51;
      HEAPF32[$50>>2] = $52;
      $53 = (($k$04) + 1)|0;
      $54 = (($j$15) + 1)|0;
      $55 = ($54|0)<($45|0);
      if ($55) {
       $j$15 = $54;$k$04 = $53;
      } else {
       $j$1$lcssa = $54;
       break;
      }
     }
    } else {
     $j$1$lcssa = $j$08;
    }
   } while(0);
   $56 = (($j$1$lcssa) + -1)|0;
   $57 = (($41) + ($56<<2)|0);
   $58 = +HEAPF32[$57>>2];
   $59 = ($j$1$lcssa|0)<($45|0);
   if ($59) {
    $j$08 = $j$1$lcssa;$last$09 = $58;
   } else {
    break;
   }
  }
 }
 $60 = (($41) + ($45<<2)|0);
 HEAPF32[$60>>2] = $16;
 $$2 = $41;
 return ($$2|0);
}
function _ov_ilog($v) {
 $v = $v|0;
 var $$01 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $ret$0$lcssa = 0, $ret$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($v|0)==(0);
 if ($0) {
  $ret$0$lcssa = 0;
 } else {
  $$01 = $v;$ret$02 = 0;
  while(1) {
   $1 = $$01 >>> 1;
   $2 = (($ret$02) + 1)|0;
   $3 = ($1|0)==(0);
   if ($3) {
    $ret$0$lcssa = $2;
    break;
   } else {
    $$01 = $1;$ret$02 = $2;
   }
  }
 }
 return ($ret$0$lcssa|0);
}
function __vorbis_block_alloc($vb,$bytes) {
 $vb = $vb|0;
 $bytes = $bytes|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($bytes) + 7)|0;
 $1 = $0 & -8;
 $2 = ((($vb)) + 72|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + ($1))|0;
 $5 = ((($vb)) + 76|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($4|0)>($6|0);
 $8 = ((($vb)) + 68|0);
 $9 = HEAP32[$8>>2]|0;
 if (!($7)) {
  $21 = $9;$22 = $3;
  $20 = (($21) + ($22)|0);
  $23 = (($22) + ($1))|0;
  HEAP32[$2>>2] = $23;
  return ($20|0);
 }
 $10 = ($9|0)==(0|0);
 if (!($10)) {
  $11 = $9;
  $12 = (_malloc(8)|0);
  $13 = ((($vb)) + 80|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = (($14) + ($3))|0;
  HEAP32[$13>>2] = $15;
  $16 = ((($vb)) + 84|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = ((($12)) + 4|0);
  HEAP32[$18>>2] = $17;
  HEAP32[$12>>2] = $11;
  HEAP32[$16>>2] = $12;
 }
 HEAP32[$5>>2] = $1;
 $19 = (_malloc($1)|0);
 HEAP32[$8>>2] = $19;
 HEAP32[$2>>2] = 0;
 $21 = $19;$22 = 0;
 $20 = (($21) + ($22)|0);
 $23 = (($22) + ($1))|0;
 HEAP32[$2>>2] = $23;
 return ($20|0);
}
function _vorbis_book_decodev_set($book,$a,$b,$n) {
 $book = $book|0;
 $a = $a|0;
 $b = $b|0;
 $n = $n|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $i$08 = 0, $i$1$lcssa = 0, $i$15 = 0, $j$04 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($book)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(0);
 $3 = ($n|0)>(0);
 if (!($2)) {
  if (!($3)) {
   $$1 = 0;
   return ($$1|0);
  }
  $4 = $n << 2;
  _memset(($a|0),0,($4|0))|0;
  $$1 = 0;
  return ($$1|0);
 }
 if (!($3)) {
  $$1 = 0;
  return ($$1|0);
 }
 $5 = ((($book)) + 16|0);
 $i$08 = 0;
 while(1) {
  $7 = (_decode_packed_entry_number($book,$b)|0);
  $8 = ($7|0)==(-1);
  if ($8) {
   $$1 = -1;
   label = 11;
   break;
  }
  $9 = HEAP32[$5>>2]|0;
  $10 = HEAP32[$book>>2]|0;
  $11 = Math_imul($10, $7)|0;
  $12 = (($9) + ($11<<2)|0);
  $13 = ($i$08|0)<($n|0);
  L14: do {
   if ($13) {
    $i$15 = $i$08;$j$04 = 0;
    while(1) {
     $14 = ($j$04|0)<($10|0);
     if (!($14)) {
      $i$1$lcssa = $i$15;
      break L14;
     }
     $15 = (($j$04) + 1)|0;
     $16 = (($12) + ($j$04<<2)|0);
     $17 = HEAP32[$16>>2]|0;
     $18 = (($i$15) + 1)|0;
     $19 = (($a) + ($i$15<<2)|0);
     HEAP32[$19>>2] = $17;
     $20 = ($18|0)<($n|0);
     if ($20) {
      $i$15 = $18;$j$04 = $15;
     } else {
      $i$1$lcssa = $18;
      break;
     }
    }
   } else {
    $i$1$lcssa = $i$08;
   }
  } while(0);
  $6 = ($i$1$lcssa|0)<($n|0);
  if ($6) {
   $i$08 = $i$1$lcssa;
  } else {
   $$1 = 0;
   label = 11;
   break;
  }
 }
 if ((label|0) == 11) {
  return ($$1|0);
 }
 return (0)|0;
}
function _decode_packed_entry_number($book,$b) {
 $book = $book|0;
 $b = $b|0;
 var $$3 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $hi$1 = 0, $hi$25 = 0, $lo$1 = 0, $lo$2$lcssa = 0, $lo$26 = 0, $lok$0$lcssa = 0;
 var $read$0$lcssa = 0, $read$07 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($book)) + 40|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($book)) + 36|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_oggpack_look($b,$3)|0);
 $5 = ($4|0)>(-1);
 do {
  if ($5) {
   $6 = ((($book)) + 32|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = (($7) + ($4<<2)|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = ($9|0)<(0);
   if ($10) {
    $11 = $9 >>> 15;
    $12 = $11 & 32767;
    $13 = ((($book)) + 8|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = $9 & 32767;
    $16 = (($14) - ($15))|0;
    $hi$1 = $16;$lo$1 = $12;
    break;
   }
   $17 = (($9) + -1)|0;
   $18 = ((($book)) + 28|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = (($19) + ($17)|0);
   $21 = HEAP8[$20>>0]|0;
   $22 = $21 << 24 >> 24;
   _oggpack_adv($b,$22);
   $$3 = $17;
   return ($$3|0);
  } else {
   $23 = ((($book)) + 8|0);
   $24 = HEAP32[$23>>2]|0;
   $hi$1 = $24;$lo$1 = 0;
  }
 } while(0);
 $25 = (_oggpack_look($b,$1)|0);
 $26 = ($25|0)<(0);
 $27 = ($1|0)>(1);
 $28 = $26 & $27;
 if ($28) {
  $read$07 = $1;
  while(1) {
   $29 = (($read$07) + -1)|0;
   $30 = (_oggpack_look($b,$29)|0);
   $31 = ($30|0)<(0);
   $32 = ($29|0)>(1);
   $33 = $31 & $32;
   if ($33) {
    $read$07 = $29;
   } else {
    $$lcssa = $31;$lok$0$lcssa = $30;$read$0$lcssa = $29;
    break;
   }
  }
 } else {
  $$lcssa = $26;$lok$0$lcssa = $25;$read$0$lcssa = $1;
 }
 if ($$lcssa) {
  $$3 = -1;
  return ($$3|0);
 }
 $34 = (_llvm_bswap_i32(($lok$0$lcssa|0))|0);
 $35 = $34 >>> 4;
 $36 = $35 & 252645135;
 $37 = $34 << 4;
 $38 = $37 & -252645136;
 $39 = $36 | $38;
 $40 = $39 >>> 2;
 $41 = $40 & 858993459;
 $42 = $39 << 2;
 $43 = $42 & -858993460;
 $44 = $41 | $43;
 $45 = $44 >>> 1;
 $46 = $45 & 1431655765;
 $47 = $44 << 1;
 $48 = $47 & -1431655766;
 $49 = $46 | $48;
 $50 = (($hi$1) - ($lo$1))|0;
 $51 = ($50|0)>(1);
 if ($51) {
  $52 = ((($book)) + 20|0);
  $53 = HEAP32[$52>>2]|0;
  $55 = $50;$hi$25 = $hi$1;$lo$26 = $lo$1;
  while(1) {
   $54 = $55 >> 1;
   $56 = (($54) + ($lo$26))|0;
   $57 = (($53) + ($56<<2)|0);
   $58 = HEAP32[$57>>2]|0;
   $59 = ($58>>>0)>($49>>>0);
   $60 = $59 ? 0 : $54;
   $61 = (($60) + ($lo$26))|0;
   $62 = $59 ? $54 : 0;
   $63 = (($hi$25) - ($62))|0;
   $64 = (($63) - ($61))|0;
   $65 = ($64|0)>(1);
   if ($65) {
    $55 = $64;$hi$25 = $63;$lo$26 = $61;
   } else {
    $lo$2$lcssa = $61;
    break;
   }
  }
 } else {
  $lo$2$lcssa = $lo$1;
 }
 $66 = ((($book)) + 28|0);
 $67 = HEAP32[$66>>2]|0;
 $68 = (($67) + ($lo$2$lcssa)|0);
 $69 = HEAP8[$68>>0]|0;
 $70 = $69 << 24 >> 24;
 $71 = ($70|0)>($read$0$lcssa|0);
 if ($71) {
  _oggpack_adv($b,$read$0$lcssa);
  $$3 = -1;
  return ($$3|0);
 } else {
  _oggpack_adv($b,$70);
  $$3 = $lo$2$lcssa;
  return ($$3|0);
 }
 return (0)|0;
}
function _floor0_inverse2($vb,$i,$memo,$out) {
 $vb = $vb|0;
 $i = $i|0;
 $memo = $memo|0;
 $out = $out|0;
 var $$$i = 0, $$0 = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0;
 var $42 = 0.0, $43 = 0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0;
 var $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0.0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0.0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $j$0$lcssa$i = 0, $j$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($i)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($vb)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($i)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($5) + ($3<<2)|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 if ($8) {
  $9 = ((($vb)) + 64|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ((($10)) + 4|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ((($12)) + 28|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = (($14) + ($3<<2)|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = (($16|0) / 2)&-1;
  $18 = HEAP32[$i>>2]|0;
  $19 = ((($1)) + 4|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = (+($20|0));
  $22 = $21 * 0.5;
  $23 = $17 << 2;
  $24 = (($23) + 4)|0;
  $25 = (_malloc($24)|0);
  HEAP32[$6>>2] = $25;
  $26 = ($16|0)>(1);
  L3: do {
   if ($26) {
    $27 = (+($18|0));
    $28 = $22 * 7.3999998858198524E-4;
    $29 = $28;
    $30 = (+Math_atan((+$29)));
    $31 = $30 * 13.100000381469727;
    $32 = $22 * $22;
    $33 = $32 * 1.8499999754340024E-8;
    $34 = $33;
    $35 = (+Math_atan((+$34)));
    $36 = $35 * 2.2400000095367432;
    $37 = $31 + $36;
    $38 = $22 * 9.9999997473787516E-5;
    $39 = $38;
    $40 = $39 + $37;
    $41 = $27 / $40;
    $42 = $41;
    $43 = HEAP32[$19>>2]|0;
    $44 = (+($43|0));
    $45 = $44 * 0.5;
    $46 = (+($17|0));
    $47 = $45 / $46;
    $48 = $42;
    $49 = HEAP32[$4>>2]|0;
    $50 = (($49) + ($3<<2)|0);
    $51 = HEAP32[$50>>2]|0;
    $71 = $18;$j$01$i = 0;
    while(1) {
     $52 = (+($j$01$i|0));
     $53 = $47 * $52;
     $54 = $53 * 7.3999998858198524E-4;
     $55 = $54;
     $56 = (+Math_atan((+$55)));
     $57 = $56 * 13.100000381469727;
     $58 = $53 * $53;
     $59 = $58 * 1.8499999754340024E-8;
     $60 = $59;
     $61 = (+Math_atan((+$60)));
     $62 = $61 * 2.2400000095367432;
     $63 = $57 + $62;
     $64 = $53 * 9.9999997473787516E-5;
     $65 = $64;
     $66 = $63 + $65;
     $67 = $48 * $66;
     $68 = (+Math_floor((+$67)));
     $69 = (~~(($68)));
     $70 = ($69|0)<($71|0);
     $72 = (($71) + -1)|0;
     $$$i = $70 ? $69 : $72;
     $73 = (($51) + ($j$01$i<<2)|0);
     HEAP32[$73>>2] = $$$i;
     $74 = (($j$01$i) + 1)|0;
     $75 = ($74|0)<($17|0);
     if (!($75)) {
      $j$0$lcssa$i = $74;
      break L3;
     }
     $$pre$i = HEAP32[$i>>2]|0;
     $71 = $$pre$i;$j$01$i = $74;
    }
   } else {
    $j$0$lcssa$i = 0;
   }
  } while(0);
  $76 = HEAP32[$4>>2]|0;
  $77 = (($76) + ($3<<2)|0);
  $78 = HEAP32[$77>>2]|0;
  $79 = (($78) + ($j$0$lcssa$i<<2)|0);
  HEAP32[$79>>2] = -1;
  $80 = (((($i)) + 12|0) + ($3<<2)|0);
  HEAP32[$80>>2] = $17;
 }
 $81 = ($memo|0)==(0|0);
 if ($81) {
  $96 = ((($i)) + 12|0);
  $97 = (($96) + ($3<<2)|0);
  $98 = HEAP32[$97>>2]|0;
  $99 = $98 << 2;
  _memset(($out|0),0,($99|0))|0;
  $$0 = 0;
  return ($$0|0);
 } else {
  $82 = ((($i)) + 4|0);
  $83 = HEAP32[$82>>2]|0;
  $84 = (($memo) + ($83<<2)|0);
  $85 = +HEAPF32[$84>>2];
  $86 = HEAP32[$4>>2]|0;
  $87 = (($86) + ($3<<2)|0);
  $88 = HEAP32[$87>>2]|0;
  $89 = ((($i)) + 12|0);
  $90 = (($89) + ($3<<2)|0);
  $91 = HEAP32[$90>>2]|0;
  $92 = HEAP32[$i>>2]|0;
  $93 = ((($1)) + 16|0);
  $94 = HEAP32[$93>>2]|0;
  $95 = (+($94|0));
  _vorbis_lsp_to_curve($out,$88,$91,$92,$memo,$83,$85,$95);
  $$0 = 1;
  return ($$0|0);
 }
 return (0)|0;
}
function _vorbis_lsp_to_curve($curve,$map,$n,$ln,$lsp,$m,$amp,$ampoffset) {
 $curve = $curve|0;
 $map = $map|0;
 $n = $n|0;
 $ln = $ln|0;
 $lsp = $lsp|0;
 $m = $m|0;
 $amp = +$amp;
 $ampoffset = +$ampoffset;
 var $$lcssa = 0, $$lcssa$us = 0, $$lcssa37 = 0.0, $$lcssa38 = 0.0, $$pn = 0.0, $$pn$us = 0.0, $$pn1 = 0.0, $$pn1$us = 0.0, $0 = 0.0, $1 = 0.0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0, $104 = 0.0, $105 = 0.0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0.0, $113 = 0.0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0.0;
 var $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0;
 var $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0.0, $6 = 0.0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0.0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0;
 var $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $exitcond = 0, $i$013 = 0, $i$110 = 0, $i$110$us = 0, $j$02$us = 0, $p$03$us = 0.0, $p$1 = 0.0, $p$1$us = 0.0, $q$04$us = 0.0, $q$1 = 0.0, $q$1$us = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+($ln|0));
 $1 = 3.1415926535897931 / $0;
 $2 = $1;
 $3 = ($m|0)>(0);
 if ($3) {
  $i$013 = 0;
  while(1) {
   $69 = (($lsp) + ($i$013<<2)|0);
   $70 = +HEAPF32[$69>>2];
   $71 = $70;
   $72 = (+Math_cos((+$71)));
   $73 = $72 * 2.0;
   $74 = $73;
   HEAPF32[$69>>2] = $74;
   $75 = (($i$013) + 1)|0;
   $exitcond = ($75|0)==($m|0);
   if ($exitcond) {
    break;
   } else {
    $i$013 = $75;
   }
  }
 }
 $4 = ($n|0)>(0);
 if (!($4)) {
  return;
 }
 $5 = ($m|0)>(1);
 $6 = $amp;
 $7 = $ampoffset;
 if (!($5)) {
  $8 = ($m|0)==(1);
  $i$110 = 0;
  while(1) {
   $77 = (($map) + ($i$110<<2)|0);
   $78 = HEAP32[$77>>2]|0;
   $79 = (+($78|0));
   $80 = $2 * $79;
   $81 = $80;
   $82 = (+Math_cos((+$81)));
   $83 = $82 * 2.0;
   $84 = $83;
   if ($8) {
    $85 = +HEAPF32[$lsp>>2];
    $86 = $84 - $85;
    $87 = $86 * 0.5;
    $88 = $84 * $84;
    $89 = 4.0 - $88;
    $90 = $87 * $87;
    $$pn1 = $89;$q$1 = $90;
   } else {
    $91 = 2.0 - $84;
    $92 = $84 + 2.0;
    $93 = $92 * 0.5;
    $94 = $93 * 0.5;
    $$pn1 = $91;$q$1 = $94;
   }
   $$pn = $$pn1 * 0.5;
   $p$1 = $$pn * 0.5;
   $95 = $q$1 + $p$1;
   $96 = $95;
   $97 = (+Math_sqrt((+$96)));
   $98 = $6 / $97;
   $99 = $98 - $7;
   $100 = $99 * 0.1151292473077774;
   $101 = (+Math_exp((+$100)));
   $102 = $101;
   $103 = (($curve) + ($i$110<<2)|0);
   $104 = +HEAPF32[$103>>2];
   $105 = $104 * $102;
   HEAPF32[$103>>2] = $105;
   $106 = (($i$110) + 1)|0;
   $107 = (($map) + ($106<<2)|0);
   $108 = HEAP32[$107>>2]|0;
   $109 = ($108|0)==($78|0);
   if ($109) {
    $111 = $106;
    while(1) {
     $110 = (($curve) + ($111<<2)|0);
     $112 = +HEAPF32[$110>>2];
     $113 = $102 * $112;
     HEAPF32[$110>>2] = $113;
     $114 = (($111) + 1)|0;
     $115 = (($map) + ($114<<2)|0);
     $116 = HEAP32[$115>>2]|0;
     $117 = ($116|0)==($78|0);
     if ($117) {
      $111 = $114;
     } else {
      $$lcssa = $114;
      break;
     }
    }
   } else {
    $$lcssa = $106;
   }
   $76 = ($$lcssa|0)<($n|0);
   if ($76) {
    $i$110 = $$lcssa;
   } else {
    break;
   }
  }
  return;
 }
 $9 = (($m) + -2)|0;
 $10 = $9 & -2;
 $11 = (($10) + 3)|0;
 $12 = $9 | 1;
 $13 = ($11|0)==($m|0);
 $14 = (($12) + 1)|0;
 $15 = (($lsp) + ($14<<2)|0);
 $i$110$us = 0;
 while(1) {
  $16 = (($map) + ($i$110$us<<2)|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = (+($17|0));
  $19 = $2 * $18;
  $20 = $19;
  $21 = (+Math_cos((+$20)));
  $22 = $21 * 2.0;
  $23 = $22;
  $j$02$us = 1;$p$03$us = 0.5;$q$04$us = 0.5;
  while(1) {
   $57 = (($j$02$us) + -1)|0;
   $58 = (($lsp) + ($57<<2)|0);
   $59 = +HEAPF32[$58>>2];
   $60 = $23 - $59;
   $61 = $q$04$us * $60;
   $62 = (($lsp) + ($j$02$us<<2)|0);
   $63 = +HEAPF32[$62>>2];
   $64 = $23 - $63;
   $65 = $p$03$us * $64;
   $66 = (($j$02$us) + 2)|0;
   $67 = ($66|0)<($m|0);
   if ($67) {
    $j$02$us = $66;$p$03$us = $65;$q$04$us = $61;
   } else {
    $$lcssa37 = $61;$$lcssa38 = $65;
    break;
   }
  }
  if ($13) {
   $28 = +HEAPF32[$15>>2];
   $29 = $23 - $28;
   $30 = $$lcssa37 * $29;
   $31 = $23 * $23;
   $32 = 4.0 - $31;
   $33 = $30 * $30;
   $$pn1$us = $32;$q$1$us = $33;
  } else {
   $24 = 2.0 - $23;
   $25 = $23 + 2.0;
   $26 = $25 * $$lcssa37;
   $27 = $$lcssa37 * $26;
   $$pn1$us = $24;$q$1$us = $27;
  }
  $$pn$us = $$lcssa38 * $$pn1$us;
  $p$1$us = $$lcssa38 * $$pn$us;
  $34 = $q$1$us + $p$1$us;
  $35 = $34;
  $36 = (+Math_sqrt((+$35)));
  $37 = $6 / $36;
  $38 = $37 - $7;
  $39 = $38 * 0.1151292473077774;
  $40 = (+Math_exp((+$39)));
  $41 = $40;
  $42 = (($curve) + ($i$110$us<<2)|0);
  $43 = +HEAPF32[$42>>2];
  $44 = $43 * $41;
  HEAPF32[$42>>2] = $44;
  $45 = (($i$110$us) + 1)|0;
  $46 = (($map) + ($45<<2)|0);
  $47 = HEAP32[$46>>2]|0;
  $48 = ($47|0)==($17|0);
  if ($48) {
   $50 = $45;
   while(1) {
    $49 = (($curve) + ($50<<2)|0);
    $51 = +HEAPF32[$49>>2];
    $52 = $41 * $51;
    HEAPF32[$49>>2] = $52;
    $53 = (($50) + 1)|0;
    $54 = (($map) + ($53<<2)|0);
    $55 = HEAP32[$54>>2]|0;
    $56 = ($55|0)==($17|0);
    if ($56) {
     $50 = $53;
    } else {
     $$lcssa$us = $53;
     break;
    }
   }
  } else {
   $$lcssa$us = $45;
  }
  $68 = ($$lcssa$us|0)<($n|0);
  if ($68) {
   $i$110$us = $$lcssa$us;
  } else {
   break;
  }
 }
 return;
}
function _floor1_pack($i,$opb) {
 $i = $i|0;
 $opb = $opb|0;
 var $$maxclass$0 = 0, $$maxclass$0$lcssa = 0, $$pre = 0, $$pre22 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $7 = 0, $8 = 0, $9 = 0, $count$03 = 0, $exitcond = 0, $exitcond21 = 0, $j$016 = 0, $j$112 = 0, $j$25 = 0, $k$09 = 0, $k$14 = 0, $k$2$lcssa = 0, $k$22 = 0, $maxclass$017 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($i)) + 836|0);
 $1 = ((($i)) + 840|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = HEAP32[$i>>2]|0;
 _oggpack_write($opb,$3,5);
 $4 = HEAP32[$i>>2]|0;
 $5 = ($4|0)>(0);
 if ($5) {
  $6 = ((($i)) + 4|0);
  $j$016 = 0;$maxclass$017 = -1;
  while(1) {
   $12 = (($6) + ($j$016<<2)|0);
   $13 = HEAP32[$12>>2]|0;
   _oggpack_write($opb,$13,4);
   $14 = HEAP32[$12>>2]|0;
   $15 = ($maxclass$017|0)<($14|0);
   $$maxclass$0 = $15 ? $14 : $maxclass$017;
   $16 = (($j$016) + 1)|0;
   $17 = HEAP32[$i>>2]|0;
   $18 = ($16|0)<($17|0);
   if ($18) {
    $j$016 = $16;$maxclass$017 = $$maxclass$0;
   } else {
    $$maxclass$0$lcssa = $$maxclass$0;
    break;
   }
  }
  $7 = ($$maxclass$0$lcssa|0)<(0);
  if (!($7)) {
   $8 = ((($i)) + 128|0);
   $9 = ((($i)) + 192|0);
   $10 = ((($i)) + 256|0);
   $11 = ((($i)) + 320|0);
   $j$112 = 0;
   while(1) {
    $19 = (($8) + ($j$112<<2)|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = (($20) + -1)|0;
    _oggpack_write($opb,$21,3);
    $22 = (($9) + ($j$112<<2)|0);
    $23 = HEAP32[$22>>2]|0;
    _oggpack_write($opb,$23,2);
    $24 = HEAP32[$22>>2]|0;
    $25 = ($24|0)==(0);
    if ($25) {
     $k$09 = 0;
     label = 8;
    } else {
     $26 = (($10) + ($j$112<<2)|0);
     $27 = HEAP32[$26>>2]|0;
     _oggpack_write($opb,$27,8);
     $$pre = HEAP32[$22>>2]|0;
     $28 = ($$pre|0)==(31);
     if (!($28)) {
      $k$09 = 0;
      label = 8;
     }
    }
    if ((label|0) == 8) {
     while(1) {
      label = 0;
      $29 = ((($11) + ($j$112<<5)|0) + ($k$09<<2)|0);
      $30 = HEAP32[$29>>2]|0;
      $31 = (($30) + 1)|0;
      _oggpack_write($opb,$31,8);
      $32 = (($k$09) + 1)|0;
      $33 = HEAP32[$22>>2]|0;
      $34 = 1 << $33;
      $35 = ($32|0)<($34|0);
      if ($35) {
       $k$09 = $32;
       label = 8;
      } else {
       break;
      }
     }
    }
    $36 = (($j$112) + 1)|0;
    $exitcond21 = ($j$112|0)==($$maxclass$0$lcssa|0);
    if ($exitcond21) {
     break;
    } else {
     $j$112 = $36;
    }
   }
  }
 }
 $37 = ((($i)) + 832|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = (($38) + -1)|0;
 _oggpack_write($opb,$39,2);
 $40 = (($2) + -1)|0;
 $41 = (_ov_ilog($40)|0);
 _oggpack_write($opb,$41,4);
 $42 = (_ov_ilog($40)|0);
 $43 = HEAP32[$i>>2]|0;
 $44 = ($43|0)>(0);
 if (!($44)) {
  return;
 }
 $45 = ((($i)) + 4|0);
 $46 = ((($i)) + 128|0);
 $60 = $43;$count$03 = 0;$j$25 = 0;$k$14 = 0;
 while(1) {
  $47 = (($45) + ($j$25<<2)|0);
  $48 = HEAP32[$47>>2]|0;
  $49 = (($46) + ($48<<2)|0);
  $50 = HEAP32[$49>>2]|0;
  $51 = (($50) + ($count$03))|0;
  $52 = ($k$14|0)<($51|0);
  if ($52) {
   $k$22 = $k$14;
   while(1) {
    $53 = (($k$22) + 2)|0;
    $54 = (($0) + ($53<<2)|0);
    $55 = HEAP32[$54>>2]|0;
    _oggpack_write($opb,$55,$42);
    $56 = (($k$22) + 1)|0;
    $exitcond = ($56|0)==($51|0);
    if ($exitcond) {
     break;
    } else {
     $k$22 = $56;
    }
   }
   $$pre22 = HEAP32[$i>>2]|0;
   $59 = $$pre22;$k$2$lcssa = $51;
  } else {
   $59 = $60;$k$2$lcssa = $k$14;
  }
  $57 = (($j$25) + 1)|0;
  $58 = ($57|0)<($59|0);
  if ($58) {
   $60 = $59;$count$03 = $51;$j$25 = $57;$k$14 = $k$2$lcssa;
  } else {
   break;
  }
 }
 return;
}
function _floor1_unpack($vi,$opb) {
 $vi = $vi|0;
 $opb = $opb|0;
 var $$01 = 0, $$lcssa = 0, $$maxclass$0 = 0, $$maxclass$0$lcssa = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phi56Z2D = 0, $$pre$phi58Z2D = 0, $$pre51 = 0, $$pre53 = 0, $$pre54 = 0, $$pre55 = 0, $$pre57 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $count$0$lcssa = 0, $count$025 = 0, $j$039 = 0, $j$134 = 0, $j$227 = 0, $j$317 = 0, $j$416 = 0, $k$031 = 0, $k$126 = 0, $k$2$lcssa = 0, $k$221 = 0, $maxclass$038 = 0, $or$cond = 0, $sortpointer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 272|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $sortpointer = sp;
 $0 = ((($vi)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_calloc(1,1120)|0);
 $3 = (_oggpack_read($opb,5)|0);
 HEAP32[$2>>2] = $3;
 $4 = ($3|0)>(0);
 L1: do {
  if ($4) {
   $5 = ((($2)) + 4|0);
   $j$039 = 0;$maxclass$038 = -1;
   while(1) {
    $12 = (_oggpack_read($opb,4)|0);
    $13 = (($5) + ($j$039<<2)|0);
    HEAP32[$13>>2] = $12;
    $14 = ($12|0)<(0);
    if ($14) {
     break L1;
    }
    $15 = ($maxclass$038|0)<($12|0);
    $$maxclass$0 = $15 ? $12 : $maxclass$038;
    $16 = (($j$039) + 1)|0;
    $17 = HEAP32[$2>>2]|0;
    $18 = ($16|0)<($17|0);
    if ($18) {
     $j$039 = $16;$maxclass$038 = $$maxclass$0;
    } else {
     $$maxclass$0$lcssa = $$maxclass$0;
     break;
    }
   }
   $6 = ($$maxclass$0$lcssa|0)<(0);
   if ($6) {
    label = 18;
   } else {
    $7 = ((($2)) + 128|0);
    $8 = ((($2)) + 192|0);
    $9 = ((($2)) + 256|0);
    $10 = ((($1)) + 24|0);
    $11 = ((($2)) + 320|0);
    $j$134 = 0;
    while(1) {
     $19 = (_oggpack_read($opb,3)|0);
     $20 = (($19) + 1)|0;
     $21 = (($7) + ($j$134<<2)|0);
     HEAP32[$21>>2] = $20;
     $22 = (_oggpack_read($opb,2)|0);
     $23 = (($8) + ($j$134<<2)|0);
     HEAP32[$23>>2] = $22;
     $24 = ($22|0)<(0);
     if ($24) {
      break L1;
     }
     $25 = ($22|0)==(0);
     if ($25) {
      $$phi$trans$insert = (($9) + ($j$134<<2)|0);
      $$pre = HEAP32[$$phi$trans$insert>>2]|0;
      $29 = $$pre;
     } else {
      $26 = (_oggpack_read($opb,8)|0);
      $27 = (($9) + ($j$134<<2)|0);
      HEAP32[$27>>2] = $26;
      $29 = $26;
     }
     $28 = ($29|0)<(0);
     if ($28) {
      break L1;
     }
     $30 = HEAP32[$10>>2]|0;
     $31 = ($29|0)<($30|0);
     if (!($31)) {
      break L1;
     }
     $32 = HEAP32[$23>>2]|0;
     $33 = ($32|0)==(31);
     if (!($33)) {
      $k$031 = 0;
      while(1) {
       $38 = (_oggpack_read($opb,8)|0);
       $39 = (($38) + -1)|0;
       $40 = ((($11) + ($j$134<<5)|0) + ($k$031<<2)|0);
       HEAP32[$40>>2] = $39;
       $41 = ($38|0)<(0);
       if ($41) {
        break L1;
       }
       $42 = HEAP32[$10>>2]|0;
       $43 = ($38|0)>($42|0);
       $37 = (($k$031) + 1)|0;
       if ($43) {
        break L1;
       }
       $34 = HEAP32[$23>>2]|0;
       $35 = 1 << $34;
       $36 = ($37|0)<($35|0);
       if ($36) {
        $k$031 = $37;
       } else {
        break;
       }
      }
     }
     $44 = (($j$134) + 1)|0;
     $45 = ($j$134|0)<($$maxclass$0$lcssa|0);
     if ($45) {
      $j$134 = $44;
     } else {
      label = 18;
      break;
     }
    }
   }
  } else {
   label = 18;
  }
 } while(0);
 L23: do {
  if ((label|0) == 18) {
   $46 = (_oggpack_read($opb,2)|0);
   $47 = (($46) + 1)|0;
   $48 = ((($2)) + 832|0);
   HEAP32[$48>>2] = $47;
   $49 = (_oggpack_read($opb,4)|0);
   $50 = ($49|0)<(0);
   if (!($50)) {
    $51 = HEAP32[$2>>2]|0;
    $52 = ($51|0)>(0);
    if ($52) {
     $53 = ((($2)) + 4|0);
     $54 = ((($2)) + 128|0);
     $55 = ((($2)) + 836|0);
     $56 = 1 << $49;
     $90 = $51;$count$025 = 0;$j$227 = 0;$k$126 = 0;
     while(1) {
      $57 = (($53) + ($j$227<<2)|0);
      $58 = HEAP32[$57>>2]|0;
      $59 = (($54) + ($58<<2)|0);
      $60 = HEAP32[$59>>2]|0;
      $61 = (($60) + ($count$025))|0;
      $62 = ($61|0)>(63);
      if ($62) {
       break L23;
      }
      $63 = ($k$126|0)<($61|0);
      if ($63) {
       $k$221 = $k$126;
       while(1) {
        $64 = (_oggpack_read($opb,$49)|0);
        $65 = (($k$221) + 2)|0;
        $66 = (($55) + ($65<<2)|0);
        HEAP32[$66>>2] = $64;
        $67 = ($64|0)>(-1);
        $68 = ($64|0)<($56|0);
        $or$cond = $67 & $68;
        if (!($or$cond)) {
         break L23;
        }
        $69 = (($k$221) + 1)|0;
        $70 = ($69|0)<($61|0);
        if ($70) {
         $k$221 = $69;
        } else {
         $$lcssa = $69;
         break;
        }
       }
       $$pre51 = HEAP32[$2>>2]|0;
       $73 = $$pre51;$k$2$lcssa = $$lcssa;
      } else {
       $73 = $90;$k$2$lcssa = $k$126;
      }
      $71 = (($j$227) + 1)|0;
      $72 = ($71|0)<($73|0);
      if ($72) {
       $90 = $73;$count$025 = $61;$j$227 = $71;$k$126 = $k$2$lcssa;
      } else {
       $$pre$phi56Z2D = $55;$$pre$phi58Z2D = $56;$count$0$lcssa = $61;
       break;
      }
     }
    } else {
     $$pre55 = ((($2)) + 836|0);
     $$pre57 = 1 << $49;
     $$pre$phi56Z2D = $$pre55;$$pre$phi58Z2D = $$pre57;$count$0$lcssa = 0;
    }
    HEAP32[$$pre$phi56Z2D>>2] = 0;
    $74 = ((($2)) + 840|0);
    HEAP32[$74>>2] = $$pre$phi58Z2D;
    $75 = (($count$0$lcssa) + 2)|0;
    $76 = ($count$0$lcssa|0)>(-2);
    if ($76) {
     $j$317 = 0;
     while(1) {
      $77 = (($$pre$phi56Z2D) + ($j$317<<2)|0);
      $78 = (($sortpointer) + ($j$317<<2)|0);
      HEAP32[$78>>2] = $77;
      $79 = (($j$317) + 1)|0;
      $80 = ($79|0)<($75|0);
      if ($80) {
       $j$317 = $79;
      } else {
       break;
      }
     }
    }
    _qsort($sortpointer,$75,4,4449);
    $81 = ($75|0)>(1);
    L43: do {
     if ($81) {
      $$pre53 = HEAP32[$sortpointer>>2]|0;
      $$pre54 = HEAP32[$$pre53>>2]|0;
      $88 = $$pre54;$j$416 = 1;
      while(1) {
       $84 = (($sortpointer) + ($j$416<<2)|0);
       $85 = HEAP32[$84>>2]|0;
       $86 = HEAP32[$85>>2]|0;
       $87 = ($88|0)==($86|0);
       $83 = (($j$416) + 1)|0;
       if ($87) {
        break;
       }
       $82 = ($83|0)<($75|0);
       if ($82) {
        $88 = $86;$j$416 = $83;
       } else {
        break L43;
       }
      }
      $89 = ($2|0)==(0|0);
      if ($89) {
       $$01 = 0;
      } else {
       break L23;
      }
      STACKTOP = sp;return ($$01|0);
     }
    } while(0);
    $$01 = $2;
    STACKTOP = sp;return ($$01|0);
   }
  }
 } while(0);
 _free($2);
 $$01 = 0;
 STACKTOP = sp;return ($$01|0);
}
function _icomp($a,$b) {
 $a = $a|0;
 $b = $b|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$a>>2]|0;
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$b>>2]|0;
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) - ($3))|0;
 return ($4|0);
}
function _floor1_look($vd,$in) {
 $vd = $vd|0;
 $in = $in|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0, $exitcond = 0, $exitcond31 = 0, $hi$05 = 0, $hi$1 = 0, $hi$1$lcssa = 0, $hx$07 = 0;
 var $hx$1 = 0, $i$026 = 0, $i$121 = 0, $i$218 = 0, $i$316 = 0, $i$413 = 0, $i$510 = 0, $indvars$iv = 0, $indvars$iv$next = 0, $j$08 = 0, $lo$04 = 0, $lo$1 = 0, $lo$1$lcssa = 0, $lx$06 = 0, $lx$1 = 0, $n$0$lcssa32343639 = 0, $n$0$lcssa33 = 0, $n$025 = 0, $or$cond = 0, $or$cond1 = 0;
 var $sortpointer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 272|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $sortpointer = sp;
 $0 = (_calloc(1,1312)|0);
 $1 = ((($0)) + 1296|0);
 HEAP32[$1>>2] = $in;
 $2 = ((($in)) + 836|0);
 $3 = ((($in)) + 840|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($0)) + 1288|0);
 HEAP32[$5>>2] = $4;
 $6 = HEAP32[$in>>2]|0;
 $7 = ($6|0)>(0);
 if ($7) {
  $9 = ((($in)) + 4|0);
  $10 = ((($in)) + 128|0);
  $i$026 = 0;$n$025 = 0;
  while(1) {
   $11 = (($9) + ($i$026<<2)|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = (($10) + ($12<<2)|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = (($14) + ($n$025))|0;
   $16 = (($i$026) + 1)|0;
   $17 = ($16|0)<($6|0);
   if ($17) {
    $i$026 = $16;$n$025 = $15;
   } else {
    $$lcssa = $15;
    break;
   }
  }
  $18 = (($$lcssa) + 2)|0;
  $19 = ((($0)) + 1284|0);
  HEAP32[$19>>2] = $18;
  $20 = ($$lcssa|0)>(-2);
  if ($20) {
   $25 = $18;$n$0$lcssa33 = $$lcssa;
   label = 7;
  } else {
   _qsort($sortpointer,$18,4,4449);
   $n$0$lcssa32343639 = $$lcssa;
  }
 } else {
  $8 = ((($0)) + 1284|0);
  HEAP32[$8>>2] = 2;
  $25 = 2;$n$0$lcssa33 = 0;
  label = 7;
 }
 if ((label|0) == 7) {
  $i$121 = 0;
  while(1) {
   $21 = (($2) + ($i$121<<2)|0);
   $22 = (($sortpointer) + ($i$121<<2)|0);
   HEAP32[$22>>2] = $21;
   $23 = (($i$121) + 1)|0;
   $24 = ($23|0)<($25|0);
   if ($24) {
    $i$121 = $23;
   } else {
    break;
   }
  }
  _qsort($sortpointer,$25,4,4449);
  $26 = $2;
  $27 = ((($0)) + 260|0);
  $i$218 = 0;
  while(1) {
   $30 = (($sortpointer) + ($i$218<<2)|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = (($31) - ($26))|0;
   $33 = $32 >> 2;
   $34 = (($27) + ($i$218<<2)|0);
   HEAP32[$34>>2] = $33;
   $35 = (($i$218) + 1)|0;
   $36 = ($35|0)<($25|0);
   if ($36) {
    $i$218 = $35;
   } else {
    break;
   }
  }
  $28 = ((($0)) + 260|0);
  $29 = ((($0)) + 520|0);
  $i$316 = 0;
  while(1) {
   $38 = (($28) + ($i$316<<2)|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = (($29) + ($39<<2)|0);
   HEAP32[$40>>2] = $i$316;
   $41 = (($i$316) + 1)|0;
   $42 = ($41|0)<($25|0);
   if ($42) {
    $i$316 = $41;
   } else {
    break;
   }
  }
  $37 = ((($0)) + 260|0);
  $i$413 = 0;
  while(1) {
   $43 = (($37) + ($i$413<<2)|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = (($2) + ($44<<2)|0);
   $46 = HEAP32[$45>>2]|0;
   $47 = (($0) + ($i$413<<2)|0);
   HEAP32[$47>>2] = $46;
   $48 = (($i$413) + 1)|0;
   $49 = ($48|0)<($25|0);
   if ($49) {
    $i$413 = $48;
   } else {
    $n$0$lcssa32343639 = $n$0$lcssa33;
    break;
   }
  }
 }
 $50 = ((($in)) + 832|0);
 $51 = HEAP32[$50>>2]|0;
 switch ($51|0) {
 case 1:  {
  $52 = ((($0)) + 1292|0);
  HEAP32[$52>>2] = 256;
  break;
 }
 case 2:  {
  $53 = ((($0)) + 1292|0);
  HEAP32[$53>>2] = 128;
  break;
 }
 case 3:  {
  $54 = ((($0)) + 1292|0);
  HEAP32[$54>>2] = 86;
  break;
 }
 case 4:  {
  $55 = ((($0)) + 1292|0);
  HEAP32[$55>>2] = 64;
  break;
 }
 default: {
 }
 }
 $56 = ($n$0$lcssa32343639|0)>(0);
 if (!($56)) {
  STACKTOP = sp;return ($0|0);
 }
 $57 = ((($0)) + 1032|0);
 $58 = ((($0)) + 780|0);
 $i$510 = 0;$indvars$iv = 2;
 while(1) {
  $59 = (($i$510) + 2)|0;
  $60 = (($2) + ($59<<2)|0);
  $61 = HEAP32[$60>>2]|0;
  $62 = HEAP32[$5>>2]|0;
  $hi$05 = 1;$hx$07 = $62;$j$08 = 0;$lo$04 = 0;$lx$06 = 0;
  while(1) {
   $63 = (($2) + ($j$08<<2)|0);
   $64 = HEAP32[$63>>2]|0;
   $65 = ($64|0)>($lx$06|0);
   $66 = ($64|0)<($61|0);
   $or$cond = $65 & $66;
   $lo$1 = $or$cond ? $j$08 : $lo$04;
   $lx$1 = $or$cond ? $64 : $lx$06;
   $67 = ($64|0)<($hx$07|0);
   $68 = ($64|0)>($61|0);
   $or$cond1 = $67 & $68;
   $hi$1 = $or$cond1 ? $j$08 : $hi$05;
   $hx$1 = $or$cond1 ? $64 : $hx$07;
   $69 = (($j$08) + 1)|0;
   $exitcond = ($69|0)==($indvars$iv|0);
   if ($exitcond) {
    $hi$1$lcssa = $hi$1;$lo$1$lcssa = $lo$1;
    break;
   } else {
    $hi$05 = $hi$1;$hx$07 = $hx$1;$j$08 = $69;$lo$04 = $lo$1;$lx$06 = $lx$1;
   }
  }
  $70 = (($57) + ($i$510<<2)|0);
  HEAP32[$70>>2] = $lo$1$lcssa;
  $71 = (($58) + ($i$510<<2)|0);
  HEAP32[$71>>2] = $hi$1$lcssa;
  $72 = (($i$510) + 1)|0;
  $indvars$iv$next = (($indvars$iv) + 1)|0;
  $exitcond31 = ($72|0)==($n$0$lcssa32343639|0);
  if ($exitcond31) {
   break;
  } else {
   $i$510 = $72;$indvars$iv = $indvars$iv$next;
  }
 }
 STACKTOP = sp;return ($0|0);
}
function _floor1_free_info($i) {
 $i = $i|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($i|0)==(0|0);
 if (!($0)) {
  _free($i);
 }
 return;
}
function _floor1_free_look($i) {
 $i = $i|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($i|0)==(0|0);
 if (!($0)) {
  _free($i);
 }
 return;
}
function _floor1_inverse1($vb,$in) {
 $vb = $vb|0;
 $in = $in|0;
 var $$0$i = 0, $$0$p$i = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cval$1$ph = 0, $cval$113 = 0, $i$017 = 0, $i$112 = 0, $ispos$i = 0, $j$016 = 0, $k$014 = 0, $neg$i = 0, $val$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($in)) + 1296|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($vb)) + 64|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 28|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($7)) + 2848|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($vb)) + 4|0);
 $11 = (_oggpack_read($10,1)|0);
 $12 = ($11|0)==(1);
 if (!($12)) {
  $$1 = 0;
  return ($$1|0);
 }
 $13 = ((($in)) + 1284|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = $14 << 2;
 $16 = (__vorbis_block_alloc($vb,$15)|0);
 $17 = ((($in)) + 1292|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = (($18) + -1)|0;
 $20 = (_ov_ilog($19)|0);
 $21 = (_oggpack_read($10,$20)|0);
 HEAP32[$16>>2] = $21;
 $22 = HEAP32[$17>>2]|0;
 $23 = (($22) + -1)|0;
 $24 = (_ov_ilog($23)|0);
 $25 = (_oggpack_read($10,$24)|0);
 $26 = ((($16)) + 4|0);
 HEAP32[$26>>2] = $25;
 $27 = HEAP32[$1>>2]|0;
 $28 = ($27|0)>(0);
 L4: do {
  if ($28) {
   $i$017 = 0;$j$016 = 2;
   L5: while(1) {
    $33 = (((($1)) + 4|0) + ($i$017<<2)|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = (((($1)) + 128|0) + ($34<<2)|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = (((($1)) + 192|0) + ($34<<2)|0);
    $38 = HEAP32[$37>>2]|0;
    $39 = 1 << $38;
    $40 = ($38|0)==(0);
    if ($40) {
     $cval$1$ph = 0;
    } else {
     $41 = (((($1)) + 256|0) + ($34<<2)|0);
     $42 = HEAP32[$41>>2]|0;
     $43 = (($9) + (($42*56)|0)|0);
     $44 = (_vorbis_book_decode($43,$10)|0);
     $45 = ($44|0)==(-1);
     if ($45) {
      $$1 = 0;
      label = 25;
      break;
     } else {
      $cval$1$ph = $44;
     }
    }
    $46 = ($36|0)>(0);
    if ($46) {
     $47 = (($39) + -1)|0;
     $cval$113 = $cval$1$ph;$k$014 = 0;
     while(1) {
      $48 = $cval$113 & $47;
      $49 = ((((($1)) + 320|0) + ($34<<5)|0) + ($48<<2)|0);
      $50 = HEAP32[$49>>2]|0;
      $51 = $cval$113 >> $38;
      $52 = ($50|0)>(-1);
      if ($52) {
       $53 = (($9) + (($50*56)|0)|0);
       $54 = (_vorbis_book_decode($53,$10)|0);
       $55 = (($k$014) + ($j$016))|0;
       $56 = (($16) + ($55<<2)|0);
       HEAP32[$56>>2] = $54;
       $57 = ($54|0)==(-1);
       if ($57) {
        $$1 = 0;
        label = 25;
        break L5;
       }
      } else {
       $58 = (($k$014) + ($j$016))|0;
       $59 = (($16) + ($58<<2)|0);
       HEAP32[$59>>2] = 0;
      }
      $60 = (($k$014) + 1)|0;
      $61 = ($60|0)<($36|0);
      if ($61) {
       $cval$113 = $51;$k$014 = $60;
      } else {
       break;
      }
     }
    }
    $62 = (($36) + ($j$016))|0;
    $63 = (($i$017) + 1)|0;
    $64 = HEAP32[$1>>2]|0;
    $65 = ($63|0)<($64|0);
    if ($65) {
     $i$017 = $63;$j$016 = $62;
    } else {
     break L4;
    }
   }
   if ((label|0) == 25) {
    return ($$1|0);
   }
  }
 } while(0);
 $29 = HEAP32[$13>>2]|0;
 $30 = ($29|0)>(2);
 if (!($30)) {
  $$1 = $16;
  return ($$1|0);
 }
 $31 = ((($in)) + 1032|0);
 $32 = ((($in)) + 780|0);
 $i$112 = 2;
 while(1) {
  $66 = (($i$112) + -2)|0;
  $67 = (($31) + ($66<<2)|0);
  $68 = HEAP32[$67>>2]|0;
  $69 = (((($1)) + 836|0) + ($68<<2)|0);
  $70 = HEAP32[$69>>2]|0;
  $71 = (($32) + ($66<<2)|0);
  $72 = HEAP32[$71>>2]|0;
  $73 = (((($1)) + 836|0) + ($72<<2)|0);
  $74 = HEAP32[$73>>2]|0;
  $75 = (($16) + ($68<<2)|0);
  $76 = HEAP32[$75>>2]|0;
  $77 = (($16) + ($72<<2)|0);
  $78 = HEAP32[$77>>2]|0;
  $79 = (((($1)) + 836|0) + ($i$112<<2)|0);
  $80 = HEAP32[$79>>2]|0;
  $81 = $76 & 32767;
  $82 = $78 & 32767;
  $83 = (($82) - ($81))|0;
  $84 = (($74) - ($70))|0;
  $ispos$i = ($83|0)>(-1);
  $neg$i = (0 - ($83))|0;
  $85 = $ispos$i ? $83 : $neg$i;
  $86 = (($80) - ($70))|0;
  $87 = Math_imul($85, $86)|0;
  $88 = (($87|0) / ($84|0))&-1;
  $89 = ($83|0)<(0);
  $90 = (0 - ($88))|0;
  $$0$p$i = $89 ? $90 : $88;
  $$0$i = (($$0$p$i) + ($81))|0;
  $91 = HEAP32[$17>>2]|0;
  $92 = (($91) - ($$0$i))|0;
  $93 = (($16) + ($i$112<<2)|0);
  $94 = HEAP32[$93>>2]|0;
  $95 = ($94|0)==(0);
  if ($95) {
   $120 = $$0$i | 32768;
   HEAP32[$93>>2] = $120;
  } else {
   $96 = ($92|0)<($$0$i|0);
   $97 = $96 ? $92 : $$0$i;
   $98 = $97 << 1;
   $99 = ($94|0)<($98|0);
   do {
    if ($99) {
     $104 = $94 & 1;
     $105 = ($104|0)==(0);
     if ($105) {
      $109 = $94 >> 1;
      $val$0 = $109;
      break;
     } else {
      $106 = (($94) + 1)|0;
      $107 = $106 >> 1;
      $108 = (0 - ($107))|0;
      $val$0 = $108;
      break;
     }
    } else {
     $100 = ($92|0)>($$0$i|0);
     if ($100) {
      $101 = (($94) - ($$0$i))|0;
      $val$0 = $101;
      break;
     } else {
      $102 = (($94) - ($92))|0;
      $103 = $102 ^ -1;
      $val$0 = $103;
      break;
     }
    }
   } while(0);
   $110 = (($val$0) + ($$0$i))|0;
   $111 = $110 & 32767;
   HEAP32[$93>>2] = $111;
   $112 = HEAP32[$67>>2]|0;
   $113 = (($16) + ($112<<2)|0);
   $114 = HEAP32[$113>>2]|0;
   $115 = $114 & 32767;
   HEAP32[$113>>2] = $115;
   $116 = HEAP32[$71>>2]|0;
   $117 = (($16) + ($116<<2)|0);
   $118 = HEAP32[$117>>2]|0;
   $119 = $118 & 32767;
   HEAP32[$117>>2] = $119;
  }
  $121 = (($i$112) + 1)|0;
  $122 = HEAP32[$13>>2]|0;
  $123 = ($121|0)<($122|0);
  if ($123) {
   $i$112 = $121;
  } else {
   $$1 = $16;
   break;
  }
 }
 return ($$1|0);
}
function _vorbis_book_decode($book,$b) {
 $book = $book|0;
 $b = $b|0;
 var $$1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($book)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(0);
 if (!($2)) {
  $$1 = -1;
  return ($$1|0);
 }
 $3 = (_decode_packed_entry_number($book,$b)|0);
 $4 = ($3|0)>(-1);
 if (!($4)) {
  $$1 = -1;
  return ($$1|0);
 }
 $5 = ((($book)) + 24|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + ($3<<2)|0);
 $8 = HEAP32[$7>>2]|0;
 $$1 = $8;
 return ($$1|0);
}
function _floor1_inverse2($vb,$in,$memo,$out) {
 $vb = $vb|0;
 $in = $in|0;
 $memo = $memo|0;
 $out = $out|0;
 var $$0 = 0, $$pn$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0.0, $55 = 0, $56 = 0.0, $57 = 0.0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0.0, $76 = 0.0, $77 = 0, $78 = 0, $8 = 0;
 var $9 = 0, $err$04$i = 0, $err$1$i = 0, $exitcond = 0, $hx$0$lcssa = 0, $hx$03 = 0, $hx$1 = 0, $ispos$i = 0, $ispos1$i = 0, $j$02 = 0, $j$11 = 0, $lx$04 = 0, $lx$1 = 0, $ly$0$lcssa = 0, $ly$05 = 0, $ly$1 = 0, $neg$i = 0, $neg2$i = 0, $x1$n$i = 0, $y$03$i = 0;
 var $y$1$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($in)) + 1296|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($vb)) + 64|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 28|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($vb)) + 28|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (($7) + ($9<<2)|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11|0) / 2)&-1;
 $13 = ($memo|0)==(0|0);
 if ($13) {
  $78 = $12 << 2;
  _memset(($out|0),0,($78|0))|0;
  $$0 = 0;
  return ($$0|0);
 }
 $14 = HEAP32[$memo>>2]|0;
 $15 = ((($1)) + 832|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = Math_imul($16, $14)|0;
 $18 = ($17|0)<(0);
 $19 = ($17|0)>(255);
 $20 = $19 ? 255 : $17;
 $21 = $18 ? 0 : $20;
 $22 = ((($in)) + 1284|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = ($23|0)>(1);
 if ($24) {
  $25 = ((($in)) + 260|0);
  $hx$03 = 0;$j$02 = 1;$lx$04 = 0;$ly$05 = $21;
  while(1) {
   $29 = (($25) + ($j$02<<2)|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = (($memo) + ($30<<2)|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = $32 & 32767;
   $34 = ($33|0)==($32|0);
   if ($34) {
    $35 = (((($1)) + 836|0) + ($30<<2)|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = Math_imul($16, $32)|0;
    $38 = ($37|0)<(0);
    $39 = ($37|0)>(255);
    $40 = $39 ? 255 : $37;
    $41 = $38 ? 0 : $40;
    $42 = (($41) - ($ly$05))|0;
    $43 = (($36) - ($lx$04))|0;
    $ispos$i = ($42|0)>(-1);
    $neg$i = (0 - ($42))|0;
    $44 = $ispos$i ? $42 : $neg$i;
    $45 = (($42|0) / ($43|0))&-1;
    $46 = $42 >> 31;
    $47 = $46 | 1;
    $48 = Math_imul($45, $43)|0;
    $ispos1$i = ($48|0)>(-1);
    $neg2$i = (0 - ($48))|0;
    $49 = $ispos1$i ? $48 : $neg2$i;
    $50 = (($44) - ($49))|0;
    $51 = ($12|0)>($36|0);
    $x1$n$i = $51 ? $36 : $12;
    $52 = ($x1$n$i|0)>($lx$04|0);
    if ($52) {
     $53 = (100240 + ($ly$05<<2)|0);
     $54 = +HEAPF32[$53>>2];
     $55 = (($out) + ($lx$04<<2)|0);
     $56 = +HEAPF32[$55>>2];
     $57 = $54 * $56;
     HEAPF32[$55>>2] = $57;
    }
    $58 = (($lx$04) + 1)|0;
    $59 = ($58|0)<($x1$n$i|0);
    if ($59) {
     $67 = $58;$err$04$i = 0;$y$03$i = $ly$05;
     while(1) {
      $60 = (($err$04$i) + ($50))|0;
      $61 = ($60|0)<($43|0);
      $62 = $61 ? 0 : $47;
      $63 = $61 ? 0 : $43;
      $err$1$i = (($60) - ($63))|0;
      $$pn$i = (($y$03$i) + ($45))|0;
      $y$1$i = (($$pn$i) + ($62))|0;
      $64 = (100240 + ($y$1$i<<2)|0);
      $65 = +HEAPF32[$64>>2];
      $66 = (($out) + ($67<<2)|0);
      $68 = +HEAPF32[$66>>2];
      $69 = $65 * $68;
      HEAPF32[$66>>2] = $69;
      $70 = (($67) + 1)|0;
      $71 = ($70|0)<($x1$n$i|0);
      if ($71) {
       $67 = $70;$err$04$i = $err$1$i;$y$03$i = $y$1$i;
      } else {
       $hx$1 = $36;$lx$1 = $36;$ly$1 = $41;
       break;
      }
     }
    } else {
     $hx$1 = $36;$lx$1 = $36;$ly$1 = $41;
    }
   } else {
    $hx$1 = $hx$03;$lx$1 = $lx$04;$ly$1 = $ly$05;
   }
   $72 = (($j$02) + 1)|0;
   $73 = ($72|0)<($23|0);
   if ($73) {
    $hx$03 = $hx$1;$j$02 = $72;$lx$04 = $lx$1;$ly$05 = $ly$1;
   } else {
    $hx$0$lcssa = $hx$1;$ly$0$lcssa = $ly$1;
    break;
   }
  }
 } else {
  $hx$0$lcssa = 0;$ly$0$lcssa = $21;
 }
 $26 = ($hx$0$lcssa|0)<($12|0);
 if (!($26)) {
  $$0 = 1;
  return ($$0|0);
 }
 $27 = (100240 + ($ly$0$lcssa<<2)|0);
 $28 = +HEAPF32[$27>>2];
 $j$11 = $hx$0$lcssa;
 while(1) {
  $74 = (($out) + ($j$11<<2)|0);
  $75 = +HEAPF32[$74>>2];
  $76 = $28 * $75;
  HEAPF32[$74>>2] = $76;
  $77 = (($j$11) + 1)|0;
  $exitcond = ($77|0)==($12|0);
  if ($exitcond) {
   $$0 = 1;
   break;
  } else {
   $j$11 = $77;
  }
 }
 return ($$0|0);
}
function _mapping0_pack($vi,$vm,$opb) {
 $vi = $vi|0;
 $vm = $vm|0;
 $opb = $opb|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$05 = 0, $i$13 = 0, $i$22 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$vm>>2]|0;
 $1 = ($0|0)>(1);
 if ($1) {
  _oggpack_write($opb,1,1);
  $2 = HEAP32[$vm>>2]|0;
  $3 = (($2) + -1)|0;
  _oggpack_write($opb,$3,4);
 } else {
  _oggpack_write($opb,0,1);
 }
 $4 = ((($vm)) + 1156|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)>(0);
 if ($6) {
  _oggpack_write($opb,1,1);
  $7 = HEAP32[$4>>2]|0;
  $8 = (($7) + -1)|0;
  _oggpack_write($opb,$8,8);
  $9 = HEAP32[$4>>2]|0;
  $10 = ($9|0)>(0);
  if ($10) {
   $11 = ((($vm)) + 1160|0);
   $12 = ((($vi)) + 4|0);
   $13 = ((($vm)) + 2184|0);
   $i$05 = 0;
   while(1) {
    $14 = (($11) + ($i$05<<2)|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = HEAP32[$12>>2]|0;
    $17 = (($16) + -1)|0;
    $18 = (_ov_ilog($17)|0);
    _oggpack_write($opb,$15,$18);
    $19 = (($13) + ($i$05<<2)|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = HEAP32[$12>>2]|0;
    $22 = (($21) + -1)|0;
    $23 = (_ov_ilog($22)|0);
    _oggpack_write($opb,$20,$23);
    $24 = (($i$05) + 1)|0;
    $25 = HEAP32[$4>>2]|0;
    $26 = ($24|0)<($25|0);
    if ($26) {
     $i$05 = $24;
    } else {
     break;
    }
   }
  }
 } else {
  _oggpack_write($opb,0,1);
 }
 _oggpack_write($opb,0,2);
 $27 = HEAP32[$vm>>2]|0;
 $28 = ($27|0)>(1);
 if ($28) {
  $29 = ((($vi)) + 4|0);
  $30 = HEAP32[$29>>2]|0;
  $31 = ($30|0)>(0);
  if ($31) {
   $32 = ((($vm)) + 4|0);
   $i$13 = 0;
   while(1) {
    $37 = (($32) + ($i$13<<2)|0);
    $38 = HEAP32[$37>>2]|0;
    _oggpack_write($opb,$38,4);
    $39 = (($i$13) + 1)|0;
    $40 = HEAP32[$29>>2]|0;
    $41 = ($39|0)<($40|0);
    if ($41) {
     $i$13 = $39;
    } else {
     break;
    }
   }
   $$pre = HEAP32[$vm>>2]|0;
   $34 = $$pre;
   label = 13;
  }
 } else {
  $34 = $27;
  label = 13;
 }
 if ((label|0) == 13) {
  $33 = ($34|0)>(0);
  if (!($33)) {
   return;
  }
 }
 $35 = ((($vm)) + 1028|0);
 $36 = ((($vm)) + 1092|0);
 $i$22 = 0;
 while(1) {
  _oggpack_write($opb,0,8);
  $42 = (($35) + ($i$22<<2)|0);
  $43 = HEAP32[$42>>2]|0;
  _oggpack_write($opb,$43,8);
  $44 = (($36) + ($i$22<<2)|0);
  $45 = HEAP32[$44>>2]|0;
  _oggpack_write($opb,$45,8);
  $46 = (($i$22) + 1)|0;
  $47 = HEAP32[$vm>>2]|0;
  $48 = ($46|0)<($47|0);
  if ($48) {
   $i$22 = $46;
  } else {
   break;
  }
 }
 return;
}
function _mapping0_unpack($vi,$opb) {
 $vi = $vi|0;
 $opb = $opb|0;
 var $$02 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $8 = 0, $9 = 0, $i$017 = 0, $i$115 = 0, $i$214 = 0;
 var $or$cond = 0, $or$cond3 = 0, $or$cond4 = 0, $or$cond5 = 0, $or$cond6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_calloc(1,3208)|0);
 $1 = ((($vi)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 _memset(($0|0),0,3208)|0;
 $3 = ((($vi)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)<(1);
 L1: do {
  if ($5) {
   label = 24;
  } else {
   $6 = (_oggpack_read($opb,1)|0);
   $7 = ($6|0)<(0);
   if ($7) {
    label = 24;
   } else {
    $8 = ($6|0)==(0);
    if ($8) {
     HEAP32[$0>>2] = 1;
    } else {
     $9 = (_oggpack_read($opb,4)|0);
     $10 = (($9) + 1)|0;
     HEAP32[$0>>2] = $10;
     $11 = ($9|0)<(0);
     if ($11) {
      break;
     }
    }
    $12 = (_oggpack_read($opb,1)|0);
    $13 = ($12|0)<(0);
    if (!($13)) {
     $14 = ($12|0)==(0);
     if (!($14)) {
      $15 = (_oggpack_read($opb,8)|0);
      $16 = (($15) + 1)|0;
      $17 = ((($0)) + 1156|0);
      HEAP32[$17>>2] = $16;
      $18 = ($15|0)<(0);
      if ($18) {
       break;
      }
      $19 = ((($0)) + 1160|0);
      $20 = ((($0)) + 2184|0);
      $$pre = HEAP32[$3>>2]|0;
      $25 = $$pre;$i$017 = 0;
      while(1) {
       $24 = (($25) + -1)|0;
       $26 = (_ov_ilog($24)|0);
       $27 = (_oggpack_read($opb,$26)|0);
       $28 = (($19) + ($i$017<<2)|0);
       HEAP32[$28>>2] = $27;
       $29 = HEAP32[$3>>2]|0;
       $30 = (($29) + -1)|0;
       $31 = (_ov_ilog($30)|0);
       $32 = (_oggpack_read($opb,$31)|0);
       $33 = (($20) + ($i$017<<2)|0);
       HEAP32[$33>>2] = $32;
       $34 = $32 | $27;
       $35 = ($34|0)<(0);
       $36 = ($27|0)==($32|0);
       $or$cond = $36 | $35;
       if ($or$cond) {
        break L1;
       }
       $37 = HEAP32[$3>>2]|0;
       $38 = ($27|0)<($37|0);
       $39 = ($32|0)<($37|0);
       $or$cond3 = $38 & $39;
       $23 = (($i$017) + 1)|0;
       if (!($or$cond3)) {
        break L1;
       }
       $21 = HEAP32[$17>>2]|0;
       $22 = ($23|0)<($21|0);
       if ($22) {
        $25 = $37;$i$017 = $23;
       } else {
        break;
       }
      }
     }
     $40 = (_oggpack_read($opb,2)|0);
     $41 = ($40|0)==(0);
     if ($41) {
      $42 = HEAP32[$0>>2]|0;
      $43 = ($42|0)>(1);
      if ($43) {
       $44 = HEAP32[$3>>2]|0;
       $45 = ($44|0)>(0);
       if ($45) {
        $46 = ((($0)) + 4|0);
        $i$115 = 0;
        while(1) {
         $56 = (_oggpack_read($opb,4)|0);
         $57 = (($46) + ($i$115<<2)|0);
         HEAP32[$57>>2] = $56;
         $58 = HEAP32[$0>>2]|0;
         $59 = ($56|0)>=($58|0);
         $60 = ($56|0)<(0);
         $or$cond4 = $60 | $59;
         $55 = (($i$115) + 1)|0;
         if ($or$cond4) {
          break L1;
         }
         $53 = HEAP32[$3>>2]|0;
         $54 = ($55|0)<($53|0);
         if ($54) {
          $i$115 = $55;
         } else {
          $48 = $58;
          label = 17;
          break;
         }
        }
       }
      } else {
       $48 = $42;
       label = 17;
      }
      if ((label|0) == 17) {
       $47 = ($48|0)>(0);
       if (!($47)) {
        $$02 = $0;
        return ($$02|0);
       }
      }
      $49 = ((($0)) + 1028|0);
      $50 = ((($2)) + 16|0);
      $51 = ((($0)) + 1092|0);
      $52 = ((($2)) + 20|0);
      $i$214 = 0;
      while(1) {
       (_oggpack_read($opb,8)|0);
       $64 = (_oggpack_read($opb,8)|0);
       $65 = (($49) + ($i$214<<2)|0);
       HEAP32[$65>>2] = $64;
       $66 = HEAP32[$50>>2]|0;
       $67 = ($64|0)>=($66|0);
       $68 = ($64|0)<(0);
       $or$cond5 = $68 | $67;
       if ($or$cond5) {
        break L1;
       }
       $69 = (_oggpack_read($opb,8)|0);
       $70 = (($51) + ($i$214<<2)|0);
       HEAP32[$70>>2] = $69;
       $71 = HEAP32[$52>>2]|0;
       $72 = ($69|0)>=($71|0);
       $73 = ($69|0)<(0);
       $or$cond6 = $73 | $72;
       $63 = (($i$214) + 1)|0;
       if ($or$cond6) {
        break L1;
       }
       $61 = HEAP32[$0>>2]|0;
       $62 = ($63|0)<($61|0);
       if ($62) {
        $i$214 = $63;
       } else {
        $$02 = $0;
        break;
       }
      }
      return ($$02|0);
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 24) {
  $74 = ($0|0)==(0|0);
  if ($74) {
   $$02 = 0;
   return ($$02|0);
  }
 }
 _free($0);
 $$02 = 0;
 return ($$02|0);
}
function _mapping0_free_info($i) {
 $i = $i|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($i|0)==(0|0);
 if (!($0)) {
  _free($i);
 }
 return;
}
function _mapping0_forward($vb) {
 $vb = $vb|0;
 var $$ = 0, $$3 = 0, $$alloca_mul = 0, $$alloca_mul1 = 0, $$alloca_mul3 = 0, $$alloca_mul5 = 0, $$lcssa = 0, $$lcssa75 = 0.0, $$pre = 0, $$pre$phi55Z2D = 0, $$pre$phi57Z2D = 0, $$pre$phiZ2D = 0, $$pre53 = 0, $$pre54 = 0, $$pre56 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0;
 var $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0.0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0;
 var $139 = 0.0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0.0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0;
 var $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0.0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0;
 var $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0;
 var $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0;
 var $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0;
 var $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0;
 var $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0;
 var $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0;
 var $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0;
 var $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0;
 var $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0;
 var $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0.0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0;
 var $427 = 0, $428 = 0, $429 = 0, $43 = 0.0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0;
 var $445 = 0.0, $446 = 0, $45 = 0, $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0.0, $78 = 0.0, $79 = 0.0;
 var $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0, $85 = 0, $86 = 0.0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0.0, $92 = 0.0, $93 = 0, $94 = 0, $95 = 0.0, $96 = 0.0, $97 = 0.0;
 var $98 = 0, $99 = 0.0, $ch_in_bundle$0$lcssa = 0, $ch_in_bundle$012 = 0, $ch_in_bundle$1 = 0, $ch_in_bundle$2$lcssa = 0, $ch_in_bundle$218 = 0, $ch_in_bundle$3 = 0, $fabs = 0.0, $fabs6$us = 0.0, $global_ampmax$0$lcssa = 0.0, $global_ampmax$042 = 0.0, $global_ampmax$042$us = 0.0, $global_ampmax$1 = 0.0, $global_ampmax$1$us = 0.0, $i$043 = 0, $i$043$us = 0, $i$134 = 0, $i$210 = 0, $i$323 = 0;
 var $j$038$us = 0, $j$128 = 0, $j$211 = 0, $j$317 = 0, $k$227 = 0, $not$ = 0, $storemerge = 0.0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($vb)) + 64|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 28|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($1)) + 104|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($vb)) + 104|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($vb)) + 36|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($3)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = $13 << 2;
 $$alloca_mul = $14;
 $15 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $16 = (__vorbis_block_alloc($vb,$14)|0);
 $17 = HEAP32[$12>>2]|0;
 $18 = $17 << 2;
 $19 = (__vorbis_block_alloc($vb,$18)|0);
 $20 = HEAP32[$12>>2]|0;
 $21 = $20 << 2;
 $22 = (__vorbis_block_alloc($vb,$21)|0);
 $23 = ((($9)) + 4|0);
 $24 = +HEAPF32[$23>>2];
 $25 = HEAP32[$12>>2]|0;
 $26 = $25 << 2;
 $$alloca_mul1 = $26;
 $27 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul1)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $28 = ((($9)) + 8|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = ((($vb)) + 28|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = (((($5)) + 544|0) + ($31<<2)|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($7)) + 56|0);
 $35 = HEAP32[$34>>2]|0;
 $36 = (($35) + (($29*52)|0)|0);
 $37 = ($31|0)!=(0);
 $38 = $37 ? 2 : 0;
 $39 = (($36) + (($38*52)|0)|0);
 $40 = ((($vb)) + 40|0);
 HEAP32[$40>>2] = $31;
 $41 = ($25|0)>(0);
 L1: do {
  if ($41) {
   $42 = (+($11|0));
   $43 = 4.0 / $42;
   $44 = (($11|0) / 2)&-1;
   $45 = $44 << 2;
   $fabs = (+Math_abs((+$43)));
   $46 = (HEAPF32[tempDoublePtr>>2]=$fabs,HEAP32[tempDoublePtr>>2]|0);
   $47 = (+($46>>>0));
   $48 = $47 * 7.1771142984289327E-7;
   $49 = $48 + -764.6162109375;
   $50 = $49;
   $51 = $50 + 0.34499999999999997;
   $52 = $51;
   $53 = ((($7)) + 4|0);
   $54 = ((($vb)) + 24|0);
   $55 = ((($vb)) + 32|0);
   $56 = (($11) + -1)|0;
   $57 = ($56|0)>(1);
   if ($57) {
    $global_ampmax$042$us = $24;$i$043$us = 0;
   } else {
    $global_ampmax$042 = $24;$i$043 = 0;
    while(1) {
     $114 = HEAP32[$vb>>2]|0;
     $115 = (($114) + ($i$043<<2)|0);
     $116 = HEAP32[$115>>2]|0;
     $117 = (__vorbis_block_alloc($vb,$45)|0);
     $118 = (($19) + ($i$043<<2)|0);
     HEAP32[$118>>2] = $117;
     $119 = (__vorbis_block_alloc($vb,$45)|0);
     $120 = (($16) + ($i$043<<2)|0);
     HEAP32[$120>>2] = $119;
     $121 = HEAP32[$54>>2]|0;
     $122 = HEAP32[$30>>2]|0;
     $123 = HEAP32[$55>>2]|0;
     __vorbis_apply_window($116,$53,$5,$121,$122,$123);
     $124 = HEAP32[$30>>2]|0;
     $125 = (((($7)) + 12|0) + ($124<<2)|0);
     $126 = HEAP32[$125>>2]|0;
     $127 = HEAP32[$126>>2]|0;
     $128 = HEAP32[$120>>2]|0;
     _mdct_forward($127,$116,$128);
     $129 = HEAP32[$30>>2]|0;
     $130 = (((($7)) + 20|0) + (($129*12)|0)|0);
     _drft_forward($130,$116);
     $131 = HEAP32[$116>>2]|0;
     $132 = $131 & 2147483647;
     $133 = (+($132>>>0));
     $134 = $133 * 7.1771142984289327E-7;
     $135 = $134 + -764.6162109375;
     $136 = $52 + $135;
     $137 = $136;
     $138 = $137 + 0.34499999999999997;
     $139 = $138;
     HEAPF32[$116>>2] = $139;
     $140 = (($27) + ($i$043<<2)|0);
     $141 = $139 > 0.0;
     $storemerge = $141 ? 0.0 : $139;
     HEAPF32[$140>>2] = $storemerge;
     $142 = $storemerge > $global_ampmax$042;
     $global_ampmax$1 = $142 ? $storemerge : $global_ampmax$042;
     $143 = (($i$043) + 1)|0;
     $144 = HEAP32[$12>>2]|0;
     $145 = ($143|0)<($144|0);
     if ($145) {
      $global_ampmax$042 = $global_ampmax$1;$i$043 = $143;
     } else {
      $$pre$phi55Z2D = $45;$$pre$phiZ2D = $44;$global_ampmax$0$lcssa = $global_ampmax$1;
      break L1;
     }
    }
   }
   while(1) {
    $58 = HEAP32[$vb>>2]|0;
    $59 = (($58) + ($i$043$us<<2)|0);
    $60 = HEAP32[$59>>2]|0;
    $61 = (__vorbis_block_alloc($vb,$45)|0);
    $62 = (($19) + ($i$043$us<<2)|0);
    HEAP32[$62>>2] = $61;
    $63 = (__vorbis_block_alloc($vb,$45)|0);
    $64 = (($16) + ($i$043$us<<2)|0);
    HEAP32[$64>>2] = $63;
    $65 = HEAP32[$54>>2]|0;
    $66 = HEAP32[$30>>2]|0;
    $67 = HEAP32[$55>>2]|0;
    __vorbis_apply_window($60,$53,$5,$65,$66,$67);
    $68 = HEAP32[$30>>2]|0;
    $69 = (((($7)) + 12|0) + ($68<<2)|0);
    $70 = HEAP32[$69>>2]|0;
    $71 = HEAP32[$70>>2]|0;
    $72 = HEAP32[$64>>2]|0;
    _mdct_forward($71,$60,$72);
    $73 = HEAP32[$30>>2]|0;
    $74 = (((($7)) + 20|0) + (($73*12)|0)|0);
    _drft_forward($74,$60);
    $75 = HEAP32[$60>>2]|0;
    $76 = $75 & 2147483647;
    $77 = (+($76>>>0));
    $78 = $77 * 7.1771142984289327E-7;
    $79 = $78 + -764.6162109375;
    $80 = $52 + $79;
    $81 = $80;
    $82 = $81 + 0.34499999999999997;
    $83 = $82;
    HEAPF32[$60>>2] = $83;
    $84 = (($27) + ($i$043$us<<2)|0);
    HEAPF32[$84>>2] = $83;
    $110 = $83;$j$038$us = 1;
    while(1) {
     $90 = (($60) + ($j$038$us<<2)|0);
     $91 = +HEAPF32[$90>>2];
     $92 = $91 * $91;
     $93 = (($j$038$us) + 1)|0;
     $94 = (($60) + ($93<<2)|0);
     $95 = +HEAPF32[$94>>2];
     $96 = $95 * $95;
     $97 = $92 + $96;
     $fabs6$us = (+Math_abs((+$97)));
     $98 = (HEAPF32[tempDoublePtr>>2]=$fabs6$us,HEAP32[tempDoublePtr>>2]|0);
     $99 = (+($98>>>0));
     $100 = $99 * 7.1771142984289327E-7;
     $101 = $100 + -764.6162109375;
     $102 = $101 * 0.5;
     $103 = $52 + $102;
     $104 = $103;
     $105 = $104 + 0.34499999999999997;
     $106 = $105;
     $107 = $93 >> 1;
     $108 = (($60) + ($107<<2)|0);
     HEAPF32[$108>>2] = $106;
     $109 = $106 > $110;
     if ($109) {
      HEAPF32[$84>>2] = $106;
      $445 = $106;
     } else {
      $445 = $110;
     }
     $111 = (($j$038$us) + 2)|0;
     $112 = ($111|0)<($56|0);
     if ($112) {
      $110 = $445;$j$038$us = $111;
     } else {
      $$lcssa75 = $445;
      break;
     }
    }
    $113 = $$lcssa75 > 0.0;
    if ($113) {
     HEAPF32[$84>>2] = 0.0;
     $86 = 0.0;
    } else {
     $86 = $$lcssa75;
    }
    $85 = $86 > $global_ampmax$042$us;
    $global_ampmax$1$us = $85 ? $86 : $global_ampmax$042$us;
    $87 = (($i$043$us) + 1)|0;
    $88 = HEAP32[$12>>2]|0;
    $89 = ($87|0)<($88|0);
    if ($89) {
     $global_ampmax$042$us = $global_ampmax$1$us;$i$043$us = $87;
    } else {
     $$pre$phi55Z2D = $45;$$pre$phiZ2D = $44;$global_ampmax$0$lcssa = $global_ampmax$1$us;
     break;
    }
   }
  } else {
   $$pre53 = (($11|0) / 2)&-1;
   $$pre54 = $$pre53 << 2;
   $$pre$phi55Z2D = $$pre54;$$pre$phiZ2D = $$pre53;$global_ampmax$0$lcssa = $24;
  }
 } while(0);
 $146 = (__vorbis_block_alloc($vb,$$pre$phi55Z2D)|0);
 $147 = (__vorbis_block_alloc($vb,$$pre$phi55Z2D)|0);
 $148 = HEAP32[$12>>2]|0;
 $149 = ($148|0)>(0);
 L19: do {
  if ($149) {
   $150 = ($11|0)>(1);
   $151 = ((($7)) + 48|0);
   $i$134 = 0;
   while(1) {
    $152 = (((($33)) + 4|0) + ($i$134<<2)|0);
    $153 = HEAP32[$152>>2]|0;
    $154 = (($16) + ($i$134<<2)|0);
    $155 = HEAP32[$154>>2]|0;
    $156 = HEAP32[$vb>>2]|0;
    $157 = (($156) + ($i$134<<2)|0);
    $158 = HEAP32[$157>>2]|0;
    $159 = (($158) + ($$pre$phiZ2D<<2)|0);
    HEAP32[$40>>2] = $31;
    $160 = (__vorbis_block_alloc($vb,60)|0);
    $161 = (($22) + ($i$134<<2)|0);
    HEAP32[$161>>2] = $160;
    dest=$160; stop=dest+60|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
    if ($150) {
     $j$128 = 0;
     while(1) {
      $162 = (($155) + ($j$128<<2)|0);
      $163 = HEAP32[$162>>2]|0;
      $164 = $163 & 2147483647;
      $165 = (+($164>>>0));
      $166 = $165 * 7.1771142984289327E-7;
      $167 = $166 + -764.6162109375;
      $168 = $167;
      $169 = $168 + 0.34499999999999997;
      $170 = $169;
      $171 = (($159) + ($j$128<<2)|0);
      HEAPF32[$171>>2] = $170;
      $172 = (($j$128) + 1)|0;
      $173 = ($172|0)<($$pre$phiZ2D|0);
      if ($173) {
       $j$128 = $172;
      } else {
       break;
      }
     }
    }
    __vp_noisemask($39,$159,$146);
    $174 = (($27) + ($i$134<<2)|0);
    $175 = +HEAPF32[$174>>2];
    __vp_tonemask($39,$158,$147,$global_ampmax$0$lcssa,$175);
    __vp_offset_and_mix($39,$146,$147,1,$158,$155,$159);
    $176 = (((($33)) + 1028|0) + ($153<<2)|0);
    $177 = HEAP32[$176>>2]|0;
    $178 = (((($5)) + 800|0) + ($177<<2)|0);
    $179 = HEAP32[$178>>2]|0;
    $180 = ($179|0)==(1);
    if (!($180)) {
     $$3 = -1;
     break;
    }
    $181 = HEAP32[$151>>2]|0;
    $182 = (($181) + ($177<<2)|0);
    $183 = HEAP32[$182>>2]|0;
    $184 = (_floor1_fit($vb,$183,$159,$158)|0);
    $185 = HEAP32[$161>>2]|0;
    $186 = ((($185)) + 28|0);
    HEAP32[$186>>2] = $184;
    $187 = (_vorbis_bitrate_managed($vb)|0);
    $188 = ($187|0)==(0);
    if (!($188)) {
     $189 = HEAP32[$161>>2]|0;
     $190 = ((($189)) + 28|0);
     $191 = HEAP32[$190>>2]|0;
     $192 = ($191|0)==(0|0);
     if (!($192)) {
      __vp_offset_and_mix($39,$146,$147,2,$158,$155,$159);
      $193 = HEAP32[$176>>2]|0;
      $194 = HEAP32[$151>>2]|0;
      $195 = (($194) + ($193<<2)|0);
      $196 = HEAP32[$195>>2]|0;
      $197 = (_floor1_fit($vb,$196,$159,$158)|0);
      $198 = HEAP32[$161>>2]|0;
      $199 = ((($198)) + 56|0);
      HEAP32[$199>>2] = $197;
      __vp_offset_and_mix($39,$146,$147,0,$158,$155,$159);
      $200 = HEAP32[$176>>2]|0;
      $201 = HEAP32[$151>>2]|0;
      $202 = (($201) + ($200<<2)|0);
      $203 = HEAP32[$202>>2]|0;
      $204 = (_floor1_fit($vb,$203,$159,$158)|0);
      $205 = HEAP32[$161>>2]|0;
      HEAP32[$205>>2] = $204;
      $206 = HEAP32[$176>>2]|0;
      $207 = HEAP32[$151>>2]|0;
      $208 = (($207) + ($206<<2)|0);
      $209 = HEAP32[$208>>2]|0;
      $210 = HEAP32[$161>>2]|0;
      $211 = HEAP32[$210>>2]|0;
      $212 = ((($210)) + 28|0);
      $213 = HEAP32[$212>>2]|0;
      $214 = (_floor1_interpolate_fit($vb,$209,$211,$213,9362)|0);
      $215 = HEAP32[$161>>2]|0;
      $216 = ((($215)) + 4|0);
      HEAP32[$216>>2] = $214;
      $217 = HEAP32[$176>>2]|0;
      $218 = HEAP32[$151>>2]|0;
      $219 = (($218) + ($217<<2)|0);
      $220 = HEAP32[$219>>2]|0;
      $221 = HEAP32[$161>>2]|0;
      $222 = HEAP32[$221>>2]|0;
      $223 = ((($221)) + 28|0);
      $224 = HEAP32[$223>>2]|0;
      $225 = (_floor1_interpolate_fit($vb,$220,$222,$224,18724)|0);
      $226 = HEAP32[$161>>2]|0;
      $227 = ((($226)) + 8|0);
      HEAP32[$227>>2] = $225;
      $228 = HEAP32[$176>>2]|0;
      $229 = HEAP32[$151>>2]|0;
      $230 = (($229) + ($228<<2)|0);
      $231 = HEAP32[$230>>2]|0;
      $232 = HEAP32[$161>>2]|0;
      $233 = HEAP32[$232>>2]|0;
      $234 = ((($232)) + 28|0);
      $235 = HEAP32[$234>>2]|0;
      $236 = (_floor1_interpolate_fit($vb,$231,$233,$235,28086)|0);
      $237 = HEAP32[$161>>2]|0;
      $238 = ((($237)) + 12|0);
      HEAP32[$238>>2] = $236;
      $239 = HEAP32[$176>>2]|0;
      $240 = HEAP32[$151>>2]|0;
      $241 = (($240) + ($239<<2)|0);
      $242 = HEAP32[$241>>2]|0;
      $243 = HEAP32[$161>>2]|0;
      $244 = HEAP32[$243>>2]|0;
      $245 = ((($243)) + 28|0);
      $246 = HEAP32[$245>>2]|0;
      $247 = (_floor1_interpolate_fit($vb,$242,$244,$246,37449)|0);
      $248 = HEAP32[$161>>2]|0;
      $249 = ((($248)) + 16|0);
      HEAP32[$249>>2] = $247;
      $250 = HEAP32[$176>>2]|0;
      $251 = HEAP32[$151>>2]|0;
      $252 = (($251) + ($250<<2)|0);
      $253 = HEAP32[$252>>2]|0;
      $254 = HEAP32[$161>>2]|0;
      $255 = HEAP32[$254>>2]|0;
      $256 = ((($254)) + 28|0);
      $257 = HEAP32[$256>>2]|0;
      $258 = (_floor1_interpolate_fit($vb,$253,$255,$257,46811)|0);
      $259 = HEAP32[$161>>2]|0;
      $260 = ((($259)) + 20|0);
      HEAP32[$260>>2] = $258;
      $261 = HEAP32[$176>>2]|0;
      $262 = HEAP32[$151>>2]|0;
      $263 = (($262) + ($261<<2)|0);
      $264 = HEAP32[$263>>2]|0;
      $265 = HEAP32[$161>>2]|0;
      $266 = HEAP32[$265>>2]|0;
      $267 = ((($265)) + 28|0);
      $268 = HEAP32[$267>>2]|0;
      $269 = (_floor1_interpolate_fit($vb,$264,$266,$268,56173)|0);
      $270 = HEAP32[$161>>2]|0;
      $271 = ((($270)) + 24|0);
      HEAP32[$271>>2] = $269;
      $272 = HEAP32[$176>>2]|0;
      $273 = HEAP32[$151>>2]|0;
      $274 = (($273) + ($272<<2)|0);
      $275 = HEAP32[$274>>2]|0;
      $276 = HEAP32[$161>>2]|0;
      $277 = ((($276)) + 28|0);
      $278 = HEAP32[$277>>2]|0;
      $279 = ((($276)) + 56|0);
      $280 = HEAP32[$279>>2]|0;
      $281 = (_floor1_interpolate_fit($vb,$275,$278,$280,9362)|0);
      $282 = HEAP32[$161>>2]|0;
      $283 = ((($282)) + 32|0);
      HEAP32[$283>>2] = $281;
      $284 = HEAP32[$176>>2]|0;
      $285 = HEAP32[$151>>2]|0;
      $286 = (($285) + ($284<<2)|0);
      $287 = HEAP32[$286>>2]|0;
      $288 = HEAP32[$161>>2]|0;
      $289 = ((($288)) + 28|0);
      $290 = HEAP32[$289>>2]|0;
      $291 = ((($288)) + 56|0);
      $292 = HEAP32[$291>>2]|0;
      $293 = (_floor1_interpolate_fit($vb,$287,$290,$292,18724)|0);
      $294 = HEAP32[$161>>2]|0;
      $295 = ((($294)) + 36|0);
      HEAP32[$295>>2] = $293;
      $296 = HEAP32[$176>>2]|0;
      $297 = HEAP32[$151>>2]|0;
      $298 = (($297) + ($296<<2)|0);
      $299 = HEAP32[$298>>2]|0;
      $300 = HEAP32[$161>>2]|0;
      $301 = ((($300)) + 28|0);
      $302 = HEAP32[$301>>2]|0;
      $303 = ((($300)) + 56|0);
      $304 = HEAP32[$303>>2]|0;
      $305 = (_floor1_interpolate_fit($vb,$299,$302,$304,28086)|0);
      $306 = HEAP32[$161>>2]|0;
      $307 = ((($306)) + 40|0);
      HEAP32[$307>>2] = $305;
      $308 = HEAP32[$176>>2]|0;
      $309 = HEAP32[$151>>2]|0;
      $310 = (($309) + ($308<<2)|0);
      $311 = HEAP32[$310>>2]|0;
      $312 = HEAP32[$161>>2]|0;
      $313 = ((($312)) + 28|0);
      $314 = HEAP32[$313>>2]|0;
      $315 = ((($312)) + 56|0);
      $316 = HEAP32[$315>>2]|0;
      $317 = (_floor1_interpolate_fit($vb,$311,$314,$316,37449)|0);
      $318 = HEAP32[$161>>2]|0;
      $319 = ((($318)) + 44|0);
      HEAP32[$319>>2] = $317;
      $320 = HEAP32[$176>>2]|0;
      $321 = HEAP32[$151>>2]|0;
      $322 = (($321) + ($320<<2)|0);
      $323 = HEAP32[$322>>2]|0;
      $324 = HEAP32[$161>>2]|0;
      $325 = ((($324)) + 28|0);
      $326 = HEAP32[$325>>2]|0;
      $327 = ((($324)) + 56|0);
      $328 = HEAP32[$327>>2]|0;
      $329 = (_floor1_interpolate_fit($vb,$323,$326,$328,46811)|0);
      $330 = HEAP32[$161>>2]|0;
      $331 = ((($330)) + 48|0);
      HEAP32[$331>>2] = $329;
      $332 = HEAP32[$176>>2]|0;
      $333 = HEAP32[$151>>2]|0;
      $334 = (($333) + ($332<<2)|0);
      $335 = HEAP32[$334>>2]|0;
      $336 = HEAP32[$161>>2]|0;
      $337 = ((($336)) + 28|0);
      $338 = HEAP32[$337>>2]|0;
      $339 = ((($336)) + 56|0);
      $340 = HEAP32[$339>>2]|0;
      $341 = (_floor1_interpolate_fit($vb,$335,$338,$340,56173)|0);
      $342 = HEAP32[$161>>2]|0;
      $343 = ((($342)) + 52|0);
      HEAP32[$343>>2] = $341;
     }
    }
    $344 = (($i$134) + 1)|0;
    $345 = HEAP32[$12>>2]|0;
    $346 = ($344|0)<($345|0);
    if ($346) {
     $i$134 = $344;
    } else {
     $$pre$phi57Z2D = $151;$348 = $345;
     break L19;
    }
   }
   STACKTOP = sp;return ($$3|0);
  } else {
   $$pre56 = ((($7)) + 48|0);
   $$pre$phi57Z2D = $$pre56;$348 = $148;
  }
 } while(0);
 HEAPF32[$23>>2] = $global_ampmax$0$lcssa;
 $347 = $348 << 2;
 $$alloca_mul3 = $347;
 $349 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul3)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $$alloca_mul5 = $347;
 $350 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul5)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $351 = (_vorbis_bitrate_managed($vb)|0);
 $352 = ($351|0)!=(0);
 $353 = $352 ? 0 : 7;
 $354 = ((($7)) + 44|0);
 $355 = ((($vb)) + 24|0);
 $356 = ((($vb)) + 32|0);
 $357 = ((($5)) + 2868|0);
 $358 = ((($7)) + 52|0);
 $k$227 = $353;
 while(1) {
  $359 = (((($9)) + 12|0) + ($k$227<<2)|0);
  $360 = HEAP32[$359>>2]|0;
  _oggpack_write($360,0,1);
  $361 = HEAP32[$354>>2]|0;
  _oggpack_write($360,$31,$361);
  $362 = HEAP32[$30>>2]|0;
  $363 = ($362|0)==(0);
  if (!($363)) {
   $364 = HEAP32[$355>>2]|0;
   _oggpack_write($360,$364,1);
   $365 = HEAP32[$356>>2]|0;
   _oggpack_write($360,$365,1);
  }
  $366 = HEAP32[$12>>2]|0;
  $367 = ($366|0)>(0);
  if ($367) {
   $i$210 = 0;
   while(1) {
    $368 = (((($33)) + 4|0) + ($i$210<<2)|0);
    $369 = HEAP32[$368>>2]|0;
    $370 = (($19) + ($i$210<<2)|0);
    $371 = HEAP32[$370>>2]|0;
    $372 = (((($33)) + 1028|0) + ($369<<2)|0);
    $373 = HEAP32[$372>>2]|0;
    $374 = HEAP32[$$pre$phi57Z2D>>2]|0;
    $375 = (($374) + ($373<<2)|0);
    $376 = HEAP32[$375>>2]|0;
    $377 = (($22) + ($i$210<<2)|0);
    $378 = HEAP32[$377>>2]|0;
    $379 = (($378) + ($k$227<<2)|0);
    $380 = HEAP32[$379>>2]|0;
    $381 = (_floor1_encode($360,$vb,$376,$380,$371)|0);
    $382 = (($15) + ($i$210<<2)|0);
    HEAP32[$382>>2] = $381;
    $383 = (($i$210) + 1)|0;
    $384 = HEAP32[$12>>2]|0;
    $385 = ($383|0)<($384|0);
    if ($385) {
     $i$210 = $383;
    } else {
     $$lcssa = $384;
     break;
    }
   }
  } else {
   $$lcssa = $366;
  }
  $386 = HEAP32[$30>>2]|0;
  $387 = ((((($5)) + 3240|0) + (($386*60)|0)|0) + ($k$227<<2)|0);
  $388 = HEAP32[$387>>2]|0;
  __vp_couple_quantize_normalize($k$227,$357,$39,$33,$16,$19,$15,$388,$$lcssa);
  $389 = HEAP32[$33>>2]|0;
  $390 = ($389|0)>(0);
  if ($390) {
   $i$323 = 0;
   while(1) {
    $391 = (((($33)) + 1092|0) + ($i$323<<2)|0);
    $392 = HEAP32[$391>>2]|0;
    $393 = HEAP32[$12>>2]|0;
    $394 = ($393|0)>(0);
    if ($394) {
     $446 = $393;$ch_in_bundle$012 = 0;$j$211 = 0;
     while(1) {
      $395 = (((($33)) + 4|0) + ($j$211<<2)|0);
      $396 = HEAP32[$395>>2]|0;
      $397 = ($396|0)==($i$323|0);
      if ($397) {
       $398 = (($350) + ($ch_in_bundle$012<<2)|0);
       $399 = (($15) + ($j$211<<2)|0);
       $400 = HEAP32[$399>>2]|0;
       $not$ = ($400|0)!=(0);
       $$ = $not$&1;
       HEAP32[$398>>2] = $$;
       $401 = (($19) + ($j$211<<2)|0);
       $402 = HEAP32[$401>>2]|0;
       $403 = (($ch_in_bundle$012) + 1)|0;
       $404 = (($349) + ($ch_in_bundle$012<<2)|0);
       HEAP32[$404>>2] = $402;
       $$pre = HEAP32[$12>>2]|0;
       $407 = $$pre;$ch_in_bundle$1 = $403;
      } else {
       $407 = $446;$ch_in_bundle$1 = $ch_in_bundle$012;
      }
      $405 = (($j$211) + 1)|0;
      $406 = ($405|0)<($407|0);
      if ($406) {
       $446 = $407;$ch_in_bundle$012 = $ch_in_bundle$1;$j$211 = $405;
      } else {
       $ch_in_bundle$0$lcssa = $ch_in_bundle$1;
       break;
      }
     }
    } else {
     $ch_in_bundle$0$lcssa = 0;
    }
    $408 = (((($5)) + 1312|0) + ($392<<2)|0);
    $409 = HEAP32[$408>>2]|0;
    $410 = (101264 + ($409<<2)|0);
    $411 = HEAP32[$410>>2]|0;
    $412 = ((($411)) + 20|0);
    $413 = HEAP32[$412>>2]|0;
    $414 = HEAP32[$358>>2]|0;
    $415 = (($414) + ($392<<2)|0);
    $416 = HEAP32[$415>>2]|0;
    $417 = (FUNCTION_TABLE_iiiiii[$413 & 8191]($vb,$416,$349,$350,$ch_in_bundle$0$lcssa)|0);
    $418 = HEAP32[$12>>2]|0;
    $419 = ($418|0)>(0);
    if ($419) {
     $ch_in_bundle$218 = 0;$j$317 = 0;
     while(1) {
      $420 = (((($33)) + 4|0) + ($j$317<<2)|0);
      $421 = HEAP32[$420>>2]|0;
      $422 = ($421|0)==($i$323|0);
      if ($422) {
       $423 = (($19) + ($j$317<<2)|0);
       $424 = HEAP32[$423>>2]|0;
       $425 = (($ch_in_bundle$218) + 1)|0;
       $426 = (($349) + ($ch_in_bundle$218<<2)|0);
       HEAP32[$426>>2] = $424;
       $ch_in_bundle$3 = $425;
      } else {
       $ch_in_bundle$3 = $ch_in_bundle$218;
      }
      $427 = (($j$317) + 1)|0;
      $428 = ($427|0)<($418|0);
      if ($428) {
       $ch_in_bundle$218 = $ch_in_bundle$3;$j$317 = $427;
      } else {
       $ch_in_bundle$2$lcssa = $ch_in_bundle$3;
       break;
      }
     }
    } else {
     $ch_in_bundle$2$lcssa = 0;
    }
    $429 = HEAP32[$408>>2]|0;
    $430 = (101264 + ($429<<2)|0);
    $431 = HEAP32[$430>>2]|0;
    $432 = ((($431)) + 24|0);
    $433 = HEAP32[$432>>2]|0;
    $434 = HEAP32[$358>>2]|0;
    $435 = (($434) + ($392<<2)|0);
    $436 = HEAP32[$435>>2]|0;
    (FUNCTION_TABLE_iiiiiiiii[$433 & 8191]($360,$vb,$436,$349,$350,$ch_in_bundle$2$lcssa,$417,$i$323)|0);
    $437 = (($i$323) + 1)|0;
    $438 = HEAP32[$33>>2]|0;
    $439 = ($437|0)<($438|0);
    if ($439) {
     $i$323 = $437;
    } else {
     break;
    }
   }
  }
  $440 = (($k$227) + 1)|0;
  $441 = (_vorbis_bitrate_managed($vb)|0);
  $442 = ($441|0)!=(0);
  $443 = $442 ? 14 : 7;
  $444 = ($k$227|0)<($443|0);
  if ($444) {
   $k$227 = $440;
  } else {
   $$3 = 0;
   break;
  }
 }
 STACKTOP = sp;return ($$3|0);
}
function __vorbis_apply_window($d,$winno,$blocksizes,$lW,$W,$nW) {
 $d = $d|0;
 $winno = $winno|0;
 $blocksizes = $blocksizes|0;
 $lW = $lW|0;
 $W = $W|0;
 $nW = $nW|0;
 var $$neg = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond16 = 0, $i$0$lcssa = 0, $i$18 = 0, $i$2$lcssa = 0, $i$25 = 0, $p$09 = 0, $p$16 = 0;
 var $p$16$in = 0, $scevgep = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($W|0)!=(0);
 $1 = $0 ? $lW : 0;
 $2 = $0 ? $nW : 0;
 $3 = (($winno) + ($1<<2)|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (67568 + ($4<<2)|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($winno) + ($2<<2)|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (67568 + ($8<<2)|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($blocksizes) + ($W<<2)|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($blocksizes) + ($1<<2)|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($blocksizes) + ($2<<2)|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = (($12|0) / 4)&-1;
 $18 = (($14|0) / 4)&-1;
 $19 = (($17) - ($18))|0;
 $20 = (($14|0) / 2)&-1;
 $21 = (($19) + ($20))|0;
 $22 = (($12|0) / 2)&-1;
 $23 = (($22) + ($17))|0;
 $$neg = (($16|0) / -4)&-1;
 $24 = (($23) + ($$neg))|0;
 $25 = (($16|0) / 2)&-1;
 $26 = (($24) + ($25))|0;
 $27 = ($19|0)>(0);
 if ($27) {
  $28 = (($17) - ($18))|0;
  $29 = $28 << 2;
  _memset(($d|0),0,($29|0))|0;
  $i$0$lcssa = $19;
 } else {
  $i$0$lcssa = 0;
 }
 $30 = ($i$0$lcssa|0)<($21|0);
 if ($30) {
  $31 = (($17) + ($20))|0;
  $32 = (($31) - ($i$0$lcssa))|0;
  $33 = (($32) - ($18))|0;
  $i$18 = $i$0$lcssa;$p$09 = 0;
  while(1) {
   $35 = (($6) + ($p$09<<2)|0);
   $36 = +HEAPF32[$35>>2];
   $37 = (($d) + ($i$18<<2)|0);
   $38 = +HEAPF32[$37>>2];
   $39 = $36 * $38;
   HEAPF32[$37>>2] = $39;
   $40 = (($i$18) + 1)|0;
   $41 = (($p$09) + 1)|0;
   $exitcond16 = ($41|0)==($33|0);
   if ($exitcond16) {
    break;
   } else {
    $i$18 = $40;$p$09 = $41;
   }
  }
 }
 $34 = ($16|0)>(1);
 if ($34) {
  $i$25 = $24;$p$16$in = $25;
  while(1) {
   $p$16 = (($p$16$in) + -1)|0;
   $45 = (($10) + ($p$16<<2)|0);
   $46 = +HEAPF32[$45>>2];
   $47 = (($d) + ($i$25<<2)|0);
   $48 = +HEAPF32[$47>>2];
   $49 = $46 * $48;
   HEAPF32[$47>>2] = $49;
   $50 = (($i$25) + 1)|0;
   $51 = ($50|0)<($26|0);
   if ($51) {
    $i$25 = $50;$p$16$in = $p$16;
   } else {
    $i$2$lcssa = $50;
    break;
   }
  }
 } else {
  $i$2$lcssa = $24;
 }
 $42 = ($12|0)>($i$2$lcssa|0);
 if (!($42)) {
  return;
 }
 $scevgep = (($d) + ($i$2$lcssa<<2)|0);
 $43 = (($12) - ($i$2$lcssa))|0;
 $44 = $43 << 2;
 _memset(($scevgep|0),0,($44|0))|0;
 return;
}
function _mdct_forward($init,$in,$out) {
 $init = $init|0;
 $in = $in|0;
 $out = $out|0;
 var $$alloca_mul = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0.0, $105 = 0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0, $112 = 0.0, $113 = 0.0, $114 = 0.0;
 var $115 = 0.0, $116 = 0.0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0.0, $132 = 0;
 var $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0, $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0.0;
 var $151 = 0.0, $152 = 0.0, $153 = 0.0, $154 = 0, $155 = 0.0, $156 = 0, $157 = 0.0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0.0, $167 = 0, $168 = 0.0, $169 = 0.0;
 var $17 = 0, $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0, $174 = 0.0, $175 = 0.0, $176 = 0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0.0, $182 = 0.0, $183 = 0.0, $184 = 0.0, $185 = 0.0, $186 = 0.0, $187 = 0.0;
 var $188 = 0, $189 = 0.0, $19 = 0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0.0, $204 = 0.0;
 var $205 = 0.0, $206 = 0, $207 = 0.0, $208 = 0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0.0, $212 = 0.0, $213 = 0.0, $214 = 0, $215 = 0.0, $216 = 0.0, $217 = 0.0, $218 = 0.0, $219 = 0.0, $22 = 0, $220 = 0.0, $221 = 0.0, $222 = 0.0;
 var $223 = 0.0, $224 = 0, $225 = 0, $226 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0;
 var $38 = 0.0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0;
 var $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $T$0$i = 0, $T$0$lcssa = 0, $T$027 = 0, $T$1$lcssa = 0, $T$118 = 0, $T$211 = 0, $T$37 = 0, $bit$0$i = 0, $exitcond = 0, $i$0$lcssa = 0, $i$028 = 0, $i$1$lcssa = 0;
 var $i$119 = 0, $i$212 = 0, $i$38 = 0, $init$idx = 0, $init$idx$val = 0, $init$idx2$val = 0, $init$idx3$val = 0, $init$idx4 = 0, $init$idx4$val = 0, $scevgep = 0, $scevgep40 = 0, $scevgep41 = 0, $scevgep42 = 0, $w$05 = 0, $w0$0$i = 0, $w1$0$i = 0, $x0$0$lcssa = 0, $x0$025 = 0, $x0$116 = 0, $x0$29 = 0;
 var $x0$36 = 0, $x1$026 = 0, $x1$1$lcssa = 0, $x1$117 = 0, $x1$210 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$init>>2]|0;
 $1 = $0 >> 1;
 $2 = $0 >> 2;
 $3 = $0 >> 3;
 $4 = $0 << 2;
 $$alloca_mul = $4;
 $5 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $6 = (($5) + ($1<<2)|0);
 $7 = (($in) + ($1<<2)|0);
 $8 = (($7) + ($2<<2)|0);
 $9 = ((($init)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + ($1<<2)|0);
 $12 = ($3|0)>(0);
 if ($12) {
  $13 = ((($8)) + 4|0);
  $14 = (($3) + -1)|0;
  $15 = $14 >>> 1;
  $16 = $15 << 1;
  $17 = (($1) + -2)|0;
  $18 = (($17) - ($16))|0;
  $19 = (($2) + ($1))|0;
  $20 = (($19) + -4)|0;
  $21 = $15 << 2;
  $22 = (($20) - ($21))|0;
  $T$027 = $11;$i$028 = 0;$x0$025 = $8;$x1$026 = $13;
  while(1) {
   $23 = ((($x0$025)) + -16|0);
   $24 = ((($T$027)) + -8|0);
   $25 = ((($x0$025)) + -8|0);
   $26 = +HEAPF32[$25>>2];
   $27 = +HEAPF32[$x1$026>>2];
   $28 = $26 + $27;
   $29 = +HEAPF32[$23>>2];
   $30 = ((($x1$026)) + 8|0);
   $31 = +HEAPF32[$30>>2];
   $32 = $29 + $31;
   $33 = ((($T$027)) + -4|0);
   $34 = +HEAPF32[$33>>2];
   $35 = $32 * $34;
   $36 = +HEAPF32[$24>>2];
   $37 = $28 * $36;
   $38 = $35 + $37;
   $39 = (($6) + ($i$028<<2)|0);
   HEAPF32[$39>>2] = $38;
   $40 = +HEAPF32[$24>>2];
   $41 = $32 * $40;
   $42 = +HEAPF32[$33>>2];
   $43 = $28 * $42;
   $44 = $41 - $43;
   $45 = $i$028 | 1;
   $46 = (($6) + ($45<<2)|0);
   HEAPF32[$46>>2] = $44;
   $47 = ((($x1$026)) + 16|0);
   $48 = (($i$028) + 2)|0;
   $49 = ($48|0)<($3|0);
   if ($49) {
    $T$027 = $24;$i$028 = $48;$x0$025 = $23;$x1$026 = $47;
   } else {
    break;
   }
  }
  $50 = (($16) + 2)|0;
  $scevgep41 = (($10) + ($18<<2)|0);
  $scevgep42 = (($in) + ($22<<2)|0);
  $T$0$lcssa = $scevgep41;$i$0$lcssa = $50;$x0$0$lcssa = $scevgep42;
 } else {
  $T$0$lcssa = $11;$i$0$lcssa = 0;$x0$0$lcssa = $8;
 }
 $51 = ((($in)) + 4|0);
 $52 = (($1) - ($3))|0;
 $53 = ($i$0$lcssa|0)<($52|0);
 if ($53) {
  $54 = (($1) + -1)|0;
  $55 = (($54) - ($i$0$lcssa))|0;
  $56 = (($55) - ($3))|0;
  $57 = $56 >>> 1;
  $58 = $57 << 1;
  $59 = (($i$0$lcssa) + ($58))|0;
  $60 = $57 << 2;
  $61 = (($60) + 5)|0;
  $62 = (-2 - ($58))|0;
  $T$118 = $T$0$lcssa;$i$119 = $i$0$lcssa;$x0$116 = $x0$0$lcssa;$x1$117 = $51;
  while(1) {
   $63 = ((($T$118)) + -8|0);
   $64 = ((($x0$116)) + -16|0);
   $65 = ((($x0$116)) + -8|0);
   $66 = +HEAPF32[$65>>2];
   $67 = +HEAPF32[$x1$117>>2];
   $68 = $66 - $67;
   $69 = +HEAPF32[$64>>2];
   $70 = ((($x1$117)) + 8|0);
   $71 = +HEAPF32[$70>>2];
   $72 = $69 - $71;
   $73 = ((($T$118)) + -4|0);
   $74 = +HEAPF32[$73>>2];
   $75 = $72 * $74;
   $76 = +HEAPF32[$63>>2];
   $77 = $68 * $76;
   $78 = $75 + $77;
   $79 = (($6) + ($i$119<<2)|0);
   HEAPF32[$79>>2] = $78;
   $80 = +HEAPF32[$63>>2];
   $81 = $72 * $80;
   $82 = +HEAPF32[$73>>2];
   $83 = $68 * $82;
   $84 = $81 - $83;
   $85 = $i$119 | 1;
   $86 = (($6) + ($85<<2)|0);
   HEAPF32[$86>>2] = $84;
   $87 = ((($x1$117)) + 16|0);
   $88 = (($i$119) + 2)|0;
   $89 = ($88|0)<($52|0);
   if ($89) {
    $T$118 = $63;$i$119 = $88;$x0$116 = $64;$x1$117 = $87;
   } else {
    break;
   }
  }
  $90 = (($59) + 2)|0;
  $scevgep = (($in) + ($61<<2)|0);
  $scevgep40 = (($T$0$lcssa) + ($62<<2)|0);
  $T$1$lcssa = $scevgep40;$i$1$lcssa = $90;$x1$1$lcssa = $scevgep;
 } else {
  $T$1$lcssa = $T$0$lcssa;$i$1$lcssa = $i$0$lcssa;$x1$1$lcssa = $51;
 }
 $91 = ($i$1$lcssa|0)<($1|0);
 if ($91) {
  $92 = (($in) + ($0<<2)|0);
  $T$211 = $T$1$lcssa;$i$212 = $i$1$lcssa;$x0$29 = $92;$x1$210 = $x1$1$lcssa;
  while(1) {
   $93 = ((($T$211)) + -8|0);
   $94 = ((($x0$29)) + -16|0);
   $95 = ((($x0$29)) + -8|0);
   $96 = +HEAPF32[$95>>2];
   $97 = -$96;
   $98 = +HEAPF32[$x1$210>>2];
   $99 = $97 - $98;
   $100 = +HEAPF32[$94>>2];
   $101 = -$100;
   $102 = ((($x1$210)) + 8|0);
   $103 = +HEAPF32[$102>>2];
   $104 = $101 - $103;
   $105 = ((($T$211)) + -4|0);
   $106 = +HEAPF32[$105>>2];
   $107 = $104 * $106;
   $108 = +HEAPF32[$93>>2];
   $109 = $99 * $108;
   $110 = $107 + $109;
   $111 = (($6) + ($i$212<<2)|0);
   HEAPF32[$111>>2] = $110;
   $112 = +HEAPF32[$93>>2];
   $113 = $104 * $112;
   $114 = +HEAPF32[$105>>2];
   $115 = $99 * $114;
   $116 = $113 - $115;
   $117 = $i$212 | 1;
   $118 = (($6) + ($117<<2)|0);
   HEAPF32[$118>>2] = $116;
   $119 = ((($x1$210)) + 16|0);
   $120 = (($i$212) + 2)|0;
   $121 = ($120|0)<($1|0);
   if ($121) {
    $T$211 = $93;$i$212 = $120;$x0$29 = $94;$x1$210 = $119;
   } else {
    break;
   }
  }
 }
 $init$idx = ((($init)) + 4|0);
 $init$idx$val = HEAP32[$init$idx>>2]|0;
 _mdct_butterflies($init$idx$val,$10,$6,$1);
 $init$idx2$val = HEAP32[$init>>2]|0;
 $init$idx3$val = HEAP32[$9>>2]|0;
 $init$idx4 = ((($init)) + 12|0);
 $init$idx4$val = HEAP32[$init$idx4>>2]|0;
 $122 = $init$idx2$val >> 1;
 $123 = (($5) + ($122<<2)|0);
 $124 = (($init$idx3$val) + ($init$idx2$val<<2)|0);
 $T$0$i = $124;$bit$0$i = $init$idx4$val;$w0$0$i = $5;$w1$0$i = $123;
 while(1) {
  $125 = HEAP32[$bit$0$i>>2]|0;
  $126 = (($123) + ($125<<2)|0);
  $127 = ((($bit$0$i)) + 4|0);
  $128 = HEAP32[$127>>2]|0;
  $129 = (($123) + ($128<<2)|0);
  $130 = ((($126)) + 4|0);
  $131 = +HEAPF32[$130>>2];
  $132 = ((($129)) + 4|0);
  $133 = +HEAPF32[$132>>2];
  $134 = $131 - $133;
  $135 = +HEAPF32[$126>>2];
  $136 = +HEAPF32[$129>>2];
  $137 = $135 + $136;
  $138 = +HEAPF32[$T$0$i>>2];
  $139 = $137 * $138;
  $140 = ((($T$0$i)) + 4|0);
  $141 = +HEAPF32[$140>>2];
  $142 = $134 * $141;
  $143 = $139 + $142;
  $144 = $137 * $141;
  $145 = $134 * $138;
  $146 = $144 - $145;
  $147 = ((($w1$0$i)) + -16|0);
  $148 = $131 + $133;
  $149 = $148 * 0.5;
  $150 = $135 - $136;
  $151 = $150 * 0.5;
  $152 = $149 + $143;
  HEAPF32[$w0$0$i>>2] = $152;
  $153 = $149 - $143;
  $154 = ((($w1$0$i)) + -8|0);
  HEAPF32[$154>>2] = $153;
  $155 = $151 + $146;
  $156 = ((($w0$0$i)) + 4|0);
  HEAPF32[$156>>2] = $155;
  $157 = $146 - $151;
  $158 = ((($w1$0$i)) + -4|0);
  HEAPF32[$158>>2] = $157;
  $159 = ((($bit$0$i)) + 8|0);
  $160 = HEAP32[$159>>2]|0;
  $161 = (($123) + ($160<<2)|0);
  $162 = ((($bit$0$i)) + 12|0);
  $163 = HEAP32[$162>>2]|0;
  $164 = (($123) + ($163<<2)|0);
  $165 = ((($161)) + 4|0);
  $166 = +HEAPF32[$165>>2];
  $167 = ((($164)) + 4|0);
  $168 = +HEAPF32[$167>>2];
  $169 = $166 - $168;
  $170 = +HEAPF32[$161>>2];
  $171 = +HEAPF32[$164>>2];
  $172 = $170 + $171;
  $173 = ((($T$0$i)) + 8|0);
  $174 = +HEAPF32[$173>>2];
  $175 = $172 * $174;
  $176 = ((($T$0$i)) + 12|0);
  $177 = +HEAPF32[$176>>2];
  $178 = $169 * $177;
  $179 = $175 + $178;
  $180 = $172 * $177;
  $181 = $169 * $174;
  $182 = $180 - $181;
  $183 = $166 + $168;
  $184 = $183 * 0.5;
  $185 = $170 - $171;
  $186 = $185 * 0.5;
  $187 = $184 + $179;
  $188 = ((($w0$0$i)) + 8|0);
  HEAPF32[$188>>2] = $187;
  $189 = $184 - $179;
  HEAPF32[$147>>2] = $189;
  $190 = $186 + $182;
  $191 = ((($w0$0$i)) + 12|0);
  HEAPF32[$191>>2] = $190;
  $192 = $182 - $186;
  $193 = ((($w1$0$i)) + -12|0);
  HEAPF32[$193>>2] = $192;
  $194 = ((($T$0$i)) + 16|0);
  $195 = ((($bit$0$i)) + 16|0);
  $196 = ((($w0$0$i)) + 16|0);
  $197 = ($196>>>0)<($147>>>0);
  if ($197) {
   $T$0$i = $194;$bit$0$i = $195;$w0$0$i = $196;$w1$0$i = $147;
  } else {
   break;
  }
 }
 $198 = ($2|0)>(0);
 if (!($198)) {
  STACKTOP = sp;return;
 }
 $199 = (($out) + ($1<<2)|0);
 $200 = (($init$idx3$val) + ($1<<2)|0);
 $201 = ((($init)) + 16|0);
 $T$37 = $200;$i$38 = 0;$w$05 = $5;$x0$36 = $199;
 while(1) {
  $202 = ((($x0$36)) + -4|0);
  $203 = +HEAPF32[$w$05>>2];
  $204 = +HEAPF32[$T$37>>2];
  $205 = $203 * $204;
  $206 = ((($w$05)) + 4|0);
  $207 = +HEAPF32[$206>>2];
  $208 = ((($T$37)) + 4|0);
  $209 = +HEAPF32[$208>>2];
  $210 = $207 * $209;
  $211 = $205 + $210;
  $212 = +HEAPF32[$201>>2];
  $213 = $212 * $211;
  $214 = (($out) + ($i$38<<2)|0);
  HEAPF32[$214>>2] = $213;
  $215 = +HEAPF32[$w$05>>2];
  $216 = +HEAPF32[$208>>2];
  $217 = $215 * $216;
  $218 = +HEAPF32[$206>>2];
  $219 = +HEAPF32[$T$37>>2];
  $220 = $218 * $219;
  $221 = $217 - $220;
  $222 = +HEAPF32[$201>>2];
  $223 = $222 * $221;
  HEAPF32[$202>>2] = $223;
  $224 = ((($w$05)) + 8|0);
  $225 = ((($T$37)) + 8|0);
  $226 = (($i$38) + 1)|0;
  $exitcond = ($226|0)==($2|0);
  if ($exitcond) {
   break;
  } else {
   $T$37 = $225;$i$38 = $226;$w$05 = $224;$x0$36 = $202;
  }
 }
 STACKTOP = sp;return;
}
function _mdct_butterflies($init$0$1$val,$init$0$2$val,$x,$points) {
 $init$0$1$val = $init$0$1$val|0;
 $init$0$2$val = $init$0$2$val|0;
 $x = $x|0;
 $points = $points|0;
 var $$0$i = 0, $$0$i3 = 0, $$pn$i = 0, $$pn$i2 = 0, $0 = 0, $1 = 0, $10 = 0.0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0.0, $12 = 0.0, $120 = 0, $121 = 0.0, $122 = 0.0, $123 = 0, $124 = 0.0, $125 = 0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0;
 var $130 = 0.0, $131 = 0, $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0.0, $140 = 0.0, $141 = 0.0, $142 = 0, $143 = 0, $144 = 0.0, $145 = 0, $146 = 0.0, $147 = 0.0, $148 = 0;
 var $149 = 0.0, $15 = 0.0, $150 = 0, $151 = 0.0, $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0.0, $160 = 0.0, $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0.0;
 var $167 = 0, $168 = 0, $169 = 0.0, $17 = 0.0, $170 = 0, $171 = 0.0, $172 = 0.0, $173 = 0, $174 = 0.0, $175 = 0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0, $18 = 0.0, $180 = 0.0, $181 = 0, $182 = 0.0, $183 = 0.0, $184 = 0.0;
 var $185 = 0.0, $186 = 0.0, $187 = 0.0, $188 = 0.0, $189 = 0.0, $19 = 0, $190 = 0.0, $191 = 0.0, $192 = 0, $193 = 0.0, $194 = 0.0, $195 = 0.0, $196 = 0, $197 = 0.0, $198 = 0, $199 = 0.0, $2 = 0, $20 = 0.0, $200 = 0.0, $201 = 0.0;
 var $202 = 0.0, $203 = 0.0, $204 = 0, $205 = 0.0, $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0.0, $21 = 0.0, $210 = 0.0, $211 = 0.0, $212 = 0.0, $213 = 0.0, $214 = 0.0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0.0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0.0, $224 = 0, $225 = 0.0, $226 = 0.0, $227 = 0, $228 = 0.0, $229 = 0, $23 = 0.0, $230 = 0.0, $231 = 0.0, $232 = 0.0, $233 = 0.0, $234 = 0, $235 = 0.0, $236 = 0, $237 = 0.0, $238 = 0.0;
 var $239 = 0, $24 = 0.0, $240 = 0.0, $241 = 0, $242 = 0.0, $243 = 0.0, $244 = 0.0, $245 = 0.0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0.0, $250 = 0.0, $251 = 0.0, $252 = 0, $253 = 0.0, $254 = 0, $255 = 0.0, $256 = 0.0;
 var $257 = 0, $258 = 0.0, $259 = 0, $26 = 0.0, $260 = 0.0, $261 = 0.0, $262 = 0.0, $263 = 0.0, $264 = 0.0, $265 = 0.0, $266 = 0.0, $267 = 0.0, $268 = 0, $269 = 0.0, $27 = 0.0, $270 = 0, $271 = 0.0, $272 = 0.0, $273 = 0, $274 = 0.0;
 var $275 = 0, $276 = 0.0, $277 = 0.0, $278 = 0.0, $279 = 0.0, $28 = 0.0, $280 = 0.0, $281 = 0.0, $282 = 0.0, $283 = 0.0, $284 = 0.0, $285 = 0.0, $286 = 0, $287 = 0.0, $288 = 0, $289 = 0.0, $29 = 0.0, $290 = 0.0, $291 = 0, $292 = 0.0;
 var $293 = 0, $294 = 0.0, $295 = 0.0, $296 = 0.0, $297 = 0.0, $298 = 0, $299 = 0.0, $3 = 0, $30 = 0, $300 = 0, $301 = 0.0, $302 = 0.0, $303 = 0, $304 = 0.0, $305 = 0, $306 = 0.0, $307 = 0.0, $308 = 0.0, $309 = 0.0, $31 = 0.0;
 var $310 = 0.0, $311 = 0.0, $312 = 0.0, $313 = 0.0, $314 = 0.0, $315 = 0.0, $316 = 0, $317 = 0.0, $318 = 0, $319 = 0.0, $32 = 0, $320 = 0.0, $321 = 0, $322 = 0.0, $323 = 0, $324 = 0.0, $325 = 0.0, $326 = 0.0, $327 = 0.0, $328 = 0.0;
 var $329 = 0.0, $33 = 0.0, $330 = 0.0, $331 = 0.0, $332 = 0.0, $333 = 0, $334 = 0.0, $335 = 0.0, $336 = 0, $337 = 0.0, $338 = 0, $339 = 0.0, $34 = 0.0, $340 = 0.0, $341 = 0.0, $342 = 0.0, $343 = 0.0, $344 = 0.0, $345 = 0.0, $346 = 0.0;
 var $347 = 0.0, $348 = 0.0, $349 = 0.0, $35 = 0, $350 = 0.0, $351 = 0.0, $352 = 0.0, $353 = 0.0, $354 = 0.0, $355 = 0.0, $356 = 0.0, $357 = 0.0, $358 = 0.0, $359 = 0.0, $36 = 0.0, $360 = 0.0, $361 = 0.0, $362 = 0.0, $363 = 0.0, $364 = 0.0;
 var $365 = 0.0, $366 = 0.0, $367 = 0.0, $368 = 0.0, $369 = 0.0, $37 = 0, $370 = 0.0, $371 = 0.0, $372 = 0.0, $373 = 0.0, $374 = 0.0, $375 = 0.0, $376 = 0.0, $377 = 0.0, $378 = 0.0, $379 = 0.0, $38 = 0.0, $380 = 0.0, $381 = 0.0, $382 = 0.0;
 var $383 = 0.0, $384 = 0.0, $385 = 0.0, $386 = 0.0, $387 = 0.0, $388 = 0.0, $389 = 0.0, $39 = 0.0, $390 = 0.0, $391 = 0.0, $392 = 0.0, $393 = 0.0, $394 = 0.0, $395 = 0.0, $396 = 0.0, $397 = 0.0, $398 = 0.0, $399 = 0.0, $4 = 0, $40 = 0.0;
 var $400 = 0.0, $401 = 0.0, $402 = 0.0, $403 = 0.0, $404 = 0.0, $405 = 0.0, $406 = 0.0, $407 = 0.0, $408 = 0.0, $409 = 0.0, $41 = 0.0, $410 = 0.0, $411 = 0.0, $412 = 0.0, $413 = 0.0, $414 = 0.0, $415 = 0.0, $416 = 0.0, $417 = 0.0, $418 = 0.0;
 var $419 = 0.0, $42 = 0.0, $420 = 0.0, $421 = 0.0, $422 = 0.0, $423 = 0.0, $424 = 0.0, $425 = 0.0, $426 = 0.0, $427 = 0.0, $428 = 0.0, $429 = 0.0, $43 = 0, $430 = 0.0, $431 = 0.0, $432 = 0.0, $433 = 0.0, $434 = 0.0, $435 = 0.0, $436 = 0.0;
 var $437 = 0.0, $438 = 0.0, $439 = 0.0, $44 = 0.0, $440 = 0.0, $441 = 0.0, $442 = 0.0, $443 = 0.0, $444 = 0.0, $445 = 0.0, $446 = 0.0, $447 = 0.0, $448 = 0.0, $449 = 0.0, $45 = 0.0, $450 = 0.0, $451 = 0.0, $452 = 0.0, $453 = 0.0, $454 = 0.0;
 var $455 = 0.0, $456 = 0.0, $457 = 0.0, $458 = 0.0, $459 = 0.0, $46 = 0, $460 = 0.0, $461 = 0.0, $462 = 0.0, $463 = 0.0, $464 = 0.0, $465 = 0.0, $466 = 0.0, $467 = 0.0, $468 = 0.0, $469 = 0.0, $47 = 0.0, $470 = 0.0, $471 = 0.0, $472 = 0.0;
 var $473 = 0.0, $474 = 0.0, $475 = 0.0, $476 = 0.0, $477 = 0.0, $478 = 0.0, $479 = 0.0, $48 = 0.0, $480 = 0.0, $481 = 0, $482 = 0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0, $56 = 0.0;
 var $57 = 0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0, $69 = 0.0, $7 = 0.0, $70 = 0.0, $71 = 0, $72 = 0.0, $73 = 0.0, $74 = 0.0;
 var $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0.0, $90 = 0.0, $91 = 0, $92 = 0.0;
 var $93 = 0.0, $94 = 0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $exitcond = 0, $i$06 = 0, $j$03 = 0, $j$12 = 0, $x1$0$i = 0, $x1$0$i4 = 0, $x2$0$i = 0, $x2$0$i1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($init$0$1$val) + -6)|0;
 $1 = ($init$0$1$val|0)>(6);
 if ($1) {
  $2 = (($x) + ($points<<2)|0);
  $3 = $points >> 1;
  $4 = (($x) + ($3<<2)|0);
  $5 = ((($4)) + -32|0);
  $$0$i = $init$0$2$val;$$pn$i = $2;$x2$0$i = $5;
  while(1) {
   $x1$0$i = ((($$pn$i)) + -32|0);
   $6 = ((($$pn$i)) + -8|0);
   $7 = +HEAPF32[$6>>2];
   $8 = ((($x2$0$i)) + 24|0);
   $9 = +HEAPF32[$8>>2];
   $10 = $7 - $9;
   $11 = ((($$pn$i)) + -4|0);
   $12 = +HEAPF32[$11>>2];
   $13 = ((($x2$0$i)) + 28|0);
   $14 = +HEAPF32[$13>>2];
   $15 = $12 - $14;
   $16 = $7 + $9;
   HEAPF32[$6>>2] = $16;
   $17 = +HEAPF32[$13>>2];
   $18 = $12 + $17;
   HEAPF32[$11>>2] = $18;
   $19 = ((($$0$i)) + 4|0);
   $20 = +HEAPF32[$19>>2];
   $21 = $15 * $20;
   $22 = +HEAPF32[$$0$i>>2];
   $23 = $10 * $22;
   $24 = $21 + $23;
   HEAPF32[$8>>2] = $24;
   $25 = +HEAPF32[$$0$i>>2];
   $26 = $15 * $25;
   $27 = +HEAPF32[$19>>2];
   $28 = $10 * $27;
   $29 = $26 - $28;
   HEAPF32[$13>>2] = $29;
   $30 = ((($$pn$i)) + -16|0);
   $31 = +HEAPF32[$30>>2];
   $32 = ((($x2$0$i)) + 16|0);
   $33 = +HEAPF32[$32>>2];
   $34 = $31 - $33;
   $35 = ((($$pn$i)) + -12|0);
   $36 = +HEAPF32[$35>>2];
   $37 = ((($x2$0$i)) + 20|0);
   $38 = +HEAPF32[$37>>2];
   $39 = $36 - $38;
   $40 = $31 + $33;
   HEAPF32[$30>>2] = $40;
   $41 = +HEAPF32[$37>>2];
   $42 = $36 + $41;
   HEAPF32[$35>>2] = $42;
   $43 = ((($$0$i)) + 20|0);
   $44 = +HEAPF32[$43>>2];
   $45 = $39 * $44;
   $46 = ((($$0$i)) + 16|0);
   $47 = +HEAPF32[$46>>2];
   $48 = $34 * $47;
   $49 = $45 + $48;
   HEAPF32[$32>>2] = $49;
   $50 = +HEAPF32[$46>>2];
   $51 = $39 * $50;
   $52 = +HEAPF32[$43>>2];
   $53 = $34 * $52;
   $54 = $51 - $53;
   HEAPF32[$37>>2] = $54;
   $55 = ((($$pn$i)) + -24|0);
   $56 = +HEAPF32[$55>>2];
   $57 = ((($x2$0$i)) + 8|0);
   $58 = +HEAPF32[$57>>2];
   $59 = $56 - $58;
   $60 = ((($$pn$i)) + -20|0);
   $61 = +HEAPF32[$60>>2];
   $62 = ((($x2$0$i)) + 12|0);
   $63 = +HEAPF32[$62>>2];
   $64 = $61 - $63;
   $65 = $56 + $58;
   HEAPF32[$55>>2] = $65;
   $66 = +HEAPF32[$62>>2];
   $67 = $61 + $66;
   HEAPF32[$60>>2] = $67;
   $68 = ((($$0$i)) + 36|0);
   $69 = +HEAPF32[$68>>2];
   $70 = $64 * $69;
   $71 = ((($$0$i)) + 32|0);
   $72 = +HEAPF32[$71>>2];
   $73 = $59 * $72;
   $74 = $70 + $73;
   HEAPF32[$57>>2] = $74;
   $75 = +HEAPF32[$71>>2];
   $76 = $64 * $75;
   $77 = +HEAPF32[$68>>2];
   $78 = $59 * $77;
   $79 = $76 - $78;
   HEAPF32[$62>>2] = $79;
   $80 = +HEAPF32[$x1$0$i>>2];
   $81 = +HEAPF32[$x2$0$i>>2];
   $82 = $80 - $81;
   $83 = ((($$pn$i)) + -28|0);
   $84 = +HEAPF32[$83>>2];
   $85 = ((($x2$0$i)) + 4|0);
   $86 = +HEAPF32[$85>>2];
   $87 = $84 - $86;
   $88 = $80 + $81;
   HEAPF32[$x1$0$i>>2] = $88;
   $89 = +HEAPF32[$85>>2];
   $90 = $84 + $89;
   HEAPF32[$83>>2] = $90;
   $91 = ((($$0$i)) + 52|0);
   $92 = +HEAPF32[$91>>2];
   $93 = $87 * $92;
   $94 = ((($$0$i)) + 48|0);
   $95 = +HEAPF32[$94>>2];
   $96 = $82 * $95;
   $97 = $93 + $96;
   HEAPF32[$x2$0$i>>2] = $97;
   $98 = +HEAPF32[$94>>2];
   $99 = $87 * $98;
   $100 = +HEAPF32[$91>>2];
   $101 = $82 * $100;
   $102 = $99 - $101;
   HEAPF32[$85>>2] = $102;
   $103 = ((($x2$0$i)) + -32|0);
   $104 = ((($$0$i)) + 64|0);
   $105 = ($103>>>0)<($x>>>0);
   if ($105) {
    break;
   } else {
    $$0$i = $104;$$pn$i = $x1$0$i;$x2$0$i = $103;
   }
  }
 }
 $106 = ($0|0)>(1);
 if ($106) {
  $i$06 = 1;
  while(1) {
   $107 = 1 << $i$06;
   $108 = ($i$06|0)==(31);
   if (!($108)) {
    $109 = $points >> $i$06;
    $110 = 4 << $i$06;
    $111 = $109 >> 1;
    $j$03 = 0;
    while(1) {
     $113 = Math_imul($j$03, $109)|0;
     $114 = (($x) + ($113<<2)|0);
     $115 = (($114) + ($109<<2)|0);
     $116 = (($114) + ($111<<2)|0);
     $117 = ((($116)) + -32|0);
     $$0$i3 = $init$0$2$val;$$pn$i2 = $115;$x2$0$i1 = $117;
     while(1) {
      $x1$0$i4 = ((($$pn$i2)) + -32|0);
      $118 = ((($$pn$i2)) + -8|0);
      $119 = +HEAPF32[$118>>2];
      $120 = ((($x2$0$i1)) + 24|0);
      $121 = +HEAPF32[$120>>2];
      $122 = $119 - $121;
      $123 = ((($$pn$i2)) + -4|0);
      $124 = +HEAPF32[$123>>2];
      $125 = ((($x2$0$i1)) + 28|0);
      $126 = +HEAPF32[$125>>2];
      $127 = $124 - $126;
      $128 = $119 + $121;
      HEAPF32[$118>>2] = $128;
      $129 = +HEAPF32[$125>>2];
      $130 = $124 + $129;
      HEAPF32[$123>>2] = $130;
      $131 = ((($$0$i3)) + 4|0);
      $132 = +HEAPF32[$131>>2];
      $133 = $127 * $132;
      $134 = +HEAPF32[$$0$i3>>2];
      $135 = $122 * $134;
      $136 = $133 + $135;
      HEAPF32[$120>>2] = $136;
      $137 = +HEAPF32[$$0$i3>>2];
      $138 = $127 * $137;
      $139 = +HEAPF32[$131>>2];
      $140 = $122 * $139;
      $141 = $138 - $140;
      HEAPF32[$125>>2] = $141;
      $142 = (($$0$i3) + ($110<<2)|0);
      $143 = ((($$pn$i2)) + -16|0);
      $144 = +HEAPF32[$143>>2];
      $145 = ((($x2$0$i1)) + 16|0);
      $146 = +HEAPF32[$145>>2];
      $147 = $144 - $146;
      $148 = ((($$pn$i2)) + -12|0);
      $149 = +HEAPF32[$148>>2];
      $150 = ((($x2$0$i1)) + 20|0);
      $151 = +HEAPF32[$150>>2];
      $152 = $149 - $151;
      $153 = $144 + $146;
      HEAPF32[$143>>2] = $153;
      $154 = +HEAPF32[$150>>2];
      $155 = $149 + $154;
      HEAPF32[$148>>2] = $155;
      $156 = ((($142)) + 4|0);
      $157 = +HEAPF32[$156>>2];
      $158 = $152 * $157;
      $159 = +HEAPF32[$142>>2];
      $160 = $147 * $159;
      $161 = $158 + $160;
      HEAPF32[$145>>2] = $161;
      $162 = +HEAPF32[$142>>2];
      $163 = $152 * $162;
      $164 = +HEAPF32[$156>>2];
      $165 = $147 * $164;
      $166 = $163 - $165;
      HEAPF32[$150>>2] = $166;
      $167 = (($142) + ($110<<2)|0);
      $168 = ((($$pn$i2)) + -24|0);
      $169 = +HEAPF32[$168>>2];
      $170 = ((($x2$0$i1)) + 8|0);
      $171 = +HEAPF32[$170>>2];
      $172 = $169 - $171;
      $173 = ((($$pn$i2)) + -20|0);
      $174 = +HEAPF32[$173>>2];
      $175 = ((($x2$0$i1)) + 12|0);
      $176 = +HEAPF32[$175>>2];
      $177 = $174 - $176;
      $178 = $169 + $171;
      HEAPF32[$168>>2] = $178;
      $179 = +HEAPF32[$175>>2];
      $180 = $174 + $179;
      HEAPF32[$173>>2] = $180;
      $181 = ((($167)) + 4|0);
      $182 = +HEAPF32[$181>>2];
      $183 = $177 * $182;
      $184 = +HEAPF32[$167>>2];
      $185 = $172 * $184;
      $186 = $183 + $185;
      HEAPF32[$170>>2] = $186;
      $187 = +HEAPF32[$167>>2];
      $188 = $177 * $187;
      $189 = +HEAPF32[$181>>2];
      $190 = $172 * $189;
      $191 = $188 - $190;
      HEAPF32[$175>>2] = $191;
      $192 = (($167) + ($110<<2)|0);
      $193 = +HEAPF32[$x1$0$i4>>2];
      $194 = +HEAPF32[$x2$0$i1>>2];
      $195 = $193 - $194;
      $196 = ((($$pn$i2)) + -28|0);
      $197 = +HEAPF32[$196>>2];
      $198 = ((($x2$0$i1)) + 4|0);
      $199 = +HEAPF32[$198>>2];
      $200 = $197 - $199;
      $201 = $193 + $194;
      HEAPF32[$x1$0$i4>>2] = $201;
      $202 = +HEAPF32[$198>>2];
      $203 = $197 + $202;
      HEAPF32[$196>>2] = $203;
      $204 = ((($192)) + 4|0);
      $205 = +HEAPF32[$204>>2];
      $206 = $200 * $205;
      $207 = +HEAPF32[$192>>2];
      $208 = $195 * $207;
      $209 = $206 + $208;
      HEAPF32[$x2$0$i1>>2] = $209;
      $210 = +HEAPF32[$192>>2];
      $211 = $200 * $210;
      $212 = +HEAPF32[$204>>2];
      $213 = $195 * $212;
      $214 = $211 - $213;
      HEAPF32[$198>>2] = $214;
      $215 = (($192) + ($110<<2)|0);
      $216 = ((($x2$0$i1)) + -32|0);
      $217 = ($216>>>0)<($114>>>0);
      if ($217) {
       break;
      } else {
       $$0$i3 = $215;$$pn$i2 = $x1$0$i4;$x2$0$i1 = $216;
      }
     }
     $218 = (($j$03) + 1)|0;
     $219 = ($218|0)<($107|0);
     if ($219) {
      $j$03 = $218;
     } else {
      break;
     }
    }
   }
   $220 = (($i$06) + 1)|0;
   $exitcond = ($220|0)==($0|0);
   if ($exitcond) {
    break;
   } else {
    $i$06 = $220;
   }
  }
 }
 $112 = ($points|0)>(0);
 if ($112) {
  $j$12 = 0;
 } else {
  return;
 }
 while(1) {
  $221 = (($x) + ($j$12<<2)|0);
  $222 = ((($221)) + 120|0);
  $223 = +HEAPF32[$222>>2];
  $224 = ((($221)) + 56|0);
  $225 = +HEAPF32[$224>>2];
  $226 = $223 - $225;
  $227 = ((($221)) + 124|0);
  $228 = +HEAPF32[$227>>2];
  $229 = ((($221)) + 60|0);
  $230 = +HEAPF32[$229>>2];
  $231 = $228 - $230;
  $232 = $223 + $225;
  HEAPF32[$222>>2] = $232;
  $233 = $228 + $230;
  HEAPF32[$227>>2] = $233;
  HEAPF32[$224>>2] = $226;
  HEAPF32[$229>>2] = $231;
  $234 = ((($221)) + 112|0);
  $235 = +HEAPF32[$234>>2];
  $236 = ((($221)) + 48|0);
  $237 = +HEAPF32[$236>>2];
  $238 = $235 - $237;
  $239 = ((($221)) + 116|0);
  $240 = +HEAPF32[$239>>2];
  $241 = ((($221)) + 52|0);
  $242 = +HEAPF32[$241>>2];
  $243 = $240 - $242;
  $244 = $235 + $237;
  HEAPF32[$234>>2] = $244;
  $245 = $240 + $242;
  HEAPF32[$239>>2] = $245;
  $246 = $238 * 0.92387950420379638;
  $247 = $243 * 0.38268342614173889;
  $248 = $246 - $247;
  HEAPF32[$236>>2] = $248;
  $249 = $238 * 0.38268342614173889;
  $250 = $243 * 0.92387950420379638;
  $251 = $249 + $250;
  HEAPF32[$241>>2] = $251;
  $252 = ((($221)) + 104|0);
  $253 = +HEAPF32[$252>>2];
  $254 = ((($221)) + 40|0);
  $255 = +HEAPF32[$254>>2];
  $256 = $253 - $255;
  $257 = ((($221)) + 108|0);
  $258 = +HEAPF32[$257>>2];
  $259 = ((($221)) + 44|0);
  $260 = +HEAPF32[$259>>2];
  $261 = $258 - $260;
  $262 = $253 + $255;
  HEAPF32[$252>>2] = $262;
  $263 = $258 + $260;
  HEAPF32[$257>>2] = $263;
  $264 = $256 - $261;
  $265 = $264 * 0.70710676908493042;
  HEAPF32[$254>>2] = $265;
  $266 = $256 + $261;
  $267 = $266 * 0.70710676908493042;
  HEAPF32[$259>>2] = $267;
  $268 = ((($221)) + 96|0);
  $269 = +HEAPF32[$268>>2];
  $270 = ((($221)) + 32|0);
  $271 = +HEAPF32[$270>>2];
  $272 = $269 - $271;
  $273 = ((($221)) + 100|0);
  $274 = +HEAPF32[$273>>2];
  $275 = ((($221)) + 36|0);
  $276 = +HEAPF32[$275>>2];
  $277 = $274 - $276;
  $278 = $269 + $271;
  HEAPF32[$268>>2] = $278;
  $279 = $274 + $276;
  HEAPF32[$273>>2] = $279;
  $280 = $272 * 0.38268342614173889;
  $281 = $277 * 0.92387950420379638;
  $282 = $280 - $281;
  $283 = $277 * 0.38268342614173889;
  $284 = $272 * 0.92387950420379638;
  $285 = $284 + $283;
  $286 = ((($221)) + 88|0);
  $287 = +HEAPF32[$286>>2];
  $288 = ((($221)) + 24|0);
  $289 = +HEAPF32[$288>>2];
  $290 = $287 - $289;
  $291 = ((($221)) + 28|0);
  $292 = +HEAPF32[$291>>2];
  $293 = ((($221)) + 92|0);
  $294 = +HEAPF32[$293>>2];
  $295 = $292 - $294;
  $296 = $287 + $289;
  HEAPF32[$286>>2] = $296;
  $297 = $292 + $294;
  HEAPF32[$293>>2] = $297;
  HEAPF32[$291>>2] = $290;
  $298 = ((($221)) + 16|0);
  $299 = +HEAPF32[$298>>2];
  $300 = ((($221)) + 80|0);
  $301 = +HEAPF32[$300>>2];
  $302 = $299 - $301;
  $303 = ((($221)) + 20|0);
  $304 = +HEAPF32[$303>>2];
  $305 = ((($221)) + 84|0);
  $306 = +HEAPF32[$305>>2];
  $307 = $304 - $306;
  $308 = $299 + $301;
  HEAPF32[$300>>2] = $308;
  $309 = $304 + $306;
  HEAPF32[$305>>2] = $309;
  $310 = $307 * 0.92387950420379638;
  $311 = $302 * 0.38268342614173889;
  $312 = $311 + $310;
  $313 = $307 * 0.38268342614173889;
  $314 = $302 * 0.92387950420379638;
  $315 = $313 - $314;
  $316 = ((($221)) + 8|0);
  $317 = +HEAPF32[$316>>2];
  $318 = ((($221)) + 72|0);
  $319 = +HEAPF32[$318>>2];
  $320 = $317 - $319;
  $321 = ((($221)) + 12|0);
  $322 = +HEAPF32[$321>>2];
  $323 = ((($221)) + 76|0);
  $324 = +HEAPF32[$323>>2];
  $325 = $322 - $324;
  $326 = $317 + $319;
  HEAPF32[$318>>2] = $326;
  $327 = $322 + $324;
  HEAPF32[$323>>2] = $327;
  $328 = $320 + $325;
  $329 = $328 * 0.70710676908493042;
  $330 = $325 - $320;
  $331 = $330 * 0.70710676908493042;
  $332 = +HEAPF32[$221>>2];
  $333 = ((($221)) + 64|0);
  $334 = +HEAPF32[$333>>2];
  $335 = $332 - $334;
  $336 = ((($221)) + 4|0);
  $337 = +HEAPF32[$336>>2];
  $338 = ((($221)) + 68|0);
  $339 = +HEAPF32[$338>>2];
  $340 = $337 - $339;
  $341 = $332 + $334;
  HEAPF32[$333>>2] = $341;
  $342 = $337 + $339;
  HEAPF32[$338>>2] = $342;
  $343 = $340 * 0.38268342614173889;
  $344 = $335 * 0.92387950420379638;
  $345 = $344 + $343;
  $346 = $340 * 0.92387950420379638;
  $347 = $335 * 0.38268342614173889;
  $348 = $346 - $347;
  $349 = $348 - $285;
  $350 = $345 - $282;
  $351 = $282 + $345;
  $352 = $285 + $348;
  $353 = $349 + $350;
  $354 = $353 * 0.70710676908493042;
  $355 = $349 - $350;
  $356 = $355 * 0.70710676908493042;
  $357 = +HEAPF32[$259>>2];
  $358 = $331 - $357;
  $359 = +HEAPF32[$254>>2];
  $360 = $359 - $329;
  $361 = $329 + $359;
  $362 = $331 + $357;
  $363 = +HEAPF32[$236>>2];
  $364 = $363 - $312;
  $365 = +HEAPF32[$241>>2];
  $366 = $365 - $315;
  $367 = $312 + $363;
  $368 = $315 + $365;
  $369 = $364 - $366;
  $370 = $369 * 0.70710676908493042;
  $371 = $364 + $366;
  $372 = $371 * 0.70710676908493042;
  $373 = +HEAPF32[$224>>2];
  $374 = $373 - $295;
  $375 = +HEAPF32[$229>>2];
  $376 = $375 - $290;
  $377 = $373 + $295;
  $378 = $375 + $290;
  $379 = $358 + $374;
  $380 = $374 - $358;
  $381 = $354 + $370;
  $382 = $370 - $354;
  $383 = $379 + $381;
  HEAPF32[$288>>2] = $383;
  $384 = $379 - $381;
  HEAPF32[$298>>2] = $384;
  $385 = $372 - $356;
  $386 = $376 - $360;
  $387 = $380 + $385;
  HEAPF32[$221>>2] = $387;
  $388 = $380 - $385;
  HEAPF32[$316>>2] = $388;
  $389 = $356 + $372;
  $390 = $360 + $376;
  $391 = $382 + $386;
  HEAPF32[$321>>2] = $391;
  $392 = $386 - $382;
  HEAPF32[$336>>2] = $392;
  $393 = $389 + $390;
  HEAPF32[$291>>2] = $393;
  $394 = $390 - $389;
  HEAPF32[$303>>2] = $394;
  $395 = $361 + $377;
  $396 = $377 - $361;
  $397 = $351 + $367;
  $398 = $367 - $351;
  $399 = $397 + $395;
  HEAPF32[$224>>2] = $399;
  $400 = $395 - $397;
  HEAPF32[$236>>2] = $400;
  $401 = $368 - $352;
  $402 = $378 - $362;
  $403 = $401 + $396;
  HEAPF32[$270>>2] = $403;
  $404 = $396 - $401;
  HEAPF32[$254>>2] = $404;
  $405 = $352 + $368;
  $406 = $362 + $378;
  $407 = $398 + $402;
  HEAPF32[$259>>2] = $407;
  $408 = $402 - $398;
  HEAPF32[$275>>2] = $408;
  $409 = $405 + $406;
  HEAPF32[$229>>2] = $409;
  $410 = $406 - $405;
  HEAPF32[$241>>2] = $410;
  $411 = +HEAPF32[$273>>2];
  $412 = $342 - $411;
  $413 = +HEAPF32[$268>>2];
  $414 = $341 - $413;
  $415 = $341 + $413;
  $416 = $342 + $411;
  $417 = $412 + $414;
  $418 = $417 * 0.70710676908493042;
  $419 = $412 - $414;
  $420 = $419 * 0.70710676908493042;
  $421 = +HEAPF32[$323>>2];
  $422 = +HEAPF32[$257>>2];
  $423 = $421 - $422;
  $424 = +HEAPF32[$252>>2];
  $425 = +HEAPF32[$318>>2];
  $426 = $424 - $425;
  $427 = $424 + $425;
  $428 = $421 + $422;
  $429 = +HEAPF32[$234>>2];
  $430 = +HEAPF32[$300>>2];
  $431 = $429 - $430;
  $432 = +HEAPF32[$239>>2];
  $433 = +HEAPF32[$305>>2];
  $434 = $432 - $433;
  $435 = $429 + $430;
  $436 = $432 + $433;
  $437 = $431 - $434;
  $438 = $437 * 0.70710676908493042;
  $439 = $431 + $434;
  $440 = $439 * 0.70710676908493042;
  $441 = +HEAPF32[$222>>2];
  $442 = +HEAPF32[$286>>2];
  $443 = $441 - $442;
  $444 = +HEAPF32[$227>>2];
  $445 = +HEAPF32[$293>>2];
  $446 = $444 - $445;
  $447 = $441 + $442;
  $448 = $444 + $445;
  $449 = $423 + $443;
  $450 = $443 - $423;
  $451 = $418 + $438;
  $452 = $438 - $418;
  $453 = $449 + $451;
  HEAPF32[$286>>2] = $453;
  $454 = $449 - $451;
  HEAPF32[$300>>2] = $454;
  $455 = $440 - $420;
  $456 = $446 - $426;
  $457 = $450 + $455;
  HEAPF32[$333>>2] = $457;
  $458 = $450 - $455;
  HEAPF32[$318>>2] = $458;
  $459 = $420 + $440;
  $460 = $426 + $446;
  $461 = $452 + $456;
  HEAPF32[$323>>2] = $461;
  $462 = $456 - $452;
  HEAPF32[$338>>2] = $462;
  $463 = $459 + $460;
  HEAPF32[$293>>2] = $463;
  $464 = $460 - $459;
  HEAPF32[$305>>2] = $464;
  $465 = $427 + $447;
  $466 = $447 - $427;
  $467 = $415 + $435;
  $468 = $435 - $415;
  $469 = $467 + $465;
  HEAPF32[$222>>2] = $469;
  $470 = $465 - $467;
  HEAPF32[$234>>2] = $470;
  $471 = $436 - $416;
  $472 = $448 - $428;
  $473 = $471 + $466;
  HEAPF32[$268>>2] = $473;
  $474 = $466 - $471;
  HEAPF32[$252>>2] = $474;
  $475 = $416 + $436;
  $476 = $428 + $448;
  $477 = $468 + $472;
  HEAPF32[$257>>2] = $477;
  $478 = $472 - $468;
  HEAPF32[$273>>2] = $478;
  $479 = $475 + $476;
  HEAPF32[$227>>2] = $479;
  $480 = $476 - $475;
  HEAPF32[$239>>2] = $480;
  $481 = (($j$12) + 32)|0;
  $482 = ($481|0)<($points|0);
  if ($482) {
   $j$12 = $481;
  } else {
   break;
  }
 }
 return;
}
function _drft_forward($l,$data) {
 $l = $l|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i = 0, $exitcond7$i = 0, $i$01$i = 0, $iw$02$i = 0, $k1$05$i = 0, $l2$04$i = 0, $na$0$$i = 0, $na$03$i = 0, $na$2$i = 0, $na$2$i$lcssa = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$l>>2]|0;
 $1 = ($0|0)==(1);
 if ($1) {
  return;
 }
 $2 = ((($l)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + ($0<<2)|0);
 $5 = ((($l)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($6)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)>(0);
 if (!($9)) {
  return;
 }
 $10 = (($8) + 1)|0;
 $iw$02$i = $0;$k1$05$i = 0;$l2$04$i = $0;$na$03$i = 1;
 while(1) {
  $11 = (($10) - ($k1$05$i))|0;
  $12 = (($6) + ($11<<2)|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = (($l2$04$i|0) / ($13|0))&-1;
  $15 = (($0|0) / ($l2$04$i|0))&-1;
  $16 = Math_imul($15, $14)|0;
  $17 = (($13) + -1)|0;
  $18 = Math_imul($15, $17)|0;
  $19 = (($iw$02$i) - ($18))|0;
  $20 = (1 - ($na$03$i))|0;
  L9: do {
   switch ($13|0) {
   case 4:  {
    $21 = (($19) + ($15))|0;
    $22 = (($21) + ($15))|0;
    $23 = ($20|0)==(0);
    $24 = (($4) + ($19<<2)|0);
    $25 = ((($24)) + -4|0);
    $26 = (($4) + ($21<<2)|0);
    $27 = ((($26)) + -4|0);
    $28 = (($4) + ($22<<2)|0);
    $29 = ((($28)) + -4|0);
    if ($23) {
     _dradf4($15,$14,$data,$3,$25,$27,$29);
     $na$2$i = 0;
     break L9;
    } else {
     _dradf4($15,$14,$3,$data,$25,$27,$29);
     $na$2$i = $20;
     break L9;
    }
    break;
   }
   case 2:  {
    $30 = ($20|0)==(0);
    $31 = (($4) + ($19<<2)|0);
    $32 = ((($31)) + -4|0);
    if ($30) {
     _dradf2($15,$14,$data,$3,$32);
     $na$2$i = 0;
     break L9;
    } else {
     _dradf2($15,$14,$3,$data,$32);
     $na$2$i = $20;
     break L9;
    }
    break;
   }
   default: {
    $33 = ($15|0)==(1);
    $na$0$$i = $33 ? $na$03$i : $20;
    $34 = ($na$0$$i|0)==(0);
    $35 = (($4) + ($19<<2)|0);
    $36 = ((($35)) + -4|0);
    if ($34) {
     _dradfg($15,$13,$14,$16,$data,$data,$data,$3,$3,$36);
     $na$2$i = 1;
     break L9;
    } else {
     _dradfg($15,$13,$14,$16,$3,$3,$3,$data,$data,$36);
     $na$2$i = 0;
     break L9;
    }
   }
   }
  } while(0);
  $37 = (($k1$05$i) + 1)|0;
  $exitcond7$i = ($37|0)==($8|0);
  if ($exitcond7$i) {
   $na$2$i$lcssa = $na$2$i;
   break;
  } else {
   $iw$02$i = $19;$k1$05$i = $37;$l2$04$i = $14;$na$03$i = $na$2$i;
  }
 }
 $38 = ($na$2$i$lcssa|0)!=(1);
 $39 = ($0|0)>(0);
 $or$cond$i = $39 & $38;
 if ($or$cond$i) {
  $i$01$i = 0;
 } else {
  return;
 }
 while(1) {
  $40 = (($3) + ($i$01$i<<2)|0);
  $41 = HEAP32[$40>>2]|0;
  $42 = (($data) + ($i$01$i<<2)|0);
  HEAP32[$42>>2] = $41;
  $43 = (($i$01$i) + 1)|0;
  $exitcond$i = ($43|0)==($0|0);
  if ($exitcond$i) {
   break;
  } else {
   $i$01$i = $43;
  }
 }
 return;
}
function _dradf4($ido,$l1,$cc,$ch,$wa1,$wa2,$wa3) {
 $ido = $ido|0;
 $l1 = $l1|0;
 $cc = $cc|0;
 $ch = $ch|0;
 $wa1 = $wa1|0;
 $wa2 = $wa2|0;
 $wa3 = $wa3|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $100 = 0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0, $105 = 0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0.0, $110 = 0, $111 = 0, $112 = 0.0, $113 = 0, $114 = 0.0, $115 = 0;
 var $116 = 0, $117 = 0.0, $118 = 0, $119 = 0.0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0.0, $124 = 0, $125 = 0.0, $126 = 0, $127 = 0, $128 = 0, $129 = 0.0, $13 = 0.0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0.0, $144 = 0, $145 = 0.0, $146 = 0.0, $147 = 0.0, $148 = 0.0, $149 = 0.0, $15 = 0.0, $150 = 0, $151 = 0;
 var $152 = 0.0, $153 = 0.0, $154 = 0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0, $159 = 0, $16 = 0.0, $160 = 0, $161 = 0, $162 = 0, $163 = 0.0, $164 = 0.0, $165 = 0, $166 = 0.0, $167 = 0.0, $168 = 0, $169 = 0, $17 = 0.0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0, $55 = 0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0, $65 = 0, $66 = 0.0, $67 = 0, $68 = 0;
 var $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0.0, $80 = 0, $81 = 0, $82 = 0.0, $83 = 0, $84 = 0, $85 = 0.0, $86 = 0.0;
 var $87 = 0, $88 = 0.0, $89 = 0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0.0, $99 = 0.0, $exitcond = 0, $exitcond23 = 0, $exitcond24 = 0, $i$09 = 0, $k$019 = 0, $k$112 = 0;
 var $k$25 = 0, $t1$018 = 0, $t1$111 = 0, $t1$24 = 0, $t2$017 = 0, $t2$18 = 0, $t2$23 = 0, $t3$016 = 0, $t4$015 = 0, $t4$17 = 0, $t4$22 = 0, $t5$06 = 0, $t6$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = Math_imul($l1, $ido)|0;
 $1 = $0 << 1;
 $2 = ($l1|0)>(0);
 if ($2) {
  $3 = ($0*3)|0;
  $4 = $ido << 2;
  $5 = (($4) + -1)|0;
  $6 = $ido << 1;
  $k$019 = 0;$t1$018 = $0;$t2$017 = $3;$t3$016 = 0;$t4$015 = $1;
  while(1) {
   $7 = (($cc) + ($t1$018<<2)|0);
   $8 = +HEAPF32[$7>>2];
   $9 = (($cc) + ($t2$017<<2)|0);
   $10 = +HEAPF32[$9>>2];
   $11 = $8 + $10;
   $12 = (($cc) + ($t3$016<<2)|0);
   $13 = +HEAPF32[$12>>2];
   $14 = (($cc) + ($t4$015<<2)|0);
   $15 = +HEAPF32[$14>>2];
   $16 = $13 + $15;
   $17 = $11 + $16;
   $18 = $t3$016 << 2;
   $19 = (($ch) + ($18<<2)|0);
   HEAPF32[$19>>2] = $17;
   $20 = $16 - $11;
   $21 = (($5) + ($18))|0;
   $22 = (($ch) + ($21<<2)|0);
   HEAPF32[$22>>2] = $20;
   $23 = +HEAPF32[$12>>2];
   $24 = +HEAPF32[$14>>2];
   $25 = $23 - $24;
   $26 = (($18) + ($6))|0;
   $27 = (($26) + -1)|0;
   $28 = (($ch) + ($27<<2)|0);
   HEAPF32[$28>>2] = $25;
   $29 = +HEAPF32[$9>>2];
   $30 = +HEAPF32[$7>>2];
   $31 = $29 - $30;
   $32 = (($ch) + ($26<<2)|0);
   HEAPF32[$32>>2] = $31;
   $33 = (($t1$018) + ($ido))|0;
   $34 = (($t2$017) + ($ido))|0;
   $35 = (($t3$016) + ($ido))|0;
   $36 = (($t4$015) + ($ido))|0;
   $37 = (($k$019) + 1)|0;
   $exitcond24 = ($37|0)==($l1|0);
   if ($exitcond24) {
    break;
   } else {
    $k$019 = $37;$t1$018 = $33;$t2$017 = $34;$t3$016 = $35;$t4$015 = $36;
   }
  }
 }
 $38 = ($ido|0)<(2);
 if ($38) {
  return;
 }
 $39 = ($ido|0)==(2);
 if (!($39)) {
  if ($2) {
   $40 = $ido << 1;
   $k$112 = 0;$t1$111 = 0;
   while(1) {
    $41 = $t1$111 << 2;
    $42 = (($41) + ($40))|0;
    $i$09 = 2;$t2$18 = $t1$111;$t4$17 = $41;$t5$06 = $42;
    while(1) {
     $43 = (($t2$18) + 2)|0;
     $44 = (($t4$17) + 2)|0;
     $45 = (($t5$06) + -2)|0;
     $46 = (($43) + ($0))|0;
     $47 = (($i$09) + -2)|0;
     $48 = (($wa1) + ($47<<2)|0);
     $49 = +HEAPF32[$48>>2];
     $50 = (($46) + -1)|0;
     $51 = (($cc) + ($50<<2)|0);
     $52 = +HEAPF32[$51>>2];
     $53 = $49 * $52;
     $54 = (($i$09) + -1)|0;
     $55 = (($wa1) + ($54<<2)|0);
     $56 = +HEAPF32[$55>>2];
     $57 = (($cc) + ($46<<2)|0);
     $58 = +HEAPF32[$57>>2];
     $59 = $56 * $58;
     $60 = $53 + $59;
     $61 = $49 * $58;
     $62 = $52 * $56;
     $63 = $61 - $62;
     $64 = (($46) + ($0))|0;
     $65 = (($wa2) + ($47<<2)|0);
     $66 = +HEAPF32[$65>>2];
     $67 = (($64) + -1)|0;
     $68 = (($cc) + ($67<<2)|0);
     $69 = +HEAPF32[$68>>2];
     $70 = $66 * $69;
     $71 = (($wa2) + ($54<<2)|0);
     $72 = +HEAPF32[$71>>2];
     $73 = (($cc) + ($64<<2)|0);
     $74 = +HEAPF32[$73>>2];
     $75 = $72 * $74;
     $76 = $70 + $75;
     $77 = $66 * $74;
     $78 = $69 * $72;
     $79 = $77 - $78;
     $80 = (($64) + ($0))|0;
     $81 = (($wa3) + ($47<<2)|0);
     $82 = +HEAPF32[$81>>2];
     $83 = (($80) + -1)|0;
     $84 = (($cc) + ($83<<2)|0);
     $85 = +HEAPF32[$84>>2];
     $86 = $82 * $85;
     $87 = (($wa3) + ($54<<2)|0);
     $88 = +HEAPF32[$87>>2];
     $89 = (($cc) + ($80<<2)|0);
     $90 = +HEAPF32[$89>>2];
     $91 = $88 * $90;
     $92 = $86 + $91;
     $93 = $82 * $90;
     $94 = $85 * $88;
     $95 = $93 - $94;
     $96 = $60 + $92;
     $97 = $92 - $60;
     $98 = $63 + $95;
     $99 = $63 - $95;
     $100 = (($cc) + ($43<<2)|0);
     $101 = +HEAPF32[$100>>2];
     $102 = $79 + $101;
     $103 = $101 - $79;
     $104 = (($t2$18) + 1)|0;
     $105 = (($cc) + ($104<<2)|0);
     $106 = +HEAPF32[$105>>2];
     $107 = $76 + $106;
     $108 = $106 - $76;
     $109 = $96 + $107;
     $110 = $t4$17 | 1;
     $111 = (($ch) + ($110<<2)|0);
     HEAPF32[$111>>2] = $109;
     $112 = $102 + $98;
     $113 = (($ch) + ($44<<2)|0);
     HEAPF32[$113>>2] = $112;
     $114 = $108 - $99;
     $115 = (($t5$06) + -3)|0;
     $116 = (($ch) + ($115<<2)|0);
     HEAPF32[$116>>2] = $114;
     $117 = $97 - $103;
     $118 = (($ch) + ($45<<2)|0);
     HEAPF32[$118>>2] = $117;
     $119 = $99 + $108;
     $120 = (($44) + ($40))|0;
     $121 = (($120) + -1)|0;
     $122 = (($ch) + ($121<<2)|0);
     HEAPF32[$122>>2] = $119;
     $123 = $103 + $97;
     $124 = (($ch) + ($120<<2)|0);
     HEAPF32[$124>>2] = $123;
     $125 = $107 - $96;
     $126 = (($45) + ($40))|0;
     $127 = (($126) + -1)|0;
     $128 = (($ch) + ($127<<2)|0);
     HEAPF32[$128>>2] = $125;
     $129 = $98 - $102;
     $130 = (($ch) + ($126<<2)|0);
     HEAPF32[$130>>2] = $129;
     $131 = (($i$09) + 2)|0;
     $132 = ($131|0)<($ido|0);
     if ($132) {
      $i$09 = $131;$t2$18 = $43;$t4$17 = $44;$t5$06 = $45;
     } else {
      break;
     }
    }
    $133 = (($t1$111) + ($ido))|0;
    $134 = (($k$112) + 1)|0;
    $exitcond23 = ($134|0)==($l1|0);
    if ($exitcond23) {
     break;
    } else {
     $k$112 = $134;$t1$111 = $133;
    }
   }
  }
  $135 = $ido & 1;
  $136 = ($135|0)==(0);
  if (!($136)) {
   return;
  }
 }
 $137 = (($ido) + -1)|0;
 $138 = (($137) + ($0))|0;
 $139 = $ido << 2;
 $140 = $ido << 1;
 if (!($2)) {
  return;
 }
 $141 = (($138) + ($1))|0;
 $k$25 = 0;$t1$24 = $138;$t2$23 = $141;$t4$22 = $ido;$t6$01 = $ido;
 while(1) {
  $142 = (($cc) + ($t1$24<<2)|0);
  $143 = +HEAPF32[$142>>2];
  $144 = (($cc) + ($t2$23<<2)|0);
  $145 = +HEAPF32[$144>>2];
  $146 = $143 + $145;
  $147 = $146 * -0.70710676908493042;
  $148 = $143 - $145;
  $149 = $148 * 0.70710676908493042;
  $150 = (($t6$01) + -1)|0;
  $151 = (($cc) + ($150<<2)|0);
  $152 = +HEAPF32[$151>>2];
  $153 = $152 + $149;
  $154 = (($t4$22) + -1)|0;
  $155 = (($ch) + ($154<<2)|0);
  HEAPF32[$155>>2] = $153;
  $156 = +HEAPF32[$151>>2];
  $157 = $156 - $149;
  $158 = (($t4$22) + ($140))|0;
  $159 = (($158) + -1)|0;
  $160 = (($ch) + ($159<<2)|0);
  HEAPF32[$160>>2] = $157;
  $161 = (($t1$24) + ($0))|0;
  $162 = (($cc) + ($161<<2)|0);
  $163 = +HEAPF32[$162>>2];
  $164 = $147 - $163;
  $165 = (($ch) + ($t4$22<<2)|0);
  HEAPF32[$165>>2] = $164;
  $166 = +HEAPF32[$162>>2];
  $167 = $147 + $166;
  $168 = (($ch) + ($158<<2)|0);
  HEAPF32[$168>>2] = $167;
  $169 = (($t1$24) + ($ido))|0;
  $170 = (($t2$23) + ($ido))|0;
  $171 = (($t4$22) + ($139))|0;
  $172 = (($t6$01) + ($ido))|0;
  $173 = (($k$25) + 1)|0;
  $exitcond = ($173|0)==($l1|0);
  if ($exitcond) {
   break;
  } else {
   $k$25 = $173;$t1$24 = $169;$t2$23 = $170;$t4$22 = $171;$t6$01 = $172;
  }
 }
 return;
}
function _dradf2($ido,$l1,$cc,$ch,$wa1) {
 $ido = $ido|0;
 $l1 = $l1|0;
 $cc = $cc|0;
 $ch = $ch|0;
 $wa1 = $wa1|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0;
 var $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0.0, $50 = 0, $51 = 0, $52 = 0, $53 = 0.0, $54 = 0.0, $55 = 0, $56 = 0, $57 = 0.0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0.0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0.0, $80 = 0;
 var $81 = 0, $9 = 0, $exitcond = 0, $exitcond23 = 0, $exitcond25 = 0, $i$05$us = 0, $k$016 = 0, $k$111$us = 0, $k$21 = 0, $t1$017 = 0, $t1$112$us = 0, $t1$22 = 0, $t2$018 = 0, $t2$113$us = 0, $t2$23 = 0, $t3$06$us = 0, $t3$14 = 0, $t4$07$us = 0, $t5$08$us = 0, $t6$09$us = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = Math_imul($l1, $ido)|0;
 $1 = $ido << 1;
 $2 = ($l1|0)>(0);
 if ($2) {
  $3 = (($1) + -1)|0;
  $k$016 = 0;$t1$017 = 0;$t2$018 = $0;
  while(1) {
   $4 = (($cc) + ($t1$017<<2)|0);
   $5 = +HEAPF32[$4>>2];
   $6 = (($cc) + ($t2$018<<2)|0);
   $7 = +HEAPF32[$6>>2];
   $8 = $5 + $7;
   $9 = $t1$017 << 1;
   $10 = (($ch) + ($9<<2)|0);
   HEAPF32[$10>>2] = $8;
   $11 = +HEAPF32[$4>>2];
   $12 = +HEAPF32[$6>>2];
   $13 = $11 - $12;
   $14 = (($3) + ($9))|0;
   $15 = (($ch) + ($14<<2)|0);
   HEAPF32[$15>>2] = $13;
   $16 = (($t1$017) + ($ido))|0;
   $17 = (($t2$018) + ($ido))|0;
   $18 = (($k$016) + 1)|0;
   $exitcond25 = ($18|0)==($l1|0);
   if ($exitcond25) {
    break;
   } else {
    $k$016 = $18;$t1$017 = $16;$t2$018 = $17;
   }
  }
 }
 $19 = ($ido|0)<(2);
 if ($19) {
  return;
 }
 $20 = ($ido|0)==(2);
 if (!($20)) {
  if ($2) {
   $k$111$us = 0;$t1$112$us = 0;$t2$113$us = $0;
   while(1) {
    $21 = $t1$112$us << 1;
    $22 = (($21) + ($1))|0;
    $i$05$us = 2;$t3$06$us = $t2$113$us;$t4$07$us = $22;$t5$08$us = $t1$112$us;$t6$09$us = $21;
    while(1) {
     $23 = (($t3$06$us) + 2)|0;
     $24 = (($t4$07$us) + -2)|0;
     $25 = (($t5$08$us) + 2)|0;
     $26 = (($t6$09$us) + 2)|0;
     $27 = (($i$05$us) + -2)|0;
     $28 = (($wa1) + ($27<<2)|0);
     $29 = +HEAPF32[$28>>2];
     $30 = (($t3$06$us) + 1)|0;
     $31 = (($cc) + ($30<<2)|0);
     $32 = +HEAPF32[$31>>2];
     $33 = $29 * $32;
     $34 = (($i$05$us) + -1)|0;
     $35 = (($wa1) + ($34<<2)|0);
     $36 = +HEAPF32[$35>>2];
     $37 = (($cc) + ($23<<2)|0);
     $38 = +HEAPF32[$37>>2];
     $39 = $36 * $38;
     $40 = $33 + $39;
     $41 = $29 * $38;
     $42 = $32 * $36;
     $43 = $41 - $42;
     $44 = (($cc) + ($25<<2)|0);
     $45 = +HEAPF32[$44>>2];
     $46 = $45 + $43;
     $47 = (($ch) + ($26<<2)|0);
     HEAPF32[$47>>2] = $46;
     $48 = +HEAPF32[$44>>2];
     $49 = $43 - $48;
     $50 = (($ch) + ($24<<2)|0);
     HEAPF32[$50>>2] = $49;
     $51 = (($t5$08$us) + 1)|0;
     $52 = (($cc) + ($51<<2)|0);
     $53 = +HEAPF32[$52>>2];
     $54 = $40 + $53;
     $55 = $t6$09$us | 1;
     $56 = (($ch) + ($55<<2)|0);
     HEAPF32[$56>>2] = $54;
     $57 = +HEAPF32[$52>>2];
     $58 = $57 - $40;
     $59 = (($t4$07$us) + -3)|0;
     $60 = (($ch) + ($59<<2)|0);
     HEAPF32[$60>>2] = $58;
     $61 = (($i$05$us) + 2)|0;
     $62 = ($61|0)<($ido|0);
     if ($62) {
      $i$05$us = $61;$t3$06$us = $23;$t4$07$us = $24;$t5$08$us = $25;$t6$09$us = $26;
     } else {
      break;
     }
    }
    $63 = (($t1$112$us) + ($ido))|0;
    $64 = (($t2$113$us) + ($ido))|0;
    $65 = (($k$111$us) + 1)|0;
    $exitcond23 = ($65|0)==($l1|0);
    if ($exitcond23) {
     break;
    } else {
     $k$111$us = $65;$t1$112$us = $63;$t2$113$us = $64;
    }
   }
  }
  $66 = (($ido|0) % 2)&-1;
  $67 = ($66|0)==(1);
  if ($67) {
   return;
  }
 }
 $68 = (($ido) + -1)|0;
 if (!($2)) {
  return;
 }
 $69 = (($0) + ($68))|0;
 $k$21 = 0;$t1$22 = $ido;$t2$23 = $69;$t3$14 = $68;
 while(1) {
  $70 = (($cc) + ($t2$23<<2)|0);
  $71 = +HEAPF32[$70>>2];
  $72 = -$71;
  $73 = (($ch) + ($t1$22<<2)|0);
  HEAPF32[$73>>2] = $72;
  $74 = (($cc) + ($t3$14<<2)|0);
  $75 = HEAP32[$74>>2]|0;
  $76 = (($t1$22) + -1)|0;
  $77 = (($ch) + ($76<<2)|0);
  HEAP32[$77>>2] = $75;
  $78 = (($t1$22) + ($1))|0;
  $79 = (($t2$23) + ($ido))|0;
  $80 = (($t3$14) + ($ido))|0;
  $81 = (($k$21) + 1)|0;
  $exitcond = ($81|0)==($l1|0);
  if ($exitcond) {
   break;
  } else {
   $k$21 = $81;$t1$22 = $78;$t2$23 = $79;$t3$14 = $80;
  }
 }
 return;
}
function _dradfg($ido,$ip,$l1,$idl1,$cc,$c1,$c2,$ch,$ch2,$wa) {
 $ido = $ido|0;
 $ip = $ip|0;
 $l1 = $l1|0;
 $idl1 = $idl1|0;
 $cc = $cc|0;
 $c1 = $c1|0;
 $c2 = $c2|0;
 $ch = $ch|0;
 $ch2 = $ch2|0;
 $wa = $wa|0;
 var $$not = 0, $$not268 = 0, $$not269 = 0, $$not270 = 0, $$not272 = 0, $$not273 = 0, $$not278 = 0, $$not279 = 0, $$not284 = 0, $$not285 = 0, $$pn = 0, $0 = 0.0, $1 = 0.0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0.0, $112 = 0, $113 = 0, $114 = 0.0, $115 = 0.0, $116 = 0, $117 = 0, $118 = 0.0, $119 = 0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0, $123 = 0.0;
 var $124 = 0.0, $125 = 0.0, $126 = 0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0.0;
 var $142 = 0, $143 = 0, $144 = 0.0, $145 = 0.0, $146 = 0, $147 = 0, $148 = 0.0, $149 = 0, $15 = 0, $150 = 0.0, $151 = 0.0, $152 = 0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0.0, $182 = 0, $183 = 0.0, $184 = 0.0, $185 = 0, $186 = 0.0, $187 = 0.0, $188 = 0.0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0.0;
 var $197 = 0, $198 = 0.0, $199 = 0.0, $2 = 0.0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0.0, $205 = 0.0, $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0.0, $21 = 0, $210 = 0, $211 = 0, $212 = 0.0, $213 = 0.0;
 var $214 = 0.0, $215 = 0.0, $216 = 0.0, $217 = 0.0, $218 = 0, $219 = 0, $22 = 0, $220 = 0.0, $221 = 0.0, $222 = 0, $223 = 0, $224 = 0.0, $225 = 0.0, $226 = 0, $227 = 0, $228 = 0.0, $229 = 0.0, $23 = 0, $230 = 0, $231 = 0;
 var $232 = 0.0, $233 = 0.0, $234 = 0, $235 = 0, $236 = 0, $237 = 0.0, $238 = 0, $239 = 0, $24 = 0, $240 = 0.0, $241 = 0.0, $242 = 0.0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0.0, $248 = 0.0, $249 = 0, $25 = 0;
 var $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0;
 var $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0;
 var $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0.0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0;
 var $304 = 0.0, $305 = 0, $306 = 0, $307 = 0, $308 = 0.0, $309 = 0.0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0.0, $314 = 0.0, $315 = 0.0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0.0, $321 = 0;
 var $322 = 0.0, $323 = 0.0, $324 = 0, $325 = 0.0, $326 = 0.0, $327 = 0.0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0;
 var $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0.0, $349 = 0, $35 = 0, $350 = 0, $351 = 0.0, $352 = 0.0, $353 = 0, $354 = 0, $355 = 0.0, $356 = 0.0, $357 = 0.0, $358 = 0;
 var $359 = 0, $36 = 0, $360 = 0, $361 = 0.0, $362 = 0, $363 = 0.0, $364 = 0.0, $365 = 0, $366 = 0.0, $367 = 0.0, $368 = 0.0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0;
 var $377 = 0, $38 = 0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0.0, $43 = 0, $44 = 0, $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0, $5 = 0.0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0;
 var $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0.0, $60 = 0.0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0.0, $75 = 0, $76 = 0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0;
 var $91 = 0.0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $ai1$0112 = 0.0, $ai2$0107$us = 0.0, $ar1$0113 = 0.0, $ar2$0108$us = 0.0, $brmerge = 0, $brmerge271 = 0, $brmerge274 = 0, $brmerge277$demorgan = 0, $brmerge280 = 0, $brmerge286 = 0, $exitcond = 0;
 var $exitcond221 = 0, $exitcond223 = 0, $exitcond224 = 0, $exitcond225 = 0, $exitcond226 = 0, $exitcond228 = 0, $exitcond229 = 0, $exitcond231 = 0, $exitcond232 = 0, $exitcond234 = 0, $exitcond235 = 0, $exitcond237 = 0, $exitcond239 = 0, $exitcond240 = 0, $exitcond241 = 0, $exitcond242 = 0, $exitcond243 = 0, $exitcond245 = 0, $exitcond246 = 0, $exitcond247 = 0;
 var $exitcond249 = 0, $exitcond250 = 0, $exitcond252 = 0, $exitcond253 = 0, $exitcond254 = 0, $exitcond255 = 0, $exitcond256 = 0, $exitcond257 = 0, $exitcond259 = 0, $i$0164$us = 0, $i$1184$us = 0, $i$2142$us = 0, $i$3152$us = 0, $i$475$us = 0, $i$572$us = 0, $i$638$us = 0, $i$728$us = 0, $idij$0165$us = 0, $idij$1185$us = 0, $ik$0202 = 0;
 var $ik$1131 = 0, $ik$294 = 0, $ik$3101$us = 0, $ik$482$us = 0, $is$0175 = 0, $is$1190 = 0, $j$0199$us = 0, $j$1036 = 0, $j$1176 = 0, $j$2191 = 0, $j$3148 = 0, $j$4161 = 0, $j$5127$us = 0, $j$6109$us = 0, $j$787$us = 0, $j$863$us = 0, $j$951 = 0, $k$0194$us = 0, $k$1027$us = 0, $k$1170$us = 0;
 var $k$2179$us = 0, $k$3136$us = 0, $k$4157$us = 0, $k$5121$us = 0, $k$680$us = 0, $k$769$us = 0, $k$856$us = 0, $k$945$us = 0, $l$0116 = 0, $or$cond = 0, $or$cond281 = 0, $or$cond282 = 0, $or$cond283 = 0, $t1$0198$us = 0, $t1$1062$us = 0, $t1$1150 = 0, $t1$1174 = 0, $t1$1235 = 0, $t1$2189 = 0, $t1$3147 = 0;
 var $t1$4160 = 0, $t1$5126$us = 0, $t1$6115 = 0, $t1$786$us = 0, $t1$879$us = 0, $t1$968$us = 0, $t2$0193$us = 0, $t2$1169$us = 0, $t2$2183$us = 0, $t2$3146 = 0, $t2$4159 = 0, $t2$5125$us = 0, $t2$6114 = 0, $t2$781$us = 0, $t2$878$us = 0, $t2$967$us = 0, $t3$0163$us = 0, $t3$1178$us = 0, $t3$2141$us = 0, $t3$3156$us = 0;
 var $t3$4120$us = 0, $t3$574$us = 0, $t3$661$us = 0, $t3$749 = 0, $t3$834 = 0, $t4$0140$us = 0, $t4$1155$us = 0, $t4$2119$us = 0, $t4$393 = 0, $t4$4106$us = 0, $t4$573$us = 0, $t4$660$us = 0, $t4$748 = 0, $t4$833 = 0, $t5$0134$us = 0, $t5$1150$us = 0, $t5$290 = 0, $t5$3105$us = 0, $t5$453$us = 0, $t5$547 = 0;
 var $t5$632 = 0, $t6$0135$us = 0, $t6$1151$us = 0, $t6$291 = 0, $t6$397$us = 0, $t6$454$us = 0, $t6$541$us = 0, $t6$623$us = 0, $t7$092 = 0, $t7$198$us = 0, $t7$255$us = 0, $t7$342$us = 0, $t7$424$us = 0, $t8$099$us = 0, $t8$143$us = 0, $t8$225$us = 0, $t9$0100$us = 0, $t9$144$us = 0, $t9$226$us = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = (+($ip|0));
 $1 = 6.2831854820251465 / $0;
 $2 = $1;
 $3 = (+Math_cos((+$2)));
 $4 = $3;
 $5 = (+Math_sin((+$2)));
 $6 = $5;
 $7 = (($ip) + 1)|0;
 $8 = $7 >> 1;
 $9 = (($ido) + -1)|0;
 $10 = $9 >> 1;
 $11 = Math_imul($l1, $ido)|0;
 $12 = Math_imul($ip, $ido)|0;
 $13 = ($ido|0)==(1);
 do {
  if (!($13)) {
   $14 = ($idl1|0)>(0);
   if ($14) {
    $ik$0202 = 0;
    while(1) {
     $25 = (($c2) + ($ik$0202<<2)|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = (($ch2) + ($ik$0202<<2)|0);
     HEAP32[$27>>2] = $26;
     $28 = (($ik$0202) + 1)|0;
     $exitcond259 = ($28|0)==($idl1|0);
     if ($exitcond259) {
      break;
     } else {
      $ik$0202 = $28;
     }
    }
   }
   $16 = ($ip|0)>(1);
   $17 = ($l1|0)>(0);
   $or$cond = $16 & $17;
   if ($or$cond) {
    $j$0199$us = 1;$t1$0198$us = 0;
    while(1) {
     $18 = (($t1$0198$us) + ($11))|0;
     $k$0194$us = 0;$t2$0193$us = $18;
     while(1) {
      $19 = (($c1) + ($t2$0193$us<<2)|0);
      $20 = HEAP32[$19>>2]|0;
      $21 = (($ch) + ($t2$0193$us<<2)|0);
      HEAP32[$21>>2] = $20;
      $22 = (($t2$0193$us) + ($ido))|0;
      $23 = (($k$0194$us) + 1)|0;
      $exitcond256 = ($23|0)==($l1|0);
      if ($exitcond256) {
       break;
      } else {
       $k$0194$us = $23;$t2$0193$us = $22;
      }
     }
     $24 = (($j$0199$us) + 1)|0;
     $exitcond257 = ($24|0)==($ip|0);
     if ($exitcond257) {
      break;
     } else {
      $j$0199$us = $24;$t1$0198$us = $18;
     }
    }
   }
   $29 = (0 - ($ido))|0;
   $30 = ($10|0)>($l1|0);
   if ($30) {
    if ($16) {
     $31 = ($l1|0)>(0);
     $32 = ($ido|0)>(2);
     $is$0175 = $29;$j$1176 = 1;$t1$1174 = 0;
     while(1) {
      $33 = (($t1$1174) + ($11))|0;
      $34 = (($is$0175) + ($ido))|0;
      if ($31) {
       $35 = (($34) + -1)|0;
       if ($32) {
        $36 = (($33) - ($ido))|0;
        $k$1170$us = 0;$t2$1169$us = $36;
        while(1) {
         $37 = (($t2$1169$us) + ($ido))|0;
         $i$0164$us = 2;$idij$0165$us = $35;$t3$0163$us = $37;
         while(1) {
          $38 = (($idij$0165$us) + 2)|0;
          $39 = (($t3$0163$us) + 2)|0;
          $40 = (($idij$0165$us) + 1)|0;
          $41 = (($wa) + ($40<<2)|0);
          $42 = +HEAPF32[$41>>2];
          $43 = (($t3$0163$us) + 1)|0;
          $44 = (($c1) + ($43<<2)|0);
          $45 = +HEAPF32[$44>>2];
          $46 = $42 * $45;
          $47 = (($wa) + ($38<<2)|0);
          $48 = +HEAPF32[$47>>2];
          $49 = (($c1) + ($39<<2)|0);
          $50 = +HEAPF32[$49>>2];
          $51 = $48 * $50;
          $52 = $46 + $51;
          $53 = (($ch) + ($43<<2)|0);
          HEAPF32[$53>>2] = $52;
          $54 = +HEAPF32[$41>>2];
          $55 = +HEAPF32[$49>>2];
          $56 = $54 * $55;
          $57 = +HEAPF32[$47>>2];
          $58 = +HEAPF32[$44>>2];
          $59 = $57 * $58;
          $60 = $56 - $59;
          $61 = (($ch) + ($39<<2)|0);
          HEAPF32[$61>>2] = $60;
          $62 = (($i$0164$us) + 2)|0;
          $63 = ($62|0)<($ido|0);
          if ($63) {
           $i$0164$us = $62;$idij$0165$us = $38;$t3$0163$us = $39;
          } else {
           break;
          }
         }
         $64 = (($k$1170$us) + 1)|0;
         $exitcond252 = ($64|0)==($l1|0);
         if ($exitcond252) {
          break;
         } else {
          $k$1170$us = $64;$t2$1169$us = $37;
         }
        }
       }
      }
      $65 = (($j$1176) + 1)|0;
      $exitcond253 = ($65|0)==($ip|0);
      if ($exitcond253) {
       break;
      } else {
       $is$0175 = $34;$j$1176 = $65;$t1$1174 = $33;
      }
     }
    }
   } else {
    if ($16) {
     $$not = ($ido|0)<(3);
     $$not268 = ($l1|0)<(1);
     $brmerge = $$not | $$not268;
     $is$1190 = $29;$j$2191 = 1;$t1$2189 = 0;
     while(1) {
      $66 = (($is$1190) + ($ido))|0;
      $67 = (($t1$2189) + ($11))|0;
      if (!($brmerge)) {
       $68 = (($66) + -1)|0;
       $i$1184$us = 2;$idij$1185$us = $68;$t2$2183$us = $67;
       while(1) {
        $69 = (($idij$1185$us) + 2)|0;
        $70 = (($t2$2183$us) + 2)|0;
        $71 = (($idij$1185$us) + 1)|0;
        $72 = (($wa) + ($71<<2)|0);
        $73 = (($wa) + ($69<<2)|0);
        $k$2179$us = 0;$t3$1178$us = $70;
        while(1) {
         $74 = +HEAPF32[$72>>2];
         $75 = (($t3$1178$us) + -1)|0;
         $76 = (($c1) + ($75<<2)|0);
         $77 = +HEAPF32[$76>>2];
         $78 = $74 * $77;
         $79 = +HEAPF32[$73>>2];
         $80 = (($c1) + ($t3$1178$us<<2)|0);
         $81 = +HEAPF32[$80>>2];
         $82 = $79 * $81;
         $83 = $78 + $82;
         $84 = (($ch) + ($75<<2)|0);
         HEAPF32[$84>>2] = $83;
         $85 = +HEAPF32[$72>>2];
         $86 = +HEAPF32[$80>>2];
         $87 = $85 * $86;
         $88 = +HEAPF32[$73>>2];
         $89 = +HEAPF32[$76>>2];
         $90 = $88 * $89;
         $91 = $87 - $90;
         $92 = (($ch) + ($t3$1178$us<<2)|0);
         HEAPF32[$92>>2] = $91;
         $93 = (($t3$1178$us) + ($ido))|0;
         $94 = (($k$2179$us) + 1)|0;
         $exitcond254 = ($94|0)==($l1|0);
         if ($exitcond254) {
          break;
         } else {
          $k$2179$us = $94;$t3$1178$us = $93;
         }
        }
        $95 = (($i$1184$us) + 2)|0;
        $96 = ($95|0)<($ido|0);
        if ($96) {
         $i$1184$us = $95;$idij$1185$us = $69;$t2$2183$us = $70;
        } else {
         break;
        }
       }
      }
      $97 = (($j$2191) + 1)|0;
      $exitcond255 = ($97|0)==($ip|0);
      if ($exitcond255) {
       break;
      } else {
       $is$1190 = $66;$j$2191 = $97;$t1$2189 = $67;
      }
     }
    }
   }
   $98 = Math_imul($11, $ip)|0;
   $99 = ($10|0)<($l1|0);
   $100 = ($8|0)>(1);
   if ($99) {
    if (!($100)) {
     break;
    }
    $$not269 = ($ido|0)<(3);
    $$not270 = ($l1|0)<(1);
    $brmerge271 = $$not269 | $$not270;
    $j$3148 = 1;$t1$3147 = 0;$t2$3146 = $98;
    while(1) {
     $101 = (($t1$3147) + ($11))|0;
     $102 = (($t2$3146) - ($11))|0;
     if (!($brmerge271)) {
      $i$2142$us = 2;$t3$2141$us = $101;$t4$0140$us = $102;
      while(1) {
       $103 = (($t3$2141$us) + 2)|0;
       $104 = (($t4$0140$us) + 2)|0;
       $105 = (($103) - ($ido))|0;
       $106 = (($104) - ($ido))|0;
       $k$3136$us = 0;$t5$0134$us = $105;$t6$0135$us = $106;
       while(1) {
        $107 = (($t5$0134$us) + ($ido))|0;
        $108 = (($t6$0135$us) + ($ido))|0;
        $109 = (($107) + -1)|0;
        $110 = (($ch) + ($109<<2)|0);
        $111 = +HEAPF32[$110>>2];
        $112 = (($108) + -1)|0;
        $113 = (($ch) + ($112<<2)|0);
        $114 = +HEAPF32[$113>>2];
        $115 = $111 + $114;
        $116 = (($c1) + ($109<<2)|0);
        HEAPF32[$116>>2] = $115;
        $117 = (($ch) + ($107<<2)|0);
        $118 = +HEAPF32[$117>>2];
        $119 = (($ch) + ($108<<2)|0);
        $120 = +HEAPF32[$119>>2];
        $121 = $118 - $120;
        $122 = (($c1) + ($112<<2)|0);
        HEAPF32[$122>>2] = $121;
        $123 = +HEAPF32[$117>>2];
        $124 = +HEAPF32[$119>>2];
        $125 = $123 + $124;
        $126 = (($c1) + ($107<<2)|0);
        HEAPF32[$126>>2] = $125;
        $127 = +HEAPF32[$113>>2];
        $128 = +HEAPF32[$110>>2];
        $129 = $127 - $128;
        $130 = (($c1) + ($108<<2)|0);
        HEAPF32[$130>>2] = $129;
        $131 = (($k$3136$us) + 1)|0;
        $exitcond246 = ($131|0)==($l1|0);
        if ($exitcond246) {
         break;
        } else {
         $k$3136$us = $131;$t5$0134$us = $107;$t6$0135$us = $108;
        }
       }
       $132 = (($i$2142$us) + 2)|0;
       $133 = ($132|0)<($ido|0);
       if ($133) {
        $i$2142$us = $132;$t3$2141$us = $103;$t4$0140$us = $104;
       } else {
        break;
       }
      }
     }
     $134 = (($j$3148) + 1)|0;
     $exitcond247 = ($134|0)==($8|0);
     if ($exitcond247) {
      break;
     } else {
      $j$3148 = $134;$t1$3147 = $101;$t2$3146 = $102;
     }
    }
   } else {
    if (!($100)) {
     break;
    }
    $$not272 = ($l1|0)<(1);
    $$not273 = ($ido|0)<(3);
    $brmerge274 = $$not272 | $$not273;
    $j$4161 = 1;$t1$4160 = 0;$t2$4159 = $98;
    while(1) {
     $135 = (($t1$4160) + ($11))|0;
     $136 = (($t2$4159) - ($11))|0;
     if (!($brmerge274)) {
      $k$4157$us = 0;$t3$3156$us = $135;$t4$1155$us = $136;
      while(1) {
       $i$3152$us = 2;$t5$1150$us = $t3$3156$us;$t6$1151$us = $t4$1155$us;
       while(1) {
        $137 = (($t5$1150$us) + 2)|0;
        $138 = (($t6$1151$us) + 2)|0;
        $139 = (($t5$1150$us) + 1)|0;
        $140 = (($ch) + ($139<<2)|0);
        $141 = +HEAPF32[$140>>2];
        $142 = (($t6$1151$us) + 1)|0;
        $143 = (($ch) + ($142<<2)|0);
        $144 = +HEAPF32[$143>>2];
        $145 = $141 + $144;
        $146 = (($c1) + ($139<<2)|0);
        HEAPF32[$146>>2] = $145;
        $147 = (($ch) + ($137<<2)|0);
        $148 = +HEAPF32[$147>>2];
        $149 = (($ch) + ($138<<2)|0);
        $150 = +HEAPF32[$149>>2];
        $151 = $148 - $150;
        $152 = (($c1) + ($142<<2)|0);
        HEAPF32[$152>>2] = $151;
        $153 = +HEAPF32[$147>>2];
        $154 = +HEAPF32[$149>>2];
        $155 = $153 + $154;
        $156 = (($c1) + ($137<<2)|0);
        HEAPF32[$156>>2] = $155;
        $157 = +HEAPF32[$143>>2];
        $158 = +HEAPF32[$140>>2];
        $159 = $157 - $158;
        $160 = (($c1) + ($138<<2)|0);
        HEAPF32[$160>>2] = $159;
        $161 = (($i$3152$us) + 2)|0;
        $162 = ($161|0)<($ido|0);
        if ($162) {
         $i$3152$us = $161;$t5$1150$us = $137;$t6$1151$us = $138;
        } else {
         break;
        }
       }
       $163 = (($t3$3156$us) + ($ido))|0;
       $164 = (($t4$1155$us) + ($ido))|0;
       $165 = (($k$4157$us) + 1)|0;
       $exitcond249 = ($165|0)==($l1|0);
       if ($exitcond249) {
        break;
       } else {
        $k$4157$us = $165;$t3$3156$us = $163;$t4$1155$us = $164;
       }
      }
     }
     $166 = (($j$4161) + 1)|0;
     $exitcond250 = ($166|0)==($8|0);
     if ($exitcond250) {
      break;
     } else {
      $j$4161 = $166;$t1$4160 = $135;$t2$4159 = $136;
     }
    }
   }
  }
 } while(0);
 $15 = ($idl1|0)>(0);
 if ($15) {
  $ik$1131 = 0;
  while(1) {
   $167 = (($ch2) + ($ik$1131<<2)|0);
   $168 = HEAP32[$167>>2]|0;
   $169 = (($c2) + ($ik$1131<<2)|0);
   HEAP32[$169>>2] = $168;
   $170 = (($ik$1131) + 1)|0;
   $exitcond245 = ($170|0)==($idl1|0);
   if ($exitcond245) {
    break;
   } else {
    $ik$1131 = $170;
   }
  }
 }
 $171 = Math_imul($idl1, $ip)|0;
 $172 = ($8|0)>(1);
 do {
  if ($172) {
   $173 = ($l1|0)>(0);
   if ($173) {
    $j$5127$us = 1;$t1$5126$us = 0;$t2$5125$us = $171;
    while(1) {
     $174 = (($t1$5126$us) + ($11))|0;
     $175 = (($t2$5125$us) - ($11))|0;
     $176 = (($174) - ($ido))|0;
     $177 = (($175) - ($ido))|0;
     $k$5121$us = 0;$t3$4120$us = $176;$t4$2119$us = $177;
     while(1) {
      $178 = (($t3$4120$us) + ($ido))|0;
      $179 = (($t4$2119$us) + ($ido))|0;
      $180 = (($ch) + ($178<<2)|0);
      $181 = +HEAPF32[$180>>2];
      $182 = (($ch) + ($179<<2)|0);
      $183 = +HEAPF32[$182>>2];
      $184 = $181 + $183;
      $185 = (($c1) + ($178<<2)|0);
      HEAPF32[$185>>2] = $184;
      $186 = +HEAPF32[$182>>2];
      $187 = +HEAPF32[$180>>2];
      $188 = $186 - $187;
      $189 = (($c1) + ($179<<2)|0);
      HEAPF32[$189>>2] = $188;
      $190 = (($k$5121$us) + 1)|0;
      $exitcond242 = ($190|0)==($l1|0);
      if ($exitcond242) {
       break;
      } else {
       $k$5121$us = $190;$t3$4120$us = $178;$t4$2119$us = $179;
      }
     }
     $191 = (($j$5127$us) + 1)|0;
     $exitcond243 = ($191|0)==($8|0);
     if ($exitcond243) {
      break;
     } else {
      $j$5127$us = $191;$t1$5126$us = $174;$t2$5125$us = $175;
     }
    }
    if (!($172)) {
     break;
    }
   }
   $$pn = (($ip) + -1)|0;
   $192 = Math_imul($$pn, $idl1)|0;
   $$not278 = ($8|0)<(3);
   $$not279 = $15 ^ 1;
   $brmerge280 = $$not278 | $$not279;
   $ai1$0112 = 0.0;$ar1$0113 = 1.0;$l$0116 = 1;$t1$6115 = 0;$t2$6114 = $171;
   while(1) {
    $202 = (($t1$6115) + ($idl1))|0;
    $203 = (($t2$6114) - ($idl1))|0;
    $204 = $4 * $ar1$0113;
    $205 = $6 * $ai1$0112;
    $206 = $204 - $205;
    $207 = $4 * $ai1$0112;
    $208 = $6 * $ar1$0113;
    $209 = $207 + $208;
    if ($15) {
     $ik$294 = 0;$t4$393 = $202;$t5$290 = $203;$t6$291 = $192;$t7$092 = $idl1;
     while(1) {
      $236 = (($c2) + ($ik$294<<2)|0);
      $237 = +HEAPF32[$236>>2];
      $238 = (($t7$092) + 1)|0;
      $239 = (($c2) + ($t7$092<<2)|0);
      $240 = +HEAPF32[$239>>2];
      $241 = $206 * $240;
      $242 = $237 + $241;
      $243 = (($t4$393) + 1)|0;
      $244 = (($ch2) + ($t4$393<<2)|0);
      HEAPF32[$244>>2] = $242;
      $245 = (($t6$291) + 1)|0;
      $246 = (($c2) + ($t6$291<<2)|0);
      $247 = +HEAPF32[$246>>2];
      $248 = $209 * $247;
      $249 = (($t5$290) + 1)|0;
      $250 = (($ch2) + ($t5$290<<2)|0);
      HEAPF32[$250>>2] = $248;
      $251 = (($ik$294) + 1)|0;
      $exitcond237 = ($251|0)==($idl1|0);
      if ($exitcond237) {
       break;
      } else {
       $ik$294 = $251;$t4$393 = $243;$t5$290 = $249;$t6$291 = $245;$t7$092 = $238;
      }
     }
    }
    if (!($brmerge280)) {
     $ai2$0107$us = $209;$ar2$0108$us = $206;$j$6109$us = 2;$t4$4106$us = $idl1;$t5$3105$us = $192;
     while(1) {
      $210 = (($t4$4106$us) + ($idl1))|0;
      $211 = (($t5$3105$us) - ($idl1))|0;
      $212 = $206 * $ar2$0108$us;
      $213 = $209 * $ai2$0107$us;
      $214 = $212 - $213;
      $215 = $206 * $ai2$0107$us;
      $216 = $209 * $ar2$0108$us;
      $217 = $215 + $216;
      $ik$3101$us = 0;$t6$397$us = $202;$t7$198$us = $203;$t8$099$us = $210;$t9$0100$us = $211;
      while(1) {
       $218 = (($t8$099$us) + 1)|0;
       $219 = (($c2) + ($t8$099$us<<2)|0);
       $220 = +HEAPF32[$219>>2];
       $221 = $214 * $220;
       $222 = (($t6$397$us) + 1)|0;
       $223 = (($ch2) + ($t6$397$us<<2)|0);
       $224 = +HEAPF32[$223>>2];
       $225 = $224 + $221;
       HEAPF32[$223>>2] = $225;
       $226 = (($t9$0100$us) + 1)|0;
       $227 = (($c2) + ($t9$0100$us<<2)|0);
       $228 = +HEAPF32[$227>>2];
       $229 = $217 * $228;
       $230 = (($t7$198$us) + 1)|0;
       $231 = (($ch2) + ($t7$198$us<<2)|0);
       $232 = +HEAPF32[$231>>2];
       $233 = $232 + $229;
       HEAPF32[$231>>2] = $233;
       $234 = (($ik$3101$us) + 1)|0;
       $exitcond239 = ($234|0)==($idl1|0);
       if ($exitcond239) {
        break;
       } else {
        $ik$3101$us = $234;$t6$397$us = $222;$t7$198$us = $230;$t8$099$us = $218;$t9$0100$us = $226;
       }
      }
      $235 = (($j$6109$us) + 1)|0;
      $exitcond240 = ($235|0)==($8|0);
      if ($exitcond240) {
       break;
      } else {
       $ai2$0107$us = $217;$ar2$0108$us = $214;$j$6109$us = $235;$t4$4106$us = $210;$t5$3105$us = $211;
      }
     }
    }
    $252 = (($l$0116) + 1)|0;
    $exitcond241 = ($252|0)==($8|0);
    if ($exitcond241) {
     break;
    } else {
     $ai1$0112 = $209;$ar1$0113 = $206;$l$0116 = $252;$t1$6115 = $202;$t2$6114 = $203;
    }
   }
   $brmerge277$demorgan = $172 & $15;
   if ($brmerge277$demorgan) {
    $j$787$us = 1;$t1$786$us = 0;
    while(1) {
     $193 = (($t1$786$us) + ($idl1))|0;
     $ik$482$us = 0;$t2$781$us = $193;
     while(1) {
      $194 = (($t2$781$us) + 1)|0;
      $195 = (($c2) + ($t2$781$us<<2)|0);
      $196 = +HEAPF32[$195>>2];
      $197 = (($ch2) + ($ik$482$us<<2)|0);
      $198 = +HEAPF32[$197>>2];
      $199 = $196 + $198;
      HEAPF32[$197>>2] = $199;
      $200 = (($ik$482$us) + 1)|0;
      $exitcond234 = ($200|0)==($idl1|0);
      if ($exitcond234) {
       break;
      } else {
       $ik$482$us = $200;$t2$781$us = $194;
      }
     }
     $201 = (($j$787$us) + 1)|0;
     $exitcond235 = ($201|0)==($8|0);
     if ($exitcond235) {
      break;
     } else {
      $j$787$us = $201;$t1$786$us = $193;
     }
    }
   }
  }
 } while(0);
 $253 = ($ido|0)<($l1|0);
 if ($253) {
  $265 = ($ido|0)>(0);
  $266 = ($l1|0)>(0);
  $or$cond282 = $265 & $266;
  if ($or$cond282) {
   $i$572$us = 0;
   while(1) {
    $k$769$us = 0;$t1$968$us = $i$572$us;$t2$967$us = $i$572$us;
    while(1) {
     $267 = (($ch) + ($t1$968$us<<2)|0);
     $268 = HEAP32[$267>>2]|0;
     $269 = (($cc) + ($t2$967$us<<2)|0);
     HEAP32[$269>>2] = $268;
     $270 = (($t1$968$us) + ($ido))|0;
     $271 = (($t2$967$us) + ($12))|0;
     $272 = (($k$769$us) + 1)|0;
     $exitcond228 = ($272|0)==($l1|0);
     if ($exitcond228) {
      break;
     } else {
      $k$769$us = $272;$t1$968$us = $270;$t2$967$us = $271;
     }
    }
    $273 = (($i$572$us) + 1)|0;
    $exitcond229 = ($273|0)==($ido|0);
    if ($exitcond229) {
     break;
    } else {
     $i$572$us = $273;
    }
   }
  }
 } else {
  $254 = ($l1|0)>(0);
  $255 = ($ido|0)>(0);
  $or$cond281 = $254 & $255;
  if ($or$cond281) {
   $k$680$us = 0;$t1$879$us = 0;$t2$878$us = 0;
   while(1) {
    $i$475$us = 0;$t3$574$us = $t1$879$us;$t4$573$us = $t2$878$us;
    while(1) {
     $256 = (($t3$574$us) + 1)|0;
     $257 = (($ch) + ($t3$574$us<<2)|0);
     $258 = HEAP32[$257>>2]|0;
     $259 = (($t4$573$us) + 1)|0;
     $260 = (($cc) + ($t4$573$us<<2)|0);
     HEAP32[$260>>2] = $258;
     $261 = (($i$475$us) + 1)|0;
     $exitcond231 = ($261|0)==($ido|0);
     if ($exitcond231) {
      break;
     } else {
      $i$475$us = $261;$t3$574$us = $256;$t4$573$us = $259;
     }
    }
    $262 = (($t1$879$us) + ($ido))|0;
    $263 = (($t2$878$us) + ($12))|0;
    $264 = (($k$680$us) + 1)|0;
    $exitcond232 = ($264|0)==($l1|0);
    if ($exitcond232) {
     break;
    } else {
     $k$680$us = $264;$t1$879$us = $262;$t2$878$us = $263;
    }
   }
  }
 }
 $274 = $ido << 1;
 $275 = Math_imul($11, $ip)|0;
 $276 = ($l1|0)>(0);
 $or$cond283 = $172 & $276;
 if ($or$cond283) {
  $j$863$us = 1;$t1$1062$us = 0;$t3$661$us = 0;$t4$660$us = $275;
  while(1) {
   $277 = (($t1$1062$us) + ($274))|0;
   $278 = (($t3$661$us) + ($11))|0;
   $279 = (($t4$660$us) - ($11))|0;
   $k$856$us = 0;$t5$453$us = $277;$t6$454$us = $278;$t7$255$us = $279;
   while(1) {
    $280 = (($ch) + ($t6$454$us<<2)|0);
    $281 = HEAP32[$280>>2]|0;
    $282 = (($t5$453$us) + -1)|0;
    $283 = (($cc) + ($282<<2)|0);
    HEAP32[$283>>2] = $281;
    $284 = (($ch) + ($t7$255$us<<2)|0);
    $285 = HEAP32[$284>>2]|0;
    $286 = (($cc) + ($t5$453$us<<2)|0);
    HEAP32[$286>>2] = $285;
    $287 = (($t5$453$us) + ($12))|0;
    $288 = (($t6$454$us) + ($ido))|0;
    $289 = (($t7$255$us) + ($ido))|0;
    $290 = (($k$856$us) + 1)|0;
    $exitcond225 = ($290|0)==($l1|0);
    if ($exitcond225) {
     break;
    } else {
     $k$856$us = $290;$t5$453$us = $287;$t6$454$us = $288;$t7$255$us = $289;
    }
   }
   $291 = (($j$863$us) + 1)|0;
   $exitcond226 = ($291|0)==($8|0);
   if ($exitcond226) {
    break;
   } else {
    $j$863$us = $291;$t1$1062$us = $277;$t3$661$us = $278;$t4$660$us = $279;
   }
  }
 }
 if ($13) {
  return;
 }
 $292 = ($10|0)<($l1|0);
 $293 = (0 - ($ido))|0;
 if (!($292)) {
  if (!($172)) {
   return;
  }
  $$not284 = ($l1|0)<(1);
  $$not285 = ($ido|0)<(3);
  $brmerge286 = $$not284 | $$not285;
  $j$951 = 1;$t1$1150 = $293;$t3$749 = 0;$t4$748 = 0;$t5$547 = $275;
  while(1) {
   $296 = (($t1$1150) + ($274))|0;
   $297 = (($t3$749) + ($274))|0;
   $298 = (($t4$748) + ($11))|0;
   $299 = (($t5$547) - ($11))|0;
   if (!($brmerge286)) {
    $k$945$us = 0;$t6$541$us = $296;$t7$342$us = $297;$t8$143$us = $298;$t9$144$us = $299;
    while(1) {
     $i$638$us = 2;
     while(1) {
      $300 = (($ido) - ($i$638$us))|0;
      $301 = (($i$638$us) + ($t8$143$us))|0;
      $302 = (($301) + -1)|0;
      $303 = (($ch) + ($302<<2)|0);
      $304 = +HEAPF32[$303>>2];
      $305 = (($i$638$us) + ($t9$144$us))|0;
      $306 = (($305) + -1)|0;
      $307 = (($ch) + ($306<<2)|0);
      $308 = +HEAPF32[$307>>2];
      $309 = $304 + $308;
      $310 = (($i$638$us) + ($t7$342$us))|0;
      $311 = (($310) + -1)|0;
      $312 = (($cc) + ($311<<2)|0);
      HEAPF32[$312>>2] = $309;
      $313 = +HEAPF32[$303>>2];
      $314 = +HEAPF32[$307>>2];
      $315 = $313 - $314;
      $316 = (($300) + ($t6$541$us))|0;
      $317 = (($316) + -1)|0;
      $318 = (($cc) + ($317<<2)|0);
      HEAPF32[$318>>2] = $315;
      $319 = (($ch) + ($301<<2)|0);
      $320 = +HEAPF32[$319>>2];
      $321 = (($ch) + ($305<<2)|0);
      $322 = +HEAPF32[$321>>2];
      $323 = $320 + $322;
      $324 = (($cc) + ($310<<2)|0);
      HEAPF32[$324>>2] = $323;
      $325 = +HEAPF32[$321>>2];
      $326 = +HEAPF32[$319>>2];
      $327 = $325 - $326;
      $328 = (($cc) + ($316<<2)|0);
      HEAPF32[$328>>2] = $327;
      $329 = (($i$638$us) + 2)|0;
      $330 = ($329|0)<($ido|0);
      if ($330) {
       $i$638$us = $329;
      } else {
       break;
      }
     }
     $331 = (($t6$541$us) + ($12))|0;
     $332 = (($t7$342$us) + ($12))|0;
     $333 = (($t8$143$us) + ($ido))|0;
     $334 = (($t9$144$us) + ($ido))|0;
     $335 = (($k$945$us) + 1)|0;
     $exitcond223 = ($335|0)==($l1|0);
     if ($exitcond223) {
      break;
     } else {
      $k$945$us = $335;$t6$541$us = $331;$t7$342$us = $332;$t8$143$us = $333;$t9$144$us = $334;
     }
    }
   }
   $336 = (($j$951) + 1)|0;
   $exitcond224 = ($336|0)==($8|0);
   if ($exitcond224) {
    break;
   } else {
    $j$951 = $336;$t1$1150 = $296;$t3$749 = $297;$t4$748 = $298;$t5$547 = $299;
   }
  }
  return;
 }
 if (!($172)) {
  return;
 }
 $294 = ($ido|0)>(2);
 $295 = ($l1|0)>(0);
 $j$1036 = 1;$t1$1235 = $293;$t3$834 = 0;$t4$833 = 0;$t5$632 = $275;
 while(1) {
  $337 = (($t1$1235) + ($274))|0;
  $338 = (($t3$834) + ($274))|0;
  $339 = (($t4$833) + ($11))|0;
  $340 = (($t5$632) - ($11))|0;
  if ($294) {
   $341 = (($337) + ($ido))|0;
   if ($295) {
    $i$728$us = 2;
    while(1) {
     $342 = (($341) - ($i$728$us))|0;
     $343 = (($i$728$us) + ($338))|0;
     $344 = (($i$728$us) + ($339))|0;
     $345 = (($i$728$us) + ($340))|0;
     $k$1027$us = 0;$t6$623$us = $342;$t7$424$us = $343;$t8$225$us = $344;$t9$226$us = $345;
     while(1) {
      $346 = (($t8$225$us) + -1)|0;
      $347 = (($ch) + ($346<<2)|0);
      $348 = +HEAPF32[$347>>2];
      $349 = (($t9$226$us) + -1)|0;
      $350 = (($ch) + ($349<<2)|0);
      $351 = +HEAPF32[$350>>2];
      $352 = $348 + $351;
      $353 = (($t7$424$us) + -1)|0;
      $354 = (($cc) + ($353<<2)|0);
      HEAPF32[$354>>2] = $352;
      $355 = +HEAPF32[$347>>2];
      $356 = +HEAPF32[$350>>2];
      $357 = $355 - $356;
      $358 = (($t6$623$us) + -1)|0;
      $359 = (($cc) + ($358<<2)|0);
      HEAPF32[$359>>2] = $357;
      $360 = (($ch) + ($t8$225$us<<2)|0);
      $361 = +HEAPF32[$360>>2];
      $362 = (($ch) + ($t9$226$us<<2)|0);
      $363 = +HEAPF32[$362>>2];
      $364 = $361 + $363;
      $365 = (($cc) + ($t7$424$us<<2)|0);
      HEAPF32[$365>>2] = $364;
      $366 = +HEAPF32[$362>>2];
      $367 = +HEAPF32[$360>>2];
      $368 = $366 - $367;
      $369 = (($cc) + ($t6$623$us<<2)|0);
      HEAPF32[$369>>2] = $368;
      $370 = (($t6$623$us) + ($12))|0;
      $371 = (($t7$424$us) + ($12))|0;
      $372 = (($t8$225$us) + ($ido))|0;
      $373 = (($t9$226$us) + ($ido))|0;
      $374 = (($k$1027$us) + 1)|0;
      $exitcond = ($374|0)==($l1|0);
      if ($exitcond) {
       break;
      } else {
       $k$1027$us = $374;$t6$623$us = $370;$t7$424$us = $371;$t8$225$us = $372;$t9$226$us = $373;
      }
     }
     $375 = (($i$728$us) + 2)|0;
     $376 = ($375|0)<($ido|0);
     if ($376) {
      $i$728$us = $375;
     } else {
      break;
     }
    }
   }
  }
  $377 = (($j$1036) + 1)|0;
  $exitcond221 = ($377|0)==($8|0);
  if ($exitcond221) {
   break;
  } else {
   $j$1036 = $377;$t1$1235 = $337;$t3$834 = $338;$t4$833 = $339;$t5$632 = $340;
  }
 }
 return;
}
function __vp_noisemask($p,$logmdct,$logmask) {
 $p = $p|0;
 $logmdct = $logmdct|0;
 $logmask = $logmask|0;
 var $$ = 0, $$$ = 0, $$alloca_mul = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0.0, $23 = 0.0;
 var $24 = 0, $25 = 0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0.0, $8 = 0;
 var $9 = 0.0, $exitcond = 0, $exitcond10 = 0, $exitcond9 = 0, $i$05 = 0, $i$12 = 0, $i$21 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$p>>2]|0;
 $1 = $0 << 2;
 $$alloca_mul = $1;
 $2 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $3 = ((($p)) + 24|0);
 $4 = HEAP32[$3>>2]|0;
 _bark_noise_hybridmp($0,$4,$logmdct,$logmask,140.0,-1);
 $5 = ($0|0)>(0);
 if ($5) {
  $i$05 = 0;
  while(1) {
   $6 = (($logmdct) + ($i$05<<2)|0);
   $7 = +HEAPF32[$6>>2];
   $8 = (($logmask) + ($i$05<<2)|0);
   $9 = +HEAPF32[$8>>2];
   $10 = $7 - $9;
   $11 = (($2) + ($i$05<<2)|0);
   HEAPF32[$11>>2] = $10;
   $12 = (($i$05) + 1)|0;
   $exitcond10 = ($12|0)==($0|0);
   if ($exitcond10) {
    break;
   } else {
    $i$05 = $12;
   }
  }
 }
 $13 = HEAP32[$3>>2]|0;
 $14 = ((($p)) + 4|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($15)) + 128|0);
 $17 = HEAP32[$16>>2]|0;
 _bark_noise_hybridmp($0,$13,$2,$logmask,0.0,$17);
 if ($5) {
  $i$12 = 0;
 } else {
  STACKTOP = sp;return;
 }
 while(1) {
  $19 = (($logmdct) + ($i$12<<2)|0);
  $20 = +HEAPF32[$19>>2];
  $21 = (($2) + ($i$12<<2)|0);
  $22 = +HEAPF32[$21>>2];
  $23 = $20 - $22;
  HEAPF32[$21>>2] = $23;
  $24 = (($i$12) + 1)|0;
  $exitcond9 = ($24|0)==($0|0);
  if ($exitcond9) {
   break;
  } else {
   $i$12 = $24;
  }
 }
 if (!($5)) {
  STACKTOP = sp;return;
 }
 $18 = HEAP32[$14>>2]|0;
 $i$21 = 0;
 while(1) {
  $25 = (($logmask) + ($i$21<<2)|0);
  $26 = +HEAPF32[$25>>2];
  $27 = $26;
  $28 = $27 + 0.5;
  $29 = (~~(($28)));
  $30 = ($29|0)>(39);
  $$ = $30 ? 39 : $29;
  $31 = ($$|0)<(0);
  $$$ = $31 ? 0 : $$;
  $32 = (($2) + ($i$21<<2)|0);
  $33 = +HEAPF32[$32>>2];
  $34 = (((($18)) + 336|0) + ($$$<<2)|0);
  $35 = +HEAPF32[$34>>2];
  $36 = $33 + $35;
  HEAPF32[$25>>2] = $36;
  $37 = (($i$21) + 1)|0;
  $exitcond = ($37|0)==($0|0);
  if ($exitcond) {
   break;
  } else {
   $i$21 = $37;
  }
 }
 STACKTOP = sp;return;
}
function _bark_noise_hybridmp($n,$b,$f,$noise,$offset,$fixed) {
 $n = $n|0;
 $b = $b|0;
 $f = $f|0;
 $noise = $noise|0;
 $offset = +$offset;
 $fixed = $fixed|0;
 var $$alloca_mul = 0, $$alloca_mul1 = 0, $$alloca_mul3 = 0, $$alloca_mul5 = 0, $$alloca_mul7 = 0, $$in = 0, $$pn = 0, $0 = 0, $1 = 0, $10 = 0.0, $100 = 0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0.0, $105 = 0, $106 = 0.0, $107 = 0, $108 = 0.0, $109 = 0.0;
 var $11 = 0.0, $110 = 0, $111 = 0.0, $112 = 0, $113 = 0.0, $114 = 0.0, $115 = 0, $116 = 0.0, $117 = 0, $118 = 0.0, $119 = 0.0, $12 = 0.0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0.0;
 var $128 = 0.0, $129 = 0.0, $13 = 0.0, $130 = 0.0, $131 = 0.0, $132 = 0, $133 = 0.0, $134 = 0, $135 = 0, $136 = 0.0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0, $145 = 0.0;
 var $146 = 0, $147 = 0, $148 = 0.0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0.0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0.0;
 var $164 = 0.0, $165 = 0, $166 = 0.0, $167 = 0, $168 = 0.0, $169 = 0.0, $17 = 0, $170 = 0, $171 = 0.0, $172 = 0, $173 = 0.0, $174 = 0.0, $175 = 0, $176 = 0.0, $177 = 0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0, $181 = 0.0;
 var $182 = 0, $183 = 0.0, $184 = 0.0, $185 = 0.0, $186 = 0.0, $187 = 0.0, $188 = 0.0, $189 = 0.0, $19 = 0.0, $190 = 0.0, $191 = 0.0, $192 = 0.0, $193 = 0.0, $194 = 0.0, $195 = 0.0, $196 = 0.0, $197 = 0.0, $198 = 0, $199 = 0.0, $2 = 0;
 var $20 = 0.0, $200 = 0, $201 = 0, $202 = 0.0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0.0, $209 = 0, $21 = 0, $210 = 0.0, $211 = 0.0, $212 = 0, $213 = 0.0, $214 = 0, $215 = 0.0, $216 = 0.0, $217 = 0;
 var $218 = 0.0, $219 = 0, $22 = 0.0, $220 = 0.0, $221 = 0.0, $222 = 0, $223 = 0.0, $224 = 0, $225 = 0.0, $226 = 0.0, $227 = 0, $228 = 0.0, $229 = 0, $23 = 0.0, $230 = 0.0, $231 = 0.0, $232 = 0.0, $233 = 0.0, $234 = 0.0, $235 = 0.0;
 var $236 = 0.0, $237 = 0.0, $238 = 0.0, $239 = 0.0, $24 = 0.0, $240 = 0.0, $241 = 0.0, $242 = 0.0, $243 = 0.0, $244 = 0.0, $245 = 0, $246 = 0.0, $247 = 0, $248 = 0, $249 = 0.0, $25 = 0.0, $250 = 0, $251 = 0.0, $252 = 0.0, $253 = 0.0;
 var $254 = 0.0, $255 = 0, $256 = 0.0, $257 = 0, $258 = 0, $259 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0.0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0.0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0.0, $56 = 0;
 var $57 = 0.0, $58 = 0, $59 = 0.0, $6 = 0.0, $60 = 0.0, $61 = 0, $62 = 0.0, $63 = 0, $64 = 0.0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0, $69 = 0.0, $7 = 0.0, $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0.0;
 var $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0, $86 = 0, $87 = 0.0, $88 = 0, $89 = 0, $9 = 0.0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0.0, $97 = 0, $98 = 0.0, $99 = 0.0, $A$0$lcssa = 0.0, $A$1$lcssa = 0.0, $A$2$lcssa = 0.0, $A$3$lcssa = 0.0, $B$0$lcssa = 0.0, $B$1$lcssa = 0.0, $B$2$lcssa = 0.0, $B$3$lcssa = 0.0, $D$0$lcssa = 0.0, $D$1$lcssa = 0.0, $D$2$lcssa = 0.0, $D$3$lcssa = 0.0, $R$0 = 0.0;
 var $R$1 = 0.0, $R$2 = 0.0, $exitcond = 0, $exitcond71 = 0, $exitcond72 = 0, $exitcond73 = 0, $exitcond74 = 0, $i$048 = 0, $i$1$lcssa = 0, $i$135 = 0, $i$2$lcssa = 0, $i$227 = 0, $i$324 = 0, $i$4$lcssa = 0, $i$415 = 0, $i$5$lcssa = 0, $i$58 = 0, $i$66 = 0, $tN$043 = 0.0, $tX$044 = 0.0;
 var $tXX$045 = 0.0, $tXY$047 = 0.0, $tY$046 = 0.0, $x$049 = 0.0, $x$1$lcssa = 0.0, $x$136 = 0.0, $x$2$lcssa = 0.0, $x$228 = 0.0, $x$325 = 0.0, $x$4$lcssa = 0.0, $x$416 = 0.0, $x$5$lcssa = 0.0, $x$59 = 0.0, $x$67 = 0.0, $y$0 = 0.0, $y$1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $n << 2;
 $$alloca_mul = $0;
 $1 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $$alloca_mul1 = $0;
 $2 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul1)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $$alloca_mul3 = $0;
 $3 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul3)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $$alloca_mul5 = $0;
 $4 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul5)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $$alloca_mul7 = $0;
 $5 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul7)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $6 = +HEAPF32[$f>>2];
 $7 = $6 + $offset;
 $8 = $7 < 1.0;
 $y$0 = $8 ? 1.0 : $7;
 $9 = $y$0 * $y$0;
 $10 = $9 * 0.5;
 $11 = $10 + 0.0;
 $12 = $y$0 * $10;
 $13 = $12 + 0.0;
 HEAPF32[$1>>2] = $11;
 HEAPF32[$2>>2] = $11;
 HEAPF32[$3>>2] = 0.0;
 HEAPF32[$4>>2] = $13;
 HEAPF32[$5>>2] = 0.0;
 $14 = ($n|0)>(1);
 if ($14) {
  $i$048 = 1;$tN$043 = $11;$tX$044 = $11;$tXX$045 = 0.0;$tXY$047 = 0.0;$tY$046 = $13;$x$049 = 1.0;
  while(1) {
   $18 = (($f) + ($i$048<<2)|0);
   $19 = +HEAPF32[$18>>2];
   $20 = $19 + $offset;
   $21 = $20 < 1.0;
   $y$1 = $21 ? 1.0 : $20;
   $22 = $y$1 * $y$1;
   $23 = $tN$043 + $22;
   $24 = $x$049 * $22;
   $25 = $tX$044 + $24;
   $26 = $x$049 * $24;
   $27 = $tXX$045 + $26;
   $28 = $y$1 * $22;
   $29 = $tY$046 + $28;
   $30 = $y$1 * $24;
   $31 = $tXY$047 + $30;
   $32 = (($1) + ($i$048<<2)|0);
   HEAPF32[$32>>2] = $23;
   $33 = (($2) + ($i$048<<2)|0);
   HEAPF32[$33>>2] = $25;
   $34 = (($3) + ($i$048<<2)|0);
   HEAPF32[$34>>2] = $27;
   $35 = (($4) + ($i$048<<2)|0);
   HEAPF32[$35>>2] = $29;
   $36 = (($5) + ($i$048<<2)|0);
   HEAPF32[$36>>2] = $31;
   $37 = (($i$048) + 1)|0;
   $38 = $x$049 + 1.0;
   $exitcond74 = ($37|0)==($n|0);
   if ($exitcond74) {
    break;
   } else {
    $i$048 = $37;$tN$043 = $23;$tX$044 = $25;$tXX$045 = $27;$tXY$047 = $31;$tY$046 = $29;$x$049 = $38;
   }
  }
 }
 $15 = HEAP32[$b>>2]|0;
 $16 = $15 >> 16;
 $17 = ($16|0)>(-1);
 if ($17) {
  $40 = $15;$A$0$lcssa = 0.0;$B$0$lcssa = 0.0;$D$0$lcssa = 1.0;$i$1$lcssa = 0;$x$1$lcssa = 0.0;
 } else {
  $43 = $15;$47 = $16;$i$135 = 0;$x$136 = 0.0;
  while(1) {
   $42 = $43 & 65535;
   $44 = (($1) + ($42<<2)|0);
   $45 = +HEAPF32[$44>>2];
   $46 = (0 - ($47))|0;
   $48 = (($1) + ($46<<2)|0);
   $49 = +HEAPF32[$48>>2];
   $50 = $45 + $49;
   $51 = (($2) + ($42<<2)|0);
   $52 = +HEAPF32[$51>>2];
   $53 = (($2) + ($46<<2)|0);
   $54 = +HEAPF32[$53>>2];
   $55 = $52 - $54;
   $56 = (($3) + ($42<<2)|0);
   $57 = +HEAPF32[$56>>2];
   $58 = (($3) + ($46<<2)|0);
   $59 = +HEAPF32[$58>>2];
   $60 = $57 + $59;
   $61 = (($4) + ($42<<2)|0);
   $62 = +HEAPF32[$61>>2];
   $63 = (($4) + ($46<<2)|0);
   $64 = +HEAPF32[$63>>2];
   $65 = $62 + $64;
   $66 = (($5) + ($42<<2)|0);
   $67 = +HEAPF32[$66>>2];
   $68 = (($5) + ($46<<2)|0);
   $69 = +HEAPF32[$68>>2];
   $70 = $67 - $69;
   $71 = $60 * $65;
   $72 = $55 * $70;
   $73 = $71 - $72;
   $74 = $50 * $70;
   $75 = $55 * $65;
   $76 = $74 - $75;
   $77 = $50 * $60;
   $78 = $55 * $55;
   $79 = $77 - $78;
   $80 = $x$136 * $76;
   $81 = $73 + $80;
   $82 = $81 / $79;
   $83 = $82 < 0.0;
   $R$0 = $83 ? 0.0 : $82;
   $84 = $R$0 - $offset;
   $85 = (($noise) + ($i$135<<2)|0);
   HEAPF32[$85>>2] = $84;
   $86 = (($i$135) + 1)|0;
   $87 = $x$136 + 1.0;
   $88 = (($b) + ($86<<2)|0);
   $89 = HEAP32[$88>>2]|0;
   $90 = $89 >> 16;
   $91 = ($90|0)>(-1);
   if ($91) {
    $40 = $89;$A$0$lcssa = $73;$B$0$lcssa = $76;$D$0$lcssa = $79;$i$1$lcssa = $86;$x$1$lcssa = $87;
    break;
   } else {
    $43 = $89;$47 = $90;$i$135 = $86;$x$136 = $87;
   }
  }
 }
 $39 = $40 & 65535;
 $41 = ($39|0)<($n|0);
 if ($41) {
  $$in = $40;$95 = $39;$i$227 = $i$1$lcssa;$x$228 = $x$1$lcssa;
  while(1) {
   $93 = $$in >> 16;
   $94 = (($1) + ($95<<2)|0);
   $96 = +HEAPF32[$94>>2];
   $97 = (($1) + ($93<<2)|0);
   $98 = +HEAPF32[$97>>2];
   $99 = $96 - $98;
   $100 = (($2) + ($95<<2)|0);
   $101 = +HEAPF32[$100>>2];
   $102 = (($2) + ($93<<2)|0);
   $103 = +HEAPF32[$102>>2];
   $104 = $101 - $103;
   $105 = (($3) + ($95<<2)|0);
   $106 = +HEAPF32[$105>>2];
   $107 = (($3) + ($93<<2)|0);
   $108 = +HEAPF32[$107>>2];
   $109 = $106 - $108;
   $110 = (($4) + ($95<<2)|0);
   $111 = +HEAPF32[$110>>2];
   $112 = (($4) + ($93<<2)|0);
   $113 = +HEAPF32[$112>>2];
   $114 = $111 - $113;
   $115 = (($5) + ($95<<2)|0);
   $116 = +HEAPF32[$115>>2];
   $117 = (($5) + ($93<<2)|0);
   $118 = +HEAPF32[$117>>2];
   $119 = $116 - $118;
   $120 = $109 * $114;
   $121 = $104 * $119;
   $122 = $120 - $121;
   $123 = $99 * $119;
   $124 = $104 * $114;
   $125 = $123 - $124;
   $126 = $99 * $109;
   $127 = $104 * $104;
   $128 = $126 - $127;
   $129 = $x$228 * $125;
   $130 = $122 + $129;
   $131 = $130 / $128;
   $132 = $131 < 0.0;
   $R$1 = $132 ? 0.0 : $131;
   $133 = $R$1 - $offset;
   $134 = (($noise) + ($i$227<<2)|0);
   HEAPF32[$134>>2] = $133;
   $135 = (($i$227) + 1)|0;
   $136 = $x$228 + 1.0;
   $137 = (($b) + ($135<<2)|0);
   $138 = HEAP32[$137>>2]|0;
   $139 = $138 & 65535;
   $140 = ($139|0)<($n|0);
   if ($140) {
    $$in = $138;$95 = $139;$i$227 = $135;$x$228 = $136;
   } else {
    $A$1$lcssa = $122;$B$1$lcssa = $125;$D$1$lcssa = $128;$i$2$lcssa = $135;$x$2$lcssa = $136;
    break;
   }
  }
 } else {
  $A$1$lcssa = $A$0$lcssa;$B$1$lcssa = $B$0$lcssa;$D$1$lcssa = $D$0$lcssa;$i$2$lcssa = $i$1$lcssa;$x$2$lcssa = $x$1$lcssa;
 }
 $92 = ($i$2$lcssa|0)<($n|0);
 if ($92) {
  $i$324 = $i$2$lcssa;$x$325 = $x$2$lcssa;
  while(1) {
   $141 = $B$1$lcssa * $x$325;
   $142 = $A$1$lcssa + $141;
   $143 = $142 / $D$1$lcssa;
   $144 = $143 < 0.0;
   $R$2 = $144 ? 0.0 : $143;
   $145 = $R$2 - $offset;
   $146 = (($noise) + ($i$324<<2)|0);
   HEAPF32[$146>>2] = $145;
   $147 = (($i$324) + 1)|0;
   $148 = $x$325 + 1.0;
   $exitcond73 = ($147|0)==($n|0);
   if ($exitcond73) {
    break;
   } else {
    $i$324 = $147;$x$325 = $148;
   }
  }
 }
 $149 = ($fixed|0)<(1);
 if ($149) {
  STACKTOP = sp;return;
 }
 $150 = (($fixed|0) / 2)&-1;
 $151 = (($150) - ($fixed))|0;
 $152 = ($151|0)>(-1);
 if ($152) {
  $A$2$lcssa = $A$1$lcssa;$B$2$lcssa = $B$1$lcssa;$D$2$lcssa = $D$1$lcssa;$i$4$lcssa = 0;$x$4$lcssa = 0.0;
 } else {
  $153 = (($fixed) - ($150))|0;
  $158 = $150;$161 = $151;$i$415 = 0;$x$416 = 0.0;
  while(1) {
   $157 = (($1) + ($158<<2)|0);
   $159 = +HEAPF32[$157>>2];
   $160 = (0 - ($161))|0;
   $162 = (($1) + ($160<<2)|0);
   $163 = +HEAPF32[$162>>2];
   $164 = $159 + $163;
   $165 = (($2) + ($158<<2)|0);
   $166 = +HEAPF32[$165>>2];
   $167 = (($2) + ($160<<2)|0);
   $168 = +HEAPF32[$167>>2];
   $169 = $166 - $168;
   $170 = (($3) + ($158<<2)|0);
   $171 = +HEAPF32[$170>>2];
   $172 = (($3) + ($160<<2)|0);
   $173 = +HEAPF32[$172>>2];
   $174 = $171 + $173;
   $175 = (($4) + ($158<<2)|0);
   $176 = +HEAPF32[$175>>2];
   $177 = (($4) + ($160<<2)|0);
   $178 = +HEAPF32[$177>>2];
   $179 = $176 + $178;
   $180 = (($5) + ($158<<2)|0);
   $181 = +HEAPF32[$180>>2];
   $182 = (($5) + ($160<<2)|0);
   $183 = +HEAPF32[$182>>2];
   $184 = $181 - $183;
   $185 = $174 * $179;
   $186 = $169 * $184;
   $187 = $185 - $186;
   $188 = $164 * $184;
   $189 = $169 * $179;
   $190 = $188 - $189;
   $191 = $164 * $174;
   $192 = $169 * $169;
   $193 = $191 - $192;
   $194 = $x$416 * $190;
   $195 = $187 + $194;
   $196 = $195 / $193;
   $197 = $196 - $offset;
   $198 = (($noise) + ($i$415<<2)|0);
   $199 = +HEAPF32[$198>>2];
   $200 = $197 < $199;
   if ($200) {
    HEAPF32[$198>>2] = $197;
   }
   $201 = (($i$415) + 1)|0;
   $202 = $x$416 + 1.0;
   $203 = (($201) + ($150))|0;
   $204 = (($203) - ($fixed))|0;
   $exitcond72 = ($201|0)==($153|0);
   if ($exitcond72) {
    $A$2$lcssa = $187;$B$2$lcssa = $190;$D$2$lcssa = $193;$i$4$lcssa = $153;$x$4$lcssa = $202;
    break;
   } else {
    $158 = $203;$161 = $204;$i$415 = $201;$x$416 = $202;
   }
  }
 }
 $154 = (($i$4$lcssa) + ($150))|0;
 $155 = ($154|0)<($n|0);
 if ($155) {
  $156 = (($n) - ($150))|0;
  $$pn = $154;$i$58 = $i$4$lcssa;$x$59 = $x$4$lcssa;
  while(1) {
   $206 = (($$pn) - ($fixed))|0;
   $207 = (($1) + ($$pn<<2)|0);
   $208 = +HEAPF32[$207>>2];
   $209 = (($1) + ($206<<2)|0);
   $210 = +HEAPF32[$209>>2];
   $211 = $208 - $210;
   $212 = (($2) + ($$pn<<2)|0);
   $213 = +HEAPF32[$212>>2];
   $214 = (($2) + ($206<<2)|0);
   $215 = +HEAPF32[$214>>2];
   $216 = $213 - $215;
   $217 = (($3) + ($$pn<<2)|0);
   $218 = +HEAPF32[$217>>2];
   $219 = (($3) + ($206<<2)|0);
   $220 = +HEAPF32[$219>>2];
   $221 = $218 - $220;
   $222 = (($4) + ($$pn<<2)|0);
   $223 = +HEAPF32[$222>>2];
   $224 = (($4) + ($206<<2)|0);
   $225 = +HEAPF32[$224>>2];
   $226 = $223 - $225;
   $227 = (($5) + ($$pn<<2)|0);
   $228 = +HEAPF32[$227>>2];
   $229 = (($5) + ($206<<2)|0);
   $230 = +HEAPF32[$229>>2];
   $231 = $228 - $230;
   $232 = $221 * $226;
   $233 = $216 * $231;
   $234 = $232 - $233;
   $235 = $211 * $231;
   $236 = $216 * $226;
   $237 = $235 - $236;
   $238 = $211 * $221;
   $239 = $216 * $216;
   $240 = $238 - $239;
   $241 = $x$59 * $237;
   $242 = $234 + $241;
   $243 = $242 / $240;
   $244 = $243 - $offset;
   $245 = (($noise) + ($i$58<<2)|0);
   $246 = +HEAPF32[$245>>2];
   $247 = $244 < $246;
   if ($247) {
    HEAPF32[$245>>2] = $244;
   }
   $248 = (($i$58) + 1)|0;
   $249 = $x$59 + 1.0;
   $250 = (($248) + ($150))|0;
   $exitcond71 = ($248|0)==($156|0);
   if ($exitcond71) {
    $A$3$lcssa = $234;$B$3$lcssa = $237;$D$3$lcssa = $240;$i$5$lcssa = $156;$x$5$lcssa = $249;
    break;
   } else {
    $$pn = $250;$i$58 = $248;$x$59 = $249;
   }
  }
 } else {
  $A$3$lcssa = $A$2$lcssa;$B$3$lcssa = $B$2$lcssa;$D$3$lcssa = $D$2$lcssa;$i$5$lcssa = $i$4$lcssa;$x$5$lcssa = $x$4$lcssa;
 }
 $205 = ($i$5$lcssa|0)<($n|0);
 if ($205) {
  $i$66 = $i$5$lcssa;$x$67 = $x$5$lcssa;
 } else {
  STACKTOP = sp;return;
 }
 while(1) {
  $251 = $B$3$lcssa * $x$67;
  $252 = $A$3$lcssa + $251;
  $253 = $252 / $D$3$lcssa;
  $254 = $253 - $offset;
  $255 = (($noise) + ($i$66<<2)|0);
  $256 = +HEAPF32[$255>>2];
  $257 = $254 < $256;
  if ($257) {
   HEAPF32[$255>>2] = $254;
  }
  $258 = (($i$66) + 1)|0;
  $259 = $x$67 + 1.0;
  $exitcond = ($258|0)==($n|0);
  if ($exitcond) {
   break;
  } else {
   $i$66 = $258;$x$67 = $259;
  }
 }
 STACKTOP = sp;return;
}
function __vp_tonemask($p,$logfft,$logmask,$global_specmax,$local_specmax) {
 $p = $p|0;
 $logfft = $logfft|0;
 $logmask = $logmask|0;
 $global_specmax = +$global_specmax;
 $local_specmax = +$local_specmax;
 var $$$$i = 0, $$$i = 0, $$alloca_mul = 0, $$in$i = 0, $$lcssa30 = 0, $$lcssa5$i = 0, $$lcssa73 = 0, $$phi$trans$insert$i = 0, $$pre = 0.0, $$pre$i = 0, $$pre$phi36Z2D = 0, $$pre$phiZ2D = 0, $$pre34 = 0, $$pre35 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0.0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0.0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0.0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0.0, $147 = 0, $148 = 0, $149 = 0, $15 = 0.0, $150 = 0, $151 = 0, $152 = 0, $153 = 0.0, $154 = 0, $155 = 0, $156 = 0.0, $157 = 0;
 var $158 = 0, $159 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0, $49 = 0.0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0.0, $80 = 0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0;
 var $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0.0, $90 = 0, $91 = 0, $92 = 0.0, $93 = 0.0, $94 = 0, $95 = 0.0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $att$0 = 0.0, $exitcond = 0, $exitcond33 = 0, $i$017 = 0, $i$02$i$i = 0, $i$04$i = 0;
 var $i$1$i = 0, $i$1$i$lcssa = 0, $i$1$ph$i = 0, $i$116 = 0, $linpos$0$lcssa$i = 0, $linpos$023$i = 0, $linpos$1$lcssa$i = 0, $linpos$116$i = 0, $linpos$27$i = 0, $max$0$ph$i = 0.0, $max$0$ph$i$lcssa = 0.0, $minV$0$ph$i = 0.0, $minV$0$ph3$lcssa$i = 0.0, $minV$0$ph313$i = 0.0, $or$cond$i$i = 0, $or$cond30$i = 0, $pos$022$i = 0, $pos$1$lcssa$i = 0, $seedptr$01$i$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$p>>2]|0;
 $1 = ((($p)) + 40|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2 << 2;
 $$alloca_mul = $3;
 $4 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $5 = ((($p)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($6)) + 4|0);
 $8 = +HEAPF32[$7>>2];
 $9 = $8 + $local_specmax;
 $10 = ($2|0)>(0);
 if ($10) {
  $i$017 = 0;
  while(1) {
   $11 = (($4) + ($i$017<<2)|0);
   HEAPF32[$11>>2] = -9999.0;
   $12 = (($i$017) + 1)|0;
   $13 = ($12|0)<($2|0);
   if ($13) {
    $i$017 = $12;
   } else {
    break;
   }
  }
 }
 $14 = ((($6)) + 8|0);
 $15 = +HEAPF32[$14>>2];
 $16 = $9 < $15;
 $att$0 = $16 ? $15 : $9;
 $17 = ($0|0)>(0);
 if ($17) {
  $18 = ((($p)) + 16|0);
  $19 = HEAP32[$18>>2]|0;
  $i$116 = 0;
  while(1) {
   $20 = (($19) + ($i$116<<2)|0);
   $21 = +HEAPF32[$20>>2];
   $22 = $att$0 + $21;
   $23 = (($logmask) + ($i$116<<2)|0);
   HEAPF32[$23>>2] = $22;
   $24 = (($i$116) + 1)|0;
   $exitcond33 = ($24|0)==($0|0);
   if ($exitcond33) {
    break;
   } else {
    $i$116 = $24;
   }
  }
  $25 = ((($p)) + 8|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = ((($6)) + 496|0);
  $28 = +HEAPF32[$27>>2];
  $29 = $28 - $global_specmax;
  if ($17) {
   $30 = ((($p)) + 20|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = ((($p)) + 32|0);
   $33 = ((($p)) + 36|0);
   $34 = ((($p)) + 28|0);
   $i$04$i = 0;
   while(1) {
    $35 = (($logfft) + ($i$04$i<<2)|0);
    $36 = +HEAPF32[$35>>2];
    $37 = (($31) + ($i$04$i<<2)|0);
    $38 = HEAP32[$37>>2]|0;
    $i$1$ph$i = $i$04$i;$max$0$ph$i = $36;
    L13: while(1) {
     $i$1$i = $i$1$ph$i;
     while(1) {
      $39 = (($i$1$i) + 1)|0;
      $40 = ($39|0)<($0|0);
      if (!($40)) {
       $$lcssa30 = 0;$$lcssa73 = $39;$i$1$i$lcssa = $i$1$i;$max$0$ph$i$lcssa = $max$0$ph$i;
       break L13;
      }
      $41 = (($31) + ($39<<2)|0);
      $42 = HEAP32[$41>>2]|0;
      $43 = ($42|0)==($38|0);
      if (!($43)) {
       $$lcssa30 = 1;$$lcssa73 = $39;$i$1$i$lcssa = $i$1$i;$max$0$ph$i$lcssa = $max$0$ph$i;
       break L13;
      }
      $44 = (($logfft) + ($39<<2)|0);
      $45 = +HEAPF32[$44>>2];
      $46 = $45 > $max$0$ph$i;
      if ($46) {
       $i$1$ph$i = $39;$max$0$ph$i = $45;
       continue L13;
      } else {
       $i$1$i = $39;
      }
     }
    }
    $47 = $max$0$ph$i$lcssa + 6.0;
    $48 = (($logmask) + ($i$1$i$lcssa<<2)|0);
    $49 = +HEAPF32[$48>>2];
    $50 = $47 > $49;
    if ($50) {
     $51 = HEAP32[$32>>2]|0;
     $52 = $38 >> $51;
     $53 = ($52|0)>(16);
     $$$i = $53 ? 16 : $52;
     $54 = ($$$i|0)<(0);
     $$$$i = $54 ? 0 : $$$i;
     $55 = (($26) + ($$$$i<<2)|0);
     $56 = HEAP32[$55>>2]|0;
     $57 = HEAP32[$1>>2]|0;
     $58 = HEAP32[$33>>2]|0;
     $59 = $29 + $max$0$ph$i$lcssa;
     $60 = $59;
     $61 = $60 + -30.0;
     $62 = $61 * 0.10000000149011612;
     $63 = (~~(($62)));
     $64 = ($63|0)<(0);
     $65 = $64 ? 0 : $63;
     $66 = ($65|0)>(7);
     $67 = $66 ? 7 : $65;
     $68 = (($56) + ($67<<2)|0);
     $69 = HEAP32[$68>>2]|0;
     $70 = ((($69)) + 8|0);
     $71 = ((($69)) + 4|0);
     $72 = +HEAPF32[$71>>2];
     $73 = (~~(($72)));
     $74 = +HEAPF32[$69>>2];
     $75 = (~~(($74)));
     $76 = ($75|0)<($73|0);
     if ($76) {
      $77 = (($31) + ($i$1$i$lcssa<<2)|0);
      $78 = HEAP32[$77>>2]|0;
      $79 = HEAP32[$34>>2]|0;
      $80 = (($78) - ($79))|0;
      $81 = (+($80|0));
      $82 = (+($58|0));
      $83 = $74 + -16.0;
      $84 = $82 * $83;
      $85 = $84 + $81;
      $86 = $58 >> 1;
      $87 = (+($86|0));
      $88 = $85 - $87;
      $89 = (~~(($88)));
      $i$02$i$i = $75;$seedptr$01$i$i = $89;
      while(1) {
       $90 = ($seedptr$01$i$i|0)>(0);
       if ($90) {
        $91 = (($70) + ($i$02$i$i<<2)|0);
        $92 = +HEAPF32[$91>>2];
        $93 = $max$0$ph$i$lcssa + $92;
        $94 = (($4) + ($seedptr$01$i$i<<2)|0);
        $95 = +HEAPF32[$94>>2];
        $96 = $95 < $93;
        if ($96) {
         HEAPF32[$94>>2] = $93;
        }
       }
       $97 = (($seedptr$01$i$i) + ($58))|0;
       $98 = ($97|0)<($57|0);
       $99 = (($i$02$i$i) + 1)|0;
       $100 = ($99|0)<($73|0);
       $or$cond$i$i = $100 & $98;
       if ($or$cond$i$i) {
        $i$02$i$i = $99;$seedptr$01$i$i = $97;
       } else {
        break;
       }
      }
     }
    }
    if ($$lcssa30) {
     $i$04$i = $$lcssa73;
    } else {
     $$pre$phi36Z2D = $30;$$pre$phiZ2D = $33;
     break;
    }
   }
  } else {
   label = 7;
  }
 } else {
  label = 7;
 }
 if ((label|0) == 7) {
  $$pre34 = ((($p)) + 36|0);
  $$pre35 = ((($p)) + 20|0);
  $$pre$phi36Z2D = $$pre35;$$pre$phiZ2D = $$pre34;
 }
 $101 = HEAP32[$1>>2]|0;
 $102 = HEAP32[$$pre$phiZ2D>>2]|0;
 _seed_chase($4,$102,$101);
 $103 = HEAP32[$p>>2]|0;
 $104 = ($103|0)>(1);
 L33: do {
  if ($104) {
   $105 = ((($p)) + 28|0);
   $106 = HEAP32[$$pre$phi36Z2D>>2]|0;
   $107 = HEAP32[$106>>2]|0;
   $108 = $102 >> 1;
   $109 = (($107) - ($108))|0;
   $110 = HEAP32[$105>>2]|0;
   $111 = (($109) - ($110))|0;
   $112 = HEAP32[$5>>2]|0;
   $113 = ((($112)) + 32|0);
   $119 = 1;$122 = $107;$linpos$023$i = 0;$pos$022$i = $111;
   while(1) {
    $116 = (($4) + ($pos$022$i<<2)|0);
    $117 = +HEAPF32[$116>>2];
    $118 = (($106) + ($119<<2)|0);
    $120 = HEAP32[$118>>2]|0;
    $121 = (($120) + ($122))|0;
    $123 = $121 >> 1;
    $124 = (($123) - ($110))|0;
    $125 = +HEAPF32[$113>>2];
    $126 = $117 > $125;
    $minV$0$ph$i = $126 ? $125 : $117;
    $127 = ($pos$022$i|0)<($124|0);
    L37: do {
     if ($127) {
      $$in$i = $pos$022$i;$minV$0$ph313$i = $minV$0$ph$i;
      while(1) {
       $128 = (($$in$i) + 1)|0;
       $129 = $minV$0$ph313$i == -9999.0;
       L40: do {
        if ($129) {
         $130 = (($4) + ($128<<2)|0);
         $$pre = +HEAPF32[$130>>2];
         $$lcssa5$i = $128;$159 = $$pre;
        } else {
         $132 = $128;
         while(1) {
          $131 = (($4) + ($132<<2)|0);
          $133 = +HEAPF32[$131>>2];
          $134 = $133 > -9999.0;
          $135 = $133 < $minV$0$ph313$i;
          $or$cond30$i = $134 & $135;
          if ($or$cond30$i) {
           $$lcssa5$i = $132;$159 = $133;
           break L40;
          }
          $136 = (($132) + 1)|0;
          $137 = ($132|0)<($124|0);
          if ($137) {
           $132 = $136;
          } else {
           $minV$0$ph3$lcssa$i = $minV$0$ph313$i;$pos$1$lcssa$i = $132;
           break L37;
          }
         }
        }
       } while(0);
       $138 = ($$lcssa5$i|0)<($124|0);
       if ($138) {
        $$in$i = $$lcssa5$i;$minV$0$ph313$i = $159;
       } else {
        $minV$0$ph3$lcssa$i = $159;$pos$1$lcssa$i = $$lcssa5$i;
        break;
       }
      }
     } else {
      $minV$0$ph3$lcssa$i = $minV$0$ph$i;$pos$1$lcssa$i = $pos$022$i;
     }
    } while(0);
    $139 = (($pos$1$lcssa$i) + ($110))|0;
    $140 = ($linpos$023$i|0)<($103|0);
    L47: do {
     if ($140) {
      $141 = HEAP32[$$pre$phi36Z2D>>2]|0;
      $linpos$116$i = $linpos$023$i;
      while(1) {
       $142 = (($141) + ($linpos$116$i<<2)|0);
       $143 = HEAP32[$142>>2]|0;
       $144 = ($143|0)>($139|0);
       if ($144) {
        $linpos$1$lcssa$i = $linpos$116$i;
        break L47;
       }
       $145 = (($logmask) + ($linpos$116$i<<2)|0);
       $146 = +HEAPF32[$145>>2];
       $147 = $146 < $minV$0$ph3$lcssa$i;
       if ($147) {
        HEAPF32[$145>>2] = $minV$0$ph3$lcssa$i;
       }
       $148 = (($linpos$116$i) + 1)|0;
       $149 = ($148|0)<($103|0);
       if ($149) {
        $linpos$116$i = $148;
       } else {
        $linpos$1$lcssa$i = $148;
        break;
       }
      }
     } else {
      $linpos$1$lcssa$i = $linpos$023$i;
     }
    } while(0);
    $114 = (($linpos$1$lcssa$i) + 1)|0;
    $115 = ($114|0)<($103|0);
    if (!($115)) {
     $linpos$0$lcssa$i = $linpos$1$lcssa$i;
     break L33;
    }
    $$phi$trans$insert$i = (($106) + ($linpos$1$lcssa$i<<2)|0);
    $$pre$i = HEAP32[$$phi$trans$insert$i>>2]|0;
    $119 = $114;$122 = $$pre$i;$linpos$023$i = $linpos$1$lcssa$i;$pos$022$i = $pos$1$lcssa$i;
   }
  } else {
   $linpos$0$lcssa$i = 0;
  }
 } while(0);
 $150 = HEAP32[$1>>2]|0;
 $151 = (($150) + -1)|0;
 $152 = (($4) + ($151<<2)|0);
 $153 = +HEAPF32[$152>>2];
 $154 = ($linpos$0$lcssa$i|0)<($103|0);
 if ($154) {
  $linpos$27$i = $linpos$0$lcssa$i;
 } else {
  STACKTOP = sp;return;
 }
 while(1) {
  $155 = (($logmask) + ($linpos$27$i<<2)|0);
  $156 = +HEAPF32[$155>>2];
  $157 = $156 < $153;
  if ($157) {
   HEAPF32[$155>>2] = $153;
  }
  $158 = (($linpos$27$i) + 1)|0;
  $exitcond = ($158|0)==($103|0);
  if ($exitcond) {
   break;
  } else {
   $linpos$27$i = $158;
  }
 }
 STACKTOP = sp;return;
}
function _seed_chase($seeds,$linesper,$n) {
 $seeds = $seeds|0;
 $linesper = $linesper|0;
 $n = $n|0;
 var $$alloca_mul = 0, $$alloca_mul1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0.0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0, $endpos$0 = 0, $exitcond = 0, $exitcond18 = 0;
 var $i$012 = 0, $i$17 = 0, $n$endpos$0 = 0, $or$cond = 0, $pos$06 = 0, $pos$1$lcssa = 0, $pos$15 = 0, $stack$011 = 0, $stack$1 = 0, $stack$1$lcssa = 0, $stack$1$lcssa22 = 0, $stack$2 = 0, $stack$2$in = 0, $stack$2$in$lcssa = 0, $stack$2$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $n << 2;
 $$alloca_mul = $0;
 $1 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $$alloca_mul1 = $0;
 $2 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul1)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $3 = ($n|0)>(0);
 if ($3) {
  $i$012 = 0;$stack$011 = 0;
 } else {
  STACKTOP = sp;return;
 }
 while(1) {
  $6 = ($stack$011|0)<(2);
  do {
   if ($6) {
    $9 = (($1) + ($stack$011<<2)|0);
    HEAP32[$9>>2] = $i$012;
    $10 = (($seeds) + ($i$012<<2)|0);
    $11 = HEAP32[$10>>2]|0;
    $12 = (($2) + ($stack$011<<2)|0);
    HEAP32[$12>>2] = $11;
    $stack$2$in = $stack$011;
   } else {
    $7 = (($seeds) + ($i$012<<2)|0);
    $8 = +HEAPF32[$7>>2];
    $stack$1 = $stack$011;
    while(1) {
     $13 = (($stack$1) + -1)|0;
     $14 = (($2) + ($13<<2)|0);
     $15 = +HEAPF32[$14>>2];
     $16 = $8 < $15;
     if ($16) {
      $stack$1$lcssa = $stack$1;
      label = 8;
      break;
     }
     $19 = (($1) + ($13<<2)|0);
     $20 = HEAP32[$19>>2]|0;
     $21 = (($20) + ($linesper))|0;
     $22 = ($i$012|0)<($21|0);
     $23 = ($stack$1|0)>(1);
     $or$cond = $23 & $22;
     if (!($or$cond)) {
      $stack$1$lcssa22 = $stack$1;
      label = 12;
      break;
     }
     $24 = (($stack$1) + -2)|0;
     $25 = (($2) + ($24<<2)|0);
     $26 = +HEAPF32[$25>>2];
     $27 = !($15 <= $26);
     if ($27) {
      $stack$1$lcssa22 = $stack$1;
      label = 12;
      break;
     }
     $28 = (($1) + ($24<<2)|0);
     $29 = HEAP32[$28>>2]|0;
     $30 = (($29) + ($linesper))|0;
     $31 = ($i$012|0)<($30|0);
     if ($31) {
      $stack$1 = $13;
     } else {
      $stack$1$lcssa22 = $stack$1;
      label = 12;
      break;
     }
    }
    if ((label|0) == 8) {
     label = 0;
     $17 = (($1) + ($stack$1$lcssa<<2)|0);
     HEAP32[$17>>2] = $i$012;
     $18 = (($2) + ($stack$1$lcssa<<2)|0);
     HEAPF32[$18>>2] = $8;
     $stack$2$in = $stack$1$lcssa;
     break;
    }
    else if ((label|0) == 12) {
     label = 0;
     $32 = (($1) + ($stack$1$lcssa22<<2)|0);
     HEAP32[$32>>2] = $i$012;
     $33 = (($2) + ($stack$1$lcssa22<<2)|0);
     HEAPF32[$33>>2] = $8;
     $stack$2$in = $stack$1$lcssa22;
     break;
    }
   }
  } while(0);
  $stack$2 = (($stack$2$in) + 1)|0;
  $34 = (($i$012) + 1)|0;
  $exitcond18 = ($34|0)==($n|0);
  if ($exitcond18) {
   $stack$2$in$lcssa = $stack$2$in;$stack$2$lcssa = $stack$2;
   break;
  } else {
   $i$012 = $34;$stack$011 = $stack$2;
  }
 }
 $4 = ($stack$2$in$lcssa|0)>(-1);
 if (!($4)) {
  STACKTOP = sp;return;
 }
 $5 = (($linesper) + 1)|0;
 $i$17 = 0;$pos$06 = 0;
 while(1) {
  $35 = ($i$17|0)<($stack$2$in$lcssa|0);
  if ($35) {
   $36 = (($i$17) + 1)|0;
   $37 = (($2) + ($36<<2)|0);
   $38 = +HEAPF32[$37>>2];
   $39 = (($2) + ($i$17<<2)|0);
   $40 = +HEAPF32[$39>>2];
   $41 = $38 > $40;
   if ($41) {
    $42 = (($1) + ($36<<2)|0);
    $43 = HEAP32[$42>>2]|0;
    $endpos$0 = $43;
   } else {
    label = 17;
   }
  } else {
   label = 17;
  }
  if ((label|0) == 17) {
   label = 0;
   $44 = (($1) + ($i$17<<2)|0);
   $45 = HEAP32[$44>>2]|0;
   $46 = (($5) + ($45))|0;
   $endpos$0 = $46;
  }
  $47 = ($endpos$0|0)>($n|0);
  $n$endpos$0 = $47 ? $n : $endpos$0;
  $48 = ($pos$06|0)<($n$endpos$0|0);
  if ($48) {
   $49 = (($2) + ($i$17<<2)|0);
   $50 = HEAP32[$49>>2]|0;
   $pos$15 = $pos$06;
   while(1) {
    $51 = (($seeds) + ($pos$15<<2)|0);
    HEAP32[$51>>2] = $50;
    $52 = (($pos$15) + 1)|0;
    $53 = ($52|0)<($n$endpos$0|0);
    if ($53) {
     $pos$15 = $52;
    } else {
     $pos$1$lcssa = $n$endpos$0;
     break;
    }
   }
  } else {
   $pos$1$lcssa = $pos$06;
  }
  $54 = (($i$17) + 1)|0;
  $exitcond = ($54|0)==($stack$2$lcssa|0);
  if ($exitcond) {
   break;
  } else {
   $i$17 = $54;$pos$06 = $pos$1$lcssa;
  }
 }
 STACKTOP = sp;return;
}
function __vp_offset_and_mix($p,$noise,$tone,$offset_select,$logmask,$mdct,$logmdct) {
 $p = $p|0;
 $noise = $noise|0;
 $tone = $tone|0;
 $offset_select = $offset_select|0;
 $logmask = $logmask|0;
 $mdct = $mdct|0;
 $logmdct = $logmdct|0;
 var $$val$0 = 0.0, $$val$0$us = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0.0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0.0, $40 = 0.0, $41 = 0, $42 = 0;
 var $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0, $47 = 0.0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0, $58 = 0, $6 = 0, $7 = 0.0, $8 = 0;
 var $9 = 0, $de$0$us = 0.0, $exitcond = 0, $exitcond3 = 0, $i$01 = 0, $i$01$us = 0, $val$0 = 0.0, $val$0$us = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$p>>2]|0;
 $1 = ((($p)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (((($2)) + 12|0) + ($offset_select<<2)|0);
 $4 = +HEAPF32[$3>>2];
 $5 = ($0|0)>(0);
 if (!($5)) {
  return;
 }
 $6 = ((($p)) + 48|0);
 $7 = +HEAPF32[$6>>2];
 $8 = ((($p)) + 12|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (($9) + ($offset_select<<2)|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($2)) + 108|0);
 $13 = ($offset_select|0)==(1);
 $14 = $7;
 if ($13) {
  $i$01$us = 0;
 } else {
  $i$01 = 0;
  while(1) {
   $46 = (($noise) + ($i$01<<2)|0);
   $47 = +HEAPF32[$46>>2];
   $48 = (($11) + ($i$01<<2)|0);
   $49 = +HEAPF32[$48>>2];
   $50 = $47 + $49;
   $51 = +HEAPF32[$12>>2];
   $52 = $50 > $51;
   $val$0 = $52 ? $51 : $50;
   $53 = (($tone) + ($i$01<<2)|0);
   $54 = +HEAPF32[$53>>2];
   $55 = $4 + $54;
   $56 = $val$0 < $55;
   $$val$0 = $56 ? $55 : $val$0;
   $57 = (($logmask) + ($i$01<<2)|0);
   HEAPF32[$57>>2] = $$val$0;
   $58 = (($i$01) + 1)|0;
   $exitcond3 = ($58|0)==($0|0);
   if ($exitcond3) {
    break;
   } else {
    $i$01 = $58;
   }
  }
  return;
 }
 while(1) {
  $15 = (($noise) + ($i$01$us<<2)|0);
  $16 = +HEAPF32[$15>>2];
  $17 = (($11) + ($i$01$us<<2)|0);
  $18 = +HEAPF32[$17>>2];
  $19 = $16 + $18;
  $20 = +HEAPF32[$12>>2];
  $21 = $19 > $20;
  $val$0$us = $21 ? $20 : $19;
  $22 = (($tone) + ($i$01$us<<2)|0);
  $23 = +HEAPF32[$22>>2];
  $24 = $4 + $23;
  $25 = $val$0$us < $24;
  $$val$0$us = $25 ? $24 : $val$0$us;
  $26 = (($logmask) + ($i$01$us<<2)|0);
  HEAPF32[$26>>2] = $$val$0$us;
  $27 = (($logmdct) + ($i$01$us<<2)|0);
  $28 = +HEAPF32[$27>>2];
  $29 = $val$0$us - $28;
  $30 = $29 > -17.200000762939453;
  $31 = $29 + 17.200000762939453;
  $32 = $31;
  if ($30) {
   $37 = $32 * 0.0050000000000000001;
   $38 = $14 * $37;
   $39 = 1.0 - $38;
   $40 = $39;
   $41 = $40 < 0.0;
   if ($41) {
    $de$0$us = 9.9999997473787516E-5;
   } else {
    $de$0$us = $40;
   }
  } else {
   $33 = $32 * 2.9999999999999997E-4;
   $34 = $14 * $33;
   $35 = 1.0 - $34;
   $36 = $35;
   $de$0$us = $36;
  }
  $42 = (($mdct) + ($i$01$us<<2)|0);
  $43 = +HEAPF32[$42>>2];
  $44 = $de$0$us * $43;
  HEAPF32[$42>>2] = $44;
  $45 = (($i$01$us) + 1)|0;
  $exitcond = ($45|0)==($0|0);
  if ($exitcond) {
   break;
  } else {
   $i$01$us = $45;
  }
 }
 return;
}
function _floor1_fit($vb,$look,$logmdct,$logmask) {
 $vb = $vb|0;
 $look = $look|0;
 $logmdct = $logmdct|0;
 $logmask = $logmask|0;
 var $$$0$i = 0, $$$i$i = 0, $$$i$i27 = 0, $$$i$i58 = 0, $$$i5$i = 0, $$0$i = 0, $$0$i$i = 0, $$0$i$i28 = 0, $$0$i$i59 = 0, $$0$i10 = 0, $$0$i54 = 0, $$0$i6$i = 0, $$0$i63 = 0, $$0$i7 = 0, $$0$i8 = 0, $$0$p$i = 0, $$pn$i = 0, $$pre = 0, $$pre118 = 0, $$pre119 = 0;
 var $$val$i = 0.0, $$val$i26 = 0.0, $$val$i61 = 0.0, $$val4$i = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0.0, $181 = 0.0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0.0, $192 = 0.0, $193 = 0.0, $194 = 0, $195 = 0.0, $196 = 0.0, $197 = 0.0, $198 = 0.0, $199 = 0, $2 = 0, $20 = 0, $200 = 0.0, $201 = 0.0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0.0, $212 = 0.0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0.0, $222 = 0.0, $223 = 0, $224 = 0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0.0, $229 = 0, $23 = 0, $230 = 0.0, $231 = 0.0, $232 = 0, $233 = 0, $234 = 0, $235 = 0.0, $236 = 0.0, $237 = 0.0, $238 = 0.0;
 var $239 = 0.0, $24 = 0.0, $240 = 0, $241 = 0.0, $242 = 0.0, $243 = 0.0, $244 = 0, $245 = 0, $246 = 0.0, $247 = 0, $248 = 0, $249 = 0, $25 = 0.0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0;
 var $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0;
 var $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0.0;
 var $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0.0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0;
 var $329 = 0, $33 = 0.0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0.0, $8 = 0;
 var $80 = 0.0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $err$010$i = 0, $err$1$i = 0, $exitcond = 0, $exitcond111 = 0, $exitcond112 = 0, $exitcond114 = 0, $exitcond116 = 0, $exitcond117 = 0, $fit_valueA = 0, $fit_valueB = 0, $fits = 0, $hineighbor = 0, $hy0 = 0, $hy1 = 0, $i$01$i = 0, $i$01$i25 = 0, $i$0106 = 0, $i$1104 = 0, $i$3100 = 0;
 var $i$595 = 0, $i$686 = 0, $i$777 = 0, $ispos$i = 0, $ispos$i56 = 0, $ispos2$i = 0, $j$079 = 0, $j$079$in = 0, $j$1 = 0, $j$182 = 0, $j$183 = 0, $loneighbor = 0, $ly0 = 0, $ly1 = 0, $memo = 0, $mse$0$lcssa$i = 0, $mse$011$i = 0, $n$0$lcssa$i = 0, $n$012$i = 0, $na$0$lcssa$i = 0;
 var $na$0$lcssa$i47 = 0, $na$07$i = 0, $na$07$i19 = 0, $na$1$i = 0, $na$1$i34 = 0, $nb$0$lcssa$i = 0, $nb$0$lcssa$i41 = 0, $nb$013$i = 0, $nb$013$i13 = 0, $nb$1$i = 0, $nb$1$i40 = 0, $neg$i = 0, $neg$i57 = 0, $neg3$i = 0, $nonzero$094 = 0, $nonzero$1 = 0, $or$cond = 0, $or$cond$i = 0, $or$cond6 = 0, $output$0 = 0;
 var $x1$$i = 0, $x2a$0$lcssa$i = 0, $x2a$0$lcssa$i50 = 0, $x2a$04$i = 0, $x2a$04$i22 = 0, $x2a$1$i = 0, $x2a$1$i31 = 0, $x2b$0$lcssa$i = 0, $x2b$0$lcssa$i44 = 0, $x2b$010$i = 0, $x2b$010$i16 = 0, $x2b$1$i = 0, $x2b$1$i37 = 0, $xa$0$lcssa$i = 0, $xa$0$lcssa$i52 = 0, $xa$02$i = 0, $xa$02$i24 = 0, $xa$1$i = 0, $xa$1$i29 = 0, $xb$0$lcssa$i = 0;
 var $xb$0$lcssa$i46 = 0, $xb$08$i = 0, $xb$08$i18 = 0, $xb$1$i = 0, $xb$1$i35 = 0, $xya$0$lcssa$i = 0, $xya$0$lcssa$i48 = 0, $xya$06$i = 0, $xya$06$i20 = 0, $xya$1$i = 0, $xya$1$i33 = 0, $xyb$0$lcssa$i = 0, $xyb$0$lcssa$i42 = 0, $xyb$012$i = 0, $xyb$012$i14 = 0, $xyb$1$i = 0, $xyb$1$i39 = 0, $y$09$i = 0, $y$1$i = 0, $y0 = 0;
 var $y1 = 0, $y2a$0$lcssa$i = 0, $y2a$0$lcssa$i49 = 0, $y2a$05$i = 0, $y2a$05$i21 = 0, $y2a$1$i = 0, $y2a$1$i32 = 0, $y2b$0$lcssa$i = 0, $y2b$0$lcssa$i43 = 0, $y2b$011$i = 0, $y2b$011$i15 = 0, $y2b$1$i = 0, $y2b$1$i38 = 0, $ya$0$lcssa$i = 0, $ya$0$lcssa$i51 = 0, $ya$03$i = 0, $ya$03$i23 = 0, $ya$1$i = 0, $ya$1$i30 = 0, $yb$0$lcssa$i = 0;
 var $yb$0$lcssa$i45 = 0, $yb$09$i = 0, $yb$09$i17 = 0, $yb$1$i = 0, $yb$1$i36 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 4912|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $fits = sp + 1328|0;
 $fit_valueA = sp + 1064|0;
 $fit_valueB = sp + 804|0;
 $loneighbor = sp + 544|0;
 $hineighbor = sp + 284|0;
 $memo = sp + 24|0;
 $y0 = sp + 20|0;
 $y1 = sp + 16|0;
 $ly0 = sp + 12|0;
 $ly1 = sp + 8|0;
 $hy0 = sp + 4|0;
 $hy1 = sp;
 $0 = ((($look)) + 1296|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($look)) + 1288|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($look)) + 1284|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)>(0);
 if ($6) {
  $i$0106 = 0;
  while(1) {
   $7 = (($fit_valueA) + ($i$0106<<2)|0);
   HEAP32[$7>>2] = -200;
   $8 = (($i$0106) + 1)|0;
   $exitcond117 = ($8|0)==($5|0);
   if ($exitcond117) {
    break;
   } else {
    $i$0106 = $8;
   }
  }
  if ($6) {
   $i$1104 = 0;
   while(1) {
    $9 = (($fit_valueB) + ($i$1104<<2)|0);
    HEAP32[$9>>2] = -200;
    $10 = (($i$1104) + 1)|0;
    $exitcond116 = ($10|0)==($5|0);
    if ($exitcond116) {
     break;
    } else {
     $i$1104 = $10;
    }
   }
   if ($6) {
    $11 = $5 << 2;
    _memset(($loneighbor|0),0,($11|0))|0;
    $i$3100 = 0;
    while(1) {
     $12 = (($hineighbor) + ($i$3100<<2)|0);
     HEAP32[$12>>2] = 1;
     $13 = (($i$3100) + 1)|0;
     $exitcond114 = ($13|0)==($5|0);
     if ($exitcond114) {
      break;
     } else {
      $i$3100 = $13;
     }
    }
    if ($6) {
     $15 = $5 << 2;
     _memset(($memo|0),-1,($15|0))|0;
     $16 = ($5|0)>(1);
     if (!($16)) {
      $output$0 = 0;
      STACKTOP = sp;return ($output$0|0);
     }
     $17 = (($3) + -1)|0;
     $18 = ((($1)) + 1112|0);
     $19 = (($5) + -1)|0;
     $$pre = HEAP32[$look>>2]|0;
     $74 = $$pre;$i$595 = 0;$nonzero$094 = 0;
     while(1) {
      $70 = (($i$595) + 1)|0;
      $71 = (($look) + ($70<<2)|0);
      $72 = HEAP32[$71>>2]|0;
      $73 = (($fits) + (($i$595*56)|0)|0);
      dest=$73; stop=dest+56|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
      HEAP32[$73>>2] = $74;
      $75 = (((($fits) + (($i$595*56)|0)|0)) + 4|0);
      HEAP32[$75>>2] = $72;
      $76 = ($72|0)<($3|0);
      $x1$$i = $76 ? $72 : $17;
      $77 = ($x1$$i|0)<($74|0);
      if ($77) {
       $na$0$lcssa$i47 = 0;$nb$0$lcssa$i41 = 0;$x2a$0$lcssa$i50 = 0;$x2b$0$lcssa$i44 = 0;$xa$0$lcssa$i52 = 0;$xb$0$lcssa$i46 = 0;$xya$0$lcssa$i48 = 0;$xyb$0$lcssa$i42 = 0;$y2a$0$lcssa$i49 = 0;$y2b$0$lcssa$i43 = 0;$ya$0$lcssa$i51 = 0;$yb$0$lcssa$i45 = 0;
      } else {
       $i$01$i25 = $74;$na$07$i19 = 0;$nb$013$i13 = 0;$x2a$04$i22 = 0;$x2b$010$i16 = 0;$xa$02$i24 = 0;$xb$08$i18 = 0;$xya$06$i20 = 0;$xyb$012$i14 = 0;$y2a$05$i21 = 0;$y2b$011$i15 = 0;$ya$03$i23 = 0;$yb$09$i17 = 0;
       while(1) {
        $78 = (($logmask) + ($i$01$i25<<2)|0);
        $$val$i26 = +HEAPF32[$78>>2];
        $79 = $$val$i26 * 7.3142857551574707;
        $80 = $79 + 1023.5;
        $81 = (~~(($80)));
        $82 = ($81|0)>(1023);
        $83 = ($81|0)<(0);
        $$$i$i27 = $83 ? 0 : $81;
        $$0$i$i28 = $82 ? 1023 : $$$i$i27;
        $84 = ($$0$i$i28|0)==(0);
        do {
         if ($84) {
          $na$1$i34 = $na$07$i19;$nb$1$i40 = $nb$013$i13;$x2a$1$i31 = $x2a$04$i22;$x2b$1$i37 = $x2b$010$i16;$xa$1$i29 = $xa$02$i24;$xb$1$i35 = $xb$08$i18;$xya$1$i33 = $xya$06$i20;$xyb$1$i39 = $xyb$012$i14;$y2a$1$i32 = $y2a$05$i21;$y2b$1$i38 = $y2b$011$i15;$ya$1$i30 = $ya$03$i23;$yb$1$i36 = $yb$09$i17;
         } else {
          $85 = (($logmdct) + ($i$01$i25<<2)|0);
          $86 = +HEAPF32[$85>>2];
          $87 = +HEAPF32[$18>>2];
          $88 = $86 + $87;
          $89 = !($88 >= $$val$i26);
          if ($89) {
           $101 = (($i$01$i25) + ($xb$08$i18))|0;
           $102 = (($$0$i$i28) + ($yb$09$i17))|0;
           $103 = Math_imul($i$01$i25, $i$01$i25)|0;
           $104 = (($103) + ($x2b$010$i16))|0;
           $105 = Math_imul($81, $81)|0;
           $106 = $83 ? 0 : $105;
           $107 = $82 ? 1046529 : $106;
           $108 = (($107) + ($y2b$011$i15))|0;
           $109 = Math_imul($$0$i$i28, $i$01$i25)|0;
           $110 = (($109) + ($xyb$012$i14))|0;
           $111 = (($nb$013$i13) + 1)|0;
           $na$1$i34 = $na$07$i19;$nb$1$i40 = $111;$x2a$1$i31 = $x2a$04$i22;$x2b$1$i37 = $104;$xa$1$i29 = $xa$02$i24;$xb$1$i35 = $101;$xya$1$i33 = $xya$06$i20;$xyb$1$i39 = $110;$y2a$1$i32 = $y2a$05$i21;$y2b$1$i38 = $108;$ya$1$i30 = $ya$03$i23;$yb$1$i36 = $102;
           break;
          } else {
           $90 = (($i$01$i25) + ($xa$02$i24))|0;
           $91 = (($$0$i$i28) + ($ya$03$i23))|0;
           $92 = Math_imul($i$01$i25, $i$01$i25)|0;
           $93 = (($92) + ($x2a$04$i22))|0;
           $94 = Math_imul($81, $81)|0;
           $95 = $83 ? 0 : $94;
           $96 = $82 ? 1046529 : $95;
           $97 = (($96) + ($y2a$05$i21))|0;
           $98 = Math_imul($$0$i$i28, $i$01$i25)|0;
           $99 = (($98) + ($xya$06$i20))|0;
           $100 = (($na$07$i19) + 1)|0;
           $na$1$i34 = $100;$nb$1$i40 = $nb$013$i13;$x2a$1$i31 = $93;$x2b$1$i37 = $x2b$010$i16;$xa$1$i29 = $90;$xb$1$i35 = $xb$08$i18;$xya$1$i33 = $99;$xyb$1$i39 = $xyb$012$i14;$y2a$1$i32 = $97;$y2b$1$i38 = $y2b$011$i15;$ya$1$i30 = $91;$yb$1$i36 = $yb$09$i17;
           break;
          }
         }
        } while(0);
        $112 = (($i$01$i25) + 1)|0;
        $113 = ($i$01$i25|0)<($x1$$i|0);
        if ($113) {
         $i$01$i25 = $112;$na$07$i19 = $na$1$i34;$nb$013$i13 = $nb$1$i40;$x2a$04$i22 = $x2a$1$i31;$x2b$010$i16 = $x2b$1$i37;$xa$02$i24 = $xa$1$i29;$xb$08$i18 = $xb$1$i35;$xya$06$i20 = $xya$1$i33;$xyb$012$i14 = $xyb$1$i39;$y2a$05$i21 = $y2a$1$i32;$y2b$011$i15 = $y2b$1$i38;$ya$03$i23 = $ya$1$i30;$yb$09$i17 = $yb$1$i36;
        } else {
         $na$0$lcssa$i47 = $na$1$i34;$nb$0$lcssa$i41 = $nb$1$i40;$x2a$0$lcssa$i50 = $x2a$1$i31;$x2b$0$lcssa$i44 = $x2b$1$i37;$xa$0$lcssa$i52 = $xa$1$i29;$xb$0$lcssa$i46 = $xb$1$i35;$xya$0$lcssa$i48 = $xya$1$i33;$xyb$0$lcssa$i42 = $xyb$1$i39;$y2a$0$lcssa$i49 = $y2a$1$i32;$y2b$0$lcssa$i43 = $y2b$1$i38;$ya$0$lcssa$i51 = $ya$1$i30;$yb$0$lcssa$i45 = $yb$1$i36;
         break;
        }
       }
      }
      $114 = (((($fits) + (($i$595*56)|0)|0)) + 8|0);
      HEAP32[$114>>2] = $xa$0$lcssa$i52;
      $115 = (((($fits) + (($i$595*56)|0)|0)) + 12|0);
      HEAP32[$115>>2] = $ya$0$lcssa$i51;
      $116 = (((($fits) + (($i$595*56)|0)|0)) + 16|0);
      HEAP32[$116>>2] = $x2a$0$lcssa$i50;
      $117 = (((($fits) + (($i$595*56)|0)|0)) + 20|0);
      HEAP32[$117>>2] = $y2a$0$lcssa$i49;
      $118 = (((($fits) + (($i$595*56)|0)|0)) + 24|0);
      HEAP32[$118>>2] = $xya$0$lcssa$i48;
      $119 = (((($fits) + (($i$595*56)|0)|0)) + 28|0);
      HEAP32[$119>>2] = $na$0$lcssa$i47;
      $120 = (((($fits) + (($i$595*56)|0)|0)) + 32|0);
      HEAP32[$120>>2] = $xb$0$lcssa$i46;
      $121 = (((($fits) + (($i$595*56)|0)|0)) + 36|0);
      HEAP32[$121>>2] = $yb$0$lcssa$i45;
      $122 = (((($fits) + (($i$595*56)|0)|0)) + 40|0);
      HEAP32[$122>>2] = $x2b$0$lcssa$i44;
      $123 = (((($fits) + (($i$595*56)|0)|0)) + 44|0);
      HEAP32[$123>>2] = $y2b$0$lcssa$i43;
      $124 = (((($fits) + (($i$595*56)|0)|0)) + 48|0);
      HEAP32[$124>>2] = $xyb$0$lcssa$i42;
      $125 = (((($fits) + (($i$595*56)|0)|0)) + 52|0);
      HEAP32[$125>>2] = $nb$0$lcssa$i41;
      $126 = (($na$0$lcssa$i47) + ($nonzero$094))|0;
      $exitcond112 = ($70|0)==($19|0);
      if ($exitcond112) {
       $nonzero$1 = $126;
       break;
      } else {
       $74 = $72;$i$595 = $70;$nonzero$094 = $126;
      }
     }
    } else {
     label = 9;
    }
   } else {
    label = 9;
   }
  } else {
   label = 9;
  }
 } else {
  label = 9;
 }
 if ((label|0) == 9) {
  $14 = ($5|0)==(0);
  if (!($14)) {
   $output$0 = 0;
   STACKTOP = sp;return ($output$0|0);
  }
  $20 = ((($fits)) + 4|0);
  dest=$fits; stop=dest+56|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
  HEAP32[$20>>2] = $3;
  $21 = ($3|0)<(1);
  if ($21) {
   $na$0$lcssa$i = 0;$nb$0$lcssa$i = 0;$x2a$0$lcssa$i = 0;$x2b$0$lcssa$i = 0;$xa$0$lcssa$i = 0;$xb$0$lcssa$i = 0;$xya$0$lcssa$i = 0;$xyb$0$lcssa$i = 0;$y2a$0$lcssa$i = 0;$y2b$0$lcssa$i = 0;$ya$0$lcssa$i = 0;$yb$0$lcssa$i = 0;
  } else {
   $22 = ((($1)) + 1112|0);
   $i$01$i = 0;$na$07$i = 0;$nb$013$i = 0;$x2a$04$i = 0;$x2b$010$i = 0;$xa$02$i = 0;$xb$08$i = 0;$xya$06$i = 0;$xyb$012$i = 0;$y2a$05$i = 0;$y2b$011$i = 0;$ya$03$i = 0;$yb$09$i = 0;
   while(1) {
    $23 = (($logmask) + ($i$01$i<<2)|0);
    $$val$i = +HEAPF32[$23>>2];
    $24 = $$val$i * 7.3142857551574707;
    $25 = $24 + 1023.5;
    $26 = (~~(($25)));
    $27 = ($26|0)>(1023);
    $28 = ($26|0)<(0);
    $$$i$i = $28 ? 0 : $26;
    $$0$i$i = $27 ? 1023 : $$$i$i;
    $29 = ($$0$i$i|0)==(0);
    do {
     if ($29) {
      $na$1$i = $na$07$i;$nb$1$i = $nb$013$i;$x2a$1$i = $x2a$04$i;$x2b$1$i = $x2b$010$i;$xa$1$i = $xa$02$i;$xb$1$i = $xb$08$i;$xya$1$i = $xya$06$i;$xyb$1$i = $xyb$012$i;$y2a$1$i = $y2a$05$i;$y2b$1$i = $y2b$011$i;$ya$1$i = $ya$03$i;$yb$1$i = $yb$09$i;
     } else {
      $30 = (($logmdct) + ($i$01$i<<2)|0);
      $31 = +HEAPF32[$30>>2];
      $32 = +HEAPF32[$22>>2];
      $33 = $31 + $32;
      $34 = !($33 >= $$val$i);
      if ($34) {
       $46 = (($i$01$i) + ($xb$08$i))|0;
       $47 = (($$0$i$i) + ($yb$09$i))|0;
       $48 = Math_imul($i$01$i, $i$01$i)|0;
       $49 = (($48) + ($x2b$010$i))|0;
       $50 = Math_imul($26, $26)|0;
       $51 = $28 ? 0 : $50;
       $52 = $27 ? 1046529 : $51;
       $53 = (($52) + ($y2b$011$i))|0;
       $54 = Math_imul($$0$i$i, $i$01$i)|0;
       $55 = (($54) + ($xyb$012$i))|0;
       $56 = (($nb$013$i) + 1)|0;
       $na$1$i = $na$07$i;$nb$1$i = $56;$x2a$1$i = $x2a$04$i;$x2b$1$i = $49;$xa$1$i = $xa$02$i;$xb$1$i = $46;$xya$1$i = $xya$06$i;$xyb$1$i = $55;$y2a$1$i = $y2a$05$i;$y2b$1$i = $53;$ya$1$i = $ya$03$i;$yb$1$i = $47;
       break;
      } else {
       $35 = (($i$01$i) + ($xa$02$i))|0;
       $36 = (($$0$i$i) + ($ya$03$i))|0;
       $37 = Math_imul($i$01$i, $i$01$i)|0;
       $38 = (($37) + ($x2a$04$i))|0;
       $39 = Math_imul($26, $26)|0;
       $40 = $28 ? 0 : $39;
       $41 = $27 ? 1046529 : $40;
       $42 = (($41) + ($y2a$05$i))|0;
       $43 = Math_imul($$0$i$i, $i$01$i)|0;
       $44 = (($43) + ($xya$06$i))|0;
       $45 = (($na$07$i) + 1)|0;
       $na$1$i = $45;$nb$1$i = $nb$013$i;$x2a$1$i = $38;$x2b$1$i = $x2b$010$i;$xa$1$i = $35;$xb$1$i = $xb$08$i;$xya$1$i = $44;$xyb$1$i = $xyb$012$i;$y2a$1$i = $42;$y2b$1$i = $y2b$011$i;$ya$1$i = $36;$yb$1$i = $yb$09$i;
       break;
      }
     }
    } while(0);
    $57 = (($i$01$i) + 1)|0;
    $exitcond111 = ($57|0)==($3|0);
    if ($exitcond111) {
     $na$0$lcssa$i = $na$1$i;$nb$0$lcssa$i = $nb$1$i;$x2a$0$lcssa$i = $x2a$1$i;$x2b$0$lcssa$i = $x2b$1$i;$xa$0$lcssa$i = $xa$1$i;$xb$0$lcssa$i = $xb$1$i;$xya$0$lcssa$i = $xya$1$i;$xyb$0$lcssa$i = $xyb$1$i;$y2a$0$lcssa$i = $y2a$1$i;$y2b$0$lcssa$i = $y2b$1$i;$ya$0$lcssa$i = $ya$1$i;$yb$0$lcssa$i = $yb$1$i;
     break;
    } else {
     $i$01$i = $57;$na$07$i = $na$1$i;$nb$013$i = $nb$1$i;$x2a$04$i = $x2a$1$i;$x2b$010$i = $x2b$1$i;$xa$02$i = $xa$1$i;$xb$08$i = $xb$1$i;$xya$06$i = $xya$1$i;$xyb$012$i = $xyb$1$i;$y2a$05$i = $y2a$1$i;$y2b$011$i = $y2b$1$i;$ya$03$i = $ya$1$i;$yb$09$i = $yb$1$i;
    }
   }
  }
  $58 = ((($fits)) + 8|0);
  HEAP32[$58>>2] = $xa$0$lcssa$i;
  $59 = ((($fits)) + 12|0);
  HEAP32[$59>>2] = $ya$0$lcssa$i;
  $60 = ((($fits)) + 16|0);
  HEAP32[$60>>2] = $x2a$0$lcssa$i;
  $61 = ((($fits)) + 20|0);
  HEAP32[$61>>2] = $y2a$0$lcssa$i;
  $62 = ((($fits)) + 24|0);
  HEAP32[$62>>2] = $xya$0$lcssa$i;
  $63 = ((($fits)) + 28|0);
  HEAP32[$63>>2] = $na$0$lcssa$i;
  $64 = ((($fits)) + 32|0);
  HEAP32[$64>>2] = $xb$0$lcssa$i;
  $65 = ((($fits)) + 36|0);
  HEAP32[$65>>2] = $yb$0$lcssa$i;
  $66 = ((($fits)) + 40|0);
  HEAP32[$66>>2] = $x2b$0$lcssa$i;
  $67 = ((($fits)) + 44|0);
  HEAP32[$67>>2] = $y2b$0$lcssa$i;
  $68 = ((($fits)) + 48|0);
  HEAP32[$68>>2] = $xyb$0$lcssa$i;
  $69 = ((($fits)) + 52|0);
  HEAP32[$69>>2] = $nb$0$lcssa$i;
  $nonzero$1 = $na$0$lcssa$i;
 }
 $127 = ($nonzero$1|0)==(0);
 if ($127) {
  $output$0 = 0;
  STACKTOP = sp;return ($output$0|0);
 }
 HEAP32[$y0>>2] = -200;
 HEAP32[$y1>>2] = -200;
 $128 = (($5) + -1)|0;
 (_fit_line($fits,$128,$y0,$y1,$1)|0);
 $129 = HEAP32[$y0>>2]|0;
 HEAP32[$fit_valueA>>2] = $129;
 HEAP32[$fit_valueB>>2] = $129;
 $130 = HEAP32[$y1>>2]|0;
 $131 = ((($fit_valueB)) + 4|0);
 HEAP32[$131>>2] = $130;
 $132 = ((($fit_valueA)) + 4|0);
 HEAP32[$132>>2] = $130;
 $133 = ($5|0)>(2);
 do {
  if ($133) {
   $134 = ((($1)) + 1112|0);
   $135 = ((($1)) + 1096|0);
   $136 = ((($1)) + 1100|0);
   $137 = ((($1)) + 1104|0);
   $i$686 = 2;
   L50: while(1) {
    $138 = (((($look)) + 520|0) + ($i$686<<2)|0);
    $139 = HEAP32[$138>>2]|0;
    $140 = (($loneighbor) + ($139<<2)|0);
    $141 = HEAP32[$140>>2]|0;
    $142 = (($hineighbor) + ($139<<2)|0);
    $143 = HEAP32[$142>>2]|0;
    $144 = (($memo) + ($141<<2)|0);
    $145 = HEAP32[$144>>2]|0;
    $146 = ($145|0)==($143|0);
    L52: do {
     if (!($146)) {
      $147 = (((($look)) + 520|0) + ($141<<2)|0);
      $148 = HEAP32[$147>>2]|0;
      $149 = (((($look)) + 520|0) + ($143<<2)|0);
      $150 = HEAP32[$149>>2]|0;
      HEAP32[$144>>2] = $143;
      $151 = (((($1)) + 836|0) + ($141<<2)|0);
      $152 = HEAP32[$151>>2]|0;
      $153 = (((($1)) + 836|0) + ($143<<2)|0);
      $154 = HEAP32[$153>>2]|0;
      $155 = (($fit_valueA) + ($141<<2)|0);
      $156 = HEAP32[$155>>2]|0;
      $157 = ($156|0)<(0);
      $158 = (($fit_valueB) + ($141<<2)|0);
      $159 = HEAP32[$158>>2]|0;
      if ($157) {
       $$0$i54 = $159;
      } else {
       $160 = ($159|0)<(0);
       if ($160) {
        $$0$i54 = $156;
       } else {
        $161 = (($159) + ($156))|0;
        $162 = $161 >> 1;
        $$0$i54 = $162;
       }
      }
      $163 = (($fit_valueA) + ($143<<2)|0);
      $164 = HEAP32[$163>>2]|0;
      $165 = ($164|0)<(0);
      $166 = (($fit_valueB) + ($143<<2)|0);
      $167 = HEAP32[$166>>2]|0;
      if ($165) {
       $$0$i63 = $167;
      } else {
       $168 = ($167|0)<(0);
       if ($168) {
        $$0$i63 = $164;
       } else {
        $169 = (($167) + ($164))|0;
        $170 = $169 >> 1;
        $$0$i63 = $170;
       }
      }
      $171 = ($$0$i54|0)==(-1);
      $172 = ($$0$i63|0)==(-1);
      $or$cond = $171 | $172;
      if ($or$cond) {
       label = 38;
       break L50;
      }
      $173 = (($$0$i63) - ($$0$i54))|0;
      $174 = (($154) - ($152))|0;
      $ispos$i56 = ($173|0)>(-1);
      $neg$i57 = (0 - ($173))|0;
      $175 = $ispos$i56 ? $173 : $neg$i57;
      $176 = (($173|0) / ($174|0))&-1;
      $177 = $173 >> 31;
      $178 = $177 | 1;
      $179 = (($logmask) + ($152<<2)|0);
      $$val4$i = +HEAPF32[$179>>2];
      $180 = $$val4$i * 7.3142857551574707;
      $181 = $180 + 1023.5;
      $182 = (~~(($181)));
      $183 = ($182|0)>(1023);
      $184 = ($182|0)<(0);
      $$$i$i58 = $184 ? 0 : $182;
      $$0$i$i59 = $183 ? 1023 : $$$i$i58;
      $185 = Math_imul($176, $174)|0;
      $ispos2$i = ($185|0)>(-1);
      $neg3$i = (0 - ($185))|0;
      $186 = $ispos2$i ? $185 : $neg3$i;
      $187 = (($175) - ($186))|0;
      $188 = (($$0$i54) - ($$0$i$i59))|0;
      $189 = Math_imul($188, $188)|0;
      $190 = (($logmdct) + ($152<<2)|0);
      $191 = +HEAPF32[$190>>2];
      $192 = +HEAPF32[$134>>2];
      $193 = $191 + $192;
      $194 = !($193 >= $$val4$i);
      if ($194) {
       label = 42;
      } else {
       $195 = (+($$0$i54|0));
       $196 = +HEAPF32[$135>>2];
       $197 = $195 + $196;
       $198 = (+($$0$i$i59|0));
       $199 = $197 < $198;
       if (!($199)) {
        $200 = +HEAPF32[$136>>2];
        $201 = $195 - $200;
        $202 = $201 > $198;
        if (!($202)) {
         label = 42;
        }
       }
      }
      L66: do {
       if ((label|0) == 42) {
        label = 0;
        $203 = (($152) + 1)|0;
        $204 = ($203|0)<($154|0);
        if ($204) {
         $210 = $203;$err$010$i = 0;$mse$011$i = $189;$n$012$i = 1;$y$09$i = $$0$i54;
         while(1) {
          $205 = (($err$010$i) + ($187))|0;
          $206 = ($205|0)<($174|0);
          $207 = $206 ? 0 : $178;
          $208 = $206 ? 0 : $174;
          $err$1$i = (($205) - ($208))|0;
          $$pn$i = (($y$09$i) + ($176))|0;
          $y$1$i = (($$pn$i) + ($207))|0;
          $209 = (($logmask) + ($210<<2)|0);
          $$val$i61 = +HEAPF32[$209>>2];
          $211 = $$val$i61 * 7.3142857551574707;
          $212 = $211 + 1023.5;
          $213 = (~~(($212)));
          $214 = ($213|0)>(1023);
          $215 = ($213|0)<(0);
          $$$i5$i = $215 ? 0 : $213;
          $$0$i6$i = $214 ? 1023 : $$$i5$i;
          $216 = (($y$1$i) - ($$0$i6$i))|0;
          $217 = Math_imul($216, $216)|0;
          $218 = (($217) + ($mse$011$i))|0;
          $219 = (($n$012$i) + 1)|0;
          $220 = (($logmdct) + ($210<<2)|0);
          $221 = +HEAPF32[$220>>2];
          $222 = $192 + $221;
          $223 = $222 >= $$val$i61;
          $224 = ($$0$i6$i|0)!=(0);
          $or$cond$i = $223 & $224;
          if ($or$cond$i) {
           $225 = (+($y$1$i|0));
           $226 = +HEAPF32[$135>>2];
           $227 = $225 + $226;
           $228 = (+($$0$i6$i|0));
           $229 = $227 < $228;
           if ($229) {
            break L66;
           }
           $230 = +HEAPF32[$136>>2];
           $231 = $225 - $230;
           $232 = $231 > $228;
           if ($232) {
            break L66;
           }
          }
          $233 = (($210) + 1)|0;
          $234 = ($233|0)<($154|0);
          if ($234) {
           $210 = $233;$err$010$i = $err$1$i;$mse$011$i = $218;$n$012$i = $219;$y$09$i = $y$1$i;
          } else {
           $mse$0$lcssa$i = $218;$n$0$lcssa$i = $219;
           break;
          }
         }
        } else {
         $mse$0$lcssa$i = $189;$n$0$lcssa$i = 1;
        }
        $235 = +HEAPF32[$135>>2];
        $236 = $235 * $235;
        $237 = (+($n$0$lcssa$i|0));
        $238 = $236 / $237;
        $239 = +HEAPF32[$137>>2];
        $240 = $238 > $239;
        if (!($240)) {
         $241 = +HEAPF32[$136>>2];
         $242 = $241 * $241;
         $243 = $242 / $237;
         $244 = $243 > $239;
         if (!($244)) {
          $245 = (($mse$0$lcssa$i|0) / ($n$0$lcssa$i|0))&-1;
          $246 = (+($245|0));
          $247 = $246 > $239;
          if ($247) {
           break;
          }
         }
        }
        $280 = (($fit_valueA) + ($i$686<<2)|0);
        HEAP32[$280>>2] = -200;
        $281 = (($fit_valueB) + ($i$686<<2)|0);
        HEAP32[$281>>2] = -200;
        break L52;
       }
      } while(0);
      HEAP32[$ly0>>2] = -200;
      HEAP32[$ly1>>2] = -200;
      HEAP32[$hy0>>2] = -200;
      HEAP32[$hy1>>2] = -200;
      $248 = (($fits) + (($148*56)|0)|0);
      $249 = (($139) - ($148))|0;
      $250 = (_fit_line($248,$249,$ly0,$ly1,$1)|0);
      $251 = (($fits) + (($139*56)|0)|0);
      $252 = (($150) - ($139))|0;
      $253 = (_fit_line($251,$252,$hy0,$hy1,$1)|0);
      $254 = ($250|0)!=(0);
      if ($254) {
       HEAP32[$ly0>>2] = $$0$i54;
       $255 = HEAP32[$hy0>>2]|0;
       HEAP32[$ly1>>2] = $255;
      }
      $256 = ($253|0)==(0);
      if ($256) {
       label = 55;
      } else {
       $257 = HEAP32[$ly1>>2]|0;
       HEAP32[$hy0>>2] = $257;
       HEAP32[$hy1>>2] = $$0$i63;
       if ($254) {
        $258 = (($fit_valueA) + ($i$686<<2)|0);
        HEAP32[$258>>2] = -200;
        $259 = (($fit_valueB) + ($i$686<<2)|0);
        HEAP32[$259>>2] = -200;
       } else {
        label = 55;
       }
      }
      L87: do {
       if ((label|0) == 55) {
        label = 0;
        $260 = HEAP32[$ly0>>2]|0;
        HEAP32[$158>>2] = $260;
        $261 = ($141|0)==(0);
        if ($261) {
         HEAP32[$fit_valueA>>2] = $260;
        }
        $262 = HEAP32[$ly1>>2]|0;
        $263 = (($fit_valueA) + ($i$686<<2)|0);
        HEAP32[$263>>2] = $262;
        $264 = HEAP32[$hy0>>2]|0;
        $265 = (($fit_valueB) + ($i$686<<2)|0);
        HEAP32[$265>>2] = $264;
        $266 = HEAP32[$hy1>>2]|0;
        HEAP32[$163>>2] = $266;
        $267 = ($143|0)==(1);
        if ($267) {
         HEAP32[$131>>2] = $266;
        }
        $268 = $264 & $262;
        $269 = ($268|0)>(-1);
        if ($269) {
         $270 = ($139|0)>(0);
         L96: do {
          if ($270) {
           $j$079$in = $139;
           while(1) {
            $j$079 = (($j$079$in) + -1)|0;
            $272 = (($hineighbor) + ($j$079<<2)|0);
            $273 = HEAP32[$272>>2]|0;
            $274 = ($273|0)==($143|0);
            if (!($274)) {
             break L96;
            }
            HEAP32[$272>>2] = $i$686;
            $275 = ($j$079$in|0)>(1);
            if ($275) {
             $j$079$in = $j$079;
            } else {
             break;
            }
           }
          }
         } while(0);
         $j$182 = (($139) + 1)|0;
         $271 = ($j$182|0)<($5|0);
         if ($271) {
          $j$183 = $j$182;
          while(1) {
           $276 = (($loneighbor) + ($j$183<<2)|0);
           $277 = HEAP32[$276>>2]|0;
           $278 = ($277|0)==($141|0);
           if (!($278)) {
            break L87;
           }
           HEAP32[$276>>2] = $i$686;
           $j$1 = (($j$183) + 1)|0;
           $279 = ($j$1|0)<($5|0);
           if ($279) {
            $j$183 = $j$1;
           } else {
            break;
           }
          }
         }
        }
       }
      } while(0);
     }
    } while(0);
    $282 = (($i$686) + 1)|0;
    $283 = ($282|0)<($5|0);
    if ($283) {
     $i$686 = $282;
    } else {
     label = 69;
     break;
    }
   }
   if ((label|0) == 38) {
    _exit(1);
    // unreachable;
   }
   else if ((label|0) == 69) {
    $$pre118 = HEAP32[$fit_valueA>>2]|0;
    $$pre119 = HEAP32[$fit_valueB>>2]|0;
    $287 = $$pre118;$289 = $$pre119;
    break;
   }
  } else {
   $287 = $129;$289 = $129;
  }
 } while(0);
 $284 = $5 << 2;
 $285 = (__vorbis_block_alloc($vb,$284)|0);
 $286 = ($287|0)<(0);
 if ($286) {
  $$0$i10 = $289;
 } else {
  $288 = ($289|0)<(0);
  if ($288) {
   $$0$i10 = $287;
  } else {
   $290 = (($289) + ($287))|0;
   $291 = $290 >> 1;
   $$0$i10 = $291;
  }
 }
 HEAP32[$285>>2] = $$0$i10;
 $292 = HEAP32[$132>>2]|0;
 $293 = ($292|0)<(0);
 $294 = HEAP32[$131>>2]|0;
 if ($293) {
  $$0$i8 = $294;
 } else {
  $295 = ($294|0)<(0);
  if ($295) {
   $$0$i8 = $292;
  } else {
   $296 = (($294) + ($292))|0;
   $297 = $296 >> 1;
   $$0$i8 = $297;
  }
 }
 $298 = ((($285)) + 4|0);
 HEAP32[$298>>2] = $$0$i8;
 if ($133) {
  $i$777 = 2;
  while(1) {
   $299 = (($i$777) + -2)|0;
   $300 = (((($look)) + 1032|0) + ($299<<2)|0);
   $301 = HEAP32[$300>>2]|0;
   $302 = (((($look)) + 780|0) + ($299<<2)|0);
   $303 = HEAP32[$302>>2]|0;
   $304 = (((($1)) + 836|0) + ($301<<2)|0);
   $305 = HEAP32[$304>>2]|0;
   $306 = (((($1)) + 836|0) + ($303<<2)|0);
   $307 = HEAP32[$306>>2]|0;
   $308 = (($285) + ($301<<2)|0);
   $309 = HEAP32[$308>>2]|0;
   $310 = (($285) + ($303<<2)|0);
   $311 = HEAP32[$310>>2]|0;
   $312 = (((($1)) + 836|0) + ($i$777<<2)|0);
   $313 = HEAP32[$312>>2]|0;
   $314 = $309 & 32767;
   $315 = $311 & 32767;
   $316 = (($315) - ($314))|0;
   $317 = (($307) - ($305))|0;
   $ispos$i = ($316|0)>(-1);
   $neg$i = (0 - ($316))|0;
   $318 = $ispos$i ? $316 : $neg$i;
   $319 = (($313) - ($305))|0;
   $320 = Math_imul($318, $319)|0;
   $321 = (($320|0) / ($317|0))&-1;
   $322 = ($316|0)<(0);
   $323 = (0 - ($321))|0;
   $$0$p$i = $322 ? $323 : $321;
   $$0$i7 = (($$0$p$i) + ($314))|0;
   $324 = (($fit_valueA) + ($i$777<<2)|0);
   $325 = HEAP32[$324>>2]|0;
   $326 = ($325|0)<(0);
   $327 = (($fit_valueB) + ($i$777<<2)|0);
   $328 = HEAP32[$327>>2]|0;
   if ($326) {
    $$0$i = $328;
   } else {
    $329 = ($328|0)<(0);
    if ($329) {
     $$0$i = $325;
    } else {
     $330 = (($328) + ($325))|0;
     $331 = $330 >> 1;
     $$0$i = $331;
    }
   }
   $332 = ($$0$i|0)<(0);
   $333 = ($$0$i7|0)==($$0$i|0);
   $or$cond6 = $332 | $333;
   $334 = $$0$i7 | 32768;
   $$$0$i = $or$cond6 ? $334 : $$0$i;
   $335 = (($285) + ($i$777<<2)|0);
   HEAP32[$335>>2] = $$$0$i;
   $336 = (($i$777) + 1)|0;
   $exitcond = ($336|0)==($5|0);
   if ($exitcond) {
    break;
   } else {
    $i$777 = $336;
   }
  }
 }
 $output$0 = $285;
 STACKTOP = sp;return ($output$0|0);
}
function _fit_line($a,$fits,$y0,$y1,$info) {
 $a = $a|0;
 $fits = $fits|0;
 $y0 = $y0|0;
 $y1 = $y1|0;
 $info = $info|0;
 var $$0 = 0, $$pr = 0, $$pre = 0, $$pre15 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $23 = 0;
 var $24 = 0.0, $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0, $47 = 0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0.0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0.0, $74 = 0, $75 = 0, $76 = 0.0, $77 = 0.0, $78 = 0.0;
 var $79 = 0.0, $8 = 0, $80 = 0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0;
 var $97 = 0.0, $98 = 0.0, $99 = 0.0, $bn$0$lcssa = 0.0, $bn$01 = 0.0, $bn$1 = 0.0, $bn$2 = 0.0, $exitcond = 0, $i$02 = 0, $x2b$0$lcssa = 0.0, $x2b$04 = 0.0, $x2b$1 = 0.0, $x2b$2 = 0.0, $xb$0$lcssa = 0.0, $xb$06 = 0.0, $xb$1 = 0.0, $xb$2 = 0.0, $xyb$0$lcssa = 0.0, $xyb$03 = 0.0, $xyb$1 = 0.0;
 var $xyb$2 = 0.0, $yb$0$lcssa = 0.0, $yb$05 = 0.0, $yb$1 = 0.0, $yb$2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$a>>2]|0;
 $1 = (($fits) + -1)|0;
 $2 = (((($a) + (($1*56)|0)|0)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($fits|0)>(0);
 if ($4) {
  $5 = ((($info)) + 1108|0);
  $6 = +HEAPF32[$5>>2];
  $bn$01 = 0.0;$i$02 = 0;$x2b$04 = 0.0;$xb$06 = 0.0;$xyb$03 = 0.0;$yb$05 = 0.0;
  while(1) {
   $7 = (((($a) + (($i$02*56)|0)|0)) + 52|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = (((($a) + (($i$02*56)|0)|0)) + 28|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = (($10) + ($8))|0;
   $12 = (+($11|0));
   $13 = $6 * $12;
   $14 = (($10) + 1)|0;
   $15 = (+($14|0));
   $16 = $13 / $15;
   $17 = $16;
   $18 = $17 + 1.0;
   $19 = (((($a) + (($i$02*56)|0)|0)) + 32|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = (+($20|0));
   $22 = (((($a) + (($i$02*56)|0)|0)) + 8|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = (+($23|0));
   $25 = $24 * $18;
   $26 = $21 + $25;
   $27 = $xb$06 + $26;
   $28 = (((($a) + (($i$02*56)|0)|0)) + 36|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = (+($29|0));
   $31 = (((($a) + (($i$02*56)|0)|0)) + 12|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = (+($32|0));
   $34 = $33 * $18;
   $35 = $30 + $34;
   $36 = $yb$05 + $35;
   $37 = (((($a) + (($i$02*56)|0)|0)) + 40|0);
   $38 = HEAP32[$37>>2]|0;
   $39 = (+($38|0));
   $40 = (((($a) + (($i$02*56)|0)|0)) + 16|0);
   $41 = HEAP32[$40>>2]|0;
   $42 = (+($41|0));
   $43 = $18 * $42;
   $44 = $39 + $43;
   $45 = $x2b$04 + $44;
   $46 = (((($a) + (($i$02*56)|0)|0)) + 48|0);
   $47 = HEAP32[$46>>2]|0;
   $48 = (+($47|0));
   $49 = (((($a) + (($i$02*56)|0)|0)) + 24|0);
   $50 = HEAP32[$49>>2]|0;
   $51 = (+($50|0));
   $52 = $18 * $51;
   $53 = $48 + $52;
   $54 = $xyb$03 + $53;
   $55 = (+($8|0));
   $56 = (+($10|0));
   $57 = $56 * $18;
   $58 = $55 + $57;
   $59 = $bn$01 + $58;
   $60 = (($i$02) + 1)|0;
   $exitcond = ($60|0)==($fits|0);
   if ($exitcond) {
    $bn$0$lcssa = $59;$x2b$0$lcssa = $45;$xb$0$lcssa = $27;$xyb$0$lcssa = $54;$yb$0$lcssa = $36;
    break;
   } else {
    $bn$01 = $59;$i$02 = $60;$x2b$04 = $45;$xb$06 = $27;$xyb$03 = $54;$yb$05 = $36;
   }
  }
 } else {
  $bn$0$lcssa = 0.0;$x2b$0$lcssa = 0.0;$xb$0$lcssa = 0.0;$xyb$0$lcssa = 0.0;$yb$0$lcssa = 0.0;
 }
 $61 = HEAP32[$y0>>2]|0;
 $62 = ($61|0)>(-1);
 if ($62) {
  $63 = (+($0|0));
  $64 = $63 + $xb$0$lcssa;
  $65 = (+($61|0));
  $66 = $yb$0$lcssa + $65;
  $67 = Math_imul($0, $0)|0;
  $68 = (+($67|0));
  $69 = $68 + $x2b$0$lcssa;
  $70 = Math_imul($61, $0)|0;
  $71 = (+($70|0));
  $72 = $xyb$0$lcssa + $71;
  $73 = $bn$0$lcssa + 1.0;
  $bn$1 = $73;$x2b$1 = $69;$xb$1 = $64;$xyb$1 = $72;$yb$1 = $66;
 } else {
  $bn$1 = $bn$0$lcssa;$x2b$1 = $x2b$0$lcssa;$xb$1 = $xb$0$lcssa;$xyb$1 = $xyb$0$lcssa;$yb$1 = $yb$0$lcssa;
 }
 $74 = HEAP32[$y1>>2]|0;
 $75 = ($74|0)>(-1);
 if ($75) {
  $76 = (+($3|0));
  $77 = $76 + $xb$1;
  $78 = (+($74|0));
  $79 = $yb$1 + $78;
  $80 = Math_imul($3, $3)|0;
  $81 = (+($80|0));
  $82 = $81 + $x2b$1;
  $83 = Math_imul($74, $3)|0;
  $84 = (+($83|0));
  $85 = $xyb$1 + $84;
  $86 = $bn$1 + 1.0;
  $bn$2 = $86;$x2b$2 = $82;$xb$2 = $77;$xyb$2 = $85;$yb$2 = $79;
 } else {
  $bn$2 = $bn$1;$x2b$2 = $x2b$1;$xb$2 = $xb$1;$xyb$2 = $xyb$1;$yb$2 = $yb$1;
 }
 $87 = $bn$2 * $x2b$2;
 $88 = $xb$2 * $xb$2;
 $89 = $87 - $88;
 $90 = $89 > 0.0;
 if (!($90)) {
  HEAP32[$y0>>2] = 0;
  HEAP32[$y1>>2] = 0;
  $$0 = 1;
  return ($$0|0);
 }
 $91 = $x2b$2 * $yb$2;
 $92 = $xyb$2 * $xb$2;
 $93 = $91 - $92;
 $94 = $93 / $89;
 $95 = $bn$2 * $xyb$2;
 $96 = $yb$2 * $xb$2;
 $97 = $95 - $96;
 $98 = $97 / $89;
 $99 = (+($0|0));
 $100 = $99 * $98;
 $101 = $94 + $100;
 $102 = (+_rint($101));
 $103 = (~~(($102)));
 HEAP32[$y0>>2] = $103;
 $104 = (+($3|0));
 $105 = $104 * $98;
 $106 = $94 + $105;
 $107 = (+_rint($106));
 $108 = (~~(($107)));
 HEAP32[$y1>>2] = $108;
 $109 = HEAP32[$y0>>2]|0;
 $110 = ($109|0)>(1023);
 if ($110) {
  HEAP32[$y0>>2] = 1023;
  $$pr = HEAP32[$y1>>2]|0;
  $112 = $$pr;$117 = 1023;
 } else {
  $112 = $108;$117 = $109;
 }
 $111 = ($112|0)>(1023);
 if ($111) {
  HEAP32[$y1>>2] = 1023;
  $$pre = HEAP32[$y0>>2]|0;
  $114 = $$pre;$118 = 1023;
 } else {
  $114 = $117;$118 = $112;
 }
 $113 = ($114|0)<(0);
 if ($113) {
  HEAP32[$y0>>2] = 0;
  $$pre15 = HEAP32[$y1>>2]|0;
  $116 = $$pre15;
 } else {
  $116 = $118;
 }
 $115 = ($116|0)<(0);
 if (!($115)) {
  $$0 = 0;
  return ($$0|0);
 }
 HEAP32[$y1>>2] = 0;
 $$0 = 0;
 return ($$0|0);
}
function _vorbis_bitrate_managed($vb) {
 $vb = $vb|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($vb)) + 64|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 104|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 80|0);
 $5 = HEAP32[$4>>2]|0;
 $not$ = ($5|0)!=(0);
 $$0 = $not$&1;
 return ($$0|0);
}
function _floor1_interpolate_fit($vb,$look,$A,$B,$del) {
 $vb = $vb|0;
 $look = $look|0;
 $A = $A|0;
 $B = $B|0;
 $del = $del|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$02 = 0, $or$cond = 0, $output$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($look)) + 1284|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($A|0)!=(0|0);
 $3 = ($B|0)!=(0|0);
 $or$cond = $2 & $3;
 if (!($or$cond)) {
  $output$0 = 0;
  return ($output$0|0);
 }
 $4 = $1 << 2;
 $5 = (__vorbis_block_alloc($vb,$4)|0);
 $6 = ($1|0)>(0);
 if (!($6)) {
  $output$0 = $5;
  return ($output$0|0);
 }
 $7 = (65536 - ($del))|0;
 $i$02 = 0;
 while(1) {
  $8 = (($A) + ($i$02<<2)|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = $9 & 32767;
  $11 = Math_imul($10, $7)|0;
  $12 = (($B) + ($i$02<<2)|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = $13 & 32767;
  $15 = Math_imul($14, $del)|0;
  $16 = (($11) + 32768)|0;
  $17 = (($16) + ($15))|0;
  $18 = $17 >> 16;
  $19 = (($5) + ($i$02<<2)|0);
  HEAP32[$19>>2] = $18;
  $20 = HEAP32[$8>>2]|0;
  $21 = $20 & 32768;
  $22 = ($21|0)==(0);
  if (!($22)) {
   $23 = HEAP32[$12>>2]|0;
   $24 = $23 & 32768;
   $25 = ($24|0)==(0);
   if (!($25)) {
    $26 = $18 | 32768;
    HEAP32[$19>>2] = $26;
   }
  }
  $27 = (($i$02) + 1)|0;
  $exitcond = ($27|0)==($1|0);
  if ($exitcond) {
   $output$0 = $5;
   break;
  } else {
   $i$02 = $27;
  }
 }
 return ($output$0|0);
}
function _floor1_encode($opb,$vb,$look,$post,$ilogmask) {
 $opb = $opb|0;
 $vb = $vb|0;
 $look = $look|0;
 $post = $post|0;
 $ilogmask = $ilogmask|0;
 var $$ = 0, $$0 = 0, $$0$i = 0, $$0$p$i = 0, $$phi$trans$insert = 0, $$pn$i = 0, $$pre = 0, $$pre50 = 0, $$pre51 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0;
 var $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0;
 var $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bookas = 0;
 var $cshift$022 = 0, $cshift$022$us = 0, $cval$0$lcssa = 0, $cval$021 = 0, $cval$021$us = 0, $err$04$i = 0, $err$1$i = 0, $exitcond = 0, $exitcond46 = 0, $exitcond47 = 0, $exitcond48 = 0, $exitcond49 = 0, $hx$0$lcssa = 0, $hx$012 = 0, $hx$1 = 0, $i$040 = 0, $i$135 = 0, $i$231 = 0, $ispos$i = 0, $ispos$i2 = 0;
 var $ispos1$i = 0, $j$030 = 0, $j$111 = 0, $j$210 = 0, $k$017 = 0, $k$123 = 0, $k$123$us = 0, $k$227 = 0, $l$019$us = 0, $l$019$us$lcssa = 0, $lx$013 = 0, $lx$1 = 0, $ly$0$lcssa = 0, $ly$014 = 0, $ly$1 = 0, $maxval = 0, $neg$i = 0, $neg$i3 = 0, $neg2$i = 0, $or$cond = 0;
 var $out = 0, $val$0 = 0, $val1$0 = 0, $x1$n$i = 0, $y$03$i = 0, $y$1$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 336|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $out = sp + 64|0;
 $bookas = sp + 32|0;
 $maxval = sp;
 $0 = ((($look)) + 1296|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($look)) + 1284|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($vb)) + 64|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($7)) + 28|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 2848|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($post|0)==(0|0);
 if ($12) {
  _oggpack_write($opb,0,1);
  $225 = ((($vb)) + 36|0);
  $226 = HEAP32[$225>>2]|0;
  $227 = (($226|0) / 2)&-1;
  $228 = $227 << 2;
  _memset(($ilogmask|0),0,($228|0))|0;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $13 = ($3|0)>(0);
 L5: do {
  if ($13) {
   $14 = ((($1)) + 832|0);
   $i$040 = 0;
   while(1) {
    $15 = (($post) + ($i$040<<2)|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = $16 & 32767;
    $18 = HEAP32[$14>>2]|0;
    switch ($18|0) {
    case 1:  {
     $19 = $17 >>> 2;
     $val$0 = $19;
     break;
    }
    case 2:  {
     $20 = $17 >>> 3;
     $val$0 = $20;
     break;
    }
    case 3:  {
     $21 = (($17>>>0) / 12)&-1;
     $val$0 = $21;
     break;
    }
    case 4:  {
     $22 = $17 >>> 4;
     $val$0 = $22;
     break;
    }
    default: {
     $val$0 = $17;
    }
    }
    $23 = $16 & 32768;
    $24 = $23 | $val$0;
    HEAP32[$15>>2] = $24;
    $25 = (($i$040) + 1)|0;
    $exitcond49 = ($25|0)==($3|0);
    if ($exitcond49) {
     break L5;
    } else {
     $i$040 = $25;
    }
   }
  }
 } while(0);
 $26 = HEAP32[$post>>2]|0;
 HEAP32[$out>>2] = $26;
 $27 = ((($post)) + 4|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ((($out)) + 4|0);
 HEAP32[$29>>2] = $28;
 $30 = ($3|0)>(2);
 $31 = ((($look)) + 1292|0);
 if ($30) {
  $i$135 = 2;
  while(1) {
   $32 = (($i$135) + -2)|0;
   $33 = (((($look)) + 1032|0) + ($32<<2)|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = (((($look)) + 780|0) + ($32<<2)|0);
   $36 = HEAP32[$35>>2]|0;
   $37 = (((($1)) + 836|0) + ($34<<2)|0);
   $38 = HEAP32[$37>>2]|0;
   $39 = (((($1)) + 836|0) + ($36<<2)|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = (($post) + ($34<<2)|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = (($post) + ($36<<2)|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = (((($1)) + 836|0) + ($i$135<<2)|0);
   $46 = HEAP32[$45>>2]|0;
   $47 = $42 & 32767;
   $48 = $44 & 32767;
   $49 = (($48) - ($47))|0;
   $50 = (($40) - ($38))|0;
   $ispos$i = ($49|0)>(-1);
   $neg$i = (0 - ($49))|0;
   $51 = $ispos$i ? $49 : $neg$i;
   $52 = (($46) - ($38))|0;
   $53 = Math_imul($51, $52)|0;
   $54 = (($53|0) / ($50|0))&-1;
   $55 = ($49|0)<(0);
   $56 = (0 - ($54))|0;
   $$0$p$i = $55 ? $56 : $54;
   $$0$i = (($$0$p$i) + ($47))|0;
   $57 = (($post) + ($i$135<<2)|0);
   $58 = HEAP32[$57>>2]|0;
   $59 = $58 & 32768;
   $60 = ($59|0)!=(0);
   $61 = ($58|0)==($$0$i|0);
   $or$cond = $60 | $61;
   if ($or$cond) {
    $62 = $$0$i | 32768;
    HEAP32[$57>>2] = $62;
    $63 = (($out) + ($i$135<<2)|0);
    HEAP32[$63>>2] = 0;
   } else {
    $64 = HEAP32[$31>>2]|0;
    $65 = (($64) - ($$0$i))|0;
    $66 = ($65|0)<($$0$i|0);
    $$ = $66 ? $65 : $$0$i;
    $67 = (($58) - ($$0$i))|0;
    $68 = ($67|0)<(0);
    do {
     if ($68) {
      $69 = (0 - ($$))|0;
      $70 = ($67|0)<($69|0);
      if ($70) {
       $71 = $67 ^ -1;
       $72 = (($$) + ($71))|0;
       $val1$0 = $72;
       break;
      } else {
       $73 = $67 << 1;
       $74 = $73 ^ -1;
       $val1$0 = $74;
       break;
      }
     } else {
      $75 = ($$|0)>($67|0);
      if ($75) {
       $77 = $67 << 1;
       $val1$0 = $77;
       break;
      } else {
       $76 = (($$) + ($67))|0;
       $val1$0 = $76;
       break;
      }
     }
    } while(0);
    $78 = (($out) + ($i$135<<2)|0);
    HEAP32[$78>>2] = $val1$0;
    HEAP32[$41>>2] = $47;
    $79 = HEAP32[$43>>2]|0;
    $80 = $79 & 32767;
    HEAP32[$43>>2] = $80;
   }
   $81 = (($i$135) + 1)|0;
   $exitcond48 = ($81|0)==($3|0);
   if ($exitcond48) {
    break;
   } else {
    $i$135 = $81;
   }
  }
  $$pre = HEAP32[$out>>2]|0;
  $$pre50 = HEAP32[$29>>2]|0;
  $95 = $$pre;$99 = $$pre50;
 } else {
  $95 = $26;$99 = $28;
 }
 _oggpack_write($opb,1,1);
 $82 = ((($look)) + 1308|0);
 $83 = HEAP32[$82>>2]|0;
 $84 = (($83) + 1)|0;
 HEAP32[$82>>2] = $84;
 $85 = HEAP32[$31>>2]|0;
 $86 = (($85) + -1)|0;
 $87 = (_ov_ilog($86)|0);
 $88 = $87 << 1;
 $89 = ((($look)) + 1304|0);
 $90 = HEAP32[$89>>2]|0;
 $91 = (($90) + ($88))|0;
 HEAP32[$89>>2] = $91;
 $92 = HEAP32[$31>>2]|0;
 $93 = (($92) + -1)|0;
 $94 = (_ov_ilog($93)|0);
 _oggpack_write($opb,$95,$94);
 $96 = HEAP32[$31>>2]|0;
 $97 = (($96) + -1)|0;
 $98 = (_ov_ilog($97)|0);
 _oggpack_write($opb,$99,$98);
 $100 = HEAP32[$1>>2]|0;
 $101 = ($100|0)>(0);
 if ($101) {
  $102 = ((($look)) + 1300|0);
  $i$231 = 0;$j$030 = 2;
  while(1) {
   $103 = (((($1)) + 4|0) + ($i$231<<2)|0);
   $104 = HEAP32[$103>>2]|0;
   $105 = (((($1)) + 128|0) + ($104<<2)|0);
   $106 = HEAP32[$105>>2]|0;
   $107 = (((($1)) + 192|0) + ($104<<2)|0);
   $108 = HEAP32[$107>>2]|0;
   $109 = 1 << $108;
   ;HEAP32[$bookas>>2]=0|0;HEAP32[$bookas+4>>2]=0|0;HEAP32[$bookas+8>>2]=0|0;HEAP32[$bookas+12>>2]=0|0;HEAP32[$bookas+16>>2]=0|0;HEAP32[$bookas+20>>2]=0|0;HEAP32[$bookas+24>>2]=0|0;HEAP32[$bookas+28>>2]=0|0;
   $110 = ($108|0)==(0);
   if (!($110)) {
    ;HEAP32[$maxval>>2]=0|0;HEAP32[$maxval+4>>2]=0|0;HEAP32[$maxval+8>>2]=0|0;HEAP32[$maxval+12>>2]=0|0;HEAP32[$maxval+16>>2]=0|0;HEAP32[$maxval+20>>2]=0|0;HEAP32[$maxval+24>>2]=0|0;HEAP32[$maxval+28>>2]=0|0;
    $111 = ($108|0)==(31);
    if (!($111)) {
     $k$017 = 0;
     while(1) {
      $127 = ((((($1)) + 320|0) + ($104<<5)|0) + ($k$017<<2)|0);
      $128 = HEAP32[$127>>2]|0;
      $129 = ($128|0)<(0);
      if ($129) {
       $$sink = 1;
      } else {
       $130 = (((($9)) + 1824|0) + ($128<<2)|0);
       $131 = HEAP32[$130>>2]|0;
       $132 = ((($131)) + 4|0);
       $133 = HEAP32[$132>>2]|0;
       $$sink = $133;
      }
      $134 = (($maxval) + ($k$017<<2)|0);
      HEAP32[$134>>2] = $$sink;
      $135 = (($k$017) + 1)|0;
      $136 = ($135|0)<($109|0);
      if ($136) {
       $k$017 = $135;
      } else {
       break;
      }
     }
    }
    $112 = ($106|0)>(0);
    L48: do {
     if ($112) {
      if ($111) {
       $cshift$022 = 0;$cval$021 = 0;$k$123 = 0;
       while(1) {
        $137 = (($bookas) + ($k$123<<2)|0);
        $138 = HEAP32[$137>>2]|0;
        $139 = $138 << $cshift$022;
        $140 = $139 | $cval$021;
        $141 = (($cshift$022) + 31)|0;
        $142 = (($k$123) + 1)|0;
        $exitcond46 = ($142|0)==($106|0);
        if ($exitcond46) {
         $cval$0$lcssa = $140;
         break L48;
        } else {
         $cshift$022 = $141;$cval$021 = $140;$k$123 = $142;
        }
       }
      } else {
       $cshift$022$us = 0;$cval$021$us = 0;$k$123$us = 0;
      }
      while(1) {
       $125 = (($k$123$us) + ($j$030))|0;
       $126 = (($out) + ($125<<2)|0);
       $116 = HEAP32[$126>>2]|0;
       $l$019$us = 0;
       while(1) {
        $113 = (($maxval) + ($l$019$us<<2)|0);
        $114 = HEAP32[$113>>2]|0;
        $115 = ($116|0)<($114|0);
        if ($115) {
         $l$019$us$lcssa = $l$019$us;
         label = 31;
         break;
        }
        $117 = (($l$019$us) + 1)|0;
        $118 = ($117|0)<($109|0);
        if ($118) {
         $l$019$us = $117;
        } else {
         label = 32;
         break;
        }
       }
       if ((label|0) == 31) {
        label = 0;
        $119 = (($bookas) + ($k$123$us<<2)|0);
        HEAP32[$119>>2] = $l$019$us$lcssa;
        $121 = $l$019$us$lcssa;
       }
       else if ((label|0) == 32) {
        label = 0;
        $$phi$trans$insert = (($bookas) + ($k$123$us<<2)|0);
        $$pre51 = HEAP32[$$phi$trans$insert>>2]|0;
        $121 = $$pre51;
       }
       $120 = $121 << $cshift$022$us;
       $122 = $120 | $cval$021$us;
       $123 = (($cshift$022$us) + ($108))|0;
       $124 = (($k$123$us) + 1)|0;
       $exitcond = ($124|0)==($106|0);
       if ($exitcond) {
        $cval$0$lcssa = $122;
        break;
       } else {
        $cshift$022$us = $123;$cval$021$us = $122;$k$123$us = $124;
       }
      }
     } else {
      $cval$0$lcssa = 0;
     }
    } while(0);
    $143 = (((($1)) + 256|0) + ($104<<2)|0);
    $144 = HEAP32[$143>>2]|0;
    $145 = (($11) + (($144*56)|0)|0);
    $146 = (_vorbis_book_encode($145,$cval$0$lcssa,$opb)|0);
    $147 = HEAP32[$102>>2]|0;
    $148 = (($147) + ($146))|0;
    HEAP32[$102>>2] = $148;
   }
   $149 = ($106|0)>(0);
   if ($149) {
    $k$227 = 0;
    while(1) {
     $150 = (($bookas) + ($k$227<<2)|0);
     $151 = HEAP32[$150>>2]|0;
     $152 = ((((($1)) + 320|0) + ($104<<5)|0) + ($151<<2)|0);
     $153 = HEAP32[$152>>2]|0;
     $154 = ($153|0)>(-1);
     if ($154) {
      $155 = (($k$227) + ($j$030))|0;
      $156 = (($out) + ($155<<2)|0);
      $157 = HEAP32[$156>>2]|0;
      $158 = (((($11) + (($153*56)|0)|0)) + 4|0);
      $159 = HEAP32[$158>>2]|0;
      $160 = ($157|0)<($159|0);
      if ($160) {
       $161 = (($11) + (($153*56)|0)|0);
       $162 = (_vorbis_book_encode($161,$157,$opb)|0);
       $163 = HEAP32[$89>>2]|0;
       $164 = (($163) + ($162))|0;
       HEAP32[$89>>2] = $164;
      }
     }
     $165 = (($k$227) + 1)|0;
     $exitcond47 = ($165|0)==($106|0);
     if ($exitcond47) {
      break;
     } else {
      $k$227 = $165;
     }
    }
   }
   $166 = (($106) + ($j$030))|0;
   $167 = (($i$231) + 1)|0;
   $168 = HEAP32[$1>>2]|0;
   $169 = ($167|0)<($168|0);
   if ($169) {
    $i$231 = $167;$j$030 = $166;
   } else {
    break;
   }
  }
 }
 $170 = HEAP32[$post>>2]|0;
 $171 = ((($1)) + 832|0);
 $172 = HEAP32[$171>>2]|0;
 $173 = Math_imul($172, $170)|0;
 $174 = ((($vb)) + 28|0);
 $175 = HEAP32[$174>>2]|0;
 $176 = (($9) + ($175<<2)|0);
 $177 = HEAP32[$176>>2]|0;
 $178 = (($177|0) / 2)&-1;
 $179 = HEAP32[$2>>2]|0;
 $180 = ($179|0)>(1);
 if ($180) {
  $hx$012 = 0;$j$111 = 1;$lx$013 = 0;$ly$014 = $173;
  while(1) {
   $185 = (((($look)) + 260|0) + ($j$111<<2)|0);
   $186 = HEAP32[$185>>2]|0;
   $187 = (($post) + ($186<<2)|0);
   $188 = HEAP32[$187>>2]|0;
   $189 = $188 & 32767;
   $190 = ($189|0)==($188|0);
   if ($190) {
    $191 = HEAP32[$171>>2]|0;
    $192 = Math_imul($191, $188)|0;
    $193 = (((($1)) + 836|0) + ($186<<2)|0);
    $194 = HEAP32[$193>>2]|0;
    $195 = (($192) - ($ly$014))|0;
    $196 = (($194) - ($lx$013))|0;
    $ispos$i2 = ($195|0)>(-1);
    $neg$i3 = (0 - ($195))|0;
    $197 = $ispos$i2 ? $195 : $neg$i3;
    $198 = (($195|0) / ($196|0))&-1;
    $199 = $195 >> 31;
    $200 = $199 | 1;
    $201 = Math_imul($198, $196)|0;
    $ispos1$i = ($201|0)>(-1);
    $neg2$i = (0 - ($201))|0;
    $202 = $ispos1$i ? $201 : $neg2$i;
    $203 = (($197) - ($202))|0;
    $204 = ($178|0)>($194|0);
    $x1$n$i = $204 ? $194 : $178;
    $205 = ($x1$n$i|0)>($lx$013|0);
    if ($205) {
     $206 = (($ilogmask) + ($lx$013<<2)|0);
     HEAP32[$206>>2] = $ly$014;
    }
    $207 = (($lx$013) + 1)|0;
    $208 = ($207|0)<($x1$n$i|0);
    if ($208) {
     $214 = $207;$err$04$i = 0;$y$03$i = $ly$014;
     while(1) {
      $209 = (($err$04$i) + ($203))|0;
      $210 = ($209|0)<($196|0);
      $211 = $210 ? 0 : $200;
      $212 = $210 ? 0 : $196;
      $err$1$i = (($209) - ($212))|0;
      $$pn$i = (($y$03$i) + ($198))|0;
      $y$1$i = (($$pn$i) + ($211))|0;
      $213 = (($ilogmask) + ($214<<2)|0);
      HEAP32[$213>>2] = $y$1$i;
      $215 = (($214) + 1)|0;
      $216 = ($215|0)<($x1$n$i|0);
      if ($216) {
       $214 = $215;$err$04$i = $err$1$i;$y$03$i = $y$1$i;
      } else {
       $hx$1 = $194;$lx$1 = $194;$ly$1 = $192;
       break;
      }
     }
    } else {
     $hx$1 = $194;$lx$1 = $194;$ly$1 = $192;
    }
   } else {
    $hx$1 = $hx$012;$lx$1 = $lx$013;$ly$1 = $ly$014;
   }
   $217 = (($j$111) + 1)|0;
   $218 = HEAP32[$2>>2]|0;
   $219 = ($217|0)<($218|0);
   if ($219) {
    $hx$012 = $hx$1;$j$111 = $217;$lx$013 = $lx$1;$ly$014 = $ly$1;
   } else {
    $hx$0$lcssa = $hx$1;$ly$0$lcssa = $ly$1;
    break;
   }
  }
 } else {
  $hx$0$lcssa = 0;$ly$0$lcssa = $173;
 }
 $181 = ((($vb)) + 36|0);
 $182 = HEAP32[$181>>2]|0;
 $183 = (($182|0) / 2)&-1;
 $184 = ($hx$0$lcssa|0)<($183|0);
 if ($184) {
  $j$210 = $hx$0$lcssa;
 } else {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 while(1) {
  $220 = (($ilogmask) + ($j$210<<2)|0);
  HEAP32[$220>>2] = $ly$0$lcssa;
  $221 = (($j$210) + 1)|0;
  $222 = HEAP32[$181>>2]|0;
  $223 = (($222|0) / 2)&-1;
  $224 = ($221|0)<($223|0);
  if ($224) {
   $j$210 = $221;
  } else {
   $$0 = 1;
   break;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _vorbis_book_encode($book,$a,$b) {
 $book = $book|0;
 $a = $a|0;
 $b = $b|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($a|0)<(0);
 if ($0) {
  $$0 = 0;
  return ($$0|0);
 }
 $1 = ((($book)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)>($a|0);
 if (!($5)) {
  $$0 = 0;
  return ($$0|0);
 }
 $6 = ((($book)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (($7) + ($a<<2)|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($2)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + ($a)|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = $13 << 24 >> 24;
 _oggpack_write($b,$9,$14);
 $15 = HEAP32[$1>>2]|0;
 $16 = ((($15)) + 8|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = (($17) + ($a)|0);
 $19 = HEAP8[$18>>0]|0;
 $20 = $19 << 24 >> 24;
 $$0 = $20;
 return ($$0|0);
}
function __vp_couple_quantize_normalize($blobno,$g,$p,$vi,$mdct,$iwork,$nonzero,$sliding_lowpass,$ch) {
 $blobno = $blobno|0;
 $g = $g|0;
 $p = $p|0;
 $vi = $vi|0;
 $mdct = $mdct|0;
 $iwork = $iwork|0;
 $nonzero = $nonzero|0;
 $sliding_lowpass = $sliding_lowpass|0;
 $ch = $ch|0;
 var $$ = 0, $$$i = 0, $$alloca_mul = 0, $$alloca_mul1 = 0, $$alloca_mul11 = 0, $$alloca_mul13 = 0, $$alloca_mul15 = 0, $$alloca_mul3 = 0, $$alloca_mul5 = 0, $$alloca_mul7 = 0, $$alloca_mul9 = 0, $$phi$trans$insert50 = 0, $$phi$trans$insert52 = 0, $$pre = 0, $$pre$phiZ2D = 0, $$pre46 = 0, $$pre47 = 0, $$pre48 = 0, $$pre51 = 0, $$pre53 = 0;
 var $$pre54 = 0, $$pre55 = 0, $$pre56 = 0, $$pre57 = 0, $$v = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0.0;
 var $111 = 0.0, $112 = 0, $113 = 0, $114 = 0.0, $115 = 0, $116 = 0.0, $117 = 0.0, $118 = 0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0.0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0.0, $174 = 0, $175 = 0.0, $176 = 0.0, $177 = 0, $178 = 0.0, $179 = 0, $18 = 0, $180 = 0.0, $181 = 0.0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0;
 var $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0.0, $208 = 0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0, $213 = 0.0, $214 = 0, $215 = 0.0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0;
 var $22 = 0, $220 = 0.0, $221 = 0, $222 = 0, $223 = 0, $224 = 0.0, $225 = 0, $226 = 0.0, $227 = 0.0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0;
 var $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0.0, $96 = 0, $97 = 0.0, $98 = 0, $99 = 0.0, $exitcond = 0, $exitcond$i = 0;
 var $exitcond44 = 0, $exitcond4472 = 0, $fabsf = 0.0, $fabsf$i = 0.0, $fabsf1 = 0.0, $fabsf2 = 0.0, $fabsf3 = 0.0, $fabsf4 = 0.0, $i$239 = 0, $i$315 = 0, $ispos = 0, $ispos5 = 0, $ispos7 = 0, $j$01$i = 0, $j$017 = 0, $j$120 = 0, $j$222 = 0, $j$330 = 0, $k$025 = 0, $neg = 0;
 var $neg6 = 0, $neg8 = 0, $not$$i = 0, $p$idx$val = 0, $p$idx9$val = 0, $postpoint$0$in = 0.0, $postpoint$0$in$in = 0, $step$036 = 0, $stereo_threshholds_limited$stereo_threshholds = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$p>>2]|0;
 $1 = ((($p)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 500|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0);
 if ($5) {
  $26 = 16;
 } else {
  $6 = ((($2)) + 508|0);
  $7 = HEAP32[$6>>2]|0;
  $26 = $7;
 }
 $8 = HEAP32[$2>>2]|0;
 $9 = ((((($g)) + 132|0) + (($8*60)|0)|0) + ($blobno<<2)|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (((($g)) + 252|0) + ($blobno<<2)|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (3304 + ($12<<3)|0);
 $14 = +HEAPF64[$13>>3];
 $15 = (((($g)) + 312|0) + ($blobno<<2)|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $ch << 2;
 $$alloca_mul = $17;
 $18 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $$alloca_mul1 = $17;
 $19 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul1)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $$alloca_mul3 = $17;
 $20 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul3)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $$alloca_mul5 = $17;
 $21 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul5)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $$alloca_mul7 = $17;
 $22 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul7)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $23 = ((($vi)) + 1156|0);
 $24 = ($0|0)>(1000);
 $stereo_threshholds_limited$stereo_threshholds = $24 ? 3376 : 3304;
 $postpoint$0$in$in = (($stereo_threshholds_limited$stereo_threshholds) + ($16<<3)|0);
 $postpoint$0$in = +HEAPF64[$postpoint$0$in$in>>3];
 $25 = Math_imul($17, $26)|0;
 $$alloca_mul9 = $25;
 $27 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul9)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 HEAP32[$18>>2] = $27;
 $$alloca_mul11 = $25;
 $28 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul11)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 HEAP32[$19>>2] = $28;
 $$alloca_mul13 = $25;
 $29 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul13)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 HEAP32[$20>>2] = $29;
 $$alloca_mul15 = $25;
 $30 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul15)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 HEAP32[$21>>2] = $30;
 $31 = ($ch|0)>(1);
 if ($31) {
  $32 = (($27) + ($26<<2)|0);
  $33 = ((($18)) + 4|0);
  HEAP32[$33>>2] = $32;
  $34 = (($28) + ($26<<2)|0);
  $35 = ((($19)) + 4|0);
  HEAP32[$35>>2] = $34;
  $36 = (($29) + ($26<<2)|0);
  $37 = ((($20)) + 4|0);
  HEAP32[$37>>2] = $36;
  $38 = (($30) + ($26<<2)|0);
  $39 = ((($21)) + 4|0);
  HEAP32[$39>>2] = $38;
  $exitcond4472 = ($ch|0)==(2);
  if (!($exitcond4472)) {
   $43 = 2;
   while(1) {
    $$pre = HEAP32[$18>>2]|0;
    $$pre46 = HEAP32[$19>>2]|0;
    $$pre47 = HEAP32[$20>>2]|0;
    $$pre48 = HEAP32[$21>>2]|0;
    $42 = Math_imul($43, $26)|0;
    $44 = (($$pre) + ($42<<2)|0);
    $45 = (($18) + ($43<<2)|0);
    HEAP32[$45>>2] = $44;
    $46 = (($$pre46) + ($42<<2)|0);
    $47 = (($19) + ($43<<2)|0);
    HEAP32[$47>>2] = $46;
    $48 = (($$pre47) + ($42<<2)|0);
    $49 = (($20) + ($43<<2)|0);
    HEAP32[$49>>2] = $48;
    $50 = (($$pre48) + ($42<<2)|0);
    $51 = (($21) + ($43<<2)|0);
    HEAP32[$51>>2] = $50;
    $52 = (($43) + 1)|0;
    $exitcond44 = ($52|0)==($ch|0);
    if ($exitcond44) {
     break;
    } else {
     $43 = $52;
    }
   }
  }
 }
 $40 = HEAP32[$23>>2]|0;
 $41 = ($0|0)>(0);
 if ($41) {
  $53 = HEAP32[$21>>2]|0;
  $54 = ($ch|0)>(0);
  $i$239 = 0;
  while(1) {
   $57 = (($0) - ($i$239))|0;
   $58 = ($26|0)>($57|0);
   $$ = $58 ? $57 : $26;
   _memcpy(($22|0),($nonzero|0),($17|0))|0;
   _memset(($53|0),0,($25|0))|0;
   if ($54) {
    $59 = ($$|0)>(0);
    $60 = (($10) - ($i$239))|0;
    $k$025 = 0;
    while(1) {
     $66 = (($iwork) + ($k$025<<2)|0);
     $67 = HEAP32[$66>>2]|0;
     $68 = (($67) + ($i$239<<2)|0);
     $69 = (($22) + ($k$025<<2)|0);
     $70 = HEAP32[$69>>2]|0;
     $71 = ($70|0)==(0);
     if ($71) {
      if ($59) {
       $74 = (($20) + ($k$025<<2)|0);
       $75 = HEAP32[$74>>2]|0;
       $76 = (($18) + ($k$025<<2)|0);
       $77 = HEAP32[$76>>2]|0;
       $78 = (($19) + ($k$025<<2)|0);
       $79 = HEAP32[$78>>2]|0;
       $80 = (($21) + ($k$025<<2)|0);
       $81 = HEAP32[$80>>2]|0;
       $j$222 = 0;
       while(1) {
        $125 = (($75) + ($j$222<<2)|0);
        HEAPF32[$125>>2] = 1.000000013351432E-10;
        $126 = (($77) + ($j$222<<2)|0);
        HEAPF32[$126>>2] = 0.0;
        $127 = (($79) + ($j$222<<2)|0);
        HEAPF32[$127>>2] = 0.0;
        $128 = (($81) + ($j$222<<2)|0);
        HEAP32[$128>>2] = 0;
        $129 = (($68) + ($j$222<<2)|0);
        HEAP32[$129>>2] = 0;
        $130 = (($j$222) + 1)|0;
        $131 = ($130|0)<($$|0);
        if ($131) {
         $j$222 = $130;
        } else {
         break;
        }
       }
      }
     } else {
      $72 = (($20) + ($k$025<<2)|0);
      $73 = HEAP32[$72>>2]|0;
      if ($59) {
       $j$017 = 0;
       while(1) {
        $82 = (($68) + ($j$017<<2)|0);
        $83 = HEAP32[$82>>2]|0;
        $84 = (100240 + ($83<<2)|0);
        $85 = HEAP32[$84>>2]|0;
        $86 = (($73) + ($j$017<<2)|0);
        HEAP32[$86>>2] = $85;
        $87 = (($j$017) + 1)|0;
        $88 = ($87|0)<($$|0);
        if ($88) {
         $j$017 = $87;
        } else {
         break;
        }
       }
       $89 = (($mdct) + ($k$025<<2)|0);
       $90 = HEAP32[$89>>2]|0;
       $91 = (($90) + ($i$239<<2)|0);
       $92 = (($21) + ($k$025<<2)|0);
       $93 = HEAP32[$92>>2]|0;
       if ($59) {
        $j$01$i = 0;
        while(1) {
         $94 = ($j$01$i|0)>=($60|0);
         $$v = $94 ? $postpoint$0$in : $14;
         $95 = $$v;
         $96 = (($91) + ($j$01$i<<2)|0);
         $97 = +HEAPF32[$96>>2];
         $fabsf$i = (+Math_abs((+$97)));
         $98 = (($73) + ($j$01$i<<2)|0);
         $99 = +HEAPF32[$98>>2];
         $100 = $fabsf$i / $99;
         $101 = (($93) + ($j$01$i<<2)|0);
         $not$$i = !($100 < $95);
         $$$i = $not$$i&1;
         HEAP32[$101>>2] = $$$i;
         $102 = (($j$01$i) + 1)|0;
         $exitcond$i = ($102|0)==($$|0);
         if ($exitcond$i) {
          break;
         } else {
          $j$01$i = $102;
         }
        }
        if ($59) {
         $103 = HEAP32[$89>>2]|0;
         $104 = (($18) + ($k$025<<2)|0);
         $105 = HEAP32[$104>>2]|0;
         $106 = (($19) + ($k$025<<2)|0);
         $107 = HEAP32[$106>>2]|0;
         $j$120 = 0;
         while(1) {
          $108 = (($j$120) + ($i$239))|0;
          $109 = (($103) + ($108<<2)|0);
          $110 = +HEAPF32[$109>>2];
          $111 = $110 * $110;
          $112 = (($105) + ($j$120<<2)|0);
          HEAPF32[$112>>2] = $111;
          $113 = (($107) + ($j$120<<2)|0);
          HEAPF32[$113>>2] = $111;
          $114 = +HEAPF32[$109>>2];
          $115 = $114 < 0.0;
          if ($115) {
           $116 = +HEAPF32[$112>>2];
           $117 = -$116;
           HEAPF32[$112>>2] = $117;
          }
          $118 = (($73) + ($j$120<<2)|0);
          $119 = +HEAPF32[$118>>2];
          $120 = $119 * $119;
          HEAPF32[$118>>2] = $120;
          $121 = (($j$120) + 1)|0;
          $122 = ($121|0)<($$|0);
          if ($122) {
           $j$120 = $121;
          } else {
           $123 = $105;$124 = $107;
           break;
          }
         }
        } else {
         label = 21;
        }
       } else {
        label = 21;
       }
      } else {
       label = 21;
      }
      if ((label|0) == 21) {
       label = 0;
       $$phi$trans$insert50 = (($18) + ($k$025<<2)|0);
       $$pre51 = HEAP32[$$phi$trans$insert50>>2]|0;
       $$phi$trans$insert52 = (($19) + ($k$025<<2)|0);
       $$pre53 = HEAP32[$$phi$trans$insert52>>2]|0;
       $123 = $$pre51;$124 = $$pre53;
      }
      $p$idx$val = HEAP32[$1>>2]|0;
      (+_noise_normalize($p$idx$val,$10,$123,$124,$73,0,$i$239,$$,$68));
     }
     $132 = (($k$025) + 1)|0;
     $exitcond = ($132|0)==($ch|0);
     if ($exitcond) {
      break;
     } else {
      $k$025 = $132;
     }
    }
   }
   $61 = HEAP32[$23>>2]|0;
   $62 = ($61|0)>(0);
   if ($62) {
    $63 = ($$|0)>(0);
    $64 = (($sliding_lowpass) - ($i$239))|0;
    $65 = (($10) - ($i$239))|0;
    $252 = $61;$step$036 = 0;
    while(1) {
     $133 = (((($vi)) + 1160|0) + ($step$036<<2)|0);
     $134 = HEAP32[$133>>2]|0;
     $135 = (((($vi)) + 2184|0) + ($step$036<<2)|0);
     $136 = HEAP32[$135>>2]|0;
     $137 = (($iwork) + ($134<<2)|0);
     $138 = HEAP32[$137>>2]|0;
     $139 = (($138) + ($i$239<<2)|0);
     $140 = (($iwork) + ($136<<2)|0);
     $141 = HEAP32[$140>>2]|0;
     $142 = (($141) + ($i$239<<2)|0);
     $143 = (($18) + ($134<<2)|0);
     $144 = HEAP32[$143>>2]|0;
     $145 = (($18) + ($136<<2)|0);
     $146 = HEAP32[$145>>2]|0;
     $147 = (($19) + ($134<<2)|0);
     $148 = HEAP32[$147>>2]|0;
     $149 = (($19) + ($136<<2)|0);
     $150 = HEAP32[$149>>2]|0;
     $151 = (($20) + ($134<<2)|0);
     $152 = HEAP32[$151>>2]|0;
     $153 = (($20) + ($136<<2)|0);
     $154 = HEAP32[$153>>2]|0;
     $155 = (($21) + ($134<<2)|0);
     $156 = HEAP32[$155>>2]|0;
     $157 = (($21) + ($136<<2)|0);
     $158 = HEAP32[$157>>2]|0;
     $159 = (($22) + ($134<<2)|0);
     $160 = HEAP32[$159>>2]|0;
     $161 = ($160|0)==(0);
     $162 = (($22) + ($136<<2)|0);
     if ($161) {
      $163 = HEAP32[$162>>2]|0;
      $164 = ($163|0)==(0);
      if ($164) {
       $232 = $252;
      } else {
       label = 31;
      }
     } else {
      label = 31;
     }
     if ((label|0) == 31) {
      label = 0;
      HEAP32[$162>>2] = 1;
      HEAP32[$159>>2] = 1;
      if ($63) {
       $j$330 = 0;
       while(1) {
        $165 = ($j$330|0)<($64|0);
        do {
         if ($165) {
          $166 = (($156) + ($j$330<<2)|0);
          $167 = HEAP32[$166>>2]|0;
          $168 = ($167|0)==(0);
          $169 = (($158) + ($j$330<<2)|0);
          if ($168) {
           $170 = HEAP32[$169>>2]|0;
           $171 = ($170|0)==(0);
           if ($171) {
            $205 = ($j$330|0)<($65|0);
            do {
             if ($205) {
              $206 = (($146) + ($j$330<<2)|0);
              $207 = +HEAPF32[$206>>2];
              $208 = (($144) + ($j$330<<2)|0);
              $209 = +HEAPF32[$208>>2];
              $210 = $207 + $209;
              HEAPF32[$208>>2] = $210;
              $fabsf2 = (+Math_abs((+$210)));
              $211 = (($148) + ($j$330<<2)|0);
              HEAPF32[$211>>2] = $fabsf2;
              $$pre$phiZ2D = $206;
             } else {
              $212 = (($144) + ($j$330<<2)|0);
              $213 = +HEAPF32[$212>>2];
              $214 = (($146) + ($j$330<<2)|0);
              $215 = +HEAPF32[$214>>2];
              $216 = $213 + $215;
              $217 = $216 < 0.0;
              $fabsf = (+Math_abs((+$213)));
              $fabsf1 = (+Math_abs((+$215)));
              $218 = $fabsf + $fabsf1;
              $219 = (($148) + ($j$330<<2)|0);
              HEAPF32[$219>>2] = $218;
              if ($217) {
               $220 = -$218;
               HEAPF32[$212>>2] = $220;
               $$pre$phiZ2D = $214;
               break;
              } else {
               HEAPF32[$212>>2] = $218;
               $$pre$phiZ2D = $214;
               break;
              }
             }
            } while(0);
            $221 = (($150) + ($j$330<<2)|0);
            HEAPF32[$221>>2] = 0.0;
            HEAPF32[$$pre$phiZ2D>>2] = 0.0;
            HEAP32[$169>>2] = 1;
            $222 = (($142) + ($j$330<<2)|0);
            HEAP32[$222>>2] = 0;
            break;
           }
          }
          $172 = (($144) + ($j$330<<2)|0);
          $173 = +HEAPF32[$172>>2];
          $fabsf3 = (+Math_abs((+$173)));
          $174 = (($146) + ($j$330<<2)|0);
          $175 = +HEAPF32[$174>>2];
          $fabsf4 = (+Math_abs((+$175)));
          $176 = $fabsf3 + $fabsf4;
          HEAPF32[$172>>2] = $176;
          $177 = (($148) + ($j$330<<2)|0);
          $178 = +HEAPF32[$177>>2];
          $179 = (($150) + ($j$330<<2)|0);
          $180 = +HEAPF32[$179>>2];
          $181 = $178 + $180;
          HEAPF32[$177>>2] = $181;
          HEAP32[$169>>2] = 1;
          HEAP32[$166>>2] = 1;
          $182 = (($139) + ($j$330<<2)|0);
          $183 = HEAP32[$182>>2]|0;
          $184 = (($142) + ($j$330<<2)|0);
          $185 = HEAP32[$184>>2]|0;
          $ispos = ($183|0)>(-1);
          $neg = (0 - ($183))|0;
          $186 = $ispos ? $183 : $neg;
          $ispos5 = ($185|0)>(-1);
          $neg6 = (0 - ($185))|0;
          $187 = $ispos5 ? $185 : $neg6;
          $188 = ($186|0)>($187|0);
          if ($188) {
           $189 = ($183|0)>(0);
           $190 = (($183) - ($185))|0;
           $191 = (($185) - ($183))|0;
           $192 = $189 ? $190 : $191;
           HEAP32[$184>>2] = $192;
           $$pre55 = HEAP32[$182>>2]|0;
           $197 = $$pre55;$201 = $192;
          } else {
           $193 = ($185|0)>(0);
           $194 = (($183) - ($185))|0;
           $195 = (($185) - ($183))|0;
           $196 = $193 ? $194 : $195;
           HEAP32[$184>>2] = $196;
           HEAP32[$182>>2] = $185;
           $$pre54 = HEAP32[$184>>2]|0;
           $197 = $185;$201 = $$pre54;
          }
          $ispos7 = ($197|0)>(-1);
          $neg8 = (0 - ($197))|0;
          $198 = $ispos7 ? $197 : $neg8;
          $199 = $198 << 1;
          $200 = ($201|0)<($199|0);
          if (!($200)) {
           $202 = (0 - ($201))|0;
           HEAP32[$184>>2] = $202;
           $203 = HEAP32[$182>>2]|0;
           $204 = (0 - ($203))|0;
           HEAP32[$182>>2] = $204;
          }
         }
        } while(0);
        $223 = (($152) + ($j$330<<2)|0);
        $224 = +HEAPF32[$223>>2];
        $225 = (($154) + ($j$330<<2)|0);
        $226 = +HEAPF32[$225>>2];
        $227 = $224 + $226;
        HEAPF32[$225>>2] = $227;
        HEAPF32[$223>>2] = $227;
        $228 = (($j$330) + 1)|0;
        $229 = ($228|0)<($$|0);
        if ($229) {
         $j$330 = $228;
        } else {
         break;
        }
       }
      }
      $p$idx9$val = HEAP32[$1>>2]|0;
      (+_noise_normalize($p$idx9$val,$10,$144,$148,$152,$156,$i$239,$$,$139));
      $$pre56 = HEAP32[$23>>2]|0;
      $232 = $$pre56;
     }
     $230 = (($step$036) + 1)|0;
     $231 = ($230|0)<($232|0);
     if ($231) {
      $252 = $232;$step$036 = $230;
     } else {
      $251 = $232;
      break;
     }
    }
   } else {
    $251 = $61;
   }
   $233 = (($i$239) + ($26))|0;
   $234 = ($0|0)>($233|0);
   if ($234) {
    $i$239 = $233;
   } else {
    $56 = $251;
    break;
   }
  }
 } else {
  $56 = $40;
 }
 $55 = ($56|0)>(0);
 if ($55) {
  $250 = $56;$i$315 = 0;
 } else {
  STACKTOP = sp;return;
 }
 while(1) {
  $235 = (((($vi)) + 1160|0) + ($i$315<<2)|0);
  $236 = HEAP32[$235>>2]|0;
  $237 = (($nonzero) + ($236<<2)|0);
  $238 = HEAP32[$237>>2]|0;
  $239 = ($238|0)==(0);
  $240 = (((($vi)) + 2184|0) + ($i$315<<2)|0);
  if ($239) {
   $241 = HEAP32[$240>>2]|0;
   $242 = (($nonzero) + ($241<<2)|0);
   $243 = HEAP32[$242>>2]|0;
   $244 = ($243|0)==(0);
   if ($244) {
    $249 = $250;
   } else {
    label = 52;
   }
  } else {
   label = 52;
  }
  if ((label|0) == 52) {
   label = 0;
   HEAP32[$237>>2] = 1;
   $245 = HEAP32[$240>>2]|0;
   $246 = (($nonzero) + ($245<<2)|0);
   HEAP32[$246>>2] = 1;
   $$pre57 = HEAP32[$23>>2]|0;
   $249 = $$pre57;
  }
  $247 = (($i$315) + 1)|0;
  $248 = ($247|0)<($249|0);
  if ($248) {
   $250 = $249;$i$315 = $247;
  } else {
   break;
  }
 }
 STACKTOP = sp;return;
}
function _noise_normalize($p$0$1$val,$limit,$r,$q,$f,$flags,$i,$n,$out) {
 $p$0$1$val = $p$0$1$val|0;
 $limit = $limit|0;
 $r = $r|0;
 $q = $q|0;
 $f = $f|0;
 $flags = $flags|0;
 $i = $i|0;
 $n = $n|0;
 $out = $out|0;
 var $$0$lcssa = 0.0, $$06 = 0.0, $$06$us = 0.0, $$2 = 0.0, $$2$us = 0.0, $$32 = 0.0, $$4 = 0.0, $$5 = 0.0, $$alloca_mul = 0, $$sink = 0, $$sink$in = 0.0, $$sink20 = 0, $$sink20$in = 0.0, $$sink21 = 0, $$sink21$in = 0.0, $$sink22 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0.0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0.0, $116 = 0, $117 = 0.0, $118 = 0, $119 = 0.0;
 var $12 = 0, $120 = 0, $121 = 0, $13 = 0.0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0, $43 = 0, $44 = 0.0, $45 = 0, $46 = 0.0;
 var $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0, $62 = 0.0, $63 = 0.0, $64 = 0;
 var $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0.0, $79 = 0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0;
 var $83 = 0.0, $84 = 0, $85 = 0, $86 = 0, $87 = 0.0, $88 = 0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0, $94 = 0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0, $99 = 0, $count$0$lcssa = 0, $count$03 = 0;
 var $count$03$us = 0, $count$2 = 0, $count$2$us = 0, $exitcond = 0, $exitcond18 = 0, $exitcond19 = 0, $j$0$lcssa = 0, $j$010 = 0, $j$010$us = 0, $j$14 = 0, $j$14$us = 0, $j$21 = 0, $n$ = 0, $or$cond = 0, $storemerge = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $n << 2;
 $$alloca_mul = $0;
 $1 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $2 = ((($p$0$1$val)) + 500|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0);
 if ($4) {
  $9 = $n;
 } else {
  $5 = ((($p$0$1$val)) + 504|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (($6) - ($i))|0;
  $9 = $7;
 }
 $8 = ($9|0)>($n|0);
 $n$ = $8 ? $n : $9;
 $10 = ($n$|0)>(0);
 L4: do {
  if ($10) {
   $11 = ($flags|0)==(0|0);
   if ($11) {
    $j$010$us = 0;
    while(1) {
     $12 = (($q) + ($j$010$us<<2)|0);
     $13 = +HEAPF32[$12>>2];
     $14 = (($f) + ($j$010$us<<2)|0);
     $15 = +HEAPF32[$14>>2];
     $16 = $13 / $15;
     $17 = (($r) + ($j$010$us<<2)|0);
     $18 = +HEAPF32[$17>>2];
     $19 = $18 < 0.0;
     $20 = $16;
     $21 = (+Math_sqrt((+$20)));
     $22 = (+_rint($21));
     $23 = -$22;
     $$sink$in = $19 ? $23 : $22;
     $$sink = (~~(($$sink$in)));
     $24 = (($out) + ($j$010$us<<2)|0);
     HEAP32[$24>>2] = $$sink;
     $25 = (($j$010$us) + 1)|0;
     $26 = ($25|0)<($n$|0);
     if ($26) {
      $j$010$us = $25;
     } else {
      $j$0$lcssa = $n$;
      break L4;
     }
    }
   } else {
    $j$010 = 0;
   }
   while(1) {
    $56 = (($flags) + ($j$010<<2)|0);
    $57 = HEAP32[$56>>2]|0;
    $58 = ($57|0)==(0);
    do {
     if ($58) {
      $59 = (($q) + ($j$010<<2)|0);
      $60 = +HEAPF32[$59>>2];
      $61 = (($f) + ($j$010<<2)|0);
      $62 = +HEAPF32[$61>>2];
      $63 = $60 / $62;
      $64 = (($r) + ($j$010<<2)|0);
      $65 = +HEAPF32[$64>>2];
      $66 = $65 < 0.0;
      $67 = $63;
      $68 = (+Math_sqrt((+$67)));
      $69 = (+_rint($68));
      if ($66) {
       $70 = -$69;
       $71 = (~~(($70)));
       $72 = (($out) + ($j$010<<2)|0);
       HEAP32[$72>>2] = $71;
       break;
      } else {
       $73 = (~~(($69)));
       $74 = (($out) + ($j$010<<2)|0);
       HEAP32[$74>>2] = $73;
       break;
      }
     }
    } while(0);
    $75 = (($j$010) + 1)|0;
    $76 = ($75|0)<($n$|0);
    if ($76) {
     $j$010 = $75;
    } else {
     $j$0$lcssa = $n$;
     break;
    }
   }
  } else {
   $j$0$lcssa = 0;
  }
 } while(0);
 $27 = ($j$0$lcssa|0)<($n|0);
 if (!($27)) {
  $$5 = 0.0;
  STACKTOP = sp;return (+$$5);
 }
 $28 = ($flags|0)==(0|0);
 $29 = (($limit) - ($i))|0;
 if ($28) {
  $$06 = 0.0;$count$03 = 0;$j$14 = $j$0$lcssa;
  while(1) {
   $77 = (($q) + ($j$14<<2)|0);
   $78 = +HEAPF32[$77>>2];
   $79 = (($f) + ($j$14<<2)|0);
   $80 = +HEAPF32[$79>>2];
   $81 = $78 / $80;
   $82 = $81 < 0.25;
   if ($82) {
    $83 = $$06 + $81;
    $84 = (($count$03) + 1)|0;
    $85 = (($1) + ($count$03<<2)|0);
    HEAP32[$85>>2] = $77;
    $$2 = $83;$count$2 = $84;
   } else {
    $86 = (($r) + ($j$14<<2)|0);
    $87 = +HEAPF32[$86>>2];
    $88 = $87 < 0.0;
    $89 = $81;
    $90 = (+Math_sqrt((+$89)));
    $91 = (+_rint($90));
    $92 = -$91;
    $$sink21$in = $88 ? $92 : $91;
    $$sink21 = (~~(($$sink21$in)));
    $93 = (($out) + ($j$14<<2)|0);
    HEAP32[$93>>2] = $$sink21;
    $94 = Math_imul($$sink21, $$sink21)|0;
    $95 = (+($94|0));
    $96 = +HEAPF32[$79>>2];
    $97 = $96 * $95;
    HEAPF32[$77>>2] = $97;
    $$2 = $$06;$count$2 = $count$03;
   }
   $98 = (($j$14) + 1)|0;
   $exitcond18 = ($98|0)==($n|0);
   if ($exitcond18) {
    $$0$lcssa = $$2;$count$0$lcssa = $count$2;
    break;
   } else {
    $$06 = $$2;$count$03 = $count$2;$j$14 = $98;
   }
  }
 } else {
  $$06$us = 0.0;$count$03$us = 0;$j$14$us = $j$0$lcssa;
  while(1) {
   $30 = (($flags) + ($j$14$us<<2)|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = ($31|0)==(0);
   do {
    if ($32) {
     $33 = (($q) + ($j$14$us<<2)|0);
     $34 = +HEAPF32[$33>>2];
     $35 = (($f) + ($j$14$us<<2)|0);
     $36 = +HEAPF32[$35>>2];
     $37 = $34 / $36;
     $38 = !($37 < 0.25);
     $39 = ($j$14$us|0)<($29|0);
     $or$cond = $38 | $39;
     if ($or$cond) {
      $43 = (($r) + ($j$14$us<<2)|0);
      $44 = +HEAPF32[$43>>2];
      $45 = $44 < 0.0;
      $46 = $37;
      $47 = (+Math_sqrt((+$46)));
      $48 = (+_rint($47));
      $49 = -$48;
      $$sink20$in = $45 ? $49 : $48;
      $$sink20 = (~~(($$sink20$in)));
      $50 = (($out) + ($j$14$us<<2)|0);
      HEAP32[$50>>2] = $$sink20;
      $51 = Math_imul($$sink20, $$sink20)|0;
      $52 = (+($51|0));
      $53 = +HEAPF32[$35>>2];
      $54 = $53 * $52;
      HEAPF32[$33>>2] = $54;
      $$2$us = $$06$us;$count$2$us = $count$03$us;
      break;
     } else {
      $40 = $$06$us + $37;
      $41 = (($count$03$us) + 1)|0;
      $42 = (($1) + ($count$03$us<<2)|0);
      HEAP32[$42>>2] = $33;
      $$2$us = $40;$count$2$us = $41;
      break;
     }
    } else {
     $$2$us = $$06$us;$count$2$us = $count$03$us;
    }
   } while(0);
   $55 = (($j$14$us) + 1)|0;
   $exitcond19 = ($55|0)==($n|0);
   if ($exitcond19) {
    $$0$lcssa = $$2$us;$count$0$lcssa = $count$2$us;
    break;
   } else {
    $$06$us = $$2$us;$count$03$us = $count$2$us;$j$14$us = $55;
   }
  }
 }
 $99 = ($count$0$lcssa|0)==(0);
 if ($99) {
  $$5 = $$0$lcssa;
  STACKTOP = sp;return (+$$5);
 }
 _qsort($1,$count$0$lcssa,4,4450);
 $100 = ($count$0$lcssa|0)>(0);
 if (!($100)) {
  $$5 = $$0$lcssa;
  STACKTOP = sp;return (+$$5);
 }
 $101 = $q;
 $102 = ((($p$0$1$val)) + 512|0);
 $103 = +HEAPF64[$102>>3];
 $$32 = $$0$lcssa;$j$21 = 0;
 while(1) {
  $104 = (($1) + ($j$21<<2)|0);
  $105 = HEAP32[$104>>2]|0;
  $106 = $105;
  $107 = (($106) - ($101))|0;
  $108 = $107 >> 2;
  $109 = $$32;
  $110 = !($109 >= $103);
  if ($110) {
   $$4 = $$32;$$sink22 = 0;$storemerge = 0.0;
  } else {
   $111 = (($r) + ($108<<2)|0);
   $112 = HEAP32[$111>>2]|0;
   $113 = $112 & -2147483648;
   $114 = $113 | 1065353216;
   $115 = (HEAP32[tempDoublePtr>>2]=$114,+HEAPF32[tempDoublePtr>>2]);
   $116 = (~~(($115)));
   $117 = $$32 + -1.0;
   $118 = (($f) + ($108<<2)|0);
   $119 = +HEAPF32[$118>>2];
   $$4 = $117;$$sink22 = $116;$storemerge = $119;
  }
  $120 = (($out) + ($108<<2)|0);
  HEAP32[$120>>2] = $$sink22;
  HEAPF32[$105>>2] = $storemerge;
  $121 = (($j$21) + 1)|0;
  $exitcond = ($121|0)==($count$0$lcssa|0);
  if ($exitcond) {
   $$5 = $$4;
   break;
  } else {
   $$32 = $$4;$j$21 = $121;
  }
 }
 STACKTOP = sp;return (+$$5);
}
function _apsort($a,$b) {
 $a = $a|0;
 $b = $b|0;
 var $0 = 0, $1 = 0.0, $2 = 0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$a>>2]|0;
 $1 = +HEAPF32[$0>>2];
 $2 = HEAP32[$b>>2]|0;
 $3 = +HEAPF32[$2>>2];
 $4 = $1 < $3;
 $5 = $4&1;
 $6 = $1 > $3;
 $7 = $6&1;
 $8 = (($5) - ($7))|0;
 return ($8|0);
}
function _res0_unpack($vi,$opb) {
 $vi = $vi|0;
 $opb = $opb|0;
 var $$0 = 0, $$01$i = 0, $$lcssa50 = 0, $$lcssa51 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0, $acc$0$lcssa37 = 0, $acc$029 = 0, $cascade$0 = 0;
 var $cascade$1 = 0, $cond = 0, $dim$021 = 0, $j$030 = 0, $j$126 = 0, $j$223 = 0, $partvals$022 = 0, $ret$0$lcssa$i = 0, $ret$02$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_calloc(1,2840)|0);
 $1 = ((($vi)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (_oggpack_read($opb,24)|0);
 HEAP32[$0>>2] = $3;
 $4 = (_oggpack_read($opb,24)|0);
 $5 = ((($0)) + 4|0);
 HEAP32[$5>>2] = $4;
 $6 = (_oggpack_read($opb,24)|0);
 $7 = (($6) + 1)|0;
 $8 = ((($0)) + 8|0);
 HEAP32[$8>>2] = $7;
 $9 = (_oggpack_read($opb,6)|0);
 $10 = (($9) + 1)|0;
 $11 = ((($0)) + 12|0);
 HEAP32[$11>>2] = $10;
 $12 = (_oggpack_read($opb,8)|0);
 $13 = ((($0)) + 20|0);
 HEAP32[$13>>2] = $12;
 $14 = ($12|0)<(0);
 L1: do {
  if ($14) {
   label = 25;
  } else {
   $15 = ($9|0)>(-1);
   if ($15) {
    $16 = ((($0)) + 24|0);
    $acc$029 = 0;$j$030 = 0;
    while(1) {
     $19 = (_oggpack_read($opb,3)|0);
     $20 = (_oggpack_read($opb,1)|0);
     $21 = ($20|0)<(0);
     if ($21) {
      label = 25;
      break L1;
     }
     $22 = ($20|0)==(0);
     if ($22) {
      $cascade$1 = $19;
     } else {
      $23 = (_oggpack_read($opb,5)|0);
      $24 = ($23|0)<(0);
      $25 = $23 << 3;
      $26 = $24 ? 0 : $25;
      $cascade$0 = $26 | $19;
      $27 = $23 >> 31;
      $28 = $27 & 2;
      $cond = ($28|0)==(0);
      if ($cond) {
       $cascade$1 = $cascade$0;
      } else {
       label = 25;
       break L1;
      }
     }
     $29 = (($16) + ($j$030<<2)|0);
     HEAP32[$29>>2] = $cascade$1;
     $30 = ($cascade$1|0)==(0);
     if ($30) {
      $ret$0$lcssa$i = 0;
     } else {
      $$01$i = $cascade$1;$ret$02$i = 0;
      while(1) {
       $31 = $$01$i & 1;
       $32 = (($31) + ($ret$02$i))|0;
       $33 = $$01$i >>> 1;
       $34 = ($33|0)==(0);
       if ($34) {
        $ret$0$lcssa$i = $32;
        break;
       } else {
        $$01$i = $33;$ret$02$i = $32;
       }
      }
     }
     $35 = (($ret$0$lcssa$i) + ($acc$029))|0;
     $36 = (($j$030) + 1)|0;
     $37 = HEAP32[$11>>2]|0;
     $38 = ($36|0)<($37|0);
     if ($38) {
      $acc$029 = $35;$j$030 = $36;
     } else {
      $$lcssa51 = $35;
      break;
     }
    }
    $17 = ($$lcssa51|0)>(0);
    if ($17) {
     $18 = ((($0)) + 280|0);
     $j$126 = 0;
     while(1) {
      $39 = (_oggpack_read($opb,8)|0);
      $40 = ($39|0)<(0);
      if ($40) {
       break L1;
      }
      $41 = (($18) + ($j$126<<2)|0);
      HEAP32[$41>>2] = $39;
      $42 = (($j$126) + 1)|0;
      $43 = ($42|0)<($$lcssa51|0);
      if ($43) {
       $j$126 = $42;
      } else {
       $72 = $17;$acc$0$lcssa37 = $$lcssa51;
       break;
      }
     }
    } else {
     $72 = 0;$acc$0$lcssa37 = $$lcssa51;
    }
   } else {
    $72 = 0;$acc$0$lcssa37 = 0;
   }
   $44 = HEAP32[$13>>2]|0;
   $45 = ((($2)) + 24|0);
   $46 = HEAP32[$45>>2]|0;
   $47 = ($44|0)<($46|0);
   if ($47) {
    if ($72) {
     $48 = ((($0)) + 280|0);
     $j$223 = 0;
     while(1) {
      $51 = (($48) + ($j$223<<2)|0);
      $52 = HEAP32[$51>>2]|0;
      $53 = ($52|0)<($46|0);
      if (!($53)) {
       break L1;
      }
      $54 = (((($2)) + 1824|0) + ($52<<2)|0);
      $55 = HEAP32[$54>>2]|0;
      $56 = ((($55)) + 12|0);
      $57 = HEAP32[$56>>2]|0;
      $58 = ($57|0)==(0);
      $50 = (($j$223) + 1)|0;
      if ($58) {
       break L1;
      }
      $49 = ($50|0)<($acc$0$lcssa37|0);
      if ($49) {
       $j$223 = $50;
      } else {
       break;
      }
     }
    }
    $59 = (((($2)) + 1824|0) + ($44<<2)|0);
    $60 = HEAP32[$59>>2]|0;
    $61 = ((($60)) + 4|0);
    $62 = HEAP32[$61>>2]|0;
    $63 = HEAP32[$60>>2]|0;
    $64 = ($63|0)<(1);
    if (!($64)) {
     $65 = HEAP32[$11>>2]|0;
     $dim$021 = $63;$partvals$022 = 1;
     while(1) {
      $68 = Math_imul($65, $partvals$022)|0;
      $69 = ($68|0)>($62|0);
      if ($69) {
       break L1;
      }
      $66 = (($dim$021) + -1)|0;
      $67 = ($dim$021|0)>(1);
      if ($67) {
       $dim$021 = $66;$partvals$022 = $68;
      } else {
       $$lcssa50 = $68;
       break;
      }
     }
     $70 = ((($0)) + 16|0);
     HEAP32[$70>>2] = $$lcssa50;
     $$0 = $0;
     return ($$0|0);
    }
   }
  }
 } while(0);
 if ((label|0) == 25) {
  $71 = ($0|0)==(0|0);
  if ($71) {
   $$0 = 0;
   return ($$0|0);
  }
 }
 _free($0);
 $$0 = 0;
 return ($$0|0);
}
function _res0_look($vd,$vr) {
 $vd = $vd|0;
 $vr = $vr|0;
 var $$cast = 0, $$lcssa = 0, $$maxstage$0 = 0, $$pn = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $8 = 0, $9 = 0, $acc$016 = 0, $acc$112 = 0;
 var $acc$2 = 0, $acc$3 = 0, $exitcond = 0, $exitcond24 = 0, $exitcond25 = 0, $j$018 = 0, $j$18 = 0, $j$24 = 0, $j$24$us = 0, $k$013 = 0, $k$11$us = 0, $maxstage$0$lcssa = 0, $maxstage$017 = 0, $maxstage$2 = 0, $mult$03$us = 0, $val$02$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_calloc(1,44)|0);
 $1 = ((($vd)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 HEAP32[$0>>2] = $vr;
 $5 = ((($vr)) + 12|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($0)) + 4|0);
 HEAP32[$7>>2] = $6;
 $8 = ((($4)) + 2848|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($0)) + 12|0);
 HEAP32[$10>>2] = $9;
 $$cast = $9;
 $11 = ((($vr)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($$cast) + (($12*56)|0)|0);
 $14 = ((($0)) + 16|0);
 HEAP32[$14>>2] = $13;
 $15 = HEAP32[$13>>2]|0;
 $16 = (_calloc($6,4)|0);
 $17 = ((($0)) + 20|0);
 HEAP32[$17>>2] = $16;
 $18 = ($6|0)>(0);
 if ($18) {
  $19 = ((($vr)) + 24|0);
  $20 = ((($vr)) + 280|0);
  $acc$016 = 0;$j$018 = 0;$maxstage$017 = 0;
  while(1) {
   $21 = (($19) + ($j$018<<2)|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = (_ov_ilog($22)|0);
   $24 = ($23|0)==(0);
   if ($24) {
    $acc$3 = $acc$016;$maxstage$2 = $maxstage$017;
   } else {
    $25 = ($23|0)>($maxstage$017|0);
    $$maxstage$0 = $25 ? $23 : $maxstage$017;
    $26 = (_calloc($23,4)|0);
    $27 = (($16) + ($j$018<<2)|0);
    HEAP32[$27>>2] = $26;
    $28 = ($23|0)>(0);
    if ($28) {
     $29 = HEAP32[$21>>2]|0;
     $30 = (($16) + ($j$018<<2)|0);
     $acc$112 = $acc$016;$k$013 = 0;
     while(1) {
      $31 = 1 << $k$013;
      $32 = $29 & $31;
      $33 = ($32|0)==(0);
      if ($33) {
       $acc$2 = $acc$112;
      } else {
       $34 = HEAP32[$8>>2]|0;
       $35 = (($acc$112) + 1)|0;
       $36 = (($20) + ($acc$112<<2)|0);
       $37 = HEAP32[$36>>2]|0;
       $38 = (($34) + (($37*56)|0)|0);
       $39 = HEAP32[$30>>2]|0;
       $40 = (($39) + ($k$013<<2)|0);
       HEAP32[$40>>2] = $38;
       $acc$2 = $35;
      }
      $41 = (($k$013) + 1)|0;
      $exitcond25 = ($41|0)==($23|0);
      if ($exitcond25) {
       $acc$3 = $acc$2;$maxstage$2 = $$maxstage$0;
       break;
      } else {
       $acc$112 = $acc$2;$k$013 = $41;
      }
     }
    } else {
     $acc$3 = $acc$016;$maxstage$2 = $$maxstage$0;
    }
   }
   $42 = (($j$018) + 1)|0;
   $43 = ($42|0)<($6|0);
   if ($43) {
    $acc$016 = $acc$3;$j$018 = $42;$maxstage$017 = $maxstage$2;
   } else {
    $maxstage$0$lcssa = $maxstage$2;
    break;
   }
  }
 } else {
  $maxstage$0$lcssa = 0;
 }
 $44 = ((($0)) + 24|0);
 HEAP32[$44>>2] = 1;
 $45 = ($15|0)>(0);
 if ($45) {
  $46 = HEAP32[$7>>2]|0;
  $48 = 1;$j$18 = 0;
  while(1) {
   $47 = Math_imul($48, $46)|0;
   $49 = (($j$18) + 1)|0;
   $exitcond24 = ($49|0)==($15|0);
   if ($exitcond24) {
    $$lcssa = $47;
    break;
   } else {
    $48 = $47;$j$18 = $49;
   }
  }
  HEAP32[$44>>2] = $$lcssa;
  $52 = $$lcssa;
 } else {
  $52 = 1;
 }
 $50 = ((($0)) + 8|0);
 HEAP32[$50>>2] = $maxstage$0$lcssa;
 $51 = $52 << 2;
 $53 = (_malloc($51)|0);
 $54 = ((($0)) + 28|0);
 HEAP32[$54>>2] = $53;
 $55 = ($52|0)>(0);
 if (!($55)) {
  return ($0|0);
 }
 $56 = $15 << 2;
 if (!($45)) {
  $j$24 = 0;
  while(1) {
   $70 = (_malloc($56)|0);
   $71 = (($53) + ($j$24<<2)|0);
   HEAP32[$71>>2] = $70;
   $72 = (($j$24) + 1)|0;
   $73 = ($72|0)<($52|0);
   if ($73) {
    $j$24 = $72;
   } else {
    break;
   }
  }
  return ($0|0);
 }
 $57 = HEAP32[$7>>2]|0;
 $58 = HEAP32[$54>>2]|0;
 $j$24$us = 0;
 while(1) {
  $59 = (_malloc($56)|0);
  $60 = (($53) + ($j$24$us<<2)|0);
  HEAP32[$60>>2] = $59;
  $61 = (($58) + ($j$24$us<<2)|0);
  $62 = HEAP32[$61>>2]|0;
  $$pn = $52;$k$11$us = 0;$val$02$us = $j$24$us;
  while(1) {
   $mult$03$us = (($$pn|0) / ($57|0))&-1;
   $63 = (($val$02$us|0) / ($mult$03$us|0))&-1;
   $64 = Math_imul($63, $mult$03$us)|0;
   $65 = (($val$02$us) - ($64))|0;
   $66 = (($62) + ($k$11$us<<2)|0);
   HEAP32[$66>>2] = $63;
   $67 = (($k$11$us) + 1)|0;
   $exitcond = ($67|0)==($15|0);
   if ($exitcond) {
    break;
   } else {
    $$pn = $mult$03$us;$k$11$us = $67;$val$02$us = $65;
   }
  }
  $68 = (($j$24$us) + 1)|0;
  $69 = ($68|0)<($52|0);
  if ($69) {
   $j$24$us = $68;
  } else {
   break;
  }
 }
 return ($0|0);
}
function _res0_free_info($i) {
 $i = $i|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($i|0)==(0|0);
 if (!($0)) {
  _free($i);
 }
 return;
}
function _res0_free_look($i) {
 $i = $i|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $j$03 = 0, $j$12 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($i|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($i)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)>(0);
 $4 = ((($i)) + 20|0);
 if ($3) {
  $24 = $2;$j$03 = 0;
  while(1) {
   $5 = HEAP32[$4>>2]|0;
   $6 = (($5) + ($j$03<<2)|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0|0);
   if ($8) {
    $11 = $24;
   } else {
    _free($7);
    $$pre = HEAP32[$1>>2]|0;
    $11 = $$pre;
   }
   $9 = (($j$03) + 1)|0;
   $10 = ($9|0)<($11|0);
   if ($10) {
    $24 = $11;$j$03 = $9;
   } else {
    break;
   }
  }
 }
 $12 = HEAP32[$4>>2]|0;
 _free($12);
 $13 = ((($i)) + 24|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ($14|0)>(0);
 $16 = ((($i)) + 28|0);
 if ($15) {
  $j$12 = 0;
  while(1) {
   $17 = HEAP32[$16>>2]|0;
   $18 = (($17) + ($j$12<<2)|0);
   $19 = HEAP32[$18>>2]|0;
   _free($19);
   $20 = (($j$12) + 1)|0;
   $21 = HEAP32[$13>>2]|0;
   $22 = ($20|0)<($21|0);
   if ($22) {
    $j$12 = $20;
   } else {
    break;
   }
  }
 }
 $23 = HEAP32[$16>>2]|0;
 _free($23);
 _free($i);
 return;
}
function _res0_inverse($vb,$vl,$in,$nonzero,$ch) {
 $vb = $vb|0;
 $vl = $vl|0;
 $in = $in|0;
 $nonzero = $nonzero|0;
 $ch = $ch|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, $used$02 = 0, $used$1 = 0, $used$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($ch|0)>(0);
 if ($0) {
  $i$01 = 0;$used$02 = 0;
 } else {
  return 0;
 }
 while(1) {
  $1 = (($nonzero) + ($i$01<<2)|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ($2|0)==(0);
  if ($3) {
   $used$1 = $used$02;
  } else {
   $4 = (($in) + ($i$01<<2)|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = (($used$02) + 1)|0;
   $7 = (($in) + ($used$02<<2)|0);
   HEAP32[$7>>2] = $5;
   $used$1 = $6;
  }
  $8 = (($i$01) + 1)|0;
  $exitcond = ($8|0)==($ch|0);
  if ($exitcond) {
   $used$1$lcssa = $used$1;
   break;
  } else {
   $i$01 = $8;$used$02 = $used$1;
  }
 }
 $9 = ($used$1$lcssa|0)==(0);
 if ($9) {
  return 0;
 }
 __01inverse($vb,$vl,$in,$used$1$lcssa,4451);
 return 0;
}
function _vorbis_book_decodevs_add($book,$a,$b,$n) {
 $book = $book|0;
 $a = $a|0;
 $b = $b|0;
 $n = $n|0;
 var $$1 = 0, $$alloca_mul = 0, $$not = 0, $$not13 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $brmerge = 0, $exitcond = 0, $i$07 = 0, $i$15$us = 0, $j$04$us = 0;
 var $o$06$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($book)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(0);
 if (!($2)) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $3 = HEAP32[$book>>2]|0;
 $4 = (($n|0) / ($3|0))&-1;
 $5 = $4 << 2;
 $$alloca_mul = $5;
 $6 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $7 = ($4|0)>(0);
 L4: do {
  if ($7) {
   $8 = ((($book)) + 16|0);
   $i$07 = 0;
   while(1) {
    $22 = (_decode_packed_entry_number($book,$b)|0);
    $23 = ($22|0)==(-1);
    if ($23) {
     $$1 = -1;
     break;
    }
    $24 = HEAP32[$8>>2]|0;
    $25 = HEAP32[$book>>2]|0;
    $26 = Math_imul($25, $22)|0;
    $27 = (($24) + ($26<<2)|0);
    $28 = (($6) + ($i$07<<2)|0);
    HEAP32[$28>>2] = $27;
    $29 = (($i$07) + 1)|0;
    $30 = ($29|0)<($4|0);
    if ($30) {
     $i$07 = $29;
    } else {
     $9 = $25;
     break L4;
    }
   }
   STACKTOP = sp;return ($$1|0);
  } else {
   $9 = $3;
  }
 } while(0);
 $$not = ($9|0)<(1);
 $$not13 = $7 ^ 1;
 $brmerge = $$not | $$not13;
 if ($brmerge) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 } else {
  $i$15$us = 0;$o$06$us = 0;
 }
 while(1) {
  $j$04$us = 0;
  while(1) {
   $10 = (($6) + ($j$04$us<<2)|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = (($11) + ($i$15$us<<2)|0);
   $13 = +HEAPF32[$12>>2];
   $14 = (($j$04$us) + ($o$06$us))|0;
   $15 = (($a) + ($14<<2)|0);
   $16 = +HEAPF32[$15>>2];
   $17 = $13 + $16;
   HEAPF32[$15>>2] = $17;
   $18 = (($j$04$us) + 1)|0;
   $exitcond = ($18|0)==($4|0);
   if ($exitcond) {
    break;
   } else {
    $j$04$us = $18;
   }
  }
  $19 = (($i$15$us) + 1)|0;
  $20 = (($o$06$us) + ($4))|0;
  $21 = ($19|0)<($9|0);
  if ($21) {
   $i$15$us = $19;$o$06$us = $20;
  } else {
   $$1 = 0;
   break;
  }
 }
 STACKTOP = sp;return ($$1|0);
}
function __01inverse($vb,$vl,$in,$ch,$decodepart) {
 $vb = $vb|0;
 $vl = $vl|0;
 $in = $in|0;
 $ch = $ch|0;
 $decodepart = $decodepart|0;
 var $$ = 0, $$alloca_mul = 0, $$not = 0, $$not34 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $brmerge = 0, $exitcond = 0, $i$026 = 0, $i$1$lcssa = 0;
 var $i$123$us = 0, $j$031 = 0, $j$118 = 0, $j$219$us = 0, $k$022$us = 0, $l$025 = 0, $or$cond$us = 0, $or$cond21 = 0, $s$029 = 0, $umax = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$vl>>2]|0;
 $1 = ((($0)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($vl)) + 16|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($vb)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 >> 1;
 $9 = ((($0)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10|0)<($8|0);
 $$ = $11 ? $10 : $8;
 $12 = HEAP32[$0>>2]|0;
 $13 = (($$) - ($12))|0;
 $14 = ($13|0)>(0);
 if (!($14)) {
  STACKTOP = sp;return;
 }
 $15 = (($13|0) / ($2|0))&-1;
 $16 = $ch << 2;
 $$alloca_mul = $16;
 $17 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $18 = ($ch|0)>(0);
 if ($18) {
  $19 = (($5) + -1)|0;
  $20 = (($19) + ($15))|0;
  $21 = (($20|0) / ($5|0))&-1;
  $22 = $21 << 2;
  $j$031 = 0;
  while(1) {
   $33 = (__vorbis_block_alloc($vb,$22)|0);
   $34 = (($17) + ($j$031<<2)|0);
   HEAP32[$34>>2] = $33;
   $35 = (($j$031) + 1)|0;
   $exitcond = ($35|0)==($ch|0);
   if ($exitcond) {
    break;
   } else {
    $j$031 = $35;
   }
  }
 }
 $23 = ((($vl)) + 8|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ($24|0)>(0);
 if (!($25)) {
  STACKTOP = sp;return;
 }
 $26 = ($15|0)>(0);
 $27 = ($5|0)>(0);
 $28 = ((($vl)) + 20|0);
 $29 = ((($vb)) + 4|0);
 $30 = ((($0)) + 16|0);
 $31 = ((($vl)) + 28|0);
 $32 = (0 - ($5))|0;
 $$not34 = ($ch|0)<(1);
 $90 = $24;$s$029 = 0;
 L12: while(1) {
  if ($26) {
   $36 = 1 << $s$029;
   $$not = ($s$029|0)!=(0);
   $brmerge = $$not | $$not34;
   $i$026 = 0;$l$025 = 0;
   while(1) {
    if (!($brmerge)) {
     $j$118 = 0;
     while(1) {
      $73 = HEAP32[$3>>2]|0;
      $74 = (_vorbis_book_decode($73,$29)|0);
      $75 = ($74|0)==(-1);
      if ($75) {
       label = 26;
       break L12;
      }
      $76 = HEAP32[$30>>2]|0;
      $77 = ($74|0)<($76|0);
      if (!($77)) {
       label = 26;
       break L12;
      }
      $78 = HEAP32[$31>>2]|0;
      $79 = (($78) + ($74<<2)|0);
      $80 = HEAP32[$79>>2]|0;
      $81 = (($17) + ($j$118<<2)|0);
      $82 = HEAP32[$81>>2]|0;
      $83 = (($82) + ($l$025<<2)|0);
      HEAP32[$83>>2] = $80;
      $84 = ($80|0)==(0);
      $72 = (($j$118) + 1)|0;
      if ($84) {
       label = 26;
       break L12;
      }
      $71 = ($72|0)<($ch|0);
      if ($71) {
       $j$118 = $72;
      } else {
       break;
      }
     }
    }
    $37 = ($i$026|0)<($15|0);
    $or$cond21 = $27 & $37;
    do {
     if ($or$cond21) {
      if ($18) {
       $i$123$us = $i$026;$k$022$us = 0;
      } else {
       $38 = (($i$026) - ($15))|0;
       $39 = ($38>>>0)<($32>>>0);
       $umax = $39 ? $32 : $38;
       $40 = (($i$026) - ($umax))|0;
       $i$1$lcssa = $40;
       break;
      }
      while(1) {
       $43 = Math_imul($i$123$us, $2)|0;
       $j$219$us = 0;
       while(1) {
        $41 = HEAP32[$0>>2]|0;
        $42 = (($41) + ($43))|0;
        $44 = (($17) + ($j$219$us<<2)|0);
        $45 = HEAP32[$44>>2]|0;
        $46 = (($45) + ($l$025<<2)|0);
        $47 = HEAP32[$46>>2]|0;
        $48 = (($47) + ($k$022$us<<2)|0);
        $49 = HEAP32[$48>>2]|0;
        $50 = (((($0)) + 24|0) + ($49<<2)|0);
        $51 = HEAP32[$50>>2]|0;
        $52 = $51 & $36;
        $53 = ($52|0)==(0);
        if (!($53)) {
         $54 = HEAP32[$28>>2]|0;
         $55 = (($54) + ($49<<2)|0);
         $56 = HEAP32[$55>>2]|0;
         $57 = (($56) + ($s$029<<2)|0);
         $58 = HEAP32[$57>>2]|0;
         $59 = ($58|0)==(0|0);
         if (!($59)) {
          $60 = (($in) + ($j$219$us<<2)|0);
          $61 = HEAP32[$60>>2]|0;
          $62 = (($61) + ($42<<2)|0);
          $63 = (FUNCTION_TABLE_iiiii[$decodepart & 8191]($58,$62,$29,$2)|0);
          $64 = ($63|0)==(-1);
          if ($64) {
           label = 26;
           break L12;
          }
         }
        }
        $65 = (($j$219$us) + 1)|0;
        $66 = ($65|0)<($ch|0);
        if ($66) {
         $j$219$us = $65;
        } else {
         break;
        }
       }
       $67 = (($k$022$us) + 1)|0;
       $68 = (($i$123$us) + 1)|0;
       $69 = ($67|0)<($5|0);
       $70 = ($68|0)<($15|0);
       $or$cond$us = $69 & $70;
       if ($or$cond$us) {
        $i$123$us = $68;$k$022$us = $67;
       } else {
        $i$1$lcssa = $68;
        break;
       }
      }
     } else {
      $i$1$lcssa = $i$026;
     }
    } while(0);
    $85 = (($l$025) + 1)|0;
    $86 = ($i$1$lcssa|0)<($15|0);
    if ($86) {
     $i$026 = $i$1$lcssa;$l$025 = $85;
    } else {
     break;
    }
   }
   $$pre = HEAP32[$23>>2]|0;
   $89 = $$pre;
  } else {
   $89 = $90;
  }
  $87 = (($s$029) + 1)|0;
  $88 = ($87|0)<($89|0);
  if ($88) {
   $90 = $89;$s$029 = $87;
  } else {
   label = 26;
   break;
  }
 }
 if ((label|0) == 26) {
  STACKTOP = sp;return;
 }
}
function _res0_pack($vr,$opb) {
 $vr = $vr|0;
 $opb = $opb|0;
 var $$01$i = 0, $$lcssa10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $acc$03 = 0, $exitcond = 0;
 var $j$02 = 0, $j$11 = 0, $ret$0$lcssa$i = 0, $ret$02$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$vr>>2]|0;
 _oggpack_write($opb,$0,24);
 $1 = ((($vr)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 _oggpack_write($opb,$2,24);
 $3 = ((($vr)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + -1)|0;
 _oggpack_write($opb,$5,24);
 $6 = ((($vr)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (($7) + -1)|0;
 _oggpack_write($opb,$8,6);
 $9 = ((($vr)) + 20|0);
 $10 = HEAP32[$9>>2]|0;
 _oggpack_write($opb,$10,8);
 $11 = HEAP32[$6>>2]|0;
 $12 = ($11|0)>(0);
 if (!($12)) {
  return;
 }
 $13 = ((($vr)) + 24|0);
 $acc$03 = 0;$j$02 = 0;
 while(1) {
  $16 = (($13) + ($j$02<<2)|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = (_ov_ilog($17)|0);
  $19 = ($18|0)>(3);
  $20 = HEAP32[$16>>2]|0;
  if ($19) {
   _oggpack_write($opb,$20,3);
   _oggpack_write($opb,1,1);
   $21 = HEAP32[$16>>2]|0;
   $22 = $21 >> 3;
   _oggpack_write($opb,$22,5);
  } else {
   _oggpack_write($opb,$20,4);
  }
  $23 = HEAP32[$16>>2]|0;
  $24 = ($23|0)==(0);
  if ($24) {
   $ret$0$lcssa$i = 0;
  } else {
   $$01$i = $23;$ret$02$i = 0;
   while(1) {
    $25 = $$01$i & 1;
    $26 = (($25) + ($ret$02$i))|0;
    $27 = $$01$i >>> 1;
    $28 = ($27|0)==(0);
    if ($28) {
     $ret$0$lcssa$i = $26;
     break;
    } else {
     $$01$i = $27;$ret$02$i = $26;
    }
   }
  }
  $29 = (($ret$0$lcssa$i) + ($acc$03))|0;
  $30 = (($j$02) + 1)|0;
  $31 = HEAP32[$6>>2]|0;
  $32 = ($30|0)<($31|0);
  if ($32) {
   $acc$03 = $29;$j$02 = $30;
  } else {
   $$lcssa10 = $29;
   break;
  }
 }
 $14 = ($$lcssa10|0)>(0);
 if (!($14)) {
  return;
 }
 $15 = ((($vr)) + 280|0);
 $j$11 = 0;
 while(1) {
  $33 = (($15) + ($j$11<<2)|0);
  $34 = HEAP32[$33>>2]|0;
  _oggpack_write($opb,$34,8);
  $35 = (($j$11) + 1)|0;
  $exitcond = ($35|0)==($$lcssa10|0);
  if ($exitcond) {
   break;
  } else {
   $j$11 = $35;
  }
 }
 return;
}
function _res1_class($vb,$vl,$in,$nonzero,$ch) {
 $vb = $vb|0;
 $vl = $vl|0;
 $in = $in|0;
 $nonzero = $nonzero|0;
 $ch = $ch|0;
 var $$0 = 0, $$lcssa = 0, $$max$0$us$i$us = 0, $$max$0$us$i$us$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0.0;
 var $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $9 = 0, $ent$06$us$i$us = 0, $exitcond = 0, $exitcond$i = 0, $exitcond47$i = 0, $exitcond48$i$us = 0, $exitcond49$i = 0, $exitcond49$i$us = 0;
 var $exitcond50$i = 0, $exitcond51$i = 0, $i$020$i = 0, $i$04 = 0, $i$117$i = 0, $ispos$us$i$us = 0, $j$015$i = 0, $j$015$us$i = 0, $j$015$us$i$us = 0, $j$015$us24$i = 0, $k$04$us$i$us = 0, $k$1$lcssa$us$i$ph$us = 0, $k$1$lcssa$us29$i = 0, $k$18$us$i$us = 0, $k$18$us27$i = 0, $max$05$us$i$us = 0, $neg$us$i$us = 0, $or$cond$us$i$us = 0, $or$cond$us28$i = 0, $used$05 = 0;
 var $used$1 = 0, $used$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($ch|0)>(0);
 if ($0) {
  $i$04 = 0;$used$05 = 0;
 } else {
  $$0 = 0;
  return ($$0|0);
 }
 while(1) {
  $1 = (($nonzero) + ($i$04<<2)|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ($2|0)==(0);
  if ($3) {
   $used$1 = $used$05;
  } else {
   $4 = (($in) + ($i$04<<2)|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = (($used$05) + 1)|0;
   $7 = (($in) + ($used$05<<2)|0);
   HEAP32[$7>>2] = $5;
   $used$1 = $6;
  }
  $8 = (($i$04) + 1)|0;
  $exitcond = ($8|0)==($ch|0);
  if ($exitcond) {
   $used$1$lcssa = $used$1;
   break;
  } else {
   $i$04 = $8;$used$05 = $used$1;
  }
 }
 $9 = ($used$1$lcssa|0)==(0);
 if ($9) {
  $$0 = 0;
  return ($$0|0);
 }
 $10 = HEAP32[$vl>>2]|0;
 $11 = ((($10)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($10)) + 12|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($10)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = HEAP32[$10>>2]|0;
 $18 = (($16) - ($17))|0;
 $19 = (($18|0) / ($12|0))&-1;
 $20 = $used$1$lcssa << 2;
 $21 = (__vorbis_block_alloc($vb,$20)|0);
 $22 = (+($12|0));
 $23 = 100.0 / $22;
 $24 = $23;
 $25 = ($used$1$lcssa|0)>(0);
 if ($25) {
  $26 = $19 << 2;
  $i$020$i = 0;
  while(1) {
   $33 = (__vorbis_block_alloc($vb,$26)|0);
   $34 = (($21) + ($i$020$i<<2)|0);
   HEAP32[$34>>2] = $33;
   _memset(($33|0),0,($26|0))|0;
   $35 = (($i$020$i) + 1)|0;
   $exitcond51$i = ($35|0)==($used$1$lcssa|0);
   if ($exitcond51$i) {
    break;
   } else {
    $i$020$i = $35;
   }
  }
 }
 $27 = ($19|0)>(0);
 if ($27) {
  $28 = ($12|0)>(0);
  $29 = (($14) + -1)|0;
  $30 = ($14|0)>(1);
  $31 = $24 * 0.0;
  $32 = (~~(($31)));
  $i$117$i = 0;
  while(1) {
   $36 = Math_imul($i$117$i, $12)|0;
   $37 = HEAP32[$10>>2]|0;
   $38 = (($37) + ($36))|0;
   L21: do {
    if ($25) {
     if (!($28)) {
      if ($30) {
       $j$015$us24$i = 0;
      } else {
       $j$015$i = 0;
       while(1) {
        $81 = (($21) + ($j$015$i<<2)|0);
        $82 = HEAP32[$81>>2]|0;
        $83 = (($82) + ($i$117$i<<2)|0);
        HEAP32[$83>>2] = 0;
        $84 = (($j$015$i) + 1)|0;
        $exitcond$i = ($84|0)==($used$1$lcssa|0);
        if ($exitcond$i) {
         break L21;
        } else {
         $j$015$i = $84;
        }
       }
      }
      while(1) {
       $k$18$us27$i = 0;
       while(1) {
        $68 = (((($10)) + 2328|0) + ($k$18$us27$i<<2)|0);
        $69 = HEAP32[$68>>2]|0;
        $70 = ($69|0)<(0);
        if (!($70)) {
         $71 = (((($10)) + 2584|0) + ($k$18$us27$i<<2)|0);
         $72 = HEAP32[$71>>2]|0;
         $73 = ($72|0)<(0);
         $74 = ($32|0)<($72|0);
         $or$cond$us28$i = $73 | $74;
         if ($or$cond$us28$i) {
          $k$1$lcssa$us29$i = $k$18$us27$i;
          break;
         }
        }
        $79 = (($k$18$us27$i) + 1)|0;
        $80 = ($79|0)<($29|0);
        if ($80) {
         $k$18$us27$i = $79;
        } else {
         $k$1$lcssa$us29$i = $79;
         break;
        }
       }
       $75 = (($21) + ($j$015$us24$i<<2)|0);
       $76 = HEAP32[$75>>2]|0;
       $77 = (($76) + ($i$117$i<<2)|0);
       HEAP32[$77>>2] = $k$1$lcssa$us29$i;
       $78 = (($j$015$us24$i) + 1)|0;
       $exitcond47$i = ($78|0)==($used$1$lcssa|0);
       if ($exitcond47$i) {
        break L21;
       } else {
        $j$015$us24$i = $78;
       }
      }
     }
     if ($30) {
      $j$015$us$i$us = 0;
     } else {
      $j$015$us$i = 0;
      while(1) {
       $64 = (($21) + ($j$015$us$i<<2)|0);
       $65 = HEAP32[$64>>2]|0;
       $66 = (($65) + ($i$117$i<<2)|0);
       HEAP32[$66>>2] = 0;
       $67 = (($j$015$us$i) + 1)|0;
       $exitcond49$i = ($67|0)==($used$1$lcssa|0);
       if ($exitcond49$i) {
        break L21;
       } else {
        $j$015$us$i = $67;
       }
      }
     }
     while(1) {
      $39 = (($in) + ($j$015$us$i$us<<2)|0);
      $40 = HEAP32[$39>>2]|0;
      $ent$06$us$i$us = 0;$k$04$us$i$us = 0;$max$05$us$i$us = 0;
      while(1) {
       $41 = (($38) + ($k$04$us$i$us))|0;
       $42 = (($40) + ($41<<2)|0);
       $43 = HEAP32[$42>>2]|0;
       $ispos$us$i$us = ($43|0)>(-1);
       $neg$us$i$us = (0 - ($43))|0;
       $44 = $ispos$us$i$us ? $43 : $neg$us$i$us;
       $45 = ($44|0)>($max$05$us$i$us|0);
       $$max$0$us$i$us = $45 ? $44 : $max$05$us$i$us;
       $46 = (($44) + ($ent$06$us$i$us))|0;
       $47 = (($k$04$us$i$us) + 1)|0;
       $exitcond48$i$us = ($47|0)==($12|0);
       if ($exitcond48$i$us) {
        $$lcssa = $46;$$max$0$us$i$us$lcssa = $$max$0$us$i$us;
        break;
       } else {
        $ent$06$us$i$us = $46;$k$04$us$i$us = $47;$max$05$us$i$us = $$max$0$us$i$us;
       }
      }
      $48 = (+($$lcssa|0));
      $49 = $24 * $48;
      $50 = (~~(($49)));
      $k$18$us$i$us = 0;
      while(1) {
       $51 = (((($10)) + 2328|0) + ($k$18$us$i$us<<2)|0);
       $52 = HEAP32[$51>>2]|0;
       $53 = ($$max$0$us$i$us$lcssa|0)>($52|0);
       if (!($53)) {
        $54 = (((($10)) + 2584|0) + ($k$18$us$i$us<<2)|0);
        $55 = HEAP32[$54>>2]|0;
        $56 = ($55|0)<(0);
        $57 = ($50|0)<($55|0);
        $or$cond$us$i$us = $56 | $57;
        if ($or$cond$us$i$us) {
         $k$1$lcssa$us$i$ph$us = $k$18$us$i$us;
         break;
        }
       }
       $58 = (($k$18$us$i$us) + 1)|0;
       $59 = ($58|0)<($29|0);
       if ($59) {
        $k$18$us$i$us = $58;
       } else {
        $k$1$lcssa$us$i$ph$us = $58;
        break;
       }
      }
      $60 = (($21) + ($j$015$us$i$us<<2)|0);
      $61 = HEAP32[$60>>2]|0;
      $62 = (($61) + ($i$117$i<<2)|0);
      HEAP32[$62>>2] = $k$1$lcssa$us$i$ph$us;
      $63 = (($j$015$us$i$us) + 1)|0;
      $exitcond49$i$us = ($63|0)==($used$1$lcssa|0);
      if ($exitcond49$i$us) {
       break;
      } else {
       $j$015$us$i$us = $63;
      }
     }
    }
   } while(0);
   $85 = (($i$117$i) + 1)|0;
   $exitcond50$i = ($85|0)==($19|0);
   if ($exitcond50$i) {
    break;
   } else {
    $i$117$i = $85;
   }
  }
 }
 $86 = ((($vl)) + 40|0);
 $87 = HEAP32[$86>>2]|0;
 $88 = (($87) + 1)|0;
 HEAP32[$86>>2] = $88;
 $$0 = $21;
 return ($$0|0);
}
function _res1_forward($opb,$vb,$vl,$in,$nonzero,$ch,$partword,$submap) {
 $opb = $opb|0;
 $vb = $vb|0;
 $vl = $vl|0;
 $in = $in|0;
 $nonzero = $nonzero|0;
 $ch = $ch|0;
 $partword = $partword|0;
 $submap = $submap|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, $used$02 = 0, $used$1 = 0, $used$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($ch|0)>(0);
 if ($0) {
  $i$01 = 0;$used$02 = 0;
 } else {
  return 0;
 }
 while(1) {
  $1 = (($nonzero) + ($i$01<<2)|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ($2|0)==(0);
  if ($3) {
   $used$1 = $used$02;
  } else {
   $4 = (($in) + ($i$01<<2)|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = (($used$02) + 1)|0;
   $7 = (($in) + ($used$02<<2)|0);
   HEAP32[$7>>2] = $5;
   $used$1 = $6;
  }
  $8 = (($i$01) + 1)|0;
  $exitcond = ($8|0)==($ch|0);
  if ($exitcond) {
   $used$1$lcssa = $used$1;
   break;
  } else {
   $i$01 = $8;$used$02 = $used$1;
  }
 }
 $9 = ($used$1$lcssa|0)==(0);
 if ($9) {
  return 0;
 }
 __01forward($opb,$vl,$in,$used$1$lcssa,$partword);
 return 0;
}
function __01forward($opb,$vl,$in,$ch,$partword) {
 $opb = $opb|0;
 $vl = $vl|0;
 $in = $in|0;
 $ch = $ch|0;
 $partword = $partword|0;
 var $$08$i$i = 0, $$lcssa$i$i = 0, $$lcssa$us$i$i = 0, $$lcssa6$i$i = 0, $$lcssa6$us$i$i = 0, $$lcssa84 = 0, $$lcssa87 = 0, $$not = 0, $$not29 = 0, $$pre$i = 0, $$pre46 = 0, $$pre47 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0;
 var $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0;
 var $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $best$019$i$i = 0, $best$019$us$i$i = 0, $best$2$i$i = 0, $best$2$us$i$i = 0, $bits$0$lcssa$i = 0, $bits$09$i = 0;
 var $brmerge = 0, $e$i$i = 0, $exitcond = 0, $exitcond$i = 0, $exitcond$i$i = 0, $exitcond20$i = 0, $exitcond21$i = 0, $exitcond41 = 0, $exitcond42 = 0, $exitcond43 = 0, $exitcond46$i$i = 0, $exitcond47$i$i = 0, $exitcond48$i$i = 0, $i$010$i = 0, $i$024 = 0, $i$033$i$i = 0, $i$1$lcssa = 0, $i$121 = 0, $i$128$i$i = 0, $i$220$i$i = 0;
 var $i$220$us$i$i = 0, $i$37$i$i = 0, $index$031$i$i = 0, $index$126$i$i = 0, $index$2$i$i = 0, $index$3$lcssa$i$i = 0, $index$318$i$i = 0, $index$318$us$i$i = 0, $index$5$i$i = 0, $index$5$us$i$i = 0, $index$6$i$i = 0, $j$013 = 0, $j$013$us = 0, $j$09$us$i$i = 0, $j$112$i$i = 0, $j$112$us$i$i = 0, $j$115 = 0, $k$012$us = 0, $k$120 = 0, $o$032$i$i = 0;
 var $o$127$i$i = 0, $or$cond = 0, $or$cond$i$i = 0, $or$cond$us$i$i = 0, $or$cond19 = 0, $or$cond36$i$i = 0, $p$i$i = 0, $resbits = 0, $resvals = 0, $s$026 = 0, $scevgep$i = 0, $this$010$us$i$i = 0, $val$011$us = 0, $val$1$us = 0, $val$1$us$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1088|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $p$i$i = sp + 1056|0;
 $e$i$i = sp + 1024|0;
 $resbits = sp + 512|0;
 $resvals = sp;
 $0 = HEAP32[$vl>>2]|0;
 $1 = ((($0)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($0)) + 12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($vl)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($0)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = HEAP32[$0>>2]|0;
 $11 = (($9) - ($10))|0;
 $12 = (($11|0) / ($2|0))&-1;
 _memset(($resbits|0),0,512)|0;
 _memset(($resvals|0),0,512)|0;
 $13 = ((($vl)) + 8|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ($14|0)>(0);
 if (!($15)) {
  STACKTOP = sp;return;
 }
 $16 = ($12|0)>(0);
 $17 = ($7|0)>(0);
 $18 = ($ch|0)>(0);
 $19 = ((($vl)) + 20|0);
 $20 = ((($vl)) + 32|0);
 $21 = ($7|0)>(1);
 $22 = ((($vl)) + 36|0);
 $$not29 = ($ch|0)<(1);
 $221 = $14;$s$026 = 0;
 while(1) {
  if ($16) {
   $23 = ($s$026|0)==(0);
   $24 = 1 << $s$026;
   $$not = $23 ^ 1;
   $brmerge = $$not29 | $$not;
   $i$024 = 0;
   while(1) {
    L10: do {
     if (!($brmerge)) {
      if ($21) {
       $j$013$us = 0;
      } else {
       $j$013 = 0;
       while(1) {
        $46 = (($partword) + ($j$013<<2)|0);
        $47 = HEAP32[$46>>2]|0;
        $48 = (($47) + ($i$024<<2)|0);
        $49 = HEAP32[$48>>2]|0;
        $50 = HEAP32[$5>>2]|0;
        $51 = ((($50)) + 4|0);
        $52 = HEAP32[$51>>2]|0;
        $53 = ($49|0)<($52|0);
        if ($53) {
         $54 = (_vorbis_book_encode($50,$49,$opb)|0);
         $55 = HEAP32[$22>>2]|0;
         $56 = (($55) + ($54))|0;
         HEAP32[$22>>2] = $56;
        }
        $57 = (($j$013) + 1)|0;
        $exitcond = ($57|0)==($ch|0);
        if ($exitcond) {
         break L10;
        } else {
         $j$013 = $57;
        }
       }
      }
      while(1) {
       $26 = (($partword) + ($j$013$us<<2)|0);
       $27 = HEAP32[$26>>2]|0;
       $28 = (($27) + ($i$024<<2)|0);
       $29 = HEAP32[$28>>2]|0;
       $k$012$us = 1;$val$011$us = $29;
       while(1) {
        $35 = Math_imul($val$011$us, $4)|0;
        $36 = (($k$012$us) + ($i$024))|0;
        $37 = ($36|0)<($12|0);
        if ($37) {
         $38 = (($27) + ($36<<2)|0);
         $39 = HEAP32[$38>>2]|0;
         $40 = (($39) + ($35))|0;
         $val$1$us = $40;
        } else {
         $val$1$us = $35;
        }
        $41 = (($k$012$us) + 1)|0;
        $exitcond41 = ($41|0)==($7|0);
        if ($exitcond41) {
         $val$1$us$lcssa = $val$1$us;
         break;
        } else {
         $k$012$us = $41;$val$011$us = $val$1$us;
        }
       }
       $30 = HEAP32[$5>>2]|0;
       $42 = ((($30)) + 4|0);
       $43 = HEAP32[$42>>2]|0;
       $44 = ($val$1$us$lcssa|0)<($43|0);
       if ($44) {
        $31 = (_vorbis_book_encode($30,$val$1$us$lcssa,$opb)|0);
        $32 = HEAP32[$22>>2]|0;
        $33 = (($32) + ($31))|0;
        HEAP32[$22>>2] = $33;
       }
       $34 = (($j$013$us) + 1)|0;
       $exitcond42 = ($34|0)==($ch|0);
       if ($exitcond42) {
        break;
       } else {
        $j$013$us = $34;
       }
      }
     }
    } while(0);
    $45 = ($i$024|0)<($12|0);
    $or$cond19 = $17 & $45;
    if ($or$cond19) {
     $i$121 = $i$024;$k$120 = 0;
     while(1) {
      $58 = Math_imul($i$121, $2)|0;
      $59 = HEAP32[$0>>2]|0;
      $60 = (($59) + ($58))|0;
      if ($18) {
       $j$115 = 0;
       while(1) {
        $61 = (($partword) + ($j$115<<2)|0);
        $62 = HEAP32[$61>>2]|0;
        $63 = (($62) + ($i$121<<2)|0);
        $64 = HEAP32[$63>>2]|0;
        if ($23) {
         $65 = (($resvals) + ($64<<2)|0);
         $66 = HEAP32[$65>>2]|0;
         $67 = (($66) + ($2))|0;
         HEAP32[$65>>2] = $67;
        }
        $68 = (((($0)) + 24|0) + ($64<<2)|0);
        $69 = HEAP32[$68>>2]|0;
        $70 = $69 & $24;
        $71 = ($70|0)==(0);
        if (!($71)) {
         $72 = HEAP32[$19>>2]|0;
         $73 = (($72) + ($64<<2)|0);
         $74 = HEAP32[$73>>2]|0;
         $75 = (($74) + ($s$026<<2)|0);
         $76 = HEAP32[$75>>2]|0;
         $77 = ($76|0)==(0|0);
         if (!($77)) {
          $78 = (($in) + ($j$115<<2)|0);
          $79 = HEAP32[$78>>2]|0;
          $80 = (($79) + ($60<<2)|0);
          $81 = HEAP32[$76>>2]|0;
          $82 = (($2|0) / ($81|0))&-1;
          $83 = ($82|0)>(0);
          if ($83) {
           $84 = ((($76)) + 48|0);
           $85 = ((($76)) + 52|0);
           $86 = ((($76)) + 44|0);
           $87 = ((($76)) + 12|0);
           $88 = ((($76)) + 4|0);
           $96 = $81;$bits$09$i = 0;$i$010$i = 0;
           while(1) {
            $89 = Math_imul($i$010$i, $81)|0;
            $scevgep$i = (($80) + ($89<<2)|0);
            $90 = HEAP32[$84>>2]|0;
            $91 = HEAP32[$85>>2]|0;
            $92 = HEAP32[$86>>2]|0;
            $93 = $92 >> 1;
            ;HEAP32[$p$i$i>>2]=0|0;HEAP32[$p$i$i+4>>2]=0|0;HEAP32[$p$i$i+8>>2]=0|0;HEAP32[$p$i$i+12>>2]=0|0;HEAP32[$p$i$i+16>>2]=0|0;HEAP32[$p$i$i+20>>2]=0|0;HEAP32[$p$i$i+24>>2]=0|0;HEAP32[$p$i$i+28>>2]=0|0;
            $94 = ($91|0)==(1);
            $95 = ($96|0)>(0);
            do {
             if ($94) {
              if (!($95)) {
               $index$2$i$i = 0;
               break;
              }
              $100 = (($92) + -1)|0;
              $101 = $96 << 2;
              _memcpy(($p$i$i|0),($scevgep$i|0),($101|0))|0;
              $i$128$i$i = 0;$index$126$i$i = 0;$o$127$i$i = $96;
              while(1) {
               $124 = (($o$127$i$i) + -1)|0;
               $125 = (($scevgep$i) + ($124<<2)|0);
               $126 = HEAP32[$125>>2]|0;
               $127 = (($126) - ($90))|0;
               $128 = ($127|0)<($93|0);
               if ($128) {
                $129 = (($93) - ($127))|0;
                $130 = $129 << 1;
                $131 = (($130) + -1)|0;
                $136 = $131;
               } else {
                $132 = (($127) - ($93))|0;
                $133 = $132 << 1;
                $136 = $133;
               }
               $134 = Math_imul($index$126$i$i, $92)|0;
               $135 = ($136|0)<(0);
               $137 = ($136|0)>=($92|0);
               $138 = $137 ? $100 : $136;
               $139 = $135 ? 0 : $138;
               $140 = (($139) + ($134))|0;
               $141 = (($i$128$i$i) + 1)|0;
               $exitcond47$i$i = ($141|0)==($96|0);
               if ($exitcond47$i$i) {
                $index$2$i$i = $140;
                break;
               } else {
                $i$128$i$i = $141;$index$126$i$i = $140;$o$127$i$i = $124;
               }
              }
             } else {
              if (!($95)) {
               $index$2$i$i = 0;
               break;
              }
              $97 = $91 >> 1;
              $98 = (($97) - ($90))|0;
              $99 = (($92) + -1)|0;
              $i$033$i$i = 0;$index$031$i$i = 0;$o$032$i$i = $96;
              while(1) {
               $102 = (($o$032$i$i) + -1)|0;
               $103 = (($scevgep$i) + ($102<<2)|0);
               $104 = HEAP32[$103>>2]|0;
               $105 = (($98) + ($104))|0;
               $106 = (($105|0) / ($91|0))&-1;
               $107 = ($106|0)<($93|0);
               if ($107) {
                $108 = (($93) - ($106))|0;
                $109 = $108 << 1;
                $110 = (($109) + -1)|0;
                $115 = $110;
               } else {
                $111 = (($106) - ($93))|0;
                $112 = $111 << 1;
                $115 = $112;
               }
               $113 = Math_imul($index$031$i$i, $92)|0;
               $114 = ($115|0)<(0);
               $116 = ($115|0)>=($92|0);
               $117 = $116 ? $99 : $115;
               $118 = $114 ? 0 : $117;
               $119 = (($118) + ($113))|0;
               $120 = Math_imul($106, $91)|0;
               $121 = (($120) + ($90))|0;
               $122 = (($p$i$i) + ($102<<2)|0);
               HEAP32[$122>>2] = $121;
               $123 = (($i$033$i$i) + 1)|0;
               $exitcond48$i$i = ($123|0)==($96|0);
               if ($exitcond48$i$i) {
                $index$2$i$i = $119;
                break;
               } else {
                $i$033$i$i = $123;$index$031$i$i = $119;$o$032$i$i = $102;
               }
              }
             }
            } while(0);
            $142 = HEAP32[$87>>2]|0;
            $143 = ((($142)) + 8|0);
            $144 = HEAP32[$143>>2]|0;
            $145 = (($144) + ($index$2$i$i)|0);
            $146 = HEAP8[$145>>0]|0;
            $147 = ($146<<24>>24)<(1);
            if ($147) {
             ;HEAP32[$e$i$i>>2]=0|0;HEAP32[$e$i$i+4>>2]=0|0;HEAP32[$e$i$i+8>>2]=0|0;HEAP32[$e$i$i+12>>2]=0|0;HEAP32[$e$i$i+16>>2]=0|0;HEAP32[$e$i$i+20>>2]=0|0;HEAP32[$e$i$i+24>>2]=0|0;HEAP32[$e$i$i+28>>2]=0|0;
             $148 = (($92) + -1)|0;
             $149 = Math_imul($148, $91)|0;
             $150 = (($149) + ($90))|0;
             $151 = HEAP32[$88>>2]|0;
             $152 = ($151|0)>(0);
             L66: do {
              if ($152) {
               if ($95) {
                $best$019$us$i$i = -1;$i$220$us$i$i = 0;$index$318$us$i$i = $index$2$i$i;
               } else {
                $best$019$i$i = -1;$i$220$i$i = 0;$index$318$i$i = $index$2$i$i;
                while(1) {
                 $178 = (($144) + ($i$220$i$i)|0);
                 $179 = HEAP8[$178>>0]|0;
                 $180 = ($179<<24>>24)>(0);
                 do {
                  if ($180) {
                   $181 = ($best$019$i$i|0)==(-1);
                   $182 = ($best$019$i$i|0)>(0);
                   $or$cond$i$i = $181 | $182;
                   if (!($or$cond$i$i)) {
                    $best$2$i$i = $best$019$i$i;$index$5$i$i = $index$318$i$i;
                    break;
                   }
                   ;HEAP32[$p$i$i>>2]=HEAP32[$e$i$i>>2]|0;HEAP32[$p$i$i+4>>2]=HEAP32[$e$i$i+4>>2]|0;HEAP32[$p$i$i+8>>2]=HEAP32[$e$i$i+8>>2]|0;HEAP32[$p$i$i+12>>2]=HEAP32[$e$i$i+12>>2]|0;HEAP32[$p$i$i+16>>2]=HEAP32[$e$i$i+16>>2]|0;HEAP32[$p$i$i+20>>2]=HEAP32[$e$i$i+20>>2]|0;HEAP32[$p$i$i+24>>2]=HEAP32[$e$i$i+24>>2]|0;HEAP32[$p$i$i+28>>2]=HEAP32[$e$i$i+28>>2]|0;
                   $best$2$i$i = 0;$index$5$i$i = $i$220$i$i;
                  } else {
                   $best$2$i$i = $best$019$i$i;$index$5$i$i = $index$318$i$i;
                  }
                 } while(0);
                 $183 = HEAP32[$e$i$i>>2]|0;
                 $184 = ($183|0)<($150|0);
                 if ($184) {
                  $$lcssa$i$i = $e$i$i;$$lcssa6$i$i = $183;
                 } else {
                  $186 = $e$i$i;$j$112$i$i = 0;
                  while(1) {
                   $185 = (($j$112$i$i) + 1)|0;
                   HEAP32[$186>>2] = 0;
                   $187 = (($e$i$i) + ($185<<2)|0);
                   $188 = HEAP32[$187>>2]|0;
                   $189 = ($188|0)<($150|0);
                   if ($189) {
                    $$lcssa$i$i = $187;$$lcssa6$i$i = $188;
                    break;
                   } else {
                    $186 = $187;$j$112$i$i = $185;
                   }
                  }
                 }
                 $190 = ($$lcssa6$i$i|0)>(-1);
                 if ($190) {
                  $191 = (($$lcssa6$i$i) + ($91))|0;
                  HEAP32[$$lcssa$i$i>>2] = $191;
                  $193 = $191;
                 } else {
                  $193 = $$lcssa6$i$i;
                 }
                 $192 = (0 - ($193))|0;
                 HEAP32[$$lcssa$i$i>>2] = $192;
                 $194 = (($i$220$i$i) + 1)|0;
                 $exitcond$i = ($194|0)==($151|0);
                 if ($exitcond$i) {
                  $index$3$lcssa$i$i = $index$5$i$i;
                  break L66;
                 } else {
                  $best$019$i$i = $best$2$i$i;$i$220$i$i = $194;$index$318$i$i = $index$5$i$i;
                 }
                }
               }
               while(1) {
                $153 = (($144) + ($i$220$us$i$i)|0);
                $154 = HEAP8[$153>>0]|0;
                $155 = ($154<<24>>24)>(0);
                do {
                 if ($155) {
                  $j$09$us$i$i = 0;$this$010$us$i$i = 0;
                  while(1) {
                   $168 = (($e$i$i) + ($j$09$us$i$i<<2)|0);
                   $169 = HEAP32[$168>>2]|0;
                   $170 = (($scevgep$i) + ($j$09$us$i$i<<2)|0);
                   $171 = HEAP32[$170>>2]|0;
                   $172 = (($169) - ($171))|0;
                   $173 = Math_imul($172, $172)|0;
                   $174 = (($173) + ($this$010$us$i$i))|0;
                   $175 = (($j$09$us$i$i) + 1)|0;
                   $exitcond46$i$i = ($175|0)==($96|0);
                   if ($exitcond46$i$i) {
                    $$lcssa84 = $174;
                    break;
                   } else {
                    $j$09$us$i$i = $175;$this$010$us$i$i = $174;
                   }
                  }
                  $176 = ($best$019$us$i$i|0)==(-1);
                  $177 = ($$lcssa84|0)<($best$019$us$i$i|0);
                  $or$cond$us$i$i = $176 | $177;
                  if (!($or$cond$us$i$i)) {
                   $best$2$us$i$i = $best$019$us$i$i;$index$5$us$i$i = $index$318$us$i$i;
                   break;
                  }
                  ;HEAP32[$p$i$i>>2]=HEAP32[$e$i$i>>2]|0;HEAP32[$p$i$i+4>>2]=HEAP32[$e$i$i+4>>2]|0;HEAP32[$p$i$i+8>>2]=HEAP32[$e$i$i+8>>2]|0;HEAP32[$p$i$i+12>>2]=HEAP32[$e$i$i+12>>2]|0;HEAP32[$p$i$i+16>>2]=HEAP32[$e$i$i+16>>2]|0;HEAP32[$p$i$i+20>>2]=HEAP32[$e$i$i+20>>2]|0;HEAP32[$p$i$i+24>>2]=HEAP32[$e$i$i+24>>2]|0;HEAP32[$p$i$i+28>>2]=HEAP32[$e$i$i+28>>2]|0;
                  $best$2$us$i$i = $$lcssa84;$index$5$us$i$i = $i$220$us$i$i;
                 } else {
                  $best$2$us$i$i = $best$019$us$i$i;$index$5$us$i$i = $index$318$us$i$i;
                 }
                } while(0);
                $156 = HEAP32[$e$i$i>>2]|0;
                $157 = ($156|0)<($150|0);
                if ($157) {
                 $$lcssa$us$i$i = $e$i$i;$$lcssa6$us$i$i = $156;
                } else {
                 $159 = $e$i$i;$j$112$us$i$i = 0;
                 while(1) {
                  $158 = (($j$112$us$i$i) + 1)|0;
                  HEAP32[$159>>2] = 0;
                  $160 = (($e$i$i) + ($158<<2)|0);
                  $161 = HEAP32[$160>>2]|0;
                  $162 = ($161|0)<($150|0);
                  if ($162) {
                   $$lcssa$us$i$i = $160;$$lcssa6$us$i$i = $161;
                   break;
                  } else {
                   $159 = $160;$j$112$us$i$i = $158;
                  }
                 }
                }
                $163 = ($$lcssa6$us$i$i|0)>(-1);
                if ($163) {
                 $164 = (($$lcssa6$us$i$i) + ($91))|0;
                 HEAP32[$$lcssa$us$i$i>>2] = $164;
                 $166 = $164;
                } else {
                 $166 = $$lcssa6$us$i$i;
                }
                $165 = (0 - ($166))|0;
                HEAP32[$$lcssa$us$i$i>>2] = $165;
                $167 = (($i$220$us$i$i) + 1)|0;
                $exitcond20$i = ($167|0)==($151|0);
                if ($exitcond20$i) {
                 $index$3$lcssa$i$i = $index$5$us$i$i;
                 break;
                } else {
                 $best$019$us$i$i = $best$2$us$i$i;$i$220$us$i$i = $167;$index$318$us$i$i = $index$5$us$i$i;
                }
               }
              } else {
               $index$3$lcssa$i$i = $index$2$i$i;
              }
             } while(0);
             $index$6$i$i = $index$3$lcssa$i$i;
            } else {
             $index$6$i$i = $index$2$i$i;
            }
            $195 = ($index$6$i$i|0)>(-1);
            $or$cond36$i$i = $95 & $195;
            if ($or$cond36$i$i) {
             $$08$i$i = $scevgep$i;$i$37$i$i = 0;
             while(1) {
              $196 = (($p$i$i) + ($i$37$i$i<<2)|0);
              $197 = HEAP32[$196>>2]|0;
              $198 = ((($$08$i$i)) + 4|0);
              $199 = HEAP32[$$08$i$i>>2]|0;
              $200 = (($199) - ($197))|0;
              HEAP32[$$08$i$i>>2] = $200;
              $201 = (($i$37$i$i) + 1)|0;
              $exitcond$i$i = ($201|0)==($96|0);
              if ($exitcond$i$i) {
               break;
              } else {
               $$08$i$i = $198;$i$37$i$i = $201;
              }
             }
            }
            $202 = (_vorbis_book_encode($76,$index$6$i$i,$opb)|0);
            $203 = (($202) + ($bits$09$i))|0;
            $204 = (($i$010$i) + 1)|0;
            $exitcond21$i = ($204|0)==($82|0);
            if ($exitcond21$i) {
             $$lcssa87 = $203;
             break;
            }
            $$pre$i = HEAP32[$76>>2]|0;
            $96 = $$pre$i;$bits$09$i = $203;$i$010$i = $204;
           }
           $$pre46 = HEAP32[$61>>2]|0;
           $208 = $$pre46;$bits$0$lcssa$i = $$lcssa87;
          } else {
           $208 = $62;$bits$0$lcssa$i = 0;
          }
          $205 = HEAP32[$20>>2]|0;
          $206 = (($205) + ($bits$0$lcssa$i))|0;
          HEAP32[$20>>2] = $206;
          $207 = (($208) + ($i$121<<2)|0);
          $209 = HEAP32[$207>>2]|0;
          $210 = (($resbits) + ($209<<2)|0);
          $211 = HEAP32[$210>>2]|0;
          $212 = (($211) + ($bits$0$lcssa$i))|0;
          HEAP32[$210>>2] = $212;
         }
        }
        $213 = (($j$115) + 1)|0;
        $exitcond43 = ($213|0)==($ch|0);
        if ($exitcond43) {
         break;
        } else {
         $j$115 = $213;
        }
       }
      }
      $214 = (($k$120) + 1)|0;
      $215 = (($i$121) + 1)|0;
      $216 = ($214|0)<($7|0);
      $217 = ($215|0)<($12|0);
      $or$cond = $216 & $217;
      if ($or$cond) {
       $i$121 = $215;$k$120 = $214;
      } else {
       $i$1$lcssa = $215;
       break;
      }
     }
    } else {
     $i$1$lcssa = $i$024;
    }
    $25 = ($i$1$lcssa|0)<($12|0);
    if ($25) {
     $i$024 = $i$1$lcssa;
    } else {
     break;
    }
   }
   $$pre47 = HEAP32[$13>>2]|0;
   $220 = $$pre47;
  } else {
   $220 = $221;
  }
  $218 = (($s$026) + 1)|0;
  $219 = ($218|0)<($220|0);
  if ($219) {
   $221 = $220;$s$026 = $218;
  } else {
   break;
  }
 }
 STACKTOP = sp;return;
}
function _res1_inverse($vb,$vl,$in,$nonzero,$ch) {
 $vb = $vb|0;
 $vl = $vl|0;
 $in = $in|0;
 $nonzero = $nonzero|0;
 $ch = $ch|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, $used$02 = 0, $used$1 = 0, $used$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($ch|0)>(0);
 if ($0) {
  $i$01 = 0;$used$02 = 0;
 } else {
  return 0;
 }
 while(1) {
  $1 = (($nonzero) + ($i$01<<2)|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ($2|0)==(0);
  if ($3) {
   $used$1 = $used$02;
  } else {
   $4 = (($in) + ($i$01<<2)|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = (($used$02) + 1)|0;
   $7 = (($in) + ($used$02<<2)|0);
   HEAP32[$7>>2] = $5;
   $used$1 = $6;
  }
  $8 = (($i$01) + 1)|0;
  $exitcond = ($8|0)==($ch|0);
  if ($exitcond) {
   $used$1$lcssa = $used$1;
   break;
  } else {
   $i$01 = $8;$used$02 = $used$1;
  }
 }
 $9 = ($used$1$lcssa|0)==(0);
 if ($9) {
  return 0;
 }
 __01inverse($vb,$vl,$in,$used$1$lcssa,4452);
 return 0;
}
function _vorbis_book_decodev_add($book,$a,$b,$n) {
 $book = $book|0;
 $a = $a|0;
 $b = $b|0;
 $n = $n|0;
 var $$1 = 0, $$lcssa144 = 0, $$lcssa145 = 0, $$lcssa146 = 0, $$lcssa147 = 0, $$lcssa148 = 0, $$lcssa149 = 0, $$lcssa150 = 0, $$lcssa151 = 0, $$lcssa152 = 0, $$lcssa153 = 0, $$lcssa154 = 0, $$lcssa155 = 0, $$lcssa156 = 0, $$lcssa157 = 0, $$lcssa158 = 0, $$lcssa159 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0.0, $45 = 0, $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0.0, $53 = 0, $54 = 0, $55 = 0.0, $56 = 0.0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0.0;
 var $61 = 0, $62 = 0, $63 = 0.0, $64 = 0.0, $65 = 0, $66 = 0, $67 = 0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0.0, $73 = 0, $74 = 0, $75 = 0.0, $76 = 0, $77 = 0, $78 = 0.0, $79 = 0.0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0.0, $88 = 0, $89 = 0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $i$022 = 0, $i$1$lcssa = 0, $i$120 = 0, $i$2$ph123 = 0, $i$3 = 0, $i$4 = 0, $i$5 = 0, $i$6 = 0, $i$7 = 0, $i$8 = 0, $i$9 = 0, $j$021 = 0, $j$1 = 0, $j$2 = 0, $j$3 = 0, $j$4 = 0, $j$5 = 0, $j$6 = 0;
 var $j$7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($book)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(0);
 if (!($2)) {
  $$1 = 0;
  return ($$1|0);
 }
 $3 = HEAP32[$book>>2]|0;
 $4 = ($3|0)>(8);
 if ($4) {
  $7 = ($n|0)>(0);
  if (!($7)) {
   $$1 = 0;
   return ($$1|0);
  }
  $8 = ((($book)) + 16|0);
  $i$022 = 0;
  while(1) {
   $10 = (_decode_packed_entry_number($book,$b)|0);
   $11 = ($10|0)==(-1);
   if ($11) {
    $$1 = -1;
    label = 28;
    break;
   }
   $12 = HEAP32[$8>>2]|0;
   $13 = HEAP32[$book>>2]|0;
   $14 = Math_imul($13, $10)|0;
   $15 = (($12) + ($14<<2)|0);
   $16 = ($13|0)>(0);
   if ($16) {
    $i$120 = $i$022;$j$021 = 0;
    while(1) {
     $17 = (($j$021) + 1)|0;
     $18 = (($15) + ($j$021<<2)|0);
     $19 = +HEAPF32[$18>>2];
     $20 = (($i$120) + 1)|0;
     $21 = (($a) + ($i$120<<2)|0);
     $22 = +HEAPF32[$21>>2];
     $23 = $19 + $22;
     HEAPF32[$21>>2] = $23;
     $24 = ($17|0)<($13|0);
     if ($24) {
      $i$120 = $20;$j$021 = $17;
     } else {
      $i$1$lcssa = $20;
      break;
     }
    }
   } else {
    $i$1$lcssa = $i$022;
   }
   $9 = ($i$1$lcssa|0)<($n|0);
   if ($9) {
    $i$022 = $i$1$lcssa;
   } else {
    $$1 = 0;
    label = 28;
    break;
   }
  }
  if ((label|0) == 28) {
   return ($$1|0);
  }
 }
 $5 = ((($book)) + 16|0);
 $6 = ($n|0)>(0);
 if ($6) {
  $i$2$ph123 = 0;
 } else {
  $$1 = 0;
  return ($$1|0);
 }
 L22: while(1) {
  L24: while(1) {
   $81 = (_decode_packed_entry_number($book,$b)|0);
   $82 = ($81|0)==(-1);
   if ($82) {
    $$1 = -1;
    label = 28;
    break L22;
   }
   $83 = HEAP32[$5>>2]|0;
   $84 = HEAP32[$book>>2]|0;
   switch ($84|0) {
   case 8:  {
    $$lcssa144 = $81;$$lcssa152 = $83;
    label = 20;
    break L24;
    break;
   }
   case 7:  {
    $$lcssa145 = $81;$$lcssa153 = $83;
    label = 21;
    break L24;
    break;
   }
   case 6:  {
    $$lcssa146 = $81;$$lcssa154 = $83;
    label = 22;
    break L24;
    break;
   }
   case 5:  {
    $$lcssa147 = $81;$$lcssa155 = $83;
    label = 23;
    break L24;
    break;
   }
   case 4:  {
    $$lcssa148 = $81;$$lcssa156 = $83;
    label = 24;
    break L24;
    break;
   }
   case 3:  {
    $$lcssa149 = $81;$$lcssa157 = $83;
    label = 25;
    break L24;
    break;
   }
   case 2:  {
    $$lcssa150 = $81;$$lcssa158 = $83;
    label = 26;
    break L24;
    break;
   }
   case 1:  {
    $$lcssa151 = $81;$$lcssa159 = $83;
    label = 27;
    break L24;
    break;
   }
   default: {
   }
   }
  }
  if ((label|0) == 20) {
   label = 0;
   $85 = $$lcssa144 << 3;
   $86 = (($$lcssa152) + ($85<<2)|0);
   $87 = +HEAPF32[$86>>2];
   $88 = (($i$2$ph123) + 1)|0;
   $89 = (($a) + ($i$2$ph123<<2)|0);
   $90 = +HEAPF32[$89>>2];
   $91 = $87 + $90;
   HEAPF32[$89>>2] = $91;
   $27 = $86;$i$3 = $88;$j$1 = 1;
   label = 10;
  }
  else if ((label|0) == 21) {
   label = 0;
   $92 = ($$lcssa145*7)|0;
   $93 = (($$lcssa153) + ($92<<2)|0);
   $27 = $93;$i$3 = $i$2$ph123;$j$1 = 0;
   label = 10;
  }
  else if ((label|0) == 22) {
   label = 0;
   $94 = ($$lcssa146*6)|0;
   $95 = (($$lcssa154) + ($94<<2)|0);
   $35 = $95;$i$4 = $i$2$ph123;$j$2 = 0;
   label = 11;
  }
  else if ((label|0) == 23) {
   label = 0;
   $96 = ($$lcssa147*5)|0;
   $97 = (($$lcssa155) + ($96<<2)|0);
   $43 = $97;$i$5 = $i$2$ph123;$j$3 = 0;
   label = 12;
  }
  else if ((label|0) == 24) {
   label = 0;
   $98 = $$lcssa148 << 2;
   $99 = (($$lcssa156) + ($98<<2)|0);
   $51 = $99;$i$6 = $i$2$ph123;$j$4 = 0;
   label = 13;
  }
  else if ((label|0) == 25) {
   label = 0;
   $100 = ($$lcssa149*3)|0;
   $101 = (($$lcssa157) + ($100<<2)|0);
   $59 = $101;$i$7 = $i$2$ph123;$j$5 = 0;
   label = 14;
  }
  else if ((label|0) == 26) {
   label = 0;
   $102 = $$lcssa150 << 1;
   $103 = (($$lcssa158) + ($102<<2)|0);
   $67 = $103;$i$8 = $i$2$ph123;$j$6 = 0;
   label = 15;
  }
  else if ((label|0) == 27) {
   label = 0;
   $104 = (($$lcssa159) + ($$lcssa151<<2)|0);
   $74 = $104;$i$9 = $i$2$ph123;$j$7 = 0;
  }
  if ((label|0) == 10) {
   label = 0;
   $25 = (($j$1) + 1)|0;
   $26 = (($27) + ($j$1<<2)|0);
   $28 = +HEAPF32[$26>>2];
   $29 = (($i$3) + 1)|0;
   $30 = (($a) + ($i$3<<2)|0);
   $31 = +HEAPF32[$30>>2];
   $32 = $28 + $31;
   HEAPF32[$30>>2] = $32;
   $35 = $27;$i$4 = $29;$j$2 = $25;
   label = 11;
  }
  if ((label|0) == 11) {
   label = 0;
   $33 = (($j$2) + 1)|0;
   $34 = (($35) + ($j$2<<2)|0);
   $36 = +HEAPF32[$34>>2];
   $37 = (($i$4) + 1)|0;
   $38 = (($a) + ($i$4<<2)|0);
   $39 = +HEAPF32[$38>>2];
   $40 = $36 + $39;
   HEAPF32[$38>>2] = $40;
   $43 = $35;$i$5 = $37;$j$3 = $33;
   label = 12;
  }
  if ((label|0) == 12) {
   label = 0;
   $41 = (($j$3) + 1)|0;
   $42 = (($43) + ($j$3<<2)|0);
   $44 = +HEAPF32[$42>>2];
   $45 = (($i$5) + 1)|0;
   $46 = (($a) + ($i$5<<2)|0);
   $47 = +HEAPF32[$46>>2];
   $48 = $44 + $47;
   HEAPF32[$46>>2] = $48;
   $51 = $43;$i$6 = $45;$j$4 = $41;
   label = 13;
  }
  if ((label|0) == 13) {
   label = 0;
   $49 = (($j$4) + 1)|0;
   $50 = (($51) + ($j$4<<2)|0);
   $52 = +HEAPF32[$50>>2];
   $53 = (($i$6) + 1)|0;
   $54 = (($a) + ($i$6<<2)|0);
   $55 = +HEAPF32[$54>>2];
   $56 = $52 + $55;
   HEAPF32[$54>>2] = $56;
   $59 = $51;$i$7 = $53;$j$5 = $49;
   label = 14;
  }
  if ((label|0) == 14) {
   label = 0;
   $57 = (($j$5) + 1)|0;
   $58 = (($59) + ($j$5<<2)|0);
   $60 = +HEAPF32[$58>>2];
   $61 = (($i$7) + 1)|0;
   $62 = (($a) + ($i$7<<2)|0);
   $63 = +HEAPF32[$62>>2];
   $64 = $60 + $63;
   HEAPF32[$62>>2] = $64;
   $67 = $59;$i$8 = $61;$j$6 = $57;
   label = 15;
  }
  if ((label|0) == 15) {
   label = 0;
   $65 = (($j$6) + 1)|0;
   $66 = (($67) + ($j$6<<2)|0);
   $68 = +HEAPF32[$66>>2];
   $69 = (($i$8) + 1)|0;
   $70 = (($a) + ($i$8<<2)|0);
   $71 = +HEAPF32[$70>>2];
   $72 = $68 + $71;
   HEAPF32[$70>>2] = $72;
   $74 = $67;$i$9 = $69;$j$7 = $65;
  }
  $73 = (($74) + ($j$7<<2)|0);
  $75 = +HEAPF32[$73>>2];
  $76 = (($i$9) + 1)|0;
  $77 = (($a) + ($i$9<<2)|0);
  $78 = +HEAPF32[$77>>2];
  $79 = $75 + $78;
  HEAPF32[$77>>2] = $79;
  $80 = ($76|0)<($n|0);
  if ($80) {
   $i$2$ph123 = $76;
  } else {
   $$1 = 0;
   label = 28;
   break;
  }
 }
 if ((label|0) == 28) {
  return ($$1|0);
 }
 return (0)|0;
}
function _res2_class($vb,$vl,$in,$nonzero,$ch) {
 $vb = $vb|0;
 $vl = $vl|0;
 $in = $in|0;
 $nonzero = $nonzero|0;
 $ch = $ch|0;
 var $$0 = 0, $$angmax$1$us$us$i$us = 0, $$angmax$1$us$us$i$us$lcssa = 0, $$angmax$1$us$us$i$us$lcssa$lcssa = 0, $$lcssa = 0, $$lcssa19 = 0, $$magmax$0$us$us$i$us = 0, $$magmax$0$us$us$i$us$lcssa = 0, $$magmax$0$us29$i = 0, $$magmax$0$us29$i$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $9 = 0, $angmax$09$us$us$i$us = 0, $angmax$15$us$us$i$us = 0, $exitcond = 0, $exitcond$i$us = 0, $exitcond47$i = 0, $exitcond47$i$us = 0, $exitcond48$i$us = 0;
 var $i$021$i$us = 0, $i$021$us$i = 0, $i$021$us$i$us = 0, $i$04 = 0, $ispos$us$us$i$us = 0, $ispos$us27$i = 0, $ispos1$us$us$i$us = 0, $j$07$us$us$i$us = 0, $j$07$us25$i = 0, $j$1$lcssa$i$ph$us = 0, $j$1$lcssa$us$i = 0, $j$1$lcssa$us$i$us = 0, $j$114$i$us = 0, $j$114$us$i = 0, $j$114$us$i$us = 0, $k$04$us$us$i$us = 0, $l$020$us$i = 0, $l$020$us$i$us = 0, $l$16$us$us$i$us = 0, $l$16$us26$i = 0;
 var $magmax$08$us$us$i$us = 0, $magmax$08$us24$i = 0, $neg$us$us$i$us = 0, $neg$us28$i = 0, $neg2$us$us$i$us = 0, $not$ = 0, $used$0$ = 0, $used$0$$lcssa = 0, $used$05 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($ch|0)>(0);
 if ($0) {
  $i$04 = 0;$used$05 = 0;
 } else {
  $$0 = 0;
  return ($$0|0);
 }
 while(1) {
  $1 = (($nonzero) + ($i$04<<2)|0);
  $2 = HEAP32[$1>>2]|0;
  $not$ = ($2|0)!=(0);
  $3 = $not$&1;
  $used$0$ = (($3) + ($used$05))|0;
  $4 = (($i$04) + 1)|0;
  $exitcond = ($4|0)==($ch|0);
  if ($exitcond) {
   $used$0$$lcssa = $used$0$;
   break;
  } else {
   $i$04 = $4;$used$05 = $used$0$;
  }
 }
 $5 = ($used$0$$lcssa|0)==(0);
 if ($5) {
  $$0 = 0;
  return ($$0|0);
 }
 $6 = HEAP32[$vl>>2]|0;
 $7 = ((($6)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($6)) + 12|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($6)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = HEAP32[$6>>2]|0;
 $14 = (($12) - ($13))|0;
 $15 = (($14|0) / ($8|0))&-1;
 $16 = (__vorbis_block_alloc($vb,4)|0);
 $17 = $15 << 2;
 $18 = (__vorbis_block_alloc($vb,$17)|0);
 HEAP32[$16>>2] = $18;
 _memset(($18|0),0,($17|0))|0;
 $19 = ($15|0)>(0);
 L9: do {
  if ($19) {
   $20 = ($8|0)>(0);
   $21 = (($10) + -1)|0;
   $22 = ($10|0)>(1);
   $23 = HEAP32[$16>>2]|0;
   if (!($20)) {
    if ($22) {
     $i$021$i$us = 0;
    } else {
     _memset(($23|0),0,($17|0))|0;
     break;
    }
    while(1) {
     $j$114$i$us = 0;
     while(1) {
      $24 = (((($6)) + 2328|0) + ($j$114$i$us<<2)|0);
      $25 = HEAP32[$24>>2]|0;
      $26 = ($25|0)<(0);
      if (!($26)) {
       $27 = (((($6)) + 2584|0) + ($j$114$i$us<<2)|0);
       $28 = HEAP32[$27>>2]|0;
       $29 = ($28|0)<(0);
       if (!($29)) {
        $j$1$lcssa$i$ph$us = $j$114$i$us;
        break;
       }
      }
      $30 = (($j$114$i$us) + 1)|0;
      $31 = ($30|0)<($21|0);
      if ($31) {
       $j$114$i$us = $30;
      } else {
       $j$1$lcssa$i$ph$us = $30;
       break;
      }
     }
     $32 = (($23) + ($i$021$i$us<<2)|0);
     HEAP32[$32>>2] = $j$1$lcssa$i$ph$us;
     $33 = (($i$021$i$us) + 1)|0;
     $exitcond48$i$us = ($33|0)==($15|0);
     if ($exitcond48$i$us) {
      break L9;
     } else {
      $i$021$i$us = $33;
     }
    }
   }
   $34 = ($ch|0)>(1);
   $35 = HEAP32[$6>>2]|0;
   $36 = (($35|0) / ($ch|0))&-1;
   if ($34) {
    $i$021$us$i$us = 0;$l$020$us$i$us = $36;
   } else {
    $i$021$us$i = 0;$l$020$us$i = $36;
    while(1) {
     $73 = HEAP32[$in>>2]|0;
     $j$07$us25$i = 0;$l$16$us26$i = $l$020$us$i;$magmax$08$us24$i = 0;
     while(1) {
      $72 = (($73) + ($l$16$us26$i<<2)|0);
      $74 = HEAP32[$72>>2]|0;
      $ispos$us27$i = ($74|0)>(-1);
      $neg$us28$i = (0 - ($74))|0;
      $75 = $ispos$us27$i ? $74 : $neg$us28$i;
      $76 = ($75|0)>($magmax$08$us24$i|0);
      $$magmax$0$us29$i = $76 ? $75 : $magmax$08$us24$i;
      $77 = (($l$16$us26$i) + 1)|0;
      $78 = (($j$07$us25$i) + ($ch))|0;
      $79 = ($78|0)<($8|0);
      if ($79) {
       $j$07$us25$i = $78;$l$16$us26$i = $77;$magmax$08$us24$i = $$magmax$0$us29$i;
      } else {
       $$lcssa19 = $77;$$magmax$0$us29$i$lcssa = $$magmax$0$us29$i;
       break;
      }
     }
     L30: do {
      if ($22) {
       $j$114$us$i = 0;
       while(1) {
        $62 = (((($6)) + 2328|0) + ($j$114$us$i<<2)|0);
        $63 = HEAP32[$62>>2]|0;
        $64 = ($$magmax$0$us29$i$lcssa|0)>($63|0);
        if (!($64)) {
         $65 = (((($6)) + 2584|0) + ($j$114$us$i<<2)|0);
         $66 = HEAP32[$65>>2]|0;
         $67 = ($66|0)<(0);
         if (!($67)) {
          $j$1$lcssa$us$i = $j$114$us$i;
          break L30;
         }
        }
        $70 = (($j$114$us$i) + 1)|0;
        $71 = ($70|0)<($21|0);
        if ($71) {
         $j$114$us$i = $70;
        } else {
         $j$1$lcssa$us$i = $70;
         break;
        }
       }
      } else {
       $j$1$lcssa$us$i = 0;
      }
     } while(0);
     $68 = (($23) + ($i$021$us$i<<2)|0);
     HEAP32[$68>>2] = $j$1$lcssa$us$i;
     $69 = (($i$021$us$i) + 1)|0;
     $exitcond47$i = ($69|0)==($15|0);
     if ($exitcond47$i) {
      break L9;
     } else {
      $i$021$us$i = $69;$l$020$us$i = $$lcssa19;
     }
    }
   }
   while(1) {
    $37 = HEAP32[$in>>2]|0;
    $angmax$09$us$us$i$us = 0;$j$07$us$us$i$us = 0;$l$16$us$us$i$us = $l$020$us$i$us;$magmax$08$us$us$i$us = 0;
    while(1) {
     $38 = (($37) + ($l$16$us$us$i$us<<2)|0);
     $39 = HEAP32[$38>>2]|0;
     $ispos$us$us$i$us = ($39|0)>(-1);
     $neg$us$us$i$us = (0 - ($39))|0;
     $40 = $ispos$us$us$i$us ? $39 : $neg$us$us$i$us;
     $angmax$15$us$us$i$us = $angmax$09$us$us$i$us;$k$04$us$us$i$us = 1;
     while(1) {
      $41 = (($in) + ($k$04$us$us$i$us<<2)|0);
      $42 = HEAP32[$41>>2]|0;
      $43 = (($42) + ($l$16$us$us$i$us<<2)|0);
      $44 = HEAP32[$43>>2]|0;
      $ispos1$us$us$i$us = ($44|0)>(-1);
      $neg2$us$us$i$us = (0 - ($44))|0;
      $45 = $ispos1$us$us$i$us ? $44 : $neg2$us$us$i$us;
      $46 = ($45|0)>($angmax$15$us$us$i$us|0);
      $$angmax$1$us$us$i$us = $46 ? $45 : $angmax$15$us$us$i$us;
      $47 = (($k$04$us$us$i$us) + 1)|0;
      $exitcond$i$us = ($47|0)==($ch|0);
      if ($exitcond$i$us) {
       $$angmax$1$us$us$i$us$lcssa = $$angmax$1$us$us$i$us;
       break;
      } else {
       $angmax$15$us$us$i$us = $$angmax$1$us$us$i$us;$k$04$us$us$i$us = $47;
      }
     }
     $48 = ($40|0)>($magmax$08$us$us$i$us|0);
     $$magmax$0$us$us$i$us = $48 ? $40 : $magmax$08$us$us$i$us;
     $49 = (($l$16$us$us$i$us) + 1)|0;
     $50 = (($j$07$us$us$i$us) + ($ch))|0;
     $51 = ($50|0)<($8|0);
     if ($51) {
      $angmax$09$us$us$i$us = $$angmax$1$us$us$i$us$lcssa;$j$07$us$us$i$us = $50;$l$16$us$us$i$us = $49;$magmax$08$us$us$i$us = $$magmax$0$us$us$i$us;
     } else {
      $$angmax$1$us$us$i$us$lcssa$lcssa = $$angmax$1$us$us$i$us$lcssa;$$lcssa = $49;$$magmax$0$us$us$i$us$lcssa = $$magmax$0$us$us$i$us;
      break;
     }
    }
    L45: do {
     if ($22) {
      $j$114$us$i$us = 0;
      while(1) {
       $52 = (((($6)) + 2328|0) + ($j$114$us$i$us<<2)|0);
       $53 = HEAP32[$52>>2]|0;
       $54 = ($$magmax$0$us$us$i$us$lcssa|0)>($53|0);
       if (!($54)) {
        $55 = (((($6)) + 2584|0) + ($j$114$us$i$us<<2)|0);
        $56 = HEAP32[$55>>2]|0;
        $57 = ($$angmax$1$us$us$i$us$lcssa$lcssa|0)>($56|0);
        if (!($57)) {
         $j$1$lcssa$us$i$us = $j$114$us$i$us;
         break L45;
        }
       }
       $58 = (($j$114$us$i$us) + 1)|0;
       $59 = ($58|0)<($21|0);
       if ($59) {
        $j$114$us$i$us = $58;
       } else {
        $j$1$lcssa$us$i$us = $58;
        break;
       }
      }
     } else {
      $j$1$lcssa$us$i$us = 0;
     }
    } while(0);
    $60 = (($23) + ($i$021$us$i$us<<2)|0);
    HEAP32[$60>>2] = $j$1$lcssa$us$i$us;
    $61 = (($i$021$us$i$us) + 1)|0;
    $exitcond47$i$us = ($61|0)==($15|0);
    if ($exitcond47$i$us) {
     break;
    } else {
     $i$021$us$i$us = $61;$l$020$us$i$us = $$lcssa;
    }
   }
  }
 } while(0);
 $80 = ((($vl)) + 40|0);
 $81 = HEAP32[$80>>2]|0;
 $82 = (($81) + 1)|0;
 HEAP32[$80>>2] = $82;
 $$0 = $16;
 return ($$0|0);
}
function _res2_forward($opb,$vb,$vl,$in,$nonzero,$ch,$partword,$submap) {
 $opb = $opb|0;
 $vb = $vb|0;
 $vl = $vl|0;
 $in = $in|0;
 $nonzero = $nonzero|0;
 $ch = $ch|0;
 $partword = $partword|0;
 $submap = $submap|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $exitcond9 = 0, $i$03 = 0, $i$03$us = 0, $j$01$us = 0, $k$02$us = 0, $not$ = 0, $not$$us = 0, $used$0$ = 0, $used$0$$us = 0, $used$0$lcssa = 0, $used$04 = 0, $used$04$us = 0, $work = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $work = sp;
 $0 = ((($vb)) + 36|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1|0) / 2)&-1;
 $3 = $ch << 2;
 $4 = Math_imul($3, $2)|0;
 $5 = (__vorbis_block_alloc($vb,$4)|0);
 HEAP32[$work>>2] = $5;
 $6 = ($ch|0)>(0);
 if (!($6)) {
  STACKTOP = sp;return 0;
 }
 $7 = ($1|0)>(1);
 if ($7) {
  $i$03$us = 0;$used$04$us = 0;
  while(1) {
   $8 = (($in) + ($i$03$us<<2)|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (($nonzero) + ($i$03$us<<2)|0);
   $11 = HEAP32[$10>>2]|0;
   $not$$us = ($11|0)!=(0);
   $12 = $not$$us&1;
   $j$01$us = 0;$k$02$us = $i$03$us;
   while(1) {
    $13 = (($9) + ($j$01$us<<2)|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = (($5) + ($k$02$us<<2)|0);
    HEAP32[$15>>2] = $14;
    $16 = (($j$01$us) + 1)|0;
    $17 = (($k$02$us) + ($ch))|0;
    $18 = ($16|0)<($2|0);
    if ($18) {
     $j$01$us = $16;$k$02$us = $17;
    } else {
     break;
    }
   }
   $used$0$$us = (($12) + ($used$04$us))|0;
   $19 = (($i$03$us) + 1)|0;
   $exitcond = ($19|0)==($ch|0);
   if ($exitcond) {
    $used$0$lcssa = $used$0$$us;
    break;
   } else {
    $i$03$us = $19;$used$04$us = $used$0$$us;
   }
  }
 } else {
  $i$03 = 0;$used$04 = 0;
  while(1) {
   $20 = (($nonzero) + ($i$03<<2)|0);
   $21 = HEAP32[$20>>2]|0;
   $not$ = ($21|0)!=(0);
   $22 = $not$&1;
   $used$0$ = (($22) + ($used$04))|0;
   $23 = (($i$03) + 1)|0;
   $exitcond9 = ($23|0)==($ch|0);
   if ($exitcond9) {
    $used$0$lcssa = $used$0$;
    break;
   } else {
    $i$03 = $23;$used$04 = $used$0$;
   }
  }
 }
 $24 = ($used$0$lcssa|0)==(0);
 if ($24) {
  STACKTOP = sp;return 0;
 }
 __01forward($opb,$vl,$work,1,$partword);
 STACKTOP = sp;return 0;
}
function _res2_inverse($vb,$vl,$in,$nonzero,$ch) {
 $vb = $vb|0;
 $vl = $vl|0;
 $in = $in|0;
 $nonzero = $nonzero|0;
 $ch = $ch|0;
 var $$ = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$0$lcssa = 0, $i$021 = 0, $i$117 = 0, $i$117$us = 0, $i$2$lcssa = 0, $i$2$lcssa$us = 0, $i$215 = 0, $i$215$us = 0, $k$014 = 0, $k$014$us = 0, $l$016 = 0, $l$016$us = 0, $or$cond = 0, $or$cond$us = 0, $or$cond13 = 0;
 var $or$cond13$us = 0, $s$019 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$vl>>2]|0;
 $1 = ((($0)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($vl)) + 16|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($vb)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = Math_imul($7, $ch)|0;
 $9 = $8 >> 1;
 $10 = ((($0)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)<($9|0);
 $$ = $12 ? $11 : $9;
 $13 = HEAP32[$0>>2]|0;
 $14 = (($$) - ($13))|0;
 $15 = ($14|0)>(0);
 if (!($15)) {
  return 0;
 }
 $16 = (($14|0) / ($2|0))&-1;
 $17 = (($5) + -1)|0;
 $18 = (($17) + ($16))|0;
 $19 = (($18|0) / ($5|0))&-1;
 $20 = $19 << 2;
 $21 = (__vorbis_block_alloc($vb,$20)|0);
 $22 = ($ch|0)>(0);
 L4: do {
  if ($22) {
   $i$021 = 0;
   while(1) {
    $23 = (($nonzero) + ($i$021<<2)|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = ($24|0)==(0);
    if (!($25)) {
     $i$0$lcssa = $i$021;
     break L4;
    }
    $26 = (($i$021) + 1)|0;
    $27 = ($26|0)<($ch|0);
    if ($27) {
     $i$021 = $26;
    } else {
     $i$0$lcssa = $26;
     break;
    }
   }
  } else {
   $i$0$lcssa = 0;
  }
 } while(0);
 $28 = ($i$0$lcssa|0)==($ch|0);
 if ($28) {
  return 0;
 }
 $29 = ((($vl)) + 8|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ($30|0)>(0);
 if (!($31)) {
  return 0;
 }
 $32 = ($16|0)>(0);
 $33 = ((($vb)) + 4|0);
 $34 = ((($0)) + 16|0);
 $35 = ((($vl)) + 28|0);
 $36 = ($5|0)>(0);
 $37 = ((($vl)) + 20|0);
 $s$019 = 0;
 L15: while(1) {
  L17: do {
   if ($32) {
    $38 = ($s$019|0)==(0);
    $39 = 1 << $s$019;
    if ($38) {
     $i$117$us = 0;$l$016$us = 0;
    } else {
     $i$117 = 0;$l$016 = 0;
     while(1) {
      $74 = ($i$117|0)<($16|0);
      $or$cond13 = $36 & $74;
      if ($or$cond13) {
       $75 = (($21) + ($l$016<<2)|0);
       $i$215 = $i$117;$k$014 = 0;
       while(1) {
        $76 = HEAP32[$75>>2]|0;
        $77 = (($76) + ($k$014<<2)|0);
        $78 = HEAP32[$77>>2]|0;
        $79 = (((($0)) + 24|0) + ($78<<2)|0);
        $80 = HEAP32[$79>>2]|0;
        $81 = $80 & $39;
        $82 = ($81|0)==(0);
        if (!($82)) {
         $83 = HEAP32[$37>>2]|0;
         $84 = (($83) + ($78<<2)|0);
         $85 = HEAP32[$84>>2]|0;
         $86 = (($85) + ($s$019<<2)|0);
         $87 = HEAP32[$86>>2]|0;
         $88 = ($87|0)==(0|0);
         if (!($88)) {
          $89 = Math_imul($i$215, $2)|0;
          $90 = HEAP32[$0>>2]|0;
          $91 = (($90) + ($89))|0;
          $92 = (_vorbis_book_decodevv_add($87,$in,$91,$ch,$33,$2)|0);
          $93 = ($92|0)==(-1);
          if ($93) {
           label = 28;
           break L15;
          }
         }
        }
        $94 = (($k$014) + 1)|0;
        $95 = (($i$215) + 1)|0;
        $96 = ($94|0)<($5|0);
        $97 = ($95|0)<($16|0);
        $or$cond = $96 & $97;
        if ($or$cond) {
         $i$215 = $95;$k$014 = $94;
        } else {
         $i$2$lcssa = $95;
         break;
        }
       }
      } else {
       $i$2$lcssa = $i$117;
      }
      $98 = (($l$016) + 1)|0;
      $99 = ($i$2$lcssa|0)<($16|0);
      if ($99) {
       $i$117 = $i$2$lcssa;$l$016 = $98;
      } else {
       break L17;
      }
     }
    }
    while(1) {
     $40 = HEAP32[$3>>2]|0;
     $41 = (_vorbis_book_decode($40,$33)|0);
     $42 = ($41|0)==(-1);
     if ($42) {
      label = 28;
      break L15;
     }
     $43 = HEAP32[$34>>2]|0;
     $44 = ($41|0)<($43|0);
     if (!($44)) {
      label = 28;
      break L15;
     }
     $45 = HEAP32[$35>>2]|0;
     $46 = (($45) + ($41<<2)|0);
     $47 = HEAP32[$46>>2]|0;
     $48 = (($21) + ($l$016$us<<2)|0);
     HEAP32[$48>>2] = $47;
     $49 = ($47|0)==(0|0);
     if ($49) {
      label = 28;
      break L15;
     }
     $73 = ($i$117$us|0)<($16|0);
     $or$cond13$us = $36 & $73;
     L36: do {
      if ($or$cond13$us) {
       $53 = $47;$i$215$us = $i$117$us;$k$014$us = 0;
       while(1) {
        $52 = (($53) + ($k$014$us<<2)|0);
        $54 = HEAP32[$52>>2]|0;
        $55 = (((($0)) + 24|0) + ($54<<2)|0);
        $56 = HEAP32[$55>>2]|0;
        $57 = $56 & $39;
        $58 = ($57|0)==(0);
        if (!($58)) {
         $59 = HEAP32[$37>>2]|0;
         $60 = (($59) + ($54<<2)|0);
         $61 = HEAP32[$60>>2]|0;
         $62 = HEAP32[$61>>2]|0;
         $63 = ($62|0)==(0|0);
         if (!($63)) {
          $64 = Math_imul($i$215$us, $2)|0;
          $65 = HEAP32[$0>>2]|0;
          $66 = (($65) + ($64))|0;
          $67 = (_vorbis_book_decodevv_add($62,$in,$66,$ch,$33,$2)|0);
          $68 = ($67|0)==(-1);
          if ($68) {
           label = 28;
           break L15;
          }
         }
        }
        $69 = (($k$014$us) + 1)|0;
        $70 = (($i$215$us) + 1)|0;
        $71 = ($69|0)<($5|0);
        $72 = ($70|0)<($16|0);
        $or$cond$us = $71 & $72;
        if (!($or$cond$us)) {
         $i$2$lcssa$us = $70;
         break L36;
        }
        $$pre = HEAP32[$48>>2]|0;
        $53 = $$pre;$i$215$us = $70;$k$014$us = $69;
       }
      } else {
       $i$2$lcssa$us = $i$117$us;
      }
     } while(0);
     $50 = (($l$016$us) + 1)|0;
     $51 = ($i$2$lcssa$us|0)<($16|0);
     if ($51) {
      $i$117$us = $i$2$lcssa$us;$l$016$us = $50;
     } else {
      break;
     }
    }
   }
  } while(0);
  $100 = (($s$019) + 1)|0;
  $101 = HEAP32[$29>>2]|0;
  $102 = ($100|0)<($101|0);
  if ($102) {
   $s$019 = $100;
  } else {
   label = 28;
   break;
  }
 }
 if ((label|0) == 28) {
  return 0;
 }
 return (0)|0;
}
function _vorbis_book_decodevv_add($book,$a,$offset,$ch,$b,$n) {
 $book = $book|0;
 $a = $a|0;
 $offset = $offset|0;
 $ch = $ch|0;
 $b = $b|0;
 $n = $n|0;
 var $$ = 0, $$0 = 0, $$i$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0.0, $23 = 0.0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $chptr$07 = 0, $chptr$1$lcssa = 0, $chptr$14 = 0, $i$06 = 0, $i$1$lcssa = 0, $i$13 = 0, $j$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($book)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(0);
 if (!($2)) {
  $$0 = 0;
  return ($$0|0);
 }
 $3 = (($offset|0) / ($ch|0))&-1;
 $4 = (($n) + ($offset))|0;
 $5 = (($4|0) / ($ch|0))&-1;
 $6 = ($3|0)<($5|0);
 if (!($6)) {
  $$0 = 0;
  return ($$0|0);
 }
 $7 = ((($book)) + 16|0);
 $chptr$07 = 0;$i$06 = $3;
 while(1) {
  $9 = (_decode_packed_entry_number($book,$b)|0);
  $10 = ($9|0)==(-1);
  if ($10) {
   $$0 = -1;
   label = 8;
   break;
  }
  $11 = HEAP32[$7>>2]|0;
  $12 = HEAP32[$book>>2]|0;
  $13 = Math_imul($12, $9)|0;
  $14 = (($11) + ($13<<2)|0);
  $15 = ($12|0)>(0);
  if ($15) {
   $chptr$14 = $chptr$07;$i$13 = $i$06;$j$02 = 0;
   while(1) {
    $16 = (($14) + ($j$02<<2)|0);
    $17 = +HEAPF32[$16>>2];
    $18 = (($chptr$14) + 1)|0;
    $19 = (($a) + ($chptr$14<<2)|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = (($20) + ($i$13<<2)|0);
    $22 = +HEAPF32[$21>>2];
    $23 = $17 + $22;
    HEAPF32[$21>>2] = $23;
    $24 = ($18|0)==($ch|0);
    $25 = $24&1;
    $$i$1 = (($25) + ($i$13))|0;
    $$ = $24 ? 0 : $18;
    $26 = (($j$02) + 1)|0;
    $27 = ($26|0)<($12|0);
    if ($27) {
     $chptr$14 = $$;$i$13 = $$i$1;$j$02 = $26;
    } else {
     $chptr$1$lcssa = $$;$i$1$lcssa = $$i$1;
     break;
    }
   }
  } else {
   $chptr$1$lcssa = $chptr$07;$i$1$lcssa = $i$06;
  }
  $8 = ($i$1$lcssa|0)<($5|0);
  if ($8) {
   $chptr$07 = $chptr$1$lcssa;$i$06 = $i$1$lcssa;
  } else {
   $$0 = 0;
   label = 8;
   break;
  }
 }
 if ((label|0) == 8) {
  return ($$0|0);
 }
 return (0)|0;
}
function _mapping0_inverse($vb,$l) {
 $vb = $vb|0;
 $l = $l|0;
 var $$ = 0, $$2 = 0, $$alloca_mul = 0, $$alloca_mul1 = 0, $$alloca_mul3 = 0, $$alloca_mul5 = 0, $$lcssa = 0, $$pre32 = 0, $$pre33 = 0, $$pre34 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0.0, $118 = 0, $119 = 0.0, $12 = 0, $120 = 0, $121 = 0, $122 = 0.0, $123 = 0.0, $124 = 0.0;
 var $125 = 0.0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $ch_in_bundle$0$lcssa = 0, $ch_in_bundle$018 = 0;
 var $ch_in_bundle$1 = 0, $i$027 = 0, $i$124 = 0, $i$222 = 0, $i$315$us = 0, $i$315$us$in = 0, $i$49 = 0, $i$58 = 0, $j$019 = 0, $j$111$us = 0, $not$ = 0, $not$3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($vb)) + 64|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 28|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($1)) + 104|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($vb)) + 28|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (($5) + ($9<<2)|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($vb)) + 36|0);
 HEAP32[$12>>2] = $11;
 $13 = ((($3)) + 4|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = $14 << 2;
 $$alloca_mul = $15;
 $16 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $$alloca_mul1 = $15;
 $17 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul1)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $$alloca_mul3 = $15;
 $18 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul3)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $$alloca_mul5 = $15;
 $19 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul5)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $20 = ($14|0)>(0);
 if ($20) {
  $21 = ((($l)) + 4|0);
  $22 = ((($l)) + 1028|0);
  $23 = ((($7)) + 48|0);
  $24 = $11 << 1;
  $25 = $24 & 2147483646;
  $i$027 = 0;
  while(1) {
   $31 = (($21) + ($i$027<<2)|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = (($22) + ($32<<2)|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = (((($5)) + 800|0) + ($34<<2)|0);
   $36 = HEAP32[$35>>2]|0;
   $37 = (101372 + ($36<<2)|0);
   $38 = HEAP32[$37>>2]|0;
   $39 = ((($38)) + 20|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = HEAP32[$23>>2]|0;
   $42 = (($41) + ($34<<2)|0);
   $43 = HEAP32[$42>>2]|0;
   $44 = (FUNCTION_TABLE_iii[$40 & 8191]($vb,$43)|0);
   $45 = (($19) + ($i$027<<2)|0);
   HEAP32[$45>>2] = $44;
   $46 = (($18) + ($i$027<<2)|0);
   $not$3 = ($44|0)!=(0|0);
   $$ = $not$3&1;
   HEAP32[$46>>2] = $$;
   $47 = HEAP32[$vb>>2]|0;
   $48 = (($47) + ($i$027<<2)|0);
   $49 = HEAP32[$48>>2]|0;
   _memset(($49|0),0,($25|0))|0;
   $50 = (($i$027) + 1)|0;
   $51 = HEAP32[$13>>2]|0;
   $52 = ($50|0)<($51|0);
   if ($52) {
    $i$027 = $50;
   } else {
    $166 = $51;
    break;
   }
  }
 } else {
  $166 = $14;
 }
 $26 = ((($l)) + 1156|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ($27|0)>(0);
 if ($28) {
  $29 = ((($l)) + 1160|0);
  $30 = ((($l)) + 2184|0);
  $i$124 = 0;
  while(1) {
   $58 = (($29) + ($i$124<<2)|0);
   $59 = HEAP32[$58>>2]|0;
   $60 = (($18) + ($59<<2)|0);
   $61 = HEAP32[$60>>2]|0;
   $62 = ($61|0)==(0);
   $63 = (($30) + ($i$124<<2)|0);
   $64 = HEAP32[$63>>2]|0;
   if ($62) {
    $65 = (($18) + ($64<<2)|0);
    $66 = HEAP32[$65>>2]|0;
    $67 = ($66|0)==(0);
    if (!($67)) {
     label = 10;
    }
   } else {
    label = 10;
   }
   if ((label|0) == 10) {
    label = 0;
    HEAP32[$60>>2] = 1;
    $68 = (($18) + ($64<<2)|0);
    HEAP32[$68>>2] = 1;
   }
   $69 = (($i$124) + 1)|0;
   $70 = ($69|0)<($27|0);
   if ($70) {
    $i$124 = $69;
   } else {
    break;
   }
  }
 }
 $53 = HEAP32[$l>>2]|0;
 $54 = ($53|0)>(0);
 if ($54) {
  $55 = ((($l)) + 1092|0);
  $56 = ((($7)) + 52|0);
  $57 = ((($l)) + 4|0);
  $72 = $166;$i$222 = 0;
  while(1) {
   $71 = ($72|0)>(0);
   if ($71) {
    $167 = $72;$ch_in_bundle$018 = 0;$j$019 = 0;
    while(1) {
     $73 = (($57) + ($j$019<<2)|0);
     $74 = HEAP32[$73>>2]|0;
     $75 = ($74|0)==($i$222|0);
     if ($75) {
      $76 = (($18) + ($j$019<<2)|0);
      $77 = HEAP32[$76>>2]|0;
      $78 = (($17) + ($ch_in_bundle$018<<2)|0);
      $not$ = ($77|0)!=(0);
      $$2 = $not$&1;
      HEAP32[$78>>2] = $$2;
      $79 = HEAP32[$vb>>2]|0;
      $80 = (($79) + ($j$019<<2)|0);
      $81 = HEAP32[$80>>2]|0;
      $82 = (($ch_in_bundle$018) + 1)|0;
      $83 = (($16) + ($ch_in_bundle$018<<2)|0);
      HEAP32[$83>>2] = $81;
      $$pre33 = HEAP32[$13>>2]|0;
      $86 = $$pre33;$ch_in_bundle$1 = $82;
     } else {
      $86 = $167;$ch_in_bundle$1 = $ch_in_bundle$018;
     }
     $84 = (($j$019) + 1)|0;
     $85 = ($84|0)<($86|0);
     if ($85) {
      $167 = $86;$ch_in_bundle$018 = $ch_in_bundle$1;$j$019 = $84;
     } else {
      $ch_in_bundle$0$lcssa = $ch_in_bundle$1;
      break;
     }
    }
   } else {
    $ch_in_bundle$0$lcssa = 0;
   }
   $87 = (($55) + ($i$222<<2)|0);
   $88 = HEAP32[$87>>2]|0;
   $89 = (((($5)) + 1312|0) + ($88<<2)|0);
   $90 = HEAP32[$89>>2]|0;
   $91 = (101264 + ($90<<2)|0);
   $92 = HEAP32[$91>>2]|0;
   $93 = ((($92)) + 28|0);
   $94 = HEAP32[$93>>2]|0;
   $95 = HEAP32[$56>>2]|0;
   $96 = (($95) + ($88<<2)|0);
   $97 = HEAP32[$96>>2]|0;
   (FUNCTION_TABLE_iiiiii[$94 & 8191]($vb,$97,$16,$17,$ch_in_bundle$0$lcssa)|0);
   $98 = (($i$222) + 1)|0;
   $99 = HEAP32[$l>>2]|0;
   $100 = ($98|0)<($99|0);
   if (!($100)) {
    break;
   }
   $$pre32 = HEAP32[$13>>2]|0;
   $72 = $$pre32;$i$222 = $98;
  }
  $$pre34 = HEAP32[$26>>2]|0;
  $102 = $$pre34;
 } else {
  $102 = $27;
 }
 $101 = ($102|0)>(0);
 if ($101) {
  $103 = ((($l)) + 1160|0);
  $104 = HEAP32[$vb>>2]|0;
  $105 = ((($l)) + 2184|0);
  $106 = (($11|0) / 2)&-1;
  $107 = ($11|0)>(1);
  if ($107) {
   $i$315$us$in = $102;
   while(1) {
    $i$315$us = (($i$315$us$in) + -1)|0;
    $108 = (($103) + ($i$315$us<<2)|0);
    $109 = HEAP32[$108>>2]|0;
    $110 = (($104) + ($109<<2)|0);
    $111 = HEAP32[$110>>2]|0;
    $112 = (($105) + ($i$315$us<<2)|0);
    $113 = HEAP32[$112>>2]|0;
    $114 = (($104) + ($113<<2)|0);
    $115 = HEAP32[$114>>2]|0;
    $j$111$us = 0;
    while(1) {
     $116 = (($111) + ($j$111$us<<2)|0);
     $117 = +HEAPF32[$116>>2];
     $118 = (($115) + ($j$111$us<<2)|0);
     $119 = +HEAPF32[$118>>2];
     $120 = $117 > 0.0;
     $121 = $119 > 0.0;
     do {
      if ($120) {
       if ($121) {
        $125 = $117 - $119;
        HEAPF32[$118>>2] = $125;
        break;
       } else {
        HEAPF32[$118>>2] = $117;
        $124 = $117 + $119;
        HEAPF32[$116>>2] = $124;
        break;
       }
      } else {
       if ($121) {
        $123 = $117 + $119;
        HEAPF32[$118>>2] = $123;
        break;
       } else {
        HEAPF32[$118>>2] = $117;
        $122 = $117 - $119;
        HEAPF32[$116>>2] = $122;
        break;
       }
      }
     } while(0);
     $126 = (($j$111$us) + 1)|0;
     $127 = ($126|0)<($106|0);
     if ($127) {
      $j$111$us = $126;
     } else {
      break;
     }
    }
    $128 = ($i$315$us$in|0)>(1);
    if ($128) {
     $i$315$us$in = $i$315$us;
    } else {
     break;
    }
   }
  }
 }
 $129 = HEAP32[$13>>2]|0;
 $130 = ($129|0)>(0);
 if (!($130)) {
  STACKTOP = sp;return 0;
 }
 $131 = ((($l)) + 4|0);
 $132 = ((($l)) + 1028|0);
 $133 = ((($7)) + 48|0);
 $i$49 = 0;
 while(1) {
  $135 = HEAP32[$vb>>2]|0;
  $136 = (($135) + ($i$49<<2)|0);
  $137 = HEAP32[$136>>2]|0;
  $138 = (($131) + ($i$49<<2)|0);
  $139 = HEAP32[$138>>2]|0;
  $140 = (($132) + ($139<<2)|0);
  $141 = HEAP32[$140>>2]|0;
  $142 = (((($5)) + 800|0) + ($141<<2)|0);
  $143 = HEAP32[$142>>2]|0;
  $144 = (101372 + ($143<<2)|0);
  $145 = HEAP32[$144>>2]|0;
  $146 = ((($145)) + 24|0);
  $147 = HEAP32[$146>>2]|0;
  $148 = HEAP32[$133>>2]|0;
  $149 = (($148) + ($141<<2)|0);
  $150 = HEAP32[$149>>2]|0;
  $151 = (($19) + ($i$49<<2)|0);
  $152 = HEAP32[$151>>2]|0;
  (FUNCTION_TABLE_iiiii[$147 & 8191]($vb,$150,$152,$137)|0);
  $153 = (($i$49) + 1)|0;
  $154 = HEAP32[$13>>2]|0;
  $155 = ($153|0)<($154|0);
  if ($155) {
   $i$49 = $153;
  } else {
   $$lcssa = $154;
   break;
  }
 }
 $134 = ($$lcssa|0)>(0);
 if (!($134)) {
  STACKTOP = sp;return 0;
 }
 $i$58 = 0;
 while(1) {
  $156 = HEAP32[$vb>>2]|0;
  $157 = (($156) + ($i$58<<2)|0);
  $158 = HEAP32[$157>>2]|0;
  $159 = HEAP32[$8>>2]|0;
  $160 = (((($7)) + 12|0) + ($159<<2)|0);
  $161 = HEAP32[$160>>2]|0;
  $162 = HEAP32[$161>>2]|0;
  _mdct_backward($162,$158,$158);
  $163 = (($i$58) + 1)|0;
  $164 = HEAP32[$13>>2]|0;
  $165 = ($163|0)<($164|0);
  if ($165) {
   $i$58 = $163;
  } else {
   break;
  }
 }
 STACKTOP = sp;return 0;
}
function _mdct_backward($init,$in,$out) {
 $init = $init|0;
 $in = $in|0;
 $out = $out|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0.0, $105 = 0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0, $114 = 0.0, $115 = 0.0;
 var $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0.0, $120 = 0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0, $128 = 0.0, $129 = 0, $13 = 0, $130 = 0.0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0.0, $14 = 0.0, $140 = 0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0, $147 = 0.0, $148 = 0.0, $149 = 0, $15 = 0.0, $150 = 0.0, $151 = 0.0;
 var $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0.0, $160 = 0.0, $161 = 0, $162 = 0.0, $163 = 0.0, $164 = 0, $165 = 0.0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0.0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0.0, $174 = 0, $175 = 0.0, $176 = 0.0, $177 = 0, $178 = 0.0, $179 = 0.0, $18 = 0, $180 = 0.0, $181 = 0.0, $182 = 0, $183 = 0.0, $184 = 0.0, $185 = 0.0, $186 = 0.0, $187 = 0.0, $188 = 0.0;
 var $189 = 0.0, $19 = 0.0, $190 = 0.0, $191 = 0, $192 = 0.0, $193 = 0, $194 = 0.0, $195 = 0.0, $196 = 0, $197 = 0.0, $198 = 0, $199 = 0.0, $2 = 0, $20 = 0.0, $200 = 0.0, $201 = 0.0, $202 = 0, $203 = 0.0, $204 = 0.0, $205 = 0.0;
 var $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0.0, $21 = 0.0, $210 = 0.0, $211 = 0, $212 = 0, $213 = 0.0, $214 = 0, $215 = 0.0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0, $22 = 0.0, $220 = 0.0, $221 = 0.0, $222 = 0.0, $223 = 0;
 var $224 = 0.0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0.0, $229 = 0.0, $23 = 0.0, $230 = 0.0, $231 = 0.0, $232 = 0, $233 = 0, $234 = 0.0, $235 = 0, $236 = 0.0, $237 = 0.0, $238 = 0, $239 = 0.0, $24 = 0.0, $240 = 0, $241 = 0.0;
 var $242 = 0.0, $243 = 0.0, $244 = 0.0, $245 = 0.0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0.0, $250 = 0.0, $251 = 0.0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0.0;
 var $260 = 0, $261 = 0.0, $262 = 0, $263 = 0.0, $264 = 0, $265 = 0.0, $266 = 0, $267 = 0.0, $268 = 0, $269 = 0, $27 = 0.0, $270 = 0.0, $271 = 0, $272 = 0.0, $273 = 0, $274 = 0.0, $275 = 0.0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0.0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0, $33 = 0.0;
 var $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0;
 var $70 = 0.0, $71 = 0.0, $72 = 0.0, $73 = 0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0;
 var $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $T$0 = 0, $T$0$i = 0, $T$1 = 0, $T$2 = 0, $bit$0$i = 0, $iX$0 = 0, $iX$1 = 0, $iX1$0 = 0;
 var $iX1$1 = 0, $iX1$2 = 0, $init$idx = 0, $init$idx$val = 0, $init$idx2$val = 0, $init$idx3$val = 0, $init$idx4 = 0, $init$idx4$val = 0, $oX$0 = 0, $oX$1 = 0, $oX1$0 = 0, $oX1$1 = 0, $oX1$2 = 0, $oX2$0 = 0, $oX2$1 = 0, $w0$0$i = 0, $w1$0$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$init>>2]|0;
 $1 = $0 >> 1;
 $2 = $0 >> 2;
 $3 = (($in) + ($1<<2)|0);
 $4 = ((($3)) + -28|0);
 $5 = (($out) + ($1<<2)|0);
 $6 = (($5) + ($2<<2)|0);
 $7 = ((($init)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (($8) + ($2<<2)|0);
 $T$0 = $9;$iX$0 = $4;$oX$0 = $6;
 while(1) {
  $10 = ((($oX$0)) + -16|0);
  $11 = ((($iX$0)) + 8|0);
  $12 = +HEAPF32[$11>>2];
  $13 = ((($T$0)) + 12|0);
  $14 = +HEAPF32[$13>>2];
  $15 = $12 * $14;
  $16 = -$15;
  $17 = +HEAPF32[$iX$0>>2];
  $18 = ((($T$0)) + 8|0);
  $19 = +HEAPF32[$18>>2];
  $20 = $17 * $19;
  $21 = $16 - $20;
  HEAPF32[$10>>2] = $21;
  $22 = +HEAPF32[$iX$0>>2];
  $23 = +HEAPF32[$13>>2];
  $24 = $22 * $23;
  $25 = +HEAPF32[$11>>2];
  $26 = +HEAPF32[$18>>2];
  $27 = $25 * $26;
  $28 = $24 - $27;
  $29 = ((($oX$0)) + -12|0);
  HEAPF32[$29>>2] = $28;
  $30 = ((($iX$0)) + 24|0);
  $31 = +HEAPF32[$30>>2];
  $32 = ((($T$0)) + 4|0);
  $33 = +HEAPF32[$32>>2];
  $34 = $31 * $33;
  $35 = -$34;
  $36 = ((($iX$0)) + 16|0);
  $37 = +HEAPF32[$36>>2];
  $38 = +HEAPF32[$T$0>>2];
  $39 = $37 * $38;
  $40 = $35 - $39;
  $41 = ((($oX$0)) + -8|0);
  HEAPF32[$41>>2] = $40;
  $42 = +HEAPF32[$36>>2];
  $43 = +HEAPF32[$32>>2];
  $44 = $42 * $43;
  $45 = +HEAPF32[$30>>2];
  $46 = +HEAPF32[$T$0>>2];
  $47 = $45 * $46;
  $48 = $44 - $47;
  $49 = ((($oX$0)) + -4|0);
  HEAPF32[$49>>2] = $48;
  $50 = ((($iX$0)) + -32|0);
  $51 = ((($T$0)) + 16|0);
  $52 = ($50>>>0)<($in>>>0);
  if ($52) {
   break;
  } else {
   $T$0 = $51;$iX$0 = $50;$oX$0 = $10;
  }
 }
 $53 = ((($3)) + -32|0);
 $T$1 = $9;$iX$1 = $53;$oX$1 = $6;
 while(1) {
  $54 = ((($T$1)) + -16|0);
  $55 = ((($iX$1)) + 16|0);
  $56 = +HEAPF32[$55>>2];
  $57 = ((($T$1)) + -4|0);
  $58 = +HEAPF32[$57>>2];
  $59 = $56 * $58;
  $60 = ((($iX$1)) + 24|0);
  $61 = +HEAPF32[$60>>2];
  $62 = ((($T$1)) + -8|0);
  $63 = +HEAPF32[$62>>2];
  $64 = $61 * $63;
  $65 = $59 + $64;
  HEAPF32[$oX$1>>2] = $65;
  $66 = +HEAPF32[$55>>2];
  $67 = +HEAPF32[$62>>2];
  $68 = $66 * $67;
  $69 = +HEAPF32[$60>>2];
  $70 = +HEAPF32[$57>>2];
  $71 = $69 * $70;
  $72 = $68 - $71;
  $73 = ((($oX$1)) + 4|0);
  HEAPF32[$73>>2] = $72;
  $74 = +HEAPF32[$iX$1>>2];
  $75 = ((($T$1)) + -12|0);
  $76 = +HEAPF32[$75>>2];
  $77 = $74 * $76;
  $78 = ((($iX$1)) + 8|0);
  $79 = +HEAPF32[$78>>2];
  $80 = +HEAPF32[$54>>2];
  $81 = $79 * $80;
  $82 = $77 + $81;
  $83 = ((($oX$1)) + 8|0);
  HEAPF32[$83>>2] = $82;
  $84 = +HEAPF32[$iX$1>>2];
  $85 = +HEAPF32[$54>>2];
  $86 = $84 * $85;
  $87 = +HEAPF32[$78>>2];
  $88 = +HEAPF32[$75>>2];
  $89 = $87 * $88;
  $90 = $86 - $89;
  $91 = ((($oX$1)) + 12|0);
  HEAPF32[$91>>2] = $90;
  $92 = ((($iX$1)) + -32|0);
  $93 = ((($oX$1)) + 16|0);
  $94 = ($92>>>0)<($in>>>0);
  if ($94) {
   break;
  } else {
   $T$1 = $54;$iX$1 = $92;$oX$1 = $93;
  }
 }
 $init$idx = ((($init)) + 4|0);
 $init$idx$val = HEAP32[$init$idx>>2]|0;
 _mdct_butterflies($init$idx$val,$8,$5,$1);
 $init$idx2$val = HEAP32[$init>>2]|0;
 $init$idx3$val = HEAP32[$7>>2]|0;
 $init$idx4 = ((($init)) + 12|0);
 $init$idx4$val = HEAP32[$init$idx4>>2]|0;
 $95 = $init$idx2$val >> 1;
 $96 = (($out) + ($95<<2)|0);
 $97 = (($init$idx3$val) + ($init$idx2$val<<2)|0);
 $T$0$i = $97;$bit$0$i = $init$idx4$val;$w0$0$i = $out;$w1$0$i = $96;
 while(1) {
  $98 = HEAP32[$bit$0$i>>2]|0;
  $99 = (($96) + ($98<<2)|0);
  $100 = ((($bit$0$i)) + 4|0);
  $101 = HEAP32[$100>>2]|0;
  $102 = (($96) + ($101<<2)|0);
  $103 = ((($99)) + 4|0);
  $104 = +HEAPF32[$103>>2];
  $105 = ((($102)) + 4|0);
  $106 = +HEAPF32[$105>>2];
  $107 = $104 - $106;
  $108 = +HEAPF32[$99>>2];
  $109 = +HEAPF32[$102>>2];
  $110 = $108 + $109;
  $111 = +HEAPF32[$T$0$i>>2];
  $112 = $110 * $111;
  $113 = ((($T$0$i)) + 4|0);
  $114 = +HEAPF32[$113>>2];
  $115 = $107 * $114;
  $116 = $112 + $115;
  $117 = $110 * $114;
  $118 = $107 * $111;
  $119 = $117 - $118;
  $120 = ((($w1$0$i)) + -16|0);
  $121 = $104 + $106;
  $122 = $121 * 0.5;
  $123 = $108 - $109;
  $124 = $123 * 0.5;
  $125 = $122 + $116;
  HEAPF32[$w0$0$i>>2] = $125;
  $126 = $122 - $116;
  $127 = ((($w1$0$i)) + -8|0);
  HEAPF32[$127>>2] = $126;
  $128 = $124 + $119;
  $129 = ((($w0$0$i)) + 4|0);
  HEAPF32[$129>>2] = $128;
  $130 = $119 - $124;
  $131 = ((($w1$0$i)) + -4|0);
  HEAPF32[$131>>2] = $130;
  $132 = ((($bit$0$i)) + 8|0);
  $133 = HEAP32[$132>>2]|0;
  $134 = (($96) + ($133<<2)|0);
  $135 = ((($bit$0$i)) + 12|0);
  $136 = HEAP32[$135>>2]|0;
  $137 = (($96) + ($136<<2)|0);
  $138 = ((($134)) + 4|0);
  $139 = +HEAPF32[$138>>2];
  $140 = ((($137)) + 4|0);
  $141 = +HEAPF32[$140>>2];
  $142 = $139 - $141;
  $143 = +HEAPF32[$134>>2];
  $144 = +HEAPF32[$137>>2];
  $145 = $143 + $144;
  $146 = ((($T$0$i)) + 8|0);
  $147 = +HEAPF32[$146>>2];
  $148 = $145 * $147;
  $149 = ((($T$0$i)) + 12|0);
  $150 = +HEAPF32[$149>>2];
  $151 = $142 * $150;
  $152 = $148 + $151;
  $153 = $145 * $150;
  $154 = $142 * $147;
  $155 = $153 - $154;
  $156 = $139 + $141;
  $157 = $156 * 0.5;
  $158 = $143 - $144;
  $159 = $158 * 0.5;
  $160 = $157 + $152;
  $161 = ((($w0$0$i)) + 8|0);
  HEAPF32[$161>>2] = $160;
  $162 = $157 - $152;
  HEAPF32[$120>>2] = $162;
  $163 = $159 + $155;
  $164 = ((($w0$0$i)) + 12|0);
  HEAPF32[$164>>2] = $163;
  $165 = $155 - $159;
  $166 = ((($w1$0$i)) + -12|0);
  HEAPF32[$166>>2] = $165;
  $167 = ((($T$0$i)) + 16|0);
  $168 = ((($bit$0$i)) + 16|0);
  $169 = ((($w0$0$i)) + 16|0);
  $170 = ($169>>>0)<($120>>>0);
  if ($170) {
   $T$0$i = $167;$bit$0$i = $168;$w0$0$i = $169;$w1$0$i = $120;
  } else {
   break;
  }
 }
 $171 = (($init$idx3$val) + ($1<<2)|0);
 $T$2 = $171;$iX1$0 = $out;$oX1$0 = $6;$oX2$0 = $6;
 while(1) {
  $172 = ((($oX1$0)) + -16|0);
  $173 = +HEAPF32[$iX1$0>>2];
  $174 = ((($T$2)) + 4|0);
  $175 = +HEAPF32[$174>>2];
  $176 = $173 * $175;
  $177 = ((($iX1$0)) + 4|0);
  $178 = +HEAPF32[$177>>2];
  $179 = +HEAPF32[$T$2>>2];
  $180 = $178 * $179;
  $181 = $176 - $180;
  $182 = ((($oX1$0)) + -4|0);
  HEAPF32[$182>>2] = $181;
  $183 = +HEAPF32[$iX1$0>>2];
  $184 = +HEAPF32[$T$2>>2];
  $185 = $183 * $184;
  $186 = +HEAPF32[$177>>2];
  $187 = +HEAPF32[$174>>2];
  $188 = $186 * $187;
  $189 = $185 + $188;
  $190 = -$189;
  HEAPF32[$oX2$0>>2] = $190;
  $191 = ((($iX1$0)) + 8|0);
  $192 = +HEAPF32[$191>>2];
  $193 = ((($T$2)) + 12|0);
  $194 = +HEAPF32[$193>>2];
  $195 = $192 * $194;
  $196 = ((($iX1$0)) + 12|0);
  $197 = +HEAPF32[$196>>2];
  $198 = ((($T$2)) + 8|0);
  $199 = +HEAPF32[$198>>2];
  $200 = $197 * $199;
  $201 = $195 - $200;
  $202 = ((($oX1$0)) + -8|0);
  HEAPF32[$202>>2] = $201;
  $203 = +HEAPF32[$191>>2];
  $204 = +HEAPF32[$198>>2];
  $205 = $203 * $204;
  $206 = +HEAPF32[$196>>2];
  $207 = +HEAPF32[$193>>2];
  $208 = $206 * $207;
  $209 = $205 + $208;
  $210 = -$209;
  $211 = ((($oX2$0)) + 4|0);
  HEAPF32[$211>>2] = $210;
  $212 = ((($iX1$0)) + 16|0);
  $213 = +HEAPF32[$212>>2];
  $214 = ((($T$2)) + 20|0);
  $215 = +HEAPF32[$214>>2];
  $216 = $213 * $215;
  $217 = ((($iX1$0)) + 20|0);
  $218 = +HEAPF32[$217>>2];
  $219 = ((($T$2)) + 16|0);
  $220 = +HEAPF32[$219>>2];
  $221 = $218 * $220;
  $222 = $216 - $221;
  $223 = ((($oX1$0)) + -12|0);
  HEAPF32[$223>>2] = $222;
  $224 = +HEAPF32[$212>>2];
  $225 = +HEAPF32[$219>>2];
  $226 = $224 * $225;
  $227 = +HEAPF32[$217>>2];
  $228 = +HEAPF32[$214>>2];
  $229 = $227 * $228;
  $230 = $226 + $229;
  $231 = -$230;
  $232 = ((($oX2$0)) + 8|0);
  HEAPF32[$232>>2] = $231;
  $233 = ((($iX1$0)) + 24|0);
  $234 = +HEAPF32[$233>>2];
  $235 = ((($T$2)) + 28|0);
  $236 = +HEAPF32[$235>>2];
  $237 = $234 * $236;
  $238 = ((($iX1$0)) + 28|0);
  $239 = +HEAPF32[$238>>2];
  $240 = ((($T$2)) + 24|0);
  $241 = +HEAPF32[$240>>2];
  $242 = $239 * $241;
  $243 = $237 - $242;
  HEAPF32[$172>>2] = $243;
  $244 = +HEAPF32[$233>>2];
  $245 = +HEAPF32[$240>>2];
  $246 = $244 * $245;
  $247 = +HEAPF32[$238>>2];
  $248 = +HEAPF32[$235>>2];
  $249 = $247 * $248;
  $250 = $246 + $249;
  $251 = -$250;
  $252 = ((($oX2$0)) + 12|0);
  HEAPF32[$252>>2] = $251;
  $253 = ((($oX2$0)) + 16|0);
  $254 = ((($iX1$0)) + 32|0);
  $255 = ((($T$2)) + 32|0);
  $256 = ($254>>>0)<($172>>>0);
  if ($256) {
   $T$2 = $255;$iX1$0 = $254;$oX1$0 = $172;$oX2$0 = $253;
  } else {
   break;
  }
 }
 $257 = (($out) + ($2<<2)|0);
 $iX1$1 = $6;$oX1$1 = $257;$oX2$1 = $257;
 while(1) {
  $258 = ((($oX1$1)) + -16|0);
  $259 = ((($iX1$1)) + -16|0);
  $260 = ((($iX1$1)) + -4|0);
  $261 = +HEAPF32[$260>>2];
  $262 = ((($oX1$1)) + -4|0);
  HEAPF32[$262>>2] = $261;
  $263 = -$261;
  HEAPF32[$oX2$1>>2] = $263;
  $264 = ((($iX1$1)) + -8|0);
  $265 = +HEAPF32[$264>>2];
  $266 = ((($oX1$1)) + -8|0);
  HEAPF32[$266>>2] = $265;
  $267 = -$265;
  $268 = ((($oX2$1)) + 4|0);
  HEAPF32[$268>>2] = $267;
  $269 = ((($iX1$1)) + -12|0);
  $270 = +HEAPF32[$269>>2];
  $271 = ((($oX1$1)) + -12|0);
  HEAPF32[$271>>2] = $270;
  $272 = -$270;
  $273 = ((($oX2$1)) + 8|0);
  HEAPF32[$273>>2] = $272;
  $274 = +HEAPF32[$259>>2];
  HEAPF32[$258>>2] = $274;
  $275 = -$274;
  $276 = ((($oX2$1)) + 12|0);
  HEAPF32[$276>>2] = $275;
  $277 = ((($oX2$1)) + 16|0);
  $278 = ($277>>>0)<($259>>>0);
  if ($278) {
   $iX1$1 = $259;$oX1$1 = $258;$oX2$1 = $277;
  } else {
   $iX1$2 = $6;$oX1$2 = $6;
   break;
  }
 }
 while(1) {
  $279 = ((($oX1$2)) + -16|0);
  $280 = ((($iX1$2)) + 12|0);
  $281 = HEAP32[$280>>2]|0;
  HEAP32[$279>>2] = $281;
  $282 = ((($iX1$2)) + 8|0);
  $283 = HEAP32[$282>>2]|0;
  $284 = ((($oX1$2)) + -12|0);
  HEAP32[$284>>2] = $283;
  $285 = ((($iX1$2)) + 4|0);
  $286 = HEAP32[$285>>2]|0;
  $287 = ((($oX1$2)) + -8|0);
  HEAP32[$287>>2] = $286;
  $288 = HEAP32[$iX1$2>>2]|0;
  $289 = ((($oX1$2)) + -4|0);
  HEAP32[$289>>2] = $288;
  $290 = ((($iX1$2)) + 16|0);
  $291 = ($279>>>0)>($5>>>0);
  if ($291) {
   $iX1$2 = $290;$oX1$2 = $279;
  } else {
   break;
  }
 }
 return;
}
function _vorbis_bitrate_clear($bm) {
 $bm = $bm|0;
 var dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 dest=$bm; stop=dest+48|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 return;
}
function _vorbis_block_init($v,$vb) {
 $v = $v|0;
 $vb = $vb|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, dest = 0, label = 0, sp = 0;
 var stop = 0;
 sp = STACKTOP;
 dest=$vb; stop=dest+112|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $0 = ((($vb)) + 64|0);
 HEAP32[$0>>2] = $v;
 $1 = ((($vb)) + 76|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($vb)) + 68|0);
 HEAP32[$2>>2] = 0;
 $3 = HEAP32[$v>>2]|0;
 $4 = ($3|0)==(0);
 if ($4) {
  return 0;
 }
 $5 = (_calloc(1,72)|0);
 $6 = ((($vb)) + 104|0);
 HEAP32[$6>>2] = $5;
 $7 = ((($5)) + 4|0);
 HEAPF32[$7>>2] = -9999.0;
 $8 = ((($vb)) + 4|0);
 $9 = ((($5)) + 12|0);
 $10 = ((($5)) + 40|0);
 $i$01 = 0;
 while(1) {
  $11 = ($i$01|0)==(7);
  if ($11) {
   HEAP32[$10>>2] = $8;
   _oggpack_writeinit($8);
   $i$01 = 8;
   continue;
  } else {
   $12 = (_calloc(1,20)|0);
   $13 = (($9) + ($i$01<<2)|0);
   HEAP32[$13>>2] = $12;
   _oggpack_writeinit($12);
   $14 = (($i$01) + 1)|0;
   $exitcond = ($14|0)==(15);
   if ($exitcond) {
    break;
   } else {
    $i$01 = $14;
    continue;
   }
  }
 }
 return 0;
}
function __vorbis_block_ripcord($vb) {
 $vb = $vb|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $reap$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($vb)) + 84|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $reap$01 = $1;
  while(1) {
   $3 = ((($reap$01)) + 4|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = HEAP32[$reap$01>>2]|0;
   _free($5);
   _free($reap$01);
   $6 = ($4|0)==(0|0);
   if ($6) {
    break;
   } else {
    $reap$01 = $4;
   }
  }
 }
 $7 = ((($vb)) + 80|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0);
 if ($9) {
  $19 = ((($vb)) + 72|0);
  HEAP32[$19>>2] = 0;
  HEAP32[$0>>2] = 0;
  return;
 }
 $10 = ((($vb)) + 68|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($vb)) + 76|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($13) + ($8))|0;
 $15 = (_realloc($11,$14)|0);
 HEAP32[$10>>2] = $15;
 $16 = HEAP32[$7>>2]|0;
 $17 = HEAP32[$12>>2]|0;
 $18 = (($17) + ($16))|0;
 HEAP32[$12>>2] = $18;
 HEAP32[$7>>2] = 0;
 $19 = ((($vb)) + 72|0);
 HEAP32[$19>>2] = 0;
 HEAP32[$0>>2] = 0;
 return;
}
function _vorbis_block_clear($vb) {
 $vb = $vb|0;
 var $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0, $reap$01$i = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($vb)) + 104|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($vb)) + 84|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  $reap$01$i = $3;
  while(1) {
   $5 = ((($reap$01$i)) + 4|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = HEAP32[$reap$01$i>>2]|0;
   _free($7);
   _free($reap$01$i);
   $8 = ($6|0)==(0|0);
   if ($8) {
    break;
   } else {
    $reap$01$i = $6;
   }
  }
 }
 $9 = ((($vb)) + 80|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10|0)==(0);
 $$phi$trans$insert = ((($vb)) + 68|0);
 $$pre = HEAP32[$$phi$trans$insert>>2]|0;
 if ($11) {
  $21 = $$pre;
 } else {
  $12 = ((($vb)) + 76|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = (($13) + ($10))|0;
  $15 = (_realloc($$pre,$14)|0);
  HEAP32[$$phi$trans$insert>>2] = $15;
  $16 = HEAP32[$9>>2]|0;
  $17 = HEAP32[$12>>2]|0;
  $18 = (($17) + ($16))|0;
  HEAP32[$12>>2] = $18;
  HEAP32[$9>>2] = 0;
  $21 = $15;
 }
 $19 = ((($vb)) + 72|0);
 HEAP32[$19>>2] = 0;
 HEAP32[$2>>2] = 0;
 $20 = ($21|0)==(0|0);
 if (!($20)) {
  _free($21);
 }
 $22 = ($1|0)==(0|0);
 if ($22) {
  dest=$vb; stop=dest+112|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
  return 0;
 } else {
  $i$01 = 0;
 }
 while(1) {
  $23 = (((($1)) + 12|0) + ($i$01<<2)|0);
  $24 = HEAP32[$23>>2]|0;
  _oggpack_writeclear($24);
  $25 = ($i$01|0)==(7);
  if ($25) {
   $i$01 = 8;
   continue;
  }
  $26 = HEAP32[$23>>2]|0;
  _free($26);
  $27 = (($i$01) + 1)|0;
  $exitcond = ($27|0)==(15);
  if ($exitcond) {
   break;
  } else {
   $i$01 = $27;
  }
 }
 _free($1);
 dest=$vb; stop=dest+112|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 return 0;
}
function __vds_shared_init($v,$vi,$encp) {
 $v = $v|0;
 $vi = $vi|0;
 $encp = $encp|0;
 var $$0 = 0, $$pre = 0, $$pre25 = 0, $$pre26 = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $i$117 = 0, $i$214 = 0, $i$37 = 0, $i$45 = 0, $i$54 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($vi)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = 1;
  return ($$0|0);
 }
 $3 = ((($1)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)<(1);
 if ($5) {
  $$0 = 1;
  return ($$0|0);
 }
 $6 = HEAP32[$1>>2]|0;
 $7 = ($6|0)<(64);
 if ($7) {
  $$0 = 1;
  return ($$0|0);
 }
 $8 = ((($1)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)<($6|0);
 if ($10) {
  $$0 = 1;
  return ($$0|0);
 }
 $11 = ((($1)) + 3656|0);
 $12 = HEAP32[$11>>2]|0;
 dest=$v; stop=dest+112|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $13 = (_calloc(1,136)|0);
 $14 = ((($v)) + 104|0);
 HEAP32[$14>>2] = $13;
 $15 = ((($v)) + 4|0);
 HEAP32[$15>>2] = $vi;
 $16 = HEAP32[$3>>2]|0;
 $17 = (($16) + -1)|0;
 $18 = (_ov_ilog($17)|0);
 $19 = ((($13)) + 44|0);
 HEAP32[$19>>2] = $18;
 $20 = (_calloc(1,4)|0);
 $21 = ((($13)) + 12|0);
 HEAP32[$21>>2] = $20;
 $22 = (_calloc(1,4)|0);
 $23 = ((($13)) + 16|0);
 HEAP32[$23>>2] = $22;
 $24 = (_calloc(1,20)|0);
 HEAP32[$20>>2] = $24;
 $25 = (_calloc(1,20)|0);
 HEAP32[$22>>2] = $25;
 $26 = HEAP32[$1>>2]|0;
 $27 = $26 >> $12;
 _mdct_init($24,$27);
 $28 = HEAP32[$23>>2]|0;
 $29 = HEAP32[$28>>2]|0;
 $30 = HEAP32[$8>>2]|0;
 $31 = $30 >> $12;
 _mdct_init($29,$31);
 $32 = HEAP32[$1>>2]|0;
 $33 = (_ov_ilog($32)|0);
 $34 = (($33) + -7)|0;
 $35 = ((($13)) + 4|0);
 HEAP32[$35>>2] = $34;
 $36 = HEAP32[$8>>2]|0;
 $37 = (_ov_ilog($36)|0);
 $38 = (($37) + -7)|0;
 $39 = ((($13)) + 8|0);
 HEAP32[$39>>2] = $38;
 $40 = ($encp|0)==(0);
 L13: do {
  if ($40) {
   $84 = ((($1)) + 2848|0);
   $85 = HEAP32[$84>>2]|0;
   $86 = ($85|0)==(0|0);
   if ($86) {
    $87 = ((($1)) + 24|0);
    $88 = HEAP32[$87>>2]|0;
    $89 = (_calloc($88,56)|0);
    HEAP32[$84>>2] = $89;
    $90 = HEAP32[$87>>2]|0;
    $91 = ($90|0)>(0);
    if ($91) {
     $174 = $90;$i$214 = 0;
     while(1) {
      $92 = (((($1)) + 1824|0) + ($i$214<<2)|0);
      $93 = HEAP32[$92>>2]|0;
      $94 = ($93|0)==(0|0);
      if ($94) {
       $96 = $174;
       break;
      }
      $97 = HEAP32[$84>>2]|0;
      $98 = (($97) + (($i$214*56)|0)|0);
      $99 = (_vorbis_book_init_decode($98,$93)|0);
      $100 = ($99|0)==(0);
      if (!($100)) {
       label = 20;
       break;
      }
      $101 = HEAP32[$92>>2]|0;
      _vorbis_staticbook_destroy($101);
      HEAP32[$92>>2] = 0;
      $102 = (($i$214) + 1)|0;
      $103 = HEAP32[$87>>2]|0;
      $104 = ($102|0)<($103|0);
      if ($104) {
       $174 = $103;$i$214 = $102;
      } else {
       break L13;
      }
     }
     if ((label|0) == 20) {
      $$pre27 = HEAP32[$87>>2]|0;
      $96 = $$pre27;
     }
     $95 = ($96|0)>(0);
     if ($95) {
      $175 = $96;$i$54 = 0;
      while(1) {
       $168 = (((($1)) + 1824|0) + ($i$54<<2)|0);
       $169 = HEAP32[$168>>2]|0;
       $170 = ($169|0)==(0|0);
       if ($170) {
        $173 = $175;
       } else {
        _vorbis_staticbook_destroy($169);
        HEAP32[$168>>2] = 0;
        $$pre28 = HEAP32[$87>>2]|0;
        $173 = $$pre28;
       }
       $171 = (($i$54) + 1)|0;
       $172 = ($171|0)<($173|0);
       if ($172) {
        $175 = $173;$i$54 = $171;
       } else {
        break;
       }
      }
     }
     _vorbis_dsp_clear($v);
     $$0 = -1;
     return ($$0|0);
    }
   }
  } else {
   $41 = ((($13)) + 20|0);
   $42 = HEAP32[$1>>2]|0;
   _drft_init($41,$42);
   $43 = ((($13)) + 32|0);
   $44 = HEAP32[$8>>2]|0;
   _drft_init($43,$44);
   $45 = ((($1)) + 2848|0);
   $46 = HEAP32[$45>>2]|0;
   $47 = ($46|0)==(0|0);
   if ($47) {
    $48 = ((($1)) + 24|0);
    $49 = HEAP32[$48>>2]|0;
    $50 = (_calloc($49,56)|0);
    HEAP32[$45>>2] = $50;
    $51 = HEAP32[$48>>2]|0;
    $52 = ($51|0)>(0);
    if ($52) {
     $53 = ((($1)) + 1824|0);
     $54 = HEAP32[$53>>2]|0;
     (_vorbis_book_init_encode($50,$54)|0);
     $55 = HEAP32[$48>>2]|0;
     $56 = ($55|0)>(1);
     if ($56) {
      $58 = 1;
      while(1) {
       $$pre = HEAP32[$45>>2]|0;
       $57 = (($$pre) + (($58*56)|0)|0);
       $59 = (((($1)) + 1824|0) + ($58<<2)|0);
       $60 = HEAP32[$59>>2]|0;
       (_vorbis_book_init_encode($57,$60)|0);
       $61 = (($58) + 1)|0;
       $62 = HEAP32[$48>>2]|0;
       $63 = ($61|0)<($62|0);
       if ($63) {
        $58 = $61;
       } else {
        break;
       }
      }
     }
    }
   }
   $64 = ((($1)) + 28|0);
   $65 = HEAP32[$64>>2]|0;
   $66 = (_calloc($65,52)|0);
   $67 = ((($13)) + 56|0);
   HEAP32[$67>>2] = $66;
   $68 = HEAP32[$64>>2]|0;
   $69 = ($68|0)>(0);
   L38: do {
    if ($69) {
     $70 = ((($1)) + 2868|0);
     $71 = ((($vi)) + 8|0);
     $73 = $66;$i$117 = 0;
     while(1) {
      $72 = (($73) + (($i$117*52)|0)|0);
      $74 = (((($1)) + 2852|0) + ($i$117<<2)|0);
      $75 = HEAP32[$74>>2]|0;
      $76 = HEAP32[$75>>2]|0;
      $77 = (($1) + ($76<<2)|0);
      $78 = HEAP32[$77>>2]|0;
      $79 = (($78|0) / 2)&-1;
      $80 = HEAP32[$71>>2]|0;
      __vp_psy_init($72,$75,$70,$79,$80);
      $81 = (($i$117) + 1)|0;
      $82 = HEAP32[$64>>2]|0;
      $83 = ($81|0)<($82|0);
      if (!($83)) {
       break L38;
      }
      $$pre25 = HEAP32[$67>>2]|0;
      $73 = $$pre25;$i$117 = $81;
     }
    }
   } while(0);
   HEAP32[$v>>2] = 1;
  }
 } while(0);
 $105 = HEAP32[$8>>2]|0;
 $106 = ((($v)) + 16|0);
 HEAP32[$106>>2] = $105;
 $107 = ((($vi)) + 4|0);
 $108 = HEAP32[$107>>2]|0;
 $109 = $108 << 2;
 $110 = (_malloc($109)|0);
 $111 = ((($v)) + 8|0);
 HEAP32[$111>>2] = $110;
 $112 = (_malloc($109)|0);
 $113 = ((($v)) + 12|0);
 HEAP32[$113>>2] = $112;
 $114 = ($108|0)>(0);
 if ($114) {
  $115 = (_calloc($105,4)|0);
  HEAP32[$110>>2] = $115;
  $116 = ($108|0)>(1);
  if ($116) {
   $119 = 1;
   while(1) {
    $$pre26 = HEAP32[$111>>2]|0;
    $117 = (_calloc($105,4)|0);
    $118 = (($$pre26) + ($119<<2)|0);
    HEAP32[$118>>2] = $117;
    $120 = (($119) + 1)|0;
    $121 = ($120|0)<($108|0);
    if ($121) {
     $119 = $120;
    } else {
     break;
    }
   }
  }
 }
 $122 = ((($v)) + 36|0);
 HEAP32[$122>>2] = 0;
 $123 = ((($v)) + 40|0);
 HEAP32[$123>>2] = 0;
 $124 = HEAP32[$8>>2]|0;
 $125 = (($124|0) / 2)&-1;
 $126 = ((($v)) + 48|0);
 HEAP32[$126>>2] = $125;
 $127 = ((($v)) + 20|0);
 HEAP32[$127>>2] = $125;
 $128 = ((($1)) + 16|0);
 $129 = HEAP32[$128>>2]|0;
 $130 = (_calloc($129,4)|0);
 $131 = ((($13)) + 48|0);
 HEAP32[$131>>2] = $130;
 $132 = ((($1)) + 20|0);
 $133 = HEAP32[$132>>2]|0;
 $134 = (_calloc($133,4)|0);
 $135 = ((($13)) + 52|0);
 HEAP32[$135>>2] = $134;
 $136 = HEAP32[$128>>2]|0;
 $137 = ($136|0)>(0);
 if ($137) {
  $i$37 = 0;
  while(1) {
   $140 = (((($1)) + 800|0) + ($i$37<<2)|0);
   $141 = HEAP32[$140>>2]|0;
   $142 = (101372 + ($141<<2)|0);
   $143 = HEAP32[$142>>2]|0;
   $144 = ((($143)) + 8|0);
   $145 = HEAP32[$144>>2]|0;
   $146 = (((($1)) + 1056|0) + ($i$37<<2)|0);
   $147 = HEAP32[$146>>2]|0;
   $148 = (FUNCTION_TABLE_iii[$145 & 8191]($v,$147)|0);
   $149 = HEAP32[$131>>2]|0;
   $150 = (($149) + ($i$37<<2)|0);
   HEAP32[$150>>2] = $148;
   $151 = (($i$37) + 1)|0;
   $152 = HEAP32[$128>>2]|0;
   $153 = ($151|0)<($152|0);
   if ($153) {
    $i$37 = $151;
   } else {
    break;
   }
  }
 }
 $138 = HEAP32[$132>>2]|0;
 $139 = ($138|0)>(0);
 if ($139) {
  $i$45 = 0;
 } else {
  $$0 = 0;
  return ($$0|0);
 }
 while(1) {
  $154 = (((($1)) + 1312|0) + ($i$45<<2)|0);
  $155 = HEAP32[$154>>2]|0;
  $156 = (101264 + ($155<<2)|0);
  $157 = HEAP32[$156>>2]|0;
  $158 = ((($157)) + 8|0);
  $159 = HEAP32[$158>>2]|0;
  $160 = (((($1)) + 1568|0) + ($i$45<<2)|0);
  $161 = HEAP32[$160>>2]|0;
  $162 = (FUNCTION_TABLE_iii[$159 & 8191]($v,$161)|0);
  $163 = HEAP32[$135>>2]|0;
  $164 = (($163) + ($i$45<<2)|0);
  HEAP32[$164>>2] = $162;
  $165 = (($i$45) + 1)|0;
  $166 = HEAP32[$132>>2]|0;
  $167 = ($165|0)<($166|0);
  if ($167) {
   $i$45 = $165;
  } else {
   $$0 = 0;
   break;
  }
 }
 return ($$0|0);
}
function _mdct_init($lookup,$n) {
 $lookup = $lookup|0;
 $n = $n|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0;
 var $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0, $44 = 0.0;
 var $45 = 0.0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0.0, $62 = 0.0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0.0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0.0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0.0, $9 = 0.0, $90 = 0, $acc$03 = 0, $acc$1 = 0, $acc$1$lcssa = 0, $i$07 = 0, $i$15 = 0, $i1$04 = 0, $j$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($n|0) / 4)&-1;
 $1 = $0 << 2;
 $2 = (_malloc($1)|0);
 $3 = (($0) + ($n))|0;
 $4 = $3 << 2;
 $5 = (_malloc($4)|0);
 $6 = $n >> 1;
 $7 = (+($n|0));
 $8 = $7;
 $9 = (+Math_log((+$8)));
 $10 = $9 / 0.69314718055994529;
 $11 = (+_rint($10));
 $12 = (~~(($11)));
 $13 = ((($lookup)) + 4|0);
 HEAP32[$13>>2] = $12;
 HEAP32[$lookup>>2] = $n;
 $14 = ((($lookup)) + 8|0);
 HEAP32[$14>>2] = $5;
 $15 = ((($lookup)) + 12|0);
 HEAP32[$15>>2] = $2;
 $16 = ($n|0)>(3);
 if (!($16)) {
  $89 = 4.0 / $7;
  $90 = ((($lookup)) + 16|0);
  HEAPF32[$90>>2] = $89;
  return;
 }
 $17 = (+($n|0));
 $18 = 3.1415926535897931 / $17;
 $19 = $n << 1;
 $20 = (+($19|0));
 $21 = 3.1415926535897931 / $20;
 $i$07 = 0;
 while(1) {
  $26 = $i$07 << 2;
  $27 = (+($26|0));
  $28 = $18 * $27;
  $29 = (+Math_cos((+$28)));
  $30 = $29;
  $31 = $i$07 << 1;
  $32 = (($5) + ($31<<2)|0);
  HEAPF32[$32>>2] = $30;
  $33 = (+Math_sin((+$28)));
  $34 = $33;
  $35 = -$34;
  $36 = $31 | 1;
  $37 = (($5) + ($36<<2)|0);
  HEAPF32[$37>>2] = $35;
  $38 = (+($36|0));
  $39 = $21 * $38;
  $40 = (+Math_cos((+$39)));
  $41 = $40;
  $42 = (($31) + ($6))|0;
  $43 = (($5) + ($42<<2)|0);
  HEAPF32[$43>>2] = $41;
  $44 = (+Math_sin((+$39)));
  $45 = $44;
  $46 = (($42) + 1)|0;
  $47 = (($5) + ($46<<2)|0);
  HEAPF32[$47>>2] = $45;
  $48 = (($i$07) + 1)|0;
  $49 = ($48|0)<($0|0);
  if ($49) {
   $i$07 = $48;
  } else {
   break;
  }
 }
 $22 = (($n|0) / 8)&-1;
 $23 = ($n|0)>(7);
 if (!($23)) {
  $89 = 4.0 / $7;
  $90 = ((($lookup)) + 16|0);
  HEAPF32[$90>>2] = $89;
  return;
 }
 $24 = (+($n|0));
 $25 = 3.1415926535897931 / $24;
 $i$15 = 0;
 while(1) {
  $50 = $i$15 << 2;
  $51 = $50 | 2;
  $52 = (+($51|0));
  $53 = $25 * $52;
  $54 = (+Math_cos((+$53)));
  $55 = $54 * 0.5;
  $56 = $55;
  $57 = $i$15 << 1;
  $58 = (($57) + ($n))|0;
  $59 = (($5) + ($58<<2)|0);
  HEAPF32[$59>>2] = $56;
  $60 = (+Math_sin((+$53)));
  $61 = $60 * -0.5;
  $62 = $61;
  $63 = (($58) + 1)|0;
  $64 = (($5) + ($63<<2)|0);
  HEAPF32[$64>>2] = $62;
  $65 = (($i$15) + 1)|0;
  $66 = ($65|0)<($22|0);
  if ($66) {
   $i$15 = $65;
  } else {
   break;
  }
 }
 $67 = (($12) + -1)|0;
 $68 = 1 << $67;
 $69 = (($68) + -1)|0;
 $70 = (($12) + -2)|0;
 $71 = 1 << $70;
 if ($23) {
  $i1$04 = 0;
 } else {
  $89 = 4.0 / $7;
  $90 = ((($lookup)) + 16|0);
  HEAPF32[$90>>2] = $89;
  return;
 }
 while(1) {
  $73 = $71;$acc$03 = 0;$j$02 = 0;
  while(1) {
   $72 = $73 & $i1$04;
   $74 = ($72|0)==(0);
   $75 = 1 << $j$02;
   $76 = $74 ? 0 : $75;
   $acc$1 = $76 | $acc$03;
   $77 = (($j$02) + 1)|0;
   $78 = $71 >> $77;
   $79 = ($78|0)==(0);
   if ($79) {
    $acc$1$lcssa = $acc$1;
    break;
   } else {
    $73 = $78;$acc$03 = $acc$1;$j$02 = $77;
   }
  }
  $80 = $acc$1$lcssa ^ -1;
  $81 = $69 & $80;
  $82 = (($81) + -1)|0;
  $83 = $i1$04 << 1;
  $84 = (($2) + ($83<<2)|0);
  HEAP32[$84>>2] = $82;
  $85 = $83 | 1;
  $86 = (($2) + ($85<<2)|0);
  HEAP32[$86>>2] = $acc$1$lcssa;
  $87 = (($i1$04) + 1)|0;
  $88 = ($87|0)<($22|0);
  if ($88) {
   $i1$04 = $87;
  } else {
   break;
  }
 }
 $89 = 4.0 / $7;
 $90 = ((($lookup)) + 16|0);
 HEAPF32[$90>>2] = $89;
 return;
}
function _drft_init($l,$n) {
 $l = $l|0;
 $n = $n|0;
 var $$lcssa = 0, $$lcssa$i$i = 0, $$lcssa7$i$i = 0, $$lcssa7$not$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0.0, $63 = 0.0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $exitcond37$i$i = 0, $exitcond38$i$i = 0, $exitcond39$i$i = 0, $fi$010$us$i$i = 0.0;
 var $i$021$i$i = 0, $i$19$us$i$i = 0, $ii$08$us$i$i = 0, $indvars$iv$i$i = 0, $indvars$iv$in$i$i = 0, $is$018$i$i = 0, $is$1$lcssa$i$i = 0, $is$112$us$i$i = 0, $j$0$i$i = 0, $j$113$us$i$i = 0, $k1$019$i$i = 0, $l1$017$i$i = 0, $ld$011$us$i$i = 0, $nf$0$i$i = 0, $nf$1$lcssa6$i$i = 0, $nf$1$us$i$i = 0, $nf$1$us$i$i$lcssa37 = 0, $nl$0$i$i = 0, $nl$1$i$i = 0, $nl$1$us$i$i = 0;
 var $ntry$0$i$i = 0, $ntry$1$ph$i$i = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$l>>2] = $n;
 $0 = ($n*3)|0;
 $1 = (_calloc($0,4)|0);
 $2 = ((($l)) + 4|0);
 HEAP32[$2>>2] = $1;
 $3 = (_calloc(32,4)|0);
 $4 = ((($l)) + 8|0);
 HEAP32[$4>>2] = $3;
 $5 = ($n|0)==(1);
 if ($5) {
  return;
 }
 $6 = (($1) + ($n<<2)|0);
 $7 = ((($3)) + 8|0);
 $j$0$i$i = -1;$nf$0$i$i = 0;$nl$0$i$i = $n;$ntry$0$i$i = 0;
 L4: while(1) {
  $8 = (($j$0$i$i) + 1)|0;
  $9 = ($8|0)<(4);
  if ($9) {
   $10 = (101384 + ($8<<2)|0);
   $11 = HEAP32[$10>>2]|0;
   $ntry$1$ph$i$i = $11;
  } else {
   $12 = (($ntry$0$i$i) + 2)|0;
   $ntry$1$ph$i$i = $12;
  }
  $13 = ($ntry$1$ph$i$i|0)==(2);
  if ($13) {
   $indvars$iv$in$i$i = $nf$0$i$i;$nl$1$i$i = $nl$0$i$i;
  } else {
   $nf$1$us$i$i = $nf$0$i$i;$nl$1$us$i$i = $nl$0$i$i;
   while(1) {
    $14 = (($nl$1$us$i$i|0) / ($ntry$1$ph$i$i|0))&-1;
    $15 = Math_imul($14, $ntry$1$ph$i$i)|0;
    $16 = ($nl$1$us$i$i|0)==($15|0);
    if (!($16)) {
     $j$0$i$i = $8;$nf$0$i$i = $nf$1$us$i$i;$nl$0$i$i = $nl$1$us$i$i;$ntry$0$i$i = $ntry$1$ph$i$i;
     continue L4;
    }
    $17 = (($nf$1$us$i$i) + 1)|0;
    $18 = (($nf$1$us$i$i) + 2)|0;
    $19 = (($3) + ($18<<2)|0);
    HEAP32[$19>>2] = $ntry$1$ph$i$i;
    $20 = ($14|0)==(1);
    if ($20) {
     $$lcssa = $17;$nf$1$us$i$i$lcssa37 = $nf$1$us$i$i;
     label = 15;
     break L4;
    } else {
     $nf$1$us$i$i = $17;$nl$1$us$i$i = $14;
    }
   }
  }
  while(1) {
   $indvars$iv$i$i = (($indvars$iv$in$i$i) + 1)|0;
   $21 = (($nl$1$i$i|0) / 2)&-1;
   $22 = $21 << 1;
   $23 = ($nl$1$i$i|0)==($22|0);
   if (!($23)) {
    $j$0$i$i = $8;$nf$0$i$i = $indvars$iv$in$i$i;$nl$0$i$i = $nl$1$i$i;$ntry$0$i$i = $ntry$1$ph$i$i;
    continue L4;
   }
   $24 = (($indvars$iv$in$i$i) + 2)|0;
   $25 = (($3) + ($24<<2)|0);
   HEAP32[$25>>2] = 2;
   $26 = ($indvars$iv$in$i$i|0)==(0);
   if (!($26)) {
    $27 = ($indvars$iv$in$i$i|0)<(1);
    if (!($27)) {
     $i$021$i$i = 1;
     while(1) {
      $28 = (($indvars$iv$i$i) - ($i$021$i$i))|0;
      $29 = (($28) + 1)|0;
      $30 = (($3) + ($29<<2)|0);
      $31 = HEAP32[$30>>2]|0;
      $32 = (($28) + 2)|0;
      $33 = (($3) + ($32<<2)|0);
      HEAP32[$33>>2] = $31;
      $34 = (($i$021$i$i) + 1)|0;
      $exitcond39$i$i = ($34|0)==($indvars$iv$i$i|0);
      if ($exitcond39$i$i) {
       break;
      } else {
       $i$021$i$i = $34;
      }
     }
    }
    HEAP32[$7>>2] = 2;
   }
   $35 = $nl$1$i$i & -2;
   $36 = ($35|0)==(2);
   if ($36) {
    $$lcssa$i$i = $indvars$iv$i$i;$$lcssa7$i$i = $26;$nf$1$lcssa6$i$i = $indvars$iv$in$i$i;
    break L4;
   } else {
    $indvars$iv$in$i$i = $indvars$iv$i$i;$nl$1$i$i = $21;
   }
  }
 }
 if ((label|0) == 15) {
  $37 = ($nf$1$us$i$i$lcssa37|0)==(0);
  $$lcssa$i$i = $$lcssa;$$lcssa7$i$i = $37;$nf$1$lcssa6$i$i = $nf$1$us$i$i$lcssa37;
 }
 HEAP32[$3>>2] = $n;
 $38 = ((($3)) + 4|0);
 HEAP32[$38>>2] = $$lcssa$i$i;
 $39 = (+($n|0));
 $40 = 6.2831854820251465 / $39;
 $$lcssa7$not$i$i = $$lcssa7$i$i ^ 1;
 $41 = ($nf$1$lcssa6$i$i|0)>(0);
 $or$cond$i$i = $41 & $$lcssa7$not$i$i;
 if ($or$cond$i$i) {
  $is$018$i$i = 0;$k1$019$i$i = 0;$l1$017$i$i = 1;
 } else {
  return;
 }
 while(1) {
  $42 = (($k1$019$i$i) + 2)|0;
  $43 = (($3) + ($42<<2)|0);
  $44 = HEAP32[$43>>2]|0;
  $45 = Math_imul($44, $l1$017$i$i)|0;
  $46 = (($n|0) / ($45|0))&-1;
  $47 = ($44|0)>(1);
  do {
   if ($47) {
    $48 = ($46|0)>(2);
    $49 = (($44) + -1)|0;
    $50 = Math_imul($46, $49)|0;
    if ($48) {
     $is$112$us$i$i = $is$018$i$i;$j$113$us$i$i = 0;$ld$011$us$i$i = 0;
    } else {
     $51 = (($50) + ($is$018$i$i))|0;
     $is$1$lcssa$i$i = $51;
     break;
    }
    while(1) {
     $52 = (($ld$011$us$i$i) + ($l1$017$i$i))|0;
     $53 = (+($52|0));
     $54 = $40 * $53;
     $fi$010$us$i$i = 0.0;$i$19$us$i$i = $is$112$us$i$i;$ii$08$us$i$i = 2;
     while(1) {
      $55 = $fi$010$us$i$i + 1.0;
      $56 = $54 * $55;
      $57 = $56;
      $58 = (+Math_cos((+$57)));
      $59 = $58;
      $60 = (($i$19$us$i$i) + 1)|0;
      $61 = (($6) + ($i$19$us$i$i<<2)|0);
      HEAPF32[$61>>2] = $59;
      $62 = (+Math_sin((+$57)));
      $63 = $62;
      $64 = (($i$19$us$i$i) + 2)|0;
      $65 = (($6) + ($60<<2)|0);
      HEAPF32[$65>>2] = $63;
      $66 = (($ii$08$us$i$i) + 2)|0;
      $67 = ($66|0)<($46|0);
      if ($67) {
       $fi$010$us$i$i = $55;$i$19$us$i$i = $64;$ii$08$us$i$i = $66;
      } else {
       break;
      }
     }
     $68 = (($is$112$us$i$i) + ($46))|0;
     $69 = (($j$113$us$i$i) + 1)|0;
     $exitcond37$i$i = ($69|0)==($49|0);
     if ($exitcond37$i$i) {
      break;
     } else {
      $is$112$us$i$i = $68;$j$113$us$i$i = $69;$ld$011$us$i$i = $52;
     }
    }
    $70 = (($50) + ($is$018$i$i))|0;
    $is$1$lcssa$i$i = $70;
   } else {
    $is$1$lcssa$i$i = $is$018$i$i;
   }
  } while(0);
  $71 = (($k1$019$i$i) + 1)|0;
  $exitcond38$i$i = ($71|0)==($nf$1$lcssa6$i$i|0);
  if ($exitcond38$i$i) {
   break;
  } else {
   $is$018$i$i = $is$1$lcssa$i$i;$k1$019$i$i = $71;$l1$017$i$i = $45;
  }
 }
 return;
}
function _vorbis_book_init_encode($c,$s) {
 $c = $c|0;
 $s = $s|0;
 var $$$i = 0.0, $$$i2 = 0.0, $$lcssa = 0, $$lcssa8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0.0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $acc$02$us$i = 0;
 var $acc1$03$us$i = 0, $exitcond = 0, $i$04$us$i = 0, $or$cond$us$i = 0, $rintf = 0.0, $rintf1 = 0.0, $vals$0$us$i = 0, $vals$0$us$i$lcssa = 0, $vals$1$us$i = 0, $vals$1$v$us$i = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 dest=$c; stop=dest+56|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $0 = ((($c)) + 12|0);
 HEAP32[$0>>2] = $s;
 $1 = ((($s)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($c)) + 4|0);
 HEAP32[$3>>2] = $2;
 $4 = ((($c)) + 8|0);
 HEAP32[$4>>2] = $2;
 $5 = HEAP32[$s>>2]|0;
 HEAP32[$c>>2] = $5;
 $6 = ((($s)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (__make_words($7,$2,0)|0);
 $9 = ((($c)) + 20|0);
 HEAP32[$9>>2] = $8;
 $10 = HEAP32[$1>>2]|0;
 $11 = HEAP32[$s>>2]|0;
 $12 = ($11|0)>(0);
 if (!($12)) {
  while(1) {
  }
 }
 $13 = (+($10|0));
 $14 = $13;
 $15 = (+($11|0));
 $16 = 1.0 / $15;
 $17 = $16;
 $18 = (+Math_pow((+$14),(+$17)));
 $19 = (+Math_floor((+$18)));
 $20 = (~~(($19)));
 $vals$0$us$i = $20;
 while(1) {
  $24 = (($vals$0$us$i) + 1)|0;
  $acc$02$us$i = 1;$acc1$03$us$i = 1;$i$04$us$i = 0;
  while(1) {
   $22 = Math_imul($acc$02$us$i, $vals$0$us$i)|0;
   $23 = Math_imul($acc1$03$us$i, $24)|0;
   $25 = (($i$04$us$i) + 1)|0;
   $exitcond = ($25|0)==($11|0);
   if ($exitcond) {
    $$lcssa = $22;$$lcssa8 = $23;
    break;
   } else {
    $acc$02$us$i = $22;$acc1$03$us$i = $23;$i$04$us$i = $25;
   }
  }
  $26 = ($$lcssa|0)<=($10|0);
  $27 = ($$lcssa8|0)>($10|0);
  $or$cond$us$i = $27 & $26;
  if ($or$cond$us$i) {
   $vals$0$us$i$lcssa = $vals$0$us$i;
   break;
  }
  $21 = ($$lcssa|0)>($10|0);
  $vals$1$v$us$i = $21 ? -1 : 1;
  $vals$1$us$i = (($vals$1$v$us$i) + ($vals$0$us$i))|0;
  $vals$0$us$i = $vals$1$us$i;
 }
 $28 = ((($c)) + 44|0);
 HEAP32[$28>>2] = $vals$0$us$i$lcssa;
 $29 = ((($s)) + 16|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = $30 & 2097151;
 $32 = (+($31|0));
 $33 = $30 >>> 21;
 $34 = $33 & 1023;
 $35 = ($30|0)<(0);
 $36 = -$32;
 $$$i = $35 ? $36 : $32;
 $37 = (($34) + -788)|0;
 $38 = (+_ldexp($$$i,$37));
 $39 = $38;
 $rintf = (+_rintf($39));
 $40 = (~~(($rintf)));
 $41 = ((($c)) + 48|0);
 HEAP32[$41>>2] = $40;
 $42 = ((($s)) + 20|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = $43 & 2097151;
 $45 = (+($44|0));
 $46 = $43 >>> 21;
 $47 = $46 & 1023;
 $48 = ($43|0)<(0);
 $49 = -$45;
 $$$i2 = $48 ? $49 : $45;
 $50 = (($47) + -788)|0;
 $51 = (+_ldexp($$$i2,$50));
 $52 = $51;
 $rintf1 = (+_rintf($52));
 $53 = (~~(($rintf1)));
 $54 = ((($c)) + 52|0);
 HEAP32[$54>>2] = $53;
 return 0;
}
function __make_words($l,$n,$sparsecount) {
 $l = $l|0;
 $n = $n|0;
 $sparsecount = $sparsecount|0;
 var $$4 = 0, $$count$0 = 0, $$lcssa52 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $9 = 0, $count$031 = 0, $count$3$ph = 0, $count$3$ph$lcssa = 0, $count$419$us = 0, $count$5$us = 0, $entry$027 = 0;
 var $exitcond = 0, $exitcond42 = 0, $i$032 = 0, $i$121 = 0, $i$220 = 0, $i$220$us = 0, $j$022$lcssa = 0, $j$02254 = 0, $j$1 = 0, $j$1$in26 = 0, $j$1$in26$phi = 0, $j$125 = 0, $j$128 = 0, $j$216 = 0, $j$216$us = 0, $marker = 0, $or$cond = 0, $temp$0$lcssa = 0, $temp$0$lcssa$us = 0, $temp$017 = 0;
 var $temp$017$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $marker = sp;
 $0 = ($sparsecount|0)!=(0);
 $1 = $0 ? $sparsecount : $n;
 $2 = $1 << 2;
 $3 = (_malloc($2)|0);
 _memset(($marker|0),0,132)|0;
 $4 = ($n|0)>(0);
 do {
  if ($4) {
   $5 = ((($marker)) + 4|0);
   $6 = ($sparsecount|0)==(0);
   $7 = $6&1;
   $count$031 = 0;$i$032 = 0;
   L3: while(1) {
    $8 = (($l) + ($i$032)|0);
    $9 = HEAP8[$8>>0]|0;
    $10 = $9 << 24 >> 24;
    $11 = ($9<<24>>24)>(0);
    L5: do {
     if ($11) {
      $12 = (($marker) + ($10<<2)|0);
      $13 = HEAP32[$12>>2]|0;
      $14 = ($9<<24>>24)>(31);
      $15 = $13 >>> $10;
      $16 = ($15|0)==(0);
      $or$cond = $14 | $16;
      if (!($or$cond)) {
       label = 15;
       break L3;
      }
      $17 = (($count$031) + 1)|0;
      $18 = (($3) + ($count$031<<2)|0);
      HEAP32[$18>>2] = $13;
      $19 = (($marker) + ($10<<2)|0);
      $20 = $13 & 1;
      $21 = ($20|0)==(0);
      L8: do {
       if ($21) {
        $31 = $13;$32 = $19;$j$02254 = $10;
        while(1) {
         $30 = (($31) + 1)|0;
         HEAP32[$32>>2] = $30;
         $33 = (($j$02254) + -1)|0;
         $34 = ($j$02254|0)>(1);
         if (!($34)) {
          break L8;
         }
         $$phi$trans$insert = (($marker) + ($33<<2)|0);
         $$pre = HEAP32[$$phi$trans$insert>>2]|0;
         $35 = (($marker) + ($33<<2)|0);
         $36 = $$pre & 1;
         $37 = ($36|0)==(0);
         if ($37) {
          $31 = $$pre;$32 = $35;$j$02254 = $33;
         } else {
          $$lcssa52 = $35;$j$022$lcssa = $33;
          label = 7;
          break;
         }
        }
       } else {
        $$lcssa52 = $19;$j$022$lcssa = $10;
        label = 7;
       }
      } while(0);
      do {
       if ((label|0) == 7) {
        label = 0;
        $23 = ($j$022$lcssa|0)==(1);
        if ($23) {
         $24 = HEAP32[$5>>2]|0;
         $25 = (($24) + 1)|0;
         HEAP32[$5>>2] = $25;
         break;
        } else {
         $26 = (($j$022$lcssa) + -1)|0;
         $27 = (($marker) + ($26<<2)|0);
         $28 = HEAP32[$27>>2]|0;
         $29 = $28 << 1;
         HEAP32[$$lcssa52>>2] = $29;
         break;
        }
       }
      } while(0);
      $j$125 = (($10) + 1)|0;
      $22 = ($j$125|0)<(33);
      if ($22) {
       $entry$027 = $13;$j$1$in26 = $10;$j$128 = $j$125;
       while(1) {
        $38 = (($marker) + ($j$128<<2)|0);
        $39 = HEAP32[$38>>2]|0;
        $40 = $39 >>> 1;
        $41 = ($40|0)==($entry$027|0);
        if (!($41)) {
         $count$3$ph = $17;
         break L5;
        }
        $42 = (($marker) + ($j$1$in26<<2)|0);
        $43 = HEAP32[$42>>2]|0;
        $44 = $43 << 1;
        HEAP32[$38>>2] = $44;
        $j$1 = (($j$128) + 1)|0;
        $45 = ($j$1|0)<(33);
        if ($45) {
         $j$1$in26$phi = $j$128;$entry$027 = $39;$j$128 = $j$1;$j$1$in26 = $j$1$in26$phi;
        } else {
         $count$3$ph = $17;
         break;
        }
       }
      } else {
       $count$3$ph = $17;
      }
     } else {
      $$count$0 = (($count$031) + ($7))|0;
      $count$3$ph = $$count$0;
     }
    } while(0);
    $46 = (($i$032) + 1)|0;
    $47 = ($46|0)<($n|0);
    if ($47) {
     $count$031 = $count$3$ph;$i$032 = $46;
    } else {
     $count$3$ph$lcssa = $count$3$ph;
     label = 17;
     break;
    }
   }
   if ((label|0) == 15) {
    _free($3);
    $$4 = 0;
    STACKTOP = sp;return ($$4|0);
   }
   else if ((label|0) == 17) {
    $48 = ($count$3$ph$lcssa|0)==(1);
    if (!($48)) {
     $i$121 = 1;
     label = 28;
     break;
    }
    $49 = ((($marker)) + 8|0);
    $50 = HEAP32[$49>>2]|0;
    $51 = ($50|0)==(2);
    if ($51) {
     break;
    } else {
     $i$121 = 1;
     label = 28;
     break;
    }
   }
  } else {
   $i$121 = 1;
   label = 28;
  }
 } while(0);
 L28: do {
  if ((label|0) == 28) {
   while(1) {
    label = 0;
    $70 = (($marker) + ($i$121<<2)|0);
    $71 = HEAP32[$70>>2]|0;
    $72 = (32 - ($i$121))|0;
    $73 = -1 >>> $72;
    $74 = $71 & $73;
    $75 = ($74|0)==(0);
    $69 = (($i$121) + 1)|0;
    if (!($75)) {
     break;
    }
    $68 = ($69|0)<(33);
    if ($68) {
     $i$121 = $69;
     label = 28;
    } else {
     break L28;
    }
   }
   _free($3);
   $$4 = 0;
   STACKTOP = sp;return ($$4|0);
  }
 } while(0);
 if (!($4)) {
  $$4 = $3;
  STACKTOP = sp;return ($$4|0);
 }
 if ($0) {
  $count$419$us = 0;$i$220$us = 0;
 } else {
  $i$220 = 0;
  while(1) {
   $76 = (($l) + ($i$220)|0);
   $77 = HEAP8[$76>>0]|0;
   $78 = ($77<<24>>24)>(0);
   $79 = (($3) + ($i$220<<2)|0);
   if ($78) {
    $80 = HEAP32[$79>>2]|0;
    $81 = $77 << 24 >> 24;
    $j$216 = 0;$temp$017 = 0;
    while(1) {
     $82 = $temp$017 << 1;
     $83 = $80 >>> $j$216;
     $84 = $83 & 1;
     $85 = $84 | $82;
     $86 = (($j$216) + 1)|0;
     $87 = ($86|0)<($81|0);
     if ($87) {
      $j$216 = $86;$temp$017 = $85;
     } else {
      $temp$0$lcssa = $85;
      break;
     }
    }
   } else {
    $temp$0$lcssa = 0;
   }
   $88 = (($i$220) + 1)|0;
   HEAP32[$79>>2] = $temp$0$lcssa;
   $exitcond42 = ($88|0)==($n|0);
   if ($exitcond42) {
    $$4 = $3;
    break;
   } else {
    $i$220 = $88;
   }
  }
  STACKTOP = sp;return ($$4|0);
 }
 while(1) {
  $52 = (($l) + ($i$220$us)|0);
  $53 = HEAP8[$52>>0]|0;
  $54 = ($53<<24>>24)>(0);
  if ($54) {
   $67 = (($3) + ($count$419$us<<2)|0);
   $61 = HEAP32[$67>>2]|0;
   $66 = $53 << 24 >> 24;
   $j$216$us = 0;$temp$017$us = 0;
   while(1) {
    $59 = $temp$017$us << 1;
    $60 = $61 >>> $j$216$us;
    $62 = $60 & 1;
    $63 = $62 | $59;
    $64 = (($j$216$us) + 1)|0;
    $65 = ($64|0)<($66|0);
    if ($65) {
     $j$216$us = $64;$temp$017$us = $63;
    } else {
     $temp$0$lcssa$us = $63;
     break;
    }
   }
  } else {
   $temp$0$lcssa$us = 0;
  }
  $55 = ($53<<24>>24)==(0);
  if ($55) {
   $count$5$us = $count$419$us;
  } else {
   $56 = (($count$419$us) + 1)|0;
   $57 = (($3) + ($count$419$us<<2)|0);
   HEAP32[$57>>2] = $temp$0$lcssa$us;
   $count$5$us = $56;
  }
  $58 = (($i$220$us) + 1)|0;
  $exitcond = ($58|0)==($n|0);
  if ($exitcond) {
   $$4 = $3;
   break;
  } else {
   $count$419$us = $count$5$us;$i$220$us = $58;
  }
 }
 STACKTOP = sp;return ($$4|0);
}
function __vp_psy_init($p,$vi,$gi,$n,$rate) {
 $p = $p|0;
 $vi = $vi|0;
 $gi = $gi|0;
 $n = $n|0;
 $rate = $rate|0;
 var $$inv = 0, $$lcssa = 0, $$lcssa74 = 0, $$not = 0, $$not66 = 0, $$phi$trans$insert = 0, $$phi$trans$insert51 = 0, $$pre = 0, $$pre50 = 0, $$pre52 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0;
 var $107 = 0.0, $108 = 0, $109 = 0.0, $11 = 0.0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0, $12 = 0.0, $120 = 0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0;
 var $125 = 0.0, $126 = 0.0, $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0.0, $130 = 0, $131 = 0.0, $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0.0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0.0, $150 = 0.0, $151 = 0.0, $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0.0, $160 = 0.0;
 var $161 = 0.0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0, $169 = 0, $17 = 0.0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0.0;
 var $18 = 0.0, $180 = 0.0, $181 = 0.0, $182 = 0.0, $183 = 0.0, $184 = 0.0, $185 = 0, $186 = 0, $187 = 0, $188 = 0.0, $189 = 0.0, $19 = 0.0, $190 = 0.0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0.0, $196 = 0.0, $197 = 0.0;
 var $198 = 0, $199 = 0.0, $2 = 0.0, $20 = 0, $200 = 0, $201 = 0.0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0.0, $212 = 0, $213 = 0.0, $214 = 0.0;
 var $215 = 0.0, $216 = 0.0, $217 = 0.0, $218 = 0.0, $219 = 0.0, $22 = 0.0, $220 = 0.0, $221 = 0.0, $222 = 0, $223 = 0, $224 = 0.0, $225 = 0.0, $226 = 0.0, $227 = 0.0, $228 = 0, $229 = 0, $23 = 0.0, $230 = 0.0, $231 = 0.0, $232 = 0.0;
 var $233 = 0, $234 = 0.0, $235 = 0.0, $236 = 0.0, $237 = 0.0, $238 = 0.0, $239 = 0, $24 = 0.0, $240 = 0, $241 = 0.0, $242 = 0.0, $243 = 0.0, $244 = 0, $245 = 0.0, $246 = 0.0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0.0, $250 = 0;
 var $251 = 0, $252 = 0.0, $253 = 0.0, $254 = 0.0, $255 = 0, $256 = 0.0, $257 = 0.0, $258 = 0.0, $259 = 0.0, $26 = 0, $260 = 0.0, $261 = 0, $262 = 0, $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0.0, $30 = 0.0, $31 = 0.0, $32 = 0.0;
 var $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0.0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0.0, $57 = 0, $58 = 0, $59 = 0, $6 = 0.0, $60 = 0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0;
 var $7 = 0.0, $70 = 0.0, $71 = 0, $72 = 0, $73 = 0.0, $74 = 0, $75 = 0.0, $76 = 0.0, $77 = 0, $78 = 0.0, $79 = 0.0, $8 = 0.0, $80 = 0, $81 = 0.0, $82 = 0, $83 = 0.0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $base$032 = 0.0, $brmerge = 0, $exitcond43 = 0, $exitcond45 = 0, $exitcond46 = 0, $exitcond47 = 0, $halfoc$0 = 0.0;
 var $halfoc$1 = 0.0, $hi$024 = 0, $hi$1$lcssa = 0, $hi$120 = 0, $i$0 = 0, $i$0$lcssa73 = 0, $i$0$ph$ph = 0, $i$126 = 0, $i$212 = 0, $i$410 = 0, $j$0$ph$ph = 0, $j$0$ph$ph$lcssa = 0, $j$133 = 0, $j$329 = 0, $lo$025 = 0, $lo$1$lcssa = 0, $lo$115 = 0, $or$cond = 0, dest = 0, label = 0;
 var sp = 0, stop = 0;
 sp = STACKTOP;
 dest=$p; stop=dest+48|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $0 = HEAP32[$gi>>2]|0;
 $1 = ((($p)) + 36|0);
 HEAP32[$1>>2] = $0;
 $2 = (+($0|0));
 $3 = $2 * 8.0;
 $4 = $3;
 $5 = (+Math_log((+$4)));
 $6 = $5 / 0.69314718055994529;
 $7 = (+_rint($6));
 $8 = $7 + -1.0;
 $9 = (~~(($8)));
 $10 = ((($p)) + 32|0);
 HEAP32[$10>>2] = $9;
 $11 = (+($rate|0));
 $12 = $11 * 0.25;
 $13 = $12;
 $14 = $13 * 0.5;
 $15 = (+($n|0));
 $16 = $14 / $15;
 $17 = (+Math_log((+$16)));
 $18 = $17 * 1.4426950216293335;
 $19 = $18 + -5.9657840728759766;
 $20 = (($9) + 1)|0;
 $21 = 1 << $20;
 $22 = (+($21|0));
 $23 = $19 * $22;
 $24 = (+($0|0));
 $25 = $23 - $24;
 $26 = (~~(($25)));
 $27 = ((($p)) + 28|0);
 HEAP32[$27>>2] = $26;
 $28 = (+($n|0));
 $29 = $28 + 0.25;
 $30 = $29 * $11;
 $31 = $30;
 $32 = $31 * 0.5;
 $33 = $32 / $15;
 $34 = (+Math_log((+$33)));
 $35 = $34 * 1.4426950216293335;
 $36 = $35 + -5.9657840728759766;
 $37 = $36 * $22;
 $38 = $37 + 0.5;
 $39 = (~~(($38)));
 $40 = (1 - ($26))|0;
 $41 = (($40) + ($39))|0;
 $42 = ((($p)) + 40|0);
 HEAP32[$42>>2] = $41;
 $43 = $n << 2;
 $44 = (_malloc($43)|0);
 $45 = ((($p)) + 16|0);
 HEAP32[$45>>2] = $44;
 $46 = (_malloc($43)|0);
 $47 = ((($p)) + 20|0);
 HEAP32[$47>>2] = $46;
 $48 = (_malloc($43)|0);
 $49 = ((($p)) + 24|0);
 HEAP32[$49>>2] = $48;
 $50 = ((($p)) + 4|0);
 HEAP32[$50>>2] = $vi;
 HEAP32[$p>>2] = $n;
 $51 = ((($p)) + 44|0);
 HEAP32[$51>>2] = $rate;
 $52 = ((($p)) + 48|0);
 HEAPF32[$52>>2] = 1.0;
 $53 = ($rate|0)<(26000);
 do {
  if ($53) {
   HEAPF32[$52>>2] = 0.0;
  } else {
   $54 = ($rate|0)<(38000);
   if ($54) {
    HEAPF32[$52>>2] = 0.93999999761581421;
    break;
   }
   $55 = ($rate|0)>(46000);
   if ($55) {
    HEAPF32[$52>>2] = 1.2749999761581421;
   }
  }
 } while(0);
 $56 = (+($rate|0));
 $i$0$ph$ph = 0;$j$0$ph$ph = 0;
 L9: while(1) {
  $$not66 = ($j$0$ph$ph|0)>=($n|0);
  $i$0 = $i$0$ph$ph;
  while(1) {
   $57 = ($i$0|0)<(87);
   if (!($57)) {
    $j$0$ph$ph$lcssa = $j$0$ph$ph;
    break L9;
   }
   $60 = (($i$0) + 1)|0;
   $61 = (+($60|0));
   $62 = $61 * 0.125;
   $63 = $62 + -2.0;
   $64 = $63 + 5.9657840728759766;
   $65 = $64 * 0.69314700365066528;
   $66 = (+Math_exp((+$65)));
   $67 = $66 * 2.0;
   $68 = $15 * $67;
   $69 = $68 / $56;
   $70 = (+_rint($69));
   $71 = (~~(($70)));
   $$not = ($71|0)<=($j$0$ph$ph|0);
   $brmerge = $$not | $$not66;
   if ($brmerge) {
    $i$0 = $60;
   } else {
    $$lcssa = $60;$$lcssa74 = $71;$i$0$lcssa73 = $i$0;
    break;
   }
  }
  $72 = (101400 + ($i$0$lcssa73<<2)|0);
  $73 = +HEAPF32[$72>>2];
  $74 = (101400 + ($$lcssa<<2)|0);
  $75 = +HEAPF32[$74>>2];
  $76 = $75 - $73;
  $77 = (($$lcssa74) - ($j$0$ph$ph))|0;
  $78 = (+($77|0));
  $79 = $76 / $78;
  $80 = HEAP32[$45>>2]|0;
  $base$032 = $73;$j$133 = $j$0$ph$ph;
  while(1) {
   $81 = $base$032 + 100.0;
   $82 = (($80) + ($j$133<<2)|0);
   HEAPF32[$82>>2] = $81;
   $83 = $79 + $base$032;
   $84 = (($j$133) + 1)|0;
   $85 = ($84|0)<($$lcssa74|0);
   $86 = ($84|0)<($n|0);
   $or$cond = $85 & $86;
   if ($or$cond) {
    $base$032 = $83;$j$133 = $84;
   } else {
    $i$0$ph$ph = $$lcssa;$j$0$ph$ph = $84;
    continue L9;
   }
  }
 }
 $58 = ($j$0$ph$ph$lcssa|0)<($n|0);
 if ($58) {
  $59 = HEAP32[$45>>2]|0;
  $j$329 = $j$0$ph$ph$lcssa;
  while(1) {
   $96 = (($j$329) + -1)|0;
   $97 = (($59) + ($96<<2)|0);
   $98 = HEAP32[$97>>2]|0;
   $99 = (($59) + ($j$329<<2)|0);
   HEAP32[$99>>2] = $98;
   $100 = (($j$329) + 1)|0;
   $exitcond47 = ($100|0)==($n|0);
   if ($exitcond47) {
    break;
   } else {
    $j$329 = $100;
   }
  }
 }
 $87 = ($n|0)>(0);
 if ($87) {
  $88 = $n << 1;
  $89 = (($rate|0) / ($88|0))&-1;
  $90 = ((($vi)) + 120|0);
  $91 = HEAP32[$90>>2]|0;
  $92 = HEAP32[$49>>2]|0;
  $93 = ((($vi)) + 124|0);
  $94 = ((($vi)) + 116|0);
  $95 = ((($vi)) + 112|0);
  $hi$024 = 1;$i$126 = 0;$lo$025 = -99;
  while(1) {
   $102 = Math_imul($89, $i$126)|0;
   $103 = (+($102|0));
   $104 = $103 * 7.3999998858198524E-4;
   $105 = $104;
   $106 = (+Math_atan((+$105)));
   $107 = $106 * 13.100000381469727;
   $108 = Math_imul($102, $102)|0;
   $109 = (+($108|0));
   $110 = $109 * 1.8499999754340024E-8;
   $111 = $110;
   $112 = (+Math_atan((+$111)));
   $113 = $112 * 2.2400000095367432;
   $114 = $107 + $113;
   $115 = $103 * 9.9999997473787516E-5;
   $116 = $115;
   $117 = $114 + $116;
   $118 = $117;
   $119 = (($91) + ($lo$025))|0;
   $120 = ($119|0)<($i$126|0);
   L27: do {
    if ($120) {
     $121 = +HEAPF32[$95>>2];
     $122 = $118 - $121;
     $123 = $122;
     $lo$115 = $lo$025;
     while(1) {
      $124 = Math_imul($lo$115, $89)|0;
      $125 = (+($124|0));
      $126 = $125 * 7.3999998858198524E-4;
      $127 = $126;
      $128 = (+Math_atan((+$127)));
      $129 = $128 * 13.100000381469727;
      $130 = Math_imul($124, $124)|0;
      $131 = (+($130|0));
      $132 = $131 * 1.8499999754340024E-8;
      $133 = $132;
      $134 = (+Math_atan((+$133)));
      $135 = $134 * 2.2400000095367432;
      $136 = $129 + $135;
      $137 = $125 * 9.9999997473787516E-5;
      $138 = $137;
      $139 = $138 + $136;
      $140 = $139 < $123;
      if (!($140)) {
       $lo$1$lcssa = $lo$115;
       break L27;
      }
      $144 = (($lo$115) + 1)|0;
      $145 = HEAP32[$90>>2]|0;
      $146 = (($145) + ($144))|0;
      $147 = ($146|0)<($i$126|0);
      if ($147) {
       $lo$115 = $144;
      } else {
       $lo$1$lcssa = $144;
       break;
      }
     }
    } else {
     $lo$1$lcssa = $lo$025;
    }
   } while(0);
   $141 = ($hi$024|0)>($n|0);
   L33: do {
    if ($141) {
     $hi$1$lcssa = $hi$024;
    } else {
     $142 = HEAP32[$93>>2]|0;
     $143 = (($142) + ($i$126))|0;
     $hi$120 = $hi$024;
     while(1) {
      $148 = ($hi$120|0)<($143|0);
      if (!($148)) {
       $149 = Math_imul($hi$120, $89)|0;
       $150 = (+($149|0));
       $151 = $150 * 7.3999998858198524E-4;
       $152 = $151;
       $153 = (+Math_atan((+$152)));
       $154 = $153 * 13.100000381469727;
       $155 = Math_imul($149, $149)|0;
       $156 = (+($155|0));
       $157 = $156 * 1.8499999754340024E-8;
       $158 = $157;
       $159 = (+Math_atan((+$158)));
       $160 = $159 * 2.2400000095367432;
       $161 = $154 + $160;
       $162 = $150 * 9.9999997473787516E-5;
       $163 = $162;
       $164 = $163 + $161;
       $165 = +HEAPF32[$94>>2];
       $166 = $118 + $165;
       $167 = $166;
       $168 = $164 < $167;
       if (!($168)) {
        $hi$1$lcssa = $hi$120;
        break L33;
       }
      }
      $169 = (($hi$120) + 1)|0;
      $170 = ($hi$120|0)<($n|0);
      if ($170) {
       $hi$120 = $169;
      } else {
       $hi$1$lcssa = $169;
       break;
      }
     }
    }
   } while(0);
   $171 = $lo$1$lcssa << 16;
   $172 = (($171) + -65537)|0;
   $173 = (($172) + ($hi$1$lcssa))|0;
   $174 = (($92) + ($i$126<<2)|0);
   HEAP32[$174>>2] = $173;
   $175 = (($i$126) + 1)|0;
   $exitcond46 = ($175|0)==($n|0);
   if ($exitcond46) {
    break;
   } else {
    $hi$024 = $hi$1$lcssa;$i$126 = $175;$lo$025 = $lo$1$lcssa;
   }
  }
  if ($87) {
   $101 = HEAP32[$47>>2]|0;
   $i$212 = 0;
   while(1) {
    $176 = (+($i$212|0));
    $177 = $176 + 0.25;
    $178 = $177;
    $179 = $178 * 0.5;
    $180 = $56 * $179;
    $181 = $180 / $15;
    $182 = (+Math_log((+$181)));
    $183 = $182 * 1.4426950216293335;
    $184 = $183 + -5.9657840728759766;
    $185 = HEAP32[$10>>2]|0;
    $186 = (($185) + 1)|0;
    $187 = 1 << $186;
    $188 = (+($187|0));
    $189 = $184 * $188;
    $190 = $189 + 0.5;
    $191 = (~~(($190)));
    $192 = (($101) + ($i$212<<2)|0);
    HEAP32[$192>>2] = $191;
    $193 = (($i$212) + 1)|0;
    $exitcond45 = ($193|0)==($n|0);
    if ($exitcond45) {
     break;
    } else {
     $i$212 = $193;
    }
   }
  }
 }
 $194 = ((($vi)) + 36|0);
 $195 = $56 * 0.5;
 $196 = $195 / $15;
 $197 = $196;
 $198 = ((($vi)) + 24|0);
 $199 = +HEAPF32[$198>>2];
 $200 = ((($vi)) + 28|0);
 $201 = +HEAPF32[$200>>2];
 $202 = (_setup_tone_curves($194,$197,$n,$199,$201)|0);
 $203 = ((($p)) + 8|0);
 HEAP32[$203>>2] = $202;
 $204 = (_malloc(12)|0);
 $205 = ((($p)) + 12|0);
 HEAP32[$205>>2] = $204;
 $206 = (_malloc($43)|0);
 HEAP32[$204>>2] = $206;
 $207 = (_malloc($43)|0);
 $208 = ((($204)) + 4|0);
 HEAP32[$208>>2] = $207;
 $209 = (_malloc($43)|0);
 $210 = ((($204)) + 8|0);
 HEAP32[$210>>2] = $209;
 if (!($87)) {
  return;
 }
 $211 = $15 * 2.0;
 $212 = HEAP32[$50>>2]|0;
 $$pre = HEAP32[$204>>2]|0;
 $$phi$trans$insert = ((($204)) + 4|0);
 $$pre50 = HEAP32[$$phi$trans$insert>>2]|0;
 $$phi$trans$insert51 = ((($204)) + 8|0);
 $$pre52 = HEAP32[$$phi$trans$insert51>>2]|0;
 $i$410 = 0;
 while(1) {
  $213 = (+($i$410|0));
  $214 = $213 + 0.5;
  $215 = $56 * $214;
  $216 = $215 / $211;
  $217 = (+Math_log((+$216)));
  $218 = $217 * 1.4426950216293335;
  $219 = $218 + -5.9657840728759766;
  $220 = $219 * 2.0;
  $221 = $220;
  $222 = $221 < 0.0;
  $halfoc$0 = $222 ? 0.0 : $221;
  $$inv = $halfoc$0 >= 16.0;
  $halfoc$1 = $$inv ? 16.0 : $halfoc$0;
  $223 = (~~(($halfoc$1)));
  $224 = (+($223|0));
  $225 = $halfoc$1 - $224;
  $226 = $225;
  $227 = 1.0 - $226;
  $228 = (($223) + 1)|0;
  $229 = (((($212)) + 132|0) + ($223<<2)|0);
  $230 = +HEAPF32[$229>>2];
  $231 = $230;
  $232 = $227 * $231;
  $233 = (((($212)) + 132|0) + ($228<<2)|0);
  $234 = +HEAPF32[$233>>2];
  $235 = $225 * $234;
  $236 = $235;
  $237 = $232 + $236;
  $238 = $237;
  $239 = (($$pre) + ($i$410<<2)|0);
  HEAPF32[$239>>2] = $238;
  $240 = (((($212)) + 200|0) + ($223<<2)|0);
  $241 = +HEAPF32[$240>>2];
  $242 = $241;
  $243 = $227 * $242;
  $244 = (((($212)) + 200|0) + ($228<<2)|0);
  $245 = +HEAPF32[$244>>2];
  $246 = $225 * $245;
  $247 = $246;
  $248 = $243 + $247;
  $249 = $248;
  $250 = (($$pre50) + ($i$410<<2)|0);
  HEAPF32[$250>>2] = $249;
  $251 = (((($212)) + 268|0) + ($223<<2)|0);
  $252 = +HEAPF32[$251>>2];
  $253 = $252;
  $254 = $227 * $253;
  $255 = (((($212)) + 268|0) + ($228<<2)|0);
  $256 = +HEAPF32[$255>>2];
  $257 = $225 * $256;
  $258 = $257;
  $259 = $254 + $258;
  $260 = $259;
  $261 = (($$pre52) + ($i$410<<2)|0);
  HEAPF32[$261>>2] = $260;
  $262 = (($i$410) + 1)|0;
  $exitcond43 = ($262|0)==($n|0);
  if ($exitcond43) {
   break;
  } else {
   $i$410 = $262;
  }
 }
 return;
}
function _setup_tone_curves($curveatt_dB,$binHz,$n,$center_boost,$center_decay_rate) {
 $curveatt_dB = $curveatt_dB|0;
 $binHz = +$binHz;
 $n = $n|0;
 $center_boost = +$center_boost;
 $center_decay_rate = +$center_decay_rate;
 var $$$$us$us$us = 0.0, $$$$us75 = 0.0, $$$us$us$us = 0.0, $$$us$us84 = 0.0, $$5 = 0, $$6 = 0, $$8 = 0, $$alloca_mul = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0.0, $106 = 0, $107 = 0.0, $108 = 0;
 var $109 = 0, $11 = 0.0, $110 = 0, $111 = 0.0, $112 = 0, $113 = 0.0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0;
 var $127 = 0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0.0, $131 = 0.0, $132 = 0.0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0, $138 = 0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0.0;
 var $145 = 0.0, $146 = 0.0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0, $161 = 0, $162 = 0.0;
 var $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0, $172 = 0.0, $173 = 0.0, $174 = 0.0, $175 = 0.0, $176 = 0.0, $177 = 0.0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0.0, $188 = 0, $189 = 0.0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0.0, $197 = 0, $198 = 0, $199 = 0;
 var $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0.0, $212 = 0.0, $213 = 0.0, $214 = 0.0, $215 = 0, $216 = 0.0;
 var $217 = 0.0, $218 = 0.0, $219 = 0.0, $22 = 0, $220 = 0.0, $221 = 0.0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0.0, $232 = 0, $233 = 0.0, $234 = 0;
 var $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0.0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0.0, $246 = 0, $247 = 0.0, $248 = 0.0, $249 = 0.0, $25 = 0, $250 = 0.0, $251 = 0.0, $252 = 0.0;
 var $253 = 0.0, $254 = 0.0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0;
 var $271 = 0.0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0.0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0.0, $281 = 0, $282 = 0, $283 = 0, $284 = 0.0, $285 = 0, $286 = 0, $287 = 0.0, $288 = 0, $289 = 0;
 var $29 = 0, $290 = 0.0, $291 = 0, $292 = 0, $293 = 0.0, $294 = 0, $295 = 0, $296 = 0.0, $297 = 0, $298 = 0, $299 = 0.0, $3 = 0, $30 = 0.0, $300 = 0, $301 = 0, $302 = 0.0, $303 = 0, $304 = 0, $305 = 0.0, $306 = 0;
 var $307 = 0, $308 = 0.0, $309 = 0, $31 = 0.0, $310 = 0, $311 = 0.0, $312 = 0, $313 = 0, $314 = 0.0, $315 = 0, $316 = 0, $317 = 0.0, $318 = 0, $319 = 0, $32 = 0.0, $320 = 0.0, $321 = 0, $322 = 0, $323 = 0.0, $324 = 0;
 var $325 = 0, $326 = 0, $327 = 0.0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0.0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0.0, $34 = 0, $340 = 0, $341 = 0, $342 = 0;
 var $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0, $45 = 0, $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0.0, $50 = 0, $51 = 0.0, $52 = 0.0;
 var $53 = 0.0, $54 = 0, $55 = 0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0.0, $63 = 0, $64 = 0, $65 = 0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0.0;
 var $71 = 0.0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0, $82 = 0.0, $83 = 0.0, $84 = 0, $85 = 0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0;
 var $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0, $93 = 0.0, $94 = 0.0, $95 = 0, $96 = 0, $97 = 0.0, $98 = 0, $99 = 0.0, $ath = 0, $athc = 0, $exitcond = 0, $exitcond$i = 0, $exitcond$i11 = 0, $exitcond$i13 = 0, $exitcond$i16 = 0, $exitcond$i18 = 0, $exitcond103 = 0;
 var $exitcond104 = 0, $exitcond105 = 0, $exitcond106 = 0, $exitcond107 = 0, $exitcond109 = 0, $exitcond111 = 0, $exitcond112 = 0, $exitcond113 = 0, $exitcond114 = 0, $exitcond115 = 0, $exitcond116 = 0, $exitcond117 = 0, $exitcond118 = 0, $exitcond119 = 0, $exitcond97 = 0, $exitcond98 = 0, $hi_bin$0 = 0, $hi_bin3$0 = 0, $i$01$i = 0, $i$01$i10 = 0;
 var $i$01$i12 = 0, $i$01$i15 = 0, $i$01$i17 = 0, $i$070 = 0, $i$1$ = 0, $i$161 = 0, $ispos = 0, $ispos$us$us$us = 0, $ispos$us$us82 = 0, $ispos$us72 = 0, $j$064 = 0, $j$10$lcssa = 0, $j$1056 = 0, $j$267 = 0, $j$267$us = 0, $j$368 = 0, $j$469 = 0, $j$532 = 0, $j$636 = 0, $j$749 = 0;
 var $j$854 = 0, $j$9$lcssa = 0.0, $k$166 = 0, $k$166$us$us$us = 0, $k$166$us$us81 = 0, $k$166$us71 = 0, $k$241 = 0, $l$038 = 0, $l$2$lcssa = 0, $l$2$lcssa$lcssa = 0, $l$234 = 0, $l$339 = 0, $l1$051 = 0, $l1$2$lcssa = 0, $l1$2$lcssa$lcssa = 0, $l1$245 = 0, $l1$352 = 0, $lo_bin$1 = 0, $lo_bin$1$l$0 = 0, $lo_bin2$1 = 0;
 var $lo_bin2$1$l1$0 = 0, $lo_curve$1 = 0, $m$057 = 0, $min$1 = 0.0, $min$1$1 = 0.0, $min$1$2 = 0.0, $min$1$3 = 0.0, $n$hi_bin$0 = 0, $n$hi_bin3$0 = 0, $neg = 0, $neg$us$us$us = 0, $neg$us$us83 = 0, $neg$us73 = 0, $or$cond7 = 0, $or$cond733 = 0, $or$cond9 = 0, $or$cond944 = 0, $workc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32480|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ath = sp + 32256|0;
 $workc = sp + 1792|0;
 $athc = sp;
 $0 = $n << 2;
 $$alloca_mul = $0;
 $1 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $2 = (_malloc(68)|0);
 _memset(($workc|0),0,30464)|0;
 $3 = $center_boost > 0.0;
 $4 = $center_boost < 0.0;
 $i$070 = 0;
 while(1) {
  $7 = $i$070 << 2;
  $j$064 = 0;
  while(1) {
   $8 = (($j$064) + ($7))|0;
   $9 = ($8|0)<(88);
   if ($9) {
    $10 = (101400 + ($8<<2)|0);
    $11 = +HEAPF32[$10>>2];
    $min$1 = $11;
   } else {
    $min$1 = -30.0;
   }
   $12 = (($8) + 1)|0;
   $13 = ($12|0)<(88);
   if ($13) {
    $326 = (101400 + ($12<<2)|0);
    $327 = +HEAPF32[$326>>2];
    $328 = $min$1 > $327;
    if ($328) {
     $min$1$1 = $327;
    } else {
     $min$1$1 = $min$1;
    }
   } else {
    $325 = $min$1 > -30.0;
    if ($325) {
     $min$1$1 = -30.0;
    } else {
     $min$1$1 = $min$1;
    }
   }
   $329 = (($8) + 2)|0;
   $330 = ($329|0)<(88);
   if ($330) {
    $332 = (101400 + ($329<<2)|0);
    $333 = +HEAPF32[$332>>2];
    $334 = $min$1$1 > $333;
    if ($334) {
     $min$1$2 = $333;
    } else {
     $min$1$2 = $min$1$1;
    }
   } else {
    $331 = $min$1$1 > -30.0;
    if ($331) {
     $min$1$2 = -30.0;
    } else {
     $min$1$2 = $min$1$1;
    }
   }
   $335 = (($8) + 3)|0;
   $336 = ($335|0)<(88);
   if ($336) {
    $338 = (101400 + ($335<<2)|0);
    $339 = +HEAPF32[$338>>2];
    $340 = $min$1$2 > $339;
    if ($340) {
     $min$1$3 = $339;
    } else {
     $min$1$3 = $min$1$2;
    }
   } else {
    $337 = $min$1$2 > -30.0;
    if ($337) {
     $min$1$3 = -30.0;
    } else {
     $min$1$3 = $min$1$2;
    }
   }
   $341 = (($ath) + ($j$064<<2)|0);
   HEAPF32[$341>>2] = $min$1$3;
   $342 = (($j$064) + 1)|0;
   $exitcond109 = ($342|0)==(56);
   if ($exitcond109) {
    break;
   } else {
    $j$064 = $342;
   }
  }
  $14 = (((($workc) + (($i$070*1792)|0)|0)) + 448|0);
  $15 = (101752 + (($i$070*1344)|0)|0);
  _memcpy(($14|0),($15|0),224)|0;
  $16 = (((($workc) + (($i$070*1792)|0)|0)) + 672|0);
  $17 = (((101752 + (($i$070*1344)|0)|0)) + 224|0);
  _memcpy(($16|0),($17|0),224)|0;
  $18 = (((($workc) + (($i$070*1792)|0)|0)) + 896|0);
  $19 = (((101752 + (($i$070*1344)|0)|0)) + 448|0);
  _memcpy(($18|0),($19|0),224)|0;
  $20 = (((($workc) + (($i$070*1792)|0)|0)) + 1120|0);
  $21 = (((101752 + (($i$070*1344)|0)|0)) + 672|0);
  _memcpy(($20|0),($21|0),224)|0;
  $22 = (((($workc) + (($i$070*1792)|0)|0)) + 1344|0);
  $23 = (((101752 + (($i$070*1344)|0)|0)) + 896|0);
  _memcpy(($22|0),($23|0),224)|0;
  $24 = (((($workc) + (($i$070*1792)|0)|0)) + 1568|0);
  $25 = (((101752 + (($i$070*1344)|0)|0)) + 1120|0);
  _memcpy(($24|0),($25|0),224)|0;
  $26 = (($workc) + (($i$070*1792)|0)|0);
  _memcpy(($26|0),($15|0),224)|0;
  $27 = (((($workc) + (($i$070*1792)|0)|0)) + 224|0);
  _memcpy(($27|0),($15|0),224)|0;
  if ($3) {
   $j$267$us = 0;
   while(1) {
    if ($4) {
     $k$166$us$us$us = 0;
     while(1) {
      $39 = (16 - ($k$166$us$us$us))|0;
      $ispos$us$us$us = ($39|0)>(-1);
      $neg$us$us$us = (0 - ($39))|0;
      $40 = $ispos$us$us$us ? $39 : $neg$us$us$us;
      $41 = (+($40|0));
      $42 = $41 * $center_decay_rate;
      $43 = $42 + $center_boost;
      $44 = $43 < 0.0;
      $$$us$us$us = $44 ? 0.0 : $43;
      $45 = $$$us$us$us > 0.0;
      $$$$us$us$us = $45 ? 0.0 : $$$us$us$us;
      $46 = (((($workc) + (($i$070*1792)|0)|0) + (($j$267$us*224)|0)|0) + ($k$166$us$us$us<<2)|0);
      $47 = +HEAPF32[$46>>2];
      $48 = $$$$us$us$us + $47;
      HEAPF32[$46>>2] = $48;
      $49 = (($k$166$us$us$us) + 1)|0;
      $exitcond115 = ($49|0)==(56);
      if ($exitcond115) {
       break;
      } else {
       $k$166$us$us$us = $49;
      }
     }
    } else {
     $k$166$us$us81 = 0;
     while(1) {
      $28 = (16 - ($k$166$us$us81))|0;
      $ispos$us$us82 = ($28|0)>(-1);
      $neg$us$us83 = (0 - ($28))|0;
      $29 = $ispos$us$us82 ? $28 : $neg$us$us83;
      $30 = (+($29|0));
      $31 = $30 * $center_decay_rate;
      $32 = $31 + $center_boost;
      $33 = $32 < 0.0;
      $$$us$us84 = $33 ? 0.0 : $32;
      $34 = (((($workc) + (($i$070*1792)|0)|0) + (($j$267$us*224)|0)|0) + ($k$166$us$us81<<2)|0);
      $35 = +HEAPF32[$34>>2];
      $36 = $$$us$us84 + $35;
      HEAPF32[$34>>2] = $36;
      $37 = (($k$166$us$us81) + 1)|0;
      $exitcond114 = ($37|0)==(56);
      if ($exitcond114) {
       break;
      } else {
       $k$166$us$us81 = $37;
      }
     }
    }
    $38 = (($j$267$us) + 1)|0;
    $exitcond116 = ($38|0)==(8);
    if ($exitcond116) {
     break;
    } else {
     $j$267$us = $38;
    }
   }
  } else {
   $j$267 = 0;
   while(1) {
    if ($4) {
     $k$166$us71 = 0;
     while(1) {
      $54 = (16 - ($k$166$us71))|0;
      $ispos$us72 = ($54|0)>(-1);
      $neg$us73 = (0 - ($54))|0;
      $55 = $ispos$us72 ? $54 : $neg$us73;
      $56 = (+($55|0));
      $57 = $56 * $center_decay_rate;
      $58 = $57 + $center_boost;
      $59 = $58 > 0.0;
      $$$$us75 = $59 ? 0.0 : $58;
      $60 = (((($workc) + (($i$070*1792)|0)|0) + (($j$267*224)|0)|0) + ($k$166$us71<<2)|0);
      $61 = +HEAPF32[$60>>2];
      $62 = $$$$us75 + $61;
      HEAPF32[$60>>2] = $62;
      $63 = (($k$166$us71) + 1)|0;
      $exitcond112 = ($63|0)==(56);
      if ($exitcond112) {
       break;
      } else {
       $k$166$us71 = $63;
      }
     }
    } else {
     $k$166 = 0;
     while(1) {
      $64 = (16 - ($k$166))|0;
      $ispos = ($64|0)>(-1);
      $neg = (0 - ($64))|0;
      $65 = $ispos ? $64 : $neg;
      $66 = (+($65|0));
      $67 = $66 * $center_decay_rate;
      $68 = $67 + $center_boost;
      $69 = (((($workc) + (($i$070*1792)|0)|0) + (($j$267*224)|0)|0) + ($k$166<<2)|0);
      $70 = +HEAPF32[$69>>2];
      $71 = $68 + $70;
      HEAPF32[$69>>2] = $71;
      $72 = (($k$166) + 1)|0;
      $exitcond111 = ($72|0)==(56);
      if ($exitcond111) {
       break;
      } else {
       $k$166 = $72;
      }
     }
    }
    $73 = (($j$267) + 1)|0;
    $exitcond113 = ($73|0)==(8);
    if ($exitcond113) {
     break;
    } else {
     $j$267 = $73;
    }
   }
  }
  $50 = (($curveatt_dB) + ($i$070<<2)|0);
  $51 = +HEAPF32[$50>>2];
  $52 = $51;
  $53 = $52 + 100.0;
  $j$368 = 0;
  while(1) {
   $74 = ($j$368|0)<(2);
   $75 = $74 ? 2 : $j$368;
   $76 = (+($75|0));
   $77 = $76 * 10.0;
   $78 = $53 - $77;
   $79 = $78 + -30.0;
   $80 = $79;
   $i$01$i = 0;
   while(1) {
    $81 = (((($workc) + (($i$070*1792)|0)|0) + (($j$368*224)|0)|0) + ($i$01$i<<2)|0);
    $82 = +HEAPF32[$81>>2];
    $83 = $80 + $82;
    HEAPF32[$81>>2] = $83;
    $84 = (($i$01$i) + 1)|0;
    $exitcond$i = ($84|0)==(56);
    if ($exitcond$i) {
     break;
    } else {
     $i$01$i = $84;
    }
   }
   $85 = (($athc) + (($j$368*224)|0)|0);
   _memcpy(($85|0),($ath|0),224)|0;
   $86 = (+($j$368|0));
   $87 = $86 * 10.0;
   $88 = $87;
   $89 = 100.0 - $88;
   $90 = $89 + -30.0;
   $91 = $90;
   $i$01$i17 = 0;
   while(1) {
    $92 = ((($athc) + (($j$368*224)|0)|0) + ($i$01$i17<<2)|0);
    $93 = +HEAPF32[$92>>2];
    $94 = $91 + $93;
    HEAPF32[$92>>2] = $94;
    $95 = (($i$01$i17) + 1)|0;
    $exitcond$i18 = ($95|0)==(56);
    if ($exitcond$i18) {
     $i$01$i15 = 0;
     break;
    } else {
     $i$01$i17 = $95;
    }
   }
   while(1) {
    $96 = (((($workc) + (($i$070*1792)|0)|0) + (($j$368*224)|0)|0) + ($i$01$i15<<2)|0);
    $97 = +HEAPF32[$96>>2];
    $98 = ((($athc) + (($j$368*224)|0)|0) + ($i$01$i15<<2)|0);
    $99 = +HEAPF32[$98>>2];
    $100 = $97 > $99;
    if ($100) {
     HEAPF32[$98>>2] = $97;
    }
    $101 = (($i$01$i15) + 1)|0;
    $exitcond$i16 = ($101|0)==(56);
    if ($exitcond$i16) {
     break;
    } else {
     $i$01$i15 = $101;
    }
   }
   $102 = (($j$368) + 1)|0;
   $exitcond117 = ($102|0)==(8);
   if ($exitcond117) {
    $j$469 = 1;
    break;
   } else {
    $j$368 = $102;
   }
  }
  while(1) {
   $103 = (($j$469) + -1)|0;
   $i$01$i12 = 0;
   while(1) {
    $104 = ((($athc) + (($103*224)|0)|0) + ($i$01$i12<<2)|0);
    $105 = +HEAPF32[$104>>2];
    $106 = ((($athc) + (($j$469*224)|0)|0) + ($i$01$i12<<2)|0);
    $107 = +HEAPF32[$106>>2];
    $108 = $105 < $107;
    if ($108) {
     HEAPF32[$106>>2] = $105;
    }
    $109 = (($i$01$i12) + 1)|0;
    $exitcond$i13 = ($109|0)==(56);
    if ($exitcond$i13) {
     $i$01$i10 = 0;
     break;
    } else {
     $i$01$i12 = $109;
    }
   }
   while(1) {
    $110 = ((($athc) + (($j$469*224)|0)|0) + ($i$01$i10<<2)|0);
    $111 = +HEAPF32[$110>>2];
    $112 = (((($workc) + (($i$070*1792)|0)|0) + (($j$469*224)|0)|0) + ($i$01$i10<<2)|0);
    $113 = +HEAPF32[$112>>2];
    $114 = $111 < $113;
    if ($114) {
     HEAPF32[$112>>2] = $111;
    }
    $115 = (($i$01$i10) + 1)|0;
    $exitcond$i11 = ($115|0)==(56);
    if ($exitcond$i11) {
     break;
    } else {
     $i$01$i10 = $115;
    }
   }
   $116 = (($j$469) + 1)|0;
   $exitcond118 = ($116|0)==(8);
   if ($exitcond118) {
    break;
   } else {
    $j$469 = $116;
   }
  }
  $117 = (($i$070) + 1)|0;
  $exitcond119 = ($117|0)==(17);
  if ($exitcond119) {
   break;
  } else {
   $i$070 = $117;
  }
 }
 $5 = $binHz;
 $6 = ($n|0)>(0);
 $i$161 = 0;
 while(1) {
  $118 = (_malloc(32)|0);
  $119 = (($2) + ($i$161<<2)|0);
  HEAP32[$119>>2] = $118;
  $120 = (+($i$161|0));
  $121 = $120 * 0.5;
  $122 = $121 + 5.9657840728759766;
  $123 = $122 * 0.69314700365066528;
  $124 = (+Math_exp((+$123)));
  $125 = $124 / $5;
  $126 = (+Math_floor((+$125)));
  $127 = (~~(($126)));
  $128 = (+($127|0));
  $129 = $128 * $binHz;
  $130 = $129 + 1.0;
  $131 = $130;
  $132 = (+Math_log((+$131)));
  $133 = $132 * 1.4426950216293335;
  $134 = $133 + -5.9657840728759766;
  $135 = $134 * 2.0;
  $136 = (+Math_ceil((+$135)));
  $137 = (~~(($136)));
  $138 = (($127) + 1)|0;
  $139 = (+($138|0));
  $140 = $139 * $binHz;
  $141 = $140;
  $142 = (+Math_log((+$141)));
  $143 = $142 * 1.4426950216293335;
  $144 = $143 + -5.9657840728759766;
  $145 = $144 * 2.0;
  $146 = (+Math_floor((+$145)));
  $147 = (~~(($146)));
  $148 = ($137|0)>($i$161|0);
  $i$1$ = $148 ? $i$161 : $137;
  $149 = ($i$1$|0)<(0);
  $lo_curve$1 = $149 ? 0 : $i$1$;
  $150 = ($147|0)>(16);
  $$5 = $150 ? 16 : $147;
  $151 = ($lo_curve$1|0)>($$5|0);
  $152 = (($i$161) + 1)|0;
  $153 = ($152|0)<(17);
  $m$057 = 0;
  while(1) {
   $154 = (_malloc(232)|0);
   $155 = (($118) + ($m$057<<2)|0);
   HEAP32[$155>>2] = $154;
   if ($6) {
    $j$532 = 0;
    while(1) {
     $156 = (($1) + ($j$532<<2)|0);
     HEAPF32[$156>>2] = 999.0;
     $157 = (($j$532) + 1)|0;
     $exitcond = ($157|0)==($n|0);
     if ($exitcond) {
      break;
     } else {
      $j$532 = $157;
     }
    }
   }
   if (!($151)) {
    $k$241 = $lo_curve$1;
    while(1) {
     $158 = (+($k$241|0));
     $159 = $158 * 0.5;
     $j$636 = 0;$l$038 = 0;
     while(1) {
      $163 = (+($j$636|0));
      $164 = $163 * 0.125;
      $165 = $159 + $164;
      $166 = $165 + -2.0625;
      $167 = $166 + 5.9657840728759766;
      $168 = $167 * 0.69314700365066528;
      $169 = (+Math_exp((+$168)));
      $170 = $169 / $5;
      $171 = (~~(($170)));
      $172 = $165 + -1.9375;
      $173 = $172 + 5.9657840728759766;
      $174 = $173 * 0.69314700365066528;
      $175 = (+Math_exp((+$174)));
      $176 = $175 / $5;
      $177 = $176 + 1.0;
      $178 = (~~(($177)));
      $179 = ($171|0)<(0);
      $$6 = $179 ? 0 : $171;
      $180 = ($$6|0)>($n|0);
      $lo_bin$1 = $180 ? $n : $$6;
      $181 = ($lo_bin$1|0)<($l$038|0);
      $lo_bin$1$l$0 = $181 ? $lo_bin$1 : $l$038;
      $182 = ($178|0)<(0);
      $hi_bin$0 = $182 ? 0 : $178;
      $183 = ($hi_bin$0|0)>($n|0);
      $n$hi_bin$0 = $183 ? $n : $hi_bin$0;
      $184 = ($lo_bin$1$l$0|0)<($n$hi_bin$0|0);
      $185 = ($lo_bin$1$l$0|0)<($n|0);
      $or$cond733 = $184 & $185;
      if ($or$cond733) {
       $186 = (((($workc) + (($k$241*1792)|0)|0) + (($m$057*224)|0)|0) + ($j$636<<2)|0);
       $187 = +HEAPF32[$186>>2];
       $l$234 = $lo_bin$1$l$0;
       while(1) {
        $188 = (($1) + ($l$234<<2)|0);
        $189 = +HEAPF32[$188>>2];
        $190 = $189 > $187;
        if ($190) {
         HEAPF32[$188>>2] = $187;
        }
        $191 = (($l$234) + 1)|0;
        $192 = ($191|0)<($n$hi_bin$0|0);
        $193 = ($191|0)<($n|0);
        $or$cond7 = $192 & $193;
        if ($or$cond7) {
         $l$234 = $191;
        } else {
         $l$2$lcssa = $191;
         break;
        }
       }
      } else {
       $l$2$lcssa = $lo_bin$1$l$0;
      }
      $194 = (($j$636) + 1)|0;
      $exitcond97 = ($194|0)==(56);
      if ($exitcond97) {
       $l$2$lcssa$lcssa = $l$2$lcssa;
       break;
      } else {
       $j$636 = $194;$l$038 = $l$2$lcssa;
      }
     }
     $160 = ($l$2$lcssa$lcssa|0)<($n|0);
     if ($160) {
      $161 = ((((($workc) + (($k$241*1792)|0)|0) + (($m$057*224)|0)|0)) + 220|0);
      $162 = +HEAPF32[$161>>2];
      $l$339 = $l$2$lcssa$lcssa;
      while(1) {
       $195 = (($1) + ($l$339<<2)|0);
       $196 = +HEAPF32[$195>>2];
       $197 = $196 > $162;
       if ($197) {
        HEAPF32[$195>>2] = $162;
       }
       $198 = (($l$339) + 1)|0;
       $exitcond98 = ($198|0)==($n|0);
       if ($exitcond98) {
        break;
       } else {
        $l$339 = $198;
       }
      }
     }
     $199 = (($k$241) + 1)|0;
     $200 = ($k$241|0)<($$5|0);
     if ($200) {
      $k$241 = $199;
     } else {
      break;
     }
    }
   }
   if ($153) {
    $j$749 = 0;$l1$051 = 0;
    while(1) {
     $207 = (+($j$749|0));
     $208 = $207 * 0.125;
     $209 = $121 + $208;
     $210 = $209 + -2.0625;
     $211 = $210 + 5.9657840728759766;
     $212 = $211 * 0.69314700365066528;
     $213 = (+Math_exp((+$212)));
     $214 = $213 / $5;
     $215 = (~~(($214)));
     $216 = $209 + -1.9375;
     $217 = $216 + 5.9657840728759766;
     $218 = $217 * 0.69314700365066528;
     $219 = (+Math_exp((+$218)));
     $220 = $219 / $5;
     $221 = $220 + 1.0;
     $222 = (~~(($221)));
     $223 = ($215|0)<(0);
     $$8 = $223 ? 0 : $215;
     $224 = ($$8|0)>($n|0);
     $lo_bin2$1 = $224 ? $n : $$8;
     $225 = ($lo_bin2$1|0)<($l1$051|0);
     $lo_bin2$1$l1$0 = $225 ? $lo_bin2$1 : $l1$051;
     $226 = ($222|0)<(0);
     $hi_bin3$0 = $226 ? 0 : $222;
     $227 = ($hi_bin3$0|0)>($n|0);
     $n$hi_bin3$0 = $227 ? $n : $hi_bin3$0;
     $228 = ($lo_bin2$1$l1$0|0)<($n$hi_bin3$0|0);
     $229 = ($lo_bin2$1$l1$0|0)<($n|0);
     $or$cond944 = $228 & $229;
     if ($or$cond944) {
      $230 = (((($workc) + (($152*1792)|0)|0) + (($m$057*224)|0)|0) + ($j$749<<2)|0);
      $231 = +HEAPF32[$230>>2];
      $l1$245 = $lo_bin2$1$l1$0;
      while(1) {
       $232 = (($1) + ($l1$245<<2)|0);
       $233 = +HEAPF32[$232>>2];
       $234 = $233 > $231;
       if ($234) {
        HEAPF32[$232>>2] = $231;
       }
       $235 = (($l1$245) + 1)|0;
       $236 = ($235|0)<($n$hi_bin3$0|0);
       $237 = ($235|0)<($n|0);
       $or$cond9 = $236 & $237;
       if ($or$cond9) {
        $l1$245 = $235;
       } else {
        $l1$2$lcssa = $235;
        break;
       }
      }
     } else {
      $l1$2$lcssa = $lo_bin2$1$l1$0;
     }
     $238 = (($j$749) + 1)|0;
     $exitcond103 = ($238|0)==(56);
     if ($exitcond103) {
      $l1$2$lcssa$lcssa = $l1$2$lcssa;
      break;
     } else {
      $j$749 = $238;$l1$051 = $l1$2$lcssa;
     }
    }
    $204 = ($l1$2$lcssa$lcssa|0)<($n|0);
    if ($204) {
     $205 = ((((($workc) + (($152*1792)|0)|0) + (($m$057*224)|0)|0)) + 220|0);
     $206 = +HEAPF32[$205>>2];
     $l1$352 = $l1$2$lcssa$lcssa;
     while(1) {
      $239 = (($1) + ($l1$352<<2)|0);
      $240 = +HEAPF32[$239>>2];
      $241 = $240 > $206;
      if ($241) {
       HEAPF32[$239>>2] = $206;
      }
      $242 = (($l1$352) + 1)|0;
      $exitcond104 = ($242|0)==($n|0);
      if ($exitcond104) {
       break;
      } else {
       $l1$352 = $242;
      }
     }
    }
   }
   $201 = (($118) + ($m$057<<2)|0);
   $202 = (($118) + ($m$057<<2)|0);
   $203 = (($118) + ($m$057<<2)|0);
   $j$854 = 0;
   while(1) {
    $247 = (+($j$854|0));
    $248 = $247 * 0.125;
    $249 = $121 + $248;
    $250 = $249 + -2.0;
    $251 = $250 + 5.9657840728759766;
    $252 = $251 * 0.69314700365066528;
    $253 = (+Math_exp((+$252)));
    $254 = $253 / $5;
    $255 = (~~(($254)));
    $256 = ($255|0)<(0);
    do {
     if ($256) {
      $257 = (($j$854) + 2)|0;
      $258 = HEAP32[$201>>2]|0;
      $259 = (($258) + ($257<<2)|0);
      HEAPF32[$259>>2] = -999.0;
     } else {
      $260 = ($255|0)<($n|0);
      if ($260) {
       $264 = (($1) + ($255<<2)|0);
       $265 = HEAP32[$264>>2]|0;
       $266 = (($j$854) + 2)|0;
       $267 = HEAP32[$202>>2]|0;
       $268 = (($267) + ($266<<2)|0);
       HEAP32[$268>>2] = $265;
       break;
      } else {
       $261 = (($j$854) + 2)|0;
       $262 = HEAP32[$203>>2]|0;
       $263 = (($262) + ($261<<2)|0);
       HEAPF32[$263>>2] = -999.0;
       break;
      }
     }
    } while(0);
    $269 = (($j$854) + 1)|0;
    $exitcond105 = ($269|0)==(56);
    if ($exitcond105) {
     break;
    } else {
     $j$854 = $269;
    }
   }
   $243 = HEAP32[$155>>2]|0;
   $244 = ((($243)) + 8|0);
   $245 = +HEAPF32[$244>>2];
   $246 = $245 > -200.0;
   do {
    if ($246) {
     $j$9$lcssa = 0.0;
    } else {
     $270 = ((($243)) + 12|0);
     $271 = +HEAPF32[$270>>2];
     $272 = $271 > -200.0;
     if ($272) {
      $j$9$lcssa = 1.0;
     } else {
      $283 = ((($243)) + 16|0);
      $284 = +HEAPF32[$283>>2];
      $285 = $284 > -200.0;
      if ($285) {
       $j$9$lcssa = 2.0;
      } else {
       $286 = ((($243)) + 20|0);
       $287 = +HEAPF32[$286>>2];
       $288 = $287 > -200.0;
       if ($288) {
        $j$9$lcssa = 3.0;
       } else {
        $289 = ((($243)) + 24|0);
        $290 = +HEAPF32[$289>>2];
        $291 = $290 > -200.0;
        if ($291) {
         $j$9$lcssa = 4.0;
        } else {
         $292 = ((($243)) + 28|0);
         $293 = +HEAPF32[$292>>2];
         $294 = $293 > -200.0;
         if ($294) {
          $j$9$lcssa = 5.0;
         } else {
          $295 = ((($243)) + 32|0);
          $296 = +HEAPF32[$295>>2];
          $297 = $296 > -200.0;
          if ($297) {
           $j$9$lcssa = 6.0;
          } else {
           $298 = ((($243)) + 36|0);
           $299 = +HEAPF32[$298>>2];
           $300 = $299 > -200.0;
           if ($300) {
            $j$9$lcssa = 7.0;
           } else {
            $301 = ((($243)) + 40|0);
            $302 = +HEAPF32[$301>>2];
            $303 = $302 > -200.0;
            if ($303) {
             $j$9$lcssa = 8.0;
            } else {
             $304 = ((($243)) + 44|0);
             $305 = +HEAPF32[$304>>2];
             $306 = $305 > -200.0;
             if ($306) {
              $j$9$lcssa = 9.0;
             } else {
              $307 = ((($243)) + 48|0);
              $308 = +HEAPF32[$307>>2];
              $309 = $308 > -200.0;
              if ($309) {
               $j$9$lcssa = 10.0;
              } else {
               $310 = ((($243)) + 52|0);
               $311 = +HEAPF32[$310>>2];
               $312 = $311 > -200.0;
               if ($312) {
                $j$9$lcssa = 11.0;
               } else {
                $313 = ((($243)) + 56|0);
                $314 = +HEAPF32[$313>>2];
                $315 = $314 > -200.0;
                if ($315) {
                 $j$9$lcssa = 12.0;
                } else {
                 $316 = ((($243)) + 60|0);
                 $317 = +HEAPF32[$316>>2];
                 $318 = $317 > -200.0;
                 if ($318) {
                  $j$9$lcssa = 13.0;
                 } else {
                  $319 = ((($243)) + 64|0);
                  $320 = +HEAPF32[$319>>2];
                  $321 = $320 > -200.0;
                  if ($321) {
                   $j$9$lcssa = 14.0;
                   break;
                  }
                  $322 = ((($243)) + 68|0);
                  $323 = +HEAPF32[$322>>2];
                  $324 = $323 > -200.0;
                  if ($324) {
                   $j$9$lcssa = 15.0;
                   break;
                  }
                  $j$9$lcssa = 16.0;
                 }
                }
               }
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    }
   } while(0);
   HEAPF32[$243>>2] = $j$9$lcssa;
   $273 = HEAP32[$155>>2]|0;
   $j$1056 = 55;
   while(1) {
    $274 = (($j$1056) + 2)|0;
    $275 = (($273) + ($274<<2)|0);
    $276 = +HEAPF32[$275>>2];
    $277 = $276 > -200.0;
    if ($277) {
     $j$10$lcssa = $j$1056;
     break;
    }
    $278 = (($j$1056) + -1)|0;
    $279 = ($278|0)>(17);
    if ($279) {
     $j$1056 = $278;
    } else {
     $j$10$lcssa = $278;
     break;
    }
   }
   $280 = (+($j$10$lcssa|0));
   $281 = ((($273)) + 4|0);
   HEAPF32[$281>>2] = $280;
   $282 = (($m$057) + 1)|0;
   $exitcond106 = ($282|0)==(8);
   if ($exitcond106) {
    break;
   } else {
    $m$057 = $282;
   }
  }
  $exitcond107 = ($152|0)==(17);
  if ($exitcond107) {
   break;
  } else {
   $i$161 = $152;
  }
 }
 STACKTOP = sp;return ($2|0);
}
function _vorbis_book_init_decode($c,$s) {
 $c = $c|0;
 $s = $s|0;
 var $$ = 0, $$01$i = 0, $$02 = 0, $$alloca_mul = 0, $$alloca_mul1 = 0, $$lcssa63 = 0, $$lo$1 = 0, $$n$0 = 0, $$pre = 0, $$pre52 = 0, $$pre54 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0;
 var $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $exitcond48 = 0, $exitcond49 = 0, $exitcond50 = 0, $hi$013 = 0, $hi$1$lcssa = 0, $hi$19 = 0;
 var $hi$2 = 0, $hival$0 = 0, $i$041 = 0, $i$137 = 0, $i$234 = 0, $i$331 = 0, $i$426 = 0, $i$520 = 0, $i$617 = 0, $i$711 = 0, $j$015 = 0, $lo$012 = 0, $lo$1 = 0, $lo$1$lcssa = 0, $lo$2 = 0, $n$040 = 0, $n$2$lcssa = 0, $n$225 = 0, $n$3 = 0, $n$4$lcssa55 = 0;
 var $n$419 = 0, $n$5 = 0, $n$5$lcssa = 0, $phitmp = 0, $ret$0$lcssa$i = 0, $ret$02$i = 0, $ret$02$i$lcssa = 0, $storemerge3 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 dest=$c; stop=dest+56|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $0 = ((($s)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(0);
 if ($2) {
  $3 = ((($s)) + 8|0);
  $4 = HEAP32[$3>>2]|0;
  $i$041 = 0;$n$040 = 0;
  while(1) {
   $5 = (($4) + ($i$041)|0);
   $6 = HEAP8[$5>>0]|0;
   $7 = ($6<<24>>24)>(0);
   $8 = $7&1;
   $$n$0 = (($8) + ($n$040))|0;
   $9 = (($i$041) + 1)|0;
   $10 = ($9|0)<($1|0);
   if ($10) {
    $i$041 = $9;$n$040 = $$n$0;
   } else {
    $13 = $$n$0;
    break;
   }
  }
 } else {
  $13 = 0;
 }
 $11 = ((($c)) + 4|0);
 HEAP32[$11>>2] = $1;
 $12 = ((($c)) + 8|0);
 HEAP32[$12>>2] = $13;
 $14 = HEAP32[$s>>2]|0;
 HEAP32[$c>>2] = $14;
 $15 = ($13|0)>(0);
 if (!($15)) {
  $$02 = 0;
  STACKTOP = sp;return ($$02|0);
 }
 $16 = ((($s)) + 8|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = (__make_words($17,$1,$13)|0);
 $19 = $13 << 2;
 $$alloca_mul = $19;
 $20 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $21 = ($18|0)==(0|0);
 if ($21) {
  $203 = ((($c)) + 16|0);
  $204 = HEAP32[$203>>2]|0;
  $205 = ($204|0)==(0|0);
  if (!($205)) {
   _free($204);
  }
  $206 = ((($c)) + 20|0);
  $207 = HEAP32[$206>>2]|0;
  $208 = ($207|0)==(0|0);
  if (!($208)) {
   _free($207);
  }
  $209 = ((($c)) + 24|0);
  $210 = HEAP32[$209>>2]|0;
  $211 = ($210|0)==(0|0);
  if (!($211)) {
   _free($210);
  }
  $212 = ((($c)) + 28|0);
  $213 = HEAP32[$212>>2]|0;
  $214 = ($213|0)==(0|0);
  if (!($214)) {
   _free($213);
  }
  $215 = ((($c)) + 32|0);
  $216 = HEAP32[$215>>2]|0;
  $217 = ($216|0)==(0|0);
  if (!($217)) {
   _free($216);
  }
  dest=$c; stop=dest+56|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
  $$02 = -1;
  STACKTOP = sp;return ($$02|0);
 } else {
  $i$137 = 0;
 }
 while(1) {
  $22 = (($18) + ($i$137<<2)|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = (_llvm_bswap_i32(($23|0))|0);
  $25 = $24 >>> 4;
  $26 = $25 & 252645135;
  $27 = $24 << 4;
  $28 = $27 & -252645136;
  $29 = $26 | $28;
  $30 = $29 >>> 2;
  $31 = $30 & 858993459;
  $32 = $29 << 2;
  $33 = $32 & -858993460;
  $34 = $31 | $33;
  $35 = $34 >>> 1;
  $36 = $35 & 1431655765;
  $37 = $34 << 1;
  $38 = $37 & -1431655766;
  $39 = $36 | $38;
  HEAP32[$22>>2] = $39;
  $40 = (($20) + ($i$137<<2)|0);
  HEAP32[$40>>2] = $22;
  $41 = (($i$137) + 1)|0;
  $exitcond50 = ($41|0)==($13|0);
  if ($exitcond50) {
   break;
  } else {
   $i$137 = $41;
  }
 }
 _qsort($20,$13,4,4453);
 $$alloca_mul1 = $19;
 $42 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul1)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $43 = (_malloc($19)|0);
 $44 = ((($c)) + 20|0);
 HEAP32[$44>>2] = $43;
 $45 = $18;
 $i$234 = 0;
 while(1) {
  $46 = (($20) + ($i$234<<2)|0);
  $47 = HEAP32[$46>>2]|0;
  $48 = (($47) - ($45))|0;
  $49 = $48 >> 2;
  $50 = (($42) + ($49<<2)|0);
  HEAP32[$50>>2] = $i$234;
  $51 = (($i$234) + 1)|0;
  $exitcond49 = ($51|0)==($13|0);
  if ($exitcond49) {
   $i$331 = 0;
   break;
  } else {
   $i$234 = $51;
  }
 }
 while(1) {
  $52 = (($18) + ($i$331<<2)|0);
  $53 = HEAP32[$52>>2]|0;
  $54 = (($42) + ($i$331<<2)|0);
  $55 = HEAP32[$54>>2]|0;
  $56 = (($43) + ($55<<2)|0);
  HEAP32[$56>>2] = $53;
  $57 = (($i$331) + 1)|0;
  $exitcond48 = ($57|0)==($13|0);
  if ($exitcond48) {
   break;
  } else {
   $i$331 = $57;
  }
 }
 _free($18);
 $58 = (__book_unquantize($s,$13,$42)|0);
 $59 = ((($c)) + 16|0);
 HEAP32[$59>>2] = $58;
 $60 = (_malloc($19)|0);
 $61 = ((($c)) + 24|0);
 HEAP32[$61>>2] = $60;
 $62 = HEAP32[$0>>2]|0;
 $63 = ($62|0)>(0);
 if ($63) {
  $$pre = HEAP32[$16>>2]|0;
  $i$426 = 0;$n$225 = 0;
  while(1) {
   $64 = (($$pre) + ($i$426)|0);
   $65 = HEAP8[$64>>0]|0;
   $66 = ($65<<24>>24)>(0);
   if ($66) {
    $67 = (($n$225) + 1)|0;
    $68 = (($42) + ($n$225<<2)|0);
    $69 = HEAP32[$68>>2]|0;
    $70 = (($60) + ($69<<2)|0);
    HEAP32[$70>>2] = $i$426;
    $n$3 = $67;
   } else {
    $n$3 = $n$225;
   }
   $71 = (($i$426) + 1)|0;
   $72 = ($71|0)<($62|0);
   if ($72) {
    $i$426 = $71;$n$225 = $n$3;
   } else {
    $n$2$lcssa = $n$3;
    break;
   }
  }
 } else {
  $n$2$lcssa = 0;
 }
 $73 = (_malloc($n$2$lcssa)|0);
 $74 = ((($c)) + 28|0);
 HEAP32[$74>>2] = $73;
 $75 = ((($c)) + 40|0);
 HEAP32[$75>>2] = 0;
 if ($63) {
  $$pre52 = HEAP32[$16>>2]|0;
  $218 = 0;$77 = $$pre52;$i$520 = 0;$n$419 = 0;
  while(1) {
   $76 = (($77) + ($i$520)|0);
   $78 = HEAP8[$76>>0]|0;
   $79 = ($78<<24>>24)>(0);
   if ($79) {
    $80 = (($n$419) + 1)|0;
    $81 = (($42) + ($n$419<<2)|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = HEAP32[$74>>2]|0;
    $84 = (($83) + ($82)|0);
    HEAP8[$84>>0] = $78;
    $85 = HEAP32[$16>>2]|0;
    $86 = (($85) + ($i$520)|0);
    $87 = HEAP8[$86>>0]|0;
    $88 = $87 << 24 >> 24;
    $89 = HEAP32[$75>>2]|0;
    $90 = ($88|0)>($89|0);
    if ($90) {
     HEAP32[$75>>2] = $88;
     $219 = $88;$220 = $85;$n$5 = $80;
    } else {
     $219 = $89;$220 = $85;$n$5 = $80;
    }
   } else {
    $219 = $218;$220 = $77;$n$5 = $n$419;
   }
   $91 = (($i$520) + 1)|0;
   $92 = HEAP32[$0>>2]|0;
   $93 = ($91|0)<($92|0);
   if ($93) {
    $218 = $219;$77 = $220;$i$520 = $91;$n$419 = $n$5;
   } else {
    $$lcssa63 = $219;$n$5$lcssa = $n$5;
    break;
   }
  }
  $94 = ($n$5$lcssa|0)==(1);
  if ($94) {
   $95 = ($$lcssa63|0)==(1);
   if ($95) {
    $96 = ((($c)) + 36|0);
    HEAP32[$96>>2] = 1;
    $97 = (_calloc(2,4)|0);
    $98 = ((($c)) + 32|0);
    HEAP32[$98>>2] = $97;
    $99 = ((($97)) + 4|0);
    HEAP32[$99>>2] = 1;
    HEAP32[$97>>2] = 1;
    $$02 = 0;
    STACKTOP = sp;return ($$02|0);
   } else {
    $n$4$lcssa55 = 1;
   }
  } else {
   $n$4$lcssa55 = $n$5$lcssa;
  }
 } else {
  $n$4$lcssa55 = 0;
 }
 $100 = HEAP32[$12>>2]|0;
 $101 = ($100|0)==(0);
 if ($101) {
  $ret$0$lcssa$i = -4;
 } else {
  $$01$i = $100;$ret$02$i = 0;
  while(1) {
   $102 = $$01$i >>> 1;
   $103 = (($ret$02$i) + 1)|0;
   $104 = ($102|0)==(0);
   if ($104) {
    $ret$02$i$lcssa = $ret$02$i;
    break;
   } else {
    $$01$i = $102;$ret$02$i = $103;
   }
  }
  $phitmp = (($ret$02$i$lcssa) + -3)|0;
  $ret$0$lcssa$i = $phitmp;
 }
 $105 = ((($c)) + 36|0);
 $106 = ($ret$0$lcssa$i|0)<(5);
 $$ = $106 ? 5 : $ret$0$lcssa$i;
 $107 = ($$|0)>(8);
 $storemerge3 = $107 ? 8 : $$;
 HEAP32[$105>>2] = $storemerge3;
 $108 = 1 << $storemerge3;
 $109 = (_calloc($108,4)|0);
 $110 = ((($c)) + 32|0);
 HEAP32[$110>>2] = $109;
 $111 = ($n$4$lcssa55|0)>(0);
 if ($111) {
  $117 = $storemerge3;$i$617 = 0;
  while(1) {
   $112 = HEAP32[$74>>2]|0;
   $113 = (($112) + ($i$617)|0);
   $114 = HEAP8[$113>>0]|0;
   $115 = $114 << 24 >> 24;
   $116 = ($117|0)<($115|0);
   if ($116) {
    $221 = $117;
   } else {
    $118 = HEAP32[$44>>2]|0;
    $119 = (($118) + ($i$617<<2)|0);
    $120 = HEAP32[$119>>2]|0;
    $121 = (_llvm_bswap_i32(($120|0))|0);
    $122 = $121 >>> 4;
    $123 = $122 & 252645135;
    $124 = $121 << 4;
    $125 = $124 & -252645136;
    $126 = $123 | $125;
    $127 = $126 >>> 2;
    $128 = $127 & 858993459;
    $129 = $126 << 2;
    $130 = $129 & -858993460;
    $131 = $128 | $130;
    $132 = $131 >>> 1;
    $133 = $132 & 1431655765;
    $134 = $131 << 1;
    $135 = $134 & -1431655766;
    $136 = $133 | $135;
    $137 = (($117) - ($115))|0;
    $138 = ($137|0)==(31);
    if ($138) {
     $221 = $117;
    } else {
     $139 = (($i$617) + 1)|0;
     $141 = $115;$j$015 = 0;
     while(1) {
      $140 = $j$015 << $141;
      $142 = $136 | $140;
      $143 = (($109) + ($142<<2)|0);
      HEAP32[$143>>2] = $139;
      $144 = (($j$015) + 1)|0;
      $145 = HEAP32[$105>>2]|0;
      $146 = HEAP8[$113>>0]|0;
      $147 = $146 << 24 >> 24;
      $148 = (($145) - ($147))|0;
      $149 = 1 << $148;
      $150 = ($144|0)<($149|0);
      if ($150) {
       $141 = $147;$j$015 = $144;
      } else {
       $221 = $145;
       break;
      }
     }
    }
   }
   $151 = (($i$617) + 1)|0;
   $exitcond = ($151|0)==($n$4$lcssa55|0);
   if ($exitcond) {
    $153 = $221;
    break;
   } else {
    $117 = $221;$i$617 = $151;
   }
  }
 } else {
  $153 = $storemerge3;
 }
 $152 = (31 - ($153))|0;
 $154 = -2 << $152;
 $155 = ($storemerge3|0)==(31);
 if ($155) {
  $$02 = 0;
  STACKTOP = sp;return ($$02|0);
 }
 $156 = HEAP32[$110>>2]|0;
 $158 = $153;$hi$013 = 0;$i$711 = 0;$lo$012 = 0;
 while(1) {
  $157 = (32 - ($158))|0;
  $159 = $i$711 << $157;
  $160 = (_llvm_bswap_i32(($159|0))|0);
  $161 = $160 >>> 4;
  $162 = $161 & 252645135;
  $163 = $160 << 4;
  $164 = $163 & -252645136;
  $165 = $162 | $164;
  $166 = $165 >>> 2;
  $167 = $166 & 858993459;
  $168 = $165 << 2;
  $169 = $168 & -858993460;
  $170 = $167 | $169;
  $171 = $170 >>> 1;
  $172 = $171 & 1431655765;
  $173 = $170 << 1;
  $174 = $173 & -1431655766;
  $175 = $172 | $174;
  $176 = (($156) + ($175<<2)|0);
  $177 = HEAP32[$176>>2]|0;
  $178 = ($177|0)==(0);
  if ($178) {
   $lo$1 = $lo$012;
   while(1) {
    $179 = (($lo$1) + 1)|0;
    $180 = ($179|0)<($n$4$lcssa55|0);
    if (!($180)) {
     $lo$1$lcssa = $lo$1;
     break;
    }
    $181 = HEAP32[$44>>2]|0;
    $182 = (($181) + ($179<<2)|0);
    $183 = HEAP32[$182>>2]|0;
    $184 = ($183>>>0)>($159>>>0);
    if ($184) {
     $lo$1$lcssa = $lo$1;
     break;
    } else {
     $lo$1 = $179;
    }
   }
   $185 = ($n$4$lcssa55|0)>($hi$013|0);
   L81: do {
    if ($185) {
     $186 = HEAP32[$44>>2]|0;
     $hi$19 = $hi$013;
     while(1) {
      $187 = (($186) + ($hi$19<<2)|0);
      $188 = HEAP32[$187>>2]|0;
      $189 = $188 & $154;
      $190 = ($159>>>0)<($189>>>0);
      if ($190) {
       $hi$1$lcssa = $hi$19;
       break L81;
      }
      $191 = (($hi$19) + 1)|0;
      $192 = ($n$4$lcssa55|0)>($191|0);
      if ($192) {
       $hi$19 = $191;
      } else {
       $hi$1$lcssa = $191;
       break;
      }
     }
    } else {
     $hi$1$lcssa = $hi$013;
    }
   } while(0);
   $193 = (($n$4$lcssa55) - ($hi$1$lcssa))|0;
   $194 = ($lo$1$lcssa>>>0)>(32767);
   $$lo$1 = $194 ? 32767 : $lo$1$lcssa;
   $195 = ($193>>>0)>(32767);
   $hival$0 = $195 ? 32767 : $193;
   $196 = $$lo$1 << 15;
   $197 = $196 | $hival$0;
   $198 = $197 | -2147483648;
   $199 = HEAP32[$110>>2]|0;
   $200 = (($199) + ($175<<2)|0);
   HEAP32[$200>>2] = $198;
   $hi$2 = $hi$1$lcssa;$lo$2 = $lo$1$lcssa;
  } else {
   $hi$2 = $hi$013;$lo$2 = $lo$012;
  }
  $201 = (($i$711) + 1)|0;
  $202 = ($201|0)<($108|0);
  if (!($202)) {
   $$02 = 0;
   break;
  }
  $$pre54 = HEAP32[$105>>2]|0;
  $158 = $$pre54;$hi$013 = $hi$2;$i$711 = $201;$lo$012 = $lo$2;
 }
 STACKTOP = sp;return ($$02|0);
}
function _sort32a($a,$b) {
 $a = $a|0;
 $b = $b|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$a>>2]|0;
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$b>>2]|0;
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1>>>0)>($3>>>0);
 $5 = $4&1;
 $6 = ($1>>>0)<($3>>>0);
 $7 = $6&1;
 $8 = (($5) - ($7))|0;
 return ($8|0);
}
function __book_unquantize($b,$n,$sparsemap) {
 $b = $b|0;
 $n = $n|0;
 $sparsemap = $sparsemap|0;
 var $$$i = 0.0, $$$i6 = 0.0, $$0 = 0, $$lcssa = 0, $$lcssa68 = 0, $$off = 0, $$pre46 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0.0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0.0, $110 = 0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0, $119 = 0, $12 = 0.0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0.0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0.0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0, $16 = 0.0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0.0, $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0.0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0.0;
 var $182 = 0.0, $183 = 0.0, $184 = 0.0, $185 = 0.0, $186 = 0.0, $187 = 0.0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0;
 var $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0, $48 = 0, $49 = 0, $5 = 0.0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0.0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0.0;
 var $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0.0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0.0, $97 = 0.0;
 var $98 = 0.0, $99 = 0.0, $acc$02$us$i = 0, $acc1$03$us$i = 0, $count$014 = 0, $count$1 = 0, $count$224 = 0, $count$3 = 0, $exitcond = 0, $fabsf = 0.0, $fabsf$us = 0.0, $fabsf$us$us = 0.0, $fabsf$us37 = 0.0, $fabsf5 = 0.0, $fabsf5$us = 0.0, $fabsf5$us33 = 0.0, $i$04$us$i = 0, $indexdiv$013 = 0, $indexdiv$013$us = 0, $indexdiv$013$us30 = 0;
 var $j$016 = 0, $j$126 = 0, $k$011 = 0, $k$011$us = 0, $k$011$us32 = 0, $k$120 = 0, $k$120$us = 0, $k$120$us$us = 0, $k$120$us36 = 0, $last$0$$us = 0.0, $last$012 = 0.0, $last$012$us = 0.0, $last1$021 = 0.0, $last1$021$us = 0.0, $or$cond$us$i = 0, $switch = 0, $vals$0$us$i = 0, $vals$0$us$i$lcssa = 0, $vals$1$us$i = 0, $vals$1$v$us$i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($b)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $$off = (($1) + -1)|0;
 $switch = ($$off>>>0)<(2);
 if (!($switch)) {
  $$0 = 0;
  return ($$0|0);
 }
 $2 = ((($b)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 2097151;
 $5 = (+($4|0));
 $6 = $3 >>> 21;
 $7 = $6 & 1023;
 $8 = ($3|0)<(0);
 $9 = -$5;
 $$$i = $8 ? $9 : $5;
 $10 = (($7) + -788)|0;
 $11 = (+_ldexp($$$i,$10));
 $12 = $11;
 $13 = ((($b)) + 20|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = $14 & 2097151;
 $16 = (+($15|0));
 $17 = $14 >>> 21;
 $18 = $17 & 1023;
 $19 = ($14|0)<(0);
 $20 = -$16;
 $$$i6 = $19 ? $20 : $16;
 $21 = (($18) + -788)|0;
 $22 = (+_ldexp($$$i6,$21));
 $23 = $22;
 $24 = HEAP32[$b>>2]|0;
 $25 = Math_imul($24, $n)|0;
 $26 = (_calloc($25,4)|0);
 switch ($1|0) {
 case 1:  {
  $37 = ((($b)) + 4|0);
  $38 = HEAP32[$37>>2]|0;
  $39 = ($24|0)>(0);
  if (!($39)) {
   while(1) {
   }
  }
  $40 = (+($38|0));
  $41 = $40;
  $42 = (+($24|0));
  $43 = 1.0 / $42;
  $44 = $43;
  $45 = (+Math_pow((+$41),(+$44)));
  $46 = (+Math_floor((+$45)));
  $47 = (~~(($46)));
  $vals$0$us$i = $47;
  while(1) {
   $51 = (($vals$0$us$i) + 1)|0;
   $acc$02$us$i = 1;$acc1$03$us$i = 1;$i$04$us$i = 0;
   while(1) {
    $49 = Math_imul($acc$02$us$i, $vals$0$us$i)|0;
    $50 = Math_imul($acc1$03$us$i, $51)|0;
    $52 = (($i$04$us$i) + 1)|0;
    $exitcond = ($52|0)==($24|0);
    if ($exitcond) {
     $$lcssa = $49;$$lcssa68 = $50;
     break;
    } else {
     $acc$02$us$i = $49;$acc1$03$us$i = $50;$i$04$us$i = $52;
    }
   }
   $53 = ($$lcssa|0)<=($38|0);
   $54 = ($$lcssa68|0)>($38|0);
   $or$cond$us$i = $54 & $53;
   if ($or$cond$us$i) {
    $vals$0$us$i$lcssa = $vals$0$us$i;
    break;
   }
   $48 = ($$lcssa|0)>($38|0);
   $vals$1$v$us$i = $48 ? -1 : 1;
   $vals$1$us$i = (($vals$1$v$us$i) + ($vals$0$us$i))|0;
   $vals$0$us$i = $vals$1$us$i;
  }
  $55 = ($38|0)>(0);
  if (!($55)) {
   $$0 = $26;
   return ($$0|0);
  }
  $56 = ($sparsemap|0)==(0|0);
  $57 = ((($b)) + 8|0);
  $58 = ((($b)) + 32|0);
  $59 = $23;
  $60 = $12;
  $61 = ((($b)) + 28|0);
  $count$014 = 0;$j$016 = 0;
  while(1) {
   if ($56) {
    $88 = HEAP32[$58>>2]|0;
    $89 = HEAP32[$61>>2]|0;
    $90 = ($89|0)==(0);
    $91 = Math_imul($24, $count$014)|0;
    if ($90) {
     $indexdiv$013$us30 = 1;$k$011$us32 = 0;
     while(1) {
      $92 = (($j$016|0) / ($indexdiv$013$us30|0))&-1;
      $93 = (($92|0) % ($vals$0$us$i$lcssa|0))&-1;
      $94 = (($88) + ($93<<2)|0);
      $95 = HEAP32[$94>>2]|0;
      $96 = (+($95|0));
      $fabsf5$us33 = (+Math_abs((+$96)));
      $97 = $fabsf5$us33;
      $98 = $59 * $97;
      $99 = $60 + $98;
      $100 = $99 + 0.0;
      $101 = $100;
      $102 = (($91) + ($k$011$us32))|0;
      $103 = (($26) + ($102<<2)|0);
      HEAPF32[$103>>2] = $101;
      $104 = Math_imul($indexdiv$013$us30, $vals$0$us$i$lcssa)|0;
      $105 = (($k$011$us32) + 1)|0;
      $106 = ($105|0)<($24|0);
      if ($106) {
       $indexdiv$013$us30 = $104;$k$011$us32 = $105;
      } else {
       label = 21;
       break;
      }
     }
    } else {
     $indexdiv$013 = 1;$k$011 = 0;$last$012 = 0.0;
     while(1) {
      $107 = (($j$016|0) / ($indexdiv$013|0))&-1;
      $108 = (($107|0) % ($vals$0$us$i$lcssa|0))&-1;
      $109 = (($88) + ($108<<2)|0);
      $110 = HEAP32[$109>>2]|0;
      $111 = (+($110|0));
      $fabsf5 = (+Math_abs((+$111)));
      $112 = $fabsf5;
      $113 = $59 * $112;
      $114 = $60 + $113;
      $115 = $last$012;
      $116 = $115 + $114;
      $117 = $116;
      $118 = (($91) + ($k$011))|0;
      $119 = (($26) + ($118<<2)|0);
      HEAPF32[$119>>2] = $117;
      $120 = Math_imul($indexdiv$013, $vals$0$us$i$lcssa)|0;
      $121 = (($k$011) + 1)|0;
      $122 = ($121|0)<($24|0);
      if ($122) {
       $indexdiv$013 = $120;$k$011 = $121;$last$012 = $117;
      } else {
       label = 21;
       break;
      }
     }
    }
   } else {
    $62 = HEAP32[$57>>2]|0;
    $63 = (($62) + ($j$016)|0);
    $64 = HEAP8[$63>>0]|0;
    $65 = ($64<<24>>24)==(0);
    if ($65) {
     $count$1 = $count$014;
    } else {
     $66 = HEAP32[$58>>2]|0;
     $67 = HEAP32[$61>>2]|0;
     $68 = ($67|0)==(0);
     $69 = (($sparsemap) + ($count$014<<2)|0);
     $70 = HEAP32[$69>>2]|0;
     $71 = Math_imul($70, $24)|0;
     $indexdiv$013$us = 1;$k$011$us = 0;$last$012$us = 0.0;
     while(1) {
      $72 = (($j$016|0) / ($indexdiv$013$us|0))&-1;
      $73 = (($72|0) % ($vals$0$us$i$lcssa|0))&-1;
      $74 = (($66) + ($73<<2)|0);
      $75 = HEAP32[$74>>2]|0;
      $76 = (+($75|0));
      $fabsf5$us = (+Math_abs((+$76)));
      $77 = $fabsf5$us;
      $78 = $59 * $77;
      $79 = $60 + $78;
      $80 = $last$012$us;
      $81 = $80 + $79;
      $82 = $81;
      $last$0$$us = $68 ? $last$012$us : $82;
      $83 = (($71) + ($k$011$us))|0;
      $84 = (($26) + ($83<<2)|0);
      HEAPF32[$84>>2] = $82;
      $85 = Math_imul($indexdiv$013$us, $vals$0$us$i$lcssa)|0;
      $86 = (($k$011$us) + 1)|0;
      $87 = ($86|0)<($24|0);
      if ($87) {
       $indexdiv$013$us = $85;$k$011$us = $86;$last$012$us = $last$0$$us;
      } else {
       label = 21;
       break;
      }
     }
    }
   }
   if ((label|0) == 21) {
    label = 0;
    $123 = (($count$014) + 1)|0;
    $count$1 = $123;
   }
   $124 = (($j$016) + 1)|0;
   $125 = ($124|0)<($38|0);
   if ($125) {
    $count$014 = $count$1;$j$016 = $124;
   } else {
    $$0 = $26;
    break;
   }
  }
  return ($$0|0);
  break;
 }
 case 2:  {
  $27 = ((($b)) + 4|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = ($28|0)>(0);
  if (!($29)) {
   $$0 = $26;
   return ($$0|0);
  }
  $30 = ($sparsemap|0)!=(0|0);
  $31 = ((($b)) + 8|0);
  $32 = ((($b)) + 32|0);
  $33 = $23;
  $34 = $12;
  $35 = ((($b)) + 28|0);
  $36 = ($24|0)>(0);
  $count$224 = 0;$j$126 = 0;
  while(1) {
   if ($30) {
    $126 = HEAP32[$31>>2]|0;
    $127 = (($126) + ($j$126)|0);
    $128 = HEAP8[$127>>0]|0;
    $129 = ($128<<24>>24)==(0);
    if ($129) {
     $count$3 = $count$224;
    } else {
     label = 25;
    }
   } else {
    label = 25;
   }
   if ((label|0) == 25) {
    label = 0;
    if ($36) {
     $130 = HEAP32[$32>>2]|0;
     $131 = HEAP32[$35>>2]|0;
     $132 = ($131|0)==(0);
     if ($30) {
      $133 = (($sparsemap) + ($count$224<<2)|0);
      $$pre46 = HEAP32[$133>>2]|0;
      $134 = Math_imul($24, $j$126)|0;
      $135 = Math_imul($$pre46, $24)|0;
      if ($132) {
       $k$120$us$us = 0;
       while(1) {
        $136 = (($134) + ($k$120$us$us))|0;
        $137 = (($130) + ($136<<2)|0);
        $138 = HEAP32[$137>>2]|0;
        $139 = (+($138|0));
        $fabsf$us$us = (+Math_abs((+$139)));
        $140 = $fabsf$us$us;
        $141 = $33 * $140;
        $142 = $34 + $141;
        $143 = $142 + 0.0;
        $144 = $143;
        $145 = (($135) + ($k$120$us$us))|0;
        $146 = (($26) + ($145<<2)|0);
        HEAPF32[$146>>2] = $144;
        $147 = (($k$120$us$us) + 1)|0;
        $148 = ($147|0)<($24|0);
        if ($148) {
         $k$120$us$us = $147;
        } else {
         break;
        }
       }
      } else {
       $k$120$us = 0;$last1$021$us = 0.0;
       while(1) {
        $149 = (($134) + ($k$120$us))|0;
        $150 = (($130) + ($149<<2)|0);
        $151 = HEAP32[$150>>2]|0;
        $152 = (+($151|0));
        $fabsf$us = (+Math_abs((+$152)));
        $153 = $fabsf$us;
        $154 = $33 * $153;
        $155 = $34 + $154;
        $156 = $last1$021$us;
        $157 = $156 + $155;
        $158 = $157;
        $159 = (($135) + ($k$120$us))|0;
        $160 = (($26) + ($159<<2)|0);
        HEAPF32[$160>>2] = $158;
        $161 = (($k$120$us) + 1)|0;
        $162 = ($161|0)<($24|0);
        if ($162) {
         $k$120$us = $161;$last1$021$us = $158;
        } else {
         break;
        }
       }
      }
     } else {
      $163 = Math_imul($24, $j$126)|0;
      $164 = Math_imul($24, $count$224)|0;
      if ($132) {
       $k$120$us36 = 0;
       while(1) {
        $165 = (($163) + ($k$120$us36))|0;
        $166 = (($130) + ($165<<2)|0);
        $167 = HEAP32[$166>>2]|0;
        $168 = (+($167|0));
        $fabsf$us37 = (+Math_abs((+$168)));
        $169 = $fabsf$us37;
        $170 = $33 * $169;
        $171 = $34 + $170;
        $172 = $171 + 0.0;
        $173 = $172;
        $174 = (($164) + ($k$120$us36))|0;
        $175 = (($26) + ($174<<2)|0);
        HEAPF32[$175>>2] = $173;
        $176 = (($k$120$us36) + 1)|0;
        $177 = ($176|0)<($24|0);
        if ($177) {
         $k$120$us36 = $176;
        } else {
         break;
        }
       }
      } else {
       $k$120 = 0;$last1$021 = 0.0;
       while(1) {
        $178 = (($163) + ($k$120))|0;
        $179 = (($130) + ($178<<2)|0);
        $180 = HEAP32[$179>>2]|0;
        $181 = (+($180|0));
        $fabsf = (+Math_abs((+$181)));
        $182 = $fabsf;
        $183 = $33 * $182;
        $184 = $34 + $183;
        $185 = $last1$021;
        $186 = $185 + $184;
        $187 = $186;
        $188 = (($164) + ($k$120))|0;
        $189 = (($26) + ($188<<2)|0);
        HEAPF32[$189>>2] = $187;
        $190 = (($k$120) + 1)|0;
        $191 = ($190|0)<($24|0);
        if ($191) {
         $k$120 = $190;$last1$021 = $187;
        } else {
         break;
        }
       }
      }
     }
    }
    $192 = (($count$224) + 1)|0;
    $count$3 = $192;
   }
   $193 = (($j$126) + 1)|0;
   $194 = ($193|0)<($28|0);
   if ($194) {
    $count$224 = $count$3;$j$126 = $193;
   } else {
    $$0 = $26;
    break;
   }
  }
  return ($$0|0);
  break;
 }
 default: {
  $$0 = $26;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function _vorbis_staticbook_destroy($b) {
 $b = $b|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($b)) + 36|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  return;
 }
 $3 = ((($b)) + 32|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if (!($5)) {
  _free($4);
 }
 $6 = ((($b)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 if (!($8)) {
  _free($7);
 }
 _free($b);
 return;
}
function _vorbis_dsp_clear($v) {
 $v = $v|0;
 var $$pre = 0, $$pre14 = 0, $$pre16 = 0, $$pre18 = 0, $$pre19 = 0, $$pre21 = 0, $$pre22 = 0, $$pre23 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $i$37 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = ($v|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($v)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)!=(0|0);
 if ($3) {
  $4 = ((($2)) + 28|0);
  $5 = HEAP32[$4>>2]|0;
  $30 = $5;
 } else {
  $30 = 0;
 }
 $6 = ((($v)) + 104|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)!=(0|0);
 if ($8) {
  $9 = HEAP32[$7>>2]|0;
  $10 = ($9|0)==(0|0);
  if (!($10)) {
   __ve_envelope_clear($9);
   $11 = HEAP32[$7>>2]|0;
   _free($11);
  }
  $12 = ((($7)) + 12|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ($13|0)==(0|0);
  if (!($14)) {
   $15 = HEAP32[$13>>2]|0;
   _mdct_clear($15);
   $16 = HEAP32[$12>>2]|0;
   $17 = HEAP32[$16>>2]|0;
   _free($17);
   $18 = HEAP32[$12>>2]|0;
   _free($18);
  }
  $19 = ((($7)) + 16|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = ($20|0)==(0|0);
  if (!($21)) {
   $22 = HEAP32[$20>>2]|0;
   _mdct_clear($22);
   $23 = HEAP32[$19>>2]|0;
   $24 = HEAP32[$23>>2]|0;
   _free($24);
   $25 = HEAP32[$19>>2]|0;
   _free($25);
  }
  $26 = ((($7)) + 48|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = ($27|0)==(0|0);
  if (!($28)) {
   $29 = ($30|0)==(0|0);
   if ($29) {
    $55 = $27;
   } else {
    $31 = ((($30)) + 16|0);
    $32 = HEAP32[$31>>2]|0;
    $33 = ($32|0)>(0);
    if ($33) {
     $34 = ((($30)) + 800|0);
     $35 = HEAP32[$34>>2]|0;
     $36 = (101372 + ($35<<2)|0);
     $37 = HEAP32[$36>>2]|0;
     $38 = ((($37)) + 16|0);
     $39 = HEAP32[$38>>2]|0;
     $40 = HEAP32[$27>>2]|0;
     FUNCTION_TABLE_vi[$39 & 8191]($40);
     $41 = HEAP32[$31>>2]|0;
     $42 = ($41|0)>(1);
     if ($42) {
      $44 = 1;
      while(1) {
       $$pre = HEAP32[$26>>2]|0;
       $43 = (($34) + ($44<<2)|0);
       $45 = HEAP32[$43>>2]|0;
       $46 = (101372 + ($45<<2)|0);
       $47 = HEAP32[$46>>2]|0;
       $48 = ((($47)) + 16|0);
       $49 = HEAP32[$48>>2]|0;
       $50 = (($$pre) + ($44<<2)|0);
       $51 = HEAP32[$50>>2]|0;
       FUNCTION_TABLE_vi[$49 & 8191]($51);
       $52 = (($44) + 1)|0;
       $53 = HEAP32[$31>>2]|0;
       $54 = ($52|0)<($53|0);
       if ($54) {
        $44 = $52;
       } else {
        break;
       }
      }
     }
     $$pre14 = HEAP32[$26>>2]|0;
     $55 = $$pre14;
    } else {
     $55 = $27;
    }
   }
   _free($55);
  }
  $56 = ((($7)) + 52|0);
  $57 = HEAP32[$56>>2]|0;
  $58 = ($57|0)==(0|0);
  if (!($58)) {
   $59 = ($30|0)==(0|0);
   if ($59) {
    $84 = $57;
   } else {
    $60 = ((($30)) + 20|0);
    $61 = HEAP32[$60>>2]|0;
    $62 = ($61|0)>(0);
    if ($62) {
     $63 = ((($30)) + 1312|0);
     $64 = HEAP32[$63>>2]|0;
     $65 = (101264 + ($64<<2)|0);
     $66 = HEAP32[$65>>2]|0;
     $67 = ((($66)) + 16|0);
     $68 = HEAP32[$67>>2]|0;
     $69 = HEAP32[$57>>2]|0;
     FUNCTION_TABLE_vi[$68 & 8191]($69);
     $70 = HEAP32[$60>>2]|0;
     $71 = ($70|0)>(1);
     if ($71) {
      $73 = 1;
      while(1) {
       $$pre16 = HEAP32[$56>>2]|0;
       $72 = (($63) + ($73<<2)|0);
       $74 = HEAP32[$72>>2]|0;
       $75 = (101264 + ($74<<2)|0);
       $76 = HEAP32[$75>>2]|0;
       $77 = ((($76)) + 16|0);
       $78 = HEAP32[$77>>2]|0;
       $79 = (($$pre16) + ($73<<2)|0);
       $80 = HEAP32[$79>>2]|0;
       FUNCTION_TABLE_vi[$78 & 8191]($80);
       $81 = (($73) + 1)|0;
       $82 = HEAP32[$60>>2]|0;
       $83 = ($81|0)<($82|0);
       if ($83) {
        $73 = $81;
       } else {
        break;
       }
      }
     }
     $$pre18 = HEAP32[$56>>2]|0;
     $84 = $$pre18;
    } else {
     $84 = $57;
    }
   }
   _free($84);
  }
  $85 = ((($7)) + 56|0);
  $86 = HEAP32[$85>>2]|0;
  $87 = ($86|0)==(0|0);
  if (!($87)) {
   $88 = ($30|0)==(0|0);
   if ($88) {
    $99 = $86;
   } else {
    $89 = ((($30)) + 28|0);
    $90 = HEAP32[$89>>2]|0;
    $91 = ($90|0)>(0);
    if ($91) {
     __vp_psy_clear($86);
     $92 = HEAP32[$89>>2]|0;
     $93 = ($92|0)>(1);
     if ($93) {
      $95 = 1;
      while(1) {
       $$pre19 = HEAP32[$85>>2]|0;
       $94 = (($$pre19) + (($95*52)|0)|0);
       __vp_psy_clear($94);
       $96 = (($95) + 1)|0;
       $97 = HEAP32[$89>>2]|0;
       $98 = ($96|0)<($97|0);
       if ($98) {
        $95 = $96;
       } else {
        break;
       }
      }
     }
     $$pre21 = HEAP32[$85>>2]|0;
     $99 = $$pre21;
    } else {
     $99 = $86;
    }
   }
   _free($99);
  }
  $100 = ((($7)) + 60|0);
  $101 = HEAP32[$100>>2]|0;
  $102 = ($101|0)==(0|0);
  if (!($102)) {
   __vp_global_free($101);
  }
  $103 = ((($7)) + 80|0);
  _vorbis_bitrate_clear($103);
  $104 = ((($7)) + 20|0);
  _drft_clear($104);
  $105 = ((($7)) + 32|0);
  _drft_clear($105);
 }
 $106 = ((($v)) + 8|0);
 $107 = HEAP32[$106>>2]|0;
 $108 = ($107|0)==(0|0);
 if (!($108)) {
  if ($3) {
   $109 = ((($2)) + 4|0);
   $110 = HEAP32[$109>>2]|0;
   $111 = ($110|0)>(0);
   if ($111) {
    $113 = $107;$132 = $110;$i$37 = 0;
    while(1) {
     $112 = (($113) + ($i$37<<2)|0);
     $114 = HEAP32[$112>>2]|0;
     $115 = ($114|0)==(0|0);
     if ($115) {
      $118 = $132;
     } else {
      _free($114);
      $$pre23 = HEAP32[$109>>2]|0;
      $118 = $$pre23;
     }
     $116 = (($i$37) + 1)|0;
     $117 = ($116|0)<($118|0);
     if (!($117)) {
      break;
     }
     $$pre22 = HEAP32[$106>>2]|0;
     $113 = $$pre22;$132 = $118;$i$37 = $116;
    }
    $$pre25 = HEAP32[$106>>2]|0;
    $119 = $$pre25;
   } else {
    $119 = $107;
   }
  } else {
   $119 = $107;
  }
  _free($119);
  $120 = ((($v)) + 12|0);
  $121 = HEAP32[$120>>2]|0;
  $122 = ($121|0)==(0|0);
  if (!($122)) {
   _free($121);
  }
 }
 if ($8) {
  $123 = ((($7)) + 64|0);
  $124 = HEAP32[$123>>2]|0;
  $125 = ($124|0)==(0|0);
  if (!($125)) {
   _free($124);
  }
  $126 = ((($7)) + 68|0);
  $127 = HEAP32[$126>>2]|0;
  $128 = ($127|0)==(0|0);
  if (!($128)) {
   _free($127);
  }
  $129 = ((($7)) + 72|0);
  $130 = HEAP32[$129>>2]|0;
  $131 = ($130|0)==(0|0);
  if (!($131)) {
   _free($130);
  }
  _free($7);
 }
 dest=$v; stop=dest+112|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 return;
}
function __ve_envelope_clear($e) {
 $e = $e|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($e)) + 16|0);
 _mdct_clear($0);
 $1 = ((($e)) + 48|0);
 $2 = HEAP32[$1>>2]|0;
 _free($2);
 $3 = ((($e)) + 64|0);
 $4 = HEAP32[$3>>2]|0;
 _free($4);
 $5 = ((($e)) + 80|0);
 $6 = HEAP32[$5>>2]|0;
 _free($6);
 $7 = ((($e)) + 96|0);
 $8 = HEAP32[$7>>2]|0;
 _free($8);
 $9 = ((($e)) + 112|0);
 $10 = HEAP32[$9>>2]|0;
 _free($10);
 $11 = ((($e)) + 128|0);
 $12 = HEAP32[$11>>2]|0;
 _free($12);
 $13 = ((($e)) + 144|0);
 $14 = HEAP32[$13>>2]|0;
 _free($14);
 $15 = ((($e)) + 36|0);
 $16 = HEAP32[$15>>2]|0;
 _free($16);
 $17 = ((($e)) + 152|0);
 $18 = HEAP32[$17>>2]|0;
 _free($18);
 $19 = ((($e)) + 160|0);
 $20 = HEAP32[$19>>2]|0;
 _free($20);
 _memset(($e|0),0,180)|0;
 return;
}
function _mdct_clear($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($l|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($l)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
  _free($2);
 }
 $4 = ((($l)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if (!($6)) {
  _free($5);
 }
 ;HEAP32[$l>>2]=0|0;HEAP32[$l+4>>2]=0|0;HEAP32[$l+8>>2]=0|0;HEAP32[$l+12>>2]=0|0;HEAP32[$l+16>>2]=0|0;
 return;
}
function __vp_psy_clear($p) {
 $p = $p|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond7 = 0, $i$05 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = ($p|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($p)) + 16|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
  _free($2);
 }
 $4 = ((($p)) + 20|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if (!($6)) {
  _free($5);
 }
 $7 = ((($p)) + 24|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 if (!($9)) {
  _free($8);
 }
 $10 = ((($p)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0|0);
 if (!($12)) {
  $14 = $11;$i$05 = 0;
  while(1) {
   $13 = (($14) + ($i$05<<2)|0);
   $15 = HEAP32[$13>>2]|0;
   $16 = HEAP32[$15>>2]|0;
   _free($16);
   $17 = HEAP32[$10>>2]|0;
   $18 = (($17) + ($i$05<<2)|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ((($19)) + 4|0);
   $21 = HEAP32[$20>>2]|0;
   _free($21);
   $22 = HEAP32[$10>>2]|0;
   $23 = (($22) + ($i$05<<2)|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = ((($24)) + 8|0);
   $26 = HEAP32[$25>>2]|0;
   _free($26);
   $27 = HEAP32[$10>>2]|0;
   $28 = (($27) + ($i$05<<2)|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = ((($29)) + 12|0);
   $31 = HEAP32[$30>>2]|0;
   _free($31);
   $32 = HEAP32[$10>>2]|0;
   $33 = (($32) + ($i$05<<2)|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = ((($34)) + 16|0);
   $36 = HEAP32[$35>>2]|0;
   _free($36);
   $37 = HEAP32[$10>>2]|0;
   $38 = (($37) + ($i$05<<2)|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = ((($39)) + 20|0);
   $41 = HEAP32[$40>>2]|0;
   _free($41);
   $42 = HEAP32[$10>>2]|0;
   $43 = (($42) + ($i$05<<2)|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = ((($44)) + 24|0);
   $46 = HEAP32[$45>>2]|0;
   _free($46);
   $47 = HEAP32[$10>>2]|0;
   $48 = (($47) + ($i$05<<2)|0);
   $49 = HEAP32[$48>>2]|0;
   $50 = ((($49)) + 28|0);
   $51 = HEAP32[$50>>2]|0;
   _free($51);
   $52 = HEAP32[$10>>2]|0;
   $53 = (($52) + ($i$05<<2)|0);
   $54 = HEAP32[$53>>2]|0;
   _free($54);
   $55 = (($i$05) + 1)|0;
   $exitcond7 = ($55|0)==(17);
   if ($exitcond7) {
    break;
   }
   $$pre = HEAP32[$10>>2]|0;
   $14 = $$pre;$i$05 = $55;
  }
  $56 = HEAP32[$10>>2]|0;
  _free($56);
 }
 $57 = ((($p)) + 12|0);
 $58 = HEAP32[$57>>2]|0;
 $59 = ($58|0)==(0|0);
 if (!($59)) {
  $60 = HEAP32[$58>>2]|0;
  _free($60);
  $61 = HEAP32[$57>>2]|0;
  $62 = ((($61)) + 4|0);
  $63 = HEAP32[$62>>2]|0;
  _free($63);
  $64 = HEAP32[$57>>2]|0;
  $65 = ((($64)) + 8|0);
  $66 = HEAP32[$65>>2]|0;
  _free($66);
  $67 = HEAP32[$57>>2]|0;
  _free($67);
 }
 dest=$p; stop=dest+52|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 return;
}
function __vp_global_free($look) {
 $look = $look|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($look|0)==(0|0);
 if ($0) {
  return;
 }
 _free($look);
 return;
}
function _drft_clear($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($l|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($l)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
  _free($2);
 }
 $4 = ((($l)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if (!($6)) {
  _free($5);
 }
 ;HEAP32[$l>>2]=0|0;HEAP32[$l+4>>2]=0|0;HEAP32[$l+8>>2]=0|0;
 return;
}
function _vorbis_synthesis_restart($v) {
 $v = $v|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($v)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($v)) + 104|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 $5 = ($1|0)==(0|0);
 $or$cond = $5 | $4;
 if ($or$cond) {
  $$0 = -1;
  return ($$0|0);
 }
 $6 = ((($1)) + 28|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 if ($8) {
  $$0 = -1;
  return ($$0|0);
 }
 $9 = ((($7)) + 3656|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($7)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10) + 1)|0;
 $14 = $12 >> $13;
 $15 = ((($v)) + 48|0);
 HEAP32[$15>>2] = $14;
 $16 = $14 >> $10;
 $17 = ((($v)) + 20|0);
 HEAP32[$17>>2] = $16;
 $18 = ((($v)) + 24|0);
 HEAP32[$18>>2] = -1;
 $19 = ((($v)) + 56|0);
 $20 = ((($v)) + 32|0);
 ;HEAP32[$19>>2]=4294967295|0;HEAP32[$19+4>>2]=4294967295|0;HEAP32[$19+8>>2]=4294967295|0;HEAP32[$19+12>>2]=4294967295|0;
 HEAP32[$20>>2] = 0;
 $21 = ((($3)) + 128|0);
 $22 = $21;
 $23 = $22;
 HEAP32[$23>>2] = -1;
 $24 = (($22) + 4)|0;
 $25 = $24;
 HEAP32[$25>>2] = -1;
 $$0 = 0;
 return ($$0|0);
}
function _vorbis_synthesis_init($v,$vi) {
 $v = $v|0;
 $vi = $vi|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__vds_shared_init($v,$vi,0)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  _vorbis_dsp_clear($v);
  $$0 = 1;
  return ($$0|0);
 }
 $2 = ((($v)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($v)) + 104|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 $7 = ($3|0)==(0|0);
 $or$cond$i = $7 | $6;
 if ($or$cond$i) {
  $$0 = 0;
  return ($$0|0);
 }
 $8 = ((($3)) + 28|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)==(0|0);
 if ($10) {
  $$0 = 0;
  return ($$0|0);
 }
 $11 = ((($9)) + 3656|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($9)) + 4|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12) + 1)|0;
 $16 = $14 >> $15;
 $17 = ((($v)) + 48|0);
 HEAP32[$17>>2] = $16;
 $18 = $16 >> $12;
 $19 = ((($v)) + 20|0);
 HEAP32[$19>>2] = $18;
 $20 = ((($v)) + 24|0);
 HEAP32[$20>>2] = -1;
 $21 = ((($v)) + 56|0);
 $22 = ((($v)) + 32|0);
 ;HEAP32[$21>>2]=4294967295|0;HEAP32[$21+4>>2]=4294967295|0;HEAP32[$21+8>>2]=4294967295|0;HEAP32[$21+12>>2]=4294967295|0;
 HEAP32[$22>>2] = 0;
 $23 = ((($5)) + 128|0);
 $24 = $23;
 $25 = $24;
 HEAP32[$25>>2] = -1;
 $26 = (($24) + 4)|0;
 $27 = $26;
 HEAP32[$27>>2] = -1;
 $$0 = 0;
 return ($$0|0);
}
function _vorbis_synthesis_blockin($v,$vb) {
 $v = $v|0;
 $vb = $vb|0;
 var $$ = 0, $$$ = 0, $$0 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$6 = 0, $$8 = 0, $$phi$trans$insert$phi$trans$insert = 0, $$pre37 = 0, $$pre38 = 0, $$pre39 = 0, $$pre40 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0;
 var $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0;
 var $177 = 0, $178 = 0.0, $179 = 0, $18 = 0, $180 = 0, $181 = 0.0, $182 = 0.0, $183 = 0, $184 = 0.0, $185 = 0, $186 = 0.0, $187 = 0.0, $188 = 0.0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0;
 var $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0.0, $204 = 0, $205 = 0, $206 = 0.0, $207 = 0.0, $208 = 0, $209 = 0.0, $21 = 0, $210 = 0, $211 = 0.0;
 var $212 = 0.0, $213 = 0.0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0.0, $23 = 0;
 var $230 = 0, $231 = 0, $232 = 0.0, $233 = 0.0, $234 = 0, $235 = 0.0, $236 = 0, $237 = 0.0, $238 = 0.0, $239 = 0.0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0.0, $248 = 0;
 var $249 = 0, $25 = 0, $250 = 0.0, $251 = 0.0, $252 = 0, $253 = 0.0, $254 = 0, $255 = 0.0, $256 = 0.0, $257 = 0.0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0;
 var $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0;
 var $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0;
 var $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0;
 var $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0;
 var $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0;
 var $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0;
 var $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0;
 var $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0;
 var $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $exitcond31 = 0, $exitcond32 = 0, $exitcond33 = 0, $exitcond34 = 0, $extra12$1 = 0, $i$014 = 0, $i$113 = 0, $i$2$lcssa = 0, $i$218 = 0;
 var $i$320 = 0, $i$416 = 0, $i$522 = 0, $j$024 = 0, $or$cond = 0, $or$cond7 = 0, $storemerge2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($v)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($v)) + 104|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($3)) + 3656|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($vb|0)==(0|0);
 if ($8) {
  $$0 = -131;
  return ($$0|0);
 }
 $9 = ((($v)) + 20|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($v)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($10|0)<=($12|0);
 $14 = ($12|0)==(-1);
 $or$cond = $13 | $14;
 if (!($or$cond)) {
  $$0 = -131;
  return ($$0|0);
 }
 $15 = ((($v)) + 40|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($v)) + 36|0);
 HEAP32[$17>>2] = $16;
 $18 = ((($vb)) + 28|0);
 $19 = HEAP32[$18>>2]|0;
 HEAP32[$15>>2] = $19;
 $20 = ((($v)) + 44|0);
 HEAP32[$20>>2] = -1;
 $21 = ((($v)) + 64|0);
 $22 = $21;
 $23 = $22;
 $24 = HEAP32[$23>>2]|0;
 $25 = (($22) + 4)|0;
 $26 = $25;
 $27 = HEAP32[$26>>2]|0;
 $28 = ($24|0)==(-1);
 $29 = ($27|0)==(-1);
 $30 = $28 & $29;
 if ($30) {
  $$phi$trans$insert$phi$trans$insert = ((($vb)) + 56|0);
  $31 = $$phi$trans$insert$phi$trans$insert;
  $32 = $31;
  $33 = HEAP32[$32>>2]|0;
  $34 = (($31) + 4)|0;
  $35 = $34;
  $36 = HEAP32[$35>>2]|0;
  $414 = $33;$415 = $36;
  label = 6;
 } else {
  $37 = (_i64Add(($24|0),($27|0),1,0)|0);
  $38 = tempRet0;
  $39 = ((($vb)) + 56|0);
  $40 = $39;
  $41 = $40;
  $42 = HEAP32[$41>>2]|0;
  $43 = (($40) + 4)|0;
  $44 = $43;
  $45 = HEAP32[$44>>2]|0;
  $46 = ($37|0)==($42|0);
  $47 = ($38|0)==($45|0);
  $48 = $46 & $47;
  if ($48) {
   $61 = $37;$64 = $38;
  } else {
   $414 = $42;$415 = $45;
   label = 6;
  }
 }
 if ((label|0) == 6) {
  $49 = ((($v)) + 56|0);
  $50 = $49;
  $51 = $50;
  HEAP32[$51>>2] = -1;
  $52 = (($50) + 4)|0;
  $53 = $52;
  HEAP32[$53>>2] = -1;
  $54 = ((($5)) + 128|0);
  $55 = $54;
  $56 = $55;
  HEAP32[$56>>2] = -1;
  $57 = (($55) + 4)|0;
  $58 = $57;
  HEAP32[$58>>2] = -1;
  $61 = $414;$64 = $415;
 }
 $59 = $21;
 $60 = $59;
 HEAP32[$60>>2] = $61;
 $62 = (($59) + 4)|0;
 $63 = $62;
 HEAP32[$63>>2] = $64;
 $65 = HEAP32[$vb>>2]|0;
 $66 = ($65|0)==(0|0);
 if ($66) {
  $350 = $10;$351 = $12;
 } else {
  $67 = (($3) + ($19<<2)|0);
  $68 = HEAP32[$67>>2]|0;
  $69 = (($7) + 1)|0;
  $70 = $68 >> $69;
  $71 = HEAP32[$3>>2]|0;
  $72 = $71 >> $69;
  $73 = ((($3)) + 4|0);
  $74 = HEAP32[$73>>2]|0;
  $75 = $74 >> $69;
  $76 = ((($vb)) + 88|0);
  $77 = HEAP32[$76>>2]|0;
  $78 = ($77|0)<(0);
  $79 = $78 << 31 >> 31;
  $80 = ((($v)) + 72|0);
  $81 = $80;
  $82 = $81;
  $83 = HEAP32[$82>>2]|0;
  $84 = (($81) + 4)|0;
  $85 = $84;
  $86 = HEAP32[$85>>2]|0;
  $87 = (_i64Add(($83|0),($86|0),($77|0),($79|0))|0);
  $88 = tempRet0;
  $89 = $80;
  $90 = $89;
  HEAP32[$90>>2] = $87;
  $91 = (($89) + 4)|0;
  $92 = $91;
  HEAP32[$92>>2] = $88;
  $93 = ((($vb)) + 92|0);
  $94 = HEAP32[$93>>2]|0;
  $95 = ($94|0)<(0);
  $96 = $95 << 31 >> 31;
  $97 = ((($v)) + 80|0);
  $98 = $97;
  $99 = $98;
  $100 = HEAP32[$99>>2]|0;
  $101 = (($98) + 4)|0;
  $102 = $101;
  $103 = HEAP32[$102>>2]|0;
  $104 = (_i64Add(($100|0),($103|0),($94|0),($96|0))|0);
  $105 = tempRet0;
  $106 = $97;
  $107 = $106;
  HEAP32[$107>>2] = $104;
  $108 = (($106) + 4)|0;
  $109 = $108;
  HEAP32[$109>>2] = $105;
  $110 = ((($vb)) + 96|0);
  $111 = HEAP32[$110>>2]|0;
  $112 = ($111|0)<(0);
  $113 = $112 << 31 >> 31;
  $114 = ((($v)) + 88|0);
  $115 = $114;
  $116 = $115;
  $117 = HEAP32[$116>>2]|0;
  $118 = (($115) + 4)|0;
  $119 = $118;
  $120 = HEAP32[$119>>2]|0;
  $121 = (_i64Add(($117|0),($120|0),($111|0),($113|0))|0);
  $122 = tempRet0;
  $123 = $114;
  $124 = $123;
  HEAP32[$124>>2] = $121;
  $125 = (($123) + 4)|0;
  $126 = $125;
  HEAP32[$126>>2] = $122;
  $127 = ((($vb)) + 100|0);
  $128 = HEAP32[$127>>2]|0;
  $129 = ($128|0)<(0);
  $130 = $129 << 31 >> 31;
  $131 = ((($v)) + 96|0);
  $132 = $131;
  $133 = $132;
  $134 = HEAP32[$133>>2]|0;
  $135 = (($132) + 4)|0;
  $136 = $135;
  $137 = HEAP32[$136>>2]|0;
  $138 = (_i64Add(($134|0),($137|0),($128|0),($130|0))|0);
  $139 = tempRet0;
  $140 = $131;
  $141 = $140;
  HEAP32[$141>>2] = $138;
  $142 = (($140) + 4)|0;
  $143 = $142;
  HEAP32[$143>>2] = $139;
  $144 = ((($v)) + 48|0);
  $145 = HEAP32[$144>>2]|0;
  $146 = ($145|0)==(0);
  $$3 = $146 ? $75 : 0;
  $$4 = $146 ? 0 : $75;
  $147 = ((($1)) + 4|0);
  $148 = HEAP32[$147>>2]|0;
  $149 = ($148|0)>(0);
  if ($149) {
   $150 = ((($5)) + 4|0);
   $151 = ((($v)) + 8|0);
   $152 = (($75|0) / 2)&-1;
   $153 = (($72|0) / 2)&-1;
   $154 = (0 - ($153))|0;
   $155 = ($72|0)>(0);
   $156 = (($152) + ($153))|0;
   $157 = (($72) + -1)|0;
   $158 = ($70|0)>(0);
   $159 = ((($5)) + 8|0);
   $160 = ($75|0)>(0);
   $161 = (($75) + -1)|0;
   $162 = (($72|0) / -2)&-1;
   $164 = $16;$166 = $19;$j$024 = 0;
   while(1) {
    $163 = ($164|0)==(0);
    $165 = ($166|0)!=(0);
    L19: do {
     if ($163) {
      $215 = HEAP32[$150>>2]|0;
      $216 = (($215) - ($7))|0;
      $217 = (__vorbis_window_get($216)|0);
      $218 = HEAP32[$151>>2]|0;
      $219 = (($218) + ($j$024<<2)|0);
      $220 = HEAP32[$219>>2]|0;
      $221 = (($220) + ($$3<<2)|0);
      $222 = HEAP32[$vb>>2]|0;
      $223 = (($222) + ($j$024<<2)|0);
      $224 = HEAP32[$223>>2]|0;
      if (!($165)) {
       if ($155) {
        $i$416 = 0;
       } else {
        $260 = $220;$262 = $224;
        break;
       }
       while(1) {
        $246 = (($221) + ($i$416<<2)|0);
        $247 = +HEAPF32[$246>>2];
        $248 = (($157) - ($i$416))|0;
        $249 = (($217) + ($248<<2)|0);
        $250 = +HEAPF32[$249>>2];
        $251 = $247 * $250;
        $252 = (($224) + ($i$416<<2)|0);
        $253 = +HEAPF32[$252>>2];
        $254 = (($217) + ($i$416<<2)|0);
        $255 = +HEAPF32[$254>>2];
        $256 = $253 * $255;
        $257 = $251 + $256;
        HEAPF32[$246>>2] = $257;
        $258 = (($i$416) + 1)|0;
        $exitcond32 = ($258|0)==($72|0);
        if ($exitcond32) {
         $260 = $220;$262 = $224;
         break L19;
        } else {
         $i$416 = $258;
        }
       }
      }
      $225 = (($224) + ($152<<2)|0);
      $226 = (($225) + ($154<<2)|0);
      if ($155) {
       $i$218 = 0;
       while(1) {
        $228 = (($221) + ($i$218<<2)|0);
        $229 = +HEAPF32[$228>>2];
        $230 = (($157) - ($i$218))|0;
        $231 = (($217) + ($230<<2)|0);
        $232 = +HEAPF32[$231>>2];
        $233 = $229 * $232;
        $234 = (($226) + ($i$218<<2)|0);
        $235 = +HEAPF32[$234>>2];
        $236 = (($217) + ($i$218<<2)|0);
        $237 = +HEAPF32[$236>>2];
        $238 = $235 * $237;
        $239 = $233 + $238;
        HEAPF32[$228>>2] = $239;
        $240 = (($i$218) + 1)|0;
        $exitcond33 = ($240|0)==($72|0);
        if ($exitcond33) {
         $i$2$lcssa = $72;
         break;
        } else {
         $i$218 = $240;
        }
       }
      } else {
       $i$2$lcssa = 0;
      }
      $227 = ($i$2$lcssa|0)<($156|0);
      if ($227) {
       $i$320 = $i$2$lcssa;
       while(1) {
        $241 = (($226) + ($i$320<<2)|0);
        $242 = HEAP32[$241>>2]|0;
        $243 = (($221) + ($i$320<<2)|0);
        HEAP32[$243>>2] = $242;
        $244 = (($i$320) + 1)|0;
        $245 = ($244|0)<($156|0);
        if ($245) {
         $i$320 = $244;
        } else {
         $260 = $220;$262 = $224;
         break;
        }
       }
      } else {
       $260 = $220;$262 = $224;
      }
     } else {
      if ($165) {
       $167 = HEAP32[$159>>2]|0;
       $168 = (($167) - ($7))|0;
       $169 = (__vorbis_window_get($168)|0);
       $170 = HEAP32[$151>>2]|0;
       $171 = (($170) + ($j$024<<2)|0);
       $172 = HEAP32[$171>>2]|0;
       $173 = (($172) + ($$3<<2)|0);
       $174 = HEAP32[$vb>>2]|0;
       $175 = (($174) + ($j$024<<2)|0);
       $176 = HEAP32[$175>>2]|0;
       if ($160) {
        $i$014 = 0;
       } else {
        $260 = $172;$262 = $176;
        break;
       }
       while(1) {
        $177 = (($173) + ($i$014<<2)|0);
        $178 = +HEAPF32[$177>>2];
        $179 = (($161) - ($i$014))|0;
        $180 = (($169) + ($179<<2)|0);
        $181 = +HEAPF32[$180>>2];
        $182 = $178 * $181;
        $183 = (($176) + ($i$014<<2)|0);
        $184 = +HEAPF32[$183>>2];
        $185 = (($169) + ($i$014<<2)|0);
        $186 = +HEAPF32[$185>>2];
        $187 = $184 * $186;
        $188 = $182 + $187;
        HEAPF32[$177>>2] = $188;
        $189 = (($i$014) + 1)|0;
        $exitcond31 = ($189|0)==($75|0);
        if ($exitcond31) {
         $260 = $172;$262 = $176;
         break;
        } else {
         $i$014 = $189;
        }
       }
      } else {
       $190 = HEAP32[$150>>2]|0;
       $191 = (($190) - ($7))|0;
       $192 = (__vorbis_window_get($191)|0);
       $193 = HEAP32[$151>>2]|0;
       $194 = (($193) + ($j$024<<2)|0);
       $195 = HEAP32[$194>>2]|0;
       $196 = (($195) + ($$3<<2)|0);
       $197 = (($196) + ($152<<2)|0);
       $198 = (($197) + ($162<<2)|0);
       $199 = HEAP32[$vb>>2]|0;
       $200 = (($199) + ($j$024<<2)|0);
       $201 = HEAP32[$200>>2]|0;
       if ($155) {
        $i$113 = 0;
       } else {
        $260 = $195;$262 = $201;
        break;
       }
       while(1) {
        $202 = (($198) + ($i$113<<2)|0);
        $203 = +HEAPF32[$202>>2];
        $204 = (($157) - ($i$113))|0;
        $205 = (($192) + ($204<<2)|0);
        $206 = +HEAPF32[$205>>2];
        $207 = $203 * $206;
        $208 = (($201) + ($i$113<<2)|0);
        $209 = +HEAPF32[$208>>2];
        $210 = (($192) + ($i$113<<2)|0);
        $211 = +HEAPF32[$210>>2];
        $212 = $209 * $211;
        $213 = $207 + $212;
        HEAPF32[$202>>2] = $213;
        $214 = (($i$113) + 1)|0;
        $exitcond = ($214|0)==($72|0);
        if ($exitcond) {
         $260 = $195;$262 = $201;
         break;
        } else {
         $i$113 = $214;
        }
       }
      }
     }
    } while(0);
    $259 = (($260) + ($$4<<2)|0);
    $261 = (($262) + ($70<<2)|0);
    if ($158) {
     $i$522 = 0;
     while(1) {
      $263 = (($261) + ($i$522<<2)|0);
      $264 = HEAP32[$263>>2]|0;
      $265 = (($259) + ($i$522<<2)|0);
      HEAP32[$265>>2] = $264;
      $266 = (($i$522) + 1)|0;
      $exitcond34 = ($266|0)==($70|0);
      if ($exitcond34) {
       break;
      } else {
       $i$522 = $266;
      }
     }
    }
    $267 = (($j$024) + 1)|0;
    $268 = HEAP32[$147>>2]|0;
    $269 = ($267|0)<($268|0);
    if (!($269)) {
     break;
    }
    $$pre37 = HEAP32[$17>>2]|0;
    $$pre38 = HEAP32[$15>>2]|0;
    $164 = $$pre37;$166 = $$pre38;$j$024 = $267;
   }
   $$pre39 = HEAP32[$144>>2]|0;
   $$pre40 = HEAP32[$11>>2]|0;
   $271 = $$pre39;$273 = $$pre40;
  } else {
   $271 = $145;$273 = $12;
  }
  $270 = ($271|0)==(0);
  $$5 = $270 ? $75 : 0;
  HEAP32[$144>>2] = $$5;
  $272 = ($273|0)==(-1);
  if ($272) {
   HEAP32[$11>>2] = $$4;
   $416 = $$4;$storemerge2 = $$4;
  } else {
   HEAP32[$11>>2] = $$3;
   $274 = HEAP32[$17>>2]|0;
   $275 = (($3) + ($274<<2)|0);
   $276 = HEAP32[$275>>2]|0;
   $277 = (($276|0) / 4)&-1;
   $278 = HEAP32[$15>>2]|0;
   $279 = (($3) + ($278<<2)|0);
   $280 = HEAP32[$279>>2]|0;
   $281 = (($280|0) / 4)&-1;
   $282 = (($281) + ($277))|0;
   $283 = $282 >> $7;
   $284 = (($283) + ($$3))|0;
   $416 = $$3;$storemerge2 = $284;
  }
  HEAP32[$9>>2] = $storemerge2;
  $350 = $storemerge2;$351 = $416;
 }
 $285 = ((($5)) + 128|0);
 $286 = $285;
 $287 = $286;
 $288 = HEAP32[$287>>2]|0;
 $289 = (($286) + 4)|0;
 $290 = $289;
 $291 = HEAP32[$290>>2]|0;
 $292 = ($288|0)==(-1);
 $293 = ($291|0)==(-1);
 $294 = $292 & $293;
 if ($294) {
  $310 = 0;$313 = 0;
 } else {
  $295 = HEAP32[$17>>2]|0;
  $296 = (($3) + ($295<<2)|0);
  $297 = HEAP32[$296>>2]|0;
  $298 = (($297|0) / 4)&-1;
  $299 = HEAP32[$15>>2]|0;
  $300 = (($3) + ($299<<2)|0);
  $301 = HEAP32[$300>>2]|0;
  $302 = (($301|0) / 4)&-1;
  $303 = (($302) + ($298))|0;
  $304 = ($303|0)<(0);
  $305 = $304 << 31 >> 31;
  $306 = (_i64Add(($303|0),($305|0),($288|0),($291|0))|0);
  $307 = tempRet0;
  $310 = $306;$313 = $307;
 }
 $308 = $285;
 $309 = $308;
 HEAP32[$309>>2] = $310;
 $311 = (($308) + 4)|0;
 $312 = $311;
 HEAP32[$312>>2] = $313;
 $314 = ((($v)) + 56|0);
 $315 = $314;
 $316 = $315;
 $317 = HEAP32[$316>>2]|0;
 $318 = (($315) + 4)|0;
 $319 = $318;
 $320 = HEAP32[$319>>2]|0;
 $321 = ($317|0)==(-1);
 $322 = ($320|0)==(-1);
 $323 = $321 & $322;
 do {
  if ($323) {
   $324 = ((($vb)) + 48|0);
   $325 = $324;
   $326 = $325;
   $327 = HEAP32[$326>>2]|0;
   $328 = (($325) + 4)|0;
   $329 = $328;
   $330 = HEAP32[$329>>2]|0;
   $331 = ($327|0)==(-1);
   $332 = ($330|0)==(-1);
   $333 = $331 & $332;
   if (!($333)) {
    $334 = $314;
    $335 = $334;
    HEAP32[$335>>2] = $327;
    $336 = (($334) + 4)|0;
    $337 = $336;
    HEAP32[$337>>2] = $330;
    $338 = ($313|0)>($330|0);
    $339 = ($310>>>0)>($327>>>0);
    $340 = ($313|0)==($330|0);
    $341 = $340 & $339;
    $342 = $338 | $341;
    if ($342) {
     $343 = (_i64Subtract(($310|0),($313|0),($327|0),($330|0))|0);
     $344 = tempRet0;
     $345 = ($343|0)<(0);
     $$ = $345 ? 0 : $343;
     $346 = ((($vb)) + 44|0);
     $347 = HEAP32[$346>>2]|0;
     $348 = ($347|0)==(0);
     if ($348) {
      $356 = $$ >> $7;
      $357 = (($351) + ($356))|0;
      $358 = ($357|0)>($350|0);
      $$6 = $358 ? $350 : $357;
      HEAP32[$11>>2] = $$6;
      break;
     } else {
      $349 = (($350) - ($351))|0;
      $352 = $349 << $7;
      $353 = ($$|0)>($352|0);
      $$$ = $353 ? $352 : $$;
      $354 = $$$ >> $7;
      $355 = (($350) - ($354))|0;
      HEAP32[$9>>2] = $355;
      break;
     }
    }
   }
  } else {
   $359 = HEAP32[$17>>2]|0;
   $360 = (($3) + ($359<<2)|0);
   $361 = HEAP32[$360>>2]|0;
   $362 = (($361|0) / 4)&-1;
   $363 = HEAP32[$15>>2]|0;
   $364 = (($3) + ($363<<2)|0);
   $365 = HEAP32[$364>>2]|0;
   $366 = (($365|0) / 4)&-1;
   $367 = (($366) + ($362))|0;
   $368 = ($367|0)<(0);
   $369 = $368 << 31 >> 31;
   $370 = (_i64Add(($367|0),($369|0),($317|0),($320|0))|0);
   $371 = tempRet0;
   $372 = $314;
   $373 = $372;
   HEAP32[$373>>2] = $370;
   $374 = (($372) + 4)|0;
   $375 = $374;
   HEAP32[$375>>2] = $371;
   $376 = ((($vb)) + 48|0);
   $377 = $376;
   $378 = $377;
   $379 = HEAP32[$378>>2]|0;
   $380 = (($377) + 4)|0;
   $381 = $380;
   $382 = HEAP32[$381>>2]|0;
   $383 = ($379|0)==(-1);
   $384 = ($382|0)==(-1);
   $385 = $383 & $384;
   $386 = ($370|0)==($379|0);
   $387 = ($371|0)==($382|0);
   $388 = $386 & $387;
   $or$cond7 = $385 | $388;
   if (!($or$cond7)) {
    $389 = ($371|0)>($382|0);
    $390 = ($370>>>0)>($379>>>0);
    $391 = ($371|0)==($382|0);
    $392 = $391 & $390;
    $393 = $389 | $392;
    if ($393) {
     $394 = (_i64Subtract(($370|0),($371|0),($379|0),($382|0))|0);
     $395 = tempRet0;
     $396 = ($394|0)==(0);
     if (!($396)) {
      $397 = ((($vb)) + 44|0);
      $398 = HEAP32[$397>>2]|0;
      $399 = ($398|0)==(0);
      if (!($399)) {
       $400 = (($350) - ($351))|0;
       $401 = $400 << $7;
       $402 = ($394|0)>($401|0);
       $$8 = $402 ? $401 : $394;
       $403 = ($$8|0)<(0);
       $extra12$1 = $403 ? 0 : $$8;
       $404 = $extra12$1 >> $7;
       $405 = (($350) - ($404))|0;
       HEAP32[$9>>2] = $405;
      }
     }
    }
    $406 = $314;
    $407 = $406;
    HEAP32[$407>>2] = $379;
    $408 = (($406) + 4)|0;
    $409 = $408;
    HEAP32[$409>>2] = $382;
   }
  }
 } while(0);
 $410 = ((($vb)) + 44|0);
 $411 = HEAP32[$410>>2]|0;
 $412 = ($411|0)==(0);
 if ($412) {
  $$0 = 0;
  return ($$0|0);
 }
 $413 = ((($v)) + 32|0);
 HEAP32[$413>>2] = 1;
 $$0 = 0;
 return ($$0|0);
}
function __vorbis_window_get($n) {
 $n = $n|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (67568 + ($n<<2)|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function _vorbis_synthesis_pcmout($v,$pcm) {
 $v = $v|0;
 $pcm = $pcm|0;
 var $$0 = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($v)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($v)) + 24|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)>(-1);
 if (!($4)) {
  $$0 = 0;
  return ($$0|0);
 }
 $5 = ((($v)) + 20|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)>($3|0);
 if (!($7)) {
  $$0 = 0;
  return ($$0|0);
 }
 $8 = ($pcm|0)==(0|0);
 if (!($8)) {
  $9 = ((($1)) + 4|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ($10|0)>(0);
  if ($11) {
   $12 = ((($v)) + 8|0);
   $13 = ((($v)) + 12|0);
   $i$01 = 0;
   while(1) {
    $14 = HEAP32[$12>>2]|0;
    $15 = (($14) + ($i$01<<2)|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = (($16) + ($3<<2)|0);
    $18 = HEAP32[$13>>2]|0;
    $19 = (($18) + ($i$01<<2)|0);
    HEAP32[$19>>2] = $17;
    $20 = (($i$01) + 1)|0;
    $21 = ($20|0)<($10|0);
    if ($21) {
     $i$01 = $20;
    } else {
     $$pre$phiZ2D = $13;
     break;
    }
   }
  } else {
   $$pre = ((($v)) + 12|0);
   $$pre$phiZ2D = $$pre;
  }
  $22 = HEAP32[$$pre$phiZ2D>>2]|0;
  HEAP32[$pcm>>2] = $22;
 }
 $23 = (($6) - ($3))|0;
 $$0 = $23;
 return ($$0|0);
}
function _vorbis_synthesis_read($v,$n) {
 $v = $v|0;
 $n = $n|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)==(0);
 $$phi$trans$insert = ((($v)) + 24|0);
 $$pre = HEAP32[$$phi$trans$insert>>2]|0;
 if (!($0)) {
  $1 = (($$pre) + ($n))|0;
  $2 = ((($v)) + 20|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ($1|0)>($3|0);
  if ($4) {
   $$0 = -131;
   return ($$0|0);
  }
 }
 $5 = (($$pre) + ($n))|0;
 HEAP32[$$phi$trans$insert>>2] = $5;
 $$0 = 0;
 return ($$0|0);
}
function __book_maptype1_quantvals($b) {
 $b = $b|0;
 var $$lcssa = 0, $$lcssa12 = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0;
 var $8 = 0.0, $9 = 0.0, $acc$02$us = 0, $acc1$03$us = 0, $i$04$us = 0, $or$cond$us = 0, $vals$0$us = 0, $vals$0$us$lcssa = 0, $vals$1$us = 0, $vals$1$v$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($b)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$b>>2]|0;
 $3 = ($2|0)>(0);
 if (!($3)) {
  while(1) {
  }
 }
 $4 = (+($1|0));
 $5 = $4;
 $6 = (+($2|0));
 $7 = 1.0 / $6;
 $8 = $7;
 $9 = (+Math_pow((+$5),(+$8)));
 $10 = (+Math_floor((+$9)));
 $11 = (~~(($10)));
 $vals$0$us = $11;
 while(1) {
  $15 = (($vals$0$us) + 1)|0;
  $acc$02$us = 1;$acc1$03$us = 1;$i$04$us = 0;
  while(1) {
   $13 = Math_imul($acc$02$us, $vals$0$us)|0;
   $14 = Math_imul($acc1$03$us, $15)|0;
   $16 = (($i$04$us) + 1)|0;
   $17 = ($16|0)<($2|0);
   if ($17) {
    $acc$02$us = $13;$acc1$03$us = $14;$i$04$us = $16;
   } else {
    $$lcssa = $13;$$lcssa12 = $14;
    break;
   }
  }
  $18 = ($$lcssa|0)<=($1|0);
  $19 = ($$lcssa12|0)>($1|0);
  $or$cond$us = $18 & $19;
  if ($or$cond$us) {
   $vals$0$us$lcssa = $vals$0$us;
   break;
  }
  $12 = ($$lcssa|0)>($1|0);
  $vals$1$v$us = $12 ? -1 : 1;
  $vals$1$us = (($vals$1$v$us) + ($vals$0$us))|0;
  $vals$0$us = $vals$1$us;
 }
 return ($vals$0$us$lcssa|0);
}
function _vorbis_staticbook_unpack($opb) {
 $opb = $opb|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $exitcond41 = 0, $i$025 = 0, $i$127 = 0, $i$233 = 0, $i$3$lcssa = 0, $i$330 = 0, $i$524 = 0, $j$029 = 0, $length$034 = 0, $or$cond = 0, $quantvals$0 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_calloc(1,40)|0);
 $1 = ((($0)) + 36|0);
 HEAP32[$1>>2] = 1;
 $2 = (_oggpack_read($opb,24)|0);
 $3 = ($2|0)==(5653314);
 L1: do {
  if ($3) {
   $4 = (_oggpack_read($opb,16)|0);
   HEAP32[$0>>2] = $4;
   $5 = (_oggpack_read($opb,24)|0);
   $6 = ((($0)) + 4|0);
   HEAP32[$6>>2] = $5;
   $7 = ($5|0)==(-1);
   if (!($7)) {
    $8 = HEAP32[$0>>2]|0;
    $9 = (_ov_ilog($8)|0);
    $10 = HEAP32[$6>>2]|0;
    $11 = (_ov_ilog($10)|0);
    $12 = (($11) + ($9))|0;
    $13 = ($12|0)>(24);
    if (!($13)) {
     $14 = (_oggpack_read($opb,1)|0);
     L5: do {
      switch ($14|0) {
      case 0:  {
       $15 = (_oggpack_read($opb,1)|0);
       $16 = HEAP32[$6>>2]|0;
       $17 = ($15|0)!=(0);
       $18 = $17 ? 1 : 5;
       $19 = Math_imul($18, $16)|0;
       $20 = (($19) + 7)|0;
       $21 = $20 >> 3;
       $22 = ((($opb)) + 16|0);
       $23 = HEAP32[$22>>2]|0;
       $24 = (_oggpack_bytes($opb)|0);
       $25 = (($23) - ($24))|0;
       $26 = ($21|0)>($25|0);
       if ($26) {
        break L1;
       }
       $27 = HEAP32[$6>>2]|0;
       $28 = (_malloc($27)|0);
       $29 = ((($0)) + 8|0);
       HEAP32[$29>>2] = $28;
       $30 = ($27|0)>(0);
       if (!($17)) {
        if ($30) {
         $i$127 = 0;
        } else {
         break L5;
        }
        while(1) {
         $44 = (_oggpack_read($opb,5)|0);
         $45 = ($44|0)==(-1);
         if ($45) {
          break L1;
         }
         $46 = (($44) + 1)|0;
         $47 = $46&255;
         $48 = HEAP32[$29>>2]|0;
         $49 = (($48) + ($i$127)|0);
         HEAP8[$49>>0] = $47;
         $50 = (($i$127) + 1)|0;
         $51 = HEAP32[$6>>2]|0;
         $52 = ($50|0)<($51|0);
         if ($52) {
          $i$127 = $50;
         } else {
          break L5;
         }
        }
       }
       if ($30) {
        $i$025 = 0;
        while(1) {
         $31 = (_oggpack_read($opb,1)|0);
         $32 = ($31|0)==(0);
         if ($32) {
          $39 = HEAP32[$29>>2]|0;
          $40 = (($39) + ($i$025)|0);
          HEAP8[$40>>0] = 0;
         } else {
          $33 = (_oggpack_read($opb,5)|0);
          $34 = ($33|0)==(-1);
          if ($34) {
           break L1;
          }
          $35 = (($33) + 1)|0;
          $36 = $35&255;
          $37 = HEAP32[$29>>2]|0;
          $38 = (($37) + ($i$025)|0);
          HEAP8[$38>>0] = $36;
         }
         $41 = (($i$025) + 1)|0;
         $42 = HEAP32[$6>>2]|0;
         $43 = ($41|0)<($42|0);
         if ($43) {
          $i$025 = $41;
         } else {
          break;
         }
        }
       }
       break;
      }
      case 1:  {
       $53 = (_oggpack_read($opb,5)|0);
       $54 = (($53) + 1)|0;
       $55 = ($54|0)==(0);
       if ($55) {
        break L1;
       }
       $56 = HEAP32[$6>>2]|0;
       $57 = (_malloc($56)|0);
       $58 = ((($0)) + 8|0);
       HEAP32[$58>>2] = $57;
       $59 = ($56|0)>(0);
       if ($59) {
        $61 = $56;$i$233 = 0;$length$034 = $54;
        while(1) {
         $60 = (($61) - ($i$233))|0;
         $62 = (_ov_ilog($60)|0);
         $63 = (_oggpack_read($opb,$62)|0);
         $64 = ($63|0)==(-1);
         $65 = ($length$034|0)>(32);
         $or$cond = $65 | $64;
         if ($or$cond) {
          break L1;
         }
         $66 = HEAP32[$6>>2]|0;
         $67 = (($66) - ($i$233))|0;
         $68 = ($63|0)>($67|0);
         if ($68) {
          break L1;
         }
         $69 = ($63|0)>(0);
         if ($69) {
          $70 = (($63) + -1)|0;
          $71 = (($length$034) + -1)|0;
          $72 = $70 >> $71;
          $73 = ($72|0)>(1);
          if ($73) {
           break L1;
          }
          $74 = $length$034&255;
          $i$330 = $i$233;$j$029 = 0;
          while(1) {
           $75 = HEAP32[$58>>2]|0;
           $76 = (($75) + ($i$330)|0);
           HEAP8[$76>>0] = $74;
           $77 = (($j$029) + 1)|0;
           $78 = (($i$330) + 1)|0;
           $exitcond41 = ($77|0)==($63|0);
           if ($exitcond41) {
            break;
           } else {
            $i$330 = $78;$j$029 = $77;
           }
          }
          $79 = (($63) + ($i$233))|0;
          $$pre = HEAP32[$6>>2]|0;
          $82 = $$pre;$i$3$lcssa = $79;
         } else {
          $82 = $66;$i$3$lcssa = $i$233;
         }
         $80 = (($length$034) + 1)|0;
         $81 = ($82|0)>($i$3$lcssa|0);
         if ($81) {
          $61 = $82;$i$233 = $i$3$lcssa;$length$034 = $80;
         } else {
          break;
         }
        }
       }
       break;
      }
      default: {
       break L1;
      }
      }
     } while(0);
     $83 = (_oggpack_read($opb,4)|0);
     $84 = ((($0)) + 12|0);
     HEAP32[$84>>2] = $83;
     switch ($83|0) {
     case 2: case 1:  {
      break;
     }
     case 0:  {
      $$0 = $0;
      return ($$0|0);
      break;
     }
     default: {
      break L1;
     }
     }
     $85 = (_oggpack_read($opb,32)|0);
     $86 = ((($0)) + 16|0);
     HEAP32[$86>>2] = $85;
     $87 = (_oggpack_read($opb,32)|0);
     $88 = ((($0)) + 20|0);
     HEAP32[$88>>2] = $87;
     $89 = (_oggpack_read($opb,4)|0);
     $90 = (($89) + 1)|0;
     $91 = ((($0)) + 24|0);
     HEAP32[$91>>2] = $90;
     $92 = (_oggpack_read($opb,1)|0);
     $93 = ((($0)) + 28|0);
     HEAP32[$93>>2] = $92;
     $94 = ($92|0)==(-1);
     if (!($94)) {
      $95 = HEAP32[$84>>2]|0;
      switch ($95|0) {
      case 1:  {
       $96 = HEAP32[$0>>2]|0;
       $97 = ($96|0)==(0);
       if ($97) {
        $quantvals$0 = 0;
       } else {
        $98 = (__book_maptype1_quantvals($0)|0);
        $quantvals$0 = $98;
       }
       break;
      }
      case 2:  {
       $99 = HEAP32[$6>>2]|0;
       $100 = HEAP32[$0>>2]|0;
       $101 = Math_imul($100, $99)|0;
       $quantvals$0 = $101;
       break;
      }
      default: {
       $quantvals$0 = 0;
      }
      }
      $102 = HEAP32[$91>>2]|0;
      $103 = Math_imul($102, $quantvals$0)|0;
      $104 = (($103) + 7)|0;
      $105 = $104 >> 3;
      $106 = ((($opb)) + 16|0);
      $107 = HEAP32[$106>>2]|0;
      $108 = (_oggpack_bytes($opb)|0);
      $109 = (($107) - ($108))|0;
      $110 = ($105|0)>($109|0);
      if (!($110)) {
       $111 = $quantvals$0 << 2;
       $112 = (_malloc($111)|0);
       $113 = ((($0)) + 32|0);
       HEAP32[$113>>2] = $112;
       $114 = ($quantvals$0|0)>(0);
       if ($114) {
        $i$524 = 0;
        while(1) {
         $115 = HEAP32[$91>>2]|0;
         $116 = (_oggpack_read($opb,$115)|0);
         $117 = HEAP32[$113>>2]|0;
         $118 = (($117) + ($i$524<<2)|0);
         HEAP32[$118>>2] = $116;
         $119 = (($i$524) + 1)|0;
         $exitcond = ($119|0)==($quantvals$0|0);
         if ($exitcond) {
          $123 = $117;
          break;
         } else {
          $i$524 = $119;
         }
        }
       } else {
        $123 = $112;
       }
       $120 = ($quantvals$0|0)==(0);
       if ($120) {
        $$0 = $0;
        return ($$0|0);
       }
       $121 = (($quantvals$0) + -1)|0;
       $122 = (($123) + ($121<<2)|0);
       $124 = HEAP32[$122>>2]|0;
       $125 = ($124|0)==(-1);
       if (!($125)) {
        $$0 = $0;
        return ($$0|0);
       }
      }
     }
    }
   }
  }
 } while(0);
 _vorbis_staticbook_destroy($0);
 $$0 = 0;
 return ($$0|0);
}
function _vorbis_comment_init($vc) {
 $vc = $vc|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$vc>>2]=0|0;HEAP32[$vc+4>>2]=0|0;HEAP32[$vc+8>>2]=0|0;HEAP32[$vc+12>>2]=0|0;
 return;
}
function _vorbis_comment_clear($vc) {
 $vc = $vc|0;
 var $$pre = 0, $$pre2 = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($vc|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = HEAP32[$vc>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = ((($vc)) + 8|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ($4|0)>(0);
  if ($5) {
   $20 = $4;$7 = $1;$i$01 = 0;
   while(1) {
    $6 = (($7) + ($i$01<<2)|0);
    $8 = HEAP32[$6>>2]|0;
    $9 = ($8|0)==(0|0);
    if ($9) {
     $12 = $20;
    } else {
     _free($8);
     $$pre2 = HEAP32[$3>>2]|0;
     $12 = $$pre2;
    }
    $10 = (($i$01) + 1)|0;
    $11 = ($10|0)<($12|0);
    if (!($11)) {
     break;
    }
    $$pre = HEAP32[$vc>>2]|0;
    $20 = $12;$7 = $$pre;$i$01 = $10;
   }
   $$pre3 = HEAP32[$vc>>2]|0;
   $13 = $$pre3;
  } else {
   $13 = $1;
  }
  _free($13);
 }
 $14 = ((($vc)) + 4|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)==(0|0);
 if (!($16)) {
  _free($15);
 }
 $17 = ((($vc)) + 12|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ($18|0)==(0|0);
 if (!($19)) {
  _free($18);
 }
 ;HEAP32[$vc>>2]=0|0;HEAP32[$vc+4>>2]=0|0;HEAP32[$vc+8>>2]=0|0;HEAP32[$vc+12>>2]=0|0;
 return;
}
function _vorbis_info_blocksize($vi,$zo) {
 $vi = $vi|0;
 $zo = $zo|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($vi)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $5 = -1;
  return ($5|0);
 }
 $3 = (($1) + ($zo<<2)|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4;
 return ($5|0);
}
function _vorbis_info_init($vi) {
 $vi = $vi|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$vi>>2]=0|0;HEAP32[$vi+4>>2]=0|0;HEAP32[$vi+8>>2]=0|0;HEAP32[$vi+12>>2]=0|0;HEAP32[$vi+16>>2]=0|0;HEAP32[$vi+20>>2]=0|0;HEAP32[$vi+24>>2]=0|0;
 $0 = (_calloc(1,3664)|0);
 $1 = ((($vi)) + 28|0);
 HEAP32[$1>>2] = $0;
 return;
}
function _vorbis_info_clear($vi) {
 $vi = $vi|0;
 var $$pre = 0, $$pre18 = 0, $$pre19 = 0, $$pre20 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $9 = 0, $i$016 = 0, $i$114 = 0, $i$212 = 0, $i$310 = 0, $i$47 = 0, $i$56 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($vi)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  ;HEAP32[$vi>>2]=0|0;HEAP32[$vi+4>>2]=0|0;HEAP32[$vi+8>>2]=0|0;HEAP32[$vi+12>>2]=0|0;HEAP32[$vi+16>>2]=0|0;HEAP32[$vi+20>>2]=0|0;HEAP32[$vi+24>>2]=0|0;HEAP32[$vi+28>>2]=0|0;
  return;
 }
 $3 = ((($1)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)>(0);
 if ($5) {
  $80 = $4;$i$016 = 0;
  while(1) {
   $9 = (((($1)) + 32|0) + ($i$016<<2)|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($10|0)==(0|0);
   if ($11) {
    $14 = $80;
   } else {
    _free($10);
    $$pre = HEAP32[$3>>2]|0;
    $14 = $$pre;
   }
   $12 = (($i$016) + 1)|0;
   $13 = ($12|0)<($14|0);
   if ($13) {
    $80 = $14;$i$016 = $12;
   } else {
    break;
   }
  }
 }
 $6 = ((($1)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)>(0);
 if ($8) {
  $81 = $7;$i$114 = 0;
  while(1) {
   $18 = (((($1)) + 544|0) + ($i$114<<2)|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ($19|0)==(0|0);
   if ($20) {
    $29 = $81;
   } else {
    $21 = (((($1)) + 288|0) + ($i$114<<2)|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = (101380 + ($22<<2)|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = ((($24)) + 8|0);
    $26 = HEAP32[$25>>2]|0;
    FUNCTION_TABLE_vi[$26 & 8191]($19);
    $$pre18 = HEAP32[$6>>2]|0;
    $29 = $$pre18;
   }
   $27 = (($i$114) + 1)|0;
   $28 = ($27|0)<($29|0);
   if ($28) {
    $81 = $29;$i$114 = $27;
   } else {
    break;
   }
  }
 }
 $15 = ((($1)) + 16|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ($16|0)>(0);
 if ($17) {
  $82 = $16;$i$212 = 0;
  while(1) {
   $33 = (((($1)) + 1056|0) + ($i$212<<2)|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = ($34|0)==(0|0);
   if ($35) {
    $44 = $82;
   } else {
    $36 = (((($1)) + 800|0) + ($i$212<<2)|0);
    $37 = HEAP32[$36>>2]|0;
    $38 = (101372 + ($37<<2)|0);
    $39 = HEAP32[$38>>2]|0;
    $40 = ((($39)) + 12|0);
    $41 = HEAP32[$40>>2]|0;
    FUNCTION_TABLE_vi[$41 & 8191]($34);
    $$pre19 = HEAP32[$15>>2]|0;
    $44 = $$pre19;
   }
   $42 = (($i$212) + 1)|0;
   $43 = ($42|0)<($44|0);
   if ($43) {
    $82 = $44;$i$212 = $42;
   } else {
    break;
   }
  }
 }
 $30 = ((($1)) + 20|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = ($31|0)>(0);
 if ($32) {
  $83 = $31;$i$310 = 0;
  while(1) {
   $49 = (((($1)) + 1568|0) + ($i$310<<2)|0);
   $50 = HEAP32[$49>>2]|0;
   $51 = ($50|0)==(0|0);
   if ($51) {
    $60 = $83;
   } else {
    $52 = (((($1)) + 1312|0) + ($i$310<<2)|0);
    $53 = HEAP32[$52>>2]|0;
    $54 = (101264 + ($53<<2)|0);
    $55 = HEAP32[$54>>2]|0;
    $56 = ((($55)) + 12|0);
    $57 = HEAP32[$56>>2]|0;
    FUNCTION_TABLE_vi[$57 & 8191]($50);
    $$pre20 = HEAP32[$30>>2]|0;
    $60 = $$pre20;
   }
   $58 = (($i$310) + 1)|0;
   $59 = ($58|0)<($60|0);
   if ($59) {
    $83 = $60;$i$310 = $58;
   } else {
    break;
   }
  }
 }
 $45 = ((($1)) + 24|0);
 $46 = HEAP32[$45>>2]|0;
 $47 = ($46|0)>(0);
 $48 = ((($1)) + 2848|0);
 if ($47) {
  $i$47 = 0;
  while(1) {
   $61 = (((($1)) + 1824|0) + ($i$47<<2)|0);
   $62 = HEAP32[$61>>2]|0;
   $63 = ($62|0)==(0|0);
   if (!($63)) {
    _vorbis_staticbook_destroy($62);
   }
   $64 = HEAP32[$48>>2]|0;
   $65 = ($64|0)==(0|0);
   if (!($65)) {
    $66 = (($64) + (($i$47*56)|0)|0);
    _vorbis_book_clear($66);
   }
   $67 = (($i$47) + 1)|0;
   $68 = HEAP32[$45>>2]|0;
   $69 = ($67|0)<($68|0);
   if ($69) {
    $i$47 = $67;
   } else {
    break;
   }
  }
 }
 $70 = HEAP32[$48>>2]|0;
 $71 = ($70|0)==(0|0);
 if (!($71)) {
  _free($70);
 }
 $72 = ((($1)) + 28|0);
 $73 = HEAP32[$72>>2]|0;
 $74 = ($73|0)>(0);
 if ($74) {
  $i$56 = 0;
  while(1) {
   $75 = (((($1)) + 2852|0) + ($i$56<<2)|0);
   $76 = HEAP32[$75>>2]|0;
   __vi_psy_free($76);
   $77 = (($i$56) + 1)|0;
   $78 = HEAP32[$72>>2]|0;
   $79 = ($77|0)<($78|0);
   if ($79) {
    $i$56 = $77;
   } else {
    break;
   }
  }
 }
 _free($1);
 ;HEAP32[$vi>>2]=0|0;HEAP32[$vi+4>>2]=0|0;HEAP32[$vi+8>>2]=0|0;HEAP32[$vi+12>>2]=0|0;HEAP32[$vi+16>>2]=0|0;HEAP32[$vi+20>>2]=0|0;HEAP32[$vi+24>>2]=0|0;HEAP32[$vi+28>>2]=0|0;
 return;
}
function _vorbis_book_clear($b) {
 $b = $b|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($b)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  _free($1);
 }
 $3 = ((($b)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if (!($5)) {
  _free($4);
 }
 $6 = ((($b)) + 24|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 if (!($8)) {
  _free($7);
 }
 $9 = ((($b)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10|0)==(0|0);
 if (!($11)) {
  _free($10);
 }
 $12 = ((($b)) + 32|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13|0)==(0|0);
 if (!($14)) {
  _free($13);
 }
 dest=$b; stop=dest+56|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 return;
}
function __vi_psy_free($i) {
 $i = $i|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($i|0)==(0|0);
 if (!($0)) {
  _free($i);
 }
 return;
}
function _vorbis_synthesis_idheader($op) {
 $op = $op|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buffer = 0, $opb = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $opb = sp;
 $buffer = sp + 20|0;
 $0 = ($op|0)==(0|0);
 if ($0) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = HEAP32[$op>>2]|0;
 $2 = ((($op)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 _oggpack_readinit($opb,$1,$3);
 $4 = ((($op)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0);
 if ($6) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $7 = (_oggpack_read($opb,8)|0);
 $8 = ($7|0)==(1);
 if (!($8)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 ;HEAP8[$buffer>>0]=0|0;HEAP8[$buffer+1>>0]=0|0;HEAP8[$buffer+2>>0]=0|0;HEAP8[$buffer+3>>0]=0|0;HEAP8[$buffer+4>>0]=0|0;
 $9 = (_oggpack_read($opb,8)|0);
 $10 = $9&255;
 $11 = ((($buffer)) + 1|0);
 HEAP8[$buffer>>0] = $10;
 $12 = (_oggpack_read($opb,8)|0);
 $13 = $12&255;
 $14 = ((($buffer)) + 2|0);
 HEAP8[$11>>0] = $13;
 $15 = (_oggpack_read($opb,8)|0);
 $16 = $15&255;
 $17 = ((($buffer)) + 3|0);
 HEAP8[$14>>0] = $16;
 $18 = (_oggpack_read($opb,8)|0);
 $19 = $18&255;
 $20 = ((($buffer)) + 4|0);
 HEAP8[$17>>0] = $19;
 $21 = (_oggpack_read($opb,8)|0);
 $22 = $21&255;
 $23 = ((($buffer)) + 5|0);
 HEAP8[$20>>0] = $22;
 $24 = (_oggpack_read($opb,8)|0);
 $25 = $24&255;
 HEAP8[$23>>0] = $25;
 $26 = (_memcmp($buffer,595830,6)|0);
 $27 = ($26|0)==(0);
 $$ = $27&1;
 $$0 = $$;
 STACKTOP = sp;return ($$0|0);
}
function _vorbis_synthesis_headerin($vi,$vc,$op) {
 $vi = $vi|0;
 $vc = $vc|0;
 $op = $op|0;
 var $$0 = 0, $$012$i$i = 0, $$012$i3$i = 0, $$03$i$i = 0, $$03$i2$i = 0, $$1 = 0, $$pre$i$i = 0, $$pre$phi$iZ2D = 0, $$pre2$i$i = 0, $$pre3$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0;
 var $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0;
 var $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0;
 var $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $buffer = 0, $i$01$i$i = 0, $i$031$i = 0;
 var $i$08$i = 0, $i$129$i = 0, $i$226$i = 0, $i$323$i = 0, $i$420$i = 0, $i$519$i = 0, $opb = 0, $or$cond$i = 0, $or$cond1$i = 0, $or$cond5$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $opb = sp;
 $buffer = sp + 20|0;
 $0 = ($op|0)==(0|0);
 if ($0) {
  $$1 = -133;
  STACKTOP = sp;return ($$1|0);
 }
 $1 = HEAP32[$op>>2]|0;
 $2 = ((($op)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 _oggpack_readinit($opb,$1,$3);
 $4 = (_oggpack_read($opb,8)|0);
 ;HEAP8[$buffer>>0]=0|0;HEAP8[$buffer+1>>0]=0|0;HEAP8[$buffer+2>>0]=0|0;HEAP8[$buffer+3>>0]=0|0;HEAP8[$buffer+4>>0]=0|0;
 $5 = (_oggpack_read($opb,8)|0);
 $6 = $5&255;
 $7 = ((($buffer)) + 1|0);
 HEAP8[$buffer>>0] = $6;
 $8 = (_oggpack_read($opb,8)|0);
 $9 = $8&255;
 $10 = ((($buffer)) + 2|0);
 HEAP8[$7>>0] = $9;
 $11 = (_oggpack_read($opb,8)|0);
 $12 = $11&255;
 $13 = ((($buffer)) + 3|0);
 HEAP8[$10>>0] = $12;
 $14 = (_oggpack_read($opb,8)|0);
 $15 = $14&255;
 $16 = ((($buffer)) + 4|0);
 HEAP8[$13>>0] = $15;
 $17 = (_oggpack_read($opb,8)|0);
 $18 = $17&255;
 $19 = ((($buffer)) + 5|0);
 HEAP8[$16>>0] = $18;
 $20 = (_oggpack_read($opb,8)|0);
 $21 = $20&255;
 HEAP8[$19>>0] = $21;
 $22 = (_memcmp($buffer,595830,6)|0);
 $23 = ($22|0)==(0);
 L4: do {
  if ($23) {
   switch ($4|0) {
   case 1:  {
    $24 = ((($op)) + 8|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = ($25|0)==(0);
    if ($26) {
     $$0 = -133;
     break L4;
    }
    $27 = ((($vi)) + 8|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = ($28|0)==(0);
    if (!($29)) {
     $$0 = -133;
     break L4;
    }
    $30 = ((($vi)) + 28|0);
    $31 = HEAP32[$30>>2]|0;
    $32 = ($31|0)==(0|0);
    if ($32) {
     $$0 = -129;
     break L4;
    }
    $33 = (_oggpack_read($opb,32)|0);
    HEAP32[$vi>>2] = $33;
    $34 = ($33|0)==(0);
    if (!($34)) {
     $$0 = -134;
     break L4;
    }
    $35 = (_oggpack_read($opb,8)|0);
    $36 = ((($vi)) + 4|0);
    HEAP32[$36>>2] = $35;
    $37 = (_oggpack_read($opb,32)|0);
    HEAP32[$27>>2] = $37;
    $38 = (_oggpack_read($opb,32)|0);
    $39 = ((($vi)) + 12|0);
    HEAP32[$39>>2] = $38;
    $40 = (_oggpack_read($opb,32)|0);
    $41 = ((($vi)) + 16|0);
    HEAP32[$41>>2] = $40;
    $42 = (_oggpack_read($opb,32)|0);
    $43 = ((($vi)) + 20|0);
    HEAP32[$43>>2] = $42;
    $44 = (_oggpack_read($opb,4)|0);
    $45 = 1 << $44;
    HEAP32[$31>>2] = $45;
    $46 = (_oggpack_read($opb,4)|0);
    $47 = 1 << $46;
    $48 = ((($31)) + 4|0);
    HEAP32[$48>>2] = $47;
    $49 = HEAP32[$27>>2]|0;
    $50 = ($49|0)<(1);
    if (!($50)) {
     $51 = HEAP32[$36>>2]|0;
     $52 = ($51|0)<(1);
     if (!($52)) {
      $53 = HEAP32[$31>>2]|0;
      $54 = ($53|0)<(64);
      $55 = ($47|0)<($53|0);
      $or$cond$i = $54 | $55;
      $56 = ($47|0)>(8192);
      $or$cond1$i = $56 | $or$cond$i;
      if (!($or$cond1$i)) {
       $57 = (_oggpack_read($opb,1)|0);
       $58 = ($57|0)==(1);
       if ($58) {
        $$0 = 0;
        break L4;
       }
      }
     }
    }
    _vorbis_info_clear($vi);
    $$0 = -133;
    break L4;
    break;
   }
   case 3:  {
    $59 = ((($vi)) + 8|0);
    $60 = HEAP32[$59>>2]|0;
    $61 = ($60|0)==(0);
    if ($61) {
     $$0 = -133;
     break L4;
    }
    $62 = ((($vc)) + 12|0);
    $63 = HEAP32[$62>>2]|0;
    $64 = ($63|0)==(0|0);
    if (!($64)) {
     $$0 = -133;
     break L4;
    }
    $65 = (_oggpack_read($opb,32)|0);
    $66 = ($65|0)<(0);
    L20: do {
     if ($66) {
      label = 30;
     } else {
      $67 = ((($opb)) + 16|0);
      $68 = HEAP32[$67>>2]|0;
      $69 = (($68) + -8)|0;
      $70 = ($65|0)>($69|0);
      if ($70) {
       label = 30;
      } else {
       $71 = (($65) + 1)|0;
       $72 = (_calloc($71,1)|0);
       HEAP32[$62>>2] = $72;
       $73 = ($65|0)==(0);
       if (!($73)) {
        $$012$i$i = $72;$$03$i$i = $65;
        while(1) {
         $74 = (($$03$i$i) + -1)|0;
         $75 = (_oggpack_read($opb,8)|0);
         $76 = $75&255;
         $77 = ((($$012$i$i)) + 1|0);
         HEAP8[$$012$i$i>>0] = $76;
         $78 = ($74|0)==(0);
         if ($78) {
          break;
         } else {
          $$012$i$i = $77;$$03$i$i = $74;
         }
        }
       }
       $79 = (_oggpack_read($opb,32)|0);
       $80 = ($79|0)<(0);
       if (!($80)) {
        $81 = HEAP32[$67>>2]|0;
        $82 = (_oggpack_bytes($opb)|0);
        $83 = (($81) - ($82))|0;
        $84 = $83 >> 2;
        $85 = ($79|0)>($84|0);
        if (!($85)) {
         $86 = ((($vc)) + 8|0);
         HEAP32[$86>>2] = $79;
         $87 = (($79) + 1)|0;
         $88 = (_calloc($87,4)|0);
         HEAP32[$vc>>2] = $88;
         $89 = (_calloc($87,4)|0);
         $90 = ((($vc)) + 4|0);
         HEAP32[$90>>2] = $89;
         $91 = ($79|0)>(0);
         if ($91) {
          $i$08$i = 0;
          while(1) {
           $92 = (_oggpack_read($opb,32)|0);
           $93 = ($92|0)<(0);
           if ($93) {
            $$pre$phi$iZ2D = $vc;
            break L20;
           }
           $94 = HEAP32[$67>>2]|0;
           $95 = (_oggpack_bytes($opb)|0);
           $96 = (($94) - ($95))|0;
           $97 = ($92|0)>($96|0);
           if ($97) {
            label = 30;
            break L20;
           }
           $98 = HEAP32[$90>>2]|0;
           $99 = (($98) + ($i$08$i<<2)|0);
           HEAP32[$99>>2] = $92;
           $100 = (($92) + 1)|0;
           $101 = (_calloc($100,1)|0);
           $102 = HEAP32[$vc>>2]|0;
           $103 = (($102) + ($i$08$i<<2)|0);
           HEAP32[$103>>2] = $101;
           $104 = ($92|0)==(0);
           if (!($104)) {
            $105 = HEAP32[$vc>>2]|0;
            $106 = (($105) + ($i$08$i<<2)|0);
            $107 = HEAP32[$106>>2]|0;
            $$012$i3$i = $107;$$03$i2$i = $92;
            while(1) {
             $108 = (($$03$i2$i) + -1)|0;
             $109 = (_oggpack_read($opb,8)|0);
             $110 = $109&255;
             $111 = ((($$012$i3$i)) + 1|0);
             HEAP8[$$012$i3$i>>0] = $110;
             $112 = ($108|0)==(0);
             if ($112) {
              break;
             } else {
              $$012$i3$i = $111;$$03$i2$i = $108;
             }
            }
           }
           $113 = (($i$08$i) + 1)|0;
           $114 = HEAP32[$86>>2]|0;
           $115 = ($113|0)<($114|0);
           if ($115) {
            $i$08$i = $113;
           } else {
            break;
           }
          }
         }
         $116 = (_oggpack_read($opb,1)|0);
         $117 = ($116|0)==(1);
         if ($117) {
          $$0 = 0;
          break L4;
         } else {
          $$pre$phi$iZ2D = $vc;
          break;
         }
        }
       }
       $118 = ($vc|0)==(0|0);
       if ($118) {
        $$0 = -133;
        break L4;
       } else {
        label = 30;
       }
      }
     }
    } while(0);
    if ((label|0) == 30) {
     $$pre$phi$iZ2D = $vc;
    }
    $119 = HEAP32[$$pre$phi$iZ2D>>2]|0;
    $120 = ($119|0)==(0|0);
    if (!($120)) {
     $121 = ((($vc)) + 8|0);
     $122 = HEAP32[$121>>2]|0;
     $123 = ($122|0)>(0);
     if ($123) {
      $125 = $119;$246 = $122;$i$01$i$i = 0;
      while(1) {
       $124 = (($125) + ($i$01$i$i<<2)|0);
       $126 = HEAP32[$124>>2]|0;
       $127 = ($126|0)==(0|0);
       if ($127) {
        $130 = $246;
       } else {
        _free($126);
        $$pre2$i$i = HEAP32[$121>>2]|0;
        $130 = $$pre2$i$i;
       }
       $128 = (($i$01$i$i) + 1)|0;
       $129 = ($128|0)<($130|0);
       if (!($129)) {
        break;
       }
       $$pre$i$i = HEAP32[$$pre$phi$iZ2D>>2]|0;
       $125 = $$pre$i$i;$246 = $130;$i$01$i$i = $128;
      }
      $$pre3$i$i = HEAP32[$vc>>2]|0;
      $131 = $$pre3$i$i;
     } else {
      $131 = $119;
     }
     _free($131);
    }
    $132 = ((($vc)) + 4|0);
    $133 = HEAP32[$132>>2]|0;
    $134 = ($133|0)==(0|0);
    if (!($134)) {
     _free($133);
    }
    $135 = HEAP32[$62>>2]|0;
    $136 = ($135|0)==(0|0);
    if (!($136)) {
     _free($135);
    }
    ;HEAP32[$vc>>2]=0|0;HEAP32[$vc+4>>2]=0|0;HEAP32[$vc+8>>2]=0|0;HEAP32[$vc+12>>2]=0|0;
    $$0 = -133;
    break L4;
    break;
   }
   case 5:  {
    $137 = ((($vi)) + 8|0);
    $138 = HEAP32[$137>>2]|0;
    $139 = ($138|0)==(0);
    if ($139) {
     $$0 = -133;
     break L4;
    }
    $140 = ((($vc)) + 12|0);
    $141 = HEAP32[$140>>2]|0;
    $142 = ($141|0)==(0|0);
    if ($142) {
     $$0 = -133;
     break L4;
    }
    $143 = ((($vi)) + 28|0);
    $144 = HEAP32[$143>>2]|0;
    $145 = ($144|0)==(0|0);
    if ($145) {
     $$0 = -129;
     break L4;
    }
    $146 = ((($144)) + 24|0);
    $147 = HEAP32[$146>>2]|0;
    $148 = ($147|0)>(0);
    if ($148) {
     $$0 = -133;
     break L4;
    }
    $149 = (_oggpack_read($opb,8)|0);
    $150 = (($149) + 1)|0;
    $151 = ((($144)) + 24|0);
    HEAP32[$151>>2] = $150;
    $152 = ($149|0)<(0);
    L68: do {
     if (!($152)) {
      $i$031$i = 0;
      while(1) {
       $156 = (_vorbis_staticbook_unpack($opb)|0);
       $157 = (((($144)) + 1824|0) + ($i$031$i<<2)|0);
       HEAP32[$157>>2] = $156;
       $158 = ($156|0)==(0|0);
       $155 = (($i$031$i) + 1)|0;
       if ($158) {
        break L68;
       }
       $153 = HEAP32[$151>>2]|0;
       $154 = ($155|0)<($153|0);
       if ($154) {
        $i$031$i = $155;
       } else {
        break;
       }
      }
      $159 = (_oggpack_read($opb,6)|0);
      $160 = ($159|0)<(0);
      if (!($160)) {
       $i$129$i = 0;
       while(1) {
        $163 = (_oggpack_read($opb,16)|0);
        $164 = ($163|0)==(0);
        if (!($164)) {
         break L68;
        }
        $161 = (($i$129$i) + 1)|0;
        $162 = ($i$129$i|0)<($159|0);
        if ($162) {
         $i$129$i = $161;
        } else {
         break;
        }
       }
       $165 = (_oggpack_read($opb,6)|0);
       $166 = (($165) + 1)|0;
       $167 = ((($144)) + 16|0);
       HEAP32[$167>>2] = $166;
       $168 = ($165|0)<(0);
       if (!($168)) {
        $i$226$i = 0;
        while(1) {
         $172 = (_oggpack_read($opb,16)|0);
         $173 = (((($144)) + 800|0) + ($i$226$i<<2)|0);
         HEAP32[$173>>2] = $172;
         $174 = ($172>>>0)>(1);
         if ($174) {
          break L68;
         }
         $175 = (101372 + ($172<<2)|0);
         $176 = HEAP32[$175>>2]|0;
         $177 = ((($176)) + 4|0);
         $178 = HEAP32[$177>>2]|0;
         $179 = (FUNCTION_TABLE_iii[$178 & 8191]($vi,$opb)|0);
         $180 = (((($144)) + 1056|0) + ($i$226$i<<2)|0);
         HEAP32[$180>>2] = $179;
         $181 = ($179|0)==(0|0);
         $171 = (($i$226$i) + 1)|0;
         if ($181) {
          break L68;
         }
         $169 = HEAP32[$167>>2]|0;
         $170 = ($171|0)<($169|0);
         if ($170) {
          $i$226$i = $171;
         } else {
          break;
         }
        }
        $182 = (_oggpack_read($opb,6)|0);
        $183 = (($182) + 1)|0;
        $184 = ((($144)) + 20|0);
        HEAP32[$184>>2] = $183;
        $185 = ($182|0)<(0);
        if (!($185)) {
         $i$323$i = 0;
         while(1) {
          $189 = (_oggpack_read($opb,16)|0);
          $190 = (((($144)) + 1312|0) + ($i$323$i<<2)|0);
          HEAP32[$190>>2] = $189;
          $191 = ($189>>>0)>(2);
          if ($191) {
           break L68;
          }
          $192 = (101264 + ($189<<2)|0);
          $193 = HEAP32[$192>>2]|0;
          $194 = ((($193)) + 4|0);
          $195 = HEAP32[$194>>2]|0;
          $196 = (FUNCTION_TABLE_iii[$195 & 8191]($vi,$opb)|0);
          $197 = (((($144)) + 1568|0) + ($i$323$i<<2)|0);
          HEAP32[$197>>2] = $196;
          $198 = ($196|0)==(0|0);
          $188 = (($i$323$i) + 1)|0;
          if ($198) {
           break L68;
          }
          $186 = HEAP32[$184>>2]|0;
          $187 = ($188|0)<($186|0);
          if ($187) {
           $i$323$i = $188;
          } else {
           break;
          }
         }
         $199 = (_oggpack_read($opb,6)|0);
         $200 = (($199) + 1)|0;
         $201 = ((($144)) + 12|0);
         HEAP32[$201>>2] = $200;
         $202 = ($199|0)<(0);
         if (!($202)) {
          $i$420$i = 0;
          while(1) {
           $206 = (_oggpack_read($opb,16)|0);
           $207 = (((($144)) + 288|0) + ($i$420$i<<2)|0);
           HEAP32[$207>>2] = $206;
           $208 = ($206|0)==(0);
           if (!($208)) {
            break L68;
           }
           $209 = HEAP32[25345]|0;
           $210 = ((($209)) + 4|0);
           $211 = HEAP32[$210>>2]|0;
           $212 = (FUNCTION_TABLE_iii[$211 & 8191]($vi,$opb)|0);
           $213 = (((($144)) + 544|0) + ($i$420$i<<2)|0);
           HEAP32[$213>>2] = $212;
           $214 = ($212|0)==(0|0);
           $205 = (($i$420$i) + 1)|0;
           if ($214) {
            break L68;
           }
           $203 = HEAP32[$201>>2]|0;
           $204 = ($205|0)<($203|0);
           if ($204) {
            $i$420$i = $205;
           } else {
            break;
           }
          }
          $215 = (_oggpack_read($opb,6)|0);
          $216 = (($215) + 1)|0;
          $217 = ((($144)) + 8|0);
          HEAP32[$217>>2] = $216;
          $218 = ($215|0)<(0);
          if (!($218)) {
           $i$519$i = 0;
           while(1) {
            $222 = (_calloc(1,16)|0);
            $223 = (((($144)) + 32|0) + ($i$519$i<<2)|0);
            HEAP32[$223>>2] = $222;
            $224 = (_oggpack_read($opb,1)|0);
            $225 = HEAP32[$223>>2]|0;
            HEAP32[$225>>2] = $224;
            $226 = (_oggpack_read($opb,16)|0);
            $227 = HEAP32[$223>>2]|0;
            $228 = ((($227)) + 4|0);
            HEAP32[$228>>2] = $226;
            $229 = (_oggpack_read($opb,16)|0);
            $230 = HEAP32[$223>>2]|0;
            $231 = ((($230)) + 8|0);
            HEAP32[$231>>2] = $229;
            $232 = (_oggpack_read($opb,8)|0);
            $233 = HEAP32[$223>>2]|0;
            $234 = ((($233)) + 12|0);
            HEAP32[$234>>2] = $232;
            $235 = ((($233)) + 4|0);
            $236 = HEAP32[$235>>2]|0;
            $237 = ($236|0)>(0);
            if ($237) {
             break L68;
            }
            $238 = ((($233)) + 8|0);
            $239 = HEAP32[$238>>2]|0;
            $240 = ($239|0)>(0);
            if ($240) {
             break L68;
            }
            $241 = HEAP32[$201>>2]|0;
            $242 = ($232|0)>=($241|0);
            $243 = ($232|0)<(0);
            $or$cond5$i = $243 | $242;
            $221 = (($i$519$i) + 1)|0;
            if ($or$cond5$i) {
             break L68;
            }
            $219 = HEAP32[$217>>2]|0;
            $220 = ($221|0)<($219|0);
            if ($220) {
             $i$519$i = $221;
            } else {
             break;
            }
           }
           $244 = (_oggpack_read($opb,1)|0);
           $245 = ($244|0)==(1);
           if ($245) {
            $$0 = 0;
            break L4;
           }
          }
         }
        }
       }
      }
     }
    } while(0);
    _vorbis_info_clear($vi);
    $$0 = -133;
    break L4;
    break;
   }
   default: {
    $$0 = -133;
    break L4;
   }
   }
  } else {
   $$0 = -132;
  }
 } while(0);
 $$1 = $$0;
 STACKTOP = sp;return ($$1|0);
}
function _vorbis_synthesis($vb,$op) {
 $vb = $vb|0;
 $op = $op|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $i$012 = 0, $or$cond = 0, $or$cond3 = 0, $or$cond5 = 0, $or$cond7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($vb|0)!=(0|0);
 if ($0) {
  $1 = ((($vb)) + 64|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ($2|0)==(0|0);
  if ($3) {
   $14 = 0;$15 = 0;$16 = 0;$18 = 0;$74 = 0;
  } else {
   $4 = ((($2)) + 104|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = ((($2)) + 4|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0|0);
   if ($8) {
    $14 = $5;$15 = 1;$16 = 0;$18 = 0;$74 = 0;
   } else {
    $9 = ((($7)) + 28|0);
    $10 = HEAP32[$9>>2]|0;
    $14 = $5;$15 = 1;$16 = 1;$18 = $10;$74 = $7;
   }
  }
 } else {
  $14 = 0;$15 = 0;$16 = 0;$18 = 0;$74 = 0;
 }
 $11 = ((($vb)) + 4|0);
 $12 = $0 ? $11 : 0;
 $13 = ($14|0)!=(0|0);
 $or$cond = $15 & $13;
 $or$cond3 = $16 & $or$cond;
 $17 = ($18|0)!=(0|0);
 $or$cond5 = $17 & $or$cond3;
 $or$cond7 = $0 & $or$cond5;
 if (!($or$cond7)) {
  $$0 = -136;
  return ($$0|0);
 }
 __vorbis_block_ripcord($vb);
 $19 = HEAP32[$op>>2]|0;
 $20 = ((($op)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 _oggpack_readinit($12,$19,$21);
 $22 = (_oggpack_read($12,1)|0);
 $23 = ($22|0)==(0);
 if (!($23)) {
  $$0 = -135;
  return ($$0|0);
 }
 $24 = ((($14)) + 44|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = (_oggpack_read($12,$25)|0);
 $27 = ($26|0)==(-1);
 if ($27) {
  $$0 = -136;
  return ($$0|0);
 }
 $28 = ((($vb)) + 40|0);
 HEAP32[$28>>2] = $26;
 $29 = (((($18)) + 32|0) + ($26<<2)|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ($30|0)==(0|0);
 if ($31) {
  $$0 = -136;
  return ($$0|0);
 }
 $32 = HEAP32[$30>>2]|0;
 $33 = ((($vb)) + 28|0);
 HEAP32[$33>>2] = $32;
 $34 = ($32|0)==(0);
 do {
  if ($34) {
   $40 = ((($vb)) + 24|0);
   HEAP32[$40>>2] = 0;
   $41 = ((($vb)) + 32|0);
   HEAP32[$41>>2] = 0;
   $70 = 0;
  } else {
   $35 = (_oggpack_read($12,1)|0);
   $36 = ((($vb)) + 24|0);
   HEAP32[$36>>2] = $35;
   $37 = (_oggpack_read($12,1)|0);
   $38 = ((($vb)) + 32|0);
   HEAP32[$38>>2] = $37;
   $39 = ($37|0)==(-1);
   if ($39) {
    $$0 = -136;
    return ($$0|0);
   } else {
    $$pre = HEAP32[$33>>2]|0;
    $70 = $$pre;
    break;
   }
  }
 } while(0);
 $42 = ((($op)) + 16|0);
 $43 = $42;
 $44 = $43;
 $45 = HEAP32[$44>>2]|0;
 $46 = (($43) + 4)|0;
 $47 = $46;
 $48 = HEAP32[$47>>2]|0;
 $49 = ((($vb)) + 48|0);
 $50 = $49;
 $51 = $50;
 HEAP32[$51>>2] = $45;
 $52 = (($50) + 4)|0;
 $53 = $52;
 HEAP32[$53>>2] = $48;
 $54 = ((($op)) + 24|0);
 $55 = $54;
 $56 = $55;
 $57 = HEAP32[$56>>2]|0;
 $58 = (($55) + 4)|0;
 $59 = $58;
 $60 = HEAP32[$59>>2]|0;
 $61 = ((($vb)) + 56|0);
 $62 = $61;
 $63 = $62;
 HEAP32[$63>>2] = $57;
 $64 = (($62) + 4)|0;
 $65 = $64;
 HEAP32[$65>>2] = $60;
 $66 = ((($op)) + 12|0);
 $67 = HEAP32[$66>>2]|0;
 $68 = ((($vb)) + 44|0);
 HEAP32[$68>>2] = $67;
 $69 = (($18) + ($70<<2)|0);
 $71 = HEAP32[$69>>2]|0;
 $72 = ((($vb)) + 36|0);
 HEAP32[$72>>2] = $71;
 $73 = ((($74)) + 4|0);
 $75 = HEAP32[$73>>2]|0;
 $76 = $75 << 2;
 $77 = (__vorbis_block_alloc($vb,$76)|0);
 HEAP32[$vb>>2] = $77;
 $78 = HEAP32[$73>>2]|0;
 $79 = ($78|0)>(0);
 if ($79) {
  $i$012 = 0;
  while(1) {
   $80 = HEAP32[$72>>2]|0;
   $81 = $80 << 2;
   $82 = (__vorbis_block_alloc($vb,$81)|0);
   $83 = HEAP32[$vb>>2]|0;
   $84 = (($83) + ($i$012<<2)|0);
   HEAP32[$84>>2] = $82;
   $85 = (($i$012) + 1)|0;
   $86 = HEAP32[$73>>2]|0;
   $87 = ($85|0)<($86|0);
   if ($87) {
    $i$012 = $85;
   } else {
    break;
   }
  }
 }
 $88 = HEAP32[$29>>2]|0;
 $89 = ((($88)) + 12|0);
 $90 = HEAP32[$89>>2]|0;
 $91 = (((($18)) + 288|0) + ($90<<2)|0);
 $92 = HEAP32[$91>>2]|0;
 $93 = (101380 + ($92<<2)|0);
 $94 = HEAP32[$93>>2]|0;
 $95 = ((($94)) + 16|0);
 $96 = HEAP32[$95>>2]|0;
 $97 = (((($18)) + 544|0) + ($90<<2)|0);
 $98 = HEAP32[$97>>2]|0;
 $99 = (FUNCTION_TABLE_iii[$96 & 8191]($vb,$98)|0);
 $$0 = $99;
 return ($$0|0);
}
function _vorbis_synthesis_trackonly($vb,$op) {
 $vb = $vb|0;
 $op = $op|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($vb)) + 64|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 104|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 28|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($vb)) + 4|0);
 __vorbis_block_ripcord($vb);
 $9 = HEAP32[$op>>2]|0;
 $10 = ((($op)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 _oggpack_readinit($8,$9,$11);
 $12 = (_oggpack_read($8,1)|0);
 $13 = ($12|0)==(0);
 if (!($13)) {
  $$0 = -135;
  return ($$0|0);
 }
 $14 = ((($3)) + 44|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = (_oggpack_read($8,$15)|0);
 $17 = ($16|0)==(-1);
 if ($17) {
  $$0 = -136;
  return ($$0|0);
 }
 $18 = ((($vb)) + 40|0);
 HEAP32[$18>>2] = $16;
 $19 = (((($7)) + 32|0) + ($16<<2)|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ($20|0)==(0|0);
 if ($21) {
  $$0 = -136;
  return ($$0|0);
 }
 $22 = HEAP32[$20>>2]|0;
 $23 = ((($vb)) + 28|0);
 HEAP32[$23>>2] = $22;
 $24 = ($22|0)==(0);
 if ($24) {
  $30 = ((($vb)) + 24|0);
  HEAP32[$30>>2] = 0;
  $31 = ((($vb)) + 32|0);
  HEAP32[$31>>2] = 0;
 } else {
  $25 = (_oggpack_read($8,1)|0);
  $26 = ((($vb)) + 24|0);
  HEAP32[$26>>2] = $25;
  $27 = (_oggpack_read($8,1)|0);
  $28 = ((($vb)) + 32|0);
  HEAP32[$28>>2] = $27;
  $29 = ($27|0)==(-1);
  if ($29) {
   $$0 = -136;
   return ($$0|0);
  }
 }
 $32 = ((($op)) + 16|0);
 $33 = $32;
 $34 = $33;
 $35 = HEAP32[$34>>2]|0;
 $36 = (($33) + 4)|0;
 $37 = $36;
 $38 = HEAP32[$37>>2]|0;
 $39 = ((($vb)) + 48|0);
 $40 = $39;
 $41 = $40;
 HEAP32[$41>>2] = $35;
 $42 = (($40) + 4)|0;
 $43 = $42;
 HEAP32[$43>>2] = $38;
 $44 = ((($op)) + 24|0);
 $45 = $44;
 $46 = $45;
 $47 = HEAP32[$46>>2]|0;
 $48 = (($45) + 4)|0;
 $49 = $48;
 $50 = HEAP32[$49>>2]|0;
 $51 = ((($vb)) + 56|0);
 $52 = $51;
 $53 = $52;
 HEAP32[$53>>2] = $47;
 $54 = (($52) + 4)|0;
 $55 = $54;
 HEAP32[$55>>2] = $50;
 $56 = ((($op)) + 12|0);
 $57 = HEAP32[$56>>2]|0;
 $58 = ((($vb)) + 44|0);
 HEAP32[$58>>2] = $57;
 $59 = ((($vb)) + 36|0);
 HEAP32[$59>>2] = 0;
 HEAP32[$vb>>2] = 0;
 $$0 = 0;
 return ($$0|0);
}
function _vorbis_packet_blocksize($vi,$op) {
 $vi = $vi|0;
 $op = $op|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $opb = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $opb = sp;
 $0 = ((($vi)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = -129;
 } else {
  $3 = ((($1)) + 8|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ($4|0)<(1);
  if ($5) {
   $$0 = -129;
  } else {
   $6 = HEAP32[$op>>2]|0;
   $7 = ((($op)) + 4|0);
   $8 = HEAP32[$7>>2]|0;
   _oggpack_readinit($opb,$6,$8);
   $9 = (_oggpack_read($opb,1)|0);
   $10 = ($9|0)==(0);
   if ($10) {
    $11 = HEAP32[$3>>2]|0;
    $12 = (($11) + -1)|0;
    $13 = (_ov_ilog($12)|0);
    $14 = (_oggpack_read($opb,$13)|0);
    $15 = ($14|0)==(-1);
    if ($15) {
     $$0 = -136;
    } else {
     $16 = (((($1)) + 32|0) + ($14<<2)|0);
     $17 = HEAP32[$16>>2]|0;
     $18 = ($17|0)==(0|0);
     if ($18) {
      $$0 = -136;
     } else {
      $19 = HEAP32[$17>>2]|0;
      $20 = (($1) + ($19<<2)|0);
      $21 = HEAP32[$20>>2]|0;
      $$0 = $21;
     }
    }
   } else {
    $$0 = -135;
   }
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _vorbis_synthesis_halfrate_p($vi) {
 $vi = $vi|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($vi)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 3656|0);
 $3 = HEAP32[$2>>2]|0;
 return ($3|0);
}
function _ov_clear($vf) {
 $vf = $vf|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $i$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($vf|0)==(0|0);
 if ($0) {
  return 0;
 }
 $1 = ((($vf)) + 592|0);
 (_vorbis_block_clear($1)|0);
 $2 = ((($vf)) + 480|0);
 _vorbis_dsp_clear($2);
 $3 = ((($vf)) + 120|0);
 (_ogg_stream_clear($3)|0);
 $4 = ((($vf)) + 72|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if (!($6)) {
  $7 = ((($vf)) + 52|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8|0)==(0);
  if (!($9)) {
   $10 = ($8|0)>(0);
   $11 = ((($vf)) + 76|0);
   if ($10) {
    $13 = $5;$i$01 = 0;
    while(1) {
     $12 = (($13) + ($i$01<<5)|0);
     _vorbis_info_clear($12);
     $14 = HEAP32[$11>>2]|0;
     $15 = (($14) + ($i$01<<4)|0);
     _vorbis_comment_clear($15);
     $16 = (($i$01) + 1)|0;
     $17 = HEAP32[$7>>2]|0;
     $18 = ($16|0)<($17|0);
     $19 = HEAP32[$4>>2]|0;
     if ($18) {
      $13 = $19;$i$01 = $16;
     } else {
      $$lcssa = $19;
      break;
     }
    }
   } else {
    $$lcssa = $5;
   }
   _free($$lcssa);
   $20 = HEAP32[$11>>2]|0;
   _free($20);
  }
 }
 $21 = ((($vf)) + 60|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ($22|0)==(0|0);
 if (!($23)) {
  _free($22);
 }
 $24 = ((($vf)) + 68|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = ($25|0)==(0|0);
 if (!($26)) {
  _free($25);
 }
 $27 = ((($vf)) + 64|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ($28|0)==(0|0);
 if (!($29)) {
  _free($28);
 }
 $30 = ((($vf)) + 56|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = ($31|0)==(0|0);
 if (!($32)) {
  _free($31);
 }
 $33 = ((($vf)) + 24|0);
 (_ogg_sync_clear($33)|0);
 $34 = HEAP32[$vf>>2]|0;
 $35 = ($34|0)==(0|0);
 if (!($35)) {
  $36 = ((($vf)) + 712|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = ($37|0)==(0|0);
  if (!($38)) {
   (FUNCTION_TABLE_ii[$37 & 8191]($34)|0);
  }
 }
 _memset(($vf|0),0,720)|0;
 return 0;
}
function _ov_open_callbacks($f,$vf,$initial,$ibytes,$callbacks) {
 $f = $f|0;
 $vf = $vf|0;
 $initial = $initial|0;
 $ibytes = $ibytes|0;
 $callbacks = $callbacks|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $callbacks$byval_copy = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $callbacks$byval_copy = sp;
 ;HEAP32[$callbacks$byval_copy>>2]=HEAP32[$callbacks>>2]|0;HEAP32[$callbacks$byval_copy+4>>2]=HEAP32[$callbacks+4>>2]|0;HEAP32[$callbacks$byval_copy+8>>2]=HEAP32[$callbacks+8>>2]|0;HEAP32[$callbacks$byval_copy+12>>2]=HEAP32[$callbacks+12>>2]|0;
 $0 = (__ov_open1($f,$vf,$initial,$ibytes,$callbacks$byval_copy)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  $$0 = $0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = (__ov_open2($vf)|0);
 $$0 = $2;
 STACKTOP = sp;return ($$0|0);
}
function __ov_open1($f,$vf,$initial,$ibytes,$callbacks) {
 $f = $f|0;
 $vf = $vf|0;
 $initial = $initial|0;
 $ibytes = $ibytes|0;
 $callbacks = $callbacks|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $serialno_list = 0, $serialno_list_size = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $serialno_list = sp + 4|0;
 $serialno_list_size = sp;
 $0 = ($f|0)==(0|0);
 if ($0) {
  $10 = -1;
 } else {
  $1 = ((($callbacks)) + 4|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ($2|0)==(0|0);
  if ($3) {
   $10 = -1;
  } else {
   $4 = (FUNCTION_TABLE_iiiii[$2 & 8191]($f,0,0,1)|0);
   $10 = $4;
  }
 }
 HEAP32[$serialno_list>>2] = 0;
 HEAP32[$serialno_list_size>>2] = 0;
 _memset(($vf|0),0,704)|0;
 HEAP32[$vf>>2] = $f;
 $5 = ((($vf)) + 704|0);
 ;HEAP32[$5>>2]=HEAP32[$callbacks>>2]|0;HEAP32[$5+4>>2]=HEAP32[$callbacks+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$callbacks+8>>2]|0;HEAP32[$5+12>>2]=HEAP32[$callbacks+12>>2]|0;
 $6 = ((($vf)) + 24|0);
 (_ogg_sync_init($6)|0);
 $7 = ($initial|0)==(0|0);
 if (!($7)) {
  $8 = (_ogg_sync_buffer($6,$ibytes)|0);
  _memcpy(($8|0),($initial|0),($ibytes|0))|0;
  (_ogg_sync_wrote($6,$ibytes)|0);
 }
 $9 = ($10|0)==(-1);
 if (!($9)) {
  $11 = ((($vf)) + 4|0);
  HEAP32[$11>>2] = 1;
 }
 $12 = ((($vf)) + 52|0);
 HEAP32[$12>>2] = 1;
 $13 = (_calloc(1,32)|0);
 $14 = ((($vf)) + 72|0);
 HEAP32[$14>>2] = $13;
 $15 = (_calloc(1,16)|0);
 $16 = ((($vf)) + 76|0);
 HEAP32[$16>>2] = $15;
 $17 = ((($vf)) + 120|0);
 (_ogg_stream_init($17,-1)|0);
 $18 = HEAP32[$14>>2]|0;
 $19 = HEAP32[$16>>2]|0;
 $20 = (__fetch_headers($vf,$18,$19,$serialno_list,$serialno_list_size,0)|0);
 $21 = ($20|0)<(0);
 if ($21) {
  HEAP32[$vf>>2] = 0;
  (_ov_clear($vf)|0);
  $$pre = HEAP32[$serialno_list>>2]|0;
  $54 = $$pre;
 } else {
  $22 = HEAP32[$serialno_list_size>>2]|0;
  $23 = (($22) + 2)|0;
  $24 = (_calloc($23,4)|0);
  $25 = ((($vf)) + 64|0);
  HEAP32[$25>>2] = $24;
  $26 = ((($vf)) + 456|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = ((($vf)) + 92|0);
  HEAP32[$28>>2] = $27;
  HEAP32[$24>>2] = $27;
  $29 = ((($24)) + 4|0);
  HEAP32[$29>>2] = $22;
  $30 = ((($24)) + 8|0);
  $31 = HEAP32[$serialno_list>>2]|0;
  $32 = $22 << 2;
  _memcpy(($30|0),($31|0),($32|0))|0;
  $33 = (_calloc(1,8)|0);
  $34 = ((($vf)) + 56|0);
  HEAP32[$34>>2] = $33;
  $35 = (_calloc(1,8)|0);
  $36 = ((($vf)) + 60|0);
  HEAP32[$36>>2] = $35;
  $37 = $33;
  $38 = $37;
  HEAP32[$38>>2] = 0;
  $39 = (($37) + 4)|0;
  $40 = $39;
  HEAP32[$40>>2] = 0;
  $41 = ((($vf)) + 8|0);
  $42 = $41;
  $43 = $42;
  $44 = HEAP32[$43>>2]|0;
  $45 = (($42) + 4)|0;
  $46 = $45;
  $47 = HEAP32[$46>>2]|0;
  $48 = $35;
  $49 = $48;
  HEAP32[$49>>2] = $44;
  $50 = (($48) + 4)|0;
  $51 = $50;
  HEAP32[$51>>2] = $47;
  $52 = ((($vf)) + 88|0);
  HEAP32[$52>>2] = 1;
  $54 = $31;
 }
 $53 = ($54|0)==(0|0);
 if ($53) {
  STACKTOP = sp;return ($20|0);
 }
 _free($54);
 STACKTOP = sp;return ($20|0);
}
function __fetch_headers($vf,$vi,$vc,$serialno_list,$serialno_n,$og_ptr) {
 $vf = $vf|0;
 $vi = $vi|0;
 $vc = $vc|0;
 $serialno_list = $serialno_list|0;
 $serialno_n = $serialno_n|0;
 $og_ptr = $og_ptr|0;
 var $$013$i$i = 0, $$12 = 0, $$2 = 0, $$i$1 = 0, $$in$i = 0, $$in$i$i = 0, $$lcssa96 = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $allbos$030 = 0, $allbos$1$ph$us = 0, $allbos$1$ph$us$lcssa98 = 0, $i$031 = 0, $i$1 = 0, $i$1$lcssa94 = 0, $og = 0, $op = 0, $or$cond$i$i = 0, $ret$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $og = sp + 32|0;
 $op = sp;
 $0 = ($og_ptr|0)==(0|0);
 if ($0) {
  $1 = (__get_next_page($vf,$og,65536,0)|0);
  $2 = tempRet0;
  $3 = ($1|0)==(-128);
  $4 = ($2|0)==(-1);
  $5 = $3 & $4;
  if ($5) {
   $$2 = -128;
   STACKTOP = sp;return ($$2|0);
  }
  $6 = ($2|0)<(0);
  if ($6) {
   $$2 = -132;
   STACKTOP = sp;return ($$2|0);
  } else {
   $$12 = $og;
  }
 } else {
  $$12 = $og_ptr;
 }
 _vorbis_info_init($vi);
 _vorbis_comment_init($vc);
 $7 = ((($vf)) + 88|0);
 HEAP32[$7>>2] = 2;
 $8 = (_ogg_page_bos($$12)|0);
 $9 = ($8|0)==(0);
 L8: do {
  if ($9) {
   label = 34;
  } else {
   $10 = ($serialno_list|0)==(0|0);
   $11 = ((($vf)) + 120|0);
   $12 = ((($vf)) + 456|0);
   L10: do {
    if ($10) {
     while(1) {
      $13 = HEAP32[$7>>2]|0;
      $14 = ($13|0)<(3);
      if ($14) {
       $15 = (_ogg_page_serialno($$12)|0);
       (_ogg_stream_reset_serialno($11,$15)|0);
       (_ogg_stream_pagein($11,$$12)|0);
       $16 = (_ogg_stream_packetout($11,$op)|0);
       $17 = ($16|0)>(0);
       if ($17) {
        $18 = (_vorbis_synthesis_idheader($op)|0);
        $19 = ($18|0)==(0);
        if (!($19)) {
         HEAP32[$7>>2] = 3;
         $20 = (_vorbis_synthesis_headerin($vi,$vc,$op)|0);
         $21 = ($20|0)==(0);
         if (!($21)) {
          $ret$8 = -133;
          break L8;
         }
        }
       }
      }
      $22 = (__get_next_page($vf,$$12,65536,0)|0);
      $23 = tempRet0;
      $24 = ($22|0)==(-128);
      $25 = ($23|0)==(-1);
      $26 = $24 & $25;
      if ($26) {
       $ret$8 = -128;
       break L8;
      }
      $27 = ($23|0)<(0);
      if ($27) {
       $ret$8 = -132;
       break L8;
      }
      $28 = HEAP32[$7>>2]|0;
      $29 = ($28|0)==(3);
      if ($29) {
       $30 = HEAP32[$12>>2]|0;
       $31 = (_ogg_page_serialno($$12)|0);
       $32 = ($30|0)==($31|0);
       if ($32) {
        break L10;
       }
      }
      $33 = (_ogg_page_bos($$12)|0);
      $34 = ($33|0)==(0);
      if ($34) {
       label = 34;
       break L8;
      }
     }
    } else {
     L23: while(1) {
      $35 = HEAP32[$serialno_list>>2]|0;
      $36 = HEAP32[$serialno_n>>2]|0;
      $37 = (_ogg_page_serialno($$12)|0);
      $38 = ($35|0)==(0|0);
      $39 = ($36|0)==(0);
      $or$cond$i$i = $38 | $39;
      if (!($or$cond$i$i)) {
       $$013$i$i = $35;$$in$i$i = $36;
       while(1) {
        $42 = (($$in$i$i) + -1)|0;
        $43 = HEAP32[$$013$i$i>>2]|0;
        $44 = ($43|0)==($37|0);
        if ($44) {
         break L23;
        }
        $40 = ((($$013$i$i)) + 4|0);
        $41 = ($42|0)==(0);
        if ($41) {
         break;
        } else {
         $$013$i$i = $40;$$in$i$i = $42;
        }
       }
      }
      $47 = (_ogg_page_serialno($$12)|0);
      $48 = HEAP32[$serialno_n>>2]|0;
      $49 = (($48) + 1)|0;
      HEAP32[$serialno_n>>2] = $49;
      $50 = HEAP32[$serialno_list>>2]|0;
      $51 = ($50|0)==(0|0);
      if ($51) {
       $54 = (_malloc(4)|0);
       HEAP32[$serialno_list>>2] = $54;
       $$in$i = $54;$56 = $49;
      } else {
       $52 = $49 << 2;
       $53 = (_realloc($50,$52)|0);
       HEAP32[$serialno_list>>2] = $53;
       $$pre$i = HEAP32[$serialno_n>>2]|0;
       $$in$i = $53;$56 = $$pre$i;
      }
      $55 = (($56) + -1)|0;
      $57 = (($$in$i) + ($55<<2)|0);
      HEAP32[$57>>2] = $47;
      $58 = HEAP32[$7>>2]|0;
      $59 = ($58|0)<(3);
      if ($59) {
       $60 = (_ogg_page_serialno($$12)|0);
       (_ogg_stream_reset_serialno($11,$60)|0);
       (_ogg_stream_pagein($11,$$12)|0);
       $61 = (_ogg_stream_packetout($11,$op)|0);
       $62 = ($61|0)>(0);
       if ($62) {
        $63 = (_vorbis_synthesis_idheader($op)|0);
        $64 = ($63|0)==(0);
        if (!($64)) {
         HEAP32[$7>>2] = 3;
         $65 = (_vorbis_synthesis_headerin($vi,$vc,$op)|0);
         $66 = ($65|0)==(0);
         if (!($66)) {
          $ret$8 = -133;
          break L8;
         }
        }
       }
      }
      $67 = (__get_next_page($vf,$$12,65536,0)|0);
      $68 = tempRet0;
      $69 = ($67|0)==(-128);
      $70 = ($68|0)==(-1);
      $71 = $69 & $70;
      if ($71) {
       $ret$8 = -128;
       break L8;
      }
      $72 = ($68|0)<(0);
      if ($72) {
       $ret$8 = -132;
       break L8;
      }
      $73 = HEAP32[$7>>2]|0;
      $74 = ($73|0)==(3);
      if ($74) {
       $75 = HEAP32[$12>>2]|0;
       $76 = (_ogg_page_serialno($$12)|0);
       $77 = ($75|0)==($76|0);
       if ($77) {
        break L10;
       }
      }
      $78 = (_ogg_page_bos($$12)|0);
      $79 = ($78|0)==(0);
      if ($79) {
       label = 34;
       break L8;
      }
     }
     $45 = HEAP32[$serialno_list>>2]|0;
     $46 = ($45|0)==(0|0);
     if (!($46)) {
      _free($45);
     }
     HEAP32[$serialno_list>>2] = 0;
     HEAP32[$serialno_n>>2] = 0;
     $ret$8 = -133;
     break L8;
    }
   } while(0);
   (_ogg_stream_pagein($11,$$12)|0);
   label = 34;
  }
 } while(0);
 L49: do {
  if ((label|0) == 34) {
   $80 = HEAP32[$7>>2]|0;
   $81 = ($80|0)==(3);
   if ($81) {
    $82 = ((($vf)) + 120|0);
    $83 = ((($vf)) + 456|0);
    $allbos$030 = 0;$i$031 = 0;
    L52: while(1) {
     $i$1 = $i$031;
     L54: while(1) {
      $84 = ($i$1|0)<(2);
      if (!($84)) {
       $$2 = 0;
       label = 48;
       break L52;
      }
      $85 = (_ogg_stream_packetout($82,$op)|0);
      switch ($85|0) {
      case -1:  {
       $ret$8 = -133;
       break L49;
       break;
      }
      case 0:  {
       $$lcssa96 = $84;$i$1$lcssa94 = $i$1;
       break L54;
       break;
      }
      default: {
      }
      }
      $94 = (_vorbis_synthesis_headerin($vi,$vc,$op)|0);
      $95 = ($94|0)==(0);
      $96 = $95&1;
      $$i$1 = (($96) + ($i$1))|0;
      if ($95) {
       $i$1 = $$i$1;
      } else {
       $ret$8 = $94;
       break L49;
      }
     }
     $allbos$1$ph$us = $allbos$030;
     L59: while(1) {
      while(1) {
       (__get_next_page($vf,$$12,65536,0)|0);
       $86 = tempRet0;
       $87 = ($86|0)<(0);
       if ($87) {
        $ret$8 = -133;
        break L49;
       }
       $88 = HEAP32[$83>>2]|0;
       $89 = (_ogg_page_serialno($$12)|0);
       $90 = ($88|0)==($89|0);
       if ($90) {
        $allbos$1$ph$us$lcssa98 = $allbos$1$ph$us;
        break L59;
       }
       $91 = (_ogg_page_bos($$12)|0);
       $92 = ($91|0)==(0);
       if (!($92)) {
        break;
       }
      }
      $93 = ($allbos$1$ph$us|0)==(0);
      if ($93) {
       $allbos$1$ph$us = 1;
      } else {
       $ret$8 = -133;
       break L49;
      }
     }
     (_ogg_stream_pagein($82,$$12)|0);
     if ($$lcssa96) {
      $allbos$030 = $allbos$1$ph$us$lcssa98;$i$031 = $i$1$lcssa94;
     } else {
      $$2 = 0;
      label = 48;
      break;
     }
    }
    if ((label|0) == 48) {
     STACKTOP = sp;return ($$2|0);
    }
   } else {
    $ret$8 = -132;
   }
  }
 } while(0);
 _vorbis_info_clear($vi);
 _vorbis_comment_clear($vc);
 HEAP32[$7>>2] = 2;
 $$2 = $ret$8;
 STACKTOP = sp;return ($$2|0);
}
function __get_next_page($vf,$og,$0,$1) {
 $vf = $vf|0;
 $og = $og|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$lcssa = 0, $$lcssa20 = 0, $$lcssa21 = 0, $$lcssa79 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($1|0)>(0);
 $3 = ($0>>>0)>(0);
 $4 = ($1|0)==(0);
 $5 = $4 & $3;
 $6 = $2 | $5;
 $7 = ((($vf)) + 8|0);
 L1: do {
  if ($6) {
   $13 = $7;
   $14 = $13;
   $15 = HEAP32[$14>>2]|0;
   $16 = (($13) + 4)|0;
   $17 = $16;
   $18 = HEAP32[$17>>2]|0;
   $19 = (_i64Add(($15|0),($18|0),($0|0),($1|0))|0);
   $20 = tempRet0;
   $21 = ($20|0)>(0);
   $22 = ($19>>>0)>(0);
   $23 = ($20|0)==(0);
   $24 = $23 & $22;
   $25 = $21 | $24;
   $26 = ((($vf)) + 24|0);
   $27 = ($19|0)==(0);
   $28 = ($20|0)==(0);
   $29 = $27 & $28;
   $30 = ((($vf)) + 704|0);
   if (!($25)) {
    if ($29) {
     $80 = $26;
     label = 16;
     break;
    } else {
     $101 = $26;$121 = $30;$124 = $vf;
     label = 15;
     break;
    }
   }
   $31 = $7;
   $32 = $31;
   $33 = HEAP32[$32>>2]|0;
   $34 = (($31) + 4)|0;
   $35 = $34;
   $36 = HEAP32[$35>>2]|0;
   $37 = ($36|0)<($20|0);
   $38 = ($33>>>0)<($19>>>0);
   $39 = ($36|0)==($20|0);
   $40 = $39 & $38;
   $41 = $37 | $40;
   if (!($41)) {
    $148 = -1;$149 = -1;
    tempRet0 = ($148);
    return ($149|0);
   }
   while(1) {
    $42 = (_ogg_sync_pageseek($26,$og)|0);
    $43 = ($42|0)<(0);
    if ($43) {
     $55 = ($42|0)<(0);
     $56 = $55 << 31 >> 31;
     $57 = $7;
     $58 = $57;
     $59 = HEAP32[$58>>2]|0;
     $60 = (($57) + 4)|0;
     $61 = $60;
     $62 = HEAP32[$61>>2]|0;
     $63 = (_i64Subtract(($59|0),($62|0),($42|0),($56|0))|0);
     $64 = tempRet0;
     $65 = $7;
     $66 = $65;
     HEAP32[$66>>2] = $63;
     $67 = (($65) + 4)|0;
     $68 = $67;
     HEAP32[$68>>2] = $64;
    } else {
     $44 = ($42|0)==(0);
     if (!($44)) {
      $$lcssa = $42;
      label = 29;
      break L1;
     }
     if ($29) {
      $148 = -1;$149 = -1;
      label = 30;
      break;
     }
     $45 = (___errno_location()|0);
     HEAP32[$45>>2] = 0;
     $46 = HEAP32[$30>>2]|0;
     $47 = ($46|0)==(0|0);
     if ($47) {
      break L1;
     }
     $48 = HEAP32[$vf>>2]|0;
     $49 = ($48|0)==(0|0);
     if ($49) {
      $148 = -1;$149 = -2;
      label = 30;
      break;
     }
     $50 = (_ogg_sync_buffer($26,2048)|0);
     $51 = HEAP32[$30>>2]|0;
     $52 = HEAP32[$vf>>2]|0;
     $53 = (FUNCTION_TABLE_iiiii[$51 & 8191]($50,1,2048,$52)|0);
     $54 = ($53|0)>(0);
     if (!($54)) {
      $$lcssa20 = $45;$$lcssa21 = $53;
      label = 25;
      break L1;
     }
     (_ogg_sync_wrote($26,$53)|0);
    }
    $69 = $7;
    $70 = $69;
    $71 = HEAP32[$70>>2]|0;
    $72 = (($69) + 4)|0;
    $73 = $72;
    $74 = HEAP32[$73>>2]|0;
    $75 = ($74|0)<($20|0);
    $76 = ($71>>>0)<($19>>>0);
    $77 = ($74|0)==($20|0);
    $78 = $77 & $76;
    $79 = $75 | $78;
    if (!($79)) {
     $148 = -1;$149 = -1;
     label = 30;
     break;
    }
   }
   if ((label|0) == 30) {
    tempRet0 = ($148);
    return ($149|0);
   }
  } else {
   $8 = ((($vf)) + 24|0);
   $9 = ($0|0)==(0);
   $10 = ($1|0)==(0);
   $11 = $9 & $10;
   $12 = ((($vf)) + 704|0);
   if ($11) {
    $80 = $8;
    label = 16;
   } else {
    $101 = $8;$121 = $12;$124 = $vf;
    label = 15;
   }
  }
 } while(0);
 L24: do {
  if ((label|0) == 15) {
   while(1) {
    $102 = (_ogg_sync_pageseek($101,$og)|0);
    $103 = ($102|0)<(0);
    if ($103) {
     $104 = ($102|0)<(0);
     $105 = $104 << 31 >> 31;
     $106 = $7;
     $107 = $106;
     $108 = HEAP32[$107>>2]|0;
     $109 = (($106) + 4)|0;
     $110 = $109;
     $111 = HEAP32[$110>>2]|0;
     $112 = (_i64Subtract(($108|0),($111|0),($102|0),($105|0))|0);
     $113 = tempRet0;
     $114 = $7;
     $115 = $114;
     HEAP32[$115>>2] = $112;
     $116 = (($114) + 4)|0;
     $117 = $116;
     HEAP32[$117>>2] = $113;
     continue;
    }
    $118 = ($102|0)==(0);
    if (!($118)) {
     $$lcssa = $102;
     label = 29;
     break L24;
    }
    $119 = (___errno_location()|0);
    HEAP32[$119>>2] = 0;
    $120 = HEAP32[$121>>2]|0;
    $122 = ($120|0)==(0|0);
    if ($122) {
     break L24;
    }
    $123 = HEAP32[$124>>2]|0;
    $125 = ($123|0)==(0|0);
    if ($125) {
     $148 = -1;$149 = -2;
     break;
    }
    $126 = (_ogg_sync_buffer($101,2048)|0);
    $127 = HEAP32[$121>>2]|0;
    $128 = HEAP32[$124>>2]|0;
    $129 = (FUNCTION_TABLE_iiiii[$127 & 8191]($126,1,2048,$128)|0);
    $130 = ($129|0)>(0);
    if (!($130)) {
     $$lcssa20 = $119;$$lcssa21 = $129;
     label = 25;
     break L24;
    }
    (_ogg_sync_wrote($101,$129)|0);
   }
   tempRet0 = ($148);
   return ($149|0);
  }
  else if ((label|0) == 16) {
   $81 = (_ogg_sync_pageseek($80,$og)|0);
   $82 = ($81|0)<(0);
   if ($82) {
    $85 = $81;
    while(1) {
     $84 = ($85|0)<(0);
     $86 = $84 << 31 >> 31;
     $87 = $7;
     $88 = $87;
     $89 = HEAP32[$88>>2]|0;
     $90 = (($87) + 4)|0;
     $91 = $90;
     $92 = HEAP32[$91>>2]|0;
     $93 = (_i64Subtract(($89|0),($92|0),($85|0),($86|0))|0);
     $94 = tempRet0;
     $95 = $7;
     $96 = $95;
     HEAP32[$96>>2] = $93;
     $97 = (($95) + 4)|0;
     $98 = $97;
     HEAP32[$98>>2] = $94;
     $99 = (_ogg_sync_pageseek($80,$og)|0);
     $100 = ($99|0)<(0);
     if ($100) {
      $85 = $99;
     } else {
      $$lcssa79 = $99;
      break;
     }
    }
   } else {
    $$lcssa79 = $81;
   }
   $83 = ($$lcssa79|0)==(0);
   if ($83) {
    $148 = -1;$149 = -1;
    tempRet0 = ($148);
    return ($149|0);
   } else {
    $$lcssa = $$lcssa79;
    label = 29;
   }
  }
 } while(0);
 if ((label|0) == 29) {
  $134 = $7;
  $135 = $134;
  $136 = HEAP32[$135>>2]|0;
  $137 = (($134) + 4)|0;
  $138 = $137;
  $139 = HEAP32[$138>>2]|0;
  $140 = ($$lcssa|0)<(0);
  $141 = $140 << 31 >> 31;
  $142 = (_i64Add(($136|0),($139|0),($$lcssa|0),($141|0))|0);
  $143 = tempRet0;
  $144 = $7;
  $145 = $144;
  HEAP32[$145>>2] = $142;
  $146 = (($144) + 4)|0;
  $147 = $146;
  HEAP32[$147>>2] = $143;
  $148 = $139;$149 = $136;
  tempRet0 = ($148);
  return ($149|0);
 }
 if ((label|0) == 25) {
  $131 = ($$lcssa21|0)==(0);
  if ($131) {
   $132 = HEAP32[$$lcssa20>>2]|0;
   $133 = ($132|0)==(0);
   if ($133) {
    $148 = -1;$149 = -2;
    tempRet0 = ($148);
    return ($149|0);
   }
  }
 }
 $148 = -1;$149 = -128;
 tempRet0 = ($148);
 return ($149|0);
}
function __ov_open2($vf) {
 $vf = $vf|0;
 var $$0 = 0, $$0$i = 0, $$0$i$ph = 0, $$0$i1 = 0, $$pre$phiZ2D = 0, $$pre$pre$phiZZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $endgran$i = 0, $endserial$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $endgran$i = sp;
 $endserial$i = sp + 8|0;
 $0 = ((($vf)) + 88|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(1);
 if (!($2)) {
  $$0 = -131;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$0>>2] = 2;
 $3 = ((($vf)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0);
 if ($5) {
  HEAP32[$0>>2] = 3;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $6 = ((($vf)) + 60|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7;
 $9 = $8;
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8) + 4)|0;
 $12 = $11;
 $13 = HEAP32[$12>>2]|0;
 $14 = $endgran$i;
 $15 = $14;
 HEAP32[$15>>2] = -1;
 $16 = (($14) + 4)|0;
 $17 = $16;
 HEAP32[$17>>2] = -1;
 $18 = ((($vf)) + 456|0);
 $19 = HEAP32[$18>>2]|0;
 HEAP32[$endserial$i>>2] = $19;
 $20 = ((($vf)) + 72|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = (__initial_pcmoffset($vf,$21)|0);
 $23 = tempRet0;
 $24 = ((($vf)) + 708|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = ($25|0)==(0|0);
 do {
  if ($26) {
   label = 5;
  } else {
   $27 = ((($vf)) + 716|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = ($28|0)==(0|0);
   if ($29) {
    label = 5;
   } else {
    $31 = HEAP32[$vf>>2]|0;
    (FUNCTION_TABLE_iiiii[$25 & 8191]($31,0,0,2)|0);
    $32 = HEAP32[$27>>2]|0;
    $33 = HEAP32[$vf>>2]|0;
    $34 = (FUNCTION_TABLE_ii[$32 & 8191]($33)|0);
    $35 = ($34|0)<(0);
    $36 = $35 << 31 >> 31;
    $37 = ((($vf)) + 16|0);
    $38 = $37;
    $39 = $38;
    HEAP32[$39>>2] = $34;
    $40 = (($38) + 4)|0;
    $41 = $40;
    HEAP32[$41>>2] = $36;
    $42 = ((($vf)) + 8|0);
    $43 = $42;
    $44 = $43;
    HEAP32[$44>>2] = $34;
    $45 = (($43) + 4)|0;
    $46 = $45;
    HEAP32[$46>>2] = $36;
    $47 = ($34|0)==(-1);
    if ($47) {
     $$0$i$ph = -131;$$pre$pre$phiZZ2D = $vf;
     label = 11;
    } else {
     $48 = ((($vf)) + 64|0);
     $49 = HEAP32[$48>>2]|0;
     $50 = ((($49)) + 8|0);
     $51 = ((($49)) + 4|0);
     $52 = HEAP32[$51>>2]|0;
     $53 = (__get_prev_page_serial($vf,$34,$36,$50,$52,$endserial$i,$endgran$i)|0);
     $54 = tempRet0;
     $55 = ($54|0)<(0);
     if ($55) {
      $$0$i = $53;
     } else {
      $56 = $endgran$i;
      $57 = $56;
      $58 = HEAP32[$57>>2]|0;
      $59 = (($56) + 4)|0;
      $60 = $59;
      $61 = HEAP32[$60>>2]|0;
      $62 = HEAP32[$endserial$i>>2]|0;
      $63 = HEAP32[$48>>2]|0;
      $64 = ((($63)) + 8|0);
      $65 = ((($63)) + 4|0);
      $66 = HEAP32[$65>>2]|0;
      $67 = (__bisect_forward_serialno($vf,0,0,$10,$13,$53,$54,$58,$61,$62,$64,$66,0)|0);
      $68 = ($67|0)<(0);
      if ($68) {
       $$0$i$ph = -128;$$pre$pre$phiZZ2D = $vf;
       label = 11;
       break;
      }
      $69 = ((($vf)) + 56|0);
      $70 = HEAP32[$69>>2]|0;
      $71 = $70;
      $72 = $71;
      HEAP32[$72>>2] = 0;
      $73 = (($71) + 4)|0;
      $74 = $73;
      HEAP32[$74>>2] = 0;
      $75 = HEAP32[$48>>2]|0;
      HEAP32[$75>>2] = $19;
      $76 = HEAP32[$6>>2]|0;
      $77 = $76;
      $78 = $77;
      HEAP32[$78>>2] = $10;
      $79 = (($77) + 4)|0;
      $80 = $79;
      HEAP32[$80>>2] = $13;
      $81 = ((($vf)) + 68|0);
      $82 = HEAP32[$81>>2]|0;
      $83 = $82;
      $84 = $83;
      HEAP32[$84>>2] = $22;
      $85 = (($83) + 4)|0;
      $86 = $85;
      HEAP32[$86>>2] = $23;
      $87 = ((($82)) + 8|0);
      $88 = $87;
      $89 = $88;
      $90 = HEAP32[$89>>2]|0;
      $91 = (($88) + 4)|0;
      $92 = $91;
      $93 = HEAP32[$92>>2]|0;
      $94 = (_i64Subtract(($90|0),($93|0),($22|0),($23|0))|0);
      $95 = tempRet0;
      $96 = ($95|0)<(0);
      $97 = $96 ? 0 : $94;
      $98 = $96 ? 0 : $95;
      $99 = $87;
      $100 = $99;
      HEAP32[$100>>2] = $97;
      $101 = (($99) + 4)|0;
      $102 = $101;
      HEAP32[$102>>2] = $98;
      $103 = (_ov_raw_seek($vf,$10,$13)|0);
      $$0$i = $103;
     }
     $104 = ($$0$i|0)==(0);
     if ($104) {
      $$0 = 0;
      STACKTOP = sp;return ($$0|0);
     } else {
      $$0$i1 = $$0$i;$$pre$phiZ2D = $vf;
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 5) {
  $30 = ((($vf)) + 8|0);
  ;HEAP32[$30>>2]=4294967295|0;HEAP32[$30+4>>2]=4294967295|0;HEAP32[$30+8>>2]=4294967295|0;HEAP32[$30+12>>2]=4294967295|0;
  $$0$i$ph = -131;$$pre$pre$phiZZ2D = $vf;
  label = 11;
 }
 if ((label|0) == 11) {
  $$0$i1 = $$0$i$ph;$$pre$phiZ2D = $$pre$pre$phiZZ2D;
 }
 HEAP32[$$pre$phiZ2D>>2] = 0;
 (_ov_clear($vf)|0);
 $$0 = $$0$i1;
 STACKTOP = sp;return ($$0|0);
}
function __initial_pcmoffset($vf,$vi) {
 $vf = $vf|0;
 $vi = $vi|0;
 var $$lcssa = 0, $$lcssa$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lastblock$09 = 0, $lastblock$1$ph = 0, $lastblock$1$ph$lcssa43 = 0, $lastblock$1$ph$ph = 0, $lastblock$1$ph44 = 0, $lastblock$3 = 0, $og = 0, $op = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $og = sp + 32|0;
 $op = sp;
 $0 = ((($vf)) + 120|0);
 $1 = ((($vf)) + 456|0);
 $2 = HEAP32[$1>>2]|0;
 (__get_next_page($vf,$og,-1,-1)|0);
 $3 = tempRet0;
 $4 = ($3|0)<(0);
 L1: do {
  if ($4) {
   $40 = 0;$42 = 0;
  } else {
   $44 = 0;$45 = 0;$lastblock$09 = -1;
   while(1) {
    $5 = (_ogg_page_bos($og)|0);
    $6 = ($5|0)==(0);
    if (!($6)) {
     $40 = $45;$42 = $44;
     break L1;
    }
    $7 = (_ogg_page_serialno($og)|0);
    $8 = ($7|0)==($2|0);
    if ($8) {
     (_ogg_stream_pagein($0,$og)|0);
     $22 = $44;$23 = $45;$lastblock$1$ph$ph = $lastblock$09;
     L7: while(1) {
      $lastblock$1$ph = $lastblock$1$ph$ph;
      while(1) {
       $9 = (_ogg_stream_packetout($0,$op)|0);
       $10 = ($9|0)==(0);
       if ($10) {
        $48 = $22;$49 = $23;$lastblock$1$ph44 = $lastblock$1$ph;
        break L7;
       } else {
        $12 = $9;
       }
       while(1) {
        $11 = ($12|0)>(0);
        if ($11) {
         $15 = (_vorbis_packet_blocksize($vi,$op)|0);
         $16 = ($15|0)>(-1);
         if ($16) {
          $$lcssa = $15;
          break;
         }
        }
        $13 = (_ogg_stream_packetout($0,$op)|0);
        $14 = ($13|0)==(0);
        if ($14) {
         $48 = $22;$49 = $23;$lastblock$1$ph44 = $lastblock$1$ph;
         break L7;
        } else {
         $12 = $13;
        }
       }
       $17 = ($lastblock$1$ph|0)==(-1);
       if ($17) {
        $lastblock$1$ph = $$lcssa;
       } else {
        $$lcssa$lcssa = $$lcssa;$lastblock$1$ph$lcssa43 = $lastblock$1$ph;
        break;
       }
      }
      $18 = (($$lcssa$lcssa) + ($lastblock$1$ph$lcssa43))|0;
      $19 = $18 >> 2;
      $20 = ($19|0)<(0);
      $21 = $20 << 31 >> 31;
      $24 = (_i64Add(($19|0),($21|0),($22|0),($23|0))|0);
      $25 = tempRet0;
      $22 = $24;$23 = $25;$lastblock$1$ph$ph = $$lcssa$lcssa;
     }
     $26 = (_ogg_page_granulepos($og)|0);
     $27 = tempRet0;
     $28 = ($26|0)==(-1);
     $29 = ($27|0)==(-1);
     $30 = $28 & $29;
     if ($30) {
      $46 = $48;$47 = $49;$lastblock$3 = $lastblock$1$ph44;
     } else {
      $33 = $48;$34 = $49;
      break;
     }
    } else {
     $46 = $44;$47 = $45;$lastblock$3 = $lastblock$09;
    }
    (__get_next_page($vf,$og,-1,-1)|0);
    $37 = tempRet0;
    $38 = ($37|0)<(0);
    if ($38) {
     $40 = $47;$42 = $46;
     break L1;
    } else {
     $44 = $46;$45 = $47;$lastblock$09 = $lastblock$3;
    }
   }
   $31 = (_ogg_page_granulepos($og)|0);
   $32 = tempRet0;
   $35 = (_i64Subtract(($31|0),($32|0),($33|0),($34|0))|0);
   $36 = tempRet0;
   $40 = $36;$42 = $35;
  }
 } while(0);
 $39 = ($40|0)<(0);
 $41 = $39 ? 0 : $42;
 $43 = $39 ? 0 : $40;
 tempRet0 = ($43);
 STACKTOP = sp;return ($41|0);
}
function __get_prev_page_serial($vf,$0,$1,$serial_list,$serial_n,$serialno,$granpos) {
 $vf = $vf|0;
 $0 = $0|0;
 $1 = $1|0;
 $serial_list = $serial_list|0;
 $serial_n = $serial_n|0;
 $serialno = $serialno|0;
 $granpos = $granpos|0;
 var $$013$i = 0, $$in$i = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $og = 0;
 var $or$cond$i = 0, $ret_serialno$0$off0$lcssa68 = 0, $ret_serialno$0$off035$ph = 0, $ret_serialno$0$off035$us$ph = 0, $ret_serialno$1$off0$lcssa = 0, $ret_serialno$1$off0$lcssa$lcssa = 0, $ret_serialno$1$off0$lcssa$us = 0, $ret_serialno$1$off023 = 0, $ret_serialno$1$off023$us$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $og = sp;
 $2 = ((($vf)) + 8|0);
 $3 = ((($vf)) + 708|0);
 $4 = ((($vf)) + 24|0);
 $5 = ($serial_list|0)==(0|0);
 $6 = ($serial_n|0)==(0);
 $or$cond$i = $5 | $6;
 L1: do {
  if ($or$cond$i) {
   $175 = -1;$176 = -1;$177 = $0;$178 = $1;$ret_serialno$0$off035$us$ph = -1;
   L2: while(1) {
    $7 = $177;$8 = $178;
    while(1) {
     $9 = (_i64Add(($7|0),($8|0),-65536,-1)|0);
     $10 = tempRet0;
     $11 = ($10|0)<(0);
     $12 = $11 ? 0 : $9;
     $13 = $11 ? 0 : $10;
     $14 = HEAP32[$vf>>2]|0;
     $15 = ($14|0)==(0|0);
     if ($15) {
      $173 = -1;$174 = -129;
      label = 35;
      break L2;
     }
     $16 = $2;
     $17 = $16;
     $18 = HEAP32[$17>>2]|0;
     $19 = (($16) + 4)|0;
     $20 = $19;
     $21 = HEAP32[$20>>2]|0;
     $22 = ($18|0)==($12|0);
     $23 = ($21|0)==($13|0);
     $24 = $22 & $23;
     if ($24) {
      $40 = $13;$42 = $12;
     } else {
      $25 = HEAP32[$3>>2]|0;
      $26 = ($25|0)==(0|0);
      if ($26) {
       $173 = -1;$174 = -128;
       label = 35;
       break L2;
      }
      $27 = (FUNCTION_TABLE_iiiii[$25 & 8191]($14,$12,$13,0)|0);
      $28 = ($27|0)==(-1);
      if ($28) {
       $173 = -1;$174 = -128;
       label = 35;
       break L2;
      }
      $29 = $2;
      $30 = $29;
      HEAP32[$30>>2] = $12;
      $31 = (($29) + 4)|0;
      $32 = $31;
      HEAP32[$32>>2] = $13;
      (_ogg_sync_reset($4)|0);
      $33 = $2;
      $34 = $33;
      $35 = HEAP32[$34>>2]|0;
      $36 = (($33) + 4)|0;
      $37 = $36;
      $38 = HEAP32[$37>>2]|0;
      $40 = $38;$42 = $35;
     }
     $39 = ($40|0)<($1|0);
     $41 = ($42>>>0)<($0>>>0);
     $43 = ($40|0)==($1|0);
     $44 = $43 & $41;
     $45 = $39 | $44;
     if ($45) {
      $185 = $42;$186 = $40;$187 = $12;$188 = $13;
      break;
     } else {
      $7 = $12;$8 = $13;
     }
    }
    $189 = $175;$190 = $176;$191 = -1;$192 = -1;$51 = $185;$52 = $186;$ret_serialno$1$off023$us$us = $ret_serialno$0$off035$us$ph;
    while(1) {
     $53 = (_i64Subtract(($0|0),($1|0),($51|0),($52|0))|0);
     $54 = tempRet0;
     $55 = (__get_next_page($vf,$og,$53,$54)|0);
     $56 = tempRet0;
     $57 = ($55|0)==(-128);
     $58 = ($56|0)==(-1);
     $59 = $57 & $58;
     if ($59) {
      $173 = -1;$174 = -128;
      label = 35;
      break L2;
     }
     $60 = ($56|0)<(0);
     if ($60) {
      $193 = $189;$194 = $190;$47 = $191;$49 = $192;$ret_serialno$1$off0$lcssa$us = $ret_serialno$1$off023$us$us;
      break;
     }
     $61 = (_ogg_page_serialno($og)|0);
     $62 = (_ogg_page_granulepos($og)|0);
     $63 = tempRet0;
     $64 = HEAP32[$serialno>>2]|0;
     $65 = ($61|0)==($64|0);
     if ($65) {
      $66 = $granpos;
      $67 = $66;
      HEAP32[$67>>2] = $62;
      $68 = (($66) + 4)|0;
      $69 = $68;
      HEAP32[$69>>2] = $63;
     }
     $70 = $2;
     $71 = $70;
     $72 = HEAP32[$71>>2]|0;
     $73 = (($70) + 4)|0;
     $74 = $73;
     $75 = HEAP32[$74>>2]|0;
     $76 = ($75|0)<($1|0);
     $77 = ($72>>>0)<($0>>>0);
     $78 = ($75|0)==($1|0);
     $79 = $78 & $77;
     $80 = $76 | $79;
     if ($80) {
      $189 = $62;$190 = $63;$191 = $55;$192 = $56;$51 = $72;$52 = $75;$ret_serialno$1$off023$us$us = $61;
     } else {
      $193 = $62;$194 = $63;$47 = $55;$49 = $56;$ret_serialno$1$off0$lcssa$us = $61;
      break;
     }
    }
    $46 = ($47|0)==(-1);
    $48 = ($49|0)==(-1);
    $50 = $46 & $48;
    if ($50) {
     $175 = $193;$176 = $194;$177 = $187;$178 = $188;$ret_serialno$0$off035$us$ph = $ret_serialno$1$off0$lcssa$us;
    } else {
     $169 = $193;$172 = $194;$195 = $47;$196 = $49;$ret_serialno$0$off0$lcssa68 = $ret_serialno$1$off0$lcssa$us;
     break L1;
    }
   }
   if ((label|0) == 35) {
    tempRet0 = ($173);
    STACKTOP = sp;return ($174|0);
   }
  } else {
   $179 = -1;$180 = -1;$181 = -1;$182 = -1;$183 = $0;$184 = $1;$ret_serialno$0$off035$ph = -1;
   L26: while(1) {
    $86 = $183;$87 = $184;
    while(1) {
     $88 = (_i64Add(($86|0),($87|0),-65536,-1)|0);
     $89 = tempRet0;
     $90 = ($89|0)<(0);
     $91 = $90 ? 0 : $88;
     $92 = $90 ? 0 : $89;
     $93 = HEAP32[$vf>>2]|0;
     $94 = ($93|0)==(0|0);
     if ($94) {
      $173 = -1;$174 = -129;
      label = 35;
      break L26;
     }
     $95 = $2;
     $96 = $95;
     $97 = HEAP32[$96>>2]|0;
     $98 = (($95) + 4)|0;
     $99 = $98;
     $100 = HEAP32[$99>>2]|0;
     $101 = ($97|0)==($91|0);
     $102 = ($100|0)==($92|0);
     $103 = $101 & $102;
     if ($103) {
      $119 = $92;$121 = $91;
     } else {
      $104 = HEAP32[$3>>2]|0;
      $105 = ($104|0)==(0|0);
      if ($105) {
       $173 = -1;$174 = -128;
       label = 35;
       break L26;
      }
      $106 = (FUNCTION_TABLE_iiiii[$104 & 8191]($93,$91,$92,0)|0);
      $107 = ($106|0)==(-1);
      if ($107) {
       $173 = -1;$174 = -128;
       label = 35;
       break L26;
      }
      $108 = $2;
      $109 = $108;
      HEAP32[$109>>2] = $91;
      $110 = (($108) + 4)|0;
      $111 = $110;
      HEAP32[$111>>2] = $92;
      (_ogg_sync_reset($4)|0);
      $112 = $2;
      $113 = $112;
      $114 = HEAP32[$113>>2]|0;
      $115 = (($112) + 4)|0;
      $116 = $115;
      $117 = HEAP32[$116>>2]|0;
      $119 = $117;$121 = $114;
     }
     $118 = ($119|0)<($1|0);
     $120 = ($121>>>0)<($0>>>0);
     $122 = ($119|0)==($1|0);
     $123 = $122 & $120;
     $124 = $118 | $123;
     if ($124) {
      $201 = $91;$202 = $92;$207 = $121;$208 = $119;
      break;
     } else {
      $86 = $91;$87 = $92;
     }
    }
    $125 = $207;$126 = $208;$209 = $179;$210 = $180;$211 = -1;$212 = -1;$213 = $181;$214 = $182;$ret_serialno$1$off023 = $ret_serialno$0$off035$ph;
    while(1) {
     $127 = (_i64Subtract(($0|0),($1|0),($125|0),($126|0))|0);
     $128 = tempRet0;
     $129 = (__get_next_page($vf,$og,$127,$128)|0);
     $130 = tempRet0;
     $131 = ($129|0)==(-128);
     $132 = ($130|0)==(-1);
     $133 = $131 & $132;
     if ($133) {
      $173 = -1;$174 = -128;
      label = 35;
      break L26;
     }
     $134 = ($130|0)<(0);
     if ($134) {
      $197 = $209;$198 = $210;$199 = $213;$200 = $214;$82 = $211;$84 = $212;$ret_serialno$1$off0$lcssa = $ret_serialno$1$off023;
      break;
     }
     $135 = (_ogg_page_serialno($og)|0);
     $136 = (_ogg_page_granulepos($og)|0);
     $137 = tempRet0;
     $138 = HEAP32[$serialno>>2]|0;
     $139 = ($135|0)==($138|0);
     if ($139) {
      $140 = $granpos;
      $141 = $140;
      HEAP32[$141>>2] = $136;
      $142 = (($140) + 4)|0;
      $143 = $142;
      HEAP32[$143>>2] = $137;
      $215 = $129;$216 = $130;
     } else {
      $215 = $213;$216 = $214;
     }
     $$013$i = $serial_list;$$in$i = $serial_n;
     while(1) {
      $157 = (($$in$i) + -1)|0;
      $158 = HEAP32[$$013$i>>2]|0;
      $159 = ($158|0)==($135|0);
      if ($159) {
       $217 = $215;$218 = $216;
       break;
      }
      $155 = ((($$013$i)) + 4|0);
      $156 = ($157|0)==(0);
      if ($156) {
       $217 = -1;$218 = -1;
       break;
      } else {
       $$013$i = $155;$$in$i = $157;
      }
     }
     $144 = $2;
     $145 = $144;
     $146 = HEAP32[$145>>2]|0;
     $147 = (($144) + 4)|0;
     $148 = $147;
     $149 = HEAP32[$148>>2]|0;
     $150 = ($149|0)<($1|0);
     $151 = ($146>>>0)<($0>>>0);
     $152 = ($149|0)==($1|0);
     $153 = $152 & $151;
     $154 = $150 | $153;
     if ($154) {
      $125 = $146;$126 = $149;$209 = $136;$210 = $137;$211 = $129;$212 = $130;$213 = $217;$214 = $218;$ret_serialno$1$off023 = $135;
     } else {
      $197 = $136;$198 = $137;$199 = $217;$200 = $218;$82 = $129;$84 = $130;$ret_serialno$1$off0$lcssa = $135;
      break;
     }
    }
    $81 = ($82|0)==(-1);
    $83 = ($84|0)==(-1);
    $85 = $81 & $83;
    if ($85) {
     $179 = $197;$180 = $198;$181 = $199;$182 = $200;$183 = $201;$184 = $202;$ret_serialno$0$off035$ph = $ret_serialno$1$off0$lcssa;
    } else {
     $161 = $200;$163 = $199;$203 = $82;$204 = $84;$205 = $197;$206 = $198;$ret_serialno$1$off0$lcssa$lcssa = $ret_serialno$1$off0$lcssa;
     label = 33;
     break;
    }
   }
   if ((label|0) == 33) {
    $160 = ($161|0)>(-1);
    $162 = ($163>>>0)>(4294967295);
    $164 = ($161|0)==(-1);
    $165 = $164 & $162;
    $166 = $160 | $165;
    if ($166) {
     $173 = $161;$174 = $163;
    } else {
     $169 = $205;$172 = $206;$195 = $203;$196 = $204;$ret_serialno$0$off0$lcssa68 = $ret_serialno$1$off0$lcssa$lcssa;
     break;
    }
    tempRet0 = ($173);
    STACKTOP = sp;return ($174|0);
   }
   else if ((label|0) == 35) {
    tempRet0 = ($173);
    STACKTOP = sp;return ($174|0);
   }
  }
 } while(0);
 HEAP32[$serialno>>2] = $ret_serialno$0$off0$lcssa68;
 $167 = $granpos;
 $168 = $167;
 HEAP32[$168>>2] = $169;
 $170 = (($167) + 4)|0;
 $171 = $170;
 HEAP32[$171>>2] = $172;
 $173 = $196;$174 = $195;
 tempRet0 = ($173);
 STACKTOP = sp;return ($174|0);
}
function __bisect_forward_serialno($vf,$0,$1,$2,$3,$4,$5,$6,$7,$endserial,$currentno_list,$currentnos,$m) {
 $vf = $vf|0;
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 $endserial = $endserial|0;
 $currentno_list = $currentno_list|0;
 $currentnos = $currentnos|0;
 $m = $m|0;
 var $$013$i = 0, $$013$i$i = 0, $$1 = 0, $$2 = 0, $$3 = 0, $$in$i = 0, $$in$i$i = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0;
 var $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0;
 var $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0;
 var $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0;
 var $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0;
 var $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $next_serialno_list = 0, $next_serialnos = 0, $og = 0;
 var $or$cond$i = 0, $searchgran = 0, $testserial = 0, $vc = 0, $vi = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $8 = sp + 8|0;
 $9 = sp + 96|0;
 $searchgran = sp;
 $og = sp + 80|0;
 $next_serialno_list = sp + 76|0;
 $next_serialnos = sp + 72|0;
 $vi = sp + 40|0;
 $vc = sp + 24|0;
 $testserial = sp + 16|0;
 $10 = $8;
 $11 = $10;
 HEAP32[$11>>2] = $6;
 $12 = (($10) + 4)|0;
 $13 = $12;
 HEAP32[$13>>2] = $7;
 HEAP32[$9>>2] = $endserial;
 $14 = $searchgran;
 $15 = $14;
 HEAP32[$15>>2] = -1;
 $16 = (($14) + 4)|0;
 $17 = $16;
 HEAP32[$17>>2] = -1;
 $18 = ((($vf)) + 456|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($currentno_list|0)==(0|0);
 $21 = ($currentnos|0)==(0);
 $or$cond$i = $20 | $21;
 L1: do {
  if ($or$cond$i) {
   label = 13;
  } else {
   $$013$i = $currentno_list;$$in$i = $currentnos;
   while(1) {
    $24 = (($$in$i) + -1)|0;
    $25 = HEAP32[$$013$i>>2]|0;
    $26 = ($25|0)==($endserial|0);
    if ($26) {
     break;
    }
    $22 = ((($$013$i)) + 4|0);
    $23 = ($24|0)==(0);
    if ($23) {
     label = 13;
     break L1;
    } else {
     $$013$i = $22;$$in$i = $24;
    }
   }
   $27 = ($19|0)==($endserial|0);
   if (!($27)) {
    $28 = $4;$29 = $5;
    while(1) {
     HEAP32[$9>>2] = $19;
     $30 = (__get_prev_page_serial($vf,$28,$29,$currentno_list,$currentnos,$9,$8)|0);
     $31 = tempRet0;
     $32 = HEAP32[$9>>2]|0;
     $33 = ($32|0)==($19|0);
     if ($33) {
      break;
     } else {
      $28 = $30;$29 = $31;
     }
    }
   }
   $34 = (($m) + 1)|0;
   $35 = ((($vf)) + 52|0);
   HEAP32[$35>>2] = $34;
   $36 = ((($vf)) + 56|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = ($37|0)==(0|0);
   if (!($38)) {
    _free($37);
   }
   $39 = ((($vf)) + 64|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = ($40|0)==(0|0);
   if (!($41)) {
    _free($40);
   }
   $42 = ((($vf)) + 60|0);
   $43 = HEAP32[$42>>2]|0;
   $44 = ($43|0)==(0|0);
   if (!($44)) {
    _free($43);
   }
   $45 = HEAP32[$35>>2]|0;
   $46 = $45 << 3;
   $47 = (($46) + 8)|0;
   $48 = (_malloc($47)|0);
   HEAP32[$36>>2] = $48;
   $49 = ((($vf)) + 72|0);
   $50 = HEAP32[$49>>2]|0;
   $51 = $45 << 5;
   $52 = (_realloc($50,$51)|0);
   HEAP32[$49>>2] = $52;
   $53 = ((($vf)) + 76|0);
   $54 = HEAP32[$53>>2]|0;
   $55 = HEAP32[$35>>2]|0;
   $56 = $55 << 4;
   $57 = (_realloc($54,$56)|0);
   HEAP32[$53>>2] = $57;
   $58 = HEAP32[$35>>2]|0;
   $59 = $58 << 2;
   $60 = (_malloc($59)|0);
   HEAP32[$39>>2] = $60;
   $61 = $58 << 3;
   $62 = (_malloc($61)|0);
   HEAP32[$42>>2] = $62;
   $63 = $58 << 4;
   $64 = (_malloc($63)|0);
   $65 = ((($vf)) + 68|0);
   HEAP32[$65>>2] = $64;
   $66 = HEAP32[$36>>2]|0;
   $67 = (($66) + ($34<<3)|0);
   $68 = $67;
   $69 = $68;
   HEAP32[$69>>2] = $4;
   $70 = (($68) + 4)|0;
   $71 = $70;
   HEAP32[$71>>2] = $5;
   $72 = (($66) + ($m<<3)|0);
   $73 = $72;
   $74 = $73;
   HEAP32[$74>>2] = $0;
   $75 = (($73) + 4)|0;
   $76 = $75;
   HEAP32[$76>>2] = $1;
   $77 = $8;
   $78 = $77;
   $79 = HEAP32[$78>>2]|0;
   $80 = (($77) + 4)|0;
   $81 = $80;
   $82 = HEAP32[$81>>2]|0;
   $83 = ($82|0)<(0);
   $84 = $83 ? 0 : $79;
   $85 = $83 ? 0 : $82;
   $86 = $m << 1;
   $87 = $86 | 1;
   $88 = (($64) + ($87<<3)|0);
   $89 = $88;
   $90 = $89;
   HEAP32[$90>>2] = $84;
   $91 = (($89) + 4)|0;
   $92 = $91;
   HEAP32[$92>>2] = $85;
  }
 } while(0);
 if ((label|0) == 13) {
  HEAP32[$next_serialno_list>>2] = 0;
  HEAP32[$next_serialnos>>2] = 0;
  $93 = (($19) + 1)|0;
  HEAP32[$testserial>>2] = $93;
  $94 = ($5|0)>($3|0);
  $95 = ($4>>>0)>($2>>>0);
  $96 = ($5|0)==($3|0);
  $97 = $96 & $95;
  $98 = $94 | $97;
  L21: do {
   if ($98) {
    $99 = ((($vf)) + 8|0);
    $100 = ((($vf)) + 708|0);
    $101 = ((($vf)) + 24|0);
    $160 = $2;$161 = $3;$348 = $4;$349 = $5;$350 = $4;$351 = $5;
    while(1) {
     if ($or$cond$i) {
      $104 = $160;$105 = $161;$352 = $350;$353 = $351;$354 = $348;$355 = $349;
      break;
     } else {
      $158 = $350;$159 = $351;$212 = $348;$214 = $349;
     }
     L25: while(1) {
      $162 = (_i64Subtract(($158|0),($159|0),($160|0),($161|0))|0);
      $163 = tempRet0;
      $164 = ($163|0)<(0);
      $165 = ($162>>>0)<(65536);
      $166 = ($163|0)==(0);
      $167 = $166 & $165;
      $168 = $164 | $167;
      if ($168) {
       $182 = $160;$184 = $161;
      } else {
       $169 = (_i64Add(($158|0),($159|0),($160|0),($161|0))|0);
       $170 = tempRet0;
       $171 = (___divdi3(($169|0),($170|0),2,0)|0);
       $172 = tempRet0;
       $182 = $171;$184 = $172;
      }
      $173 = HEAP32[$vf>>2]|0;
      $174 = ($173|0)==(0|0);
      if ($174) {
       $$1 = 0;$$2 = -129;
       break L21;
      }
      $175 = $99;
      $176 = $175;
      $177 = HEAP32[$176>>2]|0;
      $178 = (($175) + 4)|0;
      $179 = $178;
      $180 = HEAP32[$179>>2]|0;
      $181 = ($177|0)==($182|0);
      $183 = ($180|0)==($184|0);
      $185 = $181 & $183;
      if (!($185)) {
       $186 = HEAP32[$100>>2]|0;
       $187 = ($186|0)==(0|0);
       if ($187) {
        $$1 = 0;$$2 = -128;
        break L21;
       }
       $188 = (FUNCTION_TABLE_iiiii[$186 & 8191]($173,$182,$184,0)|0);
       $189 = ($188|0)==(-1);
       if ($189) {
        $$1 = 0;$$2 = -128;
        break L21;
       }
       $190 = $99;
       $191 = $190;
       HEAP32[$191>>2] = $182;
       $192 = (($190) + 4)|0;
       $193 = $192;
       HEAP32[$193>>2] = $184;
       (_ogg_sync_reset($101)|0);
      }
      $194 = (__get_next_page($vf,$og,-1,-1)|0);
      $195 = tempRet0;
      $196 = ($194|0)==(-128);
      $197 = ($195|0)==(-1);
      $198 = $196 & $197;
      if ($198) {
       $$1 = 0;$$2 = -128;
       break L21;
      }
      $199 = ($195|0)<(0);
      if (!($199)) {
       $200 = (_ogg_page_serialno($og)|0);
       $$013$i$i = $currentno_list;$$in$i$i = $currentnos;
       while(1) {
        $203 = (($$in$i$i) + -1)|0;
        $204 = HEAP32[$$013$i$i>>2]|0;
        $205 = ($204|0)==($200|0);
        if ($205) {
         $227 = $159;$229 = $158;$356 = $212;$357 = $214;
         break L25;
        }
        $201 = ((($$013$i$i)) + 4|0);
        $202 = ($203|0)==(0);
        if ($202) {
         break;
        } else {
         $$013$i$i = $201;$$in$i$i = $203;
        }
       }
      }
      $206 = ($195|0)>(-1);
      $207 = ($194>>>0)>(4294967295);
      $208 = ($195|0)==(-1);
      $209 = $208 & $207;
      $210 = $206 | $209;
      $211 = $210 ? $194 : $212;
      $213 = $210 ? $195 : $214;
      $215 = ($184|0)>($161|0);
      $216 = ($182>>>0)>($160>>>0);
      $217 = ($184|0)==($161|0);
      $218 = $217 & $216;
      $219 = $215 | $218;
      if ($219) {
       $158 = $182;$159 = $184;$212 = $211;$214 = $213;
      } else {
       $249 = $211;$251 = $213;
       label = 28;
       break L21;
      }
     }
     $220 = $99;
     $221 = $220;
     $222 = HEAP32[$221>>2]|0;
     $223 = (($220) + 4)|0;
     $224 = $223;
     $225 = HEAP32[$224>>2]|0;
     $226 = ($227|0)>($225|0);
     $228 = ($229>>>0)>($222>>>0);
     $230 = ($227|0)==($225|0);
     $231 = $230 & $228;
     $232 = $226 | $231;
     if ($232) {
      $160 = $222;$161 = $225;$348 = $356;$349 = $357;$350 = $229;$351 = $227;
     } else {
      $249 = $356;$251 = $357;
      label = 28;
      break L21;
     }
    }
    $102 = $352;$103 = $353;$150 = $354;$152 = $355;
    while(1) {
     $106 = (_i64Subtract(($102|0),($103|0),($104|0),($105|0))|0);
     $107 = tempRet0;
     $108 = ($107|0)<(0);
     $109 = ($106>>>0)<(65536);
     $110 = ($107|0)==(0);
     $111 = $110 & $109;
     $112 = $108 | $111;
     if ($112) {
      $126 = $104;$128 = $105;
     } else {
      $113 = (_i64Add(($102|0),($103|0),($104|0),($105|0))|0);
      $114 = tempRet0;
      $115 = (___divdi3(($113|0),($114|0),2,0)|0);
      $116 = tempRet0;
      $126 = $115;$128 = $116;
     }
     $117 = HEAP32[$vf>>2]|0;
     $118 = ($117|0)==(0|0);
     if ($118) {
      $$1 = 0;$$2 = -129;
      break L21;
     }
     $119 = $99;
     $120 = $119;
     $121 = HEAP32[$120>>2]|0;
     $122 = (($119) + 4)|0;
     $123 = $122;
     $124 = HEAP32[$123>>2]|0;
     $125 = ($121|0)==($126|0);
     $127 = ($124|0)==($128|0);
     $129 = $125 & $127;
     if (!($129)) {
      $130 = HEAP32[$100>>2]|0;
      $131 = ($130|0)==(0|0);
      if ($131) {
       $$1 = 0;$$2 = -128;
       break L21;
      }
      $132 = (FUNCTION_TABLE_iiiii[$130 & 8191]($117,$126,$128,0)|0);
      $133 = ($132|0)==(-1);
      if ($133) {
       $$1 = 0;$$2 = -128;
       break L21;
      }
      $134 = $99;
      $135 = $134;
      HEAP32[$135>>2] = $126;
      $136 = (($134) + 4)|0;
      $137 = $136;
      HEAP32[$137>>2] = $128;
      (_ogg_sync_reset($101)|0);
     }
     $138 = (__get_next_page($vf,$og,-1,-1)|0);
     $139 = tempRet0;
     $140 = ($138|0)==(-128);
     $141 = ($139|0)==(-1);
     $142 = $140 & $141;
     if ($142) {
      $$1 = 0;$$2 = -128;
      break L21;
     }
     $143 = ($139|0)<(0);
     if (!($143)) {
     }
     $144 = ($139|0)>(-1);
     $145 = ($138>>>0)>(4294967295);
     $146 = ($139|0)==(-1);
     $147 = $146 & $145;
     $148 = $144 | $147;
     $149 = $148 ? $138 : $150;
     $151 = $148 ? $139 : $152;
     $153 = ($128|0)>($105|0);
     $154 = ($126>>>0)>($104>>>0);
     $155 = ($128|0)==($105|0);
     $156 = $155 & $154;
     $157 = $153 | $156;
     if ($157) {
      $102 = $126;$103 = $128;$150 = $149;$152 = $151;
     } else {
      $249 = $149;$251 = $151;
      label = 28;
      break;
     }
    }
   } else {
    $249 = $4;$251 = $5;
    label = 28;
   }
  } while(0);
  do {
   if ((label|0) == 28) {
    $233 = $249;$234 = $251;
    while(1) {
     HEAP32[$testserial>>2] = $19;
     $235 = (__get_prev_page_serial($vf,$233,$234,$currentno_list,$currentnos,$testserial,$searchgran)|0);
     $236 = tempRet0;
     $237 = HEAP32[$testserial>>2]|0;
     $238 = ($237|0)==($19|0);
     if ($238) {
      break;
     } else {
      $233 = $235;$234 = $236;
     }
    }
    $239 = HEAP32[$vf>>2]|0;
    $240 = ($239|0)==(0|0);
    if ($240) {
     $$1 = 0;$$2 = -129;
    } else {
     $241 = ((($vf)) + 8|0);
     $242 = $241;
     $243 = $242;
     $244 = HEAP32[$243>>2]|0;
     $245 = (($242) + 4)|0;
     $246 = $245;
     $247 = HEAP32[$246>>2]|0;
     $248 = ($244|0)==($249|0);
     $250 = ($247|0)==($251|0);
     $252 = $248 & $250;
     if (!($252)) {
      $253 = ((($vf)) + 708|0);
      $254 = HEAP32[$253>>2]|0;
      $255 = ($254|0)==(0|0);
      if ($255) {
       $$1 = 0;$$2 = -128;
       break;
      }
      $256 = (FUNCTION_TABLE_iiiii[$254 & 8191]($239,$249,$251,0)|0);
      $257 = ($256|0)==(-1);
      if ($257) {
       $$1 = 0;$$2 = -128;
       break;
      }
      $258 = $241;
      $259 = $258;
      HEAP32[$259>>2] = $249;
      $260 = (($258) + 4)|0;
      $261 = $260;
      HEAP32[$261>>2] = $251;
      $262 = ((($vf)) + 24|0);
      (_ogg_sync_reset($262)|0);
     }
     $263 = (__fetch_headers($vf,$vi,$vc,$next_serialno_list,$next_serialnos,0)|0);
     $264 = ($263|0)==(0);
     if ($264) {
      $265 = HEAP32[$18>>2]|0;
      $266 = $241;
      $267 = $266;
      $268 = HEAP32[$267>>2]|0;
      $269 = (($266) + 4)|0;
      $270 = $269;
      $271 = HEAP32[$270>>2]|0;
      $272 = (__initial_pcmoffset($vf,$vi)|0);
      $273 = tempRet0;
      $274 = $241;
      $275 = $274;
      $276 = HEAP32[$275>>2]|0;
      $277 = (($274) + 4)|0;
      $278 = $277;
      $279 = HEAP32[$278>>2]|0;
      $280 = HEAP32[$next_serialno_list>>2]|0;
      $281 = HEAP32[$next_serialnos>>2]|0;
      $282 = (($m) + 1)|0;
      $283 = (__bisect_forward_serialno($vf,$249,$251,$276,$279,$4,$5,$6,$7,$endserial,$280,$281,$282)|0);
      $284 = ($283|0)==(0);
      if ($284) {
       $285 = HEAP32[$next_serialno_list>>2]|0;
       $286 = ($285|0)==(0|0);
       if (!($286)) {
        _free($285);
       }
       $287 = ((($vf)) + 56|0);
       $288 = HEAP32[$287>>2]|0;
       $289 = (($288) + ($282<<3)|0);
       $290 = $289;
       $291 = $290;
       HEAP32[$291>>2] = $249;
       $292 = (($290) + 4)|0;
       $293 = $292;
       HEAP32[$293>>2] = $251;
       $294 = ((($vf)) + 64|0);
       $295 = HEAP32[$294>>2]|0;
       $296 = (($295) + ($282<<2)|0);
       HEAP32[$296>>2] = $265;
       $297 = ((($vf)) + 60|0);
       $298 = HEAP32[$297>>2]|0;
       $299 = (($298) + ($282<<3)|0);
       $300 = $299;
       $301 = $300;
       HEAP32[$301>>2] = $268;
       $302 = (($300) + 4)|0;
       $303 = $302;
       HEAP32[$303>>2] = $271;
       $304 = ((($vf)) + 72|0);
       $305 = HEAP32[$304>>2]|0;
       $306 = (($305) + ($282<<5)|0);
       ;HEAP32[$306>>2]=HEAP32[$vi>>2]|0;HEAP32[$306+4>>2]=HEAP32[$vi+4>>2]|0;HEAP32[$306+8>>2]=HEAP32[$vi+8>>2]|0;HEAP32[$306+12>>2]=HEAP32[$vi+12>>2]|0;HEAP32[$306+16>>2]=HEAP32[$vi+16>>2]|0;HEAP32[$306+20>>2]=HEAP32[$vi+20>>2]|0;HEAP32[$306+24>>2]=HEAP32[$vi+24>>2]|0;HEAP32[$306+28>>2]=HEAP32[$vi+28>>2]|0;
       $307 = ((($vf)) + 76|0);
       $308 = HEAP32[$307>>2]|0;
       $309 = (($308) + ($282<<4)|0);
       ;HEAP32[$309>>2]=HEAP32[$vc>>2]|0;HEAP32[$309+4>>2]=HEAP32[$vc+4>>2]|0;HEAP32[$309+8>>2]=HEAP32[$vc+8>>2]|0;HEAP32[$309+12>>2]=HEAP32[$vc+12>>2]|0;
       $310 = $searchgran;
       $311 = $310;
       $312 = HEAP32[$311>>2]|0;
       $313 = (($310) + 4)|0;
       $314 = $313;
       $315 = HEAP32[$314>>2]|0;
       $316 = $m << 1;
       $317 = $316 | 1;
       $318 = ((($vf)) + 68|0);
       $319 = HEAP32[$318>>2]|0;
       $320 = (($319) + ($317<<3)|0);
       $321 = $320;
       $322 = $321;
       HEAP32[$322>>2] = $312;
       $323 = (($321) + 4)|0;
       $324 = $323;
       HEAP32[$324>>2] = $315;
       $325 = (($316) + 2)|0;
       $326 = (($319) + ($325<<3)|0);
       $327 = $326;
       $328 = $327;
       HEAP32[$328>>2] = $272;
       $329 = (($327) + 4)|0;
       $330 = $329;
       HEAP32[$330>>2] = $273;
       $331 = (($316) + 3)|0;
       $332 = (($319) + ($331<<3)|0);
       $333 = $332;
       $334 = $333;
       $335 = HEAP32[$334>>2]|0;
       $336 = (($333) + 4)|0;
       $337 = $336;
       $338 = HEAP32[$337>>2]|0;
       $339 = (_i64Subtract(($335|0),($338|0),($272|0),($273|0))|0);
       $340 = tempRet0;
       $341 = ($340|0)<(0);
       $342 = $341 ? 0 : $339;
       $343 = $341 ? 0 : $340;
       $344 = $332;
       $345 = $344;
       HEAP32[$345>>2] = $342;
       $346 = (($344) + 4)|0;
       $347 = $346;
       HEAP32[$347>>2] = $343;
       $$1 = 1;$$2 = 0;
      } else {
       $$1 = 0;$$2 = $283;
      }
     } else {
      $$1 = 0;$$2 = $263;
     }
    }
   }
  } while(0);
  if (!($$1)) {
   $$3 = $$2;
   STACKTOP = sp;return ($$3|0);
  }
 }
 $$3 = 0;
 STACKTOP = sp;return ($$3|0);
}
function _ov_raw_seek($vf,$0,$1) {
 $vf = $vf|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$lcssa12 = 0, $$lcssa13 = 0, $$lcssa93 = 0, $$pr$us = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $accblock$0$ph = 0, $accblock$1 = 0, $accblock$1$lcssa = 0, $exitcond = 0, $firstflag$0$ph = 0, $firstflag$0$ph9$lcssa = 0, $firstflag$0$ph971 = 0, $firstflag$0$ph974 = 0, $firstflag$2 = 0;
 var $i$023 = 0, $lastblock$0$ph = 0, $lastflag$0$ph = 0, $lastflag$0$ph8$lcssa = 0, $lastflag$0$ph868 = 0, $lastflag$0$ph873 = 0, $link1$0$lcssa$us = 0, $link1$0$lcssa$us$lcssa = 0, $link1$025$us = 0, $og = 0, $op = 0, $or$cond = 0, $thisblock$0 = 0, $work_os = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 416|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $work_os = sp + 32|0;
 $og = sp + 392|0;
 $op = sp;
 $2 = ((($vf)) + 88|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)<(2);
 if ($4) {
  $$0 = -131;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = ((($vf)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if ($7) {
  $$0 = -138;
  STACKTOP = sp;return ($$0|0);
 }
 $8 = ($1|0)<(0);
 if ($8) {
  $$0 = -131;
  STACKTOP = sp;return ($$0|0);
 }
 $9 = ((($vf)) + 16|0);
 $10 = $9;
 $11 = $10;
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10) + 4)|0;
 $14 = $13;
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)<($1|0);
 $17 = ($12>>>0)<($0>>>0);
 $18 = ($15|0)==($1|0);
 $19 = $18 & $17;
 $20 = $16 | $19;
 if ($20) {
  $$0 = -131;
  STACKTOP = sp;return ($$0|0);
 }
 $21 = ($3|0)>(2);
 do {
  if ($21) {
   $22 = ((($vf)) + 96|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = ((($vf)) + 56|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = (($25) + ($23<<3)|0);
   $27 = $26;
   $28 = $27;
   $29 = HEAP32[$28>>2]|0;
   $30 = (($27) + 4)|0;
   $31 = $30;
   $32 = HEAP32[$31>>2]|0;
   $33 = ($32|0)>($1|0);
   $34 = ($29>>>0)>($0>>>0);
   $35 = ($32|0)==($1|0);
   $36 = $35 & $34;
   $37 = $33 | $36;
   if (!($37)) {
    $38 = (($23) + 1)|0;
    $39 = (($25) + ($38<<3)|0);
    $40 = $39;
    $41 = $40;
    $42 = HEAP32[$41>>2]|0;
    $43 = (($40) + 4)|0;
    $44 = $43;
    $45 = HEAP32[$44>>2]|0;
    $46 = ($45|0)>($1|0);
    $47 = ($42>>>0)>($0>>>0);
    $48 = ($45|0)==($1|0);
    $49 = $48 & $47;
    $50 = $46 | $49;
    if ($50) {
     break;
    }
   }
   $51 = ((($vf)) + 480|0);
   _vorbis_dsp_clear($51);
   $52 = ((($vf)) + 592|0);
   (_vorbis_block_clear($52)|0);
   HEAP32[$2>>2] = 2;
  }
 } while(0);
 $53 = ((($vf)) + 80|0);
 $54 = $53;
 $55 = $54;
 HEAP32[$55>>2] = -1;
 $56 = (($54) + 4)|0;
 $57 = $56;
 HEAP32[$57>>2] = -1;
 $58 = ((($vf)) + 120|0);
 $59 = ((($vf)) + 92|0);
 $60 = HEAP32[$59>>2]|0;
 (_ogg_stream_reset_serialno($58,$60)|0);
 $61 = ((($vf)) + 480|0);
 (_vorbis_synthesis_restart($61)|0);
 $62 = HEAP32[$vf>>2]|0;
 $63 = ($62|0)==(0|0);
 do {
  if (!($63)) {
   $64 = ((($vf)) + 8|0);
   $65 = $64;
   $66 = $65;
   $67 = HEAP32[$66>>2]|0;
   $68 = (($65) + 4)|0;
   $69 = $68;
   $70 = HEAP32[$69>>2]|0;
   $71 = ($67|0)==($0|0);
   $72 = ($70|0)==($1|0);
   $73 = $71 & $72;
   if (!($73)) {
    $74 = ((($vf)) + 708|0);
    $75 = HEAP32[$74>>2]|0;
    $76 = ($75|0)==(0|0);
    if ($76) {
     break;
    }
    $77 = (FUNCTION_TABLE_iiiii[$75 & 8191]($62,$0,$1,0)|0);
    $78 = ($77|0)==(-1);
    if ($78) {
     break;
    }
    $79 = $64;
    $80 = $79;
    HEAP32[$80>>2] = $0;
    $81 = (($79) + 4)|0;
    $82 = $81;
    HEAP32[$82>>2] = $1;
    $83 = ((($vf)) + 24|0);
    (_ogg_sync_reset($83)|0);
   }
   $84 = HEAP32[$59>>2]|0;
   (_ogg_stream_init($work_os,$84)|0);
   (_ogg_stream_reset($work_os)|0);
   $85 = ((($vf)) + 96|0);
   $86 = ((($vf)) + 72|0);
   $87 = ((($vf)) + 592|0);
   $88 = ((($vf)) + 64|0);
   $89 = ((($vf)) + 52|0);
   $90 = ((($vf)) + 60|0);
   $91 = ((($op)) + 16|0);
   $accblock$0$ph = 0;$firstflag$0$ph = 0;$lastblock$0$ph = 0;$lastflag$0$ph = 0;
   L26: while(1) {
    $92 = ($lastblock$0$ph|0)==(0);
    L28: do {
     if ($92) {
      $firstflag$0$ph974 = $firstflag$0$ph;$lastflag$0$ph873 = $lastflag$0$ph;
      while(1) {
       while(1) {
        $93 = HEAP32[$2>>2]|0;
        $94 = ($93|0)>(2);
        if ($94) {
         $95 = (_ogg_stream_packetout($work_os,$op)|0);
         $96 = ($95|0)>(0);
         if ($96) {
          $97 = HEAP32[$85>>2]|0;
          $98 = HEAP32[$86>>2]|0;
          $99 = (((($98) + ($97<<5)|0)) + 28|0);
          $100 = HEAP32[$99>>2]|0;
          $101 = ($100|0)==(0|0);
          if (!($101)) {
           $$lcssa12 = $97;$$lcssa13 = $98;$firstflag$0$ph971 = $firstflag$0$ph974;$lastflag$0$ph868 = $lastflag$0$ph873;
           break L28;
          }
          (_ogg_stream_packetout($58,0)|0);
         }
        }
        $102 = (__get_next_page($vf,$og,-1,-1)|0);
        $103 = tempRet0;
        $104 = ($103|0)<(0);
        if ($104) {
         label = 46;
         break L26;
        }
        $105 = HEAP32[$2>>2]|0;
        $106 = ($105|0)>(2);
        if ($106) {
         $107 = HEAP32[$59>>2]|0;
         $108 = (_ogg_page_serialno($og)|0);
         $109 = ($107|0)==($108|0);
         if (!($109)) {
          $110 = (_ogg_page_bos($og)|0);
          $111 = ($110|0)==(0);
          if (!($111)) {
           _vorbis_dsp_clear($61);
           (_vorbis_block_clear($87)|0);
           HEAP32[$2>>2] = 2;
           (_ogg_stream_clear($work_os)|0);
          }
         }
         $$pr$us = HEAP32[$2>>2]|0;
         $112 = ($$pr$us|0)<(3);
         if (!($112)) {
          $firstflag$2 = $firstflag$0$ph974;
          break;
         }
        }
        $113 = (_ogg_page_serialno($og)|0);
        $114 = HEAP32[$89>>2]|0;
        $115 = ($114|0)>(0);
        L46: do {
         if ($115) {
          $117 = HEAP32[$88>>2]|0;
          $link1$025$us = 0;
          while(1) {
           $116 = (($117) + ($link1$025$us<<2)|0);
           $118 = HEAP32[$116>>2]|0;
           $119 = ($118|0)==($113|0);
           if ($119) {
            $link1$0$lcssa$us = $link1$025$us;
            break L46;
           }
           $120 = (($link1$025$us) + 1)|0;
           $121 = ($120|0)<($114|0);
           if ($121) {
            $link1$025$us = $120;
           } else {
            $link1$0$lcssa$us = $120;
            break;
           }
          }
         } else {
          $link1$0$lcssa$us = 0;
         }
        } while(0);
        $122 = ($link1$0$lcssa$us|0)==($114|0);
        if (!($122)) {
         $$lcssa93 = $113;$214 = $103;$216 = $102;$link1$0$lcssa$us$lcssa = $link1$0$lcssa$us;
         label = 48;
         break;
        }
       }
       if ((label|0) == 48) {
        label = 0;
        HEAP32[$85>>2] = $link1$0$lcssa$us$lcssa;
        HEAP32[$59>>2] = $$lcssa93;
        (_ogg_stream_reset_serialno($58,$$lcssa93)|0);
        (_ogg_stream_reset_serialno($work_os,$$lcssa93)|0);
        HEAP32[$2>>2] = 3;
        $205 = HEAP32[$90>>2]|0;
        $206 = (($205) + ($link1$0$lcssa$us$lcssa<<3)|0);
        $207 = $206;
        $208 = $207;
        $209 = HEAP32[$208>>2]|0;
        $210 = (($207) + 4)|0;
        $211 = $210;
        $212 = HEAP32[$211>>2]|0;
        $213 = ($214|0)<($212|0);
        $215 = ($216>>>0)<=($209>>>0);
        $217 = ($214|0)==($212|0);
        $218 = $217 & $215;
        $219 = $213 | $218;
        $220 = $219&1;
        $firstflag$2 = $220;
       }
       (_ogg_stream_pagein($58,$og)|0);
       (_ogg_stream_pagein($work_os,$og)|0);
       $221 = (_ogg_page_eos($og)|0);
       if ($92) {
        $firstflag$0$ph974 = $firstflag$2;$lastflag$0$ph873 = $221;
       } else {
        $firstflag$0$ph9$lcssa = $firstflag$2;$lastflag$0$ph8$lcssa = $221;
        label = 32;
        break;
       }
      }
     } else {
      $firstflag$0$ph9$lcssa = $firstflag$0$ph;$lastflag$0$ph8$lcssa = $lastflag$0$ph;
      label = 32;
     }
    } while(0);
    if ((label|0) == 32) {
     label = 0;
     $123 = HEAP32[$2>>2]|0;
     $124 = ($123|0)>(2);
     if (!($124)) {
      label = 47;
      break;
     }
     $125 = (_ogg_stream_packetout($work_os,$op)|0);
     $126 = ($125|0)>(0);
     if (!($126)) {
      label = 47;
      break;
     }
     $127 = HEAP32[$85>>2]|0;
     $128 = HEAP32[$86>>2]|0;
     $129 = (((($128) + ($127<<5)|0)) + 28|0);
     $130 = HEAP32[$129>>2]|0;
     $131 = ($130|0)==(0|0);
     if ($131) {
      label = 45;
      break;
     } else {
      $$lcssa12 = $127;$$lcssa13 = $128;$firstflag$0$ph971 = $firstflag$0$ph9$lcssa;$lastflag$0$ph868 = $lastflag$0$ph8$lcssa;
     }
    }
    $132 = (($$lcssa13) + ($$lcssa12<<5)|0);
    $133 = (_vorbis_packet_blocksize($132,$op)|0);
    $134 = ($133|0)<(0);
    do {
     if ($134) {
      (_ogg_stream_packetout($58,0)|0);
      $accblock$1 = $accblock$0$ph;$thisblock$0 = 0;
     } else {
      $135 = ($lastflag$0$ph868|0)==(0);
      $136 = ($firstflag$0$ph971|0)!=(0);
      $or$cond = $135 | $136;
      if (!($or$cond)) {
       (_ogg_stream_packetout($58,0)|0);
       $accblock$1 = $accblock$0$ph;$thisblock$0 = $133;
       break;
      }
      if ($92) {
       $accblock$1 = $accblock$0$ph;$thisblock$0 = $133;
      } else {
       $137 = (($133) + ($lastblock$0$ph))|0;
       $138 = $137 >> 2;
       $139 = (($138) + ($accblock$0$ph))|0;
       $accblock$1 = $139;$thisblock$0 = $133;
      }
     }
    } while(0);
    $140 = $91;
    $141 = $140;
    $142 = HEAP32[$141>>2]|0;
    $143 = (($140) + 4)|0;
    $144 = $143;
    $145 = HEAP32[$144>>2]|0;
    $146 = ($142|0)==(-1);
    $147 = ($145|0)==(-1);
    $148 = $146 & $147;
    if ($148) {
     $accblock$0$ph = $accblock$1;$firstflag$0$ph = $firstflag$0$ph971;$lastblock$0$ph = $thisblock$0;$lastflag$0$ph = $lastflag$0$ph868;
    } else {
     $160 = $142;$161 = $145;$accblock$1$lcssa = $accblock$1;
     label = 42;
     break;
    }
   }
   if ((label|0) == 42) {
    $149 = HEAP32[$85>>2]|0;
    $150 = $149 << 1;
    $151 = ((($vf)) + 68|0);
    $152 = HEAP32[$151>>2]|0;
    $153 = (($152) + ($150<<3)|0);
    $154 = $153;
    $155 = $154;
    $156 = HEAP32[$155>>2]|0;
    $157 = (($154) + 4)|0;
    $158 = $157;
    $159 = HEAP32[$158>>2]|0;
    $162 = (_i64Subtract(($160|0),($161|0),($156|0),($159|0))|0);
    $163 = tempRet0;
    $164 = ($163|0)<(0);
    $165 = $164 ? 0 : $162;
    $166 = $164 ? 0 : $163;
    $167 = ($149|0)>(0);
    if ($167) {
     $177 = $165;$178 = $166;$i$023 = 0;
     while(1) {
      $168 = $i$023 << 1;
      $169 = $168 | 1;
      $170 = (($152) + ($169<<3)|0);
      $171 = $170;
      $172 = $171;
      $173 = HEAP32[$172>>2]|0;
      $174 = (($171) + 4)|0;
      $175 = $174;
      $176 = HEAP32[$175>>2]|0;
      $179 = (_i64Add(($173|0),($176|0),($177|0),($178|0))|0);
      $180 = tempRet0;
      $181 = (($i$023) + 1)|0;
      $exitcond = ($181|0)==($149|0);
      if ($exitcond) {
       $184 = $179;$185 = $180;
       break;
      } else {
       $177 = $179;$178 = $180;$i$023 = $181;
      }
     }
    } else {
     $184 = $165;$185 = $166;
    }
    $182 = ($accblock$1$lcssa|0)<(0);
    $183 = $182 << 31 >> 31;
    $186 = (_i64Subtract(($184|0),($185|0),($accblock$1$lcssa|0),($183|0))|0);
    $187 = tempRet0;
    $188 = ($187|0)<(0);
    $189 = $188 ? 0 : $186;
    $190 = $188 ? 0 : $187;
    $191 = $53;
    $192 = $191;
    HEAP32[$192>>2] = $189;
    $193 = (($191) + 4)|0;
    $194 = $193;
    HEAP32[$194>>2] = $190;
   }
   else if ((label|0) == 45) {
    (_ogg_stream_packetout($58,0)|0);
    label = 47;
   }
   else if ((label|0) == 46) {
    $195 = (_ov_pcm_total($vf,-1)|0);
    $196 = tempRet0;
    $197 = $53;
    $198 = $197;
    HEAP32[$198>>2] = $195;
    $199 = (($197) + 4)|0;
    $200 = $199;
    HEAP32[$200>>2] = $196;
   }
   if ((label|0) == 47) {
    $201 = $53;
    $202 = $201;
    HEAP32[$202>>2] = -1;
    $203 = (($201) + 4)|0;
    $204 = $203;
    HEAP32[$204>>2] = -1;
   }
   (_ogg_stream_clear($work_os)|0);
   $222 = ((($vf)) + 104|0);
   ;HEAP32[$222>>2]=0|0;HEAP32[$222+4>>2]=0|0;HEAP32[$222+8>>2]=0|0;HEAP32[$222+12>>2]=0|0;
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 $223 = $53;
 $224 = $223;
 HEAP32[$224>>2] = -1;
 $225 = (($223) + 4)|0;
 $226 = $225;
 HEAP32[$226>>2] = -1;
 (_ogg_stream_clear($work_os)|0);
 _vorbis_dsp_clear($61);
 $227 = ((($vf)) + 592|0);
 (_vorbis_block_clear($227)|0);
 HEAP32[$2>>2] = 2;
 $$0 = -137;
 STACKTOP = sp;return ($$0|0);
}
function _ov_pcm_total($vf,$i) {
 $vf = $vf|0;
 $i = $i|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i1$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($vf)) + 88|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)<(2);
 do {
  if ($2) {
   $30 = -1;$31 = -131;
  } else {
   $3 = ((($vf)) + 4|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==(0);
   if ($5) {
    $30 = -1;$31 = -131;
   } else {
    $6 = ((($vf)) + 52|0);
    $7 = HEAP32[$6>>2]|0;
    $8 = ($7|0)>($i|0);
    if ($8) {
     $9 = ($i|0)<(0);
     if (!($9)) {
      $19 = $i << 1;
      $20 = $19 | 1;
      $21 = ((($vf)) + 68|0);
      $22 = HEAP32[$21>>2]|0;
      $23 = (($22) + ($20<<3)|0);
      $24 = $23;
      $25 = $24;
      $26 = HEAP32[$25>>2]|0;
      $27 = (($24) + 4)|0;
      $28 = $27;
      $29 = HEAP32[$28>>2]|0;
      $30 = $29;$31 = $26;
      break;
     }
     $10 = ($7|0)>(0);
     if ($10) {
      $13 = 0;$14 = 0;$i1$02 = 0;
      while(1) {
       $11 = (_ov_pcm_total($vf,$i1$02)|0);
       $12 = tempRet0;
       $15 = (_i64Add(($11|0),($12|0),($13|0),($14|0))|0);
       $16 = tempRet0;
       $17 = (($i1$02) + 1)|0;
       $18 = ($17|0)<($7|0);
       if ($18) {
        $13 = $15;$14 = $16;$i1$02 = $17;
       } else {
        $30 = $16;$31 = $15;
        break;
       }
      }
     } else {
      $30 = 0;$31 = 0;
     }
    } else {
     $30 = -1;$31 = -131;
    }
   }
  }
 } while(0);
 tempRet0 = ($30);
 return ($31|0);
}
function _ov_pcm_seek($vf,$0,$1) {
 $vf = $vf|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i8$ph = 0, $$5 = 0, $$7 = 0, $$lcssa = 0, $$pr = 0, $$pre$phi54Z2D = 0, $$pre$phi56Z2D = 0, $$pre$phi58Z2D = 0, $$pre$phi60Z2D = 0, $$pre$phi62Z2D = 0, $$pre$phiZ2D = 0, $$pre52 = 0, $$pre53 = 0, $$pre55 = 0, $$pre57 = 0, $$pre59 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0;
 var $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0;
 var $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0;
 var $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$044 = 0, $lastblock$0$ph = 0, $lastblock$2 = 0, $lastblock$4$ph = 0, $link$0$lcssa = 0, $link$0$lcssa$lcssa = 0, $link$034 = 0, $og = 0, $op = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $op = sp;
 $og = sp + 32|0;
 $2 = (_ov_pcm_seek_page($vf,$0,$1)|0);
 $3 = ($2|0)<(0);
 if ($3) {
  $$5 = $2;
  STACKTOP = sp;return ($$5|0);
 }
 $4 = ((($vf)) + 88|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)>(3);
 if ($6) {
  $$pre52 = ((($vf)) + 72|0);
  $$pre53 = ((($vf)) + 592|0);
  $$pre55 = ((($vf)) + 480|0);
  $$pre57 = ((($vf)) + 4|0);
  $$pre59 = ((($vf)) + 104|0);
  $$pre$phi54Z2D = $$pre53;$$pre$phi56Z2D = $$pre55;$$pre$phi58Z2D = $$pre57;$$pre$phi60Z2D = $$pre59;$$pre$phi62Z2D = $$pre59;$$pre$phiZ2D = $$pre52;
 } else {
  $7 = ($5|0)<(3);
  if ($7) {
   $$5 = -129;
   STACKTOP = sp;return ($$5|0);
  }
  $8 = ((($vf)) + 4|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)==(0);
  $11 = ((($vf)) + 480|0);
  $12 = ((($vf)) + 72|0);
  $13 = HEAP32[$12>>2]|0;
  if ($10) {
   $19 = (_vorbis_synthesis_init($11,$13)|0);
   $20 = ($19|0)==(0);
   if (!($20)) {
    $$5 = -137;
    STACKTOP = sp;return ($$5|0);
   }
  } else {
   $14 = ((($vf)) + 96|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = (($13) + ($15<<5)|0);
   $17 = (_vorbis_synthesis_init($11,$16)|0);
   $18 = ($17|0)==(0);
   if (!($18)) {
    $$5 = -137;
    STACKTOP = sp;return ($$5|0);
   }
  }
  $21 = ((($vf)) + 592|0);
  (_vorbis_block_init($11,$21)|0);
  HEAP32[$4>>2] = 4;
  $22 = ((($vf)) + 104|0);
  ;HEAP32[$22>>2]=0|0;HEAP32[$22+4>>2]=0|0;HEAP32[$22+8>>2]=0|0;HEAP32[$22+12>>2]=0|0;
  $$pre$phi54Z2D = $21;$$pre$phi56Z2D = $11;$$pre$phi58Z2D = $8;$$pre$phi60Z2D = $22;$$pre$phi62Z2D = $22;$$pre$phiZ2D = $12;
 }
 $23 = ((($vf)) + 120|0);
 $24 = ((($vf)) + 96|0);
 $25 = ((($vf)) + 80|0);
 $26 = ((($op)) + 16|0);
 $27 = ((($vf)) + 68|0);
 $28 = ((($vf)) + 64|0);
 $29 = ((($vf)) + 52|0);
 $30 = ((($vf)) + 92|0);
 $lastblock$0$ph = 0;
 L17: while(1) {
  $31 = (_ogg_stream_packetpeek($23,$op)|0);
  $32 = ($31|0)>(0);
  L19: do {
   if ($32) {
    label = 11;
   } else {
    $127 = $31;
    while(1) {
     $126 = ($127|0)<(0);
     $128 = ($127|0)!=(-3);
     $or$cond = $126 & $128;
     if ($or$cond) {
      break L17;
     }
     (__get_next_page($vf,$og,-1,-1)|0);
     $129 = tempRet0;
     $130 = ($129|0)<(0);
     if ($130) {
      break L17;
     }
     $131 = (_ogg_page_bos($og)|0);
     $132 = ($131|0)==(0);
     if ($132) {
      $$pr = HEAP32[$4>>2]|0;
      $133 = ($$pr|0)<(3);
      if (!($133)) {
       $lastblock$2 = $lastblock$0$ph;
       break;
      }
     } else {
      _vorbis_dsp_clear($$pre$phi56Z2D);
      (_vorbis_block_clear($$pre$phi54Z2D)|0);
      HEAP32[$4>>2] = 2;
     }
     $134 = (_ogg_page_serialno($og)|0);
     $135 = HEAP32[$29>>2]|0;
     $136 = ($135|0)>(0);
     L28: do {
      if ($136) {
       $137 = HEAP32[$28>>2]|0;
       $link$034 = 0;
       while(1) {
        $138 = (($137) + ($link$034<<2)|0);
        $139 = HEAP32[$138>>2]|0;
        $140 = ($139|0)==($134|0);
        if ($140) {
         $link$0$lcssa = $link$034;
         break L28;
        }
        $141 = (($link$034) + 1)|0;
        $142 = ($141|0)<($135|0);
        if ($142) {
         $link$034 = $141;
        } else {
         $link$0$lcssa = $141;
         break;
        }
       }
      } else {
       $link$0$lcssa = 0;
      }
     } while(0);
     $143 = ($link$0$lcssa|0)==($135|0);
     if (!($143)) {
      $$lcssa = $134;$link$0$lcssa$lcssa = $link$0$lcssa;
      label = 30;
      break;
     }
     $157 = (_ogg_stream_packetpeek($23,$op)|0);
     $158 = ($157|0)>(0);
     if ($158) {
      label = 11;
      break L19;
     } else {
      $127 = $157;
     }
    }
    if ((label|0) == 30) {
     label = 0;
     HEAP32[$24>>2] = $link$0$lcssa$lcssa;
     HEAP32[$4>>2] = 3;
     $144 = (_ogg_page_serialno($og)|0);
     HEAP32[$30>>2] = $144;
     (_ogg_stream_reset_serialno($23,$$lcssa)|0);
     $145 = HEAP32[$4>>2]|0;
     $146 = ($145|0)>(3);
     if ($146) {
      $lastblock$2 = 0;
     } else {
      $147 = ($145|0)<(3);
      if ($147) {
       $$0$i8$ph = -129;
       label = 36;
       break L17;
      }
      $148 = HEAP32[$$pre$phi58Z2D>>2]|0;
      $149 = ($148|0)==(0);
      $150 = HEAP32[$$pre$phiZ2D>>2]|0;
      if ($149) {
       $155 = (_vorbis_synthesis_init($$pre$phi56Z2D,$150)|0);
       $156 = ($155|0)==(0);
       if (!($156)) {
        $$0$i8$ph = -137;
        label = 36;
        break L17;
       }
      } else {
       $151 = HEAP32[$24>>2]|0;
       $152 = (($150) + ($151<<5)|0);
       $153 = (_vorbis_synthesis_init($$pre$phi56Z2D,$152)|0);
       $154 = ($153|0)==(0);
       if (!($154)) {
        $$0$i8$ph = -137;
        label = 36;
        break L17;
       }
      }
      (_vorbis_block_init($$pre$phi56Z2D,$$pre$phi54Z2D)|0);
      HEAP32[$4>>2] = 4;
      ;HEAP32[$$pre$phi62Z2D>>2]=0|0;HEAP32[$$pre$phi62Z2D+4>>2]=0|0;HEAP32[$$pre$phi62Z2D+8>>2]=0|0;HEAP32[$$pre$phi62Z2D+12>>2]=0|0;
      $lastblock$2 = 0;
     }
    }
    (_ogg_stream_pagein($23,$og)|0);
    $lastblock$4$ph = $lastblock$2;
   }
  } while(0);
  do {
   if ((label|0) == 11) {
    label = 0;
    $33 = HEAP32[$$pre$phiZ2D>>2]|0;
    $34 = HEAP32[$24>>2]|0;
    $35 = (($33) + ($34<<5)|0);
    $36 = (_vorbis_packet_blocksize($35,$op)|0);
    $37 = ($36|0)<(0);
    if ($37) {
     (_ogg_stream_packetout($23,0)|0);
     $lastblock$4$ph = $lastblock$0$ph;
     break;
    }
    $38 = ($lastblock$0$ph|0)==(0);
    if ($38) {
     $39 = $25;
     $40 = $39;
     $41 = HEAP32[$40>>2]|0;
     $42 = (($39) + 4)|0;
     $43 = $42;
     $44 = HEAP32[$43>>2]|0;
     $67 = $41;$68 = $44;
    } else {
     $45 = (($36) + ($lastblock$0$ph))|0;
     $46 = $45 >> 2;
     $47 = ($46|0)<(0);
     $48 = $47 << 31 >> 31;
     $49 = $25;
     $50 = $49;
     $51 = HEAP32[$50>>2]|0;
     $52 = (($49) + 4)|0;
     $53 = $52;
     $54 = HEAP32[$53>>2]|0;
     $55 = (_i64Add(($51|0),($54|0),($46|0),($48|0))|0);
     $56 = tempRet0;
     $57 = $25;
     $58 = $57;
     HEAP32[$58>>2] = $55;
     $59 = (($57) + 4)|0;
     $60 = $59;
     HEAP32[$60>>2] = $56;
     $67 = $55;$68 = $56;
    }
    $61 = HEAP32[$$pre$phiZ2D>>2]|0;
    $62 = (_vorbis_info_blocksize($61,1)|0);
    $63 = (($62) + ($36))|0;
    $64 = $63 >> 2;
    $65 = ($64|0)<(0);
    $66 = $65 << 31 >> 31;
    $69 = (_i64Add(($64|0),($66|0),($67|0),($68|0))|0);
    $70 = tempRet0;
    $71 = ($70|0)<($1|0);
    $72 = ($69>>>0)<($0>>>0);
    $73 = ($70|0)==($1|0);
    $74 = $73 & $72;
    $75 = $71 | $74;
    if (!($75)) {
     break L17;
    }
    (_ogg_stream_packetout($23,0)|0);
    (_vorbis_synthesis_trackonly($$pre$phi54Z2D,$op)|0);
    (_vorbis_synthesis_blockin($$pre$phi56Z2D,$$pre$phi54Z2D)|0);
    $76 = $26;
    $77 = $76;
    $78 = HEAP32[$77>>2]|0;
    $79 = (($76) + 4)|0;
    $80 = $79;
    $81 = HEAP32[$80>>2]|0;
    $82 = ($81|0)>(-1);
    $83 = ($78>>>0)>(4294967295);
    $84 = ($81|0)==(-1);
    $85 = $84 & $83;
    $86 = $82 | $85;
    if ($86) {
     $87 = HEAP32[$24>>2]|0;
     $88 = $87 << 1;
     $89 = HEAP32[$27>>2]|0;
     $90 = (($89) + ($88<<3)|0);
     $91 = $90;
     $92 = $91;
     $93 = HEAP32[$92>>2]|0;
     $94 = (($91) + 4)|0;
     $95 = $94;
     $96 = HEAP32[$95>>2]|0;
     $97 = (_i64Subtract(($78|0),($81|0),($93|0),($96|0))|0);
     $98 = tempRet0;
     $99 = ($98|0)<(0);
     $100 = $99 ? 0 : $97;
     $101 = $99 ? 0 : $98;
     $102 = $25;
     $103 = $102;
     HEAP32[$103>>2] = $100;
     $104 = (($102) + 4)|0;
     $105 = $104;
     HEAP32[$105>>2] = $101;
     $106 = ($87|0)>(0);
     if ($106) {
      $116 = $100;$117 = $101;$i$044 = 0;
      while(1) {
       $107 = $i$044 << 1;
       $108 = $107 | 1;
       $109 = (($89) + ($108<<3)|0);
       $110 = $109;
       $111 = $110;
       $112 = HEAP32[$111>>2]|0;
       $113 = (($110) + 4)|0;
       $114 = $113;
       $115 = HEAP32[$114>>2]|0;
       $118 = (_i64Add(($116|0),($117|0),($112|0),($115|0))|0);
       $119 = tempRet0;
       $120 = $25;
       $121 = $120;
       HEAP32[$121>>2] = $118;
       $122 = (($120) + 4)|0;
       $123 = $122;
       HEAP32[$123>>2] = $119;
       $124 = (($i$044) + 1)|0;
       $125 = ($124|0)<($87|0);
       if ($125) {
        $116 = $118;$117 = $119;$i$044 = $124;
       } else {
        $lastblock$4$ph = $36;
        break;
       }
      }
     } else {
      $lastblock$4$ph = $36;
     }
    } else {
     $lastblock$4$ph = $36;
    }
   }
  } while(0);
  $lastblock$0$ph = $lastblock$4$ph;
 }
 if ((label|0) == 36) {
  $$5 = $$0$i8$ph;
  STACKTOP = sp;return ($$5|0);
 }
 HEAPF64[$$pre$phi60Z2D>>3] = 0.0;
 $159 = ((($vf)) + 112|0);
 HEAPF64[$159>>3] = 0.0;
 $160 = HEAP32[$$pre$phiZ2D>>2]|0;
 $161 = (_vorbis_synthesis_halfrate_p($160)|0);
 $162 = $25;
 $163 = $162;
 $164 = HEAP32[$163>>2]|0;
 $165 = (($162) + 4)|0;
 $166 = $165;
 $167 = HEAP32[$166>>2]|0;
 $168 = (_bitshift64Ashr(($0|0),($1|0),($161|0))|0);
 $169 = tempRet0;
 $170 = (_bitshift64Shl(($168|0),($169|0),($161|0))|0);
 $171 = tempRet0;
 $172 = ($167|0)<($171|0);
 $173 = ($164>>>0)<($170>>>0);
 $174 = ($167|0)==($171|0);
 $175 = $174 & $173;
 $176 = $172 | $175;
 if ($176) {
  $177 = $164;$178 = $167;
 } else {
  $$5 = 0;
  STACKTOP = sp;return ($$5|0);
 }
 while(1) {
  $179 = (_i64Subtract(($0|0),($1|0),($177|0),($178|0))|0);
  $180 = tempRet0;
  $181 = (_bitshift64Ashr(($179|0),($180|0),($161|0))|0);
  $182 = tempRet0;
  $183 = (_vorbis_synthesis_pcmout($$pre$phi56Z2D,0)|0);
  $184 = ($183|0)<(0);
  $185 = $184 << 31 >> 31;
  $186 = ($185|0)>($182|0);
  $187 = ($183>>>0)>($181>>>0);
  $188 = ($185|0)==($182|0);
  $189 = $188 & $187;
  $190 = $186 | $189;
  $$7 = $190 ? $181 : $183;
  (_vorbis_synthesis_read($$pre$phi56Z2D,$$7)|0);
  $191 = $$7 << $161;
  $192 = ($191|0)<(0);
  $193 = $192 << 31 >> 31;
  $194 = $25;
  $195 = $194;
  $196 = HEAP32[$195>>2]|0;
  $197 = (($194) + 4)|0;
  $198 = $197;
  $199 = HEAP32[$198>>2]|0;
  $200 = (_i64Add(($191|0),($193|0),($196|0),($199|0))|0);
  $201 = tempRet0;
  $202 = $25;
  $203 = $202;
  HEAP32[$203>>2] = $200;
  $204 = (($202) + 4)|0;
  $205 = $204;
  HEAP32[$205>>2] = $201;
  $206 = ($$7|0)<(0);
  $207 = $206 << 31 >> 31;
  $208 = ($207|0)<($182|0);
  $209 = ($$7>>>0)<($181>>>0);
  $210 = ($207|0)==($182|0);
  $211 = $210 & $209;
  $212 = $208 | $211;
  do {
   if ($212) {
    $213 = (__fetch_and_process_packet($vf,1)|0);
    $214 = ($213|0)<(1);
    if ($214) {
     $228 = (_ov_pcm_total($vf,-1)|0);
     $229 = tempRet0;
     $230 = $25;
     $231 = $230;
     HEAP32[$231>>2] = $228;
     $232 = (($230) + 4)|0;
     $233 = $232;
     HEAP32[$233>>2] = $229;
     $222 = $229;$224 = $228;
     break;
    } else {
     $215 = $25;
     $216 = $215;
     $217 = HEAP32[$216>>2]|0;
     $218 = (($215) + 4)|0;
     $219 = $218;
     $220 = HEAP32[$219>>2]|0;
     $222 = $220;$224 = $217;
     break;
    }
   } else {
    $222 = $201;$224 = $200;
   }
  } while(0);
  $221 = ($222|0)<($171|0);
  $223 = ($224>>>0)<($170>>>0);
  $225 = ($222|0)==($171|0);
  $226 = $225 & $223;
  $227 = $221 | $226;
  if ($227) {
   $177 = $224;$178 = $222;
  } else {
   $$5 = 0;
   break;
  }
 }
 STACKTOP = sp;return ($$5|0);
}
function _ov_pcm_seek_page($vf,$0,$1) {
 $vf = $vf|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i47$ph = 0, $$3 = 0, $$phi$trans$insert = 0, $$phi$trans$insert200 = 0, $$phi$trans$insert204 = 0, $$pre = 0, $$pre$phi208Z2D = 0, $$pre$phiZ2D = 0, $$pre196 = 0, $$pre197 = 0, $$pre199 = 0, $$pre201 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0.0;
 var $143 = 0.0, $144 = 0.0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0;
 var $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0;
 var $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0;
 var $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0;
 var $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0;
 var $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0;
 var $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0;
 var $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0;
 var $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0;
 var $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0;
 var $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0;
 var $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0;
 var $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0;
 var $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0;
 var $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0;
 var $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0;
 var $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0;
 var $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0;
 var $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0;
 var $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0;
 var $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0;
 var $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0;
 var $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0;
 var $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0;
 var $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0;
 var $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0;
 var $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0;
 var $666 = 0, $667 = 0, $668 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $got_page$0$lcssa = 0;
 var $got_page$0$lcssa209 = 0, $got_page$0$ph = 0, $got_page$0$ph76 = 0, $got_page$0$ph83164 = 0, $got_page$0149 = 0, $got_page$1$be = 0, $got_page$1130 = 0, $got_page$1130$lcssa377 = 0, $link$0 = 0, $link$0$in = 0, $link$0$in$lcssa = 0, $link$0$in404 = 0, $link$0$lcssa = 0, $link$0406 = 0, $og = 0, $og2 = 0, $op = 0, $or$cond = 0, $result$0$off0$lcssa = 0, $result$0$off0$lcssa211 = 0;
 var $result$0$off0$ph = 0, $result$0$off0$ph75 = 0, $result$1$off0$be = 0, $result$10$off0$ph = 0, $result$11$off0 = 0, $result$8$off0$ph$ph = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $og = sp + 48|0;
 $og2 = sp + 32|0;
 $op = sp;
 $2 = (_ov_pcm_total($vf,-1)|0);
 $3 = tempRet0;
 $4 = ((($vf)) + 88|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)<(2);
 if ($6) {
  $$3 = -131;
  STACKTOP = sp;return ($$3|0);
 }
 $7 = ((($vf)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0);
 if ($9) {
  $$3 = -138;
  STACKTOP = sp;return ($$3|0);
 }
 $10 = ($1|0)<(0);
 $11 = ($3|0)<($1|0);
 $12 = ($2>>>0)<($0>>>0);
 $13 = ($3|0)==($1|0);
 $14 = $13 & $12;
 $15 = $11 | $14;
 $or$cond = $10 | $15;
 if ($or$cond) {
  $$3 = -131;
  STACKTOP = sp;return ($$3|0);
 }
 $16 = ((($vf)) + 52|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ((($vf)) + 68|0);
 $36 = $2;$37 = $3;$link$0$in = $17;
 while(1) {
  $link$0 = (($link$0$in) + -1)|0;
  $19 = ($link$0$in|0)>(0);
  if (!($19)) {
   $610 = $36;$611 = $37;$link$0$in$lcssa = $link$0$in;$link$0$lcssa = $link$0;
   label = 6;
   break;
  }
  $26 = $link$0 << 1;
  $27 = $26 | 1;
  $28 = HEAP32[$18>>2]|0;
  $29 = (($28) + ($27<<3)|0);
  $30 = $29;
  $31 = $30;
  $32 = HEAP32[$31>>2]|0;
  $33 = (($30) + 4)|0;
  $34 = $33;
  $35 = HEAP32[$34>>2]|0;
  $38 = (_i64Subtract(($36|0),($37|0),($32|0),($35|0))|0);
  $39 = tempRet0;
  $40 = ($39|0)>($1|0);
  $41 = ($38>>>0)>($0>>>0);
  $42 = ($39|0)==($1|0);
  $43 = $42 & $41;
  $44 = $40 | $43;
  if ($44) {
   $36 = $38;$37 = $39;$link$0$in = $link$0;
  } else {
   $$pre$phiZ2D = $26;$64 = $28;$71 = $32;$72 = $35;$75 = $38;$76 = $39;$link$0$in404 = $link$0$in;$link$0406 = $link$0;
   break;
  }
 }
 if ((label|0) == 6) {
  $$pre = HEAP32[$18>>2]|0;
  $$pre196 = $link$0$lcssa << 1;
  $$pre197 = $$pre196 | 1;
  $$phi$trans$insert204 = (($$pre) + ($$pre197<<3)|0);
  $20 = $$phi$trans$insert204;
  $21 = $20;
  $22 = HEAP32[$21>>2]|0;
  $23 = (($20) + 4)|0;
  $24 = $23;
  $25 = HEAP32[$24>>2]|0;
  $$pre$phiZ2D = $$pre196;$64 = $$pre;$71 = $22;$72 = $25;$75 = $610;$76 = $611;$link$0$in404 = $link$0$in$lcssa;$link$0406 = $link$0$lcssa;
 }
 $45 = ((($vf)) + 56|0);
 $46 = HEAP32[$45>>2]|0;
 $47 = (($46) + ($link$0$in404<<3)|0);
 $48 = $47;
 $49 = $48;
 $50 = HEAP32[$49>>2]|0;
 $51 = (($48) + 4)|0;
 $52 = $51;
 $53 = HEAP32[$52>>2]|0;
 $54 = ((($vf)) + 60|0);
 $55 = HEAP32[$54>>2]|0;
 $56 = (($55) + ($link$0406<<3)|0);
 $57 = $56;
 $58 = $57;
 $59 = HEAP32[$58>>2]|0;
 $60 = (($57) + 4)|0;
 $61 = $60;
 $62 = HEAP32[$61>>2]|0;
 $63 = (($64) + ($$pre$phiZ2D<<3)|0);
 $65 = $63;
 $66 = $65;
 $67 = HEAP32[$66>>2]|0;
 $68 = (($65) + 4)|0;
 $69 = $68;
 $70 = HEAP32[$69>>2]|0;
 $73 = (_i64Add(($71|0),($72|0),($67|0),($70|0))|0);
 $74 = tempRet0;
 $77 = (_i64Subtract(($0|0),($1|0),($75|0),($76|0))|0);
 $78 = tempRet0;
 $79 = (_i64Add(($67|0),($70|0),($77|0),($78|0))|0);
 $80 = tempRet0;
 $81 = ($59|0)==($50|0);
 $82 = ($62|0)==($53|0);
 $83 = $81 & $82;
 do {
  if ($83) {
   $84 = HEAP32[$vf>>2]|0;
   $85 = ($84|0)==(0|0);
   if ($85) {
    $result$10$off0$ph = -129;
    label = 102;
   } else {
    $86 = ((($vf)) + 8|0);
    $87 = $86;
    $88 = $87;
    $89 = HEAP32[$88>>2]|0;
    $90 = (($87) + 4)|0;
    $91 = $90;
    $92 = HEAP32[$91>>2]|0;
    $93 = ($89|0)==($50|0);
    $94 = ($92|0)==($53|0);
    $95 = $93 & $94;
    if (!($95)) {
     $96 = ((($vf)) + 708|0);
     $97 = HEAP32[$96>>2]|0;
     $98 = ($97|0)==(0|0);
     if ($98) {
      $result$10$off0$ph = -128;
      label = 102;
      break;
     }
     $99 = (FUNCTION_TABLE_iiiii[$97 & 8191]($84,$50,$53,0)|0);
     $100 = ($99|0)==(-1);
     if ($100) {
      $result$10$off0$ph = -128;
      label = 102;
      break;
     }
     $101 = $86;
     $102 = $101;
     HEAP32[$102>>2] = $50;
     $103 = (($101) + 4)|0;
     $104 = $103;
     HEAP32[$104>>2] = $53;
     $105 = ((($vf)) + 24|0);
     (_ogg_sync_reset($105)|0);
    }
    $106 = (__get_next_page($vf,$og,1,0)|0);
    $107 = tempRet0;
    $108 = ($107|0)<(0);
    if ($108) {
     $result$10$off0$ph = $106;
     label = 102;
    } else {
     $got_page$0$ph76 = 1;$result$0$off0$ph75 = $106;
     label = 15;
    }
   }
  } else {
   $got_page$0$ph76 = 0;$result$0$off0$ph75 = 0;
   label = 15;
  }
 } while(0);
 L24: do {
  if ((label|0) == 15) {
   $109 = ($53|0)>($62|0);
   $110 = ($50>>>0)>($59>>>0);
   $111 = ($53|0)==($62|0);
   $112 = $111 & $110;
   $113 = $109 | $112;
   L26: do {
    if ($113) {
     $114 = ((($vf)) + 8|0);
     $115 = ((($vf)) + 708|0);
     $116 = ((($vf)) + 24|0);
     $117 = ((($vf)) + 64|0);
     $145 = $73;$146 = $74;$612 = -1;$613 = -1;$614 = $67;$615 = $70;$616 = $59;$617 = $62;$618 = $50;$619 = $53;$got_page$0$ph83164 = $got_page$0$ph76;
     L28: while(1) {
      $127 = $618;$128 = $619;$129 = $616;$130 = $617;$138 = $614;$139 = $615;$620 = $612;$621 = $613;$got_page$0149 = $got_page$0$ph83164;
      L30: while(1) {
       $131 = (_i64Subtract(($127|0),($128|0),($129|0),($130|0))|0);
       $132 = tempRet0;
       $133 = ($132|0)<(0);
       $134 = ($131>>>0)<(65536);
       $135 = ($132|0)==(0);
       $136 = $135 & $134;
       $137 = $133 | $136;
       if ($137) {
        $175 = $129;$177 = $130;
       } else {
        $140 = (_i64Subtract(($79|0),($80|0),($138|0),($139|0))|0);
        $141 = tempRet0;
        $142 = (+($140>>>0)) + (4294967296.0*(+($141|0)));
        $143 = (+($131>>>0)) + (4294967296.0*(+($132|0)));
        $144 = $143 * $142;
        $147 = (_i64Subtract(($145|0),($146|0),($138|0),($139|0))|0);
        $148 = tempRet0;
        $149 = (+($147>>>0)) + (4294967296.0*(+($148|0)));
        $150 = $144 / $149;
        $151 = (~~$150)>>>0;
        $152 = +Math_abs($150) >= 1.0 ? $150 > 0.0 ? (~~+Math_min(+Math_floor($150 / 4294967296.0), 4294967295.0)) >>> 0 : ~~+Math_ceil(($150 - +(~~$150 >>> 0)) / 4294967296.0) >>> 0 : 0;
        $153 = (_i64Add(($129|0),($130|0),-65536,-1)|0);
        $154 = tempRet0;
        $155 = (_i64Add(($153|0),($154|0),($151|0),($152|0))|0);
        $156 = tempRet0;
        $157 = (_i64Add(($129|0),($130|0),65536,0)|0);
        $158 = tempRet0;
        $159 = ($156|0)<($158|0);
        $160 = ($155>>>0)<($157>>>0);
        $161 = ($156|0)==($158|0);
        $162 = $161 & $160;
        $163 = $159 | $162;
        $164 = $163 ? $129 : $155;
        $165 = $163 ? $130 : $156;
        $175 = $164;$177 = $165;
       }
       $166 = HEAP32[$vf>>2]|0;
       $167 = ($166|0)==(0|0);
       if ($167) {
        $result$10$off0$ph = -129;
        label = 102;
        break L24;
       }
       $168 = $114;
       $169 = $168;
       $170 = HEAP32[$169>>2]|0;
       $171 = (($168) + 4)|0;
       $172 = $171;
       $173 = HEAP32[$172>>2]|0;
       $174 = ($170|0)==($175|0);
       $176 = ($173|0)==($177|0);
       $178 = $174 & $176;
       if (!($178)) {
        $179 = HEAP32[$115>>2]|0;
        $180 = ($179|0)==(0|0);
        if ($180) {
         $result$10$off0$ph = -128;
         label = 102;
         break L24;
        }
        $181 = (FUNCTION_TABLE_iiiii[$179 & 8191]($166,$175,$177,0)|0);
        $182 = ($181|0)==(-1);
        if ($182) {
         $result$10$off0$ph = -128;
         label = 102;
         break L24;
        }
        $183 = $114;
        $184 = $183;
        HEAP32[$184>>2] = $175;
        $185 = (($183) + 4)|0;
        $186 = $185;
        HEAP32[$186>>2] = $177;
        (_ogg_sync_reset($116)|0);
       }
       $187 = ($128|0)>($130|0);
       $188 = ($127>>>0)>($129>>>0);
       $189 = ($128|0)==($130|0);
       $190 = $189 & $188;
       $191 = $187 | $190;
       L41: do {
        if ($191) {
         $198 = $127;$199 = $128;$208 = $129;$209 = $130;$213 = $177;$215 = $175;$628 = $620;$629 = $621;$630 = $138;$631 = $139;$got_page$1130 = $got_page$0149;
         while(1) {
          $192 = $114;
          $193 = $192;
          $194 = HEAP32[$193>>2]|0;
          $195 = (($192) + 4)|0;
          $196 = $195;
          $197 = HEAP32[$196>>2]|0;
          $200 = (_i64Subtract(($198|0),($199|0),($194|0),($197|0))|0);
          $201 = tempRet0;
          $202 = (__get_next_page($vf,$og,$200,$201)|0);
          $203 = tempRet0;
          $204 = ($202|0)==(-128);
          $205 = ($203|0)==(-1);
          $206 = $204 & $205;
          if ($206) {
           $result$10$off0$ph = -128;
           label = 102;
           break L24;
          }
          $207 = ($203|0)<(0);
          do {
           if ($207) {
            $210 = (_i64Add(($208|0),($209|0),1,0)|0);
            $211 = tempRet0;
            $212 = ($213|0)>($211|0);
            $214 = ($215>>>0)>($210>>>0);
            $216 = ($213|0)==($211|0);
            $217 = $216 & $214;
            $218 = $212 | $217;
            if (!($218)) {
             $632 = $202;$633 = $203;$634 = $208;$635 = $209;$636 = $628;$637 = $629;$got_page$1130$lcssa377 = $got_page$1130;
             label = 30;
             break L28;
            }
            $228 = ($215|0)==(0);
            $229 = ($213|0)==(0);
            $230 = $228 & $229;
            if ($230) {
             $644 = $202;$645 = $203;
             label = 33;
             break L28;
            }
            $231 = (_i64Add(($215|0),($213|0),-65536,-1)|0);
            $232 = tempRet0;
            $233 = ($232|0)>($209|0);
            $234 = ($231>>>0)>($208>>>0);
            $235 = ($232|0)==($209|0);
            $236 = $235 & $234;
            $237 = $233 | $236;
            $238 = $237 ? $231 : $210;
            $239 = $237 ? $232 : $211;
            $240 = HEAP32[$vf>>2]|0;
            $241 = ($240|0)==(0|0);
            if ($241) {
             $result$10$off0$ph = -129;
             label = 102;
             break L24;
            }
            $242 = $114;
            $243 = $242;
            $244 = HEAP32[$243>>2]|0;
            $245 = (($242) + 4)|0;
            $246 = $245;
            $247 = HEAP32[$246>>2]|0;
            $248 = ($244|0)==($238|0);
            $249 = ($247|0)==($239|0);
            $250 = $248 & $249;
            if ($250) {
             $220 = $209;$221 = $199;$223 = $208;$224 = $198;$638 = $238;$639 = $239;$640 = $628;$641 = $629;$642 = $630;$643 = $631;$got_page$1$be = $got_page$1130;$result$1$off0$be = 0;
             break;
            }
            $251 = HEAP32[$115>>2]|0;
            $252 = ($251|0)==(0|0);
            if ($252) {
             $result$10$off0$ph = -128;
             label = 102;
             break L24;
            }
            $253 = (FUNCTION_TABLE_iiiii[$251 & 8191]($240,$238,$239,0)|0);
            $254 = ($253|0)==(-1);
            if ($254) {
             $result$10$off0$ph = -128;
             label = 102;
             break L24;
            }
            $255 = $114;
            $256 = $255;
            HEAP32[$256>>2] = $238;
            $257 = (($255) + 4)|0;
            $258 = $257;
            HEAP32[$258>>2] = $239;
            (_ogg_sync_reset($116)|0);
            $220 = $209;$221 = $199;$223 = $208;$224 = $198;$638 = $238;$639 = $239;$640 = $628;$641 = $629;$642 = $630;$643 = $631;$got_page$1$be = $got_page$1130;$result$1$off0$be = 0;
           } else {
            $259 = (_ogg_page_serialno($og)|0);
            $260 = HEAP32[$117>>2]|0;
            $261 = (($260) + ($link$0406<<2)|0);
            $262 = HEAP32[$261>>2]|0;
            $263 = ($259|0)==($262|0);
            if (!($263)) {
             $220 = $209;$221 = $199;$223 = $208;$224 = $198;$638 = $215;$639 = $213;$640 = $628;$641 = $629;$642 = $630;$643 = $631;$got_page$1$be = 1;$result$1$off0$be = $202;
             break;
            }
            $264 = (_ogg_page_granulepos($og)|0);
            $265 = tempRet0;
            $266 = ($264|0)==(-1);
            $267 = ($265|0)==(-1);
            $268 = $266 & $267;
            if ($268) {
             $220 = $209;$221 = $199;$223 = $208;$224 = $198;$638 = $215;$639 = $213;$640 = $628;$641 = $629;$642 = $630;$643 = $631;$got_page$1$be = 1;$result$1$off0$be = $202;
             break;
            }
            $269 = ($80|0)>($265|0);
            $270 = ($79>>>0)>($264>>>0);
            $271 = ($80|0)==($265|0);
            $272 = $271 & $270;
            $273 = $269 | $272;
            if ($273) {
             $274 = $114;
             $275 = $274;
             $276 = HEAP32[$275>>2]|0;
             $277 = (($274) + 4)|0;
             $278 = $277;
             $279 = HEAP32[$278>>2]|0;
             $280 = (_i64Subtract(($79|0),($80|0),($264|0),($265|0))|0);
             $281 = tempRet0;
             $282 = ($281|0)>(0);
             $283 = ($280>>>0)>(44100);
             $284 = ($281|0)==(0);
             $285 = $284 & $283;
             $286 = $282 | $285;
             if ($286) {
              $119 = $199;$120 = $279;$122 = $198;$123 = $276;$622 = $202;$623 = $203;$624 = $264;$625 = $265;$got_page$0$ph = 1;$result$0$off0$ph = $202;
              break L41;
             } else {
              $220 = $279;$221 = $199;$223 = $276;$224 = $198;$638 = $276;$639 = $279;$640 = $202;$641 = $203;$642 = $264;$643 = $265;$got_page$1$be = 1;$result$1$off0$be = $202;
              break;
             }
            }
            $287 = (_i64Add(($208|0),($209|0),1,0)|0);
            $288 = tempRet0;
            $289 = ($213|0)>($288|0);
            $290 = ($215>>>0)>($287>>>0);
            $291 = ($213|0)==($288|0);
            $292 = $291 & $290;
            $293 = $289 | $292;
            if (!($293)) {
             $646 = $202;$647 = $203;$648 = $208;$649 = $209;$650 = $628;$651 = $629;
             label = 46;
             break L28;
            }
            $294 = $114;
            $295 = $294;
            $296 = HEAP32[$295>>2]|0;
            $297 = (($294) + 4)|0;
            $298 = $297;
            $299 = HEAP32[$298>>2]|0;
            $300 = ($198|0)==($296|0);
            $301 = ($199|0)==($299|0);
            $302 = $300 & $301;
            if (!($302)) {
             $326 = $213;$327 = $209;$329 = $215;$330 = $208;$652 = $264;$653 = $265;$654 = $202;$655 = $203;$656 = $630;$657 = $631;$658 = $628;$659 = $629;
             break L30;
            }
            $303 = (_i64Add(($215|0),($213|0),-65536,-1)|0);
            $304 = tempRet0;
            $305 = ($304|0)>($209|0);
            $306 = ($303>>>0)>($208>>>0);
            $307 = ($304|0)==($209|0);
            $308 = $307 & $306;
            $309 = $305 | $308;
            $310 = $309 ? $303 : $287;
            $311 = $309 ? $304 : $288;
            $312 = HEAP32[$vf>>2]|0;
            $313 = ($312|0)==(0|0);
            if ($313) {
             $result$10$off0$ph = -129;
             label = 102;
             break L24;
            }
            $314 = ($198|0)==($310|0);
            $315 = ($199|0)==($311|0);
            $316 = $314 & $315;
            if ($316) {
             $220 = $209;$221 = $203;$223 = $208;$224 = $202;$638 = $198;$639 = $199;$640 = $628;$641 = $629;$642 = $630;$643 = $631;$got_page$1$be = 1;$result$1$off0$be = 0;
             break;
            }
            $317 = HEAP32[$115>>2]|0;
            $318 = ($317|0)==(0|0);
            if ($318) {
             $result$10$off0$ph = -128;
             label = 102;
             break L24;
            }
            $319 = (FUNCTION_TABLE_iiiii[$317 & 8191]($312,$310,$311,0)|0);
            $320 = ($319|0)==(-1);
            if ($320) {
             $result$10$off0$ph = -128;
             label = 102;
             break L24;
            }
            $321 = $114;
            $322 = $321;
            HEAP32[$322>>2] = $310;
            $323 = (($321) + 4)|0;
            $324 = $323;
            HEAP32[$324>>2] = $311;
            (_ogg_sync_reset($116)|0);
            $220 = $209;$221 = $203;$223 = $208;$224 = $202;$638 = $310;$639 = $311;$640 = $628;$641 = $629;$642 = $630;$643 = $631;$got_page$1$be = 1;$result$1$off0$be = 0;
           }
          } while(0);
          $219 = ($220|0)<($221|0);
          $222 = ($223>>>0)<($224>>>0);
          $225 = ($220|0)==($221|0);
          $226 = $225 & $222;
          $227 = $219 | $226;
          if ($227) {
           $198 = $224;$199 = $221;$208 = $223;$209 = $220;$213 = $639;$215 = $638;$628 = $640;$629 = $641;$630 = $642;$631 = $643;$got_page$1130 = $got_page$1$be;
          } else {
           $119 = $221;$120 = $220;$122 = $224;$123 = $223;$622 = $640;$623 = $641;$624 = $642;$625 = $643;$got_page$0$ph = $got_page$1$be;$result$0$off0$ph = $result$1$off0$be;
           break;
          }
         }
        } else {
         $119 = $128;$120 = $130;$122 = $127;$123 = $129;$622 = $620;$623 = $621;$624 = $138;$625 = $139;$got_page$0$ph = $got_page$0149;$result$0$off0$ph = 0;
        }
       } while(0);
       $118 = ($119|0)>($120|0);
       $121 = ($122>>>0)>($123>>>0);
       $124 = ($119|0)==($120|0);
       $125 = $124 & $121;
       $126 = $118 | $125;
       if ($126) {
        $127 = $122;$128 = $119;$129 = $123;$130 = $120;$138 = $624;$139 = $625;$620 = $622;$621 = $623;$got_page$0149 = $got_page$0$ph;
       } else {
        $335 = $622;$337 = $623;$626 = $123;$627 = $120;$got_page$0$lcssa = $got_page$0$ph;$result$0$off0$lcssa = $result$0$off0$ph;
        break L28;
       }
      }
      $325 = ($326|0)>($327|0);
      $328 = ($329>>>0)>($330>>>0);
      $331 = ($326|0)==($327|0);
      $332 = $331 & $328;
      $333 = $325 | $332;
      if ($333) {
       $145 = $652;$146 = $653;$612 = $658;$613 = $659;$614 = $656;$615 = $657;$616 = $330;$617 = $327;$618 = $329;$619 = $326;$got_page$0$ph83164 = 1;
      } else {
       $660 = $654;$661 = $655;$662 = $330;$663 = $327;$664 = $658;$665 = $659;
       label = 54;
       break;
      }
     }
     if ((label|0) == 30) {
      $335 = $636;$337 = $637;$626 = $634;$627 = $635;$got_page$0$lcssa = $got_page$1130$lcssa377;$result$0$off0$lcssa = $632;
     }
     else if ((label|0) == 33) {
      $result$10$off0$ph = $644;
      label = 102;
      break L24;
     }
     else if ((label|0) == 46) {
      $335 = $650;$337 = $651;$626 = $648;$627 = $649;$got_page$0$lcssa = 1;$result$0$off0$lcssa = $646;
     }
     else if ((label|0) == 54) {
      $335 = $664;$337 = $665;$626 = $662;$627 = $663;$got_page$0$lcssa = 1;$result$0$off0$lcssa = $660;
     }
     $334 = ($335|0)==(-1);
     $336 = ($337|0)==(-1);
     $338 = $334 & $336;
     if ($338) {
      $349 = $626;$351 = $627;$got_page$0$lcssa209 = $got_page$0$lcssa;$result$0$off0$lcssa211 = $result$0$off0$lcssa;
      label = 56;
     } else {
      $381 = HEAP32[$vf>>2]|0;
      $382 = ($381|0)==(0|0);
      L79: do {
       if ($382) {
        $$0$i47$ph = -129;
        label = 68;
       } else {
        $383 = ((($vf)) + 8|0);
        $384 = $383;
        $385 = $384;
        $386 = HEAP32[$385>>2]|0;
        $387 = (($384) + 4)|0;
        $388 = $387;
        $389 = HEAP32[$388>>2]|0;
        $390 = ($386|0)==($335|0);
        $391 = ($389|0)==($337|0);
        $392 = $390 & $391;
        if (!($392)) {
         $393 = ((($vf)) + 708|0);
         $394 = HEAP32[$393>>2]|0;
         $395 = ($394|0)==(0|0);
         if ($395) {
          $$0$i47$ph = -128;
          label = 68;
          break;
         }
         $396 = (FUNCTION_TABLE_iiiii[$394 & 8191]($381,$335,$337,0)|0);
         $397 = ($396|0)==(-1);
         if ($397) {
          $$0$i47$ph = -128;
          label = 68;
          break;
         }
         $398 = $383;
         $399 = $398;
         HEAP32[$399>>2] = $335;
         $400 = (($398) + 4)|0;
         $401 = $400;
         HEAP32[$401>>2] = $337;
         $402 = ((($vf)) + 24|0);
         (_ogg_sync_reset($402)|0);
        }
        $408 = ((($vf)) + 80|0);
        $409 = $408;
        $410 = $409;
        HEAP32[$410>>2] = -1;
        $411 = (($409) + 4)|0;
        $412 = $411;
        HEAP32[$412>>2] = -1;
        $413 = (__get_next_page($vf,$og2,-1,-1)|0);
        $414 = tempRet0;
        $415 = ($414|0)<(0);
        if ($415) {
         $result$8$off0$ph$ph = $413;
         break;
        }
        $416 = ((($vf)) + 96|0);
        $417 = HEAP32[$416>>2]|0;
        $418 = ($link$0406|0)==($417|0);
        $419 = ((($vf)) + 480|0);
        if ($418) {
         (_vorbis_synthesis_restart($419)|0);
         $$phi$trans$insert = ((($vf)) + 92|0);
         $$pre199 = HEAP32[$$phi$trans$insert>>2]|0;
         $$pre$phi208Z2D = $$phi$trans$insert;$427 = $$pre199;
        } else {
         _vorbis_dsp_clear($419);
         $420 = ((($vf)) + 592|0);
         (_vorbis_block_clear($420)|0);
         HEAP32[$416>>2] = $link$0406;
         $421 = ((($vf)) + 64|0);
         $422 = HEAP32[$421>>2]|0;
         $423 = (($422) + ($link$0406<<2)|0);
         $424 = HEAP32[$423>>2]|0;
         $425 = ((($vf)) + 92|0);
         HEAP32[$425>>2] = $424;
         HEAP32[$4>>2] = 3;
         $$pre$phi208Z2D = $425;$427 = $424;
        }
        $426 = ((($vf)) + 120|0);
        (_ogg_stream_reset_serialno($426,$427)|0);
        (_ogg_stream_pagein($426,$og2)|0);
        $428 = ((($op)) + 16|0);
        $429 = ((($vf)) + 708|0);
        $430 = ((($vf)) + 24|0);
        $431 = ((($og2)) + 4|0);
        L93: while(1) {
         $432 = (_ogg_stream_packetpeek($426,$op)|0);
         $433 = ($432|0)<(0);
         $434 = $433 << 31 >> 31;
         $435 = ($432|0)==(0);
         if ($435) {
          $436 = HEAP32[$54>>2]|0;
          $437 = (($436) + ($link$0406<<3)|0);
          $438 = $437;
          $439 = $438;
          $440 = HEAP32[$439>>2]|0;
          $441 = (($438) + 4)|0;
          $442 = $441;
          $443 = HEAP32[$442>>2]|0;
          $444 = ($337|0)>($443|0);
          $445 = ($335>>>0)>($440>>>0);
          $446 = ($337|0)==($443|0);
          $447 = $446 & $445;
          $448 = $444 | $447;
          if ($448) {
           $487 = $337;$489 = $335;
           while(1) {
            $454 = $489;$455 = $487;
            while(1) {
             $456 = (_i64Add(($454|0),($455|0),-65536,-1)|0);
             $457 = tempRet0;
             $458 = ($457|0)<(0);
             $459 = $458 ? 0 : $456;
             $460 = $458 ? 0 : $457;
             $461 = HEAP32[$vf>>2]|0;
             $462 = ($461|0)==(0|0);
             if ($462) {
              $result$8$off0$ph$ph = -129;
              break L79;
             }
             $463 = $383;
             $464 = $463;
             $465 = HEAP32[$464>>2]|0;
             $466 = (($463) + 4)|0;
             $467 = $466;
             $468 = HEAP32[$467>>2]|0;
             $469 = ($465|0)==($459|0);
             $470 = ($468|0)==($460|0);
             $471 = $469 & $470;
             if ($471) {
              $667 = -1;$668 = -1;
             } else {
              $472 = HEAP32[$429>>2]|0;
              $473 = ($472|0)==(0|0);
              if ($473) {
               $result$8$off0$ph$ph = -128;
               break L79;
              }
              $474 = (FUNCTION_TABLE_iiiii[$472 & 8191]($461,$459,$460,0)|0);
              $475 = ($474|0)==(-1);
              if ($475) {
               $result$8$off0$ph$ph = -128;
               break L79;
              }
              $476 = $383;
              $477 = $476;
              HEAP32[$477>>2] = $459;
              $478 = (($476) + 4)|0;
              $479 = $478;
              HEAP32[$479>>2] = $460;
              (_ogg_sync_reset($430)|0);
              $667 = -1;$668 = -1;
             }
             while(1) {
              $480 = $383;
              $481 = $480;
              $482 = HEAP32[$481>>2]|0;
              $483 = (($480) + 4)|0;
              $484 = $483;
              $485 = HEAP32[$484>>2]|0;
              $486 = ($487|0)>($485|0);
              $488 = ($489>>>0)>($482>>>0);
              $490 = ($487|0)==($485|0);
              $491 = $490 & $488;
              $492 = $486 | $491;
              if (!($492)) {
               $450 = $667;$452 = $668;
               break;
              }
              ;HEAP32[$og2>>2]=0|0;HEAP32[$og2+4>>2]=0|0;HEAP32[$og2+8>>2]=0|0;HEAP32[$og2+12>>2]=0|0;
              $493 = (_i64Subtract(($489|0),($487|0),($482|0),($485|0))|0);
              $494 = tempRet0;
              $495 = (__get_next_page($vf,$og2,$493,$494)|0);
              $496 = tempRet0;
              $497 = ($495|0)==(-128);
              $498 = ($496|0)==(-1);
              $499 = $497 & $498;
              if ($499) {
               $result$8$off0$ph$ph = -128;
               break L79;
              }
              $500 = ($496|0)<(0);
              if ($500) {
               $450 = $667;$452 = $668;
               break;
              } else {
               $667 = $495;$668 = $496;
              }
             }
             $449 = ($450|0)==(-1);
             $451 = ($452|0)==(-1);
             $453 = $449 & $451;
             if ($453) {
              $454 = $459;$455 = $460;
             } else {
              $512 = $450;$514 = $452;
              break;
             }
            }
            $501 = HEAP32[$431>>2]|0;
            $502 = ($501|0)==(0);
            if ($502) {
             $503 = HEAP32[$vf>>2]|0;
             $504 = ($503|0)==(0|0);
             if ($504) {
              $result$8$off0$ph$ph = -129;
              break L79;
             }
             $505 = $383;
             $506 = $505;
             $507 = HEAP32[$506>>2]|0;
             $508 = (($505) + 4)|0;
             $509 = $508;
             $510 = HEAP32[$509>>2]|0;
             $511 = ($507|0)==($512|0);
             $513 = ($510|0)==($514|0);
             $515 = $511 & $513;
             if (!($515)) {
              $516 = HEAP32[$429>>2]|0;
              $517 = ($516|0)==(0|0);
              if ($517) {
               $result$8$off0$ph$ph = -128;
               break L79;
              }
              $518 = (FUNCTION_TABLE_iiiii[$516 & 8191]($503,$512,$514,0)|0);
              $519 = ($518|0)==(-1);
              if ($519) {
               $result$8$off0$ph$ph = -128;
               break L79;
              }
              $520 = $383;
              $521 = $520;
              HEAP32[$521>>2] = $512;
              $522 = (($520) + 4)|0;
              $523 = $522;
              HEAP32[$523>>2] = $514;
              (_ogg_sync_reset($430)|0);
             }
             (__get_next_page($vf,$og2,65536,0)|0);
             $524 = tempRet0;
             $525 = ($524|0)<(0);
             if ($525) {
              $result$8$off0$ph$ph = -129;
              break L79;
             }
            }
            $526 = (_ogg_page_serialno($og2)|0);
            $527 = HEAP32[$$pre$phi208Z2D>>2]|0;
            $528 = ($526|0)==($527|0);
            if ($528) {
             $542 = (_ogg_page_granulepos($og2)|0);
             $543 = tempRet0;
             $544 = ($543|0)>(-1);
             $545 = ($542>>>0)>(4294967295);
             $546 = ($543|0)==(-1);
             $547 = $546 & $545;
             $548 = $544 | $547;
             if ($548) {
              $585 = $512;$586 = $514;
              break L93;
             }
             $549 = (_ogg_page_continued($og2)|0);
             $550 = ($549|0)==(0);
             if ($550) {
              $585 = $512;$586 = $514;
              break L93;
             }
            }
            $529 = HEAP32[$54>>2]|0;
            $530 = (($529) + ($link$0406<<3)|0);
            $531 = $530;
            $532 = $531;
            $533 = HEAP32[$532>>2]|0;
            $534 = (($531) + 4)|0;
            $535 = $534;
            $536 = HEAP32[$535>>2]|0;
            $537 = ($514|0)>($536|0);
            $538 = ($512>>>0)>($533>>>0);
            $539 = ($514|0)==($536|0);
            $540 = $539 & $538;
            $541 = $537 | $540;
            if ($541) {
             $487 = $514;$489 = $512;
            } else {
             $552 = $514;$666 = $512;
             break;
            }
           }
          } else {
           $552 = $337;$666 = $335;
          }
         } else {
          $552 = $434;$666 = $432;
         }
         $551 = ($552|0)<(0);
         if ($551) {
          $result$8$off0$ph$ph = -136;
          break L79;
         }
         $553 = $428;
         $554 = $553;
         $555 = HEAP32[$554>>2]|0;
         $556 = (($553) + 4)|0;
         $557 = $556;
         $558 = HEAP32[$557>>2]|0;
         $559 = ($555|0)==(-1);
         $560 = ($558|0)==(-1);
         $561 = $559 & $560;
         if (!($561)) {
          $572 = $555;$573 = $558;
          label = 101;
          break;
         }
         (_ogg_stream_packetout($426,0)|0);
        }
        if ((label|0) == 101) {
         $562 = HEAP32[$416>>2]|0;
         $563 = $562 << 1;
         $564 = HEAP32[$18>>2]|0;
         $565 = (($564) + ($563<<3)|0);
         $566 = $565;
         $567 = $566;
         $568 = HEAP32[$567>>2]|0;
         $569 = (($566) + 4)|0;
         $570 = $569;
         $571 = HEAP32[$570>>2]|0;
         $574 = (_i64Subtract(($572|0),($573|0),($568|0),($571|0))|0);
         $575 = tempRet0;
         $576 = ($575|0)<(0);
         $577 = $576 ? 0 : $574;
         $578 = $576 ? 0 : $575;
         $579 = (_i64Add(($577|0),($578|0),($75|0),($76|0))|0);
         $580 = tempRet0;
         $581 = $408;
         $582 = $581;
         HEAP32[$582>>2] = $579;
         $583 = (($581) + 4)|0;
         $584 = $583;
         HEAP32[$584>>2] = $580;
         $589 = $580;$591 = $579;
         break L26;
        }
        $587 = (_ov_raw_seek($vf,$585,$586)|0);
        $$3 = $587;
        STACKTOP = sp;return ($$3|0);
       }
      } while(0);
      if ((label|0) == 68) {
       $403 = ((($vf)) + 80|0);
       $404 = $403;
       $405 = $404;
       HEAP32[$405>>2] = -1;
       $406 = (($404) + 4)|0;
       $407 = $406;
       HEAP32[$407>>2] = -1;
       $result$8$off0$ph$ph = $$0$i47$ph;
      }
      $result$11$off0 = $result$8$off0$ph$ph;
      break L24;
     }
    } else {
     $349 = $59;$351 = $62;$got_page$0$lcssa209 = $got_page$0$ph76;$result$0$off0$lcssa211 = $result$0$off0$ph75;
     label = 56;
    }
   } while(0);
   if ((label|0) == 56) {
    $339 = ($got_page$0$lcssa209|0)==(0);
    if ($339) {
     $result$10$off0$ph = $result$0$off0$lcssa211;
     label = 102;
     break;
    }
    $340 = HEAP32[$54>>2]|0;
    $341 = (($340) + ($link$0406<<3)|0);
    $342 = $341;
    $343 = $342;
    $344 = HEAP32[$343>>2]|0;
    $345 = (($342) + 4)|0;
    $346 = $345;
    $347 = HEAP32[$346>>2]|0;
    $348 = ($349|0)==($344|0);
    $350 = ($351|0)==($347|0);
    $352 = $348 & $350;
    if (!($352)) {
     $result$10$off0$ph = $result$0$off0$lcssa211;
     label = 102;
     break;
    }
    $353 = (_ogg_page_serialno($og)|0);
    $354 = ((($vf)) + 64|0);
    $355 = HEAP32[$354>>2]|0;
    $356 = (($355) + ($link$0406<<2)|0);
    $357 = HEAP32[$356>>2]|0;
    $358 = ($353|0)==($357|0);
    if (!($358)) {
     $result$10$off0$ph = $result$0$off0$lcssa211;
     label = 102;
     break;
    }
    $359 = ((($vf)) + 80|0);
    $360 = $359;
    $361 = $360;
    HEAP32[$361>>2] = $75;
    $362 = (($360) + 4)|0;
    $363 = $362;
    HEAP32[$363>>2] = $76;
    $364 = ((($vf)) + 96|0);
    $365 = HEAP32[$364>>2]|0;
    $366 = ($link$0406|0)==($365|0);
    $367 = ((($vf)) + 480|0);
    if ($366) {
     (_vorbis_synthesis_restart($367)|0);
     $$phi$trans$insert200 = ((($vf)) + 92|0);
     $$pre201 = HEAP32[$$phi$trans$insert200>>2]|0;
     $374 = $$pre201;
    } else {
     _vorbis_dsp_clear($367);
     $368 = ((($vf)) + 592|0);
     (_vorbis_block_clear($368)|0);
     HEAP32[$364>>2] = $link$0406;
     $369 = HEAP32[$354>>2]|0;
     $370 = (($369) + ($link$0406<<2)|0);
     $371 = HEAP32[$370>>2]|0;
     $372 = ((($vf)) + 92|0);
     HEAP32[$372>>2] = $371;
     HEAP32[$4>>2] = 3;
     $374 = $371;
    }
    $373 = ((($vf)) + 120|0);
    (_ogg_stream_reset_serialno($373,$374)|0);
    (_ogg_stream_pagein($373,$og)|0);
    $375 = $359;
    $376 = $375;
    $377 = HEAP32[$376>>2]|0;
    $378 = (($375) + 4)|0;
    $379 = $378;
    $380 = HEAP32[$379>>2]|0;
    $589 = $380;$591 = $377;
   }
   $588 = ($589|0)>($1|0);
   $590 = ($591>>>0)>($0>>>0);
   $592 = ($589|0)==($1|0);
   $593 = $592 & $590;
   $594 = $588 | $593;
   if ($594) {
    $result$11$off0 = -129;
   } else {
    $595 = (_ov_pcm_total($vf,-1)|0);
    $596 = tempRet0;
    $597 = ($596|0)<($1|0);
    $598 = ($595>>>0)<($0>>>0);
    $599 = ($596|0)==($1|0);
    $600 = $599 & $598;
    $601 = $597 | $600;
    if ($601) {
     $result$11$off0 = -129;
    } else {
     $602 = ((($vf)) + 104|0);
     ;HEAP32[$602>>2]=0|0;HEAP32[$602+4>>2]=0|0;HEAP32[$602+8>>2]=0|0;HEAP32[$602+12>>2]=0|0;
     $$3 = 0;
     STACKTOP = sp;return ($$3|0);
    }
   }
  }
 } while(0);
 if ((label|0) == 102) {
  $result$11$off0 = $result$10$off0$ph;
 }
 $603 = ((($vf)) + 80|0);
 $604 = $603;
 $605 = $604;
 HEAP32[$605>>2] = -1;
 $606 = (($604) + 4)|0;
 $607 = $606;
 HEAP32[$607>>2] = -1;
 $608 = ((($vf)) + 480|0);
 _vorbis_dsp_clear($608);
 $609 = ((($vf)) + 592|0);
 (_vorbis_block_clear($609)|0);
 HEAP32[$4>>2] = 2;
 $$3 = $result$11$off0;
 STACKTOP = sp;return ($$3|0);
}
function __fetch_and_process_packet($vf,$spanp) {
 $vf = $vf|0;
 $spanp = $spanp|0;
 var $$13 = 0, $$5$ph = 0, $$lcssa98 = 0, $$pr = 0, $$pr11$pr = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0.0, $47 = 0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $i$035 = 0, $link6$0$lcssa = 0, $link6$042 = 0, $og = 0, $op = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $og = sp + 32|0;
 $op = sp;
 $0 = ((($vf)) + 88|0);
 $1 = ((($vf)) + 4|0);
 $2 = ((($vf)) + 480|0);
 $3 = ((($vf)) + 72|0);
 $4 = ((($vf)) + 592|0);
 $5 = ((($vf)) + 104|0);
 $6 = ((($vf)) + 96|0);
 $7 = ((($vf)) + 120|0);
 $8 = ((($op)) + 16|0);
 $9 = ((($vf)) + 76|0);
 $10 = ((($vf)) + 456|0);
 $11 = ((($vf)) + 92|0);
 $12 = ((($vf)) + 64|0);
 $13 = ((($vf)) + 52|0);
 $14 = ((($og)) + 4|0);
 $15 = ($spanp|0)==(0);
 L1: while(1) {
  $16 = HEAP32[$0>>2]|0;
  switch ($16|0) {
  case 3:  {
   $17 = HEAP32[$1>>2]|0;
   $18 = ($17|0)==(0);
   $19 = HEAP32[$3>>2]|0;
   if ($18) {
    $24 = (_vorbis_synthesis_init($2,$19)|0);
    $25 = ($24|0)==(0);
    if (!($25)) {
     $$13 = -137;
     label = 43;
     break L1;
    }
   } else {
    $20 = HEAP32[$6>>2]|0;
    $21 = (($19) + ($20<<5)|0);
    $22 = (_vorbis_synthesis_init($2,$21)|0);
    $23 = ($22|0)==(0);
    if (!($23)) {
     $$13 = -137;
     label = 43;
     break L1;
    }
   }
   (_vorbis_block_init($2,$4)|0);
   HEAP32[$0>>2] = 4;
   ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;HEAP32[$5+12>>2]=0|0;
   label = 7;
   break;
  }
  case 4:  {
   label = 7;
   break;
  }
  default: {
   $117 = $16;
  }
  }
  if ((label|0) == 7) {
   label = 0;
   $26 = HEAP32[$3>>2]|0;
   $27 = (_vorbis_synthesis_halfrate_p($26)|0);
   $28 = (_ogg_stream_packetout($7,$op)|0);
   $29 = ($28|0)==(-1);
   if ($29) {
    $$5$ph = -3;
    break;
   } else {
    $31 = $28;
   }
   while(1) {
    $30 = ($31|0)>(0);
    if (!($30)) {
     break;
    }
    $32 = $8;
    $33 = $32;
    $34 = HEAP32[$33>>2]|0;
    $35 = (($32) + 4)|0;
    $36 = $35;
    $37 = HEAP32[$36>>2]|0;
    $38 = (_vorbis_synthesis($4,$op)|0);
    $39 = ($38|0)==(0);
    if ($39) {
     $$lcssa98 = $27;$57 = $34;$59 = $37;
     label = 11;
     break L1;
    }
    $40 = (_ogg_stream_packetout($7,$op)|0);
    $41 = ($40|0)==(-1);
    if ($41) {
     $$5$ph = -3;
     break L1;
    } else {
     $31 = $40;
    }
   }
   $$pre = HEAP32[$0>>2]|0;
   $117 = $$pre;
  }
  $116 = ($117|0)>(1);
  do {
   if ($116) {
    while(1) {
     (__get_next_page($vf,$og,-1,-1)|0);
     $118 = tempRet0;
     $119 = ($118|0)<(0);
     if ($119) {
      $$13 = -2;
      label = 43;
      break L1;
     }
     $120 = HEAP32[$14>>2]|0;
     $121 = $120 << 3;
     $122 = (+($121|0));
     $123 = +HEAPF64[$5>>3];
     $124 = $123 + $122;
     HEAPF64[$5>>3] = $124;
     $125 = HEAP32[$0>>2]|0;
     $126 = ($125|0)==(4);
     if (!($126)) {
      $138 = $125;
      break;
     }
     $127 = HEAP32[$11>>2]|0;
     $128 = (_ogg_page_serialno($og)|0);
     $129 = ($127|0)==($128|0);
     if ($129) {
      label = 31;
      break;
     }
     $130 = (_ogg_page_bos($og)|0);
     $131 = ($130|0)==(0);
     if (!($131)) {
      label = 28;
      break;
     }
    }
    if ((label|0) == 28) {
     label = 0;
     if ($15) {
      $$13 = -2;
      label = 43;
      break L1;
     }
     _vorbis_dsp_clear($2);
     (_vorbis_block_clear($4)|0);
     HEAP32[$0>>2] = 2;
     $132 = HEAP32[$1>>2]|0;
     $133 = ($132|0)==(0);
     if (!($133)) {
      label = 34;
      break;
     }
     $134 = HEAP32[$3>>2]|0;
     _vorbis_info_clear($134);
     $135 = HEAP32[$9>>2]|0;
     _vorbis_comment_clear($135);
     label = 31;
    }
    if ((label|0) == 31) {
     label = 0;
     $$pr11$pr = HEAP32[$0>>2]|0;
     $136 = ($$pr11$pr|0)==(4);
     if ($136) {
      break;
     } else {
      $138 = $$pr11$pr;
     }
    }
    $137 = ($138|0)<(3);
    if ($137) {
     label = 33;
    }
   } else {
    label = 33;
   }
  } while(0);
  if ((label|0) == 33) {
   label = 0;
   $$pr = HEAP32[$1>>2]|0;
   $139 = ($$pr|0)==(0);
   if ($139) {
    $150 = HEAP32[$3>>2]|0;
    $151 = HEAP32[$9>>2]|0;
    $152 = (__fetch_headers($vf,$150,$151,0,0,$og)|0);
    $153 = ($152|0)==(0);
    if (!($153)) {
     $$13 = $152;
     label = 43;
     break;
    }
    $154 = HEAP32[$10>>2]|0;
    HEAP32[$11>>2] = $154;
    $155 = HEAP32[$6>>2]|0;
    $156 = (($155) + 1)|0;
    HEAP32[$6>>2] = $156;
   } else {
    label = 34;
   }
  }
  if ((label|0) == 34) {
   label = 0;
   $140 = (_ogg_page_serialno($og)|0);
   $141 = HEAP32[$13>>2]|0;
   $142 = ($141|0)>(0);
   L36: do {
    if ($142) {
     $143 = HEAP32[$12>>2]|0;
     $link6$042 = 0;
     while(1) {
      $144 = (($143) + ($link6$042<<2)|0);
      $145 = HEAP32[$144>>2]|0;
      $146 = ($145|0)==($140|0);
      if ($146) {
       $link6$0$lcssa = $link6$042;
       break L36;
      }
      $147 = (($link6$042) + 1)|0;
      $148 = ($147|0)<($141|0);
      if ($148) {
       $link6$042 = $147;
      } else {
       $link6$0$lcssa = $147;
       break;
      }
     }
    } else {
     $link6$0$lcssa = 0;
    }
   } while(0);
   $149 = ($link6$0$lcssa|0)==($141|0);
   if ($149) {
    continue;
   }
   HEAP32[$11>>2] = $140;
   HEAP32[$6>>2] = $link6$0$lcssa;
   (_ogg_stream_reset_serialno($7,$140)|0);
   HEAP32[$0>>2] = 3;
  }
  (_ogg_stream_pagein($7,$og)|0);
 }
 if ((label|0) == 11) {
  $42 = (_vorbis_synthesis_pcmout($2,0)|0);
  $43 = ($42|0)==(0);
  if ($43) {
   (_vorbis_synthesis_blockin($2,$4)|0);
   $44 = (_vorbis_synthesis_pcmout($2,0)|0);
   $45 = $44 << $$lcssa98;
   $46 = (+($45|0));
   $47 = ((($vf)) + 112|0);
   $48 = +HEAPF64[$47>>3];
   $49 = $48 + $46;
   HEAPF64[$47>>3] = $49;
   $50 = ((($op)) + 4|0);
   $51 = HEAP32[$50>>2]|0;
   $52 = $51 << 3;
   $53 = (+($52|0));
   $54 = +HEAPF64[$5>>3];
   $55 = $54 + $53;
   HEAPF64[$5>>3] = $55;
   $56 = ($57|0)==(-1);
   $58 = ($59|0)==(-1);
   $60 = $56 & $58;
   if ($60) {
    $$5$ph = 1;
   } else {
    $61 = ((($op)) + 12|0);
    $62 = HEAP32[$61>>2]|0;
    $63 = ($62|0)==(0);
    if ($63) {
     $64 = HEAP32[$1>>2]|0;
     $65 = ($64|0)==(0);
     if ($65) {
      $81 = $59;$83 = $57;$92 = 0;
     } else {
      $66 = HEAP32[$6>>2]|0;
      $67 = ($66|0)>(0);
      if ($67) {
       $68 = $66 << 1;
       $69 = ((($vf)) + 68|0);
       $70 = HEAP32[$69>>2]|0;
       $71 = (($70) + ($68<<3)|0);
       $72 = $71;
       $73 = $72;
       $74 = HEAP32[$73>>2]|0;
       $75 = (($72) + 4)|0;
       $76 = $75;
       $77 = HEAP32[$76>>2]|0;
       $78 = (_i64Subtract(($57|0),($59|0),($74|0),($77|0))|0);
       $79 = tempRet0;
       $81 = $79;$83 = $78;$92 = $66;
      } else {
       $81 = $59;$83 = $57;$92 = $66;
      }
     }
     $80 = ($81|0)<(0);
     $82 = $80 ? 0 : $83;
     $84 = $80 ? 0 : $81;
     $85 = (_vorbis_synthesis_pcmout($2,0)|0);
     $86 = $85 << $$lcssa98;
     $87 = ($86|0)<(0);
     $88 = $87 << 31 >> 31;
     $89 = (_i64Subtract(($82|0),($84|0),($86|0),($88|0))|0);
     $90 = tempRet0;
     $91 = ($92|0)>(0);
     if ($91) {
      $93 = ((($vf)) + 68|0);
      $94 = HEAP32[$93>>2]|0;
      $104 = $89;$105 = $90;$i$035 = 0;
      while(1) {
       $95 = $i$035 << 1;
       $96 = $95 | 1;
       $97 = (($94) + ($96<<3)|0);
       $98 = $97;
       $99 = $98;
       $100 = HEAP32[$99>>2]|0;
       $101 = (($98) + 4)|0;
       $102 = $101;
       $103 = HEAP32[$102>>2]|0;
       $106 = (_i64Add(($100|0),($103|0),($104|0),($105|0))|0);
       $107 = tempRet0;
       $108 = (($i$035) + 1)|0;
       $exitcond = ($108|0)==($92|0);
       if ($exitcond) {
        $112 = $106;$115 = $107;
        break;
       } else {
        $104 = $106;$105 = $107;$i$035 = $108;
       }
      }
     } else {
      $112 = $89;$115 = $90;
     }
     $109 = ((($vf)) + 80|0);
     $110 = $109;
     $111 = $110;
     HEAP32[$111>>2] = $112;
     $113 = (($110) + 4)|0;
     $114 = $113;
     HEAP32[$114>>2] = $115;
     $$5$ph = 1;
    } else {
     $$5$ph = 1;
    }
   }
  } else {
   $$5$ph = -129;
  }
 }
 else if ((label|0) == 43) {
  STACKTOP = sp;return ($$13|0);
 }
 $$13 = $$5$ph;
 STACKTOP = sp;return ($$13|0);
}
function _ov_seekable($vf) {
 $vf = $vf|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($vf)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function _ov_time_total($vf,$i) {
 $vf = $vf|0;
 $i = $i|0;
 var $$0 = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0.0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $acc$01 = 0.0, $i1$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($vf)) + 88|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)<(2);
 if ($2) {
  $$0 = -131.0;
  return (+$$0);
 }
 $3 = ((($vf)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0);
 if ($5) {
  $$0 = -131.0;
  return (+$$0);
 }
 $6 = ((($vf)) + 52|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)>($i|0);
 if (!($8)) {
  $$0 = -131.0;
  return (+$$0);
 }
 $9 = ($i|0)<(0);
 if (!($9)) {
  $15 = $i << 1;
  $16 = $15 | 1;
  $17 = ((($vf)) + 68|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = (($18) + ($16<<3)|0);
  $20 = $19;
  $21 = $20;
  $22 = HEAP32[$21>>2]|0;
  $23 = (($20) + 4)|0;
  $24 = $23;
  $25 = HEAP32[$24>>2]|0;
  $26 = (+($22>>>0)) + (4294967296.0*(+($25|0)));
  $27 = ((($vf)) + 72|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = (((($28) + ($i<<5)|0)) + 8|0);
  $30 = HEAP32[$29>>2]|0;
  $31 = (+($30|0));
  $32 = $26 / $31;
  $$0 = $32;
  return (+$$0);
 }
 $10 = ($7|0)>(0);
 if ($10) {
  $acc$01 = 0.0;$i1$02 = 0;
 } else {
  $$0 = 0.0;
  return (+$$0);
 }
 while(1) {
  $11 = (+_ov_time_total($vf,$i1$02));
  $12 = $acc$01 + $11;
  $13 = (($i1$02) + 1)|0;
  $14 = ($13|0)<($7|0);
  if ($14) {
   $acc$01 = $12;$i1$02 = $13;
  } else {
   $$0 = $12;
   break;
  }
 }
 return (+$$0);
}
function _ov_time_seek($vf,$seconds) {
 $vf = $vf|0;
 $seconds = +$seconds;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0.0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0;
 var $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $link$0$lcssa = 0, $link$06 = 0, $time_total$0$lcssa = 0.0, $time_total$05 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($vf)) + 88|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)<(2);
 if ($2) {
  $$0 = -131;
  return ($$0|0);
 }
 $3 = ((($vf)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0);
 if ($5) {
  $$0 = -138;
  return ($$0|0);
 }
 $6 = $seconds < 0.0;
 if ($6) {
  $$0 = -131;
  return ($$0|0);
 }
 $7 = ((($vf)) + 52|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)>(0);
 L10: do {
  if ($9) {
   $10 = ((($vf)) + 68|0);
   $24 = 0;$25 = 0;$link$06 = 0;$time_total$05 = 0.0;
   while(1) {
    $11 = (+_ov_time_total($vf,$link$06));
    $12 = $time_total$05 + $11;
    $13 = $12 > $seconds;
    if ($13) {
     $32 = $24;$33 = $25;$link$0$lcssa = $link$06;$time_total$0$lcssa = $time_total$05;
     break L10;
    }
    $14 = $link$06 << 1;
    $15 = $14 | 1;
    $16 = HEAP32[$10>>2]|0;
    $17 = (($16) + ($15<<3)|0);
    $18 = $17;
    $19 = $18;
    $20 = HEAP32[$19>>2]|0;
    $21 = (($18) + 4)|0;
    $22 = $21;
    $23 = HEAP32[$22>>2]|0;
    $26 = (_i64Add(($20|0),($23|0),($24|0),($25|0))|0);
    $27 = tempRet0;
    $28 = (($link$06) + 1)|0;
    $29 = ($28|0)<($8|0);
    if ($29) {
     $24 = $26;$25 = $27;$link$06 = $28;$time_total$05 = $12;
    } else {
     $32 = $26;$33 = $27;$link$0$lcssa = $28;$time_total$0$lcssa = $12;
     break;
    }
   }
  } else {
   $32 = 0;$33 = 0;$link$0$lcssa = 0;$time_total$0$lcssa = 0.0;
  }
 } while(0);
 $30 = ($link$0$lcssa|0)==($8|0);
 if ($30) {
  $$0 = -131;
  return ($$0|0);
 }
 $31 = (+($32>>>0)) + (4294967296.0*(+($33|0)));
 $34 = $seconds - $time_total$0$lcssa;
 $35 = ((($vf)) + 72|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = (((($36) + ($link$0$lcssa<<5)|0)) + 8|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = (+($38|0));
 $40 = $34 * $39;
 $41 = $31 + $40;
 $42 = (~~$41)>>>0;
 $43 = +Math_abs($41) >= 1.0 ? $41 > 0.0 ? (~~+Math_min(+Math_floor($41 / 4294967296.0), 4294967295.0)) >>> 0 : ~~+Math_ceil(($41 - +(~~$41 >>> 0)) / 4294967296.0) >>> 0 : 0;
 $44 = (_ov_pcm_seek($vf,$42,$43)|0);
 $$0 = $44;
 return ($$0|0);
}
function _ov_info($vf,$link) {
 $vf = $vf|0;
 $link = $link|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($vf)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  $18 = ((($vf)) + 72|0);
  $19 = HEAP32[$18>>2]|0;
  $$0 = $19;
  return ($$0|0);
 }
 $3 = ($link|0)<(0);
 if ($3) {
  $4 = ((($vf)) + 88|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ($5|0)>(2);
  $7 = ((($vf)) + 72|0);
  $8 = HEAP32[$7>>2]|0;
  if (!($6)) {
   $$0 = $8;
   return ($$0|0);
  }
  $9 = ((($vf)) + 96|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = (($8) + ($10<<5)|0);
  $$0 = $11;
  return ($$0|0);
 } else {
  $12 = ((($vf)) + 52|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ($13|0)>($link|0);
  if (!($14)) {
   $$0 = 0;
   return ($$0|0);
  }
  $15 = ((($vf)) + 72|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = (($16) + ($link<<5)|0);
  $$0 = $17;
  return ($$0|0);
 }
 return (0)|0;
}
function _ov_comment($vf,$link) {
 $vf = $vf|0;
 $link = $link|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($vf)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  $18 = ((($vf)) + 76|0);
  $19 = HEAP32[$18>>2]|0;
  $$0 = $19;
  return ($$0|0);
 }
 $3 = ($link|0)<(0);
 if ($3) {
  $4 = ((($vf)) + 88|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ($5|0)>(2);
  $7 = ((($vf)) + 76|0);
  $8 = HEAP32[$7>>2]|0;
  if (!($6)) {
   $$0 = $8;
   return ($$0|0);
  }
  $9 = ((($vf)) + 96|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = (($8) + ($10<<4)|0);
  $$0 = $11;
  return ($$0|0);
 } else {
  $12 = ((($vf)) + 52|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ($13|0)>($link|0);
  if (!($14)) {
   $$0 = 0;
   return ($$0|0);
  }
  $15 = ((($vf)) + 76|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = (($16) + ($link<<4)|0);
  $$0 = $17;
  return ($$0|0);
 }
 return (0)|0;
}
function _ov_read_filter($vf,$buffer,$length,$bigendianp,$word,$sgned,$bitstream,$filter,$filter_param) {
 $vf = $vf|0;
 $buffer = $buffer|0;
 $length = $length|0;
 $bigendianp = $bigendianp|0;
 $word = $word|0;
 $sgned = $sgned|0;
 $bitstream = $bitstream|0;
 $filter = $filter|0;
 $filter_param = $filter_param|0;
 var $$0$i = 0, $$0124$us = 0, $$10$us = 0, $$1222$us = 0, $$2345$us = 0, $$3 = 0, $$3440$us = 0, $$6 = 0, $$7$us = 0, $$8$us = 0, $$9$us = 0, $$lcssa = 0, $$pr = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0.0;
 var $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dest$026$us = 0, $dest3$033$us = 0;
 var $exitcond = 0, $exitcond55 = 0, $exitcond57 = 0, $exitcond60 = 0, $i$021$us = 0, $i$130$us = 0, $i$237$us = 0, $i$339$us = 0, $j$023$us = 0, $j$125$us = 0, $j$232$us = 0, $j$344$us = 0, $or$cond = 0, $or$cond64 = 0, $pcm = 0, $scevgep = 0, $scevgep59 = 0, $val$0$us = 0, $val$2$us = 0, $val$3$us = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $pcm = sp;
 $0 = ((($vf)) + 88|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)<(2);
 if ($2) {
  $$3 = -131;
  STACKTOP = sp;return ($$3|0);
 }
 $3 = ((($vf)) + 480|0);
 $5 = $1;
 while(1) {
  $4 = ($5|0)==(4);
  if ($4) {
   $6 = (_vorbis_synthesis_pcmout($3,$pcm)|0);
   $7 = ($6|0)==(0);
   if (!($7)) {
    $$lcssa = $6;
    label = 8;
    break;
   }
  }
  $8 = (__fetch_and_process_packet($vf,1)|0);
  $9 = ($8|0)==(-2);
  if ($9) {
   $$3 = 0;
   label = 41;
   break;
  }
  $10 = ($8|0)<(1);
  if ($10) {
   $$3 = $8;
   label = 41;
   break;
  }
  $$pr = HEAP32[$0>>2]|0;
  $5 = $$pr;
 }
 if ((label|0) == 8) {
  $11 = ($$lcssa|0)>(0);
  if (!($11)) {
   $$3 = $$lcssa;
   STACKTOP = sp;return ($$3|0);
  }
  $12 = ((($vf)) + 4|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ($13|0)==(0);
  if ($14) {
   $22 = ((($vf)) + 72|0);
   $23 = HEAP32[$22>>2]|0;
   $$0$i = $23;
  } else {
   $15 = HEAP32[$0>>2]|0;
   $16 = ($15|0)>(2);
   $17 = ((($vf)) + 72|0);
   $18 = HEAP32[$17>>2]|0;
   if ($16) {
    $19 = ((($vf)) + 96|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = (($18) + ($20<<5)|0);
    $$0$i = $21;
   } else {
    $$0$i = $18;
   }
  }
  $24 = ((($$0$i)) + 4|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = Math_imul($25, $word)|0;
  $27 = (($length|0) / ($26|0))&-1;
  $28 = ($$lcssa|0)>($27|0);
  $$6 = $28 ? $27 : $$lcssa;
  $29 = ($$6|0)<(1);
  if ($29) {
   $$3 = -131;
   STACKTOP = sp;return ($$3|0);
  }
  $30 = ($filter|0)==(0|0);
  if (!($30)) {
   $31 = HEAP32[$pcm>>2]|0;
   FUNCTION_TABLE_viiii[$filter & 8191]($31,$25,$$6,$filter_param);
  }
  $32 = ($word|0)==(1);
  $33 = ($sgned|0)!=(0);
  L29: do {
   if ($32) {
    $34 = $33 ? 0 : 128;
    $35 = ($$6|0)>(0);
    $36 = ($25|0)>(0);
    $or$cond = $35 & $36;
    if ($or$cond) {
     $$0124$us = $buffer;$j$023$us = 0;
     while(1) {
      $$1222$us = $$0124$us;$i$021$us = 0;
      while(1) {
       $37 = HEAP32[$pcm>>2]|0;
       $38 = (($37) + ($i$021$us<<2)|0);
       $39 = HEAP32[$38>>2]|0;
       $40 = (($39) + ($j$023$us<<2)|0);
       $41 = +HEAPF32[$40>>2];
       $42 = $41 * 128.0;
       $43 = $42;
       $44 = $43 + 0.5;
       $45 = (+Math_floor((+$44)));
       $46 = (~~(($45)));
       $47 = ($46|0)>(127);
       $48 = ($46|0)<(-128);
       $$7$us = $48 ? -128 : $46;
       $val$0$us = $47 ? 127 : $$7$us;
       $49 = (($val$0$us) + ($34))|0;
       $50 = $49&255;
       $51 = ((($$1222$us)) + 1|0);
       HEAP8[$$1222$us>>0] = $50;
       $52 = (($i$021$us) + 1)|0;
       $exitcond = ($52|0)==($25|0);
       if ($exitcond) {
        break;
       } else {
        $$1222$us = $51;$i$021$us = $52;
       }
      }
      $scevgep = (($$0124$us) + ($25)|0);
      $53 = (($j$023$us) + 1)|0;
      $54 = ($53|0)<($$6|0);
      if ($54) {
       $$0124$us = $scevgep;$j$023$us = $53;
      } else {
       break;
      }
     }
    }
   } else {
    $55 = $33 ? 0 : 32768;
    $56 = ($bigendianp|0)==(0);
    if (!($56)) {
     $57 = ($$6|0)>(0);
     $58 = ($25|0)>(0);
     $or$cond64 = $57 & $58;
     if (!($or$cond64)) {
      break;
     }
     $59 = $25 << 1;
     $$2345$us = $buffer;$j$344$us = 0;
     while(1) {
      $$3440$us = $$2345$us;$i$339$us = 0;
      while(1) {
       $60 = HEAP32[$pcm>>2]|0;
       $61 = (($60) + ($i$339$us<<2)|0);
       $62 = HEAP32[$61>>2]|0;
       $63 = (($62) + ($j$344$us<<2)|0);
       $64 = +HEAPF32[$63>>2];
       $65 = $64 * 32768.0;
       $66 = $65;
       $67 = $66 + 0.5;
       $68 = (+Math_floor((+$67)));
       $69 = (~~(($68)));
       $70 = ($69|0)>(32767);
       $71 = ($69|0)<(-32768);
       $$10$us = $71 ? -32768 : $69;
       $val$3$us = $70 ? 32767 : $$10$us;
       $72 = (($val$3$us) + ($55))|0;
       $73 = $72 >>> 8;
       $74 = $73&255;
       $75 = ((($$3440$us)) + 1|0);
       HEAP8[$$3440$us>>0] = $74;
       $76 = $72&255;
       $77 = ((($$3440$us)) + 2|0);
       HEAP8[$75>>0] = $76;
       $78 = (($i$339$us) + 1)|0;
       $exitcond60 = ($78|0)==($25|0);
       if ($exitcond60) {
        break;
       } else {
        $$3440$us = $77;$i$339$us = $78;
       }
      }
      $scevgep59 = (($$2345$us) + ($59)|0);
      $79 = (($j$344$us) + 1)|0;
      $80 = ($79|0)<($$6|0);
      if ($80) {
       $$2345$us = $scevgep59;$j$344$us = $79;
      } else {
       break L29;
      }
     }
    }
    $81 = ($25|0)>(0);
    if ($33) {
     if (!($81)) {
      break;
     }
     $102 = HEAP32[$pcm>>2]|0;
     $103 = ($$6|0)>(0);
     if ($103) {
      $i$130$us = 0;
     } else {
      break;
     }
     while(1) {
      $104 = (($102) + ($i$130$us<<2)|0);
      $105 = HEAP32[$104>>2]|0;
      $106 = (($buffer) + ($i$130$us<<1)|0);
      $dest$026$us = $106;$j$125$us = 0;
      while(1) {
       $107 = (($105) + ($j$125$us<<2)|0);
       $108 = +HEAPF32[$107>>2];
       $109 = $108 * 32768.0;
       $110 = $109;
       $111 = $110 + 0.5;
       $112 = (+Math_floor((+$111)));
       $113 = (~~(($112)));
       $114 = ($113|0)>(32767);
       $115 = ($113|0)<(-32768);
       $$8$us = $115 ? -32768 : $113;
       $116 = $$8$us&65535;
       $117 = $114 ? 32767 : $116;
       HEAP16[$dest$026$us>>1] = $117;
       $118 = (($dest$026$us) + ($25<<1)|0);
       $119 = (($j$125$us) + 1)|0;
       $120 = ($119|0)<($$6|0);
       if ($120) {
        $dest$026$us = $118;$j$125$us = $119;
       } else {
        break;
       }
      }
      $121 = (($i$130$us) + 1)|0;
      $exitcond55 = ($121|0)==($25|0);
      if ($exitcond55) {
       break;
      } else {
       $i$130$us = $121;
      }
     }
    } else {
     if (!($81)) {
      break;
     }
     $82 = HEAP32[$pcm>>2]|0;
     $83 = ($$6|0)>(0);
     if ($83) {
      $i$237$us = 0;
     } else {
      break;
     }
     while(1) {
      $84 = (($82) + ($i$237$us<<2)|0);
      $85 = HEAP32[$84>>2]|0;
      $86 = (($buffer) + ($i$237$us<<1)|0);
      $dest3$033$us = $86;$j$232$us = 0;
      while(1) {
       $87 = (($85) + ($j$232$us<<2)|0);
       $88 = +HEAPF32[$87>>2];
       $89 = $88 * 32768.0;
       $90 = $89;
       $91 = $90 + 0.5;
       $92 = (+Math_floor((+$91)));
       $93 = (~~(($92)));
       $94 = ($93|0)>(32767);
       $95 = ($93|0)<(-32768);
       $$9$us = $95 ? -32768 : $93;
       $val$2$us = $94 ? 32767 : $$9$us;
       $96 = (($val$2$us) + ($55))|0;
       $97 = $96&65535;
       HEAP16[$dest3$033$us>>1] = $97;
       $98 = (($dest3$033$us) + ($25<<1)|0);
       $99 = (($j$232$us) + 1)|0;
       $100 = ($99|0)<($$6|0);
       if ($100) {
        $dest3$033$us = $98;$j$232$us = $99;
       } else {
        break;
       }
      }
      $101 = (($i$237$us) + 1)|0;
      $exitcond57 = ($101|0)==($25|0);
      if ($exitcond57) {
       break;
      } else {
       $i$237$us = $101;
      }
     }
    }
   }
  } while(0);
  (_vorbis_synthesis_read($3,$$6)|0);
  $122 = ((($vf)) + 72|0);
  $123 = HEAP32[$122>>2]|0;
  $124 = (_vorbis_synthesis_halfrate_p($123)|0);
  $125 = $$6 << $124;
  $126 = ($125|0)<(0);
  $127 = $126 << 31 >> 31;
  $128 = ((($vf)) + 80|0);
  $129 = $128;
  $130 = $129;
  $131 = HEAP32[$130>>2]|0;
  $132 = (($129) + 4)|0;
  $133 = $132;
  $134 = HEAP32[$133>>2]|0;
  $135 = (_i64Add(($125|0),($127|0),($131|0),($134|0))|0);
  $136 = tempRet0;
  $137 = $128;
  $138 = $137;
  HEAP32[$138>>2] = $135;
  $139 = (($137) + 4)|0;
  $140 = $139;
  HEAP32[$140>>2] = $136;
  $141 = ($bitstream|0)==(0|0);
  if (!($141)) {
   $142 = ((($vf)) + 96|0);
   $143 = HEAP32[$142>>2]|0;
   HEAP32[$bitstream>>2] = $143;
  }
  $144 = Math_imul($$6, $26)|0;
  $$3 = $144;
  STACKTOP = sp;return ($$3|0);
 }
 else if ((label|0) == 41) {
  STACKTOP = sp;return ($$3|0);
 }
 return (0)|0;
}
function _ov_read($vf,$buffer,$length,$bigendianp,$word,$sgned,$bitstream) {
 $vf = $vf|0;
 $buffer = $buffer|0;
 $length = $length|0;
 $bigendianp = $bigendianp|0;
 $word = $word|0;
 $sgned = $sgned|0;
 $bitstream = $bitstream|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_ov_read_filter($vf,$buffer,$length,$bigendianp,$word,$sgned,$bitstream,0,0)|0);
 return ($0|0);
}
function _oggpack_writeinit($b) {
 $b = $b|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$b>>2]=0|0;HEAP32[$b+4>>2]=0|0;HEAP32[$b+8>>2]=0|0;HEAP32[$b+12>>2]=0|0;
 $0 = (_malloc(256)|0);
 $1 = ((($b)) + 8|0);
 HEAP32[$1>>2] = $0;
 $2 = ((($b)) + 12|0);
 HEAP32[$2>>2] = $0;
 HEAP8[$0>>0] = 0;
 $3 = ((($b)) + 16|0);
 HEAP32[$3>>2] = 256;
 return;
}
function _oggpack_write($b,$value,$bits) {
 $b = $b|0;
 $value = $value|0;
 $bits = $bits|0;
 var $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($bits>>>0)>(32);
 do {
  if (!($0)) {
   $1 = HEAP32[$b>>2]|0;
   $2 = ((($b)) + 16|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = (($3) + -4)|0;
   $5 = ($1|0)<($4|0);
   $$phi$trans$insert = ((($b)) + 12|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0;
   if ($5) {
    $25 = $$pre;
   } else {
    $6 = ($$pre|0)==(0|0);
    if ($6) {
     return;
    }
    $7 = ($3|0)>(2147483391);
    if ($7) {
     break;
    }
    $8 = ((($b)) + 8|0);
    $9 = HEAP32[$8>>2]|0;
    $10 = (($3) + 256)|0;
    $11 = (_realloc($9,$10)|0);
    $12 = ($11|0)==(0|0);
    if ($12) {
     break;
    }
    HEAP32[$8>>2] = $11;
    $13 = HEAP32[$2>>2]|0;
    $14 = (($13) + 256)|0;
    HEAP32[$2>>2] = $14;
    $15 = HEAP32[$b>>2]|0;
    $16 = (($11) + ($15)|0);
    HEAP32[$$phi$trans$insert>>2] = $16;
    $25 = $16;
   }
   $17 = (124600 + ($bits<<2)|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = $18 & $value;
   $20 = ((($b)) + 4|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = (($21) + ($bits))|0;
   $23 = $19 << $21;
   $24 = HEAP8[$25>>0]|0;
   $26 = $24&255;
   $27 = $26 | $23;
   $28 = $27&255;
   HEAP8[$25>>0] = $28;
   $29 = ($22|0)>(7);
   do {
    if ($29) {
     $30 = HEAP32[$20>>2]|0;
     $31 = (8 - ($30))|0;
     $32 = $19 >>> $31;
     $33 = $32&255;
     $34 = HEAP32[$$phi$trans$insert>>2]|0;
     $35 = ((($34)) + 1|0);
     HEAP8[$35>>0] = $33;
     $36 = ($22|0)>(15);
     if ($36) {
      $37 = HEAP32[$20>>2]|0;
      $38 = (16 - ($37))|0;
      $39 = $19 >>> $38;
      $40 = $39&255;
      $41 = HEAP32[$$phi$trans$insert>>2]|0;
      $42 = ((($41)) + 2|0);
      HEAP8[$42>>0] = $40;
      $43 = ($22|0)>(23);
      if ($43) {
       $44 = HEAP32[$20>>2]|0;
       $45 = (24 - ($44))|0;
       $46 = $19 >>> $45;
       $47 = $46&255;
       $48 = HEAP32[$$phi$trans$insert>>2]|0;
       $49 = ((($48)) + 3|0);
       HEAP8[$49>>0] = $47;
       $50 = ($22|0)>(31);
       if ($50) {
        $51 = HEAP32[$20>>2]|0;
        $52 = ($51|0)==(0);
        if ($52) {
         $58 = HEAP32[$$phi$trans$insert>>2]|0;
         $59 = ((($58)) + 4|0);
         HEAP8[$59>>0] = 0;
         break;
        } else {
         $53 = (32 - ($51))|0;
         $54 = $19 >>> $53;
         $55 = $54&255;
         $56 = HEAP32[$$phi$trans$insert>>2]|0;
         $57 = ((($56)) + 4|0);
         HEAP8[$57>>0] = $55;
         break;
        }
       }
      }
     }
    }
   } while(0);
   $60 = (($22|0) / 8)&-1;
   $61 = HEAP32[$b>>2]|0;
   $62 = (($61) + ($60))|0;
   HEAP32[$b>>2] = $62;
   $63 = HEAP32[$$phi$trans$insert>>2]|0;
   $64 = (($63) + ($60)|0);
   HEAP32[$$phi$trans$insert>>2] = $64;
   $65 = $22 & 7;
   HEAP32[$20>>2] = $65;
   return;
  }
 } while(0);
 $66 = ((($b)) + 8|0);
 $67 = HEAP32[$66>>2]|0;
 $68 = ($67|0)==(0|0);
 if (!($68)) {
  _free($67);
 }
 ;HEAP32[$b>>2]=0|0;HEAP32[$b+4>>2]=0|0;HEAP32[$b+8>>2]=0|0;HEAP32[$b+12>>2]=0|0;HEAP32[$b+16>>2]=0|0;
 return;
}
function _oggpack_writeclear($b) {
 $b = $b|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($b)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  _free($1);
 }
 ;HEAP32[$b>>2]=0|0;HEAP32[$b+4>>2]=0|0;HEAP32[$b+8>>2]=0|0;HEAP32[$b+12>>2]=0|0;HEAP32[$b+16>>2]=0|0;
 return;
}
function _oggpack_readinit($b,$buf,$bytes) {
 $b = $b|0;
 $buf = $buf|0;
 $bytes = $bytes|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$b>>2]=0|0;HEAP32[$b+4>>2]=0|0;HEAP32[$b+8>>2]=0|0;HEAP32[$b+12>>2]=0|0;
 $0 = ((($b)) + 12|0);
 HEAP32[$0>>2] = $buf;
 $1 = ((($b)) + 8|0);
 HEAP32[$1>>2] = $buf;
 $2 = ((($b)) + 16|0);
 HEAP32[$2>>2] = $bytes;
 return;
}
function _oggpack_look($b,$bits) {
 $b = $b|0;
 $bits = $bits|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $ret$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($bits>>>0)>(32);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = (124600 + ($bits<<2)|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($b)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + ($bits))|0;
 $6 = HEAP32[$b>>2]|0;
 $7 = ((($b)) + 16|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (($8) + -4)|0;
 $10 = ($6|0)<($9|0);
 if (!($10)) {
  $11 = (($5) + 7)|0;
  $12 = $11 >> 3;
  $13 = (($8) - ($12))|0;
  $14 = ($6|0)>($13|0);
  if ($14) {
   $$0 = -1;
   return ($$0|0);
  }
  $15 = ($5|0)==(0);
  if ($15) {
   $$0 = 0;
   return ($$0|0);
  }
 }
 $16 = ((($b)) + 12|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = HEAP8[$17>>0]|0;
 $19 = $18&255;
 $20 = $19 >>> $4;
 $21 = ($5|0)>(8);
 if ($21) {
  $22 = ((($17)) + 1|0);
  $23 = HEAP8[$22>>0]|0;
  $24 = $23&255;
  $25 = (8 - ($4))|0;
  $26 = $24 << $25;
  $27 = $26 | $20;
  $28 = ($5|0)>(16);
  if ($28) {
   $29 = ((($17)) + 2|0);
   $30 = HEAP8[$29>>0]|0;
   $31 = $30&255;
   $32 = (16 - ($4))|0;
   $33 = $31 << $32;
   $34 = $33 | $27;
   $35 = ($5|0)>(24);
   if ($35) {
    $36 = ((($17)) + 3|0);
    $37 = HEAP8[$36>>0]|0;
    $38 = $37&255;
    $39 = (24 - ($4))|0;
    $40 = $38 << $39;
    $41 = $40 | $34;
    $42 = ($5|0)<(33);
    $43 = ($4|0)==(0);
    $or$cond = $42 | $43;
    if ($or$cond) {
     $ret$0 = $41;
    } else {
     $44 = ((($17)) + 4|0);
     $45 = HEAP8[$44>>0]|0;
     $46 = $45&255;
     $47 = (32 - ($4))|0;
     $48 = $46 << $47;
     $49 = $48 | $41;
     $ret$0 = $49;
    }
   } else {
    $ret$0 = $34;
   }
  } else {
   $ret$0 = $27;
  }
 } else {
  $ret$0 = $20;
 }
 $50 = $ret$0 & $2;
 $$0 = $50;
 return ($$0|0);
}
function _oggpack_adv($b,$bits) {
 $b = $b|0;
 $bits = $bits|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($b)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + ($bits))|0;
 $3 = HEAP32[$b>>2]|0;
 $4 = ((($b)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($2) + 7)|0;
 $7 = $6 >> 3;
 $8 = (($5) - ($7))|0;
 $9 = ($3|0)>($8|0);
 if ($9) {
  $16 = ((($b)) + 12|0);
  HEAP32[$16>>2] = 0;
  HEAP32[$b>>2] = $5;
  $storemerge = 1;
  HEAP32[$0>>2] = $storemerge;
  return;
 } else {
  $10 = (($2|0) / 8)&-1;
  $11 = ((($b)) + 12|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = (($12) + ($10)|0);
  HEAP32[$11>>2] = $13;
  $14 = (($3) + ($10))|0;
  HEAP32[$b>>2] = $14;
  $15 = $2 & 7;
  $storemerge = $15;
  HEAP32[$0>>2] = $storemerge;
  return;
 }
}
function _oggpack_read($b,$bits) {
 $b = $b|0;
 $bits = $bits|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phi3Z2D = 0, $$pre$phi5Z2D = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $ret$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($bits>>>0)>(32);
 do {
  if ($0) {
   $$phi$trans$insert = ((($b)) + 16|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0;
   $$pre4 = ((($b)) + 4|0);
   $$pre$phi3Z2D = $b;$$pre$phi5Z2D = $$pre4;$56 = $$pre;
  } else {
   $1 = (124600 + ($bits<<2)|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = ((($b)) + 4|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = (($4) + ($bits))|0;
   $6 = HEAP32[$b>>2]|0;
   $7 = ((($b)) + 16|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = (($8) + -4)|0;
   $10 = ($6|0)<($9|0);
   if (!($10)) {
    $11 = (($5) + 7)|0;
    $12 = $11 >> 3;
    $13 = (($8) - ($12))|0;
    $14 = ($6|0)>($13|0);
    if ($14) {
     $$pre$phi3Z2D = $b;$$pre$phi5Z2D = $3;$56 = $8;
     break;
    }
    $15 = ($5|0)==(0);
    if ($15) {
     $$0 = 0;
     return ($$0|0);
    }
   }
   $16 = ((($b)) + 12|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = HEAP8[$17>>0]|0;
   $19 = $18&255;
   $20 = $19 >>> $4;
   $21 = ($5|0)>(8);
   if ($21) {
    $22 = ((($17)) + 1|0);
    $23 = HEAP8[$22>>0]|0;
    $24 = $23&255;
    $25 = (8 - ($4))|0;
    $26 = $24 << $25;
    $27 = $26 | $20;
    $28 = ($5|0)>(16);
    if ($28) {
     $29 = ((($17)) + 2|0);
     $30 = HEAP8[$29>>0]|0;
     $31 = $30&255;
     $32 = (16 - ($4))|0;
     $33 = $31 << $32;
     $34 = $33 | $27;
     $35 = ($5|0)>(24);
     if ($35) {
      $36 = ((($17)) + 3|0);
      $37 = HEAP8[$36>>0]|0;
      $38 = $37&255;
      $39 = (24 - ($4))|0;
      $40 = $38 << $39;
      $41 = $40 | $34;
      $42 = ($5|0)<(33);
      $43 = ($4|0)==(0);
      $or$cond = $42 | $43;
      if ($or$cond) {
       $ret$0 = $41;
      } else {
       $44 = ((($17)) + 4|0);
       $45 = HEAP8[$44>>0]|0;
       $46 = $45&255;
       $47 = (32 - ($4))|0;
       $48 = $46 << $47;
       $49 = $48 | $41;
       $ret$0 = $49;
      }
     } else {
      $ret$0 = $34;
     }
    } else {
     $ret$0 = $27;
    }
   } else {
    $ret$0 = $20;
   }
   $50 = $ret$0 & $2;
   $51 = (($5|0) / 8)&-1;
   $52 = (($17) + ($51)|0);
   HEAP32[$16>>2] = $52;
   $53 = (($6) + ($51))|0;
   HEAP32[$b>>2] = $53;
   $54 = $5 & 7;
   HEAP32[$3>>2] = $54;
   $$0 = $50;
   return ($$0|0);
  }
 } while(0);
 $55 = ((($b)) + 12|0);
 HEAP32[$55>>2] = 0;
 HEAP32[$$pre$phi3Z2D>>2] = $56;
 HEAP32[$$pre$phi5Z2D>>2] = 1;
 $$0 = -1;
 return ($$0|0);
}
function _oggpack_bytes($b) {
 $b = $b|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$b>>2]|0;
 $1 = ((($b)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($2) + 7)|0;
 $4 = (($3|0) / 8)&-1;
 $5 = (($4) + ($0))|0;
 return ($5|0);
}
function _ogg_page_continued($og) {
 $og = $og|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$og>>2]|0;
 $1 = ((($0)) + 5|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = $2&255;
 $4 = $3 & 1;
 return ($4|0);
}
function _ogg_page_bos($og) {
 $og = $og|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$og>>2]|0;
 $1 = ((($0)) + 5|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = $2&255;
 $4 = $3 & 2;
 return ($4|0);
}
function _ogg_page_eos($og) {
 $og = $og|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$og>>2]|0;
 $1 = ((($0)) + 5|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = $2&255;
 $4 = $3 & 4;
 return ($4|0);
}
function _ogg_page_granulepos($og) {
 $og = $og|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$og>>2]|0;
 $1 = ((($0)) + 13|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = $2&255;
 $4 = (_bitshift64Shl(($3|0),0,8)|0);
 $5 = tempRet0;
 $6 = ((($0)) + 12|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 $9 = $4 | $8;
 $10 = (_bitshift64Shl(($9|0),($5|0),8)|0);
 $11 = tempRet0;
 $12 = ((($0)) + 11|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = $13&255;
 $15 = $10 | $14;
 $16 = (_bitshift64Shl(($15|0),($11|0),8)|0);
 $17 = tempRet0;
 $18 = ((($0)) + 10|0);
 $19 = HEAP8[$18>>0]|0;
 $20 = $19&255;
 $21 = $16 | $20;
 $22 = (_bitshift64Shl(($21|0),($17|0),8)|0);
 $23 = tempRet0;
 $24 = ((($0)) + 9|0);
 $25 = HEAP8[$24>>0]|0;
 $26 = $25&255;
 $27 = $22 | $26;
 $28 = (_bitshift64Shl(($27|0),($23|0),8)|0);
 $29 = tempRet0;
 $30 = ((($0)) + 8|0);
 $31 = HEAP8[$30>>0]|0;
 $32 = $31&255;
 $33 = $28 | $32;
 $34 = (_bitshift64Shl(($33|0),($29|0),8)|0);
 $35 = tempRet0;
 $36 = ((($0)) + 7|0);
 $37 = HEAP8[$36>>0]|0;
 $38 = $37&255;
 $39 = $34 | $38;
 $40 = (_bitshift64Shl(($39|0),($35|0),8)|0);
 $41 = tempRet0;
 $42 = ((($0)) + 6|0);
 $43 = HEAP8[$42>>0]|0;
 $44 = $43&255;
 $45 = $40 | $44;
 tempRet0 = ($41);
 return ($45|0);
}
function _ogg_page_serialno($og) {
 $og = $og|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$og>>2]|0;
 $1 = ((($0)) + 14|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = $2&255;
 $4 = ((($0)) + 15|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = $6 << 8;
 $8 = $7 | $3;
 $9 = ((($0)) + 16|0);
 $10 = HEAP8[$9>>0]|0;
 $11 = $10&255;
 $12 = $11 << 16;
 $13 = $8 | $12;
 $14 = ((($0)) + 17|0);
 $15 = HEAP8[$14>>0]|0;
 $16 = $15&255;
 $17 = $16 << 24;
 $18 = $13 | $17;
 return ($18|0);
}
function _ogg_stream_init($os,$serialno) {
 $os = $os|0;
 $serialno = $serialno|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ($os|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 _memset(($os|0),0,360)|0;
 $1 = ((($os)) + 4|0);
 HEAP32[$1>>2] = 16384;
 $2 = ((($os)) + 24|0);
 HEAP32[$2>>2] = 1024;
 $3 = (_malloc(16384)|0);
 HEAP32[$os>>2] = $3;
 $4 = (_malloc(4096)|0);
 $5 = ((($os)) + 16|0);
 HEAP32[$5>>2] = $4;
 $6 = (_malloc(8192)|0);
 $7 = ((($os)) + 20|0);
 HEAP32[$7>>2] = $6;
 $8 = ($3|0)==(0|0);
 do {
  if ($8) {
   $12 = $4;
  } else {
   $9 = ($4|0)==(0|0);
   $10 = ($6|0)==(0|0);
   $or$cond = $10 | $9;
   if ($or$cond) {
    _free($3);
    $$pre = HEAP32[$5>>2]|0;
    $12 = $$pre;
    break;
   }
   $15 = ((($os)) + 336|0);
   HEAP32[$15>>2] = $serialno;
   $$0 = 0;
   return ($$0|0);
  }
 } while(0);
 $11 = ($12|0)==(0|0);
 if (!($11)) {
  _free($12);
 }
 $13 = HEAP32[$7>>2]|0;
 $14 = ($13|0)==(0|0);
 if (!($14)) {
  _free($13);
 }
 _memset(($os|0),0,360)|0;
 $$0 = -1;
 return ($$0|0);
}
function _ogg_stream_clear($os) {
 $os = $os|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($os|0)==(0|0);
 if ($0) {
  return 0;
 }
 $1 = HEAP32[$os>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  _free($1);
 }
 $3 = ((($os)) + 16|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if (!($5)) {
  _free($4);
 }
 $6 = ((($os)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 if (!($8)) {
  _free($7);
 }
 _memset(($os|0),0,360)|0;
 return 0;
}
function _ogg_stream_check($os) {
 $os = $os|0;
 var $$ = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($os|0)==(0|0);
 if ($0) {
  return -1;
 } else {
  $1 = HEAP32[$os>>2]|0;
  $2 = ($1|0)==(0|0);
  $$ = $2 << 31 >> 31;
  return ($$|0);
 }
 return (0)|0;
}
function __os_lacing_expand($os,$needed) {
 $os = $os|0;
 $needed = $needed|0;
 var $$ = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($os)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) - ($needed))|0;
 $3 = ((($os)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2|0)>($4|0);
 if ($5) {
  $$1 = 0;
  return ($$1|0);
 }
 $6 = (2147483647 - ($needed))|0;
 $7 = ($1|0)>($6|0);
 if ($7) {
  $8 = HEAP32[$os>>2]|0;
  $9 = ($8|0)==(0|0);
  if (!($9)) {
   _free($8);
  }
  $10 = ((($os)) + 16|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($11|0)==(0|0);
  if (!($12)) {
   _free($11);
  }
  $13 = ((($os)) + 20|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ($14|0)==(0|0);
  if (!($15)) {
   _free($14);
  }
  _memset(($os|0),0,360)|0;
  $$1 = -1;
  return ($$1|0);
 }
 $16 = (($1) + ($needed))|0;
 $17 = ($16|0)<(2147483615);
 $18 = (($16) + 32)|0;
 $$ = $17 ? $18 : $16;
 $19 = ((($os)) + 16|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = $$ << 2;
 $22 = (_realloc($20,$21)|0);
 $23 = ($22|0)==(0|0);
 if ($23) {
  $24 = HEAP32[$os>>2]|0;
  $25 = ($24|0)==(0|0);
  if (!($25)) {
   _free($24);
  }
  $26 = HEAP32[$19>>2]|0;
  $27 = ($26|0)==(0|0);
  if (!($27)) {
   _free($26);
  }
  $28 = ((($os)) + 20|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = ($29|0)==(0|0);
  if (!($30)) {
   _free($29);
  }
  _memset(($os|0),0,360)|0;
  $$1 = -1;
  return ($$1|0);
 }
 HEAP32[$19>>2] = $22;
 $31 = ((($os)) + 20|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = $$ << 3;
 $34 = (_realloc($32,$33)|0);
 $35 = ($34|0)==(0|0);
 if (!($35)) {
  HEAP32[$31>>2] = $34;
  HEAP32[$0>>2] = $$;
  $$1 = 0;
  return ($$1|0);
 }
 $36 = HEAP32[$os>>2]|0;
 $37 = ($36|0)==(0|0);
 if (!($37)) {
  _free($36);
 }
 $38 = HEAP32[$19>>2]|0;
 $39 = ($38|0)==(0|0);
 if (!($39)) {
  _free($38);
 }
 $40 = HEAP32[$31>>2]|0;
 $41 = ($40|0)==(0|0);
 if (!($41)) {
  _free($40);
 }
 _memset(($os|0),0,360)|0;
 $$1 = -1;
 return ($$1|0);
}
function _ogg_sync_init($oy) {
 $oy = $oy|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($oy|0)==(0|0);
 if ($0) {
  return 0;
 }
 ;HEAP32[$oy>>2]=0|0;HEAP32[$oy+4>>2]=0|0;HEAP32[$oy+8>>2]=0|0;HEAP32[$oy+12>>2]=0|0;HEAP32[$oy+16>>2]=0|0;HEAP32[$oy+20>>2]=0|0;HEAP32[$oy+24>>2]=0|0;
 return 0;
}
function _ogg_sync_clear($oy) {
 $oy = $oy|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($oy|0)==(0|0);
 if ($0) {
  return 0;
 }
 $1 = HEAP32[$oy>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  _free($1);
 }
 ;HEAP32[$oy>>2]=0|0;HEAP32[$oy+4>>2]=0|0;HEAP32[$oy+8>>2]=0|0;HEAP32[$oy+12>>2]=0|0;HEAP32[$oy+16>>2]=0|0;HEAP32[$oy+20>>2]=0|0;HEAP32[$oy+24>>2]=0|0;
 return 0;
}
function _ogg_sync_buffer($oy,$size) {
 $oy = $oy|0;
 $size = $size|0;
 var $$1 = 0, $$pre = 0, $$pre$pre = 0, $$pre2 = 0, $$pre3 = 0, $$pre6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ret$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($oy)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(-1);
 if (!($2)) {
  $$1 = 0;
  return ($$1|0);
 }
 $3 = ((($oy)) + 12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0);
 $$pre6 = ((($oy)) + 8|0);
 if ($5) {
  $13 = $1;
 } else {
  $6 = HEAP32[$$pre6>>2]|0;
  $7 = (($6) - ($4))|0;
  HEAP32[$$pre6>>2] = $7;
  $8 = ($7|0)>(0);
  if ($8) {
   $9 = HEAP32[$oy>>2]|0;
   $10 = (($9) + ($4)|0);
   _memmove(($9|0),($10|0),($7|0))|0;
   $$pre$pre = HEAP32[$0>>2]|0;
   $$pre = $$pre$pre;
  } else {
   $$pre = $1;
  }
  HEAP32[$3>>2] = 0;
  $13 = $$pre;
 }
 $11 = HEAP32[$$pre6>>2]|0;
 $12 = (($13) - ($11))|0;
 $14 = ($12|0)<($size|0);
 do {
  if ($14) {
   $15 = (($size) + 4096)|0;
   $16 = (($15) + ($11))|0;
   $17 = HEAP32[$oy>>2]|0;
   $18 = ($17|0)==(0|0);
   if ($18) {
    $20 = (_malloc($16)|0);
    $ret$0 = $20;
   } else {
    $19 = (_realloc($17,$16)|0);
    $ret$0 = $19;
   }
   $21 = ($ret$0|0)==(0|0);
   if (!($21)) {
    HEAP32[$oy>>2] = $ret$0;
    HEAP32[$0>>2] = $16;
    $$pre3 = HEAP32[$$pre6>>2]|0;
    $25 = $ret$0;$26 = $$pre3;
    break;
   }
   $22 = HEAP32[$oy>>2]|0;
   $23 = ($22|0)==(0|0);
   if (!($23)) {
    _free($22);
   }
   ;HEAP32[$oy>>2]=0|0;HEAP32[$oy+4>>2]=0|0;HEAP32[$oy+8>>2]=0|0;HEAP32[$oy+12>>2]=0|0;HEAP32[$oy+16>>2]=0|0;HEAP32[$oy+20>>2]=0|0;HEAP32[$oy+24>>2]=0|0;
   $$1 = 0;
   return ($$1|0);
  } else {
   $$pre2 = HEAP32[$oy>>2]|0;
   $25 = $$pre2;$26 = $11;
  }
 } while(0);
 $24 = (($25) + ($26)|0);
 $$1 = $24;
 return ($$1|0);
}
function _ogg_sync_wrote($oy,$bytes) {
 $oy = $oy|0;
 $bytes = $bytes|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($oy)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(-1);
 if (!($2)) {
  $$0 = -1;
  return ($$0|0);
 }
 $3 = ((($oy)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + ($bytes))|0;
 $6 = ($5|0)>($1|0);
 if ($6) {
  $$0 = -1;
  return ($$0|0);
 }
 HEAP32[$3>>2] = $5;
 $$0 = 0;
 return ($$0|0);
}
function _ogg_sync_pageseek($oy,$og) {
 $oy = $oy|0;
 $og = $og|0;
 var $$2 = 0, $$pre = 0, $$pre$phi17Z2D = 0, $$pre10 = 0, $$pre12 = 0, $$pre13 = 0, $$pre16 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $chksum = 0, $crc_reg$0$lcssa$i = 0, $crc_reg$03$i = 0, $crc_reg$1$lcssa$i = 0, $crc_reg$11$i = 0, $exitcond = 0, $exitcond9 = 0, $i$04$i = 0;
 var $i$07 = 0, $i$12$i = 0, $next$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $chksum = sp;
 $0 = HEAP32[$oy>>2]|0;
 $1 = ((($oy)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($0) + ($2)|0);
 $4 = ((($oy)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($5) - ($2))|0;
 $7 = ((($oy)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)>(-1);
 if (!($9)) {
  $$2 = 0;
  STACKTOP = sp;return ($$2|0);
 }
 $10 = ((($oy)) + 20|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0);
 do {
  if ($12) {
   $13 = ($6|0)<(27);
   if ($13) {
    $$2 = 0;
    STACKTOP = sp;return ($$2|0);
   }
   $14 = (_memcmp($3,595837,4)|0);
   $15 = ($14|0)==(0);
   if (!($15)) {
    $$pre16 = ((($oy)) + 24|0);
    $$pre$phi17Z2D = $$pre16;
    break;
   }
   $16 = ((($3)) + 26|0);
   $17 = HEAP8[$16>>0]|0;
   $18 = $17&255;
   $19 = (($18) + 27)|0;
   $20 = ($6|0)<($19|0);
   if ($20) {
    $$2 = 0;
    STACKTOP = sp;return ($$2|0);
   }
   $21 = ($17<<24>>24)==(0);
   if (!($21)) {
    $22 = ((($oy)) + 24|0);
    $$pre = HEAP32[$22>>2]|0;
    $28 = $$pre;$i$07 = 0;
    while(1) {
     $23 = (($i$07) + 27)|0;
     $24 = (($3) + ($23)|0);
     $25 = HEAP8[$24>>0]|0;
     $26 = $25&255;
     $27 = (($28) + ($26))|0;
     HEAP32[$22>>2] = $27;
     $29 = (($i$07) + 1)|0;
     $30 = HEAP8[$16>>0]|0;
     $31 = $30&255;
     $32 = ($29|0)<($31|0);
     if ($32) {
      $28 = $27;$i$07 = $29;
     } else {
      break;
     }
    }
   }
   HEAP32[$10>>2] = $19;
   $36 = $19;
   label = 11;
  } else {
   $36 = $11;
   label = 11;
  }
 } while(0);
 do {
  if ((label|0) == 11) {
   $33 = ((($oy)) + 24|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = (($36) + ($34))|0;
   $37 = ($35|0)>($6|0);
   if ($37) {
    $$2 = 0;
    STACKTOP = sp;return ($$2|0);
   }
   $38 = ((($3)) + 22|0);
   $39 = HEAPU8[$38>>0]|(HEAPU8[$38+1>>0]<<8)|(HEAPU8[$38+2>>0]<<16)|(HEAPU8[$38+3>>0]<<24);
   HEAP32[$chksum>>2] = $39;
   HEAP8[$38>>0]=0&255;HEAP8[$38+1>>0]=(0>>8)&255;HEAP8[$38+2>>0]=(0>>16)&255;HEAP8[$38+3>>0]=0>>24;
   $40 = HEAP32[$10>>2]|0;
   $41 = (($3) + ($40)|0);
   $42 = HEAP32[$33>>2]|0;
   $43 = ((($3)) + 23|0);
   $44 = ((($3)) + 24|0);
   $45 = ((($3)) + 25|0);
   $46 = ($40|0)>(0);
   HEAP8[$38>>0]=0&255;HEAP8[$38+1>>0]=(0>>8)&255;HEAP8[$38+2>>0]=(0>>16)&255;HEAP8[$38+3>>0]=0>>24;
   if ($46) {
    $crc_reg$03$i = 0;$i$04$i = 0;
    while(1) {
     $48 = $crc_reg$03$i << 8;
     $49 = $crc_reg$03$i >>> 24;
     $50 = (($3) + ($i$04$i)|0);
     $51 = HEAP8[$50>>0]|0;
     $52 = $51&255;
     $53 = $52 ^ $49;
     $54 = (124732 + ($53<<2)|0);
     $55 = HEAP32[$54>>2]|0;
     $56 = $55 ^ $48;
     $57 = (($i$04$i) + 1)|0;
     $exitcond9 = ($57|0)==($40|0);
     if ($exitcond9) {
      $crc_reg$0$lcssa$i = $56;
      break;
     } else {
      $crc_reg$03$i = $56;$i$04$i = $57;
     }
    }
   } else {
    $crc_reg$0$lcssa$i = 0;
   }
   $47 = ($42|0)>(0);
   if ($47) {
    $crc_reg$11$i = $crc_reg$0$lcssa$i;$i$12$i = 0;
    while(1) {
     $58 = $crc_reg$11$i << 8;
     $59 = $crc_reg$11$i >>> 24;
     $60 = (($41) + ($i$12$i)|0);
     $61 = HEAP8[$60>>0]|0;
     $62 = $61&255;
     $63 = $62 ^ $59;
     $64 = (124732 + ($63<<2)|0);
     $65 = HEAP32[$64>>2]|0;
     $66 = $65 ^ $58;
     $67 = (($i$12$i) + 1)|0;
     $exitcond = ($67|0)==($42|0);
     if ($exitcond) {
      $crc_reg$1$lcssa$i = $66;
      break;
     } else {
      $crc_reg$11$i = $66;$i$12$i = $67;
     }
    }
   } else {
    $crc_reg$1$lcssa$i = $crc_reg$0$lcssa$i;
   }
   $68 = $crc_reg$1$lcssa$i&255;
   HEAP8[$38>>0] = $68;
   $69 = $crc_reg$1$lcssa$i >>> 8;
   $70 = $69&255;
   HEAP8[$43>>0] = $70;
   $71 = $crc_reg$1$lcssa$i >>> 16;
   $72 = $71&255;
   HEAP8[$44>>0] = $72;
   $73 = $crc_reg$1$lcssa$i >>> 24;
   $74 = $73&255;
   HEAP8[$45>>0] = $74;
   $75 = (_memcmp($chksum,$38,4)|0);
   $76 = ($75|0)==(0);
   if (!($76)) {
    HEAP8[$38>>0]=$39&255;HEAP8[$38+1>>0]=($39>>8)&255;HEAP8[$38+2>>0]=($39>>16)&255;HEAP8[$38+3>>0]=$39>>24;
    $$pre$phi17Z2D = $33;
    break;
   }
   $77 = HEAP32[$oy>>2]|0;
   $78 = HEAP32[$1>>2]|0;
   $79 = (($77) + ($78)|0);
   $80 = ($og|0)==(0|0);
   if ($80) {
    $$pre12 = HEAP32[$10>>2]|0;
    $$pre13 = HEAP32[$33>>2]|0;
    $89 = $$pre13;$90 = $$pre12;
   } else {
    HEAP32[$og>>2] = $79;
    $81 = HEAP32[$10>>2]|0;
    $82 = ((($og)) + 4|0);
    HEAP32[$82>>2] = $81;
    $83 = (($79) + ($81)|0);
    $84 = ((($og)) + 8|0);
    HEAP32[$84>>2] = $83;
    $85 = HEAP32[$33>>2]|0;
    $86 = ((($og)) + 12|0);
    HEAP32[$86>>2] = $85;
    $89 = $85;$90 = $81;
   }
   $87 = ((($oy)) + 16|0);
   HEAP32[$87>>2] = 0;
   $88 = (($89) + ($90))|0;
   $91 = (($78) + ($88))|0;
   HEAP32[$1>>2] = $91;
   HEAP32[$10>>2] = 0;
   HEAP32[$33>>2] = 0;
   $$2 = $88;
   STACKTOP = sp;return ($$2|0);
  }
 } while(0);
 HEAP32[$10>>2] = 0;
 HEAP32[$$pre$phi17Z2D>>2] = 0;
 $92 = ((($3)) + 1|0);
 $93 = (($6) + -1)|0;
 $94 = (_memchr($92,79,$93)|0);
 $95 = ($94|0)==(0|0);
 if ($95) {
  $96 = HEAP32[$oy>>2]|0;
  $97 = HEAP32[$4>>2]|0;
  $98 = (($96) + ($97)|0);
  $99 = $96;
  $102 = $99;$next$0 = $98;
 } else {
  $$pre10 = HEAP32[$oy>>2]|0;
  $102 = $$pre10;$next$0 = $94;
 }
 $100 = $next$0;
 $101 = (($100) - ($102))|0;
 HEAP32[$1>>2] = $101;
 $103 = $3;
 $104 = (($103) - ($100))|0;
 $$2 = $104;
 STACKTOP = sp;return ($$2|0);
}
function _ogg_sync_pageout($oy,$og) {
 $oy = $oy|0;
 $og = $og|0;
 var $$2 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($oy)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(-1);
 L1: do {
  if ($2) {
   $3 = ((($oy)) + 16|0);
   while(1) {
    $4 = (_ogg_sync_pageseek($oy,$og)|0);
    $5 = ($4|0)>(0);
    if ($5) {
     $$2 = 1;
     break L1;
    }
    $6 = ($4|0)==(0);
    if ($6) {
     $$2 = 0;
     break L1;
    }
    $7 = HEAP32[$3>>2]|0;
    $8 = ($7|0)==(0);
    if ($8) {
     break;
    }
   }
   HEAP32[$3>>2] = 1;
   $$2 = -1;
  } else {
   $$2 = 0;
  }
 } while(0);
 return ($$2|0);
}
function _ogg_stream_pagein($os,$og) {
 $os = $os|0;
 $og = $og|0;
 var $$$i = 0, $$0 = 0, $$lcssa = 0, $$lcssa73 = 0, $$lcssa75 = 0, $$lcssa80 = 0, $$pre = 0, $$pre40 = 0, $$pre41 = 0, $$pre42 = 0, $$pre42$pre = 0, $$pre43$pre = 0, $$pre44$pre = 0, $$promoted = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0;
 var $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0;
 var $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0;
 var $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0;
 var $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $body$018 = 0, $body$1 = 0, $bodysize$017 = 0, $bodysize$1 = 0, $bos$0 = 0;
 var $bos$1$ph12 = 0, $bos$19 = 0, $i$027 = 0, $or$cond = 0, $saved$0$ph$lcssa = 0, $saved$0$ph14 = 0, $segptr$016 = 0, $segptr$1 = 0, $segptr$2 = 0, $segptr$3$ph13 = 0, $segptr$310 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$og>>2]|0;
 $1 = ((($og)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($og)) + 12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($0)) + 4|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = ((($0)) + 5|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = $8&255;
 $10 = $9 & 1;
 $11 = $9 & 2;
 $12 = $9 & 4;
 $13 = ((($0)) + 13|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = $14&255;
 $16 = (_bitshift64Shl(($15|0),0,8)|0);
 $17 = tempRet0;
 $18 = ((($0)) + 12|0);
 $19 = HEAP8[$18>>0]|0;
 $20 = $19&255;
 $21 = $16 | $20;
 $22 = (_bitshift64Shl(($21|0),($17|0),8)|0);
 $23 = tempRet0;
 $24 = ((($0)) + 11|0);
 $25 = HEAP8[$24>>0]|0;
 $26 = $25&255;
 $27 = $22 | $26;
 $28 = (_bitshift64Shl(($27|0),($23|0),8)|0);
 $29 = tempRet0;
 $30 = ((($0)) + 10|0);
 $31 = HEAP8[$30>>0]|0;
 $32 = $31&255;
 $33 = $28 | $32;
 $34 = (_bitshift64Shl(($33|0),($29|0),8)|0);
 $35 = tempRet0;
 $36 = ((($0)) + 9|0);
 $37 = HEAP8[$36>>0]|0;
 $38 = $37&255;
 $39 = $34 | $38;
 $40 = (_bitshift64Shl(($39|0),($35|0),8)|0);
 $41 = tempRet0;
 $42 = ((($0)) + 8|0);
 $43 = HEAP8[$42>>0]|0;
 $44 = $43&255;
 $45 = $40 | $44;
 $46 = (_bitshift64Shl(($45|0),($41|0),8)|0);
 $47 = tempRet0;
 $48 = ((($0)) + 7|0);
 $49 = HEAP8[$48>>0]|0;
 $50 = $49&255;
 $51 = $46 | $50;
 $52 = (_bitshift64Shl(($51|0),($47|0),8)|0);
 $53 = tempRet0;
 $54 = ((($0)) + 6|0);
 $55 = HEAP8[$54>>0]|0;
 $56 = $55&255;
 $57 = $52 | $56;
 $58 = ((($0)) + 14|0);
 $59 = HEAP8[$58>>0]|0;
 $60 = $59&255;
 $61 = ((($0)) + 15|0);
 $62 = HEAP8[$61>>0]|0;
 $63 = $62&255;
 $64 = $63 << 8;
 $65 = $64 | $60;
 $66 = ((($0)) + 16|0);
 $67 = HEAP8[$66>>0]|0;
 $68 = $67&255;
 $69 = $68 << 16;
 $70 = $65 | $69;
 $71 = ((($0)) + 17|0);
 $72 = HEAP8[$71>>0]|0;
 $73 = $72&255;
 $74 = $73 << 24;
 $75 = $70 | $74;
 $76 = ((($0)) + 18|0);
 $77 = HEAP8[$76>>0]|0;
 $78 = $77&255;
 $79 = ((($0)) + 19|0);
 $80 = HEAP8[$79>>0]|0;
 $81 = $80&255;
 $82 = $81 << 8;
 $83 = $82 | $78;
 $84 = ((($0)) + 20|0);
 $85 = HEAP8[$84>>0]|0;
 $86 = $85&255;
 $87 = $86 << 16;
 $88 = $83 | $87;
 $89 = ((($0)) + 21|0);
 $90 = HEAP8[$89>>0]|0;
 $91 = $90&255;
 $92 = $91 << 24;
 $93 = $88 | $92;
 $94 = ((($0)) + 26|0);
 $95 = HEAP8[$94>>0]|0;
 $96 = $95&255;
 $97 = ($os|0)==(0|0);
 if ($97) {
  $$0 = -1;
  return ($$0|0);
 }
 $98 = HEAP32[$os>>2]|0;
 $99 = ($98|0)==(0|0);
 if ($99) {
  $$0 = -1;
  return ($$0|0);
 }
 $100 = ((($os)) + 36|0);
 $101 = HEAP32[$100>>2]|0;
 $102 = ((($os)) + 12|0);
 $103 = HEAP32[$102>>2]|0;
 $104 = ($103|0)==(0);
 if (!($104)) {
  $105 = ((($os)) + 8|0);
  $106 = HEAP32[$105>>2]|0;
  $107 = (($106) - ($103))|0;
  HEAP32[$105>>2] = $107;
  $108 = ($106|0)==($103|0);
  if (!($108)) {
   $109 = (($98) + ($103)|0);
   _memmove(($98|0),($109|0),($107|0))|0;
  }
  HEAP32[$102>>2] = 0;
 }
 $110 = ($101|0)==(0);
 if (!($110)) {
  $111 = ((($os)) + 28|0);
  $112 = HEAP32[$111>>2]|0;
  $113 = ($112|0)==($101|0);
  if ($113) {
   $126 = $101;
  } else {
   $114 = (($112) - ($101))|0;
   $115 = ((($os)) + 16|0);
   $116 = HEAP32[$115>>2]|0;
   $117 = (($116) + ($101<<2)|0);
   $118 = $114 << 2;
   _memmove(($116|0),($117|0),($118|0))|0;
   $119 = ((($os)) + 20|0);
   $120 = HEAP32[$119>>2]|0;
   $121 = (($120) + ($101<<3)|0);
   $122 = HEAP32[$111>>2]|0;
   $123 = (($122) - ($101))|0;
   $124 = $123 << 3;
   _memmove(($120|0),($121|0),($124|0))|0;
   $$pre = HEAP32[$111>>2]|0;
   $126 = $$pre;
  }
  $125 = (($126) - ($101))|0;
  HEAP32[$111>>2] = $125;
  $127 = ((($os)) + 32|0);
  $128 = HEAP32[$127>>2]|0;
  $129 = (($128) - ($101))|0;
  HEAP32[$127>>2] = $129;
  HEAP32[$100>>2] = 0;
 }
 $130 = ((($os)) + 336|0);
 $131 = HEAP32[$130>>2]|0;
 $132 = ($75|0)!=($131|0);
 $133 = ($6<<24>>24)!=(0);
 $or$cond = $133 | $132;
 if ($or$cond) {
  $$0 = -1;
  return ($$0|0);
 }
 $134 = (($96) + 1)|0;
 $135 = (__os_lacing_expand($os,$134)|0);
 $136 = ($135|0)==(0);
 if (!($136)) {
  $$0 = -1;
  return ($$0|0);
 }
 $137 = ((($os)) + 340|0);
 $138 = HEAP32[$137>>2]|0;
 $139 = ($93|0)==($138|0);
 if (!($139)) {
  $140 = ((($os)) + 32|0);
  $141 = HEAP32[$140>>2]|0;
  $142 = ((($os)) + 28|0);
  $143 = HEAP32[$142>>2]|0;
  $144 = ($141|0)<($143|0);
  if ($144) {
   $145 = ((($os)) + 16|0);
   $146 = HEAP32[$145>>2]|0;
   $147 = ((($os)) + 8|0);
   $$promoted = HEAP32[$147>>2]|0;
   $152 = $$promoted;$i$027 = $141;
   while(1) {
    $148 = (($146) + ($i$027<<2)|0);
    $149 = HEAP32[$148>>2]|0;
    $150 = $149 & 255;
    $151 = (($152) - ($150))|0;
    $153 = (($i$027) + 1)|0;
    $154 = ($153|0)<($143|0);
    if ($154) {
     $152 = $151;$i$027 = $153;
    } else {
     $$lcssa80 = $151;
     break;
    }
   }
   HEAP32[$147>>2] = $$lcssa80;
  }
  HEAP32[$142>>2] = $141;
  $155 = ($138|0)==(-1);
  if (!($155)) {
   $156 = (($141) + 1)|0;
   HEAP32[$142>>2] = $156;
   $157 = ((($os)) + 16|0);
   $158 = HEAP32[$157>>2]|0;
   $159 = (($158) + ($141<<2)|0);
   HEAP32[$159>>2] = 1024;
   HEAP32[$140>>2] = $156;
  }
 }
 $160 = ($10|0)==(0);
 L35: do {
  if ($160) {
   $body$1 = $2;$bodysize$1 = $4;$bos$0 = $11;$segptr$2 = 0;
  } else {
   $161 = ((($os)) + 28|0);
   $162 = HEAP32[$161>>2]|0;
   $163 = ($162|0)<(1);
   if (!($163)) {
    $164 = (($162) + -1)|0;
    $165 = ((($os)) + 16|0);
    $166 = HEAP32[$165>>2]|0;
    $167 = (($166) + ($164<<2)|0);
    $168 = HEAP32[$167>>2]|0;
    $169 = ($168|0)==(1024);
    if (!($169)) {
     $body$1 = $2;$bodysize$1 = $4;$bos$0 = $11;$segptr$2 = 0;
     break;
    }
   }
   $170 = ($95<<24>>24)==(0);
   if ($170) {
    $body$1 = $2;$bodysize$1 = $4;$bos$0 = 0;$segptr$2 = 0;
   } else {
    $body$018 = $2;$bodysize$017 = $4;$segptr$016 = 0;
    while(1) {
     $171 = (($segptr$016) + 27)|0;
     $172 = (($0) + ($171)|0);
     $173 = HEAP8[$172>>0]|0;
     $174 = $173&255;
     $175 = (($body$018) + ($174)|0);
     $176 = (($bodysize$017) - ($174))|0;
     $177 = ($173<<24>>24)==(-1);
     $178 = $177&1;
     $179 = $178 ^ 1;
     $segptr$1 = (($179) + ($segptr$016))|0;
     if (!($177)) {
      $body$1 = $175;$bodysize$1 = $176;$bos$0 = 0;$segptr$2 = $segptr$1;
      break L35;
     }
     $180 = (($segptr$1) + 1)|0;
     $181 = ($180|0)<($96|0);
     if ($181) {
      $body$018 = $175;$bodysize$017 = $176;$segptr$016 = $180;
     } else {
      $body$1 = $175;$bodysize$1 = $176;$bos$0 = 0;$segptr$2 = $180;
      break;
     }
    }
   }
  }
 } while(0);
 $182 = ($bodysize$1|0)==(0);
 if (!($182)) {
  $183 = ((($os)) + 4|0);
  $184 = HEAP32[$183>>2]|0;
  $185 = (($184) - ($bodysize$1))|0;
  $186 = ((($os)) + 8|0);
  $187 = HEAP32[$186>>2]|0;
  $188 = ($185|0)>($187|0);
  do {
   if ($188) {
    $$pre40 = HEAP32[$os>>2]|0;
    $214 = $$pre40;$215 = $187;
   } else {
    $189 = (2147483647 - ($bodysize$1))|0;
    $190 = ($184|0)>($189|0);
    if ($190) {
     $191 = HEAP32[$os>>2]|0;
     $192 = ($191|0)==(0|0);
     if (!($192)) {
      _free($191);
     }
     $193 = ((($os)) + 16|0);
     $194 = HEAP32[$193>>2]|0;
     $195 = ($194|0)==(0|0);
     if (!($195)) {
      _free($194);
     }
     $196 = ((($os)) + 20|0);
     $197 = HEAP32[$196>>2]|0;
     $198 = ($197|0)==(0|0);
     if (!($198)) {
      _free($197);
     }
     _memset(($os|0),0,360)|0;
     $$0 = -1;
     return ($$0|0);
    }
    $199 = (($184) + ($bodysize$1))|0;
    $200 = ($199|0)<(2147482623);
    $201 = (($199) + 1024)|0;
    $$$i = $200 ? $201 : $199;
    $202 = HEAP32[$os>>2]|0;
    $203 = (_realloc($202,$$$i)|0);
    $204 = ($203|0)==(0|0);
    if (!($204)) {
     HEAP32[$183>>2] = $$$i;
     HEAP32[$os>>2] = $203;
     $$pre41 = HEAP32[$186>>2]|0;
     $214 = $203;$215 = $$pre41;
     break;
    }
    $205 = HEAP32[$os>>2]|0;
    $206 = ($205|0)==(0|0);
    if (!($206)) {
     _free($205);
    }
    $207 = ((($os)) + 16|0);
    $208 = HEAP32[$207>>2]|0;
    $209 = ($208|0)==(0|0);
    if (!($209)) {
     _free($208);
    }
    $210 = ((($os)) + 20|0);
    $211 = HEAP32[$210>>2]|0;
    $212 = ($211|0)==(0|0);
    if (!($212)) {
     _free($211);
    }
    _memset(($os|0),0,360)|0;
    $$0 = -1;
    return ($$0|0);
   }
  } while(0);
  $213 = (($214) + ($215)|0);
  _memcpy(($213|0),($body$1|0),($bodysize$1|0))|0;
  $216 = HEAP32[$186>>2]|0;
  $217 = (($216) + ($bodysize$1))|0;
  HEAP32[$186>>2] = $217;
 }
 $218 = ($segptr$2|0)<($96|0);
 if ($218) {
  $219 = ((($os)) + 28|0);
  $220 = ((($os)) + 16|0);
  $221 = ((($os)) + 20|0);
  $222 = ((($os)) + 32|0);
  $$pre42$pre = HEAP32[$219>>2]|0;
  $$pre43$pre = HEAP32[$220>>2]|0;
  $$pre44$pre = HEAP32[$221>>2]|0;
  $$pre42 = $$pre42$pre;$bos$1$ph12 = $bos$0;$saved$0$ph14 = -1;$segptr$3$ph13 = $segptr$2;
  L79: while(1) {
   $228 = $$pre42;$bos$19 = $bos$1$ph12;$segptr$310 = $segptr$3$ph13;
   while(1) {
    $223 = (($segptr$310) + 27)|0;
    $224 = (($0) + ($223)|0);
    $225 = HEAP8[$224>>0]|0;
    $226 = $225&255;
    $227 = (($$pre43$pre) + ($228<<2)|0);
    HEAP32[$227>>2] = $226;
    $229 = (($$pre44$pre) + ($228<<3)|0);
    $230 = $229;
    $231 = $230;
    HEAP32[$231>>2] = -1;
    $232 = (($230) + 4)|0;
    $233 = $232;
    HEAP32[$233>>2] = -1;
    $234 = ($bos$19|0)==(0);
    if (!($234)) {
     $235 = $226 | 256;
     HEAP32[$227>>2] = $235;
    }
    $236 = ($225<<24>>24)==(-1);
    $237 = (($228) + 1)|0;
    HEAP32[$219>>2] = $237;
    $238 = (($segptr$310) + 1)|0;
    if (!($236)) {
     $$lcssa = $228;$$lcssa73 = $237;$$lcssa75 = $238;
     break;
    }
    $239 = ($238|0)<($96|0);
    if ($239) {
     $228 = $237;$bos$19 = 0;$segptr$310 = $238;
    } else {
     $saved$0$ph$lcssa = $saved$0$ph14;
     break L79;
    }
   }
   HEAP32[$222>>2] = $$lcssa73;
   $240 = ($$lcssa75|0)<($96|0);
   if ($240) {
    $$pre42 = $$lcssa73;$bos$1$ph12 = 0;$saved$0$ph14 = $$lcssa;$segptr$3$ph13 = $$lcssa75;
   } else {
    $saved$0$ph$lcssa = $$lcssa;
    break;
   }
  }
  $241 = ($saved$0$ph$lcssa|0)==(-1);
  if (!($241)) {
   $242 = ((($os)) + 20|0);
   $243 = HEAP32[$242>>2]|0;
   $244 = (($243) + ($saved$0$ph$lcssa<<3)|0);
   $245 = $244;
   $246 = $245;
   HEAP32[$246>>2] = $57;
   $247 = (($245) + 4)|0;
   $248 = $247;
   HEAP32[$248>>2] = $53;
  }
 }
 $249 = ($12|0)==(0);
 if (!($249)) {
  $250 = ((($os)) + 328|0);
  HEAP32[$250>>2] = 1;
  $251 = ((($os)) + 28|0);
  $252 = HEAP32[$251>>2]|0;
  $253 = ($252|0)>(0);
  if ($253) {
   $254 = (($252) + -1)|0;
   $255 = ((($os)) + 16|0);
   $256 = HEAP32[$255>>2]|0;
   $257 = (($256) + ($254<<2)|0);
   $258 = HEAP32[$257>>2]|0;
   $259 = $258 | 512;
   HEAP32[$257>>2] = $259;
  }
 }
 $260 = (($93) + 1)|0;
 HEAP32[$137>>2] = $260;
 $$0 = 0;
 return ($$0|0);
}
function _ogg_sync_reset($oy) {
 $oy = $oy|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($oy)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(-1);
 if (!($2)) {
  $$0 = -1;
  return ($$0|0);
 }
 $3 = ((($oy)) + 8|0);
 ;HEAP32[$3>>2]=0|0;HEAP32[$3+4>>2]=0|0;HEAP32[$3+8>>2]=0|0;HEAP32[$3+12>>2]=0|0;HEAP32[$3+16>>2]=0|0;
 $$0 = 0;
 return ($$0|0);
}
function _ogg_stream_reset($os) {
 $os = $os|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($os|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = HEAP32[$os>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = -1;
  return ($$0|0);
 }
 $3 = ((($os)) + 8|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($os)) + 12|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($os)) + 28|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($os)) + 32|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($os)) + 36|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($os)) + 324|0);
 HEAP32[$8>>2] = 0;
 $9 = ((($os)) + 328|0);
 HEAP32[$9>>2] = 0;
 $10 = ((($os)) + 332|0);
 HEAP32[$10>>2] = 0;
 $11 = ((($os)) + 340|0);
 HEAP32[$11>>2] = -1;
 $12 = ((($os)) + 344|0);
 ;HEAP32[$12>>2]=0|0;HEAP32[$12+4>>2]=0|0;HEAP32[$12+8>>2]=0|0;HEAP32[$12+12>>2]=0|0;
 $$0 = 0;
 return ($$0|0);
}
function _ogg_stream_reset_serialno($os,$serialno) {
 $os = $os|0;
 $serialno = $serialno|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($os|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = HEAP32[$os>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = -1;
  return ($$0|0);
 }
 $3 = ((($os)) + 8|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($os)) + 12|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($os)) + 28|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($os)) + 32|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($os)) + 36|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($os)) + 324|0);
 HEAP32[$8>>2] = 0;
 $9 = ((($os)) + 328|0);
 HEAP32[$9>>2] = 0;
 $10 = ((($os)) + 332|0);
 HEAP32[$10>>2] = 0;
 $11 = ((($os)) + 340|0);
 HEAP32[$11>>2] = -1;
 $12 = ((($os)) + 344|0);
 ;HEAP32[$12>>2]=0|0;HEAP32[$12+4>>2]=0|0;HEAP32[$12+8>>2]=0|0;HEAP32[$12+12>>2]=0|0;
 $13 = ((($os)) + 336|0);
 HEAP32[$13>>2] = $serialno;
 $$0 = 0;
 return ($$0|0);
}
function _ogg_stream_packetout($os,$op) {
 $os = $os|0;
 $op = $op|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$phi$trans$insert3 = 0, $$pre = 0, $$pre$phi5Z2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $bytes$0$lcssa$i = 0, $bytes$02$i = 0, $eos$0$$i = 0, $eos$0$lcssa$i = 0;
 var $eos$03$i = 0, $ptr$0$lcssa$i = 0, $ptr$04$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($os|0)==(0|0);
 if ($0) {
  $$0 = 0;
  return ($$0|0);
 }
 $1 = HEAP32[$os>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = 0;
  return ($$0|0);
 }
 $3 = ((($os)) + 36|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($os)) + 32|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)>($4|0);
 if (!($7)) {
  $$0 = 0;
  return ($$0|0);
 }
 $8 = ((($os)) + 16|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (($9) + ($4<<2)|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 & 1024;
 $13 = ($12|0)==(0);
 if (!($13)) {
  $14 = (($4) + 1)|0;
  HEAP32[$3>>2] = $14;
  $15 = ((($os)) + 344|0);
  $16 = $15;
  $17 = $16;
  $18 = HEAP32[$17>>2]|0;
  $19 = (($16) + 4)|0;
  $20 = $19;
  $21 = HEAP32[$20>>2]|0;
  $22 = (_i64Add(($18|0),($21|0),1,0)|0);
  $23 = tempRet0;
  $24 = $15;
  $25 = $24;
  HEAP32[$25>>2] = $22;
  $26 = (($24) + 4)|0;
  $27 = $26;
  HEAP32[$27>>2] = $23;
  $$0 = -1;
  return ($$0|0);
 }
 $28 = ($op|0)==(0|0);
 $29 = $11 & 255;
 $30 = $11 & 512;
 $31 = $11 & 256;
 $32 = ($29|0)==(255);
 if ($32) {
  $bytes$02$i = 255;$eos$03$i = $30;$ptr$04$i = $4;
  while(1) {
   $33 = (($ptr$04$i) + 1)|0;
   $34 = (($9) + ($33<<2)|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = $35 & 255;
   $37 = $35 & 512;
   $38 = ($37|0)==(0);
   $eos$0$$i = $38 ? $eos$03$i : 512;
   $39 = (($36) + ($bytes$02$i))|0;
   $40 = ($36|0)==(255);
   if ($40) {
    $bytes$02$i = $39;$eos$03$i = $eos$0$$i;$ptr$04$i = $33;
   } else {
    $bytes$0$lcssa$i = $39;$eos$0$lcssa$i = $eos$0$$i;$ptr$0$lcssa$i = $33;
    break;
   }
  }
 } else {
  $bytes$0$lcssa$i = $29;$eos$0$lcssa$i = $30;$ptr$0$lcssa$i = $4;
 }
 if ($28) {
  $$phi$trans$insert = ((($os)) + 12|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $$phi$trans$insert3 = ((($os)) + 344|0);
  $41 = $$phi$trans$insert3;
  $42 = $41;
  $43 = HEAP32[$42>>2]|0;
  $44 = (($41) + 4)|0;
  $45 = $44;
  $46 = HEAP32[$45>>2]|0;
  $$pre$phi5Z2D = $$phi$trans$insert3;$$pre$phiZ2D = $$phi$trans$insert;$80 = $$pre;$82 = $43;$83 = $46;
 } else {
  $47 = ((($op)) + 12|0);
  HEAP32[$47>>2] = $eos$0$lcssa$i;
  $48 = ((($op)) + 8|0);
  HEAP32[$48>>2] = $31;
  $49 = ((($os)) + 12|0);
  $50 = HEAP32[$49>>2]|0;
  $51 = (($1) + ($50)|0);
  HEAP32[$op>>2] = $51;
  $52 = ((($os)) + 344|0);
  $53 = $52;
  $54 = $53;
  $55 = HEAP32[$54>>2]|0;
  $56 = (($53) + 4)|0;
  $57 = $56;
  $58 = HEAP32[$57>>2]|0;
  $59 = ((($op)) + 24|0);
  $60 = $59;
  $61 = $60;
  HEAP32[$61>>2] = $55;
  $62 = (($60) + 4)|0;
  $63 = $62;
  HEAP32[$63>>2] = $58;
  $64 = ((($os)) + 20|0);
  $65 = HEAP32[$64>>2]|0;
  $66 = (($65) + ($ptr$0$lcssa$i<<3)|0);
  $67 = $66;
  $68 = $67;
  $69 = HEAP32[$68>>2]|0;
  $70 = (($67) + 4)|0;
  $71 = $70;
  $72 = HEAP32[$71>>2]|0;
  $73 = ((($op)) + 16|0);
  $74 = $73;
  $75 = $74;
  HEAP32[$75>>2] = $69;
  $76 = (($74) + 4)|0;
  $77 = $76;
  HEAP32[$77>>2] = $72;
  $78 = ((($op)) + 4|0);
  HEAP32[$78>>2] = $bytes$0$lcssa$i;
  $$pre$phi5Z2D = $52;$$pre$phiZ2D = $49;$80 = $50;$82 = $55;$83 = $58;
 }
 $79 = (($80) + ($bytes$0$lcssa$i))|0;
 HEAP32[$$pre$phiZ2D>>2] = $79;
 $81 = (($ptr$0$lcssa$i) + 1)|0;
 HEAP32[$3>>2] = $81;
 $84 = (_i64Add(($82|0),($83|0),1,0)|0);
 $85 = tempRet0;
 $86 = $$pre$phi5Z2D;
 $87 = $86;
 HEAP32[$87>>2] = $84;
 $88 = (($86) + 4)|0;
 $89 = $88;
 HEAP32[$89>>2] = $85;
 $$0 = 1;
 return ($$0|0);
}
function _ogg_stream_packetpeek($os,$op) {
 $os = $os|0;
 $op = $op|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0, $bytes$0$lcssa$i = 0, $bytes$02$i = 0, $eos$0$$i = 0, $eos$0$lcssa$i = 0, $eos$03$i = 0, $ptr$0$lcssa$i = 0;
 var $ptr$04$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($os|0)==(0|0);
 if ($0) {
  $$0 = 0;
  return ($$0|0);
 }
 $1 = HEAP32[$os>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = 0;
  return ($$0|0);
 }
 $3 = ((($os)) + 36|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($os)) + 32|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)>($4|0);
 if (!($7)) {
  $$0 = 0;
  return ($$0|0);
 }
 $8 = ((($os)) + 16|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (($9) + ($4<<2)|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 & 1024;
 $13 = ($12|0)==(0);
 if (!($13)) {
  $14 = (($4) + 1)|0;
  HEAP32[$3>>2] = $14;
  $15 = ((($os)) + 344|0);
  $16 = $15;
  $17 = $16;
  $18 = HEAP32[$17>>2]|0;
  $19 = (($16) + 4)|0;
  $20 = $19;
  $21 = HEAP32[$20>>2]|0;
  $22 = (_i64Add(($18|0),($21|0),1,0)|0);
  $23 = tempRet0;
  $24 = $15;
  $25 = $24;
  HEAP32[$25>>2] = $22;
  $26 = (($24) + 4)|0;
  $27 = $26;
  HEAP32[$27>>2] = $23;
  $$0 = -1;
  return ($$0|0);
 }
 $28 = ($op|0)==(0|0);
 if ($28) {
  $$0 = 1;
  return ($$0|0);
 }
 $29 = $11 & 255;
 $30 = $11 & 512;
 $31 = $11 & 256;
 $32 = ($29|0)==(255);
 if ($32) {
  $bytes$02$i = 255;$eos$03$i = $30;$ptr$04$i = $4;
  while(1) {
   $33 = (($ptr$04$i) + 1)|0;
   $34 = (($9) + ($33<<2)|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = $35 & 255;
   $37 = $35 & 512;
   $38 = ($37|0)==(0);
   $eos$0$$i = $38 ? $eos$03$i : 512;
   $39 = (($36) + ($bytes$02$i))|0;
   $40 = ($36|0)==(255);
   if ($40) {
    $bytes$02$i = $39;$eos$03$i = $eos$0$$i;$ptr$04$i = $33;
   } else {
    $bytes$0$lcssa$i = $39;$eos$0$lcssa$i = $eos$0$$i;$ptr$0$lcssa$i = $33;
    break;
   }
  }
 } else {
  $bytes$0$lcssa$i = $29;$eos$0$lcssa$i = $30;$ptr$0$lcssa$i = $4;
 }
 $41 = ((($op)) + 12|0);
 HEAP32[$41>>2] = $eos$0$lcssa$i;
 $42 = ((($op)) + 8|0);
 HEAP32[$42>>2] = $31;
 $43 = ((($os)) + 12|0);
 $44 = HEAP32[$43>>2]|0;
 $45 = (($1) + ($44)|0);
 HEAP32[$op>>2] = $45;
 $46 = ((($os)) + 344|0);
 $47 = $46;
 $48 = $47;
 $49 = HEAP32[$48>>2]|0;
 $50 = (($47) + 4)|0;
 $51 = $50;
 $52 = HEAP32[$51>>2]|0;
 $53 = ((($op)) + 24|0);
 $54 = $53;
 $55 = $54;
 HEAP32[$55>>2] = $49;
 $56 = (($54) + 4)|0;
 $57 = $56;
 HEAP32[$57>>2] = $52;
 $58 = ((($os)) + 20|0);
 $59 = HEAP32[$58>>2]|0;
 $60 = (($59) + ($ptr$0$lcssa$i<<3)|0);
 $61 = $60;
 $62 = $61;
 $63 = HEAP32[$62>>2]|0;
 $64 = (($61) + 4)|0;
 $65 = $64;
 $66 = HEAP32[$65>>2]|0;
 $67 = ((($op)) + 16|0);
 $68 = $67;
 $69 = $68;
 HEAP32[$69>>2] = $63;
 $70 = (($68) + 4)|0;
 $71 = $70;
 HEAP32[$71>>2] = $66;
 $72 = ((($op)) + 4|0);
 HEAP32[$72>>2] = $bytes$0$lcssa$i;
 $$0 = 1;
 return ($$0|0);
}
function _GLES2_CreateRenderer($window,$flags) {
 $window = $window|0;
 $flags = $flags|0;
 var $$$021$i$i = 0, $$0 = 0, $$0$i = 0, $$0$i$i$i = 0, $$021$i$i = 0, $$pre = 0, $$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0;
 var $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0;
 var $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0;
 var $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0;
 var $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0;
 var $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0;
 var $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0;
 var $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0;
 var $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0;
 var $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0;
 var $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0;
 var $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0;
 var $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0;
 var $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0.0, $403 = 0.0, $404 = 0, $405 = 0, $406 = 0.0, $407 = 0.0, $408 = 0, $409 = 0, $41 = 0, $410 = 0.0, $411 = 0.0, $412 = 0, $413 = 0, $414 = 0.0, $415 = 0.0;
 var $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0;
 var $434 = 0, $435 = 0, $436 = 0, $437 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $changed_window$0 = 0, $hasCompiler = 0, $major = 0, $minor = 0, $nFormats = 0, $or$cond = 0, $or$cond3 = 0, $or$cond5 = 0, $profile_mask = 0, $value = 0, $vararg_buffer = 0, $vararg_buffer10 = 0, $vararg_buffer102 = 0, $vararg_buffer106 = 0, $vararg_buffer110 = 0;
 var $vararg_buffer114 = 0, $vararg_buffer118 = 0, $vararg_buffer122 = 0, $vararg_buffer126 = 0, $vararg_buffer130 = 0, $vararg_buffer134 = 0, $vararg_buffer138 = 0, $vararg_buffer14 = 0, $vararg_buffer142 = 0, $vararg_buffer146 = 0, $vararg_buffer150 = 0, $vararg_buffer154 = 0, $vararg_buffer158 = 0, $vararg_buffer162 = 0, $vararg_buffer166 = 0, $vararg_buffer170 = 0, $vararg_buffer174 = 0, $vararg_buffer178 = 0, $vararg_buffer18 = 0, $vararg_buffer182 = 0;
 var $vararg_buffer186 = 0, $vararg_buffer190 = 0, $vararg_buffer194 = 0, $vararg_buffer198 = 0, $vararg_buffer2 = 0, $vararg_buffer202 = 0, $vararg_buffer206 = 0, $vararg_buffer210 = 0, $vararg_buffer214 = 0, $vararg_buffer22 = 0, $vararg_buffer26 = 0, $vararg_buffer30 = 0, $vararg_buffer34 = 0, $vararg_buffer38 = 0, $vararg_buffer42 = 0, $vararg_buffer46 = 0, $vararg_buffer50 = 0, $vararg_buffer54 = 0, $vararg_buffer58 = 0, $vararg_buffer6 = 0;
 var $vararg_buffer62 = 0, $vararg_buffer66 = 0, $vararg_buffer70 = 0, $vararg_buffer74 = 0, $vararg_buffer78 = 0, $vararg_buffer82 = 0, $vararg_buffer86 = 0, $vararg_buffer90 = 0, $vararg_buffer94 = 0, $vararg_buffer98 = 0, $vararg_ptr1 = 0, $vararg_ptr101 = 0, $vararg_ptr105 = 0, $vararg_ptr109 = 0, $vararg_ptr113 = 0, $vararg_ptr117 = 0, $vararg_ptr121 = 0, $vararg_ptr125 = 0, $vararg_ptr129 = 0, $vararg_ptr13 = 0;
 var $vararg_ptr133 = 0, $vararg_ptr137 = 0, $vararg_ptr141 = 0, $vararg_ptr145 = 0, $vararg_ptr149 = 0, $vararg_ptr153 = 0, $vararg_ptr157 = 0, $vararg_ptr161 = 0, $vararg_ptr165 = 0, $vararg_ptr169 = 0, $vararg_ptr17 = 0, $vararg_ptr173 = 0, $vararg_ptr177 = 0, $vararg_ptr181 = 0, $vararg_ptr185 = 0, $vararg_ptr189 = 0, $vararg_ptr193 = 0, $vararg_ptr197 = 0, $vararg_ptr201 = 0, $vararg_ptr205 = 0;
 var $vararg_ptr209 = 0, $vararg_ptr21 = 0, $vararg_ptr213 = 0, $vararg_ptr217 = 0, $vararg_ptr218 = 0, $vararg_ptr219 = 0, $vararg_ptr220 = 0, $vararg_ptr221 = 0, $vararg_ptr25 = 0, $vararg_ptr29 = 0, $vararg_ptr33 = 0, $vararg_ptr37 = 0, $vararg_ptr41 = 0, $vararg_ptr45 = 0, $vararg_ptr49 = 0, $vararg_ptr5 = 0, $vararg_ptr53 = 0, $vararg_ptr57 = 0, $vararg_ptr61 = 0, $vararg_ptr65 = 0;
 var $vararg_ptr69 = 0, $vararg_ptr73 = 0, $vararg_ptr77 = 0, $vararg_ptr81 = 0, $vararg_ptr85 = 0, $vararg_ptr89 = 0, $vararg_ptr9 = 0, $vararg_ptr93 = 0, $vararg_ptr97 = 0, $window_framebuffer = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 496|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer214 = sp + 432|0;
 $vararg_buffer210 = sp + 424|0;
 $vararg_buffer206 = sp + 416|0;
 $vararg_buffer202 = sp + 408|0;
 $vararg_buffer198 = sp + 400|0;
 $vararg_buffer194 = sp + 392|0;
 $vararg_buffer190 = sp + 384|0;
 $vararg_buffer186 = sp + 376|0;
 $vararg_buffer182 = sp + 368|0;
 $vararg_buffer178 = sp + 360|0;
 $vararg_buffer174 = sp + 352|0;
 $vararg_buffer170 = sp + 344|0;
 $vararg_buffer166 = sp + 336|0;
 $vararg_buffer162 = sp + 328|0;
 $vararg_buffer158 = sp + 320|0;
 $vararg_buffer154 = sp + 312|0;
 $vararg_buffer150 = sp + 304|0;
 $vararg_buffer146 = sp + 296|0;
 $vararg_buffer142 = sp + 288|0;
 $vararg_buffer138 = sp + 280|0;
 $vararg_buffer134 = sp + 272|0;
 $vararg_buffer130 = sp + 264|0;
 $vararg_buffer126 = sp + 256|0;
 $vararg_buffer122 = sp + 248|0;
 $vararg_buffer118 = sp + 240|0;
 $vararg_buffer114 = sp + 232|0;
 $vararg_buffer110 = sp + 224|0;
 $vararg_buffer106 = sp + 216|0;
 $vararg_buffer102 = sp + 208|0;
 $vararg_buffer98 = sp + 200|0;
 $vararg_buffer94 = sp + 192|0;
 $vararg_buffer90 = sp + 184|0;
 $vararg_buffer86 = sp + 176|0;
 $vararg_buffer82 = sp + 168|0;
 $vararg_buffer78 = sp + 160|0;
 $vararg_buffer74 = sp + 152|0;
 $vararg_buffer70 = sp + 144|0;
 $vararg_buffer66 = sp + 136|0;
 $vararg_buffer62 = sp + 128|0;
 $vararg_buffer58 = sp + 120|0;
 $vararg_buffer54 = sp + 112|0;
 $vararg_buffer50 = sp + 104|0;
 $vararg_buffer46 = sp + 96|0;
 $vararg_buffer42 = sp + 88|0;
 $vararg_buffer38 = sp + 80|0;
 $vararg_buffer34 = sp + 72|0;
 $vararg_buffer30 = sp + 64|0;
 $vararg_buffer26 = sp + 56|0;
 $vararg_buffer22 = sp + 48|0;
 $vararg_buffer18 = sp + 40|0;
 $vararg_buffer14 = sp + 32|0;
 $vararg_buffer10 = sp + 24|0;
 $vararg_buffer6 = sp + 16|0;
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $nFormats = sp + 476|0;
 $hasCompiler = sp + 480|0;
 $window_framebuffer = sp + 472|0;
 $value = sp + 468|0;
 $profile_mask = sp + 464|0;
 $major = sp + 460|0;
 $minor = sp + 456|0;
 HEAP32[$profile_mask>>2] = 0;
 HEAP32[$major>>2] = 0;
 HEAP32[$minor>>2] = 0;
 (_SDL_GL_GetAttribute(21,$profile_mask)|0);
 (_SDL_GL_GetAttribute(17,$major)|0);
 (_SDL_GL_GetAttribute(18,$minor)|0);
 $0 = (_SDL_GetWindowFlags($window)|0);
 $1 = $0 & 2;
 $2 = ($1|0)==(0);
 $3 = HEAP32[$profile_mask>>2]|0;
 $4 = ($3|0)!=(4);
 $or$cond = $2 | $4;
 $5 = HEAP32[$major>>2]|0;
 $6 = ($5|0)!=(2);
 $or$cond3 = $or$cond | $6;
 $7 = HEAP32[$minor>>2]|0;
 $8 = ($7|0)!=(0);
 $or$cond5 = $or$cond3 | $8;
 if ($or$cond5) {
  (_SDL_GL_SetAttribute(21,4)|0);
  (_SDL_GL_SetAttribute(17,2)|0);
  (_SDL_GL_SetAttribute(18,0)|0);
  $9 = $0 | 2;
  $10 = (_SDL_RecreateWindow($window,$9)|0);
  $11 = ($10|0)<(0);
  if (!($11)) {
   $changed_window$0 = 1;
   label = 3;
  }
 } else {
  $changed_window$0 = 0;
  label = 3;
 }
 if ((label|0) == 3) {
  $12 = (_SDL_calloc(1,320)|0);
  $13 = ($12|0)==(0|0);
  do {
   if ($13) {
    (_SDL_Error(0)|0);
   } else {
    $14 = (_SDL_calloc(1,308)|0);
    $15 = ($14|0)==(0|0);
    if ($15) {
     _GLES2_DestroyRenderer($12);
     (_SDL_Error(0)|0);
     break;
    }
    $16 = ((($12)) + 104|0);
    dest=$16; src=(125760); stop=dest+84|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
    $17 = ((($12)) + 108|0);
    HEAP32[$17>>2] = 10;
    $18 = ((($12)) + 316|0);
    HEAP32[$18>>2] = $14;
    $19 = ((($12)) + 188|0);
    HEAP32[$19>>2] = $window;
    $20 = (_SDL_GL_CreateContext($window)|0);
    HEAP32[$14>>2] = $20;
    $21 = ($20|0)==(0|0);
    if ($21) {
     _GLES2_DestroyRenderer($12);
     break;
    }
    $22 = (_SDL_GL_MakeCurrent($window,$20)|0);
    $23 = ($22|0)<(0);
    if ($23) {
     _GLES2_DestroyRenderer($12);
     break;
    }
    $24 = (_SDL_GL_GetProcAddress(616144)|0);
    $25 = ((($14)) + 16|0);
    HEAP32[$25>>2] = $24;
    $26 = ($24|0)==(0|0);
    do {
     if ($26) {
      $27 = (_SDL_GetError()|0);
      HEAP32[$vararg_buffer>>2] = 616144;
      $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
      HEAP32[$vararg_ptr1>>2] = $27;
      $28 = (_SDL_SetError(598006,$vararg_buffer)|0);
      $$0$i = $28;
      label = 119;
     } else {
      $29 = (_SDL_GL_GetProcAddress(613918)|0);
      $30 = ((($14)) + 20|0);
      HEAP32[$30>>2] = $29;
      $31 = ($29|0)==(0|0);
      if ($31) {
       $32 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer2>>2] = 613918;
       $vararg_ptr5 = ((($vararg_buffer2)) + 4|0);
       HEAP32[$vararg_ptr5>>2] = $32;
       $33 = (_SDL_SetError(598006,$vararg_buffer2)|0);
       $$0$i = $33;
       label = 119;
       break;
      }
      $34 = (_SDL_GL_GetProcAddress(615216)|0);
      $35 = ((($14)) + 24|0);
      HEAP32[$35>>2] = $34;
      $36 = ($34|0)==(0|0);
      if ($36) {
       $37 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer6>>2] = 615216;
       $vararg_ptr9 = ((($vararg_buffer6)) + 4|0);
       HEAP32[$vararg_ptr9>>2] = $37;
       $38 = (_SDL_SetError(598006,$vararg_buffer6)|0);
       $$0$i = $38;
       label = 119;
       break;
      }
      $39 = (_SDL_GL_GetProcAddress(614151)|0);
      $40 = ((($14)) + 28|0);
      HEAP32[$40>>2] = $39;
      $41 = ($39|0)==(0|0);
      if ($41) {
       $42 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer10>>2] = 614151;
       $vararg_ptr13 = ((($vararg_buffer10)) + 4|0);
       HEAP32[$vararg_ptr13>>2] = $42;
       $43 = (_SDL_SetError(598006,$vararg_buffer10)|0);
       $$0$i = $43;
       label = 119;
       break;
      }
      $44 = (_SDL_GL_GetProcAddress(616490)|0);
      $45 = ((($14)) + 32|0);
      HEAP32[$45>>2] = $44;
      $46 = ($44|0)==(0|0);
      if ($46) {
       $47 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer14>>2] = 616490;
       $vararg_ptr17 = ((($vararg_buffer14)) + 4|0);
       HEAP32[$vararg_ptr17>>2] = $47;
       $48 = (_SDL_SetError(598006,$vararg_buffer14)|0);
       $$0$i = $48;
       label = 119;
       break;
      }
      $49 = (_SDL_GL_GetProcAddress(616041)|0);
      $50 = ((($14)) + 36|0);
      HEAP32[$50>>2] = $49;
      $51 = ($49|0)==(0|0);
      if ($51) {
       $52 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer18>>2] = 616041;
       $vararg_ptr21 = ((($vararg_buffer18)) + 4|0);
       HEAP32[$vararg_ptr21>>2] = $52;
       $53 = (_SDL_SetError(598006,$vararg_buffer18)|0);
       $$0$i = $53;
       label = 119;
       break;
      }
      $54 = (_SDL_GL_GetProcAddress(616416)|0);
      $55 = ((($14)) + 40|0);
      HEAP32[$55>>2] = $54;
      $56 = ($54|0)==(0|0);
      if ($56) {
       $57 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer22>>2] = 616416;
       $vararg_ptr25 = ((($vararg_buffer22)) + 4|0);
       HEAP32[$vararg_ptr25>>2] = $57;
       $58 = (_SDL_SetError(598006,$vararg_buffer22)|0);
       $$0$i = $58;
       label = 119;
       break;
      }
      $59 = (_SDL_GL_GetProcAddress(615034)|0);
      $60 = ((($14)) + 44|0);
      HEAP32[$60>>2] = $59;
      $61 = ($59|0)==(0|0);
      if ($61) {
       $62 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer26>>2] = 615034;
       $vararg_ptr29 = ((($vararg_buffer26)) + 4|0);
       HEAP32[$vararg_ptr29>>2] = $62;
       $63 = (_SDL_SetError(598006,$vararg_buffer26)|0);
       $$0$i = $63;
       label = 119;
       break;
      }
      $64 = (_SDL_GL_GetProcAddress(613819)|0);
      $65 = ((($14)) + 48|0);
      HEAP32[$65>>2] = $64;
      $66 = ($64|0)==(0|0);
      if ($66) {
       $67 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer30>>2] = 613819;
       $vararg_ptr33 = ((($vararg_buffer30)) + 4|0);
       HEAP32[$vararg_ptr33>>2] = $67;
       $68 = (_SDL_SetError(598006,$vararg_buffer30)|0);
       $$0$i = $68;
       label = 119;
       break;
      }
      $69 = (_SDL_GL_GetProcAddress(613888)|0);
      $70 = ((($14)) + 52|0);
      HEAP32[$70>>2] = $69;
      $71 = ($69|0)==(0|0);
      if ($71) {
       $72 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer34>>2] = 613888;
       $vararg_ptr37 = ((($vararg_buffer34)) + 4|0);
       HEAP32[$vararg_ptr37>>2] = $72;
       $73 = (_SDL_SetError(598006,$vararg_buffer34)|0);
       $$0$i = $73;
       label = 119;
       break;
      }
      $74 = (_SDL_GL_GetProcAddress(615109)|0);
      $75 = ((($14)) + 56|0);
      HEAP32[$75>>2] = $74;
      $76 = ($74|0)==(0|0);
      if ($76) {
       $77 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer38>>2] = 615109;
       $vararg_ptr41 = ((($vararg_buffer38)) + 4|0);
       HEAP32[$vararg_ptr41>>2] = $77;
       $78 = (_SDL_SetError(598006,$vararg_buffer38)|0);
       $$0$i = $78;
       label = 119;
       break;
      }
      $79 = (_SDL_GL_GetProcAddress(614965)|0);
      $80 = ((($14)) + 60|0);
      HEAP32[$80>>2] = $79;
      $81 = ($79|0)==(0|0);
      if ($81) {
       $82 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer42>>2] = 614965;
       $vararg_ptr45 = ((($vararg_buffer42)) + 4|0);
       HEAP32[$vararg_ptr45>>2] = $82;
       $83 = (_SDL_SetError(598006,$vararg_buffer42)|0);
       $$0$i = $83;
       label = 119;
       break;
      }
      $84 = (_SDL_GL_GetProcAddress(614043)|0);
      $85 = ((($14)) + 64|0);
      HEAP32[$85>>2] = $84;
      $86 = ($84|0)==(0|0);
      if ($86) {
       $87 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer46>>2] = 614043;
       $vararg_ptr49 = ((($vararg_buffer46)) + 4|0);
       HEAP32[$vararg_ptr49>>2] = $87;
       $88 = (_SDL_SetError(598006,$vararg_buffer46)|0);
       $$0$i = $88;
       label = 119;
       break;
      }
      $89 = (_SDL_GL_GetProcAddress(616008)|0);
      $90 = ((($14)) + 68|0);
      HEAP32[$90>>2] = $89;
      $91 = ($89|0)==(0|0);
      if ($91) {
       $92 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer50>>2] = 616008;
       $vararg_ptr53 = ((($vararg_buffer50)) + 4|0);
       HEAP32[$vararg_ptr53>>2] = $92;
       $93 = (_SDL_SetError(598006,$vararg_buffer50)|0);
       $$0$i = $93;
       label = 119;
       break;
      }
      $94 = (_SDL_GL_GetProcAddress(615770)|0);
      $95 = ((($14)) + 72|0);
      HEAP32[$95>>2] = $94;
      $96 = ($94|0)==(0|0);
      if ($96) {
       $97 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer54>>2] = 615770;
       $vararg_ptr57 = ((($vararg_buffer54)) + 4|0);
       HEAP32[$vararg_ptr57>>2] = $97;
       $98 = (_SDL_SetError(598006,$vararg_buffer54)|0);
       $$0$i = $98;
       label = 119;
       break;
      }
      $99 = (_SDL_GL_GetProcAddress(615797)|0);
      $100 = ((($14)) + 76|0);
      HEAP32[$100>>2] = $99;
      $101 = ($99|0)==(0|0);
      if ($101) {
       $102 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer58>>2] = 615797;
       $vararg_ptr61 = ((($vararg_buffer58)) + 4|0);
       HEAP32[$vararg_ptr61>>2] = $102;
       $103 = (_SDL_SetError(598006,$vararg_buffer58)|0);
       $$0$i = $103;
       label = 119;
       break;
      }
      $104 = (_SDL_GL_GetProcAddress(615999)|0);
      $105 = ((($14)) + 80|0);
      HEAP32[$105>>2] = $104;
      $106 = ($104|0)==(0|0);
      if ($106) {
       $107 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer62>>2] = 615999;
       $vararg_ptr65 = ((($vararg_buffer62)) + 4|0);
       HEAP32[$vararg_ptr65>>2] = $107;
       $108 = (_SDL_SetError(598006,$vararg_buffer62)|0);
       $$0$i = $108;
       label = 119;
       break;
      }
      $109 = (_SDL_GL_GetProcAddress(615744)|0);
      $110 = ((($14)) + 84|0);
      HEAP32[$110>>2] = $109;
      $111 = ($109|0)==(0|0);
      if ($111) {
       $112 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer66>>2] = 615744;
       $vararg_ptr69 = ((($vararg_buffer66)) + 4|0);
       HEAP32[$vararg_ptr69>>2] = $112;
       $113 = (_SDL_SetError(598006,$vararg_buffer66)|0);
       $$0$i = $113;
       label = 119;
       break;
      }
      $114 = (_SDL_GL_GetProcAddress(615943)|0);
      $115 = ((($14)) + 88|0);
      HEAP32[$115>>2] = $114;
      $116 = ($114|0)==(0|0);
      if ($116) {
       $117 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer70>>2] = 615943;
       $vararg_ptr73 = ((($vararg_buffer70)) + 4|0);
       HEAP32[$vararg_ptr73>>2] = $117;
       $118 = (_SDL_SetError(598006,$vararg_buffer70)|0);
       $$0$i = $118;
       label = 119;
       break;
      }
      $119 = (_SDL_GL_GetProcAddress(615255)|0);
      $120 = ((($14)) + 92|0);
      HEAP32[$120>>2] = $119;
      $121 = ($119|0)==(0|0);
      if ($121) {
       $122 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer74>>2] = 615255;
       $vararg_ptr77 = ((($vararg_buffer74)) + 4|0);
       HEAP32[$vararg_ptr77>>2] = $122;
       $123 = (_SDL_SetError(598006,$vararg_buffer74)|0);
       $$0$i = $123;
       label = 119;
       break;
      }
      $124 = (_SDL_GL_GetProcAddress(614029)|0);
      $125 = ((($14)) + 96|0);
      HEAP32[$125>>2] = $124;
      $126 = ($124|0)==(0|0);
      if ($126) {
       $127 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer78>>2] = 614029;
       $vararg_ptr81 = ((($vararg_buffer78)) + 4|0);
       HEAP32[$vararg_ptr81>>2] = $127;
       $128 = (_SDL_SetError(598006,$vararg_buffer78)|0);
       $$0$i = $128;
       label = 119;
       break;
      }
      $129 = (_SDL_GL_GetProcAddress(614015)|0);
      $130 = ((($14)) + 100|0);
      HEAP32[$130>>2] = $129;
      $131 = ($129|0)==(0|0);
      if ($131) {
       $132 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer82>>2] = 614015;
       $vararg_ptr85 = ((($vararg_buffer82)) + 4|0);
       HEAP32[$vararg_ptr85>>2] = $132;
       $133 = (_SDL_SetError(598006,$vararg_buffer82)|0);
       $$0$i = $133;
       label = 119;
       break;
      }
      $134 = (_SDL_GL_GetProcAddress(613977)|0);
      $135 = ((($14)) + 104|0);
      HEAP32[$135>>2] = $134;
      $136 = ($134|0)==(0|0);
      if ($136) {
       $137 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer86>>2] = 613977;
       $vararg_ptr89 = ((($vararg_buffer86)) + 4|0);
       HEAP32[$vararg_ptr89>>2] = $137;
       $138 = (_SDL_SetError(598006,$vararg_buffer86)|0);
       $$0$i = $138;
       label = 119;
       break;
      }
      $139 = (_SDL_GL_GetProcAddress(615864)|0);
      $140 = ((($14)) + 108|0);
      HEAP32[$140>>2] = $139;
      $141 = ($139|0)==(0|0);
      if ($141) {
       $142 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer90>>2] = 615864;
       $vararg_ptr93 = ((($vararg_buffer90)) + 4|0);
       HEAP32[$vararg_ptr93>>2] = $142;
       $143 = (_SDL_SetError(598006,$vararg_buffer90)|0);
       $$0$i = $143;
       label = 119;
       break;
      }
      $144 = (_SDL_GL_GetProcAddress(613989)|0);
      $145 = ((($14)) + 112|0);
      HEAP32[$145>>2] = $144;
      $146 = ($144|0)==(0|0);
      if ($146) {
       $147 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer94>>2] = 613989;
       $vararg_ptr97 = ((($vararg_buffer94)) + 4|0);
       HEAP32[$vararg_ptr97>>2] = $147;
       $148 = (_SDL_SetError(598006,$vararg_buffer94)|0);
       $$0$i = $148;
       label = 119;
       break;
      }
      $149 = (_SDL_GL_GetProcAddress(615083)|0);
      $150 = ((($14)) + 116|0);
      HEAP32[$150>>2] = $149;
      $151 = ($149|0)==(0|0);
      if ($151) {
       $152 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer98>>2] = 615083;
       $vararg_ptr101 = ((($vararg_buffer98)) + 4|0);
       HEAP32[$vararg_ptr101>>2] = $152;
       $153 = (_SDL_SetError(598006,$vararg_buffer98)|0);
       $$0$i = $153;
       label = 119;
       break;
      }
      $154 = (_SDL_GL_GetProcAddress(615050)|0);
      $155 = ((($14)) + 120|0);
      HEAP32[$155>>2] = $154;
      $156 = ($154|0)==(0|0);
      if ($156) {
       $157 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer102>>2] = 615050;
       $vararg_ptr105 = ((($vararg_buffer102)) + 4|0);
       HEAP32[$vararg_ptr105>>2] = $157;
       $158 = (_SDL_SetError(598006,$vararg_buffer102)|0);
       $$0$i = $158;
       label = 119;
       break;
      }
      $159 = (_SDL_GL_GetProcAddress(615069)|0);
      $160 = ((($14)) + 124|0);
      HEAP32[$160>>2] = $159;
      $161 = ($159|0)==(0|0);
      if ($161) {
       $162 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer106>>2] = 615069;
       $vararg_ptr109 = ((($vararg_buffer106)) + 4|0);
       HEAP32[$vararg_ptr109>>2] = $162;
       $163 = (_SDL_SetError(598006,$vararg_buffer106)|0);
       $$0$i = $163;
       label = 119;
       break;
      }
      $164 = (_SDL_GL_GetProcAddress(614479)|0);
      $165 = ((($14)) + 128|0);
      HEAP32[$165>>2] = $164;
      $166 = ($164|0)==(0|0);
      if ($166) {
       $167 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer110>>2] = 614479;
       $vararg_ptr113 = ((($vararg_buffer110)) + 4|0);
       HEAP32[$vararg_ptr113>>2] = $167;
       $168 = (_SDL_SetError(598006,$vararg_buffer110)|0);
       $$0$i = $168;
       label = 119;
       break;
      }
      $169 = (_SDL_GL_GetProcAddress(615152)|0);
      $170 = ((($14)) + 132|0);
      HEAP32[$170>>2] = $169;
      $171 = ($169|0)==(0|0);
      if ($171) {
       $172 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer114>>2] = 615152;
       $vararg_ptr117 = ((($vararg_buffer114)) + 4|0);
       HEAP32[$vararg_ptr117>>2] = $172;
       $173 = (_SDL_SetError(598006,$vararg_buffer114)|0);
       $$0$i = $173;
       label = 119;
       break;
      }
      $174 = (_SDL_GL_GetProcAddress(613963)|0);
      $175 = ((($14)) + 136|0);
      HEAP32[$175>>2] = $174;
      $176 = ($174|0)==(0|0);
      if ($176) {
       $177 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer118>>2] = 613963;
       $vararg_ptr121 = ((($vararg_buffer118)) + 4|0);
       HEAP32[$vararg_ptr121>>2] = $177;
       $178 = (_SDL_SetError(598006,$vararg_buffer118)|0);
       $$0$i = $178;
       label = 119;
       break;
      }
      $179 = (_SDL_GL_GetProcAddress(614138)|0);
      $180 = ((($14)) + 140|0);
      HEAP32[$180>>2] = $179;
      $181 = ($179|0)==(0|0);
      if ($181) {
       $182 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer122>>2] = 614138;
       $vararg_ptr125 = ((($vararg_buffer122)) + 4|0);
       HEAP32[$vararg_ptr125>>2] = $182;
       $183 = (_SDL_SetError(598006,$vararg_buffer122)|0);
       $$0$i = $183;
       label = 119;
       break;
      }
      $184 = (_SDL_GL_GetProcAddress(616429)|0);
      $185 = ((($14)) + 144|0);
      HEAP32[$185>>2] = $184;
      $186 = ($184|0)==(0|0);
      if ($186) {
       $187 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer126>>2] = 616429;
       $vararg_ptr129 = ((($vararg_buffer126)) + 4|0);
       HEAP32[$vararg_ptr129>>2] = $187;
       $188 = (_SDL_SetError(598006,$vararg_buffer126)|0);
       $$0$i = $188;
       label = 119;
       break;
      }
      $189 = (_SDL_GL_GetProcAddress(615825)|0);
      $190 = ((($14)) + 148|0);
      HEAP32[$190>>2] = $189;
      $191 = ($189|0)==(0|0);
      if ($191) {
       $192 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer130>>2] = 615825;
       $vararg_ptr133 = ((($vararg_buffer130)) + 4|0);
       HEAP32[$vararg_ptr133>>2] = $192;
       $193 = (_SDL_SetError(598006,$vararg_buffer130)|0);
       $$0$i = $193;
       label = 119;
       break;
      }
      $194 = (_SDL_GL_GetProcAddress(615001)|0);
      $195 = ((($14)) + 152|0);
      HEAP32[$195>>2] = $194;
      $196 = ($194|0)==(0|0);
      if ($196) {
       $197 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer134>>2] = 615001;
       $vararg_ptr137 = ((($vararg_buffer134)) + 4|0);
       HEAP32[$vararg_ptr137>>2] = $197;
       $198 = (_SDL_SetError(598006,$vararg_buffer134)|0);
       $$0$i = $198;
       label = 119;
       break;
      }
      $199 = (_SDL_GL_GetProcAddress(614109)|0);
      $200 = ((($14)) + 156|0);
      HEAP32[$200>>2] = $199;
      $201 = ($199|0)==(0|0);
      if ($201) {
       $202 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer138>>2] = 614109;
       $vararg_ptr141 = ((($vararg_buffer138)) + 4|0);
       HEAP32[$vararg_ptr141>>2] = $202;
       $203 = (_SDL_SetError(598006,$vararg_buffer138)|0);
       $$0$i = $203;
       label = 119;
       break;
      }
      $204 = (_SDL_GL_GetProcAddress(616330)|0);
      $205 = ((($14)) + 160|0);
      HEAP32[$205>>2] = $204;
      $206 = ($204|0)==(0|0);
      if ($206) {
       $207 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer142>>2] = 616330;
       $vararg_ptr145 = ((($vararg_buffer142)) + 4|0);
       HEAP32[$vararg_ptr145>>2] = $207;
       $208 = (_SDL_SetError(598006,$vararg_buffer142)|0);
       $$0$i = $208;
       label = 119;
       break;
      }
      $209 = (_SDL_GL_GetProcAddress(614122)|0);
      $210 = ((($14)) + 164|0);
      HEAP32[$210>>2] = $209;
      $211 = ($209|0)==(0|0);
      if ($211) {
       $212 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer146>>2] = 614122;
       $vararg_ptr149 = ((($vararg_buffer146)) + 4|0);
       HEAP32[$vararg_ptr149>>2] = $212;
       $213 = (_SDL_SetError(598006,$vararg_buffer146)|0);
       $$0$i = $213;
       label = 119;
       break;
      }
      $214 = (_SDL_GL_GetProcAddress(614633)|0);
      $215 = ((($14)) + 168|0);
      HEAP32[$215>>2] = $214;
      $216 = ($214|0)==(0|0);
      if ($216) {
       $217 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer150>>2] = 614633;
       $vararg_ptr153 = ((($vararg_buffer150)) + 4|0);
       HEAP32[$vararg_ptr153>>2] = $217;
       $218 = (_SDL_SetError(598006,$vararg_buffer150)|0);
       $$0$i = $218;
       label = 119;
       break;
      }
      $219 = (_SDL_GL_GetProcAddress(614621)|0);
      $220 = ((($14)) + 172|0);
      HEAP32[$220>>2] = $219;
      $221 = ($219|0)==(0|0);
      if ($221) {
       $222 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer154>>2] = 614621;
       $vararg_ptr157 = ((($vararg_buffer154)) + 4|0);
       HEAP32[$vararg_ptr157>>2] = $222;
       $223 = (_SDL_SetError(598006,$vararg_buffer154)|0);
       $$0$i = $223;
       label = 119;
       break;
      }
      $224 = (_SDL_GL_GetProcAddress(614823)|0);
      $225 = ((($14)) + 176|0);
      HEAP32[$225>>2] = $224;
      $226 = ($224|0)==(0|0);
      if ($226) {
       $227 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer158>>2] = 614823;
       $vararg_ptr161 = ((($vararg_buffer158)) + 4|0);
       HEAP32[$vararg_ptr161>>2] = $227;
       $228 = (_SDL_SetError(598006,$vararg_buffer158)|0);
       $$0$i = $228;
       label = 119;
       break;
      }
      $229 = (_SDL_GL_GetProcAddress(613854)|0);
      $230 = ((($14)) + 180|0);
      HEAP32[$230>>2] = $229;
      $231 = ($229|0)==(0|0);
      if ($231) {
       $232 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer162>>2] = 613854;
       $vararg_ptr165 = ((($vararg_buffer162)) + 4|0);
       HEAP32[$vararg_ptr165>>2] = $232;
       $233 = (_SDL_SetError(598006,$vararg_buffer162)|0);
       $$0$i = $233;
       label = 119;
       break;
      }
      $234 = (_SDL_GL_GetProcAddress(615722)|0);
      $235 = ((($14)) + 184|0);
      HEAP32[$235>>2] = $234;
      $236 = ($234|0)==(0|0);
      if ($236) {
       $237 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer166>>2] = 615722;
       $vararg_ptr169 = ((($vararg_buffer166)) + 4|0);
       HEAP32[$vararg_ptr169>>2] = $237;
       $238 = (_SDL_SetError(598006,$vararg_buffer166)|0);
       $$0$i = $238;
       label = 119;
       break;
      }
      $239 = (_SDL_GL_GetProcAddress(616405)|0);
      $240 = ((($14)) + 188|0);
      HEAP32[$240>>2] = $239;
      $241 = ($239|0)==(0|0);
      if ($241) {
       $242 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer170>>2] = 616405;
       $vararg_ptr173 = ((($vararg_buffer170)) + 4|0);
       HEAP32[$vararg_ptr173>>2] = $242;
       $243 = (_SDL_SetError(598006,$vararg_buffer170)|0);
       $$0$i = $243;
       label = 119;
       break;
      }
      $244 = (_SDL_GL_GetProcAddress(615237)|0);
      $245 = ((($14)) + 192|0);
      HEAP32[$245>>2] = $244;
      $246 = ($244|0)==(0|0);
      if ($246) {
       $247 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer174>>2] = 615237;
       $vararg_ptr177 = ((($vararg_buffer174)) + 4|0);
       HEAP32[$vararg_ptr177>>2] = $247;
       $248 = (_SDL_SetError(598006,$vararg_buffer174)|0);
       $$0$i = $248;
       label = 119;
       break;
      }
      $249 = (_SDL_GL_GetProcAddress(615320)|0);
      $250 = ((($14)) + 196|0);
      HEAP32[$250>>2] = $249;
      $251 = ($249|0)==(0|0);
      if ($251) {
       $252 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer178>>2] = 615320;
       $vararg_ptr181 = ((($vararg_buffer178)) + 4|0);
       HEAP32[$vararg_ptr181>>2] = $252;
       $253 = (_SDL_SetError(598006,$vararg_buffer178)|0);
       $$0$i = $253;
       label = 119;
       break;
      }
      $254 = (_SDL_GL_GetProcAddress(616102)|0);
      $255 = ((($14)) + 200|0);
      HEAP32[$255>>2] = $254;
      $256 = ($254|0)==(0|0);
      if ($256) {
       $257 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer182>>2] = 616102;
       $vararg_ptr185 = ((($vararg_buffer182)) + 4|0);
       HEAP32[$vararg_ptr185>>2] = $257;
       $258 = (_SDL_SetError(598006,$vararg_buffer182)|0);
       $$0$i = $258;
       label = 119;
       break;
      }
      $259 = (_SDL_GL_GetProcAddress(615273)|0);
      $260 = ((($14)) + 204|0);
      HEAP32[$260>>2] = $259;
      $261 = ($259|0)==(0|0);
      if ($261) {
       $262 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer186>>2] = 615273;
       $vararg_ptr189 = ((($vararg_buffer186)) + 4|0);
       HEAP32[$vararg_ptr189>>2] = $262;
       $263 = (_SDL_SetError(598006,$vararg_buffer186)|0);
       $$0$i = $263;
       label = 119;
       break;
      }
      $264 = (_SDL_GL_GetProcAddress(614927)|0);
      $265 = ((($14)) + 208|0);
      HEAP32[$265>>2] = $264;
      $266 = ($264|0)==(0|0);
      if ($266) {
       $267 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer190>>2] = 614927;
       $vararg_ptr193 = ((($vararg_buffer190)) + 4|0);
       HEAP32[$vararg_ptr193>>2] = $267;
       $268 = (_SDL_SetError(598006,$vararg_buffer190)|0);
       $$0$i = $268;
       label = 119;
       break;
      }
      $269 = (_SDL_GL_GetProcAddress(615166)|0);
      $270 = ((($14)) + 212|0);
      HEAP32[$270>>2] = $269;
      $271 = ($269|0)==(0|0);
      if ($271) {
       $272 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer194>>2] = 615166;
       $vararg_ptr197 = ((($vararg_buffer194)) + 4|0);
       HEAP32[$vararg_ptr197>>2] = $272;
       $273 = (_SDL_SetError(598006,$vararg_buffer194)|0);
       $$0$i = $273;
       label = 119;
       break;
      }
      $274 = (_SDL_GL_GetProcAddress(614251)|0);
      $275 = ((($14)) + 216|0);
      HEAP32[$275>>2] = $274;
      $276 = ($274|0)==(0|0);
      if ($276) {
       $277 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer198>>2] = 614251;
       $vararg_ptr201 = ((($vararg_buffer198)) + 4|0);
       HEAP32[$vararg_ptr201>>2] = $277;
       $278 = (_SDL_SetError(598006,$vararg_buffer198)|0);
       $$0$i = $278;
       label = 119;
       break;
      }
      $279 = (_SDL_GL_GetProcAddress(614842)|0);
      $280 = ((($14)) + 220|0);
      HEAP32[$280>>2] = $279;
      $281 = ($279|0)==(0|0);
      if ($281) {
       $282 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer202>>2] = 614842;
       $vararg_ptr205 = ((($vararg_buffer202)) + 4|0);
       HEAP32[$vararg_ptr205>>2] = $282;
       $283 = (_SDL_SetError(598006,$vararg_buffer202)|0);
       $$0$i = $283;
       label = 119;
       break;
      }
      $284 = (_SDL_GL_GetProcAddress(614303)|0);
      $285 = ((($14)) + 224|0);
      HEAP32[$285>>2] = $284;
      $286 = ($284|0)==(0|0);
      if ($286) {
       $287 = (_SDL_GetError()|0);
       HEAP32[$vararg_buffer206>>2] = 614303;
       $vararg_ptr209 = ((($vararg_buffer206)) + 4|0);
       HEAP32[$vararg_ptr209>>2] = $287;
       $288 = (_SDL_SetError(598006,$vararg_buffer206)|0);
       $$0$i = $288;
       label = 119;
       break;
      }
      $289 = (_SDL_GL_GetProcAddress(614316)|0);
      $290 = ((($14)) + 228|0);
      HEAP32[$290>>2] = $289;
      $291 = ($289|0)==(0|0);
      if (!($291)) {
       break;
      }
      $292 = (_SDL_GetError()|0);
      HEAP32[$vararg_buffer210>>2] = 614316;
      $vararg_ptr213 = ((($vararg_buffer210)) + 4|0);
      HEAP32[$vararg_ptr213>>2] = $292;
      $293 = (_SDL_SetError(598006,$vararg_buffer210)|0);
      $$0$i = $293;
      label = 119;
     }
    } while(0);
    if ((label|0) == 119) {
     $294 = ($$0$i|0)<(0);
     if ($294) {
      _GLES2_DestroyRenderer($12);
      break;
     }
    }
    $295 = $flags & 4;
    $296 = ($295|0)==(0);
    if ($296) {
     (_SDL_GL_SetSwapInterval(0)|0);
    } else {
     (_SDL_GL_SetSwapInterval(1)|0);
    }
    $297 = (_SDL_GL_GetSwapInterval()|0);
    $298 = ($297|0)>(0);
    if ($298) {
     $299 = HEAP32[$17>>2]|0;
     $300 = $299 | 4;
     HEAP32[$17>>2] = $300;
    }
    $301 = (_SDL_GL_GetAttribute(20,$value)|0);
    $302 = ($301|0)==(0);
    if ($302) {
     $303 = HEAP32[$value>>2]|0;
     $304 = $303 & 1;
     $305 = ($304|0)==(0);
     if (!($305)) {
      $306 = ((($14)) + 4|0);
      HEAP32[$306>>2] = 1;
     }
    }
    HEAP32[$value>>2] = 0;
    $307 = ((($14)) + 112|0);
    $308 = HEAP32[$307>>2]|0;
    FUNCTION_TABLE_vii[$308 & 8191](3379,$value);
    $309 = HEAP32[$value>>2]|0;
    $310 = ((($12)) + 180|0);
    HEAP32[$310>>2] = $309;
    HEAP32[$value>>2] = 0;
    $311 = HEAP32[$307>>2]|0;
    FUNCTION_TABLE_vii[$311 & 8191](3379,$value);
    $312 = HEAP32[$value>>2]|0;
    $313 = ((($12)) + 184|0);
    HEAP32[$313>>2] = $312;
    $314 = HEAP32[$307>>2]|0;
    FUNCTION_TABLE_vii[$314 & 8191](36345,$nFormats);
    $315 = ((($14)) + 100|0);
    $316 = HEAP32[$315>>2]|0;
    FUNCTION_TABLE_vii[$316 & 8191](36346,$hasCompiler);
    $317 = HEAP8[$hasCompiler>>0]|0;
    $318 = ($317<<24>>24)==(0);
    $$pre = HEAP32[$nFormats>>2]|0;
    if ($318) {
     $320 = $$pre;
    } else {
     $319 = (($$pre) + 1)|0;
     HEAP32[$nFormats>>2] = $319;
     $320 = $319;
    }
    $321 = (_SDL_calloc($320,4)|0);
    $322 = ((($14)) + 244|0);
    HEAP32[$322>>2] = $321;
    $323 = ($321|0)==(0|0);
    if ($323) {
     _GLES2_DestroyRenderer($12);
     (_SDL_Error(0)|0);
     break;
    }
    $324 = HEAP32[$nFormats>>2]|0;
    $325 = ((($14)) + 240|0);
    HEAP32[$325>>2] = $324;
    $326 = HEAP32[$307>>2]|0;
    FUNCTION_TABLE_vii[$326 & 8191](36344,$321);
    $327 = HEAP8[$hasCompiler>>0]|0;
    $328 = ($327<<24>>24)==(0);
    if (!($328)) {
     $329 = HEAP32[$nFormats>>2]|0;
     $330 = (($329) + -1)|0;
     $331 = HEAP32[$322>>2]|0;
     $332 = (($331) + ($330<<2)|0);
     HEAP32[$332>>2] = -1;
    }
    $333 = ((($14)) + 232|0);
    HEAP32[$333>>2] = 0;
    $334 = HEAP32[$307>>2]|0;
    FUNCTION_TABLE_vii[$334 & 8191](36006,$window_framebuffer);
    $335 = HEAP32[$window_framebuffer>>2]|0;
    $336 = ((($14)) + 236|0);
    HEAP32[$336>>2] = $335;
    $337 = ((($12)) + 4|0);
    HEAP32[$337>>2] = 4454;
    $338 = ((($12)) + 8|0);
    HEAP32[$338>>2] = 4455;
    $339 = ((($12)) + 12|0);
    HEAP32[$339>>2] = 4456;
    $340 = ((($12)) + 28|0);
    HEAP32[$340>>2] = 4457;
    $341 = ((($12)) + 32|0);
    HEAP32[$341>>2] = 4458;
    $342 = ((($12)) + 36|0);
    HEAP32[$342>>2] = 4459;
    $343 = ((($12)) + 40|0);
    HEAP32[$343>>2] = 4460;
    $344 = ((($12)) + 44|0);
    HEAP32[$344>>2] = 4461;
    $345 = ((($12)) + 48|0);
    HEAP32[$345>>2] = 4462;
    $346 = ((($12)) + 52|0);
    HEAP32[$346>>2] = 4463;
    $347 = ((($12)) + 56|0);
    HEAP32[$347>>2] = 4464;
    $348 = ((($12)) + 60|0);
    HEAP32[$348>>2] = 4465;
    $349 = ((($12)) + 64|0);
    HEAP32[$349>>2] = 4466;
    $350 = ((($12)) + 68|0);
    HEAP32[$350>>2] = 4467;
    $351 = ((($12)) + 72|0);
    HEAP32[$351>>2] = 4468;
    $352 = ((($12)) + 76|0);
    HEAP32[$352>>2] = 4469;
    $353 = ((($12)) + 80|0);
    HEAP32[$353>>2] = 4470;
    $354 = ((($12)) + 84|0);
    HEAP32[$354>>2] = 4471;
    $355 = ((($12)) + 88|0);
    HEAP32[$355>>2] = 4472;
    $356 = ((($12)) + 92|0);
    HEAP32[$356>>2] = 4473;
    $357 = ((($12)) + 96|0);
    HEAP32[$357>>2] = 4474;
    $358 = ((($12)) + 100|0);
    HEAP32[$358>>2] = 4475;
    $359 = ((($12)) + 112|0);
    $360 = HEAP32[$359>>2]|0;
    $361 = (($360) + 1)|0;
    HEAP32[$359>>2] = $361;
    $362 = ((($12)) + 116|0);
    $363 = (($362) + ($360<<2)|0);
    HEAP32[$363>>2] = 842094169;
    $364 = HEAP32[$359>>2]|0;
    $365 = (($364) + 1)|0;
    HEAP32[$359>>2] = $365;
    $366 = (($362) + ($364<<2)|0);
    HEAP32[$366>>2] = 1448433993;
    $367 = HEAP32[$359>>2]|0;
    $368 = (($367) + 1)|0;
    HEAP32[$359>>2] = $368;
    $369 = (($362) + ($367<<2)|0);
    HEAP32[$369>>2] = 842094158;
    $370 = HEAP32[$359>>2]|0;
    $371 = (($370) + 1)|0;
    HEAP32[$359>>2] = $371;
    $372 = (($362) + ($370<<2)|0);
    HEAP32[$372>>2] = 825382478;
    $373 = HEAP32[$18>>2]|0;
    $374 = HEAP32[300738]|0;
    $375 = HEAP32[$373>>2]|0;
    $376 = ($374|0)==($375|0);
    if ($376) {
     (_GLES2_UpdateViewport($12)|0);
    } else {
     $377 = ((($373)) + 268|0);
     HEAP32[$377>>2] = 0;
     $378 = HEAP32[$19>>2]|0;
     $379 = (_SDL_GL_MakeCurrent($378,$375)|0);
     $380 = ($379|0)<(0);
     if (!($380)) {
      $381 = HEAP32[$373>>2]|0;
      HEAP32[300738] = $381;
      (_GLES2_UpdateViewport($12)|0);
      $$pre$i$i = HEAP32[$18>>2]|0;
      $382 = ((($$pre$i$i)) + 4|0);
      $383 = HEAP32[$382>>2]|0;
      $384 = ($383|0)==(0);
      if (!($384)) {
       $385 = ((($$pre$i$i)) + 108|0);
       while(1) {
        $386 = HEAP32[$385>>2]|0;
        $387 = (FUNCTION_TABLE_i[$386 & 8191]()|0);
        $388 = ($387|0)==(0);
        if ($388) {
         break;
        }
       }
      }
     }
    }
    $389 = ((($373)) + 8|0);
    HEAP32[$389>>2] = -1;
    $390 = ((($373)) + 12|0);
    HEAP32[$390>>2] = 0;
    $391 = ((($373)) + 16|0);
    $392 = HEAP32[$391>>2]|0;
    FUNCTION_TABLE_vi[$392 & 8191](33984);
    $393 = ((($373)) + 136|0);
    $394 = HEAP32[$393>>2]|0;
    FUNCTION_TABLE_vii[$394 & 8191](3333,1);
    $395 = HEAP32[$393>>2]|0;
    FUNCTION_TABLE_vii[$395 & 8191](3317,1);
    $396 = ((($373)) + 40|0);
    $397 = HEAP32[$396>>2]|0;
    $398 = ((($373)) + 272|0);
    $399 = HEAP32[$398>>2]|0;
    $400 = $399&65535;
    $401 = $399&255;
    $402 = (+($401&255));
    $403 = $402 * 0.0039215688593685627;
    $404 = ($400&65535) >>> 8;
    $405 = $404&255;
    $406 = (+($405&255));
    $407 = $406 * 0.0039215688593685627;
    $408 = $399 >>> 16;
    $409 = $408&255;
    $410 = (+($409&255));
    $411 = $410 * 0.0039215688593685627;
    $412 = $399 >>> 24;
    $413 = $412&255;
    $414 = (+($413&255));
    $415 = $414 * 0.0039215688593685627;
    FUNCTION_TABLE_vdddd[$397 & 8191]($403,$407,$411,$415);
    $416 = ((($373)) + 84|0);
    $417 = HEAP32[$416>>2]|0;
    FUNCTION_TABLE_vi[$417 & 8191](0);
    $418 = ((($373)) + 72|0);
    $419 = HEAP32[$418>>2]|0;
    FUNCTION_TABLE_vi[$419 & 8191](1);
    $420 = HEAP32[$18>>2]|0;
    $421 = ((($420)) + 4|0);
    $422 = HEAP32[$421>>2]|0;
    $423 = ($422|0)==(0);
    if ($423) {
     $$0 = $12;
     STACKTOP = sp;return ($$0|0);
    }
    $424 = ((($420)) + 108|0);
    $425 = HEAP32[$424>>2]|0;
    $426 = (FUNCTION_TABLE_i[$425 & 8191]()|0);
    $427 = ($426|0)==(0);
    if ($427) {
     $$0 = $12;
     STACKTOP = sp;return ($$0|0);
    } else {
     $$021$i$i = 1282466;$430 = $426;
    }
    while(1) {
     $428 = HEAP8[$$021$i$i>>0]|0;
     $429 = ($428<<24>>24)==(0);
     $$$021$i$i = $429 ? 597716 : $$021$i$i;
     switch ($430|0) {
     case 1280:  {
      $$0$i$i$i = 597787;
      break;
     }
     case 1281:  {
      $$0$i$i$i = 597770;
      break;
     }
     case 1282:  {
      $$0$i$i$i = 597749;
      break;
     }
     case 1285:  {
      $$0$i$i$i = 597732;
      break;
     }
     default: {
      $$0$i$i$i = 597724;
     }
     }
     HEAP32[$vararg_buffer214>>2] = $$$021$i$i;
     $vararg_ptr217 = ((($vararg_buffer214)) + 4|0);
     HEAP32[$vararg_ptr217>>2] = 597829;
     $vararg_ptr218 = ((($vararg_buffer214)) + 8|0);
     HEAP32[$vararg_ptr218>>2] = 1944;
     $vararg_ptr219 = ((($vararg_buffer214)) + 12|0);
     HEAP32[$vararg_ptr219>>2] = 603308;
     $vararg_ptr220 = ((($vararg_buffer214)) + 16|0);
     HEAP32[$vararg_ptr220>>2] = $$0$i$i$i;
     $vararg_ptr221 = ((($vararg_buffer214)) + 20|0);
     HEAP32[$vararg_ptr221>>2] = $430;
     (_SDL_SetError(597803,$vararg_buffer214)|0);
     $431 = HEAP32[$424>>2]|0;
     $432 = (FUNCTION_TABLE_i[$431 & 8191]()|0);
     $433 = ($432|0)==(0);
     if ($433) {
      $$0 = $12;
      break;
     } else {
      $$021$i$i = $$$021$i$i;$430 = $432;
     }
    }
    STACKTOP = sp;return ($$0|0);
   }
  } while(0);
  $434 = ($changed_window$0|0)==(0);
  if ($434) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $435 = HEAP32[$profile_mask>>2]|0;
 (_SDL_GL_SetAttribute(21,$435)|0);
 $436 = HEAP32[$major>>2]|0;
 (_SDL_GL_SetAttribute(17,$436)|0);
 $437 = HEAP32[$minor>>2]|0;
 (_SDL_GL_SetAttribute(18,$437)|0);
 (_SDL_RecreateWindow($window,$0)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GL_GetAttribute($attr,$value) {
 $attr = $attr|0;
 $value = $value|0;
 var $$ = 0, $$0 = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $8 = 0, $9 = 0, $asize = 0, $attrib$0 = 0;
 var $bsize = 0, $gsize = 0, $rsize = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer11 = 0, $vararg_buffer13 = 0, $vararg_buffer15 = 0, $vararg_buffer17 = 0, $vararg_buffer19 = 0, $vararg_buffer21 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, $vararg_buffer9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer21 = sp + 88|0;
 $vararg_buffer19 = sp + 80|0;
 $vararg_buffer17 = sp + 72|0;
 $vararg_buffer15 = sp + 64|0;
 $vararg_buffer13 = sp + 56|0;
 $vararg_buffer11 = sp + 48|0;
 $vararg_buffer9 = sp + 40|0;
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $rsize = sp + 104|0;
 $gsize = sp + 100|0;
 $bsize = sp + 96|0;
 $asize = sp + 92|0;
 HEAP32[$value>>2] = 0;
 do {
  switch ($attr|0) {
  case 0:  {
   $attrib$0 = 3410;
   break;
  }
  case 2:  {
   $attrib$0 = 3412;
   break;
  }
  case 1:  {
   $attrib$0 = 3411;
   break;
  }
  case 3:  {
   $attrib$0 = 3413;
   break;
  }
  case 5:  {
   $0 = HEAP32[300441]|0;
   $1 = ((($0)) + 296|0);
   $2 = HEAP32[$1>>2]|0;
   HEAP32[$value>>2] = $2;
   $$2 = 0;
   STACKTOP = sp;return ($$2|0);
   break;
  }
  case 6:  {
   $attrib$0 = 3414;
   break;
  }
  case 7:  {
   $attrib$0 = 3415;
   break;
  }
  case 12: case 11: case 10: case 9: case 8:  {
   HEAP32[$value>>2] = 0;
   $$2 = 0;
   STACKTOP = sp;return ($$2|0);
   break;
  }
  case 13:  {
   $attrib$0 = 32936;
   break;
  }
  case 14:  {
   $attrib$0 = 32937;
   break;
  }
  case 24:  {
   $attrib$0 = 33531;
   break;
  }
  case 4:  {
   HEAP32[$rsize>>2] = 0;
   HEAP32[$gsize>>2] = 0;
   HEAP32[$bsize>>2] = 0;
   HEAP32[$asize>>2] = 0;
   $3 = (_SDL_GL_GetAttribute(0,$rsize)|0);
   $4 = ($3|0)<(0);
   if ($4) {
    $$0 = -1;
   } else {
    $5 = (_SDL_GL_GetAttribute(1,$gsize)|0);
    $6 = ($5|0)<(0);
    if ($6) {
     $$0 = -1;
    } else {
     $7 = (_SDL_GL_GetAttribute(2,$bsize)|0);
     $8 = ($7|0)<(0);
     if ($8) {
      $$0 = -1;
     } else {
      $9 = (_SDL_GL_GetAttribute(3,$asize)|0);
      $10 = ($9|0)<(0);
      if ($10) {
       $$0 = -1;
      } else {
       $11 = HEAP32[$rsize>>2]|0;
       $12 = HEAP32[$gsize>>2]|0;
       $13 = (($12) + ($11))|0;
       $14 = HEAP32[$bsize>>2]|0;
       $15 = (($13) + ($14))|0;
       $16 = HEAP32[$asize>>2]|0;
       $17 = (($15) + ($16))|0;
       HEAP32[$value>>2] = $17;
       $$0 = 0;
      }
     }
    }
   }
   $$2 = $$0;
   STACKTOP = sp;return ($$2|0);
   break;
  }
  case 15:  {
   $18 = HEAP32[300441]|0;
   $19 = ((($18)) + 328|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = ($20|0)!=(0);
   $22 = $21&1;
   HEAP32[$value>>2] = $22;
   $$2 = 0;
   STACKTOP = sp;return ($$2|0);
   break;
  }
  case 16:  {
   $23 = HEAP32[300441]|0;
   $24 = ((($23)) + 360|0);
   $25 = HEAP32[$24>>2]|0;
   HEAP32[$value>>2] = $25;
   $$2 = 0;
   STACKTOP = sp;return ($$2|0);
   break;
  }
  case 17:  {
   $26 = HEAP32[300441]|0;
   $27 = ((($26)) + 332|0);
   $28 = HEAP32[$27>>2]|0;
   HEAP32[$value>>2] = $28;
   $$2 = 0;
   STACKTOP = sp;return ($$2|0);
   break;
  }
  case 18:  {
   $29 = HEAP32[300441]|0;
   $30 = ((($29)) + 336|0);
   $31 = HEAP32[$30>>2]|0;
   HEAP32[$value>>2] = $31;
   $$2 = 0;
   STACKTOP = sp;return ($$2|0);
   break;
  }
  case 19:  {
   $32 = HEAP32[300441]|0;
   $33 = ((($32)) + 344|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = ($34|0)==(4);
   $$ = $35&1;
   HEAP32[$value>>2] = $$;
   $$2 = 0;
   STACKTOP = sp;return ($$2|0);
   break;
  }
  case 20:  {
   $36 = HEAP32[300441]|0;
   $37 = ((($36)) + 340|0);
   $38 = HEAP32[$37>>2]|0;
   HEAP32[$value>>2] = $38;
   $$2 = 0;
   STACKTOP = sp;return ($$2|0);
   break;
  }
  case 21:  {
   $39 = HEAP32[300441]|0;
   $40 = ((($39)) + 344|0);
   $41 = HEAP32[$40>>2]|0;
   HEAP32[$value>>2] = $41;
   $$2 = 0;
   STACKTOP = sp;return ($$2|0);
   break;
  }
  case 22:  {
   $42 = HEAP32[300441]|0;
   $43 = ((($42)) + 348|0);
   $44 = HEAP32[$43>>2]|0;
   HEAP32[$value>>2] = $44;
   $$2 = 0;
   STACKTOP = sp;return ($$2|0);
   break;
  }
  case 23:  {
   $45 = HEAP32[300441]|0;
   $46 = ((($45)) + 356|0);
   $47 = HEAP32[$46>>2]|0;
   HEAP32[$value>>2] = $47;
   $$2 = 0;
   STACKTOP = sp;return ($$2|0);
   break;
  }
  default: {
   $48 = (_SDL_SetError(595842,$vararg_buffer)|0);
   $$2 = $48;
   STACKTOP = sp;return ($$2|0);
  }
  }
 } while(0);
 $49 = HEAP32[300441]|0;
 $50 = ($49|0)==(0|0);
 do {
  if ($50) {
   (_SDL_SetError(595917,$vararg_buffer1)|0);
  } else {
   $51 = ((($49)) + 144|0);
   $52 = HEAP32[$51>>2]|0;
   $53 = ($52|0)==(0|0);
   if ($53) {
    (_SDL_SetError(595987,$vararg_buffer5)|0);
    break;
   }
   $54 = ((($49)) + 364|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = ($55|0)==(0);
   if ($56) {
    (_SDL_SetError(595958,$vararg_buffer3)|0);
    break;
   }
   $57 = (FUNCTION_TABLE_iii[$52 & 8191]($49,613989)|0);
   $58 = ($57|0)==(0|0);
   if (!($58)) {
    FUNCTION_TABLE_vii[$57 & 8191]($attrib$0,$value);
    $60 = HEAP32[300441]|0;
    $61 = ($60|0)==(0|0);
    do {
     if ($61) {
      (_SDL_SetError(595917,$vararg_buffer9)|0);
     } else {
      $62 = ((($60)) + 144|0);
      $63 = HEAP32[$62>>2]|0;
      $64 = ($63|0)==(0|0);
      if ($64) {
       (_SDL_SetError(595987,$vararg_buffer13)|0);
       break;
      }
      $65 = ((($60)) + 364|0);
      $66 = HEAP32[$65>>2]|0;
      $67 = ($66|0)==(0);
      if ($67) {
       (_SDL_SetError(595958,$vararg_buffer11)|0);
       break;
      }
      $68 = (FUNCTION_TABLE_iii[$63 & 8191]($60,615864)|0);
      $69 = ($68|0)==(0|0);
      if (!($69)) {
       $71 = (FUNCTION_TABLE_i[$68 & 8191]()|0);
       switch ($71|0) {
       case 1280:  {
        $72 = (_SDL_SetError(596118,$vararg_buffer17)|0);
        $$2 = $72;
        STACKTOP = sp;return ($$2|0);
        break;
       }
       case 1281:  {
        $73 = (_SDL_SetError(596148,$vararg_buffer19)|0);
        $$2 = $73;
        STACKTOP = sp;return ($$2|0);
        break;
       }
       case 0:  {
        $$2 = 0;
        STACKTOP = sp;return ($$2|0);
        break;
       }
       default: {
        HEAP32[$vararg_buffer21>>2] = $71;
        $74 = (_SDL_SetError(596179,$vararg_buffer21)|0);
        $$2 = $74;
        STACKTOP = sp;return ($$2|0);
       }
       }
      }
     }
    } while(0);
    $70 = (_SDL_SetError(596073,$vararg_buffer15)|0);
    $$2 = $70;
    STACKTOP = sp;return ($$2|0);
   }
  }
 } while(0);
 $59 = (_SDL_SetError(596025,$vararg_buffer7)|0);
 $$2 = $59;
 STACKTOP = sp;return ($$2|0);
}
function _SDL_SetError($fmt,$varargs) {
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $$ = 0, $$01$be = 0, $$015 = 0, $$1 = 0, $$1$lcssa = 0, $$2 = 0, $$lcssa = 0, $$lcssa11 = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ap = 0, $arglist_current = 0, $arglist_current2 = 0, $arglist_current5 = 0, $arglist_current8 = 0, $arglist_next = 0, $arglist_next3 = 0, $arglist_next6 = 0, $arglist_next9 = 0, $expanded = 0, $expanded10 = 0, $expanded12 = 0, $expanded13 = 0, $expanded14 = 0, $expanded16 = 0;
 var $expanded17 = 0, $expanded19 = 0, $expanded20 = 0, $expanded21 = 0, $expanded23 = 0, $expanded24 = 0, $expanded26 = 0, $expanded27 = 0, $expanded28 = 0, $expanded3 = 0, $expanded5 = 0, $expanded6 = 0, $expanded7 = 0, $expanded9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $ap = sp + 8|0;
 $0 = ($fmt|0)==(0|0);
 if ($0) {
  STACKTOP = sp;return -1;
 }
 HEAP32[156036] = 1;
 (_SDL_strlcpy((624148),$fmt,128)|0);
 HEAP32[$ap>>2] = $varargs;
 HEAP32[(624276)>>2] = 0;
 $1 = HEAP8[$fmt>>0]|0;
 $2 = ($1<<24>>24)==(0);
 L4: do {
  if (!($2)) {
   $$015 = $fmt;$20 = 0;$5 = $1;
   while(1) {
    $3 = ((($$015)) + 1|0);
    $4 = ($5<<24>>24)==(37);
    if ($4) {
     $$1 = $3;
     while(1) {
      $6 = HEAP8[$$1>>0]|0;
      $7 = ($6<<24>>24)==(46);
      if ($7) {
       $8 = ((($$1)) + 1|0);
       $$1 = $8;
       continue;
      } else {
       $$off = (($6) + -48)<<24>>24;
       $9 = ($$off&255)<(10);
       $10 = ((($$1)) + 1|0);
       if ($9) {
        $$1 = $10;
        continue;
       } else {
        $$1$lcssa = $$1;$$lcssa = $6;$$lcssa11 = $10;
        break;
       }
      }
     }
     $11 = $$lcssa << 24 >> 24;
     switch ($11|0) {
     case 0:  {
      $$2 = $$1$lcssa;$52 = $20;
      break;
     }
     case 88: case 120: case 111: case 117: case 100: case 105: case 99:  {
      $arglist_current = HEAP32[$ap>>2]|0;
      $12 = $arglist_current;
      $13 = ((0) + 4|0);
      $expanded3 = $13;
      $expanded = (($expanded3) - 1)|0;
      $14 = (($12) + ($expanded))|0;
      $15 = ((0) + 4|0);
      $expanded7 = $15;
      $expanded6 = (($expanded7) - 1)|0;
      $expanded5 = $expanded6 ^ -1;
      $16 = $14 & $expanded5;
      $17 = $16;
      $18 = HEAP32[$17>>2]|0;
      $arglist_next = ((($17)) + 4|0);
      HEAP32[$ap>>2] = $arglist_next;
      $19 = (($20) + 1)|0;
      HEAP32[(624276)>>2] = $19;
      $21 = (624280 + ($20<<7)|0);
      HEAP32[$21>>2] = $18;
      $$2 = $$lcssa11;$52 = $19;
      break;
     }
     case 102:  {
      $arglist_current2 = HEAP32[$ap>>2]|0;
      $22 = $arglist_current2;
      $23 = ((0) + 8|0);
      $expanded10 = $23;
      $expanded9 = (($expanded10) - 1)|0;
      $24 = (($22) + ($expanded9))|0;
      $25 = ((0) + 8|0);
      $expanded14 = $25;
      $expanded13 = (($expanded14) - 1)|0;
      $expanded12 = $expanded13 ^ -1;
      $26 = $24 & $expanded12;
      $27 = $26;
      $28 = +HEAPF64[$27>>3];
      $arglist_next3 = ((($27)) + 8|0);
      HEAP32[$ap>>2] = $arglist_next3;
      $29 = (($20) + 1)|0;
      HEAP32[(624276)>>2] = $29;
      $30 = (624280 + ($20<<7)|0);
      HEAPF64[$30>>3] = $28;
      $$2 = $$lcssa11;$52 = $29;
      break;
     }
     case 112:  {
      $arglist_current5 = HEAP32[$ap>>2]|0;
      $31 = $arglist_current5;
      $32 = ((0) + 4|0);
      $expanded17 = $32;
      $expanded16 = (($expanded17) - 1)|0;
      $33 = (($31) + ($expanded16))|0;
      $34 = ((0) + 4|0);
      $expanded21 = $34;
      $expanded20 = (($expanded21) - 1)|0;
      $expanded19 = $expanded20 ^ -1;
      $35 = $33 & $expanded19;
      $36 = $35;
      $37 = HEAP32[$36>>2]|0;
      $arglist_next6 = ((($36)) + 4|0);
      HEAP32[$ap>>2] = $arglist_next6;
      $38 = (($20) + 1)|0;
      HEAP32[(624276)>>2] = $38;
      $39 = (624280 + ($20<<7)|0);
      HEAP32[$39>>2] = $37;
      $$2 = $$lcssa11;$52 = $38;
      break;
     }
     case 115:  {
      $arglist_current8 = HEAP32[$ap>>2]|0;
      $40 = $arglist_current8;
      $41 = ((0) + 4|0);
      $expanded24 = $41;
      $expanded23 = (($expanded24) - 1)|0;
      $42 = (($40) + ($expanded23))|0;
      $43 = ((0) + 4|0);
      $expanded28 = $43;
      $expanded27 = (($expanded28) - 1)|0;
      $expanded26 = $expanded27 ^ -1;
      $44 = $42 & $expanded26;
      $45 = $44;
      $46 = HEAP32[$45>>2]|0;
      $arglist_next9 = ((($45)) + 4|0);
      HEAP32[$ap>>2] = $arglist_next9;
      $47 = ($46|0)==(0|0);
      $$ = $47 ? 619383 : $46;
      $48 = (624280 + ($20<<7)|0);
      (_SDL_strlcpy($48,$$,128)|0);
      $49 = HEAP32[(624276)>>2]|0;
      $50 = (($49) + 1)|0;
      HEAP32[(624276)>>2] = $50;
      $$2 = $$lcssa11;$52 = $50;
      break;
     }
     default: {
      $$2 = $$lcssa11;$52 = $20;
     }
     }
     $51 = ($52|0)>(4);
     if ($51) {
      break L4;
     } else {
      $$01$be = $$2;$55 = $52;
     }
    } else {
     $$01$be = $3;$55 = $20;
    }
    $53 = HEAP8[$$01$be>>0]|0;
    $54 = ($53<<24>>24)==(0);
    if ($54) {
     break;
    } else {
     $$015 = $$01$be;$20 = $55;$5 = $53;
    }
   }
  }
 } while(0);
 (_SDL_GetError()|0);
 HEAP32[$vararg_buffer>>2] = 1278275;
 _SDL_LogDebug(1,595867,$vararg_buffer);
 STACKTOP = sp;return -1;
}
function _SDL_strlcpy($dst,$src,$maxlen) {
 $dst = $dst|0;
 $src = $src|0;
 $maxlen = $maxlen|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strlen($src)|0);
 $1 = ($maxlen|0)==(0);
 if ($1) {
  return ($0|0);
 }
 $2 = (($maxlen) + -1)|0;
 $3 = ($0>>>0)<($2>>>0);
 $4 = $3 ? $0 : $2;
 _memcpy(($dst|0),($src|0),($4|0))|0;
 $5 = (($dst) + ($4)|0);
 HEAP8[$5>>0] = 0;
 return ($0|0);
}
function _SDL_GetError() {
 var $$0$lcssa$i = 0, $$0$ph19$i = 0, $$013$i = 0, $$013$i$lcssa = 0, $$1$i = 0, $$lcssa85 = 0, $$lcssa86 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $7 = 0, $8 = 0, $9 = 0, $argi$0$ph22$i = 0, $argi$1$i = 0, $fmt$0$ph20$i = 0, $fmt$014$i = 0, $fmt$014$i$lcssa = 0, $fmt$1$i = 0, $fmt$1$i$lcssa = 0;
 var $fmt$1$i$lcssa80 = 0, $fmt$2$i = 0, $msg$0$lcssa$i = 0, $msg$0$ph21$i = 0, $msg$015$i = 0, $msg$015$i$lcssa = 0, $msg$1$i = 0, $msg$2$i = 0, $spot$0$i = 0, $spot$0$i$lcssa = 0, $spot$0$i$lcssa82 = 0, $spot$0$i84 = 0, $spot$1$i = 0, $tmp$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $tmp$i = sp + 32|0;
 HEAP8[1278275] = 0;
 $0 = HEAP32[156036]|0;
 $1 = ($0|0)==(0);
 if ($1) {
  STACKTOP = sp;return (1278275|0);
 }
 $2 = HEAP8[(624148)>>0]|0;
 $3 = ($2<<24>>24)==(0);
 L4: do {
  if ($3) {
   $$0$lcssa$i = 1023;$msg$0$lcssa$i = 1278275;
  } else {
   $4 = ((($tmp$i)) + 1|0);
   $5 = ((($tmp$i)) + 30|0);
   $$0$ph19$i = 1023;$65 = $2;$argi$0$ph22$i = 0;$fmt$0$ph20$i = (624148);$msg$0$ph21$i = 1278275;
   while(1) {
    $$013$i = $$0$ph19$i;$7 = $65;$fmt$014$i = $fmt$0$ph20$i;$msg$015$i = $msg$0$ph21$i;
    while(1) {
     $6 = ($7<<24>>24)==(37);
     if ($6) {
      $$013$i$lcssa = $$013$i;$fmt$014$i$lcssa = $fmt$014$i;$msg$015$i$lcssa = $msg$015$i;
      break;
     }
     $55 = ((($fmt$014$i)) + 1|0);
     $56 = ((($msg$015$i)) + 1|0);
     HEAP8[$msg$015$i>>0] = $7;
     $57 = (($$013$i) + -1)|0;
     $58 = HEAP8[$55>>0]|0;
     $59 = ($58<<24>>24)!=(0);
     $60 = ($$013$i|0)>(1);
     $61 = $60 & $59;
     if ($61) {
      $$013$i = $57;$7 = $58;$fmt$014$i = $55;$msg$015$i = $56;
     } else {
      $$0$lcssa$i = $57;$msg$0$lcssa$i = $56;
      break L4;
     }
    }
    $8 = ((($fmt$014$i$lcssa)) + 1|0);
    HEAP8[$tmp$i>>0] = 37;
    $fmt$1$i = $8;$spot$0$i = $4;
    while(1) {
     $9 = HEAP8[$fmt$1$i>>0]|0;
     $10 = ($9<<24>>24)==(46);
     if (!($10)) {
      $11 = ($9<<24>>24)>(47);
      if (!($11)) {
       $$lcssa85 = $9;$fmt$1$i$lcssa = $fmt$1$i;$spot$0$i$lcssa = $spot$0$i;
       label = 12;
       break;
      }
      $12 = ($9<<24>>24)<(58);
      if (!($12)) {
       $$lcssa86 = $9;$fmt$1$i$lcssa80 = $fmt$1$i;$spot$0$i$lcssa82 = $spot$0$i;
       label = 11;
       break;
      }
     }
     $13 = ($spot$0$i>>>0)<($5>>>0);
     $14 = ((($fmt$1$i)) + 1|0);
     $15 = ((($spot$0$i)) + 1|0);
     HEAP8[$spot$0$i>>0] = $9;
     if ($13) {
      $fmt$1$i = $14;$spot$0$i = $15;
     } else {
      $fmt$2$i = $14;$spot$0$i84 = $spot$0$i;$spot$1$i = $15;
      break;
     }
    }
    if ((label|0) == 11) {
     label = 0;
     $16 = ((($fmt$1$i$lcssa80)) + 1|0);
     $17 = ((($spot$0$i$lcssa82)) + 1|0);
     HEAP8[$spot$0$i$lcssa82>>0] = $$lcssa86;
     $fmt$2$i = $16;$spot$0$i84 = $spot$0$i$lcssa82;$spot$1$i = $17;
    }
    else if ((label|0) == 12) {
     label = 0;
     $18 = ((($fmt$1$i$lcssa)) + 1|0);
     $19 = ((($spot$0$i$lcssa)) + 1|0);
     HEAP8[$spot$0$i$lcssa>>0] = $$lcssa85;
     $fmt$2$i = $18;$spot$0$i84 = $spot$0$i$lcssa;$spot$1$i = $19;
    }
    HEAP8[$spot$1$i>>0] = 0;
    $20 = HEAP8[$spot$0$i84>>0]|0;
    $21 = $20 << 24 >> 24;
    switch ($21|0) {
    case 37:  {
     $22 = ((($msg$015$i$lcssa)) + 1|0);
     HEAP8[$msg$015$i$lcssa>>0] = 37;
     $23 = (($$013$i$lcssa) + -1)|0;
     $$1$i = $23;$argi$1$i = $argi$0$ph22$i;$msg$1$i = $22;
     break;
    }
    case 88: case 120: case 111: case 117: case 100: case 105: case 99:  {
     $24 = (($argi$0$ph22$i) + 1)|0;
     $25 = (624280 + ($argi$0$ph22$i<<7)|0);
     $26 = HEAP32[$25>>2]|0;
     HEAP32[$vararg_buffer>>2] = $26;
     $27 = (_SDL_snprintf($msg$015$i$lcssa,$$013$i$lcssa,$tmp$i,$vararg_buffer)|0);
     $28 = ($27|0)>(0);
     if ($28) {
      $29 = (($msg$015$i$lcssa) + ($27)|0);
      $30 = (($$013$i$lcssa) - ($27))|0;
      $$1$i = $30;$argi$1$i = $24;$msg$1$i = $29;
     } else {
      $$1$i = $$013$i$lcssa;$argi$1$i = $24;$msg$1$i = $msg$015$i$lcssa;
     }
     break;
    }
    case 102:  {
     $31 = (($argi$0$ph22$i) + 1)|0;
     $32 = (624280 + ($argi$0$ph22$i<<7)|0);
     $33 = +HEAPF64[$32>>3];
     HEAPF64[$vararg_buffer1>>3] = $33;
     $34 = (_SDL_snprintf($msg$015$i$lcssa,$$013$i$lcssa,$tmp$i,$vararg_buffer1)|0);
     $35 = ($34|0)>(0);
     if ($35) {
      $36 = (($msg$015$i$lcssa) + ($34)|0);
      $37 = (($$013$i$lcssa) - ($34))|0;
      $$1$i = $37;$argi$1$i = $31;$msg$1$i = $36;
     } else {
      $$1$i = $$013$i$lcssa;$argi$1$i = $31;$msg$1$i = $msg$015$i$lcssa;
     }
     break;
    }
    case 112:  {
     $38 = (($argi$0$ph22$i) + 1)|0;
     $39 = (624280 + ($argi$0$ph22$i<<7)|0);
     $40 = HEAP32[$39>>2]|0;
     HEAP32[$vararg_buffer4>>2] = $40;
     $41 = (_SDL_snprintf($msg$015$i$lcssa,$$013$i$lcssa,$tmp$i,$vararg_buffer4)|0);
     $42 = ($41|0)>(0);
     if ($42) {
      $43 = (($msg$015$i$lcssa) + ($41)|0);
      $44 = (($$013$i$lcssa) - ($41))|0;
      $$1$i = $44;$argi$1$i = $38;$msg$1$i = $43;
     } else {
      $$1$i = $$013$i$lcssa;$argi$1$i = $38;$msg$1$i = $msg$015$i$lcssa;
     }
     break;
    }
    case 115:  {
     $45 = (($argi$0$ph22$i) + 1)|0;
     $46 = (624280 + ($argi$0$ph22$i<<7)|0);
     HEAP32[$vararg_buffer7>>2] = $46;
     $47 = (_SDL_snprintf($msg$015$i$lcssa,$$013$i$lcssa,$tmp$i,$vararg_buffer7)|0);
     $48 = ($47|0)>(0);
     if ($48) {
      $49 = (($msg$015$i$lcssa) + ($47)|0);
      $50 = (($$013$i$lcssa) - ($47))|0;
      $$1$i = $50;$argi$1$i = $45;$msg$1$i = $49;
     } else {
      $$1$i = $$013$i$lcssa;$argi$1$i = $45;$msg$1$i = $msg$015$i$lcssa;
     }
     break;
    }
    default: {
     $$1$i = $$013$i$lcssa;$argi$1$i = $argi$0$ph22$i;$msg$1$i = $msg$015$i$lcssa;
    }
    }
    $51 = HEAP8[$fmt$2$i>>0]|0;
    $52 = ($51<<24>>24)!=(0);
    $53 = ($$1$i|0)>(0);
    $54 = $53 & $52;
    if ($54) {
     $$0$ph19$i = $$1$i;$65 = $51;$argi$0$ph22$i = $argi$1$i;$fmt$0$ph20$i = $fmt$2$i;$msg$0$ph21$i = $msg$1$i;
    } else {
     $$0$lcssa$i = $$1$i;$msg$0$lcssa$i = $msg$1$i;
     break L4;
    }
   }
  }
 } while(0);
 $62 = ($$0$lcssa$i|0)<(0);
 $63 = (($$0$lcssa$i) + -1)|0;
 $64 = (($msg$0$lcssa$i) + ($63)|0);
 $msg$2$i = $62 ? $64 : $msg$0$lcssa$i;
 HEAP8[$msg$2$i>>0] = 0;
 STACKTOP = sp;return (1278275|0);
}
function _SDL_snprintf($text,$maxlen,$fmt,$varargs) {
 $text = $text|0;
 $maxlen = $maxlen|0;
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $$fmt$i = 0, $0 = 0, $1 = 0, $ap = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ap = sp;
 HEAP32[$ap>>2] = $varargs;
 $0 = ($fmt|0)==(0|0);
 $$fmt$i = $0 ? 1282466 : $fmt;
 $1 = (_vsnprintf($text,$maxlen,$$fmt$i,$ap)|0);
 STACKTOP = sp;return ($1|0);
}
function _SDL_LogDebug($category,$fmt,$varargs) {
 $category = $category|0;
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $ap = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ap = sp;
 HEAP32[$ap>>2] = $varargs;
 _SDL_LogMessageV($category,2,$fmt,$ap);
 STACKTOP = sp;return;
}
function _SDL_LogMessageV($category,$priority,$fmt,$ap) {
 $category = $category|0;
 $priority = $priority|0;
 $fmt = $fmt|0;
 $ap = $ap|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $entry$0$i = 0, $entry$02$i = 0, $entry$03$i = 0, $entry$03$i$lcssa = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[31461]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  STACKTOP = sp;return;
 }
 $2 = ($priority|0)<(0);
 $3 = ($priority>>>0)>(6);
 $or$cond = $2 | $3;
 if ($or$cond) {
  STACKTOP = sp;return;
 }
 $entry$02$i = HEAP32[300442]|0;
 $4 = ($entry$02$i|0)==(0|0);
 L7: do {
  if ($4) {
   label = 7;
  } else {
   $entry$03$i = $entry$02$i;
   while(1) {
    $5 = HEAP32[$entry$03$i>>2]|0;
    $6 = ($5|0)==($category|0);
    if ($6) {
     $entry$03$i$lcssa = $entry$03$i;
     break;
    }
    $9 = ((($entry$03$i)) + 8|0);
    $entry$0$i = HEAP32[$9>>2]|0;
    $10 = ($entry$0$i|0)==(0|0);
    if ($10) {
     label = 7;
     break L7;
    } else {
     $entry$03$i = $entry$0$i;
    }
   }
   $7 = ((($entry$03$i$lcssa)) + 4|0);
   $8 = HEAP32[$7>>2]|0;
   $$0$i = $8;
  }
 } while(0);
 L12: do {
  if ((label|0) == 7) {
   switch ($category|0) {
   case 8:  {
    $$0$i = 1;
    break L12;
    break;
   }
   case 0:  {
    $11 = HEAP32[31469]|0;
    $$0$i = $11;
    break L12;
    break;
   }
   case 2:  {
    $12 = HEAP32[31470]|0;
    $$0$i = $12;
    break L12;
    break;
   }
   default: {
    $13 = HEAP32[31471]|0;
    $$0$i = $13;
    break L12;
   }
   }
  }
 } while(0);
 $14 = ($$0$i>>>0)>($priority>>>0);
 if ($14) {
  STACKTOP = sp;return;
 }
 $15 = STACKTOP; STACKTOP = STACKTOP + 4096|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 (_SDL_vsnprintf($15,4096,$fmt,$ap)|0);
 $16 = (_SDL_strlen($15)|0);
 $17 = ($16|0)==(0);
 if (!($17)) {
  $18 = (($16) + -1)|0;
  $19 = (($15) + ($18)|0);
  $20 = HEAP8[$19>>0]|0;
  $21 = ($20<<24>>24)==(10);
  if ($21) {
   HEAP8[$19>>0] = 0;
   $22 = ($18|0)==(0);
   if (!($22)) {
    $23 = (($16) + -2)|0;
    $24 = (($15) + ($23)|0);
    $25 = HEAP8[$24>>0]|0;
    $26 = ($25<<24>>24)==(13);
    if ($26) {
     HEAP8[$24>>0] = 0;
    }
   }
  }
 }
 $27 = HEAP32[31461]|0;
 $28 = HEAP32[300443]|0;
 FUNCTION_TABLE_viiii[$27 & 8191]($28,$category,$priority,$15);
 STACKTOP = sp;return;
}
function _SDL_LogOutput($userdata,$category,$priority,$message) {
 $userdata = $userdata|0;
 $category = $category|0;
 $priority = $priority|0;
 $message = $message|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP32[37544]|0;
 $1 = (125848 + ($priority<<2)|0);
 $2 = HEAP32[$1>>2]|0;
 HEAP32[$vararg_buffer>>2] = $2;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $message;
 (_fprintf($0,595909,$vararg_buffer)|0);
 STACKTOP = sp;return;
}
function _SDL_vsnprintf($text,$maxlen,$fmt,$ap) {
 $text = $text|0;
 $maxlen = $maxlen|0;
 $fmt = $fmt|0;
 $ap = $ap|0;
 var $$fmt = 0, $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($fmt|0)==(0|0);
 $$fmt = $0 ? 1282466 : $fmt;
 $1 = (_vsnprintf($text,$maxlen,$$fmt,$ap)|0);
 return ($1|0);
}
function _SDL_strlen($string) {
 $string = $string|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strlen($string)|0);
 return ($0|0);
}
function _SDL_GetWindowFlags($window) {
 $window = $window|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ((($window)) + 48|0);
   $7 = HEAP32[$6>>2]|0;
   $$0 = $7;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GL_SetAttribute($attr,$value) {
 $attr = $attr|0;
 $value = $value|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 L1: do {
  if ($1) {
   $2 = (_SDL_SetError(595917,$vararg_buffer)|0);
   $$0 = $2;
  } else {
   do {
    switch ($attr|0) {
    case 0:  {
     $3 = ((($0)) + 268|0);
     HEAP32[$3>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 1:  {
     $4 = ((($0)) + 272|0);
     HEAP32[$4>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 2:  {
     $5 = ((($0)) + 276|0);
     HEAP32[$5>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 3:  {
     $6 = ((($0)) + 280|0);
     HEAP32[$6>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 5:  {
     $7 = ((($0)) + 296|0);
     HEAP32[$7>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 4:  {
     $8 = ((($0)) + 288|0);
     HEAP32[$8>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 6:  {
     $9 = ((($0)) + 284|0);
     HEAP32[$9>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 7:  {
     $10 = ((($0)) + 292|0);
     HEAP32[$10>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 8:  {
     $11 = ((($0)) + 300|0);
     HEAP32[$11>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 9:  {
     $12 = ((($0)) + 304|0);
     HEAP32[$12>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 10:  {
     $13 = ((($0)) + 308|0);
     HEAP32[$13>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 11:  {
     $14 = ((($0)) + 312|0);
     HEAP32[$14>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 12:  {
     $15 = ((($0)) + 316|0);
     HEAP32[$15>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 13:  {
     $16 = ((($0)) + 320|0);
     HEAP32[$16>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 14:  {
     $17 = ((($0)) + 324|0);
     HEAP32[$17>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 15:  {
     $18 = ((($0)) + 328|0);
     HEAP32[$18>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 16:  {
     $19 = ((($0)) + 360|0);
     HEAP32[$19>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 17:  {
     $20 = ((($0)) + 332|0);
     HEAP32[$20>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 18:  {
     $21 = ((($0)) + 336|0);
     HEAP32[$21>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 19:  {
     $22 = ($value|0)==(0);
     $23 = ((($0)) + 344|0);
     if ($22) {
      HEAP32[$23>>2] = 0;
      $$0 = 0;
      break L1;
     } else {
      HEAP32[$23>>2] = 4;
      $$0 = 0;
      break L1;
     }
     break;
    }
    case 20:  {
     $24 = ($value>>>0)>(15);
     if ($24) {
      HEAP32[$vararg_buffer1>>2] = $value;
      $25 = (_SDL_SetError(596198,$vararg_buffer1)|0);
      $$0 = $25;
      break L1;
     } else {
      $26 = ((($0)) + 340|0);
      HEAP32[$26>>2] = $value;
      $$0 = 0;
      break L1;
     }
     break;
    }
    case 21:  {
     switch ($value|0) {
     case 0: case 1: case 2: case 4:  {
      $28 = ((($0)) + 344|0);
      HEAP32[$28>>2] = $value;
      $$0 = 0;
      break L1;
      break;
     }
     default: {
      HEAP32[$vararg_buffer3>>2] = $value;
      $27 = (_SDL_SetError(596229,$vararg_buffer3)|0);
      $$0 = $27;
      break L1;
     }
     }
     break;
    }
    case 22:  {
     $29 = ((($0)) + 348|0);
     HEAP32[$29>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 23:  {
     $30 = ((($0)) + 356|0);
     HEAP32[$30>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    case 24:  {
     $31 = ((($0)) + 352|0);
     HEAP32[$31>>2] = $value;
     $$0 = 0;
     break L1;
     break;
    }
    default: {
     $32 = (_SDL_SetError(595842,$vararg_buffer6)|0);
     $$0 = $32;
     break L1;
    }
    }
   } while(0);
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _SDL_RecreateWindow($window,$flags) {
 $window = $window|0;
 $flags = $flags|0;
 var $$0 = 0, $$0$i = 0, $$01 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre6 = 0, $$pre7 = 0, $$pre8 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $loaded_opengl$0 = 0, $or$cond = 0, $or$cond4 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer11 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, $vararg_buffer9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer11 = sp + 48|0;
 $vararg_buffer9 = sp + 40|0;
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = $flags & 2;
 $1 = ($0|0)==(0);
 $$pre = HEAP32[300441]|0;
 if (!($1)) {
  $2 = ((($$pre)) + 152|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ($3|0)==(0|0);
  if ($4) {
   $5 = (_SDL_SetError(596263,$vararg_buffer)|0);
   $$01 = $5;
   STACKTOP = sp;return ($$01|0);
  }
 }
 $6 = ((($window)) + 48|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 & 2048;
 $9 = ($8|0)==(0);
 $10 = $flags | 2048;
 $11 = $flags & -2049;
 $$0 = $9 ? $11 : $10;
 $12 = ($$pre|0)==(0|0);
 do {
  if ($12) {
   (_SDL_SetError(595917,$vararg_buffer1)|0);
  } else {
   $13 = HEAP32[$window>>2]|0;
   $14 = ((($$pre)) + 256|0);
   $15 = ($13|0)==($14|0);
   if (!($15)) {
    (_SDL_SetError(609414,$vararg_buffer3)|0);
    break;
   }
   $16 = $7 & 4;
   $17 = ($16|0)==(0);
   if (!($17)) {
    $18 = ((($window)) + 112|0);
    HEAP32[$18>>2] = 1;
    (_SDL_UpdateFullscreenMode($window,0)|0);
    $19 = HEAP32[300441]|0;
    $20 = ((($19)) + 64|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = ($21|0)==(0|0);
    if (!($22)) {
     FUNCTION_TABLE_vii[$21 & 8191]($19,$window);
    }
    HEAP32[$18>>2] = 0;
    (_SDL_SendWindowEvent($window,2,0,0)|0);
   }
  }
 } while(0);
 $23 = ((($window)) + 104|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ($24|0)==(0|0);
 if (!($25)) {
  $26 = HEAP32[$24>>2]|0;
  $27 = $26 & -5;
  HEAP32[$24>>2] = $27;
  _SDL_FreeSurface($24);
  HEAP32[$23>>2] = 0;
 }
 $28 = HEAP32[300441]|0;
 $29 = ((($28)) + 116|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ($30|0)==(0|0);
 if ($31) {
  $33 = $28;
 } else {
  FUNCTION_TABLE_vii[$30 & 8191]($28,$window);
  $$pre6 = HEAP32[300441]|0;
  $33 = $$pre6;
 }
 $32 = ((($33)) + 104|0);
 $34 = HEAP32[$32>>2]|0;
 $35 = ($34|0)!=(0|0);
 $36 = $$0 & 2048;
 $37 = ($36|0)==(0);
 $or$cond = $37 & $35;
 if ($or$cond) {
  FUNCTION_TABLE_vii[$34 & 8191]($33,$window);
 }
 $38 = HEAP32[$6>>2]|0;
 $39 = $38 & 2;
 $40 = $$0 & 2;
 $41 = ($39|0)==($40|0);
 L26: do {
  if ($41) {
   $loaded_opengl$0 = 0;
  } else {
   $42 = ($40|0)==(0);
   $43 = HEAP32[300441]|0;
   $44 = ($43|0)==(0|0);
   if ($42) {
    if ($44) {
     (_SDL_SetError(595917,$vararg_buffer9)|0);
     $loaded_opengl$0 = 0;
     break;
    }
    $64 = ((($43)) + 364|0);
    $65 = HEAP32[$64>>2]|0;
    $66 = ($65|0)>(0);
    if (!($66)) {
     $loaded_opengl$0 = 0;
     break;
    }
    $67 = (($65) + -1)|0;
    HEAP32[$64>>2] = $67;
    $68 = ($65|0)>(1);
    if ($68) {
     $loaded_opengl$0 = 0;
     break;
    }
    $69 = ((($43)) + 148|0);
    $70 = HEAP32[$69>>2]|0;
    $71 = ($70|0)==(0|0);
    if ($71) {
     $loaded_opengl$0 = 0;
     break;
    }
    FUNCTION_TABLE_vi[$70 & 8191]($43);
    $loaded_opengl$0 = 0;
    break;
   }
   L37: do {
    if ($44) {
     $45 = (_SDL_SetError(595917,$vararg_buffer5)|0);
     $$0$i = $45;
    } else {
     $46 = ((($43)) + 364|0);
     $47 = HEAP32[$46>>2]|0;
     $48 = ($47|0)==(0);
     do {
      if ($48) {
       $49 = ((($43)) + 140|0);
       $50 = HEAP32[$49>>2]|0;
       $51 = ($50|0)==(0|0);
       if ($51) {
        $52 = (_SDL_SetError(595987,$vararg_buffer7)|0);
        $$0$i = $52;
        break L37;
       }
       $53 = (FUNCTION_TABLE_iii[$50 & 8191]($43,0)|0);
       $54 = ($53|0)==(0);
       $55 = HEAP32[300441]|0;
       if ($54) {
        $$phi$trans$insert = ((($55)) + 364|0);
        $$pre7 = HEAP32[$$phi$trans$insert>>2]|0;
        $57 = $55;$59 = $$pre7;
        break;
       }
       $60 = ((($55)) + 148|0);
       $61 = HEAP32[$60>>2]|0;
       $62 = ($61|0)==(0|0);
       if ($62) {
        $$0$i = $53;
        break L37;
       }
       FUNCTION_TABLE_vi[$61 & 8191]($55);
       $$0$i = $53;
       break L37;
      } else {
       $57 = $43;$59 = $47;
      }
     } while(0);
     $56 = ((($57)) + 364|0);
     $58 = (($59) + 1)|0;
     HEAP32[$56>>2] = $58;
     $loaded_opengl$0 = 1;
     break L26;
    }
   } while(0);
   $63 = ($$0$i|0)<(0);
   if ($63) {
    $$01 = -1;
    STACKTOP = sp;return ($$01|0);
   } else {
    $loaded_opengl$0 = 1;
   }
  }
 } while(0);
 $72 = $$0 & 8242;
 $73 = $72 | 8;
 HEAP32[$6>>2] = $73;
 $74 = ((($window)) + 52|0);
 HEAP32[$74>>2] = $73;
 $75 = ((($window)) + 116|0);
 HEAP32[$75>>2] = 0;
 $76 = HEAP32[300441]|0;
 $77 = ((($76)) + 28|0);
 $78 = HEAP32[$77>>2]|0;
 $79 = ($78|0)!=(0|0);
 $or$cond4 = $37 & $79;
 if ($or$cond4) {
  $80 = (FUNCTION_TABLE_iii[$78 & 8191]($76,$window)|0);
  $81 = ($80|0)<(0);
  if ($81) {
   $82 = ($loaded_opengl$0|0)==(0);
   if ($82) {
    $$01 = -1;
    STACKTOP = sp;return ($$01|0);
   }
   $83 = HEAP32[300441]|0;
   $84 = ($83|0)==(0|0);
   if ($84) {
    (_SDL_SetError(595917,$vararg_buffer11)|0);
   } else {
    $85 = ((($83)) + 364|0);
    $86 = HEAP32[$85>>2]|0;
    $87 = ($86|0)>(0);
    if ($87) {
     $88 = (($86) + -1)|0;
     HEAP32[$85>>2] = $88;
     $89 = ($86|0)>(1);
     if (!($89)) {
      $90 = ((($83)) + 148|0);
      $91 = HEAP32[$90>>2]|0;
      $92 = ($91|0)==(0|0);
      if (!($92)) {
       FUNCTION_TABLE_vi[$91 & 8191]($83);
      }
     }
    }
   }
   $93 = HEAP32[$6>>2]|0;
   $94 = $93 & -3;
   HEAP32[$6>>2] = $94;
   $$01 = -1;
   STACKTOP = sp;return ($$01|0);
  }
 }
 if (!($37)) {
  $95 = HEAP32[$6>>2]|0;
  $96 = $95 | 2048;
  HEAP32[$6>>2] = $96;
 }
 $97 = HEAP32[300441]|0;
 $98 = ((($97)) + 36|0);
 $99 = HEAP32[$98>>2]|0;
 $100 = ($99|0)==(0|0);
 if ($100) {
  $105 = $97;
 } else {
  $101 = ((($window)) + 8|0);
  $102 = HEAP32[$101>>2]|0;
  $103 = ($102|0)==(0|0);
  if ($103) {
   $105 = $97;
  } else {
   FUNCTION_TABLE_vii[$99 & 8191]($97,$window);
   $$pre8 = HEAP32[300441]|0;
   $105 = $$pre8;
  }
 }
 $104 = ((($105)) + 40|0);
 $106 = HEAP32[$104>>2]|0;
 $107 = ($106|0)==(0|0);
 if (!($107)) {
  $108 = ((($window)) + 12|0);
  $109 = HEAP32[$108>>2]|0;
  $110 = ($109|0)==(0|0);
  if (!($110)) {
   FUNCTION_TABLE_viii[$106 & 8191]($105,$window,$109);
  }
 }
 $111 = ((($window)) + 124|0);
 $112 = HEAP32[$111>>2]|0;
 $113 = ($112|0)==(0|0);
 if (!($113)) {
  $114 = HEAP32[300441]|0;
  $115 = ((($114)) + 232|0);
  $116 = HEAP32[$115>>2]|0;
  (FUNCTION_TABLE_iii[$116 & 8191]($window,1)|0);
 }
 _SDL_FinishWindowCreation($window,$$0);
 $$01 = 0;
 STACKTOP = sp;return ($$01|0);
}
function _SDL_UpdateFullscreenMode($window,$fullscreen) {
 $window = $window|0;
 $fullscreen = $fullscreen|0;
 var $$0$i = 0, $$0$i12 = 0, $$0$i12$us = 0, $$11 = 0, $$25$ph = 0, $$5 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $current_mode$i = 0, $current_mode$i19 = 0, $display_mode$i = 0;
 var $display_mode$i18 = 0, $fullscreen_mode = 0, $not$ = 0, $or$cond = 0, $or$cond6 = 0, $other$0 = 0, $other$0$lcssa31 = 0, $other$0$us = 0, $other$032 = 0, $other$033 = 0, $other$033$us = 0, $resized$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, $vararg_buffer9 = 0, $x$i = 0, $x$i15 = 0;
 var $y$i = 0, $y$i16 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 160|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer9 = sp + 40|0;
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $display_mode$i18 = sp + 140|0;
 $current_mode$i19 = sp + 120|0;
 $x$i15 = sp + 116|0;
 $y$i16 = sp + 112|0;
 $display_mode$i = sp + 92|0;
 $current_mode$i = sp + 72|0;
 $x$i = sp + 68|0;
 $y$i = sp + 64|0;
 $fullscreen_mode = sp + 44|0;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$5 = -1;
  STACKTOP = sp;return ($$5|0);
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ((($window)) + 112|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)!=(0);
   $9 = ($fullscreen|0)!=(0);
   $or$cond = $9 & $8;
   if ($or$cond) {
    $$5 = 0;
    STACKTOP = sp;return ($$5|0);
   }
   $10 = (_SDL_GetWindowDisplayIndex($window)|0);
   $11 = ($10|0)>(-1);
   if ($11) {
    $12 = HEAP32[300441]|0;
    $13 = ((($12)) + 244|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = (($14) + (($10*68)|0)|0);
    $$0$i = $15;
   } else {
    $$0$i = 0;
   }
   $16 = ((($$0$i)) + 56|0);
   do {
    if ($9) {
     $17 = HEAP32[$16>>2]|0;
     $18 = ($17|0)==(0|0);
     $19 = ($17|0)==($window|0);
     $or$cond6 = $18 | $19;
     if (!($or$cond6)) {
      $20 = HEAP32[300441]|0;
      $21 = ($20|0)==(0|0);
      if ($21) {
       (_SDL_SetError(595917,$vararg_buffer3)|0);
       break;
      }
      $22 = HEAP32[$17>>2]|0;
      $23 = ((($20)) + 256|0);
      $24 = ($22|0)==($23|0);
      if (!($24)) {
       (_SDL_SetError(609414,$vararg_buffer5)|0);
       break;
      }
      $25 = ((($17)) + 48|0);
      $26 = HEAP32[$25>>2]|0;
      $27 = $26 & 64;
      $28 = ($27|0)==(0);
      if ($28) {
       (_SDL_UpdateFullscreenMode($17,0)|0);
       $29 = HEAP32[300441]|0;
       $30 = ((($29)) + 76|0);
       $31 = HEAP32[$30>>2]|0;
       $32 = ($31|0)==(0|0);
       if (!($32)) {
        FUNCTION_TABLE_vii[$31 & 8191]($29,$17);
       }
      }
     }
    }
   } while(0);
   $33 = HEAP32[$16>>2]|0;
   $34 = ($33|0)==($window|0);
   $35 = $34&1;
   $36 = ($35|0)==($fullscreen|0);
   if ($36) {
    $37 = ((($window)) + 52|0);
    $38 = HEAP32[$37>>2]|0;
    $39 = ((($window)) + 48|0);
    $40 = HEAP32[$39>>2]|0;
    $41 = $40 ^ $38;
    $42 = $41 & 4097;
    $43 = ($42|0)==(0);
    if ($43) {
     $$5 = 0;
     STACKTOP = sp;return ($$5|0);
    }
   }
   $44 = HEAP32[300441]|0;
   $45 = ((($44)) + 248|0);
   $other$032 = HEAP32[$45>>2]|0;
   $46 = ($other$032|0)==(0|0);
   L30: do {
    if (!($46)) {
     $47 = ($fullscreen|0)==(0);
     L32: do {
      if ($47) {
       $other$033$us = $other$032;
       while(1) {
        $48 = ($other$033$us|0)==($window|0);
        if (!($48)) {
         $49 = ((($other$033$us)) + 48|0);
         $50 = HEAP32[$49>>2]|0;
         $51 = $50 & 69;
         $52 = ($51|0)==(5);
         if ($52) {
          $53 = (_SDL_GetWindowDisplayIndex($other$033$us)|0);
          $54 = ($53|0)>(-1);
          if ($54) {
           $55 = HEAP32[300441]|0;
           $56 = ((($55)) + 244|0);
           $57 = HEAP32[$56>>2]|0;
           $58 = (($57) + (($53*68)|0)|0);
           $$0$i12$us = $58;
          } else {
           $$0$i12$us = 0;
          }
          $59 = ($$0$i12$us|0)==($$0$i|0);
          if ($59) {
           (_SDL_memset($fullscreen_mode,0,20)|0);
           $60 = (_SDL_GetWindowDisplayMode($other$033$us,$fullscreen_mode)|0);
           $61 = ($60|0)==(0);
           if ($61) {
            $other$0$lcssa31 = $other$033$us;
            break L32;
           }
          }
         }
        }
        $62 = ((($other$033$us)) + 144|0);
        $other$0$us = HEAP32[$62>>2]|0;
        $63 = ($other$0$us|0)==(0|0);
        if ($63) {
         break L30;
        } else {
         $other$033$us = $other$0$us;
        }
       }
      } else {
       $other$033 = $other$032;
       while(1) {
        $64 = ($other$033|0)==($window|0);
        if ($64) {
         label = 35;
        } else {
         $65 = ((($other$033)) + 48|0);
         $66 = HEAP32[$65>>2]|0;
         $67 = $66 & 69;
         $68 = ($67|0)==(5);
         if ($68) {
          $69 = (_SDL_GetWindowDisplayIndex($other$033)|0);
          $70 = ($69|0)>(-1);
          if ($70) {
           $71 = HEAP32[300441]|0;
           $72 = ((($71)) + 244|0);
           $73 = HEAP32[$72>>2]|0;
           $74 = (($73) + (($69*68)|0)|0);
           $$0$i12 = $74;
          } else {
           $$0$i12 = 0;
          }
          $75 = ($$0$i12|0)==($$0$i|0);
          if ($75) {
           label = 35;
          }
         }
        }
        if ((label|0) == 35) {
         label = 0;
         (_SDL_memset($fullscreen_mode,0,20)|0);
         $76 = (_SDL_GetWindowDisplayMode($other$033,$fullscreen_mode)|0);
         $77 = ($76|0)==(0);
         if ($77) {
          $other$0$lcssa31 = $other$033;
          break L32;
         }
        }
        $123 = ((($other$033)) + 144|0);
        $other$0 = HEAP32[$123>>2]|0;
        $124 = ($other$0|0)==(0|0);
        if ($124) {
         break L30;
        } else {
         $other$033 = $other$0;
        }
       }
      }
     } while(0);
     $78 = ((($other$0$lcssa31)) + 24|0);
     $79 = HEAP32[$78>>2]|0;
     $80 = ((($fullscreen_mode)) + 4|0);
     $81 = HEAP32[$80>>2]|0;
     $82 = ($79|0)==($81|0);
     if ($82) {
      $83 = ((($other$0$lcssa31)) + 28|0);
      $84 = HEAP32[$83>>2]|0;
      $85 = ((($fullscreen_mode)) + 8|0);
      $86 = HEAP32[$85>>2]|0;
      $not$ = ($84|0)!=($86|0);
      $$11 = $not$&1;
      $resized$0 = $$11;
     } else {
      $resized$0 = 1;
     }
     $87 = ((($window)) + 48|0);
     $88 = HEAP32[$87>>2]|0;
     $89 = $88 & 4097;
     $90 = ($89|0)==(4097);
     L60: do {
      if ($90) {
       $93 = ((($$0$i)) + 16|0);
       ;HEAP32[$display_mode$i>>2]=HEAP32[$93>>2]|0;HEAP32[$display_mode$i+4>>2]=HEAP32[$93+4>>2]|0;HEAP32[$display_mode$i+8>>2]=HEAP32[$93+8>>2]|0;HEAP32[$display_mode$i+12>>2]=HEAP32[$93+12>>2]|0;HEAP32[$display_mode$i+16>>2]=HEAP32[$93+16>>2]|0;
       $94 = ((($$0$i)) + 36|0);
       ;HEAP32[$current_mode$i>>2]=HEAP32[$94>>2]|0;HEAP32[$current_mode$i+4>>2]=HEAP32[$94+4>>2]|0;HEAP32[$current_mode$i+8>>2]=HEAP32[$94+8>>2]|0;HEAP32[$current_mode$i+12>>2]=HEAP32[$94+12>>2]|0;HEAP32[$current_mode$i+16>>2]=HEAP32[$94+16>>2]|0;
       $95 = (_SDL_memcmp($display_mode$i,$current_mode$i,20)|0);
       $96 = ($95|0)==(0);
       do {
        if (!($96)) {
         $97 = HEAP32[300441]|0;
         $98 = ((($97)) + 24|0);
         $99 = HEAP32[$98>>2]|0;
         $100 = ($99|0)==(0|0);
         if ($100) {
          $103 = (_SDL_SetError(596500,$vararg_buffer7)|0);
          $104 = ($103|0)<(0);
          if ($104) {
           $$25$ph = -1;
           break L60;
          } else {
           label = 47;
           break L60;
          }
         }
         $101 = (FUNCTION_TABLE_iiii[$99 & 8191]($97,$$0$i,$display_mode$i)|0);
         $102 = ($101|0)<(0);
         if ($102) {
          $$25$ph = -1;
          break L60;
         } else {
          ;HEAP32[$94>>2]=HEAP32[$display_mode$i>>2]|0;HEAP32[$94+4>>2]=HEAP32[$display_mode$i+4>>2]|0;HEAP32[$94+8>>2]=HEAP32[$display_mode$i+8>>2]|0;HEAP32[$94+12>>2]=HEAP32[$display_mode$i+12>>2]|0;HEAP32[$94+16>>2]=HEAP32[$display_mode$i+16>>2]|0;
          break;
         }
        }
       } while(0);
       label = 47;
      } else {
       $91 = (_SDL_SetDisplayModeForDisplay($$0$i,$fullscreen_mode)|0);
       $92 = ($91|0)<(0);
       if ($92) {
        $$25$ph = -1;
       } else {
        label = 47;
       }
      }
     } while(0);
     if ((label|0) == 47) {
      $105 = HEAP32[300441]|0;
      $106 = ((($105)) + 88|0);
      $107 = HEAP32[$106>>2]|0;
      $108 = ($107|0)==(0|0);
      if (!($108)) {
       FUNCTION_TABLE_viiii[$107 & 8191]($105,$other$0$lcssa31,$$0$i,1);
      }
      HEAP32[$16>>2] = $other$0$lcssa31;
      $109 = ($resized$0|0)==(0);
      if ($109) {
       $113 = ((($other$0$lcssa31)) + 108|0);
       HEAP32[$113>>2] = 0;
       $114 = HEAP32[$78>>2]|0;
       $115 = ((($other$0$lcssa31)) + 28|0);
       $116 = HEAP32[$115>>2]|0;
       (_SDL_SendWindowEvent($other$0$lcssa31,6,$114,$116)|0);
      } else {
       $110 = HEAP32[$80>>2]|0;
       $111 = ((($fullscreen_mode)) + 8|0);
       $112 = HEAP32[$111>>2]|0;
       (_SDL_SendWindowEvent($other$0$lcssa31,5,$110,$112)|0);
      }
      $117 = (_SDL_GetMouseFocus()|0);
      $118 = ($117|0)==($other$0$lcssa31|0);
      if ($118) {
       (_SDL_GetMouseState($x$i15,$y$i16)|0);
       $119 = HEAP32[$x$i15>>2]|0;
       $120 = HEAP32[$y$i16>>2]|0;
       _SDL_WarpMouseInWindow($other$0$lcssa31,$119,$120);
      }
      $121 = HEAP32[$87>>2]|0;
      $122 = ((($window)) + 52|0);
      HEAP32[$122>>2] = $121;
      $$25$ph = 0;
     }
     $$5 = $$25$ph;
     STACKTOP = sp;return ($$5|0);
    }
   } while(0);
   $125 = ((($$0$i)) + 16|0);
   ;HEAP32[$display_mode$i18>>2]=HEAP32[$125>>2]|0;HEAP32[$display_mode$i18+4>>2]=HEAP32[$125+4>>2]|0;HEAP32[$display_mode$i18+8>>2]=HEAP32[$125+8>>2]|0;HEAP32[$display_mode$i18+12>>2]=HEAP32[$125+12>>2]|0;HEAP32[$display_mode$i18+16>>2]=HEAP32[$125+16>>2]|0;
   $126 = ((($$0$i)) + 36|0);
   ;HEAP32[$current_mode$i19>>2]=HEAP32[$126>>2]|0;HEAP32[$current_mode$i19+4>>2]=HEAP32[$126+4>>2]|0;HEAP32[$current_mode$i19+8>>2]=HEAP32[$126+8>>2]|0;HEAP32[$current_mode$i19+12>>2]=HEAP32[$126+12>>2]|0;HEAP32[$current_mode$i19+16>>2]=HEAP32[$126+16>>2]|0;
   $127 = (_SDL_memcmp($display_mode$i18,$current_mode$i19,20)|0);
   $128 = ($127|0)==(0);
   do {
    if (!($128)) {
     $129 = HEAP32[300441]|0;
     $130 = ((($129)) + 24|0);
     $131 = HEAP32[$130>>2]|0;
     $132 = ($131|0)==(0|0);
     if ($132) {
      (_SDL_SetError(596500,$vararg_buffer9)|0);
      break;
     }
     $133 = (FUNCTION_TABLE_iiii[$131 & 8191]($129,$$0$i,$display_mode$i18)|0);
     $134 = ($133|0)<(0);
     if (!($134)) {
      ;HEAP32[$126>>2]=HEAP32[$display_mode$i18>>2]|0;HEAP32[$126+4>>2]=HEAP32[$display_mode$i18+4>>2]|0;HEAP32[$126+8>>2]=HEAP32[$display_mode$i18+8>>2]|0;HEAP32[$126+12>>2]=HEAP32[$display_mode$i18+12>>2]|0;HEAP32[$126+16>>2]=HEAP32[$display_mode$i18+16>>2]|0;
     }
    }
   } while(0);
   $135 = HEAP32[300441]|0;
   $136 = ((($135)) + 88|0);
   $137 = HEAP32[$136>>2]|0;
   $138 = ($137|0)==(0|0);
   if (!($138)) {
    FUNCTION_TABLE_viiii[$137 & 8191]($135,$window,$$0$i,0);
   }
   HEAP32[$16>>2] = 0;
   $139 = ((($window)) + 108|0);
   HEAP32[$139>>2] = 0;
   $140 = ((($window)) + 24|0);
   $141 = HEAP32[$140>>2]|0;
   $142 = ((($window)) + 28|0);
   $143 = HEAP32[$142>>2]|0;
   (_SDL_SendWindowEvent($window,6,$141,$143)|0);
   $144 = (_SDL_GetMouseFocus()|0);
   $145 = ($144|0)==($window|0);
   if ($145) {
    (_SDL_GetMouseState($x$i,$y$i)|0);
    $146 = HEAP32[$x$i>>2]|0;
    $147 = HEAP32[$y$i>>2]|0;
    _SDL_WarpMouseInWindow($window,$146,$147);
   }
   $148 = ((($window)) + 48|0);
   $149 = HEAP32[$148>>2]|0;
   $150 = ((($window)) + 52|0);
   HEAP32[$150>>2] = $149;
   $$5 = 0;
   STACKTOP = sp;return ($$5|0);
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 $$5 = -1;
 STACKTOP = sp;return ($$5|0);
}
function _SDL_GetWindowDisplayIndex($window) {
 $window = $window|0;
 var $$ = 0, $$1 = 0, $$2 = 0, $$closest_dist$0 = 0, $$neg5 = 0, $$neg6 = 0, $$neg8 = 0, $$neg9 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $center = 0, $closest$0$lcssa23 = 0, $closest$013 = 0, $closest_dist$014 = 0, $i$016 = 0, $i$1$closest$0 = 0, $i$1$closest$0$lcssa = 0, $i$115 = 0, $rect = 0, $switch$split2D = 0;
 var $switch$split52D = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $center = sp + 40|0;
 $rect = sp + 24|0;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$2 = -1;
  STACKTOP = sp;return ($$2|0);
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ((($window)) + 16|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = $7 & -65536;
   $switch$split2D = ($8|0)<(805240832);
   if ($switch$split2D) {
    switch ($8|0) {
    case 536805376:  {
     label = 7;
     break;
    }
    default: {
    }
    }
   } else {
    switch ($8|0) {
    case 805240832:  {
     label = 7;
     break;
    }
    default: {
    }
    }
   }
   if ((label|0) == 7) {
    $9 = $7 & 65535;
    $10 = ((($0)) + 240|0);
    $11 = HEAP32[$10>>2]|0;
    $12 = ($9|0)>=($11|0);
    $$ = $12 ? 0 : $9;
    $$2 = $$;
    STACKTOP = sp;return ($$2|0);
   }
   $13 = ((($window)) + 20|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = $14 & -65536;
   $switch$split52D = ($15|0)<(805240832);
   if ($switch$split52D) {
    switch ($15|0) {
    case 536805376:  {
     label = 11;
     break;
    }
    default: {
    }
    }
   } else {
    switch ($15|0) {
    case 805240832:  {
     label = 11;
     break;
    }
    default: {
    }
    }
   }
   if ((label|0) == 11) {
    $19 = $14 & 65535;
    $20 = ((($0)) + 240|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = ($19|0)>=($21|0);
    $$1 = $22 ? 0 : $19;
    $$2 = $$1;
    STACKTOP = sp;return ($$2|0);
   }
   $16 = ((($0)) + 240|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ($17|0)>(0);
   L22: do {
    if ($18) {
     $$phi$trans$insert = ((($0)) + 244|0);
     $$pre = HEAP32[$$phi$trans$insert>>2]|0;
     $i$016 = 0;
     while(1) {
      $25 = (((($$pre) + (($i$016*68)|0)|0)) + 56|0);
      $26 = HEAP32[$25>>2]|0;
      $27 = ($26|0)==($window|0);
      $24 = (($i$016) + 1)|0;
      if ($27) {
       $$2 = $i$016;
       break;
      }
      $23 = ($24|0)<($17|0);
      if ($23) {
       $i$016 = $24;
      } else {
       break L22;
      }
     }
     STACKTOP = sp;return ($$2|0);
    }
   } while(0);
   $28 = ((($window)) + 24|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = (($29|0) / 2)&-1;
   $31 = (($30) + ($7))|0;
   HEAP32[$center>>2] = $31;
   $32 = ((($window)) + 28|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = (($33|0) / 2)&-1;
   $35 = (($34) + ($14))|0;
   $36 = ((($center)) + 4|0);
   HEAP32[$36>>2] = $35;
   if ($18) {
    $37 = ((($rect)) + 8|0);
    $38 = ((($rect)) + 4|0);
    $39 = ((($rect)) + 12|0);
    $closest$013 = -1;$closest_dist$014 = 2147483647;$i$115 = 0;
    while(1) {
     (_SDL_GetDisplayBounds($i$115,$rect)|0);
     $40 = (_SDL_EnclosePoints($center,1,$rect,0)|0);
     $41 = ($40|0)==(0);
     if (!($41)) {
      $$2 = $i$115;
      label = 20;
      break;
     }
     $42 = HEAP32[$center>>2]|0;
     $43 = HEAP32[$rect>>2]|0;
     $44 = HEAP32[$37>>2]|0;
     $$neg5 = (($44|0) / -2)&-1;
     $$neg6 = (($42) - ($43))|0;
     $45 = (($$neg6) + ($$neg5))|0;
     $46 = HEAP32[$36>>2]|0;
     $47 = HEAP32[$38>>2]|0;
     $48 = HEAP32[$39>>2]|0;
     $$neg8 = (($48|0) / -2)&-1;
     $$neg9 = (($46) - ($47))|0;
     $49 = (($$neg9) + ($$neg8))|0;
     $50 = Math_imul($45, $45)|0;
     $51 = Math_imul($49, $49)|0;
     $52 = (($51) + ($50))|0;
     $53 = ($52|0)<($closest_dist$014|0);
     $i$1$closest$0 = $53 ? $i$115 : $closest$013;
     $$closest_dist$0 = $53 ? $52 : $closest_dist$014;
     $54 = (($i$115) + 1)|0;
     $55 = HEAP32[300441]|0;
     $56 = ((($55)) + 240|0);
     $57 = HEAP32[$56>>2]|0;
     $58 = ($54|0)<($57|0);
     if ($58) {
      $closest$013 = $i$1$closest$0;$closest_dist$014 = $$closest_dist$0;$i$115 = $54;
     } else {
      $i$1$closest$0$lcssa = $i$1$closest$0;
      break;
     }
    }
    if ((label|0) == 20) {
     STACKTOP = sp;return ($$2|0);
    }
    $59 = ($i$1$closest$0$lcssa|0)<(0);
    if ($59) {
     $closest$0$lcssa23 = $i$1$closest$0$lcssa;
    } else {
     $$2 = $i$1$closest$0$lcssa;
     STACKTOP = sp;return ($$2|0);
    }
   } else {
    $closest$0$lcssa23 = -1;
   }
   (_SDL_SetError(596351,$vararg_buffer3)|0);
   $$2 = $closest$0$lcssa23;
   STACKTOP = sp;return ($$2|0);
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 $$2 = -1;
 STACKTOP = sp;return ($$2|0);
}
function _SDL_GetDisplayBounds($displayIndex,$rect) {
 $displayIndex = $displayIndex|0;
 $rect = $rect|0;
 var $$1 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$phi3Z2D = 0, $$pre2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$1 = -1;
  STACKTOP = sp;return ($$1|0);
 }
 $2 = ($displayIndex|0)>(-1);
 $$phi$trans$insert = ((($0)) + 240|0);
 $$pre = HEAP32[$$phi$trans$insert>>2]|0;
 $3 = ($$pre|0)>($displayIndex|0);
 $or$cond = $2 & $3;
 if (!($or$cond)) {
  $4 = (($$pre) + -1)|0;
  HEAP32[$vararg_buffer1>>2] = $4;
  (_SDL_SetError(596297,$vararg_buffer1)|0);
  $$1 = -1;
  STACKTOP = sp;return ($$1|0);
 }
 $5 = ($rect|0)==(0|0);
 if ($5) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $6 = ((($0)) + 244|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($0)) + 12|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)==(0|0);
 if (!($10)) {
  $11 = (($7) + (($displayIndex*68)|0)|0);
  $12 = (FUNCTION_TABLE_iiii[$9 & 8191]($0,$11,$rect)|0);
  $13 = ($12|0)==(0);
  if ($13) {
   $$1 = 0;
   STACKTOP = sp;return ($$1|0);
  }
 }
 $14 = ($displayIndex|0)==(0);
 if ($14) {
  HEAP32[$rect>>2] = 0;
  $15 = ((($rect)) + 4|0);
  HEAP32[$15>>2] = 0;
  $$pre2 = ((($rect)) + 8|0);
  $$pre$phi3Z2D = $$pre2;
 } else {
  $16 = (($displayIndex) + -1)|0;
  (_SDL_GetDisplayBounds($16,$rect)|0);
  $17 = ((($rect)) + 8|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = HEAP32[$rect>>2]|0;
  $20 = (($19) + ($18))|0;
  HEAP32[$rect>>2] = $20;
  $$pre$phi3Z2D = $17;
 }
 $21 = (((($7) + (($displayIndex*68)|0)|0)) + 40|0);
 $22 = HEAP32[$21>>2]|0;
 HEAP32[$$pre$phi3Z2D>>2] = $22;
 $23 = (((($7) + (($displayIndex*68)|0)|0)) + 44|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($rect)) + 12|0);
 HEAP32[$25>>2] = $24;
 $$1 = 0;
 STACKTOP = sp;return ($$1|0);
}
function _SDL_EnclosePoints($points,$count,$clip,$result) {
 $points = $points|0;
 $count = $count|0;
 $clip = $clip|0;
 $result = $result|0;
 var $$1 = 0, $$demorgan = 0, $$maxx$0 = 0, $$maxx$4 = 0, $$maxy$0 = 0, $$maxy$3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $added$021 = 0, $added$1 = 0, $added$1$lcssa = 0, $exitcond = 0, $exitcond32 = 0, $i$020 = 0, $i$020$us = 0, $i$112 = 0, $maxx$022 = 0, $maxx$1 = 0, $maxx$2 = 0, $maxx$2$lcssa = 0, $maxx$413 = 0, $maxx$5 = 0, $maxx$6 = 0, $maxx$636 = 0, $maxy$019 = 0;
 var $maxy$1 = 0, $maxy$1$lcssa = 0, $maxy$311 = 0, $maxy$4 = 0, $maxy$5 = 0, $maxy$535 = 0, $minx$024 = 0, $minx$1 = 0, $minx$2 = 0, $minx$2$lcssa = 0, $minx$415 = 0, $minx$5 = 0, $minx$6 = 0, $minx$638 = 0, $miny$023 = 0, $miny$1 = 0, $miny$1$lcssa = 0, $miny$314 = 0, $miny$4 = 0, $miny$5 = 0;
 var $miny$537 = 0, $not$ = 0, $notlhs = 0, $notlhs$us = 0, $notrhs = 0, $notrhs$us = 0, $or$cond$not = 0, $or$cond$not$us = 0, $or$cond2 = 0, $or$cond2$not = 0, $or$cond2$not$us = 0, $or$cond2$us = 0, $or$cond3 = 0, $or$cond3$us = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($points|0)==(0|0);
 if ($0) {
  HEAP32[$vararg_buffer>>2] = 596338;
  (_SDL_SetError(611267,$vararg_buffer)|0);
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $1 = ($count|0)<(1);
 if ($1) {
  HEAP32[$vararg_buffer1>>2] = 596345;
  (_SDL_SetError(611267,$vararg_buffer1)|0);
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $2 = ($clip|0)==(0|0);
 if ($2) {
  $35 = ($result|0)==(0|0);
  if ($35) {
   $$1 = 1;
   STACKTOP = sp;return ($$1|0);
  }
  $36 = HEAP32[$points>>2]|0;
  $37 = ((($points)) + 4|0);
  $38 = HEAP32[$37>>2]|0;
  $39 = ($count|0)>(1);
  if ($39) {
   $i$112 = 1;$maxx$413 = $36;$maxy$311 = $38;$minx$415 = $36;$miny$314 = $38;
   while(1) {
    $40 = (($points) + ($i$112<<3)|0);
    $41 = HEAP32[$40>>2]|0;
    $42 = (((($points) + ($i$112<<3)|0)) + 4|0);
    $43 = HEAP32[$42>>2]|0;
    $44 = ($41|0)<($minx$415|0);
    $45 = ($41|0)>($maxx$413|0);
    $$maxx$4 = $45 ? $41 : $maxx$413;
    $maxx$5 = $44 ? $maxx$413 : $$maxx$4;
    $minx$5 = $44 ? $41 : $minx$415;
    $46 = ($43|0)<($miny$314|0);
    $47 = ($43|0)>($maxy$311|0);
    $$maxy$3 = $47 ? $43 : $maxy$311;
    $maxy$4 = $46 ? $maxy$311 : $$maxy$3;
    $miny$4 = $46 ? $43 : $miny$314;
    $48 = (($i$112) + 1)|0;
    $exitcond = ($48|0)==($count|0);
    if ($exitcond) {
     $maxx$6 = $maxx$5;$maxy$5 = $maxy$4;$minx$6 = $minx$5;$miny$5 = $miny$4;
     label = 19;
     break;
    } else {
     $i$112 = $48;$maxx$413 = $maxx$5;$maxy$311 = $maxy$4;$minx$415 = $minx$5;$miny$314 = $miny$4;
    }
   }
  } else {
   $maxx$636 = $36;$maxy$535 = $38;$minx$638 = $36;$miny$537 = $38;
  }
 } else {
  $3 = HEAP32[$clip>>2]|0;
  $4 = ((($clip)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ((($clip)) + 8|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (($7) + ($3))|0;
  $9 = ((($clip)) + 12|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = (($10) + ($5))|0;
  $12 = ($7|0)<(1);
  $13 = ($10|0)<(1);
  $$demorgan = $12 | $13;
  if ($$demorgan) {
   $$1 = 0;
   STACKTOP = sp;return ($$1|0);
  }
  $14 = ($result|0)==(0|0);
  if ($14) {
   $i$020$us = 0;
   while(1) {
    $15 = (($points) + ($i$020$us<<3)|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = (((($points) + ($i$020$us<<3)|0)) + 4|0);
    $18 = HEAP32[$17>>2]|0;
    $notlhs$us = ($16|0)<($3|0);
    $notrhs$us = ($16|0)>=($8|0);
    $or$cond$not$us = $notrhs$us | $notlhs$us;
    $19 = ($18|0)<($5|0);
    $or$cond2$us = $or$cond$not$us | $19;
    $or$cond2$not$us = $or$cond2$us ^ 1;
    $20 = ($18|0)<($11|0);
    $or$cond3$us = $20 & $or$cond2$not$us;
    if ($or$cond3$us) {
     $$1 = 1;
     label = 21;
     break;
    }
    $21 = (($i$020$us) + 1)|0;
    $22 = ($21|0)<($count|0);
    if ($22) {
     $i$020$us = $21;
    } else {
     $$1 = 0;
     label = 21;
     break;
    }
   }
   if ((label|0) == 21) {
    STACKTOP = sp;return ($$1|0);
   }
  } else {
   $added$021 = 0;$i$020 = 0;$maxx$022 = 0;$maxy$019 = 0;$minx$024 = 0;$miny$023 = 0;
  }
  while(1) {
   $23 = (($points) + ($i$020<<3)|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = (((($points) + ($i$020<<3)|0)) + 4|0);
   $26 = HEAP32[$25>>2]|0;
   $notlhs = ($24|0)<($3|0);
   $notrhs = ($24|0)>=($8|0);
   $or$cond$not = $notrhs | $notlhs;
   $27 = ($26|0)<($5|0);
   $or$cond2 = $or$cond$not | $27;
   $or$cond2$not = $or$cond2 ^ 1;
   $28 = ($26|0)<($11|0);
   $or$cond3 = $28 & $or$cond2$not;
   if ($or$cond3) {
    $29 = ($added$021|0)==(0);
    if ($29) {
     $added$1 = 1;$maxx$2 = $24;$maxy$1 = $26;$minx$2 = $24;$miny$1 = $26;
    } else {
     $30 = ($24|0)<($minx$024|0);
     $31 = ($24|0)>($maxx$022|0);
     $$maxx$0 = $31 ? $24 : $maxx$022;
     $maxx$1 = $30 ? $maxx$022 : $$maxx$0;
     $minx$1 = $30 ? $24 : $minx$024;
     $32 = ($26|0)<($miny$023|0);
     if ($32) {
      $added$1 = $added$021;$maxx$2 = $maxx$1;$maxy$1 = $maxy$019;$minx$2 = $minx$1;$miny$1 = $26;
     } else {
      $33 = ($26|0)>($maxy$019|0);
      $$maxy$0 = $33 ? $26 : $maxy$019;
      $added$1 = $added$021;$maxx$2 = $maxx$1;$maxy$1 = $$maxy$0;$minx$2 = $minx$1;$miny$1 = $miny$023;
     }
    }
   } else {
    $added$1 = $added$021;$maxx$2 = $maxx$022;$maxy$1 = $maxy$019;$minx$2 = $minx$024;$miny$1 = $miny$023;
   }
   $34 = (($i$020) + 1)|0;
   $exitcond32 = ($34|0)==($count|0);
   if ($exitcond32) {
    $added$1$lcssa = $added$1;$maxx$2$lcssa = $maxx$2;$maxy$1$lcssa = $maxy$1;$minx$2$lcssa = $minx$2;$miny$1$lcssa = $miny$1;
    break;
   } else {
    $added$021 = $added$1;$i$020 = $34;$maxx$022 = $maxx$2;$maxy$019 = $maxy$1;$minx$024 = $minx$2;$miny$023 = $miny$1;
   }
  }
  $not$ = ($added$1$lcssa|0)==(0);
  if ($not$) {
   $$1 = 0;
   STACKTOP = sp;return ($$1|0);
  } else {
   $maxx$6 = $maxx$2$lcssa;$maxy$5 = $maxy$1$lcssa;$minx$6 = $minx$2$lcssa;$miny$5 = $miny$1$lcssa;
   label = 19;
  }
 }
 if ((label|0) == 19) {
  $49 = ($result|0)==(0|0);
  if ($49) {
   $$1 = 1;
   STACKTOP = sp;return ($$1|0);
  } else {
   $maxx$636 = $maxx$6;$maxy$535 = $maxy$5;$minx$638 = $minx$6;$miny$537 = $miny$5;
  }
 }
 HEAP32[$result>>2] = $minx$638;
 $50 = ((($result)) + 4|0);
 HEAP32[$50>>2] = $miny$537;
 $51 = (($maxx$636) + 1)|0;
 $52 = (($51) - ($minx$638))|0;
 $53 = ((($result)) + 8|0);
 HEAP32[$53>>2] = $52;
 $54 = (($maxy$535) + 1)|0;
 $55 = (($54) - ($miny$537))|0;
 $56 = ((($result)) + 12|0);
 HEAP32[$56>>2] = $55;
 $$1 = 1;
 STACKTOP = sp;return ($$1|0);
}
function _SDL_memset($dst,$c,$len) {
 $dst = $dst|0;
 $c = $c|0;
 $len = $len|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $c&255;
 _memset(($dst|0),($0|0),($len|0))|0;
 return ($dst|0);
}
function _SDL_GetWindowDisplayMode($window,$mode) {
 $window = $window|0;
 $mode = $mode|0;
 var $$0 = 0, $$0$i = 0, $$0$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $fullscreen_mode = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $fullscreen_mode = sp + 28|0;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ($mode|0)==(0|0);
   if ($6) {
    HEAP32[$vararg_buffer3>>2] = 596378;
    $7 = (_SDL_SetError(611267,$vararg_buffer3)|0);
    $$0 = $7;
    STACKTOP = sp;return ($$0|0);
   }
   $8 = ((($window)) + 72|0);
   ;HEAP32[$fullscreen_mode>>2]=HEAP32[$8>>2]|0;HEAP32[$fullscreen_mode+4>>2]=HEAP32[$8+4>>2]|0;HEAP32[$fullscreen_mode+8>>2]=HEAP32[$8+8>>2]|0;HEAP32[$fullscreen_mode+12>>2]=HEAP32[$8+12>>2]|0;HEAP32[$fullscreen_mode+16>>2]=HEAP32[$8+16>>2]|0;
   $9 = ((($fullscreen_mode)) + 4|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($10|0)==(0);
   if ($11) {
    $12 = ((($window)) + 64|0);
    $13 = HEAP32[$12>>2]|0;
    HEAP32[$9>>2] = $13;
   }
   $14 = ((($fullscreen_mode)) + 8|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)==(0);
   if ($16) {
    $17 = ((($window)) + 68|0);
    $18 = HEAP32[$17>>2]|0;
    HEAP32[$14>>2] = $18;
   }
   $19 = (_SDL_GetWindowDisplayIndex($window)|0);
   $20 = ($19|0)>(-1);
   if ($20) {
    $21 = HEAP32[300441]|0;
    $22 = ((($21)) + 244|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = (($23) + (($19*68)|0)|0);
    $$0$i = $24;
   } else {
    $$0$i = 0;
   }
   $25 = ((($window)) + 48|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = $26 & 4097;
   $28 = ($27|0)==(4097);
   if ($28) {
    $29 = ((($$0$i)) + 16|0);
    ;HEAP32[$fullscreen_mode>>2]=HEAP32[$29>>2]|0;HEAP32[$fullscreen_mode+4>>2]=HEAP32[$29+4>>2]|0;HEAP32[$fullscreen_mode+8>>2]=HEAP32[$29+8>>2]|0;HEAP32[$fullscreen_mode+12>>2]=HEAP32[$29+12>>2]|0;HEAP32[$fullscreen_mode+16>>2]=HEAP32[$29+16>>2]|0;
   } else {
    $30 = (_SDL_GetWindowDisplayIndex($window)|0);
    $31 = ($30|0)>(-1);
    if ($31) {
     $32 = HEAP32[300441]|0;
     $33 = ((($32)) + 244|0);
     $34 = HEAP32[$33>>2]|0;
     $35 = (($34) + (($30*68)|0)|0);
     $$0$i1 = $35;
    } else {
     $$0$i1 = 0;
    }
    $36 = (_SDL_GetClosestDisplayModeForDisplay($$0$i1,$fullscreen_mode,$fullscreen_mode)|0);
    $37 = ($36|0)==(0|0);
    if ($37) {
     $38 = (_SDL_SetError(596430,$vararg_buffer5)|0);
     $$0 = $38;
     STACKTOP = sp;return ($$0|0);
    }
   }
   ;HEAP32[$mode>>2]=HEAP32[$fullscreen_mode>>2]|0;HEAP32[$mode+4>>2]=HEAP32[$fullscreen_mode+4>>2]|0;HEAP32[$mode+8>>2]=HEAP32[$fullscreen_mode+8>>2]|0;HEAP32[$mode+12>>2]=HEAP32[$fullscreen_mode+12>>2]|0;HEAP32[$mode+16>>2]=HEAP32[$fullscreen_mode+16>>2]|0;
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GetClosestDisplayModeForDisplay($display,$mode,$closest) {
 $display = $display|0;
 $mode = $mode|0;
 $closest = $closest|0;
 var $$0 = 0, $$pre = 0, $$pre$i = 0, $$sink3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $i$0 = 0, $match$0 = 0, $match$0$ = 0, $match$0$lcssa = 0, $match$1 = 0, $or$cond = 0, $or$cond2 = 0, $target_format$0 = 0, $target_refresh_rate$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($mode|0)!=(0|0);
 $1 = ($closest|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  (_SDL_SetError(596383,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = HEAP32[$mode>>2]|0;
 $3 = ($2|0)==(0);
 if ($3) {
  $4 = ((($display)) + 16|0);
  $5 = HEAP32[$4>>2]|0;
  $target_format$0 = $5;
 } else {
  $target_format$0 = $2;
 }
 $6 = ((($mode)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0);
 if ($8) {
  $9 = ((($display)) + 28|0);
  $10 = HEAP32[$9>>2]|0;
  $target_refresh_rate$0 = $10;
 } else {
  $target_refresh_rate$0 = $7;
 }
 $11 = ((($mode)) + 4|0);
 $12 = ((($display)) + 8|0);
 $13 = ((($display)) + 12|0);
 $14 = $target_format$0 >>> 8;
 $15 = $14 & 255;
 $16 = ((($mode)) + 8|0);
 $$pre = HEAP32[$12>>2]|0;
 $18 = $$pre;$i$0 = 0;$match$0 = 0;
 while(1) {
  $17 = ($18|0)==(0);
  if ($17) {
   $19 = HEAP32[300441]|0;
   $20 = ((($19)) + 20|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = ($21|0)==(0|0);
   if ($22) {
    $26 = 0;
   } else {
    FUNCTION_TABLE_vii[$21 & 8191]($19,$display);
    $23 = HEAP32[$13>>2]|0;
    $24 = HEAP32[$12>>2]|0;
    _SDL_qsort($23,$24,20,4476);
    $$pre$i = HEAP32[$12>>2]|0;
    $26 = $$pre$i;
   }
  } else {
   $26 = $18;
  }
  $25 = ($i$0|0)<($26|0);
  if (!($25)) {
   $match$0$lcssa = $match$0;
   break;
  }
  $27 = HEAP32[$13>>2]|0;
  $28 = (($27) + (($i$0*20)|0)|0);
  $29 = (((($27) + (($i$0*20)|0)|0)) + 4|0);
  $30 = HEAP32[$29>>2]|0;
  $31 = ($30|0)==(0);
  if (!($31)) {
   $32 = HEAP32[$11>>2]|0;
   $33 = ($30|0)<($32|0);
   if ($33) {
    $match$0$lcssa = $match$0;
    break;
   }
  }
  $34 = (((($27) + (($i$0*20)|0)|0)) + 8|0);
  $35 = HEAP32[$34>>2]|0;
  $36 = ($35|0)==(0);
  if ($36) {
   label = 18;
  } else {
   $37 = HEAP32[$16>>2]|0;
   $38 = ($35|0)<($37|0);
   if ($38) {
    if ($31) {
     $match$1 = $match$0;
    } else {
     $39 = HEAP32[$11>>2]|0;
     $40 = ($30|0)==($39|0);
     if ($40) {
      $match$0$lcssa = $match$0;
      break;
     } else {
      $match$1 = $match$0;
     }
    }
   } else {
    label = 18;
   }
  }
  do {
   if ((label|0) == 18) {
    label = 0;
    $41 = ($match$0|0)==(0|0);
    if ($41) {
     $match$1 = $28;
    } else {
     $42 = ((($match$0)) + 4|0);
     $43 = HEAP32[$42>>2]|0;
     $44 = ($30|0)<($43|0);
     if ($44) {
      $match$1 = $28;
     } else {
      $45 = ((($match$0)) + 8|0);
      $46 = HEAP32[$45>>2]|0;
      $47 = ($35|0)<($46|0);
      if ($47) {
       $match$1 = $28;
      } else {
       $48 = HEAP32[$28>>2]|0;
       $49 = HEAP32[$match$0>>2]|0;
       $50 = ($48|0)==($49|0);
       if ($50) {
        $58 = (((($27) + (($i$0*20)|0)|0)) + 12|0);
        $59 = HEAP32[$58>>2]|0;
        $60 = ((($match$0)) + 12|0);
        $61 = HEAP32[$60>>2]|0;
        $62 = ($59|0)==($61|0);
        $63 = ($59|0)<($target_refresh_rate$0|0);
        $or$cond2 = $63 | $62;
        $match$0$ = $or$cond2 ? $match$0 : $28;
        $match$1 = $match$0$;
        break;
       }
       $51 = ($48|0)==($target_format$0|0);
       if (!($51)) {
        $52 = $48 >>> 8;
        $53 = $52 & 255;
        $54 = ($53>>>0)<($15>>>0);
        if ($54) {
         $match$1 = $match$0;
         break;
        }
        $55 = $48 ^ $target_format$0;
        $56 = $55 & 251658240;
        $57 = ($56|0)==(0);
        if (!($57)) {
         $match$1 = $match$0;
         break;
        }
       }
       $match$1 = $28;
      }
     }
    }
   }
  } while(0);
  $64 = (($i$0) + 1)|0;
  $18 = $26;$i$0 = $64;$match$0 = $match$1;
 }
 $65 = ($match$0$lcssa|0)==(0|0);
 if ($65) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $66 = HEAP32[$match$0$lcssa>>2]|0;
 $67 = ($66|0)==(0);
 if ($67) {
  $68 = HEAP32[$mode>>2]|0;
  $69 = $68;
 } else {
  $69 = $66;
 }
 HEAP32[$closest>>2] = $69;
 $70 = ((($match$0$lcssa)) + 4|0);
 $71 = HEAP32[$70>>2]|0;
 $72 = ($71|0)==(0);
 if ($72) {
  label = 34;
 } else {
  $73 = ((($match$0$lcssa)) + 8|0);
  $74 = HEAP32[$73>>2]|0;
  $75 = ($74|0)==(0);
  if ($75) {
   label = 34;
  } else {
   $76 = ((($closest)) + 4|0);
   HEAP32[$76>>2] = $71;
   $77 = ((($closest)) + 8|0);
   HEAP32[$77>>2] = $74;
   $92 = $71;$95 = $74;
  }
 }
 if ((label|0) == 34) {
  $78 = HEAP32[$11>>2]|0;
  $79 = ((($closest)) + 4|0);
  HEAP32[$79>>2] = $78;
  $80 = HEAP32[$16>>2]|0;
  $81 = ((($closest)) + 8|0);
  HEAP32[$81>>2] = $80;
  $92 = $78;$95 = $80;
 }
 $82 = ((($match$0$lcssa)) + 12|0);
 $83 = HEAP32[$82>>2]|0;
 $84 = ($83|0)==(0);
 if ($84) {
  $85 = HEAP32[$6>>2]|0;
  $$sink3 = $85;
 } else {
  $$sink3 = $83;
 }
 $86 = ((($closest)) + 12|0);
 HEAP32[$86>>2] = $$sink3;
 $87 = ((($match$0$lcssa)) + 16|0);
 $88 = HEAP32[$87>>2]|0;
 $89 = ((($closest)) + 16|0);
 HEAP32[$89>>2] = $88;
 $90 = ($69|0)==(0);
 if ($90) {
  HEAP32[$closest>>2] = 370546692;
 }
 $91 = ($92|0)==(0);
 if ($91) {
  $93 = ((($closest)) + 4|0);
  HEAP32[$93>>2] = 640;
 }
 $94 = ($95|0)==(0);
 if (!($94)) {
  $$0 = $closest;
  STACKTOP = sp;return ($$0|0);
 }
 $96 = ((($closest)) + 8|0);
 HEAP32[$96>>2] = 480;
 $$0 = $closest;
 STACKTOP = sp;return ($$0|0);
}
function _cmpmodes($A,$B) {
 $A = $A|0;
 $B = $B|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($A|0)==($B|0);
 if ($0) {
  $$0 = 0;
  return ($$0|0);
 }
 $1 = ((($A)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($B)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==($2|0);
 if (!($5)) {
  $6 = (($4) - ($2))|0;
  $$0 = $6;
  return ($$0|0);
 }
 $7 = ((($A)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($B)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10|0)==($8|0);
 if (!($11)) {
  $12 = (($10) - ($8))|0;
  $$0 = $12;
  return ($$0|0);
 }
 $13 = HEAP32[$A>>2]|0;
 $14 = $13 >>> 8;
 $15 = $14 & 255;
 $16 = HEAP32[$B>>2]|0;
 $17 = $16 >>> 8;
 $18 = $17 & 255;
 $19 = ($18|0)==($15|0);
 if (!($19)) {
  $20 = (($18) - ($15))|0;
  $$0 = $20;
  return ($$0|0);
 }
 $21 = $13 >>> 16;
 $22 = $21 & 15;
 $23 = $16 >>> 16;
 $24 = $23 & 15;
 $25 = ($24|0)==($22|0);
 if ($25) {
  $27 = ((($A)) + 12|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = ((($B)) + 12|0);
  $30 = HEAP32[$29>>2]|0;
  $31 = ($30|0)==($28|0);
  $32 = (($30) - ($28))|0;
  $$ = $31 ? 0 : $32;
  return ($$|0);
 }
 $26 = (($24) - ($22))|0;
 $$0 = $26;
 return ($$0|0);
}
function _SDL_qsort($base,$nmemb,$size,$compare) {
 $base = $base|0;
 $nmemb = $nmemb|0;
 $size = $size|0;
 $compare = $compare|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _qsort($base,$nmemb,$size,$compare);
 return;
}
function _SDL_SetDisplayModeForDisplay($display,$mode) {
 $display = $display|0;
 $mode = $mode|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $current_mode = 0, $display_mode = 0, $vararg_buffer = 0;
 var $vararg_buffer2 = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $display_mode = sp + 32|0;
 $current_mode = sp + 12|0;
 $0 = ($mode|0)==(0|0);
 if ($0) {
  $25 = ((($display)) + 16|0);
  ;HEAP32[$display_mode>>2]=HEAP32[$25>>2]|0;HEAP32[$display_mode+4>>2]=HEAP32[$25+4>>2]|0;HEAP32[$display_mode+8>>2]=HEAP32[$25+8>>2]|0;HEAP32[$display_mode+12>>2]=HEAP32[$25+12>>2]|0;HEAP32[$display_mode+16>>2]=HEAP32[$25+16>>2]|0;
 } else {
  ;HEAP32[$display_mode>>2]=HEAP32[$mode>>2]|0;HEAP32[$display_mode+4>>2]=HEAP32[$mode+4>>2]|0;HEAP32[$display_mode+8>>2]=HEAP32[$mode+8>>2]|0;HEAP32[$display_mode+12>>2]=HEAP32[$mode+12>>2]|0;HEAP32[$display_mode+16>>2]=HEAP32[$mode+16>>2]|0;
  $1 = HEAP32[$display_mode>>2]|0;
  $2 = ($1|0)==(0);
  if ($2) {
   $3 = ((($display)) + 36|0);
   $4 = HEAP32[$3>>2]|0;
   HEAP32[$display_mode>>2] = $4;
  }
  $5 = ((($display_mode)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6|0)==(0);
  if ($7) {
   $8 = ((($display)) + 40|0);
   $9 = HEAP32[$8>>2]|0;
   HEAP32[$5>>2] = $9;
  }
  $10 = ((($display_mode)) + 8|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($11|0)==(0);
  if ($12) {
   $13 = ((($display)) + 44|0);
   $14 = HEAP32[$13>>2]|0;
   HEAP32[$10>>2] = $14;
  }
  $15 = ((($display_mode)) + 12|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ($16|0)==(0);
  if ($17) {
   $18 = ((($display)) + 48|0);
   $19 = HEAP32[$18>>2]|0;
   HEAP32[$15>>2] = $19;
  }
  $20 = (_SDL_GetClosestDisplayModeForDisplay($display,$display_mode,$display_mode)|0);
  $21 = ($20|0)==(0|0);
  if ($21) {
   $22 = HEAP32[$5>>2]|0;
   $23 = HEAP32[$10>>2]|0;
   HEAP32[$vararg_buffer>>2] = $22;
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
   HEAP32[$vararg_ptr1>>2] = $23;
   $24 = (_SDL_SetError(596463,$vararg_buffer)|0);
   $$0 = $24;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $26 = ((($display)) + 36|0);
 ;HEAP32[$current_mode>>2]=HEAP32[$26>>2]|0;HEAP32[$current_mode+4>>2]=HEAP32[$26+4>>2]|0;HEAP32[$current_mode+8>>2]=HEAP32[$26+8>>2]|0;HEAP32[$current_mode+12>>2]=HEAP32[$26+12>>2]|0;HEAP32[$current_mode+16>>2]=HEAP32[$26+16>>2]|0;
 $27 = (_SDL_memcmp($display_mode,$current_mode,20)|0);
 $28 = ($27|0)==(0);
 if ($28) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $29 = HEAP32[300441]|0;
 $30 = ((($29)) + 24|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = ($31|0)==(0|0);
 if ($32) {
  $33 = (_SDL_SetError(596500,$vararg_buffer2)|0);
  $$0 = $33;
  STACKTOP = sp;return ($$0|0);
 }
 $34 = (FUNCTION_TABLE_iiii[$31 & 8191]($29,$display,$display_mode)|0);
 $35 = ($34|0)<(0);
 if ($35) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 ;HEAP32[$26>>2]=HEAP32[$display_mode>>2]|0;HEAP32[$26+4>>2]=HEAP32[$display_mode+4>>2]|0;HEAP32[$26+8>>2]=HEAP32[$display_mode+8>>2]|0;HEAP32[$26+12>>2]=HEAP32[$display_mode+12>>2]|0;HEAP32[$26+16>>2]=HEAP32[$display_mode+16>>2]|0;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_memcmp($s1,$s2,$len) {
 $s1 = $s1|0;
 $s2 = $s2|0;
 $len = $len|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_memcmp($s1,$s2,$len)|0);
 return ($0|0);
}
function _SDL_SendWindowEvent($window,$windowevent,$data1,$data2) {
 $window = $window|0;
 $windowevent = $windowevent|0;
 $data1 = $data1|0;
 $data2 = $data2|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $event = 0, $or$cond = 0;
 var $posted$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $event = sp;
 $0 = ($window|0)==(0|0);
 if ($0) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = $windowevent&255;
 L4: do {
  switch ($1|0) {
  case 1:  {
   $2 = ((($window)) + 48|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = $3 & 4;
   $5 = ($4|0)==(0);
   if ($5) {
    $6 = $3 & -13;
    $7 = $6 | 4;
    HEAP32[$2>>2] = $7;
    _SDL_OnWindowShown($window);
    break L4;
   } else {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   break;
  }
  case 2:  {
   $8 = ((($window)) + 48|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = $9 & 4;
   $11 = ($10|0)==(0);
   if ($11) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   } else {
    $12 = $9 & -13;
    $13 = $12 | 8;
    HEAP32[$8>>2] = $13;
    _SDL_OnWindowHidden($window);
    break L4;
   }
   break;
  }
  case 4:  {
   $14 = $data1 & -65536;
   $15 = ($14|0)==(536805376);
   $16 = $data2 & -65536;
   $17 = ($16|0)==(536805376);
   $or$cond = $15 | $17;
   if ($or$cond) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   $18 = ((($window)) + 48|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = $19 & 1;
   $21 = ($20|0)==(0);
   if ($21) {
    $22 = ((($window)) + 56|0);
    HEAP32[$22>>2] = $data1;
    $23 = ((($window)) + 60|0);
    HEAP32[$23>>2] = $data2;
   }
   $24 = ((($window)) + 16|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = ($25|0)==($data1|0);
   $27 = ((($window)) + 20|0);
   if ($26) {
    $28 = HEAP32[$27>>2]|0;
    $29 = ($28|0)==($data2|0);
    if ($29) {
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
   }
   HEAP32[$24>>2] = $data1;
   HEAP32[$27>>2] = $data2;
   break;
  }
  case 5:  {
   $30 = ((($window)) + 48|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = $31 & 1;
   $33 = ($32|0)==(0);
   if ($33) {
    $34 = ((($window)) + 64|0);
    HEAP32[$34>>2] = $data1;
    $35 = ((($window)) + 68|0);
    HEAP32[$35>>2] = $data2;
   }
   $36 = ((($window)) + 24|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = ($37|0)==($data1|0);
   $39 = ((($window)) + 28|0);
   if ($38) {
    $40 = HEAP32[$39>>2]|0;
    $41 = ($40|0)==($data2|0);
    if ($41) {
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
   }
   HEAP32[$36>>2] = $data1;
   HEAP32[$39>>2] = $data2;
   _SDL_OnWindowResized($window);
   break;
  }
  case 7:  {
   $42 = ((($window)) + 48|0);
   $43 = HEAP32[$42>>2]|0;
   $44 = $43 & 64;
   $45 = ($44|0)==(0);
   if ($45) {
    $46 = $43 & -193;
    $47 = $46 | 64;
    HEAP32[$42>>2] = $47;
    _SDL_OnWindowMinimized($window);
    break L4;
   } else {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   break;
  }
  case 8:  {
   $48 = ((($window)) + 48|0);
   $49 = HEAP32[$48>>2]|0;
   $50 = $49 & 128;
   $51 = ($50|0)==(0);
   if ($51) {
    $52 = $49 & -193;
    $53 = $52 | 128;
    HEAP32[$48>>2] = $53;
    break L4;
   } else {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   break;
  }
  case 9:  {
   $54 = ((($window)) + 48|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = $55 & 192;
   $57 = ($56|0)==(0);
   if ($57) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   } else {
    $58 = $55 & -193;
    HEAP32[$54>>2] = $58;
    _SDL_OnWindowRestored($window);
    break L4;
   }
   break;
  }
  case 10:  {
   $59 = ((($window)) + 48|0);
   $60 = HEAP32[$59>>2]|0;
   $61 = $60 & 1024;
   $62 = ($61|0)==(0);
   if ($62) {
    $63 = $60 | 1024;
    HEAP32[$59>>2] = $63;
    _SDL_OnWindowEnter($window);
    break L4;
   } else {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   break;
  }
  case 11:  {
   $64 = ((($window)) + 48|0);
   $65 = HEAP32[$64>>2]|0;
   $66 = $65 & 1024;
   $67 = ($66|0)==(0);
   if ($67) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   } else {
    $68 = $65 & -1025;
    HEAP32[$64>>2] = $68;
    break L4;
   }
   break;
  }
  case 12:  {
   $69 = ((($window)) + 48|0);
   $70 = HEAP32[$69>>2]|0;
   $71 = $70 & 512;
   $72 = ($71|0)==(0);
   if ($72) {
    $73 = $70 | 512;
    HEAP32[$69>>2] = $73;
    _SDL_OnWindowFocusGained($window);
    break L4;
   } else {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   break;
  }
  case 13:  {
   $74 = ((($window)) + 48|0);
   $75 = HEAP32[$74>>2]|0;
   $76 = $75 & 512;
   $77 = ($76|0)==(0);
   if ($77) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   } else {
    $78 = $75 & -513;
    HEAP32[$74>>2] = $78;
    _SDL_OnWindowFocusLost($window);
    break L4;
   }
   break;
  }
  default: {
  }
  }
 } while(0);
 $79 = (_SDL_EventState(512,-1)|0);
 $80 = ($79<<24>>24)==(1);
 if ($80) {
  HEAP32[$event>>2] = 512;
  $81 = ((($event)) + 12|0);
  HEAP8[$81>>0] = $windowevent;
  $82 = ((($event)) + 16|0);
  HEAP32[$82>>2] = $data1;
  $83 = ((($event)) + 20|0);
  HEAP32[$83>>2] = $data2;
  $84 = ((($window)) + 4|0);
  $85 = HEAP32[$84>>2]|0;
  $86 = ((($event)) + 8|0);
  HEAP32[$86>>2] = $85;
  switch ($windowevent<<24>>24) {
  case 5:  {
   _SDL_FilterEvents(4477,$event);
   break;
  }
  case 6:  {
   _SDL_FilterEvents(4478,$event);
   break;
  }
  case 4:  {
   _SDL_FilterEvents(4479,$event);
   break;
  }
  default: {
  }
  }
  $87 = (_SDL_PushEvent($event)|0);
  $88 = ($87|0)>(0);
  $89 = $88&1;
  $posted$0 = $89;
 } else {
  $posted$0 = 0;
 }
 $90 = ($windowevent<<24>>24)==(14);
 if (!($90)) {
  $$0 = $posted$0;
  STACKTOP = sp;return ($$0|0);
 }
 $91 = ((($window)) + 140|0);
 $92 = HEAP32[$91>>2]|0;
 $93 = ($92|0)==(0|0);
 if (!($93)) {
  $$0 = $posted$0;
  STACKTOP = sp;return ($$0|0);
 }
 $94 = ((($window)) + 144|0);
 $95 = HEAP32[$94>>2]|0;
 $96 = ($95|0)==(0|0);
 if (!($96)) {
  $$0 = $posted$0;
  STACKTOP = sp;return ($$0|0);
 }
 (_SDL_SendQuit()|0);
 $$0 = $posted$0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_OnWindowShown($window) {
 $window = $window|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($window)) + 48|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 69;
 $3 = ($2|0)==(5);
 if (!($3)) {
  return;
 }
 (_SDL_UpdateFullscreenMode($window,1)|0);
 return;
}
function _SDL_OnWindowHidden($window) {
 $window = $window|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 (_SDL_UpdateFullscreenMode($window,0)|0);
 return;
}
function _SDL_OnWindowResized($window) {
 $window = $window|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($window)) + 108|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($window)) + 24|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($window)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 (_SDL_SendWindowEvent($window,6,$2,$4)|0);
 return;
}
function _SDL_OnWindowMinimized($window) {
 $window = $window|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 (_SDL_UpdateFullscreenMode($window,0)|0);
 return;
}
function _SDL_OnWindowRestored($window) {
 $window = $window|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($window)) + 48|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 69;
 $3 = ($2|0)==(5);
 if (!($3)) {
  return;
 }
 (_SDL_UpdateFullscreenMode($window,1)|0);
 return;
}
function _SDL_OnWindowEnter($window) {
 $window = $window|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[300441]|0;
 $1 = ((($0)) + 120|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  return;
 }
 FUNCTION_TABLE_vii[$2 & 8191]($0,$window);
 return;
}
function _SDL_OnWindowLeave($window) {
 $window = $window|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _SDL_OnWindowFocusGained($window) {
 $window = $window|0;
 var $$pre3$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $grabbed$01$i = 0, $or$cond$i = 0, $or$cond7$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GetMouse()|0);
 $1 = ((($window)) + 96|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
  $4 = HEAP32[300441]|0;
  $5 = ((($4)) + 92|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6|0)==(0|0);
  if (!($7)) {
   (FUNCTION_TABLE_iiii[$6 & 8191]($4,$window,$2)|0);
  }
 }
 $8 = ($0|0)==(0|0);
 if (!($8)) {
  $9 = ((($0)) + 76|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ($10|0)==(0);
  if (!($11)) {
   _SDL_SetMouseFocus($window);
   $12 = ((($window)) + 24|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = (($13|0) / 2)&-1;
   $15 = ((($window)) + 28|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = (($16|0) / 2)&-1;
   _SDL_WarpMouseInWindow($window,$14,$17);
  }
 }
 $18 = (_SDL_GetMouse()|0);
 $19 = ((($18)) + 76|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ((($window)) + 48|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = $22 & 256;
 $24 = $23 | $20;
 $25 = ($24|0)==(0);
 $26 = $22 & 512;
 $27 = ($26|0)==(0);
 $or$cond7$i = $27 | $25;
 $28 = HEAP32[300441]|0;
 $29 = ((($28)) + 252|0);
 $30 = HEAP32[$29>>2]|0;
 if ($or$cond7$i) {
  $41 = ($30|0)==($window|0);
  if ($41) {
   HEAP32[$29>>2] = 0;
   $43 = $28;$grabbed$01$i = 0;
  } else {
   $43 = $28;$grabbed$01$i = 0;
  }
 } else {
  $31 = ($30|0)==(0|0);
  $32 = ($30|0)==($window|0);
  $or$cond$i = $31 | $32;
  if ($or$cond$i) {
   $40 = $28;
  } else {
   $33 = ((($30)) + 48|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = $34 & -257;
   HEAP32[$33>>2] = $35;
   $36 = ((($28)) + 100|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = ($37|0)==(0|0);
   if ($38) {
    $40 = $28;
   } else {
    FUNCTION_TABLE_viii[$37 & 8191]($28,$30,0);
    $$pre3$i = HEAP32[300441]|0;
    $40 = $$pre3$i;
   }
  }
  $39 = ((($40)) + 252|0);
  HEAP32[$39>>2] = $window;
  $43 = $40;$grabbed$01$i = 1;
 }
 $42 = ((($43)) + 100|0);
 $44 = HEAP32[$42>>2]|0;
 $45 = ($44|0)==(0|0);
 if ($45) {
  return;
 }
 FUNCTION_TABLE_viii[$44 & 8191]($43,$window,$grabbed$01$i);
 return;
}
function _SDL_GetMouse() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (1201776|0);
}
function _SDL_SetMouseFocus($window) {
 $window = $window|0;
 var $$01$i = 0, $$01$i$in = 0, $$pr = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $notlhs$i = 0, $notrhs$i = 0, $or$cond3$not$i = 0, $or$cond5$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(1201820)>>2]|0;
 $1 = ($0|0)==($window|0);
 if ($1) {
  return;
 }
 $2 = ($0|0)==(0|0);
 if (!($2)) {
  (_SDL_SendWindowEvent($0,11,0,0)|0);
 }
 HEAP32[(1201820)>>2] = $window;
 $3 = ($window|0)==(0|0);
 if ($3) {
  label = 6;
 } else {
  (_SDL_SendWindowEvent($window,10,0,0)|0);
  $$pr = HEAP32[(1201820)>>2]|0;
  $4 = ($$pr|0)==(0|0);
  if ($4) {
   label = 6;
  } else {
   $$01$i$in = (1201876);
  }
 }
 if ((label|0) == 6) {
  $$01$i$in = (1201872);
 }
 $$01$i = HEAP32[$$01$i$in>>2]|0;
 $5 = HEAP32[(1201880)>>2]|0;
 $notlhs$i = ($$01$i|0)!=(0|0);
 $notrhs$i = ($5|0)!=(0);
 $or$cond3$not$i = $notlhs$i & $notrhs$i;
 $6 = HEAP32[(1201852)>>2]|0;
 $7 = ($6|0)==(0);
 $or$cond5$i = $or$cond3$not$i & $7;
 $8 = HEAP32[(1201784)>>2]|0;
 $9 = ($8|0)==(0|0);
 if ($or$cond5$i) {
  if ($9) {
   return;
  }
  (FUNCTION_TABLE_ii[$8 & 8191]($$01$i)|0);
  return;
 } else {
  if ($9) {
   return;
  }
  (FUNCTION_TABLE_ii[$8 & 8191](0)|0);
  return;
 }
}
function _SDL_WarpMouseInWindow($window,$x,$y) {
 $window = $window|0;
 $x = $x|0;
 $y = $y|0;
 var $$window = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($window|0)==(0|0);
 $1 = HEAP32[(1201820)>>2]|0;
 $$window = $0 ? $1 : $window;
 $2 = ($$window|0)==(0|0);
 if ($2) {
  return;
 }
 $3 = HEAP32[(1201796)>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  FUNCTION_TABLE_viii[$3 & 8191]($$window,$x,$y);
  return;
 }
 $5 = HEAP32[(1201816)>>2]|0;
 $6 = (_SDL_UpdateMouseFocus($$window,$x,$y)|0);
 $not$ = ($6|0)==(0);
 if ($not$) {
  return;
 }
 (_SDL_PrivateSendMouseMotion($$window,$5,0,$x,$y)|0);
 return;
}
function _SDL_UpdateMouseFocus($window,$x,$y) {
 $window = $window|0;
 $x = $x|0;
 $y = $y|0;
 var $$0 = 0, $$01$i$i = 0, $$01$i$i6 = 0, $$01$i$in$i = 0, $$pr$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $h = 0, $notlhs$i$i = 0, $notlhs$i$i7 = 0, $notrhs$i$i = 0;
 var $notrhs$i$i8 = 0, $or$cond = 0, $or$cond3 = 0, $or$cond3$not$i$i = 0, $or$cond3$not$i$i9 = 0, $or$cond5$i$i = 0, $or$cond5$i$i10 = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $w = sp + 4|0;
 $h = sp;
 $0 = ($window|0)==(0|0);
 if (!($0)) {
  $1 = ((($window)) + 48|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = $2 & 16384;
  $4 = ($3|0)==(0);
  if ($4) {
   _SDL_GetWindowSize($window,$w,$h);
   $5 = $y | $x;
   $6 = ($5|0)>(-1);
   $7 = HEAP32[$w>>2]|0;
   $8 = ($7|0)>($x|0);
   $or$cond = $6 & $8;
   $9 = HEAP32[$h>>2]|0;
   $10 = ($9|0)>($y|0);
   $or$cond3 = $or$cond & $10;
   if (!($or$cond3)) {
    $11 = HEAP32[(1201820)>>2]|0;
    $12 = ($11|0)==($window|0);
    if (!($12)) {
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
    $13 = HEAP32[(1201816)>>2]|0;
    (_SDL_PrivateSendMouseMotion($window,$13,0,$x,$y)|0);
    $14 = HEAP32[(1201820)>>2]|0;
    $15 = ($14|0)==(0|0);
    if ($15) {
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
    (_SDL_SendWindowEvent($14,11,0,0)|0);
    HEAP32[(1201820)>>2] = 0;
    $$01$i$i6 = HEAP32[(1201872)>>2]|0;
    $16 = HEAP32[(1201880)>>2]|0;
    $notlhs$i$i7 = ($$01$i$i6|0)!=(0|0);
    $notrhs$i$i8 = ($16|0)!=(0);
    $or$cond3$not$i$i9 = $notlhs$i$i7 & $notrhs$i$i8;
    $17 = HEAP32[(1201852)>>2]|0;
    $18 = ($17|0)==(0);
    $or$cond5$i$i10 = $or$cond3$not$i$i9 & $18;
    $19 = HEAP32[(1201784)>>2]|0;
    $20 = ($19|0)==(0|0);
    if ($or$cond5$i$i10) {
     if ($20) {
      $$0 = 0;
      STACKTOP = sp;return ($$0|0);
     }
     (FUNCTION_TABLE_ii[$19 & 8191]($$01$i$i6)|0);
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    } else {
     if ($20) {
      $$0 = 0;
      STACKTOP = sp;return ($$0|0);
     }
     (FUNCTION_TABLE_ii[$19 & 8191](0)|0);
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
   }
  }
 }
 $21 = HEAP32[(1201820)>>2]|0;
 $22 = ($21|0)==($window|0);
 if ($22) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 $23 = ($21|0)==(0|0);
 if (!($23)) {
  (_SDL_SendWindowEvent($21,11,0,0)|0);
 }
 HEAP32[(1201820)>>2] = $window;
 if ($0) {
  label = 16;
 } else {
  (_SDL_SendWindowEvent($window,10,0,0)|0);
  $$pr$i = HEAP32[(1201820)>>2]|0;
  $24 = ($$pr$i|0)==(0|0);
  if ($24) {
   label = 16;
  } else {
   $$01$i$in$i = (1201876);
  }
 }
 if ((label|0) == 16) {
  $$01$i$in$i = (1201872);
 }
 $$01$i$i = HEAP32[$$01$i$in$i>>2]|0;
 $25 = HEAP32[(1201880)>>2]|0;
 $notlhs$i$i = ($$01$i$i|0)!=(0|0);
 $notrhs$i$i = ($25|0)!=(0);
 $or$cond3$not$i$i = $notlhs$i$i & $notrhs$i$i;
 $26 = HEAP32[(1201852)>>2]|0;
 $27 = ($26|0)==(0);
 $or$cond5$i$i = $or$cond3$not$i$i & $27;
 $28 = HEAP32[(1201784)>>2]|0;
 $29 = ($28|0)==(0|0);
 if ($or$cond5$i$i) {
  if (!($29)) {
   (FUNCTION_TABLE_ii[$28 & 8191]($$01$i$i)|0);
  }
 } else {
  if (!($29)) {
   (FUNCTION_TABLE_ii[$28 & 8191](0)|0);
  }
 }
 $30 = HEAP32[(1201816)>>2]|0;
 (_SDL_PrivateSendMouseMotion($window,$30,0,$x,$y)|0);
 $$0 = 1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GetWindowSize($window,$w,$h) {
 $window = $window|0;
 $w = $w|0;
 $h = $h|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ($w|0)==(0|0);
   if (!($6)) {
    $7 = ((($window)) + 24|0);
    $8 = HEAP32[$7>>2]|0;
    HEAP32[$w>>2] = $8;
   }
   $9 = ($h|0)==(0|0);
   if ($9) {
    STACKTOP = sp;return;
   }
   $10 = ((($window)) + 28|0);
   $11 = HEAP32[$10>>2]|0;
   HEAP32[$h>>2] = $11;
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 STACKTOP = sp;return;
}
function _SDL_PrivateSendMouseMotion($window,$mouseID,$relative,$x,$y) {
 $window = $window|0;
 $mouseID = $mouseID|0;
 $relative = $relative|0;
 $x = $x|0;
 $y = $y|0;
 var $$03 = 0, $$04 = 0, $$1 = 0, $$pre = 0, $$window$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $9 = 0, $center_x = 0, $center_y = 0, $event = 0, $not$ = 0, $or$cond = 0, $or$cond7 = 0, $posted$0 = 0, $storemerge = 0, $storemerge5 = 0, $x_max = 0, $xrel$0 = 0, $y_max = 0;
 var $yrel$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $center_x = sp + 68|0;
 $center_y = sp + 64|0;
 $x_max = sp + 60|0;
 $y_max = sp + 56|0;
 $event = sp;
 $0 = HEAP32[(1201856)>>2]|0;
 $1 = ($0|0)==(0);
 if (!($1)) {
  HEAP32[$center_x>>2] = 0;
  HEAP32[$center_y>>2] = 0;
  _SDL_GetWindowSize($window,$center_x,$center_y);
  $2 = HEAP32[$center_x>>2]|0;
  $3 = (($2|0) / 2)&-1;
  HEAP32[$center_x>>2] = $3;
  $4 = HEAP32[$center_y>>2]|0;
  $5 = (($4|0) / 2)&-1;
  HEAP32[$center_y>>2] = $5;
  $6 = ($3|0)==($x|0);
  $7 = ($5|0)==($y|0);
  $or$cond = $6 & $7;
  if ($or$cond) {
   HEAP32[(1201840)>>2] = $x;
   HEAP32[(1201844)>>2] = $y;
   $$1 = 0;
   STACKTOP = sp;return ($$1|0);
  }
  $8 = ($window|0)==(0|0);
  $9 = HEAP32[(1201820)>>2]|0;
  $$window$i = $8 ? $9 : $window;
  $10 = ($$window$i|0)==(0|0);
  do {
   if (!($10)) {
    $11 = HEAP32[(1201796)>>2]|0;
    $12 = ($11|0)==(0|0);
    if (!($12)) {
     FUNCTION_TABLE_viii[$11 & 8191]($$window$i,$3,$5);
     break;
    }
    $13 = HEAP32[(1201816)>>2]|0;
    $14 = (_SDL_UpdateMouseFocus($$window$i,$3,$5)|0);
    $not$ = ($14|0)==(0);
    if (!($not$)) {
     (_SDL_PrivateSendMouseMotion($$window$i,$13,0,$3,$5)|0);
    }
   }
  } while(0);
 }
 $15 = ($relative|0)!=(0);
 $16 = HEAP32[(1201840)>>2]|0;
 if ($15) {
  $17 = (($16) + ($x))|0;
  $18 = HEAP32[(1201844)>>2]|0;
  $19 = (($18) + ($y))|0;
  $$03 = $19;$$04 = $17;$xrel$0 = $x;$yrel$0 = $y;
 } else {
  $20 = (($x) - ($16))|0;
  $21 = HEAP32[(1201844)>>2]|0;
  $22 = (($y) - ($21))|0;
  $$03 = $y;$$04 = $x;$xrel$0 = $20;$yrel$0 = $22;
 }
 $23 = $yrel$0 | $xrel$0;
 $24 = ($23|0)==(0);
 if ($24) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $25 = HEAP32[(1201852)>>2]|0;
 $26 = ($25|0)==(0);
 if ($26) {
  HEAP32[(1201824)>>2] = $$04;
  $storemerge = $$03;
 } else {
  $27 = HEAP32[(1201824)>>2]|0;
  $28 = (($27) + ($xrel$0))|0;
  HEAP32[(1201824)>>2] = $28;
  $29 = HEAP32[(1201828)>>2]|0;
  $30 = (($29) + ($yrel$0))|0;
  $storemerge = $30;
 }
 HEAP32[(1201828)>>2] = $storemerge;
 $31 = ($window|0)==(0|0);
 if ($31) {
  $56 = $25;
 } else {
  $32 = ((($window)) + 48|0);
  $33 = HEAP32[$32>>2]|0;
  $34 = $33 & 16384;
  $35 = ($34|0)==(0);
  if ($35) {
   HEAP32[$x_max>>2] = 0;
   HEAP32[$y_max>>2] = 0;
   $36 = HEAP32[(1201820)>>2]|0;
   _SDL_GetWindowSize($36,$x_max,$y_max);
   $37 = HEAP32[$x_max>>2]|0;
   $38 = (($37) + -1)|0;
   HEAP32[$x_max>>2] = $38;
   $39 = HEAP32[$y_max>>2]|0;
   $40 = (($39) + -1)|0;
   HEAP32[$y_max>>2] = $40;
   $41 = HEAP32[(1201824)>>2]|0;
   $42 = ($41|0)<($37|0);
   if ($42) {
    $44 = $41;
   } else {
    HEAP32[(1201824)>>2] = $38;
    $44 = $38;
   }
   $43 = ($44|0)<(0);
   if ($43) {
    HEAP32[(1201824)>>2] = 0;
   }
   $45 = HEAP32[(1201828)>>2]|0;
   $46 = ($45|0)<($39|0);
   if ($46) {
    $48 = $45;
   } else {
    HEAP32[(1201828)>>2] = $40;
    $48 = $40;
   }
   $47 = ($48|0)<(0);
   if ($47) {
    HEAP32[(1201828)>>2] = 0;
   }
   $$pre = HEAP32[(1201852)>>2]|0;
   $56 = $$pre;
  } else {
   $56 = $25;
  }
 }
 $49 = HEAP32[(1201832)>>2]|0;
 $50 = (($49) + ($xrel$0))|0;
 HEAP32[(1201832)>>2] = $50;
 $51 = HEAP32[(1201836)>>2]|0;
 $52 = (($51) + ($yrel$0))|0;
 HEAP32[(1201836)>>2] = $52;
 $53 = HEAP32[(1201880)>>2]|0;
 $54 = ($53|0)!=(0);
 $55 = ($56|0)==(0);
 $or$cond7 = $54 & $55;
 if ($or$cond7) {
  $57 = HEAP32[(1201788)>>2]|0;
  $58 = ($57|0)==(0|0);
  if (!($58)) {
   $59 = HEAP32[(1201876)>>2]|0;
   $60 = ($59|0)==(0|0);
   if (!($60)) {
    FUNCTION_TABLE_vi[$57 & 8191]($59);
   }
  }
 }
 $61 = (_SDL_EventState(1024,-1)|0);
 $62 = ($61<<24>>24)==(1);
 if ($62) {
  HEAP32[$event>>2] = 1024;
  $63 = HEAP32[(1201820)>>2]|0;
  $64 = ($63|0)==(0|0);
  if ($64) {
   $68 = 0;
  } else {
   $65 = ((($63)) + 4|0);
   $66 = HEAP32[$65>>2]|0;
   $68 = $66;
  }
  $67 = ((($event)) + 8|0);
  HEAP32[$67>>2] = $68;
  $69 = ((($event)) + 12|0);
  HEAP32[$69>>2] = $mouseID;
  $70 = HEAP32[(1201848)>>2]|0;
  $71 = ((($event)) + 16|0);
  HEAP32[$71>>2] = $70;
  $72 = HEAP32[(1201824)>>2]|0;
  $73 = ((($event)) + 20|0);
  HEAP32[$73>>2] = $72;
  $74 = HEAP32[(1201828)>>2]|0;
  $75 = ((($event)) + 24|0);
  HEAP32[$75>>2] = $74;
  $76 = ((($event)) + 28|0);
  HEAP32[$76>>2] = $xrel$0;
  $77 = ((($event)) + 32|0);
  HEAP32[$77>>2] = $yrel$0;
  $78 = (_SDL_PushEvent($event)|0);
  $79 = ($78|0)>(0);
  $80 = $79&1;
  $posted$0 = $80;
 } else {
  $posted$0 = 0;
 }
 if ($15) {
  $81 = HEAP32[(1201824)>>2]|0;
  HEAP32[(1201840)>>2] = $81;
  $82 = HEAP32[(1201828)>>2]|0;
  $storemerge5 = $82;
 } else {
  HEAP32[(1201840)>>2] = $$04;
  $storemerge5 = $$03;
 }
 HEAP32[(1201844)>>2] = $storemerge5;
 $$1 = $posted$0;
 STACKTOP = sp;return ($$1|0);
}
function _SDL_EventState($type,$state) {
 $type = $type|0;
 $state = $state|0;
 var $$pre$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $7 = 0, $8 = 0, $9 = 0, $current_state$0 = 0, $entry$01$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $type >>> 8;
 $1 = $0 & 255;
 $2 = (1201888 + ($1<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  label = 3;
 } else {
  $5 = $type >>> 5;
  $6 = $5 & 7;
  $7 = (($3) + ($6<<2)|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = $type & 31;
  $10 = 1 << $9;
  $11 = $8 & $10;
  $12 = ($11|0)==(0);
  if ($12) {
   label = 3;
  } else {
   $current_state$0 = 0;
  }
 }
 if ((label|0) == 3) {
  $current_state$0 = 1;
 }
 $13 = $current_state$0&255;
 $14 = ($13|0)==($state|0);
 if ($14) {
  return ($current_state$0|0);
 }
 switch ($state|0) {
 case 0:  {
  if ($4) {
   $15 = (_SDL_calloc(1,32)|0);
   HEAP32[$2>>2] = $15;
   $16 = ($15|0)==(0|0);
   if ($16) {
    return ($current_state$0|0);
   } else {
    $22 = $15;
   }
  } else {
   $22 = $3;
  }
  $17 = $type & 31;
  $18 = 1 << $17;
  $19 = $type >>> 5;
  $20 = $19 & 7;
  $21 = (($22) + ($20<<2)|0);
  $23 = HEAP32[$21>>2]|0;
  $24 = $23 | $18;
  HEAP32[$21>>2] = $24;
  $25 = HEAP32[(125892)>>2]|0;
  $26 = ($25|0)==(0);
  if ($26) {
   return ($current_state$0|0);
  }
  $27 = HEAP32[31472]|0;
  $28 = (_SDL_LockMutex($27)|0);
  $29 = ($28|0)==(0);
  if (!($29)) {
   return ($current_state$0|0);
  }
  $30 = HEAP32[(125904)>>2]|0;
  $31 = ($30|0)==(0|0);
  if (!($31)) {
   $entry$01$i$i = $30;
   while(1) {
    $32 = ((($entry$01$i$i)) + 72|0);
    $33 = HEAP32[$32>>2]|0;
    $34 = HEAP32[$entry$01$i$i>>2]|0;
    $35 = ($34|0)==($type|0);
    $36 = $33;
    if ($35) {
     $38 = ((($entry$01$i$i)) + 68|0);
     $39 = HEAP32[$38>>2]|0;
     $40 = ($39|0)==(0|0);
     $41 = $39;
     if ($40) {
      $44 = $33;
     } else {
      $42 = ((($39)) + 72|0);
      HEAP32[$42>>2] = $36;
      $$pre$i$i = HEAP32[$32>>2]|0;
      $44 = $$pre$i$i;
     }
     $43 = ($44|0)==(0|0);
     $45 = $44;
     if (!($43)) {
      $46 = ((($44)) + 68|0);
      HEAP32[$46>>2] = $41;
     }
     $47 = HEAP32[(125904)>>2]|0;
     $48 = ($47|0)==($entry$01$i$i|0);
     if ($48) {
      HEAP32[(125904)>>2] = $45;
     }
     $49 = HEAP32[(125908)>>2]|0;
     $50 = ($49|0)==($entry$01$i$i|0);
     if ($50) {
      $51 = HEAP32[$38>>2]|0;
      HEAP32[(125908)>>2] = $51;
     }
     $52 = HEAP32[(125912)>>2]|0;
     HEAP32[$32>>2] = $52;
     HEAP32[(125912)>>2] = $entry$01$i$i;
     $53 = HEAP32[(125896)>>2]|0;
     $54 = (($53) + -1)|0;
     HEAP32[(125896)>>2] = $54;
    }
    $37 = ($33|0)==(0|0);
    if ($37) {
     break;
    } else {
     $entry$01$i$i = $33;
    }
   }
  }
  return ($current_state$0|0);
  break;
 }
 case 1:  {
  $55 = $type & 31;
  $56 = 1 << $55;
  $57 = $56 ^ -1;
  $58 = $type >>> 5;
  $59 = $58 & 7;
  $60 = (($3) + ($59<<2)|0);
  $61 = HEAP32[$60>>2]|0;
  $62 = $61 & $57;
  HEAP32[$60>>2] = $62;
  return ($current_state$0|0);
  break;
 }
 default: {
  return ($current_state$0|0);
 }
 }
 return (0)|0;
}
function _SDL_calloc($nmemb,$size) {
 $nmemb = $nmemb|0;
 $size = $size|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_calloc($nmemb,$size)|0);
 return ($0|0);
}
function _SDL_LockMutex($mutex) {
 $mutex = $mutex|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _SDL_UnlockMutex($mutex) {
 $mutex = $mutex|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _SDL_PushEvent($event) {
 $event = $event|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $curr$0 = 0, $curr$01 = 0, $curr$02 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GetTicks()|0);
 $1 = ((($event)) + 4|0);
 HEAP32[$1>>2] = $0;
 $2 = HEAP32[300435]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
  $4 = HEAP32[300436]|0;
  $5 = (FUNCTION_TABLE_iii[$2 & 8191]($4,$event)|0);
  $6 = ($5|0)==(0);
  if ($6) {
   $$0 = 0;
   return ($$0|0);
  }
 }
 $curr$01 = HEAP32[300734]|0;
 $7 = ($curr$01|0)==(0|0);
 if (!($7)) {
  $curr$02 = $curr$01;
  while(1) {
   $8 = HEAP32[$curr$02>>2]|0;
   $9 = ((($curr$02)) + 4|0);
   $10 = HEAP32[$9>>2]|0;
   (FUNCTION_TABLE_iii[$8 & 8191]($10,$event)|0);
   $11 = ((($curr$02)) + 8|0);
   $curr$0 = HEAP32[$11>>2]|0;
   $12 = ($curr$0|0)==(0|0);
   if ($12) {
    break;
   } else {
    $curr$02 = $curr$0;
   }
  }
 }
 $13 = (_SDL_PeepEvents($event,1,0,0,0)|0);
 $14 = ($13|0)<(1);
 if ($14) {
  $$0 = -1;
  return ($$0|0);
 }
 _SDL_GestureProcessEvent($event);
 $$0 = 1;
 return ($$0|0);
}
function _SDL_GetTicks() {
 var $$pr = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $now = 0, $now1 = 0, $ticks$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $now = sp + 8|0;
 $now1 = sp;
 $0 = HEAP32[300728]|0;
 $1 = ($0|0)==(0);
 do {
  if ($1) {
   HEAP32[300728] = 1;
   $2 = (_clock_gettime(4,(1202916|0))|0);
   $3 = ($2|0)==(0);
   if ($3) {
    HEAP32[300731] = 1;
    break;
   } else {
    (_gettimeofday((1202928|0),(0|0))|0);
    label = 5;
    break;
   }
  } else {
   label = 5;
  }
 } while(0);
 if ((label|0) == 5) {
  $$pr = HEAP32[300731]|0;
  $4 = ($$pr|0)==(0);
  if ($4) {
   (_gettimeofday(($now1|0),(0|0))|0);
   $15 = HEAP32[$now1>>2]|0;
   $16 = HEAP32[300732]|0;
   $17 = (($15) - ($16))|0;
   $18 = ($17*1000)|0;
   $19 = ((($now1)) + 4|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = HEAP32[(1202932)>>2]|0;
   $22 = (($20) - ($21))|0;
   $23 = (($22|0) / 1000)&-1;
   $24 = (($23) + ($18))|0;
   $ticks$0 = $24;
   STACKTOP = sp;return ($ticks$0|0);
  }
 }
 (_clock_gettime(4,($now|0))|0);
 $5 = HEAP32[$now>>2]|0;
 $6 = HEAP32[300729]|0;
 $7 = (($5) - ($6))|0;
 $8 = ($7*1000)|0;
 $9 = ((($now)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = HEAP32[(1202920)>>2]|0;
 $12 = (($10) - ($11))|0;
 $13 = (($12|0) / 1000000)&-1;
 $14 = (($13) + ($8))|0;
 $ticks$0 = $14;
 STACKTOP = sp;return ($ticks$0|0);
}
function _SDL_PeepEvents($events,$numevents,$action,$minType,$maxType) {
 $events = $events|0;
 $numevents = $numevents|0;
 $action = $action|0;
 $minType = $minType|0;
 $maxType = $maxType|0;
 var $$0 = 0, $$0$i = 0, $$lcssa81 = 0, $$lcssa82 = 0, $$lcssa83 = 0, $$lcssa87 = 0, $$lcssa88 = 0, $$numevents = 0, $$pre = 0, $$pre34 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $entry$0$i = 0, $entry$0$ph$us62 = 0, $entry$0$ph64 = 0, $entry$0$us = 0, $entry$0$us$lcssa86 = 0, $entry$0$us$us = 0, $entry$0$us$us$lcssa80 = 0, $exitcond = 0, $i$08 = 0, $not$ = 0, $or$cond4$us = 0, $or$cond4$us$us = 0, $tmpevent = 0, $tmpevent$events = 0;
 var $used$010 = 0, $used$1$ph$lcssa = 0, $used$1$ph$us61 = 0, $used$1$ph63 = 0, $used$3 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $wmmsg$011 = 0, $wmmsg$011$lcssa = 0, $wmmsg$1 = 0, $wmmsg$1$us = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 72|0;
 $vararg_buffer1 = sp + 64|0;
 $vararg_buffer = sp + 56|0;
 $tmpevent = sp;
 $0 = HEAP32[(125892)>>2]|0;
 $1 = ($0|0)==(0);
 if ($1) {
  $2 = ($action|0)==(0);
  if ($2) {
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  }
  (_SDL_SetError(596551,$vararg_buffer)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = HEAP32[31472]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  $5 = (_SDL_LockMutex($3)|0);
  $6 = ($5|0)==(0);
  if (!($6)) {
   $113 = (_SDL_SetError(596619,$vararg_buffer3)|0);
   $$0 = $113;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $7 = ($action|0)==(0);
 if ($7) {
  $8 = ($numevents|0)>(0);
  if ($8) {
   $i$08 = 0;$used$010 = 0;
   while(1) {
    $9 = (($events) + (($i$08*56)|0)|0);
    $10 = HEAP32[(125896)>>2]|0;
    $11 = ($10|0)>(65534);
    do {
     if ($11) {
      $12 = HEAP32[(125896)>>2]|0;
      HEAP32[$vararg_buffer1>>2] = $12;
      (_SDL_SetError(596587,$vararg_buffer1)|0);
      $$0$i = 0;
     } else {
      $13 = HEAP32[(125912)>>2]|0;
      $14 = ($13|0)==(0|0);
      if ($14) {
       $15 = (_SDL_malloc(80)|0);
       $16 = ($15|0)==(0|0);
       if ($16) {
        $$0$i = 0;
        break;
       } else {
        $entry$0$i = $15;
       }
      } else {
       $17 = ((($13)) + 72|0);
       $18 = HEAP32[$17>>2]|0;
       HEAP32[(125912)>>2] = $18;
       $entry$0$i = $13;
      }
      dest=$entry$0$i; src=$9; stop=dest+56|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
      $19 = HEAP32[$9>>2]|0;
      $20 = ($19|0)==(513);
      if ($20) {
       $21 = ((($entry$0$i)) + 56|0);
       $22 = (((($events) + (($i$08*56)|0)|0)) + 8|0);
       $23 = HEAP32[$22>>2]|0;
       ;HEAP32[$21>>2]=HEAP32[$23>>2]|0;HEAP32[$21+4>>2]=HEAP32[$23+4>>2]|0;HEAP32[$21+8>>2]=HEAP32[$23+8>>2]|0;
       $24 = ((($entry$0$i)) + 8|0);
       HEAP32[$24>>2] = $21;
      }
      $25 = HEAP32[(125908)>>2]|0;
      $26 = ($25|0)==(0|0);
      if ($26) {
       HEAP32[(125904)>>2] = $entry$0$i;
       HEAP32[(125908)>>2] = $entry$0$i;
       $30 = ((($entry$0$i)) + 68|0);
       HEAP32[$30>>2] = 0;
      } else {
       $27 = $25;
       $28 = ((($25)) + 72|0);
       HEAP32[$28>>2] = $entry$0$i;
       $29 = ((($entry$0$i)) + 68|0);
       HEAP32[$29>>2] = $27;
       HEAP32[(125908)>>2] = $entry$0$i;
      }
      $31 = ((($entry$0$i)) + 72|0);
      HEAP32[$31>>2] = 0;
      $32 = HEAP32[(125896)>>2]|0;
      $33 = (($32) + 1)|0;
      HEAP32[(125896)>>2] = $33;
      $34 = HEAP32[(125896)>>2]|0;
      $35 = HEAP32[(125900)>>2]|0;
      $36 = ($34|0)>($35|0);
      if ($36) {
       $37 = HEAP32[(125896)>>2]|0;
       HEAP32[(125900)>>2] = $37;
       $$0$i = 1;
      } else {
       $$0$i = 1;
      }
     }
    } while(0);
    $38 = (($$0$i) + ($used$010))|0;
    $39 = (($i$08) + 1)|0;
    $exitcond = ($39|0)==($numevents|0);
    if ($exitcond) {
     $used$3 = $38;
     break;
    } else {
     $i$08 = $39;$used$010 = $38;
    }
   }
  } else {
   $used$3 = 0;
  }
 } else {
  $40 = ($events|0)==(0|0);
  $$numevents = $40 ? 1 : $numevents;
  $tmpevent$events = $40 ? $tmpevent : $events;
  $41 = HEAP32[(125916)>>2]|0;
  $42 = ($41|0)==(0|0);
  if (!($42)) {
   $$pre = HEAP32[(125920)>>2]|0;
   $45 = $$pre;$wmmsg$011 = $41;
   while(1) {
    $43 = ((($wmmsg$011)) + 12|0);
    $44 = HEAP32[$43>>2]|0;
    HEAP32[$43>>2] = $45;
    $46 = ($44|0)==(0|0);
    $47 = $wmmsg$011;
    if ($46) {
     $wmmsg$011$lcssa = $wmmsg$011;
     break;
    } else {
     $45 = $47;$wmmsg$011 = $44;
    }
   }
   HEAP32[(125920)>>2] = $wmmsg$011$lcssa;
  }
  HEAP32[(125916)>>2] = 0;
  $48 = HEAP32[(125904)>>2]|0;
  $49 = ($action|0)==(2);
  $not$ = $40 ^ 1;
  $50 = $49 & $not$;
  $51 = ($$numevents|0)>(0);
  L40: do {
   if ($50) {
    if ($51) {
     $entry$0$ph$us62 = $48;$used$1$ph$us61 = 0;
     while(1) {
      $entry$0$us$us = $entry$0$ph$us62;
      while(1) {
       $82 = ($entry$0$us$us|0)==(0|0);
       if ($82) {
        $used$1$ph$lcssa = $used$1$ph$us61;
        break L40;
       }
       $83 = ((($entry$0$us$us)) + 72|0);
       $84 = HEAP32[$83>>2]|0;
       $85 = HEAP32[$entry$0$us$us>>2]|0;
       $86 = ($85>>>0)<($minType>>>0);
       $87 = ($85>>>0)>($maxType>>>0);
       $or$cond4$us$us = $86 | $87;
       if ($or$cond4$us$us) {
        $entry$0$us$us = $84;
       } else {
        $$lcssa81 = $83;$$lcssa82 = $84;$$lcssa83 = $entry$0$us$us;$entry$0$us$us$lcssa80 = $entry$0$us$us;
        break;
       }
      }
      $88 = (($tmpevent$events) + (($used$1$ph$us61*56)|0)|0);
      dest=$88; src=$entry$0$us$us$lcssa80; stop=dest+56|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
      $89 = HEAP32[$$lcssa83>>2]|0;
      $90 = ($89|0)==(513);
      if ($90) {
       $52 = HEAP32[(125920)>>2]|0;
       $53 = ($52|0)==(0|0);
       if ($53) {
        $56 = (_SDL_malloc(16)|0);
        $wmmsg$1$us = $56;
       } else {
        $54 = ((($52)) + 12|0);
        $55 = HEAP32[$54>>2]|0;
        HEAP32[(125920)>>2] = $55;
        $wmmsg$1$us = $52;
       }
       $57 = ((($entry$0$us$us$lcssa80)) + 8|0);
       $58 = HEAP32[$57>>2]|0;
       ;HEAP32[$wmmsg$1$us>>2]=HEAP32[$58>>2]|0;HEAP32[$wmmsg$1$us+4>>2]=HEAP32[$58+4>>2]|0;HEAP32[$wmmsg$1$us+8>>2]=HEAP32[$58+8>>2]|0;
       $59 = HEAP32[(125916)>>2]|0;
       $60 = ((($wmmsg$1$us)) + 12|0);
       HEAP32[$60>>2] = $59;
       HEAP32[(125916)>>2] = $wmmsg$1$us;
       $61 = (((($tmpevent$events) + (($used$1$ph$us61*56)|0)|0)) + 8|0);
       HEAP32[$61>>2] = $wmmsg$1$us;
      }
      $62 = (($used$1$ph$us61) + 1)|0;
      $63 = ((($entry$0$us$us$lcssa80)) + 68|0);
      $64 = HEAP32[$63>>2]|0;
      $65 = ($64|0)==(0|0);
      $66 = $64;
      if (!($65)) {
       $67 = HEAP32[$$lcssa81>>2]|0;
       $68 = ((($64)) + 72|0);
       HEAP32[$68>>2] = $67;
      }
      $69 = HEAP32[$$lcssa81>>2]|0;
      $70 = ($69|0)==(0|0);
      $71 = $69;
      if (!($70)) {
       $72 = ((($69)) + 68|0);
       HEAP32[$72>>2] = $66;
      }
      $73 = HEAP32[(125904)>>2]|0;
      $74 = ($73|0)==($entry$0$us$us$lcssa80|0);
      if ($74) {
       HEAP32[(125904)>>2] = $71;
      }
      $75 = HEAP32[(125908)>>2]|0;
      $76 = ($75|0)==($entry$0$us$us$lcssa80|0);
      if ($76) {
       $77 = HEAP32[$63>>2]|0;
       HEAP32[(125908)>>2] = $77;
      }
      $78 = HEAP32[(125912)>>2]|0;
      HEAP32[$$lcssa81>>2] = $78;
      HEAP32[(125912)>>2] = $entry$0$us$us$lcssa80;
      $79 = HEAP32[(125896)>>2]|0;
      $80 = (($79) + -1)|0;
      HEAP32[(125896)>>2] = $80;
      $81 = ($62|0)<($$numevents|0);
      if ($81) {
       $entry$0$ph$us62 = $$lcssa82;$used$1$ph$us61 = $62;
      } else {
       $used$1$ph$lcssa = $62;
       break;
      }
     }
    } else {
     $used$1$ph$lcssa = 0;
    }
   } else {
    if ($51) {
     $114 = 0;$entry$0$ph64 = $48;$used$1$ph63 = 0;
     while(1) {
      $entry$0$us = $entry$0$ph64;
      while(1) {
       $91 = ($entry$0$us|0)==(0|0);
       if ($91) {
        $used$1$ph$lcssa = $used$1$ph63;
        break L40;
       }
       $92 = ((($entry$0$us)) + 72|0);
       $93 = HEAP32[$92>>2]|0;
       $94 = HEAP32[$entry$0$us>>2]|0;
       $95 = ($94>>>0)<($minType>>>0);
       $96 = ($94>>>0)>($maxType>>>0);
       $or$cond4$us = $95 | $96;
       if ($or$cond4$us) {
        $entry$0$us = $93;
       } else {
        $$lcssa87 = $93;$$lcssa88 = $entry$0$us;$entry$0$us$lcssa86 = $entry$0$us;
        break;
       }
      }
      $97 = (($tmpevent$events) + (($used$1$ph63*56)|0)|0);
      dest=$97; src=$entry$0$us$lcssa86; stop=dest+56|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
      $98 = HEAP32[$$lcssa88>>2]|0;
      $99 = ($98|0)==(513);
      if ($99) {
       $100 = HEAP32[(125920)>>2]|0;
       $101 = ($100|0)==(0|0);
       if ($101) {
        $104 = (_SDL_malloc(16)|0);
        $$pre34 = HEAP32[(125916)>>2]|0;
        $108 = $$pre34;$wmmsg$1 = $104;
       } else {
        $102 = ((($100)) + 12|0);
        $103 = HEAP32[$102>>2]|0;
        HEAP32[(125920)>>2] = $103;
        $108 = $114;$wmmsg$1 = $100;
       }
       $105 = ((($entry$0$us$lcssa86)) + 8|0);
       $106 = HEAP32[$105>>2]|0;
       ;HEAP32[$wmmsg$1>>2]=HEAP32[$106>>2]|0;HEAP32[$wmmsg$1+4>>2]=HEAP32[$106+4>>2]|0;HEAP32[$wmmsg$1+8>>2]=HEAP32[$106+8>>2]|0;
       $107 = ((($wmmsg$1)) + 12|0);
       HEAP32[$107>>2] = $108;
       HEAP32[(125916)>>2] = $wmmsg$1;
       $109 = (((($tmpevent$events) + (($used$1$ph63*56)|0)|0)) + 8|0);
       HEAP32[$109>>2] = $wmmsg$1;
       $110 = $wmmsg$1;
       $115 = $110;
      } else {
       $115 = $114;
      }
      $111 = (($used$1$ph63) + 1)|0;
      $112 = ($111|0)<($$numevents|0);
      if ($112) {
       $114 = $115;$entry$0$ph64 = $$lcssa87;$used$1$ph63 = $111;
      } else {
       $used$1$ph$lcssa = $111;
       break;
      }
     }
    } else {
     $used$1$ph$lcssa = 0;
    }
   }
  } while(0);
  $used$3 = $used$1$ph$lcssa;
 }
 $$0 = $used$3;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_malloc($size) {
 $size = $size|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_malloc($size)|0);
 return ($0|0);
}
function _SDL_GestureProcessEvent($event) {
 $event = $event|0;
 var $$0$i4 = 0, $$0$i4$ph = 0, $$0$i5 = 0, $$lcssa$i$i = 0, $$lcssa120 = 0, $$lcssa121 = 0, $$lcssa125 = 0, $$lcssa127 = 0, $$off = 0, $$pr = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0.0;
 var $141 = 0, $142 = 0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0.0, $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0.0, $152 = 0, $153 = 0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0;
 var $16 = 0, $160 = 0.0, $161 = 0, $162 = 0, $163 = 0.0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0;
 var $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0.0, $204 = 0, $205 = 0.0, $206 = 0.0, $207 = 0.0, $208 = 0.0, $209 = 0.0, $21 = 0, $210 = 0, $211 = 0.0, $212 = 0.0;
 var $213 = 0.0, $214 = 0.0, $215 = 0, $216 = 0.0, $217 = 0, $218 = 0.0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0.0;
 var $231 = 0, $232 = 0, $233 = 0.0, $234 = 0.0, $235 = 0, $236 = 0.0, $237 = 0, $238 = 0.0, $239 = 0.0, $24 = 0, $240 = 0.0, $241 = 0.0, $242 = 0.0, $243 = 0.0, $244 = 0.0, $245 = 0.0, $246 = 0, $247 = 0.0, $248 = 0.0, $249 = 0;
 var $25 = 0.0, $250 = 0, $251 = 0, $252 = 0.0, $253 = 0.0, $254 = 0, $255 = 0, $256 = 0.0, $257 = 0.0, $258 = 0.0, $259 = 0.0, $26 = 0, $260 = 0.0, $261 = 0, $262 = 0.0, $263 = 0.0, $264 = 0.0, $265 = 0.0, $266 = 0.0, $267 = 0.0;
 var $268 = 0.0, $269 = 0.0, $27 = 0.0, $270 = 0.0, $271 = 0.0, $272 = 0.0, $273 = 0.0, $274 = 0.0, $275 = 0.0, $276 = 0.0, $277 = 0.0, $278 = 0.0, $279 = 0.0, $28 = 0, $280 = 0.0, $281 = 0.0, $282 = 0.0, $283 = 0.0, $284 = 0.0, $285 = 0.0;
 var $286 = 0.0, $287 = 0.0, $288 = 0.0, $289 = 0.0, $29 = 0, $290 = 0.0, $291 = 0.0, $292 = 0.0, $293 = 0.0, $294 = 0.0, $295 = 0.0, $296 = 0.0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0;
 var $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0;
 var $321 = 0, $322 = 0.0, $323 = 0, $324 = 0, $325 = 0.0, $326 = 0.0, $327 = 0.0, $328 = 0.0, $329 = 0.0, $33 = 0, $330 = 0, $331 = 0.0, $332 = 0.0, $333 = 0.0, $334 = 0.0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0;
 var $34 = 0, $340 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0.0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0.0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0.0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0.0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bestDiff$010$i = 0.0, $bestDiff$1$i = 0.0, $bestDiff$1$i$lcssa = 0.0, $bestTempl$0 = 0, $bestTempl$1 = 0, $bestTempl$1$lcssa = 0, $dDist$0 = 0.0;
 var $dtheta$0 = 0.0, $event$i = 0, $event$i17 = 0, $event$i18 = 0, $event$i20 = 0, $exitcond$i$i = 0, $exitcond$i$i13 = 0, $f1$0$lcssa$i$i = 0.0, $f1$0$ph20$i$i = 0.0, $f1$09$i$i = 0.0, $f1$09$i$i$lcssa = 0.0, $f1$09$i$i$phi = 0.0, $f2$0$lcssa$i$i = 0.0, $f2$0$ph22$i$i = 0.0, $f2$011$i$i = 0.0, $hash$01$i$i = 0, $hash$01$i$i12 = 0, $i$01$i = 0, $i$01$i$lcssa = 0, $i$01$i2 = 0;
 var $i$011$i = 0, $i$02$i$i = 0, $i$02$i$i11 = 0, $i$043 = 0, $index$0 = 0, $path = 0, $points$i = 0, $switch = 0, $ta$0$ph18$i$i = 0.0, $ta$07$i$i = 0.0, $ta$07$i$i$lcssa = 0.0, $tb$0$ph17$i$i = 0.0, $vararg_buffer = 0, $x1$0$ph19$i$i = 0.0, $x1$08$i$i = 0.0, $x1$08$i$i$lcssa = 0.0, $x1$08$i$i$phi = 0.0, $x2$0$ph21$i$i = 0.0, $x2$010$i$i = 0.0, $x2$010$i$i$lcssa = 0.0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1264|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp + 224|0;
 $event$i20 = sp + 168|0;
 $points$i = sp + 744|0;
 $event$i18 = sp + 112|0;
 $event$i17 = sp + 56|0;
 $event$i = sp;
 $path = sp + 232|0;
 $0 = HEAP32[$event>>2]|0;
 $$off = (($0) + -1792)|0;
 $switch = ($$off>>>0)<(3);
 if (!($switch)) {
  STACKTOP = sp;return;
 }
 $1 = ((($event)) + 8|0);
 $2 = $1;
 $3 = $2;
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2) + 4)|0;
 $6 = $5;
 $7 = HEAP32[$6>>2]|0;
 $8 = HEAP32[300437]|0;
 $9 = ($8|0)>(0);
 if (!($9)) {
  STACKTOP = sp;return;
 }
 $10 = HEAP32[300439]|0;
 $i$01$i = 0;
 while(1) {
  $13 = (($10) + (($i$01$i*8232)|0)|0);
  $14 = $13;
  $15 = $14;
  $16 = HEAP32[$15>>2]|0;
  $17 = (($14) + 4)|0;
  $18 = $17;
  $19 = HEAP32[$18>>2]|0;
  $20 = ($16|0)==($4|0);
  $21 = ($19|0)==($7|0);
  $22 = $20 & $21;
  $12 = (($i$01$i) + 1)|0;
  if ($22) {
   $$lcssa125 = $13;$$lcssa127 = $13;$i$01$i$lcssa = $i$01$i;
   break;
  }
  $11 = ($12|0)<($8|0);
  if ($11) {
   $i$01$i = $12;
  } else {
   label = 52;
   break;
  }
 }
 if ((label|0) == 52) {
  STACKTOP = sp;return;
 }
 $23 = ($$lcssa125|0)==(0|0);
 if ($23) {
  STACKTOP = sp;return;
 }
 $24 = ((($event)) + 24|0);
 $25 = +HEAPF32[$24>>2];
 $26 = ((($event)) + 28|0);
 $27 = +HEAPF32[$26>>2];
 switch ($0|0) {
 case 1793:  {
  $28 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 8216|0);
  $29 = HEAP16[$28>>1]|0;
  $30 = (($29) + -1)<<16>>16;
  HEAP16[$28>>1] = $30;
  $31 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 8228|0);
  $32 = HEAP32[$31>>2]|0;
  $33 = ($32|0)==(0);
  do {
   if ($33) {
    $131 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 16|0);
    (_SDL_memset($points$i,0,512)|0);
    _dollarNormalize($131,$points$i);
    $132 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 8220|0);
    $133 = HEAP32[$132>>2]|0;
    $134 = ($133|0)>(0);
    if (!($134)) {
     break;
    }
    $135 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 8224|0);
    $bestDiff$010$i = 1.0E+4;$bestTempl$0 = -1;$i$011$i = 0;
    while(1) {
     $136 = HEAP32[$135>>2]|0;
     $137 = (($136) + (($i$011$i*516)|0)|0);
     $138 = (+_dollarDifference($points$i,$137,-0.185407355427742));
     $139 = (+_dollarDifference($points$i,$137,0.185407355427742));
     $140 = (+_SDL_fabs(-1.5707963267948966));
     $141 = $140 > 0.034906585039886591;
     $142 = $138 < $139;
     L25: do {
      if ($141) {
       $339 = $142;$f1$0$ph20$i$i = $138;$f2$0$ph22$i$i = $139;$ta$0$ph18$i$i = -0.78539816339744828;$tb$0$ph17$i$i = 0.78539816339744828;$x1$0$ph19$i$i = -0.185407355427742;$x2$0$ph21$i$i = 0.185407355427742;
       while(1) {
        $143 = $tb$0$ph17$i$i * 0.61803398899999995;
        $340 = $339;$f1$09$i$i = $f1$0$ph20$i$i;$f2$011$i$i = $f2$0$ph22$i$i;$ta$07$i$i = $ta$0$ph18$i$i;$x1$08$i$i = $x1$0$ph19$i$i;$x2$010$i$i = $x2$0$ph21$i$i;
        while(1) {
         if ($340) {
          $f1$09$i$i$lcssa = $f1$09$i$i;$ta$07$i$i$lcssa = $ta$07$i$i;$x1$08$i$i$lcssa = $x1$08$i$i;$x2$010$i$i$lcssa = $x2$010$i$i;
          break;
         }
         $154 = $x1$08$i$i;
         $155 = $154 * 0.38196601100000005;
         $156 = $143 + $155;
         $157 = $156;
         $158 = (+_dollarDifference($points$i,$137,$157));
         $159 = $154 - $tb$0$ph17$i$i;
         $160 = (+_SDL_fabs($159));
         $161 = $160 > 0.034906585039886591;
         $162 = $f2$011$i$i < $158;
         if ($161) {
          $x1$08$i$i$phi = $x2$010$i$i;$f1$09$i$i$phi = $f2$011$i$i;$340 = $162;$f2$011$i$i = $158;$ta$07$i$i = $154;$x2$010$i$i = $157;$x1$08$i$i = $x1$08$i$i$phi;$f1$09$i$i = $f1$09$i$i$phi;
         } else {
          $$lcssa$i$i = $162;$f1$0$lcssa$i$i = $f2$011$i$i;$f2$0$lcssa$i$i = $158;
          break L25;
         }
        }
        $144 = $x2$010$i$i$lcssa;
        $145 = $ta$07$i$i$lcssa * 0.61803398899999995;
        $146 = $144 * 0.38196601100000005;
        $147 = $146 + $145;
        $148 = $147;
        $149 = (+_dollarDifference($points$i,$137,$148));
        $150 = $ta$07$i$i$lcssa - $144;
        $151 = (+_SDL_fabs($150));
        $152 = $151 > 0.034906585039886591;
        $153 = $149 < $f1$09$i$i$lcssa;
        if ($152) {
         $339 = $153;$f1$0$ph20$i$i = $149;$f2$0$ph22$i$i = $f1$09$i$i$lcssa;$ta$0$ph18$i$i = $ta$07$i$i$lcssa;$tb$0$ph17$i$i = $144;$x1$0$ph19$i$i = $148;$x2$0$ph21$i$i = $x1$08$i$i$lcssa;
        } else {
         $$lcssa$i$i = $153;$f1$0$lcssa$i$i = $149;$f2$0$lcssa$i$i = $f1$09$i$i$lcssa;
         break;
        }
       }
      } else {
       $$lcssa$i$i = $142;$f1$0$lcssa$i$i = $138;$f2$0$lcssa$i$i = $139;
      }
     } while(0);
     $163 = $$lcssa$i$i ? $f1$0$lcssa$i$i : $f2$0$lcssa$i$i;
     $164 = $163 < $bestDiff$010$i;
     $bestTempl$1 = $164 ? $i$011$i : $bestTempl$0;
     $bestDiff$1$i = $164 ? $163 : $bestDiff$010$i;
     $165 = (($i$011$i) + 1)|0;
     $166 = HEAP32[$132>>2]|0;
     $167 = ($165|0)<($166|0);
     if ($167) {
      $bestDiff$010$i = $bestDiff$1$i;$bestTempl$0 = $bestTempl$1;$i$011$i = $165;
     } else {
      $bestDiff$1$i$lcssa = $bestDiff$1$i;$bestTempl$1$lcssa = $bestTempl$1;
      break;
     }
    }
    $168 = ($bestTempl$1$lcssa|0)>(-1);
    if ($168) {
     $169 = HEAP32[$135>>2]|0;
     $170 = (((($169) + (($bestTempl$1$lcssa*516)|0)|0)) + 512|0);
     $171 = HEAP32[$170>>2]|0;
     HEAP32[$event$i20>>2] = 2048;
     $172 = $$lcssa127;
     $173 = $172;
     $174 = HEAP32[$173>>2]|0;
     $175 = (($172) + 4)|0;
     $176 = $175;
     $177 = HEAP32[$176>>2]|0;
     $178 = ((($event$i20)) + 8|0);
     $179 = $178;
     $180 = $179;
     HEAP32[$180>>2] = $174;
     $181 = (($179) + 4)|0;
     $182 = $181;
     HEAP32[$182>>2] = $177;
     $183 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 8|0);
     $184 = HEAP32[$183>>2]|0;
     $185 = ((($event$i20)) + 32|0);
     HEAP32[$185>>2] = $184;
     $186 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 12|0);
     $187 = HEAP32[$186>>2]|0;
     $188 = ((($event$i20)) + 36|0);
     HEAP32[$188>>2] = $187;
     $189 = ((($event$i20)) + 16|0);
     $190 = $189;
     $191 = $190;
     HEAP32[$191>>2] = $171;
     $192 = (($190) + 4)|0;
     $193 = $192;
     HEAP32[$193>>2] = 0;
     $194 = ((($event$i20)) + 28|0);
     HEAPF32[$194>>2] = $bestDiff$1$i$lcssa;
     $195 = HEAP16[$28>>1]|0;
     $196 = $195&65535;
     $197 = (($196) + 1)|0;
     $198 = ((($event$i20)) + 24|0);
     HEAP32[$198>>2] = $197;
     (_SDL_PushEvent($event$i20)|0);
    }
   } else {
    HEAP32[$31>>2] = 0;
    $34 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 16|0);
    _dollarNormalize($34,$path);
    $35 = HEAP32[300438]|0;
    $36 = ($35|0)==(0);
    do {
     if ($36) {
      $76 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 8220|0);
      $77 = HEAP32[$76>>2]|0;
      $78 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 8224|0);
      $79 = HEAP32[$78>>2]|0;
      $80 = ($77*516)|0;
      $81 = (($80) + 516)|0;
      $82 = (_SDL_realloc($79,$81)|0);
      $83 = ($82|0)==(0|0);
      if ($83) {
       $84 = (_SDL_Error(0)|0);
       $index$0 = $84;
       break;
      }
      HEAP32[$78>>2] = $82;
      $85 = (($82) + (($77*516)|0)|0);
      (_SDL_memcpy($85,$path,512)|0);
      $hash$01$i$i12 = 5381;$i$02$i$i11 = 0;
      while(1) {
       $86 = ($hash$01$i$i12*33)|0;
       $87 = ((($82) + (($77*516)|0)|0) + ($i$02$i$i11<<3)|0);
       $88 = +HEAPF32[$87>>2];
       $89 = (~~(($88))>>>0);
       $90 = (($89) + ($86))|0;
       $91 = ($90*33)|0;
       $92 = ((((($82) + (($77*516)|0)|0) + ($i$02$i$i11<<3)|0)) + 4|0);
       $93 = +HEAPF32[$92>>2];
       $94 = (~~(($93))>>>0);
       $95 = (($91) + ($94))|0;
       $96 = (($i$02$i$i11) + 1)|0;
       $exitcond$i$i13 = ($96|0)==(64);
       if ($exitcond$i$i13) {
        $$lcssa120 = $95;
        break;
       } else {
        $hash$01$i$i12 = $95;$i$02$i$i11 = $96;
       }
      }
      $97 = (((($82) + (($77*516)|0)|0)) + 512|0);
      HEAP32[$97>>2] = $$lcssa120;
      $98 = HEAP32[$76>>2]|0;
      $99 = (($98) + 1)|0;
      HEAP32[$76>>2] = $99;
      $index$0 = $77;
     } else {
      $37 = HEAP32[300437]|0;
      $38 = ($37|0)==(0);
      L45: do {
       if ($38) {
        $40 = (_SDL_SetError(596668,$vararg_buffer)|0);
        $$0$i4$ph = $40;
        label = 20;
       } else {
        $39 = ($37|0)>(0);
        if ($39) {
         $i$01$i2 = 0;
         while(1) {
          $44 = HEAP32[300439]|0;
          $45 = (((($44) + (($i$01$i2*8232)|0)|0)) + 8220|0);
          $46 = HEAP32[$45>>2]|0;
          $47 = (((($44) + (($i$01$i2*8232)|0)|0)) + 8224|0);
          $48 = HEAP32[$47>>2]|0;
          $49 = ($46*516)|0;
          $50 = (($49) + 516)|0;
          $51 = (_SDL_realloc($48,$50)|0);
          $52 = ($51|0)==(0|0);
          if ($52) {
           $53 = (_SDL_Error(0)|0);
           $$0$i5 = $53;
          } else {
           HEAP32[$47>>2] = $51;
           $54 = (($51) + (($46*516)|0)|0);
           (_SDL_memcpy($54,$path,512)|0);
           $hash$01$i$i = 5381;$i$02$i$i = 0;
           while(1) {
            $55 = ($hash$01$i$i*33)|0;
            $56 = ((($51) + (($46*516)|0)|0) + ($i$02$i$i<<3)|0);
            $57 = +HEAPF32[$56>>2];
            $58 = (~~(($57))>>>0);
            $59 = (($58) + ($55))|0;
            $60 = ($59*33)|0;
            $61 = ((((($51) + (($46*516)|0)|0) + ($i$02$i$i<<3)|0)) + 4|0);
            $62 = +HEAPF32[$61>>2];
            $63 = (~~(($62))>>>0);
            $64 = (($60) + ($63))|0;
            $65 = (($i$02$i$i) + 1)|0;
            $exitcond$i$i = ($65|0)==(64);
            if ($exitcond$i$i) {
             $$lcssa121 = $64;
             break;
            } else {
             $hash$01$i$i = $64;$i$02$i$i = $65;
            }
           }
           $66 = (((($51) + (($46*516)|0)|0)) + 512|0);
           HEAP32[$66>>2] = $$lcssa121;
           $67 = HEAP32[$45>>2]|0;
           $68 = (($67) + 1)|0;
           HEAP32[$45>>2] = $68;
           $$0$i5 = $46;
          }
          $69 = ($$0$i5|0)<(0);
          $43 = (($i$01$i2) + 1)|0;
          if ($69) {
           $$0$i4$ph = -1;
           label = 20;
           break L45;
          }
          $41 = HEAP32[300437]|0;
          $42 = ($43|0)<($41|0);
          if ($42) {
           $i$01$i2 = $43;
          } else {
           $$0$i4 = $$0$i5;$71 = $41;
           break;
          }
         }
        } else {
         $$0$i4$ph = -1;
         label = 20;
        }
       }
      } while(0);
      if ((label|0) == 20) {
       $$pr = HEAP32[300437]|0;
       $$0$i4 = $$0$i4$ph;$71 = $$pr;
      }
      $70 = ($71|0)>(0);
      if ($70) {
       $72 = HEAP32[300439]|0;
       $i$043 = 0;
       while(1) {
        $73 = (((($72) + (($i$043*8232)|0)|0)) + 8228|0);
        HEAP32[$73>>2] = 0;
        $74 = (($i$043) + 1)|0;
        $75 = ($74|0)<($71|0);
        if ($75) {
         $i$043 = $74;
        } else {
         $index$0 = $$0$i4;
         break;
        }
       }
      } else {
       $index$0 = $$0$i4;
      }
     }
    } while(0);
    $100 = ($index$0|0)>(-1);
    $101 = $$lcssa127;
    $102 = $101;
    $103 = HEAP32[$102>>2]|0;
    $104 = (($101) + 4)|0;
    $105 = $104;
    $106 = HEAP32[$105>>2]|0;
    if ($100) {
     $107 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 8224|0);
     $108 = HEAP32[$107>>2]|0;
     $109 = (((($108) + (($index$0*516)|0)|0)) + 512|0);
     $110 = HEAP32[$109>>2]|0;
     HEAP32[$event$i17>>2] = 2049;
     $111 = ((($event$i17)) + 8|0);
     $112 = $111;
     $113 = $112;
     HEAP32[$113>>2] = $103;
     $114 = (($112) + 4)|0;
     $115 = $114;
     HEAP32[$115>>2] = $106;
     $116 = ((($event$i17)) + 16|0);
     $117 = $116;
     $118 = $117;
     HEAP32[$118>>2] = $110;
     $119 = (($117) + 4)|0;
     $120 = $119;
     HEAP32[$120>>2] = 0;
     (_SDL_PushEvent($event$i17)|0);
     break;
    } else {
     HEAP32[$event$i18>>2] = 2049;
     $121 = ((($event$i18)) + 8|0);
     $122 = $121;
     $123 = $122;
     HEAP32[$123>>2] = $103;
     $124 = (($122) + 4)|0;
     $125 = $124;
     HEAP32[$125>>2] = $106;
     $126 = ((($event$i18)) + 16|0);
     $127 = $126;
     $128 = $127;
     HEAP32[$128>>2] = -1;
     $129 = (($127) + 4)|0;
     $130 = $129;
     HEAP32[$130>>2] = -1;
     (_SDL_PushEvent($event$i18)|0);
     break;
    }
   }
  } while(0);
  $199 = HEAP16[$28>>1]|0;
  $200 = ($199<<16>>16)==(0);
  if (!($200)) {
   $201 = $199&65535;
   $202 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 8|0);
   $203 = +HEAPF32[$202>>2];
   $204 = (($201) + 1)|0;
   $205 = (+($204|0));
   $206 = $205 * $203;
   $207 = $206 - $25;
   $208 = (+($199&65535));
   $209 = $207 / $208;
   HEAPF32[$202>>2] = $209;
   $210 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 12|0);
   $211 = +HEAPF32[$210>>2];
   $212 = $211 * $205;
   $213 = $212 - $27;
   $214 = $213 / $208;
   HEAPF32[$210>>2] = $214;
  }
  STACKTOP = sp;return;
  break;
 }
 case 1794:  {
  $215 = ((($event)) + 32|0);
  $216 = +HEAPF32[$215>>2];
  $217 = ((($event)) + 36|0);
  $218 = +HEAPF32[$217>>2];
  $219 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 20|0);
  $220 = HEAP32[$219>>2]|0;
  $221 = ($220|0)<(1024);
  if ($221) {
   $222 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 8|0);
   $223 = HEAP32[$222>>2]|0;
   $224 = ((((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 24|0) + ($220<<3)|0);
   HEAP32[$224>>2] = $223;
   $225 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 12|0);
   $226 = HEAP32[$225>>2]|0;
   $227 = ((((((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 24|0) + ($220<<3)|0)) + 4|0);
   HEAP32[$227>>2] = $226;
   $228 = HEAP32[$219>>2]|0;
   $229 = ((((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 24|0) + ($228<<3)|0);
   $230 = +HEAPF32[$229>>2];
   $231 = (($228) + -1)|0;
   $232 = ((((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 24|0) + ($231<<3)|0);
   $233 = +HEAPF32[$232>>2];
   $234 = $230 - $233;
   $235 = ((((((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 24|0) + ($228<<3)|0)) + 4|0);
   $236 = +HEAPF32[$235>>2];
   $237 = ((((((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 24|0) + ($231<<3)|0)) + 4|0);
   $238 = +HEAPF32[$237>>2];
   $239 = $236 - $238;
   $240 = $234 * $234;
   $241 = $239 * $239;
   $242 = $240 + $241;
   $243 = $242;
   $244 = (+_SDL_sqrt($243));
   $245 = $244;
   $246 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 16|0);
   $247 = +HEAPF32[$246>>2];
   $248 = $247 + $245;
   HEAPF32[$246>>2] = $248;
   $249 = HEAP32[$219>>2]|0;
   $250 = (($249) + 1)|0;
   HEAP32[$219>>2] = $250;
   $$pre$phiZ2D = $225;
  } else {
   $$pre = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 12|0);
   $$pre$phiZ2D = $$pre;
  }
  $251 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 8|0);
  $252 = +HEAPF32[$251>>2];
  $253 = +HEAPF32[$$pre$phiZ2D>>2];
  $254 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 8216|0);
  $255 = HEAP16[$254>>1]|0;
  $256 = (+($255&65535));
  $257 = $216 / $256;
  $258 = $252 + $257;
  HEAPF32[$251>>2] = $258;
  $259 = $218 / $256;
  $260 = $253 + $259;
  HEAPF32[$$pre$phiZ2D>>2] = $260;
  $261 = ($255&65535)>(1);
  if (!($261)) {
   STACKTOP = sp;return;
  }
  $262 = $27 - $218;
  $263 = $25 - $216;
  $264 = $263 - $252;
  $265 = $262 - $253;
  $266 = $264 * $264;
  $267 = $265 * $265;
  $268 = $266 + $267;
  $269 = $268;
  $270 = (+_SDL_sqrt($269));
  $271 = $270;
  $272 = +HEAPF32[$251>>2];
  $273 = $25 - $272;
  $274 = +HEAPF32[$$pre$phiZ2D>>2];
  $275 = $27 - $274;
  $276 = $273 * $273;
  $277 = $275 * $275;
  $278 = $276 + $277;
  $279 = $278;
  $280 = (+_SDL_sqrt($279));
  $281 = $280;
  $282 = $264 / $271;
  $283 = $265 / $271;
  $284 = $273 / $281;
  $285 = $275 / $281;
  $286 = $282 * $285;
  $287 = $283 * $284;
  $288 = $286 - $287;
  $289 = $288;
  $290 = $282 * $284;
  $291 = $283 * $285;
  $292 = $290 + $291;
  $293 = $292;
  $294 = (+_SDL_atan2($289,$293));
  $295 = $294;
  $296 = $281 - $271;
  $297 = $271 == 0.0;
  $dDist$0 = $297 ? 0.0 : $296;
  $dtheta$0 = $297 ? 0.0 : $295;
  HEAP32[$event$i>>2] = 2050;
  $298 = $$lcssa127;
  $299 = $298;
  $300 = HEAP32[$299>>2]|0;
  $301 = (($298) + 4)|0;
  $302 = $301;
  $303 = HEAP32[$302>>2]|0;
  $304 = ((($event$i)) + 8|0);
  $305 = $304;
  $306 = $305;
  HEAP32[$306>>2] = $300;
  $307 = (($305) + 4)|0;
  $308 = $307;
  HEAP32[$308>>2] = $303;
  $309 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 8|0);
  $310 = HEAP32[$309>>2]|0;
  $311 = ((($event$i)) + 24|0);
  HEAP32[$311>>2] = $310;
  $312 = HEAP32[$$pre$phiZ2D>>2]|0;
  $313 = ((($event$i)) + 28|0);
  HEAP32[$313>>2] = $312;
  $314 = ((($event$i)) + 16|0);
  HEAPF32[$314>>2] = $dtheta$0;
  $315 = ((($event$i)) + 20|0);
  HEAPF32[$315>>2] = $dDist$0;
  $316 = HEAP16[$254>>1]|0;
  $317 = ((($event$i)) + 32|0);
  HEAP16[$317>>1] = $316;
  (_SDL_PushEvent($event$i)|0);
  STACKTOP = sp;return;
  break;
 }
 case 1792:  {
  $318 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 8216|0);
  $319 = HEAP16[$318>>1]|0;
  $320 = (($319) + 1)<<16>>16;
  HEAP16[$318>>1] = $320;
  $321 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 8|0);
  $322 = +HEAPF32[$321>>2];
  $323 = $320&65535;
  $324 = (($323) + -1)|0;
  $325 = (+($324|0));
  $326 = $322 * $325;
  $327 = $25 + $326;
  $328 = (+($320&65535));
  $329 = $327 / $328;
  HEAPF32[$321>>2] = $329;
  $330 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 12|0);
  $331 = +HEAPF32[$330>>2];
  $332 = $331 * $325;
  $333 = $27 + $332;
  $334 = $333 / $328;
  HEAPF32[$330>>2] = $334;
  $335 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 16|0);
  HEAPF32[$335>>2] = 0.0;
  $336 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 24|0);
  HEAPF32[$336>>2] = $25;
  $337 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 28|0);
  HEAPF32[$337>>2] = $27;
  $338 = (((($10) + (($i$01$i$lcssa*8232)|0)|0)) + 20|0);
  HEAP32[$338>>2] = 1;
  STACKTOP = sp;return;
  break;
 }
 default: {
  STACKTOP = sp;return;
 }
 }
}
function _dollarNormalize($path,$points) {
 $path = $path|0;
 $points = $points|0;
 var $$lcssa = 0.0, $$lcssa65 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0.0, $1 = 0, $10 = 0.0, $100 = 0.0, $101 = 0.0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0.0, $107 = 0.0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0.0;
 var $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0.0, $12 = 0.0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0.0, $126 = 0.0, $127 = 0, $128 = 0.0, $129 = 0.0, $13 = 0.0;
 var $130 = 0.0, $131 = 0.0, $132 = 0, $133 = 0.0, $134 = 0.0, $135 = 0.0, $136 = 0.0, $137 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0, $24 = 0.0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0.0, $40 = 0.0, $41 = 0.0, $42 = 0.0;
 var $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0.0;
 var $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0.0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0.0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0.0, $80 = 0, $81 = 0, $82 = 0, $83 = 0.0, $84 = 0.0, $85 = 0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0;
 var $98 = 0.0, $99 = 0, $centroid$sroa$0$016 = 0.0, $centroid$sroa$0$1$lcssa = 0.0, $centroid$sroa$0$1$lcssa$lcssa = 0.0, $centroid$sroa$0$19 = 0.0, $centroid$sroa$14$017 = 0.0, $centroid$sroa$14$1$lcssa = 0.0, $centroid$sroa$14$1$lcssa$lcssa = 0.0, $centroid$sroa$14$110 = 0.0, $dist$019 = 0.0, $dist$112 = 0.0, $exitcond = 0, $exitcond38 = 0, $i$029 = 0, $i$120 = 0, $i$28 = 0, $i$33 = 0, $length$028 = 0.0, $length$1 = 0.0;
 var $numPoints$0$lcssa39 = 0, $numPoints$018 = 0, $numPoints$1$lcssa = 0, $numPoints$1$lcssa$lcssa = 0, $numPoints$111 = 0, $vararg_buffer = 0, $xmax$06 = 0.0, $xmax$1 = 0.0, $xmax$1$lcssa = 0.0, $xmin$07 = 0.0, $xmin$1 = 0.0, $xmin$1$lcssa = 0.0, $ymax$05 = 0.0, $ymax$1 = 0.0, $ymax$1$lcssa = 0.0, $ymin$04 = 0.0, $ymin$1 = 0.0, $ymin$1$lcssa = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = +HEAPF32[$path>>2];
 $1 = !($0 <= 0.0);
 $$phi$trans$insert = ((($path)) + 4|0);
 $$pre = HEAP32[$$phi$trans$insert>>2]|0;
 if ($1) {
  $26 = $$pre;$length$1 = $0;
  label = 4;
 } else {
  $2 = ($$pre|0)>(1);
  if ($2) {
   $i$029 = 1;$length$028 = $0;
   while(1) {
    $3 = (((($path)) + 8|0) + ($i$029<<3)|0);
    $4 = +HEAPF32[$3>>2];
    $5 = (($i$029) + -1)|0;
    $6 = (((($path)) + 8|0) + ($5<<3)|0);
    $7 = +HEAPF32[$6>>2];
    $8 = $4 - $7;
    $9 = (((((($path)) + 8|0) + ($i$029<<3)|0)) + 4|0);
    $10 = +HEAPF32[$9>>2];
    $11 = (((((($path)) + 8|0) + ($5<<3)|0)) + 4|0);
    $12 = +HEAPF32[$11>>2];
    $13 = $10 - $12;
    $14 = $8 * $8;
    $15 = $13 * $13;
    $16 = $14 + $15;
    $17 = $16;
    $18 = (+_SDL_sqrt($17));
    $19 = $18;
    $20 = $length$028 + $19;
    $21 = (($i$029) + 1)|0;
    $22 = HEAP32[$$phi$trans$insert>>2]|0;
    $23 = ($21|0)<($22|0);
    if ($23) {
     $i$029 = $21;$length$028 = $20;
    } else {
     $26 = $22;$length$1 = $20;
     label = 4;
     break;
    }
   }
  } else {
   $numPoints$0$lcssa39 = 0;
  }
 }
 if ((label|0) == 4) {
  $24 = $length$1 / 63.0;
  $25 = ($26|0)>(1);
  if ($25) {
   $centroid$sroa$0$016 = 0.0;$centroid$sroa$14$017 = 0.0;$dist$019 = $24;$i$120 = 1;$numPoints$018 = 0;
   while(1) {
    $27 = (($i$120) + -1)|0;
    $28 = (((($path)) + 8|0) + ($27<<3)|0);
    $29 = +HEAPF32[$28>>2];
    $30 = (((($path)) + 8|0) + ($i$120<<3)|0);
    $31 = +HEAPF32[$30>>2];
    $32 = $29 - $31;
    $33 = $32 * $32;
    $34 = (((((($path)) + 8|0) + ($27<<3)|0)) + 4|0);
    $35 = +HEAPF32[$34>>2];
    $36 = (((((($path)) + 8|0) + ($i$120<<3)|0)) + 4|0);
    $37 = +HEAPF32[$36>>2];
    $38 = $35 - $37;
    $39 = $38 * $38;
    $40 = $33 + $39;
    $41 = $40;
    $42 = (+_SDL_sqrt($41));
    $43 = $42;
    $44 = $43 + $dist$019;
    $45 = $44 > $24;
    if ($45) {
     $centroid$sroa$0$19 = $centroid$sroa$0$016;$centroid$sroa$14$110 = $centroid$sroa$14$017;$dist$112 = $dist$019;$numPoints$111 = $numPoints$018;
     while(1) {
      $46 = +HEAPF32[$28>>2];
      $47 = $24 - $dist$112;
      $48 = $47 / $43;
      $49 = +HEAPF32[$30>>2];
      $50 = $49 - $46;
      $51 = $48 * $50;
      $52 = $46 + $51;
      $53 = (($points) + ($numPoints$111<<3)|0);
      HEAPF32[$53>>2] = $52;
      $54 = +HEAPF32[$34>>2];
      $55 = +HEAPF32[$36>>2];
      $56 = $55 - $54;
      $57 = $48 * $56;
      $58 = $54 + $57;
      $59 = (((($points) + ($numPoints$111<<3)|0)) + 4|0);
      HEAPF32[$59>>2] = $58;
      $60 = $centroid$sroa$0$19 + $52;
      $61 = $centroid$sroa$14$110 + $58;
      $62 = (($numPoints$111) + 1)|0;
      $63 = $dist$112 - $24;
      $64 = $43 + $63;
      $65 = $64 > $24;
      if ($65) {
       $centroid$sroa$0$19 = $60;$centroid$sroa$14$110 = $61;$dist$112 = $63;$numPoints$111 = $62;
      } else {
       $$lcssa = $64;$centroid$sroa$0$1$lcssa = $60;$centroid$sroa$14$1$lcssa = $61;$numPoints$1$lcssa = $62;
       break;
      }
     }
    } else {
     $$lcssa = $44;$centroid$sroa$0$1$lcssa = $centroid$sroa$0$016;$centroid$sroa$14$1$lcssa = $centroid$sroa$14$017;$numPoints$1$lcssa = $numPoints$018;
    }
    $66 = (($i$120) + 1)|0;
    $67 = HEAP32[$$phi$trans$insert>>2]|0;
    $68 = ($66|0)<($67|0);
    if ($68) {
     $centroid$sroa$0$016 = $centroid$sroa$0$1$lcssa;$centroid$sroa$14$017 = $centroid$sroa$14$1$lcssa;$dist$019 = $$lcssa;$i$120 = $66;$numPoints$018 = $numPoints$1$lcssa;
    } else {
     $$lcssa65 = $67;$centroid$sroa$0$1$lcssa$lcssa = $centroid$sroa$0$1$lcssa;$centroid$sroa$14$1$lcssa$lcssa = $centroid$sroa$14$1$lcssa;$numPoints$1$lcssa$lcssa = $numPoints$1$lcssa;
     break;
    }
   }
   $69 = ($numPoints$1$lcssa$lcssa|0)<(63);
   if ($69) {
    $numPoints$0$lcssa39 = $numPoints$1$lcssa$lcssa;
   } else {
    $70 = ((($points)) + 504|0);
    $71 = (($$lcssa65) + -1)|0;
    $72 = (((($path)) + 8|0) + ($71<<3)|0);
    $73 = $72;
    $74 = $73;
    $75 = HEAP32[$74>>2]|0;
    $76 = (($73) + 4)|0;
    $77 = $76;
    $78 = HEAP32[$77>>2]|0;
    $79 = $70;
    $80 = $79;
    HEAP32[$80>>2] = $75;
    $81 = (($79) + 4)|0;
    $82 = $81;
    HEAP32[$82>>2] = $78;
    $83 = $centroid$sroa$0$1$lcssa$lcssa * 0.015625;
    $84 = $centroid$sroa$14$1$lcssa$lcssa * 0.015625;
    $85 = ((($points)) + 4|0);
    $86 = +HEAPF32[$85>>2];
    $87 = $84 - $86;
    $88 = $87;
    $89 = +HEAPF32[$points>>2];
    $90 = $83 - $89;
    $91 = $90;
    $92 = (+_SDL_atan2($88,$91));
    $93 = $92;
    $94 = $93;
    $95 = $83;
    $96 = $84;
    $i$28 = 0;$xmax$06 = $83;$xmin$07 = $83;$ymax$05 = $84;$ymin$04 = $84;
    while(1) {
     $97 = (($points) + ($i$28<<3)|0);
     $98 = +HEAPF32[$97>>2];
     $99 = (((($points) + ($i$28<<3)|0)) + 4|0);
     $100 = +HEAPF32[$99>>2];
     $101 = $98 - $83;
     $102 = $101;
     $103 = (+_SDL_cos($94));
     $104 = $102 * $103;
     $105 = $100 - $84;
     $106 = $105;
     $107 = (+_SDL_sin($94));
     $108 = $106 * $107;
     $109 = $104 - $108;
     $110 = $95 + $109;
     $111 = $110;
     HEAPF32[$97>>2] = $111;
     $112 = (+_SDL_sin($94));
     $113 = $102 * $112;
     $114 = (+_SDL_cos($94));
     $115 = $106 * $114;
     $116 = $113 + $115;
     $117 = $96 + $116;
     $118 = $117;
     HEAPF32[$99>>2] = $118;
     $119 = +HEAPF32[$97>>2];
     $120 = $119 < $xmin$07;
     $xmin$1 = $120 ? $119 : $xmin$07;
     $121 = $119 > $xmax$06;
     $xmax$1 = $121 ? $119 : $xmax$06;
     $122 = $118 < $ymin$04;
     $ymin$1 = $122 ? $118 : $ymin$04;
     $123 = $118 > $ymax$05;
     $ymax$1 = $123 ? $118 : $ymax$05;
     $124 = (($i$28) + 1)|0;
     $exitcond38 = ($124|0)==(64);
     if ($exitcond38) {
      $xmax$1$lcssa = $xmax$1;$xmin$1$lcssa = $xmin$1;$ymax$1$lcssa = $ymax$1;$ymin$1$lcssa = $ymin$1;
      break;
     } else {
      $i$28 = $124;$xmax$06 = $xmax$1;$xmin$07 = $xmin$1;$ymax$05 = $ymax$1;$ymin$04 = $ymin$1;
     }
    }
    $125 = $xmax$1$lcssa - $xmin$1$lcssa;
    $126 = $ymax$1$lcssa - $ymin$1$lcssa;
    $i$33 = 0;
    while(1) {
     $127 = (($points) + ($i$33<<3)|0);
     $128 = +HEAPF32[$127>>2];
     $129 = $128 - $83;
     $130 = $129 * 256.0;
     $131 = $130 / $125;
     HEAPF32[$127>>2] = $131;
     $132 = (((($points) + ($i$33<<3)|0)) + 4|0);
     $133 = +HEAPF32[$132>>2];
     $134 = $133 - $84;
     $135 = $134 * 256.0;
     $136 = $135 / $126;
     HEAPF32[$132>>2] = $136;
     $137 = (($i$33) + 1)|0;
     $exitcond = ($137|0)==(64);
     if ($exitcond) {
      break;
     } else {
      $i$33 = $137;
     }
    }
    STACKTOP = sp;return;
   }
  } else {
   $numPoints$0$lcssa39 = 0;
  }
 }
 HEAP32[$vararg_buffer>>2] = $numPoints$0$lcssa39;
 (_SDL_SetError(596645,$vararg_buffer)|0);
 STACKTOP = sp;return;
}
function _SDL_sqrt($x) {
 $x = +$x;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+Math_sqrt((+$x)));
 return (+$0);
}
function _SDL_atan2($x,$y) {
 $x = +$x;
 $y = +$y;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+Math_atan2((+$x),(+$y)));
 return (+$0);
}
function _SDL_cos($x) {
 $x = +$x;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+Math_cos((+$x)));
 return (+$0);
}
function _SDL_sin($x) {
 $x = +$x;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+Math_sin((+$x)));
 return (+$0);
}
function _SDL_realloc($ptr,$size) {
 $ptr = $ptr|0;
 $size = $size|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_realloc($ptr,$size)|0);
 return ($0|0);
}
function _SDL_Error($code) {
 $code = $code|0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, $vararg_buffer9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer9 = sp + 40|0;
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 switch ($code|0) {
 case 0:  {
  (_SDL_SetError(596704,$vararg_buffer)|0);
  break;
 }
 case 1:  {
  (_SDL_SetError(596718,$vararg_buffer1)|0);
  break;
 }
 case 2:  {
  (_SDL_SetError(596748,$vararg_buffer3)|0);
  break;
 }
 case 3:  {
  (_SDL_SetError(596776,$vararg_buffer5)|0);
  break;
 }
 case 4:  {
  (_SDL_SetError(596804,$vararg_buffer7)|0);
  break;
 }
 default: {
  (_SDL_SetError(596836,$vararg_buffer9)|0);
 }
 }
 STACKTOP = sp;return -1;
}
function _SDL_memcpy($dst,$src,$len) {
 $dst = $dst|0;
 $src = $src|0;
 $len = $len|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _memcpy(($dst|0),($src|0),($len|0))|0;
 return ($dst|0);
}
function _dollarDifference($points,$templ,$ang) {
 $points = $points|0;
 $templ = $templ|0;
 $ang = +$ang;
 var $$lcssa = 0.0, $0 = 0.0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0;
 var $26 = 0.0, $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0.0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0.0, $4 = 0.0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $dist$01 = 0.0;
 var $exitcond = 0, $i$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $ang;
 $dist$01 = 0.0;$i$02 = 0;
 while(1) {
  $1 = (($points) + ($i$02<<3)|0);
  $2 = +HEAPF32[$1>>2];
  $3 = $2;
  $4 = (+_SDL_cos($0));
  $5 = $3 * $4;
  $6 = (((($points) + ($i$02<<3)|0)) + 4|0);
  $7 = +HEAPF32[$6>>2];
  $8 = $7;
  $9 = (+_SDL_sin($0));
  $10 = $8 * $9;
  $11 = $5 - $10;
  $12 = $11;
  $13 = +HEAPF32[$1>>2];
  $14 = $13;
  $15 = (+_SDL_sin($0));
  $16 = $14 * $15;
  $17 = +HEAPF32[$6>>2];
  $18 = $17;
  $19 = (+_SDL_cos($0));
  $20 = $18 * $19;
  $21 = $16 + $20;
  $22 = $21;
  $23 = (($templ) + ($i$02<<3)|0);
  $24 = +HEAPF32[$23>>2];
  $25 = $12 - $24;
  $26 = $25 * $25;
  $27 = (((($templ) + ($i$02<<3)|0)) + 4|0);
  $28 = +HEAPF32[$27>>2];
  $29 = $22 - $28;
  $30 = $29 * $29;
  $31 = $26 + $30;
  $32 = $31;
  $33 = (+_SDL_sqrt($32));
  $34 = $33;
  $35 = $dist$01 + $34;
  $36 = (($i$02) + 1)|0;
  $exitcond = ($36|0)==(64);
  if ($exitcond) {
   $$lcssa = $35;
   break;
  } else {
   $dist$01 = $35;$i$02 = $36;
  }
 }
 $37 = $$lcssa * 0.015625;
 return (+$37);
}
function _SDL_fabs($x) {
 $x = +$x;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+Math_abs((+$x)));
 return (+$0);
}
function _SDL_OnWindowFocusLost($window) {
 $window = $window|0;
 var $$pre3$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $grabbed$01$i = 0;
 var $not$$i = 0, $or$cond$i = 0, $or$cond7$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($window)) + 96|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[300441]|0;
  $4 = ((($3)) + 92|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ($5|0)==(0|0);
  if (!($6)) {
   $7 = ((($window)) + 100|0);
   $8 = HEAP32[$7>>2]|0;
   (FUNCTION_TABLE_iiii[$5 & 8191]($3,$window,$8)|0);
  }
 }
 $9 = (_SDL_GetMouse()|0);
 $10 = ((($9)) + 76|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($window)) + 48|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = $13 & 256;
 $15 = $14 | $11;
 $16 = ($15|0)==(0);
 $17 = $13 & 512;
 $18 = ($17|0)==(0);
 $or$cond7$i = $18 | $16;
 $19 = HEAP32[300441]|0;
 $20 = ((($19)) + 252|0);
 $21 = HEAP32[$20>>2]|0;
 if ($or$cond7$i) {
  $32 = ($21|0)==($window|0);
  if ($32) {
   HEAP32[$20>>2] = 0;
   $34 = $19;$grabbed$01$i = 0;
  } else {
   $34 = $19;$grabbed$01$i = 0;
  }
 } else {
  $22 = ($21|0)==(0|0);
  $23 = ($21|0)==($window|0);
  $or$cond$i = $22 | $23;
  if ($or$cond$i) {
   $31 = $19;
  } else {
   $24 = ((($21)) + 48|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = $25 & -257;
   HEAP32[$24>>2] = $26;
   $27 = ((($19)) + 100|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = ($28|0)==(0|0);
   if ($29) {
    $31 = $19;
   } else {
    FUNCTION_TABLE_viii[$28 & 8191]($19,$21,0);
    $$pre3$i = HEAP32[300441]|0;
    $31 = $$pre3$i;
   }
  }
  $30 = ((($31)) + 252|0);
  HEAP32[$30>>2] = $window;
  $34 = $31;$grabbed$01$i = 1;
 }
 $33 = ((($34)) + 100|0);
 $35 = HEAP32[$33>>2]|0;
 $36 = ($35|0)==(0|0);
 if (!($36)) {
  FUNCTION_TABLE_viii[$35 & 8191]($34,$window,$grabbed$01$i);
 }
 $37 = HEAP32[$12>>2]|0;
 $38 = $37 & 1;
 $39 = ($38|0)==(0);
 if ($39) {
  STACKTOP = sp;return;
 }
 $40 = ((($window)) + 116|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = ($41|0)==(0);
 if (!($42)) {
  STACKTOP = sp;return;
 }
 $43 = (_SDL_GetHint(604513)|0);
 $44 = ($43|0)==(0|0);
 if (!($44)) {
  $45 = HEAP8[$43>>0]|0;
  $not$$i = ($45<<24>>24)==(48);
  if ($not$$i) {
   STACKTOP = sp;return;
  }
 }
 $46 = HEAP32[300441]|0;
 $47 = ($46|0)==(0|0);
 if ($47) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $48 = HEAP32[$window>>2]|0;
 $49 = ((($46)) + 256|0);
 $50 = ($48|0)==($49|0);
 if (!($50)) {
  (_SDL_SetError(609414,$vararg_buffer1)|0);
  STACKTOP = sp;return;
 }
 $51 = HEAP32[$12>>2]|0;
 $52 = $51 & 64;
 $53 = ($52|0)==(0);
 if (!($53)) {
  STACKTOP = sp;return;
 }
 (_SDL_UpdateFullscreenMode($window,0)|0);
 $54 = HEAP32[300441]|0;
 $55 = ((($54)) + 76|0);
 $56 = HEAP32[$55>>2]|0;
 $57 = ($56|0)==(0|0);
 if ($57) {
  STACKTOP = sp;return;
 }
 FUNCTION_TABLE_vii[$56 & 8191]($54,$window);
 STACKTOP = sp;return;
}
function _SDL_GetHint($name) {
 $name = $name|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hint$0 = 0, $hint$02 = 0, $hint$03 = 0, $hint$03$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_getenv($name)|0);
 $hint$02 = HEAP32[300735]|0;
 $1 = ($hint$02|0)==(0|0);
 if ($1) {
  $$0 = $0;
  return ($$0|0);
 } else {
  $hint$03 = $hint$02;
 }
 while(1) {
  $2 = HEAP32[$hint$03>>2]|0;
  $3 = (_SDL_strcmp($name,$2)|0);
  $4 = ($3|0)==(0);
  if ($4) {
   $hint$03$lcssa = $hint$03;
   break;
  }
  $11 = ((($hint$03)) + 16|0);
  $hint$0 = HEAP32[$11>>2]|0;
  $12 = ($hint$0|0)==(0|0);
  if ($12) {
   $$0 = $0;
   label = 7;
   break;
  } else {
   $hint$03 = $hint$0;
  }
 }
 if ((label|0) == 7) {
  return ($$0|0);
 }
 $5 = ($0|0)==(0|0);
 if (!($5)) {
  $6 = ((($hint$03$lcssa)) + 8|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)==(2);
  if (!($8)) {
   $$0 = $0;
   return ($$0|0);
  }
 }
 $9 = ((($hint$03$lcssa)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 $$0 = $10;
 return ($$0|0);
}
function _SDL_getenv($name) {
 $name = $name|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($name|0)==(0|0);
 if ($0) {
  $$0 = 0;
 } else {
  $1 = (_SDL_strlen($name)|0);
  $2 = ($1|0)==(0);
  if ($2) {
   $$0 = 0;
  } else {
   $3 = (_getenv(($name|0))|0);
   $$0 = $3;
  }
 }
 return ($$0|0);
}
function _SDL_strcmp($str1,$str2) {
 $str1 = $str1|0;
 $str2 = $str2|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strcmp($str1,$str2)|0);
 return ($0|0);
}
function _RemovePendingResizedEvents($userdata,$event) {
 $userdata = $userdata|0;
 $event = $event|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$event>>2]|0;
 $1 = ($0|0)==(512);
 if ($1) {
  $2 = ((($event)) + 12|0);
  $3 = HEAP8[$2>>0]|0;
  $4 = ($3<<24>>24)==(5);
  if ($4) {
   $5 = ((($event)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ((($userdata)) + 8|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ($6|0)==($8|0);
   if ($9) {
    $$0 = 0;
    return ($$0|0);
   }
  }
 }
 $$0 = 1;
 return ($$0|0);
}
function _SDL_FilterEvents($filter,$userdata) {
 $filter = $filter|0;
 $userdata = $userdata|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $entry$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[31472]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  return;
 }
 $2 = (_SDL_LockMutex($0)|0);
 $3 = ($2|0)==(0);
 if (!($3)) {
  return;
 }
 $4 = HEAP32[(125904)>>2]|0;
 $5 = ($4|0)==(0|0);
 if (!($5)) {
  $entry$01 = $4;
  while(1) {
   $6 = ((($entry$01)) + 72|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = (FUNCTION_TABLE_iii[$filter & 8191]($userdata,$entry$01)|0);
   $9 = ($8|0)==(0);
   if ($9) {
    $11 = ((($entry$01)) + 68|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = ($12|0)==(0|0);
    $14 = $12;
    if (!($13)) {
     $15 = HEAP32[$6>>2]|0;
     $16 = ((($12)) + 72|0);
     HEAP32[$16>>2] = $15;
    }
    $17 = HEAP32[$6>>2]|0;
    $18 = ($17|0)==(0|0);
    $19 = $17;
    if (!($18)) {
     $20 = ((($17)) + 68|0);
     HEAP32[$20>>2] = $14;
    }
    $21 = HEAP32[(125904)>>2]|0;
    $22 = ($21|0)==($entry$01|0);
    if ($22) {
     HEAP32[(125904)>>2] = $19;
    }
    $23 = HEAP32[(125908)>>2]|0;
    $24 = ($23|0)==($entry$01|0);
    if ($24) {
     $25 = HEAP32[$11>>2]|0;
     HEAP32[(125908)>>2] = $25;
    }
    $26 = HEAP32[(125912)>>2]|0;
    HEAP32[$6>>2] = $26;
    HEAP32[(125912)>>2] = $entry$01;
    $27 = HEAP32[(125896)>>2]|0;
    $28 = (($27) + -1)|0;
    HEAP32[(125896)>>2] = $28;
   }
   $10 = ($7|0)==(0|0);
   if ($10) {
    break;
   } else {
    $entry$01 = $7;
   }
  }
 }
 return;
}
function _RemovePendingSizeChangedEvents($userdata,$event) {
 $userdata = $userdata|0;
 $event = $event|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$event>>2]|0;
 $1 = ($0|0)==(512);
 if ($1) {
  $2 = ((($event)) + 12|0);
  $3 = HEAP8[$2>>0]|0;
  $4 = ($3<<24>>24)==(6);
  if ($4) {
   $5 = ((($event)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ((($userdata)) + 8|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ($6|0)==($8|0);
   if ($9) {
    $$0 = 0;
    return ($$0|0);
   }
  }
 }
 $$0 = 1;
 return ($$0|0);
}
function _RemovePendingMoveEvents($userdata,$event) {
 $userdata = $userdata|0;
 $event = $event|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$event>>2]|0;
 $1 = ($0|0)==(512);
 if ($1) {
  $2 = ((($event)) + 12|0);
  $3 = HEAP8[$2>>0]|0;
  $4 = ($3<<24>>24)==(4);
  if ($4) {
   $5 = ((($event)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ((($userdata)) + 8|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ($6|0)==($8|0);
   if ($9) {
    $$0 = 0;
    return ($$0|0);
   }
  }
 }
 $$0 = 1;
 return ($$0|0);
}
function _SDL_SendQuit() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[300736] = 0;
 $0 = (_SDL_SendAppEvent(256)|0);
 return ($0|0);
}
function _SDL_SendAppEvent($eventType) {
 $eventType = $eventType|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $curr$0$i = 0, $curr$01$i = 0, $curr$02$i = 0, $event = 0, $posted$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $event = sp;
 $0 = $eventType >>> 8;
 $1 = $0 & 255;
 $2 = (1201888 + ($1<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  $5 = $eventType >>> 5;
  $6 = $5 & 7;
  $7 = (($3) + ($6<<2)|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = $eventType & 31;
  $10 = 1 << $9;
  $11 = $8 & $10;
  $12 = ($11|0)==(0);
  if (!($12)) {
   $posted$0 = 0;
   STACKTOP = sp;return ($posted$0|0);
  }
 }
 HEAP32[$event>>2] = $eventType;
 $13 = (_SDL_GetTicks()|0);
 $14 = ((($event)) + 4|0);
 HEAP32[$14>>2] = $13;
 $15 = HEAP32[300435]|0;
 $16 = ($15|0)==(0|0);
 if ($16) {
  label = 5;
 } else {
  $17 = HEAP32[300436]|0;
  $18 = (FUNCTION_TABLE_iii[$15 & 8191]($17,$event)|0);
  $19 = ($18|0)==(0);
  if ($19) {
   $$0$i = 0;
  } else {
   label = 5;
  }
 }
 if ((label|0) == 5) {
  $curr$01$i = HEAP32[300734]|0;
  $20 = ($curr$01$i|0)==(0|0);
  if (!($20)) {
   $curr$02$i = $curr$01$i;
   while(1) {
    $21 = HEAP32[$curr$02$i>>2]|0;
    $22 = ((($curr$02$i)) + 4|0);
    $23 = HEAP32[$22>>2]|0;
    (FUNCTION_TABLE_iii[$21 & 8191]($23,$event)|0);
    $24 = ((($curr$02$i)) + 8|0);
    $curr$0$i = HEAP32[$24>>2]|0;
    $25 = ($curr$0$i|0)==(0|0);
    if ($25) {
     break;
    } else {
     $curr$02$i = $curr$0$i;
    }
   }
  }
  $26 = (_SDL_PeepEvents($event,1,0,0,0)|0);
  $27 = ($26|0)<(1);
  if ($27) {
   $$0$i = 0;
  } else {
   _SDL_GestureProcessEvent($event);
   $$0$i = 1;
  }
 }
 $posted$0 = $$0$i;
 STACKTOP = sp;return ($posted$0|0);
}
function _SDL_GetMouseFocus() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(1201820)>>2]|0;
 return ($0|0);
}
function _SDL_GetMouseState($x,$y) {
 $x = $x|0;
 $y = $y|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($x|0)==(0|0);
 if (!($0)) {
  $1 = HEAP32[(1201824)>>2]|0;
  HEAP32[$x>>2] = $1;
 }
 $2 = ($y|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[(1201828)>>2]|0;
  HEAP32[$y>>2] = $3;
 }
 $4 = HEAP32[(1201848)>>2]|0;
 return ($4|0);
}
function _SDL_FreeSurface($surface) {
 $surface = $surface|0;
 var $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($surface|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = HEAP32[$surface>>2]|0;
 $2 = $1 & 4;
 $3 = ($2|0)==(0);
 if (!($3)) {
  return;
 }
 $4 = ((($surface)) + 56|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($5) + -1)|0;
 HEAP32[$4>>2] = $6;
 $7 = ($5|0)>(1);
 if ($7) {
  return;
 }
 $8 = ((($surface)) + 28|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)>(0);
 if ($10) {
  $12 = $9;
  while(1) {
   $11 = (($12) + -1)|0;
   HEAP32[$8>>2] = $11;
   $13 = ($12|0)>(1);
   if ($13) {
    $15 = $11;
   } else {
    $16 = HEAP32[$surface>>2]|0;
    $17 = $16 & 2;
    $18 = ($17|0)==(0);
    if ($18) {
     $15 = $11;
    } else {
     $19 = $16 & -3;
     HEAP32[$surface>>2] = $19;
     (_SDL_RLESurface($surface)|0);
     $$pre = HEAP32[$8>>2]|0;
     $15 = $$pre;
    }
   }
   $14 = ($15|0)>(0);
   if ($14) {
    $12 = $15;
   } else {
    break;
   }
  }
  $$pre1 = HEAP32[$surface>>2]|0;
  $21 = $$pre1;
 } else {
  $21 = $1;
 }
 $20 = $21 & 2;
 $22 = ($20|0)==(0);
 if (!($22)) {
  _SDL_UnRLESurface($surface,0);
 }
 $23 = ((($surface)) + 4|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ($24|0)==(0|0);
 if (!($25)) {
  $26 = (_SDL_SetPixelFormatPalette($24,0)|0);
  $27 = ($26|0)<(0);
  if (!($27)) {
   $28 = ((($surface)) + 52|0);
   $29 = HEAP32[$28>>2]|0;
   _SDL_InvalidateMap($29);
  }
  $30 = HEAP32[$23>>2]|0;
  _SDL_FreeFormat($30);
  HEAP32[$23>>2] = 0;
 }
 $31 = ((($surface)) + 52|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ($32|0)==(0|0);
 if (!($33)) {
  _SDL_FreeBlitMap($32);
  HEAP32[$31>>2] = 0;
 }
 $34 = HEAP32[$surface>>2]|0;
 $35 = $34 & 1;
 $36 = ($35|0)==(0);
 if ($36) {
  $37 = ((($surface)) + 20|0);
  $38 = HEAP32[$37>>2]|0;
  _SDL_free($38);
 }
 _SDL_free($surface);
 return;
}
function _SDL_RLESurface($surface) {
 $surface = $surface|0;
 var $$$i = 0, $$$i35 = 0, $$0 = 0, $$blankline$0$i = 0, $$blankline$0$i23 = 0, $$blankline$0$i23$lcssa = 0, $$lcssa$i = 0, $$lcssa$i21 = 0, $$lcssa7$i = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$i = 0, $$pre63 = 0, $$pre64 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0;
 var $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0;
 var $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0;
 var $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0;
 var $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0;
 var $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0;
 var $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0;
 var $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0;
 var $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0;
 var $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0;
 var $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0;
 var $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0;
 var $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0;
 var $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0;
 var $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $blankline$0$i = 0, $blankline$0$i18 = 0, $blankline$2$i = 0, $copy_opaque$1$i = 0, $copy_transl$0$i = 0, $copy_transl$1$i = 0, $dst$027$i = 0, $dst$054$i = 0, $dst$1$i = 0, $dst$1$i16 = 0, $dst$10$i = 0, $dst$2$lcssa$i = 0;
 var $dst$2$lcssa$i26 = 0, $dst$212$i = 0, $dst$212$us$i = 0, $dst$217$i = 0, $dst$3$i = 0, $dst$4$i = 0, $dst$4$i28 = 0, $dst$5$lastline$1$i = 0, $dst$5$lastline$1$i$lcssa = 0, $dst$5$lcssa$i = 0, $dst$5$lcssa$i$lcssa = 0, $dst$5$lcssa$i29 = 0, $dst$5$lcssa$i29$lcssa = 0, $dst$517$i = 0, $dst$517$us$i = 0, $dst$522$i = 0, $dst$6$i = 0, $dst$7$i = 0, $dst$7$i30 = 0, $dst$8$lcssa$i = 0;
 var $dst$841$i = 0, $dst$9$lastline$1$i = 0, $dst$9$lastline$1$i$lcssa = 0, $dst$9$lastline$1$i$lcssa$lcssa = 0, $dst$9$lcssa$i = 0, $dst$9$lcssa$i$lcssa = 0, $dst$947$i = 0, $exitcond$i = 0, $exitcond$i33 = 0, $lastline$0$lcssa$i = 0, $lastline$0$lcssa$i34 = 0, $lastline$024$i = 0, $lastline$057$i = 0, $lastline$1$i = 0, $lastline$1$i31 = 0, $maxsize$0$i = 0, $maxsize$0$i13 = 0, $or$cond$i = 0, $or$cond12 = 0, $or$cond2$i = 0;
 var $or$cond3$i = 0, $or$cond4$i = 0, $or$cond9 = 0, $run$018$i = 0, $run$018$us$i = 0, $run$024$i = 0, $run1$049$i = 0, $runstart$019$i = 0, $runstart$019$us$i = 0, $runstart$023$i = 0, $runstart$148$i = 0, $skip$0$lcssa$i = 0, $skip$0$lcssa$i25 = 0, $skip$013$i = 0, $skip$013$us$i = 0, $skip$018$i = 0, $skip2$0$lcssa$i = 0, $skip2$042$i = 0, $src$056$i = 0, $srcbuf$025$i = 0;
 var $switch$split2D = 0, $tmp$i = 0, $tmp2$i = 0, $x$0$i = 0, $x$0$i17 = 0, $x$1$lcssa$i = 0, $x$1$lcssa$i20 = 0, $x$13$i = 0, $x$18$i = 0, $x$2$lcssa$i = 0, $x$2$lcssa$i22 = 0, $x$211$i = 0, $x$26$i = 0, $x$3$i = 0, $x$4$lcssa$i = 0, $x$429$i = 0, $x$5$lcssa$i = 0, $x$535$i = 0, $y$026$i = 0, $y$055$i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$surface>>2]|0;
 $1 = $0 & 2;
 $2 = ($1|0)==(0);
 if (!($2)) {
  _SDL_UnRLESurface($surface,1);
 }
 $3 = ((($surface)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 8|0);
 $6 = HEAP16[$5>>1]|0;
 $7 = $6&255;
 $8 = ($7&255)<(8);
 $9 = ($6&65535) >>> 8;
 if ($8) {
  $$0 = -1;
  return ($$0|0);
 }
 $10 = ((($surface)) + 20|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0|0);
 if ($12) {
  $$0 = -1;
  return ($$0|0);
 }
 $13 = ((($surface)) + 52|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($14)) + 68|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 & 272;
 $18 = ($17|0)!=(0);
 $19 = $16 & 1;
 $20 = ($19|0)==(0);
 $or$cond12 = $18 & $20;
 if (!($or$cond12)) {
  $$0 = -1;
  return ($$0|0);
 }
 $21 = $16 & 2;
 $22 = ($21|0)==(0);
 do {
  if ($22) {
   $28 = $16 & 608;
   $29 = ($28|0)==(0);
   if ($29) {
    $$phi$trans$insert = ((($4)) + 24|0);
    $$pre = HEAP32[$$phi$trans$insert>>2]|0;
    $31 = $$pre;
    break;
   } else {
    $$0 = -1;
    return ($$0|0);
   }
  } else {
   $23 = ((($4)) + 24|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = $16 & 608;
   $26 = $25 | $24;
   $27 = ($26|0)==(0);
   if ($27) {
    $31 = $24;
   } else {
    $$0 = -1;
    return ($$0|0);
   }
  }
 } while(0);
 $30 = ($31|0)==(0);
 $32 = $16 & 16;
 $33 = ($32|0)==(0);
 $or$cond9 = $33 | $30;
 if ($or$cond9) {
  $34 = ((($14)) + 4|0);
  $35 = HEAP32[$34>>2]|0;
  $36 = ($35|0)==(0);
  if ($36) {
   $$0 = -1;
   return ($$0|0);
  }
  $37 = $9&65535;
  switch ($37|0) {
  case 1:  {
   $38 = ((($surface)) + 12|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = ($39*3)|0;
   $41 = ((($surface)) + 8|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = (($42|0) / 2)&-1;
   $44 = (($43) + 1)|0;
   $45 = Math_imul($40, $44)|0;
   $46 = (($45) + 2)|0;
   $maxsize$0$i = $46;
   break;
  }
  case 3: case 2:  {
   $47 = ((($surface)) + 12|0);
   $48 = HEAP32[$47>>2]|0;
   $49 = ((($surface)) + 8|0);
   $50 = HEAP32[$49>>2]|0;
   $51 = (($50|0) / 255)&-1;
   $52 = $51 << 1;
   $53 = Math_imul($50, $37)|0;
   $54 = (($53) + 2)|0;
   $55 = (($54) + ($52))|0;
   $56 = Math_imul($55, $48)|0;
   $57 = (($56) + 2)|0;
   $maxsize$0$i = $57;
   break;
  }
  case 4:  {
   $58 = ((($surface)) + 12|0);
   $59 = HEAP32[$58>>2]|0;
   $60 = ((($surface)) + 8|0);
   $61 = HEAP32[$60>>2]|0;
   $62 = (($61|0) / 65535)&-1;
   $tmp$i = (($62) + ($61))|0;
   $tmp2$i = $tmp$i << 2;
   $63 = (($tmp2$i) + 4)|0;
   $64 = Math_imul($63, $59)|0;
   $65 = (($64) + 4)|0;
   $maxsize$0$i = $65;
   break;
  }
  default: {
   $maxsize$0$i = 0;
  }
  }
  $66 = (_SDL_malloc($maxsize$0$i)|0);
  $67 = ($66|0)==(0|0);
  do {
   if ($67) {
    $177 = (_SDL_Error(0)|0);
    $178 = ($177|0)<(0);
    if ($178) {
     $$0 = -1;
     return ($$0|0);
    } else {
     $$pre64 = HEAP32[$13>>2]|0;
     $180 = $$pre64;
     break;
    }
   } else {
    $68 = ($9<<16>>16)==(4);
    $69 = $68 ? 65535 : 255;
    $70 = HEAP32[$3>>2]|0;
    $71 = ((($70)) + 24|0);
    $72 = HEAP32[$71>>2]|0;
    $73 = $72 ^ -1;
    $74 = HEAP32[$13>>2]|0;
    $75 = ((($74)) + 72|0);
    $76 = HEAP32[$75>>2]|0;
    $77 = $76 & $73;
    $78 = (($37) + -1)|0;
    $79 = (125960 + ($78<<2)|0);
    $80 = HEAP32[$79>>2]|0;
    $81 = ((($surface)) + 8|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = ((($surface)) + 12|0);
    $84 = HEAP32[$83>>2]|0;
    $85 = ($84|0)>(0);
    if ($85) {
     $86 = HEAP32[$10>>2]|0;
     $87 = $69&65535;
     $88 = ((($surface)) + 16|0);
     $dst$027$i = $66;$lastline$024$i = $66;$srcbuf$025$i = $86;$y$026$i = 0;
     while(1) {
      $blankline$0$i = 0;$dst$1$i = $dst$027$i;$lastline$1$i = $lastline$024$i;$x$0$i = 0;
      while(1) {
       $89 = ($x$0$i|0)<($82|0);
       L43: do {
        if ($89) {
         $x$13$i = $x$0$i;
         while(1) {
          $90 = Math_imul($x$13$i, $37)|0;
          $91 = (($srcbuf$025$i) + ($90)|0);
          $92 = (FUNCTION_TABLE_ii[$80 & 8191]($91)|0);
          $93 = $92 & $73;
          $94 = ($93|0)==($77|0);
          if (!($94)) {
           $x$1$lcssa$i = $x$13$i;
           break L43;
          }
          $96 = (($x$13$i) + 1)|0;
          $97 = ($96|0)<($82|0);
          if ($97) {
           $x$13$i = $96;
          } else {
           $x$1$lcssa$i = $96;
           break;
          }
         }
        } else {
         $x$1$lcssa$i = $x$0$i;
        }
       } while(0);
       $95 = ($x$1$lcssa$i|0)<($82|0);
       L48: do {
        if ($95) {
         $x$26$i = $x$1$lcssa$i;
         while(1) {
          $98 = Math_imul($x$26$i, $37)|0;
          $99 = (($srcbuf$025$i) + ($98)|0);
          $100 = (FUNCTION_TABLE_ii[$80 & 8191]($99)|0);
          $101 = $100 & $73;
          $102 = ($101|0)==($77|0);
          if ($102) {
           $$lcssa$i = 1;$x$2$lcssa$i = $x$26$i;
           break L48;
          }
          $103 = (($x$26$i) + 1)|0;
          $104 = ($103|0)<($82|0);
          if ($104) {
           $x$26$i = $103;
          } else {
           $$lcssa$i = 0;$x$2$lcssa$i = $103;
           break;
          }
         }
        } else {
         $$lcssa$i = 0;$x$2$lcssa$i = $x$1$lcssa$i;
        }
       } while(0);
       $105 = (($x$1$lcssa$i) - ($x$0$i))|0;
       $106 = ($105|0)==($82|0);
       $$blankline$0$i = $106 ? 1 : $blankline$0$i;
       $107 = (($x$2$lcssa$i) - ($x$1$lcssa$i))|0;
       $108 = ($105|0)>($69|0);
       if ($108) {
        if ($68) {
         $dst$212$us$i = $dst$1$i;$skip$013$us$i = $105;
         while(1) {
          HEAP16[$dst$212$us$i>>1] = $87;
          $109 = ((($dst$212$us$i)) + 2|0);
          HEAP16[$109>>1] = 0;
          $110 = ((($dst$212$us$i)) + 4|0);
          $111 = (($skip$013$us$i) - ($69))|0;
          $112 = ($111|0)>($69|0);
          if ($112) {
           $dst$212$us$i = $110;$skip$013$us$i = $111;
          } else {
           $dst$2$lcssa$i = $110;$skip$0$lcssa$i = $111;
           break;
          }
         }
        } else {
         $dst$212$i = $dst$1$i;$skip$013$i = $105;
         while(1) {
          HEAP8[$dst$212$i>>0] = -1;
          $113 = ((($dst$212$i)) + 1|0);
          HEAP8[$113>>0] = 0;
          $114 = ((($dst$212$i)) + 2|0);
          $115 = (($skip$013$i) - ($69))|0;
          $116 = ($115|0)>($69|0);
          if ($116) {
           $dst$212$i = $114;$skip$013$i = $115;
          } else {
           $dst$2$lcssa$i = $114;$skip$0$lcssa$i = $115;
           break;
          }
         }
        }
       } else {
        $dst$2$lcssa$i = $dst$1$i;$skip$0$lcssa$i = $105;
       }
       $117 = ($107|0)<($69|0);
       $118 = $117 ? $107 : $69;
       if ($68) {
        $119 = $skip$0$lcssa$i&65535;
        HEAP16[$dst$2$lcssa$i>>1] = $119;
        $120 = $118&65535;
        $121 = ((($dst$2$lcssa$i)) + 2|0);
        HEAP16[$121>>1] = $120;
        $122 = ((($dst$2$lcssa$i)) + 4|0);
        $dst$4$i = $122;
       } else {
        $123 = $skip$0$lcssa$i&255;
        HEAP8[$dst$2$lcssa$i>>0] = $123;
        $124 = $118&255;
        $125 = ((($dst$2$lcssa$i)) + 1|0);
        HEAP8[$125>>0] = $124;
        $126 = ((($dst$2$lcssa$i)) + 2|0);
        $dst$4$i = $126;
       }
       $127 = Math_imul($x$1$lcssa$i, $37)|0;
       $128 = (($srcbuf$025$i) + ($127)|0);
       $129 = Math_imul($118, $37)|0;
       (_SDL_memcpy($dst$4$i,$128,$129)|0);
       $130 = (($dst$4$i) + ($129)|0);
       $131 = (($107) - ($118))|0;
       $132 = (($118) + ($x$1$lcssa$i))|0;
       $133 = ($107|0)==($118|0);
       if ($133) {
        $dst$5$lcssa$i = $130;
       } else {
        if ($68) {
         $dst$517$us$i = $130;$run$018$us$i = $131;$runstart$019$us$i = $132;
         while(1) {
          $134 = ($run$018$us$i|0)<($69|0);
          $135 = $134 ? $run$018$us$i : $69;
          HEAP16[$dst$517$us$i>>1] = 0;
          $136 = $135&65535;
          $137 = ((($dst$517$us$i)) + 2|0);
          HEAP16[$137>>1] = $136;
          $138 = ((($dst$517$us$i)) + 4|0);
          $139 = Math_imul($runstart$019$us$i, $37)|0;
          $140 = (($srcbuf$025$i) + ($139)|0);
          $141 = Math_imul($135, $37)|0;
          (_SDL_memcpy($138,$140,$141)|0);
          $142 = (($138) + ($141)|0);
          $143 = (($135) + ($runstart$019$us$i))|0;
          $144 = (($run$018$us$i) - ($135))|0;
          $145 = ($run$018$us$i|0)==($135|0);
          if ($145) {
           $dst$5$lcssa$i = $142;
           break;
          } else {
           $dst$517$us$i = $142;$run$018$us$i = $144;$runstart$019$us$i = $143;
          }
         }
        } else {
         $dst$517$i = $130;$run$018$i = $131;$runstart$019$i = $132;
         while(1) {
          $146 = ($run$018$i|0)<($69|0);
          $147 = $146 ? $run$018$i : $69;
          HEAP8[$dst$517$i>>0] = 0;
          $148 = $147&255;
          $149 = ((($dst$517$i)) + 1|0);
          HEAP8[$149>>0] = $148;
          $150 = ((($dst$517$i)) + 2|0);
          $151 = Math_imul($runstart$019$i, $37)|0;
          $152 = (($srcbuf$025$i) + ($151)|0);
          $153 = Math_imul($147, $37)|0;
          (_SDL_memcpy($150,$152,$153)|0);
          $154 = (($150) + ($153)|0);
          $155 = (($147) + ($runstart$019$i))|0;
          $156 = (($run$018$i) - ($147))|0;
          $157 = ($run$018$i|0)==($147|0);
          if ($157) {
           $dst$5$lcssa$i = $154;
           break;
          } else {
           $dst$517$i = $154;$run$018$i = $156;$runstart$019$i = $155;
          }
         }
        }
       }
       $158 = ($$blankline$0$i|0)==(0);
       $dst$5$lastline$1$i = $158 ? $dst$5$lcssa$i : $lastline$1$i;
       if ($$lcssa$i) {
        $blankline$0$i = $$blankline$0$i;$dst$1$i = $dst$5$lcssa$i;$lastline$1$i = $dst$5$lastline$1$i;$x$0$i = $x$2$lcssa$i;
       } else {
        $dst$5$lastline$1$i$lcssa = $dst$5$lastline$1$i;$dst$5$lcssa$i$lcssa = $dst$5$lcssa$i;
        break;
       }
      }
      $159 = HEAP32[$88>>2]|0;
      $160 = (($srcbuf$025$i) + ($159)|0);
      $161 = (($y$026$i) + 1)|0;
      $exitcond$i = ($161|0)==($84|0);
      if ($exitcond$i) {
       $lastline$0$lcssa$i = $dst$5$lastline$1$i$lcssa;
       break;
      } else {
       $dst$027$i = $dst$5$lcssa$i$lcssa;$lastline$024$i = $dst$5$lastline$1$i$lcssa;$srcbuf$025$i = $160;$y$026$i = $161;
      }
     }
    } else {
     $lastline$0$lcssa$i = $66;
    }
    if ($68) {
     HEAP16[$lastline$0$lcssa$i>>1] = 0;
     $162 = ((($lastline$0$lcssa$i)) + 2|0);
     HEAP16[$162>>1] = 0;
     $163 = ((($lastline$0$lcssa$i)) + 4|0);
     $dst$7$i = $163;
    } else {
     HEAP8[$lastline$0$lcssa$i>>0] = 0;
     $164 = ((($lastline$0$lcssa$i)) + 1|0);
     HEAP8[$164>>0] = 0;
     $165 = ((($lastline$0$lcssa$i)) + 2|0);
     $dst$7$i = $165;
    }
    $166 = HEAP32[$surface>>2]|0;
    $167 = $166 & 1;
    $168 = ($167|0)==(0);
    if ($168) {
     $169 = HEAP32[$10>>2]|0;
     _SDL_free($169);
     HEAP32[$10>>2] = 0;
    }
    $170 = $dst$7$i;
    $171 = $66;
    $172 = (($170) - ($171))|0;
    $173 = (_SDL_realloc($66,$172)|0);
    $174 = ($173|0)==(0|0);
    $$$i = $174 ? $66 : $173;
    $175 = HEAP32[$13>>2]|0;
    $176 = ((($175)) + 12|0);
    HEAP32[$176>>2] = $$$i;
    $180 = $175;
   }
  } while(0);
  $179 = ((($180)) + 8|0);
  HEAP32[$179>>2] = 4480;
  $181 = ((($180)) + 68|0);
  $182 = HEAP32[$181>>2]|0;
  $183 = $182 | 8192;
  HEAP32[$181>>2] = $183;
 } else {
  $184 = HEAP32[$14>>2]|0;
  $185 = ($184|0)==(0|0);
  if ($185) {
   $$0 = -1;
   return ($$0|0);
  }
  $186 = ((($184)) + 4|0);
  $187 = HEAP32[$186>>2]|0;
  $188 = ($7<<24>>24)==(32);
  if (!($188)) {
   $$0 = -1;
   return ($$0|0);
  }
  $189 = ((($187)) + 12|0);
  $190 = HEAP32[$189>>2]|0;
  $191 = ((($187)) + 16|0);
  $192 = HEAP32[$191>>2]|0;
  $193 = $192 | $190;
  $194 = ((($187)) + 20|0);
  $195 = HEAP32[$194>>2]|0;
  $196 = $193 | $195;
  $197 = ((($187)) + 9|0);
  $198 = HEAP8[$197>>0]|0;
  $199 = $198&255;
  L90: do {
   switch ($199|0) {
   case 2:  {
    $switch$split2D = ($196|0)<(65535);
    if ($switch$split2D) {
     switch ($196|0) {
     case 32767:  {
      break;
     }
     default: {
      $$0 = -1;
      return ($$0|0);
     }
     }
     $203 = ($192|0)==(992);
     $204 = ($190|0)==(992);
     $or$cond3$i = $204 | $203;
     $205 = ($195|0)==(992);
     $or$cond4$i = $or$cond3$i | $205;
     if ($or$cond4$i) {
      $copy_transl$0$i = 4483;
     } else {
      $$0 = -1;
      return ($$0|0);
     }
    } else {
     switch ($196|0) {
     case 65535:  {
      break;
     }
     default: {
      $$0 = -1;
      return ($$0|0);
     }
     }
     $200 = ($192|0)==(2016);
     $201 = ($190|0)==(2016);
     $or$cond$i = $201 | $200;
     $202 = ($195|0)==(2016);
     $or$cond2$i = $or$cond$i | $202;
     if ($or$cond2$i) {
      $copy_transl$0$i = 4482;
     } else {
      $$0 = -1;
      return ($$0|0);
     }
    }
    $206 = ((($surface)) + 12|0);
    $207 = HEAP32[$206>>2]|0;
    $208 = ((($surface)) + 8|0);
    $209 = HEAP32[$208>>2]|0;
    $210 = ($209*6)|0;
    $211 = (($210) + 8)|0;
    $212 = Math_imul($211, $207)|0;
    $213 = (($212) + 2)|0;
    $copy_opaque$1$i = 4484;$copy_transl$1$i = $copy_transl$0$i;$maxsize$0$i13 = $213;
    break;
   }
   case 4:  {
    $214 = ($196|0)==(16777215);
    if ($214) {
     $215 = ((($surface)) + 12|0);
     $216 = HEAP32[$215>>2]|0;
     $217 = $216 << 3;
     $218 = ((($surface)) + 8|0);
     $219 = HEAP32[$218>>2]|0;
     $220 = (($219) + 1)|0;
     $221 = Math_imul($217, $220)|0;
     $222 = $221 | 4;
     $copy_opaque$1$i = 4485;$copy_transl$1$i = 4485;$maxsize$0$i13 = $222;
     break L90;
    } else {
     $$0 = -1;
     return ($$0|0);
    }
    break;
   }
   default: {
    $$0 = -1;
    return ($$0|0);
   }
   }
  } while(0);
  $223 = (($maxsize$0$i13) + 28)|0;
  $224 = (_SDL_malloc($223)|0);
  $225 = ($224|0)==(0|0);
  do {
   if ($225) {
    $425 = (_SDL_Error(0)|0);
    $426 = ($425|0)<(0);
    if ($426) {
     $$0 = -1;
     return ($$0|0);
    } else {
     $$pre63 = HEAP32[$13>>2]|0;
     $428 = $$pre63;
     break;
    }
   } else {
    $226 = HEAP8[$197>>0]|0;
    HEAP8[$224>>0] = $226;
    $227 = HEAP32[$189>>2]|0;
    $228 = ((($224)) + 4|0);
    HEAP32[$228>>2] = $227;
    $229 = HEAP32[$191>>2]|0;
    $230 = ((($224)) + 8|0);
    HEAP32[$230>>2] = $229;
    $231 = HEAP32[$194>>2]|0;
    $232 = ((($224)) + 12|0);
    HEAP32[$232>>2] = $231;
    $233 = ((($187)) + 24|0);
    $234 = HEAP32[$233>>2]|0;
    $235 = ((($224)) + 16|0);
    HEAP32[$235>>2] = $234;
    $236 = ((($187)) + 28|0);
    $237 = HEAP32[$236>>2]|0;
    $238 = $237&65535;
    $239 = $237&255;
    $240 = ((($224)) + 20|0);
    HEAP8[$240>>0] = $239;
    $241 = ($238&65535) >>> 8;
    $242 = $241&255;
    $243 = ((($224)) + 21|0);
    HEAP8[$243>>0] = $242;
    $244 = $237 >>> 16;
    $245 = $244&255;
    $246 = ((($224)) + 22|0);
    HEAP8[$246>>0] = $245;
    $247 = $237 >>> 24;
    $248 = $247&255;
    $249 = ((($224)) + 23|0);
    HEAP8[$249>>0] = $248;
    $250 = ((($187)) + 32|0);
    $251 = HEAP32[$250>>2]|0;
    $252 = $251&65535;
    $253 = $251&255;
    $254 = ((($224)) + 24|0);
    HEAP8[$254>>0] = $253;
    $255 = ($252&65535) >>> 8;
    $256 = $255&255;
    $257 = ((($224)) + 25|0);
    HEAP8[$257>>0] = $256;
    $258 = $251 >>> 16;
    $259 = $258&255;
    $260 = ((($224)) + 26|0);
    HEAP8[$260>>0] = $259;
    $261 = $251 >>> 24;
    $262 = $261&255;
    $263 = ((($224)) + 27|0);
    HEAP8[$263>>0] = $262;
    $264 = ((($224)) + 28|0);
    $265 = ((($surface)) + 12|0);
    $266 = HEAP32[$265>>2]|0;
    $267 = ((($surface)) + 8|0);
    $268 = HEAP32[$267>>2]|0;
    $269 = HEAP32[$3>>2]|0;
    $270 = ($266|0)>(0);
    if ($270) {
     $271 = HEAP32[$10>>2]|0;
     $272 = ((($269)) + 24|0);
     $273 = ((($269)) + 35|0);
     $274 = ((($surface)) + 16|0);
     $dst$054$i = $264;$lastline$057$i = $264;$src$056$i = $271;$y$055$i = 0;
     while(1) {
      $blankline$0$i18 = 0;$dst$1$i16 = $dst$054$i;$x$0$i17 = 0;
      while(1) {
       $275 = ($x$0$i17|0)<($268|0);
       L122: do {
        if ($275) {
         $276 = HEAP32[$272>>2]|0;
         $277 = HEAP8[$273>>0]|0;
         $278 = $277&255;
         $x$18$i = $x$0$i17;
         while(1) {
          $279 = (($src$056$i) + ($x$18$i<<2)|0);
          $280 = HEAP32[$279>>2]|0;
          $281 = $280 & $276;
          $282 = $281 >>> $278;
          $283 = ($282|0)==(255);
          if ($283) {
           $x$1$lcssa$i20 = $x$18$i;
           break L122;
          }
          $288 = (($x$18$i) + 1)|0;
          $289 = ($288|0)<($268|0);
          if ($289) {
           $x$18$i = $288;
          } else {
           $x$1$lcssa$i20 = $288;
           break;
          }
         }
        } else {
         $x$1$lcssa$i20 = $x$0$i17;
        }
       } while(0);
       $284 = ($x$1$lcssa$i20|0)<($268|0);
       L128: do {
        if ($284) {
         $285 = HEAP32[$272>>2]|0;
         $286 = HEAP8[$273>>0]|0;
         $287 = $286&255;
         $x$211$i = $x$1$lcssa$i20;
         while(1) {
          $290 = (($src$056$i) + ($x$211$i<<2)|0);
          $291 = HEAP32[$290>>2]|0;
          $292 = $291 & $285;
          $293 = $292 >>> $287;
          $294 = ($293|0)==(255);
          if (!($294)) {
           $$lcssa$i21 = 1;$x$2$lcssa$i22 = $x$211$i;
           break L128;
          }
          $295 = (($x$211$i) + 1)|0;
          $296 = ($295|0)<($268|0);
          if ($296) {
           $x$211$i = $295;
          } else {
           $$lcssa$i21 = 0;$x$2$lcssa$i22 = $295;
           break;
          }
         }
        } else {
         $$lcssa$i21 = 0;$x$2$lcssa$i22 = $x$1$lcssa$i20;
        }
       } while(0);
       $297 = (($x$1$lcssa$i20) - ($x$0$i17))|0;
       $298 = ($297|0)==($268|0);
       $$blankline$0$i23 = $298 ? 1 : $blankline$0$i18;
       $299 = (($x$2$lcssa$i22) - ($x$1$lcssa$i20))|0;
       $300 = ($297|0)>(255);
       if ($300) {
        $dst$217$i = $dst$1$i16;$skip$018$i = $297;
        while(1) {
         $301 = HEAP8[$197>>0]|0;
         $302 = ($301<<24>>24)==(4);
         if ($302) {
          HEAP16[$dst$217$i>>1] = 255;
          $303 = ((($dst$217$i)) + 2|0);
          HEAP16[$303>>1] = 0;
          $304 = ((($dst$217$i)) + 4|0);
          $dst$3$i = $304;
         } else {
          HEAP8[$dst$217$i>>0] = -1;
          $305 = ((($dst$217$i)) + 1|0);
          HEAP8[$305>>0] = 0;
          $306 = ((($dst$217$i)) + 2|0);
          $dst$3$i = $306;
         }
         $307 = (($skip$018$i) + -255)|0;
         $308 = ($307|0)>(255);
         if ($308) {
          $dst$217$i = $dst$3$i;$skip$018$i = $307;
         } else {
          $dst$2$lcssa$i26 = $dst$3$i;$skip$0$lcssa$i25 = $307;
          break;
         }
        }
       } else {
        $dst$2$lcssa$i26 = $dst$1$i16;$skip$0$lcssa$i25 = $297;
       }
       $309 = ($299|0)<(255);
       $310 = $309 ? $299 : 255;
       $311 = HEAP8[$197>>0]|0;
       $312 = ($311<<24>>24)==(4);
       if ($312) {
        $313 = $skip$0$lcssa$i25&65535;
        HEAP16[$dst$2$lcssa$i26>>1] = $313;
        $314 = $310&65535;
        $315 = ((($dst$2$lcssa$i26)) + 2|0);
        HEAP16[$315>>1] = $314;
        $316 = ((($dst$2$lcssa$i26)) + 4|0);
        $dst$4$i28 = $316;
       } else {
        $317 = $skip$0$lcssa$i25&255;
        HEAP8[$dst$2$lcssa$i26>>0] = $317;
        $318 = $310&255;
        $319 = ((($dst$2$lcssa$i26)) + 1|0);
        HEAP8[$319>>0] = $318;
        $320 = ((($dst$2$lcssa$i26)) + 2|0);
        $dst$4$i28 = $320;
       }
       $321 = (($src$056$i) + ($x$1$lcssa$i20<<2)|0);
       $322 = (FUNCTION_TABLE_iiiiii[$copy_opaque$1$i & 8191]($dst$4$i28,$321,$310,$269,$187)|0);
       $323 = (($dst$4$i28) + ($322)|0);
       $324 = ($299|0)==($310|0);
       if ($324) {
        $dst$5$lcssa$i29 = $323;
       } else {
        $325 = (($299) - ($310))|0;
        $326 = (($310) + ($x$1$lcssa$i20))|0;
        $dst$522$i = $323;$run$024$i = $325;$runstart$023$i = $326;
        while(1) {
         $327 = ($run$024$i|0)<(255);
         $328 = $327 ? $run$024$i : 255;
         $329 = HEAP8[$197>>0]|0;
         $330 = ($329<<24>>24)==(4);
         if ($330) {
          HEAP16[$dst$522$i>>1] = 0;
          $331 = $328&65535;
          $332 = ((($dst$522$i)) + 2|0);
          HEAP16[$332>>1] = $331;
          $333 = ((($dst$522$i)) + 4|0);
          $dst$6$i = $333;
         } else {
          HEAP8[$dst$522$i>>0] = 0;
          $334 = $328&255;
          $335 = ((($dst$522$i)) + 1|0);
          HEAP8[$335>>0] = $334;
          $336 = ((($dst$522$i)) + 2|0);
          $dst$6$i = $336;
         }
         $337 = (($src$056$i) + ($runstart$023$i<<2)|0);
         $338 = (FUNCTION_TABLE_iiiiii[$copy_opaque$1$i & 8191]($dst$6$i,$337,$328,$269,$187)|0);
         $339 = (($dst$6$i) + ($338)|0);
         $340 = (($328) + ($runstart$023$i))|0;
         $341 = (($run$024$i) - ($328))|0;
         $342 = ($run$024$i|0)==($328|0);
         if ($342) {
          $dst$5$lcssa$i29 = $339;
          break;
         } else {
          $dst$522$i = $339;$run$024$i = $341;$runstart$023$i = $340;
         }
        }
       }
       if ($$lcssa$i21) {
        $blankline$0$i18 = $$blankline$0$i23;$dst$1$i16 = $dst$5$lcssa$i29;$x$0$i17 = $x$2$lcssa$i22;
       } else {
        $$blankline$0$i23$lcssa = $$blankline$0$i23;$dst$5$lcssa$i29$lcssa = $dst$5$lcssa$i29;
        break;
       }
      }
      $343 = $dst$5$lcssa$i29$lcssa;
      $344 = $343 & 2;
      $345 = (($dst$5$lcssa$i29$lcssa) + ($344)|0);
      $blankline$2$i = $$blankline$0$i23$lcssa;$dst$7$i30 = $345;$lastline$1$i31 = $lastline$057$i;$x$3$i = 0;
      while(1) {
       $346 = ($x$3$i|0)<($268|0);
       L158: do {
        if ($346) {
         $347 = HEAP32[$272>>2]|0;
         $348 = HEAP8[$273>>0]|0;
         $349 = $348&255;
         $x$429$i = $x$3$i;
         while(1) {
          $350 = (($src$056$i) + ($x$429$i<<2)|0);
          $351 = HEAP32[$350>>2]|0;
          $352 = $351 & $347;
          $353 = $352 >>> $349;
          $354 = (($353) + -1)|0;
          $355 = ($354>>>0)>(253);
          if (!($355)) {
           $x$4$lcssa$i = $x$429$i;
           break L158;
          }
          $360 = (($x$429$i) + 1)|0;
          $361 = ($360|0)<($268|0);
          if ($361) {
           $x$429$i = $360;
          } else {
           $x$4$lcssa$i = $360;
           break;
          }
         }
        } else {
         $x$4$lcssa$i = $x$3$i;
        }
       } while(0);
       $356 = ($x$4$lcssa$i|0)<($268|0);
       L164: do {
        if ($356) {
         $357 = HEAP32[$272>>2]|0;
         $358 = HEAP8[$273>>0]|0;
         $359 = $358&255;
         $x$535$i = $x$4$lcssa$i;
         while(1) {
          $362 = (($src$056$i) + ($x$535$i<<2)|0);
          $363 = HEAP32[$362>>2]|0;
          $364 = $363 & $357;
          $365 = $364 >>> $359;
          $366 = (($365) + -1)|0;
          $367 = ($366>>>0)<(254);
          if (!($367)) {
           $$lcssa7$i = 1;$x$5$lcssa$i = $x$535$i;
           break L164;
          }
          $368 = (($x$535$i) + 1)|0;
          $369 = ($368|0)<($268|0);
          if ($369) {
           $x$535$i = $368;
          } else {
           $$lcssa7$i = 0;$x$5$lcssa$i = $368;
           break;
          }
         }
        } else {
         $$lcssa7$i = 0;$x$5$lcssa$i = $x$4$lcssa$i;
        }
       } while(0);
       $370 = (($x$4$lcssa$i) - ($x$3$i))|0;
       $371 = ($370|0)==($268|0);
       $372 = $371&1;
       $373 = $372 & $blankline$2$i;
       $374 = (($x$5$lcssa$i) - ($x$4$lcssa$i))|0;
       $375 = ($370|0)>(65535);
       if ($375) {
        $dst$841$i = $dst$7$i30;$skip2$042$i = $370;
        while(1) {
         HEAP16[$dst$841$i>>1] = -1;
         $376 = ((($dst$841$i)) + 2|0);
         HEAP16[$376>>1] = 0;
         $377 = ((($dst$841$i)) + 4|0);
         $378 = (($skip2$042$i) + -65535)|0;
         $379 = ($378|0)>(65535);
         if ($379) {
          $dst$841$i = $377;$skip2$042$i = $378;
         } else {
          $dst$8$lcssa$i = $377;$skip2$0$lcssa$i = $378;
          break;
         }
        }
       } else {
        $dst$8$lcssa$i = $dst$7$i30;$skip2$0$lcssa$i = $370;
       }
       $380 = ($374|0)<(65535);
       $381 = $380 ? $374 : 65535;
       $382 = $skip2$0$lcssa$i&65535;
       HEAP16[$dst$8$lcssa$i>>1] = $382;
       $383 = $381&65535;
       $384 = ((($dst$8$lcssa$i)) + 2|0);
       HEAP16[$384>>1] = $383;
       $385 = ((($dst$8$lcssa$i)) + 4|0);
       $386 = (($src$056$i) + ($x$4$lcssa$i<<2)|0);
       $387 = (FUNCTION_TABLE_iiiiii[$copy_transl$1$i & 8191]($385,$386,$381,$269,$187)|0);
       $388 = (($385) + ($387)|0);
       $389 = ($374|0)==($381|0);
       if ($389) {
        $dst$9$lcssa$i = $388;
       } else {
        $390 = (($374) - ($381))|0;
        $391 = (($381) + ($x$4$lcssa$i))|0;
        $dst$947$i = $388;$run1$049$i = $390;$runstart$148$i = $391;
        while(1) {
         $392 = ($run1$049$i|0)<(65535);
         $393 = $392 ? $run1$049$i : 65535;
         HEAP16[$dst$947$i>>1] = 0;
         $394 = $393&65535;
         $395 = ((($dst$947$i)) + 2|0);
         HEAP16[$395>>1] = $394;
         $396 = ((($dst$947$i)) + 4|0);
         $397 = (($src$056$i) + ($runstart$148$i<<2)|0);
         $398 = (FUNCTION_TABLE_iiiiii[$copy_transl$1$i & 8191]($396,$397,$393,$269,$187)|0);
         $399 = (($396) + ($398)|0);
         $400 = (($393) + ($runstart$148$i))|0;
         $401 = (($run1$049$i) - ($393))|0;
         $402 = ($run1$049$i|0)==($393|0);
         if ($402) {
          $dst$9$lcssa$i = $399;
          break;
         } else {
          $dst$947$i = $399;$run1$049$i = $401;$runstart$148$i = $400;
         }
        }
       }
       $403 = ($373|0)==(0);
       $dst$9$lastline$1$i = $403 ? $dst$9$lcssa$i : $lastline$1$i31;
       if ($$lcssa7$i) {
        $blankline$2$i = $373;$dst$7$i30 = $dst$9$lcssa$i;$lastline$1$i31 = $dst$9$lastline$1$i;$x$3$i = $x$5$lcssa$i;
       } else {
        $dst$9$lastline$1$i$lcssa = $dst$9$lastline$1$i;$dst$9$lcssa$i$lcssa = $dst$9$lcssa$i;
        break;
       }
      }
      $404 = HEAP32[$274>>2]|0;
      $405 = $404 >> 2;
      $406 = (($src$056$i) + ($405<<2)|0);
      $407 = (($y$055$i) + 1)|0;
      $exitcond$i33 = ($407|0)==($266|0);
      if ($exitcond$i33) {
       $dst$9$lastline$1$i$lcssa$lcssa = $dst$9$lastline$1$i$lcssa;
       break;
      } else {
       $dst$054$i = $dst$9$lcssa$i$lcssa;$lastline$057$i = $dst$9$lastline$1$i$lcssa;$src$056$i = $406;$y$055$i = $407;
      }
     }
     $$pre$i = HEAP8[$197>>0]|0;
     $409 = $$pre$i;$lastline$0$lcssa$i34 = $dst$9$lastline$1$i$lcssa$lcssa;
    } else {
     $409 = $226;$lastline$0$lcssa$i34 = $264;
    }
    $408 = ($409<<24>>24)==(4);
    if ($408) {
     HEAP16[$lastline$0$lcssa$i34>>1] = 0;
     $410 = ((($lastline$0$lcssa$i34)) + 2|0);
     HEAP16[$410>>1] = 0;
     $411 = ((($lastline$0$lcssa$i34)) + 4|0);
     $dst$10$i = $411;
    } else {
     HEAP8[$lastline$0$lcssa$i34>>0] = 0;
     $412 = ((($lastline$0$lcssa$i34)) + 1|0);
     HEAP8[$412>>0] = 0;
     $413 = ((($lastline$0$lcssa$i34)) + 2|0);
     $dst$10$i = $413;
    }
    $414 = HEAP32[$surface>>2]|0;
    $415 = $414 & 1;
    $416 = ($415|0)==(0);
    if ($416) {
     $417 = HEAP32[$10>>2]|0;
     _SDL_free($417);
     HEAP32[$10>>2] = 0;
    }
    $418 = $dst$10$i;
    $419 = $224;
    $420 = (($418) - ($419))|0;
    $421 = (_SDL_realloc($224,$420)|0);
    $422 = ($421|0)==(0|0);
    $$$i35 = $422 ? $224 : $421;
    $423 = HEAP32[$13>>2]|0;
    $424 = ((($423)) + 12|0);
    HEAP32[$424>>2] = $$$i35;
    $428 = $423;
   }
  } while(0);
  $427 = ((($428)) + 8|0);
  HEAP32[$427>>2] = 4481;
  $429 = ((($428)) + 68|0);
  $430 = HEAP32[$429>>2]|0;
  $431 = $430 | 16384;
  HEAP32[$429>>2] = $431;
 }
 $432 = HEAP32[$surface>>2]|0;
 $433 = $432 | 2;
 HEAP32[$surface>>2] = $433;
 $$0 = 0;
 return ($$0|0);
}
function _SDL_UnRLESurface($surface,$recode) {
 $surface = $surface|0;
 $recode = $recode|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$0$i = 0, $dst$0$us$i = 0, $full = 0, $ofs$0$i = 0, $ofs$0$us$us$i = 0, $ofs$1$i = 0, $ofs$1$us$us$i = 0, $ofs$2$i = 0, $ofs$2$us$us$i = 0, $ofs$4$i = 0, $ofs$4$us$i = 0, $ofs$5$i = 0, $ofs$5$us$i = 0, $or$cond = 0, $srcbuf$0$i = 0, $srcbuf$0$us$i = 0;
 var $srcbuf$1$i = 0, $srcbuf$1$us$us$i = 0, $srcbuf$3$i = 0, $srcbuf$3$us$us$i = 0, $srcbuf$3$us$us$i$lcssa = 0, $srcbuf$6$i = 0, $srcbuf$6$us$i = 0, $srcbuf$7$i = 0, $srcbuf$7$i$lcssa = 0, $srcbuf$7$us$i = 0, $srcbuf$7$us$i$lcssa = 0, $uncopy_opaque_16$uncopy_32$i = 0, $uncopy_transl_16$uncopy_32$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $full = sp;
 $0 = HEAP32[$surface>>2]|0;
 $1 = $0 & 2;
 $2 = ($1|0)==(0);
 if ($2) {
  STACKTOP = sp;return;
 }
 $3 = $0 & -3;
 HEAP32[$surface>>2] = $3;
 $4 = ($recode|0)!=(0);
 $5 = $0 & 1;
 $6 = ($5|0)==(0);
 $or$cond = $4 & $6;
 $7 = ((($surface)) + 52|0);
 L4: do {
  if ($or$cond) {
   $8 = HEAP32[$7>>2]|0;
   $9 = ((($8)) + 68|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = $10 & 8192;
   $12 = ($11|0)==(0);
   if (!($12)) {
    $13 = ((($surface)) + 12|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = ((($surface)) + 16|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = Math_imul($16, $14)|0;
    $18 = (_SDL_malloc($17)|0);
    $19 = ((($surface)) + 20|0);
    HEAP32[$19>>2] = $18;
    $20 = ($18|0)==(0|0);
    if (!($20)) {
     $23 = HEAP32[$7>>2]|0;
     $24 = ((($23)) + 72|0);
     $25 = HEAP32[$24>>2]|0;
     (_SDL_FillRect($surface,0,$25)|0);
     $26 = ((($full)) + 4|0);
     HEAP32[$26>>2] = 0;
     HEAP32[$full>>2] = 0;
     $27 = ((($surface)) + 8|0);
     $28 = HEAP32[$27>>2]|0;
     $29 = ((($full)) + 8|0);
     HEAP32[$29>>2] = $28;
     $30 = HEAP32[$13>>2]|0;
     $31 = ((($full)) + 12|0);
     HEAP32[$31>>2] = $30;
     (_SDL_RLEBlit($surface,$full,$surface,$full)|0);
     break;
    }
    $21 = HEAP32[$surface>>2]|0;
    $22 = $21 | 2;
    HEAP32[$surface>>2] = $22;
    STACKTOP = sp;return;
   }
   $32 = ((($surface)) + 4|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ((($8)) + 12|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = ((($surface)) + 8|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = HEAP8[$35>>0]|0;
   $39 = ($38<<24>>24)==(2);
   $uncopy_opaque_16$uncopy_32$i = $39 ? 4486 : 4487;
   $uncopy_transl_16$uncopy_32$i = $39 ? 4488 : 4487;
   $40 = ((($surface)) + 12|0);
   $41 = HEAP32[$40>>2]|0;
   $42 = ((($surface)) + 16|0);
   $43 = HEAP32[$42>>2]|0;
   $44 = Math_imul($43, $41)|0;
   $45 = (_SDL_malloc($44)|0);
   $46 = ((($surface)) + 20|0);
   HEAP32[$46>>2] = $45;
   $47 = ($45|0)==(0|0);
   if ($47) {
    $114 = HEAP32[$surface>>2]|0;
    $115 = $114 | 2;
    HEAP32[$surface>>2] = $115;
    STACKTOP = sp;return;
   }
   $48 = HEAP32[$40>>2]|0;
   $49 = HEAP32[$42>>2]|0;
   $50 = Math_imul($49, $48)|0;
   (_SDL_memset($45,0,$50)|0);
   $51 = HEAP32[$46>>2]|0;
   $52 = ((($35)) + 28|0);
   if ($39) {
    $dst$0$us$i = $51;$srcbuf$0$us$i = $52;
   } else {
    $dst$0$i = $51;$srcbuf$0$i = $52;
    while(1) {
     $ofs$0$i = 0;$srcbuf$1$i = $srcbuf$0$i;
     while(1) {
      $85 = HEAP16[$srcbuf$1$i>>1]|0;
      $86 = $85&65535;
      $87 = ((($srcbuf$1$i)) + 2|0);
      $88 = HEAP16[$87>>1]|0;
      $89 = $88&65535;
      $90 = ((($srcbuf$1$i)) + 4|0);
      $ofs$1$i = (($86) + ($ofs$0$i))|0;
      $91 = ($88<<16>>16)==(0);
      if ($91) {
       $96 = ($ofs$1$i|0)==(0);
       if ($96) {
        break L4;
       } else {
        $ofs$2$i = $ofs$1$i;$srcbuf$3$i = $90;
       }
      } else {
       $92 = (($dst$0$i) + ($ofs$1$i<<2)|0);
       $93 = (FUNCTION_TABLE_iiiiii[$uncopy_opaque_16$uncopy_32$i & 8191]($92,$90,$89,$35,$33)|0);
       $94 = (($90) + ($93)|0);
       $95 = (($ofs$1$i) + ($89))|0;
       $ofs$2$i = $95;$srcbuf$3$i = $94;
      }
      $97 = ($ofs$2$i|0)<($37|0);
      if ($97) {
       $ofs$0$i = $ofs$2$i;$srcbuf$1$i = $srcbuf$3$i;
      } else {
       $ofs$4$i = 0;$srcbuf$6$i = $srcbuf$3$i;
       break;
      }
     }
     while(1) {
      $98 = HEAP16[$srcbuf$6$i>>1]|0;
      $99 = $98&65535;
      $100 = (($99) + ($ofs$4$i))|0;
      $101 = ((($srcbuf$6$i)) + 2|0);
      $102 = HEAP16[$101>>1]|0;
      $103 = $102&65535;
      $104 = ((($srcbuf$6$i)) + 4|0);
      $105 = ($102<<16>>16)==(0);
      if ($105) {
       $ofs$5$i = $100;$srcbuf$7$i = $104;
      } else {
       $106 = (($dst$0$i) + ($100<<2)|0);
       $107 = (FUNCTION_TABLE_iiiiii[$uncopy_transl_16$uncopy_32$i & 8191]($106,$104,$103,$35,$33)|0);
       $108 = (($104) + ($107)|0);
       $109 = (($100) + ($103))|0;
       $ofs$5$i = $109;$srcbuf$7$i = $108;
      }
      $110 = ($ofs$5$i|0)<($37|0);
      if ($110) {
       $ofs$4$i = $ofs$5$i;$srcbuf$6$i = $srcbuf$7$i;
      } else {
       $srcbuf$7$i$lcssa = $srcbuf$7$i;
       break;
      }
     }
     $111 = HEAP32[$42>>2]|0;
     $112 = $111 >> 2;
     $113 = (($dst$0$i) + ($112<<2)|0);
     $dst$0$i = $113;$srcbuf$0$i = $srcbuf$7$i$lcssa;
    }
   }
   while(1) {
    $ofs$0$us$us$i = 0;$srcbuf$1$us$us$i = $srcbuf$0$us$i;
    while(1) {
     $69 = HEAP8[$srcbuf$1$us$us$i>>0]|0;
     $70 = $69&255;
     $71 = ((($srcbuf$1$us$us$i)) + 1|0);
     $72 = HEAP8[$71>>0]|0;
     $73 = $72&255;
     $74 = ((($srcbuf$1$us$us$i)) + 2|0);
     $ofs$1$us$us$i = (($70) + ($ofs$0$us$us$i))|0;
     $75 = ($72<<24>>24)==(0);
     if ($75) {
      $80 = ($ofs$1$us$us$i|0)==(0);
      if ($80) {
       break L4;
      } else {
       $ofs$2$us$us$i = $ofs$1$us$us$i;$srcbuf$3$us$us$i = $74;
      }
     } else {
      $76 = (($dst$0$us$i) + ($ofs$1$us$us$i<<2)|0);
      $77 = (FUNCTION_TABLE_iiiiii[$uncopy_opaque_16$uncopy_32$i & 8191]($76,$74,$73,$35,$33)|0);
      $78 = (($74) + ($77)|0);
      $79 = (($ofs$1$us$us$i) + ($73))|0;
      $ofs$2$us$us$i = $79;$srcbuf$3$us$us$i = $78;
     }
     $81 = ($ofs$2$us$us$i|0)<($37|0);
     if ($81) {
      $ofs$0$us$us$i = $ofs$2$us$us$i;$srcbuf$1$us$us$i = $srcbuf$3$us$us$i;
     } else {
      $srcbuf$3$us$us$i$lcssa = $srcbuf$3$us$us$i;
      break;
     }
    }
    $82 = $srcbuf$3$us$us$i$lcssa;
    $83 = $82 & 2;
    $84 = (($srcbuf$3$us$us$i$lcssa) + ($83)|0);
    $ofs$4$us$i = 0;$srcbuf$6$us$i = $84;
    while(1) {
     $53 = HEAP16[$srcbuf$6$us$i>>1]|0;
     $54 = $53&65535;
     $55 = (($54) + ($ofs$4$us$i))|0;
     $56 = ((($srcbuf$6$us$i)) + 2|0);
     $57 = HEAP16[$56>>1]|0;
     $58 = $57&65535;
     $59 = ((($srcbuf$6$us$i)) + 4|0);
     $60 = ($57<<16>>16)==(0);
     if ($60) {
      $ofs$5$us$i = $55;$srcbuf$7$us$i = $59;
     } else {
      $61 = (($dst$0$us$i) + ($55<<2)|0);
      $62 = (FUNCTION_TABLE_iiiiii[$uncopy_transl_16$uncopy_32$i & 8191]($61,$59,$58,$35,$33)|0);
      $63 = (($59) + ($62)|0);
      $64 = (($55) + ($58))|0;
      $ofs$5$us$i = $64;$srcbuf$7$us$i = $63;
     }
     $65 = ($ofs$5$us$i|0)<($37|0);
     if ($65) {
      $ofs$4$us$i = $ofs$5$us$i;$srcbuf$6$us$i = $srcbuf$7$us$i;
     } else {
      $srcbuf$7$us$i$lcssa = $srcbuf$7$us$i;
      break;
     }
    }
    $66 = HEAP32[$42>>2]|0;
    $67 = $66 >> 2;
    $68 = (($dst$0$us$i) + ($67<<2)|0);
    $dst$0$us$i = $68;$srcbuf$0$us$i = $srcbuf$7$us$i$lcssa;
   }
  }
 } while(0);
 $116 = HEAP32[$7>>2]|0;
 $117 = ((($116)) + 68|0);
 $118 = HEAP32[$117>>2]|0;
 $119 = $118 & -24577;
 HEAP32[$117>>2] = $119;
 $120 = ((($116)) + 12|0);
 $121 = HEAP32[$120>>2]|0;
 _SDL_free($121);
 $122 = HEAP32[$7>>2]|0;
 $123 = ((($122)) + 12|0);
 HEAP32[$123>>2] = 0;
 STACKTOP = sp;return;
}
function _SDL_FillRect($dst,$rect,$color) {
 $dst = $dst|0;
 $rect = $rect|0;
 $color = $color|0;
 var $$0 = 0, $$01 = 0, $$012$i = 0, $$012$i$us = 0, $$012$i22$us34 = 0, $$012$i22$us38 = 0, $$012$i22$us43 = 0, $$012$i4 = 0, $$012$us$i = 0, $$012$us$i8 = 0, $$012$us17$i = 0, $$012$us17$i$us = 0, $$012$us5$i = 0, $$03$i = 0, $$03$i$us = 0, $$03$us16$i = 0, $$03$us16$i$us = 0, $$03$us4$i = 0, $$04$i = 0, $$in$i = 0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $_n$0$i$us$i = 0, $_n$0$i$us$i13 = 0;
 var $_n$0$i$us7$i = 0, $_n$1$i$us$i = 0, $_n$1$i$us$i15 = 0, $_n$2$i$us$i = 0, $_n$2$i$us$i17 = 0, $_n$3$i$us$i = 0, $_n$3$i$us$i19 = 0, $_p$0$i$us$i = 0, $_p$0$i$us$i12 = 0, $_p$0$i$us6$i = 0, $_p$1$i$us$i = 0, $_p$1$i$us$i14 = 0, $_p$1$i$us20$i32 = 0, $_p$2$i$i103 = 0, $_p$2$i$us$i = 0, $_p$2$i$us$i16 = 0, $_p$3$i$i$us104 = 0, $_p$3$i$us$i = 0, $_p$3$i$us$i18 = 0, $clipped = 0;
 var $n$0$us$i = 0, $n$0$us$i10 = 0, $n$02$i = 0, $n$1$us$i = 0, $n$2$us$i = 0, $or$cond$i = 0, $p$0$us$i = 0, $p$0$us$i11 = 0, $p$03$i = 0, $p$1$us$i = 0, $p$2$us$i = 0, $p$4$us$i = 0, $p$5$us$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $clipped = sp + 24|0;
 $0 = ($dst|0)==(0|0);
 if ($0) {
  $1 = (_SDL_SetError(603923,$vararg_buffer)|0);
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ((($dst)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 8|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ($5&255)<(8);
 if ($6) {
  $7 = (_SDL_SetError(596854,$vararg_buffer1)|0);
  $$0 = $7;
  STACKTOP = sp;return ($$0|0);
 }
 $8 = ($rect|0)==(0|0);
 $9 = ((($dst)) + 36|0);
 if ($8) {
  $12 = ((($dst)) + 44|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ($13|0)<(1);
  if ($14) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
  $15 = ((($dst)) + 48|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ($16|0)<(1);
  if ($17) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  } else {
   $$01 = $9;
  }
 } else {
  $10 = (_SDL_IntersectRect($rect,$9,$clipped)|0);
  $11 = ($10|0)==(0);
  if ($11) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  } else {
   $$01 = $clipped;
  }
 }
 $18 = ((($dst)) + 20|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($19|0)==(0|0);
 if ($20) {
  $21 = (_SDL_SetError(596908,$vararg_buffer3)|0);
  $$0 = $21;
  STACKTOP = sp;return ($$0|0);
 }
 $22 = ((($$01)) + 4|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = ((($dst)) + 16|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = Math_imul($25, $23)|0;
 $27 = (($19) + ($26)|0);
 $28 = HEAP32[$$01>>2]|0;
 $29 = HEAP32[$2>>2]|0;
 $30 = ((($29)) + 9|0);
 $31 = HEAP8[$30>>0]|0;
 $32 = $31&255;
 $33 = Math_imul($32, $28)|0;
 $34 = (($27) + ($33)|0);
 switch ($32|0) {
 case 1:  {
  $35 = $color << 8;
  $36 = $35 | $color;
  $37 = $36 << 16;
  $38 = $37 | $36;
  $39 = ((($$01)) + 8|0);
  $40 = HEAP32[$39>>2]|0;
  $41 = ((($$01)) + 12|0);
  $42 = HEAP32[$41>>2]|0;
  $43 = (($42) + -1)|0;
  $44 = ($42|0)==(0);
  if ($44) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
  $45 = ($40|0)>(3);
  $46 = $color&255;
  $47 = (($40) + -1)|0;
  if ($45) {
   $$012$us$i8 = $34;$94 = $43;
  } else {
   $48 = $40 & 3;
   $49 = $40 & -4;
   switch ($48|0) {
   case 3:  {
    $$012$i22$us34 = $34;$55 = $43;
    while(1) {
     $50 = (($$012$i22$us34) + ($49)|0);
     $51 = ((($50)) + 1|0);
     HEAP8[$50>>0] = $46;
     $52 = ((($51)) + 1|0);
     HEAP8[$51>>0] = $46;
     HEAP8[$52>>0] = $46;
     $53 = (($$012$i22$us34) + ($25)|0);
     $54 = (($55) + -1)|0;
     $56 = ($55|0)==(0);
     if ($56) {
      $$0 = 0;
      break;
     } else {
      $$012$i22$us34 = $53;$55 = $54;
     }
    }
    STACKTOP = sp;return ($$0|0);
    break;
   }
   case 2:  {
    $$012$i22$us38 = $34;$61 = $43;
    while(1) {
     $57 = (($$012$i22$us38) + ($49)|0);
     $58 = ((($57)) + 1|0);
     HEAP8[$57>>0] = $46;
     HEAP8[$58>>0] = $46;
     $59 = (($$012$i22$us38) + ($25)|0);
     $60 = (($61) + -1)|0;
     $62 = ($61|0)==(0);
     if ($62) {
      $$0 = 0;
      break;
     } else {
      $$012$i22$us38 = $59;$61 = $60;
     }
    }
    STACKTOP = sp;return ($$0|0);
    break;
   }
   case 1:  {
    $$012$i22$us43 = $34;$66 = $43;
    while(1) {
     $63 = (($$012$i22$us43) + ($49)|0);
     HEAP8[$63>>0] = $46;
     $64 = (($$012$i22$us43) + ($25)|0);
     $65 = (($66) + -1)|0;
     $67 = ($66|0)==(0);
     if ($67) {
      $$0 = 0;
      break;
     } else {
      $$012$i22$us43 = $64;$66 = $65;
     }
    }
    STACKTOP = sp;return ($$0|0);
    break;
   }
   case 0:  {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
    break;
   }
   default: {
    // unreachable;
   }
   }
  }
  L41: while(1) {
   $68 = $$012$us$i8;
   $69 = $68 & 3;
   switch ($69|0) {
   case 1:  {
    $70 = ((($$012$us$i8)) + 1|0);
    HEAP8[$$012$us$i8>>0] = $46;
    $n$0$us$i10 = $47;$p$0$us$i11 = $70;
    label = 20;
    break;
   }
   case 2:  {
    $n$0$us$i10 = $40;$p$0$us$i11 = $$012$us$i8;
    label = 20;
    break;
   }
   case 3:  {
    $n$1$us$i = $40;$p$1$us$i = $$012$us$i8;
    label = 21;
    break;
   }
   default: {
    $n$2$us$i = $40;$p$2$us$i = $$012$us$i8;
   }
   }
   if ((label|0) == 20) {
    label = 0;
    $71 = ((($p$0$us$i11)) + 1|0);
    HEAP8[$p$0$us$i11>>0] = $46;
    $72 = (($n$0$us$i10) + -1)|0;
    $n$1$us$i = $72;$p$1$us$i = $71;
    label = 21;
   }
   if ((label|0) == 21) {
    label = 0;
    $73 = ((($p$1$us$i)) + 1|0);
    HEAP8[$p$1$us$i>>0] = $46;
    $74 = (($n$1$us$i) + -1)|0;
    $n$2$us$i = $74;$p$2$us$i = $73;
   }
   $75 = $n$2$us$i >> 2;
   $76 = (($75) + 3)|0;
   $77 = $76 >>> 2;
   $78 = ($75|0)==(0);
   if (!($78)) {
    $79 = $75 & 3;
    switch ($79|0) {
    case 0:  {
     $_n$0$i$us$i13 = $77;$_p$0$i$us$i12 = $p$2$us$i;
     label = 24;
     break;
    }
    case 3:  {
     $_n$1$i$us$i15 = $77;$_p$1$i$us$i14 = $p$2$us$i;
     label = 25;
     break;
    }
    case 2:  {
     $_n$2$i$us$i17 = $77;$_p$2$i$us$i16 = $p$2$us$i;
     label = 26;
     break;
    }
    case 1:  {
     $_n$3$i$us$i19 = $77;$_p$3$i$us$i18 = $p$2$us$i;
     label = 27;
     break;
    }
    default: {
     label = 34;
     break L41;
    }
    }
    while(1) {
     if ((label|0) == 24) {
      label = 0;
      $80 = ((($_p$0$i$us$i12)) + 4|0);
      HEAP32[$_p$0$i$us$i12>>2] = $38;
      $_n$1$i$us$i15 = $_n$0$i$us$i13;$_p$1$i$us$i14 = $80;
      label = 25;
      continue;
     }
     else if ((label|0) == 25) {
      label = 0;
      $81 = ((($_p$1$i$us$i14)) + 4|0);
      HEAP32[$_p$1$i$us$i14>>2] = $38;
      $_n$2$i$us$i17 = $_n$1$i$us$i15;$_p$2$i$us$i16 = $81;
      label = 26;
      continue;
     }
     else if ((label|0) == 26) {
      label = 0;
      $82 = ((($_p$2$i$us$i16)) + 4|0);
      HEAP32[$_p$2$i$us$i16>>2] = $38;
      $_n$3$i$us$i19 = $_n$2$i$us$i17;$_p$3$i$us$i18 = $82;
      label = 27;
      continue;
     }
     else if ((label|0) == 27) {
      label = 0;
      $83 = ((($_p$3$i$us$i18)) + 4|0);
      HEAP32[$_p$3$i$us$i18>>2] = $38;
      $84 = (($_n$3$i$us$i19) + -1)|0;
      $85 = ($84|0)==(0);
      if ($85) {
       break;
      } else {
       $_n$0$i$us$i13 = $84;$_p$0$i$us$i12 = $83;
       label = 24;
       continue;
      }
     }
    }
   }
   $86 = $n$2$us$i & 3;
   $87 = ($86|0)==(0);
   L59: do {
    if (!($87)) {
     $88 = $n$2$us$i & -4;
     $89 = (($p$2$us$i) + ($88)|0);
     switch ($86|0) {
     case 3:  {
      $90 = ((($89)) + 1|0);
      HEAP8[$89>>0] = $46;
      $p$4$us$i = $90;
      label = 31;
      break;
     }
     case 2:  {
      $p$4$us$i = $89;
      label = 31;
      break;
     }
     case 1:  {
      $p$5$us$i = $89;
      break;
     }
     default: {
      break L59;
     }
     }
     if ((label|0) == 31) {
      label = 0;
      $91 = ((($p$4$us$i)) + 1|0);
      HEAP8[$p$4$us$i>>0] = $46;
      $p$5$us$i = $91;
     }
     HEAP8[$p$5$us$i>>0] = $46;
    }
   } while(0);
   $92 = (($$012$us$i8) + ($25)|0);
   $93 = (($94) + -1)|0;
   $95 = ($94|0)==(0);
   if ($95) {
    $$0 = 0;
    label = 75;
    break;
   } else {
    $$012$us$i8 = $92;$94 = $93;
   }
  }
  if ((label|0) == 34) {
   // unreachable;
  }
  else if ((label|0) == 75) {
   STACKTOP = sp;return ($$0|0);
  }
  break;
 }
 case 2:  {
  $96 = $color << 16;
  $97 = $96 | $color;
  $98 = ((($$01)) + 8|0);
  $99 = HEAP32[$98>>2]|0;
  $100 = ((($$01)) + 12|0);
  $101 = HEAP32[$100>>2]|0;
  $102 = (($101) + -1)|0;
  $103 = ($101|0)==(0);
  if ($103) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
  $104 = ($99|0)>(1);
  $105 = $color&65535;
  $106 = (($99) + -1)|0;
  if ($104) {
   $$012$us$i = $34;$130 = $102;
  } else {
   $107 = $99 & 1;
   $108 = ($107|0)==(0);
   if ($108) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   } else {
    $$012$i4 = $34;$135 = $102;
   }
   while(1) {
    $132 = (($$012$i4) + ($106<<1)|0);
    HEAP16[$132>>1] = $105;
    $133 = (($$012$i4) + ($25)|0);
    $134 = (($135) + -1)|0;
    $136 = ($135|0)==(0);
    if ($136) {
     $$0 = 0;
     break;
    } else {
     $$012$i4 = $133;$135 = $134;
    }
   }
   STACKTOP = sp;return ($$0|0);
  }
  L81: while(1) {
   $109 = $$012$us$i;
   $110 = $109 & 2;
   $111 = ($110|0)==(0);
   if ($111) {
    $n$0$us$i = $99;$p$0$us$i = $$012$us$i;
   } else {
    $112 = ((($$012$us$i)) + 2|0);
    HEAP16[$$012$us$i>>1] = $105;
    $n$0$us$i = $106;$p$0$us$i = $112;
   }
   $113 = $n$0$us$i >> 1;
   $114 = (($113) + 3)|0;
   $115 = $114 >>> 2;
   $116 = ($113|0)==(0);
   if (!($116)) {
    $117 = $113 & 3;
    switch ($117|0) {
    case 0:  {
     $_n$0$i$us$i = $115;$_p$0$i$us$i = $p$0$us$i;
     label = 43;
     break;
    }
    case 3:  {
     $_n$1$i$us$i = $115;$_p$1$i$us$i = $p$0$us$i;
     label = 44;
     break;
    }
    case 2:  {
     $_n$2$i$us$i = $115;$_p$2$i$us$i = $p$0$us$i;
     label = 45;
     break;
    }
    case 1:  {
     $_n$3$i$us$i = $115;$_p$3$i$us$i = $p$0$us$i;
     label = 46;
     break;
    }
    default: {
     label = 50;
     break L81;
    }
    }
    while(1) {
     if ((label|0) == 43) {
      label = 0;
      $118 = ((($_p$0$i$us$i)) + 4|0);
      HEAP32[$_p$0$i$us$i>>2] = $97;
      $_n$1$i$us$i = $_n$0$i$us$i;$_p$1$i$us$i = $118;
      label = 44;
      continue;
     }
     else if ((label|0) == 44) {
      label = 0;
      $119 = ((($_p$1$i$us$i)) + 4|0);
      HEAP32[$_p$1$i$us$i>>2] = $97;
      $_n$2$i$us$i = $_n$1$i$us$i;$_p$2$i$us$i = $119;
      label = 45;
      continue;
     }
     else if ((label|0) == 45) {
      label = 0;
      $120 = ((($_p$2$i$us$i)) + 4|0);
      HEAP32[$_p$2$i$us$i>>2] = $97;
      $_n$3$i$us$i = $_n$2$i$us$i;$_p$3$i$us$i = $120;
      label = 46;
      continue;
     }
     else if ((label|0) == 46) {
      label = 0;
      $121 = ((($_p$3$i$us$i)) + 4|0);
      HEAP32[$_p$3$i$us$i>>2] = $97;
      $122 = (($_n$3$i$us$i) + -1)|0;
      $123 = ($122|0)==(0);
      if ($123) {
       break;
      } else {
       $_n$0$i$us$i = $122;$_p$0$i$us$i = $121;
       label = 43;
       continue;
      }
     }
    }
   }
   $124 = $n$0$us$i & 1;
   $125 = ($124|0)==(0);
   if (!($125)) {
    $126 = (($n$0$us$i) + -1)|0;
    $127 = (($p$0$us$i) + ($126<<1)|0);
    HEAP16[$127>>1] = $105;
   }
   $128 = (($$012$us$i) + ($25)|0);
   $129 = (($130) + -1)|0;
   $131 = ($130|0)==(0);
   if ($131) {
    $$0 = 0;
    label = 75;
    break;
   } else {
    $$012$us$i = $128;$130 = $129;
   }
  }
  if ((label|0) == 50) {
   // unreachable;
  }
  else if ((label|0) == 75) {
   STACKTOP = sp;return ($$0|0);
  }
  break;
 }
 case 3:  {
  $137 = ((($$01)) + 8|0);
  $138 = HEAP32[$137>>2]|0;
  $139 = ((($$01)) + 12|0);
  $140 = HEAP32[$139>>2]|0;
  $141 = $color&255;
  $142 = $color >>> 8;
  $143 = $142&255;
  $144 = $color >>> 16;
  $145 = $144&255;
  $146 = ($140|0)==(0);
  $147 = ($138|0)==(0);
  $or$cond$i = $147 | $146;
  if ($or$cond$i) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  } else {
   $$04$i = $34;$$in$i = $140;
  }
  while(1) {
   $n$02$i = $138;$p$03$i = $$04$i;
   while(1) {
    $148 = (($n$02$i) + -1)|0;
    $149 = ((($p$03$i)) + 1|0);
    HEAP8[$p$03$i>>0] = $141;
    $150 = ((($p$03$i)) + 2|0);
    HEAP8[$149>>0] = $143;
    $151 = ((($p$03$i)) + 3|0);
    HEAP8[$150>>0] = $145;
    $152 = ($148|0)==(0);
    if ($152) {
     break;
    } else {
     $n$02$i = $148;$p$03$i = $151;
    }
   }
   $153 = (($$in$i) + -1)|0;
   $154 = (($$04$i) + ($25)|0);
   $155 = ($153|0)==(0);
   if ($155) {
    $$0 = 0;
    break;
   } else {
    $$04$i = $154;$$in$i = $153;
   }
  }
  STACKTOP = sp;return ($$0|0);
  break;
 }
 case 4:  {
  $156 = ((($$01)) + 8|0);
  $157 = HEAP32[$156>>2]|0;
  $158 = ((($$01)) + 12|0);
  $159 = HEAP32[$158>>2]|0;
  $160 = ($159|0)==(0);
  if ($160) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
  $161 = (($157) + 3)|0;
  $162 = $161 >>> 2;
  $163 = ($157|0)==(0);
  if ($163) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
  $164 = $157 & 3;
  switch ($164|0) {
  case 0:  {
   $$012$us5$i = $34;$$03$us4$i = $159;
   while(1) {
    $_n$0$i$us7$i = $162;$_p$0$i$us6$i = $$012$us5$i;
    while(1) {
     $186 = ((($_p$0$i$us6$i)) + 4|0);
     HEAP32[$_p$0$i$us6$i>>2] = $color;
     $187 = ((($_p$0$i$us6$i)) + 8|0);
     HEAP32[$186>>2] = $color;
     $188 = ((($_p$0$i$us6$i)) + 12|0);
     HEAP32[$187>>2] = $color;
     $189 = ((($_p$0$i$us6$i)) + 16|0);
     HEAP32[$188>>2] = $color;
     $190 = (($_n$0$i$us7$i) + -1)|0;
     $191 = ($190|0)==(0);
     if ($191) {
      break;
     } else {
      $_n$0$i$us7$i = $190;$_p$0$i$us6$i = $189;
     }
    }
    $192 = (($$03$us4$i) + -1)|0;
    $193 = (($$012$us5$i) + ($25)|0);
    $194 = ($192|0)==(0);
    if ($194) {
     $$0 = 0;
     break;
    } else {
     $$012$us5$i = $193;$$03$us4$i = $192;
    }
   }
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 3:  {
   $169 = (($162) + -1)|0;
   $170 = ($169|0)==(0);
   if ($170) {
    $$012$us17$i$us = $34;$$03$us16$i$us = $159;
    while(1) {
     $171 = ((($$012$us17$i$us)) + 4|0);
     HEAP32[$$012$us17$i$us>>2] = $color;
     $172 = ((($$012$us17$i$us)) + 8|0);
     HEAP32[$171>>2] = $color;
     HEAP32[$172>>2] = $color;
     $173 = (($$03$us16$i$us) + -1)|0;
     $174 = (($$012$us17$i$us) + ($25)|0);
     $175 = ($173|0)==(0);
     if ($175) {
      $$0 = 0;
      break;
     } else {
      $$012$us17$i$us = $174;$$03$us16$i$us = $173;
     }
    }
    STACKTOP = sp;return ($$0|0);
   } else {
    $$012$us17$i = $34;$$03$us16$i = $159;
   }
   while(1) {
    $195 = ((($$012$us17$i)) + 4|0);
    HEAP32[$$012$us17$i>>2] = $color;
    $196 = ((($$012$us17$i)) + 8|0);
    HEAP32[$195>>2] = $color;
    HEAP32[$196>>2] = $color;
    $202 = $169;$_p$1$i$us20$i32 = $$012$us17$i;
    while(1) {
     $197 = ((($_p$1$i$us20$i32)) + 12|0);
     $198 = ((($_p$1$i$us20$i32)) + 16|0);
     HEAP32[$197>>2] = $color;
     $199 = ((($_p$1$i$us20$i32)) + 20|0);
     HEAP32[$198>>2] = $color;
     $200 = ((($_p$1$i$us20$i32)) + 24|0);
     HEAP32[$199>>2] = $color;
     HEAP32[$200>>2] = $color;
     $201 = (($202) + -1)|0;
     $203 = ($201|0)==(0);
     if ($203) {
      break;
     } else {
      $202 = $201;$_p$1$i$us20$i32 = $198;
     }
    }
    $204 = (($$03$us16$i) + -1)|0;
    $205 = (($$012$us17$i) + ($25)|0);
    $206 = ($204|0)==(0);
    if ($206) {
     $$0 = 0;
     break;
    } else {
     $$012$us17$i = $205;$$03$us16$i = $204;
    }
   }
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 1:  {
   $165 = (($162) + -1)|0;
   $166 = ($165|0)==(0);
   $$012$i$us = $34;$$03$i$us = $159;
   while(1) {
    HEAP32[$$012$i$us>>2] = $color;
    if (!($166)) {
     $181 = $165;$_p$3$i$i$us104 = $$012$i$us;
     while(1) {
      $176 = ((($_p$3$i$i$us104)) + 4|0);
      $177 = ((($_p$3$i$i$us104)) + 8|0);
      HEAP32[$176>>2] = $color;
      $178 = ((($_p$3$i$i$us104)) + 12|0);
      HEAP32[$177>>2] = $color;
      $179 = ((($_p$3$i$i$us104)) + 16|0);
      HEAP32[$178>>2] = $color;
      HEAP32[$179>>2] = $color;
      $180 = (($181) + -1)|0;
      $182 = ($180|0)==(0);
      if ($182) {
       break;
      } else {
       $181 = $180;$_p$3$i$i$us104 = $179;
      }
     }
    }
    $183 = (($$03$i$us) + -1)|0;
    $184 = (($$012$i$us) + ($25)|0);
    $185 = ($183|0)==(0);
    if ($185) {
     $$0 = 0;
     break;
    } else {
     $$012$i$us = $184;$$03$i$us = $183;
    }
   }
   STACKTOP = sp;return ($$0|0);
   break;
  }
  default: {
   $167 = (($162) + -1)|0;
   $168 = ($167|0)==(0);
   $$012$i = $34;$$03$i = $159;
   while(1) {
    $207 = ((($$012$i)) + 4|0);
    HEAP32[$$012$i>>2] = $color;
    HEAP32[$207>>2] = $color;
    if (!($168)) {
     $213 = $167;$_p$2$i$i103 = $$012$i;
     while(1) {
      $208 = ((($_p$2$i$i103)) + 8|0);
      $209 = ((($_p$2$i$i103)) + 12|0);
      HEAP32[$208>>2] = $color;
      $210 = ((($_p$2$i$i103)) + 16|0);
      HEAP32[$209>>2] = $color;
      $211 = ((($_p$2$i$i103)) + 20|0);
      HEAP32[$210>>2] = $color;
      HEAP32[$211>>2] = $color;
      $212 = (($213) + -1)|0;
      $214 = ($212|0)==(0);
      if ($214) {
       break;
      } else {
       $213 = $212;$_p$2$i$i103 = $210;
      }
     }
    }
    $215 = (($$03$i) + -1)|0;
    $216 = (($$012$i) + ($25)|0);
    $217 = ($215|0)==(0);
    if ($217) {
     $$0 = 0;
     break;
    } else {
     $$012$i = $216;$$03$i = $215;
    }
   }
   STACKTOP = sp;return ($$0|0);
  }
  }
  break;
 }
 default: {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 }
 return (0)|0;
}
function _SDL_IntersectRect($A,$B,$result) {
 $A = $A|0;
 $B = $B|0;
 $result = $result|0;
 var $$ = 0, $$0 = 0, $$1 = 0, $$demorgan = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $Amax$0 = 0, $Amax$1 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($A|0)==(0|0);
 if ($0) {
  HEAP32[$vararg_buffer>>2] = 596897;
  (_SDL_SetError(611267,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ($B|0)==(0|0);
 if ($1) {
  HEAP32[$vararg_buffer1>>2] = 596899;
  (_SDL_SetError(611267,$vararg_buffer1)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($result|0)==(0|0);
 if ($2) {
  HEAP32[$vararg_buffer4>>2] = 596901;
  (_SDL_SetError(611267,$vararg_buffer4)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ((($A)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)<(1);
 if (!($5)) {
  $6 = ((($A)) + 12|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)<(1);
  if (!($8)) {
   $9 = ((($B)) + 8|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($10|0)<(1);
   if (!($11)) {
    $12 = ((($B)) + 12|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ($13|0)<(1);
    if (!($14)) {
     $17 = HEAP32[$A>>2]|0;
     $18 = (($4) + ($17))|0;
     $19 = HEAP32[$B>>2]|0;
     $20 = (($10) + ($19))|0;
     $21 = ($19|0)>($17|0);
     $$ = $21 ? $19 : $17;
     HEAP32[$result>>2] = $$;
     $22 = ($20|0)<($18|0);
     $Amax$0 = $22 ? $20 : $18;
     $23 = (($Amax$0) - ($$))|0;
     $24 = ((($result)) + 8|0);
     HEAP32[$24>>2] = $23;
     $25 = ((($A)) + 4|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = (($7) + ($26))|0;
     $28 = ((($B)) + 4|0);
     $29 = HEAP32[$28>>2]|0;
     $30 = (($13) + ($29))|0;
     $31 = ($29|0)>($26|0);
     $$1 = $31 ? $29 : $26;
     $32 = ((($result)) + 4|0);
     HEAP32[$32>>2] = $$1;
     $33 = ($30|0)<($27|0);
     $Amax$1 = $33 ? $30 : $27;
     $34 = (($Amax$1) - ($$1))|0;
     $35 = ((($result)) + 12|0);
     HEAP32[$35>>2] = $34;
     $36 = ($23|0)<(1);
     $37 = ($34|0)<(1);
     $$demorgan = $36 | $37;
     $38 = $$demorgan&1;
     $39 = $38 ^ 1;
     $$0 = $39;
     STACKTOP = sp;return ($$0|0);
    }
   }
  }
 }
 $15 = ((($result)) + 8|0);
 HEAP32[$15>>2] = 0;
 $16 = ((($result)) + 12|0);
 HEAP32[$16>>2] = 0;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_RLEBlit($surf_src,$srcrect,$surf_dst,$dstrect) {
 $surf_src = $surf_src|0;
 $srcrect = $srcrect|0;
 $surf_dst = $surf_dst|0;
 $dstrect = $dstrect|0;
 var $$0 = 0, $$0$i = 0, $$0$ph$i = 0, $$013$ph$i = 0, $$1$i = 0, $$1$i$lcssa = 0, $$10$i = 0, $$10$i$lcssa = 0, $$12$i = 0, $$12$ph$i = 0, $$1225$ph$i = 0, $$13$i = 0, $$13$i$lcssa = 0, $$15$i = 0, $$15$ph$i = 0, $$1528$ph$i = 0, $$16$i = 0, $$16$i$lcssa = 0, $$18$i = 0, $$18$ph$i = 0;
 var $$1831$ph$i = 0, $$19$i = 0, $$19$i$lcssa = 0, $$21$i = 0, $$21$ph$i = 0, $$2134$ph$i = 0, $$22$i = 0, $$22$i$lcssa = 0, $$24$i = 0, $$24$ph$i = 0, $$2437$ph$i = 0, $$25$i = 0, $$25$i$lcssa = 0, $$27$i = 0, $$27$ph$i = 0, $$2740$ph$i = 0, $$28$i = 0, $$28$i$lcssa = 0, $$3$i = 0, $$3$ph$i = 0;
 var $$30$i = 0, $$30$ph$i = 0, $$3043$ph$i = 0, $$31$i = 0, $$31$i$lcssa = 0, $$316$ph$i = 0, $$33$i = 0, $$33$ph$i = 0, $$3346$ph$i = 0, $$34$i = 0, $$34$i$lcssa = 0, $$36$i = 0, $$36$ph$i = 0, $$3649$ph$i = 0, $$37$i = 0, $$37$i$lcssa = 0, $$4$i = 0, $$4$i$lcssa = 0, $$6$i = 0, $$6$ph$i = 0;
 var $$619$ph$i = 0, $$7$i = 0, $$7$i$lcssa = 0, $$9$i = 0, $$9$ph$i = 0, $$922$ph$i = 0, $$len$0$i = 0, $$len105$0$i = 0, $$len120$0$i = 0, $$len141$0$i = 0, $$len156$0$i = 0, $$len171$0$i = 0, $$len18$0$i = 0, $$len28$0$i = 0, $$len38$0$i = 0, $$len54$0$i = 0, $$len68$0$i = 0, $$len8$0$i = 0, $$len89$0$i = 0, $$pre = 0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0;
 var $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0;
 var $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0;
 var $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0;
 var $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0;
 var $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0;
 var $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0;
 var $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0;
 var $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0;
 var $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0;
 var $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0;
 var $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0;
 var $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0;
 var $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0;
 var $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0, $1267 = 0;
 var $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0, $1285 = 0;
 var $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0, $1302 = 0;
 var $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0, $1320 = 0;
 var $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0, $1339 = 0;
 var $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0, $1357 = 0;
 var $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0, $1375 = 0;
 var $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0, $1393 = 0;
 var $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0, $1410 = 0;
 var $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0, $1429 = 0;
 var $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0, $1447 = 0;
 var $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0, $1465 = 0;
 var $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0, $1483 = 0;
 var $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0, $1500 = 0;
 var $1501 = 0, $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0, $1511 = 0, $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0, $1518 = 0, $1519 = 0;
 var $152 = 0, $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0, $153 = 0, $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0, $1536 = 0, $1537 = 0;
 var $1538 = 0, $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0, $1548 = 0, $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0, $1554 = 0, $1555 = 0;
 var $1556 = 0, $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $1565 = 0, $1566 = 0, $1567 = 0, $1568 = 0, $1569 = 0, $157 = 0, $1570 = 0, $1571 = 0, $1572 = 0, $1573 = 0;
 var $1574 = 0, $1575 = 0, $1576 = 0, $1577 = 0, $1578 = 0, $1579 = 0, $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0, $1583 = 0, $1584 = 0, $1585 = 0, $1586 = 0, $1587 = 0, $1588 = 0, $1589 = 0, $159 = 0, $1590 = 0, $1591 = 0;
 var $1592 = 0, $1593 = 0, $1594 = 0, $1595 = 0, $1596 = 0, $1597 = 0, $1598 = 0, $1599 = 0, $16 = 0, $160 = 0, $1600 = 0, $1601 = 0, $1602 = 0, $1603 = 0, $1604 = 0, $1605 = 0, $1606 = 0, $1607 = 0, $1608 = 0, $1609 = 0;
 var $161 = 0, $1610 = 0, $1611 = 0, $1612 = 0, $1613 = 0, $1614 = 0, $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0, $1619 = 0, $162 = 0, $1620 = 0, $1621 = 0, $1622 = 0, $1623 = 0, $1624 = 0, $1625 = 0, $1626 = 0, $1627 = 0;
 var $1628 = 0, $1629 = 0, $163 = 0, $1630 = 0, $1631 = 0, $1632 = 0, $1633 = 0, $1634 = 0, $1635 = 0, $1636 = 0, $1637 = 0, $1638 = 0, $1639 = 0, $164 = 0, $1640 = 0, $1641 = 0, $1642 = 0, $1643 = 0, $1644 = 0, $1645 = 0;
 var $1646 = 0, $1647 = 0, $1648 = 0, $1649 = 0, $165 = 0, $1650 = 0, $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0, $1655 = 0, $1656 = 0, $1657 = 0, $1658 = 0, $1659 = 0, $166 = 0, $1660 = 0, $1661 = 0, $1662 = 0, $1663 = 0;
 var $1664 = 0, $1665 = 0, $1666 = 0, $1667 = 0, $1668 = 0, $1669 = 0, $167 = 0, $1670 = 0, $1671 = 0, $1672 = 0, $1673 = 0, $1674 = 0, $1675 = 0, $1676 = 0, $1677 = 0, $1678 = 0, $1679 = 0, $168 = 0, $1680 = 0, $1681 = 0;
 var $1682 = 0, $1683 = 0, $1684 = 0, $1685 = 0, $1686 = 0, $1687 = 0, $1688 = 0, $1689 = 0, $169 = 0, $1690 = 0, $1691 = 0, $1692 = 0, $1693 = 0, $1694 = 0, $1695 = 0, $1696 = 0, $1697 = 0, $1698 = 0, $1699 = 0, $17 = 0;
 var $170 = 0, $1700 = 0, $1701 = 0, $1702 = 0, $1703 = 0, $1704 = 0, $1705 = 0, $1706 = 0, $1707 = 0, $1708 = 0, $1709 = 0, $171 = 0, $1710 = 0, $1711 = 0, $1712 = 0, $1713 = 0, $1714 = 0, $1715 = 0, $1716 = 0, $1717 = 0;
 var $1718 = 0, $1719 = 0, $172 = 0, $1720 = 0, $1721 = 0, $1722 = 0, $1723 = 0, $1724 = 0, $1725 = 0, $1726 = 0, $1727 = 0, $1728 = 0, $1729 = 0, $173 = 0, $1730 = 0, $1731 = 0, $1732 = 0, $1733 = 0, $1734 = 0, $1735 = 0;
 var $1736 = 0, $1737 = 0, $1738 = 0, $1739 = 0, $174 = 0, $1740 = 0, $1741 = 0, $1742 = 0, $1743 = 0, $1744 = 0, $1745 = 0, $1746 = 0, $1747 = 0, $1748 = 0, $1749 = 0, $175 = 0, $1750 = 0, $1751 = 0, $1752 = 0, $1753 = 0;
 var $1754 = 0, $1755 = 0, $1756 = 0, $1757 = 0, $1758 = 0, $1759 = 0, $176 = 0, $1760 = 0, $1761 = 0, $1762 = 0, $1763 = 0, $1764 = 0, $1765 = 0, $1766 = 0, $1767 = 0, $1768 = 0, $1769 = 0, $177 = 0, $1770 = 0, $1771 = 0;
 var $1772 = 0, $1773 = 0, $1774 = 0, $1775 = 0, $1776 = 0, $1777 = 0, $1778 = 0, $1779 = 0, $178 = 0, $1780 = 0, $1781 = 0, $1782 = 0, $1783 = 0, $1784 = 0, $1785 = 0, $1786 = 0, $1787 = 0, $1788 = 0, $1789 = 0, $179 = 0;
 var $1790 = 0, $1791 = 0, $1792 = 0, $1793 = 0, $1794 = 0, $1795 = 0, $1796 = 0, $1797 = 0, $1798 = 0, $1799 = 0, $18 = 0, $180 = 0, $1800 = 0, $1801 = 0, $1802 = 0, $1803 = 0, $1804 = 0, $1805 = 0, $1806 = 0, $1807 = 0;
 var $1808 = 0, $1809 = 0, $181 = 0, $1810 = 0, $1811 = 0, $1812 = 0, $1813 = 0, $1814 = 0, $1815 = 0, $1816 = 0, $1817 = 0, $1818 = 0, $1819 = 0, $182 = 0, $1820 = 0, $1821 = 0, $1822 = 0, $1823 = 0, $1824 = 0, $1825 = 0;
 var $1826 = 0, $1827 = 0, $1828 = 0, $1829 = 0, $183 = 0, $1830 = 0, $1831 = 0, $1832 = 0, $1833 = 0, $1834 = 0, $1835 = 0, $1836 = 0, $1837 = 0, $1838 = 0, $1839 = 0, $184 = 0, $1840 = 0, $1841 = 0, $1842 = 0, $1843 = 0;
 var $1844 = 0, $1845 = 0, $1846 = 0, $1847 = 0, $1848 = 0, $1849 = 0, $185 = 0, $1850 = 0, $1851 = 0, $1852 = 0, $1853 = 0, $1854 = 0, $1855 = 0, $1856 = 0, $1857 = 0, $1858 = 0, $1859 = 0, $186 = 0, $1860 = 0, $1861 = 0;
 var $1862 = 0, $1863 = 0, $1864 = 0, $1865 = 0, $1866 = 0, $1867 = 0, $1868 = 0, $1869 = 0, $187 = 0, $1870 = 0, $1871 = 0, $1872 = 0, $1873 = 0, $1874 = 0, $1875 = 0, $1876 = 0, $1877 = 0, $1878 = 0, $1879 = 0, $188 = 0;
 var $1880 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0;
 var $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0;
 var $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0;
 var $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0;
 var $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0;
 var $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0;
 var $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0;
 var $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0;
 var $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0;
 var $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0;
 var $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0;
 var $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0;
 var $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0;
 var $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0;
 var $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0;
 var $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0;
 var $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0;
 var $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0;
 var $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0;
 var $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0;
 var $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0;
 var $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0;
 var $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0;
 var $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0;
 var $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0;
 var $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0;
 var $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0;
 var $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0;
 var $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0;
 var $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0;
 var $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0;
 var $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0;
 var $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0;
 var $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0;
 var $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0;
 var $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0;
 var $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0;
 var $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0;
 var $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0;
 var $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0;
 var $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0;
 var $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0;
 var $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0;
 var $999 = 0, $dst$0155 = 0, $dst$0205$i = 0, $dst$1$lcssa = 0, $dst$1$lcssa$i = 0, $dst$1$ph = 0, $dst$1$ph$i = 0, $dst$1159 = 0, $dst$1209$i = 0, $dst104$0139 = 0, $dst109$0203$i = 0, $dst113$0145 = 0, $dst124$0239$i = 0, $dst145$0192$i = 0, $dst160$0195$i = 0, $dst175$0199$i = 0, $dst37$0166 = 0, $dst46$0172 = 0, $dst46$1$lcssa = 0, $dst46$1$ph = 0;
 var $dst46$1176 = 0, $dst57$0215$i = 0, $dst61$0184 = 0, $dst71$0151 = 0, $dst72$0220$i = 0, $dst72$1$lcssa$i = 0, $dst72$1$ph$i = 0, $dst72$1224$i = 0, $dst80$0192 = 0, $dst93$0232$i = 0, $dst95$0136 = 0, $dstbuf$0$ph = 0, $dstbuf$12$ph = 0, $dstbuf$15$ph = 0, $dstbuf$18$ph = 0, $dstbuf$21$ph = 0, $dstbuf$24$ph = 0, $dstbuf$27$ph = 0, $dstbuf$3$ph = 0, $dstbuf$30$ph = 0;
 var $dstbuf$33$ph = 0, $dstbuf$36$ph = 0, $dstbuf$6$ph = 0, $dstbuf$9$ph = 0, $exitcond = 0, $exitcond259 = 0, $exitcond260 = 0, $exitcond261 = 0, $exitcond265 = 0, $exitcond270 = 0, $exitcond271 = 0, $i$0168 = 0, $i$0217$i = 0, $i102$0137 = 0, $i107$0201$i = 0, $i111$0143 = 0, $i122$0237$i = 0, $i143$0190$i = 0, $i158$0193$i = 0, $i173$0197$i = 0;
 var $i59$0186 = 0, $i69$0149 = 0, $i78$0190 = 0, $i91$0234$i = 0, $i93$0134 = 0, $len$0$i = 0, $len105$0$i = 0, $len120$0$i = 0, $len141$0$i = 0, $len156$0$i = 0, $len171$0$i = 0, $len18$0$i = 0, $len28$0$i = 0, $len38$0$i = 0, $len54$0$i = 0, $len68$0$i = 0, $len8$0$i = 0, $len89$0$i = 0, $linecount$0$ph = 0, $linecount$0$ph$i = 0;
 var $linecount108$0$ph = 0, $linecount11$0$ph = 0, $linecount114$0$ph$i = 0, $linecount12$0$ph$i = 0, $linecount135$0$ph$i = 0, $linecount15$0$ph = 0, $linecount150$0$ph$i = 0, $linecount165$0$ph$i = 0, $linecount19$0$ph = 0, $linecount2$0$ph$i = 0, $linecount22$0$ph$i = 0, $linecount23$0$ph = 0, $linecount32$0$ph$i = 0, $linecount33$0$ph = 0, $linecount41$0$ph = 0, $linecount48$0$ph$i = 0, $linecount56$0$ph = 0, $linecount62$0$ph$i = 0, $linecount66$0$ph = 0, $linecount75$0$ph = 0;
 var $linecount83$0$ph$i = 0, $linecount90$0$ph = 0, $linecount99$0$ph = 0, $linecount99$0$ph$i = 0, $n$0157 = 0, $n$0207$i = 0, $n$1$lcssa = 0, $n$1$lcssa$i = 0, $n$1$ph = 0, $n$1$ph$i = 0, $n$1161 = 0, $n$1211$i = 0, $n44$0174 = 0, $n44$1$lcssa = 0, $n44$1$ph = 0, $n44$1178 = 0, $n70$0222$i = 0, $n70$1$lcssa$i = 0, $n70$1$ph$i = 0, $n70$1226$i = 0;
 var $ofs$0 = 0, $ofs$0$i = 0, $ofs$1 = 0, $ofs$1$i = 0, $ofs$12 = 0, $ofs$13 = 0, $ofs$4 = 0, $ofs$5 = 0, $ofs$8 = 0, $ofs$9 = 0, $ofs100$0 = 0, $ofs100$0$i = 0, $ofs100$1 = 0, $ofs100$1$i = 0, $ofs109$0 = 0, $ofs109$1 = 0, $ofs115$0$i = 0, $ofs115$1$i = 0, $ofs12$0 = 0, $ofs12$1 = 0;
 var $ofs13$0$i = 0, $ofs13$1$i = 0, $ofs136$0$i = 0, $ofs136$1$i = 0, $ofs151$0$i = 0, $ofs151$1$i = 0, $ofs16$0 = 0, $ofs16$1 = 0, $ofs166$0$i = 0, $ofs166$1$i = 0, $ofs20$0 = 0, $ofs20$1 = 0, $ofs23$0$i = 0, $ofs23$1$i = 0, $ofs24$0 = 0, $ofs24$1 = 0, $ofs3$0$i = 0, $ofs3$1$i = 0, $ofs33$0$i = 0, $ofs33$1$i = 0;
 var $ofs34$0 = 0, $ofs34$1 = 0, $ofs42$0 = 0, $ofs42$1 = 0, $ofs49$0$i = 0, $ofs49$1$i = 0, $ofs57$0 = 0, $ofs57$1 = 0, $ofs63$0$i = 0, $ofs63$1$i = 0, $ofs67$0 = 0, $ofs67$1 = 0, $ofs76$0 = 0, $ofs76$1 = 0, $ofs8$0 = 0, $ofs8$1 = 0, $ofs84$0$i = 0, $ofs84$1$i = 0, $ofs91$0 = 0, $ofs91$1 = 0;
 var $or$cond = 0, $or$cond$i = 0, $or$cond14 = 0, $or$cond15 = 0, $or$cond16 = 0, $or$cond17 = 0, $or$cond18 = 0, $or$cond78$i = 0, $or$cond79$i = 0, $or$cond80$i = 0, $or$cond81$i = 0, $or$cond82$i = 0, $scevgep = 0, $scevgep$i = 0, $scevgep262 = 0, $scevgep263 = 0, $scevgep264 = 0, $scevgep266 = 0, $scevgep267 = 0, $scevgep268 = 0;
 var $scevgep269 = 0, $scevgep272$i = 0, $scevgep273$i = 0, $scevgep274$i = 0, $scevgep275$i = 0, $scevgep276$i = 0, $scevgep277$i = 0, $scevgep278$i = 0, $src$0156 = 0, $src$0206$i = 0, $src$1$lcssa = 0, $src$1$lcssa$i = 0, $src$1$ph = 0, $src$1$ph$i = 0, $src$1160 = 0, $src$1210$i = 0, $src103$0138 = 0, $src108$0202$i = 0, $src112$0144 = 0, $src123$0238$i = 0;
 var $src144$0191$i = 0, $src159$0194$i = 0, $src174$0198$i = 0, $src36$0167 = 0, $src45$0173 = 0, $src45$1$lcssa = 0, $src45$1$ph = 0, $src45$1177 = 0, $src56$0216$i = 0, $src60$0185 = 0, $src70$0150 = 0, $src71$0221$i = 0, $src71$1$lcssa$i = 0, $src71$1$ph$i = 0, $src71$1225$i = 0, $src79$0191 = 0, $src92$0233$i = 0, $src94$0135 = 0, $srcbuf$0 = 0, $srcbuf$0$ph = 0;
 var $srcbuf$1 = 0, $srcbuf$1$lcssa = 0, $srcbuf$10 = 0, $srcbuf$10$lcssa = 0, $srcbuf$13$ph = 0, $srcbuf$14 = 0, $srcbuf$14$ph = 0, $srcbuf$15 = 0, $srcbuf$15$lcssa = 0, $srcbuf$17 = 0, $srcbuf$17$ph = 0, $srcbuf$18 = 0, $srcbuf$18$lcssa = 0, $srcbuf$20 = 0, $srcbuf$20$ph = 0, $srcbuf$21 = 0, $srcbuf$21$lcssa = 0, $srcbuf$23 = 0, $srcbuf$23$ph = 0, $srcbuf$24 = 0;
 var $srcbuf$24$lcssa = 0, $srcbuf$26 = 0, $srcbuf$26$ph = 0, $srcbuf$27 = 0, $srcbuf$27$lcssa = 0, $srcbuf$29 = 0, $srcbuf$29$ph = 0, $srcbuf$3 = 0, $srcbuf$3$ph = 0, $srcbuf$30 = 0, $srcbuf$30$lcssa = 0, $srcbuf$32 = 0, $srcbuf$32$ph = 0, $srcbuf$33 = 0, $srcbuf$33$lcssa = 0, $srcbuf$35 = 0, $srcbuf$35$ph = 0, $srcbuf$36 = 0, $srcbuf$36$lcssa = 0, $srcbuf$38 = 0;
 var $srcbuf$38$ph = 0, $srcbuf$39 = 0, $srcbuf$39$lcssa = 0, $srcbuf$4 = 0, $srcbuf$4$lcssa = 0, $srcbuf$41 = 0, $srcbuf$41$ph = 0, $srcbuf$42 = 0, $srcbuf$42$lcssa = 0, $srcbuf$44 = 0, $srcbuf$44$ph = 0, $srcbuf$45 = 0, $srcbuf$45$lcssa = 0, $srcbuf$47 = 0, $srcbuf$47$ph = 0, $srcbuf$48 = 0, $srcbuf$48$lcssa = 0, $srcbuf$50 = 0, $srcbuf$50$ph = 0, $srcbuf$51 = 0;
 var $srcbuf$51$lcssa = 0, $srcbuf$6 = 0, $srcbuf$6$ph = 0, $srcbuf$7 = 0, $srcbuf$7$lcssa = 0, $srcbuf$9 = 0, $srcbuf$9$ph = 0, $start$0$i = 0, $start104$0$i = 0, $start119$0$i = 0, $start140$0$i = 0, $start155$0$i = 0, $start17$0$i = 0, $start170$0$i = 0, $start27$0$i = 0, $start37$0$i = 0, $start53$0$i = 0, $start67$0$i = 0, $start7$0$i = 0, $start88$0$i = 0;
 var $switch$split12D = 0, $switch$split2D = 0, $vskip$0$ph = 0, $vskip$3$ph = 0, $vskip$6$ph = 0, $vskip$9$ph = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($surf_src)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$surf_dst>>2]|0;
 $3 = $2 & 2;
 $4 = ($3|0)==(0);
 if (!($4)) {
  $5 = (_SDL_LockSurface($surf_dst)|0);
  $6 = ($5|0)<(0);
  if ($6) {
   $$0 = -1;
   return ($$0|0);
  }
 }
 $7 = HEAP32[$dstrect>>2]|0;
 $8 = ((($dstrect)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($surf_dst)) + 20|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($surf_dst)) + 16|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = Math_imul($13, $9)|0;
 $15 = (($11) + ($14)|0);
 $16 = ((($surf_src)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ((($17)) + 9|0);
 $19 = HEAP8[$18>>0]|0;
 $20 = $19&255;
 $21 = Math_imul($20, $7)|0;
 $22 = (($15) + ($21)|0);
 $23 = ((($surf_src)) + 52|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($24)) + 12|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ((($srcrect)) + 4|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ($28|0)==(0);
 L5: do {
  if ($29) {
   $srcbuf$13$ph = $26;
   label = 29;
  } else {
   switch ($20|0) {
   case 1:  {
    $srcbuf$0$ph = $26;$vskip$0$ph = $28;
    while(1) {
     $ofs$0 = 0;$srcbuf$0 = $srcbuf$0$ph;
     while(1) {
      $30 = HEAP8[$srcbuf$0>>0]|0;
      $31 = $30&255;
      $32 = (($31) + ($ofs$0))|0;
      $33 = ((($srcbuf$0)) + 1|0);
      $34 = HEAP8[$33>>0]|0;
      $35 = $34&255;
      $36 = ((($srcbuf$0)) + 2|0);
      $37 = ($34<<24>>24)==(0);
      if ($37) {
       $40 = ($32|0)==(0);
       if ($40) {
        break L5;
       } else {
        $ofs$1 = $32;$srcbuf$1 = $36;
       }
      } else {
       $38 = (($36) + ($35)|0);
       $39 = (($32) + ($35))|0;
       $ofs$1 = $39;$srcbuf$1 = $38;
      }
      $41 = ($ofs$1|0)==($1|0);
      if ($41) {
       $srcbuf$1$lcssa = $srcbuf$1;
       break;
      } else {
       $ofs$0 = $ofs$1;$srcbuf$0 = $srcbuf$1;
      }
     }
     $42 = (($vskip$0$ph) + -1)|0;
     $43 = ($42|0)==(0);
     if ($43) {
      $srcbuf$13$ph = $srcbuf$1$lcssa;
      label = 29;
      break;
     } else {
      $srcbuf$0$ph = $srcbuf$1$lcssa;$vskip$0$ph = $42;
     }
    }
    break;
   }
   case 2:  {
    $srcbuf$3$ph = $26;$vskip$3$ph = $28;
    while(1) {
     $ofs$4 = 0;$srcbuf$3 = $srcbuf$3$ph;
     while(1) {
      $44 = HEAP8[$srcbuf$3>>0]|0;
      $45 = $44&255;
      $46 = (($45) + ($ofs$4))|0;
      $47 = ((($srcbuf$3)) + 1|0);
      $48 = HEAP8[$47>>0]|0;
      $49 = $48&255;
      $50 = ((($srcbuf$3)) + 2|0);
      $51 = ($48<<24>>24)==(0);
      if ($51) {
       $55 = ($46|0)==(0);
       if ($55) {
        break L5;
       } else {
        $ofs$5 = $46;$srcbuf$4 = $50;
       }
      } else {
       $52 = $49 << 1;
       $53 = (($50) + ($52)|0);
       $54 = (($46) + ($49))|0;
       $ofs$5 = $54;$srcbuf$4 = $53;
      }
      $56 = ($ofs$5|0)==($1|0);
      if ($56) {
       $srcbuf$4$lcssa = $srcbuf$4;
       break;
      } else {
       $ofs$4 = $ofs$5;$srcbuf$3 = $srcbuf$4;
      }
     }
     $57 = (($vskip$3$ph) + -1)|0;
     $58 = ($57|0)==(0);
     if ($58) {
      $srcbuf$13$ph = $srcbuf$4$lcssa;
      label = 29;
      break;
     } else {
      $srcbuf$3$ph = $srcbuf$4$lcssa;$vskip$3$ph = $57;
     }
    }
    break;
   }
   case 3:  {
    $srcbuf$6$ph = $26;$vskip$6$ph = $28;
    while(1) {
     $ofs$8 = 0;$srcbuf$6 = $srcbuf$6$ph;
     while(1) {
      $59 = HEAP8[$srcbuf$6>>0]|0;
      $60 = $59&255;
      $61 = (($60) + ($ofs$8))|0;
      $62 = ((($srcbuf$6)) + 1|0);
      $63 = HEAP8[$62>>0]|0;
      $64 = $63&255;
      $65 = ((($srcbuf$6)) + 2|0);
      $66 = ($63<<24>>24)==(0);
      if ($66) {
       $70 = ($61|0)==(0);
       if ($70) {
        break L5;
       } else {
        $ofs$9 = $61;$srcbuf$7 = $65;
       }
      } else {
       $67 = ($64*3)|0;
       $68 = (($65) + ($67)|0);
       $69 = (($61) + ($64))|0;
       $ofs$9 = $69;$srcbuf$7 = $68;
      }
      $71 = ($ofs$9|0)==($1|0);
      if ($71) {
       $srcbuf$7$lcssa = $srcbuf$7;
       break;
      } else {
       $ofs$8 = $ofs$9;$srcbuf$6 = $srcbuf$7;
      }
     }
     $72 = (($vskip$6$ph) + -1)|0;
     $73 = ($72|0)==(0);
     if ($73) {
      $srcbuf$13$ph = $srcbuf$7$lcssa;
      label = 29;
      break;
     } else {
      $srcbuf$6$ph = $srcbuf$7$lcssa;$vskip$6$ph = $72;
     }
    }
    break;
   }
   case 4:  {
    $srcbuf$9$ph = $26;$vskip$9$ph = $28;
    while(1) {
     $ofs$12 = 0;$srcbuf$9 = $srcbuf$9$ph;
     while(1) {
      $74 = HEAP16[$srcbuf$9>>1]|0;
      $75 = $74&65535;
      $76 = (($75) + ($ofs$12))|0;
      $77 = ((($srcbuf$9)) + 2|0);
      $78 = HEAP16[$77>>1]|0;
      $79 = $78&65535;
      $80 = ((($srcbuf$9)) + 4|0);
      $81 = ($78<<16>>16)==(0);
      if ($81) {
       $85 = ($76|0)==(0);
       if ($85) {
        break L5;
       } else {
        $ofs$13 = $76;$srcbuf$10 = $80;
       }
      } else {
       $82 = $79 << 2;
       $83 = (($80) + ($82)|0);
       $84 = (($76) + ($79))|0;
       $ofs$13 = $84;$srcbuf$10 = $83;
      }
      $86 = ($ofs$13|0)==($1|0);
      if ($86) {
       $srcbuf$10$lcssa = $srcbuf$10;
       break;
      } else {
       $ofs$12 = $ofs$13;$srcbuf$9 = $srcbuf$10;
      }
     }
     $87 = (($vskip$9$ph) + -1)|0;
     $88 = ($87|0)==(0);
     if ($88) {
      $srcbuf$13$ph = $srcbuf$10$lcssa;
      label = 29;
      break;
     } else {
      $srcbuf$9$ph = $srcbuf$10$lcssa;$vskip$9$ph = $87;
     }
    }
    break;
   }
   default: {
    $srcbuf$13$ph = $26;
    label = 29;
    break L5;
   }
   }
  }
 } while(0);
 L44: do {
  if ((label|0) == 29) {
   $89 = ((($24)) + 79|0);
   $90 = HEAP8[$89>>0]|0;
   $91 = $90&255;
   $92 = HEAP32[$srcrect>>2]|0;
   $93 = ($92|0)==(0);
   if ($93) {
    $94 = ((($srcrect)) + 8|0);
    $95 = HEAP32[$94>>2]|0;
    $96 = HEAP32[$0>>2]|0;
    $97 = ($95|0)==($96|0);
    if ($97) {
     $1092 = ($90<<24>>24)==(-1);
     if ($1092) {
      switch ($20|0) {
      case 1:  {
       $1093 = ((($srcrect)) + 12|0);
       $1094 = HEAP32[$1093>>2]|0;
       $dstbuf$0$ph = $22;$linecount$0$ph = $1094;$srcbuf$14$ph = $srcbuf$13$ph;
       while(1) {
        $ofs8$0 = 0;$srcbuf$14 = $srcbuf$14$ph;
        while(1) {
         $1095 = HEAP8[$srcbuf$14>>0]|0;
         $1096 = $1095&255;
         $1097 = (($1096) + ($ofs8$0))|0;
         $1098 = ((($srcbuf$14)) + 1|0);
         $1099 = HEAP8[$1098>>0]|0;
         $1100 = $1099&255;
         $1101 = ((($srcbuf$14)) + 2|0);
         $1102 = ($1099<<24>>24)==(0);
         if ($1102) {
          $1106 = ($1097|0)==(0);
          if ($1106) {
           break L44;
          } else {
           $ofs8$1 = $1097;$srcbuf$15 = $1101;
          }
         } else {
          $1103 = (($dstbuf$0$ph) + ($1097)|0);
          (_SDL_memcpy($1103,$1101,$1100)|0);
          $1104 = (($1101) + ($1100)|0);
          $1105 = (($1097) + ($1100))|0;
          $ofs8$1 = $1105;$srcbuf$15 = $1104;
         }
         $1107 = ($ofs8$1|0)==($1|0);
         if ($1107) {
          $srcbuf$15$lcssa = $srcbuf$15;
          break;
         } else {
          $ofs8$0 = $ofs8$1;$srcbuf$14 = $srcbuf$15;
         }
        }
        $1108 = HEAP32[$12>>2]|0;
        $1109 = (($dstbuf$0$ph) + ($1108)|0);
        $1110 = (($linecount$0$ph) + -1)|0;
        $1111 = ($1110|0)==(0);
        if ($1111) {
         break L44;
        } else {
         $dstbuf$0$ph = $1109;$linecount$0$ph = $1110;$srcbuf$14$ph = $srcbuf$15$lcssa;
        }
       }
       break;
      }
      case 2:  {
       $1112 = ((($srcrect)) + 12|0);
       $1113 = HEAP32[$1112>>2]|0;
       $dstbuf$3$ph = $22;$linecount11$0$ph = $1113;$srcbuf$17$ph = $srcbuf$13$ph;
       while(1) {
        $ofs12$0 = 0;$srcbuf$17 = $srcbuf$17$ph;
        while(1) {
         $1114 = HEAP8[$srcbuf$17>>0]|0;
         $1115 = $1114&255;
         $1116 = (($1115) + ($ofs12$0))|0;
         $1117 = ((($srcbuf$17)) + 1|0);
         $1118 = HEAP8[$1117>>0]|0;
         $1119 = $1118&255;
         $1120 = ((($srcbuf$17)) + 2|0);
         $1121 = ($1118<<24>>24)==(0);
         if ($1121) {
          $1127 = ($1116|0)==(0);
          if ($1127) {
           break L44;
          } else {
           $ofs12$1 = $1116;$srcbuf$18 = $1120;
          }
         } else {
          $1122 = $1116 << 1;
          $1123 = (($dstbuf$3$ph) + ($1122)|0);
          $1124 = $1119 << 1;
          (_SDL_memcpy($1123,$1120,$1124)|0);
          $1125 = (($1120) + ($1124)|0);
          $1126 = (($1116) + ($1119))|0;
          $ofs12$1 = $1126;$srcbuf$18 = $1125;
         }
         $1128 = ($ofs12$1|0)==($1|0);
         if ($1128) {
          $srcbuf$18$lcssa = $srcbuf$18;
          break;
         } else {
          $ofs12$0 = $ofs12$1;$srcbuf$17 = $srcbuf$18;
         }
        }
        $1129 = HEAP32[$12>>2]|0;
        $1130 = (($dstbuf$3$ph) + ($1129)|0);
        $1131 = (($linecount11$0$ph) + -1)|0;
        $1132 = ($1131|0)==(0);
        if ($1132) {
         break L44;
        } else {
         $dstbuf$3$ph = $1130;$linecount11$0$ph = $1131;$srcbuf$17$ph = $srcbuf$18$lcssa;
        }
       }
       break;
      }
      case 3:  {
       $1133 = ((($srcrect)) + 12|0);
       $1134 = HEAP32[$1133>>2]|0;
       $dstbuf$6$ph = $22;$linecount15$0$ph = $1134;$srcbuf$20$ph = $srcbuf$13$ph;
       while(1) {
        $ofs16$0 = 0;$srcbuf$20 = $srcbuf$20$ph;
        while(1) {
         $1135 = HEAP8[$srcbuf$20>>0]|0;
         $1136 = $1135&255;
         $1137 = (($1136) + ($ofs16$0))|0;
         $1138 = ((($srcbuf$20)) + 1|0);
         $1139 = HEAP8[$1138>>0]|0;
         $1140 = $1139&255;
         $1141 = ((($srcbuf$20)) + 2|0);
         $1142 = ($1139<<24>>24)==(0);
         if ($1142) {
          $1148 = ($1137|0)==(0);
          if ($1148) {
           break L44;
          } else {
           $ofs16$1 = $1137;$srcbuf$21 = $1141;
          }
         } else {
          $1143 = ($1137*3)|0;
          $1144 = (($dstbuf$6$ph) + ($1143)|0);
          $1145 = ($1140*3)|0;
          (_SDL_memcpy($1144,$1141,$1145)|0);
          $1146 = (($1141) + ($1145)|0);
          $1147 = (($1137) + ($1140))|0;
          $ofs16$1 = $1147;$srcbuf$21 = $1146;
         }
         $1149 = ($ofs16$1|0)==($1|0);
         if ($1149) {
          $srcbuf$21$lcssa = $srcbuf$21;
          break;
         } else {
          $ofs16$0 = $ofs16$1;$srcbuf$20 = $srcbuf$21;
         }
        }
        $1150 = HEAP32[$12>>2]|0;
        $1151 = (($dstbuf$6$ph) + ($1150)|0);
        $1152 = (($linecount15$0$ph) + -1)|0;
        $1153 = ($1152|0)==(0);
        if ($1153) {
         break L44;
        } else {
         $dstbuf$6$ph = $1151;$linecount15$0$ph = $1152;$srcbuf$20$ph = $srcbuf$21$lcssa;
        }
       }
       break;
      }
      case 4:  {
       $1154 = ((($srcrect)) + 12|0);
       $1155 = HEAP32[$1154>>2]|0;
       $dstbuf$9$ph = $22;$linecount19$0$ph = $1155;$srcbuf$23$ph = $srcbuf$13$ph;
       while(1) {
        $ofs20$0 = 0;$srcbuf$23 = $srcbuf$23$ph;
        while(1) {
         $1156 = HEAP16[$srcbuf$23>>1]|0;
         $1157 = $1156&65535;
         $1158 = (($1157) + ($ofs20$0))|0;
         $1159 = ((($srcbuf$23)) + 2|0);
         $1160 = HEAP16[$1159>>1]|0;
         $1161 = $1160&65535;
         $1162 = ((($srcbuf$23)) + 4|0);
         $1163 = ($1160<<16>>16)==(0);
         if ($1163) {
          $1169 = ($1158|0)==(0);
          if ($1169) {
           break L44;
          } else {
           $ofs20$1 = $1158;$srcbuf$24 = $1162;
          }
         } else {
          $1164 = $1158 << 2;
          $1165 = (($dstbuf$9$ph) + ($1164)|0);
          $1166 = $1161 << 2;
          (_SDL_memcpy($1165,$1162,$1166)|0);
          $1167 = (($1162) + ($1166)|0);
          $1168 = (($1158) + ($1161))|0;
          $ofs20$1 = $1168;$srcbuf$24 = $1167;
         }
         $1170 = ($ofs20$1|0)==($1|0);
         if ($1170) {
          $srcbuf$24$lcssa = $srcbuf$24;
          break;
         } else {
          $ofs20$0 = $ofs20$1;$srcbuf$23 = $srcbuf$24;
         }
        }
        $1171 = HEAP32[$12>>2]|0;
        $1172 = (($dstbuf$9$ph) + ($1171)|0);
        $1173 = (($linecount19$0$ph) + -1)|0;
        $1174 = ($1173|0)==(0);
        if ($1174) {
         break L44;
        } else {
         $dstbuf$9$ph = $1172;$linecount19$0$ph = $1173;$srcbuf$23$ph = $srcbuf$24$lcssa;
        }
       }
       break;
      }
      default: {
       break L44;
      }
      }
     }
     switch ($20|0) {
     case 4:  {
      $1691 = ((($17)) + 12|0);
      $1692 = HEAP32[$1691>>2]|0;
      $1693 = ((($17)) + 16|0);
      $1694 = HEAP32[$1693>>2]|0;
      $1695 = $1694 | $1692;
      $1696 = ((($17)) + 20|0);
      $1697 = HEAP32[$1696>>2]|0;
      $1698 = $1695 | $1697;
      $1699 = ($1698|0)==(16777215);
      if ($1699) {
       $1700 = ($1694|0)==(65280);
       $1701 = ($1692|0)==(65280);
       $or$cond17 = $1701 | $1700;
       $1702 = ($1697|0)==(65280);
       $or$cond18 = $or$cond17 | $1702;
       if ($or$cond18) {
        $1703 = ($90<<24>>24)==(-128);
        $1704 = ((($srcrect)) + 12|0);
        $1705 = HEAP32[$1704>>2]|0;
        if ($1703) {
         $dstbuf$30$ph = $22;$linecount90$0$ph = $1705;$srcbuf$44$ph = $srcbuf$13$ph;
         while(1) {
          $ofs91$0 = 0;$srcbuf$44 = $srcbuf$44$ph;
          while(1) {
           $1706 = HEAP16[$srcbuf$44>>1]|0;
           $1707 = $1706&65535;
           $1708 = (($1707) + ($ofs91$0))|0;
           $1709 = ((($srcbuf$44)) + 2|0);
           $1710 = HEAP16[$1709>>1]|0;
           $1711 = $1710&65535;
           $1712 = ((($srcbuf$44)) + 4|0);
           $1713 = ($1710<<16>>16)==(0);
           if ($1713) {
            $1731 = ($1708|0)==(0);
            if ($1731) {
             break L44;
            } else {
             $ofs91$1 = $1708;$srcbuf$45 = $1712;
            }
           } else {
            $1714 = $1708 << 2;
            $1715 = (($dstbuf$30$ph) + ($1714)|0);
            $dst95$0136 = $1715;$i93$0134 = 0;$src94$0135 = $1712;
            while(1) {
             $1716 = ((($src94$0135)) + 4|0);
             $1717 = HEAP32[$src94$0135>>2]|0;
             $1718 = HEAP32[$dst95$0136>>2]|0;
             $1719 = $1717 & 16711422;
             $1720 = $1718 & 16711422;
             $1721 = (($1720) + ($1719))|0;
             $1722 = $1721 >>> 1;
             $1723 = $1717 & 65793;
             $1724 = $1723 & $1718;
             $1725 = (($1722) + ($1724))|0;
             $1726 = ((($dst95$0136)) + 4|0);
             HEAP32[$dst95$0136>>2] = $1725;
             $1727 = (($i93$0134) + 1)|0;
             $exitcond = ($1727|0)==($1711|0);
             if ($exitcond) {
              break;
             } else {
              $dst95$0136 = $1726;$i93$0134 = $1727;$src94$0135 = $1716;
             }
            }
            $1728 = $1711 << 2;
            $1729 = (($1712) + ($1728)|0);
            $1730 = (($1708) + ($1711))|0;
            $ofs91$1 = $1730;$srcbuf$45 = $1729;
           }
           $1732 = ($ofs91$1|0)==($1|0);
           if ($1732) {
            $srcbuf$45$lcssa = $srcbuf$45;
            break;
           } else {
            $ofs91$0 = $ofs91$1;$srcbuf$44 = $srcbuf$45;
           }
          }
          $1733 = HEAP32[$12>>2]|0;
          $1734 = (($dstbuf$30$ph) + ($1733)|0);
          $1735 = (($linecount90$0$ph) + -1)|0;
          $1736 = ($1735|0)==(0);
          if ($1736) {
           break L44;
          } else {
           $dstbuf$30$ph = $1734;$linecount90$0$ph = $1735;$srcbuf$44$ph = $srcbuf$45$lcssa;
          }
         }
        } else {
         $dstbuf$33$ph = $22;$linecount99$0$ph = $1705;$srcbuf$47$ph = $srcbuf$13$ph;
         while(1) {
          $ofs100$0 = 0;$srcbuf$47 = $srcbuf$47$ph;
          while(1) {
           $1737 = HEAP16[$srcbuf$47>>1]|0;
           $1738 = $1737&65535;
           $1739 = (($1738) + ($ofs100$0))|0;
           $1740 = ((($srcbuf$47)) + 2|0);
           $1741 = HEAP16[$1740>>1]|0;
           $1742 = $1741&65535;
           $1743 = ((($srcbuf$47)) + 4|0);
           $1744 = ($1741<<16>>16)==(0);
           if ($1744) {
            $1770 = ($1739|0)==(0);
            if ($1770) {
             break L44;
            } else {
             $ofs100$1 = $1739;$srcbuf$48 = $1743;
            }
           } else {
            $1745 = $1739 << 2;
            $1746 = (($dstbuf$33$ph) + ($1745)|0);
            $dst104$0139 = $1746;$i102$0137 = 0;$src103$0138 = $1743;
            while(1) {
             $1747 = ((($src103$0138)) + 4|0);
             $1748 = HEAP32[$src103$0138>>2]|0;
             $1749 = HEAP32[$dst104$0139>>2]|0;
             $1750 = $1748 & 16711935;
             $1751 = $1749 & 16711935;
             $1752 = (($1750) - ($1751))|0;
             $1753 = Math_imul($1752, $91)|0;
             $1754 = $1753 >>> 8;
             $1755 = (($1754) + ($1751))|0;
             $1756 = $1755 & 16711935;
             $1757 = $1748 & 65280;
             $1758 = $1749 & 65280;
             $1759 = (($1757) - ($1758))|0;
             $1760 = Math_imul($1759, $91)|0;
             $1761 = $1760 >>> 8;
             $1762 = (($1761) + ($1758))|0;
             $1763 = $1762 & 65280;
             $1764 = $1756 | $1763;
             $1765 = ((($dst104$0139)) + 4|0);
             HEAP32[$dst104$0139>>2] = $1764;
             $1766 = (($i102$0137) + 1)|0;
             $exitcond259 = ($1766|0)==($1742|0);
             if ($exitcond259) {
              break;
             } else {
              $dst104$0139 = $1765;$i102$0137 = $1766;$src103$0138 = $1747;
             }
            }
            $1767 = $1742 << 2;
            $1768 = (($1743) + ($1767)|0);
            $1769 = (($1739) + ($1742))|0;
            $ofs100$1 = $1769;$srcbuf$48 = $1768;
           }
           $1771 = ($ofs100$1|0)==($1|0);
           if ($1771) {
            $srcbuf$48$lcssa = $srcbuf$48;
            break;
           } else {
            $ofs100$0 = $ofs100$1;$srcbuf$47 = $srcbuf$48;
           }
          }
          $1772 = HEAP32[$12>>2]|0;
          $1773 = (($dstbuf$33$ph) + ($1772)|0);
          $1774 = (($linecount99$0$ph) + -1)|0;
          $1775 = ($1774|0)==(0);
          if ($1775) {
           break L44;
          } else {
           $dstbuf$33$ph = $1773;$linecount99$0$ph = $1774;$srcbuf$47$ph = $srcbuf$48$lcssa;
          }
         }
        }
       }
      }
      $1776 = ((($srcrect)) + 12|0);
      $1777 = HEAP32[$1776>>2]|0;
      $1778 = ((($17)) + 32|0);
      $1779 = ((($17)) + 28|0);
      $1780 = ((($17)) + 24|0);
      $dstbuf$36$ph = $22;$linecount108$0$ph = $1777;$srcbuf$50$ph = $srcbuf$13$ph;
      while(1) {
       $ofs109$0 = 0;$srcbuf$50 = $srcbuf$50$ph;
       while(1) {
        $1781 = HEAP16[$srcbuf$50>>1]|0;
        $1782 = $1781&65535;
        $1783 = (($1782) + ($ofs109$0))|0;
        $1784 = ((($srcbuf$50)) + 2|0);
        $1785 = HEAP16[$1784>>1]|0;
        $1786 = $1785&65535;
        $1787 = ((($srcbuf$50)) + 4|0);
        $1788 = ($1785<<16>>16)==(0);
        if ($1788) {
         $1872 = ($1783|0)==(0);
         if ($1872) {
          break L44;
         } else {
          $ofs109$1 = $1783;$srcbuf$51 = $1787;
         }
        } else {
         $1789 = $1783 << 2;
         $1790 = (($dstbuf$36$ph) + ($1789)|0);
         $dst113$0145 = $1790;$i111$0143 = 0;$src112$0144 = $1787;
         while(1) {
          $1791 = HEAP32[$src112$0144>>2]|0;
          $1792 = HEAP32[$dst113$0145>>2]|0;
          $1793 = HEAP32[$1691>>2]|0;
          $1794 = $1793 & $1791;
          $1795 = HEAP32[$1778>>2]|0;
          $1796 = $1795 & 255;
          $1797 = $1794 >>> $1796;
          $1798 = HEAP32[$1779>>2]|0;
          $1799 = $1798 & 255;
          $1800 = (125924 + ($1799<<2)|0);
          $1801 = HEAP32[$1800>>2]|0;
          $1802 = (($1801) + ($1797)|0);
          $1803 = HEAP8[$1802>>0]|0;
          $1804 = $1803&255;
          $1805 = HEAP32[$1693>>2]|0;
          $1806 = $1805 & $1791;
          $1807 = $1795 >>> 8;
          $1808 = $1807 & 255;
          $1809 = $1806 >>> $1808;
          $1810 = $1798 >>> 8;
          $1811 = $1810 & 255;
          $1812 = (125924 + ($1811<<2)|0);
          $1813 = HEAP32[$1812>>2]|0;
          $1814 = (($1813) + ($1809)|0);
          $1815 = HEAP8[$1814>>0]|0;
          $1816 = $1815&255;
          $1817 = HEAP32[$1696>>2]|0;
          $1818 = $1817 & $1791;
          $1819 = $1795 >>> 16;
          $1820 = $1819 & 255;
          $1821 = $1818 >>> $1820;
          $1822 = $1798 >>> 16;
          $1823 = $1822 & 255;
          $1824 = (125924 + ($1823<<2)|0);
          $1825 = HEAP32[$1824>>2]|0;
          $1826 = (($1825) + ($1821)|0);
          $1827 = HEAP8[$1826>>0]|0;
          $1828 = $1827&255;
          $1829 = $1793 & $1792;
          $1830 = $1829 >>> $1796;
          $1831 = (($1801) + ($1830)|0);
          $1832 = HEAP8[$1831>>0]|0;
          $1833 = $1832&255;
          $1834 = $1805 & $1792;
          $1835 = $1834 >>> $1808;
          $1836 = (($1813) + ($1835)|0);
          $1837 = HEAP8[$1836>>0]|0;
          $1838 = $1837&255;
          $1839 = $1817 & $1792;
          $1840 = $1839 >>> $1820;
          $1841 = (($1825) + ($1840)|0);
          $1842 = HEAP8[$1841>>0]|0;
          $1843 = $1842&255;
          $1844 = (($1804) - ($1833))|0;
          $1845 = Math_imul($1844, $91)|0;
          $1846 = $1845 >>> 8;
          $1847 = (($1846) + ($1833))|0;
          $1848 = (($1816) - ($1838))|0;
          $1849 = Math_imul($1848, $91)|0;
          $1850 = $1849 >>> 8;
          $1851 = (($1850) + ($1838))|0;
          $1852 = (($1828) - ($1843))|0;
          $1853 = Math_imul($1852, $91)|0;
          $1854 = $1853 >>> 8;
          $1855 = (($1854) + ($1843))|0;
          $1856 = $1847 >>> $1799;
          $1857 = $1856 << $1796;
          $1858 = $1851 >>> $1811;
          $1859 = $1858 << $1808;
          $1860 = $1855 >>> $1823;
          $1861 = $1860 << $1820;
          $1862 = HEAP32[$1780>>2]|0;
          $1863 = $1857 | $1862;
          $1864 = $1863 | $1859;
          $1865 = $1864 | $1861;
          HEAP32[$dst113$0145>>2] = $1865;
          $1866 = ((($src112$0144)) + 4|0);
          $1867 = ((($dst113$0145)) + 4|0);
          $1868 = (($i111$0143) + 1)|0;
          $exitcond260 = ($1868|0)==($1786|0);
          if ($exitcond260) {
           break;
          } else {
           $dst113$0145 = $1867;$i111$0143 = $1868;$src112$0144 = $1866;
          }
         }
         $1869 = $1786 << 2;
         $1870 = (($1787) + ($1869)|0);
         $1871 = (($1783) + ($1786))|0;
         $ofs109$1 = $1871;$srcbuf$51 = $1870;
        }
        $1873 = ($ofs109$1|0)==($1|0);
        if ($1873) {
         $srcbuf$51$lcssa = $srcbuf$51;
         break;
        } else {
         $ofs109$0 = $ofs109$1;$srcbuf$50 = $srcbuf$51;
        }
       }
       $1874 = HEAP32[$12>>2]|0;
       $1875 = (($dstbuf$36$ph) + ($1874)|0);
       $1876 = (($linecount108$0$ph) + -1)|0;
       $1877 = ($1876|0)==(0);
       if ($1877) {
        break L44;
       } else {
        $dstbuf$36$ph = $1875;$linecount108$0$ph = $1876;$srcbuf$50$ph = $srcbuf$51$lcssa;
       }
      }
      break;
     }
     case 2:  {
      $1175 = ((($17)) + 12|0);
      $1176 = HEAP32[$1175>>2]|0;
      $1177 = ((($17)) + 16|0);
      $1178 = HEAP32[$1177>>2]|0;
      $1179 = $1178 | $1176;
      $1180 = ((($17)) + 20|0);
      $1181 = HEAP32[$1180>>2]|0;
      $1182 = $1179 | $1181;
      $switch$split12D = ($1182|0)<(65535);
      L137: do {
       if ($switch$split12D) {
        switch ($1182|0) {
        case 32767:  {
         break;
        }
        default: {
         break L137;
        }
        }
        $1320 = ($1178|0)==(992);
        $1321 = ($1176|0)==(992);
        $or$cond15 = $1321 | $1320;
        $1322 = ($1181|0)==(992);
        $or$cond16 = $or$cond15 | $1322;
        if ($or$cond16) {
         $1323 = ($90<<24>>24)==(-128);
         $1324 = ((($srcrect)) + 12|0);
         $1325 = HEAP32[$1324>>2]|0;
         if ($1323) {
          $dstbuf$18$ph = $22;$linecount41$0$ph = $1325;$srcbuf$32$ph = $srcbuf$13$ph;
         } else {
          $1326 = $91 >>> 3;
          $dstbuf$21$ph = $22;$linecount56$0$ph = $1325;$srcbuf$35$ph = $srcbuf$13$ph;
          while(1) {
           $ofs57$0 = 0;$srcbuf$35 = $srcbuf$35$ph;
           while(1) {
            $1418 = HEAP8[$srcbuf$35>>0]|0;
            $1419 = $1418&255;
            $1420 = (($1419) + ($ofs57$0))|0;
            $1421 = ((($srcbuf$35)) + 1|0);
            $1422 = HEAP8[$1421>>0]|0;
            $1423 = $1422&255;
            $1424 = ((($srcbuf$35)) + 2|0);
            $1425 = ($1422<<24>>24)==(0);
            if ($1425) {
             $1452 = ($1420|0)==(0);
             if ($1452) {
              break L44;
             } else {
              $ofs57$1 = $1420;$srcbuf$36 = $1424;
             }
            } else {
             $1426 = $1420 << 1;
             $1427 = (($dstbuf$21$ph) + ($1426)|0);
             $dst61$0184 = $1427;$i59$0186 = 0;$src60$0185 = $1424;
             while(1) {
              $1428 = ((($src60$0185)) + 2|0);
              $1429 = HEAP16[$src60$0185>>1]|0;
              $1430 = $1429&65535;
              $1431 = HEAP16[$dst61$0184>>1]|0;
              $1432 = $1431&65535;
              $1433 = $1430 << 16;
              $1434 = $1433 | $1430;
              $1435 = $1434 & 65043487;
              $1436 = $1432 << 16;
              $1437 = $1436 | $1432;
              $1438 = $1437 & 65043487;
              $1439 = (($1435) - ($1438))|0;
              $1440 = Math_imul($1439, $1326)|0;
              $1441 = $1440 >>> 5;
              $1442 = (($1441) + ($1438))|0;
              $1443 = $1442 & 65043487;
              $1444 = $1443 >>> 16;
              $1445 = $1444 | $1443;
              $1446 = $1445&65535;
              $1447 = ((($dst61$0184)) + 2|0);
              HEAP16[$dst61$0184>>1] = $1446;
              $1448 = (($i59$0186) + 1)|0;
              $exitcond270 = ($1448|0)==($1423|0);
              if ($exitcond270) {
               break;
              } else {
               $dst61$0184 = $1447;$i59$0186 = $1448;$src60$0185 = $1428;
              }
             }
             $1449 = $1423 << 1;
             $1450 = (($1424) + ($1449)|0);
             $1451 = (($1420) + ($1423))|0;
             $ofs57$1 = $1451;$srcbuf$36 = $1450;
            }
            $1453 = ($ofs57$1|0)==($1|0);
            if ($1453) {
             $srcbuf$36$lcssa = $srcbuf$36;
             break;
            } else {
             $ofs57$0 = $ofs57$1;$srcbuf$35 = $srcbuf$36;
            }
           }
           $1454 = (($dstbuf$21$ph) + ($13)|0);
           $1455 = (($linecount56$0$ph) + -1)|0;
           $1456 = ($1455|0)==(0);
           if ($1456) {
            break L44;
           } else {
            $dstbuf$21$ph = $1454;$linecount56$0$ph = $1455;$srcbuf$35$ph = $srcbuf$36$lcssa;
           }
          }
         }
         while(1) {
          $ofs42$0 = 0;$srcbuf$32 = $srcbuf$32$ph;
          while(1) {
           $1327 = HEAP8[$srcbuf$32>>0]|0;
           $1328 = $1327&255;
           $1329 = (($1328) + ($ofs42$0))|0;
           $1330 = ((($srcbuf$32)) + 1|0);
           $1331 = HEAP8[$1330>>0]|0;
           $1332 = $1331&255;
           $1333 = ((($srcbuf$32)) + 2|0);
           $1334 = ($1331<<24>>24)==(0);
           if ($1334) {
            $1412 = ($1329|0)==(0);
            if ($1412) {
             break L44;
            } else {
             $ofs42$1 = $1329;$srcbuf$33 = $1333;
            }
           } else {
            $1335 = $1329 << 1;
            $1336 = (($dstbuf$18$ph) + ($1335)|0);
            $1337 = $1333;
            $1338 = $1336;
            $1339 = $1338 ^ $1337;
            $1340 = $1339 & 3;
            $1341 = ($1340|0)==(0);
            do {
             if ($1341) {
              $1359 = $1337 & 3;
              $1360 = ($1359|0)==(0);
              if ($1360) {
               $dst46$1$ph = $1336;$n44$1$ph = $1332;$src45$1$ph = $1333;
              } else {
               $1361 = ((($srcbuf$32)) + 4|0);
               $1362 = HEAP16[$1333>>1]|0;
               $1363 = $1362&65535;
               $1364 = HEAP16[$1336>>1]|0;
               $1365 = $1364&65535;
               $1366 = $1363 & 64478;
               $1367 = $1365 & 64478;
               $1368 = (($1367) + ($1366))|0;
               $1369 = $1368 >>> 1;
               $1370 = $1364 & $1362;
               $1371 = $1370&65535;
               $1372 = $1371 & 1057;
               $1373 = (($1369) + ($1372))|0;
               $1374 = $1373&65535;
               $1375 = ((($1336)) + 2|0);
               HEAP16[$1336>>1] = $1374;
               $1376 = (($1332) + -1)|0;
               $dst46$1$ph = $1375;$n44$1$ph = $1376;$src45$1$ph = $1361;
              }
              $1377 = ($n44$1$ph>>>0)>(1);
              if ($1377) {
               $1378 = (($n44$1$ph) + -2)|0;
               $1379 = $1378 & -2;
               $scevgep266 = ((($dst46$1$ph)) + 4|0);
               $scevgep268 = ((($src45$1$ph)) + 4|0);
               $dst46$1176 = $dst46$1$ph;$n44$1178 = $n44$1$ph;$src45$1177 = $src45$1$ph;
               while(1) {
                $1380 = HEAP32[$src45$1177>>2]|0;
                $1381 = HEAP32[$dst46$1176>>2]|0;
                $1382 = $1380 >>> 1;
                $1383 = $1382 & 2112847343;
                $1384 = $1381 >>> 1;
                $1385 = $1384 & 2112847343;
                $1386 = (($1385) + ($1383))|0;
                $1387 = $1380 & 69272609;
                $1388 = $1387 & $1381;
                $1389 = (($1386) + ($1388))|0;
                HEAP32[$dst46$1176>>2] = $1389;
                $1390 = ((($src45$1177)) + 4|0);
                $1391 = ((($dst46$1176)) + 4|0);
                $1392 = (($n44$1178) + -2)|0;
                $1393 = ($1392>>>0)>(1);
                if ($1393) {
                 $dst46$1176 = $1391;$n44$1178 = $1392;$src45$1177 = $1390;
                } else {
                 break;
                }
               }
               $1394 = (($1378) - ($1379))|0;
               $scevgep267 = (($scevgep266) + ($1379<<1)|0);
               $scevgep269 = (($scevgep268) + ($1379<<1)|0);
               $dst46$1$lcssa = $scevgep267;$n44$1$lcssa = $1394;$src45$1$lcssa = $scevgep269;
              } else {
               $dst46$1$lcssa = $dst46$1$ph;$n44$1$lcssa = $n44$1$ph;$src45$1$lcssa = $src45$1$ph;
              }
              $1395 = ($n44$1$lcssa|0)==(0);
              if ($1395) {
               break;
              }
              $1396 = HEAP16[$src45$1$lcssa>>1]|0;
              $1397 = $1396&65535;
              $1398 = HEAP16[$dst46$1$lcssa>>1]|0;
              $1399 = $1398&65535;
              $1400 = $1397 & 64478;
              $1401 = $1399 & 64478;
              $1402 = (($1401) + ($1400))|0;
              $1403 = $1402 >>> 1;
              $1404 = $1398 & $1396;
              $1405 = $1404&65535;
              $1406 = $1405 & 1057;
              $1407 = (($1403) + ($1406))|0;
              $1408 = $1407&65535;
              HEAP16[$dst46$1$lcssa>>1] = $1408;
             } else {
              $dst46$0172 = $1336;$n44$0174 = $1332;$src45$0173 = $1333;
              while(1) {
               $1342 = (($n44$0174) + -1)|0;
               $1343 = ((($src45$0173)) + 2|0);
               $1344 = HEAP16[$src45$0173>>1]|0;
               $1345 = $1344&65535;
               $1346 = HEAP16[$dst46$0172>>1]|0;
               $1347 = $1346&65535;
               $1348 = $1345 & 64478;
               $1349 = $1347 & 64478;
               $1350 = (($1349) + ($1348))|0;
               $1351 = $1350 >>> 1;
               $1352 = $1346 & $1344;
               $1353 = $1352&65535;
               $1354 = $1353 & 1057;
               $1355 = (($1351) + ($1354))|0;
               $1356 = $1355&65535;
               $1357 = ((($dst46$0172)) + 2|0);
               HEAP16[$dst46$0172>>1] = $1356;
               $1358 = ($1342|0)==(0);
               if ($1358) {
                break;
               } else {
                $dst46$0172 = $1357;$n44$0174 = $1342;$src45$0173 = $1343;
               }
              }
             }
            } while(0);
            $1409 = $1332 << 1;
            $1410 = (($1333) + ($1409)|0);
            $1411 = (($1329) + ($1332))|0;
            $ofs42$1 = $1411;$srcbuf$33 = $1410;
           }
           $1413 = ($ofs42$1|0)==($1|0);
           if ($1413) {
            $srcbuf$33$lcssa = $srcbuf$33;
            break;
           } else {
            $ofs42$0 = $ofs42$1;$srcbuf$32 = $srcbuf$33;
           }
          }
          $1414 = HEAP32[$12>>2]|0;
          $1415 = (($dstbuf$18$ph) + ($1414)|0);
          $1416 = (($linecount41$0$ph) + -1)|0;
          $1417 = ($1416|0)==(0);
          if ($1417) {
           break L44;
          } else {
           $dstbuf$18$ph = $1415;$linecount41$0$ph = $1416;$srcbuf$32$ph = $srcbuf$33$lcssa;
          }
         }
        }
       } else {
        switch ($1182|0) {
        case 65535:  {
         break;
        }
        default: {
         break L137;
        }
        }
        $1183 = ($1178|0)==(2016);
        $1184 = ($1176|0)==(2016);
        $or$cond = $1184 | $1183;
        $1185 = ($1181|0)==(2016);
        $or$cond14 = $or$cond | $1185;
        if ($or$cond14) {
         $1186 = ($90<<24>>24)==(-128);
         $1187 = ((($srcrect)) + 12|0);
         $1188 = HEAP32[$1187>>2]|0;
         if ($1186) {
          $dstbuf$12$ph = $22;$linecount23$0$ph = $1188;$srcbuf$26$ph = $srcbuf$13$ph;
         } else {
          $1189 = $91 >>> 3;
          $dstbuf$15$ph = $22;$linecount33$0$ph = $1188;$srcbuf$29$ph = $srcbuf$13$ph;
          while(1) {
           $ofs34$0 = 0;$srcbuf$29 = $srcbuf$29$ph;
           while(1) {
            $1281 = HEAP8[$srcbuf$29>>0]|0;
            $1282 = $1281&255;
            $1283 = (($1282) + ($ofs34$0))|0;
            $1284 = ((($srcbuf$29)) + 1|0);
            $1285 = HEAP8[$1284>>0]|0;
            $1286 = $1285&255;
            $1287 = ((($srcbuf$29)) + 2|0);
            $1288 = ($1285<<24>>24)==(0);
            if ($1288) {
             $1315 = ($1283|0)==(0);
             if ($1315) {
              break L44;
             } else {
              $ofs34$1 = $1283;$srcbuf$30 = $1287;
             }
            } else {
             $1289 = $1283 << 1;
             $1290 = (($dstbuf$15$ph) + ($1289)|0);
             $dst37$0166 = $1290;$i$0168 = 0;$src36$0167 = $1287;
             while(1) {
              $1291 = ((($src36$0167)) + 2|0);
              $1292 = HEAP16[$src36$0167>>1]|0;
              $1293 = $1292&65535;
              $1294 = HEAP16[$dst37$0166>>1]|0;
              $1295 = $1294&65535;
              $1296 = $1293 << 16;
              $1297 = $1296 | $1293;
              $1298 = $1297 & 132184095;
              $1299 = $1295 << 16;
              $1300 = $1299 | $1295;
              $1301 = $1300 & 132184095;
              $1302 = (($1298) - ($1301))|0;
              $1303 = Math_imul($1302, $1189)|0;
              $1304 = $1303 >>> 5;
              $1305 = (($1304) + ($1301))|0;
              $1306 = $1305 & 132184095;
              $1307 = $1306 >>> 16;
              $1308 = $1307 | $1306;
              $1309 = $1308&65535;
              $1310 = ((($dst37$0166)) + 2|0);
              HEAP16[$dst37$0166>>1] = $1309;
              $1311 = (($i$0168) + 1)|0;
              $exitcond265 = ($1311|0)==($1286|0);
              if ($exitcond265) {
               break;
              } else {
               $dst37$0166 = $1310;$i$0168 = $1311;$src36$0167 = $1291;
              }
             }
             $1312 = $1286 << 1;
             $1313 = (($1287) + ($1312)|0);
             $1314 = (($1283) + ($1286))|0;
             $ofs34$1 = $1314;$srcbuf$30 = $1313;
            }
            $1316 = ($ofs34$1|0)==($1|0);
            if ($1316) {
             $srcbuf$30$lcssa = $srcbuf$30;
             break;
            } else {
             $ofs34$0 = $ofs34$1;$srcbuf$29 = $srcbuf$30;
            }
           }
           $1317 = (($dstbuf$15$ph) + ($13)|0);
           $1318 = (($linecount33$0$ph) + -1)|0;
           $1319 = ($1318|0)==(0);
           if ($1319) {
            break L44;
           } else {
            $dstbuf$15$ph = $1317;$linecount33$0$ph = $1318;$srcbuf$29$ph = $srcbuf$30$lcssa;
           }
          }
         }
         while(1) {
          $ofs24$0 = 0;$srcbuf$26 = $srcbuf$26$ph;
          while(1) {
           $1190 = HEAP8[$srcbuf$26>>0]|0;
           $1191 = $1190&255;
           $1192 = (($1191) + ($ofs24$0))|0;
           $1193 = ((($srcbuf$26)) + 1|0);
           $1194 = HEAP8[$1193>>0]|0;
           $1195 = $1194&255;
           $1196 = ((($srcbuf$26)) + 2|0);
           $1197 = ($1194<<24>>24)==(0);
           if ($1197) {
            $1275 = ($1192|0)==(0);
            if ($1275) {
             break L44;
            } else {
             $ofs24$1 = $1192;$srcbuf$27 = $1196;
            }
           } else {
            $1198 = $1192 << 1;
            $1199 = (($dstbuf$12$ph) + ($1198)|0);
            $1200 = $1196;
            $1201 = $1199;
            $1202 = $1201 ^ $1200;
            $1203 = $1202 & 3;
            $1204 = ($1203|0)==(0);
            do {
             if ($1204) {
              $1222 = $1200 & 3;
              $1223 = ($1222|0)==(0);
              if ($1223) {
               $dst$1$ph = $1199;$n$1$ph = $1195;$src$1$ph = $1196;
              } else {
               $1224 = ((($srcbuf$26)) + 4|0);
               $1225 = HEAP16[$1196>>1]|0;
               $1226 = $1225&65535;
               $1227 = HEAP16[$1199>>1]|0;
               $1228 = $1227&65535;
               $1229 = $1226 & 63454;
               $1230 = $1228 & 63454;
               $1231 = (($1230) + ($1229))|0;
               $1232 = $1231 >>> 1;
               $1233 = $1227 & $1225;
               $1234 = $1233&65535;
               $1235 = $1234 & 2081;
               $1236 = (($1232) + ($1235))|0;
               $1237 = $1236&65535;
               $1238 = ((($1199)) + 2|0);
               HEAP16[$1199>>1] = $1237;
               $1239 = (($1195) + -1)|0;
               $dst$1$ph = $1238;$n$1$ph = $1239;$src$1$ph = $1224;
              }
              $1240 = ($n$1$ph>>>0)>(1);
              if ($1240) {
               $1241 = (($n$1$ph) + -2)|0;
               $1242 = $1241 & -2;
               $scevgep = ((($dst$1$ph)) + 4|0);
               $scevgep263 = ((($src$1$ph)) + 4|0);
               $dst$1159 = $dst$1$ph;$n$1161 = $n$1$ph;$src$1160 = $src$1$ph;
               while(1) {
                $1243 = HEAP32[$src$1160>>2]|0;
                $1244 = HEAP32[$dst$1159>>2]|0;
                $1245 = $1243 >>> 1;
                $1246 = $1245 & 2079292399;
                $1247 = $1244 >>> 1;
                $1248 = $1247 & 2079292399;
                $1249 = (($1248) + ($1246))|0;
                $1250 = $1243 & 136382497;
                $1251 = $1250 & $1244;
                $1252 = (($1249) + ($1251))|0;
                HEAP32[$dst$1159>>2] = $1252;
                $1253 = ((($src$1160)) + 4|0);
                $1254 = ((($dst$1159)) + 4|0);
                $1255 = (($n$1161) + -2)|0;
                $1256 = ($1255>>>0)>(1);
                if ($1256) {
                 $dst$1159 = $1254;$n$1161 = $1255;$src$1160 = $1253;
                } else {
                 break;
                }
               }
               $1257 = (($1241) - ($1242))|0;
               $scevgep262 = (($scevgep) + ($1242<<1)|0);
               $scevgep264 = (($scevgep263) + ($1242<<1)|0);
               $dst$1$lcssa = $scevgep262;$n$1$lcssa = $1257;$src$1$lcssa = $scevgep264;
              } else {
               $dst$1$lcssa = $dst$1$ph;$n$1$lcssa = $n$1$ph;$src$1$lcssa = $src$1$ph;
              }
              $1258 = ($n$1$lcssa|0)==(0);
              if ($1258) {
               break;
              }
              $1259 = HEAP16[$src$1$lcssa>>1]|0;
              $1260 = $1259&65535;
              $1261 = HEAP16[$dst$1$lcssa>>1]|0;
              $1262 = $1261&65535;
              $1263 = $1260 & 63454;
              $1264 = $1262 & 63454;
              $1265 = (($1264) + ($1263))|0;
              $1266 = $1265 >>> 1;
              $1267 = $1261 & $1259;
              $1268 = $1267&65535;
              $1269 = $1268 & 2081;
              $1270 = (($1266) + ($1269))|0;
              $1271 = $1270&65535;
              HEAP16[$dst$1$lcssa>>1] = $1271;
             } else {
              $dst$0155 = $1199;$n$0157 = $1195;$src$0156 = $1196;
              while(1) {
               $1205 = (($n$0157) + -1)|0;
               $1206 = ((($src$0156)) + 2|0);
               $1207 = HEAP16[$src$0156>>1]|0;
               $1208 = $1207&65535;
               $1209 = HEAP16[$dst$0155>>1]|0;
               $1210 = $1209&65535;
               $1211 = $1208 & 63454;
               $1212 = $1210 & 63454;
               $1213 = (($1212) + ($1211))|0;
               $1214 = $1213 >>> 1;
               $1215 = $1209 & $1207;
               $1216 = $1215&65535;
               $1217 = $1216 & 2081;
               $1218 = (($1214) + ($1217))|0;
               $1219 = $1218&65535;
               $1220 = ((($dst$0155)) + 2|0);
               HEAP16[$dst$0155>>1] = $1219;
               $1221 = ($1205|0)==(0);
               if ($1221) {
                break;
               } else {
                $dst$0155 = $1220;$n$0157 = $1205;$src$0156 = $1206;
               }
              }
             }
            } while(0);
            $1272 = $1195 << 1;
            $1273 = (($1196) + ($1272)|0);
            $1274 = (($1192) + ($1195))|0;
            $ofs24$1 = $1274;$srcbuf$27 = $1273;
           }
           $1276 = ($ofs24$1|0)==($1|0);
           if ($1276) {
            $srcbuf$27$lcssa = $srcbuf$27;
            break;
           } else {
            $ofs24$0 = $ofs24$1;$srcbuf$26 = $srcbuf$27;
           }
          }
          $1277 = HEAP32[$12>>2]|0;
          $1278 = (($dstbuf$12$ph) + ($1277)|0);
          $1279 = (($linecount23$0$ph) + -1)|0;
          $1280 = ($1279|0)==(0);
          if ($1280) {
           break L44;
          } else {
           $dstbuf$12$ph = $1278;$linecount23$0$ph = $1279;$srcbuf$26$ph = $srcbuf$27$lcssa;
          }
         }
        }
       }
      } while(0);
      $1457 = ((($srcrect)) + 12|0);
      $1458 = HEAP32[$1457>>2]|0;
      $1459 = ((($17)) + 32|0);
      $1460 = ((($17)) + 28|0);
      $1461 = ((($17)) + 24|0);
      $dstbuf$24$ph = $22;$linecount66$0$ph = $1458;$srcbuf$38$ph = $srcbuf$13$ph;
      while(1) {
       $ofs67$0 = 0;$srcbuf$38 = $srcbuf$38$ph;
       while(1) {
        $1462 = HEAP8[$srcbuf$38>>0]|0;
        $1463 = $1462&255;
        $1464 = (($1463) + ($ofs67$0))|0;
        $1465 = ((($srcbuf$38)) + 1|0);
        $1466 = HEAP8[$1465>>0]|0;
        $1467 = $1466&255;
        $1468 = ((($srcbuf$38)) + 2|0);
        $1469 = ($1466<<24>>24)==(0);
        if ($1469) {
         $1553 = ($1464|0)==(0);
         if ($1553) {
          break L44;
         } else {
          $ofs67$1 = $1464;$srcbuf$39 = $1468;
         }
        } else {
         $1470 = $1464 << 1;
         $1471 = (($dstbuf$24$ph) + ($1470)|0);
         $dst71$0151 = $1471;$i69$0149 = 0;$src70$0150 = $1468;
         while(1) {
          $1472 = HEAP16[$src70$0150>>1]|0;
          $1473 = $1472&65535;
          $1474 = HEAP16[$dst71$0151>>1]|0;
          $1475 = $1474&65535;
          $1476 = $1176 & $1473;
          $1477 = HEAP32[$1459>>2]|0;
          $1478 = $1477 & 255;
          $1479 = $1476 >>> $1478;
          $1480 = HEAP32[$1460>>2]|0;
          $1481 = $1480 & 255;
          $1482 = (125924 + ($1481<<2)|0);
          $1483 = HEAP32[$1482>>2]|0;
          $1484 = (($1483) + ($1479)|0);
          $1485 = HEAP8[$1484>>0]|0;
          $1486 = $1485&255;
          $1487 = $1178 & $1473;
          $1488 = $1477 >>> 8;
          $1489 = $1488 & 255;
          $1490 = $1487 >>> $1489;
          $1491 = $1480 >>> 8;
          $1492 = $1491 & 255;
          $1493 = (125924 + ($1492<<2)|0);
          $1494 = HEAP32[$1493>>2]|0;
          $1495 = (($1494) + ($1490)|0);
          $1496 = HEAP8[$1495>>0]|0;
          $1497 = $1496&255;
          $1498 = $1181 & $1473;
          $1499 = $1477 >>> 16;
          $1500 = $1499 & 255;
          $1501 = $1498 >>> $1500;
          $1502 = $1480 >>> 16;
          $1503 = $1502 & 255;
          $1504 = (125924 + ($1503<<2)|0);
          $1505 = HEAP32[$1504>>2]|0;
          $1506 = (($1505) + ($1501)|0);
          $1507 = HEAP8[$1506>>0]|0;
          $1508 = $1507&255;
          $1509 = $1475 & $1176;
          $1510 = $1509 >>> $1478;
          $1511 = (($1483) + ($1510)|0);
          $1512 = HEAP8[$1511>>0]|0;
          $1513 = $1512&255;
          $1514 = $1178 & $1475;
          $1515 = $1514 >>> $1489;
          $1516 = (($1494) + ($1515)|0);
          $1517 = HEAP8[$1516>>0]|0;
          $1518 = $1517&255;
          $1519 = $1181 & $1475;
          $1520 = $1519 >>> $1500;
          $1521 = (($1505) + ($1520)|0);
          $1522 = HEAP8[$1521>>0]|0;
          $1523 = $1522&255;
          $1524 = (($1486) - ($1513))|0;
          $1525 = Math_imul($1524, $91)|0;
          $1526 = $1525 >>> 8;
          $1527 = (($1526) + ($1513))|0;
          $1528 = (($1497) - ($1518))|0;
          $1529 = Math_imul($1528, $91)|0;
          $1530 = $1529 >>> 8;
          $1531 = (($1530) + ($1518))|0;
          $1532 = (($1508) - ($1523))|0;
          $1533 = Math_imul($1532, $91)|0;
          $1534 = $1533 >>> 8;
          $1535 = (($1534) + ($1523))|0;
          $1536 = $1527 >>> $1481;
          $1537 = $1536 << $1478;
          $1538 = $1531 >>> $1492;
          $1539 = $1538 << $1489;
          $1540 = $1535 >>> $1503;
          $1541 = $1540 << $1500;
          $1542 = HEAP32[$1461>>2]|0;
          $1543 = $1537 | $1542;
          $1544 = $1543 | $1539;
          $1545 = $1544 | $1541;
          $1546 = $1545&65535;
          HEAP16[$dst71$0151>>1] = $1546;
          $1547 = ((($src70$0150)) + 2|0);
          $1548 = ((($dst71$0151)) + 2|0);
          $1549 = (($i69$0149) + 1)|0;
          $exitcond261 = ($1549|0)==($1467|0);
          if ($exitcond261) {
           break;
          } else {
           $dst71$0151 = $1548;$i69$0149 = $1549;$src70$0150 = $1547;
          }
         }
         $1550 = $1467 << 1;
         $1551 = (($1468) + ($1550)|0);
         $1552 = (($1464) + ($1467))|0;
         $ofs67$1 = $1552;$srcbuf$39 = $1551;
        }
        $1554 = ($ofs67$1|0)==($1|0);
        if ($1554) {
         $srcbuf$39$lcssa = $srcbuf$39;
         break;
        } else {
         $ofs67$0 = $ofs67$1;$srcbuf$38 = $srcbuf$39;
        }
       }
       $1555 = HEAP32[$12>>2]|0;
       $1556 = (($dstbuf$24$ph) + ($1555)|0);
       $1557 = (($linecount66$0$ph) + -1)|0;
       $1558 = ($1557|0)==(0);
       if ($1558) {
        break L44;
       } else {
        $dstbuf$24$ph = $1556;$linecount66$0$ph = $1557;$srcbuf$38$ph = $srcbuf$39$lcssa;
       }
      }
      break;
     }
     case 3:  {
      $1559 = ((($srcrect)) + 12|0);
      $1560 = HEAP32[$1559>>2]|0;
      $1561 = ((($17)) + 12|0);
      $1562 = ((($17)) + 32|0);
      $1563 = ((($17)) + 28|0);
      $1564 = ((($17)) + 16|0);
      $1565 = ((($17)) + 20|0);
      $1566 = ((($17)) + 24|0);
      $dstbuf$27$ph = $22;$linecount75$0$ph = $1560;$srcbuf$41$ph = $srcbuf$13$ph;
      while(1) {
       $ofs76$0 = 0;$srcbuf$41 = $srcbuf$41$ph;
       while(1) {
        $1567 = HEAP8[$srcbuf$41>>0]|0;
        $1568 = $1567&255;
        $1569 = (($1568) + ($ofs76$0))|0;
        $1570 = ((($srcbuf$41)) + 1|0);
        $1571 = HEAP8[$1570>>0]|0;
        $1572 = $1571&255;
        $1573 = ((($srcbuf$41)) + 2|0);
        $1574 = ($1571<<24>>24)==(0);
        if ($1574) {
         $1685 = ($1569|0)==(0);
         if ($1685) {
          break L44;
         } else {
          $ofs76$1 = $1569;$srcbuf$42 = $1573;
         }
        } else {
         $1575 = ($1569*3)|0;
         $1576 = (($dstbuf$27$ph) + ($1575)|0);
         $dst80$0192 = $1576;$i78$0190 = 0;$src79$0191 = $1573;
         while(1) {
          $1577 = ((($src79$0191)) + 2|0);
          $1578 = HEAP8[$1577>>0]|0;
          $1579 = $1578&255;
          $1580 = $1579 << 16;
          $1581 = ((($src79$0191)) + 1|0);
          $1582 = HEAP8[$1581>>0]|0;
          $1583 = $1582&255;
          $1584 = $1583 << 8;
          $1585 = $1584 | $1580;
          $1586 = HEAP8[$src79$0191>>0]|0;
          $1587 = $1586&255;
          $1588 = $1585 | $1587;
          $1589 = ((($dst80$0192)) + 2|0);
          $1590 = HEAP8[$1589>>0]|0;
          $1591 = $1590&255;
          $1592 = $1591 << 16;
          $1593 = ((($dst80$0192)) + 1|0);
          $1594 = HEAP8[$1593>>0]|0;
          $1595 = $1594&255;
          $1596 = $1595 << 8;
          $1597 = $1596 | $1592;
          $1598 = HEAP8[$dst80$0192>>0]|0;
          $1599 = $1598&255;
          $1600 = $1597 | $1599;
          $1601 = HEAP32[$1561>>2]|0;
          $1602 = $1601 & $1588;
          $1603 = HEAP32[$1562>>2]|0;
          $1604 = $1603 & 255;
          $1605 = $1602 >>> $1604;
          $1606 = HEAP32[$1563>>2]|0;
          $1607 = $1606 & 255;
          $1608 = (125924 + ($1607<<2)|0);
          $1609 = HEAP32[$1608>>2]|0;
          $1610 = (($1609) + ($1605)|0);
          $1611 = HEAP8[$1610>>0]|0;
          $1612 = $1611&255;
          $1613 = HEAP32[$1564>>2]|0;
          $1614 = $1613 & $1588;
          $1615 = $1603 >>> 8;
          $1616 = $1615 & 255;
          $1617 = $1614 >>> $1616;
          $1618 = $1606 >>> 8;
          $1619 = $1618 & 255;
          $1620 = (125924 + ($1619<<2)|0);
          $1621 = HEAP32[$1620>>2]|0;
          $1622 = (($1621) + ($1617)|0);
          $1623 = HEAP8[$1622>>0]|0;
          $1624 = $1623&255;
          $1625 = HEAP32[$1565>>2]|0;
          $1626 = $1625 & $1588;
          $1627 = $1603 >>> 16;
          $1628 = $1627 & 255;
          $1629 = $1626 >>> $1628;
          $1630 = $1606 >>> 16;
          $1631 = $1630 & 255;
          $1632 = (125924 + ($1631<<2)|0);
          $1633 = HEAP32[$1632>>2]|0;
          $1634 = (($1633) + ($1629)|0);
          $1635 = HEAP8[$1634>>0]|0;
          $1636 = $1635&255;
          $1637 = $1600 & $1601;
          $1638 = $1637 >>> $1604;
          $1639 = (($1609) + ($1638)|0);
          $1640 = HEAP8[$1639>>0]|0;
          $1641 = $1640&255;
          $1642 = $1613 & $1600;
          $1643 = $1642 >>> $1616;
          $1644 = (($1621) + ($1643)|0);
          $1645 = HEAP8[$1644>>0]|0;
          $1646 = $1645&255;
          $1647 = $1625 & $1600;
          $1648 = $1647 >>> $1628;
          $1649 = (($1633) + ($1648)|0);
          $1650 = HEAP8[$1649>>0]|0;
          $1651 = $1650&255;
          $1652 = (($1612) - ($1641))|0;
          $1653 = Math_imul($1652, $91)|0;
          $1654 = $1653 >>> 8;
          $1655 = (($1654) + ($1641))|0;
          $1656 = (($1624) - ($1646))|0;
          $1657 = Math_imul($1656, $91)|0;
          $1658 = $1657 >>> 8;
          $1659 = (($1658) + ($1646))|0;
          $1660 = (($1636) - ($1651))|0;
          $1661 = Math_imul($1660, $91)|0;
          $1662 = $1661 >>> 8;
          $1663 = (($1662) + ($1651))|0;
          $1664 = $1655 >>> $1607;
          $1665 = $1664 << $1604;
          $1666 = $1659 >>> $1619;
          $1667 = $1666 << $1616;
          $1668 = $1663 >>> $1631;
          $1669 = $1668 << $1628;
          $1670 = HEAP32[$1566>>2]|0;
          $1671 = $1665 | $1670;
          $1672 = $1671 | $1667;
          $1673 = $1672 | $1669;
          $1674 = $1673&255;
          HEAP8[$dst80$0192>>0] = $1674;
          $1675 = $1673 >>> 8;
          $1676 = $1675&255;
          HEAP8[$1593>>0] = $1676;
          $1677 = $1673 >>> 16;
          $1678 = $1677&255;
          HEAP8[$1589>>0] = $1678;
          $1679 = ((($src79$0191)) + 3|0);
          $1680 = ((($dst80$0192)) + 3|0);
          $1681 = (($i78$0190) + 1)|0;
          $exitcond271 = ($1681|0)==($1572|0);
          if ($exitcond271) {
           break;
          } else {
           $dst80$0192 = $1680;$i78$0190 = $1681;$src79$0191 = $1679;
          }
         }
         $1682 = ($1572*3)|0;
         $1683 = (($1573) + ($1682)|0);
         $1684 = (($1569) + ($1572))|0;
         $ofs76$1 = $1684;$srcbuf$42 = $1683;
        }
        $1686 = ($ofs76$1|0)==($1|0);
        if ($1686) {
         $srcbuf$42$lcssa = $srcbuf$42;
         break;
        } else {
         $ofs76$0 = $ofs76$1;$srcbuf$41 = $srcbuf$42;
        }
       }
       $1687 = HEAP32[$12>>2]|0;
       $1688 = (($dstbuf$27$ph) + ($1687)|0);
       $1689 = (($linecount75$0$ph) + -1)|0;
       $1690 = ($1689|0)==(0);
       if ($1690) {
        break L44;
       } else {
        $dstbuf$27$ph = $1688;$linecount75$0$ph = $1689;$srcbuf$41$ph = $srcbuf$42$lcssa;
       }
      }
      break;
     }
     default: {
      break L44;
     }
     }
    }
   }
   $98 = ((($surf_dst)) + 4|0);
   $99 = HEAP32[$98>>2]|0;
   $100 = ($90<<24>>24)==(-1);
   $101 = ((($99)) + 9|0);
   $102 = HEAP8[$101>>0]|0;
   $103 = $102&255;
   if ($100) {
    switch ($103|0) {
    case 1:  {
     $104 = ((($srcrect)) + 12|0);
     $105 = HEAP32[$104>>2]|0;
     $106 = ((($srcrect)) + 8|0);
     $107 = HEAP32[$106>>2]|0;
     $108 = (($107) + ($92))|0;
     $109 = (0 - ($92))|0;
     $110 = (($22) + ($109)|0);
     $$0$ph$i = $srcbuf$13$ph;$$013$ph$i = $110;$linecount$0$ph$i = $105;
     while(1) {
      $$0$i = $$0$ph$i;$ofs$0$i = 0;
      while(1) {
       $111 = HEAP8[$$0$i>>0]|0;
       $112 = $111&255;
       $113 = (($112) + ($ofs$0$i))|0;
       $114 = ((($$0$i)) + 1|0);
       $115 = HEAP8[$114>>0]|0;
       $116 = $115&255;
       $117 = ((($$0$i)) + 2|0);
       $118 = ($115<<24>>24)==(0);
       if ($118) {
        $131 = ($113|0)==(0);
        if ($131) {
         break L44;
        } else {
         $$1$i = $117;$ofs$1$i = $113;
        }
       } else {
        $119 = ($113|0)<($108|0);
        do {
         if ($119) {
          $120 = (($92) - ($113))|0;
          $121 = ($120|0)>(0);
          if ($121) {
           $122 = (($116) - ($120))|0;
           $123 = ($122|0)<(1);
           if ($123) {
            break;
           } else {
            $len$0$i = $122;$start$0$i = $120;
           }
          } else {
           $len$0$i = $116;$start$0$i = 0;
          }
          $124 = (($start$0$i) + ($113))|0;
          $125 = (($108) - ($124))|0;
          $126 = ($len$0$i|0)>($125|0);
          $$len$0$i = $126 ? $125 : $len$0$i;
          $127 = (($$013$ph$i) + ($124)|0);
          $128 = (($117) + ($start$0$i)|0);
          (_SDL_memcpy($127,$128,$$len$0$i)|0);
         }
        } while(0);
        $129 = (($117) + ($116)|0);
        $130 = (($113) + ($116))|0;
        $$1$i = $129;$ofs$1$i = $130;
       }
       $132 = ($ofs$1$i|0)==($1|0);
       if ($132) {
        $$1$i$lcssa = $$1$i;
        break;
       } else {
        $$0$i = $$1$i;$ofs$0$i = $ofs$1$i;
       }
      }
      $133 = HEAP32[$12>>2]|0;
      $134 = (($$013$ph$i) + ($133)|0);
      $135 = (($linecount$0$ph$i) + -1)|0;
      $136 = ($135|0)==(0);
      if ($136) {
       break L44;
      } else {
       $$0$ph$i = $$1$i$lcssa;$$013$ph$i = $134;$linecount$0$ph$i = $135;
      }
     }
     break;
    }
    case 2:  {
     $137 = ((($srcrect)) + 12|0);
     $138 = HEAP32[$137>>2]|0;
     $139 = ((($srcrect)) + 8|0);
     $140 = HEAP32[$139>>2]|0;
     $141 = (($140) + ($92))|0;
     $142 = $92 << 1;
     $143 = (0 - ($142))|0;
     $144 = (($22) + ($143)|0);
     $$3$ph$i = $srcbuf$13$ph;$$316$ph$i = $144;$linecount2$0$ph$i = $138;
     while(1) {
      $$3$i = $$3$ph$i;$ofs3$0$i = 0;
      while(1) {
       $145 = HEAP8[$$3$i>>0]|0;
       $146 = $145&255;
       $147 = (($146) + ($ofs3$0$i))|0;
       $148 = ((($$3$i)) + 1|0);
       $149 = HEAP8[$148>>0]|0;
       $150 = $149&255;
       $151 = ((($$3$i)) + 2|0);
       $152 = ($149<<24>>24)==(0);
       if ($152) {
        $169 = ($147|0)==(0);
        if ($169) {
         break L44;
        } else {
         $$4$i = $151;$ofs3$1$i = $147;
        }
       } else {
        $153 = ($147|0)<($141|0);
        do {
         if ($153) {
          $154 = (($92) - ($147))|0;
          $155 = ($154|0)>(0);
          if ($155) {
           $156 = (($150) - ($154))|0;
           $157 = ($156|0)<(1);
           if ($157) {
            break;
           } else {
            $len8$0$i = $156;$start7$0$i = $154;
           }
          } else {
           $len8$0$i = $150;$start7$0$i = 0;
          }
          $158 = (($start7$0$i) + ($147))|0;
          $159 = (($141) - ($158))|0;
          $160 = ($len8$0$i|0)>($159|0);
          $$len8$0$i = $160 ? $159 : $len8$0$i;
          $161 = $158 << 1;
          $162 = (($$316$ph$i) + ($161)|0);
          $163 = $start7$0$i << 1;
          $164 = (($151) + ($163)|0);
          $165 = $$len8$0$i << 1;
          (_SDL_memcpy($162,$164,$165)|0);
         }
        } while(0);
        $166 = $150 << 1;
        $167 = (($151) + ($166)|0);
        $168 = (($147) + ($150))|0;
        $$4$i = $167;$ofs3$1$i = $168;
       }
       $170 = ($ofs3$1$i|0)==($1|0);
       if ($170) {
        $$4$i$lcssa = $$4$i;
        break;
       } else {
        $$3$i = $$4$i;$ofs3$0$i = $ofs3$1$i;
       }
      }
      $171 = HEAP32[$12>>2]|0;
      $172 = (($$316$ph$i) + ($171)|0);
      $173 = (($linecount2$0$ph$i) + -1)|0;
      $174 = ($173|0)==(0);
      if ($174) {
       break L44;
      } else {
       $$3$ph$i = $$4$i$lcssa;$$316$ph$i = $172;$linecount2$0$ph$i = $173;
      }
     }
     break;
    }
    case 3:  {
     $175 = ((($srcrect)) + 12|0);
     $176 = HEAP32[$175>>2]|0;
     $177 = ((($srcrect)) + 8|0);
     $178 = HEAP32[$177>>2]|0;
     $179 = (($178) + ($92))|0;
     $180 = Math_imul($92, -3)|0;
     $181 = (($22) + ($180)|0);
     $$6$ph$i = $srcbuf$13$ph;$$619$ph$i = $181;$linecount12$0$ph$i = $176;
     while(1) {
      $$6$i = $$6$ph$i;$ofs13$0$i = 0;
      while(1) {
       $182 = HEAP8[$$6$i>>0]|0;
       $183 = $182&255;
       $184 = (($183) + ($ofs13$0$i))|0;
       $185 = ((($$6$i)) + 1|0);
       $186 = HEAP8[$185>>0]|0;
       $187 = $186&255;
       $188 = ((($$6$i)) + 2|0);
       $189 = ($186<<24>>24)==(0);
       if ($189) {
        $206 = ($184|0)==(0);
        if ($206) {
         break L44;
        } else {
         $$7$i = $188;$ofs13$1$i = $184;
        }
       } else {
        $190 = ($184|0)<($179|0);
        do {
         if ($190) {
          $191 = (($92) - ($184))|0;
          $192 = ($191|0)>(0);
          if ($192) {
           $193 = (($187) - ($191))|0;
           $194 = ($193|0)<(1);
           if ($194) {
            break;
           } else {
            $len18$0$i = $193;$start17$0$i = $191;
           }
          } else {
           $len18$0$i = $187;$start17$0$i = 0;
          }
          $195 = (($start17$0$i) + ($184))|0;
          $196 = (($179) - ($195))|0;
          $197 = ($len18$0$i|0)>($196|0);
          $$len18$0$i = $197 ? $196 : $len18$0$i;
          $198 = ($195*3)|0;
          $199 = (($$619$ph$i) + ($198)|0);
          $200 = ($start17$0$i*3)|0;
          $201 = (($188) + ($200)|0);
          $202 = ($$len18$0$i*3)|0;
          (_SDL_memcpy($199,$201,$202)|0);
         }
        } while(0);
        $203 = ($187*3)|0;
        $204 = (($188) + ($203)|0);
        $205 = (($184) + ($187))|0;
        $$7$i = $204;$ofs13$1$i = $205;
       }
       $207 = ($ofs13$1$i|0)==($1|0);
       if ($207) {
        $$7$i$lcssa = $$7$i;
        break;
       } else {
        $$6$i = $$7$i;$ofs13$0$i = $ofs13$1$i;
       }
      }
      $208 = HEAP32[$12>>2]|0;
      $209 = (($$619$ph$i) + ($208)|0);
      $210 = (($linecount12$0$ph$i) + -1)|0;
      $211 = ($210|0)==(0);
      if ($211) {
       break L44;
      } else {
       $$6$ph$i = $$7$i$lcssa;$$619$ph$i = $209;$linecount12$0$ph$i = $210;
      }
     }
     break;
    }
    case 4:  {
     $212 = ((($srcrect)) + 12|0);
     $213 = HEAP32[$212>>2]|0;
     $214 = ((($srcrect)) + 8|0);
     $215 = HEAP32[$214>>2]|0;
     $216 = (($215) + ($92))|0;
     $217 = $92 << 2;
     $218 = (0 - ($217))|0;
     $219 = (($22) + ($218)|0);
     $$9$ph$i = $srcbuf$13$ph;$$922$ph$i = $219;$linecount22$0$ph$i = $213;
     while(1) {
      $$9$i = $$9$ph$i;$ofs23$0$i = 0;
      while(1) {
       $220 = HEAP16[$$9$i>>1]|0;
       $221 = $220&65535;
       $222 = (($221) + ($ofs23$0$i))|0;
       $223 = ((($$9$i)) + 2|0);
       $224 = HEAP16[$223>>1]|0;
       $225 = $224&65535;
       $226 = ((($$9$i)) + 4|0);
       $227 = ($224<<16>>16)==(0);
       if ($227) {
        $244 = ($222|0)==(0);
        if ($244) {
         break L44;
        } else {
         $$10$i = $226;$ofs23$1$i = $222;
        }
       } else {
        $228 = ($222|0)<($216|0);
        do {
         if ($228) {
          $229 = (($92) - ($222))|0;
          $230 = ($229|0)>(0);
          if ($230) {
           $231 = (($225) - ($229))|0;
           $232 = ($231|0)<(1);
           if ($232) {
            break;
           } else {
            $len28$0$i = $231;$start27$0$i = $229;
           }
          } else {
           $len28$0$i = $225;$start27$0$i = 0;
          }
          $233 = (($start27$0$i) + ($222))|0;
          $234 = (($216) - ($233))|0;
          $235 = ($len28$0$i|0)>($234|0);
          $$len28$0$i = $235 ? $234 : $len28$0$i;
          $236 = $233 << 2;
          $237 = (($$922$ph$i) + ($236)|0);
          $238 = $start27$0$i << 2;
          $239 = (($226) + ($238)|0);
          $240 = $$len28$0$i << 2;
          (_SDL_memcpy($237,$239,$240)|0);
         }
        } while(0);
        $241 = $225 << 2;
        $242 = (($226) + ($241)|0);
        $243 = (($222) + ($225))|0;
        $$10$i = $242;$ofs23$1$i = $243;
       }
       $245 = ($ofs23$1$i|0)==($1|0);
       if ($245) {
        $$10$i$lcssa = $$10$i;
        break;
       } else {
        $$9$i = $$10$i;$ofs23$0$i = $ofs23$1$i;
       }
      }
      $246 = HEAP32[$12>>2]|0;
      $247 = (($$922$ph$i) + ($246)|0);
      $248 = (($linecount22$0$ph$i) + -1)|0;
      $249 = ($248|0)==(0);
      if ($249) {
       break L44;
      } else {
       $$9$ph$i = $$10$i$lcssa;$$922$ph$i = $247;$linecount22$0$ph$i = $248;
      }
     }
     break;
    }
    default: {
     break L44;
    }
    }
   }
   switch ($103|0) {
   case 4:  {
    $857 = ((($99)) + 12|0);
    $858 = HEAP32[$857>>2]|0;
    $859 = ((($99)) + 16|0);
    $860 = HEAP32[$859>>2]|0;
    $861 = $860 | $858;
    $862 = ((($99)) + 20|0);
    $863 = HEAP32[$862>>2]|0;
    $864 = $861 | $863;
    $865 = ($864|0)==(16777215);
    if ($865) {
     $866 = ($860|0)==(65280);
     $867 = ($858|0)==(65280);
     $or$cond81$i = $867 | $866;
     $868 = ($863|0)==(65280);
     $or$cond82$i = $or$cond81$i | $868;
     if ($or$cond82$i) {
      $869 = ($90<<24>>24)==(-128);
      $870 = ((($srcrect)) + 12|0);
      $871 = HEAP32[$870>>2]|0;
      $872 = ((($srcrect)) + 8|0);
      $873 = HEAP32[$872>>2]|0;
      $874 = (($873) + ($92))|0;
      $875 = $92 << 2;
      $876 = (0 - ($875))|0;
      $877 = (($22) + ($876)|0);
      if ($869) {
       $$30$ph$i = $srcbuf$13$ph;$$3043$ph$i = $877;$linecount135$0$ph$i = $871;
       while(1) {
        $$30$i = $$30$ph$i;$ofs136$0$i = 0;
        while(1) {
         $878 = HEAP16[$$30$i>>1]|0;
         $879 = $878&65535;
         $880 = (($879) + ($ofs136$0$i))|0;
         $881 = ((($$30$i)) + 2|0);
         $882 = HEAP16[$881>>1]|0;
         $883 = $882&65535;
         $884 = ((($$30$i)) + 4|0);
         $885 = ($882<<16>>16)==(0);
         if ($885) {
          $915 = ($880|0)==(0);
          if ($915) {
           break L44;
          } else {
           $$31$i = $884;$ofs136$1$i = $880;
          }
         } else {
          $886 = ($880|0)<($874|0);
          do {
           if ($886) {
            $887 = (($92) - ($880))|0;
            $888 = ($887|0)>(0);
            if ($888) {
             $889 = (($883) - ($887))|0;
             $890 = ($889|0)<(1);
             if ($890) {
              break;
             } else {
              $len141$0$i = $889;$start140$0$i = $887;
             }
            } else {
             $len141$0$i = $883;$start140$0$i = 0;
            }
            $891 = (($start140$0$i) + ($880))|0;
            $892 = (($874) - ($891))|0;
            $893 = ($len141$0$i|0)>($892|0);
            $$len141$0$i = $893 ? $892 : $len141$0$i;
            $894 = ($$len141$0$i|0)>(0);
            if ($894) {
             $895 = $891 << 2;
             $896 = (($$3043$ph$i) + ($895)|0);
             $897 = $start140$0$i << 2;
             $898 = (($884) + ($897)|0);
             $dst145$0192$i = $896;$i143$0190$i = 0;$src144$0191$i = $898;
             while(1) {
              $899 = ((($src144$0191$i)) + 4|0);
              $900 = HEAP32[$src144$0191$i>>2]|0;
              $901 = HEAP32[$dst145$0192$i>>2]|0;
              $902 = $900 & 16711422;
              $903 = $901 & 16711422;
              $904 = (($903) + ($902))|0;
              $905 = $904 >>> 1;
              $906 = $900 & 65793;
              $907 = $906 & $901;
              $908 = (($905) + ($907))|0;
              $909 = ((($dst145$0192$i)) + 4|0);
              HEAP32[$dst145$0192$i>>2] = $908;
              $910 = (($i143$0190$i) + 1)|0;
              $911 = ($910|0)<($$len141$0$i|0);
              if ($911) {
               $dst145$0192$i = $909;$i143$0190$i = $910;$src144$0191$i = $899;
              } else {
               break;
              }
             }
            }
           }
          } while(0);
          $912 = $883 << 2;
          $913 = (($884) + ($912)|0);
          $914 = (($880) + ($883))|0;
          $$31$i = $913;$ofs136$1$i = $914;
         }
         $916 = ($ofs136$1$i|0)==($1|0);
         if ($916) {
          $$31$i$lcssa = $$31$i;
          break;
         } else {
          $$30$i = $$31$i;$ofs136$0$i = $ofs136$1$i;
         }
        }
        $917 = HEAP32[$12>>2]|0;
        $918 = (($$3043$ph$i) + ($917)|0);
        $919 = (($linecount135$0$ph$i) + -1)|0;
        $920 = ($919|0)==(0);
        if ($920) {
         break L44;
        } else {
         $$30$ph$i = $$31$i$lcssa;$$3043$ph$i = $918;$linecount135$0$ph$i = $919;
        }
       }
      } else {
       $$33$ph$i = $srcbuf$13$ph;$$3346$ph$i = $877;$linecount150$0$ph$i = $871;
       while(1) {
        $$33$i = $$33$ph$i;$ofs151$0$i = 0;
        while(1) {
         $921 = HEAP16[$$33$i>>1]|0;
         $922 = $921&65535;
         $923 = (($922) + ($ofs151$0$i))|0;
         $924 = ((($$33$i)) + 2|0);
         $925 = HEAP16[$924>>1]|0;
         $926 = $925&65535;
         $927 = ((($$33$i)) + 4|0);
         $928 = ($925<<16>>16)==(0);
         if ($928) {
          $966 = ($923|0)==(0);
          if ($966) {
           break L44;
          } else {
           $$34$i = $927;$ofs151$1$i = $923;
          }
         } else {
          $929 = ($923|0)<($874|0);
          do {
           if ($929) {
            $930 = (($92) - ($923))|0;
            $931 = ($930|0)>(0);
            if ($931) {
             $932 = (($926) - ($930))|0;
             $933 = ($932|0)<(1);
             if ($933) {
              break;
             } else {
              $len156$0$i = $932;$start155$0$i = $930;
             }
            } else {
             $len156$0$i = $926;$start155$0$i = 0;
            }
            $934 = (($start155$0$i) + ($923))|0;
            $935 = (($874) - ($934))|0;
            $936 = ($len156$0$i|0)>($935|0);
            $$len156$0$i = $936 ? $935 : $len156$0$i;
            $937 = ($$len156$0$i|0)>(0);
            if ($937) {
             $938 = $934 << 2;
             $939 = (($$3346$ph$i) + ($938)|0);
             $940 = $start155$0$i << 2;
             $941 = (($927) + ($940)|0);
             $dst160$0195$i = $939;$i158$0193$i = 0;$src159$0194$i = $941;
             while(1) {
              $942 = ((($src159$0194$i)) + 4|0);
              $943 = HEAP32[$src159$0194$i>>2]|0;
              $944 = HEAP32[$dst160$0195$i>>2]|0;
              $945 = $943 & 16711935;
              $946 = $944 & 16711935;
              $947 = (($945) - ($946))|0;
              $948 = Math_imul($947, $91)|0;
              $949 = $948 >>> 8;
              $950 = (($949) + ($946))|0;
              $951 = $950 & 16711935;
              $952 = $943 & 65280;
              $953 = $944 & 65280;
              $954 = (($952) - ($953))|0;
              $955 = Math_imul($954, $91)|0;
              $956 = $955 >>> 8;
              $957 = (($956) + ($953))|0;
              $958 = $957 & 65280;
              $959 = $951 | $958;
              $960 = ((($dst160$0195$i)) + 4|0);
              HEAP32[$dst160$0195$i>>2] = $959;
              $961 = (($i158$0193$i) + 1)|0;
              $962 = ($961|0)<($$len156$0$i|0);
              if ($962) {
               $dst160$0195$i = $960;$i158$0193$i = $961;$src159$0194$i = $942;
              } else {
               break;
              }
             }
            }
           }
          } while(0);
          $963 = $926 << 2;
          $964 = (($927) + ($963)|0);
          $965 = (($923) + ($926))|0;
          $$34$i = $964;$ofs151$1$i = $965;
         }
         $967 = ($ofs151$1$i|0)==($1|0);
         if ($967) {
          $$34$i$lcssa = $$34$i;
          break;
         } else {
          $$33$i = $$34$i;$ofs151$0$i = $ofs151$1$i;
         }
        }
        $968 = HEAP32[$12>>2]|0;
        $969 = (($$3346$ph$i) + ($968)|0);
        $970 = (($linecount150$0$ph$i) + -1)|0;
        $971 = ($970|0)==(0);
        if ($971) {
         break L44;
        } else {
         $$33$ph$i = $$34$i$lcssa;$$3346$ph$i = $969;$linecount150$0$ph$i = $970;
        }
       }
      }
     }
    }
    $972 = ((($srcrect)) + 12|0);
    $973 = HEAP32[$972>>2]|0;
    $974 = ((($srcrect)) + 8|0);
    $975 = HEAP32[$974>>2]|0;
    $976 = (($975) + ($92))|0;
    $977 = $92 << 2;
    $978 = (0 - ($977))|0;
    $979 = (($22) + ($978)|0);
    $980 = ((($99)) + 32|0);
    $981 = ((($99)) + 28|0);
    $982 = ((($99)) + 24|0);
    $$36$ph$i = $srcbuf$13$ph;$$3649$ph$i = $979;$linecount165$0$ph$i = $973;
    while(1) {
     $$36$i = $$36$ph$i;$ofs166$0$i = 0;
     while(1) {
      $983 = HEAP16[$$36$i>>1]|0;
      $984 = $983&65535;
      $985 = (($984) + ($ofs166$0$i))|0;
      $986 = ((($$36$i)) + 2|0);
      $987 = HEAP16[$986>>1]|0;
      $988 = $987&65535;
      $989 = ((($$36$i)) + 4|0);
      $990 = ($987<<16>>16)==(0);
      if ($990) {
       $1086 = ($985|0)==(0);
       if ($1086) {
        break L44;
       } else {
        $$37$i = $989;$ofs166$1$i = $985;
       }
      } else {
       $991 = ($985|0)<($976|0);
       do {
        if ($991) {
         $992 = (($92) - ($985))|0;
         $993 = ($992|0)>(0);
         if ($993) {
          $994 = (($988) - ($992))|0;
          $995 = ($994|0)<(1);
          if ($995) {
           break;
          } else {
           $len171$0$i = $994;$start170$0$i = $992;
          }
         } else {
          $len171$0$i = $988;$start170$0$i = 0;
         }
         $996 = (($start170$0$i) + ($985))|0;
         $997 = (($976) - ($996))|0;
         $998 = ($len171$0$i|0)>($997|0);
         $$len171$0$i = $998 ? $997 : $len171$0$i;
         $999 = ($$len171$0$i|0)>(0);
         if ($999) {
          $1000 = $996 << 2;
          $1001 = (($$3649$ph$i) + ($1000)|0);
          $1002 = $start170$0$i << 2;
          $1003 = (($989) + ($1002)|0);
          $dst175$0199$i = $1001;$i173$0197$i = 0;$src174$0198$i = $1003;
          while(1) {
           $1004 = HEAP32[$src174$0198$i>>2]|0;
           $1005 = HEAP32[$dst175$0199$i>>2]|0;
           $1006 = HEAP32[$857>>2]|0;
           $1007 = $1006 & $1004;
           $1008 = HEAP32[$980>>2]|0;
           $1009 = $1008 & 255;
           $1010 = $1007 >>> $1009;
           $1011 = HEAP32[$981>>2]|0;
           $1012 = $1011 & 255;
           $1013 = (125924 + ($1012<<2)|0);
           $1014 = HEAP32[$1013>>2]|0;
           $1015 = (($1014) + ($1010)|0);
           $1016 = HEAP8[$1015>>0]|0;
           $1017 = $1016&255;
           $1018 = HEAP32[$859>>2]|0;
           $1019 = $1018 & $1004;
           $1020 = $1008 >>> 8;
           $1021 = $1020 & 255;
           $1022 = $1019 >>> $1021;
           $1023 = $1011 >>> 8;
           $1024 = $1023 & 255;
           $1025 = (125924 + ($1024<<2)|0);
           $1026 = HEAP32[$1025>>2]|0;
           $1027 = (($1026) + ($1022)|0);
           $1028 = HEAP8[$1027>>0]|0;
           $1029 = $1028&255;
           $1030 = HEAP32[$862>>2]|0;
           $1031 = $1030 & $1004;
           $1032 = $1008 >>> 16;
           $1033 = $1032 & 255;
           $1034 = $1031 >>> $1033;
           $1035 = $1011 >>> 16;
           $1036 = $1035 & 255;
           $1037 = (125924 + ($1036<<2)|0);
           $1038 = HEAP32[$1037>>2]|0;
           $1039 = (($1038) + ($1034)|0);
           $1040 = HEAP8[$1039>>0]|0;
           $1041 = $1040&255;
           $1042 = $1006 & $1005;
           $1043 = $1042 >>> $1009;
           $1044 = (($1014) + ($1043)|0);
           $1045 = HEAP8[$1044>>0]|0;
           $1046 = $1045&255;
           $1047 = $1018 & $1005;
           $1048 = $1047 >>> $1021;
           $1049 = (($1026) + ($1048)|0);
           $1050 = HEAP8[$1049>>0]|0;
           $1051 = $1050&255;
           $1052 = $1030 & $1005;
           $1053 = $1052 >>> $1033;
           $1054 = (($1038) + ($1053)|0);
           $1055 = HEAP8[$1054>>0]|0;
           $1056 = $1055&255;
           $1057 = (($1017) - ($1046))|0;
           $1058 = Math_imul($1057, $91)|0;
           $1059 = $1058 >>> 8;
           $1060 = (($1059) + ($1046))|0;
           $1061 = (($1029) - ($1051))|0;
           $1062 = Math_imul($1061, $91)|0;
           $1063 = $1062 >>> 8;
           $1064 = (($1063) + ($1051))|0;
           $1065 = (($1041) - ($1056))|0;
           $1066 = Math_imul($1065, $91)|0;
           $1067 = $1066 >>> 8;
           $1068 = (($1067) + ($1056))|0;
           $1069 = $1060 >>> $1012;
           $1070 = $1069 << $1009;
           $1071 = $1064 >>> $1024;
           $1072 = $1071 << $1021;
           $1073 = $1068 >>> $1036;
           $1074 = $1073 << $1033;
           $1075 = HEAP32[$982>>2]|0;
           $1076 = $1070 | $1075;
           $1077 = $1076 | $1072;
           $1078 = $1077 | $1074;
           HEAP32[$dst175$0199$i>>2] = $1078;
           $1079 = ((($src174$0198$i)) + 4|0);
           $1080 = ((($dst175$0199$i)) + 4|0);
           $1081 = (($i173$0197$i) + 1)|0;
           $1082 = ($1081|0)<($$len171$0$i|0);
           if ($1082) {
            $dst175$0199$i = $1080;$i173$0197$i = $1081;$src174$0198$i = $1079;
           } else {
            break;
           }
          }
         }
        }
       } while(0);
       $1083 = $988 << 2;
       $1084 = (($989) + ($1083)|0);
       $1085 = (($985) + ($988))|0;
       $$37$i = $1084;$ofs166$1$i = $1085;
      }
      $1087 = ($ofs166$1$i|0)==($1|0);
      if ($1087) {
       $$37$i$lcssa = $$37$i;
       break;
      } else {
       $$36$i = $$37$i;$ofs166$0$i = $ofs166$1$i;
      }
     }
     $1088 = HEAP32[$12>>2]|0;
     $1089 = (($$3649$ph$i) + ($1088)|0);
     $1090 = (($linecount165$0$ph$i) + -1)|0;
     $1091 = ($1090|0)==(0);
     if ($1091) {
      break;
     } else {
      $$36$ph$i = $$37$i$lcssa;$$3649$ph$i = $1089;$linecount165$0$ph$i = $1090;
     }
    }
    break;
   }
   case 2:  {
    $250 = ((($99)) + 12|0);
    $251 = HEAP32[$250>>2]|0;
    $252 = ((($99)) + 16|0);
    $253 = HEAP32[$252>>2]|0;
    $254 = $253 | $251;
    $255 = ((($99)) + 20|0);
    $256 = HEAP32[$255>>2]|0;
    $257 = $254 | $256;
    $switch$split2D = ($257|0)<(65535);
    L377: do {
     if ($switch$split2D) {
      switch ($257|0) {
      case 32767:  {
       break;
      }
      default: {
       break L377;
      }
      }
      $424 = ($253|0)==(992);
      $425 = ($251|0)==(992);
      $or$cond79$i = $425 | $424;
      $426 = ($256|0)==(992);
      $or$cond80$i = $or$cond79$i | $426;
      if ($or$cond80$i) {
       $427 = ($90<<24>>24)==(-128);
       $428 = ((($srcrect)) + 12|0);
       $429 = HEAP32[$428>>2]|0;
       $430 = ((($srcrect)) + 8|0);
       $431 = HEAP32[$430>>2]|0;
       $432 = (($431) + ($92))|0;
       $433 = $92 << 1;
       $434 = (0 - ($433))|0;
       $435 = (($22) + ($434)|0);
       if ($427) {
        $$18$ph$i = $srcbuf$13$ph;$$1831$ph$i = $435;$linecount62$0$ph$i = $429;
       } else {
        $436 = $91 >>> 3;
        $$21$ph$i = $srcbuf$13$ph;$$2134$ph$i = $435;$linecount83$0$ph$i = $429;
        while(1) {
         $$21$i = $$21$ph$i;$ofs84$0$i = 0;
         while(1) {
          $539 = HEAP8[$$21$i>>0]|0;
          $540 = $539&255;
          $541 = (($540) + ($ofs84$0$i))|0;
          $542 = ((($$21$i)) + 1|0);
          $543 = HEAP8[$542>>0]|0;
          $544 = $543&255;
          $545 = ((($$21$i)) + 2|0);
          $546 = ($543<<24>>24)==(0);
          if ($546) {
           $585 = ($541|0)==(0);
           if ($585) {
            break L44;
           } else {
            $$22$i = $545;$ofs84$1$i = $541;
           }
          } else {
           $547 = ($541|0)<($432|0);
           do {
            if ($547) {
             $548 = (($92) - ($541))|0;
             $549 = ($548|0)>(0);
             if ($549) {
              $550 = (($544) - ($548))|0;
              $551 = ($550|0)<(1);
              if ($551) {
               break;
              } else {
               $len89$0$i = $550;$start88$0$i = $548;
              }
             } else {
              $len89$0$i = $544;$start88$0$i = 0;
             }
             $552 = (($start88$0$i) + ($541))|0;
             $553 = (($432) - ($552))|0;
             $554 = ($len89$0$i|0)>($553|0);
             $$len89$0$i = $554 ? $553 : $len89$0$i;
             $555 = ($$len89$0$i|0)>(0);
             if ($555) {
              $556 = $552 << 1;
              $557 = (($$2134$ph$i) + ($556)|0);
              $558 = $start88$0$i << 1;
              $559 = (($545) + ($558)|0);
              $dst93$0232$i = $557;$i91$0234$i = 0;$src92$0233$i = $559;
              while(1) {
               $560 = ((($src92$0233$i)) + 2|0);
               $561 = HEAP16[$src92$0233$i>>1]|0;
               $562 = $561&65535;
               $563 = HEAP16[$dst93$0232$i>>1]|0;
               $564 = $563&65535;
               $565 = $562 << 16;
               $566 = $565 | $562;
               $567 = $566 & 65043487;
               $568 = $564 << 16;
               $569 = $568 | $564;
               $570 = $569 & 65043487;
               $571 = (($567) - ($570))|0;
               $572 = Math_imul($571, $436)|0;
               $573 = $572 >>> 5;
               $574 = (($573) + ($570))|0;
               $575 = $574 & 65043487;
               $576 = $575 >>> 16;
               $577 = $576 | $575;
               $578 = $577&65535;
               $579 = ((($dst93$0232$i)) + 2|0);
               HEAP16[$dst93$0232$i>>1] = $578;
               $580 = (($i91$0234$i) + 1)|0;
               $581 = ($580|0)<($$len89$0$i|0);
               if ($581) {
                $dst93$0232$i = $579;$i91$0234$i = $580;$src92$0233$i = $560;
               } else {
                break;
               }
              }
             }
            }
           } while(0);
           $582 = $544 << 1;
           $583 = (($545) + ($582)|0);
           $584 = (($541) + ($544))|0;
           $$22$i = $583;$ofs84$1$i = $584;
          }
          $586 = ($ofs84$1$i|0)==($1|0);
          if ($586) {
           $$22$i$lcssa = $$22$i;
           break;
          } else {
           $$21$i = $$22$i;$ofs84$0$i = $ofs84$1$i;
          }
         }
         $587 = (($$2134$ph$i) + ($13)|0);
         $588 = (($linecount83$0$ph$i) + -1)|0;
         $589 = ($588|0)==(0);
         if ($589) {
          break L44;
         } else {
          $$21$ph$i = $$22$i$lcssa;$$2134$ph$i = $587;$linecount83$0$ph$i = $588;
         }
        }
       }
       while(1) {
        $$18$i = $$18$ph$i;$ofs63$0$i = 0;
        while(1) {
         $437 = HEAP8[$$18$i>>0]|0;
         $438 = $437&255;
         $439 = (($438) + ($ofs63$0$i))|0;
         $440 = ((($$18$i)) + 1|0);
         $441 = HEAP8[$440>>0]|0;
         $442 = $441&255;
         $443 = ((($$18$i)) + 2|0);
         $444 = ($441<<24>>24)==(0);
         if ($444) {
          $533 = ($439|0)==(0);
          if ($533) {
           break L44;
          } else {
           $$19$i = $443;$ofs63$1$i = $439;
          }
         } else {
          $445 = ($439|0)<($432|0);
          L408: do {
           if ($445) {
            $446 = (($92) - ($439))|0;
            $447 = ($446|0)>(0);
            if ($447) {
             $448 = (($442) - ($446))|0;
             $449 = ($448|0)<(1);
             if ($449) {
              break;
             } else {
              $len68$0$i = $448;$start67$0$i = $446;
             }
            } else {
             $len68$0$i = $442;$start67$0$i = 0;
            }
            $450 = (($start67$0$i) + ($439))|0;
            $451 = (($432) - ($450))|0;
            $452 = ($len68$0$i|0)>($451|0);
            $$len68$0$i = $452 ? $451 : $len68$0$i;
            $453 = $start67$0$i << 1;
            $454 = (($443) + ($453)|0);
            $455 = $450 << 1;
            $456 = (($$1831$ph$i) + ($455)|0);
            $457 = $454;
            $458 = $456;
            $459 = $458 ^ $457;
            $460 = $459 & 3;
            $461 = ($460|0)==(0);
            if (!($461)) {
             $462 = ($$len68$0$i|0)==(0);
             if ($462) {
              break;
             } else {
              $dst72$0220$i = $456;$n70$0222$i = $$len68$0$i;$src71$0221$i = $454;
             }
             while(1) {
              $463 = (($n70$0222$i) + -1)|0;
              $464 = ((($src71$0221$i)) + 2|0);
              $465 = HEAP16[$src71$0221$i>>1]|0;
              $466 = $465&65535;
              $467 = HEAP16[$dst72$0220$i>>1]|0;
              $468 = $467&65535;
              $469 = $466 & 64478;
              $470 = $468 & 64478;
              $471 = (($470) + ($469))|0;
              $472 = $471 >>> 1;
              $473 = $467 & $465;
              $474 = $473&65535;
              $475 = $474 & 1057;
              $476 = (($472) + ($475))|0;
              $477 = $476&65535;
              $478 = ((($dst72$0220$i)) + 2|0);
              HEAP16[$dst72$0220$i>>1] = $477;
              $479 = ($463|0)==(0);
              if ($479) {
               break L408;
              } else {
               $dst72$0220$i = $478;$n70$0222$i = $463;$src71$0221$i = $464;
              }
             }
            }
            $480 = $457 & 3;
            $481 = ($480|0)==(0);
            if ($481) {
             $dst72$1$ph$i = $456;$n70$1$ph$i = $$len68$0$i;$src71$1$ph$i = $454;
            } else {
             $482 = ((($454)) + 2|0);
             $483 = HEAP16[$454>>1]|0;
             $484 = $483&65535;
             $485 = HEAP16[$456>>1]|0;
             $486 = $485&65535;
             $487 = $484 & 64478;
             $488 = $486 & 64478;
             $489 = (($488) + ($487))|0;
             $490 = $489 >>> 1;
             $491 = $485 & $483;
             $492 = $491&65535;
             $493 = $492 & 1057;
             $494 = (($490) + ($493))|0;
             $495 = $494&65535;
             $496 = ((($456)) + 2|0);
             HEAP16[$456>>1] = $495;
             $497 = (($$len68$0$i) + -1)|0;
             $dst72$1$ph$i = $496;$n70$1$ph$i = $497;$src71$1$ph$i = $482;
            }
            $498 = ($n70$1$ph$i>>>0)>(1);
            if ($498) {
             $499 = (($n70$1$ph$i) + -2)|0;
             $scevgep275$i = ((($dst72$1$ph$i)) + 4|0);
             $dst72$1224$i = $dst72$1$ph$i;$n70$1226$i = $n70$1$ph$i;$src71$1225$i = $src71$1$ph$i;
             while(1) {
              $500 = HEAP32[$src71$1225$i>>2]|0;
              $501 = HEAP32[$dst72$1224$i>>2]|0;
              $502 = $500 >>> 1;
              $503 = $502 & 2112847343;
              $504 = $501 >>> 1;
              $505 = $504 & 2112847343;
              $506 = $500 & 69272609;
              $507 = $506 & $501;
              $508 = (($507) + ($503))|0;
              $509 = (($508) + ($505))|0;
              HEAP32[$dst72$1224$i>>2] = $509;
              $510 = ((($src71$1225$i)) + 4|0);
              $511 = ((($dst72$1224$i)) + 4|0);
              $512 = (($n70$1226$i) + -2)|0;
              $513 = ($512>>>0)>(1);
              if ($513) {
               $dst72$1224$i = $511;$n70$1226$i = $512;$src71$1225$i = $510;
              } else {
               break;
              }
             }
             $514 = $499 & -2;
             $scevgep277$i = ((($src71$1$ph$i)) + 4|0);
             $515 = (($499) - ($514))|0;
             $scevgep276$i = (($scevgep275$i) + ($514<<1)|0);
             $scevgep278$i = (($scevgep277$i) + ($514<<1)|0);
             $dst72$1$lcssa$i = $scevgep276$i;$n70$1$lcssa$i = $515;$src71$1$lcssa$i = $scevgep278$i;
            } else {
             $dst72$1$lcssa$i = $dst72$1$ph$i;$n70$1$lcssa$i = $n70$1$ph$i;$src71$1$lcssa$i = $src71$1$ph$i;
            }
            $516 = ($n70$1$lcssa$i|0)==(0);
            if ($516) {
             break;
            }
            $517 = HEAP16[$src71$1$lcssa$i>>1]|0;
            $518 = $517&65535;
            $519 = HEAP16[$dst72$1$lcssa$i>>1]|0;
            $520 = $519&65535;
            $521 = $518 & 64478;
            $522 = $520 & 64478;
            $523 = (($522) + ($521))|0;
            $524 = $523 >>> 1;
            $525 = $519 & $517;
            $526 = $525&65535;
            $527 = $526 & 1057;
            $528 = (($524) + ($527))|0;
            $529 = $528&65535;
            HEAP16[$dst72$1$lcssa$i>>1] = $529;
           }
          } while(0);
          $530 = $442 << 1;
          $531 = (($443) + ($530)|0);
          $532 = (($439) + ($442))|0;
          $$19$i = $531;$ofs63$1$i = $532;
         }
         $534 = ($ofs63$1$i|0)==($1|0);
         if ($534) {
          $$19$i$lcssa = $$19$i;
          break;
         } else {
          $$18$i = $$19$i;$ofs63$0$i = $ofs63$1$i;
         }
        }
        $535 = HEAP32[$12>>2]|0;
        $536 = (($$1831$ph$i) + ($535)|0);
        $537 = (($linecount62$0$ph$i) + -1)|0;
        $538 = ($537|0)==(0);
        if ($538) {
         break L44;
        } else {
         $$18$ph$i = $$19$i$lcssa;$$1831$ph$i = $536;$linecount62$0$ph$i = $537;
        }
       }
      }
     } else {
      switch ($257|0) {
      case 65535:  {
       break;
      }
      default: {
       break L377;
      }
      }
      $258 = ($253|0)==(2016);
      $259 = ($251|0)==(2016);
      $or$cond$i = $259 | $258;
      $260 = ($256|0)==(2016);
      $or$cond78$i = $or$cond$i | $260;
      if ($or$cond78$i) {
       $261 = ($90<<24>>24)==(-128);
       $262 = ((($srcrect)) + 12|0);
       $263 = HEAP32[$262>>2]|0;
       $264 = ((($srcrect)) + 8|0);
       $265 = HEAP32[$264>>2]|0;
       $266 = (($265) + ($92))|0;
       $267 = $92 << 1;
       $268 = (0 - ($267))|0;
       $269 = (($22) + ($268)|0);
       if ($261) {
        $$12$ph$i = $srcbuf$13$ph;$$1225$ph$i = $269;$linecount32$0$ph$i = $263;
       } else {
        $270 = $91 >>> 3;
        $$15$ph$i = $srcbuf$13$ph;$$1528$ph$i = $269;$linecount48$0$ph$i = $263;
        while(1) {
         $$15$i = $$15$ph$i;$ofs49$0$i = 0;
         while(1) {
          $373 = HEAP8[$$15$i>>0]|0;
          $374 = $373&255;
          $375 = (($374) + ($ofs49$0$i))|0;
          $376 = ((($$15$i)) + 1|0);
          $377 = HEAP8[$376>>0]|0;
          $378 = $377&255;
          $379 = ((($$15$i)) + 2|0);
          $380 = ($377<<24>>24)==(0);
          if ($380) {
           $419 = ($375|0)==(0);
           if ($419) {
            break L44;
           } else {
            $$16$i = $379;$ofs49$1$i = $375;
           }
          } else {
           $381 = ($375|0)<($266|0);
           do {
            if ($381) {
             $382 = (($92) - ($375))|0;
             $383 = ($382|0)>(0);
             if ($383) {
              $384 = (($378) - ($382))|0;
              $385 = ($384|0)<(1);
              if ($385) {
               break;
              } else {
               $len54$0$i = $384;$start53$0$i = $382;
              }
             } else {
              $len54$0$i = $378;$start53$0$i = 0;
             }
             $386 = (($start53$0$i) + ($375))|0;
             $387 = (($266) - ($386))|0;
             $388 = ($len54$0$i|0)>($387|0);
             $$len54$0$i = $388 ? $387 : $len54$0$i;
             $389 = ($$len54$0$i|0)>(0);
             if ($389) {
              $390 = $386 << 1;
              $391 = (($$1528$ph$i) + ($390)|0);
              $392 = $start53$0$i << 1;
              $393 = (($379) + ($392)|0);
              $dst57$0215$i = $391;$i$0217$i = 0;$src56$0216$i = $393;
              while(1) {
               $394 = ((($src56$0216$i)) + 2|0);
               $395 = HEAP16[$src56$0216$i>>1]|0;
               $396 = $395&65535;
               $397 = HEAP16[$dst57$0215$i>>1]|0;
               $398 = $397&65535;
               $399 = $396 << 16;
               $400 = $399 | $396;
               $401 = $400 & 132184095;
               $402 = $398 << 16;
               $403 = $402 | $398;
               $404 = $403 & 132184095;
               $405 = (($401) - ($404))|0;
               $406 = Math_imul($405, $270)|0;
               $407 = $406 >>> 5;
               $408 = (($407) + ($404))|0;
               $409 = $408 & 132184095;
               $410 = $409 >>> 16;
               $411 = $410 | $409;
               $412 = $411&65535;
               $413 = ((($dst57$0215$i)) + 2|0);
               HEAP16[$dst57$0215$i>>1] = $412;
               $414 = (($i$0217$i) + 1)|0;
               $415 = ($414|0)<($$len54$0$i|0);
               if ($415) {
                $dst57$0215$i = $413;$i$0217$i = $414;$src56$0216$i = $394;
               } else {
                break;
               }
              }
             }
            }
           } while(0);
           $416 = $378 << 1;
           $417 = (($379) + ($416)|0);
           $418 = (($375) + ($378))|0;
           $$16$i = $417;$ofs49$1$i = $418;
          }
          $420 = ($ofs49$1$i|0)==($1|0);
          if ($420) {
           $$16$i$lcssa = $$16$i;
           break;
          } else {
           $$15$i = $$16$i;$ofs49$0$i = $ofs49$1$i;
          }
         }
         $421 = (($$1528$ph$i) + ($13)|0);
         $422 = (($linecount48$0$ph$i) + -1)|0;
         $423 = ($422|0)==(0);
         if ($423) {
          break L44;
         } else {
          $$15$ph$i = $$16$i$lcssa;$$1528$ph$i = $421;$linecount48$0$ph$i = $422;
         }
        }
       }
       while(1) {
        $$12$i = $$12$ph$i;$ofs33$0$i = 0;
        while(1) {
         $271 = HEAP8[$$12$i>>0]|0;
         $272 = $271&255;
         $273 = (($272) + ($ofs33$0$i))|0;
         $274 = ((($$12$i)) + 1|0);
         $275 = HEAP8[$274>>0]|0;
         $276 = $275&255;
         $277 = ((($$12$i)) + 2|0);
         $278 = ($275<<24>>24)==(0);
         if ($278) {
          $367 = ($273|0)==(0);
          if ($367) {
           break L44;
          } else {
           $$13$i = $277;$ofs33$1$i = $273;
          }
         } else {
          $279 = ($273|0)<($266|0);
          L461: do {
           if ($279) {
            $280 = (($92) - ($273))|0;
            $281 = ($280|0)>(0);
            if ($281) {
             $282 = (($276) - ($280))|0;
             $283 = ($282|0)<(1);
             if ($283) {
              break;
             } else {
              $len38$0$i = $282;$start37$0$i = $280;
             }
            } else {
             $len38$0$i = $276;$start37$0$i = 0;
            }
            $284 = (($start37$0$i) + ($273))|0;
            $285 = (($266) - ($284))|0;
            $286 = ($len38$0$i|0)>($285|0);
            $$len38$0$i = $286 ? $285 : $len38$0$i;
            $287 = $start37$0$i << 1;
            $288 = (($277) + ($287)|0);
            $289 = $284 << 1;
            $290 = (($$1225$ph$i) + ($289)|0);
            $291 = $288;
            $292 = $290;
            $293 = $292 ^ $291;
            $294 = $293 & 3;
            $295 = ($294|0)==(0);
            if (!($295)) {
             $296 = ($$len38$0$i|0)==(0);
             if ($296) {
              break;
             } else {
              $dst$0205$i = $290;$n$0207$i = $$len38$0$i;$src$0206$i = $288;
             }
             while(1) {
              $297 = (($n$0207$i) + -1)|0;
              $298 = ((($src$0206$i)) + 2|0);
              $299 = HEAP16[$src$0206$i>>1]|0;
              $300 = $299&65535;
              $301 = HEAP16[$dst$0205$i>>1]|0;
              $302 = $301&65535;
              $303 = $300 & 63454;
              $304 = $302 & 63454;
              $305 = (($304) + ($303))|0;
              $306 = $305 >>> 1;
              $307 = $301 & $299;
              $308 = $307&65535;
              $309 = $308 & 2081;
              $310 = (($306) + ($309))|0;
              $311 = $310&65535;
              $312 = ((($dst$0205$i)) + 2|0);
              HEAP16[$dst$0205$i>>1] = $311;
              $313 = ($297|0)==(0);
              if ($313) {
               break L461;
              } else {
               $dst$0205$i = $312;$n$0207$i = $297;$src$0206$i = $298;
              }
             }
            }
            $314 = $291 & 3;
            $315 = ($314|0)==(0);
            if ($315) {
             $dst$1$ph$i = $290;$n$1$ph$i = $$len38$0$i;$src$1$ph$i = $288;
            } else {
             $316 = ((($288)) + 2|0);
             $317 = HEAP16[$288>>1]|0;
             $318 = $317&65535;
             $319 = HEAP16[$290>>1]|0;
             $320 = $319&65535;
             $321 = $318 & 63454;
             $322 = $320 & 63454;
             $323 = (($322) + ($321))|0;
             $324 = $323 >>> 1;
             $325 = $319 & $317;
             $326 = $325&65535;
             $327 = $326 & 2081;
             $328 = (($324) + ($327))|0;
             $329 = $328&65535;
             $330 = ((($290)) + 2|0);
             HEAP16[$290>>1] = $329;
             $331 = (($$len38$0$i) + -1)|0;
             $dst$1$ph$i = $330;$n$1$ph$i = $331;$src$1$ph$i = $316;
            }
            $332 = ($n$1$ph$i>>>0)>(1);
            if ($332) {
             $333 = (($n$1$ph$i) + -2)|0;
             $scevgep$i = ((($dst$1$ph$i)) + 4|0);
             $dst$1209$i = $dst$1$ph$i;$n$1211$i = $n$1$ph$i;$src$1210$i = $src$1$ph$i;
             while(1) {
              $334 = HEAP32[$src$1210$i>>2]|0;
              $335 = HEAP32[$dst$1209$i>>2]|0;
              $336 = $334 >>> 1;
              $337 = $336 & 2079292399;
              $338 = $335 >>> 1;
              $339 = $338 & 2079292399;
              $340 = $334 & 136382497;
              $341 = $340 & $335;
              $342 = (($341) + ($337))|0;
              $343 = (($342) + ($339))|0;
              HEAP32[$dst$1209$i>>2] = $343;
              $344 = ((($src$1210$i)) + 4|0);
              $345 = ((($dst$1209$i)) + 4|0);
              $346 = (($n$1211$i) + -2)|0;
              $347 = ($346>>>0)>(1);
              if ($347) {
               $dst$1209$i = $345;$n$1211$i = $346;$src$1210$i = $344;
              } else {
               break;
              }
             }
             $348 = $333 & -2;
             $scevgep273$i = ((($src$1$ph$i)) + 4|0);
             $349 = (($333) - ($348))|0;
             $scevgep272$i = (($scevgep$i) + ($348<<1)|0);
             $scevgep274$i = (($scevgep273$i) + ($348<<1)|0);
             $dst$1$lcssa$i = $scevgep272$i;$n$1$lcssa$i = $349;$src$1$lcssa$i = $scevgep274$i;
            } else {
             $dst$1$lcssa$i = $dst$1$ph$i;$n$1$lcssa$i = $n$1$ph$i;$src$1$lcssa$i = $src$1$ph$i;
            }
            $350 = ($n$1$lcssa$i|0)==(0);
            if ($350) {
             break;
            }
            $351 = HEAP16[$src$1$lcssa$i>>1]|0;
            $352 = $351&65535;
            $353 = HEAP16[$dst$1$lcssa$i>>1]|0;
            $354 = $353&65535;
            $355 = $352 & 63454;
            $356 = $354 & 63454;
            $357 = (($356) + ($355))|0;
            $358 = $357 >>> 1;
            $359 = $353 & $351;
            $360 = $359&65535;
            $361 = $360 & 2081;
            $362 = (($358) + ($361))|0;
            $363 = $362&65535;
            HEAP16[$dst$1$lcssa$i>>1] = $363;
           }
          } while(0);
          $364 = $276 << 1;
          $365 = (($277) + ($364)|0);
          $366 = (($273) + ($276))|0;
          $$13$i = $365;$ofs33$1$i = $366;
         }
         $368 = ($ofs33$1$i|0)==($1|0);
         if ($368) {
          $$13$i$lcssa = $$13$i;
          break;
         } else {
          $$12$i = $$13$i;$ofs33$0$i = $ofs33$1$i;
         }
        }
        $369 = HEAP32[$12>>2]|0;
        $370 = (($$1225$ph$i) + ($369)|0);
        $371 = (($linecount32$0$ph$i) + -1)|0;
        $372 = ($371|0)==(0);
        if ($372) {
         break L44;
        } else {
         $$12$ph$i = $$13$i$lcssa;$$1225$ph$i = $370;$linecount32$0$ph$i = $371;
        }
       }
      }
     }
    } while(0);
    $590 = ((($srcrect)) + 12|0);
    $591 = HEAP32[$590>>2]|0;
    $592 = ((($srcrect)) + 8|0);
    $593 = HEAP32[$592>>2]|0;
    $594 = (($593) + ($92))|0;
    $595 = $92 << 1;
    $596 = (0 - ($595))|0;
    $597 = (($22) + ($596)|0);
    $598 = ((($99)) + 32|0);
    $599 = ((($99)) + 28|0);
    $600 = ((($99)) + 24|0);
    $$24$ph$i = $srcbuf$13$ph;$$2437$ph$i = $597;$linecount99$0$ph$i = $591;
    while(1) {
     $$24$i = $$24$ph$i;$ofs100$0$i = 0;
     while(1) {
      $601 = HEAP8[$$24$i>>0]|0;
      $602 = $601&255;
      $603 = (($602) + ($ofs100$0$i))|0;
      $604 = ((($$24$i)) + 1|0);
      $605 = HEAP8[$604>>0]|0;
      $606 = $605&255;
      $607 = ((($$24$i)) + 2|0);
      $608 = ($605<<24>>24)==(0);
      if ($608) {
       $703 = ($603|0)==(0);
       if ($703) {
        break L44;
       } else {
        $$25$i = $607;$ofs100$1$i = $603;
       }
      } else {
       $609 = ($603|0)<($594|0);
       do {
        if ($609) {
         $610 = (($92) - ($603))|0;
         $611 = ($610|0)>(0);
         if ($611) {
          $612 = (($606) - ($610))|0;
          $613 = ($612|0)<(1);
          if ($613) {
           break;
          } else {
           $len105$0$i = $612;$start104$0$i = $610;
          }
         } else {
          $len105$0$i = $606;$start104$0$i = 0;
         }
         $614 = (($start104$0$i) + ($603))|0;
         $615 = (($594) - ($614))|0;
         $616 = ($len105$0$i|0)>($615|0);
         $$len105$0$i = $616 ? $615 : $len105$0$i;
         $617 = ($$len105$0$i|0)>(0);
         if ($617) {
          $618 = $614 << 1;
          $619 = (($$2437$ph$i) + ($618)|0);
          $620 = $start104$0$i << 1;
          $621 = (($607) + ($620)|0);
          $$pre = HEAP32[$600>>2]|0;
          $dst109$0203$i = $619;$i107$0201$i = 0;$src108$0202$i = $621;
          while(1) {
           $622 = HEAP16[$src108$0202$i>>1]|0;
           $623 = $622&65535;
           $624 = HEAP16[$dst109$0203$i>>1]|0;
           $625 = $624&65535;
           $626 = $623 & $251;
           $627 = HEAP32[$598>>2]|0;
           $628 = $627 & 255;
           $629 = $626 >>> $628;
           $630 = HEAP32[$599>>2]|0;
           $631 = $630 & 255;
           $632 = (125924 + ($631<<2)|0);
           $633 = HEAP32[$632>>2]|0;
           $634 = (($633) + ($629)|0);
           $635 = HEAP8[$634>>0]|0;
           $636 = $635&255;
           $637 = $623 & $253;
           $638 = $627 >>> 8;
           $639 = $638 & 255;
           $640 = $637 >>> $639;
           $641 = $630 >>> 8;
           $642 = $641 & 255;
           $643 = (125924 + ($642<<2)|0);
           $644 = HEAP32[$643>>2]|0;
           $645 = (($644) + ($640)|0);
           $646 = HEAP8[$645>>0]|0;
           $647 = $646&255;
           $648 = $623 & $256;
           $649 = $627 >>> 16;
           $650 = $649 & 255;
           $651 = $648 >>> $650;
           $652 = $630 >>> 16;
           $653 = $652 & 255;
           $654 = (125924 + ($653<<2)|0);
           $655 = HEAP32[$654>>2]|0;
           $656 = (($655) + ($651)|0);
           $657 = HEAP8[$656>>0]|0;
           $658 = $657&255;
           $659 = $625 & $251;
           $660 = $659 >>> $628;
           $661 = (($633) + ($660)|0);
           $662 = HEAP8[$661>>0]|0;
           $663 = $662&255;
           $664 = $625 & $253;
           $665 = $664 >>> $639;
           $666 = (($644) + ($665)|0);
           $667 = HEAP8[$666>>0]|0;
           $668 = $667&255;
           $669 = $625 & $256;
           $670 = $669 >>> $650;
           $671 = (($655) + ($670)|0);
           $672 = HEAP8[$671>>0]|0;
           $673 = $672&255;
           $674 = (($636) - ($663))|0;
           $675 = Math_imul($674, $91)|0;
           $676 = $675 >>> 8;
           $677 = (($676) + ($663))|0;
           $678 = (($647) - ($668))|0;
           $679 = Math_imul($678, $91)|0;
           $680 = $679 >>> 8;
           $681 = (($680) + ($668))|0;
           $682 = (($658) - ($673))|0;
           $683 = Math_imul($682, $91)|0;
           $684 = $683 >>> 8;
           $685 = (($684) + ($673))|0;
           $686 = $677 >>> $631;
           $687 = $686 << $628;
           $688 = $681 >>> $642;
           $689 = $688 << $639;
           $690 = $685 >>> $653;
           $691 = $690 << $650;
           $692 = $687 | $$pre;
           $693 = $692 | $689;
           $694 = $693 | $691;
           $695 = $694&65535;
           HEAP16[$dst109$0203$i>>1] = $695;
           $696 = ((($src108$0202$i)) + 2|0);
           $697 = ((($dst109$0203$i)) + 2|0);
           $698 = (($i107$0201$i) + 1)|0;
           $699 = ($698|0)<($$len105$0$i|0);
           if ($699) {
            $dst109$0203$i = $697;$i107$0201$i = $698;$src108$0202$i = $696;
           } else {
            break;
           }
          }
         }
        }
       } while(0);
       $700 = $606 << 1;
       $701 = (($607) + ($700)|0);
       $702 = (($603) + ($606))|0;
       $$25$i = $701;$ofs100$1$i = $702;
      }
      $704 = ($ofs100$1$i|0)==($1|0);
      if ($704) {
       $$25$i$lcssa = $$25$i;
       break;
      } else {
       $$24$i = $$25$i;$ofs100$0$i = $ofs100$1$i;
      }
     }
     $705 = (($$2437$ph$i) + ($13)|0);
     $706 = (($linecount99$0$ph$i) + -1)|0;
     $707 = ($706|0)==(0);
     if ($707) {
      break;
     } else {
      $$24$ph$i = $$25$i$lcssa;$$2437$ph$i = $705;$linecount99$0$ph$i = $706;
     }
    }
    break;
   }
   case 3:  {
    $708 = ((($srcrect)) + 12|0);
    $709 = HEAP32[$708>>2]|0;
    $710 = ((($srcrect)) + 8|0);
    $711 = HEAP32[$710>>2]|0;
    $712 = (($711) + ($92))|0;
    $713 = Math_imul($92, -3)|0;
    $714 = (($22) + ($713)|0);
    $715 = ((($99)) + 12|0);
    $716 = ((($99)) + 32|0);
    $717 = ((($99)) + 28|0);
    $718 = ((($99)) + 16|0);
    $719 = ((($99)) + 20|0);
    $720 = ((($99)) + 24|0);
    $$27$ph$i = $srcbuf$13$ph;$$2740$ph$i = $714;$linecount114$0$ph$i = $709;
    while(1) {
     $$27$i = $$27$ph$i;$ofs115$0$i = 0;
     while(1) {
      $721 = HEAP8[$$27$i>>0]|0;
      $722 = $721&255;
      $723 = (($722) + ($ofs115$0$i))|0;
      $724 = ((($$27$i)) + 1|0);
      $725 = HEAP8[$724>>0]|0;
      $726 = $725&255;
      $727 = ((($$27$i)) + 2|0);
      $728 = ($725<<24>>24)==(0);
      if ($728) {
       $851 = ($723|0)==(0);
       if ($851) {
        break L44;
       } else {
        $$28$i = $727;$ofs115$1$i = $723;
       }
      } else {
       $729 = ($723|0)<($712|0);
       do {
        if ($729) {
         $730 = (($92) - ($723))|0;
         $731 = ($730|0)>(0);
         if ($731) {
          $732 = (($726) - ($730))|0;
          $733 = ($732|0)<(1);
          if ($733) {
           break;
          } else {
           $len120$0$i = $732;$start119$0$i = $730;
          }
         } else {
          $len120$0$i = $726;$start119$0$i = 0;
         }
         $734 = (($start119$0$i) + ($723))|0;
         $735 = (($712) - ($734))|0;
         $736 = ($len120$0$i|0)>($735|0);
         $$len120$0$i = $736 ? $735 : $len120$0$i;
         $737 = ($$len120$0$i|0)>(0);
         if ($737) {
          $738 = ($734*3)|0;
          $739 = (($$2740$ph$i) + ($738)|0);
          $740 = ($start119$0$i*3)|0;
          $741 = (($727) + ($740)|0);
          $dst124$0239$i = $739;$i122$0237$i = 0;$src123$0238$i = $741;
          while(1) {
           $742 = ((($src123$0238$i)) + 2|0);
           $743 = HEAP8[$742>>0]|0;
           $744 = $743&255;
           $745 = $744 << 16;
           $746 = ((($src123$0238$i)) + 1|0);
           $747 = HEAP8[$746>>0]|0;
           $748 = $747&255;
           $749 = $748 << 8;
           $750 = $749 | $745;
           $751 = HEAP8[$src123$0238$i>>0]|0;
           $752 = $751&255;
           $753 = $750 | $752;
           $754 = ((($dst124$0239$i)) + 2|0);
           $755 = HEAP8[$754>>0]|0;
           $756 = $755&255;
           $757 = $756 << 16;
           $758 = ((($dst124$0239$i)) + 1|0);
           $759 = HEAP8[$758>>0]|0;
           $760 = $759&255;
           $761 = $760 << 8;
           $762 = $761 | $757;
           $763 = HEAP8[$dst124$0239$i>>0]|0;
           $764 = $763&255;
           $765 = $762 | $764;
           $766 = HEAP32[$715>>2]|0;
           $767 = $766 & $753;
           $768 = HEAP32[$716>>2]|0;
           $769 = $768 & 255;
           $770 = $767 >>> $769;
           $771 = HEAP32[$717>>2]|0;
           $772 = $771 & 255;
           $773 = (125924 + ($772<<2)|0);
           $774 = HEAP32[$773>>2]|0;
           $775 = (($774) + ($770)|0);
           $776 = HEAP8[$775>>0]|0;
           $777 = $776&255;
           $778 = HEAP32[$718>>2]|0;
           $779 = $778 & $753;
           $780 = $768 >>> 8;
           $781 = $780 & 255;
           $782 = $779 >>> $781;
           $783 = $771 >>> 8;
           $784 = $783 & 255;
           $785 = (125924 + ($784<<2)|0);
           $786 = HEAP32[$785>>2]|0;
           $787 = (($786) + ($782)|0);
           $788 = HEAP8[$787>>0]|0;
           $789 = $788&255;
           $790 = HEAP32[$719>>2]|0;
           $791 = $790 & $753;
           $792 = $768 >>> 16;
           $793 = $792 & 255;
           $794 = $791 >>> $793;
           $795 = $771 >>> 16;
           $796 = $795 & 255;
           $797 = (125924 + ($796<<2)|0);
           $798 = HEAP32[$797>>2]|0;
           $799 = (($798) + ($794)|0);
           $800 = HEAP8[$799>>0]|0;
           $801 = $800&255;
           $802 = $765 & $766;
           $803 = $802 >>> $769;
           $804 = (($774) + ($803)|0);
           $805 = HEAP8[$804>>0]|0;
           $806 = $805&255;
           $807 = $778 & $765;
           $808 = $807 >>> $781;
           $809 = (($786) + ($808)|0);
           $810 = HEAP8[$809>>0]|0;
           $811 = $810&255;
           $812 = $790 & $765;
           $813 = $812 >>> $793;
           $814 = (($798) + ($813)|0);
           $815 = HEAP8[$814>>0]|0;
           $816 = $815&255;
           $817 = (($777) - ($806))|0;
           $818 = Math_imul($817, $91)|0;
           $819 = $818 >>> 8;
           $820 = (($819) + ($806))|0;
           $821 = (($789) - ($811))|0;
           $822 = Math_imul($821, $91)|0;
           $823 = $822 >>> 8;
           $824 = (($823) + ($811))|0;
           $825 = (($801) - ($816))|0;
           $826 = Math_imul($825, $91)|0;
           $827 = $826 >>> 8;
           $828 = (($827) + ($816))|0;
           $829 = $820 >>> $772;
           $830 = $829 << $769;
           $831 = $824 >>> $784;
           $832 = $831 << $781;
           $833 = $828 >>> $796;
           $834 = $833 << $793;
           $835 = HEAP32[$720>>2]|0;
           $836 = $830 | $835;
           $837 = $836 | $832;
           $838 = $837 | $834;
           $839 = $838&255;
           HEAP8[$dst124$0239$i>>0] = $839;
           $840 = $838 >>> 8;
           $841 = $840&255;
           HEAP8[$758>>0] = $841;
           $842 = $838 >>> 16;
           $843 = $842&255;
           HEAP8[$754>>0] = $843;
           $844 = ((($src123$0238$i)) + 3|0);
           $845 = ((($dst124$0239$i)) + 3|0);
           $846 = (($i122$0237$i) + 1)|0;
           $847 = ($846|0)<($$len120$0$i|0);
           if ($847) {
            $dst124$0239$i = $845;$i122$0237$i = $846;$src123$0238$i = $844;
           } else {
            break;
           }
          }
         }
        }
       } while(0);
       $848 = ($726*3)|0;
       $849 = (($727) + ($848)|0);
       $850 = (($723) + ($726))|0;
       $$28$i = $849;$ofs115$1$i = $850;
      }
      $852 = ($ofs115$1$i|0)==($1|0);
      if ($852) {
       $$28$i$lcssa = $$28$i;
       break;
      } else {
       $$27$i = $$28$i;$ofs115$0$i = $ofs115$1$i;
      }
     }
     $853 = HEAP32[$12>>2]|0;
     $854 = (($$2740$ph$i) + ($853)|0);
     $855 = (($linecount114$0$ph$i) + -1)|0;
     $856 = ($855|0)==(0);
     if ($856) {
      break;
     } else {
      $$27$ph$i = $$28$i$lcssa;$$2740$ph$i = $854;$linecount114$0$ph$i = $855;
     }
    }
    break;
   }
   default: {
    break L44;
   }
   }
  }
 } while(0);
 $1878 = HEAP32[$surf_dst>>2]|0;
 $1879 = $1878 & 2;
 $1880 = ($1879|0)==(0);
 if ($1880) {
  $$0 = 0;
  return ($$0|0);
 }
 _SDL_UnlockSurface($surf_dst);
 $$0 = 0;
 return ($$0|0);
}
function _SDL_LockSurface($surface) {
 $surface = $surface|0;
 var $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($surface)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  $3 = HEAP32[$surface>>2]|0;
  $4 = $3 & 2;
  $5 = ($4|0)==(0);
  if ($5) {
   $9 = 0;
  } else {
   _SDL_UnRLESurface($surface,1);
   $6 = HEAP32[$surface>>2]|0;
   $7 = $6 | 2;
   HEAP32[$surface>>2] = $7;
   $$pre = HEAP32[$0>>2]|0;
   $9 = $$pre;
  }
 } else {
  $9 = $1;
 }
 $8 = (($9) + 1)|0;
 HEAP32[$0>>2] = $8;
 return 0;
}
function _SDL_UnlockSurface($surface) {
 $surface = $surface|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($surface)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  return;
 }
 $3 = (($1) + -1)|0;
 HEAP32[$0>>2] = $3;
 $4 = ($1|0)>(1);
 if ($4) {
  return;
 }
 $5 = HEAP32[$surface>>2]|0;
 $6 = $5 & 2;
 $7 = ($6|0)==(0);
 if ($7) {
  return;
 }
 $8 = $5 & -3;
 HEAP32[$surface>>2] = $8;
 (_SDL_RLESurface($surface)|0);
 return;
}
function _uncopy_opaque_16($dst,$src,$n,$sfmt,$dfmt) {
 $dst = $dst|0;
 $src = $src|0;
 $n = $n|0;
 $sfmt = $sfmt|0;
 $dfmt = $dfmt|0;
 var $$03 = 0, $$phi$trans$insert = 0, $$phi$trans$insert10 = 0, $$phi$trans$insert11 = 0, $$phi$trans$insert5 = 0, $$phi$trans$insert8 = 0, $$pre = 0, $$pre12 = 0, $$pre4 = 0, $$pre6 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $exitcond = 0, $i$02 = 0, $s$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)>(0);
 if (!($0)) {
  $68 = $n << 1;
  return ($68|0);
 }
 $1 = ((($dfmt)) + 24|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)!=(0);
 $4 = $3 ? 255 : 0;
 $5 = ((($sfmt)) + 4|0);
 $6 = ((($sfmt)) + 24|0);
 $7 = ((($sfmt)) + 20|0);
 $8 = ((($sfmt)) + 8|0);
 $9 = ((($sfmt)) + 12|0);
 $10 = ((($dfmt)) + 28|0);
 $11 = ((($dfmt)) + 32|0);
 $$pre = HEAP32[$6>>2]|0;
 $$pre4 = HEAP32[$7>>2]|0;
 $$phi$trans$insert = $$pre4 & 255;
 $$phi$trans$insert5 = (125924 + ($$phi$trans$insert<<2)|0);
 $$pre6 = HEAP32[$$phi$trans$insert5>>2]|0;
 $12 = $$pre >>> 8;
 $13 = $12 & 255;
 $14 = $$pre4 >>> 8;
 $15 = $14 & 255;
 $$phi$trans$insert8 = (125924 + ($15<<2)|0);
 $$pre9 = HEAP32[$$phi$trans$insert8>>2]|0;
 $16 = $$pre >>> 16;
 $17 = $$pre4 >>> 16;
 $$phi$trans$insert10 = $17 & 255;
 $$phi$trans$insert11 = (125924 + ($$phi$trans$insert10<<2)|0);
 $$pre12 = HEAP32[$$phi$trans$insert11>>2]|0;
 $18 = HEAP32[$10>>2]|0;
 $19 = HEAP32[$11>>2]|0;
 $20 = $18 >>> 8;
 $21 = $20 & 255;
 $22 = $19 >>> 8;
 $23 = $22 & 255;
 $24 = $18 >>> 16;
 $25 = $19 >>> 16;
 $26 = $18 >>> 24;
 $27 = $19 >>> 24;
 $28 = $$pre & 255;
 $29 = $16 & 255;
 $30 = $18 & 255;
 $31 = $19 & 255;
 $32 = $24 & 255;
 $33 = $25 & 255;
 $34 = $4 >>> $26;
 $35 = $34 << $27;
 $$03 = $dst;$i$02 = 0;$s$01 = $src;
 while(1) {
  $36 = HEAP16[$s$01>>1]|0;
  $37 = $36&65535;
  $38 = HEAP32[$5>>2]|0;
  $39 = $37 & $38;
  $40 = $39 >>> $28;
  $41 = (($$pre6) + ($40)|0);
  $42 = HEAP8[$41>>0]|0;
  $43 = $42&255;
  $44 = HEAP32[$8>>2]|0;
  $45 = $44 & $37;
  $46 = $45 >>> $13;
  $47 = (($$pre9) + ($46)|0);
  $48 = HEAP8[$47>>0]|0;
  $49 = $48&255;
  $50 = HEAP32[$9>>2]|0;
  $51 = $50 & $37;
  $52 = $51 >>> $29;
  $53 = (($$pre12) + ($52)|0);
  $54 = HEAP8[$53>>0]|0;
  $55 = $54&255;
  $56 = $43 >>> $30;
  $57 = $56 << $31;
  $58 = $49 >>> $21;
  $59 = $58 << $23;
  $60 = $59 | $57;
  $61 = $55 >>> $32;
  $62 = $61 << $33;
  $63 = $60 | $62;
  $64 = $63 | $35;
  HEAP32[$$03>>2] = $64;
  $65 = ((($s$01)) + 2|0);
  $66 = ((($$03)) + 4|0);
  $67 = (($i$02) + 1)|0;
  $exitcond = ($67|0)==($n|0);
  if ($exitcond) {
   break;
  } else {
   $$03 = $66;$i$02 = $67;$s$01 = $65;
  }
 }
 $68 = $n << 1;
 return ($68|0);
}
function _uncopy_32($dst,$src,$n,$sfmt,$dfmt) {
 $dst = $dst|0;
 $src = $src|0;
 $n = $n|0;
 $sfmt = $sfmt|0;
 $dfmt = $dfmt|0;
 var $$03 = 0, $$phi$trans$insert = 0, $$phi$trans$insert10 = 0, $$phi$trans$insert11 = 0, $$phi$trans$insert5 = 0, $$phi$trans$insert8 = 0, $$pre = 0, $$pre12 = 0, $$pre4 = 0, $$pre6 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$02 = 0, $s$01 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)>(0);
 if (!($0)) {
  $64 = $n << 2;
  return ($64|0);
 }
 $1 = ((($sfmt)) + 4|0);
 $2 = ((($sfmt)) + 24|0);
 $3 = ((($sfmt)) + 20|0);
 $4 = ((($sfmt)) + 8|0);
 $5 = ((($sfmt)) + 12|0);
 $6 = ((($dfmt)) + 28|0);
 $7 = ((($dfmt)) + 32|0);
 $$pre = HEAP32[$2>>2]|0;
 $$pre4 = HEAP32[$3>>2]|0;
 $$phi$trans$insert = $$pre4 & 255;
 $$phi$trans$insert5 = (125924 + ($$phi$trans$insert<<2)|0);
 $$pre6 = HEAP32[$$phi$trans$insert5>>2]|0;
 $8 = $$pre >>> 8;
 $9 = $8 & 255;
 $10 = $$pre4 >>> 8;
 $11 = $10 & 255;
 $$phi$trans$insert8 = (125924 + ($11<<2)|0);
 $$pre9 = HEAP32[$$phi$trans$insert8>>2]|0;
 $12 = $$pre >>> 16;
 $13 = $$pre4 >>> 16;
 $$phi$trans$insert10 = $13 & 255;
 $$phi$trans$insert11 = (125924 + ($$phi$trans$insert10<<2)|0);
 $$pre12 = HEAP32[$$phi$trans$insert11>>2]|0;
 $14 = HEAP32[$6>>2]|0;
 $15 = HEAP32[$7>>2]|0;
 $16 = $14 >>> 8;
 $17 = $16 & 255;
 $18 = $15 >>> 8;
 $19 = $18 & 255;
 $20 = $14 >>> 16;
 $21 = $15 >>> 16;
 $22 = $14 >>> 24;
 $23 = $15 >>> 24;
 $24 = $$pre & 255;
 $25 = $12 & 255;
 $26 = $14 & 255;
 $27 = $15 & 255;
 $28 = $20 & 255;
 $29 = $21 & 255;
 $$03 = $dst;$i$02 = 0;$s$01 = $src;
 while(1) {
  $30 = ((($s$01)) + 4|0);
  $31 = HEAP32[$s$01>>2]|0;
  $32 = HEAP32[$1>>2]|0;
  $33 = $32 & $31;
  $34 = $33 >>> $24;
  $35 = (($$pre6) + ($34)|0);
  $36 = HEAP8[$35>>0]|0;
  $37 = $36&255;
  $38 = HEAP32[$4>>2]|0;
  $39 = $38 & $31;
  $40 = $39 >>> $9;
  $41 = (($$pre9) + ($40)|0);
  $42 = HEAP8[$41>>0]|0;
  $43 = $42&255;
  $44 = HEAP32[$5>>2]|0;
  $45 = $44 & $31;
  $46 = $45 >>> $25;
  $47 = (($$pre12) + ($46)|0);
  $48 = HEAP8[$47>>0]|0;
  $49 = $48&255;
  $50 = $31 >>> 24;
  $51 = $37 >>> $26;
  $52 = $51 << $27;
  $53 = $43 >>> $17;
  $54 = $53 << $19;
  $55 = $54 | $52;
  $56 = $49 >>> $28;
  $57 = $56 << $29;
  $58 = $55 | $57;
  $59 = $50 >>> $22;
  $60 = $59 << $23;
  $61 = $58 | $60;
  HEAP32[$$03>>2] = $61;
  $62 = ((($$03)) + 4|0);
  $63 = (($i$02) + 1)|0;
  $exitcond = ($63|0)==($n|0);
  if ($exitcond) {
   break;
  } else {
   $$03 = $62;$i$02 = $63;$s$01 = $30;
  }
 }
 $64 = $n << 2;
 return ($64|0);
}
function _uncopy_transl_16($dst,$src,$n,$sfmt,$dfmt) {
 $dst = $dst|0;
 $src = $src|0;
 $n = $n|0;
 $sfmt = $sfmt|0;
 $dfmt = $dfmt|0;
 var $$03 = 0, $$phi$trans$insert = 0, $$phi$trans$insert10 = 0, $$phi$trans$insert11 = 0, $$phi$trans$insert5 = 0, $$phi$trans$insert8 = 0, $$pre = 0, $$pre12 = 0, $$pre4 = 0, $$pre6 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $exitcond = 0, $i$02 = 0, $s$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)>(0);
 if (!($0)) {
  $68 = $n << 2;
  return ($68|0);
 }
 $1 = ((($sfmt)) + 4|0);
 $2 = ((($sfmt)) + 24|0);
 $3 = ((($sfmt)) + 20|0);
 $4 = ((($sfmt)) + 8|0);
 $5 = ((($sfmt)) + 12|0);
 $6 = ((($dfmt)) + 28|0);
 $7 = ((($dfmt)) + 32|0);
 $$pre = HEAP32[$2>>2]|0;
 $$pre4 = HEAP32[$3>>2]|0;
 $$phi$trans$insert = $$pre4 & 255;
 $$phi$trans$insert5 = (125924 + ($$phi$trans$insert<<2)|0);
 $$pre6 = HEAP32[$$phi$trans$insert5>>2]|0;
 $8 = $$pre >>> 8;
 $9 = $8 & 255;
 $10 = $$pre4 >>> 8;
 $11 = $10 & 255;
 $$phi$trans$insert8 = (125924 + ($11<<2)|0);
 $$pre9 = HEAP32[$$phi$trans$insert8>>2]|0;
 $12 = $$pre >>> 16;
 $13 = $$pre4 >>> 16;
 $$phi$trans$insert10 = $13 & 255;
 $$phi$trans$insert11 = (125924 + ($$phi$trans$insert10<<2)|0);
 $$pre12 = HEAP32[$$phi$trans$insert11>>2]|0;
 $14 = HEAP32[$6>>2]|0;
 $15 = HEAP32[$7>>2]|0;
 $16 = $14 >>> 8;
 $17 = $16 & 255;
 $18 = $15 >>> 8;
 $19 = $18 & 255;
 $20 = $14 >>> 16;
 $21 = $15 >>> 16;
 $22 = $14 >>> 24;
 $23 = $15 >>> 24;
 $24 = $$pre & 255;
 $25 = $12 & 255;
 $26 = $14 & 255;
 $27 = $15 & 255;
 $28 = $20 & 255;
 $29 = $21 & 255;
 $$03 = $dst;$i$02 = 0;$s$01 = $src;
 while(1) {
  $30 = ((($s$01)) + 4|0);
  $31 = HEAP32[$s$01>>2]|0;
  $32 = $31 >>> 2;
  $33 = $32 & 248;
  $34 = $31 & -993;
  $35 = $31 >>> 16;
  $36 = $34 | $35;
  $37 = HEAP32[$1>>2]|0;
  $38 = $36 & $37;
  $39 = $38 >>> $24;
  $40 = (($$pre6) + ($39)|0);
  $41 = HEAP8[$40>>0]|0;
  $42 = $41&255;
  $43 = HEAP32[$4>>2]|0;
  $44 = $43 & $36;
  $45 = $44 >>> $9;
  $46 = (($$pre9) + ($45)|0);
  $47 = HEAP8[$46>>0]|0;
  $48 = $47&255;
  $49 = HEAP32[$5>>2]|0;
  $50 = $49 & $36;
  $51 = $50 >>> $25;
  $52 = (($$pre12) + ($51)|0);
  $53 = HEAP8[$52>>0]|0;
  $54 = $53&255;
  $55 = $42 >>> $26;
  $56 = $55 << $27;
  $57 = $48 >>> $17;
  $58 = $57 << $19;
  $59 = $58 | $56;
  $60 = $54 >>> $28;
  $61 = $60 << $29;
  $62 = $59 | $61;
  $63 = $33 >>> $22;
  $64 = $63 << $23;
  $65 = $62 | $64;
  HEAP32[$$03>>2] = $65;
  $66 = ((($$03)) + 4|0);
  $67 = (($i$02) + 1)|0;
  $exitcond = ($67|0)==($n|0);
  if ($exitcond) {
   break;
  } else {
   $$03 = $66;$i$02 = $67;$s$01 = $30;
  }
 }
 $68 = $n << 2;
 return ($68|0);
}
function _SDL_free($ptr) {
 $ptr = $ptr|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _free($ptr);
 return;
}
function _getpix_8($srcbuf) {
 $srcbuf = $srcbuf|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$srcbuf>>0]|0;
 $1 = $0&255;
 return ($1|0);
}
function _getpix_16($srcbuf) {
 $srcbuf = $srcbuf|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP16[$srcbuf>>1]|0;
 $1 = $0&65535;
 return ($1|0);
}
function _getpix_24($srcbuf) {
 $srcbuf = $srcbuf|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$srcbuf>>0]|0;
 $1 = $0&255;
 $2 = ((($srcbuf)) + 1|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = $4 << 8;
 $6 = $5 | $1;
 $7 = ((($srcbuf)) + 2|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = $8&255;
 $10 = $9 << 16;
 $11 = $6 | $10;
 return ($11|0);
}
function _getpix_32($srcbuf) {
 $srcbuf = $srcbuf|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$srcbuf>>2]|0;
 return ($0|0);
}
function _copy_transl_565($dst,$src,$n,$sfmt,$dfmt) {
 $dst = $dst|0;
 $src = $src|0;
 $n = $n|0;
 $sfmt = $sfmt|0;
 $dfmt = $dfmt|0;
 var $$03 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $d$01 = 0, $exitcond = 0, $i$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)>(0);
 if (!($0)) {
  $61 = $n << 2;
  return ($61|0);
 }
 $1 = ((($sfmt)) + 12|0);
 $2 = ((($sfmt)) + 32|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = ((($sfmt)) + 16|0);
 $6 = $3 >>> 8;
 $7 = $6 & 255;
 $8 = ((($sfmt)) + 20|0);
 $9 = $3 >>> 16;
 $10 = $9 & 255;
 $11 = ((($sfmt)) + 24|0);
 $12 = $3 >>> 24;
 $13 = ((($dfmt)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = $14 & 255;
 $16 = ((($dfmt)) + 32|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 & 255;
 $19 = $14 >>> 8;
 $20 = $19 & 255;
 $21 = $17 >>> 8;
 $22 = $21 & 255;
 $23 = $14 >>> 16;
 $24 = $23 & 255;
 $25 = $17 >>> 16;
 $26 = $25 & 255;
 $27 = ((($dfmt)) + 24|0);
 $$03 = $src;$d$01 = $dst;$i$02 = 0;
 while(1) {
  $28 = HEAP32[$$03>>2]|0;
  $29 = HEAP32[$1>>2]|0;
  $30 = $29 & $28;
  $31 = $30 >>> $4;
  $32 = HEAP32[$5>>2]|0;
  $33 = $32 & $28;
  $34 = $33 >>> $7;
  $35 = HEAP32[$8>>2]|0;
  $36 = $35 & $28;
  $37 = $36 >>> $10;
  $38 = HEAP32[$11>>2]|0;
  $39 = $38 & $28;
  $40 = $39 >>> $12;
  $41 = $31 >>> $15;
  $42 = $41 << $18;
  $43 = $34 >>> $20;
  $44 = $43 << $22;
  $45 = $44 | $42;
  $46 = $37 >>> $24;
  $47 = $46 << $26;
  $48 = HEAP32[$27>>2]|0;
  $49 = $45 | $48;
  $50 = $49 | $47;
  $51 = $50 << 16;
  $52 = $51 & 132120576;
  $53 = $50 & 63519;
  $54 = $40 << 2;
  $55 = $54 & 2016;
  $56 = $53 | $55;
  $57 = $56 | $52;
  HEAP32[$d$01>>2] = $57;
  $58 = ((($$03)) + 4|0);
  $59 = ((($d$01)) + 4|0);
  $60 = (($i$02) + 1)|0;
  $exitcond = ($60|0)==($n|0);
  if ($exitcond) {
   break;
  } else {
   $$03 = $58;$d$01 = $59;$i$02 = $60;
  }
 }
 $61 = $n << 2;
 return ($61|0);
}
function _copy_transl_555($dst,$src,$n,$sfmt,$dfmt) {
 $dst = $dst|0;
 $src = $src|0;
 $n = $n|0;
 $sfmt = $sfmt|0;
 $dfmt = $dfmt|0;
 var $$03 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $d$01 = 0, $exitcond = 0, $i$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)>(0);
 if (!($0)) {
  $61 = $n << 2;
  return ($61|0);
 }
 $1 = ((($sfmt)) + 12|0);
 $2 = ((($sfmt)) + 32|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = ((($sfmt)) + 16|0);
 $6 = $3 >>> 8;
 $7 = $6 & 255;
 $8 = ((($sfmt)) + 20|0);
 $9 = $3 >>> 16;
 $10 = $9 & 255;
 $11 = ((($sfmt)) + 24|0);
 $12 = $3 >>> 24;
 $13 = ((($dfmt)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = $14 & 255;
 $16 = ((($dfmt)) + 32|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 & 255;
 $19 = $14 >>> 8;
 $20 = $19 & 255;
 $21 = $17 >>> 8;
 $22 = $21 & 255;
 $23 = $14 >>> 16;
 $24 = $23 & 255;
 $25 = $17 >>> 16;
 $26 = $25 & 255;
 $27 = ((($dfmt)) + 24|0);
 $$03 = $src;$d$01 = $dst;$i$02 = 0;
 while(1) {
  $28 = HEAP32[$$03>>2]|0;
  $29 = HEAP32[$1>>2]|0;
  $30 = $29 & $28;
  $31 = $30 >>> $4;
  $32 = HEAP32[$5>>2]|0;
  $33 = $32 & $28;
  $34 = $33 >>> $7;
  $35 = HEAP32[$8>>2]|0;
  $36 = $35 & $28;
  $37 = $36 >>> $10;
  $38 = HEAP32[$11>>2]|0;
  $39 = $38 & $28;
  $40 = $39 >>> $12;
  $41 = $31 >>> $15;
  $42 = $41 << $18;
  $43 = $34 >>> $20;
  $44 = $43 << $22;
  $45 = $44 | $42;
  $46 = $37 >>> $24;
  $47 = $46 << $26;
  $48 = HEAP32[$27>>2]|0;
  $49 = $45 | $48;
  $50 = $49 | $47;
  $51 = $50 << 16;
  $52 = $51 & 65011712;
  $53 = $50 & 64543;
  $54 = $40 << 2;
  $55 = $54 & 992;
  $56 = $53 | $55;
  $57 = $56 | $52;
  HEAP32[$d$01>>2] = $57;
  $58 = ((($$03)) + 4|0);
  $59 = ((($d$01)) + 4|0);
  $60 = (($i$02) + 1)|0;
  $exitcond = ($60|0)==($n|0);
  if ($exitcond) {
   break;
  } else {
   $$03 = $58;$d$01 = $59;$i$02 = $60;
  }
 }
 $61 = $n << 2;
 return ($61|0);
}
function _copy_32($dst,$src,$n,$sfmt,$dfmt) {
 $dst = $dst|0;
 $src = $src|0;
 $n = $n|0;
 $sfmt = $sfmt|0;
 $dfmt = $dfmt|0;
 var $$03 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $d$01 = 0, $exitcond = 0, $i$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)>(0);
 if (!($0)) {
  $53 = $n << 2;
  return ($53|0);
 }
 $1 = ((($sfmt)) + 12|0);
 $2 = ((($sfmt)) + 32|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = ((($sfmt)) + 16|0);
 $6 = $3 >>> 8;
 $7 = $6 & 255;
 $8 = ((($sfmt)) + 20|0);
 $9 = $3 >>> 16;
 $10 = $9 & 255;
 $11 = ((($sfmt)) + 24|0);
 $12 = $3 >>> 24;
 $13 = ((($dfmt)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = $14 & 255;
 $16 = ((($dfmt)) + 32|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 & 255;
 $19 = $14 >>> 8;
 $20 = $19 & 255;
 $21 = $17 >>> 8;
 $22 = $21 & 255;
 $23 = $14 >>> 16;
 $24 = $23 & 255;
 $25 = $17 >>> 16;
 $26 = $25 & 255;
 $$03 = $src;$d$01 = $dst;$i$02 = 0;
 while(1) {
  $27 = HEAP32[$$03>>2]|0;
  $28 = HEAP32[$1>>2]|0;
  $29 = $28 & $27;
  $30 = $29 >>> $4;
  $31 = HEAP32[$5>>2]|0;
  $32 = $31 & $27;
  $33 = $32 >>> $7;
  $34 = HEAP32[$8>>2]|0;
  $35 = $34 & $27;
  $36 = $35 >>> $10;
  $37 = HEAP32[$11>>2]|0;
  $38 = $37 & $27;
  $39 = $38 >>> $12;
  $40 = $30 >>> $15;
  $41 = $40 << $18;
  $42 = $33 >>> $20;
  $43 = $42 << $22;
  $44 = $36 >>> $24;
  $45 = $44 << $26;
  $46 = $39 << 24;
  $47 = $41 | $46;
  $48 = $47 | $43;
  $49 = $48 | $45;
  HEAP32[$d$01>>2] = $49;
  $50 = ((($d$01)) + 4|0);
  $51 = ((($$03)) + 4|0);
  $52 = (($i$02) + 1)|0;
  $exitcond = ($52|0)==($n|0);
  if ($exitcond) {
   break;
  } else {
   $$03 = $51;$d$01 = $50;$i$02 = $52;
  }
 }
 $53 = $n << 2;
 return ($53|0);
}
function _copy_opaque_16($dst,$src,$n,$sfmt,$dfmt) {
 $dst = $dst|0;
 $src = $src|0;
 $n = $n|0;
 $sfmt = $sfmt|0;
 $dfmt = $dfmt|0;
 var $$03 = 0, $$phi$trans$insert = 0, $$phi$trans$insert10 = 0, $$phi$trans$insert13 = 0, $$phi$trans$insert14 = 0, $$phi$trans$insert6 = 0, $$pre = 0, $$pre11 = 0, $$pre12 = 0, $$pre15 = 0, $$pre16 = 0, $$pre17 = 0, $$pre18 = 0, $$pre4 = 0, $$pre5 = 0, $$pre7 = 0, $$pre8 = 0, $0 = 0, $1 = 0, $10 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $d$01 = 0, $exitcond = 0, $i$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)>(0);
 if (!($0)) {
  $56 = $n << 1;
  return ($56|0);
 }
 $1 = ((($sfmt)) + 12|0);
 $2 = ((($sfmt)) + 32|0);
 $3 = ((($sfmt)) + 28|0);
 $4 = ((($sfmt)) + 16|0);
 $5 = ((($sfmt)) + 20|0);
 $6 = ((($dfmt)) + 28|0);
 $7 = ((($dfmt)) + 32|0);
 $8 = ((($dfmt)) + 24|0);
 $$pre = HEAP32[$1>>2]|0;
 $$pre4 = HEAP32[$2>>2]|0;
 $$pre5 = HEAP32[$3>>2]|0;
 $$phi$trans$insert = $$pre5 & 255;
 $$phi$trans$insert6 = (125924 + ($$phi$trans$insert<<2)|0);
 $$pre7 = HEAP32[$$phi$trans$insert6>>2]|0;
 $$pre8 = HEAP32[$4>>2]|0;
 $9 = $$pre4 >>> 8;
 $10 = $9 & 255;
 $11 = $$pre5 >>> 8;
 $12 = $11 & 255;
 $$phi$trans$insert10 = (125924 + ($12<<2)|0);
 $$pre11 = HEAP32[$$phi$trans$insert10>>2]|0;
 $$pre12 = HEAP32[$5>>2]|0;
 $13 = $$pre4 >>> 16;
 $14 = $$pre5 >>> 16;
 $$phi$trans$insert13 = $14 & 255;
 $$phi$trans$insert14 = (125924 + ($$phi$trans$insert13<<2)|0);
 $$pre15 = HEAP32[$$phi$trans$insert14>>2]|0;
 $$pre16 = HEAP32[$6>>2]|0;
 $$pre17 = HEAP32[$7>>2]|0;
 $15 = $$pre16 >>> 8;
 $16 = $15 & 255;
 $17 = $$pre17 >>> 8;
 $18 = $17 & 255;
 $19 = $$pre16 >>> 16;
 $20 = $$pre17 >>> 16;
 $$pre18 = HEAP32[$8>>2]|0;
 $21 = $$pre4 & 255;
 $22 = $13 & 255;
 $23 = $$pre16 & 255;
 $24 = $$pre17 & 255;
 $25 = $19 & 255;
 $26 = $20 & 255;
 $$03 = $src;$d$01 = $dst;$i$02 = 0;
 while(1) {
  $27 = HEAP32[$$03>>2]|0;
  $28 = $$pre & $27;
  $29 = $28 >>> $21;
  $30 = (($$pre7) + ($29)|0);
  $31 = HEAP8[$30>>0]|0;
  $32 = $31&255;
  $33 = $$pre8 & $27;
  $34 = $33 >>> $10;
  $35 = (($$pre11) + ($34)|0);
  $36 = HEAP8[$35>>0]|0;
  $37 = $36&255;
  $38 = $$pre12 & $27;
  $39 = $38 >>> $22;
  $40 = (($$pre15) + ($39)|0);
  $41 = HEAP8[$40>>0]|0;
  $42 = $41&255;
  $43 = $32 >>> $23;
  $44 = $43 << $24;
  $45 = $37 >>> $16;
  $46 = $45 << $18;
  $47 = $46 | $44;
  $48 = $42 >>> $25;
  $49 = $48 << $26;
  $50 = $47 | $$pre18;
  $51 = $50 | $49;
  $52 = $51&65535;
  HEAP16[$d$01>>1] = $52;
  $53 = ((($$03)) + 4|0);
  $54 = ((($d$01)) + 2|0);
  $55 = (($i$02) + 1)|0;
  $exitcond = ($55|0)==($n|0);
  if ($exitcond) {
   break;
  } else {
   $$03 = $53;$d$01 = $54;$i$02 = $55;
  }
 }
 $56 = $n << 1;
 return ($56|0);
}
function _SDL_RLEAlphaBlit($surf_src,$srcrect,$surf_dst,$dstrect) {
 $surf_src = $surf_src|0;
 $srcrect = $srcrect|0;
 $surf_dst = $surf_dst|0;
 $dstrect = $dstrect|0;
 var $$$i = 0, $$0 = 0, $$0$i = 0, $$09$i = 0, $$1$i = 0, $$11$i = 0, $$12$i = 0, $$12$i$lcssa = 0, $$14$i = 0, $$15$i = 0, $$16$i = 0, $$16$i$lcssa = 0, $$18$i = 0, $$19$i = 0, $$19$i$lcssa = 0, $$2$i = 0, $$2$i$lcssa = 0, $$211$i = 0, $$24$i = 0, $$25$i = 0;
 var $$26$i = 0, $$27$i = 0, $$28$i = 0, $$29$i = 0, $$30$i = 0, $$31$i = 0, $$32$i = 0, $$33$i = 0, $$34$i = 0, $$4$i = 0, $$413$i = 0, $$5$i = 0, $$5$i$lcssa = 0, $$7$i = 0, $$8$i = 0, $$9$i = 0, $$9$i$lcssa = 0, $$lcssa = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0;
 var $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0;
 var $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0;
 var $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0;
 var $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0;
 var $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0;
 var $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0;
 var $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0;
 var $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0;
 var $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0;
 var $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0;
 var $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0;
 var $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0;
 var $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0;
 var $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0;
 var $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0;
 var $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0;
 var $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0;
 var $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $crun$1$i = 0, $crun12$1$i = 0;
 var $crun16$1$i = 0, $crun3$1$i = 0, $crun31$1$i = 0, $crun35$1$i = 0, $dst$066 = 0, $dst17$068 = 0, $dst30$075 = 0, $dstbuf$0 = 0, $dstbuf$2 = 0, $dstbuf$4 = 0, $exitcond = 0, $exitcond102 = 0, $exitcond105 = 0, $i$061$i = 0, $i$065 = 0, $i18$069 = 0, $i19$062$i = 0, $i31$076 = 0, $i38$064$i = 0, $linecount$0 = 0;
 var $linecount$0$i = 0, $linecount12$0 = 0, $linecount25$0 = 0, $linecount25$0$i = 0, $linecount6$0$i = 0, $ofs$0 = 0, $ofs$0$i = 0, $ofs$1 = 0, $ofs$1$i = 0, $ofs$3 = 0, $ofs$3$i = 0, $ofs$4 = 0, $ofs$4$i = 0, $ofs$5 = 0, $ofs13$0 = 0, $ofs13$1 = 0, $ofs13$3 = 0, $ofs13$4 = 0, $ofs26$0 = 0, $ofs26$1 = 0;
 var $ofs26$3 = 0, $ofs26$4 = 0, $ofs28$0$i = 0, $ofs28$1$i = 0, $ofs28$3$i = 0, $ofs28$4$i = 0, $ofs6$0 = 0, $ofs6$1 = 0, $ofs6$3 = 0, $ofs6$4 = 0, $ofs9$0$i = 0, $ofs9$1$i = 0, $ofs9$3$i = 0, $ofs9$4$i = 0, $scevgep101 = 0, $scevgep104 = 0, $scevgep99 = 0, $srcbuf$0 = 0, $srcbuf$1 = 0, $srcbuf$12$ph = 0;
 var $srcbuf$13 = 0, $srcbuf$14 = 0, $srcbuf$15 = 0, $srcbuf$15$lcssa = 0, $srcbuf$17 = 0, $srcbuf$1867 = 0, $srcbuf$19 = 0, $srcbuf$19$lcssa = 0, $srcbuf$2 = 0, $srcbuf$2$lcssa = 0, $srcbuf$21 = 0, $srcbuf$22 = 0, $srcbuf$23 = 0, $srcbuf$23$lcssa = 0, $srcbuf$25 = 0, $srcbuf$2670 = 0, $srcbuf$27 = 0, $srcbuf$27$lcssa = 0, $srcbuf$29 = 0, $srcbuf$30 = 0;
 var $srcbuf$31 = 0, $srcbuf$31$lcssa = 0, $srcbuf$33 = 0, $srcbuf$3477 = 0, $srcbuf$35 = 0, $srcbuf$35$lcssa = 0, $srcbuf$4 = 0, $srcbuf$5 = 0, $srcbuf$6 = 0, $srcbuf$7 = 0, $srcbuf$7$lcssa = 0, $vskip$0 = 0, $vskip$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($surf_src)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($surf_dst)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$surf_dst>>2]|0;
 $5 = $4 & 2;
 $6 = ($5|0)==(0);
 if (!($6)) {
  $7 = (_SDL_LockSurface($surf_dst)|0);
  $8 = ($7|0)<(0);
  if ($8) {
   $$0 = -1;
   return ($$0|0);
  }
 }
 $9 = HEAP32[$dstrect>>2]|0;
 $10 = ((($dstrect)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($surf_dst)) + 20|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($surf_dst)) + 16|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = Math_imul($15, $11)|0;
 $17 = (($13) + ($16)|0);
 $18 = ((($3)) + 9|0);
 $19 = HEAP8[$18>>0]|0;
 $20 = $19&255;
 $21 = Math_imul($20, $9)|0;
 $22 = (($17) + ($21)|0);
 $23 = ((($surf_src)) + 52|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($24)) + 12|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ((($26)) + 28|0);
 $28 = ((($srcrect)) + 4|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = ($29|0)==(0);
 L5: do {
  if ($30) {
   $srcbuf$12$ph = $27;
   label = 20;
  } else {
   $31 = ($19<<24>>24)==(2);
   if ($31) {
    $srcbuf$0 = $27;$vskip$0 = $29;
   } else {
    $61 = $29 << 1;
    $srcbuf$5 = $27;$vskip$1 = $61;
    while(1) {
     $ofs$4 = 0;$srcbuf$6 = $srcbuf$5;
     while(1) {
      $62 = HEAP16[$srcbuf$6>>1]|0;
      $63 = $62&65535;
      $64 = (($63) + ($ofs$4))|0;
      $65 = ((($srcbuf$6)) + 2|0);
      $66 = HEAP16[$65>>1]|0;
      $67 = $66&65535;
      $68 = ((($srcbuf$6)) + 4|0);
      $69 = ($66<<16>>16)==(0);
      if ($69) {
       $73 = ($64|0)==(0);
       if ($73) {
        break L5;
       } else {
        $ofs$5 = $64;$srcbuf$7 = $68;
       }
      } else {
       $70 = $67 << 2;
       $71 = (($68) + ($70)|0);
       $72 = (($64) + ($67))|0;
       $ofs$5 = $72;$srcbuf$7 = $71;
      }
      $74 = ($ofs$5|0)<($1|0);
      if ($74) {
       $ofs$4 = $ofs$5;$srcbuf$6 = $srcbuf$7;
      } else {
       $srcbuf$7$lcssa = $srcbuf$7;
       break;
      }
     }
     $75 = (($vskip$1) + -1)|0;
     $76 = ($75|0)==(0);
     if ($76) {
      $srcbuf$12$ph = $srcbuf$7$lcssa;
      label = 20;
      break L5;
     } else {
      $srcbuf$5 = $srcbuf$7$lcssa;$vskip$1 = $75;
     }
    }
   }
   while(1) {
    $ofs$0 = 0;$srcbuf$1 = $srcbuf$0;
    while(1) {
     $32 = HEAP8[$srcbuf$1>>0]|0;
     $33 = $32&255;
     $34 = (($33) + ($ofs$0))|0;
     $35 = ((($srcbuf$1)) + 1|0);
     $36 = HEAP8[$35>>0]|0;
     $37 = $36&255;
     $38 = ((($srcbuf$1)) + 2|0);
     $39 = ($36<<24>>24)==(0);
     if ($39) {
      $43 = ($34|0)==(0);
      if ($43) {
       break L5;
      } else {
       $ofs$1 = $34;$srcbuf$2 = $38;
      }
     } else {
      $40 = $37 << 1;
      $41 = (($38) + ($40)|0);
      $42 = (($34) + ($37))|0;
      $ofs$1 = $42;$srcbuf$2 = $41;
     }
     $44 = ($ofs$1|0)<($1|0);
     if ($44) {
      $ofs$0 = $ofs$1;$srcbuf$1 = $srcbuf$2;
     } else {
      $srcbuf$2$lcssa = $srcbuf$2;
      break;
     }
    }
    $45 = $srcbuf$2$lcssa;
    $46 = $45 & 2;
    $47 = (($srcbuf$2$lcssa) + ($46)|0);
    $ofs$3 = 0;$srcbuf$4 = $47;
    while(1) {
     $48 = HEAP16[$srcbuf$4>>1]|0;
     $49 = $48&65535;
     $50 = (($49) + ($ofs$3))|0;
     $51 = ((($srcbuf$4)) + 2|0);
     $52 = HEAP16[$51>>1]|0;
     $53 = $52&65535;
     $54 = $53 << 2;
     $55 = (($54) + 4)|0;
     $56 = (($srcbuf$4) + ($55)|0);
     $57 = (($50) + ($53))|0;
     $58 = ($57|0)<($1|0);
     if ($58) {
      $ofs$3 = $57;$srcbuf$4 = $56;
     } else {
      $$lcssa = $56;
      break;
     }
    }
    $59 = (($vskip$0) + -1)|0;
    $60 = ($59|0)==(0);
    if ($60) {
     $srcbuf$12$ph = $$lcssa;
     label = 20;
     break;
    } else {
     $srcbuf$0 = $$lcssa;$vskip$0 = $59;
    }
   }
  }
 } while(0);
 L30: do {
  if ((label|0) == 20) {
   $77 = HEAP32[$srcrect>>2]|0;
   $78 = ($77|0)==(0);
   if ($78) {
    $79 = ((($srcrect)) + 8|0);
    $80 = HEAP32[$79>>2]|0;
    $81 = HEAP32[$0>>2]|0;
    $82 = ($80|0)==($81|0);
    if ($82) {
     switch ($20|0) {
     case 2:  {
      break;
     }
     case 4:  {
      $462 = ((($srcrect)) + 12|0);
      $463 = HEAP32[$462>>2]|0;
      $dstbuf$4 = $22;$linecount25$0 = $463;$srcbuf$29 = $srcbuf$12$ph;
      while(1) {
       $ofs26$0 = 0;$srcbuf$30 = $srcbuf$29;
       while(1) {
        $464 = HEAP16[$srcbuf$30>>1]|0;
        $465 = $464&65535;
        $466 = (($465) + ($ofs26$0))|0;
        $467 = ((($srcbuf$30)) + 2|0);
        $468 = HEAP16[$467>>1]|0;
        $469 = $468&65535;
        $470 = ((($srcbuf$30)) + 4|0);
        $471 = ($468<<16>>16)==(0);
        if ($471) {
         $477 = ($466|0)==(0);
         if ($477) {
          break L30;
         } else {
          $ofs26$1 = $466;$srcbuf$31 = $470;
         }
        } else {
         $472 = $466 << 2;
         $473 = (($dstbuf$4) + ($472)|0);
         $474 = $469 << 2;
         (_SDL_memcpy($473,$470,$474)|0);
         $475 = (($470) + ($474)|0);
         $476 = (($466) + ($469))|0;
         $ofs26$1 = $476;$srcbuf$31 = $475;
        }
        $478 = ($ofs26$1|0)<($1|0);
        if ($478) {
         $ofs26$0 = $ofs26$1;$srcbuf$30 = $srcbuf$31;
        } else {
         $srcbuf$31$lcssa = $srcbuf$31;
         break;
        }
       }
       $ofs26$3 = 0;$srcbuf$33 = $srcbuf$31$lcssa;
       while(1) {
        $479 = HEAP16[$srcbuf$33>>1]|0;
        $480 = $479&65535;
        $481 = (($480) + ($ofs26$3))|0;
        $482 = ((($srcbuf$33)) + 2|0);
        $483 = HEAP16[$482>>1]|0;
        $484 = $483&65535;
        $485 = ((($srcbuf$33)) + 4|0);
        $486 = ($483<<16>>16)==(0);
        if ($486) {
         $ofs26$4 = $481;$srcbuf$35 = $485;
        } else {
         $487 = (($dstbuf$4) + ($481<<2)|0);
         $488 = $484 << 2;
         $dst30$075 = $487;$i31$076 = 0;$srcbuf$3477 = $485;
         while(1) {
          $489 = HEAP32[$srcbuf$3477>>2]|0;
          $490 = HEAP32[$dst30$075>>2]|0;
          $491 = $489 >>> 24;
          $492 = $489 & 16711935;
          $493 = $490 & 16711935;
          $494 = (($492) - ($493))|0;
          $495 = Math_imul($494, $491)|0;
          $496 = $495 >>> 8;
          $497 = (($496) + ($493))|0;
          $498 = $497 & 16711935;
          $499 = $489 & 65280;
          $500 = $490 & 65280;
          $501 = (($499) - ($500))|0;
          $502 = Math_imul($501, $491)|0;
          $503 = $502 >>> 8;
          $504 = (($503) + ($500))|0;
          $505 = $504 & 65280;
          $506 = $505 | $498;
          $507 = $506 | -16777216;
          HEAP32[$dst30$075>>2] = $507;
          $508 = ((($srcbuf$3477)) + 4|0);
          $509 = ((($dst30$075)) + 4|0);
          $510 = (($i31$076) + 1)|0;
          $exitcond105 = ($510|0)==($484|0);
          if ($exitcond105) {
           break;
          } else {
           $dst30$075 = $509;$i31$076 = $510;$srcbuf$3477 = $508;
          }
         }
         $scevgep104 = (($485) + ($488)|0);
         $511 = (($481) + ($484))|0;
         $ofs26$4 = $511;$srcbuf$35 = $scevgep104;
        }
        $512 = ($ofs26$4|0)<($1|0);
        if ($512) {
         $ofs26$3 = $ofs26$4;$srcbuf$33 = $srcbuf$35;
        } else {
         $srcbuf$35$lcssa = $srcbuf$35;
         break;
        }
       }
       $513 = HEAP32[$14>>2]|0;
       $514 = (($dstbuf$4) + ($513)|0);
       $515 = (($linecount25$0) + -1)|0;
       $516 = ($515|0)==(0);
       if ($516) {
        break L30;
       } else {
        $dstbuf$4 = $514;$linecount25$0 = $515;$srcbuf$29 = $srcbuf$35$lcssa;
       }
      }
      break;
     }
     default: {
      break L30;
     }
     }
     $341 = ((($3)) + 16|0);
     $342 = HEAP32[$341>>2]|0;
     $343 = ($342|0)==(2016);
     if (!($343)) {
      $344 = ((($3)) + 12|0);
      $345 = HEAP32[$344>>2]|0;
      $346 = ($345|0)==(2016);
      if (!($346)) {
       $347 = ((($3)) + 20|0);
       $348 = HEAP32[$347>>2]|0;
       $349 = ($348|0)==(2016);
       if (!($349)) {
        $406 = ((($srcrect)) + 12|0);
        $407 = HEAP32[$406>>2]|0;
        $dstbuf$2 = $22;$linecount12$0 = $407;$srcbuf$21 = $srcbuf$12$ph;
        while(1) {
         $ofs13$0 = 0;$srcbuf$22 = $srcbuf$21;
         while(1) {
          $408 = HEAP8[$srcbuf$22>>0]|0;
          $409 = $408&255;
          $410 = (($409) + ($ofs13$0))|0;
          $411 = ((($srcbuf$22)) + 1|0);
          $412 = HEAP8[$411>>0]|0;
          $413 = $412&255;
          $414 = ((($srcbuf$22)) + 2|0);
          $415 = ($412<<24>>24)==(0);
          if ($415) {
           $421 = ($410|0)==(0);
           if ($421) {
            break L30;
           } else {
            $ofs13$1 = $410;$srcbuf$23 = $414;
           }
          } else {
           $416 = $410 << 1;
           $417 = (($dstbuf$2) + ($416)|0);
           $418 = $413 << 1;
           (_SDL_memcpy($417,$414,$418)|0);
           $419 = (($414) + ($418)|0);
           $420 = (($410) + ($413))|0;
           $ofs13$1 = $420;$srcbuf$23 = $419;
          }
          $422 = ($ofs13$1|0)<($1|0);
          if ($422) {
           $ofs13$0 = $ofs13$1;$srcbuf$22 = $srcbuf$23;
          } else {
           $srcbuf$23$lcssa = $srcbuf$23;
           break;
          }
         }
         $423 = $srcbuf$23$lcssa;
         $424 = $423 & 2;
         $425 = (($srcbuf$23$lcssa) + ($424)|0);
         $ofs13$3 = 0;$srcbuf$25 = $425;
         while(1) {
          $426 = HEAP16[$srcbuf$25>>1]|0;
          $427 = $426&65535;
          $428 = (($427) + ($ofs13$3))|0;
          $429 = ((($srcbuf$25)) + 2|0);
          $430 = HEAP16[$429>>1]|0;
          $431 = $430&65535;
          $432 = ((($srcbuf$25)) + 4|0);
          $433 = ($430<<16>>16)==(0);
          if ($433) {
           $ofs13$4 = $428;$srcbuf$27 = $432;
          } else {
           $434 = (($dstbuf$2) + ($428<<1)|0);
           $435 = $431 << 2;
           $dst17$068 = $434;$i18$069 = 0;$srcbuf$2670 = $432;
           while(1) {
            $436 = HEAP32[$srcbuf$2670>>2]|0;
            $437 = HEAP16[$dst17$068>>1]|0;
            $438 = $437&65535;
            $439 = $436 >>> 5;
            $440 = $439 & 31;
            $441 = $436 & 65043487;
            $442 = $438 << 16;
            $443 = $442 | $438;
            $444 = $443 & 65043487;
            $445 = (($441) - ($444))|0;
            $446 = Math_imul($445, $440)|0;
            $447 = $446 >>> 5;
            $448 = (($447) + ($444))|0;
            $449 = $448 & 65043487;
            $450 = $449 >>> 16;
            $451 = $450 | $449;
            $452 = $451&65535;
            HEAP16[$dst17$068>>1] = $452;
            $453 = ((($srcbuf$2670)) + 4|0);
            $454 = ((($dst17$068)) + 2|0);
            $455 = (($i18$069) + 1)|0;
            $exitcond102 = ($455|0)==($431|0);
            if ($exitcond102) {
             break;
            } else {
             $dst17$068 = $454;$i18$069 = $455;$srcbuf$2670 = $453;
            }
           }
           $scevgep101 = (($432) + ($435)|0);
           $456 = (($428) + ($431))|0;
           $ofs13$4 = $456;$srcbuf$27 = $scevgep101;
          }
          $457 = ($ofs13$4|0)<($1|0);
          if ($457) {
           $ofs13$3 = $ofs13$4;$srcbuf$25 = $srcbuf$27;
          } else {
           $srcbuf$27$lcssa = $srcbuf$27;
           break;
          }
         }
         $458 = HEAP32[$14>>2]|0;
         $459 = (($dstbuf$2) + ($458)|0);
         $460 = (($linecount12$0) + -1)|0;
         $461 = ($460|0)==(0);
         if ($461) {
          break L30;
         } else {
          $dstbuf$2 = $459;$linecount12$0 = $460;$srcbuf$21 = $srcbuf$27$lcssa;
         }
        }
       }
      }
     }
     $350 = ((($srcrect)) + 12|0);
     $351 = HEAP32[$350>>2]|0;
     $dstbuf$0 = $22;$linecount$0 = $351;$srcbuf$13 = $srcbuf$12$ph;
     while(1) {
      $ofs6$0 = 0;$srcbuf$14 = $srcbuf$13;
      while(1) {
       $352 = HEAP8[$srcbuf$14>>0]|0;
       $353 = $352&255;
       $354 = (($353) + ($ofs6$0))|0;
       $355 = ((($srcbuf$14)) + 1|0);
       $356 = HEAP8[$355>>0]|0;
       $357 = $356&255;
       $358 = ((($srcbuf$14)) + 2|0);
       $359 = ($356<<24>>24)==(0);
       if ($359) {
        $365 = ($354|0)==(0);
        if ($365) {
         break L30;
        } else {
         $ofs6$1 = $354;$srcbuf$15 = $358;
        }
       } else {
        $360 = $354 << 1;
        $361 = (($dstbuf$0) + ($360)|0);
        $362 = $357 << 1;
        (_SDL_memcpy($361,$358,$362)|0);
        $363 = (($358) + ($362)|0);
        $364 = (($354) + ($357))|0;
        $ofs6$1 = $364;$srcbuf$15 = $363;
       }
       $366 = ($ofs6$1|0)<($1|0);
       if ($366) {
        $ofs6$0 = $ofs6$1;$srcbuf$14 = $srcbuf$15;
       } else {
        $srcbuf$15$lcssa = $srcbuf$15;
        break;
       }
      }
      $367 = $srcbuf$15$lcssa;
      $368 = $367 & 2;
      $369 = (($srcbuf$15$lcssa) + ($368)|0);
      $ofs6$3 = 0;$srcbuf$17 = $369;
      while(1) {
       $370 = HEAP16[$srcbuf$17>>1]|0;
       $371 = $370&65535;
       $372 = (($371) + ($ofs6$3))|0;
       $373 = ((($srcbuf$17)) + 2|0);
       $374 = HEAP16[$373>>1]|0;
       $375 = $374&65535;
       $376 = ((($srcbuf$17)) + 4|0);
       $377 = ($374<<16>>16)==(0);
       if ($377) {
        $ofs6$4 = $372;$srcbuf$19 = $376;
       } else {
        $378 = (($dstbuf$0) + ($372<<1)|0);
        $379 = $375 << 2;
        $dst$066 = $378;$i$065 = 0;$srcbuf$1867 = $376;
        while(1) {
         $380 = HEAP32[$srcbuf$1867>>2]|0;
         $381 = HEAP16[$dst$066>>1]|0;
         $382 = $381&65535;
         $383 = $380 >>> 5;
         $384 = $383 & 31;
         $385 = $380 & 132184095;
         $386 = $382 << 16;
         $387 = $386 | $382;
         $388 = $387 & 132184095;
         $389 = (($385) - ($388))|0;
         $390 = Math_imul($389, $384)|0;
         $391 = $390 >>> 5;
         $392 = (($391) + ($388))|0;
         $393 = $392 & 132184095;
         $394 = $393 >>> 16;
         $395 = $394 | $393;
         $396 = $395&65535;
         HEAP16[$dst$066>>1] = $396;
         $397 = ((($srcbuf$1867)) + 4|0);
         $398 = ((($dst$066)) + 2|0);
         $399 = (($i$065) + 1)|0;
         $exitcond = ($399|0)==($375|0);
         if ($exitcond) {
          break;
         } else {
          $dst$066 = $398;$i$065 = $399;$srcbuf$1867 = $397;
         }
        }
        $scevgep99 = (($376) + ($379)|0);
        $400 = (($372) + ($375))|0;
        $ofs6$4 = $400;$srcbuf$19 = $scevgep99;
       }
       $401 = ($ofs6$4|0)<($1|0);
       if ($401) {
        $ofs6$3 = $ofs6$4;$srcbuf$17 = $srcbuf$19;
       } else {
        $srcbuf$19$lcssa = $srcbuf$19;
        break;
       }
      }
      $402 = HEAP32[$14>>2]|0;
      $403 = (($dstbuf$0) + ($402)|0);
      $404 = (($linecount$0) + -1)|0;
      $405 = ($404|0)==(0);
      if ($405) {
       break L30;
      } else {
       $dstbuf$0 = $403;$linecount$0 = $404;$srcbuf$13 = $srcbuf$19$lcssa;
      }
     }
    }
   }
   $83 = HEAP32[$2>>2]|0;
   $84 = ((($83)) + 9|0);
   $85 = HEAP8[$84>>0]|0;
   $86 = $85&255;
   switch ($86|0) {
   case 2:  {
    break;
   }
   case 4:  {
    $260 = ((($srcrect)) + 12|0);
    $261 = HEAP32[$260>>2]|0;
    $262 = ((($srcrect)) + 8|0);
    $263 = HEAP32[$262>>2]|0;
    $264 = (($263) + ($77))|0;
    $265 = $77 << 2;
    $266 = (0 - ($265))|0;
    $267 = (($22) + ($266)|0);
    $$14$i = $srcbuf$12$ph;$$413$i = $267;$linecount25$0$i = $261;
    while(1) {
     $$15$i = $$14$i;$ofs28$0$i = 0;
     while(1) {
      $268 = HEAP16[$$15$i>>1]|0;
      $269 = $268&65535;
      $270 = (($269) + ($ofs28$0$i))|0;
      $271 = ((($$15$i)) + 2|0);
      $272 = HEAP16[$271>>1]|0;
      $273 = $272&65535;
      $274 = ((($$15$i)) + 4|0);
      $275 = ($272<<16>>16)==(0);
      if ($275) {
       $291 = ($270|0)==(0);
       if ($291) {
        break L30;
       } else {
        $$16$i = $274;$ofs28$1$i = $270;
       }
      } else {
       $276 = (($77) - ($270))|0;
       $277 = ($276|0)>(0);
       $$31$i = $277 ? $77 : $270;
       $278 = $277 ? $276 : 0;
       $$32$i = (($273) - ($278))|0;
       $279 = (($264) - ($$31$i))|0;
       $280 = ($$32$i|0)>($279|0);
       $crun31$1$i = $280 ? $279 : $$32$i;
       $281 = ($crun31$1$i|0)>(0);
       if ($281) {
        $282 = $$31$i << 2;
        $283 = (($$413$i) + ($282)|0);
        $284 = (($$31$i) - ($270))|0;
        $285 = $284 << 2;
        $286 = (($274) + ($285)|0);
        $287 = $crun31$1$i << 2;
        (_SDL_memcpy($283,$286,$287)|0);
       }
       $288 = $273 << 2;
       $289 = (($274) + ($288)|0);
       $290 = (($270) + ($273))|0;
       $$16$i = $289;$ofs28$1$i = $290;
      }
      $292 = ($ofs28$1$i|0)<($1|0);
      if ($292) {
       $$15$i = $$16$i;$ofs28$0$i = $ofs28$1$i;
      } else {
       $$16$i$lcssa = $$16$i;
       break;
      }
     }
     $$18$i = $$16$i$lcssa;$ofs28$3$i = 0;
     while(1) {
      $293 = HEAP16[$$18$i>>1]|0;
      $294 = $293&65535;
      $295 = (($294) + ($ofs28$3$i))|0;
      $296 = ((($$18$i)) + 2|0);
      $297 = HEAP16[$296>>1]|0;
      $298 = $297&65535;
      $299 = ((($$18$i)) + 4|0);
      $300 = ($297<<16>>16)==(0);
      if ($300) {
       $$19$i = $299;$ofs28$4$i = $295;
      } else {
       $301 = (($77) - ($295))|0;
       $302 = ($301|0)>(0);
       $$33$i = $302 ? $77 : $295;
       $303 = $302 ? $301 : 0;
       $$34$i = (($298) - ($303))|0;
       $304 = (($264) - ($$33$i))|0;
       $305 = ($$34$i|0)>($304|0);
       $crun35$1$i = $305 ? $304 : $$34$i;
       $306 = ($crun35$1$i|0)>(0);
       if ($306) {
        $307 = (($$413$i) + ($$33$i<<2)|0);
        $308 = (($$33$i) - ($295))|0;
        $309 = (($299) + ($308<<2)|0);
        $i38$064$i = 0;
        while(1) {
         $310 = (($309) + ($i38$064$i<<2)|0);
         $311 = HEAP32[$310>>2]|0;
         $312 = (($307) + ($i38$064$i<<2)|0);
         $313 = HEAP32[$312>>2]|0;
         $314 = $311 >>> 24;
         $315 = $311 & 16711935;
         $316 = $313 & 16711935;
         $317 = (($315) - ($316))|0;
         $318 = Math_imul($317, $314)|0;
         $319 = $318 >>> 8;
         $320 = (($319) + ($316))|0;
         $321 = $320 & 16711935;
         $322 = $311 & 65280;
         $323 = $313 & 65280;
         $324 = (($322) - ($323))|0;
         $325 = Math_imul($324, $314)|0;
         $326 = $325 >>> 8;
         $327 = (($326) + ($323))|0;
         $328 = $327 & 65280;
         $329 = $321 | $328;
         $330 = $329 | -16777216;
         HEAP32[$312>>2] = $330;
         $331 = (($i38$064$i) + 1)|0;
         $332 = ($331|0)<($crun35$1$i|0);
         if ($332) {
          $i38$064$i = $331;
         } else {
          break;
         }
        }
       }
       $333 = $298 << 2;
       $334 = (($299) + ($333)|0);
       $335 = (($295) + ($298))|0;
       $$19$i = $334;$ofs28$4$i = $335;
      }
      $336 = ($ofs28$4$i|0)<($1|0);
      if ($336) {
       $$18$i = $$19$i;$ofs28$3$i = $ofs28$4$i;
      } else {
       $$19$i$lcssa = $$19$i;
       break;
      }
     }
     $337 = HEAP32[$14>>2]|0;
     $338 = (($$413$i) + ($337)|0);
     $339 = (($linecount25$0$i) + -1)|0;
     $340 = ($339|0)==(0);
     if ($340) {
      break L30;
     } else {
      $$14$i = $$19$i$lcssa;$$413$i = $338;$linecount25$0$i = $339;
     }
    }
    break;
   }
   default: {
    break L30;
   }
   }
   $87 = ((($83)) + 16|0);
   $88 = HEAP32[$87>>2]|0;
   $89 = ($88|0)==(2016);
   if (!($89)) {
    $90 = ((($83)) + 12|0);
    $91 = HEAP32[$90>>2]|0;
    $92 = ($91|0)==(2016);
    if (!($92)) {
     $93 = ((($83)) + 20|0);
     $94 = HEAP32[$93>>2]|0;
     $95 = ($94|0)==(2016);
     if (!($95)) {
      $178 = ((($srcrect)) + 12|0);
      $179 = HEAP32[$178>>2]|0;
      $180 = ((($srcrect)) + 8|0);
      $181 = HEAP32[$180>>2]|0;
      $182 = (($181) + ($77))|0;
      $183 = $77 << 1;
      $184 = (0 - ($183))|0;
      $185 = (($22) + ($184)|0);
      $$211$i = $185;$$7$i = $srcbuf$12$ph;$linecount6$0$i = $179;
      while(1) {
       $$8$i = $$7$i;$ofs9$0$i = 0;
       while(1) {
        $186 = HEAP8[$$8$i>>0]|0;
        $187 = $186&255;
        $188 = (($187) + ($ofs9$0$i))|0;
        $189 = ((($$8$i)) + 1|0);
        $190 = HEAP8[$189>>0]|0;
        $191 = $190&255;
        $192 = ((($$8$i)) + 2|0);
        $193 = ($190<<24>>24)==(0);
        if ($193) {
         $209 = ($188|0)==(0);
         if ($209) {
          break L30;
         } else {
          $$9$i = $192;$ofs9$1$i = $188;
         }
        } else {
         $194 = (($77) - ($188))|0;
         $195 = ($194|0)>(0);
         $196 = $195 ? $194 : 0;
         $$27$i = (($191) - ($196))|0;
         $$28$i = $195 ? $77 : $188;
         $197 = (($182) - ($$28$i))|0;
         $198 = ($$27$i|0)>($197|0);
         $crun12$1$i = $198 ? $197 : $$27$i;
         $199 = ($crun12$1$i|0)>(0);
         if ($199) {
          $200 = $$28$i << 1;
          $201 = (($$211$i) + ($200)|0);
          $202 = (($$28$i) - ($188))|0;
          $203 = $202 << 1;
          $204 = (($192) + ($203)|0);
          $205 = $crun12$1$i << 1;
          (_SDL_memcpy($201,$204,$205)|0);
         }
         $206 = $191 << 1;
         $207 = (($192) + ($206)|0);
         $208 = (($188) + ($191))|0;
         $$9$i = $207;$ofs9$1$i = $208;
        }
        $210 = ($ofs9$1$i|0)<($1|0);
        if ($210) {
         $$8$i = $$9$i;$ofs9$0$i = $ofs9$1$i;
        } else {
         $$9$i$lcssa = $$9$i;
         break;
        }
       }
       $211 = $$9$i$lcssa;
       $212 = $211 & 2;
       $213 = (($$9$i$lcssa) + ($212)|0);
       $$11$i = $213;$ofs9$3$i = 0;
       while(1) {
        $214 = HEAP16[$$11$i>>1]|0;
        $215 = $214&65535;
        $216 = (($215) + ($ofs9$3$i))|0;
        $217 = ((($$11$i)) + 2|0);
        $218 = HEAP16[$217>>1]|0;
        $219 = $218&65535;
        $220 = ((($$11$i)) + 4|0);
        $221 = ($218<<16>>16)==(0);
        if ($221) {
         $$12$i = $220;$ofs9$4$i = $216;
        } else {
         $222 = (($77) - ($216))|0;
         $223 = ($222|0)>(0);
         $224 = $223 ? $222 : 0;
         $$29$i = (($219) - ($224))|0;
         $$30$i = $223 ? $77 : $216;
         $225 = (($182) - ($$30$i))|0;
         $226 = ($$29$i|0)>($225|0);
         $crun16$1$i = $226 ? $225 : $$29$i;
         $227 = ($crun16$1$i|0)>(0);
         if ($227) {
          $228 = (($$211$i) + ($$30$i<<1)|0);
          $229 = (($$30$i) - ($216))|0;
          $230 = (($220) + ($229<<2)|0);
          $i19$062$i = 0;
          while(1) {
           $231 = (($230) + ($i19$062$i<<2)|0);
           $232 = HEAP32[$231>>2]|0;
           $233 = (($228) + ($i19$062$i<<1)|0);
           $234 = HEAP16[$233>>1]|0;
           $235 = $234&65535;
           $236 = $232 >>> 5;
           $237 = $236 & 31;
           $238 = $232 & 65043487;
           $239 = $235 << 16;
           $240 = $239 | $235;
           $241 = $240 & 65043487;
           $242 = (($238) - ($241))|0;
           $243 = Math_imul($242, $237)|0;
           $244 = $243 >>> 5;
           $245 = (($244) + ($241))|0;
           $246 = $245 & 65043487;
           $247 = $246 >>> 16;
           $248 = $247 | $246;
           $249 = $248&65535;
           HEAP16[$233>>1] = $249;
           $250 = (($i19$062$i) + 1)|0;
           $251 = ($250|0)<($crun16$1$i|0);
           if ($251) {
            $i19$062$i = $250;
           } else {
            break;
           }
          }
         }
         $252 = $219 << 2;
         $253 = (($220) + ($252)|0);
         $254 = (($216) + ($219))|0;
         $$12$i = $253;$ofs9$4$i = $254;
        }
        $255 = ($ofs9$4$i|0)<($1|0);
        if ($255) {
         $$11$i = $$12$i;$ofs9$3$i = $ofs9$4$i;
        } else {
         $$12$i$lcssa = $$12$i;
         break;
        }
       }
       $256 = HEAP32[$14>>2]|0;
       $257 = (($$211$i) + ($256)|0);
       $258 = (($linecount6$0$i) + -1)|0;
       $259 = ($258|0)==(0);
       if ($259) {
        break L30;
       } else {
        $$211$i = $257;$$7$i = $$12$i$lcssa;$linecount6$0$i = $258;
       }
      }
     }
    }
   }
   $96 = ((($srcrect)) + 12|0);
   $97 = HEAP32[$96>>2]|0;
   $98 = ((($srcrect)) + 8|0);
   $99 = HEAP32[$98>>2]|0;
   $100 = (($99) + ($77))|0;
   $101 = $77 << 1;
   $102 = (0 - ($101))|0;
   $103 = (($22) + ($102)|0);
   $$0$i = $srcbuf$12$ph;$$09$i = $103;$linecount$0$i = $97;
   while(1) {
    $$1$i = $$0$i;$ofs$0$i = 0;
    while(1) {
     $104 = HEAP8[$$1$i>>0]|0;
     $105 = $104&255;
     $106 = (($105) + ($ofs$0$i))|0;
     $107 = ((($$1$i)) + 1|0);
     $108 = HEAP8[$107>>0]|0;
     $109 = $108&255;
     $110 = ((($$1$i)) + 2|0);
     $111 = ($108<<24>>24)==(0);
     if ($111) {
      $127 = ($106|0)==(0);
      if ($127) {
       break L30;
      } else {
       $$2$i = $110;$ofs$1$i = $106;
      }
     } else {
      $112 = (($77) - ($106))|0;
      $113 = ($112|0)>(0);
      $114 = $113 ? $112 : 0;
      $$$i = (($109) - ($114))|0;
      $$24$i = $113 ? $77 : $106;
      $115 = (($100) - ($$24$i))|0;
      $116 = ($$$i|0)>($115|0);
      $crun$1$i = $116 ? $115 : $$$i;
      $117 = ($crun$1$i|0)>(0);
      if ($117) {
       $118 = $$24$i << 1;
       $119 = (($$09$i) + ($118)|0);
       $120 = (($$24$i) - ($106))|0;
       $121 = $120 << 1;
       $122 = (($110) + ($121)|0);
       $123 = $crun$1$i << 1;
       (_SDL_memcpy($119,$122,$123)|0);
      }
      $124 = $109 << 1;
      $125 = (($110) + ($124)|0);
      $126 = (($106) + ($109))|0;
      $$2$i = $125;$ofs$1$i = $126;
     }
     $128 = ($ofs$1$i|0)<($1|0);
     if ($128) {
      $$1$i = $$2$i;$ofs$0$i = $ofs$1$i;
     } else {
      $$2$i$lcssa = $$2$i;
      break;
     }
    }
    $129 = $$2$i$lcssa;
    $130 = $129 & 2;
    $131 = (($$2$i$lcssa) + ($130)|0);
    $$4$i = $131;$ofs$3$i = 0;
    while(1) {
     $132 = HEAP16[$$4$i>>1]|0;
     $133 = $132&65535;
     $134 = (($133) + ($ofs$3$i))|0;
     $135 = ((($$4$i)) + 2|0);
     $136 = HEAP16[$135>>1]|0;
     $137 = $136&65535;
     $138 = ((($$4$i)) + 4|0);
     $139 = ($136<<16>>16)==(0);
     if ($139) {
      $$5$i = $138;$ofs$4$i = $134;
     } else {
      $140 = (($77) - ($134))|0;
      $141 = ($140|0)>(0);
      $142 = $141 ? $140 : 0;
      $$25$i = (($137) - ($142))|0;
      $$26$i = $141 ? $77 : $134;
      $143 = (($100) - ($$26$i))|0;
      $144 = ($$25$i|0)>($143|0);
      $crun3$1$i = $144 ? $143 : $$25$i;
      $145 = ($crun3$1$i|0)>(0);
      if ($145) {
       $146 = (($$09$i) + ($$26$i<<1)|0);
       $147 = (($$26$i) - ($134))|0;
       $148 = (($138) + ($147<<2)|0);
       $i$061$i = 0;
       while(1) {
        $149 = (($148) + ($i$061$i<<2)|0);
        $150 = HEAP32[$149>>2]|0;
        $151 = (($146) + ($i$061$i<<1)|0);
        $152 = HEAP16[$151>>1]|0;
        $153 = $152&65535;
        $154 = $150 >>> 5;
        $155 = $154 & 31;
        $156 = $150 & 132184095;
        $157 = $153 << 16;
        $158 = $157 | $153;
        $159 = $158 & 132184095;
        $160 = (($156) - ($159))|0;
        $161 = Math_imul($160, $155)|0;
        $162 = $161 >>> 5;
        $163 = (($162) + ($159))|0;
        $164 = $163 & 132184095;
        $165 = $164 >>> 16;
        $166 = $165 | $164;
        $167 = $166&65535;
        HEAP16[$151>>1] = $167;
        $168 = (($i$061$i) + 1)|0;
        $169 = ($168|0)<($crun3$1$i|0);
        if ($169) {
         $i$061$i = $168;
        } else {
         break;
        }
       }
      }
      $170 = $137 << 2;
      $171 = (($138) + ($170)|0);
      $172 = (($134) + ($137))|0;
      $$5$i = $171;$ofs$4$i = $172;
     }
     $173 = ($ofs$4$i|0)<($1|0);
     if ($173) {
      $$4$i = $$5$i;$ofs$3$i = $ofs$4$i;
     } else {
      $$5$i$lcssa = $$5$i;
      break;
     }
    }
    $174 = HEAP32[$14>>2]|0;
    $175 = (($$09$i) + ($174)|0);
    $176 = (($linecount$0$i) + -1)|0;
    $177 = ($176|0)==(0);
    if ($177) {
     break;
    } else {
     $$0$i = $$5$i$lcssa;$$09$i = $175;$linecount$0$i = $176;
    }
   }
  }
 } while(0);
 $517 = HEAP32[$surf_dst>>2]|0;
 $518 = $517 & 2;
 $519 = ($518|0)==(0);
 if ($519) {
  $$0 = 0;
  return ($$0|0);
 }
 _SDL_UnlockSurface($surf_dst);
 $$0 = 0;
 return ($$0|0);
}
function _SDL_SetPixelFormatPalette($format,$palette) {
 $format = $format|0;
 $palette = $palette|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($format|0)==(0|0);
 if ($0) {
  $1 = (_SDL_SetError(597461,$vararg_buffer)|0);
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($palette|0)!=(0|0);
 if ($2) {
  $3 = HEAP32[$palette>>2]|0;
  $4 = ((($format)) + 8|0);
  $5 = HEAP8[$4>>0]|0;
  $6 = $5&255;
  $7 = 1 << $6;
  $8 = ($3|0)==($7|0);
  if (!($8)) {
   $9 = (_SDL_SetError(597508,$vararg_buffer1)|0);
   $$0 = $9;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $10 = ((($format)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==($palette|0);
 if ($12) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $13 = ($11|0)==(0|0);
 if (!($13)) {
  $14 = ((($11)) + 12|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = (($15) + -1)|0;
  HEAP32[$14>>2] = $16;
  $17 = ($15|0)>(1);
  if (!($17)) {
   $18 = ((($11)) + 4|0);
   $19 = HEAP32[$18>>2]|0;
   _SDL_free($19);
   _SDL_free($11);
  }
 }
 HEAP32[$10>>2] = $palette;
 if (!($2)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $20 = ((($palette)) + 12|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = (($21) + 1)|0;
 HEAP32[$20>>2] = $22;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_InvalidateMap($map) {
 $map = $map|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($map|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = HEAP32[$map>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = ((($1)) + 56|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = (($4) + -1)|0;
  HEAP32[$3>>2] = $5;
  $6 = ($4|0)<(2);
  if ($6) {
   _SDL_FreeSurface($1);
  }
 }
 HEAP32[$map>>2] = 0;
 $7 = ((($map)) + 84|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($map)) + 80|0);
 HEAP32[$8>>2] = 0;
 $9 = ((($map)) + 64|0);
 $10 = HEAP32[$9>>2]|0;
 _SDL_free($10);
 HEAP32[$9>>2] = 0;
 return;
}
function _SDL_FreeFormat($format) {
 $format = $format|0;
 var $$lcssa6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $prev$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($format|0)==(0|0);
 if ($0) {
  HEAP32[$vararg_buffer>>2] = 597583;
  (_SDL_SetError(611267,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $1 = ((($format)) + 36|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($2) + -1)|0;
 HEAP32[$1>>2] = $3;
 $4 = ($2|0)>(1);
 if ($4) {
  STACKTOP = sp;return;
 }
 $5 = HEAP32[300737]|0;
 $6 = ($5|0)==($format|0);
 L8: do {
  if ($6) {
   $7 = ((($format)) + 40|0);
   $8 = HEAP32[$7>>2]|0;
   HEAP32[300737] = $8;
  } else {
   $9 = ($5|0)==(0|0);
   if (!($9)) {
    $prev$0 = $5;
    while(1) {
     $10 = ((($prev$0)) + 40|0);
     $11 = HEAP32[$10>>2]|0;
     $12 = ($11|0)==(0|0);
     if ($12) {
      break L8;
     }
     $13 = ($11|0)==($format|0);
     if ($13) {
      $$lcssa6 = $10;
      break;
     } else {
      $prev$0 = $11;
     }
    }
    $14 = ((($format)) + 40|0);
    $15 = HEAP32[$14>>2]|0;
    HEAP32[$$lcssa6>>2] = $15;
   }
  }
 } while(0);
 $16 = ((($format)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ($17|0)==(0|0);
 if (!($18)) {
  $19 = ((($17)) + 12|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = (($20) + -1)|0;
  HEAP32[$19>>2] = $21;
  $22 = ($20|0)>(1);
  if (!($22)) {
   $23 = ((($17)) + 4|0);
   $24 = HEAP32[$23>>2]|0;
   _SDL_free($24);
   _SDL_free($17);
  }
 }
 _SDL_free($format);
 STACKTOP = sp;return;
}
function _SDL_FreeBlitMap($map) {
 $map = $map|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($map|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = HEAP32[$map>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = ((($1)) + 56|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = (($4) + -1)|0;
  HEAP32[$3>>2] = $5;
  $6 = ($4|0)<(2);
  if ($6) {
   _SDL_FreeSurface($1);
  }
 }
 HEAP32[$map>>2] = 0;
 $7 = ((($map)) + 84|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($map)) + 80|0);
 HEAP32[$8>>2] = 0;
 $9 = ((($map)) + 64|0);
 $10 = HEAP32[$9>>2]|0;
 _SDL_free($10);
 HEAP32[$9>>2] = 0;
 _SDL_free($map);
 return;
}
function _SDL_FinishWindowCreation($window,$flags) {
 $window = $window|0;
 $flags = $flags|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $9 = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer11 = 0, $vararg_buffer13 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, $vararg_buffer9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer13 = sp + 56|0;
 $vararg_buffer11 = sp + 48|0;
 $vararg_buffer9 = sp + 40|0;
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($window)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($window)) + 56|0);
 HEAP32[$2>>2] = $1;
 $3 = ((($window)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($window)) + 60|0);
 HEAP32[$5>>2] = $4;
 $6 = ((($window)) + 24|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($window)) + 64|0);
 HEAP32[$8>>2] = $7;
 $9 = ((($window)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($window)) + 68|0);
 HEAP32[$11>>2] = $10;
 $12 = $flags & 128;
 $13 = ($12|0)==(0);
 do {
  if (!($13)) {
   $14 = HEAP32[300441]|0;
   $15 = ($14|0)==(0|0);
   if ($15) {
    (_SDL_SetError(595917,$vararg_buffer)|0);
    break;
   }
   $16 = HEAP32[$window>>2]|0;
   $17 = ((($14)) + 256|0);
   $18 = ($16|0)==($17|0);
   if (!($18)) {
    (_SDL_SetError(609414,$vararg_buffer1)|0);
    break;
   }
   $19 = ((($window)) + 48|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = $20 & 128;
   $22 = ($21|0)==(0);
   if ($22) {
    $23 = ((($14)) + 72|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = ($24|0)==(0|0);
    if (!($25)) {
     FUNCTION_TABLE_vii[$24 & 8191]($14,$window);
    }
   }
  }
 } while(0);
 $26 = $flags & 64;
 $27 = ($26|0)==(0);
 do {
  if (!($27)) {
   $28 = HEAP32[300441]|0;
   $29 = ($28|0)==(0|0);
   if ($29) {
    (_SDL_SetError(595917,$vararg_buffer3)|0);
    break;
   }
   $30 = HEAP32[$window>>2]|0;
   $31 = ((($28)) + 256|0);
   $32 = ($30|0)==($31|0);
   if (!($32)) {
    (_SDL_SetError(609414,$vararg_buffer5)|0);
    break;
   }
   $33 = ((($window)) + 48|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = $34 & 64;
   $36 = ($35|0)==(0);
   if ($36) {
    (_SDL_UpdateFullscreenMode($window,0)|0);
    $37 = HEAP32[300441]|0;
    $38 = ((($37)) + 76|0);
    $39 = HEAP32[$38>>2]|0;
    $40 = ($39|0)==(0|0);
    if (!($40)) {
     FUNCTION_TABLE_vii[$39 & 8191]($37,$window);
    }
   }
  }
 } while(0);
 $41 = $flags & 1;
 $42 = ($41|0)==(0);
 do {
  if (!($42)) {
   $43 = HEAP32[300441]|0;
   $44 = ($43|0)==(0|0);
   if ($44) {
    (_SDL_SetError(595917,$vararg_buffer7)|0);
    break;
   }
   $45 = HEAP32[$window>>2]|0;
   $46 = ((($43)) + 256|0);
   $47 = ($45|0)==($46|0);
   if (!($47)) {
    (_SDL_SetError(609414,$vararg_buffer9)|0);
    break;
   }
   $48 = $flags & 4097;
   $49 = ((($window)) + 48|0);
   $50 = HEAP32[$49>>2]|0;
   $51 = $50 & 4097;
   $52 = ($48|0)==($51|0);
   if (!($52)) {
    $53 = $50 & -4098;
    $54 = $53 | $48;
    HEAP32[$49>>2] = $54;
    $55 = $50 & 68;
    $56 = ($55|0)==(4);
    $57 = $56&1;
    $58 = (_SDL_UpdateFullscreenMode($window,$57)|0);
    $59 = ($58|0)==(0);
    if (!($59)) {
     $60 = HEAP32[$49>>2]|0;
     $61 = $60 & -4098;
     $62 = $61 | $51;
     HEAP32[$49>>2] = $62;
    }
   }
  }
 } while(0);
 $63 = $flags & 256;
 $64 = ($63|0)==(0);
 if (!($64)) {
  _SDL_SetWindowGrab($window,1);
 }
 $65 = $flags & 8;
 $66 = ($65|0)==(0);
 if (!($66)) {
  STACKTOP = sp;return;
 }
 $67 = HEAP32[300441]|0;
 $68 = ($67|0)==(0|0);
 if ($68) {
  (_SDL_SetError(595917,$vararg_buffer11)|0);
  STACKTOP = sp;return;
 }
 $69 = ($window|0)==(0|0);
 if (!($69)) {
  $70 = HEAP32[$window>>2]|0;
  $71 = ((($67)) + 256|0);
  $72 = ($70|0)==($71|0);
  if ($72) {
   $73 = ((($window)) + 48|0);
   $74 = HEAP32[$73>>2]|0;
   $75 = $74 & 4;
   $76 = ($75|0)==(0);
   if (!($76)) {
    STACKTOP = sp;return;
   }
   $77 = ((($67)) + 60|0);
   $78 = HEAP32[$77>>2]|0;
   $79 = ($78|0)==(0|0);
   if (!($79)) {
    FUNCTION_TABLE_vii[$78 & 8191]($67,$window);
   }
   (_SDL_SendWindowEvent($window,1,0,0)|0);
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(609414,$vararg_buffer13)|0);
 STACKTOP = sp;return;
}
function _SDL_SetWindowGrab($window,$grabbed) {
 $window = $window|0;
 $grabbed = $grabbed|0;
 var $$pre3$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $grabbed$01$i = 0, $or$cond$i = 0, $or$cond7$i = 0, $storemerge = 0, $tmp = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ($grabbed|0)!=(0);
   $7 = ((($window)) + 48|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = $8 & 256;
   $10 = ($9|0)!=(0);
   $tmp = $6 ^ $10;
   if (!($tmp)) {
    STACKTOP = sp;return;
   }
   $11 = $8 & -257;
   $12 = $8 | 256;
   $storemerge = $6 ? $12 : $11;
   HEAP32[$7>>2] = $storemerge;
   $13 = (_SDL_GetMouse()|0);
   $14 = ((($13)) + 76|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = HEAP32[$7>>2]|0;
   $17 = $16 & 256;
   $18 = $17 | $15;
   $19 = ($18|0)==(0);
   $20 = $16 & 512;
   $21 = ($20|0)==(0);
   $or$cond7$i = $21 | $19;
   $22 = HEAP32[300441]|0;
   $23 = ((($22)) + 252|0);
   $24 = HEAP32[$23>>2]|0;
   if ($or$cond7$i) {
    $35 = ($24|0)==($window|0);
    if ($35) {
     HEAP32[$23>>2] = 0;
     $37 = $22;$grabbed$01$i = 0;
    } else {
     $37 = $22;$grabbed$01$i = 0;
    }
   } else {
    $25 = ($24|0)==(0|0);
    $26 = ($24|0)==($window|0);
    $or$cond$i = $25 | $26;
    if ($or$cond$i) {
     $34 = $22;
    } else {
     $27 = ((($24)) + 48|0);
     $28 = HEAP32[$27>>2]|0;
     $29 = $28 & -257;
     HEAP32[$27>>2] = $29;
     $30 = ((($22)) + 100|0);
     $31 = HEAP32[$30>>2]|0;
     $32 = ($31|0)==(0|0);
     if ($32) {
      $34 = $22;
     } else {
      FUNCTION_TABLE_viii[$31 & 8191]($22,$24,0);
      $$pre3$i = HEAP32[300441]|0;
      $34 = $$pre3$i;
     }
    }
    $33 = ((($34)) + 252|0);
    HEAP32[$33>>2] = $window;
    $37 = $34;$grabbed$01$i = 1;
   }
   $36 = ((($37)) + 100|0);
   $38 = HEAP32[$36>>2]|0;
   $39 = ($38|0)==(0|0);
   if ($39) {
    STACKTOP = sp;return;
   }
   FUNCTION_TABLE_viii[$38 & 8191]($37,$window,$grabbed$01$i);
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 STACKTOP = sp;return;
}
function _GLES2_DestroyRenderer($renderer) {
 $renderer = $renderer|0;
 var $$$021$i = 0, $$0$i$i = 0, $$021$i = 0, $$cast = 0, $$pre = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $7 = 0, $8 = 0, $9 = 0, $entry$05 = 0, $entry1$01 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0;
 var $vararg_ptr4 = 0, $vararg_ptr5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  _SDL_free($renderer);
  STACKTOP = sp;return;
 }
 $3 = HEAP32[300738]|0;
 $4 = HEAP32[$1>>2]|0;
 $5 = ($3|0)==($4|0);
 if ($5) {
  $13 = $1;
  label = 5;
 } else {
  $6 = ((($1)) + 268|0);
  HEAP32[$6>>2] = 0;
  $7 = ((($renderer)) + 188|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = (_SDL_GL_MakeCurrent($8,$4)|0);
  $10 = ($9|0)<(0);
  if (!($10)) {
   $11 = HEAP32[$1>>2]|0;
   HEAP32[300738] = $11;
   (_GLES2_UpdateViewport($renderer)|0);
   $$pre$i = HEAP32[$0>>2]|0;
   $13 = $$pre$i;
   label = 5;
  }
 }
 if ((label|0) == 5) {
  $12 = ((($13)) + 4|0);
  $14 = HEAP32[$12>>2]|0;
  $15 = ($14|0)==(0);
  if (!($15)) {
   $16 = ((($13)) + 108|0);
   while(1) {
    $17 = HEAP32[$16>>2]|0;
    $18 = (FUNCTION_TABLE_i[$17 & 8191]()|0);
    $19 = ($18|0)==(0);
    if ($19) {
     break;
    }
   }
  }
 }
 $20 = ((($1)) + 252|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ($21|0)==(0|0);
 if (!($22)) {
  $23 = ((($1)) + 60|0);
  $entry$05 = $21;
  while(1) {
   $24 = HEAP32[$23>>2]|0;
   $25 = HEAP32[$entry$05>>2]|0;
   FUNCTION_TABLE_vi[$24 & 8191]($25);
   $26 = ((($entry$05)) + 24|0);
   $27 = HEAP32[$26>>2]|0;
   _SDL_free($entry$05);
   $28 = ($27|0)==(0|0);
   if ($28) {
    break;
   } else {
    $entry$05 = $27;
   }
  }
 }
 $29 = ((($1)) + 260|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ($30|0)==(0|0);
 if (!($31)) {
  $32 = ((($1)) + 56|0);
  $entry1$01 = $30;
  while(1) {
   $33 = HEAP32[$32>>2]|0;
   $34 = HEAP32[$entry1$01>>2]|0;
   FUNCTION_TABLE_vi[$33 & 8191]($34);
   $35 = ((($entry1$01)) + 156|0);
   $36 = HEAP32[$35>>2]|0;
   _SDL_free($entry1$01);
   $37 = ($36|0)==(0|0);
   if ($37) {
    break;
   } else {
    $entry1$01 = $36;
   }
  }
 }
 $38 = HEAP32[$1>>2]|0;
 $39 = ($38|0)==(0|0);
 if (!($39)) {
  $40 = ((($1)) + 232|0);
  $41 = HEAP32[$40>>2]|0;
  $42 = ($41|0)==(0|0);
  if ($42) {
   $65 = $38;
  } else {
   $43 = ((($1)) + 204|0);
   $45 = $41;
   while(1) {
    $44 = ((($45)) + 12|0);
    $46 = HEAP32[$44>>2]|0;
    $47 = HEAP32[$43>>2]|0;
    $48 = ((($45)) + 8|0);
    FUNCTION_TABLE_vii[$47 & 8191](1,$48);
    $49 = HEAP32[$0>>2]|0;
    $50 = ((($49)) + 4|0);
    $51 = HEAP32[$50>>2]|0;
    $52 = ($51|0)==(0);
    L29: do {
     if (!($52)) {
      $53 = ((($49)) + 108|0);
      $54 = HEAP32[$53>>2]|0;
      $55 = (FUNCTION_TABLE_i[$54 & 8191]()|0);
      $56 = ($55|0)==(0);
      if (!($56)) {
       $$021$i = 1282466;$59 = $55;
       while(1) {
        $57 = HEAP8[$$021$i>>0]|0;
        $58 = ($57<<24>>24)==(0);
        $$$021$i = $58 ? 597716 : $$021$i;
        switch ($59|0) {
        case 1280:  {
         $$0$i$i = 597787;
         break;
        }
        case 1281:  {
         $$0$i$i = 597770;
         break;
        }
        case 1282:  {
         $$0$i$i = 597749;
         break;
        }
        case 1285:  {
         $$0$i$i = 597732;
         break;
        }
        default: {
         $$0$i$i = 597724;
        }
        }
        HEAP32[$vararg_buffer>>2] = $$$021$i;
        $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
        HEAP32[$vararg_ptr1>>2] = 597829;
        $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
        HEAP32[$vararg_ptr2>>2] = 464;
        $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
        HEAP32[$vararg_ptr3>>2] = 597963;
        $vararg_ptr4 = ((($vararg_buffer)) + 16|0);
        HEAP32[$vararg_ptr4>>2] = $$0$i$i;
        $vararg_ptr5 = ((($vararg_buffer)) + 20|0);
        HEAP32[$vararg_ptr5>>2] = $59;
        (_SDL_SetError(597803,$vararg_buffer)|0);
        $60 = HEAP32[$53>>2]|0;
        $61 = (FUNCTION_TABLE_i[$60 & 8191]()|0);
        $62 = ($61|0)==(0);
        if ($62) {
         break L29;
        } else {
         $$021$i = $$$021$i;$59 = $61;
        }
       }
      }
     }
    } while(0);
    $63 = HEAP32[$40>>2]|0;
    _SDL_free($63);
    HEAP32[$40>>2] = $46;
    $$cast = $46;
    $64 = ($46|0)==(0);
    if ($64) {
     break;
    } else {
     $45 = $$cast;
    }
   }
   $$pre = HEAP32[$1>>2]|0;
   $65 = $$pre;
  }
  _SDL_GL_DeleteContext($65);
 }
 $66 = ((($1)) + 244|0);
 $67 = HEAP32[$66>>2]|0;
 _SDL_free($67);
 _SDL_free($1);
 _SDL_free($renderer);
 STACKTOP = sp;return;
}
function _SDL_GL_MakeCurrent($window,$ctx) {
 $window = $window|0;
 $ctx = $ctx|0;
 var $$0 = 0, $$0$i = 0, $$0$i2 = 0, $$01 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0$i = 0;
 } else {
  $2 = ((($0)) + 636|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = (_SDL_TLSGet($3)|0);
  $$0$i = $4;
 }
 $5 = ($$0$i|0)==($window|0);
 if ($5) {
  $6 = HEAP32[300441]|0;
  $7 = ($6|0)==(0|0);
  if ($7) {
   (_SDL_SetError(595917,$vararg_buffer1)|0);
   $$0$i2 = 0;
  } else {
   $8 = ((($6)) + 640|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (_SDL_TLSGet($9)|0);
   $$0$i2 = $10;
  }
  $11 = ($$0$i2|0)==($ctx|0);
  if ($11) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $12 = ($ctx|0)==(0|0);
 $$pre = HEAP32[300441]|0;
 do {
  if ($12) {
   $$01 = 0;
  } else {
   $13 = ($$pre|0)==(0|0);
   if ($13) {
    (_SDL_SetError(595917,$vararg_buffer3)|0);
    $$0 = -1;
    STACKTOP = sp;return ($$0|0);
   }
   $14 = ($window|0)==(0|0);
   if (!($14)) {
    $15 = HEAP32[$window>>2]|0;
    $16 = ((($$pre)) + 256|0);
    $17 = ($15|0)==($16|0);
    if ($17) {
     $18 = ((($window)) + 48|0);
     $19 = HEAP32[$18>>2]|0;
     $20 = $19 & 2;
     $21 = ($20|0)==(0);
     if (!($21)) {
      $$01 = $window;
      break;
     }
     $22 = (_SDL_SetError(597590,$vararg_buffer7)|0);
     $$0 = $22;
     STACKTOP = sp;return ($$0|0);
    }
   }
   (_SDL_SetError(609414,$vararg_buffer5)|0);
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 $23 = ((($$pre)) + 156|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = (FUNCTION_TABLE_iiii[$24 & 8191]($$pre,$$01,$ctx)|0);
 $26 = ($25|0)==(0);
 if (!($26)) {
  $$0 = $25;
  STACKTOP = sp;return ($$0|0);
 }
 $27 = HEAP32[300441]|0;
 $28 = ((($27)) + 628|0);
 HEAP32[$28>>2] = $$01;
 $29 = ((($27)) + 632|0);
 HEAP32[$29>>2] = $ctx;
 $30 = ((($27)) + 636|0);
 $31 = HEAP32[$30>>2]|0;
 (_SDL_TLSSet($31,$$01,0)|0);
 $32 = HEAP32[300441]|0;
 $33 = ((($32)) + 640|0);
 $34 = HEAP32[$33>>2]|0;
 (_SDL_TLSSet($34,$ctx,0)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_TLSGet($id) {
 $id = $id|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_SYS_GetTLSData()|0);
 $1 = ($0|0)==(0|0);
 $2 = ($id|0)==(0);
 $or$cond = $2 | $1;
 if ($or$cond) {
  $$0 = 0;
  return ($$0|0);
 }
 $3 = HEAP32[$0>>2]|0;
 $4 = ($3>>>0)<($id>>>0);
 if ($4) {
  $$0 = 0;
  return ($$0|0);
 }
 $5 = (($id) + -1)|0;
 $6 = (((($0)) + 4|0) + ($5<<3)|0);
 $7 = HEAP32[$6>>2]|0;
 $$0 = $7;
 return ($$0|0);
}
function _SDL_SYS_GetTLSData() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_Generic_GetTLSData()|0);
 return ($0|0);
}
function _SDL_Generic_GetTLSData() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $_tmp = 0, $entry$0 = 0, $entry$02 = 0, $entry$03 = 0, $entry$03$lcssa = 0, $storage$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $_tmp = sp;
 $0 = (_SDL_ThreadID()|0);
 HEAP32[$_tmp>>2] = 0;
 _SDL_AtomicLock($_tmp);
 _SDL_AtomicUnlock($_tmp);
 $entry$02 = HEAP32[300739]|0;
 $1 = ($entry$02|0)==(0|0);
 if ($1) {
  $storage$0 = 0;
  STACKTOP = sp;return ($storage$0|0);
 } else {
  $entry$03 = $entry$02;
 }
 while(1) {
  $2 = HEAP32[$entry$03>>2]|0;
  $3 = ($2|0)==($0|0);
  if ($3) {
   $entry$03$lcssa = $entry$03;
   break;
  }
  $6 = ((($entry$03)) + 8|0);
  $entry$0 = HEAP32[$6>>2]|0;
  $7 = ($entry$0|0)==(0|0);
  if ($7) {
   $storage$0 = 0;
   label = 5;
   break;
  } else {
   $entry$03 = $entry$0;
  }
 }
 if ((label|0) == 5) {
  STACKTOP = sp;return ($storage$0|0);
 }
 $4 = ((($entry$03$lcssa)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $storage$0 = $5;
 STACKTOP = sp;return ($storage$0|0);
}
function _SDL_ThreadID() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _SDL_AtomicLock($lock) {
 $lock = $lock|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$lock>>2]|0;HEAP32[$lock>>2] = 1;
 $1 = ($0|0)==(0);
 if ($1) {
  return;
 }
 while(1) {
  _SDL_Delay(0);
  $2 = HEAP32[$lock>>2]|0;HEAP32[$lock>>2] = 1;
  $3 = ($2|0)==(0);
  if ($3) {
   break;
  }
 }
 return;
}
function _SDL_Delay($ms) {
 $ms = $ms|0;
 var $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $elapsed = 0, $tv = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $elapsed = sp + 8|0;
 $tv = sp;
 $0 = (($ms>>>0) / 1000)&-1;
 HEAP32[$elapsed>>2] = $0;
 $1 = (($ms>>>0) % 1000)&-1;
 $2 = ($1*1000000)|0;
 $3 = ((($elapsed)) + 4|0);
 HEAP32[$3>>2] = $2;
 $4 = (___errno_location()|0);
 $5 = ((($tv)) + 4|0);
 HEAP32[$4>>2] = 0;
 HEAP32[$tv>>2] = $0;
 HEAP32[$5>>2] = $2;
 $6 = (_nanosleep(($tv|0),($elapsed|0))|0);
 $7 = ($6|0)==(0);
 if ($7) {
  STACKTOP = sp;return;
 }
 while(1) {
  $8 = HEAP32[$4>>2]|0;
  $9 = ($8|0)==(4);
  if (!($9)) {
   label = 4;
   break;
  }
  $$pre = HEAP32[$elapsed>>2]|0;
  $$pre1 = HEAP32[$3>>2]|0;
  HEAP32[$4>>2] = 0;
  HEAP32[$tv>>2] = $$pre;
  HEAP32[$5>>2] = $$pre1;
  $10 = (_nanosleep(($tv|0),($elapsed|0))|0);
  $11 = ($10|0)==(0);
  if ($11) {
   label = 4;
   break;
  }
 }
 if ((label|0) == 4) {
  STACKTOP = sp;return;
 }
}
function _SDL_AtomicUnlock($lock) {
 $lock = $lock|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$lock>>2] = 0;
 return;
}
function _SDL_TLSSet($id,$value,$destructor) {
 $id = $id|0;
 $value = $value|0;
 $destructor = $destructor|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$03 = 0, $storage$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($id|0)==(0);
 if ($0) {
  HEAP32[$vararg_buffer>>2] = 597634;
  $1 = (_SDL_SetError(611267,$vararg_buffer)|0);
  $$1 = $1;
  STACKTOP = sp;return ($$1|0);
 }
 $2 = (_SDL_SYS_GetTLSData()|0);
 $3 = ($2|0)==(0|0);
 if ($3) {
  $13 = 0;
  label = 5;
 } else {
  $4 = HEAP32[$2>>2]|0;
  $5 = ($4>>>0)<($id>>>0);
  if ($5) {
   $13 = $4;
   label = 5;
  } else {
   $storage$0 = $2;
  }
 }
 if ((label|0) == 5) {
  $6 = (($id) + 4)|0;
  $7 = $id << 3;
  $8 = (($7) + 36)|0;
  $9 = (_SDL_realloc($2,$8)|0);
  $10 = ($9|0)==(0|0);
  if ($10) {
   $11 = (_SDL_Error(0)|0);
   $$1 = $11;
   STACKTOP = sp;return ($$1|0);
  }
  HEAP32[$9>>2] = $6;
  $12 = ($13>>>0)<($6>>>0);
  if ($12) {
   $14 = ((($9)) + 4|0);
   $i$03 = $13;
   while(1) {
    $15 = (($14) + ($i$03<<3)|0);
    HEAP32[$15>>2] = 0;
    $16 = (((($14) + ($i$03<<3)|0)) + 4|0);
    HEAP32[$16>>2] = 0;
    $17 = (($i$03) + 1)|0;
    $exitcond = ($17|0)==($6|0);
    if ($exitcond) {
     break;
    } else {
     $i$03 = $17;
    }
   }
  }
  $18 = (_SDL_SYS_SetTLSData($9)|0);
  $19 = ($18|0)==(0);
  if ($19) {
   $storage$0 = $9;
  } else {
   $$1 = -1;
   STACKTOP = sp;return ($$1|0);
  }
 }
 $20 = (($id) + -1)|0;
 $21 = (((($storage$0)) + 4|0) + ($20<<3)|0);
 HEAP32[$21>>2] = $value;
 $22 = (((((($storage$0)) + 4|0) + ($20<<3)|0)) + 4|0);
 HEAP32[$22>>2] = $destructor;
 $$1 = 0;
 STACKTOP = sp;return ($$1|0);
}
function _SDL_SYS_SetTLSData($data) {
 $data = $data|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_Generic_SetTLSData($data)|0);
 return ($0|0);
}
function _SDL_Generic_SetTLSData($storage) {
 $storage = $storage|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $entry$0 = 0;
 var $entry$03 = 0, $entry$05 = 0, $entry$05$lcssa = 0, $prev$04 = 0, $prev$04$lcssa = 0, $prev$04$phi = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_ThreadID()|0);
 $entry$03 = HEAP32[300739]|0;
 $1 = ($entry$03|0)==(0|0);
 L1: do {
  if ($1) {
   label = 10;
  } else {
   $entry$05 = $entry$03;$prev$04 = 0;
   while(1) {
    $2 = HEAP32[$entry$05>>2]|0;
    $3 = ($2|0)==($0|0);
    if ($3) {
     $entry$05$lcssa = $entry$05;$prev$04$lcssa = $prev$04;
     break;
    }
    $10 = ((($entry$05)) + 8|0);
    $entry$0 = HEAP32[$10>>2]|0;
    $11 = ($entry$0|0)==(0|0);
    if ($11) {
     label = 10;
     break L1;
    } else {
     $prev$04$phi = $entry$05;$entry$05 = $entry$0;$prev$04 = $prev$04$phi;
    }
   }
   $4 = ($storage|0)==(0|0);
   if (!($4)) {
    $5 = ((($entry$05$lcssa)) + 4|0);
    HEAP32[$5>>2] = $storage;
    break;
   }
   $6 = ($prev$04$lcssa|0)==(0|0);
   $7 = ((($entry$05$lcssa)) + 8|0);
   $8 = HEAP32[$7>>2]|0;
   if ($6) {
    HEAP32[300739] = $8;
   } else {
    $9 = ((($prev$04$lcssa)) + 8|0);
    HEAP32[$9>>2] = $8;
   }
   _SDL_free($entry$05$lcssa);
  }
 } while(0);
 do {
  if ((label|0) == 10) {
   $12 = (_SDL_malloc(12)|0);
   $13 = ($12|0)==(0|0);
   if (!($13)) {
    HEAP32[$12>>2] = $0;
    $14 = ((($12)) + 4|0);
    HEAP32[$14>>2] = $storage;
    $15 = HEAP32[300739]|0;
    $16 = ((($12)) + 8|0);
    HEAP32[$16>>2] = $15;
    HEAP32[300739] = $12;
    break;
   }
   $17 = (_SDL_Error(0)|0);
   $$0 = $17;
   return ($$0|0);
  }
 } while(0);
 $$0 = 0;
 return ($$0|0);
}
function _GLES2_UpdateViewport($renderer) {
 $renderer = $renderer|0;
 var $$$021$i = 0, $$$i = 0.0, $$0 = 0, $$0$i$i = 0, $$021$i = 0, $$1$i = 0.0, $$sink$i = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0, $45 = 0, $46 = 0.0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $h = 0, $projection$i = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0, $vararg_ptr4 = 0, $vararg_ptr5 = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $projection$i = sp + 32|0;
 $w = sp + 28|0;
 $h = sp + 24|0;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[300738]|0;
 $3 = HEAP32[$1>>2]|0;
 $4 = ($2|0)==($3|0);
 if (!($4)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = ((($renderer)) + 304|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0|0);
 if ($7) {
  (_SDL_GetRendererOutputSize($renderer,$w,$h)|0);
  $18 = ((($1)) + 188|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = ((($renderer)) + 212|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = HEAP32[$h>>2]|0;
  $23 = ((($renderer)) + 216|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = (($22) - ($24))|0;
  $26 = ((($renderer)) + 224|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = (($25) - ($27))|0;
  $29 = ((($renderer)) + 220|0);
  $30 = HEAP32[$29>>2]|0;
  FUNCTION_TABLE_viiii[$19 & 8191]($21,$28,$30,$27);
 } else {
  $8 = ((($1)) + 188|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($renderer)) + 212|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ((($renderer)) + 216|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ((($renderer)) + 220|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ((($renderer)) + 224|0);
  $17 = HEAP32[$16>>2]|0;
  FUNCTION_TABLE_viiii[$9 & 8191]($11,$13,$15,$17);
 }
 $31 = ((($1)) + 268|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ($32|0)==(0|0);
 if (!($33)) {
  $34 = HEAP32[$0>>2]|0;
  $35 = ((($renderer)) + 220|0);
  $36 = HEAP32[$35>>2]|0;
  $37 = ($36|0)==(0);
  if (!($37)) {
   $38 = ((($renderer)) + 224|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = ($39|0)==(0);
   if (!($40)) {
    $41 = (+($36|0));
    $42 = 2.0 / $41;
    HEAPF32[$projection$i>>2] = $42;
    $43 = ((($projection$i)) + 4|0);
    ;HEAP32[$43>>2]=0|0;HEAP32[$43+4>>2]=0|0;HEAP32[$43+8>>2]=0|0;HEAP32[$43+12>>2]=0|0;
    $44 = HEAP32[$5>>2]|0;
    $45 = ($44|0)==(0|0);
    $46 = (+($39|0));
    $$1$i = $45 ? -2.0 : 2.0;
    $$sink$i = $$1$i / $46;
    $47 = ((($projection$i)) + 20|0);
    HEAPF32[$47>>2] = $$sink$i;
    $48 = ((($projection$i)) + 24|0);
    $49 = ((($projection$i)) + 48|0);
    ;HEAP32[$48>>2]=0|0;HEAP32[$48+4>>2]=0|0;HEAP32[$48+8>>2]=0|0;HEAP32[$48+12>>2]=0|0;HEAP32[$48+16>>2]=0|0;HEAP32[$48+20>>2]=0|0;
    HEAPF32[$49>>2] = -1.0;
    $50 = ((($projection$i)) + 52|0);
    $$$i = $45 ? 1.0 : -1.0;
    HEAPF32[$50>>2] = $$$i;
    $51 = ((($projection$i)) + 56|0);
    HEAPF32[$51>>2] = 0.0;
    $52 = ((($projection$i)) + 60|0);
    HEAPF32[$52>>2] = 1.0;
    $53 = ((($34)) + 268|0);
    $54 = HEAP32[$53>>2]|0;
    $55 = ((($54)) + 88|0);
    $56 = (_SDL_memcmp($55,$projection$i,64)|0);
    $57 = ($56|0)==(0);
    if (!($57)) {
     $58 = HEAP32[$53>>2]|0;
     $59 = ((($58)) + 16|0);
     $60 = HEAP32[$59>>2]|0;
     $61 = ((($34)) + 176|0);
     $62 = HEAP32[$61>>2]|0;
     FUNCTION_TABLE_viiii[$62 & 8191]($60,1,0,$projection$i);
     $63 = HEAP32[$53>>2]|0;
     $64 = ((($63)) + 88|0);
     (_SDL_memcpy($64,$projection$i,64)|0);
    }
   }
  }
 }
 $65 = HEAP32[$0>>2]|0;
 $66 = ((($65)) + 4|0);
 $67 = HEAP32[$66>>2]|0;
 $68 = ($67|0)==(0);
 if ($68) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $69 = ((($65)) + 108|0);
 $70 = HEAP32[$69>>2]|0;
 $71 = (FUNCTION_TABLE_i[$70 & 8191]()|0);
 $72 = ($71|0)==(0);
 if ($72) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $$021$i = 1282466;$75 = $71;
 }
 while(1) {
  $73 = HEAP8[$$021$i>>0]|0;
  $74 = ($73<<24>>24)==(0);
  $$$021$i = $74 ? 597716 : $$021$i;
  switch ($75|0) {
  case 1280:  {
   $$0$i$i = 597787;
   break;
  }
  case 1281:  {
   $$0$i$i = 597770;
   break;
  }
  case 1282:  {
   $$0$i$i = 597749;
   break;
  }
  case 1285:  {
   $$0$i$i = 597732;
   break;
  }
  default: {
   $$0$i$i = 597724;
  }
  }
  HEAP32[$vararg_buffer>>2] = $$$021$i;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = 597829;
  $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
  HEAP32[$vararg_ptr2>>2] = 399;
  $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
  HEAP32[$vararg_ptr3>>2] = 597922;
  $vararg_ptr4 = ((($vararg_buffer)) + 16|0);
  HEAP32[$vararg_ptr4>>2] = $$0$i$i;
  $vararg_ptr5 = ((($vararg_buffer)) + 20|0);
  HEAP32[$vararg_ptr5>>2] = $75;
  (_SDL_SetError(597803,$vararg_buffer)|0);
  $76 = HEAP32[$69>>2]|0;
  $77 = (FUNCTION_TABLE_i[$76 & 8191]()|0);
  $78 = ($77|0)==(0);
  if ($78) {
   $$0 = -1;
   break;
  } else {
   $$021$i = $$$021$i;$75 = $77;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GetRendererOutputSize($renderer,$w,$h) {
 $renderer = $renderer|0;
 $w = $w|0;
 $h = $h|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($renderer|0)==(0|0);
 do {
  if ($0) {
   label = 3;
  } else {
   $1 = HEAP32[$renderer>>2]|0;
   $2 = ($1|0)==(1279299|0);
   if ($2) {
    $3 = ((($renderer)) + 304|0);
    $4 = HEAP32[$3>>2]|0;
    $5 = ($4|0)==(0|0);
    if ($5) {
     $14 = ((($renderer)) + 8|0);
     $15 = HEAP32[$14>>2]|0;
     $16 = ($15|0)==(0|0);
     if (!($16)) {
      $17 = (FUNCTION_TABLE_iiii[$15 & 8191]($renderer,$w,$h)|0);
      $$0 = $17;
      break;
     }
     $18 = ((($renderer)) + 188|0);
     $19 = HEAP32[$18>>2]|0;
     $20 = ($19|0)==(0|0);
     if ($20) {
      $21 = (_SDL_SetError(597670,$vararg_buffer3)|0);
      $$0 = $21;
      break;
     } else {
      _SDL_GetWindowSize($19,$w,$h);
      $$0 = 0;
      break;
     }
    }
    $6 = HEAP32[$4>>2]|0;
    $7 = ($6|0)==(1279300|0);
    if (!($7)) {
     (_SDL_SetError(597654,$vararg_buffer1)|0);
     $$0 = -1;
     break;
    }
    $8 = ($w|0)==(0|0);
    if (!($8)) {
     $9 = ((($4)) + 12|0);
     $10 = HEAP32[$9>>2]|0;
     HEAP32[$w>>2] = $10;
    }
    $11 = ($h|0)==(0|0);
    if ($11) {
     $$0 = 0;
    } else {
     $12 = ((($4)) + 16|0);
     $13 = HEAP32[$12>>2]|0;
     HEAP32[$h>>2] = $13;
     $$0 = 0;
    }
   } else {
    label = 3;
   }
  }
 } while(0);
 if ((label|0) == 3) {
  (_SDL_SetError(597637,$vararg_buffer)|0);
  $$0 = -1;
 }
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GL_DeleteContext($context) {
 $context = $context|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)!=(0|0);
 $2 = ($context|0)!=(0|0);
 $or$cond = $2 & $1;
 if (!($or$cond)) {
  return;
 }
 $3 = ((($0)) + 640|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (_SDL_TLSGet($4)|0);
 $6 = ($5|0)==($context|0);
 if ($6) {
  (_SDL_GL_MakeCurrent(0,0)|0);
 }
 $7 = HEAP32[300441]|0;
 $8 = ((($7)) + 176|0);
 $9 = HEAP32[$8>>2]|0;
 FUNCTION_TABLE_vii[$9 & 8191]($7,$context);
 return;
}
function _SDL_GL_CreateContext($window) {
 $window = $window|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ((($window)) + 48|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = $7 & 2;
   $9 = ($8|0)==(0);
   if ($9) {
    (_SDL_SetError(597590,$vararg_buffer3)|0);
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   $10 = ((($0)) + 152|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = (FUNCTION_TABLE_iii[$11 & 8191]($0,$window)|0);
   $13 = ($12|0)==(0|0);
   if ($13) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   $14 = HEAP32[300441]|0;
   $15 = ((($14)) + 628|0);
   HEAP32[$15>>2] = $window;
   $16 = ((($14)) + 632|0);
   HEAP32[$16>>2] = $12;
   $17 = ((($14)) + 636|0);
   $18 = HEAP32[$17>>2]|0;
   (_SDL_TLSSet($18,$window,0)|0);
   $19 = HEAP32[300441]|0;
   $20 = ((($19)) + 640|0);
   $21 = HEAP32[$20>>2]|0;
   (_SDL_TLSSet($21,$12,0)|0);
   $$0 = $12;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GL_GetProcAddress($proc) {
 $proc = $proc|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ((($0)) + 144|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  (_SDL_SetError(595987,$vararg_buffer3)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = ((($0)) + 364|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if ($7) {
  (_SDL_SetError(595958,$vararg_buffer1)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $8 = (FUNCTION_TABLE_iii[$3 & 8191]($0,$proc)|0);
  $$0 = $8;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_GL_SetSwapInterval($interval) {
 $interval = $interval|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  $2 = (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ((($0)) + 640|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (_SDL_TLSGet($4)|0);
 $6 = ($5|0)==(0|0);
 if ($6) {
  $7 = (_SDL_SetError(598043,$vararg_buffer1)|0);
  $$0 = $7;
  STACKTOP = sp;return ($$0|0);
 }
 $8 = HEAP32[300441]|0;
 $9 = ((($8)) + 164|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10|0)==(0|0);
 if ($11) {
  $13 = (_SDL_SetError(598083,$vararg_buffer3)|0);
  $$0 = $13;
  STACKTOP = sp;return ($$0|0);
 } else {
  $12 = (FUNCTION_TABLE_iii[$10 & 8191]($8,$interval)|0);
  $$0 = $12;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_GL_GetSwapInterval() {
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  $$0 = 0;
  return ($$0|0);
 }
 $2 = ((($0)) + 640|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_SDL_TLSGet($3)|0);
 $5 = ($4|0)==(0|0);
 if ($5) {
  $$0 = 0;
  return ($$0|0);
 }
 $6 = HEAP32[300441]|0;
 $7 = ((($6)) + 168|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 if ($9) {
  $$0 = 0;
  return ($$0|0);
 }
 $10 = (FUNCTION_TABLE_ii[$8 & 8191]($6)|0);
 $$0 = $10;
 return ($$0|0);
}
function _GLES2_WindowEvent($renderer,$event) {
 $renderer = $renderer|0;
 $event = $event|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($event)) + 12|0);
 $1 = HEAP8[$0>>0]|0;
 switch ($1<<24>>24) {
 case 2: case 1: case 6:  {
  HEAP32[300738] = 0;
  return;
  break;
 }
 case 7:  {
  $2 = ((($renderer)) + 316|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($3)) + 88|0);
  $5 = HEAP32[$4>>2]|0;
  FUNCTION_TABLE_v[$5 & 8191]();
  return;
  break;
 }
 default: {
  return;
 }
 }
}
function _GLES2_GetOutputSize($renderer,$w,$h) {
 $renderer = $renderer|0;
 $w = $w|0;
 $h = $h|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($renderer)) + 188|0);
 $1 = HEAP32[$0>>2]|0;
 _SDL_GL_GetDrawableSize($1,$w,$h);
 return 0;
}
function _SDL_GL_GetDrawableSize($window,$w,$h) {
 $window = $window|0;
 $w = $w|0;
 $h = $h|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ((($0)) + 160|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0|0);
   if (!($8)) {
    FUNCTION_TABLE_viiii[$7 & 8191]($0,$window,$w,$h);
    STACKTOP = sp;return;
   }
   $9 = ($w|0)==(0|0);
   if (!($9)) {
    $10 = ((($window)) + 24|0);
    $11 = HEAP32[$10>>2]|0;
    HEAP32[$w>>2] = $11;
   }
   $12 = ($h|0)==(0|0);
   if ($12) {
    STACKTOP = sp;return;
   }
   $13 = ((($window)) + 28|0);
   $14 = HEAP32[$13>>2]|0;
   HEAP32[$h>>2] = $14;
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 STACKTOP = sp;return;
}
function _GLES2_CreateTexture($renderer,$texture) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 var $$ = 0, $$$021$i = 0, $$$021$i61 = 0, $$$i = 0, $$0$i$i = 0, $$0$i$i15 = 0, $$0$i$i23 = 0, $$0$i$i31 = 0, $$0$i$i39 = 0, $$0$i$i47 = 0, $$0$i$i55 = 0, $$0$i$i64 = 0, $$0$i$i7 = 0, $$021$i = 0, $$021$i61 = 0, $$1 = 0, $$mask = 0, $$pre = 0, $$pre$i = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0;
 var $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0;
 var $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0;
 var $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cond$i = 0, $cond4$i = 0, $format$0 = 0, $or$cond = 0, $phitmp = 0, $result$0$i = 0, $result$03$i = 0, $result$05$i = 0, $result$1$i = 0, $size$0 = 0, $size$1 = 0, $switch$split2D = 0, $switch$split722D = 0, $switch$split752D = 0, $switch$split782D = 0;
 var $switch$split812D = 0, $switch$split842D = 0, $switch$split872D = 0, $switch$split902D = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer16 = 0, $vararg_buffer24 = 0, $vararg_buffer32 = 0, $vararg_buffer40 = 0, $vararg_buffer48 = 0, $vararg_buffer56 = 0, $vararg_buffer64 = 0, $vararg_buffer8 = 0, $vararg_ptr11 = 0, $vararg_ptr12 = 0, $vararg_ptr13 = 0, $vararg_ptr14 = 0, $vararg_ptr15 = 0, $vararg_ptr19 = 0;
 var $vararg_ptr20 = 0, $vararg_ptr21 = 0, $vararg_ptr22 = 0, $vararg_ptr23 = 0, $vararg_ptr27 = 0, $vararg_ptr28 = 0, $vararg_ptr29 = 0, $vararg_ptr3 = 0, $vararg_ptr30 = 0, $vararg_ptr31 = 0, $vararg_ptr35 = 0, $vararg_ptr36 = 0, $vararg_ptr37 = 0, $vararg_ptr38 = 0, $vararg_ptr39 = 0, $vararg_ptr4 = 0, $vararg_ptr43 = 0, $vararg_ptr44 = 0, $vararg_ptr45 = 0, $vararg_ptr46 = 0;
 var $vararg_ptr47 = 0, $vararg_ptr5 = 0, $vararg_ptr51 = 0, $vararg_ptr52 = 0, $vararg_ptr53 = 0, $vararg_ptr54 = 0, $vararg_ptr55 = 0, $vararg_ptr59 = 0, $vararg_ptr6 = 0, $vararg_ptr60 = 0, $vararg_ptr61 = 0, $vararg_ptr62 = 0, $vararg_ptr63 = 0, $vararg_ptr67 = 0, $vararg_ptr68 = 0, $vararg_ptr69 = 0, $vararg_ptr7 = 0, $vararg_ptr70 = 0, $vararg_ptr71 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer64 = sp + 200|0;
 $vararg_buffer56 = sp + 176|0;
 $vararg_buffer48 = sp + 152|0;
 $vararg_buffer40 = sp + 128|0;
 $vararg_buffer32 = sp + 104|0;
 $vararg_buffer24 = sp + 80|0;
 $vararg_buffer16 = sp + 56|0;
 $vararg_buffer8 = sp + 32|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[300738]|0;
 $3 = HEAP32[$1>>2]|0;
 $4 = ($2|0)==($3|0);
 if ($4) {
  $12 = $1;
  label = 4;
 } else {
  $5 = ((($1)) + 268|0);
  HEAP32[$5>>2] = 0;
  $6 = ((($renderer)) + 188|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (_SDL_GL_MakeCurrent($7,$3)|0);
  $9 = ($8|0)<(0);
  if (!($9)) {
   $10 = HEAP32[$1>>2]|0;
   HEAP32[300738] = $10;
   (_GLES2_UpdateViewport($renderer)|0);
   $$pre$i = HEAP32[$0>>2]|0;
   $12 = $$pre$i;
   label = 4;
  }
 }
 if ((label|0) == 4) {
  $11 = ((($12)) + 4|0);
  $13 = HEAP32[$11>>2]|0;
  $14 = ($13|0)==(0);
  if (!($14)) {
   $15 = ((($12)) + 108|0);
   while(1) {
    $16 = HEAP32[$15>>2]|0;
    $17 = (FUNCTION_TABLE_i[$16 & 8191]()|0);
    $18 = ($17|0)==(0);
    if ($18) {
     break;
    }
   }
  }
 }
 $19 = ((($texture)) + 4|0);
 $20 = HEAP32[$19>>2]|0;
 $switch$split2D = ($20|0)<(825382478);
 L10: do {
  if ($switch$split2D) {
   $switch$split752D = ($20|0)<(374740996);
   if ($switch$split752D) {
    $switch$split812D = ($20|0)<(372645892);
    if ($switch$split812D) {
     switch ($20|0) {
     case 370546692:  {
      $format$0 = 6408;
      break L10;
      break;
     }
     default: {
      label = 9;
      break L10;
     }
     }
    } else {
     switch ($20|0) {
     case 372645892:  {
      $format$0 = 6408;
      break L10;
      break;
     }
     default: {
      label = 9;
      break L10;
     }
     }
    }
   } else {
    $switch$split842D = ($20|0)<(376840196);
    if ($switch$split842D) {
     switch ($20|0) {
     case 374740996:  {
      $format$0 = 6408;
      break L10;
      break;
     }
     default: {
      label = 9;
      break L10;
     }
     }
    } else {
     switch ($20|0) {
     case 376840196:  {
      $format$0 = 6408;
      break L10;
      break;
     }
     default: {
      label = 9;
      break L10;
     }
     }
    }
   }
  } else {
   $switch$split782D = ($20|0)<(842094169);
   L22: do {
    if ($switch$split782D) {
     $switch$split872D = ($20|0)<(842094158);
     if ($switch$split872D) {
      switch ($20|0) {
      case 825382478:  {
       break L22;
       break;
      }
      default: {
       label = 9;
       break L10;
      }
      }
     } else {
      switch ($20|0) {
      case 842094158:  {
       break L22;
       break;
      }
      default: {
       label = 9;
       break L10;
      }
      }
     }
    } else {
     $switch$split902D = ($20|0)<(1448433993);
     if ($switch$split902D) {
      switch ($20|0) {
      case 842094169:  {
       break L22;
       break;
      }
      default: {
       label = 9;
       break L10;
      }
      }
     } else {
      switch ($20|0) {
      case 1448433993:  {
       break L22;
       break;
      }
      default: {
       label = 9;
       break L10;
      }
      }
     }
    }
   } while(0);
   $format$0 = 6409;
  }
 } while(0);
 if ((label|0) == 9) {
  $21 = (_SDL_SetError(598126,$vararg_buffer)|0);
  $$1 = $21;
  STACKTOP = sp;return ($$1|0);
 }
 $22 = (_SDL_calloc(1,44)|0);
 $23 = ($22|0)==(0|0);
 if ($23) {
  $24 = (_SDL_Error(0)|0);
  $$1 = $24;
  STACKTOP = sp;return ($$1|0);
 }
 HEAP32[$22>>2] = 0;
 $25 = ((($22)) + 4|0);
 HEAP32[$25>>2] = 3553;
 $26 = ((($22)) + 8|0);
 HEAP32[$26>>2] = $format$0;
 $27 = ((($22)) + 12|0);
 HEAP32[$27>>2] = 5121;
 $28 = HEAP32[$19>>2]|0;
 $29 = ($28|0)==(1448433993);
 $30 = ($28|0)==(842094169);
 $$ = $29 | $30;
 $31 = $$&1;
 $32 = ((($22)) + 24|0);
 HEAP32[$32>>2] = $31;
 $33 = ($28|0)==(842094158);
 $34 = ($28|0)==(825382478);
 $35 = $33 | $34;
 $36 = $35&1;
 $37 = ((($22)) + 28|0);
 HEAP32[$37>>2] = $36;
 $38 = ((($22)) + 36|0);
 HEAP32[$38>>2] = 0;
 $39 = ((($22)) + 32|0);
 HEAP32[$39>>2] = 0;
 $40 = (_SDL_GetHint(604329)|0);
 $41 = ($40|0)==(0|0);
 if ($41) {
  $113 = 9728;
 } else {
  $42 = HEAP8[$40>>0]|0;
  $43 = ($42<<24>>24)==(48);
  if ($43) {
   $113 = 9728;
  } else {
   $44 = (_SDL_strcasecmp($40,604354)|0);
   $45 = ($44|0)==(0);
   $$$i = $45 ? 9728 : 9729;
   $113 = $$$i;
  }
 }
 $46 = ((($texture)) + 8|0);
 $47 = HEAP32[$46>>2]|0;
 $48 = ($47|0)==(1);
 if ($48) {
  $49 = ((($texture)) + 12|0);
  $50 = HEAP32[$49>>2]|0;
  $51 = HEAP32[$19>>2]|0;
  $52 = ($51|0)==(0);
  $$mask = $51 & -268435456;
  $53 = ($$mask|0)==(268435456);
  $or$cond = $52 | $53;
  L46: do {
   if ($or$cond) {
    $55 = $51 & 255;
    $57 = $55;
   } else {
    $switch$split722D = ($51|0)<(1498831189);
    if ($switch$split722D) {
     switch ($51|0) {
     case 844715353:  {
      $57 = 2;
      break L46;
      break;
     }
     default: {
     }
     }
    } else {
     switch ($51|0) {
     case 1498831189:  {
      $57 = 2;
      break L46;
      break;
     }
     default: {
     }
     }
    }
    $54 = ($51|0)==(1431918169);
    $phitmp = $54 ? 2 : 1;
    $57 = $phitmp;
   }
  } while(0);
  $56 = Math_imul($57, $50)|0;
  $58 = ((($22)) + 20|0);
  HEAP32[$58>>2] = $56;
  $59 = ((($texture)) + 16|0);
  $60 = HEAP32[$59>>2]|0;
  $61 = Math_imul($60, $56)|0;
  $62 = HEAP32[$32>>2]|0;
  $63 = ($62|0)==(0);
  if ($63) {
   $size$0 = $61;
  } else {
   $64 = (($61|0) / 2)&-1;
   $65 = (($64) + ($61))|0;
   $size$0 = $65;
  }
  $66 = HEAP32[$37>>2]|0;
  $67 = ($66|0)==(0);
  if ($67) {
   $size$1 = $size$0;
  } else {
   $68 = (($61|0) / 2)&-1;
   $69 = (($68) + ($size$0))|0;
   $size$1 = $69;
  }
  $70 = (_SDL_calloc(1,$size$1)|0);
  $71 = ((($22)) + 16|0);
  HEAP32[$71>>2] = $70;
  $72 = ($70|0)==(0|0);
  if ($72) {
   _SDL_free($22);
   $73 = (_SDL_Error(0)|0);
   $$1 = $73;
   STACKTOP = sp;return ($$1|0);
  }
 }
 $74 = HEAP32[$0>>2]|0;
 $75 = ((($74)) + 4|0);
 $76 = HEAP32[$75>>2]|0;
 $77 = ($76|0)==(0);
 L63: do {
  if (!($77)) {
   $78 = ((($74)) + 108|0);
   $79 = HEAP32[$78>>2]|0;
   $80 = (FUNCTION_TABLE_i[$79 & 8191]()|0);
   $81 = ($80|0)==(0);
   if (!($81)) {
    $$021$i = 1282466;$84 = $80;
    while(1) {
     $82 = HEAP8[$$021$i>>0]|0;
     $83 = ($82<<24>>24)==(0);
     $$$021$i = $83 ? 597716 : $$021$i;
     switch ($84|0) {
     case 1280:  {
      $$0$i$i = 597787;
      break;
     }
     case 1281:  {
      $$0$i$i = 597770;
      break;
     }
     case 1282:  {
      $$0$i$i = 597749;
      break;
     }
     case 1285:  {
      $$0$i$i = 597732;
      break;
     }
     default: {
      $$0$i$i = 597724;
     }
     }
     HEAP32[$vararg_buffer1>>2] = $$$021$i;
     $vararg_ptr3 = ((($vararg_buffer1)) + 4|0);
     HEAP32[$vararg_ptr3>>2] = 597829;
     $vararg_ptr4 = ((($vararg_buffer1)) + 8|0);
     HEAP32[$vararg_ptr4>>2] = 574;
     $vararg_ptr5 = ((($vararg_buffer1)) + 12|0);
     HEAP32[$vararg_ptr5>>2] = 598155;
     $vararg_ptr6 = ((($vararg_buffer1)) + 16|0);
     HEAP32[$vararg_ptr6>>2] = $$0$i$i;
     $vararg_ptr7 = ((($vararg_buffer1)) + 20|0);
     HEAP32[$vararg_ptr7>>2] = $84;
     (_SDL_SetError(597803,$vararg_buffer1)|0);
     $85 = HEAP32[$78>>2]|0;
     $86 = (FUNCTION_TABLE_i[$85 & 8191]()|0);
     $87 = ($86|0)==(0);
     if ($87) {
      break L63;
     } else {
      $$021$i = $$$021$i;$84 = $86;
     }
    }
   }
  }
 } while(0);
 $88 = HEAP32[$32>>2]|0;
 $89 = ($88|0)==(0);
 if (!($89)) {
  $90 = ((($1)) + 96|0);
  $91 = HEAP32[$90>>2]|0;
  FUNCTION_TABLE_vii[$91 & 8191](1,$39);
  $92 = HEAP32[$0>>2]|0;
  $93 = ((($92)) + 4|0);
  $94 = HEAP32[$93>>2]|0;
  $95 = ($94|0)==(0);
  if (!($95)) {
   $96 = ((($92)) + 108|0);
   $97 = HEAP32[$96>>2]|0;
   $98 = (FUNCTION_TABLE_i[$97 & 8191]()|0);
   $99 = ($98|0)==(0);
   if (!($99)) {
    $100 = $98;
    while(1) {
     switch ($100|0) {
     case 1280:  {
      $$0$i$i7 = 597787;
      break;
     }
     case 1281:  {
      $$0$i$i7 = 597770;
      break;
     }
     case 1282:  {
      $$0$i$i7 = 597749;
      break;
     }
     case 1285:  {
      $$0$i$i7 = 597732;
      break;
     }
     default: {
      $$0$i$i7 = 597724;
     }
     }
     HEAP32[$vararg_buffer8>>2] = 598210;
     $vararg_ptr11 = ((($vararg_buffer8)) + 4|0);
     HEAP32[$vararg_ptr11>>2] = 597829;
     $vararg_ptr12 = ((($vararg_buffer8)) + 8|0);
     HEAP32[$vararg_ptr12>>2] = 578;
     $vararg_ptr13 = ((($vararg_buffer8)) + 12|0);
     HEAP32[$vararg_ptr13>>2] = 598155;
     $vararg_ptr14 = ((($vararg_buffer8)) + 16|0);
     HEAP32[$vararg_ptr14>>2] = $$0$i$i7;
     $vararg_ptr15 = ((($vararg_buffer8)) + 20|0);
     HEAP32[$vararg_ptr15>>2] = $100;
     (_SDL_SetError(597803,$vararg_buffer8)|0);
     $101 = HEAP32[$96>>2]|0;
     $102 = (FUNCTION_TABLE_i[$101 & 8191]()|0);
     $103 = ($102|0)==(0);
     if ($103) {
      $$1 = -1;
      break;
     } else {
      $100 = $102;
     }
    }
    STACKTOP = sp;return ($$1|0);
   }
  }
  $104 = ((($1)) + 16|0);
  $105 = HEAP32[$104>>2]|0;
  FUNCTION_TABLE_vi[$105 & 8191](33986);
  $106 = ((($1)) + 28|0);
  $107 = HEAP32[$106>>2]|0;
  $108 = HEAP32[$25>>2]|0;
  $109 = HEAP32[$39>>2]|0;
  FUNCTION_TABLE_vii[$107 & 8191]($108,$109);
  $110 = ((($1)) + 160|0);
  $111 = HEAP32[$110>>2]|0;
  $112 = HEAP32[$25>>2]|0;
  FUNCTION_TABLE_viii[$111 & 8191]($112,10241,$113);
  $114 = HEAP32[$110>>2]|0;
  $115 = HEAP32[$25>>2]|0;
  FUNCTION_TABLE_viii[$114 & 8191]($115,10240,$113);
  $116 = HEAP32[$110>>2]|0;
  $117 = HEAP32[$25>>2]|0;
  FUNCTION_TABLE_viii[$116 & 8191]($117,10242,33071);
  $118 = HEAP32[$110>>2]|0;
  $119 = HEAP32[$25>>2]|0;
  FUNCTION_TABLE_viii[$118 & 8191]($119,10243,33071);
  $120 = ((($1)) + 156|0);
  $121 = HEAP32[$120>>2]|0;
  $122 = HEAP32[$25>>2]|0;
  $123 = ((($texture)) + 12|0);
  $124 = HEAP32[$123>>2]|0;
  $125 = (($124|0) / 2)&-1;
  $126 = ((($texture)) + 16|0);
  $127 = HEAP32[$126>>2]|0;
  $128 = (($127|0) / 2)&-1;
  FUNCTION_TABLE_viiiiiiiii[$121 & 8191]($122,0,$format$0,$125,$128,0,$format$0,5121,0);
  $129 = HEAP32[$90>>2]|0;
  FUNCTION_TABLE_vii[$129 & 8191](1,$38);
  $130 = HEAP32[$0>>2]|0;
  $131 = ((($130)) + 4|0);
  $132 = HEAP32[$131>>2]|0;
  $133 = ($132|0)==(0);
  if (!($133)) {
   $134 = ((($130)) + 108|0);
   $135 = HEAP32[$134>>2]|0;
   $136 = (FUNCTION_TABLE_i[$135 & 8191]()|0);
   $137 = ($136|0)==(0);
   if (!($137)) {
    $138 = $136;
    while(1) {
     switch ($138|0) {
     case 1280:  {
      $$0$i$i15 = 597787;
      break;
     }
     case 1281:  {
      $$0$i$i15 = 597770;
      break;
     }
     case 1282:  {
      $$0$i$i15 = 597749;
      break;
     }
     case 1285:  {
      $$0$i$i15 = 597732;
      break;
     }
     default: {
      $$0$i$i15 = 597724;
     }
     }
     HEAP32[$vararg_buffer16>>2] = 598210;
     $vararg_ptr19 = ((($vararg_buffer16)) + 4|0);
     HEAP32[$vararg_ptr19>>2] = 597829;
     $vararg_ptr20 = ((($vararg_buffer16)) + 8|0);
     HEAP32[$vararg_ptr20>>2] = 590;
     $vararg_ptr21 = ((($vararg_buffer16)) + 12|0);
     HEAP32[$vararg_ptr21>>2] = 598155;
     $vararg_ptr22 = ((($vararg_buffer16)) + 16|0);
     HEAP32[$vararg_ptr22>>2] = $$0$i$i15;
     $vararg_ptr23 = ((($vararg_buffer16)) + 20|0);
     HEAP32[$vararg_ptr23>>2] = $138;
     (_SDL_SetError(597803,$vararg_buffer16)|0);
     $139 = HEAP32[$134>>2]|0;
     $140 = (FUNCTION_TABLE_i[$139 & 8191]()|0);
     $141 = ($140|0)==(0);
     if ($141) {
      $$1 = -1;
      break;
     } else {
      $138 = $140;
     }
    }
    STACKTOP = sp;return ($$1|0);
   }
  }
  $142 = HEAP32[$104>>2]|0;
  FUNCTION_TABLE_vi[$142 & 8191](33985);
  $143 = HEAP32[$106>>2]|0;
  $144 = HEAP32[$25>>2]|0;
  $145 = HEAP32[$38>>2]|0;
  FUNCTION_TABLE_vii[$143 & 8191]($144,$145);
  $146 = HEAP32[$110>>2]|0;
  $147 = HEAP32[$25>>2]|0;
  FUNCTION_TABLE_viii[$146 & 8191]($147,10241,$113);
  $148 = HEAP32[$110>>2]|0;
  $149 = HEAP32[$25>>2]|0;
  FUNCTION_TABLE_viii[$148 & 8191]($149,10240,$113);
  $150 = HEAP32[$110>>2]|0;
  $151 = HEAP32[$25>>2]|0;
  FUNCTION_TABLE_viii[$150 & 8191]($151,10242,33071);
  $152 = HEAP32[$110>>2]|0;
  $153 = HEAP32[$25>>2]|0;
  FUNCTION_TABLE_viii[$152 & 8191]($153,10243,33071);
  $154 = HEAP32[$120>>2]|0;
  $155 = HEAP32[$25>>2]|0;
  $156 = HEAP32[$123>>2]|0;
  $157 = (($156|0) / 2)&-1;
  $158 = HEAP32[$126>>2]|0;
  $159 = (($158|0) / 2)&-1;
  FUNCTION_TABLE_viiiiiiiii[$154 & 8191]($155,0,$format$0,$157,$159,0,$format$0,5121,0);
  $160 = HEAP32[$0>>2]|0;
  $161 = ((($160)) + 4|0);
  $162 = HEAP32[$161>>2]|0;
  $163 = ($162|0)==(0);
  if (!($163)) {
   $164 = ((($160)) + 108|0);
   $165 = HEAP32[$164>>2]|0;
   $166 = (FUNCTION_TABLE_i[$165 & 8191]()|0);
   $167 = ($166|0)==(0);
   if (!($167)) {
    $168 = $166;
    while(1) {
     switch ($168|0) {
     case 1280:  {
      $$0$i$i23 = 597787;
      break;
     }
     case 1281:  {
      $$0$i$i23 = 597770;
      break;
     }
     case 1282:  {
      $$0$i$i23 = 597749;
      break;
     }
     case 1285:  {
      $$0$i$i23 = 597732;
      break;
     }
     default: {
      $$0$i$i23 = 597724;
     }
     }
     HEAP32[$vararg_buffer24>>2] = 598225;
     $vararg_ptr27 = ((($vararg_buffer24)) + 4|0);
     HEAP32[$vararg_ptr27>>2] = 597829;
     $vararg_ptr28 = ((($vararg_buffer24)) + 8|0);
     HEAP32[$vararg_ptr28>>2] = 600;
     $vararg_ptr29 = ((($vararg_buffer24)) + 12|0);
     HEAP32[$vararg_ptr29>>2] = 598155;
     $vararg_ptr30 = ((($vararg_buffer24)) + 16|0);
     HEAP32[$vararg_ptr30>>2] = $$0$i$i23;
     $vararg_ptr31 = ((($vararg_buffer24)) + 20|0);
     HEAP32[$vararg_ptr31>>2] = $168;
     (_SDL_SetError(597803,$vararg_buffer24)|0);
     $169 = HEAP32[$164>>2]|0;
     $170 = (FUNCTION_TABLE_i[$169 & 8191]()|0);
     $171 = ($170|0)==(0);
     if ($171) {
      $$1 = -1;
      break;
     } else {
      $168 = $170;
     }
    }
    STACKTOP = sp;return ($$1|0);
   }
  }
 }
 $172 = HEAP32[$37>>2]|0;
 $173 = ($172|0)==(0);
 $$pre = ((($1)) + 96|0);
 if (!($173)) {
  $174 = HEAP32[$$pre>>2]|0;
  FUNCTION_TABLE_vii[$174 & 8191](1,$38);
  $175 = HEAP32[$0>>2]|0;
  $176 = ((($175)) + 4|0);
  $177 = HEAP32[$176>>2]|0;
  $178 = ($177|0)==(0);
  if (!($178)) {
   $179 = ((($175)) + 108|0);
   $180 = HEAP32[$179>>2]|0;
   $181 = (FUNCTION_TABLE_i[$180 & 8191]()|0);
   $182 = ($181|0)==(0);
   if (!($182)) {
    $183 = $181;
    while(1) {
     switch ($183|0) {
     case 1280:  {
      $$0$i$i31 = 597787;
      break;
     }
     case 1281:  {
      $$0$i$i31 = 597770;
      break;
     }
     case 1282:  {
      $$0$i$i31 = 597749;
      break;
     }
     case 1285:  {
      $$0$i$i31 = 597732;
      break;
     }
     default: {
      $$0$i$i31 = 597724;
     }
     }
     HEAP32[$vararg_buffer32>>2] = 598210;
     $vararg_ptr35 = ((($vararg_buffer32)) + 4|0);
     HEAP32[$vararg_ptr35>>2] = 597829;
     $vararg_ptr36 = ((($vararg_buffer32)) + 8|0);
     HEAP32[$vararg_ptr36>>2] = 607;
     $vararg_ptr37 = ((($vararg_buffer32)) + 12|0);
     HEAP32[$vararg_ptr37>>2] = 598155;
     $vararg_ptr38 = ((($vararg_buffer32)) + 16|0);
     HEAP32[$vararg_ptr38>>2] = $$0$i$i31;
     $vararg_ptr39 = ((($vararg_buffer32)) + 20|0);
     HEAP32[$vararg_ptr39>>2] = $183;
     (_SDL_SetError(597803,$vararg_buffer32)|0);
     $184 = HEAP32[$179>>2]|0;
     $185 = (FUNCTION_TABLE_i[$184 & 8191]()|0);
     $186 = ($185|0)==(0);
     if ($186) {
      $$1 = -1;
      break;
     } else {
      $183 = $185;
     }
    }
    STACKTOP = sp;return ($$1|0);
   }
  }
  $187 = ((($1)) + 16|0);
  $188 = HEAP32[$187>>2]|0;
  FUNCTION_TABLE_vi[$188 & 8191](33985);
  $189 = ((($1)) + 28|0);
  $190 = HEAP32[$189>>2]|0;
  $191 = HEAP32[$25>>2]|0;
  $192 = HEAP32[$38>>2]|0;
  FUNCTION_TABLE_vii[$190 & 8191]($191,$192);
  $193 = ((($1)) + 160|0);
  $194 = HEAP32[$193>>2]|0;
  $195 = HEAP32[$25>>2]|0;
  FUNCTION_TABLE_viii[$194 & 8191]($195,10241,$113);
  $196 = HEAP32[$193>>2]|0;
  $197 = HEAP32[$25>>2]|0;
  FUNCTION_TABLE_viii[$196 & 8191]($197,10240,$113);
  $198 = HEAP32[$193>>2]|0;
  $199 = HEAP32[$25>>2]|0;
  FUNCTION_TABLE_viii[$198 & 8191]($199,10242,33071);
  $200 = HEAP32[$193>>2]|0;
  $201 = HEAP32[$25>>2]|0;
  FUNCTION_TABLE_viii[$200 & 8191]($201,10243,33071);
  $202 = ((($1)) + 156|0);
  $203 = HEAP32[$202>>2]|0;
  $204 = HEAP32[$25>>2]|0;
  $205 = ((($texture)) + 12|0);
  $206 = HEAP32[$205>>2]|0;
  $207 = (($206|0) / 2)&-1;
  $208 = ((($texture)) + 16|0);
  $209 = HEAP32[$208>>2]|0;
  $210 = (($209|0) / 2)&-1;
  FUNCTION_TABLE_viiiiiiiii[$203 & 8191]($204,0,6410,$207,$210,0,6410,5121,0);
  $211 = HEAP32[$0>>2]|0;
  $212 = ((($211)) + 4|0);
  $213 = HEAP32[$212>>2]|0;
  $214 = ($213|0)==(0);
  if (!($214)) {
   $215 = ((($211)) + 108|0);
   $216 = HEAP32[$215>>2]|0;
   $217 = (FUNCTION_TABLE_i[$216 & 8191]()|0);
   $218 = ($217|0)==(0);
   if (!($218)) {
    $219 = $217;
    while(1) {
     switch ($219|0) {
     case 1280:  {
      $$0$i$i39 = 597787;
      break;
     }
     case 1281:  {
      $$0$i$i39 = 597770;
      break;
     }
     case 1282:  {
      $$0$i$i39 = 597749;
      break;
     }
     case 1285:  {
      $$0$i$i39 = 597732;
      break;
     }
     default: {
      $$0$i$i39 = 597724;
     }
     }
     HEAP32[$vararg_buffer40>>2] = 598225;
     $vararg_ptr43 = ((($vararg_buffer40)) + 4|0);
     HEAP32[$vararg_ptr43>>2] = 597829;
     $vararg_ptr44 = ((($vararg_buffer40)) + 8|0);
     HEAP32[$vararg_ptr44>>2] = 617;
     $vararg_ptr45 = ((($vararg_buffer40)) + 12|0);
     HEAP32[$vararg_ptr45>>2] = 598155;
     $vararg_ptr46 = ((($vararg_buffer40)) + 16|0);
     HEAP32[$vararg_ptr46>>2] = $$0$i$i39;
     $vararg_ptr47 = ((($vararg_buffer40)) + 20|0);
     HEAP32[$vararg_ptr47>>2] = $219;
     (_SDL_SetError(597803,$vararg_buffer40)|0);
     $220 = HEAP32[$215>>2]|0;
     $221 = (FUNCTION_TABLE_i[$220 & 8191]()|0);
     $222 = ($221|0)==(0);
     if ($222) {
      $$1 = -1;
      break;
     } else {
      $219 = $221;
     }
    }
    STACKTOP = sp;return ($$1|0);
   }
  }
 }
 $223 = HEAP32[$$pre>>2]|0;
 FUNCTION_TABLE_vii[$223 & 8191](1,$22);
 $224 = HEAP32[$0>>2]|0;
 $225 = ((($224)) + 4|0);
 $226 = HEAP32[$225>>2]|0;
 $227 = ($226|0)==(0);
 if (!($227)) {
  $228 = ((($224)) + 108|0);
  $229 = HEAP32[$228>>2]|0;
  $230 = (FUNCTION_TABLE_i[$229 & 8191]()|0);
  $231 = ($230|0)==(0);
  if (!($231)) {
   $232 = $230;
   while(1) {
    switch ($232|0) {
    case 1280:  {
     $$0$i$i47 = 597787;
     break;
    }
    case 1281:  {
     $$0$i$i47 = 597770;
     break;
    }
    case 1282:  {
     $$0$i$i47 = 597749;
     break;
    }
    case 1285:  {
     $$0$i$i47 = 597732;
     break;
    }
    default: {
     $$0$i$i47 = 597724;
    }
    }
    HEAP32[$vararg_buffer48>>2] = 598210;
    $vararg_ptr51 = ((($vararg_buffer48)) + 4|0);
    HEAP32[$vararg_ptr51>>2] = 597829;
    $vararg_ptr52 = ((($vararg_buffer48)) + 8|0);
    HEAP32[$vararg_ptr52>>2] = 623;
    $vararg_ptr53 = ((($vararg_buffer48)) + 12|0);
    HEAP32[$vararg_ptr53>>2] = 598155;
    $vararg_ptr54 = ((($vararg_buffer48)) + 16|0);
    HEAP32[$vararg_ptr54>>2] = $$0$i$i47;
    $vararg_ptr55 = ((($vararg_buffer48)) + 20|0);
    HEAP32[$vararg_ptr55>>2] = $232;
    (_SDL_SetError(597803,$vararg_buffer48)|0);
    $233 = HEAP32[$228>>2]|0;
    $234 = (FUNCTION_TABLE_i[$233 & 8191]()|0);
    $235 = ($234|0)==(0);
    if ($235) {
     $$1 = -1;
     break;
    } else {
     $232 = $234;
    }
   }
   STACKTOP = sp;return ($$1|0);
  }
 }
 $236 = ((($texture)) + 68|0);
 HEAP32[$236>>2] = $22;
 $237 = ((($1)) + 16|0);
 $238 = HEAP32[$237>>2]|0;
 FUNCTION_TABLE_vi[$238 & 8191](33984);
 $239 = ((($1)) + 28|0);
 $240 = HEAP32[$239>>2]|0;
 $241 = HEAP32[$25>>2]|0;
 $242 = HEAP32[$22>>2]|0;
 FUNCTION_TABLE_vii[$240 & 8191]($241,$242);
 $243 = ((($1)) + 160|0);
 $244 = HEAP32[$243>>2]|0;
 $245 = HEAP32[$25>>2]|0;
 FUNCTION_TABLE_viii[$244 & 8191]($245,10241,$113);
 $246 = HEAP32[$243>>2]|0;
 $247 = HEAP32[$25>>2]|0;
 FUNCTION_TABLE_viii[$246 & 8191]($247,10240,$113);
 $248 = HEAP32[$243>>2]|0;
 $249 = HEAP32[$25>>2]|0;
 FUNCTION_TABLE_viii[$248 & 8191]($249,10242,33071);
 $250 = HEAP32[$243>>2]|0;
 $251 = HEAP32[$25>>2]|0;
 FUNCTION_TABLE_viii[$250 & 8191]($251,10243,33071);
 $252 = ((($1)) + 156|0);
 $253 = HEAP32[$252>>2]|0;
 $254 = HEAP32[$25>>2]|0;
 $255 = ((($texture)) + 12|0);
 $256 = HEAP32[$255>>2]|0;
 $257 = ((($texture)) + 16|0);
 $258 = HEAP32[$257>>2]|0;
 FUNCTION_TABLE_viiiiiiiii[$253 & 8191]($254,0,$format$0,$256,$258,0,$format$0,5121,0);
 $259 = HEAP32[$0>>2]|0;
 $260 = ((($259)) + 4|0);
 $261 = HEAP32[$260>>2]|0;
 $262 = ($261|0)==(0);
 if (!($262)) {
  $263 = ((($259)) + 108|0);
  $264 = HEAP32[$263>>2]|0;
  $265 = (FUNCTION_TABLE_i[$264 & 8191]()|0);
  $266 = ($265|0)==(0);
  if (!($266)) {
   $267 = $265;
   while(1) {
    switch ($267|0) {
    case 1280:  {
     $$0$i$i55 = 597787;
     break;
    }
    case 1281:  {
     $$0$i$i55 = 597770;
     break;
    }
    case 1282:  {
     $$0$i$i55 = 597749;
     break;
    }
    case 1285:  {
     $$0$i$i55 = 597732;
     break;
    }
    default: {
     $$0$i$i55 = 597724;
    }
    }
    HEAP32[$vararg_buffer56>>2] = 598225;
    $vararg_ptr59 = ((($vararg_buffer56)) + 4|0);
    HEAP32[$vararg_ptr59>>2] = 597829;
    $vararg_ptr60 = ((($vararg_buffer56)) + 8|0);
    HEAP32[$vararg_ptr60>>2] = 634;
    $vararg_ptr61 = ((($vararg_buffer56)) + 12|0);
    HEAP32[$vararg_ptr61>>2] = 598155;
    $vararg_ptr62 = ((($vararg_buffer56)) + 16|0);
    HEAP32[$vararg_ptr62>>2] = $$0$i$i55;
    $vararg_ptr63 = ((($vararg_buffer56)) + 20|0);
    HEAP32[$vararg_ptr63>>2] = $267;
    (_SDL_SetError(597803,$vararg_buffer56)|0);
    $268 = HEAP32[$263>>2]|0;
    $269 = (FUNCTION_TABLE_i[$268 & 8191]()|0);
    $270 = ($269|0)==(0);
    if ($270) {
     $$1 = -1;
     break;
    } else {
     $267 = $269;
    }
   }
   STACKTOP = sp;return ($$1|0);
  }
 }
 $271 = HEAP32[$46>>2]|0;
 $272 = ($271|0)==(2);
 if ($272) {
  $273 = HEAP32[$0>>2]|0;
  $274 = HEAP32[$255>>2]|0;
  $275 = HEAP32[$257>>2]|0;
  $276 = ((($273)) + 232|0);
  $result$03$i = HEAP32[$276>>2]|0;
  $cond4$i = ($result$03$i|0)==(0|0);
  L162: do {
   if ($cond4$i) {
    label = 97;
   } else {
    $result$05$i = $result$03$i;
    while(1) {
     $277 = HEAP32[$result$05$i>>2]|0;
     $278 = ($277|0)==($274|0);
     if ($278) {
      $279 = ((($result$05$i)) + 4|0);
      $280 = HEAP32[$279>>2]|0;
      $281 = ($280|0)==($275|0);
      if ($281) {
       $result$1$i = $result$05$i;
       break L162;
      }
     }
     $282 = ((($result$05$i)) + 12|0);
     $result$0$i = HEAP32[$282>>2]|0;
     $cond$i = ($result$0$i|0)==(0|0);
     if ($cond$i) {
      label = 97;
      break;
     } else {
      $result$05$i = $result$0$i;
     }
    }
   }
  } while(0);
  if ((label|0) == 97) {
   $283 = (_SDL_malloc(16)|0);
   HEAP32[$283>>2] = $274;
   $284 = ((($283)) + 4|0);
   HEAP32[$284>>2] = $275;
   $285 = ((($273)) + 92|0);
   $286 = HEAP32[$285>>2]|0;
   $287 = ((($283)) + 8|0);
   FUNCTION_TABLE_vii[$286 & 8191](1,$287);
   $288 = HEAP32[$276>>2]|0;
   $289 = ((($283)) + 12|0);
   HEAP32[$289>>2] = $288;
   HEAP32[$276>>2] = $283;
   $result$1$i = $283;
  }
  $290 = ((($22)) + 40|0);
  HEAP32[$290>>2] = $result$1$i;
 } else {
  $291 = ((($22)) + 40|0);
  HEAP32[$291>>2] = 0;
 }
 $292 = HEAP32[$0>>2]|0;
 $293 = ((($292)) + 4|0);
 $294 = HEAP32[$293>>2]|0;
 $295 = ($294|0)==(0);
 if ($295) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $296 = ((($292)) + 108|0);
 $297 = HEAP32[$296>>2]|0;
 $298 = (FUNCTION_TABLE_i[$297 & 8191]()|0);
 $299 = ($298|0)==(0);
 if ($299) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 } else {
  $$021$i61 = 1282466;$302 = $298;
 }
 while(1) {
  $300 = HEAP8[$$021$i61>>0]|0;
  $301 = ($300<<24>>24)==(0);
  $$$021$i61 = $301 ? 597716 : $$021$i61;
  switch ($302|0) {
  case 1280:  {
   $$0$i$i64 = 597787;
   break;
  }
  case 1281:  {
   $$0$i$i64 = 597770;
   break;
  }
  case 1282:  {
   $$0$i$i64 = 597749;
   break;
  }
  case 1285:  {
   $$0$i$i64 = 597732;
   break;
  }
  default: {
   $$0$i$i64 = 597724;
  }
  }
  HEAP32[$vararg_buffer64>>2] = $$$021$i61;
  $vararg_ptr67 = ((($vararg_buffer64)) + 4|0);
  HEAP32[$vararg_ptr67>>2] = 597829;
  $vararg_ptr68 = ((($vararg_buffer64)) + 8|0);
  HEAP32[$vararg_ptr68>>2] = 644;
  $vararg_ptr69 = ((($vararg_buffer64)) + 12|0);
  HEAP32[$vararg_ptr69>>2] = 598155;
  $vararg_ptr70 = ((($vararg_buffer64)) + 16|0);
  HEAP32[$vararg_ptr70>>2] = $$0$i$i64;
  $vararg_ptr71 = ((($vararg_buffer64)) + 20|0);
  HEAP32[$vararg_ptr71>>2] = $302;
  (_SDL_SetError(597803,$vararg_buffer64)|0);
  $303 = HEAP32[$296>>2]|0;
  $304 = (FUNCTION_TABLE_i[$303 & 8191]()|0);
  $305 = ($304|0)==(0);
  if ($305) {
   $$1 = -1;
   break;
  } else {
   $$021$i61 = $$$021$i61;$302 = $304;
  }
 }
 STACKTOP = sp;return ($$1|0);
}
function _SDL_strcasecmp($str1,$str2) {
 $str1 = $str1|0;
 $str2 = $str2|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strcasecmp($str1,$str2)|0);
 return ($0|0);
}
function _GLES2_UpdateTexture($renderer,$texture,$rect,$pixels,$pitch) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 $rect = $rect|0;
 $pixels = $pixels|0;
 $pitch = $pitch|0;
 var $$0 = 0, $$0$i$i = 0, $$01 = 0, $$011$i = 0, $$011$i16 = 0, $$011$i26 = 0, $$011$i6 = 0, $$mask = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $blob$0$i = 0, $blob$0$i19 = 0, $blob$0$i29 = 0, $blob$0$i9 = 0, $exitcond$i = 0, $exitcond$i17 = 0, $exitcond$i27 = 0, $exitcond$i7 = 0, $or$cond = 0, $phitmp = 0, $src$02$i = 0, $src$02$i15 = 0, $src$02$i25 = 0, $src$02$i5 = 0, $src$1$i = 0, $src$1$i10 = 0;
 var $src$1$i20 = 0, $src$1$i30 = 0, $switch$split2D = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0, $vararg_ptr4 = 0, $vararg_ptr5 = 0, $y$03$i = 0, $y$03$i14 = 0, $y$03$i24 = 0, $y$03$i4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($texture)) + 68|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[300738]|0;
 $5 = HEAP32[$1>>2]|0;
 $6 = ($4|0)==($5|0);
 if ($6) {
  $14 = $1;
  label = 4;
 } else {
  $7 = ((($1)) + 268|0);
  HEAP32[$7>>2] = 0;
  $8 = ((($renderer)) + 188|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = (_SDL_GL_MakeCurrent($9,$5)|0);
  $11 = ($10|0)<(0);
  if (!($11)) {
   $12 = HEAP32[$1>>2]|0;
   HEAP32[300738] = $12;
   (_GLES2_UpdateViewport($renderer)|0);
   $$pre$i = HEAP32[$0>>2]|0;
   $14 = $$pre$i;
   label = 4;
  }
 }
 if ((label|0) == 4) {
  $13 = ((($14)) + 4|0);
  $15 = HEAP32[$13>>2]|0;
  $16 = ($15|0)==(0);
  if (!($16)) {
   $17 = ((($14)) + 108|0);
   while(1) {
    $18 = HEAP32[$17>>2]|0;
    $19 = (FUNCTION_TABLE_i[$18 & 8191]()|0);
    $20 = ($19|0)==(0);
    if ($20) {
     break;
    }
   }
  }
 }
 $21 = ((($rect)) + 8|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ($22|0)<(1);
 if ($23) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $24 = ((($rect)) + 12|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = ($25|0)<(1);
 if ($26) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $27 = ((($1)) + 28|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ((($3)) + 4|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = HEAP32[$3>>2]|0;
 FUNCTION_TABLE_vii[$28 & 8191]($30,$31);
 $32 = HEAP32[$29>>2]|0;
 $33 = HEAP32[$rect>>2]|0;
 $34 = ((($rect)) + 4|0);
 $35 = HEAP32[$34>>2]|0;
 $36 = HEAP32[$21>>2]|0;
 $37 = HEAP32[$24>>2]|0;
 $38 = ((($3)) + 8|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ((($3)) + 12|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = ((($texture)) + 4|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = ($43|0)==(0);
 $$mask = $43 & -268435456;
 $45 = ($$mask|0)==(268435456);
 $or$cond = $44 | $45;
 L16: do {
  if ($or$cond) {
   $47 = $43 & 255;
   $49 = $47;
  } else {
   $switch$split2D = ($43|0)<(1498831189);
   if ($switch$split2D) {
    switch ($43|0) {
    case 844715353:  {
     $49 = 2;
     break L16;
     break;
    }
    default: {
    }
    }
   } else {
    switch ($43|0) {
    case 1498831189:  {
     $49 = 2;
     break L16;
     break;
    }
    default: {
    }
    }
   }
   $46 = ($43|0)==(1431918169);
   $phitmp = $46 ? 2 : 1;
   $49 = $phitmp;
  }
 } while(0);
 $48 = Math_imul($49, $36)|0;
 $50 = ($48|0)==($pitch|0);
 do {
  if ($50) {
   $blob$0$i = 0;$src$1$i = $pixels;
   label = 18;
  } else {
   $51 = Math_imul($48, $37)|0;
   $52 = (_SDL_malloc($51)|0);
   $53 = ($52|0)==(0|0);
   if ($53) {
    (_SDL_Error(0)|0);
    break;
   }
   $54 = ($37|0)>(0);
   if ($54) {
    $$011$i = $pixels;$src$02$i = $52;$y$03$i = 0;
    while(1) {
     (_SDL_memcpy($src$02$i,$$011$i,$48)|0);
     $55 = (($src$02$i) + ($48)|0);
     $56 = (($$011$i) + ($pitch)|0);
     $57 = (($y$03$i) + 1)|0;
     $exitcond$i = ($57|0)==($37|0);
     if ($exitcond$i) {
      $blob$0$i = $52;$src$1$i = $52;
      label = 18;
      break;
     } else {
      $$011$i = $56;$src$02$i = $55;$y$03$i = $57;
     }
    }
   } else {
    $blob$0$i = $52;$src$1$i = $52;
    label = 18;
   }
  }
 } while(0);
 if ((label|0) == 18) {
  $58 = ((($1)) + 164|0);
  $59 = HEAP32[$58>>2]|0;
  FUNCTION_TABLE_viiiiiiiii[$59 & 8191]($32,0,$33,$35,$36,$37,$39,$41,$src$1$i);
  $60 = ($blob$0$i|0)==(0|0);
  if (!($60)) {
   _SDL_free($blob$0$i);
  }
 }
 $61 = ((($3)) + 24|0);
 $62 = HEAP32[$61>>2]|0;
 $63 = ($62|0)==(0);
 do {
  if ($63) {
   $$01 = $pixels;
  } else {
   $64 = HEAP32[$24>>2]|0;
   $65 = Math_imul($64, $pitch)|0;
   $66 = (($pixels) + ($65)|0);
   $67 = HEAP32[$42>>2]|0;
   $68 = ($67|0)==(842094169);
   $69 = HEAP32[$27>>2]|0;
   $70 = HEAP32[$29>>2]|0;
   if ($68) {
    $71 = ((($3)) + 32|0);
    $72 = HEAP32[$71>>2]|0;
    FUNCTION_TABLE_vii[$69 & 8191]($70,$72);
   } else {
    $73 = ((($3)) + 36|0);
    $74 = HEAP32[$73>>2]|0;
    FUNCTION_TABLE_vii[$69 & 8191]($70,$74);
   }
   $75 = HEAP32[$29>>2]|0;
   $76 = HEAP32[$rect>>2]|0;
   $77 = (($76|0) / 2)&-1;
   $78 = HEAP32[$34>>2]|0;
   $79 = (($78|0) / 2)&-1;
   $80 = HEAP32[$21>>2]|0;
   $81 = (($80|0) / 2)&-1;
   $82 = HEAP32[$24>>2]|0;
   $83 = (($82|0) / 2)&-1;
   $84 = HEAP32[$38>>2]|0;
   $85 = HEAP32[$40>>2]|0;
   $86 = (($pitch|0) / 2)&-1;
   $87 = ($81|0)==($86|0);
   do {
    if ($87) {
     $blob$0$i9 = 0;$src$1$i10 = $66;
     label = 29;
    } else {
     $88 = Math_imul($83, $81)|0;
     $89 = (_SDL_malloc($88)|0);
     $90 = ($89|0)==(0|0);
     if ($90) {
      (_SDL_Error(0)|0);
      break;
     }
     $91 = ($82|0)>(1);
     if ($91) {
      $$011$i6 = $66;$src$02$i5 = $89;$y$03$i4 = 0;
      while(1) {
       (_SDL_memcpy($src$02$i5,$$011$i6,$81)|0);
       $92 = (($src$02$i5) + ($81)|0);
       $93 = (($$011$i6) + ($86)|0);
       $94 = (($y$03$i4) + 1)|0;
       $exitcond$i7 = ($94|0)==($83|0);
       if ($exitcond$i7) {
        $blob$0$i9 = $89;$src$1$i10 = $89;
        label = 29;
        break;
       } else {
        $$011$i6 = $93;$src$02$i5 = $92;$y$03$i4 = $94;
       }
      }
     } else {
      $blob$0$i9 = $89;$src$1$i10 = $89;
      label = 29;
     }
    }
   } while(0);
   if ((label|0) == 29) {
    $95 = ((($1)) + 164|0);
    $96 = HEAP32[$95>>2]|0;
    FUNCTION_TABLE_viiiiiiiii[$96 & 8191]($75,0,$77,$79,$81,$83,$84,$85,$src$1$i10);
    $97 = ($blob$0$i9|0)==(0|0);
    if (!($97)) {
     _SDL_free($blob$0$i9);
    }
   }
   $98 = HEAP32[$24>>2]|0;
   $99 = Math_imul($98, $pitch)|0;
   $100 = (($99|0) / 4)&-1;
   $101 = (($66) + ($100)|0);
   $102 = HEAP32[$42>>2]|0;
   $103 = ($102|0)==(842094169);
   $104 = HEAP32[$27>>2]|0;
   $105 = HEAP32[$29>>2]|0;
   if ($103) {
    $106 = ((($3)) + 36|0);
    $107 = HEAP32[$106>>2]|0;
    FUNCTION_TABLE_vii[$104 & 8191]($105,$107);
   } else {
    $108 = ((($3)) + 32|0);
    $109 = HEAP32[$108>>2]|0;
    FUNCTION_TABLE_vii[$104 & 8191]($105,$109);
   }
   $110 = HEAP32[$29>>2]|0;
   $111 = HEAP32[$rect>>2]|0;
   $112 = (($111|0) / 2)&-1;
   $113 = HEAP32[$34>>2]|0;
   $114 = (($113|0) / 2)&-1;
   $115 = HEAP32[$21>>2]|0;
   $116 = (($115|0) / 2)&-1;
   $117 = HEAP32[$24>>2]|0;
   $118 = (($117|0) / 2)&-1;
   $119 = HEAP32[$38>>2]|0;
   $120 = HEAP32[$40>>2]|0;
   $121 = ($116|0)==($86|0);
   if ($121) {
    $blob$0$i19 = 0;$src$1$i20 = $101;
   } else {
    $122 = Math_imul($118, $116)|0;
    $123 = (_SDL_malloc($122)|0);
    $124 = ($123|0)==(0|0);
    if ($124) {
     (_SDL_Error(0)|0);
     $$01 = $101;
     break;
    }
    $125 = ($117|0)>(1);
    if ($125) {
     $$011$i16 = $101;$src$02$i15 = $123;$y$03$i14 = 0;
     while(1) {
      (_SDL_memcpy($src$02$i15,$$011$i16,$116)|0);
      $126 = (($src$02$i15) + ($116)|0);
      $127 = (($$011$i16) + ($86)|0);
      $128 = (($y$03$i14) + 1)|0;
      $exitcond$i17 = ($128|0)==($118|0);
      if ($exitcond$i17) {
       $blob$0$i19 = $123;$src$1$i20 = $123;
       break;
      } else {
       $$011$i16 = $127;$src$02$i15 = $126;$y$03$i14 = $128;
      }
     }
    } else {
     $blob$0$i19 = $123;$src$1$i20 = $123;
    }
   }
   $129 = ((($1)) + 164|0);
   $130 = HEAP32[$129>>2]|0;
   FUNCTION_TABLE_viiiiiiiii[$130 & 8191]($110,0,$112,$114,$116,$118,$119,$120,$src$1$i20);
   $131 = ($blob$0$i19|0)==(0|0);
   if ($131) {
    $$01 = $101;
   } else {
    _SDL_free($blob$0$i19);
    $$01 = $101;
   }
  }
 } while(0);
 $132 = ((($3)) + 28|0);
 $133 = HEAP32[$132>>2]|0;
 $134 = ($133|0)==(0);
 do {
  if (!($134)) {
   $135 = HEAP32[$24>>2]|0;
   $136 = Math_imul($135, $pitch)|0;
   $137 = (($$01) + ($136)|0);
   $138 = HEAP32[$27>>2]|0;
   $139 = HEAP32[$29>>2]|0;
   $140 = ((($3)) + 36|0);
   $141 = HEAP32[$140>>2]|0;
   FUNCTION_TABLE_vii[$138 & 8191]($139,$141);
   $142 = HEAP32[$29>>2]|0;
   $143 = HEAP32[$rect>>2]|0;
   $144 = (($143|0) / 2)&-1;
   $145 = HEAP32[$34>>2]|0;
   $146 = (($145|0) / 2)&-1;
   $147 = HEAP32[$21>>2]|0;
   $148 = (($147|0) / 2)&-1;
   $149 = HEAP32[$24>>2]|0;
   $150 = (($149|0) / 2)&-1;
   $151 = $148 << 1;
   $152 = ($151|0)==($pitch|0);
   if ($152) {
    $blob$0$i29 = 0;$src$1$i30 = $137;
   } else {
    $153 = Math_imul($151, $150)|0;
    $154 = (_SDL_malloc($153)|0);
    $155 = ($154|0)==(0|0);
    if ($155) {
     (_SDL_Error(0)|0);
     break;
    }
    $156 = ($149|0)>(1);
    if ($156) {
     $$011$i26 = $137;$src$02$i25 = $154;$y$03$i24 = 0;
     while(1) {
      (_SDL_memcpy($src$02$i25,$$011$i26,$151)|0);
      $157 = (($src$02$i25) + ($151)|0);
      $158 = (($$011$i26) + ($pitch)|0);
      $159 = (($y$03$i24) + 1)|0;
      $exitcond$i27 = ($159|0)==($150|0);
      if ($exitcond$i27) {
       $blob$0$i29 = $154;$src$1$i30 = $154;
       break;
      } else {
       $$011$i26 = $158;$src$02$i25 = $157;$y$03$i24 = $159;
      }
     }
    } else {
     $blob$0$i29 = $154;$src$1$i30 = $154;
    }
   }
   $160 = ((($1)) + 164|0);
   $161 = HEAP32[$160>>2]|0;
   FUNCTION_TABLE_viiiiiiiii[$161 & 8191]($142,0,$144,$146,$148,$150,6410,5121,$src$1$i30);
   $162 = ($blob$0$i29|0)==(0|0);
   if (!($162)) {
    _SDL_free($blob$0$i29);
   }
  }
 } while(0);
 $163 = HEAP32[$0>>2]|0;
 $164 = ((($163)) + 4|0);
 $165 = HEAP32[$164>>2]|0;
 $166 = ($165|0)==(0);
 if ($166) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $167 = ((($163)) + 108|0);
 $168 = HEAP32[$167>>2]|0;
 $169 = (FUNCTION_TABLE_i[$168 & 8191]()|0);
 $170 = ($169|0)==(0);
 if ($170) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $171 = $169;
 }
 while(1) {
  switch ($171|0) {
  case 1280:  {
   $$0$i$i = 597787;
   break;
  }
  case 1281:  {
   $$0$i$i = 597770;
   break;
  }
  case 1282:  {
   $$0$i$i = 597749;
   break;
  }
  case 1285:  {
   $$0$i$i = 597732;
   break;
  }
  default: {
   $$0$i$i = 597724;
  }
  }
  HEAP32[$vararg_buffer>>2] = 598240;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = 597829;
  $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
  HEAP32[$vararg_ptr2>>2] = 753;
  $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
  HEAP32[$vararg_ptr3>>2] = 598258;
  $vararg_ptr4 = ((($vararg_buffer)) + 16|0);
  HEAP32[$vararg_ptr4>>2] = $$0$i$i;
  $vararg_ptr5 = ((($vararg_buffer)) + 20|0);
  HEAP32[$vararg_ptr5>>2] = $171;
  (_SDL_SetError(597803,$vararg_buffer)|0);
  $172 = HEAP32[$167>>2]|0;
  $173 = (FUNCTION_TABLE_i[$172 & 8191]()|0);
  $174 = ($173|0)==(0);
  if ($174) {
   $$0 = -1;
   break;
  } else {
   $171 = $173;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _GLES2_UpdateTextureYUV($renderer,$texture,$rect,$Yplane,$Ypitch,$Uplane,$Upitch,$Vplane,$Vpitch) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 $rect = $rect|0;
 $Yplane = $Yplane|0;
 $Ypitch = $Ypitch|0;
 $Uplane = $Uplane|0;
 $Upitch = $Upitch|0;
 $Vplane = $Vplane|0;
 $Vpitch = $Vpitch|0;
 var $$0 = 0, $$0$i$i = 0, $$011$i = 0, $$011$i14 = 0, $$011$i4 = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $blob$0$i = 0, $blob$0$i17 = 0, $blob$0$i7 = 0, $exitcond$i = 0, $exitcond$i15 = 0, $exitcond$i5 = 0, $src$02$i = 0, $src$02$i13 = 0, $src$02$i3 = 0, $src$1$i = 0, $src$1$i18 = 0, $src$1$i8 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0, $vararg_ptr4 = 0;
 var $vararg_ptr5 = 0, $y$03$i = 0, $y$03$i12 = 0, $y$03$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($texture)) + 68|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[300738]|0;
 $5 = HEAP32[$1>>2]|0;
 $6 = ($4|0)==($5|0);
 if ($6) {
  $14 = $1;
  label = 4;
 } else {
  $7 = ((($1)) + 268|0);
  HEAP32[$7>>2] = 0;
  $8 = ((($renderer)) + 188|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = (_SDL_GL_MakeCurrent($9,$5)|0);
  $11 = ($10|0)<(0);
  if (!($11)) {
   $12 = HEAP32[$1>>2]|0;
   HEAP32[300738] = $12;
   (_GLES2_UpdateViewport($renderer)|0);
   $$pre$i = HEAP32[$0>>2]|0;
   $14 = $$pre$i;
   label = 4;
  }
 }
 if ((label|0) == 4) {
  $13 = ((($14)) + 4|0);
  $15 = HEAP32[$13>>2]|0;
  $16 = ($15|0)==(0);
  if (!($16)) {
   $17 = ((($14)) + 108|0);
   while(1) {
    $18 = HEAP32[$17>>2]|0;
    $19 = (FUNCTION_TABLE_i[$18 & 8191]()|0);
    $20 = ($19|0)==(0);
    if ($20) {
     break;
    }
   }
  }
 }
 $21 = ((($rect)) + 8|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ($22|0)<(1);
 if ($23) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $24 = ((($rect)) + 12|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = ($25|0)<(1);
 if ($26) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $27 = ((($1)) + 28|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ((($3)) + 4|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ((($3)) + 32|0);
 $32 = HEAP32[$31>>2]|0;
 FUNCTION_TABLE_vii[$28 & 8191]($30,$32);
 $33 = HEAP32[$29>>2]|0;
 $34 = HEAP32[$rect>>2]|0;
 $35 = (($34|0) / 2)&-1;
 $36 = ((($rect)) + 4|0);
 $37 = HEAP32[$36>>2]|0;
 $38 = (($37|0) / 2)&-1;
 $39 = HEAP32[$21>>2]|0;
 $40 = (($39|0) / 2)&-1;
 $41 = HEAP32[$24>>2]|0;
 $42 = (($41|0) / 2)&-1;
 $43 = ((($3)) + 8|0);
 $44 = HEAP32[$43>>2]|0;
 $45 = ((($3)) + 12|0);
 $46 = HEAP32[$45>>2]|0;
 $47 = ($40|0)==($Vpitch|0);
 do {
  if ($47) {
   $blob$0$i = 0;$src$1$i = $Vplane;
   label = 14;
  } else {
   $48 = Math_imul($42, $40)|0;
   $49 = (_SDL_malloc($48)|0);
   $50 = ($49|0)==(0|0);
   if ($50) {
    (_SDL_Error(0)|0);
    break;
   }
   $51 = ($41|0)>(1);
   if ($51) {
    $$011$i = $Vplane;$src$02$i = $49;$y$03$i = 0;
    while(1) {
     (_SDL_memcpy($src$02$i,$$011$i,$40)|0);
     $52 = (($src$02$i) + ($40)|0);
     $53 = (($$011$i) + ($Vpitch)|0);
     $54 = (($y$03$i) + 1)|0;
     $exitcond$i = ($54|0)==($42|0);
     if ($exitcond$i) {
      $blob$0$i = $49;$src$1$i = $49;
      label = 14;
      break;
     } else {
      $$011$i = $53;$src$02$i = $52;$y$03$i = $54;
     }
    }
   } else {
    $blob$0$i = $49;$src$1$i = $49;
    label = 14;
   }
  }
 } while(0);
 if ((label|0) == 14) {
  $55 = ((($1)) + 164|0);
  $56 = HEAP32[$55>>2]|0;
  FUNCTION_TABLE_viiiiiiiii[$56 & 8191]($33,0,$35,$38,$40,$42,$44,$46,$src$1$i);
  $57 = ($blob$0$i|0)==(0|0);
  if (!($57)) {
   _SDL_free($blob$0$i);
  }
 }
 $58 = HEAP32[$27>>2]|0;
 $59 = HEAP32[$29>>2]|0;
 $60 = ((($3)) + 36|0);
 $61 = HEAP32[$60>>2]|0;
 FUNCTION_TABLE_vii[$58 & 8191]($59,$61);
 $62 = HEAP32[$29>>2]|0;
 $63 = HEAP32[$rect>>2]|0;
 $64 = (($63|0) / 2)&-1;
 $65 = HEAP32[$36>>2]|0;
 $66 = (($65|0) / 2)&-1;
 $67 = HEAP32[$21>>2]|0;
 $68 = (($67|0) / 2)&-1;
 $69 = HEAP32[$24>>2]|0;
 $70 = (($69|0) / 2)&-1;
 $71 = HEAP32[$43>>2]|0;
 $72 = HEAP32[$45>>2]|0;
 $73 = ($68|0)==($Upitch|0);
 do {
  if ($73) {
   $blob$0$i7 = 0;$src$1$i8 = $Uplane;
   label = 21;
  } else {
   $74 = Math_imul($70, $68)|0;
   $75 = (_SDL_malloc($74)|0);
   $76 = ($75|0)==(0|0);
   if ($76) {
    (_SDL_Error(0)|0);
    break;
   }
   $77 = ($69|0)>(1);
   if ($77) {
    $$011$i4 = $Uplane;$src$02$i3 = $75;$y$03$i2 = 0;
    while(1) {
     (_SDL_memcpy($src$02$i3,$$011$i4,$68)|0);
     $78 = (($src$02$i3) + ($68)|0);
     $79 = (($$011$i4) + ($Upitch)|0);
     $80 = (($y$03$i2) + 1)|0;
     $exitcond$i5 = ($80|0)==($70|0);
     if ($exitcond$i5) {
      $blob$0$i7 = $75;$src$1$i8 = $75;
      label = 21;
      break;
     } else {
      $$011$i4 = $79;$src$02$i3 = $78;$y$03$i2 = $80;
     }
    }
   } else {
    $blob$0$i7 = $75;$src$1$i8 = $75;
    label = 21;
   }
  }
 } while(0);
 if ((label|0) == 21) {
  $81 = ((($1)) + 164|0);
  $82 = HEAP32[$81>>2]|0;
  FUNCTION_TABLE_viiiiiiiii[$82 & 8191]($62,0,$64,$66,$68,$70,$71,$72,$src$1$i8);
  $83 = ($blob$0$i7|0)==(0|0);
  if (!($83)) {
   _SDL_free($blob$0$i7);
  }
 }
 $84 = HEAP32[$27>>2]|0;
 $85 = HEAP32[$29>>2]|0;
 $86 = HEAP32[$3>>2]|0;
 FUNCTION_TABLE_vii[$84 & 8191]($85,$86);
 $87 = HEAP32[$29>>2]|0;
 $88 = HEAP32[$rect>>2]|0;
 $89 = HEAP32[$36>>2]|0;
 $90 = HEAP32[$21>>2]|0;
 $91 = HEAP32[$24>>2]|0;
 $92 = HEAP32[$43>>2]|0;
 $93 = HEAP32[$45>>2]|0;
 $94 = ($90|0)==($Ypitch|0);
 do {
  if ($94) {
   $blob$0$i17 = 0;$src$1$i18 = $Yplane;
   label = 28;
  } else {
   $95 = Math_imul($91, $90)|0;
   $96 = (_SDL_malloc($95)|0);
   $97 = ($96|0)==(0|0);
   if ($97) {
    (_SDL_Error(0)|0);
    break;
   }
   $98 = ($91|0)>(0);
   if ($98) {
    $$011$i14 = $Yplane;$src$02$i13 = $96;$y$03$i12 = 0;
    while(1) {
     (_SDL_memcpy($src$02$i13,$$011$i14,$90)|0);
     $99 = (($src$02$i13) + ($90)|0);
     $100 = (($$011$i14) + ($Ypitch)|0);
     $101 = (($y$03$i12) + 1)|0;
     $exitcond$i15 = ($101|0)==($91|0);
     if ($exitcond$i15) {
      $blob$0$i17 = $96;$src$1$i18 = $96;
      label = 28;
      break;
     } else {
      $$011$i14 = $100;$src$02$i13 = $99;$y$03$i12 = $101;
     }
    }
   } else {
    $blob$0$i17 = $96;$src$1$i18 = $96;
    label = 28;
   }
  }
 } while(0);
 if ((label|0) == 28) {
  $102 = ((($1)) + 164|0);
  $103 = HEAP32[$102>>2]|0;
  FUNCTION_TABLE_viiiiiiiii[$103 & 8191]($87,0,$88,$89,$90,$91,$92,$93,$src$1$i18);
  $104 = ($blob$0$i17|0)==(0|0);
  if (!($104)) {
   _SDL_free($blob$0$i17);
  }
 }
 $105 = HEAP32[$0>>2]|0;
 $106 = ((($105)) + 4|0);
 $107 = HEAP32[$106>>2]|0;
 $108 = ($107|0)==(0);
 if ($108) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $109 = ((($105)) + 108|0);
 $110 = HEAP32[$109>>2]|0;
 $111 = (FUNCTION_TABLE_i[$110 & 8191]()|0);
 $112 = ($111|0)==(0);
 if ($112) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $113 = $111;
 }
 while(1) {
  switch ($113|0) {
  case 1280:  {
   $$0$i$i = 597787;
   break;
  }
  case 1281:  {
   $$0$i$i = 597770;
   break;
  }
  case 1282:  {
   $$0$i$i = 597749;
   break;
  }
  case 1285:  {
   $$0$i$i = 597732;
   break;
  }
  default: {
   $$0$i$i = 597724;
  }
  }
  HEAP32[$vararg_buffer>>2] = 598240;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = 597829;
  $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
  HEAP32[$vararg_ptr2>>2] = 803;
  $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
  HEAP32[$vararg_ptr3>>2] = 598350;
  $vararg_ptr4 = ((($vararg_buffer)) + 16|0);
  HEAP32[$vararg_ptr4>>2] = $$0$i$i;
  $vararg_ptr5 = ((($vararg_buffer)) + 20|0);
  HEAP32[$vararg_ptr5>>2] = $113;
  (_SDL_SetError(597803,$vararg_buffer)|0);
  $114 = HEAP32[$109>>2]|0;
  $115 = (FUNCTION_TABLE_i[$114 & 8191]()|0);
  $116 = ($115|0)==(0);
  if ($116) {
   $$0 = -1;
   break;
  } else {
   $113 = $115;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _GLES2_LockTexture($renderer,$texture,$rect,$pixels,$pitch) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 $rect = $rect|0;
 $pixels = $pixels|0;
 $pitch = $pitch|0;
 var $$mask = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $or$cond = 0, $phitmp = 0, $switch$split2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($texture)) + 68|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 20|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($rect)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = Math_imul($7, $5)|0;
 $9 = (($3) + ($8)|0);
 $10 = HEAP32[$rect>>2]|0;
 $11 = ((($texture)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)==(0);
 $$mask = $12 & -268435456;
 $14 = ($$mask|0)==(268435456);
 $or$cond = $13 | $14;
 L1: do {
  if ($or$cond) {
   $16 = $12 & 255;
   $18 = $16;
  } else {
   $switch$split2D = ($12|0)<(1498831189);
   if ($switch$split2D) {
    switch ($12|0) {
    case 844715353:  {
     $18 = 2;
     break L1;
     break;
    }
    default: {
    }
    }
   } else {
    switch ($12|0) {
    case 1498831189:  {
     $18 = 2;
     break L1;
     break;
    }
    default: {
    }
    }
   }
   $15 = ($12|0)==(1431918169);
   $phitmp = $15 ? 2 : 1;
   $18 = $phitmp;
  }
 } while(0);
 $17 = Math_imul($18, $10)|0;
 $19 = (($9) + ($17)|0);
 HEAP32[$pixels>>2] = $19;
 HEAP32[$pitch>>2] = $5;
 return 0;
}
function _GLES2_UnlockTexture($renderer,$texture) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $rect = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $rect = sp;
 $0 = ((($texture)) + 68|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$rect>>2] = 0;
 $2 = ((($rect)) + 4|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($texture)) + 12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($rect)) + 8|0);
 HEAP32[$5>>2] = $4;
 $6 = ((($texture)) + 16|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($rect)) + 12|0);
 HEAP32[$8>>2] = $7;
 $9 = ((($1)) + 16|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($1)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 (_GLES2_UpdateTexture($renderer,$texture,$rect,$10,$12)|0);
 STACKTOP = sp;return;
}
function _GLES2_SetRenderTarget($renderer,$texture) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($texture|0)==(0|0);
 if ($2) {
  $3 = ((($1)) + 192|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ((($1)) + 236|0);
  $6 = HEAP32[$5>>2]|0;
  FUNCTION_TABLE_vii[$4 & 8191](36160,$6);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $7 = ((($texture)) + 68|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($1)) + 192|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($8)) + 40|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($12)) + 8|0);
 $14 = HEAP32[$13>>2]|0;
 FUNCTION_TABLE_vii[$10 & 8191](36160,$14);
 $15 = ((($1)) + 196|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($8)) + 4|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = HEAP32[$8>>2]|0;
 FUNCTION_TABLE_viiiii[$16 & 8191](36160,36064,$18,$19,0);
 $20 = ((($1)) + 200|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = (FUNCTION_TABLE_ii[$21 & 8191](36160)|0);
 $23 = ($22|0)==(36053);
 if ($23) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $24 = (_SDL_SetError(598486,$vararg_buffer)|0);
 $$0 = $24;
 STACKTOP = sp;return ($$0|0);
}
function _GLES2_UpdateClipRect($renderer) {
 $renderer = $renderer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $h = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $w = sp + 4|0;
 $h = sp;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[300738]|0;
 $3 = HEAP32[$1>>2]|0;
 $4 = ($2|0)==($3|0);
 if (!($4)) {
  STACKTOP = sp;return 0;
 }
 $5 = ((($renderer)) + 276|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if ($7) {
  $47 = ((($1)) + 68|0);
  $48 = HEAP32[$47>>2]|0;
  FUNCTION_TABLE_vi[$48 & 8191](3089);
  STACKTOP = sp;return 0;
 }
 $8 = ((($renderer)) + 244|0);
 $9 = ((($1)) + 80|0);
 $10 = HEAP32[$9>>2]|0;
 FUNCTION_TABLE_vi[$10 & 8191](3089);
 $11 = ((($renderer)) + 304|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)==(0|0);
 if ($13) {
  (_SDL_GetRendererOutputSize($renderer,$w,$h)|0);
  $29 = ((($1)) + 144|0);
  $30 = HEAP32[$29>>2]|0;
  $31 = ((($renderer)) + 212|0);
  $32 = HEAP32[$31>>2]|0;
  $33 = HEAP32[$8>>2]|0;
  $34 = (($33) + ($32))|0;
  $35 = HEAP32[$h>>2]|0;
  $36 = ((($renderer)) + 216|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = (($35) - ($37))|0;
  $39 = ((($renderer)) + 248|0);
  $40 = HEAP32[$39>>2]|0;
  $41 = (($38) - ($40))|0;
  $42 = ((($renderer)) + 256|0);
  $43 = HEAP32[$42>>2]|0;
  $44 = (($41) - ($43))|0;
  $45 = ((($renderer)) + 252|0);
  $46 = HEAP32[$45>>2]|0;
  FUNCTION_TABLE_viiii[$30 & 8191]($34,$44,$46,$43);
  STACKTOP = sp;return 0;
 } else {
  $14 = ((($1)) + 144|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ((($renderer)) + 212|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = HEAP32[$8>>2]|0;
  $19 = (($18) + ($17))|0;
  $20 = ((($renderer)) + 216|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = ((($renderer)) + 248|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = (($23) + ($21))|0;
  $25 = ((($renderer)) + 252|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = ((($renderer)) + 256|0);
  $28 = HEAP32[$27>>2]|0;
  FUNCTION_TABLE_viiii[$15 & 8191]($19,$24,$26,$28);
  STACKTOP = sp;return 0;
 }
 return (0)|0;
}
function _GLES2_RenderClear($renderer) {
 $renderer = $renderer|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $8 = 0, $9 = 0, $b$0 = 0, $b$0$pre = 0, $r$0 = 0, $switch$split2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[300738]|0;
 $3 = HEAP32[$1>>2]|0;
 $4 = ($2|0)==($3|0);
 if ($4) {
  $12 = $1;
  label = 4;
 } else {
  $5 = ((($1)) + 268|0);
  HEAP32[$5>>2] = 0;
  $6 = ((($renderer)) + 188|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (_SDL_GL_MakeCurrent($7,$3)|0);
  $9 = ($8|0)<(0);
  if (!($9)) {
   $10 = HEAP32[$1>>2]|0;
   HEAP32[300738] = $10;
   (_GLES2_UpdateViewport($renderer)|0);
   $$pre$i = HEAP32[$0>>2]|0;
   $12 = $$pre$i;
   label = 4;
  }
 }
 if ((label|0) == 4) {
  $11 = ((($12)) + 4|0);
  $13 = HEAP32[$11>>2]|0;
  $14 = ($13|0)==(0);
  if (!($14)) {
   $15 = ((($12)) + 108|0);
   while(1) {
    $16 = HEAP32[$15>>2]|0;
    $17 = (FUNCTION_TABLE_i[$16 & 8191]()|0);
    $18 = ($17|0)==(0);
    if ($18) {
     break;
    }
   }
  }
 }
 $19 = ((($1)) + 272|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ((($1)) + 273|0);
 $22 = ((($1)) + 274|0);
 $23 = ((($1)) + 275|0);
 $24 = ((($renderer)) + 308|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = $25&65535;
 $27 = ($26&65535) >>> 8;
 $28 = $27&255;
 $29 = ((($renderer)) + 310|0);
 $30 = $25 >>> 16;
 $31 = $30&255;
 $32 = $25 >>> 24;
 $33 = $32&255;
 $34 = (_llvm_bswap_i32(($20|0))|0);
 $35 = $25 << 24;
 $36 = $27&65535;
 $37 = $36 << 16;
 $38 = $37 | $35;
 $39 = $30 << 8;
 $40 = $39 & 65280;
 $41 = $38 | $40;
 $42 = $41 | $32;
 $43 = ($34|0)==($42|0);
 $44 = $25&255;
 if ($43) {
  $69 = ((($1)) + 36|0);
  $70 = HEAP32[$69>>2]|0;
  FUNCTION_TABLE_vi[$70 & 8191](16384);
  return 0;
 }
 $45 = ((($renderer)) + 304|0);
 $46 = HEAP32[$45>>2]|0;
 $47 = ($46|0)==(0|0);
 L13: do {
  if ($47) {
   label = 10;
  } else {
   $48 = ((($46)) + 4|0);
   $49 = HEAP32[$48>>2]|0;
   $switch$split2D = ($49|0)<(372645892);
   if ($switch$split2D) {
    switch ($49|0) {
    case 370546692:  {
     $b$0 = $44;$r$0 = $31;
     break L13;
     break;
    }
    default: {
     label = 10;
     break L13;
    }
    }
   } else {
    switch ($49|0) {
    case 372645892:  {
     $b$0 = $44;$r$0 = $31;
     break L13;
     break;
    }
    default: {
     label = 10;
     break L13;
    }
    }
   }
  }
 } while(0);
 if ((label|0) == 10) {
  $b$0$pre = HEAP8[$29>>0]|0;
  $b$0 = $b$0$pre;$r$0 = $44;
 }
 $50 = ((($1)) + 40|0);
 $51 = HEAP32[$50>>2]|0;
 $52 = (+($r$0&255));
 $53 = $52 * 0.0039215688593685627;
 $54 = (+($28&255));
 $55 = $54 * 0.0039215688593685627;
 $56 = (+($b$0&255));
 $57 = $56 * 0.0039215688593685627;
 $58 = (+($33&255));
 $59 = $58 * 0.0039215688593685627;
 FUNCTION_TABLE_vdddd[$51 & 8191]($53,$55,$57,$59);
 $60 = HEAP32[$24>>2]|0;
 $61 = $60&65535;
 $62 = $60&255;
 HEAP8[$19>>0] = $62;
 $63 = ($61&65535) >>> 8;
 $64 = $63&255;
 HEAP8[$21>>0] = $64;
 $65 = $60 >>> 16;
 $66 = $65&255;
 HEAP8[$22>>0] = $66;
 $67 = $60 >>> 24;
 $68 = $67&255;
 HEAP8[$23>>0] = $68;
 $69 = ((($1)) + 36|0);
 $70 = HEAP32[$69>>2]|0;
 FUNCTION_TABLE_vi[$70 & 8191](16384);
 return 0;
}
function _GLES2_RenderDrawPoints($renderer,$points,$count) {
 $renderer = $renderer|0;
 $points = $points|0;
 $count = $count|0;
 var $$0 = 0, $$alloca_mul = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0;
 var $9 = 0.0, $exitcond = 0, $idx$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_GLES2_SetDrawingState($renderer)|0);
 $3 = ($2|0)<(0);
 if ($3) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = $count << 3;
 $$alloca_mul = $4;
 $5 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $6 = ($count|0)>(0);
 if ($6) {
  $idx$01 = 0;
  while(1) {
   $7 = (($points) + ($idx$01<<3)|0);
   $8 = +HEAPF32[$7>>2];
   $9 = $8 + 0.5;
   $10 = (((($points) + ($idx$01<<3)|0)) + 4|0);
   $11 = +HEAPF32[$10>>2];
   $12 = $11 + 0.5;
   $13 = $idx$01 << 1;
   $14 = (($5) + ($13<<2)|0);
   HEAPF32[$14>>2] = $9;
   $15 = $13 | 1;
   $16 = (($5) + ($15<<2)|0);
   HEAPF32[$16>>2] = $12;
   $17 = (($idx$01) + 1)|0;
   $exitcond = ($17|0)==($count|0);
   if ($exitcond) {
    break;
   } else {
    $idx$01 = $17;
   }
  }
 }
 $18 = HEAP32[$0>>2]|0;
 $19 = ((($18)) + 276|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ($20|0)==(0);
 if ($21) {
  $22 = ((($18)) + 216|0);
  $23 = HEAP32[$22>>2]|0;
  FUNCTION_TABLE_vii[$23 & 8191](1,$19);
  $$pre$i = HEAP32[$19>>2]|0;
  $26 = $$pre$i;
 } else {
  $26 = $20;
 }
 $24 = ((($18)) + 220|0);
 $25 = HEAP32[$24>>2]|0;
 FUNCTION_TABLE_vii[$25 & 8191](34962,$26);
 $27 = ((($18)) + 292|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ($28>>>0)<($4>>>0);
 if ($29) {
  $30 = ((($18)) + 224|0);
  $31 = HEAP32[$30>>2]|0;
  FUNCTION_TABLE_viiii[$31 & 8191](34962,$4,$5,35040);
  HEAP32[$27>>2] = $4;
 } else {
  $32 = ((($18)) + 228|0);
  $33 = HEAP32[$32>>2]|0;
  FUNCTION_TABLE_viiii[$33 & 8191](34962,0,$4,$5);
 }
 $34 = ((($18)) + 184|0);
 $35 = HEAP32[$34>>2]|0;
 FUNCTION_TABLE_viiiiii[$35 & 8191](0,2,5126,0,0,0);
 $36 = ((($1)) + 76|0);
 $37 = HEAP32[$36>>2]|0;
 FUNCTION_TABLE_viii[$37 & 8191](0,0,$count);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _GLES2_SetDrawingState($renderer) {
 $renderer = $renderer|0;
 var $$0 = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0.0, $82 = 0.0, $83 = 0.0, $84 = 0.0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $9 = 0, $b$0 = 0, $b$0$in = 0, $r$0 = 0, $r$0$in = 0, $switch$split2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($renderer)) + 312|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[300738]|0;
 $5 = HEAP32[$1>>2]|0;
 $6 = ($4|0)==($5|0);
 if ($6) {
  $14 = $1;
  label = 4;
 } else {
  $7 = ((($1)) + 268|0);
  HEAP32[$7>>2] = 0;
  $8 = ((($renderer)) + 188|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = (_SDL_GL_MakeCurrent($9,$5)|0);
  $11 = ($10|0)<(0);
  if (!($11)) {
   $12 = HEAP32[$1>>2]|0;
   HEAP32[300738] = $12;
   (_GLES2_UpdateViewport($renderer)|0);
   $$pre$i = HEAP32[$0>>2]|0;
   $14 = $$pre$i;
   label = 4;
  }
 }
 if ((label|0) == 4) {
  $13 = ((($14)) + 4|0);
  $15 = HEAP32[$13>>2]|0;
  $16 = ($15|0)==(0);
  if (!($16)) {
   $17 = ((($14)) + 108|0);
   while(1) {
    $18 = HEAP32[$17>>2]|0;
    $19 = (FUNCTION_TABLE_i[$18 & 8191]()|0);
    $20 = ($19|0)==(0);
    if ($20) {
     break;
    }
   }
  }
 }
 $21 = ((($1)) + 8|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ($22|0)==($3|0);
 if (!($23)) {
  switch ($3|0) {
  case 4:  {
   $34 = ((($1)) + 80|0);
   $35 = HEAP32[$34>>2]|0;
   FUNCTION_TABLE_vi[$35 & 8191](3042);
   $36 = ((($1)) + 32|0);
   $37 = HEAP32[$36>>2]|0;
   FUNCTION_TABLE_viiii[$37 & 8191](0,768,0,1);
   break;
  }
  case 1:  {
   $26 = ((($1)) + 80|0);
   $27 = HEAP32[$26>>2]|0;
   FUNCTION_TABLE_vi[$27 & 8191](3042);
   $28 = ((($1)) + 32|0);
   $29 = HEAP32[$28>>2]|0;
   FUNCTION_TABLE_viiii[$29 & 8191](770,771,1,771);
   break;
  }
  case 2:  {
   $30 = ((($1)) + 80|0);
   $31 = HEAP32[$30>>2]|0;
   FUNCTION_TABLE_vi[$31 & 8191](3042);
   $32 = ((($1)) + 32|0);
   $33 = HEAP32[$32>>2]|0;
   FUNCTION_TABLE_viiii[$33 & 8191](770,1,0,1);
   break;
  }
  default: {
   $24 = ((($1)) + 68|0);
   $25 = HEAP32[$24>>2]|0;
   FUNCTION_TABLE_vi[$25 & 8191](3042);
  }
  }
  HEAP32[$21>>2] = $3;
 }
 $38 = ((($1)) + 12|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = ($39|0)==(0);
 if (!($40)) {
  $41 = ((($1)) + 72|0);
  $42 = HEAP32[$41>>2]|0;
  FUNCTION_TABLE_vi[$42 & 8191](1);
  HEAP32[$38>>2] = 0;
 }
 $43 = (_GLES2_SelectProgram($renderer,0,$3)|0);
 $44 = ($43|0)<(0);
 if ($44) {
  $$0 = -1;
  return ($$0|0);
 }
 $45 = ((($renderer)) + 309|0);
 $46 = HEAP8[$45>>0]|0;
 $47 = ((($renderer)) + 311|0);
 $48 = HEAP8[$47>>0]|0;
 $49 = ((($renderer)) + 304|0);
 $50 = HEAP32[$49>>2]|0;
 $51 = ($50|0)==(0|0);
 L25: do {
  if ($51) {
   label = 20;
  } else {
   $52 = ((($50)) + 4|0);
   $53 = HEAP32[$52>>2]|0;
   $switch$split2D = ($53|0)<(372645892);
   if ($switch$split2D) {
    switch ($53|0) {
    case 370546692:  {
     break;
    }
    default: {
     label = 20;
     break L25;
    }
    }
   } else {
    switch ($53|0) {
    case 372645892:  {
     break;
    }
    default: {
     label = 20;
     break L25;
    }
    }
   }
   $54 = ((($renderer)) + 310|0);
   $55 = ((($renderer)) + 308|0);
   $b$0$in = $55;$r$0$in = $54;
  }
 } while(0);
 if ((label|0) == 20) {
  $56 = ((($renderer)) + 308|0);
  $57 = ((($renderer)) + 310|0);
  $b$0$in = $57;$r$0$in = $56;
 }
 $b$0 = HEAP8[$b$0$in>>0]|0;
 $r$0 = HEAP8[$r$0$in>>0]|0;
 $58 = ((($1)) + 268|0);
 $59 = HEAP32[$58>>2]|0;
 $60 = ((($59)) + 80|0);
 $61 = HEAP32[$60>>2]|0;
 $62 = (_llvm_bswap_i32(($61|0))|0);
 $63 = $r$0&255;
 $64 = $63 << 24;
 $65 = $46&255;
 $66 = $65 << 16;
 $67 = $b$0&255;
 $68 = $67 << 8;
 $69 = $48&255;
 $70 = $69 | $66;
 $71 = $70 | $68;
 $72 = $71 | $64;
 $73 = ($62|0)==($72|0);
 if ($73) {
  $$0 = 0;
  return ($$0|0);
 }
 $74 = ((($59)) + 83|0);
 $75 = ((($59)) + 82|0);
 $76 = ((($59)) + 81|0);
 $77 = ((($1)) + 172|0);
 $78 = HEAP32[$77>>2]|0;
 $79 = ((($59)) + 28|0);
 $80 = HEAP32[$79>>2]|0;
 $81 = (+($r$0&255));
 $82 = $81 * 0.0039215688593685627;
 $83 = (+($46&255));
 $84 = $83 * 0.0039215688593685627;
 $85 = (+($b$0&255));
 $86 = $85 * 0.0039215688593685627;
 $87 = (+($48&255));
 $88 = $87 * 0.0039215688593685627;
 FUNCTION_TABLE_vidddd[$78 & 8191]($80,$82,$84,$86,$88);
 HEAP8[$60>>0] = $r$0;
 HEAP8[$76>>0] = $46;
 HEAP8[$75>>0] = $b$0;
 HEAP8[$74>>0] = $48;
 $$0 = 0;
 return ($$0|0);
}
function _GLES2_SelectProgram($renderer,$source,$blendMode) {
 $renderer = $renderer|0;
 $source = $source|0;
 $blendMode = $blendMode|0;
 var $$$i = 0.0, $$0 = 0, $$0$i$ph = 0, $$0$i6 = 0, $$1$i = 0.0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i1 = 0, $$pre$i1$i = 0, $$pre$i3 = 0, $$pre10$i = 0, $$pre8$i = 0, $$sink$i = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0;
 var $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0;
 var $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0;
 var $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0;
 var $212 = 0, $213 = 0, $214 = 0.0, $215 = 0.0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0.0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0;
 var $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0;
 var $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0;
 var $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0;
 var $285 = 0, $286 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $entry$0$i = 0, $entry$05$i = 0, $entry$06$i = 0, $entry$06$i$lcssa = 0, $fragment$011 = 0, $ftype$0 = 0, $linkSuccessful$i = 0, $projection$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $projection$i = sp + 8|0;
 $linkSuccessful$i = sp + 4|0;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 switch ($source|0) {
 case 0:  {
  $ftype$0 = 1;
  label = 9;
  break;
 }
 case 1:  {
  $ftype$0 = 2;
  label = 9;
  break;
 }
 case 2:  {
  $ftype$0 = 3;
  label = 9;
  break;
 }
 case 3:  {
  $ftype$0 = 5;
  label = 9;
  break;
 }
 case 4:  {
  $ftype$0 = 4;
  label = 9;
  break;
 }
 case 5:  {
  $ftype$0 = 6;
  label = 9;
  break;
 }
 case 6:  {
  $ftype$0 = 7;
  label = 9;
  break;
 }
 case 7:  {
  $ftype$0 = 8;
  label = 9;
  break;
 }
 default: {
 }
 }
 if ((label|0) == 9) {
  $2 = (_GLES2_CacheShader($renderer,0,$blendMode)|0);
  $3 = ($2|0)==(0|0);
  if (!($3)) {
   $4 = (_GLES2_CacheShader($renderer,$ftype$0,$blendMode)|0);
   $5 = ($4|0)==(0|0);
   L12: do {
    if ($5) {
     $fragment$011 = 0;
    } else {
     $6 = ((($1)) + 268|0);
     $7 = HEAP32[$6>>2]|0;
     $8 = ($7|0)==(0|0);
     if (!($8)) {
      $9 = ((($7)) + 8|0);
      $10 = HEAP32[$9>>2]|0;
      $11 = ($10|0)==($2|0);
      if ($11) {
       $12 = ((($7)) + 12|0);
       $13 = HEAP32[$12>>2]|0;
       $14 = ($13|0)==($4|0);
       if ($14) {
        $$0 = 0;
        STACKTOP = sp;return ($$0|0);
       }
      }
     }
     $15 = HEAP32[$0>>2]|0;
     $16 = ((($15)) + 260|0);
     $entry$05$i = HEAP32[$16>>2]|0;
     $17 = ($entry$05$i|0)==(0|0);
     $18 = $entry$05$i;
     L19: do {
      if ($17) {
       label = 24;
      } else {
       $entry$06$i = $entry$05$i;
       while(1) {
        $19 = ((($entry$06$i)) + 8|0);
        $20 = HEAP32[$19>>2]|0;
        $21 = ($20|0)==($2|0);
        if ($21) {
         $22 = ((($entry$06$i)) + 12|0);
         $23 = HEAP32[$22>>2]|0;
         $24 = ($23|0)==($4|0);
         if ($24) {
          $entry$06$i$lcssa = $entry$06$i;
          break;
         }
        }
        $25 = ((($entry$06$i)) + 156|0);
        $entry$0$i = HEAP32[$25>>2]|0;
        $26 = ($entry$0$i|0)==(0|0);
        if ($26) {
         label = 24;
         break L19;
        } else {
         $entry$06$i = $entry$0$i;
        }
       }
       $27 = ($entry$05$i|0)==($entry$06$i$lcssa|0);
       if ($27) {
        $$0$i6 = $entry$05$i;
        break;
       }
       $28 = ((($entry$06$i$lcssa)) + 156|0);
       $29 = HEAP32[$28>>2]|0;
       $30 = ($29|0)==(0|0);
       $31 = $29;
       $$pre10$i = ((($entry$06$i$lcssa)) + 152|0);
       if (!($30)) {
        $32 = HEAP32[$$pre10$i>>2]|0;
        $33 = ((($29)) + 152|0);
        HEAP32[$33>>2] = $32;
       }
       $34 = HEAP32[$$pre10$i>>2]|0;
       $35 = ($34|0)==(0|0);
       if (!($35)) {
        $36 = ((($34)) + 156|0);
        HEAP32[$36>>2] = $31;
       }
       HEAP32[$$pre10$i>>2] = 0;
       HEAP32[$28>>2] = $18;
       $37 = ((($entry$05$i)) + 152|0);
       HEAP32[$37>>2] = $entry$06$i$lcssa;
       HEAP32[$16>>2] = $entry$06$i$lcssa;
       $$0$i$ph = $entry$06$i$lcssa;
       label = 49;
      }
     } while(0);
     L35: do {
      if ((label|0) == 24) {
       $38 = (_SDL_calloc(1,160)|0);
       $39 = ($38|0)==(0|0);
       do {
        if ($39) {
         (_SDL_Error(0)|0);
        } else {
         $40 = ((($38)) + 8|0);
         HEAP32[$40>>2] = $2;
         $41 = ((($38)) + 12|0);
         HEAP32[$41>>2] = $4;
         $42 = ((($38)) + 4|0);
         HEAP32[$42>>2] = $blendMode;
         $43 = ((($15)) + 48|0);
         $44 = HEAP32[$43>>2]|0;
         $45 = (FUNCTION_TABLE_i[$44 & 8191]()|0);
         HEAP32[$38>>2] = $45;
         $46 = ((($15)) + 20|0);
         $47 = HEAP32[$46>>2]|0;
         $48 = HEAP32[$2>>2]|0;
         FUNCTION_TABLE_vii[$47 & 8191]($45,$48);
         $49 = HEAP32[$46>>2]|0;
         $50 = HEAP32[$38>>2]|0;
         $51 = HEAP32[$4>>2]|0;
         FUNCTION_TABLE_vii[$49 & 8191]($50,$51);
         $52 = ((($15)) + 24|0);
         $53 = HEAP32[$52>>2]|0;
         $54 = HEAP32[$38>>2]|0;
         FUNCTION_TABLE_viii[$53 & 8191]($54,0,602456);
         $55 = HEAP32[$52>>2]|0;
         $56 = HEAP32[$38>>2]|0;
         FUNCTION_TABLE_viii[$55 & 8191]($56,1,602467);
         $57 = HEAP32[$52>>2]|0;
         $58 = HEAP32[$38>>2]|0;
         FUNCTION_TABLE_viii[$57 & 8191]($58,2,602478);
         $59 = HEAP32[$52>>2]|0;
         $60 = HEAP32[$38>>2]|0;
         FUNCTION_TABLE_viii[$59 & 8191]($60,3,602486);
         $61 = ((($15)) + 132|0);
         $62 = HEAP32[$61>>2]|0;
         $63 = HEAP32[$38>>2]|0;
         FUNCTION_TABLE_vi[$62 & 8191]($63);
         $64 = ((($15)) + 116|0);
         $65 = HEAP32[$64>>2]|0;
         $66 = HEAP32[$38>>2]|0;
         FUNCTION_TABLE_viii[$65 & 8191]($66,35714,$linkSuccessful$i);
         $67 = HEAP32[$linkSuccessful$i>>2]|0;
         $68 = ($67|0)==(0);
         if ($68) {
          $69 = ((($15)) + 56|0);
          $70 = HEAP32[$69>>2]|0;
          $71 = HEAP32[$38>>2]|0;
          FUNCTION_TABLE_vi[$70 & 8191]($71);
          _SDL_free($38);
          (_SDL_SetError(602495,$vararg_buffer)|0);
          break;
         }
         $72 = ((($15)) + 128|0);
         $73 = HEAP32[$72>>2]|0;
         $74 = HEAP32[$38>>2]|0;
         $75 = (FUNCTION_TABLE_iii[$73 & 8191]($74,602525)|0);
         $76 = ((($38)) + 16|0);
         HEAP32[$76>>2] = $75;
         $77 = HEAP32[$72>>2]|0;
         $78 = HEAP32[$38>>2]|0;
         $79 = (FUNCTION_TABLE_iii[$77 & 8191]($78,602538)|0);
         $80 = ((($38)) + 36|0);
         HEAP32[$80>>2] = $79;
         $81 = HEAP32[$72>>2]|0;
         $82 = HEAP32[$38>>2]|0;
         $83 = (FUNCTION_TABLE_iii[$81 & 8191]($82,602550)|0);
         $84 = ((($38)) + 32|0);
         HEAP32[$84>>2] = $83;
         $85 = HEAP32[$72>>2]|0;
         $86 = HEAP32[$38>>2]|0;
         $87 = (FUNCTION_TABLE_iii[$85 & 8191]($86,602562)|0);
         $88 = ((($38)) + 20|0);
         HEAP32[$88>>2] = $87;
         $89 = HEAP32[$72>>2]|0;
         $90 = HEAP32[$38>>2]|0;
         $91 = (FUNCTION_TABLE_iii[$89 & 8191]($90,602572)|0);
         $92 = ((($38)) + 24|0);
         HEAP32[$92>>2] = $91;
         $93 = HEAP32[$72>>2]|0;
         $94 = HEAP32[$38>>2]|0;
         $95 = (FUNCTION_TABLE_iii[$93 & 8191]($94,602585)|0);
         $96 = ((($38)) + 28|0);
         HEAP32[$96>>2] = $95;
         $97 = ((($38)) + 80|0);
         $98 = ((($15)) + 180|0);
         $99 = $97;
         $100 = $99;
         HEAP32[$100>>2] = -1;
         $101 = (($99) + 4)|0;
         $102 = $101;
         HEAP32[$102>>2] = -1;
         $103 = HEAP32[$98>>2]|0;
         $104 = HEAP32[$38>>2]|0;
         FUNCTION_TABLE_vi[$103 & 8191]($104);
         $105 = ((($15)) + 168|0);
         $106 = HEAP32[$105>>2]|0;
         $107 = HEAP32[$80>>2]|0;
         FUNCTION_TABLE_vii[$106 & 8191]($107,2);
         $108 = HEAP32[$105>>2]|0;
         $109 = HEAP32[$84>>2]|0;
         FUNCTION_TABLE_vii[$108 & 8191]($109,1);
         $110 = HEAP32[$105>>2]|0;
         $111 = HEAP32[$88>>2]|0;
         FUNCTION_TABLE_vii[$110 & 8191]($111,0);
         $112 = ((($15)) + 176|0);
         $113 = HEAP32[$112>>2]|0;
         $114 = HEAP32[$76>>2]|0;
         $115 = ((($38)) + 88|0);
         FUNCTION_TABLE_viiii[$113 & 8191]($114,1,0,$115);
         $116 = ((($15)) + 172|0);
         $117 = HEAP32[$116>>2]|0;
         $118 = HEAP32[$92>>2]|0;
         FUNCTION_TABLE_vidddd[$117 & 8191]($118,1.0,1.0,1.0,1.0);
         $119 = HEAP32[$116>>2]|0;
         $120 = HEAP32[$96>>2]|0;
         FUNCTION_TABLE_vidddd[$119 & 8191]($120,1.0,1.0,1.0,1.0);
         $121 = HEAP32[$16>>2]|0;
         $122 = ($121|0)==(0|0);
         if ($122) {
          $125 = ((($15)) + 264|0);
          HEAP32[$125>>2] = $38;
         } else {
          $123 = ((($38)) + 156|0);
          HEAP32[$123>>2] = $121;
          $124 = ((($121)) + 152|0);
          HEAP32[$124>>2] = $38;
         }
         HEAP32[$16>>2] = $38;
         $126 = ((($15)) + 256|0);
         $127 = HEAP32[$126>>2]|0;
         $128 = (($127) + 1)|0;
         HEAP32[$126>>2] = $128;
         $129 = ((($2)) + 12|0);
         $130 = HEAP32[$129>>2]|0;
         $131 = (($130) + 1)|0;
         HEAP32[$129>>2] = $131;
         $132 = ((($4)) + 12|0);
         $133 = HEAP32[$132>>2]|0;
         $134 = (($133) + 1)|0;
         HEAP32[$132>>2] = $134;
         $135 = ($127|0)>(7);
         if (!($135)) {
          $$0$i$ph = $38;
          label = 49;
          break L35;
         }
         $136 = ((($15)) + 264|0);
         $137 = HEAP32[$136>>2]|0;
         $138 = ((($137)) + 8|0);
         $139 = HEAP32[$138>>2]|0;
         $140 = ((($139)) + 12|0);
         $141 = HEAP32[$140>>2]|0;
         $142 = (($141) + -1)|0;
         HEAP32[$140>>2] = $142;
         $143 = ($141|0)<(2);
         if ($143) {
          $144 = HEAP32[$0>>2]|0;
          $145 = ((($139)) + 24|0);
          $146 = HEAP32[$145>>2]|0;
          $147 = ($146|0)==(0|0);
          $148 = $146;
          $$pre$i$i = ((($139)) + 20|0);
          if (!($147)) {
           $149 = HEAP32[$$pre$i$i>>2]|0;
           $150 = ((($146)) + 20|0);
           HEAP32[$150>>2] = $149;
          }
          $151 = HEAP32[$$pre$i$i>>2]|0;
          $152 = ($151|0)==(0|0);
          if (!($152)) {
           $153 = ((($151)) + 24|0);
           HEAP32[$153>>2] = $148;
          }
          $154 = ((($144)) + 252|0);
          $155 = HEAP32[$154>>2]|0;
          $156 = ($155|0)==($139|0);
          if ($156) {
           $157 = HEAP32[$145>>2]|0;
           HEAP32[$154>>2] = $157;
          }
          $158 = ((($144)) + 248|0);
          $159 = HEAP32[$158>>2]|0;
          $160 = (($159) + -1)|0;
          HEAP32[$158>>2] = $160;
          $161 = ((($144)) + 60|0);
          $162 = HEAP32[$161>>2]|0;
          $163 = HEAP32[$139>>2]|0;
          FUNCTION_TABLE_vi[$162 & 8191]($163);
          _SDL_free($139);
          $$pre$i = HEAP32[$136>>2]|0;
          $165 = $$pre$i;
         } else {
          $165 = $137;
         }
         $164 = ((($165)) + 12|0);
         $166 = HEAP32[$164>>2]|0;
         $167 = ((($166)) + 12|0);
         $168 = HEAP32[$167>>2]|0;
         $169 = (($168) + -1)|0;
         HEAP32[$167>>2] = $169;
         $170 = ($168|0)<(2);
         if ($170) {
          $171 = HEAP32[$0>>2]|0;
          $172 = ((($166)) + 24|0);
          $173 = HEAP32[$172>>2]|0;
          $174 = ($173|0)==(0|0);
          $175 = $173;
          $$pre$i1$i = ((($166)) + 20|0);
          if (!($174)) {
           $176 = HEAP32[$$pre$i1$i>>2]|0;
           $177 = ((($173)) + 20|0);
           HEAP32[$177>>2] = $176;
          }
          $178 = HEAP32[$$pre$i1$i>>2]|0;
          $179 = ($178|0)==(0|0);
          if (!($179)) {
           $180 = ((($178)) + 24|0);
           HEAP32[$180>>2] = $175;
          }
          $181 = ((($171)) + 252|0);
          $182 = HEAP32[$181>>2]|0;
          $183 = ($182|0)==($166|0);
          if ($183) {
           $184 = HEAP32[$172>>2]|0;
           HEAP32[$181>>2] = $184;
          }
          $185 = ((($171)) + 248|0);
          $186 = HEAP32[$185>>2]|0;
          $187 = (($186) + -1)|0;
          HEAP32[$185>>2] = $187;
          $188 = ((($171)) + 60|0);
          $189 = HEAP32[$188>>2]|0;
          $190 = HEAP32[$166>>2]|0;
          FUNCTION_TABLE_vi[$189 & 8191]($190);
          _SDL_free($166);
          $$pre8$i = HEAP32[$136>>2]|0;
          $194 = $$pre8$i;
         } else {
          $194 = $165;
         }
         $191 = ((($15)) + 56|0);
         $192 = HEAP32[$191>>2]|0;
         $193 = HEAP32[$194>>2]|0;
         FUNCTION_TABLE_vi[$192 & 8191]($193);
         $195 = HEAP32[$136>>2]|0;
         $196 = ((($195)) + 152|0);
         $197 = HEAP32[$196>>2]|0;
         HEAP32[$136>>2] = $197;
         $198 = ((($197)) + 156|0);
         $199 = HEAP32[$198>>2]|0;
         _SDL_free($199);
         $200 = HEAP32[$136>>2]|0;
         $201 = ((($200)) + 156|0);
         HEAP32[$201>>2] = 0;
         $202 = HEAP32[$126>>2]|0;
         $203 = (($202) + -1)|0;
         HEAP32[$126>>2] = $203;
         $$0$i$ph = $38;
         label = 49;
         break L35;
        }
       } while(0);
       $fragment$011 = $4;
       break L12;
      }
     } while(0);
     if ((label|0) == 49) {
      $$0$i6 = $$0$i$ph;
     }
     $204 = ((($1)) + 180|0);
     $205 = HEAP32[$204>>2]|0;
     $206 = HEAP32[$$0$i6>>2]|0;
     FUNCTION_TABLE_vi[$205 & 8191]($206);
     HEAP32[$6>>2] = $$0$i6;
     $207 = HEAP32[$0>>2]|0;
     $208 = ((($renderer)) + 220|0);
     $209 = HEAP32[$208>>2]|0;
     $210 = ($209|0)==(0);
     if (!($210)) {
      $211 = ((($renderer)) + 224|0);
      $212 = HEAP32[$211>>2]|0;
      $213 = ($212|0)==(0);
      if (!($213)) {
       $214 = (+($209|0));
       $215 = 2.0 / $214;
       HEAPF32[$projection$i>>2] = $215;
       $216 = ((($projection$i)) + 4|0);
       $217 = ((($renderer)) + 304|0);
       ;HEAP32[$216>>2]=0|0;HEAP32[$216+4>>2]=0|0;HEAP32[$216+8>>2]=0|0;HEAP32[$216+12>>2]=0|0;
       $218 = HEAP32[$217>>2]|0;
       $219 = ($218|0)==(0|0);
       $220 = (+($212|0));
       $$1$i = $219 ? -2.0 : 2.0;
       $$sink$i = $$1$i / $220;
       $221 = ((($projection$i)) + 20|0);
       HEAPF32[$221>>2] = $$sink$i;
       $222 = ((($projection$i)) + 24|0);
       $223 = ((($projection$i)) + 48|0);
       ;HEAP32[$222>>2]=0|0;HEAP32[$222+4>>2]=0|0;HEAP32[$222+8>>2]=0|0;HEAP32[$222+12>>2]=0|0;HEAP32[$222+16>>2]=0|0;HEAP32[$222+20>>2]=0|0;
       HEAPF32[$223>>2] = -1.0;
       $224 = ((($projection$i)) + 52|0);
       $$$i = $219 ? 1.0 : -1.0;
       HEAPF32[$224>>2] = $$$i;
       $225 = ((($projection$i)) + 56|0);
       HEAPF32[$225>>2] = 0.0;
       $226 = ((($projection$i)) + 60|0);
       HEAPF32[$226>>2] = 1.0;
       $227 = ((($207)) + 268|0);
       $228 = HEAP32[$227>>2]|0;
       $229 = ((($228)) + 88|0);
       $230 = (_SDL_memcmp($229,$projection$i,64)|0);
       $231 = ($230|0)==(0);
       if (!($231)) {
        $232 = HEAP32[$227>>2]|0;
        $233 = ((($232)) + 16|0);
        $234 = HEAP32[$233>>2]|0;
        $235 = ((($207)) + 176|0);
        $236 = HEAP32[$235>>2]|0;
        FUNCTION_TABLE_viiii[$236 & 8191]($234,1,0,$projection$i);
        $237 = HEAP32[$227>>2]|0;
        $238 = ((($237)) + 88|0);
        (_SDL_memcpy($238,$projection$i,64)|0);
       }
      }
     }
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
   } while(0);
   $239 = ((($2)) + 12|0);
   $240 = HEAP32[$239>>2]|0;
   $241 = ($240|0)<(1);
   if ($241) {
    $242 = HEAP32[$0>>2]|0;
    $243 = ((($2)) + 24|0);
    $244 = HEAP32[$243>>2]|0;
    $245 = ($244|0)==(0|0);
    $246 = $244;
    $$pre$i1 = ((($2)) + 20|0);
    if (!($245)) {
     $247 = HEAP32[$$pre$i1>>2]|0;
     $248 = ((($244)) + 20|0);
     HEAP32[$248>>2] = $247;
    }
    $249 = HEAP32[$$pre$i1>>2]|0;
    $250 = ($249|0)==(0|0);
    if (!($250)) {
     $251 = ((($249)) + 24|0);
     HEAP32[$251>>2] = $246;
    }
    $252 = ((($242)) + 252|0);
    $253 = HEAP32[$252>>2]|0;
    $254 = ($253|0)==($2|0);
    if ($254) {
     $255 = HEAP32[$243>>2]|0;
     HEAP32[$252>>2] = $255;
    }
    $256 = ((($242)) + 248|0);
    $257 = HEAP32[$256>>2]|0;
    $258 = (($257) + -1)|0;
    HEAP32[$256>>2] = $258;
    $259 = ((($242)) + 60|0);
    $260 = HEAP32[$259>>2]|0;
    $261 = HEAP32[$2>>2]|0;
    FUNCTION_TABLE_vi[$260 & 8191]($261);
    _SDL_free($2);
   }
   $262 = ($fragment$011|0)==(0|0);
   if (!($262)) {
    $263 = ((($fragment$011)) + 12|0);
    $264 = HEAP32[$263>>2]|0;
    $265 = ($264|0)<(1);
    if ($265) {
     $266 = HEAP32[$0>>2]|0;
     $267 = ((($fragment$011)) + 24|0);
     $268 = HEAP32[$267>>2]|0;
     $269 = ($268|0)==(0|0);
     $270 = $268;
     $$pre$i3 = ((($fragment$011)) + 20|0);
     if (!($269)) {
      $271 = HEAP32[$$pre$i3>>2]|0;
      $272 = ((($268)) + 20|0);
      HEAP32[$272>>2] = $271;
     }
     $273 = HEAP32[$$pre$i3>>2]|0;
     $274 = ($273|0)==(0|0);
     if (!($274)) {
      $275 = ((($273)) + 24|0);
      HEAP32[$275>>2] = $270;
     }
     $276 = ((($266)) + 252|0);
     $277 = HEAP32[$276>>2]|0;
     $278 = ($277|0)==($fragment$011|0);
     if ($278) {
      $279 = HEAP32[$267>>2]|0;
      HEAP32[$276>>2] = $279;
     }
     $280 = ((($266)) + 248|0);
     $281 = HEAP32[$280>>2]|0;
     $282 = (($281) + -1)|0;
     HEAP32[$280>>2] = $282;
     $283 = ((($266)) + 60|0);
     $284 = HEAP32[$283>>2]|0;
     $285 = HEAP32[$fragment$011>>2]|0;
     FUNCTION_TABLE_vi[$284 & 8191]($285);
     _SDL_free($fragment$011);
    }
   }
  }
 }
 $286 = ((($1)) + 268|0);
 HEAP32[$286>>2] = 0;
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _GLES2_CacheShader($renderer,$type,$blendMode) {
 $renderer = $renderer|0;
 $type = $type|0;
 $blendMode = $blendMode|0;
 var $$0 = 0, $$alloca_mul = 0, $$instance$1$us20 = 0, $$pr = 0, $$pre = 0, $$pre30 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $8 = 0;
 var $9 = 0, $compileSuccessful = 0, $entry$0 = 0, $entry$07 = 0, $entry$08 = 0, $i$014$us = 0, $instance$1$lcssa$us = 0, $j$011$us18 = 0, $length = 0, $or$cond$us = 0, $or$cond2$us22 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $compileSuccessful = sp + 32|0;
 $length = sp + 28|0;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$compileSuccessful>>2] = 0;
 $2 = (_GLES2_GetShader($type,$blendMode)|0);
 $3 = ($2|0)==(0|0);
 if ($3) {
  (_SDL_SetError(602279,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = HEAP32[$2>>2]|0;
 $5 = ($4|0)>(0);
 if ($5) {
  $6 = ((($1)) + 240|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)>(0);
  $9 = ((($1)) + 244|0);
  if ($8) {
   $i$014$us = 0;
   while(1) {
    $20 = (((($2)) + 4|0) + ($i$014$us<<2)|0);
    $16 = HEAP32[$20>>2]|0;
    $21 = ($16|0)==(0|0);
    if ($21) {
     $instance$1$lcssa$us = 0;
    } else {
     $22 = ((($16)) + 4|0);
     $$pre = HEAP32[$22>>2]|0;
     $$pre30 = HEAP32[$9>>2]|0;
     $j$011$us18 = 0;
     while(1) {
      $13 = (($$pre30) + ($j$011$us18<<2)|0);
      $14 = HEAP32[$13>>2]|0;
      $15 = ($$pre|0)==($14|0);
      $$instance$1$us20 = $15 ? $16 : 0;
      $17 = (($j$011$us18) + 1)|0;
      $18 = ($17|0)<($7|0);
      $19 = ($$instance$1$us20|0)==(0|0);
      $or$cond2$us22 = $19 & $18;
      if ($or$cond2$us22) {
       $j$011$us18 = $17;
      } else {
       $instance$1$lcssa$us = $$instance$1$us20;
       break;
      }
     }
    }
    $10 = (($i$014$us) + 1)|0;
    $11 = ($10|0)<($4|0);
    $12 = ($instance$1$lcssa$us|0)==(0|0);
    $or$cond$us = $12 & $11;
    if ($or$cond$us) {
     $i$014$us = $10;
    } else {
     break;
    }
   }
   if (!($12)) {
    $23 = ((($1)) + 252|0);
    $entry$07 = HEAP32[$23>>2]|0;
    $24 = ($entry$07|0)==(0|0);
    L16: do {
     if (!($24)) {
      $entry$08 = $entry$07;
      while(1) {
       $25 = ((($entry$08)) + 8|0);
       $26 = HEAP32[$25>>2]|0;
       $27 = ($26|0)==($instance$1$lcssa$us|0);
       if ($27) {
        $$0 = $entry$08;
        break;
       }
       $28 = ((($entry$08)) + 24|0);
       $entry$0 = HEAP32[$28>>2]|0;
       $29 = ($entry$0|0)==(0|0);
       if ($29) {
        break L16;
       } else {
        $entry$08 = $entry$0;
       }
      }
      STACKTOP = sp;return ($$0|0);
     }
    } while(0);
    $30 = (_SDL_calloc(1,28)|0);
    $31 = ($30|0)==(0|0);
    if ($31) {
     (_SDL_Error(0)|0);
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
    $32 = ((($30)) + 4|0);
    HEAP32[$32>>2] = $type;
    $33 = ((($30)) + 8|0);
    HEAP32[$33>>2] = $instance$1$lcssa$us;
    $34 = ((($1)) + 52|0);
    $35 = HEAP32[$34>>2]|0;
    $36 = HEAP32[$instance$1$lcssa$us>>2]|0;
    $37 = (FUNCTION_TABLE_ii[$35 & 8191]($36)|0);
    HEAP32[$30>>2] = $37;
    $38 = ((($instance$1$lcssa$us)) + 4|0);
    $39 = HEAP32[$38>>2]|0;
    $40 = ($39|0)==(-1);
    if ($40) {
     $47 = ((($1)) + 152|0);
     $48 = HEAP32[$47>>2]|0;
     $49 = ((($instance$1$lcssa$us)) + 12|0);
     FUNCTION_TABLE_viiii[$48 & 8191]($37,1,$49,0);
     $50 = ((($1)) + 44|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = HEAP32[$30>>2]|0;
     FUNCTION_TABLE_vi[$51 & 8191]($52);
     $53 = ((($1)) + 124|0);
     $54 = HEAP32[$53>>2]|0;
     $55 = HEAP32[$30>>2]|0;
     FUNCTION_TABLE_viii[$54 & 8191]($55,35713,$compileSuccessful);
     $$pr = HEAP32[$compileSuccessful>>2]|0;
     $56 = ($$pr|0)==(0);
     if ($56) {
      HEAP32[$length>>2] = 0;
      $57 = HEAP32[$53>>2]|0;
      $58 = HEAP32[$30>>2]|0;
      FUNCTION_TABLE_viii[$57 & 8191]($58,35716,$length);
      $59 = HEAP32[$length>>2]|0;
      $60 = ($59|0)>(0);
      if ($60) {
       $$alloca_mul = $59;
       $61 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
       $62 = ((($1)) + 120|0);
       $63 = HEAP32[$62>>2]|0;
       $64 = HEAP32[$30>>2]|0;
       FUNCTION_TABLE_viiii[$63 & 8191]($64,$59,$length,$61);
       HEAP32[$vararg_buffer3>>2] = $61;
       (_SDL_SetError(602400,$vararg_buffer3)|0);
      } else {
       (_SDL_SetError(602430,$vararg_buffer5)|0);
      }
      $65 = ((($1)) + 60|0);
      $66 = HEAP32[$65>>2]|0;
      $67 = HEAP32[$30>>2]|0;
      FUNCTION_TABLE_vi[$66 & 8191]($67);
      _SDL_free($30);
      $$0 = 0;
      STACKTOP = sp;return ($$0|0);
     }
    } else {
     $41 = ((($1)) + 148|0);
     $42 = HEAP32[$41>>2]|0;
     $43 = ((($instance$1$lcssa$us)) + 12|0);
     $44 = HEAP32[$43>>2]|0;
     $45 = ((($instance$1$lcssa$us)) + 8|0);
     $46 = HEAP32[$45>>2]|0;
     FUNCTION_TABLE_viiiii[$42 & 8191](1,$30,$39,$44,$46);
     HEAP32[$compileSuccessful>>2] = 1;
    }
    $68 = HEAP32[$23>>2]|0;
    $69 = ($68|0)==(0|0);
    if (!($69)) {
     $70 = ((($30)) + 24|0);
     HEAP32[$70>>2] = $68;
     $71 = ((($68)) + 20|0);
     HEAP32[$71>>2] = $30;
    }
    HEAP32[$23>>2] = $30;
    $72 = ((($1)) + 248|0);
    $73 = HEAP32[$72>>2]|0;
    $74 = (($73) + 1)|0;
    HEAP32[$72>>2] = $74;
    $$0 = $30;
    STACKTOP = sp;return ($$0|0);
   }
  }
 }
 (_SDL_SetError(602338,$vararg_buffer1)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _GLES2_GetShader($type,$blendMode) {
 $type = $type|0;
 $blendMode = $blendMode|0;
 var $$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 L1: do {
  switch ($type|0) {
  case 0:  {
   $$0 = 126464;
   break;
  }
  case 1:  {
   switch ($blendMode|0) {
   case 0:  {
    $$0 = 126500;
    break L1;
    break;
   }
   case 1:  {
    $$0 = 126444;
    break L1;
    break;
   }
   case 2:  {
    $$0 = 126424;
    break L1;
    break;
   }
   case 4:  {
    $$0 = 126388;
    break L1;
    break;
   }
   default: {
    $$0 = 0;
    break L1;
   }
   }
   break;
  }
  case 2:  {
   switch ($blendMode|0) {
   case 0:  {
    $$0 = 126520;
    break L1;
    break;
   }
   case 1:  {
    $$0 = 126368;
    break L1;
    break;
   }
   case 2:  {
    $$0 = 126348;
    break L1;
    break;
   }
   case 4:  {
    $$0 = 126312;
    break L1;
    break;
   }
   default: {
    $$0 = 0;
    break L1;
   }
   }
   break;
  }
  case 3:  {
   switch ($blendMode|0) {
   case 0:  {
    $$0 = 126540;
    break L1;
    break;
   }
   case 1:  {
    $$0 = 126292;
    break L1;
    break;
   }
   case 2:  {
    $$0 = 126272;
    break L1;
    break;
   }
   case 4:  {
    $$0 = 126236;
    break L1;
    break;
   }
   default: {
    $$0 = 0;
    break L1;
   }
   }
   break;
  }
  case 5:  {
   switch ($blendMode|0) {
   case 0:  {
    $$0 = 126560;
    break L1;
    break;
   }
   case 1:  {
    $$0 = 126216;
    break L1;
    break;
   }
   case 2:  {
    $$0 = 126196;
    break L1;
    break;
   }
   case 4:  {
    $$0 = 126160;
    break L1;
    break;
   }
   default: {
    $$0 = 0;
    break L1;
   }
   }
   break;
  }
  case 4:  {
   switch ($blendMode|0) {
   case 0:  {
    $$0 = 126580;
    break L1;
    break;
   }
   case 1:  {
    $$0 = 126140;
    break L1;
    break;
   }
   case 2:  {
    $$0 = 126120;
    break L1;
    break;
   }
   case 4:  {
    $$0 = 126084;
    break L1;
    break;
   }
   default: {
    $$0 = 0;
    break L1;
   }
   }
   break;
  }
  case 6:  {
   $$0 = 126048;
   break;
  }
  case 7:  {
   $$0 = 126012;
   break;
  }
  case 8:  {
   $$0 = 125976;
   break;
  }
  default: {
   $$0 = 0;
  }
  }
 } while(0);
 return ($$0|0);
}
function _GLES2_RenderDrawLines($renderer,$points,$count) {
 $renderer = $renderer|0;
 $points = $points|0;
 $count = $count|0;
 var $$$021$i = 0, $$0 = 0, $$0$i$i = 0, $$021$i = 0, $$alloca_mul = 0, $$pre$i = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0.0, $4 = 0, $40 = 0, $41 = 0, $42 = 0.0, $43 = 0, $44 = 0, $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0.0, $9 = 0.0, $exitcond = 0, $idx$01 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0, $vararg_ptr4 = 0, $vararg_ptr5 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_GLES2_SetDrawingState($renderer)|0);
 $3 = ($2|0)<(0);
 if ($3) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = $count << 3;
 $$alloca_mul = $4;
 $5 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $6 = ($count|0)>(0);
 if ($6) {
  $idx$01 = 0;
  while(1) {
   $7 = (($points) + ($idx$01<<3)|0);
   $8 = +HEAPF32[$7>>2];
   $9 = $8 + 0.5;
   $10 = (((($points) + ($idx$01<<3)|0)) + 4|0);
   $11 = +HEAPF32[$10>>2];
   $12 = $11 + 0.5;
   $13 = $idx$01 << 1;
   $14 = (($5) + ($13<<2)|0);
   HEAPF32[$14>>2] = $9;
   $15 = $13 | 1;
   $16 = (($5) + ($15<<2)|0);
   HEAPF32[$16>>2] = $12;
   $17 = (($idx$01) + 1)|0;
   $exitcond = ($17|0)==($count|0);
   if ($exitcond) {
    break;
   } else {
    $idx$01 = $17;
   }
  }
 }
 $18 = HEAP32[$0>>2]|0;
 $19 = ((($18)) + 276|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ($20|0)==(0);
 if ($21) {
  $22 = ((($18)) + 216|0);
  $23 = HEAP32[$22>>2]|0;
  FUNCTION_TABLE_vii[$23 & 8191](1,$19);
  $$pre$i = HEAP32[$19>>2]|0;
  $26 = $$pre$i;
 } else {
  $26 = $20;
 }
 $24 = ((($18)) + 220|0);
 $25 = HEAP32[$24>>2]|0;
 FUNCTION_TABLE_vii[$25 & 8191](34962,$26);
 $27 = ((($18)) + 292|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ($28>>>0)<($4>>>0);
 if ($29) {
  $30 = ((($18)) + 224|0);
  $31 = HEAP32[$30>>2]|0;
  FUNCTION_TABLE_viiii[$31 & 8191](34962,$4,$5,35040);
  HEAP32[$27>>2] = $4;
 } else {
  $32 = ((($18)) + 228|0);
  $33 = HEAP32[$32>>2]|0;
  FUNCTION_TABLE_viiii[$33 & 8191](34962,0,$4,$5);
 }
 $34 = ((($18)) + 184|0);
 $35 = HEAP32[$34>>2]|0;
 FUNCTION_TABLE_viiiiii[$35 & 8191](0,2,5126,0,0,0);
 $36 = ((($1)) + 76|0);
 $37 = HEAP32[$36>>2]|0;
 FUNCTION_TABLE_viii[$37 & 8191](3,0,$count);
 $38 = ($count|0)==(2);
 if ($38) {
  $$pre$phiZ2D = 1;
  label = 12;
 } else {
  $39 = +HEAPF32[$points>>2];
  $40 = (($count) + -1)|0;
  $41 = (($points) + ($40<<3)|0);
  $42 = +HEAPF32[$41>>2];
  $43 = $39 != $42;
  if ($43) {
   $$pre$phiZ2D = $40;
   label = 12;
  } else {
   $44 = ((($points)) + 4|0);
   $45 = +HEAPF32[$44>>2];
   $46 = (((($points) + ($40<<3)|0)) + 4|0);
   $47 = +HEAPF32[$46>>2];
   $48 = $45 != $47;
   if ($48) {
    $$pre$phiZ2D = $40;
    label = 12;
   }
  }
 }
 if ((label|0) == 12) {
  $49 = HEAP32[$36>>2]|0;
  FUNCTION_TABLE_viii[$49 & 8191](0,$$pre$phiZ2D,1);
 }
 $50 = HEAP32[$0>>2]|0;
 $51 = ((($50)) + 4|0);
 $52 = HEAP32[$51>>2]|0;
 $53 = ($52|0)==(0);
 if ($53) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $54 = ((($50)) + 108|0);
 $55 = HEAP32[$54>>2]|0;
 $56 = (FUNCTION_TABLE_i[$55 & 8191]()|0);
 $57 = ($56|0)==(0);
 if ($57) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $$021$i = 1282466;$60 = $56;
 }
 while(1) {
  $58 = HEAP8[$$021$i>>0]|0;
  $59 = ($58<<24>>24)==(0);
  $$$021$i = $59 ? 597716 : $$021$i;
  switch ($60|0) {
  case 1280:  {
   $$0$i$i = 597787;
   break;
  }
  case 1281:  {
   $$0$i$i = 597770;
   break;
  }
  case 1282:  {
   $$0$i$i = 597749;
   break;
  }
  case 1285:  {
   $$0$i$i = 597732;
   break;
  }
  default: {
   $$0$i$i = 597724;
  }
  }
  HEAP32[$vararg_buffer>>2] = $$$021$i;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = 597829;
  $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
  HEAP32[$vararg_ptr2>>2] = 1506;
  $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
  HEAP32[$vararg_ptr3>>2] = 602593;
  $vararg_ptr4 = ((($vararg_buffer)) + 16|0);
  HEAP32[$vararg_ptr4>>2] = $$0$i$i;
  $vararg_ptr5 = ((($vararg_buffer)) + 20|0);
  HEAP32[$vararg_ptr5>>2] = $60;
  (_SDL_SetError(597803,$vararg_buffer)|0);
  $61 = HEAP32[$54>>2]|0;
  $62 = (FUNCTION_TABLE_i[$61 & 8191]()|0);
  $63 = ($62|0)==(0);
  if ($63) {
   $$0 = -1;
   break;
  } else {
   $$021$i = $$$021$i;$60 = $62;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _GLES2_RenderFillRects($renderer,$rects,$count) {
 $renderer = $renderer|0;
 $rects = $rects|0;
 $count = $count|0;
 var $$$021$i = 0, $$0 = 0, $$0$i$i = 0, $$021$i = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0;
 var $22 = 0.0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $exitcond = 0, $idx$01 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0, $vararg_ptr4 = 0, $vararg_ptr5 = 0, $vertices = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $vertices = sp + 24|0;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_GLES2_SetDrawingState($renderer)|0);
 $3 = ($2|0)<(0);
 if ($3) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ($count|0)>(0);
 if ($4) {
  $5 = ((($vertices)) + 4|0);
  $6 = ((($vertices)) + 8|0);
  $7 = ((($vertices)) + 12|0);
  $8 = ((($vertices)) + 16|0);
  $9 = ((($vertices)) + 20|0);
  $10 = ((($vertices)) + 24|0);
  $11 = ((($vertices)) + 28|0);
  $12 = ((($1)) + 76|0);
  $idx$01 = 0;
  while(1) {
   $13 = (($rects) + ($idx$01<<4)|0);
   $14 = +HEAPF32[$13>>2];
   $15 = (((($rects) + ($idx$01<<4)|0)) + 8|0);
   $16 = +HEAPF32[$15>>2];
   $17 = $14 + $16;
   $18 = (((($rects) + ($idx$01<<4)|0)) + 4|0);
   $19 = +HEAPF32[$18>>2];
   $20 = (((($rects) + ($idx$01<<4)|0)) + 12|0);
   $21 = +HEAPF32[$20>>2];
   $22 = $19 + $21;
   HEAPF32[$vertices>>2] = $14;
   HEAPF32[$5>>2] = $19;
   HEAPF32[$6>>2] = $17;
   HEAPF32[$7>>2] = $19;
   HEAPF32[$8>>2] = $14;
   HEAPF32[$9>>2] = $22;
   HEAPF32[$10>>2] = $17;
   HEAPF32[$11>>2] = $22;
   $23 = HEAP32[$0>>2]|0;
   $24 = ((($23)) + 276|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = ($25|0)==(0);
   if ($26) {
    $27 = ((($23)) + 216|0);
    $28 = HEAP32[$27>>2]|0;
    FUNCTION_TABLE_vii[$28 & 8191](1,$24);
    $$pre$i = HEAP32[$24>>2]|0;
    $31 = $$pre$i;
   } else {
    $31 = $25;
   }
   $29 = ((($23)) + 220|0);
   $30 = HEAP32[$29>>2]|0;
   FUNCTION_TABLE_vii[$30 & 8191](34962,$31);
   $32 = ((($23)) + 292|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ($33>>>0)<(32);
   if ($34) {
    $35 = ((($23)) + 224|0);
    $36 = HEAP32[$35>>2]|0;
    FUNCTION_TABLE_viiii[$36 & 8191](34962,32,$vertices,35040);
    HEAP32[$32>>2] = 32;
   } else {
    $37 = ((($23)) + 228|0);
    $38 = HEAP32[$37>>2]|0;
    FUNCTION_TABLE_viiii[$38 & 8191](34962,0,32,$vertices);
   }
   $39 = ((($23)) + 184|0);
   $40 = HEAP32[$39>>2]|0;
   FUNCTION_TABLE_viiiiii[$40 & 8191](0,2,5126,0,0,0);
   $41 = HEAP32[$12>>2]|0;
   FUNCTION_TABLE_viii[$41 & 8191](5,0,4);
   $42 = (($idx$01) + 1)|0;
   $exitcond = ($42|0)==($count|0);
   if ($exitcond) {
    break;
   } else {
    $idx$01 = $42;
   }
  }
 }
 $43 = HEAP32[$0>>2]|0;
 $44 = ((($43)) + 4|0);
 $45 = HEAP32[$44>>2]|0;
 $46 = ($45|0)==(0);
 if ($46) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $47 = ((($43)) + 108|0);
 $48 = HEAP32[$47>>2]|0;
 $49 = (FUNCTION_TABLE_i[$48 & 8191]()|0);
 $50 = ($49|0)==(0);
 if ($50) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $$021$i = 1282466;$53 = $49;
 }
 while(1) {
  $51 = HEAP8[$$021$i>>0]|0;
  $52 = ($51<<24>>24)==(0);
  $$$021$i = $52 ? 597716 : $$021$i;
  switch ($53|0) {
  case 1280:  {
   $$0$i$i = 597787;
   break;
  }
  case 1281:  {
   $$0$i$i = 597770;
   break;
  }
  case 1282:  {
   $$0$i$i = 597749;
   break;
  }
  case 1285:  {
   $$0$i$i = 597732;
   break;
  }
  default: {
   $$0$i$i = 597724;
  }
  }
  HEAP32[$vararg_buffer>>2] = $$$021$i;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = 597829;
  $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
  HEAP32[$vararg_ptr2>>2] = 1541;
  $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
  HEAP32[$vararg_ptr3>>2] = 602660;
  $vararg_ptr4 = ((($vararg_buffer)) + 16|0);
  HEAP32[$vararg_ptr4>>2] = $$0$i$i;
  $vararg_ptr5 = ((($vararg_buffer)) + 20|0);
  HEAP32[$vararg_ptr5>>2] = $53;
  (_SDL_SetError(597803,$vararg_buffer)|0);
  $54 = HEAP32[$47>>2]|0;
  $55 = (FUNCTION_TABLE_i[$54 & 8191]()|0);
  $56 = ($55|0)==(0);
  if ($56) {
   $$0 = -1;
   break;
  } else {
   $$021$i = $$$021$i;$53 = $55;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _GLES2_RenderCopy($renderer,$texture,$srcrect,$dstrect) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 $srcrect = $srcrect|0;
 $dstrect = $dstrect|0;
 var $$$021$i = 0, $$0 = 0, $$0$i$i = 0, $$021$i = 0, $$pre$i = 0, $$pre$i1 = 0, $$pre$i2 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0.0, $26 = 0, $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0.0, $59 = 0, $6 = 0, $60 = 0, $61 = 0.0, $62 = 0.0, $63 = 0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0.0, $83 = 0.0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $texCoords = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0, $vararg_ptr4 = 0, $vararg_ptr5 = 0, $vertices = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $vertices = sp + 56|0;
 $texCoords = sp + 24|0;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[300738]|0;
 $3 = HEAP32[$1>>2]|0;
 $4 = ($2|0)==($3|0);
 if ($4) {
  $12 = $1;
  label = 4;
 } else {
  $5 = ((($1)) + 268|0);
  HEAP32[$5>>2] = 0;
  $6 = ((($renderer)) + 188|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (_SDL_GL_MakeCurrent($7,$3)|0);
  $9 = ($8|0)<(0);
  if (!($9)) {
   $10 = HEAP32[$1>>2]|0;
   HEAP32[300738] = $10;
   (_GLES2_UpdateViewport($renderer)|0);
   $$pre$i = HEAP32[$0>>2]|0;
   $12 = $$pre$i;
   label = 4;
  }
 }
 if ((label|0) == 4) {
  $11 = ((($12)) + 4|0);
  $13 = HEAP32[$11>>2]|0;
  $14 = ($13|0)==(0);
  if (!($14)) {
   $15 = ((($12)) + 108|0);
   while(1) {
    $16 = HEAP32[$15>>2]|0;
    $17 = (FUNCTION_TABLE_i[$16 & 8191]()|0);
    $18 = ($17|0)==(0);
    if ($18) {
     break;
    }
   }
  }
 }
 $19 = (_GLES2_SetupCopy($renderer,$texture)|0);
 $20 = ($19|0)<(0);
 if ($20) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $21 = HEAP32[$dstrect>>2]|0;
 HEAP32[$vertices>>2] = $21;
 $22 = ((($dstrect)) + 4|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = ((($vertices)) + 4|0);
 HEAP32[$24>>2] = $23;
 $25 = (HEAP32[tempDoublePtr>>2]=$21,+HEAPF32[tempDoublePtr>>2]);
 $26 = ((($dstrect)) + 8|0);
 $27 = +HEAPF32[$26>>2];
 $28 = $25 + $27;
 $29 = ((($vertices)) + 8|0);
 HEAPF32[$29>>2] = $28;
 $30 = ((($vertices)) + 12|0);
 HEAP32[$30>>2] = $23;
 $31 = ((($vertices)) + 16|0);
 HEAP32[$31>>2] = $21;
 $32 = (HEAP32[tempDoublePtr>>2]=$23,+HEAPF32[tempDoublePtr>>2]);
 $33 = ((($dstrect)) + 12|0);
 $34 = +HEAPF32[$33>>2];
 $35 = $32 + $34;
 $36 = ((($vertices)) + 20|0);
 HEAPF32[$36>>2] = $35;
 $37 = ((($vertices)) + 24|0);
 HEAPF32[$37>>2] = $28;
 $38 = ((($vertices)) + 28|0);
 HEAPF32[$38>>2] = $35;
 $39 = HEAP32[$0>>2]|0;
 $40 = ((($39)) + 276|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = ($41|0)==(0);
 if ($42) {
  $43 = ((($39)) + 216|0);
  $44 = HEAP32[$43>>2]|0;
  FUNCTION_TABLE_vii[$44 & 8191](1,$40);
  $$pre$i1 = HEAP32[$40>>2]|0;
  $47 = $$pre$i1;
 } else {
  $47 = $41;
 }
 $45 = ((($39)) + 220|0);
 $46 = HEAP32[$45>>2]|0;
 FUNCTION_TABLE_vii[$46 & 8191](34962,$47);
 $48 = ((($39)) + 292|0);
 $49 = HEAP32[$48>>2]|0;
 $50 = ($49>>>0)<(32);
 if ($50) {
  $51 = ((($39)) + 224|0);
  $52 = HEAP32[$51>>2]|0;
  FUNCTION_TABLE_viiii[$52 & 8191](34962,32,$vertices,35040);
  HEAP32[$48>>2] = 32;
 } else {
  $53 = ((($39)) + 228|0);
  $54 = HEAP32[$53>>2]|0;
  FUNCTION_TABLE_viiii[$54 & 8191](34962,0,32,$vertices);
 }
 $55 = ((($39)) + 184|0);
 $56 = HEAP32[$55>>2]|0;
 FUNCTION_TABLE_viiiiii[$56 & 8191](0,2,5126,0,0,0);
 $57 = HEAP32[$srcrect>>2]|0;
 $58 = (+($57|0));
 $59 = ((($texture)) + 12|0);
 $60 = HEAP32[$59>>2]|0;
 $61 = (+($60|0));
 $62 = $58 / $61;
 HEAPF32[$texCoords>>2] = $62;
 $63 = ((($srcrect)) + 4|0);
 $64 = HEAP32[$63>>2]|0;
 $65 = (+($64|0));
 $66 = ((($texture)) + 16|0);
 $67 = HEAP32[$66>>2]|0;
 $68 = (+($67|0));
 $69 = $65 / $68;
 $70 = ((($texCoords)) + 4|0);
 HEAPF32[$70>>2] = $69;
 $71 = ((($srcrect)) + 8|0);
 $72 = HEAP32[$71>>2]|0;
 $73 = (($72) + ($57))|0;
 $74 = (+($73|0));
 $75 = $74 / $61;
 $76 = ((($texCoords)) + 8|0);
 HEAPF32[$76>>2] = $75;
 $77 = ((($texCoords)) + 12|0);
 HEAPF32[$77>>2] = $69;
 $78 = ((($texCoords)) + 16|0);
 HEAPF32[$78>>2] = $62;
 $79 = ((($srcrect)) + 12|0);
 $80 = HEAP32[$79>>2]|0;
 $81 = (($80) + ($64))|0;
 $82 = (+($81|0));
 $83 = $82 / $68;
 $84 = ((($texCoords)) + 20|0);
 HEAPF32[$84>>2] = $83;
 $85 = ((($texCoords)) + 24|0);
 HEAPF32[$85>>2] = $75;
 $86 = ((($texCoords)) + 28|0);
 HEAPF32[$86>>2] = $83;
 $87 = HEAP32[$0>>2]|0;
 $88 = ((($87)) + 280|0);
 $89 = HEAP32[$88>>2]|0;
 $90 = ($89|0)==(0);
 if ($90) {
  $91 = ((($87)) + 216|0);
  $92 = HEAP32[$91>>2]|0;
  FUNCTION_TABLE_vii[$92 & 8191](1,$88);
  $$pre$i2 = HEAP32[$88>>2]|0;
  $95 = $$pre$i2;
 } else {
  $95 = $89;
 }
 $93 = ((($87)) + 220|0);
 $94 = HEAP32[$93>>2]|0;
 FUNCTION_TABLE_vii[$94 & 8191](34962,$95);
 $96 = ((($87)) + 296|0);
 $97 = HEAP32[$96>>2]|0;
 $98 = ($97>>>0)<(32);
 if ($98) {
  $99 = ((($87)) + 224|0);
  $100 = HEAP32[$99>>2]|0;
  FUNCTION_TABLE_viiii[$100 & 8191](34962,32,$texCoords,35040);
  HEAP32[$96>>2] = 32;
 } else {
  $101 = ((($87)) + 228|0);
  $102 = HEAP32[$101>>2]|0;
  FUNCTION_TABLE_viiii[$102 & 8191](34962,0,32,$texCoords);
 }
 $103 = ((($87)) + 184|0);
 $104 = HEAP32[$103>>2]|0;
 FUNCTION_TABLE_viiiiii[$104 & 8191](1,2,5126,0,0,0);
 $105 = ((($1)) + 76|0);
 $106 = HEAP32[$105>>2]|0;
 FUNCTION_TABLE_viii[$106 & 8191](5,0,4);
 $107 = HEAP32[$0>>2]|0;
 $108 = ((($107)) + 4|0);
 $109 = HEAP32[$108>>2]|0;
 $110 = ($109|0)==(0);
 if ($110) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $111 = ((($107)) + 108|0);
 $112 = HEAP32[$111>>2]|0;
 $113 = (FUNCTION_TABLE_i[$112 & 8191]()|0);
 $114 = ($113|0)==(0);
 if ($114) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $$021$i = 1282466;$117 = $113;
 }
 while(1) {
  $115 = HEAP8[$$021$i>>0]|0;
  $116 = ($115<<24>>24)==(0);
  $$$021$i = $116 ? 597716 : $$021$i;
  switch ($117|0) {
  case 1280:  {
   $$0$i$i = 597787;
   break;
  }
  case 1281:  {
   $$0$i$i = 597770;
   break;
  }
  case 1282:  {
   $$0$i$i = 597749;
   break;
  }
  case 1285:  {
   $$0$i$i = 597732;
   break;
  }
  default: {
   $$0$i$i = 597724;
  }
  }
  HEAP32[$vararg_buffer>>2] = $$$021$i;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = 597829;
  $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
  HEAP32[$vararg_ptr2>>2] = 1743;
  $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
  HEAP32[$vararg_ptr3>>2] = 602753;
  $vararg_ptr4 = ((($vararg_buffer)) + 16|0);
  HEAP32[$vararg_ptr4>>2] = $$0$i$i;
  $vararg_ptr5 = ((($vararg_buffer)) + 20|0);
  HEAP32[$vararg_ptr5>>2] = $117;
  (_SDL_SetError(597803,$vararg_buffer)|0);
  $118 = HEAP32[$111>>2]|0;
  $119 = (FUNCTION_TABLE_i[$118 & 8191]()|0);
  $120 = ($119|0)==(0);
  if ($120) {
   $$0 = -1;
   break;
  } else {
   $$021$i = $$$021$i;$117 = $119;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _GLES2_SetupCopy($renderer,$texture) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 var $$0 = 0, $$pre = 0, $$pre$phi2Z2D = 0, $$pre$phiZ2D = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $b$0 = 0, $b$0$in = 0, $r$0 = 0, $r$0$in = 0, $sourceType$0 = 0, $switch$split122D = 0, $switch$split152D = 0, $switch$split182D = 0, $switch$split212D = 0, $switch$split242D = 0, $switch$split272D = 0, $switch$split2D = 0, $switch$split302D = 0, $switch$split32D = 0, $switch$split332D = 0, $switch$split362D = 0, $switch$split392D = 0, $switch$split422D = 0;
 var $switch$split452D = 0, $switch$split482D = 0, $switch$split512D = 0, $switch$split542D = 0, $switch$split572D = 0, $switch$split602D = 0, $switch$split62D = 0, $switch$split92D = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($texture)) + 68|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($texture)) + 24|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($renderer)) + 304|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 L1: do {
  if ($8) {
   $15 = ((($texture)) + 4|0);
   $16 = HEAP32[$15>>2]|0;
   $switch$split152D = ($16|0)<(825382478);
   L3: do {
    if ($switch$split152D) {
     $switch$split332D = ($16|0)<(374740996);
     if ($switch$split332D) {
      $switch$split512D = ($16|0)<(372645892);
      if (!($switch$split512D)) {
       switch ($16|0) {
       case 372645892:  {
        $sourceType$0 = 2;
        break L1;
        break;
       }
       default: {
        break L3;
       }
       }
      }
      switch ($16|0) {
      case 370546692:  {
       break;
      }
      default: {
       break L3;
      }
      }
      $sourceType$0 = 3;
      break L1;
     }
     $switch$split542D = ($16|0)<(376840196);
     if ($switch$split542D) {
      switch ($16|0) {
      case 374740996:  {
       break;
      }
      default: {
       break L3;
      }
      }
      $sourceType$0 = 4;
      break L1;
     } else {
      switch ($16|0) {
      case 376840196:  {
       break;
      }
      default: {
       break L3;
      }
      }
      $sourceType$0 = 1;
      break L1;
     }
    } else {
     $switch$split362D = ($16|0)<(842094169);
     if (!($switch$split362D)) {
      $switch$split602D = ($16|0)<(1448433993);
      if ($switch$split602D) {
       switch ($16|0) {
       case 842094169:  {
        break;
       }
       default: {
        break L3;
       }
       }
      } else {
       switch ($16|0) {
       case 1448433993:  {
        break;
       }
       default: {
        break L3;
       }
       }
      }
      $sourceType$0 = 5;
      break L1;
     }
     $switch$split572D = ($16|0)<(842094158);
     if ($switch$split572D) {
      switch ($16|0) {
      case 825382478:  {
       break;
      }
      default: {
       break L3;
      }
      }
      $sourceType$0 = 7;
      break L1;
     } else {
      switch ($16|0) {
      case 842094158:  {
       break;
      }
      default: {
       break L3;
      }
      }
      $sourceType$0 = 6;
      break L1;
     }
    }
   } while(0);
   $17 = (_SDL_SetError(602726,$vararg_buffer1)|0);
   $$0 = $17;
   STACKTOP = sp;return ($$0|0);
  } else {
   $9 = ((($7)) + 4|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ((($texture)) + 4|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($10|0)==($12|0);
   if ($13) {
    $sourceType$0 = 1;
   } else {
    $switch$split2D = ($12|0)<(825382478);
    L34: do {
     if ($switch$split2D) {
      $switch$split212D = ($12|0)<(374740996);
      if ($switch$split212D) {
       $switch$split392D = ($12|0)<(372645892);
       if (!($switch$split392D)) {
        switch ($12|0) {
        case 372645892:  {
         break;
        }
        default: {
         break L34;
        }
        }
        $switch$split32D = ($10|0)<(376840196);
        if ($switch$split32D) {
         switch ($10|0) {
         case 374740996:  {
          break;
         }
         default: {
          $sourceType$0 = 1;
          break L1;
         }
         }
        } else {
         switch ($10|0) {
         case 376840196:  {
          break;
         }
         default: {
          $sourceType$0 = 1;
          break L1;
         }
         }
        }
        $sourceType$0 = 2;
        break L1;
       }
       switch ($12|0) {
       case 370546692:  {
        break;
       }
       default: {
        break L34;
       }
       }
       $switch$split92D = ($10|0)<(374740996);
       if ($switch$split92D) {
        switch ($10|0) {
        case 372645892:  {
         break;
        }
        default: {
         $sourceType$0 = 1;
         break L1;
        }
        }
        $sourceType$0 = 4;
        break L1;
       }
       $switch$split272D = ($10|0)<(376840196);
       if ($switch$split272D) {
        switch ($10|0) {
        case 374740996:  {
         break;
        }
        default: {
         $sourceType$0 = 1;
         break L1;
        }
        }
        $sourceType$0 = 2;
        break L1;
       } else {
        switch ($10|0) {
        case 376840196:  {
         break;
        }
        default: {
         $sourceType$0 = 1;
         break L1;
        }
        }
        $sourceType$0 = 2;
        break L1;
       }
      } else {
       $switch$split422D = ($12|0)<(376840196);
       if (!($switch$split422D)) {
        switch ($12|0) {
        case 376840196:  {
         break;
        }
        default: {
         break L34;
        }
        }
        $switch$split62D = ($10|0)<(372645892);
        if ($switch$split62D) {
         switch ($10|0) {
         case 370546692:  {
          break;
         }
         default: {
          $sourceType$0 = 1;
          break L1;
         }
         }
        } else {
         switch ($10|0) {
         case 372645892:  {
          break;
         }
         default: {
          $sourceType$0 = 1;
          break L1;
         }
         }
        }
        $sourceType$0 = 2;
        break L1;
       }
       switch ($12|0) {
       case 374740996:  {
        break;
       }
       default: {
        break L34;
       }
       }
       $switch$split122D = ($10|0)<(372645892);
       if ($switch$split122D) {
        switch ($10|0) {
        case 370546692:  {
         break;
        }
        default: {
         $sourceType$0 = 1;
         break L1;
        }
        }
        $sourceType$0 = 2;
        break L1;
       }
       $switch$split302D = ($10|0)<(376840196);
       if ($switch$split302D) {
        switch ($10|0) {
        case 372645892:  {
         break;
        }
        default: {
         $sourceType$0 = 1;
         break L1;
        }
        }
        $sourceType$0 = 3;
        break L1;
       } else {
        switch ($10|0) {
        case 376840196:  {
         break;
        }
        default: {
         $sourceType$0 = 1;
         break L1;
        }
        }
        $sourceType$0 = 4;
        break L1;
       }
      }
     } else {
      $switch$split242D = ($12|0)<(842094169);
      if (!($switch$split242D)) {
       $switch$split482D = ($12|0)<(1448433993);
       if ($switch$split482D) {
        switch ($12|0) {
        case 842094169:  {
         $sourceType$0 = 5;
         break L1;
         break;
        }
        default: {
         break L34;
        }
        }
       } else {
        switch ($12|0) {
        case 1448433993:  {
         $sourceType$0 = 5;
         break L1;
         break;
        }
        default: {
         break L34;
        }
        }
       }
      }
      $switch$split452D = ($12|0)<(842094158);
      if ($switch$split452D) {
       switch ($12|0) {
       case 825382478:  {
        break;
       }
       default: {
        break L34;
       }
       }
       $sourceType$0 = 7;
       break L1;
      } else {
       switch ($12|0) {
       case 842094158:  {
        break;
       }
       default: {
        break L34;
       }
       }
       $sourceType$0 = 6;
       break L1;
      }
     }
    } while(0);
    $14 = (_SDL_SetError(602726,$vararg_buffer)|0);
    $$0 = $14;
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 $18 = (_GLES2_SelectProgram($renderer,$sourceType$0,$5)|0);
 $19 = ($18|0)<(0);
 if ($19) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $20 = ((($3)) + 24|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ($21|0)==(0);
 if (!($22)) {
  $23 = ((($1)) + 16|0);
  $24 = HEAP32[$23>>2]|0;
  FUNCTION_TABLE_vi[$24 & 8191](33986);
  $25 = ((($1)) + 28|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = ((($3)) + 4|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = ((($3)) + 32|0);
  $30 = HEAP32[$29>>2]|0;
  FUNCTION_TABLE_vii[$26 & 8191]($28,$30);
  $31 = HEAP32[$23>>2]|0;
  FUNCTION_TABLE_vi[$31 & 8191](33985);
  $32 = HEAP32[$25>>2]|0;
  $33 = HEAP32[$27>>2]|0;
  $34 = ((($3)) + 36|0);
  $35 = HEAP32[$34>>2]|0;
  FUNCTION_TABLE_vii[$32 & 8191]($33,$35);
  $36 = HEAP32[$23>>2]|0;
  FUNCTION_TABLE_vi[$36 & 8191](33984);
 }
 $37 = ((($3)) + 28|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = ($38|0)==(0);
 if ($39) {
  $$pre = ((($1)) + 28|0);
  $$pre1 = ((($3)) + 4|0);
  $$pre$phi2Z2D = $$pre1;$$pre$phiZ2D = $$pre;
 } else {
  $40 = ((($1)) + 16|0);
  $41 = HEAP32[$40>>2]|0;
  FUNCTION_TABLE_vi[$41 & 8191](33985);
  $42 = ((($1)) + 28|0);
  $43 = HEAP32[$42>>2]|0;
  $44 = ((($3)) + 4|0);
  $45 = HEAP32[$44>>2]|0;
  $46 = ((($3)) + 36|0);
  $47 = HEAP32[$46>>2]|0;
  FUNCTION_TABLE_vii[$43 & 8191]($45,$47);
  $48 = HEAP32[$40>>2]|0;
  FUNCTION_TABLE_vi[$48 & 8191](33984);
  $$pre$phi2Z2D = $44;$$pre$phiZ2D = $42;
 }
 $49 = HEAP32[$$pre$phiZ2D>>2]|0;
 $50 = HEAP32[$$pre$phi2Z2D>>2]|0;
 $51 = HEAP32[$3>>2]|0;
 FUNCTION_TABLE_vii[$49 & 8191]($50,$51);
 $52 = ((($texture)) + 29|0);
 $53 = HEAP8[$52>>0]|0;
 $54 = ((($texture)) + 31|0);
 $55 = HEAP8[$54>>0]|0;
 $56 = HEAP32[$6>>2]|0;
 $57 = ($56|0)==(0|0);
 L100: do {
  if ($57) {
   label = 36;
  } else {
   $58 = ((($56)) + 4|0);
   $59 = HEAP32[$58>>2]|0;
   $switch$split182D = ($59|0)<(372645892);
   if ($switch$split182D) {
    switch ($59|0) {
    case 370546692:  {
     break;
    }
    default: {
     label = 36;
     break L100;
    }
    }
   } else {
    switch ($59|0) {
    case 372645892:  {
     break;
    }
    default: {
     label = 36;
     break L100;
    }
    }
   }
   $60 = ((($texture)) + 30|0);
   $61 = ((($texture)) + 28|0);
   $b$0$in = $61;$r$0$in = $60;
  }
 } while(0);
 if ((label|0) == 36) {
  $62 = ((($texture)) + 28|0);
  $63 = ((($texture)) + 30|0);
  $b$0$in = $63;$r$0$in = $62;
 }
 $b$0 = HEAP8[$b$0$in>>0]|0;
 $r$0 = HEAP8[$r$0$in>>0]|0;
 $64 = ((($1)) + 268|0);
 $65 = HEAP32[$64>>2]|0;
 $66 = ((($65)) + 84|0);
 $67 = HEAP32[$66>>2]|0;
 $68 = (_llvm_bswap_i32(($67|0))|0);
 $69 = $r$0&255;
 $70 = $69 << 24;
 $71 = $53&255;
 $72 = $71 << 16;
 $73 = $b$0&255;
 $74 = $73 << 8;
 $75 = $55&255;
 $76 = $75 | $72;
 $77 = $76 | $74;
 $78 = $77 | $70;
 $79 = ($68|0)==($78|0);
 if (!($79)) {
  $80 = ((($65)) + 87|0);
  $81 = ((($65)) + 86|0);
  $82 = ((($65)) + 85|0);
  $83 = ((($1)) + 172|0);
  $84 = HEAP32[$83>>2]|0;
  $85 = ((($65)) + 24|0);
  $86 = HEAP32[$85>>2]|0;
  $87 = (+($r$0&255));
  $88 = $87 * 0.0039215688593685627;
  $89 = (+($53&255));
  $90 = $89 * 0.0039215688593685627;
  $91 = (+($b$0&255));
  $92 = $91 * 0.0039215688593685627;
  $93 = (+($55&255));
  $94 = $93 * 0.0039215688593685627;
  FUNCTION_TABLE_vidddd[$84 & 8191]($86,$88,$90,$92,$94);
  HEAP8[$66>>0] = $r$0;
  HEAP8[$82>>0] = $53;
  HEAP8[$81>>0] = $b$0;
  HEAP8[$80>>0] = $55;
 }
 $95 = ((($1)) + 8|0);
 $96 = HEAP32[$95>>2]|0;
 $97 = ($96|0)==($5|0);
 if (!($97)) {
  switch ($5|0) {
  case 4:  {
   $108 = ((($1)) + 80|0);
   $109 = HEAP32[$108>>2]|0;
   FUNCTION_TABLE_vi[$109 & 8191](3042);
   $110 = ((($1)) + 32|0);
   $111 = HEAP32[$110>>2]|0;
   FUNCTION_TABLE_viiii[$111 & 8191](0,768,0,1);
   break;
  }
  case 1:  {
   $100 = ((($1)) + 80|0);
   $101 = HEAP32[$100>>2]|0;
   FUNCTION_TABLE_vi[$101 & 8191](3042);
   $102 = ((($1)) + 32|0);
   $103 = HEAP32[$102>>2]|0;
   FUNCTION_TABLE_viiii[$103 & 8191](770,771,1,771);
   break;
  }
  case 2:  {
   $104 = ((($1)) + 80|0);
   $105 = HEAP32[$104>>2]|0;
   FUNCTION_TABLE_vi[$105 & 8191](3042);
   $106 = ((($1)) + 32|0);
   $107 = HEAP32[$106>>2]|0;
   FUNCTION_TABLE_viiii[$107 & 8191](770,1,0,1);
   break;
  }
  default: {
   $98 = ((($1)) + 68|0);
   $99 = HEAP32[$98>>2]|0;
   FUNCTION_TABLE_vi[$99 & 8191](3042);
  }
  }
  HEAP32[$95>>2] = $5;
 }
 $112 = ((($1)) + 12|0);
 $113 = HEAP32[$112>>2]|0;
 $114 = ($113|0)==(1);
 if ($114) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $115 = ((($1)) + 84|0);
 $116 = HEAP32[$115>>2]|0;
 FUNCTION_TABLE_vi[$116 & 8191](1);
 HEAP32[$112>>2] = 1;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _GLES2_RenderCopyEx($renderer,$texture,$srcrect,$dstrect,$angle,$center,$flip) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 $srcrect = $srcrect|0;
 $dstrect = $dstrect|0;
 $angle = +$angle;
 $center = $center|0;
 $flip = $flip|0;
 var $$$021$i = 0, $$0 = 0, $$0$i$i = 0, $$021$i = 0, $$pre$i = 0, $$pre$i1 = 0, $$pre$i2 = 0, $$pre$i4 = 0, $$pre$i6 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0.0, $117 = 0, $118 = 0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0, $123 = 0.0, $124 = 0, $125 = 0;
 var $126 = 0.0, $127 = 0.0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0.0, $133 = 0.0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0, $181 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0.0, $47 = 0.0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0.0;
 var $53 = 0.0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fAngle = 0, $texCoords = 0, $translate = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0, $vararg_ptr4 = 0, $vararg_ptr5 = 0;
 var $vertices = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $vertices = sp + 104|0;
 $texCoords = sp + 72|0;
 $translate = sp + 40|0;
 $fAngle = sp + 24|0;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[300738]|0;
 $3 = HEAP32[$1>>2]|0;
 $4 = ($2|0)==($3|0);
 if ($4) {
  $12 = $1;
  label = 4;
 } else {
  $5 = ((($1)) + 268|0);
  HEAP32[$5>>2] = 0;
  $6 = ((($renderer)) + 188|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (_SDL_GL_MakeCurrent($7,$3)|0);
  $9 = ($8|0)<(0);
  if (!($9)) {
   $10 = HEAP32[$1>>2]|0;
   HEAP32[300738] = $10;
   (_GLES2_UpdateViewport($renderer)|0);
   $$pre$i = HEAP32[$0>>2]|0;
   $12 = $$pre$i;
   label = 4;
  }
 }
 if ((label|0) == 4) {
  $11 = ((($12)) + 4|0);
  $13 = HEAP32[$11>>2]|0;
  $14 = ($13|0)==(0);
  if (!($14)) {
   $15 = ((($12)) + 108|0);
   while(1) {
    $16 = HEAP32[$15>>2]|0;
    $17 = (FUNCTION_TABLE_i[$16 & 8191]()|0);
    $18 = ($17|0)==(0);
    if ($18) {
     break;
    }
   }
  }
 }
 $19 = (_GLES2_SetupCopy($renderer,$texture)|0);
 $20 = ($19|0)<(0);
 if ($20) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $21 = ((($1)) + 84|0);
 $22 = HEAP32[$21>>2]|0;
 FUNCTION_TABLE_vi[$22 & 8191](3);
 $23 = HEAP32[$21>>2]|0;
 FUNCTION_TABLE_vi[$23 & 8191](2);
 $24 = 360.0 - $angle;
 $25 = $24;
 $26 = ((($fAngle)) + 12|0);
 HEAPF32[$26>>2] = $25;
 $27 = ((($fAngle)) + 8|0);
 HEAPF32[$27>>2] = $25;
 $28 = ((($fAngle)) + 4|0);
 HEAPF32[$28>>2] = $25;
 HEAPF32[$fAngle>>2] = $25;
 $29 = +HEAPF32[$center>>2];
 $30 = +HEAPF32[$dstrect>>2];
 $31 = $29 + $30;
 $32 = ((($translate)) + 24|0);
 HEAPF32[$32>>2] = $31;
 $33 = ((($translate)) + 16|0);
 HEAPF32[$33>>2] = $31;
 $34 = ((($translate)) + 8|0);
 HEAPF32[$34>>2] = $31;
 HEAPF32[$translate>>2] = $31;
 $35 = ((($center)) + 4|0);
 $36 = +HEAPF32[$35>>2];
 $37 = ((($dstrect)) + 4|0);
 $38 = +HEAPF32[$37>>2];
 $39 = $36 + $38;
 $40 = ((($translate)) + 28|0);
 HEAPF32[$40>>2] = $39;
 $41 = ((($translate)) + 20|0);
 HEAPF32[$41>>2] = $39;
 $42 = ((($translate)) + 12|0);
 HEAPF32[$42>>2] = $39;
 $43 = ((($translate)) + 4|0);
 HEAPF32[$43>>2] = $39;
 HEAPF32[$vertices>>2] = $30;
 $44 = ((($vertices)) + 4|0);
 HEAPF32[$44>>2] = $38;
 $45 = ((($dstrect)) + 8|0);
 $46 = +HEAPF32[$45>>2];
 $47 = $30 + $46;
 $48 = ((($vertices)) + 8|0);
 HEAPF32[$48>>2] = $47;
 $49 = ((($vertices)) + 12|0);
 HEAPF32[$49>>2] = $38;
 $50 = ((($vertices)) + 16|0);
 HEAPF32[$50>>2] = $30;
 $51 = ((($dstrect)) + 12|0);
 $52 = +HEAPF32[$51>>2];
 $53 = $38 + $52;
 $54 = ((($vertices)) + 20|0);
 HEAPF32[$54>>2] = $53;
 $55 = ((($vertices)) + 24|0);
 HEAPF32[$55>>2] = $47;
 $56 = ((($vertices)) + 28|0);
 HEAPF32[$56>>2] = $53;
 $57 = $flip & 1;
 $58 = ($57|0)==(0);
 if (!($58)) {
  HEAPF32[$50>>2] = $47;
  HEAPF32[$vertices>>2] = $47;
  HEAPF32[$55>>2] = $30;
  HEAPF32[$48>>2] = $30;
 }
 $59 = $flip & 2;
 $60 = ($59|0)==(0);
 if (!($60)) {
  HEAPF32[$49>>2] = $53;
  HEAPF32[$44>>2] = $53;
  HEAPF32[$56>>2] = $38;
  HEAPF32[$54>>2] = $38;
 }
 $61 = HEAP32[$0>>2]|0;
 $62 = ((($61)) + 284|0);
 $63 = HEAP32[$62>>2]|0;
 $64 = ($63|0)==(0);
 if ($64) {
  $65 = ((($61)) + 216|0);
  $66 = HEAP32[$65>>2]|0;
  FUNCTION_TABLE_vii[$66 & 8191](1,$62);
  $$pre$i1 = HEAP32[$62>>2]|0;
  $69 = $$pre$i1;
 } else {
  $69 = $63;
 }
 $67 = ((($61)) + 220|0);
 $68 = HEAP32[$67>>2]|0;
 FUNCTION_TABLE_vii[$68 & 8191](34962,$69);
 $70 = ((($61)) + 300|0);
 $71 = HEAP32[$70>>2]|0;
 $72 = ($71>>>0)<(16);
 if ($72) {
  $73 = ((($61)) + 224|0);
  $74 = HEAP32[$73>>2]|0;
  FUNCTION_TABLE_viiii[$74 & 8191](34962,16,$fAngle,35040);
  HEAP32[$70>>2] = 16;
 } else {
  $75 = ((($61)) + 228|0);
  $76 = HEAP32[$75>>2]|0;
  FUNCTION_TABLE_viiii[$76 & 8191](34962,0,16,$fAngle);
 }
 $77 = ((($61)) + 184|0);
 $78 = HEAP32[$77>>2]|0;
 FUNCTION_TABLE_viiiiii[$78 & 8191](2,1,5126,0,0,0);
 $79 = HEAP32[$0>>2]|0;
 $80 = ((($79)) + 288|0);
 $81 = HEAP32[$80>>2]|0;
 $82 = ($81|0)==(0);
 if ($82) {
  $83 = ((($79)) + 216|0);
  $84 = HEAP32[$83>>2]|0;
  FUNCTION_TABLE_vii[$84 & 8191](1,$80);
  $$pre$i2 = HEAP32[$80>>2]|0;
  $87 = $$pre$i2;
 } else {
  $87 = $81;
 }
 $85 = ((($79)) + 220|0);
 $86 = HEAP32[$85>>2]|0;
 FUNCTION_TABLE_vii[$86 & 8191](34962,$87);
 $88 = ((($79)) + 304|0);
 $89 = HEAP32[$88>>2]|0;
 $90 = ($89>>>0)<(32);
 if ($90) {
  $91 = ((($79)) + 224|0);
  $92 = HEAP32[$91>>2]|0;
  FUNCTION_TABLE_viiii[$92 & 8191](34962,32,$translate,35040);
  HEAP32[$88>>2] = 32;
 } else {
  $93 = ((($79)) + 228|0);
  $94 = HEAP32[$93>>2]|0;
  FUNCTION_TABLE_viiii[$94 & 8191](34962,0,32,$translate);
 }
 $95 = ((($79)) + 184|0);
 $96 = HEAP32[$95>>2]|0;
 FUNCTION_TABLE_viiiiii[$96 & 8191](3,2,5126,0,0,0);
 $97 = HEAP32[$0>>2]|0;
 $98 = ((($97)) + 276|0);
 $99 = HEAP32[$98>>2]|0;
 $100 = ($99|0)==(0);
 if ($100) {
  $101 = ((($97)) + 216|0);
  $102 = HEAP32[$101>>2]|0;
  FUNCTION_TABLE_vii[$102 & 8191](1,$98);
  $$pre$i4 = HEAP32[$98>>2]|0;
  $105 = $$pre$i4;
 } else {
  $105 = $99;
 }
 $103 = ((($97)) + 220|0);
 $104 = HEAP32[$103>>2]|0;
 FUNCTION_TABLE_vii[$104 & 8191](34962,$105);
 $106 = ((($97)) + 292|0);
 $107 = HEAP32[$106>>2]|0;
 $108 = ($107>>>0)<(32);
 if ($108) {
  $109 = ((($97)) + 224|0);
  $110 = HEAP32[$109>>2]|0;
  FUNCTION_TABLE_viiii[$110 & 8191](34962,32,$vertices,35040);
  HEAP32[$106>>2] = 32;
 } else {
  $111 = ((($97)) + 228|0);
  $112 = HEAP32[$111>>2]|0;
  FUNCTION_TABLE_viiii[$112 & 8191](34962,0,32,$vertices);
 }
 $113 = ((($97)) + 184|0);
 $114 = HEAP32[$113>>2]|0;
 FUNCTION_TABLE_viiiiii[$114 & 8191](0,2,5126,0,0,0);
 $115 = HEAP32[$srcrect>>2]|0;
 $116 = (+($115|0));
 $117 = ((($texture)) + 12|0);
 $118 = HEAP32[$117>>2]|0;
 $119 = (+($118|0));
 $120 = $116 / $119;
 HEAPF32[$texCoords>>2] = $120;
 $121 = ((($srcrect)) + 4|0);
 $122 = HEAP32[$121>>2]|0;
 $123 = (+($122|0));
 $124 = ((($texture)) + 16|0);
 $125 = HEAP32[$124>>2]|0;
 $126 = (+($125|0));
 $127 = $123 / $126;
 $128 = ((($texCoords)) + 4|0);
 HEAPF32[$128>>2] = $127;
 $129 = ((($srcrect)) + 8|0);
 $130 = HEAP32[$129>>2]|0;
 $131 = (($130) + ($115))|0;
 $132 = (+($131|0));
 $133 = $132 / $119;
 $134 = ((($texCoords)) + 8|0);
 HEAPF32[$134>>2] = $133;
 $135 = ((($texCoords)) + 12|0);
 HEAPF32[$135>>2] = $127;
 $136 = ((($texCoords)) + 16|0);
 HEAPF32[$136>>2] = $120;
 $137 = ((($srcrect)) + 12|0);
 $138 = HEAP32[$137>>2]|0;
 $139 = (($138) + ($122))|0;
 $140 = (+($139|0));
 $141 = $140 / $126;
 $142 = ((($texCoords)) + 20|0);
 HEAPF32[$142>>2] = $141;
 $143 = ((($texCoords)) + 24|0);
 HEAPF32[$143>>2] = $133;
 $144 = ((($texCoords)) + 28|0);
 HEAPF32[$144>>2] = $141;
 $145 = HEAP32[$0>>2]|0;
 $146 = ((($145)) + 280|0);
 $147 = HEAP32[$146>>2]|0;
 $148 = ($147|0)==(0);
 if ($148) {
  $149 = ((($145)) + 216|0);
  $150 = HEAP32[$149>>2]|0;
  FUNCTION_TABLE_vii[$150 & 8191](1,$146);
  $$pre$i6 = HEAP32[$146>>2]|0;
  $153 = $$pre$i6;
 } else {
  $153 = $147;
 }
 $151 = ((($145)) + 220|0);
 $152 = HEAP32[$151>>2]|0;
 FUNCTION_TABLE_vii[$152 & 8191](34962,$153);
 $154 = ((($145)) + 296|0);
 $155 = HEAP32[$154>>2]|0;
 $156 = ($155>>>0)<(32);
 if ($156) {
  $157 = ((($145)) + 224|0);
  $158 = HEAP32[$157>>2]|0;
  FUNCTION_TABLE_viiii[$158 & 8191](34962,32,$texCoords,35040);
  HEAP32[$154>>2] = 32;
 } else {
  $159 = ((($145)) + 228|0);
  $160 = HEAP32[$159>>2]|0;
  FUNCTION_TABLE_viiii[$160 & 8191](34962,0,32,$texCoords);
 }
 $161 = ((($145)) + 184|0);
 $162 = HEAP32[$161>>2]|0;
 FUNCTION_TABLE_viiiiii[$162 & 8191](1,2,5126,0,0,0);
 $163 = ((($1)) + 76|0);
 $164 = HEAP32[$163>>2]|0;
 FUNCTION_TABLE_viii[$164 & 8191](5,0,4);
 $165 = ((($1)) + 72|0);
 $166 = HEAP32[$165>>2]|0;
 FUNCTION_TABLE_vi[$166 & 8191](3);
 $167 = HEAP32[$165>>2]|0;
 FUNCTION_TABLE_vi[$167 & 8191](2);
 $168 = HEAP32[$0>>2]|0;
 $169 = ((($168)) + 4|0);
 $170 = HEAP32[$169>>2]|0;
 $171 = ($170|0)==(0);
 if ($171) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $172 = ((($168)) + 108|0);
 $173 = HEAP32[$172>>2]|0;
 $174 = (FUNCTION_TABLE_i[$173 & 8191]()|0);
 $175 = ($174|0)==(0);
 if ($175) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $$021$i = 1282466;$178 = $174;
 }
 while(1) {
  $176 = HEAP8[$$021$i>>0]|0;
  $177 = ($176<<24>>24)==(0);
  $$$021$i = $177 ? 597716 : $$021$i;
  switch ($178|0) {
  case 1280:  {
   $$0$i$i = 597787;
   break;
  }
  case 1281:  {
   $$0$i$i = 597770;
   break;
  }
  case 1282:  {
   $$0$i$i = 597749;
   break;
  }
  case 1285:  {
   $$0$i$i = 597732;
   break;
  }
  default: {
   $$0$i$i = 597724;
  }
  }
  HEAP32[$vararg_buffer>>2] = $$$021$i;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = 597829;
  $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
  HEAP32[$vararg_ptr2>>2] = 1812;
  $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
  HEAP32[$vararg_ptr3>>2] = 602842;
  $vararg_ptr4 = ((($vararg_buffer)) + 16|0);
  HEAP32[$vararg_ptr4>>2] = $$0$i$i;
  $vararg_ptr5 = ((($vararg_buffer)) + 20|0);
  HEAP32[$vararg_ptr5>>2] = $178;
  (_SDL_SetError(597803,$vararg_buffer)|0);
  $179 = HEAP32[$172>>2]|0;
  $180 = (FUNCTION_TABLE_i[$179 & 8191]()|0);
  $181 = ($180|0)==(0);
  if ($181) {
   $$0 = -1;
   break;
  } else {
   $$021$i = $$$021$i;$178 = $180;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _GLES2_RenderReadPixels($renderer,$rect,$pixel_format,$pixels,$pitch) {
 $renderer = $renderer|0;
 $rect = $rect|0;
 $pixel_format = $pixel_format|0;
 $pixels = $pixels|0;
 $pitch = $pitch|0;
 var $$0 = 0, $$0$i$i = 0, $$alloca_mul = 0, $$off = 0, $$pre = 0, $$pre$i = 0, $$pre8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$05 = 0, $h = 0, $rows$07 = 0, $src$06 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0;
 var $vararg_ptr2 = 0, $vararg_ptr3 = 0, $vararg_ptr4 = 0, $vararg_ptr5 = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $w = sp + 28|0;
 $h = sp + 24|0;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[300738]|0;
 $3 = HEAP32[$1>>2]|0;
 $4 = ($2|0)==($3|0);
 if ($4) {
  $12 = $1;
  label = 4;
 } else {
  $5 = ((($1)) + 268|0);
  HEAP32[$5>>2] = 0;
  $6 = ((($renderer)) + 188|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (_SDL_GL_MakeCurrent($7,$3)|0);
  $9 = ($8|0)<(0);
  if (!($9)) {
   $10 = HEAP32[$1>>2]|0;
   HEAP32[300738] = $10;
   (_GLES2_UpdateViewport($renderer)|0);
   $$pre$i = HEAP32[$0>>2]|0;
   $12 = $$pre$i;
   label = 4;
  }
 }
 if ((label|0) == 4) {
  $11 = ((($12)) + 4|0);
  $13 = HEAP32[$11>>2]|0;
  $14 = ($13|0)==(0);
  if (!($14)) {
   $15 = ((($12)) + 108|0);
   while(1) {
    $16 = HEAP32[$15>>2]|0;
    $17 = (FUNCTION_TABLE_i[$16 & 8191]()|0);
    $18 = ($17|0)==(0);
    if ($18) {
     break;
    }
   }
  }
 }
 $19 = ((($rect)) + 8|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = $20 << 2;
 $22 = ((($rect)) + 12|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = Math_imul($23, $21)|0;
 $25 = (_SDL_malloc($24)|0);
 $26 = ($25|0)==(0|0);
 if ($26) {
  $27 = (_SDL_Error(0)|0);
  $$0 = $27;
  STACKTOP = sp;return ($$0|0);
 }
 (_SDL_GetRendererOutputSize($renderer,$w,$h)|0);
 $28 = ((($1)) + 140|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = HEAP32[$rect>>2]|0;
 $31 = HEAP32[$h>>2]|0;
 $32 = ((($rect)) + 4|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = (($31) - ($33))|0;
 $35 = HEAP32[$22>>2]|0;
 $36 = (($34) - ($35))|0;
 $37 = HEAP32[$19>>2]|0;
 FUNCTION_TABLE_viiiiiii[$29 & 8191]($30,$36,$37,$35,6408,5121,$25);
 $38 = HEAP32[$0>>2]|0;
 $39 = ((($38)) + 4|0);
 $40 = HEAP32[$39>>2]|0;
 $41 = ($40|0)==(0);
 if (!($41)) {
  $42 = ((($38)) + 108|0);
  $43 = HEAP32[$42>>2]|0;
  $44 = (FUNCTION_TABLE_i[$43 & 8191]()|0);
  $45 = ($44|0)==(0);
  if (!($45)) {
   $46 = $44;
   while(1) {
    switch ($46|0) {
    case 1280:  {
     $$0$i$i = 597787;
     break;
    }
    case 1281:  {
     $$0$i$i = 597770;
     break;
    }
    case 1282:  {
     $$0$i$i = 597749;
     break;
    }
    case 1285:  {
     $$0$i$i = 597732;
     break;
    }
    default: {
     $$0$i$i = 597724;
    }
    }
    HEAP32[$vararg_buffer>>2] = 602991;
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
    HEAP32[$vararg_ptr1>>2] = 597829;
    $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
    HEAP32[$vararg_ptr2>>2] = 1839;
    $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
    HEAP32[$vararg_ptr3>>2] = 603006;
    $vararg_ptr4 = ((($vararg_buffer)) + 16|0);
    HEAP32[$vararg_ptr4>>2] = $$0$i$i;
    $vararg_ptr5 = ((($vararg_buffer)) + 20|0);
    HEAP32[$vararg_ptr5>>2] = $46;
    (_SDL_SetError(597803,$vararg_buffer)|0);
    $47 = HEAP32[$42>>2]|0;
    $48 = (FUNCTION_TABLE_i[$47 & 8191]()|0);
    $49 = ($48|0)==(0);
    if ($49) {
     $$0 = -1;
     break;
    } else {
     $46 = $48;
    }
   }
   STACKTOP = sp;return ($$0|0);
  }
 }
 $50 = HEAP32[$19>>2]|0;
 $51 = $50 << 2;
 $52 = HEAP32[$22>>2]|0;
 $$alloca_mul = $51;
 $53 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $$off = (($52) + 1)|0;
 $54 = ($$off>>>0)<(3);
 if ($54) {
  $64 = $50;$65 = $52;
 } else {
  $55 = (($52|0) / 2)&-1;
  $56 = (($52) + -1)|0;
  $57 = Math_imul($56, $21)|0;
  $58 = (($25) + ($57)|0);
  $59 = (0 - ($21))|0;
  $dst$05 = $25;$rows$07 = $55;$src$06 = $58;
  while(1) {
   $60 = (($rows$07) + -1)|0;
   (_SDL_memcpy($53,$dst$05,$51)|0);
   (_SDL_memcpy($dst$05,$src$06,$51)|0);
   (_SDL_memcpy($src$06,$53,$51)|0);
   $61 = (($dst$05) + ($21)|0);
   $62 = (($src$06) + ($59)|0);
   $63 = ($60|0)==(0);
   if ($63) {
    break;
   } else {
    $dst$05 = $61;$rows$07 = $60;$src$06 = $62;
   }
  }
  $$pre = HEAP32[$19>>2]|0;
  $$pre8 = HEAP32[$22>>2]|0;
  $64 = $$pre;$65 = $$pre8;
 }
 $66 = (_SDL_ConvertPixels($64,$65,376840196,$25,$21,$pixel_format,$pixels,$pitch)|0);
 _SDL_free($25);
 $$0 = $66;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_ConvertPixels($width,$height,$src_format,$src,$src_pitch,$dst_format,$dst,$dst_pitch) {
 $width = $width|0;
 $height = $height|0;
 $src_format = $src_format|0;
 $src = $src|0;
 $src_pitch = $src_pitch|0;
 $dst_format = $dst_format|0;
 $dst = $dst|0;
 $dst_pitch = $dst_pitch|0;
 var $$06$lcssa = 0, $$0633 = 0, $$08$lcssa = 0, $$0832 = 0, $$1 = 0, $$1726 = 0, $$1925 = 0, $$21028 = 0, $$229 = 0, $$mask$i17 = 0, $$mask11 = 0, $$not = 0, $$off$i = 0, $$off$i19 = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $9 = 0, $bpp$0 = 0, $dst_blitmap = 0, $dst_fmt = 0, $dst_surface = 0, $height$off = 0, $height$off37 = 0, $i$034 = 0, $i$127 = 0, $i$230 = 0, $or$cond$i18 = 0, $or$cond13 = 0, $rect = 0, $scevgep = 0, $scevgep40 = 0, $src_blitmap = 0, $src_fmt = 0, $src_surface = 0;
 var $switch$i = 0, $switch$i20 = 0, $switch$split102D = 0, $switch$split132D = 0, $switch$split162D = 0, $switch$split192D = 0, $switch$split222D = 0, $switch$split252D = 0, $switch$split282D = 0, $switch$split2D = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, $vararg_buffer6 = 0, $vararg_buffer8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 448|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer8 = sp + 32|0;
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $src_surface = sp + 380|0;
 $dst_surface = sp + 320|0;
 $src_fmt = sp + 276|0;
 $dst_fmt = sp + 232|0;
 $src_blitmap = sp + 144|0;
 $dst_blitmap = sp + 56|0;
 $rect = sp + 40|0;
 $0 = ($dst|0)==(0|0);
 if ($0) {
  HEAP32[$vararg_buffer>>2] = 603088;
  $1 = (_SDL_SetError(611267,$vararg_buffer)|0);
  $$1 = $1;
  STACKTOP = sp;return ($$1|0);
 }
 $2 = ($dst_pitch|0)==(0);
 if ($2) {
  HEAP32[$vararg_buffer1>>2] = 603092;
  $3 = (_SDL_SetError(611267,$vararg_buffer1)|0);
  $$1 = $3;
  STACKTOP = sp;return ($$1|0);
 }
 $4 = ($src_format|0)==($dst_format|0);
 $$not = ($src_format|0)==(0);
 $$mask11 = $src_format & -268435456;
 $5 = ($$mask11|0)==(268435456);
 $or$cond13 = $$not | $5;
 if (!($4)) {
  if ($or$cond13) {
   $32 = $src_format >>> 24;
   $33 = $32 & 15;
   $$off$i = (($33) + -1)|0;
   $switch$i = ($$off$i>>>0)<(3);
   if ($switch$i) {
    (_SDL_SetError(603130,$vararg_buffer6)|0);
    $$1 = -1;
    STACKTOP = sp;return ($$1|0);
   }
  }
  $34 = (_SDL_InitFormat($src_fmt,$src_format)|0);
  $35 = ($34|0)<(0);
  if ($35) {
   $$1 = -1;
   STACKTOP = sp;return ($$1|0);
  }
  (_SDL_memset($src_surface,0,60)|0);
  HEAP32[$src_surface>>2] = 1;
  $36 = ((($src_surface)) + 4|0);
  HEAP32[$36>>2] = $src_fmt;
  $37 = ((($src_surface)) + 20|0);
  HEAP32[$37>>2] = $src;
  $38 = ((($src_surface)) + 8|0);
  HEAP32[$38>>2] = $width;
  $39 = ((($src_surface)) + 12|0);
  HEAP32[$39>>2] = $height;
  $40 = ((($src_surface)) + 16|0);
  HEAP32[$40>>2] = $src_pitch;
  (_SDL_memset($src_blitmap,0,88)|0);
  $41 = ((($src_blitmap)) + 76|0);
  $42 = ((($src_surface)) + 52|0);
  HEAP32[$41>>2] = -1;
  HEAP32[$42>>2] = $src_blitmap;
  $43 = ((($src_surface)) + 56|0);
  HEAP32[$43>>2] = 1;
  $44 = ($dst_format|0)==(0);
  $$mask$i17 = $dst_format & -268435456;
  $45 = ($$mask$i17|0)==(268435456);
  $or$cond$i18 = $44 | $45;
  if ($or$cond$i18) {
   $46 = $dst_format >>> 24;
   $47 = $46 & 15;
   $$off$i19 = (($47) + -1)|0;
   $switch$i20 = ($$off$i19>>>0)<(3);
   if ($switch$i20) {
    (_SDL_SetError(603130,$vararg_buffer8)|0);
    $$1 = -1;
    STACKTOP = sp;return ($$1|0);
   }
  }
  $48 = (_SDL_InitFormat($dst_fmt,$dst_format)|0);
  $49 = ($48|0)<(0);
  if ($49) {
   $$1 = -1;
   STACKTOP = sp;return ($$1|0);
  }
  (_SDL_memset($dst_surface,0,60)|0);
  HEAP32[$dst_surface>>2] = 1;
  $50 = ((($dst_surface)) + 4|0);
  HEAP32[$50>>2] = $dst_fmt;
  $51 = ((($dst_surface)) + 20|0);
  HEAP32[$51>>2] = $dst;
  $52 = ((($dst_surface)) + 8|0);
  HEAP32[$52>>2] = $width;
  $53 = ((($dst_surface)) + 12|0);
  HEAP32[$53>>2] = $height;
  $54 = ((($dst_surface)) + 16|0);
  HEAP32[$54>>2] = $dst_pitch;
  (_SDL_memset($dst_blitmap,0,88)|0);
  $55 = ((($dst_blitmap)) + 76|0);
  $56 = ((($dst_surface)) + 52|0);
  HEAP32[$55>>2] = -1;
  HEAP32[$56>>2] = $dst_blitmap;
  $57 = ((($dst_surface)) + 56|0);
  HEAP32[$57>>2] = 1;
  HEAP32[$rect>>2] = 0;
  $58 = ((($rect)) + 4|0);
  HEAP32[$58>>2] = 0;
  $59 = ((($rect)) + 8|0);
  HEAP32[$59>>2] = $width;
  $60 = ((($rect)) + 12|0);
  HEAP32[$60>>2] = $height;
  $61 = HEAP32[$42>>2]|0;
  $62 = HEAP32[$61>>2]|0;
  $63 = ($62|0)==($dst_surface|0);
  do {
   if ($63) {
    $64 = HEAP32[$50>>2]|0;
    $65 = ((($64)) + 4|0);
    $66 = HEAP32[$65>>2]|0;
    $67 = ($66|0)==(0|0);
    if (!($67)) {
     $68 = ((($61)) + 80|0);
     $69 = HEAP32[$68>>2]|0;
     $70 = ((($66)) + 8|0);
     $71 = HEAP32[$70>>2]|0;
     $72 = ($69|0)==($71|0);
     if (!($72)) {
      label = 35;
      break;
     }
    }
    $73 = HEAP32[$36>>2]|0;
    $74 = ((($73)) + 4|0);
    $75 = HEAP32[$74>>2]|0;
    $76 = ($75|0)==(0|0);
    if ($76) {
     $85 = $61;
    } else {
     $77 = ((($61)) + 84|0);
     $78 = HEAP32[$77>>2]|0;
     $79 = ((($75)) + 8|0);
     $80 = HEAP32[$79>>2]|0;
     $81 = ($78|0)==($80|0);
     if ($81) {
      $85 = $61;
     } else {
      label = 35;
     }
    }
   } else {
    label = 35;
   }
  } while(0);
  do {
   if ((label|0) == 35) {
    $82 = (_SDL_MapSurface($src_surface,$dst_surface)|0);
    $83 = ($82|0)<(0);
    if ($83) {
     $$1 = -1;
     STACKTOP = sp;return ($$1|0);
    } else {
     $$pre$i = HEAP32[$42>>2]|0;
     $85 = $$pre$i;
     break;
    }
   }
  } while(0);
  $84 = ((($85)) + 8|0);
  $86 = HEAP32[$84>>2]|0;
  $87 = (FUNCTION_TABLE_iiiii[$86 & 8191]($src_surface,$rect,$dst_surface,$rect)|0);
  $$1 = $87;
  STACKTOP = sp;return ($$1|0);
 }
 L41: do {
  if ($or$cond13) {
   $7 = $src_format & 255;
   $bpp$0 = $7;
  } else {
   $switch$split2D = ($src_format|0)<(844715353);
   L44: do {
    if ($switch$split2D) {
     $switch$split132D = ($src_format|0)<(842094158);
     if ($switch$split132D) {
      switch ($src_format|0) {
      case 825382478:  {
       label = 8;
       break L44;
       break;
      }
      default: {
       break L44;
      }
      }
     } else {
      switch ($src_format|0) {
      case 842094158: case 842094169:  {
       label = 8;
       break L44;
       break;
      }
      default: {
       break L44;
      }
      }
     }
    } else {
     $switch$split162D = ($src_format|0)<(1448433993);
     if ($switch$split162D) {
      $switch$split252D = ($src_format|0)<(1431918169);
      if ($switch$split252D) {
       switch ($src_format|0) {
       case 844715353:  {
        $bpp$0 = 2;
        break L41;
        break;
       }
       default: {
        break L44;
       }
       }
      } else {
       switch ($src_format|0) {
       case 1431918169:  {
        $bpp$0 = 2;
        break L41;
        break;
       }
       default: {
        break L44;
       }
       }
      }
     } else {
      $switch$split282D = ($src_format|0)<(1498831189);
      if ($switch$split282D) {
       switch ($src_format|0) {
       case 1448433993:  {
        label = 8;
        break L44;
        break;
       }
       default: {
        break L44;
       }
       }
      } else {
       switch ($src_format|0) {
       case 1498831189:  {
        $bpp$0 = 2;
        break L41;
        break;
       }
       default: {
        break L44;
       }
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 8) {
    $bpp$0 = 1;
    break;
   }
   $6 = (_SDL_SetError(603102,$vararg_buffer4)|0);
   $$1 = $6;
   STACKTOP = sp;return ($$1|0);
  }
 } while(0);
 $8 = Math_imul($bpp$0, $width)|0;
 $9 = ($height|0)==(0);
 if ($9) {
  $$06$lcssa = $dst;$$08$lcssa = $src;
 } else {
  $10 = Math_imul($dst_pitch, $height)|0;
  $11 = Math_imul($src_pitch, $height)|0;
  $$0633 = $dst;$$0832 = $src;$i$034 = $height;
  while(1) {
   $12 = (($i$034) + -1)|0;
   (_SDL_memcpy($$0633,$$0832,$8)|0);
   $13 = (($$0832) + ($src_pitch)|0);
   $14 = (($$0633) + ($dst_pitch)|0);
   $15 = ($12|0)==(0);
   if ($15) {
    break;
   } else {
    $$0633 = $14;$$0832 = $13;$i$034 = $12;
   }
  }
  $scevgep = (($dst) + ($10)|0);
  $scevgep40 = (($src) + ($11)|0);
  $$06$lcssa = $scevgep;$$08$lcssa = $scevgep40;
 }
 $switch$split102D = ($src_format|0)<(842094169);
 if ($switch$split102D) {
  $switch$split192D = ($src_format|0)<(842094158);
  L72: do {
   if ($switch$split192D) {
    switch ($src_format|0) {
    case 825382478:  {
     break L72;
     break;
    }
    default: {
     $$1 = 0;
    }
    }
    STACKTOP = sp;return ($$1|0);
   } else {
    switch ($src_format|0) {
    case 842094158:  {
     break L72;
     break;
    }
    default: {
     $$1 = 0;
    }
    }
    STACKTOP = sp;return ($$1|0);
   }
  } while(0);
  $height$off = (($height) + 1)|0;
  $26 = ($height$off>>>0)<(3);
  if ($26) {
   $$1 = 0;
   STACKTOP = sp;return ($$1|0);
  }
  $27 = (($height|0) / 2)&-1;
  $$21028 = $$08$lcssa;$$229 = $$06$lcssa;$i$230 = $27;
  while(1) {
   $28 = (($i$230) + -1)|0;
   (_SDL_memcpy($$229,$$21028,$8)|0);
   $29 = (($$21028) + ($src_pitch)|0);
   $30 = (($$229) + ($dst_pitch)|0);
   $31 = ($28|0)==(0);
   if ($31) {
    $$1 = 0;
    break;
   } else {
    $$21028 = $29;$$229 = $30;$i$230 = $28;
   }
  }
  STACKTOP = sp;return ($$1|0);
 } else {
  $switch$split222D = ($src_format|0)<(1448433993);
  L85: do {
   if ($switch$split222D) {
    switch ($src_format|0) {
    case 842094169:  {
     break L85;
     break;
    }
    default: {
     $$1 = 0;
    }
    }
    STACKTOP = sp;return ($$1|0);
   } else {
    switch ($src_format|0) {
    case 1448433993:  {
     break L85;
     break;
    }
    default: {
     $$1 = 0;
    }
    }
    STACKTOP = sp;return ($$1|0);
   }
  } while(0);
  $16 = (($8|0) / 2)&-1;
  $17 = (($src_pitch|0) / 2)&-1;
  $18 = (($dst_pitch|0) / 2)&-1;
  $height$off37 = (($height) + 1)|0;
  $19 = ($height$off37>>>0)<(3);
  if ($19) {
   $$1 = 0;
   STACKTOP = sp;return ($$1|0);
  }
  $20 = (($height|0) / 2)&-1;
  $21 = $20 << 1;
  $$1726 = $$06$lcssa;$$1925 = $$08$lcssa;$i$127 = $21;
  while(1) {
   $22 = (($i$127) + -1)|0;
   (_SDL_memcpy($$1726,$$1925,$16)|0);
   $23 = (($$1925) + ($17)|0);
   $24 = (($$1726) + ($18)|0);
   $25 = ($22|0)==(0);
   if ($25) {
    $$1 = 0;
    break;
   } else {
    $$1726 = $24;$$1925 = $23;$i$127 = $22;
   }
  }
  STACKTOP = sp;return ($$1|0);
 }
 return (0)|0;
}
function _SDL_InitFormat($format,$pixel_format) {
 $format = $format|0;
 $pixel_format = $pixel_format|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa87 = 0, $$lcssa88 = 0, $$lcssa89 = 0, $$lcssa90 = 0, $$lcssa91 = 0, $$lcssa92 = 0, $$lcssa93 = 0, $$lcssa94 = 0, $$lcssa95 = 0, $$lcssa96 = 0, $$lcssa97 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $9 = 0, $Amask = 0, $Bmask = 0, $Gmask = 0, $Rmask = 0, $bpp = 0, $mask$0$lcssa = 0, $mask$026 = 0, $mask$124 = 0, $mask$2$lcssa = 0, $mask$221 = 0, $mask$319 = 0, $mask$4$lcssa = 0, $mask$416 = 0, $mask$514 = 0, $mask$6$lcssa = 0, $mask$612 = 0, $mask$711 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $bpp = sp + 16|0;
 $Rmask = sp + 12|0;
 $Gmask = sp + 8|0;
 $Bmask = sp + 4|0;
 $Amask = sp;
 $0 = (_SDL_PixelFormatEnumToMasks($pixel_format,$bpp,$Rmask,$Gmask,$Bmask,$Amask)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 (_SDL_memset($format,0,44)|0);
 HEAP32[$format>>2] = $pixel_format;
 $2 = HEAP32[$bpp>>2]|0;
 $3 = $2&255;
 $4 = ((($format)) + 8|0);
 HEAP8[$4>>0] = $3;
 $5 = (($2) + 7)|0;
 $6 = (($5|0) / 8)&-1;
 $7 = $6&255;
 $8 = ((($format)) + 9|0);
 HEAP8[$8>>0] = $7;
 $9 = HEAP32[$Rmask>>2]|0;
 $10 = ((($format)) + 12|0);
 HEAP32[$10>>2] = $9;
 $11 = ((($format)) + 32|0);
 HEAP8[$11>>0] = 0;
 $12 = ((($format)) + 28|0);
 HEAP8[$12>>0] = 8;
 $13 = ($9|0)==(0);
 if (!($13)) {
  $14 = $9 & 1;
  $15 = ($14|0)==(0);
  if ($15) {
   $19 = 0;$mask$026 = $9;
   while(1) {
    $18 = (($19) + 1)<<24>>24;
    $20 = $mask$026 >>> 1;
    $21 = $20 & 1;
    $22 = ($21|0)==(0);
    if ($22) {
     $19 = $18;$mask$026 = $20;
    } else {
     $$lcssa96 = $18;$$lcssa97 = $20;
     break;
    }
   }
   HEAP8[$11>>0] = $$lcssa96;
   $mask$0$lcssa = $$lcssa97;
  } else {
   $mask$0$lcssa = $9;
  }
  $16 = $mask$0$lcssa & 1;
  $17 = ($16|0)==(0);
  if (!($17)) {
   $24 = 8;$mask$124 = $mask$0$lcssa;
   while(1) {
    $23 = (($24) + -1)<<24>>24;
    $25 = $mask$124 >>> 1;
    $26 = $25 & 1;
    $27 = ($26|0)==(0);
    if ($27) {
     $$lcssa95 = $23;
     break;
    } else {
     $24 = $23;$mask$124 = $25;
    }
   }
   HEAP8[$12>>0] = $$lcssa95;
  }
 }
 $28 = HEAP32[$Gmask>>2]|0;
 $29 = ((($format)) + 16|0);
 HEAP32[$29>>2] = $28;
 $30 = ((($format)) + 33|0);
 HEAP8[$30>>0] = 0;
 $31 = ((($format)) + 29|0);
 HEAP8[$31>>0] = 8;
 $32 = ($28|0)==(0);
 if (!($32)) {
  $33 = $28 & 1;
  $34 = ($33|0)==(0);
  if ($34) {
   $38 = 0;$mask$221 = $28;
   while(1) {
    $37 = (($38) + 1)<<24>>24;
    $39 = $mask$221 >>> 1;
    $40 = $39 & 1;
    $41 = ($40|0)==(0);
    if ($41) {
     $38 = $37;$mask$221 = $39;
    } else {
     $$lcssa93 = $37;$$lcssa94 = $39;
     break;
    }
   }
   HEAP8[$30>>0] = $$lcssa93;
   $mask$2$lcssa = $$lcssa94;
  } else {
   $mask$2$lcssa = $28;
  }
  $35 = $mask$2$lcssa & 1;
  $36 = ($35|0)==(0);
  if (!($36)) {
   $43 = 8;$mask$319 = $mask$2$lcssa;
   while(1) {
    $42 = (($43) + -1)<<24>>24;
    $44 = $mask$319 >>> 1;
    $45 = $44 & 1;
    $46 = ($45|0)==(0);
    if ($46) {
     $$lcssa92 = $42;
     break;
    } else {
     $43 = $42;$mask$319 = $44;
    }
   }
   HEAP8[$31>>0] = $$lcssa92;
  }
 }
 $47 = HEAP32[$Bmask>>2]|0;
 $48 = ((($format)) + 20|0);
 HEAP32[$48>>2] = $47;
 $49 = ((($format)) + 34|0);
 HEAP8[$49>>0] = 0;
 $50 = ((($format)) + 30|0);
 HEAP8[$50>>0] = 8;
 $51 = ($47|0)==(0);
 if (!($51)) {
  $52 = $47 & 1;
  $53 = ($52|0)==(0);
  if ($53) {
   $57 = 0;$mask$416 = $47;
   while(1) {
    $56 = (($57) + 1)<<24>>24;
    $58 = $mask$416 >>> 1;
    $59 = $58 & 1;
    $60 = ($59|0)==(0);
    if ($60) {
     $57 = $56;$mask$416 = $58;
    } else {
     $$lcssa90 = $56;$$lcssa91 = $58;
     break;
    }
   }
   HEAP8[$49>>0] = $$lcssa90;
   $mask$4$lcssa = $$lcssa91;
  } else {
   $mask$4$lcssa = $47;
  }
  $54 = $mask$4$lcssa & 1;
  $55 = ($54|0)==(0);
  if (!($55)) {
   $62 = 8;$mask$514 = $mask$4$lcssa;
   while(1) {
    $61 = (($62) + -1)<<24>>24;
    $63 = $mask$514 >>> 1;
    $64 = $63 & 1;
    $65 = ($64|0)==(0);
    if ($65) {
     $$lcssa89 = $61;
     break;
    } else {
     $62 = $61;$mask$514 = $63;
    }
   }
   HEAP8[$50>>0] = $$lcssa89;
  }
 }
 $66 = HEAP32[$Amask>>2]|0;
 $67 = ((($format)) + 24|0);
 HEAP32[$67>>2] = $66;
 $68 = ((($format)) + 35|0);
 HEAP8[$68>>0] = 0;
 $69 = ((($format)) + 31|0);
 HEAP8[$69>>0] = 8;
 $70 = ($66|0)==(0);
 if (!($70)) {
  $71 = $66 & 1;
  $72 = ($71|0)==(0);
  if ($72) {
   $76 = 0;$mask$612 = $66;
   while(1) {
    $75 = (($76) + 1)<<24>>24;
    $77 = $mask$612 >>> 1;
    $78 = $77 & 1;
    $79 = ($78|0)==(0);
    if ($79) {
     $76 = $75;$mask$612 = $77;
    } else {
     $$lcssa87 = $75;$$lcssa88 = $77;
     break;
    }
   }
   HEAP8[$68>>0] = $$lcssa87;
   $mask$6$lcssa = $$lcssa88;
  } else {
   $mask$6$lcssa = $66;
  }
  $73 = $mask$6$lcssa & 1;
  $74 = ($73|0)==(0);
  if (!($74)) {
   $81 = 8;$mask$711 = $mask$6$lcssa;
   while(1) {
    $80 = (($81) + -1)<<24>>24;
    $82 = $mask$711 >>> 1;
    $83 = $82 & 1;
    $84 = ($83|0)==(0);
    if ($84) {
     $$lcssa = $80;
     break;
    } else {
     $81 = $80;$mask$711 = $82;
    }
   }
   HEAP8[$69>>0] = $$lcssa;
  }
 }
 $85 = ((($format)) + 4|0);
 HEAP32[$85>>2] = 0;
 $86 = ((($format)) + 36|0);
 HEAP32[$86>>2] = 1;
 $87 = ((($format)) + 40|0);
 HEAP32[$87>>2] = 0;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_PixelFormatEnumToMasks($format,$bpp,$Rmask,$Gmask,$Bmask,$Amask) {
 $format = $format|0;
 $bpp = $bpp|0;
 $Rmask = $Rmask|0;
 $Gmask = $Gmask|0;
 $Bmask = $Bmask|0;
 $Amask = $Amask|0;
 var $$0 = 0, $$mask5 = 0, $$not = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $masks$sroa$0$0 = 0, $masks$sroa$16$0 = 0, $masks$sroa$32$0 = 0;
 var $masks$sroa$48$0 = 0, $or$cond7 = 0, $storemerge = 0, $switch = 0, $switch$split2D = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $$not = ($format|0)==(0);
 $$mask5 = $format & -268435456;
 $0 = ($$mask5|0)==(268435456);
 $or$cond7 = $$not | $0;
 if (!($or$cond7)) {
  (_SDL_SetError(603166,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = $format & 255;
 $2 = ($1>>>0)<(3);
 $3 = $format << 3;
 $4 = $3 & 2040;
 $5 = $format >>> 8;
 $6 = $5 & 255;
 $storemerge = $2 ? $6 : $4;
 HEAP32[$bpp>>2] = $storemerge;
 HEAP32[$Amask>>2] = 0;
 HEAP32[$Bmask>>2] = 0;
 HEAP32[$Gmask>>2] = 0;
 HEAP32[$Rmask>>2] = 0;
 $switch$split2D = ($format|0)<(390076419);
 L5: do {
  if ($switch$split2D) {
   switch ($format|0) {
   case 386930691:  {
    break;
   }
   default: {
    break L5;
   }
   }
   HEAP32[$Rmask>>2] = 255;
   HEAP32[$Gmask>>2] = 65280;
   HEAP32[$Bmask>>2] = 16711680;
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
  } else {
   switch ($format|0) {
   case 390076419:  {
    break;
   }
   default: {
    break L5;
   }
   }
   HEAP32[$Rmask>>2] = 16711680;
   HEAP32[$Gmask>>2] = 65280;
   HEAP32[$Bmask>>2] = 255;
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 $7 = $format >>> 24;
 $8 = $7 & 15;
 $$off = (($8) + -4)|0;
 $switch = ($$off>>>0)<(3);
 if (!($switch)) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 $9 = $format >>> 16;
 $10 = $9 & 15;
 switch ($10|0) {
 case 1:  {
  $masks$sroa$0$0 = 0;$masks$sroa$16$0 = 224;$masks$sroa$32$0 = 28;$masks$sroa$48$0 = 3;
  break;
 }
 case 2:  {
  $masks$sroa$0$0 = 61440;$masks$sroa$16$0 = 3840;$masks$sroa$32$0 = 240;$masks$sroa$48$0 = 15;
  break;
 }
 case 3:  {
  $masks$sroa$0$0 = 32768;$masks$sroa$16$0 = 31744;$masks$sroa$32$0 = 992;$masks$sroa$48$0 = 31;
  break;
 }
 case 4:  {
  $masks$sroa$0$0 = 63488;$masks$sroa$16$0 = 1984;$masks$sroa$32$0 = 62;$masks$sroa$48$0 = 1;
  break;
 }
 case 5:  {
  $masks$sroa$0$0 = 0;$masks$sroa$16$0 = 63488;$masks$sroa$32$0 = 2016;$masks$sroa$48$0 = 31;
  break;
 }
 case 6:  {
  $masks$sroa$0$0 = -16777216;$masks$sroa$16$0 = 16711680;$masks$sroa$32$0 = 65280;$masks$sroa$48$0 = 255;
  break;
 }
 case 7:  {
  $masks$sroa$0$0 = -1073741824;$masks$sroa$16$0 = 1072693248;$masks$sroa$32$0 = 1047552;$masks$sroa$48$0 = 1023;
  break;
 }
 case 8:  {
  $masks$sroa$0$0 = -4194304;$masks$sroa$16$0 = 4190208;$masks$sroa$32$0 = 4092;$masks$sroa$48$0 = 3;
  break;
 }
 default: {
  (_SDL_SetError(603356,$vararg_buffer1)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 }
 $11 = $format >>> 20;
 $12 = $11 & 15;
 switch ($12|0) {
 case 1:  {
  HEAP32[$Rmask>>2] = $masks$sroa$16$0;
  HEAP32[$Gmask>>2] = $masks$sroa$32$0;
  HEAP32[$Bmask>>2] = $masks$sroa$48$0;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 case 2:  {
  HEAP32[$Rmask>>2] = $masks$sroa$0$0;
  HEAP32[$Gmask>>2] = $masks$sroa$16$0;
  HEAP32[$Bmask>>2] = $masks$sroa$32$0;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 case 3:  {
  HEAP32[$Amask>>2] = $masks$sroa$0$0;
  HEAP32[$Rmask>>2] = $masks$sroa$16$0;
  HEAP32[$Gmask>>2] = $masks$sroa$32$0;
  HEAP32[$Bmask>>2] = $masks$sroa$48$0;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 case 4:  {
  HEAP32[$Rmask>>2] = $masks$sroa$0$0;
  HEAP32[$Gmask>>2] = $masks$sroa$16$0;
  HEAP32[$Bmask>>2] = $masks$sroa$32$0;
  HEAP32[$Amask>>2] = $masks$sroa$48$0;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 case 5:  {
  HEAP32[$Bmask>>2] = $masks$sroa$16$0;
  HEAP32[$Gmask>>2] = $masks$sroa$32$0;
  HEAP32[$Rmask>>2] = $masks$sroa$48$0;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 case 6:  {
  HEAP32[$Bmask>>2] = $masks$sroa$0$0;
  HEAP32[$Gmask>>2] = $masks$sroa$16$0;
  HEAP32[$Rmask>>2] = $masks$sroa$32$0;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 case 8:  {
  HEAP32[$Bmask>>2] = $masks$sroa$0$0;
  HEAP32[$Gmask>>2] = $masks$sroa$16$0;
  HEAP32[$Rmask>>2] = $masks$sroa$32$0;
  HEAP32[$Amask>>2] = $masks$sroa$48$0;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 case 7:  {
  HEAP32[$Amask>>2] = $masks$sroa$0$0;
  HEAP32[$Bmask>>2] = $masks$sroa$16$0;
  HEAP32[$Gmask>>2] = $masks$sroa$32$0;
  HEAP32[$Rmask>>2] = $masks$sroa$48$0;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 default: {
  (_SDL_SetError(603356,$vararg_buffer3)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 }
 return (0)|0;
}
function _SDL_MapSurface($src,$dst) {
 $src = $src|0;
 $dst = $dst|0;
 var $$$i = 0, $$0 = 0, $$idx = 0, $$idx$val = 0, $$idx14 = 0, $$idx14$val = 0, $$mask = 0, $$mask1 = 0, $$mask2 = 0, $$off = 0, $$off10 = 0, $$off12 = 0, $$sink = 0, $$sink16 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0;
 var $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0;
 var $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0;
 var $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0;
 var $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $colors$i = 0, $dithered$i = 0, $div$i = 0, $div1$i = 0, $div2$i = 0, $exitcond$i$i = 0;
 var $i$01$i = 0, $i$01$i$i = 0, $or$cond = 0, $or$cond5 = 0, $or$cond6 = 0, $or$cond8 = 0, $or$cond9 = 0, $switch = 0, $switch11 = 0, $switch13 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1040|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $dithered$i = sp;
 $colors$i = sp + 16|0;
 $0 = ((($src)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$src>>2]|0;
 $3 = $2 & 2;
 $4 = ($3|0)==(0);
 if (!($4)) {
  _SDL_UnRLESurface($src,1);
 }
 $5 = ($1|0)==(0|0);
 if (!($5)) {
  $6 = HEAP32[$1>>2]|0;
  $7 = ($6|0)==(0|0);
  if (!($7)) {
   $8 = ((($6)) + 56|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (($9) + -1)|0;
   HEAP32[$8>>2] = $10;
   $11 = ($9|0)<(2);
   if ($11) {
    _SDL_FreeSurface($6);
   }
  }
  HEAP32[$1>>2] = 0;
  $12 = ((($1)) + 84|0);
  HEAP32[$12>>2] = 0;
  $13 = ((($1)) + 80|0);
  HEAP32[$13>>2] = 0;
  $14 = ((($1)) + 64|0);
  $15 = HEAP32[$14>>2]|0;
  _SDL_free($15);
  HEAP32[$14>>2] = 0;
 }
 $16 = ((($1)) + 4|0);
 HEAP32[$16>>2] = 0;
 $17 = ((($src)) + 4|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($dst)) + 4|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = HEAP32[$18>>2]|0;
 $22 = ($21|0)==(0);
 $$mask2 = $21 & -268435456;
 $23 = ($$mask2|0)==(268435456);
 $or$cond = $22 | $23;
 do {
  if ($or$cond) {
   $24 = $21 >>> 24;
   $25 = $24 & 15;
   $$off = (($25) + -1)|0;
   $switch = ($$off>>>0)<(3);
   if ($switch) {
    $26 = HEAP32[$20>>2]|0;
    $27 = ($26|0)==(0);
    $$mask1 = $26 & -268435456;
    $28 = ($$mask1|0)==(268435456);
    $or$cond5 = $27 | $28;
    if ($or$cond5) {
     $29 = $26 >>> 24;
     $30 = $29 & 15;
     $$off10 = (($30) + -1)|0;
     $switch11 = ($$off10>>>0)<(3);
     if ($switch11) {
      $31 = ((($18)) + 4|0);
      $32 = HEAP32[$31>>2]|0;
      $33 = ((($20)) + 4|0);
      $34 = HEAP32[$33>>2]|0;
      $35 = (_Map1to1($32,$34,$16)|0);
      $36 = ((($1)) + 64|0);
      HEAP32[$36>>2] = $35;
      $37 = HEAP32[$16>>2]|0;
      $38 = ($37|0)==(0);
      $39 = ($35|0)==(0|0);
      $or$cond6 = $39 & $38;
      if ($or$cond6) {
       $$0 = -1;
       STACKTOP = sp;return ($$0|0);
      }
      $40 = ((($18)) + 8|0);
      $41 = HEAP8[$40>>0]|0;
      $42 = ((($20)) + 8|0);
      $43 = HEAP8[$42>>0]|0;
      $44 = ($41<<24>>24)==($43<<24>>24);
      if ($44) {
       label = 35;
       break;
      }
      HEAP32[$16>>2] = 0;
      label = 35;
      break;
     }
    }
    $45 = HEAP32[$0>>2]|0;
    $46 = ((($45)) + 76|0);
    $47 = HEAP32[$46>>2]|0;
    $48 = $47 >>> 8;
    $49 = $48 & 255;
    $50 = $47 >>> 16;
    $51 = $47 >>> 24;
    $$idx = ((($18)) + 4|0);
    $$idx$val = HEAP32[$$idx>>2]|0;
    $52 = ((($20)) + 9|0);
    $53 = HEAP8[$52>>0]|0;
    $54 = ($53<<24>>24)==(3);
    $55 = $53&255;
    $$$i = $54 ? 4 : $55;
    $56 = HEAP32[$$idx$val>>2]|0;
    $57 = Math_imul($$$i, $56)|0;
    $58 = (_SDL_malloc($57)|0);
    $59 = ($58|0)==(0|0);
    if ($59) {
     (_SDL_Error(0)|0);
     $70 = ((($1)) + 64|0);
     HEAP32[$70>>2] = 0;
     $$0 = -1;
     STACKTOP = sp;return ($$0|0);
    }
    $60 = HEAP32[$$idx$val>>2]|0;
    $61 = ($60|0)>(0);
    if (!($61)) {
     $62 = ((($1)) + 64|0);
     HEAP32[$62>>2] = $58;
     label = 35;
     break;
    }
    $63 = ((($$idx$val)) + 4|0);
    $64 = $47 & 255;
    $65 = $50 & 255;
    $66 = ((($20)) + 28|0);
    $67 = ((($20)) + 32|0);
    $68 = ((($20)) + 33|0);
    $69 = ((($20)) + 34|0);
    $i$01$i = 0;
    while(1) {
     $71 = HEAP32[$63>>2]|0;
     $72 = (($71) + ($i$01$i<<2)|0);
     $73 = HEAP8[$72>>0]|0;
     $74 = $73&255;
     $75 = Math_imul($74, $64)|0;
     $76 = (($75>>>0) / 255)&-1;
     $77 = (((($71) + ($i$01$i<<2)|0)) + 1|0);
     $78 = HEAP8[$77>>0]|0;
     $79 = $78&255;
     $80 = Math_imul($79, $49)|0;
     $81 = (($80>>>0) / 255)&-1;
     $82 = (((($71) + ($i$01$i<<2)|0)) + 2|0);
     $83 = HEAP8[$82>>0]|0;
     $84 = $83&255;
     $85 = Math_imul($84, $65)|0;
     $86 = (($85>>>0) / 255)&-1;
     $87 = (((($71) + ($i$01$i<<2)|0)) + 3|0);
     $88 = HEAP8[$87>>0]|0;
     $89 = $88&255;
     $90 = Math_imul($89, $51)|0;
     $91 = (($90>>>0) / 255)&-1;
     $92 = HEAP8[$52>>0]|0;
     $93 = $92&255;
     switch ($93|0) {
     case 1:  {
      $94 = $76 & 255;
      $95 = HEAP32[$66>>2]|0;
      $96 = $95 & 255;
      $97 = $94 >>> $96;
      $98 = HEAP32[$67>>2]|0;
      $99 = $98 & 255;
      $100 = $97 << $99;
      $101 = $81 & 255;
      $102 = $95 >>> 8;
      $103 = $102 & 255;
      $104 = $101 >>> $103;
      $105 = $98 >>> 8;
      $106 = $105 & 255;
      $107 = $104 << $106;
      $108 = $86 & 255;
      $109 = $95 >>> 16;
      $110 = $109 & 255;
      $111 = $108 >>> $110;
      $112 = $98 >>> 16;
      $113 = $112 & 255;
      $114 = $111 << $113;
      $115 = $91 & 255;
      $116 = $95 >>> 24;
      $117 = $115 >>> $116;
      $118 = $98 >>> 24;
      $119 = $117 << $118;
      $120 = $119 | $100;
      $121 = $120 | $107;
      $122 = $121 | $114;
      $123 = $122&255;
      $124 = Math_imul($i$01$i, $$$i)|0;
      $125 = (($58) + ($124)|0);
      HEAP8[$125>>0] = $123;
      break;
     }
     case 2:  {
      $126 = $76 & 255;
      $127 = HEAP32[$66>>2]|0;
      $128 = $127 & 255;
      $129 = $126 >>> $128;
      $130 = HEAP32[$67>>2]|0;
      $131 = $130 & 255;
      $132 = $129 << $131;
      $133 = $81 & 255;
      $134 = $127 >>> 8;
      $135 = $134 & 255;
      $136 = $133 >>> $135;
      $137 = $130 >>> 8;
      $138 = $137 & 255;
      $139 = $136 << $138;
      $140 = $86 & 255;
      $141 = $127 >>> 16;
      $142 = $141 & 255;
      $143 = $140 >>> $142;
      $144 = $130 >>> 16;
      $145 = $144 & 255;
      $146 = $143 << $145;
      $147 = $91 & 255;
      $148 = $127 >>> 24;
      $149 = $147 >>> $148;
      $150 = $130 >>> 24;
      $151 = $149 << $150;
      $152 = $151 | $132;
      $153 = $152 | $139;
      $154 = $153 | $146;
      $155 = $154&65535;
      $156 = Math_imul($i$01$i, $$$i)|0;
      $157 = (($58) + ($156)|0);
      HEAP16[$157>>1] = $155;
      break;
     }
     case 3:  {
      $158 = $86&255;
      $159 = $81&255;
      $160 = $76&255;
      $161 = Math_imul($i$01$i, $$$i)|0;
      $162 = (($58) + ($161)|0);
      $163 = HEAP8[$67>>0]|0;
      $div$i = ($163&255) >>> 3;
      $164 = $div$i&255;
      $165 = (($162) + ($164)|0);
      HEAP8[$165>>0] = $160;
      $166 = HEAP8[$68>>0]|0;
      $div1$i = ($166&255) >>> 3;
      $167 = $div1$i&255;
      $168 = (($162) + ($167)|0);
      HEAP8[$168>>0] = $159;
      $169 = HEAP8[$69>>0]|0;
      $div2$i = ($169&255) >>> 3;
      $170 = $div2$i&255;
      $171 = (($162) + ($170)|0);
      HEAP8[$171>>0] = $158;
      break;
     }
     case 4:  {
      $172 = $76 & 255;
      $173 = HEAP32[$66>>2]|0;
      $174 = $173 & 255;
      $175 = $172 >>> $174;
      $176 = HEAP32[$67>>2]|0;
      $177 = $176 & 255;
      $178 = $175 << $177;
      $179 = $81 & 255;
      $180 = $173 >>> 8;
      $181 = $180 & 255;
      $182 = $179 >>> $181;
      $183 = $176 >>> 8;
      $184 = $183 & 255;
      $185 = $182 << $184;
      $186 = $86 & 255;
      $187 = $173 >>> 16;
      $188 = $187 & 255;
      $189 = $186 >>> $188;
      $190 = $176 >>> 16;
      $191 = $190 & 255;
      $192 = $189 << $191;
      $193 = $91 & 255;
      $194 = $173 >>> 24;
      $195 = $193 >>> $194;
      $196 = $176 >>> 24;
      $197 = $195 << $196;
      $198 = $197 | $178;
      $199 = $198 | $185;
      $200 = $199 | $192;
      $201 = Math_imul($i$01$i, $$$i)|0;
      $202 = (($58) + ($201)|0);
      HEAP32[$202>>2] = $200;
      break;
     }
     default: {
     }
     }
     $203 = (($i$01$i) + 1)|0;
     $204 = HEAP32[$$idx$val>>2]|0;
     $205 = ($203|0)<($204|0);
     if ($205) {
      $i$01$i = $203;
     } else {
      break;
     }
    }
    $206 = ((($1)) + 64|0);
    HEAP32[$206>>2] = $58;
    label = 36;
   } else {
    label = 27;
   }
  } else {
   label = 27;
  }
 } while(0);
 do {
  if ((label|0) == 27) {
   $207 = HEAP32[$20>>2]|0;
   $208 = ($207|0)==(0);
   $$mask = $207 & -268435456;
   $209 = ($$mask|0)==(268435456);
   $or$cond8 = $208 | $209;
   if ($or$cond8) {
    $210 = $207 >>> 24;
    $211 = $210 & 15;
    $$off12 = (($211) + -1)|0;
    $switch13 = ($$off12>>>0)<(3);
    if ($switch13) {
     $$idx14 = ((($20)) + 4|0);
     $$idx14$val = HEAP32[$$idx14>>2]|0;
     HEAP32[$dithered$i>>2] = 256;
     $i$01$i$i = 0;
     while(1) {
      $212 = $i$01$i$i & 224;
      $213 = $212 >>> 3;
      $214 = $212 >>> 6;
      $215 = $214 | $212;
      $216 = $215 | $213;
      $217 = $216&255;
      $218 = (($colors$i) + ($i$01$i$i<<2)|0);
      HEAP8[$218>>0] = $217;
      $219 = $i$01$i$i << 3;
      $220 = $219 & 224;
      $221 = $220 >>> 3;
      $222 = $220 >>> 6;
      $223 = $222 | $220;
      $224 = $223 | $221;
      $225 = $224&255;
      $226 = (((($colors$i) + ($i$01$i$i<<2)|0)) + 1|0);
      HEAP8[$226>>0] = $225;
      $227 = $i$01$i$i & 3;
      $228 = $227 << 2;
      $229 = $228 | $227;
      $230 = $229 << 4;
      $231 = $230 | $229;
      $232 = $231&255;
      $233 = (((($colors$i) + ($i$01$i$i<<2)|0)) + 2|0);
      HEAP8[$233>>0] = $232;
      $234 = (((($colors$i) + ($i$01$i$i<<2)|0)) + 3|0);
      HEAP8[$234>>0] = -1;
      $235 = (($i$01$i$i) + 1)|0;
      $exitcond$i$i = ($235|0)==(256);
      if ($exitcond$i$i) {
       break;
      } else {
       $i$01$i$i = $235;
      }
     }
     $236 = ((($dithered$i)) + 4|0);
     HEAP32[$236>>2] = $colors$i;
     $237 = (_Map1to1($dithered$i,$$idx14$val,$16)|0);
     $238 = ((($1)) + 64|0);
     HEAP32[$238>>2] = $237;
     $239 = HEAP32[$16>>2]|0;
     $240 = ($239|0)==(0);
     $241 = ($237|0)==(0|0);
     $or$cond9 = $241 & $240;
     if ($or$cond9) {
      $$0 = -1;
      STACKTOP = sp;return ($$0|0);
     } else {
      HEAP32[$16>>2] = 0;
      label = 36;
      break;
     }
    }
   }
   $242 = ($18|0)==($20|0);
   if ($242) {
    HEAP32[$16>>2] = 1;
    label = 35;
   } else {
    label = 35;
   }
  }
 } while(0);
 if ((label|0) == 35) {
  HEAP32[$1>>2] = $dst;
  label = 37;
 }
 else if ((label|0) == 36) {
  HEAP32[$1>>2] = $dst;
  $243 = ($dst|0)==(0|0);
  if (!($243)) {
   label = 37;
  }
 }
 if ((label|0) == 37) {
  $244 = ((($dst)) + 56|0);
  $245 = HEAP32[$244>>2]|0;
  $246 = (($245) + 1)|0;
  HEAP32[$244>>2] = $246;
 }
 $247 = ((($20)) + 4|0);
 $248 = HEAP32[$247>>2]|0;
 $249 = ($248|0)==(0|0);
 if ($249) {
  $$sink = 0;
 } else {
  $250 = ((($248)) + 8|0);
  $251 = HEAP32[$250>>2]|0;
  $$sink = $251;
 }
 $252 = ((($1)) + 80|0);
 HEAP32[$252>>2] = $$sink;
 $253 = ((($18)) + 4|0);
 $254 = HEAP32[$253>>2]|0;
 $255 = ($254|0)==(0|0);
 if ($255) {
  $$sink16 = 0;
 } else {
  $256 = ((($254)) + 8|0);
  $257 = HEAP32[$256>>2]|0;
  $$sink16 = $257;
 }
 $258 = ((($1)) + 84|0);
 HEAP32[$258>>2] = $$sink16;
 $259 = (_SDL_CalculateBlit($src)|0);
 $$0 = $259;
 STACKTOP = sp;return ($$0|0);
}
function _Map1to1($src,$dst,$identical) {
 $src = $src|0;
 $dst = $dst|0;
 $identical = $identical|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $i$02$i = 0, $pixel$03$i = 0, $pixel$1$i = 0, $pixel$2$i = 0, $smallest$01$i = 0, $smallest$1$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($identical|0)==(0|0);
 if (!($0)) {
  $1 = HEAP32[$src>>2]|0;
  $2 = HEAP32[$dst>>2]|0;
  $3 = ($1|0)>($2|0);
  do {
   if (!($3)) {
    $4 = ($src|0)==($dst|0);
    if (!($4)) {
     $5 = ((($src)) + 4|0);
     $6 = HEAP32[$5>>2]|0;
     $7 = ((($dst)) + 4|0);
     $8 = HEAP32[$7>>2]|0;
     $9 = $1 << 2;
     $10 = (_SDL_memcmp($6,$8,$9)|0);
     $11 = ($10|0)==(0);
     if (!($11)) {
      break;
     }
    }
    HEAP32[$identical>>2] = 1;
    $$0 = 0;
    return ($$0|0);
   }
  } while(0);
  HEAP32[$identical>>2] = 0;
 }
 $12 = HEAP32[$src>>2]|0;
 $13 = (_SDL_malloc($12)|0);
 $14 = ($13|0)==(0|0);
 if ($14) {
  (_SDL_Error(0)|0);
  $$0 = 0;
  return ($$0|0);
 }
 $15 = HEAP32[$src>>2]|0;
 $16 = ($15|0)>(0);
 if (!($16)) {
  $$0 = $13;
  return ($$0|0);
 }
 $17 = ((($src)) + 4|0);
 $18 = ((($dst)) + 4|0);
 $i$01 = 0;
 while(1) {
  $19 = HEAP32[$17>>2]|0;
  $20 = HEAP32[$dst>>2]|0;
  $21 = ($20|0)>(0);
  L20: do {
   if ($21) {
    $22 = (((($19) + ($i$01<<2)|0)) + 3|0);
    $23 = HEAP8[$22>>0]|0;
    $24 = (((($19) + ($i$01<<2)|0)) + 2|0);
    $25 = HEAP8[$24>>0]|0;
    $26 = (((($19) + ($i$01<<2)|0)) + 1|0);
    $27 = HEAP8[$26>>0]|0;
    $28 = (($19) + ($i$01<<2)|0);
    $29 = HEAP8[$28>>0]|0;
    $30 = HEAP32[$18>>2]|0;
    $31 = $29&255;
    $32 = $27&255;
    $33 = $25&255;
    $34 = $23&255;
    $i$02$i = 0;$pixel$03$i = 0;$smallest$01$i = -1;
    while(1) {
     $35 = (($30) + ($i$02$i<<2)|0);
     $36 = HEAP8[$35>>0]|0;
     $37 = $36&255;
     $38 = (($37) - ($31))|0;
     $39 = (((($30) + ($i$02$i<<2)|0)) + 1|0);
     $40 = HEAP8[$39>>0]|0;
     $41 = $40&255;
     $42 = (($41) - ($32))|0;
     $43 = (((($30) + ($i$02$i<<2)|0)) + 2|0);
     $44 = HEAP8[$43>>0]|0;
     $45 = $44&255;
     $46 = (($45) - ($33))|0;
     $47 = (((($30) + ($i$02$i<<2)|0)) + 3|0);
     $48 = HEAP8[$47>>0]|0;
     $49 = $48&255;
     $50 = (($49) - ($34))|0;
     $51 = Math_imul($38, $38)|0;
     $52 = Math_imul($42, $42)|0;
     $53 = (($52) + ($51))|0;
     $54 = Math_imul($46, $46)|0;
     $55 = (($53) + ($54))|0;
     $56 = Math_imul($50, $50)|0;
     $57 = (($55) + ($56))|0;
     $58 = ($57>>>0)<($smallest$01$i>>>0);
     if ($58) {
      $59 = $i$02$i&255;
      $60 = ($57|0)==(0);
      if ($60) {
       $pixel$2$i = $59;
       break L20;
      } else {
       $pixel$1$i = $59;$smallest$1$i = $57;
      }
     } else {
      $pixel$1$i = $pixel$03$i;$smallest$1$i = $smallest$01$i;
     }
     $61 = (($i$02$i) + 1)|0;
     $62 = ($61|0)<($20|0);
     if ($62) {
      $i$02$i = $61;$pixel$03$i = $pixel$1$i;$smallest$01$i = $smallest$1$i;
     } else {
      $pixel$2$i = $pixel$1$i;
      break;
     }
    }
   } else {
    $pixel$2$i = 0;
   }
  } while(0);
  $63 = (($13) + ($i$01)|0);
  HEAP8[$63>>0] = $pixel$2$i;
  $64 = (($i$01) + 1)|0;
  $65 = HEAP32[$src>>2]|0;
  $66 = ($64|0)<($65|0);
  if ($66) {
   $i$01 = $64;
  } else {
   $$0 = $13;
   break;
  }
 }
 return ($$0|0);
}
function _SDL_CalculateBlit($surface) {
 $surface = $surface|0;
 var $$0 = 0, $$mask = 0, $$mask1 = 0, $$mask3 = 0, $$mask4 = 0, $$mask5 = 0, $$not = 0, $$not14 = 0, $$off = 0, $$off20 = 0, $$off22 = 0, $$off24 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $blit$0 = 0, $blit$3 = 0, $i$01$i = 0, $or$cond = 0, $or$cond10 = 0, $or$cond16 = 0, $or$cond8 = 0, $switch = 0;
 var $switch21 = 0, $switch23 = 0, $switch25 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($surface)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = HEAP32[$surface>>2]|0;
 $4 = $3 & 2;
 $5 = ($4|0)==(0);
 if (!($5)) {
  _SDL_UnRLESurface($surface,1);
 }
 $6 = ((($1)) + 8|0);
 HEAP32[$6>>2] = 4489;
 $7 = ((($surface)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($1)) + 56|0);
 HEAP32[$9>>2] = $8;
 $10 = ((($surface)) + 16|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($1)) + 28|0);
 HEAP32[$12>>2] = $11;
 $13 = ((($2)) + 4|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($1)) + 60|0);
 HEAP32[$15>>2] = $14;
 $16 = ((($2)) + 16|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ((($1)) + 48|0);
 HEAP32[$18>>2] = $17;
 $19 = ((($1)) + 68|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = $20 & 4096;
 $22 = ($21|0)==(0);
 if (!($22)) {
  $23 = (_SDL_RLESurface($surface)|0);
  $24 = ($23|0)==(0);
  if ($24) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $25 = ((($1)) + 4|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ($26|0)==(0);
 if ($27) {
  label = 7;
 } else {
  $28 = HEAP32[$19>>2]|0;
  $29 = $28 & -4097;
  $30 = ($29|0)==(0);
  if ($30) {
   $blit$3 = 4490;
  } else {
   label = 7;
  }
 }
 L10: do {
  if ((label|0) == 7) {
   $31 = HEAP32[$7>>2]|0;
   $32 = ((($31)) + 8|0);
   $33 = HEAP16[$32>>1]|0;
   $34 = $33&255;
   $35 = ($34&255)<(8);
   if ($35) {
    $36 = HEAP32[$31>>2]|0;
    $37 = ($36|0)==(0);
    $$mask5 = $36 & -268435456;
    $38 = ($$mask5|0)==(268435456);
    $or$cond = $37 | $38;
    if ($or$cond) {
     $39 = $36 >>> 24;
     $40 = $39 & 15;
     $$off = (($40) + -1)|0;
     $switch = ($$off>>>0)<(3);
     if ($switch) {
      $41 = (_SDL_CalculateBlit0($surface)|0);
      $blit$0 = $41;
     } else {
      label = 11;
     }
    } else {
     label = 11;
    }
   } else {
    label = 11;
   }
   do {
    if ((label|0) == 11) {
     $$mask = $33 & -256;
     $42 = ($$mask<<16>>16)==(256);
     if ($42) {
      $43 = HEAP32[$31>>2]|0;
      $44 = ($43|0)==(0);
      $$mask4 = $43 & -268435456;
      $45 = ($$mask4|0)==(268435456);
      $or$cond8 = $44 | $45;
      if ($or$cond8) {
       $46 = $43 >>> 24;
       $47 = $46 & 15;
       $$off20 = (($47) + -1)|0;
       $switch21 = ($$off20>>>0)<(3);
       if ($switch21) {
        $48 = (_SDL_CalculateBlit1($surface)|0);
        $blit$0 = $48;
        break;
       }
      }
     }
     $49 = HEAP32[$19>>2]|0;
     $50 = $49 & 16;
     $51 = ($50|0)==(0);
     if ($51) {
      $53 = (_SDL_CalculateBlitN($surface)|0);
      $blit$0 = $53;
      break;
     } else {
      $52 = (_SDL_CalculateBlitA($surface)|0);
      $blit$0 = $52;
      break;
     }
    }
   } while(0);
   $54 = ($blit$0|0)==(0|0);
   if ($54) {
    $55 = HEAP32[$7>>2]|0;
    $56 = HEAP32[$55>>2]|0;
    $57 = HEAP32[$13>>2]|0;
    $58 = HEAP32[$57>>2]|0;
    $59 = HEAP32[$19>>2]|0;
    $60 = HEAP32[33823]|0;
    $61 = ($60|0)==(-1);
    do {
     if ($61) {
      $62 = (_SDL_getenv(603205)|0);
      HEAP32[33823] = 0;
      $63 = ($62|0)==(0|0);
      if (!($63)) {
       HEAP32[$vararg_buffer>>2] = 135292;
       (_SDL_sscanf($62,603227,$vararg_buffer)|0);
       break;
      }
      $64 = (_SDL_HasMMX()|0);
      $65 = ($64|0)==(0);
      if (!($65)) {
       $66 = HEAP32[33823]|0;
       $67 = $66 | 1;
       HEAP32[33823] = $67;
      }
      $68 = (_SDL_Has3DNow()|0);
      $69 = ($68|0)==(0);
      if (!($69)) {
       $70 = HEAP32[33823]|0;
       $71 = $70 | 2;
       HEAP32[33823] = $71;
      }
      $72 = (_SDL_HasSSE()|0);
      $73 = ($72|0)==(0);
      if (!($73)) {
       $74 = HEAP32[33823]|0;
       $75 = $74 | 4;
       HEAP32[33823] = $75;
      }
      $76 = (_SDL_HasSSE2()|0);
      $77 = ($76|0)==(0);
      if (!($77)) {
       $78 = HEAP32[33823]|0;
       $79 = $78 | 8;
       HEAP32[33823] = $79;
      }
      $80 = (_SDL_HasAltiVec()|0);
      $81 = ($80|0)==(0);
      if (!($81)) {
       $82 = HEAP32[33823]|0;
       $83 = $82 | 16;
       HEAP32[33823] = $83;
      }
     }
    } while(0);
    $84 = HEAP32[(135312)>>2]|0;
    $85 = ($84|0)==(0|0);
    if (!($85)) {
     $86 = HEAP32[33823]|0;
     $87 = $59 & 883;
     $119 = $84;$i$01$i = 0;
     while(1) {
      $88 = (135296 + (($i$01$i*20)|0)|0);
      $89 = HEAP32[$88>>2]|0;
      $90 = ($89|0)==($56|0);
      if ($90) {
       $91 = (((135296 + (($i$01$i*20)|0)|0)) + 4|0);
       $92 = HEAP32[$91>>2]|0;
       $93 = ($92|0)==($58|0);
       if ($93) {
        $94 = (((135296 + (($i$01$i*20)|0)|0)) + 8|0);
        $95 = HEAP32[$94>>2]|0;
        $96 = $95 & $87;
        $97 = ($96|0)==($87|0);
        if ($97) {
         $98 = (((135296 + (($i$01$i*20)|0)|0)) + 12|0);
         $99 = HEAP32[$98>>2]|0;
         $100 = $99 & $86;
         $101 = ($100|0)==($99|0);
         if ($101) {
          $blit$3 = $119;
          break L10;
         }
        }
       }
      }
      $102 = (($i$01$i) + 1)|0;
      $103 = (((135296 + (($102*20)|0)|0)) + 16|0);
      $104 = HEAP32[$103>>2]|0;
      $105 = ($104|0)==(0|0);
      if ($105) {
       break;
      } else {
       $119 = $104;$i$01$i = $102;
      }
     }
    }
    $106 = HEAP32[$7>>2]|0;
    $107 = HEAP32[$106>>2]|0;
    $108 = HEAP32[$13>>2]|0;
    $109 = HEAP32[$108>>2]|0;
    $$not = ($107|0)==(0);
    $$mask3 = $107 & -268435456;
    $110 = ($$mask3|0)==(268435456);
    $or$cond10 = $$not | $110;
    if ($or$cond10) {
     $111 = $107 >>> 24;
     $112 = $111 & 15;
     $$off22 = (($112) + -1)|0;
     $switch23 = ($$off22>>>0)<(3);
     if (!($switch23)) {
      $$not14 = ($109|0)==(0);
      $$mask1 = $109 & -268435456;
      $113 = ($$mask1|0)==(268435456);
      $or$cond16 = $$not14 | $113;
      if ($or$cond16) {
       $114 = $109 >>> 24;
       $115 = $114 & 15;
       $$off24 = (($115) + -1)|0;
       $switch25 = ($$off24>>>0)<(3);
       if (!($switch25)) {
        $blit$3 = 4491;
        break;
       }
      }
     }
    }
    $117 = ((($1)) + 12|0);
    HEAP32[$117>>2] = 0;
    _SDL_InvalidateMap($1);
    $118 = (_SDL_SetError(603230,$vararg_buffer1)|0);
    $$0 = $118;
    STACKTOP = sp;return ($$0|0);
   } else {
    $blit$3 = $blit$0;
   }
  }
 } while(0);
 $116 = ((($1)) + 12|0);
 HEAP32[$116>>2] = $blit$3;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SoftBlit($src,$srcrect,$dst,$dstrect) {
 $src = $src|0;
 $srcrect = $srcrect|0;
 $dst = $dst|0;
 $dstrect = $dstrect|0;
 var $$lobit = 0, $$lobit$not = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $8 = 0, $9 = 0;
 var $dst_locked$0 = 0, $okay$0 = 0, $src_locked$0 = 0, $src_locked$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$dst>>2]|0;
 $1 = $0 & 2;
 $2 = ($1|0)==(0);
 if ($2) {
  $dst_locked$0 = 0;$okay$0 = 1;
 } else {
  $3 = (_SDL_LockSurface($dst)|0);
  $$lobit = $3 >>> 31;
  $$lobit$not = $$lobit ^ 1;
  $dst_locked$0 = $$lobit$not;$okay$0 = $$lobit$not;
 }
 $4 = HEAP32[$src>>2]|0;
 $5 = $4 & 2;
 $6 = ($5|0)==(0);
 if ($6) {
  $src_locked$0 = 0;
  label = 5;
 } else {
  $7 = (_SDL_LockSurface($src)|0);
  $8 = ($7|0)<(0);
  if ($8) {
   $77 = -1;$src_locked$02 = 0;
  } else {
   $src_locked$0 = 1;
   label = 5;
  }
 }
 if ((label|0) == 5) {
  $9 = ($okay$0|0)==(0);
  if ($9) {
   $77 = -1;$src_locked$02 = $src_locked$0;
  } else {
   $10 = ($srcrect|0)==(0|0);
   if ($10) {
    $77 = 0;$src_locked$02 = $src_locked$0;
   } else {
    $11 = ((($srcrect)) + 8|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = ($12|0)<(1);
    if ($13) {
     $77 = 0;$src_locked$02 = $src_locked$0;
    } else {
     $14 = ((($srcrect)) + 12|0);
     $15 = HEAP32[$14>>2]|0;
     $16 = ($15|0)<(1);
     if ($16) {
      $77 = 0;$src_locked$02 = $src_locked$0;
     } else {
      $17 = ((($src)) + 52|0);
      $18 = HEAP32[$17>>2]|0;
      $19 = ((($18)) + 16|0);
      $20 = ((($src)) + 20|0);
      $21 = HEAP32[$20>>2]|0;
      $22 = ((($srcrect)) + 4|0);
      $23 = HEAP32[$22>>2]|0;
      $24 = $23 & 65535;
      $25 = ((($src)) + 16|0);
      $26 = HEAP32[$25>>2]|0;
      $27 = Math_imul($24, $26)|0;
      $28 = (($21) + ($27)|0);
      $29 = HEAP32[$srcrect>>2]|0;
      $30 = $29 & 65535;
      $31 = ((($18)) + 56|0);
      $32 = HEAP32[$31>>2]|0;
      $33 = ((($32)) + 9|0);
      $34 = HEAP8[$33>>0]|0;
      $35 = $34&255;
      $36 = Math_imul($35, $30)|0;
      $37 = (($28) + ($36)|0);
      HEAP32[$19>>2] = $37;
      $38 = ((($18)) + 20|0);
      HEAP32[$38>>2] = $12;
      $39 = ((($18)) + 24|0);
      HEAP32[$39>>2] = $15;
      $40 = ((($18)) + 28|0);
      HEAP32[$40>>2] = $26;
      $41 = Math_imul($35, $12)|0;
      $42 = (($26) - ($41))|0;
      $43 = ((($18)) + 32|0);
      HEAP32[$43>>2] = $42;
      $44 = ((($dst)) + 20|0);
      $45 = HEAP32[$44>>2]|0;
      $46 = ((($dstrect)) + 4|0);
      $47 = HEAP32[$46>>2]|0;
      $48 = $47 & 65535;
      $49 = ((($dst)) + 16|0);
      $50 = HEAP32[$49>>2]|0;
      $51 = Math_imul($48, $50)|0;
      $52 = (($45) + ($51)|0);
      $53 = HEAP32[$dstrect>>2]|0;
      $54 = $53 & 65535;
      $55 = ((($18)) + 60|0);
      $56 = HEAP32[$55>>2]|0;
      $57 = ((($56)) + 9|0);
      $58 = HEAP8[$57>>0]|0;
      $59 = $58&255;
      $60 = Math_imul($59, $54)|0;
      $61 = (($52) + ($60)|0);
      $62 = ((($18)) + 36|0);
      HEAP32[$62>>2] = $61;
      $63 = ((($dstrect)) + 8|0);
      $64 = HEAP32[$63>>2]|0;
      $65 = ((($18)) + 40|0);
      HEAP32[$65>>2] = $64;
      $66 = ((($dstrect)) + 12|0);
      $67 = HEAP32[$66>>2]|0;
      $68 = ((($18)) + 44|0);
      HEAP32[$68>>2] = $67;
      $69 = ((($18)) + 48|0);
      HEAP32[$69>>2] = $50;
      $70 = Math_imul($59, $64)|0;
      $71 = (($50) - ($70))|0;
      $72 = ((($18)) + 52|0);
      HEAP32[$72>>2] = $71;
      $73 = ((($18)) + 12|0);
      $74 = HEAP32[$73>>2]|0;
      FUNCTION_TABLE_vi[$74 & 8191]($19);
      $77 = 0;$src_locked$02 = $src_locked$0;
     }
    }
   }
  }
 }
 $75 = ($dst_locked$0|0)==(0);
 if (!($75)) {
  _SDL_UnlockSurface($dst);
 }
 $76 = ($src_locked$02|0)==(0);
 if ($76) {
  return ($77|0);
 }
 _SDL_UnlockSurface($src);
 return ($77|0);
}
function _SDL_CalculateBlit0($surface) {
 $surface = $surface|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $which$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($surface)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = ($3<<24>>24)==(1);
 if (!($4)) {
  $$0 = 0;
  return ($$0|0);
 }
 $5 = ((($surface)) + 52|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($7)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 8|0);
 $11 = HEAP16[$10>>1]|0;
 $12 = $11&255;
 $13 = ($12&255)<(8);
 $14 = ($11&65535) >>> 8;
 $15 = $14&65535;
 $which$0 = $13 ? 0 : $15;
 $16 = ((($6)) + 68|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 & -28673;
 switch ($18|0) {
 case 0:  {
  $19 = (126600 + ($which$0<<2)|0);
  $20 = HEAP32[$19>>2]|0;
  $$0 = $20;
  return ($$0|0);
  break;
 }
 case 256:  {
  $21 = (126620 + ($which$0<<2)|0);
  $22 = HEAP32[$21>>2]|0;
  $$0 = $22;
  return ($$0|0);
  break;
 }
 case 18:  {
  $23 = ($which$0>>>0)>(1);
  $24 = $23 ? 4492 : 0;
  $$0 = $24;
  return ($$0|0);
  break;
 }
 case 274:  {
  $25 = ($which$0>>>0)>(1);
  $26 = $25 ? 4493 : 0;
  $$0 = $26;
  return ($$0|0);
  break;
 }
 default: {
  $$0 = 0;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function _BlitBto1($info) {
 $info = $info|0;
 var $$not29 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $byte$014$us = 0, $byte$1$us = 0, $byte1$07$us = 0, $byte1$1$us = 0, $c$015$us = 0, $c$18$us = 0;
 var $dst$020$us = 0, $dst$112$us = 0, $dst$210$us = 0, $dst$35$us = 0, $exitcond = 0, $exitcond27 = 0, $or$cond30 = 0, $scevgep = 0, $scevgep26 = 0, $src$021$us = 0, $src$113$us = 0, $src$2$us = 0, $src$2$us$lcssa = 0, $src$311$us = 0, $src$46$us = 0, $src$5$us = 0, $src$5$us$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$info>>2]|0;
 $5 = ((($info)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($info)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($info)) + 36|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($info)) + 48|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($1) + 7)|0;
 $14 = (($13|0) / 8)&-1;
 $15 = (($1) - ($14))|0;
 $16 = (($6) + ($15))|0;
 $17 = ($12|0)==(0|0);
 $18 = (($3) + -1)|0;
 $$not29 = ($3|0)!=(0);
 $19 = ($1|0)>(0);
 $or$cond30 = $$not29 & $19;
 if ($17) {
  if ($or$cond30) {
   $51 = $18;$dst$210$us = $8;$src$311$us = $4;
  } else {
   return;
  }
  while(1) {
   $byte1$07$us = 0;$c$18$us = 0;$dst$35$us = $dst$210$us;$src$46$us = $src$311$us;
   while(1) {
    $38 = $c$18$us & 7;
    $39 = ($38|0)==(0);
    if ($39) {
     $40 = ((($src$46$us)) + 1|0);
     $41 = HEAP8[$src$46$us>>0]|0;
     $byte1$1$us = $41;$src$5$us = $40;
    } else {
     $byte1$1$us = $byte1$07$us;$src$5$us = $src$46$us;
    }
    $42 = $byte1$1$us&255;
    $43 = ($byte1$1$us&255) >>> 7;
    HEAP8[$dst$35$us>>0] = $43;
    $44 = ((($dst$35$us)) + 1|0);
    $45 = $42 << 1;
    $46 = $45&255;
    $47 = (($c$18$us) + 1)|0;
    $exitcond = ($47|0)==($1|0);
    if ($exitcond) {
     $src$5$us$lcssa = $src$5$us;
     break;
    } else {
     $byte1$07$us = $46;$c$18$us = $47;$dst$35$us = $44;$src$46$us = $src$5$us;
    }
   }
   $scevgep = (($dst$210$us) + ($1)|0);
   $48 = (($src$5$us$lcssa) + ($16)|0);
   $49 = (($scevgep) + ($10)|0);
   $50 = (($51) + -1)|0;
   $52 = ($51|0)==(0);
   if ($52) {
    break;
   } else {
    $51 = $50;$dst$210$us = $49;$src$311$us = $48;
   }
  }
  return;
 } else {
  if ($or$cond30) {
   $36 = $18;$dst$020$us = $8;$src$021$us = $4;
  } else {
   return;
  }
  while(1) {
   $byte$014$us = 0;$c$015$us = 0;$dst$112$us = $dst$020$us;$src$113$us = $src$021$us;
   while(1) {
    $20 = $c$015$us & 7;
    $21 = ($20|0)==(0);
    if ($21) {
     $22 = ((($src$113$us)) + 1|0);
     $23 = HEAP8[$src$113$us>>0]|0;
     $byte$1$us = $23;$src$2$us = $22;
    } else {
     $byte$1$us = $byte$014$us;$src$2$us = $src$113$us;
    }
    $24 = $byte$1$us&255;
    $25 = ($byte$1$us&255) >>> 7;
    $26 = $25&255;
    $27 = (($12) + ($26)|0);
    $28 = HEAP8[$27>>0]|0;
    HEAP8[$dst$112$us>>0] = $28;
    $29 = ((($dst$112$us)) + 1|0);
    $30 = $24 << 1;
    $31 = $30&255;
    $32 = (($c$015$us) + 1)|0;
    $exitcond27 = ($32|0)==($1|0);
    if ($exitcond27) {
     $src$2$us$lcssa = $src$2$us;
     break;
    } else {
     $byte$014$us = $31;$c$015$us = $32;$dst$112$us = $29;$src$113$us = $src$2$us;
    }
   }
   $scevgep26 = (($dst$020$us) + ($1)|0);
   $33 = (($src$2$us$lcssa) + ($16)|0);
   $34 = (($scevgep26) + ($10)|0);
   $35 = (($36) + -1)|0;
   $37 = ($36|0)==(0);
   if ($37) {
    break;
   } else {
    $36 = $35;$dst$020$us = $34;$src$021$us = $33;
   }
  }
  return;
 }
}
function _BlitBto2($info) {
 $info = $info|0;
 var $$in = 0, $$neg = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $byte$03$us = 0, $byte$1$us = 0;
 var $c$04$us = 0, $dst$07$us = 0, $dst$12$us = 0, $exitcond = 0, $or$cond = 0, $scevgep = 0, $src$06$us = 0, $src$11$us = 0, $src$2$us = 0, $src$2$us$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (($7|0) / 2)&-1;
 $9 = ((($info)) + 48|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($1) + 7)|0;
 $$neg = (($11|0) / -8)&-1;
 $12 = (($$neg) + ($1))|0;
 $13 = (($12) + ($5))|0;
 $14 = ($3|0)!=(0);
 $15 = ($1|0)>(0);
 $or$cond = $14 & $15;
 if (!($or$cond)) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = HEAP32[$info>>2]|0;
 $$in = $3;$dst$07$us = $17;$src$06$us = $18;
 while(1) {
  $32 = (($$in) + -1)|0;
  $byte$03$us = 0;$c$04$us = 0;$dst$12$us = $dst$07$us;$src$11$us = $src$06$us;
  while(1) {
   $19 = $c$04$us & 7;
   $20 = ($19|0)==(0);
   if ($20) {
    $21 = ((($src$11$us)) + 1|0);
    $22 = HEAP8[$src$11$us>>0]|0;
    $byte$1$us = $22;$src$2$us = $21;
   } else {
    $byte$1$us = $byte$03$us;$src$2$us = $src$11$us;
   }
   $23 = $byte$1$us&255;
   $24 = ($byte$1$us&255) >>> 7;
   $25 = $24&255;
   $26 = (($10) + ($25<<1)|0);
   $27 = HEAP16[$26>>1]|0;
   HEAP16[$dst$12$us>>1] = $27;
   $28 = $23 << 1;
   $29 = $28&255;
   $30 = ((($dst$12$us)) + 2|0);
   $31 = (($c$04$us) + 1)|0;
   $exitcond = ($31|0)==($1|0);
   if ($exitcond) {
    $src$2$us$lcssa = $src$2$us;
    break;
   } else {
    $byte$03$us = $29;$c$04$us = $31;$dst$12$us = $30;$src$11$us = $src$2$us;
   }
  }
  $scevgep = (($dst$07$us) + ($1<<1)|0);
  $33 = (($src$2$us$lcssa) + ($13)|0);
  $34 = (($scevgep) + ($8<<1)|0);
  $35 = ($32|0)==(0);
  if ($35) {
   break;
  } else {
   $$in = $32;$dst$07$us = $34;$src$06$us = $33;
  }
 }
 return;
}
function _BlitBto3($info) {
 $info = $info|0;
 var $$in = 0, $$neg = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $byte$03$us = 0, $byte$1$us = 0, $c$04$us = 0, $dst$07$us = 0, $dst$12$us = 0, $exitcond = 0, $or$cond = 0, $scevgep = 0, $src$06$us = 0, $src$11$us = 0, $src$2$us = 0, $src$2$us$lcssa = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 48|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (($1) + 7)|0;
 $$neg = (($10|0) / -8)&-1;
 $11 = (($$neg) + ($1))|0;
 $12 = (($11) + ($5))|0;
 $13 = ($3|0)!=(0);
 $14 = ($1|0)>(0);
 $or$cond = $13 & $14;
 if (!($or$cond)) {
  return;
 }
 $15 = ((($info)) + 20|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = HEAP32[$info>>2]|0;
 $18 = ($1*3)|0;
 $$in = $3;$dst$07$us = $16;$src$06$us = $17;
 while(1) {
  $41 = (($$in) + -1)|0;
  $byte$03$us = 0;$c$04$us = 0;$dst$12$us = $dst$07$us;$src$11$us = $src$06$us;
  while(1) {
   $19 = $c$04$us & 7;
   $20 = ($19|0)==(0);
   if ($20) {
    $21 = ((($src$11$us)) + 1|0);
    $22 = HEAP8[$src$11$us>>0]|0;
    $byte$1$us = $22;$src$2$us = $21;
   } else {
    $byte$1$us = $byte$03$us;$src$2$us = $src$11$us;
   }
   $23 = $byte$1$us&255;
   $24 = ($byte$1$us&255) >>> 7;
   $25 = $24&255;
   $26 = $25 << 2;
   $27 = $26 | 1;
   $28 = (($9) + ($26)|0);
   $29 = HEAP8[$28>>0]|0;
   HEAP8[$dst$12$us>>0] = $29;
   $30 = (($27) + 1)|0;
   $31 = (($9) + ($27)|0);
   $32 = HEAP8[$31>>0]|0;
   $33 = ((($dst$12$us)) + 1|0);
   HEAP8[$33>>0] = $32;
   $34 = (($9) + ($30)|0);
   $35 = HEAP8[$34>>0]|0;
   $36 = ((($dst$12$us)) + 2|0);
   HEAP8[$36>>0] = $35;
   $37 = $23 << 1;
   $38 = $37&255;
   $39 = ((($dst$12$us)) + 3|0);
   $40 = (($c$04$us) + 1)|0;
   $exitcond = ($40|0)==($1|0);
   if ($exitcond) {
    $src$2$us$lcssa = $src$2$us;
    break;
   } else {
    $byte$03$us = $38;$c$04$us = $40;$dst$12$us = $39;$src$11$us = $src$2$us;
   }
  }
  $scevgep = (($dst$07$us) + ($18)|0);
  $42 = (($src$2$us$lcssa) + ($12)|0);
  $43 = (($scevgep) + ($7)|0);
  $44 = ($41|0)==(0);
  if ($44) {
   break;
  } else {
   $$in = $41;$dst$07$us = $43;$src$06$us = $42;
  }
 }
 return;
}
function _BlitBto4($info) {
 $info = $info|0;
 var $$in = 0, $$neg = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $byte$04$us = 0, $byte$1$us = 0;
 var $c$03$us = 0, $dst$06$us = 0, $dst$11$us = 0, $exitcond = 0, $or$cond = 0, $scevgep = 0, $src$07$us = 0, $src$12$us = 0, $src$2$us = 0, $src$2$us$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (($7|0) / 4)&-1;
 $9 = ((($info)) + 48|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($1) + 7)|0;
 $$neg = (($11|0) / -8)&-1;
 $12 = (($$neg) + ($1))|0;
 $13 = (($12) + ($5))|0;
 $14 = ($3|0)!=(0);
 $15 = ($1|0)>(0);
 $or$cond = $14 & $15;
 if (!($or$cond)) {
  return;
 }
 $16 = HEAP32[$info>>2]|0;
 $17 = ((($info)) + 20|0);
 $18 = HEAP32[$17>>2]|0;
 $$in = $3;$dst$06$us = $18;$src$07$us = $16;
 while(1) {
  $32 = (($$in) + -1)|0;
  $byte$04$us = 0;$c$03$us = 0;$dst$11$us = $dst$06$us;$src$12$us = $src$07$us;
  while(1) {
   $19 = $c$03$us & 7;
   $20 = ($19|0)==(0);
   if ($20) {
    $21 = ((($src$12$us)) + 1|0);
    $22 = HEAP8[$src$12$us>>0]|0;
    $byte$1$us = $22;$src$2$us = $21;
   } else {
    $byte$1$us = $byte$04$us;$src$2$us = $src$12$us;
   }
   $23 = $byte$1$us&255;
   $24 = ($byte$1$us&255) >>> 7;
   $25 = $24&255;
   $26 = (($10) + ($25<<2)|0);
   $27 = HEAP32[$26>>2]|0;
   HEAP32[$dst$11$us>>2] = $27;
   $28 = $23 << 1;
   $29 = $28&255;
   $30 = ((($dst$11$us)) + 4|0);
   $31 = (($c$03$us) + 1)|0;
   $exitcond = ($31|0)==($1|0);
   if ($exitcond) {
    $src$2$us$lcssa = $src$2$us;
    break;
   } else {
    $byte$04$us = $29;$c$03$us = $31;$dst$11$us = $30;$src$12$us = $src$2$us;
   }
  }
  $scevgep = (($dst$06$us) + ($1<<2)|0);
  $33 = (($src$2$us$lcssa) + ($13)|0);
  $34 = (($scevgep) + ($8<<2)|0);
  $35 = ($32|0)==(0);
  if ($35) {
   break;
  } else {
   $$in = $32;$dst$06$us = $34;$src$07$us = $33;
  }
 }
 return;
}
function _BlitBto1Key($info) {
 $info = $info|0;
 var $$not29 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $byte$014$us = 0;
 var $byte$1$us = 0, $byte1$08$us = 0, $byte1$1$us = 0, $c$013$us = 0, $c$16$us = 0, $dst$020$us = 0, $dst$112$us = 0, $dst$210$us = 0, $dst$35$us = 0, $exitcond = 0, $exitcond27 = 0, $or$cond30 = 0, $scevgep = 0, $scevgep26 = 0, $src$021$us = 0, $src$115$us = 0, $src$2$us = 0, $src$2$us$lcssa = 0, $src$311$us = 0, $src$47$us = 0;
 var $src$5$us = 0, $src$5$us$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$info>>2]|0;
 $5 = ((($info)) + 20|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($info)) + 16|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($info)) + 36|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($info)) + 56|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($info)) + 48|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($1) + 7)|0;
 $16 = (($15|0) / 8)&-1;
 $17 = (($1) - ($16))|0;
 $18 = (($8) + ($17))|0;
 $19 = ($14|0)==(0|0);
 $20 = (($3) + -1)|0;
 $$not29 = ($3|0)!=(0);
 $21 = ($1|0)>(0);
 $or$cond30 = $$not29 & $21;
 if ($19) {
  if ($or$cond30) {
   $56 = $20;$dst$210$us = $6;$src$311$us = $4;
  } else {
   return;
  }
  while(1) {
   $byte1$08$us = 0;$c$16$us = 0;$dst$35$us = $dst$210$us;$src$47$us = $src$311$us;
   while(1) {
    $41 = $c$16$us & 7;
    $42 = ($41|0)==(0);
    if ($42) {
     $43 = ((($src$47$us)) + 1|0);
     $44 = HEAP8[$src$47$us>>0]|0;
     $byte1$1$us = $44;$src$5$us = $43;
    } else {
     $byte1$1$us = $byte1$08$us;$src$5$us = $src$47$us;
    }
    $45 = $byte1$1$us&255;
    $46 = ($byte1$1$us&255) >>> 7;
    $47 = $46&255;
    $48 = ($47|0)==($12|0);
    if (!($48)) {
     HEAP8[$dst$35$us>>0] = $46;
    }
    $49 = ((($dst$35$us)) + 1|0);
    $50 = $45 << 1;
    $51 = $50&255;
    $52 = (($c$16$us) + 1)|0;
    $exitcond = ($52|0)==($1|0);
    if ($exitcond) {
     $src$5$us$lcssa = $src$5$us;
     break;
    } else {
     $byte1$08$us = $51;$c$16$us = $52;$dst$35$us = $49;$src$47$us = $src$5$us;
    }
   }
   $scevgep = (($dst$210$us) + ($1)|0);
   $53 = (($src$5$us$lcssa) + ($18)|0);
   $54 = (($scevgep) + ($10)|0);
   $55 = (($56) + -1)|0;
   $57 = ($56|0)==(0);
   if ($57) {
    break;
   } else {
    $56 = $55;$dst$210$us = $54;$src$311$us = $53;
   }
  }
  return;
 } else {
  if ($or$cond30) {
   $39 = $20;$dst$020$us = $6;$src$021$us = $4;
  } else {
   return;
  }
  while(1) {
   $byte$014$us = 0;$c$013$us = 0;$dst$112$us = $dst$020$us;$src$115$us = $src$021$us;
   while(1) {
    $22 = $c$013$us & 7;
    $23 = ($22|0)==(0);
    if ($23) {
     $24 = ((($src$115$us)) + 1|0);
     $25 = HEAP8[$src$115$us>>0]|0;
     $byte$1$us = $25;$src$2$us = $24;
    } else {
     $byte$1$us = $byte$014$us;$src$2$us = $src$115$us;
    }
    $26 = $byte$1$us&255;
    $27 = ($byte$1$us&255) >>> 7;
    $28 = $27&255;
    $29 = ($28|0)==($12|0);
    if (!($29)) {
     $30 = (($14) + ($28)|0);
     $31 = HEAP8[$30>>0]|0;
     HEAP8[$dst$112$us>>0] = $31;
    }
    $32 = ((($dst$112$us)) + 1|0);
    $33 = $26 << 1;
    $34 = $33&255;
    $35 = (($c$013$us) + 1)|0;
    $exitcond27 = ($35|0)==($1|0);
    if ($exitcond27) {
     $src$2$us$lcssa = $src$2$us;
     break;
    } else {
     $byte$014$us = $34;$c$013$us = $35;$dst$112$us = $32;$src$115$us = $src$2$us;
    }
   }
   $scevgep26 = (($dst$020$us) + ($1)|0);
   $36 = (($src$2$us$lcssa) + ($18)|0);
   $37 = (($scevgep26) + ($10)|0);
   $38 = (($39) + -1)|0;
   $40 = ($39|0)==(0);
   if ($40) {
    break;
   } else {
    $39 = $38;$dst$020$us = $37;$src$021$us = $36;
   }
  }
  return;
 }
}
function _BlitBto2Key($info) {
 $info = $info|0;
 var $$in = 0, $$neg = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $byte$04$us = 0, $byte$1$us = 0, $c$03$us = 0, $dstp$06$us = 0, $dstp$11$us = 0, $exitcond = 0, $or$cond = 0, $scevgep = 0, $src$07$us = 0, $src$12$us = 0, $src$2$us = 0, $src$2$us$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 56|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($info)) + 48|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($1) + 7)|0;
 $$neg = (($12|0) / -8)&-1;
 $13 = (($$neg) + ($1))|0;
 $14 = (($13) + ($5))|0;
 $15 = (($7|0) / 2)&-1;
 $16 = ($3|0)!=(0);
 $17 = ($1|0)>(0);
 $or$cond = $16 & $17;
 if (!($or$cond)) {
  return;
 }
 $18 = HEAP32[$info>>2]|0;
 $19 = ((($info)) + 20|0);
 $20 = HEAP32[$19>>2]|0;
 $$in = $3;$dstp$06$us = $20;$src$07$us = $18;
 while(1) {
  $35 = (($$in) + -1)|0;
  $byte$04$us = 0;$c$03$us = 0;$dstp$11$us = $dstp$06$us;$src$12$us = $src$07$us;
  while(1) {
   $21 = $c$03$us & 7;
   $22 = ($21|0)==(0);
   if ($22) {
    $23 = ((($src$12$us)) + 1|0);
    $24 = HEAP8[$src$12$us>>0]|0;
    $byte$1$us = $24;$src$2$us = $23;
   } else {
    $byte$1$us = $byte$04$us;$src$2$us = $src$12$us;
   }
   $25 = $byte$1$us&255;
   $26 = ($byte$1$us&255) >>> 7;
   $27 = $26&255;
   $28 = ($27|0)==($9|0);
   if (!($28)) {
    $29 = (($11) + ($27<<1)|0);
    $30 = HEAP16[$29>>1]|0;
    HEAP16[$dstp$11$us>>1] = $30;
   }
   $31 = $25 << 1;
   $32 = $31&255;
   $33 = ((($dstp$11$us)) + 2|0);
   $34 = (($c$03$us) + 1)|0;
   $exitcond = ($34|0)==($1|0);
   if ($exitcond) {
    $src$2$us$lcssa = $src$2$us;
    break;
   } else {
    $byte$04$us = $32;$c$03$us = $34;$dstp$11$us = $33;$src$12$us = $src$2$us;
   }
  }
  $scevgep = (($dstp$06$us) + ($1<<1)|0);
  $36 = (($src$2$us$lcssa) + ($14)|0);
  $37 = (($scevgep) + ($15<<1)|0);
  $38 = ($35|0)==(0);
  if ($38) {
   break;
  } else {
   $$in = $35;$dstp$06$us = $37;$src$07$us = $36;
  }
 }
 return;
}
function _BlitBto3Key($info) {
 $info = $info|0;
 var $$in = 0, $$neg = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $byte$04$us = 0, $byte$1$us = 0, $c$03$us = 0, $dst$06$us = 0, $dst$11$us = 0, $exitcond = 0, $or$cond = 0, $scevgep = 0, $src$07$us = 0, $src$12$us = 0, $src$2$us = 0, $src$2$us$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 56|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($info)) + 48|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($1) + 7)|0;
 $$neg = (($12|0) / -8)&-1;
 $13 = (($$neg) + ($1))|0;
 $14 = (($13) + ($5))|0;
 $15 = ($3|0)!=(0);
 $16 = ($1|0)>(0);
 $or$cond = $15 & $16;
 if (!($or$cond)) {
  return;
 }
 $17 = HEAP32[$info>>2]|0;
 $18 = ((($info)) + 20|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($1*3)|0;
 $$in = $3;$dst$06$us = $19;$src$07$us = $17;
 while(1) {
  $35 = (($$in) + -1)|0;
  $byte$04$us = 0;$c$03$us = 0;$dst$11$us = $dst$06$us;$src$12$us = $src$07$us;
  while(1) {
   $21 = $c$03$us & 7;
   $22 = ($21|0)==(0);
   if ($22) {
    $23 = ((($src$12$us)) + 1|0);
    $24 = HEAP8[$src$12$us>>0]|0;
    $byte$1$us = $24;$src$2$us = $23;
   } else {
    $byte$1$us = $byte$04$us;$src$2$us = $src$12$us;
   }
   $25 = $byte$1$us&255;
   $26 = ($byte$1$us&255) >>> 7;
   $27 = $26&255;
   $28 = ($27|0)==($9|0);
   if (!($28)) {
    $29 = $27 << 2;
    $30 = (($11) + ($29)|0);
    (_SDL_memcpy($dst$11$us,$30,3)|0);
   }
   $31 = $25 << 1;
   $32 = $31&255;
   $33 = ((($dst$11$us)) + 3|0);
   $34 = (($c$03$us) + 1)|0;
   $exitcond = ($34|0)==($1|0);
   if ($exitcond) {
    $src$2$us$lcssa = $src$2$us;
    break;
   } else {
    $byte$04$us = $32;$c$03$us = $34;$dst$11$us = $33;$src$12$us = $src$2$us;
   }
  }
  $scevgep = (($dst$06$us) + ($20)|0);
  $36 = (($src$2$us$lcssa) + ($14)|0);
  $37 = (($scevgep) + ($7)|0);
  $38 = ($35|0)==(0);
  if ($38) {
   break;
  } else {
   $$in = $35;$dst$06$us = $37;$src$07$us = $36;
  }
 }
 return;
}
function _BlitBto4Key($info) {
 $info = $info|0;
 var $$in = 0, $$neg = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $byte$04$us = 0, $byte$1$us = 0, $c$03$us = 0, $dstp$06$us = 0, $dstp$11$us = 0, $exitcond = 0, $or$cond = 0, $scevgep = 0, $src$07$us = 0, $src$12$us = 0, $src$2$us = 0, $src$2$us$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 56|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($info)) + 48|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($1) + 7)|0;
 $$neg = (($12|0) / -8)&-1;
 $13 = (($$neg) + ($1))|0;
 $14 = (($13) + ($5))|0;
 $15 = (($7|0) / 4)&-1;
 $16 = ($3|0)!=(0);
 $17 = ($1|0)>(0);
 $or$cond = $16 & $17;
 if (!($or$cond)) {
  return;
 }
 $18 = HEAP32[$info>>2]|0;
 $19 = ((($info)) + 20|0);
 $20 = HEAP32[$19>>2]|0;
 $$in = $3;$dstp$06$us = $20;$src$07$us = $18;
 while(1) {
  $35 = (($$in) + -1)|0;
  $byte$04$us = 0;$c$03$us = 0;$dstp$11$us = $dstp$06$us;$src$12$us = $src$07$us;
  while(1) {
   $21 = $c$03$us & 7;
   $22 = ($21|0)==(0);
   if ($22) {
    $23 = ((($src$12$us)) + 1|0);
    $24 = HEAP8[$src$12$us>>0]|0;
    $byte$1$us = $24;$src$2$us = $23;
   } else {
    $byte$1$us = $byte$04$us;$src$2$us = $src$12$us;
   }
   $25 = $byte$1$us&255;
   $26 = ($byte$1$us&255) >>> 7;
   $27 = $26&255;
   $28 = ($27|0)==($9|0);
   if (!($28)) {
    $29 = (($11) + ($27<<2)|0);
    $30 = HEAP32[$29>>2]|0;
    HEAP32[$dstp$11$us>>2] = $30;
   }
   $31 = $25 << 1;
   $32 = $31&255;
   $33 = ((($dstp$11$us)) + 4|0);
   $34 = (($c$03$us) + 1)|0;
   $exitcond = ($34|0)==($1|0);
   if ($exitcond) {
    $src$2$us$lcssa = $src$2$us;
    break;
   } else {
    $byte$04$us = $32;$c$03$us = $34;$dstp$11$us = $33;$src$12$us = $src$2$us;
   }
  }
  $scevgep = (($dstp$06$us) + ($1<<2)|0);
  $36 = (($src$2$us$lcssa) + ($14)|0);
  $37 = (($scevgep) + ($15<<2)|0);
  $38 = ($35|0)==(0);
  if ($38) {
   break;
  } else {
   $$in = $35;$dstp$06$us = $37;$src$07$us = $36;
  }
 }
 return;
}
function _BlitBtoNAlpha($info) {
 $info = $info|0;
 var $$in = 0, $$neg = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0;
 var $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0;
 var $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0;
 var $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $byte$08 = 0, $byte$1 = 0, $c$07 = 0, $dA$0 = 0, $dA$0$shrunk = 0, $dB$0 = 0, $dB$0$shrunk = 0, $dG$0 = 0;
 var $dG$0$shrunk = 0, $dR$0 = 0, $dR$0$shrunk = 0, $div = 0, $div1 = 0, $div2 = 0, $div3 = 0, $div4 = 0, $dst$012 = 0, $dst$1$lcssa = 0, $dst$19 = 0, $exitcond = 0, $scevgep = 0, $src$013 = 0, $src$1$lcssa = 0, $src$110 = 0, $src$2 = 0, $src$2$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 40|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($11)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($info)) + 44|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($info)) + 63|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = $17&255;
 $19 = ((($15)) + 9|0);
 $20 = HEAP8[$19>>0]|0;
 $21 = $20&255;
 $22 = (($1) + 7)|0;
 $$neg = (($22|0) / -8)&-1;
 $23 = (($$neg) + ($1))|0;
 $24 = (($23) + ($5))|0;
 $25 = ($3|0)==(0);
 if ($25) {
  return;
 }
 $26 = ((($info)) + 20|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = HEAP32[$info>>2]|0;
 $29 = ($1|0)>(0);
 $30 = ((($15)) + 28|0);
 $31 = ((($15)) + 32|0);
 $32 = ((($15)) + 33|0);
 $33 = ((($15)) + 34|0);
 $34 = ((($15)) + 12|0);
 $35 = ((($15)) + 16|0);
 $36 = ((($15)) + 20|0);
 $37 = ((($15)) + 24|0);
 $38 = Math_imul($1, $21)|0;
 $$in = $3;$dst$012 = $27;$src$013 = $28;
 while(1) {
  $39 = (($$in) + -1)|0;
  if ($29) {
   $byte$08 = 0;$c$07 = 0;$dst$19 = $dst$012;$src$110 = $src$013;
   while(1) {
    $40 = $c$07 & 7;
    $41 = ($40|0)==(0);
    if ($41) {
     $42 = ((($src$110)) + 1|0);
     $43 = HEAP8[$src$110>>0]|0;
     $byte$1 = $43;$src$2 = $42;
    } else {
     $byte$1 = $byte$08;$src$2 = $src$110;
    }
    $44 = $byte$1&255;
    $45 = ($byte$1&255) >>> 7;
    $46 = $45&255;
    $47 = (($13) + ($46<<2)|0);
    $48 = HEAP8[$47>>0]|0;
    $49 = $48&255;
    $50 = (((($13) + ($46<<2)|0)) + 1|0);
    $51 = HEAP8[$50>>0]|0;
    $52 = $51&255;
    $53 = (((($13) + ($46<<2)|0)) + 2|0);
    $54 = HEAP8[$53>>0]|0;
    $55 = $54&255;
    switch ($21|0) {
    case 1:  {
     $56 = HEAP8[$dst$19>>0]|0;
     $57 = $56&255;
     $58 = HEAP32[$34>>2]|0;
     $59 = $57 & $58;
     $60 = HEAP32[$31>>2]|0;
     $61 = $60 & 255;
     $62 = $59 >>> $61;
     $63 = HEAP32[$30>>2]|0;
     $64 = $63 & 255;
     $65 = (125924 + ($64<<2)|0);
     $66 = HEAP32[$65>>2]|0;
     $67 = (($66) + ($62)|0);
     $68 = HEAP8[$67>>0]|0;
     $69 = HEAP32[$35>>2]|0;
     $70 = $69 & $57;
     $71 = $60 >>> 8;
     $72 = $71 & 255;
     $73 = $70 >>> $72;
     $74 = $63 >>> 8;
     $75 = $74 & 255;
     $76 = (125924 + ($75<<2)|0);
     $77 = HEAP32[$76>>2]|0;
     $78 = (($77) + ($73)|0);
     $79 = HEAP8[$78>>0]|0;
     $80 = HEAP32[$36>>2]|0;
     $81 = $80 & $57;
     $82 = $60 >>> 16;
     $83 = $82 & 255;
     $84 = $81 >>> $83;
     $85 = $63 >>> 16;
     $86 = $85 & 255;
     $87 = (125924 + ($86<<2)|0);
     $88 = HEAP32[$87>>2]|0;
     $89 = (($88) + ($84)|0);
     $90 = HEAP8[$89>>0]|0;
     $91 = HEAP32[$37>>2]|0;
     $92 = $91 & $57;
     $93 = $60 >>> 24;
     $94 = $92 >>> $93;
     $95 = $63 >>> 24;
     $96 = (125924 + ($95<<2)|0);
     $97 = HEAP32[$96>>2]|0;
     $98 = (($97) + ($94)|0);
     $99 = HEAP8[$98>>0]|0;
     $dA$0$shrunk = $99;$dB$0$shrunk = $90;$dG$0$shrunk = $79;$dR$0$shrunk = $68;
     break;
    }
    case 2:  {
     $100 = HEAP16[$dst$19>>1]|0;
     $101 = $100&65535;
     $102 = HEAP32[$34>>2]|0;
     $103 = $101 & $102;
     $104 = HEAP32[$31>>2]|0;
     $105 = $104 & 255;
     $106 = $103 >>> $105;
     $107 = HEAP32[$30>>2]|0;
     $108 = $107 & 255;
     $109 = (125924 + ($108<<2)|0);
     $110 = HEAP32[$109>>2]|0;
     $111 = (($110) + ($106)|0);
     $112 = HEAP8[$111>>0]|0;
     $113 = HEAP32[$35>>2]|0;
     $114 = $113 & $101;
     $115 = $104 >>> 8;
     $116 = $115 & 255;
     $117 = $114 >>> $116;
     $118 = $107 >>> 8;
     $119 = $118 & 255;
     $120 = (125924 + ($119<<2)|0);
     $121 = HEAP32[$120>>2]|0;
     $122 = (($121) + ($117)|0);
     $123 = HEAP8[$122>>0]|0;
     $124 = HEAP32[$36>>2]|0;
     $125 = $124 & $101;
     $126 = $104 >>> 16;
     $127 = $126 & 255;
     $128 = $125 >>> $127;
     $129 = $107 >>> 16;
     $130 = $129 & 255;
     $131 = (125924 + ($130<<2)|0);
     $132 = HEAP32[$131>>2]|0;
     $133 = (($132) + ($128)|0);
     $134 = HEAP8[$133>>0]|0;
     $135 = HEAP32[$37>>2]|0;
     $136 = $135 & $101;
     $137 = $104 >>> 24;
     $138 = $136 >>> $137;
     $139 = $107 >>> 24;
     $140 = (125924 + ($139<<2)|0);
     $141 = HEAP32[$140>>2]|0;
     $142 = (($141) + ($138)|0);
     $143 = HEAP8[$142>>0]|0;
     $dA$0$shrunk = $143;$dB$0$shrunk = $134;$dG$0$shrunk = $123;$dR$0$shrunk = $112;
     break;
    }
    case 3:  {
     $144 = HEAP32[$31>>2]|0;
     $div3 = $144 >>> 3;
     $145 = $div3 & 31;
     $146 = (($dst$19) + ($145)|0);
     $147 = HEAP8[$146>>0]|0;
     $148 = $144 >>> 11;
     $div4 = $148 & 31;
     $149 = (($dst$19) + ($div4)|0);
     $150 = HEAP8[$149>>0]|0;
     $151 = $144 >>> 19;
     $152 = $151 & 31;
     $153 = (($dst$19) + ($152)|0);
     $154 = HEAP8[$153>>0]|0;
     $dA$0$shrunk = -1;$dB$0$shrunk = $154;$dG$0$shrunk = $150;$dR$0$shrunk = $147;
     break;
    }
    case 4:  {
     $155 = HEAP32[$dst$19>>2]|0;
     $156 = HEAP32[$34>>2]|0;
     $157 = $156 & $155;
     $158 = HEAP32[$31>>2]|0;
     $159 = $158 & 255;
     $160 = $157 >>> $159;
     $161 = HEAP32[$30>>2]|0;
     $162 = $161 & 255;
     $163 = (125924 + ($162<<2)|0);
     $164 = HEAP32[$163>>2]|0;
     $165 = (($164) + ($160)|0);
     $166 = HEAP8[$165>>0]|0;
     $167 = HEAP32[$35>>2]|0;
     $168 = $167 & $155;
     $169 = $158 >>> 8;
     $170 = $169 & 255;
     $171 = $168 >>> $170;
     $172 = $161 >>> 8;
     $173 = $172 & 255;
     $174 = (125924 + ($173<<2)|0);
     $175 = HEAP32[$174>>2]|0;
     $176 = (($175) + ($171)|0);
     $177 = HEAP8[$176>>0]|0;
     $178 = HEAP32[$36>>2]|0;
     $179 = $178 & $155;
     $180 = $158 >>> 16;
     $181 = $180 & 255;
     $182 = $179 >>> $181;
     $183 = $161 >>> 16;
     $184 = $183 & 255;
     $185 = (125924 + ($184<<2)|0);
     $186 = HEAP32[$185>>2]|0;
     $187 = (($186) + ($182)|0);
     $188 = HEAP8[$187>>0]|0;
     $189 = HEAP32[$37>>2]|0;
     $190 = $189 & $155;
     $191 = $158 >>> 24;
     $192 = $190 >>> $191;
     $193 = $161 >>> 24;
     $194 = (125924 + ($193<<2)|0);
     $195 = HEAP32[$194>>2]|0;
     $196 = (($195) + ($192)|0);
     $197 = HEAP8[$196>>0]|0;
     $dA$0$shrunk = $197;$dB$0$shrunk = $188;$dG$0$shrunk = $177;$dR$0$shrunk = $166;
     break;
    }
    default: {
     $dA$0$shrunk = 0;$dB$0$shrunk = 0;$dG$0$shrunk = 0;$dR$0$shrunk = 0;
    }
    }
    $dA$0 = $dA$0$shrunk&255;
    $dB$0 = $dB$0$shrunk&255;
    $dG$0 = $dG$0$shrunk&255;
    $dR$0 = $dR$0$shrunk&255;
    $198 = (($49) - ($dR$0))|0;
    $199 = Math_imul($198, $18)|0;
    $200 = (($199>>>0) / 255)&-1;
    $201 = (($200) + ($dR$0))|0;
    $202 = (($52) - ($dG$0))|0;
    $203 = Math_imul($202, $18)|0;
    $204 = (($203>>>0) / 255)&-1;
    $205 = (($204) + ($dG$0))|0;
    $206 = (($55) - ($dB$0))|0;
    $207 = Math_imul($206, $18)|0;
    $208 = (($207>>>0) / 255)&-1;
    $209 = (($208) + ($dB$0))|0;
    $210 = (($dA$0) + ($18))|0;
    $211 = Math_imul($dA$0, $18)|0;
    $212 = (($211>>>0) / 255)&-1;
    $213 = (($210) - ($212))|0;
    switch ($21|0) {
    case 1:  {
     $214 = HEAP32[$30>>2]|0;
     $215 = $214 & 255;
     $216 = $201 >>> $215;
     $217 = HEAP32[$31>>2]|0;
     $218 = $217 & 255;
     $219 = $216 << $218;
     $220 = $214 >>> 8;
     $221 = $220 & 255;
     $222 = $205 >>> $221;
     $223 = $217 >>> 8;
     $224 = $223 & 255;
     $225 = $222 << $224;
     $226 = $225 | $219;
     $227 = $214 >>> 16;
     $228 = $227 & 255;
     $229 = $209 >>> $228;
     $230 = $217 >>> 16;
     $231 = $230 & 255;
     $232 = $229 << $231;
     $233 = $226 | $232;
     $234 = $214 >>> 24;
     $235 = $213 >>> $234;
     $236 = $217 >>> 24;
     $237 = $235 << $236;
     $238 = $233 | $237;
     $239 = $238&255;
     HEAP8[$dst$19>>0] = $239;
     break;
    }
    case 2:  {
     $240 = HEAP32[$30>>2]|0;
     $241 = $240 & 255;
     $242 = $201 >>> $241;
     $243 = HEAP32[$31>>2]|0;
     $244 = $243 & 255;
     $245 = $242 << $244;
     $246 = $240 >>> 8;
     $247 = $246 & 255;
     $248 = $205 >>> $247;
     $249 = $243 >>> 8;
     $250 = $249 & 255;
     $251 = $248 << $250;
     $252 = $251 | $245;
     $253 = $240 >>> 16;
     $254 = $253 & 255;
     $255 = $209 >>> $254;
     $256 = $243 >>> 16;
     $257 = $256 & 255;
     $258 = $255 << $257;
     $259 = $252 | $258;
     $260 = $240 >>> 24;
     $261 = $213 >>> $260;
     $262 = $243 >>> 24;
     $263 = $261 << $262;
     $264 = $259 | $263;
     $265 = $264&65535;
     HEAP16[$dst$19>>1] = $265;
     break;
    }
    case 3:  {
     $266 = $201&255;
     $267 = HEAP8[$31>>0]|0;
     $div = ($267&255) >>> 3;
     $268 = $div&255;
     $269 = (($dst$19) + ($268)|0);
     HEAP8[$269>>0] = $266;
     $270 = $205&255;
     $271 = HEAP8[$32>>0]|0;
     $div1 = ($271&255) >>> 3;
     $272 = $div1&255;
     $273 = (($dst$19) + ($272)|0);
     HEAP8[$273>>0] = $270;
     $274 = $209&255;
     $275 = HEAP8[$33>>0]|0;
     $div2 = ($275&255) >>> 3;
     $276 = $div2&255;
     $277 = (($dst$19) + ($276)|0);
     HEAP8[$277>>0] = $274;
     break;
    }
    case 4:  {
     $278 = HEAP32[$30>>2]|0;
     $279 = $278 & 255;
     $280 = $201 >>> $279;
     $281 = HEAP32[$31>>2]|0;
     $282 = $281 & 255;
     $283 = $280 << $282;
     $284 = $278 >>> 8;
     $285 = $284 & 255;
     $286 = $205 >>> $285;
     $287 = $281 >>> 8;
     $288 = $287 & 255;
     $289 = $286 << $288;
     $290 = $289 | $283;
     $291 = $278 >>> 16;
     $292 = $291 & 255;
     $293 = $209 >>> $292;
     $294 = $281 >>> 16;
     $295 = $294 & 255;
     $296 = $293 << $295;
     $297 = $290 | $296;
     $298 = $278 >>> 24;
     $299 = $213 >>> $298;
     $300 = $281 >>> 24;
     $301 = $299 << $300;
     $302 = $297 | $301;
     HEAP32[$dst$19>>2] = $302;
     break;
    }
    default: {
    }
    }
    $303 = $44 << 1;
    $304 = $303&255;
    $305 = (($dst$19) + ($21)|0);
    $306 = (($c$07) + 1)|0;
    $exitcond = ($306|0)==($1|0);
    if ($exitcond) {
     $src$2$lcssa = $src$2;
     break;
    } else {
     $byte$08 = $304;$c$07 = $306;$dst$19 = $305;$src$110 = $src$2;
    }
   }
   $scevgep = (($dst$012) + ($38)|0);
   $dst$1$lcssa = $scevgep;$src$1$lcssa = $src$2$lcssa;
  } else {
   $dst$1$lcssa = $dst$012;$src$1$lcssa = $src$013;
  }
  $307 = (($src$1$lcssa) + ($24)|0);
  $308 = (($dst$1$lcssa) + ($7)|0);
  $309 = ($39|0)==(0);
  if ($309) {
   break;
  } else {
   $$in = $39;$dst$012 = $308;$src$013 = $307;
  }
 }
 return;
}
function _BlitBtoNAlphaKey($info) {
 $info = $info|0;
 var $$in = 0, $$neg = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0;
 var $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0;
 var $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0;
 var $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0;
 var $312 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $byte$08 = 0, $byte$1 = 0, $c$07 = 0, $dA$0 = 0, $dA$0$shrunk = 0;
 var $dB$0 = 0, $dB$0$shrunk = 0, $dG$0 = 0, $dG$0$shrunk = 0, $dR$0 = 0, $dR$0$shrunk = 0, $div = 0, $div1 = 0, $div2 = 0, $div3 = 0, $div4 = 0, $dst$013 = 0, $dst$1$lcssa = 0, $dst$19 = 0, $exitcond = 0, $scevgep = 0, $src$014 = 0, $src$1$lcssa = 0, $src$111 = 0, $src$2 = 0;
 var $src$2$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 40|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($info)) + 44|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($9)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($13)) + 4|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($info)) + 63|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = $17&255;
 $19 = ((($info)) + 56|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ((($11)) + 9|0);
 $22 = HEAP8[$21>>0]|0;
 $23 = $22&255;
 $24 = (($1) + 7)|0;
 $$neg = (($24|0) / -8)&-1;
 $25 = (($$neg) + ($1))|0;
 $26 = (($25) + ($5))|0;
 $27 = ($3|0)==(0);
 if ($27) {
  return;
 }
 $28 = ((($info)) + 20|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = HEAP32[$info>>2]|0;
 $31 = ($1|0)>(0);
 $32 = ((($11)) + 28|0);
 $33 = ((($11)) + 32|0);
 $34 = ((($11)) + 33|0);
 $35 = ((($11)) + 34|0);
 $36 = ((($11)) + 12|0);
 $37 = ((($11)) + 16|0);
 $38 = ((($11)) + 20|0);
 $39 = ((($11)) + 24|0);
 $40 = Math_imul($1, $23)|0;
 $$in = $3;$dst$013 = $29;$src$014 = $30;
 while(1) {
  $41 = (($$in) + -1)|0;
  if ($31) {
   $byte$08 = 0;$c$07 = 0;$dst$19 = $dst$013;$src$111 = $src$014;
   while(1) {
    $42 = $c$07 & 7;
    $43 = ($42|0)==(0);
    if ($43) {
     $44 = ((($src$111)) + 1|0);
     $45 = HEAP8[$src$111>>0]|0;
     $byte$1 = $45;$src$2 = $44;
    } else {
     $byte$1 = $byte$08;$src$2 = $src$111;
    }
    $46 = $byte$1&255;
    $47 = ($byte$1&255) >>> 7;
    $48 = $47&255;
    $49 = ($48|0)==($20|0);
    L12: do {
     if (!($49)) {
      $50 = (($15) + ($48<<2)|0);
      $51 = HEAP8[$50>>0]|0;
      $52 = $51&255;
      $53 = (((($15) + ($48<<2)|0)) + 1|0);
      $54 = HEAP8[$53>>0]|0;
      $55 = $54&255;
      $56 = (((($15) + ($48<<2)|0)) + 2|0);
      $57 = HEAP8[$56>>0]|0;
      $58 = $57&255;
      switch ($23|0) {
      case 1:  {
       $59 = HEAP8[$dst$19>>0]|0;
       $60 = $59&255;
       $61 = HEAP32[$36>>2]|0;
       $62 = $60 & $61;
       $63 = HEAP32[$33>>2]|0;
       $64 = $63 & 255;
       $65 = $62 >>> $64;
       $66 = HEAP32[$32>>2]|0;
       $67 = $66 & 255;
       $68 = (125924 + ($67<<2)|0);
       $69 = HEAP32[$68>>2]|0;
       $70 = (($69) + ($65)|0);
       $71 = HEAP8[$70>>0]|0;
       $72 = HEAP32[$37>>2]|0;
       $73 = $72 & $60;
       $74 = $63 >>> 8;
       $75 = $74 & 255;
       $76 = $73 >>> $75;
       $77 = $66 >>> 8;
       $78 = $77 & 255;
       $79 = (125924 + ($78<<2)|0);
       $80 = HEAP32[$79>>2]|0;
       $81 = (($80) + ($76)|0);
       $82 = HEAP8[$81>>0]|0;
       $83 = HEAP32[$38>>2]|0;
       $84 = $83 & $60;
       $85 = $63 >>> 16;
       $86 = $85 & 255;
       $87 = $84 >>> $86;
       $88 = $66 >>> 16;
       $89 = $88 & 255;
       $90 = (125924 + ($89<<2)|0);
       $91 = HEAP32[$90>>2]|0;
       $92 = (($91) + ($87)|0);
       $93 = HEAP8[$92>>0]|0;
       $94 = HEAP32[$39>>2]|0;
       $95 = $94 & $60;
       $96 = $63 >>> 24;
       $97 = $95 >>> $96;
       $98 = $66 >>> 24;
       $99 = (125924 + ($98<<2)|0);
       $100 = HEAP32[$99>>2]|0;
       $101 = (($100) + ($97)|0);
       $102 = HEAP8[$101>>0]|0;
       $dA$0$shrunk = $102;$dB$0$shrunk = $93;$dG$0$shrunk = $82;$dR$0$shrunk = $71;
       break;
      }
      case 2:  {
       $103 = HEAP16[$dst$19>>1]|0;
       $104 = $103&65535;
       $105 = HEAP32[$36>>2]|0;
       $106 = $104 & $105;
       $107 = HEAP32[$33>>2]|0;
       $108 = $107 & 255;
       $109 = $106 >>> $108;
       $110 = HEAP32[$32>>2]|0;
       $111 = $110 & 255;
       $112 = (125924 + ($111<<2)|0);
       $113 = HEAP32[$112>>2]|0;
       $114 = (($113) + ($109)|0);
       $115 = HEAP8[$114>>0]|0;
       $116 = HEAP32[$37>>2]|0;
       $117 = $116 & $104;
       $118 = $107 >>> 8;
       $119 = $118 & 255;
       $120 = $117 >>> $119;
       $121 = $110 >>> 8;
       $122 = $121 & 255;
       $123 = (125924 + ($122<<2)|0);
       $124 = HEAP32[$123>>2]|0;
       $125 = (($124) + ($120)|0);
       $126 = HEAP8[$125>>0]|0;
       $127 = HEAP32[$38>>2]|0;
       $128 = $127 & $104;
       $129 = $107 >>> 16;
       $130 = $129 & 255;
       $131 = $128 >>> $130;
       $132 = $110 >>> 16;
       $133 = $132 & 255;
       $134 = (125924 + ($133<<2)|0);
       $135 = HEAP32[$134>>2]|0;
       $136 = (($135) + ($131)|0);
       $137 = HEAP8[$136>>0]|0;
       $138 = HEAP32[$39>>2]|0;
       $139 = $138 & $104;
       $140 = $107 >>> 24;
       $141 = $139 >>> $140;
       $142 = $110 >>> 24;
       $143 = (125924 + ($142<<2)|0);
       $144 = HEAP32[$143>>2]|0;
       $145 = (($144) + ($141)|0);
       $146 = HEAP8[$145>>0]|0;
       $dA$0$shrunk = $146;$dB$0$shrunk = $137;$dG$0$shrunk = $126;$dR$0$shrunk = $115;
       break;
      }
      case 3:  {
       $147 = HEAP32[$33>>2]|0;
       $div3 = $147 >>> 3;
       $148 = $div3 & 31;
       $149 = (($dst$19) + ($148)|0);
       $150 = HEAP8[$149>>0]|0;
       $151 = $147 >>> 11;
       $div4 = $151 & 31;
       $152 = (($dst$19) + ($div4)|0);
       $153 = HEAP8[$152>>0]|0;
       $154 = $147 >>> 19;
       $155 = $154 & 31;
       $156 = (($dst$19) + ($155)|0);
       $157 = HEAP8[$156>>0]|0;
       $dA$0$shrunk = -1;$dB$0$shrunk = $157;$dG$0$shrunk = $153;$dR$0$shrunk = $150;
       break;
      }
      case 4:  {
       $158 = HEAP32[$dst$19>>2]|0;
       $159 = HEAP32[$36>>2]|0;
       $160 = $159 & $158;
       $161 = HEAP32[$33>>2]|0;
       $162 = $161 & 255;
       $163 = $160 >>> $162;
       $164 = HEAP32[$32>>2]|0;
       $165 = $164 & 255;
       $166 = (125924 + ($165<<2)|0);
       $167 = HEAP32[$166>>2]|0;
       $168 = (($167) + ($163)|0);
       $169 = HEAP8[$168>>0]|0;
       $170 = HEAP32[$37>>2]|0;
       $171 = $170 & $158;
       $172 = $161 >>> 8;
       $173 = $172 & 255;
       $174 = $171 >>> $173;
       $175 = $164 >>> 8;
       $176 = $175 & 255;
       $177 = (125924 + ($176<<2)|0);
       $178 = HEAP32[$177>>2]|0;
       $179 = (($178) + ($174)|0);
       $180 = HEAP8[$179>>0]|0;
       $181 = HEAP32[$38>>2]|0;
       $182 = $181 & $158;
       $183 = $161 >>> 16;
       $184 = $183 & 255;
       $185 = $182 >>> $184;
       $186 = $164 >>> 16;
       $187 = $186 & 255;
       $188 = (125924 + ($187<<2)|0);
       $189 = HEAP32[$188>>2]|0;
       $190 = (($189) + ($185)|0);
       $191 = HEAP8[$190>>0]|0;
       $192 = HEAP32[$39>>2]|0;
       $193 = $192 & $158;
       $194 = $161 >>> 24;
       $195 = $193 >>> $194;
       $196 = $164 >>> 24;
       $197 = (125924 + ($196<<2)|0);
       $198 = HEAP32[$197>>2]|0;
       $199 = (($198) + ($195)|0);
       $200 = HEAP8[$199>>0]|0;
       $dA$0$shrunk = $200;$dB$0$shrunk = $191;$dG$0$shrunk = $180;$dR$0$shrunk = $169;
       break;
      }
      default: {
       $dA$0$shrunk = 0;$dB$0$shrunk = 0;$dG$0$shrunk = 0;$dR$0$shrunk = 0;
      }
      }
      $dA$0 = $dA$0$shrunk&255;
      $dB$0 = $dB$0$shrunk&255;
      $dG$0 = $dG$0$shrunk&255;
      $dR$0 = $dR$0$shrunk&255;
      $201 = (($52) - ($dR$0))|0;
      $202 = Math_imul($201, $18)|0;
      $203 = (($202>>>0) / 255)&-1;
      $204 = (($203) + ($dR$0))|0;
      $205 = (($55) - ($dG$0))|0;
      $206 = Math_imul($205, $18)|0;
      $207 = (($206>>>0) / 255)&-1;
      $208 = (($207) + ($dG$0))|0;
      $209 = (($58) - ($dB$0))|0;
      $210 = Math_imul($209, $18)|0;
      $211 = (($210>>>0) / 255)&-1;
      $212 = (($211) + ($dB$0))|0;
      $213 = (($dA$0) + ($18))|0;
      $214 = Math_imul($dA$0, $18)|0;
      $215 = (($214>>>0) / 255)&-1;
      $216 = (($213) - ($215))|0;
      switch ($23|0) {
      case 1:  {
       $217 = HEAP32[$32>>2]|0;
       $218 = $217 & 255;
       $219 = $204 >>> $218;
       $220 = HEAP32[$33>>2]|0;
       $221 = $220 & 255;
       $222 = $219 << $221;
       $223 = $217 >>> 8;
       $224 = $223 & 255;
       $225 = $208 >>> $224;
       $226 = $220 >>> 8;
       $227 = $226 & 255;
       $228 = $225 << $227;
       $229 = $228 | $222;
       $230 = $217 >>> 16;
       $231 = $230 & 255;
       $232 = $212 >>> $231;
       $233 = $220 >>> 16;
       $234 = $233 & 255;
       $235 = $232 << $234;
       $236 = $229 | $235;
       $237 = $217 >>> 24;
       $238 = $216 >>> $237;
       $239 = $220 >>> 24;
       $240 = $238 << $239;
       $241 = $236 | $240;
       $242 = $241&255;
       HEAP8[$dst$19>>0] = $242;
       break L12;
       break;
      }
      case 2:  {
       $243 = HEAP32[$32>>2]|0;
       $244 = $243 & 255;
       $245 = $204 >>> $244;
       $246 = HEAP32[$33>>2]|0;
       $247 = $246 & 255;
       $248 = $245 << $247;
       $249 = $243 >>> 8;
       $250 = $249 & 255;
       $251 = $208 >>> $250;
       $252 = $246 >>> 8;
       $253 = $252 & 255;
       $254 = $251 << $253;
       $255 = $254 | $248;
       $256 = $243 >>> 16;
       $257 = $256 & 255;
       $258 = $212 >>> $257;
       $259 = $246 >>> 16;
       $260 = $259 & 255;
       $261 = $258 << $260;
       $262 = $255 | $261;
       $263 = $243 >>> 24;
       $264 = $216 >>> $263;
       $265 = $246 >>> 24;
       $266 = $264 << $265;
       $267 = $262 | $266;
       $268 = $267&65535;
       HEAP16[$dst$19>>1] = $268;
       break L12;
       break;
      }
      case 3:  {
       $269 = $204&255;
       $270 = HEAP8[$33>>0]|0;
       $div = ($270&255) >>> 3;
       $271 = $div&255;
       $272 = (($dst$19) + ($271)|0);
       HEAP8[$272>>0] = $269;
       $273 = $208&255;
       $274 = HEAP8[$34>>0]|0;
       $div1 = ($274&255) >>> 3;
       $275 = $div1&255;
       $276 = (($dst$19) + ($275)|0);
       HEAP8[$276>>0] = $273;
       $277 = $212&255;
       $278 = HEAP8[$35>>0]|0;
       $div2 = ($278&255) >>> 3;
       $279 = $div2&255;
       $280 = (($dst$19) + ($279)|0);
       HEAP8[$280>>0] = $277;
       break L12;
       break;
      }
      case 4:  {
       $281 = HEAP32[$32>>2]|0;
       $282 = $281 & 255;
       $283 = $204 >>> $282;
       $284 = HEAP32[$33>>2]|0;
       $285 = $284 & 255;
       $286 = $283 << $285;
       $287 = $281 >>> 8;
       $288 = $287 & 255;
       $289 = $208 >>> $288;
       $290 = $284 >>> 8;
       $291 = $290 & 255;
       $292 = $289 << $291;
       $293 = $292 | $286;
       $294 = $281 >>> 16;
       $295 = $294 & 255;
       $296 = $212 >>> $295;
       $297 = $284 >>> 16;
       $298 = $297 & 255;
       $299 = $296 << $298;
       $300 = $293 | $299;
       $301 = $281 >>> 24;
       $302 = $216 >>> $301;
       $303 = $284 >>> 24;
       $304 = $302 << $303;
       $305 = $300 | $304;
       HEAP32[$dst$19>>2] = $305;
       break L12;
       break;
      }
      default: {
       break L12;
      }
      }
     }
    } while(0);
    $306 = $46 << 1;
    $307 = $306&255;
    $308 = (($dst$19) + ($23)|0);
    $309 = (($c$07) + 1)|0;
    $exitcond = ($309|0)==($1|0);
    if ($exitcond) {
     $src$2$lcssa = $src$2;
     break;
    } else {
     $byte$08 = $307;$c$07 = $309;$dst$19 = $308;$src$111 = $src$2;
    }
   }
   $scevgep = (($dst$013) + ($40)|0);
   $dst$1$lcssa = $scevgep;$src$1$lcssa = $src$2$lcssa;
  } else {
   $dst$1$lcssa = $dst$013;$src$1$lcssa = $src$014;
  }
  $310 = (($src$1$lcssa) + ($26)|0);
  $311 = (($dst$1$lcssa) + ($7)|0);
  $312 = ($41|0)==(0);
  if ($312) {
   break;
  } else {
   $$in = $41;$dst$013 = $311;$src$014 = $310;
  }
 }
 return;
}
function _SDL_CalculateBlit1($surface) {
 $surface = $surface|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $which$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($surface)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 8|0);
 $6 = HEAP16[$5>>1]|0;
 $7 = $6&255;
 $8 = ($7&255)<(8);
 $9 = ($6&65535) >>> 8;
 $10 = $9&65535;
 $which$0 = $8 ? 0 : $10;
 $11 = ((($1)) + 68|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = $12 & -28673;
 switch ($13|0) {
 case 0:  {
  $14 = (126640 + ($which$0<<2)|0);
  $15 = HEAP32[$14>>2]|0;
  $$0 = $15;
  return ($$0|0);
  break;
 }
 case 256:  {
  $16 = (126660 + ($which$0<<2)|0);
  $17 = HEAP32[$16>>2]|0;
  $$0 = $17;
  return ($$0|0);
  break;
 }
 case 18:  {
  $18 = ($which$0>>>0)>(1);
  $19 = $18 ? 4494 : 0;
  $$0 = $19;
  return ($$0|0);
  break;
 }
 case 274:  {
  $20 = ($which$0>>>0)>(1);
  $21 = $20 ? 4495 : 0;
  $$0 = $21;
  return ($$0|0);
  break;
 }
 default: {
  $$0 = 0;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function _Blit1to1($info) {
 $info = $info|0;
 var $$lcssa = 0, $$lcssa55 = 0, $$lcssa56 = 0, $$lcssa57 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$01 = 0, $dst$01$us = 0, $dst$01$us5 = 0, $dst$1$us = 0, $dst$2$us9 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $n$0$us = 0, $n$1$us11 = 0;
 var $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $src$02 = 0, $src$02$us = 0, $src$02$us4 = 0, $src$1$us = 0, $src$2$us10 = 0, $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$info>>2]|0;
 $5 = ((($info)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($info)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($info)) + 36|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($info)) + 48|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($3) + -1)|0;
 $14 = ($3|0)==(0);
 if ($14) {
  return;
 }
 $15 = (($1) + 7)|0;
 $16 = (($15|0) / 8)&-1;
 $17 = $1 & 7;
 switch ($17|0) {
 case 0:  {
  $71 = $13;$dst$01$us = $8;$src$02$us = $4;
  while(1) {
   $dst$1$us = $dst$01$us;$n$0$us = $16;$src$1$us = $src$02$us;
   while(1) {
    $18 = HEAP8[$src$1$us>>0]|0;
    $19 = $18&255;
    $20 = (($12) + ($19)|0);
    $21 = HEAP8[$20>>0]|0;
    HEAP8[$dst$1$us>>0] = $21;
    $22 = ((($dst$1$us)) + 1|0);
    $23 = ((($src$1$us)) + 1|0);
    $24 = HEAP8[$23>>0]|0;
    $25 = $24&255;
    $26 = (($12) + ($25)|0);
    $27 = HEAP8[$26>>0]|0;
    HEAP8[$22>>0] = $27;
    $28 = ((($dst$1$us)) + 2|0);
    $29 = ((($src$1$us)) + 2|0);
    $30 = HEAP8[$29>>0]|0;
    $31 = $30&255;
    $32 = (($12) + ($31)|0);
    $33 = HEAP8[$32>>0]|0;
    HEAP8[$28>>0] = $33;
    $34 = ((($dst$1$us)) + 3|0);
    $35 = ((($src$1$us)) + 3|0);
    $36 = HEAP8[$35>>0]|0;
    $37 = $36&255;
    $38 = (($12) + ($37)|0);
    $39 = HEAP8[$38>>0]|0;
    HEAP8[$34>>0] = $39;
    $40 = ((($dst$1$us)) + 4|0);
    $41 = ((($src$1$us)) + 4|0);
    $42 = HEAP8[$41>>0]|0;
    $43 = $42&255;
    $44 = (($12) + ($43)|0);
    $45 = HEAP8[$44>>0]|0;
    HEAP8[$40>>0] = $45;
    $46 = ((($dst$1$us)) + 5|0);
    $47 = ((($src$1$us)) + 5|0);
    $48 = HEAP8[$47>>0]|0;
    $49 = $48&255;
    $50 = (($12) + ($49)|0);
    $51 = HEAP8[$50>>0]|0;
    HEAP8[$46>>0] = $51;
    $52 = ((($dst$1$us)) + 6|0);
    $53 = ((($src$1$us)) + 6|0);
    $54 = HEAP8[$53>>0]|0;
    $55 = $54&255;
    $56 = (($12) + ($55)|0);
    $57 = HEAP8[$56>>0]|0;
    HEAP8[$52>>0] = $57;
    $58 = ((($dst$1$us)) + 7|0);
    $59 = ((($src$1$us)) + 7|0);
    $60 = HEAP8[$59>>0]|0;
    $61 = $60&255;
    $62 = (($12) + ($61)|0);
    $63 = HEAP8[$62>>0]|0;
    HEAP8[$58>>0] = $63;
    $64 = ((($dst$1$us)) + 8|0);
    $65 = ((($src$1$us)) + 8|0);
    $66 = (($n$0$us) + -1)|0;
    $67 = ($n$0$us|0)>(1);
    if ($67) {
     $dst$1$us = $64;$n$0$us = $66;$src$1$us = $65;
    } else {
     $$lcssa = $64;$$lcssa55 = $65;
     break;
    }
   }
   $68 = (($$lcssa55) + ($6)|0);
   $69 = (($$lcssa) + ($10)|0);
   $70 = (($71) + -1)|0;
   $72 = ($71|0)==(0);
   if ($72) {
    break;
   } else {
    $71 = $70;$dst$01$us = $69;$src$02$us = $68;
   }
  }
  return;
  break;
 }
 case 7:  {
  $126 = $13;$dst$01$us5 = $8;$src$02$us4 = $4;
  while(1) {
   $dst$2$us9 = $dst$01$us5;$n$1$us11 = $16;$src$2$us10 = $src$02$us4;
   while(1) {
    $82 = HEAP8[$src$2$us10>>0]|0;
    $83 = $82&255;
    $84 = (($12) + ($83)|0);
    $85 = HEAP8[$84>>0]|0;
    HEAP8[$dst$2$us9>>0] = $85;
    $86 = ((($dst$2$us9)) + 1|0);
    $87 = ((($src$2$us10)) + 1|0);
    $88 = HEAP8[$87>>0]|0;
    $89 = $88&255;
    $90 = (($12) + ($89)|0);
    $91 = HEAP8[$90>>0]|0;
    HEAP8[$86>>0] = $91;
    $92 = ((($dst$2$us9)) + 2|0);
    $93 = ((($src$2$us10)) + 2|0);
    $94 = HEAP8[$93>>0]|0;
    $95 = $94&255;
    $96 = (($12) + ($95)|0);
    $97 = HEAP8[$96>>0]|0;
    HEAP8[$92>>0] = $97;
    $98 = ((($dst$2$us9)) + 3|0);
    $99 = ((($src$2$us10)) + 3|0);
    $100 = HEAP8[$99>>0]|0;
    $101 = $100&255;
    $102 = (($12) + ($101)|0);
    $103 = HEAP8[$102>>0]|0;
    HEAP8[$98>>0] = $103;
    $104 = ((($dst$2$us9)) + 4|0);
    $105 = ((($src$2$us10)) + 4|0);
    $106 = HEAP8[$105>>0]|0;
    $107 = $106&255;
    $108 = (($12) + ($107)|0);
    $109 = HEAP8[$108>>0]|0;
    HEAP8[$104>>0] = $109;
    $110 = ((($dst$2$us9)) + 5|0);
    $111 = ((($src$2$us10)) + 5|0);
    $112 = HEAP8[$111>>0]|0;
    $113 = $112&255;
    $114 = (($12) + ($113)|0);
    $115 = HEAP8[$114>>0]|0;
    HEAP8[$110>>0] = $115;
    $116 = ((($dst$2$us9)) + 6|0);
    $117 = ((($src$2$us10)) + 6|0);
    $118 = HEAP8[$117>>0]|0;
    $119 = $118&255;
    $120 = (($12) + ($119)|0);
    $121 = HEAP8[$120>>0]|0;
    HEAP8[$116>>0] = $121;
    $79 = ((($dst$2$us9)) + 7|0);
    $75 = ((($src$2$us10)) + 7|0);
    $122 = ($n$1$us11|0)>(1);
    if (!($122)) {
     $$lcssa56 = $79;$$lcssa57 = $75;
     break;
    }
    $73 = (($n$1$us11) + -1)|0;
    $74 = HEAP8[$75>>0]|0;
    $76 = $74&255;
    $77 = (($12) + ($76)|0);
    $78 = HEAP8[$77>>0]|0;
    HEAP8[$79>>0] = $78;
    $80 = ((($dst$2$us9)) + 8|0);
    $81 = ((($src$2$us10)) + 8|0);
    $dst$2$us9 = $80;$n$1$us11 = $73;$src$2$us10 = $81;
   }
   $123 = (($$lcssa57) + ($6)|0);
   $124 = (($$lcssa56) + ($10)|0);
   $125 = (($126) + -1)|0;
   $127 = ($126|0)==(0);
   if ($127) {
    break;
   } else {
    $126 = $125;$dst$01$us5 = $124;$src$02$us4 = $123;
   }
  }
  return;
  break;
 }
 default: {
  $181 = $13;$dst$01 = $8;$src$02 = $4;
  L5: while(1) {
   switch ($17|0) {
   case 1:  {
    $dst$8 = $dst$01;$n$7 = $16;$src$8 = $src$02;
    label = 17;
    break;
   }
   case 2:  {
    $dst$7 = $dst$01;$n$6 = $16;$src$7 = $src$02;
    label = 16;
    break;
   }
   case 6:  {
    $dst$3 = $dst$01;$n$2 = $16;$src$3 = $src$02;
    label = 12;
    break;
   }
   case 5:  {
    $dst$4 = $dst$01;$n$3 = $16;$src$4 = $src$02;
    label = 13;
    break;
   }
   case 4:  {
    $dst$5 = $dst$01;$n$4 = $16;$src$5 = $src$02;
    label = 14;
    break;
   }
   case 3:  {
    $dst$6 = $dst$01;$n$5 = $16;$src$6 = $src$02;
    label = 15;
    break;
   }
   default: {
    label = 18;
    break L5;
   }
   }
   while(1) {
    if ((label|0) == 12) {
     label = 0;
     $143 = HEAP8[$src$3>>0]|0;
     $144 = $143&255;
     $145 = (($12) + ($144)|0);
     $146 = HEAP8[$145>>0]|0;
     HEAP8[$dst$3>>0] = $146;
     $147 = ((($dst$3)) + 1|0);
     $148 = ((($src$3)) + 1|0);
     $dst$4 = $147;$n$3 = $n$2;$src$4 = $148;
     label = 13;
     continue;
    }
    else if ((label|0) == 13) {
     label = 0;
     $149 = HEAP8[$src$4>>0]|0;
     $150 = $149&255;
     $151 = (($12) + ($150)|0);
     $152 = HEAP8[$151>>0]|0;
     HEAP8[$dst$4>>0] = $152;
     $153 = ((($dst$4)) + 1|0);
     $154 = ((($src$4)) + 1|0);
     $dst$5 = $153;$n$4 = $n$3;$src$5 = $154;
     label = 14;
     continue;
    }
    else if ((label|0) == 14) {
     label = 0;
     $155 = HEAP8[$src$5>>0]|0;
     $156 = $155&255;
     $157 = (($12) + ($156)|0);
     $158 = HEAP8[$157>>0]|0;
     HEAP8[$dst$5>>0] = $158;
     $159 = ((($dst$5)) + 1|0);
     $160 = ((($src$5)) + 1|0);
     $dst$6 = $159;$n$5 = $n$4;$src$6 = $160;
     label = 15;
     continue;
    }
    else if ((label|0) == 15) {
     label = 0;
     $161 = HEAP8[$src$6>>0]|0;
     $162 = $161&255;
     $163 = (($12) + ($162)|0);
     $164 = HEAP8[$163>>0]|0;
     HEAP8[$dst$6>>0] = $164;
     $165 = ((($dst$6)) + 1|0);
     $166 = ((($src$6)) + 1|0);
     $dst$7 = $165;$n$6 = $n$5;$src$7 = $166;
     label = 16;
     continue;
    }
    else if ((label|0) == 16) {
     label = 0;
     $167 = HEAP8[$src$7>>0]|0;
     $168 = $167&255;
     $169 = (($12) + ($168)|0);
     $170 = HEAP8[$169>>0]|0;
     HEAP8[$dst$7>>0] = $170;
     $171 = ((($dst$7)) + 1|0);
     $172 = ((($src$7)) + 1|0);
     $dst$8 = $171;$n$7 = $n$6;$src$8 = $172;
     label = 17;
     continue;
    }
    else if ((label|0) == 17) {
     label = 0;
     $173 = HEAP8[$src$8>>0]|0;
     $174 = $173&255;
     $175 = (($12) + ($174)|0);
     $176 = HEAP8[$175>>0]|0;
     HEAP8[$dst$8>>0] = $176;
     $134 = ((($dst$8)) + 1|0);
     $130 = ((($src$8)) + 1|0);
     $177 = ($n$7|0)>(1);
     if (!($177)) {
      break;
     }
     $128 = (($n$7) + -1)|0;
     $129 = HEAP8[$130>>0]|0;
     $131 = $129&255;
     $132 = (($12) + ($131)|0);
     $133 = HEAP8[$132>>0]|0;
     HEAP8[$134>>0] = $133;
     $135 = ((($dst$8)) + 2|0);
     $136 = ((($src$8)) + 2|0);
     $137 = HEAP8[$136>>0]|0;
     $138 = $137&255;
     $139 = (($12) + ($138)|0);
     $140 = HEAP8[$139>>0]|0;
     HEAP8[$135>>0] = $140;
     $141 = ((($dst$8)) + 3|0);
     $142 = ((($src$8)) + 3|0);
     $dst$3 = $141;$n$2 = $128;$src$3 = $142;
     label = 12;
     continue;
    }
   }
   $178 = (($130) + ($6)|0);
   $179 = (($134) + ($10)|0);
   $180 = (($181) + -1)|0;
   $182 = ($181|0)==(0);
   if ($182) {
    label = 20;
    break;
   } else {
    $181 = $180;$dst$01 = $179;$src$02 = $178;
   }
  }
  if ((label|0) == 18) {
   // unreachable;
  }
  else if ((label|0) == 20) {
   return;
  }
 }
 }
}
function _Blit1to2($info) {
 $info = $info|0;
 var $$lcssa = 0, $$lcssa14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$01 = 0, $dst$01$us = 0, $dst$1$us = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0;
 var $dst$8 = 0, $n$0$us = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $src$02 = 0, $src$02$us = 0, $src$1$us = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$info>>2]|0;
 $5 = ((($info)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($info)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($info)) + 36|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($info)) + 48|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($3) + -1)|0;
 $14 = ($3|0)==(0);
 if ($14) {
  return;
 }
 $15 = (($1) + 7)|0;
 $16 = (($15|0) / 8)&-1;
 $17 = $1 & 7;
 $18 = ($17|0)==(0);
 if ($18) {
  $72 = $13;$dst$01$us = $8;$src$02$us = $4;
  while(1) {
   $dst$1$us = $dst$01$us;$n$0$us = $16;$src$1$us = $src$02$us;
   while(1) {
    $19 = ((($src$1$us)) + 1|0);
    $20 = HEAP8[$src$1$us>>0]|0;
    $21 = $20&255;
    $22 = (($12) + ($21<<1)|0);
    $23 = HEAP16[$22>>1]|0;
    HEAP16[$dst$1$us>>1] = $23;
    $24 = ((($dst$1$us)) + 2|0);
    $25 = ((($src$1$us)) + 2|0);
    $26 = HEAP8[$19>>0]|0;
    $27 = $26&255;
    $28 = (($12) + ($27<<1)|0);
    $29 = HEAP16[$28>>1]|0;
    HEAP16[$24>>1] = $29;
    $30 = ((($dst$1$us)) + 4|0);
    $31 = ((($src$1$us)) + 3|0);
    $32 = HEAP8[$25>>0]|0;
    $33 = $32&255;
    $34 = (($12) + ($33<<1)|0);
    $35 = HEAP16[$34>>1]|0;
    HEAP16[$30>>1] = $35;
    $36 = ((($dst$1$us)) + 6|0);
    $37 = ((($src$1$us)) + 4|0);
    $38 = HEAP8[$31>>0]|0;
    $39 = $38&255;
    $40 = (($12) + ($39<<1)|0);
    $41 = HEAP16[$40>>1]|0;
    HEAP16[$36>>1] = $41;
    $42 = ((($dst$1$us)) + 8|0);
    $43 = ((($src$1$us)) + 5|0);
    $44 = HEAP8[$37>>0]|0;
    $45 = $44&255;
    $46 = (($12) + ($45<<1)|0);
    $47 = HEAP16[$46>>1]|0;
    HEAP16[$42>>1] = $47;
    $48 = ((($dst$1$us)) + 10|0);
    $49 = ((($src$1$us)) + 6|0);
    $50 = HEAP8[$43>>0]|0;
    $51 = $50&255;
    $52 = (($12) + ($51<<1)|0);
    $53 = HEAP16[$52>>1]|0;
    HEAP16[$48>>1] = $53;
    $54 = ((($dst$1$us)) + 12|0);
    $55 = ((($src$1$us)) + 7|0);
    $56 = HEAP8[$49>>0]|0;
    $57 = $56&255;
    $58 = (($12) + ($57<<1)|0);
    $59 = HEAP16[$58>>1]|0;
    HEAP16[$54>>1] = $59;
    $60 = ((($dst$1$us)) + 14|0);
    $61 = ((($src$1$us)) + 8|0);
    $62 = HEAP8[$55>>0]|0;
    $63 = $62&255;
    $64 = (($12) + ($63<<1)|0);
    $65 = HEAP16[$64>>1]|0;
    HEAP16[$60>>1] = $65;
    $66 = ((($dst$1$us)) + 16|0);
    $67 = (($n$0$us) + -1)|0;
    $68 = ($n$0$us|0)>(1);
    if ($68) {
     $dst$1$us = $66;$n$0$us = $67;$src$1$us = $61;
    } else {
     $$lcssa = $61;$$lcssa14 = $66;
     break;
    }
   }
   $69 = (($$lcssa) + ($6)|0);
   $70 = (($$lcssa14) + ($10)|0);
   $71 = (($72) + -1)|0;
   $73 = ($72|0)==(0);
   if ($73) {
    break;
   } else {
    $72 = $71;$dst$01$us = $70;$src$02$us = $69;
   }
  }
  return;
 } else {
  $127 = $13;$dst$01 = $8;$src$02 = $4;
 }
 L11: while(1) {
  switch ($17|0) {
  case 1:  {
   $dst$8 = $dst$01;$n$7 = $16;$src$8 = $src$02;
   label = 14;
   break;
  }
  case 7:  {
   $dst$2 = $dst$01;$n$1 = $16;$src$2 = $src$02;
   label = 8;
   break;
  }
  case 6:  {
   $dst$3 = $dst$01;$n$2 = $16;$src$3 = $src$02;
   label = 9;
   break;
  }
  case 5:  {
   $dst$4 = $dst$01;$n$3 = $16;$src$4 = $src$02;
   label = 10;
   break;
  }
  case 4:  {
   $dst$5 = $dst$01;$n$4 = $16;$src$5 = $src$02;
   label = 11;
   break;
  }
  case 3:  {
   $dst$6 = $dst$01;$n$5 = $16;$src$6 = $src$02;
   label = 12;
   break;
  }
  case 2:  {
   $dst$7 = $dst$01;$n$6 = $16;$src$7 = $src$02;
   label = 13;
   break;
  }
  default: {
   label = 15;
   break L11;
  }
  }
  while(1) {
   if ((label|0) == 8) {
    label = 0;
    $83 = ((($src$2)) + 1|0);
    $84 = HEAP8[$src$2>>0]|0;
    $85 = $84&255;
    $86 = (($12) + ($85<<1)|0);
    $87 = HEAP16[$86>>1]|0;
    HEAP16[$dst$2>>1] = $87;
    $88 = ((($dst$2)) + 2|0);
    $dst$3 = $88;$n$2 = $n$1;$src$3 = $83;
    label = 9;
    continue;
   }
   else if ((label|0) == 9) {
    label = 0;
    $89 = ((($src$3)) + 1|0);
    $90 = HEAP8[$src$3>>0]|0;
    $91 = $90&255;
    $92 = (($12) + ($91<<1)|0);
    $93 = HEAP16[$92>>1]|0;
    HEAP16[$dst$3>>1] = $93;
    $94 = ((($dst$3)) + 2|0);
    $dst$4 = $94;$n$3 = $n$2;$src$4 = $89;
    label = 10;
    continue;
   }
   else if ((label|0) == 10) {
    label = 0;
    $95 = ((($src$4)) + 1|0);
    $96 = HEAP8[$src$4>>0]|0;
    $97 = $96&255;
    $98 = (($12) + ($97<<1)|0);
    $99 = HEAP16[$98>>1]|0;
    HEAP16[$dst$4>>1] = $99;
    $100 = ((($dst$4)) + 2|0);
    $dst$5 = $100;$n$4 = $n$3;$src$5 = $95;
    label = 11;
    continue;
   }
   else if ((label|0) == 11) {
    label = 0;
    $101 = ((($src$5)) + 1|0);
    $102 = HEAP8[$src$5>>0]|0;
    $103 = $102&255;
    $104 = (($12) + ($103<<1)|0);
    $105 = HEAP16[$104>>1]|0;
    HEAP16[$dst$5>>1] = $105;
    $106 = ((($dst$5)) + 2|0);
    $dst$6 = $106;$n$5 = $n$4;$src$6 = $101;
    label = 12;
    continue;
   }
   else if ((label|0) == 12) {
    label = 0;
    $107 = ((($src$6)) + 1|0);
    $108 = HEAP8[$src$6>>0]|0;
    $109 = $108&255;
    $110 = (($12) + ($109<<1)|0);
    $111 = HEAP16[$110>>1]|0;
    HEAP16[$dst$6>>1] = $111;
    $112 = ((($dst$6)) + 2|0);
    $dst$7 = $112;$n$6 = $n$5;$src$7 = $107;
    label = 13;
    continue;
   }
   else if ((label|0) == 13) {
    label = 0;
    $113 = ((($src$7)) + 1|0);
    $114 = HEAP8[$src$7>>0]|0;
    $115 = $114&255;
    $116 = (($12) + ($115<<1)|0);
    $117 = HEAP16[$116>>1]|0;
    HEAP16[$dst$7>>1] = $117;
    $118 = ((($dst$7)) + 2|0);
    $dst$8 = $118;$n$7 = $n$6;$src$8 = $113;
    label = 14;
    continue;
   }
   else if ((label|0) == 14) {
    label = 0;
    $77 = ((($src$8)) + 1|0);
    $119 = HEAP8[$src$8>>0]|0;
    $120 = $119&255;
    $121 = (($12) + ($120<<1)|0);
    $122 = HEAP16[$121>>1]|0;
    HEAP16[$dst$8>>1] = $122;
    $81 = ((($dst$8)) + 2|0);
    $123 = ($n$7|0)>(1);
    if (!($123)) {
     break;
    }
    $74 = (($n$7) + -1)|0;
    $75 = ((($src$8)) + 2|0);
    $76 = HEAP8[$77>>0]|0;
    $78 = $76&255;
    $79 = (($12) + ($78<<1)|0);
    $80 = HEAP16[$79>>1]|0;
    HEAP16[$81>>1] = $80;
    $82 = ((($dst$8)) + 4|0);
    $dst$2 = $82;$n$1 = $74;$src$2 = $75;
    label = 8;
    continue;
   }
  }
  $124 = (($77) + ($6)|0);
  $125 = (($81) + ($10)|0);
  $126 = (($127) + -1)|0;
  $128 = ($127|0)==(0);
  if ($128) {
   label = 17;
   break;
  } else {
   $127 = $126;$dst$01 = $125;$src$02 = $124;
  }
 }
 if ((label|0) == 15) {
  // unreachable;
 }
 else if ((label|0) == 17) {
  return;
 }
}
function _Blit1to3($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$02 = 0, $dst$1 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0;
 var $n$7 = 0, $src$01 = 0, $src$1 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 48|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($3|0)==(0);
 if ($10) {
  return;
 }
 $11 = ((($info)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = HEAP32[$info>>2]|0;
 $14 = (($1) + 7)|0;
 $15 = (($14|0) / 8)&-1;
 $16 = $1 & 7;
 $$in = $3;$dst$02 = $12;$src$01 = $13;
 L4: while(1) {
  $17 = (($$in) + -1)|0;
  switch ($16|0) {
  case 0:  {
   $dst$1 = $dst$02;$n$0 = $15;$src$1 = $src$01;
   label = 4;
   break;
  }
  case 7:  {
   $dst$2 = $dst$02;$n$1 = $15;$src$2 = $src$01;
   label = 5;
   break;
  }
  case 6:  {
   $dst$3 = $dst$02;$n$2 = $15;$src$3 = $src$01;
   label = 6;
   break;
  }
  case 5:  {
   $dst$4 = $dst$02;$n$3 = $15;$src$4 = $src$01;
   label = 7;
   break;
  }
  case 4:  {
   $dst$5 = $dst$02;$n$4 = $15;$src$5 = $src$01;
   label = 8;
   break;
  }
  case 3:  {
   $dst$6 = $dst$02;$n$5 = $15;$src$6 = $src$01;
   label = 9;
   break;
  }
  case 2:  {
   $dst$7 = $dst$02;$n$6 = $15;$src$7 = $src$01;
   label = 10;
   break;
  }
  case 1:  {
   $dst$8 = $dst$02;$n$7 = $15;$src$8 = $src$01;
   label = 11;
   break;
  }
  default: {
   label = 12;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    $18 = HEAP8[$src$1>>0]|0;
    $19 = $18&255;
    $20 = $19 << 2;
    $21 = $20 | 1;
    $22 = (($9) + ($20)|0);
    $23 = HEAP8[$22>>0]|0;
    HEAP8[$dst$1>>0] = $23;
    $24 = (($21) + 1)|0;
    $25 = (($9) + ($21)|0);
    $26 = HEAP8[$25>>0]|0;
    $27 = ((($dst$1)) + 1|0);
    HEAP8[$27>>0] = $26;
    $28 = (($9) + ($24)|0);
    $29 = HEAP8[$28>>0]|0;
    $30 = ((($dst$1)) + 2|0);
    HEAP8[$30>>0] = $29;
    $31 = ((($src$1)) + 1|0);
    $32 = ((($dst$1)) + 3|0);
    $dst$2 = $32;$n$1 = $n$0;$src$2 = $31;
    label = 5;
    continue;
   }
   else if ((label|0) == 5) {
    label = 0;
    $33 = HEAP8[$src$2>>0]|0;
    $34 = $33&255;
    $35 = $34 << 2;
    $36 = $35 | 1;
    $37 = (($9) + ($35)|0);
    $38 = HEAP8[$37>>0]|0;
    HEAP8[$dst$2>>0] = $38;
    $39 = (($36) + 1)|0;
    $40 = (($9) + ($36)|0);
    $41 = HEAP8[$40>>0]|0;
    $42 = ((($dst$2)) + 1|0);
    HEAP8[$42>>0] = $41;
    $43 = (($9) + ($39)|0);
    $44 = HEAP8[$43>>0]|0;
    $45 = ((($dst$2)) + 2|0);
    HEAP8[$45>>0] = $44;
    $46 = ((($src$2)) + 1|0);
    $47 = ((($dst$2)) + 3|0);
    $dst$3 = $47;$n$2 = $n$1;$src$3 = $46;
    label = 6;
    continue;
   }
   else if ((label|0) == 6) {
    label = 0;
    $48 = HEAP8[$src$3>>0]|0;
    $49 = $48&255;
    $50 = $49 << 2;
    $51 = $50 | 1;
    $52 = (($9) + ($50)|0);
    $53 = HEAP8[$52>>0]|0;
    HEAP8[$dst$3>>0] = $53;
    $54 = (($51) + 1)|0;
    $55 = (($9) + ($51)|0);
    $56 = HEAP8[$55>>0]|0;
    $57 = ((($dst$3)) + 1|0);
    HEAP8[$57>>0] = $56;
    $58 = (($9) + ($54)|0);
    $59 = HEAP8[$58>>0]|0;
    $60 = ((($dst$3)) + 2|0);
    HEAP8[$60>>0] = $59;
    $61 = ((($src$3)) + 1|0);
    $62 = ((($dst$3)) + 3|0);
    $dst$4 = $62;$n$3 = $n$2;$src$4 = $61;
    label = 7;
    continue;
   }
   else if ((label|0) == 7) {
    label = 0;
    $63 = HEAP8[$src$4>>0]|0;
    $64 = $63&255;
    $65 = $64 << 2;
    $66 = $65 | 1;
    $67 = (($9) + ($65)|0);
    $68 = HEAP8[$67>>0]|0;
    HEAP8[$dst$4>>0] = $68;
    $69 = (($66) + 1)|0;
    $70 = (($9) + ($66)|0);
    $71 = HEAP8[$70>>0]|0;
    $72 = ((($dst$4)) + 1|0);
    HEAP8[$72>>0] = $71;
    $73 = (($9) + ($69)|0);
    $74 = HEAP8[$73>>0]|0;
    $75 = ((($dst$4)) + 2|0);
    HEAP8[$75>>0] = $74;
    $76 = ((($src$4)) + 1|0);
    $77 = ((($dst$4)) + 3|0);
    $dst$5 = $77;$n$4 = $n$3;$src$5 = $76;
    label = 8;
    continue;
   }
   else if ((label|0) == 8) {
    label = 0;
    $78 = HEAP8[$src$5>>0]|0;
    $79 = $78&255;
    $80 = $79 << 2;
    $81 = $80 | 1;
    $82 = (($9) + ($80)|0);
    $83 = HEAP8[$82>>0]|0;
    HEAP8[$dst$5>>0] = $83;
    $84 = (($81) + 1)|0;
    $85 = (($9) + ($81)|0);
    $86 = HEAP8[$85>>0]|0;
    $87 = ((($dst$5)) + 1|0);
    HEAP8[$87>>0] = $86;
    $88 = (($9) + ($84)|0);
    $89 = HEAP8[$88>>0]|0;
    $90 = ((($dst$5)) + 2|0);
    HEAP8[$90>>0] = $89;
    $91 = ((($src$5)) + 1|0);
    $92 = ((($dst$5)) + 3|0);
    $dst$6 = $92;$n$5 = $n$4;$src$6 = $91;
    label = 9;
    continue;
   }
   else if ((label|0) == 9) {
    label = 0;
    $93 = HEAP8[$src$6>>0]|0;
    $94 = $93&255;
    $95 = $94 << 2;
    $96 = $95 | 1;
    $97 = (($9) + ($95)|0);
    $98 = HEAP8[$97>>0]|0;
    HEAP8[$dst$6>>0] = $98;
    $99 = (($96) + 1)|0;
    $100 = (($9) + ($96)|0);
    $101 = HEAP8[$100>>0]|0;
    $102 = ((($dst$6)) + 1|0);
    HEAP8[$102>>0] = $101;
    $103 = (($9) + ($99)|0);
    $104 = HEAP8[$103>>0]|0;
    $105 = ((($dst$6)) + 2|0);
    HEAP8[$105>>0] = $104;
    $106 = ((($src$6)) + 1|0);
    $107 = ((($dst$6)) + 3|0);
    $dst$7 = $107;$n$6 = $n$5;$src$7 = $106;
    label = 10;
    continue;
   }
   else if ((label|0) == 10) {
    label = 0;
    $108 = HEAP8[$src$7>>0]|0;
    $109 = $108&255;
    $110 = $109 << 2;
    $111 = $110 | 1;
    $112 = (($9) + ($110)|0);
    $113 = HEAP8[$112>>0]|0;
    HEAP8[$dst$7>>0] = $113;
    $114 = (($111) + 1)|0;
    $115 = (($9) + ($111)|0);
    $116 = HEAP8[$115>>0]|0;
    $117 = ((($dst$7)) + 1|0);
    HEAP8[$117>>0] = $116;
    $118 = (($9) + ($114)|0);
    $119 = HEAP8[$118>>0]|0;
    $120 = ((($dst$7)) + 2|0);
    HEAP8[$120>>0] = $119;
    $121 = ((($src$7)) + 1|0);
    $122 = ((($dst$7)) + 3|0);
    $dst$8 = $122;$n$7 = $n$6;$src$8 = $121;
    label = 11;
    continue;
   }
   else if ((label|0) == 11) {
    label = 0;
    $123 = HEAP8[$src$8>>0]|0;
    $124 = $123&255;
    $125 = $124 << 2;
    $126 = $125 | 1;
    $127 = (($9) + ($125)|0);
    $128 = HEAP8[$127>>0]|0;
    HEAP8[$dst$8>>0] = $128;
    $129 = (($126) + 1)|0;
    $130 = (($9) + ($126)|0);
    $131 = HEAP8[$130>>0]|0;
    $132 = ((($dst$8)) + 1|0);
    HEAP8[$132>>0] = $131;
    $133 = (($9) + ($129)|0);
    $134 = HEAP8[$133>>0]|0;
    $135 = ((($dst$8)) + 2|0);
    HEAP8[$135>>0] = $134;
    $136 = ((($src$8)) + 1|0);
    $137 = ((($dst$8)) + 3|0);
    $138 = (($n$7) + -1)|0;
    $139 = ($n$7|0)>(1);
    if ($139) {
     $dst$1 = $137;$n$0 = $138;$src$1 = $136;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $140 = (($136) + ($5)|0);
  $141 = (($137) + ($7)|0);
  $142 = ($17|0)==(0);
  if ($142) {
   label = 14;
   break;
  } else {
   $$in = $17;$dst$02 = $141;$src$01 = $140;
  }
 }
 if ((label|0) == 12) {
  // unreachable;
 }
 else if ((label|0) == 14) {
  return;
 }
}
function _Blit1to4($info) {
 $info = $info|0;
 var $$lcssa = 0, $$lcssa14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$01 = 0, $dst$01$us = 0, $dst$1$us = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0;
 var $dst$7 = 0, $dst$8 = 0, $n$0$us = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $src$02 = 0, $src$02$us = 0, $src$1$us = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$info>>2]|0;
 $5 = ((($info)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($info)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($info)) + 36|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10|0) / 4)&-1;
 $12 = ((($info)) + 48|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($3) + -1)|0;
 $15 = ($3|0)==(0);
 if ($15) {
  return;
 }
 $16 = (($1) + 7)|0;
 $17 = (($16|0) / 8)&-1;
 $18 = $1 & 7;
 $19 = ($18|0)==(0);
 if ($19) {
  $73 = $14;$dst$01$us = $8;$src$02$us = $4;
  while(1) {
   $dst$1$us = $dst$01$us;$n$0$us = $17;$src$1$us = $src$02$us;
   while(1) {
    $20 = ((($src$1$us)) + 1|0);
    $21 = HEAP8[$src$1$us>>0]|0;
    $22 = $21&255;
    $23 = (($13) + ($22<<2)|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = ((($dst$1$us)) + 4|0);
    HEAP32[$dst$1$us>>2] = $24;
    $26 = ((($src$1$us)) + 2|0);
    $27 = HEAP8[$20>>0]|0;
    $28 = $27&255;
    $29 = (($13) + ($28<<2)|0);
    $30 = HEAP32[$29>>2]|0;
    $31 = ((($dst$1$us)) + 8|0);
    HEAP32[$25>>2] = $30;
    $32 = ((($src$1$us)) + 3|0);
    $33 = HEAP8[$26>>0]|0;
    $34 = $33&255;
    $35 = (($13) + ($34<<2)|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ((($dst$1$us)) + 12|0);
    HEAP32[$31>>2] = $36;
    $38 = ((($src$1$us)) + 4|0);
    $39 = HEAP8[$32>>0]|0;
    $40 = $39&255;
    $41 = (($13) + ($40<<2)|0);
    $42 = HEAP32[$41>>2]|0;
    $43 = ((($dst$1$us)) + 16|0);
    HEAP32[$37>>2] = $42;
    $44 = ((($src$1$us)) + 5|0);
    $45 = HEAP8[$38>>0]|0;
    $46 = $45&255;
    $47 = (($13) + ($46<<2)|0);
    $48 = HEAP32[$47>>2]|0;
    $49 = ((($dst$1$us)) + 20|0);
    HEAP32[$43>>2] = $48;
    $50 = ((($src$1$us)) + 6|0);
    $51 = HEAP8[$44>>0]|0;
    $52 = $51&255;
    $53 = (($13) + ($52<<2)|0);
    $54 = HEAP32[$53>>2]|0;
    $55 = ((($dst$1$us)) + 24|0);
    HEAP32[$49>>2] = $54;
    $56 = ((($src$1$us)) + 7|0);
    $57 = HEAP8[$50>>0]|0;
    $58 = $57&255;
    $59 = (($13) + ($58<<2)|0);
    $60 = HEAP32[$59>>2]|0;
    $61 = ((($dst$1$us)) + 28|0);
    HEAP32[$55>>2] = $60;
    $62 = ((($src$1$us)) + 8|0);
    $63 = HEAP8[$56>>0]|0;
    $64 = $63&255;
    $65 = (($13) + ($64<<2)|0);
    $66 = HEAP32[$65>>2]|0;
    $67 = ((($dst$1$us)) + 32|0);
    HEAP32[$61>>2] = $66;
    $68 = (($n$0$us) + -1)|0;
    $69 = ($n$0$us|0)>(1);
    if ($69) {
     $dst$1$us = $67;$n$0$us = $68;$src$1$us = $62;
    } else {
     $$lcssa = $62;$$lcssa14 = $67;
     break;
    }
   }
   $70 = (($$lcssa) + ($6)|0);
   $71 = (($$lcssa14) + ($11<<2)|0);
   $72 = (($73) + -1)|0;
   $74 = ($73|0)==(0);
   if ($74) {
    break;
   } else {
    $73 = $72;$dst$01$us = $71;$src$02$us = $70;
   }
  }
  return;
 } else {
  $128 = $14;$dst$01 = $8;$src$02 = $4;
 }
 L11: while(1) {
  switch ($18|0) {
  case 1:  {
   $dst$8 = $dst$01;$n$7 = $17;$src$8 = $src$02;
   label = 14;
   break;
  }
  case 7:  {
   $dst$2 = $dst$01;$n$1 = $17;$src$2 = $src$02;
   label = 8;
   break;
  }
  case 6:  {
   $dst$3 = $dst$01;$n$2 = $17;$src$3 = $src$02;
   label = 9;
   break;
  }
  case 5:  {
   $dst$4 = $dst$01;$n$3 = $17;$src$4 = $src$02;
   label = 10;
   break;
  }
  case 4:  {
   $dst$5 = $dst$01;$n$4 = $17;$src$5 = $src$02;
   label = 11;
   break;
  }
  case 3:  {
   $dst$6 = $dst$01;$n$5 = $17;$src$6 = $src$02;
   label = 12;
   break;
  }
  case 2:  {
   $dst$7 = $dst$01;$n$6 = $17;$src$7 = $src$02;
   label = 13;
   break;
  }
  default: {
   label = 15;
   break L11;
  }
  }
  while(1) {
   if ((label|0) == 8) {
    label = 0;
    $84 = ((($src$2)) + 1|0);
    $85 = HEAP8[$src$2>>0]|0;
    $86 = $85&255;
    $87 = (($13) + ($86<<2)|0);
    $88 = HEAP32[$87>>2]|0;
    $89 = ((($dst$2)) + 4|0);
    HEAP32[$dst$2>>2] = $88;
    $dst$3 = $89;$n$2 = $n$1;$src$3 = $84;
    label = 9;
    continue;
   }
   else if ((label|0) == 9) {
    label = 0;
    $90 = ((($src$3)) + 1|0);
    $91 = HEAP8[$src$3>>0]|0;
    $92 = $91&255;
    $93 = (($13) + ($92<<2)|0);
    $94 = HEAP32[$93>>2]|0;
    $95 = ((($dst$3)) + 4|0);
    HEAP32[$dst$3>>2] = $94;
    $dst$4 = $95;$n$3 = $n$2;$src$4 = $90;
    label = 10;
    continue;
   }
   else if ((label|0) == 10) {
    label = 0;
    $96 = ((($src$4)) + 1|0);
    $97 = HEAP8[$src$4>>0]|0;
    $98 = $97&255;
    $99 = (($13) + ($98<<2)|0);
    $100 = HEAP32[$99>>2]|0;
    $101 = ((($dst$4)) + 4|0);
    HEAP32[$dst$4>>2] = $100;
    $dst$5 = $101;$n$4 = $n$3;$src$5 = $96;
    label = 11;
    continue;
   }
   else if ((label|0) == 11) {
    label = 0;
    $102 = ((($src$5)) + 1|0);
    $103 = HEAP8[$src$5>>0]|0;
    $104 = $103&255;
    $105 = (($13) + ($104<<2)|0);
    $106 = HEAP32[$105>>2]|0;
    $107 = ((($dst$5)) + 4|0);
    HEAP32[$dst$5>>2] = $106;
    $dst$6 = $107;$n$5 = $n$4;$src$6 = $102;
    label = 12;
    continue;
   }
   else if ((label|0) == 12) {
    label = 0;
    $108 = ((($src$6)) + 1|0);
    $109 = HEAP8[$src$6>>0]|0;
    $110 = $109&255;
    $111 = (($13) + ($110<<2)|0);
    $112 = HEAP32[$111>>2]|0;
    $113 = ((($dst$6)) + 4|0);
    HEAP32[$dst$6>>2] = $112;
    $dst$7 = $113;$n$6 = $n$5;$src$7 = $108;
    label = 13;
    continue;
   }
   else if ((label|0) == 13) {
    label = 0;
    $114 = ((($src$7)) + 1|0);
    $115 = HEAP8[$src$7>>0]|0;
    $116 = $115&255;
    $117 = (($13) + ($116<<2)|0);
    $118 = HEAP32[$117>>2]|0;
    $119 = ((($dst$7)) + 4|0);
    HEAP32[$dst$7>>2] = $118;
    $dst$8 = $119;$n$7 = $n$6;$src$8 = $114;
    label = 14;
    continue;
   }
   else if ((label|0) == 14) {
    label = 0;
    $78 = ((($src$8)) + 1|0);
    $120 = HEAP8[$src$8>>0]|0;
    $121 = $120&255;
    $122 = (($13) + ($121<<2)|0);
    $123 = HEAP32[$122>>2]|0;
    $83 = ((($dst$8)) + 4|0);
    HEAP32[$dst$8>>2] = $123;
    $124 = ($n$7|0)>(1);
    if (!($124)) {
     break;
    }
    $75 = (($n$7) + -1)|0;
    $76 = ((($src$8)) + 2|0);
    $77 = HEAP8[$78>>0]|0;
    $79 = $77&255;
    $80 = (($13) + ($79<<2)|0);
    $81 = HEAP32[$80>>2]|0;
    $82 = ((($dst$8)) + 8|0);
    HEAP32[$83>>2] = $81;
    $dst$2 = $82;$n$1 = $75;$src$2 = $76;
    label = 8;
    continue;
   }
  }
  $125 = (($78) + ($6)|0);
  $126 = (($83) + ($11<<2)|0);
  $127 = (($128) + -1)|0;
  $129 = ($128|0)==(0);
  if ($129) {
   label = 17;
   break;
  } else {
   $128 = $127;$dst$01 = $126;$src$02 = $125;
  }
 }
 if ((label|0) == 15) {
  // unreachable;
 }
 else if ((label|0) == 17) {
  return;
 }
}
function _Blit1to1Key($info) {
 $info = $info|0;
 var $$lcssa = 0, $$lcssa39 = 0, $$lcssa40 = 0, $$lcssa41 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$05 = 0, $dst$05$us = 0, $dst$1$us = 0, $dst$10$us = 0, $dst$11 = 0, $dst$12 = 0, $dst$13 = 0, $dst$14 = 0, $dst$15 = 0, $dst$16 = 0;
 var $dst$17 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $dst$93 = 0, $dst$93$us = 0, $n$0$us = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $n1$0$us = 0, $n1$1 = 0;
 var $n1$2 = 0, $n1$3 = 0, $n1$4 = 0, $n1$5 = 0, $n1$6 = 0, $n1$7 = 0, $src$06 = 0, $src$06$us = 0, $src$1$us = 0, $src$10$us = 0, $src$11 = 0, $src$12 = 0, $src$13 = 0, $src$14 = 0, $src$15 = 0, $src$16 = 0, $src$17 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0;
 var $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, $src$94 = 0, $src$94$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$info>>2]|0;
 $5 = ((($info)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($info)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($info)) + 36|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($info)) + 48|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($info)) + 56|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ($12|0)==(0|0);
 $16 = (($3) + -1)|0;
 $17 = ($3|0)==(0);
 if ($15) {
  if ($17) {
   return;
  }
  $85 = (($1) + 7)|0;
  $86 = (($85|0) / 8)&-1;
  $87 = $1 & 7;
  $88 = ($87|0)==(0);
  if ($88) {
   $134 = $16;$dst$93$us = $8;$src$94$us = $4;
   while(1) {
    $dst$10$us = $dst$93$us;$n1$0$us = $86;$src$10$us = $src$94$us;
    while(1) {
     $89 = HEAP8[$src$10$us>>0]|0;
     $90 = $89&255;
     $91 = ($90|0)==($14|0);
     if (!($91)) {
      HEAP8[$dst$10$us>>0] = $89;
     }
     $92 = ((($src$10$us)) + 1|0);
     $93 = HEAP8[$92>>0]|0;
     $94 = $93&255;
     $95 = ($94|0)==($14|0);
     if (!($95)) {
      $96 = ((($dst$10$us)) + 1|0);
      HEAP8[$96>>0] = $93;
     }
     $97 = ((($src$10$us)) + 2|0);
     $98 = HEAP8[$97>>0]|0;
     $99 = $98&255;
     $100 = ($99|0)==($14|0);
     if (!($100)) {
      $101 = ((($dst$10$us)) + 2|0);
      HEAP8[$101>>0] = $98;
     }
     $102 = ((($src$10$us)) + 3|0);
     $103 = HEAP8[$102>>0]|0;
     $104 = $103&255;
     $105 = ($104|0)==($14|0);
     if (!($105)) {
      $106 = ((($dst$10$us)) + 3|0);
      HEAP8[$106>>0] = $103;
     }
     $107 = ((($src$10$us)) + 4|0);
     $108 = HEAP8[$107>>0]|0;
     $109 = $108&255;
     $110 = ($109|0)==($14|0);
     if (!($110)) {
      $111 = ((($dst$10$us)) + 4|0);
      HEAP8[$111>>0] = $108;
     }
     $112 = ((($src$10$us)) + 5|0);
     $113 = HEAP8[$112>>0]|0;
     $114 = $113&255;
     $115 = ($114|0)==($14|0);
     if (!($115)) {
      $116 = ((($dst$10$us)) + 5|0);
      HEAP8[$116>>0] = $113;
     }
     $117 = ((($src$10$us)) + 6|0);
     $118 = HEAP8[$117>>0]|0;
     $119 = $118&255;
     $120 = ($119|0)==($14|0);
     if (!($120)) {
      $121 = ((($dst$10$us)) + 6|0);
      HEAP8[$121>>0] = $118;
     }
     $122 = ((($src$10$us)) + 7|0);
     $123 = HEAP8[$122>>0]|0;
     $124 = $123&255;
     $125 = ($124|0)==($14|0);
     if (!($125)) {
      $126 = ((($dst$10$us)) + 7|0);
      HEAP8[$126>>0] = $123;
     }
     $127 = ((($dst$10$us)) + 8|0);
     $128 = ((($src$10$us)) + 8|0);
     $129 = (($n1$0$us) + -1)|0;
     $130 = ($n1$0$us|0)>(1);
     if ($130) {
      $dst$10$us = $127;$n1$0$us = $129;$src$10$us = $128;
     } else {
      $$lcssa = $127;$$lcssa39 = $128;
      break;
     }
    }
    $131 = (($$lcssa39) + ($6)|0);
    $132 = (($$lcssa) + ($10)|0);
    $133 = (($134) + -1)|0;
    $135 = ($134|0)==(0);
    if ($135) {
     break;
    } else {
     $134 = $133;$dst$93$us = $132;$src$94$us = $131;
    }
   }
   return;
  } else {
   $244 = $16;$dst$93 = $8;$src$94 = $4;
  }
  L37: while(1) {
   switch ($87|0) {
   case 1:  {
    $dst$17 = $dst$93;$n1$7 = $86;$src$17 = $src$94;
    label = 93;
    break;
   }
   case 7:  {
    $dst$11 = $dst$93;$n1$1 = $86;$src$11 = $src$94;
    label = 75;
    break;
   }
   case 6:  {
    $dst$12 = $dst$93;$n1$2 = $86;$src$12 = $src$94;
    label = 78;
    break;
   }
   case 5:  {
    $dst$13 = $dst$93;$n1$3 = $86;$src$13 = $src$94;
    label = 81;
    break;
   }
   case 4:  {
    $dst$14 = $dst$93;$n1$4 = $86;$src$14 = $src$94;
    label = 84;
    break;
   }
   case 3:  {
    $dst$15 = $dst$93;$n1$5 = $86;$src$15 = $src$94;
    label = 87;
    break;
   }
   case 2:  {
    $dst$16 = $dst$93;$n1$6 = $86;$src$16 = $src$94;
    label = 90;
    break;
   }
   default: {
    label = 96;
    break L37;
   }
   }
   while(1) {
    if ((label|0) == 75) {
     label = 0;
     $206 = HEAP8[$src$11>>0]|0;
     $207 = $206&255;
     $208 = ($207|0)==($14|0);
     if (!($208)) {
      HEAP8[$dst$11>>0] = $206;
     }
     $209 = ((($dst$11)) + 1|0);
     $210 = ((($src$11)) + 1|0);
     $dst$12 = $209;$n1$2 = $n1$1;$src$12 = $210;
     label = 78;
     continue;
    }
    else if ((label|0) == 78) {
     label = 0;
     $211 = HEAP8[$src$12>>0]|0;
     $212 = $211&255;
     $213 = ($212|0)==($14|0);
     if (!($213)) {
      HEAP8[$dst$12>>0] = $211;
     }
     $214 = ((($dst$12)) + 1|0);
     $215 = ((($src$12)) + 1|0);
     $dst$13 = $214;$n1$3 = $n1$2;$src$13 = $215;
     label = 81;
     continue;
    }
    else if ((label|0) == 81) {
     label = 0;
     $216 = HEAP8[$src$13>>0]|0;
     $217 = $216&255;
     $218 = ($217|0)==($14|0);
     if (!($218)) {
      HEAP8[$dst$13>>0] = $216;
     }
     $219 = ((($dst$13)) + 1|0);
     $220 = ((($src$13)) + 1|0);
     $dst$14 = $219;$n1$4 = $n1$3;$src$14 = $220;
     label = 84;
     continue;
    }
    else if ((label|0) == 84) {
     label = 0;
     $221 = HEAP8[$src$14>>0]|0;
     $222 = $221&255;
     $223 = ($222|0)==($14|0);
     if (!($223)) {
      HEAP8[$dst$14>>0] = $221;
     }
     $224 = ((($dst$14)) + 1|0);
     $225 = ((($src$14)) + 1|0);
     $dst$15 = $224;$n1$5 = $n1$4;$src$15 = $225;
     label = 87;
     continue;
    }
    else if ((label|0) == 87) {
     label = 0;
     $226 = HEAP8[$src$15>>0]|0;
     $227 = $226&255;
     $228 = ($227|0)==($14|0);
     if (!($228)) {
      HEAP8[$dst$15>>0] = $226;
     }
     $229 = ((($dst$15)) + 1|0);
     $230 = ((($src$15)) + 1|0);
     $dst$16 = $229;$n1$6 = $n1$5;$src$16 = $230;
     label = 90;
     continue;
    }
    else if ((label|0) == 90) {
     label = 0;
     $231 = HEAP8[$src$16>>0]|0;
     $232 = $231&255;
     $233 = ($232|0)==($14|0);
     if (!($233)) {
      HEAP8[$dst$16>>0] = $231;
     }
     $234 = ((($dst$16)) + 1|0);
     $235 = ((($src$16)) + 1|0);
     $dst$17 = $234;$n1$7 = $n1$6;$src$17 = $235;
     label = 93;
     continue;
    }
    else if ((label|0) == 93) {
     label = 0;
     $236 = HEAP8[$src$17>>0]|0;
     $237 = $236&255;
     $238 = ($237|0)==($14|0);
     if (!($238)) {
      HEAP8[$dst$17>>0] = $236;
     }
     $203 = ((($dst$17)) + 1|0);
     $200 = ((($src$17)) + 1|0);
     $239 = (($n1$7) + -1)|0;
     $240 = ($n1$7|0)>(1);
     if (!($240)) {
      break;
     }
     $199 = HEAP8[$200>>0]|0;
     $201 = $199&255;
     $202 = ($201|0)==($14|0);
     if (!($202)) {
      HEAP8[$203>>0] = $199;
     }
     $204 = ((($dst$17)) + 2|0);
     $205 = ((($src$17)) + 2|0);
     $dst$11 = $204;$n1$1 = $239;$src$11 = $205;
     label = 75;
     continue;
    }
   }
   $241 = (($200) + ($6)|0);
   $242 = (($203) + ($10)|0);
   $243 = (($244) + -1)|0;
   $245 = ($244|0)==(0);
   if ($245) {
    label = 98;
    break;
   } else {
    $244 = $243;$dst$93 = $242;$src$94 = $241;
   }
  }
  if ((label|0) == 96) {
   // unreachable;
  }
  else if ((label|0) == 98) {
   return;
  }
 } else {
  if ($17) {
   return;
  }
  $18 = (($1) + 7)|0;
  $19 = (($18|0) / 8)&-1;
  $20 = $1 & 7;
  $21 = ($20|0)==(0);
  if ($21) {
   $83 = $16;$dst$05$us = $8;$src$06$us = $4;
   while(1) {
    $dst$1$us = $dst$05$us;$n$0$us = $19;$src$1$us = $src$06$us;
    while(1) {
     $22 = HEAP8[$src$1$us>>0]|0;
     $23 = $22&255;
     $24 = ($23|0)==($14|0);
     if (!($24)) {
      $25 = (($12) + ($23)|0);
      $26 = HEAP8[$25>>0]|0;
      HEAP8[$dst$1$us>>0] = $26;
     }
     $27 = ((($src$1$us)) + 1|0);
     $28 = HEAP8[$27>>0]|0;
     $29 = $28&255;
     $30 = ($29|0)==($14|0);
     if (!($30)) {
      $31 = ((($dst$1$us)) + 1|0);
      $32 = (($12) + ($29)|0);
      $33 = HEAP8[$32>>0]|0;
      HEAP8[$31>>0] = $33;
     }
     $34 = ((($src$1$us)) + 2|0);
     $35 = HEAP8[$34>>0]|0;
     $36 = $35&255;
     $37 = ($36|0)==($14|0);
     if (!($37)) {
      $38 = ((($dst$1$us)) + 2|0);
      $39 = (($12) + ($36)|0);
      $40 = HEAP8[$39>>0]|0;
      HEAP8[$38>>0] = $40;
     }
     $41 = ((($src$1$us)) + 3|0);
     $42 = HEAP8[$41>>0]|0;
     $43 = $42&255;
     $44 = ($43|0)==($14|0);
     if (!($44)) {
      $45 = ((($dst$1$us)) + 3|0);
      $46 = (($12) + ($43)|0);
      $47 = HEAP8[$46>>0]|0;
      HEAP8[$45>>0] = $47;
     }
     $48 = ((($src$1$us)) + 4|0);
     $49 = HEAP8[$48>>0]|0;
     $50 = $49&255;
     $51 = ($50|0)==($14|0);
     if (!($51)) {
      $52 = ((($dst$1$us)) + 4|0);
      $53 = (($12) + ($50)|0);
      $54 = HEAP8[$53>>0]|0;
      HEAP8[$52>>0] = $54;
     }
     $55 = ((($src$1$us)) + 5|0);
     $56 = HEAP8[$55>>0]|0;
     $57 = $56&255;
     $58 = ($57|0)==($14|0);
     if (!($58)) {
      $59 = ((($dst$1$us)) + 5|0);
      $60 = (($12) + ($57)|0);
      $61 = HEAP8[$60>>0]|0;
      HEAP8[$59>>0] = $61;
     }
     $62 = ((($src$1$us)) + 6|0);
     $63 = HEAP8[$62>>0]|0;
     $64 = $63&255;
     $65 = ($64|0)==($14|0);
     if (!($65)) {
      $66 = ((($dst$1$us)) + 6|0);
      $67 = (($12) + ($64)|0);
      $68 = HEAP8[$67>>0]|0;
      HEAP8[$66>>0] = $68;
     }
     $69 = ((($src$1$us)) + 7|0);
     $70 = HEAP8[$69>>0]|0;
     $71 = $70&255;
     $72 = ($71|0)==($14|0);
     if (!($72)) {
      $73 = ((($dst$1$us)) + 7|0);
      $74 = (($12) + ($71)|0);
      $75 = HEAP8[$74>>0]|0;
      HEAP8[$73>>0] = $75;
     }
     $76 = ((($dst$1$us)) + 8|0);
     $77 = ((($src$1$us)) + 8|0);
     $78 = (($n$0$us) + -1)|0;
     $79 = ($n$0$us|0)>(1);
     if ($79) {
      $dst$1$us = $76;$n$0$us = $78;$src$1$us = $77;
     } else {
      $$lcssa40 = $76;$$lcssa41 = $77;
      break;
     }
    }
    $80 = (($$lcssa41) + ($6)|0);
    $81 = (($$lcssa40) + ($10)|0);
    $82 = (($83) + -1)|0;
    $84 = ($83|0)==(0);
    if ($84) {
     break;
    } else {
     $83 = $82;$dst$05$us = $81;$src$06$us = $80;
    }
   }
   return;
  } else {
   $197 = $16;$dst$05 = $8;$src$06 = $4;
  }
  L112: while(1) {
   switch ($20|0) {
   case 1:  {
    $dst$8 = $dst$05;$n$7 = $19;$src$8 = $src$06;
    label = 66;
    break;
   }
   case 7:  {
    $dst$2 = $dst$05;$n$1 = $19;$src$2 = $src$06;
    label = 48;
    break;
   }
   case 6:  {
    $dst$3 = $dst$05;$n$2 = $19;$src$3 = $src$06;
    label = 51;
    break;
   }
   case 5:  {
    $dst$4 = $dst$05;$n$3 = $19;$src$4 = $src$06;
    label = 54;
    break;
   }
   case 4:  {
    $dst$5 = $dst$05;$n$4 = $19;$src$5 = $src$06;
    label = 57;
    break;
   }
   case 3:  {
    $dst$6 = $dst$05;$n$5 = $19;$src$6 = $src$06;
    label = 60;
    break;
   }
   case 2:  {
    $dst$7 = $dst$05;$n$6 = $19;$src$7 = $src$06;
    label = 63;
    break;
   }
   default: {
    label = 69;
    break L112;
   }
   }
   while(1) {
    if ((label|0) == 48) {
     label = 0;
     $145 = HEAP8[$src$2>>0]|0;
     $146 = $145&255;
     $147 = ($146|0)==($14|0);
     if (!($147)) {
      $148 = (($12) + ($146)|0);
      $149 = HEAP8[$148>>0]|0;
      HEAP8[$dst$2>>0] = $149;
     }
     $150 = ((($dst$2)) + 1|0);
     $151 = ((($src$2)) + 1|0);
     $dst$3 = $150;$n$2 = $n$1;$src$3 = $151;
     label = 51;
     continue;
    }
    else if ((label|0) == 51) {
     label = 0;
     $152 = HEAP8[$src$3>>0]|0;
     $153 = $152&255;
     $154 = ($153|0)==($14|0);
     if (!($154)) {
      $155 = (($12) + ($153)|0);
      $156 = HEAP8[$155>>0]|0;
      HEAP8[$dst$3>>0] = $156;
     }
     $157 = ((($dst$3)) + 1|0);
     $158 = ((($src$3)) + 1|0);
     $dst$4 = $157;$n$3 = $n$2;$src$4 = $158;
     label = 54;
     continue;
    }
    else if ((label|0) == 54) {
     label = 0;
     $159 = HEAP8[$src$4>>0]|0;
     $160 = $159&255;
     $161 = ($160|0)==($14|0);
     if (!($161)) {
      $162 = (($12) + ($160)|0);
      $163 = HEAP8[$162>>0]|0;
      HEAP8[$dst$4>>0] = $163;
     }
     $164 = ((($dst$4)) + 1|0);
     $165 = ((($src$4)) + 1|0);
     $dst$5 = $164;$n$4 = $n$3;$src$5 = $165;
     label = 57;
     continue;
    }
    else if ((label|0) == 57) {
     label = 0;
     $166 = HEAP8[$src$5>>0]|0;
     $167 = $166&255;
     $168 = ($167|0)==($14|0);
     if (!($168)) {
      $169 = (($12) + ($167)|0);
      $170 = HEAP8[$169>>0]|0;
      HEAP8[$dst$5>>0] = $170;
     }
     $171 = ((($dst$5)) + 1|0);
     $172 = ((($src$5)) + 1|0);
     $dst$6 = $171;$n$5 = $n$4;$src$6 = $172;
     label = 60;
     continue;
    }
    else if ((label|0) == 60) {
     label = 0;
     $173 = HEAP8[$src$6>>0]|0;
     $174 = $173&255;
     $175 = ($174|0)==($14|0);
     if (!($175)) {
      $176 = (($12) + ($174)|0);
      $177 = HEAP8[$176>>0]|0;
      HEAP8[$dst$6>>0] = $177;
     }
     $178 = ((($dst$6)) + 1|0);
     $179 = ((($src$6)) + 1|0);
     $dst$7 = $178;$n$6 = $n$5;$src$7 = $179;
     label = 63;
     continue;
    }
    else if ((label|0) == 63) {
     label = 0;
     $180 = HEAP8[$src$7>>0]|0;
     $181 = $180&255;
     $182 = ($181|0)==($14|0);
     if (!($182)) {
      $183 = (($12) + ($181)|0);
      $184 = HEAP8[$183>>0]|0;
      HEAP8[$dst$7>>0] = $184;
     }
     $185 = ((($dst$7)) + 1|0);
     $186 = ((($src$7)) + 1|0);
     $dst$8 = $185;$n$7 = $n$6;$src$8 = $186;
     label = 66;
     continue;
    }
    else if ((label|0) == 66) {
     label = 0;
     $187 = HEAP8[$src$8>>0]|0;
     $188 = $187&255;
     $189 = ($188|0)==($14|0);
     if (!($189)) {
      $190 = (($12) + ($188)|0);
      $191 = HEAP8[$190>>0]|0;
      HEAP8[$dst$8>>0] = $191;
     }
     $142 = ((($dst$8)) + 1|0);
     $137 = ((($src$8)) + 1|0);
     $192 = (($n$7) + -1)|0;
     $193 = ($n$7|0)>(1);
     if (!($193)) {
      break;
     }
     $136 = HEAP8[$137>>0]|0;
     $138 = $136&255;
     $139 = ($138|0)==($14|0);
     if (!($139)) {
      $140 = (($12) + ($138)|0);
      $141 = HEAP8[$140>>0]|0;
      HEAP8[$142>>0] = $141;
     }
     $143 = ((($dst$8)) + 2|0);
     $144 = ((($src$8)) + 2|0);
     $dst$2 = $143;$n$1 = $192;$src$2 = $144;
     label = 48;
     continue;
    }
   }
   $194 = (($137) + ($6)|0);
   $195 = (($142) + ($10)|0);
   $196 = (($197) + -1)|0;
   $198 = ($197|0)==(0);
   if ($198) {
    label = 98;
    break;
   } else {
    $197 = $196;$dst$05 = $195;$src$06 = $194;
   }
  }
  if ((label|0) == 69) {
   // unreachable;
  }
  else if ((label|0) == 98) {
   return;
  }
 }
}
function _Blit1to2Key($info) {
 $info = $info|0;
 var $$lcssa = 0, $$lcssa14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dstp$01 = 0, $dstp$01$us = 0, $dstp$1$us = 0, $dstp$2 = 0, $dstp$3 = 0, $dstp$4 = 0, $dstp$5 = 0, $dstp$6 = 0, $dstp$7 = 0, $dstp$8 = 0;
 var $n$0$us = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $src$02 = 0, $src$02$us = 0, $src$1$us = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$info>>2]|0;
 $5 = ((($info)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($info)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($info)) + 36|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($info)) + 48|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($info)) + 56|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($10|0) / 2)&-1;
 $16 = (($3) + -1)|0;
 $17 = ($3|0)==(0);
 if ($17) {
  return;
 }
 $18 = (($1) + 7)|0;
 $19 = (($18|0) / 8)&-1;
 $20 = $1 & 7;
 $21 = ($20|0)==(0);
 if ($21) {
  $83 = $16;$dstp$01$us = $8;$src$02$us = $4;
  while(1) {
   $dstp$1$us = $dstp$01$us;$n$0$us = $19;$src$1$us = $src$02$us;
   while(1) {
    $22 = HEAP8[$src$1$us>>0]|0;
    $23 = $22&255;
    $24 = ($23|0)==($14|0);
    if (!($24)) {
     $25 = (($12) + ($23<<1)|0);
     $26 = HEAP16[$25>>1]|0;
     HEAP16[$dstp$1$us>>1] = $26;
    }
    $27 = ((($src$1$us)) + 1|0);
    $28 = HEAP8[$27>>0]|0;
    $29 = $28&255;
    $30 = ($29|0)==($14|0);
    if (!($30)) {
     $31 = ((($dstp$1$us)) + 2|0);
     $32 = (($12) + ($29<<1)|0);
     $33 = HEAP16[$32>>1]|0;
     HEAP16[$31>>1] = $33;
    }
    $34 = ((($src$1$us)) + 2|0);
    $35 = HEAP8[$34>>0]|0;
    $36 = $35&255;
    $37 = ($36|0)==($14|0);
    if (!($37)) {
     $38 = ((($dstp$1$us)) + 4|0);
     $39 = (($12) + ($36<<1)|0);
     $40 = HEAP16[$39>>1]|0;
     HEAP16[$38>>1] = $40;
    }
    $41 = ((($src$1$us)) + 3|0);
    $42 = HEAP8[$41>>0]|0;
    $43 = $42&255;
    $44 = ($43|0)==($14|0);
    if (!($44)) {
     $45 = ((($dstp$1$us)) + 6|0);
     $46 = (($12) + ($43<<1)|0);
     $47 = HEAP16[$46>>1]|0;
     HEAP16[$45>>1] = $47;
    }
    $48 = ((($src$1$us)) + 4|0);
    $49 = HEAP8[$48>>0]|0;
    $50 = $49&255;
    $51 = ($50|0)==($14|0);
    if (!($51)) {
     $52 = ((($dstp$1$us)) + 8|0);
     $53 = (($12) + ($50<<1)|0);
     $54 = HEAP16[$53>>1]|0;
     HEAP16[$52>>1] = $54;
    }
    $55 = ((($src$1$us)) + 5|0);
    $56 = HEAP8[$55>>0]|0;
    $57 = $56&255;
    $58 = ($57|0)==($14|0);
    if (!($58)) {
     $59 = ((($dstp$1$us)) + 10|0);
     $60 = (($12) + ($57<<1)|0);
     $61 = HEAP16[$60>>1]|0;
     HEAP16[$59>>1] = $61;
    }
    $62 = ((($src$1$us)) + 6|0);
    $63 = HEAP8[$62>>0]|0;
    $64 = $63&255;
    $65 = ($64|0)==($14|0);
    if (!($65)) {
     $66 = ((($dstp$1$us)) + 12|0);
     $67 = (($12) + ($64<<1)|0);
     $68 = HEAP16[$67>>1]|0;
     HEAP16[$66>>1] = $68;
    }
    $69 = ((($src$1$us)) + 7|0);
    $70 = HEAP8[$69>>0]|0;
    $71 = $70&255;
    $72 = ($71|0)==($14|0);
    if (!($72)) {
     $73 = ((($dstp$1$us)) + 14|0);
     $74 = (($12) + ($71<<1)|0);
     $75 = HEAP16[$74>>1]|0;
     HEAP16[$73>>1] = $75;
    }
    $76 = ((($src$1$us)) + 8|0);
    $77 = ((($dstp$1$us)) + 16|0);
    $78 = (($n$0$us) + -1)|0;
    $79 = ($n$0$us|0)>(1);
    if ($79) {
     $dstp$1$us = $77;$n$0$us = $78;$src$1$us = $76;
    } else {
     $$lcssa = $76;$$lcssa14 = $77;
     break;
    }
   }
   $80 = (($$lcssa) + ($6)|0);
   $81 = (($$lcssa14) + ($15<<1)|0);
   $82 = (($83) + -1)|0;
   $84 = ($83|0)==(0);
   if ($84) {
    break;
   } else {
    $83 = $82;$dstp$01$us = $81;$src$02$us = $80;
   }
  }
  return;
 } else {
  $146 = $16;$dstp$01 = $8;$src$02 = $4;
 }
 L35: while(1) {
  switch ($20|0) {
  case 1:  {
   $dstp$8 = $dstp$01;$n$7 = $19;$src$8 = $src$02;
   label = 44;
   break;
  }
  case 7:  {
   $dstp$2 = $dstp$01;$n$1 = $19;$src$2 = $src$02;
   label = 26;
   break;
  }
  case 6:  {
   $dstp$3 = $dstp$01;$n$2 = $19;$src$3 = $src$02;
   label = 29;
   break;
  }
  case 5:  {
   $dstp$4 = $dstp$01;$n$3 = $19;$src$4 = $src$02;
   label = 32;
   break;
  }
  case 4:  {
   $dstp$5 = $dstp$01;$n$4 = $19;$src$5 = $src$02;
   label = 35;
   break;
  }
  case 3:  {
   $dstp$6 = $dstp$01;$n$5 = $19;$src$6 = $src$02;
   label = 38;
   break;
  }
  case 2:  {
   $dstp$7 = $dstp$01;$n$6 = $19;$src$7 = $src$02;
   label = 41;
   break;
  }
  default: {
   label = 47;
   break L35;
  }
  }
  while(1) {
   if ((label|0) == 26) {
    label = 0;
    $94 = HEAP8[$src$2>>0]|0;
    $95 = $94&255;
    $96 = ($95|0)==($14|0);
    if (!($96)) {
     $97 = (($12) + ($95<<1)|0);
     $98 = HEAP16[$97>>1]|0;
     HEAP16[$dstp$2>>1] = $98;
    }
    $99 = ((($src$2)) + 1|0);
    $100 = ((($dstp$2)) + 2|0);
    $dstp$3 = $100;$n$2 = $n$1;$src$3 = $99;
    label = 29;
    continue;
   }
   else if ((label|0) == 29) {
    label = 0;
    $101 = HEAP8[$src$3>>0]|0;
    $102 = $101&255;
    $103 = ($102|0)==($14|0);
    if (!($103)) {
     $104 = (($12) + ($102<<1)|0);
     $105 = HEAP16[$104>>1]|0;
     HEAP16[$dstp$3>>1] = $105;
    }
    $106 = ((($src$3)) + 1|0);
    $107 = ((($dstp$3)) + 2|0);
    $dstp$4 = $107;$n$3 = $n$2;$src$4 = $106;
    label = 32;
    continue;
   }
   else if ((label|0) == 32) {
    label = 0;
    $108 = HEAP8[$src$4>>0]|0;
    $109 = $108&255;
    $110 = ($109|0)==($14|0);
    if (!($110)) {
     $111 = (($12) + ($109<<1)|0);
     $112 = HEAP16[$111>>1]|0;
     HEAP16[$dstp$4>>1] = $112;
    }
    $113 = ((($src$4)) + 1|0);
    $114 = ((($dstp$4)) + 2|0);
    $dstp$5 = $114;$n$4 = $n$3;$src$5 = $113;
    label = 35;
    continue;
   }
   else if ((label|0) == 35) {
    label = 0;
    $115 = HEAP8[$src$5>>0]|0;
    $116 = $115&255;
    $117 = ($116|0)==($14|0);
    if (!($117)) {
     $118 = (($12) + ($116<<1)|0);
     $119 = HEAP16[$118>>1]|0;
     HEAP16[$dstp$5>>1] = $119;
    }
    $120 = ((($src$5)) + 1|0);
    $121 = ((($dstp$5)) + 2|0);
    $dstp$6 = $121;$n$5 = $n$4;$src$6 = $120;
    label = 38;
    continue;
   }
   else if ((label|0) == 38) {
    label = 0;
    $122 = HEAP8[$src$6>>0]|0;
    $123 = $122&255;
    $124 = ($123|0)==($14|0);
    if (!($124)) {
     $125 = (($12) + ($123<<1)|0);
     $126 = HEAP16[$125>>1]|0;
     HEAP16[$dstp$6>>1] = $126;
    }
    $127 = ((($src$6)) + 1|0);
    $128 = ((($dstp$6)) + 2|0);
    $dstp$7 = $128;$n$6 = $n$5;$src$7 = $127;
    label = 41;
    continue;
   }
   else if ((label|0) == 41) {
    label = 0;
    $129 = HEAP8[$src$7>>0]|0;
    $130 = $129&255;
    $131 = ($130|0)==($14|0);
    if (!($131)) {
     $132 = (($12) + ($130<<1)|0);
     $133 = HEAP16[$132>>1]|0;
     HEAP16[$dstp$7>>1] = $133;
    }
    $134 = ((($src$7)) + 1|0);
    $135 = ((($dstp$7)) + 2|0);
    $dstp$8 = $135;$n$7 = $n$6;$src$8 = $134;
    label = 44;
    continue;
   }
   else if ((label|0) == 44) {
    label = 0;
    $136 = HEAP8[$src$8>>0]|0;
    $137 = $136&255;
    $138 = ($137|0)==($14|0);
    if (!($138)) {
     $139 = (($12) + ($137<<1)|0);
     $140 = HEAP16[$139>>1]|0;
     HEAP16[$dstp$8>>1] = $140;
    }
    $86 = ((($src$8)) + 1|0);
    $91 = ((($dstp$8)) + 2|0);
    $141 = (($n$7) + -1)|0;
    $142 = ($n$7|0)>(1);
    if (!($142)) {
     break;
    }
    $85 = HEAP8[$86>>0]|0;
    $87 = $85&255;
    $88 = ($87|0)==($14|0);
    if (!($88)) {
     $89 = (($12) + ($87<<1)|0);
     $90 = HEAP16[$89>>1]|0;
     HEAP16[$91>>1] = $90;
    }
    $92 = ((($src$8)) + 2|0);
    $93 = ((($dstp$8)) + 4|0);
    $dstp$2 = $93;$n$1 = $141;$src$2 = $92;
    label = 26;
    continue;
   }
  }
  $143 = (($86) + ($6)|0);
  $144 = (($91) + ($15<<1)|0);
  $145 = (($146) + -1)|0;
  $147 = ($146|0)==(0);
  if ($147) {
   label = 49;
   break;
  } else {
   $146 = $145;$dstp$01 = $144;$src$02 = $143;
  }
 }
 if ((label|0) == 47) {
  // unreachable;
 }
 else if ((label|0) == 49) {
  return;
 }
}
function _Blit1to3Key($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$01 = 0, $dst$1 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0;
 var $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $src$02 = 0, $src$1 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 48|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($info)) + 56|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($3|0)==(0);
 if ($12) {
  return;
 }
 $13 = ((($info)) + 20|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = HEAP32[$info>>2]|0;
 $16 = (($1) + 7)|0;
 $17 = (($16|0) / 8)&-1;
 $18 = $1 & 7;
 $$in = $3;$dst$01 = $14;$src$02 = $15;
 L4: while(1) {
  $19 = (($$in) + -1)|0;
  switch ($18|0) {
  case 0:  {
   $dst$1 = $dst$01;$n$0 = $17;$src$1 = $src$02;
   label = 4;
   break;
  }
  case 7:  {
   $dst$2 = $dst$01;$n$1 = $17;$src$2 = $src$02;
   label = 7;
   break;
  }
  case 6:  {
   $dst$3 = $dst$01;$n$2 = $17;$src$3 = $src$02;
   label = 10;
   break;
  }
  case 5:  {
   $dst$4 = $dst$01;$n$3 = $17;$src$4 = $src$02;
   label = 13;
   break;
  }
  case 4:  {
   $dst$5 = $dst$01;$n$4 = $17;$src$5 = $src$02;
   label = 16;
   break;
  }
  case 3:  {
   $dst$6 = $dst$01;$n$5 = $17;$src$6 = $src$02;
   label = 19;
   break;
  }
  case 2:  {
   $dst$7 = $dst$01;$n$6 = $17;$src$7 = $src$02;
   label = 22;
   break;
  }
  case 1:  {
   $dst$8 = $dst$01;$n$7 = $17;$src$8 = $src$02;
   label = 25;
   break;
  }
  default: {
   label = 28;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    $20 = HEAP8[$src$1>>0]|0;
    $21 = $20&255;
    $22 = ($21|0)==($11|0);
    if (!($22)) {
     $23 = $21 << 2;
     $24 = $23 | 1;
     $25 = (($9) + ($23)|0);
     $26 = HEAP8[$25>>0]|0;
     HEAP8[$dst$1>>0] = $26;
     $27 = (($24) + 1)|0;
     $28 = (($9) + ($24)|0);
     $29 = HEAP8[$28>>0]|0;
     $30 = ((($dst$1)) + 1|0);
     HEAP8[$30>>0] = $29;
     $31 = (($9) + ($27)|0);
     $32 = HEAP8[$31>>0]|0;
     $33 = ((($dst$1)) + 2|0);
     HEAP8[$33>>0] = $32;
    }
    $34 = ((($src$1)) + 1|0);
    $35 = ((($dst$1)) + 3|0);
    $dst$2 = $35;$n$1 = $n$0;$src$2 = $34;
    label = 7;
    continue;
   }
   else if ((label|0) == 7) {
    label = 0;
    $36 = HEAP8[$src$2>>0]|0;
    $37 = $36&255;
    $38 = ($37|0)==($11|0);
    if (!($38)) {
     $39 = $37 << 2;
     $40 = $39 | 1;
     $41 = (($9) + ($39)|0);
     $42 = HEAP8[$41>>0]|0;
     HEAP8[$dst$2>>0] = $42;
     $43 = (($40) + 1)|0;
     $44 = (($9) + ($40)|0);
     $45 = HEAP8[$44>>0]|0;
     $46 = ((($dst$2)) + 1|0);
     HEAP8[$46>>0] = $45;
     $47 = (($9) + ($43)|0);
     $48 = HEAP8[$47>>0]|0;
     $49 = ((($dst$2)) + 2|0);
     HEAP8[$49>>0] = $48;
    }
    $50 = ((($src$2)) + 1|0);
    $51 = ((($dst$2)) + 3|0);
    $dst$3 = $51;$n$2 = $n$1;$src$3 = $50;
    label = 10;
    continue;
   }
   else if ((label|0) == 10) {
    label = 0;
    $52 = HEAP8[$src$3>>0]|0;
    $53 = $52&255;
    $54 = ($53|0)==($11|0);
    if (!($54)) {
     $55 = $53 << 2;
     $56 = $55 | 1;
     $57 = (($9) + ($55)|0);
     $58 = HEAP8[$57>>0]|0;
     HEAP8[$dst$3>>0] = $58;
     $59 = (($56) + 1)|0;
     $60 = (($9) + ($56)|0);
     $61 = HEAP8[$60>>0]|0;
     $62 = ((($dst$3)) + 1|0);
     HEAP8[$62>>0] = $61;
     $63 = (($9) + ($59)|0);
     $64 = HEAP8[$63>>0]|0;
     $65 = ((($dst$3)) + 2|0);
     HEAP8[$65>>0] = $64;
    }
    $66 = ((($src$3)) + 1|0);
    $67 = ((($dst$3)) + 3|0);
    $dst$4 = $67;$n$3 = $n$2;$src$4 = $66;
    label = 13;
    continue;
   }
   else if ((label|0) == 13) {
    label = 0;
    $68 = HEAP8[$src$4>>0]|0;
    $69 = $68&255;
    $70 = ($69|0)==($11|0);
    if (!($70)) {
     $71 = $69 << 2;
     $72 = $71 | 1;
     $73 = (($9) + ($71)|0);
     $74 = HEAP8[$73>>0]|0;
     HEAP8[$dst$4>>0] = $74;
     $75 = (($72) + 1)|0;
     $76 = (($9) + ($72)|0);
     $77 = HEAP8[$76>>0]|0;
     $78 = ((($dst$4)) + 1|0);
     HEAP8[$78>>0] = $77;
     $79 = (($9) + ($75)|0);
     $80 = HEAP8[$79>>0]|0;
     $81 = ((($dst$4)) + 2|0);
     HEAP8[$81>>0] = $80;
    }
    $82 = ((($src$4)) + 1|0);
    $83 = ((($dst$4)) + 3|0);
    $dst$5 = $83;$n$4 = $n$3;$src$5 = $82;
    label = 16;
    continue;
   }
   else if ((label|0) == 16) {
    label = 0;
    $84 = HEAP8[$src$5>>0]|0;
    $85 = $84&255;
    $86 = ($85|0)==($11|0);
    if (!($86)) {
     $87 = $85 << 2;
     $88 = $87 | 1;
     $89 = (($9) + ($87)|0);
     $90 = HEAP8[$89>>0]|0;
     HEAP8[$dst$5>>0] = $90;
     $91 = (($88) + 1)|0;
     $92 = (($9) + ($88)|0);
     $93 = HEAP8[$92>>0]|0;
     $94 = ((($dst$5)) + 1|0);
     HEAP8[$94>>0] = $93;
     $95 = (($9) + ($91)|0);
     $96 = HEAP8[$95>>0]|0;
     $97 = ((($dst$5)) + 2|0);
     HEAP8[$97>>0] = $96;
    }
    $98 = ((($src$5)) + 1|0);
    $99 = ((($dst$5)) + 3|0);
    $dst$6 = $99;$n$5 = $n$4;$src$6 = $98;
    label = 19;
    continue;
   }
   else if ((label|0) == 19) {
    label = 0;
    $100 = HEAP8[$src$6>>0]|0;
    $101 = $100&255;
    $102 = ($101|0)==($11|0);
    if (!($102)) {
     $103 = $101 << 2;
     $104 = $103 | 1;
     $105 = (($9) + ($103)|0);
     $106 = HEAP8[$105>>0]|0;
     HEAP8[$dst$6>>0] = $106;
     $107 = (($104) + 1)|0;
     $108 = (($9) + ($104)|0);
     $109 = HEAP8[$108>>0]|0;
     $110 = ((($dst$6)) + 1|0);
     HEAP8[$110>>0] = $109;
     $111 = (($9) + ($107)|0);
     $112 = HEAP8[$111>>0]|0;
     $113 = ((($dst$6)) + 2|0);
     HEAP8[$113>>0] = $112;
    }
    $114 = ((($src$6)) + 1|0);
    $115 = ((($dst$6)) + 3|0);
    $dst$7 = $115;$n$6 = $n$5;$src$7 = $114;
    label = 22;
    continue;
   }
   else if ((label|0) == 22) {
    label = 0;
    $116 = HEAP8[$src$7>>0]|0;
    $117 = $116&255;
    $118 = ($117|0)==($11|0);
    if (!($118)) {
     $119 = $117 << 2;
     $120 = $119 | 1;
     $121 = (($9) + ($119)|0);
     $122 = HEAP8[$121>>0]|0;
     HEAP8[$dst$7>>0] = $122;
     $123 = (($120) + 1)|0;
     $124 = (($9) + ($120)|0);
     $125 = HEAP8[$124>>0]|0;
     $126 = ((($dst$7)) + 1|0);
     HEAP8[$126>>0] = $125;
     $127 = (($9) + ($123)|0);
     $128 = HEAP8[$127>>0]|0;
     $129 = ((($dst$7)) + 2|0);
     HEAP8[$129>>0] = $128;
    }
    $130 = ((($src$7)) + 1|0);
    $131 = ((($dst$7)) + 3|0);
    $dst$8 = $131;$n$7 = $n$6;$src$8 = $130;
    label = 25;
    continue;
   }
   else if ((label|0) == 25) {
    label = 0;
    $132 = HEAP8[$src$8>>0]|0;
    $133 = $132&255;
    $134 = ($133|0)==($11|0);
    if (!($134)) {
     $135 = $133 << 2;
     $136 = $135 | 1;
     $137 = (($9) + ($135)|0);
     $138 = HEAP8[$137>>0]|0;
     HEAP8[$dst$8>>0] = $138;
     $139 = (($136) + 1)|0;
     $140 = (($9) + ($136)|0);
     $141 = HEAP8[$140>>0]|0;
     $142 = ((($dst$8)) + 1|0);
     HEAP8[$142>>0] = $141;
     $143 = (($9) + ($139)|0);
     $144 = HEAP8[$143>>0]|0;
     $145 = ((($dst$8)) + 2|0);
     HEAP8[$145>>0] = $144;
    }
    $146 = ((($src$8)) + 1|0);
    $147 = ((($dst$8)) + 3|0);
    $148 = (($n$7) + -1)|0;
    $149 = ($n$7|0)>(1);
    if ($149) {
     $dst$1 = $147;$n$0 = $148;$src$1 = $146;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $150 = (($146) + ($5)|0);
  $151 = (($147) + ($7)|0);
  $152 = ($19|0)==(0);
  if ($152) {
   label = 30;
   break;
  } else {
   $$in = $19;$dst$01 = $151;$src$02 = $150;
  }
 }
 if ((label|0) == 28) {
  // unreachable;
 }
 else if ((label|0) == 30) {
  return;
 }
}
function _Blit1to4Key($info) {
 $info = $info|0;
 var $$lcssa = 0, $$lcssa14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dstp$01 = 0, $dstp$01$us = 0, $dstp$1$us = 0, $dstp$2 = 0, $dstp$3 = 0, $dstp$4 = 0, $dstp$5 = 0, $dstp$6 = 0, $dstp$7 = 0, $dstp$8 = 0;
 var $n$0$us = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $src$02 = 0, $src$02$us = 0, $src$1$us = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$info>>2]|0;
 $5 = ((($info)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($info)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($info)) + 36|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($info)) + 48|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($info)) + 56|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($10|0) / 4)&-1;
 $16 = (($3) + -1)|0;
 $17 = ($3|0)==(0);
 if ($17) {
  return;
 }
 $18 = (($1) + 7)|0;
 $19 = (($18|0) / 8)&-1;
 $20 = $1 & 7;
 $21 = ($20|0)==(0);
 if ($21) {
  $83 = $16;$dstp$01$us = $8;$src$02$us = $4;
  while(1) {
   $dstp$1$us = $dstp$01$us;$n$0$us = $19;$src$1$us = $src$02$us;
   while(1) {
    $22 = HEAP8[$src$1$us>>0]|0;
    $23 = $22&255;
    $24 = ($23|0)==($14|0);
    if (!($24)) {
     $25 = (($12) + ($23<<2)|0);
     $26 = HEAP32[$25>>2]|0;
     HEAP32[$dstp$1$us>>2] = $26;
    }
    $27 = ((($src$1$us)) + 1|0);
    $28 = HEAP8[$27>>0]|0;
    $29 = $28&255;
    $30 = ($29|0)==($14|0);
    if (!($30)) {
     $31 = ((($dstp$1$us)) + 4|0);
     $32 = (($12) + ($29<<2)|0);
     $33 = HEAP32[$32>>2]|0;
     HEAP32[$31>>2] = $33;
    }
    $34 = ((($src$1$us)) + 2|0);
    $35 = HEAP8[$34>>0]|0;
    $36 = $35&255;
    $37 = ($36|0)==($14|0);
    if (!($37)) {
     $38 = ((($dstp$1$us)) + 8|0);
     $39 = (($12) + ($36<<2)|0);
     $40 = HEAP32[$39>>2]|0;
     HEAP32[$38>>2] = $40;
    }
    $41 = ((($src$1$us)) + 3|0);
    $42 = HEAP8[$41>>0]|0;
    $43 = $42&255;
    $44 = ($43|0)==($14|0);
    if (!($44)) {
     $45 = ((($dstp$1$us)) + 12|0);
     $46 = (($12) + ($43<<2)|0);
     $47 = HEAP32[$46>>2]|0;
     HEAP32[$45>>2] = $47;
    }
    $48 = ((($src$1$us)) + 4|0);
    $49 = HEAP8[$48>>0]|0;
    $50 = $49&255;
    $51 = ($50|0)==($14|0);
    if (!($51)) {
     $52 = ((($dstp$1$us)) + 16|0);
     $53 = (($12) + ($50<<2)|0);
     $54 = HEAP32[$53>>2]|0;
     HEAP32[$52>>2] = $54;
    }
    $55 = ((($src$1$us)) + 5|0);
    $56 = HEAP8[$55>>0]|0;
    $57 = $56&255;
    $58 = ($57|0)==($14|0);
    if (!($58)) {
     $59 = ((($dstp$1$us)) + 20|0);
     $60 = (($12) + ($57<<2)|0);
     $61 = HEAP32[$60>>2]|0;
     HEAP32[$59>>2] = $61;
    }
    $62 = ((($src$1$us)) + 6|0);
    $63 = HEAP8[$62>>0]|0;
    $64 = $63&255;
    $65 = ($64|0)==($14|0);
    if (!($65)) {
     $66 = ((($dstp$1$us)) + 24|0);
     $67 = (($12) + ($64<<2)|0);
     $68 = HEAP32[$67>>2]|0;
     HEAP32[$66>>2] = $68;
    }
    $69 = ((($src$1$us)) + 7|0);
    $70 = HEAP8[$69>>0]|0;
    $71 = $70&255;
    $72 = ($71|0)==($14|0);
    if (!($72)) {
     $73 = ((($dstp$1$us)) + 28|0);
     $74 = (($12) + ($71<<2)|0);
     $75 = HEAP32[$74>>2]|0;
     HEAP32[$73>>2] = $75;
    }
    $76 = ((($src$1$us)) + 8|0);
    $77 = ((($dstp$1$us)) + 32|0);
    $78 = (($n$0$us) + -1)|0;
    $79 = ($n$0$us|0)>(1);
    if ($79) {
     $dstp$1$us = $77;$n$0$us = $78;$src$1$us = $76;
    } else {
     $$lcssa = $76;$$lcssa14 = $77;
     break;
    }
   }
   $80 = (($$lcssa) + ($6)|0);
   $81 = (($$lcssa14) + ($15<<2)|0);
   $82 = (($83) + -1)|0;
   $84 = ($83|0)==(0);
   if ($84) {
    break;
   } else {
    $83 = $82;$dstp$01$us = $81;$src$02$us = $80;
   }
  }
  return;
 } else {
  $146 = $16;$dstp$01 = $8;$src$02 = $4;
 }
 L35: while(1) {
  switch ($20|0) {
  case 1:  {
   $dstp$8 = $dstp$01;$n$7 = $19;$src$8 = $src$02;
   label = 44;
   break;
  }
  case 7:  {
   $dstp$2 = $dstp$01;$n$1 = $19;$src$2 = $src$02;
   label = 26;
   break;
  }
  case 6:  {
   $dstp$3 = $dstp$01;$n$2 = $19;$src$3 = $src$02;
   label = 29;
   break;
  }
  case 5:  {
   $dstp$4 = $dstp$01;$n$3 = $19;$src$4 = $src$02;
   label = 32;
   break;
  }
  case 4:  {
   $dstp$5 = $dstp$01;$n$4 = $19;$src$5 = $src$02;
   label = 35;
   break;
  }
  case 3:  {
   $dstp$6 = $dstp$01;$n$5 = $19;$src$6 = $src$02;
   label = 38;
   break;
  }
  case 2:  {
   $dstp$7 = $dstp$01;$n$6 = $19;$src$7 = $src$02;
   label = 41;
   break;
  }
  default: {
   label = 47;
   break L35;
  }
  }
  while(1) {
   if ((label|0) == 26) {
    label = 0;
    $94 = HEAP8[$src$2>>0]|0;
    $95 = $94&255;
    $96 = ($95|0)==($14|0);
    if (!($96)) {
     $97 = (($12) + ($95<<2)|0);
     $98 = HEAP32[$97>>2]|0;
     HEAP32[$dstp$2>>2] = $98;
    }
    $99 = ((($src$2)) + 1|0);
    $100 = ((($dstp$2)) + 4|0);
    $dstp$3 = $100;$n$2 = $n$1;$src$3 = $99;
    label = 29;
    continue;
   }
   else if ((label|0) == 29) {
    label = 0;
    $101 = HEAP8[$src$3>>0]|0;
    $102 = $101&255;
    $103 = ($102|0)==($14|0);
    if (!($103)) {
     $104 = (($12) + ($102<<2)|0);
     $105 = HEAP32[$104>>2]|0;
     HEAP32[$dstp$3>>2] = $105;
    }
    $106 = ((($src$3)) + 1|0);
    $107 = ((($dstp$3)) + 4|0);
    $dstp$4 = $107;$n$3 = $n$2;$src$4 = $106;
    label = 32;
    continue;
   }
   else if ((label|0) == 32) {
    label = 0;
    $108 = HEAP8[$src$4>>0]|0;
    $109 = $108&255;
    $110 = ($109|0)==($14|0);
    if (!($110)) {
     $111 = (($12) + ($109<<2)|0);
     $112 = HEAP32[$111>>2]|0;
     HEAP32[$dstp$4>>2] = $112;
    }
    $113 = ((($src$4)) + 1|0);
    $114 = ((($dstp$4)) + 4|0);
    $dstp$5 = $114;$n$4 = $n$3;$src$5 = $113;
    label = 35;
    continue;
   }
   else if ((label|0) == 35) {
    label = 0;
    $115 = HEAP8[$src$5>>0]|0;
    $116 = $115&255;
    $117 = ($116|0)==($14|0);
    if (!($117)) {
     $118 = (($12) + ($116<<2)|0);
     $119 = HEAP32[$118>>2]|0;
     HEAP32[$dstp$5>>2] = $119;
    }
    $120 = ((($src$5)) + 1|0);
    $121 = ((($dstp$5)) + 4|0);
    $dstp$6 = $121;$n$5 = $n$4;$src$6 = $120;
    label = 38;
    continue;
   }
   else if ((label|0) == 38) {
    label = 0;
    $122 = HEAP8[$src$6>>0]|0;
    $123 = $122&255;
    $124 = ($123|0)==($14|0);
    if (!($124)) {
     $125 = (($12) + ($123<<2)|0);
     $126 = HEAP32[$125>>2]|0;
     HEAP32[$dstp$6>>2] = $126;
    }
    $127 = ((($src$6)) + 1|0);
    $128 = ((($dstp$6)) + 4|0);
    $dstp$7 = $128;$n$6 = $n$5;$src$7 = $127;
    label = 41;
    continue;
   }
   else if ((label|0) == 41) {
    label = 0;
    $129 = HEAP8[$src$7>>0]|0;
    $130 = $129&255;
    $131 = ($130|0)==($14|0);
    if (!($131)) {
     $132 = (($12) + ($130<<2)|0);
     $133 = HEAP32[$132>>2]|0;
     HEAP32[$dstp$7>>2] = $133;
    }
    $134 = ((($src$7)) + 1|0);
    $135 = ((($dstp$7)) + 4|0);
    $dstp$8 = $135;$n$7 = $n$6;$src$8 = $134;
    label = 44;
    continue;
   }
   else if ((label|0) == 44) {
    label = 0;
    $136 = HEAP8[$src$8>>0]|0;
    $137 = $136&255;
    $138 = ($137|0)==($14|0);
    if (!($138)) {
     $139 = (($12) + ($137<<2)|0);
     $140 = HEAP32[$139>>2]|0;
     HEAP32[$dstp$8>>2] = $140;
    }
    $86 = ((($src$8)) + 1|0);
    $91 = ((($dstp$8)) + 4|0);
    $141 = (($n$7) + -1)|0;
    $142 = ($n$7|0)>(1);
    if (!($142)) {
     break;
    }
    $85 = HEAP8[$86>>0]|0;
    $87 = $85&255;
    $88 = ($87|0)==($14|0);
    if (!($88)) {
     $89 = (($12) + ($87<<2)|0);
     $90 = HEAP32[$89>>2]|0;
     HEAP32[$91>>2] = $90;
    }
    $92 = ((($src$8)) + 2|0);
    $93 = ((($dstp$8)) + 8|0);
    $dstp$2 = $93;$n$1 = $141;$src$2 = $92;
    label = 26;
    continue;
   }
  }
  $143 = (($86) + ($6)|0);
  $144 = (($91) + ($15<<2)|0);
  $145 = (($146) + -1)|0;
  $147 = ($146|0)==(0);
  if ($147) {
   label = 49;
   break;
  } else {
   $146 = $145;$dstp$01 = $144;$src$02 = $143;
  }
 }
 if ((label|0) == 47) {
  // unreachable;
 }
 else if ((label|0) == 49) {
  return;
 }
}
function _Blit1toNAlpha($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0;
 var $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0;
 var $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0;
 var $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0;
 var $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0;
 var $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0;
 var $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0;
 var $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0;
 var $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0;
 var $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0;
 var $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0;
 var $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0;
 var $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0;
 var $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0;
 var $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0;
 var $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0;
 var $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0;
 var $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0;
 var $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0;
 var $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0;
 var $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0;
 var $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0;
 var $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0;
 var $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0;
 var $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0;
 var $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0;
 var $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0;
 var $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0;
 var $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0;
 var $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0;
 var $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0;
 var $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0;
 var $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0;
 var $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0;
 var $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0;
 var $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0;
 var $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0;
 var $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0;
 var $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0;
 var $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0;
 var $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0;
 var $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0;
 var $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0;
 var $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0;
 var $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0;
 var $996 = 0, $997 = 0, $998 = 0, $999 = 0, $dA$0 = 0, $dA$0$shrunk = 0, $dA$1 = 0, $dA$1$shrunk = 0, $dA$2 = 0, $dA$2$shrunk = 0, $dA$3 = 0, $dA$3$shrunk = 0, $dB$0 = 0, $dB$0$shrunk = 0, $dB$1 = 0, $dB$1$shrunk = 0, $dB$2 = 0, $dB$2$shrunk = 0, $dB$3 = 0, $dB$3$shrunk = 0;
 var $dG$0 = 0, $dG$0$shrunk = 0, $dG$1 = 0, $dG$1$shrunk = 0, $dG$2 = 0, $dG$2$shrunk = 0, $dG$3 = 0, $dG$3$shrunk = 0, $dR$0 = 0, $dR$0$shrunk = 0, $dR$1 = 0, $dR$1$shrunk = 0, $dR$2 = 0, $dR$2$shrunk = 0, $dR$3 = 0, $dR$3$shrunk = 0, $div = 0, $div1 = 0, $div10 = 0, $div12 = 0;
 var $div13 = 0, $div14 = 0, $div15 = 0, $div16 = 0, $div18 = 0, $div19 = 0, $div2 = 0, $div20 = 0, $div21 = 0, $div22 = 0, $div3 = 0, $div4 = 0, $div6 = 0, $div7 = 0, $div8 = 0, $div9 = 0, $dst$027 = 0, $dst$1 = 0, $dst$2 = 0, $dst$3 = 0;
 var $dst$4 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $src$028 = 0, $src$1 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 44|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($info)) + 40|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($11)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($13)) + 4|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($info)) + 63|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = $17&255;
 $19 = ((($9)) + 9|0);
 $20 = HEAP8[$19>>0]|0;
 $21 = $20&255;
 $22 = ($3|0)==(0);
 if ($22) {
  return;
 }
 $23 = ((($info)) + 20|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = HEAP32[$info>>2]|0;
 $26 = (($1) + 3)|0;
 $27 = (($26|0) / 4)&-1;
 $28 = $1 & 3;
 $29 = ((($9)) + 28|0);
 $30 = ((($9)) + 32|0);
 $31 = ((($9)) + 33|0);
 $32 = ((($9)) + 34|0);
 $33 = ((($9)) + 12|0);
 $34 = ((($9)) + 16|0);
 $35 = ((($9)) + 20|0);
 $36 = ((($9)) + 24|0);
 $$in = $3;$dst$027 = $24;$src$028 = $25;
 L4: while(1) {
  $37 = (($$in) + -1)|0;
  switch ($28|0) {
  case 0:  {
   $dst$1 = $dst$027;$n$0 = $27;$src$1 = $src$028;
   label = 4;
   break;
  }
  case 3:  {
   $dst$2 = $dst$027;$n$1 = $27;$src$2 = $src$028;
   label = 15;
   break;
  }
  case 2:  {
   $dst$3 = $dst$027;$n$2 = $27;$src$3 = $src$028;
   label = 26;
   break;
  }
  case 1:  {
   $dst$4 = $dst$027;$n$3 = $27;$src$4 = $src$028;
   label = 37;
   break;
  }
  default: {
   label = 48;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    $38 = HEAP8[$src$1>>0]|0;
    $39 = $38&255;
    $40 = (($15) + ($39<<2)|0);
    $41 = HEAP8[$40>>0]|0;
    $42 = $41&255;
    $43 = (((($15) + ($39<<2)|0)) + 1|0);
    $44 = HEAP8[$43>>0]|0;
    $45 = $44&255;
    $46 = (((($15) + ($39<<2)|0)) + 2|0);
    $47 = HEAP8[$46>>0]|0;
    $48 = $47&255;
    switch ($21|0) {
    case 1:  {
     $49 = HEAP8[$dst$1>>0]|0;
     $50 = $49&255;
     $51 = HEAP32[$33>>2]|0;
     $52 = $50 & $51;
     $53 = HEAP32[$30>>2]|0;
     $54 = $53 & 255;
     $55 = $52 >>> $54;
     $56 = HEAP32[$29>>2]|0;
     $57 = $56 & 255;
     $58 = (125924 + ($57<<2)|0);
     $59 = HEAP32[$58>>2]|0;
     $60 = (($59) + ($55)|0);
     $61 = HEAP8[$60>>0]|0;
     $62 = HEAP32[$34>>2]|0;
     $63 = $62 & $50;
     $64 = $53 >>> 8;
     $65 = $64 & 255;
     $66 = $63 >>> $65;
     $67 = $56 >>> 8;
     $68 = $67 & 255;
     $69 = (125924 + ($68<<2)|0);
     $70 = HEAP32[$69>>2]|0;
     $71 = (($70) + ($66)|0);
     $72 = HEAP8[$71>>0]|0;
     $73 = HEAP32[$35>>2]|0;
     $74 = $73 & $50;
     $75 = $53 >>> 16;
     $76 = $75 & 255;
     $77 = $74 >>> $76;
     $78 = $56 >>> 16;
     $79 = $78 & 255;
     $80 = (125924 + ($79<<2)|0);
     $81 = HEAP32[$80>>2]|0;
     $82 = (($81) + ($77)|0);
     $83 = HEAP8[$82>>0]|0;
     $84 = HEAP32[$36>>2]|0;
     $85 = $84 & $50;
     $86 = $53 >>> 24;
     $87 = $85 >>> $86;
     $88 = $56 >>> 24;
     $89 = (125924 + ($88<<2)|0);
     $90 = HEAP32[$89>>2]|0;
     $91 = (($90) + ($87)|0);
     $92 = HEAP8[$91>>0]|0;
     $dA$0$shrunk = $92;$dB$0$shrunk = $83;$dG$0$shrunk = $72;$dR$0$shrunk = $61;
     break;
    }
    case 2:  {
     $93 = HEAP16[$dst$1>>1]|0;
     $94 = $93&65535;
     $95 = HEAP32[$33>>2]|0;
     $96 = $94 & $95;
     $97 = HEAP32[$30>>2]|0;
     $98 = $97 & 255;
     $99 = $96 >>> $98;
     $100 = HEAP32[$29>>2]|0;
     $101 = $100 & 255;
     $102 = (125924 + ($101<<2)|0);
     $103 = HEAP32[$102>>2]|0;
     $104 = (($103) + ($99)|0);
     $105 = HEAP8[$104>>0]|0;
     $106 = HEAP32[$34>>2]|0;
     $107 = $106 & $94;
     $108 = $97 >>> 8;
     $109 = $108 & 255;
     $110 = $107 >>> $109;
     $111 = $100 >>> 8;
     $112 = $111 & 255;
     $113 = (125924 + ($112<<2)|0);
     $114 = HEAP32[$113>>2]|0;
     $115 = (($114) + ($110)|0);
     $116 = HEAP8[$115>>0]|0;
     $117 = HEAP32[$35>>2]|0;
     $118 = $117 & $94;
     $119 = $97 >>> 16;
     $120 = $119 & 255;
     $121 = $118 >>> $120;
     $122 = $100 >>> 16;
     $123 = $122 & 255;
     $124 = (125924 + ($123<<2)|0);
     $125 = HEAP32[$124>>2]|0;
     $126 = (($125) + ($121)|0);
     $127 = HEAP8[$126>>0]|0;
     $128 = HEAP32[$36>>2]|0;
     $129 = $128 & $94;
     $130 = $97 >>> 24;
     $131 = $129 >>> $130;
     $132 = $100 >>> 24;
     $133 = (125924 + ($132<<2)|0);
     $134 = HEAP32[$133>>2]|0;
     $135 = (($134) + ($131)|0);
     $136 = HEAP8[$135>>0]|0;
     $dA$0$shrunk = $136;$dB$0$shrunk = $127;$dG$0$shrunk = $116;$dR$0$shrunk = $105;
     break;
    }
    case 3:  {
     $137 = HEAP32[$30>>2]|0;
     $div15 = $137 >>> 3;
     $138 = $div15 & 31;
     $139 = (($dst$1) + ($138)|0);
     $140 = HEAP8[$139>>0]|0;
     $141 = $137 >>> 11;
     $div16 = $141 & 31;
     $142 = (($dst$1) + ($div16)|0);
     $143 = HEAP8[$142>>0]|0;
     $144 = $137 >>> 19;
     $145 = $144 & 31;
     $146 = (($dst$1) + ($145)|0);
     $147 = HEAP8[$146>>0]|0;
     $dA$0$shrunk = -1;$dB$0$shrunk = $147;$dG$0$shrunk = $143;$dR$0$shrunk = $140;
     break;
    }
    case 4:  {
     $148 = HEAP32[$dst$1>>2]|0;
     $149 = HEAP32[$33>>2]|0;
     $150 = $149 & $148;
     $151 = HEAP32[$30>>2]|0;
     $152 = $151 & 255;
     $153 = $150 >>> $152;
     $154 = HEAP32[$29>>2]|0;
     $155 = $154 & 255;
     $156 = (125924 + ($155<<2)|0);
     $157 = HEAP32[$156>>2]|0;
     $158 = (($157) + ($153)|0);
     $159 = HEAP8[$158>>0]|0;
     $160 = HEAP32[$34>>2]|0;
     $161 = $160 & $148;
     $162 = $151 >>> 8;
     $163 = $162 & 255;
     $164 = $161 >>> $163;
     $165 = $154 >>> 8;
     $166 = $165 & 255;
     $167 = (125924 + ($166<<2)|0);
     $168 = HEAP32[$167>>2]|0;
     $169 = (($168) + ($164)|0);
     $170 = HEAP8[$169>>0]|0;
     $171 = HEAP32[$35>>2]|0;
     $172 = $171 & $148;
     $173 = $151 >>> 16;
     $174 = $173 & 255;
     $175 = $172 >>> $174;
     $176 = $154 >>> 16;
     $177 = $176 & 255;
     $178 = (125924 + ($177<<2)|0);
     $179 = HEAP32[$178>>2]|0;
     $180 = (($179) + ($175)|0);
     $181 = HEAP8[$180>>0]|0;
     $182 = HEAP32[$36>>2]|0;
     $183 = $182 & $148;
     $184 = $151 >>> 24;
     $185 = $183 >>> $184;
     $186 = $154 >>> 24;
     $187 = (125924 + ($186<<2)|0);
     $188 = HEAP32[$187>>2]|0;
     $189 = (($188) + ($185)|0);
     $190 = HEAP8[$189>>0]|0;
     $dA$0$shrunk = $190;$dB$0$shrunk = $181;$dG$0$shrunk = $170;$dR$0$shrunk = $159;
     break;
    }
    default: {
     $dA$0$shrunk = 0;$dB$0$shrunk = 0;$dG$0$shrunk = 0;$dR$0$shrunk = 0;
    }
    }
    $dR$0 = $dR$0$shrunk&255;
    $dG$0 = $dG$0$shrunk&255;
    $dB$0 = $dB$0$shrunk&255;
    $dA$0 = $dA$0$shrunk&255;
    $191 = (($42) - ($dR$0))|0;
    $192 = Math_imul($191, $18)|0;
    $193 = (($192>>>0) / 255)&-1;
    $194 = (($193) + ($dR$0))|0;
    $195 = (($45) - ($dG$0))|0;
    $196 = Math_imul($195, $18)|0;
    $197 = (($196>>>0) / 255)&-1;
    $198 = (($197) + ($dG$0))|0;
    $199 = (($48) - ($dB$0))|0;
    $200 = Math_imul($199, $18)|0;
    $201 = (($200>>>0) / 255)&-1;
    $202 = (($201) + ($dB$0))|0;
    $203 = (($dA$0) + ($18))|0;
    $204 = Math_imul($dA$0, $18)|0;
    $205 = (($204>>>0) / 255)&-1;
    $206 = (($203) - ($205))|0;
    switch ($21|0) {
    case 1:  {
     $207 = HEAP32[$29>>2]|0;
     $208 = $207 & 255;
     $209 = $194 >>> $208;
     $210 = HEAP32[$30>>2]|0;
     $211 = $210 & 255;
     $212 = $209 << $211;
     $213 = $207 >>> 8;
     $214 = $213 & 255;
     $215 = $198 >>> $214;
     $216 = $210 >>> 8;
     $217 = $216 & 255;
     $218 = $215 << $217;
     $219 = $218 | $212;
     $220 = $207 >>> 16;
     $221 = $220 & 255;
     $222 = $202 >>> $221;
     $223 = $210 >>> 16;
     $224 = $223 & 255;
     $225 = $222 << $224;
     $226 = $219 | $225;
     $227 = $207 >>> 24;
     $228 = $206 >>> $227;
     $229 = $210 >>> 24;
     $230 = $228 << $229;
     $231 = $226 | $230;
     $232 = $231&255;
     HEAP8[$dst$1>>0] = $232;
     break;
    }
    case 2:  {
     $233 = HEAP32[$29>>2]|0;
     $234 = $233 & 255;
     $235 = $194 >>> $234;
     $236 = HEAP32[$30>>2]|0;
     $237 = $236 & 255;
     $238 = $235 << $237;
     $239 = $233 >>> 8;
     $240 = $239 & 255;
     $241 = $198 >>> $240;
     $242 = $236 >>> 8;
     $243 = $242 & 255;
     $244 = $241 << $243;
     $245 = $244 | $238;
     $246 = $233 >>> 16;
     $247 = $246 & 255;
     $248 = $202 >>> $247;
     $249 = $236 >>> 16;
     $250 = $249 & 255;
     $251 = $248 << $250;
     $252 = $245 | $251;
     $253 = $233 >>> 24;
     $254 = $206 >>> $253;
     $255 = $236 >>> 24;
     $256 = $254 << $255;
     $257 = $252 | $256;
     $258 = $257&65535;
     HEAP16[$dst$1>>1] = $258;
     break;
    }
    case 3:  {
     $259 = $194&255;
     $260 = HEAP8[$30>>0]|0;
     $div12 = ($260&255) >>> 3;
     $261 = $div12&255;
     $262 = (($dst$1) + ($261)|0);
     HEAP8[$262>>0] = $259;
     $263 = $198&255;
     $264 = HEAP8[$31>>0]|0;
     $div13 = ($264&255) >>> 3;
     $265 = $div13&255;
     $266 = (($dst$1) + ($265)|0);
     HEAP8[$266>>0] = $263;
     $267 = $202&255;
     $268 = HEAP8[$32>>0]|0;
     $div14 = ($268&255) >>> 3;
     $269 = $div14&255;
     $270 = (($dst$1) + ($269)|0);
     HEAP8[$270>>0] = $267;
     break;
    }
    case 4:  {
     $271 = HEAP32[$29>>2]|0;
     $272 = $271 & 255;
     $273 = $194 >>> $272;
     $274 = HEAP32[$30>>2]|0;
     $275 = $274 & 255;
     $276 = $273 << $275;
     $277 = $271 >>> 8;
     $278 = $277 & 255;
     $279 = $198 >>> $278;
     $280 = $274 >>> 8;
     $281 = $280 & 255;
     $282 = $279 << $281;
     $283 = $282 | $276;
     $284 = $271 >>> 16;
     $285 = $284 & 255;
     $286 = $202 >>> $285;
     $287 = $274 >>> 16;
     $288 = $287 & 255;
     $289 = $286 << $288;
     $290 = $283 | $289;
     $291 = $271 >>> 24;
     $292 = $206 >>> $291;
     $293 = $274 >>> 24;
     $294 = $292 << $293;
     $295 = $290 | $294;
     HEAP32[$dst$1>>2] = $295;
     break;
    }
    default: {
    }
    }
    $296 = ((($src$1)) + 1|0);
    $297 = (($dst$1) + ($21)|0);
    $dst$2 = $297;$n$1 = $n$0;$src$2 = $296;
    label = 15;
    continue;
   }
   else if ((label|0) == 15) {
    label = 0;
    $298 = HEAP8[$src$2>>0]|0;
    $299 = $298&255;
    $300 = (($15) + ($299<<2)|0);
    $301 = HEAP8[$300>>0]|0;
    $302 = $301&255;
    $303 = (((($15) + ($299<<2)|0)) + 1|0);
    $304 = HEAP8[$303>>0]|0;
    $305 = $304&255;
    $306 = (((($15) + ($299<<2)|0)) + 2|0);
    $307 = HEAP8[$306>>0]|0;
    $308 = $307&255;
    switch ($21|0) {
    case 1:  {
     $309 = HEAP8[$dst$2>>0]|0;
     $310 = $309&255;
     $311 = HEAP32[$33>>2]|0;
     $312 = $310 & $311;
     $313 = HEAP32[$30>>2]|0;
     $314 = $313 & 255;
     $315 = $312 >>> $314;
     $316 = HEAP32[$29>>2]|0;
     $317 = $316 & 255;
     $318 = (125924 + ($317<<2)|0);
     $319 = HEAP32[$318>>2]|0;
     $320 = (($319) + ($315)|0);
     $321 = HEAP8[$320>>0]|0;
     $322 = HEAP32[$34>>2]|0;
     $323 = $322 & $310;
     $324 = $313 >>> 8;
     $325 = $324 & 255;
     $326 = $323 >>> $325;
     $327 = $316 >>> 8;
     $328 = $327 & 255;
     $329 = (125924 + ($328<<2)|0);
     $330 = HEAP32[$329>>2]|0;
     $331 = (($330) + ($326)|0);
     $332 = HEAP8[$331>>0]|0;
     $333 = HEAP32[$35>>2]|0;
     $334 = $333 & $310;
     $335 = $313 >>> 16;
     $336 = $335 & 255;
     $337 = $334 >>> $336;
     $338 = $316 >>> 16;
     $339 = $338 & 255;
     $340 = (125924 + ($339<<2)|0);
     $341 = HEAP32[$340>>2]|0;
     $342 = (($341) + ($337)|0);
     $343 = HEAP8[$342>>0]|0;
     $344 = HEAP32[$36>>2]|0;
     $345 = $344 & $310;
     $346 = $313 >>> 24;
     $347 = $345 >>> $346;
     $348 = $316 >>> 24;
     $349 = (125924 + ($348<<2)|0);
     $350 = HEAP32[$349>>2]|0;
     $351 = (($350) + ($347)|0);
     $352 = HEAP8[$351>>0]|0;
     $dA$1$shrunk = $352;$dB$1$shrunk = $343;$dG$1$shrunk = $332;$dR$1$shrunk = $321;
     break;
    }
    case 2:  {
     $353 = HEAP16[$dst$2>>1]|0;
     $354 = $353&65535;
     $355 = HEAP32[$33>>2]|0;
     $356 = $354 & $355;
     $357 = HEAP32[$30>>2]|0;
     $358 = $357 & 255;
     $359 = $356 >>> $358;
     $360 = HEAP32[$29>>2]|0;
     $361 = $360 & 255;
     $362 = (125924 + ($361<<2)|0);
     $363 = HEAP32[$362>>2]|0;
     $364 = (($363) + ($359)|0);
     $365 = HEAP8[$364>>0]|0;
     $366 = HEAP32[$34>>2]|0;
     $367 = $366 & $354;
     $368 = $357 >>> 8;
     $369 = $368 & 255;
     $370 = $367 >>> $369;
     $371 = $360 >>> 8;
     $372 = $371 & 255;
     $373 = (125924 + ($372<<2)|0);
     $374 = HEAP32[$373>>2]|0;
     $375 = (($374) + ($370)|0);
     $376 = HEAP8[$375>>0]|0;
     $377 = HEAP32[$35>>2]|0;
     $378 = $377 & $354;
     $379 = $357 >>> 16;
     $380 = $379 & 255;
     $381 = $378 >>> $380;
     $382 = $360 >>> 16;
     $383 = $382 & 255;
     $384 = (125924 + ($383<<2)|0);
     $385 = HEAP32[$384>>2]|0;
     $386 = (($385) + ($381)|0);
     $387 = HEAP8[$386>>0]|0;
     $388 = HEAP32[$36>>2]|0;
     $389 = $388 & $354;
     $390 = $357 >>> 24;
     $391 = $389 >>> $390;
     $392 = $360 >>> 24;
     $393 = (125924 + ($392<<2)|0);
     $394 = HEAP32[$393>>2]|0;
     $395 = (($394) + ($391)|0);
     $396 = HEAP8[$395>>0]|0;
     $dA$1$shrunk = $396;$dB$1$shrunk = $387;$dG$1$shrunk = $376;$dR$1$shrunk = $365;
     break;
    }
    case 3:  {
     $397 = HEAP32[$30>>2]|0;
     $div9 = $397 >>> 3;
     $398 = $div9 & 31;
     $399 = (($dst$2) + ($398)|0);
     $400 = HEAP8[$399>>0]|0;
     $401 = $397 >>> 11;
     $div10 = $401 & 31;
     $402 = (($dst$2) + ($div10)|0);
     $403 = HEAP8[$402>>0]|0;
     $404 = $397 >>> 19;
     $405 = $404 & 31;
     $406 = (($dst$2) + ($405)|0);
     $407 = HEAP8[$406>>0]|0;
     $dA$1$shrunk = -1;$dB$1$shrunk = $407;$dG$1$shrunk = $403;$dR$1$shrunk = $400;
     break;
    }
    case 4:  {
     $408 = HEAP32[$dst$2>>2]|0;
     $409 = HEAP32[$33>>2]|0;
     $410 = $409 & $408;
     $411 = HEAP32[$30>>2]|0;
     $412 = $411 & 255;
     $413 = $410 >>> $412;
     $414 = HEAP32[$29>>2]|0;
     $415 = $414 & 255;
     $416 = (125924 + ($415<<2)|0);
     $417 = HEAP32[$416>>2]|0;
     $418 = (($417) + ($413)|0);
     $419 = HEAP8[$418>>0]|0;
     $420 = HEAP32[$34>>2]|0;
     $421 = $420 & $408;
     $422 = $411 >>> 8;
     $423 = $422 & 255;
     $424 = $421 >>> $423;
     $425 = $414 >>> 8;
     $426 = $425 & 255;
     $427 = (125924 + ($426<<2)|0);
     $428 = HEAP32[$427>>2]|0;
     $429 = (($428) + ($424)|0);
     $430 = HEAP8[$429>>0]|0;
     $431 = HEAP32[$35>>2]|0;
     $432 = $431 & $408;
     $433 = $411 >>> 16;
     $434 = $433 & 255;
     $435 = $432 >>> $434;
     $436 = $414 >>> 16;
     $437 = $436 & 255;
     $438 = (125924 + ($437<<2)|0);
     $439 = HEAP32[$438>>2]|0;
     $440 = (($439) + ($435)|0);
     $441 = HEAP8[$440>>0]|0;
     $442 = HEAP32[$36>>2]|0;
     $443 = $442 & $408;
     $444 = $411 >>> 24;
     $445 = $443 >>> $444;
     $446 = $414 >>> 24;
     $447 = (125924 + ($446<<2)|0);
     $448 = HEAP32[$447>>2]|0;
     $449 = (($448) + ($445)|0);
     $450 = HEAP8[$449>>0]|0;
     $dA$1$shrunk = $450;$dB$1$shrunk = $441;$dG$1$shrunk = $430;$dR$1$shrunk = $419;
     break;
    }
    default: {
     $dA$1$shrunk = 0;$dB$1$shrunk = 0;$dG$1$shrunk = 0;$dR$1$shrunk = 0;
    }
    }
    $dR$1 = $dR$1$shrunk&255;
    $dG$1 = $dG$1$shrunk&255;
    $dB$1 = $dB$1$shrunk&255;
    $dA$1 = $dA$1$shrunk&255;
    $451 = (($302) - ($dR$1))|0;
    $452 = Math_imul($451, $18)|0;
    $453 = (($452>>>0) / 255)&-1;
    $454 = (($453) + ($dR$1))|0;
    $455 = (($305) - ($dG$1))|0;
    $456 = Math_imul($455, $18)|0;
    $457 = (($456>>>0) / 255)&-1;
    $458 = (($457) + ($dG$1))|0;
    $459 = (($308) - ($dB$1))|0;
    $460 = Math_imul($459, $18)|0;
    $461 = (($460>>>0) / 255)&-1;
    $462 = (($461) + ($dB$1))|0;
    $463 = (($dA$1) + ($18))|0;
    $464 = Math_imul($dA$1, $18)|0;
    $465 = (($464>>>0) / 255)&-1;
    $466 = (($463) - ($465))|0;
    switch ($21|0) {
    case 1:  {
     $467 = HEAP32[$29>>2]|0;
     $468 = $467 & 255;
     $469 = $454 >>> $468;
     $470 = HEAP32[$30>>2]|0;
     $471 = $470 & 255;
     $472 = $469 << $471;
     $473 = $467 >>> 8;
     $474 = $473 & 255;
     $475 = $458 >>> $474;
     $476 = $470 >>> 8;
     $477 = $476 & 255;
     $478 = $475 << $477;
     $479 = $478 | $472;
     $480 = $467 >>> 16;
     $481 = $480 & 255;
     $482 = $462 >>> $481;
     $483 = $470 >>> 16;
     $484 = $483 & 255;
     $485 = $482 << $484;
     $486 = $479 | $485;
     $487 = $467 >>> 24;
     $488 = $466 >>> $487;
     $489 = $470 >>> 24;
     $490 = $488 << $489;
     $491 = $486 | $490;
     $492 = $491&255;
     HEAP8[$dst$2>>0] = $492;
     break;
    }
    case 2:  {
     $493 = HEAP32[$29>>2]|0;
     $494 = $493 & 255;
     $495 = $454 >>> $494;
     $496 = HEAP32[$30>>2]|0;
     $497 = $496 & 255;
     $498 = $495 << $497;
     $499 = $493 >>> 8;
     $500 = $499 & 255;
     $501 = $458 >>> $500;
     $502 = $496 >>> 8;
     $503 = $502 & 255;
     $504 = $501 << $503;
     $505 = $504 | $498;
     $506 = $493 >>> 16;
     $507 = $506 & 255;
     $508 = $462 >>> $507;
     $509 = $496 >>> 16;
     $510 = $509 & 255;
     $511 = $508 << $510;
     $512 = $505 | $511;
     $513 = $493 >>> 24;
     $514 = $466 >>> $513;
     $515 = $496 >>> 24;
     $516 = $514 << $515;
     $517 = $512 | $516;
     $518 = $517&65535;
     HEAP16[$dst$2>>1] = $518;
     break;
    }
    case 3:  {
     $519 = $454&255;
     $520 = HEAP8[$30>>0]|0;
     $div6 = ($520&255) >>> 3;
     $521 = $div6&255;
     $522 = (($dst$2) + ($521)|0);
     HEAP8[$522>>0] = $519;
     $523 = $458&255;
     $524 = HEAP8[$31>>0]|0;
     $div7 = ($524&255) >>> 3;
     $525 = $div7&255;
     $526 = (($dst$2) + ($525)|0);
     HEAP8[$526>>0] = $523;
     $527 = $462&255;
     $528 = HEAP8[$32>>0]|0;
     $div8 = ($528&255) >>> 3;
     $529 = $div8&255;
     $530 = (($dst$2) + ($529)|0);
     HEAP8[$530>>0] = $527;
     break;
    }
    case 4:  {
     $531 = HEAP32[$29>>2]|0;
     $532 = $531 & 255;
     $533 = $454 >>> $532;
     $534 = HEAP32[$30>>2]|0;
     $535 = $534 & 255;
     $536 = $533 << $535;
     $537 = $531 >>> 8;
     $538 = $537 & 255;
     $539 = $458 >>> $538;
     $540 = $534 >>> 8;
     $541 = $540 & 255;
     $542 = $539 << $541;
     $543 = $542 | $536;
     $544 = $531 >>> 16;
     $545 = $544 & 255;
     $546 = $462 >>> $545;
     $547 = $534 >>> 16;
     $548 = $547 & 255;
     $549 = $546 << $548;
     $550 = $543 | $549;
     $551 = $531 >>> 24;
     $552 = $466 >>> $551;
     $553 = $534 >>> 24;
     $554 = $552 << $553;
     $555 = $550 | $554;
     HEAP32[$dst$2>>2] = $555;
     break;
    }
    default: {
    }
    }
    $556 = ((($src$2)) + 1|0);
    $557 = (($dst$2) + ($21)|0);
    $dst$3 = $557;$n$2 = $n$1;$src$3 = $556;
    label = 26;
    continue;
   }
   else if ((label|0) == 26) {
    label = 0;
    $558 = HEAP8[$src$3>>0]|0;
    $559 = $558&255;
    $560 = (($15) + ($559<<2)|0);
    $561 = HEAP8[$560>>0]|0;
    $562 = $561&255;
    $563 = (((($15) + ($559<<2)|0)) + 1|0);
    $564 = HEAP8[$563>>0]|0;
    $565 = $564&255;
    $566 = (((($15) + ($559<<2)|0)) + 2|0);
    $567 = HEAP8[$566>>0]|0;
    $568 = $567&255;
    switch ($21|0) {
    case 1:  {
     $569 = HEAP8[$dst$3>>0]|0;
     $570 = $569&255;
     $571 = HEAP32[$33>>2]|0;
     $572 = $570 & $571;
     $573 = HEAP32[$30>>2]|0;
     $574 = $573 & 255;
     $575 = $572 >>> $574;
     $576 = HEAP32[$29>>2]|0;
     $577 = $576 & 255;
     $578 = (125924 + ($577<<2)|0);
     $579 = HEAP32[$578>>2]|0;
     $580 = (($579) + ($575)|0);
     $581 = HEAP8[$580>>0]|0;
     $582 = HEAP32[$34>>2]|0;
     $583 = $582 & $570;
     $584 = $573 >>> 8;
     $585 = $584 & 255;
     $586 = $583 >>> $585;
     $587 = $576 >>> 8;
     $588 = $587 & 255;
     $589 = (125924 + ($588<<2)|0);
     $590 = HEAP32[$589>>2]|0;
     $591 = (($590) + ($586)|0);
     $592 = HEAP8[$591>>0]|0;
     $593 = HEAP32[$35>>2]|0;
     $594 = $593 & $570;
     $595 = $573 >>> 16;
     $596 = $595 & 255;
     $597 = $594 >>> $596;
     $598 = $576 >>> 16;
     $599 = $598 & 255;
     $600 = (125924 + ($599<<2)|0);
     $601 = HEAP32[$600>>2]|0;
     $602 = (($601) + ($597)|0);
     $603 = HEAP8[$602>>0]|0;
     $604 = HEAP32[$36>>2]|0;
     $605 = $604 & $570;
     $606 = $573 >>> 24;
     $607 = $605 >>> $606;
     $608 = $576 >>> 24;
     $609 = (125924 + ($608<<2)|0);
     $610 = HEAP32[$609>>2]|0;
     $611 = (($610) + ($607)|0);
     $612 = HEAP8[$611>>0]|0;
     $dA$2$shrunk = $612;$dB$2$shrunk = $603;$dG$2$shrunk = $592;$dR$2$shrunk = $581;
     break;
    }
    case 2:  {
     $613 = HEAP16[$dst$3>>1]|0;
     $614 = $613&65535;
     $615 = HEAP32[$33>>2]|0;
     $616 = $614 & $615;
     $617 = HEAP32[$30>>2]|0;
     $618 = $617 & 255;
     $619 = $616 >>> $618;
     $620 = HEAP32[$29>>2]|0;
     $621 = $620 & 255;
     $622 = (125924 + ($621<<2)|0);
     $623 = HEAP32[$622>>2]|0;
     $624 = (($623) + ($619)|0);
     $625 = HEAP8[$624>>0]|0;
     $626 = HEAP32[$34>>2]|0;
     $627 = $626 & $614;
     $628 = $617 >>> 8;
     $629 = $628 & 255;
     $630 = $627 >>> $629;
     $631 = $620 >>> 8;
     $632 = $631 & 255;
     $633 = (125924 + ($632<<2)|0);
     $634 = HEAP32[$633>>2]|0;
     $635 = (($634) + ($630)|0);
     $636 = HEAP8[$635>>0]|0;
     $637 = HEAP32[$35>>2]|0;
     $638 = $637 & $614;
     $639 = $617 >>> 16;
     $640 = $639 & 255;
     $641 = $638 >>> $640;
     $642 = $620 >>> 16;
     $643 = $642 & 255;
     $644 = (125924 + ($643<<2)|0);
     $645 = HEAP32[$644>>2]|0;
     $646 = (($645) + ($641)|0);
     $647 = HEAP8[$646>>0]|0;
     $648 = HEAP32[$36>>2]|0;
     $649 = $648 & $614;
     $650 = $617 >>> 24;
     $651 = $649 >>> $650;
     $652 = $620 >>> 24;
     $653 = (125924 + ($652<<2)|0);
     $654 = HEAP32[$653>>2]|0;
     $655 = (($654) + ($651)|0);
     $656 = HEAP8[$655>>0]|0;
     $dA$2$shrunk = $656;$dB$2$shrunk = $647;$dG$2$shrunk = $636;$dR$2$shrunk = $625;
     break;
    }
    case 3:  {
     $657 = HEAP32[$30>>2]|0;
     $div3 = $657 >>> 3;
     $658 = $div3 & 31;
     $659 = (($dst$3) + ($658)|0);
     $660 = HEAP8[$659>>0]|0;
     $661 = $657 >>> 11;
     $div4 = $661 & 31;
     $662 = (($dst$3) + ($div4)|0);
     $663 = HEAP8[$662>>0]|0;
     $664 = $657 >>> 19;
     $665 = $664 & 31;
     $666 = (($dst$3) + ($665)|0);
     $667 = HEAP8[$666>>0]|0;
     $dA$2$shrunk = -1;$dB$2$shrunk = $667;$dG$2$shrunk = $663;$dR$2$shrunk = $660;
     break;
    }
    case 4:  {
     $668 = HEAP32[$dst$3>>2]|0;
     $669 = HEAP32[$33>>2]|0;
     $670 = $669 & $668;
     $671 = HEAP32[$30>>2]|0;
     $672 = $671 & 255;
     $673 = $670 >>> $672;
     $674 = HEAP32[$29>>2]|0;
     $675 = $674 & 255;
     $676 = (125924 + ($675<<2)|0);
     $677 = HEAP32[$676>>2]|0;
     $678 = (($677) + ($673)|0);
     $679 = HEAP8[$678>>0]|0;
     $680 = HEAP32[$34>>2]|0;
     $681 = $680 & $668;
     $682 = $671 >>> 8;
     $683 = $682 & 255;
     $684 = $681 >>> $683;
     $685 = $674 >>> 8;
     $686 = $685 & 255;
     $687 = (125924 + ($686<<2)|0);
     $688 = HEAP32[$687>>2]|0;
     $689 = (($688) + ($684)|0);
     $690 = HEAP8[$689>>0]|0;
     $691 = HEAP32[$35>>2]|0;
     $692 = $691 & $668;
     $693 = $671 >>> 16;
     $694 = $693 & 255;
     $695 = $692 >>> $694;
     $696 = $674 >>> 16;
     $697 = $696 & 255;
     $698 = (125924 + ($697<<2)|0);
     $699 = HEAP32[$698>>2]|0;
     $700 = (($699) + ($695)|0);
     $701 = HEAP8[$700>>0]|0;
     $702 = HEAP32[$36>>2]|0;
     $703 = $702 & $668;
     $704 = $671 >>> 24;
     $705 = $703 >>> $704;
     $706 = $674 >>> 24;
     $707 = (125924 + ($706<<2)|0);
     $708 = HEAP32[$707>>2]|0;
     $709 = (($708) + ($705)|0);
     $710 = HEAP8[$709>>0]|0;
     $dA$2$shrunk = $710;$dB$2$shrunk = $701;$dG$2$shrunk = $690;$dR$2$shrunk = $679;
     break;
    }
    default: {
     $dA$2$shrunk = 0;$dB$2$shrunk = 0;$dG$2$shrunk = 0;$dR$2$shrunk = 0;
    }
    }
    $dR$2 = $dR$2$shrunk&255;
    $dG$2 = $dG$2$shrunk&255;
    $dB$2 = $dB$2$shrunk&255;
    $dA$2 = $dA$2$shrunk&255;
    $711 = (($562) - ($dR$2))|0;
    $712 = Math_imul($711, $18)|0;
    $713 = (($712>>>0) / 255)&-1;
    $714 = (($713) + ($dR$2))|0;
    $715 = (($565) - ($dG$2))|0;
    $716 = Math_imul($715, $18)|0;
    $717 = (($716>>>0) / 255)&-1;
    $718 = (($717) + ($dG$2))|0;
    $719 = (($568) - ($dB$2))|0;
    $720 = Math_imul($719, $18)|0;
    $721 = (($720>>>0) / 255)&-1;
    $722 = (($721) + ($dB$2))|0;
    $723 = (($dA$2) + ($18))|0;
    $724 = Math_imul($dA$2, $18)|0;
    $725 = (($724>>>0) / 255)&-1;
    $726 = (($723) - ($725))|0;
    switch ($21|0) {
    case 1:  {
     $727 = HEAP32[$29>>2]|0;
     $728 = $727 & 255;
     $729 = $714 >>> $728;
     $730 = HEAP32[$30>>2]|0;
     $731 = $730 & 255;
     $732 = $729 << $731;
     $733 = $727 >>> 8;
     $734 = $733 & 255;
     $735 = $718 >>> $734;
     $736 = $730 >>> 8;
     $737 = $736 & 255;
     $738 = $735 << $737;
     $739 = $738 | $732;
     $740 = $727 >>> 16;
     $741 = $740 & 255;
     $742 = $722 >>> $741;
     $743 = $730 >>> 16;
     $744 = $743 & 255;
     $745 = $742 << $744;
     $746 = $739 | $745;
     $747 = $727 >>> 24;
     $748 = $726 >>> $747;
     $749 = $730 >>> 24;
     $750 = $748 << $749;
     $751 = $746 | $750;
     $752 = $751&255;
     HEAP8[$dst$3>>0] = $752;
     break;
    }
    case 2:  {
     $753 = HEAP32[$29>>2]|0;
     $754 = $753 & 255;
     $755 = $714 >>> $754;
     $756 = HEAP32[$30>>2]|0;
     $757 = $756 & 255;
     $758 = $755 << $757;
     $759 = $753 >>> 8;
     $760 = $759 & 255;
     $761 = $718 >>> $760;
     $762 = $756 >>> 8;
     $763 = $762 & 255;
     $764 = $761 << $763;
     $765 = $764 | $758;
     $766 = $753 >>> 16;
     $767 = $766 & 255;
     $768 = $722 >>> $767;
     $769 = $756 >>> 16;
     $770 = $769 & 255;
     $771 = $768 << $770;
     $772 = $765 | $771;
     $773 = $753 >>> 24;
     $774 = $726 >>> $773;
     $775 = $756 >>> 24;
     $776 = $774 << $775;
     $777 = $772 | $776;
     $778 = $777&65535;
     HEAP16[$dst$3>>1] = $778;
     break;
    }
    case 3:  {
     $779 = $714&255;
     $780 = HEAP8[$30>>0]|0;
     $div = ($780&255) >>> 3;
     $781 = $div&255;
     $782 = (($dst$3) + ($781)|0);
     HEAP8[$782>>0] = $779;
     $783 = $718&255;
     $784 = HEAP8[$31>>0]|0;
     $div1 = ($784&255) >>> 3;
     $785 = $div1&255;
     $786 = (($dst$3) + ($785)|0);
     HEAP8[$786>>0] = $783;
     $787 = $722&255;
     $788 = HEAP8[$32>>0]|0;
     $div2 = ($788&255) >>> 3;
     $789 = $div2&255;
     $790 = (($dst$3) + ($789)|0);
     HEAP8[$790>>0] = $787;
     break;
    }
    case 4:  {
     $791 = HEAP32[$29>>2]|0;
     $792 = $791 & 255;
     $793 = $714 >>> $792;
     $794 = HEAP32[$30>>2]|0;
     $795 = $794 & 255;
     $796 = $793 << $795;
     $797 = $791 >>> 8;
     $798 = $797 & 255;
     $799 = $718 >>> $798;
     $800 = $794 >>> 8;
     $801 = $800 & 255;
     $802 = $799 << $801;
     $803 = $802 | $796;
     $804 = $791 >>> 16;
     $805 = $804 & 255;
     $806 = $722 >>> $805;
     $807 = $794 >>> 16;
     $808 = $807 & 255;
     $809 = $806 << $808;
     $810 = $803 | $809;
     $811 = $791 >>> 24;
     $812 = $726 >>> $811;
     $813 = $794 >>> 24;
     $814 = $812 << $813;
     $815 = $810 | $814;
     HEAP32[$dst$3>>2] = $815;
     break;
    }
    default: {
    }
    }
    $816 = ((($src$3)) + 1|0);
    $817 = (($dst$3) + ($21)|0);
    $dst$4 = $817;$n$3 = $n$2;$src$4 = $816;
    label = 37;
    continue;
   }
   else if ((label|0) == 37) {
    label = 0;
    $818 = HEAP8[$src$4>>0]|0;
    $819 = $818&255;
    $820 = (($15) + ($819<<2)|0);
    $821 = HEAP8[$820>>0]|0;
    $822 = $821&255;
    $823 = (((($15) + ($819<<2)|0)) + 1|0);
    $824 = HEAP8[$823>>0]|0;
    $825 = $824&255;
    $826 = (((($15) + ($819<<2)|0)) + 2|0);
    $827 = HEAP8[$826>>0]|0;
    $828 = $827&255;
    switch ($21|0) {
    case 1:  {
     $829 = HEAP8[$dst$4>>0]|0;
     $830 = $829&255;
     $831 = HEAP32[$33>>2]|0;
     $832 = $830 & $831;
     $833 = HEAP32[$30>>2]|0;
     $834 = $833 & 255;
     $835 = $832 >>> $834;
     $836 = HEAP32[$29>>2]|0;
     $837 = $836 & 255;
     $838 = (125924 + ($837<<2)|0);
     $839 = HEAP32[$838>>2]|0;
     $840 = (($839) + ($835)|0);
     $841 = HEAP8[$840>>0]|0;
     $842 = HEAP32[$34>>2]|0;
     $843 = $842 & $830;
     $844 = $833 >>> 8;
     $845 = $844 & 255;
     $846 = $843 >>> $845;
     $847 = $836 >>> 8;
     $848 = $847 & 255;
     $849 = (125924 + ($848<<2)|0);
     $850 = HEAP32[$849>>2]|0;
     $851 = (($850) + ($846)|0);
     $852 = HEAP8[$851>>0]|0;
     $853 = HEAP32[$35>>2]|0;
     $854 = $853 & $830;
     $855 = $833 >>> 16;
     $856 = $855 & 255;
     $857 = $854 >>> $856;
     $858 = $836 >>> 16;
     $859 = $858 & 255;
     $860 = (125924 + ($859<<2)|0);
     $861 = HEAP32[$860>>2]|0;
     $862 = (($861) + ($857)|0);
     $863 = HEAP8[$862>>0]|0;
     $864 = HEAP32[$36>>2]|0;
     $865 = $864 & $830;
     $866 = $833 >>> 24;
     $867 = $865 >>> $866;
     $868 = $836 >>> 24;
     $869 = (125924 + ($868<<2)|0);
     $870 = HEAP32[$869>>2]|0;
     $871 = (($870) + ($867)|0);
     $872 = HEAP8[$871>>0]|0;
     $dA$3$shrunk = $872;$dB$3$shrunk = $863;$dG$3$shrunk = $852;$dR$3$shrunk = $841;
     break;
    }
    case 2:  {
     $873 = HEAP16[$dst$4>>1]|0;
     $874 = $873&65535;
     $875 = HEAP32[$33>>2]|0;
     $876 = $874 & $875;
     $877 = HEAP32[$30>>2]|0;
     $878 = $877 & 255;
     $879 = $876 >>> $878;
     $880 = HEAP32[$29>>2]|0;
     $881 = $880 & 255;
     $882 = (125924 + ($881<<2)|0);
     $883 = HEAP32[$882>>2]|0;
     $884 = (($883) + ($879)|0);
     $885 = HEAP8[$884>>0]|0;
     $886 = HEAP32[$34>>2]|0;
     $887 = $886 & $874;
     $888 = $877 >>> 8;
     $889 = $888 & 255;
     $890 = $887 >>> $889;
     $891 = $880 >>> 8;
     $892 = $891 & 255;
     $893 = (125924 + ($892<<2)|0);
     $894 = HEAP32[$893>>2]|0;
     $895 = (($894) + ($890)|0);
     $896 = HEAP8[$895>>0]|0;
     $897 = HEAP32[$35>>2]|0;
     $898 = $897 & $874;
     $899 = $877 >>> 16;
     $900 = $899 & 255;
     $901 = $898 >>> $900;
     $902 = $880 >>> 16;
     $903 = $902 & 255;
     $904 = (125924 + ($903<<2)|0);
     $905 = HEAP32[$904>>2]|0;
     $906 = (($905) + ($901)|0);
     $907 = HEAP8[$906>>0]|0;
     $908 = HEAP32[$36>>2]|0;
     $909 = $908 & $874;
     $910 = $877 >>> 24;
     $911 = $909 >>> $910;
     $912 = $880 >>> 24;
     $913 = (125924 + ($912<<2)|0);
     $914 = HEAP32[$913>>2]|0;
     $915 = (($914) + ($911)|0);
     $916 = HEAP8[$915>>0]|0;
     $dA$3$shrunk = $916;$dB$3$shrunk = $907;$dG$3$shrunk = $896;$dR$3$shrunk = $885;
     break;
    }
    case 3:  {
     $917 = HEAP32[$30>>2]|0;
     $div21 = $917 >>> 3;
     $918 = $div21 & 31;
     $919 = (($dst$4) + ($918)|0);
     $920 = HEAP8[$919>>0]|0;
     $921 = $917 >>> 11;
     $div22 = $921 & 31;
     $922 = (($dst$4) + ($div22)|0);
     $923 = HEAP8[$922>>0]|0;
     $924 = $917 >>> 19;
     $925 = $924 & 31;
     $926 = (($dst$4) + ($925)|0);
     $927 = HEAP8[$926>>0]|0;
     $dA$3$shrunk = -1;$dB$3$shrunk = $927;$dG$3$shrunk = $923;$dR$3$shrunk = $920;
     break;
    }
    case 4:  {
     $928 = HEAP32[$dst$4>>2]|0;
     $929 = HEAP32[$33>>2]|0;
     $930 = $929 & $928;
     $931 = HEAP32[$30>>2]|0;
     $932 = $931 & 255;
     $933 = $930 >>> $932;
     $934 = HEAP32[$29>>2]|0;
     $935 = $934 & 255;
     $936 = (125924 + ($935<<2)|0);
     $937 = HEAP32[$936>>2]|0;
     $938 = (($937) + ($933)|0);
     $939 = HEAP8[$938>>0]|0;
     $940 = HEAP32[$34>>2]|0;
     $941 = $940 & $928;
     $942 = $931 >>> 8;
     $943 = $942 & 255;
     $944 = $941 >>> $943;
     $945 = $934 >>> 8;
     $946 = $945 & 255;
     $947 = (125924 + ($946<<2)|0);
     $948 = HEAP32[$947>>2]|0;
     $949 = (($948) + ($944)|0);
     $950 = HEAP8[$949>>0]|0;
     $951 = HEAP32[$35>>2]|0;
     $952 = $951 & $928;
     $953 = $931 >>> 16;
     $954 = $953 & 255;
     $955 = $952 >>> $954;
     $956 = $934 >>> 16;
     $957 = $956 & 255;
     $958 = (125924 + ($957<<2)|0);
     $959 = HEAP32[$958>>2]|0;
     $960 = (($959) + ($955)|0);
     $961 = HEAP8[$960>>0]|0;
     $962 = HEAP32[$36>>2]|0;
     $963 = $962 & $928;
     $964 = $931 >>> 24;
     $965 = $963 >>> $964;
     $966 = $934 >>> 24;
     $967 = (125924 + ($966<<2)|0);
     $968 = HEAP32[$967>>2]|0;
     $969 = (($968) + ($965)|0);
     $970 = HEAP8[$969>>0]|0;
     $dA$3$shrunk = $970;$dB$3$shrunk = $961;$dG$3$shrunk = $950;$dR$3$shrunk = $939;
     break;
    }
    default: {
     $dA$3$shrunk = 0;$dB$3$shrunk = 0;$dG$3$shrunk = 0;$dR$3$shrunk = 0;
    }
    }
    $dR$3 = $dR$3$shrunk&255;
    $dG$3 = $dG$3$shrunk&255;
    $dB$3 = $dB$3$shrunk&255;
    $dA$3 = $dA$3$shrunk&255;
    $971 = (($822) - ($dR$3))|0;
    $972 = Math_imul($971, $18)|0;
    $973 = (($972>>>0) / 255)&-1;
    $974 = (($973) + ($dR$3))|0;
    $975 = (($825) - ($dG$3))|0;
    $976 = Math_imul($975, $18)|0;
    $977 = (($976>>>0) / 255)&-1;
    $978 = (($977) + ($dG$3))|0;
    $979 = (($828) - ($dB$3))|0;
    $980 = Math_imul($979, $18)|0;
    $981 = (($980>>>0) / 255)&-1;
    $982 = (($981) + ($dB$3))|0;
    $983 = (($dA$3) + ($18))|0;
    $984 = Math_imul($dA$3, $18)|0;
    $985 = (($984>>>0) / 255)&-1;
    $986 = (($983) - ($985))|0;
    switch ($21|0) {
    case 1:  {
     $987 = HEAP32[$29>>2]|0;
     $988 = $987 & 255;
     $989 = $974 >>> $988;
     $990 = HEAP32[$30>>2]|0;
     $991 = $990 & 255;
     $992 = $989 << $991;
     $993 = $987 >>> 8;
     $994 = $993 & 255;
     $995 = $978 >>> $994;
     $996 = $990 >>> 8;
     $997 = $996 & 255;
     $998 = $995 << $997;
     $999 = $998 | $992;
     $1000 = $987 >>> 16;
     $1001 = $1000 & 255;
     $1002 = $982 >>> $1001;
     $1003 = $990 >>> 16;
     $1004 = $1003 & 255;
     $1005 = $1002 << $1004;
     $1006 = $999 | $1005;
     $1007 = $987 >>> 24;
     $1008 = $986 >>> $1007;
     $1009 = $990 >>> 24;
     $1010 = $1008 << $1009;
     $1011 = $1006 | $1010;
     $1012 = $1011&255;
     HEAP8[$dst$4>>0] = $1012;
     break;
    }
    case 2:  {
     $1013 = HEAP32[$29>>2]|0;
     $1014 = $1013 & 255;
     $1015 = $974 >>> $1014;
     $1016 = HEAP32[$30>>2]|0;
     $1017 = $1016 & 255;
     $1018 = $1015 << $1017;
     $1019 = $1013 >>> 8;
     $1020 = $1019 & 255;
     $1021 = $978 >>> $1020;
     $1022 = $1016 >>> 8;
     $1023 = $1022 & 255;
     $1024 = $1021 << $1023;
     $1025 = $1024 | $1018;
     $1026 = $1013 >>> 16;
     $1027 = $1026 & 255;
     $1028 = $982 >>> $1027;
     $1029 = $1016 >>> 16;
     $1030 = $1029 & 255;
     $1031 = $1028 << $1030;
     $1032 = $1025 | $1031;
     $1033 = $1013 >>> 24;
     $1034 = $986 >>> $1033;
     $1035 = $1016 >>> 24;
     $1036 = $1034 << $1035;
     $1037 = $1032 | $1036;
     $1038 = $1037&65535;
     HEAP16[$dst$4>>1] = $1038;
     break;
    }
    case 3:  {
     $1039 = $974&255;
     $1040 = HEAP8[$30>>0]|0;
     $div18 = ($1040&255) >>> 3;
     $1041 = $div18&255;
     $1042 = (($dst$4) + ($1041)|0);
     HEAP8[$1042>>0] = $1039;
     $1043 = $978&255;
     $1044 = HEAP8[$31>>0]|0;
     $div19 = ($1044&255) >>> 3;
     $1045 = $div19&255;
     $1046 = (($dst$4) + ($1045)|0);
     HEAP8[$1046>>0] = $1043;
     $1047 = $982&255;
     $1048 = HEAP8[$32>>0]|0;
     $div20 = ($1048&255) >>> 3;
     $1049 = $div20&255;
     $1050 = (($dst$4) + ($1049)|0);
     HEAP8[$1050>>0] = $1047;
     break;
    }
    case 4:  {
     $1051 = HEAP32[$29>>2]|0;
     $1052 = $1051 & 255;
     $1053 = $974 >>> $1052;
     $1054 = HEAP32[$30>>2]|0;
     $1055 = $1054 & 255;
     $1056 = $1053 << $1055;
     $1057 = $1051 >>> 8;
     $1058 = $1057 & 255;
     $1059 = $978 >>> $1058;
     $1060 = $1054 >>> 8;
     $1061 = $1060 & 255;
     $1062 = $1059 << $1061;
     $1063 = $1062 | $1056;
     $1064 = $1051 >>> 16;
     $1065 = $1064 & 255;
     $1066 = $982 >>> $1065;
     $1067 = $1054 >>> 16;
     $1068 = $1067 & 255;
     $1069 = $1066 << $1068;
     $1070 = $1063 | $1069;
     $1071 = $1051 >>> 24;
     $1072 = $986 >>> $1071;
     $1073 = $1054 >>> 24;
     $1074 = $1072 << $1073;
     $1075 = $1070 | $1074;
     HEAP32[$dst$4>>2] = $1075;
     break;
    }
    default: {
    }
    }
    $1076 = ((($src$4)) + 1|0);
    $1077 = (($dst$4) + ($21)|0);
    $1078 = (($n$3) + -1)|0;
    $1079 = ($n$3|0)>(1);
    if ($1079) {
     $dst$1 = $1077;$n$0 = $1078;$src$1 = $1076;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $1080 = (($1076) + ($5)|0);
  $1081 = (($1077) + ($7)|0);
  $1082 = ($37|0)==(0);
  if ($1082) {
   label = 50;
   break;
  } else {
   $$in = $37;$dst$027 = $1081;$src$028 = $1080;
  }
 }
 if ((label|0) == 48) {
  // unreachable;
 }
 else if ((label|0) == 50) {
  return;
 }
}
function _Blit1toNAlphaKey($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0;
 var $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0;
 var $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0;
 var $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0;
 var $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0;
 var $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0;
 var $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0;
 var $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0;
 var $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0;
 var $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0;
 var $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0;
 var $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0;
 var $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0;
 var $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0;
 var $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0;
 var $1267 = 0, $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0;
 var $1285 = 0, $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0;
 var $1302 = 0, $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0;
 var $1320 = 0, $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0;
 var $1339 = 0, $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0;
 var $1357 = 0, $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0;
 var $1375 = 0, $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0;
 var $1393 = 0, $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0;
 var $1410 = 0, $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0;
 var $1429 = 0, $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0;
 var $1447 = 0, $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0;
 var $1465 = 0, $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0;
 var $1483 = 0, $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0;
 var $1500 = 0, $1501 = 0, $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0, $1511 = 0, $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0, $1518 = 0;
 var $1519 = 0, $152 = 0, $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0, $153 = 0, $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0, $1536 = 0;
 var $1537 = 0, $1538 = 0, $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0, $1548 = 0, $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0, $1554 = 0;
 var $1555 = 0, $1556 = 0, $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $1565 = 0, $1566 = 0, $1567 = 0, $1568 = 0, $1569 = 0, $157 = 0, $1570 = 0, $1571 = 0, $1572 = 0;
 var $1573 = 0, $1574 = 0, $1575 = 0, $1576 = 0, $1577 = 0, $1578 = 0, $1579 = 0, $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0, $1583 = 0, $1584 = 0, $1585 = 0, $1586 = 0, $1587 = 0, $1588 = 0, $1589 = 0, $159 = 0, $1590 = 0;
 var $1591 = 0, $1592 = 0, $1593 = 0, $1594 = 0, $1595 = 0, $1596 = 0, $1597 = 0, $1598 = 0, $1599 = 0, $16 = 0, $160 = 0, $1600 = 0, $1601 = 0, $1602 = 0, $1603 = 0, $1604 = 0, $1605 = 0, $1606 = 0, $1607 = 0, $1608 = 0;
 var $1609 = 0, $161 = 0, $1610 = 0, $1611 = 0, $1612 = 0, $1613 = 0, $1614 = 0, $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0, $1619 = 0, $162 = 0, $1620 = 0, $1621 = 0, $1622 = 0, $1623 = 0, $1624 = 0, $1625 = 0, $1626 = 0;
 var $1627 = 0, $1628 = 0, $1629 = 0, $163 = 0, $1630 = 0, $1631 = 0, $1632 = 0, $1633 = 0, $1634 = 0, $1635 = 0, $1636 = 0, $1637 = 0, $1638 = 0, $1639 = 0, $164 = 0, $1640 = 0, $1641 = 0, $1642 = 0, $1643 = 0, $1644 = 0;
 var $1645 = 0, $1646 = 0, $1647 = 0, $1648 = 0, $1649 = 0, $165 = 0, $1650 = 0, $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0, $1655 = 0, $1656 = 0, $1657 = 0, $1658 = 0, $1659 = 0, $166 = 0, $1660 = 0, $1661 = 0, $1662 = 0;
 var $1663 = 0, $1664 = 0, $1665 = 0, $1666 = 0, $1667 = 0, $1668 = 0, $1669 = 0, $167 = 0, $1670 = 0, $1671 = 0, $1672 = 0, $1673 = 0, $1674 = 0, $1675 = 0, $1676 = 0, $1677 = 0, $1678 = 0, $1679 = 0, $168 = 0, $1680 = 0;
 var $1681 = 0, $1682 = 0, $1683 = 0, $1684 = 0, $1685 = 0, $1686 = 0, $1687 = 0, $1688 = 0, $1689 = 0, $169 = 0, $1690 = 0, $1691 = 0, $1692 = 0, $1693 = 0, $1694 = 0, $1695 = 0, $1696 = 0, $1697 = 0, $1698 = 0, $1699 = 0;
 var $17 = 0, $170 = 0, $1700 = 0, $1701 = 0, $1702 = 0, $1703 = 0, $1704 = 0, $1705 = 0, $1706 = 0, $1707 = 0, $1708 = 0, $1709 = 0, $171 = 0, $1710 = 0, $1711 = 0, $1712 = 0, $1713 = 0, $1714 = 0, $1715 = 0, $1716 = 0;
 var $1717 = 0, $1718 = 0, $1719 = 0, $172 = 0, $1720 = 0, $1721 = 0, $1722 = 0, $1723 = 0, $1724 = 0, $1725 = 0, $1726 = 0, $1727 = 0, $1728 = 0, $1729 = 0, $173 = 0, $1730 = 0, $1731 = 0, $1732 = 0, $1733 = 0, $1734 = 0;
 var $1735 = 0, $1736 = 0, $1737 = 0, $1738 = 0, $1739 = 0, $174 = 0, $1740 = 0, $1741 = 0, $1742 = 0, $1743 = 0, $1744 = 0, $1745 = 0, $1746 = 0, $1747 = 0, $1748 = 0, $1749 = 0, $175 = 0, $1750 = 0, $1751 = 0, $1752 = 0;
 var $1753 = 0, $1754 = 0, $1755 = 0, $1756 = 0, $1757 = 0, $1758 = 0, $1759 = 0, $176 = 0, $1760 = 0, $1761 = 0, $1762 = 0, $1763 = 0, $1764 = 0, $1765 = 0, $1766 = 0, $1767 = 0, $1768 = 0, $1769 = 0, $177 = 0, $1770 = 0;
 var $1771 = 0, $1772 = 0, $1773 = 0, $1774 = 0, $1775 = 0, $1776 = 0, $1777 = 0, $1778 = 0, $1779 = 0, $178 = 0, $1780 = 0, $1781 = 0, $1782 = 0, $1783 = 0, $1784 = 0, $1785 = 0, $1786 = 0, $1787 = 0, $1788 = 0, $1789 = 0;
 var $179 = 0, $1790 = 0, $1791 = 0, $1792 = 0, $1793 = 0, $1794 = 0, $1795 = 0, $1796 = 0, $1797 = 0, $1798 = 0, $1799 = 0, $18 = 0, $180 = 0, $1800 = 0, $1801 = 0, $1802 = 0, $1803 = 0, $1804 = 0, $1805 = 0, $1806 = 0;
 var $1807 = 0, $1808 = 0, $1809 = 0, $181 = 0, $1810 = 0, $1811 = 0, $1812 = 0, $1813 = 0, $1814 = 0, $1815 = 0, $1816 = 0, $1817 = 0, $1818 = 0, $1819 = 0, $182 = 0, $1820 = 0, $1821 = 0, $1822 = 0, $1823 = 0, $1824 = 0;
 var $1825 = 0, $1826 = 0, $1827 = 0, $1828 = 0, $1829 = 0, $183 = 0, $1830 = 0, $1831 = 0, $1832 = 0, $1833 = 0, $1834 = 0, $1835 = 0, $1836 = 0, $1837 = 0, $1838 = 0, $1839 = 0, $184 = 0, $1840 = 0, $1841 = 0, $1842 = 0;
 var $1843 = 0, $1844 = 0, $1845 = 0, $1846 = 0, $1847 = 0, $1848 = 0, $1849 = 0, $185 = 0, $1850 = 0, $1851 = 0, $1852 = 0, $1853 = 0, $1854 = 0, $1855 = 0, $1856 = 0, $1857 = 0, $1858 = 0, $1859 = 0, $186 = 0, $1860 = 0;
 var $1861 = 0, $1862 = 0, $1863 = 0, $1864 = 0, $1865 = 0, $1866 = 0, $1867 = 0, $1868 = 0, $1869 = 0, $187 = 0, $1870 = 0, $1871 = 0, $1872 = 0, $1873 = 0, $1874 = 0, $1875 = 0, $1876 = 0, $1877 = 0, $1878 = 0, $1879 = 0;
 var $188 = 0, $1880 = 0, $1881 = 0, $1882 = 0, $1883 = 0, $1884 = 0, $1885 = 0, $1886 = 0, $1887 = 0, $1888 = 0, $1889 = 0, $189 = 0, $1890 = 0, $1891 = 0, $1892 = 0, $1893 = 0, $1894 = 0, $1895 = 0, $1896 = 0, $1897 = 0;
 var $1898 = 0, $1899 = 0, $19 = 0, $190 = 0, $1900 = 0, $1901 = 0, $1902 = 0, $1903 = 0, $1904 = 0, $1905 = 0, $1906 = 0, $1907 = 0, $1908 = 0, $1909 = 0, $191 = 0, $1910 = 0, $1911 = 0, $1912 = 0, $1913 = 0, $1914 = 0;
 var $1915 = 0, $1916 = 0, $1917 = 0, $1918 = 0, $1919 = 0, $192 = 0, $1920 = 0, $1921 = 0, $1922 = 0, $1923 = 0, $1924 = 0, $1925 = 0, $1926 = 0, $1927 = 0, $1928 = 0, $1929 = 0, $193 = 0, $1930 = 0, $1931 = 0, $1932 = 0;
 var $1933 = 0, $1934 = 0, $1935 = 0, $1936 = 0, $1937 = 0, $1938 = 0, $1939 = 0, $194 = 0, $1940 = 0, $1941 = 0, $1942 = 0, $1943 = 0, $1944 = 0, $1945 = 0, $1946 = 0, $1947 = 0, $1948 = 0, $1949 = 0, $195 = 0, $1950 = 0;
 var $1951 = 0, $1952 = 0, $1953 = 0, $1954 = 0, $1955 = 0, $1956 = 0, $1957 = 0, $1958 = 0, $1959 = 0, $196 = 0, $1960 = 0, $1961 = 0, $1962 = 0, $1963 = 0, $1964 = 0, $1965 = 0, $1966 = 0, $1967 = 0, $1968 = 0, $1969 = 0;
 var $197 = 0, $1970 = 0, $1971 = 0, $1972 = 0, $1973 = 0, $1974 = 0, $1975 = 0, $1976 = 0, $1977 = 0, $1978 = 0, $1979 = 0, $198 = 0, $1980 = 0, $1981 = 0, $1982 = 0, $1983 = 0, $1984 = 0, $1985 = 0, $1986 = 0, $1987 = 0;
 var $1988 = 0, $1989 = 0, $199 = 0, $1990 = 0, $1991 = 0, $1992 = 0, $1993 = 0, $1994 = 0, $1995 = 0, $1996 = 0, $1997 = 0, $1998 = 0, $1999 = 0, $2 = 0, $20 = 0, $200 = 0, $2000 = 0, $2001 = 0, $2002 = 0, $2003 = 0;
 var $2004 = 0, $2005 = 0, $2006 = 0, $2007 = 0, $2008 = 0, $2009 = 0, $201 = 0, $2010 = 0, $2011 = 0, $2012 = 0, $2013 = 0, $2014 = 0, $2015 = 0, $2016 = 0, $2017 = 0, $2018 = 0, $2019 = 0, $202 = 0, $2020 = 0, $2021 = 0;
 var $2022 = 0, $2023 = 0, $2024 = 0, $2025 = 0, $2026 = 0, $2027 = 0, $2028 = 0, $2029 = 0, $203 = 0, $2030 = 0, $2031 = 0, $2032 = 0, $2033 = 0, $2034 = 0, $2035 = 0, $2036 = 0, $2037 = 0, $2038 = 0, $2039 = 0, $204 = 0;
 var $2040 = 0, $2041 = 0, $2042 = 0, $2043 = 0, $2044 = 0, $2045 = 0, $2046 = 0, $2047 = 0, $2048 = 0, $2049 = 0, $205 = 0, $2050 = 0, $2051 = 0, $2052 = 0, $2053 = 0, $2054 = 0, $2055 = 0, $2056 = 0, $2057 = 0, $2058 = 0;
 var $2059 = 0, $206 = 0, $2060 = 0, $2061 = 0, $2062 = 0, $2063 = 0, $2064 = 0, $2065 = 0, $2066 = 0, $2067 = 0, $2068 = 0, $2069 = 0, $207 = 0, $2070 = 0, $2071 = 0, $2072 = 0, $2073 = 0, $2074 = 0, $2075 = 0, $2076 = 0;
 var $2077 = 0, $2078 = 0, $2079 = 0, $208 = 0, $2080 = 0, $2081 = 0, $2082 = 0, $2083 = 0, $2084 = 0, $2085 = 0, $2086 = 0, $2087 = 0, $2088 = 0, $2089 = 0, $209 = 0, $2090 = 0, $2091 = 0, $2092 = 0, $2093 = 0, $2094 = 0;
 var $2095 = 0, $2096 = 0, $2097 = 0, $2098 = 0, $2099 = 0, $21 = 0, $210 = 0, $2100 = 0, $2101 = 0, $2102 = 0, $2103 = 0, $2104 = 0, $2105 = 0, $2106 = 0, $2107 = 0, $2108 = 0, $2109 = 0, $211 = 0, $2110 = 0, $2111 = 0;
 var $2112 = 0, $2113 = 0, $2114 = 0, $2115 = 0, $2116 = 0, $2117 = 0, $2118 = 0, $2119 = 0, $212 = 0, $2120 = 0, $2121 = 0, $2122 = 0, $2123 = 0, $2124 = 0, $2125 = 0, $2126 = 0, $2127 = 0, $2128 = 0, $2129 = 0, $213 = 0;
 var $2130 = 0, $2131 = 0, $2132 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0;
 var $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0;
 var $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0;
 var $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0;
 var $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0;
 var $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0;
 var $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0;
 var $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0;
 var $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0;
 var $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0;
 var $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0;
 var $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0;
 var $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0;
 var $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0;
 var $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0;
 var $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0;
 var $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0;
 var $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0;
 var $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0;
 var $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0;
 var $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0;
 var $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0;
 var $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0;
 var $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0;
 var $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0;
 var $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0;
 var $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0;
 var $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0;
 var $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0;
 var $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0;
 var $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0;
 var $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0;
 var $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0;
 var $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0;
 var $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0;
 var $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0;
 var $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0;
 var $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0;
 var $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0;
 var $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0;
 var $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0;
 var $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0;
 var $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0;
 var $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $dA$0 = 0, $dA$0$shrunk = 0, $dA$1 = 0, $dA$1$shrunk = 0, $dA$2 = 0, $dA$2$shrunk = 0;
 var $dA$3 = 0, $dA$3$shrunk = 0, $dA$4 = 0, $dA$4$shrunk = 0, $dA$5 = 0, $dA$5$shrunk = 0, $dA$6 = 0, $dA$6$shrunk = 0, $dA$7 = 0, $dA$7$shrunk = 0, $dB$0 = 0, $dB$0$shrunk = 0, $dB$1 = 0, $dB$1$shrunk = 0, $dB$2 = 0, $dB$2$shrunk = 0, $dB$3 = 0, $dB$3$shrunk = 0, $dB$4 = 0, $dB$4$shrunk = 0;
 var $dB$5 = 0, $dB$5$shrunk = 0, $dB$6 = 0, $dB$6$shrunk = 0, $dB$7 = 0, $dB$7$shrunk = 0, $dG$0 = 0, $dG$0$shrunk = 0, $dG$1 = 0, $dG$1$shrunk = 0, $dG$2 = 0, $dG$2$shrunk = 0, $dG$3 = 0, $dG$3$shrunk = 0, $dG$4 = 0, $dG$4$shrunk = 0, $dG$5 = 0, $dG$5$shrunk = 0, $dG$6 = 0, $dG$6$shrunk = 0;
 var $dG$7 = 0, $dG$7$shrunk = 0, $dR$0 = 0, $dR$0$shrunk = 0, $dR$1 = 0, $dR$1$shrunk = 0, $dR$2 = 0, $dR$2$shrunk = 0, $dR$3 = 0, $dR$3$shrunk = 0, $dR$4 = 0, $dR$4$shrunk = 0, $dR$5 = 0, $dR$5$shrunk = 0, $dR$6 = 0, $dR$6$shrunk = 0, $dR$7 = 0, $dR$7$shrunk = 0, $div = 0, $div1 = 0;
 var $div10 = 0, $div12 = 0, $div13 = 0, $div14 = 0, $div15 = 0, $div16 = 0, $div18 = 0, $div19 = 0, $div2 = 0, $div20 = 0, $div21 = 0, $div22 = 0, $div24 = 0, $div25 = 0, $div26 = 0, $div27 = 0, $div28 = 0, $div3 = 0, $div30 = 0, $div31 = 0;
 var $div32 = 0, $div33 = 0, $div34 = 0, $div36 = 0, $div37 = 0, $div38 = 0, $div39 = 0, $div4 = 0, $div40 = 0, $div42 = 0, $div43 = 0, $div44 = 0, $div45 = 0, $div46 = 0, $div6 = 0, $div7 = 0, $div8 = 0, $div9 = 0, $dst$055 = 0, $dst$1 = 0;
 var $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $src$056 = 0, $src$1 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0;
 var $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 44|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($info)) + 40|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($11)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($13)) + 4|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($info)) + 56|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ((($info)) + 63|0);
 $19 = HEAP8[$18>>0]|0;
 $20 = $19&255;
 $21 = ((($9)) + 9|0);
 $22 = HEAP8[$21>>0]|0;
 $23 = $22&255;
 $24 = ($3|0)==(0);
 if ($24) {
  return;
 }
 $25 = ((($info)) + 20|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = HEAP32[$info>>2]|0;
 $28 = (($1) + 7)|0;
 $29 = (($28|0) / 8)&-1;
 $30 = $1 & 7;
 $31 = ((($9)) + 28|0);
 $32 = ((($9)) + 32|0);
 $33 = ((($9)) + 33|0);
 $34 = ((($9)) + 34|0);
 $35 = ((($9)) + 12|0);
 $36 = ((($9)) + 16|0);
 $37 = ((($9)) + 20|0);
 $38 = ((($9)) + 24|0);
 $$in = $3;$dst$055 = $26;$src$056 = $27;
 L4: while(1) {
  $39 = (($$in) + -1)|0;
  switch ($30|0) {
  case 0:  {
   $dst$1 = $dst$055;$n$0 = $29;$src$1 = $src$056;
   label = 4;
   break;
  }
  case 7:  {
   $dst$2 = $dst$055;$n$1 = $29;$src$2 = $src$056;
   label = 16;
   break;
  }
  case 6:  {
   $dst$3 = $dst$055;$n$2 = $29;$src$3 = $src$056;
   label = 28;
   break;
  }
  case 5:  {
   $dst$4 = $dst$055;$n$3 = $29;$src$4 = $src$056;
   label = 40;
   break;
  }
  case 4:  {
   $dst$5 = $dst$055;$n$4 = $29;$src$5 = $src$056;
   label = 52;
   break;
  }
  case 3:  {
   $dst$6 = $dst$055;$n$5 = $29;$src$6 = $src$056;
   label = 64;
   break;
  }
  case 2:  {
   $dst$7 = $dst$055;$n$6 = $29;$src$7 = $src$056;
   label = 76;
   break;
  }
  case 1:  {
   $dst$8 = $dst$055;$n$7 = $29;$src$8 = $src$056;
   label = 88;
   break;
  }
  default: {
   label = 100;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    $40 = HEAP8[$src$1>>0]|0;
    $41 = $40&255;
    $42 = ($41|0)==($17|0);
    L9: do {
     if (!($42)) {
      $43 = (($15) + ($41<<2)|0);
      $44 = HEAP8[$43>>0]|0;
      $45 = $44&255;
      $46 = (((($15) + ($41<<2)|0)) + 1|0);
      $47 = HEAP8[$46>>0]|0;
      $48 = $47&255;
      $49 = (((($15) + ($41<<2)|0)) + 2|0);
      $50 = HEAP8[$49>>0]|0;
      $51 = $50&255;
      switch ($23|0) {
      case 1:  {
       $52 = HEAP8[$dst$1>>0]|0;
       $53 = $52&255;
       $54 = HEAP32[$35>>2]|0;
       $55 = $53 & $54;
       $56 = HEAP32[$32>>2]|0;
       $57 = $56 & 255;
       $58 = $55 >>> $57;
       $59 = HEAP32[$31>>2]|0;
       $60 = $59 & 255;
       $61 = (125924 + ($60<<2)|0);
       $62 = HEAP32[$61>>2]|0;
       $63 = (($62) + ($58)|0);
       $64 = HEAP8[$63>>0]|0;
       $65 = HEAP32[$36>>2]|0;
       $66 = $65 & $53;
       $67 = $56 >>> 8;
       $68 = $67 & 255;
       $69 = $66 >>> $68;
       $70 = $59 >>> 8;
       $71 = $70 & 255;
       $72 = (125924 + ($71<<2)|0);
       $73 = HEAP32[$72>>2]|0;
       $74 = (($73) + ($69)|0);
       $75 = HEAP8[$74>>0]|0;
       $76 = HEAP32[$37>>2]|0;
       $77 = $76 & $53;
       $78 = $56 >>> 16;
       $79 = $78 & 255;
       $80 = $77 >>> $79;
       $81 = $59 >>> 16;
       $82 = $81 & 255;
       $83 = (125924 + ($82<<2)|0);
       $84 = HEAP32[$83>>2]|0;
       $85 = (($84) + ($80)|0);
       $86 = HEAP8[$85>>0]|0;
       $87 = HEAP32[$38>>2]|0;
       $88 = $87 & $53;
       $89 = $56 >>> 24;
       $90 = $88 >>> $89;
       $91 = $59 >>> 24;
       $92 = (125924 + ($91<<2)|0);
       $93 = HEAP32[$92>>2]|0;
       $94 = (($93) + ($90)|0);
       $95 = HEAP8[$94>>0]|0;
       $dA$0$shrunk = $95;$dB$0$shrunk = $86;$dG$0$shrunk = $75;$dR$0$shrunk = $64;
       break;
      }
      case 2:  {
       $96 = HEAP16[$dst$1>>1]|0;
       $97 = $96&65535;
       $98 = HEAP32[$35>>2]|0;
       $99 = $97 & $98;
       $100 = HEAP32[$32>>2]|0;
       $101 = $100 & 255;
       $102 = $99 >>> $101;
       $103 = HEAP32[$31>>2]|0;
       $104 = $103 & 255;
       $105 = (125924 + ($104<<2)|0);
       $106 = HEAP32[$105>>2]|0;
       $107 = (($106) + ($102)|0);
       $108 = HEAP8[$107>>0]|0;
       $109 = HEAP32[$36>>2]|0;
       $110 = $109 & $97;
       $111 = $100 >>> 8;
       $112 = $111 & 255;
       $113 = $110 >>> $112;
       $114 = $103 >>> 8;
       $115 = $114 & 255;
       $116 = (125924 + ($115<<2)|0);
       $117 = HEAP32[$116>>2]|0;
       $118 = (($117) + ($113)|0);
       $119 = HEAP8[$118>>0]|0;
       $120 = HEAP32[$37>>2]|0;
       $121 = $120 & $97;
       $122 = $100 >>> 16;
       $123 = $122 & 255;
       $124 = $121 >>> $123;
       $125 = $103 >>> 16;
       $126 = $125 & 255;
       $127 = (125924 + ($126<<2)|0);
       $128 = HEAP32[$127>>2]|0;
       $129 = (($128) + ($124)|0);
       $130 = HEAP8[$129>>0]|0;
       $131 = HEAP32[$38>>2]|0;
       $132 = $131 & $97;
       $133 = $100 >>> 24;
       $134 = $132 >>> $133;
       $135 = $103 >>> 24;
       $136 = (125924 + ($135<<2)|0);
       $137 = HEAP32[$136>>2]|0;
       $138 = (($137) + ($134)|0);
       $139 = HEAP8[$138>>0]|0;
       $dA$0$shrunk = $139;$dB$0$shrunk = $130;$dG$0$shrunk = $119;$dR$0$shrunk = $108;
       break;
      }
      case 3:  {
       $140 = HEAP32[$32>>2]|0;
       $div39 = $140 >>> 3;
       $141 = $div39 & 31;
       $142 = (($dst$1) + ($141)|0);
       $143 = HEAP8[$142>>0]|0;
       $144 = $140 >>> 11;
       $div40 = $144 & 31;
       $145 = (($dst$1) + ($div40)|0);
       $146 = HEAP8[$145>>0]|0;
       $147 = $140 >>> 19;
       $148 = $147 & 31;
       $149 = (($dst$1) + ($148)|0);
       $150 = HEAP8[$149>>0]|0;
       $dA$0$shrunk = -1;$dB$0$shrunk = $150;$dG$0$shrunk = $146;$dR$0$shrunk = $143;
       break;
      }
      case 4:  {
       $151 = HEAP32[$dst$1>>2]|0;
       $152 = HEAP32[$35>>2]|0;
       $153 = $152 & $151;
       $154 = HEAP32[$32>>2]|0;
       $155 = $154 & 255;
       $156 = $153 >>> $155;
       $157 = HEAP32[$31>>2]|0;
       $158 = $157 & 255;
       $159 = (125924 + ($158<<2)|0);
       $160 = HEAP32[$159>>2]|0;
       $161 = (($160) + ($156)|0);
       $162 = HEAP8[$161>>0]|0;
       $163 = HEAP32[$36>>2]|0;
       $164 = $163 & $151;
       $165 = $154 >>> 8;
       $166 = $165 & 255;
       $167 = $164 >>> $166;
       $168 = $157 >>> 8;
       $169 = $168 & 255;
       $170 = (125924 + ($169<<2)|0);
       $171 = HEAP32[$170>>2]|0;
       $172 = (($171) + ($167)|0);
       $173 = HEAP8[$172>>0]|0;
       $174 = HEAP32[$37>>2]|0;
       $175 = $174 & $151;
       $176 = $154 >>> 16;
       $177 = $176 & 255;
       $178 = $175 >>> $177;
       $179 = $157 >>> 16;
       $180 = $179 & 255;
       $181 = (125924 + ($180<<2)|0);
       $182 = HEAP32[$181>>2]|0;
       $183 = (($182) + ($178)|0);
       $184 = HEAP8[$183>>0]|0;
       $185 = HEAP32[$38>>2]|0;
       $186 = $185 & $151;
       $187 = $154 >>> 24;
       $188 = $186 >>> $187;
       $189 = $157 >>> 24;
       $190 = (125924 + ($189<<2)|0);
       $191 = HEAP32[$190>>2]|0;
       $192 = (($191) + ($188)|0);
       $193 = HEAP8[$192>>0]|0;
       $dA$0$shrunk = $193;$dB$0$shrunk = $184;$dG$0$shrunk = $173;$dR$0$shrunk = $162;
       break;
      }
      default: {
       $dA$0$shrunk = 0;$dB$0$shrunk = 0;$dG$0$shrunk = 0;$dR$0$shrunk = 0;
      }
      }
      $dR$0 = $dR$0$shrunk&255;
      $dG$0 = $dG$0$shrunk&255;
      $dB$0 = $dB$0$shrunk&255;
      $dA$0 = $dA$0$shrunk&255;
      $194 = (($45) - ($dR$0))|0;
      $195 = Math_imul($194, $20)|0;
      $196 = (($195>>>0) / 255)&-1;
      $197 = (($196) + ($dR$0))|0;
      $198 = (($48) - ($dG$0))|0;
      $199 = Math_imul($198, $20)|0;
      $200 = (($199>>>0) / 255)&-1;
      $201 = (($200) + ($dG$0))|0;
      $202 = (($51) - ($dB$0))|0;
      $203 = Math_imul($202, $20)|0;
      $204 = (($203>>>0) / 255)&-1;
      $205 = (($204) + ($dB$0))|0;
      $206 = (($dA$0) + ($20))|0;
      $207 = Math_imul($dA$0, $20)|0;
      $208 = (($207>>>0) / 255)&-1;
      $209 = (($206) - ($208))|0;
      switch ($23|0) {
      case 1:  {
       $210 = HEAP32[$31>>2]|0;
       $211 = $210 & 255;
       $212 = $197 >>> $211;
       $213 = HEAP32[$32>>2]|0;
       $214 = $213 & 255;
       $215 = $212 << $214;
       $216 = $210 >>> 8;
       $217 = $216 & 255;
       $218 = $201 >>> $217;
       $219 = $213 >>> 8;
       $220 = $219 & 255;
       $221 = $218 << $220;
       $222 = $221 | $215;
       $223 = $210 >>> 16;
       $224 = $223 & 255;
       $225 = $205 >>> $224;
       $226 = $213 >>> 16;
       $227 = $226 & 255;
       $228 = $225 << $227;
       $229 = $222 | $228;
       $230 = $210 >>> 24;
       $231 = $209 >>> $230;
       $232 = $213 >>> 24;
       $233 = $231 << $232;
       $234 = $229 | $233;
       $235 = $234&255;
       HEAP8[$dst$1>>0] = $235;
       break L9;
       break;
      }
      case 2:  {
       $236 = HEAP32[$31>>2]|0;
       $237 = $236 & 255;
       $238 = $197 >>> $237;
       $239 = HEAP32[$32>>2]|0;
       $240 = $239 & 255;
       $241 = $238 << $240;
       $242 = $236 >>> 8;
       $243 = $242 & 255;
       $244 = $201 >>> $243;
       $245 = $239 >>> 8;
       $246 = $245 & 255;
       $247 = $244 << $246;
       $248 = $247 | $241;
       $249 = $236 >>> 16;
       $250 = $249 & 255;
       $251 = $205 >>> $250;
       $252 = $239 >>> 16;
       $253 = $252 & 255;
       $254 = $251 << $253;
       $255 = $248 | $254;
       $256 = $236 >>> 24;
       $257 = $209 >>> $256;
       $258 = $239 >>> 24;
       $259 = $257 << $258;
       $260 = $255 | $259;
       $261 = $260&65535;
       HEAP16[$dst$1>>1] = $261;
       break L9;
       break;
      }
      case 3:  {
       $262 = $197&255;
       $263 = HEAP8[$32>>0]|0;
       $div36 = ($263&255) >>> 3;
       $264 = $div36&255;
       $265 = (($dst$1) + ($264)|0);
       HEAP8[$265>>0] = $262;
       $266 = $201&255;
       $267 = HEAP8[$33>>0]|0;
       $div37 = ($267&255) >>> 3;
       $268 = $div37&255;
       $269 = (($dst$1) + ($268)|0);
       HEAP8[$269>>0] = $266;
       $270 = $205&255;
       $271 = HEAP8[$34>>0]|0;
       $div38 = ($271&255) >>> 3;
       $272 = $div38&255;
       $273 = (($dst$1) + ($272)|0);
       HEAP8[$273>>0] = $270;
       break L9;
       break;
      }
      case 4:  {
       $274 = HEAP32[$31>>2]|0;
       $275 = $274 & 255;
       $276 = $197 >>> $275;
       $277 = HEAP32[$32>>2]|0;
       $278 = $277 & 255;
       $279 = $276 << $278;
       $280 = $274 >>> 8;
       $281 = $280 & 255;
       $282 = $201 >>> $281;
       $283 = $277 >>> 8;
       $284 = $283 & 255;
       $285 = $282 << $284;
       $286 = $285 | $279;
       $287 = $274 >>> 16;
       $288 = $287 & 255;
       $289 = $205 >>> $288;
       $290 = $277 >>> 16;
       $291 = $290 & 255;
       $292 = $289 << $291;
       $293 = $286 | $292;
       $294 = $274 >>> 24;
       $295 = $209 >>> $294;
       $296 = $277 >>> 24;
       $297 = $295 << $296;
       $298 = $293 | $297;
       HEAP32[$dst$1>>2] = $298;
       break L9;
       break;
      }
      default: {
       break L9;
      }
      }
     }
    } while(0);
    $299 = ((($src$1)) + 1|0);
    $300 = (($dst$1) + ($23)|0);
    $dst$2 = $300;$n$1 = $n$0;$src$2 = $299;
    label = 16;
    continue;
   }
   else if ((label|0) == 16) {
    label = 0;
    $301 = HEAP8[$src$2>>0]|0;
    $302 = $301&255;
    $303 = ($302|0)==($17|0);
    L24: do {
     if (!($303)) {
      $304 = (($15) + ($302<<2)|0);
      $305 = HEAP8[$304>>0]|0;
      $306 = $305&255;
      $307 = (((($15) + ($302<<2)|0)) + 1|0);
      $308 = HEAP8[$307>>0]|0;
      $309 = $308&255;
      $310 = (((($15) + ($302<<2)|0)) + 2|0);
      $311 = HEAP8[$310>>0]|0;
      $312 = $311&255;
      switch ($23|0) {
      case 1:  {
       $313 = HEAP8[$dst$2>>0]|0;
       $314 = $313&255;
       $315 = HEAP32[$35>>2]|0;
       $316 = $314 & $315;
       $317 = HEAP32[$32>>2]|0;
       $318 = $317 & 255;
       $319 = $316 >>> $318;
       $320 = HEAP32[$31>>2]|0;
       $321 = $320 & 255;
       $322 = (125924 + ($321<<2)|0);
       $323 = HEAP32[$322>>2]|0;
       $324 = (($323) + ($319)|0);
       $325 = HEAP8[$324>>0]|0;
       $326 = HEAP32[$36>>2]|0;
       $327 = $326 & $314;
       $328 = $317 >>> 8;
       $329 = $328 & 255;
       $330 = $327 >>> $329;
       $331 = $320 >>> 8;
       $332 = $331 & 255;
       $333 = (125924 + ($332<<2)|0);
       $334 = HEAP32[$333>>2]|0;
       $335 = (($334) + ($330)|0);
       $336 = HEAP8[$335>>0]|0;
       $337 = HEAP32[$37>>2]|0;
       $338 = $337 & $314;
       $339 = $317 >>> 16;
       $340 = $339 & 255;
       $341 = $338 >>> $340;
       $342 = $320 >>> 16;
       $343 = $342 & 255;
       $344 = (125924 + ($343<<2)|0);
       $345 = HEAP32[$344>>2]|0;
       $346 = (($345) + ($341)|0);
       $347 = HEAP8[$346>>0]|0;
       $348 = HEAP32[$38>>2]|0;
       $349 = $348 & $314;
       $350 = $317 >>> 24;
       $351 = $349 >>> $350;
       $352 = $320 >>> 24;
       $353 = (125924 + ($352<<2)|0);
       $354 = HEAP32[$353>>2]|0;
       $355 = (($354) + ($351)|0);
       $356 = HEAP8[$355>>0]|0;
       $dA$1$shrunk = $356;$dB$1$shrunk = $347;$dG$1$shrunk = $336;$dR$1$shrunk = $325;
       break;
      }
      case 2:  {
       $357 = HEAP16[$dst$2>>1]|0;
       $358 = $357&65535;
       $359 = HEAP32[$35>>2]|0;
       $360 = $358 & $359;
       $361 = HEAP32[$32>>2]|0;
       $362 = $361 & 255;
       $363 = $360 >>> $362;
       $364 = HEAP32[$31>>2]|0;
       $365 = $364 & 255;
       $366 = (125924 + ($365<<2)|0);
       $367 = HEAP32[$366>>2]|0;
       $368 = (($367) + ($363)|0);
       $369 = HEAP8[$368>>0]|0;
       $370 = HEAP32[$36>>2]|0;
       $371 = $370 & $358;
       $372 = $361 >>> 8;
       $373 = $372 & 255;
       $374 = $371 >>> $373;
       $375 = $364 >>> 8;
       $376 = $375 & 255;
       $377 = (125924 + ($376<<2)|0);
       $378 = HEAP32[$377>>2]|0;
       $379 = (($378) + ($374)|0);
       $380 = HEAP8[$379>>0]|0;
       $381 = HEAP32[$37>>2]|0;
       $382 = $381 & $358;
       $383 = $361 >>> 16;
       $384 = $383 & 255;
       $385 = $382 >>> $384;
       $386 = $364 >>> 16;
       $387 = $386 & 255;
       $388 = (125924 + ($387<<2)|0);
       $389 = HEAP32[$388>>2]|0;
       $390 = (($389) + ($385)|0);
       $391 = HEAP8[$390>>0]|0;
       $392 = HEAP32[$38>>2]|0;
       $393 = $392 & $358;
       $394 = $361 >>> 24;
       $395 = $393 >>> $394;
       $396 = $364 >>> 24;
       $397 = (125924 + ($396<<2)|0);
       $398 = HEAP32[$397>>2]|0;
       $399 = (($398) + ($395)|0);
       $400 = HEAP8[$399>>0]|0;
       $dA$1$shrunk = $400;$dB$1$shrunk = $391;$dG$1$shrunk = $380;$dR$1$shrunk = $369;
       break;
      }
      case 3:  {
       $401 = HEAP32[$32>>2]|0;
       $div33 = $401 >>> 3;
       $402 = $div33 & 31;
       $403 = (($dst$2) + ($402)|0);
       $404 = HEAP8[$403>>0]|0;
       $405 = $401 >>> 11;
       $div34 = $405 & 31;
       $406 = (($dst$2) + ($div34)|0);
       $407 = HEAP8[$406>>0]|0;
       $408 = $401 >>> 19;
       $409 = $408 & 31;
       $410 = (($dst$2) + ($409)|0);
       $411 = HEAP8[$410>>0]|0;
       $dA$1$shrunk = -1;$dB$1$shrunk = $411;$dG$1$shrunk = $407;$dR$1$shrunk = $404;
       break;
      }
      case 4:  {
       $412 = HEAP32[$dst$2>>2]|0;
       $413 = HEAP32[$35>>2]|0;
       $414 = $413 & $412;
       $415 = HEAP32[$32>>2]|0;
       $416 = $415 & 255;
       $417 = $414 >>> $416;
       $418 = HEAP32[$31>>2]|0;
       $419 = $418 & 255;
       $420 = (125924 + ($419<<2)|0);
       $421 = HEAP32[$420>>2]|0;
       $422 = (($421) + ($417)|0);
       $423 = HEAP8[$422>>0]|0;
       $424 = HEAP32[$36>>2]|0;
       $425 = $424 & $412;
       $426 = $415 >>> 8;
       $427 = $426 & 255;
       $428 = $425 >>> $427;
       $429 = $418 >>> 8;
       $430 = $429 & 255;
       $431 = (125924 + ($430<<2)|0);
       $432 = HEAP32[$431>>2]|0;
       $433 = (($432) + ($428)|0);
       $434 = HEAP8[$433>>0]|0;
       $435 = HEAP32[$37>>2]|0;
       $436 = $435 & $412;
       $437 = $415 >>> 16;
       $438 = $437 & 255;
       $439 = $436 >>> $438;
       $440 = $418 >>> 16;
       $441 = $440 & 255;
       $442 = (125924 + ($441<<2)|0);
       $443 = HEAP32[$442>>2]|0;
       $444 = (($443) + ($439)|0);
       $445 = HEAP8[$444>>0]|0;
       $446 = HEAP32[$38>>2]|0;
       $447 = $446 & $412;
       $448 = $415 >>> 24;
       $449 = $447 >>> $448;
       $450 = $418 >>> 24;
       $451 = (125924 + ($450<<2)|0);
       $452 = HEAP32[$451>>2]|0;
       $453 = (($452) + ($449)|0);
       $454 = HEAP8[$453>>0]|0;
       $dA$1$shrunk = $454;$dB$1$shrunk = $445;$dG$1$shrunk = $434;$dR$1$shrunk = $423;
       break;
      }
      default: {
       $dA$1$shrunk = 0;$dB$1$shrunk = 0;$dG$1$shrunk = 0;$dR$1$shrunk = 0;
      }
      }
      $dR$1 = $dR$1$shrunk&255;
      $dG$1 = $dG$1$shrunk&255;
      $dB$1 = $dB$1$shrunk&255;
      $dA$1 = $dA$1$shrunk&255;
      $455 = (($306) - ($dR$1))|0;
      $456 = Math_imul($455, $20)|0;
      $457 = (($456>>>0) / 255)&-1;
      $458 = (($457) + ($dR$1))|0;
      $459 = (($309) - ($dG$1))|0;
      $460 = Math_imul($459, $20)|0;
      $461 = (($460>>>0) / 255)&-1;
      $462 = (($461) + ($dG$1))|0;
      $463 = (($312) - ($dB$1))|0;
      $464 = Math_imul($463, $20)|0;
      $465 = (($464>>>0) / 255)&-1;
      $466 = (($465) + ($dB$1))|0;
      $467 = (($dA$1) + ($20))|0;
      $468 = Math_imul($dA$1, $20)|0;
      $469 = (($468>>>0) / 255)&-1;
      $470 = (($467) - ($469))|0;
      switch ($23|0) {
      case 1:  {
       $471 = HEAP32[$31>>2]|0;
       $472 = $471 & 255;
       $473 = $458 >>> $472;
       $474 = HEAP32[$32>>2]|0;
       $475 = $474 & 255;
       $476 = $473 << $475;
       $477 = $471 >>> 8;
       $478 = $477 & 255;
       $479 = $462 >>> $478;
       $480 = $474 >>> 8;
       $481 = $480 & 255;
       $482 = $479 << $481;
       $483 = $482 | $476;
       $484 = $471 >>> 16;
       $485 = $484 & 255;
       $486 = $466 >>> $485;
       $487 = $474 >>> 16;
       $488 = $487 & 255;
       $489 = $486 << $488;
       $490 = $483 | $489;
       $491 = $471 >>> 24;
       $492 = $470 >>> $491;
       $493 = $474 >>> 24;
       $494 = $492 << $493;
       $495 = $490 | $494;
       $496 = $495&255;
       HEAP8[$dst$2>>0] = $496;
       break L24;
       break;
      }
      case 2:  {
       $497 = HEAP32[$31>>2]|0;
       $498 = $497 & 255;
       $499 = $458 >>> $498;
       $500 = HEAP32[$32>>2]|0;
       $501 = $500 & 255;
       $502 = $499 << $501;
       $503 = $497 >>> 8;
       $504 = $503 & 255;
       $505 = $462 >>> $504;
       $506 = $500 >>> 8;
       $507 = $506 & 255;
       $508 = $505 << $507;
       $509 = $508 | $502;
       $510 = $497 >>> 16;
       $511 = $510 & 255;
       $512 = $466 >>> $511;
       $513 = $500 >>> 16;
       $514 = $513 & 255;
       $515 = $512 << $514;
       $516 = $509 | $515;
       $517 = $497 >>> 24;
       $518 = $470 >>> $517;
       $519 = $500 >>> 24;
       $520 = $518 << $519;
       $521 = $516 | $520;
       $522 = $521&65535;
       HEAP16[$dst$2>>1] = $522;
       break L24;
       break;
      }
      case 3:  {
       $523 = $458&255;
       $524 = HEAP8[$32>>0]|0;
       $div30 = ($524&255) >>> 3;
       $525 = $div30&255;
       $526 = (($dst$2) + ($525)|0);
       HEAP8[$526>>0] = $523;
       $527 = $462&255;
       $528 = HEAP8[$33>>0]|0;
       $div31 = ($528&255) >>> 3;
       $529 = $div31&255;
       $530 = (($dst$2) + ($529)|0);
       HEAP8[$530>>0] = $527;
       $531 = $466&255;
       $532 = HEAP8[$34>>0]|0;
       $div32 = ($532&255) >>> 3;
       $533 = $div32&255;
       $534 = (($dst$2) + ($533)|0);
       HEAP8[$534>>0] = $531;
       break L24;
       break;
      }
      case 4:  {
       $535 = HEAP32[$31>>2]|0;
       $536 = $535 & 255;
       $537 = $458 >>> $536;
       $538 = HEAP32[$32>>2]|0;
       $539 = $538 & 255;
       $540 = $537 << $539;
       $541 = $535 >>> 8;
       $542 = $541 & 255;
       $543 = $462 >>> $542;
       $544 = $538 >>> 8;
       $545 = $544 & 255;
       $546 = $543 << $545;
       $547 = $546 | $540;
       $548 = $535 >>> 16;
       $549 = $548 & 255;
       $550 = $466 >>> $549;
       $551 = $538 >>> 16;
       $552 = $551 & 255;
       $553 = $550 << $552;
       $554 = $547 | $553;
       $555 = $535 >>> 24;
       $556 = $470 >>> $555;
       $557 = $538 >>> 24;
       $558 = $556 << $557;
       $559 = $554 | $558;
       HEAP32[$dst$2>>2] = $559;
       break L24;
       break;
      }
      default: {
       break L24;
      }
      }
     }
    } while(0);
    $560 = ((($src$2)) + 1|0);
    $561 = (($dst$2) + ($23)|0);
    $dst$3 = $561;$n$2 = $n$1;$src$3 = $560;
    label = 28;
    continue;
   }
   else if ((label|0) == 28) {
    label = 0;
    $562 = HEAP8[$src$3>>0]|0;
    $563 = $562&255;
    $564 = ($563|0)==($17|0);
    L39: do {
     if (!($564)) {
      $565 = (($15) + ($563<<2)|0);
      $566 = HEAP8[$565>>0]|0;
      $567 = $566&255;
      $568 = (((($15) + ($563<<2)|0)) + 1|0);
      $569 = HEAP8[$568>>0]|0;
      $570 = $569&255;
      $571 = (((($15) + ($563<<2)|0)) + 2|0);
      $572 = HEAP8[$571>>0]|0;
      $573 = $572&255;
      switch ($23|0) {
      case 1:  {
       $574 = HEAP8[$dst$3>>0]|0;
       $575 = $574&255;
       $576 = HEAP32[$35>>2]|0;
       $577 = $575 & $576;
       $578 = HEAP32[$32>>2]|0;
       $579 = $578 & 255;
       $580 = $577 >>> $579;
       $581 = HEAP32[$31>>2]|0;
       $582 = $581 & 255;
       $583 = (125924 + ($582<<2)|0);
       $584 = HEAP32[$583>>2]|0;
       $585 = (($584) + ($580)|0);
       $586 = HEAP8[$585>>0]|0;
       $587 = HEAP32[$36>>2]|0;
       $588 = $587 & $575;
       $589 = $578 >>> 8;
       $590 = $589 & 255;
       $591 = $588 >>> $590;
       $592 = $581 >>> 8;
       $593 = $592 & 255;
       $594 = (125924 + ($593<<2)|0);
       $595 = HEAP32[$594>>2]|0;
       $596 = (($595) + ($591)|0);
       $597 = HEAP8[$596>>0]|0;
       $598 = HEAP32[$37>>2]|0;
       $599 = $598 & $575;
       $600 = $578 >>> 16;
       $601 = $600 & 255;
       $602 = $599 >>> $601;
       $603 = $581 >>> 16;
       $604 = $603 & 255;
       $605 = (125924 + ($604<<2)|0);
       $606 = HEAP32[$605>>2]|0;
       $607 = (($606) + ($602)|0);
       $608 = HEAP8[$607>>0]|0;
       $609 = HEAP32[$38>>2]|0;
       $610 = $609 & $575;
       $611 = $578 >>> 24;
       $612 = $610 >>> $611;
       $613 = $581 >>> 24;
       $614 = (125924 + ($613<<2)|0);
       $615 = HEAP32[$614>>2]|0;
       $616 = (($615) + ($612)|0);
       $617 = HEAP8[$616>>0]|0;
       $dA$2$shrunk = $617;$dB$2$shrunk = $608;$dG$2$shrunk = $597;$dR$2$shrunk = $586;
       break;
      }
      case 2:  {
       $618 = HEAP16[$dst$3>>1]|0;
       $619 = $618&65535;
       $620 = HEAP32[$35>>2]|0;
       $621 = $619 & $620;
       $622 = HEAP32[$32>>2]|0;
       $623 = $622 & 255;
       $624 = $621 >>> $623;
       $625 = HEAP32[$31>>2]|0;
       $626 = $625 & 255;
       $627 = (125924 + ($626<<2)|0);
       $628 = HEAP32[$627>>2]|0;
       $629 = (($628) + ($624)|0);
       $630 = HEAP8[$629>>0]|0;
       $631 = HEAP32[$36>>2]|0;
       $632 = $631 & $619;
       $633 = $622 >>> 8;
       $634 = $633 & 255;
       $635 = $632 >>> $634;
       $636 = $625 >>> 8;
       $637 = $636 & 255;
       $638 = (125924 + ($637<<2)|0);
       $639 = HEAP32[$638>>2]|0;
       $640 = (($639) + ($635)|0);
       $641 = HEAP8[$640>>0]|0;
       $642 = HEAP32[$37>>2]|0;
       $643 = $642 & $619;
       $644 = $622 >>> 16;
       $645 = $644 & 255;
       $646 = $643 >>> $645;
       $647 = $625 >>> 16;
       $648 = $647 & 255;
       $649 = (125924 + ($648<<2)|0);
       $650 = HEAP32[$649>>2]|0;
       $651 = (($650) + ($646)|0);
       $652 = HEAP8[$651>>0]|0;
       $653 = HEAP32[$38>>2]|0;
       $654 = $653 & $619;
       $655 = $622 >>> 24;
       $656 = $654 >>> $655;
       $657 = $625 >>> 24;
       $658 = (125924 + ($657<<2)|0);
       $659 = HEAP32[$658>>2]|0;
       $660 = (($659) + ($656)|0);
       $661 = HEAP8[$660>>0]|0;
       $dA$2$shrunk = $661;$dB$2$shrunk = $652;$dG$2$shrunk = $641;$dR$2$shrunk = $630;
       break;
      }
      case 3:  {
       $662 = HEAP32[$32>>2]|0;
       $div27 = $662 >>> 3;
       $663 = $div27 & 31;
       $664 = (($dst$3) + ($663)|0);
       $665 = HEAP8[$664>>0]|0;
       $666 = $662 >>> 11;
       $div28 = $666 & 31;
       $667 = (($dst$3) + ($div28)|0);
       $668 = HEAP8[$667>>0]|0;
       $669 = $662 >>> 19;
       $670 = $669 & 31;
       $671 = (($dst$3) + ($670)|0);
       $672 = HEAP8[$671>>0]|0;
       $dA$2$shrunk = -1;$dB$2$shrunk = $672;$dG$2$shrunk = $668;$dR$2$shrunk = $665;
       break;
      }
      case 4:  {
       $673 = HEAP32[$dst$3>>2]|0;
       $674 = HEAP32[$35>>2]|0;
       $675 = $674 & $673;
       $676 = HEAP32[$32>>2]|0;
       $677 = $676 & 255;
       $678 = $675 >>> $677;
       $679 = HEAP32[$31>>2]|0;
       $680 = $679 & 255;
       $681 = (125924 + ($680<<2)|0);
       $682 = HEAP32[$681>>2]|0;
       $683 = (($682) + ($678)|0);
       $684 = HEAP8[$683>>0]|0;
       $685 = HEAP32[$36>>2]|0;
       $686 = $685 & $673;
       $687 = $676 >>> 8;
       $688 = $687 & 255;
       $689 = $686 >>> $688;
       $690 = $679 >>> 8;
       $691 = $690 & 255;
       $692 = (125924 + ($691<<2)|0);
       $693 = HEAP32[$692>>2]|0;
       $694 = (($693) + ($689)|0);
       $695 = HEAP8[$694>>0]|0;
       $696 = HEAP32[$37>>2]|0;
       $697 = $696 & $673;
       $698 = $676 >>> 16;
       $699 = $698 & 255;
       $700 = $697 >>> $699;
       $701 = $679 >>> 16;
       $702 = $701 & 255;
       $703 = (125924 + ($702<<2)|0);
       $704 = HEAP32[$703>>2]|0;
       $705 = (($704) + ($700)|0);
       $706 = HEAP8[$705>>0]|0;
       $707 = HEAP32[$38>>2]|0;
       $708 = $707 & $673;
       $709 = $676 >>> 24;
       $710 = $708 >>> $709;
       $711 = $679 >>> 24;
       $712 = (125924 + ($711<<2)|0);
       $713 = HEAP32[$712>>2]|0;
       $714 = (($713) + ($710)|0);
       $715 = HEAP8[$714>>0]|0;
       $dA$2$shrunk = $715;$dB$2$shrunk = $706;$dG$2$shrunk = $695;$dR$2$shrunk = $684;
       break;
      }
      default: {
       $dA$2$shrunk = 0;$dB$2$shrunk = 0;$dG$2$shrunk = 0;$dR$2$shrunk = 0;
      }
      }
      $dR$2 = $dR$2$shrunk&255;
      $dG$2 = $dG$2$shrunk&255;
      $dB$2 = $dB$2$shrunk&255;
      $dA$2 = $dA$2$shrunk&255;
      $716 = (($567) - ($dR$2))|0;
      $717 = Math_imul($716, $20)|0;
      $718 = (($717>>>0) / 255)&-1;
      $719 = (($718) + ($dR$2))|0;
      $720 = (($570) - ($dG$2))|0;
      $721 = Math_imul($720, $20)|0;
      $722 = (($721>>>0) / 255)&-1;
      $723 = (($722) + ($dG$2))|0;
      $724 = (($573) - ($dB$2))|0;
      $725 = Math_imul($724, $20)|0;
      $726 = (($725>>>0) / 255)&-1;
      $727 = (($726) + ($dB$2))|0;
      $728 = (($dA$2) + ($20))|0;
      $729 = Math_imul($dA$2, $20)|0;
      $730 = (($729>>>0) / 255)&-1;
      $731 = (($728) - ($730))|0;
      switch ($23|0) {
      case 1:  {
       $732 = HEAP32[$31>>2]|0;
       $733 = $732 & 255;
       $734 = $719 >>> $733;
       $735 = HEAP32[$32>>2]|0;
       $736 = $735 & 255;
       $737 = $734 << $736;
       $738 = $732 >>> 8;
       $739 = $738 & 255;
       $740 = $723 >>> $739;
       $741 = $735 >>> 8;
       $742 = $741 & 255;
       $743 = $740 << $742;
       $744 = $743 | $737;
       $745 = $732 >>> 16;
       $746 = $745 & 255;
       $747 = $727 >>> $746;
       $748 = $735 >>> 16;
       $749 = $748 & 255;
       $750 = $747 << $749;
       $751 = $744 | $750;
       $752 = $732 >>> 24;
       $753 = $731 >>> $752;
       $754 = $735 >>> 24;
       $755 = $753 << $754;
       $756 = $751 | $755;
       $757 = $756&255;
       HEAP8[$dst$3>>0] = $757;
       break L39;
       break;
      }
      case 2:  {
       $758 = HEAP32[$31>>2]|0;
       $759 = $758 & 255;
       $760 = $719 >>> $759;
       $761 = HEAP32[$32>>2]|0;
       $762 = $761 & 255;
       $763 = $760 << $762;
       $764 = $758 >>> 8;
       $765 = $764 & 255;
       $766 = $723 >>> $765;
       $767 = $761 >>> 8;
       $768 = $767 & 255;
       $769 = $766 << $768;
       $770 = $769 | $763;
       $771 = $758 >>> 16;
       $772 = $771 & 255;
       $773 = $727 >>> $772;
       $774 = $761 >>> 16;
       $775 = $774 & 255;
       $776 = $773 << $775;
       $777 = $770 | $776;
       $778 = $758 >>> 24;
       $779 = $731 >>> $778;
       $780 = $761 >>> 24;
       $781 = $779 << $780;
       $782 = $777 | $781;
       $783 = $782&65535;
       HEAP16[$dst$3>>1] = $783;
       break L39;
       break;
      }
      case 3:  {
       $784 = $719&255;
       $785 = HEAP8[$32>>0]|0;
       $div24 = ($785&255) >>> 3;
       $786 = $div24&255;
       $787 = (($dst$3) + ($786)|0);
       HEAP8[$787>>0] = $784;
       $788 = $723&255;
       $789 = HEAP8[$33>>0]|0;
       $div25 = ($789&255) >>> 3;
       $790 = $div25&255;
       $791 = (($dst$3) + ($790)|0);
       HEAP8[$791>>0] = $788;
       $792 = $727&255;
       $793 = HEAP8[$34>>0]|0;
       $div26 = ($793&255) >>> 3;
       $794 = $div26&255;
       $795 = (($dst$3) + ($794)|0);
       HEAP8[$795>>0] = $792;
       break L39;
       break;
      }
      case 4:  {
       $796 = HEAP32[$31>>2]|0;
       $797 = $796 & 255;
       $798 = $719 >>> $797;
       $799 = HEAP32[$32>>2]|0;
       $800 = $799 & 255;
       $801 = $798 << $800;
       $802 = $796 >>> 8;
       $803 = $802 & 255;
       $804 = $723 >>> $803;
       $805 = $799 >>> 8;
       $806 = $805 & 255;
       $807 = $804 << $806;
       $808 = $807 | $801;
       $809 = $796 >>> 16;
       $810 = $809 & 255;
       $811 = $727 >>> $810;
       $812 = $799 >>> 16;
       $813 = $812 & 255;
       $814 = $811 << $813;
       $815 = $808 | $814;
       $816 = $796 >>> 24;
       $817 = $731 >>> $816;
       $818 = $799 >>> 24;
       $819 = $817 << $818;
       $820 = $815 | $819;
       HEAP32[$dst$3>>2] = $820;
       break L39;
       break;
      }
      default: {
       break L39;
      }
      }
     }
    } while(0);
    $821 = ((($src$3)) + 1|0);
    $822 = (($dst$3) + ($23)|0);
    $dst$4 = $822;$n$3 = $n$2;$src$4 = $821;
    label = 40;
    continue;
   }
   else if ((label|0) == 40) {
    label = 0;
    $823 = HEAP8[$src$4>>0]|0;
    $824 = $823&255;
    $825 = ($824|0)==($17|0);
    L54: do {
     if (!($825)) {
      $826 = (($15) + ($824<<2)|0);
      $827 = HEAP8[$826>>0]|0;
      $828 = $827&255;
      $829 = (((($15) + ($824<<2)|0)) + 1|0);
      $830 = HEAP8[$829>>0]|0;
      $831 = $830&255;
      $832 = (((($15) + ($824<<2)|0)) + 2|0);
      $833 = HEAP8[$832>>0]|0;
      $834 = $833&255;
      switch ($23|0) {
      case 1:  {
       $835 = HEAP8[$dst$4>>0]|0;
       $836 = $835&255;
       $837 = HEAP32[$35>>2]|0;
       $838 = $836 & $837;
       $839 = HEAP32[$32>>2]|0;
       $840 = $839 & 255;
       $841 = $838 >>> $840;
       $842 = HEAP32[$31>>2]|0;
       $843 = $842 & 255;
       $844 = (125924 + ($843<<2)|0);
       $845 = HEAP32[$844>>2]|0;
       $846 = (($845) + ($841)|0);
       $847 = HEAP8[$846>>0]|0;
       $848 = HEAP32[$36>>2]|0;
       $849 = $848 & $836;
       $850 = $839 >>> 8;
       $851 = $850 & 255;
       $852 = $849 >>> $851;
       $853 = $842 >>> 8;
       $854 = $853 & 255;
       $855 = (125924 + ($854<<2)|0);
       $856 = HEAP32[$855>>2]|0;
       $857 = (($856) + ($852)|0);
       $858 = HEAP8[$857>>0]|0;
       $859 = HEAP32[$37>>2]|0;
       $860 = $859 & $836;
       $861 = $839 >>> 16;
       $862 = $861 & 255;
       $863 = $860 >>> $862;
       $864 = $842 >>> 16;
       $865 = $864 & 255;
       $866 = (125924 + ($865<<2)|0);
       $867 = HEAP32[$866>>2]|0;
       $868 = (($867) + ($863)|0);
       $869 = HEAP8[$868>>0]|0;
       $870 = HEAP32[$38>>2]|0;
       $871 = $870 & $836;
       $872 = $839 >>> 24;
       $873 = $871 >>> $872;
       $874 = $842 >>> 24;
       $875 = (125924 + ($874<<2)|0);
       $876 = HEAP32[$875>>2]|0;
       $877 = (($876) + ($873)|0);
       $878 = HEAP8[$877>>0]|0;
       $dA$3$shrunk = $878;$dB$3$shrunk = $869;$dG$3$shrunk = $858;$dR$3$shrunk = $847;
       break;
      }
      case 2:  {
       $879 = HEAP16[$dst$4>>1]|0;
       $880 = $879&65535;
       $881 = HEAP32[$35>>2]|0;
       $882 = $880 & $881;
       $883 = HEAP32[$32>>2]|0;
       $884 = $883 & 255;
       $885 = $882 >>> $884;
       $886 = HEAP32[$31>>2]|0;
       $887 = $886 & 255;
       $888 = (125924 + ($887<<2)|0);
       $889 = HEAP32[$888>>2]|0;
       $890 = (($889) + ($885)|0);
       $891 = HEAP8[$890>>0]|0;
       $892 = HEAP32[$36>>2]|0;
       $893 = $892 & $880;
       $894 = $883 >>> 8;
       $895 = $894 & 255;
       $896 = $893 >>> $895;
       $897 = $886 >>> 8;
       $898 = $897 & 255;
       $899 = (125924 + ($898<<2)|0);
       $900 = HEAP32[$899>>2]|0;
       $901 = (($900) + ($896)|0);
       $902 = HEAP8[$901>>0]|0;
       $903 = HEAP32[$37>>2]|0;
       $904 = $903 & $880;
       $905 = $883 >>> 16;
       $906 = $905 & 255;
       $907 = $904 >>> $906;
       $908 = $886 >>> 16;
       $909 = $908 & 255;
       $910 = (125924 + ($909<<2)|0);
       $911 = HEAP32[$910>>2]|0;
       $912 = (($911) + ($907)|0);
       $913 = HEAP8[$912>>0]|0;
       $914 = HEAP32[$38>>2]|0;
       $915 = $914 & $880;
       $916 = $883 >>> 24;
       $917 = $915 >>> $916;
       $918 = $886 >>> 24;
       $919 = (125924 + ($918<<2)|0);
       $920 = HEAP32[$919>>2]|0;
       $921 = (($920) + ($917)|0);
       $922 = HEAP8[$921>>0]|0;
       $dA$3$shrunk = $922;$dB$3$shrunk = $913;$dG$3$shrunk = $902;$dR$3$shrunk = $891;
       break;
      }
      case 3:  {
       $923 = HEAP32[$32>>2]|0;
       $div21 = $923 >>> 3;
       $924 = $div21 & 31;
       $925 = (($dst$4) + ($924)|0);
       $926 = HEAP8[$925>>0]|0;
       $927 = $923 >>> 11;
       $div22 = $927 & 31;
       $928 = (($dst$4) + ($div22)|0);
       $929 = HEAP8[$928>>0]|0;
       $930 = $923 >>> 19;
       $931 = $930 & 31;
       $932 = (($dst$4) + ($931)|0);
       $933 = HEAP8[$932>>0]|0;
       $dA$3$shrunk = -1;$dB$3$shrunk = $933;$dG$3$shrunk = $929;$dR$3$shrunk = $926;
       break;
      }
      case 4:  {
       $934 = HEAP32[$dst$4>>2]|0;
       $935 = HEAP32[$35>>2]|0;
       $936 = $935 & $934;
       $937 = HEAP32[$32>>2]|0;
       $938 = $937 & 255;
       $939 = $936 >>> $938;
       $940 = HEAP32[$31>>2]|0;
       $941 = $940 & 255;
       $942 = (125924 + ($941<<2)|0);
       $943 = HEAP32[$942>>2]|0;
       $944 = (($943) + ($939)|0);
       $945 = HEAP8[$944>>0]|0;
       $946 = HEAP32[$36>>2]|0;
       $947 = $946 & $934;
       $948 = $937 >>> 8;
       $949 = $948 & 255;
       $950 = $947 >>> $949;
       $951 = $940 >>> 8;
       $952 = $951 & 255;
       $953 = (125924 + ($952<<2)|0);
       $954 = HEAP32[$953>>2]|0;
       $955 = (($954) + ($950)|0);
       $956 = HEAP8[$955>>0]|0;
       $957 = HEAP32[$37>>2]|0;
       $958 = $957 & $934;
       $959 = $937 >>> 16;
       $960 = $959 & 255;
       $961 = $958 >>> $960;
       $962 = $940 >>> 16;
       $963 = $962 & 255;
       $964 = (125924 + ($963<<2)|0);
       $965 = HEAP32[$964>>2]|0;
       $966 = (($965) + ($961)|0);
       $967 = HEAP8[$966>>0]|0;
       $968 = HEAP32[$38>>2]|0;
       $969 = $968 & $934;
       $970 = $937 >>> 24;
       $971 = $969 >>> $970;
       $972 = $940 >>> 24;
       $973 = (125924 + ($972<<2)|0);
       $974 = HEAP32[$973>>2]|0;
       $975 = (($974) + ($971)|0);
       $976 = HEAP8[$975>>0]|0;
       $dA$3$shrunk = $976;$dB$3$shrunk = $967;$dG$3$shrunk = $956;$dR$3$shrunk = $945;
       break;
      }
      default: {
       $dA$3$shrunk = 0;$dB$3$shrunk = 0;$dG$3$shrunk = 0;$dR$3$shrunk = 0;
      }
      }
      $dR$3 = $dR$3$shrunk&255;
      $dG$3 = $dG$3$shrunk&255;
      $dB$3 = $dB$3$shrunk&255;
      $dA$3 = $dA$3$shrunk&255;
      $977 = (($828) - ($dR$3))|0;
      $978 = Math_imul($977, $20)|0;
      $979 = (($978>>>0) / 255)&-1;
      $980 = (($979) + ($dR$3))|0;
      $981 = (($831) - ($dG$3))|0;
      $982 = Math_imul($981, $20)|0;
      $983 = (($982>>>0) / 255)&-1;
      $984 = (($983) + ($dG$3))|0;
      $985 = (($834) - ($dB$3))|0;
      $986 = Math_imul($985, $20)|0;
      $987 = (($986>>>0) / 255)&-1;
      $988 = (($987) + ($dB$3))|0;
      $989 = (($dA$3) + ($20))|0;
      $990 = Math_imul($dA$3, $20)|0;
      $991 = (($990>>>0) / 255)&-1;
      $992 = (($989) - ($991))|0;
      switch ($23|0) {
      case 1:  {
       $993 = HEAP32[$31>>2]|0;
       $994 = $993 & 255;
       $995 = $980 >>> $994;
       $996 = HEAP32[$32>>2]|0;
       $997 = $996 & 255;
       $998 = $995 << $997;
       $999 = $993 >>> 8;
       $1000 = $999 & 255;
       $1001 = $984 >>> $1000;
       $1002 = $996 >>> 8;
       $1003 = $1002 & 255;
       $1004 = $1001 << $1003;
       $1005 = $1004 | $998;
       $1006 = $993 >>> 16;
       $1007 = $1006 & 255;
       $1008 = $988 >>> $1007;
       $1009 = $996 >>> 16;
       $1010 = $1009 & 255;
       $1011 = $1008 << $1010;
       $1012 = $1005 | $1011;
       $1013 = $993 >>> 24;
       $1014 = $992 >>> $1013;
       $1015 = $996 >>> 24;
       $1016 = $1014 << $1015;
       $1017 = $1012 | $1016;
       $1018 = $1017&255;
       HEAP8[$dst$4>>0] = $1018;
       break L54;
       break;
      }
      case 2:  {
       $1019 = HEAP32[$31>>2]|0;
       $1020 = $1019 & 255;
       $1021 = $980 >>> $1020;
       $1022 = HEAP32[$32>>2]|0;
       $1023 = $1022 & 255;
       $1024 = $1021 << $1023;
       $1025 = $1019 >>> 8;
       $1026 = $1025 & 255;
       $1027 = $984 >>> $1026;
       $1028 = $1022 >>> 8;
       $1029 = $1028 & 255;
       $1030 = $1027 << $1029;
       $1031 = $1030 | $1024;
       $1032 = $1019 >>> 16;
       $1033 = $1032 & 255;
       $1034 = $988 >>> $1033;
       $1035 = $1022 >>> 16;
       $1036 = $1035 & 255;
       $1037 = $1034 << $1036;
       $1038 = $1031 | $1037;
       $1039 = $1019 >>> 24;
       $1040 = $992 >>> $1039;
       $1041 = $1022 >>> 24;
       $1042 = $1040 << $1041;
       $1043 = $1038 | $1042;
       $1044 = $1043&65535;
       HEAP16[$dst$4>>1] = $1044;
       break L54;
       break;
      }
      case 3:  {
       $1045 = $980&255;
       $1046 = HEAP8[$32>>0]|0;
       $div18 = ($1046&255) >>> 3;
       $1047 = $div18&255;
       $1048 = (($dst$4) + ($1047)|0);
       HEAP8[$1048>>0] = $1045;
       $1049 = $984&255;
       $1050 = HEAP8[$33>>0]|0;
       $div19 = ($1050&255) >>> 3;
       $1051 = $div19&255;
       $1052 = (($dst$4) + ($1051)|0);
       HEAP8[$1052>>0] = $1049;
       $1053 = $988&255;
       $1054 = HEAP8[$34>>0]|0;
       $div20 = ($1054&255) >>> 3;
       $1055 = $div20&255;
       $1056 = (($dst$4) + ($1055)|0);
       HEAP8[$1056>>0] = $1053;
       break L54;
       break;
      }
      case 4:  {
       $1057 = HEAP32[$31>>2]|0;
       $1058 = $1057 & 255;
       $1059 = $980 >>> $1058;
       $1060 = HEAP32[$32>>2]|0;
       $1061 = $1060 & 255;
       $1062 = $1059 << $1061;
       $1063 = $1057 >>> 8;
       $1064 = $1063 & 255;
       $1065 = $984 >>> $1064;
       $1066 = $1060 >>> 8;
       $1067 = $1066 & 255;
       $1068 = $1065 << $1067;
       $1069 = $1068 | $1062;
       $1070 = $1057 >>> 16;
       $1071 = $1070 & 255;
       $1072 = $988 >>> $1071;
       $1073 = $1060 >>> 16;
       $1074 = $1073 & 255;
       $1075 = $1072 << $1074;
       $1076 = $1069 | $1075;
       $1077 = $1057 >>> 24;
       $1078 = $992 >>> $1077;
       $1079 = $1060 >>> 24;
       $1080 = $1078 << $1079;
       $1081 = $1076 | $1080;
       HEAP32[$dst$4>>2] = $1081;
       break L54;
       break;
      }
      default: {
       break L54;
      }
      }
     }
    } while(0);
    $1082 = ((($src$4)) + 1|0);
    $1083 = (($dst$4) + ($23)|0);
    $dst$5 = $1083;$n$4 = $n$3;$src$5 = $1082;
    label = 52;
    continue;
   }
   else if ((label|0) == 52) {
    label = 0;
    $1084 = HEAP8[$src$5>>0]|0;
    $1085 = $1084&255;
    $1086 = ($1085|0)==($17|0);
    L69: do {
     if (!($1086)) {
      $1087 = (($15) + ($1085<<2)|0);
      $1088 = HEAP8[$1087>>0]|0;
      $1089 = $1088&255;
      $1090 = (((($15) + ($1085<<2)|0)) + 1|0);
      $1091 = HEAP8[$1090>>0]|0;
      $1092 = $1091&255;
      $1093 = (((($15) + ($1085<<2)|0)) + 2|0);
      $1094 = HEAP8[$1093>>0]|0;
      $1095 = $1094&255;
      switch ($23|0) {
      case 1:  {
       $1096 = HEAP8[$dst$5>>0]|0;
       $1097 = $1096&255;
       $1098 = HEAP32[$35>>2]|0;
       $1099 = $1097 & $1098;
       $1100 = HEAP32[$32>>2]|0;
       $1101 = $1100 & 255;
       $1102 = $1099 >>> $1101;
       $1103 = HEAP32[$31>>2]|0;
       $1104 = $1103 & 255;
       $1105 = (125924 + ($1104<<2)|0);
       $1106 = HEAP32[$1105>>2]|0;
       $1107 = (($1106) + ($1102)|0);
       $1108 = HEAP8[$1107>>0]|0;
       $1109 = HEAP32[$36>>2]|0;
       $1110 = $1109 & $1097;
       $1111 = $1100 >>> 8;
       $1112 = $1111 & 255;
       $1113 = $1110 >>> $1112;
       $1114 = $1103 >>> 8;
       $1115 = $1114 & 255;
       $1116 = (125924 + ($1115<<2)|0);
       $1117 = HEAP32[$1116>>2]|0;
       $1118 = (($1117) + ($1113)|0);
       $1119 = HEAP8[$1118>>0]|0;
       $1120 = HEAP32[$37>>2]|0;
       $1121 = $1120 & $1097;
       $1122 = $1100 >>> 16;
       $1123 = $1122 & 255;
       $1124 = $1121 >>> $1123;
       $1125 = $1103 >>> 16;
       $1126 = $1125 & 255;
       $1127 = (125924 + ($1126<<2)|0);
       $1128 = HEAP32[$1127>>2]|0;
       $1129 = (($1128) + ($1124)|0);
       $1130 = HEAP8[$1129>>0]|0;
       $1131 = HEAP32[$38>>2]|0;
       $1132 = $1131 & $1097;
       $1133 = $1100 >>> 24;
       $1134 = $1132 >>> $1133;
       $1135 = $1103 >>> 24;
       $1136 = (125924 + ($1135<<2)|0);
       $1137 = HEAP32[$1136>>2]|0;
       $1138 = (($1137) + ($1134)|0);
       $1139 = HEAP8[$1138>>0]|0;
       $dA$4$shrunk = $1139;$dB$4$shrunk = $1130;$dG$4$shrunk = $1119;$dR$4$shrunk = $1108;
       break;
      }
      case 2:  {
       $1140 = HEAP16[$dst$5>>1]|0;
       $1141 = $1140&65535;
       $1142 = HEAP32[$35>>2]|0;
       $1143 = $1141 & $1142;
       $1144 = HEAP32[$32>>2]|0;
       $1145 = $1144 & 255;
       $1146 = $1143 >>> $1145;
       $1147 = HEAP32[$31>>2]|0;
       $1148 = $1147 & 255;
       $1149 = (125924 + ($1148<<2)|0);
       $1150 = HEAP32[$1149>>2]|0;
       $1151 = (($1150) + ($1146)|0);
       $1152 = HEAP8[$1151>>0]|0;
       $1153 = HEAP32[$36>>2]|0;
       $1154 = $1153 & $1141;
       $1155 = $1144 >>> 8;
       $1156 = $1155 & 255;
       $1157 = $1154 >>> $1156;
       $1158 = $1147 >>> 8;
       $1159 = $1158 & 255;
       $1160 = (125924 + ($1159<<2)|0);
       $1161 = HEAP32[$1160>>2]|0;
       $1162 = (($1161) + ($1157)|0);
       $1163 = HEAP8[$1162>>0]|0;
       $1164 = HEAP32[$37>>2]|0;
       $1165 = $1164 & $1141;
       $1166 = $1144 >>> 16;
       $1167 = $1166 & 255;
       $1168 = $1165 >>> $1167;
       $1169 = $1147 >>> 16;
       $1170 = $1169 & 255;
       $1171 = (125924 + ($1170<<2)|0);
       $1172 = HEAP32[$1171>>2]|0;
       $1173 = (($1172) + ($1168)|0);
       $1174 = HEAP8[$1173>>0]|0;
       $1175 = HEAP32[$38>>2]|0;
       $1176 = $1175 & $1141;
       $1177 = $1144 >>> 24;
       $1178 = $1176 >>> $1177;
       $1179 = $1147 >>> 24;
       $1180 = (125924 + ($1179<<2)|0);
       $1181 = HEAP32[$1180>>2]|0;
       $1182 = (($1181) + ($1178)|0);
       $1183 = HEAP8[$1182>>0]|0;
       $dA$4$shrunk = $1183;$dB$4$shrunk = $1174;$dG$4$shrunk = $1163;$dR$4$shrunk = $1152;
       break;
      }
      case 3:  {
       $1184 = HEAP32[$32>>2]|0;
       $div15 = $1184 >>> 3;
       $1185 = $div15 & 31;
       $1186 = (($dst$5) + ($1185)|0);
       $1187 = HEAP8[$1186>>0]|0;
       $1188 = $1184 >>> 11;
       $div16 = $1188 & 31;
       $1189 = (($dst$5) + ($div16)|0);
       $1190 = HEAP8[$1189>>0]|0;
       $1191 = $1184 >>> 19;
       $1192 = $1191 & 31;
       $1193 = (($dst$5) + ($1192)|0);
       $1194 = HEAP8[$1193>>0]|0;
       $dA$4$shrunk = -1;$dB$4$shrunk = $1194;$dG$4$shrunk = $1190;$dR$4$shrunk = $1187;
       break;
      }
      case 4:  {
       $1195 = HEAP32[$dst$5>>2]|0;
       $1196 = HEAP32[$35>>2]|0;
       $1197 = $1196 & $1195;
       $1198 = HEAP32[$32>>2]|0;
       $1199 = $1198 & 255;
       $1200 = $1197 >>> $1199;
       $1201 = HEAP32[$31>>2]|0;
       $1202 = $1201 & 255;
       $1203 = (125924 + ($1202<<2)|0);
       $1204 = HEAP32[$1203>>2]|0;
       $1205 = (($1204) + ($1200)|0);
       $1206 = HEAP8[$1205>>0]|0;
       $1207 = HEAP32[$36>>2]|0;
       $1208 = $1207 & $1195;
       $1209 = $1198 >>> 8;
       $1210 = $1209 & 255;
       $1211 = $1208 >>> $1210;
       $1212 = $1201 >>> 8;
       $1213 = $1212 & 255;
       $1214 = (125924 + ($1213<<2)|0);
       $1215 = HEAP32[$1214>>2]|0;
       $1216 = (($1215) + ($1211)|0);
       $1217 = HEAP8[$1216>>0]|0;
       $1218 = HEAP32[$37>>2]|0;
       $1219 = $1218 & $1195;
       $1220 = $1198 >>> 16;
       $1221 = $1220 & 255;
       $1222 = $1219 >>> $1221;
       $1223 = $1201 >>> 16;
       $1224 = $1223 & 255;
       $1225 = (125924 + ($1224<<2)|0);
       $1226 = HEAP32[$1225>>2]|0;
       $1227 = (($1226) + ($1222)|0);
       $1228 = HEAP8[$1227>>0]|0;
       $1229 = HEAP32[$38>>2]|0;
       $1230 = $1229 & $1195;
       $1231 = $1198 >>> 24;
       $1232 = $1230 >>> $1231;
       $1233 = $1201 >>> 24;
       $1234 = (125924 + ($1233<<2)|0);
       $1235 = HEAP32[$1234>>2]|0;
       $1236 = (($1235) + ($1232)|0);
       $1237 = HEAP8[$1236>>0]|0;
       $dA$4$shrunk = $1237;$dB$4$shrunk = $1228;$dG$4$shrunk = $1217;$dR$4$shrunk = $1206;
       break;
      }
      default: {
       $dA$4$shrunk = 0;$dB$4$shrunk = 0;$dG$4$shrunk = 0;$dR$4$shrunk = 0;
      }
      }
      $dR$4 = $dR$4$shrunk&255;
      $dG$4 = $dG$4$shrunk&255;
      $dB$4 = $dB$4$shrunk&255;
      $dA$4 = $dA$4$shrunk&255;
      $1238 = (($1089) - ($dR$4))|0;
      $1239 = Math_imul($1238, $20)|0;
      $1240 = (($1239>>>0) / 255)&-1;
      $1241 = (($1240) + ($dR$4))|0;
      $1242 = (($1092) - ($dG$4))|0;
      $1243 = Math_imul($1242, $20)|0;
      $1244 = (($1243>>>0) / 255)&-1;
      $1245 = (($1244) + ($dG$4))|0;
      $1246 = (($1095) - ($dB$4))|0;
      $1247 = Math_imul($1246, $20)|0;
      $1248 = (($1247>>>0) / 255)&-1;
      $1249 = (($1248) + ($dB$4))|0;
      $1250 = (($dA$4) + ($20))|0;
      $1251 = Math_imul($dA$4, $20)|0;
      $1252 = (($1251>>>0) / 255)&-1;
      $1253 = (($1250) - ($1252))|0;
      switch ($23|0) {
      case 1:  {
       $1254 = HEAP32[$31>>2]|0;
       $1255 = $1254 & 255;
       $1256 = $1241 >>> $1255;
       $1257 = HEAP32[$32>>2]|0;
       $1258 = $1257 & 255;
       $1259 = $1256 << $1258;
       $1260 = $1254 >>> 8;
       $1261 = $1260 & 255;
       $1262 = $1245 >>> $1261;
       $1263 = $1257 >>> 8;
       $1264 = $1263 & 255;
       $1265 = $1262 << $1264;
       $1266 = $1265 | $1259;
       $1267 = $1254 >>> 16;
       $1268 = $1267 & 255;
       $1269 = $1249 >>> $1268;
       $1270 = $1257 >>> 16;
       $1271 = $1270 & 255;
       $1272 = $1269 << $1271;
       $1273 = $1266 | $1272;
       $1274 = $1254 >>> 24;
       $1275 = $1253 >>> $1274;
       $1276 = $1257 >>> 24;
       $1277 = $1275 << $1276;
       $1278 = $1273 | $1277;
       $1279 = $1278&255;
       HEAP8[$dst$5>>0] = $1279;
       break L69;
       break;
      }
      case 2:  {
       $1280 = HEAP32[$31>>2]|0;
       $1281 = $1280 & 255;
       $1282 = $1241 >>> $1281;
       $1283 = HEAP32[$32>>2]|0;
       $1284 = $1283 & 255;
       $1285 = $1282 << $1284;
       $1286 = $1280 >>> 8;
       $1287 = $1286 & 255;
       $1288 = $1245 >>> $1287;
       $1289 = $1283 >>> 8;
       $1290 = $1289 & 255;
       $1291 = $1288 << $1290;
       $1292 = $1291 | $1285;
       $1293 = $1280 >>> 16;
       $1294 = $1293 & 255;
       $1295 = $1249 >>> $1294;
       $1296 = $1283 >>> 16;
       $1297 = $1296 & 255;
       $1298 = $1295 << $1297;
       $1299 = $1292 | $1298;
       $1300 = $1280 >>> 24;
       $1301 = $1253 >>> $1300;
       $1302 = $1283 >>> 24;
       $1303 = $1301 << $1302;
       $1304 = $1299 | $1303;
       $1305 = $1304&65535;
       HEAP16[$dst$5>>1] = $1305;
       break L69;
       break;
      }
      case 3:  {
       $1306 = $1241&255;
       $1307 = HEAP8[$32>>0]|0;
       $div12 = ($1307&255) >>> 3;
       $1308 = $div12&255;
       $1309 = (($dst$5) + ($1308)|0);
       HEAP8[$1309>>0] = $1306;
       $1310 = $1245&255;
       $1311 = HEAP8[$33>>0]|0;
       $div13 = ($1311&255) >>> 3;
       $1312 = $div13&255;
       $1313 = (($dst$5) + ($1312)|0);
       HEAP8[$1313>>0] = $1310;
       $1314 = $1249&255;
       $1315 = HEAP8[$34>>0]|0;
       $div14 = ($1315&255) >>> 3;
       $1316 = $div14&255;
       $1317 = (($dst$5) + ($1316)|0);
       HEAP8[$1317>>0] = $1314;
       break L69;
       break;
      }
      case 4:  {
       $1318 = HEAP32[$31>>2]|0;
       $1319 = $1318 & 255;
       $1320 = $1241 >>> $1319;
       $1321 = HEAP32[$32>>2]|0;
       $1322 = $1321 & 255;
       $1323 = $1320 << $1322;
       $1324 = $1318 >>> 8;
       $1325 = $1324 & 255;
       $1326 = $1245 >>> $1325;
       $1327 = $1321 >>> 8;
       $1328 = $1327 & 255;
       $1329 = $1326 << $1328;
       $1330 = $1329 | $1323;
       $1331 = $1318 >>> 16;
       $1332 = $1331 & 255;
       $1333 = $1249 >>> $1332;
       $1334 = $1321 >>> 16;
       $1335 = $1334 & 255;
       $1336 = $1333 << $1335;
       $1337 = $1330 | $1336;
       $1338 = $1318 >>> 24;
       $1339 = $1253 >>> $1338;
       $1340 = $1321 >>> 24;
       $1341 = $1339 << $1340;
       $1342 = $1337 | $1341;
       HEAP32[$dst$5>>2] = $1342;
       break L69;
       break;
      }
      default: {
       break L69;
      }
      }
     }
    } while(0);
    $1343 = ((($src$5)) + 1|0);
    $1344 = (($dst$5) + ($23)|0);
    $dst$6 = $1344;$n$5 = $n$4;$src$6 = $1343;
    label = 64;
    continue;
   }
   else if ((label|0) == 64) {
    label = 0;
    $1345 = HEAP8[$src$6>>0]|0;
    $1346 = $1345&255;
    $1347 = ($1346|0)==($17|0);
    L84: do {
     if (!($1347)) {
      $1348 = (($15) + ($1346<<2)|0);
      $1349 = HEAP8[$1348>>0]|0;
      $1350 = $1349&255;
      $1351 = (((($15) + ($1346<<2)|0)) + 1|0);
      $1352 = HEAP8[$1351>>0]|0;
      $1353 = $1352&255;
      $1354 = (((($15) + ($1346<<2)|0)) + 2|0);
      $1355 = HEAP8[$1354>>0]|0;
      $1356 = $1355&255;
      switch ($23|0) {
      case 1:  {
       $1357 = HEAP8[$dst$6>>0]|0;
       $1358 = $1357&255;
       $1359 = HEAP32[$35>>2]|0;
       $1360 = $1358 & $1359;
       $1361 = HEAP32[$32>>2]|0;
       $1362 = $1361 & 255;
       $1363 = $1360 >>> $1362;
       $1364 = HEAP32[$31>>2]|0;
       $1365 = $1364 & 255;
       $1366 = (125924 + ($1365<<2)|0);
       $1367 = HEAP32[$1366>>2]|0;
       $1368 = (($1367) + ($1363)|0);
       $1369 = HEAP8[$1368>>0]|0;
       $1370 = HEAP32[$36>>2]|0;
       $1371 = $1370 & $1358;
       $1372 = $1361 >>> 8;
       $1373 = $1372 & 255;
       $1374 = $1371 >>> $1373;
       $1375 = $1364 >>> 8;
       $1376 = $1375 & 255;
       $1377 = (125924 + ($1376<<2)|0);
       $1378 = HEAP32[$1377>>2]|0;
       $1379 = (($1378) + ($1374)|0);
       $1380 = HEAP8[$1379>>0]|0;
       $1381 = HEAP32[$37>>2]|0;
       $1382 = $1381 & $1358;
       $1383 = $1361 >>> 16;
       $1384 = $1383 & 255;
       $1385 = $1382 >>> $1384;
       $1386 = $1364 >>> 16;
       $1387 = $1386 & 255;
       $1388 = (125924 + ($1387<<2)|0);
       $1389 = HEAP32[$1388>>2]|0;
       $1390 = (($1389) + ($1385)|0);
       $1391 = HEAP8[$1390>>0]|0;
       $1392 = HEAP32[$38>>2]|0;
       $1393 = $1392 & $1358;
       $1394 = $1361 >>> 24;
       $1395 = $1393 >>> $1394;
       $1396 = $1364 >>> 24;
       $1397 = (125924 + ($1396<<2)|0);
       $1398 = HEAP32[$1397>>2]|0;
       $1399 = (($1398) + ($1395)|0);
       $1400 = HEAP8[$1399>>0]|0;
       $dA$5$shrunk = $1400;$dB$5$shrunk = $1391;$dG$5$shrunk = $1380;$dR$5$shrunk = $1369;
       break;
      }
      case 2:  {
       $1401 = HEAP16[$dst$6>>1]|0;
       $1402 = $1401&65535;
       $1403 = HEAP32[$35>>2]|0;
       $1404 = $1402 & $1403;
       $1405 = HEAP32[$32>>2]|0;
       $1406 = $1405 & 255;
       $1407 = $1404 >>> $1406;
       $1408 = HEAP32[$31>>2]|0;
       $1409 = $1408 & 255;
       $1410 = (125924 + ($1409<<2)|0);
       $1411 = HEAP32[$1410>>2]|0;
       $1412 = (($1411) + ($1407)|0);
       $1413 = HEAP8[$1412>>0]|0;
       $1414 = HEAP32[$36>>2]|0;
       $1415 = $1414 & $1402;
       $1416 = $1405 >>> 8;
       $1417 = $1416 & 255;
       $1418 = $1415 >>> $1417;
       $1419 = $1408 >>> 8;
       $1420 = $1419 & 255;
       $1421 = (125924 + ($1420<<2)|0);
       $1422 = HEAP32[$1421>>2]|0;
       $1423 = (($1422) + ($1418)|0);
       $1424 = HEAP8[$1423>>0]|0;
       $1425 = HEAP32[$37>>2]|0;
       $1426 = $1425 & $1402;
       $1427 = $1405 >>> 16;
       $1428 = $1427 & 255;
       $1429 = $1426 >>> $1428;
       $1430 = $1408 >>> 16;
       $1431 = $1430 & 255;
       $1432 = (125924 + ($1431<<2)|0);
       $1433 = HEAP32[$1432>>2]|0;
       $1434 = (($1433) + ($1429)|0);
       $1435 = HEAP8[$1434>>0]|0;
       $1436 = HEAP32[$38>>2]|0;
       $1437 = $1436 & $1402;
       $1438 = $1405 >>> 24;
       $1439 = $1437 >>> $1438;
       $1440 = $1408 >>> 24;
       $1441 = (125924 + ($1440<<2)|0);
       $1442 = HEAP32[$1441>>2]|0;
       $1443 = (($1442) + ($1439)|0);
       $1444 = HEAP8[$1443>>0]|0;
       $dA$5$shrunk = $1444;$dB$5$shrunk = $1435;$dG$5$shrunk = $1424;$dR$5$shrunk = $1413;
       break;
      }
      case 3:  {
       $1445 = HEAP32[$32>>2]|0;
       $div9 = $1445 >>> 3;
       $1446 = $div9 & 31;
       $1447 = (($dst$6) + ($1446)|0);
       $1448 = HEAP8[$1447>>0]|0;
       $1449 = $1445 >>> 11;
       $div10 = $1449 & 31;
       $1450 = (($dst$6) + ($div10)|0);
       $1451 = HEAP8[$1450>>0]|0;
       $1452 = $1445 >>> 19;
       $1453 = $1452 & 31;
       $1454 = (($dst$6) + ($1453)|0);
       $1455 = HEAP8[$1454>>0]|0;
       $dA$5$shrunk = -1;$dB$5$shrunk = $1455;$dG$5$shrunk = $1451;$dR$5$shrunk = $1448;
       break;
      }
      case 4:  {
       $1456 = HEAP32[$dst$6>>2]|0;
       $1457 = HEAP32[$35>>2]|0;
       $1458 = $1457 & $1456;
       $1459 = HEAP32[$32>>2]|0;
       $1460 = $1459 & 255;
       $1461 = $1458 >>> $1460;
       $1462 = HEAP32[$31>>2]|0;
       $1463 = $1462 & 255;
       $1464 = (125924 + ($1463<<2)|0);
       $1465 = HEAP32[$1464>>2]|0;
       $1466 = (($1465) + ($1461)|0);
       $1467 = HEAP8[$1466>>0]|0;
       $1468 = HEAP32[$36>>2]|0;
       $1469 = $1468 & $1456;
       $1470 = $1459 >>> 8;
       $1471 = $1470 & 255;
       $1472 = $1469 >>> $1471;
       $1473 = $1462 >>> 8;
       $1474 = $1473 & 255;
       $1475 = (125924 + ($1474<<2)|0);
       $1476 = HEAP32[$1475>>2]|0;
       $1477 = (($1476) + ($1472)|0);
       $1478 = HEAP8[$1477>>0]|0;
       $1479 = HEAP32[$37>>2]|0;
       $1480 = $1479 & $1456;
       $1481 = $1459 >>> 16;
       $1482 = $1481 & 255;
       $1483 = $1480 >>> $1482;
       $1484 = $1462 >>> 16;
       $1485 = $1484 & 255;
       $1486 = (125924 + ($1485<<2)|0);
       $1487 = HEAP32[$1486>>2]|0;
       $1488 = (($1487) + ($1483)|0);
       $1489 = HEAP8[$1488>>0]|0;
       $1490 = HEAP32[$38>>2]|0;
       $1491 = $1490 & $1456;
       $1492 = $1459 >>> 24;
       $1493 = $1491 >>> $1492;
       $1494 = $1462 >>> 24;
       $1495 = (125924 + ($1494<<2)|0);
       $1496 = HEAP32[$1495>>2]|0;
       $1497 = (($1496) + ($1493)|0);
       $1498 = HEAP8[$1497>>0]|0;
       $dA$5$shrunk = $1498;$dB$5$shrunk = $1489;$dG$5$shrunk = $1478;$dR$5$shrunk = $1467;
       break;
      }
      default: {
       $dA$5$shrunk = 0;$dB$5$shrunk = 0;$dG$5$shrunk = 0;$dR$5$shrunk = 0;
      }
      }
      $dR$5 = $dR$5$shrunk&255;
      $dG$5 = $dG$5$shrunk&255;
      $dB$5 = $dB$5$shrunk&255;
      $dA$5 = $dA$5$shrunk&255;
      $1499 = (($1350) - ($dR$5))|0;
      $1500 = Math_imul($1499, $20)|0;
      $1501 = (($1500>>>0) / 255)&-1;
      $1502 = (($1501) + ($dR$5))|0;
      $1503 = (($1353) - ($dG$5))|0;
      $1504 = Math_imul($1503, $20)|0;
      $1505 = (($1504>>>0) / 255)&-1;
      $1506 = (($1505) + ($dG$5))|0;
      $1507 = (($1356) - ($dB$5))|0;
      $1508 = Math_imul($1507, $20)|0;
      $1509 = (($1508>>>0) / 255)&-1;
      $1510 = (($1509) + ($dB$5))|0;
      $1511 = (($dA$5) + ($20))|0;
      $1512 = Math_imul($dA$5, $20)|0;
      $1513 = (($1512>>>0) / 255)&-1;
      $1514 = (($1511) - ($1513))|0;
      switch ($23|0) {
      case 1:  {
       $1515 = HEAP32[$31>>2]|0;
       $1516 = $1515 & 255;
       $1517 = $1502 >>> $1516;
       $1518 = HEAP32[$32>>2]|0;
       $1519 = $1518 & 255;
       $1520 = $1517 << $1519;
       $1521 = $1515 >>> 8;
       $1522 = $1521 & 255;
       $1523 = $1506 >>> $1522;
       $1524 = $1518 >>> 8;
       $1525 = $1524 & 255;
       $1526 = $1523 << $1525;
       $1527 = $1526 | $1520;
       $1528 = $1515 >>> 16;
       $1529 = $1528 & 255;
       $1530 = $1510 >>> $1529;
       $1531 = $1518 >>> 16;
       $1532 = $1531 & 255;
       $1533 = $1530 << $1532;
       $1534 = $1527 | $1533;
       $1535 = $1515 >>> 24;
       $1536 = $1514 >>> $1535;
       $1537 = $1518 >>> 24;
       $1538 = $1536 << $1537;
       $1539 = $1534 | $1538;
       $1540 = $1539&255;
       HEAP8[$dst$6>>0] = $1540;
       break L84;
       break;
      }
      case 2:  {
       $1541 = HEAP32[$31>>2]|0;
       $1542 = $1541 & 255;
       $1543 = $1502 >>> $1542;
       $1544 = HEAP32[$32>>2]|0;
       $1545 = $1544 & 255;
       $1546 = $1543 << $1545;
       $1547 = $1541 >>> 8;
       $1548 = $1547 & 255;
       $1549 = $1506 >>> $1548;
       $1550 = $1544 >>> 8;
       $1551 = $1550 & 255;
       $1552 = $1549 << $1551;
       $1553 = $1552 | $1546;
       $1554 = $1541 >>> 16;
       $1555 = $1554 & 255;
       $1556 = $1510 >>> $1555;
       $1557 = $1544 >>> 16;
       $1558 = $1557 & 255;
       $1559 = $1556 << $1558;
       $1560 = $1553 | $1559;
       $1561 = $1541 >>> 24;
       $1562 = $1514 >>> $1561;
       $1563 = $1544 >>> 24;
       $1564 = $1562 << $1563;
       $1565 = $1560 | $1564;
       $1566 = $1565&65535;
       HEAP16[$dst$6>>1] = $1566;
       break L84;
       break;
      }
      case 3:  {
       $1567 = $1502&255;
       $1568 = HEAP8[$32>>0]|0;
       $div6 = ($1568&255) >>> 3;
       $1569 = $div6&255;
       $1570 = (($dst$6) + ($1569)|0);
       HEAP8[$1570>>0] = $1567;
       $1571 = $1506&255;
       $1572 = HEAP8[$33>>0]|0;
       $div7 = ($1572&255) >>> 3;
       $1573 = $div7&255;
       $1574 = (($dst$6) + ($1573)|0);
       HEAP8[$1574>>0] = $1571;
       $1575 = $1510&255;
       $1576 = HEAP8[$34>>0]|0;
       $div8 = ($1576&255) >>> 3;
       $1577 = $div8&255;
       $1578 = (($dst$6) + ($1577)|0);
       HEAP8[$1578>>0] = $1575;
       break L84;
       break;
      }
      case 4:  {
       $1579 = HEAP32[$31>>2]|0;
       $1580 = $1579 & 255;
       $1581 = $1502 >>> $1580;
       $1582 = HEAP32[$32>>2]|0;
       $1583 = $1582 & 255;
       $1584 = $1581 << $1583;
       $1585 = $1579 >>> 8;
       $1586 = $1585 & 255;
       $1587 = $1506 >>> $1586;
       $1588 = $1582 >>> 8;
       $1589 = $1588 & 255;
       $1590 = $1587 << $1589;
       $1591 = $1590 | $1584;
       $1592 = $1579 >>> 16;
       $1593 = $1592 & 255;
       $1594 = $1510 >>> $1593;
       $1595 = $1582 >>> 16;
       $1596 = $1595 & 255;
       $1597 = $1594 << $1596;
       $1598 = $1591 | $1597;
       $1599 = $1579 >>> 24;
       $1600 = $1514 >>> $1599;
       $1601 = $1582 >>> 24;
       $1602 = $1600 << $1601;
       $1603 = $1598 | $1602;
       HEAP32[$dst$6>>2] = $1603;
       break L84;
       break;
      }
      default: {
       break L84;
      }
      }
     }
    } while(0);
    $1604 = ((($src$6)) + 1|0);
    $1605 = (($dst$6) + ($23)|0);
    $dst$7 = $1605;$n$6 = $n$5;$src$7 = $1604;
    label = 76;
    continue;
   }
   else if ((label|0) == 76) {
    label = 0;
    $1606 = HEAP8[$src$7>>0]|0;
    $1607 = $1606&255;
    $1608 = ($1607|0)==($17|0);
    L99: do {
     if (!($1608)) {
      $1609 = (($15) + ($1607<<2)|0);
      $1610 = HEAP8[$1609>>0]|0;
      $1611 = $1610&255;
      $1612 = (((($15) + ($1607<<2)|0)) + 1|0);
      $1613 = HEAP8[$1612>>0]|0;
      $1614 = $1613&255;
      $1615 = (((($15) + ($1607<<2)|0)) + 2|0);
      $1616 = HEAP8[$1615>>0]|0;
      $1617 = $1616&255;
      switch ($23|0) {
      case 1:  {
       $1618 = HEAP8[$dst$7>>0]|0;
       $1619 = $1618&255;
       $1620 = HEAP32[$35>>2]|0;
       $1621 = $1619 & $1620;
       $1622 = HEAP32[$32>>2]|0;
       $1623 = $1622 & 255;
       $1624 = $1621 >>> $1623;
       $1625 = HEAP32[$31>>2]|0;
       $1626 = $1625 & 255;
       $1627 = (125924 + ($1626<<2)|0);
       $1628 = HEAP32[$1627>>2]|0;
       $1629 = (($1628) + ($1624)|0);
       $1630 = HEAP8[$1629>>0]|0;
       $1631 = HEAP32[$36>>2]|0;
       $1632 = $1631 & $1619;
       $1633 = $1622 >>> 8;
       $1634 = $1633 & 255;
       $1635 = $1632 >>> $1634;
       $1636 = $1625 >>> 8;
       $1637 = $1636 & 255;
       $1638 = (125924 + ($1637<<2)|0);
       $1639 = HEAP32[$1638>>2]|0;
       $1640 = (($1639) + ($1635)|0);
       $1641 = HEAP8[$1640>>0]|0;
       $1642 = HEAP32[$37>>2]|0;
       $1643 = $1642 & $1619;
       $1644 = $1622 >>> 16;
       $1645 = $1644 & 255;
       $1646 = $1643 >>> $1645;
       $1647 = $1625 >>> 16;
       $1648 = $1647 & 255;
       $1649 = (125924 + ($1648<<2)|0);
       $1650 = HEAP32[$1649>>2]|0;
       $1651 = (($1650) + ($1646)|0);
       $1652 = HEAP8[$1651>>0]|0;
       $1653 = HEAP32[$38>>2]|0;
       $1654 = $1653 & $1619;
       $1655 = $1622 >>> 24;
       $1656 = $1654 >>> $1655;
       $1657 = $1625 >>> 24;
       $1658 = (125924 + ($1657<<2)|0);
       $1659 = HEAP32[$1658>>2]|0;
       $1660 = (($1659) + ($1656)|0);
       $1661 = HEAP8[$1660>>0]|0;
       $dA$6$shrunk = $1661;$dB$6$shrunk = $1652;$dG$6$shrunk = $1641;$dR$6$shrunk = $1630;
       break;
      }
      case 2:  {
       $1662 = HEAP16[$dst$7>>1]|0;
       $1663 = $1662&65535;
       $1664 = HEAP32[$35>>2]|0;
       $1665 = $1663 & $1664;
       $1666 = HEAP32[$32>>2]|0;
       $1667 = $1666 & 255;
       $1668 = $1665 >>> $1667;
       $1669 = HEAP32[$31>>2]|0;
       $1670 = $1669 & 255;
       $1671 = (125924 + ($1670<<2)|0);
       $1672 = HEAP32[$1671>>2]|0;
       $1673 = (($1672) + ($1668)|0);
       $1674 = HEAP8[$1673>>0]|0;
       $1675 = HEAP32[$36>>2]|0;
       $1676 = $1675 & $1663;
       $1677 = $1666 >>> 8;
       $1678 = $1677 & 255;
       $1679 = $1676 >>> $1678;
       $1680 = $1669 >>> 8;
       $1681 = $1680 & 255;
       $1682 = (125924 + ($1681<<2)|0);
       $1683 = HEAP32[$1682>>2]|0;
       $1684 = (($1683) + ($1679)|0);
       $1685 = HEAP8[$1684>>0]|0;
       $1686 = HEAP32[$37>>2]|0;
       $1687 = $1686 & $1663;
       $1688 = $1666 >>> 16;
       $1689 = $1688 & 255;
       $1690 = $1687 >>> $1689;
       $1691 = $1669 >>> 16;
       $1692 = $1691 & 255;
       $1693 = (125924 + ($1692<<2)|0);
       $1694 = HEAP32[$1693>>2]|0;
       $1695 = (($1694) + ($1690)|0);
       $1696 = HEAP8[$1695>>0]|0;
       $1697 = HEAP32[$38>>2]|0;
       $1698 = $1697 & $1663;
       $1699 = $1666 >>> 24;
       $1700 = $1698 >>> $1699;
       $1701 = $1669 >>> 24;
       $1702 = (125924 + ($1701<<2)|0);
       $1703 = HEAP32[$1702>>2]|0;
       $1704 = (($1703) + ($1700)|0);
       $1705 = HEAP8[$1704>>0]|0;
       $dA$6$shrunk = $1705;$dB$6$shrunk = $1696;$dG$6$shrunk = $1685;$dR$6$shrunk = $1674;
       break;
      }
      case 3:  {
       $1706 = HEAP32[$32>>2]|0;
       $div3 = $1706 >>> 3;
       $1707 = $div3 & 31;
       $1708 = (($dst$7) + ($1707)|0);
       $1709 = HEAP8[$1708>>0]|0;
       $1710 = $1706 >>> 11;
       $div4 = $1710 & 31;
       $1711 = (($dst$7) + ($div4)|0);
       $1712 = HEAP8[$1711>>0]|0;
       $1713 = $1706 >>> 19;
       $1714 = $1713 & 31;
       $1715 = (($dst$7) + ($1714)|0);
       $1716 = HEAP8[$1715>>0]|0;
       $dA$6$shrunk = -1;$dB$6$shrunk = $1716;$dG$6$shrunk = $1712;$dR$6$shrunk = $1709;
       break;
      }
      case 4:  {
       $1717 = HEAP32[$dst$7>>2]|0;
       $1718 = HEAP32[$35>>2]|0;
       $1719 = $1718 & $1717;
       $1720 = HEAP32[$32>>2]|0;
       $1721 = $1720 & 255;
       $1722 = $1719 >>> $1721;
       $1723 = HEAP32[$31>>2]|0;
       $1724 = $1723 & 255;
       $1725 = (125924 + ($1724<<2)|0);
       $1726 = HEAP32[$1725>>2]|0;
       $1727 = (($1726) + ($1722)|0);
       $1728 = HEAP8[$1727>>0]|0;
       $1729 = HEAP32[$36>>2]|0;
       $1730 = $1729 & $1717;
       $1731 = $1720 >>> 8;
       $1732 = $1731 & 255;
       $1733 = $1730 >>> $1732;
       $1734 = $1723 >>> 8;
       $1735 = $1734 & 255;
       $1736 = (125924 + ($1735<<2)|0);
       $1737 = HEAP32[$1736>>2]|0;
       $1738 = (($1737) + ($1733)|0);
       $1739 = HEAP8[$1738>>0]|0;
       $1740 = HEAP32[$37>>2]|0;
       $1741 = $1740 & $1717;
       $1742 = $1720 >>> 16;
       $1743 = $1742 & 255;
       $1744 = $1741 >>> $1743;
       $1745 = $1723 >>> 16;
       $1746 = $1745 & 255;
       $1747 = (125924 + ($1746<<2)|0);
       $1748 = HEAP32[$1747>>2]|0;
       $1749 = (($1748) + ($1744)|0);
       $1750 = HEAP8[$1749>>0]|0;
       $1751 = HEAP32[$38>>2]|0;
       $1752 = $1751 & $1717;
       $1753 = $1720 >>> 24;
       $1754 = $1752 >>> $1753;
       $1755 = $1723 >>> 24;
       $1756 = (125924 + ($1755<<2)|0);
       $1757 = HEAP32[$1756>>2]|0;
       $1758 = (($1757) + ($1754)|0);
       $1759 = HEAP8[$1758>>0]|0;
       $dA$6$shrunk = $1759;$dB$6$shrunk = $1750;$dG$6$shrunk = $1739;$dR$6$shrunk = $1728;
       break;
      }
      default: {
       $dA$6$shrunk = 0;$dB$6$shrunk = 0;$dG$6$shrunk = 0;$dR$6$shrunk = 0;
      }
      }
      $dR$6 = $dR$6$shrunk&255;
      $dG$6 = $dG$6$shrunk&255;
      $dB$6 = $dB$6$shrunk&255;
      $dA$6 = $dA$6$shrunk&255;
      $1760 = (($1611) - ($dR$6))|0;
      $1761 = Math_imul($1760, $20)|0;
      $1762 = (($1761>>>0) / 255)&-1;
      $1763 = (($1762) + ($dR$6))|0;
      $1764 = (($1614) - ($dG$6))|0;
      $1765 = Math_imul($1764, $20)|0;
      $1766 = (($1765>>>0) / 255)&-1;
      $1767 = (($1766) + ($dG$6))|0;
      $1768 = (($1617) - ($dB$6))|0;
      $1769 = Math_imul($1768, $20)|0;
      $1770 = (($1769>>>0) / 255)&-1;
      $1771 = (($1770) + ($dB$6))|0;
      $1772 = (($dA$6) + ($20))|0;
      $1773 = Math_imul($dA$6, $20)|0;
      $1774 = (($1773>>>0) / 255)&-1;
      $1775 = (($1772) - ($1774))|0;
      switch ($23|0) {
      case 1:  {
       $1776 = HEAP32[$31>>2]|0;
       $1777 = $1776 & 255;
       $1778 = $1763 >>> $1777;
       $1779 = HEAP32[$32>>2]|0;
       $1780 = $1779 & 255;
       $1781 = $1778 << $1780;
       $1782 = $1776 >>> 8;
       $1783 = $1782 & 255;
       $1784 = $1767 >>> $1783;
       $1785 = $1779 >>> 8;
       $1786 = $1785 & 255;
       $1787 = $1784 << $1786;
       $1788 = $1787 | $1781;
       $1789 = $1776 >>> 16;
       $1790 = $1789 & 255;
       $1791 = $1771 >>> $1790;
       $1792 = $1779 >>> 16;
       $1793 = $1792 & 255;
       $1794 = $1791 << $1793;
       $1795 = $1788 | $1794;
       $1796 = $1776 >>> 24;
       $1797 = $1775 >>> $1796;
       $1798 = $1779 >>> 24;
       $1799 = $1797 << $1798;
       $1800 = $1795 | $1799;
       $1801 = $1800&255;
       HEAP8[$dst$7>>0] = $1801;
       break L99;
       break;
      }
      case 2:  {
       $1802 = HEAP32[$31>>2]|0;
       $1803 = $1802 & 255;
       $1804 = $1763 >>> $1803;
       $1805 = HEAP32[$32>>2]|0;
       $1806 = $1805 & 255;
       $1807 = $1804 << $1806;
       $1808 = $1802 >>> 8;
       $1809 = $1808 & 255;
       $1810 = $1767 >>> $1809;
       $1811 = $1805 >>> 8;
       $1812 = $1811 & 255;
       $1813 = $1810 << $1812;
       $1814 = $1813 | $1807;
       $1815 = $1802 >>> 16;
       $1816 = $1815 & 255;
       $1817 = $1771 >>> $1816;
       $1818 = $1805 >>> 16;
       $1819 = $1818 & 255;
       $1820 = $1817 << $1819;
       $1821 = $1814 | $1820;
       $1822 = $1802 >>> 24;
       $1823 = $1775 >>> $1822;
       $1824 = $1805 >>> 24;
       $1825 = $1823 << $1824;
       $1826 = $1821 | $1825;
       $1827 = $1826&65535;
       HEAP16[$dst$7>>1] = $1827;
       break L99;
       break;
      }
      case 3:  {
       $1828 = $1763&255;
       $1829 = HEAP8[$32>>0]|0;
       $div = ($1829&255) >>> 3;
       $1830 = $div&255;
       $1831 = (($dst$7) + ($1830)|0);
       HEAP8[$1831>>0] = $1828;
       $1832 = $1767&255;
       $1833 = HEAP8[$33>>0]|0;
       $div1 = ($1833&255) >>> 3;
       $1834 = $div1&255;
       $1835 = (($dst$7) + ($1834)|0);
       HEAP8[$1835>>0] = $1832;
       $1836 = $1771&255;
       $1837 = HEAP8[$34>>0]|0;
       $div2 = ($1837&255) >>> 3;
       $1838 = $div2&255;
       $1839 = (($dst$7) + ($1838)|0);
       HEAP8[$1839>>0] = $1836;
       break L99;
       break;
      }
      case 4:  {
       $1840 = HEAP32[$31>>2]|0;
       $1841 = $1840 & 255;
       $1842 = $1763 >>> $1841;
       $1843 = HEAP32[$32>>2]|0;
       $1844 = $1843 & 255;
       $1845 = $1842 << $1844;
       $1846 = $1840 >>> 8;
       $1847 = $1846 & 255;
       $1848 = $1767 >>> $1847;
       $1849 = $1843 >>> 8;
       $1850 = $1849 & 255;
       $1851 = $1848 << $1850;
       $1852 = $1851 | $1845;
       $1853 = $1840 >>> 16;
       $1854 = $1853 & 255;
       $1855 = $1771 >>> $1854;
       $1856 = $1843 >>> 16;
       $1857 = $1856 & 255;
       $1858 = $1855 << $1857;
       $1859 = $1852 | $1858;
       $1860 = $1840 >>> 24;
       $1861 = $1775 >>> $1860;
       $1862 = $1843 >>> 24;
       $1863 = $1861 << $1862;
       $1864 = $1859 | $1863;
       HEAP32[$dst$7>>2] = $1864;
       break L99;
       break;
      }
      default: {
       break L99;
      }
      }
     }
    } while(0);
    $1865 = ((($src$7)) + 1|0);
    $1866 = (($dst$7) + ($23)|0);
    $dst$8 = $1866;$n$7 = $n$6;$src$8 = $1865;
    label = 88;
    continue;
   }
   else if ((label|0) == 88) {
    label = 0;
    $1867 = HEAP8[$src$8>>0]|0;
    $1868 = $1867&255;
    $1869 = ($1868|0)==($17|0);
    L114: do {
     if (!($1869)) {
      $1870 = (($15) + ($1868<<2)|0);
      $1871 = HEAP8[$1870>>0]|0;
      $1872 = $1871&255;
      $1873 = (((($15) + ($1868<<2)|0)) + 1|0);
      $1874 = HEAP8[$1873>>0]|0;
      $1875 = $1874&255;
      $1876 = (((($15) + ($1868<<2)|0)) + 2|0);
      $1877 = HEAP8[$1876>>0]|0;
      $1878 = $1877&255;
      switch ($23|0) {
      case 1:  {
       $1879 = HEAP8[$dst$8>>0]|0;
       $1880 = $1879&255;
       $1881 = HEAP32[$35>>2]|0;
       $1882 = $1880 & $1881;
       $1883 = HEAP32[$32>>2]|0;
       $1884 = $1883 & 255;
       $1885 = $1882 >>> $1884;
       $1886 = HEAP32[$31>>2]|0;
       $1887 = $1886 & 255;
       $1888 = (125924 + ($1887<<2)|0);
       $1889 = HEAP32[$1888>>2]|0;
       $1890 = (($1889) + ($1885)|0);
       $1891 = HEAP8[$1890>>0]|0;
       $1892 = HEAP32[$36>>2]|0;
       $1893 = $1892 & $1880;
       $1894 = $1883 >>> 8;
       $1895 = $1894 & 255;
       $1896 = $1893 >>> $1895;
       $1897 = $1886 >>> 8;
       $1898 = $1897 & 255;
       $1899 = (125924 + ($1898<<2)|0);
       $1900 = HEAP32[$1899>>2]|0;
       $1901 = (($1900) + ($1896)|0);
       $1902 = HEAP8[$1901>>0]|0;
       $1903 = HEAP32[$37>>2]|0;
       $1904 = $1903 & $1880;
       $1905 = $1883 >>> 16;
       $1906 = $1905 & 255;
       $1907 = $1904 >>> $1906;
       $1908 = $1886 >>> 16;
       $1909 = $1908 & 255;
       $1910 = (125924 + ($1909<<2)|0);
       $1911 = HEAP32[$1910>>2]|0;
       $1912 = (($1911) + ($1907)|0);
       $1913 = HEAP8[$1912>>0]|0;
       $1914 = HEAP32[$38>>2]|0;
       $1915 = $1914 & $1880;
       $1916 = $1883 >>> 24;
       $1917 = $1915 >>> $1916;
       $1918 = $1886 >>> 24;
       $1919 = (125924 + ($1918<<2)|0);
       $1920 = HEAP32[$1919>>2]|0;
       $1921 = (($1920) + ($1917)|0);
       $1922 = HEAP8[$1921>>0]|0;
       $dA$7$shrunk = $1922;$dB$7$shrunk = $1913;$dG$7$shrunk = $1902;$dR$7$shrunk = $1891;
       break;
      }
      case 2:  {
       $1923 = HEAP16[$dst$8>>1]|0;
       $1924 = $1923&65535;
       $1925 = HEAP32[$35>>2]|0;
       $1926 = $1924 & $1925;
       $1927 = HEAP32[$32>>2]|0;
       $1928 = $1927 & 255;
       $1929 = $1926 >>> $1928;
       $1930 = HEAP32[$31>>2]|0;
       $1931 = $1930 & 255;
       $1932 = (125924 + ($1931<<2)|0);
       $1933 = HEAP32[$1932>>2]|0;
       $1934 = (($1933) + ($1929)|0);
       $1935 = HEAP8[$1934>>0]|0;
       $1936 = HEAP32[$36>>2]|0;
       $1937 = $1936 & $1924;
       $1938 = $1927 >>> 8;
       $1939 = $1938 & 255;
       $1940 = $1937 >>> $1939;
       $1941 = $1930 >>> 8;
       $1942 = $1941 & 255;
       $1943 = (125924 + ($1942<<2)|0);
       $1944 = HEAP32[$1943>>2]|0;
       $1945 = (($1944) + ($1940)|0);
       $1946 = HEAP8[$1945>>0]|0;
       $1947 = HEAP32[$37>>2]|0;
       $1948 = $1947 & $1924;
       $1949 = $1927 >>> 16;
       $1950 = $1949 & 255;
       $1951 = $1948 >>> $1950;
       $1952 = $1930 >>> 16;
       $1953 = $1952 & 255;
       $1954 = (125924 + ($1953<<2)|0);
       $1955 = HEAP32[$1954>>2]|0;
       $1956 = (($1955) + ($1951)|0);
       $1957 = HEAP8[$1956>>0]|0;
       $1958 = HEAP32[$38>>2]|0;
       $1959 = $1958 & $1924;
       $1960 = $1927 >>> 24;
       $1961 = $1959 >>> $1960;
       $1962 = $1930 >>> 24;
       $1963 = (125924 + ($1962<<2)|0);
       $1964 = HEAP32[$1963>>2]|0;
       $1965 = (($1964) + ($1961)|0);
       $1966 = HEAP8[$1965>>0]|0;
       $dA$7$shrunk = $1966;$dB$7$shrunk = $1957;$dG$7$shrunk = $1946;$dR$7$shrunk = $1935;
       break;
      }
      case 3:  {
       $1967 = HEAP32[$32>>2]|0;
       $div45 = $1967 >>> 3;
       $1968 = $div45 & 31;
       $1969 = (($dst$8) + ($1968)|0);
       $1970 = HEAP8[$1969>>0]|0;
       $1971 = $1967 >>> 11;
       $div46 = $1971 & 31;
       $1972 = (($dst$8) + ($div46)|0);
       $1973 = HEAP8[$1972>>0]|0;
       $1974 = $1967 >>> 19;
       $1975 = $1974 & 31;
       $1976 = (($dst$8) + ($1975)|0);
       $1977 = HEAP8[$1976>>0]|0;
       $dA$7$shrunk = -1;$dB$7$shrunk = $1977;$dG$7$shrunk = $1973;$dR$7$shrunk = $1970;
       break;
      }
      case 4:  {
       $1978 = HEAP32[$dst$8>>2]|0;
       $1979 = HEAP32[$35>>2]|0;
       $1980 = $1979 & $1978;
       $1981 = HEAP32[$32>>2]|0;
       $1982 = $1981 & 255;
       $1983 = $1980 >>> $1982;
       $1984 = HEAP32[$31>>2]|0;
       $1985 = $1984 & 255;
       $1986 = (125924 + ($1985<<2)|0);
       $1987 = HEAP32[$1986>>2]|0;
       $1988 = (($1987) + ($1983)|0);
       $1989 = HEAP8[$1988>>0]|0;
       $1990 = HEAP32[$36>>2]|0;
       $1991 = $1990 & $1978;
       $1992 = $1981 >>> 8;
       $1993 = $1992 & 255;
       $1994 = $1991 >>> $1993;
       $1995 = $1984 >>> 8;
       $1996 = $1995 & 255;
       $1997 = (125924 + ($1996<<2)|0);
       $1998 = HEAP32[$1997>>2]|0;
       $1999 = (($1998) + ($1994)|0);
       $2000 = HEAP8[$1999>>0]|0;
       $2001 = HEAP32[$37>>2]|0;
       $2002 = $2001 & $1978;
       $2003 = $1981 >>> 16;
       $2004 = $2003 & 255;
       $2005 = $2002 >>> $2004;
       $2006 = $1984 >>> 16;
       $2007 = $2006 & 255;
       $2008 = (125924 + ($2007<<2)|0);
       $2009 = HEAP32[$2008>>2]|0;
       $2010 = (($2009) + ($2005)|0);
       $2011 = HEAP8[$2010>>0]|0;
       $2012 = HEAP32[$38>>2]|0;
       $2013 = $2012 & $1978;
       $2014 = $1981 >>> 24;
       $2015 = $2013 >>> $2014;
       $2016 = $1984 >>> 24;
       $2017 = (125924 + ($2016<<2)|0);
       $2018 = HEAP32[$2017>>2]|0;
       $2019 = (($2018) + ($2015)|0);
       $2020 = HEAP8[$2019>>0]|0;
       $dA$7$shrunk = $2020;$dB$7$shrunk = $2011;$dG$7$shrunk = $2000;$dR$7$shrunk = $1989;
       break;
      }
      default: {
       $dA$7$shrunk = 0;$dB$7$shrunk = 0;$dG$7$shrunk = 0;$dR$7$shrunk = 0;
      }
      }
      $dR$7 = $dR$7$shrunk&255;
      $dG$7 = $dG$7$shrunk&255;
      $dB$7 = $dB$7$shrunk&255;
      $dA$7 = $dA$7$shrunk&255;
      $2021 = (($1872) - ($dR$7))|0;
      $2022 = Math_imul($2021, $20)|0;
      $2023 = (($2022>>>0) / 255)&-1;
      $2024 = (($2023) + ($dR$7))|0;
      $2025 = (($1875) - ($dG$7))|0;
      $2026 = Math_imul($2025, $20)|0;
      $2027 = (($2026>>>0) / 255)&-1;
      $2028 = (($2027) + ($dG$7))|0;
      $2029 = (($1878) - ($dB$7))|0;
      $2030 = Math_imul($2029, $20)|0;
      $2031 = (($2030>>>0) / 255)&-1;
      $2032 = (($2031) + ($dB$7))|0;
      $2033 = (($dA$7) + ($20))|0;
      $2034 = Math_imul($dA$7, $20)|0;
      $2035 = (($2034>>>0) / 255)&-1;
      $2036 = (($2033) - ($2035))|0;
      switch ($23|0) {
      case 1:  {
       $2037 = HEAP32[$31>>2]|0;
       $2038 = $2037 & 255;
       $2039 = $2024 >>> $2038;
       $2040 = HEAP32[$32>>2]|0;
       $2041 = $2040 & 255;
       $2042 = $2039 << $2041;
       $2043 = $2037 >>> 8;
       $2044 = $2043 & 255;
       $2045 = $2028 >>> $2044;
       $2046 = $2040 >>> 8;
       $2047 = $2046 & 255;
       $2048 = $2045 << $2047;
       $2049 = $2048 | $2042;
       $2050 = $2037 >>> 16;
       $2051 = $2050 & 255;
       $2052 = $2032 >>> $2051;
       $2053 = $2040 >>> 16;
       $2054 = $2053 & 255;
       $2055 = $2052 << $2054;
       $2056 = $2049 | $2055;
       $2057 = $2037 >>> 24;
       $2058 = $2036 >>> $2057;
       $2059 = $2040 >>> 24;
       $2060 = $2058 << $2059;
       $2061 = $2056 | $2060;
       $2062 = $2061&255;
       HEAP8[$dst$8>>0] = $2062;
       break L114;
       break;
      }
      case 2:  {
       $2063 = HEAP32[$31>>2]|0;
       $2064 = $2063 & 255;
       $2065 = $2024 >>> $2064;
       $2066 = HEAP32[$32>>2]|0;
       $2067 = $2066 & 255;
       $2068 = $2065 << $2067;
       $2069 = $2063 >>> 8;
       $2070 = $2069 & 255;
       $2071 = $2028 >>> $2070;
       $2072 = $2066 >>> 8;
       $2073 = $2072 & 255;
       $2074 = $2071 << $2073;
       $2075 = $2074 | $2068;
       $2076 = $2063 >>> 16;
       $2077 = $2076 & 255;
       $2078 = $2032 >>> $2077;
       $2079 = $2066 >>> 16;
       $2080 = $2079 & 255;
       $2081 = $2078 << $2080;
       $2082 = $2075 | $2081;
       $2083 = $2063 >>> 24;
       $2084 = $2036 >>> $2083;
       $2085 = $2066 >>> 24;
       $2086 = $2084 << $2085;
       $2087 = $2082 | $2086;
       $2088 = $2087&65535;
       HEAP16[$dst$8>>1] = $2088;
       break L114;
       break;
      }
      case 3:  {
       $2089 = $2024&255;
       $2090 = HEAP8[$32>>0]|0;
       $div42 = ($2090&255) >>> 3;
       $2091 = $div42&255;
       $2092 = (($dst$8) + ($2091)|0);
       HEAP8[$2092>>0] = $2089;
       $2093 = $2028&255;
       $2094 = HEAP8[$33>>0]|0;
       $div43 = ($2094&255) >>> 3;
       $2095 = $div43&255;
       $2096 = (($dst$8) + ($2095)|0);
       HEAP8[$2096>>0] = $2093;
       $2097 = $2032&255;
       $2098 = HEAP8[$34>>0]|0;
       $div44 = ($2098&255) >>> 3;
       $2099 = $div44&255;
       $2100 = (($dst$8) + ($2099)|0);
       HEAP8[$2100>>0] = $2097;
       break L114;
       break;
      }
      case 4:  {
       $2101 = HEAP32[$31>>2]|0;
       $2102 = $2101 & 255;
       $2103 = $2024 >>> $2102;
       $2104 = HEAP32[$32>>2]|0;
       $2105 = $2104 & 255;
       $2106 = $2103 << $2105;
       $2107 = $2101 >>> 8;
       $2108 = $2107 & 255;
       $2109 = $2028 >>> $2108;
       $2110 = $2104 >>> 8;
       $2111 = $2110 & 255;
       $2112 = $2109 << $2111;
       $2113 = $2112 | $2106;
       $2114 = $2101 >>> 16;
       $2115 = $2114 & 255;
       $2116 = $2032 >>> $2115;
       $2117 = $2104 >>> 16;
       $2118 = $2117 & 255;
       $2119 = $2116 << $2118;
       $2120 = $2113 | $2119;
       $2121 = $2101 >>> 24;
       $2122 = $2036 >>> $2121;
       $2123 = $2104 >>> 24;
       $2124 = $2122 << $2123;
       $2125 = $2120 | $2124;
       HEAP32[$dst$8>>2] = $2125;
       break L114;
       break;
      }
      default: {
       break L114;
      }
      }
     }
    } while(0);
    $2126 = ((($src$8)) + 1|0);
    $2127 = (($dst$8) + ($23)|0);
    $2128 = (($n$7) + -1)|0;
    $2129 = ($n$7|0)>(1);
    if ($2129) {
     $dst$1 = $2127;$n$0 = $2128;$src$1 = $2126;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $2130 = (($2126) + ($5)|0);
  $2131 = (($2127) + ($7)|0);
  $2132 = ($39|0)==(0);
  if ($2132) {
   label = 102;
   break;
  } else {
   $$in = $39;$dst$055 = $2131;$src$056 = $2130;
  }
 }
 if ((label|0) == 100) {
  // unreachable;
 }
 else if ((label|0) == 102) {
  return;
 }
}
function _SDL_CalculateBlitA($surface) {
 $surface = $surface|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $BlitNto1SurfaceAlphaKey$BlitNtoNSurfaceAlphaKey = 0, $switch$select = 0, $switch$select2 = 0, $switch$select4 = 0;
 var $switch$select6 = 0, $switch$selectcmp = 0, $switch$selectcmp1 = 0, $switch$selectcmp3 = 0, $switch$selectcmp5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($surface)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($surface)) + 52|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($3)) + 68|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $8 & -28673;
 switch ($9|0) {
 case 16:  {
  $10 = ((($6)) + 9|0);
  $11 = HEAP8[$10>>0]|0;
  $12 = $11&255;
  switch ($12|0) {
  case 2:  {
   $13 = ((($1)) + 9|0);
   $14 = HEAP8[$13>>0]|0;
   $15 = ($14<<24>>24)==(4);
   if (!($15)) {
    $$0 = 4497;
    return ($$0|0);
   }
   $16 = ((($1)) + 24|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ($17|0)==(-16777216);
   if (!($18)) {
    $$0 = 4497;
    return ($$0|0);
   }
   $19 = ((($1)) + 16|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = ($20|0)==(65280);
   if (!($21)) {
    $$0 = 4497;
    return ($$0|0);
   }
   $22 = ((($1)) + 12|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = ($23|0)==(255);
   if ($24) {
    $25 = ((($6)) + 12|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ($26|0)==(31);
    if (!($27)) {
     label = 8;
    }
   } else {
    label = 8;
   }
   if ((label|0) == 8) {
    $28 = ((($1)) + 20|0);
    $29 = HEAP32[$28>>2]|0;
    $30 = ($29|0)==(255);
    if (!($30)) {
     $$0 = 4497;
     return ($$0|0);
    }
    $31 = ((($6)) + 20|0);
    $32 = HEAP32[$31>>2]|0;
    $33 = ($32|0)==(31);
    if (!($33)) {
     $$0 = 4497;
     return ($$0|0);
    }
   }
   $34 = ((($6)) + 16|0);
   $35 = HEAP32[$34>>2]|0;
   $switch$selectcmp = ($35|0)==(992);
   $switch$select = $switch$selectcmp ? 4496 : 4497;
   $switch$selectcmp1 = ($35|0)==(2016);
   $switch$select2 = $switch$selectcmp1 ? 4498 : $switch$select;
   $$0 = $switch$select2;
   return ($$0|0);
   break;
  }
  case 4:  {
   $36 = ((($1)) + 12|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = ((($6)) + 12|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = ($37|0)==($39|0);
   if ($40) {
    $41 = ((($1)) + 16|0);
    $42 = HEAP32[$41>>2]|0;
    $43 = ((($6)) + 16|0);
    $44 = HEAP32[$43>>2]|0;
    $45 = ($42|0)==($44|0);
    if ($45) {
     $46 = ((($1)) + 20|0);
     $47 = HEAP32[$46>>2]|0;
     $48 = ((($6)) + 20|0);
     $49 = HEAP32[$48>>2]|0;
     $50 = ($47|0)==($49|0);
     if ($50) {
      $51 = ((($1)) + 9|0);
      $52 = HEAP8[$51>>0]|0;
      $53 = ($52<<24>>24)==(4);
      if ($53) {
       $54 = ((($1)) + 24|0);
       $55 = HEAP32[$54>>2]|0;
       $56 = ($55|0)==(-16777216);
       if ($56) {
        $$0 = 4505;
        return ($$0|0);
       }
      }
     }
    }
   }
   $$0 = 4497;
   return ($$0|0);
   break;
  }
  case 1:  {
   $$0 = 4504;
   return ($$0|0);
   break;
  }
  default: {
   $$0 = 4497;
   return ($$0|0);
  }
  }
  break;
 }
 case 18:  {
  $57 = ((($1)) + 24|0);
  $58 = HEAP32[$57>>2]|0;
  $59 = ($58|0)==(0);
  if (!($59)) {
   $$0 = 0;
   return ($$0|0);
  }
  $60 = ((($6)) + 9|0);
  $61 = HEAP8[$60>>0]|0;
  $62 = $61&255;
  switch ($62|0) {
  case 2:  {
   $63 = ((($3)) + 4|0);
   $64 = HEAP32[$63>>2]|0;
   $65 = ($64|0)==(0);
   if ($65) {
    $$0 = 4500;
    return ($$0|0);
   }
   $66 = ((($6)) + 16|0);
   $67 = HEAP32[$66>>2]|0;
   $switch$selectcmp3 = ($67|0)==(992);
   $switch$select4 = $switch$selectcmp3 ? 4499 : 4500;
   $switch$selectcmp5 = ($67|0)==(2016);
   $switch$select6 = $switch$selectcmp5 ? 4501 : $switch$select4;
   $$0 = $switch$select6;
   return ($$0|0);
   break;
  }
  case 4:  {
   $68 = ((($1)) + 12|0);
   $69 = HEAP32[$68>>2]|0;
   $70 = ((($6)) + 12|0);
   $71 = HEAP32[$70>>2]|0;
   $72 = ($69|0)==($71|0);
   if ($72) {
    $73 = ((($1)) + 16|0);
    $74 = HEAP32[$73>>2]|0;
    $75 = ((($6)) + 16|0);
    $76 = HEAP32[$75>>2]|0;
    $77 = ($74|0)==($76|0);
    if ($77) {
     $78 = ((($1)) + 20|0);
     $79 = HEAP32[$78>>2]|0;
     $80 = ((($6)) + 20|0);
     $81 = HEAP32[$80>>2]|0;
     $82 = ($79|0)==($81|0);
     if ($82) {
      $83 = ((($1)) + 9|0);
      $84 = HEAP8[$83>>0]|0;
      $85 = ($84<<24>>24)==(4);
      if ($85) {
       $86 = $74 | $69;
       $87 = $86 | $79;
       $88 = ($87|0)==(16777215);
       if ($88) {
        $$0 = 4507;
        return ($$0|0);
       }
      }
     }
    }
   }
   $$0 = 4500;
   return ($$0|0);
   break;
  }
  case 1:  {
   $$0 = 4506;
   return ($$0|0);
   break;
  }
  default: {
   $$0 = 4500;
   return ($$0|0);
  }
  }
  break;
 }
 case 274:  {
  $89 = ((($1)) + 24|0);
  $90 = HEAP32[$89>>2]|0;
  $91 = ($90|0)==(0);
  if (!($91)) {
   $$0 = 0;
   return ($$0|0);
  }
  $92 = ((($6)) + 9|0);
  $93 = HEAP8[$92>>0]|0;
  $94 = ($93<<24>>24)==(1);
  $BlitNto1SurfaceAlphaKey$BlitNtoNSurfaceAlphaKey = $94 ? 4502 : 4503;
  $$0 = $BlitNto1SurfaceAlphaKey$BlitNtoNSurfaceAlphaKey;
  return ($$0|0);
  break;
 }
 default: {
  $$0 = 0;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function _BlitARGBto555PixelAlpha($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $dstp$01 = 0, $dstp$1 = 0, $dstp$2 = 0, $dstp$3 = 0, $dstp$4 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $srcp$02 = 0, $srcp$1 = 0, $srcp$2 = 0, $srcp$3 = 0, $srcp$4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $5 >> 2;
 $7 = ((($info)) + 36|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $8 >> 1;
 $10 = ($3|0)==(0);
 if ($10) {
  return;
 }
 $11 = ((($info)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = HEAP32[$info>>2]|0;
 $14 = (($1) + 3)|0;
 $15 = (($14|0) / 4)&-1;
 $16 = $1 & 3;
 $$in = $3;$dstp$01 = $12;$srcp$02 = $13;
 L4: while(1) {
  $17 = (($$in) + -1)|0;
  switch ($16|0) {
  case 0:  {
   $dstp$1 = $dstp$01;$n$0 = $15;$srcp$1 = $srcp$02;
   label = 4;
   break;
  }
  case 3:  {
   $dstp$2 = $dstp$01;$n$1 = $15;$srcp$2 = $srcp$02;
   label = 8;
   break;
  }
  case 2:  {
   $dstp$3 = $dstp$01;$n$2 = $15;$srcp$3 = $srcp$02;
   label = 12;
   break;
  }
  case 1:  {
   $dstp$4 = $dstp$01;$n$3 = $15;$srcp$4 = $srcp$02;
   label = 16;
   break;
  }
  default: {
   label = 20;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    $18 = HEAP32[$srcp$1>>2]|0;
    $19 = $18 >>> 27;
    switch ($19|0) {
    case 0:  {
     break;
    }
    case 31:  {
     $20 = $18 >>> 9;
     $21 = $20 & 31744;
     $22 = $18 >>> 6;
     $23 = $22 & 992;
     $24 = $21 | $23;
     $25 = $18 >>> 3;
     $26 = $25 & 31;
     $27 = $24 | $26;
     $28 = $27&65535;
     HEAP16[$dstp$1>>1] = $28;
     break;
    }
    default: {
     $29 = HEAP16[$dstp$1>>1]|0;
     $30 = $29&65535;
     $31 = $18 << 10;
     $32 = $31 & 65011712;
     $33 = $18 >>> 9;
     $34 = $33 & 31744;
     $35 = $32 | $34;
     $36 = $18 >>> 3;
     $37 = $36 & 31;
     $38 = $35 | $37;
     $39 = $30 << 16;
     $40 = $39 | $30;
     $41 = $40 & 65043487;
     $42 = (($38) - ($41))|0;
     $43 = Math_imul($42, $19)|0;
     $44 = $43 >>> 5;
     $45 = (($44) + ($41))|0;
     $46 = $45 & 65043487;
     $47 = $46 >>> 16;
     $48 = $47 | $46;
     $49 = $48&65535;
     HEAP16[$dstp$1>>1] = $49;
    }
    }
    $50 = ((($srcp$1)) + 4|0);
    $51 = ((($dstp$1)) + 2|0);
    $dstp$2 = $51;$n$1 = $n$0;$srcp$2 = $50;
    label = 8;
    continue;
   }
   else if ((label|0) == 8) {
    label = 0;
    $52 = HEAP32[$srcp$2>>2]|0;
    $53 = $52 >>> 27;
    switch ($53|0) {
    case 0:  {
     break;
    }
    case 31:  {
     $54 = $52 >>> 9;
     $55 = $54 & 31744;
     $56 = $52 >>> 6;
     $57 = $56 & 992;
     $58 = $55 | $57;
     $59 = $52 >>> 3;
     $60 = $59 & 31;
     $61 = $58 | $60;
     $62 = $61&65535;
     HEAP16[$dstp$2>>1] = $62;
     break;
    }
    default: {
     $63 = HEAP16[$dstp$2>>1]|0;
     $64 = $63&65535;
     $65 = $52 << 10;
     $66 = $65 & 65011712;
     $67 = $52 >>> 9;
     $68 = $67 & 31744;
     $69 = $66 | $68;
     $70 = $52 >>> 3;
     $71 = $70 & 31;
     $72 = $69 | $71;
     $73 = $64 << 16;
     $74 = $73 | $64;
     $75 = $74 & 65043487;
     $76 = (($72) - ($75))|0;
     $77 = Math_imul($76, $53)|0;
     $78 = $77 >>> 5;
     $79 = (($78) + ($75))|0;
     $80 = $79 & 65043487;
     $81 = $80 >>> 16;
     $82 = $81 | $80;
     $83 = $82&65535;
     HEAP16[$dstp$2>>1] = $83;
    }
    }
    $84 = ((($srcp$2)) + 4|0);
    $85 = ((($dstp$2)) + 2|0);
    $dstp$3 = $85;$n$2 = $n$1;$srcp$3 = $84;
    label = 12;
    continue;
   }
   else if ((label|0) == 12) {
    label = 0;
    $86 = HEAP32[$srcp$3>>2]|0;
    $87 = $86 >>> 27;
    switch ($87|0) {
    case 0:  {
     break;
    }
    case 31:  {
     $88 = $86 >>> 9;
     $89 = $88 & 31744;
     $90 = $86 >>> 6;
     $91 = $90 & 992;
     $92 = $89 | $91;
     $93 = $86 >>> 3;
     $94 = $93 & 31;
     $95 = $92 | $94;
     $96 = $95&65535;
     HEAP16[$dstp$3>>1] = $96;
     break;
    }
    default: {
     $97 = HEAP16[$dstp$3>>1]|0;
     $98 = $97&65535;
     $99 = $86 << 10;
     $100 = $99 & 65011712;
     $101 = $86 >>> 9;
     $102 = $101 & 31744;
     $103 = $100 | $102;
     $104 = $86 >>> 3;
     $105 = $104 & 31;
     $106 = $103 | $105;
     $107 = $98 << 16;
     $108 = $107 | $98;
     $109 = $108 & 65043487;
     $110 = (($106) - ($109))|0;
     $111 = Math_imul($110, $87)|0;
     $112 = $111 >>> 5;
     $113 = (($112) + ($109))|0;
     $114 = $113 & 65043487;
     $115 = $114 >>> 16;
     $116 = $115 | $114;
     $117 = $116&65535;
     HEAP16[$dstp$3>>1] = $117;
    }
    }
    $118 = ((($srcp$3)) + 4|0);
    $119 = ((($dstp$3)) + 2|0);
    $dstp$4 = $119;$n$3 = $n$2;$srcp$4 = $118;
    label = 16;
    continue;
   }
   else if ((label|0) == 16) {
    label = 0;
    $120 = HEAP32[$srcp$4>>2]|0;
    $121 = $120 >>> 27;
    switch ($121|0) {
    case 0:  {
     break;
    }
    case 31:  {
     $122 = $120 >>> 9;
     $123 = $122 & 31744;
     $124 = $120 >>> 6;
     $125 = $124 & 992;
     $126 = $123 | $125;
     $127 = $120 >>> 3;
     $128 = $127 & 31;
     $129 = $126 | $128;
     $130 = $129&65535;
     HEAP16[$dstp$4>>1] = $130;
     break;
    }
    default: {
     $131 = HEAP16[$dstp$4>>1]|0;
     $132 = $131&65535;
     $133 = $120 << 10;
     $134 = $133 & 65011712;
     $135 = $120 >>> 9;
     $136 = $135 & 31744;
     $137 = $134 | $136;
     $138 = $120 >>> 3;
     $139 = $138 & 31;
     $140 = $137 | $139;
     $141 = $132 << 16;
     $142 = $141 | $132;
     $143 = $142 & 65043487;
     $144 = (($140) - ($143))|0;
     $145 = Math_imul($144, $121)|0;
     $146 = $145 >>> 5;
     $147 = (($146) + ($143))|0;
     $148 = $147 & 65043487;
     $149 = $148 >>> 16;
     $150 = $149 | $148;
     $151 = $150&65535;
     HEAP16[$dstp$4>>1] = $151;
    }
    }
    $152 = ((($srcp$4)) + 4|0);
    $153 = ((($dstp$4)) + 2|0);
    $154 = (($n$3) + -1)|0;
    $155 = ($n$3|0)>(1);
    if ($155) {
     $dstp$1 = $153;$n$0 = $154;$srcp$1 = $152;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $156 = (($152) + ($6<<2)|0);
  $157 = (($153) + ($9<<1)|0);
  $158 = ($17|0)==(0);
  if ($158) {
   label = 22;
   break;
  } else {
   $$in = $17;$dstp$01 = $157;$srcp$02 = $156;
  }
 }
 if ((label|0) == 20) {
  // unreachable;
 }
 else if ((label|0) == 22) {
  return;
 }
}
function _BlitNtoNPixelAlpha($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0;
 var $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0;
 var $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0;
 var $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0;
 var $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0;
 var $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0;
 var $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0;
 var $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0;
 var $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0;
 var $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0;
 var $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0;
 var $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0;
 var $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0;
 var $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0;
 var $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0;
 var $1267 = 0, $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0;
 var $1285 = 0, $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0;
 var $1302 = 0, $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0;
 var $1320 = 0, $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0;
 var $1339 = 0, $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0;
 var $1357 = 0, $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0;
 var $1375 = 0, $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0;
 var $1393 = 0, $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0;
 var $1410 = 0, $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0;
 var $1429 = 0, $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0;
 var $1447 = 0, $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0;
 var $1465 = 0, $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0;
 var $1483 = 0, $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0;
 var $1500 = 0, $1501 = 0, $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0, $1511 = 0, $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0, $1518 = 0;
 var $1519 = 0, $152 = 0, $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0, $153 = 0, $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0, $1536 = 0;
 var $1537 = 0, $1538 = 0, $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0, $1548 = 0, $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0, $1554 = 0;
 var $1555 = 0, $1556 = 0, $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0;
 var $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0;
 var $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0;
 var $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0;
 var $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0;
 var $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0;
 var $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0;
 var $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0;
 var $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0;
 var $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0;
 var $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0;
 var $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0;
 var $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0;
 var $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0;
 var $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0;
 var $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0;
 var $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0;
 var $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0;
 var $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0;
 var $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0;
 var $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0;
 var $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0;
 var $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0;
 var $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0;
 var $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0;
 var $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0;
 var $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0;
 var $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0;
 var $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0;
 var $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0;
 var $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0;
 var $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0;
 var $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0;
 var $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0;
 var $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0;
 var $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0;
 var $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0;
 var $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0;
 var $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0;
 var $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0;
 var $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0;
 var $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0;
 var $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0;
 var $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $dA$0 = 0, $dA$0$shrunk = 0, $dA$1 = 0, $dA$1$shrunk = 0, $dA$2 = 0, $dA$2$shrunk = 0, $dA$3 = 0, $dA$3$shrunk = 0, $dB$0 = 0, $dB$0$shrunk = 0, $dB$1 = 0, $dB$1$shrunk = 0, $dB$2 = 0, $dB$2$shrunk = 0;
 var $dB$3 = 0, $dB$3$shrunk = 0, $dG$0 = 0, $dG$0$shrunk = 0, $dG$1 = 0, $dG$1$shrunk = 0, $dG$2 = 0, $dG$2$shrunk = 0, $dG$3 = 0, $dG$3$shrunk = 0, $dR$0 = 0, $dR$0$shrunk = 0, $dR$1 = 0, $dR$1$shrunk = 0, $dR$2 = 0, $dR$2$shrunk = 0, $dR$3 = 0, $dR$3$shrunk = 0, $div = 0, $div1 = 0;
 var $div10 = 0, $div11 = 0, $div12 = 0, $div13 = 0, $div15 = 0, $div16 = 0, $div18 = 0, $div19 = 0, $div2 = 0, $div20 = 0, $div21 = 0, $div22 = 0, $div24 = 0, $div25 = 0, $div27 = 0, $div28 = 0, $div29 = 0, $div3 = 0, $div30 = 0, $div31 = 0;
 var $div33 = 0, $div34 = 0, $div4 = 0, $div6 = 0, $div7 = 0, $div9 = 0, $dst$094 = 0, $dst$1 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $sA$0 = 0, $sA$0$shrunk = 0, $sA$0$shrunk$in = 0, $sA$048 = 0, $sA$1 = 0;
 var $sA$1$shrunk = 0, $sA$1$shrunk$in = 0, $sA$162 = 0, $sA$2 = 0, $sA$2$shrunk = 0, $sA$2$shrunk$in = 0, $sA$276 = 0, $sA$3 = 0, $sA$3$shrunk = 0, $sA$3$shrunk$in = 0, $sA$390 = 0, $sB$0 = 0, $sB$0$shrunk = 0, $sB$0$shrunk$in = 0, $sB$043 = 0, $sB$047 = 0, $sB$1 = 0, $sB$1$shrunk = 0, $sB$1$shrunk$in = 0, $sB$157 = 0;
 var $sB$161 = 0, $sB$2 = 0, $sB$2$shrunk = 0, $sB$2$shrunk$in = 0, $sB$271 = 0, $sB$275 = 0, $sB$3 = 0, $sB$3$shrunk = 0, $sB$3$shrunk$in = 0, $sB$385 = 0, $sB$389 = 0, $sG$0 = 0, $sG$0$shrunk = 0, $sG$0$shrunk$in = 0, $sG$042 = 0, $sG$046 = 0, $sG$1 = 0, $sG$1$shrunk = 0, $sG$1$shrunk$in = 0, $sG$156 = 0;
 var $sG$160 = 0, $sG$2 = 0, $sG$2$shrunk = 0, $sG$2$shrunk$in = 0, $sG$270 = 0, $sG$274 = 0, $sG$3 = 0, $sG$3$shrunk = 0, $sG$3$shrunk$in = 0, $sG$384 = 0, $sG$388 = 0, $sR$0 = 0, $sR$0$shrunk = 0, $sR$0$shrunk$in = 0, $sR$041 = 0, $sR$045 = 0, $sR$1 = 0, $sR$1$shrunk = 0, $sR$1$shrunk$in = 0, $sR$155 = 0;
 var $sR$159 = 0, $sR$2 = 0, $sR$2$shrunk = 0, $sR$2$shrunk$in = 0, $sR$269 = 0, $sR$273 = 0, $sR$3 = 0, $sR$3$shrunk = 0, $sR$3$shrunk$in = 0, $sR$383 = 0, $sR$387 = 0, $src$095 = 0, $src$1 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 40|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($info)) + 44|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($9)) + 9|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = $13&255;
 $15 = ((($11)) + 9|0);
 $16 = HEAP8[$15>>0]|0;
 $17 = $16&255;
 $18 = ($3|0)==(0);
 if ($18) {
  return;
 }
 $19 = ((($info)) + 20|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = HEAP32[$info>>2]|0;
 $22 = (($1) + 3)|0;
 $23 = (($22|0) / 4)&-1;
 $24 = $1 & 3;
 $25 = ((($9)) + 12|0);
 $26 = ((($9)) + 32|0);
 $27 = ((($9)) + 28|0);
 $28 = ((($9)) + 16|0);
 $29 = ((($9)) + 20|0);
 $30 = ((($9)) + 24|0);
 $31 = ((($11)) + 28|0);
 $32 = ((($11)) + 32|0);
 $33 = ((($11)) + 33|0);
 $34 = ((($11)) + 34|0);
 $35 = ((($11)) + 12|0);
 $36 = ((($11)) + 16|0);
 $37 = ((($11)) + 20|0);
 $38 = ((($11)) + 24|0);
 $$in = $3;$dst$094 = $20;$src$095 = $21;
 L4: while(1) {
  $39 = (($$in) + -1)|0;
  switch ($24|0) {
  case 0:  {
   $dst$1 = $dst$094;$n$0 = $23;$src$1 = $src$095;
   label = 4;
   break;
  }
  case 3:  {
   $dst$2 = $dst$094;$n$1 = $23;$src$2 = $src$095;
   label = 21;
   break;
  }
  case 2:  {
   $dst$3 = $dst$094;$n$2 = $23;$src$3 = $src$095;
   label = 38;
   break;
  }
  case 1:  {
   $dst$4 = $dst$094;$n$3 = $23;$src$4 = $src$095;
   label = 55;
   break;
  }
  default: {
   label = 72;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $40 = HEAP8[$src$1>>0]|0;
     $41 = $40&255;
     $42 = HEAP32[$25>>2]|0;
     $43 = $41 & $42;
     $44 = HEAP32[$26>>2]|0;
     $45 = $44 & 255;
     $46 = $43 >>> $45;
     $47 = HEAP32[$27>>2]|0;
     $48 = $47 & 255;
     $49 = (125924 + ($48<<2)|0);
     $50 = HEAP32[$49>>2]|0;
     $51 = (($50) + ($46)|0);
     $52 = HEAP32[$28>>2]|0;
     $53 = $52 & $41;
     $54 = $44 >>> 8;
     $55 = $54 & 255;
     $56 = $53 >>> $55;
     $57 = $47 >>> 8;
     $58 = $57 & 255;
     $59 = (125924 + ($58<<2)|0);
     $60 = HEAP32[$59>>2]|0;
     $61 = (($60) + ($56)|0);
     $62 = HEAP32[$29>>2]|0;
     $63 = $62 & $41;
     $64 = $44 >>> 16;
     $65 = $64 & 255;
     $66 = $63 >>> $65;
     $67 = $47 >>> 16;
     $68 = $67 & 255;
     $69 = (125924 + ($68<<2)|0);
     $70 = HEAP32[$69>>2]|0;
     $71 = (($70) + ($66)|0);
     $72 = HEAP32[$30>>2]|0;
     $73 = $72 & $41;
     $74 = $44 >>> 24;
     $75 = $73 >>> $74;
     $76 = $47 >>> 24;
     $77 = (125924 + ($76<<2)|0);
     $78 = HEAP32[$77>>2]|0;
     $79 = (($78) + ($75)|0);
     $sA$0$shrunk$in = $79;$sB$0$shrunk$in = $71;$sG$0$shrunk$in = $61;$sR$0$shrunk$in = $51;
     label = 9;
     break;
    }
    case 2:  {
     $80 = HEAP16[$src$1>>1]|0;
     $81 = $80&65535;
     $82 = HEAP32[$25>>2]|0;
     $83 = $81 & $82;
     $84 = HEAP32[$26>>2]|0;
     $85 = $84 & 255;
     $86 = $83 >>> $85;
     $87 = HEAP32[$27>>2]|0;
     $88 = $87 & 255;
     $89 = (125924 + ($88<<2)|0);
     $90 = HEAP32[$89>>2]|0;
     $91 = (($90) + ($86)|0);
     $92 = HEAP32[$28>>2]|0;
     $93 = $92 & $81;
     $94 = $84 >>> 8;
     $95 = $94 & 255;
     $96 = $93 >>> $95;
     $97 = $87 >>> 8;
     $98 = $97 & 255;
     $99 = (125924 + ($98<<2)|0);
     $100 = HEAP32[$99>>2]|0;
     $101 = (($100) + ($96)|0);
     $102 = HEAP32[$29>>2]|0;
     $103 = $102 & $81;
     $104 = $84 >>> 16;
     $105 = $104 & 255;
     $106 = $103 >>> $105;
     $107 = $87 >>> 16;
     $108 = $107 & 255;
     $109 = (125924 + ($108<<2)|0);
     $110 = HEAP32[$109>>2]|0;
     $111 = (($110) + ($106)|0);
     $112 = HEAP32[$30>>2]|0;
     $113 = $112 & $81;
     $114 = $84 >>> 24;
     $115 = $113 >>> $114;
     $116 = $87 >>> 24;
     $117 = (125924 + ($116<<2)|0);
     $118 = HEAP32[$117>>2]|0;
     $119 = (($118) + ($115)|0);
     $sA$0$shrunk$in = $119;$sB$0$shrunk$in = $111;$sG$0$shrunk$in = $101;$sR$0$shrunk$in = $91;
     label = 9;
     break;
    }
    case 3:  {
     $120 = HEAP32[$26>>2]|0;
     $div24 = $120 >>> 3;
     $121 = $div24 & 31;
     $122 = (($src$1) + ($121)|0);
     $123 = HEAP8[$122>>0]|0;
     $124 = $120 >>> 11;
     $div25 = $124 & 31;
     $125 = (($src$1) + ($div25)|0);
     $126 = HEAP8[$125>>0]|0;
     $127 = $120 >>> 19;
     $128 = $127 & 31;
     $129 = (($src$1) + ($128)|0);
     $130 = HEAP8[$129>>0]|0;
     $sR$041 = $123&255;
     $sG$042 = $126&255;
     $sB$043 = $130&255;
     $sA$048 = 255;$sB$047 = $sB$043;$sG$046 = $sG$042;$sR$045 = $sR$041;
     label = 10;
     break;
    }
    case 4:  {
     $131 = HEAP32[$src$1>>2]|0;
     $132 = HEAP32[$25>>2]|0;
     $133 = $132 & $131;
     $134 = HEAP32[$26>>2]|0;
     $135 = $134 & 255;
     $136 = $133 >>> $135;
     $137 = HEAP32[$27>>2]|0;
     $138 = $137 & 255;
     $139 = (125924 + ($138<<2)|0);
     $140 = HEAP32[$139>>2]|0;
     $141 = (($140) + ($136)|0);
     $142 = HEAP32[$28>>2]|0;
     $143 = $142 & $131;
     $144 = $134 >>> 8;
     $145 = $144 & 255;
     $146 = $143 >>> $145;
     $147 = $137 >>> 8;
     $148 = $147 & 255;
     $149 = (125924 + ($148<<2)|0);
     $150 = HEAP32[$149>>2]|0;
     $151 = (($150) + ($146)|0);
     $152 = HEAP32[$29>>2]|0;
     $153 = $152 & $131;
     $154 = $134 >>> 16;
     $155 = $154 & 255;
     $156 = $153 >>> $155;
     $157 = $137 >>> 16;
     $158 = $157 & 255;
     $159 = (125924 + ($158<<2)|0);
     $160 = HEAP32[$159>>2]|0;
     $161 = (($160) + ($156)|0);
     $162 = HEAP32[$30>>2]|0;
     $163 = $162 & $131;
     $164 = $134 >>> 24;
     $165 = $163 >>> $164;
     $166 = $137 >>> 24;
     $167 = (125924 + ($166<<2)|0);
     $168 = HEAP32[$167>>2]|0;
     $169 = (($168) + ($165)|0);
     $sA$0$shrunk$in = $169;$sB$0$shrunk$in = $161;$sG$0$shrunk$in = $151;$sR$0$shrunk$in = $141;
     label = 9;
     break;
    }
    default: {
    }
    }
    if ((label|0) == 9) {
     label = 0;
     $sR$0$shrunk = HEAP8[$sR$0$shrunk$in>>0]|0;
     $sG$0$shrunk = HEAP8[$sG$0$shrunk$in>>0]|0;
     $sB$0$shrunk = HEAP8[$sB$0$shrunk$in>>0]|0;
     $sA$0$shrunk = HEAP8[$sA$0$shrunk$in>>0]|0;
     $sR$0 = $sR$0$shrunk&255;
     $sG$0 = $sG$0$shrunk&255;
     $sB$0 = $sB$0$shrunk&255;
     $sA$0 = $sA$0$shrunk&255;
     $170 = ($sA$0$shrunk<<24>>24)==(0);
     if (!($170)) {
      $sA$048 = $sA$0;$sB$047 = $sB$0;$sG$046 = $sG$0;$sR$045 = $sR$0;
      label = 10;
     }
    }
    L16: do {
     if ((label|0) == 10) {
      label = 0;
      switch ($17|0) {
      case 1:  {
       $171 = HEAP8[$dst$1>>0]|0;
       $172 = $171&255;
       $173 = HEAP32[$35>>2]|0;
       $174 = $172 & $173;
       $175 = HEAP32[$32>>2]|0;
       $176 = $175 & 255;
       $177 = $174 >>> $176;
       $178 = HEAP32[$31>>2]|0;
       $179 = $178 & 255;
       $180 = (125924 + ($179<<2)|0);
       $181 = HEAP32[$180>>2]|0;
       $182 = (($181) + ($177)|0);
       $183 = HEAP8[$182>>0]|0;
       $184 = HEAP32[$36>>2]|0;
       $185 = $184 & $172;
       $186 = $175 >>> 8;
       $187 = $186 & 255;
       $188 = $185 >>> $187;
       $189 = $178 >>> 8;
       $190 = $189 & 255;
       $191 = (125924 + ($190<<2)|0);
       $192 = HEAP32[$191>>2]|0;
       $193 = (($192) + ($188)|0);
       $194 = HEAP8[$193>>0]|0;
       $195 = HEAP32[$37>>2]|0;
       $196 = $195 & $172;
       $197 = $175 >>> 16;
       $198 = $197 & 255;
       $199 = $196 >>> $198;
       $200 = $178 >>> 16;
       $201 = $200 & 255;
       $202 = (125924 + ($201<<2)|0);
       $203 = HEAP32[$202>>2]|0;
       $204 = (($203) + ($199)|0);
       $205 = HEAP8[$204>>0]|0;
       $206 = HEAP32[$38>>2]|0;
       $207 = $206 & $172;
       $208 = $175 >>> 24;
       $209 = $207 >>> $208;
       $210 = $178 >>> 24;
       $211 = (125924 + ($210<<2)|0);
       $212 = HEAP32[$211>>2]|0;
       $213 = (($212) + ($209)|0);
       $214 = HEAP8[$213>>0]|0;
       $dA$0$shrunk = $214;$dB$0$shrunk = $205;$dG$0$shrunk = $194;$dR$0$shrunk = $183;
       break;
      }
      case 2:  {
       $215 = HEAP16[$dst$1>>1]|0;
       $216 = $215&65535;
       $217 = HEAP32[$35>>2]|0;
       $218 = $216 & $217;
       $219 = HEAP32[$32>>2]|0;
       $220 = $219 & 255;
       $221 = $218 >>> $220;
       $222 = HEAP32[$31>>2]|0;
       $223 = $222 & 255;
       $224 = (125924 + ($223<<2)|0);
       $225 = HEAP32[$224>>2]|0;
       $226 = (($225) + ($221)|0);
       $227 = HEAP8[$226>>0]|0;
       $228 = HEAP32[$36>>2]|0;
       $229 = $228 & $216;
       $230 = $219 >>> 8;
       $231 = $230 & 255;
       $232 = $229 >>> $231;
       $233 = $222 >>> 8;
       $234 = $233 & 255;
       $235 = (125924 + ($234<<2)|0);
       $236 = HEAP32[$235>>2]|0;
       $237 = (($236) + ($232)|0);
       $238 = HEAP8[$237>>0]|0;
       $239 = HEAP32[$37>>2]|0;
       $240 = $239 & $216;
       $241 = $219 >>> 16;
       $242 = $241 & 255;
       $243 = $240 >>> $242;
       $244 = $222 >>> 16;
       $245 = $244 & 255;
       $246 = (125924 + ($245<<2)|0);
       $247 = HEAP32[$246>>2]|0;
       $248 = (($247) + ($243)|0);
       $249 = HEAP8[$248>>0]|0;
       $250 = HEAP32[$38>>2]|0;
       $251 = $250 & $216;
       $252 = $219 >>> 24;
       $253 = $251 >>> $252;
       $254 = $222 >>> 24;
       $255 = (125924 + ($254<<2)|0);
       $256 = HEAP32[$255>>2]|0;
       $257 = (($256) + ($253)|0);
       $258 = HEAP8[$257>>0]|0;
       $dA$0$shrunk = $258;$dB$0$shrunk = $249;$dG$0$shrunk = $238;$dR$0$shrunk = $227;
       break;
      }
      case 3:  {
       $259 = HEAP32[$32>>2]|0;
       $div21 = $259 >>> 3;
       $260 = $div21 & 31;
       $261 = (($dst$1) + ($260)|0);
       $262 = HEAP8[$261>>0]|0;
       $263 = $259 >>> 11;
       $div22 = $263 & 31;
       $264 = (($dst$1) + ($div22)|0);
       $265 = HEAP8[$264>>0]|0;
       $266 = $259 >>> 19;
       $267 = $266 & 31;
       $268 = (($dst$1) + ($267)|0);
       $269 = HEAP8[$268>>0]|0;
       $dA$0$shrunk = -1;$dB$0$shrunk = $269;$dG$0$shrunk = $265;$dR$0$shrunk = $262;
       break;
      }
      case 4:  {
       $270 = HEAP32[$dst$1>>2]|0;
       $271 = HEAP32[$35>>2]|0;
       $272 = $271 & $270;
       $273 = HEAP32[$32>>2]|0;
       $274 = $273 & 255;
       $275 = $272 >>> $274;
       $276 = HEAP32[$31>>2]|0;
       $277 = $276 & 255;
       $278 = (125924 + ($277<<2)|0);
       $279 = HEAP32[$278>>2]|0;
       $280 = (($279) + ($275)|0);
       $281 = HEAP8[$280>>0]|0;
       $282 = HEAP32[$36>>2]|0;
       $283 = $282 & $270;
       $284 = $273 >>> 8;
       $285 = $284 & 255;
       $286 = $283 >>> $285;
       $287 = $276 >>> 8;
       $288 = $287 & 255;
       $289 = (125924 + ($288<<2)|0);
       $290 = HEAP32[$289>>2]|0;
       $291 = (($290) + ($286)|0);
       $292 = HEAP8[$291>>0]|0;
       $293 = HEAP32[$37>>2]|0;
       $294 = $293 & $270;
       $295 = $273 >>> 16;
       $296 = $295 & 255;
       $297 = $294 >>> $296;
       $298 = $276 >>> 16;
       $299 = $298 & 255;
       $300 = (125924 + ($299<<2)|0);
       $301 = HEAP32[$300>>2]|0;
       $302 = (($301) + ($297)|0);
       $303 = HEAP8[$302>>0]|0;
       $304 = HEAP32[$38>>2]|0;
       $305 = $304 & $270;
       $306 = $273 >>> 24;
       $307 = $305 >>> $306;
       $308 = $276 >>> 24;
       $309 = (125924 + ($308<<2)|0);
       $310 = HEAP32[$309>>2]|0;
       $311 = (($310) + ($307)|0);
       $312 = HEAP8[$311>>0]|0;
       $dA$0$shrunk = $312;$dB$0$shrunk = $303;$dG$0$shrunk = $292;$dR$0$shrunk = $281;
       break;
      }
      default: {
       $dA$0$shrunk = 0;$dB$0$shrunk = 0;$dG$0$shrunk = 0;$dR$0$shrunk = 0;
      }
      }
      $dR$0 = $dR$0$shrunk&255;
      $dG$0 = $dG$0$shrunk&255;
      $dB$0 = $dB$0$shrunk&255;
      $dA$0 = $dA$0$shrunk&255;
      $313 = (($sR$045) - ($dR$0))|0;
      $314 = Math_imul($313, $sA$048)|0;
      $315 = (($314>>>0) / 255)&-1;
      $316 = (($315) + ($dR$0))|0;
      $317 = (($sG$046) - ($dG$0))|0;
      $318 = Math_imul($317, $sA$048)|0;
      $319 = (($318>>>0) / 255)&-1;
      $320 = (($319) + ($dG$0))|0;
      $321 = (($sB$047) - ($dB$0))|0;
      $322 = Math_imul($321, $sA$048)|0;
      $323 = (($322>>>0) / 255)&-1;
      $324 = (($323) + ($dB$0))|0;
      $325 = (($dA$0) + ($sA$048))|0;
      $326 = Math_imul($dA$0, $sA$048)|0;
      $327 = (($326>>>0) / 255)&-1;
      $328 = (($325) - ($327))|0;
      switch ($17|0) {
      case 1:  {
       $329 = HEAP32[$31>>2]|0;
       $330 = $329 & 255;
       $331 = $316 >>> $330;
       $332 = HEAP32[$32>>2]|0;
       $333 = $332 & 255;
       $334 = $331 << $333;
       $335 = $329 >>> 8;
       $336 = $335 & 255;
       $337 = $320 >>> $336;
       $338 = $332 >>> 8;
       $339 = $338 & 255;
       $340 = $337 << $339;
       $341 = $340 | $334;
       $342 = $329 >>> 16;
       $343 = $342 & 255;
       $344 = $324 >>> $343;
       $345 = $332 >>> 16;
       $346 = $345 & 255;
       $347 = $344 << $346;
       $348 = $341 | $347;
       $349 = $329 >>> 24;
       $350 = $328 >>> $349;
       $351 = $332 >>> 24;
       $352 = $350 << $351;
       $353 = $348 | $352;
       $354 = $353&255;
       HEAP8[$dst$1>>0] = $354;
       break L16;
       break;
      }
      case 2:  {
       $355 = HEAP32[$31>>2]|0;
       $356 = $355 & 255;
       $357 = $316 >>> $356;
       $358 = HEAP32[$32>>2]|0;
       $359 = $358 & 255;
       $360 = $357 << $359;
       $361 = $355 >>> 8;
       $362 = $361 & 255;
       $363 = $320 >>> $362;
       $364 = $358 >>> 8;
       $365 = $364 & 255;
       $366 = $363 << $365;
       $367 = $366 | $360;
       $368 = $355 >>> 16;
       $369 = $368 & 255;
       $370 = $324 >>> $369;
       $371 = $358 >>> 16;
       $372 = $371 & 255;
       $373 = $370 << $372;
       $374 = $367 | $373;
       $375 = $355 >>> 24;
       $376 = $328 >>> $375;
       $377 = $358 >>> 24;
       $378 = $376 << $377;
       $379 = $374 | $378;
       $380 = $379&65535;
       HEAP16[$dst$1>>1] = $380;
       break L16;
       break;
      }
      case 3:  {
       $381 = $316&255;
       $382 = HEAP8[$32>>0]|0;
       $div18 = ($382&255) >>> 3;
       $383 = $div18&255;
       $384 = (($dst$1) + ($383)|0);
       HEAP8[$384>>0] = $381;
       $385 = $320&255;
       $386 = HEAP8[$33>>0]|0;
       $div19 = ($386&255) >>> 3;
       $387 = $div19&255;
       $388 = (($dst$1) + ($387)|0);
       HEAP8[$388>>0] = $385;
       $389 = $324&255;
       $390 = HEAP8[$34>>0]|0;
       $div20 = ($390&255) >>> 3;
       $391 = $div20&255;
       $392 = (($dst$1) + ($391)|0);
       HEAP8[$392>>0] = $389;
       break L16;
       break;
      }
      case 4:  {
       $393 = HEAP32[$31>>2]|0;
       $394 = $393 & 255;
       $395 = $316 >>> $394;
       $396 = HEAP32[$32>>2]|0;
       $397 = $396 & 255;
       $398 = $395 << $397;
       $399 = $393 >>> 8;
       $400 = $399 & 255;
       $401 = $320 >>> $400;
       $402 = $396 >>> 8;
       $403 = $402 & 255;
       $404 = $401 << $403;
       $405 = $404 | $398;
       $406 = $393 >>> 16;
       $407 = $406 & 255;
       $408 = $324 >>> $407;
       $409 = $396 >>> 16;
       $410 = $409 & 255;
       $411 = $408 << $410;
       $412 = $405 | $411;
       $413 = $393 >>> 24;
       $414 = $328 >>> $413;
       $415 = $396 >>> 24;
       $416 = $414 << $415;
       $417 = $412 | $416;
       HEAP32[$dst$1>>2] = $417;
       break L16;
       break;
      }
      default: {
       break L16;
      }
      }
     }
    } while(0);
    $418 = (($src$1) + ($14)|0);
    $419 = (($dst$1) + ($17)|0);
    $dst$2 = $419;$n$1 = $n$0;$src$2 = $418;
    label = 21;
    continue;
   }
   else if ((label|0) == 21) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $420 = HEAP8[$src$2>>0]|0;
     $421 = $420&255;
     $422 = HEAP32[$25>>2]|0;
     $423 = $421 & $422;
     $424 = HEAP32[$26>>2]|0;
     $425 = $424 & 255;
     $426 = $423 >>> $425;
     $427 = HEAP32[$27>>2]|0;
     $428 = $427 & 255;
     $429 = (125924 + ($428<<2)|0);
     $430 = HEAP32[$429>>2]|0;
     $431 = (($430) + ($426)|0);
     $432 = HEAP32[$28>>2]|0;
     $433 = $432 & $421;
     $434 = $424 >>> 8;
     $435 = $434 & 255;
     $436 = $433 >>> $435;
     $437 = $427 >>> 8;
     $438 = $437 & 255;
     $439 = (125924 + ($438<<2)|0);
     $440 = HEAP32[$439>>2]|0;
     $441 = (($440) + ($436)|0);
     $442 = HEAP32[$29>>2]|0;
     $443 = $442 & $421;
     $444 = $424 >>> 16;
     $445 = $444 & 255;
     $446 = $443 >>> $445;
     $447 = $427 >>> 16;
     $448 = $447 & 255;
     $449 = (125924 + ($448<<2)|0);
     $450 = HEAP32[$449>>2]|0;
     $451 = (($450) + ($446)|0);
     $452 = HEAP32[$30>>2]|0;
     $453 = $452 & $421;
     $454 = $424 >>> 24;
     $455 = $453 >>> $454;
     $456 = $427 >>> 24;
     $457 = (125924 + ($456<<2)|0);
     $458 = HEAP32[$457>>2]|0;
     $459 = (($458) + ($455)|0);
     $sA$1$shrunk$in = $459;$sB$1$shrunk$in = $451;$sG$1$shrunk$in = $441;$sR$1$shrunk$in = $431;
     label = 26;
     break;
    }
    case 2:  {
     $460 = HEAP16[$src$2>>1]|0;
     $461 = $460&65535;
     $462 = HEAP32[$25>>2]|0;
     $463 = $461 & $462;
     $464 = HEAP32[$26>>2]|0;
     $465 = $464 & 255;
     $466 = $463 >>> $465;
     $467 = HEAP32[$27>>2]|0;
     $468 = $467 & 255;
     $469 = (125924 + ($468<<2)|0);
     $470 = HEAP32[$469>>2]|0;
     $471 = (($470) + ($466)|0);
     $472 = HEAP32[$28>>2]|0;
     $473 = $472 & $461;
     $474 = $464 >>> 8;
     $475 = $474 & 255;
     $476 = $473 >>> $475;
     $477 = $467 >>> 8;
     $478 = $477 & 255;
     $479 = (125924 + ($478<<2)|0);
     $480 = HEAP32[$479>>2]|0;
     $481 = (($480) + ($476)|0);
     $482 = HEAP32[$29>>2]|0;
     $483 = $482 & $461;
     $484 = $464 >>> 16;
     $485 = $484 & 255;
     $486 = $483 >>> $485;
     $487 = $467 >>> 16;
     $488 = $487 & 255;
     $489 = (125924 + ($488<<2)|0);
     $490 = HEAP32[$489>>2]|0;
     $491 = (($490) + ($486)|0);
     $492 = HEAP32[$30>>2]|0;
     $493 = $492 & $461;
     $494 = $464 >>> 24;
     $495 = $493 >>> $494;
     $496 = $467 >>> 24;
     $497 = (125924 + ($496<<2)|0);
     $498 = HEAP32[$497>>2]|0;
     $499 = (($498) + ($495)|0);
     $sA$1$shrunk$in = $499;$sB$1$shrunk$in = $491;$sG$1$shrunk$in = $481;$sR$1$shrunk$in = $471;
     label = 26;
     break;
    }
    case 3:  {
     $500 = HEAP32[$26>>2]|0;
     $div15 = $500 >>> 3;
     $501 = $div15 & 31;
     $502 = (($src$2) + ($501)|0);
     $503 = HEAP8[$502>>0]|0;
     $504 = $500 >>> 11;
     $div16 = $504 & 31;
     $505 = (($src$2) + ($div16)|0);
     $506 = HEAP8[$505>>0]|0;
     $507 = $500 >>> 19;
     $508 = $507 & 31;
     $509 = (($src$2) + ($508)|0);
     $510 = HEAP8[$509>>0]|0;
     $sR$155 = $503&255;
     $sG$156 = $506&255;
     $sB$157 = $510&255;
     $sA$162 = 255;$sB$161 = $sB$157;$sG$160 = $sG$156;$sR$159 = $sR$155;
     label = 27;
     break;
    }
    case 4:  {
     $511 = HEAP32[$src$2>>2]|0;
     $512 = HEAP32[$25>>2]|0;
     $513 = $512 & $511;
     $514 = HEAP32[$26>>2]|0;
     $515 = $514 & 255;
     $516 = $513 >>> $515;
     $517 = HEAP32[$27>>2]|0;
     $518 = $517 & 255;
     $519 = (125924 + ($518<<2)|0);
     $520 = HEAP32[$519>>2]|0;
     $521 = (($520) + ($516)|0);
     $522 = HEAP32[$28>>2]|0;
     $523 = $522 & $511;
     $524 = $514 >>> 8;
     $525 = $524 & 255;
     $526 = $523 >>> $525;
     $527 = $517 >>> 8;
     $528 = $527 & 255;
     $529 = (125924 + ($528<<2)|0);
     $530 = HEAP32[$529>>2]|0;
     $531 = (($530) + ($526)|0);
     $532 = HEAP32[$29>>2]|0;
     $533 = $532 & $511;
     $534 = $514 >>> 16;
     $535 = $534 & 255;
     $536 = $533 >>> $535;
     $537 = $517 >>> 16;
     $538 = $537 & 255;
     $539 = (125924 + ($538<<2)|0);
     $540 = HEAP32[$539>>2]|0;
     $541 = (($540) + ($536)|0);
     $542 = HEAP32[$30>>2]|0;
     $543 = $542 & $511;
     $544 = $514 >>> 24;
     $545 = $543 >>> $544;
     $546 = $517 >>> 24;
     $547 = (125924 + ($546<<2)|0);
     $548 = HEAP32[$547>>2]|0;
     $549 = (($548) + ($545)|0);
     $sA$1$shrunk$in = $549;$sB$1$shrunk$in = $541;$sG$1$shrunk$in = $531;$sR$1$shrunk$in = $521;
     label = 26;
     break;
    }
    default: {
    }
    }
    if ((label|0) == 26) {
     label = 0;
     $sR$1$shrunk = HEAP8[$sR$1$shrunk$in>>0]|0;
     $sG$1$shrunk = HEAP8[$sG$1$shrunk$in>>0]|0;
     $sB$1$shrunk = HEAP8[$sB$1$shrunk$in>>0]|0;
     $sA$1$shrunk = HEAP8[$sA$1$shrunk$in>>0]|0;
     $sR$1 = $sR$1$shrunk&255;
     $sG$1 = $sG$1$shrunk&255;
     $sB$1 = $sB$1$shrunk&255;
     $sA$1 = $sA$1$shrunk&255;
     $550 = ($sA$1$shrunk<<24>>24)==(0);
     if (!($550)) {
      $sA$162 = $sA$1;$sB$161 = $sB$1;$sG$160 = $sG$1;$sR$159 = $sR$1;
      label = 27;
     }
    }
    L38: do {
     if ((label|0) == 27) {
      label = 0;
      switch ($17|0) {
      case 1:  {
       $551 = HEAP8[$dst$2>>0]|0;
       $552 = $551&255;
       $553 = HEAP32[$35>>2]|0;
       $554 = $552 & $553;
       $555 = HEAP32[$32>>2]|0;
       $556 = $555 & 255;
       $557 = $554 >>> $556;
       $558 = HEAP32[$31>>2]|0;
       $559 = $558 & 255;
       $560 = (125924 + ($559<<2)|0);
       $561 = HEAP32[$560>>2]|0;
       $562 = (($561) + ($557)|0);
       $563 = HEAP8[$562>>0]|0;
       $564 = HEAP32[$36>>2]|0;
       $565 = $564 & $552;
       $566 = $555 >>> 8;
       $567 = $566 & 255;
       $568 = $565 >>> $567;
       $569 = $558 >>> 8;
       $570 = $569 & 255;
       $571 = (125924 + ($570<<2)|0);
       $572 = HEAP32[$571>>2]|0;
       $573 = (($572) + ($568)|0);
       $574 = HEAP8[$573>>0]|0;
       $575 = HEAP32[$37>>2]|0;
       $576 = $575 & $552;
       $577 = $555 >>> 16;
       $578 = $577 & 255;
       $579 = $576 >>> $578;
       $580 = $558 >>> 16;
       $581 = $580 & 255;
       $582 = (125924 + ($581<<2)|0);
       $583 = HEAP32[$582>>2]|0;
       $584 = (($583) + ($579)|0);
       $585 = HEAP8[$584>>0]|0;
       $586 = HEAP32[$38>>2]|0;
       $587 = $586 & $552;
       $588 = $555 >>> 24;
       $589 = $587 >>> $588;
       $590 = $558 >>> 24;
       $591 = (125924 + ($590<<2)|0);
       $592 = HEAP32[$591>>2]|0;
       $593 = (($592) + ($589)|0);
       $594 = HEAP8[$593>>0]|0;
       $dA$1$shrunk = $594;$dB$1$shrunk = $585;$dG$1$shrunk = $574;$dR$1$shrunk = $563;
       break;
      }
      case 2:  {
       $595 = HEAP16[$dst$2>>1]|0;
       $596 = $595&65535;
       $597 = HEAP32[$35>>2]|0;
       $598 = $596 & $597;
       $599 = HEAP32[$32>>2]|0;
       $600 = $599 & 255;
       $601 = $598 >>> $600;
       $602 = HEAP32[$31>>2]|0;
       $603 = $602 & 255;
       $604 = (125924 + ($603<<2)|0);
       $605 = HEAP32[$604>>2]|0;
       $606 = (($605) + ($601)|0);
       $607 = HEAP8[$606>>0]|0;
       $608 = HEAP32[$36>>2]|0;
       $609 = $608 & $596;
       $610 = $599 >>> 8;
       $611 = $610 & 255;
       $612 = $609 >>> $611;
       $613 = $602 >>> 8;
       $614 = $613 & 255;
       $615 = (125924 + ($614<<2)|0);
       $616 = HEAP32[$615>>2]|0;
       $617 = (($616) + ($612)|0);
       $618 = HEAP8[$617>>0]|0;
       $619 = HEAP32[$37>>2]|0;
       $620 = $619 & $596;
       $621 = $599 >>> 16;
       $622 = $621 & 255;
       $623 = $620 >>> $622;
       $624 = $602 >>> 16;
       $625 = $624 & 255;
       $626 = (125924 + ($625<<2)|0);
       $627 = HEAP32[$626>>2]|0;
       $628 = (($627) + ($623)|0);
       $629 = HEAP8[$628>>0]|0;
       $630 = HEAP32[$38>>2]|0;
       $631 = $630 & $596;
       $632 = $599 >>> 24;
       $633 = $631 >>> $632;
       $634 = $602 >>> 24;
       $635 = (125924 + ($634<<2)|0);
       $636 = HEAP32[$635>>2]|0;
       $637 = (($636) + ($633)|0);
       $638 = HEAP8[$637>>0]|0;
       $dA$1$shrunk = $638;$dB$1$shrunk = $629;$dG$1$shrunk = $618;$dR$1$shrunk = $607;
       break;
      }
      case 3:  {
       $639 = HEAP32[$32>>2]|0;
       $div12 = $639 >>> 3;
       $640 = $div12 & 31;
       $641 = (($dst$2) + ($640)|0);
       $642 = HEAP8[$641>>0]|0;
       $643 = $639 >>> 11;
       $div13 = $643 & 31;
       $644 = (($dst$2) + ($div13)|0);
       $645 = HEAP8[$644>>0]|0;
       $646 = $639 >>> 19;
       $647 = $646 & 31;
       $648 = (($dst$2) + ($647)|0);
       $649 = HEAP8[$648>>0]|0;
       $dA$1$shrunk = -1;$dB$1$shrunk = $649;$dG$1$shrunk = $645;$dR$1$shrunk = $642;
       break;
      }
      case 4:  {
       $650 = HEAP32[$dst$2>>2]|0;
       $651 = HEAP32[$35>>2]|0;
       $652 = $651 & $650;
       $653 = HEAP32[$32>>2]|0;
       $654 = $653 & 255;
       $655 = $652 >>> $654;
       $656 = HEAP32[$31>>2]|0;
       $657 = $656 & 255;
       $658 = (125924 + ($657<<2)|0);
       $659 = HEAP32[$658>>2]|0;
       $660 = (($659) + ($655)|0);
       $661 = HEAP8[$660>>0]|0;
       $662 = HEAP32[$36>>2]|0;
       $663 = $662 & $650;
       $664 = $653 >>> 8;
       $665 = $664 & 255;
       $666 = $663 >>> $665;
       $667 = $656 >>> 8;
       $668 = $667 & 255;
       $669 = (125924 + ($668<<2)|0);
       $670 = HEAP32[$669>>2]|0;
       $671 = (($670) + ($666)|0);
       $672 = HEAP8[$671>>0]|0;
       $673 = HEAP32[$37>>2]|0;
       $674 = $673 & $650;
       $675 = $653 >>> 16;
       $676 = $675 & 255;
       $677 = $674 >>> $676;
       $678 = $656 >>> 16;
       $679 = $678 & 255;
       $680 = (125924 + ($679<<2)|0);
       $681 = HEAP32[$680>>2]|0;
       $682 = (($681) + ($677)|0);
       $683 = HEAP8[$682>>0]|0;
       $684 = HEAP32[$38>>2]|0;
       $685 = $684 & $650;
       $686 = $653 >>> 24;
       $687 = $685 >>> $686;
       $688 = $656 >>> 24;
       $689 = (125924 + ($688<<2)|0);
       $690 = HEAP32[$689>>2]|0;
       $691 = (($690) + ($687)|0);
       $692 = HEAP8[$691>>0]|0;
       $dA$1$shrunk = $692;$dB$1$shrunk = $683;$dG$1$shrunk = $672;$dR$1$shrunk = $661;
       break;
      }
      default: {
       $dA$1$shrunk = 0;$dB$1$shrunk = 0;$dG$1$shrunk = 0;$dR$1$shrunk = 0;
      }
      }
      $dR$1 = $dR$1$shrunk&255;
      $dG$1 = $dG$1$shrunk&255;
      $dB$1 = $dB$1$shrunk&255;
      $dA$1 = $dA$1$shrunk&255;
      $693 = (($sR$159) - ($dR$1))|0;
      $694 = Math_imul($693, $sA$162)|0;
      $695 = (($694>>>0) / 255)&-1;
      $696 = (($695) + ($dR$1))|0;
      $697 = (($sG$160) - ($dG$1))|0;
      $698 = Math_imul($697, $sA$162)|0;
      $699 = (($698>>>0) / 255)&-1;
      $700 = (($699) + ($dG$1))|0;
      $701 = (($sB$161) - ($dB$1))|0;
      $702 = Math_imul($701, $sA$162)|0;
      $703 = (($702>>>0) / 255)&-1;
      $704 = (($703) + ($dB$1))|0;
      $705 = (($dA$1) + ($sA$162))|0;
      $706 = Math_imul($dA$1, $sA$162)|0;
      $707 = (($706>>>0) / 255)&-1;
      $708 = (($705) - ($707))|0;
      switch ($17|0) {
      case 1:  {
       $709 = HEAP32[$31>>2]|0;
       $710 = $709 & 255;
       $711 = $696 >>> $710;
       $712 = HEAP32[$32>>2]|0;
       $713 = $712 & 255;
       $714 = $711 << $713;
       $715 = $709 >>> 8;
       $716 = $715 & 255;
       $717 = $700 >>> $716;
       $718 = $712 >>> 8;
       $719 = $718 & 255;
       $720 = $717 << $719;
       $721 = $720 | $714;
       $722 = $709 >>> 16;
       $723 = $722 & 255;
       $724 = $704 >>> $723;
       $725 = $712 >>> 16;
       $726 = $725 & 255;
       $727 = $724 << $726;
       $728 = $721 | $727;
       $729 = $709 >>> 24;
       $730 = $708 >>> $729;
       $731 = $712 >>> 24;
       $732 = $730 << $731;
       $733 = $728 | $732;
       $734 = $733&255;
       HEAP8[$dst$2>>0] = $734;
       break L38;
       break;
      }
      case 2:  {
       $735 = HEAP32[$31>>2]|0;
       $736 = $735 & 255;
       $737 = $696 >>> $736;
       $738 = HEAP32[$32>>2]|0;
       $739 = $738 & 255;
       $740 = $737 << $739;
       $741 = $735 >>> 8;
       $742 = $741 & 255;
       $743 = $700 >>> $742;
       $744 = $738 >>> 8;
       $745 = $744 & 255;
       $746 = $743 << $745;
       $747 = $746 | $740;
       $748 = $735 >>> 16;
       $749 = $748 & 255;
       $750 = $704 >>> $749;
       $751 = $738 >>> 16;
       $752 = $751 & 255;
       $753 = $750 << $752;
       $754 = $747 | $753;
       $755 = $735 >>> 24;
       $756 = $708 >>> $755;
       $757 = $738 >>> 24;
       $758 = $756 << $757;
       $759 = $754 | $758;
       $760 = $759&65535;
       HEAP16[$dst$2>>1] = $760;
       break L38;
       break;
      }
      case 3:  {
       $761 = $696&255;
       $762 = HEAP8[$32>>0]|0;
       $div9 = ($762&255) >>> 3;
       $763 = $div9&255;
       $764 = (($dst$2) + ($763)|0);
       HEAP8[$764>>0] = $761;
       $765 = $700&255;
       $766 = HEAP8[$33>>0]|0;
       $div10 = ($766&255) >>> 3;
       $767 = $div10&255;
       $768 = (($dst$2) + ($767)|0);
       HEAP8[$768>>0] = $765;
       $769 = $704&255;
       $770 = HEAP8[$34>>0]|0;
       $div11 = ($770&255) >>> 3;
       $771 = $div11&255;
       $772 = (($dst$2) + ($771)|0);
       HEAP8[$772>>0] = $769;
       break L38;
       break;
      }
      case 4:  {
       $773 = HEAP32[$31>>2]|0;
       $774 = $773 & 255;
       $775 = $696 >>> $774;
       $776 = HEAP32[$32>>2]|0;
       $777 = $776 & 255;
       $778 = $775 << $777;
       $779 = $773 >>> 8;
       $780 = $779 & 255;
       $781 = $700 >>> $780;
       $782 = $776 >>> 8;
       $783 = $782 & 255;
       $784 = $781 << $783;
       $785 = $784 | $778;
       $786 = $773 >>> 16;
       $787 = $786 & 255;
       $788 = $704 >>> $787;
       $789 = $776 >>> 16;
       $790 = $789 & 255;
       $791 = $788 << $790;
       $792 = $785 | $791;
       $793 = $773 >>> 24;
       $794 = $708 >>> $793;
       $795 = $776 >>> 24;
       $796 = $794 << $795;
       $797 = $792 | $796;
       HEAP32[$dst$2>>2] = $797;
       break L38;
       break;
      }
      default: {
       break L38;
      }
      }
     }
    } while(0);
    $798 = (($src$2) + ($14)|0);
    $799 = (($dst$2) + ($17)|0);
    $dst$3 = $799;$n$2 = $n$1;$src$3 = $798;
    label = 38;
    continue;
   }
   else if ((label|0) == 38) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $800 = HEAP8[$src$3>>0]|0;
     $801 = $800&255;
     $802 = HEAP32[$25>>2]|0;
     $803 = $801 & $802;
     $804 = HEAP32[$26>>2]|0;
     $805 = $804 & 255;
     $806 = $803 >>> $805;
     $807 = HEAP32[$27>>2]|0;
     $808 = $807 & 255;
     $809 = (125924 + ($808<<2)|0);
     $810 = HEAP32[$809>>2]|0;
     $811 = (($810) + ($806)|0);
     $812 = HEAP32[$28>>2]|0;
     $813 = $812 & $801;
     $814 = $804 >>> 8;
     $815 = $814 & 255;
     $816 = $813 >>> $815;
     $817 = $807 >>> 8;
     $818 = $817 & 255;
     $819 = (125924 + ($818<<2)|0);
     $820 = HEAP32[$819>>2]|0;
     $821 = (($820) + ($816)|0);
     $822 = HEAP32[$29>>2]|0;
     $823 = $822 & $801;
     $824 = $804 >>> 16;
     $825 = $824 & 255;
     $826 = $823 >>> $825;
     $827 = $807 >>> 16;
     $828 = $827 & 255;
     $829 = (125924 + ($828<<2)|0);
     $830 = HEAP32[$829>>2]|0;
     $831 = (($830) + ($826)|0);
     $832 = HEAP32[$30>>2]|0;
     $833 = $832 & $801;
     $834 = $804 >>> 24;
     $835 = $833 >>> $834;
     $836 = $807 >>> 24;
     $837 = (125924 + ($836<<2)|0);
     $838 = HEAP32[$837>>2]|0;
     $839 = (($838) + ($835)|0);
     $sA$2$shrunk$in = $839;$sB$2$shrunk$in = $831;$sG$2$shrunk$in = $821;$sR$2$shrunk$in = $811;
     label = 43;
     break;
    }
    case 2:  {
     $840 = HEAP16[$src$3>>1]|0;
     $841 = $840&65535;
     $842 = HEAP32[$25>>2]|0;
     $843 = $841 & $842;
     $844 = HEAP32[$26>>2]|0;
     $845 = $844 & 255;
     $846 = $843 >>> $845;
     $847 = HEAP32[$27>>2]|0;
     $848 = $847 & 255;
     $849 = (125924 + ($848<<2)|0);
     $850 = HEAP32[$849>>2]|0;
     $851 = (($850) + ($846)|0);
     $852 = HEAP32[$28>>2]|0;
     $853 = $852 & $841;
     $854 = $844 >>> 8;
     $855 = $854 & 255;
     $856 = $853 >>> $855;
     $857 = $847 >>> 8;
     $858 = $857 & 255;
     $859 = (125924 + ($858<<2)|0);
     $860 = HEAP32[$859>>2]|0;
     $861 = (($860) + ($856)|0);
     $862 = HEAP32[$29>>2]|0;
     $863 = $862 & $841;
     $864 = $844 >>> 16;
     $865 = $864 & 255;
     $866 = $863 >>> $865;
     $867 = $847 >>> 16;
     $868 = $867 & 255;
     $869 = (125924 + ($868<<2)|0);
     $870 = HEAP32[$869>>2]|0;
     $871 = (($870) + ($866)|0);
     $872 = HEAP32[$30>>2]|0;
     $873 = $872 & $841;
     $874 = $844 >>> 24;
     $875 = $873 >>> $874;
     $876 = $847 >>> 24;
     $877 = (125924 + ($876<<2)|0);
     $878 = HEAP32[$877>>2]|0;
     $879 = (($878) + ($875)|0);
     $sA$2$shrunk$in = $879;$sB$2$shrunk$in = $871;$sG$2$shrunk$in = $861;$sR$2$shrunk$in = $851;
     label = 43;
     break;
    }
    case 3:  {
     $880 = HEAP32[$26>>2]|0;
     $div6 = $880 >>> 3;
     $881 = $div6 & 31;
     $882 = (($src$3) + ($881)|0);
     $883 = HEAP8[$882>>0]|0;
     $884 = $880 >>> 11;
     $div7 = $884 & 31;
     $885 = (($src$3) + ($div7)|0);
     $886 = HEAP8[$885>>0]|0;
     $887 = $880 >>> 19;
     $888 = $887 & 31;
     $889 = (($src$3) + ($888)|0);
     $890 = HEAP8[$889>>0]|0;
     $sR$269 = $883&255;
     $sG$270 = $886&255;
     $sB$271 = $890&255;
     $sA$276 = 255;$sB$275 = $sB$271;$sG$274 = $sG$270;$sR$273 = $sR$269;
     label = 44;
     break;
    }
    case 4:  {
     $891 = HEAP32[$src$3>>2]|0;
     $892 = HEAP32[$25>>2]|0;
     $893 = $892 & $891;
     $894 = HEAP32[$26>>2]|0;
     $895 = $894 & 255;
     $896 = $893 >>> $895;
     $897 = HEAP32[$27>>2]|0;
     $898 = $897 & 255;
     $899 = (125924 + ($898<<2)|0);
     $900 = HEAP32[$899>>2]|0;
     $901 = (($900) + ($896)|0);
     $902 = HEAP32[$28>>2]|0;
     $903 = $902 & $891;
     $904 = $894 >>> 8;
     $905 = $904 & 255;
     $906 = $903 >>> $905;
     $907 = $897 >>> 8;
     $908 = $907 & 255;
     $909 = (125924 + ($908<<2)|0);
     $910 = HEAP32[$909>>2]|0;
     $911 = (($910) + ($906)|0);
     $912 = HEAP32[$29>>2]|0;
     $913 = $912 & $891;
     $914 = $894 >>> 16;
     $915 = $914 & 255;
     $916 = $913 >>> $915;
     $917 = $897 >>> 16;
     $918 = $917 & 255;
     $919 = (125924 + ($918<<2)|0);
     $920 = HEAP32[$919>>2]|0;
     $921 = (($920) + ($916)|0);
     $922 = HEAP32[$30>>2]|0;
     $923 = $922 & $891;
     $924 = $894 >>> 24;
     $925 = $923 >>> $924;
     $926 = $897 >>> 24;
     $927 = (125924 + ($926<<2)|0);
     $928 = HEAP32[$927>>2]|0;
     $929 = (($928) + ($925)|0);
     $sA$2$shrunk$in = $929;$sB$2$shrunk$in = $921;$sG$2$shrunk$in = $911;$sR$2$shrunk$in = $901;
     label = 43;
     break;
    }
    default: {
    }
    }
    if ((label|0) == 43) {
     label = 0;
     $sR$2$shrunk = HEAP8[$sR$2$shrunk$in>>0]|0;
     $sG$2$shrunk = HEAP8[$sG$2$shrunk$in>>0]|0;
     $sB$2$shrunk = HEAP8[$sB$2$shrunk$in>>0]|0;
     $sA$2$shrunk = HEAP8[$sA$2$shrunk$in>>0]|0;
     $sR$2 = $sR$2$shrunk&255;
     $sG$2 = $sG$2$shrunk&255;
     $sB$2 = $sB$2$shrunk&255;
     $sA$2 = $sA$2$shrunk&255;
     $930 = ($sA$2$shrunk<<24>>24)==(0);
     if (!($930)) {
      $sA$276 = $sA$2;$sB$275 = $sB$2;$sG$274 = $sG$2;$sR$273 = $sR$2;
      label = 44;
     }
    }
    L60: do {
     if ((label|0) == 44) {
      label = 0;
      switch ($17|0) {
      case 1:  {
       $931 = HEAP8[$dst$3>>0]|0;
       $932 = $931&255;
       $933 = HEAP32[$35>>2]|0;
       $934 = $932 & $933;
       $935 = HEAP32[$32>>2]|0;
       $936 = $935 & 255;
       $937 = $934 >>> $936;
       $938 = HEAP32[$31>>2]|0;
       $939 = $938 & 255;
       $940 = (125924 + ($939<<2)|0);
       $941 = HEAP32[$940>>2]|0;
       $942 = (($941) + ($937)|0);
       $943 = HEAP8[$942>>0]|0;
       $944 = HEAP32[$36>>2]|0;
       $945 = $944 & $932;
       $946 = $935 >>> 8;
       $947 = $946 & 255;
       $948 = $945 >>> $947;
       $949 = $938 >>> 8;
       $950 = $949 & 255;
       $951 = (125924 + ($950<<2)|0);
       $952 = HEAP32[$951>>2]|0;
       $953 = (($952) + ($948)|0);
       $954 = HEAP8[$953>>0]|0;
       $955 = HEAP32[$37>>2]|0;
       $956 = $955 & $932;
       $957 = $935 >>> 16;
       $958 = $957 & 255;
       $959 = $956 >>> $958;
       $960 = $938 >>> 16;
       $961 = $960 & 255;
       $962 = (125924 + ($961<<2)|0);
       $963 = HEAP32[$962>>2]|0;
       $964 = (($963) + ($959)|0);
       $965 = HEAP8[$964>>0]|0;
       $966 = HEAP32[$38>>2]|0;
       $967 = $966 & $932;
       $968 = $935 >>> 24;
       $969 = $967 >>> $968;
       $970 = $938 >>> 24;
       $971 = (125924 + ($970<<2)|0);
       $972 = HEAP32[$971>>2]|0;
       $973 = (($972) + ($969)|0);
       $974 = HEAP8[$973>>0]|0;
       $dA$2$shrunk = $974;$dB$2$shrunk = $965;$dG$2$shrunk = $954;$dR$2$shrunk = $943;
       break;
      }
      case 2:  {
       $975 = HEAP16[$dst$3>>1]|0;
       $976 = $975&65535;
       $977 = HEAP32[$35>>2]|0;
       $978 = $976 & $977;
       $979 = HEAP32[$32>>2]|0;
       $980 = $979 & 255;
       $981 = $978 >>> $980;
       $982 = HEAP32[$31>>2]|0;
       $983 = $982 & 255;
       $984 = (125924 + ($983<<2)|0);
       $985 = HEAP32[$984>>2]|0;
       $986 = (($985) + ($981)|0);
       $987 = HEAP8[$986>>0]|0;
       $988 = HEAP32[$36>>2]|0;
       $989 = $988 & $976;
       $990 = $979 >>> 8;
       $991 = $990 & 255;
       $992 = $989 >>> $991;
       $993 = $982 >>> 8;
       $994 = $993 & 255;
       $995 = (125924 + ($994<<2)|0);
       $996 = HEAP32[$995>>2]|0;
       $997 = (($996) + ($992)|0);
       $998 = HEAP8[$997>>0]|0;
       $999 = HEAP32[$37>>2]|0;
       $1000 = $999 & $976;
       $1001 = $979 >>> 16;
       $1002 = $1001 & 255;
       $1003 = $1000 >>> $1002;
       $1004 = $982 >>> 16;
       $1005 = $1004 & 255;
       $1006 = (125924 + ($1005<<2)|0);
       $1007 = HEAP32[$1006>>2]|0;
       $1008 = (($1007) + ($1003)|0);
       $1009 = HEAP8[$1008>>0]|0;
       $1010 = HEAP32[$38>>2]|0;
       $1011 = $1010 & $976;
       $1012 = $979 >>> 24;
       $1013 = $1011 >>> $1012;
       $1014 = $982 >>> 24;
       $1015 = (125924 + ($1014<<2)|0);
       $1016 = HEAP32[$1015>>2]|0;
       $1017 = (($1016) + ($1013)|0);
       $1018 = HEAP8[$1017>>0]|0;
       $dA$2$shrunk = $1018;$dB$2$shrunk = $1009;$dG$2$shrunk = $998;$dR$2$shrunk = $987;
       break;
      }
      case 3:  {
       $1019 = HEAP32[$32>>2]|0;
       $div3 = $1019 >>> 3;
       $1020 = $div3 & 31;
       $1021 = (($dst$3) + ($1020)|0);
       $1022 = HEAP8[$1021>>0]|0;
       $1023 = $1019 >>> 11;
       $div4 = $1023 & 31;
       $1024 = (($dst$3) + ($div4)|0);
       $1025 = HEAP8[$1024>>0]|0;
       $1026 = $1019 >>> 19;
       $1027 = $1026 & 31;
       $1028 = (($dst$3) + ($1027)|0);
       $1029 = HEAP8[$1028>>0]|0;
       $dA$2$shrunk = -1;$dB$2$shrunk = $1029;$dG$2$shrunk = $1025;$dR$2$shrunk = $1022;
       break;
      }
      case 4:  {
       $1030 = HEAP32[$dst$3>>2]|0;
       $1031 = HEAP32[$35>>2]|0;
       $1032 = $1031 & $1030;
       $1033 = HEAP32[$32>>2]|0;
       $1034 = $1033 & 255;
       $1035 = $1032 >>> $1034;
       $1036 = HEAP32[$31>>2]|0;
       $1037 = $1036 & 255;
       $1038 = (125924 + ($1037<<2)|0);
       $1039 = HEAP32[$1038>>2]|0;
       $1040 = (($1039) + ($1035)|0);
       $1041 = HEAP8[$1040>>0]|0;
       $1042 = HEAP32[$36>>2]|0;
       $1043 = $1042 & $1030;
       $1044 = $1033 >>> 8;
       $1045 = $1044 & 255;
       $1046 = $1043 >>> $1045;
       $1047 = $1036 >>> 8;
       $1048 = $1047 & 255;
       $1049 = (125924 + ($1048<<2)|0);
       $1050 = HEAP32[$1049>>2]|0;
       $1051 = (($1050) + ($1046)|0);
       $1052 = HEAP8[$1051>>0]|0;
       $1053 = HEAP32[$37>>2]|0;
       $1054 = $1053 & $1030;
       $1055 = $1033 >>> 16;
       $1056 = $1055 & 255;
       $1057 = $1054 >>> $1056;
       $1058 = $1036 >>> 16;
       $1059 = $1058 & 255;
       $1060 = (125924 + ($1059<<2)|0);
       $1061 = HEAP32[$1060>>2]|0;
       $1062 = (($1061) + ($1057)|0);
       $1063 = HEAP8[$1062>>0]|0;
       $1064 = HEAP32[$38>>2]|0;
       $1065 = $1064 & $1030;
       $1066 = $1033 >>> 24;
       $1067 = $1065 >>> $1066;
       $1068 = $1036 >>> 24;
       $1069 = (125924 + ($1068<<2)|0);
       $1070 = HEAP32[$1069>>2]|0;
       $1071 = (($1070) + ($1067)|0);
       $1072 = HEAP8[$1071>>0]|0;
       $dA$2$shrunk = $1072;$dB$2$shrunk = $1063;$dG$2$shrunk = $1052;$dR$2$shrunk = $1041;
       break;
      }
      default: {
       $dA$2$shrunk = 0;$dB$2$shrunk = 0;$dG$2$shrunk = 0;$dR$2$shrunk = 0;
      }
      }
      $dR$2 = $dR$2$shrunk&255;
      $dG$2 = $dG$2$shrunk&255;
      $dB$2 = $dB$2$shrunk&255;
      $dA$2 = $dA$2$shrunk&255;
      $1073 = (($sR$273) - ($dR$2))|0;
      $1074 = Math_imul($1073, $sA$276)|0;
      $1075 = (($1074>>>0) / 255)&-1;
      $1076 = (($1075) + ($dR$2))|0;
      $1077 = (($sG$274) - ($dG$2))|0;
      $1078 = Math_imul($1077, $sA$276)|0;
      $1079 = (($1078>>>0) / 255)&-1;
      $1080 = (($1079) + ($dG$2))|0;
      $1081 = (($sB$275) - ($dB$2))|0;
      $1082 = Math_imul($1081, $sA$276)|0;
      $1083 = (($1082>>>0) / 255)&-1;
      $1084 = (($1083) + ($dB$2))|0;
      $1085 = (($dA$2) + ($sA$276))|0;
      $1086 = Math_imul($dA$2, $sA$276)|0;
      $1087 = (($1086>>>0) / 255)&-1;
      $1088 = (($1085) - ($1087))|0;
      switch ($17|0) {
      case 1:  {
       $1089 = HEAP32[$31>>2]|0;
       $1090 = $1089 & 255;
       $1091 = $1076 >>> $1090;
       $1092 = HEAP32[$32>>2]|0;
       $1093 = $1092 & 255;
       $1094 = $1091 << $1093;
       $1095 = $1089 >>> 8;
       $1096 = $1095 & 255;
       $1097 = $1080 >>> $1096;
       $1098 = $1092 >>> 8;
       $1099 = $1098 & 255;
       $1100 = $1097 << $1099;
       $1101 = $1100 | $1094;
       $1102 = $1089 >>> 16;
       $1103 = $1102 & 255;
       $1104 = $1084 >>> $1103;
       $1105 = $1092 >>> 16;
       $1106 = $1105 & 255;
       $1107 = $1104 << $1106;
       $1108 = $1101 | $1107;
       $1109 = $1089 >>> 24;
       $1110 = $1088 >>> $1109;
       $1111 = $1092 >>> 24;
       $1112 = $1110 << $1111;
       $1113 = $1108 | $1112;
       $1114 = $1113&255;
       HEAP8[$dst$3>>0] = $1114;
       break L60;
       break;
      }
      case 2:  {
       $1115 = HEAP32[$31>>2]|0;
       $1116 = $1115 & 255;
       $1117 = $1076 >>> $1116;
       $1118 = HEAP32[$32>>2]|0;
       $1119 = $1118 & 255;
       $1120 = $1117 << $1119;
       $1121 = $1115 >>> 8;
       $1122 = $1121 & 255;
       $1123 = $1080 >>> $1122;
       $1124 = $1118 >>> 8;
       $1125 = $1124 & 255;
       $1126 = $1123 << $1125;
       $1127 = $1126 | $1120;
       $1128 = $1115 >>> 16;
       $1129 = $1128 & 255;
       $1130 = $1084 >>> $1129;
       $1131 = $1118 >>> 16;
       $1132 = $1131 & 255;
       $1133 = $1130 << $1132;
       $1134 = $1127 | $1133;
       $1135 = $1115 >>> 24;
       $1136 = $1088 >>> $1135;
       $1137 = $1118 >>> 24;
       $1138 = $1136 << $1137;
       $1139 = $1134 | $1138;
       $1140 = $1139&65535;
       HEAP16[$dst$3>>1] = $1140;
       break L60;
       break;
      }
      case 3:  {
       $1141 = $1076&255;
       $1142 = HEAP8[$32>>0]|0;
       $div = ($1142&255) >>> 3;
       $1143 = $div&255;
       $1144 = (($dst$3) + ($1143)|0);
       HEAP8[$1144>>0] = $1141;
       $1145 = $1080&255;
       $1146 = HEAP8[$33>>0]|0;
       $div1 = ($1146&255) >>> 3;
       $1147 = $div1&255;
       $1148 = (($dst$3) + ($1147)|0);
       HEAP8[$1148>>0] = $1145;
       $1149 = $1084&255;
       $1150 = HEAP8[$34>>0]|0;
       $div2 = ($1150&255) >>> 3;
       $1151 = $div2&255;
       $1152 = (($dst$3) + ($1151)|0);
       HEAP8[$1152>>0] = $1149;
       break L60;
       break;
      }
      case 4:  {
       $1153 = HEAP32[$31>>2]|0;
       $1154 = $1153 & 255;
       $1155 = $1076 >>> $1154;
       $1156 = HEAP32[$32>>2]|0;
       $1157 = $1156 & 255;
       $1158 = $1155 << $1157;
       $1159 = $1153 >>> 8;
       $1160 = $1159 & 255;
       $1161 = $1080 >>> $1160;
       $1162 = $1156 >>> 8;
       $1163 = $1162 & 255;
       $1164 = $1161 << $1163;
       $1165 = $1164 | $1158;
       $1166 = $1153 >>> 16;
       $1167 = $1166 & 255;
       $1168 = $1084 >>> $1167;
       $1169 = $1156 >>> 16;
       $1170 = $1169 & 255;
       $1171 = $1168 << $1170;
       $1172 = $1165 | $1171;
       $1173 = $1153 >>> 24;
       $1174 = $1088 >>> $1173;
       $1175 = $1156 >>> 24;
       $1176 = $1174 << $1175;
       $1177 = $1172 | $1176;
       HEAP32[$dst$3>>2] = $1177;
       break L60;
       break;
      }
      default: {
       break L60;
      }
      }
     }
    } while(0);
    $1178 = (($src$3) + ($14)|0);
    $1179 = (($dst$3) + ($17)|0);
    $dst$4 = $1179;$n$3 = $n$2;$src$4 = $1178;
    label = 55;
    continue;
   }
   else if ((label|0) == 55) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $1180 = HEAP8[$src$4>>0]|0;
     $1181 = $1180&255;
     $1182 = HEAP32[$25>>2]|0;
     $1183 = $1181 & $1182;
     $1184 = HEAP32[$26>>2]|0;
     $1185 = $1184 & 255;
     $1186 = $1183 >>> $1185;
     $1187 = HEAP32[$27>>2]|0;
     $1188 = $1187 & 255;
     $1189 = (125924 + ($1188<<2)|0);
     $1190 = HEAP32[$1189>>2]|0;
     $1191 = (($1190) + ($1186)|0);
     $1192 = HEAP32[$28>>2]|0;
     $1193 = $1192 & $1181;
     $1194 = $1184 >>> 8;
     $1195 = $1194 & 255;
     $1196 = $1193 >>> $1195;
     $1197 = $1187 >>> 8;
     $1198 = $1197 & 255;
     $1199 = (125924 + ($1198<<2)|0);
     $1200 = HEAP32[$1199>>2]|0;
     $1201 = (($1200) + ($1196)|0);
     $1202 = HEAP32[$29>>2]|0;
     $1203 = $1202 & $1181;
     $1204 = $1184 >>> 16;
     $1205 = $1204 & 255;
     $1206 = $1203 >>> $1205;
     $1207 = $1187 >>> 16;
     $1208 = $1207 & 255;
     $1209 = (125924 + ($1208<<2)|0);
     $1210 = HEAP32[$1209>>2]|0;
     $1211 = (($1210) + ($1206)|0);
     $1212 = HEAP32[$30>>2]|0;
     $1213 = $1212 & $1181;
     $1214 = $1184 >>> 24;
     $1215 = $1213 >>> $1214;
     $1216 = $1187 >>> 24;
     $1217 = (125924 + ($1216<<2)|0);
     $1218 = HEAP32[$1217>>2]|0;
     $1219 = (($1218) + ($1215)|0);
     $sA$3$shrunk$in = $1219;$sB$3$shrunk$in = $1211;$sG$3$shrunk$in = $1201;$sR$3$shrunk$in = $1191;
     label = 60;
     break;
    }
    case 2:  {
     $1220 = HEAP16[$src$4>>1]|0;
     $1221 = $1220&65535;
     $1222 = HEAP32[$25>>2]|0;
     $1223 = $1221 & $1222;
     $1224 = HEAP32[$26>>2]|0;
     $1225 = $1224 & 255;
     $1226 = $1223 >>> $1225;
     $1227 = HEAP32[$27>>2]|0;
     $1228 = $1227 & 255;
     $1229 = (125924 + ($1228<<2)|0);
     $1230 = HEAP32[$1229>>2]|0;
     $1231 = (($1230) + ($1226)|0);
     $1232 = HEAP32[$28>>2]|0;
     $1233 = $1232 & $1221;
     $1234 = $1224 >>> 8;
     $1235 = $1234 & 255;
     $1236 = $1233 >>> $1235;
     $1237 = $1227 >>> 8;
     $1238 = $1237 & 255;
     $1239 = (125924 + ($1238<<2)|0);
     $1240 = HEAP32[$1239>>2]|0;
     $1241 = (($1240) + ($1236)|0);
     $1242 = HEAP32[$29>>2]|0;
     $1243 = $1242 & $1221;
     $1244 = $1224 >>> 16;
     $1245 = $1244 & 255;
     $1246 = $1243 >>> $1245;
     $1247 = $1227 >>> 16;
     $1248 = $1247 & 255;
     $1249 = (125924 + ($1248<<2)|0);
     $1250 = HEAP32[$1249>>2]|0;
     $1251 = (($1250) + ($1246)|0);
     $1252 = HEAP32[$30>>2]|0;
     $1253 = $1252 & $1221;
     $1254 = $1224 >>> 24;
     $1255 = $1253 >>> $1254;
     $1256 = $1227 >>> 24;
     $1257 = (125924 + ($1256<<2)|0);
     $1258 = HEAP32[$1257>>2]|0;
     $1259 = (($1258) + ($1255)|0);
     $sA$3$shrunk$in = $1259;$sB$3$shrunk$in = $1251;$sG$3$shrunk$in = $1241;$sR$3$shrunk$in = $1231;
     label = 60;
     break;
    }
    case 3:  {
     $1260 = HEAP32[$26>>2]|0;
     $div33 = $1260 >>> 3;
     $1261 = $div33 & 31;
     $1262 = (($src$4) + ($1261)|0);
     $1263 = HEAP8[$1262>>0]|0;
     $1264 = $1260 >>> 11;
     $div34 = $1264 & 31;
     $1265 = (($src$4) + ($div34)|0);
     $1266 = HEAP8[$1265>>0]|0;
     $1267 = $1260 >>> 19;
     $1268 = $1267 & 31;
     $1269 = (($src$4) + ($1268)|0);
     $1270 = HEAP8[$1269>>0]|0;
     $sR$383 = $1263&255;
     $sG$384 = $1266&255;
     $sB$385 = $1270&255;
     $sA$390 = 255;$sB$389 = $sB$385;$sG$388 = $sG$384;$sR$387 = $sR$383;
     label = 61;
     break;
    }
    case 4:  {
     $1271 = HEAP32[$src$4>>2]|0;
     $1272 = HEAP32[$25>>2]|0;
     $1273 = $1272 & $1271;
     $1274 = HEAP32[$26>>2]|0;
     $1275 = $1274 & 255;
     $1276 = $1273 >>> $1275;
     $1277 = HEAP32[$27>>2]|0;
     $1278 = $1277 & 255;
     $1279 = (125924 + ($1278<<2)|0);
     $1280 = HEAP32[$1279>>2]|0;
     $1281 = (($1280) + ($1276)|0);
     $1282 = HEAP32[$28>>2]|0;
     $1283 = $1282 & $1271;
     $1284 = $1274 >>> 8;
     $1285 = $1284 & 255;
     $1286 = $1283 >>> $1285;
     $1287 = $1277 >>> 8;
     $1288 = $1287 & 255;
     $1289 = (125924 + ($1288<<2)|0);
     $1290 = HEAP32[$1289>>2]|0;
     $1291 = (($1290) + ($1286)|0);
     $1292 = HEAP32[$29>>2]|0;
     $1293 = $1292 & $1271;
     $1294 = $1274 >>> 16;
     $1295 = $1294 & 255;
     $1296 = $1293 >>> $1295;
     $1297 = $1277 >>> 16;
     $1298 = $1297 & 255;
     $1299 = (125924 + ($1298<<2)|0);
     $1300 = HEAP32[$1299>>2]|0;
     $1301 = (($1300) + ($1296)|0);
     $1302 = HEAP32[$30>>2]|0;
     $1303 = $1302 & $1271;
     $1304 = $1274 >>> 24;
     $1305 = $1303 >>> $1304;
     $1306 = $1277 >>> 24;
     $1307 = (125924 + ($1306<<2)|0);
     $1308 = HEAP32[$1307>>2]|0;
     $1309 = (($1308) + ($1305)|0);
     $sA$3$shrunk$in = $1309;$sB$3$shrunk$in = $1301;$sG$3$shrunk$in = $1291;$sR$3$shrunk$in = $1281;
     label = 60;
     break;
    }
    default: {
    }
    }
    if ((label|0) == 60) {
     label = 0;
     $sR$3$shrunk = HEAP8[$sR$3$shrunk$in>>0]|0;
     $sG$3$shrunk = HEAP8[$sG$3$shrunk$in>>0]|0;
     $sB$3$shrunk = HEAP8[$sB$3$shrunk$in>>0]|0;
     $sA$3$shrunk = HEAP8[$sA$3$shrunk$in>>0]|0;
     $sR$3 = $sR$3$shrunk&255;
     $sG$3 = $sG$3$shrunk&255;
     $sB$3 = $sB$3$shrunk&255;
     $sA$3 = $sA$3$shrunk&255;
     $1310 = ($sA$3$shrunk<<24>>24)==(0);
     if (!($1310)) {
      $sA$390 = $sA$3;$sB$389 = $sB$3;$sG$388 = $sG$3;$sR$387 = $sR$3;
      label = 61;
     }
    }
    L82: do {
     if ((label|0) == 61) {
      label = 0;
      switch ($17|0) {
      case 1:  {
       $1311 = HEAP8[$dst$4>>0]|0;
       $1312 = $1311&255;
       $1313 = HEAP32[$35>>2]|0;
       $1314 = $1312 & $1313;
       $1315 = HEAP32[$32>>2]|0;
       $1316 = $1315 & 255;
       $1317 = $1314 >>> $1316;
       $1318 = HEAP32[$31>>2]|0;
       $1319 = $1318 & 255;
       $1320 = (125924 + ($1319<<2)|0);
       $1321 = HEAP32[$1320>>2]|0;
       $1322 = (($1321) + ($1317)|0);
       $1323 = HEAP8[$1322>>0]|0;
       $1324 = HEAP32[$36>>2]|0;
       $1325 = $1324 & $1312;
       $1326 = $1315 >>> 8;
       $1327 = $1326 & 255;
       $1328 = $1325 >>> $1327;
       $1329 = $1318 >>> 8;
       $1330 = $1329 & 255;
       $1331 = (125924 + ($1330<<2)|0);
       $1332 = HEAP32[$1331>>2]|0;
       $1333 = (($1332) + ($1328)|0);
       $1334 = HEAP8[$1333>>0]|0;
       $1335 = HEAP32[$37>>2]|0;
       $1336 = $1335 & $1312;
       $1337 = $1315 >>> 16;
       $1338 = $1337 & 255;
       $1339 = $1336 >>> $1338;
       $1340 = $1318 >>> 16;
       $1341 = $1340 & 255;
       $1342 = (125924 + ($1341<<2)|0);
       $1343 = HEAP32[$1342>>2]|0;
       $1344 = (($1343) + ($1339)|0);
       $1345 = HEAP8[$1344>>0]|0;
       $1346 = HEAP32[$38>>2]|0;
       $1347 = $1346 & $1312;
       $1348 = $1315 >>> 24;
       $1349 = $1347 >>> $1348;
       $1350 = $1318 >>> 24;
       $1351 = (125924 + ($1350<<2)|0);
       $1352 = HEAP32[$1351>>2]|0;
       $1353 = (($1352) + ($1349)|0);
       $1354 = HEAP8[$1353>>0]|0;
       $dA$3$shrunk = $1354;$dB$3$shrunk = $1345;$dG$3$shrunk = $1334;$dR$3$shrunk = $1323;
       break;
      }
      case 2:  {
       $1355 = HEAP16[$dst$4>>1]|0;
       $1356 = $1355&65535;
       $1357 = HEAP32[$35>>2]|0;
       $1358 = $1356 & $1357;
       $1359 = HEAP32[$32>>2]|0;
       $1360 = $1359 & 255;
       $1361 = $1358 >>> $1360;
       $1362 = HEAP32[$31>>2]|0;
       $1363 = $1362 & 255;
       $1364 = (125924 + ($1363<<2)|0);
       $1365 = HEAP32[$1364>>2]|0;
       $1366 = (($1365) + ($1361)|0);
       $1367 = HEAP8[$1366>>0]|0;
       $1368 = HEAP32[$36>>2]|0;
       $1369 = $1368 & $1356;
       $1370 = $1359 >>> 8;
       $1371 = $1370 & 255;
       $1372 = $1369 >>> $1371;
       $1373 = $1362 >>> 8;
       $1374 = $1373 & 255;
       $1375 = (125924 + ($1374<<2)|0);
       $1376 = HEAP32[$1375>>2]|0;
       $1377 = (($1376) + ($1372)|0);
       $1378 = HEAP8[$1377>>0]|0;
       $1379 = HEAP32[$37>>2]|0;
       $1380 = $1379 & $1356;
       $1381 = $1359 >>> 16;
       $1382 = $1381 & 255;
       $1383 = $1380 >>> $1382;
       $1384 = $1362 >>> 16;
       $1385 = $1384 & 255;
       $1386 = (125924 + ($1385<<2)|0);
       $1387 = HEAP32[$1386>>2]|0;
       $1388 = (($1387) + ($1383)|0);
       $1389 = HEAP8[$1388>>0]|0;
       $1390 = HEAP32[$38>>2]|0;
       $1391 = $1390 & $1356;
       $1392 = $1359 >>> 24;
       $1393 = $1391 >>> $1392;
       $1394 = $1362 >>> 24;
       $1395 = (125924 + ($1394<<2)|0);
       $1396 = HEAP32[$1395>>2]|0;
       $1397 = (($1396) + ($1393)|0);
       $1398 = HEAP8[$1397>>0]|0;
       $dA$3$shrunk = $1398;$dB$3$shrunk = $1389;$dG$3$shrunk = $1378;$dR$3$shrunk = $1367;
       break;
      }
      case 3:  {
       $1399 = HEAP32[$32>>2]|0;
       $div30 = $1399 >>> 3;
       $1400 = $div30 & 31;
       $1401 = (($dst$4) + ($1400)|0);
       $1402 = HEAP8[$1401>>0]|0;
       $1403 = $1399 >>> 11;
       $div31 = $1403 & 31;
       $1404 = (($dst$4) + ($div31)|0);
       $1405 = HEAP8[$1404>>0]|0;
       $1406 = $1399 >>> 19;
       $1407 = $1406 & 31;
       $1408 = (($dst$4) + ($1407)|0);
       $1409 = HEAP8[$1408>>0]|0;
       $dA$3$shrunk = -1;$dB$3$shrunk = $1409;$dG$3$shrunk = $1405;$dR$3$shrunk = $1402;
       break;
      }
      case 4:  {
       $1410 = HEAP32[$dst$4>>2]|0;
       $1411 = HEAP32[$35>>2]|0;
       $1412 = $1411 & $1410;
       $1413 = HEAP32[$32>>2]|0;
       $1414 = $1413 & 255;
       $1415 = $1412 >>> $1414;
       $1416 = HEAP32[$31>>2]|0;
       $1417 = $1416 & 255;
       $1418 = (125924 + ($1417<<2)|0);
       $1419 = HEAP32[$1418>>2]|0;
       $1420 = (($1419) + ($1415)|0);
       $1421 = HEAP8[$1420>>0]|0;
       $1422 = HEAP32[$36>>2]|0;
       $1423 = $1422 & $1410;
       $1424 = $1413 >>> 8;
       $1425 = $1424 & 255;
       $1426 = $1423 >>> $1425;
       $1427 = $1416 >>> 8;
       $1428 = $1427 & 255;
       $1429 = (125924 + ($1428<<2)|0);
       $1430 = HEAP32[$1429>>2]|0;
       $1431 = (($1430) + ($1426)|0);
       $1432 = HEAP8[$1431>>0]|0;
       $1433 = HEAP32[$37>>2]|0;
       $1434 = $1433 & $1410;
       $1435 = $1413 >>> 16;
       $1436 = $1435 & 255;
       $1437 = $1434 >>> $1436;
       $1438 = $1416 >>> 16;
       $1439 = $1438 & 255;
       $1440 = (125924 + ($1439<<2)|0);
       $1441 = HEAP32[$1440>>2]|0;
       $1442 = (($1441) + ($1437)|0);
       $1443 = HEAP8[$1442>>0]|0;
       $1444 = HEAP32[$38>>2]|0;
       $1445 = $1444 & $1410;
       $1446 = $1413 >>> 24;
       $1447 = $1445 >>> $1446;
       $1448 = $1416 >>> 24;
       $1449 = (125924 + ($1448<<2)|0);
       $1450 = HEAP32[$1449>>2]|0;
       $1451 = (($1450) + ($1447)|0);
       $1452 = HEAP8[$1451>>0]|0;
       $dA$3$shrunk = $1452;$dB$3$shrunk = $1443;$dG$3$shrunk = $1432;$dR$3$shrunk = $1421;
       break;
      }
      default: {
       $dA$3$shrunk = 0;$dB$3$shrunk = 0;$dG$3$shrunk = 0;$dR$3$shrunk = 0;
      }
      }
      $dR$3 = $dR$3$shrunk&255;
      $dG$3 = $dG$3$shrunk&255;
      $dB$3 = $dB$3$shrunk&255;
      $dA$3 = $dA$3$shrunk&255;
      $1453 = (($sR$387) - ($dR$3))|0;
      $1454 = Math_imul($1453, $sA$390)|0;
      $1455 = (($1454>>>0) / 255)&-1;
      $1456 = (($1455) + ($dR$3))|0;
      $1457 = (($sG$388) - ($dG$3))|0;
      $1458 = Math_imul($1457, $sA$390)|0;
      $1459 = (($1458>>>0) / 255)&-1;
      $1460 = (($1459) + ($dG$3))|0;
      $1461 = (($sB$389) - ($dB$3))|0;
      $1462 = Math_imul($1461, $sA$390)|0;
      $1463 = (($1462>>>0) / 255)&-1;
      $1464 = (($1463) + ($dB$3))|0;
      $1465 = (($dA$3) + ($sA$390))|0;
      $1466 = Math_imul($dA$3, $sA$390)|0;
      $1467 = (($1466>>>0) / 255)&-1;
      $1468 = (($1465) - ($1467))|0;
      switch ($17|0) {
      case 1:  {
       $1469 = HEAP32[$31>>2]|0;
       $1470 = $1469 & 255;
       $1471 = $1456 >>> $1470;
       $1472 = HEAP32[$32>>2]|0;
       $1473 = $1472 & 255;
       $1474 = $1471 << $1473;
       $1475 = $1469 >>> 8;
       $1476 = $1475 & 255;
       $1477 = $1460 >>> $1476;
       $1478 = $1472 >>> 8;
       $1479 = $1478 & 255;
       $1480 = $1477 << $1479;
       $1481 = $1480 | $1474;
       $1482 = $1469 >>> 16;
       $1483 = $1482 & 255;
       $1484 = $1464 >>> $1483;
       $1485 = $1472 >>> 16;
       $1486 = $1485 & 255;
       $1487 = $1484 << $1486;
       $1488 = $1481 | $1487;
       $1489 = $1469 >>> 24;
       $1490 = $1468 >>> $1489;
       $1491 = $1472 >>> 24;
       $1492 = $1490 << $1491;
       $1493 = $1488 | $1492;
       $1494 = $1493&255;
       HEAP8[$dst$4>>0] = $1494;
       break L82;
       break;
      }
      case 2:  {
       $1495 = HEAP32[$31>>2]|0;
       $1496 = $1495 & 255;
       $1497 = $1456 >>> $1496;
       $1498 = HEAP32[$32>>2]|0;
       $1499 = $1498 & 255;
       $1500 = $1497 << $1499;
       $1501 = $1495 >>> 8;
       $1502 = $1501 & 255;
       $1503 = $1460 >>> $1502;
       $1504 = $1498 >>> 8;
       $1505 = $1504 & 255;
       $1506 = $1503 << $1505;
       $1507 = $1506 | $1500;
       $1508 = $1495 >>> 16;
       $1509 = $1508 & 255;
       $1510 = $1464 >>> $1509;
       $1511 = $1498 >>> 16;
       $1512 = $1511 & 255;
       $1513 = $1510 << $1512;
       $1514 = $1507 | $1513;
       $1515 = $1495 >>> 24;
       $1516 = $1468 >>> $1515;
       $1517 = $1498 >>> 24;
       $1518 = $1516 << $1517;
       $1519 = $1514 | $1518;
       $1520 = $1519&65535;
       HEAP16[$dst$4>>1] = $1520;
       break L82;
       break;
      }
      case 3:  {
       $1521 = $1456&255;
       $1522 = HEAP8[$32>>0]|0;
       $div27 = ($1522&255) >>> 3;
       $1523 = $div27&255;
       $1524 = (($dst$4) + ($1523)|0);
       HEAP8[$1524>>0] = $1521;
       $1525 = $1460&255;
       $1526 = HEAP8[$33>>0]|0;
       $div28 = ($1526&255) >>> 3;
       $1527 = $div28&255;
       $1528 = (($dst$4) + ($1527)|0);
       HEAP8[$1528>>0] = $1525;
       $1529 = $1464&255;
       $1530 = HEAP8[$34>>0]|0;
       $div29 = ($1530&255) >>> 3;
       $1531 = $div29&255;
       $1532 = (($dst$4) + ($1531)|0);
       HEAP8[$1532>>0] = $1529;
       break L82;
       break;
      }
      case 4:  {
       $1533 = HEAP32[$31>>2]|0;
       $1534 = $1533 & 255;
       $1535 = $1456 >>> $1534;
       $1536 = HEAP32[$32>>2]|0;
       $1537 = $1536 & 255;
       $1538 = $1535 << $1537;
       $1539 = $1533 >>> 8;
       $1540 = $1539 & 255;
       $1541 = $1460 >>> $1540;
       $1542 = $1536 >>> 8;
       $1543 = $1542 & 255;
       $1544 = $1541 << $1543;
       $1545 = $1544 | $1538;
       $1546 = $1533 >>> 16;
       $1547 = $1546 & 255;
       $1548 = $1464 >>> $1547;
       $1549 = $1536 >>> 16;
       $1550 = $1549 & 255;
       $1551 = $1548 << $1550;
       $1552 = $1545 | $1551;
       $1553 = $1533 >>> 24;
       $1554 = $1468 >>> $1553;
       $1555 = $1536 >>> 24;
       $1556 = $1554 << $1555;
       $1557 = $1552 | $1556;
       HEAP32[$dst$4>>2] = $1557;
       break L82;
       break;
      }
      default: {
       break L82;
      }
      }
     }
    } while(0);
    $1558 = (($src$4) + ($14)|0);
    $1559 = (($dst$4) + ($17)|0);
    $1560 = (($n$3) + -1)|0;
    $1561 = ($n$3|0)>(1);
    if ($1561) {
     $dst$1 = $1559;$n$0 = $1560;$src$1 = $1558;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $1562 = (($1558) + ($5)|0);
  $1563 = (($1559) + ($7)|0);
  $1564 = ($39|0)==(0);
  if ($1564) {
   label = 74;
   break;
  } else {
   $$in = $39;$dst$094 = $1563;$src$095 = $1562;
  }
 }
 if ((label|0) == 72) {
  // unreachable;
 }
 else if ((label|0) == 74) {
  return;
 }
}
function _BlitARGBto565PixelAlpha($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $dstp$01 = 0, $dstp$1 = 0, $dstp$2 = 0, $dstp$3 = 0, $dstp$4 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $srcp$02 = 0, $srcp$1 = 0, $srcp$2 = 0, $srcp$3 = 0, $srcp$4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $5 >> 2;
 $7 = ((($info)) + 36|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $8 >> 1;
 $10 = ($3|0)==(0);
 if ($10) {
  return;
 }
 $11 = ((($info)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = HEAP32[$info>>2]|0;
 $14 = (($1) + 3)|0;
 $15 = (($14|0) / 4)&-1;
 $16 = $1 & 3;
 $$in = $3;$dstp$01 = $12;$srcp$02 = $13;
 L4: while(1) {
  $17 = (($$in) + -1)|0;
  switch ($16|0) {
  case 0:  {
   $dstp$1 = $dstp$01;$n$0 = $15;$srcp$1 = $srcp$02;
   label = 4;
   break;
  }
  case 3:  {
   $dstp$2 = $dstp$01;$n$1 = $15;$srcp$2 = $srcp$02;
   label = 8;
   break;
  }
  case 2:  {
   $dstp$3 = $dstp$01;$n$2 = $15;$srcp$3 = $srcp$02;
   label = 12;
   break;
  }
  case 1:  {
   $dstp$4 = $dstp$01;$n$3 = $15;$srcp$4 = $srcp$02;
   label = 16;
   break;
  }
  default: {
   label = 20;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    $18 = HEAP32[$srcp$1>>2]|0;
    $19 = $18 >>> 27;
    switch ($19|0) {
    case 0:  {
     break;
    }
    case 31:  {
     $20 = $18 >>> 8;
     $21 = $20 & 63488;
     $22 = $18 >>> 5;
     $23 = $22 & 2016;
     $24 = $21 | $23;
     $25 = $18 >>> 3;
     $26 = $25 & 31;
     $27 = $24 | $26;
     $28 = $27&65535;
     HEAP16[$dstp$1>>1] = $28;
     break;
    }
    default: {
     $29 = HEAP16[$dstp$1>>1]|0;
     $30 = $29&65535;
     $31 = $18 << 11;
     $32 = $31 & 132120576;
     $33 = $18 >>> 8;
     $34 = $33 & 63488;
     $35 = $32 | $34;
     $36 = $18 >>> 3;
     $37 = $36 & 31;
     $38 = $35 | $37;
     $39 = $30 << 16;
     $40 = $39 | $30;
     $41 = $40 & 132184095;
     $42 = (($38) - ($41))|0;
     $43 = Math_imul($42, $19)|0;
     $44 = $43 >>> 5;
     $45 = (($44) + ($41))|0;
     $46 = $45 & 132184095;
     $47 = $46 >>> 16;
     $48 = $47 | $46;
     $49 = $48&65535;
     HEAP16[$dstp$1>>1] = $49;
    }
    }
    $50 = ((($srcp$1)) + 4|0);
    $51 = ((($dstp$1)) + 2|0);
    $dstp$2 = $51;$n$1 = $n$0;$srcp$2 = $50;
    label = 8;
    continue;
   }
   else if ((label|0) == 8) {
    label = 0;
    $52 = HEAP32[$srcp$2>>2]|0;
    $53 = $52 >>> 27;
    switch ($53|0) {
    case 0:  {
     break;
    }
    case 31:  {
     $54 = $52 >>> 8;
     $55 = $54 & 63488;
     $56 = $52 >>> 5;
     $57 = $56 & 2016;
     $58 = $55 | $57;
     $59 = $52 >>> 3;
     $60 = $59 & 31;
     $61 = $58 | $60;
     $62 = $61&65535;
     HEAP16[$dstp$2>>1] = $62;
     break;
    }
    default: {
     $63 = HEAP16[$dstp$2>>1]|0;
     $64 = $63&65535;
     $65 = $52 << 11;
     $66 = $65 & 132120576;
     $67 = $52 >>> 8;
     $68 = $67 & 63488;
     $69 = $66 | $68;
     $70 = $52 >>> 3;
     $71 = $70 & 31;
     $72 = $69 | $71;
     $73 = $64 << 16;
     $74 = $73 | $64;
     $75 = $74 & 132184095;
     $76 = (($72) - ($75))|0;
     $77 = Math_imul($76, $53)|0;
     $78 = $77 >>> 5;
     $79 = (($78) + ($75))|0;
     $80 = $79 & 132184095;
     $81 = $80 >>> 16;
     $82 = $81 | $80;
     $83 = $82&65535;
     HEAP16[$dstp$2>>1] = $83;
    }
    }
    $84 = ((($srcp$2)) + 4|0);
    $85 = ((($dstp$2)) + 2|0);
    $dstp$3 = $85;$n$2 = $n$1;$srcp$3 = $84;
    label = 12;
    continue;
   }
   else if ((label|0) == 12) {
    label = 0;
    $86 = HEAP32[$srcp$3>>2]|0;
    $87 = $86 >>> 27;
    switch ($87|0) {
    case 0:  {
     break;
    }
    case 31:  {
     $88 = $86 >>> 8;
     $89 = $88 & 63488;
     $90 = $86 >>> 5;
     $91 = $90 & 2016;
     $92 = $89 | $91;
     $93 = $86 >>> 3;
     $94 = $93 & 31;
     $95 = $92 | $94;
     $96 = $95&65535;
     HEAP16[$dstp$3>>1] = $96;
     break;
    }
    default: {
     $97 = HEAP16[$dstp$3>>1]|0;
     $98 = $97&65535;
     $99 = $86 << 11;
     $100 = $99 & 132120576;
     $101 = $86 >>> 8;
     $102 = $101 & 63488;
     $103 = $100 | $102;
     $104 = $86 >>> 3;
     $105 = $104 & 31;
     $106 = $103 | $105;
     $107 = $98 << 16;
     $108 = $107 | $98;
     $109 = $108 & 132184095;
     $110 = (($106) - ($109))|0;
     $111 = Math_imul($110, $87)|0;
     $112 = $111 >>> 5;
     $113 = (($112) + ($109))|0;
     $114 = $113 & 132184095;
     $115 = $114 >>> 16;
     $116 = $115 | $114;
     $117 = $116&65535;
     HEAP16[$dstp$3>>1] = $117;
    }
    }
    $118 = ((($srcp$3)) + 4|0);
    $119 = ((($dstp$3)) + 2|0);
    $dstp$4 = $119;$n$3 = $n$2;$srcp$4 = $118;
    label = 16;
    continue;
   }
   else if ((label|0) == 16) {
    label = 0;
    $120 = HEAP32[$srcp$4>>2]|0;
    $121 = $120 >>> 27;
    switch ($121|0) {
    case 0:  {
     break;
    }
    case 31:  {
     $122 = $120 >>> 8;
     $123 = $122 & 63488;
     $124 = $120 >>> 5;
     $125 = $124 & 2016;
     $126 = $123 | $125;
     $127 = $120 >>> 3;
     $128 = $127 & 31;
     $129 = $126 | $128;
     $130 = $129&65535;
     HEAP16[$dstp$4>>1] = $130;
     break;
    }
    default: {
     $131 = HEAP16[$dstp$4>>1]|0;
     $132 = $131&65535;
     $133 = $120 << 11;
     $134 = $133 & 132120576;
     $135 = $120 >>> 8;
     $136 = $135 & 63488;
     $137 = $134 | $136;
     $138 = $120 >>> 3;
     $139 = $138 & 31;
     $140 = $137 | $139;
     $141 = $132 << 16;
     $142 = $141 | $132;
     $143 = $142 & 132184095;
     $144 = (($140) - ($143))|0;
     $145 = Math_imul($144, $121)|0;
     $146 = $145 >>> 5;
     $147 = (($146) + ($143))|0;
     $148 = $147 & 132184095;
     $149 = $148 >>> 16;
     $150 = $149 | $148;
     $151 = $150&65535;
     HEAP16[$dstp$4>>1] = $151;
    }
    }
    $152 = ((($srcp$4)) + 4|0);
    $153 = ((($dstp$4)) + 2|0);
    $154 = (($n$3) + -1)|0;
    $155 = ($n$3|0)>(1);
    if ($155) {
     $dstp$1 = $153;$n$0 = $154;$srcp$1 = $152;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $156 = (($152) + ($6<<2)|0);
  $157 = (($153) + ($9<<1)|0);
  $158 = ($17|0)==(0);
  if ($158) {
   label = 22;
   break;
  } else {
   $$in = $17;$dstp$01 = $157;$srcp$02 = $156;
  }
 }
 if ((label|0) == 20) {
  // unreachable;
 }
 else if ((label|0) == 22) {
  return;
 }
}
function _Blit555to555SurfaceAlpha($info) {
 $info = $info|0;
 var $$lcssa = 0, $$lcssa13 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $dstp$01 = 0, $dstp$01$us = 0, $dstp$1$us = 0, $dstp$2 = 0, $dstp$3 = 0, $dstp$4 = 0, $n$0$us = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $srcp$02 = 0, $srcp$02$us = 0, $srcp$1$us = 0, $srcp$2 = 0, $srcp$3 = 0, $srcp$4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 63|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = ($1<<24>>24)==(-128);
 if ($2) {
  _Blit16to16SurfaceAlpha128($info,-1058);
  return;
 }
 $3 = $1&255;
 $4 = ((($info)) + 24|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 28|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = HEAP32[$info>>2]|0;
 $9 = ((($info)) + 16|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 >> 1;
 $12 = ((($info)) + 20|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($info)) + 36|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = $15 >> 1;
 $17 = $3 >>> 3;
 $18 = (($7) + -1)|0;
 $19 = ($7|0)==(0);
 if ($19) {
  return;
 }
 $20 = (($5) + 3)|0;
 $21 = (($20|0) / 4)&-1;
 $22 = $5 & 3;
 $23 = ($22|0)==(0);
 if ($23) {
  $109 = $18;$dstp$01$us = $13;$srcp$02$us = $8;
  while(1) {
   $dstp$1$us = $dstp$01$us;$n$0$us = $21;$srcp$1$us = $srcp$02$us;
   while(1) {
    $24 = ((($srcp$1$us)) + 2|0);
    $25 = HEAP16[$srcp$1$us>>1]|0;
    $26 = $25&65535;
    $27 = HEAP16[$dstp$1$us>>1]|0;
    $28 = $27&65535;
    $29 = $26 << 16;
    $30 = $29 | $26;
    $31 = $30 & 65043487;
    $32 = $28 << 16;
    $33 = $32 | $28;
    $34 = $33 & 65043487;
    $35 = (($31) - ($34))|0;
    $36 = Math_imul($35, $17)|0;
    $37 = $36 >>> 5;
    $38 = (($37) + ($34))|0;
    $39 = $38 & 65043487;
    $40 = $39 >>> 16;
    $41 = $40 | $39;
    $42 = $41&65535;
    $43 = ((($dstp$1$us)) + 2|0);
    HEAP16[$dstp$1$us>>1] = $42;
    $44 = ((($srcp$1$us)) + 4|0);
    $45 = HEAP16[$24>>1]|0;
    $46 = $45&65535;
    $47 = HEAP16[$43>>1]|0;
    $48 = $47&65535;
    $49 = $46 << 16;
    $50 = $49 | $46;
    $51 = $50 & 65043487;
    $52 = $48 << 16;
    $53 = $52 | $48;
    $54 = $53 & 65043487;
    $55 = (($51) - ($54))|0;
    $56 = Math_imul($55, $17)|0;
    $57 = $56 >>> 5;
    $58 = (($57) + ($54))|0;
    $59 = $58 & 65043487;
    $60 = $59 >>> 16;
    $61 = $60 | $59;
    $62 = $61&65535;
    $63 = ((($dstp$1$us)) + 4|0);
    HEAP16[$43>>1] = $62;
    $64 = ((($srcp$1$us)) + 6|0);
    $65 = HEAP16[$44>>1]|0;
    $66 = $65&65535;
    $67 = HEAP16[$63>>1]|0;
    $68 = $67&65535;
    $69 = $66 << 16;
    $70 = $69 | $66;
    $71 = $70 & 65043487;
    $72 = $68 << 16;
    $73 = $72 | $68;
    $74 = $73 & 65043487;
    $75 = (($71) - ($74))|0;
    $76 = Math_imul($75, $17)|0;
    $77 = $76 >>> 5;
    $78 = (($77) + ($74))|0;
    $79 = $78 & 65043487;
    $80 = $79 >>> 16;
    $81 = $80 | $79;
    $82 = $81&65535;
    $83 = ((($dstp$1$us)) + 6|0);
    HEAP16[$63>>1] = $82;
    $84 = ((($srcp$1$us)) + 8|0);
    $85 = HEAP16[$64>>1]|0;
    $86 = $85&65535;
    $87 = HEAP16[$83>>1]|0;
    $88 = $87&65535;
    $89 = $86 << 16;
    $90 = $89 | $86;
    $91 = $90 & 65043487;
    $92 = $88 << 16;
    $93 = $92 | $88;
    $94 = $93 & 65043487;
    $95 = (($91) - ($94))|0;
    $96 = Math_imul($95, $17)|0;
    $97 = $96 >>> 5;
    $98 = (($97) + ($94))|0;
    $99 = $98 & 65043487;
    $100 = $99 >>> 16;
    $101 = $100 | $99;
    $102 = $101&65535;
    $103 = ((($dstp$1$us)) + 8|0);
    HEAP16[$83>>1] = $102;
    $104 = (($n$0$us) + -1)|0;
    $105 = ($n$0$us|0)>(1);
    if ($105) {
     $dstp$1$us = $103;$n$0$us = $104;$srcp$1$us = $84;
    } else {
     $$lcssa = $84;$$lcssa13 = $103;
     break;
    }
   }
   $106 = (($$lcssa) + ($11<<1)|0);
   $107 = (($$lcssa13) + ($16<<1)|0);
   $108 = (($109) + -1)|0;
   $110 = ($109|0)==(0);
   if ($110) {
    break;
   } else {
    $109 = $108;$dstp$01$us = $107;$srcp$02$us = $106;
   }
  }
  return;
 } else {
  $196 = $18;$dstp$01 = $13;$srcp$02 = $8;
 }
 L15: while(1) {
  switch ($22|0) {
  case 1:  {
   $dstp$4 = $dstp$01;$n$3 = $21;$srcp$4 = $srcp$02;
   label = 12;
   break;
  }
  case 3:  {
   $dstp$2 = $dstp$01;$n$1 = $21;$srcp$2 = $srcp$02;
   label = 10;
   break;
  }
  case 2:  {
   $dstp$3 = $dstp$01;$n$2 = $21;$srcp$3 = $srcp$02;
   label = 11;
   break;
  }
  default: {
   label = 13;
   break L15;
  }
  }
  while(1) {
   if ((label|0) == 10) {
    label = 0;
    $134 = ((($srcp$2)) + 2|0);
    $135 = HEAP16[$srcp$2>>1]|0;
    $136 = $135&65535;
    $137 = HEAP16[$dstp$2>>1]|0;
    $138 = $137&65535;
    $139 = $136 << 16;
    $140 = $139 | $136;
    $141 = $140 & 65043487;
    $142 = $138 << 16;
    $143 = $142 | $138;
    $144 = $143 & 65043487;
    $145 = (($141) - ($144))|0;
    $146 = Math_imul($145, $17)|0;
    $147 = $146 >>> 5;
    $148 = (($147) + ($144))|0;
    $149 = $148 & 65043487;
    $150 = $149 >>> 16;
    $151 = $150 | $149;
    $152 = $151&65535;
    $153 = ((($dstp$2)) + 2|0);
    HEAP16[$dstp$2>>1] = $152;
    $dstp$3 = $153;$n$2 = $n$1;$srcp$3 = $134;
    label = 11;
    continue;
   }
   else if ((label|0) == 11) {
    label = 0;
    $154 = ((($srcp$3)) + 2|0);
    $155 = HEAP16[$srcp$3>>1]|0;
    $156 = $155&65535;
    $157 = HEAP16[$dstp$3>>1]|0;
    $158 = $157&65535;
    $159 = $156 << 16;
    $160 = $159 | $156;
    $161 = $160 & 65043487;
    $162 = $158 << 16;
    $163 = $162 | $158;
    $164 = $163 & 65043487;
    $165 = (($161) - ($164))|0;
    $166 = Math_imul($165, $17)|0;
    $167 = $166 >>> 5;
    $168 = (($167) + ($164))|0;
    $169 = $168 & 65043487;
    $170 = $169 >>> 16;
    $171 = $170 | $169;
    $172 = $171&65535;
    $173 = ((($dstp$3)) + 2|0);
    HEAP16[$dstp$3>>1] = $172;
    $dstp$4 = $173;$n$3 = $n$2;$srcp$4 = $154;
    label = 12;
    continue;
   }
   else if ((label|0) == 12) {
    label = 0;
    $114 = ((($srcp$4)) + 2|0);
    $174 = HEAP16[$srcp$4>>1]|0;
    $175 = $174&65535;
    $176 = HEAP16[$dstp$4>>1]|0;
    $177 = $176&65535;
    $178 = $175 << 16;
    $179 = $178 | $175;
    $180 = $179 & 65043487;
    $181 = $177 << 16;
    $182 = $181 | $177;
    $183 = $182 & 65043487;
    $184 = (($180) - ($183))|0;
    $185 = Math_imul($184, $17)|0;
    $186 = $185 >>> 5;
    $187 = (($186) + ($183))|0;
    $188 = $187 & 65043487;
    $189 = $188 >>> 16;
    $190 = $189 | $188;
    $191 = $190&65535;
    $117 = ((($dstp$4)) + 2|0);
    HEAP16[$dstp$4>>1] = $191;
    $192 = ($n$3|0)>(1);
    if (!($192)) {
     break;
    }
    $111 = (($n$3) + -1)|0;
    $112 = ((($srcp$4)) + 4|0);
    $113 = HEAP16[$114>>1]|0;
    $115 = $113&65535;
    $116 = HEAP16[$117>>1]|0;
    $118 = $116&65535;
    $119 = $115 << 16;
    $120 = $119 | $115;
    $121 = $120 & 65043487;
    $122 = $118 << 16;
    $123 = $122 | $118;
    $124 = $123 & 65043487;
    $125 = (($121) - ($124))|0;
    $126 = Math_imul($125, $17)|0;
    $127 = $126 >>> 5;
    $128 = (($127) + ($124))|0;
    $129 = $128 & 65043487;
    $130 = $129 >>> 16;
    $131 = $130 | $129;
    $132 = $131&65535;
    $133 = ((($dstp$4)) + 4|0);
    HEAP16[$117>>1] = $132;
    $dstp$2 = $133;$n$1 = $111;$srcp$2 = $112;
    label = 10;
    continue;
   }
  }
  $193 = (($114) + ($11<<1)|0);
  $194 = (($117) + ($16<<1)|0);
  $195 = (($196) + -1)|0;
  $197 = ($196|0)==(0);
  if ($197) {
   label = 15;
   break;
  } else {
   $196 = $195;$dstp$01 = $194;$srcp$02 = $193;
  }
 }
 if ((label|0) == 13) {
  // unreachable;
 }
 else if ((label|0) == 15) {
  return;
 }
}
function _Blit16to16SurfaceAlpha128($info,$mask) {
 $info = $info|0;
 $mask = $mask|0;
 var $$in = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $dstp$016 = 0, $dstp$1 = 0, $dstp$2$lcssa = 0, $dstp$23 = 0, $dstp$3 = 0, $dstp$3$sink = 0, $dstp$4$lcssa = 0, $dstp$4$ph = 0, $dstp$49 = 0, $dstp$5 = 0, $prev_sw$0$lcssa = 0, $prev_sw$02 = 0, $scevgep = 0, $scevgep27 = 0, $scevgep28 = 0, $scevgep29 = 0, $scevgep30 = 0, $scevgep31 = 0, $scevgep32 = 0, $scevgep33 = 0;
 var $srcp$0$be = 0, $srcp$017 = 0, $srcp$1 = 0, $srcp$2$lcssa = 0, $srcp$24 = 0, $srcp$3 = 0, $srcp$4$lcssa = 0, $srcp$4$ph = 0, $srcp$410 = 0, $srcp$5 = 0, $w$0 = 0, $w$1$lcssa = 0, $w$11 = 0, $w4$0$lcssa = 0, $w4$0$ph = 0, $w4$08 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $5 >> 1;
 $7 = ((($info)) + 36|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $8 >> 1;
 $10 = ($3|0)==(0);
 if ($10) {
  return;
 }
 $11 = ((($info)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = HEAP32[$info>>2]|0;
 $14 = $mask&65535;
 $15 = $14 ^ 65535;
 $16 = (($1) + -1)|0;
 $17 = $14 << 16;
 $18 = $17 | $14;
 $19 = $18 ^ -1;
 $20 = (($6) + -1)|0;
 $$in = $3;$dstp$016 = $12;$srcp$017 = $13;
 while(1) {
  $21 = (($$in) + -1)|0;
  $22 = $srcp$017;
  $23 = $dstp$016;
  $24 = $22 ^ $23;
  $25 = $24 & 2;
  $26 = ($25|0)==(0);
  if ($26) {
   $84 = $22 & 2;
   $85 = ($84|0)==(0);
   if ($85) {
    $dstp$4$ph = $dstp$016;$srcp$4$ph = $srcp$017;$w4$0$ph = $1;
   } else {
    $86 = HEAP16[$dstp$016>>1]|0;
    $87 = HEAP16[$srcp$017>>1]|0;
    $88 = $87 & $mask;
    $89 = $88&65535;
    $90 = $86 & $mask;
    $91 = $90&65535;
    $92 = (($89) + ($91))|0;
    $93 = $92 >>> 1;
    $94 = $87 & $86;
    $95 = $94&65535;
    $96 = $95 & $15;
    $97 = (($93) + ($96))|0;
    $98 = $97&65535;
    HEAP16[$dstp$016>>1] = $98;
    $99 = ((($srcp$017)) + 2|0);
    $100 = ((($dstp$016)) + 2|0);
    $dstp$4$ph = $100;$srcp$4$ph = $99;$w4$0$ph = $16;
   }
   $101 = ($w4$0$ph|0)>(1);
   if ($101) {
    $102 = (($w4$0$ph) + -2)|0;
    $103 = $102 & -2;
    $scevgep30 = ((($dstp$4$ph)) + 4|0);
    $scevgep32 = ((($srcp$4$ph)) + 4|0);
    $dstp$49 = $dstp$4$ph;$srcp$410 = $srcp$4$ph;$w4$08 = $w4$0$ph;
    while(1) {
     $104 = HEAP32[$srcp$410>>2]|0;
     $105 = HEAP32[$dstp$49>>2]|0;
     $106 = $104 & $18;
     $107 = $106 >>> 1;
     $108 = $105 & $18;
     $109 = $108 >>> 1;
     $110 = (($109) + ($107))|0;
     $111 = $104 & $19;
     $112 = $111 & $105;
     $113 = (($110) + ($112))|0;
     HEAP32[$dstp$49>>2] = $113;
     $114 = ((($srcp$410)) + 4|0);
     $115 = ((($dstp$49)) + 4|0);
     $116 = (($w4$08) + -2)|0;
     $117 = ($116|0)>(1);
     if ($117) {
      $dstp$49 = $115;$srcp$410 = $114;$w4$08 = $116;
     } else {
      break;
     }
    }
    $118 = (($102) - ($103))|0;
    $scevgep31 = (($scevgep30) + ($103<<1)|0);
    $scevgep33 = (($scevgep32) + ($103<<1)|0);
    $dstp$4$lcssa = $scevgep31;$srcp$4$lcssa = $scevgep33;$w4$0$lcssa = $118;
   } else {
    $dstp$4$lcssa = $dstp$4$ph;$srcp$4$lcssa = $srcp$4$ph;$w4$0$lcssa = $w4$0$ph;
   }
   $119 = ($w4$0$lcssa|0)==(0);
   if ($119) {
    $dstp$5 = $dstp$4$lcssa;$srcp$5 = $srcp$4$lcssa;
   } else {
    $120 = HEAP16[$dstp$4$lcssa>>1]|0;
    $121 = HEAP16[$srcp$4$lcssa>>1]|0;
    $122 = $121 & $mask;
    $123 = $122&65535;
    $124 = $120 & $mask;
    $125 = $124&65535;
    $126 = (($123) + ($125))|0;
    $127 = $126 >>> 1;
    $128 = $121 & $120;
    $129 = $128&65535;
    $130 = $129 & $15;
    $131 = (($127) + ($130))|0;
    $132 = $131&65535;
    HEAP16[$dstp$4$lcssa>>1] = $132;
    $133 = ((($srcp$4$lcssa)) + 2|0);
    $134 = ((($dstp$4$lcssa)) + 2|0);
    $dstp$5 = $134;$srcp$5 = $133;
   }
   $135 = (($srcp$5) + ($6<<1)|0);
   $dstp$3$sink = $dstp$5;$srcp$0$be = $135;
  } else {
   $27 = $23 & 2;
   $28 = ($27|0)==(0);
   if ($28) {
    $dstp$1 = $dstp$016;$srcp$1 = $srcp$017;$w$0 = $1;
   } else {
    $29 = HEAP16[$dstp$016>>1]|0;
    $30 = HEAP16[$srcp$017>>1]|0;
    $31 = $30 & $mask;
    $32 = $31&65535;
    $33 = $29 & $mask;
    $34 = $33&65535;
    $35 = (($32) + ($34))|0;
    $36 = $35 >>> 1;
    $37 = $30 & $29;
    $38 = $37&65535;
    $39 = $38 & $15;
    $40 = (($36) + ($39))|0;
    $41 = $40&65535;
    HEAP16[$dstp$016>>1] = $41;
    $42 = ((($dstp$016)) + 2|0);
    $43 = ((($srcp$017)) + 2|0);
    $dstp$1 = $42;$srcp$1 = $43;$w$0 = $16;
   }
   $44 = ((($srcp$1)) + 2|0);
   $45 = ((($srcp$1)) + -2|0);
   $46 = HEAP32[$45>>2]|0;
   $47 = ($w$0|0)>(1);
   if ($47) {
    $48 = (($w$0) + -2)|0;
    $49 = $48 & -2;
    $scevgep = ((($srcp$1)) + 6|0);
    $scevgep28 = ((($dstp$1)) + 4|0);
    $dstp$23 = $dstp$1;$prev_sw$02 = $46;$srcp$24 = $44;$w$11 = $w$0;
    while(1) {
     $50 = HEAP32[$srcp$24>>2]|0;
     $51 = HEAP32[$dstp$23>>2]|0;
     $52 = $prev_sw$02 >>> 16;
     $53 = $50 << 16;
     $54 = $53 | $52;
     $55 = $54 & $18;
     $56 = $55 >>> 1;
     $57 = $51 & $18;
     $58 = $57 >>> 1;
     $59 = (($56) + ($58))|0;
     $60 = $51 & $19;
     $61 = $60 & $54;
     $62 = (($59) + ($61))|0;
     HEAP32[$dstp$23>>2] = $62;
     $63 = ((($dstp$23)) + 4|0);
     $64 = ((($srcp$24)) + 4|0);
     $65 = (($w$11) + -2)|0;
     $66 = ($65|0)>(1);
     if ($66) {
      $dstp$23 = $63;$prev_sw$02 = $50;$srcp$24 = $64;$w$11 = $65;
     } else {
      $$lcssa = $50;
      break;
     }
    }
    $67 = (($48) - ($49))|0;
    $scevgep27 = (($scevgep) + ($49<<1)|0);
    $scevgep29 = (($scevgep28) + ($49<<1)|0);
    $dstp$2$lcssa = $scevgep29;$prev_sw$0$lcssa = $$lcssa;$srcp$2$lcssa = $scevgep27;$w$1$lcssa = $67;
   } else {
    $dstp$2$lcssa = $dstp$1;$prev_sw$0$lcssa = $46;$srcp$2$lcssa = $44;$w$1$lcssa = $w$0;
   }
   $68 = ($w$1$lcssa|0)==(0);
   if ($68) {
    $dstp$3 = $dstp$2$lcssa;$srcp$3 = $srcp$2$lcssa;
   } else {
    $69 = HEAP16[$dstp$2$lcssa>>1]|0;
    $70 = $prev_sw$0$lcssa >>> 16;
    $71 = $70 & $14;
    $72 = $69&65535;
    $73 = $69 & $mask;
    $74 = $73&65535;
    $75 = (($74) + ($71))|0;
    $76 = $75 >>> 1;
    $77 = $70 & $15;
    $78 = $77 & $72;
    $79 = (($76) + ($78))|0;
    $80 = $79&65535;
    HEAP16[$dstp$2$lcssa>>1] = $80;
    $81 = ((($srcp$2$lcssa)) + 2|0);
    $82 = ((($dstp$2$lcssa)) + 2|0);
    $dstp$3 = $82;$srcp$3 = $81;
   }
   $83 = (($srcp$3) + ($20<<1)|0);
   $dstp$3$sink = $dstp$3;$srcp$0$be = $83;
  }
  $136 = (($dstp$3$sink) + ($9<<1)|0);
  $137 = ($21|0)==(0);
  if ($137) {
   break;
  } else {
   $$in = $21;$dstp$016 = $136;$srcp$017 = $srcp$0$be;
  }
 }
 return;
}
function _BlitNtoNSurfaceAlpha($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0;
 var $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0;
 var $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0;
 var $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0;
 var $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0;
 var $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0;
 var $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0;
 var $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0;
 var $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0;
 var $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0;
 var $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0;
 var $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0;
 var $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0;
 var $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0;
 var $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0;
 var $1267 = 0, $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0;
 var $1285 = 0, $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0;
 var $1302 = 0, $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0;
 var $1320 = 0, $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0;
 var $1339 = 0, $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0;
 var $1357 = 0, $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0;
 var $1375 = 0, $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0;
 var $1393 = 0, $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0;
 var $1410 = 0, $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0;
 var $1429 = 0, $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0;
 var $1447 = 0, $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0;
 var $1465 = 0, $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0;
 var $1483 = 0, $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0;
 var $1500 = 0, $1501 = 0, $1502 = 0, $1503 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0;
 var $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0;
 var $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0;
 var $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0;
 var $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0;
 var $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0;
 var $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0;
 var $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0;
 var $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0;
 var $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0;
 var $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0;
 var $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0;
 var $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0;
 var $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0;
 var $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0;
 var $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0;
 var $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0;
 var $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0;
 var $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0;
 var $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0;
 var $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0;
 var $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0;
 var $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0;
 var $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0;
 var $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0;
 var $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0;
 var $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0;
 var $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0;
 var $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0;
 var $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0;
 var $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0;
 var $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0;
 var $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0;
 var $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0;
 var $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0;
 var $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0;
 var $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0;
 var $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0;
 var $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0;
 var $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0;
 var $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0;
 var $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0;
 var $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0;
 var $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0;
 var $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0;
 var $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $dA$0 = 0, $dA$0$shrunk = 0, $dA$1 = 0, $dA$1$shrunk = 0, $dA$2 = 0, $dA$2$shrunk = 0, $dA$3 = 0, $dA$3$shrunk = 0, $dB$0 = 0, $dB$0$shrunk = 0, $dB$1 = 0, $dB$1$shrunk = 0, $dB$2 = 0, $dB$2$shrunk = 0, $dB$3 = 0;
 var $dB$3$shrunk = 0, $dG$0 = 0, $dG$0$shrunk = 0, $dG$1 = 0, $dG$1$shrunk = 0, $dG$2 = 0, $dG$2$shrunk = 0, $dG$3 = 0, $dG$3$shrunk = 0, $dR$0 = 0, $dR$0$shrunk = 0, $dR$1 = 0, $dR$1$shrunk = 0, $dR$2 = 0, $dR$2$shrunk = 0, $dR$3 = 0, $dR$3$shrunk = 0, $div = 0, $div1 = 0, $div10 = 0;
 var $div11 = 0, $div12 = 0, $div13 = 0, $div15 = 0, $div16 = 0, $div18 = 0, $div19 = 0, $div2 = 0, $div20 = 0, $div21 = 0, $div22 = 0, $div24 = 0, $div25 = 0, $div27 = 0, $div28 = 0, $div29 = 0, $div3 = 0, $div30 = 0, $div31 = 0, $div33 = 0;
 var $div34 = 0, $div4 = 0, $div6 = 0, $div7 = 0, $div9 = 0, $dst$039 = 0, $dst$1 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $or$cond = 0, $sB$0 = 0, $sB$0$shrunk = 0, $sB$1 = 0, $sB$1$shrunk = 0, $sB$2 = 0;
 var $sB$2$shrunk = 0, $sB$3 = 0, $sB$3$shrunk = 0, $sG$0 = 0, $sG$0$shrunk = 0, $sG$1 = 0, $sG$1$shrunk = 0, $sG$2 = 0, $sG$2$shrunk = 0, $sG$3 = 0, $sG$3$shrunk = 0, $sR$0 = 0, $sR$0$shrunk = 0, $sR$1 = 0, $sR$1$shrunk = 0, $sR$2 = 0, $sR$2$shrunk = 0, $sR$3 = 0, $sR$3$shrunk = 0, $src$040 = 0;
 var $src$1 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 40|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($info)) + 44|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($9)) + 9|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = $13&255;
 $15 = ((($11)) + 9|0);
 $16 = HEAP8[$15>>0]|0;
 $17 = $16&255;
 $18 = ((($info)) + 63|0);
 $19 = HEAP8[$18>>0]|0;
 $20 = $19&255;
 $21 = ($19<<24>>24)==(0);
 $22 = ($3|0)==(0);
 $or$cond = $21 | $22;
 if ($or$cond) {
  return;
 }
 $23 = ((($info)) + 20|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = HEAP32[$info>>2]|0;
 $26 = (($1) + 3)|0;
 $27 = (($26|0) / 4)&-1;
 $28 = $1 & 3;
 $29 = ((($11)) + 28|0);
 $30 = ((($11)) + 32|0);
 $31 = ((($11)) + 33|0);
 $32 = ((($11)) + 34|0);
 $33 = ((($11)) + 12|0);
 $34 = ((($11)) + 16|0);
 $35 = ((($11)) + 20|0);
 $36 = ((($11)) + 24|0);
 $37 = ((($9)) + 12|0);
 $38 = ((($9)) + 32|0);
 $39 = ((($9)) + 28|0);
 $40 = ((($9)) + 16|0);
 $41 = ((($9)) + 20|0);
 $$in = $3;$dst$039 = $24;$src$040 = $25;
 L4: while(1) {
  $42 = (($$in) + -1)|0;
  switch ($28|0) {
  case 0:  {
   $dst$1 = $dst$039;$n$0 = $27;$src$1 = $src$040;
   label = 4;
   break;
  }
  case 3:  {
   $dst$2 = $dst$039;$n$1 = $27;$src$2 = $src$040;
   label = 20;
   break;
  }
  case 2:  {
   $dst$3 = $dst$039;$n$2 = $27;$src$3 = $src$040;
   label = 36;
   break;
  }
  case 1:  {
   $dst$4 = $dst$039;$n$3 = $27;$src$4 = $src$040;
   label = 52;
   break;
  }
  default: {
   label = 68;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $43 = HEAP8[$src$1>>0]|0;
     $44 = $43&255;
     $45 = HEAP32[$37>>2]|0;
     $46 = $44 & $45;
     $47 = HEAP32[$38>>2]|0;
     $48 = $47 & 255;
     $49 = $46 >>> $48;
     $50 = HEAP32[$39>>2]|0;
     $51 = $50 & 255;
     $52 = (125924 + ($51<<2)|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = (($53) + ($49)|0);
     $55 = HEAP8[$54>>0]|0;
     $56 = HEAP32[$40>>2]|0;
     $57 = $56 & $44;
     $58 = $47 >>> 8;
     $59 = $58 & 255;
     $60 = $57 >>> $59;
     $61 = $50 >>> 8;
     $62 = $61 & 255;
     $63 = (125924 + ($62<<2)|0);
     $64 = HEAP32[$63>>2]|0;
     $65 = (($64) + ($60)|0);
     $66 = HEAP8[$65>>0]|0;
     $67 = HEAP32[$41>>2]|0;
     $68 = $67 & $44;
     $69 = $47 >>> 16;
     $70 = $69 & 255;
     $71 = $68 >>> $70;
     $72 = $50 >>> 16;
     $73 = $72 & 255;
     $74 = (125924 + ($73<<2)|0);
     $75 = HEAP32[$74>>2]|0;
     $76 = (($75) + ($71)|0);
     $77 = HEAP8[$76>>0]|0;
     $sB$0$shrunk = $77;$sG$0$shrunk = $66;$sR$0$shrunk = $55;
     break;
    }
    case 2:  {
     $78 = HEAP16[$src$1>>1]|0;
     $79 = $78&65535;
     $80 = HEAP32[$37>>2]|0;
     $81 = $79 & $80;
     $82 = HEAP32[$38>>2]|0;
     $83 = $82 & 255;
     $84 = $81 >>> $83;
     $85 = HEAP32[$39>>2]|0;
     $86 = $85 & 255;
     $87 = (125924 + ($86<<2)|0);
     $88 = HEAP32[$87>>2]|0;
     $89 = (($88) + ($84)|0);
     $90 = HEAP8[$89>>0]|0;
     $91 = HEAP32[$40>>2]|0;
     $92 = $91 & $79;
     $93 = $82 >>> 8;
     $94 = $93 & 255;
     $95 = $92 >>> $94;
     $96 = $85 >>> 8;
     $97 = $96 & 255;
     $98 = (125924 + ($97<<2)|0);
     $99 = HEAP32[$98>>2]|0;
     $100 = (($99) + ($95)|0);
     $101 = HEAP8[$100>>0]|0;
     $102 = HEAP32[$41>>2]|0;
     $103 = $102 & $79;
     $104 = $82 >>> 16;
     $105 = $104 & 255;
     $106 = $103 >>> $105;
     $107 = $85 >>> 16;
     $108 = $107 & 255;
     $109 = (125924 + ($108<<2)|0);
     $110 = HEAP32[$109>>2]|0;
     $111 = (($110) + ($106)|0);
     $112 = HEAP8[$111>>0]|0;
     $sB$0$shrunk = $112;$sG$0$shrunk = $101;$sR$0$shrunk = $90;
     break;
    }
    case 3:  {
     $113 = HEAP32[$38>>2]|0;
     $div24 = $113 >>> 3;
     $114 = $div24 & 31;
     $115 = (($src$1) + ($114)|0);
     $116 = HEAP8[$115>>0]|0;
     $117 = $113 >>> 11;
     $div25 = $117 & 31;
     $118 = (($src$1) + ($div25)|0);
     $119 = HEAP8[$118>>0]|0;
     $120 = $113 >>> 19;
     $121 = $120 & 31;
     $122 = (($src$1) + ($121)|0);
     $123 = HEAP8[$122>>0]|0;
     $sB$0$shrunk = $123;$sG$0$shrunk = $119;$sR$0$shrunk = $116;
     break;
    }
    case 4:  {
     $124 = HEAP32[$src$1>>2]|0;
     $125 = HEAP32[$37>>2]|0;
     $126 = $125 & $124;
     $127 = HEAP32[$38>>2]|0;
     $128 = $127 & 255;
     $129 = $126 >>> $128;
     $130 = HEAP32[$39>>2]|0;
     $131 = $130 & 255;
     $132 = (125924 + ($131<<2)|0);
     $133 = HEAP32[$132>>2]|0;
     $134 = (($133) + ($129)|0);
     $135 = HEAP8[$134>>0]|0;
     $136 = HEAP32[$40>>2]|0;
     $137 = $136 & $124;
     $138 = $127 >>> 8;
     $139 = $138 & 255;
     $140 = $137 >>> $139;
     $141 = $130 >>> 8;
     $142 = $141 & 255;
     $143 = (125924 + ($142<<2)|0);
     $144 = HEAP32[$143>>2]|0;
     $145 = (($144) + ($140)|0);
     $146 = HEAP8[$145>>0]|0;
     $147 = HEAP32[$41>>2]|0;
     $148 = $147 & $124;
     $149 = $127 >>> 16;
     $150 = $149 & 255;
     $151 = $148 >>> $150;
     $152 = $130 >>> 16;
     $153 = $152 & 255;
     $154 = (125924 + ($153<<2)|0);
     $155 = HEAP32[$154>>2]|0;
     $156 = (($155) + ($151)|0);
     $157 = HEAP8[$156>>0]|0;
     $sB$0$shrunk = $157;$sG$0$shrunk = $146;$sR$0$shrunk = $135;
     break;
    }
    default: {
     $sB$0$shrunk = 0;$sG$0$shrunk = 0;$sR$0$shrunk = 0;
    }
    }
    $sR$0 = $sR$0$shrunk&255;
    $sG$0 = $sG$0$shrunk&255;
    $sB$0 = $sB$0$shrunk&255;
    switch ($17|0) {
    case 1:  {
     $158 = HEAP8[$dst$1>>0]|0;
     $159 = $158&255;
     $160 = HEAP32[$33>>2]|0;
     $161 = $159 & $160;
     $162 = HEAP32[$30>>2]|0;
     $163 = $162 & 255;
     $164 = $161 >>> $163;
     $165 = HEAP32[$29>>2]|0;
     $166 = $165 & 255;
     $167 = (125924 + ($166<<2)|0);
     $168 = HEAP32[$167>>2]|0;
     $169 = (($168) + ($164)|0);
     $170 = HEAP8[$169>>0]|0;
     $171 = HEAP32[$34>>2]|0;
     $172 = $171 & $159;
     $173 = $162 >>> 8;
     $174 = $173 & 255;
     $175 = $172 >>> $174;
     $176 = $165 >>> 8;
     $177 = $176 & 255;
     $178 = (125924 + ($177<<2)|0);
     $179 = HEAP32[$178>>2]|0;
     $180 = (($179) + ($175)|0);
     $181 = HEAP8[$180>>0]|0;
     $182 = HEAP32[$35>>2]|0;
     $183 = $182 & $159;
     $184 = $162 >>> 16;
     $185 = $184 & 255;
     $186 = $183 >>> $185;
     $187 = $165 >>> 16;
     $188 = $187 & 255;
     $189 = (125924 + ($188<<2)|0);
     $190 = HEAP32[$189>>2]|0;
     $191 = (($190) + ($186)|0);
     $192 = HEAP8[$191>>0]|0;
     $193 = HEAP32[$36>>2]|0;
     $194 = $193 & $159;
     $195 = $162 >>> 24;
     $196 = $194 >>> $195;
     $197 = $165 >>> 24;
     $198 = (125924 + ($197<<2)|0);
     $199 = HEAP32[$198>>2]|0;
     $200 = (($199) + ($196)|0);
     $201 = HEAP8[$200>>0]|0;
     $dA$0$shrunk = $201;$dB$0$shrunk = $192;$dG$0$shrunk = $181;$dR$0$shrunk = $170;
     break;
    }
    case 2:  {
     $202 = HEAP16[$dst$1>>1]|0;
     $203 = $202&65535;
     $204 = HEAP32[$33>>2]|0;
     $205 = $203 & $204;
     $206 = HEAP32[$30>>2]|0;
     $207 = $206 & 255;
     $208 = $205 >>> $207;
     $209 = HEAP32[$29>>2]|0;
     $210 = $209 & 255;
     $211 = (125924 + ($210<<2)|0);
     $212 = HEAP32[$211>>2]|0;
     $213 = (($212) + ($208)|0);
     $214 = HEAP8[$213>>0]|0;
     $215 = HEAP32[$34>>2]|0;
     $216 = $215 & $203;
     $217 = $206 >>> 8;
     $218 = $217 & 255;
     $219 = $216 >>> $218;
     $220 = $209 >>> 8;
     $221 = $220 & 255;
     $222 = (125924 + ($221<<2)|0);
     $223 = HEAP32[$222>>2]|0;
     $224 = (($223) + ($219)|0);
     $225 = HEAP8[$224>>0]|0;
     $226 = HEAP32[$35>>2]|0;
     $227 = $226 & $203;
     $228 = $206 >>> 16;
     $229 = $228 & 255;
     $230 = $227 >>> $229;
     $231 = $209 >>> 16;
     $232 = $231 & 255;
     $233 = (125924 + ($232<<2)|0);
     $234 = HEAP32[$233>>2]|0;
     $235 = (($234) + ($230)|0);
     $236 = HEAP8[$235>>0]|0;
     $237 = HEAP32[$36>>2]|0;
     $238 = $237 & $203;
     $239 = $206 >>> 24;
     $240 = $238 >>> $239;
     $241 = $209 >>> 24;
     $242 = (125924 + ($241<<2)|0);
     $243 = HEAP32[$242>>2]|0;
     $244 = (($243) + ($240)|0);
     $245 = HEAP8[$244>>0]|0;
     $dA$0$shrunk = $245;$dB$0$shrunk = $236;$dG$0$shrunk = $225;$dR$0$shrunk = $214;
     break;
    }
    case 3:  {
     $246 = HEAP32[$30>>2]|0;
     $div21 = $246 >>> 3;
     $247 = $div21 & 31;
     $248 = (($dst$1) + ($247)|0);
     $249 = HEAP8[$248>>0]|0;
     $250 = $246 >>> 11;
     $div22 = $250 & 31;
     $251 = (($dst$1) + ($div22)|0);
     $252 = HEAP8[$251>>0]|0;
     $253 = $246 >>> 19;
     $254 = $253 & 31;
     $255 = (($dst$1) + ($254)|0);
     $256 = HEAP8[$255>>0]|0;
     $dA$0$shrunk = -1;$dB$0$shrunk = $256;$dG$0$shrunk = $252;$dR$0$shrunk = $249;
     break;
    }
    case 4:  {
     $257 = HEAP32[$dst$1>>2]|0;
     $258 = HEAP32[$33>>2]|0;
     $259 = $258 & $257;
     $260 = HEAP32[$30>>2]|0;
     $261 = $260 & 255;
     $262 = $259 >>> $261;
     $263 = HEAP32[$29>>2]|0;
     $264 = $263 & 255;
     $265 = (125924 + ($264<<2)|0);
     $266 = HEAP32[$265>>2]|0;
     $267 = (($266) + ($262)|0);
     $268 = HEAP8[$267>>0]|0;
     $269 = HEAP32[$34>>2]|0;
     $270 = $269 & $257;
     $271 = $260 >>> 8;
     $272 = $271 & 255;
     $273 = $270 >>> $272;
     $274 = $263 >>> 8;
     $275 = $274 & 255;
     $276 = (125924 + ($275<<2)|0);
     $277 = HEAP32[$276>>2]|0;
     $278 = (($277) + ($273)|0);
     $279 = HEAP8[$278>>0]|0;
     $280 = HEAP32[$35>>2]|0;
     $281 = $280 & $257;
     $282 = $260 >>> 16;
     $283 = $282 & 255;
     $284 = $281 >>> $283;
     $285 = $263 >>> 16;
     $286 = $285 & 255;
     $287 = (125924 + ($286<<2)|0);
     $288 = HEAP32[$287>>2]|0;
     $289 = (($288) + ($284)|0);
     $290 = HEAP8[$289>>0]|0;
     $291 = HEAP32[$36>>2]|0;
     $292 = $291 & $257;
     $293 = $260 >>> 24;
     $294 = $292 >>> $293;
     $295 = $263 >>> 24;
     $296 = (125924 + ($295<<2)|0);
     $297 = HEAP32[$296>>2]|0;
     $298 = (($297) + ($294)|0);
     $299 = HEAP8[$298>>0]|0;
     $dA$0$shrunk = $299;$dB$0$shrunk = $290;$dG$0$shrunk = $279;$dR$0$shrunk = $268;
     break;
    }
    default: {
     $dA$0$shrunk = 0;$dB$0$shrunk = 0;$dG$0$shrunk = 0;$dR$0$shrunk = 0;
    }
    }
    $dR$0 = $dR$0$shrunk&255;
    $dG$0 = $dG$0$shrunk&255;
    $dB$0 = $dB$0$shrunk&255;
    $dA$0 = $dA$0$shrunk&255;
    $300 = (($sR$0) - ($dR$0))|0;
    $301 = Math_imul($300, $20)|0;
    $302 = (($301>>>0) / 255)&-1;
    $303 = (($302) + ($dR$0))|0;
    $304 = (($sG$0) - ($dG$0))|0;
    $305 = Math_imul($304, $20)|0;
    $306 = (($305>>>0) / 255)&-1;
    $307 = (($306) + ($dG$0))|0;
    $308 = (($sB$0) - ($dB$0))|0;
    $309 = Math_imul($308, $20)|0;
    $310 = (($309>>>0) / 255)&-1;
    $311 = (($310) + ($dB$0))|0;
    $312 = (($dA$0) + ($20))|0;
    $313 = Math_imul($dA$0, $20)|0;
    $314 = (($313>>>0) / 255)&-1;
    $315 = (($312) - ($314))|0;
    switch ($17|0) {
    case 1:  {
     $316 = HEAP32[$29>>2]|0;
     $317 = $316 & 255;
     $318 = $303 >>> $317;
     $319 = HEAP32[$30>>2]|0;
     $320 = $319 & 255;
     $321 = $318 << $320;
     $322 = $316 >>> 8;
     $323 = $322 & 255;
     $324 = $307 >>> $323;
     $325 = $319 >>> 8;
     $326 = $325 & 255;
     $327 = $324 << $326;
     $328 = $327 | $321;
     $329 = $316 >>> 16;
     $330 = $329 & 255;
     $331 = $311 >>> $330;
     $332 = $319 >>> 16;
     $333 = $332 & 255;
     $334 = $331 << $333;
     $335 = $328 | $334;
     $336 = $316 >>> 24;
     $337 = $315 >>> $336;
     $338 = $319 >>> 24;
     $339 = $337 << $338;
     $340 = $335 | $339;
     $341 = $340&255;
     HEAP8[$dst$1>>0] = $341;
     break;
    }
    case 2:  {
     $342 = HEAP32[$29>>2]|0;
     $343 = $342 & 255;
     $344 = $303 >>> $343;
     $345 = HEAP32[$30>>2]|0;
     $346 = $345 & 255;
     $347 = $344 << $346;
     $348 = $342 >>> 8;
     $349 = $348 & 255;
     $350 = $307 >>> $349;
     $351 = $345 >>> 8;
     $352 = $351 & 255;
     $353 = $350 << $352;
     $354 = $353 | $347;
     $355 = $342 >>> 16;
     $356 = $355 & 255;
     $357 = $311 >>> $356;
     $358 = $345 >>> 16;
     $359 = $358 & 255;
     $360 = $357 << $359;
     $361 = $354 | $360;
     $362 = $342 >>> 24;
     $363 = $315 >>> $362;
     $364 = $345 >>> 24;
     $365 = $363 << $364;
     $366 = $361 | $365;
     $367 = $366&65535;
     HEAP16[$dst$1>>1] = $367;
     break;
    }
    case 3:  {
     $368 = $303&255;
     $369 = HEAP8[$30>>0]|0;
     $div18 = ($369&255) >>> 3;
     $370 = $div18&255;
     $371 = (($dst$1) + ($370)|0);
     HEAP8[$371>>0] = $368;
     $372 = $307&255;
     $373 = HEAP8[$31>>0]|0;
     $div19 = ($373&255) >>> 3;
     $374 = $div19&255;
     $375 = (($dst$1) + ($374)|0);
     HEAP8[$375>>0] = $372;
     $376 = $311&255;
     $377 = HEAP8[$32>>0]|0;
     $div20 = ($377&255) >>> 3;
     $378 = $div20&255;
     $379 = (($dst$1) + ($378)|0);
     HEAP8[$379>>0] = $376;
     break;
    }
    case 4:  {
     $380 = HEAP32[$29>>2]|0;
     $381 = $380 & 255;
     $382 = $303 >>> $381;
     $383 = HEAP32[$30>>2]|0;
     $384 = $383 & 255;
     $385 = $382 << $384;
     $386 = $380 >>> 8;
     $387 = $386 & 255;
     $388 = $307 >>> $387;
     $389 = $383 >>> 8;
     $390 = $389 & 255;
     $391 = $388 << $390;
     $392 = $391 | $385;
     $393 = $380 >>> 16;
     $394 = $393 & 255;
     $395 = $311 >>> $394;
     $396 = $383 >>> 16;
     $397 = $396 & 255;
     $398 = $395 << $397;
     $399 = $392 | $398;
     $400 = $380 >>> 24;
     $401 = $315 >>> $400;
     $402 = $383 >>> 24;
     $403 = $401 << $402;
     $404 = $399 | $403;
     HEAP32[$dst$1>>2] = $404;
     break;
    }
    default: {
    }
    }
    $405 = (($src$1) + ($14)|0);
    $406 = (($dst$1) + ($17)|0);
    $dst$2 = $406;$n$1 = $n$0;$src$2 = $405;
    label = 20;
    continue;
   }
   else if ((label|0) == 20) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $407 = HEAP8[$src$2>>0]|0;
     $408 = $407&255;
     $409 = HEAP32[$37>>2]|0;
     $410 = $408 & $409;
     $411 = HEAP32[$38>>2]|0;
     $412 = $411 & 255;
     $413 = $410 >>> $412;
     $414 = HEAP32[$39>>2]|0;
     $415 = $414 & 255;
     $416 = (125924 + ($415<<2)|0);
     $417 = HEAP32[$416>>2]|0;
     $418 = (($417) + ($413)|0);
     $419 = HEAP8[$418>>0]|0;
     $420 = HEAP32[$40>>2]|0;
     $421 = $420 & $408;
     $422 = $411 >>> 8;
     $423 = $422 & 255;
     $424 = $421 >>> $423;
     $425 = $414 >>> 8;
     $426 = $425 & 255;
     $427 = (125924 + ($426<<2)|0);
     $428 = HEAP32[$427>>2]|0;
     $429 = (($428) + ($424)|0);
     $430 = HEAP8[$429>>0]|0;
     $431 = HEAP32[$41>>2]|0;
     $432 = $431 & $408;
     $433 = $411 >>> 16;
     $434 = $433 & 255;
     $435 = $432 >>> $434;
     $436 = $414 >>> 16;
     $437 = $436 & 255;
     $438 = (125924 + ($437<<2)|0);
     $439 = HEAP32[$438>>2]|0;
     $440 = (($439) + ($435)|0);
     $441 = HEAP8[$440>>0]|0;
     $sB$1$shrunk = $441;$sG$1$shrunk = $430;$sR$1$shrunk = $419;
     break;
    }
    case 2:  {
     $442 = HEAP16[$src$2>>1]|0;
     $443 = $442&65535;
     $444 = HEAP32[$37>>2]|0;
     $445 = $443 & $444;
     $446 = HEAP32[$38>>2]|0;
     $447 = $446 & 255;
     $448 = $445 >>> $447;
     $449 = HEAP32[$39>>2]|0;
     $450 = $449 & 255;
     $451 = (125924 + ($450<<2)|0);
     $452 = HEAP32[$451>>2]|0;
     $453 = (($452) + ($448)|0);
     $454 = HEAP8[$453>>0]|0;
     $455 = HEAP32[$40>>2]|0;
     $456 = $455 & $443;
     $457 = $446 >>> 8;
     $458 = $457 & 255;
     $459 = $456 >>> $458;
     $460 = $449 >>> 8;
     $461 = $460 & 255;
     $462 = (125924 + ($461<<2)|0);
     $463 = HEAP32[$462>>2]|0;
     $464 = (($463) + ($459)|0);
     $465 = HEAP8[$464>>0]|0;
     $466 = HEAP32[$41>>2]|0;
     $467 = $466 & $443;
     $468 = $446 >>> 16;
     $469 = $468 & 255;
     $470 = $467 >>> $469;
     $471 = $449 >>> 16;
     $472 = $471 & 255;
     $473 = (125924 + ($472<<2)|0);
     $474 = HEAP32[$473>>2]|0;
     $475 = (($474) + ($470)|0);
     $476 = HEAP8[$475>>0]|0;
     $sB$1$shrunk = $476;$sG$1$shrunk = $465;$sR$1$shrunk = $454;
     break;
    }
    case 3:  {
     $477 = HEAP32[$38>>2]|0;
     $div15 = $477 >>> 3;
     $478 = $div15 & 31;
     $479 = (($src$2) + ($478)|0);
     $480 = HEAP8[$479>>0]|0;
     $481 = $477 >>> 11;
     $div16 = $481 & 31;
     $482 = (($src$2) + ($div16)|0);
     $483 = HEAP8[$482>>0]|0;
     $484 = $477 >>> 19;
     $485 = $484 & 31;
     $486 = (($src$2) + ($485)|0);
     $487 = HEAP8[$486>>0]|0;
     $sB$1$shrunk = $487;$sG$1$shrunk = $483;$sR$1$shrunk = $480;
     break;
    }
    case 4:  {
     $488 = HEAP32[$src$2>>2]|0;
     $489 = HEAP32[$37>>2]|0;
     $490 = $489 & $488;
     $491 = HEAP32[$38>>2]|0;
     $492 = $491 & 255;
     $493 = $490 >>> $492;
     $494 = HEAP32[$39>>2]|0;
     $495 = $494 & 255;
     $496 = (125924 + ($495<<2)|0);
     $497 = HEAP32[$496>>2]|0;
     $498 = (($497) + ($493)|0);
     $499 = HEAP8[$498>>0]|0;
     $500 = HEAP32[$40>>2]|0;
     $501 = $500 & $488;
     $502 = $491 >>> 8;
     $503 = $502 & 255;
     $504 = $501 >>> $503;
     $505 = $494 >>> 8;
     $506 = $505 & 255;
     $507 = (125924 + ($506<<2)|0);
     $508 = HEAP32[$507>>2]|0;
     $509 = (($508) + ($504)|0);
     $510 = HEAP8[$509>>0]|0;
     $511 = HEAP32[$41>>2]|0;
     $512 = $511 & $488;
     $513 = $491 >>> 16;
     $514 = $513 & 255;
     $515 = $512 >>> $514;
     $516 = $494 >>> 16;
     $517 = $516 & 255;
     $518 = (125924 + ($517<<2)|0);
     $519 = HEAP32[$518>>2]|0;
     $520 = (($519) + ($515)|0);
     $521 = HEAP8[$520>>0]|0;
     $sB$1$shrunk = $521;$sG$1$shrunk = $510;$sR$1$shrunk = $499;
     break;
    }
    default: {
     $sB$1$shrunk = 0;$sG$1$shrunk = 0;$sR$1$shrunk = 0;
    }
    }
    $sR$1 = $sR$1$shrunk&255;
    $sG$1 = $sG$1$shrunk&255;
    $sB$1 = $sB$1$shrunk&255;
    switch ($17|0) {
    case 1:  {
     $522 = HEAP8[$dst$2>>0]|0;
     $523 = $522&255;
     $524 = HEAP32[$33>>2]|0;
     $525 = $523 & $524;
     $526 = HEAP32[$30>>2]|0;
     $527 = $526 & 255;
     $528 = $525 >>> $527;
     $529 = HEAP32[$29>>2]|0;
     $530 = $529 & 255;
     $531 = (125924 + ($530<<2)|0);
     $532 = HEAP32[$531>>2]|0;
     $533 = (($532) + ($528)|0);
     $534 = HEAP8[$533>>0]|0;
     $535 = HEAP32[$34>>2]|0;
     $536 = $535 & $523;
     $537 = $526 >>> 8;
     $538 = $537 & 255;
     $539 = $536 >>> $538;
     $540 = $529 >>> 8;
     $541 = $540 & 255;
     $542 = (125924 + ($541<<2)|0);
     $543 = HEAP32[$542>>2]|0;
     $544 = (($543) + ($539)|0);
     $545 = HEAP8[$544>>0]|0;
     $546 = HEAP32[$35>>2]|0;
     $547 = $546 & $523;
     $548 = $526 >>> 16;
     $549 = $548 & 255;
     $550 = $547 >>> $549;
     $551 = $529 >>> 16;
     $552 = $551 & 255;
     $553 = (125924 + ($552<<2)|0);
     $554 = HEAP32[$553>>2]|0;
     $555 = (($554) + ($550)|0);
     $556 = HEAP8[$555>>0]|0;
     $557 = HEAP32[$36>>2]|0;
     $558 = $557 & $523;
     $559 = $526 >>> 24;
     $560 = $558 >>> $559;
     $561 = $529 >>> 24;
     $562 = (125924 + ($561<<2)|0);
     $563 = HEAP32[$562>>2]|0;
     $564 = (($563) + ($560)|0);
     $565 = HEAP8[$564>>0]|0;
     $dA$1$shrunk = $565;$dB$1$shrunk = $556;$dG$1$shrunk = $545;$dR$1$shrunk = $534;
     break;
    }
    case 2:  {
     $566 = HEAP16[$dst$2>>1]|0;
     $567 = $566&65535;
     $568 = HEAP32[$33>>2]|0;
     $569 = $567 & $568;
     $570 = HEAP32[$30>>2]|0;
     $571 = $570 & 255;
     $572 = $569 >>> $571;
     $573 = HEAP32[$29>>2]|0;
     $574 = $573 & 255;
     $575 = (125924 + ($574<<2)|0);
     $576 = HEAP32[$575>>2]|0;
     $577 = (($576) + ($572)|0);
     $578 = HEAP8[$577>>0]|0;
     $579 = HEAP32[$34>>2]|0;
     $580 = $579 & $567;
     $581 = $570 >>> 8;
     $582 = $581 & 255;
     $583 = $580 >>> $582;
     $584 = $573 >>> 8;
     $585 = $584 & 255;
     $586 = (125924 + ($585<<2)|0);
     $587 = HEAP32[$586>>2]|0;
     $588 = (($587) + ($583)|0);
     $589 = HEAP8[$588>>0]|0;
     $590 = HEAP32[$35>>2]|0;
     $591 = $590 & $567;
     $592 = $570 >>> 16;
     $593 = $592 & 255;
     $594 = $591 >>> $593;
     $595 = $573 >>> 16;
     $596 = $595 & 255;
     $597 = (125924 + ($596<<2)|0);
     $598 = HEAP32[$597>>2]|0;
     $599 = (($598) + ($594)|0);
     $600 = HEAP8[$599>>0]|0;
     $601 = HEAP32[$36>>2]|0;
     $602 = $601 & $567;
     $603 = $570 >>> 24;
     $604 = $602 >>> $603;
     $605 = $573 >>> 24;
     $606 = (125924 + ($605<<2)|0);
     $607 = HEAP32[$606>>2]|0;
     $608 = (($607) + ($604)|0);
     $609 = HEAP8[$608>>0]|0;
     $dA$1$shrunk = $609;$dB$1$shrunk = $600;$dG$1$shrunk = $589;$dR$1$shrunk = $578;
     break;
    }
    case 3:  {
     $610 = HEAP32[$30>>2]|0;
     $div12 = $610 >>> 3;
     $611 = $div12 & 31;
     $612 = (($dst$2) + ($611)|0);
     $613 = HEAP8[$612>>0]|0;
     $614 = $610 >>> 11;
     $div13 = $614 & 31;
     $615 = (($dst$2) + ($div13)|0);
     $616 = HEAP8[$615>>0]|0;
     $617 = $610 >>> 19;
     $618 = $617 & 31;
     $619 = (($dst$2) + ($618)|0);
     $620 = HEAP8[$619>>0]|0;
     $dA$1$shrunk = -1;$dB$1$shrunk = $620;$dG$1$shrunk = $616;$dR$1$shrunk = $613;
     break;
    }
    case 4:  {
     $621 = HEAP32[$dst$2>>2]|0;
     $622 = HEAP32[$33>>2]|0;
     $623 = $622 & $621;
     $624 = HEAP32[$30>>2]|0;
     $625 = $624 & 255;
     $626 = $623 >>> $625;
     $627 = HEAP32[$29>>2]|0;
     $628 = $627 & 255;
     $629 = (125924 + ($628<<2)|0);
     $630 = HEAP32[$629>>2]|0;
     $631 = (($630) + ($626)|0);
     $632 = HEAP8[$631>>0]|0;
     $633 = HEAP32[$34>>2]|0;
     $634 = $633 & $621;
     $635 = $624 >>> 8;
     $636 = $635 & 255;
     $637 = $634 >>> $636;
     $638 = $627 >>> 8;
     $639 = $638 & 255;
     $640 = (125924 + ($639<<2)|0);
     $641 = HEAP32[$640>>2]|0;
     $642 = (($641) + ($637)|0);
     $643 = HEAP8[$642>>0]|0;
     $644 = HEAP32[$35>>2]|0;
     $645 = $644 & $621;
     $646 = $624 >>> 16;
     $647 = $646 & 255;
     $648 = $645 >>> $647;
     $649 = $627 >>> 16;
     $650 = $649 & 255;
     $651 = (125924 + ($650<<2)|0);
     $652 = HEAP32[$651>>2]|0;
     $653 = (($652) + ($648)|0);
     $654 = HEAP8[$653>>0]|0;
     $655 = HEAP32[$36>>2]|0;
     $656 = $655 & $621;
     $657 = $624 >>> 24;
     $658 = $656 >>> $657;
     $659 = $627 >>> 24;
     $660 = (125924 + ($659<<2)|0);
     $661 = HEAP32[$660>>2]|0;
     $662 = (($661) + ($658)|0);
     $663 = HEAP8[$662>>0]|0;
     $dA$1$shrunk = $663;$dB$1$shrunk = $654;$dG$1$shrunk = $643;$dR$1$shrunk = $632;
     break;
    }
    default: {
     $dA$1$shrunk = 0;$dB$1$shrunk = 0;$dG$1$shrunk = 0;$dR$1$shrunk = 0;
    }
    }
    $dR$1 = $dR$1$shrunk&255;
    $dG$1 = $dG$1$shrunk&255;
    $dB$1 = $dB$1$shrunk&255;
    $dA$1 = $dA$1$shrunk&255;
    $664 = (($sR$1) - ($dR$1))|0;
    $665 = Math_imul($664, $20)|0;
    $666 = (($665>>>0) / 255)&-1;
    $667 = (($666) + ($dR$1))|0;
    $668 = (($sG$1) - ($dG$1))|0;
    $669 = Math_imul($668, $20)|0;
    $670 = (($669>>>0) / 255)&-1;
    $671 = (($670) + ($dG$1))|0;
    $672 = (($sB$1) - ($dB$1))|0;
    $673 = Math_imul($672, $20)|0;
    $674 = (($673>>>0) / 255)&-1;
    $675 = (($674) + ($dB$1))|0;
    $676 = (($dA$1) + ($20))|0;
    $677 = Math_imul($dA$1, $20)|0;
    $678 = (($677>>>0) / 255)&-1;
    $679 = (($676) - ($678))|0;
    switch ($17|0) {
    case 1:  {
     $680 = HEAP32[$29>>2]|0;
     $681 = $680 & 255;
     $682 = $667 >>> $681;
     $683 = HEAP32[$30>>2]|0;
     $684 = $683 & 255;
     $685 = $682 << $684;
     $686 = $680 >>> 8;
     $687 = $686 & 255;
     $688 = $671 >>> $687;
     $689 = $683 >>> 8;
     $690 = $689 & 255;
     $691 = $688 << $690;
     $692 = $691 | $685;
     $693 = $680 >>> 16;
     $694 = $693 & 255;
     $695 = $675 >>> $694;
     $696 = $683 >>> 16;
     $697 = $696 & 255;
     $698 = $695 << $697;
     $699 = $692 | $698;
     $700 = $680 >>> 24;
     $701 = $679 >>> $700;
     $702 = $683 >>> 24;
     $703 = $701 << $702;
     $704 = $699 | $703;
     $705 = $704&255;
     HEAP8[$dst$2>>0] = $705;
     break;
    }
    case 2:  {
     $706 = HEAP32[$29>>2]|0;
     $707 = $706 & 255;
     $708 = $667 >>> $707;
     $709 = HEAP32[$30>>2]|0;
     $710 = $709 & 255;
     $711 = $708 << $710;
     $712 = $706 >>> 8;
     $713 = $712 & 255;
     $714 = $671 >>> $713;
     $715 = $709 >>> 8;
     $716 = $715 & 255;
     $717 = $714 << $716;
     $718 = $717 | $711;
     $719 = $706 >>> 16;
     $720 = $719 & 255;
     $721 = $675 >>> $720;
     $722 = $709 >>> 16;
     $723 = $722 & 255;
     $724 = $721 << $723;
     $725 = $718 | $724;
     $726 = $706 >>> 24;
     $727 = $679 >>> $726;
     $728 = $709 >>> 24;
     $729 = $727 << $728;
     $730 = $725 | $729;
     $731 = $730&65535;
     HEAP16[$dst$2>>1] = $731;
     break;
    }
    case 3:  {
     $732 = $667&255;
     $733 = HEAP8[$30>>0]|0;
     $div9 = ($733&255) >>> 3;
     $734 = $div9&255;
     $735 = (($dst$2) + ($734)|0);
     HEAP8[$735>>0] = $732;
     $736 = $671&255;
     $737 = HEAP8[$31>>0]|0;
     $div10 = ($737&255) >>> 3;
     $738 = $div10&255;
     $739 = (($dst$2) + ($738)|0);
     HEAP8[$739>>0] = $736;
     $740 = $675&255;
     $741 = HEAP8[$32>>0]|0;
     $div11 = ($741&255) >>> 3;
     $742 = $div11&255;
     $743 = (($dst$2) + ($742)|0);
     HEAP8[$743>>0] = $740;
     break;
    }
    case 4:  {
     $744 = HEAP32[$29>>2]|0;
     $745 = $744 & 255;
     $746 = $667 >>> $745;
     $747 = HEAP32[$30>>2]|0;
     $748 = $747 & 255;
     $749 = $746 << $748;
     $750 = $744 >>> 8;
     $751 = $750 & 255;
     $752 = $671 >>> $751;
     $753 = $747 >>> 8;
     $754 = $753 & 255;
     $755 = $752 << $754;
     $756 = $755 | $749;
     $757 = $744 >>> 16;
     $758 = $757 & 255;
     $759 = $675 >>> $758;
     $760 = $747 >>> 16;
     $761 = $760 & 255;
     $762 = $759 << $761;
     $763 = $756 | $762;
     $764 = $744 >>> 24;
     $765 = $679 >>> $764;
     $766 = $747 >>> 24;
     $767 = $765 << $766;
     $768 = $763 | $767;
     HEAP32[$dst$2>>2] = $768;
     break;
    }
    default: {
    }
    }
    $769 = (($src$2) + ($14)|0);
    $770 = (($dst$2) + ($17)|0);
    $dst$3 = $770;$n$2 = $n$1;$src$3 = $769;
    label = 36;
    continue;
   }
   else if ((label|0) == 36) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $771 = HEAP8[$src$3>>0]|0;
     $772 = $771&255;
     $773 = HEAP32[$37>>2]|0;
     $774 = $772 & $773;
     $775 = HEAP32[$38>>2]|0;
     $776 = $775 & 255;
     $777 = $774 >>> $776;
     $778 = HEAP32[$39>>2]|0;
     $779 = $778 & 255;
     $780 = (125924 + ($779<<2)|0);
     $781 = HEAP32[$780>>2]|0;
     $782 = (($781) + ($777)|0);
     $783 = HEAP8[$782>>0]|0;
     $784 = HEAP32[$40>>2]|0;
     $785 = $784 & $772;
     $786 = $775 >>> 8;
     $787 = $786 & 255;
     $788 = $785 >>> $787;
     $789 = $778 >>> 8;
     $790 = $789 & 255;
     $791 = (125924 + ($790<<2)|0);
     $792 = HEAP32[$791>>2]|0;
     $793 = (($792) + ($788)|0);
     $794 = HEAP8[$793>>0]|0;
     $795 = HEAP32[$41>>2]|0;
     $796 = $795 & $772;
     $797 = $775 >>> 16;
     $798 = $797 & 255;
     $799 = $796 >>> $798;
     $800 = $778 >>> 16;
     $801 = $800 & 255;
     $802 = (125924 + ($801<<2)|0);
     $803 = HEAP32[$802>>2]|0;
     $804 = (($803) + ($799)|0);
     $805 = HEAP8[$804>>0]|0;
     $sB$2$shrunk = $805;$sG$2$shrunk = $794;$sR$2$shrunk = $783;
     break;
    }
    case 2:  {
     $806 = HEAP16[$src$3>>1]|0;
     $807 = $806&65535;
     $808 = HEAP32[$37>>2]|0;
     $809 = $807 & $808;
     $810 = HEAP32[$38>>2]|0;
     $811 = $810 & 255;
     $812 = $809 >>> $811;
     $813 = HEAP32[$39>>2]|0;
     $814 = $813 & 255;
     $815 = (125924 + ($814<<2)|0);
     $816 = HEAP32[$815>>2]|0;
     $817 = (($816) + ($812)|0);
     $818 = HEAP8[$817>>0]|0;
     $819 = HEAP32[$40>>2]|0;
     $820 = $819 & $807;
     $821 = $810 >>> 8;
     $822 = $821 & 255;
     $823 = $820 >>> $822;
     $824 = $813 >>> 8;
     $825 = $824 & 255;
     $826 = (125924 + ($825<<2)|0);
     $827 = HEAP32[$826>>2]|0;
     $828 = (($827) + ($823)|0);
     $829 = HEAP8[$828>>0]|0;
     $830 = HEAP32[$41>>2]|0;
     $831 = $830 & $807;
     $832 = $810 >>> 16;
     $833 = $832 & 255;
     $834 = $831 >>> $833;
     $835 = $813 >>> 16;
     $836 = $835 & 255;
     $837 = (125924 + ($836<<2)|0);
     $838 = HEAP32[$837>>2]|0;
     $839 = (($838) + ($834)|0);
     $840 = HEAP8[$839>>0]|0;
     $sB$2$shrunk = $840;$sG$2$shrunk = $829;$sR$2$shrunk = $818;
     break;
    }
    case 3:  {
     $841 = HEAP32[$38>>2]|0;
     $div6 = $841 >>> 3;
     $842 = $div6 & 31;
     $843 = (($src$3) + ($842)|0);
     $844 = HEAP8[$843>>0]|0;
     $845 = $841 >>> 11;
     $div7 = $845 & 31;
     $846 = (($src$3) + ($div7)|0);
     $847 = HEAP8[$846>>0]|0;
     $848 = $841 >>> 19;
     $849 = $848 & 31;
     $850 = (($src$3) + ($849)|0);
     $851 = HEAP8[$850>>0]|0;
     $sB$2$shrunk = $851;$sG$2$shrunk = $847;$sR$2$shrunk = $844;
     break;
    }
    case 4:  {
     $852 = HEAP32[$src$3>>2]|0;
     $853 = HEAP32[$37>>2]|0;
     $854 = $853 & $852;
     $855 = HEAP32[$38>>2]|0;
     $856 = $855 & 255;
     $857 = $854 >>> $856;
     $858 = HEAP32[$39>>2]|0;
     $859 = $858 & 255;
     $860 = (125924 + ($859<<2)|0);
     $861 = HEAP32[$860>>2]|0;
     $862 = (($861) + ($857)|0);
     $863 = HEAP8[$862>>0]|0;
     $864 = HEAP32[$40>>2]|0;
     $865 = $864 & $852;
     $866 = $855 >>> 8;
     $867 = $866 & 255;
     $868 = $865 >>> $867;
     $869 = $858 >>> 8;
     $870 = $869 & 255;
     $871 = (125924 + ($870<<2)|0);
     $872 = HEAP32[$871>>2]|0;
     $873 = (($872) + ($868)|0);
     $874 = HEAP8[$873>>0]|0;
     $875 = HEAP32[$41>>2]|0;
     $876 = $875 & $852;
     $877 = $855 >>> 16;
     $878 = $877 & 255;
     $879 = $876 >>> $878;
     $880 = $858 >>> 16;
     $881 = $880 & 255;
     $882 = (125924 + ($881<<2)|0);
     $883 = HEAP32[$882>>2]|0;
     $884 = (($883) + ($879)|0);
     $885 = HEAP8[$884>>0]|0;
     $sB$2$shrunk = $885;$sG$2$shrunk = $874;$sR$2$shrunk = $863;
     break;
    }
    default: {
     $sB$2$shrunk = 0;$sG$2$shrunk = 0;$sR$2$shrunk = 0;
    }
    }
    $sR$2 = $sR$2$shrunk&255;
    $sG$2 = $sG$2$shrunk&255;
    $sB$2 = $sB$2$shrunk&255;
    switch ($17|0) {
    case 1:  {
     $886 = HEAP8[$dst$3>>0]|0;
     $887 = $886&255;
     $888 = HEAP32[$33>>2]|0;
     $889 = $887 & $888;
     $890 = HEAP32[$30>>2]|0;
     $891 = $890 & 255;
     $892 = $889 >>> $891;
     $893 = HEAP32[$29>>2]|0;
     $894 = $893 & 255;
     $895 = (125924 + ($894<<2)|0);
     $896 = HEAP32[$895>>2]|0;
     $897 = (($896) + ($892)|0);
     $898 = HEAP8[$897>>0]|0;
     $899 = HEAP32[$34>>2]|0;
     $900 = $899 & $887;
     $901 = $890 >>> 8;
     $902 = $901 & 255;
     $903 = $900 >>> $902;
     $904 = $893 >>> 8;
     $905 = $904 & 255;
     $906 = (125924 + ($905<<2)|0);
     $907 = HEAP32[$906>>2]|0;
     $908 = (($907) + ($903)|0);
     $909 = HEAP8[$908>>0]|0;
     $910 = HEAP32[$35>>2]|0;
     $911 = $910 & $887;
     $912 = $890 >>> 16;
     $913 = $912 & 255;
     $914 = $911 >>> $913;
     $915 = $893 >>> 16;
     $916 = $915 & 255;
     $917 = (125924 + ($916<<2)|0);
     $918 = HEAP32[$917>>2]|0;
     $919 = (($918) + ($914)|0);
     $920 = HEAP8[$919>>0]|0;
     $921 = HEAP32[$36>>2]|0;
     $922 = $921 & $887;
     $923 = $890 >>> 24;
     $924 = $922 >>> $923;
     $925 = $893 >>> 24;
     $926 = (125924 + ($925<<2)|0);
     $927 = HEAP32[$926>>2]|0;
     $928 = (($927) + ($924)|0);
     $929 = HEAP8[$928>>0]|0;
     $dA$2$shrunk = $929;$dB$2$shrunk = $920;$dG$2$shrunk = $909;$dR$2$shrunk = $898;
     break;
    }
    case 2:  {
     $930 = HEAP16[$dst$3>>1]|0;
     $931 = $930&65535;
     $932 = HEAP32[$33>>2]|0;
     $933 = $931 & $932;
     $934 = HEAP32[$30>>2]|0;
     $935 = $934 & 255;
     $936 = $933 >>> $935;
     $937 = HEAP32[$29>>2]|0;
     $938 = $937 & 255;
     $939 = (125924 + ($938<<2)|0);
     $940 = HEAP32[$939>>2]|0;
     $941 = (($940) + ($936)|0);
     $942 = HEAP8[$941>>0]|0;
     $943 = HEAP32[$34>>2]|0;
     $944 = $943 & $931;
     $945 = $934 >>> 8;
     $946 = $945 & 255;
     $947 = $944 >>> $946;
     $948 = $937 >>> 8;
     $949 = $948 & 255;
     $950 = (125924 + ($949<<2)|0);
     $951 = HEAP32[$950>>2]|0;
     $952 = (($951) + ($947)|0);
     $953 = HEAP8[$952>>0]|0;
     $954 = HEAP32[$35>>2]|0;
     $955 = $954 & $931;
     $956 = $934 >>> 16;
     $957 = $956 & 255;
     $958 = $955 >>> $957;
     $959 = $937 >>> 16;
     $960 = $959 & 255;
     $961 = (125924 + ($960<<2)|0);
     $962 = HEAP32[$961>>2]|0;
     $963 = (($962) + ($958)|0);
     $964 = HEAP8[$963>>0]|0;
     $965 = HEAP32[$36>>2]|0;
     $966 = $965 & $931;
     $967 = $934 >>> 24;
     $968 = $966 >>> $967;
     $969 = $937 >>> 24;
     $970 = (125924 + ($969<<2)|0);
     $971 = HEAP32[$970>>2]|0;
     $972 = (($971) + ($968)|0);
     $973 = HEAP8[$972>>0]|0;
     $dA$2$shrunk = $973;$dB$2$shrunk = $964;$dG$2$shrunk = $953;$dR$2$shrunk = $942;
     break;
    }
    case 3:  {
     $974 = HEAP32[$30>>2]|0;
     $div3 = $974 >>> 3;
     $975 = $div3 & 31;
     $976 = (($dst$3) + ($975)|0);
     $977 = HEAP8[$976>>0]|0;
     $978 = $974 >>> 11;
     $div4 = $978 & 31;
     $979 = (($dst$3) + ($div4)|0);
     $980 = HEAP8[$979>>0]|0;
     $981 = $974 >>> 19;
     $982 = $981 & 31;
     $983 = (($dst$3) + ($982)|0);
     $984 = HEAP8[$983>>0]|0;
     $dA$2$shrunk = -1;$dB$2$shrunk = $984;$dG$2$shrunk = $980;$dR$2$shrunk = $977;
     break;
    }
    case 4:  {
     $985 = HEAP32[$dst$3>>2]|0;
     $986 = HEAP32[$33>>2]|0;
     $987 = $986 & $985;
     $988 = HEAP32[$30>>2]|0;
     $989 = $988 & 255;
     $990 = $987 >>> $989;
     $991 = HEAP32[$29>>2]|0;
     $992 = $991 & 255;
     $993 = (125924 + ($992<<2)|0);
     $994 = HEAP32[$993>>2]|0;
     $995 = (($994) + ($990)|0);
     $996 = HEAP8[$995>>0]|0;
     $997 = HEAP32[$34>>2]|0;
     $998 = $997 & $985;
     $999 = $988 >>> 8;
     $1000 = $999 & 255;
     $1001 = $998 >>> $1000;
     $1002 = $991 >>> 8;
     $1003 = $1002 & 255;
     $1004 = (125924 + ($1003<<2)|0);
     $1005 = HEAP32[$1004>>2]|0;
     $1006 = (($1005) + ($1001)|0);
     $1007 = HEAP8[$1006>>0]|0;
     $1008 = HEAP32[$35>>2]|0;
     $1009 = $1008 & $985;
     $1010 = $988 >>> 16;
     $1011 = $1010 & 255;
     $1012 = $1009 >>> $1011;
     $1013 = $991 >>> 16;
     $1014 = $1013 & 255;
     $1015 = (125924 + ($1014<<2)|0);
     $1016 = HEAP32[$1015>>2]|0;
     $1017 = (($1016) + ($1012)|0);
     $1018 = HEAP8[$1017>>0]|0;
     $1019 = HEAP32[$36>>2]|0;
     $1020 = $1019 & $985;
     $1021 = $988 >>> 24;
     $1022 = $1020 >>> $1021;
     $1023 = $991 >>> 24;
     $1024 = (125924 + ($1023<<2)|0);
     $1025 = HEAP32[$1024>>2]|0;
     $1026 = (($1025) + ($1022)|0);
     $1027 = HEAP8[$1026>>0]|0;
     $dA$2$shrunk = $1027;$dB$2$shrunk = $1018;$dG$2$shrunk = $1007;$dR$2$shrunk = $996;
     break;
    }
    default: {
     $dA$2$shrunk = 0;$dB$2$shrunk = 0;$dG$2$shrunk = 0;$dR$2$shrunk = 0;
    }
    }
    $dR$2 = $dR$2$shrunk&255;
    $dG$2 = $dG$2$shrunk&255;
    $dB$2 = $dB$2$shrunk&255;
    $dA$2 = $dA$2$shrunk&255;
    $1028 = (($sR$2) - ($dR$2))|0;
    $1029 = Math_imul($1028, $20)|0;
    $1030 = (($1029>>>0) / 255)&-1;
    $1031 = (($1030) + ($dR$2))|0;
    $1032 = (($sG$2) - ($dG$2))|0;
    $1033 = Math_imul($1032, $20)|0;
    $1034 = (($1033>>>0) / 255)&-1;
    $1035 = (($1034) + ($dG$2))|0;
    $1036 = (($sB$2) - ($dB$2))|0;
    $1037 = Math_imul($1036, $20)|0;
    $1038 = (($1037>>>0) / 255)&-1;
    $1039 = (($1038) + ($dB$2))|0;
    $1040 = (($dA$2) + ($20))|0;
    $1041 = Math_imul($dA$2, $20)|0;
    $1042 = (($1041>>>0) / 255)&-1;
    $1043 = (($1040) - ($1042))|0;
    switch ($17|0) {
    case 1:  {
     $1044 = HEAP32[$29>>2]|0;
     $1045 = $1044 & 255;
     $1046 = $1031 >>> $1045;
     $1047 = HEAP32[$30>>2]|0;
     $1048 = $1047 & 255;
     $1049 = $1046 << $1048;
     $1050 = $1044 >>> 8;
     $1051 = $1050 & 255;
     $1052 = $1035 >>> $1051;
     $1053 = $1047 >>> 8;
     $1054 = $1053 & 255;
     $1055 = $1052 << $1054;
     $1056 = $1055 | $1049;
     $1057 = $1044 >>> 16;
     $1058 = $1057 & 255;
     $1059 = $1039 >>> $1058;
     $1060 = $1047 >>> 16;
     $1061 = $1060 & 255;
     $1062 = $1059 << $1061;
     $1063 = $1056 | $1062;
     $1064 = $1044 >>> 24;
     $1065 = $1043 >>> $1064;
     $1066 = $1047 >>> 24;
     $1067 = $1065 << $1066;
     $1068 = $1063 | $1067;
     $1069 = $1068&255;
     HEAP8[$dst$3>>0] = $1069;
     break;
    }
    case 2:  {
     $1070 = HEAP32[$29>>2]|0;
     $1071 = $1070 & 255;
     $1072 = $1031 >>> $1071;
     $1073 = HEAP32[$30>>2]|0;
     $1074 = $1073 & 255;
     $1075 = $1072 << $1074;
     $1076 = $1070 >>> 8;
     $1077 = $1076 & 255;
     $1078 = $1035 >>> $1077;
     $1079 = $1073 >>> 8;
     $1080 = $1079 & 255;
     $1081 = $1078 << $1080;
     $1082 = $1081 | $1075;
     $1083 = $1070 >>> 16;
     $1084 = $1083 & 255;
     $1085 = $1039 >>> $1084;
     $1086 = $1073 >>> 16;
     $1087 = $1086 & 255;
     $1088 = $1085 << $1087;
     $1089 = $1082 | $1088;
     $1090 = $1070 >>> 24;
     $1091 = $1043 >>> $1090;
     $1092 = $1073 >>> 24;
     $1093 = $1091 << $1092;
     $1094 = $1089 | $1093;
     $1095 = $1094&65535;
     HEAP16[$dst$3>>1] = $1095;
     break;
    }
    case 3:  {
     $1096 = $1031&255;
     $1097 = HEAP8[$30>>0]|0;
     $div = ($1097&255) >>> 3;
     $1098 = $div&255;
     $1099 = (($dst$3) + ($1098)|0);
     HEAP8[$1099>>0] = $1096;
     $1100 = $1035&255;
     $1101 = HEAP8[$31>>0]|0;
     $div1 = ($1101&255) >>> 3;
     $1102 = $div1&255;
     $1103 = (($dst$3) + ($1102)|0);
     HEAP8[$1103>>0] = $1100;
     $1104 = $1039&255;
     $1105 = HEAP8[$32>>0]|0;
     $div2 = ($1105&255) >>> 3;
     $1106 = $div2&255;
     $1107 = (($dst$3) + ($1106)|0);
     HEAP8[$1107>>0] = $1104;
     break;
    }
    case 4:  {
     $1108 = HEAP32[$29>>2]|0;
     $1109 = $1108 & 255;
     $1110 = $1031 >>> $1109;
     $1111 = HEAP32[$30>>2]|0;
     $1112 = $1111 & 255;
     $1113 = $1110 << $1112;
     $1114 = $1108 >>> 8;
     $1115 = $1114 & 255;
     $1116 = $1035 >>> $1115;
     $1117 = $1111 >>> 8;
     $1118 = $1117 & 255;
     $1119 = $1116 << $1118;
     $1120 = $1119 | $1113;
     $1121 = $1108 >>> 16;
     $1122 = $1121 & 255;
     $1123 = $1039 >>> $1122;
     $1124 = $1111 >>> 16;
     $1125 = $1124 & 255;
     $1126 = $1123 << $1125;
     $1127 = $1120 | $1126;
     $1128 = $1108 >>> 24;
     $1129 = $1043 >>> $1128;
     $1130 = $1111 >>> 24;
     $1131 = $1129 << $1130;
     $1132 = $1127 | $1131;
     HEAP32[$dst$3>>2] = $1132;
     break;
    }
    default: {
    }
    }
    $1133 = (($src$3) + ($14)|0);
    $1134 = (($dst$3) + ($17)|0);
    $dst$4 = $1134;$n$3 = $n$2;$src$4 = $1133;
    label = 52;
    continue;
   }
   else if ((label|0) == 52) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $1135 = HEAP8[$src$4>>0]|0;
     $1136 = $1135&255;
     $1137 = HEAP32[$37>>2]|0;
     $1138 = $1136 & $1137;
     $1139 = HEAP32[$38>>2]|0;
     $1140 = $1139 & 255;
     $1141 = $1138 >>> $1140;
     $1142 = HEAP32[$39>>2]|0;
     $1143 = $1142 & 255;
     $1144 = (125924 + ($1143<<2)|0);
     $1145 = HEAP32[$1144>>2]|0;
     $1146 = (($1145) + ($1141)|0);
     $1147 = HEAP8[$1146>>0]|0;
     $1148 = HEAP32[$40>>2]|0;
     $1149 = $1148 & $1136;
     $1150 = $1139 >>> 8;
     $1151 = $1150 & 255;
     $1152 = $1149 >>> $1151;
     $1153 = $1142 >>> 8;
     $1154 = $1153 & 255;
     $1155 = (125924 + ($1154<<2)|0);
     $1156 = HEAP32[$1155>>2]|0;
     $1157 = (($1156) + ($1152)|0);
     $1158 = HEAP8[$1157>>0]|0;
     $1159 = HEAP32[$41>>2]|0;
     $1160 = $1159 & $1136;
     $1161 = $1139 >>> 16;
     $1162 = $1161 & 255;
     $1163 = $1160 >>> $1162;
     $1164 = $1142 >>> 16;
     $1165 = $1164 & 255;
     $1166 = (125924 + ($1165<<2)|0);
     $1167 = HEAP32[$1166>>2]|0;
     $1168 = (($1167) + ($1163)|0);
     $1169 = HEAP8[$1168>>0]|0;
     $sB$3$shrunk = $1169;$sG$3$shrunk = $1158;$sR$3$shrunk = $1147;
     break;
    }
    case 2:  {
     $1170 = HEAP16[$src$4>>1]|0;
     $1171 = $1170&65535;
     $1172 = HEAP32[$37>>2]|0;
     $1173 = $1171 & $1172;
     $1174 = HEAP32[$38>>2]|0;
     $1175 = $1174 & 255;
     $1176 = $1173 >>> $1175;
     $1177 = HEAP32[$39>>2]|0;
     $1178 = $1177 & 255;
     $1179 = (125924 + ($1178<<2)|0);
     $1180 = HEAP32[$1179>>2]|0;
     $1181 = (($1180) + ($1176)|0);
     $1182 = HEAP8[$1181>>0]|0;
     $1183 = HEAP32[$40>>2]|0;
     $1184 = $1183 & $1171;
     $1185 = $1174 >>> 8;
     $1186 = $1185 & 255;
     $1187 = $1184 >>> $1186;
     $1188 = $1177 >>> 8;
     $1189 = $1188 & 255;
     $1190 = (125924 + ($1189<<2)|0);
     $1191 = HEAP32[$1190>>2]|0;
     $1192 = (($1191) + ($1187)|0);
     $1193 = HEAP8[$1192>>0]|0;
     $1194 = HEAP32[$41>>2]|0;
     $1195 = $1194 & $1171;
     $1196 = $1174 >>> 16;
     $1197 = $1196 & 255;
     $1198 = $1195 >>> $1197;
     $1199 = $1177 >>> 16;
     $1200 = $1199 & 255;
     $1201 = (125924 + ($1200<<2)|0);
     $1202 = HEAP32[$1201>>2]|0;
     $1203 = (($1202) + ($1198)|0);
     $1204 = HEAP8[$1203>>0]|0;
     $sB$3$shrunk = $1204;$sG$3$shrunk = $1193;$sR$3$shrunk = $1182;
     break;
    }
    case 3:  {
     $1205 = HEAP32[$38>>2]|0;
     $div33 = $1205 >>> 3;
     $1206 = $div33 & 31;
     $1207 = (($src$4) + ($1206)|0);
     $1208 = HEAP8[$1207>>0]|0;
     $1209 = $1205 >>> 11;
     $div34 = $1209 & 31;
     $1210 = (($src$4) + ($div34)|0);
     $1211 = HEAP8[$1210>>0]|0;
     $1212 = $1205 >>> 19;
     $1213 = $1212 & 31;
     $1214 = (($src$4) + ($1213)|0);
     $1215 = HEAP8[$1214>>0]|0;
     $sB$3$shrunk = $1215;$sG$3$shrunk = $1211;$sR$3$shrunk = $1208;
     break;
    }
    case 4:  {
     $1216 = HEAP32[$src$4>>2]|0;
     $1217 = HEAP32[$37>>2]|0;
     $1218 = $1217 & $1216;
     $1219 = HEAP32[$38>>2]|0;
     $1220 = $1219 & 255;
     $1221 = $1218 >>> $1220;
     $1222 = HEAP32[$39>>2]|0;
     $1223 = $1222 & 255;
     $1224 = (125924 + ($1223<<2)|0);
     $1225 = HEAP32[$1224>>2]|0;
     $1226 = (($1225) + ($1221)|0);
     $1227 = HEAP8[$1226>>0]|0;
     $1228 = HEAP32[$40>>2]|0;
     $1229 = $1228 & $1216;
     $1230 = $1219 >>> 8;
     $1231 = $1230 & 255;
     $1232 = $1229 >>> $1231;
     $1233 = $1222 >>> 8;
     $1234 = $1233 & 255;
     $1235 = (125924 + ($1234<<2)|0);
     $1236 = HEAP32[$1235>>2]|0;
     $1237 = (($1236) + ($1232)|0);
     $1238 = HEAP8[$1237>>0]|0;
     $1239 = HEAP32[$41>>2]|0;
     $1240 = $1239 & $1216;
     $1241 = $1219 >>> 16;
     $1242 = $1241 & 255;
     $1243 = $1240 >>> $1242;
     $1244 = $1222 >>> 16;
     $1245 = $1244 & 255;
     $1246 = (125924 + ($1245<<2)|0);
     $1247 = HEAP32[$1246>>2]|0;
     $1248 = (($1247) + ($1243)|0);
     $1249 = HEAP8[$1248>>0]|0;
     $sB$3$shrunk = $1249;$sG$3$shrunk = $1238;$sR$3$shrunk = $1227;
     break;
    }
    default: {
     $sB$3$shrunk = 0;$sG$3$shrunk = 0;$sR$3$shrunk = 0;
    }
    }
    $sR$3 = $sR$3$shrunk&255;
    $sG$3 = $sG$3$shrunk&255;
    $sB$3 = $sB$3$shrunk&255;
    switch ($17|0) {
    case 1:  {
     $1250 = HEAP8[$dst$4>>0]|0;
     $1251 = $1250&255;
     $1252 = HEAP32[$33>>2]|0;
     $1253 = $1251 & $1252;
     $1254 = HEAP32[$30>>2]|0;
     $1255 = $1254 & 255;
     $1256 = $1253 >>> $1255;
     $1257 = HEAP32[$29>>2]|0;
     $1258 = $1257 & 255;
     $1259 = (125924 + ($1258<<2)|0);
     $1260 = HEAP32[$1259>>2]|0;
     $1261 = (($1260) + ($1256)|0);
     $1262 = HEAP8[$1261>>0]|0;
     $1263 = HEAP32[$34>>2]|0;
     $1264 = $1263 & $1251;
     $1265 = $1254 >>> 8;
     $1266 = $1265 & 255;
     $1267 = $1264 >>> $1266;
     $1268 = $1257 >>> 8;
     $1269 = $1268 & 255;
     $1270 = (125924 + ($1269<<2)|0);
     $1271 = HEAP32[$1270>>2]|0;
     $1272 = (($1271) + ($1267)|0);
     $1273 = HEAP8[$1272>>0]|0;
     $1274 = HEAP32[$35>>2]|0;
     $1275 = $1274 & $1251;
     $1276 = $1254 >>> 16;
     $1277 = $1276 & 255;
     $1278 = $1275 >>> $1277;
     $1279 = $1257 >>> 16;
     $1280 = $1279 & 255;
     $1281 = (125924 + ($1280<<2)|0);
     $1282 = HEAP32[$1281>>2]|0;
     $1283 = (($1282) + ($1278)|0);
     $1284 = HEAP8[$1283>>0]|0;
     $1285 = HEAP32[$36>>2]|0;
     $1286 = $1285 & $1251;
     $1287 = $1254 >>> 24;
     $1288 = $1286 >>> $1287;
     $1289 = $1257 >>> 24;
     $1290 = (125924 + ($1289<<2)|0);
     $1291 = HEAP32[$1290>>2]|0;
     $1292 = (($1291) + ($1288)|0);
     $1293 = HEAP8[$1292>>0]|0;
     $dA$3$shrunk = $1293;$dB$3$shrunk = $1284;$dG$3$shrunk = $1273;$dR$3$shrunk = $1262;
     break;
    }
    case 2:  {
     $1294 = HEAP16[$dst$4>>1]|0;
     $1295 = $1294&65535;
     $1296 = HEAP32[$33>>2]|0;
     $1297 = $1295 & $1296;
     $1298 = HEAP32[$30>>2]|0;
     $1299 = $1298 & 255;
     $1300 = $1297 >>> $1299;
     $1301 = HEAP32[$29>>2]|0;
     $1302 = $1301 & 255;
     $1303 = (125924 + ($1302<<2)|0);
     $1304 = HEAP32[$1303>>2]|0;
     $1305 = (($1304) + ($1300)|0);
     $1306 = HEAP8[$1305>>0]|0;
     $1307 = HEAP32[$34>>2]|0;
     $1308 = $1307 & $1295;
     $1309 = $1298 >>> 8;
     $1310 = $1309 & 255;
     $1311 = $1308 >>> $1310;
     $1312 = $1301 >>> 8;
     $1313 = $1312 & 255;
     $1314 = (125924 + ($1313<<2)|0);
     $1315 = HEAP32[$1314>>2]|0;
     $1316 = (($1315) + ($1311)|0);
     $1317 = HEAP8[$1316>>0]|0;
     $1318 = HEAP32[$35>>2]|0;
     $1319 = $1318 & $1295;
     $1320 = $1298 >>> 16;
     $1321 = $1320 & 255;
     $1322 = $1319 >>> $1321;
     $1323 = $1301 >>> 16;
     $1324 = $1323 & 255;
     $1325 = (125924 + ($1324<<2)|0);
     $1326 = HEAP32[$1325>>2]|0;
     $1327 = (($1326) + ($1322)|0);
     $1328 = HEAP8[$1327>>0]|0;
     $1329 = HEAP32[$36>>2]|0;
     $1330 = $1329 & $1295;
     $1331 = $1298 >>> 24;
     $1332 = $1330 >>> $1331;
     $1333 = $1301 >>> 24;
     $1334 = (125924 + ($1333<<2)|0);
     $1335 = HEAP32[$1334>>2]|0;
     $1336 = (($1335) + ($1332)|0);
     $1337 = HEAP8[$1336>>0]|0;
     $dA$3$shrunk = $1337;$dB$3$shrunk = $1328;$dG$3$shrunk = $1317;$dR$3$shrunk = $1306;
     break;
    }
    case 3:  {
     $1338 = HEAP32[$30>>2]|0;
     $div30 = $1338 >>> 3;
     $1339 = $div30 & 31;
     $1340 = (($dst$4) + ($1339)|0);
     $1341 = HEAP8[$1340>>0]|0;
     $1342 = $1338 >>> 11;
     $div31 = $1342 & 31;
     $1343 = (($dst$4) + ($div31)|0);
     $1344 = HEAP8[$1343>>0]|0;
     $1345 = $1338 >>> 19;
     $1346 = $1345 & 31;
     $1347 = (($dst$4) + ($1346)|0);
     $1348 = HEAP8[$1347>>0]|0;
     $dA$3$shrunk = -1;$dB$3$shrunk = $1348;$dG$3$shrunk = $1344;$dR$3$shrunk = $1341;
     break;
    }
    case 4:  {
     $1349 = HEAP32[$dst$4>>2]|0;
     $1350 = HEAP32[$33>>2]|0;
     $1351 = $1350 & $1349;
     $1352 = HEAP32[$30>>2]|0;
     $1353 = $1352 & 255;
     $1354 = $1351 >>> $1353;
     $1355 = HEAP32[$29>>2]|0;
     $1356 = $1355 & 255;
     $1357 = (125924 + ($1356<<2)|0);
     $1358 = HEAP32[$1357>>2]|0;
     $1359 = (($1358) + ($1354)|0);
     $1360 = HEAP8[$1359>>0]|0;
     $1361 = HEAP32[$34>>2]|0;
     $1362 = $1361 & $1349;
     $1363 = $1352 >>> 8;
     $1364 = $1363 & 255;
     $1365 = $1362 >>> $1364;
     $1366 = $1355 >>> 8;
     $1367 = $1366 & 255;
     $1368 = (125924 + ($1367<<2)|0);
     $1369 = HEAP32[$1368>>2]|0;
     $1370 = (($1369) + ($1365)|0);
     $1371 = HEAP8[$1370>>0]|0;
     $1372 = HEAP32[$35>>2]|0;
     $1373 = $1372 & $1349;
     $1374 = $1352 >>> 16;
     $1375 = $1374 & 255;
     $1376 = $1373 >>> $1375;
     $1377 = $1355 >>> 16;
     $1378 = $1377 & 255;
     $1379 = (125924 + ($1378<<2)|0);
     $1380 = HEAP32[$1379>>2]|0;
     $1381 = (($1380) + ($1376)|0);
     $1382 = HEAP8[$1381>>0]|0;
     $1383 = HEAP32[$36>>2]|0;
     $1384 = $1383 & $1349;
     $1385 = $1352 >>> 24;
     $1386 = $1384 >>> $1385;
     $1387 = $1355 >>> 24;
     $1388 = (125924 + ($1387<<2)|0);
     $1389 = HEAP32[$1388>>2]|0;
     $1390 = (($1389) + ($1386)|0);
     $1391 = HEAP8[$1390>>0]|0;
     $dA$3$shrunk = $1391;$dB$3$shrunk = $1382;$dG$3$shrunk = $1371;$dR$3$shrunk = $1360;
     break;
    }
    default: {
     $dA$3$shrunk = 0;$dB$3$shrunk = 0;$dG$3$shrunk = 0;$dR$3$shrunk = 0;
    }
    }
    $dR$3 = $dR$3$shrunk&255;
    $dG$3 = $dG$3$shrunk&255;
    $dB$3 = $dB$3$shrunk&255;
    $dA$3 = $dA$3$shrunk&255;
    $1392 = (($sR$3) - ($dR$3))|0;
    $1393 = Math_imul($1392, $20)|0;
    $1394 = (($1393>>>0) / 255)&-1;
    $1395 = (($1394) + ($dR$3))|0;
    $1396 = (($sG$3) - ($dG$3))|0;
    $1397 = Math_imul($1396, $20)|0;
    $1398 = (($1397>>>0) / 255)&-1;
    $1399 = (($1398) + ($dG$3))|0;
    $1400 = (($sB$3) - ($dB$3))|0;
    $1401 = Math_imul($1400, $20)|0;
    $1402 = (($1401>>>0) / 255)&-1;
    $1403 = (($1402) + ($dB$3))|0;
    $1404 = (($dA$3) + ($20))|0;
    $1405 = Math_imul($dA$3, $20)|0;
    $1406 = (($1405>>>0) / 255)&-1;
    $1407 = (($1404) - ($1406))|0;
    switch ($17|0) {
    case 1:  {
     $1408 = HEAP32[$29>>2]|0;
     $1409 = $1408 & 255;
     $1410 = $1395 >>> $1409;
     $1411 = HEAP32[$30>>2]|0;
     $1412 = $1411 & 255;
     $1413 = $1410 << $1412;
     $1414 = $1408 >>> 8;
     $1415 = $1414 & 255;
     $1416 = $1399 >>> $1415;
     $1417 = $1411 >>> 8;
     $1418 = $1417 & 255;
     $1419 = $1416 << $1418;
     $1420 = $1419 | $1413;
     $1421 = $1408 >>> 16;
     $1422 = $1421 & 255;
     $1423 = $1403 >>> $1422;
     $1424 = $1411 >>> 16;
     $1425 = $1424 & 255;
     $1426 = $1423 << $1425;
     $1427 = $1420 | $1426;
     $1428 = $1408 >>> 24;
     $1429 = $1407 >>> $1428;
     $1430 = $1411 >>> 24;
     $1431 = $1429 << $1430;
     $1432 = $1427 | $1431;
     $1433 = $1432&255;
     HEAP8[$dst$4>>0] = $1433;
     break;
    }
    case 2:  {
     $1434 = HEAP32[$29>>2]|0;
     $1435 = $1434 & 255;
     $1436 = $1395 >>> $1435;
     $1437 = HEAP32[$30>>2]|0;
     $1438 = $1437 & 255;
     $1439 = $1436 << $1438;
     $1440 = $1434 >>> 8;
     $1441 = $1440 & 255;
     $1442 = $1399 >>> $1441;
     $1443 = $1437 >>> 8;
     $1444 = $1443 & 255;
     $1445 = $1442 << $1444;
     $1446 = $1445 | $1439;
     $1447 = $1434 >>> 16;
     $1448 = $1447 & 255;
     $1449 = $1403 >>> $1448;
     $1450 = $1437 >>> 16;
     $1451 = $1450 & 255;
     $1452 = $1449 << $1451;
     $1453 = $1446 | $1452;
     $1454 = $1434 >>> 24;
     $1455 = $1407 >>> $1454;
     $1456 = $1437 >>> 24;
     $1457 = $1455 << $1456;
     $1458 = $1453 | $1457;
     $1459 = $1458&65535;
     HEAP16[$dst$4>>1] = $1459;
     break;
    }
    case 3:  {
     $1460 = $1395&255;
     $1461 = HEAP8[$30>>0]|0;
     $div27 = ($1461&255) >>> 3;
     $1462 = $div27&255;
     $1463 = (($dst$4) + ($1462)|0);
     HEAP8[$1463>>0] = $1460;
     $1464 = $1399&255;
     $1465 = HEAP8[$31>>0]|0;
     $div28 = ($1465&255) >>> 3;
     $1466 = $div28&255;
     $1467 = (($dst$4) + ($1466)|0);
     HEAP8[$1467>>0] = $1464;
     $1468 = $1403&255;
     $1469 = HEAP8[$32>>0]|0;
     $div29 = ($1469&255) >>> 3;
     $1470 = $div29&255;
     $1471 = (($dst$4) + ($1470)|0);
     HEAP8[$1471>>0] = $1468;
     break;
    }
    case 4:  {
     $1472 = HEAP32[$29>>2]|0;
     $1473 = $1472 & 255;
     $1474 = $1395 >>> $1473;
     $1475 = HEAP32[$30>>2]|0;
     $1476 = $1475 & 255;
     $1477 = $1474 << $1476;
     $1478 = $1472 >>> 8;
     $1479 = $1478 & 255;
     $1480 = $1399 >>> $1479;
     $1481 = $1475 >>> 8;
     $1482 = $1481 & 255;
     $1483 = $1480 << $1482;
     $1484 = $1483 | $1477;
     $1485 = $1472 >>> 16;
     $1486 = $1485 & 255;
     $1487 = $1403 >>> $1486;
     $1488 = $1475 >>> 16;
     $1489 = $1488 & 255;
     $1490 = $1487 << $1489;
     $1491 = $1484 | $1490;
     $1492 = $1472 >>> 24;
     $1493 = $1407 >>> $1492;
     $1494 = $1475 >>> 24;
     $1495 = $1493 << $1494;
     $1496 = $1491 | $1495;
     HEAP32[$dst$4>>2] = $1496;
     break;
    }
    default: {
    }
    }
    $1497 = (($src$4) + ($14)|0);
    $1498 = (($dst$4) + ($17)|0);
    $1499 = (($n$3) + -1)|0;
    $1500 = ($n$3|0)>(1);
    if ($1500) {
     $dst$1 = $1498;$n$0 = $1499;$src$1 = $1497;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $1501 = (($1497) + ($5)|0);
  $1502 = (($1498) + ($7)|0);
  $1503 = ($42|0)==(0);
  if ($1503) {
   label = 70;
   break;
  } else {
   $$in = $42;$dst$039 = $1502;$src$040 = $1501;
  }
 }
 if ((label|0) == 68) {
  // unreachable;
 }
 else if ((label|0) == 70) {
  return;
 }
}
function _Blit565to565SurfaceAlpha($info) {
 $info = $info|0;
 var $$lcssa = 0, $$lcssa13 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $dstp$01 = 0, $dstp$01$us = 0, $dstp$1$us = 0, $dstp$2 = 0, $dstp$3 = 0, $dstp$4 = 0, $n$0$us = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $srcp$02 = 0, $srcp$02$us = 0, $srcp$1$us = 0, $srcp$2 = 0, $srcp$3 = 0, $srcp$4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 63|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = ($1<<24>>24)==(-128);
 if ($2) {
  _Blit16to16SurfaceAlpha128($info,-2082);
  return;
 }
 $3 = $1&255;
 $4 = ((($info)) + 24|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 28|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = HEAP32[$info>>2]|0;
 $9 = ((($info)) + 16|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 >> 1;
 $12 = ((($info)) + 20|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($info)) + 36|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = $15 >> 1;
 $17 = $3 >>> 3;
 $18 = (($7) + -1)|0;
 $19 = ($7|0)==(0);
 if ($19) {
  return;
 }
 $20 = (($5) + 3)|0;
 $21 = (($20|0) / 4)&-1;
 $22 = $5 & 3;
 $23 = ($22|0)==(0);
 if ($23) {
  $109 = $18;$dstp$01$us = $13;$srcp$02$us = $8;
  while(1) {
   $dstp$1$us = $dstp$01$us;$n$0$us = $21;$srcp$1$us = $srcp$02$us;
   while(1) {
    $24 = ((($srcp$1$us)) + 2|0);
    $25 = HEAP16[$srcp$1$us>>1]|0;
    $26 = $25&65535;
    $27 = HEAP16[$dstp$1$us>>1]|0;
    $28 = $27&65535;
    $29 = $26 << 16;
    $30 = $29 | $26;
    $31 = $30 & 132184095;
    $32 = $28 << 16;
    $33 = $32 | $28;
    $34 = $33 & 132184095;
    $35 = (($31) - ($34))|0;
    $36 = Math_imul($35, $17)|0;
    $37 = $36 >>> 5;
    $38 = (($37) + ($34))|0;
    $39 = $38 & 132184095;
    $40 = $39 >>> 16;
    $41 = $40 | $39;
    $42 = $41&65535;
    $43 = ((($dstp$1$us)) + 2|0);
    HEAP16[$dstp$1$us>>1] = $42;
    $44 = ((($srcp$1$us)) + 4|0);
    $45 = HEAP16[$24>>1]|0;
    $46 = $45&65535;
    $47 = HEAP16[$43>>1]|0;
    $48 = $47&65535;
    $49 = $46 << 16;
    $50 = $49 | $46;
    $51 = $50 & 132184095;
    $52 = $48 << 16;
    $53 = $52 | $48;
    $54 = $53 & 132184095;
    $55 = (($51) - ($54))|0;
    $56 = Math_imul($55, $17)|0;
    $57 = $56 >>> 5;
    $58 = (($57) + ($54))|0;
    $59 = $58 & 132184095;
    $60 = $59 >>> 16;
    $61 = $60 | $59;
    $62 = $61&65535;
    $63 = ((($dstp$1$us)) + 4|0);
    HEAP16[$43>>1] = $62;
    $64 = ((($srcp$1$us)) + 6|0);
    $65 = HEAP16[$44>>1]|0;
    $66 = $65&65535;
    $67 = HEAP16[$63>>1]|0;
    $68 = $67&65535;
    $69 = $66 << 16;
    $70 = $69 | $66;
    $71 = $70 & 132184095;
    $72 = $68 << 16;
    $73 = $72 | $68;
    $74 = $73 & 132184095;
    $75 = (($71) - ($74))|0;
    $76 = Math_imul($75, $17)|0;
    $77 = $76 >>> 5;
    $78 = (($77) + ($74))|0;
    $79 = $78 & 132184095;
    $80 = $79 >>> 16;
    $81 = $80 | $79;
    $82 = $81&65535;
    $83 = ((($dstp$1$us)) + 6|0);
    HEAP16[$63>>1] = $82;
    $84 = ((($srcp$1$us)) + 8|0);
    $85 = HEAP16[$64>>1]|0;
    $86 = $85&65535;
    $87 = HEAP16[$83>>1]|0;
    $88 = $87&65535;
    $89 = $86 << 16;
    $90 = $89 | $86;
    $91 = $90 & 132184095;
    $92 = $88 << 16;
    $93 = $92 | $88;
    $94 = $93 & 132184095;
    $95 = (($91) - ($94))|0;
    $96 = Math_imul($95, $17)|0;
    $97 = $96 >>> 5;
    $98 = (($97) + ($94))|0;
    $99 = $98 & 132184095;
    $100 = $99 >>> 16;
    $101 = $100 | $99;
    $102 = $101&65535;
    $103 = ((($dstp$1$us)) + 8|0);
    HEAP16[$83>>1] = $102;
    $104 = (($n$0$us) + -1)|0;
    $105 = ($n$0$us|0)>(1);
    if ($105) {
     $dstp$1$us = $103;$n$0$us = $104;$srcp$1$us = $84;
    } else {
     $$lcssa = $84;$$lcssa13 = $103;
     break;
    }
   }
   $106 = (($$lcssa) + ($11<<1)|0);
   $107 = (($$lcssa13) + ($16<<1)|0);
   $108 = (($109) + -1)|0;
   $110 = ($109|0)==(0);
   if ($110) {
    break;
   } else {
    $109 = $108;$dstp$01$us = $107;$srcp$02$us = $106;
   }
  }
  return;
 } else {
  $196 = $18;$dstp$01 = $13;$srcp$02 = $8;
 }
 L15: while(1) {
  switch ($22|0) {
  case 1:  {
   $dstp$4 = $dstp$01;$n$3 = $21;$srcp$4 = $srcp$02;
   label = 12;
   break;
  }
  case 3:  {
   $dstp$2 = $dstp$01;$n$1 = $21;$srcp$2 = $srcp$02;
   label = 10;
   break;
  }
  case 2:  {
   $dstp$3 = $dstp$01;$n$2 = $21;$srcp$3 = $srcp$02;
   label = 11;
   break;
  }
  default: {
   label = 13;
   break L15;
  }
  }
  while(1) {
   if ((label|0) == 10) {
    label = 0;
    $134 = ((($srcp$2)) + 2|0);
    $135 = HEAP16[$srcp$2>>1]|0;
    $136 = $135&65535;
    $137 = HEAP16[$dstp$2>>1]|0;
    $138 = $137&65535;
    $139 = $136 << 16;
    $140 = $139 | $136;
    $141 = $140 & 132184095;
    $142 = $138 << 16;
    $143 = $142 | $138;
    $144 = $143 & 132184095;
    $145 = (($141) - ($144))|0;
    $146 = Math_imul($145, $17)|0;
    $147 = $146 >>> 5;
    $148 = (($147) + ($144))|0;
    $149 = $148 & 132184095;
    $150 = $149 >>> 16;
    $151 = $150 | $149;
    $152 = $151&65535;
    $153 = ((($dstp$2)) + 2|0);
    HEAP16[$dstp$2>>1] = $152;
    $dstp$3 = $153;$n$2 = $n$1;$srcp$3 = $134;
    label = 11;
    continue;
   }
   else if ((label|0) == 11) {
    label = 0;
    $154 = ((($srcp$3)) + 2|0);
    $155 = HEAP16[$srcp$3>>1]|0;
    $156 = $155&65535;
    $157 = HEAP16[$dstp$3>>1]|0;
    $158 = $157&65535;
    $159 = $156 << 16;
    $160 = $159 | $156;
    $161 = $160 & 132184095;
    $162 = $158 << 16;
    $163 = $162 | $158;
    $164 = $163 & 132184095;
    $165 = (($161) - ($164))|0;
    $166 = Math_imul($165, $17)|0;
    $167 = $166 >>> 5;
    $168 = (($167) + ($164))|0;
    $169 = $168 & 132184095;
    $170 = $169 >>> 16;
    $171 = $170 | $169;
    $172 = $171&65535;
    $173 = ((($dstp$3)) + 2|0);
    HEAP16[$dstp$3>>1] = $172;
    $dstp$4 = $173;$n$3 = $n$2;$srcp$4 = $154;
    label = 12;
    continue;
   }
   else if ((label|0) == 12) {
    label = 0;
    $114 = ((($srcp$4)) + 2|0);
    $174 = HEAP16[$srcp$4>>1]|0;
    $175 = $174&65535;
    $176 = HEAP16[$dstp$4>>1]|0;
    $177 = $176&65535;
    $178 = $175 << 16;
    $179 = $178 | $175;
    $180 = $179 & 132184095;
    $181 = $177 << 16;
    $182 = $181 | $177;
    $183 = $182 & 132184095;
    $184 = (($180) - ($183))|0;
    $185 = Math_imul($184, $17)|0;
    $186 = $185 >>> 5;
    $187 = (($186) + ($183))|0;
    $188 = $187 & 132184095;
    $189 = $188 >>> 16;
    $190 = $189 | $188;
    $191 = $190&65535;
    $117 = ((($dstp$4)) + 2|0);
    HEAP16[$dstp$4>>1] = $191;
    $192 = ($n$3|0)>(1);
    if (!($192)) {
     break;
    }
    $111 = (($n$3) + -1)|0;
    $112 = ((($srcp$4)) + 4|0);
    $113 = HEAP16[$114>>1]|0;
    $115 = $113&65535;
    $116 = HEAP16[$117>>1]|0;
    $118 = $116&65535;
    $119 = $115 << 16;
    $120 = $119 | $115;
    $121 = $120 & 132184095;
    $122 = $118 << 16;
    $123 = $122 | $118;
    $124 = $123 & 132184095;
    $125 = (($121) - ($124))|0;
    $126 = Math_imul($125, $17)|0;
    $127 = $126 >>> 5;
    $128 = (($127) + ($124))|0;
    $129 = $128 & 132184095;
    $130 = $129 >>> 16;
    $131 = $130 | $129;
    $132 = $131&65535;
    $133 = ((($dstp$4)) + 4|0);
    HEAP16[$117>>1] = $132;
    $dstp$2 = $133;$n$1 = $111;$srcp$2 = $112;
    label = 10;
    continue;
   }
  }
  $193 = (($114) + ($11<<1)|0);
  $194 = (($117) + ($16<<1)|0);
  $195 = (($196) + -1)|0;
  $197 = ($196|0)==(0);
  if ($197) {
   label = 15;
   break;
  } else {
   $196 = $195;$dstp$01 = $194;$srcp$02 = $193;
  }
 }
 if ((label|0) == 13) {
  // unreachable;
 }
 else if ((label|0) == 15) {
  return;
 }
}
function _BlitNto1SurfaceAlphaKey($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0;
 var $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0;
 var $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0;
 var $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0;
 var $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0;
 var $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0;
 var $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0;
 var $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0;
 var $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0;
 var $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0;
 var $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0;
 var $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0;
 var $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0;
 var $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0;
 var $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0;
 var $1267 = 0, $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0;
 var $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0;
 var $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0;
 var $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0;
 var $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0;
 var $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0;
 var $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0;
 var $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0;
 var $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0;
 var $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0;
 var $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0;
 var $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0;
 var $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0;
 var $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0;
 var $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0;
 var $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0;
 var $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0;
 var $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0;
 var $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0;
 var $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0;
 var $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0;
 var $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0;
 var $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0;
 var $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0;
 var $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0;
 var $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0;
 var $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0;
 var $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0;
 var $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0;
 var $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0;
 var $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0;
 var $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0;
 var $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0;
 var $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0;
 var $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0;
 var $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0;
 var $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0;
 var $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0;
 var $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0;
 var $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0;
 var $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0;
 var $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0;
 var $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0;
 var $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $Pixel$0 = 0, $Pixel$1 = 0, $Pixel$2 = 0, $Pixel$3 = 0, $Pixel$4 = 0;
 var $Pixel$5 = 0, $Pixel$6 = 0, $Pixel$7 = 0, $div = 0, $div1 = 0, $div10 = 0, $div12 = 0, $div13 = 0, $div15 = 0, $div16 = 0, $div18 = 0, $div19 = 0, $div21 = 0, $div22 = 0, $div3 = 0, $div4 = 0, $div6 = 0, $div7 = 0, $div9 = 0, $dst$024 = 0;
 var $dst$1 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $sB$0 = 0, $sB$0$shrunk = 0, $sB$1 = 0, $sB$1$shrunk = 0;
 var $sB$2 = 0, $sB$2$shrunk = 0, $sB$3 = 0, $sB$3$shrunk = 0, $sB$4 = 0, $sB$4$shrunk = 0, $sB$5 = 0, $sB$5$shrunk = 0, $sB$6 = 0, $sB$6$shrunk = 0, $sB$7 = 0, $sB$7$shrunk = 0, $sG$0 = 0, $sG$0$shrunk = 0, $sG$1 = 0, $sG$1$shrunk = 0, $sG$2 = 0, $sG$2$shrunk = 0, $sG$3 = 0, $sG$3$shrunk = 0;
 var $sG$4 = 0, $sG$4$shrunk = 0, $sG$5 = 0, $sG$5$shrunk = 0, $sG$6 = 0, $sG$6$shrunk = 0, $sG$7 = 0, $sG$7$shrunk = 0, $sR$0 = 0, $sR$0$shrunk = 0, $sR$1 = 0, $sR$1$shrunk = 0, $sR$2 = 0, $sR$2$shrunk = 0, $sR$3 = 0, $sR$3$shrunk = 0, $sR$4 = 0, $sR$4$shrunk = 0, $sR$5 = 0, $sR$5$shrunk = 0;
 var $sR$6 = 0, $sR$6$shrunk = 0, $sR$7 = 0, $sR$7$shrunk = 0, $src$025 = 0, $src$1 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 48|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($info)) + 40|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($11)) + 9|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = $13&255;
 $15 = ((($info)) + 56|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($info)) + 63|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = $18&255;
 $20 = ($3|0)==(0);
 if ($20) {
  return;
 }
 $21 = ((($info)) + 44|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ((($info)) + 20|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = HEAP32[$info>>2]|0;
 $26 = (($1) + 7)|0;
 $27 = (($26|0) / 8)&-1;
 $28 = $1 & 7;
 $29 = ((($22)) + 4|0);
 $30 = ($9|0)==(0|0);
 $31 = ((($11)) + 12|0);
 $32 = ((($11)) + 32|0);
 $33 = ((($11)) + 28|0);
 $34 = ((($11)) + 16|0);
 $35 = ((($11)) + 20|0);
 $$in = $3;$dst$024 = $24;$src$025 = $25;
 L4: while(1) {
  $36 = (($$in) + -1)|0;
  switch ($28|0) {
  case 0:  {
   $dst$1 = $dst$024;$n$0 = $27;$src$1 = $src$025;
   label = 4;
   break;
  }
  case 7:  {
   $dst$2 = $dst$024;$n$1 = $27;$src$2 = $src$025;
   label = 14;
   break;
  }
  case 6:  {
   $dst$3 = $dst$024;$n$2 = $27;$src$3 = $src$025;
   label = 24;
   break;
  }
  case 5:  {
   $dst$4 = $dst$024;$n$3 = $27;$src$4 = $src$025;
   label = 34;
   break;
  }
  case 4:  {
   $dst$5 = $dst$024;$n$4 = $27;$src$5 = $src$025;
   label = 44;
   break;
  }
  case 3:  {
   $dst$6 = $dst$024;$n$5 = $27;$src$6 = $src$025;
   label = 54;
   break;
  }
  case 2:  {
   $dst$7 = $dst$024;$n$6 = $27;$src$7 = $src$025;
   label = 64;
   break;
  }
  case 1:  {
   $dst$8 = $dst$024;$n$7 = $27;$src$8 = $src$025;
   label = 74;
   break;
  }
  default: {
   label = 84;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $37 = HEAP8[$src$1>>0]|0;
     $38 = $37&255;
     $39 = HEAP32[$31>>2]|0;
     $40 = $38 & $39;
     $41 = HEAP32[$32>>2]|0;
     $42 = $41 & 255;
     $43 = $40 >>> $42;
     $44 = HEAP32[$33>>2]|0;
     $45 = $44 & 255;
     $46 = (125924 + ($45<<2)|0);
     $47 = HEAP32[$46>>2]|0;
     $48 = (($47) + ($43)|0);
     $49 = HEAP8[$48>>0]|0;
     $50 = HEAP32[$34>>2]|0;
     $51 = $50 & $38;
     $52 = $41 >>> 8;
     $53 = $52 & 255;
     $54 = $51 >>> $53;
     $55 = $44 >>> 8;
     $56 = $55 & 255;
     $57 = (125924 + ($56<<2)|0);
     $58 = HEAP32[$57>>2]|0;
     $59 = (($58) + ($54)|0);
     $60 = HEAP8[$59>>0]|0;
     $61 = HEAP32[$35>>2]|0;
     $62 = $61 & $38;
     $63 = $41 >>> 16;
     $64 = $63 & 255;
     $65 = $62 >>> $64;
     $66 = $44 >>> 16;
     $67 = $66 & 255;
     $68 = (125924 + ($67<<2)|0);
     $69 = HEAP32[$68>>2]|0;
     $70 = (($69) + ($65)|0);
     $71 = HEAP8[$70>>0]|0;
     $Pixel$0 = $38;$sB$0$shrunk = $71;$sG$0$shrunk = $60;$sR$0$shrunk = $49;
     break;
    }
    case 2:  {
     $72 = HEAP16[$src$1>>1]|0;
     $73 = $72&65535;
     $74 = HEAP32[$31>>2]|0;
     $75 = $73 & $74;
     $76 = HEAP32[$32>>2]|0;
     $77 = $76 & 255;
     $78 = $75 >>> $77;
     $79 = HEAP32[$33>>2]|0;
     $80 = $79 & 255;
     $81 = (125924 + ($80<<2)|0);
     $82 = HEAP32[$81>>2]|0;
     $83 = (($82) + ($78)|0);
     $84 = HEAP8[$83>>0]|0;
     $85 = HEAP32[$34>>2]|0;
     $86 = $85 & $73;
     $87 = $76 >>> 8;
     $88 = $87 & 255;
     $89 = $86 >>> $88;
     $90 = $79 >>> 8;
     $91 = $90 & 255;
     $92 = (125924 + ($91<<2)|0);
     $93 = HEAP32[$92>>2]|0;
     $94 = (($93) + ($89)|0);
     $95 = HEAP8[$94>>0]|0;
     $96 = HEAP32[$35>>2]|0;
     $97 = $96 & $73;
     $98 = $76 >>> 16;
     $99 = $98 & 255;
     $100 = $97 >>> $99;
     $101 = $79 >>> 16;
     $102 = $101 & 255;
     $103 = (125924 + ($102<<2)|0);
     $104 = HEAP32[$103>>2]|0;
     $105 = (($104) + ($100)|0);
     $106 = HEAP8[$105>>0]|0;
     $Pixel$0 = $73;$sB$0$shrunk = $106;$sG$0$shrunk = $95;$sR$0$shrunk = $84;
     break;
    }
    case 3:  {
     $107 = HEAP32[$32>>2]|0;
     $div18 = $107 >>> 3;
     $108 = $div18 & 31;
     $109 = (($src$1) + ($108)|0);
     $110 = HEAP8[$109>>0]|0;
     $111 = $107 >>> 11;
     $div19 = $111 & 31;
     $112 = (($src$1) + ($div19)|0);
     $113 = HEAP8[$112>>0]|0;
     $114 = $107 >>> 19;
     $115 = $114 & 31;
     $116 = (($src$1) + ($115)|0);
     $117 = HEAP8[$116>>0]|0;
     $Pixel$0 = 0;$sB$0$shrunk = $117;$sG$0$shrunk = $113;$sR$0$shrunk = $110;
     break;
    }
    case 4:  {
     $118 = HEAP32[$src$1>>2]|0;
     $119 = HEAP32[$31>>2]|0;
     $120 = $119 & $118;
     $121 = HEAP32[$32>>2]|0;
     $122 = $121 & 255;
     $123 = $120 >>> $122;
     $124 = HEAP32[$33>>2]|0;
     $125 = $124 & 255;
     $126 = (125924 + ($125<<2)|0);
     $127 = HEAP32[$126>>2]|0;
     $128 = (($127) + ($123)|0);
     $129 = HEAP8[$128>>0]|0;
     $130 = HEAP32[$34>>2]|0;
     $131 = $130 & $118;
     $132 = $121 >>> 8;
     $133 = $132 & 255;
     $134 = $131 >>> $133;
     $135 = $124 >>> 8;
     $136 = $135 & 255;
     $137 = (125924 + ($136<<2)|0);
     $138 = HEAP32[$137>>2]|0;
     $139 = (($138) + ($134)|0);
     $140 = HEAP8[$139>>0]|0;
     $141 = HEAP32[$35>>2]|0;
     $142 = $141 & $118;
     $143 = $121 >>> 16;
     $144 = $143 & 255;
     $145 = $142 >>> $144;
     $146 = $124 >>> 16;
     $147 = $146 & 255;
     $148 = (125924 + ($147<<2)|0);
     $149 = HEAP32[$148>>2]|0;
     $150 = (($149) + ($145)|0);
     $151 = HEAP8[$150>>0]|0;
     $Pixel$0 = $118;$sB$0$shrunk = $151;$sG$0$shrunk = $140;$sR$0$shrunk = $129;
     break;
    }
    default: {
     $Pixel$0 = 0;$sB$0$shrunk = 0;$sG$0$shrunk = 0;$sR$0$shrunk = 0;
    }
    }
    $152 = ($Pixel$0|0)==($16|0);
    do {
     if (!($152)) {
      $sB$0 = $sB$0$shrunk&255;
      $sG$0 = $sG$0$shrunk&255;
      $sR$0 = $sR$0$shrunk&255;
      $153 = HEAP8[$dst$1>>0]|0;
      $154 = $153&255;
      $155 = HEAP32[$29>>2]|0;
      $156 = ((($155)) + 4|0);
      $157 = HEAP32[$156>>2]|0;
      $158 = (($157) + ($154<<2)|0);
      $159 = HEAP8[$158>>0]|0;
      $160 = $159&255;
      $161 = (((($157) + ($154<<2)|0)) + 1|0);
      $162 = HEAP8[$161>>0]|0;
      $163 = $162&255;
      $164 = (((($157) + ($154<<2)|0)) + 2|0);
      $165 = HEAP8[$164>>0]|0;
      $166 = $165&255;
      $167 = (($sR$0) - ($160))|0;
      $168 = Math_imul($167, $19)|0;
      $169 = (($168>>>0) / 255)&-1;
      $170 = (($169) + ($160))|0;
      $171 = (($sG$0) - ($163))|0;
      $172 = Math_imul($171, $19)|0;
      $173 = (($172>>>0) / 255)&-1;
      $174 = (($173) + ($163))|0;
      $175 = (($sB$0) - ($166))|0;
      $176 = Math_imul($175, $19)|0;
      $177 = (($176>>>0) / 255)&-1;
      $178 = (($177) + ($166))|0;
      $179 = $170 & 224;
      $180 = $174 >>> 3;
      $181 = $180 & 28;
      $182 = $181 | $179;
      $183 = $178 >>> 6;
      $184 = $183 & 3;
      $185 = $182 | $184;
      if ($30) {
       $186 = $185&255;
       HEAP8[$dst$1>>0] = $186;
       break;
      } else {
       $187 = (($9) + ($185)|0);
       $188 = HEAP8[$187>>0]|0;
       HEAP8[$dst$1>>0] = $188;
       break;
      }
     }
    } while(0);
    $189 = ((($dst$1)) + 1|0);
    $190 = (($src$1) + ($14)|0);
    $dst$2 = $189;$n$1 = $n$0;$src$2 = $190;
    label = 14;
    continue;
   }
   else if ((label|0) == 14) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $191 = HEAP8[$src$2>>0]|0;
     $192 = $191&255;
     $193 = HEAP32[$31>>2]|0;
     $194 = $192 & $193;
     $195 = HEAP32[$32>>2]|0;
     $196 = $195 & 255;
     $197 = $194 >>> $196;
     $198 = HEAP32[$33>>2]|0;
     $199 = $198 & 255;
     $200 = (125924 + ($199<<2)|0);
     $201 = HEAP32[$200>>2]|0;
     $202 = (($201) + ($197)|0);
     $203 = HEAP8[$202>>0]|0;
     $204 = HEAP32[$34>>2]|0;
     $205 = $204 & $192;
     $206 = $195 >>> 8;
     $207 = $206 & 255;
     $208 = $205 >>> $207;
     $209 = $198 >>> 8;
     $210 = $209 & 255;
     $211 = (125924 + ($210<<2)|0);
     $212 = HEAP32[$211>>2]|0;
     $213 = (($212) + ($208)|0);
     $214 = HEAP8[$213>>0]|0;
     $215 = HEAP32[$35>>2]|0;
     $216 = $215 & $192;
     $217 = $195 >>> 16;
     $218 = $217 & 255;
     $219 = $216 >>> $218;
     $220 = $198 >>> 16;
     $221 = $220 & 255;
     $222 = (125924 + ($221<<2)|0);
     $223 = HEAP32[$222>>2]|0;
     $224 = (($223) + ($219)|0);
     $225 = HEAP8[$224>>0]|0;
     $Pixel$1 = $192;$sB$1$shrunk = $225;$sG$1$shrunk = $214;$sR$1$shrunk = $203;
     break;
    }
    case 2:  {
     $226 = HEAP16[$src$2>>1]|0;
     $227 = $226&65535;
     $228 = HEAP32[$31>>2]|0;
     $229 = $227 & $228;
     $230 = HEAP32[$32>>2]|0;
     $231 = $230 & 255;
     $232 = $229 >>> $231;
     $233 = HEAP32[$33>>2]|0;
     $234 = $233 & 255;
     $235 = (125924 + ($234<<2)|0);
     $236 = HEAP32[$235>>2]|0;
     $237 = (($236) + ($232)|0);
     $238 = HEAP8[$237>>0]|0;
     $239 = HEAP32[$34>>2]|0;
     $240 = $239 & $227;
     $241 = $230 >>> 8;
     $242 = $241 & 255;
     $243 = $240 >>> $242;
     $244 = $233 >>> 8;
     $245 = $244 & 255;
     $246 = (125924 + ($245<<2)|0);
     $247 = HEAP32[$246>>2]|0;
     $248 = (($247) + ($243)|0);
     $249 = HEAP8[$248>>0]|0;
     $250 = HEAP32[$35>>2]|0;
     $251 = $250 & $227;
     $252 = $230 >>> 16;
     $253 = $252 & 255;
     $254 = $251 >>> $253;
     $255 = $233 >>> 16;
     $256 = $255 & 255;
     $257 = (125924 + ($256<<2)|0);
     $258 = HEAP32[$257>>2]|0;
     $259 = (($258) + ($254)|0);
     $260 = HEAP8[$259>>0]|0;
     $Pixel$1 = $227;$sB$1$shrunk = $260;$sG$1$shrunk = $249;$sR$1$shrunk = $238;
     break;
    }
    case 3:  {
     $261 = HEAP32[$32>>2]|0;
     $div15 = $261 >>> 3;
     $262 = $div15 & 31;
     $263 = (($src$2) + ($262)|0);
     $264 = HEAP8[$263>>0]|0;
     $265 = $261 >>> 11;
     $div16 = $265 & 31;
     $266 = (($src$2) + ($div16)|0);
     $267 = HEAP8[$266>>0]|0;
     $268 = $261 >>> 19;
     $269 = $268 & 31;
     $270 = (($src$2) + ($269)|0);
     $271 = HEAP8[$270>>0]|0;
     $Pixel$1 = 0;$sB$1$shrunk = $271;$sG$1$shrunk = $267;$sR$1$shrunk = $264;
     break;
    }
    case 4:  {
     $272 = HEAP32[$src$2>>2]|0;
     $273 = HEAP32[$31>>2]|0;
     $274 = $273 & $272;
     $275 = HEAP32[$32>>2]|0;
     $276 = $275 & 255;
     $277 = $274 >>> $276;
     $278 = HEAP32[$33>>2]|0;
     $279 = $278 & 255;
     $280 = (125924 + ($279<<2)|0);
     $281 = HEAP32[$280>>2]|0;
     $282 = (($281) + ($277)|0);
     $283 = HEAP8[$282>>0]|0;
     $284 = HEAP32[$34>>2]|0;
     $285 = $284 & $272;
     $286 = $275 >>> 8;
     $287 = $286 & 255;
     $288 = $285 >>> $287;
     $289 = $278 >>> 8;
     $290 = $289 & 255;
     $291 = (125924 + ($290<<2)|0);
     $292 = HEAP32[$291>>2]|0;
     $293 = (($292) + ($288)|0);
     $294 = HEAP8[$293>>0]|0;
     $295 = HEAP32[$35>>2]|0;
     $296 = $295 & $272;
     $297 = $275 >>> 16;
     $298 = $297 & 255;
     $299 = $296 >>> $298;
     $300 = $278 >>> 16;
     $301 = $300 & 255;
     $302 = (125924 + ($301<<2)|0);
     $303 = HEAP32[$302>>2]|0;
     $304 = (($303) + ($299)|0);
     $305 = HEAP8[$304>>0]|0;
     $Pixel$1 = $272;$sB$1$shrunk = $305;$sG$1$shrunk = $294;$sR$1$shrunk = $283;
     break;
    }
    default: {
     $Pixel$1 = 0;$sB$1$shrunk = 0;$sG$1$shrunk = 0;$sR$1$shrunk = 0;
    }
    }
    $306 = ($Pixel$1|0)==($16|0);
    do {
     if (!($306)) {
      $sB$1 = $sB$1$shrunk&255;
      $sG$1 = $sG$1$shrunk&255;
      $sR$1 = $sR$1$shrunk&255;
      $307 = HEAP8[$dst$2>>0]|0;
      $308 = $307&255;
      $309 = HEAP32[$29>>2]|0;
      $310 = ((($309)) + 4|0);
      $311 = HEAP32[$310>>2]|0;
      $312 = (($311) + ($308<<2)|0);
      $313 = HEAP8[$312>>0]|0;
      $314 = $313&255;
      $315 = (((($311) + ($308<<2)|0)) + 1|0);
      $316 = HEAP8[$315>>0]|0;
      $317 = $316&255;
      $318 = (((($311) + ($308<<2)|0)) + 2|0);
      $319 = HEAP8[$318>>0]|0;
      $320 = $319&255;
      $321 = (($sR$1) - ($314))|0;
      $322 = Math_imul($321, $19)|0;
      $323 = (($322>>>0) / 255)&-1;
      $324 = (($323) + ($314))|0;
      $325 = (($sG$1) - ($317))|0;
      $326 = Math_imul($325, $19)|0;
      $327 = (($326>>>0) / 255)&-1;
      $328 = (($327) + ($317))|0;
      $329 = (($sB$1) - ($320))|0;
      $330 = Math_imul($329, $19)|0;
      $331 = (($330>>>0) / 255)&-1;
      $332 = (($331) + ($320))|0;
      $333 = $324 & 224;
      $334 = $328 >>> 3;
      $335 = $334 & 28;
      $336 = $335 | $333;
      $337 = $332 >>> 6;
      $338 = $337 & 3;
      $339 = $336 | $338;
      if ($30) {
       $340 = $339&255;
       HEAP8[$dst$2>>0] = $340;
       break;
      } else {
       $341 = (($9) + ($339)|0);
       $342 = HEAP8[$341>>0]|0;
       HEAP8[$dst$2>>0] = $342;
       break;
      }
     }
    } while(0);
    $343 = ((($dst$2)) + 1|0);
    $344 = (($src$2) + ($14)|0);
    $dst$3 = $343;$n$2 = $n$1;$src$3 = $344;
    label = 24;
    continue;
   }
   else if ((label|0) == 24) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $345 = HEAP8[$src$3>>0]|0;
     $346 = $345&255;
     $347 = HEAP32[$31>>2]|0;
     $348 = $346 & $347;
     $349 = HEAP32[$32>>2]|0;
     $350 = $349 & 255;
     $351 = $348 >>> $350;
     $352 = HEAP32[$33>>2]|0;
     $353 = $352 & 255;
     $354 = (125924 + ($353<<2)|0);
     $355 = HEAP32[$354>>2]|0;
     $356 = (($355) + ($351)|0);
     $357 = HEAP8[$356>>0]|0;
     $358 = HEAP32[$34>>2]|0;
     $359 = $358 & $346;
     $360 = $349 >>> 8;
     $361 = $360 & 255;
     $362 = $359 >>> $361;
     $363 = $352 >>> 8;
     $364 = $363 & 255;
     $365 = (125924 + ($364<<2)|0);
     $366 = HEAP32[$365>>2]|0;
     $367 = (($366) + ($362)|0);
     $368 = HEAP8[$367>>0]|0;
     $369 = HEAP32[$35>>2]|0;
     $370 = $369 & $346;
     $371 = $349 >>> 16;
     $372 = $371 & 255;
     $373 = $370 >>> $372;
     $374 = $352 >>> 16;
     $375 = $374 & 255;
     $376 = (125924 + ($375<<2)|0);
     $377 = HEAP32[$376>>2]|0;
     $378 = (($377) + ($373)|0);
     $379 = HEAP8[$378>>0]|0;
     $Pixel$2 = $346;$sB$2$shrunk = $379;$sG$2$shrunk = $368;$sR$2$shrunk = $357;
     break;
    }
    case 2:  {
     $380 = HEAP16[$src$3>>1]|0;
     $381 = $380&65535;
     $382 = HEAP32[$31>>2]|0;
     $383 = $381 & $382;
     $384 = HEAP32[$32>>2]|0;
     $385 = $384 & 255;
     $386 = $383 >>> $385;
     $387 = HEAP32[$33>>2]|0;
     $388 = $387 & 255;
     $389 = (125924 + ($388<<2)|0);
     $390 = HEAP32[$389>>2]|0;
     $391 = (($390) + ($386)|0);
     $392 = HEAP8[$391>>0]|0;
     $393 = HEAP32[$34>>2]|0;
     $394 = $393 & $381;
     $395 = $384 >>> 8;
     $396 = $395 & 255;
     $397 = $394 >>> $396;
     $398 = $387 >>> 8;
     $399 = $398 & 255;
     $400 = (125924 + ($399<<2)|0);
     $401 = HEAP32[$400>>2]|0;
     $402 = (($401) + ($397)|0);
     $403 = HEAP8[$402>>0]|0;
     $404 = HEAP32[$35>>2]|0;
     $405 = $404 & $381;
     $406 = $384 >>> 16;
     $407 = $406 & 255;
     $408 = $405 >>> $407;
     $409 = $387 >>> 16;
     $410 = $409 & 255;
     $411 = (125924 + ($410<<2)|0);
     $412 = HEAP32[$411>>2]|0;
     $413 = (($412) + ($408)|0);
     $414 = HEAP8[$413>>0]|0;
     $Pixel$2 = $381;$sB$2$shrunk = $414;$sG$2$shrunk = $403;$sR$2$shrunk = $392;
     break;
    }
    case 3:  {
     $415 = HEAP32[$32>>2]|0;
     $div12 = $415 >>> 3;
     $416 = $div12 & 31;
     $417 = (($src$3) + ($416)|0);
     $418 = HEAP8[$417>>0]|0;
     $419 = $415 >>> 11;
     $div13 = $419 & 31;
     $420 = (($src$3) + ($div13)|0);
     $421 = HEAP8[$420>>0]|0;
     $422 = $415 >>> 19;
     $423 = $422 & 31;
     $424 = (($src$3) + ($423)|0);
     $425 = HEAP8[$424>>0]|0;
     $Pixel$2 = 0;$sB$2$shrunk = $425;$sG$2$shrunk = $421;$sR$2$shrunk = $418;
     break;
    }
    case 4:  {
     $426 = HEAP32[$src$3>>2]|0;
     $427 = HEAP32[$31>>2]|0;
     $428 = $427 & $426;
     $429 = HEAP32[$32>>2]|0;
     $430 = $429 & 255;
     $431 = $428 >>> $430;
     $432 = HEAP32[$33>>2]|0;
     $433 = $432 & 255;
     $434 = (125924 + ($433<<2)|0);
     $435 = HEAP32[$434>>2]|0;
     $436 = (($435) + ($431)|0);
     $437 = HEAP8[$436>>0]|0;
     $438 = HEAP32[$34>>2]|0;
     $439 = $438 & $426;
     $440 = $429 >>> 8;
     $441 = $440 & 255;
     $442 = $439 >>> $441;
     $443 = $432 >>> 8;
     $444 = $443 & 255;
     $445 = (125924 + ($444<<2)|0);
     $446 = HEAP32[$445>>2]|0;
     $447 = (($446) + ($442)|0);
     $448 = HEAP8[$447>>0]|0;
     $449 = HEAP32[$35>>2]|0;
     $450 = $449 & $426;
     $451 = $429 >>> 16;
     $452 = $451 & 255;
     $453 = $450 >>> $452;
     $454 = $432 >>> 16;
     $455 = $454 & 255;
     $456 = (125924 + ($455<<2)|0);
     $457 = HEAP32[$456>>2]|0;
     $458 = (($457) + ($453)|0);
     $459 = HEAP8[$458>>0]|0;
     $Pixel$2 = $426;$sB$2$shrunk = $459;$sG$2$shrunk = $448;$sR$2$shrunk = $437;
     break;
    }
    default: {
     $Pixel$2 = 0;$sB$2$shrunk = 0;$sG$2$shrunk = 0;$sR$2$shrunk = 0;
    }
    }
    $460 = ($Pixel$2|0)==($16|0);
    do {
     if (!($460)) {
      $sB$2 = $sB$2$shrunk&255;
      $sG$2 = $sG$2$shrunk&255;
      $sR$2 = $sR$2$shrunk&255;
      $461 = HEAP8[$dst$3>>0]|0;
      $462 = $461&255;
      $463 = HEAP32[$29>>2]|0;
      $464 = ((($463)) + 4|0);
      $465 = HEAP32[$464>>2]|0;
      $466 = (($465) + ($462<<2)|0);
      $467 = HEAP8[$466>>0]|0;
      $468 = $467&255;
      $469 = (((($465) + ($462<<2)|0)) + 1|0);
      $470 = HEAP8[$469>>0]|0;
      $471 = $470&255;
      $472 = (((($465) + ($462<<2)|0)) + 2|0);
      $473 = HEAP8[$472>>0]|0;
      $474 = $473&255;
      $475 = (($sR$2) - ($468))|0;
      $476 = Math_imul($475, $19)|0;
      $477 = (($476>>>0) / 255)&-1;
      $478 = (($477) + ($468))|0;
      $479 = (($sG$2) - ($471))|0;
      $480 = Math_imul($479, $19)|0;
      $481 = (($480>>>0) / 255)&-1;
      $482 = (($481) + ($471))|0;
      $483 = (($sB$2) - ($474))|0;
      $484 = Math_imul($483, $19)|0;
      $485 = (($484>>>0) / 255)&-1;
      $486 = (($485) + ($474))|0;
      $487 = $478 & 224;
      $488 = $482 >>> 3;
      $489 = $488 & 28;
      $490 = $489 | $487;
      $491 = $486 >>> 6;
      $492 = $491 & 3;
      $493 = $490 | $492;
      if ($30) {
       $494 = $493&255;
       HEAP8[$dst$3>>0] = $494;
       break;
      } else {
       $495 = (($9) + ($493)|0);
       $496 = HEAP8[$495>>0]|0;
       HEAP8[$dst$3>>0] = $496;
       break;
      }
     }
    } while(0);
    $497 = ((($dst$3)) + 1|0);
    $498 = (($src$3) + ($14)|0);
    $dst$4 = $497;$n$3 = $n$2;$src$4 = $498;
    label = 34;
    continue;
   }
   else if ((label|0) == 34) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $499 = HEAP8[$src$4>>0]|0;
     $500 = $499&255;
     $501 = HEAP32[$31>>2]|0;
     $502 = $500 & $501;
     $503 = HEAP32[$32>>2]|0;
     $504 = $503 & 255;
     $505 = $502 >>> $504;
     $506 = HEAP32[$33>>2]|0;
     $507 = $506 & 255;
     $508 = (125924 + ($507<<2)|0);
     $509 = HEAP32[$508>>2]|0;
     $510 = (($509) + ($505)|0);
     $511 = HEAP8[$510>>0]|0;
     $512 = HEAP32[$34>>2]|0;
     $513 = $512 & $500;
     $514 = $503 >>> 8;
     $515 = $514 & 255;
     $516 = $513 >>> $515;
     $517 = $506 >>> 8;
     $518 = $517 & 255;
     $519 = (125924 + ($518<<2)|0);
     $520 = HEAP32[$519>>2]|0;
     $521 = (($520) + ($516)|0);
     $522 = HEAP8[$521>>0]|0;
     $523 = HEAP32[$35>>2]|0;
     $524 = $523 & $500;
     $525 = $503 >>> 16;
     $526 = $525 & 255;
     $527 = $524 >>> $526;
     $528 = $506 >>> 16;
     $529 = $528 & 255;
     $530 = (125924 + ($529<<2)|0);
     $531 = HEAP32[$530>>2]|0;
     $532 = (($531) + ($527)|0);
     $533 = HEAP8[$532>>0]|0;
     $Pixel$3 = $500;$sB$3$shrunk = $533;$sG$3$shrunk = $522;$sR$3$shrunk = $511;
     break;
    }
    case 2:  {
     $534 = HEAP16[$src$4>>1]|0;
     $535 = $534&65535;
     $536 = HEAP32[$31>>2]|0;
     $537 = $535 & $536;
     $538 = HEAP32[$32>>2]|0;
     $539 = $538 & 255;
     $540 = $537 >>> $539;
     $541 = HEAP32[$33>>2]|0;
     $542 = $541 & 255;
     $543 = (125924 + ($542<<2)|0);
     $544 = HEAP32[$543>>2]|0;
     $545 = (($544) + ($540)|0);
     $546 = HEAP8[$545>>0]|0;
     $547 = HEAP32[$34>>2]|0;
     $548 = $547 & $535;
     $549 = $538 >>> 8;
     $550 = $549 & 255;
     $551 = $548 >>> $550;
     $552 = $541 >>> 8;
     $553 = $552 & 255;
     $554 = (125924 + ($553<<2)|0);
     $555 = HEAP32[$554>>2]|0;
     $556 = (($555) + ($551)|0);
     $557 = HEAP8[$556>>0]|0;
     $558 = HEAP32[$35>>2]|0;
     $559 = $558 & $535;
     $560 = $538 >>> 16;
     $561 = $560 & 255;
     $562 = $559 >>> $561;
     $563 = $541 >>> 16;
     $564 = $563 & 255;
     $565 = (125924 + ($564<<2)|0);
     $566 = HEAP32[$565>>2]|0;
     $567 = (($566) + ($562)|0);
     $568 = HEAP8[$567>>0]|0;
     $Pixel$3 = $535;$sB$3$shrunk = $568;$sG$3$shrunk = $557;$sR$3$shrunk = $546;
     break;
    }
    case 3:  {
     $569 = HEAP32[$32>>2]|0;
     $div9 = $569 >>> 3;
     $570 = $div9 & 31;
     $571 = (($src$4) + ($570)|0);
     $572 = HEAP8[$571>>0]|0;
     $573 = $569 >>> 11;
     $div10 = $573 & 31;
     $574 = (($src$4) + ($div10)|0);
     $575 = HEAP8[$574>>0]|0;
     $576 = $569 >>> 19;
     $577 = $576 & 31;
     $578 = (($src$4) + ($577)|0);
     $579 = HEAP8[$578>>0]|0;
     $Pixel$3 = 0;$sB$3$shrunk = $579;$sG$3$shrunk = $575;$sR$3$shrunk = $572;
     break;
    }
    case 4:  {
     $580 = HEAP32[$src$4>>2]|0;
     $581 = HEAP32[$31>>2]|0;
     $582 = $581 & $580;
     $583 = HEAP32[$32>>2]|0;
     $584 = $583 & 255;
     $585 = $582 >>> $584;
     $586 = HEAP32[$33>>2]|0;
     $587 = $586 & 255;
     $588 = (125924 + ($587<<2)|0);
     $589 = HEAP32[$588>>2]|0;
     $590 = (($589) + ($585)|0);
     $591 = HEAP8[$590>>0]|0;
     $592 = HEAP32[$34>>2]|0;
     $593 = $592 & $580;
     $594 = $583 >>> 8;
     $595 = $594 & 255;
     $596 = $593 >>> $595;
     $597 = $586 >>> 8;
     $598 = $597 & 255;
     $599 = (125924 + ($598<<2)|0);
     $600 = HEAP32[$599>>2]|0;
     $601 = (($600) + ($596)|0);
     $602 = HEAP8[$601>>0]|0;
     $603 = HEAP32[$35>>2]|0;
     $604 = $603 & $580;
     $605 = $583 >>> 16;
     $606 = $605 & 255;
     $607 = $604 >>> $606;
     $608 = $586 >>> 16;
     $609 = $608 & 255;
     $610 = (125924 + ($609<<2)|0);
     $611 = HEAP32[$610>>2]|0;
     $612 = (($611) + ($607)|0);
     $613 = HEAP8[$612>>0]|0;
     $Pixel$3 = $580;$sB$3$shrunk = $613;$sG$3$shrunk = $602;$sR$3$shrunk = $591;
     break;
    }
    default: {
     $Pixel$3 = 0;$sB$3$shrunk = 0;$sG$3$shrunk = 0;$sR$3$shrunk = 0;
    }
    }
    $614 = ($Pixel$3|0)==($16|0);
    do {
     if (!($614)) {
      $sB$3 = $sB$3$shrunk&255;
      $sG$3 = $sG$3$shrunk&255;
      $sR$3 = $sR$3$shrunk&255;
      $615 = HEAP8[$dst$4>>0]|0;
      $616 = $615&255;
      $617 = HEAP32[$29>>2]|0;
      $618 = ((($617)) + 4|0);
      $619 = HEAP32[$618>>2]|0;
      $620 = (($619) + ($616<<2)|0);
      $621 = HEAP8[$620>>0]|0;
      $622 = $621&255;
      $623 = (((($619) + ($616<<2)|0)) + 1|0);
      $624 = HEAP8[$623>>0]|0;
      $625 = $624&255;
      $626 = (((($619) + ($616<<2)|0)) + 2|0);
      $627 = HEAP8[$626>>0]|0;
      $628 = $627&255;
      $629 = (($sR$3) - ($622))|0;
      $630 = Math_imul($629, $19)|0;
      $631 = (($630>>>0) / 255)&-1;
      $632 = (($631) + ($622))|0;
      $633 = (($sG$3) - ($625))|0;
      $634 = Math_imul($633, $19)|0;
      $635 = (($634>>>0) / 255)&-1;
      $636 = (($635) + ($625))|0;
      $637 = (($sB$3) - ($628))|0;
      $638 = Math_imul($637, $19)|0;
      $639 = (($638>>>0) / 255)&-1;
      $640 = (($639) + ($628))|0;
      $641 = $632 & 224;
      $642 = $636 >>> 3;
      $643 = $642 & 28;
      $644 = $643 | $641;
      $645 = $640 >>> 6;
      $646 = $645 & 3;
      $647 = $644 | $646;
      if ($30) {
       $648 = $647&255;
       HEAP8[$dst$4>>0] = $648;
       break;
      } else {
       $649 = (($9) + ($647)|0);
       $650 = HEAP8[$649>>0]|0;
       HEAP8[$dst$4>>0] = $650;
       break;
      }
     }
    } while(0);
    $651 = ((($dst$4)) + 1|0);
    $652 = (($src$4) + ($14)|0);
    $dst$5 = $651;$n$4 = $n$3;$src$5 = $652;
    label = 44;
    continue;
   }
   else if ((label|0) == 44) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $653 = HEAP8[$src$5>>0]|0;
     $654 = $653&255;
     $655 = HEAP32[$31>>2]|0;
     $656 = $654 & $655;
     $657 = HEAP32[$32>>2]|0;
     $658 = $657 & 255;
     $659 = $656 >>> $658;
     $660 = HEAP32[$33>>2]|0;
     $661 = $660 & 255;
     $662 = (125924 + ($661<<2)|0);
     $663 = HEAP32[$662>>2]|0;
     $664 = (($663) + ($659)|0);
     $665 = HEAP8[$664>>0]|0;
     $666 = HEAP32[$34>>2]|0;
     $667 = $666 & $654;
     $668 = $657 >>> 8;
     $669 = $668 & 255;
     $670 = $667 >>> $669;
     $671 = $660 >>> 8;
     $672 = $671 & 255;
     $673 = (125924 + ($672<<2)|0);
     $674 = HEAP32[$673>>2]|0;
     $675 = (($674) + ($670)|0);
     $676 = HEAP8[$675>>0]|0;
     $677 = HEAP32[$35>>2]|0;
     $678 = $677 & $654;
     $679 = $657 >>> 16;
     $680 = $679 & 255;
     $681 = $678 >>> $680;
     $682 = $660 >>> 16;
     $683 = $682 & 255;
     $684 = (125924 + ($683<<2)|0);
     $685 = HEAP32[$684>>2]|0;
     $686 = (($685) + ($681)|0);
     $687 = HEAP8[$686>>0]|0;
     $Pixel$4 = $654;$sB$4$shrunk = $687;$sG$4$shrunk = $676;$sR$4$shrunk = $665;
     break;
    }
    case 2:  {
     $688 = HEAP16[$src$5>>1]|0;
     $689 = $688&65535;
     $690 = HEAP32[$31>>2]|0;
     $691 = $689 & $690;
     $692 = HEAP32[$32>>2]|0;
     $693 = $692 & 255;
     $694 = $691 >>> $693;
     $695 = HEAP32[$33>>2]|0;
     $696 = $695 & 255;
     $697 = (125924 + ($696<<2)|0);
     $698 = HEAP32[$697>>2]|0;
     $699 = (($698) + ($694)|0);
     $700 = HEAP8[$699>>0]|0;
     $701 = HEAP32[$34>>2]|0;
     $702 = $701 & $689;
     $703 = $692 >>> 8;
     $704 = $703 & 255;
     $705 = $702 >>> $704;
     $706 = $695 >>> 8;
     $707 = $706 & 255;
     $708 = (125924 + ($707<<2)|0);
     $709 = HEAP32[$708>>2]|0;
     $710 = (($709) + ($705)|0);
     $711 = HEAP8[$710>>0]|0;
     $712 = HEAP32[$35>>2]|0;
     $713 = $712 & $689;
     $714 = $692 >>> 16;
     $715 = $714 & 255;
     $716 = $713 >>> $715;
     $717 = $695 >>> 16;
     $718 = $717 & 255;
     $719 = (125924 + ($718<<2)|0);
     $720 = HEAP32[$719>>2]|0;
     $721 = (($720) + ($716)|0);
     $722 = HEAP8[$721>>0]|0;
     $Pixel$4 = $689;$sB$4$shrunk = $722;$sG$4$shrunk = $711;$sR$4$shrunk = $700;
     break;
    }
    case 3:  {
     $723 = HEAP32[$32>>2]|0;
     $div6 = $723 >>> 3;
     $724 = $div6 & 31;
     $725 = (($src$5) + ($724)|0);
     $726 = HEAP8[$725>>0]|0;
     $727 = $723 >>> 11;
     $div7 = $727 & 31;
     $728 = (($src$5) + ($div7)|0);
     $729 = HEAP8[$728>>0]|0;
     $730 = $723 >>> 19;
     $731 = $730 & 31;
     $732 = (($src$5) + ($731)|0);
     $733 = HEAP8[$732>>0]|0;
     $Pixel$4 = 0;$sB$4$shrunk = $733;$sG$4$shrunk = $729;$sR$4$shrunk = $726;
     break;
    }
    case 4:  {
     $734 = HEAP32[$src$5>>2]|0;
     $735 = HEAP32[$31>>2]|0;
     $736 = $735 & $734;
     $737 = HEAP32[$32>>2]|0;
     $738 = $737 & 255;
     $739 = $736 >>> $738;
     $740 = HEAP32[$33>>2]|0;
     $741 = $740 & 255;
     $742 = (125924 + ($741<<2)|0);
     $743 = HEAP32[$742>>2]|0;
     $744 = (($743) + ($739)|0);
     $745 = HEAP8[$744>>0]|0;
     $746 = HEAP32[$34>>2]|0;
     $747 = $746 & $734;
     $748 = $737 >>> 8;
     $749 = $748 & 255;
     $750 = $747 >>> $749;
     $751 = $740 >>> 8;
     $752 = $751 & 255;
     $753 = (125924 + ($752<<2)|0);
     $754 = HEAP32[$753>>2]|0;
     $755 = (($754) + ($750)|0);
     $756 = HEAP8[$755>>0]|0;
     $757 = HEAP32[$35>>2]|0;
     $758 = $757 & $734;
     $759 = $737 >>> 16;
     $760 = $759 & 255;
     $761 = $758 >>> $760;
     $762 = $740 >>> 16;
     $763 = $762 & 255;
     $764 = (125924 + ($763<<2)|0);
     $765 = HEAP32[$764>>2]|0;
     $766 = (($765) + ($761)|0);
     $767 = HEAP8[$766>>0]|0;
     $Pixel$4 = $734;$sB$4$shrunk = $767;$sG$4$shrunk = $756;$sR$4$shrunk = $745;
     break;
    }
    default: {
     $Pixel$4 = 0;$sB$4$shrunk = 0;$sG$4$shrunk = 0;$sR$4$shrunk = 0;
    }
    }
    $768 = ($Pixel$4|0)==($16|0);
    do {
     if (!($768)) {
      $sB$4 = $sB$4$shrunk&255;
      $sG$4 = $sG$4$shrunk&255;
      $sR$4 = $sR$4$shrunk&255;
      $769 = HEAP8[$dst$5>>0]|0;
      $770 = $769&255;
      $771 = HEAP32[$29>>2]|0;
      $772 = ((($771)) + 4|0);
      $773 = HEAP32[$772>>2]|0;
      $774 = (($773) + ($770<<2)|0);
      $775 = HEAP8[$774>>0]|0;
      $776 = $775&255;
      $777 = (((($773) + ($770<<2)|0)) + 1|0);
      $778 = HEAP8[$777>>0]|0;
      $779 = $778&255;
      $780 = (((($773) + ($770<<2)|0)) + 2|0);
      $781 = HEAP8[$780>>0]|0;
      $782 = $781&255;
      $783 = (($sR$4) - ($776))|0;
      $784 = Math_imul($783, $19)|0;
      $785 = (($784>>>0) / 255)&-1;
      $786 = (($785) + ($776))|0;
      $787 = (($sG$4) - ($779))|0;
      $788 = Math_imul($787, $19)|0;
      $789 = (($788>>>0) / 255)&-1;
      $790 = (($789) + ($779))|0;
      $791 = (($sB$4) - ($782))|0;
      $792 = Math_imul($791, $19)|0;
      $793 = (($792>>>0) / 255)&-1;
      $794 = (($793) + ($782))|0;
      $795 = $786 & 224;
      $796 = $790 >>> 3;
      $797 = $796 & 28;
      $798 = $797 | $795;
      $799 = $794 >>> 6;
      $800 = $799 & 3;
      $801 = $798 | $800;
      if ($30) {
       $802 = $801&255;
       HEAP8[$dst$5>>0] = $802;
       break;
      } else {
       $803 = (($9) + ($801)|0);
       $804 = HEAP8[$803>>0]|0;
       HEAP8[$dst$5>>0] = $804;
       break;
      }
     }
    } while(0);
    $805 = ((($dst$5)) + 1|0);
    $806 = (($src$5) + ($14)|0);
    $dst$6 = $805;$n$5 = $n$4;$src$6 = $806;
    label = 54;
    continue;
   }
   else if ((label|0) == 54) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $807 = HEAP8[$src$6>>0]|0;
     $808 = $807&255;
     $809 = HEAP32[$31>>2]|0;
     $810 = $808 & $809;
     $811 = HEAP32[$32>>2]|0;
     $812 = $811 & 255;
     $813 = $810 >>> $812;
     $814 = HEAP32[$33>>2]|0;
     $815 = $814 & 255;
     $816 = (125924 + ($815<<2)|0);
     $817 = HEAP32[$816>>2]|0;
     $818 = (($817) + ($813)|0);
     $819 = HEAP8[$818>>0]|0;
     $820 = HEAP32[$34>>2]|0;
     $821 = $820 & $808;
     $822 = $811 >>> 8;
     $823 = $822 & 255;
     $824 = $821 >>> $823;
     $825 = $814 >>> 8;
     $826 = $825 & 255;
     $827 = (125924 + ($826<<2)|0);
     $828 = HEAP32[$827>>2]|0;
     $829 = (($828) + ($824)|0);
     $830 = HEAP8[$829>>0]|0;
     $831 = HEAP32[$35>>2]|0;
     $832 = $831 & $808;
     $833 = $811 >>> 16;
     $834 = $833 & 255;
     $835 = $832 >>> $834;
     $836 = $814 >>> 16;
     $837 = $836 & 255;
     $838 = (125924 + ($837<<2)|0);
     $839 = HEAP32[$838>>2]|0;
     $840 = (($839) + ($835)|0);
     $841 = HEAP8[$840>>0]|0;
     $Pixel$5 = $808;$sB$5$shrunk = $841;$sG$5$shrunk = $830;$sR$5$shrunk = $819;
     break;
    }
    case 2:  {
     $842 = HEAP16[$src$6>>1]|0;
     $843 = $842&65535;
     $844 = HEAP32[$31>>2]|0;
     $845 = $843 & $844;
     $846 = HEAP32[$32>>2]|0;
     $847 = $846 & 255;
     $848 = $845 >>> $847;
     $849 = HEAP32[$33>>2]|0;
     $850 = $849 & 255;
     $851 = (125924 + ($850<<2)|0);
     $852 = HEAP32[$851>>2]|0;
     $853 = (($852) + ($848)|0);
     $854 = HEAP8[$853>>0]|0;
     $855 = HEAP32[$34>>2]|0;
     $856 = $855 & $843;
     $857 = $846 >>> 8;
     $858 = $857 & 255;
     $859 = $856 >>> $858;
     $860 = $849 >>> 8;
     $861 = $860 & 255;
     $862 = (125924 + ($861<<2)|0);
     $863 = HEAP32[$862>>2]|0;
     $864 = (($863) + ($859)|0);
     $865 = HEAP8[$864>>0]|0;
     $866 = HEAP32[$35>>2]|0;
     $867 = $866 & $843;
     $868 = $846 >>> 16;
     $869 = $868 & 255;
     $870 = $867 >>> $869;
     $871 = $849 >>> 16;
     $872 = $871 & 255;
     $873 = (125924 + ($872<<2)|0);
     $874 = HEAP32[$873>>2]|0;
     $875 = (($874) + ($870)|0);
     $876 = HEAP8[$875>>0]|0;
     $Pixel$5 = $843;$sB$5$shrunk = $876;$sG$5$shrunk = $865;$sR$5$shrunk = $854;
     break;
    }
    case 3:  {
     $877 = HEAP32[$32>>2]|0;
     $div3 = $877 >>> 3;
     $878 = $div3 & 31;
     $879 = (($src$6) + ($878)|0);
     $880 = HEAP8[$879>>0]|0;
     $881 = $877 >>> 11;
     $div4 = $881 & 31;
     $882 = (($src$6) + ($div4)|0);
     $883 = HEAP8[$882>>0]|0;
     $884 = $877 >>> 19;
     $885 = $884 & 31;
     $886 = (($src$6) + ($885)|0);
     $887 = HEAP8[$886>>0]|0;
     $Pixel$5 = 0;$sB$5$shrunk = $887;$sG$5$shrunk = $883;$sR$5$shrunk = $880;
     break;
    }
    case 4:  {
     $888 = HEAP32[$src$6>>2]|0;
     $889 = HEAP32[$31>>2]|0;
     $890 = $889 & $888;
     $891 = HEAP32[$32>>2]|0;
     $892 = $891 & 255;
     $893 = $890 >>> $892;
     $894 = HEAP32[$33>>2]|0;
     $895 = $894 & 255;
     $896 = (125924 + ($895<<2)|0);
     $897 = HEAP32[$896>>2]|0;
     $898 = (($897) + ($893)|0);
     $899 = HEAP8[$898>>0]|0;
     $900 = HEAP32[$34>>2]|0;
     $901 = $900 & $888;
     $902 = $891 >>> 8;
     $903 = $902 & 255;
     $904 = $901 >>> $903;
     $905 = $894 >>> 8;
     $906 = $905 & 255;
     $907 = (125924 + ($906<<2)|0);
     $908 = HEAP32[$907>>2]|0;
     $909 = (($908) + ($904)|0);
     $910 = HEAP8[$909>>0]|0;
     $911 = HEAP32[$35>>2]|0;
     $912 = $911 & $888;
     $913 = $891 >>> 16;
     $914 = $913 & 255;
     $915 = $912 >>> $914;
     $916 = $894 >>> 16;
     $917 = $916 & 255;
     $918 = (125924 + ($917<<2)|0);
     $919 = HEAP32[$918>>2]|0;
     $920 = (($919) + ($915)|0);
     $921 = HEAP8[$920>>0]|0;
     $Pixel$5 = $888;$sB$5$shrunk = $921;$sG$5$shrunk = $910;$sR$5$shrunk = $899;
     break;
    }
    default: {
     $Pixel$5 = 0;$sB$5$shrunk = 0;$sG$5$shrunk = 0;$sR$5$shrunk = 0;
    }
    }
    $922 = ($Pixel$5|0)==($16|0);
    do {
     if (!($922)) {
      $sB$5 = $sB$5$shrunk&255;
      $sG$5 = $sG$5$shrunk&255;
      $sR$5 = $sR$5$shrunk&255;
      $923 = HEAP8[$dst$6>>0]|0;
      $924 = $923&255;
      $925 = HEAP32[$29>>2]|0;
      $926 = ((($925)) + 4|0);
      $927 = HEAP32[$926>>2]|0;
      $928 = (($927) + ($924<<2)|0);
      $929 = HEAP8[$928>>0]|0;
      $930 = $929&255;
      $931 = (((($927) + ($924<<2)|0)) + 1|0);
      $932 = HEAP8[$931>>0]|0;
      $933 = $932&255;
      $934 = (((($927) + ($924<<2)|0)) + 2|0);
      $935 = HEAP8[$934>>0]|0;
      $936 = $935&255;
      $937 = (($sR$5) - ($930))|0;
      $938 = Math_imul($937, $19)|0;
      $939 = (($938>>>0) / 255)&-1;
      $940 = (($939) + ($930))|0;
      $941 = (($sG$5) - ($933))|0;
      $942 = Math_imul($941, $19)|0;
      $943 = (($942>>>0) / 255)&-1;
      $944 = (($943) + ($933))|0;
      $945 = (($sB$5) - ($936))|0;
      $946 = Math_imul($945, $19)|0;
      $947 = (($946>>>0) / 255)&-1;
      $948 = (($947) + ($936))|0;
      $949 = $940 & 224;
      $950 = $944 >>> 3;
      $951 = $950 & 28;
      $952 = $951 | $949;
      $953 = $948 >>> 6;
      $954 = $953 & 3;
      $955 = $952 | $954;
      if ($30) {
       $956 = $955&255;
       HEAP8[$dst$6>>0] = $956;
       break;
      } else {
       $957 = (($9) + ($955)|0);
       $958 = HEAP8[$957>>0]|0;
       HEAP8[$dst$6>>0] = $958;
       break;
      }
     }
    } while(0);
    $959 = ((($dst$6)) + 1|0);
    $960 = (($src$6) + ($14)|0);
    $dst$7 = $959;$n$6 = $n$5;$src$7 = $960;
    label = 64;
    continue;
   }
   else if ((label|0) == 64) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $961 = HEAP8[$src$7>>0]|0;
     $962 = $961&255;
     $963 = HEAP32[$31>>2]|0;
     $964 = $962 & $963;
     $965 = HEAP32[$32>>2]|0;
     $966 = $965 & 255;
     $967 = $964 >>> $966;
     $968 = HEAP32[$33>>2]|0;
     $969 = $968 & 255;
     $970 = (125924 + ($969<<2)|0);
     $971 = HEAP32[$970>>2]|0;
     $972 = (($971) + ($967)|0);
     $973 = HEAP8[$972>>0]|0;
     $974 = HEAP32[$34>>2]|0;
     $975 = $974 & $962;
     $976 = $965 >>> 8;
     $977 = $976 & 255;
     $978 = $975 >>> $977;
     $979 = $968 >>> 8;
     $980 = $979 & 255;
     $981 = (125924 + ($980<<2)|0);
     $982 = HEAP32[$981>>2]|0;
     $983 = (($982) + ($978)|0);
     $984 = HEAP8[$983>>0]|0;
     $985 = HEAP32[$35>>2]|0;
     $986 = $985 & $962;
     $987 = $965 >>> 16;
     $988 = $987 & 255;
     $989 = $986 >>> $988;
     $990 = $968 >>> 16;
     $991 = $990 & 255;
     $992 = (125924 + ($991<<2)|0);
     $993 = HEAP32[$992>>2]|0;
     $994 = (($993) + ($989)|0);
     $995 = HEAP8[$994>>0]|0;
     $Pixel$6 = $962;$sB$6$shrunk = $995;$sG$6$shrunk = $984;$sR$6$shrunk = $973;
     break;
    }
    case 2:  {
     $996 = HEAP16[$src$7>>1]|0;
     $997 = $996&65535;
     $998 = HEAP32[$31>>2]|0;
     $999 = $997 & $998;
     $1000 = HEAP32[$32>>2]|0;
     $1001 = $1000 & 255;
     $1002 = $999 >>> $1001;
     $1003 = HEAP32[$33>>2]|0;
     $1004 = $1003 & 255;
     $1005 = (125924 + ($1004<<2)|0);
     $1006 = HEAP32[$1005>>2]|0;
     $1007 = (($1006) + ($1002)|0);
     $1008 = HEAP8[$1007>>0]|0;
     $1009 = HEAP32[$34>>2]|0;
     $1010 = $1009 & $997;
     $1011 = $1000 >>> 8;
     $1012 = $1011 & 255;
     $1013 = $1010 >>> $1012;
     $1014 = $1003 >>> 8;
     $1015 = $1014 & 255;
     $1016 = (125924 + ($1015<<2)|0);
     $1017 = HEAP32[$1016>>2]|0;
     $1018 = (($1017) + ($1013)|0);
     $1019 = HEAP8[$1018>>0]|0;
     $1020 = HEAP32[$35>>2]|0;
     $1021 = $1020 & $997;
     $1022 = $1000 >>> 16;
     $1023 = $1022 & 255;
     $1024 = $1021 >>> $1023;
     $1025 = $1003 >>> 16;
     $1026 = $1025 & 255;
     $1027 = (125924 + ($1026<<2)|0);
     $1028 = HEAP32[$1027>>2]|0;
     $1029 = (($1028) + ($1024)|0);
     $1030 = HEAP8[$1029>>0]|0;
     $Pixel$6 = $997;$sB$6$shrunk = $1030;$sG$6$shrunk = $1019;$sR$6$shrunk = $1008;
     break;
    }
    case 3:  {
     $1031 = HEAP32[$32>>2]|0;
     $div = $1031 >>> 3;
     $1032 = $div & 31;
     $1033 = (($src$7) + ($1032)|0);
     $1034 = HEAP8[$1033>>0]|0;
     $1035 = $1031 >>> 11;
     $div1 = $1035 & 31;
     $1036 = (($src$7) + ($div1)|0);
     $1037 = HEAP8[$1036>>0]|0;
     $1038 = $1031 >>> 19;
     $1039 = $1038 & 31;
     $1040 = (($src$7) + ($1039)|0);
     $1041 = HEAP8[$1040>>0]|0;
     $Pixel$6 = 0;$sB$6$shrunk = $1041;$sG$6$shrunk = $1037;$sR$6$shrunk = $1034;
     break;
    }
    case 4:  {
     $1042 = HEAP32[$src$7>>2]|0;
     $1043 = HEAP32[$31>>2]|0;
     $1044 = $1043 & $1042;
     $1045 = HEAP32[$32>>2]|0;
     $1046 = $1045 & 255;
     $1047 = $1044 >>> $1046;
     $1048 = HEAP32[$33>>2]|0;
     $1049 = $1048 & 255;
     $1050 = (125924 + ($1049<<2)|0);
     $1051 = HEAP32[$1050>>2]|0;
     $1052 = (($1051) + ($1047)|0);
     $1053 = HEAP8[$1052>>0]|0;
     $1054 = HEAP32[$34>>2]|0;
     $1055 = $1054 & $1042;
     $1056 = $1045 >>> 8;
     $1057 = $1056 & 255;
     $1058 = $1055 >>> $1057;
     $1059 = $1048 >>> 8;
     $1060 = $1059 & 255;
     $1061 = (125924 + ($1060<<2)|0);
     $1062 = HEAP32[$1061>>2]|0;
     $1063 = (($1062) + ($1058)|0);
     $1064 = HEAP8[$1063>>0]|0;
     $1065 = HEAP32[$35>>2]|0;
     $1066 = $1065 & $1042;
     $1067 = $1045 >>> 16;
     $1068 = $1067 & 255;
     $1069 = $1066 >>> $1068;
     $1070 = $1048 >>> 16;
     $1071 = $1070 & 255;
     $1072 = (125924 + ($1071<<2)|0);
     $1073 = HEAP32[$1072>>2]|0;
     $1074 = (($1073) + ($1069)|0);
     $1075 = HEAP8[$1074>>0]|0;
     $Pixel$6 = $1042;$sB$6$shrunk = $1075;$sG$6$shrunk = $1064;$sR$6$shrunk = $1053;
     break;
    }
    default: {
     $Pixel$6 = 0;$sB$6$shrunk = 0;$sG$6$shrunk = 0;$sR$6$shrunk = 0;
    }
    }
    $1076 = ($Pixel$6|0)==($16|0);
    do {
     if (!($1076)) {
      $sB$6 = $sB$6$shrunk&255;
      $sG$6 = $sG$6$shrunk&255;
      $sR$6 = $sR$6$shrunk&255;
      $1077 = HEAP8[$dst$7>>0]|0;
      $1078 = $1077&255;
      $1079 = HEAP32[$29>>2]|0;
      $1080 = ((($1079)) + 4|0);
      $1081 = HEAP32[$1080>>2]|0;
      $1082 = (($1081) + ($1078<<2)|0);
      $1083 = HEAP8[$1082>>0]|0;
      $1084 = $1083&255;
      $1085 = (((($1081) + ($1078<<2)|0)) + 1|0);
      $1086 = HEAP8[$1085>>0]|0;
      $1087 = $1086&255;
      $1088 = (((($1081) + ($1078<<2)|0)) + 2|0);
      $1089 = HEAP8[$1088>>0]|0;
      $1090 = $1089&255;
      $1091 = (($sR$6) - ($1084))|0;
      $1092 = Math_imul($1091, $19)|0;
      $1093 = (($1092>>>0) / 255)&-1;
      $1094 = (($1093) + ($1084))|0;
      $1095 = (($sG$6) - ($1087))|0;
      $1096 = Math_imul($1095, $19)|0;
      $1097 = (($1096>>>0) / 255)&-1;
      $1098 = (($1097) + ($1087))|0;
      $1099 = (($sB$6) - ($1090))|0;
      $1100 = Math_imul($1099, $19)|0;
      $1101 = (($1100>>>0) / 255)&-1;
      $1102 = (($1101) + ($1090))|0;
      $1103 = $1094 & 224;
      $1104 = $1098 >>> 3;
      $1105 = $1104 & 28;
      $1106 = $1105 | $1103;
      $1107 = $1102 >>> 6;
      $1108 = $1107 & 3;
      $1109 = $1106 | $1108;
      if ($30) {
       $1110 = $1109&255;
       HEAP8[$dst$7>>0] = $1110;
       break;
      } else {
       $1111 = (($9) + ($1109)|0);
       $1112 = HEAP8[$1111>>0]|0;
       HEAP8[$dst$7>>0] = $1112;
       break;
      }
     }
    } while(0);
    $1113 = ((($dst$7)) + 1|0);
    $1114 = (($src$7) + ($14)|0);
    $dst$8 = $1113;$n$7 = $n$6;$src$8 = $1114;
    label = 74;
    continue;
   }
   else if ((label|0) == 74) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $1115 = HEAP8[$src$8>>0]|0;
     $1116 = $1115&255;
     $1117 = HEAP32[$31>>2]|0;
     $1118 = $1116 & $1117;
     $1119 = HEAP32[$32>>2]|0;
     $1120 = $1119 & 255;
     $1121 = $1118 >>> $1120;
     $1122 = HEAP32[$33>>2]|0;
     $1123 = $1122 & 255;
     $1124 = (125924 + ($1123<<2)|0);
     $1125 = HEAP32[$1124>>2]|0;
     $1126 = (($1125) + ($1121)|0);
     $1127 = HEAP8[$1126>>0]|0;
     $1128 = HEAP32[$34>>2]|0;
     $1129 = $1128 & $1116;
     $1130 = $1119 >>> 8;
     $1131 = $1130 & 255;
     $1132 = $1129 >>> $1131;
     $1133 = $1122 >>> 8;
     $1134 = $1133 & 255;
     $1135 = (125924 + ($1134<<2)|0);
     $1136 = HEAP32[$1135>>2]|0;
     $1137 = (($1136) + ($1132)|0);
     $1138 = HEAP8[$1137>>0]|0;
     $1139 = HEAP32[$35>>2]|0;
     $1140 = $1139 & $1116;
     $1141 = $1119 >>> 16;
     $1142 = $1141 & 255;
     $1143 = $1140 >>> $1142;
     $1144 = $1122 >>> 16;
     $1145 = $1144 & 255;
     $1146 = (125924 + ($1145<<2)|0);
     $1147 = HEAP32[$1146>>2]|0;
     $1148 = (($1147) + ($1143)|0);
     $1149 = HEAP8[$1148>>0]|0;
     $Pixel$7 = $1116;$sB$7$shrunk = $1149;$sG$7$shrunk = $1138;$sR$7$shrunk = $1127;
     break;
    }
    case 2:  {
     $1150 = HEAP16[$src$8>>1]|0;
     $1151 = $1150&65535;
     $1152 = HEAP32[$31>>2]|0;
     $1153 = $1151 & $1152;
     $1154 = HEAP32[$32>>2]|0;
     $1155 = $1154 & 255;
     $1156 = $1153 >>> $1155;
     $1157 = HEAP32[$33>>2]|0;
     $1158 = $1157 & 255;
     $1159 = (125924 + ($1158<<2)|0);
     $1160 = HEAP32[$1159>>2]|0;
     $1161 = (($1160) + ($1156)|0);
     $1162 = HEAP8[$1161>>0]|0;
     $1163 = HEAP32[$34>>2]|0;
     $1164 = $1163 & $1151;
     $1165 = $1154 >>> 8;
     $1166 = $1165 & 255;
     $1167 = $1164 >>> $1166;
     $1168 = $1157 >>> 8;
     $1169 = $1168 & 255;
     $1170 = (125924 + ($1169<<2)|0);
     $1171 = HEAP32[$1170>>2]|0;
     $1172 = (($1171) + ($1167)|0);
     $1173 = HEAP8[$1172>>0]|0;
     $1174 = HEAP32[$35>>2]|0;
     $1175 = $1174 & $1151;
     $1176 = $1154 >>> 16;
     $1177 = $1176 & 255;
     $1178 = $1175 >>> $1177;
     $1179 = $1157 >>> 16;
     $1180 = $1179 & 255;
     $1181 = (125924 + ($1180<<2)|0);
     $1182 = HEAP32[$1181>>2]|0;
     $1183 = (($1182) + ($1178)|0);
     $1184 = HEAP8[$1183>>0]|0;
     $Pixel$7 = $1151;$sB$7$shrunk = $1184;$sG$7$shrunk = $1173;$sR$7$shrunk = $1162;
     break;
    }
    case 3:  {
     $1185 = HEAP32[$32>>2]|0;
     $div21 = $1185 >>> 3;
     $1186 = $div21 & 31;
     $1187 = (($src$8) + ($1186)|0);
     $1188 = HEAP8[$1187>>0]|0;
     $1189 = $1185 >>> 11;
     $div22 = $1189 & 31;
     $1190 = (($src$8) + ($div22)|0);
     $1191 = HEAP8[$1190>>0]|0;
     $1192 = $1185 >>> 19;
     $1193 = $1192 & 31;
     $1194 = (($src$8) + ($1193)|0);
     $1195 = HEAP8[$1194>>0]|0;
     $Pixel$7 = 0;$sB$7$shrunk = $1195;$sG$7$shrunk = $1191;$sR$7$shrunk = $1188;
     break;
    }
    case 4:  {
     $1196 = HEAP32[$src$8>>2]|0;
     $1197 = HEAP32[$31>>2]|0;
     $1198 = $1197 & $1196;
     $1199 = HEAP32[$32>>2]|0;
     $1200 = $1199 & 255;
     $1201 = $1198 >>> $1200;
     $1202 = HEAP32[$33>>2]|0;
     $1203 = $1202 & 255;
     $1204 = (125924 + ($1203<<2)|0);
     $1205 = HEAP32[$1204>>2]|0;
     $1206 = (($1205) + ($1201)|0);
     $1207 = HEAP8[$1206>>0]|0;
     $1208 = HEAP32[$34>>2]|0;
     $1209 = $1208 & $1196;
     $1210 = $1199 >>> 8;
     $1211 = $1210 & 255;
     $1212 = $1209 >>> $1211;
     $1213 = $1202 >>> 8;
     $1214 = $1213 & 255;
     $1215 = (125924 + ($1214<<2)|0);
     $1216 = HEAP32[$1215>>2]|0;
     $1217 = (($1216) + ($1212)|0);
     $1218 = HEAP8[$1217>>0]|0;
     $1219 = HEAP32[$35>>2]|0;
     $1220 = $1219 & $1196;
     $1221 = $1199 >>> 16;
     $1222 = $1221 & 255;
     $1223 = $1220 >>> $1222;
     $1224 = $1202 >>> 16;
     $1225 = $1224 & 255;
     $1226 = (125924 + ($1225<<2)|0);
     $1227 = HEAP32[$1226>>2]|0;
     $1228 = (($1227) + ($1223)|0);
     $1229 = HEAP8[$1228>>0]|0;
     $Pixel$7 = $1196;$sB$7$shrunk = $1229;$sG$7$shrunk = $1218;$sR$7$shrunk = $1207;
     break;
    }
    default: {
     $Pixel$7 = 0;$sB$7$shrunk = 0;$sG$7$shrunk = 0;$sR$7$shrunk = 0;
    }
    }
    $1230 = ($Pixel$7|0)==($16|0);
    do {
     if (!($1230)) {
      $sB$7 = $sB$7$shrunk&255;
      $sG$7 = $sG$7$shrunk&255;
      $sR$7 = $sR$7$shrunk&255;
      $1231 = HEAP8[$dst$8>>0]|0;
      $1232 = $1231&255;
      $1233 = HEAP32[$29>>2]|0;
      $1234 = ((($1233)) + 4|0);
      $1235 = HEAP32[$1234>>2]|0;
      $1236 = (($1235) + ($1232<<2)|0);
      $1237 = HEAP8[$1236>>0]|0;
      $1238 = $1237&255;
      $1239 = (((($1235) + ($1232<<2)|0)) + 1|0);
      $1240 = HEAP8[$1239>>0]|0;
      $1241 = $1240&255;
      $1242 = (((($1235) + ($1232<<2)|0)) + 2|0);
      $1243 = HEAP8[$1242>>0]|0;
      $1244 = $1243&255;
      $1245 = (($sR$7) - ($1238))|0;
      $1246 = Math_imul($1245, $19)|0;
      $1247 = (($1246>>>0) / 255)&-1;
      $1248 = (($1247) + ($1238))|0;
      $1249 = (($sG$7) - ($1241))|0;
      $1250 = Math_imul($1249, $19)|0;
      $1251 = (($1250>>>0) / 255)&-1;
      $1252 = (($1251) + ($1241))|0;
      $1253 = (($sB$7) - ($1244))|0;
      $1254 = Math_imul($1253, $19)|0;
      $1255 = (($1254>>>0) / 255)&-1;
      $1256 = (($1255) + ($1244))|0;
      $1257 = $1248 & 224;
      $1258 = $1252 >>> 3;
      $1259 = $1258 & 28;
      $1260 = $1259 | $1257;
      $1261 = $1256 >>> 6;
      $1262 = $1261 & 3;
      $1263 = $1260 | $1262;
      if ($30) {
       $1264 = $1263&255;
       HEAP8[$dst$8>>0] = $1264;
       break;
      } else {
       $1265 = (($9) + ($1263)|0);
       $1266 = HEAP8[$1265>>0]|0;
       HEAP8[$dst$8>>0] = $1266;
       break;
      }
     }
    } while(0);
    $1267 = ((($dst$8)) + 1|0);
    $1268 = (($src$8) + ($14)|0);
    $1269 = (($n$7) + -1)|0;
    $1270 = ($n$7|0)>(1);
    if ($1270) {
     $dst$1 = $1267;$n$0 = $1269;$src$1 = $1268;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $1271 = (($1268) + ($5)|0);
  $1272 = (($1267) + ($7)|0);
  $1273 = ($36|0)==(0);
  if ($1273) {
   label = 86;
   break;
  } else {
   $$in = $36;$dst$024 = $1272;$src$025 = $1271;
  }
 }
 if ((label|0) == 84) {
  // unreachable;
 }
 else if ((label|0) == 86) {
  return;
 }
}
function _BlitNtoNSurfaceAlphaKey($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0;
 var $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0;
 var $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0;
 var $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0;
 var $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0;
 var $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0;
 var $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0;
 var $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0;
 var $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0;
 var $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0;
 var $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0;
 var $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0;
 var $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0;
 var $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0;
 var $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0;
 var $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0;
 var $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0;
 var $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0;
 var $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0;
 var $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0;
 var $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0;
 var $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0;
 var $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0;
 var $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0;
 var $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0;
 var $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0;
 var $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0;
 var $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0;
 var $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0;
 var $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0;
 var $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0;
 var $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0;
 var $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0;
 var $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0;
 var $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0;
 var $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0;
 var $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0;
 var $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0;
 var $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0;
 var $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0;
 var $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0;
 var $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0;
 var $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0;
 var $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0;
 var $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0;
 var $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0;
 var $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0;
 var $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0;
 var $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0;
 var $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0;
 var $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0;
 var $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0;
 var $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0;
 var $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0;
 var $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0;
 var $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0;
 var $997 = 0, $998 = 0, $999 = 0, $Pixel$0 = 0, $Pixel$1 = 0, $Pixel$2 = 0, $Pixel$3 = 0, $dA$0 = 0, $dA$0$shrunk = 0, $dA$1 = 0, $dA$1$shrunk = 0, $dA$2 = 0, $dA$2$shrunk = 0, $dA$3 = 0, $dA$3$shrunk = 0, $dB$0 = 0, $dB$0$shrunk = 0, $dB$1 = 0, $dB$1$shrunk = 0, $dB$2 = 0;
 var $dB$2$shrunk = 0, $dB$3 = 0, $dB$3$shrunk = 0, $dG$0 = 0, $dG$0$shrunk = 0, $dG$1 = 0, $dG$1$shrunk = 0, $dG$2 = 0, $dG$2$shrunk = 0, $dG$3 = 0, $dG$3$shrunk = 0, $dR$0 = 0, $dR$0$shrunk = 0, $dR$1 = 0, $dR$1$shrunk = 0, $dR$2 = 0, $dR$2$shrunk = 0, $dR$3 = 0, $dR$3$shrunk = 0, $div = 0;
 var $div1 = 0, $div10 = 0, $div12 = 0, $div13 = 0, $div14 = 0, $div15 = 0, $div16 = 0, $div18 = 0, $div19 = 0, $div2 = 0, $div20 = 0, $div21 = 0, $div22 = 0, $div3 = 0, $div4 = 0, $div6 = 0, $div7 = 0, $div8 = 0, $div9 = 0, $dst$030 = 0;
 var $dst$1 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $or$cond = 0, $or$cond24 = 0, $or$cond25 = 0, $or$cond26 = 0, $src$031 = 0, $src$1 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 40|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($info)) + 44|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($info)) + 56|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($9)) + 9|0);
 $15 = HEAP8[$14>>0]|0;
 $16 = $15&255;
 $17 = ((($11)) + 9|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = $18&255;
 $20 = ((($info)) + 63|0);
 $21 = HEAP8[$20>>0]|0;
 $22 = $21&255;
 $23 = ($3|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($info)) + 20|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = HEAP32[$info>>2]|0;
 $27 = (($1) + 3)|0;
 $28 = (($27|0) / 4)&-1;
 $29 = $1 & 3;
 $30 = ($21<<24>>24)==(0);
 $31 = ((($9)) + 12|0);
 $32 = ((($9)) + 32|0);
 $33 = ((($9)) + 28|0);
 $34 = ((($9)) + 16|0);
 $35 = ((($9)) + 20|0);
 $36 = ((($11)) + 28|0);
 $37 = ((($11)) + 32|0);
 $38 = ((($11)) + 33|0);
 $39 = ((($11)) + 34|0);
 $40 = ((($11)) + 12|0);
 $41 = ((($11)) + 16|0);
 $42 = ((($11)) + 20|0);
 $43 = ((($11)) + 24|0);
 $$in = $3;$dst$030 = $25;$src$031 = $26;
 L4: while(1) {
  $44 = (($$in) + -1)|0;
  switch ($29|0) {
  case 0:  {
   $dst$1 = $dst$030;$n$0 = $28;$src$1 = $src$031;
   label = 4;
   break;
  }
  case 3:  {
   $dst$2 = $dst$030;$n$1 = $28;$src$2 = $src$031;
   label = 21;
   break;
  }
  case 2:  {
   $dst$3 = $dst$030;$n$2 = $28;$src$3 = $src$031;
   label = 38;
   break;
  }
  case 1:  {
   $dst$4 = $dst$030;$n$3 = $28;$src$4 = $src$031;
   label = 55;
   break;
  }
  default: {
   label = 72;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    switch ($16|0) {
    case 1:  {
     $45 = HEAP8[$src$1>>0]|0;
     $46 = $45&255;
     $Pixel$0 = $46;
     break;
    }
    case 2:  {
     $47 = HEAP16[$src$1>>1]|0;
     $48 = $47&65535;
     $Pixel$0 = $48;
     break;
    }
    case 3:  {
     $49 = HEAP8[$src$1>>0]|0;
     $50 = $49&255;
     $51 = ((($src$1)) + 1|0);
     $52 = HEAP8[$51>>0]|0;
     $53 = $52&255;
     $54 = $53 << 8;
     $55 = $54 | $50;
     $56 = ((($src$1)) + 2|0);
     $57 = HEAP8[$56>>0]|0;
     $58 = $57&255;
     $59 = $58 << 16;
     $60 = $55 | $59;
     $Pixel$0 = $60;
     break;
    }
    case 4:  {
     $61 = HEAP32[$src$1>>2]|0;
     $Pixel$0 = $61;
     break;
    }
    default: {
     $Pixel$0 = 0;
    }
    }
    $62 = ($Pixel$0|0)==($13|0);
    $or$cond = $30 | $62;
    L15: do {
     if (!($or$cond)) {
      $63 = HEAP32[$31>>2]|0;
      $64 = $63 & $Pixel$0;
      $65 = HEAP32[$32>>2]|0;
      $66 = $65 & 255;
      $67 = $64 >>> $66;
      $68 = HEAP32[$33>>2]|0;
      $69 = $68 & 255;
      $70 = (125924 + ($69<<2)|0);
      $71 = HEAP32[$70>>2]|0;
      $72 = (($71) + ($67)|0);
      $73 = HEAP8[$72>>0]|0;
      $74 = $73&255;
      $75 = HEAP32[$34>>2]|0;
      $76 = $75 & $Pixel$0;
      $77 = $65 >>> 8;
      $78 = $77 & 255;
      $79 = $76 >>> $78;
      $80 = $68 >>> 8;
      $81 = $80 & 255;
      $82 = (125924 + ($81<<2)|0);
      $83 = HEAP32[$82>>2]|0;
      $84 = (($83) + ($79)|0);
      $85 = HEAP8[$84>>0]|0;
      $86 = $85&255;
      $87 = HEAP32[$35>>2]|0;
      $88 = $87 & $Pixel$0;
      $89 = $65 >>> 16;
      $90 = $89 & 255;
      $91 = $88 >>> $90;
      $92 = $68 >>> 16;
      $93 = $92 & 255;
      $94 = (125924 + ($93<<2)|0);
      $95 = HEAP32[$94>>2]|0;
      $96 = (($95) + ($91)|0);
      $97 = HEAP8[$96>>0]|0;
      $98 = $97&255;
      switch ($19|0) {
      case 1:  {
       $99 = HEAP8[$dst$1>>0]|0;
       $100 = $99&255;
       $101 = HEAP32[$40>>2]|0;
       $102 = $100 & $101;
       $103 = HEAP32[$37>>2]|0;
       $104 = $103 & 255;
       $105 = $102 >>> $104;
       $106 = HEAP32[$36>>2]|0;
       $107 = $106 & 255;
       $108 = (125924 + ($107<<2)|0);
       $109 = HEAP32[$108>>2]|0;
       $110 = (($109) + ($105)|0);
       $111 = HEAP8[$110>>0]|0;
       $112 = HEAP32[$41>>2]|0;
       $113 = $112 & $100;
       $114 = $103 >>> 8;
       $115 = $114 & 255;
       $116 = $113 >>> $115;
       $117 = $106 >>> 8;
       $118 = $117 & 255;
       $119 = (125924 + ($118<<2)|0);
       $120 = HEAP32[$119>>2]|0;
       $121 = (($120) + ($116)|0);
       $122 = HEAP8[$121>>0]|0;
       $123 = HEAP32[$42>>2]|0;
       $124 = $123 & $100;
       $125 = $103 >>> 16;
       $126 = $125 & 255;
       $127 = $124 >>> $126;
       $128 = $106 >>> 16;
       $129 = $128 & 255;
       $130 = (125924 + ($129<<2)|0);
       $131 = HEAP32[$130>>2]|0;
       $132 = (($131) + ($127)|0);
       $133 = HEAP8[$132>>0]|0;
       $134 = HEAP32[$43>>2]|0;
       $135 = $134 & $100;
       $136 = $103 >>> 24;
       $137 = $135 >>> $136;
       $138 = $106 >>> 24;
       $139 = (125924 + ($138<<2)|0);
       $140 = HEAP32[$139>>2]|0;
       $141 = (($140) + ($137)|0);
       $142 = HEAP8[$141>>0]|0;
       $dA$0$shrunk = $142;$dB$0$shrunk = $133;$dG$0$shrunk = $122;$dR$0$shrunk = $111;
       break;
      }
      case 2:  {
       $143 = HEAP16[$dst$1>>1]|0;
       $144 = $143&65535;
       $145 = HEAP32[$40>>2]|0;
       $146 = $144 & $145;
       $147 = HEAP32[$37>>2]|0;
       $148 = $147 & 255;
       $149 = $146 >>> $148;
       $150 = HEAP32[$36>>2]|0;
       $151 = $150 & 255;
       $152 = (125924 + ($151<<2)|0);
       $153 = HEAP32[$152>>2]|0;
       $154 = (($153) + ($149)|0);
       $155 = HEAP8[$154>>0]|0;
       $156 = HEAP32[$41>>2]|0;
       $157 = $156 & $144;
       $158 = $147 >>> 8;
       $159 = $158 & 255;
       $160 = $157 >>> $159;
       $161 = $150 >>> 8;
       $162 = $161 & 255;
       $163 = (125924 + ($162<<2)|0);
       $164 = HEAP32[$163>>2]|0;
       $165 = (($164) + ($160)|0);
       $166 = HEAP8[$165>>0]|0;
       $167 = HEAP32[$42>>2]|0;
       $168 = $167 & $144;
       $169 = $147 >>> 16;
       $170 = $169 & 255;
       $171 = $168 >>> $170;
       $172 = $150 >>> 16;
       $173 = $172 & 255;
       $174 = (125924 + ($173<<2)|0);
       $175 = HEAP32[$174>>2]|0;
       $176 = (($175) + ($171)|0);
       $177 = HEAP8[$176>>0]|0;
       $178 = HEAP32[$43>>2]|0;
       $179 = $178 & $144;
       $180 = $147 >>> 24;
       $181 = $179 >>> $180;
       $182 = $150 >>> 24;
       $183 = (125924 + ($182<<2)|0);
       $184 = HEAP32[$183>>2]|0;
       $185 = (($184) + ($181)|0);
       $186 = HEAP8[$185>>0]|0;
       $dA$0$shrunk = $186;$dB$0$shrunk = $177;$dG$0$shrunk = $166;$dR$0$shrunk = $155;
       break;
      }
      case 3:  {
       $187 = HEAP32[$37>>2]|0;
       $div15 = $187 >>> 3;
       $188 = $div15 & 31;
       $189 = (($dst$1) + ($188)|0);
       $190 = HEAP8[$189>>0]|0;
       $191 = $187 >>> 11;
       $div16 = $191 & 31;
       $192 = (($dst$1) + ($div16)|0);
       $193 = HEAP8[$192>>0]|0;
       $194 = $187 >>> 19;
       $195 = $194 & 31;
       $196 = (($dst$1) + ($195)|0);
       $197 = HEAP8[$196>>0]|0;
       $dA$0$shrunk = -1;$dB$0$shrunk = $197;$dG$0$shrunk = $193;$dR$0$shrunk = $190;
       break;
      }
      case 4:  {
       $198 = HEAP32[$dst$1>>2]|0;
       $199 = HEAP32[$40>>2]|0;
       $200 = $199 & $198;
       $201 = HEAP32[$37>>2]|0;
       $202 = $201 & 255;
       $203 = $200 >>> $202;
       $204 = HEAP32[$36>>2]|0;
       $205 = $204 & 255;
       $206 = (125924 + ($205<<2)|0);
       $207 = HEAP32[$206>>2]|0;
       $208 = (($207) + ($203)|0);
       $209 = HEAP8[$208>>0]|0;
       $210 = HEAP32[$41>>2]|0;
       $211 = $210 & $198;
       $212 = $201 >>> 8;
       $213 = $212 & 255;
       $214 = $211 >>> $213;
       $215 = $204 >>> 8;
       $216 = $215 & 255;
       $217 = (125924 + ($216<<2)|0);
       $218 = HEAP32[$217>>2]|0;
       $219 = (($218) + ($214)|0);
       $220 = HEAP8[$219>>0]|0;
       $221 = HEAP32[$42>>2]|0;
       $222 = $221 & $198;
       $223 = $201 >>> 16;
       $224 = $223 & 255;
       $225 = $222 >>> $224;
       $226 = $204 >>> 16;
       $227 = $226 & 255;
       $228 = (125924 + ($227<<2)|0);
       $229 = HEAP32[$228>>2]|0;
       $230 = (($229) + ($225)|0);
       $231 = HEAP8[$230>>0]|0;
       $232 = HEAP32[$43>>2]|0;
       $233 = $232 & $198;
       $234 = $201 >>> 24;
       $235 = $233 >>> $234;
       $236 = $204 >>> 24;
       $237 = (125924 + ($236<<2)|0);
       $238 = HEAP32[$237>>2]|0;
       $239 = (($238) + ($235)|0);
       $240 = HEAP8[$239>>0]|0;
       $dA$0$shrunk = $240;$dB$0$shrunk = $231;$dG$0$shrunk = $220;$dR$0$shrunk = $209;
       break;
      }
      default: {
       $dA$0$shrunk = 0;$dB$0$shrunk = 0;$dG$0$shrunk = 0;$dR$0$shrunk = 0;
      }
      }
      $dR$0 = $dR$0$shrunk&255;
      $dG$0 = $dG$0$shrunk&255;
      $dB$0 = $dB$0$shrunk&255;
      $dA$0 = $dA$0$shrunk&255;
      $241 = (($74) - ($dR$0))|0;
      $242 = Math_imul($241, $22)|0;
      $243 = (($242>>>0) / 255)&-1;
      $244 = (($243) + ($dR$0))|0;
      $245 = (($86) - ($dG$0))|0;
      $246 = Math_imul($245, $22)|0;
      $247 = (($246>>>0) / 255)&-1;
      $248 = (($247) + ($dG$0))|0;
      $249 = (($98) - ($dB$0))|0;
      $250 = Math_imul($249, $22)|0;
      $251 = (($250>>>0) / 255)&-1;
      $252 = (($251) + ($dB$0))|0;
      $253 = (($dA$0) + ($22))|0;
      $254 = Math_imul($dA$0, $22)|0;
      $255 = (($254>>>0) / 255)&-1;
      $256 = (($253) - ($255))|0;
      switch ($19|0) {
      case 1:  {
       $257 = HEAP32[$36>>2]|0;
       $258 = $257 & 255;
       $259 = $244 >>> $258;
       $260 = HEAP32[$37>>2]|0;
       $261 = $260 & 255;
       $262 = $259 << $261;
       $263 = $257 >>> 8;
       $264 = $263 & 255;
       $265 = $248 >>> $264;
       $266 = $260 >>> 8;
       $267 = $266 & 255;
       $268 = $265 << $267;
       $269 = $268 | $262;
       $270 = $257 >>> 16;
       $271 = $270 & 255;
       $272 = $252 >>> $271;
       $273 = $260 >>> 16;
       $274 = $273 & 255;
       $275 = $272 << $274;
       $276 = $269 | $275;
       $277 = $257 >>> 24;
       $278 = $256 >>> $277;
       $279 = $260 >>> 24;
       $280 = $278 << $279;
       $281 = $276 | $280;
       $282 = $281&255;
       HEAP8[$dst$1>>0] = $282;
       break L15;
       break;
      }
      case 2:  {
       $283 = HEAP32[$36>>2]|0;
       $284 = $283 & 255;
       $285 = $244 >>> $284;
       $286 = HEAP32[$37>>2]|0;
       $287 = $286 & 255;
       $288 = $285 << $287;
       $289 = $283 >>> 8;
       $290 = $289 & 255;
       $291 = $248 >>> $290;
       $292 = $286 >>> 8;
       $293 = $292 & 255;
       $294 = $291 << $293;
       $295 = $294 | $288;
       $296 = $283 >>> 16;
       $297 = $296 & 255;
       $298 = $252 >>> $297;
       $299 = $286 >>> 16;
       $300 = $299 & 255;
       $301 = $298 << $300;
       $302 = $295 | $301;
       $303 = $283 >>> 24;
       $304 = $256 >>> $303;
       $305 = $286 >>> 24;
       $306 = $304 << $305;
       $307 = $302 | $306;
       $308 = $307&65535;
       HEAP16[$dst$1>>1] = $308;
       break L15;
       break;
      }
      case 3:  {
       $309 = $244&255;
       $310 = HEAP8[$37>>0]|0;
       $div12 = ($310&255) >>> 3;
       $311 = $div12&255;
       $312 = (($dst$1) + ($311)|0);
       HEAP8[$312>>0] = $309;
       $313 = $248&255;
       $314 = HEAP8[$38>>0]|0;
       $div13 = ($314&255) >>> 3;
       $315 = $div13&255;
       $316 = (($dst$1) + ($315)|0);
       HEAP8[$316>>0] = $313;
       $317 = $252&255;
       $318 = HEAP8[$39>>0]|0;
       $div14 = ($318&255) >>> 3;
       $319 = $div14&255;
       $320 = (($dst$1) + ($319)|0);
       HEAP8[$320>>0] = $317;
       break L15;
       break;
      }
      case 4:  {
       $321 = HEAP32[$36>>2]|0;
       $322 = $321 & 255;
       $323 = $244 >>> $322;
       $324 = HEAP32[$37>>2]|0;
       $325 = $324 & 255;
       $326 = $323 << $325;
       $327 = $321 >>> 8;
       $328 = $327 & 255;
       $329 = $248 >>> $328;
       $330 = $324 >>> 8;
       $331 = $330 & 255;
       $332 = $329 << $331;
       $333 = $332 | $326;
       $334 = $321 >>> 16;
       $335 = $334 & 255;
       $336 = $252 >>> $335;
       $337 = $324 >>> 16;
       $338 = $337 & 255;
       $339 = $336 << $338;
       $340 = $333 | $339;
       $341 = $321 >>> 24;
       $342 = $256 >>> $341;
       $343 = $324 >>> 24;
       $344 = $342 << $343;
       $345 = $340 | $344;
       HEAP32[$dst$1>>2] = $345;
       break L15;
       break;
      }
      default: {
       break L15;
      }
      }
     }
    } while(0);
    $346 = (($src$1) + ($16)|0);
    $347 = (($dst$1) + ($19)|0);
    $dst$2 = $347;$n$1 = $n$0;$src$2 = $346;
    label = 21;
    continue;
   }
   else if ((label|0) == 21) {
    label = 0;
    switch ($16|0) {
    case 1:  {
     $348 = HEAP8[$src$2>>0]|0;
     $349 = $348&255;
     $Pixel$1 = $349;
     break;
    }
    case 2:  {
     $350 = HEAP16[$src$2>>1]|0;
     $351 = $350&65535;
     $Pixel$1 = $351;
     break;
    }
    case 3:  {
     $352 = HEAP8[$src$2>>0]|0;
     $353 = $352&255;
     $354 = ((($src$2)) + 1|0);
     $355 = HEAP8[$354>>0]|0;
     $356 = $355&255;
     $357 = $356 << 8;
     $358 = $357 | $353;
     $359 = ((($src$2)) + 2|0);
     $360 = HEAP8[$359>>0]|0;
     $361 = $360&255;
     $362 = $361 << 16;
     $363 = $358 | $362;
     $Pixel$1 = $363;
     break;
    }
    case 4:  {
     $364 = HEAP32[$src$2>>2]|0;
     $Pixel$1 = $364;
     break;
    }
    default: {
     $Pixel$1 = 0;
    }
    }
    $365 = ($Pixel$1|0)==($13|0);
    $or$cond24 = $30 | $365;
    L36: do {
     if (!($or$cond24)) {
      $366 = HEAP32[$31>>2]|0;
      $367 = $366 & $Pixel$1;
      $368 = HEAP32[$32>>2]|0;
      $369 = $368 & 255;
      $370 = $367 >>> $369;
      $371 = HEAP32[$33>>2]|0;
      $372 = $371 & 255;
      $373 = (125924 + ($372<<2)|0);
      $374 = HEAP32[$373>>2]|0;
      $375 = (($374) + ($370)|0);
      $376 = HEAP8[$375>>0]|0;
      $377 = $376&255;
      $378 = HEAP32[$34>>2]|0;
      $379 = $378 & $Pixel$1;
      $380 = $368 >>> 8;
      $381 = $380 & 255;
      $382 = $379 >>> $381;
      $383 = $371 >>> 8;
      $384 = $383 & 255;
      $385 = (125924 + ($384<<2)|0);
      $386 = HEAP32[$385>>2]|0;
      $387 = (($386) + ($382)|0);
      $388 = HEAP8[$387>>0]|0;
      $389 = $388&255;
      $390 = HEAP32[$35>>2]|0;
      $391 = $390 & $Pixel$1;
      $392 = $368 >>> 16;
      $393 = $392 & 255;
      $394 = $391 >>> $393;
      $395 = $371 >>> 16;
      $396 = $395 & 255;
      $397 = (125924 + ($396<<2)|0);
      $398 = HEAP32[$397>>2]|0;
      $399 = (($398) + ($394)|0);
      $400 = HEAP8[$399>>0]|0;
      $401 = $400&255;
      switch ($19|0) {
      case 1:  {
       $402 = HEAP8[$dst$2>>0]|0;
       $403 = $402&255;
       $404 = HEAP32[$40>>2]|0;
       $405 = $403 & $404;
       $406 = HEAP32[$37>>2]|0;
       $407 = $406 & 255;
       $408 = $405 >>> $407;
       $409 = HEAP32[$36>>2]|0;
       $410 = $409 & 255;
       $411 = (125924 + ($410<<2)|0);
       $412 = HEAP32[$411>>2]|0;
       $413 = (($412) + ($408)|0);
       $414 = HEAP8[$413>>0]|0;
       $415 = HEAP32[$41>>2]|0;
       $416 = $415 & $403;
       $417 = $406 >>> 8;
       $418 = $417 & 255;
       $419 = $416 >>> $418;
       $420 = $409 >>> 8;
       $421 = $420 & 255;
       $422 = (125924 + ($421<<2)|0);
       $423 = HEAP32[$422>>2]|0;
       $424 = (($423) + ($419)|0);
       $425 = HEAP8[$424>>0]|0;
       $426 = HEAP32[$42>>2]|0;
       $427 = $426 & $403;
       $428 = $406 >>> 16;
       $429 = $428 & 255;
       $430 = $427 >>> $429;
       $431 = $409 >>> 16;
       $432 = $431 & 255;
       $433 = (125924 + ($432<<2)|0);
       $434 = HEAP32[$433>>2]|0;
       $435 = (($434) + ($430)|0);
       $436 = HEAP8[$435>>0]|0;
       $437 = HEAP32[$43>>2]|0;
       $438 = $437 & $403;
       $439 = $406 >>> 24;
       $440 = $438 >>> $439;
       $441 = $409 >>> 24;
       $442 = (125924 + ($441<<2)|0);
       $443 = HEAP32[$442>>2]|0;
       $444 = (($443) + ($440)|0);
       $445 = HEAP8[$444>>0]|0;
       $dA$1$shrunk = $445;$dB$1$shrunk = $436;$dG$1$shrunk = $425;$dR$1$shrunk = $414;
       break;
      }
      case 2:  {
       $446 = HEAP16[$dst$2>>1]|0;
       $447 = $446&65535;
       $448 = HEAP32[$40>>2]|0;
       $449 = $447 & $448;
       $450 = HEAP32[$37>>2]|0;
       $451 = $450 & 255;
       $452 = $449 >>> $451;
       $453 = HEAP32[$36>>2]|0;
       $454 = $453 & 255;
       $455 = (125924 + ($454<<2)|0);
       $456 = HEAP32[$455>>2]|0;
       $457 = (($456) + ($452)|0);
       $458 = HEAP8[$457>>0]|0;
       $459 = HEAP32[$41>>2]|0;
       $460 = $459 & $447;
       $461 = $450 >>> 8;
       $462 = $461 & 255;
       $463 = $460 >>> $462;
       $464 = $453 >>> 8;
       $465 = $464 & 255;
       $466 = (125924 + ($465<<2)|0);
       $467 = HEAP32[$466>>2]|0;
       $468 = (($467) + ($463)|0);
       $469 = HEAP8[$468>>0]|0;
       $470 = HEAP32[$42>>2]|0;
       $471 = $470 & $447;
       $472 = $450 >>> 16;
       $473 = $472 & 255;
       $474 = $471 >>> $473;
       $475 = $453 >>> 16;
       $476 = $475 & 255;
       $477 = (125924 + ($476<<2)|0);
       $478 = HEAP32[$477>>2]|0;
       $479 = (($478) + ($474)|0);
       $480 = HEAP8[$479>>0]|0;
       $481 = HEAP32[$43>>2]|0;
       $482 = $481 & $447;
       $483 = $450 >>> 24;
       $484 = $482 >>> $483;
       $485 = $453 >>> 24;
       $486 = (125924 + ($485<<2)|0);
       $487 = HEAP32[$486>>2]|0;
       $488 = (($487) + ($484)|0);
       $489 = HEAP8[$488>>0]|0;
       $dA$1$shrunk = $489;$dB$1$shrunk = $480;$dG$1$shrunk = $469;$dR$1$shrunk = $458;
       break;
      }
      case 3:  {
       $490 = HEAP32[$37>>2]|0;
       $div9 = $490 >>> 3;
       $491 = $div9 & 31;
       $492 = (($dst$2) + ($491)|0);
       $493 = HEAP8[$492>>0]|0;
       $494 = $490 >>> 11;
       $div10 = $494 & 31;
       $495 = (($dst$2) + ($div10)|0);
       $496 = HEAP8[$495>>0]|0;
       $497 = $490 >>> 19;
       $498 = $497 & 31;
       $499 = (($dst$2) + ($498)|0);
       $500 = HEAP8[$499>>0]|0;
       $dA$1$shrunk = -1;$dB$1$shrunk = $500;$dG$1$shrunk = $496;$dR$1$shrunk = $493;
       break;
      }
      case 4:  {
       $501 = HEAP32[$dst$2>>2]|0;
       $502 = HEAP32[$40>>2]|0;
       $503 = $502 & $501;
       $504 = HEAP32[$37>>2]|0;
       $505 = $504 & 255;
       $506 = $503 >>> $505;
       $507 = HEAP32[$36>>2]|0;
       $508 = $507 & 255;
       $509 = (125924 + ($508<<2)|0);
       $510 = HEAP32[$509>>2]|0;
       $511 = (($510) + ($506)|0);
       $512 = HEAP8[$511>>0]|0;
       $513 = HEAP32[$41>>2]|0;
       $514 = $513 & $501;
       $515 = $504 >>> 8;
       $516 = $515 & 255;
       $517 = $514 >>> $516;
       $518 = $507 >>> 8;
       $519 = $518 & 255;
       $520 = (125924 + ($519<<2)|0);
       $521 = HEAP32[$520>>2]|0;
       $522 = (($521) + ($517)|0);
       $523 = HEAP8[$522>>0]|0;
       $524 = HEAP32[$42>>2]|0;
       $525 = $524 & $501;
       $526 = $504 >>> 16;
       $527 = $526 & 255;
       $528 = $525 >>> $527;
       $529 = $507 >>> 16;
       $530 = $529 & 255;
       $531 = (125924 + ($530<<2)|0);
       $532 = HEAP32[$531>>2]|0;
       $533 = (($532) + ($528)|0);
       $534 = HEAP8[$533>>0]|0;
       $535 = HEAP32[$43>>2]|0;
       $536 = $535 & $501;
       $537 = $504 >>> 24;
       $538 = $536 >>> $537;
       $539 = $507 >>> 24;
       $540 = (125924 + ($539<<2)|0);
       $541 = HEAP32[$540>>2]|0;
       $542 = (($541) + ($538)|0);
       $543 = HEAP8[$542>>0]|0;
       $dA$1$shrunk = $543;$dB$1$shrunk = $534;$dG$1$shrunk = $523;$dR$1$shrunk = $512;
       break;
      }
      default: {
       $dA$1$shrunk = 0;$dB$1$shrunk = 0;$dG$1$shrunk = 0;$dR$1$shrunk = 0;
      }
      }
      $dR$1 = $dR$1$shrunk&255;
      $dG$1 = $dG$1$shrunk&255;
      $dB$1 = $dB$1$shrunk&255;
      $dA$1 = $dA$1$shrunk&255;
      $544 = (($377) - ($dR$1))|0;
      $545 = Math_imul($544, $22)|0;
      $546 = (($545>>>0) / 255)&-1;
      $547 = (($546) + ($dR$1))|0;
      $548 = (($389) - ($dG$1))|0;
      $549 = Math_imul($548, $22)|0;
      $550 = (($549>>>0) / 255)&-1;
      $551 = (($550) + ($dG$1))|0;
      $552 = (($401) - ($dB$1))|0;
      $553 = Math_imul($552, $22)|0;
      $554 = (($553>>>0) / 255)&-1;
      $555 = (($554) + ($dB$1))|0;
      $556 = (($dA$1) + ($22))|0;
      $557 = Math_imul($dA$1, $22)|0;
      $558 = (($557>>>0) / 255)&-1;
      $559 = (($556) - ($558))|0;
      switch ($19|0) {
      case 1:  {
       $560 = HEAP32[$36>>2]|0;
       $561 = $560 & 255;
       $562 = $547 >>> $561;
       $563 = HEAP32[$37>>2]|0;
       $564 = $563 & 255;
       $565 = $562 << $564;
       $566 = $560 >>> 8;
       $567 = $566 & 255;
       $568 = $551 >>> $567;
       $569 = $563 >>> 8;
       $570 = $569 & 255;
       $571 = $568 << $570;
       $572 = $571 | $565;
       $573 = $560 >>> 16;
       $574 = $573 & 255;
       $575 = $555 >>> $574;
       $576 = $563 >>> 16;
       $577 = $576 & 255;
       $578 = $575 << $577;
       $579 = $572 | $578;
       $580 = $560 >>> 24;
       $581 = $559 >>> $580;
       $582 = $563 >>> 24;
       $583 = $581 << $582;
       $584 = $579 | $583;
       $585 = $584&255;
       HEAP8[$dst$2>>0] = $585;
       break L36;
       break;
      }
      case 2:  {
       $586 = HEAP32[$36>>2]|0;
       $587 = $586 & 255;
       $588 = $547 >>> $587;
       $589 = HEAP32[$37>>2]|0;
       $590 = $589 & 255;
       $591 = $588 << $590;
       $592 = $586 >>> 8;
       $593 = $592 & 255;
       $594 = $551 >>> $593;
       $595 = $589 >>> 8;
       $596 = $595 & 255;
       $597 = $594 << $596;
       $598 = $597 | $591;
       $599 = $586 >>> 16;
       $600 = $599 & 255;
       $601 = $555 >>> $600;
       $602 = $589 >>> 16;
       $603 = $602 & 255;
       $604 = $601 << $603;
       $605 = $598 | $604;
       $606 = $586 >>> 24;
       $607 = $559 >>> $606;
       $608 = $589 >>> 24;
       $609 = $607 << $608;
       $610 = $605 | $609;
       $611 = $610&65535;
       HEAP16[$dst$2>>1] = $611;
       break L36;
       break;
      }
      case 3:  {
       $612 = $547&255;
       $613 = HEAP8[$37>>0]|0;
       $div6 = ($613&255) >>> 3;
       $614 = $div6&255;
       $615 = (($dst$2) + ($614)|0);
       HEAP8[$615>>0] = $612;
       $616 = $551&255;
       $617 = HEAP8[$38>>0]|0;
       $div7 = ($617&255) >>> 3;
       $618 = $div7&255;
       $619 = (($dst$2) + ($618)|0);
       HEAP8[$619>>0] = $616;
       $620 = $555&255;
       $621 = HEAP8[$39>>0]|0;
       $div8 = ($621&255) >>> 3;
       $622 = $div8&255;
       $623 = (($dst$2) + ($622)|0);
       HEAP8[$623>>0] = $620;
       break L36;
       break;
      }
      case 4:  {
       $624 = HEAP32[$36>>2]|0;
       $625 = $624 & 255;
       $626 = $547 >>> $625;
       $627 = HEAP32[$37>>2]|0;
       $628 = $627 & 255;
       $629 = $626 << $628;
       $630 = $624 >>> 8;
       $631 = $630 & 255;
       $632 = $551 >>> $631;
       $633 = $627 >>> 8;
       $634 = $633 & 255;
       $635 = $632 << $634;
       $636 = $635 | $629;
       $637 = $624 >>> 16;
       $638 = $637 & 255;
       $639 = $555 >>> $638;
       $640 = $627 >>> 16;
       $641 = $640 & 255;
       $642 = $639 << $641;
       $643 = $636 | $642;
       $644 = $624 >>> 24;
       $645 = $559 >>> $644;
       $646 = $627 >>> 24;
       $647 = $645 << $646;
       $648 = $643 | $647;
       HEAP32[$dst$2>>2] = $648;
       break L36;
       break;
      }
      default: {
       break L36;
      }
      }
     }
    } while(0);
    $649 = (($src$2) + ($16)|0);
    $650 = (($dst$2) + ($19)|0);
    $dst$3 = $650;$n$2 = $n$1;$src$3 = $649;
    label = 38;
    continue;
   }
   else if ((label|0) == 38) {
    label = 0;
    switch ($16|0) {
    case 1:  {
     $651 = HEAP8[$src$3>>0]|0;
     $652 = $651&255;
     $Pixel$2 = $652;
     break;
    }
    case 2:  {
     $653 = HEAP16[$src$3>>1]|0;
     $654 = $653&65535;
     $Pixel$2 = $654;
     break;
    }
    case 3:  {
     $655 = HEAP8[$src$3>>0]|0;
     $656 = $655&255;
     $657 = ((($src$3)) + 1|0);
     $658 = HEAP8[$657>>0]|0;
     $659 = $658&255;
     $660 = $659 << 8;
     $661 = $660 | $656;
     $662 = ((($src$3)) + 2|0);
     $663 = HEAP8[$662>>0]|0;
     $664 = $663&255;
     $665 = $664 << 16;
     $666 = $661 | $665;
     $Pixel$2 = $666;
     break;
    }
    case 4:  {
     $667 = HEAP32[$src$3>>2]|0;
     $Pixel$2 = $667;
     break;
    }
    default: {
     $Pixel$2 = 0;
    }
    }
    $668 = ($Pixel$2|0)==($13|0);
    $or$cond25 = $30 | $668;
    L57: do {
     if (!($or$cond25)) {
      $669 = HEAP32[$31>>2]|0;
      $670 = $669 & $Pixel$2;
      $671 = HEAP32[$32>>2]|0;
      $672 = $671 & 255;
      $673 = $670 >>> $672;
      $674 = HEAP32[$33>>2]|0;
      $675 = $674 & 255;
      $676 = (125924 + ($675<<2)|0);
      $677 = HEAP32[$676>>2]|0;
      $678 = (($677) + ($673)|0);
      $679 = HEAP8[$678>>0]|0;
      $680 = $679&255;
      $681 = HEAP32[$34>>2]|0;
      $682 = $681 & $Pixel$2;
      $683 = $671 >>> 8;
      $684 = $683 & 255;
      $685 = $682 >>> $684;
      $686 = $674 >>> 8;
      $687 = $686 & 255;
      $688 = (125924 + ($687<<2)|0);
      $689 = HEAP32[$688>>2]|0;
      $690 = (($689) + ($685)|0);
      $691 = HEAP8[$690>>0]|0;
      $692 = $691&255;
      $693 = HEAP32[$35>>2]|0;
      $694 = $693 & $Pixel$2;
      $695 = $671 >>> 16;
      $696 = $695 & 255;
      $697 = $694 >>> $696;
      $698 = $674 >>> 16;
      $699 = $698 & 255;
      $700 = (125924 + ($699<<2)|0);
      $701 = HEAP32[$700>>2]|0;
      $702 = (($701) + ($697)|0);
      $703 = HEAP8[$702>>0]|0;
      $704 = $703&255;
      switch ($19|0) {
      case 1:  {
       $705 = HEAP8[$dst$3>>0]|0;
       $706 = $705&255;
       $707 = HEAP32[$40>>2]|0;
       $708 = $706 & $707;
       $709 = HEAP32[$37>>2]|0;
       $710 = $709 & 255;
       $711 = $708 >>> $710;
       $712 = HEAP32[$36>>2]|0;
       $713 = $712 & 255;
       $714 = (125924 + ($713<<2)|0);
       $715 = HEAP32[$714>>2]|0;
       $716 = (($715) + ($711)|0);
       $717 = HEAP8[$716>>0]|0;
       $718 = HEAP32[$41>>2]|0;
       $719 = $718 & $706;
       $720 = $709 >>> 8;
       $721 = $720 & 255;
       $722 = $719 >>> $721;
       $723 = $712 >>> 8;
       $724 = $723 & 255;
       $725 = (125924 + ($724<<2)|0);
       $726 = HEAP32[$725>>2]|0;
       $727 = (($726) + ($722)|0);
       $728 = HEAP8[$727>>0]|0;
       $729 = HEAP32[$42>>2]|0;
       $730 = $729 & $706;
       $731 = $709 >>> 16;
       $732 = $731 & 255;
       $733 = $730 >>> $732;
       $734 = $712 >>> 16;
       $735 = $734 & 255;
       $736 = (125924 + ($735<<2)|0);
       $737 = HEAP32[$736>>2]|0;
       $738 = (($737) + ($733)|0);
       $739 = HEAP8[$738>>0]|0;
       $740 = HEAP32[$43>>2]|0;
       $741 = $740 & $706;
       $742 = $709 >>> 24;
       $743 = $741 >>> $742;
       $744 = $712 >>> 24;
       $745 = (125924 + ($744<<2)|0);
       $746 = HEAP32[$745>>2]|0;
       $747 = (($746) + ($743)|0);
       $748 = HEAP8[$747>>0]|0;
       $dA$2$shrunk = $748;$dB$2$shrunk = $739;$dG$2$shrunk = $728;$dR$2$shrunk = $717;
       break;
      }
      case 2:  {
       $749 = HEAP16[$dst$3>>1]|0;
       $750 = $749&65535;
       $751 = HEAP32[$40>>2]|0;
       $752 = $750 & $751;
       $753 = HEAP32[$37>>2]|0;
       $754 = $753 & 255;
       $755 = $752 >>> $754;
       $756 = HEAP32[$36>>2]|0;
       $757 = $756 & 255;
       $758 = (125924 + ($757<<2)|0);
       $759 = HEAP32[$758>>2]|0;
       $760 = (($759) + ($755)|0);
       $761 = HEAP8[$760>>0]|0;
       $762 = HEAP32[$41>>2]|0;
       $763 = $762 & $750;
       $764 = $753 >>> 8;
       $765 = $764 & 255;
       $766 = $763 >>> $765;
       $767 = $756 >>> 8;
       $768 = $767 & 255;
       $769 = (125924 + ($768<<2)|0);
       $770 = HEAP32[$769>>2]|0;
       $771 = (($770) + ($766)|0);
       $772 = HEAP8[$771>>0]|0;
       $773 = HEAP32[$42>>2]|0;
       $774 = $773 & $750;
       $775 = $753 >>> 16;
       $776 = $775 & 255;
       $777 = $774 >>> $776;
       $778 = $756 >>> 16;
       $779 = $778 & 255;
       $780 = (125924 + ($779<<2)|0);
       $781 = HEAP32[$780>>2]|0;
       $782 = (($781) + ($777)|0);
       $783 = HEAP8[$782>>0]|0;
       $784 = HEAP32[$43>>2]|0;
       $785 = $784 & $750;
       $786 = $753 >>> 24;
       $787 = $785 >>> $786;
       $788 = $756 >>> 24;
       $789 = (125924 + ($788<<2)|0);
       $790 = HEAP32[$789>>2]|0;
       $791 = (($790) + ($787)|0);
       $792 = HEAP8[$791>>0]|0;
       $dA$2$shrunk = $792;$dB$2$shrunk = $783;$dG$2$shrunk = $772;$dR$2$shrunk = $761;
       break;
      }
      case 3:  {
       $793 = HEAP32[$37>>2]|0;
       $div3 = $793 >>> 3;
       $794 = $div3 & 31;
       $795 = (($dst$3) + ($794)|0);
       $796 = HEAP8[$795>>0]|0;
       $797 = $793 >>> 11;
       $div4 = $797 & 31;
       $798 = (($dst$3) + ($div4)|0);
       $799 = HEAP8[$798>>0]|0;
       $800 = $793 >>> 19;
       $801 = $800 & 31;
       $802 = (($dst$3) + ($801)|0);
       $803 = HEAP8[$802>>0]|0;
       $dA$2$shrunk = -1;$dB$2$shrunk = $803;$dG$2$shrunk = $799;$dR$2$shrunk = $796;
       break;
      }
      case 4:  {
       $804 = HEAP32[$dst$3>>2]|0;
       $805 = HEAP32[$40>>2]|0;
       $806 = $805 & $804;
       $807 = HEAP32[$37>>2]|0;
       $808 = $807 & 255;
       $809 = $806 >>> $808;
       $810 = HEAP32[$36>>2]|0;
       $811 = $810 & 255;
       $812 = (125924 + ($811<<2)|0);
       $813 = HEAP32[$812>>2]|0;
       $814 = (($813) + ($809)|0);
       $815 = HEAP8[$814>>0]|0;
       $816 = HEAP32[$41>>2]|0;
       $817 = $816 & $804;
       $818 = $807 >>> 8;
       $819 = $818 & 255;
       $820 = $817 >>> $819;
       $821 = $810 >>> 8;
       $822 = $821 & 255;
       $823 = (125924 + ($822<<2)|0);
       $824 = HEAP32[$823>>2]|0;
       $825 = (($824) + ($820)|0);
       $826 = HEAP8[$825>>0]|0;
       $827 = HEAP32[$42>>2]|0;
       $828 = $827 & $804;
       $829 = $807 >>> 16;
       $830 = $829 & 255;
       $831 = $828 >>> $830;
       $832 = $810 >>> 16;
       $833 = $832 & 255;
       $834 = (125924 + ($833<<2)|0);
       $835 = HEAP32[$834>>2]|0;
       $836 = (($835) + ($831)|0);
       $837 = HEAP8[$836>>0]|0;
       $838 = HEAP32[$43>>2]|0;
       $839 = $838 & $804;
       $840 = $807 >>> 24;
       $841 = $839 >>> $840;
       $842 = $810 >>> 24;
       $843 = (125924 + ($842<<2)|0);
       $844 = HEAP32[$843>>2]|0;
       $845 = (($844) + ($841)|0);
       $846 = HEAP8[$845>>0]|0;
       $dA$2$shrunk = $846;$dB$2$shrunk = $837;$dG$2$shrunk = $826;$dR$2$shrunk = $815;
       break;
      }
      default: {
       $dA$2$shrunk = 0;$dB$2$shrunk = 0;$dG$2$shrunk = 0;$dR$2$shrunk = 0;
      }
      }
      $dR$2 = $dR$2$shrunk&255;
      $dG$2 = $dG$2$shrunk&255;
      $dB$2 = $dB$2$shrunk&255;
      $dA$2 = $dA$2$shrunk&255;
      $847 = (($680) - ($dR$2))|0;
      $848 = Math_imul($847, $22)|0;
      $849 = (($848>>>0) / 255)&-1;
      $850 = (($849) + ($dR$2))|0;
      $851 = (($692) - ($dG$2))|0;
      $852 = Math_imul($851, $22)|0;
      $853 = (($852>>>0) / 255)&-1;
      $854 = (($853) + ($dG$2))|0;
      $855 = (($704) - ($dB$2))|0;
      $856 = Math_imul($855, $22)|0;
      $857 = (($856>>>0) / 255)&-1;
      $858 = (($857) + ($dB$2))|0;
      $859 = (($dA$2) + ($22))|0;
      $860 = Math_imul($dA$2, $22)|0;
      $861 = (($860>>>0) / 255)&-1;
      $862 = (($859) - ($861))|0;
      switch ($19|0) {
      case 1:  {
       $863 = HEAP32[$36>>2]|0;
       $864 = $863 & 255;
       $865 = $850 >>> $864;
       $866 = HEAP32[$37>>2]|0;
       $867 = $866 & 255;
       $868 = $865 << $867;
       $869 = $863 >>> 8;
       $870 = $869 & 255;
       $871 = $854 >>> $870;
       $872 = $866 >>> 8;
       $873 = $872 & 255;
       $874 = $871 << $873;
       $875 = $874 | $868;
       $876 = $863 >>> 16;
       $877 = $876 & 255;
       $878 = $858 >>> $877;
       $879 = $866 >>> 16;
       $880 = $879 & 255;
       $881 = $878 << $880;
       $882 = $875 | $881;
       $883 = $863 >>> 24;
       $884 = $862 >>> $883;
       $885 = $866 >>> 24;
       $886 = $884 << $885;
       $887 = $882 | $886;
       $888 = $887&255;
       HEAP8[$dst$3>>0] = $888;
       break L57;
       break;
      }
      case 2:  {
       $889 = HEAP32[$36>>2]|0;
       $890 = $889 & 255;
       $891 = $850 >>> $890;
       $892 = HEAP32[$37>>2]|0;
       $893 = $892 & 255;
       $894 = $891 << $893;
       $895 = $889 >>> 8;
       $896 = $895 & 255;
       $897 = $854 >>> $896;
       $898 = $892 >>> 8;
       $899 = $898 & 255;
       $900 = $897 << $899;
       $901 = $900 | $894;
       $902 = $889 >>> 16;
       $903 = $902 & 255;
       $904 = $858 >>> $903;
       $905 = $892 >>> 16;
       $906 = $905 & 255;
       $907 = $904 << $906;
       $908 = $901 | $907;
       $909 = $889 >>> 24;
       $910 = $862 >>> $909;
       $911 = $892 >>> 24;
       $912 = $910 << $911;
       $913 = $908 | $912;
       $914 = $913&65535;
       HEAP16[$dst$3>>1] = $914;
       break L57;
       break;
      }
      case 3:  {
       $915 = $850&255;
       $916 = HEAP8[$37>>0]|0;
       $div = ($916&255) >>> 3;
       $917 = $div&255;
       $918 = (($dst$3) + ($917)|0);
       HEAP8[$918>>0] = $915;
       $919 = $854&255;
       $920 = HEAP8[$38>>0]|0;
       $div1 = ($920&255) >>> 3;
       $921 = $div1&255;
       $922 = (($dst$3) + ($921)|0);
       HEAP8[$922>>0] = $919;
       $923 = $858&255;
       $924 = HEAP8[$39>>0]|0;
       $div2 = ($924&255) >>> 3;
       $925 = $div2&255;
       $926 = (($dst$3) + ($925)|0);
       HEAP8[$926>>0] = $923;
       break L57;
       break;
      }
      case 4:  {
       $927 = HEAP32[$36>>2]|0;
       $928 = $927 & 255;
       $929 = $850 >>> $928;
       $930 = HEAP32[$37>>2]|0;
       $931 = $930 & 255;
       $932 = $929 << $931;
       $933 = $927 >>> 8;
       $934 = $933 & 255;
       $935 = $854 >>> $934;
       $936 = $930 >>> 8;
       $937 = $936 & 255;
       $938 = $935 << $937;
       $939 = $938 | $932;
       $940 = $927 >>> 16;
       $941 = $940 & 255;
       $942 = $858 >>> $941;
       $943 = $930 >>> 16;
       $944 = $943 & 255;
       $945 = $942 << $944;
       $946 = $939 | $945;
       $947 = $927 >>> 24;
       $948 = $862 >>> $947;
       $949 = $930 >>> 24;
       $950 = $948 << $949;
       $951 = $946 | $950;
       HEAP32[$dst$3>>2] = $951;
       break L57;
       break;
      }
      default: {
       break L57;
      }
      }
     }
    } while(0);
    $952 = (($src$3) + ($16)|0);
    $953 = (($dst$3) + ($19)|0);
    $dst$4 = $953;$n$3 = $n$2;$src$4 = $952;
    label = 55;
    continue;
   }
   else if ((label|0) == 55) {
    label = 0;
    switch ($16|0) {
    case 1:  {
     $954 = HEAP8[$src$4>>0]|0;
     $955 = $954&255;
     $Pixel$3 = $955;
     break;
    }
    case 2:  {
     $956 = HEAP16[$src$4>>1]|0;
     $957 = $956&65535;
     $Pixel$3 = $957;
     break;
    }
    case 3:  {
     $958 = HEAP8[$src$4>>0]|0;
     $959 = $958&255;
     $960 = ((($src$4)) + 1|0);
     $961 = HEAP8[$960>>0]|0;
     $962 = $961&255;
     $963 = $962 << 8;
     $964 = $963 | $959;
     $965 = ((($src$4)) + 2|0);
     $966 = HEAP8[$965>>0]|0;
     $967 = $966&255;
     $968 = $967 << 16;
     $969 = $964 | $968;
     $Pixel$3 = $969;
     break;
    }
    case 4:  {
     $970 = HEAP32[$src$4>>2]|0;
     $Pixel$3 = $970;
     break;
    }
    default: {
     $Pixel$3 = 0;
    }
    }
    $971 = ($Pixel$3|0)==($13|0);
    $or$cond26 = $30 | $971;
    L78: do {
     if (!($or$cond26)) {
      $972 = HEAP32[$31>>2]|0;
      $973 = $972 & $Pixel$3;
      $974 = HEAP32[$32>>2]|0;
      $975 = $974 & 255;
      $976 = $973 >>> $975;
      $977 = HEAP32[$33>>2]|0;
      $978 = $977 & 255;
      $979 = (125924 + ($978<<2)|0);
      $980 = HEAP32[$979>>2]|0;
      $981 = (($980) + ($976)|0);
      $982 = HEAP8[$981>>0]|0;
      $983 = $982&255;
      $984 = HEAP32[$34>>2]|0;
      $985 = $984 & $Pixel$3;
      $986 = $974 >>> 8;
      $987 = $986 & 255;
      $988 = $985 >>> $987;
      $989 = $977 >>> 8;
      $990 = $989 & 255;
      $991 = (125924 + ($990<<2)|0);
      $992 = HEAP32[$991>>2]|0;
      $993 = (($992) + ($988)|0);
      $994 = HEAP8[$993>>0]|0;
      $995 = $994&255;
      $996 = HEAP32[$35>>2]|0;
      $997 = $996 & $Pixel$3;
      $998 = $974 >>> 16;
      $999 = $998 & 255;
      $1000 = $997 >>> $999;
      $1001 = $977 >>> 16;
      $1002 = $1001 & 255;
      $1003 = (125924 + ($1002<<2)|0);
      $1004 = HEAP32[$1003>>2]|0;
      $1005 = (($1004) + ($1000)|0);
      $1006 = HEAP8[$1005>>0]|0;
      $1007 = $1006&255;
      switch ($19|0) {
      case 1:  {
       $1008 = HEAP8[$dst$4>>0]|0;
       $1009 = $1008&255;
       $1010 = HEAP32[$40>>2]|0;
       $1011 = $1009 & $1010;
       $1012 = HEAP32[$37>>2]|0;
       $1013 = $1012 & 255;
       $1014 = $1011 >>> $1013;
       $1015 = HEAP32[$36>>2]|0;
       $1016 = $1015 & 255;
       $1017 = (125924 + ($1016<<2)|0);
       $1018 = HEAP32[$1017>>2]|0;
       $1019 = (($1018) + ($1014)|0);
       $1020 = HEAP8[$1019>>0]|0;
       $1021 = HEAP32[$41>>2]|0;
       $1022 = $1021 & $1009;
       $1023 = $1012 >>> 8;
       $1024 = $1023 & 255;
       $1025 = $1022 >>> $1024;
       $1026 = $1015 >>> 8;
       $1027 = $1026 & 255;
       $1028 = (125924 + ($1027<<2)|0);
       $1029 = HEAP32[$1028>>2]|0;
       $1030 = (($1029) + ($1025)|0);
       $1031 = HEAP8[$1030>>0]|0;
       $1032 = HEAP32[$42>>2]|0;
       $1033 = $1032 & $1009;
       $1034 = $1012 >>> 16;
       $1035 = $1034 & 255;
       $1036 = $1033 >>> $1035;
       $1037 = $1015 >>> 16;
       $1038 = $1037 & 255;
       $1039 = (125924 + ($1038<<2)|0);
       $1040 = HEAP32[$1039>>2]|0;
       $1041 = (($1040) + ($1036)|0);
       $1042 = HEAP8[$1041>>0]|0;
       $1043 = HEAP32[$43>>2]|0;
       $1044 = $1043 & $1009;
       $1045 = $1012 >>> 24;
       $1046 = $1044 >>> $1045;
       $1047 = $1015 >>> 24;
       $1048 = (125924 + ($1047<<2)|0);
       $1049 = HEAP32[$1048>>2]|0;
       $1050 = (($1049) + ($1046)|0);
       $1051 = HEAP8[$1050>>0]|0;
       $dA$3$shrunk = $1051;$dB$3$shrunk = $1042;$dG$3$shrunk = $1031;$dR$3$shrunk = $1020;
       break;
      }
      case 2:  {
       $1052 = HEAP16[$dst$4>>1]|0;
       $1053 = $1052&65535;
       $1054 = HEAP32[$40>>2]|0;
       $1055 = $1053 & $1054;
       $1056 = HEAP32[$37>>2]|0;
       $1057 = $1056 & 255;
       $1058 = $1055 >>> $1057;
       $1059 = HEAP32[$36>>2]|0;
       $1060 = $1059 & 255;
       $1061 = (125924 + ($1060<<2)|0);
       $1062 = HEAP32[$1061>>2]|0;
       $1063 = (($1062) + ($1058)|0);
       $1064 = HEAP8[$1063>>0]|0;
       $1065 = HEAP32[$41>>2]|0;
       $1066 = $1065 & $1053;
       $1067 = $1056 >>> 8;
       $1068 = $1067 & 255;
       $1069 = $1066 >>> $1068;
       $1070 = $1059 >>> 8;
       $1071 = $1070 & 255;
       $1072 = (125924 + ($1071<<2)|0);
       $1073 = HEAP32[$1072>>2]|0;
       $1074 = (($1073) + ($1069)|0);
       $1075 = HEAP8[$1074>>0]|0;
       $1076 = HEAP32[$42>>2]|0;
       $1077 = $1076 & $1053;
       $1078 = $1056 >>> 16;
       $1079 = $1078 & 255;
       $1080 = $1077 >>> $1079;
       $1081 = $1059 >>> 16;
       $1082 = $1081 & 255;
       $1083 = (125924 + ($1082<<2)|0);
       $1084 = HEAP32[$1083>>2]|0;
       $1085 = (($1084) + ($1080)|0);
       $1086 = HEAP8[$1085>>0]|0;
       $1087 = HEAP32[$43>>2]|0;
       $1088 = $1087 & $1053;
       $1089 = $1056 >>> 24;
       $1090 = $1088 >>> $1089;
       $1091 = $1059 >>> 24;
       $1092 = (125924 + ($1091<<2)|0);
       $1093 = HEAP32[$1092>>2]|0;
       $1094 = (($1093) + ($1090)|0);
       $1095 = HEAP8[$1094>>0]|0;
       $dA$3$shrunk = $1095;$dB$3$shrunk = $1086;$dG$3$shrunk = $1075;$dR$3$shrunk = $1064;
       break;
      }
      case 3:  {
       $1096 = HEAP32[$37>>2]|0;
       $div21 = $1096 >>> 3;
       $1097 = $div21 & 31;
       $1098 = (($dst$4) + ($1097)|0);
       $1099 = HEAP8[$1098>>0]|0;
       $1100 = $1096 >>> 11;
       $div22 = $1100 & 31;
       $1101 = (($dst$4) + ($div22)|0);
       $1102 = HEAP8[$1101>>0]|0;
       $1103 = $1096 >>> 19;
       $1104 = $1103 & 31;
       $1105 = (($dst$4) + ($1104)|0);
       $1106 = HEAP8[$1105>>0]|0;
       $dA$3$shrunk = -1;$dB$3$shrunk = $1106;$dG$3$shrunk = $1102;$dR$3$shrunk = $1099;
       break;
      }
      case 4:  {
       $1107 = HEAP32[$dst$4>>2]|0;
       $1108 = HEAP32[$40>>2]|0;
       $1109 = $1108 & $1107;
       $1110 = HEAP32[$37>>2]|0;
       $1111 = $1110 & 255;
       $1112 = $1109 >>> $1111;
       $1113 = HEAP32[$36>>2]|0;
       $1114 = $1113 & 255;
       $1115 = (125924 + ($1114<<2)|0);
       $1116 = HEAP32[$1115>>2]|0;
       $1117 = (($1116) + ($1112)|0);
       $1118 = HEAP8[$1117>>0]|0;
       $1119 = HEAP32[$41>>2]|0;
       $1120 = $1119 & $1107;
       $1121 = $1110 >>> 8;
       $1122 = $1121 & 255;
       $1123 = $1120 >>> $1122;
       $1124 = $1113 >>> 8;
       $1125 = $1124 & 255;
       $1126 = (125924 + ($1125<<2)|0);
       $1127 = HEAP32[$1126>>2]|0;
       $1128 = (($1127) + ($1123)|0);
       $1129 = HEAP8[$1128>>0]|0;
       $1130 = HEAP32[$42>>2]|0;
       $1131 = $1130 & $1107;
       $1132 = $1110 >>> 16;
       $1133 = $1132 & 255;
       $1134 = $1131 >>> $1133;
       $1135 = $1113 >>> 16;
       $1136 = $1135 & 255;
       $1137 = (125924 + ($1136<<2)|0);
       $1138 = HEAP32[$1137>>2]|0;
       $1139 = (($1138) + ($1134)|0);
       $1140 = HEAP8[$1139>>0]|0;
       $1141 = HEAP32[$43>>2]|0;
       $1142 = $1141 & $1107;
       $1143 = $1110 >>> 24;
       $1144 = $1142 >>> $1143;
       $1145 = $1113 >>> 24;
       $1146 = (125924 + ($1145<<2)|0);
       $1147 = HEAP32[$1146>>2]|0;
       $1148 = (($1147) + ($1144)|0);
       $1149 = HEAP8[$1148>>0]|0;
       $dA$3$shrunk = $1149;$dB$3$shrunk = $1140;$dG$3$shrunk = $1129;$dR$3$shrunk = $1118;
       break;
      }
      default: {
       $dA$3$shrunk = 0;$dB$3$shrunk = 0;$dG$3$shrunk = 0;$dR$3$shrunk = 0;
      }
      }
      $dR$3 = $dR$3$shrunk&255;
      $dG$3 = $dG$3$shrunk&255;
      $dB$3 = $dB$3$shrunk&255;
      $dA$3 = $dA$3$shrunk&255;
      $1150 = (($983) - ($dR$3))|0;
      $1151 = Math_imul($1150, $22)|0;
      $1152 = (($1151>>>0) / 255)&-1;
      $1153 = (($1152) + ($dR$3))|0;
      $1154 = (($995) - ($dG$3))|0;
      $1155 = Math_imul($1154, $22)|0;
      $1156 = (($1155>>>0) / 255)&-1;
      $1157 = (($1156) + ($dG$3))|0;
      $1158 = (($1007) - ($dB$3))|0;
      $1159 = Math_imul($1158, $22)|0;
      $1160 = (($1159>>>0) / 255)&-1;
      $1161 = (($1160) + ($dB$3))|0;
      $1162 = (($dA$3) + ($22))|0;
      $1163 = Math_imul($dA$3, $22)|0;
      $1164 = (($1163>>>0) / 255)&-1;
      $1165 = (($1162) - ($1164))|0;
      switch ($19|0) {
      case 1:  {
       $1166 = HEAP32[$36>>2]|0;
       $1167 = $1166 & 255;
       $1168 = $1153 >>> $1167;
       $1169 = HEAP32[$37>>2]|0;
       $1170 = $1169 & 255;
       $1171 = $1168 << $1170;
       $1172 = $1166 >>> 8;
       $1173 = $1172 & 255;
       $1174 = $1157 >>> $1173;
       $1175 = $1169 >>> 8;
       $1176 = $1175 & 255;
       $1177 = $1174 << $1176;
       $1178 = $1177 | $1171;
       $1179 = $1166 >>> 16;
       $1180 = $1179 & 255;
       $1181 = $1161 >>> $1180;
       $1182 = $1169 >>> 16;
       $1183 = $1182 & 255;
       $1184 = $1181 << $1183;
       $1185 = $1178 | $1184;
       $1186 = $1166 >>> 24;
       $1187 = $1165 >>> $1186;
       $1188 = $1169 >>> 24;
       $1189 = $1187 << $1188;
       $1190 = $1185 | $1189;
       $1191 = $1190&255;
       HEAP8[$dst$4>>0] = $1191;
       break L78;
       break;
      }
      case 2:  {
       $1192 = HEAP32[$36>>2]|0;
       $1193 = $1192 & 255;
       $1194 = $1153 >>> $1193;
       $1195 = HEAP32[$37>>2]|0;
       $1196 = $1195 & 255;
       $1197 = $1194 << $1196;
       $1198 = $1192 >>> 8;
       $1199 = $1198 & 255;
       $1200 = $1157 >>> $1199;
       $1201 = $1195 >>> 8;
       $1202 = $1201 & 255;
       $1203 = $1200 << $1202;
       $1204 = $1203 | $1197;
       $1205 = $1192 >>> 16;
       $1206 = $1205 & 255;
       $1207 = $1161 >>> $1206;
       $1208 = $1195 >>> 16;
       $1209 = $1208 & 255;
       $1210 = $1207 << $1209;
       $1211 = $1204 | $1210;
       $1212 = $1192 >>> 24;
       $1213 = $1165 >>> $1212;
       $1214 = $1195 >>> 24;
       $1215 = $1213 << $1214;
       $1216 = $1211 | $1215;
       $1217 = $1216&65535;
       HEAP16[$dst$4>>1] = $1217;
       break L78;
       break;
      }
      case 3:  {
       $1218 = $1153&255;
       $1219 = HEAP8[$37>>0]|0;
       $div18 = ($1219&255) >>> 3;
       $1220 = $div18&255;
       $1221 = (($dst$4) + ($1220)|0);
       HEAP8[$1221>>0] = $1218;
       $1222 = $1157&255;
       $1223 = HEAP8[$38>>0]|0;
       $div19 = ($1223&255) >>> 3;
       $1224 = $div19&255;
       $1225 = (($dst$4) + ($1224)|0);
       HEAP8[$1225>>0] = $1222;
       $1226 = $1161&255;
       $1227 = HEAP8[$39>>0]|0;
       $div20 = ($1227&255) >>> 3;
       $1228 = $div20&255;
       $1229 = (($dst$4) + ($1228)|0);
       HEAP8[$1229>>0] = $1226;
       break L78;
       break;
      }
      case 4:  {
       $1230 = HEAP32[$36>>2]|0;
       $1231 = $1230 & 255;
       $1232 = $1153 >>> $1231;
       $1233 = HEAP32[$37>>2]|0;
       $1234 = $1233 & 255;
       $1235 = $1232 << $1234;
       $1236 = $1230 >>> 8;
       $1237 = $1236 & 255;
       $1238 = $1157 >>> $1237;
       $1239 = $1233 >>> 8;
       $1240 = $1239 & 255;
       $1241 = $1238 << $1240;
       $1242 = $1241 | $1235;
       $1243 = $1230 >>> 16;
       $1244 = $1243 & 255;
       $1245 = $1161 >>> $1244;
       $1246 = $1233 >>> 16;
       $1247 = $1246 & 255;
       $1248 = $1245 << $1247;
       $1249 = $1242 | $1248;
       $1250 = $1230 >>> 24;
       $1251 = $1165 >>> $1250;
       $1252 = $1233 >>> 24;
       $1253 = $1251 << $1252;
       $1254 = $1249 | $1253;
       HEAP32[$dst$4>>2] = $1254;
       break L78;
       break;
      }
      default: {
       break L78;
      }
      }
     }
    } while(0);
    $1255 = (($src$4) + ($16)|0);
    $1256 = (($dst$4) + ($19)|0);
    $1257 = (($n$3) + -1)|0;
    $1258 = ($n$3|0)>(1);
    if ($1258) {
     $dst$1 = $1256;$n$0 = $1257;$src$1 = $1255;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $1259 = (($1255) + ($5)|0);
  $1260 = (($1256) + ($7)|0);
  $1261 = ($44|0)==(0);
  if ($1261) {
   label = 74;
   break;
  } else {
   $$in = $44;$dst$030 = $1260;$src$031 = $1259;
  }
 }
 if ((label|0) == 72) {
  // unreachable;
 }
 else if ((label|0) == 74) {
  return;
 }
}
function _BlitNto1PixelAlpha($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0;
 var $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0;
 var $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0;
 var $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0;
 var $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0;
 var $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0;
 var $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0;
 var $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0;
 var $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0;
 var $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0;
 var $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0;
 var $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0;
 var $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0;
 var $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0;
 var $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0;
 var $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0;
 var $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0;
 var $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0;
 var $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0;
 var $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0;
 var $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0;
 var $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0;
 var $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0;
 var $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0;
 var $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0;
 var $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0;
 var $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0;
 var $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0;
 var $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0;
 var $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $div = 0;
 var $div10 = 0, $div12 = 0, $div13 = 0, $div4 = 0, $div6 = 0, $div7 = 0, $div9 = 0, $dst$015 = 0, $dst$1 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $sA$0 = 0, $sA$0$shrunk = 0, $sA$1 = 0, $sA$1$shrunk = 0;
 var $sA$2 = 0, $sA$2$shrunk = 0, $sA$3 = 0, $sA$3$shrunk = 0, $sB$0 = 0, $sB$0$shrunk = 0, $sB$1 = 0, $sB$1$shrunk = 0, $sB$2 = 0, $sB$2$shrunk = 0, $sB$3 = 0, $sB$3$shrunk = 0, $sG$0 = 0, $sG$0$shrunk = 0, $sG$1 = 0, $sG$1$shrunk = 0, $sG$2 = 0, $sG$2$shrunk = 0, $sG$3 = 0, $sG$3$shrunk = 0;
 var $sR$0 = 0, $sR$0$shrunk = 0, $sR$1 = 0, $sR$1$shrunk = 0, $sR$2 = 0, $sR$2$shrunk = 0, $sR$3 = 0, $sR$3$shrunk = 0, $src$016 = 0, $src$1 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, $storemerge = 0, $storemerge1 = 0, $storemerge2 = 0, $storemerge3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 48|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($info)) + 40|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($11)) + 9|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = $13&255;
 $15 = ($3|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 44|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ((($info)) + 20|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = HEAP32[$info>>2]|0;
 $21 = (($1) + 3)|0;
 $22 = (($21|0) / 4)&-1;
 $23 = $1 & 3;
 $24 = ((($17)) + 4|0);
 $25 = ($9|0)==(0|0);
 $26 = ((($11)) + 12|0);
 $27 = ((($11)) + 32|0);
 $28 = ((($11)) + 28|0);
 $29 = ((($11)) + 16|0);
 $30 = ((($11)) + 20|0);
 $31 = ((($11)) + 24|0);
 $$in = $3;$dst$015 = $19;$src$016 = $20;
 L4: while(1) {
  $32 = (($$in) + -1)|0;
  switch ($23|0) {
  case 0:  {
   $dst$1 = $dst$015;$n$0 = $22;$src$1 = $src$016;
   label = 4;
   break;
  }
  case 3:  {
   $dst$2 = $dst$015;$n$1 = $22;$src$2 = $src$016;
   label = 13;
   break;
  }
  case 2:  {
   $dst$3 = $dst$015;$n$2 = $22;$src$3 = $src$016;
   label = 22;
   break;
  }
  case 1:  {
   $dst$4 = $dst$015;$n$3 = $22;$src$4 = $src$016;
   label = 31;
   break;
  }
  default: {
   label = 40;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $33 = HEAP8[$src$1>>0]|0;
     $34 = $33&255;
     $35 = HEAP32[$26>>2]|0;
     $36 = $34 & $35;
     $37 = HEAP32[$27>>2]|0;
     $38 = $37 & 255;
     $39 = $36 >>> $38;
     $40 = HEAP32[$28>>2]|0;
     $41 = $40 & 255;
     $42 = (125924 + ($41<<2)|0);
     $43 = HEAP32[$42>>2]|0;
     $44 = (($43) + ($39)|0);
     $45 = HEAP8[$44>>0]|0;
     $46 = HEAP32[$29>>2]|0;
     $47 = $46 & $34;
     $48 = $37 >>> 8;
     $49 = $48 & 255;
     $50 = $47 >>> $49;
     $51 = $40 >>> 8;
     $52 = $51 & 255;
     $53 = (125924 + ($52<<2)|0);
     $54 = HEAP32[$53>>2]|0;
     $55 = (($54) + ($50)|0);
     $56 = HEAP8[$55>>0]|0;
     $57 = HEAP32[$30>>2]|0;
     $58 = $57 & $34;
     $59 = $37 >>> 16;
     $60 = $59 & 255;
     $61 = $58 >>> $60;
     $62 = $40 >>> 16;
     $63 = $62 & 255;
     $64 = (125924 + ($63<<2)|0);
     $65 = HEAP32[$64>>2]|0;
     $66 = (($65) + ($61)|0);
     $67 = HEAP8[$66>>0]|0;
     $68 = HEAP32[$31>>2]|0;
     $69 = $68 & $34;
     $70 = $37 >>> 24;
     $71 = $69 >>> $70;
     $72 = $40 >>> 24;
     $73 = (125924 + ($72<<2)|0);
     $74 = HEAP32[$73>>2]|0;
     $75 = (($74) + ($71)|0);
     $76 = HEAP8[$75>>0]|0;
     $sA$0$shrunk = $76;$sB$0$shrunk = $67;$sG$0$shrunk = $56;$sR$0$shrunk = $45;
     break;
    }
    case 2:  {
     $77 = HEAP16[$src$1>>1]|0;
     $78 = $77&65535;
     $79 = HEAP32[$26>>2]|0;
     $80 = $78 & $79;
     $81 = HEAP32[$27>>2]|0;
     $82 = $81 & 255;
     $83 = $80 >>> $82;
     $84 = HEAP32[$28>>2]|0;
     $85 = $84 & 255;
     $86 = (125924 + ($85<<2)|0);
     $87 = HEAP32[$86>>2]|0;
     $88 = (($87) + ($83)|0);
     $89 = HEAP8[$88>>0]|0;
     $90 = HEAP32[$29>>2]|0;
     $91 = $90 & $78;
     $92 = $81 >>> 8;
     $93 = $92 & 255;
     $94 = $91 >>> $93;
     $95 = $84 >>> 8;
     $96 = $95 & 255;
     $97 = (125924 + ($96<<2)|0);
     $98 = HEAP32[$97>>2]|0;
     $99 = (($98) + ($94)|0);
     $100 = HEAP8[$99>>0]|0;
     $101 = HEAP32[$30>>2]|0;
     $102 = $101 & $78;
     $103 = $81 >>> 16;
     $104 = $103 & 255;
     $105 = $102 >>> $104;
     $106 = $84 >>> 16;
     $107 = $106 & 255;
     $108 = (125924 + ($107<<2)|0);
     $109 = HEAP32[$108>>2]|0;
     $110 = (($109) + ($105)|0);
     $111 = HEAP8[$110>>0]|0;
     $112 = HEAP32[$31>>2]|0;
     $113 = $112 & $78;
     $114 = $81 >>> 24;
     $115 = $113 >>> $114;
     $116 = $84 >>> 24;
     $117 = (125924 + ($116<<2)|0);
     $118 = HEAP32[$117>>2]|0;
     $119 = (($118) + ($115)|0);
     $120 = HEAP8[$119>>0]|0;
     $sA$0$shrunk = $120;$sB$0$shrunk = $111;$sG$0$shrunk = $100;$sR$0$shrunk = $89;
     break;
    }
    case 3:  {
     $121 = HEAP32[$27>>2]|0;
     $div9 = $121 >>> 3;
     $122 = $div9 & 31;
     $123 = (($src$1) + ($122)|0);
     $124 = HEAP8[$123>>0]|0;
     $125 = $121 >>> 11;
     $div10 = $125 & 31;
     $126 = (($src$1) + ($div10)|0);
     $127 = HEAP8[$126>>0]|0;
     $128 = $121 >>> 19;
     $129 = $128 & 31;
     $130 = (($src$1) + ($129)|0);
     $131 = HEAP8[$130>>0]|0;
     $sA$0$shrunk = -1;$sB$0$shrunk = $131;$sG$0$shrunk = $127;$sR$0$shrunk = $124;
     break;
    }
    case 4:  {
     $132 = HEAP32[$src$1>>2]|0;
     $133 = HEAP32[$26>>2]|0;
     $134 = $133 & $132;
     $135 = HEAP32[$27>>2]|0;
     $136 = $135 & 255;
     $137 = $134 >>> $136;
     $138 = HEAP32[$28>>2]|0;
     $139 = $138 & 255;
     $140 = (125924 + ($139<<2)|0);
     $141 = HEAP32[$140>>2]|0;
     $142 = (($141) + ($137)|0);
     $143 = HEAP8[$142>>0]|0;
     $144 = HEAP32[$29>>2]|0;
     $145 = $144 & $132;
     $146 = $135 >>> 8;
     $147 = $146 & 255;
     $148 = $145 >>> $147;
     $149 = $138 >>> 8;
     $150 = $149 & 255;
     $151 = (125924 + ($150<<2)|0);
     $152 = HEAP32[$151>>2]|0;
     $153 = (($152) + ($148)|0);
     $154 = HEAP8[$153>>0]|0;
     $155 = HEAP32[$30>>2]|0;
     $156 = $155 & $132;
     $157 = $135 >>> 16;
     $158 = $157 & 255;
     $159 = $156 >>> $158;
     $160 = $138 >>> 16;
     $161 = $160 & 255;
     $162 = (125924 + ($161<<2)|0);
     $163 = HEAP32[$162>>2]|0;
     $164 = (($163) + ($159)|0);
     $165 = HEAP8[$164>>0]|0;
     $166 = HEAP32[$31>>2]|0;
     $167 = $166 & $132;
     $168 = $135 >>> 24;
     $169 = $167 >>> $168;
     $170 = $138 >>> 24;
     $171 = (125924 + ($170<<2)|0);
     $172 = HEAP32[$171>>2]|0;
     $173 = (($172) + ($169)|0);
     $174 = HEAP8[$173>>0]|0;
     $sA$0$shrunk = $174;$sB$0$shrunk = $165;$sG$0$shrunk = $154;$sR$0$shrunk = $143;
     break;
    }
    default: {
     $sA$0$shrunk = 0;$sB$0$shrunk = 0;$sG$0$shrunk = 0;$sR$0$shrunk = 0;
    }
    }
    $sA$0 = $sA$0$shrunk&255;
    $sB$0 = $sB$0$shrunk&255;
    $sG$0 = $sG$0$shrunk&255;
    $sR$0 = $sR$0$shrunk&255;
    $175 = HEAP8[$dst$1>>0]|0;
    $176 = $175&255;
    $177 = HEAP32[$24>>2]|0;
    $178 = ((($177)) + 4|0);
    $179 = HEAP32[$178>>2]|0;
    $180 = (($179) + ($176<<2)|0);
    $181 = HEAP8[$180>>0]|0;
    $182 = $181&255;
    $183 = (((($179) + ($176<<2)|0)) + 1|0);
    $184 = HEAP8[$183>>0]|0;
    $185 = $184&255;
    $186 = (((($179) + ($176<<2)|0)) + 2|0);
    $187 = HEAP8[$186>>0]|0;
    $188 = $187&255;
    $189 = (($sR$0) - ($182))|0;
    $190 = Math_imul($189, $sA$0)|0;
    $191 = (($190>>>0) / 255)&-1;
    $192 = (($191) + ($182))|0;
    $193 = (($sG$0) - ($185))|0;
    $194 = Math_imul($193, $sA$0)|0;
    $195 = (($194>>>0) / 255)&-1;
    $196 = (($195) + ($185))|0;
    $197 = (($sB$0) - ($188))|0;
    $198 = Math_imul($197, $sA$0)|0;
    $199 = (($198>>>0) / 255)&-1;
    $200 = (($199) + ($188))|0;
    $201 = $192 & 224;
    $202 = $196 >>> 3;
    $203 = $202 & 28;
    $204 = $203 | $201;
    $205 = $200 >>> 6;
    $206 = $205 & 3;
    $207 = $204 | $206;
    if ($25) {
     $208 = $207&255;
     $storemerge1 = $208;
    } else {
     $209 = (($9) + ($207)|0);
     $210 = HEAP8[$209>>0]|0;
     $storemerge1 = $210;
    }
    HEAP8[$dst$1>>0] = $storemerge1;
    $211 = ((($dst$1)) + 1|0);
    $212 = (($src$1) + ($14)|0);
    $dst$2 = $211;$n$1 = $n$0;$src$2 = $212;
    label = 13;
    continue;
   }
   else if ((label|0) == 13) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $213 = HEAP8[$src$2>>0]|0;
     $214 = $213&255;
     $215 = HEAP32[$26>>2]|0;
     $216 = $214 & $215;
     $217 = HEAP32[$27>>2]|0;
     $218 = $217 & 255;
     $219 = $216 >>> $218;
     $220 = HEAP32[$28>>2]|0;
     $221 = $220 & 255;
     $222 = (125924 + ($221<<2)|0);
     $223 = HEAP32[$222>>2]|0;
     $224 = (($223) + ($219)|0);
     $225 = HEAP8[$224>>0]|0;
     $226 = HEAP32[$29>>2]|0;
     $227 = $226 & $214;
     $228 = $217 >>> 8;
     $229 = $228 & 255;
     $230 = $227 >>> $229;
     $231 = $220 >>> 8;
     $232 = $231 & 255;
     $233 = (125924 + ($232<<2)|0);
     $234 = HEAP32[$233>>2]|0;
     $235 = (($234) + ($230)|0);
     $236 = HEAP8[$235>>0]|0;
     $237 = HEAP32[$30>>2]|0;
     $238 = $237 & $214;
     $239 = $217 >>> 16;
     $240 = $239 & 255;
     $241 = $238 >>> $240;
     $242 = $220 >>> 16;
     $243 = $242 & 255;
     $244 = (125924 + ($243<<2)|0);
     $245 = HEAP32[$244>>2]|0;
     $246 = (($245) + ($241)|0);
     $247 = HEAP8[$246>>0]|0;
     $248 = HEAP32[$31>>2]|0;
     $249 = $248 & $214;
     $250 = $217 >>> 24;
     $251 = $249 >>> $250;
     $252 = $220 >>> 24;
     $253 = (125924 + ($252<<2)|0);
     $254 = HEAP32[$253>>2]|0;
     $255 = (($254) + ($251)|0);
     $256 = HEAP8[$255>>0]|0;
     $sA$1$shrunk = $256;$sB$1$shrunk = $247;$sG$1$shrunk = $236;$sR$1$shrunk = $225;
     break;
    }
    case 2:  {
     $257 = HEAP16[$src$2>>1]|0;
     $258 = $257&65535;
     $259 = HEAP32[$26>>2]|0;
     $260 = $258 & $259;
     $261 = HEAP32[$27>>2]|0;
     $262 = $261 & 255;
     $263 = $260 >>> $262;
     $264 = HEAP32[$28>>2]|0;
     $265 = $264 & 255;
     $266 = (125924 + ($265<<2)|0);
     $267 = HEAP32[$266>>2]|0;
     $268 = (($267) + ($263)|0);
     $269 = HEAP8[$268>>0]|0;
     $270 = HEAP32[$29>>2]|0;
     $271 = $270 & $258;
     $272 = $261 >>> 8;
     $273 = $272 & 255;
     $274 = $271 >>> $273;
     $275 = $264 >>> 8;
     $276 = $275 & 255;
     $277 = (125924 + ($276<<2)|0);
     $278 = HEAP32[$277>>2]|0;
     $279 = (($278) + ($274)|0);
     $280 = HEAP8[$279>>0]|0;
     $281 = HEAP32[$30>>2]|0;
     $282 = $281 & $258;
     $283 = $261 >>> 16;
     $284 = $283 & 255;
     $285 = $282 >>> $284;
     $286 = $264 >>> 16;
     $287 = $286 & 255;
     $288 = (125924 + ($287<<2)|0);
     $289 = HEAP32[$288>>2]|0;
     $290 = (($289) + ($285)|0);
     $291 = HEAP8[$290>>0]|0;
     $292 = HEAP32[$31>>2]|0;
     $293 = $292 & $258;
     $294 = $261 >>> 24;
     $295 = $293 >>> $294;
     $296 = $264 >>> 24;
     $297 = (125924 + ($296<<2)|0);
     $298 = HEAP32[$297>>2]|0;
     $299 = (($298) + ($295)|0);
     $300 = HEAP8[$299>>0]|0;
     $sA$1$shrunk = $300;$sB$1$shrunk = $291;$sG$1$shrunk = $280;$sR$1$shrunk = $269;
     break;
    }
    case 3:  {
     $301 = HEAP32[$27>>2]|0;
     $div6 = $301 >>> 3;
     $302 = $div6 & 31;
     $303 = (($src$2) + ($302)|0);
     $304 = HEAP8[$303>>0]|0;
     $305 = $301 >>> 11;
     $div7 = $305 & 31;
     $306 = (($src$2) + ($div7)|0);
     $307 = HEAP8[$306>>0]|0;
     $308 = $301 >>> 19;
     $309 = $308 & 31;
     $310 = (($src$2) + ($309)|0);
     $311 = HEAP8[$310>>0]|0;
     $sA$1$shrunk = -1;$sB$1$shrunk = $311;$sG$1$shrunk = $307;$sR$1$shrunk = $304;
     break;
    }
    case 4:  {
     $312 = HEAP32[$src$2>>2]|0;
     $313 = HEAP32[$26>>2]|0;
     $314 = $313 & $312;
     $315 = HEAP32[$27>>2]|0;
     $316 = $315 & 255;
     $317 = $314 >>> $316;
     $318 = HEAP32[$28>>2]|0;
     $319 = $318 & 255;
     $320 = (125924 + ($319<<2)|0);
     $321 = HEAP32[$320>>2]|0;
     $322 = (($321) + ($317)|0);
     $323 = HEAP8[$322>>0]|0;
     $324 = HEAP32[$29>>2]|0;
     $325 = $324 & $312;
     $326 = $315 >>> 8;
     $327 = $326 & 255;
     $328 = $325 >>> $327;
     $329 = $318 >>> 8;
     $330 = $329 & 255;
     $331 = (125924 + ($330<<2)|0);
     $332 = HEAP32[$331>>2]|0;
     $333 = (($332) + ($328)|0);
     $334 = HEAP8[$333>>0]|0;
     $335 = HEAP32[$30>>2]|0;
     $336 = $335 & $312;
     $337 = $315 >>> 16;
     $338 = $337 & 255;
     $339 = $336 >>> $338;
     $340 = $318 >>> 16;
     $341 = $340 & 255;
     $342 = (125924 + ($341<<2)|0);
     $343 = HEAP32[$342>>2]|0;
     $344 = (($343) + ($339)|0);
     $345 = HEAP8[$344>>0]|0;
     $346 = HEAP32[$31>>2]|0;
     $347 = $346 & $312;
     $348 = $315 >>> 24;
     $349 = $347 >>> $348;
     $350 = $318 >>> 24;
     $351 = (125924 + ($350<<2)|0);
     $352 = HEAP32[$351>>2]|0;
     $353 = (($352) + ($349)|0);
     $354 = HEAP8[$353>>0]|0;
     $sA$1$shrunk = $354;$sB$1$shrunk = $345;$sG$1$shrunk = $334;$sR$1$shrunk = $323;
     break;
    }
    default: {
     $sA$1$shrunk = 0;$sB$1$shrunk = 0;$sG$1$shrunk = 0;$sR$1$shrunk = 0;
    }
    }
    $sA$1 = $sA$1$shrunk&255;
    $sB$1 = $sB$1$shrunk&255;
    $sG$1 = $sG$1$shrunk&255;
    $sR$1 = $sR$1$shrunk&255;
    $355 = HEAP8[$dst$2>>0]|0;
    $356 = $355&255;
    $357 = HEAP32[$24>>2]|0;
    $358 = ((($357)) + 4|0);
    $359 = HEAP32[$358>>2]|0;
    $360 = (($359) + ($356<<2)|0);
    $361 = HEAP8[$360>>0]|0;
    $362 = $361&255;
    $363 = (((($359) + ($356<<2)|0)) + 1|0);
    $364 = HEAP8[$363>>0]|0;
    $365 = $364&255;
    $366 = (((($359) + ($356<<2)|0)) + 2|0);
    $367 = HEAP8[$366>>0]|0;
    $368 = $367&255;
    $369 = (($sR$1) - ($362))|0;
    $370 = Math_imul($369, $sA$1)|0;
    $371 = (($370>>>0) / 255)&-1;
    $372 = (($371) + ($362))|0;
    $373 = (($sG$1) - ($365))|0;
    $374 = Math_imul($373, $sA$1)|0;
    $375 = (($374>>>0) / 255)&-1;
    $376 = (($375) + ($365))|0;
    $377 = (($sB$1) - ($368))|0;
    $378 = Math_imul($377, $sA$1)|0;
    $379 = (($378>>>0) / 255)&-1;
    $380 = (($379) + ($368))|0;
    $381 = $372 & 224;
    $382 = $376 >>> 3;
    $383 = $382 & 28;
    $384 = $383 | $381;
    $385 = $380 >>> 6;
    $386 = $385 & 3;
    $387 = $384 | $386;
    if ($25) {
     $388 = $387&255;
     $storemerge2 = $388;
    } else {
     $389 = (($9) + ($387)|0);
     $390 = HEAP8[$389>>0]|0;
     $storemerge2 = $390;
    }
    HEAP8[$dst$2>>0] = $storemerge2;
    $391 = ((($dst$2)) + 1|0);
    $392 = (($src$2) + ($14)|0);
    $dst$3 = $391;$n$2 = $n$1;$src$3 = $392;
    label = 22;
    continue;
   }
   else if ((label|0) == 22) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $393 = HEAP8[$src$3>>0]|0;
     $394 = $393&255;
     $395 = HEAP32[$26>>2]|0;
     $396 = $394 & $395;
     $397 = HEAP32[$27>>2]|0;
     $398 = $397 & 255;
     $399 = $396 >>> $398;
     $400 = HEAP32[$28>>2]|0;
     $401 = $400 & 255;
     $402 = (125924 + ($401<<2)|0);
     $403 = HEAP32[$402>>2]|0;
     $404 = (($403) + ($399)|0);
     $405 = HEAP8[$404>>0]|0;
     $406 = HEAP32[$29>>2]|0;
     $407 = $406 & $394;
     $408 = $397 >>> 8;
     $409 = $408 & 255;
     $410 = $407 >>> $409;
     $411 = $400 >>> 8;
     $412 = $411 & 255;
     $413 = (125924 + ($412<<2)|0);
     $414 = HEAP32[$413>>2]|0;
     $415 = (($414) + ($410)|0);
     $416 = HEAP8[$415>>0]|0;
     $417 = HEAP32[$30>>2]|0;
     $418 = $417 & $394;
     $419 = $397 >>> 16;
     $420 = $419 & 255;
     $421 = $418 >>> $420;
     $422 = $400 >>> 16;
     $423 = $422 & 255;
     $424 = (125924 + ($423<<2)|0);
     $425 = HEAP32[$424>>2]|0;
     $426 = (($425) + ($421)|0);
     $427 = HEAP8[$426>>0]|0;
     $428 = HEAP32[$31>>2]|0;
     $429 = $428 & $394;
     $430 = $397 >>> 24;
     $431 = $429 >>> $430;
     $432 = $400 >>> 24;
     $433 = (125924 + ($432<<2)|0);
     $434 = HEAP32[$433>>2]|0;
     $435 = (($434) + ($431)|0);
     $436 = HEAP8[$435>>0]|0;
     $sA$2$shrunk = $436;$sB$2$shrunk = $427;$sG$2$shrunk = $416;$sR$2$shrunk = $405;
     break;
    }
    case 2:  {
     $437 = HEAP16[$src$3>>1]|0;
     $438 = $437&65535;
     $439 = HEAP32[$26>>2]|0;
     $440 = $438 & $439;
     $441 = HEAP32[$27>>2]|0;
     $442 = $441 & 255;
     $443 = $440 >>> $442;
     $444 = HEAP32[$28>>2]|0;
     $445 = $444 & 255;
     $446 = (125924 + ($445<<2)|0);
     $447 = HEAP32[$446>>2]|0;
     $448 = (($447) + ($443)|0);
     $449 = HEAP8[$448>>0]|0;
     $450 = HEAP32[$29>>2]|0;
     $451 = $450 & $438;
     $452 = $441 >>> 8;
     $453 = $452 & 255;
     $454 = $451 >>> $453;
     $455 = $444 >>> 8;
     $456 = $455 & 255;
     $457 = (125924 + ($456<<2)|0);
     $458 = HEAP32[$457>>2]|0;
     $459 = (($458) + ($454)|0);
     $460 = HEAP8[$459>>0]|0;
     $461 = HEAP32[$30>>2]|0;
     $462 = $461 & $438;
     $463 = $441 >>> 16;
     $464 = $463 & 255;
     $465 = $462 >>> $464;
     $466 = $444 >>> 16;
     $467 = $466 & 255;
     $468 = (125924 + ($467<<2)|0);
     $469 = HEAP32[$468>>2]|0;
     $470 = (($469) + ($465)|0);
     $471 = HEAP8[$470>>0]|0;
     $472 = HEAP32[$31>>2]|0;
     $473 = $472 & $438;
     $474 = $441 >>> 24;
     $475 = $473 >>> $474;
     $476 = $444 >>> 24;
     $477 = (125924 + ($476<<2)|0);
     $478 = HEAP32[$477>>2]|0;
     $479 = (($478) + ($475)|0);
     $480 = HEAP8[$479>>0]|0;
     $sA$2$shrunk = $480;$sB$2$shrunk = $471;$sG$2$shrunk = $460;$sR$2$shrunk = $449;
     break;
    }
    case 3:  {
     $481 = HEAP32[$27>>2]|0;
     $div = $481 >>> 3;
     $482 = $div & 31;
     $483 = (($src$3) + ($482)|0);
     $484 = HEAP8[$483>>0]|0;
     $485 = $481 >>> 11;
     $div4 = $485 & 31;
     $486 = (($src$3) + ($div4)|0);
     $487 = HEAP8[$486>>0]|0;
     $488 = $481 >>> 19;
     $489 = $488 & 31;
     $490 = (($src$3) + ($489)|0);
     $491 = HEAP8[$490>>0]|0;
     $sA$2$shrunk = -1;$sB$2$shrunk = $491;$sG$2$shrunk = $487;$sR$2$shrunk = $484;
     break;
    }
    case 4:  {
     $492 = HEAP32[$src$3>>2]|0;
     $493 = HEAP32[$26>>2]|0;
     $494 = $493 & $492;
     $495 = HEAP32[$27>>2]|0;
     $496 = $495 & 255;
     $497 = $494 >>> $496;
     $498 = HEAP32[$28>>2]|0;
     $499 = $498 & 255;
     $500 = (125924 + ($499<<2)|0);
     $501 = HEAP32[$500>>2]|0;
     $502 = (($501) + ($497)|0);
     $503 = HEAP8[$502>>0]|0;
     $504 = HEAP32[$29>>2]|0;
     $505 = $504 & $492;
     $506 = $495 >>> 8;
     $507 = $506 & 255;
     $508 = $505 >>> $507;
     $509 = $498 >>> 8;
     $510 = $509 & 255;
     $511 = (125924 + ($510<<2)|0);
     $512 = HEAP32[$511>>2]|0;
     $513 = (($512) + ($508)|0);
     $514 = HEAP8[$513>>0]|0;
     $515 = HEAP32[$30>>2]|0;
     $516 = $515 & $492;
     $517 = $495 >>> 16;
     $518 = $517 & 255;
     $519 = $516 >>> $518;
     $520 = $498 >>> 16;
     $521 = $520 & 255;
     $522 = (125924 + ($521<<2)|0);
     $523 = HEAP32[$522>>2]|0;
     $524 = (($523) + ($519)|0);
     $525 = HEAP8[$524>>0]|0;
     $526 = HEAP32[$31>>2]|0;
     $527 = $526 & $492;
     $528 = $495 >>> 24;
     $529 = $527 >>> $528;
     $530 = $498 >>> 24;
     $531 = (125924 + ($530<<2)|0);
     $532 = HEAP32[$531>>2]|0;
     $533 = (($532) + ($529)|0);
     $534 = HEAP8[$533>>0]|0;
     $sA$2$shrunk = $534;$sB$2$shrunk = $525;$sG$2$shrunk = $514;$sR$2$shrunk = $503;
     break;
    }
    default: {
     $sA$2$shrunk = 0;$sB$2$shrunk = 0;$sG$2$shrunk = 0;$sR$2$shrunk = 0;
    }
    }
    $sA$2 = $sA$2$shrunk&255;
    $sB$2 = $sB$2$shrunk&255;
    $sG$2 = $sG$2$shrunk&255;
    $sR$2 = $sR$2$shrunk&255;
    $535 = HEAP8[$dst$3>>0]|0;
    $536 = $535&255;
    $537 = HEAP32[$24>>2]|0;
    $538 = ((($537)) + 4|0);
    $539 = HEAP32[$538>>2]|0;
    $540 = (($539) + ($536<<2)|0);
    $541 = HEAP8[$540>>0]|0;
    $542 = $541&255;
    $543 = (((($539) + ($536<<2)|0)) + 1|0);
    $544 = HEAP8[$543>>0]|0;
    $545 = $544&255;
    $546 = (((($539) + ($536<<2)|0)) + 2|0);
    $547 = HEAP8[$546>>0]|0;
    $548 = $547&255;
    $549 = (($sR$2) - ($542))|0;
    $550 = Math_imul($549, $sA$2)|0;
    $551 = (($550>>>0) / 255)&-1;
    $552 = (($551) + ($542))|0;
    $553 = (($sG$2) - ($545))|0;
    $554 = Math_imul($553, $sA$2)|0;
    $555 = (($554>>>0) / 255)&-1;
    $556 = (($555) + ($545))|0;
    $557 = (($sB$2) - ($548))|0;
    $558 = Math_imul($557, $sA$2)|0;
    $559 = (($558>>>0) / 255)&-1;
    $560 = (($559) + ($548))|0;
    $561 = $552 & 224;
    $562 = $556 >>> 3;
    $563 = $562 & 28;
    $564 = $563 | $561;
    $565 = $560 >>> 6;
    $566 = $565 & 3;
    $567 = $564 | $566;
    if ($25) {
     $568 = $567&255;
     $storemerge3 = $568;
    } else {
     $569 = (($9) + ($567)|0);
     $570 = HEAP8[$569>>0]|0;
     $storemerge3 = $570;
    }
    HEAP8[$dst$3>>0] = $storemerge3;
    $571 = ((($dst$3)) + 1|0);
    $572 = (($src$3) + ($14)|0);
    $dst$4 = $571;$n$3 = $n$2;$src$4 = $572;
    label = 31;
    continue;
   }
   else if ((label|0) == 31) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $573 = HEAP8[$src$4>>0]|0;
     $574 = $573&255;
     $575 = HEAP32[$26>>2]|0;
     $576 = $574 & $575;
     $577 = HEAP32[$27>>2]|0;
     $578 = $577 & 255;
     $579 = $576 >>> $578;
     $580 = HEAP32[$28>>2]|0;
     $581 = $580 & 255;
     $582 = (125924 + ($581<<2)|0);
     $583 = HEAP32[$582>>2]|0;
     $584 = (($583) + ($579)|0);
     $585 = HEAP8[$584>>0]|0;
     $586 = HEAP32[$29>>2]|0;
     $587 = $586 & $574;
     $588 = $577 >>> 8;
     $589 = $588 & 255;
     $590 = $587 >>> $589;
     $591 = $580 >>> 8;
     $592 = $591 & 255;
     $593 = (125924 + ($592<<2)|0);
     $594 = HEAP32[$593>>2]|0;
     $595 = (($594) + ($590)|0);
     $596 = HEAP8[$595>>0]|0;
     $597 = HEAP32[$30>>2]|0;
     $598 = $597 & $574;
     $599 = $577 >>> 16;
     $600 = $599 & 255;
     $601 = $598 >>> $600;
     $602 = $580 >>> 16;
     $603 = $602 & 255;
     $604 = (125924 + ($603<<2)|0);
     $605 = HEAP32[$604>>2]|0;
     $606 = (($605) + ($601)|0);
     $607 = HEAP8[$606>>0]|0;
     $608 = HEAP32[$31>>2]|0;
     $609 = $608 & $574;
     $610 = $577 >>> 24;
     $611 = $609 >>> $610;
     $612 = $580 >>> 24;
     $613 = (125924 + ($612<<2)|0);
     $614 = HEAP32[$613>>2]|0;
     $615 = (($614) + ($611)|0);
     $616 = HEAP8[$615>>0]|0;
     $sA$3$shrunk = $616;$sB$3$shrunk = $607;$sG$3$shrunk = $596;$sR$3$shrunk = $585;
     break;
    }
    case 2:  {
     $617 = HEAP16[$src$4>>1]|0;
     $618 = $617&65535;
     $619 = HEAP32[$26>>2]|0;
     $620 = $618 & $619;
     $621 = HEAP32[$27>>2]|0;
     $622 = $621 & 255;
     $623 = $620 >>> $622;
     $624 = HEAP32[$28>>2]|0;
     $625 = $624 & 255;
     $626 = (125924 + ($625<<2)|0);
     $627 = HEAP32[$626>>2]|0;
     $628 = (($627) + ($623)|0);
     $629 = HEAP8[$628>>0]|0;
     $630 = HEAP32[$29>>2]|0;
     $631 = $630 & $618;
     $632 = $621 >>> 8;
     $633 = $632 & 255;
     $634 = $631 >>> $633;
     $635 = $624 >>> 8;
     $636 = $635 & 255;
     $637 = (125924 + ($636<<2)|0);
     $638 = HEAP32[$637>>2]|0;
     $639 = (($638) + ($634)|0);
     $640 = HEAP8[$639>>0]|0;
     $641 = HEAP32[$30>>2]|0;
     $642 = $641 & $618;
     $643 = $621 >>> 16;
     $644 = $643 & 255;
     $645 = $642 >>> $644;
     $646 = $624 >>> 16;
     $647 = $646 & 255;
     $648 = (125924 + ($647<<2)|0);
     $649 = HEAP32[$648>>2]|0;
     $650 = (($649) + ($645)|0);
     $651 = HEAP8[$650>>0]|0;
     $652 = HEAP32[$31>>2]|0;
     $653 = $652 & $618;
     $654 = $621 >>> 24;
     $655 = $653 >>> $654;
     $656 = $624 >>> 24;
     $657 = (125924 + ($656<<2)|0);
     $658 = HEAP32[$657>>2]|0;
     $659 = (($658) + ($655)|0);
     $660 = HEAP8[$659>>0]|0;
     $sA$3$shrunk = $660;$sB$3$shrunk = $651;$sG$3$shrunk = $640;$sR$3$shrunk = $629;
     break;
    }
    case 3:  {
     $661 = HEAP32[$27>>2]|0;
     $div12 = $661 >>> 3;
     $662 = $div12 & 31;
     $663 = (($src$4) + ($662)|0);
     $664 = HEAP8[$663>>0]|0;
     $665 = $661 >>> 11;
     $div13 = $665 & 31;
     $666 = (($src$4) + ($div13)|0);
     $667 = HEAP8[$666>>0]|0;
     $668 = $661 >>> 19;
     $669 = $668 & 31;
     $670 = (($src$4) + ($669)|0);
     $671 = HEAP8[$670>>0]|0;
     $sA$3$shrunk = -1;$sB$3$shrunk = $671;$sG$3$shrunk = $667;$sR$3$shrunk = $664;
     break;
    }
    case 4:  {
     $672 = HEAP32[$src$4>>2]|0;
     $673 = HEAP32[$26>>2]|0;
     $674 = $673 & $672;
     $675 = HEAP32[$27>>2]|0;
     $676 = $675 & 255;
     $677 = $674 >>> $676;
     $678 = HEAP32[$28>>2]|0;
     $679 = $678 & 255;
     $680 = (125924 + ($679<<2)|0);
     $681 = HEAP32[$680>>2]|0;
     $682 = (($681) + ($677)|0);
     $683 = HEAP8[$682>>0]|0;
     $684 = HEAP32[$29>>2]|0;
     $685 = $684 & $672;
     $686 = $675 >>> 8;
     $687 = $686 & 255;
     $688 = $685 >>> $687;
     $689 = $678 >>> 8;
     $690 = $689 & 255;
     $691 = (125924 + ($690<<2)|0);
     $692 = HEAP32[$691>>2]|0;
     $693 = (($692) + ($688)|0);
     $694 = HEAP8[$693>>0]|0;
     $695 = HEAP32[$30>>2]|0;
     $696 = $695 & $672;
     $697 = $675 >>> 16;
     $698 = $697 & 255;
     $699 = $696 >>> $698;
     $700 = $678 >>> 16;
     $701 = $700 & 255;
     $702 = (125924 + ($701<<2)|0);
     $703 = HEAP32[$702>>2]|0;
     $704 = (($703) + ($699)|0);
     $705 = HEAP8[$704>>0]|0;
     $706 = HEAP32[$31>>2]|0;
     $707 = $706 & $672;
     $708 = $675 >>> 24;
     $709 = $707 >>> $708;
     $710 = $678 >>> 24;
     $711 = (125924 + ($710<<2)|0);
     $712 = HEAP32[$711>>2]|0;
     $713 = (($712) + ($709)|0);
     $714 = HEAP8[$713>>0]|0;
     $sA$3$shrunk = $714;$sB$3$shrunk = $705;$sG$3$shrunk = $694;$sR$3$shrunk = $683;
     break;
    }
    default: {
     $sA$3$shrunk = 0;$sB$3$shrunk = 0;$sG$3$shrunk = 0;$sR$3$shrunk = 0;
    }
    }
    $sA$3 = $sA$3$shrunk&255;
    $sB$3 = $sB$3$shrunk&255;
    $sG$3 = $sG$3$shrunk&255;
    $sR$3 = $sR$3$shrunk&255;
    $715 = HEAP8[$dst$4>>0]|0;
    $716 = $715&255;
    $717 = HEAP32[$24>>2]|0;
    $718 = ((($717)) + 4|0);
    $719 = HEAP32[$718>>2]|0;
    $720 = (($719) + ($716<<2)|0);
    $721 = HEAP8[$720>>0]|0;
    $722 = $721&255;
    $723 = (((($719) + ($716<<2)|0)) + 1|0);
    $724 = HEAP8[$723>>0]|0;
    $725 = $724&255;
    $726 = (((($719) + ($716<<2)|0)) + 2|0);
    $727 = HEAP8[$726>>0]|0;
    $728 = $727&255;
    $729 = (($sR$3) - ($722))|0;
    $730 = Math_imul($729, $sA$3)|0;
    $731 = (($730>>>0) / 255)&-1;
    $732 = (($731) + ($722))|0;
    $733 = (($sG$3) - ($725))|0;
    $734 = Math_imul($733, $sA$3)|0;
    $735 = (($734>>>0) / 255)&-1;
    $736 = (($735) + ($725))|0;
    $737 = (($sB$3) - ($728))|0;
    $738 = Math_imul($737, $sA$3)|0;
    $739 = (($738>>>0) / 255)&-1;
    $740 = (($739) + ($728))|0;
    $741 = $732 & 224;
    $742 = $736 >>> 3;
    $743 = $742 & 28;
    $744 = $743 | $741;
    $745 = $740 >>> 6;
    $746 = $745 & 3;
    $747 = $744 | $746;
    if ($25) {
     $748 = $747&255;
     $storemerge = $748;
    } else {
     $749 = (($9) + ($747)|0);
     $750 = HEAP8[$749>>0]|0;
     $storemerge = $750;
    }
    HEAP8[$dst$4>>0] = $storemerge;
    $751 = ((($dst$4)) + 1|0);
    $752 = (($src$4) + ($14)|0);
    $753 = (($n$3) + -1)|0;
    $754 = ($n$3|0)>(1);
    if ($754) {
     $dst$1 = $751;$n$0 = $753;$src$1 = $752;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $755 = (($752) + ($5)|0);
  $756 = (($751) + ($7)|0);
  $757 = ($32|0)==(0);
  if ($757) {
   label = 42;
   break;
  } else {
   $$in = $32;$dst$015 = $756;$src$016 = $755;
  }
 }
 if ((label|0) == 40) {
  // unreachable;
 }
 else if ((label|0) == 42) {
  return;
 }
}
function _BlitRGBtoRGBPixelAlpha($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dstp$01 = 0, $dstp$1 = 0, $dstp$2 = 0, $dstp$3 = 0, $dstp$4 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0;
 var $n$3 = 0, $srcp$02 = 0, $srcp$1 = 0, $srcp$2 = 0, $srcp$3 = 0, $srcp$4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $5 >> 2;
 $7 = ((($info)) + 36|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $8 >> 2;
 $10 = ($3|0)==(0);
 if ($10) {
  return;
 }
 $11 = ((($info)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = HEAP32[$info>>2]|0;
 $14 = (($1) + 3)|0;
 $15 = (($14|0) / 4)&-1;
 $16 = $1 & 3;
 $$in = $3;$dstp$01 = $12;$srcp$02 = $13;
 L4: while(1) {
  $17 = (($$in) + -1)|0;
  switch ($16|0) {
  case 0:  {
   $dstp$1 = $dstp$01;$n$0 = $15;$srcp$1 = $srcp$02;
   label = 4;
   break;
  }
  case 3:  {
   $dstp$2 = $dstp$01;$n$1 = $15;$srcp$2 = $srcp$02;
   label = 8;
   break;
  }
  case 2:  {
   $dstp$3 = $dstp$01;$n$2 = $15;$srcp$3 = $srcp$02;
   label = 12;
   break;
  }
  case 1:  {
   $dstp$4 = $dstp$01;$n$3 = $15;$srcp$4 = $srcp$02;
   label = 16;
   break;
  }
  default: {
   label = 20;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    $18 = HEAP32[$srcp$1>>2]|0;
    $19 = $18 >>> 24;
    switch ($19|0) {
    case 0:  {
     break;
    }
    case 255:  {
     HEAP32[$dstp$1>>2] = $18;
     break;
    }
    default: {
     $20 = HEAP32[$dstp$1>>2]|0;
     $21 = $20 >>> 24;
     $22 = $18 & 16711935;
     $23 = $20 & 16711935;
     $24 = (($22) - ($23))|0;
     $25 = Math_imul($24, $19)|0;
     $26 = $25 >>> 8;
     $27 = (($26) + ($23))|0;
     $28 = $27 & 16711935;
     $29 = $18 & 65280;
     $30 = $20 & 65280;
     $31 = (($29) - ($30))|0;
     $32 = Math_imul($31, $19)|0;
     $33 = $32 >>> 8;
     $34 = (($33) + ($30))|0;
     $35 = $34 & 65280;
     $36 = $19 ^ 255;
     $37 = Math_imul($21, $36)|0;
     $38 = $37 >>> 8;
     $39 = (($38) + ($19))|0;
     $40 = $39 << 24;
     $41 = $35 | $40;
     $42 = $41 | $28;
     HEAP32[$dstp$1>>2] = $42;
    }
    }
    $43 = ((($srcp$1)) + 4|0);
    $44 = ((($dstp$1)) + 4|0);
    $dstp$2 = $44;$n$1 = $n$0;$srcp$2 = $43;
    label = 8;
    continue;
   }
   else if ((label|0) == 8) {
    label = 0;
    $45 = HEAP32[$srcp$2>>2]|0;
    $46 = $45 >>> 24;
    switch ($46|0) {
    case 0:  {
     break;
    }
    case 255:  {
     HEAP32[$dstp$2>>2] = $45;
     break;
    }
    default: {
     $47 = HEAP32[$dstp$2>>2]|0;
     $48 = $47 >>> 24;
     $49 = $45 & 16711935;
     $50 = $47 & 16711935;
     $51 = (($49) - ($50))|0;
     $52 = Math_imul($51, $46)|0;
     $53 = $52 >>> 8;
     $54 = (($53) + ($50))|0;
     $55 = $54 & 16711935;
     $56 = $45 & 65280;
     $57 = $47 & 65280;
     $58 = (($56) - ($57))|0;
     $59 = Math_imul($58, $46)|0;
     $60 = $59 >>> 8;
     $61 = (($60) + ($57))|0;
     $62 = $61 & 65280;
     $63 = $46 ^ 255;
     $64 = Math_imul($48, $63)|0;
     $65 = $64 >>> 8;
     $66 = (($65) + ($46))|0;
     $67 = $66 << 24;
     $68 = $62 | $67;
     $69 = $68 | $55;
     HEAP32[$dstp$2>>2] = $69;
    }
    }
    $70 = ((($srcp$2)) + 4|0);
    $71 = ((($dstp$2)) + 4|0);
    $dstp$3 = $71;$n$2 = $n$1;$srcp$3 = $70;
    label = 12;
    continue;
   }
   else if ((label|0) == 12) {
    label = 0;
    $72 = HEAP32[$srcp$3>>2]|0;
    $73 = $72 >>> 24;
    switch ($73|0) {
    case 0:  {
     break;
    }
    case 255:  {
     HEAP32[$dstp$3>>2] = $72;
     break;
    }
    default: {
     $74 = HEAP32[$dstp$3>>2]|0;
     $75 = $74 >>> 24;
     $76 = $72 & 16711935;
     $77 = $74 & 16711935;
     $78 = (($76) - ($77))|0;
     $79 = Math_imul($78, $73)|0;
     $80 = $79 >>> 8;
     $81 = (($80) + ($77))|0;
     $82 = $81 & 16711935;
     $83 = $72 & 65280;
     $84 = $74 & 65280;
     $85 = (($83) - ($84))|0;
     $86 = Math_imul($85, $73)|0;
     $87 = $86 >>> 8;
     $88 = (($87) + ($84))|0;
     $89 = $88 & 65280;
     $90 = $73 ^ 255;
     $91 = Math_imul($75, $90)|0;
     $92 = $91 >>> 8;
     $93 = (($92) + ($73))|0;
     $94 = $93 << 24;
     $95 = $89 | $94;
     $96 = $95 | $82;
     HEAP32[$dstp$3>>2] = $96;
    }
    }
    $97 = ((($srcp$3)) + 4|0);
    $98 = ((($dstp$3)) + 4|0);
    $dstp$4 = $98;$n$3 = $n$2;$srcp$4 = $97;
    label = 16;
    continue;
   }
   else if ((label|0) == 16) {
    label = 0;
    $99 = HEAP32[$srcp$4>>2]|0;
    $100 = $99 >>> 24;
    switch ($100|0) {
    case 0:  {
     break;
    }
    case 255:  {
     HEAP32[$dstp$4>>2] = $99;
     break;
    }
    default: {
     $101 = HEAP32[$dstp$4>>2]|0;
     $102 = $101 >>> 24;
     $103 = $99 & 16711935;
     $104 = $101 & 16711935;
     $105 = (($103) - ($104))|0;
     $106 = Math_imul($105, $100)|0;
     $107 = $106 >>> 8;
     $108 = (($107) + ($104))|0;
     $109 = $108 & 16711935;
     $110 = $99 & 65280;
     $111 = $101 & 65280;
     $112 = (($110) - ($111))|0;
     $113 = Math_imul($112, $100)|0;
     $114 = $113 >>> 8;
     $115 = (($114) + ($111))|0;
     $116 = $115 & 65280;
     $117 = $100 ^ 255;
     $118 = Math_imul($102, $117)|0;
     $119 = $118 >>> 8;
     $120 = (($119) + ($100))|0;
     $121 = $120 << 24;
     $122 = $116 | $121;
     $123 = $122 | $109;
     HEAP32[$dstp$4>>2] = $123;
    }
    }
    $124 = ((($srcp$4)) + 4|0);
    $125 = ((($dstp$4)) + 4|0);
    $126 = (($n$3) + -1)|0;
    $127 = ($n$3|0)>(1);
    if ($127) {
     $dstp$1 = $125;$n$0 = $126;$srcp$1 = $124;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $128 = (($124) + ($6<<2)|0);
  $129 = (($125) + ($9<<2)|0);
  $130 = ($17|0)==(0);
  if ($130) {
   label = 22;
   break;
  } else {
   $$in = $17;$dstp$01 = $129;$srcp$02 = $128;
  }
 }
 if ((label|0) == 20) {
  // unreachable;
 }
 else if ((label|0) == 22) {
  return;
 }
}
function _BlitNto1SurfaceAlpha($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0;
 var $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0;
 var $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0;
 var $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0;
 var $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0;
 var $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0;
 var $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0;
 var $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0;
 var $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0;
 var $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0;
 var $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0;
 var $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0;
 var $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0;
 var $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0;
 var $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0;
 var $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0;
 var $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0;
 var $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0;
 var $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0;
 var $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0;
 var $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0;
 var $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0;
 var $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0;
 var $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $div = 0, $div10 = 0, $div12 = 0, $div13 = 0, $div4 = 0, $div6 = 0, $div7 = 0;
 var $div9 = 0, $dst$015 = 0, $dst$1 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $sB$0 = 0, $sB$0$shrunk = 0, $sB$1 = 0, $sB$1$shrunk = 0, $sB$2 = 0, $sB$2$shrunk = 0, $sB$3 = 0, $sB$3$shrunk = 0, $sG$0 = 0, $sG$0$shrunk = 0;
 var $sG$1 = 0, $sG$1$shrunk = 0, $sG$2 = 0, $sG$2$shrunk = 0, $sG$3 = 0, $sG$3$shrunk = 0, $sR$0 = 0, $sR$0$shrunk = 0, $sR$1 = 0, $sR$1$shrunk = 0, $sR$2 = 0, $sR$2$shrunk = 0, $sR$3 = 0, $sR$3$shrunk = 0, $src$016 = 0, $src$1 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, $storemerge = 0;
 var $storemerge1 = 0, $storemerge2 = 0, $storemerge3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 48|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($info)) + 40|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($11)) + 9|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = $13&255;
 $15 = ((($info)) + 63|0);
 $16 = HEAP8[$15>>0]|0;
 $17 = $16&255;
 $18 = ($3|0)==(0);
 if ($18) {
  return;
 }
 $19 = ((($info)) + 44|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ((($info)) + 20|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = HEAP32[$info>>2]|0;
 $24 = (($1) + 3)|0;
 $25 = (($24|0) / 4)&-1;
 $26 = $1 & 3;
 $27 = ((($20)) + 4|0);
 $28 = ($9|0)==(0|0);
 $29 = ((($11)) + 12|0);
 $30 = ((($11)) + 32|0);
 $31 = ((($11)) + 28|0);
 $32 = ((($11)) + 16|0);
 $33 = ((($11)) + 20|0);
 $$in = $3;$dst$015 = $22;$src$016 = $23;
 L4: while(1) {
  $34 = (($$in) + -1)|0;
  switch ($26|0) {
  case 0:  {
   $dst$1 = $dst$015;$n$0 = $25;$src$1 = $src$016;
   label = 4;
   break;
  }
  case 3:  {
   $dst$2 = $dst$015;$n$1 = $25;$src$2 = $src$016;
   label = 13;
   break;
  }
  case 2:  {
   $dst$3 = $dst$015;$n$2 = $25;$src$3 = $src$016;
   label = 22;
   break;
  }
  case 1:  {
   $dst$4 = $dst$015;$n$3 = $25;$src$4 = $src$016;
   label = 31;
   break;
  }
  default: {
   label = 40;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $35 = HEAP8[$src$1>>0]|0;
     $36 = $35&255;
     $37 = HEAP32[$29>>2]|0;
     $38 = $36 & $37;
     $39 = HEAP32[$30>>2]|0;
     $40 = $39 & 255;
     $41 = $38 >>> $40;
     $42 = HEAP32[$31>>2]|0;
     $43 = $42 & 255;
     $44 = (125924 + ($43<<2)|0);
     $45 = HEAP32[$44>>2]|0;
     $46 = (($45) + ($41)|0);
     $47 = HEAP8[$46>>0]|0;
     $48 = HEAP32[$32>>2]|0;
     $49 = $48 & $36;
     $50 = $39 >>> 8;
     $51 = $50 & 255;
     $52 = $49 >>> $51;
     $53 = $42 >>> 8;
     $54 = $53 & 255;
     $55 = (125924 + ($54<<2)|0);
     $56 = HEAP32[$55>>2]|0;
     $57 = (($56) + ($52)|0);
     $58 = HEAP8[$57>>0]|0;
     $59 = HEAP32[$33>>2]|0;
     $60 = $59 & $36;
     $61 = $39 >>> 16;
     $62 = $61 & 255;
     $63 = $60 >>> $62;
     $64 = $42 >>> 16;
     $65 = $64 & 255;
     $66 = (125924 + ($65<<2)|0);
     $67 = HEAP32[$66>>2]|0;
     $68 = (($67) + ($63)|0);
     $69 = HEAP8[$68>>0]|0;
     $sB$0$shrunk = $69;$sG$0$shrunk = $58;$sR$0$shrunk = $47;
     break;
    }
    case 2:  {
     $70 = HEAP16[$src$1>>1]|0;
     $71 = $70&65535;
     $72 = HEAP32[$29>>2]|0;
     $73 = $71 & $72;
     $74 = HEAP32[$30>>2]|0;
     $75 = $74 & 255;
     $76 = $73 >>> $75;
     $77 = HEAP32[$31>>2]|0;
     $78 = $77 & 255;
     $79 = (125924 + ($78<<2)|0);
     $80 = HEAP32[$79>>2]|0;
     $81 = (($80) + ($76)|0);
     $82 = HEAP8[$81>>0]|0;
     $83 = HEAP32[$32>>2]|0;
     $84 = $83 & $71;
     $85 = $74 >>> 8;
     $86 = $85 & 255;
     $87 = $84 >>> $86;
     $88 = $77 >>> 8;
     $89 = $88 & 255;
     $90 = (125924 + ($89<<2)|0);
     $91 = HEAP32[$90>>2]|0;
     $92 = (($91) + ($87)|0);
     $93 = HEAP8[$92>>0]|0;
     $94 = HEAP32[$33>>2]|0;
     $95 = $94 & $71;
     $96 = $74 >>> 16;
     $97 = $96 & 255;
     $98 = $95 >>> $97;
     $99 = $77 >>> 16;
     $100 = $99 & 255;
     $101 = (125924 + ($100<<2)|0);
     $102 = HEAP32[$101>>2]|0;
     $103 = (($102) + ($98)|0);
     $104 = HEAP8[$103>>0]|0;
     $sB$0$shrunk = $104;$sG$0$shrunk = $93;$sR$0$shrunk = $82;
     break;
    }
    case 3:  {
     $105 = HEAP32[$30>>2]|0;
     $div9 = $105 >>> 3;
     $106 = $div9 & 31;
     $107 = (($src$1) + ($106)|0);
     $108 = HEAP8[$107>>0]|0;
     $109 = $105 >>> 11;
     $div10 = $109 & 31;
     $110 = (($src$1) + ($div10)|0);
     $111 = HEAP8[$110>>0]|0;
     $112 = $105 >>> 19;
     $113 = $112 & 31;
     $114 = (($src$1) + ($113)|0);
     $115 = HEAP8[$114>>0]|0;
     $sB$0$shrunk = $115;$sG$0$shrunk = $111;$sR$0$shrunk = $108;
     break;
    }
    case 4:  {
     $116 = HEAP32[$src$1>>2]|0;
     $117 = HEAP32[$29>>2]|0;
     $118 = $117 & $116;
     $119 = HEAP32[$30>>2]|0;
     $120 = $119 & 255;
     $121 = $118 >>> $120;
     $122 = HEAP32[$31>>2]|0;
     $123 = $122 & 255;
     $124 = (125924 + ($123<<2)|0);
     $125 = HEAP32[$124>>2]|0;
     $126 = (($125) + ($121)|0);
     $127 = HEAP8[$126>>0]|0;
     $128 = HEAP32[$32>>2]|0;
     $129 = $128 & $116;
     $130 = $119 >>> 8;
     $131 = $130 & 255;
     $132 = $129 >>> $131;
     $133 = $122 >>> 8;
     $134 = $133 & 255;
     $135 = (125924 + ($134<<2)|0);
     $136 = HEAP32[$135>>2]|0;
     $137 = (($136) + ($132)|0);
     $138 = HEAP8[$137>>0]|0;
     $139 = HEAP32[$33>>2]|0;
     $140 = $139 & $116;
     $141 = $119 >>> 16;
     $142 = $141 & 255;
     $143 = $140 >>> $142;
     $144 = $122 >>> 16;
     $145 = $144 & 255;
     $146 = (125924 + ($145<<2)|0);
     $147 = HEAP32[$146>>2]|0;
     $148 = (($147) + ($143)|0);
     $149 = HEAP8[$148>>0]|0;
     $sB$0$shrunk = $149;$sG$0$shrunk = $138;$sR$0$shrunk = $127;
     break;
    }
    default: {
     $sB$0$shrunk = 0;$sG$0$shrunk = 0;$sR$0$shrunk = 0;
    }
    }
    $sB$0 = $sB$0$shrunk&255;
    $sG$0 = $sG$0$shrunk&255;
    $sR$0 = $sR$0$shrunk&255;
    $150 = HEAP8[$dst$1>>0]|0;
    $151 = $150&255;
    $152 = HEAP32[$27>>2]|0;
    $153 = ((($152)) + 4|0);
    $154 = HEAP32[$153>>2]|0;
    $155 = (($154) + ($151<<2)|0);
    $156 = HEAP8[$155>>0]|0;
    $157 = $156&255;
    $158 = (((($154) + ($151<<2)|0)) + 1|0);
    $159 = HEAP8[$158>>0]|0;
    $160 = $159&255;
    $161 = (((($154) + ($151<<2)|0)) + 2|0);
    $162 = HEAP8[$161>>0]|0;
    $163 = $162&255;
    $164 = (($sR$0) - ($157))|0;
    $165 = Math_imul($164, $17)|0;
    $166 = (($165>>>0) / 255)&-1;
    $167 = (($166) + ($157))|0;
    $168 = (($sG$0) - ($160))|0;
    $169 = Math_imul($168, $17)|0;
    $170 = (($169>>>0) / 255)&-1;
    $171 = (($170) + ($160))|0;
    $172 = (($sB$0) - ($163))|0;
    $173 = Math_imul($172, $17)|0;
    $174 = (($173>>>0) / 255)&-1;
    $175 = (($174) + ($163))|0;
    $176 = $167 & 224;
    $177 = $171 >>> 3;
    $178 = $177 & 28;
    $179 = $178 | $176;
    $180 = $175 >>> 6;
    $181 = $180 & 3;
    $182 = $179 | $181;
    if ($28) {
     $183 = $182&255;
     $storemerge1 = $183;
    } else {
     $184 = (($9) + ($182)|0);
     $185 = HEAP8[$184>>0]|0;
     $storemerge1 = $185;
    }
    HEAP8[$dst$1>>0] = $storemerge1;
    $186 = ((($dst$1)) + 1|0);
    $187 = (($src$1) + ($14)|0);
    $dst$2 = $186;$n$1 = $n$0;$src$2 = $187;
    label = 13;
    continue;
   }
   else if ((label|0) == 13) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $188 = HEAP8[$src$2>>0]|0;
     $189 = $188&255;
     $190 = HEAP32[$29>>2]|0;
     $191 = $189 & $190;
     $192 = HEAP32[$30>>2]|0;
     $193 = $192 & 255;
     $194 = $191 >>> $193;
     $195 = HEAP32[$31>>2]|0;
     $196 = $195 & 255;
     $197 = (125924 + ($196<<2)|0);
     $198 = HEAP32[$197>>2]|0;
     $199 = (($198) + ($194)|0);
     $200 = HEAP8[$199>>0]|0;
     $201 = HEAP32[$32>>2]|0;
     $202 = $201 & $189;
     $203 = $192 >>> 8;
     $204 = $203 & 255;
     $205 = $202 >>> $204;
     $206 = $195 >>> 8;
     $207 = $206 & 255;
     $208 = (125924 + ($207<<2)|0);
     $209 = HEAP32[$208>>2]|0;
     $210 = (($209) + ($205)|0);
     $211 = HEAP8[$210>>0]|0;
     $212 = HEAP32[$33>>2]|0;
     $213 = $212 & $189;
     $214 = $192 >>> 16;
     $215 = $214 & 255;
     $216 = $213 >>> $215;
     $217 = $195 >>> 16;
     $218 = $217 & 255;
     $219 = (125924 + ($218<<2)|0);
     $220 = HEAP32[$219>>2]|0;
     $221 = (($220) + ($216)|0);
     $222 = HEAP8[$221>>0]|0;
     $sB$1$shrunk = $222;$sG$1$shrunk = $211;$sR$1$shrunk = $200;
     break;
    }
    case 2:  {
     $223 = HEAP16[$src$2>>1]|0;
     $224 = $223&65535;
     $225 = HEAP32[$29>>2]|0;
     $226 = $224 & $225;
     $227 = HEAP32[$30>>2]|0;
     $228 = $227 & 255;
     $229 = $226 >>> $228;
     $230 = HEAP32[$31>>2]|0;
     $231 = $230 & 255;
     $232 = (125924 + ($231<<2)|0);
     $233 = HEAP32[$232>>2]|0;
     $234 = (($233) + ($229)|0);
     $235 = HEAP8[$234>>0]|0;
     $236 = HEAP32[$32>>2]|0;
     $237 = $236 & $224;
     $238 = $227 >>> 8;
     $239 = $238 & 255;
     $240 = $237 >>> $239;
     $241 = $230 >>> 8;
     $242 = $241 & 255;
     $243 = (125924 + ($242<<2)|0);
     $244 = HEAP32[$243>>2]|0;
     $245 = (($244) + ($240)|0);
     $246 = HEAP8[$245>>0]|0;
     $247 = HEAP32[$33>>2]|0;
     $248 = $247 & $224;
     $249 = $227 >>> 16;
     $250 = $249 & 255;
     $251 = $248 >>> $250;
     $252 = $230 >>> 16;
     $253 = $252 & 255;
     $254 = (125924 + ($253<<2)|0);
     $255 = HEAP32[$254>>2]|0;
     $256 = (($255) + ($251)|0);
     $257 = HEAP8[$256>>0]|0;
     $sB$1$shrunk = $257;$sG$1$shrunk = $246;$sR$1$shrunk = $235;
     break;
    }
    case 3:  {
     $258 = HEAP32[$30>>2]|0;
     $div6 = $258 >>> 3;
     $259 = $div6 & 31;
     $260 = (($src$2) + ($259)|0);
     $261 = HEAP8[$260>>0]|0;
     $262 = $258 >>> 11;
     $div7 = $262 & 31;
     $263 = (($src$2) + ($div7)|0);
     $264 = HEAP8[$263>>0]|0;
     $265 = $258 >>> 19;
     $266 = $265 & 31;
     $267 = (($src$2) + ($266)|0);
     $268 = HEAP8[$267>>0]|0;
     $sB$1$shrunk = $268;$sG$1$shrunk = $264;$sR$1$shrunk = $261;
     break;
    }
    case 4:  {
     $269 = HEAP32[$src$2>>2]|0;
     $270 = HEAP32[$29>>2]|0;
     $271 = $270 & $269;
     $272 = HEAP32[$30>>2]|0;
     $273 = $272 & 255;
     $274 = $271 >>> $273;
     $275 = HEAP32[$31>>2]|0;
     $276 = $275 & 255;
     $277 = (125924 + ($276<<2)|0);
     $278 = HEAP32[$277>>2]|0;
     $279 = (($278) + ($274)|0);
     $280 = HEAP8[$279>>0]|0;
     $281 = HEAP32[$32>>2]|0;
     $282 = $281 & $269;
     $283 = $272 >>> 8;
     $284 = $283 & 255;
     $285 = $282 >>> $284;
     $286 = $275 >>> 8;
     $287 = $286 & 255;
     $288 = (125924 + ($287<<2)|0);
     $289 = HEAP32[$288>>2]|0;
     $290 = (($289) + ($285)|0);
     $291 = HEAP8[$290>>0]|0;
     $292 = HEAP32[$33>>2]|0;
     $293 = $292 & $269;
     $294 = $272 >>> 16;
     $295 = $294 & 255;
     $296 = $293 >>> $295;
     $297 = $275 >>> 16;
     $298 = $297 & 255;
     $299 = (125924 + ($298<<2)|0);
     $300 = HEAP32[$299>>2]|0;
     $301 = (($300) + ($296)|0);
     $302 = HEAP8[$301>>0]|0;
     $sB$1$shrunk = $302;$sG$1$shrunk = $291;$sR$1$shrunk = $280;
     break;
    }
    default: {
     $sB$1$shrunk = 0;$sG$1$shrunk = 0;$sR$1$shrunk = 0;
    }
    }
    $sB$1 = $sB$1$shrunk&255;
    $sG$1 = $sG$1$shrunk&255;
    $sR$1 = $sR$1$shrunk&255;
    $303 = HEAP8[$dst$2>>0]|0;
    $304 = $303&255;
    $305 = HEAP32[$27>>2]|0;
    $306 = ((($305)) + 4|0);
    $307 = HEAP32[$306>>2]|0;
    $308 = (($307) + ($304<<2)|0);
    $309 = HEAP8[$308>>0]|0;
    $310 = $309&255;
    $311 = (((($307) + ($304<<2)|0)) + 1|0);
    $312 = HEAP8[$311>>0]|0;
    $313 = $312&255;
    $314 = (((($307) + ($304<<2)|0)) + 2|0);
    $315 = HEAP8[$314>>0]|0;
    $316 = $315&255;
    $317 = (($sR$1) - ($310))|0;
    $318 = Math_imul($317, $17)|0;
    $319 = (($318>>>0) / 255)&-1;
    $320 = (($319) + ($310))|0;
    $321 = (($sG$1) - ($313))|0;
    $322 = Math_imul($321, $17)|0;
    $323 = (($322>>>0) / 255)&-1;
    $324 = (($323) + ($313))|0;
    $325 = (($sB$1) - ($316))|0;
    $326 = Math_imul($325, $17)|0;
    $327 = (($326>>>0) / 255)&-1;
    $328 = (($327) + ($316))|0;
    $329 = $320 & 224;
    $330 = $324 >>> 3;
    $331 = $330 & 28;
    $332 = $331 | $329;
    $333 = $328 >>> 6;
    $334 = $333 & 3;
    $335 = $332 | $334;
    if ($28) {
     $336 = $335&255;
     $storemerge2 = $336;
    } else {
     $337 = (($9) + ($335)|0);
     $338 = HEAP8[$337>>0]|0;
     $storemerge2 = $338;
    }
    HEAP8[$dst$2>>0] = $storemerge2;
    $339 = ((($dst$2)) + 1|0);
    $340 = (($src$2) + ($14)|0);
    $dst$3 = $339;$n$2 = $n$1;$src$3 = $340;
    label = 22;
    continue;
   }
   else if ((label|0) == 22) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $341 = HEAP8[$src$3>>0]|0;
     $342 = $341&255;
     $343 = HEAP32[$29>>2]|0;
     $344 = $342 & $343;
     $345 = HEAP32[$30>>2]|0;
     $346 = $345 & 255;
     $347 = $344 >>> $346;
     $348 = HEAP32[$31>>2]|0;
     $349 = $348 & 255;
     $350 = (125924 + ($349<<2)|0);
     $351 = HEAP32[$350>>2]|0;
     $352 = (($351) + ($347)|0);
     $353 = HEAP8[$352>>0]|0;
     $354 = HEAP32[$32>>2]|0;
     $355 = $354 & $342;
     $356 = $345 >>> 8;
     $357 = $356 & 255;
     $358 = $355 >>> $357;
     $359 = $348 >>> 8;
     $360 = $359 & 255;
     $361 = (125924 + ($360<<2)|0);
     $362 = HEAP32[$361>>2]|0;
     $363 = (($362) + ($358)|0);
     $364 = HEAP8[$363>>0]|0;
     $365 = HEAP32[$33>>2]|0;
     $366 = $365 & $342;
     $367 = $345 >>> 16;
     $368 = $367 & 255;
     $369 = $366 >>> $368;
     $370 = $348 >>> 16;
     $371 = $370 & 255;
     $372 = (125924 + ($371<<2)|0);
     $373 = HEAP32[$372>>2]|0;
     $374 = (($373) + ($369)|0);
     $375 = HEAP8[$374>>0]|0;
     $sB$2$shrunk = $375;$sG$2$shrunk = $364;$sR$2$shrunk = $353;
     break;
    }
    case 2:  {
     $376 = HEAP16[$src$3>>1]|0;
     $377 = $376&65535;
     $378 = HEAP32[$29>>2]|0;
     $379 = $377 & $378;
     $380 = HEAP32[$30>>2]|0;
     $381 = $380 & 255;
     $382 = $379 >>> $381;
     $383 = HEAP32[$31>>2]|0;
     $384 = $383 & 255;
     $385 = (125924 + ($384<<2)|0);
     $386 = HEAP32[$385>>2]|0;
     $387 = (($386) + ($382)|0);
     $388 = HEAP8[$387>>0]|0;
     $389 = HEAP32[$32>>2]|0;
     $390 = $389 & $377;
     $391 = $380 >>> 8;
     $392 = $391 & 255;
     $393 = $390 >>> $392;
     $394 = $383 >>> 8;
     $395 = $394 & 255;
     $396 = (125924 + ($395<<2)|0);
     $397 = HEAP32[$396>>2]|0;
     $398 = (($397) + ($393)|0);
     $399 = HEAP8[$398>>0]|0;
     $400 = HEAP32[$33>>2]|0;
     $401 = $400 & $377;
     $402 = $380 >>> 16;
     $403 = $402 & 255;
     $404 = $401 >>> $403;
     $405 = $383 >>> 16;
     $406 = $405 & 255;
     $407 = (125924 + ($406<<2)|0);
     $408 = HEAP32[$407>>2]|0;
     $409 = (($408) + ($404)|0);
     $410 = HEAP8[$409>>0]|0;
     $sB$2$shrunk = $410;$sG$2$shrunk = $399;$sR$2$shrunk = $388;
     break;
    }
    case 3:  {
     $411 = HEAP32[$30>>2]|0;
     $div = $411 >>> 3;
     $412 = $div & 31;
     $413 = (($src$3) + ($412)|0);
     $414 = HEAP8[$413>>0]|0;
     $415 = $411 >>> 11;
     $div4 = $415 & 31;
     $416 = (($src$3) + ($div4)|0);
     $417 = HEAP8[$416>>0]|0;
     $418 = $411 >>> 19;
     $419 = $418 & 31;
     $420 = (($src$3) + ($419)|0);
     $421 = HEAP8[$420>>0]|0;
     $sB$2$shrunk = $421;$sG$2$shrunk = $417;$sR$2$shrunk = $414;
     break;
    }
    case 4:  {
     $422 = HEAP32[$src$3>>2]|0;
     $423 = HEAP32[$29>>2]|0;
     $424 = $423 & $422;
     $425 = HEAP32[$30>>2]|0;
     $426 = $425 & 255;
     $427 = $424 >>> $426;
     $428 = HEAP32[$31>>2]|0;
     $429 = $428 & 255;
     $430 = (125924 + ($429<<2)|0);
     $431 = HEAP32[$430>>2]|0;
     $432 = (($431) + ($427)|0);
     $433 = HEAP8[$432>>0]|0;
     $434 = HEAP32[$32>>2]|0;
     $435 = $434 & $422;
     $436 = $425 >>> 8;
     $437 = $436 & 255;
     $438 = $435 >>> $437;
     $439 = $428 >>> 8;
     $440 = $439 & 255;
     $441 = (125924 + ($440<<2)|0);
     $442 = HEAP32[$441>>2]|0;
     $443 = (($442) + ($438)|0);
     $444 = HEAP8[$443>>0]|0;
     $445 = HEAP32[$33>>2]|0;
     $446 = $445 & $422;
     $447 = $425 >>> 16;
     $448 = $447 & 255;
     $449 = $446 >>> $448;
     $450 = $428 >>> 16;
     $451 = $450 & 255;
     $452 = (125924 + ($451<<2)|0);
     $453 = HEAP32[$452>>2]|0;
     $454 = (($453) + ($449)|0);
     $455 = HEAP8[$454>>0]|0;
     $sB$2$shrunk = $455;$sG$2$shrunk = $444;$sR$2$shrunk = $433;
     break;
    }
    default: {
     $sB$2$shrunk = 0;$sG$2$shrunk = 0;$sR$2$shrunk = 0;
    }
    }
    $sB$2 = $sB$2$shrunk&255;
    $sG$2 = $sG$2$shrunk&255;
    $sR$2 = $sR$2$shrunk&255;
    $456 = HEAP8[$dst$3>>0]|0;
    $457 = $456&255;
    $458 = HEAP32[$27>>2]|0;
    $459 = ((($458)) + 4|0);
    $460 = HEAP32[$459>>2]|0;
    $461 = (($460) + ($457<<2)|0);
    $462 = HEAP8[$461>>0]|0;
    $463 = $462&255;
    $464 = (((($460) + ($457<<2)|0)) + 1|0);
    $465 = HEAP8[$464>>0]|0;
    $466 = $465&255;
    $467 = (((($460) + ($457<<2)|0)) + 2|0);
    $468 = HEAP8[$467>>0]|0;
    $469 = $468&255;
    $470 = (($sR$2) - ($463))|0;
    $471 = Math_imul($470, $17)|0;
    $472 = (($471>>>0) / 255)&-1;
    $473 = (($472) + ($463))|0;
    $474 = (($sG$2) - ($466))|0;
    $475 = Math_imul($474, $17)|0;
    $476 = (($475>>>0) / 255)&-1;
    $477 = (($476) + ($466))|0;
    $478 = (($sB$2) - ($469))|0;
    $479 = Math_imul($478, $17)|0;
    $480 = (($479>>>0) / 255)&-1;
    $481 = (($480) + ($469))|0;
    $482 = $473 & 224;
    $483 = $477 >>> 3;
    $484 = $483 & 28;
    $485 = $484 | $482;
    $486 = $481 >>> 6;
    $487 = $486 & 3;
    $488 = $485 | $487;
    if ($28) {
     $489 = $488&255;
     $storemerge3 = $489;
    } else {
     $490 = (($9) + ($488)|0);
     $491 = HEAP8[$490>>0]|0;
     $storemerge3 = $491;
    }
    HEAP8[$dst$3>>0] = $storemerge3;
    $492 = ((($dst$3)) + 1|0);
    $493 = (($src$3) + ($14)|0);
    $dst$4 = $492;$n$3 = $n$2;$src$4 = $493;
    label = 31;
    continue;
   }
   else if ((label|0) == 31) {
    label = 0;
    switch ($14|0) {
    case 1:  {
     $494 = HEAP8[$src$4>>0]|0;
     $495 = $494&255;
     $496 = HEAP32[$29>>2]|0;
     $497 = $495 & $496;
     $498 = HEAP32[$30>>2]|0;
     $499 = $498 & 255;
     $500 = $497 >>> $499;
     $501 = HEAP32[$31>>2]|0;
     $502 = $501 & 255;
     $503 = (125924 + ($502<<2)|0);
     $504 = HEAP32[$503>>2]|0;
     $505 = (($504) + ($500)|0);
     $506 = HEAP8[$505>>0]|0;
     $507 = HEAP32[$32>>2]|0;
     $508 = $507 & $495;
     $509 = $498 >>> 8;
     $510 = $509 & 255;
     $511 = $508 >>> $510;
     $512 = $501 >>> 8;
     $513 = $512 & 255;
     $514 = (125924 + ($513<<2)|0);
     $515 = HEAP32[$514>>2]|0;
     $516 = (($515) + ($511)|0);
     $517 = HEAP8[$516>>0]|0;
     $518 = HEAP32[$33>>2]|0;
     $519 = $518 & $495;
     $520 = $498 >>> 16;
     $521 = $520 & 255;
     $522 = $519 >>> $521;
     $523 = $501 >>> 16;
     $524 = $523 & 255;
     $525 = (125924 + ($524<<2)|0);
     $526 = HEAP32[$525>>2]|0;
     $527 = (($526) + ($522)|0);
     $528 = HEAP8[$527>>0]|0;
     $sB$3$shrunk = $528;$sG$3$shrunk = $517;$sR$3$shrunk = $506;
     break;
    }
    case 2:  {
     $529 = HEAP16[$src$4>>1]|0;
     $530 = $529&65535;
     $531 = HEAP32[$29>>2]|0;
     $532 = $530 & $531;
     $533 = HEAP32[$30>>2]|0;
     $534 = $533 & 255;
     $535 = $532 >>> $534;
     $536 = HEAP32[$31>>2]|0;
     $537 = $536 & 255;
     $538 = (125924 + ($537<<2)|0);
     $539 = HEAP32[$538>>2]|0;
     $540 = (($539) + ($535)|0);
     $541 = HEAP8[$540>>0]|0;
     $542 = HEAP32[$32>>2]|0;
     $543 = $542 & $530;
     $544 = $533 >>> 8;
     $545 = $544 & 255;
     $546 = $543 >>> $545;
     $547 = $536 >>> 8;
     $548 = $547 & 255;
     $549 = (125924 + ($548<<2)|0);
     $550 = HEAP32[$549>>2]|0;
     $551 = (($550) + ($546)|0);
     $552 = HEAP8[$551>>0]|0;
     $553 = HEAP32[$33>>2]|0;
     $554 = $553 & $530;
     $555 = $533 >>> 16;
     $556 = $555 & 255;
     $557 = $554 >>> $556;
     $558 = $536 >>> 16;
     $559 = $558 & 255;
     $560 = (125924 + ($559<<2)|0);
     $561 = HEAP32[$560>>2]|0;
     $562 = (($561) + ($557)|0);
     $563 = HEAP8[$562>>0]|0;
     $sB$3$shrunk = $563;$sG$3$shrunk = $552;$sR$3$shrunk = $541;
     break;
    }
    case 3:  {
     $564 = HEAP32[$30>>2]|0;
     $div12 = $564 >>> 3;
     $565 = $div12 & 31;
     $566 = (($src$4) + ($565)|0);
     $567 = HEAP8[$566>>0]|0;
     $568 = $564 >>> 11;
     $div13 = $568 & 31;
     $569 = (($src$4) + ($div13)|0);
     $570 = HEAP8[$569>>0]|0;
     $571 = $564 >>> 19;
     $572 = $571 & 31;
     $573 = (($src$4) + ($572)|0);
     $574 = HEAP8[$573>>0]|0;
     $sB$3$shrunk = $574;$sG$3$shrunk = $570;$sR$3$shrunk = $567;
     break;
    }
    case 4:  {
     $575 = HEAP32[$src$4>>2]|0;
     $576 = HEAP32[$29>>2]|0;
     $577 = $576 & $575;
     $578 = HEAP32[$30>>2]|0;
     $579 = $578 & 255;
     $580 = $577 >>> $579;
     $581 = HEAP32[$31>>2]|0;
     $582 = $581 & 255;
     $583 = (125924 + ($582<<2)|0);
     $584 = HEAP32[$583>>2]|0;
     $585 = (($584) + ($580)|0);
     $586 = HEAP8[$585>>0]|0;
     $587 = HEAP32[$32>>2]|0;
     $588 = $587 & $575;
     $589 = $578 >>> 8;
     $590 = $589 & 255;
     $591 = $588 >>> $590;
     $592 = $581 >>> 8;
     $593 = $592 & 255;
     $594 = (125924 + ($593<<2)|0);
     $595 = HEAP32[$594>>2]|0;
     $596 = (($595) + ($591)|0);
     $597 = HEAP8[$596>>0]|0;
     $598 = HEAP32[$33>>2]|0;
     $599 = $598 & $575;
     $600 = $578 >>> 16;
     $601 = $600 & 255;
     $602 = $599 >>> $601;
     $603 = $581 >>> 16;
     $604 = $603 & 255;
     $605 = (125924 + ($604<<2)|0);
     $606 = HEAP32[$605>>2]|0;
     $607 = (($606) + ($602)|0);
     $608 = HEAP8[$607>>0]|0;
     $sB$3$shrunk = $608;$sG$3$shrunk = $597;$sR$3$shrunk = $586;
     break;
    }
    default: {
     $sB$3$shrunk = 0;$sG$3$shrunk = 0;$sR$3$shrunk = 0;
    }
    }
    $sB$3 = $sB$3$shrunk&255;
    $sG$3 = $sG$3$shrunk&255;
    $sR$3 = $sR$3$shrunk&255;
    $609 = HEAP8[$dst$4>>0]|0;
    $610 = $609&255;
    $611 = HEAP32[$27>>2]|0;
    $612 = ((($611)) + 4|0);
    $613 = HEAP32[$612>>2]|0;
    $614 = (($613) + ($610<<2)|0);
    $615 = HEAP8[$614>>0]|0;
    $616 = $615&255;
    $617 = (((($613) + ($610<<2)|0)) + 1|0);
    $618 = HEAP8[$617>>0]|0;
    $619 = $618&255;
    $620 = (((($613) + ($610<<2)|0)) + 2|0);
    $621 = HEAP8[$620>>0]|0;
    $622 = $621&255;
    $623 = (($sR$3) - ($616))|0;
    $624 = Math_imul($623, $17)|0;
    $625 = (($624>>>0) / 255)&-1;
    $626 = (($625) + ($616))|0;
    $627 = (($sG$3) - ($619))|0;
    $628 = Math_imul($627, $17)|0;
    $629 = (($628>>>0) / 255)&-1;
    $630 = (($629) + ($619))|0;
    $631 = (($sB$3) - ($622))|0;
    $632 = Math_imul($631, $17)|0;
    $633 = (($632>>>0) / 255)&-1;
    $634 = (($633) + ($622))|0;
    $635 = $626 & 224;
    $636 = $630 >>> 3;
    $637 = $636 & 28;
    $638 = $637 | $635;
    $639 = $634 >>> 6;
    $640 = $639 & 3;
    $641 = $638 | $640;
    if ($28) {
     $642 = $641&255;
     $storemerge = $642;
    } else {
     $643 = (($9) + ($641)|0);
     $644 = HEAP8[$643>>0]|0;
     $storemerge = $644;
    }
    HEAP8[$dst$4>>0] = $storemerge;
    $645 = ((($dst$4)) + 1|0);
    $646 = (($src$4) + ($14)|0);
    $647 = (($n$3) + -1)|0;
    $648 = ($n$3|0)>(1);
    if ($648) {
     $dst$1 = $645;$n$0 = $647;$src$1 = $646;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $649 = (($646) + ($5)|0);
  $650 = (($645) + ($7)|0);
  $651 = ($34|0)==(0);
  if ($651) {
   label = 42;
   break;
  } else {
   $$in = $34;$dst$015 = $650;$src$016 = $649;
  }
 }
 if ((label|0) == 40) {
  // unreachable;
 }
 else if ((label|0) == 42) {
  return;
 }
}
function _BlitRGBtoRGBSurfaceAlpha($info) {
 $info = $info|0;
 var $$lcssa = 0, $$lcssa33 = 0, $$lcssa34 = 0, $$lcssa51 = 0, $$lcssa54 = 0, $$lcssa55 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0;
 var $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0;
 var $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0;
 var $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0;
 var $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0;
 var $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0;
 var $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0;
 var $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0;
 var $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dstp$01$i = 0, $dstp$01$i$us = 0, $dstp$01$us$i = 0, $dstp$05 = 0, $dstp$05$us = 0, $dstp$1$us = 0, $dstp$1$us$i = 0, $dstp$2 = 0, $dstp$2$i = 0, $dstp$3 = 0, $dstp$3$i = 0, $dstp$4 = 0, $dstp$4$i = 0, $dstp$4$i$us38 = 0;
 var $n$0$us = 0, $n$0$us$i = 0, $n$1 = 0, $n$1$i = 0, $n$2 = 0, $n$2$i = 0, $n$3 = 0, $n$3$i = 0, $n$3$i$us37 = 0, $srcp$02$i = 0, $srcp$02$i$us = 0, $srcp$02$us$i = 0, $srcp$04 = 0, $srcp$04$us = 0, $srcp$1$us = 0, $srcp$1$us$i = 0, $srcp$2 = 0, $srcp$2$i = 0, $srcp$3 = 0, $srcp$3$i = 0;
 var $srcp$4 = 0, $srcp$4$i = 0, $srcp$4$i$us39 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 63|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = $1&255;
 $3 = ($1<<24>>24)==(-128);
 $4 = ((($info)) + 24|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 28|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = HEAP32[$info>>2]|0;
 $9 = ((($info)) + 16|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 >> 2;
 $12 = ((($info)) + 20|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($info)) + 36|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = $15 >> 2;
 $17 = (($7) + -1)|0;
 $18 = ($7|0)==(0);
 if (!($3)) {
  if ($18) {
   return;
  }
  $19 = (($5) + 3)|0;
  $20 = (($19|0) / 4)&-1;
  $21 = $5 & 3;
  $22 = ($21|0)==(0);
  if ($22) {
   $108 = $17;$dstp$05$us = $13;$srcp$04$us = $8;
   while(1) {
    $dstp$1$us = $dstp$05$us;$n$0$us = $20;$srcp$1$us = $srcp$04$us;
    while(1) {
     $23 = HEAP32[$srcp$1$us>>2]|0;
     $24 = HEAP32[$dstp$1$us>>2]|0;
     $25 = $23 & 16711935;
     $26 = $24 & 16711935;
     $27 = (($25) - ($26))|0;
     $28 = Math_imul($27, $2)|0;
     $29 = $28 >>> 8;
     $30 = (($29) + ($26))|0;
     $31 = $30 & 16711935;
     $32 = $23 & 65280;
     $33 = $24 & 65280;
     $34 = (($32) - ($33))|0;
     $35 = Math_imul($34, $2)|0;
     $36 = $35 >>> 8;
     $37 = (($36) + ($33))|0;
     $38 = $37 & 65280;
     $39 = $38 | $31;
     $40 = $39 | -16777216;
     HEAP32[$dstp$1$us>>2] = $40;
     $41 = ((($srcp$1$us)) + 4|0);
     $42 = ((($dstp$1$us)) + 4|0);
     $43 = HEAP32[$41>>2]|0;
     $44 = HEAP32[$42>>2]|0;
     $45 = $43 & 16711935;
     $46 = $44 & 16711935;
     $47 = (($45) - ($46))|0;
     $48 = Math_imul($47, $2)|0;
     $49 = $48 >>> 8;
     $50 = (($49) + ($46))|0;
     $51 = $50 & 16711935;
     $52 = $43 & 65280;
     $53 = $44 & 65280;
     $54 = (($52) - ($53))|0;
     $55 = Math_imul($54, $2)|0;
     $56 = $55 >>> 8;
     $57 = (($56) + ($53))|0;
     $58 = $57 & 65280;
     $59 = $58 | $51;
     $60 = $59 | -16777216;
     HEAP32[$42>>2] = $60;
     $61 = ((($srcp$1$us)) + 8|0);
     $62 = ((($dstp$1$us)) + 8|0);
     $63 = HEAP32[$61>>2]|0;
     $64 = HEAP32[$62>>2]|0;
     $65 = $63 & 16711935;
     $66 = $64 & 16711935;
     $67 = (($65) - ($66))|0;
     $68 = Math_imul($67, $2)|0;
     $69 = $68 >>> 8;
     $70 = (($69) + ($66))|0;
     $71 = $70 & 16711935;
     $72 = $63 & 65280;
     $73 = $64 & 65280;
     $74 = (($72) - ($73))|0;
     $75 = Math_imul($74, $2)|0;
     $76 = $75 >>> 8;
     $77 = (($76) + ($73))|0;
     $78 = $77 & 65280;
     $79 = $78 | $71;
     $80 = $79 | -16777216;
     HEAP32[$62>>2] = $80;
     $81 = ((($srcp$1$us)) + 12|0);
     $82 = ((($dstp$1$us)) + 12|0);
     $83 = HEAP32[$81>>2]|0;
     $84 = HEAP32[$82>>2]|0;
     $85 = $83 & 16711935;
     $86 = $84 & 16711935;
     $87 = (($85) - ($86))|0;
     $88 = Math_imul($87, $2)|0;
     $89 = $88 >>> 8;
     $90 = (($89) + ($86))|0;
     $91 = $90 & 16711935;
     $92 = $83 & 65280;
     $93 = $84 & 65280;
     $94 = (($92) - ($93))|0;
     $95 = Math_imul($94, $2)|0;
     $96 = $95 >>> 8;
     $97 = (($96) + ($93))|0;
     $98 = $97 & 65280;
     $99 = $98 | $91;
     $100 = $99 | -16777216;
     HEAP32[$82>>2] = $100;
     $101 = ((($srcp$1$us)) + 16|0);
     $102 = ((($dstp$1$us)) + 16|0);
     $103 = (($n$0$us) + -1)|0;
     $104 = ($n$0$us|0)>(1);
     if ($104) {
      $dstp$1$us = $102;$n$0$us = $103;$srcp$1$us = $101;
     } else {
      $$lcssa54 = $101;$$lcssa55 = $102;
      break;
     }
    }
    $105 = (($$lcssa54) + ($11<<2)|0);
    $106 = (($$lcssa55) + ($16<<2)|0);
    $107 = (($108) + -1)|0;
    $109 = ($108|0)==(0);
    if ($109) {
     break;
    } else {
     $108 = $107;$dstp$05$us = $106;$srcp$04$us = $105;
    }
   }
   return;
  } else {
   $378 = $17;$dstp$05 = $13;$srcp$04 = $8;
  }
  L13: while(1) {
   switch ($21|0) {
   case 1:  {
    $dstp$4 = $dstp$05;$n$3 = $20;$srcp$4 = $srcp$04;
    label = 27;
    break;
   }
   case 3:  {
    $dstp$2 = $dstp$05;$n$1 = $20;$srcp$2 = $srcp$04;
    label = 25;
    break;
   }
   case 2:  {
    $dstp$3 = $dstp$05;$n$2 = $20;$srcp$3 = $srcp$04;
    label = 26;
    break;
   }
   default: {
    label = 28;
    break L13;
   }
   }
   while(1) {
    if ((label|0) == 25) {
     label = 0;
     $316 = HEAP32[$srcp$2>>2]|0;
     $317 = HEAP32[$dstp$2>>2]|0;
     $318 = $316 & 16711935;
     $319 = $317 & 16711935;
     $320 = (($318) - ($319))|0;
     $321 = Math_imul($320, $2)|0;
     $322 = $321 >>> 8;
     $323 = (($322) + ($319))|0;
     $324 = $323 & 16711935;
     $325 = $316 & 65280;
     $326 = $317 & 65280;
     $327 = (($325) - ($326))|0;
     $328 = Math_imul($327, $2)|0;
     $329 = $328 >>> 8;
     $330 = (($329) + ($326))|0;
     $331 = $330 & 65280;
     $332 = $331 | $324;
     $333 = $332 | -16777216;
     HEAP32[$dstp$2>>2] = $333;
     $334 = ((($srcp$2)) + 4|0);
     $335 = ((($dstp$2)) + 4|0);
     $dstp$3 = $335;$n$2 = $n$1;$srcp$3 = $334;
     label = 26;
     continue;
    }
    else if ((label|0) == 26) {
     label = 0;
     $336 = HEAP32[$srcp$3>>2]|0;
     $337 = HEAP32[$dstp$3>>2]|0;
     $338 = $336 & 16711935;
     $339 = $337 & 16711935;
     $340 = (($338) - ($339))|0;
     $341 = Math_imul($340, $2)|0;
     $342 = $341 >>> 8;
     $343 = (($342) + ($339))|0;
     $344 = $343 & 16711935;
     $345 = $336 & 65280;
     $346 = $337 & 65280;
     $347 = (($345) - ($346))|0;
     $348 = Math_imul($347, $2)|0;
     $349 = $348 >>> 8;
     $350 = (($349) + ($346))|0;
     $351 = $350 & 65280;
     $352 = $351 | $344;
     $353 = $352 | -16777216;
     HEAP32[$dstp$3>>2] = $353;
     $354 = ((($srcp$3)) + 4|0);
     $355 = ((($dstp$3)) + 4|0);
     $dstp$4 = $355;$n$3 = $n$2;$srcp$4 = $354;
     label = 27;
     continue;
    }
    else if ((label|0) == 27) {
     label = 0;
     $356 = HEAP32[$srcp$4>>2]|0;
     $357 = HEAP32[$dstp$4>>2]|0;
     $358 = $356 & 16711935;
     $359 = $357 & 16711935;
     $360 = (($358) - ($359))|0;
     $361 = Math_imul($360, $2)|0;
     $362 = $361 >>> 8;
     $363 = (($362) + ($359))|0;
     $364 = $363 & 16711935;
     $365 = $356 & 65280;
     $366 = $357 & 65280;
     $367 = (($365) - ($366))|0;
     $368 = Math_imul($367, $2)|0;
     $369 = $368 >>> 8;
     $370 = (($369) + ($366))|0;
     $371 = $370 & 65280;
     $372 = $371 | $364;
     $373 = $372 | -16777216;
     HEAP32[$dstp$4>>2] = $373;
     $295 = ((($srcp$4)) + 4|0);
     $297 = ((($dstp$4)) + 4|0);
     $374 = ($n$3|0)>(1);
     if (!($374)) {
      break;
     }
     $293 = (($n$3) + -1)|0;
     $294 = HEAP32[$295>>2]|0;
     $296 = HEAP32[$297>>2]|0;
     $298 = $294 & 16711935;
     $299 = $296 & 16711935;
     $300 = (($298) - ($299))|0;
     $301 = Math_imul($300, $2)|0;
     $302 = $301 >>> 8;
     $303 = (($302) + ($299))|0;
     $304 = $303 & 16711935;
     $305 = $294 & 65280;
     $306 = $296 & 65280;
     $307 = (($305) - ($306))|0;
     $308 = Math_imul($307, $2)|0;
     $309 = $308 >>> 8;
     $310 = (($309) + ($306))|0;
     $311 = $310 & 65280;
     $312 = $311 | $304;
     $313 = $312 | -16777216;
     HEAP32[$297>>2] = $313;
     $314 = ((($srcp$4)) + 8|0);
     $315 = ((($dstp$4)) + 8|0);
     $dstp$2 = $315;$n$1 = $293;$srcp$2 = $314;
     label = 25;
     continue;
    }
   }
   $375 = (($295) + ($11<<2)|0);
   $376 = (($297) + ($16<<2)|0);
   $377 = (($378) + -1)|0;
   $379 = ($378|0)==(0);
   if ($379) {
    label = 30;
    break;
   } else {
    $378 = $377;$dstp$05 = $376;$srcp$04 = $375;
   }
  }
  if ((label|0) == 28) {
   // unreachable;
  }
  else if ((label|0) == 30) {
   return;
  }
 }
 if ($18) {
  return;
 }
 $110 = (($5) + 3)|0;
 $111 = (($110|0) / 4)&-1;
 $112 = $5 & 3;
 switch ($112|0) {
 case 0:  {
  $236 = $17;$dstp$01$us$i = $13;$srcp$02$us$i = $8;
  while(1) {
   $dstp$1$us$i = $dstp$01$us$i;$n$0$us$i = $111;$srcp$1$us$i = $srcp$02$us$i;
   while(1) {
    $183 = ((($srcp$1$us$i)) + 4|0);
    $184 = HEAP32[$srcp$1$us$i>>2]|0;
    $185 = HEAP32[$dstp$1$us$i>>2]|0;
    $186 = $184 & 16711422;
    $187 = $185 & 16711422;
    $188 = (($187) + ($186))|0;
    $189 = $188 >>> 1;
    $190 = $184 & 65793;
    $191 = $190 & $185;
    $192 = (($189) + ($191))|0;
    $193 = $192 | -16777216;
    $194 = ((($dstp$1$us$i)) + 4|0);
    HEAP32[$dstp$1$us$i>>2] = $193;
    $195 = ((($srcp$1$us$i)) + 8|0);
    $196 = HEAP32[$183>>2]|0;
    $197 = HEAP32[$194>>2]|0;
    $198 = $196 & 16711422;
    $199 = $197 & 16711422;
    $200 = (($199) + ($198))|0;
    $201 = $200 >>> 1;
    $202 = $196 & 65793;
    $203 = $202 & $197;
    $204 = (($201) + ($203))|0;
    $205 = $204 | -16777216;
    $206 = ((($dstp$1$us$i)) + 8|0);
    HEAP32[$194>>2] = $205;
    $207 = ((($srcp$1$us$i)) + 12|0);
    $208 = HEAP32[$195>>2]|0;
    $209 = HEAP32[$206>>2]|0;
    $210 = $208 & 16711422;
    $211 = $209 & 16711422;
    $212 = (($211) + ($210))|0;
    $213 = $212 >>> 1;
    $214 = $208 & 65793;
    $215 = $214 & $209;
    $216 = (($213) + ($215))|0;
    $217 = $216 | -16777216;
    $218 = ((($dstp$1$us$i)) + 12|0);
    HEAP32[$206>>2] = $217;
    $219 = ((($srcp$1$us$i)) + 16|0);
    $220 = HEAP32[$207>>2]|0;
    $221 = HEAP32[$218>>2]|0;
    $222 = $220 & 16711422;
    $223 = $221 & 16711422;
    $224 = (($223) + ($222))|0;
    $225 = $224 >>> 1;
    $226 = $220 & 65793;
    $227 = $226 & $221;
    $228 = (($225) + ($227))|0;
    $229 = $228 | -16777216;
    $230 = ((($dstp$1$us$i)) + 16|0);
    HEAP32[$218>>2] = $229;
    $231 = (($n$0$us$i) + -1)|0;
    $232 = ($n$0$us$i|0)>(1);
    if ($232) {
     $dstp$1$us$i = $230;$n$0$us$i = $231;$srcp$1$us$i = $219;
    } else {
     $$lcssa = $219;$$lcssa51 = $230;
     break;
    }
   }
   $233 = (($$lcssa) + ($11<<2)|0);
   $234 = (($$lcssa51) + ($16<<2)|0);
   $235 = (($236) + -1)|0;
   $237 = ($236|0)==(0);
   if ($237) {
    break;
   } else {
    $236 = $235;$dstp$01$us$i = $234;$srcp$02$us$i = $233;
   }
  }
  return;
  break;
 }
 case 1:  {
  $113 = ($110|0)>(7);
  $129 = $17;$dstp$01$i$us = $13;$srcp$02$i$us = $8;
  while(1) {
   $114 = ((($srcp$02$i$us)) + 4|0);
   $115 = HEAP32[$srcp$02$i$us>>2]|0;
   $116 = HEAP32[$dstp$01$i$us>>2]|0;
   $117 = $115 & 16711422;
   $118 = $116 & 16711422;
   $119 = (($118) + ($117))|0;
   $120 = $119 >>> 1;
   $121 = $115 & 65793;
   $122 = $121 & $116;
   $123 = (($120) + ($122))|0;
   $124 = $123 | -16777216;
   $125 = ((($dstp$01$i$us)) + 4|0);
   HEAP32[$dstp$01$i$us>>2] = $124;
   if ($113) {
    $134 = $114;$136 = $125;$dstp$4$i$us38 = $dstp$01$i$us;$n$3$i$us37 = $111;$srcp$4$i$us39 = $srcp$02$i$us;
    while(1) {
     $131 = (($n$3$i$us37) + -1)|0;
     $132 = ((($srcp$4$i$us39)) + 8|0);
     $133 = HEAP32[$134>>2]|0;
     $135 = HEAP32[$136>>2]|0;
     $137 = $133 & 16711422;
     $138 = $135 & 16711422;
     $139 = (($138) + ($137))|0;
     $140 = $139 >>> 1;
     $141 = $133 & 65793;
     $142 = $141 & $135;
     $143 = (($140) + ($142))|0;
     $144 = $143 | -16777216;
     $145 = ((($dstp$4$i$us38)) + 8|0);
     HEAP32[$136>>2] = $144;
     $146 = ((($srcp$4$i$us39)) + 12|0);
     $147 = HEAP32[$132>>2]|0;
     $148 = HEAP32[$145>>2]|0;
     $149 = $147 & 16711422;
     $150 = $148 & 16711422;
     $151 = (($150) + ($149))|0;
     $152 = $151 >>> 1;
     $153 = $147 & 65793;
     $154 = $153 & $148;
     $155 = (($152) + ($154))|0;
     $156 = $155 | -16777216;
     $157 = ((($dstp$4$i$us38)) + 12|0);
     HEAP32[$145>>2] = $156;
     $158 = ((($srcp$4$i$us39)) + 16|0);
     $159 = HEAP32[$146>>2]|0;
     $160 = HEAP32[$157>>2]|0;
     $161 = $159 & 16711422;
     $162 = $160 & 16711422;
     $163 = (($162) + ($161))|0;
     $164 = $163 >>> 1;
     $165 = $159 & 65793;
     $166 = $165 & $160;
     $167 = (($164) + ($166))|0;
     $168 = $167 | -16777216;
     $169 = ((($dstp$4$i$us38)) + 16|0);
     HEAP32[$157>>2] = $168;
     $170 = ((($srcp$4$i$us39)) + 20|0);
     $171 = HEAP32[$158>>2]|0;
     $172 = HEAP32[$169>>2]|0;
     $173 = $171 & 16711422;
     $174 = $172 & 16711422;
     $175 = (($174) + ($173))|0;
     $176 = $175 >>> 1;
     $177 = $171 & 65793;
     $178 = $177 & $172;
     $179 = (($176) + ($178))|0;
     $180 = $179 | -16777216;
     $181 = ((($dstp$4$i$us38)) + 20|0);
     HEAP32[$169>>2] = $180;
     $182 = ($131|0)>(1);
     if ($182) {
      $134 = $170;$136 = $181;$dstp$4$i$us38 = $169;$n$3$i$us37 = $131;$srcp$4$i$us39 = $158;
     } else {
      $$lcssa33 = $170;$$lcssa34 = $181;
      break;
     }
    }
   } else {
    $$lcssa33 = $114;$$lcssa34 = $125;
   }
   $126 = (($$lcssa33) + ($11<<2)|0);
   $127 = (($$lcssa34) + ($16<<2)|0);
   $128 = (($129) + -1)|0;
   $130 = ($129|0)==(0);
   if ($130) {
    break;
   } else {
    $129 = $128;$dstp$01$i$us = $127;$srcp$02$i$us = $126;
   }
  }
  return;
  break;
 }
 default: {
  $291 = $17;$dstp$01$i = $13;$srcp$02$i = $8;
  L30: while(1) {
   switch ($112|0) {
   case 1:  {
    $dstp$4$i = $dstp$01$i;$n$3$i = $111;$srcp$4$i = $srcp$02$i;
    label = 20;
    break;
   }
   case 3:  {
    $dstp$2$i = $dstp$01$i;$n$1$i = $111;$srcp$2$i = $srcp$02$i;
    label = 18;
    break;
   }
   case 2:  {
    $dstp$3$i = $dstp$01$i;$n$2$i = $111;$srcp$3$i = $srcp$02$i;
    label = 19;
    break;
   }
   default: {
    label = 21;
    break L30;
   }
   }
   while(1) {
    if ((label|0) == 18) {
     label = 0;
     $253 = ((($srcp$2$i)) + 4|0);
     $254 = HEAP32[$srcp$2$i>>2]|0;
     $255 = HEAP32[$dstp$2$i>>2]|0;
     $256 = $254 & 16711422;
     $257 = $255 & 16711422;
     $258 = (($257) + ($256))|0;
     $259 = $258 >>> 1;
     $260 = $254 & 65793;
     $261 = $260 & $255;
     $262 = (($259) + ($261))|0;
     $263 = $262 | -16777216;
     $264 = ((($dstp$2$i)) + 4|0);
     HEAP32[$dstp$2$i>>2] = $263;
     $dstp$3$i = $264;$n$2$i = $n$1$i;$srcp$3$i = $253;
     label = 19;
     continue;
    }
    else if ((label|0) == 19) {
     label = 0;
     $265 = ((($srcp$3$i)) + 4|0);
     $266 = HEAP32[$srcp$3$i>>2]|0;
     $267 = HEAP32[$dstp$3$i>>2]|0;
     $268 = $266 & 16711422;
     $269 = $267 & 16711422;
     $270 = (($269) + ($268))|0;
     $271 = $270 >>> 1;
     $272 = $266 & 65793;
     $273 = $272 & $267;
     $274 = (($271) + ($273))|0;
     $275 = $274 | -16777216;
     $276 = ((($dstp$3$i)) + 4|0);
     HEAP32[$dstp$3$i>>2] = $275;
     $dstp$4$i = $276;$n$3$i = $n$2$i;$srcp$4$i = $265;
     label = 20;
     continue;
    }
    else if ((label|0) == 20) {
     label = 0;
     $241 = ((($srcp$4$i)) + 4|0);
     $277 = HEAP32[$srcp$4$i>>2]|0;
     $278 = HEAP32[$dstp$4$i>>2]|0;
     $279 = $277 & 16711422;
     $280 = $278 & 16711422;
     $281 = (($280) + ($279))|0;
     $282 = $281 >>> 1;
     $283 = $277 & 65793;
     $284 = $283 & $278;
     $285 = (($282) + ($284))|0;
     $286 = $285 | -16777216;
     $243 = ((($dstp$4$i)) + 4|0);
     HEAP32[$dstp$4$i>>2] = $286;
     $287 = ($n$3$i|0)>(1);
     if (!($287)) {
      break;
     }
     $238 = (($n$3$i) + -1)|0;
     $239 = ((($srcp$4$i)) + 8|0);
     $240 = HEAP32[$241>>2]|0;
     $242 = HEAP32[$243>>2]|0;
     $244 = $240 & 16711422;
     $245 = $242 & 16711422;
     $246 = (($245) + ($244))|0;
     $247 = $246 >>> 1;
     $248 = $240 & 65793;
     $249 = $248 & $242;
     $250 = (($247) + ($249))|0;
     $251 = $250 | -16777216;
     $252 = ((($dstp$4$i)) + 8|0);
     HEAP32[$243>>2] = $251;
     $dstp$2$i = $252;$n$1$i = $238;$srcp$2$i = $239;
     label = 18;
     continue;
    }
   }
   $288 = (($241) + ($11<<2)|0);
   $289 = (($243) + ($16<<2)|0);
   $290 = (($291) + -1)|0;
   $292 = ($291|0)==(0);
   if ($292) {
    label = 30;
    break;
   } else {
    $291 = $290;$dstp$01$i = $289;$srcp$02$i = $288;
   }
  }
  if ((label|0) == 21) {
   // unreachable;
  }
  else if ((label|0) == 30) {
   return;
  }
 }
 }
}
function _SDL_CalculateBlitN($surface) {
 $surface = $surface|0;
 var $$0 = 0, $$mask = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $Blit4to4CopyAlpha$BlitNtoNCopyAlpha = 0, $BlitNtoNCopyAlpha$ = 0, $a_need$0 = 0, $or$cond = 0, $or$cond1 = 0, $or$cond2 = 0, $or$cond3 = 0;
 var $or$cond4 = 0, $or$cond5 = 0, $switch$split2D = 0, $which$0$lcssa = 0, $which$06 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($surface)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($surface)) + 52|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($6)) + 8|0);
 $8 = HEAP16[$7>>1]|0;
 $9 = $8&255;
 $10 = ($9&255)<(8);
 if ($10) {
  $$0 = 0;
  return ($$0|0);
 }
 $11 = ((($3)) + 68|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = $12 & -28673;
 switch ($13|0) {
 case 0:  {
  $14 = ($9<<24>>24)==(8);
  if ($14) {
   $15 = ((($1)) + 9|0);
   $16 = HEAP8[$15>>0]|0;
   $17 = ($16<<24>>24)==(4);
   L8: do {
    if ($17) {
     $18 = ((($1)) + 12|0);
     $19 = HEAP32[$18>>2]|0;
     $switch$split2D = ($19|0)<(1072693248);
     if ($switch$split2D) {
      switch ($19|0) {
      case 16711680:  {
       break;
      }
      default: {
       break L8;
      }
      }
      $20 = ((($1)) + 16|0);
      $21 = HEAP32[$20>>2]|0;
      $22 = ($21|0)==(65280);
      if (!($22)) {
       break;
      }
      $23 = ((($1)) + 20|0);
      $24 = HEAP32[$23>>2]|0;
      $25 = ($24|0)==(255);
      if ($25) {
       $$0 = 4510;
      } else {
       break;
      }
      return ($$0|0);
     } else {
      switch ($19|0) {
      case 1072693248:  {
       break;
      }
      default: {
       break L8;
      }
      }
      $26 = ((($1)) + 16|0);
      $27 = HEAP32[$26>>2]|0;
      $28 = ($27|0)==(1047552);
      if (!($28)) {
       break;
      }
      $29 = ((($1)) + 20|0);
      $30 = HEAP32[$29>>2]|0;
      $31 = ($30|0)==(1023);
      if ($31) {
       $$0 = 4511;
      } else {
       break;
      }
      return ($$0|0);
     }
    }
   } while(0);
   $$0 = 4512;
   return ($$0|0);
  }
  $32 = ((($6)) + 24|0);
  $33 = HEAP32[$32>>2]|0;
  $34 = ($33|0)==(0);
  if ($34) {
   $a_need$0 = 1;
  } else {
   $35 = ((($1)) + 24|0);
   $36 = HEAP32[$35>>2]|0;
   $37 = ($36|0)!=(0);
   $38 = $37 ? 4 : 2;
   $a_need$0 = $38;
  }
  $39 = ((($1)) + 9|0);
  $40 = HEAP8[$39>>0]|0;
  $41 = $40&255;
  $42 = (($41) + -1)|0;
  $43 = (126680 + ($42<<2)|0);
  $44 = HEAP32[$43>>2]|0;
  $45 = ((($44)) + 12|0);
  $46 = HEAP32[$45>>2]|0;
  $47 = ($46|0)==(0);
  L25: do {
   if ($47) {
    $which$0$lcssa = 0;
   } else {
    $48 = ((($1)) + 12|0);
    $49 = ((($1)) + 16|0);
    $50 = ((($1)) + 20|0);
    $51 = ((($6)) + 12|0);
    $52 = ((($6)) + 16|0);
    $53 = ((($6)) + 20|0);
    $54 = ((($6)) + 9|0);
    $88 = $46;$which$06 = 0;
    while(1) {
     $55 = HEAP32[$48>>2]|0;
     $56 = (($44) + (($which$06*40)|0)|0);
     $57 = HEAP32[$56>>2]|0;
     $58 = ($55|0)==($57|0);
     $59 = ($57|0)==(0);
     $or$cond = $58 | $59;
     if ($or$cond) {
      $60 = HEAP32[$49>>2]|0;
      $61 = (((($44) + (($which$06*40)|0)|0)) + 4|0);
      $62 = HEAP32[$61>>2]|0;
      $63 = ($60|0)==($62|0);
      $64 = ($62|0)==(0);
      $or$cond1 = $63 | $64;
      if ($or$cond1) {
       $65 = HEAP32[$50>>2]|0;
       $66 = (((($44) + (($which$06*40)|0)|0)) + 8|0);
       $67 = HEAP32[$66>>2]|0;
       $68 = ($65|0)==($67|0);
       $69 = ($67|0)==(0);
       $or$cond2 = $68 | $69;
       if ($or$cond2) {
        $70 = HEAP32[$51>>2]|0;
        $71 = (((($44) + (($which$06*40)|0)|0)) + 16|0);
        $72 = HEAP32[$71>>2]|0;
        $73 = ($70|0)==($72|0);
        $74 = ($72|0)==(0);
        $or$cond3 = $73 | $74;
        if ($or$cond3) {
         $75 = HEAP32[$52>>2]|0;
         $76 = (((($44) + (($which$06*40)|0)|0)) + 20|0);
         $77 = HEAP32[$76>>2]|0;
         $78 = ($75|0)==($77|0);
         $79 = ($77|0)==(0);
         $or$cond4 = $78 | $79;
         if ($or$cond4) {
          $80 = HEAP32[$53>>2]|0;
          $81 = (((($44) + (($which$06*40)|0)|0)) + 24|0);
          $82 = HEAP32[$81>>2]|0;
          $83 = ($80|0)==($82|0);
          $84 = ($82|0)==(0);
          $or$cond5 = $83 | $84;
          if ($or$cond5) {
           $85 = HEAP8[$54>>0]|0;
           $86 = $85&255;
           $87 = ($86|0)==($88|0);
           if ($87) {
            $89 = (((($44) + (($which$06*40)|0)|0)) + 36|0);
            $90 = HEAP32[$89>>2]|0;
            $91 = $90 & $a_need$0;
            $92 = ($91|0)==($a_need$0|0);
            if ($92) {
             $93 = (((($44) + (($which$06*40)|0)|0)) + 28|0);
             $94 = HEAP32[$93>>2]|0;
             $95 = (_SDL_HasMMX()|0);
             $96 = ($95|0)!=(0);
             $97 = $96&1;
             $98 = $97 & $94;
             $99 = HEAP32[$93>>2]|0;
             $100 = ($98|0)==($99|0);
             if ($100) {
              $which$0$lcssa = $which$06;
              break L25;
             }
            }
           }
          }
         }
        }
       }
      }
     }
     $101 = (($which$06) + 1)|0;
     $102 = (((($44) + (($101*40)|0)|0)) + 12|0);
     $103 = HEAP32[$102>>2]|0;
     $104 = ($103|0)==(0);
     if ($104) {
      $which$0$lcssa = $101;
      break;
     } else {
      $88 = $103;$which$06 = $101;
     }
    }
   }
  } while(0);
  $105 = (((($44) + (($which$0$lcssa*40)|0)|0)) + 32|0);
  $106 = HEAP32[$105>>2]|0;
  $107 = ($106|0)==(2858|0);
  if (!($107)) {
   $$0 = $106;
   return ($$0|0);
  }
  $108 = HEAP32[$1>>2]|0;
  $109 = ($108|0)==(372711428);
  if ($109) {
   $$0 = 4513;
   return ($$0|0);
  }
  $110 = HEAP32[$6>>2]|0;
  $111 = ($110|0)==(372711428);
  if ($111) {
   $$0 = 4514;
   return ($$0|0);
  }
  $112 = HEAP8[$39>>0]|0;
  $113 = ($112<<24>>24)==(4);
  if (!($113)) {
   $137 = ($a_need$0|0)==(4);
   $BlitNtoNCopyAlpha$ = $137 ? 4509 : 2858;
   return ($BlitNtoNCopyAlpha$|0);
  }
  $114 = ((($6)) + 9|0);
  $115 = HEAP8[$114>>0]|0;
  $116 = ($115<<24>>24)==(4);
  if (!($116)) {
   $137 = ($a_need$0|0)==(4);
   $BlitNtoNCopyAlpha$ = $137 ? 4509 : 2858;
   return ($BlitNtoNCopyAlpha$|0);
  }
  $117 = ((($1)) + 12|0);
  $118 = HEAP32[$117>>2]|0;
  $119 = ((($6)) + 12|0);
  $120 = HEAP32[$119>>2]|0;
  $121 = ($118|0)==($120|0);
  if (!($121)) {
   $137 = ($a_need$0|0)==(4);
   $BlitNtoNCopyAlpha$ = $137 ? 4509 : 2858;
   return ($BlitNtoNCopyAlpha$|0);
  }
  $122 = ((($1)) + 16|0);
  $123 = HEAP32[$122>>2]|0;
  $124 = ((($6)) + 16|0);
  $125 = HEAP32[$124>>2]|0;
  $126 = ($123|0)==($125|0);
  if (!($126)) {
   $137 = ($a_need$0|0)==(4);
   $BlitNtoNCopyAlpha$ = $137 ? 4509 : 2858;
   return ($BlitNtoNCopyAlpha$|0);
  }
  $127 = ((($1)) + 20|0);
  $128 = HEAP32[$127>>2]|0;
  $129 = ((($6)) + 20|0);
  $130 = HEAP32[$129>>2]|0;
  $131 = ($128|0)==($130|0);
  if (!($131)) {
   $137 = ($a_need$0|0)==(4);
   $BlitNtoNCopyAlpha$ = $137 ? 4509 : 2858;
   return ($BlitNtoNCopyAlpha$|0);
  }
  $132 = ($a_need$0|0)==(4);
  if (!($132)) {
   $$0 = 4515;
   return ($$0|0);
  }
  $133 = ((($1)) + 24|0);
  $134 = HEAP32[$133>>2]|0;
  $135 = HEAP32[$32>>2]|0;
  $136 = ($134|0)==($135|0);
  $Blit4to4CopyAlpha$BlitNtoNCopyAlpha = $136 ? 4508 : 4509;
  $$0 = $Blit4to4CopyAlpha$BlitNtoNCopyAlpha;
  return ($$0|0);
  break;
 }
 case 256:  {
  $138 = ((($1)) + 9|0);
  $139 = HEAP8[$138>>0]|0;
  $140 = ($139<<24>>24)==(2);
  if ($140) {
   $141 = ((($3)) + 4|0);
   $142 = HEAP32[$141>>2]|0;
   $143 = ($142|0)==(0);
   if (!($143)) {
    $$0 = 4516;
    return ($$0|0);
   }
  }
  $$mask = $8 & -256;
  $144 = ($$mask<<16>>16)==(256);
  if ($144) {
   $$0 = 4517;
   return ($$0|0);
  }
  $145 = ((($1)) + 24|0);
  $146 = HEAP32[$145>>2]|0;
  $147 = ($146|0)==(0);
  if (!($147)) {
   $148 = ((($6)) + 24|0);
   $149 = HEAP32[$148>>2]|0;
   $150 = ($149|0)==(0);
   if (!($150)) {
    $$0 = 4518;
    return ($$0|0);
   }
  }
  $$0 = 4519;
  return ($$0|0);
  break;
 }
 default: {
  $$0 = 0;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function _BlitNtoN($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0;
 var $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0;
 var $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0;
 var $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0;
 var $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0;
 var $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0;
 var $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0;
 var $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0;
 var $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0;
 var $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0;
 var $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0;
 var $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0;
 var $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0;
 var $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0;
 var $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0;
 var $1267 = 0, $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0;
 var $1285 = 0, $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0;
 var $1302 = 0, $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0;
 var $1320 = 0, $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0;
 var $1339 = 0, $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0;
 var $1357 = 0, $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0;
 var $1375 = 0, $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0;
 var $1393 = 0, $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0;
 var $1410 = 0, $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0;
 var $1429 = 0, $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0;
 var $1447 = 0, $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0;
 var $1465 = 0, $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0;
 var $1483 = 0, $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0;
 var $1500 = 0, $1501 = 0, $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0, $1511 = 0, $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0, $1518 = 0;
 var $1519 = 0, $152 = 0, $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0, $153 = 0, $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0, $1536 = 0;
 var $1537 = 0, $1538 = 0, $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0, $1548 = 0, $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0, $1554 = 0;
 var $1555 = 0, $1556 = 0, $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $1565 = 0, $1566 = 0, $1567 = 0, $1568 = 0, $1569 = 0, $157 = 0, $1570 = 0, $1571 = 0, $1572 = 0;
 var $1573 = 0, $1574 = 0, $1575 = 0, $1576 = 0, $1577 = 0, $1578 = 0, $1579 = 0, $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0, $1583 = 0, $1584 = 0, $1585 = 0, $1586 = 0, $1587 = 0, $1588 = 0, $1589 = 0, $159 = 0, $1590 = 0;
 var $1591 = 0, $1592 = 0, $1593 = 0, $1594 = 0, $1595 = 0, $1596 = 0, $1597 = 0, $1598 = 0, $1599 = 0, $16 = 0, $160 = 0, $1600 = 0, $1601 = 0, $1602 = 0, $1603 = 0, $1604 = 0, $1605 = 0, $1606 = 0, $1607 = 0, $1608 = 0;
 var $1609 = 0, $161 = 0, $1610 = 0, $1611 = 0, $1612 = 0, $1613 = 0, $1614 = 0, $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0, $1619 = 0, $162 = 0, $1620 = 0, $1621 = 0, $1622 = 0, $1623 = 0, $1624 = 0, $1625 = 0, $1626 = 0;
 var $1627 = 0, $1628 = 0, $1629 = 0, $163 = 0, $1630 = 0, $1631 = 0, $1632 = 0, $1633 = 0, $1634 = 0, $1635 = 0, $1636 = 0, $1637 = 0, $1638 = 0, $1639 = 0, $164 = 0, $1640 = 0, $1641 = 0, $1642 = 0, $1643 = 0, $1644 = 0;
 var $1645 = 0, $1646 = 0, $1647 = 0, $1648 = 0, $1649 = 0, $165 = 0, $1650 = 0, $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0, $1655 = 0, $1656 = 0, $1657 = 0, $1658 = 0, $1659 = 0, $166 = 0, $1660 = 0, $1661 = 0, $1662 = 0;
 var $1663 = 0, $1664 = 0, $1665 = 0, $1666 = 0, $1667 = 0, $1668 = 0, $1669 = 0, $167 = 0, $1670 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0;
 var $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0;
 var $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0;
 var $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0;
 var $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0;
 var $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0;
 var $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0;
 var $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0;
 var $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0;
 var $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0;
 var $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0;
 var $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0;
 var $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0;
 var $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0;
 var $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0;
 var $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0;
 var $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0;
 var $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0;
 var $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0;
 var $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0;
 var $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0;
 var $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0;
 var $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0;
 var $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0;
 var $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0;
 var $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0;
 var $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0;
 var $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0;
 var $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0;
 var $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0;
 var $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0;
 var $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0;
 var $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0;
 var $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0;
 var $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0;
 var $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0;
 var $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0;
 var $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0;
 var $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0;
 var $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0;
 var $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0;
 var $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0;
 var $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0;
 var $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0;
 var $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0;
 var $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $div = 0, $div1 = 0, $div10 = 0, $div12 = 0, $div13 = 0, $div14 = 0, $div15 = 0, $div16 = 0;
 var $div18 = 0, $div19 = 0, $div2 = 0, $div20 = 0, $div21 = 0, $div22 = 0, $div24 = 0, $div25 = 0, $div26 = 0, $div27 = 0, $div28 = 0, $div3 = 0, $div30 = 0, $div31 = 0, $div32 = 0, $div33 = 0, $div34 = 0, $div36 = 0, $div37 = 0, $div38 = 0;
 var $div39 = 0, $div4 = 0, $div40 = 0, $div42 = 0, $div43 = 0, $div44 = 0, $div45 = 0, $div46 = 0, $div6 = 0, $div7 = 0, $div8 = 0, $div9 = 0, $dst$048 = 0, $dst$1 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0;
 var $dst$8 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $sB$0 = 0, $sB$0$shrunk = 0, $sB13$0 = 0, $sB13$0$shrunk = 0, $sB20$0 = 0, $sB20$0$shrunk = 0, $sB27$0 = 0, $sB27$0$shrunk = 0, $sB34$0 = 0, $sB34$0$shrunk = 0, $sB41$0 = 0;
 var $sB41$0$shrunk = 0, $sB48$0 = 0, $sB48$0$shrunk = 0, $sB6$0 = 0, $sB6$0$shrunk = 0, $sG$0 = 0, $sG$0$shrunk = 0, $sG12$0 = 0, $sG12$0$shrunk = 0, $sG19$0 = 0, $sG19$0$shrunk = 0, $sG26$0 = 0, $sG26$0$shrunk = 0, $sG33$0 = 0, $sG33$0$shrunk = 0, $sG40$0 = 0, $sG40$0$shrunk = 0, $sG47$0 = 0, $sG47$0$shrunk = 0, $sG5$0 = 0;
 var $sG5$0$shrunk = 0, $sR$0 = 0, $sR$0$shrunk = 0, $sR11$0 = 0, $sR11$0$shrunk = 0, $sR18$0 = 0, $sR18$0$shrunk = 0, $sR25$0 = 0, $sR25$0$shrunk = 0, $sR32$0 = 0, $sR32$0$shrunk = 0, $sR39$0 = 0, $sR39$0$shrunk = 0, $sR4$0 = 0, $sR4$0$shrunk = 0, $sR46$0 = 0, $sR46$0$shrunk = 0, $src$049 = 0, $src$1 = 0, $src$2 = 0;
 var $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$info>>2]|0;
 $5 = ((($info)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($info)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($info)) + 36|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($info)) + 40|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($12)) + 9|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = $14&255;
 $16 = ((($info)) + 44|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ((($17)) + 9|0);
 $19 = HEAP8[$18>>0]|0;
 $20 = $19&255;
 $21 = ((($17)) + 24|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ($22|0)==(0);
 if ($23) {
  $178 = 0;
 } else {
  $24 = ((($info)) + 63|0);
  $25 = HEAP8[$24>>0]|0;
  $26 = $25&255;
  $178 = $26;
 }
 $27 = ($3|0)==(0);
 if ($27) {
  return;
 }
 $28 = (($1) + 7)|0;
 $29 = (($28|0) / 8)&-1;
 $30 = $1 & 7;
 $31 = ((($17)) + 28|0);
 $32 = ((($17)) + 32|0);
 $33 = ((($17)) + 33|0);
 $34 = ((($17)) + 34|0);
 $35 = ((($12)) + 12|0);
 $36 = ((($12)) + 32|0);
 $37 = ((($12)) + 28|0);
 $38 = ((($12)) + 16|0);
 $39 = ((($12)) + 20|0);
 $$in = $3;$dst$048 = $8;$src$049 = $4;
 L7: while(1) {
  $40 = (($$in) + -1)|0;
  switch ($30|0) {
  case 0:  {
   $dst$1 = $dst$048;$n$0 = $29;$src$1 = $src$049;
   label = 6;
   break;
  }
  case 7:  {
   $dst$2 = $dst$048;$n$1 = $29;$src$2 = $src$049;
   label = 17;
   break;
  }
  case 6:  {
   $dst$3 = $dst$048;$n$2 = $29;$src$3 = $src$049;
   label = 28;
   break;
  }
  case 5:  {
   $dst$4 = $dst$048;$n$3 = $29;$src$4 = $src$049;
   label = 39;
   break;
  }
  case 4:  {
   $dst$5 = $dst$048;$n$4 = $29;$src$5 = $src$049;
   label = 50;
   break;
  }
  case 3:  {
   $dst$6 = $dst$048;$n$5 = $29;$src$6 = $src$049;
   label = 61;
   break;
  }
  case 2:  {
   $dst$7 = $dst$048;$n$6 = $29;$src$7 = $src$049;
   label = 72;
   break;
  }
  case 1:  {
   $dst$8 = $dst$048;$n$7 = $29;$src$8 = $src$049;
   label = 83;
   break;
  }
  default: {
   label = 94;
   break L7;
  }
  }
  while(1) {
   if ((label|0) == 6) {
    label = 0;
    switch ($15|0) {
    case 1:  {
     $41 = HEAP8[$src$1>>0]|0;
     $42 = $41&255;
     $43 = HEAP32[$35>>2]|0;
     $44 = $42 & $43;
     $45 = HEAP32[$36>>2]|0;
     $46 = $45 & 255;
     $47 = $44 >>> $46;
     $48 = HEAP32[$37>>2]|0;
     $49 = $48 & 255;
     $50 = (125924 + ($49<<2)|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = (($51) + ($47)|0);
     $53 = HEAP8[$52>>0]|0;
     $54 = HEAP32[$38>>2]|0;
     $55 = $54 & $42;
     $56 = $45 >>> 8;
     $57 = $56 & 255;
     $58 = $55 >>> $57;
     $59 = $48 >>> 8;
     $60 = $59 & 255;
     $61 = (125924 + ($60<<2)|0);
     $62 = HEAP32[$61>>2]|0;
     $63 = (($62) + ($58)|0);
     $64 = HEAP8[$63>>0]|0;
     $65 = HEAP32[$39>>2]|0;
     $66 = $65 & $42;
     $67 = $45 >>> 16;
     $68 = $67 & 255;
     $69 = $66 >>> $68;
     $70 = $48 >>> 16;
     $71 = $70 & 255;
     $72 = (125924 + ($71<<2)|0);
     $73 = HEAP32[$72>>2]|0;
     $74 = (($73) + ($69)|0);
     $75 = HEAP8[$74>>0]|0;
     $sB$0$shrunk = $75;$sG$0$shrunk = $64;$sR$0$shrunk = $53;
     break;
    }
    case 2:  {
     $76 = HEAP16[$src$1>>1]|0;
     $77 = $76&65535;
     $78 = HEAP32[$35>>2]|0;
     $79 = $77 & $78;
     $80 = HEAP32[$36>>2]|0;
     $81 = $80 & 255;
     $82 = $79 >>> $81;
     $83 = HEAP32[$37>>2]|0;
     $84 = $83 & 255;
     $85 = (125924 + ($84<<2)|0);
     $86 = HEAP32[$85>>2]|0;
     $87 = (($86) + ($82)|0);
     $88 = HEAP8[$87>>0]|0;
     $89 = HEAP32[$38>>2]|0;
     $90 = $89 & $77;
     $91 = $80 >>> 8;
     $92 = $91 & 255;
     $93 = $90 >>> $92;
     $94 = $83 >>> 8;
     $95 = $94 & 255;
     $96 = (125924 + ($95<<2)|0);
     $97 = HEAP32[$96>>2]|0;
     $98 = (($97) + ($93)|0);
     $99 = HEAP8[$98>>0]|0;
     $100 = HEAP32[$39>>2]|0;
     $101 = $100 & $77;
     $102 = $80 >>> 16;
     $103 = $102 & 255;
     $104 = $101 >>> $103;
     $105 = $83 >>> 16;
     $106 = $105 & 255;
     $107 = (125924 + ($106<<2)|0);
     $108 = HEAP32[$107>>2]|0;
     $109 = (($108) + ($104)|0);
     $110 = HEAP8[$109>>0]|0;
     $sB$0$shrunk = $110;$sG$0$shrunk = $99;$sR$0$shrunk = $88;
     break;
    }
    case 3:  {
     $111 = HEAP32[$36>>2]|0;
     $div39 = $111 >>> 3;
     $112 = $div39 & 31;
     $113 = (($src$1) + ($112)|0);
     $114 = HEAP8[$113>>0]|0;
     $115 = $111 >>> 11;
     $div40 = $115 & 31;
     $116 = (($src$1) + ($div40)|0);
     $117 = HEAP8[$116>>0]|0;
     $118 = $111 >>> 19;
     $119 = $118 & 31;
     $120 = (($src$1) + ($119)|0);
     $121 = HEAP8[$120>>0]|0;
     $sB$0$shrunk = $121;$sG$0$shrunk = $117;$sR$0$shrunk = $114;
     break;
    }
    case 4:  {
     $122 = HEAP32[$src$1>>2]|0;
     $123 = HEAP32[$35>>2]|0;
     $124 = $123 & $122;
     $125 = HEAP32[$36>>2]|0;
     $126 = $125 & 255;
     $127 = $124 >>> $126;
     $128 = HEAP32[$37>>2]|0;
     $129 = $128 & 255;
     $130 = (125924 + ($129<<2)|0);
     $131 = HEAP32[$130>>2]|0;
     $132 = (($131) + ($127)|0);
     $133 = HEAP8[$132>>0]|0;
     $134 = HEAP32[$38>>2]|0;
     $135 = $134 & $122;
     $136 = $125 >>> 8;
     $137 = $136 & 255;
     $138 = $135 >>> $137;
     $139 = $128 >>> 8;
     $140 = $139 & 255;
     $141 = (125924 + ($140<<2)|0);
     $142 = HEAP32[$141>>2]|0;
     $143 = (($142) + ($138)|0);
     $144 = HEAP8[$143>>0]|0;
     $145 = HEAP32[$39>>2]|0;
     $146 = $145 & $122;
     $147 = $125 >>> 16;
     $148 = $147 & 255;
     $149 = $146 >>> $148;
     $150 = $128 >>> 16;
     $151 = $150 & 255;
     $152 = (125924 + ($151<<2)|0);
     $153 = HEAP32[$152>>2]|0;
     $154 = (($153) + ($149)|0);
     $155 = HEAP8[$154>>0]|0;
     $sB$0$shrunk = $155;$sG$0$shrunk = $144;$sR$0$shrunk = $133;
     break;
    }
    default: {
     $sB$0$shrunk = 0;$sG$0$shrunk = 0;$sR$0$shrunk = 0;
    }
    }
    $sR$0 = $sR$0$shrunk&255;
    $sG$0 = $sG$0$shrunk&255;
    $sB$0 = $sB$0$shrunk&255;
    switch ($20|0) {
    case 1:  {
     $156 = HEAP32[$31>>2]|0;
     $157 = $156 & 255;
     $158 = $sR$0 >>> $157;
     $159 = HEAP32[$32>>2]|0;
     $160 = $159 & 255;
     $161 = $158 << $160;
     $162 = $156 >>> 8;
     $163 = $162 & 255;
     $164 = $sG$0 >>> $163;
     $165 = $159 >>> 8;
     $166 = $165 & 255;
     $167 = $164 << $166;
     $168 = $167 | $161;
     $169 = $156 >>> 16;
     $170 = $169 & 255;
     $171 = $sB$0 >>> $170;
     $172 = $159 >>> 16;
     $173 = $172 & 255;
     $174 = $171 << $173;
     $175 = $168 | $174;
     $176 = $156 >>> 24;
     $177 = $178 >>> $176;
     $179 = $159 >>> 24;
     $180 = $177 << $179;
     $181 = $175 | $180;
     $182 = $181&255;
     HEAP8[$dst$1>>0] = $182;
     break;
    }
    case 2:  {
     $183 = HEAP32[$31>>2]|0;
     $184 = $183 & 255;
     $185 = $sR$0 >>> $184;
     $186 = HEAP32[$32>>2]|0;
     $187 = $186 & 255;
     $188 = $185 << $187;
     $189 = $183 >>> 8;
     $190 = $189 & 255;
     $191 = $sG$0 >>> $190;
     $192 = $186 >>> 8;
     $193 = $192 & 255;
     $194 = $191 << $193;
     $195 = $194 | $188;
     $196 = $183 >>> 16;
     $197 = $196 & 255;
     $198 = $sB$0 >>> $197;
     $199 = $186 >>> 16;
     $200 = $199 & 255;
     $201 = $198 << $200;
     $202 = $195 | $201;
     $203 = $183 >>> 24;
     $204 = $178 >>> $203;
     $205 = $186 >>> 24;
     $206 = $204 << $205;
     $207 = $202 | $206;
     $208 = $207&65535;
     HEAP16[$dst$1>>1] = $208;
     break;
    }
    case 3:  {
     $209 = HEAP8[$32>>0]|0;
     $div36 = ($209&255) >>> 3;
     $210 = $div36&255;
     $211 = (($dst$1) + ($210)|0);
     HEAP8[$211>>0] = $sR$0$shrunk;
     $212 = HEAP8[$33>>0]|0;
     $div37 = ($212&255) >>> 3;
     $213 = $div37&255;
     $214 = (($dst$1) + ($213)|0);
     HEAP8[$214>>0] = $sG$0$shrunk;
     $215 = HEAP8[$34>>0]|0;
     $div38 = ($215&255) >>> 3;
     $216 = $div38&255;
     $217 = (($dst$1) + ($216)|0);
     HEAP8[$217>>0] = $sB$0$shrunk;
     break;
    }
    case 4:  {
     $218 = HEAP32[$31>>2]|0;
     $219 = $218 & 255;
     $220 = $sR$0 >>> $219;
     $221 = HEAP32[$32>>2]|0;
     $222 = $221 & 255;
     $223 = $220 << $222;
     $224 = $218 >>> 8;
     $225 = $224 & 255;
     $226 = $sG$0 >>> $225;
     $227 = $221 >>> 8;
     $228 = $227 & 255;
     $229 = $226 << $228;
     $230 = $229 | $223;
     $231 = $218 >>> 16;
     $232 = $231 & 255;
     $233 = $sB$0 >>> $232;
     $234 = $221 >>> 16;
     $235 = $234 & 255;
     $236 = $233 << $235;
     $237 = $230 | $236;
     $238 = $218 >>> 24;
     $239 = $178 >>> $238;
     $240 = $221 >>> 24;
     $241 = $239 << $240;
     $242 = $237 | $241;
     HEAP32[$dst$1>>2] = $242;
     break;
    }
    default: {
    }
    }
    $243 = (($dst$1) + ($20)|0);
    $244 = (($src$1) + ($15)|0);
    $dst$2 = $243;$n$1 = $n$0;$src$2 = $244;
    label = 17;
    continue;
   }
   else if ((label|0) == 17) {
    label = 0;
    switch ($15|0) {
    case 1:  {
     $245 = HEAP8[$src$2>>0]|0;
     $246 = $245&255;
     $247 = HEAP32[$35>>2]|0;
     $248 = $246 & $247;
     $249 = HEAP32[$36>>2]|0;
     $250 = $249 & 255;
     $251 = $248 >>> $250;
     $252 = HEAP32[$37>>2]|0;
     $253 = $252 & 255;
     $254 = (125924 + ($253<<2)|0);
     $255 = HEAP32[$254>>2]|0;
     $256 = (($255) + ($251)|0);
     $257 = HEAP8[$256>>0]|0;
     $258 = HEAP32[$38>>2]|0;
     $259 = $258 & $246;
     $260 = $249 >>> 8;
     $261 = $260 & 255;
     $262 = $259 >>> $261;
     $263 = $252 >>> 8;
     $264 = $263 & 255;
     $265 = (125924 + ($264<<2)|0);
     $266 = HEAP32[$265>>2]|0;
     $267 = (($266) + ($262)|0);
     $268 = HEAP8[$267>>0]|0;
     $269 = HEAP32[$39>>2]|0;
     $270 = $269 & $246;
     $271 = $249 >>> 16;
     $272 = $271 & 255;
     $273 = $270 >>> $272;
     $274 = $252 >>> 16;
     $275 = $274 & 255;
     $276 = (125924 + ($275<<2)|0);
     $277 = HEAP32[$276>>2]|0;
     $278 = (($277) + ($273)|0);
     $279 = HEAP8[$278>>0]|0;
     $sB6$0$shrunk = $279;$sG5$0$shrunk = $268;$sR4$0$shrunk = $257;
     break;
    }
    case 2:  {
     $280 = HEAP16[$src$2>>1]|0;
     $281 = $280&65535;
     $282 = HEAP32[$35>>2]|0;
     $283 = $281 & $282;
     $284 = HEAP32[$36>>2]|0;
     $285 = $284 & 255;
     $286 = $283 >>> $285;
     $287 = HEAP32[$37>>2]|0;
     $288 = $287 & 255;
     $289 = (125924 + ($288<<2)|0);
     $290 = HEAP32[$289>>2]|0;
     $291 = (($290) + ($286)|0);
     $292 = HEAP8[$291>>0]|0;
     $293 = HEAP32[$38>>2]|0;
     $294 = $293 & $281;
     $295 = $284 >>> 8;
     $296 = $295 & 255;
     $297 = $294 >>> $296;
     $298 = $287 >>> 8;
     $299 = $298 & 255;
     $300 = (125924 + ($299<<2)|0);
     $301 = HEAP32[$300>>2]|0;
     $302 = (($301) + ($297)|0);
     $303 = HEAP8[$302>>0]|0;
     $304 = HEAP32[$39>>2]|0;
     $305 = $304 & $281;
     $306 = $284 >>> 16;
     $307 = $306 & 255;
     $308 = $305 >>> $307;
     $309 = $287 >>> 16;
     $310 = $309 & 255;
     $311 = (125924 + ($310<<2)|0);
     $312 = HEAP32[$311>>2]|0;
     $313 = (($312) + ($308)|0);
     $314 = HEAP8[$313>>0]|0;
     $sB6$0$shrunk = $314;$sG5$0$shrunk = $303;$sR4$0$shrunk = $292;
     break;
    }
    case 3:  {
     $315 = HEAP32[$36>>2]|0;
     $div33 = $315 >>> 3;
     $316 = $div33 & 31;
     $317 = (($src$2) + ($316)|0);
     $318 = HEAP8[$317>>0]|0;
     $319 = $315 >>> 11;
     $div34 = $319 & 31;
     $320 = (($src$2) + ($div34)|0);
     $321 = HEAP8[$320>>0]|0;
     $322 = $315 >>> 19;
     $323 = $322 & 31;
     $324 = (($src$2) + ($323)|0);
     $325 = HEAP8[$324>>0]|0;
     $sB6$0$shrunk = $325;$sG5$0$shrunk = $321;$sR4$0$shrunk = $318;
     break;
    }
    case 4:  {
     $326 = HEAP32[$src$2>>2]|0;
     $327 = HEAP32[$35>>2]|0;
     $328 = $327 & $326;
     $329 = HEAP32[$36>>2]|0;
     $330 = $329 & 255;
     $331 = $328 >>> $330;
     $332 = HEAP32[$37>>2]|0;
     $333 = $332 & 255;
     $334 = (125924 + ($333<<2)|0);
     $335 = HEAP32[$334>>2]|0;
     $336 = (($335) + ($331)|0);
     $337 = HEAP8[$336>>0]|0;
     $338 = HEAP32[$38>>2]|0;
     $339 = $338 & $326;
     $340 = $329 >>> 8;
     $341 = $340 & 255;
     $342 = $339 >>> $341;
     $343 = $332 >>> 8;
     $344 = $343 & 255;
     $345 = (125924 + ($344<<2)|0);
     $346 = HEAP32[$345>>2]|0;
     $347 = (($346) + ($342)|0);
     $348 = HEAP8[$347>>0]|0;
     $349 = HEAP32[$39>>2]|0;
     $350 = $349 & $326;
     $351 = $329 >>> 16;
     $352 = $351 & 255;
     $353 = $350 >>> $352;
     $354 = $332 >>> 16;
     $355 = $354 & 255;
     $356 = (125924 + ($355<<2)|0);
     $357 = HEAP32[$356>>2]|0;
     $358 = (($357) + ($353)|0);
     $359 = HEAP8[$358>>0]|0;
     $sB6$0$shrunk = $359;$sG5$0$shrunk = $348;$sR4$0$shrunk = $337;
     break;
    }
    default: {
     $sB6$0$shrunk = 0;$sG5$0$shrunk = 0;$sR4$0$shrunk = 0;
    }
    }
    $sR4$0 = $sR4$0$shrunk&255;
    $sG5$0 = $sG5$0$shrunk&255;
    $sB6$0 = $sB6$0$shrunk&255;
    switch ($20|0) {
    case 1:  {
     $360 = HEAP32[$31>>2]|0;
     $361 = $360 & 255;
     $362 = $sR4$0 >>> $361;
     $363 = HEAP32[$32>>2]|0;
     $364 = $363 & 255;
     $365 = $362 << $364;
     $366 = $360 >>> 8;
     $367 = $366 & 255;
     $368 = $sG5$0 >>> $367;
     $369 = $363 >>> 8;
     $370 = $369 & 255;
     $371 = $368 << $370;
     $372 = $371 | $365;
     $373 = $360 >>> 16;
     $374 = $373 & 255;
     $375 = $sB6$0 >>> $374;
     $376 = $363 >>> 16;
     $377 = $376 & 255;
     $378 = $375 << $377;
     $379 = $372 | $378;
     $380 = $360 >>> 24;
     $381 = $178 >>> $380;
     $382 = $363 >>> 24;
     $383 = $381 << $382;
     $384 = $379 | $383;
     $385 = $384&255;
     HEAP8[$dst$2>>0] = $385;
     break;
    }
    case 2:  {
     $386 = HEAP32[$31>>2]|0;
     $387 = $386 & 255;
     $388 = $sR4$0 >>> $387;
     $389 = HEAP32[$32>>2]|0;
     $390 = $389 & 255;
     $391 = $388 << $390;
     $392 = $386 >>> 8;
     $393 = $392 & 255;
     $394 = $sG5$0 >>> $393;
     $395 = $389 >>> 8;
     $396 = $395 & 255;
     $397 = $394 << $396;
     $398 = $397 | $391;
     $399 = $386 >>> 16;
     $400 = $399 & 255;
     $401 = $sB6$0 >>> $400;
     $402 = $389 >>> 16;
     $403 = $402 & 255;
     $404 = $401 << $403;
     $405 = $398 | $404;
     $406 = $386 >>> 24;
     $407 = $178 >>> $406;
     $408 = $389 >>> 24;
     $409 = $407 << $408;
     $410 = $405 | $409;
     $411 = $410&65535;
     HEAP16[$dst$2>>1] = $411;
     break;
    }
    case 3:  {
     $412 = HEAP8[$32>>0]|0;
     $div30 = ($412&255) >>> 3;
     $413 = $div30&255;
     $414 = (($dst$2) + ($413)|0);
     HEAP8[$414>>0] = $sR4$0$shrunk;
     $415 = HEAP8[$33>>0]|0;
     $div31 = ($415&255) >>> 3;
     $416 = $div31&255;
     $417 = (($dst$2) + ($416)|0);
     HEAP8[$417>>0] = $sG5$0$shrunk;
     $418 = HEAP8[$34>>0]|0;
     $div32 = ($418&255) >>> 3;
     $419 = $div32&255;
     $420 = (($dst$2) + ($419)|0);
     HEAP8[$420>>0] = $sB6$0$shrunk;
     break;
    }
    case 4:  {
     $421 = HEAP32[$31>>2]|0;
     $422 = $421 & 255;
     $423 = $sR4$0 >>> $422;
     $424 = HEAP32[$32>>2]|0;
     $425 = $424 & 255;
     $426 = $423 << $425;
     $427 = $421 >>> 8;
     $428 = $427 & 255;
     $429 = $sG5$0 >>> $428;
     $430 = $424 >>> 8;
     $431 = $430 & 255;
     $432 = $429 << $431;
     $433 = $432 | $426;
     $434 = $421 >>> 16;
     $435 = $434 & 255;
     $436 = $sB6$0 >>> $435;
     $437 = $424 >>> 16;
     $438 = $437 & 255;
     $439 = $436 << $438;
     $440 = $433 | $439;
     $441 = $421 >>> 24;
     $442 = $178 >>> $441;
     $443 = $424 >>> 24;
     $444 = $442 << $443;
     $445 = $440 | $444;
     HEAP32[$dst$2>>2] = $445;
     break;
    }
    default: {
    }
    }
    $446 = (($dst$2) + ($20)|0);
    $447 = (($src$2) + ($15)|0);
    $dst$3 = $446;$n$2 = $n$1;$src$3 = $447;
    label = 28;
    continue;
   }
   else if ((label|0) == 28) {
    label = 0;
    switch ($15|0) {
    case 1:  {
     $448 = HEAP8[$src$3>>0]|0;
     $449 = $448&255;
     $450 = HEAP32[$35>>2]|0;
     $451 = $449 & $450;
     $452 = HEAP32[$36>>2]|0;
     $453 = $452 & 255;
     $454 = $451 >>> $453;
     $455 = HEAP32[$37>>2]|0;
     $456 = $455 & 255;
     $457 = (125924 + ($456<<2)|0);
     $458 = HEAP32[$457>>2]|0;
     $459 = (($458) + ($454)|0);
     $460 = HEAP8[$459>>0]|0;
     $461 = HEAP32[$38>>2]|0;
     $462 = $461 & $449;
     $463 = $452 >>> 8;
     $464 = $463 & 255;
     $465 = $462 >>> $464;
     $466 = $455 >>> 8;
     $467 = $466 & 255;
     $468 = (125924 + ($467<<2)|0);
     $469 = HEAP32[$468>>2]|0;
     $470 = (($469) + ($465)|0);
     $471 = HEAP8[$470>>0]|0;
     $472 = HEAP32[$39>>2]|0;
     $473 = $472 & $449;
     $474 = $452 >>> 16;
     $475 = $474 & 255;
     $476 = $473 >>> $475;
     $477 = $455 >>> 16;
     $478 = $477 & 255;
     $479 = (125924 + ($478<<2)|0);
     $480 = HEAP32[$479>>2]|0;
     $481 = (($480) + ($476)|0);
     $482 = HEAP8[$481>>0]|0;
     $sB13$0$shrunk = $482;$sG12$0$shrunk = $471;$sR11$0$shrunk = $460;
     break;
    }
    case 2:  {
     $483 = HEAP16[$src$3>>1]|0;
     $484 = $483&65535;
     $485 = HEAP32[$35>>2]|0;
     $486 = $484 & $485;
     $487 = HEAP32[$36>>2]|0;
     $488 = $487 & 255;
     $489 = $486 >>> $488;
     $490 = HEAP32[$37>>2]|0;
     $491 = $490 & 255;
     $492 = (125924 + ($491<<2)|0);
     $493 = HEAP32[$492>>2]|0;
     $494 = (($493) + ($489)|0);
     $495 = HEAP8[$494>>0]|0;
     $496 = HEAP32[$38>>2]|0;
     $497 = $496 & $484;
     $498 = $487 >>> 8;
     $499 = $498 & 255;
     $500 = $497 >>> $499;
     $501 = $490 >>> 8;
     $502 = $501 & 255;
     $503 = (125924 + ($502<<2)|0);
     $504 = HEAP32[$503>>2]|0;
     $505 = (($504) + ($500)|0);
     $506 = HEAP8[$505>>0]|0;
     $507 = HEAP32[$39>>2]|0;
     $508 = $507 & $484;
     $509 = $487 >>> 16;
     $510 = $509 & 255;
     $511 = $508 >>> $510;
     $512 = $490 >>> 16;
     $513 = $512 & 255;
     $514 = (125924 + ($513<<2)|0);
     $515 = HEAP32[$514>>2]|0;
     $516 = (($515) + ($511)|0);
     $517 = HEAP8[$516>>0]|0;
     $sB13$0$shrunk = $517;$sG12$0$shrunk = $506;$sR11$0$shrunk = $495;
     break;
    }
    case 3:  {
     $518 = HEAP32[$36>>2]|0;
     $div27 = $518 >>> 3;
     $519 = $div27 & 31;
     $520 = (($src$3) + ($519)|0);
     $521 = HEAP8[$520>>0]|0;
     $522 = $518 >>> 11;
     $div28 = $522 & 31;
     $523 = (($src$3) + ($div28)|0);
     $524 = HEAP8[$523>>0]|0;
     $525 = $518 >>> 19;
     $526 = $525 & 31;
     $527 = (($src$3) + ($526)|0);
     $528 = HEAP8[$527>>0]|0;
     $sB13$0$shrunk = $528;$sG12$0$shrunk = $524;$sR11$0$shrunk = $521;
     break;
    }
    case 4:  {
     $529 = HEAP32[$src$3>>2]|0;
     $530 = HEAP32[$35>>2]|0;
     $531 = $530 & $529;
     $532 = HEAP32[$36>>2]|0;
     $533 = $532 & 255;
     $534 = $531 >>> $533;
     $535 = HEAP32[$37>>2]|0;
     $536 = $535 & 255;
     $537 = (125924 + ($536<<2)|0);
     $538 = HEAP32[$537>>2]|0;
     $539 = (($538) + ($534)|0);
     $540 = HEAP8[$539>>0]|0;
     $541 = HEAP32[$38>>2]|0;
     $542 = $541 & $529;
     $543 = $532 >>> 8;
     $544 = $543 & 255;
     $545 = $542 >>> $544;
     $546 = $535 >>> 8;
     $547 = $546 & 255;
     $548 = (125924 + ($547<<2)|0);
     $549 = HEAP32[$548>>2]|0;
     $550 = (($549) + ($545)|0);
     $551 = HEAP8[$550>>0]|0;
     $552 = HEAP32[$39>>2]|0;
     $553 = $552 & $529;
     $554 = $532 >>> 16;
     $555 = $554 & 255;
     $556 = $553 >>> $555;
     $557 = $535 >>> 16;
     $558 = $557 & 255;
     $559 = (125924 + ($558<<2)|0);
     $560 = HEAP32[$559>>2]|0;
     $561 = (($560) + ($556)|0);
     $562 = HEAP8[$561>>0]|0;
     $sB13$0$shrunk = $562;$sG12$0$shrunk = $551;$sR11$0$shrunk = $540;
     break;
    }
    default: {
     $sB13$0$shrunk = 0;$sG12$0$shrunk = 0;$sR11$0$shrunk = 0;
    }
    }
    $sR11$0 = $sR11$0$shrunk&255;
    $sG12$0 = $sG12$0$shrunk&255;
    $sB13$0 = $sB13$0$shrunk&255;
    switch ($20|0) {
    case 1:  {
     $563 = HEAP32[$31>>2]|0;
     $564 = $563 & 255;
     $565 = $sR11$0 >>> $564;
     $566 = HEAP32[$32>>2]|0;
     $567 = $566 & 255;
     $568 = $565 << $567;
     $569 = $563 >>> 8;
     $570 = $569 & 255;
     $571 = $sG12$0 >>> $570;
     $572 = $566 >>> 8;
     $573 = $572 & 255;
     $574 = $571 << $573;
     $575 = $574 | $568;
     $576 = $563 >>> 16;
     $577 = $576 & 255;
     $578 = $sB13$0 >>> $577;
     $579 = $566 >>> 16;
     $580 = $579 & 255;
     $581 = $578 << $580;
     $582 = $575 | $581;
     $583 = $563 >>> 24;
     $584 = $178 >>> $583;
     $585 = $566 >>> 24;
     $586 = $584 << $585;
     $587 = $582 | $586;
     $588 = $587&255;
     HEAP8[$dst$3>>0] = $588;
     break;
    }
    case 2:  {
     $589 = HEAP32[$31>>2]|0;
     $590 = $589 & 255;
     $591 = $sR11$0 >>> $590;
     $592 = HEAP32[$32>>2]|0;
     $593 = $592 & 255;
     $594 = $591 << $593;
     $595 = $589 >>> 8;
     $596 = $595 & 255;
     $597 = $sG12$0 >>> $596;
     $598 = $592 >>> 8;
     $599 = $598 & 255;
     $600 = $597 << $599;
     $601 = $600 | $594;
     $602 = $589 >>> 16;
     $603 = $602 & 255;
     $604 = $sB13$0 >>> $603;
     $605 = $592 >>> 16;
     $606 = $605 & 255;
     $607 = $604 << $606;
     $608 = $601 | $607;
     $609 = $589 >>> 24;
     $610 = $178 >>> $609;
     $611 = $592 >>> 24;
     $612 = $610 << $611;
     $613 = $608 | $612;
     $614 = $613&65535;
     HEAP16[$dst$3>>1] = $614;
     break;
    }
    case 3:  {
     $615 = HEAP8[$32>>0]|0;
     $div24 = ($615&255) >>> 3;
     $616 = $div24&255;
     $617 = (($dst$3) + ($616)|0);
     HEAP8[$617>>0] = $sR11$0$shrunk;
     $618 = HEAP8[$33>>0]|0;
     $div25 = ($618&255) >>> 3;
     $619 = $div25&255;
     $620 = (($dst$3) + ($619)|0);
     HEAP8[$620>>0] = $sG12$0$shrunk;
     $621 = HEAP8[$34>>0]|0;
     $div26 = ($621&255) >>> 3;
     $622 = $div26&255;
     $623 = (($dst$3) + ($622)|0);
     HEAP8[$623>>0] = $sB13$0$shrunk;
     break;
    }
    case 4:  {
     $624 = HEAP32[$31>>2]|0;
     $625 = $624 & 255;
     $626 = $sR11$0 >>> $625;
     $627 = HEAP32[$32>>2]|0;
     $628 = $627 & 255;
     $629 = $626 << $628;
     $630 = $624 >>> 8;
     $631 = $630 & 255;
     $632 = $sG12$0 >>> $631;
     $633 = $627 >>> 8;
     $634 = $633 & 255;
     $635 = $632 << $634;
     $636 = $635 | $629;
     $637 = $624 >>> 16;
     $638 = $637 & 255;
     $639 = $sB13$0 >>> $638;
     $640 = $627 >>> 16;
     $641 = $640 & 255;
     $642 = $639 << $641;
     $643 = $636 | $642;
     $644 = $624 >>> 24;
     $645 = $178 >>> $644;
     $646 = $627 >>> 24;
     $647 = $645 << $646;
     $648 = $643 | $647;
     HEAP32[$dst$3>>2] = $648;
     break;
    }
    default: {
    }
    }
    $649 = (($dst$3) + ($20)|0);
    $650 = (($src$3) + ($15)|0);
    $dst$4 = $649;$n$3 = $n$2;$src$4 = $650;
    label = 39;
    continue;
   }
   else if ((label|0) == 39) {
    label = 0;
    switch ($15|0) {
    case 1:  {
     $651 = HEAP8[$src$4>>0]|0;
     $652 = $651&255;
     $653 = HEAP32[$35>>2]|0;
     $654 = $652 & $653;
     $655 = HEAP32[$36>>2]|0;
     $656 = $655 & 255;
     $657 = $654 >>> $656;
     $658 = HEAP32[$37>>2]|0;
     $659 = $658 & 255;
     $660 = (125924 + ($659<<2)|0);
     $661 = HEAP32[$660>>2]|0;
     $662 = (($661) + ($657)|0);
     $663 = HEAP8[$662>>0]|0;
     $664 = HEAP32[$38>>2]|0;
     $665 = $664 & $652;
     $666 = $655 >>> 8;
     $667 = $666 & 255;
     $668 = $665 >>> $667;
     $669 = $658 >>> 8;
     $670 = $669 & 255;
     $671 = (125924 + ($670<<2)|0);
     $672 = HEAP32[$671>>2]|0;
     $673 = (($672) + ($668)|0);
     $674 = HEAP8[$673>>0]|0;
     $675 = HEAP32[$39>>2]|0;
     $676 = $675 & $652;
     $677 = $655 >>> 16;
     $678 = $677 & 255;
     $679 = $676 >>> $678;
     $680 = $658 >>> 16;
     $681 = $680 & 255;
     $682 = (125924 + ($681<<2)|0);
     $683 = HEAP32[$682>>2]|0;
     $684 = (($683) + ($679)|0);
     $685 = HEAP8[$684>>0]|0;
     $sB20$0$shrunk = $685;$sG19$0$shrunk = $674;$sR18$0$shrunk = $663;
     break;
    }
    case 2:  {
     $686 = HEAP16[$src$4>>1]|0;
     $687 = $686&65535;
     $688 = HEAP32[$35>>2]|0;
     $689 = $687 & $688;
     $690 = HEAP32[$36>>2]|0;
     $691 = $690 & 255;
     $692 = $689 >>> $691;
     $693 = HEAP32[$37>>2]|0;
     $694 = $693 & 255;
     $695 = (125924 + ($694<<2)|0);
     $696 = HEAP32[$695>>2]|0;
     $697 = (($696) + ($692)|0);
     $698 = HEAP8[$697>>0]|0;
     $699 = HEAP32[$38>>2]|0;
     $700 = $699 & $687;
     $701 = $690 >>> 8;
     $702 = $701 & 255;
     $703 = $700 >>> $702;
     $704 = $693 >>> 8;
     $705 = $704 & 255;
     $706 = (125924 + ($705<<2)|0);
     $707 = HEAP32[$706>>2]|0;
     $708 = (($707) + ($703)|0);
     $709 = HEAP8[$708>>0]|0;
     $710 = HEAP32[$39>>2]|0;
     $711 = $710 & $687;
     $712 = $690 >>> 16;
     $713 = $712 & 255;
     $714 = $711 >>> $713;
     $715 = $693 >>> 16;
     $716 = $715 & 255;
     $717 = (125924 + ($716<<2)|0);
     $718 = HEAP32[$717>>2]|0;
     $719 = (($718) + ($714)|0);
     $720 = HEAP8[$719>>0]|0;
     $sB20$0$shrunk = $720;$sG19$0$shrunk = $709;$sR18$0$shrunk = $698;
     break;
    }
    case 3:  {
     $721 = HEAP32[$36>>2]|0;
     $div21 = $721 >>> 3;
     $722 = $div21 & 31;
     $723 = (($src$4) + ($722)|0);
     $724 = HEAP8[$723>>0]|0;
     $725 = $721 >>> 11;
     $div22 = $725 & 31;
     $726 = (($src$4) + ($div22)|0);
     $727 = HEAP8[$726>>0]|0;
     $728 = $721 >>> 19;
     $729 = $728 & 31;
     $730 = (($src$4) + ($729)|0);
     $731 = HEAP8[$730>>0]|0;
     $sB20$0$shrunk = $731;$sG19$0$shrunk = $727;$sR18$0$shrunk = $724;
     break;
    }
    case 4:  {
     $732 = HEAP32[$src$4>>2]|0;
     $733 = HEAP32[$35>>2]|0;
     $734 = $733 & $732;
     $735 = HEAP32[$36>>2]|0;
     $736 = $735 & 255;
     $737 = $734 >>> $736;
     $738 = HEAP32[$37>>2]|0;
     $739 = $738 & 255;
     $740 = (125924 + ($739<<2)|0);
     $741 = HEAP32[$740>>2]|0;
     $742 = (($741) + ($737)|0);
     $743 = HEAP8[$742>>0]|0;
     $744 = HEAP32[$38>>2]|0;
     $745 = $744 & $732;
     $746 = $735 >>> 8;
     $747 = $746 & 255;
     $748 = $745 >>> $747;
     $749 = $738 >>> 8;
     $750 = $749 & 255;
     $751 = (125924 + ($750<<2)|0);
     $752 = HEAP32[$751>>2]|0;
     $753 = (($752) + ($748)|0);
     $754 = HEAP8[$753>>0]|0;
     $755 = HEAP32[$39>>2]|0;
     $756 = $755 & $732;
     $757 = $735 >>> 16;
     $758 = $757 & 255;
     $759 = $756 >>> $758;
     $760 = $738 >>> 16;
     $761 = $760 & 255;
     $762 = (125924 + ($761<<2)|0);
     $763 = HEAP32[$762>>2]|0;
     $764 = (($763) + ($759)|0);
     $765 = HEAP8[$764>>0]|0;
     $sB20$0$shrunk = $765;$sG19$0$shrunk = $754;$sR18$0$shrunk = $743;
     break;
    }
    default: {
     $sB20$0$shrunk = 0;$sG19$0$shrunk = 0;$sR18$0$shrunk = 0;
    }
    }
    $sB20$0 = $sB20$0$shrunk&255;
    $sG19$0 = $sG19$0$shrunk&255;
    $sR18$0 = $sR18$0$shrunk&255;
    switch ($20|0) {
    case 1:  {
     $766 = HEAP32[$31>>2]|0;
     $767 = $766 & 255;
     $768 = $sR18$0 >>> $767;
     $769 = HEAP32[$32>>2]|0;
     $770 = $769 & 255;
     $771 = $768 << $770;
     $772 = $766 >>> 8;
     $773 = $772 & 255;
     $774 = $sG19$0 >>> $773;
     $775 = $769 >>> 8;
     $776 = $775 & 255;
     $777 = $774 << $776;
     $778 = $777 | $771;
     $779 = $766 >>> 16;
     $780 = $779 & 255;
     $781 = $sB20$0 >>> $780;
     $782 = $769 >>> 16;
     $783 = $782 & 255;
     $784 = $781 << $783;
     $785 = $778 | $784;
     $786 = $766 >>> 24;
     $787 = $178 >>> $786;
     $788 = $769 >>> 24;
     $789 = $787 << $788;
     $790 = $785 | $789;
     $791 = $790&255;
     HEAP8[$dst$4>>0] = $791;
     break;
    }
    case 2:  {
     $792 = HEAP32[$31>>2]|0;
     $793 = $792 & 255;
     $794 = $sR18$0 >>> $793;
     $795 = HEAP32[$32>>2]|0;
     $796 = $795 & 255;
     $797 = $794 << $796;
     $798 = $792 >>> 8;
     $799 = $798 & 255;
     $800 = $sG19$0 >>> $799;
     $801 = $795 >>> 8;
     $802 = $801 & 255;
     $803 = $800 << $802;
     $804 = $803 | $797;
     $805 = $792 >>> 16;
     $806 = $805 & 255;
     $807 = $sB20$0 >>> $806;
     $808 = $795 >>> 16;
     $809 = $808 & 255;
     $810 = $807 << $809;
     $811 = $804 | $810;
     $812 = $792 >>> 24;
     $813 = $178 >>> $812;
     $814 = $795 >>> 24;
     $815 = $813 << $814;
     $816 = $811 | $815;
     $817 = $816&65535;
     HEAP16[$dst$4>>1] = $817;
     break;
    }
    case 3:  {
     $818 = HEAP8[$32>>0]|0;
     $div18 = ($818&255) >>> 3;
     $819 = $div18&255;
     $820 = (($dst$4) + ($819)|0);
     HEAP8[$820>>0] = $sR18$0$shrunk;
     $821 = HEAP8[$33>>0]|0;
     $div19 = ($821&255) >>> 3;
     $822 = $div19&255;
     $823 = (($dst$4) + ($822)|0);
     HEAP8[$823>>0] = $sG19$0$shrunk;
     $824 = HEAP8[$34>>0]|0;
     $div20 = ($824&255) >>> 3;
     $825 = $div20&255;
     $826 = (($dst$4) + ($825)|0);
     HEAP8[$826>>0] = $sB20$0$shrunk;
     break;
    }
    case 4:  {
     $827 = HEAP32[$31>>2]|0;
     $828 = $827 & 255;
     $829 = $sR18$0 >>> $828;
     $830 = HEAP32[$32>>2]|0;
     $831 = $830 & 255;
     $832 = $829 << $831;
     $833 = $827 >>> 8;
     $834 = $833 & 255;
     $835 = $sG19$0 >>> $834;
     $836 = $830 >>> 8;
     $837 = $836 & 255;
     $838 = $835 << $837;
     $839 = $838 | $832;
     $840 = $827 >>> 16;
     $841 = $840 & 255;
     $842 = $sB20$0 >>> $841;
     $843 = $830 >>> 16;
     $844 = $843 & 255;
     $845 = $842 << $844;
     $846 = $839 | $845;
     $847 = $827 >>> 24;
     $848 = $178 >>> $847;
     $849 = $830 >>> 24;
     $850 = $848 << $849;
     $851 = $846 | $850;
     HEAP32[$dst$4>>2] = $851;
     break;
    }
    default: {
    }
    }
    $852 = (($dst$4) + ($20)|0);
    $853 = (($src$4) + ($15)|0);
    $dst$5 = $852;$n$4 = $n$3;$src$5 = $853;
    label = 50;
    continue;
   }
   else if ((label|0) == 50) {
    label = 0;
    switch ($15|0) {
    case 1:  {
     $854 = HEAP8[$src$5>>0]|0;
     $855 = $854&255;
     $856 = HEAP32[$35>>2]|0;
     $857 = $855 & $856;
     $858 = HEAP32[$36>>2]|0;
     $859 = $858 & 255;
     $860 = $857 >>> $859;
     $861 = HEAP32[$37>>2]|0;
     $862 = $861 & 255;
     $863 = (125924 + ($862<<2)|0);
     $864 = HEAP32[$863>>2]|0;
     $865 = (($864) + ($860)|0);
     $866 = HEAP8[$865>>0]|0;
     $867 = HEAP32[$38>>2]|0;
     $868 = $867 & $855;
     $869 = $858 >>> 8;
     $870 = $869 & 255;
     $871 = $868 >>> $870;
     $872 = $861 >>> 8;
     $873 = $872 & 255;
     $874 = (125924 + ($873<<2)|0);
     $875 = HEAP32[$874>>2]|0;
     $876 = (($875) + ($871)|0);
     $877 = HEAP8[$876>>0]|0;
     $878 = HEAP32[$39>>2]|0;
     $879 = $878 & $855;
     $880 = $858 >>> 16;
     $881 = $880 & 255;
     $882 = $879 >>> $881;
     $883 = $861 >>> 16;
     $884 = $883 & 255;
     $885 = (125924 + ($884<<2)|0);
     $886 = HEAP32[$885>>2]|0;
     $887 = (($886) + ($882)|0);
     $888 = HEAP8[$887>>0]|0;
     $sB27$0$shrunk = $888;$sG26$0$shrunk = $877;$sR25$0$shrunk = $866;
     break;
    }
    case 2:  {
     $889 = HEAP16[$src$5>>1]|0;
     $890 = $889&65535;
     $891 = HEAP32[$35>>2]|0;
     $892 = $890 & $891;
     $893 = HEAP32[$36>>2]|0;
     $894 = $893 & 255;
     $895 = $892 >>> $894;
     $896 = HEAP32[$37>>2]|0;
     $897 = $896 & 255;
     $898 = (125924 + ($897<<2)|0);
     $899 = HEAP32[$898>>2]|0;
     $900 = (($899) + ($895)|0);
     $901 = HEAP8[$900>>0]|0;
     $902 = HEAP32[$38>>2]|0;
     $903 = $902 & $890;
     $904 = $893 >>> 8;
     $905 = $904 & 255;
     $906 = $903 >>> $905;
     $907 = $896 >>> 8;
     $908 = $907 & 255;
     $909 = (125924 + ($908<<2)|0);
     $910 = HEAP32[$909>>2]|0;
     $911 = (($910) + ($906)|0);
     $912 = HEAP8[$911>>0]|0;
     $913 = HEAP32[$39>>2]|0;
     $914 = $913 & $890;
     $915 = $893 >>> 16;
     $916 = $915 & 255;
     $917 = $914 >>> $916;
     $918 = $896 >>> 16;
     $919 = $918 & 255;
     $920 = (125924 + ($919<<2)|0);
     $921 = HEAP32[$920>>2]|0;
     $922 = (($921) + ($917)|0);
     $923 = HEAP8[$922>>0]|0;
     $sB27$0$shrunk = $923;$sG26$0$shrunk = $912;$sR25$0$shrunk = $901;
     break;
    }
    case 3:  {
     $924 = HEAP32[$36>>2]|0;
     $div15 = $924 >>> 3;
     $925 = $div15 & 31;
     $926 = (($src$5) + ($925)|0);
     $927 = HEAP8[$926>>0]|0;
     $928 = $924 >>> 11;
     $div16 = $928 & 31;
     $929 = (($src$5) + ($div16)|0);
     $930 = HEAP8[$929>>0]|0;
     $931 = $924 >>> 19;
     $932 = $931 & 31;
     $933 = (($src$5) + ($932)|0);
     $934 = HEAP8[$933>>0]|0;
     $sB27$0$shrunk = $934;$sG26$0$shrunk = $930;$sR25$0$shrunk = $927;
     break;
    }
    case 4:  {
     $935 = HEAP32[$src$5>>2]|0;
     $936 = HEAP32[$35>>2]|0;
     $937 = $936 & $935;
     $938 = HEAP32[$36>>2]|0;
     $939 = $938 & 255;
     $940 = $937 >>> $939;
     $941 = HEAP32[$37>>2]|0;
     $942 = $941 & 255;
     $943 = (125924 + ($942<<2)|0);
     $944 = HEAP32[$943>>2]|0;
     $945 = (($944) + ($940)|0);
     $946 = HEAP8[$945>>0]|0;
     $947 = HEAP32[$38>>2]|0;
     $948 = $947 & $935;
     $949 = $938 >>> 8;
     $950 = $949 & 255;
     $951 = $948 >>> $950;
     $952 = $941 >>> 8;
     $953 = $952 & 255;
     $954 = (125924 + ($953<<2)|0);
     $955 = HEAP32[$954>>2]|0;
     $956 = (($955) + ($951)|0);
     $957 = HEAP8[$956>>0]|0;
     $958 = HEAP32[$39>>2]|0;
     $959 = $958 & $935;
     $960 = $938 >>> 16;
     $961 = $960 & 255;
     $962 = $959 >>> $961;
     $963 = $941 >>> 16;
     $964 = $963 & 255;
     $965 = (125924 + ($964<<2)|0);
     $966 = HEAP32[$965>>2]|0;
     $967 = (($966) + ($962)|0);
     $968 = HEAP8[$967>>0]|0;
     $sB27$0$shrunk = $968;$sG26$0$shrunk = $957;$sR25$0$shrunk = $946;
     break;
    }
    default: {
     $sB27$0$shrunk = 0;$sG26$0$shrunk = 0;$sR25$0$shrunk = 0;
    }
    }
    $sB27$0 = $sB27$0$shrunk&255;
    $sG26$0 = $sG26$0$shrunk&255;
    $sR25$0 = $sR25$0$shrunk&255;
    switch ($20|0) {
    case 1:  {
     $969 = HEAP32[$31>>2]|0;
     $970 = $969 & 255;
     $971 = $sR25$0 >>> $970;
     $972 = HEAP32[$32>>2]|0;
     $973 = $972 & 255;
     $974 = $971 << $973;
     $975 = $969 >>> 8;
     $976 = $975 & 255;
     $977 = $sG26$0 >>> $976;
     $978 = $972 >>> 8;
     $979 = $978 & 255;
     $980 = $977 << $979;
     $981 = $980 | $974;
     $982 = $969 >>> 16;
     $983 = $982 & 255;
     $984 = $sB27$0 >>> $983;
     $985 = $972 >>> 16;
     $986 = $985 & 255;
     $987 = $984 << $986;
     $988 = $981 | $987;
     $989 = $969 >>> 24;
     $990 = $178 >>> $989;
     $991 = $972 >>> 24;
     $992 = $990 << $991;
     $993 = $988 | $992;
     $994 = $993&255;
     HEAP8[$dst$5>>0] = $994;
     break;
    }
    case 2:  {
     $995 = HEAP32[$31>>2]|0;
     $996 = $995 & 255;
     $997 = $sR25$0 >>> $996;
     $998 = HEAP32[$32>>2]|0;
     $999 = $998 & 255;
     $1000 = $997 << $999;
     $1001 = $995 >>> 8;
     $1002 = $1001 & 255;
     $1003 = $sG26$0 >>> $1002;
     $1004 = $998 >>> 8;
     $1005 = $1004 & 255;
     $1006 = $1003 << $1005;
     $1007 = $1006 | $1000;
     $1008 = $995 >>> 16;
     $1009 = $1008 & 255;
     $1010 = $sB27$0 >>> $1009;
     $1011 = $998 >>> 16;
     $1012 = $1011 & 255;
     $1013 = $1010 << $1012;
     $1014 = $1007 | $1013;
     $1015 = $995 >>> 24;
     $1016 = $178 >>> $1015;
     $1017 = $998 >>> 24;
     $1018 = $1016 << $1017;
     $1019 = $1014 | $1018;
     $1020 = $1019&65535;
     HEAP16[$dst$5>>1] = $1020;
     break;
    }
    case 3:  {
     $1021 = HEAP8[$32>>0]|0;
     $div12 = ($1021&255) >>> 3;
     $1022 = $div12&255;
     $1023 = (($dst$5) + ($1022)|0);
     HEAP8[$1023>>0] = $sR25$0$shrunk;
     $1024 = HEAP8[$33>>0]|0;
     $div13 = ($1024&255) >>> 3;
     $1025 = $div13&255;
     $1026 = (($dst$5) + ($1025)|0);
     HEAP8[$1026>>0] = $sG26$0$shrunk;
     $1027 = HEAP8[$34>>0]|0;
     $div14 = ($1027&255) >>> 3;
     $1028 = $div14&255;
     $1029 = (($dst$5) + ($1028)|0);
     HEAP8[$1029>>0] = $sB27$0$shrunk;
     break;
    }
    case 4:  {
     $1030 = HEAP32[$31>>2]|0;
     $1031 = $1030 & 255;
     $1032 = $sR25$0 >>> $1031;
     $1033 = HEAP32[$32>>2]|0;
     $1034 = $1033 & 255;
     $1035 = $1032 << $1034;
     $1036 = $1030 >>> 8;
     $1037 = $1036 & 255;
     $1038 = $sG26$0 >>> $1037;
     $1039 = $1033 >>> 8;
     $1040 = $1039 & 255;
     $1041 = $1038 << $1040;
     $1042 = $1041 | $1035;
     $1043 = $1030 >>> 16;
     $1044 = $1043 & 255;
     $1045 = $sB27$0 >>> $1044;
     $1046 = $1033 >>> 16;
     $1047 = $1046 & 255;
     $1048 = $1045 << $1047;
     $1049 = $1042 | $1048;
     $1050 = $1030 >>> 24;
     $1051 = $178 >>> $1050;
     $1052 = $1033 >>> 24;
     $1053 = $1051 << $1052;
     $1054 = $1049 | $1053;
     HEAP32[$dst$5>>2] = $1054;
     break;
    }
    default: {
    }
    }
    $1055 = (($dst$5) + ($20)|0);
    $1056 = (($src$5) + ($15)|0);
    $dst$6 = $1055;$n$5 = $n$4;$src$6 = $1056;
    label = 61;
    continue;
   }
   else if ((label|0) == 61) {
    label = 0;
    switch ($15|0) {
    case 1:  {
     $1057 = HEAP8[$src$6>>0]|0;
     $1058 = $1057&255;
     $1059 = HEAP32[$35>>2]|0;
     $1060 = $1058 & $1059;
     $1061 = HEAP32[$36>>2]|0;
     $1062 = $1061 & 255;
     $1063 = $1060 >>> $1062;
     $1064 = HEAP32[$37>>2]|0;
     $1065 = $1064 & 255;
     $1066 = (125924 + ($1065<<2)|0);
     $1067 = HEAP32[$1066>>2]|0;
     $1068 = (($1067) + ($1063)|0);
     $1069 = HEAP8[$1068>>0]|0;
     $1070 = HEAP32[$38>>2]|0;
     $1071 = $1070 & $1058;
     $1072 = $1061 >>> 8;
     $1073 = $1072 & 255;
     $1074 = $1071 >>> $1073;
     $1075 = $1064 >>> 8;
     $1076 = $1075 & 255;
     $1077 = (125924 + ($1076<<2)|0);
     $1078 = HEAP32[$1077>>2]|0;
     $1079 = (($1078) + ($1074)|0);
     $1080 = HEAP8[$1079>>0]|0;
     $1081 = HEAP32[$39>>2]|0;
     $1082 = $1081 & $1058;
     $1083 = $1061 >>> 16;
     $1084 = $1083 & 255;
     $1085 = $1082 >>> $1084;
     $1086 = $1064 >>> 16;
     $1087 = $1086 & 255;
     $1088 = (125924 + ($1087<<2)|0);
     $1089 = HEAP32[$1088>>2]|0;
     $1090 = (($1089) + ($1085)|0);
     $1091 = HEAP8[$1090>>0]|0;
     $sB34$0$shrunk = $1091;$sG33$0$shrunk = $1080;$sR32$0$shrunk = $1069;
     break;
    }
    case 2:  {
     $1092 = HEAP16[$src$6>>1]|0;
     $1093 = $1092&65535;
     $1094 = HEAP32[$35>>2]|0;
     $1095 = $1093 & $1094;
     $1096 = HEAP32[$36>>2]|0;
     $1097 = $1096 & 255;
     $1098 = $1095 >>> $1097;
     $1099 = HEAP32[$37>>2]|0;
     $1100 = $1099 & 255;
     $1101 = (125924 + ($1100<<2)|0);
     $1102 = HEAP32[$1101>>2]|0;
     $1103 = (($1102) + ($1098)|0);
     $1104 = HEAP8[$1103>>0]|0;
     $1105 = HEAP32[$38>>2]|0;
     $1106 = $1105 & $1093;
     $1107 = $1096 >>> 8;
     $1108 = $1107 & 255;
     $1109 = $1106 >>> $1108;
     $1110 = $1099 >>> 8;
     $1111 = $1110 & 255;
     $1112 = (125924 + ($1111<<2)|0);
     $1113 = HEAP32[$1112>>2]|0;
     $1114 = (($1113) + ($1109)|0);
     $1115 = HEAP8[$1114>>0]|0;
     $1116 = HEAP32[$39>>2]|0;
     $1117 = $1116 & $1093;
     $1118 = $1096 >>> 16;
     $1119 = $1118 & 255;
     $1120 = $1117 >>> $1119;
     $1121 = $1099 >>> 16;
     $1122 = $1121 & 255;
     $1123 = (125924 + ($1122<<2)|0);
     $1124 = HEAP32[$1123>>2]|0;
     $1125 = (($1124) + ($1120)|0);
     $1126 = HEAP8[$1125>>0]|0;
     $sB34$0$shrunk = $1126;$sG33$0$shrunk = $1115;$sR32$0$shrunk = $1104;
     break;
    }
    case 3:  {
     $1127 = HEAP32[$36>>2]|0;
     $div9 = $1127 >>> 3;
     $1128 = $div9 & 31;
     $1129 = (($src$6) + ($1128)|0);
     $1130 = HEAP8[$1129>>0]|0;
     $1131 = $1127 >>> 11;
     $div10 = $1131 & 31;
     $1132 = (($src$6) + ($div10)|0);
     $1133 = HEAP8[$1132>>0]|0;
     $1134 = $1127 >>> 19;
     $1135 = $1134 & 31;
     $1136 = (($src$6) + ($1135)|0);
     $1137 = HEAP8[$1136>>0]|0;
     $sB34$0$shrunk = $1137;$sG33$0$shrunk = $1133;$sR32$0$shrunk = $1130;
     break;
    }
    case 4:  {
     $1138 = HEAP32[$src$6>>2]|0;
     $1139 = HEAP32[$35>>2]|0;
     $1140 = $1139 & $1138;
     $1141 = HEAP32[$36>>2]|0;
     $1142 = $1141 & 255;
     $1143 = $1140 >>> $1142;
     $1144 = HEAP32[$37>>2]|0;
     $1145 = $1144 & 255;
     $1146 = (125924 + ($1145<<2)|0);
     $1147 = HEAP32[$1146>>2]|0;
     $1148 = (($1147) + ($1143)|0);
     $1149 = HEAP8[$1148>>0]|0;
     $1150 = HEAP32[$38>>2]|0;
     $1151 = $1150 & $1138;
     $1152 = $1141 >>> 8;
     $1153 = $1152 & 255;
     $1154 = $1151 >>> $1153;
     $1155 = $1144 >>> 8;
     $1156 = $1155 & 255;
     $1157 = (125924 + ($1156<<2)|0);
     $1158 = HEAP32[$1157>>2]|0;
     $1159 = (($1158) + ($1154)|0);
     $1160 = HEAP8[$1159>>0]|0;
     $1161 = HEAP32[$39>>2]|0;
     $1162 = $1161 & $1138;
     $1163 = $1141 >>> 16;
     $1164 = $1163 & 255;
     $1165 = $1162 >>> $1164;
     $1166 = $1144 >>> 16;
     $1167 = $1166 & 255;
     $1168 = (125924 + ($1167<<2)|0);
     $1169 = HEAP32[$1168>>2]|0;
     $1170 = (($1169) + ($1165)|0);
     $1171 = HEAP8[$1170>>0]|0;
     $sB34$0$shrunk = $1171;$sG33$0$shrunk = $1160;$sR32$0$shrunk = $1149;
     break;
    }
    default: {
     $sB34$0$shrunk = 0;$sG33$0$shrunk = 0;$sR32$0$shrunk = 0;
    }
    }
    $sB34$0 = $sB34$0$shrunk&255;
    $sG33$0 = $sG33$0$shrunk&255;
    $sR32$0 = $sR32$0$shrunk&255;
    switch ($20|0) {
    case 1:  {
     $1172 = HEAP32[$31>>2]|0;
     $1173 = $1172 & 255;
     $1174 = $sR32$0 >>> $1173;
     $1175 = HEAP32[$32>>2]|0;
     $1176 = $1175 & 255;
     $1177 = $1174 << $1176;
     $1178 = $1172 >>> 8;
     $1179 = $1178 & 255;
     $1180 = $sG33$0 >>> $1179;
     $1181 = $1175 >>> 8;
     $1182 = $1181 & 255;
     $1183 = $1180 << $1182;
     $1184 = $1183 | $1177;
     $1185 = $1172 >>> 16;
     $1186 = $1185 & 255;
     $1187 = $sB34$0 >>> $1186;
     $1188 = $1175 >>> 16;
     $1189 = $1188 & 255;
     $1190 = $1187 << $1189;
     $1191 = $1184 | $1190;
     $1192 = $1172 >>> 24;
     $1193 = $178 >>> $1192;
     $1194 = $1175 >>> 24;
     $1195 = $1193 << $1194;
     $1196 = $1191 | $1195;
     $1197 = $1196&255;
     HEAP8[$dst$6>>0] = $1197;
     break;
    }
    case 2:  {
     $1198 = HEAP32[$31>>2]|0;
     $1199 = $1198 & 255;
     $1200 = $sR32$0 >>> $1199;
     $1201 = HEAP32[$32>>2]|0;
     $1202 = $1201 & 255;
     $1203 = $1200 << $1202;
     $1204 = $1198 >>> 8;
     $1205 = $1204 & 255;
     $1206 = $sG33$0 >>> $1205;
     $1207 = $1201 >>> 8;
     $1208 = $1207 & 255;
     $1209 = $1206 << $1208;
     $1210 = $1209 | $1203;
     $1211 = $1198 >>> 16;
     $1212 = $1211 & 255;
     $1213 = $sB34$0 >>> $1212;
     $1214 = $1201 >>> 16;
     $1215 = $1214 & 255;
     $1216 = $1213 << $1215;
     $1217 = $1210 | $1216;
     $1218 = $1198 >>> 24;
     $1219 = $178 >>> $1218;
     $1220 = $1201 >>> 24;
     $1221 = $1219 << $1220;
     $1222 = $1217 | $1221;
     $1223 = $1222&65535;
     HEAP16[$dst$6>>1] = $1223;
     break;
    }
    case 3:  {
     $1224 = HEAP8[$32>>0]|0;
     $div6 = ($1224&255) >>> 3;
     $1225 = $div6&255;
     $1226 = (($dst$6) + ($1225)|0);
     HEAP8[$1226>>0] = $sR32$0$shrunk;
     $1227 = HEAP8[$33>>0]|0;
     $div7 = ($1227&255) >>> 3;
     $1228 = $div7&255;
     $1229 = (($dst$6) + ($1228)|0);
     HEAP8[$1229>>0] = $sG33$0$shrunk;
     $1230 = HEAP8[$34>>0]|0;
     $div8 = ($1230&255) >>> 3;
     $1231 = $div8&255;
     $1232 = (($dst$6) + ($1231)|0);
     HEAP8[$1232>>0] = $sB34$0$shrunk;
     break;
    }
    case 4:  {
     $1233 = HEAP32[$31>>2]|0;
     $1234 = $1233 & 255;
     $1235 = $sR32$0 >>> $1234;
     $1236 = HEAP32[$32>>2]|0;
     $1237 = $1236 & 255;
     $1238 = $1235 << $1237;
     $1239 = $1233 >>> 8;
     $1240 = $1239 & 255;
     $1241 = $sG33$0 >>> $1240;
     $1242 = $1236 >>> 8;
     $1243 = $1242 & 255;
     $1244 = $1241 << $1243;
     $1245 = $1244 | $1238;
     $1246 = $1233 >>> 16;
     $1247 = $1246 & 255;
     $1248 = $sB34$0 >>> $1247;
     $1249 = $1236 >>> 16;
     $1250 = $1249 & 255;
     $1251 = $1248 << $1250;
     $1252 = $1245 | $1251;
     $1253 = $1233 >>> 24;
     $1254 = $178 >>> $1253;
     $1255 = $1236 >>> 24;
     $1256 = $1254 << $1255;
     $1257 = $1252 | $1256;
     HEAP32[$dst$6>>2] = $1257;
     break;
    }
    default: {
    }
    }
    $1258 = (($dst$6) + ($20)|0);
    $1259 = (($src$6) + ($15)|0);
    $dst$7 = $1258;$n$6 = $n$5;$src$7 = $1259;
    label = 72;
    continue;
   }
   else if ((label|0) == 72) {
    label = 0;
    switch ($15|0) {
    case 1:  {
     $1260 = HEAP8[$src$7>>0]|0;
     $1261 = $1260&255;
     $1262 = HEAP32[$35>>2]|0;
     $1263 = $1261 & $1262;
     $1264 = HEAP32[$36>>2]|0;
     $1265 = $1264 & 255;
     $1266 = $1263 >>> $1265;
     $1267 = HEAP32[$37>>2]|0;
     $1268 = $1267 & 255;
     $1269 = (125924 + ($1268<<2)|0);
     $1270 = HEAP32[$1269>>2]|0;
     $1271 = (($1270) + ($1266)|0);
     $1272 = HEAP8[$1271>>0]|0;
     $1273 = HEAP32[$38>>2]|0;
     $1274 = $1273 & $1261;
     $1275 = $1264 >>> 8;
     $1276 = $1275 & 255;
     $1277 = $1274 >>> $1276;
     $1278 = $1267 >>> 8;
     $1279 = $1278 & 255;
     $1280 = (125924 + ($1279<<2)|0);
     $1281 = HEAP32[$1280>>2]|0;
     $1282 = (($1281) + ($1277)|0);
     $1283 = HEAP8[$1282>>0]|0;
     $1284 = HEAP32[$39>>2]|0;
     $1285 = $1284 & $1261;
     $1286 = $1264 >>> 16;
     $1287 = $1286 & 255;
     $1288 = $1285 >>> $1287;
     $1289 = $1267 >>> 16;
     $1290 = $1289 & 255;
     $1291 = (125924 + ($1290<<2)|0);
     $1292 = HEAP32[$1291>>2]|0;
     $1293 = (($1292) + ($1288)|0);
     $1294 = HEAP8[$1293>>0]|0;
     $sB41$0$shrunk = $1294;$sG40$0$shrunk = $1283;$sR39$0$shrunk = $1272;
     break;
    }
    case 2:  {
     $1295 = HEAP16[$src$7>>1]|0;
     $1296 = $1295&65535;
     $1297 = HEAP32[$35>>2]|0;
     $1298 = $1296 & $1297;
     $1299 = HEAP32[$36>>2]|0;
     $1300 = $1299 & 255;
     $1301 = $1298 >>> $1300;
     $1302 = HEAP32[$37>>2]|0;
     $1303 = $1302 & 255;
     $1304 = (125924 + ($1303<<2)|0);
     $1305 = HEAP32[$1304>>2]|0;
     $1306 = (($1305) + ($1301)|0);
     $1307 = HEAP8[$1306>>0]|0;
     $1308 = HEAP32[$38>>2]|0;
     $1309 = $1308 & $1296;
     $1310 = $1299 >>> 8;
     $1311 = $1310 & 255;
     $1312 = $1309 >>> $1311;
     $1313 = $1302 >>> 8;
     $1314 = $1313 & 255;
     $1315 = (125924 + ($1314<<2)|0);
     $1316 = HEAP32[$1315>>2]|0;
     $1317 = (($1316) + ($1312)|0);
     $1318 = HEAP8[$1317>>0]|0;
     $1319 = HEAP32[$39>>2]|0;
     $1320 = $1319 & $1296;
     $1321 = $1299 >>> 16;
     $1322 = $1321 & 255;
     $1323 = $1320 >>> $1322;
     $1324 = $1302 >>> 16;
     $1325 = $1324 & 255;
     $1326 = (125924 + ($1325<<2)|0);
     $1327 = HEAP32[$1326>>2]|0;
     $1328 = (($1327) + ($1323)|0);
     $1329 = HEAP8[$1328>>0]|0;
     $sB41$0$shrunk = $1329;$sG40$0$shrunk = $1318;$sR39$0$shrunk = $1307;
     break;
    }
    case 3:  {
     $1330 = HEAP32[$36>>2]|0;
     $div3 = $1330 >>> 3;
     $1331 = $div3 & 31;
     $1332 = (($src$7) + ($1331)|0);
     $1333 = HEAP8[$1332>>0]|0;
     $1334 = $1330 >>> 11;
     $div4 = $1334 & 31;
     $1335 = (($src$7) + ($div4)|0);
     $1336 = HEAP8[$1335>>0]|0;
     $1337 = $1330 >>> 19;
     $1338 = $1337 & 31;
     $1339 = (($src$7) + ($1338)|0);
     $1340 = HEAP8[$1339>>0]|0;
     $sB41$0$shrunk = $1340;$sG40$0$shrunk = $1336;$sR39$0$shrunk = $1333;
     break;
    }
    case 4:  {
     $1341 = HEAP32[$src$7>>2]|0;
     $1342 = HEAP32[$35>>2]|0;
     $1343 = $1342 & $1341;
     $1344 = HEAP32[$36>>2]|0;
     $1345 = $1344 & 255;
     $1346 = $1343 >>> $1345;
     $1347 = HEAP32[$37>>2]|0;
     $1348 = $1347 & 255;
     $1349 = (125924 + ($1348<<2)|0);
     $1350 = HEAP32[$1349>>2]|0;
     $1351 = (($1350) + ($1346)|0);
     $1352 = HEAP8[$1351>>0]|0;
     $1353 = HEAP32[$38>>2]|0;
     $1354 = $1353 & $1341;
     $1355 = $1344 >>> 8;
     $1356 = $1355 & 255;
     $1357 = $1354 >>> $1356;
     $1358 = $1347 >>> 8;
     $1359 = $1358 & 255;
     $1360 = (125924 + ($1359<<2)|0);
     $1361 = HEAP32[$1360>>2]|0;
     $1362 = (($1361) + ($1357)|0);
     $1363 = HEAP8[$1362>>0]|0;
     $1364 = HEAP32[$39>>2]|0;
     $1365 = $1364 & $1341;
     $1366 = $1344 >>> 16;
     $1367 = $1366 & 255;
     $1368 = $1365 >>> $1367;
     $1369 = $1347 >>> 16;
     $1370 = $1369 & 255;
     $1371 = (125924 + ($1370<<2)|0);
     $1372 = HEAP32[$1371>>2]|0;
     $1373 = (($1372) + ($1368)|0);
     $1374 = HEAP8[$1373>>0]|0;
     $sB41$0$shrunk = $1374;$sG40$0$shrunk = $1363;$sR39$0$shrunk = $1352;
     break;
    }
    default: {
     $sB41$0$shrunk = 0;$sG40$0$shrunk = 0;$sR39$0$shrunk = 0;
    }
    }
    $sB41$0 = $sB41$0$shrunk&255;
    $sG40$0 = $sG40$0$shrunk&255;
    $sR39$0 = $sR39$0$shrunk&255;
    switch ($20|0) {
    case 1:  {
     $1375 = HEAP32[$31>>2]|0;
     $1376 = $1375 & 255;
     $1377 = $sR39$0 >>> $1376;
     $1378 = HEAP32[$32>>2]|0;
     $1379 = $1378 & 255;
     $1380 = $1377 << $1379;
     $1381 = $1375 >>> 8;
     $1382 = $1381 & 255;
     $1383 = $sG40$0 >>> $1382;
     $1384 = $1378 >>> 8;
     $1385 = $1384 & 255;
     $1386 = $1383 << $1385;
     $1387 = $1386 | $1380;
     $1388 = $1375 >>> 16;
     $1389 = $1388 & 255;
     $1390 = $sB41$0 >>> $1389;
     $1391 = $1378 >>> 16;
     $1392 = $1391 & 255;
     $1393 = $1390 << $1392;
     $1394 = $1387 | $1393;
     $1395 = $1375 >>> 24;
     $1396 = $178 >>> $1395;
     $1397 = $1378 >>> 24;
     $1398 = $1396 << $1397;
     $1399 = $1394 | $1398;
     $1400 = $1399&255;
     HEAP8[$dst$7>>0] = $1400;
     break;
    }
    case 2:  {
     $1401 = HEAP32[$31>>2]|0;
     $1402 = $1401 & 255;
     $1403 = $sR39$0 >>> $1402;
     $1404 = HEAP32[$32>>2]|0;
     $1405 = $1404 & 255;
     $1406 = $1403 << $1405;
     $1407 = $1401 >>> 8;
     $1408 = $1407 & 255;
     $1409 = $sG40$0 >>> $1408;
     $1410 = $1404 >>> 8;
     $1411 = $1410 & 255;
     $1412 = $1409 << $1411;
     $1413 = $1412 | $1406;
     $1414 = $1401 >>> 16;
     $1415 = $1414 & 255;
     $1416 = $sB41$0 >>> $1415;
     $1417 = $1404 >>> 16;
     $1418 = $1417 & 255;
     $1419 = $1416 << $1418;
     $1420 = $1413 | $1419;
     $1421 = $1401 >>> 24;
     $1422 = $178 >>> $1421;
     $1423 = $1404 >>> 24;
     $1424 = $1422 << $1423;
     $1425 = $1420 | $1424;
     $1426 = $1425&65535;
     HEAP16[$dst$7>>1] = $1426;
     break;
    }
    case 3:  {
     $1427 = HEAP8[$32>>0]|0;
     $div = ($1427&255) >>> 3;
     $1428 = $div&255;
     $1429 = (($dst$7) + ($1428)|0);
     HEAP8[$1429>>0] = $sR39$0$shrunk;
     $1430 = HEAP8[$33>>0]|0;
     $div1 = ($1430&255) >>> 3;
     $1431 = $div1&255;
     $1432 = (($dst$7) + ($1431)|0);
     HEAP8[$1432>>0] = $sG40$0$shrunk;
     $1433 = HEAP8[$34>>0]|0;
     $div2 = ($1433&255) >>> 3;
     $1434 = $div2&255;
     $1435 = (($dst$7) + ($1434)|0);
     HEAP8[$1435>>0] = $sB41$0$shrunk;
     break;
    }
    case 4:  {
     $1436 = HEAP32[$31>>2]|0;
     $1437 = $1436 & 255;
     $1438 = $sR39$0 >>> $1437;
     $1439 = HEAP32[$32>>2]|0;
     $1440 = $1439 & 255;
     $1441 = $1438 << $1440;
     $1442 = $1436 >>> 8;
     $1443 = $1442 & 255;
     $1444 = $sG40$0 >>> $1443;
     $1445 = $1439 >>> 8;
     $1446 = $1445 & 255;
     $1447 = $1444 << $1446;
     $1448 = $1447 | $1441;
     $1449 = $1436 >>> 16;
     $1450 = $1449 & 255;
     $1451 = $sB41$0 >>> $1450;
     $1452 = $1439 >>> 16;
     $1453 = $1452 & 255;
     $1454 = $1451 << $1453;
     $1455 = $1448 | $1454;
     $1456 = $1436 >>> 24;
     $1457 = $178 >>> $1456;
     $1458 = $1439 >>> 24;
     $1459 = $1457 << $1458;
     $1460 = $1455 | $1459;
     HEAP32[$dst$7>>2] = $1460;
     break;
    }
    default: {
    }
    }
    $1461 = (($dst$7) + ($20)|0);
    $1462 = (($src$7) + ($15)|0);
    $dst$8 = $1461;$n$7 = $n$6;$src$8 = $1462;
    label = 83;
    continue;
   }
   else if ((label|0) == 83) {
    label = 0;
    switch ($15|0) {
    case 1:  {
     $1463 = HEAP8[$src$8>>0]|0;
     $1464 = $1463&255;
     $1465 = HEAP32[$35>>2]|0;
     $1466 = $1464 & $1465;
     $1467 = HEAP32[$36>>2]|0;
     $1468 = $1467 & 255;
     $1469 = $1466 >>> $1468;
     $1470 = HEAP32[$37>>2]|0;
     $1471 = $1470 & 255;
     $1472 = (125924 + ($1471<<2)|0);
     $1473 = HEAP32[$1472>>2]|0;
     $1474 = (($1473) + ($1469)|0);
     $1475 = HEAP8[$1474>>0]|0;
     $1476 = HEAP32[$38>>2]|0;
     $1477 = $1476 & $1464;
     $1478 = $1467 >>> 8;
     $1479 = $1478 & 255;
     $1480 = $1477 >>> $1479;
     $1481 = $1470 >>> 8;
     $1482 = $1481 & 255;
     $1483 = (125924 + ($1482<<2)|0);
     $1484 = HEAP32[$1483>>2]|0;
     $1485 = (($1484) + ($1480)|0);
     $1486 = HEAP8[$1485>>0]|0;
     $1487 = HEAP32[$39>>2]|0;
     $1488 = $1487 & $1464;
     $1489 = $1467 >>> 16;
     $1490 = $1489 & 255;
     $1491 = $1488 >>> $1490;
     $1492 = $1470 >>> 16;
     $1493 = $1492 & 255;
     $1494 = (125924 + ($1493<<2)|0);
     $1495 = HEAP32[$1494>>2]|0;
     $1496 = (($1495) + ($1491)|0);
     $1497 = HEAP8[$1496>>0]|0;
     $sB48$0$shrunk = $1497;$sG47$0$shrunk = $1486;$sR46$0$shrunk = $1475;
     break;
    }
    case 2:  {
     $1498 = HEAP16[$src$8>>1]|0;
     $1499 = $1498&65535;
     $1500 = HEAP32[$35>>2]|0;
     $1501 = $1499 & $1500;
     $1502 = HEAP32[$36>>2]|0;
     $1503 = $1502 & 255;
     $1504 = $1501 >>> $1503;
     $1505 = HEAP32[$37>>2]|0;
     $1506 = $1505 & 255;
     $1507 = (125924 + ($1506<<2)|0);
     $1508 = HEAP32[$1507>>2]|0;
     $1509 = (($1508) + ($1504)|0);
     $1510 = HEAP8[$1509>>0]|0;
     $1511 = HEAP32[$38>>2]|0;
     $1512 = $1511 & $1499;
     $1513 = $1502 >>> 8;
     $1514 = $1513 & 255;
     $1515 = $1512 >>> $1514;
     $1516 = $1505 >>> 8;
     $1517 = $1516 & 255;
     $1518 = (125924 + ($1517<<2)|0);
     $1519 = HEAP32[$1518>>2]|0;
     $1520 = (($1519) + ($1515)|0);
     $1521 = HEAP8[$1520>>0]|0;
     $1522 = HEAP32[$39>>2]|0;
     $1523 = $1522 & $1499;
     $1524 = $1502 >>> 16;
     $1525 = $1524 & 255;
     $1526 = $1523 >>> $1525;
     $1527 = $1505 >>> 16;
     $1528 = $1527 & 255;
     $1529 = (125924 + ($1528<<2)|0);
     $1530 = HEAP32[$1529>>2]|0;
     $1531 = (($1530) + ($1526)|0);
     $1532 = HEAP8[$1531>>0]|0;
     $sB48$0$shrunk = $1532;$sG47$0$shrunk = $1521;$sR46$0$shrunk = $1510;
     break;
    }
    case 3:  {
     $1533 = HEAP32[$36>>2]|0;
     $div45 = $1533 >>> 3;
     $1534 = $div45 & 31;
     $1535 = (($src$8) + ($1534)|0);
     $1536 = HEAP8[$1535>>0]|0;
     $1537 = $1533 >>> 11;
     $div46 = $1537 & 31;
     $1538 = (($src$8) + ($div46)|0);
     $1539 = HEAP8[$1538>>0]|0;
     $1540 = $1533 >>> 19;
     $1541 = $1540 & 31;
     $1542 = (($src$8) + ($1541)|0);
     $1543 = HEAP8[$1542>>0]|0;
     $sB48$0$shrunk = $1543;$sG47$0$shrunk = $1539;$sR46$0$shrunk = $1536;
     break;
    }
    case 4:  {
     $1544 = HEAP32[$src$8>>2]|0;
     $1545 = HEAP32[$35>>2]|0;
     $1546 = $1545 & $1544;
     $1547 = HEAP32[$36>>2]|0;
     $1548 = $1547 & 255;
     $1549 = $1546 >>> $1548;
     $1550 = HEAP32[$37>>2]|0;
     $1551 = $1550 & 255;
     $1552 = (125924 + ($1551<<2)|0);
     $1553 = HEAP32[$1552>>2]|0;
     $1554 = (($1553) + ($1549)|0);
     $1555 = HEAP8[$1554>>0]|0;
     $1556 = HEAP32[$38>>2]|0;
     $1557 = $1556 & $1544;
     $1558 = $1547 >>> 8;
     $1559 = $1558 & 255;
     $1560 = $1557 >>> $1559;
     $1561 = $1550 >>> 8;
     $1562 = $1561 & 255;
     $1563 = (125924 + ($1562<<2)|0);
     $1564 = HEAP32[$1563>>2]|0;
     $1565 = (($1564) + ($1560)|0);
     $1566 = HEAP8[$1565>>0]|0;
     $1567 = HEAP32[$39>>2]|0;
     $1568 = $1567 & $1544;
     $1569 = $1547 >>> 16;
     $1570 = $1569 & 255;
     $1571 = $1568 >>> $1570;
     $1572 = $1550 >>> 16;
     $1573 = $1572 & 255;
     $1574 = (125924 + ($1573<<2)|0);
     $1575 = HEAP32[$1574>>2]|0;
     $1576 = (($1575) + ($1571)|0);
     $1577 = HEAP8[$1576>>0]|0;
     $sB48$0$shrunk = $1577;$sG47$0$shrunk = $1566;$sR46$0$shrunk = $1555;
     break;
    }
    default: {
     $sB48$0$shrunk = 0;$sG47$0$shrunk = 0;$sR46$0$shrunk = 0;
    }
    }
    $sB48$0 = $sB48$0$shrunk&255;
    $sG47$0 = $sG47$0$shrunk&255;
    $sR46$0 = $sR46$0$shrunk&255;
    switch ($20|0) {
    case 1:  {
     $1578 = HEAP32[$31>>2]|0;
     $1579 = $1578 & 255;
     $1580 = $sR46$0 >>> $1579;
     $1581 = HEAP32[$32>>2]|0;
     $1582 = $1581 & 255;
     $1583 = $1580 << $1582;
     $1584 = $1578 >>> 8;
     $1585 = $1584 & 255;
     $1586 = $sG47$0 >>> $1585;
     $1587 = $1581 >>> 8;
     $1588 = $1587 & 255;
     $1589 = $1586 << $1588;
     $1590 = $1589 | $1583;
     $1591 = $1578 >>> 16;
     $1592 = $1591 & 255;
     $1593 = $sB48$0 >>> $1592;
     $1594 = $1581 >>> 16;
     $1595 = $1594 & 255;
     $1596 = $1593 << $1595;
     $1597 = $1590 | $1596;
     $1598 = $1578 >>> 24;
     $1599 = $178 >>> $1598;
     $1600 = $1581 >>> 24;
     $1601 = $1599 << $1600;
     $1602 = $1597 | $1601;
     $1603 = $1602&255;
     HEAP8[$dst$8>>0] = $1603;
     break;
    }
    case 2:  {
     $1604 = HEAP32[$31>>2]|0;
     $1605 = $1604 & 255;
     $1606 = $sR46$0 >>> $1605;
     $1607 = HEAP32[$32>>2]|0;
     $1608 = $1607 & 255;
     $1609 = $1606 << $1608;
     $1610 = $1604 >>> 8;
     $1611 = $1610 & 255;
     $1612 = $sG47$0 >>> $1611;
     $1613 = $1607 >>> 8;
     $1614 = $1613 & 255;
     $1615 = $1612 << $1614;
     $1616 = $1615 | $1609;
     $1617 = $1604 >>> 16;
     $1618 = $1617 & 255;
     $1619 = $sB48$0 >>> $1618;
     $1620 = $1607 >>> 16;
     $1621 = $1620 & 255;
     $1622 = $1619 << $1621;
     $1623 = $1616 | $1622;
     $1624 = $1604 >>> 24;
     $1625 = $178 >>> $1624;
     $1626 = $1607 >>> 24;
     $1627 = $1625 << $1626;
     $1628 = $1623 | $1627;
     $1629 = $1628&65535;
     HEAP16[$dst$8>>1] = $1629;
     break;
    }
    case 3:  {
     $1630 = HEAP8[$32>>0]|0;
     $div42 = ($1630&255) >>> 3;
     $1631 = $div42&255;
     $1632 = (($dst$8) + ($1631)|0);
     HEAP8[$1632>>0] = $sR46$0$shrunk;
     $1633 = HEAP8[$33>>0]|0;
     $div43 = ($1633&255) >>> 3;
     $1634 = $div43&255;
     $1635 = (($dst$8) + ($1634)|0);
     HEAP8[$1635>>0] = $sG47$0$shrunk;
     $1636 = HEAP8[$34>>0]|0;
     $div44 = ($1636&255) >>> 3;
     $1637 = $div44&255;
     $1638 = (($dst$8) + ($1637)|0);
     HEAP8[$1638>>0] = $sB48$0$shrunk;
     break;
    }
    case 4:  {
     $1639 = HEAP32[$31>>2]|0;
     $1640 = $1639 & 255;
     $1641 = $sR46$0 >>> $1640;
     $1642 = HEAP32[$32>>2]|0;
     $1643 = $1642 & 255;
     $1644 = $1641 << $1643;
     $1645 = $1639 >>> 8;
     $1646 = $1645 & 255;
     $1647 = $sG47$0 >>> $1646;
     $1648 = $1642 >>> 8;
     $1649 = $1648 & 255;
     $1650 = $1647 << $1649;
     $1651 = $1650 | $1644;
     $1652 = $1639 >>> 16;
     $1653 = $1652 & 255;
     $1654 = $sB48$0 >>> $1653;
     $1655 = $1642 >>> 16;
     $1656 = $1655 & 255;
     $1657 = $1654 << $1656;
     $1658 = $1651 | $1657;
     $1659 = $1639 >>> 24;
     $1660 = $178 >>> $1659;
     $1661 = $1642 >>> 24;
     $1662 = $1660 << $1661;
     $1663 = $1658 | $1662;
     HEAP32[$dst$8>>2] = $1663;
     break;
    }
    default: {
    }
    }
    $1664 = (($dst$8) + ($20)|0);
    $1665 = (($src$8) + ($15)|0);
    $1666 = (($n$7) + -1)|0;
    $1667 = ($n$7|0)>(1);
    if ($1667) {
     $dst$1 = $1664;$n$0 = $1666;$src$1 = $1665;
     label = 6;
     continue;
    } else {
     break;
    }
   }
  }
  $1668 = (($1665) + ($6)|0);
  $1669 = (($1664) + ($10)|0);
  $1670 = ($40|0)==(0);
  if ($1670) {
   label = 96;
   break;
  } else {
   $$in = $40;$dst$048 = $1669;$src$049 = $1668;
  }
 }
 if ((label|0) == 94) {
  // unreachable;
 }
 else if ((label|0) == 96) {
  return;
 }
}
function _Blit_RGB565_ARGB8888($info) {
 $info = $info|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _Blit_RGB565_32($info,126936);
 return;
}
function _Blit_RGB565_32($info,$map) {
 $info = $info|0;
 $map = $map|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $dst$01 = 0, $dst$1 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0;
 var $src$02 = 0, $src$1 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (($7|0) / 4)&-1;
 $9 = ($3|0)==(0);
 if ($9) {
  return;
 }
 $10 = ((($info)) + 20|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = HEAP32[$info>>2]|0;
 $13 = (($1) + 7)|0;
 $14 = (($13|0) / 8)&-1;
 $15 = $1 & 7;
 $$in = $3;$dst$01 = $11;$src$02 = $12;
 L4: while(1) {
  $16 = (($$in) + -1)|0;
  switch ($15|0) {
  case 0:  {
   $dst$1 = $dst$01;$n$0 = $14;$src$1 = $src$02;
   label = 4;
   break;
  }
  case 7:  {
   $dst$2 = $dst$01;$n$1 = $14;$src$2 = $src$02;
   label = 5;
   break;
  }
  case 6:  {
   $dst$3 = $dst$01;$n$2 = $14;$src$3 = $src$02;
   label = 6;
   break;
  }
  case 5:  {
   $dst$4 = $dst$01;$n$3 = $14;$src$4 = $src$02;
   label = 7;
   break;
  }
  case 4:  {
   $dst$5 = $dst$01;$n$4 = $14;$src$5 = $src$02;
   label = 8;
   break;
  }
  case 3:  {
   $dst$6 = $dst$01;$n$5 = $14;$src$6 = $src$02;
   label = 9;
   break;
  }
  case 2:  {
   $dst$7 = $dst$01;$n$6 = $14;$src$7 = $src$02;
   label = 10;
   break;
  }
  case 1:  {
   $dst$8 = $dst$01;$n$7 = $14;$src$8 = $src$02;
   label = 11;
   break;
  }
  default: {
   label = 12;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    $17 = HEAP8[$src$1>>0]|0;
    $18 = $17&255;
    $19 = $18 << 1;
    $20 = (($map) + ($19<<2)|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = ((($src$1)) + 1|0);
    $23 = HEAP8[$22>>0]|0;
    $24 = $23&255;
    $25 = $24 << 1;
    $26 = $25 | 1;
    $27 = (($map) + ($26<<2)|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = (($28) + ($21))|0;
    $30 = ((($dst$1)) + 4|0);
    HEAP32[$dst$1>>2] = $29;
    $31 = ((($src$1)) + 2|0);
    $dst$2 = $30;$n$1 = $n$0;$src$2 = $31;
    label = 5;
    continue;
   }
   else if ((label|0) == 5) {
    label = 0;
    $32 = HEAP8[$src$2>>0]|0;
    $33 = $32&255;
    $34 = $33 << 1;
    $35 = (($map) + ($34<<2)|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ((($src$2)) + 1|0);
    $38 = HEAP8[$37>>0]|0;
    $39 = $38&255;
    $40 = $39 << 1;
    $41 = $40 | 1;
    $42 = (($map) + ($41<<2)|0);
    $43 = HEAP32[$42>>2]|0;
    $44 = (($43) + ($36))|0;
    $45 = ((($dst$2)) + 4|0);
    HEAP32[$dst$2>>2] = $44;
    $46 = ((($src$2)) + 2|0);
    $dst$3 = $45;$n$2 = $n$1;$src$3 = $46;
    label = 6;
    continue;
   }
   else if ((label|0) == 6) {
    label = 0;
    $47 = HEAP8[$src$3>>0]|0;
    $48 = $47&255;
    $49 = $48 << 1;
    $50 = (($map) + ($49<<2)|0);
    $51 = HEAP32[$50>>2]|0;
    $52 = ((($src$3)) + 1|0);
    $53 = HEAP8[$52>>0]|0;
    $54 = $53&255;
    $55 = $54 << 1;
    $56 = $55 | 1;
    $57 = (($map) + ($56<<2)|0);
    $58 = HEAP32[$57>>2]|0;
    $59 = (($58) + ($51))|0;
    $60 = ((($dst$3)) + 4|0);
    HEAP32[$dst$3>>2] = $59;
    $61 = ((($src$3)) + 2|0);
    $dst$4 = $60;$n$3 = $n$2;$src$4 = $61;
    label = 7;
    continue;
   }
   else if ((label|0) == 7) {
    label = 0;
    $62 = HEAP8[$src$4>>0]|0;
    $63 = $62&255;
    $64 = $63 << 1;
    $65 = (($map) + ($64<<2)|0);
    $66 = HEAP32[$65>>2]|0;
    $67 = ((($src$4)) + 1|0);
    $68 = HEAP8[$67>>0]|0;
    $69 = $68&255;
    $70 = $69 << 1;
    $71 = $70 | 1;
    $72 = (($map) + ($71<<2)|0);
    $73 = HEAP32[$72>>2]|0;
    $74 = (($73) + ($66))|0;
    $75 = ((($dst$4)) + 4|0);
    HEAP32[$dst$4>>2] = $74;
    $76 = ((($src$4)) + 2|0);
    $dst$5 = $75;$n$4 = $n$3;$src$5 = $76;
    label = 8;
    continue;
   }
   else if ((label|0) == 8) {
    label = 0;
    $77 = HEAP8[$src$5>>0]|0;
    $78 = $77&255;
    $79 = $78 << 1;
    $80 = (($map) + ($79<<2)|0);
    $81 = HEAP32[$80>>2]|0;
    $82 = ((($src$5)) + 1|0);
    $83 = HEAP8[$82>>0]|0;
    $84 = $83&255;
    $85 = $84 << 1;
    $86 = $85 | 1;
    $87 = (($map) + ($86<<2)|0);
    $88 = HEAP32[$87>>2]|0;
    $89 = (($88) + ($81))|0;
    $90 = ((($dst$5)) + 4|0);
    HEAP32[$dst$5>>2] = $89;
    $91 = ((($src$5)) + 2|0);
    $dst$6 = $90;$n$5 = $n$4;$src$6 = $91;
    label = 9;
    continue;
   }
   else if ((label|0) == 9) {
    label = 0;
    $92 = HEAP8[$src$6>>0]|0;
    $93 = $92&255;
    $94 = $93 << 1;
    $95 = (($map) + ($94<<2)|0);
    $96 = HEAP32[$95>>2]|0;
    $97 = ((($src$6)) + 1|0);
    $98 = HEAP8[$97>>0]|0;
    $99 = $98&255;
    $100 = $99 << 1;
    $101 = $100 | 1;
    $102 = (($map) + ($101<<2)|0);
    $103 = HEAP32[$102>>2]|0;
    $104 = (($103) + ($96))|0;
    $105 = ((($dst$6)) + 4|0);
    HEAP32[$dst$6>>2] = $104;
    $106 = ((($src$6)) + 2|0);
    $dst$7 = $105;$n$6 = $n$5;$src$7 = $106;
    label = 10;
    continue;
   }
   else if ((label|0) == 10) {
    label = 0;
    $107 = HEAP8[$src$7>>0]|0;
    $108 = $107&255;
    $109 = $108 << 1;
    $110 = (($map) + ($109<<2)|0);
    $111 = HEAP32[$110>>2]|0;
    $112 = ((($src$7)) + 1|0);
    $113 = HEAP8[$112>>0]|0;
    $114 = $113&255;
    $115 = $114 << 1;
    $116 = $115 | 1;
    $117 = (($map) + ($116<<2)|0);
    $118 = HEAP32[$117>>2]|0;
    $119 = (($118) + ($111))|0;
    $120 = ((($dst$7)) + 4|0);
    HEAP32[$dst$7>>2] = $119;
    $121 = ((($src$7)) + 2|0);
    $dst$8 = $120;$n$7 = $n$6;$src$8 = $121;
    label = 11;
    continue;
   }
   else if ((label|0) == 11) {
    label = 0;
    $122 = HEAP8[$src$8>>0]|0;
    $123 = $122&255;
    $124 = $123 << 1;
    $125 = (($map) + ($124<<2)|0);
    $126 = HEAP32[$125>>2]|0;
    $127 = ((($src$8)) + 1|0);
    $128 = HEAP8[$127>>0]|0;
    $129 = $128&255;
    $130 = $129 << 1;
    $131 = $130 | 1;
    $132 = (($map) + ($131<<2)|0);
    $133 = HEAP32[$132>>2]|0;
    $134 = (($133) + ($126))|0;
    $135 = ((($dst$8)) + 4|0);
    HEAP32[$dst$8>>2] = $134;
    $136 = ((($src$8)) + 2|0);
    $137 = (($n$7) + -1)|0;
    $138 = ($n$7|0)>(1);
    if ($138) {
     $dst$1 = $135;$n$0 = $137;$src$1 = $136;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $139 = (($136) + ($5)|0);
  $140 = (($135) + ($8<<2)|0);
  $141 = ($16|0)==(0);
  if ($141) {
   label = 14;
   break;
  } else {
   $$in = $16;$dst$01 = $140;$src$02 = $139;
  }
 }
 if ((label|0) == 12) {
  // unreachable;
 }
 else if ((label|0) == 14) {
  return;
 }
}
function _Blit_RGB565_ABGR8888($info) {
 $info = $info|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _Blit_RGB565_32($info,128984);
 return;
}
function _Blit_RGB565_RGBA8888($info) {
 $info = $info|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _Blit_RGB565_32($info,131032);
 return;
}
function _Blit_RGB565_BGRA8888($info) {
 $info = $info|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _Blit_RGB565_32($info,133080);
 return;
}
function _Blit_RGB888_RGB565($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $dst$01 = 0, $dst$1 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $src$02 = 0, $src$1 = 0, $src$2 = 0;
 var $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($5|0) / 4)&-1;
 $7 = ((($info)) + 36|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (($8|0) / 2)&-1;
 $10 = ($3|0)==(0);
 if ($10) {
  return;
 }
 $11 = ((($info)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = HEAP32[$info>>2]|0;
 $14 = (($1) + 7)|0;
 $15 = (($14|0) / 8)&-1;
 $16 = $1 & 7;
 $$in = $3;$dst$01 = $12;$src$02 = $13;
 L4: while(1) {
  $17 = (($$in) + -1)|0;
  switch ($16|0) {
  case 0:  {
   $dst$1 = $dst$01;$n$0 = $15;$src$1 = $src$02;
   label = 4;
   break;
  }
  case 7:  {
   $dst$2 = $dst$01;$n$1 = $15;$src$2 = $src$02;
   label = 5;
   break;
  }
  case 6:  {
   $dst$3 = $dst$01;$n$2 = $15;$src$3 = $src$02;
   label = 6;
   break;
  }
  case 5:  {
   $dst$4 = $dst$01;$n$3 = $15;$src$4 = $src$02;
   label = 7;
   break;
  }
  case 4:  {
   $dst$5 = $dst$01;$n$4 = $15;$src$5 = $src$02;
   label = 8;
   break;
  }
  case 3:  {
   $dst$6 = $dst$01;$n$5 = $15;$src$6 = $src$02;
   label = 9;
   break;
  }
  case 2:  {
   $dst$7 = $dst$01;$n$6 = $15;$src$7 = $src$02;
   label = 10;
   break;
  }
  case 1:  {
   $dst$8 = $dst$01;$n$7 = $15;$src$8 = $src$02;
   label = 11;
   break;
  }
  default: {
   label = 12;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    $18 = HEAP32[$src$1>>2]|0;
    $19 = $18 >>> 8;
    $20 = $19 & 63488;
    $21 = $18 >>> 5;
    $22 = $21 & 2016;
    $23 = $20 | $22;
    $24 = $18 >>> 3;
    $25 = $24 & 31;
    $26 = $23 | $25;
    $27 = $26&65535;
    HEAP16[$dst$1>>1] = $27;
    $28 = ((($src$1)) + 4|0);
    $29 = ((($dst$1)) + 2|0);
    $dst$2 = $29;$n$1 = $n$0;$src$2 = $28;
    label = 5;
    continue;
   }
   else if ((label|0) == 5) {
    label = 0;
    $30 = HEAP32[$src$2>>2]|0;
    $31 = $30 >>> 8;
    $32 = $31 & 63488;
    $33 = $30 >>> 5;
    $34 = $33 & 2016;
    $35 = $32 | $34;
    $36 = $30 >>> 3;
    $37 = $36 & 31;
    $38 = $35 | $37;
    $39 = $38&65535;
    HEAP16[$dst$2>>1] = $39;
    $40 = ((($src$2)) + 4|0);
    $41 = ((($dst$2)) + 2|0);
    $dst$3 = $41;$n$2 = $n$1;$src$3 = $40;
    label = 6;
    continue;
   }
   else if ((label|0) == 6) {
    label = 0;
    $42 = HEAP32[$src$3>>2]|0;
    $43 = $42 >>> 8;
    $44 = $43 & 63488;
    $45 = $42 >>> 5;
    $46 = $45 & 2016;
    $47 = $44 | $46;
    $48 = $42 >>> 3;
    $49 = $48 & 31;
    $50 = $47 | $49;
    $51 = $50&65535;
    HEAP16[$dst$3>>1] = $51;
    $52 = ((($src$3)) + 4|0);
    $53 = ((($dst$3)) + 2|0);
    $dst$4 = $53;$n$3 = $n$2;$src$4 = $52;
    label = 7;
    continue;
   }
   else if ((label|0) == 7) {
    label = 0;
    $54 = HEAP32[$src$4>>2]|0;
    $55 = $54 >>> 8;
    $56 = $55 & 63488;
    $57 = $54 >>> 5;
    $58 = $57 & 2016;
    $59 = $56 | $58;
    $60 = $54 >>> 3;
    $61 = $60 & 31;
    $62 = $59 | $61;
    $63 = $62&65535;
    HEAP16[$dst$4>>1] = $63;
    $64 = ((($src$4)) + 4|0);
    $65 = ((($dst$4)) + 2|0);
    $dst$5 = $65;$n$4 = $n$3;$src$5 = $64;
    label = 8;
    continue;
   }
   else if ((label|0) == 8) {
    label = 0;
    $66 = HEAP32[$src$5>>2]|0;
    $67 = $66 >>> 8;
    $68 = $67 & 63488;
    $69 = $66 >>> 5;
    $70 = $69 & 2016;
    $71 = $68 | $70;
    $72 = $66 >>> 3;
    $73 = $72 & 31;
    $74 = $71 | $73;
    $75 = $74&65535;
    HEAP16[$dst$5>>1] = $75;
    $76 = ((($src$5)) + 4|0);
    $77 = ((($dst$5)) + 2|0);
    $dst$6 = $77;$n$5 = $n$4;$src$6 = $76;
    label = 9;
    continue;
   }
   else if ((label|0) == 9) {
    label = 0;
    $78 = HEAP32[$src$6>>2]|0;
    $79 = $78 >>> 8;
    $80 = $79 & 63488;
    $81 = $78 >>> 5;
    $82 = $81 & 2016;
    $83 = $80 | $82;
    $84 = $78 >>> 3;
    $85 = $84 & 31;
    $86 = $83 | $85;
    $87 = $86&65535;
    HEAP16[$dst$6>>1] = $87;
    $88 = ((($src$6)) + 4|0);
    $89 = ((($dst$6)) + 2|0);
    $dst$7 = $89;$n$6 = $n$5;$src$7 = $88;
    label = 10;
    continue;
   }
   else if ((label|0) == 10) {
    label = 0;
    $90 = HEAP32[$src$7>>2]|0;
    $91 = $90 >>> 8;
    $92 = $91 & 63488;
    $93 = $90 >>> 5;
    $94 = $93 & 2016;
    $95 = $92 | $94;
    $96 = $90 >>> 3;
    $97 = $96 & 31;
    $98 = $95 | $97;
    $99 = $98&65535;
    HEAP16[$dst$7>>1] = $99;
    $100 = ((($src$7)) + 4|0);
    $101 = ((($dst$7)) + 2|0);
    $dst$8 = $101;$n$7 = $n$6;$src$8 = $100;
    label = 11;
    continue;
   }
   else if ((label|0) == 11) {
    label = 0;
    $102 = HEAP32[$src$8>>2]|0;
    $103 = $102 >>> 8;
    $104 = $103 & 63488;
    $105 = $102 >>> 5;
    $106 = $105 & 2016;
    $107 = $104 | $106;
    $108 = $102 >>> 3;
    $109 = $108 & 31;
    $110 = $107 | $109;
    $111 = $110&65535;
    HEAP16[$dst$8>>1] = $111;
    $112 = ((($src$8)) + 4|0);
    $113 = ((($dst$8)) + 2|0);
    $114 = (($n$7) + -1)|0;
    $115 = ($n$7|0)>(1);
    if ($115) {
     $dst$1 = $113;$n$0 = $114;$src$1 = $112;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $116 = (($112) + ($6<<2)|0);
  $117 = (($113) + ($9<<1)|0);
  $118 = ($17|0)==(0);
  if ($118) {
   label = 14;
   break;
  } else {
   $$in = $17;$dst$01 = $117;$src$02 = $116;
  }
 }
 if ((label|0) == 12) {
  // unreachable;
 }
 else if ((label|0) == 14) {
  return;
 }
}
function _Blit_RGB888_RGB555($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $dst$01 = 0, $dst$1 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $src$02 = 0, $src$1 = 0, $src$2 = 0;
 var $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($5|0) / 4)&-1;
 $7 = ((($info)) + 36|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (($8|0) / 2)&-1;
 $10 = ($3|0)==(0);
 if ($10) {
  return;
 }
 $11 = ((($info)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = HEAP32[$info>>2]|0;
 $14 = (($1) + 7)|0;
 $15 = (($14|0) / 8)&-1;
 $16 = $1 & 7;
 $$in = $3;$dst$01 = $12;$src$02 = $13;
 L4: while(1) {
  $17 = (($$in) + -1)|0;
  switch ($16|0) {
  case 0:  {
   $dst$1 = $dst$01;$n$0 = $15;$src$1 = $src$02;
   label = 4;
   break;
  }
  case 7:  {
   $dst$2 = $dst$01;$n$1 = $15;$src$2 = $src$02;
   label = 5;
   break;
  }
  case 6:  {
   $dst$3 = $dst$01;$n$2 = $15;$src$3 = $src$02;
   label = 6;
   break;
  }
  case 5:  {
   $dst$4 = $dst$01;$n$3 = $15;$src$4 = $src$02;
   label = 7;
   break;
  }
  case 4:  {
   $dst$5 = $dst$01;$n$4 = $15;$src$5 = $src$02;
   label = 8;
   break;
  }
  case 3:  {
   $dst$6 = $dst$01;$n$5 = $15;$src$6 = $src$02;
   label = 9;
   break;
  }
  case 2:  {
   $dst$7 = $dst$01;$n$6 = $15;$src$7 = $src$02;
   label = 10;
   break;
  }
  case 1:  {
   $dst$8 = $dst$01;$n$7 = $15;$src$8 = $src$02;
   label = 11;
   break;
  }
  default: {
   label = 12;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    $18 = HEAP32[$src$1>>2]|0;
    $19 = $18 >>> 9;
    $20 = $19 & 31744;
    $21 = $18 >>> 6;
    $22 = $21 & 992;
    $23 = $20 | $22;
    $24 = $18 >>> 3;
    $25 = $24 & 31;
    $26 = $23 | $25;
    $27 = $26&65535;
    HEAP16[$dst$1>>1] = $27;
    $28 = ((($src$1)) + 4|0);
    $29 = ((($dst$1)) + 2|0);
    $dst$2 = $29;$n$1 = $n$0;$src$2 = $28;
    label = 5;
    continue;
   }
   else if ((label|0) == 5) {
    label = 0;
    $30 = HEAP32[$src$2>>2]|0;
    $31 = $30 >>> 9;
    $32 = $31 & 31744;
    $33 = $30 >>> 6;
    $34 = $33 & 992;
    $35 = $32 | $34;
    $36 = $30 >>> 3;
    $37 = $36 & 31;
    $38 = $35 | $37;
    $39 = $38&65535;
    HEAP16[$dst$2>>1] = $39;
    $40 = ((($src$2)) + 4|0);
    $41 = ((($dst$2)) + 2|0);
    $dst$3 = $41;$n$2 = $n$1;$src$3 = $40;
    label = 6;
    continue;
   }
   else if ((label|0) == 6) {
    label = 0;
    $42 = HEAP32[$src$3>>2]|0;
    $43 = $42 >>> 9;
    $44 = $43 & 31744;
    $45 = $42 >>> 6;
    $46 = $45 & 992;
    $47 = $44 | $46;
    $48 = $42 >>> 3;
    $49 = $48 & 31;
    $50 = $47 | $49;
    $51 = $50&65535;
    HEAP16[$dst$3>>1] = $51;
    $52 = ((($src$3)) + 4|0);
    $53 = ((($dst$3)) + 2|0);
    $dst$4 = $53;$n$3 = $n$2;$src$4 = $52;
    label = 7;
    continue;
   }
   else if ((label|0) == 7) {
    label = 0;
    $54 = HEAP32[$src$4>>2]|0;
    $55 = $54 >>> 9;
    $56 = $55 & 31744;
    $57 = $54 >>> 6;
    $58 = $57 & 992;
    $59 = $56 | $58;
    $60 = $54 >>> 3;
    $61 = $60 & 31;
    $62 = $59 | $61;
    $63 = $62&65535;
    HEAP16[$dst$4>>1] = $63;
    $64 = ((($src$4)) + 4|0);
    $65 = ((($dst$4)) + 2|0);
    $dst$5 = $65;$n$4 = $n$3;$src$5 = $64;
    label = 8;
    continue;
   }
   else if ((label|0) == 8) {
    label = 0;
    $66 = HEAP32[$src$5>>2]|0;
    $67 = $66 >>> 9;
    $68 = $67 & 31744;
    $69 = $66 >>> 6;
    $70 = $69 & 992;
    $71 = $68 | $70;
    $72 = $66 >>> 3;
    $73 = $72 & 31;
    $74 = $71 | $73;
    $75 = $74&65535;
    HEAP16[$dst$5>>1] = $75;
    $76 = ((($src$5)) + 4|0);
    $77 = ((($dst$5)) + 2|0);
    $dst$6 = $77;$n$5 = $n$4;$src$6 = $76;
    label = 9;
    continue;
   }
   else if ((label|0) == 9) {
    label = 0;
    $78 = HEAP32[$src$6>>2]|0;
    $79 = $78 >>> 9;
    $80 = $79 & 31744;
    $81 = $78 >>> 6;
    $82 = $81 & 992;
    $83 = $80 | $82;
    $84 = $78 >>> 3;
    $85 = $84 & 31;
    $86 = $83 | $85;
    $87 = $86&65535;
    HEAP16[$dst$6>>1] = $87;
    $88 = ((($src$6)) + 4|0);
    $89 = ((($dst$6)) + 2|0);
    $dst$7 = $89;$n$6 = $n$5;$src$7 = $88;
    label = 10;
    continue;
   }
   else if ((label|0) == 10) {
    label = 0;
    $90 = HEAP32[$src$7>>2]|0;
    $91 = $90 >>> 9;
    $92 = $91 & 31744;
    $93 = $90 >>> 6;
    $94 = $93 & 992;
    $95 = $92 | $94;
    $96 = $90 >>> 3;
    $97 = $96 & 31;
    $98 = $95 | $97;
    $99 = $98&65535;
    HEAP16[$dst$7>>1] = $99;
    $100 = ((($src$7)) + 4|0);
    $101 = ((($dst$7)) + 2|0);
    $dst$8 = $101;$n$7 = $n$6;$src$8 = $100;
    label = 11;
    continue;
   }
   else if ((label|0) == 11) {
    label = 0;
    $102 = HEAP32[$src$8>>2]|0;
    $103 = $102 >>> 9;
    $104 = $103 & 31744;
    $105 = $102 >>> 6;
    $106 = $105 & 992;
    $107 = $104 | $106;
    $108 = $102 >>> 3;
    $109 = $108 & 31;
    $110 = $107 | $109;
    $111 = $110&65535;
    HEAP16[$dst$8>>1] = $111;
    $112 = ((($src$8)) + 4|0);
    $113 = ((($dst$8)) + 2|0);
    $114 = (($n$7) + -1)|0;
    $115 = ($n$7|0)>(1);
    if ($115) {
     $dst$1 = $113;$n$0 = $114;$src$1 = $112;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $116 = (($112) + ($6<<2)|0);
  $117 = (($113) + ($9<<1)|0);
  $118 = ($17|0)==(0);
  if ($118) {
   label = 14;
   break;
  } else {
   $$in = $17;$dst$01 = $117;$src$02 = $116;
  }
 }
 if ((label|0) == 12) {
  // unreachable;
 }
 else if ((label|0) == 14) {
  return;
 }
}
function _SDL_HasMMX() {
 var $$lobit = 0, $$pre$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $altivec$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $altivec$i$i = sp;
 $0 = HEAP32[33822]|0;
 $1 = ($0|0)==(-1);
 if ($1) {
  HEAP32[33822] = 0;
  HEAP32[$altivec$i$i>>2] = 0;
  $2 = HEAP32[$altivec$i$i>>2]|0;
  $3 = ($2|0)==(0);
  $$pre$i = HEAP32[33822]|0;
  if ($3) {
   $6 = $$pre$i;
  } else {
   $4 = $$pre$i | 2;
   HEAP32[33822] = $4;
   $6 = $4;
  }
 } else {
  $6 = $0;
 }
 $5 = $6 >>> 2;
 $$lobit = $5 & 1;
 STACKTOP = sp;return ($$lobit|0);
}
function _Blit4to4CopyAlpha($info) {
 $info = $info|0;
 var $$lcssa = 0, $$lcssa48 = 0, $$lcssa49 = 0, $$lcssa61 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$02 = 0, $dst$02$us = 0, $dst$02$us4 = 0;
 var $dst$1$us = 0, $dst$2$us1051 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $n$0$us = 0, $n$1$us1152 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $src$01 = 0, $src$01$us = 0, $src$01$us5 = 0, $src$1$us = 0;
 var $src$2$us950 = 0, $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$info>>2]|0;
 $5 = ((($info)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($info)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($info)) + 36|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($3) + -1)|0;
 $12 = ($3|0)==(0);
 if ($12) {
  return;
 }
 $13 = (($1) + 7)|0;
 $14 = (($13|0) / 8)&-1;
 $15 = $1 & 7;
 switch ($15|0) {
 case 0:  {
  $46 = $11;$dst$02$us = $8;$src$01$us = $4;
  while(1) {
   $dst$1$us = $dst$02$us;$n$0$us = $14;$src$1$us = $src$01$us;
   while(1) {
    $17 = HEAP32[$src$1$us>>2]|0;
    HEAP32[$dst$1$us>>2] = $17;
    $18 = ((($dst$1$us)) + 4|0);
    $19 = ((($src$1$us)) + 4|0);
    $20 = HEAP32[$19>>2]|0;
    HEAP32[$18>>2] = $20;
    $21 = ((($dst$1$us)) + 8|0);
    $22 = ((($src$1$us)) + 8|0);
    $23 = HEAP32[$22>>2]|0;
    HEAP32[$21>>2] = $23;
    $24 = ((($dst$1$us)) + 12|0);
    $25 = ((($src$1$us)) + 12|0);
    $26 = HEAP32[$25>>2]|0;
    HEAP32[$24>>2] = $26;
    $27 = ((($dst$1$us)) + 16|0);
    $28 = ((($src$1$us)) + 16|0);
    $29 = HEAP32[$28>>2]|0;
    HEAP32[$27>>2] = $29;
    $30 = ((($dst$1$us)) + 20|0);
    $31 = ((($src$1$us)) + 20|0);
    $32 = HEAP32[$31>>2]|0;
    HEAP32[$30>>2] = $32;
    $33 = ((($dst$1$us)) + 24|0);
    $34 = ((($src$1$us)) + 24|0);
    $35 = HEAP32[$34>>2]|0;
    HEAP32[$33>>2] = $35;
    $36 = ((($dst$1$us)) + 28|0);
    $37 = ((($src$1$us)) + 28|0);
    $38 = HEAP32[$37>>2]|0;
    HEAP32[$36>>2] = $38;
    $39 = ((($dst$1$us)) + 32|0);
    $40 = ((($src$1$us)) + 32|0);
    $41 = (($n$0$us) + -1)|0;
    $42 = ($n$0$us|0)>(1);
    if ($42) {
     $dst$1$us = $39;$n$0$us = $41;$src$1$us = $40;
    } else {
     $$lcssa = $39;$$lcssa61 = $40;
     break;
    }
   }
   $43 = (($$lcssa61) + ($6)|0);
   $44 = (($$lcssa) + ($10)|0);
   $45 = (($46) + -1)|0;
   $47 = ($46|0)==(0);
   if ($47) {
    break;
   } else {
    $46 = $45;$dst$02$us = $44;$src$01$us = $43;
   }
  }
  return;
  break;
 }
 case 7:  {
  $16 = ($13|0)>(15);
  $100 = $11;$dst$02$us4 = $8;$src$01$us5 = $4;
  while(1) {
   $48 = HEAP32[$src$01$us5>>2]|0;
   HEAP32[$dst$02$us4>>2] = $48;
   $49 = ((($dst$02$us4)) + 4|0);
   $50 = ((($src$01$us5)) + 4|0);
   $51 = HEAP32[$50>>2]|0;
   HEAP32[$49>>2] = $51;
   $52 = ((($dst$02$us4)) + 8|0);
   $53 = ((($src$01$us5)) + 8|0);
   $54 = HEAP32[$53>>2]|0;
   HEAP32[$52>>2] = $54;
   $55 = ((($dst$02$us4)) + 12|0);
   $56 = ((($src$01$us5)) + 12|0);
   $57 = HEAP32[$56>>2]|0;
   HEAP32[$55>>2] = $57;
   $58 = ((($dst$02$us4)) + 16|0);
   $59 = ((($src$01$us5)) + 16|0);
   $60 = HEAP32[$59>>2]|0;
   HEAP32[$58>>2] = $60;
   $61 = ((($dst$02$us4)) + 20|0);
   $62 = ((($src$01$us5)) + 20|0);
   $63 = HEAP32[$62>>2]|0;
   HEAP32[$61>>2] = $63;
   $64 = ((($dst$02$us4)) + 24|0);
   $65 = ((($src$01$us5)) + 24|0);
   $66 = HEAP32[$65>>2]|0;
   HEAP32[$64>>2] = $66;
   $67 = ((($dst$02$us4)) + 28|0);
   $68 = ((($src$01$us5)) + 28|0);
   if ($16) {
    $71 = $68;$72 = $67;$dst$2$us1051 = $dst$02$us4;$n$1$us1152 = $14;$src$2$us950 = $src$01$us5;
    while(1) {
     $69 = (($n$1$us1152) + -1)|0;
     $70 = HEAP32[$71>>2]|0;
     HEAP32[$72>>2] = $70;
     $73 = ((($dst$2$us1051)) + 32|0);
     $74 = ((($src$2$us950)) + 32|0);
     $75 = HEAP32[$74>>2]|0;
     HEAP32[$73>>2] = $75;
     $76 = ((($dst$2$us1051)) + 36|0);
     $77 = ((($src$2$us950)) + 36|0);
     $78 = HEAP32[$77>>2]|0;
     HEAP32[$76>>2] = $78;
     $79 = ((($dst$2$us1051)) + 40|0);
     $80 = ((($src$2$us950)) + 40|0);
     $81 = HEAP32[$80>>2]|0;
     HEAP32[$79>>2] = $81;
     $82 = ((($dst$2$us1051)) + 44|0);
     $83 = ((($src$2$us950)) + 44|0);
     $84 = HEAP32[$83>>2]|0;
     HEAP32[$82>>2] = $84;
     $85 = ((($dst$2$us1051)) + 48|0);
     $86 = ((($src$2$us950)) + 48|0);
     $87 = HEAP32[$86>>2]|0;
     HEAP32[$85>>2] = $87;
     $88 = ((($dst$2$us1051)) + 52|0);
     $89 = ((($src$2$us950)) + 52|0);
     $90 = HEAP32[$89>>2]|0;
     HEAP32[$88>>2] = $90;
     $91 = ((($dst$2$us1051)) + 56|0);
     $92 = ((($src$2$us950)) + 56|0);
     $93 = HEAP32[$92>>2]|0;
     HEAP32[$91>>2] = $93;
     $94 = ((($dst$2$us1051)) + 60|0);
     $95 = ((($src$2$us950)) + 60|0);
     $96 = ($69|0)>(1);
     if ($96) {
      $71 = $95;$72 = $94;$dst$2$us1051 = $73;$n$1$us1152 = $69;$src$2$us950 = $74;
     } else {
      $$lcssa48 = $94;$$lcssa49 = $95;
      break;
     }
    }
   } else {
    $$lcssa48 = $67;$$lcssa49 = $68;
   }
   $97 = (($$lcssa49) + ($6)|0);
   $98 = (($$lcssa48) + ($10)|0);
   $99 = (($100) + -1)|0;
   $101 = ($100|0)==(0);
   if ($101) {
    break;
   } else {
    $100 = $99;$dst$02$us4 = $98;$src$01$us5 = $97;
   }
  }
  return;
  break;
 }
 default: {
  $131 = $11;$dst$02 = $8;$src$01 = $4;
  L5: while(1) {
   switch ($15|0) {
   case 1:  {
    $dst$8 = $dst$02;$n$7 = $14;$src$8 = $src$01;
    label = 17;
    break;
   }
   case 2:  {
    $dst$7 = $dst$02;$n$6 = $14;$src$7 = $src$01;
    label = 16;
    break;
   }
   case 6:  {
    $dst$3 = $dst$02;$n$2 = $14;$src$3 = $src$01;
    label = 12;
    break;
   }
   case 5:  {
    $dst$4 = $dst$02;$n$3 = $14;$src$4 = $src$01;
    label = 13;
    break;
   }
   case 4:  {
    $dst$5 = $dst$02;$n$4 = $14;$src$5 = $src$01;
    label = 14;
    break;
   }
   case 3:  {
    $dst$6 = $dst$02;$n$5 = $14;$src$6 = $src$01;
    label = 15;
    break;
   }
   default: {
    label = 18;
    break L5;
   }
   }
   while(1) {
    if ((label|0) == 12) {
     label = 0;
     $111 = HEAP32[$src$3>>2]|0;
     HEAP32[$dst$3>>2] = $111;
     $112 = ((($dst$3)) + 4|0);
     $113 = ((($src$3)) + 4|0);
     $dst$4 = $112;$n$3 = $n$2;$src$4 = $113;
     label = 13;
     continue;
    }
    else if ((label|0) == 13) {
     label = 0;
     $114 = HEAP32[$src$4>>2]|0;
     HEAP32[$dst$4>>2] = $114;
     $115 = ((($dst$4)) + 4|0);
     $116 = ((($src$4)) + 4|0);
     $dst$5 = $115;$n$4 = $n$3;$src$5 = $116;
     label = 14;
     continue;
    }
    else if ((label|0) == 14) {
     label = 0;
     $117 = HEAP32[$src$5>>2]|0;
     HEAP32[$dst$5>>2] = $117;
     $118 = ((($dst$5)) + 4|0);
     $119 = ((($src$5)) + 4|0);
     $dst$6 = $118;$n$5 = $n$4;$src$6 = $119;
     label = 15;
     continue;
    }
    else if ((label|0) == 15) {
     label = 0;
     $120 = HEAP32[$src$6>>2]|0;
     HEAP32[$dst$6>>2] = $120;
     $121 = ((($dst$6)) + 4|0);
     $122 = ((($src$6)) + 4|0);
     $dst$7 = $121;$n$6 = $n$5;$src$7 = $122;
     label = 16;
     continue;
    }
    else if ((label|0) == 16) {
     label = 0;
     $123 = HEAP32[$src$7>>2]|0;
     HEAP32[$dst$7>>2] = $123;
     $124 = ((($dst$7)) + 4|0);
     $125 = ((($src$7)) + 4|0);
     $dst$8 = $124;$n$7 = $n$6;$src$8 = $125;
     label = 17;
     continue;
    }
    else if ((label|0) == 17) {
     label = 0;
     $126 = HEAP32[$src$8>>2]|0;
     HEAP32[$dst$8>>2] = $126;
     $105 = ((($dst$8)) + 4|0);
     $104 = ((($src$8)) + 4|0);
     $127 = ($n$7|0)>(1);
     if (!($127)) {
      break;
     }
     $102 = (($n$7) + -1)|0;
     $103 = HEAP32[$104>>2]|0;
     HEAP32[$105>>2] = $103;
     $106 = ((($dst$8)) + 8|0);
     $107 = ((($src$8)) + 8|0);
     $108 = HEAP32[$107>>2]|0;
     HEAP32[$106>>2] = $108;
     $109 = ((($dst$8)) + 12|0);
     $110 = ((($src$8)) + 12|0);
     $dst$3 = $109;$n$2 = $102;$src$3 = $110;
     label = 12;
     continue;
    }
   }
   $128 = (($104) + ($6)|0);
   $129 = (($105) + ($10)|0);
   $130 = (($131) + -1)|0;
   $132 = ($131|0)==(0);
   if ($132) {
    label = 20;
    break;
   } else {
    $131 = $130;$dst$02 = $129;$src$01 = $128;
   }
  }
  if ((label|0) == 18) {
   // unreachable;
  }
  else if ((label|0) == 20) {
   return;
  }
 }
 }
}
function _BlitNtoNCopyAlpha($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0;
 var $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0;
 var $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $c$06 = 0, $div = 0, $div1 = 0, $div2 = 0, $div3 = 0, $div4 = 0, $dst$010 = 0, $dst$1$lcssa = 0;
 var $dst$17 = 0, $sA$0 = 0, $sA$0$shrunk = 0, $sB$0 = 0, $sB$0$shrunk = 0, $sG$0 = 0, $sG$0$shrunk = 0, $sR$0 = 0, $sR$0$shrunk = 0, $scevgep = 0, $scevgep14 = 0, $src$011 = 0, $src$1$lcssa = 0, $src$18 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 40|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 9|0);
 $11 = HEAP8[$10>>0]|0;
 $12 = $11&255;
 $13 = ((($info)) + 44|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($14)) + 9|0);
 $16 = HEAP8[$15>>0]|0;
 $17 = $16&255;
 $18 = ($3|0)==(0);
 if ($18) {
  return;
 }
 $19 = ((($info)) + 20|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = HEAP32[$info>>2]|0;
 $22 = ($1|0)==(0);
 $23 = ((($14)) + 28|0);
 $24 = ((($14)) + 32|0);
 $25 = ((($14)) + 33|0);
 $26 = ((($14)) + 34|0);
 $27 = ((($9)) + 12|0);
 $28 = ((($9)) + 32|0);
 $29 = ((($9)) + 28|0);
 $30 = ((($9)) + 16|0);
 $31 = ((($9)) + 20|0);
 $32 = ((($9)) + 24|0);
 $33 = Math_imul($1, $12)|0;
 $34 = Math_imul($1, $17)|0;
 $$in = $3;$dst$010 = $20;$src$011 = $21;
 while(1) {
  $35 = (($$in) + -1)|0;
  if ($22) {
   $dst$1$lcssa = $dst$010;$src$1$lcssa = $src$011;
  } else {
   $scevgep = (($src$011) + ($33)|0);
   $c$06 = $1;$dst$17 = $dst$010;$src$18 = $src$011;
   while(1) {
    switch ($12|0) {
    case 1:  {
     $36 = HEAP8[$src$18>>0]|0;
     $37 = $36&255;
     $38 = HEAP32[$27>>2]|0;
     $39 = $37 & $38;
     $40 = HEAP32[$28>>2]|0;
     $41 = $40 & 255;
     $42 = $39 >>> $41;
     $43 = HEAP32[$29>>2]|0;
     $44 = $43 & 255;
     $45 = (125924 + ($44<<2)|0);
     $46 = HEAP32[$45>>2]|0;
     $47 = (($46) + ($42)|0);
     $48 = HEAP8[$47>>0]|0;
     $49 = HEAP32[$30>>2]|0;
     $50 = $49 & $37;
     $51 = $40 >>> 8;
     $52 = $51 & 255;
     $53 = $50 >>> $52;
     $54 = $43 >>> 8;
     $55 = $54 & 255;
     $56 = (125924 + ($55<<2)|0);
     $57 = HEAP32[$56>>2]|0;
     $58 = (($57) + ($53)|0);
     $59 = HEAP8[$58>>0]|0;
     $60 = HEAP32[$31>>2]|0;
     $61 = $60 & $37;
     $62 = $40 >>> 16;
     $63 = $62 & 255;
     $64 = $61 >>> $63;
     $65 = $43 >>> 16;
     $66 = $65 & 255;
     $67 = (125924 + ($66<<2)|0);
     $68 = HEAP32[$67>>2]|0;
     $69 = (($68) + ($64)|0);
     $70 = HEAP8[$69>>0]|0;
     $71 = HEAP32[$32>>2]|0;
     $72 = $71 & $37;
     $73 = $40 >>> 24;
     $74 = $72 >>> $73;
     $75 = $43 >>> 24;
     $76 = (125924 + ($75<<2)|0);
     $77 = HEAP32[$76>>2]|0;
     $78 = (($77) + ($74)|0);
     $79 = HEAP8[$78>>0]|0;
     $sA$0$shrunk = $79;$sB$0$shrunk = $70;$sG$0$shrunk = $59;$sR$0$shrunk = $48;
     break;
    }
    case 2:  {
     $80 = HEAP16[$src$18>>1]|0;
     $81 = $80&65535;
     $82 = HEAP32[$27>>2]|0;
     $83 = $81 & $82;
     $84 = HEAP32[$28>>2]|0;
     $85 = $84 & 255;
     $86 = $83 >>> $85;
     $87 = HEAP32[$29>>2]|0;
     $88 = $87 & 255;
     $89 = (125924 + ($88<<2)|0);
     $90 = HEAP32[$89>>2]|0;
     $91 = (($90) + ($86)|0);
     $92 = HEAP8[$91>>0]|0;
     $93 = HEAP32[$30>>2]|0;
     $94 = $93 & $81;
     $95 = $84 >>> 8;
     $96 = $95 & 255;
     $97 = $94 >>> $96;
     $98 = $87 >>> 8;
     $99 = $98 & 255;
     $100 = (125924 + ($99<<2)|0);
     $101 = HEAP32[$100>>2]|0;
     $102 = (($101) + ($97)|0);
     $103 = HEAP8[$102>>0]|0;
     $104 = HEAP32[$31>>2]|0;
     $105 = $104 & $81;
     $106 = $84 >>> 16;
     $107 = $106 & 255;
     $108 = $105 >>> $107;
     $109 = $87 >>> 16;
     $110 = $109 & 255;
     $111 = (125924 + ($110<<2)|0);
     $112 = HEAP32[$111>>2]|0;
     $113 = (($112) + ($108)|0);
     $114 = HEAP8[$113>>0]|0;
     $115 = HEAP32[$32>>2]|0;
     $116 = $115 & $81;
     $117 = $84 >>> 24;
     $118 = $116 >>> $117;
     $119 = $87 >>> 24;
     $120 = (125924 + ($119<<2)|0);
     $121 = HEAP32[$120>>2]|0;
     $122 = (($121) + ($118)|0);
     $123 = HEAP8[$122>>0]|0;
     $sA$0$shrunk = $123;$sB$0$shrunk = $114;$sG$0$shrunk = $103;$sR$0$shrunk = $92;
     break;
    }
    case 3:  {
     $124 = HEAP32[$28>>2]|0;
     $div3 = $124 >>> 3;
     $125 = $div3 & 31;
     $126 = (($src$18) + ($125)|0);
     $127 = HEAP8[$126>>0]|0;
     $128 = $124 >>> 11;
     $div4 = $128 & 31;
     $129 = (($src$18) + ($div4)|0);
     $130 = HEAP8[$129>>0]|0;
     $131 = $124 >>> 19;
     $132 = $131 & 31;
     $133 = (($src$18) + ($132)|0);
     $134 = HEAP8[$133>>0]|0;
     $sA$0$shrunk = -1;$sB$0$shrunk = $134;$sG$0$shrunk = $130;$sR$0$shrunk = $127;
     break;
    }
    case 4:  {
     $135 = HEAP32[$src$18>>2]|0;
     $136 = HEAP32[$27>>2]|0;
     $137 = $136 & $135;
     $138 = HEAP32[$28>>2]|0;
     $139 = $138 & 255;
     $140 = $137 >>> $139;
     $141 = HEAP32[$29>>2]|0;
     $142 = $141 & 255;
     $143 = (125924 + ($142<<2)|0);
     $144 = HEAP32[$143>>2]|0;
     $145 = (($144) + ($140)|0);
     $146 = HEAP8[$145>>0]|0;
     $147 = HEAP32[$30>>2]|0;
     $148 = $147 & $135;
     $149 = $138 >>> 8;
     $150 = $149 & 255;
     $151 = $148 >>> $150;
     $152 = $141 >>> 8;
     $153 = $152 & 255;
     $154 = (125924 + ($153<<2)|0);
     $155 = HEAP32[$154>>2]|0;
     $156 = (($155) + ($151)|0);
     $157 = HEAP8[$156>>0]|0;
     $158 = HEAP32[$31>>2]|0;
     $159 = $158 & $135;
     $160 = $138 >>> 16;
     $161 = $160 & 255;
     $162 = $159 >>> $161;
     $163 = $141 >>> 16;
     $164 = $163 & 255;
     $165 = (125924 + ($164<<2)|0);
     $166 = HEAP32[$165>>2]|0;
     $167 = (($166) + ($162)|0);
     $168 = HEAP8[$167>>0]|0;
     $169 = HEAP32[$32>>2]|0;
     $170 = $169 & $135;
     $171 = $138 >>> 24;
     $172 = $170 >>> $171;
     $173 = $141 >>> 24;
     $174 = (125924 + ($173<<2)|0);
     $175 = HEAP32[$174>>2]|0;
     $176 = (($175) + ($172)|0);
     $177 = HEAP8[$176>>0]|0;
     $sA$0$shrunk = $177;$sB$0$shrunk = $168;$sG$0$shrunk = $157;$sR$0$shrunk = $146;
     break;
    }
    default: {
     $sA$0$shrunk = 0;$sB$0$shrunk = 0;$sG$0$shrunk = 0;$sR$0$shrunk = 0;
    }
    }
    $sA$0 = $sA$0$shrunk&255;
    $sB$0 = $sB$0$shrunk&255;
    $sG$0 = $sG$0$shrunk&255;
    $sR$0 = $sR$0$shrunk&255;
    switch ($17|0) {
    case 1:  {
     $178 = HEAP32[$23>>2]|0;
     $179 = $178 & 255;
     $180 = $sR$0 >>> $179;
     $181 = HEAP32[$24>>2]|0;
     $182 = $181 & 255;
     $183 = $180 << $182;
     $184 = $178 >>> 8;
     $185 = $184 & 255;
     $186 = $sG$0 >>> $185;
     $187 = $181 >>> 8;
     $188 = $187 & 255;
     $189 = $186 << $188;
     $190 = $189 | $183;
     $191 = $178 >>> 16;
     $192 = $191 & 255;
     $193 = $sB$0 >>> $192;
     $194 = $181 >>> 16;
     $195 = $194 & 255;
     $196 = $193 << $195;
     $197 = $190 | $196;
     $198 = $178 >>> 24;
     $199 = $sA$0 >>> $198;
     $200 = $181 >>> 24;
     $201 = $199 << $200;
     $202 = $197 | $201;
     $203 = $202&255;
     HEAP8[$dst$17>>0] = $203;
     break;
    }
    case 2:  {
     $204 = HEAP32[$23>>2]|0;
     $205 = $204 & 255;
     $206 = $sR$0 >>> $205;
     $207 = HEAP32[$24>>2]|0;
     $208 = $207 & 255;
     $209 = $206 << $208;
     $210 = $204 >>> 8;
     $211 = $210 & 255;
     $212 = $sG$0 >>> $211;
     $213 = $207 >>> 8;
     $214 = $213 & 255;
     $215 = $212 << $214;
     $216 = $215 | $209;
     $217 = $204 >>> 16;
     $218 = $217 & 255;
     $219 = $sB$0 >>> $218;
     $220 = $207 >>> 16;
     $221 = $220 & 255;
     $222 = $219 << $221;
     $223 = $216 | $222;
     $224 = $204 >>> 24;
     $225 = $sA$0 >>> $224;
     $226 = $207 >>> 24;
     $227 = $225 << $226;
     $228 = $223 | $227;
     $229 = $228&65535;
     HEAP16[$dst$17>>1] = $229;
     break;
    }
    case 3:  {
     $230 = HEAP8[$24>>0]|0;
     $div = ($230&255) >>> 3;
     $231 = $div&255;
     $232 = (($dst$17) + ($231)|0);
     HEAP8[$232>>0] = $sR$0$shrunk;
     $233 = HEAP8[$25>>0]|0;
     $div1 = ($233&255) >>> 3;
     $234 = $div1&255;
     $235 = (($dst$17) + ($234)|0);
     HEAP8[$235>>0] = $sG$0$shrunk;
     $236 = HEAP8[$26>>0]|0;
     $div2 = ($236&255) >>> 3;
     $237 = $div2&255;
     $238 = (($dst$17) + ($237)|0);
     HEAP8[$238>>0] = $sB$0$shrunk;
     break;
    }
    case 4:  {
     $239 = HEAP32[$23>>2]|0;
     $240 = $239 & 255;
     $241 = $sR$0 >>> $240;
     $242 = HEAP32[$24>>2]|0;
     $243 = $242 & 255;
     $244 = $241 << $243;
     $245 = $239 >>> 8;
     $246 = $245 & 255;
     $247 = $sG$0 >>> $246;
     $248 = $242 >>> 8;
     $249 = $248 & 255;
     $250 = $247 << $249;
     $251 = $250 | $244;
     $252 = $239 >>> 16;
     $253 = $252 & 255;
     $254 = $sB$0 >>> $253;
     $255 = $242 >>> 16;
     $256 = $255 & 255;
     $257 = $254 << $256;
     $258 = $251 | $257;
     $259 = $239 >>> 24;
     $260 = $sA$0 >>> $259;
     $261 = $242 >>> 24;
     $262 = $260 << $261;
     $263 = $258 | $262;
     HEAP32[$dst$17>>2] = $263;
     break;
    }
    default: {
    }
    }
    $264 = (($dst$17) + ($17)|0);
    $265 = (($src$18) + ($12)|0);
    $266 = (($c$06) + -1)|0;
    $267 = ($266|0)==(0);
    if ($267) {
     break;
    } else {
     $c$06 = $266;$dst$17 = $264;$src$18 = $265;
    }
   }
   $scevgep14 = (($dst$010) + ($34)|0);
   $dst$1$lcssa = $scevgep14;$src$1$lcssa = $scevgep;
  }
  $268 = (($src$1$lcssa) + ($5)|0);
  $269 = (($dst$1$lcssa) + ($7)|0);
  $270 = ($35|0)==(0);
  if ($270) {
   break;
  } else {
   $$in = $35;$dst$010 = $269;$src$011 = $268;
  }
 }
 return;
}
function _BlitNtoNKey($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0;
 var $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0;
 var $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0;
 var $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0;
 var $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0;
 var $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0;
 var $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0;
 var $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0;
 var $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0;
 var $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0;
 var $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0;
 var $1195 = 0, $1196 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0;
 var $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0;
 var $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0;
 var $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0;
 var $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0;
 var $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0;
 var $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0;
 var $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0;
 var $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0;
 var $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0;
 var $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0;
 var $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0;
 var $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0;
 var $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0;
 var $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0;
 var $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0;
 var $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0;
 var $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0;
 var $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0;
 var $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0;
 var $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0;
 var $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0;
 var $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0;
 var $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0;
 var $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0;
 var $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0;
 var $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0;
 var $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0;
 var $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0;
 var $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0;
 var $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0;
 var $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0;
 var $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0;
 var $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0;
 var $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0;
 var $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0;
 var $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0;
 var $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0;
 var $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0;
 var $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0;
 var $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0;
 var $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0;
 var $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0;
 var $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0;
 var $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $Pixel$0 = 0, $Pixel11$0 = 0;
 var $Pixel19$0 = 0, $Pixel27$0 = 0, $Pixel3$0 = 0, $Pixel35$0 = 0, $Pixel43$0 = 0, $Pixel51$0 = 0, $div = 0, $div1 = 0, $div10 = 0, $div11 = 0, $div12 = 0, $div13 = 0, $div14 = 0, $div15 = 0, $div16 = 0, $div17 = 0, $div18 = 0, $div19 = 0, $div2 = 0, $div20 = 0;
 var $div21 = 0, $div22 = 0, $div23 = 0, $div3 = 0, $div4 = 0, $div5 = 0, $div6 = 0, $div7 = 0, $div8 = 0, $div9 = 0, $dst$024 = 0, $dst$1 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $n$0 = 0;
 var $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $src$025 = 0, $src$1 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$info>>2]|0;
 $5 = ((($info)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($info)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($info)) + 36|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($info)) + 56|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($info)) + 40|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($info)) + 44|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($14)) + 9|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = $18&255;
 $20 = ((($16)) + 9|0);
 $21 = HEAP8[$20>>0]|0;
 $22 = $21&255;
 $23 = ((($16)) + 24|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ($24|0)==(0);
 if ($25) {
  $124 = 0;
 } else {
  $26 = ((($info)) + 63|0);
  $27 = HEAP8[$26>>0]|0;
  $28 = $27&255;
  $124 = $28;
 }
 $29 = ((($14)) + 24|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = $30 ^ -1;
 $32 = $12 & $31;
 $33 = ($3|0)==(0);
 if ($33) {
  return;
 }
 $34 = (($1) + 7)|0;
 $35 = (($34|0) / 8)&-1;
 $36 = $1 & 7;
 $37 = ((($14)) + 12|0);
 $38 = ((($14)) + 32|0);
 $39 = ((($14)) + 28|0);
 $40 = ((($14)) + 16|0);
 $41 = ((($14)) + 20|0);
 $42 = ((($16)) + 28|0);
 $43 = ((($16)) + 32|0);
 $44 = ((($16)) + 33|0);
 $45 = ((($16)) + 34|0);
 $$in = $3;$dst$024 = $8;$src$025 = $4;
 L7: while(1) {
  $46 = (($$in) + -1)|0;
  switch ($36|0) {
  case 0:  {
   $dst$1 = $dst$024;$n$0 = $35;$src$1 = $src$025;
   label = 6;
   break;
  }
  case 7:  {
   $dst$2 = $dst$024;$n$1 = $35;$src$2 = $src$025;
   label = 18;
   break;
  }
  case 6:  {
   $dst$3 = $dst$024;$n$2 = $35;$src$3 = $src$025;
   label = 30;
   break;
  }
  case 5:  {
   $dst$4 = $dst$024;$n$3 = $35;$src$4 = $src$025;
   label = 42;
   break;
  }
  case 4:  {
   $dst$5 = $dst$024;$n$4 = $35;$src$5 = $src$025;
   label = 54;
   break;
  }
  case 3:  {
   $dst$6 = $dst$024;$n$5 = $35;$src$6 = $src$025;
   label = 66;
   break;
  }
  case 2:  {
   $dst$7 = $dst$024;$n$6 = $35;$src$7 = $src$025;
   label = 78;
   break;
  }
  case 1:  {
   $dst$8 = $dst$024;$n$7 = $35;$src$8 = $src$025;
   label = 90;
   break;
  }
  default: {
   label = 102;
   break L7;
  }
  }
  while(1) {
   if ((label|0) == 6) {
    label = 0;
    switch ($19|0) {
    case 1:  {
     $47 = HEAP8[$src$1>>0]|0;
     $48 = $47&255;
     $Pixel$0 = $48;
     break;
    }
    case 2:  {
     $49 = HEAP16[$src$1>>1]|0;
     $50 = $49&65535;
     $Pixel$0 = $50;
     break;
    }
    case 3:  {
     $51 = HEAP8[$src$1>>0]|0;
     $52 = $51&255;
     $53 = ((($src$1)) + 1|0);
     $54 = HEAP8[$53>>0]|0;
     $55 = $54&255;
     $56 = $55 << 8;
     $57 = $56 | $52;
     $58 = ((($src$1)) + 2|0);
     $59 = HEAP8[$58>>0]|0;
     $60 = $59&255;
     $61 = $60 << 16;
     $62 = $57 | $61;
     $Pixel$0 = $62;
     break;
    }
    case 4:  {
     $63 = HEAP32[$src$1>>2]|0;
     $Pixel$0 = $63;
     break;
    }
    default: {
     $Pixel$0 = 0;
    }
    }
    $64 = $Pixel$0 & $31;
    $65 = ($64|0)==($32|0);
    L18: do {
     if (!($65)) {
      $66 = HEAP32[$37>>2]|0;
      $67 = $66 & $Pixel$0;
      $68 = HEAP32[$38>>2]|0;
      $69 = $68 & 255;
      $70 = $67 >>> $69;
      $71 = HEAP32[$39>>2]|0;
      $72 = $71 & 255;
      $73 = (125924 + ($72<<2)|0);
      $74 = HEAP32[$73>>2]|0;
      $75 = (($74) + ($70)|0);
      $76 = HEAP8[$75>>0]|0;
      $77 = $76&255;
      $78 = HEAP32[$40>>2]|0;
      $79 = $78 & $Pixel$0;
      $80 = $68 >>> 8;
      $81 = $80 & 255;
      $82 = $79 >>> $81;
      $83 = $71 >>> 8;
      $84 = $83 & 255;
      $85 = (125924 + ($84<<2)|0);
      $86 = HEAP32[$85>>2]|0;
      $87 = (($86) + ($82)|0);
      $88 = HEAP8[$87>>0]|0;
      $89 = $88&255;
      $90 = HEAP32[$41>>2]|0;
      $91 = $90 & $Pixel$0;
      $92 = $68 >>> 16;
      $93 = $92 & 255;
      $94 = $91 >>> $93;
      $95 = $71 >>> 16;
      $96 = $95 & 255;
      $97 = (125924 + ($96<<2)|0);
      $98 = HEAP32[$97>>2]|0;
      $99 = (($98) + ($94)|0);
      $100 = HEAP8[$99>>0]|0;
      $101 = $100&255;
      switch ($22|0) {
      case 1:  {
       $102 = HEAP32[$42>>2]|0;
       $103 = $102 & 255;
       $104 = $77 >>> $103;
       $105 = HEAP32[$43>>2]|0;
       $106 = $105 & 255;
       $107 = $104 << $106;
       $108 = $102 >>> 8;
       $109 = $108 & 255;
       $110 = $89 >>> $109;
       $111 = $105 >>> 8;
       $112 = $111 & 255;
       $113 = $110 << $112;
       $114 = $113 | $107;
       $115 = $102 >>> 16;
       $116 = $115 & 255;
       $117 = $101 >>> $116;
       $118 = $105 >>> 16;
       $119 = $118 & 255;
       $120 = $117 << $119;
       $121 = $114 | $120;
       $122 = $102 >>> 24;
       $123 = $124 >>> $122;
       $125 = $105 >>> 24;
       $126 = $123 << $125;
       $127 = $121 | $126;
       $128 = $127&255;
       HEAP8[$dst$1>>0] = $128;
       break L18;
       break;
      }
      case 2:  {
       $129 = HEAP32[$42>>2]|0;
       $130 = $129 & 255;
       $131 = $77 >>> $130;
       $132 = HEAP32[$43>>2]|0;
       $133 = $132 & 255;
       $134 = $131 << $133;
       $135 = $129 >>> 8;
       $136 = $135 & 255;
       $137 = $89 >>> $136;
       $138 = $132 >>> 8;
       $139 = $138 & 255;
       $140 = $137 << $139;
       $141 = $140 | $134;
       $142 = $129 >>> 16;
       $143 = $142 & 255;
       $144 = $101 >>> $143;
       $145 = $132 >>> 16;
       $146 = $145 & 255;
       $147 = $144 << $146;
       $148 = $141 | $147;
       $149 = $129 >>> 24;
       $150 = $124 >>> $149;
       $151 = $132 >>> 24;
       $152 = $150 << $151;
       $153 = $148 | $152;
       $154 = $153&65535;
       HEAP16[$dst$1>>1] = $154;
       break L18;
       break;
      }
      case 3:  {
       $155 = HEAP8[$43>>0]|0;
       $div18 = ($155&255) >>> 3;
       $156 = $div18&255;
       $157 = (($dst$1) + ($156)|0);
       HEAP8[$157>>0] = $76;
       $158 = HEAP8[$44>>0]|0;
       $div19 = ($158&255) >>> 3;
       $159 = $div19&255;
       $160 = (($dst$1) + ($159)|0);
       HEAP8[$160>>0] = $88;
       $161 = HEAP8[$45>>0]|0;
       $div20 = ($161&255) >>> 3;
       $162 = $div20&255;
       $163 = (($dst$1) + ($162)|0);
       HEAP8[$163>>0] = $100;
       break L18;
       break;
      }
      case 4:  {
       $164 = HEAP32[$42>>2]|0;
       $165 = $164 & 255;
       $166 = $77 >>> $165;
       $167 = HEAP32[$43>>2]|0;
       $168 = $167 & 255;
       $169 = $166 << $168;
       $170 = $164 >>> 8;
       $171 = $170 & 255;
       $172 = $89 >>> $171;
       $173 = $167 >>> 8;
       $174 = $173 & 255;
       $175 = $172 << $174;
       $176 = $175 | $169;
       $177 = $164 >>> 16;
       $178 = $177 & 255;
       $179 = $101 >>> $178;
       $180 = $167 >>> 16;
       $181 = $180 & 255;
       $182 = $179 << $181;
       $183 = $176 | $182;
       $184 = $164 >>> 24;
       $185 = $124 >>> $184;
       $186 = $167 >>> 24;
       $187 = $185 << $186;
       $188 = $183 | $187;
       HEAP32[$dst$1>>2] = $188;
       break L18;
       break;
      }
      default: {
       break L18;
      }
      }
     }
    } while(0);
    $189 = (($dst$1) + ($22)|0);
    $190 = (($src$1) + ($19)|0);
    $dst$2 = $189;$n$1 = $n$0;$src$2 = $190;
    label = 18;
    continue;
   }
   else if ((label|0) == 18) {
    label = 0;
    switch ($19|0) {
    case 1:  {
     $191 = HEAP8[$src$2>>0]|0;
     $192 = $191&255;
     $Pixel3$0 = $192;
     break;
    }
    case 2:  {
     $193 = HEAP16[$src$2>>1]|0;
     $194 = $193&65535;
     $Pixel3$0 = $194;
     break;
    }
    case 3:  {
     $195 = HEAP8[$src$2>>0]|0;
     $196 = $195&255;
     $197 = ((($src$2)) + 1|0);
     $198 = HEAP8[$197>>0]|0;
     $199 = $198&255;
     $200 = $199 << 8;
     $201 = $200 | $196;
     $202 = ((($src$2)) + 2|0);
     $203 = HEAP8[$202>>0]|0;
     $204 = $203&255;
     $205 = $204 << 16;
     $206 = $201 | $205;
     $Pixel3$0 = $206;
     break;
    }
    case 4:  {
     $207 = HEAP32[$src$2>>2]|0;
     $Pixel3$0 = $207;
     break;
    }
    default: {
     $Pixel3$0 = 0;
    }
    }
    $208 = $Pixel3$0 & $31;
    $209 = ($208|0)==($32|0);
    L33: do {
     if (!($209)) {
      $210 = HEAP32[$37>>2]|0;
      $211 = $210 & $Pixel3$0;
      $212 = HEAP32[$38>>2]|0;
      $213 = $212 & 255;
      $214 = $211 >>> $213;
      $215 = HEAP32[$39>>2]|0;
      $216 = $215 & 255;
      $217 = (125924 + ($216<<2)|0);
      $218 = HEAP32[$217>>2]|0;
      $219 = (($218) + ($214)|0);
      $220 = HEAP8[$219>>0]|0;
      $221 = $220&255;
      $222 = HEAP32[$40>>2]|0;
      $223 = $222 & $Pixel3$0;
      $224 = $212 >>> 8;
      $225 = $224 & 255;
      $226 = $223 >>> $225;
      $227 = $215 >>> 8;
      $228 = $227 & 255;
      $229 = (125924 + ($228<<2)|0);
      $230 = HEAP32[$229>>2]|0;
      $231 = (($230) + ($226)|0);
      $232 = HEAP8[$231>>0]|0;
      $233 = $232&255;
      $234 = HEAP32[$41>>2]|0;
      $235 = $234 & $Pixel3$0;
      $236 = $212 >>> 16;
      $237 = $236 & 255;
      $238 = $235 >>> $237;
      $239 = $215 >>> 16;
      $240 = $239 & 255;
      $241 = (125924 + ($240<<2)|0);
      $242 = HEAP32[$241>>2]|0;
      $243 = (($242) + ($238)|0);
      $244 = HEAP8[$243>>0]|0;
      $245 = $244&255;
      switch ($22|0) {
      case 1:  {
       $246 = HEAP32[$42>>2]|0;
       $247 = $246 & 255;
       $248 = $221 >>> $247;
       $249 = HEAP32[$43>>2]|0;
       $250 = $249 & 255;
       $251 = $248 << $250;
       $252 = $246 >>> 8;
       $253 = $252 & 255;
       $254 = $233 >>> $253;
       $255 = $249 >>> 8;
       $256 = $255 & 255;
       $257 = $254 << $256;
       $258 = $257 | $251;
       $259 = $246 >>> 16;
       $260 = $259 & 255;
       $261 = $245 >>> $260;
       $262 = $249 >>> 16;
       $263 = $262 & 255;
       $264 = $261 << $263;
       $265 = $258 | $264;
       $266 = $246 >>> 24;
       $267 = $124 >>> $266;
       $268 = $249 >>> 24;
       $269 = $267 << $268;
       $270 = $265 | $269;
       $271 = $270&255;
       HEAP8[$dst$2>>0] = $271;
       break L33;
       break;
      }
      case 2:  {
       $272 = HEAP32[$42>>2]|0;
       $273 = $272 & 255;
       $274 = $221 >>> $273;
       $275 = HEAP32[$43>>2]|0;
       $276 = $275 & 255;
       $277 = $274 << $276;
       $278 = $272 >>> 8;
       $279 = $278 & 255;
       $280 = $233 >>> $279;
       $281 = $275 >>> 8;
       $282 = $281 & 255;
       $283 = $280 << $282;
       $284 = $283 | $277;
       $285 = $272 >>> 16;
       $286 = $285 & 255;
       $287 = $245 >>> $286;
       $288 = $275 >>> 16;
       $289 = $288 & 255;
       $290 = $287 << $289;
       $291 = $284 | $290;
       $292 = $272 >>> 24;
       $293 = $124 >>> $292;
       $294 = $275 >>> 24;
       $295 = $293 << $294;
       $296 = $291 | $295;
       $297 = $296&65535;
       HEAP16[$dst$2>>1] = $297;
       break L33;
       break;
      }
      case 3:  {
       $298 = HEAP8[$43>>0]|0;
       $div15 = ($298&255) >>> 3;
       $299 = $div15&255;
       $300 = (($dst$2) + ($299)|0);
       HEAP8[$300>>0] = $220;
       $301 = HEAP8[$44>>0]|0;
       $div16 = ($301&255) >>> 3;
       $302 = $div16&255;
       $303 = (($dst$2) + ($302)|0);
       HEAP8[$303>>0] = $232;
       $304 = HEAP8[$45>>0]|0;
       $div17 = ($304&255) >>> 3;
       $305 = $div17&255;
       $306 = (($dst$2) + ($305)|0);
       HEAP8[$306>>0] = $244;
       break L33;
       break;
      }
      case 4:  {
       $307 = HEAP32[$42>>2]|0;
       $308 = $307 & 255;
       $309 = $221 >>> $308;
       $310 = HEAP32[$43>>2]|0;
       $311 = $310 & 255;
       $312 = $309 << $311;
       $313 = $307 >>> 8;
       $314 = $313 & 255;
       $315 = $233 >>> $314;
       $316 = $310 >>> 8;
       $317 = $316 & 255;
       $318 = $315 << $317;
       $319 = $318 | $312;
       $320 = $307 >>> 16;
       $321 = $320 & 255;
       $322 = $245 >>> $321;
       $323 = $310 >>> 16;
       $324 = $323 & 255;
       $325 = $322 << $324;
       $326 = $319 | $325;
       $327 = $307 >>> 24;
       $328 = $124 >>> $327;
       $329 = $310 >>> 24;
       $330 = $328 << $329;
       $331 = $326 | $330;
       HEAP32[$dst$2>>2] = $331;
       break L33;
       break;
      }
      default: {
       break L33;
      }
      }
     }
    } while(0);
    $332 = (($dst$2) + ($22)|0);
    $333 = (($src$2) + ($19)|0);
    $dst$3 = $332;$n$2 = $n$1;$src$3 = $333;
    label = 30;
    continue;
   }
   else if ((label|0) == 30) {
    label = 0;
    switch ($19|0) {
    case 1:  {
     $334 = HEAP8[$src$3>>0]|0;
     $335 = $334&255;
     $Pixel11$0 = $335;
     break;
    }
    case 2:  {
     $336 = HEAP16[$src$3>>1]|0;
     $337 = $336&65535;
     $Pixel11$0 = $337;
     break;
    }
    case 3:  {
     $338 = HEAP8[$src$3>>0]|0;
     $339 = $338&255;
     $340 = ((($src$3)) + 1|0);
     $341 = HEAP8[$340>>0]|0;
     $342 = $341&255;
     $343 = $342 << 8;
     $344 = $343 | $339;
     $345 = ((($src$3)) + 2|0);
     $346 = HEAP8[$345>>0]|0;
     $347 = $346&255;
     $348 = $347 << 16;
     $349 = $344 | $348;
     $Pixel11$0 = $349;
     break;
    }
    case 4:  {
     $350 = HEAP32[$src$3>>2]|0;
     $Pixel11$0 = $350;
     break;
    }
    default: {
     $Pixel11$0 = 0;
    }
    }
    $351 = $Pixel11$0 & $31;
    $352 = ($351|0)==($32|0);
    L48: do {
     if (!($352)) {
      $353 = HEAP32[$37>>2]|0;
      $354 = $353 & $Pixel11$0;
      $355 = HEAP32[$38>>2]|0;
      $356 = $355 & 255;
      $357 = $354 >>> $356;
      $358 = HEAP32[$39>>2]|0;
      $359 = $358 & 255;
      $360 = (125924 + ($359<<2)|0);
      $361 = HEAP32[$360>>2]|0;
      $362 = (($361) + ($357)|0);
      $363 = HEAP8[$362>>0]|0;
      $364 = $363&255;
      $365 = HEAP32[$40>>2]|0;
      $366 = $365 & $Pixel11$0;
      $367 = $355 >>> 8;
      $368 = $367 & 255;
      $369 = $366 >>> $368;
      $370 = $358 >>> 8;
      $371 = $370 & 255;
      $372 = (125924 + ($371<<2)|0);
      $373 = HEAP32[$372>>2]|0;
      $374 = (($373) + ($369)|0);
      $375 = HEAP8[$374>>0]|0;
      $376 = $375&255;
      $377 = HEAP32[$41>>2]|0;
      $378 = $377 & $Pixel11$0;
      $379 = $355 >>> 16;
      $380 = $379 & 255;
      $381 = $378 >>> $380;
      $382 = $358 >>> 16;
      $383 = $382 & 255;
      $384 = (125924 + ($383<<2)|0);
      $385 = HEAP32[$384>>2]|0;
      $386 = (($385) + ($381)|0);
      $387 = HEAP8[$386>>0]|0;
      $388 = $387&255;
      switch ($22|0) {
      case 1:  {
       $389 = HEAP32[$42>>2]|0;
       $390 = $389 & 255;
       $391 = $364 >>> $390;
       $392 = HEAP32[$43>>2]|0;
       $393 = $392 & 255;
       $394 = $391 << $393;
       $395 = $389 >>> 8;
       $396 = $395 & 255;
       $397 = $376 >>> $396;
       $398 = $392 >>> 8;
       $399 = $398 & 255;
       $400 = $397 << $399;
       $401 = $400 | $394;
       $402 = $389 >>> 16;
       $403 = $402 & 255;
       $404 = $388 >>> $403;
       $405 = $392 >>> 16;
       $406 = $405 & 255;
       $407 = $404 << $406;
       $408 = $401 | $407;
       $409 = $389 >>> 24;
       $410 = $124 >>> $409;
       $411 = $392 >>> 24;
       $412 = $410 << $411;
       $413 = $408 | $412;
       $414 = $413&255;
       HEAP8[$dst$3>>0] = $414;
       break L48;
       break;
      }
      case 2:  {
       $415 = HEAP32[$42>>2]|0;
       $416 = $415 & 255;
       $417 = $364 >>> $416;
       $418 = HEAP32[$43>>2]|0;
       $419 = $418 & 255;
       $420 = $417 << $419;
       $421 = $415 >>> 8;
       $422 = $421 & 255;
       $423 = $376 >>> $422;
       $424 = $418 >>> 8;
       $425 = $424 & 255;
       $426 = $423 << $425;
       $427 = $426 | $420;
       $428 = $415 >>> 16;
       $429 = $428 & 255;
       $430 = $388 >>> $429;
       $431 = $418 >>> 16;
       $432 = $431 & 255;
       $433 = $430 << $432;
       $434 = $427 | $433;
       $435 = $415 >>> 24;
       $436 = $124 >>> $435;
       $437 = $418 >>> 24;
       $438 = $436 << $437;
       $439 = $434 | $438;
       $440 = $439&65535;
       HEAP16[$dst$3>>1] = $440;
       break L48;
       break;
      }
      case 3:  {
       $441 = HEAP8[$43>>0]|0;
       $div12 = ($441&255) >>> 3;
       $442 = $div12&255;
       $443 = (($dst$3) + ($442)|0);
       HEAP8[$443>>0] = $363;
       $444 = HEAP8[$44>>0]|0;
       $div13 = ($444&255) >>> 3;
       $445 = $div13&255;
       $446 = (($dst$3) + ($445)|0);
       HEAP8[$446>>0] = $375;
       $447 = HEAP8[$45>>0]|0;
       $div14 = ($447&255) >>> 3;
       $448 = $div14&255;
       $449 = (($dst$3) + ($448)|0);
       HEAP8[$449>>0] = $387;
       break L48;
       break;
      }
      case 4:  {
       $450 = HEAP32[$42>>2]|0;
       $451 = $450 & 255;
       $452 = $364 >>> $451;
       $453 = HEAP32[$43>>2]|0;
       $454 = $453 & 255;
       $455 = $452 << $454;
       $456 = $450 >>> 8;
       $457 = $456 & 255;
       $458 = $376 >>> $457;
       $459 = $453 >>> 8;
       $460 = $459 & 255;
       $461 = $458 << $460;
       $462 = $461 | $455;
       $463 = $450 >>> 16;
       $464 = $463 & 255;
       $465 = $388 >>> $464;
       $466 = $453 >>> 16;
       $467 = $466 & 255;
       $468 = $465 << $467;
       $469 = $462 | $468;
       $470 = $450 >>> 24;
       $471 = $124 >>> $470;
       $472 = $453 >>> 24;
       $473 = $471 << $472;
       $474 = $469 | $473;
       HEAP32[$dst$3>>2] = $474;
       break L48;
       break;
      }
      default: {
       break L48;
      }
      }
     }
    } while(0);
    $475 = (($dst$3) + ($22)|0);
    $476 = (($src$3) + ($19)|0);
    $dst$4 = $475;$n$3 = $n$2;$src$4 = $476;
    label = 42;
    continue;
   }
   else if ((label|0) == 42) {
    label = 0;
    switch ($19|0) {
    case 1:  {
     $477 = HEAP8[$src$4>>0]|0;
     $478 = $477&255;
     $Pixel19$0 = $478;
     break;
    }
    case 2:  {
     $479 = HEAP16[$src$4>>1]|0;
     $480 = $479&65535;
     $Pixel19$0 = $480;
     break;
    }
    case 3:  {
     $481 = HEAP8[$src$4>>0]|0;
     $482 = $481&255;
     $483 = ((($src$4)) + 1|0);
     $484 = HEAP8[$483>>0]|0;
     $485 = $484&255;
     $486 = $485 << 8;
     $487 = $486 | $482;
     $488 = ((($src$4)) + 2|0);
     $489 = HEAP8[$488>>0]|0;
     $490 = $489&255;
     $491 = $490 << 16;
     $492 = $487 | $491;
     $Pixel19$0 = $492;
     break;
    }
    case 4:  {
     $493 = HEAP32[$src$4>>2]|0;
     $Pixel19$0 = $493;
     break;
    }
    default: {
     $Pixel19$0 = 0;
    }
    }
    $494 = $Pixel19$0 & $31;
    $495 = ($494|0)==($32|0);
    L63: do {
     if (!($495)) {
      $496 = HEAP32[$37>>2]|0;
      $497 = $496 & $Pixel19$0;
      $498 = HEAP32[$38>>2]|0;
      $499 = $498 & 255;
      $500 = $497 >>> $499;
      $501 = HEAP32[$39>>2]|0;
      $502 = $501 & 255;
      $503 = (125924 + ($502<<2)|0);
      $504 = HEAP32[$503>>2]|0;
      $505 = (($504) + ($500)|0);
      $506 = HEAP8[$505>>0]|0;
      $507 = $506&255;
      $508 = HEAP32[$40>>2]|0;
      $509 = $508 & $Pixel19$0;
      $510 = $498 >>> 8;
      $511 = $510 & 255;
      $512 = $509 >>> $511;
      $513 = $501 >>> 8;
      $514 = $513 & 255;
      $515 = (125924 + ($514<<2)|0);
      $516 = HEAP32[$515>>2]|0;
      $517 = (($516) + ($512)|0);
      $518 = HEAP8[$517>>0]|0;
      $519 = $518&255;
      $520 = HEAP32[$41>>2]|0;
      $521 = $520 & $Pixel19$0;
      $522 = $498 >>> 16;
      $523 = $522 & 255;
      $524 = $521 >>> $523;
      $525 = $501 >>> 16;
      $526 = $525 & 255;
      $527 = (125924 + ($526<<2)|0);
      $528 = HEAP32[$527>>2]|0;
      $529 = (($528) + ($524)|0);
      $530 = HEAP8[$529>>0]|0;
      $531 = $530&255;
      switch ($22|0) {
      case 1:  {
       $532 = HEAP32[$42>>2]|0;
       $533 = $532 & 255;
       $534 = $507 >>> $533;
       $535 = HEAP32[$43>>2]|0;
       $536 = $535 & 255;
       $537 = $534 << $536;
       $538 = $532 >>> 8;
       $539 = $538 & 255;
       $540 = $519 >>> $539;
       $541 = $535 >>> 8;
       $542 = $541 & 255;
       $543 = $540 << $542;
       $544 = $543 | $537;
       $545 = $532 >>> 16;
       $546 = $545 & 255;
       $547 = $531 >>> $546;
       $548 = $535 >>> 16;
       $549 = $548 & 255;
       $550 = $547 << $549;
       $551 = $544 | $550;
       $552 = $532 >>> 24;
       $553 = $124 >>> $552;
       $554 = $535 >>> 24;
       $555 = $553 << $554;
       $556 = $551 | $555;
       $557 = $556&255;
       HEAP8[$dst$4>>0] = $557;
       break L63;
       break;
      }
      case 2:  {
       $558 = HEAP32[$42>>2]|0;
       $559 = $558 & 255;
       $560 = $507 >>> $559;
       $561 = HEAP32[$43>>2]|0;
       $562 = $561 & 255;
       $563 = $560 << $562;
       $564 = $558 >>> 8;
       $565 = $564 & 255;
       $566 = $519 >>> $565;
       $567 = $561 >>> 8;
       $568 = $567 & 255;
       $569 = $566 << $568;
       $570 = $569 | $563;
       $571 = $558 >>> 16;
       $572 = $571 & 255;
       $573 = $531 >>> $572;
       $574 = $561 >>> 16;
       $575 = $574 & 255;
       $576 = $573 << $575;
       $577 = $570 | $576;
       $578 = $558 >>> 24;
       $579 = $124 >>> $578;
       $580 = $561 >>> 24;
       $581 = $579 << $580;
       $582 = $577 | $581;
       $583 = $582&65535;
       HEAP16[$dst$4>>1] = $583;
       break L63;
       break;
      }
      case 3:  {
       $584 = HEAP8[$43>>0]|0;
       $div9 = ($584&255) >>> 3;
       $585 = $div9&255;
       $586 = (($dst$4) + ($585)|0);
       HEAP8[$586>>0] = $506;
       $587 = HEAP8[$44>>0]|0;
       $div10 = ($587&255) >>> 3;
       $588 = $div10&255;
       $589 = (($dst$4) + ($588)|0);
       HEAP8[$589>>0] = $518;
       $590 = HEAP8[$45>>0]|0;
       $div11 = ($590&255) >>> 3;
       $591 = $div11&255;
       $592 = (($dst$4) + ($591)|0);
       HEAP8[$592>>0] = $530;
       break L63;
       break;
      }
      case 4:  {
       $593 = HEAP32[$42>>2]|0;
       $594 = $593 & 255;
       $595 = $507 >>> $594;
       $596 = HEAP32[$43>>2]|0;
       $597 = $596 & 255;
       $598 = $595 << $597;
       $599 = $593 >>> 8;
       $600 = $599 & 255;
       $601 = $519 >>> $600;
       $602 = $596 >>> 8;
       $603 = $602 & 255;
       $604 = $601 << $603;
       $605 = $604 | $598;
       $606 = $593 >>> 16;
       $607 = $606 & 255;
       $608 = $531 >>> $607;
       $609 = $596 >>> 16;
       $610 = $609 & 255;
       $611 = $608 << $610;
       $612 = $605 | $611;
       $613 = $593 >>> 24;
       $614 = $124 >>> $613;
       $615 = $596 >>> 24;
       $616 = $614 << $615;
       $617 = $612 | $616;
       HEAP32[$dst$4>>2] = $617;
       break L63;
       break;
      }
      default: {
       break L63;
      }
      }
     }
    } while(0);
    $618 = (($dst$4) + ($22)|0);
    $619 = (($src$4) + ($19)|0);
    $dst$5 = $618;$n$4 = $n$3;$src$5 = $619;
    label = 54;
    continue;
   }
   else if ((label|0) == 54) {
    label = 0;
    switch ($19|0) {
    case 1:  {
     $620 = HEAP8[$src$5>>0]|0;
     $621 = $620&255;
     $Pixel27$0 = $621;
     break;
    }
    case 2:  {
     $622 = HEAP16[$src$5>>1]|0;
     $623 = $622&65535;
     $Pixel27$0 = $623;
     break;
    }
    case 3:  {
     $624 = HEAP8[$src$5>>0]|0;
     $625 = $624&255;
     $626 = ((($src$5)) + 1|0);
     $627 = HEAP8[$626>>0]|0;
     $628 = $627&255;
     $629 = $628 << 8;
     $630 = $629 | $625;
     $631 = ((($src$5)) + 2|0);
     $632 = HEAP8[$631>>0]|0;
     $633 = $632&255;
     $634 = $633 << 16;
     $635 = $630 | $634;
     $Pixel27$0 = $635;
     break;
    }
    case 4:  {
     $636 = HEAP32[$src$5>>2]|0;
     $Pixel27$0 = $636;
     break;
    }
    default: {
     $Pixel27$0 = 0;
    }
    }
    $637 = $Pixel27$0 & $31;
    $638 = ($637|0)==($32|0);
    L78: do {
     if (!($638)) {
      $639 = HEAP32[$37>>2]|0;
      $640 = $639 & $Pixel27$0;
      $641 = HEAP32[$38>>2]|0;
      $642 = $641 & 255;
      $643 = $640 >>> $642;
      $644 = HEAP32[$39>>2]|0;
      $645 = $644 & 255;
      $646 = (125924 + ($645<<2)|0);
      $647 = HEAP32[$646>>2]|0;
      $648 = (($647) + ($643)|0);
      $649 = HEAP8[$648>>0]|0;
      $650 = $649&255;
      $651 = HEAP32[$40>>2]|0;
      $652 = $651 & $Pixel27$0;
      $653 = $641 >>> 8;
      $654 = $653 & 255;
      $655 = $652 >>> $654;
      $656 = $644 >>> 8;
      $657 = $656 & 255;
      $658 = (125924 + ($657<<2)|0);
      $659 = HEAP32[$658>>2]|0;
      $660 = (($659) + ($655)|0);
      $661 = HEAP8[$660>>0]|0;
      $662 = $661&255;
      $663 = HEAP32[$41>>2]|0;
      $664 = $663 & $Pixel27$0;
      $665 = $641 >>> 16;
      $666 = $665 & 255;
      $667 = $664 >>> $666;
      $668 = $644 >>> 16;
      $669 = $668 & 255;
      $670 = (125924 + ($669<<2)|0);
      $671 = HEAP32[$670>>2]|0;
      $672 = (($671) + ($667)|0);
      $673 = HEAP8[$672>>0]|0;
      $674 = $673&255;
      switch ($22|0) {
      case 1:  {
       $675 = HEAP32[$42>>2]|0;
       $676 = $675 & 255;
       $677 = $650 >>> $676;
       $678 = HEAP32[$43>>2]|0;
       $679 = $678 & 255;
       $680 = $677 << $679;
       $681 = $675 >>> 8;
       $682 = $681 & 255;
       $683 = $662 >>> $682;
       $684 = $678 >>> 8;
       $685 = $684 & 255;
       $686 = $683 << $685;
       $687 = $686 | $680;
       $688 = $675 >>> 16;
       $689 = $688 & 255;
       $690 = $674 >>> $689;
       $691 = $678 >>> 16;
       $692 = $691 & 255;
       $693 = $690 << $692;
       $694 = $687 | $693;
       $695 = $675 >>> 24;
       $696 = $124 >>> $695;
       $697 = $678 >>> 24;
       $698 = $696 << $697;
       $699 = $694 | $698;
       $700 = $699&255;
       HEAP8[$dst$5>>0] = $700;
       break L78;
       break;
      }
      case 2:  {
       $701 = HEAP32[$42>>2]|0;
       $702 = $701 & 255;
       $703 = $650 >>> $702;
       $704 = HEAP32[$43>>2]|0;
       $705 = $704 & 255;
       $706 = $703 << $705;
       $707 = $701 >>> 8;
       $708 = $707 & 255;
       $709 = $662 >>> $708;
       $710 = $704 >>> 8;
       $711 = $710 & 255;
       $712 = $709 << $711;
       $713 = $712 | $706;
       $714 = $701 >>> 16;
       $715 = $714 & 255;
       $716 = $674 >>> $715;
       $717 = $704 >>> 16;
       $718 = $717 & 255;
       $719 = $716 << $718;
       $720 = $713 | $719;
       $721 = $701 >>> 24;
       $722 = $124 >>> $721;
       $723 = $704 >>> 24;
       $724 = $722 << $723;
       $725 = $720 | $724;
       $726 = $725&65535;
       HEAP16[$dst$5>>1] = $726;
       break L78;
       break;
      }
      case 3:  {
       $727 = HEAP8[$43>>0]|0;
       $div6 = ($727&255) >>> 3;
       $728 = $div6&255;
       $729 = (($dst$5) + ($728)|0);
       HEAP8[$729>>0] = $649;
       $730 = HEAP8[$44>>0]|0;
       $div7 = ($730&255) >>> 3;
       $731 = $div7&255;
       $732 = (($dst$5) + ($731)|0);
       HEAP8[$732>>0] = $661;
       $733 = HEAP8[$45>>0]|0;
       $div8 = ($733&255) >>> 3;
       $734 = $div8&255;
       $735 = (($dst$5) + ($734)|0);
       HEAP8[$735>>0] = $673;
       break L78;
       break;
      }
      case 4:  {
       $736 = HEAP32[$42>>2]|0;
       $737 = $736 & 255;
       $738 = $650 >>> $737;
       $739 = HEAP32[$43>>2]|0;
       $740 = $739 & 255;
       $741 = $738 << $740;
       $742 = $736 >>> 8;
       $743 = $742 & 255;
       $744 = $662 >>> $743;
       $745 = $739 >>> 8;
       $746 = $745 & 255;
       $747 = $744 << $746;
       $748 = $747 | $741;
       $749 = $736 >>> 16;
       $750 = $749 & 255;
       $751 = $674 >>> $750;
       $752 = $739 >>> 16;
       $753 = $752 & 255;
       $754 = $751 << $753;
       $755 = $748 | $754;
       $756 = $736 >>> 24;
       $757 = $124 >>> $756;
       $758 = $739 >>> 24;
       $759 = $757 << $758;
       $760 = $755 | $759;
       HEAP32[$dst$5>>2] = $760;
       break L78;
       break;
      }
      default: {
       break L78;
      }
      }
     }
    } while(0);
    $761 = (($dst$5) + ($22)|0);
    $762 = (($src$5) + ($19)|0);
    $dst$6 = $761;$n$5 = $n$4;$src$6 = $762;
    label = 66;
    continue;
   }
   else if ((label|0) == 66) {
    label = 0;
    switch ($19|0) {
    case 1:  {
     $763 = HEAP8[$src$6>>0]|0;
     $764 = $763&255;
     $Pixel35$0 = $764;
     break;
    }
    case 2:  {
     $765 = HEAP16[$src$6>>1]|0;
     $766 = $765&65535;
     $Pixel35$0 = $766;
     break;
    }
    case 3:  {
     $767 = HEAP8[$src$6>>0]|0;
     $768 = $767&255;
     $769 = ((($src$6)) + 1|0);
     $770 = HEAP8[$769>>0]|0;
     $771 = $770&255;
     $772 = $771 << 8;
     $773 = $772 | $768;
     $774 = ((($src$6)) + 2|0);
     $775 = HEAP8[$774>>0]|0;
     $776 = $775&255;
     $777 = $776 << 16;
     $778 = $773 | $777;
     $Pixel35$0 = $778;
     break;
    }
    case 4:  {
     $779 = HEAP32[$src$6>>2]|0;
     $Pixel35$0 = $779;
     break;
    }
    default: {
     $Pixel35$0 = 0;
    }
    }
    $780 = $Pixel35$0 & $31;
    $781 = ($780|0)==($32|0);
    L93: do {
     if (!($781)) {
      $782 = HEAP32[$37>>2]|0;
      $783 = $782 & $Pixel35$0;
      $784 = HEAP32[$38>>2]|0;
      $785 = $784 & 255;
      $786 = $783 >>> $785;
      $787 = HEAP32[$39>>2]|0;
      $788 = $787 & 255;
      $789 = (125924 + ($788<<2)|0);
      $790 = HEAP32[$789>>2]|0;
      $791 = (($790) + ($786)|0);
      $792 = HEAP8[$791>>0]|0;
      $793 = $792&255;
      $794 = HEAP32[$40>>2]|0;
      $795 = $794 & $Pixel35$0;
      $796 = $784 >>> 8;
      $797 = $796 & 255;
      $798 = $795 >>> $797;
      $799 = $787 >>> 8;
      $800 = $799 & 255;
      $801 = (125924 + ($800<<2)|0);
      $802 = HEAP32[$801>>2]|0;
      $803 = (($802) + ($798)|0);
      $804 = HEAP8[$803>>0]|0;
      $805 = $804&255;
      $806 = HEAP32[$41>>2]|0;
      $807 = $806 & $Pixel35$0;
      $808 = $784 >>> 16;
      $809 = $808 & 255;
      $810 = $807 >>> $809;
      $811 = $787 >>> 16;
      $812 = $811 & 255;
      $813 = (125924 + ($812<<2)|0);
      $814 = HEAP32[$813>>2]|0;
      $815 = (($814) + ($810)|0);
      $816 = HEAP8[$815>>0]|0;
      $817 = $816&255;
      switch ($22|0) {
      case 1:  {
       $818 = HEAP32[$42>>2]|0;
       $819 = $818 & 255;
       $820 = $793 >>> $819;
       $821 = HEAP32[$43>>2]|0;
       $822 = $821 & 255;
       $823 = $820 << $822;
       $824 = $818 >>> 8;
       $825 = $824 & 255;
       $826 = $805 >>> $825;
       $827 = $821 >>> 8;
       $828 = $827 & 255;
       $829 = $826 << $828;
       $830 = $829 | $823;
       $831 = $818 >>> 16;
       $832 = $831 & 255;
       $833 = $817 >>> $832;
       $834 = $821 >>> 16;
       $835 = $834 & 255;
       $836 = $833 << $835;
       $837 = $830 | $836;
       $838 = $818 >>> 24;
       $839 = $124 >>> $838;
       $840 = $821 >>> 24;
       $841 = $839 << $840;
       $842 = $837 | $841;
       $843 = $842&255;
       HEAP8[$dst$6>>0] = $843;
       break L93;
       break;
      }
      case 2:  {
       $844 = HEAP32[$42>>2]|0;
       $845 = $844 & 255;
       $846 = $793 >>> $845;
       $847 = HEAP32[$43>>2]|0;
       $848 = $847 & 255;
       $849 = $846 << $848;
       $850 = $844 >>> 8;
       $851 = $850 & 255;
       $852 = $805 >>> $851;
       $853 = $847 >>> 8;
       $854 = $853 & 255;
       $855 = $852 << $854;
       $856 = $855 | $849;
       $857 = $844 >>> 16;
       $858 = $857 & 255;
       $859 = $817 >>> $858;
       $860 = $847 >>> 16;
       $861 = $860 & 255;
       $862 = $859 << $861;
       $863 = $856 | $862;
       $864 = $844 >>> 24;
       $865 = $124 >>> $864;
       $866 = $847 >>> 24;
       $867 = $865 << $866;
       $868 = $863 | $867;
       $869 = $868&65535;
       HEAP16[$dst$6>>1] = $869;
       break L93;
       break;
      }
      case 3:  {
       $870 = HEAP8[$43>>0]|0;
       $div3 = ($870&255) >>> 3;
       $871 = $div3&255;
       $872 = (($dst$6) + ($871)|0);
       HEAP8[$872>>0] = $792;
       $873 = HEAP8[$44>>0]|0;
       $div4 = ($873&255) >>> 3;
       $874 = $div4&255;
       $875 = (($dst$6) + ($874)|0);
       HEAP8[$875>>0] = $804;
       $876 = HEAP8[$45>>0]|0;
       $div5 = ($876&255) >>> 3;
       $877 = $div5&255;
       $878 = (($dst$6) + ($877)|0);
       HEAP8[$878>>0] = $816;
       break L93;
       break;
      }
      case 4:  {
       $879 = HEAP32[$42>>2]|0;
       $880 = $879 & 255;
       $881 = $793 >>> $880;
       $882 = HEAP32[$43>>2]|0;
       $883 = $882 & 255;
       $884 = $881 << $883;
       $885 = $879 >>> 8;
       $886 = $885 & 255;
       $887 = $805 >>> $886;
       $888 = $882 >>> 8;
       $889 = $888 & 255;
       $890 = $887 << $889;
       $891 = $890 | $884;
       $892 = $879 >>> 16;
       $893 = $892 & 255;
       $894 = $817 >>> $893;
       $895 = $882 >>> 16;
       $896 = $895 & 255;
       $897 = $894 << $896;
       $898 = $891 | $897;
       $899 = $879 >>> 24;
       $900 = $124 >>> $899;
       $901 = $882 >>> 24;
       $902 = $900 << $901;
       $903 = $898 | $902;
       HEAP32[$dst$6>>2] = $903;
       break L93;
       break;
      }
      default: {
       break L93;
      }
      }
     }
    } while(0);
    $904 = (($dst$6) + ($22)|0);
    $905 = (($src$6) + ($19)|0);
    $dst$7 = $904;$n$6 = $n$5;$src$7 = $905;
    label = 78;
    continue;
   }
   else if ((label|0) == 78) {
    label = 0;
    switch ($19|0) {
    case 1:  {
     $906 = HEAP8[$src$7>>0]|0;
     $907 = $906&255;
     $Pixel43$0 = $907;
     break;
    }
    case 2:  {
     $908 = HEAP16[$src$7>>1]|0;
     $909 = $908&65535;
     $Pixel43$0 = $909;
     break;
    }
    case 3:  {
     $910 = HEAP8[$src$7>>0]|0;
     $911 = $910&255;
     $912 = ((($src$7)) + 1|0);
     $913 = HEAP8[$912>>0]|0;
     $914 = $913&255;
     $915 = $914 << 8;
     $916 = $915 | $911;
     $917 = ((($src$7)) + 2|0);
     $918 = HEAP8[$917>>0]|0;
     $919 = $918&255;
     $920 = $919 << 16;
     $921 = $916 | $920;
     $Pixel43$0 = $921;
     break;
    }
    case 4:  {
     $922 = HEAP32[$src$7>>2]|0;
     $Pixel43$0 = $922;
     break;
    }
    default: {
     $Pixel43$0 = 0;
    }
    }
    $923 = $Pixel43$0 & $31;
    $924 = ($923|0)==($32|0);
    L108: do {
     if (!($924)) {
      $925 = HEAP32[$37>>2]|0;
      $926 = $925 & $Pixel43$0;
      $927 = HEAP32[$38>>2]|0;
      $928 = $927 & 255;
      $929 = $926 >>> $928;
      $930 = HEAP32[$39>>2]|0;
      $931 = $930 & 255;
      $932 = (125924 + ($931<<2)|0);
      $933 = HEAP32[$932>>2]|0;
      $934 = (($933) + ($929)|0);
      $935 = HEAP8[$934>>0]|0;
      $936 = $935&255;
      $937 = HEAP32[$40>>2]|0;
      $938 = $937 & $Pixel43$0;
      $939 = $927 >>> 8;
      $940 = $939 & 255;
      $941 = $938 >>> $940;
      $942 = $930 >>> 8;
      $943 = $942 & 255;
      $944 = (125924 + ($943<<2)|0);
      $945 = HEAP32[$944>>2]|0;
      $946 = (($945) + ($941)|0);
      $947 = HEAP8[$946>>0]|0;
      $948 = $947&255;
      $949 = HEAP32[$41>>2]|0;
      $950 = $949 & $Pixel43$0;
      $951 = $927 >>> 16;
      $952 = $951 & 255;
      $953 = $950 >>> $952;
      $954 = $930 >>> 16;
      $955 = $954 & 255;
      $956 = (125924 + ($955<<2)|0);
      $957 = HEAP32[$956>>2]|0;
      $958 = (($957) + ($953)|0);
      $959 = HEAP8[$958>>0]|0;
      $960 = $959&255;
      switch ($22|0) {
      case 1:  {
       $961 = HEAP32[$42>>2]|0;
       $962 = $961 & 255;
       $963 = $936 >>> $962;
       $964 = HEAP32[$43>>2]|0;
       $965 = $964 & 255;
       $966 = $963 << $965;
       $967 = $961 >>> 8;
       $968 = $967 & 255;
       $969 = $948 >>> $968;
       $970 = $964 >>> 8;
       $971 = $970 & 255;
       $972 = $969 << $971;
       $973 = $972 | $966;
       $974 = $961 >>> 16;
       $975 = $974 & 255;
       $976 = $960 >>> $975;
       $977 = $964 >>> 16;
       $978 = $977 & 255;
       $979 = $976 << $978;
       $980 = $973 | $979;
       $981 = $961 >>> 24;
       $982 = $124 >>> $981;
       $983 = $964 >>> 24;
       $984 = $982 << $983;
       $985 = $980 | $984;
       $986 = $985&255;
       HEAP8[$dst$7>>0] = $986;
       break L108;
       break;
      }
      case 2:  {
       $987 = HEAP32[$42>>2]|0;
       $988 = $987 & 255;
       $989 = $936 >>> $988;
       $990 = HEAP32[$43>>2]|0;
       $991 = $990 & 255;
       $992 = $989 << $991;
       $993 = $987 >>> 8;
       $994 = $993 & 255;
       $995 = $948 >>> $994;
       $996 = $990 >>> 8;
       $997 = $996 & 255;
       $998 = $995 << $997;
       $999 = $998 | $992;
       $1000 = $987 >>> 16;
       $1001 = $1000 & 255;
       $1002 = $960 >>> $1001;
       $1003 = $990 >>> 16;
       $1004 = $1003 & 255;
       $1005 = $1002 << $1004;
       $1006 = $999 | $1005;
       $1007 = $987 >>> 24;
       $1008 = $124 >>> $1007;
       $1009 = $990 >>> 24;
       $1010 = $1008 << $1009;
       $1011 = $1006 | $1010;
       $1012 = $1011&65535;
       HEAP16[$dst$7>>1] = $1012;
       break L108;
       break;
      }
      case 3:  {
       $1013 = HEAP8[$43>>0]|0;
       $div = ($1013&255) >>> 3;
       $1014 = $div&255;
       $1015 = (($dst$7) + ($1014)|0);
       HEAP8[$1015>>0] = $935;
       $1016 = HEAP8[$44>>0]|0;
       $div1 = ($1016&255) >>> 3;
       $1017 = $div1&255;
       $1018 = (($dst$7) + ($1017)|0);
       HEAP8[$1018>>0] = $947;
       $1019 = HEAP8[$45>>0]|0;
       $div2 = ($1019&255) >>> 3;
       $1020 = $div2&255;
       $1021 = (($dst$7) + ($1020)|0);
       HEAP8[$1021>>0] = $959;
       break L108;
       break;
      }
      case 4:  {
       $1022 = HEAP32[$42>>2]|0;
       $1023 = $1022 & 255;
       $1024 = $936 >>> $1023;
       $1025 = HEAP32[$43>>2]|0;
       $1026 = $1025 & 255;
       $1027 = $1024 << $1026;
       $1028 = $1022 >>> 8;
       $1029 = $1028 & 255;
       $1030 = $948 >>> $1029;
       $1031 = $1025 >>> 8;
       $1032 = $1031 & 255;
       $1033 = $1030 << $1032;
       $1034 = $1033 | $1027;
       $1035 = $1022 >>> 16;
       $1036 = $1035 & 255;
       $1037 = $960 >>> $1036;
       $1038 = $1025 >>> 16;
       $1039 = $1038 & 255;
       $1040 = $1037 << $1039;
       $1041 = $1034 | $1040;
       $1042 = $1022 >>> 24;
       $1043 = $124 >>> $1042;
       $1044 = $1025 >>> 24;
       $1045 = $1043 << $1044;
       $1046 = $1041 | $1045;
       HEAP32[$dst$7>>2] = $1046;
       break L108;
       break;
      }
      default: {
       break L108;
      }
      }
     }
    } while(0);
    $1047 = (($dst$7) + ($22)|0);
    $1048 = (($src$7) + ($19)|0);
    $dst$8 = $1047;$n$7 = $n$6;$src$8 = $1048;
    label = 90;
    continue;
   }
   else if ((label|0) == 90) {
    label = 0;
    switch ($19|0) {
    case 1:  {
     $1049 = HEAP8[$src$8>>0]|0;
     $1050 = $1049&255;
     $Pixel51$0 = $1050;
     break;
    }
    case 2:  {
     $1051 = HEAP16[$src$8>>1]|0;
     $1052 = $1051&65535;
     $Pixel51$0 = $1052;
     break;
    }
    case 3:  {
     $1053 = HEAP8[$src$8>>0]|0;
     $1054 = $1053&255;
     $1055 = ((($src$8)) + 1|0);
     $1056 = HEAP8[$1055>>0]|0;
     $1057 = $1056&255;
     $1058 = $1057 << 8;
     $1059 = $1058 | $1054;
     $1060 = ((($src$8)) + 2|0);
     $1061 = HEAP8[$1060>>0]|0;
     $1062 = $1061&255;
     $1063 = $1062 << 16;
     $1064 = $1059 | $1063;
     $Pixel51$0 = $1064;
     break;
    }
    case 4:  {
     $1065 = HEAP32[$src$8>>2]|0;
     $Pixel51$0 = $1065;
     break;
    }
    default: {
     $Pixel51$0 = 0;
    }
    }
    $1066 = $Pixel51$0 & $31;
    $1067 = ($1066|0)==($32|0);
    L123: do {
     if (!($1067)) {
      $1068 = HEAP32[$37>>2]|0;
      $1069 = $1068 & $Pixel51$0;
      $1070 = HEAP32[$38>>2]|0;
      $1071 = $1070 & 255;
      $1072 = $1069 >>> $1071;
      $1073 = HEAP32[$39>>2]|0;
      $1074 = $1073 & 255;
      $1075 = (125924 + ($1074<<2)|0);
      $1076 = HEAP32[$1075>>2]|0;
      $1077 = (($1076) + ($1072)|0);
      $1078 = HEAP8[$1077>>0]|0;
      $1079 = $1078&255;
      $1080 = HEAP32[$40>>2]|0;
      $1081 = $1080 & $Pixel51$0;
      $1082 = $1070 >>> 8;
      $1083 = $1082 & 255;
      $1084 = $1081 >>> $1083;
      $1085 = $1073 >>> 8;
      $1086 = $1085 & 255;
      $1087 = (125924 + ($1086<<2)|0);
      $1088 = HEAP32[$1087>>2]|0;
      $1089 = (($1088) + ($1084)|0);
      $1090 = HEAP8[$1089>>0]|0;
      $1091 = $1090&255;
      $1092 = HEAP32[$41>>2]|0;
      $1093 = $1092 & $Pixel51$0;
      $1094 = $1070 >>> 16;
      $1095 = $1094 & 255;
      $1096 = $1093 >>> $1095;
      $1097 = $1073 >>> 16;
      $1098 = $1097 & 255;
      $1099 = (125924 + ($1098<<2)|0);
      $1100 = HEAP32[$1099>>2]|0;
      $1101 = (($1100) + ($1096)|0);
      $1102 = HEAP8[$1101>>0]|0;
      $1103 = $1102&255;
      switch ($22|0) {
      case 1:  {
       $1104 = HEAP32[$42>>2]|0;
       $1105 = $1104 & 255;
       $1106 = $1079 >>> $1105;
       $1107 = HEAP32[$43>>2]|0;
       $1108 = $1107 & 255;
       $1109 = $1106 << $1108;
       $1110 = $1104 >>> 8;
       $1111 = $1110 & 255;
       $1112 = $1091 >>> $1111;
       $1113 = $1107 >>> 8;
       $1114 = $1113 & 255;
       $1115 = $1112 << $1114;
       $1116 = $1115 | $1109;
       $1117 = $1104 >>> 16;
       $1118 = $1117 & 255;
       $1119 = $1103 >>> $1118;
       $1120 = $1107 >>> 16;
       $1121 = $1120 & 255;
       $1122 = $1119 << $1121;
       $1123 = $1116 | $1122;
       $1124 = $1104 >>> 24;
       $1125 = $124 >>> $1124;
       $1126 = $1107 >>> 24;
       $1127 = $1125 << $1126;
       $1128 = $1123 | $1127;
       $1129 = $1128&255;
       HEAP8[$dst$8>>0] = $1129;
       break L123;
       break;
      }
      case 2:  {
       $1130 = HEAP32[$42>>2]|0;
       $1131 = $1130 & 255;
       $1132 = $1079 >>> $1131;
       $1133 = HEAP32[$43>>2]|0;
       $1134 = $1133 & 255;
       $1135 = $1132 << $1134;
       $1136 = $1130 >>> 8;
       $1137 = $1136 & 255;
       $1138 = $1091 >>> $1137;
       $1139 = $1133 >>> 8;
       $1140 = $1139 & 255;
       $1141 = $1138 << $1140;
       $1142 = $1141 | $1135;
       $1143 = $1130 >>> 16;
       $1144 = $1143 & 255;
       $1145 = $1103 >>> $1144;
       $1146 = $1133 >>> 16;
       $1147 = $1146 & 255;
       $1148 = $1145 << $1147;
       $1149 = $1142 | $1148;
       $1150 = $1130 >>> 24;
       $1151 = $124 >>> $1150;
       $1152 = $1133 >>> 24;
       $1153 = $1151 << $1152;
       $1154 = $1149 | $1153;
       $1155 = $1154&65535;
       HEAP16[$dst$8>>1] = $1155;
       break L123;
       break;
      }
      case 3:  {
       $1156 = HEAP8[$43>>0]|0;
       $div21 = ($1156&255) >>> 3;
       $1157 = $div21&255;
       $1158 = (($dst$8) + ($1157)|0);
       HEAP8[$1158>>0] = $1078;
       $1159 = HEAP8[$44>>0]|0;
       $div22 = ($1159&255) >>> 3;
       $1160 = $div22&255;
       $1161 = (($dst$8) + ($1160)|0);
       HEAP8[$1161>>0] = $1090;
       $1162 = HEAP8[$45>>0]|0;
       $div23 = ($1162&255) >>> 3;
       $1163 = $div23&255;
       $1164 = (($dst$8) + ($1163)|0);
       HEAP8[$1164>>0] = $1102;
       break L123;
       break;
      }
      case 4:  {
       $1165 = HEAP32[$42>>2]|0;
       $1166 = $1165 & 255;
       $1167 = $1079 >>> $1166;
       $1168 = HEAP32[$43>>2]|0;
       $1169 = $1168 & 255;
       $1170 = $1167 << $1169;
       $1171 = $1165 >>> 8;
       $1172 = $1171 & 255;
       $1173 = $1091 >>> $1172;
       $1174 = $1168 >>> 8;
       $1175 = $1174 & 255;
       $1176 = $1173 << $1175;
       $1177 = $1176 | $1170;
       $1178 = $1165 >>> 16;
       $1179 = $1178 & 255;
       $1180 = $1103 >>> $1179;
       $1181 = $1168 >>> 16;
       $1182 = $1181 & 255;
       $1183 = $1180 << $1182;
       $1184 = $1177 | $1183;
       $1185 = $1165 >>> 24;
       $1186 = $124 >>> $1185;
       $1187 = $1168 >>> 24;
       $1188 = $1186 << $1187;
       $1189 = $1184 | $1188;
       HEAP32[$dst$8>>2] = $1189;
       break L123;
       break;
      }
      default: {
       break L123;
      }
      }
     }
    } while(0);
    $1190 = (($dst$8) + ($22)|0);
    $1191 = (($src$8) + ($19)|0);
    $1192 = (($n$7) + -1)|0;
    $1193 = ($n$7|0)>(1);
    if ($1193) {
     $dst$1 = $1190;$n$0 = $1192;$src$1 = $1191;
     label = 6;
     continue;
    } else {
     break;
    }
   }
  }
  $1194 = (($1191) + ($6)|0);
  $1195 = (($1190) + ($10)|0);
  $1196 = ($46|0)==(0);
  if ($1196) {
   label = 104;
   break;
  } else {
   $$in = $46;$dst$024 = $1195;$src$025 = $1194;
  }
 }
 if ((label|0) == 102) {
  // unreachable;
 }
 else if ((label|0) == 104) {
  return;
 }
}
function _BlitNto1($info) {
 $info = $info|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0;
 var $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0;
 var $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0;
 var $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0;
 var $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0;
 var $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0;
 var $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0;
 var $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0;
 var $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0;
 var $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0;
 var $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0;
 var $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0;
 var $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0;
 var $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0;
 var $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0, $1267 = 0;
 var $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0, $1285 = 0;
 var $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0, $1302 = 0;
 var $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0, $1320 = 0;
 var $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0, $1339 = 0;
 var $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0, $1357 = 0;
 var $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0, $1375 = 0;
 var $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0, $1393 = 0;
 var $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0, $1410 = 0;
 var $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0, $1429 = 0;
 var $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0, $1447 = 0;
 var $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0, $1465 = 0;
 var $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0, $1483 = 0;
 var $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0, $1500 = 0;
 var $1501 = 0, $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0, $1511 = 0, $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0, $1518 = 0, $1519 = 0;
 var $152 = 0, $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0, $153 = 0, $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0, $1536 = 0, $1537 = 0;
 var $1538 = 0, $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0, $1548 = 0, $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0, $1554 = 0, $1555 = 0;
 var $1556 = 0, $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $1565 = 0, $1566 = 0, $1567 = 0, $1568 = 0, $1569 = 0, $157 = 0, $1570 = 0, $1571 = 0, $1572 = 0, $1573 = 0;
 var $1574 = 0, $1575 = 0, $1576 = 0, $1577 = 0, $1578 = 0, $1579 = 0, $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0, $1583 = 0, $1584 = 0, $1585 = 0, $1586 = 0, $1587 = 0, $1588 = 0, $1589 = 0, $159 = 0, $1590 = 0, $1591 = 0;
 var $1592 = 0, $1593 = 0, $1594 = 0, $1595 = 0, $1596 = 0, $1597 = 0, $1598 = 0, $1599 = 0, $16 = 0, $160 = 0, $1600 = 0, $1601 = 0, $1602 = 0, $1603 = 0, $1604 = 0, $1605 = 0, $1606 = 0, $1607 = 0, $1608 = 0, $1609 = 0;
 var $161 = 0, $1610 = 0, $1611 = 0, $1612 = 0, $1613 = 0, $1614 = 0, $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0, $1619 = 0, $162 = 0, $1620 = 0, $1621 = 0, $1622 = 0, $1623 = 0, $1624 = 0, $1625 = 0, $1626 = 0, $1627 = 0;
 var $1628 = 0, $1629 = 0, $163 = 0, $1630 = 0, $1631 = 0, $1632 = 0, $1633 = 0, $1634 = 0, $1635 = 0, $1636 = 0, $1637 = 0, $1638 = 0, $1639 = 0, $164 = 0, $1640 = 0, $1641 = 0, $1642 = 0, $1643 = 0, $1644 = 0, $1645 = 0;
 var $1646 = 0, $1647 = 0, $1648 = 0, $1649 = 0, $165 = 0, $1650 = 0, $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0, $1655 = 0, $1656 = 0, $1657 = 0, $1658 = 0, $1659 = 0, $166 = 0, $1660 = 0, $1661 = 0, $1662 = 0, $1663 = 0;
 var $1664 = 0, $1665 = 0, $1666 = 0, $1667 = 0, $1668 = 0, $1669 = 0, $167 = 0, $1670 = 0, $1671 = 0, $1672 = 0, $1673 = 0, $1674 = 0, $1675 = 0, $1676 = 0, $1677 = 0, $1678 = 0, $1679 = 0, $168 = 0, $1680 = 0, $1681 = 0;
 var $1682 = 0, $1683 = 0, $1684 = 0, $1685 = 0, $1686 = 0, $1687 = 0, $1688 = 0, $1689 = 0, $169 = 0, $1690 = 0, $1691 = 0, $1692 = 0, $1693 = 0, $1694 = 0, $1695 = 0, $1696 = 0, $1697 = 0, $1698 = 0, $1699 = 0, $17 = 0;
 var $170 = 0, $1700 = 0, $1701 = 0, $1702 = 0, $1703 = 0, $1704 = 0, $1705 = 0, $1706 = 0, $1707 = 0, $1708 = 0, $1709 = 0, $171 = 0, $1710 = 0, $1711 = 0, $1712 = 0, $1713 = 0, $1714 = 0, $1715 = 0, $1716 = 0, $1717 = 0;
 var $1718 = 0, $1719 = 0, $172 = 0, $1720 = 0, $1721 = 0, $1722 = 0, $1723 = 0, $1724 = 0, $1725 = 0, $1726 = 0, $1727 = 0, $1728 = 0, $1729 = 0, $173 = 0, $1730 = 0, $1731 = 0, $1732 = 0, $1733 = 0, $1734 = 0, $1735 = 0;
 var $1736 = 0, $1737 = 0, $1738 = 0, $1739 = 0, $174 = 0, $1740 = 0, $1741 = 0, $1742 = 0, $1743 = 0, $1744 = 0, $1745 = 0, $1746 = 0, $1747 = 0, $1748 = 0, $1749 = 0, $175 = 0, $1750 = 0, $1751 = 0, $1752 = 0, $1753 = 0;
 var $1754 = 0, $1755 = 0, $1756 = 0, $1757 = 0, $1758 = 0, $1759 = 0, $176 = 0, $1760 = 0, $1761 = 0, $1762 = 0, $1763 = 0, $1764 = 0, $1765 = 0, $1766 = 0, $1767 = 0, $1768 = 0, $1769 = 0, $177 = 0, $1770 = 0, $1771 = 0;
 var $1772 = 0, $1773 = 0, $1774 = 0, $1775 = 0, $1776 = 0, $1777 = 0, $1778 = 0, $1779 = 0, $178 = 0, $1780 = 0, $1781 = 0, $1782 = 0, $1783 = 0, $1784 = 0, $1785 = 0, $1786 = 0, $1787 = 0, $1788 = 0, $1789 = 0, $179 = 0;
 var $1790 = 0, $1791 = 0, $1792 = 0, $1793 = 0, $1794 = 0, $1795 = 0, $1796 = 0, $1797 = 0, $1798 = 0, $1799 = 0, $18 = 0, $180 = 0, $1800 = 0, $1801 = 0, $1802 = 0, $1803 = 0, $1804 = 0, $1805 = 0, $1806 = 0, $1807 = 0;
 var $1808 = 0, $1809 = 0, $181 = 0, $1810 = 0, $1811 = 0, $1812 = 0, $1813 = 0, $1814 = 0, $1815 = 0, $1816 = 0, $1817 = 0, $1818 = 0, $1819 = 0, $182 = 0, $1820 = 0, $1821 = 0, $1822 = 0, $1823 = 0, $1824 = 0, $1825 = 0;
 var $1826 = 0, $1827 = 0, $1828 = 0, $1829 = 0, $183 = 0, $1830 = 0, $1831 = 0, $1832 = 0, $1833 = 0, $1834 = 0, $1835 = 0, $1836 = 0, $1837 = 0, $1838 = 0, $1839 = 0, $184 = 0, $1840 = 0, $1841 = 0, $1842 = 0, $1843 = 0;
 var $1844 = 0, $1845 = 0, $1846 = 0, $1847 = 0, $1848 = 0, $1849 = 0, $185 = 0, $1850 = 0, $1851 = 0, $1852 = 0, $1853 = 0, $1854 = 0, $1855 = 0, $1856 = 0, $1857 = 0, $1858 = 0, $1859 = 0, $186 = 0, $1860 = 0, $1861 = 0;
 var $1862 = 0, $1863 = 0, $1864 = 0, $1865 = 0, $1866 = 0, $1867 = 0, $1868 = 0, $1869 = 0, $187 = 0, $1870 = 0, $1871 = 0, $1872 = 0, $1873 = 0, $1874 = 0, $1875 = 0, $1876 = 0, $1877 = 0, $1878 = 0, $1879 = 0, $188 = 0;
 var $1880 = 0, $1881 = 0, $1882 = 0, $1883 = 0, $1884 = 0, $1885 = 0, $1886 = 0, $1887 = 0, $1888 = 0, $1889 = 0, $189 = 0, $1890 = 0, $1891 = 0, $1892 = 0, $1893 = 0, $1894 = 0, $1895 = 0, $1896 = 0, $1897 = 0, $1898 = 0;
 var $1899 = 0, $19 = 0, $190 = 0, $1900 = 0, $1901 = 0, $1902 = 0, $1903 = 0, $1904 = 0, $1905 = 0, $1906 = 0, $1907 = 0, $1908 = 0, $1909 = 0, $191 = 0, $1910 = 0, $1911 = 0, $1912 = 0, $1913 = 0, $1914 = 0, $1915 = 0;
 var $1916 = 0, $1917 = 0, $1918 = 0, $1919 = 0, $192 = 0, $1920 = 0, $1921 = 0, $1922 = 0, $1923 = 0, $1924 = 0, $1925 = 0, $1926 = 0, $1927 = 0, $1928 = 0, $1929 = 0, $193 = 0, $1930 = 0, $1931 = 0, $1932 = 0, $1933 = 0;
 var $1934 = 0, $1935 = 0, $1936 = 0, $1937 = 0, $1938 = 0, $1939 = 0, $194 = 0, $1940 = 0, $1941 = 0, $1942 = 0, $1943 = 0, $1944 = 0, $1945 = 0, $1946 = 0, $1947 = 0, $1948 = 0, $1949 = 0, $195 = 0, $1950 = 0, $1951 = 0;
 var $1952 = 0, $1953 = 0, $1954 = 0, $1955 = 0, $1956 = 0, $1957 = 0, $1958 = 0, $1959 = 0, $196 = 0, $1960 = 0, $1961 = 0, $1962 = 0, $1963 = 0, $1964 = 0, $1965 = 0, $1966 = 0, $1967 = 0, $1968 = 0, $1969 = 0, $197 = 0;
 var $1970 = 0, $1971 = 0, $1972 = 0, $1973 = 0, $1974 = 0, $1975 = 0, $1976 = 0, $1977 = 0, $1978 = 0, $1979 = 0, $198 = 0, $1980 = 0, $1981 = 0, $1982 = 0, $1983 = 0, $1984 = 0, $1985 = 0, $1986 = 0, $1987 = 0, $1988 = 0;
 var $1989 = 0, $199 = 0, $1990 = 0, $1991 = 0, $1992 = 0, $1993 = 0, $1994 = 0, $1995 = 0, $1996 = 0, $1997 = 0, $1998 = 0, $1999 = 0, $2 = 0, $20 = 0, $200 = 0, $2000 = 0, $2001 = 0, $2002 = 0, $2003 = 0, $2004 = 0;
 var $2005 = 0, $2006 = 0, $2007 = 0, $2008 = 0, $2009 = 0, $201 = 0, $2010 = 0, $2011 = 0, $2012 = 0, $2013 = 0, $2014 = 0, $2015 = 0, $2016 = 0, $2017 = 0, $2018 = 0, $2019 = 0, $202 = 0, $2020 = 0, $2021 = 0, $2022 = 0;
 var $2023 = 0, $2024 = 0, $2025 = 0, $2026 = 0, $2027 = 0, $2028 = 0, $2029 = 0, $203 = 0, $2030 = 0, $2031 = 0, $2032 = 0, $2033 = 0, $2034 = 0, $2035 = 0, $2036 = 0, $2037 = 0, $2038 = 0, $2039 = 0, $204 = 0, $2040 = 0;
 var $2041 = 0, $2042 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0;
 var $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0;
 var $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0;
 var $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0;
 var $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0;
 var $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0;
 var $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0;
 var $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0;
 var $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0;
 var $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0;
 var $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0;
 var $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0;
 var $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0;
 var $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0;
 var $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0;
 var $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0;
 var $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0;
 var $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0;
 var $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0;
 var $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0;
 var $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0;
 var $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0;
 var $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0;
 var $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0;
 var $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0;
 var $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0;
 var $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0;
 var $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0;
 var $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0;
 var $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0;
 var $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0;
 var $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0;
 var $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0;
 var $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0;
 var $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0;
 var $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0;
 var $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0;
 var $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0;
 var $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0;
 var $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0;
 var $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0;
 var $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0;
 var $997 = 0, $998 = 0, $999 = 0, $div = 0, $div1 = 0, $div10 = 0, $div12 = 0, $div13 = 0, $div15 = 0, $div16 = 0, $div18 = 0, $div19 = 0, $div21 = 0, $div22 = 0, $div24 = 0, $div25 = 0, $div27 = 0, $div28 = 0, $div3 = 0, $div30 = 0;
 var $div31 = 0, $div33 = 0, $div34 = 0, $div36 = 0, $div37 = 0, $div39 = 0, $div4 = 0, $div40 = 0, $div42 = 0, $div43 = 0, $div45 = 0, $div46 = 0, $div6 = 0, $div7 = 0, $div9 = 0, $dst$050 = 0, $dst$1 = 0, $dst$10 = 0, $dst$11 = 0, $dst$12 = 0;
 var $dst$13 = 0, $dst$14 = 0, $dst$15 = 0, $dst$16 = 0, $dst$17 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $dst$952 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0;
 var $n$7 = 0, $n1$0 = 0, $n1$1 = 0, $n1$2 = 0, $n1$3 = 0, $n1$4 = 0, $n1$5 = 0, $n1$6 = 0, $n1$7 = 0, $sB$0 = 0, $sB$0$shrunk = 0, $sB$1 = 0, $sB$1$shrunk = 0, $sB$10 = 0, $sB$10$shrunk = 0, $sB$11 = 0, $sB$11$shrunk = 0, $sB$12 = 0, $sB$12$shrunk = 0, $sB$13 = 0;
 var $sB$13$shrunk = 0, $sB$14 = 0, $sB$14$shrunk = 0, $sB$15 = 0, $sB$15$shrunk = 0, $sB$2 = 0, $sB$2$shrunk = 0, $sB$3 = 0, $sB$3$shrunk = 0, $sB$4 = 0, $sB$4$shrunk = 0, $sB$5 = 0, $sB$5$shrunk = 0, $sB$6 = 0, $sB$6$shrunk = 0, $sB$7 = 0, $sB$7$shrunk = 0, $sB$8 = 0, $sB$8$shrunk = 0, $sB$9 = 0;
 var $sB$9$shrunk = 0, $sG$0 = 0, $sG$0$shrunk = 0, $sG$1 = 0, $sG$1$shrunk = 0, $sG$10 = 0, $sG$10$shrunk = 0, $sG$11 = 0, $sG$11$shrunk = 0, $sG$12 = 0, $sG$12$shrunk = 0, $sG$13 = 0, $sG$13$shrunk = 0, $sG$14 = 0, $sG$14$shrunk = 0, $sG$15 = 0, $sG$15$shrunk = 0, $sG$2 = 0, $sG$2$shrunk = 0, $sG$3 = 0;
 var $sG$3$shrunk = 0, $sG$4 = 0, $sG$4$shrunk = 0, $sG$5 = 0, $sG$5$shrunk = 0, $sG$6 = 0, $sG$6$shrunk = 0, $sG$7 = 0, $sG$7$shrunk = 0, $sG$8 = 0, $sG$8$shrunk = 0, $sG$9 = 0, $sG$9$shrunk = 0, $sR$0 = 0, $sR$0$shrunk = 0, $sR$1 = 0, $sR$1$shrunk = 0, $sR$10 = 0, $sR$10$shrunk = 0, $sR$11 = 0;
 var $sR$11$shrunk = 0, $sR$12 = 0, $sR$12$shrunk = 0, $sR$13 = 0, $sR$13$shrunk = 0, $sR$14 = 0, $sR$14$shrunk = 0, $sR$15 = 0, $sR$15$shrunk = 0, $sR$2 = 0, $sR$2$shrunk = 0, $sR$3 = 0, $sR$3$shrunk = 0, $sR$4 = 0, $sR$4$shrunk = 0, $sR$5 = 0, $sR$5$shrunk = 0, $sR$6 = 0, $sR$6$shrunk = 0, $sR$7 = 0;
 var $sR$7$shrunk = 0, $sR$8 = 0, $sR$8$shrunk = 0, $sR$9 = 0, $sR$9$shrunk = 0, $src$051 = 0, $src$1 = 0, $src$10 = 0, $src$11 = 0, $src$12 = 0, $src$13 = 0, $src$14 = 0, $src$15 = 0, $src$16 = 0, $src$17 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0;
 var $src$7 = 0, $src$8 = 0, $src$953 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$info>>2]|0;
 $5 = ((($info)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($info)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($info)) + 36|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($info)) + 48|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($info)) + 40|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($14)) + 9|0);
 $16 = HEAP8[$15>>0]|0;
 $17 = $16&255;
 $18 = ($12|0)==(0|0);
 $19 = (($3) + -1)|0;
 $20 = ($3|0)==(0);
 if ($18) {
  if ($20) {
   return;
  }
  $29 = (($1) + 7)|0;
  $30 = (($29|0) / 8)&-1;
  $31 = $1 & 7;
  $32 = ((($14)) + 12|0);
  $33 = ((($14)) + 32|0);
  $34 = ((($14)) + 28|0);
  $35 = ((($14)) + 16|0);
  $36 = ((($14)) + 20|0);
  $1034 = $19;$dst$050 = $8;$src$051 = $4;
  L6: while(1) {
   switch ($31|0) {
   case 0:  {
    $dst$1 = $dst$050;$n$0 = $30;$src$1 = $src$051;
    label = 7;
    break;
   }
   case 7:  {
    $dst$2 = $dst$050;$n$1 = $30;$src$2 = $src$051;
    label = 13;
    break;
   }
   case 6:  {
    $dst$3 = $dst$050;$n$2 = $30;$src$3 = $src$051;
    label = 19;
    break;
   }
   case 5:  {
    $dst$4 = $dst$050;$n$3 = $30;$src$4 = $src$051;
    label = 25;
    break;
   }
   case 4:  {
    $dst$5 = $dst$050;$n$4 = $30;$src$5 = $src$051;
    label = 31;
    break;
   }
   case 3:  {
    $dst$6 = $dst$050;$n$5 = $30;$src$6 = $src$051;
    label = 37;
    break;
   }
   case 2:  {
    $dst$7 = $dst$050;$n$6 = $30;$src$7 = $src$051;
    label = 43;
    break;
   }
   case 1:  {
    $dst$8 = $dst$050;$n$7 = $30;$src$8 = $src$051;
    label = 49;
    break;
   }
   default: {
    label = 55;
    break L6;
   }
   }
   while(1) {
    if ((label|0) == 7) {
     label = 0;
     switch ($17|0) {
     case 1:  {
      $37 = HEAP8[$src$1>>0]|0;
      $38 = $37&255;
      $39 = HEAP32[$32>>2]|0;
      $40 = $38 & $39;
      $41 = HEAP32[$33>>2]|0;
      $42 = $41 & 255;
      $43 = $40 >>> $42;
      $44 = HEAP32[$34>>2]|0;
      $45 = $44 & 255;
      $46 = (125924 + ($45<<2)|0);
      $47 = HEAP32[$46>>2]|0;
      $48 = (($47) + ($43)|0);
      $49 = HEAP8[$48>>0]|0;
      $50 = HEAP32[$35>>2]|0;
      $51 = $50 & $38;
      $52 = $41 >>> 8;
      $53 = $52 & 255;
      $54 = $51 >>> $53;
      $55 = $44 >>> 8;
      $56 = $55 & 255;
      $57 = (125924 + ($56<<2)|0);
      $58 = HEAP32[$57>>2]|0;
      $59 = (($58) + ($54)|0);
      $60 = HEAP8[$59>>0]|0;
      $61 = HEAP32[$36>>2]|0;
      $62 = $61 & $38;
      $63 = $41 >>> 16;
      $64 = $63 & 255;
      $65 = $62 >>> $64;
      $66 = $44 >>> 16;
      $67 = $66 & 255;
      $68 = (125924 + ($67<<2)|0);
      $69 = HEAP32[$68>>2]|0;
      $70 = (($69) + ($65)|0);
      $71 = HEAP8[$70>>0]|0;
      $sB$0$shrunk = $71;$sG$0$shrunk = $60;$sR$0$shrunk = $49;
      break;
     }
     case 2:  {
      $72 = HEAP16[$src$1>>1]|0;
      $73 = $72&65535;
      $74 = HEAP32[$32>>2]|0;
      $75 = $73 & $74;
      $76 = HEAP32[$33>>2]|0;
      $77 = $76 & 255;
      $78 = $75 >>> $77;
      $79 = HEAP32[$34>>2]|0;
      $80 = $79 & 255;
      $81 = (125924 + ($80<<2)|0);
      $82 = HEAP32[$81>>2]|0;
      $83 = (($82) + ($78)|0);
      $84 = HEAP8[$83>>0]|0;
      $85 = HEAP32[$35>>2]|0;
      $86 = $85 & $73;
      $87 = $76 >>> 8;
      $88 = $87 & 255;
      $89 = $86 >>> $88;
      $90 = $79 >>> 8;
      $91 = $90 & 255;
      $92 = (125924 + ($91<<2)|0);
      $93 = HEAP32[$92>>2]|0;
      $94 = (($93) + ($89)|0);
      $95 = HEAP8[$94>>0]|0;
      $96 = HEAP32[$36>>2]|0;
      $97 = $96 & $73;
      $98 = $76 >>> 16;
      $99 = $98 & 255;
      $100 = $97 >>> $99;
      $101 = $79 >>> 16;
      $102 = $101 & 255;
      $103 = (125924 + ($102<<2)|0);
      $104 = HEAP32[$103>>2]|0;
      $105 = (($104) + ($100)|0);
      $106 = HEAP8[$105>>0]|0;
      $sB$0$shrunk = $106;$sG$0$shrunk = $95;$sR$0$shrunk = $84;
      break;
     }
     case 3:  {
      $107 = HEAP32[$33>>2]|0;
      $div42 = $107 >>> 3;
      $108 = $div42 & 31;
      $109 = (($src$1) + ($108)|0);
      $110 = HEAP8[$109>>0]|0;
      $111 = $107 >>> 11;
      $div43 = $111 & 31;
      $112 = (($src$1) + ($div43)|0);
      $113 = HEAP8[$112>>0]|0;
      $114 = $107 >>> 19;
      $115 = $114 & 31;
      $116 = (($src$1) + ($115)|0);
      $117 = HEAP8[$116>>0]|0;
      $sB$0$shrunk = $117;$sG$0$shrunk = $113;$sR$0$shrunk = $110;
      break;
     }
     case 4:  {
      $118 = HEAP32[$src$1>>2]|0;
      $119 = HEAP32[$32>>2]|0;
      $120 = $119 & $118;
      $121 = HEAP32[$33>>2]|0;
      $122 = $121 & 255;
      $123 = $120 >>> $122;
      $124 = HEAP32[$34>>2]|0;
      $125 = $124 & 255;
      $126 = (125924 + ($125<<2)|0);
      $127 = HEAP32[$126>>2]|0;
      $128 = (($127) + ($123)|0);
      $129 = HEAP8[$128>>0]|0;
      $130 = HEAP32[$35>>2]|0;
      $131 = $130 & $118;
      $132 = $121 >>> 8;
      $133 = $132 & 255;
      $134 = $131 >>> $133;
      $135 = $124 >>> 8;
      $136 = $135 & 255;
      $137 = (125924 + ($136<<2)|0);
      $138 = HEAP32[$137>>2]|0;
      $139 = (($138) + ($134)|0);
      $140 = HEAP8[$139>>0]|0;
      $141 = HEAP32[$36>>2]|0;
      $142 = $141 & $118;
      $143 = $121 >>> 16;
      $144 = $143 & 255;
      $145 = $142 >>> $144;
      $146 = $124 >>> 16;
      $147 = $146 & 255;
      $148 = (125924 + ($147<<2)|0);
      $149 = HEAP32[$148>>2]|0;
      $150 = (($149) + ($145)|0);
      $151 = HEAP8[$150>>0]|0;
      $sB$0$shrunk = $151;$sG$0$shrunk = $140;$sR$0$shrunk = $129;
      break;
     }
     default: {
      $sB$0$shrunk = 0;$sG$0$shrunk = 0;$sR$0$shrunk = 0;
     }
     }
     $sB$0 = $sB$0$shrunk&255;
     $sG$0 = $sG$0$shrunk&255;
     $sR$0 = $sR$0$shrunk&255;
     $152 = $sR$0 & 224;
     $153 = $sG$0 >>> 5;
     $154 = $153 << 2;
     $155 = $154 | $152;
     $156 = $sB$0 >>> 6;
     $157 = $155 | $156;
     $158 = $157&255;
     HEAP8[$dst$1>>0] = $158;
     $159 = ((($dst$1)) + 1|0);
     $160 = (($src$1) + ($17)|0);
     $dst$2 = $159;$n$1 = $n$0;$src$2 = $160;
     label = 13;
     continue;
    }
    else if ((label|0) == 13) {
     label = 0;
     switch ($17|0) {
     case 1:  {
      $161 = HEAP8[$src$2>>0]|0;
      $162 = $161&255;
      $163 = HEAP32[$32>>2]|0;
      $164 = $162 & $163;
      $165 = HEAP32[$33>>2]|0;
      $166 = $165 & 255;
      $167 = $164 >>> $166;
      $168 = HEAP32[$34>>2]|0;
      $169 = $168 & 255;
      $170 = (125924 + ($169<<2)|0);
      $171 = HEAP32[$170>>2]|0;
      $172 = (($171) + ($167)|0);
      $173 = HEAP8[$172>>0]|0;
      $174 = HEAP32[$35>>2]|0;
      $175 = $174 & $162;
      $176 = $165 >>> 8;
      $177 = $176 & 255;
      $178 = $175 >>> $177;
      $179 = $168 >>> 8;
      $180 = $179 & 255;
      $181 = (125924 + ($180<<2)|0);
      $182 = HEAP32[$181>>2]|0;
      $183 = (($182) + ($178)|0);
      $184 = HEAP8[$183>>0]|0;
      $185 = HEAP32[$36>>2]|0;
      $186 = $185 & $162;
      $187 = $165 >>> 16;
      $188 = $187 & 255;
      $189 = $186 >>> $188;
      $190 = $168 >>> 16;
      $191 = $190 & 255;
      $192 = (125924 + ($191<<2)|0);
      $193 = HEAP32[$192>>2]|0;
      $194 = (($193) + ($189)|0);
      $195 = HEAP8[$194>>0]|0;
      $sB$1$shrunk = $195;$sG$1$shrunk = $184;$sR$1$shrunk = $173;
      break;
     }
     case 2:  {
      $196 = HEAP16[$src$2>>1]|0;
      $197 = $196&65535;
      $198 = HEAP32[$32>>2]|0;
      $199 = $197 & $198;
      $200 = HEAP32[$33>>2]|0;
      $201 = $200 & 255;
      $202 = $199 >>> $201;
      $203 = HEAP32[$34>>2]|0;
      $204 = $203 & 255;
      $205 = (125924 + ($204<<2)|0);
      $206 = HEAP32[$205>>2]|0;
      $207 = (($206) + ($202)|0);
      $208 = HEAP8[$207>>0]|0;
      $209 = HEAP32[$35>>2]|0;
      $210 = $209 & $197;
      $211 = $200 >>> 8;
      $212 = $211 & 255;
      $213 = $210 >>> $212;
      $214 = $203 >>> 8;
      $215 = $214 & 255;
      $216 = (125924 + ($215<<2)|0);
      $217 = HEAP32[$216>>2]|0;
      $218 = (($217) + ($213)|0);
      $219 = HEAP8[$218>>0]|0;
      $220 = HEAP32[$36>>2]|0;
      $221 = $220 & $197;
      $222 = $200 >>> 16;
      $223 = $222 & 255;
      $224 = $221 >>> $223;
      $225 = $203 >>> 16;
      $226 = $225 & 255;
      $227 = (125924 + ($226<<2)|0);
      $228 = HEAP32[$227>>2]|0;
      $229 = (($228) + ($224)|0);
      $230 = HEAP8[$229>>0]|0;
      $sB$1$shrunk = $230;$sG$1$shrunk = $219;$sR$1$shrunk = $208;
      break;
     }
     case 3:  {
      $231 = HEAP32[$33>>2]|0;
      $div39 = $231 >>> 3;
      $232 = $div39 & 31;
      $233 = (($src$2) + ($232)|0);
      $234 = HEAP8[$233>>0]|0;
      $235 = $231 >>> 11;
      $div40 = $235 & 31;
      $236 = (($src$2) + ($div40)|0);
      $237 = HEAP8[$236>>0]|0;
      $238 = $231 >>> 19;
      $239 = $238 & 31;
      $240 = (($src$2) + ($239)|0);
      $241 = HEAP8[$240>>0]|0;
      $sB$1$shrunk = $241;$sG$1$shrunk = $237;$sR$1$shrunk = $234;
      break;
     }
     case 4:  {
      $242 = HEAP32[$src$2>>2]|0;
      $243 = HEAP32[$32>>2]|0;
      $244 = $243 & $242;
      $245 = HEAP32[$33>>2]|0;
      $246 = $245 & 255;
      $247 = $244 >>> $246;
      $248 = HEAP32[$34>>2]|0;
      $249 = $248 & 255;
      $250 = (125924 + ($249<<2)|0);
      $251 = HEAP32[$250>>2]|0;
      $252 = (($251) + ($247)|0);
      $253 = HEAP8[$252>>0]|0;
      $254 = HEAP32[$35>>2]|0;
      $255 = $254 & $242;
      $256 = $245 >>> 8;
      $257 = $256 & 255;
      $258 = $255 >>> $257;
      $259 = $248 >>> 8;
      $260 = $259 & 255;
      $261 = (125924 + ($260<<2)|0);
      $262 = HEAP32[$261>>2]|0;
      $263 = (($262) + ($258)|0);
      $264 = HEAP8[$263>>0]|0;
      $265 = HEAP32[$36>>2]|0;
      $266 = $265 & $242;
      $267 = $245 >>> 16;
      $268 = $267 & 255;
      $269 = $266 >>> $268;
      $270 = $248 >>> 16;
      $271 = $270 & 255;
      $272 = (125924 + ($271<<2)|0);
      $273 = HEAP32[$272>>2]|0;
      $274 = (($273) + ($269)|0);
      $275 = HEAP8[$274>>0]|0;
      $sB$1$shrunk = $275;$sG$1$shrunk = $264;$sR$1$shrunk = $253;
      break;
     }
     default: {
      $sB$1$shrunk = 0;$sG$1$shrunk = 0;$sR$1$shrunk = 0;
     }
     }
     $sB$1 = $sB$1$shrunk&255;
     $sG$1 = $sG$1$shrunk&255;
     $sR$1 = $sR$1$shrunk&255;
     $276 = $sR$1 & 224;
     $277 = $sG$1 >>> 5;
     $278 = $277 << 2;
     $279 = $278 | $276;
     $280 = $sB$1 >>> 6;
     $281 = $279 | $280;
     $282 = $281&255;
     HEAP8[$dst$2>>0] = $282;
     $283 = ((($dst$2)) + 1|0);
     $284 = (($src$2) + ($17)|0);
     $dst$3 = $283;$n$2 = $n$1;$src$3 = $284;
     label = 19;
     continue;
    }
    else if ((label|0) == 19) {
     label = 0;
     switch ($17|0) {
     case 1:  {
      $285 = HEAP8[$src$3>>0]|0;
      $286 = $285&255;
      $287 = HEAP32[$32>>2]|0;
      $288 = $286 & $287;
      $289 = HEAP32[$33>>2]|0;
      $290 = $289 & 255;
      $291 = $288 >>> $290;
      $292 = HEAP32[$34>>2]|0;
      $293 = $292 & 255;
      $294 = (125924 + ($293<<2)|0);
      $295 = HEAP32[$294>>2]|0;
      $296 = (($295) + ($291)|0);
      $297 = HEAP8[$296>>0]|0;
      $298 = HEAP32[$35>>2]|0;
      $299 = $298 & $286;
      $300 = $289 >>> 8;
      $301 = $300 & 255;
      $302 = $299 >>> $301;
      $303 = $292 >>> 8;
      $304 = $303 & 255;
      $305 = (125924 + ($304<<2)|0);
      $306 = HEAP32[$305>>2]|0;
      $307 = (($306) + ($302)|0);
      $308 = HEAP8[$307>>0]|0;
      $309 = HEAP32[$36>>2]|0;
      $310 = $309 & $286;
      $311 = $289 >>> 16;
      $312 = $311 & 255;
      $313 = $310 >>> $312;
      $314 = $292 >>> 16;
      $315 = $314 & 255;
      $316 = (125924 + ($315<<2)|0);
      $317 = HEAP32[$316>>2]|0;
      $318 = (($317) + ($313)|0);
      $319 = HEAP8[$318>>0]|0;
      $sB$2$shrunk = $319;$sG$2$shrunk = $308;$sR$2$shrunk = $297;
      break;
     }
     case 2:  {
      $320 = HEAP16[$src$3>>1]|0;
      $321 = $320&65535;
      $322 = HEAP32[$32>>2]|0;
      $323 = $321 & $322;
      $324 = HEAP32[$33>>2]|0;
      $325 = $324 & 255;
      $326 = $323 >>> $325;
      $327 = HEAP32[$34>>2]|0;
      $328 = $327 & 255;
      $329 = (125924 + ($328<<2)|0);
      $330 = HEAP32[$329>>2]|0;
      $331 = (($330) + ($326)|0);
      $332 = HEAP8[$331>>0]|0;
      $333 = HEAP32[$35>>2]|0;
      $334 = $333 & $321;
      $335 = $324 >>> 8;
      $336 = $335 & 255;
      $337 = $334 >>> $336;
      $338 = $327 >>> 8;
      $339 = $338 & 255;
      $340 = (125924 + ($339<<2)|0);
      $341 = HEAP32[$340>>2]|0;
      $342 = (($341) + ($337)|0);
      $343 = HEAP8[$342>>0]|0;
      $344 = HEAP32[$36>>2]|0;
      $345 = $344 & $321;
      $346 = $324 >>> 16;
      $347 = $346 & 255;
      $348 = $345 >>> $347;
      $349 = $327 >>> 16;
      $350 = $349 & 255;
      $351 = (125924 + ($350<<2)|0);
      $352 = HEAP32[$351>>2]|0;
      $353 = (($352) + ($348)|0);
      $354 = HEAP8[$353>>0]|0;
      $sB$2$shrunk = $354;$sG$2$shrunk = $343;$sR$2$shrunk = $332;
      break;
     }
     case 3:  {
      $355 = HEAP32[$33>>2]|0;
      $div36 = $355 >>> 3;
      $356 = $div36 & 31;
      $357 = (($src$3) + ($356)|0);
      $358 = HEAP8[$357>>0]|0;
      $359 = $355 >>> 11;
      $div37 = $359 & 31;
      $360 = (($src$3) + ($div37)|0);
      $361 = HEAP8[$360>>0]|0;
      $362 = $355 >>> 19;
      $363 = $362 & 31;
      $364 = (($src$3) + ($363)|0);
      $365 = HEAP8[$364>>0]|0;
      $sB$2$shrunk = $365;$sG$2$shrunk = $361;$sR$2$shrunk = $358;
      break;
     }
     case 4:  {
      $366 = HEAP32[$src$3>>2]|0;
      $367 = HEAP32[$32>>2]|0;
      $368 = $367 & $366;
      $369 = HEAP32[$33>>2]|0;
      $370 = $369 & 255;
      $371 = $368 >>> $370;
      $372 = HEAP32[$34>>2]|0;
      $373 = $372 & 255;
      $374 = (125924 + ($373<<2)|0);
      $375 = HEAP32[$374>>2]|0;
      $376 = (($375) + ($371)|0);
      $377 = HEAP8[$376>>0]|0;
      $378 = HEAP32[$35>>2]|0;
      $379 = $378 & $366;
      $380 = $369 >>> 8;
      $381 = $380 & 255;
      $382 = $379 >>> $381;
      $383 = $372 >>> 8;
      $384 = $383 & 255;
      $385 = (125924 + ($384<<2)|0);
      $386 = HEAP32[$385>>2]|0;
      $387 = (($386) + ($382)|0);
      $388 = HEAP8[$387>>0]|0;
      $389 = HEAP32[$36>>2]|0;
      $390 = $389 & $366;
      $391 = $369 >>> 16;
      $392 = $391 & 255;
      $393 = $390 >>> $392;
      $394 = $372 >>> 16;
      $395 = $394 & 255;
      $396 = (125924 + ($395<<2)|0);
      $397 = HEAP32[$396>>2]|0;
      $398 = (($397) + ($393)|0);
      $399 = HEAP8[$398>>0]|0;
      $sB$2$shrunk = $399;$sG$2$shrunk = $388;$sR$2$shrunk = $377;
      break;
     }
     default: {
      $sB$2$shrunk = 0;$sG$2$shrunk = 0;$sR$2$shrunk = 0;
     }
     }
     $sB$2 = $sB$2$shrunk&255;
     $sG$2 = $sG$2$shrunk&255;
     $sR$2 = $sR$2$shrunk&255;
     $400 = $sR$2 & 224;
     $401 = $sG$2 >>> 5;
     $402 = $401 << 2;
     $403 = $402 | $400;
     $404 = $sB$2 >>> 6;
     $405 = $403 | $404;
     $406 = $405&255;
     HEAP8[$dst$3>>0] = $406;
     $407 = ((($dst$3)) + 1|0);
     $408 = (($src$3) + ($17)|0);
     $dst$4 = $407;$n$3 = $n$2;$src$4 = $408;
     label = 25;
     continue;
    }
    else if ((label|0) == 25) {
     label = 0;
     switch ($17|0) {
     case 1:  {
      $409 = HEAP8[$src$4>>0]|0;
      $410 = $409&255;
      $411 = HEAP32[$32>>2]|0;
      $412 = $410 & $411;
      $413 = HEAP32[$33>>2]|0;
      $414 = $413 & 255;
      $415 = $412 >>> $414;
      $416 = HEAP32[$34>>2]|0;
      $417 = $416 & 255;
      $418 = (125924 + ($417<<2)|0);
      $419 = HEAP32[$418>>2]|0;
      $420 = (($419) + ($415)|0);
      $421 = HEAP8[$420>>0]|0;
      $422 = HEAP32[$35>>2]|0;
      $423 = $422 & $410;
      $424 = $413 >>> 8;
      $425 = $424 & 255;
      $426 = $423 >>> $425;
      $427 = $416 >>> 8;
      $428 = $427 & 255;
      $429 = (125924 + ($428<<2)|0);
      $430 = HEAP32[$429>>2]|0;
      $431 = (($430) + ($426)|0);
      $432 = HEAP8[$431>>0]|0;
      $433 = HEAP32[$36>>2]|0;
      $434 = $433 & $410;
      $435 = $413 >>> 16;
      $436 = $435 & 255;
      $437 = $434 >>> $436;
      $438 = $416 >>> 16;
      $439 = $438 & 255;
      $440 = (125924 + ($439<<2)|0);
      $441 = HEAP32[$440>>2]|0;
      $442 = (($441) + ($437)|0);
      $443 = HEAP8[$442>>0]|0;
      $sB$3$shrunk = $443;$sG$3$shrunk = $432;$sR$3$shrunk = $421;
      break;
     }
     case 2:  {
      $444 = HEAP16[$src$4>>1]|0;
      $445 = $444&65535;
      $446 = HEAP32[$32>>2]|0;
      $447 = $445 & $446;
      $448 = HEAP32[$33>>2]|0;
      $449 = $448 & 255;
      $450 = $447 >>> $449;
      $451 = HEAP32[$34>>2]|0;
      $452 = $451 & 255;
      $453 = (125924 + ($452<<2)|0);
      $454 = HEAP32[$453>>2]|0;
      $455 = (($454) + ($450)|0);
      $456 = HEAP8[$455>>0]|0;
      $457 = HEAP32[$35>>2]|0;
      $458 = $457 & $445;
      $459 = $448 >>> 8;
      $460 = $459 & 255;
      $461 = $458 >>> $460;
      $462 = $451 >>> 8;
      $463 = $462 & 255;
      $464 = (125924 + ($463<<2)|0);
      $465 = HEAP32[$464>>2]|0;
      $466 = (($465) + ($461)|0);
      $467 = HEAP8[$466>>0]|0;
      $468 = HEAP32[$36>>2]|0;
      $469 = $468 & $445;
      $470 = $448 >>> 16;
      $471 = $470 & 255;
      $472 = $469 >>> $471;
      $473 = $451 >>> 16;
      $474 = $473 & 255;
      $475 = (125924 + ($474<<2)|0);
      $476 = HEAP32[$475>>2]|0;
      $477 = (($476) + ($472)|0);
      $478 = HEAP8[$477>>0]|0;
      $sB$3$shrunk = $478;$sG$3$shrunk = $467;$sR$3$shrunk = $456;
      break;
     }
     case 3:  {
      $479 = HEAP32[$33>>2]|0;
      $div33 = $479 >>> 3;
      $480 = $div33 & 31;
      $481 = (($src$4) + ($480)|0);
      $482 = HEAP8[$481>>0]|0;
      $483 = $479 >>> 11;
      $div34 = $483 & 31;
      $484 = (($src$4) + ($div34)|0);
      $485 = HEAP8[$484>>0]|0;
      $486 = $479 >>> 19;
      $487 = $486 & 31;
      $488 = (($src$4) + ($487)|0);
      $489 = HEAP8[$488>>0]|0;
      $sB$3$shrunk = $489;$sG$3$shrunk = $485;$sR$3$shrunk = $482;
      break;
     }
     case 4:  {
      $490 = HEAP32[$src$4>>2]|0;
      $491 = HEAP32[$32>>2]|0;
      $492 = $491 & $490;
      $493 = HEAP32[$33>>2]|0;
      $494 = $493 & 255;
      $495 = $492 >>> $494;
      $496 = HEAP32[$34>>2]|0;
      $497 = $496 & 255;
      $498 = (125924 + ($497<<2)|0);
      $499 = HEAP32[$498>>2]|0;
      $500 = (($499) + ($495)|0);
      $501 = HEAP8[$500>>0]|0;
      $502 = HEAP32[$35>>2]|0;
      $503 = $502 & $490;
      $504 = $493 >>> 8;
      $505 = $504 & 255;
      $506 = $503 >>> $505;
      $507 = $496 >>> 8;
      $508 = $507 & 255;
      $509 = (125924 + ($508<<2)|0);
      $510 = HEAP32[$509>>2]|0;
      $511 = (($510) + ($506)|0);
      $512 = HEAP8[$511>>0]|0;
      $513 = HEAP32[$36>>2]|0;
      $514 = $513 & $490;
      $515 = $493 >>> 16;
      $516 = $515 & 255;
      $517 = $514 >>> $516;
      $518 = $496 >>> 16;
      $519 = $518 & 255;
      $520 = (125924 + ($519<<2)|0);
      $521 = HEAP32[$520>>2]|0;
      $522 = (($521) + ($517)|0);
      $523 = HEAP8[$522>>0]|0;
      $sB$3$shrunk = $523;$sG$3$shrunk = $512;$sR$3$shrunk = $501;
      break;
     }
     default: {
      $sB$3$shrunk = 0;$sG$3$shrunk = 0;$sR$3$shrunk = 0;
     }
     }
     $sB$3 = $sB$3$shrunk&255;
     $sG$3 = $sG$3$shrunk&255;
     $sR$3 = $sR$3$shrunk&255;
     $524 = $sR$3 & 224;
     $525 = $sG$3 >>> 5;
     $526 = $525 << 2;
     $527 = $526 | $524;
     $528 = $sB$3 >>> 6;
     $529 = $527 | $528;
     $530 = $529&255;
     HEAP8[$dst$4>>0] = $530;
     $531 = ((($dst$4)) + 1|0);
     $532 = (($src$4) + ($17)|0);
     $dst$5 = $531;$n$4 = $n$3;$src$5 = $532;
     label = 31;
     continue;
    }
    else if ((label|0) == 31) {
     label = 0;
     switch ($17|0) {
     case 1:  {
      $533 = HEAP8[$src$5>>0]|0;
      $534 = $533&255;
      $535 = HEAP32[$32>>2]|0;
      $536 = $534 & $535;
      $537 = HEAP32[$33>>2]|0;
      $538 = $537 & 255;
      $539 = $536 >>> $538;
      $540 = HEAP32[$34>>2]|0;
      $541 = $540 & 255;
      $542 = (125924 + ($541<<2)|0);
      $543 = HEAP32[$542>>2]|0;
      $544 = (($543) + ($539)|0);
      $545 = HEAP8[$544>>0]|0;
      $546 = HEAP32[$35>>2]|0;
      $547 = $546 & $534;
      $548 = $537 >>> 8;
      $549 = $548 & 255;
      $550 = $547 >>> $549;
      $551 = $540 >>> 8;
      $552 = $551 & 255;
      $553 = (125924 + ($552<<2)|0);
      $554 = HEAP32[$553>>2]|0;
      $555 = (($554) + ($550)|0);
      $556 = HEAP8[$555>>0]|0;
      $557 = HEAP32[$36>>2]|0;
      $558 = $557 & $534;
      $559 = $537 >>> 16;
      $560 = $559 & 255;
      $561 = $558 >>> $560;
      $562 = $540 >>> 16;
      $563 = $562 & 255;
      $564 = (125924 + ($563<<2)|0);
      $565 = HEAP32[$564>>2]|0;
      $566 = (($565) + ($561)|0);
      $567 = HEAP8[$566>>0]|0;
      $sB$4$shrunk = $567;$sG$4$shrunk = $556;$sR$4$shrunk = $545;
      break;
     }
     case 2:  {
      $568 = HEAP16[$src$5>>1]|0;
      $569 = $568&65535;
      $570 = HEAP32[$32>>2]|0;
      $571 = $569 & $570;
      $572 = HEAP32[$33>>2]|0;
      $573 = $572 & 255;
      $574 = $571 >>> $573;
      $575 = HEAP32[$34>>2]|0;
      $576 = $575 & 255;
      $577 = (125924 + ($576<<2)|0);
      $578 = HEAP32[$577>>2]|0;
      $579 = (($578) + ($574)|0);
      $580 = HEAP8[$579>>0]|0;
      $581 = HEAP32[$35>>2]|0;
      $582 = $581 & $569;
      $583 = $572 >>> 8;
      $584 = $583 & 255;
      $585 = $582 >>> $584;
      $586 = $575 >>> 8;
      $587 = $586 & 255;
      $588 = (125924 + ($587<<2)|0);
      $589 = HEAP32[$588>>2]|0;
      $590 = (($589) + ($585)|0);
      $591 = HEAP8[$590>>0]|0;
      $592 = HEAP32[$36>>2]|0;
      $593 = $592 & $569;
      $594 = $572 >>> 16;
      $595 = $594 & 255;
      $596 = $593 >>> $595;
      $597 = $575 >>> 16;
      $598 = $597 & 255;
      $599 = (125924 + ($598<<2)|0);
      $600 = HEAP32[$599>>2]|0;
      $601 = (($600) + ($596)|0);
      $602 = HEAP8[$601>>0]|0;
      $sB$4$shrunk = $602;$sG$4$shrunk = $591;$sR$4$shrunk = $580;
      break;
     }
     case 3:  {
      $603 = HEAP32[$33>>2]|0;
      $div30 = $603 >>> 3;
      $604 = $div30 & 31;
      $605 = (($src$5) + ($604)|0);
      $606 = HEAP8[$605>>0]|0;
      $607 = $603 >>> 11;
      $div31 = $607 & 31;
      $608 = (($src$5) + ($div31)|0);
      $609 = HEAP8[$608>>0]|0;
      $610 = $603 >>> 19;
      $611 = $610 & 31;
      $612 = (($src$5) + ($611)|0);
      $613 = HEAP8[$612>>0]|0;
      $sB$4$shrunk = $613;$sG$4$shrunk = $609;$sR$4$shrunk = $606;
      break;
     }
     case 4:  {
      $614 = HEAP32[$src$5>>2]|0;
      $615 = HEAP32[$32>>2]|0;
      $616 = $615 & $614;
      $617 = HEAP32[$33>>2]|0;
      $618 = $617 & 255;
      $619 = $616 >>> $618;
      $620 = HEAP32[$34>>2]|0;
      $621 = $620 & 255;
      $622 = (125924 + ($621<<2)|0);
      $623 = HEAP32[$622>>2]|0;
      $624 = (($623) + ($619)|0);
      $625 = HEAP8[$624>>0]|0;
      $626 = HEAP32[$35>>2]|0;
      $627 = $626 & $614;
      $628 = $617 >>> 8;
      $629 = $628 & 255;
      $630 = $627 >>> $629;
      $631 = $620 >>> 8;
      $632 = $631 & 255;
      $633 = (125924 + ($632<<2)|0);
      $634 = HEAP32[$633>>2]|0;
      $635 = (($634) + ($630)|0);
      $636 = HEAP8[$635>>0]|0;
      $637 = HEAP32[$36>>2]|0;
      $638 = $637 & $614;
      $639 = $617 >>> 16;
      $640 = $639 & 255;
      $641 = $638 >>> $640;
      $642 = $620 >>> 16;
      $643 = $642 & 255;
      $644 = (125924 + ($643<<2)|0);
      $645 = HEAP32[$644>>2]|0;
      $646 = (($645) + ($641)|0);
      $647 = HEAP8[$646>>0]|0;
      $sB$4$shrunk = $647;$sG$4$shrunk = $636;$sR$4$shrunk = $625;
      break;
     }
     default: {
      $sB$4$shrunk = 0;$sG$4$shrunk = 0;$sR$4$shrunk = 0;
     }
     }
     $sB$4 = $sB$4$shrunk&255;
     $sG$4 = $sG$4$shrunk&255;
     $sR$4 = $sR$4$shrunk&255;
     $648 = $sR$4 & 224;
     $649 = $sG$4 >>> 5;
     $650 = $649 << 2;
     $651 = $650 | $648;
     $652 = $sB$4 >>> 6;
     $653 = $651 | $652;
     $654 = $653&255;
     HEAP8[$dst$5>>0] = $654;
     $655 = ((($dst$5)) + 1|0);
     $656 = (($src$5) + ($17)|0);
     $dst$6 = $655;$n$5 = $n$4;$src$6 = $656;
     label = 37;
     continue;
    }
    else if ((label|0) == 37) {
     label = 0;
     switch ($17|0) {
     case 1:  {
      $657 = HEAP8[$src$6>>0]|0;
      $658 = $657&255;
      $659 = HEAP32[$32>>2]|0;
      $660 = $658 & $659;
      $661 = HEAP32[$33>>2]|0;
      $662 = $661 & 255;
      $663 = $660 >>> $662;
      $664 = HEAP32[$34>>2]|0;
      $665 = $664 & 255;
      $666 = (125924 + ($665<<2)|0);
      $667 = HEAP32[$666>>2]|0;
      $668 = (($667) + ($663)|0);
      $669 = HEAP8[$668>>0]|0;
      $670 = HEAP32[$35>>2]|0;
      $671 = $670 & $658;
      $672 = $661 >>> 8;
      $673 = $672 & 255;
      $674 = $671 >>> $673;
      $675 = $664 >>> 8;
      $676 = $675 & 255;
      $677 = (125924 + ($676<<2)|0);
      $678 = HEAP32[$677>>2]|0;
      $679 = (($678) + ($674)|0);
      $680 = HEAP8[$679>>0]|0;
      $681 = HEAP32[$36>>2]|0;
      $682 = $681 & $658;
      $683 = $661 >>> 16;
      $684 = $683 & 255;
      $685 = $682 >>> $684;
      $686 = $664 >>> 16;
      $687 = $686 & 255;
      $688 = (125924 + ($687<<2)|0);
      $689 = HEAP32[$688>>2]|0;
      $690 = (($689) + ($685)|0);
      $691 = HEAP8[$690>>0]|0;
      $sB$5$shrunk = $691;$sG$5$shrunk = $680;$sR$5$shrunk = $669;
      break;
     }
     case 2:  {
      $692 = HEAP16[$src$6>>1]|0;
      $693 = $692&65535;
      $694 = HEAP32[$32>>2]|0;
      $695 = $693 & $694;
      $696 = HEAP32[$33>>2]|0;
      $697 = $696 & 255;
      $698 = $695 >>> $697;
      $699 = HEAP32[$34>>2]|0;
      $700 = $699 & 255;
      $701 = (125924 + ($700<<2)|0);
      $702 = HEAP32[$701>>2]|0;
      $703 = (($702) + ($698)|0);
      $704 = HEAP8[$703>>0]|0;
      $705 = HEAP32[$35>>2]|0;
      $706 = $705 & $693;
      $707 = $696 >>> 8;
      $708 = $707 & 255;
      $709 = $706 >>> $708;
      $710 = $699 >>> 8;
      $711 = $710 & 255;
      $712 = (125924 + ($711<<2)|0);
      $713 = HEAP32[$712>>2]|0;
      $714 = (($713) + ($709)|0);
      $715 = HEAP8[$714>>0]|0;
      $716 = HEAP32[$36>>2]|0;
      $717 = $716 & $693;
      $718 = $696 >>> 16;
      $719 = $718 & 255;
      $720 = $717 >>> $719;
      $721 = $699 >>> 16;
      $722 = $721 & 255;
      $723 = (125924 + ($722<<2)|0);
      $724 = HEAP32[$723>>2]|0;
      $725 = (($724) + ($720)|0);
      $726 = HEAP8[$725>>0]|0;
      $sB$5$shrunk = $726;$sG$5$shrunk = $715;$sR$5$shrunk = $704;
      break;
     }
     case 3:  {
      $727 = HEAP32[$33>>2]|0;
      $div27 = $727 >>> 3;
      $728 = $div27 & 31;
      $729 = (($src$6) + ($728)|0);
      $730 = HEAP8[$729>>0]|0;
      $731 = $727 >>> 11;
      $div28 = $731 & 31;
      $732 = (($src$6) + ($div28)|0);
      $733 = HEAP8[$732>>0]|0;
      $734 = $727 >>> 19;
      $735 = $734 & 31;
      $736 = (($src$6) + ($735)|0);
      $737 = HEAP8[$736>>0]|0;
      $sB$5$shrunk = $737;$sG$5$shrunk = $733;$sR$5$shrunk = $730;
      break;
     }
     case 4:  {
      $738 = HEAP32[$src$6>>2]|0;
      $739 = HEAP32[$32>>2]|0;
      $740 = $739 & $738;
      $741 = HEAP32[$33>>2]|0;
      $742 = $741 & 255;
      $743 = $740 >>> $742;
      $744 = HEAP32[$34>>2]|0;
      $745 = $744 & 255;
      $746 = (125924 + ($745<<2)|0);
      $747 = HEAP32[$746>>2]|0;
      $748 = (($747) + ($743)|0);
      $749 = HEAP8[$748>>0]|0;
      $750 = HEAP32[$35>>2]|0;
      $751 = $750 & $738;
      $752 = $741 >>> 8;
      $753 = $752 & 255;
      $754 = $751 >>> $753;
      $755 = $744 >>> 8;
      $756 = $755 & 255;
      $757 = (125924 + ($756<<2)|0);
      $758 = HEAP32[$757>>2]|0;
      $759 = (($758) + ($754)|0);
      $760 = HEAP8[$759>>0]|0;
      $761 = HEAP32[$36>>2]|0;
      $762 = $761 & $738;
      $763 = $741 >>> 16;
      $764 = $763 & 255;
      $765 = $762 >>> $764;
      $766 = $744 >>> 16;
      $767 = $766 & 255;
      $768 = (125924 + ($767<<2)|0);
      $769 = HEAP32[$768>>2]|0;
      $770 = (($769) + ($765)|0);
      $771 = HEAP8[$770>>0]|0;
      $sB$5$shrunk = $771;$sG$5$shrunk = $760;$sR$5$shrunk = $749;
      break;
     }
     default: {
      $sB$5$shrunk = 0;$sG$5$shrunk = 0;$sR$5$shrunk = 0;
     }
     }
     $sB$5 = $sB$5$shrunk&255;
     $sG$5 = $sG$5$shrunk&255;
     $sR$5 = $sR$5$shrunk&255;
     $772 = $sR$5 & 224;
     $773 = $sG$5 >>> 5;
     $774 = $773 << 2;
     $775 = $774 | $772;
     $776 = $sB$5 >>> 6;
     $777 = $775 | $776;
     $778 = $777&255;
     HEAP8[$dst$6>>0] = $778;
     $779 = ((($dst$6)) + 1|0);
     $780 = (($src$6) + ($17)|0);
     $dst$7 = $779;$n$6 = $n$5;$src$7 = $780;
     label = 43;
     continue;
    }
    else if ((label|0) == 43) {
     label = 0;
     switch ($17|0) {
     case 1:  {
      $781 = HEAP8[$src$7>>0]|0;
      $782 = $781&255;
      $783 = HEAP32[$32>>2]|0;
      $784 = $782 & $783;
      $785 = HEAP32[$33>>2]|0;
      $786 = $785 & 255;
      $787 = $784 >>> $786;
      $788 = HEAP32[$34>>2]|0;
      $789 = $788 & 255;
      $790 = (125924 + ($789<<2)|0);
      $791 = HEAP32[$790>>2]|0;
      $792 = (($791) + ($787)|0);
      $793 = HEAP8[$792>>0]|0;
      $794 = HEAP32[$35>>2]|0;
      $795 = $794 & $782;
      $796 = $785 >>> 8;
      $797 = $796 & 255;
      $798 = $795 >>> $797;
      $799 = $788 >>> 8;
      $800 = $799 & 255;
      $801 = (125924 + ($800<<2)|0);
      $802 = HEAP32[$801>>2]|0;
      $803 = (($802) + ($798)|0);
      $804 = HEAP8[$803>>0]|0;
      $805 = HEAP32[$36>>2]|0;
      $806 = $805 & $782;
      $807 = $785 >>> 16;
      $808 = $807 & 255;
      $809 = $806 >>> $808;
      $810 = $788 >>> 16;
      $811 = $810 & 255;
      $812 = (125924 + ($811<<2)|0);
      $813 = HEAP32[$812>>2]|0;
      $814 = (($813) + ($809)|0);
      $815 = HEAP8[$814>>0]|0;
      $sB$6$shrunk = $815;$sG$6$shrunk = $804;$sR$6$shrunk = $793;
      break;
     }
     case 2:  {
      $816 = HEAP16[$src$7>>1]|0;
      $817 = $816&65535;
      $818 = HEAP32[$32>>2]|0;
      $819 = $817 & $818;
      $820 = HEAP32[$33>>2]|0;
      $821 = $820 & 255;
      $822 = $819 >>> $821;
      $823 = HEAP32[$34>>2]|0;
      $824 = $823 & 255;
      $825 = (125924 + ($824<<2)|0);
      $826 = HEAP32[$825>>2]|0;
      $827 = (($826) + ($822)|0);
      $828 = HEAP8[$827>>0]|0;
      $829 = HEAP32[$35>>2]|0;
      $830 = $829 & $817;
      $831 = $820 >>> 8;
      $832 = $831 & 255;
      $833 = $830 >>> $832;
      $834 = $823 >>> 8;
      $835 = $834 & 255;
      $836 = (125924 + ($835<<2)|0);
      $837 = HEAP32[$836>>2]|0;
      $838 = (($837) + ($833)|0);
      $839 = HEAP8[$838>>0]|0;
      $840 = HEAP32[$36>>2]|0;
      $841 = $840 & $817;
      $842 = $820 >>> 16;
      $843 = $842 & 255;
      $844 = $841 >>> $843;
      $845 = $823 >>> 16;
      $846 = $845 & 255;
      $847 = (125924 + ($846<<2)|0);
      $848 = HEAP32[$847>>2]|0;
      $849 = (($848) + ($844)|0);
      $850 = HEAP8[$849>>0]|0;
      $sB$6$shrunk = $850;$sG$6$shrunk = $839;$sR$6$shrunk = $828;
      break;
     }
     case 3:  {
      $851 = HEAP32[$33>>2]|0;
      $div24 = $851 >>> 3;
      $852 = $div24 & 31;
      $853 = (($src$7) + ($852)|0);
      $854 = HEAP8[$853>>0]|0;
      $855 = $851 >>> 11;
      $div25 = $855 & 31;
      $856 = (($src$7) + ($div25)|0);
      $857 = HEAP8[$856>>0]|0;
      $858 = $851 >>> 19;
      $859 = $858 & 31;
      $860 = (($src$7) + ($859)|0);
      $861 = HEAP8[$860>>0]|0;
      $sB$6$shrunk = $861;$sG$6$shrunk = $857;$sR$6$shrunk = $854;
      break;
     }
     case 4:  {
      $862 = HEAP32[$src$7>>2]|0;
      $863 = HEAP32[$32>>2]|0;
      $864 = $863 & $862;
      $865 = HEAP32[$33>>2]|0;
      $866 = $865 & 255;
      $867 = $864 >>> $866;
      $868 = HEAP32[$34>>2]|0;
      $869 = $868 & 255;
      $870 = (125924 + ($869<<2)|0);
      $871 = HEAP32[$870>>2]|0;
      $872 = (($871) + ($867)|0);
      $873 = HEAP8[$872>>0]|0;
      $874 = HEAP32[$35>>2]|0;
      $875 = $874 & $862;
      $876 = $865 >>> 8;
      $877 = $876 & 255;
      $878 = $875 >>> $877;
      $879 = $868 >>> 8;
      $880 = $879 & 255;
      $881 = (125924 + ($880<<2)|0);
      $882 = HEAP32[$881>>2]|0;
      $883 = (($882) + ($878)|0);
      $884 = HEAP8[$883>>0]|0;
      $885 = HEAP32[$36>>2]|0;
      $886 = $885 & $862;
      $887 = $865 >>> 16;
      $888 = $887 & 255;
      $889 = $886 >>> $888;
      $890 = $868 >>> 16;
      $891 = $890 & 255;
      $892 = (125924 + ($891<<2)|0);
      $893 = HEAP32[$892>>2]|0;
      $894 = (($893) + ($889)|0);
      $895 = HEAP8[$894>>0]|0;
      $sB$6$shrunk = $895;$sG$6$shrunk = $884;$sR$6$shrunk = $873;
      break;
     }
     default: {
      $sB$6$shrunk = 0;$sG$6$shrunk = 0;$sR$6$shrunk = 0;
     }
     }
     $sB$6 = $sB$6$shrunk&255;
     $sG$6 = $sG$6$shrunk&255;
     $sR$6 = $sR$6$shrunk&255;
     $896 = $sR$6 & 224;
     $897 = $sG$6 >>> 5;
     $898 = $897 << 2;
     $899 = $898 | $896;
     $900 = $sB$6 >>> 6;
     $901 = $899 | $900;
     $902 = $901&255;
     HEAP8[$dst$7>>0] = $902;
     $903 = ((($dst$7)) + 1|0);
     $904 = (($src$7) + ($17)|0);
     $dst$8 = $903;$n$7 = $n$6;$src$8 = $904;
     label = 49;
     continue;
    }
    else if ((label|0) == 49) {
     label = 0;
     switch ($17|0) {
     case 1:  {
      $905 = HEAP8[$src$8>>0]|0;
      $906 = $905&255;
      $907 = HEAP32[$32>>2]|0;
      $908 = $906 & $907;
      $909 = HEAP32[$33>>2]|0;
      $910 = $909 & 255;
      $911 = $908 >>> $910;
      $912 = HEAP32[$34>>2]|0;
      $913 = $912 & 255;
      $914 = (125924 + ($913<<2)|0);
      $915 = HEAP32[$914>>2]|0;
      $916 = (($915) + ($911)|0);
      $917 = HEAP8[$916>>0]|0;
      $918 = HEAP32[$35>>2]|0;
      $919 = $918 & $906;
      $920 = $909 >>> 8;
      $921 = $920 & 255;
      $922 = $919 >>> $921;
      $923 = $912 >>> 8;
      $924 = $923 & 255;
      $925 = (125924 + ($924<<2)|0);
      $926 = HEAP32[$925>>2]|0;
      $927 = (($926) + ($922)|0);
      $928 = HEAP8[$927>>0]|0;
      $929 = HEAP32[$36>>2]|0;
      $930 = $929 & $906;
      $931 = $909 >>> 16;
      $932 = $931 & 255;
      $933 = $930 >>> $932;
      $934 = $912 >>> 16;
      $935 = $934 & 255;
      $936 = (125924 + ($935<<2)|0);
      $937 = HEAP32[$936>>2]|0;
      $938 = (($937) + ($933)|0);
      $939 = HEAP8[$938>>0]|0;
      $sB$7$shrunk = $939;$sG$7$shrunk = $928;$sR$7$shrunk = $917;
      break;
     }
     case 2:  {
      $940 = HEAP16[$src$8>>1]|0;
      $941 = $940&65535;
      $942 = HEAP32[$32>>2]|0;
      $943 = $941 & $942;
      $944 = HEAP32[$33>>2]|0;
      $945 = $944 & 255;
      $946 = $943 >>> $945;
      $947 = HEAP32[$34>>2]|0;
      $948 = $947 & 255;
      $949 = (125924 + ($948<<2)|0);
      $950 = HEAP32[$949>>2]|0;
      $951 = (($950) + ($946)|0);
      $952 = HEAP8[$951>>0]|0;
      $953 = HEAP32[$35>>2]|0;
      $954 = $953 & $941;
      $955 = $944 >>> 8;
      $956 = $955 & 255;
      $957 = $954 >>> $956;
      $958 = $947 >>> 8;
      $959 = $958 & 255;
      $960 = (125924 + ($959<<2)|0);
      $961 = HEAP32[$960>>2]|0;
      $962 = (($961) + ($957)|0);
      $963 = HEAP8[$962>>0]|0;
      $964 = HEAP32[$36>>2]|0;
      $965 = $964 & $941;
      $966 = $944 >>> 16;
      $967 = $966 & 255;
      $968 = $965 >>> $967;
      $969 = $947 >>> 16;
      $970 = $969 & 255;
      $971 = (125924 + ($970<<2)|0);
      $972 = HEAP32[$971>>2]|0;
      $973 = (($972) + ($968)|0);
      $974 = HEAP8[$973>>0]|0;
      $sB$7$shrunk = $974;$sG$7$shrunk = $963;$sR$7$shrunk = $952;
      break;
     }
     case 3:  {
      $975 = HEAP32[$33>>2]|0;
      $div45 = $975 >>> 3;
      $976 = $div45 & 31;
      $977 = (($src$8) + ($976)|0);
      $978 = HEAP8[$977>>0]|0;
      $979 = $975 >>> 11;
      $div46 = $979 & 31;
      $980 = (($src$8) + ($div46)|0);
      $981 = HEAP8[$980>>0]|0;
      $982 = $975 >>> 19;
      $983 = $982 & 31;
      $984 = (($src$8) + ($983)|0);
      $985 = HEAP8[$984>>0]|0;
      $sB$7$shrunk = $985;$sG$7$shrunk = $981;$sR$7$shrunk = $978;
      break;
     }
     case 4:  {
      $986 = HEAP32[$src$8>>2]|0;
      $987 = HEAP32[$32>>2]|0;
      $988 = $987 & $986;
      $989 = HEAP32[$33>>2]|0;
      $990 = $989 & 255;
      $991 = $988 >>> $990;
      $992 = HEAP32[$34>>2]|0;
      $993 = $992 & 255;
      $994 = (125924 + ($993<<2)|0);
      $995 = HEAP32[$994>>2]|0;
      $996 = (($995) + ($991)|0);
      $997 = HEAP8[$996>>0]|0;
      $998 = HEAP32[$35>>2]|0;
      $999 = $998 & $986;
      $1000 = $989 >>> 8;
      $1001 = $1000 & 255;
      $1002 = $999 >>> $1001;
      $1003 = $992 >>> 8;
      $1004 = $1003 & 255;
      $1005 = (125924 + ($1004<<2)|0);
      $1006 = HEAP32[$1005>>2]|0;
      $1007 = (($1006) + ($1002)|0);
      $1008 = HEAP8[$1007>>0]|0;
      $1009 = HEAP32[$36>>2]|0;
      $1010 = $1009 & $986;
      $1011 = $989 >>> 16;
      $1012 = $1011 & 255;
      $1013 = $1010 >>> $1012;
      $1014 = $992 >>> 16;
      $1015 = $1014 & 255;
      $1016 = (125924 + ($1015<<2)|0);
      $1017 = HEAP32[$1016>>2]|0;
      $1018 = (($1017) + ($1013)|0);
      $1019 = HEAP8[$1018>>0]|0;
      $sB$7$shrunk = $1019;$sG$7$shrunk = $1008;$sR$7$shrunk = $997;
      break;
     }
     default: {
      $sB$7$shrunk = 0;$sG$7$shrunk = 0;$sR$7$shrunk = 0;
     }
     }
     $sB$7 = $sB$7$shrunk&255;
     $sG$7 = $sG$7$shrunk&255;
     $sR$7 = $sR$7$shrunk&255;
     $1020 = $sR$7 & 224;
     $1021 = $sG$7 >>> 5;
     $1022 = $1021 << 2;
     $1023 = $1022 | $1020;
     $1024 = $sB$7 >>> 6;
     $1025 = $1023 | $1024;
     $1026 = $1025&255;
     HEAP8[$dst$8>>0] = $1026;
     $1027 = ((($dst$8)) + 1|0);
     $1028 = (($src$8) + ($17)|0);
     $1029 = (($n$7) + -1)|0;
     $1030 = ($n$7|0)>(1);
     if ($1030) {
      $dst$1 = $1027;$n$0 = $1029;$src$1 = $1028;
      label = 7;
      continue;
     } else {
      break;
     }
    }
   }
   $1031 = (($1028) + ($6)|0);
   $1032 = (($1027) + ($10)|0);
   $1033 = (($1034) + -1)|0;
   $1035 = ($1034|0)==(0);
   if ($1035) {
    label = 108;
    break;
   } else {
    $1034 = $1033;$dst$050 = $1032;$src$051 = $1031;
   }
  }
  if ((label|0) == 55) {
   // unreachable;
  }
  else if ((label|0) == 108) {
   return;
  }
 } else {
  if ($20) {
   return;
  }
  $21 = (($1) + 7)|0;
  $22 = (($21|0) / 8)&-1;
  $23 = $1 & 7;
  $24 = ((($14)) + 12|0);
  $25 = ((($14)) + 32|0);
  $26 = ((($14)) + 28|0);
  $27 = ((($14)) + 16|0);
  $28 = ((($14)) + 20|0);
  $2041 = $19;$dst$952 = $8;$src$953 = $4;
  L74: while(1) {
   switch ($23|0) {
   case 0:  {
    $dst$10 = $dst$952;$n1$0 = $22;$src$10 = $src$953;
    label = 58;
    break;
   }
   case 7:  {
    $dst$11 = $dst$952;$n1$1 = $22;$src$11 = $src$953;
    label = 64;
    break;
   }
   case 6:  {
    $dst$12 = $dst$952;$n1$2 = $22;$src$12 = $src$953;
    label = 70;
    break;
   }
   case 5:  {
    $dst$13 = $dst$952;$n1$3 = $22;$src$13 = $src$953;
    label = 76;
    break;
   }
   case 4:  {
    $dst$14 = $dst$952;$n1$4 = $22;$src$14 = $src$953;
    label = 82;
    break;
   }
   case 3:  {
    $dst$15 = $dst$952;$n1$5 = $22;$src$15 = $src$953;
    label = 88;
    break;
   }
   case 2:  {
    $dst$16 = $dst$952;$n1$6 = $22;$src$16 = $src$953;
    label = 94;
    break;
   }
   case 1:  {
    $dst$17 = $dst$952;$n1$7 = $22;$src$17 = $src$953;
    label = 100;
    break;
   }
   default: {
    label = 106;
    break L74;
   }
   }
   while(1) {
    if ((label|0) == 58) {
     label = 0;
     switch ($17|0) {
     case 1:  {
      $1036 = HEAP8[$src$10>>0]|0;
      $1037 = $1036&255;
      $1038 = HEAP32[$24>>2]|0;
      $1039 = $1037 & $1038;
      $1040 = HEAP32[$25>>2]|0;
      $1041 = $1040 & 255;
      $1042 = $1039 >>> $1041;
      $1043 = HEAP32[$26>>2]|0;
      $1044 = $1043 & 255;
      $1045 = (125924 + ($1044<<2)|0);
      $1046 = HEAP32[$1045>>2]|0;
      $1047 = (($1046) + ($1042)|0);
      $1048 = HEAP8[$1047>>0]|0;
      $1049 = HEAP32[$27>>2]|0;
      $1050 = $1049 & $1037;
      $1051 = $1040 >>> 8;
      $1052 = $1051 & 255;
      $1053 = $1050 >>> $1052;
      $1054 = $1043 >>> 8;
      $1055 = $1054 & 255;
      $1056 = (125924 + ($1055<<2)|0);
      $1057 = HEAP32[$1056>>2]|0;
      $1058 = (($1057) + ($1053)|0);
      $1059 = HEAP8[$1058>>0]|0;
      $1060 = HEAP32[$28>>2]|0;
      $1061 = $1060 & $1037;
      $1062 = $1040 >>> 16;
      $1063 = $1062 & 255;
      $1064 = $1061 >>> $1063;
      $1065 = $1043 >>> 16;
      $1066 = $1065 & 255;
      $1067 = (125924 + ($1066<<2)|0);
      $1068 = HEAP32[$1067>>2]|0;
      $1069 = (($1068) + ($1064)|0);
      $1070 = HEAP8[$1069>>0]|0;
      $sB$8$shrunk = $1070;$sG$8$shrunk = $1059;$sR$8$shrunk = $1048;
      break;
     }
     case 2:  {
      $1071 = HEAP16[$src$10>>1]|0;
      $1072 = $1071&65535;
      $1073 = HEAP32[$24>>2]|0;
      $1074 = $1072 & $1073;
      $1075 = HEAP32[$25>>2]|0;
      $1076 = $1075 & 255;
      $1077 = $1074 >>> $1076;
      $1078 = HEAP32[$26>>2]|0;
      $1079 = $1078 & 255;
      $1080 = (125924 + ($1079<<2)|0);
      $1081 = HEAP32[$1080>>2]|0;
      $1082 = (($1081) + ($1077)|0);
      $1083 = HEAP8[$1082>>0]|0;
      $1084 = HEAP32[$27>>2]|0;
      $1085 = $1084 & $1072;
      $1086 = $1075 >>> 8;
      $1087 = $1086 & 255;
      $1088 = $1085 >>> $1087;
      $1089 = $1078 >>> 8;
      $1090 = $1089 & 255;
      $1091 = (125924 + ($1090<<2)|0);
      $1092 = HEAP32[$1091>>2]|0;
      $1093 = (($1092) + ($1088)|0);
      $1094 = HEAP8[$1093>>0]|0;
      $1095 = HEAP32[$28>>2]|0;
      $1096 = $1095 & $1072;
      $1097 = $1075 >>> 16;
      $1098 = $1097 & 255;
      $1099 = $1096 >>> $1098;
      $1100 = $1078 >>> 16;
      $1101 = $1100 & 255;
      $1102 = (125924 + ($1101<<2)|0);
      $1103 = HEAP32[$1102>>2]|0;
      $1104 = (($1103) + ($1099)|0);
      $1105 = HEAP8[$1104>>0]|0;
      $sB$8$shrunk = $1105;$sG$8$shrunk = $1094;$sR$8$shrunk = $1083;
      break;
     }
     case 3:  {
      $1106 = HEAP32[$25>>2]|0;
      $div18 = $1106 >>> 3;
      $1107 = $div18 & 31;
      $1108 = (($src$10) + ($1107)|0);
      $1109 = HEAP8[$1108>>0]|0;
      $1110 = $1106 >>> 11;
      $div19 = $1110 & 31;
      $1111 = (($src$10) + ($div19)|0);
      $1112 = HEAP8[$1111>>0]|0;
      $1113 = $1106 >>> 19;
      $1114 = $1113 & 31;
      $1115 = (($src$10) + ($1114)|0);
      $1116 = HEAP8[$1115>>0]|0;
      $sB$8$shrunk = $1116;$sG$8$shrunk = $1112;$sR$8$shrunk = $1109;
      break;
     }
     case 4:  {
      $1117 = HEAP32[$src$10>>2]|0;
      $1118 = HEAP32[$24>>2]|0;
      $1119 = $1118 & $1117;
      $1120 = HEAP32[$25>>2]|0;
      $1121 = $1120 & 255;
      $1122 = $1119 >>> $1121;
      $1123 = HEAP32[$26>>2]|0;
      $1124 = $1123 & 255;
      $1125 = (125924 + ($1124<<2)|0);
      $1126 = HEAP32[$1125>>2]|0;
      $1127 = (($1126) + ($1122)|0);
      $1128 = HEAP8[$1127>>0]|0;
      $1129 = HEAP32[$27>>2]|0;
      $1130 = $1129 & $1117;
      $1131 = $1120 >>> 8;
      $1132 = $1131 & 255;
      $1133 = $1130 >>> $1132;
      $1134 = $1123 >>> 8;
      $1135 = $1134 & 255;
      $1136 = (125924 + ($1135<<2)|0);
      $1137 = HEAP32[$1136>>2]|0;
      $1138 = (($1137) + ($1133)|0);
      $1139 = HEAP8[$1138>>0]|0;
      $1140 = HEAP32[$28>>2]|0;
      $1141 = $1140 & $1117;
      $1142 = $1120 >>> 16;
      $1143 = $1142 & 255;
      $1144 = $1141 >>> $1143;
      $1145 = $1123 >>> 16;
      $1146 = $1145 & 255;
      $1147 = (125924 + ($1146<<2)|0);
      $1148 = HEAP32[$1147>>2]|0;
      $1149 = (($1148) + ($1144)|0);
      $1150 = HEAP8[$1149>>0]|0;
      $sB$8$shrunk = $1150;$sG$8$shrunk = $1139;$sR$8$shrunk = $1128;
      break;
     }
     default: {
      $sB$8$shrunk = 0;$sG$8$shrunk = 0;$sR$8$shrunk = 0;
     }
     }
     $sB$8 = $sB$8$shrunk&255;
     $sG$8 = $sG$8$shrunk&255;
     $sR$8 = $sR$8$shrunk&255;
     $1151 = $sR$8 & 224;
     $1152 = $sG$8 >>> 5;
     $1153 = $1152 << 2;
     $1154 = $1153 | $1151;
     $1155 = $sB$8 >>> 6;
     $1156 = $1154 | $1155;
     $1157 = (($12) + ($1156)|0);
     $1158 = HEAP8[$1157>>0]|0;
     HEAP8[$dst$10>>0] = $1158;
     $1159 = ((($dst$10)) + 1|0);
     $1160 = (($src$10) + ($17)|0);
     $dst$11 = $1159;$n1$1 = $n1$0;$src$11 = $1160;
     label = 64;
     continue;
    }
    else if ((label|0) == 64) {
     label = 0;
     switch ($17|0) {
     case 1:  {
      $1161 = HEAP8[$src$11>>0]|0;
      $1162 = $1161&255;
      $1163 = HEAP32[$24>>2]|0;
      $1164 = $1162 & $1163;
      $1165 = HEAP32[$25>>2]|0;
      $1166 = $1165 & 255;
      $1167 = $1164 >>> $1166;
      $1168 = HEAP32[$26>>2]|0;
      $1169 = $1168 & 255;
      $1170 = (125924 + ($1169<<2)|0);
      $1171 = HEAP32[$1170>>2]|0;
      $1172 = (($1171) + ($1167)|0);
      $1173 = HEAP8[$1172>>0]|0;
      $1174 = HEAP32[$27>>2]|0;
      $1175 = $1174 & $1162;
      $1176 = $1165 >>> 8;
      $1177 = $1176 & 255;
      $1178 = $1175 >>> $1177;
      $1179 = $1168 >>> 8;
      $1180 = $1179 & 255;
      $1181 = (125924 + ($1180<<2)|0);
      $1182 = HEAP32[$1181>>2]|0;
      $1183 = (($1182) + ($1178)|0);
      $1184 = HEAP8[$1183>>0]|0;
      $1185 = HEAP32[$28>>2]|0;
      $1186 = $1185 & $1162;
      $1187 = $1165 >>> 16;
      $1188 = $1187 & 255;
      $1189 = $1186 >>> $1188;
      $1190 = $1168 >>> 16;
      $1191 = $1190 & 255;
      $1192 = (125924 + ($1191<<2)|0);
      $1193 = HEAP32[$1192>>2]|0;
      $1194 = (($1193) + ($1189)|0);
      $1195 = HEAP8[$1194>>0]|0;
      $sB$9$shrunk = $1195;$sG$9$shrunk = $1184;$sR$9$shrunk = $1173;
      break;
     }
     case 2:  {
      $1196 = HEAP16[$src$11>>1]|0;
      $1197 = $1196&65535;
      $1198 = HEAP32[$24>>2]|0;
      $1199 = $1197 & $1198;
      $1200 = HEAP32[$25>>2]|0;
      $1201 = $1200 & 255;
      $1202 = $1199 >>> $1201;
      $1203 = HEAP32[$26>>2]|0;
      $1204 = $1203 & 255;
      $1205 = (125924 + ($1204<<2)|0);
      $1206 = HEAP32[$1205>>2]|0;
      $1207 = (($1206) + ($1202)|0);
      $1208 = HEAP8[$1207>>0]|0;
      $1209 = HEAP32[$27>>2]|0;
      $1210 = $1209 & $1197;
      $1211 = $1200 >>> 8;
      $1212 = $1211 & 255;
      $1213 = $1210 >>> $1212;
      $1214 = $1203 >>> 8;
      $1215 = $1214 & 255;
      $1216 = (125924 + ($1215<<2)|0);
      $1217 = HEAP32[$1216>>2]|0;
      $1218 = (($1217) + ($1213)|0);
      $1219 = HEAP8[$1218>>0]|0;
      $1220 = HEAP32[$28>>2]|0;
      $1221 = $1220 & $1197;
      $1222 = $1200 >>> 16;
      $1223 = $1222 & 255;
      $1224 = $1221 >>> $1223;
      $1225 = $1203 >>> 16;
      $1226 = $1225 & 255;
      $1227 = (125924 + ($1226<<2)|0);
      $1228 = HEAP32[$1227>>2]|0;
      $1229 = (($1228) + ($1224)|0);
      $1230 = HEAP8[$1229>>0]|0;
      $sB$9$shrunk = $1230;$sG$9$shrunk = $1219;$sR$9$shrunk = $1208;
      break;
     }
     case 3:  {
      $1231 = HEAP32[$25>>2]|0;
      $div15 = $1231 >>> 3;
      $1232 = $div15 & 31;
      $1233 = (($src$11) + ($1232)|0);
      $1234 = HEAP8[$1233>>0]|0;
      $1235 = $1231 >>> 11;
      $div16 = $1235 & 31;
      $1236 = (($src$11) + ($div16)|0);
      $1237 = HEAP8[$1236>>0]|0;
      $1238 = $1231 >>> 19;
      $1239 = $1238 & 31;
      $1240 = (($src$11) + ($1239)|0);
      $1241 = HEAP8[$1240>>0]|0;
      $sB$9$shrunk = $1241;$sG$9$shrunk = $1237;$sR$9$shrunk = $1234;
      break;
     }
     case 4:  {
      $1242 = HEAP32[$src$11>>2]|0;
      $1243 = HEAP32[$24>>2]|0;
      $1244 = $1243 & $1242;
      $1245 = HEAP32[$25>>2]|0;
      $1246 = $1245 & 255;
      $1247 = $1244 >>> $1246;
      $1248 = HEAP32[$26>>2]|0;
      $1249 = $1248 & 255;
      $1250 = (125924 + ($1249<<2)|0);
      $1251 = HEAP32[$1250>>2]|0;
      $1252 = (($1251) + ($1247)|0);
      $1253 = HEAP8[$1252>>0]|0;
      $1254 = HEAP32[$27>>2]|0;
      $1255 = $1254 & $1242;
      $1256 = $1245 >>> 8;
      $1257 = $1256 & 255;
      $1258 = $1255 >>> $1257;
      $1259 = $1248 >>> 8;
      $1260 = $1259 & 255;
      $1261 = (125924 + ($1260<<2)|0);
      $1262 = HEAP32[$1261>>2]|0;
      $1263 = (($1262) + ($1258)|0);
      $1264 = HEAP8[$1263>>0]|0;
      $1265 = HEAP32[$28>>2]|0;
      $1266 = $1265 & $1242;
      $1267 = $1245 >>> 16;
      $1268 = $1267 & 255;
      $1269 = $1266 >>> $1268;
      $1270 = $1248 >>> 16;
      $1271 = $1270 & 255;
      $1272 = (125924 + ($1271<<2)|0);
      $1273 = HEAP32[$1272>>2]|0;
      $1274 = (($1273) + ($1269)|0);
      $1275 = HEAP8[$1274>>0]|0;
      $sB$9$shrunk = $1275;$sG$9$shrunk = $1264;$sR$9$shrunk = $1253;
      break;
     }
     default: {
      $sB$9$shrunk = 0;$sG$9$shrunk = 0;$sR$9$shrunk = 0;
     }
     }
     $sB$9 = $sB$9$shrunk&255;
     $sG$9 = $sG$9$shrunk&255;
     $sR$9 = $sR$9$shrunk&255;
     $1276 = $sR$9 & 224;
     $1277 = $sG$9 >>> 5;
     $1278 = $1277 << 2;
     $1279 = $1278 | $1276;
     $1280 = $sB$9 >>> 6;
     $1281 = $1279 | $1280;
     $1282 = (($12) + ($1281)|0);
     $1283 = HEAP8[$1282>>0]|0;
     HEAP8[$dst$11>>0] = $1283;
     $1284 = ((($dst$11)) + 1|0);
     $1285 = (($src$11) + ($17)|0);
     $dst$12 = $1284;$n1$2 = $n1$1;$src$12 = $1285;
     label = 70;
     continue;
    }
    else if ((label|0) == 70) {
     label = 0;
     switch ($17|0) {
     case 1:  {
      $1286 = HEAP8[$src$12>>0]|0;
      $1287 = $1286&255;
      $1288 = HEAP32[$24>>2]|0;
      $1289 = $1287 & $1288;
      $1290 = HEAP32[$25>>2]|0;
      $1291 = $1290 & 255;
      $1292 = $1289 >>> $1291;
      $1293 = HEAP32[$26>>2]|0;
      $1294 = $1293 & 255;
      $1295 = (125924 + ($1294<<2)|0);
      $1296 = HEAP32[$1295>>2]|0;
      $1297 = (($1296) + ($1292)|0);
      $1298 = HEAP8[$1297>>0]|0;
      $1299 = HEAP32[$27>>2]|0;
      $1300 = $1299 & $1287;
      $1301 = $1290 >>> 8;
      $1302 = $1301 & 255;
      $1303 = $1300 >>> $1302;
      $1304 = $1293 >>> 8;
      $1305 = $1304 & 255;
      $1306 = (125924 + ($1305<<2)|0);
      $1307 = HEAP32[$1306>>2]|0;
      $1308 = (($1307) + ($1303)|0);
      $1309 = HEAP8[$1308>>0]|0;
      $1310 = HEAP32[$28>>2]|0;
      $1311 = $1310 & $1287;
      $1312 = $1290 >>> 16;
      $1313 = $1312 & 255;
      $1314 = $1311 >>> $1313;
      $1315 = $1293 >>> 16;
      $1316 = $1315 & 255;
      $1317 = (125924 + ($1316<<2)|0);
      $1318 = HEAP32[$1317>>2]|0;
      $1319 = (($1318) + ($1314)|0);
      $1320 = HEAP8[$1319>>0]|0;
      $sB$10$shrunk = $1320;$sG$10$shrunk = $1309;$sR$10$shrunk = $1298;
      break;
     }
     case 2:  {
      $1321 = HEAP16[$src$12>>1]|0;
      $1322 = $1321&65535;
      $1323 = HEAP32[$24>>2]|0;
      $1324 = $1322 & $1323;
      $1325 = HEAP32[$25>>2]|0;
      $1326 = $1325 & 255;
      $1327 = $1324 >>> $1326;
      $1328 = HEAP32[$26>>2]|0;
      $1329 = $1328 & 255;
      $1330 = (125924 + ($1329<<2)|0);
      $1331 = HEAP32[$1330>>2]|0;
      $1332 = (($1331) + ($1327)|0);
      $1333 = HEAP8[$1332>>0]|0;
      $1334 = HEAP32[$27>>2]|0;
      $1335 = $1334 & $1322;
      $1336 = $1325 >>> 8;
      $1337 = $1336 & 255;
      $1338 = $1335 >>> $1337;
      $1339 = $1328 >>> 8;
      $1340 = $1339 & 255;
      $1341 = (125924 + ($1340<<2)|0);
      $1342 = HEAP32[$1341>>2]|0;
      $1343 = (($1342) + ($1338)|0);
      $1344 = HEAP8[$1343>>0]|0;
      $1345 = HEAP32[$28>>2]|0;
      $1346 = $1345 & $1322;
      $1347 = $1325 >>> 16;
      $1348 = $1347 & 255;
      $1349 = $1346 >>> $1348;
      $1350 = $1328 >>> 16;
      $1351 = $1350 & 255;
      $1352 = (125924 + ($1351<<2)|0);
      $1353 = HEAP32[$1352>>2]|0;
      $1354 = (($1353) + ($1349)|0);
      $1355 = HEAP8[$1354>>0]|0;
      $sB$10$shrunk = $1355;$sG$10$shrunk = $1344;$sR$10$shrunk = $1333;
      break;
     }
     case 3:  {
      $1356 = HEAP32[$25>>2]|0;
      $div12 = $1356 >>> 3;
      $1357 = $div12 & 31;
      $1358 = (($src$12) + ($1357)|0);
      $1359 = HEAP8[$1358>>0]|0;
      $1360 = $1356 >>> 11;
      $div13 = $1360 & 31;
      $1361 = (($src$12) + ($div13)|0);
      $1362 = HEAP8[$1361>>0]|0;
      $1363 = $1356 >>> 19;
      $1364 = $1363 & 31;
      $1365 = (($src$12) + ($1364)|0);
      $1366 = HEAP8[$1365>>0]|0;
      $sB$10$shrunk = $1366;$sG$10$shrunk = $1362;$sR$10$shrunk = $1359;
      break;
     }
     case 4:  {
      $1367 = HEAP32[$src$12>>2]|0;
      $1368 = HEAP32[$24>>2]|0;
      $1369 = $1368 & $1367;
      $1370 = HEAP32[$25>>2]|0;
      $1371 = $1370 & 255;
      $1372 = $1369 >>> $1371;
      $1373 = HEAP32[$26>>2]|0;
      $1374 = $1373 & 255;
      $1375 = (125924 + ($1374<<2)|0);
      $1376 = HEAP32[$1375>>2]|0;
      $1377 = (($1376) + ($1372)|0);
      $1378 = HEAP8[$1377>>0]|0;
      $1379 = HEAP32[$27>>2]|0;
      $1380 = $1379 & $1367;
      $1381 = $1370 >>> 8;
      $1382 = $1381 & 255;
      $1383 = $1380 >>> $1382;
      $1384 = $1373 >>> 8;
      $1385 = $1384 & 255;
      $1386 = (125924 + ($1385<<2)|0);
      $1387 = HEAP32[$1386>>2]|0;
      $1388 = (($1387) + ($1383)|0);
      $1389 = HEAP8[$1388>>0]|0;
      $1390 = HEAP32[$28>>2]|0;
      $1391 = $1390 & $1367;
      $1392 = $1370 >>> 16;
      $1393 = $1392 & 255;
      $1394 = $1391 >>> $1393;
      $1395 = $1373 >>> 16;
      $1396 = $1395 & 255;
      $1397 = (125924 + ($1396<<2)|0);
      $1398 = HEAP32[$1397>>2]|0;
      $1399 = (($1398) + ($1394)|0);
      $1400 = HEAP8[$1399>>0]|0;
      $sB$10$shrunk = $1400;$sG$10$shrunk = $1389;$sR$10$shrunk = $1378;
      break;
     }
     default: {
      $sB$10$shrunk = 0;$sG$10$shrunk = 0;$sR$10$shrunk = 0;
     }
     }
     $sB$10 = $sB$10$shrunk&255;
     $sG$10 = $sG$10$shrunk&255;
     $sR$10 = $sR$10$shrunk&255;
     $1401 = $sR$10 & 224;
     $1402 = $sG$10 >>> 5;
     $1403 = $1402 << 2;
     $1404 = $1403 | $1401;
     $1405 = $sB$10 >>> 6;
     $1406 = $1404 | $1405;
     $1407 = (($12) + ($1406)|0);
     $1408 = HEAP8[$1407>>0]|0;
     HEAP8[$dst$12>>0] = $1408;
     $1409 = ((($dst$12)) + 1|0);
     $1410 = (($src$12) + ($17)|0);
     $dst$13 = $1409;$n1$3 = $n1$2;$src$13 = $1410;
     label = 76;
     continue;
    }
    else if ((label|0) == 76) {
     label = 0;
     switch ($17|0) {
     case 1:  {
      $1411 = HEAP8[$src$13>>0]|0;
      $1412 = $1411&255;
      $1413 = HEAP32[$24>>2]|0;
      $1414 = $1412 & $1413;
      $1415 = HEAP32[$25>>2]|0;
      $1416 = $1415 & 255;
      $1417 = $1414 >>> $1416;
      $1418 = HEAP32[$26>>2]|0;
      $1419 = $1418 & 255;
      $1420 = (125924 + ($1419<<2)|0);
      $1421 = HEAP32[$1420>>2]|0;
      $1422 = (($1421) + ($1417)|0);
      $1423 = HEAP8[$1422>>0]|0;
      $1424 = HEAP32[$27>>2]|0;
      $1425 = $1424 & $1412;
      $1426 = $1415 >>> 8;
      $1427 = $1426 & 255;
      $1428 = $1425 >>> $1427;
      $1429 = $1418 >>> 8;
      $1430 = $1429 & 255;
      $1431 = (125924 + ($1430<<2)|0);
      $1432 = HEAP32[$1431>>2]|0;
      $1433 = (($1432) + ($1428)|0);
      $1434 = HEAP8[$1433>>0]|0;
      $1435 = HEAP32[$28>>2]|0;
      $1436 = $1435 & $1412;
      $1437 = $1415 >>> 16;
      $1438 = $1437 & 255;
      $1439 = $1436 >>> $1438;
      $1440 = $1418 >>> 16;
      $1441 = $1440 & 255;
      $1442 = (125924 + ($1441<<2)|0);
      $1443 = HEAP32[$1442>>2]|0;
      $1444 = (($1443) + ($1439)|0);
      $1445 = HEAP8[$1444>>0]|0;
      $sB$11$shrunk = $1445;$sG$11$shrunk = $1434;$sR$11$shrunk = $1423;
      break;
     }
     case 2:  {
      $1446 = HEAP16[$src$13>>1]|0;
      $1447 = $1446&65535;
      $1448 = HEAP32[$24>>2]|0;
      $1449 = $1447 & $1448;
      $1450 = HEAP32[$25>>2]|0;
      $1451 = $1450 & 255;
      $1452 = $1449 >>> $1451;
      $1453 = HEAP32[$26>>2]|0;
      $1454 = $1453 & 255;
      $1455 = (125924 + ($1454<<2)|0);
      $1456 = HEAP32[$1455>>2]|0;
      $1457 = (($1456) + ($1452)|0);
      $1458 = HEAP8[$1457>>0]|0;
      $1459 = HEAP32[$27>>2]|0;
      $1460 = $1459 & $1447;
      $1461 = $1450 >>> 8;
      $1462 = $1461 & 255;
      $1463 = $1460 >>> $1462;
      $1464 = $1453 >>> 8;
      $1465 = $1464 & 255;
      $1466 = (125924 + ($1465<<2)|0);
      $1467 = HEAP32[$1466>>2]|0;
      $1468 = (($1467) + ($1463)|0);
      $1469 = HEAP8[$1468>>0]|0;
      $1470 = HEAP32[$28>>2]|0;
      $1471 = $1470 & $1447;
      $1472 = $1450 >>> 16;
      $1473 = $1472 & 255;
      $1474 = $1471 >>> $1473;
      $1475 = $1453 >>> 16;
      $1476 = $1475 & 255;
      $1477 = (125924 + ($1476<<2)|0);
      $1478 = HEAP32[$1477>>2]|0;
      $1479 = (($1478) + ($1474)|0);
      $1480 = HEAP8[$1479>>0]|0;
      $sB$11$shrunk = $1480;$sG$11$shrunk = $1469;$sR$11$shrunk = $1458;
      break;
     }
     case 3:  {
      $1481 = HEAP32[$25>>2]|0;
      $div9 = $1481 >>> 3;
      $1482 = $div9 & 31;
      $1483 = (($src$13) + ($1482)|0);
      $1484 = HEAP8[$1483>>0]|0;
      $1485 = $1481 >>> 11;
      $div10 = $1485 & 31;
      $1486 = (($src$13) + ($div10)|0);
      $1487 = HEAP8[$1486>>0]|0;
      $1488 = $1481 >>> 19;
      $1489 = $1488 & 31;
      $1490 = (($src$13) + ($1489)|0);
      $1491 = HEAP8[$1490>>0]|0;
      $sB$11$shrunk = $1491;$sG$11$shrunk = $1487;$sR$11$shrunk = $1484;
      break;
     }
     case 4:  {
      $1492 = HEAP32[$src$13>>2]|0;
      $1493 = HEAP32[$24>>2]|0;
      $1494 = $1493 & $1492;
      $1495 = HEAP32[$25>>2]|0;
      $1496 = $1495 & 255;
      $1497 = $1494 >>> $1496;
      $1498 = HEAP32[$26>>2]|0;
      $1499 = $1498 & 255;
      $1500 = (125924 + ($1499<<2)|0);
      $1501 = HEAP32[$1500>>2]|0;
      $1502 = (($1501) + ($1497)|0);
      $1503 = HEAP8[$1502>>0]|0;
      $1504 = HEAP32[$27>>2]|0;
      $1505 = $1504 & $1492;
      $1506 = $1495 >>> 8;
      $1507 = $1506 & 255;
      $1508 = $1505 >>> $1507;
      $1509 = $1498 >>> 8;
      $1510 = $1509 & 255;
      $1511 = (125924 + ($1510<<2)|0);
      $1512 = HEAP32[$1511>>2]|0;
      $1513 = (($1512) + ($1508)|0);
      $1514 = HEAP8[$1513>>0]|0;
      $1515 = HEAP32[$28>>2]|0;
      $1516 = $1515 & $1492;
      $1517 = $1495 >>> 16;
      $1518 = $1517 & 255;
      $1519 = $1516 >>> $1518;
      $1520 = $1498 >>> 16;
      $1521 = $1520 & 255;
      $1522 = (125924 + ($1521<<2)|0);
      $1523 = HEAP32[$1522>>2]|0;
      $1524 = (($1523) + ($1519)|0);
      $1525 = HEAP8[$1524>>0]|0;
      $sB$11$shrunk = $1525;$sG$11$shrunk = $1514;$sR$11$shrunk = $1503;
      break;
     }
     default: {
      $sB$11$shrunk = 0;$sG$11$shrunk = 0;$sR$11$shrunk = 0;
     }
     }
     $sB$11 = $sB$11$shrunk&255;
     $sG$11 = $sG$11$shrunk&255;
     $sR$11 = $sR$11$shrunk&255;
     $1526 = $sR$11 & 224;
     $1527 = $sG$11 >>> 5;
     $1528 = $1527 << 2;
     $1529 = $1528 | $1526;
     $1530 = $sB$11 >>> 6;
     $1531 = $1529 | $1530;
     $1532 = (($12) + ($1531)|0);
     $1533 = HEAP8[$1532>>0]|0;
     HEAP8[$dst$13>>0] = $1533;
     $1534 = ((($dst$13)) + 1|0);
     $1535 = (($src$13) + ($17)|0);
     $dst$14 = $1534;$n1$4 = $n1$3;$src$14 = $1535;
     label = 82;
     continue;
    }
    else if ((label|0) == 82) {
     label = 0;
     switch ($17|0) {
     case 1:  {
      $1536 = HEAP8[$src$14>>0]|0;
      $1537 = $1536&255;
      $1538 = HEAP32[$24>>2]|0;
      $1539 = $1537 & $1538;
      $1540 = HEAP32[$25>>2]|0;
      $1541 = $1540 & 255;
      $1542 = $1539 >>> $1541;
      $1543 = HEAP32[$26>>2]|0;
      $1544 = $1543 & 255;
      $1545 = (125924 + ($1544<<2)|0);
      $1546 = HEAP32[$1545>>2]|0;
      $1547 = (($1546) + ($1542)|0);
      $1548 = HEAP8[$1547>>0]|0;
      $1549 = HEAP32[$27>>2]|0;
      $1550 = $1549 & $1537;
      $1551 = $1540 >>> 8;
      $1552 = $1551 & 255;
      $1553 = $1550 >>> $1552;
      $1554 = $1543 >>> 8;
      $1555 = $1554 & 255;
      $1556 = (125924 + ($1555<<2)|0);
      $1557 = HEAP32[$1556>>2]|0;
      $1558 = (($1557) + ($1553)|0);
      $1559 = HEAP8[$1558>>0]|0;
      $1560 = HEAP32[$28>>2]|0;
      $1561 = $1560 & $1537;
      $1562 = $1540 >>> 16;
      $1563 = $1562 & 255;
      $1564 = $1561 >>> $1563;
      $1565 = $1543 >>> 16;
      $1566 = $1565 & 255;
      $1567 = (125924 + ($1566<<2)|0);
      $1568 = HEAP32[$1567>>2]|0;
      $1569 = (($1568) + ($1564)|0);
      $1570 = HEAP8[$1569>>0]|0;
      $sB$12$shrunk = $1570;$sG$12$shrunk = $1559;$sR$12$shrunk = $1548;
      break;
     }
     case 2:  {
      $1571 = HEAP16[$src$14>>1]|0;
      $1572 = $1571&65535;
      $1573 = HEAP32[$24>>2]|0;
      $1574 = $1572 & $1573;
      $1575 = HEAP32[$25>>2]|0;
      $1576 = $1575 & 255;
      $1577 = $1574 >>> $1576;
      $1578 = HEAP32[$26>>2]|0;
      $1579 = $1578 & 255;
      $1580 = (125924 + ($1579<<2)|0);
      $1581 = HEAP32[$1580>>2]|0;
      $1582 = (($1581) + ($1577)|0);
      $1583 = HEAP8[$1582>>0]|0;
      $1584 = HEAP32[$27>>2]|0;
      $1585 = $1584 & $1572;
      $1586 = $1575 >>> 8;
      $1587 = $1586 & 255;
      $1588 = $1585 >>> $1587;
      $1589 = $1578 >>> 8;
      $1590 = $1589 & 255;
      $1591 = (125924 + ($1590<<2)|0);
      $1592 = HEAP32[$1591>>2]|0;
      $1593 = (($1592) + ($1588)|0);
      $1594 = HEAP8[$1593>>0]|0;
      $1595 = HEAP32[$28>>2]|0;
      $1596 = $1595 & $1572;
      $1597 = $1575 >>> 16;
      $1598 = $1597 & 255;
      $1599 = $1596 >>> $1598;
      $1600 = $1578 >>> 16;
      $1601 = $1600 & 255;
      $1602 = (125924 + ($1601<<2)|0);
      $1603 = HEAP32[$1602>>2]|0;
      $1604 = (($1603) + ($1599)|0);
      $1605 = HEAP8[$1604>>0]|0;
      $sB$12$shrunk = $1605;$sG$12$shrunk = $1594;$sR$12$shrunk = $1583;
      break;
     }
     case 3:  {
      $1606 = HEAP32[$25>>2]|0;
      $div6 = $1606 >>> 3;
      $1607 = $div6 & 31;
      $1608 = (($src$14) + ($1607)|0);
      $1609 = HEAP8[$1608>>0]|0;
      $1610 = $1606 >>> 11;
      $div7 = $1610 & 31;
      $1611 = (($src$14) + ($div7)|0);
      $1612 = HEAP8[$1611>>0]|0;
      $1613 = $1606 >>> 19;
      $1614 = $1613 & 31;
      $1615 = (($src$14) + ($1614)|0);
      $1616 = HEAP8[$1615>>0]|0;
      $sB$12$shrunk = $1616;$sG$12$shrunk = $1612;$sR$12$shrunk = $1609;
      break;
     }
     case 4:  {
      $1617 = HEAP32[$src$14>>2]|0;
      $1618 = HEAP32[$24>>2]|0;
      $1619 = $1618 & $1617;
      $1620 = HEAP32[$25>>2]|0;
      $1621 = $1620 & 255;
      $1622 = $1619 >>> $1621;
      $1623 = HEAP32[$26>>2]|0;
      $1624 = $1623 & 255;
      $1625 = (125924 + ($1624<<2)|0);
      $1626 = HEAP32[$1625>>2]|0;
      $1627 = (($1626) + ($1622)|0);
      $1628 = HEAP8[$1627>>0]|0;
      $1629 = HEAP32[$27>>2]|0;
      $1630 = $1629 & $1617;
      $1631 = $1620 >>> 8;
      $1632 = $1631 & 255;
      $1633 = $1630 >>> $1632;
      $1634 = $1623 >>> 8;
      $1635 = $1634 & 255;
      $1636 = (125924 + ($1635<<2)|0);
      $1637 = HEAP32[$1636>>2]|0;
      $1638 = (($1637) + ($1633)|0);
      $1639 = HEAP8[$1638>>0]|0;
      $1640 = HEAP32[$28>>2]|0;
      $1641 = $1640 & $1617;
      $1642 = $1620 >>> 16;
      $1643 = $1642 & 255;
      $1644 = $1641 >>> $1643;
      $1645 = $1623 >>> 16;
      $1646 = $1645 & 255;
      $1647 = (125924 + ($1646<<2)|0);
      $1648 = HEAP32[$1647>>2]|0;
      $1649 = (($1648) + ($1644)|0);
      $1650 = HEAP8[$1649>>0]|0;
      $sB$12$shrunk = $1650;$sG$12$shrunk = $1639;$sR$12$shrunk = $1628;
      break;
     }
     default: {
      $sB$12$shrunk = 0;$sG$12$shrunk = 0;$sR$12$shrunk = 0;
     }
     }
     $sB$12 = $sB$12$shrunk&255;
     $sG$12 = $sG$12$shrunk&255;
     $sR$12 = $sR$12$shrunk&255;
     $1651 = $sR$12 & 224;
     $1652 = $sG$12 >>> 5;
     $1653 = $1652 << 2;
     $1654 = $1653 | $1651;
     $1655 = $sB$12 >>> 6;
     $1656 = $1654 | $1655;
     $1657 = (($12) + ($1656)|0);
     $1658 = HEAP8[$1657>>0]|0;
     HEAP8[$dst$14>>0] = $1658;
     $1659 = ((($dst$14)) + 1|0);
     $1660 = (($src$14) + ($17)|0);
     $dst$15 = $1659;$n1$5 = $n1$4;$src$15 = $1660;
     label = 88;
     continue;
    }
    else if ((label|0) == 88) {
     label = 0;
     switch ($17|0) {
     case 1:  {
      $1661 = HEAP8[$src$15>>0]|0;
      $1662 = $1661&255;
      $1663 = HEAP32[$24>>2]|0;
      $1664 = $1662 & $1663;
      $1665 = HEAP32[$25>>2]|0;
      $1666 = $1665 & 255;
      $1667 = $1664 >>> $1666;
      $1668 = HEAP32[$26>>2]|0;
      $1669 = $1668 & 255;
      $1670 = (125924 + ($1669<<2)|0);
      $1671 = HEAP32[$1670>>2]|0;
      $1672 = (($1671) + ($1667)|0);
      $1673 = HEAP8[$1672>>0]|0;
      $1674 = HEAP32[$27>>2]|0;
      $1675 = $1674 & $1662;
      $1676 = $1665 >>> 8;
      $1677 = $1676 & 255;
      $1678 = $1675 >>> $1677;
      $1679 = $1668 >>> 8;
      $1680 = $1679 & 255;
      $1681 = (125924 + ($1680<<2)|0);
      $1682 = HEAP32[$1681>>2]|0;
      $1683 = (($1682) + ($1678)|0);
      $1684 = HEAP8[$1683>>0]|0;
      $1685 = HEAP32[$28>>2]|0;
      $1686 = $1685 & $1662;
      $1687 = $1665 >>> 16;
      $1688 = $1687 & 255;
      $1689 = $1686 >>> $1688;
      $1690 = $1668 >>> 16;
      $1691 = $1690 & 255;
      $1692 = (125924 + ($1691<<2)|0);
      $1693 = HEAP32[$1692>>2]|0;
      $1694 = (($1693) + ($1689)|0);
      $1695 = HEAP8[$1694>>0]|0;
      $sB$13$shrunk = $1695;$sG$13$shrunk = $1684;$sR$13$shrunk = $1673;
      break;
     }
     case 2:  {
      $1696 = HEAP16[$src$15>>1]|0;
      $1697 = $1696&65535;
      $1698 = HEAP32[$24>>2]|0;
      $1699 = $1697 & $1698;
      $1700 = HEAP32[$25>>2]|0;
      $1701 = $1700 & 255;
      $1702 = $1699 >>> $1701;
      $1703 = HEAP32[$26>>2]|0;
      $1704 = $1703 & 255;
      $1705 = (125924 + ($1704<<2)|0);
      $1706 = HEAP32[$1705>>2]|0;
      $1707 = (($1706) + ($1702)|0);
      $1708 = HEAP8[$1707>>0]|0;
      $1709 = HEAP32[$27>>2]|0;
      $1710 = $1709 & $1697;
      $1711 = $1700 >>> 8;
      $1712 = $1711 & 255;
      $1713 = $1710 >>> $1712;
      $1714 = $1703 >>> 8;
      $1715 = $1714 & 255;
      $1716 = (125924 + ($1715<<2)|0);
      $1717 = HEAP32[$1716>>2]|0;
      $1718 = (($1717) + ($1713)|0);
      $1719 = HEAP8[$1718>>0]|0;
      $1720 = HEAP32[$28>>2]|0;
      $1721 = $1720 & $1697;
      $1722 = $1700 >>> 16;
      $1723 = $1722 & 255;
      $1724 = $1721 >>> $1723;
      $1725 = $1703 >>> 16;
      $1726 = $1725 & 255;
      $1727 = (125924 + ($1726<<2)|0);
      $1728 = HEAP32[$1727>>2]|0;
      $1729 = (($1728) + ($1724)|0);
      $1730 = HEAP8[$1729>>0]|0;
      $sB$13$shrunk = $1730;$sG$13$shrunk = $1719;$sR$13$shrunk = $1708;
      break;
     }
     case 3:  {
      $1731 = HEAP32[$25>>2]|0;
      $div3 = $1731 >>> 3;
      $1732 = $div3 & 31;
      $1733 = (($src$15) + ($1732)|0);
      $1734 = HEAP8[$1733>>0]|0;
      $1735 = $1731 >>> 11;
      $div4 = $1735 & 31;
      $1736 = (($src$15) + ($div4)|0);
      $1737 = HEAP8[$1736>>0]|0;
      $1738 = $1731 >>> 19;
      $1739 = $1738 & 31;
      $1740 = (($src$15) + ($1739)|0);
      $1741 = HEAP8[$1740>>0]|0;
      $sB$13$shrunk = $1741;$sG$13$shrunk = $1737;$sR$13$shrunk = $1734;
      break;
     }
     case 4:  {
      $1742 = HEAP32[$src$15>>2]|0;
      $1743 = HEAP32[$24>>2]|0;
      $1744 = $1743 & $1742;
      $1745 = HEAP32[$25>>2]|0;
      $1746 = $1745 & 255;
      $1747 = $1744 >>> $1746;
      $1748 = HEAP32[$26>>2]|0;
      $1749 = $1748 & 255;
      $1750 = (125924 + ($1749<<2)|0);
      $1751 = HEAP32[$1750>>2]|0;
      $1752 = (($1751) + ($1747)|0);
      $1753 = HEAP8[$1752>>0]|0;
      $1754 = HEAP32[$27>>2]|0;
      $1755 = $1754 & $1742;
      $1756 = $1745 >>> 8;
      $1757 = $1756 & 255;
      $1758 = $1755 >>> $1757;
      $1759 = $1748 >>> 8;
      $1760 = $1759 & 255;
      $1761 = (125924 + ($1760<<2)|0);
      $1762 = HEAP32[$1761>>2]|0;
      $1763 = (($1762) + ($1758)|0);
      $1764 = HEAP8[$1763>>0]|0;
      $1765 = HEAP32[$28>>2]|0;
      $1766 = $1765 & $1742;
      $1767 = $1745 >>> 16;
      $1768 = $1767 & 255;
      $1769 = $1766 >>> $1768;
      $1770 = $1748 >>> 16;
      $1771 = $1770 & 255;
      $1772 = (125924 + ($1771<<2)|0);
      $1773 = HEAP32[$1772>>2]|0;
      $1774 = (($1773) + ($1769)|0);
      $1775 = HEAP8[$1774>>0]|0;
      $sB$13$shrunk = $1775;$sG$13$shrunk = $1764;$sR$13$shrunk = $1753;
      break;
     }
     default: {
      $sB$13$shrunk = 0;$sG$13$shrunk = 0;$sR$13$shrunk = 0;
     }
     }
     $sB$13 = $sB$13$shrunk&255;
     $sG$13 = $sG$13$shrunk&255;
     $sR$13 = $sR$13$shrunk&255;
     $1776 = $sR$13 & 224;
     $1777 = $sG$13 >>> 5;
     $1778 = $1777 << 2;
     $1779 = $1778 | $1776;
     $1780 = $sB$13 >>> 6;
     $1781 = $1779 | $1780;
     $1782 = (($12) + ($1781)|0);
     $1783 = HEAP8[$1782>>0]|0;
     HEAP8[$dst$15>>0] = $1783;
     $1784 = ((($dst$15)) + 1|0);
     $1785 = (($src$15) + ($17)|0);
     $dst$16 = $1784;$n1$6 = $n1$5;$src$16 = $1785;
     label = 94;
     continue;
    }
    else if ((label|0) == 94) {
     label = 0;
     switch ($17|0) {
     case 1:  {
      $1786 = HEAP8[$src$16>>0]|0;
      $1787 = $1786&255;
      $1788 = HEAP32[$24>>2]|0;
      $1789 = $1787 & $1788;
      $1790 = HEAP32[$25>>2]|0;
      $1791 = $1790 & 255;
      $1792 = $1789 >>> $1791;
      $1793 = HEAP32[$26>>2]|0;
      $1794 = $1793 & 255;
      $1795 = (125924 + ($1794<<2)|0);
      $1796 = HEAP32[$1795>>2]|0;
      $1797 = (($1796) + ($1792)|0);
      $1798 = HEAP8[$1797>>0]|0;
      $1799 = HEAP32[$27>>2]|0;
      $1800 = $1799 & $1787;
      $1801 = $1790 >>> 8;
      $1802 = $1801 & 255;
      $1803 = $1800 >>> $1802;
      $1804 = $1793 >>> 8;
      $1805 = $1804 & 255;
      $1806 = (125924 + ($1805<<2)|0);
      $1807 = HEAP32[$1806>>2]|0;
      $1808 = (($1807) + ($1803)|0);
      $1809 = HEAP8[$1808>>0]|0;
      $1810 = HEAP32[$28>>2]|0;
      $1811 = $1810 & $1787;
      $1812 = $1790 >>> 16;
      $1813 = $1812 & 255;
      $1814 = $1811 >>> $1813;
      $1815 = $1793 >>> 16;
      $1816 = $1815 & 255;
      $1817 = (125924 + ($1816<<2)|0);
      $1818 = HEAP32[$1817>>2]|0;
      $1819 = (($1818) + ($1814)|0);
      $1820 = HEAP8[$1819>>0]|0;
      $sB$14$shrunk = $1820;$sG$14$shrunk = $1809;$sR$14$shrunk = $1798;
      break;
     }
     case 2:  {
      $1821 = HEAP16[$src$16>>1]|0;
      $1822 = $1821&65535;
      $1823 = HEAP32[$24>>2]|0;
      $1824 = $1822 & $1823;
      $1825 = HEAP32[$25>>2]|0;
      $1826 = $1825 & 255;
      $1827 = $1824 >>> $1826;
      $1828 = HEAP32[$26>>2]|0;
      $1829 = $1828 & 255;
      $1830 = (125924 + ($1829<<2)|0);
      $1831 = HEAP32[$1830>>2]|0;
      $1832 = (($1831) + ($1827)|0);
      $1833 = HEAP8[$1832>>0]|0;
      $1834 = HEAP32[$27>>2]|0;
      $1835 = $1834 & $1822;
      $1836 = $1825 >>> 8;
      $1837 = $1836 & 255;
      $1838 = $1835 >>> $1837;
      $1839 = $1828 >>> 8;
      $1840 = $1839 & 255;
      $1841 = (125924 + ($1840<<2)|0);
      $1842 = HEAP32[$1841>>2]|0;
      $1843 = (($1842) + ($1838)|0);
      $1844 = HEAP8[$1843>>0]|0;
      $1845 = HEAP32[$28>>2]|0;
      $1846 = $1845 & $1822;
      $1847 = $1825 >>> 16;
      $1848 = $1847 & 255;
      $1849 = $1846 >>> $1848;
      $1850 = $1828 >>> 16;
      $1851 = $1850 & 255;
      $1852 = (125924 + ($1851<<2)|0);
      $1853 = HEAP32[$1852>>2]|0;
      $1854 = (($1853) + ($1849)|0);
      $1855 = HEAP8[$1854>>0]|0;
      $sB$14$shrunk = $1855;$sG$14$shrunk = $1844;$sR$14$shrunk = $1833;
      break;
     }
     case 3:  {
      $1856 = HEAP32[$25>>2]|0;
      $div = $1856 >>> 3;
      $1857 = $div & 31;
      $1858 = (($src$16) + ($1857)|0);
      $1859 = HEAP8[$1858>>0]|0;
      $1860 = $1856 >>> 11;
      $div1 = $1860 & 31;
      $1861 = (($src$16) + ($div1)|0);
      $1862 = HEAP8[$1861>>0]|0;
      $1863 = $1856 >>> 19;
      $1864 = $1863 & 31;
      $1865 = (($src$16) + ($1864)|0);
      $1866 = HEAP8[$1865>>0]|0;
      $sB$14$shrunk = $1866;$sG$14$shrunk = $1862;$sR$14$shrunk = $1859;
      break;
     }
     case 4:  {
      $1867 = HEAP32[$src$16>>2]|0;
      $1868 = HEAP32[$24>>2]|0;
      $1869 = $1868 & $1867;
      $1870 = HEAP32[$25>>2]|0;
      $1871 = $1870 & 255;
      $1872 = $1869 >>> $1871;
      $1873 = HEAP32[$26>>2]|0;
      $1874 = $1873 & 255;
      $1875 = (125924 + ($1874<<2)|0);
      $1876 = HEAP32[$1875>>2]|0;
      $1877 = (($1876) + ($1872)|0);
      $1878 = HEAP8[$1877>>0]|0;
      $1879 = HEAP32[$27>>2]|0;
      $1880 = $1879 & $1867;
      $1881 = $1870 >>> 8;
      $1882 = $1881 & 255;
      $1883 = $1880 >>> $1882;
      $1884 = $1873 >>> 8;
      $1885 = $1884 & 255;
      $1886 = (125924 + ($1885<<2)|0);
      $1887 = HEAP32[$1886>>2]|0;
      $1888 = (($1887) + ($1883)|0);
      $1889 = HEAP8[$1888>>0]|0;
      $1890 = HEAP32[$28>>2]|0;
      $1891 = $1890 & $1867;
      $1892 = $1870 >>> 16;
      $1893 = $1892 & 255;
      $1894 = $1891 >>> $1893;
      $1895 = $1873 >>> 16;
      $1896 = $1895 & 255;
      $1897 = (125924 + ($1896<<2)|0);
      $1898 = HEAP32[$1897>>2]|0;
      $1899 = (($1898) + ($1894)|0);
      $1900 = HEAP8[$1899>>0]|0;
      $sB$14$shrunk = $1900;$sG$14$shrunk = $1889;$sR$14$shrunk = $1878;
      break;
     }
     default: {
      $sB$14$shrunk = 0;$sG$14$shrunk = 0;$sR$14$shrunk = 0;
     }
     }
     $sB$14 = $sB$14$shrunk&255;
     $sG$14 = $sG$14$shrunk&255;
     $sR$14 = $sR$14$shrunk&255;
     $1901 = $sR$14 & 224;
     $1902 = $sG$14 >>> 5;
     $1903 = $1902 << 2;
     $1904 = $1903 | $1901;
     $1905 = $sB$14 >>> 6;
     $1906 = $1904 | $1905;
     $1907 = (($12) + ($1906)|0);
     $1908 = HEAP8[$1907>>0]|0;
     HEAP8[$dst$16>>0] = $1908;
     $1909 = ((($dst$16)) + 1|0);
     $1910 = (($src$16) + ($17)|0);
     $dst$17 = $1909;$n1$7 = $n1$6;$src$17 = $1910;
     label = 100;
     continue;
    }
    else if ((label|0) == 100) {
     label = 0;
     switch ($17|0) {
     case 1:  {
      $1911 = HEAP8[$src$17>>0]|0;
      $1912 = $1911&255;
      $1913 = HEAP32[$24>>2]|0;
      $1914 = $1912 & $1913;
      $1915 = HEAP32[$25>>2]|0;
      $1916 = $1915 & 255;
      $1917 = $1914 >>> $1916;
      $1918 = HEAP32[$26>>2]|0;
      $1919 = $1918 & 255;
      $1920 = (125924 + ($1919<<2)|0);
      $1921 = HEAP32[$1920>>2]|0;
      $1922 = (($1921) + ($1917)|0);
      $1923 = HEAP8[$1922>>0]|0;
      $1924 = HEAP32[$27>>2]|0;
      $1925 = $1924 & $1912;
      $1926 = $1915 >>> 8;
      $1927 = $1926 & 255;
      $1928 = $1925 >>> $1927;
      $1929 = $1918 >>> 8;
      $1930 = $1929 & 255;
      $1931 = (125924 + ($1930<<2)|0);
      $1932 = HEAP32[$1931>>2]|0;
      $1933 = (($1932) + ($1928)|0);
      $1934 = HEAP8[$1933>>0]|0;
      $1935 = HEAP32[$28>>2]|0;
      $1936 = $1935 & $1912;
      $1937 = $1915 >>> 16;
      $1938 = $1937 & 255;
      $1939 = $1936 >>> $1938;
      $1940 = $1918 >>> 16;
      $1941 = $1940 & 255;
      $1942 = (125924 + ($1941<<2)|0);
      $1943 = HEAP32[$1942>>2]|0;
      $1944 = (($1943) + ($1939)|0);
      $1945 = HEAP8[$1944>>0]|0;
      $sB$15$shrunk = $1945;$sG$15$shrunk = $1934;$sR$15$shrunk = $1923;
      break;
     }
     case 2:  {
      $1946 = HEAP16[$src$17>>1]|0;
      $1947 = $1946&65535;
      $1948 = HEAP32[$24>>2]|0;
      $1949 = $1947 & $1948;
      $1950 = HEAP32[$25>>2]|0;
      $1951 = $1950 & 255;
      $1952 = $1949 >>> $1951;
      $1953 = HEAP32[$26>>2]|0;
      $1954 = $1953 & 255;
      $1955 = (125924 + ($1954<<2)|0);
      $1956 = HEAP32[$1955>>2]|0;
      $1957 = (($1956) + ($1952)|0);
      $1958 = HEAP8[$1957>>0]|0;
      $1959 = HEAP32[$27>>2]|0;
      $1960 = $1959 & $1947;
      $1961 = $1950 >>> 8;
      $1962 = $1961 & 255;
      $1963 = $1960 >>> $1962;
      $1964 = $1953 >>> 8;
      $1965 = $1964 & 255;
      $1966 = (125924 + ($1965<<2)|0);
      $1967 = HEAP32[$1966>>2]|0;
      $1968 = (($1967) + ($1963)|0);
      $1969 = HEAP8[$1968>>0]|0;
      $1970 = HEAP32[$28>>2]|0;
      $1971 = $1970 & $1947;
      $1972 = $1950 >>> 16;
      $1973 = $1972 & 255;
      $1974 = $1971 >>> $1973;
      $1975 = $1953 >>> 16;
      $1976 = $1975 & 255;
      $1977 = (125924 + ($1976<<2)|0);
      $1978 = HEAP32[$1977>>2]|0;
      $1979 = (($1978) + ($1974)|0);
      $1980 = HEAP8[$1979>>0]|0;
      $sB$15$shrunk = $1980;$sG$15$shrunk = $1969;$sR$15$shrunk = $1958;
      break;
     }
     case 3:  {
      $1981 = HEAP32[$25>>2]|0;
      $div21 = $1981 >>> 3;
      $1982 = $div21 & 31;
      $1983 = (($src$17) + ($1982)|0);
      $1984 = HEAP8[$1983>>0]|0;
      $1985 = $1981 >>> 11;
      $div22 = $1985 & 31;
      $1986 = (($src$17) + ($div22)|0);
      $1987 = HEAP8[$1986>>0]|0;
      $1988 = $1981 >>> 19;
      $1989 = $1988 & 31;
      $1990 = (($src$17) + ($1989)|0);
      $1991 = HEAP8[$1990>>0]|0;
      $sB$15$shrunk = $1991;$sG$15$shrunk = $1987;$sR$15$shrunk = $1984;
      break;
     }
     case 4:  {
      $1992 = HEAP32[$src$17>>2]|0;
      $1993 = HEAP32[$24>>2]|0;
      $1994 = $1993 & $1992;
      $1995 = HEAP32[$25>>2]|0;
      $1996 = $1995 & 255;
      $1997 = $1994 >>> $1996;
      $1998 = HEAP32[$26>>2]|0;
      $1999 = $1998 & 255;
      $2000 = (125924 + ($1999<<2)|0);
      $2001 = HEAP32[$2000>>2]|0;
      $2002 = (($2001) + ($1997)|0);
      $2003 = HEAP8[$2002>>0]|0;
      $2004 = HEAP32[$27>>2]|0;
      $2005 = $2004 & $1992;
      $2006 = $1995 >>> 8;
      $2007 = $2006 & 255;
      $2008 = $2005 >>> $2007;
      $2009 = $1998 >>> 8;
      $2010 = $2009 & 255;
      $2011 = (125924 + ($2010<<2)|0);
      $2012 = HEAP32[$2011>>2]|0;
      $2013 = (($2012) + ($2008)|0);
      $2014 = HEAP8[$2013>>0]|0;
      $2015 = HEAP32[$28>>2]|0;
      $2016 = $2015 & $1992;
      $2017 = $1995 >>> 16;
      $2018 = $2017 & 255;
      $2019 = $2016 >>> $2018;
      $2020 = $1998 >>> 16;
      $2021 = $2020 & 255;
      $2022 = (125924 + ($2021<<2)|0);
      $2023 = HEAP32[$2022>>2]|0;
      $2024 = (($2023) + ($2019)|0);
      $2025 = HEAP8[$2024>>0]|0;
      $sB$15$shrunk = $2025;$sG$15$shrunk = $2014;$sR$15$shrunk = $2003;
      break;
     }
     default: {
      $sB$15$shrunk = 0;$sG$15$shrunk = 0;$sR$15$shrunk = 0;
     }
     }
     $sB$15 = $sB$15$shrunk&255;
     $sG$15 = $sG$15$shrunk&255;
     $sR$15 = $sR$15$shrunk&255;
     $2026 = $sR$15 & 224;
     $2027 = $sG$15 >>> 5;
     $2028 = $2027 << 2;
     $2029 = $2028 | $2026;
     $2030 = $sB$15 >>> 6;
     $2031 = $2029 | $2030;
     $2032 = (($12) + ($2031)|0);
     $2033 = HEAP8[$2032>>0]|0;
     HEAP8[$dst$17>>0] = $2033;
     $2034 = ((($dst$17)) + 1|0);
     $2035 = (($src$17) + ($17)|0);
     $2036 = (($n1$7) + -1)|0;
     $2037 = ($n1$7|0)>(1);
     if ($2037) {
      $dst$10 = $2034;$n1$0 = $2036;$src$10 = $2035;
      label = 58;
      continue;
     } else {
      break;
     }
    }
   }
   $2038 = (($2035) + ($6)|0);
   $2039 = (($2034) + ($10)|0);
   $2040 = (($2041) + -1)|0;
   $2042 = ($2041|0)==(0);
   if ($2042) {
    label = 108;
    break;
   } else {
    $2041 = $2040;$dst$952 = $2039;$src$953 = $2038;
   }
  }
  if ((label|0) == 106) {
   // unreachable;
  }
  else if ((label|0) == 108) {
   return;
  }
 }
}
function _Blit_RGB888_index8($info) {
 $info = $info|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$03 = 0, $dst$1 = 0, $dst$10 = 0, $dst$11 = 0, $dst$12 = 0, $dst$13 = 0, $dst$14 = 0, $dst$15 = 0, $dst$16 = 0, $dst$17 = 0, $dst$2 = 0;
 var $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $dst$95 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $n1$0 = 0, $n1$1 = 0, $n1$2 = 0, $n1$3 = 0, $n1$4 = 0;
 var $n1$5 = 0, $n1$6 = 0, $n1$7 = 0, $src$04 = 0, $src$16 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, $src$9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$info>>2]|0;
 $5 = ((($info)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6|0) / 4)&-1;
 $8 = ((($info)) + 20|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($info)) + 36|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($info)) + 48|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13|0)==(0|0);
 $15 = (($3) + -1)|0;
 $16 = ($3|0)==(0);
 if ($14) {
  if ($16) {
   return;
  }
  $20 = (($1) + 7)|0;
  $21 = (($20|0) / 8)&-1;
  $22 = $1 & 7;
  $116 = $15;$dst$03 = $9;$src$04 = $4;
  L6: while(1) {
   switch ($22|0) {
   case 0:  {
    $dst$1 = $dst$03;$n$0 = $21;
    label = 7;
    break;
   }
   case 7:  {
    $dst$2 = $dst$03;$n$1 = $21;
    label = 8;
    break;
   }
   case 6:  {
    $dst$3 = $dst$03;$n$2 = $21;
    label = 9;
    break;
   }
   case 5:  {
    $dst$4 = $dst$03;$n$3 = $21;
    label = 10;
    break;
   }
   case 4:  {
    $dst$5 = $dst$03;$n$4 = $21;
    label = 11;
    break;
   }
   case 3:  {
    $dst$6 = $dst$03;$n$5 = $21;
    label = 12;
    break;
   }
   case 2:  {
    $dst$7 = $dst$03;$n$6 = $21;
    label = 13;
    break;
   }
   case 1:  {
    $dst$8 = $dst$03;$n$7 = $21;
    label = 14;
    break;
   }
   default: {
    label = 15;
    break L6;
   }
   }
   while(1) {
    if ((label|0) == 7) {
     label = 0;
     $23 = HEAP32[$src$04>>2]|0;
     $24 = $23 >>> 16;
     $25 = $24 & 224;
     $26 = $23 >>> 11;
     $27 = $26 & 28;
     $28 = $25 | $27;
     $29 = $23 >>> 6;
     $30 = $29 & 3;
     $31 = $28 | $30;
     $32 = $31&255;
     $33 = ((($dst$1)) + 1|0);
     HEAP8[$dst$1>>0] = $32;
     $dst$2 = $33;$n$1 = $n$0;
     label = 8;
     continue;
    }
    else if ((label|0) == 8) {
     label = 0;
     $34 = HEAP32[$src$04>>2]|0;
     $35 = $34 >>> 16;
     $36 = $35 & 224;
     $37 = $34 >>> 11;
     $38 = $37 & 28;
     $39 = $36 | $38;
     $40 = $34 >>> 6;
     $41 = $40 & 3;
     $42 = $39 | $41;
     $43 = $42&255;
     $44 = ((($dst$2)) + 1|0);
     HEAP8[$dst$2>>0] = $43;
     $dst$3 = $44;$n$2 = $n$1;
     label = 9;
     continue;
    }
    else if ((label|0) == 9) {
     label = 0;
     $45 = HEAP32[$src$04>>2]|0;
     $46 = $45 >>> 16;
     $47 = $46 & 224;
     $48 = $45 >>> 11;
     $49 = $48 & 28;
     $50 = $47 | $49;
     $51 = $45 >>> 6;
     $52 = $51 & 3;
     $53 = $50 | $52;
     $54 = $53&255;
     $55 = ((($dst$3)) + 1|0);
     HEAP8[$dst$3>>0] = $54;
     $dst$4 = $55;$n$3 = $n$2;
     label = 10;
     continue;
    }
    else if ((label|0) == 10) {
     label = 0;
     $56 = HEAP32[$src$04>>2]|0;
     $57 = $56 >>> 16;
     $58 = $57 & 224;
     $59 = $56 >>> 11;
     $60 = $59 & 28;
     $61 = $58 | $60;
     $62 = $56 >>> 6;
     $63 = $62 & 3;
     $64 = $61 | $63;
     $65 = $64&255;
     $66 = ((($dst$4)) + 1|0);
     HEAP8[$dst$4>>0] = $65;
     $dst$5 = $66;$n$4 = $n$3;
     label = 11;
     continue;
    }
    else if ((label|0) == 11) {
     label = 0;
     $67 = HEAP32[$src$04>>2]|0;
     $68 = $67 >>> 16;
     $69 = $68 & 224;
     $70 = $67 >>> 11;
     $71 = $70 & 28;
     $72 = $69 | $71;
     $73 = $67 >>> 6;
     $74 = $73 & 3;
     $75 = $72 | $74;
     $76 = $75&255;
     $77 = ((($dst$5)) + 1|0);
     HEAP8[$dst$5>>0] = $76;
     $dst$6 = $77;$n$5 = $n$4;
     label = 12;
     continue;
    }
    else if ((label|0) == 12) {
     label = 0;
     $78 = HEAP32[$src$04>>2]|0;
     $79 = $78 >>> 16;
     $80 = $79 & 224;
     $81 = $78 >>> 11;
     $82 = $81 & 28;
     $83 = $80 | $82;
     $84 = $78 >>> 6;
     $85 = $84 & 3;
     $86 = $83 | $85;
     $87 = $86&255;
     $88 = ((($dst$6)) + 1|0);
     HEAP8[$dst$6>>0] = $87;
     $dst$7 = $88;$n$6 = $n$5;
     label = 13;
     continue;
    }
    else if ((label|0) == 13) {
     label = 0;
     $89 = HEAP32[$src$04>>2]|0;
     $90 = $89 >>> 16;
     $91 = $90 & 224;
     $92 = $89 >>> 11;
     $93 = $92 & 28;
     $94 = $91 | $93;
     $95 = $89 >>> 6;
     $96 = $95 & 3;
     $97 = $94 | $96;
     $98 = $97&255;
     $99 = ((($dst$7)) + 1|0);
     HEAP8[$dst$7>>0] = $98;
     $dst$8 = $99;$n$7 = $n$6;
     label = 14;
     continue;
    }
    else if ((label|0) == 14) {
     label = 0;
     $100 = HEAP32[$src$04>>2]|0;
     $101 = $100 >>> 16;
     $102 = $101 & 224;
     $103 = $100 >>> 11;
     $104 = $103 & 28;
     $105 = $102 | $104;
     $106 = $100 >>> 6;
     $107 = $106 & 3;
     $108 = $105 | $107;
     $109 = $108&255;
     $110 = ((($dst$8)) + 1|0);
     HEAP8[$dst$8>>0] = $109;
     $111 = (($n$7) + -1)|0;
     $112 = ($n$7|0)>(1);
     if ($112) {
      $dst$1 = $110;$n$0 = $111;
      label = 7;
      continue;
     } else {
      break;
     }
    }
   }
   $113 = (($src$04) + ($7<<2)|0);
   $114 = (($110) + ($11)|0);
   $115 = (($116) + -1)|0;
   $117 = ($116|0)==(0);
   if ($117) {
    label = 28;
    break;
   } else {
    $116 = $115;$dst$03 = $114;$src$04 = $113;
   }
  }
  if ((label|0) == 15) {
   // unreachable;
  }
  else if ((label|0) == 28) {
   return;
  }
 } else {
  if ($16) {
   return;
  }
  $17 = (($1) + 7)|0;
  $18 = (($17|0) / 8)&-1;
  $19 = $1 & 7;
  $227 = $15;$dst$95 = $9;$src$16 = $4;
  L26: while(1) {
   switch ($19|0) {
   case 0:  {
    $dst$10 = $dst$95;$n1$0 = $18;$src$2 = $src$16;
    label = 18;
    break;
   }
   case 7:  {
    $dst$11 = $dst$95;$n1$1 = $18;$src$3 = $src$16;
    label = 19;
    break;
   }
   case 6:  {
    $dst$12 = $dst$95;$n1$2 = $18;$src$4 = $src$16;
    label = 20;
    break;
   }
   case 5:  {
    $dst$13 = $dst$95;$n1$3 = $18;$src$5 = $src$16;
    label = 21;
    break;
   }
   case 4:  {
    $dst$14 = $dst$95;$n1$4 = $18;$src$6 = $src$16;
    label = 22;
    break;
   }
   case 3:  {
    $dst$15 = $dst$95;$n1$5 = $18;$src$7 = $src$16;
    label = 23;
    break;
   }
   case 2:  {
    $dst$16 = $dst$95;$n1$6 = $18;$src$8 = $src$16;
    label = 24;
    break;
   }
   case 1:  {
    $dst$17 = $dst$95;$n1$7 = $18;$src$9 = $src$16;
    label = 25;
    break;
   }
   default: {
    label = 26;
    break L26;
   }
   }
   while(1) {
    if ((label|0) == 18) {
     label = 0;
     $118 = HEAP32[$src$2>>2]|0;
     $119 = $118 >>> 16;
     $120 = $119 & 224;
     $121 = $118 >>> 11;
     $122 = $121 & 28;
     $123 = $120 | $122;
     $124 = $118 >>> 6;
     $125 = $124 & 3;
     $126 = $123 | $125;
     $127 = (($13) + ($126)|0);
     $128 = HEAP8[$127>>0]|0;
     $129 = ((($dst$10)) + 1|0);
     HEAP8[$dst$10>>0] = $128;
     $130 = ((($src$2)) + 4|0);
     $dst$11 = $129;$n1$1 = $n1$0;$src$3 = $130;
     label = 19;
     continue;
    }
    else if ((label|0) == 19) {
     label = 0;
     $131 = HEAP32[$src$3>>2]|0;
     $132 = $131 >>> 16;
     $133 = $132 & 224;
     $134 = $131 >>> 11;
     $135 = $134 & 28;
     $136 = $133 | $135;
     $137 = $131 >>> 6;
     $138 = $137 & 3;
     $139 = $136 | $138;
     $140 = (($13) + ($139)|0);
     $141 = HEAP8[$140>>0]|0;
     $142 = ((($dst$11)) + 1|0);
     HEAP8[$dst$11>>0] = $141;
     $143 = ((($src$3)) + 4|0);
     $dst$12 = $142;$n1$2 = $n1$1;$src$4 = $143;
     label = 20;
     continue;
    }
    else if ((label|0) == 20) {
     label = 0;
     $144 = HEAP32[$src$4>>2]|0;
     $145 = $144 >>> 16;
     $146 = $145 & 224;
     $147 = $144 >>> 11;
     $148 = $147 & 28;
     $149 = $146 | $148;
     $150 = $144 >>> 6;
     $151 = $150 & 3;
     $152 = $149 | $151;
     $153 = (($13) + ($152)|0);
     $154 = HEAP8[$153>>0]|0;
     $155 = ((($dst$12)) + 1|0);
     HEAP8[$dst$12>>0] = $154;
     $156 = ((($src$4)) + 4|0);
     $dst$13 = $155;$n1$3 = $n1$2;$src$5 = $156;
     label = 21;
     continue;
    }
    else if ((label|0) == 21) {
     label = 0;
     $157 = HEAP32[$src$5>>2]|0;
     $158 = $157 >>> 16;
     $159 = $158 & 224;
     $160 = $157 >>> 11;
     $161 = $160 & 28;
     $162 = $159 | $161;
     $163 = $157 >>> 6;
     $164 = $163 & 3;
     $165 = $162 | $164;
     $166 = (($13) + ($165)|0);
     $167 = HEAP8[$166>>0]|0;
     $168 = ((($dst$13)) + 1|0);
     HEAP8[$dst$13>>0] = $167;
     $169 = ((($src$5)) + 4|0);
     $dst$14 = $168;$n1$4 = $n1$3;$src$6 = $169;
     label = 22;
     continue;
    }
    else if ((label|0) == 22) {
     label = 0;
     $170 = HEAP32[$src$6>>2]|0;
     $171 = $170 >>> 16;
     $172 = $171 & 224;
     $173 = $170 >>> 11;
     $174 = $173 & 28;
     $175 = $172 | $174;
     $176 = $170 >>> 6;
     $177 = $176 & 3;
     $178 = $175 | $177;
     $179 = (($13) + ($178)|0);
     $180 = HEAP8[$179>>0]|0;
     $181 = ((($dst$14)) + 1|0);
     HEAP8[$dst$14>>0] = $180;
     $182 = ((($src$6)) + 4|0);
     $dst$15 = $181;$n1$5 = $n1$4;$src$7 = $182;
     label = 23;
     continue;
    }
    else if ((label|0) == 23) {
     label = 0;
     $183 = HEAP32[$src$7>>2]|0;
     $184 = $183 >>> 16;
     $185 = $184 & 224;
     $186 = $183 >>> 11;
     $187 = $186 & 28;
     $188 = $185 | $187;
     $189 = $183 >>> 6;
     $190 = $189 & 3;
     $191 = $188 | $190;
     $192 = (($13) + ($191)|0);
     $193 = HEAP8[$192>>0]|0;
     $194 = ((($dst$15)) + 1|0);
     HEAP8[$dst$15>>0] = $193;
     $195 = ((($src$7)) + 4|0);
     $dst$16 = $194;$n1$6 = $n1$5;$src$8 = $195;
     label = 24;
     continue;
    }
    else if ((label|0) == 24) {
     label = 0;
     $196 = HEAP32[$src$8>>2]|0;
     $197 = $196 >>> 16;
     $198 = $197 & 224;
     $199 = $196 >>> 11;
     $200 = $199 & 28;
     $201 = $198 | $200;
     $202 = $196 >>> 6;
     $203 = $202 & 3;
     $204 = $201 | $203;
     $205 = (($13) + ($204)|0);
     $206 = HEAP8[$205>>0]|0;
     $207 = ((($dst$16)) + 1|0);
     HEAP8[$dst$16>>0] = $206;
     $208 = ((($src$8)) + 4|0);
     $dst$17 = $207;$n1$7 = $n1$6;$src$9 = $208;
     label = 25;
     continue;
    }
    else if ((label|0) == 25) {
     label = 0;
     $209 = HEAP32[$src$9>>2]|0;
     $210 = $209 >>> 16;
     $211 = $210 & 224;
     $212 = $209 >>> 11;
     $213 = $212 & 28;
     $214 = $211 | $213;
     $215 = $209 >>> 6;
     $216 = $215 & 3;
     $217 = $214 | $216;
     $218 = (($13) + ($217)|0);
     $219 = HEAP8[$218>>0]|0;
     $220 = ((($dst$17)) + 1|0);
     HEAP8[$dst$17>>0] = $219;
     $221 = ((($src$9)) + 4|0);
     $222 = (($n1$7) + -1)|0;
     $223 = ($n1$7|0)>(1);
     if ($223) {
      $dst$10 = $220;$n1$0 = $222;$src$2 = $221;
      label = 18;
      continue;
     } else {
      break;
     }
    }
   }
   $224 = (($221) + ($7<<2)|0);
   $225 = (($220) + ($11)|0);
   $226 = (($227) + -1)|0;
   $228 = ($227|0)==(0);
   if ($228) {
    label = 28;
    break;
   } else {
    $227 = $226;$dst$95 = $225;$src$16 = $224;
   }
  }
  if ((label|0) == 26) {
   // unreachable;
  }
  else if ((label|0) == 28) {
   return;
  }
 }
}
function _Blit_RGB101010_index8($info) {
 $info = $info|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$03 = 0, $dst$1 = 0, $dst$10 = 0, $dst$11 = 0, $dst$12 = 0, $dst$13 = 0, $dst$14 = 0, $dst$15 = 0, $dst$16 = 0, $dst$17 = 0, $dst$2 = 0;
 var $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $dst$95 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $n1$0 = 0, $n1$1 = 0, $n1$2 = 0, $n1$3 = 0, $n1$4 = 0;
 var $n1$5 = 0, $n1$6 = 0, $n1$7 = 0, $src$04 = 0, $src$16 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, $src$9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$info>>2]|0;
 $5 = ((($info)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6|0) / 4)&-1;
 $8 = ((($info)) + 20|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($info)) + 36|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($info)) + 48|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13|0)==(0|0);
 $15 = (($3) + -1)|0;
 $16 = ($3|0)==(0);
 if ($14) {
  if ($16) {
   return;
  }
  $20 = (($1) + 7)|0;
  $21 = (($20|0) / 8)&-1;
  $22 = $1 & 7;
  $116 = $15;$dst$03 = $9;$src$04 = $4;
  L6: while(1) {
   switch ($22|0) {
   case 0:  {
    $dst$1 = $dst$03;$n$0 = $21;
    label = 7;
    break;
   }
   case 7:  {
    $dst$2 = $dst$03;$n$1 = $21;
    label = 8;
    break;
   }
   case 6:  {
    $dst$3 = $dst$03;$n$2 = $21;
    label = 9;
    break;
   }
   case 5:  {
    $dst$4 = $dst$03;$n$3 = $21;
    label = 10;
    break;
   }
   case 4:  {
    $dst$5 = $dst$03;$n$4 = $21;
    label = 11;
    break;
   }
   case 3:  {
    $dst$6 = $dst$03;$n$5 = $21;
    label = 12;
    break;
   }
   case 2:  {
    $dst$7 = $dst$03;$n$6 = $21;
    label = 13;
    break;
   }
   case 1:  {
    $dst$8 = $dst$03;$n$7 = $21;
    label = 14;
    break;
   }
   default: {
    label = 15;
    break L6;
   }
   }
   while(1) {
    if ((label|0) == 7) {
     label = 0;
     $23 = HEAP32[$src$04>>2]|0;
     $24 = $23 >>> 22;
     $25 = $24 & 224;
     $26 = $23 >>> 15;
     $27 = $26 & 28;
     $28 = $25 | $27;
     $29 = $23 >>> 8;
     $30 = $29 & 3;
     $31 = $28 | $30;
     $32 = $31&255;
     $33 = ((($dst$1)) + 1|0);
     HEAP8[$dst$1>>0] = $32;
     $dst$2 = $33;$n$1 = $n$0;
     label = 8;
     continue;
    }
    else if ((label|0) == 8) {
     label = 0;
     $34 = HEAP32[$src$04>>2]|0;
     $35 = $34 >>> 22;
     $36 = $35 & 224;
     $37 = $34 >>> 15;
     $38 = $37 & 28;
     $39 = $36 | $38;
     $40 = $34 >>> 8;
     $41 = $40 & 3;
     $42 = $39 | $41;
     $43 = $42&255;
     $44 = ((($dst$2)) + 1|0);
     HEAP8[$dst$2>>0] = $43;
     $dst$3 = $44;$n$2 = $n$1;
     label = 9;
     continue;
    }
    else if ((label|0) == 9) {
     label = 0;
     $45 = HEAP32[$src$04>>2]|0;
     $46 = $45 >>> 22;
     $47 = $46 & 224;
     $48 = $45 >>> 15;
     $49 = $48 & 28;
     $50 = $47 | $49;
     $51 = $45 >>> 8;
     $52 = $51 & 3;
     $53 = $50 | $52;
     $54 = $53&255;
     $55 = ((($dst$3)) + 1|0);
     HEAP8[$dst$3>>0] = $54;
     $dst$4 = $55;$n$3 = $n$2;
     label = 10;
     continue;
    }
    else if ((label|0) == 10) {
     label = 0;
     $56 = HEAP32[$src$04>>2]|0;
     $57 = $56 >>> 22;
     $58 = $57 & 224;
     $59 = $56 >>> 15;
     $60 = $59 & 28;
     $61 = $58 | $60;
     $62 = $56 >>> 8;
     $63 = $62 & 3;
     $64 = $61 | $63;
     $65 = $64&255;
     $66 = ((($dst$4)) + 1|0);
     HEAP8[$dst$4>>0] = $65;
     $dst$5 = $66;$n$4 = $n$3;
     label = 11;
     continue;
    }
    else if ((label|0) == 11) {
     label = 0;
     $67 = HEAP32[$src$04>>2]|0;
     $68 = $67 >>> 22;
     $69 = $68 & 224;
     $70 = $67 >>> 15;
     $71 = $70 & 28;
     $72 = $69 | $71;
     $73 = $67 >>> 8;
     $74 = $73 & 3;
     $75 = $72 | $74;
     $76 = $75&255;
     $77 = ((($dst$5)) + 1|0);
     HEAP8[$dst$5>>0] = $76;
     $dst$6 = $77;$n$5 = $n$4;
     label = 12;
     continue;
    }
    else if ((label|0) == 12) {
     label = 0;
     $78 = HEAP32[$src$04>>2]|0;
     $79 = $78 >>> 22;
     $80 = $79 & 224;
     $81 = $78 >>> 15;
     $82 = $81 & 28;
     $83 = $80 | $82;
     $84 = $78 >>> 8;
     $85 = $84 & 3;
     $86 = $83 | $85;
     $87 = $86&255;
     $88 = ((($dst$6)) + 1|0);
     HEAP8[$dst$6>>0] = $87;
     $dst$7 = $88;$n$6 = $n$5;
     label = 13;
     continue;
    }
    else if ((label|0) == 13) {
     label = 0;
     $89 = HEAP32[$src$04>>2]|0;
     $90 = $89 >>> 22;
     $91 = $90 & 224;
     $92 = $89 >>> 15;
     $93 = $92 & 28;
     $94 = $91 | $93;
     $95 = $89 >>> 8;
     $96 = $95 & 3;
     $97 = $94 | $96;
     $98 = $97&255;
     $99 = ((($dst$7)) + 1|0);
     HEAP8[$dst$7>>0] = $98;
     $dst$8 = $99;$n$7 = $n$6;
     label = 14;
     continue;
    }
    else if ((label|0) == 14) {
     label = 0;
     $100 = HEAP32[$src$04>>2]|0;
     $101 = $100 >>> 22;
     $102 = $101 & 224;
     $103 = $100 >>> 15;
     $104 = $103 & 28;
     $105 = $102 | $104;
     $106 = $100 >>> 8;
     $107 = $106 & 3;
     $108 = $105 | $107;
     $109 = $108&255;
     $110 = ((($dst$8)) + 1|0);
     HEAP8[$dst$8>>0] = $109;
     $111 = (($n$7) + -1)|0;
     $112 = ($n$7|0)>(1);
     if ($112) {
      $dst$1 = $110;$n$0 = $111;
      label = 7;
      continue;
     } else {
      break;
     }
    }
   }
   $113 = (($src$04) + ($7<<2)|0);
   $114 = (($110) + ($11)|0);
   $115 = (($116) + -1)|0;
   $117 = ($116|0)==(0);
   if ($117) {
    label = 28;
    break;
   } else {
    $116 = $115;$dst$03 = $114;$src$04 = $113;
   }
  }
  if ((label|0) == 15) {
   // unreachable;
  }
  else if ((label|0) == 28) {
   return;
  }
 } else {
  if ($16) {
   return;
  }
  $17 = (($1) + 7)|0;
  $18 = (($17|0) / 8)&-1;
  $19 = $1 & 7;
  $227 = $15;$dst$95 = $9;$src$16 = $4;
  L26: while(1) {
   switch ($19|0) {
   case 0:  {
    $dst$10 = $dst$95;$n1$0 = $18;$src$2 = $src$16;
    label = 18;
    break;
   }
   case 7:  {
    $dst$11 = $dst$95;$n1$1 = $18;$src$3 = $src$16;
    label = 19;
    break;
   }
   case 6:  {
    $dst$12 = $dst$95;$n1$2 = $18;$src$4 = $src$16;
    label = 20;
    break;
   }
   case 5:  {
    $dst$13 = $dst$95;$n1$3 = $18;$src$5 = $src$16;
    label = 21;
    break;
   }
   case 4:  {
    $dst$14 = $dst$95;$n1$4 = $18;$src$6 = $src$16;
    label = 22;
    break;
   }
   case 3:  {
    $dst$15 = $dst$95;$n1$5 = $18;$src$7 = $src$16;
    label = 23;
    break;
   }
   case 2:  {
    $dst$16 = $dst$95;$n1$6 = $18;$src$8 = $src$16;
    label = 24;
    break;
   }
   case 1:  {
    $dst$17 = $dst$95;$n1$7 = $18;$src$9 = $src$16;
    label = 25;
    break;
   }
   default: {
    label = 26;
    break L26;
   }
   }
   while(1) {
    if ((label|0) == 18) {
     label = 0;
     $118 = HEAP32[$src$2>>2]|0;
     $119 = $118 >>> 22;
     $120 = $119 & 224;
     $121 = $118 >>> 15;
     $122 = $121 & 28;
     $123 = $120 | $122;
     $124 = $118 >>> 8;
     $125 = $124 & 3;
     $126 = $123 | $125;
     $127 = (($13) + ($126)|0);
     $128 = HEAP8[$127>>0]|0;
     $129 = ((($dst$10)) + 1|0);
     HEAP8[$dst$10>>0] = $128;
     $130 = ((($src$2)) + 4|0);
     $dst$11 = $129;$n1$1 = $n1$0;$src$3 = $130;
     label = 19;
     continue;
    }
    else if ((label|0) == 19) {
     label = 0;
     $131 = HEAP32[$src$3>>2]|0;
     $132 = $131 >>> 22;
     $133 = $132 & 224;
     $134 = $131 >>> 15;
     $135 = $134 & 28;
     $136 = $133 | $135;
     $137 = $131 >>> 8;
     $138 = $137 & 3;
     $139 = $136 | $138;
     $140 = (($13) + ($139)|0);
     $141 = HEAP8[$140>>0]|0;
     $142 = ((($dst$11)) + 1|0);
     HEAP8[$dst$11>>0] = $141;
     $143 = ((($src$3)) + 4|0);
     $dst$12 = $142;$n1$2 = $n1$1;$src$4 = $143;
     label = 20;
     continue;
    }
    else if ((label|0) == 20) {
     label = 0;
     $144 = HEAP32[$src$4>>2]|0;
     $145 = $144 >>> 22;
     $146 = $145 & 224;
     $147 = $144 >>> 15;
     $148 = $147 & 28;
     $149 = $146 | $148;
     $150 = $144 >>> 8;
     $151 = $150 & 3;
     $152 = $149 | $151;
     $153 = (($13) + ($152)|0);
     $154 = HEAP8[$153>>0]|0;
     $155 = ((($dst$12)) + 1|0);
     HEAP8[$dst$12>>0] = $154;
     $156 = ((($src$4)) + 4|0);
     $dst$13 = $155;$n1$3 = $n1$2;$src$5 = $156;
     label = 21;
     continue;
    }
    else if ((label|0) == 21) {
     label = 0;
     $157 = HEAP32[$src$5>>2]|0;
     $158 = $157 >>> 22;
     $159 = $158 & 224;
     $160 = $157 >>> 15;
     $161 = $160 & 28;
     $162 = $159 | $161;
     $163 = $157 >>> 8;
     $164 = $163 & 3;
     $165 = $162 | $164;
     $166 = (($13) + ($165)|0);
     $167 = HEAP8[$166>>0]|0;
     $168 = ((($dst$13)) + 1|0);
     HEAP8[$dst$13>>0] = $167;
     $169 = ((($src$5)) + 4|0);
     $dst$14 = $168;$n1$4 = $n1$3;$src$6 = $169;
     label = 22;
     continue;
    }
    else if ((label|0) == 22) {
     label = 0;
     $170 = HEAP32[$src$6>>2]|0;
     $171 = $170 >>> 22;
     $172 = $171 & 224;
     $173 = $170 >>> 15;
     $174 = $173 & 28;
     $175 = $172 | $174;
     $176 = $170 >>> 8;
     $177 = $176 & 3;
     $178 = $175 | $177;
     $179 = (($13) + ($178)|0);
     $180 = HEAP8[$179>>0]|0;
     $181 = ((($dst$14)) + 1|0);
     HEAP8[$dst$14>>0] = $180;
     $182 = ((($src$6)) + 4|0);
     $dst$15 = $181;$n1$5 = $n1$4;$src$7 = $182;
     label = 23;
     continue;
    }
    else if ((label|0) == 23) {
     label = 0;
     $183 = HEAP32[$src$7>>2]|0;
     $184 = $183 >>> 22;
     $185 = $184 & 224;
     $186 = $183 >>> 15;
     $187 = $186 & 28;
     $188 = $185 | $187;
     $189 = $183 >>> 8;
     $190 = $189 & 3;
     $191 = $188 | $190;
     $192 = (($13) + ($191)|0);
     $193 = HEAP8[$192>>0]|0;
     $194 = ((($dst$15)) + 1|0);
     HEAP8[$dst$15>>0] = $193;
     $195 = ((($src$7)) + 4|0);
     $dst$16 = $194;$n1$6 = $n1$5;$src$8 = $195;
     label = 24;
     continue;
    }
    else if ((label|0) == 24) {
     label = 0;
     $196 = HEAP32[$src$8>>2]|0;
     $197 = $196 >>> 22;
     $198 = $197 & 224;
     $199 = $196 >>> 15;
     $200 = $199 & 28;
     $201 = $198 | $200;
     $202 = $196 >>> 8;
     $203 = $202 & 3;
     $204 = $201 | $203;
     $205 = (($13) + ($204)|0);
     $206 = HEAP8[$205>>0]|0;
     $207 = ((($dst$16)) + 1|0);
     HEAP8[$dst$16>>0] = $206;
     $208 = ((($src$8)) + 4|0);
     $dst$17 = $207;$n1$7 = $n1$6;$src$9 = $208;
     label = 25;
     continue;
    }
    else if ((label|0) == 25) {
     label = 0;
     $209 = HEAP32[$src$9>>2]|0;
     $210 = $209 >>> 22;
     $211 = $210 & 224;
     $212 = $209 >>> 15;
     $213 = $212 & 28;
     $214 = $211 | $213;
     $215 = $209 >>> 8;
     $216 = $215 & 3;
     $217 = $214 | $216;
     $218 = (($13) + ($217)|0);
     $219 = HEAP8[$218>>0]|0;
     $220 = ((($dst$17)) + 1|0);
     HEAP8[$dst$17>>0] = $219;
     $221 = ((($src$9)) + 4|0);
     $222 = (($n1$7) + -1)|0;
     $223 = ($n1$7|0)>(1);
     if ($223) {
      $dst$10 = $220;$n1$0 = $222;$src$2 = $221;
      label = 18;
      continue;
     } else {
      break;
     }
    }
   }
   $224 = (($221) + ($7<<2)|0);
   $225 = (($220) + ($11)|0);
   $226 = (($227) + -1)|0;
   $228 = ($227|0)==(0);
   if ($228) {
    label = 28;
    break;
   } else {
    $227 = $226;$dst$95 = $225;$src$16 = $224;
   }
  }
  if ((label|0) == 26) {
   // unreachable;
  }
  else if ((label|0) == 28) {
   return;
  }
 }
}
function _Blit2101010toN($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0;
 var $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0;
 var $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0;
 var $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0;
 var $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0;
 var $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0;
 var $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0;
 var $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0;
 var $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0;
 var $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0;
 var $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0;
 var $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0;
 var $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0;
 var $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0;
 var $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0;
 var $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0;
 var $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0;
 var $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0;
 var $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0;
 var $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0;
 var $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0;
 var $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0;
 var $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0;
 var $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0;
 var $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0;
 var $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0;
 var $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0;
 var $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0;
 var $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0;
 var $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0;
 var $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0;
 var $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0;
 var $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0;
 var $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0;
 var $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $div = 0, $div1 = 0, $div10 = 0, $div11 = 0, $div12 = 0;
 var $div13 = 0, $div14 = 0, $div15 = 0, $div16 = 0, $div17 = 0, $div18 = 0, $div19 = 0, $div2 = 0, $div20 = 0, $div21 = 0, $div22 = 0, $div23 = 0, $div3 = 0, $div4 = 0, $div5 = 0, $div6 = 0, $div7 = 0, $div8 = 0, $div9 = 0, $dst$024 = 0;
 var $dst$1 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $src$025 = 0, $src$1 = 0, $src$2 = 0, $src$3 = 0;
 var $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 44|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 9|0);
 $11 = HEAP8[$10>>0]|0;
 $12 = $11&255;
 $13 = ($3|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = HEAP32[$info>>2]|0;
 $17 = (($1) + 7)|0;
 $18 = (($17|0) / 8)&-1;
 $19 = $1 & 7;
 $20 = ((($9)) + 28|0);
 $21 = ((($9)) + 32|0);
 $22 = ((($9)) + 33|0);
 $23 = ((($9)) + 34|0);
 $$in = $3;$dst$024 = $15;$src$025 = $16;
 L4: while(1) {
  $24 = (($$in) + -1)|0;
  switch ($19|0) {
  case 0:  {
   $dst$1 = $dst$024;$n$0 = $18;$src$1 = $src$025;
   label = 4;
   break;
  }
  case 7:  {
   $dst$2 = $dst$024;$n$1 = $18;$src$2 = $src$025;
   label = 10;
   break;
  }
  case 6:  {
   $dst$3 = $dst$024;$n$2 = $18;$src$3 = $src$025;
   label = 16;
   break;
  }
  case 5:  {
   $dst$4 = $dst$024;$n$3 = $18;$src$4 = $src$025;
   label = 22;
   break;
  }
  case 4:  {
   $dst$5 = $dst$024;$n$4 = $18;$src$5 = $src$025;
   label = 28;
   break;
  }
  case 3:  {
   $dst$6 = $dst$024;$n$5 = $18;$src$6 = $src$025;
   label = 34;
   break;
  }
  case 2:  {
   $dst$7 = $dst$024;$n$6 = $18;$src$7 = $src$025;
   label = 40;
   break;
  }
  case 1:  {
   $dst$8 = $dst$024;$n$7 = $18;$src$8 = $src$025;
   label = 46;
   break;
  }
  default: {
   label = 52;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    $25 = HEAP32[$src$1>>2]|0;
    $26 = $25 >>> 22;
    $27 = $26 & 255;
    $28 = $25 >>> 12;
    $29 = $28 & 255;
    $30 = $25 >>> 2;
    $31 = $30 & 255;
    $32 = $25 >>> 30;
    $33 = HEAP32[(125948)>>2]|0;
    $34 = (($33) + ($32)|0);
    $35 = HEAP8[$34>>0]|0;
    $36 = $35&255;
    switch ($12|0) {
    case 1:  {
     $37 = HEAP32[$20>>2]|0;
     $38 = $37 & 255;
     $39 = $27 >>> $38;
     $40 = HEAP32[$21>>2]|0;
     $41 = $40 & 255;
     $42 = $39 << $41;
     $43 = $37 >>> 8;
     $44 = $43 & 255;
     $45 = $29 >>> $44;
     $46 = $40 >>> 8;
     $47 = $46 & 255;
     $48 = $45 << $47;
     $49 = $48 | $42;
     $50 = $37 >>> 16;
     $51 = $50 & 255;
     $52 = $31 >>> $51;
     $53 = $40 >>> 16;
     $54 = $53 & 255;
     $55 = $52 << $54;
     $56 = $49 | $55;
     $57 = $37 >>> 24;
     $58 = $36 >>> $57;
     $59 = $40 >>> 24;
     $60 = $58 << $59;
     $61 = $56 | $60;
     $62 = $61&255;
     HEAP8[$dst$1>>0] = $62;
     break;
    }
    case 2:  {
     $63 = HEAP32[$20>>2]|0;
     $64 = $63 & 255;
     $65 = $27 >>> $64;
     $66 = HEAP32[$21>>2]|0;
     $67 = $66 & 255;
     $68 = $65 << $67;
     $69 = $63 >>> 8;
     $70 = $69 & 255;
     $71 = $29 >>> $70;
     $72 = $66 >>> 8;
     $73 = $72 & 255;
     $74 = $71 << $73;
     $75 = $74 | $68;
     $76 = $63 >>> 16;
     $77 = $76 & 255;
     $78 = $31 >>> $77;
     $79 = $66 >>> 16;
     $80 = $79 & 255;
     $81 = $78 << $80;
     $82 = $75 | $81;
     $83 = $63 >>> 24;
     $84 = $36 >>> $83;
     $85 = $66 >>> 24;
     $86 = $84 << $85;
     $87 = $82 | $86;
     $88 = $87&65535;
     HEAP16[$dst$1>>1] = $88;
     break;
    }
    case 3:  {
     $89 = $26&255;
     $90 = HEAP8[$21>>0]|0;
     $div18 = ($90&255) >>> 3;
     $91 = $div18&255;
     $92 = (($dst$1) + ($91)|0);
     HEAP8[$92>>0] = $89;
     $93 = $28&255;
     $94 = HEAP8[$22>>0]|0;
     $div19 = ($94&255) >>> 3;
     $95 = $div19&255;
     $96 = (($dst$1) + ($95)|0);
     HEAP8[$96>>0] = $93;
     $97 = $30&255;
     $98 = HEAP8[$23>>0]|0;
     $div20 = ($98&255) >>> 3;
     $99 = $div20&255;
     $100 = (($dst$1) + ($99)|0);
     HEAP8[$100>>0] = $97;
     break;
    }
    case 4:  {
     $101 = HEAP32[$20>>2]|0;
     $102 = $101 & 255;
     $103 = $27 >>> $102;
     $104 = HEAP32[$21>>2]|0;
     $105 = $104 & 255;
     $106 = $103 << $105;
     $107 = $101 >>> 8;
     $108 = $107 & 255;
     $109 = $29 >>> $108;
     $110 = $104 >>> 8;
     $111 = $110 & 255;
     $112 = $109 << $111;
     $113 = $112 | $106;
     $114 = $101 >>> 16;
     $115 = $114 & 255;
     $116 = $31 >>> $115;
     $117 = $104 >>> 16;
     $118 = $117 & 255;
     $119 = $116 << $118;
     $120 = $113 | $119;
     $121 = $101 >>> 24;
     $122 = $36 >>> $121;
     $123 = $104 >>> 24;
     $124 = $122 << $123;
     $125 = $120 | $124;
     HEAP32[$dst$1>>2] = $125;
     break;
    }
    default: {
    }
    }
    $126 = (($dst$1) + ($12)|0);
    $127 = ((($src$1)) + 4|0);
    $dst$2 = $126;$n$1 = $n$0;$src$2 = $127;
    label = 10;
    continue;
   }
   else if ((label|0) == 10) {
    label = 0;
    $128 = HEAP32[$src$2>>2]|0;
    $129 = $128 >>> 22;
    $130 = $129 & 255;
    $131 = $128 >>> 12;
    $132 = $131 & 255;
    $133 = $128 >>> 2;
    $134 = $133 & 255;
    $135 = $128 >>> 30;
    $136 = HEAP32[(125948)>>2]|0;
    $137 = (($136) + ($135)|0);
    $138 = HEAP8[$137>>0]|0;
    $139 = $138&255;
    switch ($12|0) {
    case 1:  {
     $140 = HEAP32[$20>>2]|0;
     $141 = $140 & 255;
     $142 = $130 >>> $141;
     $143 = HEAP32[$21>>2]|0;
     $144 = $143 & 255;
     $145 = $142 << $144;
     $146 = $140 >>> 8;
     $147 = $146 & 255;
     $148 = $132 >>> $147;
     $149 = $143 >>> 8;
     $150 = $149 & 255;
     $151 = $148 << $150;
     $152 = $151 | $145;
     $153 = $140 >>> 16;
     $154 = $153 & 255;
     $155 = $134 >>> $154;
     $156 = $143 >>> 16;
     $157 = $156 & 255;
     $158 = $155 << $157;
     $159 = $152 | $158;
     $160 = $140 >>> 24;
     $161 = $139 >>> $160;
     $162 = $143 >>> 24;
     $163 = $161 << $162;
     $164 = $159 | $163;
     $165 = $164&255;
     HEAP8[$dst$2>>0] = $165;
     break;
    }
    case 2:  {
     $166 = HEAP32[$20>>2]|0;
     $167 = $166 & 255;
     $168 = $130 >>> $167;
     $169 = HEAP32[$21>>2]|0;
     $170 = $169 & 255;
     $171 = $168 << $170;
     $172 = $166 >>> 8;
     $173 = $172 & 255;
     $174 = $132 >>> $173;
     $175 = $169 >>> 8;
     $176 = $175 & 255;
     $177 = $174 << $176;
     $178 = $177 | $171;
     $179 = $166 >>> 16;
     $180 = $179 & 255;
     $181 = $134 >>> $180;
     $182 = $169 >>> 16;
     $183 = $182 & 255;
     $184 = $181 << $183;
     $185 = $178 | $184;
     $186 = $166 >>> 24;
     $187 = $139 >>> $186;
     $188 = $169 >>> 24;
     $189 = $187 << $188;
     $190 = $185 | $189;
     $191 = $190&65535;
     HEAP16[$dst$2>>1] = $191;
     break;
    }
    case 3:  {
     $192 = $129&255;
     $193 = HEAP8[$21>>0]|0;
     $div15 = ($193&255) >>> 3;
     $194 = $div15&255;
     $195 = (($dst$2) + ($194)|0);
     HEAP8[$195>>0] = $192;
     $196 = $131&255;
     $197 = HEAP8[$22>>0]|0;
     $div16 = ($197&255) >>> 3;
     $198 = $div16&255;
     $199 = (($dst$2) + ($198)|0);
     HEAP8[$199>>0] = $196;
     $200 = $133&255;
     $201 = HEAP8[$23>>0]|0;
     $div17 = ($201&255) >>> 3;
     $202 = $div17&255;
     $203 = (($dst$2) + ($202)|0);
     HEAP8[$203>>0] = $200;
     break;
    }
    case 4:  {
     $204 = HEAP32[$20>>2]|0;
     $205 = $204 & 255;
     $206 = $130 >>> $205;
     $207 = HEAP32[$21>>2]|0;
     $208 = $207 & 255;
     $209 = $206 << $208;
     $210 = $204 >>> 8;
     $211 = $210 & 255;
     $212 = $132 >>> $211;
     $213 = $207 >>> 8;
     $214 = $213 & 255;
     $215 = $212 << $214;
     $216 = $215 | $209;
     $217 = $204 >>> 16;
     $218 = $217 & 255;
     $219 = $134 >>> $218;
     $220 = $207 >>> 16;
     $221 = $220 & 255;
     $222 = $219 << $221;
     $223 = $216 | $222;
     $224 = $204 >>> 24;
     $225 = $139 >>> $224;
     $226 = $207 >>> 24;
     $227 = $225 << $226;
     $228 = $223 | $227;
     HEAP32[$dst$2>>2] = $228;
     break;
    }
    default: {
    }
    }
    $229 = (($dst$2) + ($12)|0);
    $230 = ((($src$2)) + 4|0);
    $dst$3 = $229;$n$2 = $n$1;$src$3 = $230;
    label = 16;
    continue;
   }
   else if ((label|0) == 16) {
    label = 0;
    $231 = HEAP32[$src$3>>2]|0;
    $232 = $231 >>> 22;
    $233 = $232 & 255;
    $234 = $231 >>> 12;
    $235 = $234 & 255;
    $236 = $231 >>> 2;
    $237 = $236 & 255;
    $238 = $231 >>> 30;
    $239 = HEAP32[(125948)>>2]|0;
    $240 = (($239) + ($238)|0);
    $241 = HEAP8[$240>>0]|0;
    $242 = $241&255;
    switch ($12|0) {
    case 1:  {
     $243 = HEAP32[$20>>2]|0;
     $244 = $243 & 255;
     $245 = $233 >>> $244;
     $246 = HEAP32[$21>>2]|0;
     $247 = $246 & 255;
     $248 = $245 << $247;
     $249 = $243 >>> 8;
     $250 = $249 & 255;
     $251 = $235 >>> $250;
     $252 = $246 >>> 8;
     $253 = $252 & 255;
     $254 = $251 << $253;
     $255 = $254 | $248;
     $256 = $243 >>> 16;
     $257 = $256 & 255;
     $258 = $237 >>> $257;
     $259 = $246 >>> 16;
     $260 = $259 & 255;
     $261 = $258 << $260;
     $262 = $255 | $261;
     $263 = $243 >>> 24;
     $264 = $242 >>> $263;
     $265 = $246 >>> 24;
     $266 = $264 << $265;
     $267 = $262 | $266;
     $268 = $267&255;
     HEAP8[$dst$3>>0] = $268;
     break;
    }
    case 2:  {
     $269 = HEAP32[$20>>2]|0;
     $270 = $269 & 255;
     $271 = $233 >>> $270;
     $272 = HEAP32[$21>>2]|0;
     $273 = $272 & 255;
     $274 = $271 << $273;
     $275 = $269 >>> 8;
     $276 = $275 & 255;
     $277 = $235 >>> $276;
     $278 = $272 >>> 8;
     $279 = $278 & 255;
     $280 = $277 << $279;
     $281 = $280 | $274;
     $282 = $269 >>> 16;
     $283 = $282 & 255;
     $284 = $237 >>> $283;
     $285 = $272 >>> 16;
     $286 = $285 & 255;
     $287 = $284 << $286;
     $288 = $281 | $287;
     $289 = $269 >>> 24;
     $290 = $242 >>> $289;
     $291 = $272 >>> 24;
     $292 = $290 << $291;
     $293 = $288 | $292;
     $294 = $293&65535;
     HEAP16[$dst$3>>1] = $294;
     break;
    }
    case 3:  {
     $295 = $232&255;
     $296 = HEAP8[$21>>0]|0;
     $div12 = ($296&255) >>> 3;
     $297 = $div12&255;
     $298 = (($dst$3) + ($297)|0);
     HEAP8[$298>>0] = $295;
     $299 = $234&255;
     $300 = HEAP8[$22>>0]|0;
     $div13 = ($300&255) >>> 3;
     $301 = $div13&255;
     $302 = (($dst$3) + ($301)|0);
     HEAP8[$302>>0] = $299;
     $303 = $236&255;
     $304 = HEAP8[$23>>0]|0;
     $div14 = ($304&255) >>> 3;
     $305 = $div14&255;
     $306 = (($dst$3) + ($305)|0);
     HEAP8[$306>>0] = $303;
     break;
    }
    case 4:  {
     $307 = HEAP32[$20>>2]|0;
     $308 = $307 & 255;
     $309 = $233 >>> $308;
     $310 = HEAP32[$21>>2]|0;
     $311 = $310 & 255;
     $312 = $309 << $311;
     $313 = $307 >>> 8;
     $314 = $313 & 255;
     $315 = $235 >>> $314;
     $316 = $310 >>> 8;
     $317 = $316 & 255;
     $318 = $315 << $317;
     $319 = $318 | $312;
     $320 = $307 >>> 16;
     $321 = $320 & 255;
     $322 = $237 >>> $321;
     $323 = $310 >>> 16;
     $324 = $323 & 255;
     $325 = $322 << $324;
     $326 = $319 | $325;
     $327 = $307 >>> 24;
     $328 = $242 >>> $327;
     $329 = $310 >>> 24;
     $330 = $328 << $329;
     $331 = $326 | $330;
     HEAP32[$dst$3>>2] = $331;
     break;
    }
    default: {
    }
    }
    $332 = (($dst$3) + ($12)|0);
    $333 = ((($src$3)) + 4|0);
    $dst$4 = $332;$n$3 = $n$2;$src$4 = $333;
    label = 22;
    continue;
   }
   else if ((label|0) == 22) {
    label = 0;
    $334 = HEAP32[$src$4>>2]|0;
    $335 = $334 >>> 22;
    $336 = $335 & 255;
    $337 = $334 >>> 12;
    $338 = $337 & 255;
    $339 = $334 >>> 2;
    $340 = $339 & 255;
    $341 = $334 >>> 30;
    $342 = HEAP32[(125948)>>2]|0;
    $343 = (($342) + ($341)|0);
    $344 = HEAP8[$343>>0]|0;
    $345 = $344&255;
    switch ($12|0) {
    case 1:  {
     $346 = HEAP32[$20>>2]|0;
     $347 = $346 & 255;
     $348 = $336 >>> $347;
     $349 = HEAP32[$21>>2]|0;
     $350 = $349 & 255;
     $351 = $348 << $350;
     $352 = $346 >>> 8;
     $353 = $352 & 255;
     $354 = $338 >>> $353;
     $355 = $349 >>> 8;
     $356 = $355 & 255;
     $357 = $354 << $356;
     $358 = $357 | $351;
     $359 = $346 >>> 16;
     $360 = $359 & 255;
     $361 = $340 >>> $360;
     $362 = $349 >>> 16;
     $363 = $362 & 255;
     $364 = $361 << $363;
     $365 = $358 | $364;
     $366 = $346 >>> 24;
     $367 = $345 >>> $366;
     $368 = $349 >>> 24;
     $369 = $367 << $368;
     $370 = $365 | $369;
     $371 = $370&255;
     HEAP8[$dst$4>>0] = $371;
     break;
    }
    case 2:  {
     $372 = HEAP32[$20>>2]|0;
     $373 = $372 & 255;
     $374 = $336 >>> $373;
     $375 = HEAP32[$21>>2]|0;
     $376 = $375 & 255;
     $377 = $374 << $376;
     $378 = $372 >>> 8;
     $379 = $378 & 255;
     $380 = $338 >>> $379;
     $381 = $375 >>> 8;
     $382 = $381 & 255;
     $383 = $380 << $382;
     $384 = $383 | $377;
     $385 = $372 >>> 16;
     $386 = $385 & 255;
     $387 = $340 >>> $386;
     $388 = $375 >>> 16;
     $389 = $388 & 255;
     $390 = $387 << $389;
     $391 = $384 | $390;
     $392 = $372 >>> 24;
     $393 = $345 >>> $392;
     $394 = $375 >>> 24;
     $395 = $393 << $394;
     $396 = $391 | $395;
     $397 = $396&65535;
     HEAP16[$dst$4>>1] = $397;
     break;
    }
    case 3:  {
     $398 = $335&255;
     $399 = HEAP8[$21>>0]|0;
     $div9 = ($399&255) >>> 3;
     $400 = $div9&255;
     $401 = (($dst$4) + ($400)|0);
     HEAP8[$401>>0] = $398;
     $402 = $337&255;
     $403 = HEAP8[$22>>0]|0;
     $div10 = ($403&255) >>> 3;
     $404 = $div10&255;
     $405 = (($dst$4) + ($404)|0);
     HEAP8[$405>>0] = $402;
     $406 = $339&255;
     $407 = HEAP8[$23>>0]|0;
     $div11 = ($407&255) >>> 3;
     $408 = $div11&255;
     $409 = (($dst$4) + ($408)|0);
     HEAP8[$409>>0] = $406;
     break;
    }
    case 4:  {
     $410 = HEAP32[$20>>2]|0;
     $411 = $410 & 255;
     $412 = $336 >>> $411;
     $413 = HEAP32[$21>>2]|0;
     $414 = $413 & 255;
     $415 = $412 << $414;
     $416 = $410 >>> 8;
     $417 = $416 & 255;
     $418 = $338 >>> $417;
     $419 = $413 >>> 8;
     $420 = $419 & 255;
     $421 = $418 << $420;
     $422 = $421 | $415;
     $423 = $410 >>> 16;
     $424 = $423 & 255;
     $425 = $340 >>> $424;
     $426 = $413 >>> 16;
     $427 = $426 & 255;
     $428 = $425 << $427;
     $429 = $422 | $428;
     $430 = $410 >>> 24;
     $431 = $345 >>> $430;
     $432 = $413 >>> 24;
     $433 = $431 << $432;
     $434 = $429 | $433;
     HEAP32[$dst$4>>2] = $434;
     break;
    }
    default: {
    }
    }
    $435 = (($dst$4) + ($12)|0);
    $436 = ((($src$4)) + 4|0);
    $dst$5 = $435;$n$4 = $n$3;$src$5 = $436;
    label = 28;
    continue;
   }
   else if ((label|0) == 28) {
    label = 0;
    $437 = HEAP32[$src$5>>2]|0;
    $438 = $437 >>> 22;
    $439 = $438 & 255;
    $440 = $437 >>> 12;
    $441 = $440 & 255;
    $442 = $437 >>> 2;
    $443 = $442 & 255;
    $444 = $437 >>> 30;
    $445 = HEAP32[(125948)>>2]|0;
    $446 = (($445) + ($444)|0);
    $447 = HEAP8[$446>>0]|0;
    $448 = $447&255;
    switch ($12|0) {
    case 1:  {
     $449 = HEAP32[$20>>2]|0;
     $450 = $449 & 255;
     $451 = $439 >>> $450;
     $452 = HEAP32[$21>>2]|0;
     $453 = $452 & 255;
     $454 = $451 << $453;
     $455 = $449 >>> 8;
     $456 = $455 & 255;
     $457 = $441 >>> $456;
     $458 = $452 >>> 8;
     $459 = $458 & 255;
     $460 = $457 << $459;
     $461 = $460 | $454;
     $462 = $449 >>> 16;
     $463 = $462 & 255;
     $464 = $443 >>> $463;
     $465 = $452 >>> 16;
     $466 = $465 & 255;
     $467 = $464 << $466;
     $468 = $461 | $467;
     $469 = $449 >>> 24;
     $470 = $448 >>> $469;
     $471 = $452 >>> 24;
     $472 = $470 << $471;
     $473 = $468 | $472;
     $474 = $473&255;
     HEAP8[$dst$5>>0] = $474;
     break;
    }
    case 2:  {
     $475 = HEAP32[$20>>2]|0;
     $476 = $475 & 255;
     $477 = $439 >>> $476;
     $478 = HEAP32[$21>>2]|0;
     $479 = $478 & 255;
     $480 = $477 << $479;
     $481 = $475 >>> 8;
     $482 = $481 & 255;
     $483 = $441 >>> $482;
     $484 = $478 >>> 8;
     $485 = $484 & 255;
     $486 = $483 << $485;
     $487 = $486 | $480;
     $488 = $475 >>> 16;
     $489 = $488 & 255;
     $490 = $443 >>> $489;
     $491 = $478 >>> 16;
     $492 = $491 & 255;
     $493 = $490 << $492;
     $494 = $487 | $493;
     $495 = $475 >>> 24;
     $496 = $448 >>> $495;
     $497 = $478 >>> 24;
     $498 = $496 << $497;
     $499 = $494 | $498;
     $500 = $499&65535;
     HEAP16[$dst$5>>1] = $500;
     break;
    }
    case 3:  {
     $501 = $438&255;
     $502 = HEAP8[$21>>0]|0;
     $div6 = ($502&255) >>> 3;
     $503 = $div6&255;
     $504 = (($dst$5) + ($503)|0);
     HEAP8[$504>>0] = $501;
     $505 = $440&255;
     $506 = HEAP8[$22>>0]|0;
     $div7 = ($506&255) >>> 3;
     $507 = $div7&255;
     $508 = (($dst$5) + ($507)|0);
     HEAP8[$508>>0] = $505;
     $509 = $442&255;
     $510 = HEAP8[$23>>0]|0;
     $div8 = ($510&255) >>> 3;
     $511 = $div8&255;
     $512 = (($dst$5) + ($511)|0);
     HEAP8[$512>>0] = $509;
     break;
    }
    case 4:  {
     $513 = HEAP32[$20>>2]|0;
     $514 = $513 & 255;
     $515 = $439 >>> $514;
     $516 = HEAP32[$21>>2]|0;
     $517 = $516 & 255;
     $518 = $515 << $517;
     $519 = $513 >>> 8;
     $520 = $519 & 255;
     $521 = $441 >>> $520;
     $522 = $516 >>> 8;
     $523 = $522 & 255;
     $524 = $521 << $523;
     $525 = $524 | $518;
     $526 = $513 >>> 16;
     $527 = $526 & 255;
     $528 = $443 >>> $527;
     $529 = $516 >>> 16;
     $530 = $529 & 255;
     $531 = $528 << $530;
     $532 = $525 | $531;
     $533 = $513 >>> 24;
     $534 = $448 >>> $533;
     $535 = $516 >>> 24;
     $536 = $534 << $535;
     $537 = $532 | $536;
     HEAP32[$dst$5>>2] = $537;
     break;
    }
    default: {
    }
    }
    $538 = (($dst$5) + ($12)|0);
    $539 = ((($src$5)) + 4|0);
    $dst$6 = $538;$n$5 = $n$4;$src$6 = $539;
    label = 34;
    continue;
   }
   else if ((label|0) == 34) {
    label = 0;
    $540 = HEAP32[$src$6>>2]|0;
    $541 = $540 >>> 22;
    $542 = $541 & 255;
    $543 = $540 >>> 12;
    $544 = $543 & 255;
    $545 = $540 >>> 2;
    $546 = $545 & 255;
    $547 = $540 >>> 30;
    $548 = HEAP32[(125948)>>2]|0;
    $549 = (($548) + ($547)|0);
    $550 = HEAP8[$549>>0]|0;
    $551 = $550&255;
    switch ($12|0) {
    case 1:  {
     $552 = HEAP32[$20>>2]|0;
     $553 = $552 & 255;
     $554 = $542 >>> $553;
     $555 = HEAP32[$21>>2]|0;
     $556 = $555 & 255;
     $557 = $554 << $556;
     $558 = $552 >>> 8;
     $559 = $558 & 255;
     $560 = $544 >>> $559;
     $561 = $555 >>> 8;
     $562 = $561 & 255;
     $563 = $560 << $562;
     $564 = $563 | $557;
     $565 = $552 >>> 16;
     $566 = $565 & 255;
     $567 = $546 >>> $566;
     $568 = $555 >>> 16;
     $569 = $568 & 255;
     $570 = $567 << $569;
     $571 = $564 | $570;
     $572 = $552 >>> 24;
     $573 = $551 >>> $572;
     $574 = $555 >>> 24;
     $575 = $573 << $574;
     $576 = $571 | $575;
     $577 = $576&255;
     HEAP8[$dst$6>>0] = $577;
     break;
    }
    case 2:  {
     $578 = HEAP32[$20>>2]|0;
     $579 = $578 & 255;
     $580 = $542 >>> $579;
     $581 = HEAP32[$21>>2]|0;
     $582 = $581 & 255;
     $583 = $580 << $582;
     $584 = $578 >>> 8;
     $585 = $584 & 255;
     $586 = $544 >>> $585;
     $587 = $581 >>> 8;
     $588 = $587 & 255;
     $589 = $586 << $588;
     $590 = $589 | $583;
     $591 = $578 >>> 16;
     $592 = $591 & 255;
     $593 = $546 >>> $592;
     $594 = $581 >>> 16;
     $595 = $594 & 255;
     $596 = $593 << $595;
     $597 = $590 | $596;
     $598 = $578 >>> 24;
     $599 = $551 >>> $598;
     $600 = $581 >>> 24;
     $601 = $599 << $600;
     $602 = $597 | $601;
     $603 = $602&65535;
     HEAP16[$dst$6>>1] = $603;
     break;
    }
    case 3:  {
     $604 = $541&255;
     $605 = HEAP8[$21>>0]|0;
     $div3 = ($605&255) >>> 3;
     $606 = $div3&255;
     $607 = (($dst$6) + ($606)|0);
     HEAP8[$607>>0] = $604;
     $608 = $543&255;
     $609 = HEAP8[$22>>0]|0;
     $div4 = ($609&255) >>> 3;
     $610 = $div4&255;
     $611 = (($dst$6) + ($610)|0);
     HEAP8[$611>>0] = $608;
     $612 = $545&255;
     $613 = HEAP8[$23>>0]|0;
     $div5 = ($613&255) >>> 3;
     $614 = $div5&255;
     $615 = (($dst$6) + ($614)|0);
     HEAP8[$615>>0] = $612;
     break;
    }
    case 4:  {
     $616 = HEAP32[$20>>2]|0;
     $617 = $616 & 255;
     $618 = $542 >>> $617;
     $619 = HEAP32[$21>>2]|0;
     $620 = $619 & 255;
     $621 = $618 << $620;
     $622 = $616 >>> 8;
     $623 = $622 & 255;
     $624 = $544 >>> $623;
     $625 = $619 >>> 8;
     $626 = $625 & 255;
     $627 = $624 << $626;
     $628 = $627 | $621;
     $629 = $616 >>> 16;
     $630 = $629 & 255;
     $631 = $546 >>> $630;
     $632 = $619 >>> 16;
     $633 = $632 & 255;
     $634 = $631 << $633;
     $635 = $628 | $634;
     $636 = $616 >>> 24;
     $637 = $551 >>> $636;
     $638 = $619 >>> 24;
     $639 = $637 << $638;
     $640 = $635 | $639;
     HEAP32[$dst$6>>2] = $640;
     break;
    }
    default: {
    }
    }
    $641 = (($dst$6) + ($12)|0);
    $642 = ((($src$6)) + 4|0);
    $dst$7 = $641;$n$6 = $n$5;$src$7 = $642;
    label = 40;
    continue;
   }
   else if ((label|0) == 40) {
    label = 0;
    $643 = HEAP32[$src$7>>2]|0;
    $644 = $643 >>> 22;
    $645 = $644 & 255;
    $646 = $643 >>> 12;
    $647 = $646 & 255;
    $648 = $643 >>> 2;
    $649 = $648 & 255;
    $650 = $643 >>> 30;
    $651 = HEAP32[(125948)>>2]|0;
    $652 = (($651) + ($650)|0);
    $653 = HEAP8[$652>>0]|0;
    $654 = $653&255;
    switch ($12|0) {
    case 1:  {
     $655 = HEAP32[$20>>2]|0;
     $656 = $655 & 255;
     $657 = $645 >>> $656;
     $658 = HEAP32[$21>>2]|0;
     $659 = $658 & 255;
     $660 = $657 << $659;
     $661 = $655 >>> 8;
     $662 = $661 & 255;
     $663 = $647 >>> $662;
     $664 = $658 >>> 8;
     $665 = $664 & 255;
     $666 = $663 << $665;
     $667 = $666 | $660;
     $668 = $655 >>> 16;
     $669 = $668 & 255;
     $670 = $649 >>> $669;
     $671 = $658 >>> 16;
     $672 = $671 & 255;
     $673 = $670 << $672;
     $674 = $667 | $673;
     $675 = $655 >>> 24;
     $676 = $654 >>> $675;
     $677 = $658 >>> 24;
     $678 = $676 << $677;
     $679 = $674 | $678;
     $680 = $679&255;
     HEAP8[$dst$7>>0] = $680;
     break;
    }
    case 2:  {
     $681 = HEAP32[$20>>2]|0;
     $682 = $681 & 255;
     $683 = $645 >>> $682;
     $684 = HEAP32[$21>>2]|0;
     $685 = $684 & 255;
     $686 = $683 << $685;
     $687 = $681 >>> 8;
     $688 = $687 & 255;
     $689 = $647 >>> $688;
     $690 = $684 >>> 8;
     $691 = $690 & 255;
     $692 = $689 << $691;
     $693 = $692 | $686;
     $694 = $681 >>> 16;
     $695 = $694 & 255;
     $696 = $649 >>> $695;
     $697 = $684 >>> 16;
     $698 = $697 & 255;
     $699 = $696 << $698;
     $700 = $693 | $699;
     $701 = $681 >>> 24;
     $702 = $654 >>> $701;
     $703 = $684 >>> 24;
     $704 = $702 << $703;
     $705 = $700 | $704;
     $706 = $705&65535;
     HEAP16[$dst$7>>1] = $706;
     break;
    }
    case 3:  {
     $707 = $644&255;
     $708 = HEAP8[$21>>0]|0;
     $div = ($708&255) >>> 3;
     $709 = $div&255;
     $710 = (($dst$7) + ($709)|0);
     HEAP8[$710>>0] = $707;
     $711 = $646&255;
     $712 = HEAP8[$22>>0]|0;
     $div1 = ($712&255) >>> 3;
     $713 = $div1&255;
     $714 = (($dst$7) + ($713)|0);
     HEAP8[$714>>0] = $711;
     $715 = $648&255;
     $716 = HEAP8[$23>>0]|0;
     $div2 = ($716&255) >>> 3;
     $717 = $div2&255;
     $718 = (($dst$7) + ($717)|0);
     HEAP8[$718>>0] = $715;
     break;
    }
    case 4:  {
     $719 = HEAP32[$20>>2]|0;
     $720 = $719 & 255;
     $721 = $645 >>> $720;
     $722 = HEAP32[$21>>2]|0;
     $723 = $722 & 255;
     $724 = $721 << $723;
     $725 = $719 >>> 8;
     $726 = $725 & 255;
     $727 = $647 >>> $726;
     $728 = $722 >>> 8;
     $729 = $728 & 255;
     $730 = $727 << $729;
     $731 = $730 | $724;
     $732 = $719 >>> 16;
     $733 = $732 & 255;
     $734 = $649 >>> $733;
     $735 = $722 >>> 16;
     $736 = $735 & 255;
     $737 = $734 << $736;
     $738 = $731 | $737;
     $739 = $719 >>> 24;
     $740 = $654 >>> $739;
     $741 = $722 >>> 24;
     $742 = $740 << $741;
     $743 = $738 | $742;
     HEAP32[$dst$7>>2] = $743;
     break;
    }
    default: {
    }
    }
    $744 = (($dst$7) + ($12)|0);
    $745 = ((($src$7)) + 4|0);
    $dst$8 = $744;$n$7 = $n$6;$src$8 = $745;
    label = 46;
    continue;
   }
   else if ((label|0) == 46) {
    label = 0;
    $746 = HEAP32[$src$8>>2]|0;
    $747 = $746 >>> 22;
    $748 = $747 & 255;
    $749 = $746 >>> 12;
    $750 = $749 & 255;
    $751 = $746 >>> 2;
    $752 = $751 & 255;
    $753 = $746 >>> 30;
    $754 = HEAP32[(125948)>>2]|0;
    $755 = (($754) + ($753)|0);
    $756 = HEAP8[$755>>0]|0;
    $757 = $756&255;
    switch ($12|0) {
    case 1:  {
     $758 = HEAP32[$20>>2]|0;
     $759 = $758 & 255;
     $760 = $748 >>> $759;
     $761 = HEAP32[$21>>2]|0;
     $762 = $761 & 255;
     $763 = $760 << $762;
     $764 = $758 >>> 8;
     $765 = $764 & 255;
     $766 = $750 >>> $765;
     $767 = $761 >>> 8;
     $768 = $767 & 255;
     $769 = $766 << $768;
     $770 = $769 | $763;
     $771 = $758 >>> 16;
     $772 = $771 & 255;
     $773 = $752 >>> $772;
     $774 = $761 >>> 16;
     $775 = $774 & 255;
     $776 = $773 << $775;
     $777 = $770 | $776;
     $778 = $758 >>> 24;
     $779 = $757 >>> $778;
     $780 = $761 >>> 24;
     $781 = $779 << $780;
     $782 = $777 | $781;
     $783 = $782&255;
     HEAP8[$dst$8>>0] = $783;
     break;
    }
    case 2:  {
     $784 = HEAP32[$20>>2]|0;
     $785 = $784 & 255;
     $786 = $748 >>> $785;
     $787 = HEAP32[$21>>2]|0;
     $788 = $787 & 255;
     $789 = $786 << $788;
     $790 = $784 >>> 8;
     $791 = $790 & 255;
     $792 = $750 >>> $791;
     $793 = $787 >>> 8;
     $794 = $793 & 255;
     $795 = $792 << $794;
     $796 = $795 | $789;
     $797 = $784 >>> 16;
     $798 = $797 & 255;
     $799 = $752 >>> $798;
     $800 = $787 >>> 16;
     $801 = $800 & 255;
     $802 = $799 << $801;
     $803 = $796 | $802;
     $804 = $784 >>> 24;
     $805 = $757 >>> $804;
     $806 = $787 >>> 24;
     $807 = $805 << $806;
     $808 = $803 | $807;
     $809 = $808&65535;
     HEAP16[$dst$8>>1] = $809;
     break;
    }
    case 3:  {
     $810 = $747&255;
     $811 = HEAP8[$21>>0]|0;
     $div21 = ($811&255) >>> 3;
     $812 = $div21&255;
     $813 = (($dst$8) + ($812)|0);
     HEAP8[$813>>0] = $810;
     $814 = $749&255;
     $815 = HEAP8[$22>>0]|0;
     $div22 = ($815&255) >>> 3;
     $816 = $div22&255;
     $817 = (($dst$8) + ($816)|0);
     HEAP8[$817>>0] = $814;
     $818 = $751&255;
     $819 = HEAP8[$23>>0]|0;
     $div23 = ($819&255) >>> 3;
     $820 = $div23&255;
     $821 = (($dst$8) + ($820)|0);
     HEAP8[$821>>0] = $818;
     break;
    }
    case 4:  {
     $822 = HEAP32[$20>>2]|0;
     $823 = $822 & 255;
     $824 = $748 >>> $823;
     $825 = HEAP32[$21>>2]|0;
     $826 = $825 & 255;
     $827 = $824 << $826;
     $828 = $822 >>> 8;
     $829 = $828 & 255;
     $830 = $750 >>> $829;
     $831 = $825 >>> 8;
     $832 = $831 & 255;
     $833 = $830 << $832;
     $834 = $833 | $827;
     $835 = $822 >>> 16;
     $836 = $835 & 255;
     $837 = $752 >>> $836;
     $838 = $825 >>> 16;
     $839 = $838 & 255;
     $840 = $837 << $839;
     $841 = $834 | $840;
     $842 = $822 >>> 24;
     $843 = $757 >>> $842;
     $844 = $825 >>> 24;
     $845 = $843 << $844;
     $846 = $841 | $845;
     HEAP32[$dst$8>>2] = $846;
     break;
    }
    default: {
    }
    }
    $847 = (($dst$8) + ($12)|0);
    $848 = ((($src$8)) + 4|0);
    $849 = (($n$7) + -1)|0;
    $850 = ($n$7|0)>(1);
    if ($850) {
     $dst$1 = $847;$n$0 = $849;$src$1 = $848;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $851 = (($848) + ($5)|0);
  $852 = (($847) + ($7)|0);
  $853 = ($24|0)==(0);
  if ($853) {
   label = 54;
   break;
  } else {
   $$in = $24;$dst$024 = $852;$src$025 = $851;
  }
 }
 if ((label|0) == 52) {
  // unreachable;
 }
 else if ((label|0) == 54) {
  return;
 }
}
function _BlitNto2101010($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0;
 var $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0;
 var $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0;
 var $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0;
 var $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0;
 var $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0;
 var $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0;
 var $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0;
 var $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0;
 var $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0;
 var $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0;
 var $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0;
 var $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0;
 var $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0;
 var $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0;
 var $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0;
 var $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0;
 var $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0;
 var $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0;
 var $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0;
 var $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0;
 var $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0;
 var $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0;
 var $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0;
 var $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0;
 var $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0;
 var $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0;
 var $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0;
 var $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0;
 var $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0;
 var $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0;
 var $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0;
 var $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0;
 var $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0;
 var $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0;
 var $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0;
 var $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0;
 var $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0;
 var $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0;
 var $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0;
 var $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0;
 var $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0;
 var $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0;
 var $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0;
 var $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0;
 var $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0;
 var $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0;
 var $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0;
 var $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0;
 var $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0;
 var $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0;
 var $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0;
 var $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0;
 var $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0;
 var $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0;
 var $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0;
 var $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0;
 var $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $div = 0, $div1 = 0, $div11 = 0, $div12 = 0, $div14 = 0, $div15 = 0, $div17 = 0, $div18 = 0, $div2 = 0, $div20 = 0, $div21 = 0, $div23 = 0, $div24 = 0, $div26 = 0, $div27 = 0;
 var $div29 = 0, $div3 = 0, $div30 = 0, $div4 = 0, $div5 = 0, $div6 = 0, $div7 = 0, $div8 = 0, $div9 = 0, $dst$095 = 0, $dst$1 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $n$0 = 0, $n$1 = 0;
 var $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $sA$0$shrunk = 0, $sA$0$shrunk36 = 0, $sA$1$shrunk = 0, $sA$1$shrunk44 = 0, $sA$2$shrunk = 0, $sA$2$shrunk52 = 0, $sA$3$shrunk = 0, $sA$3$shrunk60 = 0, $sA$4$shrunk = 0, $sA$4$shrunk68 = 0, $sA$5$shrunk = 0, $sA$5$shrunk76 = 0, $sA$6$shrunk = 0, $sA$6$shrunk84 = 0;
 var $sA$7$shrunk = 0, $sA$7$shrunk92 = 0, $sB$0 = 0, $sB$0$shrunk = 0, $sB$0$shrunk$in = 0, $sB$0$shrunk35 = 0, $sB$037 = 0, $sB$1 = 0, $sB$1$shrunk = 0, $sB$1$shrunk$in = 0, $sB$1$shrunk43 = 0, $sB$145 = 0, $sB$2 = 0, $sB$2$shrunk = 0, $sB$2$shrunk$in = 0, $sB$2$shrunk51 = 0, $sB$253 = 0, $sB$3 = 0, $sB$3$shrunk = 0, $sB$3$shrunk$in = 0;
 var $sB$3$shrunk59 = 0, $sB$361 = 0, $sB$4 = 0, $sB$4$shrunk = 0, $sB$4$shrunk$in = 0, $sB$4$shrunk67 = 0, $sB$469 = 0, $sB$5 = 0, $sB$5$shrunk = 0, $sB$5$shrunk$in = 0, $sB$5$shrunk75 = 0, $sB$577 = 0, $sB$6 = 0, $sB$6$shrunk = 0, $sB$6$shrunk$in = 0, $sB$6$shrunk83 = 0, $sB$685 = 0, $sB$7 = 0, $sB$7$shrunk = 0, $sB$7$shrunk$in = 0;
 var $sB$7$shrunk91 = 0, $sB$793 = 0, $sG$0 = 0, $sG$0$shrunk = 0, $sG$0$shrunk$in = 0, $sG$0$shrunk34 = 0, $sG$038 = 0, $sG$1 = 0, $sG$1$shrunk = 0, $sG$1$shrunk$in = 0, $sG$1$shrunk42 = 0, $sG$146 = 0, $sG$2 = 0, $sG$2$shrunk = 0, $sG$2$shrunk$in = 0, $sG$2$shrunk50 = 0, $sG$254 = 0, $sG$3 = 0, $sG$3$shrunk = 0, $sG$3$shrunk$in = 0;
 var $sG$3$shrunk58 = 0, $sG$362 = 0, $sG$4 = 0, $sG$4$shrunk = 0, $sG$4$shrunk$in = 0, $sG$4$shrunk66 = 0, $sG$470 = 0, $sG$5 = 0, $sG$5$shrunk = 0, $sG$5$shrunk$in = 0, $sG$5$shrunk74 = 0, $sG$578 = 0, $sG$6 = 0, $sG$6$shrunk = 0, $sG$6$shrunk$in = 0, $sG$6$shrunk82 = 0, $sG$686 = 0, $sG$7 = 0, $sG$7$shrunk = 0, $sG$7$shrunk$in = 0;
 var $sG$7$shrunk90 = 0, $sG$794 = 0, $sR$0 = 0, $sR$0$shrunk = 0, $sR$0$shrunk$in = 0, $sR$1 = 0, $sR$1$shrunk = 0, $sR$1$shrunk$in = 0, $sR$2 = 0, $sR$2$shrunk = 0, $sR$2$shrunk$in = 0, $sR$3 = 0, $sR$3$shrunk = 0, $sR$3$shrunk$in = 0, $sR$4 = 0, $sR$4$shrunk = 0, $sR$4$shrunk$in = 0, $sR$5 = 0, $sR$5$shrunk = 0, $sR$5$shrunk$in = 0;
 var $sR$6 = 0, $sR$6$shrunk = 0, $sR$6$shrunk$in = 0, $sR$7 = 0, $sR$7$shrunk = 0, $sR$7$shrunk$in = 0, $src$096 = 0, $src$1 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 40|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 9|0);
 $11 = HEAP8[$10>>0]|0;
 $12 = $11&255;
 $13 = ($3|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = HEAP32[$info>>2]|0;
 $17 = (($1) + 7)|0;
 $18 = (($17|0) / 8)&-1;
 $19 = $1 & 7;
 $20 = ((($9)) + 12|0);
 $21 = ((($9)) + 32|0);
 $22 = ((($9)) + 28|0);
 $23 = ((($9)) + 16|0);
 $24 = ((($9)) + 20|0);
 $25 = ((($9)) + 24|0);
 $$in = $3;$dst$095 = $15;$src$096 = $16;
 L4: while(1) {
  $26 = (($$in) + -1)|0;
  switch ($19|0) {
  case 0:  {
   $dst$1 = $dst$095;$n$0 = $18;$src$1 = $src$096;
   label = 4;
   break;
  }
  case 7:  {
   $dst$2 = $dst$095;$n$1 = $18;$src$2 = $src$096;
   label = 12;
   break;
  }
  case 6:  {
   $dst$3 = $dst$095;$n$2 = $18;$src$3 = $src$096;
   label = 20;
   break;
  }
  case 5:  {
   $dst$4 = $dst$095;$n$3 = $18;$src$4 = $src$096;
   label = 28;
   break;
  }
  case 4:  {
   $dst$5 = $dst$095;$n$4 = $18;$src$5 = $src$096;
   label = 36;
   break;
  }
  case 3:  {
   $dst$6 = $dst$095;$n$5 = $18;$src$6 = $src$096;
   label = 44;
   break;
  }
  case 2:  {
   $dst$7 = $dst$095;$n$6 = $18;$src$7 = $src$096;
   label = 52;
   break;
  }
  case 1:  {
   $dst$8 = $dst$095;$n$7 = $18;$src$8 = $src$096;
   label = 60;
   break;
  }
  default: {
   label = 68;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    switch ($12|0) {
    case 1:  {
     $27 = HEAP8[$src$1>>0]|0;
     $28 = $27&255;
     $29 = HEAP32[$20>>2]|0;
     $30 = $28 & $29;
     $31 = HEAP32[$21>>2]|0;
     $32 = $31 & 255;
     $33 = $30 >>> $32;
     $34 = HEAP32[$22>>2]|0;
     $35 = $34 & 255;
     $36 = (125924 + ($35<<2)|0);
     $37 = HEAP32[$36>>2]|0;
     $38 = (($37) + ($33)|0);
     $39 = HEAP32[$23>>2]|0;
     $40 = $39 & $28;
     $41 = $31 >>> 8;
     $42 = $41 & 255;
     $43 = $40 >>> $42;
     $44 = $34 >>> 8;
     $45 = $44 & 255;
     $46 = (125924 + ($45<<2)|0);
     $47 = HEAP32[$46>>2]|0;
     $48 = (($47) + ($43)|0);
     $49 = HEAP32[$24>>2]|0;
     $50 = $49 & $28;
     $51 = $31 >>> 16;
     $52 = $51 & 255;
     $53 = $50 >>> $52;
     $54 = $34 >>> 16;
     $55 = $54 & 255;
     $56 = (125924 + ($55<<2)|0);
     $57 = HEAP32[$56>>2]|0;
     $58 = (($57) + ($53)|0);
     $59 = HEAP32[$25>>2]|0;
     $60 = $59 & $28;
     $61 = $31 >>> 24;
     $62 = $60 >>> $61;
     $63 = $34 >>> 24;
     $64 = (125924 + ($63<<2)|0);
     $65 = HEAP32[$64>>2]|0;
     $66 = (($65) + ($62)|0);
     $67 = HEAP8[$66>>0]|0;
     $sA$0$shrunk = $67;$sB$0$shrunk$in = $58;$sG$0$shrunk$in = $48;$sR$0$shrunk$in = $38;
     label = 9;
     break;
    }
    case 2:  {
     $68 = HEAP16[$src$1>>1]|0;
     $69 = $68&65535;
     $70 = HEAP32[$20>>2]|0;
     $71 = $69 & $70;
     $72 = HEAP32[$21>>2]|0;
     $73 = $72 & 255;
     $74 = $71 >>> $73;
     $75 = HEAP32[$22>>2]|0;
     $76 = $75 & 255;
     $77 = (125924 + ($76<<2)|0);
     $78 = HEAP32[$77>>2]|0;
     $79 = (($78) + ($74)|0);
     $80 = HEAP32[$23>>2]|0;
     $81 = $80 & $69;
     $82 = $72 >>> 8;
     $83 = $82 & 255;
     $84 = $81 >>> $83;
     $85 = $75 >>> 8;
     $86 = $85 & 255;
     $87 = (125924 + ($86<<2)|0);
     $88 = HEAP32[$87>>2]|0;
     $89 = (($88) + ($84)|0);
     $90 = HEAP32[$24>>2]|0;
     $91 = $90 & $69;
     $92 = $72 >>> 16;
     $93 = $92 & 255;
     $94 = $91 >>> $93;
     $95 = $75 >>> 16;
     $96 = $95 & 255;
     $97 = (125924 + ($96<<2)|0);
     $98 = HEAP32[$97>>2]|0;
     $99 = (($98) + ($94)|0);
     $100 = HEAP32[$25>>2]|0;
     $101 = $100 & $69;
     $102 = $72 >>> 24;
     $103 = $101 >>> $102;
     $104 = $75 >>> 24;
     $105 = (125924 + ($104<<2)|0);
     $106 = HEAP32[$105>>2]|0;
     $107 = (($106) + ($103)|0);
     $108 = HEAP8[$107>>0]|0;
     $sA$0$shrunk = $108;$sB$0$shrunk$in = $99;$sG$0$shrunk$in = $89;$sR$0$shrunk$in = $79;
     label = 9;
     break;
    }
    case 3:  {
     $109 = HEAP32[$21>>2]|0;
     $div26 = $109 >>> 3;
     $110 = $div26 & 31;
     $111 = (($src$1) + ($110)|0);
     $112 = $109 >>> 11;
     $div27 = $112 & 31;
     $113 = (($src$1) + ($div27)|0);
     $114 = $109 >>> 19;
     $115 = $114 & 31;
     $116 = (($src$1) + ($115)|0);
     $sA$0$shrunk = -1;$sB$0$shrunk$in = $116;$sG$0$shrunk$in = $113;$sR$0$shrunk$in = $111;
     label = 9;
     break;
    }
    case 4:  {
     $117 = HEAP32[$src$1>>2]|0;
     $118 = HEAP32[$20>>2]|0;
     $119 = $118 & $117;
     $120 = HEAP32[$21>>2]|0;
     $121 = $120 & 255;
     $122 = $119 >>> $121;
     $123 = HEAP32[$22>>2]|0;
     $124 = $123 & 255;
     $125 = (125924 + ($124<<2)|0);
     $126 = HEAP32[$125>>2]|0;
     $127 = (($126) + ($122)|0);
     $128 = HEAP32[$23>>2]|0;
     $129 = $128 & $117;
     $130 = $120 >>> 8;
     $131 = $130 & 255;
     $132 = $129 >>> $131;
     $133 = $123 >>> 8;
     $134 = $133 & 255;
     $135 = (125924 + ($134<<2)|0);
     $136 = HEAP32[$135>>2]|0;
     $137 = (($136) + ($132)|0);
     $138 = HEAP32[$24>>2]|0;
     $139 = $138 & $117;
     $140 = $120 >>> 16;
     $141 = $140 & 255;
     $142 = $139 >>> $141;
     $143 = $123 >>> 16;
     $144 = $143 & 255;
     $145 = (125924 + ($144<<2)|0);
     $146 = HEAP32[$145>>2]|0;
     $147 = (($146) + ($142)|0);
     $148 = HEAP32[$25>>2]|0;
     $149 = $148 & $117;
     $150 = $120 >>> 24;
     $151 = $149 >>> $150;
     $152 = $123 >>> 24;
     $153 = (125924 + ($152<<2)|0);
     $154 = HEAP32[$153>>2]|0;
     $155 = (($154) + ($151)|0);
     $156 = HEAP8[$155>>0]|0;
     $sA$0$shrunk = $156;$sB$0$shrunk$in = $147;$sG$0$shrunk$in = $137;$sR$0$shrunk$in = $127;
     label = 9;
     break;
    }
    default: {
     $171 = 0;$sA$0$shrunk36 = 0;$sB$0$shrunk35 = 0;$sB$037 = 0;$sG$0$shrunk34 = 0;$sG$038 = 0;
    }
    }
    if ((label|0) == 9) {
     label = 0;
     $sB$0$shrunk = HEAP8[$sB$0$shrunk$in>>0]|0;
     $sG$0$shrunk = HEAP8[$sG$0$shrunk$in>>0]|0;
     $sR$0$shrunk = HEAP8[$sR$0$shrunk$in>>0]|0;
     $sB$0 = $sB$0$shrunk&255;
     $sG$0 = $sG$0$shrunk&255;
     $157 = ($sR$0$shrunk<<24>>24)==(0);
     if ($157) {
      $171 = 0;$sA$0$shrunk36 = $sA$0$shrunk;$sB$0$shrunk35 = $sB$0$shrunk;$sB$037 = $sB$0;$sG$0$shrunk34 = $sG$0$shrunk;$sG$038 = $sG$0;
     } else {
      $sR$0 = $sR$0$shrunk&255;
      $158 = $sR$0 << 22;
      $159 = $158 | 3145728;
      $171 = $159;$sA$0$shrunk36 = $sA$0$shrunk;$sB$0$shrunk35 = $sB$0$shrunk;$sB$037 = $sB$0;$sG$0$shrunk34 = $sG$0$shrunk;$sG$038 = $sG$0;
     }
    }
    $160 = ($sG$0$shrunk34<<24>>24)==(0);
    $161 = $sG$038 << 12;
    $162 = $161 | 3072;
    $163 = $160 ? 0 : $162;
    $164 = ($sB$0$shrunk35<<24>>24)==(0);
    $165 = $sB$037 << 2;
    $166 = $165 | 3;
    $167 = $164 ? 0 : $166;
    $div1 = (($sA$0$shrunk36&255) / 85)&-1;
    $168 = $div1&255;
    $169 = $168 << 30;
    $170 = $169 | $171;
    $172 = $170 | $167;
    $173 = $172 | $163;
    HEAP32[$dst$1>>2] = $173;
    $174 = ((($dst$1)) + 4|0);
    $175 = (($src$1) + ($12)|0);
    $dst$2 = $174;$n$1 = $n$0;$src$2 = $175;
    label = 12;
    continue;
   }
   else if ((label|0) == 12) {
    label = 0;
    switch ($12|0) {
    case 1:  {
     $176 = HEAP8[$src$2>>0]|0;
     $177 = $176&255;
     $178 = HEAP32[$20>>2]|0;
     $179 = $177 & $178;
     $180 = HEAP32[$21>>2]|0;
     $181 = $180 & 255;
     $182 = $179 >>> $181;
     $183 = HEAP32[$22>>2]|0;
     $184 = $183 & 255;
     $185 = (125924 + ($184<<2)|0);
     $186 = HEAP32[$185>>2]|0;
     $187 = (($186) + ($182)|0);
     $188 = HEAP32[$23>>2]|0;
     $189 = $188 & $177;
     $190 = $180 >>> 8;
     $191 = $190 & 255;
     $192 = $189 >>> $191;
     $193 = $183 >>> 8;
     $194 = $193 & 255;
     $195 = (125924 + ($194<<2)|0);
     $196 = HEAP32[$195>>2]|0;
     $197 = (($196) + ($192)|0);
     $198 = HEAP32[$24>>2]|0;
     $199 = $198 & $177;
     $200 = $180 >>> 16;
     $201 = $200 & 255;
     $202 = $199 >>> $201;
     $203 = $183 >>> 16;
     $204 = $203 & 255;
     $205 = (125924 + ($204<<2)|0);
     $206 = HEAP32[$205>>2]|0;
     $207 = (($206) + ($202)|0);
     $208 = HEAP32[$25>>2]|0;
     $209 = $208 & $177;
     $210 = $180 >>> 24;
     $211 = $209 >>> $210;
     $212 = $183 >>> 24;
     $213 = (125924 + ($212<<2)|0);
     $214 = HEAP32[$213>>2]|0;
     $215 = (($214) + ($211)|0);
     $216 = HEAP8[$215>>0]|0;
     $sA$1$shrunk = $216;$sB$1$shrunk$in = $207;$sG$1$shrunk$in = $197;$sR$1$shrunk$in = $187;
     label = 17;
     break;
    }
    case 2:  {
     $217 = HEAP16[$src$2>>1]|0;
     $218 = $217&65535;
     $219 = HEAP32[$20>>2]|0;
     $220 = $218 & $219;
     $221 = HEAP32[$21>>2]|0;
     $222 = $221 & 255;
     $223 = $220 >>> $222;
     $224 = HEAP32[$22>>2]|0;
     $225 = $224 & 255;
     $226 = (125924 + ($225<<2)|0);
     $227 = HEAP32[$226>>2]|0;
     $228 = (($227) + ($223)|0);
     $229 = HEAP32[$23>>2]|0;
     $230 = $229 & $218;
     $231 = $221 >>> 8;
     $232 = $231 & 255;
     $233 = $230 >>> $232;
     $234 = $224 >>> 8;
     $235 = $234 & 255;
     $236 = (125924 + ($235<<2)|0);
     $237 = HEAP32[$236>>2]|0;
     $238 = (($237) + ($233)|0);
     $239 = HEAP32[$24>>2]|0;
     $240 = $239 & $218;
     $241 = $221 >>> 16;
     $242 = $241 & 255;
     $243 = $240 >>> $242;
     $244 = $224 >>> 16;
     $245 = $244 & 255;
     $246 = (125924 + ($245<<2)|0);
     $247 = HEAP32[$246>>2]|0;
     $248 = (($247) + ($243)|0);
     $249 = HEAP32[$25>>2]|0;
     $250 = $249 & $218;
     $251 = $221 >>> 24;
     $252 = $250 >>> $251;
     $253 = $224 >>> 24;
     $254 = (125924 + ($253<<2)|0);
     $255 = HEAP32[$254>>2]|0;
     $256 = (($255) + ($252)|0);
     $257 = HEAP8[$256>>0]|0;
     $sA$1$shrunk = $257;$sB$1$shrunk$in = $248;$sG$1$shrunk$in = $238;$sR$1$shrunk$in = $228;
     label = 17;
     break;
    }
    case 3:  {
     $258 = HEAP32[$21>>2]|0;
     $div23 = $258 >>> 3;
     $259 = $div23 & 31;
     $260 = (($src$2) + ($259)|0);
     $261 = $258 >>> 11;
     $div24 = $261 & 31;
     $262 = (($src$2) + ($div24)|0);
     $263 = $258 >>> 19;
     $264 = $263 & 31;
     $265 = (($src$2) + ($264)|0);
     $sA$1$shrunk = -1;$sB$1$shrunk$in = $265;$sG$1$shrunk$in = $262;$sR$1$shrunk$in = $260;
     label = 17;
     break;
    }
    case 4:  {
     $266 = HEAP32[$src$2>>2]|0;
     $267 = HEAP32[$20>>2]|0;
     $268 = $267 & $266;
     $269 = HEAP32[$21>>2]|0;
     $270 = $269 & 255;
     $271 = $268 >>> $270;
     $272 = HEAP32[$22>>2]|0;
     $273 = $272 & 255;
     $274 = (125924 + ($273<<2)|0);
     $275 = HEAP32[$274>>2]|0;
     $276 = (($275) + ($271)|0);
     $277 = HEAP32[$23>>2]|0;
     $278 = $277 & $266;
     $279 = $269 >>> 8;
     $280 = $279 & 255;
     $281 = $278 >>> $280;
     $282 = $272 >>> 8;
     $283 = $282 & 255;
     $284 = (125924 + ($283<<2)|0);
     $285 = HEAP32[$284>>2]|0;
     $286 = (($285) + ($281)|0);
     $287 = HEAP32[$24>>2]|0;
     $288 = $287 & $266;
     $289 = $269 >>> 16;
     $290 = $289 & 255;
     $291 = $288 >>> $290;
     $292 = $272 >>> 16;
     $293 = $292 & 255;
     $294 = (125924 + ($293<<2)|0);
     $295 = HEAP32[$294>>2]|0;
     $296 = (($295) + ($291)|0);
     $297 = HEAP32[$25>>2]|0;
     $298 = $297 & $266;
     $299 = $269 >>> 24;
     $300 = $298 >>> $299;
     $301 = $272 >>> 24;
     $302 = (125924 + ($301<<2)|0);
     $303 = HEAP32[$302>>2]|0;
     $304 = (($303) + ($300)|0);
     $305 = HEAP8[$304>>0]|0;
     $sA$1$shrunk = $305;$sB$1$shrunk$in = $296;$sG$1$shrunk$in = $286;$sR$1$shrunk$in = $276;
     label = 17;
     break;
    }
    default: {
     $320 = 0;$sA$1$shrunk44 = 0;$sB$1$shrunk43 = 0;$sB$145 = 0;$sG$1$shrunk42 = 0;$sG$146 = 0;
    }
    }
    if ((label|0) == 17) {
     label = 0;
     $sB$1$shrunk = HEAP8[$sB$1$shrunk$in>>0]|0;
     $sG$1$shrunk = HEAP8[$sG$1$shrunk$in>>0]|0;
     $sR$1$shrunk = HEAP8[$sR$1$shrunk$in>>0]|0;
     $sB$1 = $sB$1$shrunk&255;
     $sG$1 = $sG$1$shrunk&255;
     $306 = ($sR$1$shrunk<<24>>24)==(0);
     if ($306) {
      $320 = 0;$sA$1$shrunk44 = $sA$1$shrunk;$sB$1$shrunk43 = $sB$1$shrunk;$sB$145 = $sB$1;$sG$1$shrunk42 = $sG$1$shrunk;$sG$146 = $sG$1;
     } else {
      $sR$1 = $sR$1$shrunk&255;
      $307 = $sR$1 << 22;
      $308 = $307 | 3145728;
      $320 = $308;$sA$1$shrunk44 = $sA$1$shrunk;$sB$1$shrunk43 = $sB$1$shrunk;$sB$145 = $sB$1;$sG$1$shrunk42 = $sG$1$shrunk;$sG$146 = $sG$1;
     }
    }
    $309 = ($sG$1$shrunk42<<24>>24)==(0);
    $310 = $sG$146 << 12;
    $311 = $310 | 3072;
    $312 = $309 ? 0 : $311;
    $313 = ($sB$1$shrunk43<<24>>24)==(0);
    $314 = $sB$145 << 2;
    $315 = $314 | 3;
    $316 = $313 ? 0 : $315;
    $div2 = (($sA$1$shrunk44&255) / 85)&-1;
    $317 = $div2&255;
    $318 = $317 << 30;
    $319 = $318 | $320;
    $321 = $319 | $316;
    $322 = $321 | $312;
    HEAP32[$dst$2>>2] = $322;
    $323 = ((($dst$2)) + 4|0);
    $324 = (($src$2) + ($12)|0);
    $dst$3 = $323;$n$2 = $n$1;$src$3 = $324;
    label = 20;
    continue;
   }
   else if ((label|0) == 20) {
    label = 0;
    switch ($12|0) {
    case 1:  {
     $325 = HEAP8[$src$3>>0]|0;
     $326 = $325&255;
     $327 = HEAP32[$20>>2]|0;
     $328 = $326 & $327;
     $329 = HEAP32[$21>>2]|0;
     $330 = $329 & 255;
     $331 = $328 >>> $330;
     $332 = HEAP32[$22>>2]|0;
     $333 = $332 & 255;
     $334 = (125924 + ($333<<2)|0);
     $335 = HEAP32[$334>>2]|0;
     $336 = (($335) + ($331)|0);
     $337 = HEAP32[$23>>2]|0;
     $338 = $337 & $326;
     $339 = $329 >>> 8;
     $340 = $339 & 255;
     $341 = $338 >>> $340;
     $342 = $332 >>> 8;
     $343 = $342 & 255;
     $344 = (125924 + ($343<<2)|0);
     $345 = HEAP32[$344>>2]|0;
     $346 = (($345) + ($341)|0);
     $347 = HEAP32[$24>>2]|0;
     $348 = $347 & $326;
     $349 = $329 >>> 16;
     $350 = $349 & 255;
     $351 = $348 >>> $350;
     $352 = $332 >>> 16;
     $353 = $352 & 255;
     $354 = (125924 + ($353<<2)|0);
     $355 = HEAP32[$354>>2]|0;
     $356 = (($355) + ($351)|0);
     $357 = HEAP32[$25>>2]|0;
     $358 = $357 & $326;
     $359 = $329 >>> 24;
     $360 = $358 >>> $359;
     $361 = $332 >>> 24;
     $362 = (125924 + ($361<<2)|0);
     $363 = HEAP32[$362>>2]|0;
     $364 = (($363) + ($360)|0);
     $365 = HEAP8[$364>>0]|0;
     $sA$2$shrunk = $365;$sB$2$shrunk$in = $356;$sG$2$shrunk$in = $346;$sR$2$shrunk$in = $336;
     label = 25;
     break;
    }
    case 2:  {
     $366 = HEAP16[$src$3>>1]|0;
     $367 = $366&65535;
     $368 = HEAP32[$20>>2]|0;
     $369 = $367 & $368;
     $370 = HEAP32[$21>>2]|0;
     $371 = $370 & 255;
     $372 = $369 >>> $371;
     $373 = HEAP32[$22>>2]|0;
     $374 = $373 & 255;
     $375 = (125924 + ($374<<2)|0);
     $376 = HEAP32[$375>>2]|0;
     $377 = (($376) + ($372)|0);
     $378 = HEAP32[$23>>2]|0;
     $379 = $378 & $367;
     $380 = $370 >>> 8;
     $381 = $380 & 255;
     $382 = $379 >>> $381;
     $383 = $373 >>> 8;
     $384 = $383 & 255;
     $385 = (125924 + ($384<<2)|0);
     $386 = HEAP32[$385>>2]|0;
     $387 = (($386) + ($382)|0);
     $388 = HEAP32[$24>>2]|0;
     $389 = $388 & $367;
     $390 = $370 >>> 16;
     $391 = $390 & 255;
     $392 = $389 >>> $391;
     $393 = $373 >>> 16;
     $394 = $393 & 255;
     $395 = (125924 + ($394<<2)|0);
     $396 = HEAP32[$395>>2]|0;
     $397 = (($396) + ($392)|0);
     $398 = HEAP32[$25>>2]|0;
     $399 = $398 & $367;
     $400 = $370 >>> 24;
     $401 = $399 >>> $400;
     $402 = $373 >>> 24;
     $403 = (125924 + ($402<<2)|0);
     $404 = HEAP32[$403>>2]|0;
     $405 = (($404) + ($401)|0);
     $406 = HEAP8[$405>>0]|0;
     $sA$2$shrunk = $406;$sB$2$shrunk$in = $397;$sG$2$shrunk$in = $387;$sR$2$shrunk$in = $377;
     label = 25;
     break;
    }
    case 3:  {
     $407 = HEAP32[$21>>2]|0;
     $div20 = $407 >>> 3;
     $408 = $div20 & 31;
     $409 = (($src$3) + ($408)|0);
     $410 = $407 >>> 11;
     $div21 = $410 & 31;
     $411 = (($src$3) + ($div21)|0);
     $412 = $407 >>> 19;
     $413 = $412 & 31;
     $414 = (($src$3) + ($413)|0);
     $sA$2$shrunk = -1;$sB$2$shrunk$in = $414;$sG$2$shrunk$in = $411;$sR$2$shrunk$in = $409;
     label = 25;
     break;
    }
    case 4:  {
     $415 = HEAP32[$src$3>>2]|0;
     $416 = HEAP32[$20>>2]|0;
     $417 = $416 & $415;
     $418 = HEAP32[$21>>2]|0;
     $419 = $418 & 255;
     $420 = $417 >>> $419;
     $421 = HEAP32[$22>>2]|0;
     $422 = $421 & 255;
     $423 = (125924 + ($422<<2)|0);
     $424 = HEAP32[$423>>2]|0;
     $425 = (($424) + ($420)|0);
     $426 = HEAP32[$23>>2]|0;
     $427 = $426 & $415;
     $428 = $418 >>> 8;
     $429 = $428 & 255;
     $430 = $427 >>> $429;
     $431 = $421 >>> 8;
     $432 = $431 & 255;
     $433 = (125924 + ($432<<2)|0);
     $434 = HEAP32[$433>>2]|0;
     $435 = (($434) + ($430)|0);
     $436 = HEAP32[$24>>2]|0;
     $437 = $436 & $415;
     $438 = $418 >>> 16;
     $439 = $438 & 255;
     $440 = $437 >>> $439;
     $441 = $421 >>> 16;
     $442 = $441 & 255;
     $443 = (125924 + ($442<<2)|0);
     $444 = HEAP32[$443>>2]|0;
     $445 = (($444) + ($440)|0);
     $446 = HEAP32[$25>>2]|0;
     $447 = $446 & $415;
     $448 = $418 >>> 24;
     $449 = $447 >>> $448;
     $450 = $421 >>> 24;
     $451 = (125924 + ($450<<2)|0);
     $452 = HEAP32[$451>>2]|0;
     $453 = (($452) + ($449)|0);
     $454 = HEAP8[$453>>0]|0;
     $sA$2$shrunk = $454;$sB$2$shrunk$in = $445;$sG$2$shrunk$in = $435;$sR$2$shrunk$in = $425;
     label = 25;
     break;
    }
    default: {
     $469 = 0;$sA$2$shrunk52 = 0;$sB$2$shrunk51 = 0;$sB$253 = 0;$sG$2$shrunk50 = 0;$sG$254 = 0;
    }
    }
    if ((label|0) == 25) {
     label = 0;
     $sB$2$shrunk = HEAP8[$sB$2$shrunk$in>>0]|0;
     $sG$2$shrunk = HEAP8[$sG$2$shrunk$in>>0]|0;
     $sR$2$shrunk = HEAP8[$sR$2$shrunk$in>>0]|0;
     $sB$2 = $sB$2$shrunk&255;
     $sG$2 = $sG$2$shrunk&255;
     $455 = ($sR$2$shrunk<<24>>24)==(0);
     if ($455) {
      $469 = 0;$sA$2$shrunk52 = $sA$2$shrunk;$sB$2$shrunk51 = $sB$2$shrunk;$sB$253 = $sB$2;$sG$2$shrunk50 = $sG$2$shrunk;$sG$254 = $sG$2;
     } else {
      $sR$2 = $sR$2$shrunk&255;
      $456 = $sR$2 << 22;
      $457 = $456 | 3145728;
      $469 = $457;$sA$2$shrunk52 = $sA$2$shrunk;$sB$2$shrunk51 = $sB$2$shrunk;$sB$253 = $sB$2;$sG$2$shrunk50 = $sG$2$shrunk;$sG$254 = $sG$2;
     }
    }
    $458 = ($sG$2$shrunk50<<24>>24)==(0);
    $459 = $sG$254 << 12;
    $460 = $459 | 3072;
    $461 = $458 ? 0 : $460;
    $462 = ($sB$2$shrunk51<<24>>24)==(0);
    $463 = $sB$253 << 2;
    $464 = $463 | 3;
    $465 = $462 ? 0 : $464;
    $div3 = (($sA$2$shrunk52&255) / 85)&-1;
    $466 = $div3&255;
    $467 = $466 << 30;
    $468 = $467 | $469;
    $470 = $468 | $465;
    $471 = $470 | $461;
    HEAP32[$dst$3>>2] = $471;
    $472 = ((($dst$3)) + 4|0);
    $473 = (($src$3) + ($12)|0);
    $dst$4 = $472;$n$3 = $n$2;$src$4 = $473;
    label = 28;
    continue;
   }
   else if ((label|0) == 28) {
    label = 0;
    switch ($12|0) {
    case 1:  {
     $474 = HEAP8[$src$4>>0]|0;
     $475 = $474&255;
     $476 = HEAP32[$20>>2]|0;
     $477 = $475 & $476;
     $478 = HEAP32[$21>>2]|0;
     $479 = $478 & 255;
     $480 = $477 >>> $479;
     $481 = HEAP32[$22>>2]|0;
     $482 = $481 & 255;
     $483 = (125924 + ($482<<2)|0);
     $484 = HEAP32[$483>>2]|0;
     $485 = (($484) + ($480)|0);
     $486 = HEAP32[$23>>2]|0;
     $487 = $486 & $475;
     $488 = $478 >>> 8;
     $489 = $488 & 255;
     $490 = $487 >>> $489;
     $491 = $481 >>> 8;
     $492 = $491 & 255;
     $493 = (125924 + ($492<<2)|0);
     $494 = HEAP32[$493>>2]|0;
     $495 = (($494) + ($490)|0);
     $496 = HEAP32[$24>>2]|0;
     $497 = $496 & $475;
     $498 = $478 >>> 16;
     $499 = $498 & 255;
     $500 = $497 >>> $499;
     $501 = $481 >>> 16;
     $502 = $501 & 255;
     $503 = (125924 + ($502<<2)|0);
     $504 = HEAP32[$503>>2]|0;
     $505 = (($504) + ($500)|0);
     $506 = HEAP32[$25>>2]|0;
     $507 = $506 & $475;
     $508 = $478 >>> 24;
     $509 = $507 >>> $508;
     $510 = $481 >>> 24;
     $511 = (125924 + ($510<<2)|0);
     $512 = HEAP32[$511>>2]|0;
     $513 = (($512) + ($509)|0);
     $514 = HEAP8[$513>>0]|0;
     $sA$3$shrunk = $514;$sB$3$shrunk$in = $505;$sG$3$shrunk$in = $495;$sR$3$shrunk$in = $485;
     label = 33;
     break;
    }
    case 2:  {
     $515 = HEAP16[$src$4>>1]|0;
     $516 = $515&65535;
     $517 = HEAP32[$20>>2]|0;
     $518 = $516 & $517;
     $519 = HEAP32[$21>>2]|0;
     $520 = $519 & 255;
     $521 = $518 >>> $520;
     $522 = HEAP32[$22>>2]|0;
     $523 = $522 & 255;
     $524 = (125924 + ($523<<2)|0);
     $525 = HEAP32[$524>>2]|0;
     $526 = (($525) + ($521)|0);
     $527 = HEAP32[$23>>2]|0;
     $528 = $527 & $516;
     $529 = $519 >>> 8;
     $530 = $529 & 255;
     $531 = $528 >>> $530;
     $532 = $522 >>> 8;
     $533 = $532 & 255;
     $534 = (125924 + ($533<<2)|0);
     $535 = HEAP32[$534>>2]|0;
     $536 = (($535) + ($531)|0);
     $537 = HEAP32[$24>>2]|0;
     $538 = $537 & $516;
     $539 = $519 >>> 16;
     $540 = $539 & 255;
     $541 = $538 >>> $540;
     $542 = $522 >>> 16;
     $543 = $542 & 255;
     $544 = (125924 + ($543<<2)|0);
     $545 = HEAP32[$544>>2]|0;
     $546 = (($545) + ($541)|0);
     $547 = HEAP32[$25>>2]|0;
     $548 = $547 & $516;
     $549 = $519 >>> 24;
     $550 = $548 >>> $549;
     $551 = $522 >>> 24;
     $552 = (125924 + ($551<<2)|0);
     $553 = HEAP32[$552>>2]|0;
     $554 = (($553) + ($550)|0);
     $555 = HEAP8[$554>>0]|0;
     $sA$3$shrunk = $555;$sB$3$shrunk$in = $546;$sG$3$shrunk$in = $536;$sR$3$shrunk$in = $526;
     label = 33;
     break;
    }
    case 3:  {
     $556 = HEAP32[$21>>2]|0;
     $div17 = $556 >>> 3;
     $557 = $div17 & 31;
     $558 = (($src$4) + ($557)|0);
     $559 = $556 >>> 11;
     $div18 = $559 & 31;
     $560 = (($src$4) + ($div18)|0);
     $561 = $556 >>> 19;
     $562 = $561 & 31;
     $563 = (($src$4) + ($562)|0);
     $sA$3$shrunk = -1;$sB$3$shrunk$in = $563;$sG$3$shrunk$in = $560;$sR$3$shrunk$in = $558;
     label = 33;
     break;
    }
    case 4:  {
     $564 = HEAP32[$src$4>>2]|0;
     $565 = HEAP32[$20>>2]|0;
     $566 = $565 & $564;
     $567 = HEAP32[$21>>2]|0;
     $568 = $567 & 255;
     $569 = $566 >>> $568;
     $570 = HEAP32[$22>>2]|0;
     $571 = $570 & 255;
     $572 = (125924 + ($571<<2)|0);
     $573 = HEAP32[$572>>2]|0;
     $574 = (($573) + ($569)|0);
     $575 = HEAP32[$23>>2]|0;
     $576 = $575 & $564;
     $577 = $567 >>> 8;
     $578 = $577 & 255;
     $579 = $576 >>> $578;
     $580 = $570 >>> 8;
     $581 = $580 & 255;
     $582 = (125924 + ($581<<2)|0);
     $583 = HEAP32[$582>>2]|0;
     $584 = (($583) + ($579)|0);
     $585 = HEAP32[$24>>2]|0;
     $586 = $585 & $564;
     $587 = $567 >>> 16;
     $588 = $587 & 255;
     $589 = $586 >>> $588;
     $590 = $570 >>> 16;
     $591 = $590 & 255;
     $592 = (125924 + ($591<<2)|0);
     $593 = HEAP32[$592>>2]|0;
     $594 = (($593) + ($589)|0);
     $595 = HEAP32[$25>>2]|0;
     $596 = $595 & $564;
     $597 = $567 >>> 24;
     $598 = $596 >>> $597;
     $599 = $570 >>> 24;
     $600 = (125924 + ($599<<2)|0);
     $601 = HEAP32[$600>>2]|0;
     $602 = (($601) + ($598)|0);
     $603 = HEAP8[$602>>0]|0;
     $sA$3$shrunk = $603;$sB$3$shrunk$in = $594;$sG$3$shrunk$in = $584;$sR$3$shrunk$in = $574;
     label = 33;
     break;
    }
    default: {
     $618 = 0;$sA$3$shrunk60 = 0;$sB$3$shrunk59 = 0;$sB$361 = 0;$sG$3$shrunk58 = 0;$sG$362 = 0;
    }
    }
    if ((label|0) == 33) {
     label = 0;
     $sB$3$shrunk = HEAP8[$sB$3$shrunk$in>>0]|0;
     $sG$3$shrunk = HEAP8[$sG$3$shrunk$in>>0]|0;
     $sR$3$shrunk = HEAP8[$sR$3$shrunk$in>>0]|0;
     $sB$3 = $sB$3$shrunk&255;
     $sG$3 = $sG$3$shrunk&255;
     $604 = ($sR$3$shrunk<<24>>24)==(0);
     if ($604) {
      $618 = 0;$sA$3$shrunk60 = $sA$3$shrunk;$sB$3$shrunk59 = $sB$3$shrunk;$sB$361 = $sB$3;$sG$3$shrunk58 = $sG$3$shrunk;$sG$362 = $sG$3;
     } else {
      $sR$3 = $sR$3$shrunk&255;
      $605 = $sR$3 << 22;
      $606 = $605 | 3145728;
      $618 = $606;$sA$3$shrunk60 = $sA$3$shrunk;$sB$3$shrunk59 = $sB$3$shrunk;$sB$361 = $sB$3;$sG$3$shrunk58 = $sG$3$shrunk;$sG$362 = $sG$3;
     }
    }
    $607 = ($sG$3$shrunk58<<24>>24)==(0);
    $608 = $sG$362 << 12;
    $609 = $608 | 3072;
    $610 = $607 ? 0 : $609;
    $611 = ($sB$3$shrunk59<<24>>24)==(0);
    $612 = $sB$361 << 2;
    $613 = $612 | 3;
    $614 = $611 ? 0 : $613;
    $div4 = (($sA$3$shrunk60&255) / 85)&-1;
    $615 = $div4&255;
    $616 = $615 << 30;
    $617 = $616 | $618;
    $619 = $617 | $614;
    $620 = $619 | $610;
    HEAP32[$dst$4>>2] = $620;
    $621 = ((($dst$4)) + 4|0);
    $622 = (($src$4) + ($12)|0);
    $dst$5 = $621;$n$4 = $n$3;$src$5 = $622;
    label = 36;
    continue;
   }
   else if ((label|0) == 36) {
    label = 0;
    switch ($12|0) {
    case 1:  {
     $623 = HEAP8[$src$5>>0]|0;
     $624 = $623&255;
     $625 = HEAP32[$20>>2]|0;
     $626 = $624 & $625;
     $627 = HEAP32[$21>>2]|0;
     $628 = $627 & 255;
     $629 = $626 >>> $628;
     $630 = HEAP32[$22>>2]|0;
     $631 = $630 & 255;
     $632 = (125924 + ($631<<2)|0);
     $633 = HEAP32[$632>>2]|0;
     $634 = (($633) + ($629)|0);
     $635 = HEAP32[$23>>2]|0;
     $636 = $635 & $624;
     $637 = $627 >>> 8;
     $638 = $637 & 255;
     $639 = $636 >>> $638;
     $640 = $630 >>> 8;
     $641 = $640 & 255;
     $642 = (125924 + ($641<<2)|0);
     $643 = HEAP32[$642>>2]|0;
     $644 = (($643) + ($639)|0);
     $645 = HEAP32[$24>>2]|0;
     $646 = $645 & $624;
     $647 = $627 >>> 16;
     $648 = $647 & 255;
     $649 = $646 >>> $648;
     $650 = $630 >>> 16;
     $651 = $650 & 255;
     $652 = (125924 + ($651<<2)|0);
     $653 = HEAP32[$652>>2]|0;
     $654 = (($653) + ($649)|0);
     $655 = HEAP32[$25>>2]|0;
     $656 = $655 & $624;
     $657 = $627 >>> 24;
     $658 = $656 >>> $657;
     $659 = $630 >>> 24;
     $660 = (125924 + ($659<<2)|0);
     $661 = HEAP32[$660>>2]|0;
     $662 = (($661) + ($658)|0);
     $663 = HEAP8[$662>>0]|0;
     $sA$4$shrunk = $663;$sB$4$shrunk$in = $654;$sG$4$shrunk$in = $644;$sR$4$shrunk$in = $634;
     label = 41;
     break;
    }
    case 2:  {
     $664 = HEAP16[$src$5>>1]|0;
     $665 = $664&65535;
     $666 = HEAP32[$20>>2]|0;
     $667 = $665 & $666;
     $668 = HEAP32[$21>>2]|0;
     $669 = $668 & 255;
     $670 = $667 >>> $669;
     $671 = HEAP32[$22>>2]|0;
     $672 = $671 & 255;
     $673 = (125924 + ($672<<2)|0);
     $674 = HEAP32[$673>>2]|0;
     $675 = (($674) + ($670)|0);
     $676 = HEAP32[$23>>2]|0;
     $677 = $676 & $665;
     $678 = $668 >>> 8;
     $679 = $678 & 255;
     $680 = $677 >>> $679;
     $681 = $671 >>> 8;
     $682 = $681 & 255;
     $683 = (125924 + ($682<<2)|0);
     $684 = HEAP32[$683>>2]|0;
     $685 = (($684) + ($680)|0);
     $686 = HEAP32[$24>>2]|0;
     $687 = $686 & $665;
     $688 = $668 >>> 16;
     $689 = $688 & 255;
     $690 = $687 >>> $689;
     $691 = $671 >>> 16;
     $692 = $691 & 255;
     $693 = (125924 + ($692<<2)|0);
     $694 = HEAP32[$693>>2]|0;
     $695 = (($694) + ($690)|0);
     $696 = HEAP32[$25>>2]|0;
     $697 = $696 & $665;
     $698 = $668 >>> 24;
     $699 = $697 >>> $698;
     $700 = $671 >>> 24;
     $701 = (125924 + ($700<<2)|0);
     $702 = HEAP32[$701>>2]|0;
     $703 = (($702) + ($699)|0);
     $704 = HEAP8[$703>>0]|0;
     $sA$4$shrunk = $704;$sB$4$shrunk$in = $695;$sG$4$shrunk$in = $685;$sR$4$shrunk$in = $675;
     label = 41;
     break;
    }
    case 3:  {
     $705 = HEAP32[$21>>2]|0;
     $div14 = $705 >>> 3;
     $706 = $div14 & 31;
     $707 = (($src$5) + ($706)|0);
     $708 = $705 >>> 11;
     $div15 = $708 & 31;
     $709 = (($src$5) + ($div15)|0);
     $710 = $705 >>> 19;
     $711 = $710 & 31;
     $712 = (($src$5) + ($711)|0);
     $sA$4$shrunk = -1;$sB$4$shrunk$in = $712;$sG$4$shrunk$in = $709;$sR$4$shrunk$in = $707;
     label = 41;
     break;
    }
    case 4:  {
     $713 = HEAP32[$src$5>>2]|0;
     $714 = HEAP32[$20>>2]|0;
     $715 = $714 & $713;
     $716 = HEAP32[$21>>2]|0;
     $717 = $716 & 255;
     $718 = $715 >>> $717;
     $719 = HEAP32[$22>>2]|0;
     $720 = $719 & 255;
     $721 = (125924 + ($720<<2)|0);
     $722 = HEAP32[$721>>2]|0;
     $723 = (($722) + ($718)|0);
     $724 = HEAP32[$23>>2]|0;
     $725 = $724 & $713;
     $726 = $716 >>> 8;
     $727 = $726 & 255;
     $728 = $725 >>> $727;
     $729 = $719 >>> 8;
     $730 = $729 & 255;
     $731 = (125924 + ($730<<2)|0);
     $732 = HEAP32[$731>>2]|0;
     $733 = (($732) + ($728)|0);
     $734 = HEAP32[$24>>2]|0;
     $735 = $734 & $713;
     $736 = $716 >>> 16;
     $737 = $736 & 255;
     $738 = $735 >>> $737;
     $739 = $719 >>> 16;
     $740 = $739 & 255;
     $741 = (125924 + ($740<<2)|0);
     $742 = HEAP32[$741>>2]|0;
     $743 = (($742) + ($738)|0);
     $744 = HEAP32[$25>>2]|0;
     $745 = $744 & $713;
     $746 = $716 >>> 24;
     $747 = $745 >>> $746;
     $748 = $719 >>> 24;
     $749 = (125924 + ($748<<2)|0);
     $750 = HEAP32[$749>>2]|0;
     $751 = (($750) + ($747)|0);
     $752 = HEAP8[$751>>0]|0;
     $sA$4$shrunk = $752;$sB$4$shrunk$in = $743;$sG$4$shrunk$in = $733;$sR$4$shrunk$in = $723;
     label = 41;
     break;
    }
    default: {
     $767 = 0;$sA$4$shrunk68 = 0;$sB$4$shrunk67 = 0;$sB$469 = 0;$sG$4$shrunk66 = 0;$sG$470 = 0;
    }
    }
    if ((label|0) == 41) {
     label = 0;
     $sB$4$shrunk = HEAP8[$sB$4$shrunk$in>>0]|0;
     $sG$4$shrunk = HEAP8[$sG$4$shrunk$in>>0]|0;
     $sR$4$shrunk = HEAP8[$sR$4$shrunk$in>>0]|0;
     $sB$4 = $sB$4$shrunk&255;
     $sG$4 = $sG$4$shrunk&255;
     $753 = ($sR$4$shrunk<<24>>24)==(0);
     if ($753) {
      $767 = 0;$sA$4$shrunk68 = $sA$4$shrunk;$sB$4$shrunk67 = $sB$4$shrunk;$sB$469 = $sB$4;$sG$4$shrunk66 = $sG$4$shrunk;$sG$470 = $sG$4;
     } else {
      $sR$4 = $sR$4$shrunk&255;
      $754 = $sR$4 << 22;
      $755 = $754 | 3145728;
      $767 = $755;$sA$4$shrunk68 = $sA$4$shrunk;$sB$4$shrunk67 = $sB$4$shrunk;$sB$469 = $sB$4;$sG$4$shrunk66 = $sG$4$shrunk;$sG$470 = $sG$4;
     }
    }
    $756 = ($sG$4$shrunk66<<24>>24)==(0);
    $757 = $sG$470 << 12;
    $758 = $757 | 3072;
    $759 = $756 ? 0 : $758;
    $760 = ($sB$4$shrunk67<<24>>24)==(0);
    $761 = $sB$469 << 2;
    $762 = $761 | 3;
    $763 = $760 ? 0 : $762;
    $div5 = (($sA$4$shrunk68&255) / 85)&-1;
    $764 = $div5&255;
    $765 = $764 << 30;
    $766 = $765 | $767;
    $768 = $766 | $763;
    $769 = $768 | $759;
    HEAP32[$dst$5>>2] = $769;
    $770 = ((($dst$5)) + 4|0);
    $771 = (($src$5) + ($12)|0);
    $dst$6 = $770;$n$5 = $n$4;$src$6 = $771;
    label = 44;
    continue;
   }
   else if ((label|0) == 44) {
    label = 0;
    switch ($12|0) {
    case 1:  {
     $772 = HEAP8[$src$6>>0]|0;
     $773 = $772&255;
     $774 = HEAP32[$20>>2]|0;
     $775 = $773 & $774;
     $776 = HEAP32[$21>>2]|0;
     $777 = $776 & 255;
     $778 = $775 >>> $777;
     $779 = HEAP32[$22>>2]|0;
     $780 = $779 & 255;
     $781 = (125924 + ($780<<2)|0);
     $782 = HEAP32[$781>>2]|0;
     $783 = (($782) + ($778)|0);
     $784 = HEAP32[$23>>2]|0;
     $785 = $784 & $773;
     $786 = $776 >>> 8;
     $787 = $786 & 255;
     $788 = $785 >>> $787;
     $789 = $779 >>> 8;
     $790 = $789 & 255;
     $791 = (125924 + ($790<<2)|0);
     $792 = HEAP32[$791>>2]|0;
     $793 = (($792) + ($788)|0);
     $794 = HEAP32[$24>>2]|0;
     $795 = $794 & $773;
     $796 = $776 >>> 16;
     $797 = $796 & 255;
     $798 = $795 >>> $797;
     $799 = $779 >>> 16;
     $800 = $799 & 255;
     $801 = (125924 + ($800<<2)|0);
     $802 = HEAP32[$801>>2]|0;
     $803 = (($802) + ($798)|0);
     $804 = HEAP32[$25>>2]|0;
     $805 = $804 & $773;
     $806 = $776 >>> 24;
     $807 = $805 >>> $806;
     $808 = $779 >>> 24;
     $809 = (125924 + ($808<<2)|0);
     $810 = HEAP32[$809>>2]|0;
     $811 = (($810) + ($807)|0);
     $812 = HEAP8[$811>>0]|0;
     $sA$5$shrunk = $812;$sB$5$shrunk$in = $803;$sG$5$shrunk$in = $793;$sR$5$shrunk$in = $783;
     label = 49;
     break;
    }
    case 2:  {
     $813 = HEAP16[$src$6>>1]|0;
     $814 = $813&65535;
     $815 = HEAP32[$20>>2]|0;
     $816 = $814 & $815;
     $817 = HEAP32[$21>>2]|0;
     $818 = $817 & 255;
     $819 = $816 >>> $818;
     $820 = HEAP32[$22>>2]|0;
     $821 = $820 & 255;
     $822 = (125924 + ($821<<2)|0);
     $823 = HEAP32[$822>>2]|0;
     $824 = (($823) + ($819)|0);
     $825 = HEAP32[$23>>2]|0;
     $826 = $825 & $814;
     $827 = $817 >>> 8;
     $828 = $827 & 255;
     $829 = $826 >>> $828;
     $830 = $820 >>> 8;
     $831 = $830 & 255;
     $832 = (125924 + ($831<<2)|0);
     $833 = HEAP32[$832>>2]|0;
     $834 = (($833) + ($829)|0);
     $835 = HEAP32[$24>>2]|0;
     $836 = $835 & $814;
     $837 = $817 >>> 16;
     $838 = $837 & 255;
     $839 = $836 >>> $838;
     $840 = $820 >>> 16;
     $841 = $840 & 255;
     $842 = (125924 + ($841<<2)|0);
     $843 = HEAP32[$842>>2]|0;
     $844 = (($843) + ($839)|0);
     $845 = HEAP32[$25>>2]|0;
     $846 = $845 & $814;
     $847 = $817 >>> 24;
     $848 = $846 >>> $847;
     $849 = $820 >>> 24;
     $850 = (125924 + ($849<<2)|0);
     $851 = HEAP32[$850>>2]|0;
     $852 = (($851) + ($848)|0);
     $853 = HEAP8[$852>>0]|0;
     $sA$5$shrunk = $853;$sB$5$shrunk$in = $844;$sG$5$shrunk$in = $834;$sR$5$shrunk$in = $824;
     label = 49;
     break;
    }
    case 3:  {
     $854 = HEAP32[$21>>2]|0;
     $div11 = $854 >>> 3;
     $855 = $div11 & 31;
     $856 = (($src$6) + ($855)|0);
     $857 = $854 >>> 11;
     $div12 = $857 & 31;
     $858 = (($src$6) + ($div12)|0);
     $859 = $854 >>> 19;
     $860 = $859 & 31;
     $861 = (($src$6) + ($860)|0);
     $sA$5$shrunk = -1;$sB$5$shrunk$in = $861;$sG$5$shrunk$in = $858;$sR$5$shrunk$in = $856;
     label = 49;
     break;
    }
    case 4:  {
     $862 = HEAP32[$src$6>>2]|0;
     $863 = HEAP32[$20>>2]|0;
     $864 = $863 & $862;
     $865 = HEAP32[$21>>2]|0;
     $866 = $865 & 255;
     $867 = $864 >>> $866;
     $868 = HEAP32[$22>>2]|0;
     $869 = $868 & 255;
     $870 = (125924 + ($869<<2)|0);
     $871 = HEAP32[$870>>2]|0;
     $872 = (($871) + ($867)|0);
     $873 = HEAP32[$23>>2]|0;
     $874 = $873 & $862;
     $875 = $865 >>> 8;
     $876 = $875 & 255;
     $877 = $874 >>> $876;
     $878 = $868 >>> 8;
     $879 = $878 & 255;
     $880 = (125924 + ($879<<2)|0);
     $881 = HEAP32[$880>>2]|0;
     $882 = (($881) + ($877)|0);
     $883 = HEAP32[$24>>2]|0;
     $884 = $883 & $862;
     $885 = $865 >>> 16;
     $886 = $885 & 255;
     $887 = $884 >>> $886;
     $888 = $868 >>> 16;
     $889 = $888 & 255;
     $890 = (125924 + ($889<<2)|0);
     $891 = HEAP32[$890>>2]|0;
     $892 = (($891) + ($887)|0);
     $893 = HEAP32[$25>>2]|0;
     $894 = $893 & $862;
     $895 = $865 >>> 24;
     $896 = $894 >>> $895;
     $897 = $868 >>> 24;
     $898 = (125924 + ($897<<2)|0);
     $899 = HEAP32[$898>>2]|0;
     $900 = (($899) + ($896)|0);
     $901 = HEAP8[$900>>0]|0;
     $sA$5$shrunk = $901;$sB$5$shrunk$in = $892;$sG$5$shrunk$in = $882;$sR$5$shrunk$in = $872;
     label = 49;
     break;
    }
    default: {
     $916 = 0;$sA$5$shrunk76 = 0;$sB$5$shrunk75 = 0;$sB$577 = 0;$sG$5$shrunk74 = 0;$sG$578 = 0;
    }
    }
    if ((label|0) == 49) {
     label = 0;
     $sB$5$shrunk = HEAP8[$sB$5$shrunk$in>>0]|0;
     $sG$5$shrunk = HEAP8[$sG$5$shrunk$in>>0]|0;
     $sR$5$shrunk = HEAP8[$sR$5$shrunk$in>>0]|0;
     $sB$5 = $sB$5$shrunk&255;
     $sG$5 = $sG$5$shrunk&255;
     $902 = ($sR$5$shrunk<<24>>24)==(0);
     if ($902) {
      $916 = 0;$sA$5$shrunk76 = $sA$5$shrunk;$sB$5$shrunk75 = $sB$5$shrunk;$sB$577 = $sB$5;$sG$5$shrunk74 = $sG$5$shrunk;$sG$578 = $sG$5;
     } else {
      $sR$5 = $sR$5$shrunk&255;
      $903 = $sR$5 << 22;
      $904 = $903 | 3145728;
      $916 = $904;$sA$5$shrunk76 = $sA$5$shrunk;$sB$5$shrunk75 = $sB$5$shrunk;$sB$577 = $sB$5;$sG$5$shrunk74 = $sG$5$shrunk;$sG$578 = $sG$5;
     }
    }
    $905 = ($sG$5$shrunk74<<24>>24)==(0);
    $906 = $sG$578 << 12;
    $907 = $906 | 3072;
    $908 = $905 ? 0 : $907;
    $909 = ($sB$5$shrunk75<<24>>24)==(0);
    $910 = $sB$577 << 2;
    $911 = $910 | 3;
    $912 = $909 ? 0 : $911;
    $div6 = (($sA$5$shrunk76&255) / 85)&-1;
    $913 = $div6&255;
    $914 = $913 << 30;
    $915 = $914 | $916;
    $917 = $915 | $912;
    $918 = $917 | $908;
    HEAP32[$dst$6>>2] = $918;
    $919 = ((($dst$6)) + 4|0);
    $920 = (($src$6) + ($12)|0);
    $dst$7 = $919;$n$6 = $n$5;$src$7 = $920;
    label = 52;
    continue;
   }
   else if ((label|0) == 52) {
    label = 0;
    switch ($12|0) {
    case 1:  {
     $921 = HEAP8[$src$7>>0]|0;
     $922 = $921&255;
     $923 = HEAP32[$20>>2]|0;
     $924 = $922 & $923;
     $925 = HEAP32[$21>>2]|0;
     $926 = $925 & 255;
     $927 = $924 >>> $926;
     $928 = HEAP32[$22>>2]|0;
     $929 = $928 & 255;
     $930 = (125924 + ($929<<2)|0);
     $931 = HEAP32[$930>>2]|0;
     $932 = (($931) + ($927)|0);
     $933 = HEAP32[$23>>2]|0;
     $934 = $933 & $922;
     $935 = $925 >>> 8;
     $936 = $935 & 255;
     $937 = $934 >>> $936;
     $938 = $928 >>> 8;
     $939 = $938 & 255;
     $940 = (125924 + ($939<<2)|0);
     $941 = HEAP32[$940>>2]|0;
     $942 = (($941) + ($937)|0);
     $943 = HEAP32[$24>>2]|0;
     $944 = $943 & $922;
     $945 = $925 >>> 16;
     $946 = $945 & 255;
     $947 = $944 >>> $946;
     $948 = $928 >>> 16;
     $949 = $948 & 255;
     $950 = (125924 + ($949<<2)|0);
     $951 = HEAP32[$950>>2]|0;
     $952 = (($951) + ($947)|0);
     $953 = HEAP32[$25>>2]|0;
     $954 = $953 & $922;
     $955 = $925 >>> 24;
     $956 = $954 >>> $955;
     $957 = $928 >>> 24;
     $958 = (125924 + ($957<<2)|0);
     $959 = HEAP32[$958>>2]|0;
     $960 = (($959) + ($956)|0);
     $961 = HEAP8[$960>>0]|0;
     $sA$6$shrunk = $961;$sB$6$shrunk$in = $952;$sG$6$shrunk$in = $942;$sR$6$shrunk$in = $932;
     label = 57;
     break;
    }
    case 2:  {
     $962 = HEAP16[$src$7>>1]|0;
     $963 = $962&65535;
     $964 = HEAP32[$20>>2]|0;
     $965 = $963 & $964;
     $966 = HEAP32[$21>>2]|0;
     $967 = $966 & 255;
     $968 = $965 >>> $967;
     $969 = HEAP32[$22>>2]|0;
     $970 = $969 & 255;
     $971 = (125924 + ($970<<2)|0);
     $972 = HEAP32[$971>>2]|0;
     $973 = (($972) + ($968)|0);
     $974 = HEAP32[$23>>2]|0;
     $975 = $974 & $963;
     $976 = $966 >>> 8;
     $977 = $976 & 255;
     $978 = $975 >>> $977;
     $979 = $969 >>> 8;
     $980 = $979 & 255;
     $981 = (125924 + ($980<<2)|0);
     $982 = HEAP32[$981>>2]|0;
     $983 = (($982) + ($978)|0);
     $984 = HEAP32[$24>>2]|0;
     $985 = $984 & $963;
     $986 = $966 >>> 16;
     $987 = $986 & 255;
     $988 = $985 >>> $987;
     $989 = $969 >>> 16;
     $990 = $989 & 255;
     $991 = (125924 + ($990<<2)|0);
     $992 = HEAP32[$991>>2]|0;
     $993 = (($992) + ($988)|0);
     $994 = HEAP32[$25>>2]|0;
     $995 = $994 & $963;
     $996 = $966 >>> 24;
     $997 = $995 >>> $996;
     $998 = $969 >>> 24;
     $999 = (125924 + ($998<<2)|0);
     $1000 = HEAP32[$999>>2]|0;
     $1001 = (($1000) + ($997)|0);
     $1002 = HEAP8[$1001>>0]|0;
     $sA$6$shrunk = $1002;$sB$6$shrunk$in = $993;$sG$6$shrunk$in = $983;$sR$6$shrunk$in = $973;
     label = 57;
     break;
    }
    case 3:  {
     $1003 = HEAP32[$21>>2]|0;
     $div8 = $1003 >>> 3;
     $1004 = $div8 & 31;
     $1005 = (($src$7) + ($1004)|0);
     $1006 = $1003 >>> 11;
     $div9 = $1006 & 31;
     $1007 = (($src$7) + ($div9)|0);
     $1008 = $1003 >>> 19;
     $1009 = $1008 & 31;
     $1010 = (($src$7) + ($1009)|0);
     $sA$6$shrunk = -1;$sB$6$shrunk$in = $1010;$sG$6$shrunk$in = $1007;$sR$6$shrunk$in = $1005;
     label = 57;
     break;
    }
    case 4:  {
     $1011 = HEAP32[$src$7>>2]|0;
     $1012 = HEAP32[$20>>2]|0;
     $1013 = $1012 & $1011;
     $1014 = HEAP32[$21>>2]|0;
     $1015 = $1014 & 255;
     $1016 = $1013 >>> $1015;
     $1017 = HEAP32[$22>>2]|0;
     $1018 = $1017 & 255;
     $1019 = (125924 + ($1018<<2)|0);
     $1020 = HEAP32[$1019>>2]|0;
     $1021 = (($1020) + ($1016)|0);
     $1022 = HEAP32[$23>>2]|0;
     $1023 = $1022 & $1011;
     $1024 = $1014 >>> 8;
     $1025 = $1024 & 255;
     $1026 = $1023 >>> $1025;
     $1027 = $1017 >>> 8;
     $1028 = $1027 & 255;
     $1029 = (125924 + ($1028<<2)|0);
     $1030 = HEAP32[$1029>>2]|0;
     $1031 = (($1030) + ($1026)|0);
     $1032 = HEAP32[$24>>2]|0;
     $1033 = $1032 & $1011;
     $1034 = $1014 >>> 16;
     $1035 = $1034 & 255;
     $1036 = $1033 >>> $1035;
     $1037 = $1017 >>> 16;
     $1038 = $1037 & 255;
     $1039 = (125924 + ($1038<<2)|0);
     $1040 = HEAP32[$1039>>2]|0;
     $1041 = (($1040) + ($1036)|0);
     $1042 = HEAP32[$25>>2]|0;
     $1043 = $1042 & $1011;
     $1044 = $1014 >>> 24;
     $1045 = $1043 >>> $1044;
     $1046 = $1017 >>> 24;
     $1047 = (125924 + ($1046<<2)|0);
     $1048 = HEAP32[$1047>>2]|0;
     $1049 = (($1048) + ($1045)|0);
     $1050 = HEAP8[$1049>>0]|0;
     $sA$6$shrunk = $1050;$sB$6$shrunk$in = $1041;$sG$6$shrunk$in = $1031;$sR$6$shrunk$in = $1021;
     label = 57;
     break;
    }
    default: {
     $1065 = 0;$sA$6$shrunk84 = 0;$sB$6$shrunk83 = 0;$sB$685 = 0;$sG$6$shrunk82 = 0;$sG$686 = 0;
    }
    }
    if ((label|0) == 57) {
     label = 0;
     $sB$6$shrunk = HEAP8[$sB$6$shrunk$in>>0]|0;
     $sG$6$shrunk = HEAP8[$sG$6$shrunk$in>>0]|0;
     $sR$6$shrunk = HEAP8[$sR$6$shrunk$in>>0]|0;
     $sB$6 = $sB$6$shrunk&255;
     $sG$6 = $sG$6$shrunk&255;
     $1051 = ($sR$6$shrunk<<24>>24)==(0);
     if ($1051) {
      $1065 = 0;$sA$6$shrunk84 = $sA$6$shrunk;$sB$6$shrunk83 = $sB$6$shrunk;$sB$685 = $sB$6;$sG$6$shrunk82 = $sG$6$shrunk;$sG$686 = $sG$6;
     } else {
      $sR$6 = $sR$6$shrunk&255;
      $1052 = $sR$6 << 22;
      $1053 = $1052 | 3145728;
      $1065 = $1053;$sA$6$shrunk84 = $sA$6$shrunk;$sB$6$shrunk83 = $sB$6$shrunk;$sB$685 = $sB$6;$sG$6$shrunk82 = $sG$6$shrunk;$sG$686 = $sG$6;
     }
    }
    $1054 = ($sG$6$shrunk82<<24>>24)==(0);
    $1055 = $sG$686 << 12;
    $1056 = $1055 | 3072;
    $1057 = $1054 ? 0 : $1056;
    $1058 = ($sB$6$shrunk83<<24>>24)==(0);
    $1059 = $sB$685 << 2;
    $1060 = $1059 | 3;
    $1061 = $1058 ? 0 : $1060;
    $div7 = (($sA$6$shrunk84&255) / 85)&-1;
    $1062 = $div7&255;
    $1063 = $1062 << 30;
    $1064 = $1063 | $1065;
    $1066 = $1064 | $1061;
    $1067 = $1066 | $1057;
    HEAP32[$dst$7>>2] = $1067;
    $1068 = ((($dst$7)) + 4|0);
    $1069 = (($src$7) + ($12)|0);
    $dst$8 = $1068;$n$7 = $n$6;$src$8 = $1069;
    label = 60;
    continue;
   }
   else if ((label|0) == 60) {
    label = 0;
    switch ($12|0) {
    case 1:  {
     $1070 = HEAP8[$src$8>>0]|0;
     $1071 = $1070&255;
     $1072 = HEAP32[$20>>2]|0;
     $1073 = $1071 & $1072;
     $1074 = HEAP32[$21>>2]|0;
     $1075 = $1074 & 255;
     $1076 = $1073 >>> $1075;
     $1077 = HEAP32[$22>>2]|0;
     $1078 = $1077 & 255;
     $1079 = (125924 + ($1078<<2)|0);
     $1080 = HEAP32[$1079>>2]|0;
     $1081 = (($1080) + ($1076)|0);
     $1082 = HEAP32[$23>>2]|0;
     $1083 = $1082 & $1071;
     $1084 = $1074 >>> 8;
     $1085 = $1084 & 255;
     $1086 = $1083 >>> $1085;
     $1087 = $1077 >>> 8;
     $1088 = $1087 & 255;
     $1089 = (125924 + ($1088<<2)|0);
     $1090 = HEAP32[$1089>>2]|0;
     $1091 = (($1090) + ($1086)|0);
     $1092 = HEAP32[$24>>2]|0;
     $1093 = $1092 & $1071;
     $1094 = $1074 >>> 16;
     $1095 = $1094 & 255;
     $1096 = $1093 >>> $1095;
     $1097 = $1077 >>> 16;
     $1098 = $1097 & 255;
     $1099 = (125924 + ($1098<<2)|0);
     $1100 = HEAP32[$1099>>2]|0;
     $1101 = (($1100) + ($1096)|0);
     $1102 = HEAP32[$25>>2]|0;
     $1103 = $1102 & $1071;
     $1104 = $1074 >>> 24;
     $1105 = $1103 >>> $1104;
     $1106 = $1077 >>> 24;
     $1107 = (125924 + ($1106<<2)|0);
     $1108 = HEAP32[$1107>>2]|0;
     $1109 = (($1108) + ($1105)|0);
     $1110 = HEAP8[$1109>>0]|0;
     $sA$7$shrunk = $1110;$sB$7$shrunk$in = $1101;$sG$7$shrunk$in = $1091;$sR$7$shrunk$in = $1081;
     label = 65;
     break;
    }
    case 2:  {
     $1111 = HEAP16[$src$8>>1]|0;
     $1112 = $1111&65535;
     $1113 = HEAP32[$20>>2]|0;
     $1114 = $1112 & $1113;
     $1115 = HEAP32[$21>>2]|0;
     $1116 = $1115 & 255;
     $1117 = $1114 >>> $1116;
     $1118 = HEAP32[$22>>2]|0;
     $1119 = $1118 & 255;
     $1120 = (125924 + ($1119<<2)|0);
     $1121 = HEAP32[$1120>>2]|0;
     $1122 = (($1121) + ($1117)|0);
     $1123 = HEAP32[$23>>2]|0;
     $1124 = $1123 & $1112;
     $1125 = $1115 >>> 8;
     $1126 = $1125 & 255;
     $1127 = $1124 >>> $1126;
     $1128 = $1118 >>> 8;
     $1129 = $1128 & 255;
     $1130 = (125924 + ($1129<<2)|0);
     $1131 = HEAP32[$1130>>2]|0;
     $1132 = (($1131) + ($1127)|0);
     $1133 = HEAP32[$24>>2]|0;
     $1134 = $1133 & $1112;
     $1135 = $1115 >>> 16;
     $1136 = $1135 & 255;
     $1137 = $1134 >>> $1136;
     $1138 = $1118 >>> 16;
     $1139 = $1138 & 255;
     $1140 = (125924 + ($1139<<2)|0);
     $1141 = HEAP32[$1140>>2]|0;
     $1142 = (($1141) + ($1137)|0);
     $1143 = HEAP32[$25>>2]|0;
     $1144 = $1143 & $1112;
     $1145 = $1115 >>> 24;
     $1146 = $1144 >>> $1145;
     $1147 = $1118 >>> 24;
     $1148 = (125924 + ($1147<<2)|0);
     $1149 = HEAP32[$1148>>2]|0;
     $1150 = (($1149) + ($1146)|0);
     $1151 = HEAP8[$1150>>0]|0;
     $sA$7$shrunk = $1151;$sB$7$shrunk$in = $1142;$sG$7$shrunk$in = $1132;$sR$7$shrunk$in = $1122;
     label = 65;
     break;
    }
    case 3:  {
     $1152 = HEAP32[$21>>2]|0;
     $div29 = $1152 >>> 3;
     $1153 = $div29 & 31;
     $1154 = (($src$8) + ($1153)|0);
     $1155 = $1152 >>> 11;
     $div30 = $1155 & 31;
     $1156 = (($src$8) + ($div30)|0);
     $1157 = $1152 >>> 19;
     $1158 = $1157 & 31;
     $1159 = (($src$8) + ($1158)|0);
     $sA$7$shrunk = -1;$sB$7$shrunk$in = $1159;$sG$7$shrunk$in = $1156;$sR$7$shrunk$in = $1154;
     label = 65;
     break;
    }
    case 4:  {
     $1160 = HEAP32[$src$8>>2]|0;
     $1161 = HEAP32[$20>>2]|0;
     $1162 = $1161 & $1160;
     $1163 = HEAP32[$21>>2]|0;
     $1164 = $1163 & 255;
     $1165 = $1162 >>> $1164;
     $1166 = HEAP32[$22>>2]|0;
     $1167 = $1166 & 255;
     $1168 = (125924 + ($1167<<2)|0);
     $1169 = HEAP32[$1168>>2]|0;
     $1170 = (($1169) + ($1165)|0);
     $1171 = HEAP32[$23>>2]|0;
     $1172 = $1171 & $1160;
     $1173 = $1163 >>> 8;
     $1174 = $1173 & 255;
     $1175 = $1172 >>> $1174;
     $1176 = $1166 >>> 8;
     $1177 = $1176 & 255;
     $1178 = (125924 + ($1177<<2)|0);
     $1179 = HEAP32[$1178>>2]|0;
     $1180 = (($1179) + ($1175)|0);
     $1181 = HEAP32[$24>>2]|0;
     $1182 = $1181 & $1160;
     $1183 = $1163 >>> 16;
     $1184 = $1183 & 255;
     $1185 = $1182 >>> $1184;
     $1186 = $1166 >>> 16;
     $1187 = $1186 & 255;
     $1188 = (125924 + ($1187<<2)|0);
     $1189 = HEAP32[$1188>>2]|0;
     $1190 = (($1189) + ($1185)|0);
     $1191 = HEAP32[$25>>2]|0;
     $1192 = $1191 & $1160;
     $1193 = $1163 >>> 24;
     $1194 = $1192 >>> $1193;
     $1195 = $1166 >>> 24;
     $1196 = (125924 + ($1195<<2)|0);
     $1197 = HEAP32[$1196>>2]|0;
     $1198 = (($1197) + ($1194)|0);
     $1199 = HEAP8[$1198>>0]|0;
     $sA$7$shrunk = $1199;$sB$7$shrunk$in = $1190;$sG$7$shrunk$in = $1180;$sR$7$shrunk$in = $1170;
     label = 65;
     break;
    }
    default: {
     $1214 = 0;$sA$7$shrunk92 = 0;$sB$7$shrunk91 = 0;$sB$793 = 0;$sG$7$shrunk90 = 0;$sG$794 = 0;
    }
    }
    if ((label|0) == 65) {
     label = 0;
     $sB$7$shrunk = HEAP8[$sB$7$shrunk$in>>0]|0;
     $sG$7$shrunk = HEAP8[$sG$7$shrunk$in>>0]|0;
     $sR$7$shrunk = HEAP8[$sR$7$shrunk$in>>0]|0;
     $sB$7 = $sB$7$shrunk&255;
     $sG$7 = $sG$7$shrunk&255;
     $1200 = ($sR$7$shrunk<<24>>24)==(0);
     if ($1200) {
      $1214 = 0;$sA$7$shrunk92 = $sA$7$shrunk;$sB$7$shrunk91 = $sB$7$shrunk;$sB$793 = $sB$7;$sG$7$shrunk90 = $sG$7$shrunk;$sG$794 = $sG$7;
     } else {
      $sR$7 = $sR$7$shrunk&255;
      $1201 = $sR$7 << 22;
      $1202 = $1201 | 3145728;
      $1214 = $1202;$sA$7$shrunk92 = $sA$7$shrunk;$sB$7$shrunk91 = $sB$7$shrunk;$sB$793 = $sB$7;$sG$7$shrunk90 = $sG$7$shrunk;$sG$794 = $sG$7;
     }
    }
    $1203 = ($sG$7$shrunk90<<24>>24)==(0);
    $1204 = $sG$794 << 12;
    $1205 = $1204 | 3072;
    $1206 = $1203 ? 0 : $1205;
    $1207 = ($sB$7$shrunk91<<24>>24)==(0);
    $1208 = $sB$793 << 2;
    $1209 = $1208 | 3;
    $1210 = $1207 ? 0 : $1209;
    $div = (($sA$7$shrunk92&255) / 85)&-1;
    $1211 = $div&255;
    $1212 = $1211 << 30;
    $1213 = $1212 | $1214;
    $1215 = $1213 | $1210;
    $1216 = $1215 | $1206;
    HEAP32[$dst$8>>2] = $1216;
    $1217 = ((($dst$8)) + 4|0);
    $1218 = (($src$8) + ($12)|0);
    $1219 = (($n$7) + -1)|0;
    $1220 = ($n$7|0)>(1);
    if ($1220) {
     $dst$1 = $1217;$n$0 = $1219;$src$1 = $1218;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $1221 = (($1218) + ($5)|0);
  $1222 = (($1217) + ($7)|0);
  $1223 = ($26|0)==(0);
  if ($1223) {
   label = 70;
   break;
  } else {
   $$in = $26;$dst$095 = $1222;$src$096 = $1221;
  }
 }
 if ((label|0) == 68) {
  // unreachable;
 }
 else if ((label|0) == 70) {
  return;
 }
}
function _Blit4to4MaskAlpha($info) {
 $info = $info|0;
 var $$lcssa = 0, $$lcssa120 = 0, $$lcssa121 = 0, $$lcssa122 = 0, $$lcssa123 = 0, $$lcssa124 = 0, $$lcssa125 = 0, $$lcssa126 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0;
 var $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0;
 var $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0;
 var $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0;
 var $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0;
 var $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$04 = 0, $dst$04$us = 0, $dst$04$us39 = 0, $dst$1$us = 0, $dst$10$us = 0, $dst$11$us12 = 0, $dst$12 = 0, $dst$13 = 0, $dst$14 = 0, $dst$15 = 0, $dst$16 = 0;
 var $dst$17 = 0, $dst$2$us43 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $dst$92 = 0, $dst$92$us = 0, $dst$92$us8 = 0, $n$0$us = 0, $n$1$us45 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $n2$0$us = 0;
 var $n2$1$us14 = 0, $n2$2 = 0, $n2$3 = 0, $n2$4 = 0, $n2$5 = 0, $n2$6 = 0, $n2$7 = 0, $src$05 = 0, $src$05$us = 0, $src$05$us38 = 0, $src$1$us = 0, $src$10$us = 0, $src$11$us13 = 0, $src$12 = 0, $src$13 = 0, $src$14 = 0, $src$15 = 0, $src$16 = 0, $src$17 = 0, $src$2$us44 = 0;
 var $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, $src$93 = 0, $src$93$us = 0, $src$93$us7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$info>>2]|0;
 $5 = ((($info)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($info)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($info)) + 36|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($info)) + 40|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($info)) + 44|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($14)) + 24|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ($16|0)==(0);
 if ($17) {
  $151 = ((($12)) + 12|0);
  $152 = HEAP32[$151>>2]|0;
  $153 = ((($12)) + 16|0);
  $154 = HEAP32[$153>>2]|0;
  $155 = $154 | $152;
  $156 = ((($12)) + 20|0);
  $157 = HEAP32[$156>>2]|0;
  $158 = $155 | $157;
  $159 = (($3) + -1)|0;
  $160 = ($3|0)==(0);
  if ($160) {
   return;
  }
  $161 = (($1) + 7)|0;
  $162 = (($161|0) / 8)&-1;
  $163 = $1 & 7;
  switch ($163|0) {
  case 0:  {
   $201 = $159;$dst$92$us = $8;$src$93$us = $4;
   while(1) {
    $dst$10$us = $dst$92$us;$n2$0$us = $162;$src$10$us = $src$93$us;
    while(1) {
     $164 = HEAP32[$src$10$us>>2]|0;
     $165 = $164 & $158;
     HEAP32[$dst$10$us>>2] = $165;
     $166 = ((($dst$10$us)) + 4|0);
     $167 = ((($src$10$us)) + 4|0);
     $168 = HEAP32[$167>>2]|0;
     $169 = $168 & $158;
     HEAP32[$166>>2] = $169;
     $170 = ((($dst$10$us)) + 8|0);
     $171 = ((($src$10$us)) + 8|0);
     $172 = HEAP32[$171>>2]|0;
     $173 = $172 & $158;
     HEAP32[$170>>2] = $173;
     $174 = ((($dst$10$us)) + 12|0);
     $175 = ((($src$10$us)) + 12|0);
     $176 = HEAP32[$175>>2]|0;
     $177 = $176 & $158;
     HEAP32[$174>>2] = $177;
     $178 = ((($dst$10$us)) + 16|0);
     $179 = ((($src$10$us)) + 16|0);
     $180 = HEAP32[$179>>2]|0;
     $181 = $180 & $158;
     HEAP32[$178>>2] = $181;
     $182 = ((($dst$10$us)) + 20|0);
     $183 = ((($src$10$us)) + 20|0);
     $184 = HEAP32[$183>>2]|0;
     $185 = $184 & $158;
     HEAP32[$182>>2] = $185;
     $186 = ((($dst$10$us)) + 24|0);
     $187 = ((($src$10$us)) + 24|0);
     $188 = HEAP32[$187>>2]|0;
     $189 = $188 & $158;
     HEAP32[$186>>2] = $189;
     $190 = ((($dst$10$us)) + 28|0);
     $191 = ((($src$10$us)) + 28|0);
     $192 = HEAP32[$191>>2]|0;
     $193 = $192 & $158;
     HEAP32[$190>>2] = $193;
     $194 = ((($dst$10$us)) + 32|0);
     $195 = ((($src$10$us)) + 32|0);
     $196 = (($n2$0$us) + -1)|0;
     $197 = ($n2$0$us|0)>(1);
     if ($197) {
      $dst$10$us = $194;$n2$0$us = $196;$src$10$us = $195;
     } else {
      $$lcssa = $194;$$lcssa120 = $195;
      break;
     }
    }
    $198 = (($$lcssa120) + ($6)|0);
    $199 = (($$lcssa) + ($10)|0);
    $200 = (($201) + -1)|0;
    $202 = ($201|0)==(0);
    if ($202) {
     break;
    } else {
     $201 = $200;$dst$92$us = $199;$src$93$us = $198;
    }
   }
   return;
   break;
  }
  case 7:  {
   $240 = $159;$dst$92$us8 = $8;$src$93$us7 = $4;
   while(1) {
    $dst$11$us12 = $dst$92$us8;$n2$1$us14 = $162;$src$11$us13 = $src$93$us7;
    while(1) {
     $210 = HEAP32[$src$11$us13>>2]|0;
     $211 = $210 & $158;
     HEAP32[$dst$11$us12>>2] = $211;
     $212 = ((($dst$11$us12)) + 4|0);
     $213 = ((($src$11$us13)) + 4|0);
     $214 = HEAP32[$213>>2]|0;
     $215 = $214 & $158;
     HEAP32[$212>>2] = $215;
     $216 = ((($dst$11$us12)) + 8|0);
     $217 = ((($src$11$us13)) + 8|0);
     $218 = HEAP32[$217>>2]|0;
     $219 = $218 & $158;
     HEAP32[$216>>2] = $219;
     $220 = ((($dst$11$us12)) + 12|0);
     $221 = ((($src$11$us13)) + 12|0);
     $222 = HEAP32[$221>>2]|0;
     $223 = $222 & $158;
     HEAP32[$220>>2] = $223;
     $224 = ((($dst$11$us12)) + 16|0);
     $225 = ((($src$11$us13)) + 16|0);
     $226 = HEAP32[$225>>2]|0;
     $227 = $226 & $158;
     HEAP32[$224>>2] = $227;
     $228 = ((($dst$11$us12)) + 20|0);
     $229 = ((($src$11$us13)) + 20|0);
     $230 = HEAP32[$229>>2]|0;
     $231 = $230 & $158;
     HEAP32[$228>>2] = $231;
     $232 = ((($dst$11$us12)) + 24|0);
     $233 = ((($src$11$us13)) + 24|0);
     $234 = HEAP32[$233>>2]|0;
     $235 = $234 & $158;
     HEAP32[$232>>2] = $235;
     $207 = ((($dst$11$us12)) + 28|0);
     $205 = ((($src$11$us13)) + 28|0);
     $236 = ($n2$1$us14|0)>(1);
     if (!($236)) {
      $$lcssa121 = $207;$$lcssa122 = $205;
      break;
     }
     $203 = (($n2$1$us14) + -1)|0;
     $204 = HEAP32[$205>>2]|0;
     $206 = $204 & $158;
     HEAP32[$207>>2] = $206;
     $208 = ((($dst$11$us12)) + 32|0);
     $209 = ((($src$11$us13)) + 32|0);
     $dst$11$us12 = $208;$n2$1$us14 = $203;$src$11$us13 = $209;
    }
    $237 = (($$lcssa122) + ($6)|0);
    $238 = (($$lcssa121) + ($10)|0);
    $239 = (($240) + -1)|0;
    $241 = ($240|0)==(0);
    if ($241) {
     break;
    } else {
     $240 = $239;$dst$92$us8 = $238;$src$93$us7 = $237;
    }
   }
   return;
   break;
  }
  default: {
   $279 = $159;$dst$92 = $8;$src$93 = $4;
   L7: while(1) {
    switch ($163|0) {
    case 1:  {
     $dst$17 = $dst$92;$n2$7 = $162;$src$17 = $src$93;
     label = 37;
     break;
    }
    case 2:  {
     $dst$16 = $dst$92;$n2$6 = $162;$src$16 = $src$93;
     label = 36;
     break;
    }
    case 6:  {
     $dst$12 = $dst$92;$n2$2 = $162;$src$12 = $src$93;
     label = 32;
     break;
    }
    case 5:  {
     $dst$13 = $dst$92;$n2$3 = $162;$src$13 = $src$93;
     label = 33;
     break;
    }
    case 4:  {
     $dst$14 = $dst$92;$n2$4 = $162;$src$14 = $src$93;
     label = 34;
     break;
    }
    case 3:  {
     $dst$15 = $dst$92;$n2$5 = $162;$src$15 = $src$93;
     label = 35;
     break;
    }
    default: {
     label = 38;
     break L7;
    }
    }
    while(1) {
     if ((label|0) == 32) {
      label = 0;
      $253 = HEAP32[$src$12>>2]|0;
      $254 = $253 & $158;
      HEAP32[$dst$12>>2] = $254;
      $255 = ((($dst$12)) + 4|0);
      $256 = ((($src$12)) + 4|0);
      $dst$13 = $255;$n2$3 = $n2$2;$src$13 = $256;
      label = 33;
      continue;
     }
     else if ((label|0) == 33) {
      label = 0;
      $257 = HEAP32[$src$13>>2]|0;
      $258 = $257 & $158;
      HEAP32[$dst$13>>2] = $258;
      $259 = ((($dst$13)) + 4|0);
      $260 = ((($src$13)) + 4|0);
      $dst$14 = $259;$n2$4 = $n2$3;$src$14 = $260;
      label = 34;
      continue;
     }
     else if ((label|0) == 34) {
      label = 0;
      $261 = HEAP32[$src$14>>2]|0;
      $262 = $261 & $158;
      HEAP32[$dst$14>>2] = $262;
      $263 = ((($dst$14)) + 4|0);
      $264 = ((($src$14)) + 4|0);
      $dst$15 = $263;$n2$5 = $n2$4;$src$15 = $264;
      label = 35;
      continue;
     }
     else if ((label|0) == 35) {
      label = 0;
      $265 = HEAP32[$src$15>>2]|0;
      $266 = $265 & $158;
      HEAP32[$dst$15>>2] = $266;
      $267 = ((($dst$15)) + 4|0);
      $268 = ((($src$15)) + 4|0);
      $dst$16 = $267;$n2$6 = $n2$5;$src$16 = $268;
      label = 36;
      continue;
     }
     else if ((label|0) == 36) {
      label = 0;
      $269 = HEAP32[$src$16>>2]|0;
      $270 = $269 & $158;
      HEAP32[$dst$16>>2] = $270;
      $271 = ((($dst$16)) + 4|0);
      $272 = ((($src$16)) + 4|0);
      $dst$17 = $271;$n2$7 = $n2$6;$src$17 = $272;
      label = 37;
      continue;
     }
     else if ((label|0) == 37) {
      label = 0;
      $273 = HEAP32[$src$17>>2]|0;
      $274 = $273 & $158;
      HEAP32[$dst$17>>2] = $274;
      $246 = ((($dst$17)) + 4|0);
      $244 = ((($src$17)) + 4|0);
      $275 = ($n2$7|0)>(1);
      if (!($275)) {
       break;
      }
      $242 = (($n2$7) + -1)|0;
      $243 = HEAP32[$244>>2]|0;
      $245 = $243 & $158;
      HEAP32[$246>>2] = $245;
      $247 = ((($dst$17)) + 8|0);
      $248 = ((($src$17)) + 8|0);
      $249 = HEAP32[$248>>2]|0;
      $250 = $249 & $158;
      HEAP32[$247>>2] = $250;
      $251 = ((($dst$17)) + 12|0);
      $252 = ((($src$17)) + 12|0);
      $dst$12 = $251;$n2$2 = $242;$src$12 = $252;
      label = 32;
      continue;
     }
    }
    $276 = (($244) + ($6)|0);
    $277 = (($246) + ($10)|0);
    $278 = (($279) + -1)|0;
    $280 = ($279|0)==(0);
    if ($280) {
     label = 40;
     break;
    } else {
     $279 = $278;$dst$92 = $277;$src$93 = $276;
    }
   }
   if ((label|0) == 38) {
    // unreachable;
   }
   else if ((label|0) == 40) {
    return;
   }
  }
  }
 } else {
  $18 = ((($info)) + 63|0);
  $19 = HEAP8[$18>>0]|0;
  $20 = $19&255;
  $21 = ((($14)) + 31|0);
  $22 = HEAP8[$21>>0]|0;
  $23 = $22&255;
  $24 = $20 >>> $23;
  $25 = ((($14)) + 35|0);
  $26 = HEAP8[$25>>0]|0;
  $27 = $26&255;
  $28 = $24 << $27;
  $29 = (($3) + -1)|0;
  $30 = ($3|0)==(0);
  if ($30) {
   return;
  }
  $31 = (($1) + 7)|0;
  $32 = (($31|0) / 8)&-1;
  $33 = $1 & 7;
  switch ($33|0) {
  case 0:  {
   $71 = $29;$dst$04$us = $8;$src$05$us = $4;
   while(1) {
    $dst$1$us = $dst$04$us;$n$0$us = $32;$src$1$us = $src$05$us;
    while(1) {
     $34 = HEAP32[$src$1$us>>2]|0;
     $35 = $34 | $28;
     HEAP32[$dst$1$us>>2] = $35;
     $36 = ((($dst$1$us)) + 4|0);
     $37 = ((($src$1$us)) + 4|0);
     $38 = HEAP32[$37>>2]|0;
     $39 = $38 | $28;
     HEAP32[$36>>2] = $39;
     $40 = ((($dst$1$us)) + 8|0);
     $41 = ((($src$1$us)) + 8|0);
     $42 = HEAP32[$41>>2]|0;
     $43 = $42 | $28;
     HEAP32[$40>>2] = $43;
     $44 = ((($dst$1$us)) + 12|0);
     $45 = ((($src$1$us)) + 12|0);
     $46 = HEAP32[$45>>2]|0;
     $47 = $46 | $28;
     HEAP32[$44>>2] = $47;
     $48 = ((($dst$1$us)) + 16|0);
     $49 = ((($src$1$us)) + 16|0);
     $50 = HEAP32[$49>>2]|0;
     $51 = $50 | $28;
     HEAP32[$48>>2] = $51;
     $52 = ((($dst$1$us)) + 20|0);
     $53 = ((($src$1$us)) + 20|0);
     $54 = HEAP32[$53>>2]|0;
     $55 = $54 | $28;
     HEAP32[$52>>2] = $55;
     $56 = ((($dst$1$us)) + 24|0);
     $57 = ((($src$1$us)) + 24|0);
     $58 = HEAP32[$57>>2]|0;
     $59 = $58 | $28;
     HEAP32[$56>>2] = $59;
     $60 = ((($dst$1$us)) + 28|0);
     $61 = ((($src$1$us)) + 28|0);
     $62 = HEAP32[$61>>2]|0;
     $63 = $62 | $28;
     HEAP32[$60>>2] = $63;
     $64 = ((($dst$1$us)) + 32|0);
     $65 = ((($src$1$us)) + 32|0);
     $66 = (($n$0$us) + -1)|0;
     $67 = ($n$0$us|0)>(1);
     if ($67) {
      $dst$1$us = $64;$n$0$us = $66;$src$1$us = $65;
     } else {
      $$lcssa123 = $64;$$lcssa124 = $65;
      break;
     }
    }
    $68 = (($$lcssa124) + ($6)|0);
    $69 = (($$lcssa123) + ($10)|0);
    $70 = (($71) + -1)|0;
    $72 = ($71|0)==(0);
    if ($72) {
     break;
    } else {
     $71 = $70;$dst$04$us = $69;$src$05$us = $68;
    }
   }
   return;
   break;
  }
  case 7:  {
   $110 = $29;$dst$04$us39 = $8;$src$05$us38 = $4;
   while(1) {
    $dst$2$us43 = $dst$04$us39;$n$1$us45 = $32;$src$2$us44 = $src$05$us38;
    while(1) {
     $80 = HEAP32[$src$2$us44>>2]|0;
     $81 = $80 | $28;
     HEAP32[$dst$2$us43>>2] = $81;
     $82 = ((($dst$2$us43)) + 4|0);
     $83 = ((($src$2$us44)) + 4|0);
     $84 = HEAP32[$83>>2]|0;
     $85 = $84 | $28;
     HEAP32[$82>>2] = $85;
     $86 = ((($dst$2$us43)) + 8|0);
     $87 = ((($src$2$us44)) + 8|0);
     $88 = HEAP32[$87>>2]|0;
     $89 = $88 | $28;
     HEAP32[$86>>2] = $89;
     $90 = ((($dst$2$us43)) + 12|0);
     $91 = ((($src$2$us44)) + 12|0);
     $92 = HEAP32[$91>>2]|0;
     $93 = $92 | $28;
     HEAP32[$90>>2] = $93;
     $94 = ((($dst$2$us43)) + 16|0);
     $95 = ((($src$2$us44)) + 16|0);
     $96 = HEAP32[$95>>2]|0;
     $97 = $96 | $28;
     HEAP32[$94>>2] = $97;
     $98 = ((($dst$2$us43)) + 20|0);
     $99 = ((($src$2$us44)) + 20|0);
     $100 = HEAP32[$99>>2]|0;
     $101 = $100 | $28;
     HEAP32[$98>>2] = $101;
     $102 = ((($dst$2$us43)) + 24|0);
     $103 = ((($src$2$us44)) + 24|0);
     $104 = HEAP32[$103>>2]|0;
     $105 = $104 | $28;
     HEAP32[$102>>2] = $105;
     $77 = ((($dst$2$us43)) + 28|0);
     $75 = ((($src$2$us44)) + 28|0);
     $106 = ($n$1$us45|0)>(1);
     if (!($106)) {
      $$lcssa125 = $77;$$lcssa126 = $75;
      break;
     }
     $73 = (($n$1$us45) + -1)|0;
     $74 = HEAP32[$75>>2]|0;
     $76 = $74 | $28;
     HEAP32[$77>>2] = $76;
     $78 = ((($dst$2$us43)) + 32|0);
     $79 = ((($src$2$us44)) + 32|0);
     $dst$2$us43 = $78;$n$1$us45 = $73;$src$2$us44 = $79;
    }
    $107 = (($$lcssa126) + ($6)|0);
    $108 = (($$lcssa125) + ($10)|0);
    $109 = (($110) + -1)|0;
    $111 = ($110|0)==(0);
    if ($111) {
     break;
    } else {
     $110 = $109;$dst$04$us39 = $108;$src$05$us38 = $107;
    }
   }
   return;
   break;
  }
  default: {
   $149 = $29;$dst$04 = $8;$src$05 = $4;
   L40: while(1) {
    switch ($33|0) {
    case 1:  {
     $dst$8 = $dst$04;$n$7 = $32;$src$8 = $src$05;
     label = 18;
     break;
    }
    case 2:  {
     $dst$7 = $dst$04;$n$6 = $32;$src$7 = $src$05;
     label = 17;
     break;
    }
    case 6:  {
     $dst$3 = $dst$04;$n$2 = $32;$src$3 = $src$05;
     label = 13;
     break;
    }
    case 5:  {
     $dst$4 = $dst$04;$n$3 = $32;$src$4 = $src$05;
     label = 14;
     break;
    }
    case 4:  {
     $dst$5 = $dst$04;$n$4 = $32;$src$5 = $src$05;
     label = 15;
     break;
    }
    case 3:  {
     $dst$6 = $dst$04;$n$5 = $32;$src$6 = $src$05;
     label = 16;
     break;
    }
    default: {
     label = 19;
     break L40;
    }
    }
    while(1) {
     if ((label|0) == 13) {
      label = 0;
      $123 = HEAP32[$src$3>>2]|0;
      $124 = $123 | $28;
      HEAP32[$dst$3>>2] = $124;
      $125 = ((($dst$3)) + 4|0);
      $126 = ((($src$3)) + 4|0);
      $dst$4 = $125;$n$3 = $n$2;$src$4 = $126;
      label = 14;
      continue;
     }
     else if ((label|0) == 14) {
      label = 0;
      $127 = HEAP32[$src$4>>2]|0;
      $128 = $127 | $28;
      HEAP32[$dst$4>>2] = $128;
      $129 = ((($dst$4)) + 4|0);
      $130 = ((($src$4)) + 4|0);
      $dst$5 = $129;$n$4 = $n$3;$src$5 = $130;
      label = 15;
      continue;
     }
     else if ((label|0) == 15) {
      label = 0;
      $131 = HEAP32[$src$5>>2]|0;
      $132 = $131 | $28;
      HEAP32[$dst$5>>2] = $132;
      $133 = ((($dst$5)) + 4|0);
      $134 = ((($src$5)) + 4|0);
      $dst$6 = $133;$n$5 = $n$4;$src$6 = $134;
      label = 16;
      continue;
     }
     else if ((label|0) == 16) {
      label = 0;
      $135 = HEAP32[$src$6>>2]|0;
      $136 = $135 | $28;
      HEAP32[$dst$6>>2] = $136;
      $137 = ((($dst$6)) + 4|0);
      $138 = ((($src$6)) + 4|0);
      $dst$7 = $137;$n$6 = $n$5;$src$7 = $138;
      label = 17;
      continue;
     }
     else if ((label|0) == 17) {
      label = 0;
      $139 = HEAP32[$src$7>>2]|0;
      $140 = $139 | $28;
      HEAP32[$dst$7>>2] = $140;
      $141 = ((($dst$7)) + 4|0);
      $142 = ((($src$7)) + 4|0);
      $dst$8 = $141;$n$7 = $n$6;$src$8 = $142;
      label = 18;
      continue;
     }
     else if ((label|0) == 18) {
      label = 0;
      $143 = HEAP32[$src$8>>2]|0;
      $144 = $143 | $28;
      HEAP32[$dst$8>>2] = $144;
      $116 = ((($dst$8)) + 4|0);
      $114 = ((($src$8)) + 4|0);
      $145 = ($n$7|0)>(1);
      if (!($145)) {
       break;
      }
      $112 = (($n$7) + -1)|0;
      $113 = HEAP32[$114>>2]|0;
      $115 = $113 | $28;
      HEAP32[$116>>2] = $115;
      $117 = ((($dst$8)) + 8|0);
      $118 = ((($src$8)) + 8|0);
      $119 = HEAP32[$118>>2]|0;
      $120 = $119 | $28;
      HEAP32[$117>>2] = $120;
      $121 = ((($dst$8)) + 12|0);
      $122 = ((($src$8)) + 12|0);
      $dst$3 = $121;$n$2 = $112;$src$3 = $122;
      label = 13;
      continue;
     }
    }
    $146 = (($114) + ($6)|0);
    $147 = (($116) + ($10)|0);
    $148 = (($149) + -1)|0;
    $150 = ($149|0)==(0);
    if ($150) {
     label = 40;
     break;
    } else {
     $149 = $148;$dst$04 = $147;$src$05 = $146;
    }
   }
   if ((label|0) == 19) {
    // unreachable;
   }
   else if ((label|0) == 40) {
    return;
   }
  }
  }
 }
}
function _Blit2to2Key($info) {
 $info = $info|0;
 var $$lcssa = 0, $$lcssa14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dstp$02 = 0;
 var $dstp$02$us = 0, $dstp$1$us = 0, $dstp$2 = 0, $dstp$3 = 0, $dstp$4 = 0, $dstp$5 = 0, $dstp$6 = 0, $dstp$7 = 0, $dstp$8 = 0, $n$0$us = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $srcp$01 = 0, $srcp$01$us = 0, $srcp$1$us = 0;
 var $srcp$2 = 0, $srcp$3 = 0, $srcp$4 = 0, $srcp$5 = 0, $srcp$6 = 0, $srcp$7 = 0, $srcp$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$info>>2]|0;
 $5 = ((($info)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($info)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($info)) + 36|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($info)) + 56|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($info)) + 40|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($14)) + 24|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 ^ -1;
 $18 = (($6|0) / 2)&-1;
 $19 = (($10|0) / 2)&-1;
 $20 = $12 & $17;
 $21 = (($3) + -1)|0;
 $22 = ($3|0)==(0);
 if ($22) {
  return;
 }
 $23 = (($1) + 7)|0;
 $24 = (($23|0) / 8)&-1;
 $25 = $1 & 7;
 $26 = ($25|0)==(0);
 if ($26) {
  $80 = $21;$dstp$02$us = $8;$srcp$01$us = $4;
  while(1) {
   $dstp$1$us = $dstp$02$us;$n$0$us = $24;$srcp$1$us = $srcp$01$us;
   while(1) {
    $27 = HEAP16[$srcp$1$us>>1]|0;
    $28 = $27&65535;
    $29 = $28 & $17;
    $30 = ($29|0)==($20|0);
    if (!($30)) {
     HEAP16[$dstp$1$us>>1] = $27;
    }
    $31 = ((($srcp$1$us)) + 2|0);
    $32 = HEAP16[$31>>1]|0;
    $33 = $32&65535;
    $34 = $33 & $17;
    $35 = ($34|0)==($20|0);
    if (!($35)) {
     $36 = ((($dstp$1$us)) + 2|0);
     HEAP16[$36>>1] = $32;
    }
    $37 = ((($srcp$1$us)) + 4|0);
    $38 = HEAP16[$37>>1]|0;
    $39 = $38&65535;
    $40 = $39 & $17;
    $41 = ($40|0)==($20|0);
    if (!($41)) {
     $42 = ((($dstp$1$us)) + 4|0);
     HEAP16[$42>>1] = $38;
    }
    $43 = ((($srcp$1$us)) + 6|0);
    $44 = HEAP16[$43>>1]|0;
    $45 = $44&65535;
    $46 = $45 & $17;
    $47 = ($46|0)==($20|0);
    if (!($47)) {
     $48 = ((($dstp$1$us)) + 6|0);
     HEAP16[$48>>1] = $44;
    }
    $49 = ((($srcp$1$us)) + 8|0);
    $50 = HEAP16[$49>>1]|0;
    $51 = $50&65535;
    $52 = $51 & $17;
    $53 = ($52|0)==($20|0);
    if (!($53)) {
     $54 = ((($dstp$1$us)) + 8|0);
     HEAP16[$54>>1] = $50;
    }
    $55 = ((($srcp$1$us)) + 10|0);
    $56 = HEAP16[$55>>1]|0;
    $57 = $56&65535;
    $58 = $57 & $17;
    $59 = ($58|0)==($20|0);
    if (!($59)) {
     $60 = ((($dstp$1$us)) + 10|0);
     HEAP16[$60>>1] = $56;
    }
    $61 = ((($srcp$1$us)) + 12|0);
    $62 = HEAP16[$61>>1]|0;
    $63 = $62&65535;
    $64 = $63 & $17;
    $65 = ($64|0)==($20|0);
    if (!($65)) {
     $66 = ((($dstp$1$us)) + 12|0);
     HEAP16[$66>>1] = $62;
    }
    $67 = ((($srcp$1$us)) + 14|0);
    $68 = HEAP16[$67>>1]|0;
    $69 = $68&65535;
    $70 = $69 & $17;
    $71 = ($70|0)==($20|0);
    if (!($71)) {
     $72 = ((($dstp$1$us)) + 14|0);
     HEAP16[$72>>1] = $68;
    }
    $73 = ((($dstp$1$us)) + 16|0);
    $74 = ((($srcp$1$us)) + 16|0);
    $75 = (($n$0$us) + -1)|0;
    $76 = ($n$0$us|0)>(1);
    if ($76) {
     $dstp$1$us = $73;$n$0$us = $75;$srcp$1$us = $74;
    } else {
     $$lcssa = $73;$$lcssa14 = $74;
     break;
    }
   }
   $77 = (($$lcssa14) + ($18<<1)|0);
   $78 = (($$lcssa) + ($19<<1)|0);
   $79 = (($80) + -1)|0;
   $81 = ($80|0)==(0);
   if ($81) {
    break;
   } else {
    $80 = $79;$dstp$02$us = $78;$srcp$01$us = $77;
   }
  }
  return;
 } else {
  $135 = $21;$dstp$02 = $8;$srcp$01 = $4;
 }
 L35: while(1) {
  switch ($25|0) {
  case 1:  {
   $dstp$8 = $dstp$02;$n$7 = $24;$srcp$8 = $srcp$01;
   label = 44;
   break;
  }
  case 7:  {
   $dstp$2 = $dstp$02;$n$1 = $24;$srcp$2 = $srcp$01;
   label = 26;
   break;
  }
  case 6:  {
   $dstp$3 = $dstp$02;$n$2 = $24;$srcp$3 = $srcp$01;
   label = 29;
   break;
  }
  case 5:  {
   $dstp$4 = $dstp$02;$n$3 = $24;$srcp$4 = $srcp$01;
   label = 32;
   break;
  }
  case 4:  {
   $dstp$5 = $dstp$02;$n$4 = $24;$srcp$5 = $srcp$01;
   label = 35;
   break;
  }
  case 3:  {
   $dstp$6 = $dstp$02;$n$5 = $24;$srcp$6 = $srcp$01;
   label = 38;
   break;
  }
  case 2:  {
   $dstp$7 = $dstp$02;$n$6 = $24;$srcp$7 = $srcp$01;
   label = 41;
   break;
  }
  default: {
   label = 47;
   break L35;
  }
  }
  while(1) {
   if ((label|0) == 26) {
    label = 0;
    $90 = HEAP16[$srcp$2>>1]|0;
    $91 = $90&65535;
    $92 = $91 & $17;
    $93 = ($92|0)==($20|0);
    if (!($93)) {
     HEAP16[$dstp$2>>1] = $90;
    }
    $94 = ((($dstp$2)) + 2|0);
    $95 = ((($srcp$2)) + 2|0);
    $dstp$3 = $94;$n$2 = $n$1;$srcp$3 = $95;
    label = 29;
    continue;
   }
   else if ((label|0) == 29) {
    label = 0;
    $96 = HEAP16[$srcp$3>>1]|0;
    $97 = $96&65535;
    $98 = $97 & $17;
    $99 = ($98|0)==($20|0);
    if (!($99)) {
     HEAP16[$dstp$3>>1] = $96;
    }
    $100 = ((($dstp$3)) + 2|0);
    $101 = ((($srcp$3)) + 2|0);
    $dstp$4 = $100;$n$3 = $n$2;$srcp$4 = $101;
    label = 32;
    continue;
   }
   else if ((label|0) == 32) {
    label = 0;
    $102 = HEAP16[$srcp$4>>1]|0;
    $103 = $102&65535;
    $104 = $103 & $17;
    $105 = ($104|0)==($20|0);
    if (!($105)) {
     HEAP16[$dstp$4>>1] = $102;
    }
    $106 = ((($dstp$4)) + 2|0);
    $107 = ((($srcp$4)) + 2|0);
    $dstp$5 = $106;$n$4 = $n$3;$srcp$5 = $107;
    label = 35;
    continue;
   }
   else if ((label|0) == 35) {
    label = 0;
    $108 = HEAP16[$srcp$5>>1]|0;
    $109 = $108&65535;
    $110 = $109 & $17;
    $111 = ($110|0)==($20|0);
    if (!($111)) {
     HEAP16[$dstp$5>>1] = $108;
    }
    $112 = ((($dstp$5)) + 2|0);
    $113 = ((($srcp$5)) + 2|0);
    $dstp$6 = $112;$n$5 = $n$4;$srcp$6 = $113;
    label = 38;
    continue;
   }
   else if ((label|0) == 38) {
    label = 0;
    $114 = HEAP16[$srcp$6>>1]|0;
    $115 = $114&65535;
    $116 = $115 & $17;
    $117 = ($116|0)==($20|0);
    if (!($117)) {
     HEAP16[$dstp$6>>1] = $114;
    }
    $118 = ((($dstp$6)) + 2|0);
    $119 = ((($srcp$6)) + 2|0);
    $dstp$7 = $118;$n$6 = $n$5;$srcp$7 = $119;
    label = 41;
    continue;
   }
   else if ((label|0) == 41) {
    label = 0;
    $120 = HEAP16[$srcp$7>>1]|0;
    $121 = $120&65535;
    $122 = $121 & $17;
    $123 = ($122|0)==($20|0);
    if (!($123)) {
     HEAP16[$dstp$7>>1] = $120;
    }
    $124 = ((($dstp$7)) + 2|0);
    $125 = ((($srcp$7)) + 2|0);
    $dstp$8 = $124;$n$7 = $n$6;$srcp$8 = $125;
    label = 44;
    continue;
   }
   else if ((label|0) == 44) {
    label = 0;
    $126 = HEAP16[$srcp$8>>1]|0;
    $127 = $126&65535;
    $128 = $127 & $17;
    $129 = ($128|0)==($20|0);
    if (!($129)) {
     HEAP16[$dstp$8>>1] = $126;
    }
    $87 = ((($dstp$8)) + 2|0);
    $83 = ((($srcp$8)) + 2|0);
    $130 = (($n$7) + -1)|0;
    $131 = ($n$7|0)>(1);
    if (!($131)) {
     break;
    }
    $82 = HEAP16[$83>>1]|0;
    $84 = $82&65535;
    $85 = $84 & $17;
    $86 = ($85|0)==($20|0);
    if (!($86)) {
     HEAP16[$87>>1] = $82;
    }
    $88 = ((($dstp$8)) + 4|0);
    $89 = ((($srcp$8)) + 4|0);
    $dstp$2 = $88;$n$1 = $130;$srcp$2 = $89;
    label = 26;
    continue;
   }
  }
  $132 = (($83) + ($18<<1)|0);
  $133 = (($87) + ($19<<1)|0);
  $134 = (($135) + -1)|0;
  $136 = ($135|0)==(0);
  if ($136) {
   label = 49;
   break;
  } else {
   $135 = $134;$dstp$02 = $133;$srcp$01 = $132;
  }
 }
 if ((label|0) == 47) {
  // unreachable;
 }
 else if ((label|0) == 49) {
  return;
 }
}
function _BlitNto1Key($info) {
 $info = $info|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0;
 var $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0;
 var $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0;
 var $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0;
 var $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0;
 var $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0;
 var $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0;
 var $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0;
 var $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0;
 var $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0;
 var $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0;
 var $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0;
 var $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0;
 var $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0;
 var $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0, $1267 = 0;
 var $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0, $1285 = 0;
 var $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0, $1302 = 0;
 var $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0, $1320 = 0;
 var $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0, $1339 = 0;
 var $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0, $1357 = 0;
 var $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0, $1375 = 0;
 var $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0, $1393 = 0;
 var $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0, $1410 = 0;
 var $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0, $1429 = 0;
 var $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0, $1447 = 0;
 var $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0, $1465 = 0;
 var $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0, $1483 = 0;
 var $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0, $1500 = 0;
 var $1501 = 0, $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0, $1511 = 0, $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0, $1518 = 0, $1519 = 0;
 var $152 = 0, $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0, $153 = 0, $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0, $1536 = 0, $1537 = 0;
 var $1538 = 0, $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0, $1548 = 0, $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0, $1554 = 0, $1555 = 0;
 var $1556 = 0, $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $1565 = 0, $1566 = 0, $1567 = 0, $1568 = 0, $1569 = 0, $157 = 0, $1570 = 0, $1571 = 0, $1572 = 0, $1573 = 0;
 var $1574 = 0, $1575 = 0, $1576 = 0, $1577 = 0, $1578 = 0, $1579 = 0, $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0, $1583 = 0, $1584 = 0, $1585 = 0, $1586 = 0, $1587 = 0, $1588 = 0, $1589 = 0, $159 = 0, $1590 = 0, $1591 = 0;
 var $1592 = 0, $1593 = 0, $1594 = 0, $1595 = 0, $1596 = 0, $1597 = 0, $1598 = 0, $1599 = 0, $16 = 0, $160 = 0, $1600 = 0, $1601 = 0, $1602 = 0, $1603 = 0, $1604 = 0, $1605 = 0, $1606 = 0, $1607 = 0, $1608 = 0, $1609 = 0;
 var $161 = 0, $1610 = 0, $1611 = 0, $1612 = 0, $1613 = 0, $1614 = 0, $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0, $1619 = 0, $162 = 0, $1620 = 0, $1621 = 0, $1622 = 0, $1623 = 0, $1624 = 0, $1625 = 0, $1626 = 0, $1627 = 0;
 var $1628 = 0, $1629 = 0, $163 = 0, $1630 = 0, $1631 = 0, $1632 = 0, $1633 = 0, $1634 = 0, $1635 = 0, $1636 = 0, $1637 = 0, $1638 = 0, $1639 = 0, $164 = 0, $1640 = 0, $1641 = 0, $1642 = 0, $1643 = 0, $1644 = 0, $1645 = 0;
 var $1646 = 0, $1647 = 0, $1648 = 0, $1649 = 0, $165 = 0, $1650 = 0, $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0, $1655 = 0, $1656 = 0, $1657 = 0, $1658 = 0, $1659 = 0, $166 = 0, $1660 = 0, $1661 = 0, $1662 = 0, $1663 = 0;
 var $1664 = 0, $1665 = 0, $1666 = 0, $1667 = 0, $1668 = 0, $1669 = 0, $167 = 0, $1670 = 0, $1671 = 0, $1672 = 0, $1673 = 0, $1674 = 0, $1675 = 0, $1676 = 0, $1677 = 0, $1678 = 0, $1679 = 0, $168 = 0, $1680 = 0, $1681 = 0;
 var $1682 = 0, $1683 = 0, $1684 = 0, $1685 = 0, $1686 = 0, $1687 = 0, $1688 = 0, $1689 = 0, $169 = 0, $1690 = 0, $1691 = 0, $1692 = 0, $1693 = 0, $1694 = 0, $1695 = 0, $1696 = 0, $1697 = 0, $1698 = 0, $1699 = 0, $17 = 0;
 var $170 = 0, $1700 = 0, $1701 = 0, $1702 = 0, $1703 = 0, $1704 = 0, $1705 = 0, $1706 = 0, $1707 = 0, $1708 = 0, $1709 = 0, $171 = 0, $1710 = 0, $1711 = 0, $1712 = 0, $1713 = 0, $1714 = 0, $1715 = 0, $1716 = 0, $1717 = 0;
 var $1718 = 0, $1719 = 0, $172 = 0, $1720 = 0, $1721 = 0, $1722 = 0, $1723 = 0, $1724 = 0, $1725 = 0, $1726 = 0, $1727 = 0, $1728 = 0, $1729 = 0, $173 = 0, $1730 = 0, $1731 = 0, $1732 = 0, $1733 = 0, $1734 = 0, $1735 = 0;
 var $1736 = 0, $1737 = 0, $1738 = 0, $1739 = 0, $174 = 0, $1740 = 0, $1741 = 0, $1742 = 0, $1743 = 0, $1744 = 0, $1745 = 0, $1746 = 0, $1747 = 0, $1748 = 0, $1749 = 0, $175 = 0, $1750 = 0, $1751 = 0, $1752 = 0, $1753 = 0;
 var $1754 = 0, $1755 = 0, $1756 = 0, $1757 = 0, $1758 = 0, $1759 = 0, $176 = 0, $1760 = 0, $1761 = 0, $1762 = 0, $1763 = 0, $1764 = 0, $1765 = 0, $1766 = 0, $1767 = 0, $1768 = 0, $1769 = 0, $177 = 0, $1770 = 0, $1771 = 0;
 var $1772 = 0, $1773 = 0, $1774 = 0, $1775 = 0, $1776 = 0, $1777 = 0, $1778 = 0, $1779 = 0, $178 = 0, $1780 = 0, $1781 = 0, $1782 = 0, $1783 = 0, $1784 = 0, $1785 = 0, $1786 = 0, $1787 = 0, $1788 = 0, $1789 = 0, $179 = 0;
 var $1790 = 0, $1791 = 0, $1792 = 0, $1793 = 0, $1794 = 0, $1795 = 0, $1796 = 0, $1797 = 0, $1798 = 0, $1799 = 0, $18 = 0, $180 = 0, $1800 = 0, $1801 = 0, $1802 = 0, $1803 = 0, $1804 = 0, $1805 = 0, $1806 = 0, $1807 = 0;
 var $1808 = 0, $1809 = 0, $181 = 0, $1810 = 0, $1811 = 0, $1812 = 0, $1813 = 0, $1814 = 0, $1815 = 0, $1816 = 0, $1817 = 0, $1818 = 0, $1819 = 0, $182 = 0, $1820 = 0, $1821 = 0, $1822 = 0, $1823 = 0, $1824 = 0, $1825 = 0;
 var $1826 = 0, $1827 = 0, $1828 = 0, $1829 = 0, $183 = 0, $1830 = 0, $1831 = 0, $1832 = 0, $1833 = 0, $1834 = 0, $1835 = 0, $1836 = 0, $1837 = 0, $1838 = 0, $1839 = 0, $184 = 0, $1840 = 0, $1841 = 0, $1842 = 0, $1843 = 0;
 var $1844 = 0, $1845 = 0, $1846 = 0, $1847 = 0, $1848 = 0, $1849 = 0, $185 = 0, $1850 = 0, $1851 = 0, $1852 = 0, $1853 = 0, $1854 = 0, $1855 = 0, $1856 = 0, $1857 = 0, $1858 = 0, $1859 = 0, $186 = 0, $1860 = 0, $1861 = 0;
 var $1862 = 0, $1863 = 0, $1864 = 0, $1865 = 0, $1866 = 0, $1867 = 0, $1868 = 0, $1869 = 0, $187 = 0, $1870 = 0, $1871 = 0, $1872 = 0, $1873 = 0, $1874 = 0, $1875 = 0, $1876 = 0, $1877 = 0, $1878 = 0, $1879 = 0, $188 = 0;
 var $1880 = 0, $1881 = 0, $1882 = 0, $1883 = 0, $1884 = 0, $1885 = 0, $1886 = 0, $1887 = 0, $1888 = 0, $1889 = 0, $189 = 0, $1890 = 0, $1891 = 0, $1892 = 0, $1893 = 0, $1894 = 0, $1895 = 0, $1896 = 0, $1897 = 0, $1898 = 0;
 var $1899 = 0, $19 = 0, $190 = 0, $1900 = 0, $1901 = 0, $1902 = 0, $1903 = 0, $1904 = 0, $1905 = 0, $1906 = 0, $1907 = 0, $1908 = 0, $1909 = 0, $191 = 0, $1910 = 0, $1911 = 0, $1912 = 0, $1913 = 0, $1914 = 0, $1915 = 0;
 var $1916 = 0, $1917 = 0, $1918 = 0, $1919 = 0, $192 = 0, $1920 = 0, $1921 = 0, $1922 = 0, $1923 = 0, $1924 = 0, $1925 = 0, $1926 = 0, $1927 = 0, $1928 = 0, $1929 = 0, $193 = 0, $1930 = 0, $1931 = 0, $1932 = 0, $1933 = 0;
 var $1934 = 0, $1935 = 0, $1936 = 0, $1937 = 0, $1938 = 0, $1939 = 0, $194 = 0, $1940 = 0, $1941 = 0, $1942 = 0, $1943 = 0, $1944 = 0, $1945 = 0, $1946 = 0, $1947 = 0, $1948 = 0, $1949 = 0, $195 = 0, $1950 = 0, $1951 = 0;
 var $1952 = 0, $1953 = 0, $1954 = 0, $1955 = 0, $1956 = 0, $1957 = 0, $1958 = 0, $1959 = 0, $196 = 0, $1960 = 0, $1961 = 0, $1962 = 0, $1963 = 0, $1964 = 0, $1965 = 0, $1966 = 0, $1967 = 0, $1968 = 0, $1969 = 0, $197 = 0;
 var $1970 = 0, $1971 = 0, $1972 = 0, $1973 = 0, $1974 = 0, $1975 = 0, $1976 = 0, $1977 = 0, $1978 = 0, $1979 = 0, $198 = 0, $1980 = 0, $1981 = 0, $1982 = 0, $1983 = 0, $1984 = 0, $1985 = 0, $1986 = 0, $1987 = 0, $1988 = 0;
 var $1989 = 0, $199 = 0, $1990 = 0, $1991 = 0, $1992 = 0, $1993 = 0, $1994 = 0, $1995 = 0, $1996 = 0, $1997 = 0, $1998 = 0, $1999 = 0, $2 = 0, $20 = 0, $200 = 0, $2000 = 0, $2001 = 0, $2002 = 0, $2003 = 0, $2004 = 0;
 var $2005 = 0, $2006 = 0, $2007 = 0, $2008 = 0, $2009 = 0, $201 = 0, $2010 = 0, $2011 = 0, $2012 = 0, $2013 = 0, $2014 = 0, $2015 = 0, $2016 = 0, $2017 = 0, $2018 = 0, $2019 = 0, $202 = 0, $2020 = 0, $2021 = 0, $2022 = 0;
 var $2023 = 0, $2024 = 0, $2025 = 0, $2026 = 0, $2027 = 0, $2028 = 0, $2029 = 0, $203 = 0, $2030 = 0, $2031 = 0, $2032 = 0, $2033 = 0, $2034 = 0, $2035 = 0, $2036 = 0, $2037 = 0, $2038 = 0, $2039 = 0, $204 = 0, $2040 = 0;
 var $2041 = 0, $2042 = 0, $2043 = 0, $2044 = 0, $2045 = 0, $2046 = 0, $2047 = 0, $2048 = 0, $2049 = 0, $205 = 0, $2050 = 0, $2051 = 0, $2052 = 0, $2053 = 0, $2054 = 0, $2055 = 0, $2056 = 0, $2057 = 0, $2058 = 0, $2059 = 0;
 var $206 = 0, $2060 = 0, $2061 = 0, $2062 = 0, $2063 = 0, $2064 = 0, $2065 = 0, $2066 = 0, $2067 = 0, $2068 = 0, $2069 = 0, $207 = 0, $2070 = 0, $2071 = 0, $2072 = 0, $2073 = 0, $2074 = 0, $2075 = 0, $2076 = 0, $2077 = 0;
 var $2078 = 0, $2079 = 0, $208 = 0, $2080 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0;
 var $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0;
 var $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0;
 var $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0;
 var $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0;
 var $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0;
 var $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0;
 var $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0;
 var $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0;
 var $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0;
 var $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0;
 var $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0;
 var $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0;
 var $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0;
 var $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0;
 var $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0;
 var $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0;
 var $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0;
 var $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0;
 var $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0;
 var $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0;
 var $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0;
 var $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0;
 var $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0;
 var $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0;
 var $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0;
 var $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0;
 var $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0;
 var $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0;
 var $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0;
 var $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0;
 var $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0;
 var $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0;
 var $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0;
 var $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0;
 var $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0;
 var $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0;
 var $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0;
 var $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0;
 var $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0;
 var $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0;
 var $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0;
 var $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0;
 var $999 = 0, $Pixel$0 = 0, $Pixel$1 = 0, $Pixel$10 = 0, $Pixel$11 = 0, $Pixel$12 = 0, $Pixel$13 = 0, $Pixel$14 = 0, $Pixel$15 = 0, $Pixel$2 = 0, $Pixel$3 = 0, $Pixel$4 = 0, $Pixel$5 = 0, $Pixel$6 = 0, $Pixel$7 = 0, $Pixel$8 = 0, $Pixel$9 = 0, $div = 0, $div1 = 0, $div10 = 0;
 var $div12 = 0, $div13 = 0, $div15 = 0, $div16 = 0, $div18 = 0, $div19 = 0, $div21 = 0, $div22 = 0, $div24 = 0, $div25 = 0, $div27 = 0, $div28 = 0, $div3 = 0, $div30 = 0, $div31 = 0, $div33 = 0, $div34 = 0, $div36 = 0, $div37 = 0, $div39 = 0;
 var $div4 = 0, $div40 = 0, $div42 = 0, $div43 = 0, $div45 = 0, $div46 = 0, $div6 = 0, $div7 = 0, $div9 = 0, $dst$050 = 0, $dst$1 = 0, $dst$10 = 0, $dst$11 = 0, $dst$12 = 0, $dst$13 = 0, $dst$14 = 0, $dst$15 = 0, $dst$16 = 0, $dst$17 = 0, $dst$2 = 0;
 var $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0, $dst$952 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $n1$0 = 0, $n1$1 = 0, $n1$2 = 0, $n1$3 = 0, $n1$4 = 0;
 var $n1$5 = 0, $n1$6 = 0, $n1$7 = 0, $sB$0 = 0, $sB$0$shrunk = 0, $sB$1 = 0, $sB$1$shrunk = 0, $sB$10 = 0, $sB$10$shrunk = 0, $sB$11 = 0, $sB$11$shrunk = 0, $sB$12 = 0, $sB$12$shrunk = 0, $sB$13 = 0, $sB$13$shrunk = 0, $sB$14 = 0, $sB$14$shrunk = 0, $sB$15 = 0, $sB$15$shrunk = 0, $sB$2 = 0;
 var $sB$2$shrunk = 0, $sB$3 = 0, $sB$3$shrunk = 0, $sB$4 = 0, $sB$4$shrunk = 0, $sB$5 = 0, $sB$5$shrunk = 0, $sB$6 = 0, $sB$6$shrunk = 0, $sB$7 = 0, $sB$7$shrunk = 0, $sB$8 = 0, $sB$8$shrunk = 0, $sB$9 = 0, $sB$9$shrunk = 0, $sG$0 = 0, $sG$0$shrunk = 0, $sG$1 = 0, $sG$1$shrunk = 0, $sG$10 = 0;
 var $sG$10$shrunk = 0, $sG$11 = 0, $sG$11$shrunk = 0, $sG$12 = 0, $sG$12$shrunk = 0, $sG$13 = 0, $sG$13$shrunk = 0, $sG$14 = 0, $sG$14$shrunk = 0, $sG$15 = 0, $sG$15$shrunk = 0, $sG$2 = 0, $sG$2$shrunk = 0, $sG$3 = 0, $sG$3$shrunk = 0, $sG$4 = 0, $sG$4$shrunk = 0, $sG$5 = 0, $sG$5$shrunk = 0, $sG$6 = 0;
 var $sG$6$shrunk = 0, $sG$7 = 0, $sG$7$shrunk = 0, $sG$8 = 0, $sG$8$shrunk = 0, $sG$9 = 0, $sG$9$shrunk = 0, $sR$0 = 0, $sR$0$shrunk = 0, $sR$1 = 0, $sR$1$shrunk = 0, $sR$10 = 0, $sR$10$shrunk = 0, $sR$11 = 0, $sR$11$shrunk = 0, $sR$12 = 0, $sR$12$shrunk = 0, $sR$13 = 0, $sR$13$shrunk = 0, $sR$14 = 0;
 var $sR$14$shrunk = 0, $sR$15 = 0, $sR$15$shrunk = 0, $sR$2 = 0, $sR$2$shrunk = 0, $sR$3 = 0, $sR$3$shrunk = 0, $sR$4 = 0, $sR$4$shrunk = 0, $sR$5 = 0, $sR$5$shrunk = 0, $sR$6 = 0, $sR$6$shrunk = 0, $sR$7 = 0, $sR$7$shrunk = 0, $sR$8 = 0, $sR$8$shrunk = 0, $sR$9 = 0, $sR$9$shrunk = 0, $src$051 = 0;
 var $src$1 = 0, $src$10 = 0, $src$11 = 0, $src$12 = 0, $src$13 = 0, $src$14 = 0, $src$15 = 0, $src$16 = 0, $src$17 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0, $src$8 = 0, $src$953 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$info>>2]|0;
 $5 = ((($info)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($info)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($info)) + 36|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($info)) + 40|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($info)) + 48|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($info)) + 56|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($12)) + 24|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = $18 ^ -1;
 $20 = ((($12)) + 9|0);
 $21 = HEAP8[$20>>0]|0;
 $22 = $21&255;
 $23 = $16 & $19;
 $24 = ($14|0)==(0|0);
 $25 = (($3) + -1)|0;
 $26 = ($3|0)==(0);
 if ($24) {
  if ($26) {
   return;
  }
  $35 = (($1) + 7)|0;
  $36 = (($35|0) / 8)&-1;
  $37 = $1 & 7;
  $38 = ((($12)) + 12|0);
  $39 = ((($12)) + 32|0);
  $40 = ((($12)) + 28|0);
  $41 = ((($12)) + 16|0);
  $42 = ((($12)) + 20|0);
  $1056 = $25;$dst$050 = $8;$src$051 = $4;
  L6: while(1) {
   switch ($37|0) {
   case 0:  {
    $dst$1 = $dst$050;$n$0 = $36;$src$1 = $src$051;
    label = 7;
    break;
   }
   case 7:  {
    $dst$2 = $dst$050;$n$1 = $36;$src$2 = $src$051;
    label = 15;
    break;
   }
   case 6:  {
    $dst$3 = $dst$050;$n$2 = $36;$src$3 = $src$051;
    label = 23;
    break;
   }
   case 5:  {
    $dst$4 = $dst$050;$n$3 = $36;$src$4 = $src$051;
    label = 31;
    break;
   }
   case 4:  {
    $dst$5 = $dst$050;$n$4 = $36;$src$5 = $src$051;
    label = 39;
    break;
   }
   case 3:  {
    $dst$6 = $dst$050;$n$5 = $36;$src$6 = $src$051;
    label = 47;
    break;
   }
   case 2:  {
    $dst$7 = $dst$050;$n$6 = $36;$src$7 = $src$051;
    label = 55;
    break;
   }
   case 1:  {
    $dst$8 = $dst$050;$n$7 = $36;$src$8 = $src$051;
    label = 63;
    break;
   }
   default: {
    label = 71;
    break L6;
   }
   }
   while(1) {
    if ((label|0) == 7) {
     label = 0;
     switch ($22|0) {
     case 1:  {
      $43 = HEAP8[$src$1>>0]|0;
      $44 = $43&255;
      $45 = HEAP32[$38>>2]|0;
      $46 = $44 & $45;
      $47 = HEAP32[$39>>2]|0;
      $48 = $47 & 255;
      $49 = $46 >>> $48;
      $50 = HEAP32[$40>>2]|0;
      $51 = $50 & 255;
      $52 = (125924 + ($51<<2)|0);
      $53 = HEAP32[$52>>2]|0;
      $54 = (($53) + ($49)|0);
      $55 = HEAP8[$54>>0]|0;
      $56 = HEAP32[$41>>2]|0;
      $57 = $56 & $44;
      $58 = $47 >>> 8;
      $59 = $58 & 255;
      $60 = $57 >>> $59;
      $61 = $50 >>> 8;
      $62 = $61 & 255;
      $63 = (125924 + ($62<<2)|0);
      $64 = HEAP32[$63>>2]|0;
      $65 = (($64) + ($60)|0);
      $66 = HEAP8[$65>>0]|0;
      $67 = HEAP32[$42>>2]|0;
      $68 = $67 & $44;
      $69 = $47 >>> 16;
      $70 = $69 & 255;
      $71 = $68 >>> $70;
      $72 = $50 >>> 16;
      $73 = $72 & 255;
      $74 = (125924 + ($73<<2)|0);
      $75 = HEAP32[$74>>2]|0;
      $76 = (($75) + ($71)|0);
      $77 = HEAP8[$76>>0]|0;
      $Pixel$0 = $44;$sB$0$shrunk = $77;$sG$0$shrunk = $66;$sR$0$shrunk = $55;
      break;
     }
     case 2:  {
      $78 = HEAP16[$src$1>>1]|0;
      $79 = $78&65535;
      $80 = HEAP32[$38>>2]|0;
      $81 = $79 & $80;
      $82 = HEAP32[$39>>2]|0;
      $83 = $82 & 255;
      $84 = $81 >>> $83;
      $85 = HEAP32[$40>>2]|0;
      $86 = $85 & 255;
      $87 = (125924 + ($86<<2)|0);
      $88 = HEAP32[$87>>2]|0;
      $89 = (($88) + ($84)|0);
      $90 = HEAP8[$89>>0]|0;
      $91 = HEAP32[$41>>2]|0;
      $92 = $91 & $79;
      $93 = $82 >>> 8;
      $94 = $93 & 255;
      $95 = $92 >>> $94;
      $96 = $85 >>> 8;
      $97 = $96 & 255;
      $98 = (125924 + ($97<<2)|0);
      $99 = HEAP32[$98>>2]|0;
      $100 = (($99) + ($95)|0);
      $101 = HEAP8[$100>>0]|0;
      $102 = HEAP32[$42>>2]|0;
      $103 = $102 & $79;
      $104 = $82 >>> 16;
      $105 = $104 & 255;
      $106 = $103 >>> $105;
      $107 = $85 >>> 16;
      $108 = $107 & 255;
      $109 = (125924 + ($108<<2)|0);
      $110 = HEAP32[$109>>2]|0;
      $111 = (($110) + ($106)|0);
      $112 = HEAP8[$111>>0]|0;
      $Pixel$0 = $79;$sB$0$shrunk = $112;$sG$0$shrunk = $101;$sR$0$shrunk = $90;
      break;
     }
     case 3:  {
      $113 = HEAP32[$39>>2]|0;
      $div42 = $113 >>> 3;
      $114 = $div42 & 31;
      $115 = (($src$1) + ($114)|0);
      $116 = HEAP8[$115>>0]|0;
      $117 = $113 >>> 11;
      $div43 = $117 & 31;
      $118 = (($src$1) + ($div43)|0);
      $119 = HEAP8[$118>>0]|0;
      $120 = $113 >>> 19;
      $121 = $120 & 31;
      $122 = (($src$1) + ($121)|0);
      $123 = HEAP8[$122>>0]|0;
      $Pixel$0 = 0;$sB$0$shrunk = $123;$sG$0$shrunk = $119;$sR$0$shrunk = $116;
      break;
     }
     case 4:  {
      $124 = HEAP32[$src$1>>2]|0;
      $125 = HEAP32[$38>>2]|0;
      $126 = $125 & $124;
      $127 = HEAP32[$39>>2]|0;
      $128 = $127 & 255;
      $129 = $126 >>> $128;
      $130 = HEAP32[$40>>2]|0;
      $131 = $130 & 255;
      $132 = (125924 + ($131<<2)|0);
      $133 = HEAP32[$132>>2]|0;
      $134 = (($133) + ($129)|0);
      $135 = HEAP8[$134>>0]|0;
      $136 = HEAP32[$41>>2]|0;
      $137 = $136 & $124;
      $138 = $127 >>> 8;
      $139 = $138 & 255;
      $140 = $137 >>> $139;
      $141 = $130 >>> 8;
      $142 = $141 & 255;
      $143 = (125924 + ($142<<2)|0);
      $144 = HEAP32[$143>>2]|0;
      $145 = (($144) + ($140)|0);
      $146 = HEAP8[$145>>0]|0;
      $147 = HEAP32[$42>>2]|0;
      $148 = $147 & $124;
      $149 = $127 >>> 16;
      $150 = $149 & 255;
      $151 = $148 >>> $150;
      $152 = $130 >>> 16;
      $153 = $152 & 255;
      $154 = (125924 + ($153<<2)|0);
      $155 = HEAP32[$154>>2]|0;
      $156 = (($155) + ($151)|0);
      $157 = HEAP8[$156>>0]|0;
      $Pixel$0 = $124;$sB$0$shrunk = $157;$sG$0$shrunk = $146;$sR$0$shrunk = $135;
      break;
     }
     default: {
      $Pixel$0 = 0;$sB$0$shrunk = 0;$sG$0$shrunk = 0;$sR$0$shrunk = 0;
     }
     }
     $158 = $Pixel$0 & $19;
     $159 = ($158|0)==($23|0);
     if (!($159)) {
      $sB$0 = $sB$0$shrunk&255;
      $sG$0 = $sG$0$shrunk&255;
      $sR$0 = $sR$0$shrunk&255;
      $160 = $sR$0 & 224;
      $161 = $sG$0 >>> 5;
      $162 = $161 << 2;
      $163 = $162 | $160;
      $164 = $sB$0 >>> 6;
      $165 = $163 | $164;
      $166 = $165&255;
      HEAP8[$dst$1>>0] = $166;
     }
     $167 = ((($dst$1)) + 1|0);
     $168 = (($src$1) + ($22)|0);
     $dst$2 = $167;$n$1 = $n$0;$src$2 = $168;
     label = 15;
     continue;
    }
    else if ((label|0) == 15) {
     label = 0;
     switch ($22|0) {
     case 1:  {
      $169 = HEAP8[$src$2>>0]|0;
      $170 = $169&255;
      $171 = HEAP32[$38>>2]|0;
      $172 = $170 & $171;
      $173 = HEAP32[$39>>2]|0;
      $174 = $173 & 255;
      $175 = $172 >>> $174;
      $176 = HEAP32[$40>>2]|0;
      $177 = $176 & 255;
      $178 = (125924 + ($177<<2)|0);
      $179 = HEAP32[$178>>2]|0;
      $180 = (($179) + ($175)|0);
      $181 = HEAP8[$180>>0]|0;
      $182 = HEAP32[$41>>2]|0;
      $183 = $182 & $170;
      $184 = $173 >>> 8;
      $185 = $184 & 255;
      $186 = $183 >>> $185;
      $187 = $176 >>> 8;
      $188 = $187 & 255;
      $189 = (125924 + ($188<<2)|0);
      $190 = HEAP32[$189>>2]|0;
      $191 = (($190) + ($186)|0);
      $192 = HEAP8[$191>>0]|0;
      $193 = HEAP32[$42>>2]|0;
      $194 = $193 & $170;
      $195 = $173 >>> 16;
      $196 = $195 & 255;
      $197 = $194 >>> $196;
      $198 = $176 >>> 16;
      $199 = $198 & 255;
      $200 = (125924 + ($199<<2)|0);
      $201 = HEAP32[$200>>2]|0;
      $202 = (($201) + ($197)|0);
      $203 = HEAP8[$202>>0]|0;
      $Pixel$1 = $170;$sB$1$shrunk = $203;$sG$1$shrunk = $192;$sR$1$shrunk = $181;
      break;
     }
     case 2:  {
      $204 = HEAP16[$src$2>>1]|0;
      $205 = $204&65535;
      $206 = HEAP32[$38>>2]|0;
      $207 = $205 & $206;
      $208 = HEAP32[$39>>2]|0;
      $209 = $208 & 255;
      $210 = $207 >>> $209;
      $211 = HEAP32[$40>>2]|0;
      $212 = $211 & 255;
      $213 = (125924 + ($212<<2)|0);
      $214 = HEAP32[$213>>2]|0;
      $215 = (($214) + ($210)|0);
      $216 = HEAP8[$215>>0]|0;
      $217 = HEAP32[$41>>2]|0;
      $218 = $217 & $205;
      $219 = $208 >>> 8;
      $220 = $219 & 255;
      $221 = $218 >>> $220;
      $222 = $211 >>> 8;
      $223 = $222 & 255;
      $224 = (125924 + ($223<<2)|0);
      $225 = HEAP32[$224>>2]|0;
      $226 = (($225) + ($221)|0);
      $227 = HEAP8[$226>>0]|0;
      $228 = HEAP32[$42>>2]|0;
      $229 = $228 & $205;
      $230 = $208 >>> 16;
      $231 = $230 & 255;
      $232 = $229 >>> $231;
      $233 = $211 >>> 16;
      $234 = $233 & 255;
      $235 = (125924 + ($234<<2)|0);
      $236 = HEAP32[$235>>2]|0;
      $237 = (($236) + ($232)|0);
      $238 = HEAP8[$237>>0]|0;
      $Pixel$1 = $205;$sB$1$shrunk = $238;$sG$1$shrunk = $227;$sR$1$shrunk = $216;
      break;
     }
     case 3:  {
      $239 = HEAP32[$39>>2]|0;
      $div39 = $239 >>> 3;
      $240 = $div39 & 31;
      $241 = (($src$2) + ($240)|0);
      $242 = HEAP8[$241>>0]|0;
      $243 = $239 >>> 11;
      $div40 = $243 & 31;
      $244 = (($src$2) + ($div40)|0);
      $245 = HEAP8[$244>>0]|0;
      $246 = $239 >>> 19;
      $247 = $246 & 31;
      $248 = (($src$2) + ($247)|0);
      $249 = HEAP8[$248>>0]|0;
      $Pixel$1 = 0;$sB$1$shrunk = $249;$sG$1$shrunk = $245;$sR$1$shrunk = $242;
      break;
     }
     case 4:  {
      $250 = HEAP32[$src$2>>2]|0;
      $251 = HEAP32[$38>>2]|0;
      $252 = $251 & $250;
      $253 = HEAP32[$39>>2]|0;
      $254 = $253 & 255;
      $255 = $252 >>> $254;
      $256 = HEAP32[$40>>2]|0;
      $257 = $256 & 255;
      $258 = (125924 + ($257<<2)|0);
      $259 = HEAP32[$258>>2]|0;
      $260 = (($259) + ($255)|0);
      $261 = HEAP8[$260>>0]|0;
      $262 = HEAP32[$41>>2]|0;
      $263 = $262 & $250;
      $264 = $253 >>> 8;
      $265 = $264 & 255;
      $266 = $263 >>> $265;
      $267 = $256 >>> 8;
      $268 = $267 & 255;
      $269 = (125924 + ($268<<2)|0);
      $270 = HEAP32[$269>>2]|0;
      $271 = (($270) + ($266)|0);
      $272 = HEAP8[$271>>0]|0;
      $273 = HEAP32[$42>>2]|0;
      $274 = $273 & $250;
      $275 = $253 >>> 16;
      $276 = $275 & 255;
      $277 = $274 >>> $276;
      $278 = $256 >>> 16;
      $279 = $278 & 255;
      $280 = (125924 + ($279<<2)|0);
      $281 = HEAP32[$280>>2]|0;
      $282 = (($281) + ($277)|0);
      $283 = HEAP8[$282>>0]|0;
      $Pixel$1 = $250;$sB$1$shrunk = $283;$sG$1$shrunk = $272;$sR$1$shrunk = $261;
      break;
     }
     default: {
      $Pixel$1 = 0;$sB$1$shrunk = 0;$sG$1$shrunk = 0;$sR$1$shrunk = 0;
     }
     }
     $284 = $Pixel$1 & $19;
     $285 = ($284|0)==($23|0);
     if (!($285)) {
      $sB$1 = $sB$1$shrunk&255;
      $sG$1 = $sG$1$shrunk&255;
      $sR$1 = $sR$1$shrunk&255;
      $286 = $sR$1 & 224;
      $287 = $sG$1 >>> 5;
      $288 = $287 << 2;
      $289 = $288 | $286;
      $290 = $sB$1 >>> 6;
      $291 = $289 | $290;
      $292 = $291&255;
      HEAP8[$dst$2>>0] = $292;
     }
     $293 = ((($dst$2)) + 1|0);
     $294 = (($src$2) + ($22)|0);
     $dst$3 = $293;$n$2 = $n$1;$src$3 = $294;
     label = 23;
     continue;
    }
    else if ((label|0) == 23) {
     label = 0;
     switch ($22|0) {
     case 1:  {
      $295 = HEAP8[$src$3>>0]|0;
      $296 = $295&255;
      $297 = HEAP32[$38>>2]|0;
      $298 = $296 & $297;
      $299 = HEAP32[$39>>2]|0;
      $300 = $299 & 255;
      $301 = $298 >>> $300;
      $302 = HEAP32[$40>>2]|0;
      $303 = $302 & 255;
      $304 = (125924 + ($303<<2)|0);
      $305 = HEAP32[$304>>2]|0;
      $306 = (($305) + ($301)|0);
      $307 = HEAP8[$306>>0]|0;
      $308 = HEAP32[$41>>2]|0;
      $309 = $308 & $296;
      $310 = $299 >>> 8;
      $311 = $310 & 255;
      $312 = $309 >>> $311;
      $313 = $302 >>> 8;
      $314 = $313 & 255;
      $315 = (125924 + ($314<<2)|0);
      $316 = HEAP32[$315>>2]|0;
      $317 = (($316) + ($312)|0);
      $318 = HEAP8[$317>>0]|0;
      $319 = HEAP32[$42>>2]|0;
      $320 = $319 & $296;
      $321 = $299 >>> 16;
      $322 = $321 & 255;
      $323 = $320 >>> $322;
      $324 = $302 >>> 16;
      $325 = $324 & 255;
      $326 = (125924 + ($325<<2)|0);
      $327 = HEAP32[$326>>2]|0;
      $328 = (($327) + ($323)|0);
      $329 = HEAP8[$328>>0]|0;
      $Pixel$2 = $296;$sB$2$shrunk = $329;$sG$2$shrunk = $318;$sR$2$shrunk = $307;
      break;
     }
     case 2:  {
      $330 = HEAP16[$src$3>>1]|0;
      $331 = $330&65535;
      $332 = HEAP32[$38>>2]|0;
      $333 = $331 & $332;
      $334 = HEAP32[$39>>2]|0;
      $335 = $334 & 255;
      $336 = $333 >>> $335;
      $337 = HEAP32[$40>>2]|0;
      $338 = $337 & 255;
      $339 = (125924 + ($338<<2)|0);
      $340 = HEAP32[$339>>2]|0;
      $341 = (($340) + ($336)|0);
      $342 = HEAP8[$341>>0]|0;
      $343 = HEAP32[$41>>2]|0;
      $344 = $343 & $331;
      $345 = $334 >>> 8;
      $346 = $345 & 255;
      $347 = $344 >>> $346;
      $348 = $337 >>> 8;
      $349 = $348 & 255;
      $350 = (125924 + ($349<<2)|0);
      $351 = HEAP32[$350>>2]|0;
      $352 = (($351) + ($347)|0);
      $353 = HEAP8[$352>>0]|0;
      $354 = HEAP32[$42>>2]|0;
      $355 = $354 & $331;
      $356 = $334 >>> 16;
      $357 = $356 & 255;
      $358 = $355 >>> $357;
      $359 = $337 >>> 16;
      $360 = $359 & 255;
      $361 = (125924 + ($360<<2)|0);
      $362 = HEAP32[$361>>2]|0;
      $363 = (($362) + ($358)|0);
      $364 = HEAP8[$363>>0]|0;
      $Pixel$2 = $331;$sB$2$shrunk = $364;$sG$2$shrunk = $353;$sR$2$shrunk = $342;
      break;
     }
     case 3:  {
      $365 = HEAP32[$39>>2]|0;
      $div36 = $365 >>> 3;
      $366 = $div36 & 31;
      $367 = (($src$3) + ($366)|0);
      $368 = HEAP8[$367>>0]|0;
      $369 = $365 >>> 11;
      $div37 = $369 & 31;
      $370 = (($src$3) + ($div37)|0);
      $371 = HEAP8[$370>>0]|0;
      $372 = $365 >>> 19;
      $373 = $372 & 31;
      $374 = (($src$3) + ($373)|0);
      $375 = HEAP8[$374>>0]|0;
      $Pixel$2 = 0;$sB$2$shrunk = $375;$sG$2$shrunk = $371;$sR$2$shrunk = $368;
      break;
     }
     case 4:  {
      $376 = HEAP32[$src$3>>2]|0;
      $377 = HEAP32[$38>>2]|0;
      $378 = $377 & $376;
      $379 = HEAP32[$39>>2]|0;
      $380 = $379 & 255;
      $381 = $378 >>> $380;
      $382 = HEAP32[$40>>2]|0;
      $383 = $382 & 255;
      $384 = (125924 + ($383<<2)|0);
      $385 = HEAP32[$384>>2]|0;
      $386 = (($385) + ($381)|0);
      $387 = HEAP8[$386>>0]|0;
      $388 = HEAP32[$41>>2]|0;
      $389 = $388 & $376;
      $390 = $379 >>> 8;
      $391 = $390 & 255;
      $392 = $389 >>> $391;
      $393 = $382 >>> 8;
      $394 = $393 & 255;
      $395 = (125924 + ($394<<2)|0);
      $396 = HEAP32[$395>>2]|0;
      $397 = (($396) + ($392)|0);
      $398 = HEAP8[$397>>0]|0;
      $399 = HEAP32[$42>>2]|0;
      $400 = $399 & $376;
      $401 = $379 >>> 16;
      $402 = $401 & 255;
      $403 = $400 >>> $402;
      $404 = $382 >>> 16;
      $405 = $404 & 255;
      $406 = (125924 + ($405<<2)|0);
      $407 = HEAP32[$406>>2]|0;
      $408 = (($407) + ($403)|0);
      $409 = HEAP8[$408>>0]|0;
      $Pixel$2 = $376;$sB$2$shrunk = $409;$sG$2$shrunk = $398;$sR$2$shrunk = $387;
      break;
     }
     default: {
      $Pixel$2 = 0;$sB$2$shrunk = 0;$sG$2$shrunk = 0;$sR$2$shrunk = 0;
     }
     }
     $410 = $Pixel$2 & $19;
     $411 = ($410|0)==($23|0);
     if (!($411)) {
      $sB$2 = $sB$2$shrunk&255;
      $sG$2 = $sG$2$shrunk&255;
      $sR$2 = $sR$2$shrunk&255;
      $412 = $sR$2 & 224;
      $413 = $sG$2 >>> 5;
      $414 = $413 << 2;
      $415 = $414 | $412;
      $416 = $sB$2 >>> 6;
      $417 = $415 | $416;
      $418 = $417&255;
      HEAP8[$dst$3>>0] = $418;
     }
     $419 = ((($dst$3)) + 1|0);
     $420 = (($src$3) + ($22)|0);
     $dst$4 = $419;$n$3 = $n$2;$src$4 = $420;
     label = 31;
     continue;
    }
    else if ((label|0) == 31) {
     label = 0;
     switch ($22|0) {
     case 1:  {
      $421 = HEAP8[$src$4>>0]|0;
      $422 = $421&255;
      $423 = HEAP32[$38>>2]|0;
      $424 = $422 & $423;
      $425 = HEAP32[$39>>2]|0;
      $426 = $425 & 255;
      $427 = $424 >>> $426;
      $428 = HEAP32[$40>>2]|0;
      $429 = $428 & 255;
      $430 = (125924 + ($429<<2)|0);
      $431 = HEAP32[$430>>2]|0;
      $432 = (($431) + ($427)|0);
      $433 = HEAP8[$432>>0]|0;
      $434 = HEAP32[$41>>2]|0;
      $435 = $434 & $422;
      $436 = $425 >>> 8;
      $437 = $436 & 255;
      $438 = $435 >>> $437;
      $439 = $428 >>> 8;
      $440 = $439 & 255;
      $441 = (125924 + ($440<<2)|0);
      $442 = HEAP32[$441>>2]|0;
      $443 = (($442) + ($438)|0);
      $444 = HEAP8[$443>>0]|0;
      $445 = HEAP32[$42>>2]|0;
      $446 = $445 & $422;
      $447 = $425 >>> 16;
      $448 = $447 & 255;
      $449 = $446 >>> $448;
      $450 = $428 >>> 16;
      $451 = $450 & 255;
      $452 = (125924 + ($451<<2)|0);
      $453 = HEAP32[$452>>2]|0;
      $454 = (($453) + ($449)|0);
      $455 = HEAP8[$454>>0]|0;
      $Pixel$3 = $422;$sB$3$shrunk = $455;$sG$3$shrunk = $444;$sR$3$shrunk = $433;
      break;
     }
     case 2:  {
      $456 = HEAP16[$src$4>>1]|0;
      $457 = $456&65535;
      $458 = HEAP32[$38>>2]|0;
      $459 = $457 & $458;
      $460 = HEAP32[$39>>2]|0;
      $461 = $460 & 255;
      $462 = $459 >>> $461;
      $463 = HEAP32[$40>>2]|0;
      $464 = $463 & 255;
      $465 = (125924 + ($464<<2)|0);
      $466 = HEAP32[$465>>2]|0;
      $467 = (($466) + ($462)|0);
      $468 = HEAP8[$467>>0]|0;
      $469 = HEAP32[$41>>2]|0;
      $470 = $469 & $457;
      $471 = $460 >>> 8;
      $472 = $471 & 255;
      $473 = $470 >>> $472;
      $474 = $463 >>> 8;
      $475 = $474 & 255;
      $476 = (125924 + ($475<<2)|0);
      $477 = HEAP32[$476>>2]|0;
      $478 = (($477) + ($473)|0);
      $479 = HEAP8[$478>>0]|0;
      $480 = HEAP32[$42>>2]|0;
      $481 = $480 & $457;
      $482 = $460 >>> 16;
      $483 = $482 & 255;
      $484 = $481 >>> $483;
      $485 = $463 >>> 16;
      $486 = $485 & 255;
      $487 = (125924 + ($486<<2)|0);
      $488 = HEAP32[$487>>2]|0;
      $489 = (($488) + ($484)|0);
      $490 = HEAP8[$489>>0]|0;
      $Pixel$3 = $457;$sB$3$shrunk = $490;$sG$3$shrunk = $479;$sR$3$shrunk = $468;
      break;
     }
     case 3:  {
      $491 = HEAP32[$39>>2]|0;
      $div33 = $491 >>> 3;
      $492 = $div33 & 31;
      $493 = (($src$4) + ($492)|0);
      $494 = HEAP8[$493>>0]|0;
      $495 = $491 >>> 11;
      $div34 = $495 & 31;
      $496 = (($src$4) + ($div34)|0);
      $497 = HEAP8[$496>>0]|0;
      $498 = $491 >>> 19;
      $499 = $498 & 31;
      $500 = (($src$4) + ($499)|0);
      $501 = HEAP8[$500>>0]|0;
      $Pixel$3 = 0;$sB$3$shrunk = $501;$sG$3$shrunk = $497;$sR$3$shrunk = $494;
      break;
     }
     case 4:  {
      $502 = HEAP32[$src$4>>2]|0;
      $503 = HEAP32[$38>>2]|0;
      $504 = $503 & $502;
      $505 = HEAP32[$39>>2]|0;
      $506 = $505 & 255;
      $507 = $504 >>> $506;
      $508 = HEAP32[$40>>2]|0;
      $509 = $508 & 255;
      $510 = (125924 + ($509<<2)|0);
      $511 = HEAP32[$510>>2]|0;
      $512 = (($511) + ($507)|0);
      $513 = HEAP8[$512>>0]|0;
      $514 = HEAP32[$41>>2]|0;
      $515 = $514 & $502;
      $516 = $505 >>> 8;
      $517 = $516 & 255;
      $518 = $515 >>> $517;
      $519 = $508 >>> 8;
      $520 = $519 & 255;
      $521 = (125924 + ($520<<2)|0);
      $522 = HEAP32[$521>>2]|0;
      $523 = (($522) + ($518)|0);
      $524 = HEAP8[$523>>0]|0;
      $525 = HEAP32[$42>>2]|0;
      $526 = $525 & $502;
      $527 = $505 >>> 16;
      $528 = $527 & 255;
      $529 = $526 >>> $528;
      $530 = $508 >>> 16;
      $531 = $530 & 255;
      $532 = (125924 + ($531<<2)|0);
      $533 = HEAP32[$532>>2]|0;
      $534 = (($533) + ($529)|0);
      $535 = HEAP8[$534>>0]|0;
      $Pixel$3 = $502;$sB$3$shrunk = $535;$sG$3$shrunk = $524;$sR$3$shrunk = $513;
      break;
     }
     default: {
      $Pixel$3 = 0;$sB$3$shrunk = 0;$sG$3$shrunk = 0;$sR$3$shrunk = 0;
     }
     }
     $536 = $Pixel$3 & $19;
     $537 = ($536|0)==($23|0);
     if (!($537)) {
      $sB$3 = $sB$3$shrunk&255;
      $sG$3 = $sG$3$shrunk&255;
      $sR$3 = $sR$3$shrunk&255;
      $538 = $sR$3 & 224;
      $539 = $sG$3 >>> 5;
      $540 = $539 << 2;
      $541 = $540 | $538;
      $542 = $sB$3 >>> 6;
      $543 = $541 | $542;
      $544 = $543&255;
      HEAP8[$dst$4>>0] = $544;
     }
     $545 = ((($dst$4)) + 1|0);
     $546 = (($src$4) + ($22)|0);
     $dst$5 = $545;$n$4 = $n$3;$src$5 = $546;
     label = 39;
     continue;
    }
    else if ((label|0) == 39) {
     label = 0;
     switch ($22|0) {
     case 1:  {
      $547 = HEAP8[$src$5>>0]|0;
      $548 = $547&255;
      $549 = HEAP32[$38>>2]|0;
      $550 = $548 & $549;
      $551 = HEAP32[$39>>2]|0;
      $552 = $551 & 255;
      $553 = $550 >>> $552;
      $554 = HEAP32[$40>>2]|0;
      $555 = $554 & 255;
      $556 = (125924 + ($555<<2)|0);
      $557 = HEAP32[$556>>2]|0;
      $558 = (($557) + ($553)|0);
      $559 = HEAP8[$558>>0]|0;
      $560 = HEAP32[$41>>2]|0;
      $561 = $560 & $548;
      $562 = $551 >>> 8;
      $563 = $562 & 255;
      $564 = $561 >>> $563;
      $565 = $554 >>> 8;
      $566 = $565 & 255;
      $567 = (125924 + ($566<<2)|0);
      $568 = HEAP32[$567>>2]|0;
      $569 = (($568) + ($564)|0);
      $570 = HEAP8[$569>>0]|0;
      $571 = HEAP32[$42>>2]|0;
      $572 = $571 & $548;
      $573 = $551 >>> 16;
      $574 = $573 & 255;
      $575 = $572 >>> $574;
      $576 = $554 >>> 16;
      $577 = $576 & 255;
      $578 = (125924 + ($577<<2)|0);
      $579 = HEAP32[$578>>2]|0;
      $580 = (($579) + ($575)|0);
      $581 = HEAP8[$580>>0]|0;
      $Pixel$4 = $548;$sB$4$shrunk = $581;$sG$4$shrunk = $570;$sR$4$shrunk = $559;
      break;
     }
     case 2:  {
      $582 = HEAP16[$src$5>>1]|0;
      $583 = $582&65535;
      $584 = HEAP32[$38>>2]|0;
      $585 = $583 & $584;
      $586 = HEAP32[$39>>2]|0;
      $587 = $586 & 255;
      $588 = $585 >>> $587;
      $589 = HEAP32[$40>>2]|0;
      $590 = $589 & 255;
      $591 = (125924 + ($590<<2)|0);
      $592 = HEAP32[$591>>2]|0;
      $593 = (($592) + ($588)|0);
      $594 = HEAP8[$593>>0]|0;
      $595 = HEAP32[$41>>2]|0;
      $596 = $595 & $583;
      $597 = $586 >>> 8;
      $598 = $597 & 255;
      $599 = $596 >>> $598;
      $600 = $589 >>> 8;
      $601 = $600 & 255;
      $602 = (125924 + ($601<<2)|0);
      $603 = HEAP32[$602>>2]|0;
      $604 = (($603) + ($599)|0);
      $605 = HEAP8[$604>>0]|0;
      $606 = HEAP32[$42>>2]|0;
      $607 = $606 & $583;
      $608 = $586 >>> 16;
      $609 = $608 & 255;
      $610 = $607 >>> $609;
      $611 = $589 >>> 16;
      $612 = $611 & 255;
      $613 = (125924 + ($612<<2)|0);
      $614 = HEAP32[$613>>2]|0;
      $615 = (($614) + ($610)|0);
      $616 = HEAP8[$615>>0]|0;
      $Pixel$4 = $583;$sB$4$shrunk = $616;$sG$4$shrunk = $605;$sR$4$shrunk = $594;
      break;
     }
     case 3:  {
      $617 = HEAP32[$39>>2]|0;
      $div30 = $617 >>> 3;
      $618 = $div30 & 31;
      $619 = (($src$5) + ($618)|0);
      $620 = HEAP8[$619>>0]|0;
      $621 = $617 >>> 11;
      $div31 = $621 & 31;
      $622 = (($src$5) + ($div31)|0);
      $623 = HEAP8[$622>>0]|0;
      $624 = $617 >>> 19;
      $625 = $624 & 31;
      $626 = (($src$5) + ($625)|0);
      $627 = HEAP8[$626>>0]|0;
      $Pixel$4 = 0;$sB$4$shrunk = $627;$sG$4$shrunk = $623;$sR$4$shrunk = $620;
      break;
     }
     case 4:  {
      $628 = HEAP32[$src$5>>2]|0;
      $629 = HEAP32[$38>>2]|0;
      $630 = $629 & $628;
      $631 = HEAP32[$39>>2]|0;
      $632 = $631 & 255;
      $633 = $630 >>> $632;
      $634 = HEAP32[$40>>2]|0;
      $635 = $634 & 255;
      $636 = (125924 + ($635<<2)|0);
      $637 = HEAP32[$636>>2]|0;
      $638 = (($637) + ($633)|0);
      $639 = HEAP8[$638>>0]|0;
      $640 = HEAP32[$41>>2]|0;
      $641 = $640 & $628;
      $642 = $631 >>> 8;
      $643 = $642 & 255;
      $644 = $641 >>> $643;
      $645 = $634 >>> 8;
      $646 = $645 & 255;
      $647 = (125924 + ($646<<2)|0);
      $648 = HEAP32[$647>>2]|0;
      $649 = (($648) + ($644)|0);
      $650 = HEAP8[$649>>0]|0;
      $651 = HEAP32[$42>>2]|0;
      $652 = $651 & $628;
      $653 = $631 >>> 16;
      $654 = $653 & 255;
      $655 = $652 >>> $654;
      $656 = $634 >>> 16;
      $657 = $656 & 255;
      $658 = (125924 + ($657<<2)|0);
      $659 = HEAP32[$658>>2]|0;
      $660 = (($659) + ($655)|0);
      $661 = HEAP8[$660>>0]|0;
      $Pixel$4 = $628;$sB$4$shrunk = $661;$sG$4$shrunk = $650;$sR$4$shrunk = $639;
      break;
     }
     default: {
      $Pixel$4 = 0;$sB$4$shrunk = 0;$sG$4$shrunk = 0;$sR$4$shrunk = 0;
     }
     }
     $662 = $Pixel$4 & $19;
     $663 = ($662|0)==($23|0);
     if (!($663)) {
      $sB$4 = $sB$4$shrunk&255;
      $sG$4 = $sG$4$shrunk&255;
      $sR$4 = $sR$4$shrunk&255;
      $664 = $sR$4 & 224;
      $665 = $sG$4 >>> 5;
      $666 = $665 << 2;
      $667 = $666 | $664;
      $668 = $sB$4 >>> 6;
      $669 = $667 | $668;
      $670 = $669&255;
      HEAP8[$dst$5>>0] = $670;
     }
     $671 = ((($dst$5)) + 1|0);
     $672 = (($src$5) + ($22)|0);
     $dst$6 = $671;$n$5 = $n$4;$src$6 = $672;
     label = 47;
     continue;
    }
    else if ((label|0) == 47) {
     label = 0;
     switch ($22|0) {
     case 1:  {
      $673 = HEAP8[$src$6>>0]|0;
      $674 = $673&255;
      $675 = HEAP32[$38>>2]|0;
      $676 = $674 & $675;
      $677 = HEAP32[$39>>2]|0;
      $678 = $677 & 255;
      $679 = $676 >>> $678;
      $680 = HEAP32[$40>>2]|0;
      $681 = $680 & 255;
      $682 = (125924 + ($681<<2)|0);
      $683 = HEAP32[$682>>2]|0;
      $684 = (($683) + ($679)|0);
      $685 = HEAP8[$684>>0]|0;
      $686 = HEAP32[$41>>2]|0;
      $687 = $686 & $674;
      $688 = $677 >>> 8;
      $689 = $688 & 255;
      $690 = $687 >>> $689;
      $691 = $680 >>> 8;
      $692 = $691 & 255;
      $693 = (125924 + ($692<<2)|0);
      $694 = HEAP32[$693>>2]|0;
      $695 = (($694) + ($690)|0);
      $696 = HEAP8[$695>>0]|0;
      $697 = HEAP32[$42>>2]|0;
      $698 = $697 & $674;
      $699 = $677 >>> 16;
      $700 = $699 & 255;
      $701 = $698 >>> $700;
      $702 = $680 >>> 16;
      $703 = $702 & 255;
      $704 = (125924 + ($703<<2)|0);
      $705 = HEAP32[$704>>2]|0;
      $706 = (($705) + ($701)|0);
      $707 = HEAP8[$706>>0]|0;
      $Pixel$5 = $674;$sB$5$shrunk = $707;$sG$5$shrunk = $696;$sR$5$shrunk = $685;
      break;
     }
     case 2:  {
      $708 = HEAP16[$src$6>>1]|0;
      $709 = $708&65535;
      $710 = HEAP32[$38>>2]|0;
      $711 = $709 & $710;
      $712 = HEAP32[$39>>2]|0;
      $713 = $712 & 255;
      $714 = $711 >>> $713;
      $715 = HEAP32[$40>>2]|0;
      $716 = $715 & 255;
      $717 = (125924 + ($716<<2)|0);
      $718 = HEAP32[$717>>2]|0;
      $719 = (($718) + ($714)|0);
      $720 = HEAP8[$719>>0]|0;
      $721 = HEAP32[$41>>2]|0;
      $722 = $721 & $709;
      $723 = $712 >>> 8;
      $724 = $723 & 255;
      $725 = $722 >>> $724;
      $726 = $715 >>> 8;
      $727 = $726 & 255;
      $728 = (125924 + ($727<<2)|0);
      $729 = HEAP32[$728>>2]|0;
      $730 = (($729) + ($725)|0);
      $731 = HEAP8[$730>>0]|0;
      $732 = HEAP32[$42>>2]|0;
      $733 = $732 & $709;
      $734 = $712 >>> 16;
      $735 = $734 & 255;
      $736 = $733 >>> $735;
      $737 = $715 >>> 16;
      $738 = $737 & 255;
      $739 = (125924 + ($738<<2)|0);
      $740 = HEAP32[$739>>2]|0;
      $741 = (($740) + ($736)|0);
      $742 = HEAP8[$741>>0]|0;
      $Pixel$5 = $709;$sB$5$shrunk = $742;$sG$5$shrunk = $731;$sR$5$shrunk = $720;
      break;
     }
     case 3:  {
      $743 = HEAP32[$39>>2]|0;
      $div27 = $743 >>> 3;
      $744 = $div27 & 31;
      $745 = (($src$6) + ($744)|0);
      $746 = HEAP8[$745>>0]|0;
      $747 = $743 >>> 11;
      $div28 = $747 & 31;
      $748 = (($src$6) + ($div28)|0);
      $749 = HEAP8[$748>>0]|0;
      $750 = $743 >>> 19;
      $751 = $750 & 31;
      $752 = (($src$6) + ($751)|0);
      $753 = HEAP8[$752>>0]|0;
      $Pixel$5 = 0;$sB$5$shrunk = $753;$sG$5$shrunk = $749;$sR$5$shrunk = $746;
      break;
     }
     case 4:  {
      $754 = HEAP32[$src$6>>2]|0;
      $755 = HEAP32[$38>>2]|0;
      $756 = $755 & $754;
      $757 = HEAP32[$39>>2]|0;
      $758 = $757 & 255;
      $759 = $756 >>> $758;
      $760 = HEAP32[$40>>2]|0;
      $761 = $760 & 255;
      $762 = (125924 + ($761<<2)|0);
      $763 = HEAP32[$762>>2]|0;
      $764 = (($763) + ($759)|0);
      $765 = HEAP8[$764>>0]|0;
      $766 = HEAP32[$41>>2]|0;
      $767 = $766 & $754;
      $768 = $757 >>> 8;
      $769 = $768 & 255;
      $770 = $767 >>> $769;
      $771 = $760 >>> 8;
      $772 = $771 & 255;
      $773 = (125924 + ($772<<2)|0);
      $774 = HEAP32[$773>>2]|0;
      $775 = (($774) + ($770)|0);
      $776 = HEAP8[$775>>0]|0;
      $777 = HEAP32[$42>>2]|0;
      $778 = $777 & $754;
      $779 = $757 >>> 16;
      $780 = $779 & 255;
      $781 = $778 >>> $780;
      $782 = $760 >>> 16;
      $783 = $782 & 255;
      $784 = (125924 + ($783<<2)|0);
      $785 = HEAP32[$784>>2]|0;
      $786 = (($785) + ($781)|0);
      $787 = HEAP8[$786>>0]|0;
      $Pixel$5 = $754;$sB$5$shrunk = $787;$sG$5$shrunk = $776;$sR$5$shrunk = $765;
      break;
     }
     default: {
      $Pixel$5 = 0;$sB$5$shrunk = 0;$sG$5$shrunk = 0;$sR$5$shrunk = 0;
     }
     }
     $788 = $Pixel$5 & $19;
     $789 = ($788|0)==($23|0);
     if (!($789)) {
      $sB$5 = $sB$5$shrunk&255;
      $sG$5 = $sG$5$shrunk&255;
      $sR$5 = $sR$5$shrunk&255;
      $790 = $sR$5 & 224;
      $791 = $sG$5 >>> 5;
      $792 = $791 << 2;
      $793 = $792 | $790;
      $794 = $sB$5 >>> 6;
      $795 = $793 | $794;
      $796 = $795&255;
      HEAP8[$dst$6>>0] = $796;
     }
     $797 = ((($dst$6)) + 1|0);
     $798 = (($src$6) + ($22)|0);
     $dst$7 = $797;$n$6 = $n$5;$src$7 = $798;
     label = 55;
     continue;
    }
    else if ((label|0) == 55) {
     label = 0;
     switch ($22|0) {
     case 1:  {
      $799 = HEAP8[$src$7>>0]|0;
      $800 = $799&255;
      $801 = HEAP32[$38>>2]|0;
      $802 = $800 & $801;
      $803 = HEAP32[$39>>2]|0;
      $804 = $803 & 255;
      $805 = $802 >>> $804;
      $806 = HEAP32[$40>>2]|0;
      $807 = $806 & 255;
      $808 = (125924 + ($807<<2)|0);
      $809 = HEAP32[$808>>2]|0;
      $810 = (($809) + ($805)|0);
      $811 = HEAP8[$810>>0]|0;
      $812 = HEAP32[$41>>2]|0;
      $813 = $812 & $800;
      $814 = $803 >>> 8;
      $815 = $814 & 255;
      $816 = $813 >>> $815;
      $817 = $806 >>> 8;
      $818 = $817 & 255;
      $819 = (125924 + ($818<<2)|0);
      $820 = HEAP32[$819>>2]|0;
      $821 = (($820) + ($816)|0);
      $822 = HEAP8[$821>>0]|0;
      $823 = HEAP32[$42>>2]|0;
      $824 = $823 & $800;
      $825 = $803 >>> 16;
      $826 = $825 & 255;
      $827 = $824 >>> $826;
      $828 = $806 >>> 16;
      $829 = $828 & 255;
      $830 = (125924 + ($829<<2)|0);
      $831 = HEAP32[$830>>2]|0;
      $832 = (($831) + ($827)|0);
      $833 = HEAP8[$832>>0]|0;
      $Pixel$6 = $800;$sB$6$shrunk = $833;$sG$6$shrunk = $822;$sR$6$shrunk = $811;
      break;
     }
     case 2:  {
      $834 = HEAP16[$src$7>>1]|0;
      $835 = $834&65535;
      $836 = HEAP32[$38>>2]|0;
      $837 = $835 & $836;
      $838 = HEAP32[$39>>2]|0;
      $839 = $838 & 255;
      $840 = $837 >>> $839;
      $841 = HEAP32[$40>>2]|0;
      $842 = $841 & 255;
      $843 = (125924 + ($842<<2)|0);
      $844 = HEAP32[$843>>2]|0;
      $845 = (($844) + ($840)|0);
      $846 = HEAP8[$845>>0]|0;
      $847 = HEAP32[$41>>2]|0;
      $848 = $847 & $835;
      $849 = $838 >>> 8;
      $850 = $849 & 255;
      $851 = $848 >>> $850;
      $852 = $841 >>> 8;
      $853 = $852 & 255;
      $854 = (125924 + ($853<<2)|0);
      $855 = HEAP32[$854>>2]|0;
      $856 = (($855) + ($851)|0);
      $857 = HEAP8[$856>>0]|0;
      $858 = HEAP32[$42>>2]|0;
      $859 = $858 & $835;
      $860 = $838 >>> 16;
      $861 = $860 & 255;
      $862 = $859 >>> $861;
      $863 = $841 >>> 16;
      $864 = $863 & 255;
      $865 = (125924 + ($864<<2)|0);
      $866 = HEAP32[$865>>2]|0;
      $867 = (($866) + ($862)|0);
      $868 = HEAP8[$867>>0]|0;
      $Pixel$6 = $835;$sB$6$shrunk = $868;$sG$6$shrunk = $857;$sR$6$shrunk = $846;
      break;
     }
     case 3:  {
      $869 = HEAP32[$39>>2]|0;
      $div24 = $869 >>> 3;
      $870 = $div24 & 31;
      $871 = (($src$7) + ($870)|0);
      $872 = HEAP8[$871>>0]|0;
      $873 = $869 >>> 11;
      $div25 = $873 & 31;
      $874 = (($src$7) + ($div25)|0);
      $875 = HEAP8[$874>>0]|0;
      $876 = $869 >>> 19;
      $877 = $876 & 31;
      $878 = (($src$7) + ($877)|0);
      $879 = HEAP8[$878>>0]|0;
      $Pixel$6 = 0;$sB$6$shrunk = $879;$sG$6$shrunk = $875;$sR$6$shrunk = $872;
      break;
     }
     case 4:  {
      $880 = HEAP32[$src$7>>2]|0;
      $881 = HEAP32[$38>>2]|0;
      $882 = $881 & $880;
      $883 = HEAP32[$39>>2]|0;
      $884 = $883 & 255;
      $885 = $882 >>> $884;
      $886 = HEAP32[$40>>2]|0;
      $887 = $886 & 255;
      $888 = (125924 + ($887<<2)|0);
      $889 = HEAP32[$888>>2]|0;
      $890 = (($889) + ($885)|0);
      $891 = HEAP8[$890>>0]|0;
      $892 = HEAP32[$41>>2]|0;
      $893 = $892 & $880;
      $894 = $883 >>> 8;
      $895 = $894 & 255;
      $896 = $893 >>> $895;
      $897 = $886 >>> 8;
      $898 = $897 & 255;
      $899 = (125924 + ($898<<2)|0);
      $900 = HEAP32[$899>>2]|0;
      $901 = (($900) + ($896)|0);
      $902 = HEAP8[$901>>0]|0;
      $903 = HEAP32[$42>>2]|0;
      $904 = $903 & $880;
      $905 = $883 >>> 16;
      $906 = $905 & 255;
      $907 = $904 >>> $906;
      $908 = $886 >>> 16;
      $909 = $908 & 255;
      $910 = (125924 + ($909<<2)|0);
      $911 = HEAP32[$910>>2]|0;
      $912 = (($911) + ($907)|0);
      $913 = HEAP8[$912>>0]|0;
      $Pixel$6 = $880;$sB$6$shrunk = $913;$sG$6$shrunk = $902;$sR$6$shrunk = $891;
      break;
     }
     default: {
      $Pixel$6 = 0;$sB$6$shrunk = 0;$sG$6$shrunk = 0;$sR$6$shrunk = 0;
     }
     }
     $914 = $Pixel$6 & $19;
     $915 = ($914|0)==($23|0);
     if (!($915)) {
      $sB$6 = $sB$6$shrunk&255;
      $sG$6 = $sG$6$shrunk&255;
      $sR$6 = $sR$6$shrunk&255;
      $916 = $sR$6 & 224;
      $917 = $sG$6 >>> 5;
      $918 = $917 << 2;
      $919 = $918 | $916;
      $920 = $sB$6 >>> 6;
      $921 = $919 | $920;
      $922 = $921&255;
      HEAP8[$dst$7>>0] = $922;
     }
     $923 = ((($dst$7)) + 1|0);
     $924 = (($src$7) + ($22)|0);
     $dst$8 = $923;$n$7 = $n$6;$src$8 = $924;
     label = 63;
     continue;
    }
    else if ((label|0) == 63) {
     label = 0;
     switch ($22|0) {
     case 1:  {
      $925 = HEAP8[$src$8>>0]|0;
      $926 = $925&255;
      $927 = HEAP32[$38>>2]|0;
      $928 = $926 & $927;
      $929 = HEAP32[$39>>2]|0;
      $930 = $929 & 255;
      $931 = $928 >>> $930;
      $932 = HEAP32[$40>>2]|0;
      $933 = $932 & 255;
      $934 = (125924 + ($933<<2)|0);
      $935 = HEAP32[$934>>2]|0;
      $936 = (($935) + ($931)|0);
      $937 = HEAP8[$936>>0]|0;
      $938 = HEAP32[$41>>2]|0;
      $939 = $938 & $926;
      $940 = $929 >>> 8;
      $941 = $940 & 255;
      $942 = $939 >>> $941;
      $943 = $932 >>> 8;
      $944 = $943 & 255;
      $945 = (125924 + ($944<<2)|0);
      $946 = HEAP32[$945>>2]|0;
      $947 = (($946) + ($942)|0);
      $948 = HEAP8[$947>>0]|0;
      $949 = HEAP32[$42>>2]|0;
      $950 = $949 & $926;
      $951 = $929 >>> 16;
      $952 = $951 & 255;
      $953 = $950 >>> $952;
      $954 = $932 >>> 16;
      $955 = $954 & 255;
      $956 = (125924 + ($955<<2)|0);
      $957 = HEAP32[$956>>2]|0;
      $958 = (($957) + ($953)|0);
      $959 = HEAP8[$958>>0]|0;
      $Pixel$7 = $926;$sB$7$shrunk = $959;$sG$7$shrunk = $948;$sR$7$shrunk = $937;
      break;
     }
     case 2:  {
      $960 = HEAP16[$src$8>>1]|0;
      $961 = $960&65535;
      $962 = HEAP32[$38>>2]|0;
      $963 = $961 & $962;
      $964 = HEAP32[$39>>2]|0;
      $965 = $964 & 255;
      $966 = $963 >>> $965;
      $967 = HEAP32[$40>>2]|0;
      $968 = $967 & 255;
      $969 = (125924 + ($968<<2)|0);
      $970 = HEAP32[$969>>2]|0;
      $971 = (($970) + ($966)|0);
      $972 = HEAP8[$971>>0]|0;
      $973 = HEAP32[$41>>2]|0;
      $974 = $973 & $961;
      $975 = $964 >>> 8;
      $976 = $975 & 255;
      $977 = $974 >>> $976;
      $978 = $967 >>> 8;
      $979 = $978 & 255;
      $980 = (125924 + ($979<<2)|0);
      $981 = HEAP32[$980>>2]|0;
      $982 = (($981) + ($977)|0);
      $983 = HEAP8[$982>>0]|0;
      $984 = HEAP32[$42>>2]|0;
      $985 = $984 & $961;
      $986 = $964 >>> 16;
      $987 = $986 & 255;
      $988 = $985 >>> $987;
      $989 = $967 >>> 16;
      $990 = $989 & 255;
      $991 = (125924 + ($990<<2)|0);
      $992 = HEAP32[$991>>2]|0;
      $993 = (($992) + ($988)|0);
      $994 = HEAP8[$993>>0]|0;
      $Pixel$7 = $961;$sB$7$shrunk = $994;$sG$7$shrunk = $983;$sR$7$shrunk = $972;
      break;
     }
     case 3:  {
      $995 = HEAP32[$39>>2]|0;
      $div45 = $995 >>> 3;
      $996 = $div45 & 31;
      $997 = (($src$8) + ($996)|0);
      $998 = HEAP8[$997>>0]|0;
      $999 = $995 >>> 11;
      $div46 = $999 & 31;
      $1000 = (($src$8) + ($div46)|0);
      $1001 = HEAP8[$1000>>0]|0;
      $1002 = $995 >>> 19;
      $1003 = $1002 & 31;
      $1004 = (($src$8) + ($1003)|0);
      $1005 = HEAP8[$1004>>0]|0;
      $Pixel$7 = 0;$sB$7$shrunk = $1005;$sG$7$shrunk = $1001;$sR$7$shrunk = $998;
      break;
     }
     case 4:  {
      $1006 = HEAP32[$src$8>>2]|0;
      $1007 = HEAP32[$38>>2]|0;
      $1008 = $1007 & $1006;
      $1009 = HEAP32[$39>>2]|0;
      $1010 = $1009 & 255;
      $1011 = $1008 >>> $1010;
      $1012 = HEAP32[$40>>2]|0;
      $1013 = $1012 & 255;
      $1014 = (125924 + ($1013<<2)|0);
      $1015 = HEAP32[$1014>>2]|0;
      $1016 = (($1015) + ($1011)|0);
      $1017 = HEAP8[$1016>>0]|0;
      $1018 = HEAP32[$41>>2]|0;
      $1019 = $1018 & $1006;
      $1020 = $1009 >>> 8;
      $1021 = $1020 & 255;
      $1022 = $1019 >>> $1021;
      $1023 = $1012 >>> 8;
      $1024 = $1023 & 255;
      $1025 = (125924 + ($1024<<2)|0);
      $1026 = HEAP32[$1025>>2]|0;
      $1027 = (($1026) + ($1022)|0);
      $1028 = HEAP8[$1027>>0]|0;
      $1029 = HEAP32[$42>>2]|0;
      $1030 = $1029 & $1006;
      $1031 = $1009 >>> 16;
      $1032 = $1031 & 255;
      $1033 = $1030 >>> $1032;
      $1034 = $1012 >>> 16;
      $1035 = $1034 & 255;
      $1036 = (125924 + ($1035<<2)|0);
      $1037 = HEAP32[$1036>>2]|0;
      $1038 = (($1037) + ($1033)|0);
      $1039 = HEAP8[$1038>>0]|0;
      $Pixel$7 = $1006;$sB$7$shrunk = $1039;$sG$7$shrunk = $1028;$sR$7$shrunk = $1017;
      break;
     }
     default: {
      $Pixel$7 = 0;$sB$7$shrunk = 0;$sG$7$shrunk = 0;$sR$7$shrunk = 0;
     }
     }
     $1040 = $Pixel$7 & $19;
     $1041 = ($1040|0)==($23|0);
     if (!($1041)) {
      $sB$7 = $sB$7$shrunk&255;
      $sG$7 = $sG$7$shrunk&255;
      $sR$7 = $sR$7$shrunk&255;
      $1042 = $sR$7 & 224;
      $1043 = $sG$7 >>> 5;
      $1044 = $1043 << 2;
      $1045 = $1044 | $1042;
      $1046 = $sB$7 >>> 6;
      $1047 = $1045 | $1046;
      $1048 = $1047&255;
      HEAP8[$dst$8>>0] = $1048;
     }
     $1049 = ((($dst$8)) + 1|0);
     $1050 = (($src$8) + ($22)|0);
     $1051 = (($n$7) + -1)|0;
     $1052 = ($n$7|0)>(1);
     if ($1052) {
      $dst$1 = $1049;$n$0 = $1051;$src$1 = $1050;
      label = 7;
      continue;
     } else {
      break;
     }
    }
   }
   $1053 = (($1050) + ($6)|0);
   $1054 = (($1049) + ($10)|0);
   $1055 = (($1056) + -1)|0;
   $1057 = ($1056|0)==(0);
   if ($1057) {
    label = 140;
    break;
   } else {
    $1056 = $1055;$dst$050 = $1054;$src$051 = $1053;
   }
  }
  if ((label|0) == 71) {
   // unreachable;
  }
  else if ((label|0) == 140) {
   return;
  }
 } else {
  if ($26) {
   return;
  }
  $27 = (($1) + 7)|0;
  $28 = (($27|0) / 8)&-1;
  $29 = $1 & 7;
  $30 = ((($12)) + 12|0);
  $31 = ((($12)) + 32|0);
  $32 = ((($12)) + 28|0);
  $33 = ((($12)) + 16|0);
  $34 = ((($12)) + 20|0);
  $2079 = $25;$dst$952 = $8;$src$953 = $4;
  L98: while(1) {
   switch ($29|0) {
   case 0:  {
    $dst$10 = $dst$952;$n1$0 = $28;$src$10 = $src$953;
    label = 74;
    break;
   }
   case 7:  {
    $dst$11 = $dst$952;$n1$1 = $28;$src$11 = $src$953;
    label = 82;
    break;
   }
   case 6:  {
    $dst$12 = $dst$952;$n1$2 = $28;$src$12 = $src$953;
    label = 90;
    break;
   }
   case 5:  {
    $dst$13 = $dst$952;$n1$3 = $28;$src$13 = $src$953;
    label = 98;
    break;
   }
   case 4:  {
    $dst$14 = $dst$952;$n1$4 = $28;$src$14 = $src$953;
    label = 106;
    break;
   }
   case 3:  {
    $dst$15 = $dst$952;$n1$5 = $28;$src$15 = $src$953;
    label = 114;
    break;
   }
   case 2:  {
    $dst$16 = $dst$952;$n1$6 = $28;$src$16 = $src$953;
    label = 122;
    break;
   }
   case 1:  {
    $dst$17 = $dst$952;$n1$7 = $28;$src$17 = $src$953;
    label = 130;
    break;
   }
   default: {
    label = 138;
    break L98;
   }
   }
   while(1) {
    if ((label|0) == 74) {
     label = 0;
     switch ($22|0) {
     case 1:  {
      $1058 = HEAP8[$src$10>>0]|0;
      $1059 = $1058&255;
      $1060 = HEAP32[$30>>2]|0;
      $1061 = $1059 & $1060;
      $1062 = HEAP32[$31>>2]|0;
      $1063 = $1062 & 255;
      $1064 = $1061 >>> $1063;
      $1065 = HEAP32[$32>>2]|0;
      $1066 = $1065 & 255;
      $1067 = (125924 + ($1066<<2)|0);
      $1068 = HEAP32[$1067>>2]|0;
      $1069 = (($1068) + ($1064)|0);
      $1070 = HEAP8[$1069>>0]|0;
      $1071 = HEAP32[$33>>2]|0;
      $1072 = $1071 & $1059;
      $1073 = $1062 >>> 8;
      $1074 = $1073 & 255;
      $1075 = $1072 >>> $1074;
      $1076 = $1065 >>> 8;
      $1077 = $1076 & 255;
      $1078 = (125924 + ($1077<<2)|0);
      $1079 = HEAP32[$1078>>2]|0;
      $1080 = (($1079) + ($1075)|0);
      $1081 = HEAP8[$1080>>0]|0;
      $1082 = HEAP32[$34>>2]|0;
      $1083 = $1082 & $1059;
      $1084 = $1062 >>> 16;
      $1085 = $1084 & 255;
      $1086 = $1083 >>> $1085;
      $1087 = $1065 >>> 16;
      $1088 = $1087 & 255;
      $1089 = (125924 + ($1088<<2)|0);
      $1090 = HEAP32[$1089>>2]|0;
      $1091 = (($1090) + ($1086)|0);
      $1092 = HEAP8[$1091>>0]|0;
      $Pixel$8 = $1059;$sB$8$shrunk = $1092;$sG$8$shrunk = $1081;$sR$8$shrunk = $1070;
      break;
     }
     case 2:  {
      $1093 = HEAP16[$src$10>>1]|0;
      $1094 = $1093&65535;
      $1095 = HEAP32[$30>>2]|0;
      $1096 = $1094 & $1095;
      $1097 = HEAP32[$31>>2]|0;
      $1098 = $1097 & 255;
      $1099 = $1096 >>> $1098;
      $1100 = HEAP32[$32>>2]|0;
      $1101 = $1100 & 255;
      $1102 = (125924 + ($1101<<2)|0);
      $1103 = HEAP32[$1102>>2]|0;
      $1104 = (($1103) + ($1099)|0);
      $1105 = HEAP8[$1104>>0]|0;
      $1106 = HEAP32[$33>>2]|0;
      $1107 = $1106 & $1094;
      $1108 = $1097 >>> 8;
      $1109 = $1108 & 255;
      $1110 = $1107 >>> $1109;
      $1111 = $1100 >>> 8;
      $1112 = $1111 & 255;
      $1113 = (125924 + ($1112<<2)|0);
      $1114 = HEAP32[$1113>>2]|0;
      $1115 = (($1114) + ($1110)|0);
      $1116 = HEAP8[$1115>>0]|0;
      $1117 = HEAP32[$34>>2]|0;
      $1118 = $1117 & $1094;
      $1119 = $1097 >>> 16;
      $1120 = $1119 & 255;
      $1121 = $1118 >>> $1120;
      $1122 = $1100 >>> 16;
      $1123 = $1122 & 255;
      $1124 = (125924 + ($1123<<2)|0);
      $1125 = HEAP32[$1124>>2]|0;
      $1126 = (($1125) + ($1121)|0);
      $1127 = HEAP8[$1126>>0]|0;
      $Pixel$8 = $1094;$sB$8$shrunk = $1127;$sG$8$shrunk = $1116;$sR$8$shrunk = $1105;
      break;
     }
     case 3:  {
      $1128 = HEAP32[$31>>2]|0;
      $div18 = $1128 >>> 3;
      $1129 = $div18 & 31;
      $1130 = (($src$10) + ($1129)|0);
      $1131 = HEAP8[$1130>>0]|0;
      $1132 = $1128 >>> 11;
      $div19 = $1132 & 31;
      $1133 = (($src$10) + ($div19)|0);
      $1134 = HEAP8[$1133>>0]|0;
      $1135 = $1128 >>> 19;
      $1136 = $1135 & 31;
      $1137 = (($src$10) + ($1136)|0);
      $1138 = HEAP8[$1137>>0]|0;
      $Pixel$8 = 0;$sB$8$shrunk = $1138;$sG$8$shrunk = $1134;$sR$8$shrunk = $1131;
      break;
     }
     case 4:  {
      $1139 = HEAP32[$src$10>>2]|0;
      $1140 = HEAP32[$30>>2]|0;
      $1141 = $1140 & $1139;
      $1142 = HEAP32[$31>>2]|0;
      $1143 = $1142 & 255;
      $1144 = $1141 >>> $1143;
      $1145 = HEAP32[$32>>2]|0;
      $1146 = $1145 & 255;
      $1147 = (125924 + ($1146<<2)|0);
      $1148 = HEAP32[$1147>>2]|0;
      $1149 = (($1148) + ($1144)|0);
      $1150 = HEAP8[$1149>>0]|0;
      $1151 = HEAP32[$33>>2]|0;
      $1152 = $1151 & $1139;
      $1153 = $1142 >>> 8;
      $1154 = $1153 & 255;
      $1155 = $1152 >>> $1154;
      $1156 = $1145 >>> 8;
      $1157 = $1156 & 255;
      $1158 = (125924 + ($1157<<2)|0);
      $1159 = HEAP32[$1158>>2]|0;
      $1160 = (($1159) + ($1155)|0);
      $1161 = HEAP8[$1160>>0]|0;
      $1162 = HEAP32[$34>>2]|0;
      $1163 = $1162 & $1139;
      $1164 = $1142 >>> 16;
      $1165 = $1164 & 255;
      $1166 = $1163 >>> $1165;
      $1167 = $1145 >>> 16;
      $1168 = $1167 & 255;
      $1169 = (125924 + ($1168<<2)|0);
      $1170 = HEAP32[$1169>>2]|0;
      $1171 = (($1170) + ($1166)|0);
      $1172 = HEAP8[$1171>>0]|0;
      $Pixel$8 = $1139;$sB$8$shrunk = $1172;$sG$8$shrunk = $1161;$sR$8$shrunk = $1150;
      break;
     }
     default: {
      $Pixel$8 = 0;$sB$8$shrunk = 0;$sG$8$shrunk = 0;$sR$8$shrunk = 0;
     }
     }
     $1173 = $Pixel$8 & $19;
     $1174 = ($1173|0)==($23|0);
     if (!($1174)) {
      $sB$8 = $sB$8$shrunk&255;
      $sG$8 = $sG$8$shrunk&255;
      $sR$8 = $sR$8$shrunk&255;
      $1175 = $sR$8 & 224;
      $1176 = $sG$8 >>> 5;
      $1177 = $1176 << 2;
      $1178 = $1177 | $1175;
      $1179 = $sB$8 >>> 6;
      $1180 = $1178 | $1179;
      $1181 = (($14) + ($1180)|0);
      $1182 = HEAP8[$1181>>0]|0;
      HEAP8[$dst$10>>0] = $1182;
     }
     $1183 = ((($dst$10)) + 1|0);
     $1184 = (($src$10) + ($22)|0);
     $dst$11 = $1183;$n1$1 = $n1$0;$src$11 = $1184;
     label = 82;
     continue;
    }
    else if ((label|0) == 82) {
     label = 0;
     switch ($22|0) {
     case 1:  {
      $1185 = HEAP8[$src$11>>0]|0;
      $1186 = $1185&255;
      $1187 = HEAP32[$30>>2]|0;
      $1188 = $1186 & $1187;
      $1189 = HEAP32[$31>>2]|0;
      $1190 = $1189 & 255;
      $1191 = $1188 >>> $1190;
      $1192 = HEAP32[$32>>2]|0;
      $1193 = $1192 & 255;
      $1194 = (125924 + ($1193<<2)|0);
      $1195 = HEAP32[$1194>>2]|0;
      $1196 = (($1195) + ($1191)|0);
      $1197 = HEAP8[$1196>>0]|0;
      $1198 = HEAP32[$33>>2]|0;
      $1199 = $1198 & $1186;
      $1200 = $1189 >>> 8;
      $1201 = $1200 & 255;
      $1202 = $1199 >>> $1201;
      $1203 = $1192 >>> 8;
      $1204 = $1203 & 255;
      $1205 = (125924 + ($1204<<2)|0);
      $1206 = HEAP32[$1205>>2]|0;
      $1207 = (($1206) + ($1202)|0);
      $1208 = HEAP8[$1207>>0]|0;
      $1209 = HEAP32[$34>>2]|0;
      $1210 = $1209 & $1186;
      $1211 = $1189 >>> 16;
      $1212 = $1211 & 255;
      $1213 = $1210 >>> $1212;
      $1214 = $1192 >>> 16;
      $1215 = $1214 & 255;
      $1216 = (125924 + ($1215<<2)|0);
      $1217 = HEAP32[$1216>>2]|0;
      $1218 = (($1217) + ($1213)|0);
      $1219 = HEAP8[$1218>>0]|0;
      $Pixel$9 = $1186;$sB$9$shrunk = $1219;$sG$9$shrunk = $1208;$sR$9$shrunk = $1197;
      break;
     }
     case 2:  {
      $1220 = HEAP16[$src$11>>1]|0;
      $1221 = $1220&65535;
      $1222 = HEAP32[$30>>2]|0;
      $1223 = $1221 & $1222;
      $1224 = HEAP32[$31>>2]|0;
      $1225 = $1224 & 255;
      $1226 = $1223 >>> $1225;
      $1227 = HEAP32[$32>>2]|0;
      $1228 = $1227 & 255;
      $1229 = (125924 + ($1228<<2)|0);
      $1230 = HEAP32[$1229>>2]|0;
      $1231 = (($1230) + ($1226)|0);
      $1232 = HEAP8[$1231>>0]|0;
      $1233 = HEAP32[$33>>2]|0;
      $1234 = $1233 & $1221;
      $1235 = $1224 >>> 8;
      $1236 = $1235 & 255;
      $1237 = $1234 >>> $1236;
      $1238 = $1227 >>> 8;
      $1239 = $1238 & 255;
      $1240 = (125924 + ($1239<<2)|0);
      $1241 = HEAP32[$1240>>2]|0;
      $1242 = (($1241) + ($1237)|0);
      $1243 = HEAP8[$1242>>0]|0;
      $1244 = HEAP32[$34>>2]|0;
      $1245 = $1244 & $1221;
      $1246 = $1224 >>> 16;
      $1247 = $1246 & 255;
      $1248 = $1245 >>> $1247;
      $1249 = $1227 >>> 16;
      $1250 = $1249 & 255;
      $1251 = (125924 + ($1250<<2)|0);
      $1252 = HEAP32[$1251>>2]|0;
      $1253 = (($1252) + ($1248)|0);
      $1254 = HEAP8[$1253>>0]|0;
      $Pixel$9 = $1221;$sB$9$shrunk = $1254;$sG$9$shrunk = $1243;$sR$9$shrunk = $1232;
      break;
     }
     case 3:  {
      $1255 = HEAP32[$31>>2]|0;
      $div15 = $1255 >>> 3;
      $1256 = $div15 & 31;
      $1257 = (($src$11) + ($1256)|0);
      $1258 = HEAP8[$1257>>0]|0;
      $1259 = $1255 >>> 11;
      $div16 = $1259 & 31;
      $1260 = (($src$11) + ($div16)|0);
      $1261 = HEAP8[$1260>>0]|0;
      $1262 = $1255 >>> 19;
      $1263 = $1262 & 31;
      $1264 = (($src$11) + ($1263)|0);
      $1265 = HEAP8[$1264>>0]|0;
      $Pixel$9 = 0;$sB$9$shrunk = $1265;$sG$9$shrunk = $1261;$sR$9$shrunk = $1258;
      break;
     }
     case 4:  {
      $1266 = HEAP32[$src$11>>2]|0;
      $1267 = HEAP32[$30>>2]|0;
      $1268 = $1267 & $1266;
      $1269 = HEAP32[$31>>2]|0;
      $1270 = $1269 & 255;
      $1271 = $1268 >>> $1270;
      $1272 = HEAP32[$32>>2]|0;
      $1273 = $1272 & 255;
      $1274 = (125924 + ($1273<<2)|0);
      $1275 = HEAP32[$1274>>2]|0;
      $1276 = (($1275) + ($1271)|0);
      $1277 = HEAP8[$1276>>0]|0;
      $1278 = HEAP32[$33>>2]|0;
      $1279 = $1278 & $1266;
      $1280 = $1269 >>> 8;
      $1281 = $1280 & 255;
      $1282 = $1279 >>> $1281;
      $1283 = $1272 >>> 8;
      $1284 = $1283 & 255;
      $1285 = (125924 + ($1284<<2)|0);
      $1286 = HEAP32[$1285>>2]|0;
      $1287 = (($1286) + ($1282)|0);
      $1288 = HEAP8[$1287>>0]|0;
      $1289 = HEAP32[$34>>2]|0;
      $1290 = $1289 & $1266;
      $1291 = $1269 >>> 16;
      $1292 = $1291 & 255;
      $1293 = $1290 >>> $1292;
      $1294 = $1272 >>> 16;
      $1295 = $1294 & 255;
      $1296 = (125924 + ($1295<<2)|0);
      $1297 = HEAP32[$1296>>2]|0;
      $1298 = (($1297) + ($1293)|0);
      $1299 = HEAP8[$1298>>0]|0;
      $Pixel$9 = $1266;$sB$9$shrunk = $1299;$sG$9$shrunk = $1288;$sR$9$shrunk = $1277;
      break;
     }
     default: {
      $Pixel$9 = 0;$sB$9$shrunk = 0;$sG$9$shrunk = 0;$sR$9$shrunk = 0;
     }
     }
     $1300 = $Pixel$9 & $19;
     $1301 = ($1300|0)==($23|0);
     if (!($1301)) {
      $sB$9 = $sB$9$shrunk&255;
      $sG$9 = $sG$9$shrunk&255;
      $sR$9 = $sR$9$shrunk&255;
      $1302 = $sR$9 & 224;
      $1303 = $sG$9 >>> 5;
      $1304 = $1303 << 2;
      $1305 = $1304 | $1302;
      $1306 = $sB$9 >>> 6;
      $1307 = $1305 | $1306;
      $1308 = (($14) + ($1307)|0);
      $1309 = HEAP8[$1308>>0]|0;
      HEAP8[$dst$11>>0] = $1309;
     }
     $1310 = ((($dst$11)) + 1|0);
     $1311 = (($src$11) + ($22)|0);
     $dst$12 = $1310;$n1$2 = $n1$1;$src$12 = $1311;
     label = 90;
     continue;
    }
    else if ((label|0) == 90) {
     label = 0;
     switch ($22|0) {
     case 1:  {
      $1312 = HEAP8[$src$12>>0]|0;
      $1313 = $1312&255;
      $1314 = HEAP32[$30>>2]|0;
      $1315 = $1313 & $1314;
      $1316 = HEAP32[$31>>2]|0;
      $1317 = $1316 & 255;
      $1318 = $1315 >>> $1317;
      $1319 = HEAP32[$32>>2]|0;
      $1320 = $1319 & 255;
      $1321 = (125924 + ($1320<<2)|0);
      $1322 = HEAP32[$1321>>2]|0;
      $1323 = (($1322) + ($1318)|0);
      $1324 = HEAP8[$1323>>0]|0;
      $1325 = HEAP32[$33>>2]|0;
      $1326 = $1325 & $1313;
      $1327 = $1316 >>> 8;
      $1328 = $1327 & 255;
      $1329 = $1326 >>> $1328;
      $1330 = $1319 >>> 8;
      $1331 = $1330 & 255;
      $1332 = (125924 + ($1331<<2)|0);
      $1333 = HEAP32[$1332>>2]|0;
      $1334 = (($1333) + ($1329)|0);
      $1335 = HEAP8[$1334>>0]|0;
      $1336 = HEAP32[$34>>2]|0;
      $1337 = $1336 & $1313;
      $1338 = $1316 >>> 16;
      $1339 = $1338 & 255;
      $1340 = $1337 >>> $1339;
      $1341 = $1319 >>> 16;
      $1342 = $1341 & 255;
      $1343 = (125924 + ($1342<<2)|0);
      $1344 = HEAP32[$1343>>2]|0;
      $1345 = (($1344) + ($1340)|0);
      $1346 = HEAP8[$1345>>0]|0;
      $Pixel$10 = $1313;$sB$10$shrunk = $1346;$sG$10$shrunk = $1335;$sR$10$shrunk = $1324;
      break;
     }
     case 2:  {
      $1347 = HEAP16[$src$12>>1]|0;
      $1348 = $1347&65535;
      $1349 = HEAP32[$30>>2]|0;
      $1350 = $1348 & $1349;
      $1351 = HEAP32[$31>>2]|0;
      $1352 = $1351 & 255;
      $1353 = $1350 >>> $1352;
      $1354 = HEAP32[$32>>2]|0;
      $1355 = $1354 & 255;
      $1356 = (125924 + ($1355<<2)|0);
      $1357 = HEAP32[$1356>>2]|0;
      $1358 = (($1357) + ($1353)|0);
      $1359 = HEAP8[$1358>>0]|0;
      $1360 = HEAP32[$33>>2]|0;
      $1361 = $1360 & $1348;
      $1362 = $1351 >>> 8;
      $1363 = $1362 & 255;
      $1364 = $1361 >>> $1363;
      $1365 = $1354 >>> 8;
      $1366 = $1365 & 255;
      $1367 = (125924 + ($1366<<2)|0);
      $1368 = HEAP32[$1367>>2]|0;
      $1369 = (($1368) + ($1364)|0);
      $1370 = HEAP8[$1369>>0]|0;
      $1371 = HEAP32[$34>>2]|0;
      $1372 = $1371 & $1348;
      $1373 = $1351 >>> 16;
      $1374 = $1373 & 255;
      $1375 = $1372 >>> $1374;
      $1376 = $1354 >>> 16;
      $1377 = $1376 & 255;
      $1378 = (125924 + ($1377<<2)|0);
      $1379 = HEAP32[$1378>>2]|0;
      $1380 = (($1379) + ($1375)|0);
      $1381 = HEAP8[$1380>>0]|0;
      $Pixel$10 = $1348;$sB$10$shrunk = $1381;$sG$10$shrunk = $1370;$sR$10$shrunk = $1359;
      break;
     }
     case 3:  {
      $1382 = HEAP32[$31>>2]|0;
      $div12 = $1382 >>> 3;
      $1383 = $div12 & 31;
      $1384 = (($src$12) + ($1383)|0);
      $1385 = HEAP8[$1384>>0]|0;
      $1386 = $1382 >>> 11;
      $div13 = $1386 & 31;
      $1387 = (($src$12) + ($div13)|0);
      $1388 = HEAP8[$1387>>0]|0;
      $1389 = $1382 >>> 19;
      $1390 = $1389 & 31;
      $1391 = (($src$12) + ($1390)|0);
      $1392 = HEAP8[$1391>>0]|0;
      $Pixel$10 = 0;$sB$10$shrunk = $1392;$sG$10$shrunk = $1388;$sR$10$shrunk = $1385;
      break;
     }
     case 4:  {
      $1393 = HEAP32[$src$12>>2]|0;
      $1394 = HEAP32[$30>>2]|0;
      $1395 = $1394 & $1393;
      $1396 = HEAP32[$31>>2]|0;
      $1397 = $1396 & 255;
      $1398 = $1395 >>> $1397;
      $1399 = HEAP32[$32>>2]|0;
      $1400 = $1399 & 255;
      $1401 = (125924 + ($1400<<2)|0);
      $1402 = HEAP32[$1401>>2]|0;
      $1403 = (($1402) + ($1398)|0);
      $1404 = HEAP8[$1403>>0]|0;
      $1405 = HEAP32[$33>>2]|0;
      $1406 = $1405 & $1393;
      $1407 = $1396 >>> 8;
      $1408 = $1407 & 255;
      $1409 = $1406 >>> $1408;
      $1410 = $1399 >>> 8;
      $1411 = $1410 & 255;
      $1412 = (125924 + ($1411<<2)|0);
      $1413 = HEAP32[$1412>>2]|0;
      $1414 = (($1413) + ($1409)|0);
      $1415 = HEAP8[$1414>>0]|0;
      $1416 = HEAP32[$34>>2]|0;
      $1417 = $1416 & $1393;
      $1418 = $1396 >>> 16;
      $1419 = $1418 & 255;
      $1420 = $1417 >>> $1419;
      $1421 = $1399 >>> 16;
      $1422 = $1421 & 255;
      $1423 = (125924 + ($1422<<2)|0);
      $1424 = HEAP32[$1423>>2]|0;
      $1425 = (($1424) + ($1420)|0);
      $1426 = HEAP8[$1425>>0]|0;
      $Pixel$10 = $1393;$sB$10$shrunk = $1426;$sG$10$shrunk = $1415;$sR$10$shrunk = $1404;
      break;
     }
     default: {
      $Pixel$10 = 0;$sB$10$shrunk = 0;$sG$10$shrunk = 0;$sR$10$shrunk = 0;
     }
     }
     $1427 = $Pixel$10 & $19;
     $1428 = ($1427|0)==($23|0);
     if (!($1428)) {
      $sB$10 = $sB$10$shrunk&255;
      $sG$10 = $sG$10$shrunk&255;
      $sR$10 = $sR$10$shrunk&255;
      $1429 = $sR$10 & 224;
      $1430 = $sG$10 >>> 5;
      $1431 = $1430 << 2;
      $1432 = $1431 | $1429;
      $1433 = $sB$10 >>> 6;
      $1434 = $1432 | $1433;
      $1435 = (($14) + ($1434)|0);
      $1436 = HEAP8[$1435>>0]|0;
      HEAP8[$dst$12>>0] = $1436;
     }
     $1437 = ((($dst$12)) + 1|0);
     $1438 = (($src$12) + ($22)|0);
     $dst$13 = $1437;$n1$3 = $n1$2;$src$13 = $1438;
     label = 98;
     continue;
    }
    else if ((label|0) == 98) {
     label = 0;
     switch ($22|0) {
     case 1:  {
      $1439 = HEAP8[$src$13>>0]|0;
      $1440 = $1439&255;
      $1441 = HEAP32[$30>>2]|0;
      $1442 = $1440 & $1441;
      $1443 = HEAP32[$31>>2]|0;
      $1444 = $1443 & 255;
      $1445 = $1442 >>> $1444;
      $1446 = HEAP32[$32>>2]|0;
      $1447 = $1446 & 255;
      $1448 = (125924 + ($1447<<2)|0);
      $1449 = HEAP32[$1448>>2]|0;
      $1450 = (($1449) + ($1445)|0);
      $1451 = HEAP8[$1450>>0]|0;
      $1452 = HEAP32[$33>>2]|0;
      $1453 = $1452 & $1440;
      $1454 = $1443 >>> 8;
      $1455 = $1454 & 255;
      $1456 = $1453 >>> $1455;
      $1457 = $1446 >>> 8;
      $1458 = $1457 & 255;
      $1459 = (125924 + ($1458<<2)|0);
      $1460 = HEAP32[$1459>>2]|0;
      $1461 = (($1460) + ($1456)|0);
      $1462 = HEAP8[$1461>>0]|0;
      $1463 = HEAP32[$34>>2]|0;
      $1464 = $1463 & $1440;
      $1465 = $1443 >>> 16;
      $1466 = $1465 & 255;
      $1467 = $1464 >>> $1466;
      $1468 = $1446 >>> 16;
      $1469 = $1468 & 255;
      $1470 = (125924 + ($1469<<2)|0);
      $1471 = HEAP32[$1470>>2]|0;
      $1472 = (($1471) + ($1467)|0);
      $1473 = HEAP8[$1472>>0]|0;
      $Pixel$11 = $1440;$sB$11$shrunk = $1473;$sG$11$shrunk = $1462;$sR$11$shrunk = $1451;
      break;
     }
     case 2:  {
      $1474 = HEAP16[$src$13>>1]|0;
      $1475 = $1474&65535;
      $1476 = HEAP32[$30>>2]|0;
      $1477 = $1475 & $1476;
      $1478 = HEAP32[$31>>2]|0;
      $1479 = $1478 & 255;
      $1480 = $1477 >>> $1479;
      $1481 = HEAP32[$32>>2]|0;
      $1482 = $1481 & 255;
      $1483 = (125924 + ($1482<<2)|0);
      $1484 = HEAP32[$1483>>2]|0;
      $1485 = (($1484) + ($1480)|0);
      $1486 = HEAP8[$1485>>0]|0;
      $1487 = HEAP32[$33>>2]|0;
      $1488 = $1487 & $1475;
      $1489 = $1478 >>> 8;
      $1490 = $1489 & 255;
      $1491 = $1488 >>> $1490;
      $1492 = $1481 >>> 8;
      $1493 = $1492 & 255;
      $1494 = (125924 + ($1493<<2)|0);
      $1495 = HEAP32[$1494>>2]|0;
      $1496 = (($1495) + ($1491)|0);
      $1497 = HEAP8[$1496>>0]|0;
      $1498 = HEAP32[$34>>2]|0;
      $1499 = $1498 & $1475;
      $1500 = $1478 >>> 16;
      $1501 = $1500 & 255;
      $1502 = $1499 >>> $1501;
      $1503 = $1481 >>> 16;
      $1504 = $1503 & 255;
      $1505 = (125924 + ($1504<<2)|0);
      $1506 = HEAP32[$1505>>2]|0;
      $1507 = (($1506) + ($1502)|0);
      $1508 = HEAP8[$1507>>0]|0;
      $Pixel$11 = $1475;$sB$11$shrunk = $1508;$sG$11$shrunk = $1497;$sR$11$shrunk = $1486;
      break;
     }
     case 3:  {
      $1509 = HEAP32[$31>>2]|0;
      $div9 = $1509 >>> 3;
      $1510 = $div9 & 31;
      $1511 = (($src$13) + ($1510)|0);
      $1512 = HEAP8[$1511>>0]|0;
      $1513 = $1509 >>> 11;
      $div10 = $1513 & 31;
      $1514 = (($src$13) + ($div10)|0);
      $1515 = HEAP8[$1514>>0]|0;
      $1516 = $1509 >>> 19;
      $1517 = $1516 & 31;
      $1518 = (($src$13) + ($1517)|0);
      $1519 = HEAP8[$1518>>0]|0;
      $Pixel$11 = 0;$sB$11$shrunk = $1519;$sG$11$shrunk = $1515;$sR$11$shrunk = $1512;
      break;
     }
     case 4:  {
      $1520 = HEAP32[$src$13>>2]|0;
      $1521 = HEAP32[$30>>2]|0;
      $1522 = $1521 & $1520;
      $1523 = HEAP32[$31>>2]|0;
      $1524 = $1523 & 255;
      $1525 = $1522 >>> $1524;
      $1526 = HEAP32[$32>>2]|0;
      $1527 = $1526 & 255;
      $1528 = (125924 + ($1527<<2)|0);
      $1529 = HEAP32[$1528>>2]|0;
      $1530 = (($1529) + ($1525)|0);
      $1531 = HEAP8[$1530>>0]|0;
      $1532 = HEAP32[$33>>2]|0;
      $1533 = $1532 & $1520;
      $1534 = $1523 >>> 8;
      $1535 = $1534 & 255;
      $1536 = $1533 >>> $1535;
      $1537 = $1526 >>> 8;
      $1538 = $1537 & 255;
      $1539 = (125924 + ($1538<<2)|0);
      $1540 = HEAP32[$1539>>2]|0;
      $1541 = (($1540) + ($1536)|0);
      $1542 = HEAP8[$1541>>0]|0;
      $1543 = HEAP32[$34>>2]|0;
      $1544 = $1543 & $1520;
      $1545 = $1523 >>> 16;
      $1546 = $1545 & 255;
      $1547 = $1544 >>> $1546;
      $1548 = $1526 >>> 16;
      $1549 = $1548 & 255;
      $1550 = (125924 + ($1549<<2)|0);
      $1551 = HEAP32[$1550>>2]|0;
      $1552 = (($1551) + ($1547)|0);
      $1553 = HEAP8[$1552>>0]|0;
      $Pixel$11 = $1520;$sB$11$shrunk = $1553;$sG$11$shrunk = $1542;$sR$11$shrunk = $1531;
      break;
     }
     default: {
      $Pixel$11 = 0;$sB$11$shrunk = 0;$sG$11$shrunk = 0;$sR$11$shrunk = 0;
     }
     }
     $1554 = $Pixel$11 & $19;
     $1555 = ($1554|0)==($23|0);
     if (!($1555)) {
      $sB$11 = $sB$11$shrunk&255;
      $sG$11 = $sG$11$shrunk&255;
      $sR$11 = $sR$11$shrunk&255;
      $1556 = $sR$11 & 224;
      $1557 = $sG$11 >>> 5;
      $1558 = $1557 << 2;
      $1559 = $1558 | $1556;
      $1560 = $sB$11 >>> 6;
      $1561 = $1559 | $1560;
      $1562 = (($14) + ($1561)|0);
      $1563 = HEAP8[$1562>>0]|0;
      HEAP8[$dst$13>>0] = $1563;
     }
     $1564 = ((($dst$13)) + 1|0);
     $1565 = (($src$13) + ($22)|0);
     $dst$14 = $1564;$n1$4 = $n1$3;$src$14 = $1565;
     label = 106;
     continue;
    }
    else if ((label|0) == 106) {
     label = 0;
     switch ($22|0) {
     case 1:  {
      $1566 = HEAP8[$src$14>>0]|0;
      $1567 = $1566&255;
      $1568 = HEAP32[$30>>2]|0;
      $1569 = $1567 & $1568;
      $1570 = HEAP32[$31>>2]|0;
      $1571 = $1570 & 255;
      $1572 = $1569 >>> $1571;
      $1573 = HEAP32[$32>>2]|0;
      $1574 = $1573 & 255;
      $1575 = (125924 + ($1574<<2)|0);
      $1576 = HEAP32[$1575>>2]|0;
      $1577 = (($1576) + ($1572)|0);
      $1578 = HEAP8[$1577>>0]|0;
      $1579 = HEAP32[$33>>2]|0;
      $1580 = $1579 & $1567;
      $1581 = $1570 >>> 8;
      $1582 = $1581 & 255;
      $1583 = $1580 >>> $1582;
      $1584 = $1573 >>> 8;
      $1585 = $1584 & 255;
      $1586 = (125924 + ($1585<<2)|0);
      $1587 = HEAP32[$1586>>2]|0;
      $1588 = (($1587) + ($1583)|0);
      $1589 = HEAP8[$1588>>0]|0;
      $1590 = HEAP32[$34>>2]|0;
      $1591 = $1590 & $1567;
      $1592 = $1570 >>> 16;
      $1593 = $1592 & 255;
      $1594 = $1591 >>> $1593;
      $1595 = $1573 >>> 16;
      $1596 = $1595 & 255;
      $1597 = (125924 + ($1596<<2)|0);
      $1598 = HEAP32[$1597>>2]|0;
      $1599 = (($1598) + ($1594)|0);
      $1600 = HEAP8[$1599>>0]|0;
      $Pixel$12 = $1567;$sB$12$shrunk = $1600;$sG$12$shrunk = $1589;$sR$12$shrunk = $1578;
      break;
     }
     case 2:  {
      $1601 = HEAP16[$src$14>>1]|0;
      $1602 = $1601&65535;
      $1603 = HEAP32[$30>>2]|0;
      $1604 = $1602 & $1603;
      $1605 = HEAP32[$31>>2]|0;
      $1606 = $1605 & 255;
      $1607 = $1604 >>> $1606;
      $1608 = HEAP32[$32>>2]|0;
      $1609 = $1608 & 255;
      $1610 = (125924 + ($1609<<2)|0);
      $1611 = HEAP32[$1610>>2]|0;
      $1612 = (($1611) + ($1607)|0);
      $1613 = HEAP8[$1612>>0]|0;
      $1614 = HEAP32[$33>>2]|0;
      $1615 = $1614 & $1602;
      $1616 = $1605 >>> 8;
      $1617 = $1616 & 255;
      $1618 = $1615 >>> $1617;
      $1619 = $1608 >>> 8;
      $1620 = $1619 & 255;
      $1621 = (125924 + ($1620<<2)|0);
      $1622 = HEAP32[$1621>>2]|0;
      $1623 = (($1622) + ($1618)|0);
      $1624 = HEAP8[$1623>>0]|0;
      $1625 = HEAP32[$34>>2]|0;
      $1626 = $1625 & $1602;
      $1627 = $1605 >>> 16;
      $1628 = $1627 & 255;
      $1629 = $1626 >>> $1628;
      $1630 = $1608 >>> 16;
      $1631 = $1630 & 255;
      $1632 = (125924 + ($1631<<2)|0);
      $1633 = HEAP32[$1632>>2]|0;
      $1634 = (($1633) + ($1629)|0);
      $1635 = HEAP8[$1634>>0]|0;
      $Pixel$12 = $1602;$sB$12$shrunk = $1635;$sG$12$shrunk = $1624;$sR$12$shrunk = $1613;
      break;
     }
     case 3:  {
      $1636 = HEAP32[$31>>2]|0;
      $div6 = $1636 >>> 3;
      $1637 = $div6 & 31;
      $1638 = (($src$14) + ($1637)|0);
      $1639 = HEAP8[$1638>>0]|0;
      $1640 = $1636 >>> 11;
      $div7 = $1640 & 31;
      $1641 = (($src$14) + ($div7)|0);
      $1642 = HEAP8[$1641>>0]|0;
      $1643 = $1636 >>> 19;
      $1644 = $1643 & 31;
      $1645 = (($src$14) + ($1644)|0);
      $1646 = HEAP8[$1645>>0]|0;
      $Pixel$12 = 0;$sB$12$shrunk = $1646;$sG$12$shrunk = $1642;$sR$12$shrunk = $1639;
      break;
     }
     case 4:  {
      $1647 = HEAP32[$src$14>>2]|0;
      $1648 = HEAP32[$30>>2]|0;
      $1649 = $1648 & $1647;
      $1650 = HEAP32[$31>>2]|0;
      $1651 = $1650 & 255;
      $1652 = $1649 >>> $1651;
      $1653 = HEAP32[$32>>2]|0;
      $1654 = $1653 & 255;
      $1655 = (125924 + ($1654<<2)|0);
      $1656 = HEAP32[$1655>>2]|0;
      $1657 = (($1656) + ($1652)|0);
      $1658 = HEAP8[$1657>>0]|0;
      $1659 = HEAP32[$33>>2]|0;
      $1660 = $1659 & $1647;
      $1661 = $1650 >>> 8;
      $1662 = $1661 & 255;
      $1663 = $1660 >>> $1662;
      $1664 = $1653 >>> 8;
      $1665 = $1664 & 255;
      $1666 = (125924 + ($1665<<2)|0);
      $1667 = HEAP32[$1666>>2]|0;
      $1668 = (($1667) + ($1663)|0);
      $1669 = HEAP8[$1668>>0]|0;
      $1670 = HEAP32[$34>>2]|0;
      $1671 = $1670 & $1647;
      $1672 = $1650 >>> 16;
      $1673 = $1672 & 255;
      $1674 = $1671 >>> $1673;
      $1675 = $1653 >>> 16;
      $1676 = $1675 & 255;
      $1677 = (125924 + ($1676<<2)|0);
      $1678 = HEAP32[$1677>>2]|0;
      $1679 = (($1678) + ($1674)|0);
      $1680 = HEAP8[$1679>>0]|0;
      $Pixel$12 = $1647;$sB$12$shrunk = $1680;$sG$12$shrunk = $1669;$sR$12$shrunk = $1658;
      break;
     }
     default: {
      $Pixel$12 = 0;$sB$12$shrunk = 0;$sG$12$shrunk = 0;$sR$12$shrunk = 0;
     }
     }
     $1681 = $Pixel$12 & $19;
     $1682 = ($1681|0)==($23|0);
     if (!($1682)) {
      $sB$12 = $sB$12$shrunk&255;
      $sG$12 = $sG$12$shrunk&255;
      $sR$12 = $sR$12$shrunk&255;
      $1683 = $sR$12 & 224;
      $1684 = $sG$12 >>> 5;
      $1685 = $1684 << 2;
      $1686 = $1685 | $1683;
      $1687 = $sB$12 >>> 6;
      $1688 = $1686 | $1687;
      $1689 = (($14) + ($1688)|0);
      $1690 = HEAP8[$1689>>0]|0;
      HEAP8[$dst$14>>0] = $1690;
     }
     $1691 = ((($dst$14)) + 1|0);
     $1692 = (($src$14) + ($22)|0);
     $dst$15 = $1691;$n1$5 = $n1$4;$src$15 = $1692;
     label = 114;
     continue;
    }
    else if ((label|0) == 114) {
     label = 0;
     switch ($22|0) {
     case 1:  {
      $1693 = HEAP8[$src$15>>0]|0;
      $1694 = $1693&255;
      $1695 = HEAP32[$30>>2]|0;
      $1696 = $1694 & $1695;
      $1697 = HEAP32[$31>>2]|0;
      $1698 = $1697 & 255;
      $1699 = $1696 >>> $1698;
      $1700 = HEAP32[$32>>2]|0;
      $1701 = $1700 & 255;
      $1702 = (125924 + ($1701<<2)|0);
      $1703 = HEAP32[$1702>>2]|0;
      $1704 = (($1703) + ($1699)|0);
      $1705 = HEAP8[$1704>>0]|0;
      $1706 = HEAP32[$33>>2]|0;
      $1707 = $1706 & $1694;
      $1708 = $1697 >>> 8;
      $1709 = $1708 & 255;
      $1710 = $1707 >>> $1709;
      $1711 = $1700 >>> 8;
      $1712 = $1711 & 255;
      $1713 = (125924 + ($1712<<2)|0);
      $1714 = HEAP32[$1713>>2]|0;
      $1715 = (($1714) + ($1710)|0);
      $1716 = HEAP8[$1715>>0]|0;
      $1717 = HEAP32[$34>>2]|0;
      $1718 = $1717 & $1694;
      $1719 = $1697 >>> 16;
      $1720 = $1719 & 255;
      $1721 = $1718 >>> $1720;
      $1722 = $1700 >>> 16;
      $1723 = $1722 & 255;
      $1724 = (125924 + ($1723<<2)|0);
      $1725 = HEAP32[$1724>>2]|0;
      $1726 = (($1725) + ($1721)|0);
      $1727 = HEAP8[$1726>>0]|0;
      $Pixel$13 = $1694;$sB$13$shrunk = $1727;$sG$13$shrunk = $1716;$sR$13$shrunk = $1705;
      break;
     }
     case 2:  {
      $1728 = HEAP16[$src$15>>1]|0;
      $1729 = $1728&65535;
      $1730 = HEAP32[$30>>2]|0;
      $1731 = $1729 & $1730;
      $1732 = HEAP32[$31>>2]|0;
      $1733 = $1732 & 255;
      $1734 = $1731 >>> $1733;
      $1735 = HEAP32[$32>>2]|0;
      $1736 = $1735 & 255;
      $1737 = (125924 + ($1736<<2)|0);
      $1738 = HEAP32[$1737>>2]|0;
      $1739 = (($1738) + ($1734)|0);
      $1740 = HEAP8[$1739>>0]|0;
      $1741 = HEAP32[$33>>2]|0;
      $1742 = $1741 & $1729;
      $1743 = $1732 >>> 8;
      $1744 = $1743 & 255;
      $1745 = $1742 >>> $1744;
      $1746 = $1735 >>> 8;
      $1747 = $1746 & 255;
      $1748 = (125924 + ($1747<<2)|0);
      $1749 = HEAP32[$1748>>2]|0;
      $1750 = (($1749) + ($1745)|0);
      $1751 = HEAP8[$1750>>0]|0;
      $1752 = HEAP32[$34>>2]|0;
      $1753 = $1752 & $1729;
      $1754 = $1732 >>> 16;
      $1755 = $1754 & 255;
      $1756 = $1753 >>> $1755;
      $1757 = $1735 >>> 16;
      $1758 = $1757 & 255;
      $1759 = (125924 + ($1758<<2)|0);
      $1760 = HEAP32[$1759>>2]|0;
      $1761 = (($1760) + ($1756)|0);
      $1762 = HEAP8[$1761>>0]|0;
      $Pixel$13 = $1729;$sB$13$shrunk = $1762;$sG$13$shrunk = $1751;$sR$13$shrunk = $1740;
      break;
     }
     case 3:  {
      $1763 = HEAP32[$31>>2]|0;
      $div3 = $1763 >>> 3;
      $1764 = $div3 & 31;
      $1765 = (($src$15) + ($1764)|0);
      $1766 = HEAP8[$1765>>0]|0;
      $1767 = $1763 >>> 11;
      $div4 = $1767 & 31;
      $1768 = (($src$15) + ($div4)|0);
      $1769 = HEAP8[$1768>>0]|0;
      $1770 = $1763 >>> 19;
      $1771 = $1770 & 31;
      $1772 = (($src$15) + ($1771)|0);
      $1773 = HEAP8[$1772>>0]|0;
      $Pixel$13 = 0;$sB$13$shrunk = $1773;$sG$13$shrunk = $1769;$sR$13$shrunk = $1766;
      break;
     }
     case 4:  {
      $1774 = HEAP32[$src$15>>2]|0;
      $1775 = HEAP32[$30>>2]|0;
      $1776 = $1775 & $1774;
      $1777 = HEAP32[$31>>2]|0;
      $1778 = $1777 & 255;
      $1779 = $1776 >>> $1778;
      $1780 = HEAP32[$32>>2]|0;
      $1781 = $1780 & 255;
      $1782 = (125924 + ($1781<<2)|0);
      $1783 = HEAP32[$1782>>2]|0;
      $1784 = (($1783) + ($1779)|0);
      $1785 = HEAP8[$1784>>0]|0;
      $1786 = HEAP32[$33>>2]|0;
      $1787 = $1786 & $1774;
      $1788 = $1777 >>> 8;
      $1789 = $1788 & 255;
      $1790 = $1787 >>> $1789;
      $1791 = $1780 >>> 8;
      $1792 = $1791 & 255;
      $1793 = (125924 + ($1792<<2)|0);
      $1794 = HEAP32[$1793>>2]|0;
      $1795 = (($1794) + ($1790)|0);
      $1796 = HEAP8[$1795>>0]|0;
      $1797 = HEAP32[$34>>2]|0;
      $1798 = $1797 & $1774;
      $1799 = $1777 >>> 16;
      $1800 = $1799 & 255;
      $1801 = $1798 >>> $1800;
      $1802 = $1780 >>> 16;
      $1803 = $1802 & 255;
      $1804 = (125924 + ($1803<<2)|0);
      $1805 = HEAP32[$1804>>2]|0;
      $1806 = (($1805) + ($1801)|0);
      $1807 = HEAP8[$1806>>0]|0;
      $Pixel$13 = $1774;$sB$13$shrunk = $1807;$sG$13$shrunk = $1796;$sR$13$shrunk = $1785;
      break;
     }
     default: {
      $Pixel$13 = 0;$sB$13$shrunk = 0;$sG$13$shrunk = 0;$sR$13$shrunk = 0;
     }
     }
     $1808 = $Pixel$13 & $19;
     $1809 = ($1808|0)==($23|0);
     if (!($1809)) {
      $sB$13 = $sB$13$shrunk&255;
      $sG$13 = $sG$13$shrunk&255;
      $sR$13 = $sR$13$shrunk&255;
      $1810 = $sR$13 & 224;
      $1811 = $sG$13 >>> 5;
      $1812 = $1811 << 2;
      $1813 = $1812 | $1810;
      $1814 = $sB$13 >>> 6;
      $1815 = $1813 | $1814;
      $1816 = (($14) + ($1815)|0);
      $1817 = HEAP8[$1816>>0]|0;
      HEAP8[$dst$15>>0] = $1817;
     }
     $1818 = ((($dst$15)) + 1|0);
     $1819 = (($src$15) + ($22)|0);
     $dst$16 = $1818;$n1$6 = $n1$5;$src$16 = $1819;
     label = 122;
     continue;
    }
    else if ((label|0) == 122) {
     label = 0;
     switch ($22|0) {
     case 1:  {
      $1820 = HEAP8[$src$16>>0]|0;
      $1821 = $1820&255;
      $1822 = HEAP32[$30>>2]|0;
      $1823 = $1821 & $1822;
      $1824 = HEAP32[$31>>2]|0;
      $1825 = $1824 & 255;
      $1826 = $1823 >>> $1825;
      $1827 = HEAP32[$32>>2]|0;
      $1828 = $1827 & 255;
      $1829 = (125924 + ($1828<<2)|0);
      $1830 = HEAP32[$1829>>2]|0;
      $1831 = (($1830) + ($1826)|0);
      $1832 = HEAP8[$1831>>0]|0;
      $1833 = HEAP32[$33>>2]|0;
      $1834 = $1833 & $1821;
      $1835 = $1824 >>> 8;
      $1836 = $1835 & 255;
      $1837 = $1834 >>> $1836;
      $1838 = $1827 >>> 8;
      $1839 = $1838 & 255;
      $1840 = (125924 + ($1839<<2)|0);
      $1841 = HEAP32[$1840>>2]|0;
      $1842 = (($1841) + ($1837)|0);
      $1843 = HEAP8[$1842>>0]|0;
      $1844 = HEAP32[$34>>2]|0;
      $1845 = $1844 & $1821;
      $1846 = $1824 >>> 16;
      $1847 = $1846 & 255;
      $1848 = $1845 >>> $1847;
      $1849 = $1827 >>> 16;
      $1850 = $1849 & 255;
      $1851 = (125924 + ($1850<<2)|0);
      $1852 = HEAP32[$1851>>2]|0;
      $1853 = (($1852) + ($1848)|0);
      $1854 = HEAP8[$1853>>0]|0;
      $Pixel$14 = $1821;$sB$14$shrunk = $1854;$sG$14$shrunk = $1843;$sR$14$shrunk = $1832;
      break;
     }
     case 2:  {
      $1855 = HEAP16[$src$16>>1]|0;
      $1856 = $1855&65535;
      $1857 = HEAP32[$30>>2]|0;
      $1858 = $1856 & $1857;
      $1859 = HEAP32[$31>>2]|0;
      $1860 = $1859 & 255;
      $1861 = $1858 >>> $1860;
      $1862 = HEAP32[$32>>2]|0;
      $1863 = $1862 & 255;
      $1864 = (125924 + ($1863<<2)|0);
      $1865 = HEAP32[$1864>>2]|0;
      $1866 = (($1865) + ($1861)|0);
      $1867 = HEAP8[$1866>>0]|0;
      $1868 = HEAP32[$33>>2]|0;
      $1869 = $1868 & $1856;
      $1870 = $1859 >>> 8;
      $1871 = $1870 & 255;
      $1872 = $1869 >>> $1871;
      $1873 = $1862 >>> 8;
      $1874 = $1873 & 255;
      $1875 = (125924 + ($1874<<2)|0);
      $1876 = HEAP32[$1875>>2]|0;
      $1877 = (($1876) + ($1872)|0);
      $1878 = HEAP8[$1877>>0]|0;
      $1879 = HEAP32[$34>>2]|0;
      $1880 = $1879 & $1856;
      $1881 = $1859 >>> 16;
      $1882 = $1881 & 255;
      $1883 = $1880 >>> $1882;
      $1884 = $1862 >>> 16;
      $1885 = $1884 & 255;
      $1886 = (125924 + ($1885<<2)|0);
      $1887 = HEAP32[$1886>>2]|0;
      $1888 = (($1887) + ($1883)|0);
      $1889 = HEAP8[$1888>>0]|0;
      $Pixel$14 = $1856;$sB$14$shrunk = $1889;$sG$14$shrunk = $1878;$sR$14$shrunk = $1867;
      break;
     }
     case 3:  {
      $1890 = HEAP32[$31>>2]|0;
      $div = $1890 >>> 3;
      $1891 = $div & 31;
      $1892 = (($src$16) + ($1891)|0);
      $1893 = HEAP8[$1892>>0]|0;
      $1894 = $1890 >>> 11;
      $div1 = $1894 & 31;
      $1895 = (($src$16) + ($div1)|0);
      $1896 = HEAP8[$1895>>0]|0;
      $1897 = $1890 >>> 19;
      $1898 = $1897 & 31;
      $1899 = (($src$16) + ($1898)|0);
      $1900 = HEAP8[$1899>>0]|0;
      $Pixel$14 = 0;$sB$14$shrunk = $1900;$sG$14$shrunk = $1896;$sR$14$shrunk = $1893;
      break;
     }
     case 4:  {
      $1901 = HEAP32[$src$16>>2]|0;
      $1902 = HEAP32[$30>>2]|0;
      $1903 = $1902 & $1901;
      $1904 = HEAP32[$31>>2]|0;
      $1905 = $1904 & 255;
      $1906 = $1903 >>> $1905;
      $1907 = HEAP32[$32>>2]|0;
      $1908 = $1907 & 255;
      $1909 = (125924 + ($1908<<2)|0);
      $1910 = HEAP32[$1909>>2]|0;
      $1911 = (($1910) + ($1906)|0);
      $1912 = HEAP8[$1911>>0]|0;
      $1913 = HEAP32[$33>>2]|0;
      $1914 = $1913 & $1901;
      $1915 = $1904 >>> 8;
      $1916 = $1915 & 255;
      $1917 = $1914 >>> $1916;
      $1918 = $1907 >>> 8;
      $1919 = $1918 & 255;
      $1920 = (125924 + ($1919<<2)|0);
      $1921 = HEAP32[$1920>>2]|0;
      $1922 = (($1921) + ($1917)|0);
      $1923 = HEAP8[$1922>>0]|0;
      $1924 = HEAP32[$34>>2]|0;
      $1925 = $1924 & $1901;
      $1926 = $1904 >>> 16;
      $1927 = $1926 & 255;
      $1928 = $1925 >>> $1927;
      $1929 = $1907 >>> 16;
      $1930 = $1929 & 255;
      $1931 = (125924 + ($1930<<2)|0);
      $1932 = HEAP32[$1931>>2]|0;
      $1933 = (($1932) + ($1928)|0);
      $1934 = HEAP8[$1933>>0]|0;
      $Pixel$14 = $1901;$sB$14$shrunk = $1934;$sG$14$shrunk = $1923;$sR$14$shrunk = $1912;
      break;
     }
     default: {
      $Pixel$14 = 0;$sB$14$shrunk = 0;$sG$14$shrunk = 0;$sR$14$shrunk = 0;
     }
     }
     $1935 = $Pixel$14 & $19;
     $1936 = ($1935|0)==($23|0);
     if (!($1936)) {
      $sB$14 = $sB$14$shrunk&255;
      $sG$14 = $sG$14$shrunk&255;
      $sR$14 = $sR$14$shrunk&255;
      $1937 = $sR$14 & 224;
      $1938 = $sG$14 >>> 5;
      $1939 = $1938 << 2;
      $1940 = $1939 | $1937;
      $1941 = $sB$14 >>> 6;
      $1942 = $1940 | $1941;
      $1943 = (($14) + ($1942)|0);
      $1944 = HEAP8[$1943>>0]|0;
      HEAP8[$dst$16>>0] = $1944;
     }
     $1945 = ((($dst$16)) + 1|0);
     $1946 = (($src$16) + ($22)|0);
     $dst$17 = $1945;$n1$7 = $n1$6;$src$17 = $1946;
     label = 130;
     continue;
    }
    else if ((label|0) == 130) {
     label = 0;
     switch ($22|0) {
     case 1:  {
      $1947 = HEAP8[$src$17>>0]|0;
      $1948 = $1947&255;
      $1949 = HEAP32[$30>>2]|0;
      $1950 = $1948 & $1949;
      $1951 = HEAP32[$31>>2]|0;
      $1952 = $1951 & 255;
      $1953 = $1950 >>> $1952;
      $1954 = HEAP32[$32>>2]|0;
      $1955 = $1954 & 255;
      $1956 = (125924 + ($1955<<2)|0);
      $1957 = HEAP32[$1956>>2]|0;
      $1958 = (($1957) + ($1953)|0);
      $1959 = HEAP8[$1958>>0]|0;
      $1960 = HEAP32[$33>>2]|0;
      $1961 = $1960 & $1948;
      $1962 = $1951 >>> 8;
      $1963 = $1962 & 255;
      $1964 = $1961 >>> $1963;
      $1965 = $1954 >>> 8;
      $1966 = $1965 & 255;
      $1967 = (125924 + ($1966<<2)|0);
      $1968 = HEAP32[$1967>>2]|0;
      $1969 = (($1968) + ($1964)|0);
      $1970 = HEAP8[$1969>>0]|0;
      $1971 = HEAP32[$34>>2]|0;
      $1972 = $1971 & $1948;
      $1973 = $1951 >>> 16;
      $1974 = $1973 & 255;
      $1975 = $1972 >>> $1974;
      $1976 = $1954 >>> 16;
      $1977 = $1976 & 255;
      $1978 = (125924 + ($1977<<2)|0);
      $1979 = HEAP32[$1978>>2]|0;
      $1980 = (($1979) + ($1975)|0);
      $1981 = HEAP8[$1980>>0]|0;
      $Pixel$15 = $1948;$sB$15$shrunk = $1981;$sG$15$shrunk = $1970;$sR$15$shrunk = $1959;
      break;
     }
     case 2:  {
      $1982 = HEAP16[$src$17>>1]|0;
      $1983 = $1982&65535;
      $1984 = HEAP32[$30>>2]|0;
      $1985 = $1983 & $1984;
      $1986 = HEAP32[$31>>2]|0;
      $1987 = $1986 & 255;
      $1988 = $1985 >>> $1987;
      $1989 = HEAP32[$32>>2]|0;
      $1990 = $1989 & 255;
      $1991 = (125924 + ($1990<<2)|0);
      $1992 = HEAP32[$1991>>2]|0;
      $1993 = (($1992) + ($1988)|0);
      $1994 = HEAP8[$1993>>0]|0;
      $1995 = HEAP32[$33>>2]|0;
      $1996 = $1995 & $1983;
      $1997 = $1986 >>> 8;
      $1998 = $1997 & 255;
      $1999 = $1996 >>> $1998;
      $2000 = $1989 >>> 8;
      $2001 = $2000 & 255;
      $2002 = (125924 + ($2001<<2)|0);
      $2003 = HEAP32[$2002>>2]|0;
      $2004 = (($2003) + ($1999)|0);
      $2005 = HEAP8[$2004>>0]|0;
      $2006 = HEAP32[$34>>2]|0;
      $2007 = $2006 & $1983;
      $2008 = $1986 >>> 16;
      $2009 = $2008 & 255;
      $2010 = $2007 >>> $2009;
      $2011 = $1989 >>> 16;
      $2012 = $2011 & 255;
      $2013 = (125924 + ($2012<<2)|0);
      $2014 = HEAP32[$2013>>2]|0;
      $2015 = (($2014) + ($2010)|0);
      $2016 = HEAP8[$2015>>0]|0;
      $Pixel$15 = $1983;$sB$15$shrunk = $2016;$sG$15$shrunk = $2005;$sR$15$shrunk = $1994;
      break;
     }
     case 3:  {
      $2017 = HEAP32[$31>>2]|0;
      $div21 = $2017 >>> 3;
      $2018 = $div21 & 31;
      $2019 = (($src$17) + ($2018)|0);
      $2020 = HEAP8[$2019>>0]|0;
      $2021 = $2017 >>> 11;
      $div22 = $2021 & 31;
      $2022 = (($src$17) + ($div22)|0);
      $2023 = HEAP8[$2022>>0]|0;
      $2024 = $2017 >>> 19;
      $2025 = $2024 & 31;
      $2026 = (($src$17) + ($2025)|0);
      $2027 = HEAP8[$2026>>0]|0;
      $Pixel$15 = 0;$sB$15$shrunk = $2027;$sG$15$shrunk = $2023;$sR$15$shrunk = $2020;
      break;
     }
     case 4:  {
      $2028 = HEAP32[$src$17>>2]|0;
      $2029 = HEAP32[$30>>2]|0;
      $2030 = $2029 & $2028;
      $2031 = HEAP32[$31>>2]|0;
      $2032 = $2031 & 255;
      $2033 = $2030 >>> $2032;
      $2034 = HEAP32[$32>>2]|0;
      $2035 = $2034 & 255;
      $2036 = (125924 + ($2035<<2)|0);
      $2037 = HEAP32[$2036>>2]|0;
      $2038 = (($2037) + ($2033)|0);
      $2039 = HEAP8[$2038>>0]|0;
      $2040 = HEAP32[$33>>2]|0;
      $2041 = $2040 & $2028;
      $2042 = $2031 >>> 8;
      $2043 = $2042 & 255;
      $2044 = $2041 >>> $2043;
      $2045 = $2034 >>> 8;
      $2046 = $2045 & 255;
      $2047 = (125924 + ($2046<<2)|0);
      $2048 = HEAP32[$2047>>2]|0;
      $2049 = (($2048) + ($2044)|0);
      $2050 = HEAP8[$2049>>0]|0;
      $2051 = HEAP32[$34>>2]|0;
      $2052 = $2051 & $2028;
      $2053 = $2031 >>> 16;
      $2054 = $2053 & 255;
      $2055 = $2052 >>> $2054;
      $2056 = $2034 >>> 16;
      $2057 = $2056 & 255;
      $2058 = (125924 + ($2057<<2)|0);
      $2059 = HEAP32[$2058>>2]|0;
      $2060 = (($2059) + ($2055)|0);
      $2061 = HEAP8[$2060>>0]|0;
      $Pixel$15 = $2028;$sB$15$shrunk = $2061;$sG$15$shrunk = $2050;$sR$15$shrunk = $2039;
      break;
     }
     default: {
      $Pixel$15 = 0;$sB$15$shrunk = 0;$sG$15$shrunk = 0;$sR$15$shrunk = 0;
     }
     }
     $2062 = $Pixel$15 & $19;
     $2063 = ($2062|0)==($23|0);
     if (!($2063)) {
      $sB$15 = $sB$15$shrunk&255;
      $sG$15 = $sG$15$shrunk&255;
      $sR$15 = $sR$15$shrunk&255;
      $2064 = $sR$15 & 224;
      $2065 = $sG$15 >>> 5;
      $2066 = $2065 << 2;
      $2067 = $2066 | $2064;
      $2068 = $sB$15 >>> 6;
      $2069 = $2067 | $2068;
      $2070 = (($14) + ($2069)|0);
      $2071 = HEAP8[$2070>>0]|0;
      HEAP8[$dst$17>>0] = $2071;
     }
     $2072 = ((($dst$17)) + 1|0);
     $2073 = (($src$17) + ($22)|0);
     $2074 = (($n1$7) + -1)|0;
     $2075 = ($n1$7|0)>(1);
     if ($2075) {
      $dst$10 = $2072;$n1$0 = $2074;$src$10 = $2073;
      label = 74;
      continue;
     } else {
      break;
     }
    }
   }
   $2076 = (($2073) + ($6)|0);
   $2077 = (($2072) + ($10)|0);
   $2078 = (($2079) + -1)|0;
   $2080 = ($2079|0)==(0);
   if ($2080) {
    label = 140;
    break;
   } else {
    $2079 = $2078;$dst$952 = $2077;$src$953 = $2076;
   }
  }
  if ((label|0) == 138) {
   // unreachable;
  }
  else if ((label|0) == 140) {
   return;
  }
 }
}
function _BlitNtoNKeyCopyAlpha($info) {
 $info = $info|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0;
 var $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0;
 var $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0;
 var $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0;
 var $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0;
 var $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0;
 var $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0;
 var $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0;
 var $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0;
 var $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0;
 var $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0;
 var $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0;
 var $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0;
 var $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0;
 var $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0;
 var $1267 = 0, $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0;
 var $1285 = 0, $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0;
 var $1302 = 0, $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0;
 var $1320 = 0, $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0;
 var $1339 = 0, $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0;
 var $1357 = 0, $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0;
 var $1375 = 0, $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0;
 var $1393 = 0, $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0;
 var $1410 = 0, $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0;
 var $1429 = 0, $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0;
 var $1447 = 0, $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0;
 var $1465 = 0, $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0;
 var $1483 = 0, $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0;
 var $1500 = 0, $1501 = 0, $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0, $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0, $1511 = 0, $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0, $1518 = 0;
 var $1519 = 0, $152 = 0, $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0, $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0, $153 = 0, $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0, $1536 = 0;
 var $1537 = 0, $1538 = 0, $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0, $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0, $1548 = 0, $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0, $1554 = 0;
 var $1555 = 0, $1556 = 0, $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0, $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $1565 = 0, $1566 = 0, $1567 = 0, $1568 = 0, $1569 = 0, $157 = 0, $1570 = 0, $1571 = 0, $1572 = 0;
 var $1573 = 0, $1574 = 0, $1575 = 0, $1576 = 0, $1577 = 0, $1578 = 0, $1579 = 0, $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0, $1583 = 0, $1584 = 0, $1585 = 0, $1586 = 0, $1587 = 0, $1588 = 0, $1589 = 0, $159 = 0, $1590 = 0;
 var $1591 = 0, $1592 = 0, $1593 = 0, $1594 = 0, $1595 = 0, $1596 = 0, $1597 = 0, $1598 = 0, $1599 = 0, $16 = 0, $160 = 0, $1600 = 0, $1601 = 0, $1602 = 0, $1603 = 0, $1604 = 0, $1605 = 0, $1606 = 0, $1607 = 0, $1608 = 0;
 var $1609 = 0, $161 = 0, $1610 = 0, $1611 = 0, $1612 = 0, $1613 = 0, $1614 = 0, $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0, $1619 = 0, $162 = 0, $1620 = 0, $1621 = 0, $1622 = 0, $1623 = 0, $1624 = 0, $1625 = 0, $1626 = 0;
 var $1627 = 0, $1628 = 0, $1629 = 0, $163 = 0, $1630 = 0, $1631 = 0, $1632 = 0, $1633 = 0, $1634 = 0, $1635 = 0, $1636 = 0, $1637 = 0, $1638 = 0, $1639 = 0, $164 = 0, $1640 = 0, $1641 = 0, $1642 = 0, $1643 = 0, $1644 = 0;
 var $1645 = 0, $1646 = 0, $1647 = 0, $1648 = 0, $1649 = 0, $165 = 0, $1650 = 0, $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0, $1655 = 0, $1656 = 0, $1657 = 0, $1658 = 0, $1659 = 0, $166 = 0, $1660 = 0, $1661 = 0, $1662 = 0;
 var $1663 = 0, $1664 = 0, $1665 = 0, $1666 = 0, $1667 = 0, $1668 = 0, $1669 = 0, $167 = 0, $1670 = 0, $1671 = 0, $1672 = 0, $1673 = 0, $1674 = 0, $1675 = 0, $1676 = 0, $1677 = 0, $1678 = 0, $1679 = 0, $168 = 0, $1680 = 0;
 var $1681 = 0, $1682 = 0, $1683 = 0, $1684 = 0, $1685 = 0, $1686 = 0, $1687 = 0, $1688 = 0, $1689 = 0, $169 = 0, $1690 = 0, $1691 = 0, $1692 = 0, $1693 = 0, $1694 = 0, $1695 = 0, $1696 = 0, $1697 = 0, $1698 = 0, $1699 = 0;
 var $17 = 0, $170 = 0, $1700 = 0, $1701 = 0, $1702 = 0, $1703 = 0, $1704 = 0, $1705 = 0, $1706 = 0, $1707 = 0, $1708 = 0, $1709 = 0, $171 = 0, $1710 = 0, $1711 = 0, $1712 = 0, $1713 = 0, $1714 = 0, $1715 = 0, $1716 = 0;
 var $1717 = 0, $1718 = 0, $1719 = 0, $172 = 0, $1720 = 0, $1721 = 0, $1722 = 0, $1723 = 0, $1724 = 0, $1725 = 0, $1726 = 0, $1727 = 0, $1728 = 0, $1729 = 0, $173 = 0, $1730 = 0, $1731 = 0, $1732 = 0, $1733 = 0, $1734 = 0;
 var $1735 = 0, $1736 = 0, $1737 = 0, $1738 = 0, $1739 = 0, $174 = 0, $1740 = 0, $1741 = 0, $1742 = 0, $1743 = 0, $1744 = 0, $1745 = 0, $1746 = 0, $1747 = 0, $1748 = 0, $1749 = 0, $175 = 0, $1750 = 0, $1751 = 0, $1752 = 0;
 var $1753 = 0, $1754 = 0, $1755 = 0, $1756 = 0, $1757 = 0, $1758 = 0, $1759 = 0, $176 = 0, $1760 = 0, $1761 = 0, $1762 = 0, $1763 = 0, $1764 = 0, $1765 = 0, $1766 = 0, $1767 = 0, $1768 = 0, $1769 = 0, $177 = 0, $1770 = 0;
 var $1771 = 0, $1772 = 0, $1773 = 0, $1774 = 0, $1775 = 0, $1776 = 0, $1777 = 0, $1778 = 0, $1779 = 0, $178 = 0, $1780 = 0, $1781 = 0, $1782 = 0, $1783 = 0, $1784 = 0, $1785 = 0, $1786 = 0, $1787 = 0, $1788 = 0, $1789 = 0;
 var $179 = 0, $1790 = 0, $1791 = 0, $1792 = 0, $1793 = 0, $1794 = 0, $1795 = 0, $1796 = 0, $1797 = 0, $1798 = 0, $1799 = 0, $18 = 0, $180 = 0, $1800 = 0, $1801 = 0, $1802 = 0, $1803 = 0, $1804 = 0, $1805 = 0, $1806 = 0;
 var $1807 = 0, $1808 = 0, $1809 = 0, $181 = 0, $1810 = 0, $1811 = 0, $1812 = 0, $1813 = 0, $1814 = 0, $1815 = 0, $1816 = 0, $1817 = 0, $1818 = 0, $1819 = 0, $182 = 0, $1820 = 0, $1821 = 0, $1822 = 0, $1823 = 0, $1824 = 0;
 var $1825 = 0, $1826 = 0, $1827 = 0, $1828 = 0, $1829 = 0, $183 = 0, $1830 = 0, $1831 = 0, $1832 = 0, $1833 = 0, $1834 = 0, $1835 = 0, $1836 = 0, $1837 = 0, $1838 = 0, $1839 = 0, $184 = 0, $1840 = 0, $1841 = 0, $1842 = 0;
 var $1843 = 0, $1844 = 0, $1845 = 0, $1846 = 0, $1847 = 0, $1848 = 0, $1849 = 0, $185 = 0, $1850 = 0, $1851 = 0, $1852 = 0, $1853 = 0, $1854 = 0, $1855 = 0, $1856 = 0, $1857 = 0, $1858 = 0, $1859 = 0, $186 = 0, $1860 = 0;
 var $1861 = 0, $1862 = 0, $1863 = 0, $1864 = 0, $1865 = 0, $1866 = 0, $1867 = 0, $1868 = 0, $1869 = 0, $187 = 0, $1870 = 0, $1871 = 0, $1872 = 0, $1873 = 0, $1874 = 0, $1875 = 0, $1876 = 0, $1877 = 0, $1878 = 0, $1879 = 0;
 var $188 = 0, $1880 = 0, $1881 = 0, $1882 = 0, $1883 = 0, $1884 = 0, $1885 = 0, $1886 = 0, $1887 = 0, $1888 = 0, $1889 = 0, $189 = 0, $1890 = 0, $1891 = 0, $1892 = 0, $1893 = 0, $1894 = 0, $1895 = 0, $1896 = 0, $1897 = 0;
 var $1898 = 0, $1899 = 0, $19 = 0, $190 = 0, $1900 = 0, $1901 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0;
 var $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0;
 var $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0;
 var $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0;
 var $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0;
 var $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0;
 var $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0;
 var $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0;
 var $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0;
 var $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0;
 var $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0;
 var $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0;
 var $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0;
 var $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0;
 var $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0;
 var $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0;
 var $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0;
 var $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0;
 var $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0;
 var $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0;
 var $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0;
 var $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0;
 var $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0;
 var $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0;
 var $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0;
 var $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0;
 var $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0;
 var $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0;
 var $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0;
 var $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0;
 var $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0;
 var $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0;
 var $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0;
 var $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0;
 var $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0;
 var $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0;
 var $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0;
 var $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0;
 var $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0;
 var $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0;
 var $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0;
 var $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0;
 var $997 = 0, $998 = 0, $999 = 0, $Pixel$0 = 0, $Pixel$1 = 0, $Pixel$2 = 0, $Pixel$3 = 0, $Pixel$4 = 0, $Pixel$5 = 0, $Pixel$6 = 0, $Pixel$7 = 0, $div = 0, $div1 = 0, $div10 = 0, $div12 = 0, $div13 = 0, $div14 = 0, $div15 = 0, $div16 = 0, $div18 = 0;
 var $div19 = 0, $div2 = 0, $div20 = 0, $div21 = 0, $div22 = 0, $div24 = 0, $div25 = 0, $div26 = 0, $div27 = 0, $div28 = 0, $div3 = 0, $div30 = 0, $div31 = 0, $div32 = 0, $div33 = 0, $div34 = 0, $div36 = 0, $div37 = 0, $div38 = 0, $div39 = 0;
 var $div4 = 0, $div40 = 0, $div42 = 0, $div43 = 0, $div44 = 0, $div45 = 0, $div46 = 0, $div6 = 0, $div7 = 0, $div8 = 0, $div9 = 0, $dst$048 = 0, $dst$1 = 0, $dst$2 = 0, $dst$3 = 0, $dst$4 = 0, $dst$5 = 0, $dst$6 = 0, $dst$7 = 0, $dst$8 = 0;
 var $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n$4 = 0, $n$5 = 0, $n$6 = 0, $n$7 = 0, $sA$0 = 0, $sA$0$shrunk = 0, $sA$1 = 0, $sA$1$shrunk = 0, $sA$2 = 0, $sA$2$shrunk = 0, $sA$3 = 0, $sA$3$shrunk = 0, $sA$4 = 0, $sA$4$shrunk = 0, $sA$5 = 0, $sA$5$shrunk = 0;
 var $sA$6 = 0, $sA$6$shrunk = 0, $sA$7 = 0, $sA$7$shrunk = 0, $sB$0 = 0, $sB$0$shrunk = 0, $sB$1 = 0, $sB$1$shrunk = 0, $sB$2 = 0, $sB$2$shrunk = 0, $sB$3 = 0, $sB$3$shrunk = 0, $sB$4 = 0, $sB$4$shrunk = 0, $sB$5 = 0, $sB$5$shrunk = 0, $sB$6 = 0, $sB$6$shrunk = 0, $sB$7 = 0, $sB$7$shrunk = 0;
 var $sG$0 = 0, $sG$0$shrunk = 0, $sG$1 = 0, $sG$1$shrunk = 0, $sG$2 = 0, $sG$2$shrunk = 0, $sG$3 = 0, $sG$3$shrunk = 0, $sG$4 = 0, $sG$4$shrunk = 0, $sG$5 = 0, $sG$5$shrunk = 0, $sG$6 = 0, $sG$6$shrunk = 0, $sG$7 = 0, $sG$7$shrunk = 0, $sR$0 = 0, $sR$0$shrunk = 0, $sR$1 = 0, $sR$1$shrunk = 0;
 var $sR$2 = 0, $sR$2$shrunk = 0, $sR$3 = 0, $sR$3$shrunk = 0, $sR$4 = 0, $sR$4$shrunk = 0, $sR$5 = 0, $sR$5$shrunk = 0, $sR$6 = 0, $sR$6$shrunk = 0, $sR$7 = 0, $sR$7$shrunk = 0, $src$049 = 0, $src$1 = 0, $src$2 = 0, $src$3 = 0, $src$4 = 0, $src$5 = 0, $src$6 = 0, $src$7 = 0;
 var $src$8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($info)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($info)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($info)) + 56|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($info)) + 40|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($info)) + 44|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($11)) + 24|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = $15 ^ -1;
 $17 = $9 & $16;
 $18 = ($3|0)==(0);
 if ($18) {
  return;
 }
 $19 = ((($13)) + 9|0);
 $20 = HEAP8[$19>>0]|0;
 $21 = ((($11)) + 9|0);
 $22 = HEAP8[$21>>0]|0;
 $23 = ((($info)) + 20|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = HEAP32[$info>>2]|0;
 $26 = (($1) + 7)|0;
 $27 = (($26|0) / 8)&-1;
 $28 = $1 & 7;
 $29 = $22&255;
 $30 = $20&255;
 $31 = ((($13)) + 28|0);
 $32 = ((($13)) + 32|0);
 $33 = ((($13)) + 33|0);
 $34 = ((($13)) + 34|0);
 $35 = ((($11)) + 12|0);
 $36 = ((($11)) + 32|0);
 $37 = ((($11)) + 28|0);
 $38 = ((($11)) + 16|0);
 $39 = ((($11)) + 20|0);
 $$in = $3;$dst$048 = $24;$src$049 = $25;
 L4: while(1) {
  $40 = (($$in) + -1)|0;
  switch ($28|0) {
  case 0:  {
   $dst$1 = $dst$048;$n$0 = $27;$src$1 = $src$049;
   label = 4;
   break;
  }
  case 7:  {
   $dst$2 = $dst$048;$n$1 = $27;$src$2 = $src$049;
   label = 16;
   break;
  }
  case 6:  {
   $dst$3 = $dst$048;$n$2 = $27;$src$3 = $src$049;
   label = 28;
   break;
  }
  case 5:  {
   $dst$4 = $dst$048;$n$3 = $27;$src$4 = $src$049;
   label = 40;
   break;
  }
  case 4:  {
   $dst$5 = $dst$048;$n$4 = $27;$src$5 = $src$049;
   label = 52;
   break;
  }
  case 3:  {
   $dst$6 = $dst$048;$n$5 = $27;$src$6 = $src$049;
   label = 64;
   break;
  }
  case 2:  {
   $dst$7 = $dst$048;$n$6 = $27;$src$7 = $src$049;
   label = 76;
   break;
  }
  case 1:  {
   $dst$8 = $dst$048;$n$7 = $27;$src$8 = $src$049;
   label = 88;
   break;
  }
  default: {
   label = 100;
   break L4;
  }
  }
  while(1) {
   if ((label|0) == 4) {
    label = 0;
    switch ($29|0) {
    case 1:  {
     $41 = HEAP8[$src$1>>0]|0;
     $42 = $41&255;
     $43 = HEAP32[$35>>2]|0;
     $44 = $42 & $43;
     $45 = HEAP32[$36>>2]|0;
     $46 = $45 & 255;
     $47 = $44 >>> $46;
     $48 = HEAP32[$37>>2]|0;
     $49 = $48 & 255;
     $50 = (125924 + ($49<<2)|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = (($51) + ($47)|0);
     $53 = HEAP8[$52>>0]|0;
     $54 = HEAP32[$38>>2]|0;
     $55 = $54 & $42;
     $56 = $45 >>> 8;
     $57 = $56 & 255;
     $58 = $55 >>> $57;
     $59 = $48 >>> 8;
     $60 = $59 & 255;
     $61 = (125924 + ($60<<2)|0);
     $62 = HEAP32[$61>>2]|0;
     $63 = (($62) + ($58)|0);
     $64 = HEAP8[$63>>0]|0;
     $65 = HEAP32[$39>>2]|0;
     $66 = $65 & $42;
     $67 = $45 >>> 16;
     $68 = $67 & 255;
     $69 = $66 >>> $68;
     $70 = $48 >>> 16;
     $71 = $70 & 255;
     $72 = (125924 + ($71<<2)|0);
     $73 = HEAP32[$72>>2]|0;
     $74 = (($73) + ($69)|0);
     $75 = HEAP8[$74>>0]|0;
     $76 = HEAP32[$14>>2]|0;
     $77 = $76 & $42;
     $78 = $45 >>> 24;
     $79 = $77 >>> $78;
     $80 = $48 >>> 24;
     $81 = (125924 + ($80<<2)|0);
     $82 = HEAP32[$81>>2]|0;
     $83 = (($82) + ($79)|0);
     $84 = HEAP8[$83>>0]|0;
     $Pixel$0 = $42;$sA$0$shrunk = $84;$sB$0$shrunk = $75;$sG$0$shrunk = $64;$sR$0$shrunk = $53;
     break;
    }
    case 2:  {
     $85 = HEAP16[$src$1>>1]|0;
     $86 = $85&65535;
     $87 = HEAP32[$35>>2]|0;
     $88 = $86 & $87;
     $89 = HEAP32[$36>>2]|0;
     $90 = $89 & 255;
     $91 = $88 >>> $90;
     $92 = HEAP32[$37>>2]|0;
     $93 = $92 & 255;
     $94 = (125924 + ($93<<2)|0);
     $95 = HEAP32[$94>>2]|0;
     $96 = (($95) + ($91)|0);
     $97 = HEAP8[$96>>0]|0;
     $98 = HEAP32[$38>>2]|0;
     $99 = $98 & $86;
     $100 = $89 >>> 8;
     $101 = $100 & 255;
     $102 = $99 >>> $101;
     $103 = $92 >>> 8;
     $104 = $103 & 255;
     $105 = (125924 + ($104<<2)|0);
     $106 = HEAP32[$105>>2]|0;
     $107 = (($106) + ($102)|0);
     $108 = HEAP8[$107>>0]|0;
     $109 = HEAP32[$39>>2]|0;
     $110 = $109 & $86;
     $111 = $89 >>> 16;
     $112 = $111 & 255;
     $113 = $110 >>> $112;
     $114 = $92 >>> 16;
     $115 = $114 & 255;
     $116 = (125924 + ($115<<2)|0);
     $117 = HEAP32[$116>>2]|0;
     $118 = (($117) + ($113)|0);
     $119 = HEAP8[$118>>0]|0;
     $120 = HEAP32[$14>>2]|0;
     $121 = $120 & $86;
     $122 = $89 >>> 24;
     $123 = $121 >>> $122;
     $124 = $92 >>> 24;
     $125 = (125924 + ($124<<2)|0);
     $126 = HEAP32[$125>>2]|0;
     $127 = (($126) + ($123)|0);
     $128 = HEAP8[$127>>0]|0;
     $Pixel$0 = $86;$sA$0$shrunk = $128;$sB$0$shrunk = $119;$sG$0$shrunk = $108;$sR$0$shrunk = $97;
     break;
    }
    case 3:  {
     $129 = HEAP32[$36>>2]|0;
     $div39 = $129 >>> 3;
     $130 = $div39 & 31;
     $131 = (($src$1) + ($130)|0);
     $132 = HEAP8[$131>>0]|0;
     $133 = $129 >>> 11;
     $div40 = $133 & 31;
     $134 = (($src$1) + ($div40)|0);
     $135 = HEAP8[$134>>0]|0;
     $136 = $129 >>> 19;
     $137 = $136 & 31;
     $138 = (($src$1) + ($137)|0);
     $139 = HEAP8[$138>>0]|0;
     $Pixel$0 = 0;$sA$0$shrunk = -1;$sB$0$shrunk = $139;$sG$0$shrunk = $135;$sR$0$shrunk = $132;
     break;
    }
    case 4:  {
     $140 = HEAP32[$src$1>>2]|0;
     $141 = HEAP32[$35>>2]|0;
     $142 = $141 & $140;
     $143 = HEAP32[$36>>2]|0;
     $144 = $143 & 255;
     $145 = $142 >>> $144;
     $146 = HEAP32[$37>>2]|0;
     $147 = $146 & 255;
     $148 = (125924 + ($147<<2)|0);
     $149 = HEAP32[$148>>2]|0;
     $150 = (($149) + ($145)|0);
     $151 = HEAP8[$150>>0]|0;
     $152 = HEAP32[$38>>2]|0;
     $153 = $152 & $140;
     $154 = $143 >>> 8;
     $155 = $154 & 255;
     $156 = $153 >>> $155;
     $157 = $146 >>> 8;
     $158 = $157 & 255;
     $159 = (125924 + ($158<<2)|0);
     $160 = HEAP32[$159>>2]|0;
     $161 = (($160) + ($156)|0);
     $162 = HEAP8[$161>>0]|0;
     $163 = HEAP32[$39>>2]|0;
     $164 = $163 & $140;
     $165 = $143 >>> 16;
     $166 = $165 & 255;
     $167 = $164 >>> $166;
     $168 = $146 >>> 16;
     $169 = $168 & 255;
     $170 = (125924 + ($169<<2)|0);
     $171 = HEAP32[$170>>2]|0;
     $172 = (($171) + ($167)|0);
     $173 = HEAP8[$172>>0]|0;
     $174 = HEAP32[$14>>2]|0;
     $175 = $174 & $140;
     $176 = $143 >>> 24;
     $177 = $175 >>> $176;
     $178 = $146 >>> 24;
     $179 = (125924 + ($178<<2)|0);
     $180 = HEAP32[$179>>2]|0;
     $181 = (($180) + ($177)|0);
     $182 = HEAP8[$181>>0]|0;
     $Pixel$0 = $140;$sA$0$shrunk = $182;$sB$0$shrunk = $173;$sG$0$shrunk = $162;$sR$0$shrunk = $151;
     break;
    }
    default: {
     $Pixel$0 = 0;$sA$0$shrunk = 0;$sB$0$shrunk = 0;$sG$0$shrunk = 0;$sR$0$shrunk = 0;
    }
    }
    $sR$0 = $sR$0$shrunk&255;
    $sG$0 = $sG$0$shrunk&255;
    $sB$0 = $sB$0$shrunk&255;
    $sA$0 = $sA$0$shrunk&255;
    $183 = $Pixel$0 & $16;
    $184 = ($183|0)==($17|0);
    L15: do {
     if (!($184)) {
      switch ($30|0) {
      case 1:  {
       $185 = HEAP32[$31>>2]|0;
       $186 = $185 & 255;
       $187 = $sR$0 >>> $186;
       $188 = HEAP32[$32>>2]|0;
       $189 = $188 & 255;
       $190 = $187 << $189;
       $191 = $185 >>> 8;
       $192 = $191 & 255;
       $193 = $sG$0 >>> $192;
       $194 = $188 >>> 8;
       $195 = $194 & 255;
       $196 = $193 << $195;
       $197 = $196 | $190;
       $198 = $185 >>> 16;
       $199 = $198 & 255;
       $200 = $sB$0 >>> $199;
       $201 = $188 >>> 16;
       $202 = $201 & 255;
       $203 = $200 << $202;
       $204 = $197 | $203;
       $205 = $185 >>> 24;
       $206 = $sA$0 >>> $205;
       $207 = $188 >>> 24;
       $208 = $206 << $207;
       $209 = $204 | $208;
       $210 = $209&255;
       HEAP8[$dst$1>>0] = $210;
       break L15;
       break;
      }
      case 2:  {
       $211 = HEAP32[$31>>2]|0;
       $212 = $211 & 255;
       $213 = $sR$0 >>> $212;
       $214 = HEAP32[$32>>2]|0;
       $215 = $214 & 255;
       $216 = $213 << $215;
       $217 = $211 >>> 8;
       $218 = $217 & 255;
       $219 = $sG$0 >>> $218;
       $220 = $214 >>> 8;
       $221 = $220 & 255;
       $222 = $219 << $221;
       $223 = $222 | $216;
       $224 = $211 >>> 16;
       $225 = $224 & 255;
       $226 = $sB$0 >>> $225;
       $227 = $214 >>> 16;
       $228 = $227 & 255;
       $229 = $226 << $228;
       $230 = $223 | $229;
       $231 = $211 >>> 24;
       $232 = $sA$0 >>> $231;
       $233 = $214 >>> 24;
       $234 = $232 << $233;
       $235 = $230 | $234;
       $236 = $235&65535;
       HEAP16[$dst$1>>1] = $236;
       break L15;
       break;
      }
      case 3:  {
       $237 = HEAP8[$32>>0]|0;
       $div36 = ($237&255) >>> 3;
       $238 = $div36&255;
       $239 = (($dst$1) + ($238)|0);
       HEAP8[$239>>0] = $sR$0$shrunk;
       $240 = HEAP8[$33>>0]|0;
       $div37 = ($240&255) >>> 3;
       $241 = $div37&255;
       $242 = (($dst$1) + ($241)|0);
       HEAP8[$242>>0] = $sG$0$shrunk;
       $243 = HEAP8[$34>>0]|0;
       $div38 = ($243&255) >>> 3;
       $244 = $div38&255;
       $245 = (($dst$1) + ($244)|0);
       HEAP8[$245>>0] = $sB$0$shrunk;
       break L15;
       break;
      }
      case 4:  {
       $246 = HEAP32[$31>>2]|0;
       $247 = $246 & 255;
       $248 = $sR$0 >>> $247;
       $249 = HEAP32[$32>>2]|0;
       $250 = $249 & 255;
       $251 = $248 << $250;
       $252 = $246 >>> 8;
       $253 = $252 & 255;
       $254 = $sG$0 >>> $253;
       $255 = $249 >>> 8;
       $256 = $255 & 255;
       $257 = $254 << $256;
       $258 = $257 | $251;
       $259 = $246 >>> 16;
       $260 = $259 & 255;
       $261 = $sB$0 >>> $260;
       $262 = $249 >>> 16;
       $263 = $262 & 255;
       $264 = $261 << $263;
       $265 = $258 | $264;
       $266 = $246 >>> 24;
       $267 = $sA$0 >>> $266;
       $268 = $249 >>> 24;
       $269 = $267 << $268;
       $270 = $265 | $269;
       HEAP32[$dst$1>>2] = $270;
       break L15;
       break;
      }
      default: {
       break L15;
      }
      }
     }
    } while(0);
    $271 = (($dst$1) + ($30)|0);
    $272 = (($src$1) + ($29)|0);
    $dst$2 = $271;$n$1 = $n$0;$src$2 = $272;
    label = 16;
    continue;
   }
   else if ((label|0) == 16) {
    label = 0;
    switch ($29|0) {
    case 1:  {
     $273 = HEAP8[$src$2>>0]|0;
     $274 = $273&255;
     $275 = HEAP32[$35>>2]|0;
     $276 = $274 & $275;
     $277 = HEAP32[$36>>2]|0;
     $278 = $277 & 255;
     $279 = $276 >>> $278;
     $280 = HEAP32[$37>>2]|0;
     $281 = $280 & 255;
     $282 = (125924 + ($281<<2)|0);
     $283 = HEAP32[$282>>2]|0;
     $284 = (($283) + ($279)|0);
     $285 = HEAP8[$284>>0]|0;
     $286 = HEAP32[$38>>2]|0;
     $287 = $286 & $274;
     $288 = $277 >>> 8;
     $289 = $288 & 255;
     $290 = $287 >>> $289;
     $291 = $280 >>> 8;
     $292 = $291 & 255;
     $293 = (125924 + ($292<<2)|0);
     $294 = HEAP32[$293>>2]|0;
     $295 = (($294) + ($290)|0);
     $296 = HEAP8[$295>>0]|0;
     $297 = HEAP32[$39>>2]|0;
     $298 = $297 & $274;
     $299 = $277 >>> 16;
     $300 = $299 & 255;
     $301 = $298 >>> $300;
     $302 = $280 >>> 16;
     $303 = $302 & 255;
     $304 = (125924 + ($303<<2)|0);
     $305 = HEAP32[$304>>2]|0;
     $306 = (($305) + ($301)|0);
     $307 = HEAP8[$306>>0]|0;
     $308 = HEAP32[$14>>2]|0;
     $309 = $308 & $274;
     $310 = $277 >>> 24;
     $311 = $309 >>> $310;
     $312 = $280 >>> 24;
     $313 = (125924 + ($312<<2)|0);
     $314 = HEAP32[$313>>2]|0;
     $315 = (($314) + ($311)|0);
     $316 = HEAP8[$315>>0]|0;
     $Pixel$1 = $274;$sA$1$shrunk = $316;$sB$1$shrunk = $307;$sG$1$shrunk = $296;$sR$1$shrunk = $285;
     break;
    }
    case 2:  {
     $317 = HEAP16[$src$2>>1]|0;
     $318 = $317&65535;
     $319 = HEAP32[$35>>2]|0;
     $320 = $318 & $319;
     $321 = HEAP32[$36>>2]|0;
     $322 = $321 & 255;
     $323 = $320 >>> $322;
     $324 = HEAP32[$37>>2]|0;
     $325 = $324 & 255;
     $326 = (125924 + ($325<<2)|0);
     $327 = HEAP32[$326>>2]|0;
     $328 = (($327) + ($323)|0);
     $329 = HEAP8[$328>>0]|0;
     $330 = HEAP32[$38>>2]|0;
     $331 = $330 & $318;
     $332 = $321 >>> 8;
     $333 = $332 & 255;
     $334 = $331 >>> $333;
     $335 = $324 >>> 8;
     $336 = $335 & 255;
     $337 = (125924 + ($336<<2)|0);
     $338 = HEAP32[$337>>2]|0;
     $339 = (($338) + ($334)|0);
     $340 = HEAP8[$339>>0]|0;
     $341 = HEAP32[$39>>2]|0;
     $342 = $341 & $318;
     $343 = $321 >>> 16;
     $344 = $343 & 255;
     $345 = $342 >>> $344;
     $346 = $324 >>> 16;
     $347 = $346 & 255;
     $348 = (125924 + ($347<<2)|0);
     $349 = HEAP32[$348>>2]|0;
     $350 = (($349) + ($345)|0);
     $351 = HEAP8[$350>>0]|0;
     $352 = HEAP32[$14>>2]|0;
     $353 = $352 & $318;
     $354 = $321 >>> 24;
     $355 = $353 >>> $354;
     $356 = $324 >>> 24;
     $357 = (125924 + ($356<<2)|0);
     $358 = HEAP32[$357>>2]|0;
     $359 = (($358) + ($355)|0);
     $360 = HEAP8[$359>>0]|0;
     $Pixel$1 = $318;$sA$1$shrunk = $360;$sB$1$shrunk = $351;$sG$1$shrunk = $340;$sR$1$shrunk = $329;
     break;
    }
    case 3:  {
     $361 = HEAP32[$36>>2]|0;
     $div33 = $361 >>> 3;
     $362 = $div33 & 31;
     $363 = (($src$2) + ($362)|0);
     $364 = HEAP8[$363>>0]|0;
     $365 = $361 >>> 11;
     $div34 = $365 & 31;
     $366 = (($src$2) + ($div34)|0);
     $367 = HEAP8[$366>>0]|0;
     $368 = $361 >>> 19;
     $369 = $368 & 31;
     $370 = (($src$2) + ($369)|0);
     $371 = HEAP8[$370>>0]|0;
     $Pixel$1 = 0;$sA$1$shrunk = -1;$sB$1$shrunk = $371;$sG$1$shrunk = $367;$sR$1$shrunk = $364;
     break;
    }
    case 4:  {
     $372 = HEAP32[$src$2>>2]|0;
     $373 = HEAP32[$35>>2]|0;
     $374 = $373 & $372;
     $375 = HEAP32[$36>>2]|0;
     $376 = $375 & 255;
     $377 = $374 >>> $376;
     $378 = HEAP32[$37>>2]|0;
     $379 = $378 & 255;
     $380 = (125924 + ($379<<2)|0);
     $381 = HEAP32[$380>>2]|0;
     $382 = (($381) + ($377)|0);
     $383 = HEAP8[$382>>0]|0;
     $384 = HEAP32[$38>>2]|0;
     $385 = $384 & $372;
     $386 = $375 >>> 8;
     $387 = $386 & 255;
     $388 = $385 >>> $387;
     $389 = $378 >>> 8;
     $390 = $389 & 255;
     $391 = (125924 + ($390<<2)|0);
     $392 = HEAP32[$391>>2]|0;
     $393 = (($392) + ($388)|0);
     $394 = HEAP8[$393>>0]|0;
     $395 = HEAP32[$39>>2]|0;
     $396 = $395 & $372;
     $397 = $375 >>> 16;
     $398 = $397 & 255;
     $399 = $396 >>> $398;
     $400 = $378 >>> 16;
     $401 = $400 & 255;
     $402 = (125924 + ($401<<2)|0);
     $403 = HEAP32[$402>>2]|0;
     $404 = (($403) + ($399)|0);
     $405 = HEAP8[$404>>0]|0;
     $406 = HEAP32[$14>>2]|0;
     $407 = $406 & $372;
     $408 = $375 >>> 24;
     $409 = $407 >>> $408;
     $410 = $378 >>> 24;
     $411 = (125924 + ($410<<2)|0);
     $412 = HEAP32[$411>>2]|0;
     $413 = (($412) + ($409)|0);
     $414 = HEAP8[$413>>0]|0;
     $Pixel$1 = $372;$sA$1$shrunk = $414;$sB$1$shrunk = $405;$sG$1$shrunk = $394;$sR$1$shrunk = $383;
     break;
    }
    default: {
     $Pixel$1 = 0;$sA$1$shrunk = 0;$sB$1$shrunk = 0;$sG$1$shrunk = 0;$sR$1$shrunk = 0;
    }
    }
    $sR$1 = $sR$1$shrunk&255;
    $sG$1 = $sG$1$shrunk&255;
    $sB$1 = $sB$1$shrunk&255;
    $sA$1 = $sA$1$shrunk&255;
    $415 = $Pixel$1 & $16;
    $416 = ($415|0)==($17|0);
    L30: do {
     if (!($416)) {
      switch ($30|0) {
      case 1:  {
       $417 = HEAP32[$31>>2]|0;
       $418 = $417 & 255;
       $419 = $sR$1 >>> $418;
       $420 = HEAP32[$32>>2]|0;
       $421 = $420 & 255;
       $422 = $419 << $421;
       $423 = $417 >>> 8;
       $424 = $423 & 255;
       $425 = $sG$1 >>> $424;
       $426 = $420 >>> 8;
       $427 = $426 & 255;
       $428 = $425 << $427;
       $429 = $428 | $422;
       $430 = $417 >>> 16;
       $431 = $430 & 255;
       $432 = $sB$1 >>> $431;
       $433 = $420 >>> 16;
       $434 = $433 & 255;
       $435 = $432 << $434;
       $436 = $429 | $435;
       $437 = $417 >>> 24;
       $438 = $sA$1 >>> $437;
       $439 = $420 >>> 24;
       $440 = $438 << $439;
       $441 = $436 | $440;
       $442 = $441&255;
       HEAP8[$dst$2>>0] = $442;
       break L30;
       break;
      }
      case 2:  {
       $443 = HEAP32[$31>>2]|0;
       $444 = $443 & 255;
       $445 = $sR$1 >>> $444;
       $446 = HEAP32[$32>>2]|0;
       $447 = $446 & 255;
       $448 = $445 << $447;
       $449 = $443 >>> 8;
       $450 = $449 & 255;
       $451 = $sG$1 >>> $450;
       $452 = $446 >>> 8;
       $453 = $452 & 255;
       $454 = $451 << $453;
       $455 = $454 | $448;
       $456 = $443 >>> 16;
       $457 = $456 & 255;
       $458 = $sB$1 >>> $457;
       $459 = $446 >>> 16;
       $460 = $459 & 255;
       $461 = $458 << $460;
       $462 = $455 | $461;
       $463 = $443 >>> 24;
       $464 = $sA$1 >>> $463;
       $465 = $446 >>> 24;
       $466 = $464 << $465;
       $467 = $462 | $466;
       $468 = $467&65535;
       HEAP16[$dst$2>>1] = $468;
       break L30;
       break;
      }
      case 3:  {
       $469 = HEAP8[$32>>0]|0;
       $div30 = ($469&255) >>> 3;
       $470 = $div30&255;
       $471 = (($dst$2) + ($470)|0);
       HEAP8[$471>>0] = $sR$1$shrunk;
       $472 = HEAP8[$33>>0]|0;
       $div31 = ($472&255) >>> 3;
       $473 = $div31&255;
       $474 = (($dst$2) + ($473)|0);
       HEAP8[$474>>0] = $sG$1$shrunk;
       $475 = HEAP8[$34>>0]|0;
       $div32 = ($475&255) >>> 3;
       $476 = $div32&255;
       $477 = (($dst$2) + ($476)|0);
       HEAP8[$477>>0] = $sB$1$shrunk;
       break L30;
       break;
      }
      case 4:  {
       $478 = HEAP32[$31>>2]|0;
       $479 = $478 & 255;
       $480 = $sR$1 >>> $479;
       $481 = HEAP32[$32>>2]|0;
       $482 = $481 & 255;
       $483 = $480 << $482;
       $484 = $478 >>> 8;
       $485 = $484 & 255;
       $486 = $sG$1 >>> $485;
       $487 = $481 >>> 8;
       $488 = $487 & 255;
       $489 = $486 << $488;
       $490 = $489 | $483;
       $491 = $478 >>> 16;
       $492 = $491 & 255;
       $493 = $sB$1 >>> $492;
       $494 = $481 >>> 16;
       $495 = $494 & 255;
       $496 = $493 << $495;
       $497 = $490 | $496;
       $498 = $478 >>> 24;
       $499 = $sA$1 >>> $498;
       $500 = $481 >>> 24;
       $501 = $499 << $500;
       $502 = $497 | $501;
       HEAP32[$dst$2>>2] = $502;
       break L30;
       break;
      }
      default: {
       break L30;
      }
      }
     }
    } while(0);
    $503 = (($dst$2) + ($30)|0);
    $504 = (($src$2) + ($29)|0);
    $dst$3 = $503;$n$2 = $n$1;$src$3 = $504;
    label = 28;
    continue;
   }
   else if ((label|0) == 28) {
    label = 0;
    switch ($29|0) {
    case 1:  {
     $505 = HEAP8[$src$3>>0]|0;
     $506 = $505&255;
     $507 = HEAP32[$35>>2]|0;
     $508 = $506 & $507;
     $509 = HEAP32[$36>>2]|0;
     $510 = $509 & 255;
     $511 = $508 >>> $510;
     $512 = HEAP32[$37>>2]|0;
     $513 = $512 & 255;
     $514 = (125924 + ($513<<2)|0);
     $515 = HEAP32[$514>>2]|0;
     $516 = (($515) + ($511)|0);
     $517 = HEAP8[$516>>0]|0;
     $518 = HEAP32[$38>>2]|0;
     $519 = $518 & $506;
     $520 = $509 >>> 8;
     $521 = $520 & 255;
     $522 = $519 >>> $521;
     $523 = $512 >>> 8;
     $524 = $523 & 255;
     $525 = (125924 + ($524<<2)|0);
     $526 = HEAP32[$525>>2]|0;
     $527 = (($526) + ($522)|0);
     $528 = HEAP8[$527>>0]|0;
     $529 = HEAP32[$39>>2]|0;
     $530 = $529 & $506;
     $531 = $509 >>> 16;
     $532 = $531 & 255;
     $533 = $530 >>> $532;
     $534 = $512 >>> 16;
     $535 = $534 & 255;
     $536 = (125924 + ($535<<2)|0);
     $537 = HEAP32[$536>>2]|0;
     $538 = (($537) + ($533)|0);
     $539 = HEAP8[$538>>0]|0;
     $540 = HEAP32[$14>>2]|0;
     $541 = $540 & $506;
     $542 = $509 >>> 24;
     $543 = $541 >>> $542;
     $544 = $512 >>> 24;
     $545 = (125924 + ($544<<2)|0);
     $546 = HEAP32[$545>>2]|0;
     $547 = (($546) + ($543)|0);
     $548 = HEAP8[$547>>0]|0;
     $Pixel$2 = $506;$sA$2$shrunk = $548;$sB$2$shrunk = $539;$sG$2$shrunk = $528;$sR$2$shrunk = $517;
     break;
    }
    case 2:  {
     $549 = HEAP16[$src$3>>1]|0;
     $550 = $549&65535;
     $551 = HEAP32[$35>>2]|0;
     $552 = $550 & $551;
     $553 = HEAP32[$36>>2]|0;
     $554 = $553 & 255;
     $555 = $552 >>> $554;
     $556 = HEAP32[$37>>2]|0;
     $557 = $556 & 255;
     $558 = (125924 + ($557<<2)|0);
     $559 = HEAP32[$558>>2]|0;
     $560 = (($559) + ($555)|0);
     $561 = HEAP8[$560>>0]|0;
     $562 = HEAP32[$38>>2]|0;
     $563 = $562 & $550;
     $564 = $553 >>> 8;
     $565 = $564 & 255;
     $566 = $563 >>> $565;
     $567 = $556 >>> 8;
     $568 = $567 & 255;
     $569 = (125924 + ($568<<2)|0);
     $570 = HEAP32[$569>>2]|0;
     $571 = (($570) + ($566)|0);
     $572 = HEAP8[$571>>0]|0;
     $573 = HEAP32[$39>>2]|0;
     $574 = $573 & $550;
     $575 = $553 >>> 16;
     $576 = $575 & 255;
     $577 = $574 >>> $576;
     $578 = $556 >>> 16;
     $579 = $578 & 255;
     $580 = (125924 + ($579<<2)|0);
     $581 = HEAP32[$580>>2]|0;
     $582 = (($581) + ($577)|0);
     $583 = HEAP8[$582>>0]|0;
     $584 = HEAP32[$14>>2]|0;
     $585 = $584 & $550;
     $586 = $553 >>> 24;
     $587 = $585 >>> $586;
     $588 = $556 >>> 24;
     $589 = (125924 + ($588<<2)|0);
     $590 = HEAP32[$589>>2]|0;
     $591 = (($590) + ($587)|0);
     $592 = HEAP8[$591>>0]|0;
     $Pixel$2 = $550;$sA$2$shrunk = $592;$sB$2$shrunk = $583;$sG$2$shrunk = $572;$sR$2$shrunk = $561;
     break;
    }
    case 3:  {
     $593 = HEAP32[$36>>2]|0;
     $div27 = $593 >>> 3;
     $594 = $div27 & 31;
     $595 = (($src$3) + ($594)|0);
     $596 = HEAP8[$595>>0]|0;
     $597 = $593 >>> 11;
     $div28 = $597 & 31;
     $598 = (($src$3) + ($div28)|0);
     $599 = HEAP8[$598>>0]|0;
     $600 = $593 >>> 19;
     $601 = $600 & 31;
     $602 = (($src$3) + ($601)|0);
     $603 = HEAP8[$602>>0]|0;
     $Pixel$2 = 0;$sA$2$shrunk = -1;$sB$2$shrunk = $603;$sG$2$shrunk = $599;$sR$2$shrunk = $596;
     break;
    }
    case 4:  {
     $604 = HEAP32[$src$3>>2]|0;
     $605 = HEAP32[$35>>2]|0;
     $606 = $605 & $604;
     $607 = HEAP32[$36>>2]|0;
     $608 = $607 & 255;
     $609 = $606 >>> $608;
     $610 = HEAP32[$37>>2]|0;
     $611 = $610 & 255;
     $612 = (125924 + ($611<<2)|0);
     $613 = HEAP32[$612>>2]|0;
     $614 = (($613) + ($609)|0);
     $615 = HEAP8[$614>>0]|0;
     $616 = HEAP32[$38>>2]|0;
     $617 = $616 & $604;
     $618 = $607 >>> 8;
     $619 = $618 & 255;
     $620 = $617 >>> $619;
     $621 = $610 >>> 8;
     $622 = $621 & 255;
     $623 = (125924 + ($622<<2)|0);
     $624 = HEAP32[$623>>2]|0;
     $625 = (($624) + ($620)|0);
     $626 = HEAP8[$625>>0]|0;
     $627 = HEAP32[$39>>2]|0;
     $628 = $627 & $604;
     $629 = $607 >>> 16;
     $630 = $629 & 255;
     $631 = $628 >>> $630;
     $632 = $610 >>> 16;
     $633 = $632 & 255;
     $634 = (125924 + ($633<<2)|0);
     $635 = HEAP32[$634>>2]|0;
     $636 = (($635) + ($631)|0);
     $637 = HEAP8[$636>>0]|0;
     $638 = HEAP32[$14>>2]|0;
     $639 = $638 & $604;
     $640 = $607 >>> 24;
     $641 = $639 >>> $640;
     $642 = $610 >>> 24;
     $643 = (125924 + ($642<<2)|0);
     $644 = HEAP32[$643>>2]|0;
     $645 = (($644) + ($641)|0);
     $646 = HEAP8[$645>>0]|0;
     $Pixel$2 = $604;$sA$2$shrunk = $646;$sB$2$shrunk = $637;$sG$2$shrunk = $626;$sR$2$shrunk = $615;
     break;
    }
    default: {
     $Pixel$2 = 0;$sA$2$shrunk = 0;$sB$2$shrunk = 0;$sG$2$shrunk = 0;$sR$2$shrunk = 0;
    }
    }
    $sR$2 = $sR$2$shrunk&255;
    $sG$2 = $sG$2$shrunk&255;
    $sB$2 = $sB$2$shrunk&255;
    $sA$2 = $sA$2$shrunk&255;
    $647 = $Pixel$2 & $16;
    $648 = ($647|0)==($17|0);
    L45: do {
     if (!($648)) {
      switch ($30|0) {
      case 1:  {
       $649 = HEAP32[$31>>2]|0;
       $650 = $649 & 255;
       $651 = $sR$2 >>> $650;
       $652 = HEAP32[$32>>2]|0;
       $653 = $652 & 255;
       $654 = $651 << $653;
       $655 = $649 >>> 8;
       $656 = $655 & 255;
       $657 = $sG$2 >>> $656;
       $658 = $652 >>> 8;
       $659 = $658 & 255;
       $660 = $657 << $659;
       $661 = $660 | $654;
       $662 = $649 >>> 16;
       $663 = $662 & 255;
       $664 = $sB$2 >>> $663;
       $665 = $652 >>> 16;
       $666 = $665 & 255;
       $667 = $664 << $666;
       $668 = $661 | $667;
       $669 = $649 >>> 24;
       $670 = $sA$2 >>> $669;
       $671 = $652 >>> 24;
       $672 = $670 << $671;
       $673 = $668 | $672;
       $674 = $673&255;
       HEAP8[$dst$3>>0] = $674;
       break L45;
       break;
      }
      case 2:  {
       $675 = HEAP32[$31>>2]|0;
       $676 = $675 & 255;
       $677 = $sR$2 >>> $676;
       $678 = HEAP32[$32>>2]|0;
       $679 = $678 & 255;
       $680 = $677 << $679;
       $681 = $675 >>> 8;
       $682 = $681 & 255;
       $683 = $sG$2 >>> $682;
       $684 = $678 >>> 8;
       $685 = $684 & 255;
       $686 = $683 << $685;
       $687 = $686 | $680;
       $688 = $675 >>> 16;
       $689 = $688 & 255;
       $690 = $sB$2 >>> $689;
       $691 = $678 >>> 16;
       $692 = $691 & 255;
       $693 = $690 << $692;
       $694 = $687 | $693;
       $695 = $675 >>> 24;
       $696 = $sA$2 >>> $695;
       $697 = $678 >>> 24;
       $698 = $696 << $697;
       $699 = $694 | $698;
       $700 = $699&65535;
       HEAP16[$dst$3>>1] = $700;
       break L45;
       break;
      }
      case 3:  {
       $701 = HEAP8[$32>>0]|0;
       $div24 = ($701&255) >>> 3;
       $702 = $div24&255;
       $703 = (($dst$3) + ($702)|0);
       HEAP8[$703>>0] = $sR$2$shrunk;
       $704 = HEAP8[$33>>0]|0;
       $div25 = ($704&255) >>> 3;
       $705 = $div25&255;
       $706 = (($dst$3) + ($705)|0);
       HEAP8[$706>>0] = $sG$2$shrunk;
       $707 = HEAP8[$34>>0]|0;
       $div26 = ($707&255) >>> 3;
       $708 = $div26&255;
       $709 = (($dst$3) + ($708)|0);
       HEAP8[$709>>0] = $sB$2$shrunk;
       break L45;
       break;
      }
      case 4:  {
       $710 = HEAP32[$31>>2]|0;
       $711 = $710 & 255;
       $712 = $sR$2 >>> $711;
       $713 = HEAP32[$32>>2]|0;
       $714 = $713 & 255;
       $715 = $712 << $714;
       $716 = $710 >>> 8;
       $717 = $716 & 255;
       $718 = $sG$2 >>> $717;
       $719 = $713 >>> 8;
       $720 = $719 & 255;
       $721 = $718 << $720;
       $722 = $721 | $715;
       $723 = $710 >>> 16;
       $724 = $723 & 255;
       $725 = $sB$2 >>> $724;
       $726 = $713 >>> 16;
       $727 = $726 & 255;
       $728 = $725 << $727;
       $729 = $722 | $728;
       $730 = $710 >>> 24;
       $731 = $sA$2 >>> $730;
       $732 = $713 >>> 24;
       $733 = $731 << $732;
       $734 = $729 | $733;
       HEAP32[$dst$3>>2] = $734;
       break L45;
       break;
      }
      default: {
       break L45;
      }
      }
     }
    } while(0);
    $735 = (($dst$3) + ($30)|0);
    $736 = (($src$3) + ($29)|0);
    $dst$4 = $735;$n$3 = $n$2;$src$4 = $736;
    label = 40;
    continue;
   }
   else if ((label|0) == 40) {
    label = 0;
    switch ($29|0) {
    case 1:  {
     $737 = HEAP8[$src$4>>0]|0;
     $738 = $737&255;
     $739 = HEAP32[$35>>2]|0;
     $740 = $738 & $739;
     $741 = HEAP32[$36>>2]|0;
     $742 = $741 & 255;
     $743 = $740 >>> $742;
     $744 = HEAP32[$37>>2]|0;
     $745 = $744 & 255;
     $746 = (125924 + ($745<<2)|0);
     $747 = HEAP32[$746>>2]|0;
     $748 = (($747) + ($743)|0);
     $749 = HEAP8[$748>>0]|0;
     $750 = HEAP32[$38>>2]|0;
     $751 = $750 & $738;
     $752 = $741 >>> 8;
     $753 = $752 & 255;
     $754 = $751 >>> $753;
     $755 = $744 >>> 8;
     $756 = $755 & 255;
     $757 = (125924 + ($756<<2)|0);
     $758 = HEAP32[$757>>2]|0;
     $759 = (($758) + ($754)|0);
     $760 = HEAP8[$759>>0]|0;
     $761 = HEAP32[$39>>2]|0;
     $762 = $761 & $738;
     $763 = $741 >>> 16;
     $764 = $763 & 255;
     $765 = $762 >>> $764;
     $766 = $744 >>> 16;
     $767 = $766 & 255;
     $768 = (125924 + ($767<<2)|0);
     $769 = HEAP32[$768>>2]|0;
     $770 = (($769) + ($765)|0);
     $771 = HEAP8[$770>>0]|0;
     $772 = HEAP32[$14>>2]|0;
     $773 = $772 & $738;
     $774 = $741 >>> 24;
     $775 = $773 >>> $774;
     $776 = $744 >>> 24;
     $777 = (125924 + ($776<<2)|0);
     $778 = HEAP32[$777>>2]|0;
     $779 = (($778) + ($775)|0);
     $780 = HEAP8[$779>>0]|0;
     $Pixel$3 = $738;$sA$3$shrunk = $780;$sB$3$shrunk = $771;$sG$3$shrunk = $760;$sR$3$shrunk = $749;
     break;
    }
    case 2:  {
     $781 = HEAP16[$src$4>>1]|0;
     $782 = $781&65535;
     $783 = HEAP32[$35>>2]|0;
     $784 = $782 & $783;
     $785 = HEAP32[$36>>2]|0;
     $786 = $785 & 255;
     $787 = $784 >>> $786;
     $788 = HEAP32[$37>>2]|0;
     $789 = $788 & 255;
     $790 = (125924 + ($789<<2)|0);
     $791 = HEAP32[$790>>2]|0;
     $792 = (($791) + ($787)|0);
     $793 = HEAP8[$792>>0]|0;
     $794 = HEAP32[$38>>2]|0;
     $795 = $794 & $782;
     $796 = $785 >>> 8;
     $797 = $796 & 255;
     $798 = $795 >>> $797;
     $799 = $788 >>> 8;
     $800 = $799 & 255;
     $801 = (125924 + ($800<<2)|0);
     $802 = HEAP32[$801>>2]|0;
     $803 = (($802) + ($798)|0);
     $804 = HEAP8[$803>>0]|0;
     $805 = HEAP32[$39>>2]|0;
     $806 = $805 & $782;
     $807 = $785 >>> 16;
     $808 = $807 & 255;
     $809 = $806 >>> $808;
     $810 = $788 >>> 16;
     $811 = $810 & 255;
     $812 = (125924 + ($811<<2)|0);
     $813 = HEAP32[$812>>2]|0;
     $814 = (($813) + ($809)|0);
     $815 = HEAP8[$814>>0]|0;
     $816 = HEAP32[$14>>2]|0;
     $817 = $816 & $782;
     $818 = $785 >>> 24;
     $819 = $817 >>> $818;
     $820 = $788 >>> 24;
     $821 = (125924 + ($820<<2)|0);
     $822 = HEAP32[$821>>2]|0;
     $823 = (($822) + ($819)|0);
     $824 = HEAP8[$823>>0]|0;
     $Pixel$3 = $782;$sA$3$shrunk = $824;$sB$3$shrunk = $815;$sG$3$shrunk = $804;$sR$3$shrunk = $793;
     break;
    }
    case 3:  {
     $825 = HEAP32[$36>>2]|0;
     $div21 = $825 >>> 3;
     $826 = $div21 & 31;
     $827 = (($src$4) + ($826)|0);
     $828 = HEAP8[$827>>0]|0;
     $829 = $825 >>> 11;
     $div22 = $829 & 31;
     $830 = (($src$4) + ($div22)|0);
     $831 = HEAP8[$830>>0]|0;
     $832 = $825 >>> 19;
     $833 = $832 & 31;
     $834 = (($src$4) + ($833)|0);
     $835 = HEAP8[$834>>0]|0;
     $Pixel$3 = 0;$sA$3$shrunk = -1;$sB$3$shrunk = $835;$sG$3$shrunk = $831;$sR$3$shrunk = $828;
     break;
    }
    case 4:  {
     $836 = HEAP32[$src$4>>2]|0;
     $837 = HEAP32[$35>>2]|0;
     $838 = $837 & $836;
     $839 = HEAP32[$36>>2]|0;
     $840 = $839 & 255;
     $841 = $838 >>> $840;
     $842 = HEAP32[$37>>2]|0;
     $843 = $842 & 255;
     $844 = (125924 + ($843<<2)|0);
     $845 = HEAP32[$844>>2]|0;
     $846 = (($845) + ($841)|0);
     $847 = HEAP8[$846>>0]|0;
     $848 = HEAP32[$38>>2]|0;
     $849 = $848 & $836;
     $850 = $839 >>> 8;
     $851 = $850 & 255;
     $852 = $849 >>> $851;
     $853 = $842 >>> 8;
     $854 = $853 & 255;
     $855 = (125924 + ($854<<2)|0);
     $856 = HEAP32[$855>>2]|0;
     $857 = (($856) + ($852)|0);
     $858 = HEAP8[$857>>0]|0;
     $859 = HEAP32[$39>>2]|0;
     $860 = $859 & $836;
     $861 = $839 >>> 16;
     $862 = $861 & 255;
     $863 = $860 >>> $862;
     $864 = $842 >>> 16;
     $865 = $864 & 255;
     $866 = (125924 + ($865<<2)|0);
     $867 = HEAP32[$866>>2]|0;
     $868 = (($867) + ($863)|0);
     $869 = HEAP8[$868>>0]|0;
     $870 = HEAP32[$14>>2]|0;
     $871 = $870 & $836;
     $872 = $839 >>> 24;
     $873 = $871 >>> $872;
     $874 = $842 >>> 24;
     $875 = (125924 + ($874<<2)|0);
     $876 = HEAP32[$875>>2]|0;
     $877 = (($876) + ($873)|0);
     $878 = HEAP8[$877>>0]|0;
     $Pixel$3 = $836;$sA$3$shrunk = $878;$sB$3$shrunk = $869;$sG$3$shrunk = $858;$sR$3$shrunk = $847;
     break;
    }
    default: {
     $Pixel$3 = 0;$sA$3$shrunk = 0;$sB$3$shrunk = 0;$sG$3$shrunk = 0;$sR$3$shrunk = 0;
    }
    }
    $sR$3 = $sR$3$shrunk&255;
    $sG$3 = $sG$3$shrunk&255;
    $sB$3 = $sB$3$shrunk&255;
    $sA$3 = $sA$3$shrunk&255;
    $879 = $Pixel$3 & $16;
    $880 = ($879|0)==($17|0);
    L60: do {
     if (!($880)) {
      switch ($30|0) {
      case 1:  {
       $881 = HEAP32[$31>>2]|0;
       $882 = $881 & 255;
       $883 = $sR$3 >>> $882;
       $884 = HEAP32[$32>>2]|0;
       $885 = $884 & 255;
       $886 = $883 << $885;
       $887 = $881 >>> 8;
       $888 = $887 & 255;
       $889 = $sG$3 >>> $888;
       $890 = $884 >>> 8;
       $891 = $890 & 255;
       $892 = $889 << $891;
       $893 = $892 | $886;
       $894 = $881 >>> 16;
       $895 = $894 & 255;
       $896 = $sB$3 >>> $895;
       $897 = $884 >>> 16;
       $898 = $897 & 255;
       $899 = $896 << $898;
       $900 = $893 | $899;
       $901 = $881 >>> 24;
       $902 = $sA$3 >>> $901;
       $903 = $884 >>> 24;
       $904 = $902 << $903;
       $905 = $900 | $904;
       $906 = $905&255;
       HEAP8[$dst$4>>0] = $906;
       break L60;
       break;
      }
      case 2:  {
       $907 = HEAP32[$31>>2]|0;
       $908 = $907 & 255;
       $909 = $sR$3 >>> $908;
       $910 = HEAP32[$32>>2]|0;
       $911 = $910 & 255;
       $912 = $909 << $911;
       $913 = $907 >>> 8;
       $914 = $913 & 255;
       $915 = $sG$3 >>> $914;
       $916 = $910 >>> 8;
       $917 = $916 & 255;
       $918 = $915 << $917;
       $919 = $918 | $912;
       $920 = $907 >>> 16;
       $921 = $920 & 255;
       $922 = $sB$3 >>> $921;
       $923 = $910 >>> 16;
       $924 = $923 & 255;
       $925 = $922 << $924;
       $926 = $919 | $925;
       $927 = $907 >>> 24;
       $928 = $sA$3 >>> $927;
       $929 = $910 >>> 24;
       $930 = $928 << $929;
       $931 = $926 | $930;
       $932 = $931&65535;
       HEAP16[$dst$4>>1] = $932;
       break L60;
       break;
      }
      case 3:  {
       $933 = HEAP8[$32>>0]|0;
       $div18 = ($933&255) >>> 3;
       $934 = $div18&255;
       $935 = (($dst$4) + ($934)|0);
       HEAP8[$935>>0] = $sR$3$shrunk;
       $936 = HEAP8[$33>>0]|0;
       $div19 = ($936&255) >>> 3;
       $937 = $div19&255;
       $938 = (($dst$4) + ($937)|0);
       HEAP8[$938>>0] = $sG$3$shrunk;
       $939 = HEAP8[$34>>0]|0;
       $div20 = ($939&255) >>> 3;
       $940 = $div20&255;
       $941 = (($dst$4) + ($940)|0);
       HEAP8[$941>>0] = $sB$3$shrunk;
       break L60;
       break;
      }
      case 4:  {
       $942 = HEAP32[$31>>2]|0;
       $943 = $942 & 255;
       $944 = $sR$3 >>> $943;
       $945 = HEAP32[$32>>2]|0;
       $946 = $945 & 255;
       $947 = $944 << $946;
       $948 = $942 >>> 8;
       $949 = $948 & 255;
       $950 = $sG$3 >>> $949;
       $951 = $945 >>> 8;
       $952 = $951 & 255;
       $953 = $950 << $952;
       $954 = $953 | $947;
       $955 = $942 >>> 16;
       $956 = $955 & 255;
       $957 = $sB$3 >>> $956;
       $958 = $945 >>> 16;
       $959 = $958 & 255;
       $960 = $957 << $959;
       $961 = $954 | $960;
       $962 = $942 >>> 24;
       $963 = $sA$3 >>> $962;
       $964 = $945 >>> 24;
       $965 = $963 << $964;
       $966 = $961 | $965;
       HEAP32[$dst$4>>2] = $966;
       break L60;
       break;
      }
      default: {
       break L60;
      }
      }
     }
    } while(0);
    $967 = (($dst$4) + ($30)|0);
    $968 = (($src$4) + ($29)|0);
    $dst$5 = $967;$n$4 = $n$3;$src$5 = $968;
    label = 52;
    continue;
   }
   else if ((label|0) == 52) {
    label = 0;
    switch ($29|0) {
    case 1:  {
     $969 = HEAP8[$src$5>>0]|0;
     $970 = $969&255;
     $971 = HEAP32[$35>>2]|0;
     $972 = $970 & $971;
     $973 = HEAP32[$36>>2]|0;
     $974 = $973 & 255;
     $975 = $972 >>> $974;
     $976 = HEAP32[$37>>2]|0;
     $977 = $976 & 255;
     $978 = (125924 + ($977<<2)|0);
     $979 = HEAP32[$978>>2]|0;
     $980 = (($979) + ($975)|0);
     $981 = HEAP8[$980>>0]|0;
     $982 = HEAP32[$38>>2]|0;
     $983 = $982 & $970;
     $984 = $973 >>> 8;
     $985 = $984 & 255;
     $986 = $983 >>> $985;
     $987 = $976 >>> 8;
     $988 = $987 & 255;
     $989 = (125924 + ($988<<2)|0);
     $990 = HEAP32[$989>>2]|0;
     $991 = (($990) + ($986)|0);
     $992 = HEAP8[$991>>0]|0;
     $993 = HEAP32[$39>>2]|0;
     $994 = $993 & $970;
     $995 = $973 >>> 16;
     $996 = $995 & 255;
     $997 = $994 >>> $996;
     $998 = $976 >>> 16;
     $999 = $998 & 255;
     $1000 = (125924 + ($999<<2)|0);
     $1001 = HEAP32[$1000>>2]|0;
     $1002 = (($1001) + ($997)|0);
     $1003 = HEAP8[$1002>>0]|0;
     $1004 = HEAP32[$14>>2]|0;
     $1005 = $1004 & $970;
     $1006 = $973 >>> 24;
     $1007 = $1005 >>> $1006;
     $1008 = $976 >>> 24;
     $1009 = (125924 + ($1008<<2)|0);
     $1010 = HEAP32[$1009>>2]|0;
     $1011 = (($1010) + ($1007)|0);
     $1012 = HEAP8[$1011>>0]|0;
     $Pixel$4 = $970;$sA$4$shrunk = $1012;$sB$4$shrunk = $1003;$sG$4$shrunk = $992;$sR$4$shrunk = $981;
     break;
    }
    case 2:  {
     $1013 = HEAP16[$src$5>>1]|0;
     $1014 = $1013&65535;
     $1015 = HEAP32[$35>>2]|0;
     $1016 = $1014 & $1015;
     $1017 = HEAP32[$36>>2]|0;
     $1018 = $1017 & 255;
     $1019 = $1016 >>> $1018;
     $1020 = HEAP32[$37>>2]|0;
     $1021 = $1020 & 255;
     $1022 = (125924 + ($1021<<2)|0);
     $1023 = HEAP32[$1022>>2]|0;
     $1024 = (($1023) + ($1019)|0);
     $1025 = HEAP8[$1024>>0]|0;
     $1026 = HEAP32[$38>>2]|0;
     $1027 = $1026 & $1014;
     $1028 = $1017 >>> 8;
     $1029 = $1028 & 255;
     $1030 = $1027 >>> $1029;
     $1031 = $1020 >>> 8;
     $1032 = $1031 & 255;
     $1033 = (125924 + ($1032<<2)|0);
     $1034 = HEAP32[$1033>>2]|0;
     $1035 = (($1034) + ($1030)|0);
     $1036 = HEAP8[$1035>>0]|0;
     $1037 = HEAP32[$39>>2]|0;
     $1038 = $1037 & $1014;
     $1039 = $1017 >>> 16;
     $1040 = $1039 & 255;
     $1041 = $1038 >>> $1040;
     $1042 = $1020 >>> 16;
     $1043 = $1042 & 255;
     $1044 = (125924 + ($1043<<2)|0);
     $1045 = HEAP32[$1044>>2]|0;
     $1046 = (($1045) + ($1041)|0);
     $1047 = HEAP8[$1046>>0]|0;
     $1048 = HEAP32[$14>>2]|0;
     $1049 = $1048 & $1014;
     $1050 = $1017 >>> 24;
     $1051 = $1049 >>> $1050;
     $1052 = $1020 >>> 24;
     $1053 = (125924 + ($1052<<2)|0);
     $1054 = HEAP32[$1053>>2]|0;
     $1055 = (($1054) + ($1051)|0);
     $1056 = HEAP8[$1055>>0]|0;
     $Pixel$4 = $1014;$sA$4$shrunk = $1056;$sB$4$shrunk = $1047;$sG$4$shrunk = $1036;$sR$4$shrunk = $1025;
     break;
    }
    case 3:  {
     $1057 = HEAP32[$36>>2]|0;
     $div15 = $1057 >>> 3;
     $1058 = $div15 & 31;
     $1059 = (($src$5) + ($1058)|0);
     $1060 = HEAP8[$1059>>0]|0;
     $1061 = $1057 >>> 11;
     $div16 = $1061 & 31;
     $1062 = (($src$5) + ($div16)|0);
     $1063 = HEAP8[$1062>>0]|0;
     $1064 = $1057 >>> 19;
     $1065 = $1064 & 31;
     $1066 = (($src$5) + ($1065)|0);
     $1067 = HEAP8[$1066>>0]|0;
     $Pixel$4 = 0;$sA$4$shrunk = -1;$sB$4$shrunk = $1067;$sG$4$shrunk = $1063;$sR$4$shrunk = $1060;
     break;
    }
    case 4:  {
     $1068 = HEAP32[$src$5>>2]|0;
     $1069 = HEAP32[$35>>2]|0;
     $1070 = $1069 & $1068;
     $1071 = HEAP32[$36>>2]|0;
     $1072 = $1071 & 255;
     $1073 = $1070 >>> $1072;
     $1074 = HEAP32[$37>>2]|0;
     $1075 = $1074 & 255;
     $1076 = (125924 + ($1075<<2)|0);
     $1077 = HEAP32[$1076>>2]|0;
     $1078 = (($1077) + ($1073)|0);
     $1079 = HEAP8[$1078>>0]|0;
     $1080 = HEAP32[$38>>2]|0;
     $1081 = $1080 & $1068;
     $1082 = $1071 >>> 8;
     $1083 = $1082 & 255;
     $1084 = $1081 >>> $1083;
     $1085 = $1074 >>> 8;
     $1086 = $1085 & 255;
     $1087 = (125924 + ($1086<<2)|0);
     $1088 = HEAP32[$1087>>2]|0;
     $1089 = (($1088) + ($1084)|0);
     $1090 = HEAP8[$1089>>0]|0;
     $1091 = HEAP32[$39>>2]|0;
     $1092 = $1091 & $1068;
     $1093 = $1071 >>> 16;
     $1094 = $1093 & 255;
     $1095 = $1092 >>> $1094;
     $1096 = $1074 >>> 16;
     $1097 = $1096 & 255;
     $1098 = (125924 + ($1097<<2)|0);
     $1099 = HEAP32[$1098>>2]|0;
     $1100 = (($1099) + ($1095)|0);
     $1101 = HEAP8[$1100>>0]|0;
     $1102 = HEAP32[$14>>2]|0;
     $1103 = $1102 & $1068;
     $1104 = $1071 >>> 24;
     $1105 = $1103 >>> $1104;
     $1106 = $1074 >>> 24;
     $1107 = (125924 + ($1106<<2)|0);
     $1108 = HEAP32[$1107>>2]|0;
     $1109 = (($1108) + ($1105)|0);
     $1110 = HEAP8[$1109>>0]|0;
     $Pixel$4 = $1068;$sA$4$shrunk = $1110;$sB$4$shrunk = $1101;$sG$4$shrunk = $1090;$sR$4$shrunk = $1079;
     break;
    }
    default: {
     $Pixel$4 = 0;$sA$4$shrunk = 0;$sB$4$shrunk = 0;$sG$4$shrunk = 0;$sR$4$shrunk = 0;
    }
    }
    $sR$4 = $sR$4$shrunk&255;
    $sG$4 = $sG$4$shrunk&255;
    $sB$4 = $sB$4$shrunk&255;
    $sA$4 = $sA$4$shrunk&255;
    $1111 = $Pixel$4 & $16;
    $1112 = ($1111|0)==($17|0);
    L75: do {
     if (!($1112)) {
      switch ($30|0) {
      case 1:  {
       $1113 = HEAP32[$31>>2]|0;
       $1114 = $1113 & 255;
       $1115 = $sR$4 >>> $1114;
       $1116 = HEAP32[$32>>2]|0;
       $1117 = $1116 & 255;
       $1118 = $1115 << $1117;
       $1119 = $1113 >>> 8;
       $1120 = $1119 & 255;
       $1121 = $sG$4 >>> $1120;
       $1122 = $1116 >>> 8;
       $1123 = $1122 & 255;
       $1124 = $1121 << $1123;
       $1125 = $1124 | $1118;
       $1126 = $1113 >>> 16;
       $1127 = $1126 & 255;
       $1128 = $sB$4 >>> $1127;
       $1129 = $1116 >>> 16;
       $1130 = $1129 & 255;
       $1131 = $1128 << $1130;
       $1132 = $1125 | $1131;
       $1133 = $1113 >>> 24;
       $1134 = $sA$4 >>> $1133;
       $1135 = $1116 >>> 24;
       $1136 = $1134 << $1135;
       $1137 = $1132 | $1136;
       $1138 = $1137&255;
       HEAP8[$dst$5>>0] = $1138;
       break L75;
       break;
      }
      case 2:  {
       $1139 = HEAP32[$31>>2]|0;
       $1140 = $1139 & 255;
       $1141 = $sR$4 >>> $1140;
       $1142 = HEAP32[$32>>2]|0;
       $1143 = $1142 & 255;
       $1144 = $1141 << $1143;
       $1145 = $1139 >>> 8;
       $1146 = $1145 & 255;
       $1147 = $sG$4 >>> $1146;
       $1148 = $1142 >>> 8;
       $1149 = $1148 & 255;
       $1150 = $1147 << $1149;
       $1151 = $1150 | $1144;
       $1152 = $1139 >>> 16;
       $1153 = $1152 & 255;
       $1154 = $sB$4 >>> $1153;
       $1155 = $1142 >>> 16;
       $1156 = $1155 & 255;
       $1157 = $1154 << $1156;
       $1158 = $1151 | $1157;
       $1159 = $1139 >>> 24;
       $1160 = $sA$4 >>> $1159;
       $1161 = $1142 >>> 24;
       $1162 = $1160 << $1161;
       $1163 = $1158 | $1162;
       $1164 = $1163&65535;
       HEAP16[$dst$5>>1] = $1164;
       break L75;
       break;
      }
      case 3:  {
       $1165 = HEAP8[$32>>0]|0;
       $div12 = ($1165&255) >>> 3;
       $1166 = $div12&255;
       $1167 = (($dst$5) + ($1166)|0);
       HEAP8[$1167>>0] = $sR$4$shrunk;
       $1168 = HEAP8[$33>>0]|0;
       $div13 = ($1168&255) >>> 3;
       $1169 = $div13&255;
       $1170 = (($dst$5) + ($1169)|0);
       HEAP8[$1170>>0] = $sG$4$shrunk;
       $1171 = HEAP8[$34>>0]|0;
       $div14 = ($1171&255) >>> 3;
       $1172 = $div14&255;
       $1173 = (($dst$5) + ($1172)|0);
       HEAP8[$1173>>0] = $sB$4$shrunk;
       break L75;
       break;
      }
      case 4:  {
       $1174 = HEAP32[$31>>2]|0;
       $1175 = $1174 & 255;
       $1176 = $sR$4 >>> $1175;
       $1177 = HEAP32[$32>>2]|0;
       $1178 = $1177 & 255;
       $1179 = $1176 << $1178;
       $1180 = $1174 >>> 8;
       $1181 = $1180 & 255;
       $1182 = $sG$4 >>> $1181;
       $1183 = $1177 >>> 8;
       $1184 = $1183 & 255;
       $1185 = $1182 << $1184;
       $1186 = $1185 | $1179;
       $1187 = $1174 >>> 16;
       $1188 = $1187 & 255;
       $1189 = $sB$4 >>> $1188;
       $1190 = $1177 >>> 16;
       $1191 = $1190 & 255;
       $1192 = $1189 << $1191;
       $1193 = $1186 | $1192;
       $1194 = $1174 >>> 24;
       $1195 = $sA$4 >>> $1194;
       $1196 = $1177 >>> 24;
       $1197 = $1195 << $1196;
       $1198 = $1193 | $1197;
       HEAP32[$dst$5>>2] = $1198;
       break L75;
       break;
      }
      default: {
       break L75;
      }
      }
     }
    } while(0);
    $1199 = (($dst$5) + ($30)|0);
    $1200 = (($src$5) + ($29)|0);
    $dst$6 = $1199;$n$5 = $n$4;$src$6 = $1200;
    label = 64;
    continue;
   }
   else if ((label|0) == 64) {
    label = 0;
    switch ($29|0) {
    case 1:  {
     $1201 = HEAP8[$src$6>>0]|0;
     $1202 = $1201&255;
     $1203 = HEAP32[$35>>2]|0;
     $1204 = $1202 & $1203;
     $1205 = HEAP32[$36>>2]|0;
     $1206 = $1205 & 255;
     $1207 = $1204 >>> $1206;
     $1208 = HEAP32[$37>>2]|0;
     $1209 = $1208 & 255;
     $1210 = (125924 + ($1209<<2)|0);
     $1211 = HEAP32[$1210>>2]|0;
     $1212 = (($1211) + ($1207)|0);
     $1213 = HEAP8[$1212>>0]|0;
     $1214 = HEAP32[$38>>2]|0;
     $1215 = $1214 & $1202;
     $1216 = $1205 >>> 8;
     $1217 = $1216 & 255;
     $1218 = $1215 >>> $1217;
     $1219 = $1208 >>> 8;
     $1220 = $1219 & 255;
     $1221 = (125924 + ($1220<<2)|0);
     $1222 = HEAP32[$1221>>2]|0;
     $1223 = (($1222) + ($1218)|0);
     $1224 = HEAP8[$1223>>0]|0;
     $1225 = HEAP32[$39>>2]|0;
     $1226 = $1225 & $1202;
     $1227 = $1205 >>> 16;
     $1228 = $1227 & 255;
     $1229 = $1226 >>> $1228;
     $1230 = $1208 >>> 16;
     $1231 = $1230 & 255;
     $1232 = (125924 + ($1231<<2)|0);
     $1233 = HEAP32[$1232>>2]|0;
     $1234 = (($1233) + ($1229)|0);
     $1235 = HEAP8[$1234>>0]|0;
     $1236 = HEAP32[$14>>2]|0;
     $1237 = $1236 & $1202;
     $1238 = $1205 >>> 24;
     $1239 = $1237 >>> $1238;
     $1240 = $1208 >>> 24;
     $1241 = (125924 + ($1240<<2)|0);
     $1242 = HEAP32[$1241>>2]|0;
     $1243 = (($1242) + ($1239)|0);
     $1244 = HEAP8[$1243>>0]|0;
     $Pixel$5 = $1202;$sA$5$shrunk = $1244;$sB$5$shrunk = $1235;$sG$5$shrunk = $1224;$sR$5$shrunk = $1213;
     break;
    }
    case 2:  {
     $1245 = HEAP16[$src$6>>1]|0;
     $1246 = $1245&65535;
     $1247 = HEAP32[$35>>2]|0;
     $1248 = $1246 & $1247;
     $1249 = HEAP32[$36>>2]|0;
     $1250 = $1249 & 255;
     $1251 = $1248 >>> $1250;
     $1252 = HEAP32[$37>>2]|0;
     $1253 = $1252 & 255;
     $1254 = (125924 + ($1253<<2)|0);
     $1255 = HEAP32[$1254>>2]|0;
     $1256 = (($1255) + ($1251)|0);
     $1257 = HEAP8[$1256>>0]|0;
     $1258 = HEAP32[$38>>2]|0;
     $1259 = $1258 & $1246;
     $1260 = $1249 >>> 8;
     $1261 = $1260 & 255;
     $1262 = $1259 >>> $1261;
     $1263 = $1252 >>> 8;
     $1264 = $1263 & 255;
     $1265 = (125924 + ($1264<<2)|0);
     $1266 = HEAP32[$1265>>2]|0;
     $1267 = (($1266) + ($1262)|0);
     $1268 = HEAP8[$1267>>0]|0;
     $1269 = HEAP32[$39>>2]|0;
     $1270 = $1269 & $1246;
     $1271 = $1249 >>> 16;
     $1272 = $1271 & 255;
     $1273 = $1270 >>> $1272;
     $1274 = $1252 >>> 16;
     $1275 = $1274 & 255;
     $1276 = (125924 + ($1275<<2)|0);
     $1277 = HEAP32[$1276>>2]|0;
     $1278 = (($1277) + ($1273)|0);
     $1279 = HEAP8[$1278>>0]|0;
     $1280 = HEAP32[$14>>2]|0;
     $1281 = $1280 & $1246;
     $1282 = $1249 >>> 24;
     $1283 = $1281 >>> $1282;
     $1284 = $1252 >>> 24;
     $1285 = (125924 + ($1284<<2)|0);
     $1286 = HEAP32[$1285>>2]|0;
     $1287 = (($1286) + ($1283)|0);
     $1288 = HEAP8[$1287>>0]|0;
     $Pixel$5 = $1246;$sA$5$shrunk = $1288;$sB$5$shrunk = $1279;$sG$5$shrunk = $1268;$sR$5$shrunk = $1257;
     break;
    }
    case 3:  {
     $1289 = HEAP32[$36>>2]|0;
     $div9 = $1289 >>> 3;
     $1290 = $div9 & 31;
     $1291 = (($src$6) + ($1290)|0);
     $1292 = HEAP8[$1291>>0]|0;
     $1293 = $1289 >>> 11;
     $div10 = $1293 & 31;
     $1294 = (($src$6) + ($div10)|0);
     $1295 = HEAP8[$1294>>0]|0;
     $1296 = $1289 >>> 19;
     $1297 = $1296 & 31;
     $1298 = (($src$6) + ($1297)|0);
     $1299 = HEAP8[$1298>>0]|0;
     $Pixel$5 = 0;$sA$5$shrunk = -1;$sB$5$shrunk = $1299;$sG$5$shrunk = $1295;$sR$5$shrunk = $1292;
     break;
    }
    case 4:  {
     $1300 = HEAP32[$src$6>>2]|0;
     $1301 = HEAP32[$35>>2]|0;
     $1302 = $1301 & $1300;
     $1303 = HEAP32[$36>>2]|0;
     $1304 = $1303 & 255;
     $1305 = $1302 >>> $1304;
     $1306 = HEAP32[$37>>2]|0;
     $1307 = $1306 & 255;
     $1308 = (125924 + ($1307<<2)|0);
     $1309 = HEAP32[$1308>>2]|0;
     $1310 = (($1309) + ($1305)|0);
     $1311 = HEAP8[$1310>>0]|0;
     $1312 = HEAP32[$38>>2]|0;
     $1313 = $1312 & $1300;
     $1314 = $1303 >>> 8;
     $1315 = $1314 & 255;
     $1316 = $1313 >>> $1315;
     $1317 = $1306 >>> 8;
     $1318 = $1317 & 255;
     $1319 = (125924 + ($1318<<2)|0);
     $1320 = HEAP32[$1319>>2]|0;
     $1321 = (($1320) + ($1316)|0);
     $1322 = HEAP8[$1321>>0]|0;
     $1323 = HEAP32[$39>>2]|0;
     $1324 = $1323 & $1300;
     $1325 = $1303 >>> 16;
     $1326 = $1325 & 255;
     $1327 = $1324 >>> $1326;
     $1328 = $1306 >>> 16;
     $1329 = $1328 & 255;
     $1330 = (125924 + ($1329<<2)|0);
     $1331 = HEAP32[$1330>>2]|0;
     $1332 = (($1331) + ($1327)|0);
     $1333 = HEAP8[$1332>>0]|0;
     $1334 = HEAP32[$14>>2]|0;
     $1335 = $1334 & $1300;
     $1336 = $1303 >>> 24;
     $1337 = $1335 >>> $1336;
     $1338 = $1306 >>> 24;
     $1339 = (125924 + ($1338<<2)|0);
     $1340 = HEAP32[$1339>>2]|0;
     $1341 = (($1340) + ($1337)|0);
     $1342 = HEAP8[$1341>>0]|0;
     $Pixel$5 = $1300;$sA$5$shrunk = $1342;$sB$5$shrunk = $1333;$sG$5$shrunk = $1322;$sR$5$shrunk = $1311;
     break;
    }
    default: {
     $Pixel$5 = 0;$sA$5$shrunk = 0;$sB$5$shrunk = 0;$sG$5$shrunk = 0;$sR$5$shrunk = 0;
    }
    }
    $sR$5 = $sR$5$shrunk&255;
    $sG$5 = $sG$5$shrunk&255;
    $sB$5 = $sB$5$shrunk&255;
    $sA$5 = $sA$5$shrunk&255;
    $1343 = $Pixel$5 & $16;
    $1344 = ($1343|0)==($17|0);
    L90: do {
     if (!($1344)) {
      switch ($30|0) {
      case 1:  {
       $1345 = HEAP32[$31>>2]|0;
       $1346 = $1345 & 255;
       $1347 = $sR$5 >>> $1346;
       $1348 = HEAP32[$32>>2]|0;
       $1349 = $1348 & 255;
       $1350 = $1347 << $1349;
       $1351 = $1345 >>> 8;
       $1352 = $1351 & 255;
       $1353 = $sG$5 >>> $1352;
       $1354 = $1348 >>> 8;
       $1355 = $1354 & 255;
       $1356 = $1353 << $1355;
       $1357 = $1356 | $1350;
       $1358 = $1345 >>> 16;
       $1359 = $1358 & 255;
       $1360 = $sB$5 >>> $1359;
       $1361 = $1348 >>> 16;
       $1362 = $1361 & 255;
       $1363 = $1360 << $1362;
       $1364 = $1357 | $1363;
       $1365 = $1345 >>> 24;
       $1366 = $sA$5 >>> $1365;
       $1367 = $1348 >>> 24;
       $1368 = $1366 << $1367;
       $1369 = $1364 | $1368;
       $1370 = $1369&255;
       HEAP8[$dst$6>>0] = $1370;
       break L90;
       break;
      }
      case 2:  {
       $1371 = HEAP32[$31>>2]|0;
       $1372 = $1371 & 255;
       $1373 = $sR$5 >>> $1372;
       $1374 = HEAP32[$32>>2]|0;
       $1375 = $1374 & 255;
       $1376 = $1373 << $1375;
       $1377 = $1371 >>> 8;
       $1378 = $1377 & 255;
       $1379 = $sG$5 >>> $1378;
       $1380 = $1374 >>> 8;
       $1381 = $1380 & 255;
       $1382 = $1379 << $1381;
       $1383 = $1382 | $1376;
       $1384 = $1371 >>> 16;
       $1385 = $1384 & 255;
       $1386 = $sB$5 >>> $1385;
       $1387 = $1374 >>> 16;
       $1388 = $1387 & 255;
       $1389 = $1386 << $1388;
       $1390 = $1383 | $1389;
       $1391 = $1371 >>> 24;
       $1392 = $sA$5 >>> $1391;
       $1393 = $1374 >>> 24;
       $1394 = $1392 << $1393;
       $1395 = $1390 | $1394;
       $1396 = $1395&65535;
       HEAP16[$dst$6>>1] = $1396;
       break L90;
       break;
      }
      case 3:  {
       $1397 = HEAP8[$32>>0]|0;
       $div6 = ($1397&255) >>> 3;
       $1398 = $div6&255;
       $1399 = (($dst$6) + ($1398)|0);
       HEAP8[$1399>>0] = $sR$5$shrunk;
       $1400 = HEAP8[$33>>0]|0;
       $div7 = ($1400&255) >>> 3;
       $1401 = $div7&255;
       $1402 = (($dst$6) + ($1401)|0);
       HEAP8[$1402>>0] = $sG$5$shrunk;
       $1403 = HEAP8[$34>>0]|0;
       $div8 = ($1403&255) >>> 3;
       $1404 = $div8&255;
       $1405 = (($dst$6) + ($1404)|0);
       HEAP8[$1405>>0] = $sB$5$shrunk;
       break L90;
       break;
      }
      case 4:  {
       $1406 = HEAP32[$31>>2]|0;
       $1407 = $1406 & 255;
       $1408 = $sR$5 >>> $1407;
       $1409 = HEAP32[$32>>2]|0;
       $1410 = $1409 & 255;
       $1411 = $1408 << $1410;
       $1412 = $1406 >>> 8;
       $1413 = $1412 & 255;
       $1414 = $sG$5 >>> $1413;
       $1415 = $1409 >>> 8;
       $1416 = $1415 & 255;
       $1417 = $1414 << $1416;
       $1418 = $1417 | $1411;
       $1419 = $1406 >>> 16;
       $1420 = $1419 & 255;
       $1421 = $sB$5 >>> $1420;
       $1422 = $1409 >>> 16;
       $1423 = $1422 & 255;
       $1424 = $1421 << $1423;
       $1425 = $1418 | $1424;
       $1426 = $1406 >>> 24;
       $1427 = $sA$5 >>> $1426;
       $1428 = $1409 >>> 24;
       $1429 = $1427 << $1428;
       $1430 = $1425 | $1429;
       HEAP32[$dst$6>>2] = $1430;
       break L90;
       break;
      }
      default: {
       break L90;
      }
      }
     }
    } while(0);
    $1431 = (($dst$6) + ($30)|0);
    $1432 = (($src$6) + ($29)|0);
    $dst$7 = $1431;$n$6 = $n$5;$src$7 = $1432;
    label = 76;
    continue;
   }
   else if ((label|0) == 76) {
    label = 0;
    switch ($29|0) {
    case 1:  {
     $1433 = HEAP8[$src$7>>0]|0;
     $1434 = $1433&255;
     $1435 = HEAP32[$35>>2]|0;
     $1436 = $1434 & $1435;
     $1437 = HEAP32[$36>>2]|0;
     $1438 = $1437 & 255;
     $1439 = $1436 >>> $1438;
     $1440 = HEAP32[$37>>2]|0;
     $1441 = $1440 & 255;
     $1442 = (125924 + ($1441<<2)|0);
     $1443 = HEAP32[$1442>>2]|0;
     $1444 = (($1443) + ($1439)|0);
     $1445 = HEAP8[$1444>>0]|0;
     $1446 = HEAP32[$38>>2]|0;
     $1447 = $1446 & $1434;
     $1448 = $1437 >>> 8;
     $1449 = $1448 & 255;
     $1450 = $1447 >>> $1449;
     $1451 = $1440 >>> 8;
     $1452 = $1451 & 255;
     $1453 = (125924 + ($1452<<2)|0);
     $1454 = HEAP32[$1453>>2]|0;
     $1455 = (($1454) + ($1450)|0);
     $1456 = HEAP8[$1455>>0]|0;
     $1457 = HEAP32[$39>>2]|0;
     $1458 = $1457 & $1434;
     $1459 = $1437 >>> 16;
     $1460 = $1459 & 255;
     $1461 = $1458 >>> $1460;
     $1462 = $1440 >>> 16;
     $1463 = $1462 & 255;
     $1464 = (125924 + ($1463<<2)|0);
     $1465 = HEAP32[$1464>>2]|0;
     $1466 = (($1465) + ($1461)|0);
     $1467 = HEAP8[$1466>>0]|0;
     $1468 = HEAP32[$14>>2]|0;
     $1469 = $1468 & $1434;
     $1470 = $1437 >>> 24;
     $1471 = $1469 >>> $1470;
     $1472 = $1440 >>> 24;
     $1473 = (125924 + ($1472<<2)|0);
     $1474 = HEAP32[$1473>>2]|0;
     $1475 = (($1474) + ($1471)|0);
     $1476 = HEAP8[$1475>>0]|0;
     $Pixel$6 = $1434;$sA$6$shrunk = $1476;$sB$6$shrunk = $1467;$sG$6$shrunk = $1456;$sR$6$shrunk = $1445;
     break;
    }
    case 2:  {
     $1477 = HEAP16[$src$7>>1]|0;
     $1478 = $1477&65535;
     $1479 = HEAP32[$35>>2]|0;
     $1480 = $1478 & $1479;
     $1481 = HEAP32[$36>>2]|0;
     $1482 = $1481 & 255;
     $1483 = $1480 >>> $1482;
     $1484 = HEAP32[$37>>2]|0;
     $1485 = $1484 & 255;
     $1486 = (125924 + ($1485<<2)|0);
     $1487 = HEAP32[$1486>>2]|0;
     $1488 = (($1487) + ($1483)|0);
     $1489 = HEAP8[$1488>>0]|0;
     $1490 = HEAP32[$38>>2]|0;
     $1491 = $1490 & $1478;
     $1492 = $1481 >>> 8;
     $1493 = $1492 & 255;
     $1494 = $1491 >>> $1493;
     $1495 = $1484 >>> 8;
     $1496 = $1495 & 255;
     $1497 = (125924 + ($1496<<2)|0);
     $1498 = HEAP32[$1497>>2]|0;
     $1499 = (($1498) + ($1494)|0);
     $1500 = HEAP8[$1499>>0]|0;
     $1501 = HEAP32[$39>>2]|0;
     $1502 = $1501 & $1478;
     $1503 = $1481 >>> 16;
     $1504 = $1503 & 255;
     $1505 = $1502 >>> $1504;
     $1506 = $1484 >>> 16;
     $1507 = $1506 & 255;
     $1508 = (125924 + ($1507<<2)|0);
     $1509 = HEAP32[$1508>>2]|0;
     $1510 = (($1509) + ($1505)|0);
     $1511 = HEAP8[$1510>>0]|0;
     $1512 = HEAP32[$14>>2]|0;
     $1513 = $1512 & $1478;
     $1514 = $1481 >>> 24;
     $1515 = $1513 >>> $1514;
     $1516 = $1484 >>> 24;
     $1517 = (125924 + ($1516<<2)|0);
     $1518 = HEAP32[$1517>>2]|0;
     $1519 = (($1518) + ($1515)|0);
     $1520 = HEAP8[$1519>>0]|0;
     $Pixel$6 = $1478;$sA$6$shrunk = $1520;$sB$6$shrunk = $1511;$sG$6$shrunk = $1500;$sR$6$shrunk = $1489;
     break;
    }
    case 3:  {
     $1521 = HEAP32[$36>>2]|0;
     $div3 = $1521 >>> 3;
     $1522 = $div3 & 31;
     $1523 = (($src$7) + ($1522)|0);
     $1524 = HEAP8[$1523>>0]|0;
     $1525 = $1521 >>> 11;
     $div4 = $1525 & 31;
     $1526 = (($src$7) + ($div4)|0);
     $1527 = HEAP8[$1526>>0]|0;
     $1528 = $1521 >>> 19;
     $1529 = $1528 & 31;
     $1530 = (($src$7) + ($1529)|0);
     $1531 = HEAP8[$1530>>0]|0;
     $Pixel$6 = 0;$sA$6$shrunk = -1;$sB$6$shrunk = $1531;$sG$6$shrunk = $1527;$sR$6$shrunk = $1524;
     break;
    }
    case 4:  {
     $1532 = HEAP32[$src$7>>2]|0;
     $1533 = HEAP32[$35>>2]|0;
     $1534 = $1533 & $1532;
     $1535 = HEAP32[$36>>2]|0;
     $1536 = $1535 & 255;
     $1537 = $1534 >>> $1536;
     $1538 = HEAP32[$37>>2]|0;
     $1539 = $1538 & 255;
     $1540 = (125924 + ($1539<<2)|0);
     $1541 = HEAP32[$1540>>2]|0;
     $1542 = (($1541) + ($1537)|0);
     $1543 = HEAP8[$1542>>0]|0;
     $1544 = HEAP32[$38>>2]|0;
     $1545 = $1544 & $1532;
     $1546 = $1535 >>> 8;
     $1547 = $1546 & 255;
     $1548 = $1545 >>> $1547;
     $1549 = $1538 >>> 8;
     $1550 = $1549 & 255;
     $1551 = (125924 + ($1550<<2)|0);
     $1552 = HEAP32[$1551>>2]|0;
     $1553 = (($1552) + ($1548)|0);
     $1554 = HEAP8[$1553>>0]|0;
     $1555 = HEAP32[$39>>2]|0;
     $1556 = $1555 & $1532;
     $1557 = $1535 >>> 16;
     $1558 = $1557 & 255;
     $1559 = $1556 >>> $1558;
     $1560 = $1538 >>> 16;
     $1561 = $1560 & 255;
     $1562 = (125924 + ($1561<<2)|0);
     $1563 = HEAP32[$1562>>2]|0;
     $1564 = (($1563) + ($1559)|0);
     $1565 = HEAP8[$1564>>0]|0;
     $1566 = HEAP32[$14>>2]|0;
     $1567 = $1566 & $1532;
     $1568 = $1535 >>> 24;
     $1569 = $1567 >>> $1568;
     $1570 = $1538 >>> 24;
     $1571 = (125924 + ($1570<<2)|0);
     $1572 = HEAP32[$1571>>2]|0;
     $1573 = (($1572) + ($1569)|0);
     $1574 = HEAP8[$1573>>0]|0;
     $Pixel$6 = $1532;$sA$6$shrunk = $1574;$sB$6$shrunk = $1565;$sG$6$shrunk = $1554;$sR$6$shrunk = $1543;
     break;
    }
    default: {
     $Pixel$6 = 0;$sA$6$shrunk = 0;$sB$6$shrunk = 0;$sG$6$shrunk = 0;$sR$6$shrunk = 0;
    }
    }
    $sR$6 = $sR$6$shrunk&255;
    $sG$6 = $sG$6$shrunk&255;
    $sB$6 = $sB$6$shrunk&255;
    $sA$6 = $sA$6$shrunk&255;
    $1575 = $Pixel$6 & $16;
    $1576 = ($1575|0)==($17|0);
    L105: do {
     if (!($1576)) {
      switch ($30|0) {
      case 1:  {
       $1577 = HEAP32[$31>>2]|0;
       $1578 = $1577 & 255;
       $1579 = $sR$6 >>> $1578;
       $1580 = HEAP32[$32>>2]|0;
       $1581 = $1580 & 255;
       $1582 = $1579 << $1581;
       $1583 = $1577 >>> 8;
       $1584 = $1583 & 255;
       $1585 = $sG$6 >>> $1584;
       $1586 = $1580 >>> 8;
       $1587 = $1586 & 255;
       $1588 = $1585 << $1587;
       $1589 = $1588 | $1582;
       $1590 = $1577 >>> 16;
       $1591 = $1590 & 255;
       $1592 = $sB$6 >>> $1591;
       $1593 = $1580 >>> 16;
       $1594 = $1593 & 255;
       $1595 = $1592 << $1594;
       $1596 = $1589 | $1595;
       $1597 = $1577 >>> 24;
       $1598 = $sA$6 >>> $1597;
       $1599 = $1580 >>> 24;
       $1600 = $1598 << $1599;
       $1601 = $1596 | $1600;
       $1602 = $1601&255;
       HEAP8[$dst$7>>0] = $1602;
       break L105;
       break;
      }
      case 2:  {
       $1603 = HEAP32[$31>>2]|0;
       $1604 = $1603 & 255;
       $1605 = $sR$6 >>> $1604;
       $1606 = HEAP32[$32>>2]|0;
       $1607 = $1606 & 255;
       $1608 = $1605 << $1607;
       $1609 = $1603 >>> 8;
       $1610 = $1609 & 255;
       $1611 = $sG$6 >>> $1610;
       $1612 = $1606 >>> 8;
       $1613 = $1612 & 255;
       $1614 = $1611 << $1613;
       $1615 = $1614 | $1608;
       $1616 = $1603 >>> 16;
       $1617 = $1616 & 255;
       $1618 = $sB$6 >>> $1617;
       $1619 = $1606 >>> 16;
       $1620 = $1619 & 255;
       $1621 = $1618 << $1620;
       $1622 = $1615 | $1621;
       $1623 = $1603 >>> 24;
       $1624 = $sA$6 >>> $1623;
       $1625 = $1606 >>> 24;
       $1626 = $1624 << $1625;
       $1627 = $1622 | $1626;
       $1628 = $1627&65535;
       HEAP16[$dst$7>>1] = $1628;
       break L105;
       break;
      }
      case 3:  {
       $1629 = HEAP8[$32>>0]|0;
       $div = ($1629&255) >>> 3;
       $1630 = $div&255;
       $1631 = (($dst$7) + ($1630)|0);
       HEAP8[$1631>>0] = $sR$6$shrunk;
       $1632 = HEAP8[$33>>0]|0;
       $div1 = ($1632&255) >>> 3;
       $1633 = $div1&255;
       $1634 = (($dst$7) + ($1633)|0);
       HEAP8[$1634>>0] = $sG$6$shrunk;
       $1635 = HEAP8[$34>>0]|0;
       $div2 = ($1635&255) >>> 3;
       $1636 = $div2&255;
       $1637 = (($dst$7) + ($1636)|0);
       HEAP8[$1637>>0] = $sB$6$shrunk;
       break L105;
       break;
      }
      case 4:  {
       $1638 = HEAP32[$31>>2]|0;
       $1639 = $1638 & 255;
       $1640 = $sR$6 >>> $1639;
       $1641 = HEAP32[$32>>2]|0;
       $1642 = $1641 & 255;
       $1643 = $1640 << $1642;
       $1644 = $1638 >>> 8;
       $1645 = $1644 & 255;
       $1646 = $sG$6 >>> $1645;
       $1647 = $1641 >>> 8;
       $1648 = $1647 & 255;
       $1649 = $1646 << $1648;
       $1650 = $1649 | $1643;
       $1651 = $1638 >>> 16;
       $1652 = $1651 & 255;
       $1653 = $sB$6 >>> $1652;
       $1654 = $1641 >>> 16;
       $1655 = $1654 & 255;
       $1656 = $1653 << $1655;
       $1657 = $1650 | $1656;
       $1658 = $1638 >>> 24;
       $1659 = $sA$6 >>> $1658;
       $1660 = $1641 >>> 24;
       $1661 = $1659 << $1660;
       $1662 = $1657 | $1661;
       HEAP32[$dst$7>>2] = $1662;
       break L105;
       break;
      }
      default: {
       break L105;
      }
      }
     }
    } while(0);
    $1663 = (($dst$7) + ($30)|0);
    $1664 = (($src$7) + ($29)|0);
    $dst$8 = $1663;$n$7 = $n$6;$src$8 = $1664;
    label = 88;
    continue;
   }
   else if ((label|0) == 88) {
    label = 0;
    switch ($29|0) {
    case 1:  {
     $1665 = HEAP8[$src$8>>0]|0;
     $1666 = $1665&255;
     $1667 = HEAP32[$35>>2]|0;
     $1668 = $1666 & $1667;
     $1669 = HEAP32[$36>>2]|0;
     $1670 = $1669 & 255;
     $1671 = $1668 >>> $1670;
     $1672 = HEAP32[$37>>2]|0;
     $1673 = $1672 & 255;
     $1674 = (125924 + ($1673<<2)|0);
     $1675 = HEAP32[$1674>>2]|0;
     $1676 = (($1675) + ($1671)|0);
     $1677 = HEAP8[$1676>>0]|0;
     $1678 = HEAP32[$38>>2]|0;
     $1679 = $1678 & $1666;
     $1680 = $1669 >>> 8;
     $1681 = $1680 & 255;
     $1682 = $1679 >>> $1681;
     $1683 = $1672 >>> 8;
     $1684 = $1683 & 255;
     $1685 = (125924 + ($1684<<2)|0);
     $1686 = HEAP32[$1685>>2]|0;
     $1687 = (($1686) + ($1682)|0);
     $1688 = HEAP8[$1687>>0]|0;
     $1689 = HEAP32[$39>>2]|0;
     $1690 = $1689 & $1666;
     $1691 = $1669 >>> 16;
     $1692 = $1691 & 255;
     $1693 = $1690 >>> $1692;
     $1694 = $1672 >>> 16;
     $1695 = $1694 & 255;
     $1696 = (125924 + ($1695<<2)|0);
     $1697 = HEAP32[$1696>>2]|0;
     $1698 = (($1697) + ($1693)|0);
     $1699 = HEAP8[$1698>>0]|0;
     $1700 = HEAP32[$14>>2]|0;
     $1701 = $1700 & $1666;
     $1702 = $1669 >>> 24;
     $1703 = $1701 >>> $1702;
     $1704 = $1672 >>> 24;
     $1705 = (125924 + ($1704<<2)|0);
     $1706 = HEAP32[$1705>>2]|0;
     $1707 = (($1706) + ($1703)|0);
     $1708 = HEAP8[$1707>>0]|0;
     $Pixel$7 = $1666;$sA$7$shrunk = $1708;$sB$7$shrunk = $1699;$sG$7$shrunk = $1688;$sR$7$shrunk = $1677;
     break;
    }
    case 2:  {
     $1709 = HEAP16[$src$8>>1]|0;
     $1710 = $1709&65535;
     $1711 = HEAP32[$35>>2]|0;
     $1712 = $1710 & $1711;
     $1713 = HEAP32[$36>>2]|0;
     $1714 = $1713 & 255;
     $1715 = $1712 >>> $1714;
     $1716 = HEAP32[$37>>2]|0;
     $1717 = $1716 & 255;
     $1718 = (125924 + ($1717<<2)|0);
     $1719 = HEAP32[$1718>>2]|0;
     $1720 = (($1719) + ($1715)|0);
     $1721 = HEAP8[$1720>>0]|0;
     $1722 = HEAP32[$38>>2]|0;
     $1723 = $1722 & $1710;
     $1724 = $1713 >>> 8;
     $1725 = $1724 & 255;
     $1726 = $1723 >>> $1725;
     $1727 = $1716 >>> 8;
     $1728 = $1727 & 255;
     $1729 = (125924 + ($1728<<2)|0);
     $1730 = HEAP32[$1729>>2]|0;
     $1731 = (($1730) + ($1726)|0);
     $1732 = HEAP8[$1731>>0]|0;
     $1733 = HEAP32[$39>>2]|0;
     $1734 = $1733 & $1710;
     $1735 = $1713 >>> 16;
     $1736 = $1735 & 255;
     $1737 = $1734 >>> $1736;
     $1738 = $1716 >>> 16;
     $1739 = $1738 & 255;
     $1740 = (125924 + ($1739<<2)|0);
     $1741 = HEAP32[$1740>>2]|0;
     $1742 = (($1741) + ($1737)|0);
     $1743 = HEAP8[$1742>>0]|0;
     $1744 = HEAP32[$14>>2]|0;
     $1745 = $1744 & $1710;
     $1746 = $1713 >>> 24;
     $1747 = $1745 >>> $1746;
     $1748 = $1716 >>> 24;
     $1749 = (125924 + ($1748<<2)|0);
     $1750 = HEAP32[$1749>>2]|0;
     $1751 = (($1750) + ($1747)|0);
     $1752 = HEAP8[$1751>>0]|0;
     $Pixel$7 = $1710;$sA$7$shrunk = $1752;$sB$7$shrunk = $1743;$sG$7$shrunk = $1732;$sR$7$shrunk = $1721;
     break;
    }
    case 3:  {
     $1753 = HEAP32[$36>>2]|0;
     $div45 = $1753 >>> 3;
     $1754 = $div45 & 31;
     $1755 = (($src$8) + ($1754)|0);
     $1756 = HEAP8[$1755>>0]|0;
     $1757 = $1753 >>> 11;
     $div46 = $1757 & 31;
     $1758 = (($src$8) + ($div46)|0);
     $1759 = HEAP8[$1758>>0]|0;
     $1760 = $1753 >>> 19;
     $1761 = $1760 & 31;
     $1762 = (($src$8) + ($1761)|0);
     $1763 = HEAP8[$1762>>0]|0;
     $Pixel$7 = 0;$sA$7$shrunk = -1;$sB$7$shrunk = $1763;$sG$7$shrunk = $1759;$sR$7$shrunk = $1756;
     break;
    }
    case 4:  {
     $1764 = HEAP32[$src$8>>2]|0;
     $1765 = HEAP32[$35>>2]|0;
     $1766 = $1765 & $1764;
     $1767 = HEAP32[$36>>2]|0;
     $1768 = $1767 & 255;
     $1769 = $1766 >>> $1768;
     $1770 = HEAP32[$37>>2]|0;
     $1771 = $1770 & 255;
     $1772 = (125924 + ($1771<<2)|0);
     $1773 = HEAP32[$1772>>2]|0;
     $1774 = (($1773) + ($1769)|0);
     $1775 = HEAP8[$1774>>0]|0;
     $1776 = HEAP32[$38>>2]|0;
     $1777 = $1776 & $1764;
     $1778 = $1767 >>> 8;
     $1779 = $1778 & 255;
     $1780 = $1777 >>> $1779;
     $1781 = $1770 >>> 8;
     $1782 = $1781 & 255;
     $1783 = (125924 + ($1782<<2)|0);
     $1784 = HEAP32[$1783>>2]|0;
     $1785 = (($1784) + ($1780)|0);
     $1786 = HEAP8[$1785>>0]|0;
     $1787 = HEAP32[$39>>2]|0;
     $1788 = $1787 & $1764;
     $1789 = $1767 >>> 16;
     $1790 = $1789 & 255;
     $1791 = $1788 >>> $1790;
     $1792 = $1770 >>> 16;
     $1793 = $1792 & 255;
     $1794 = (125924 + ($1793<<2)|0);
     $1795 = HEAP32[$1794>>2]|0;
     $1796 = (($1795) + ($1791)|0);
     $1797 = HEAP8[$1796>>0]|0;
     $1798 = HEAP32[$14>>2]|0;
     $1799 = $1798 & $1764;
     $1800 = $1767 >>> 24;
     $1801 = $1799 >>> $1800;
     $1802 = $1770 >>> 24;
     $1803 = (125924 + ($1802<<2)|0);
     $1804 = HEAP32[$1803>>2]|0;
     $1805 = (($1804) + ($1801)|0);
     $1806 = HEAP8[$1805>>0]|0;
     $Pixel$7 = $1764;$sA$7$shrunk = $1806;$sB$7$shrunk = $1797;$sG$7$shrunk = $1786;$sR$7$shrunk = $1775;
     break;
    }
    default: {
     $Pixel$7 = 0;$sA$7$shrunk = 0;$sB$7$shrunk = 0;$sG$7$shrunk = 0;$sR$7$shrunk = 0;
    }
    }
    $sR$7 = $sR$7$shrunk&255;
    $sG$7 = $sG$7$shrunk&255;
    $sB$7 = $sB$7$shrunk&255;
    $sA$7 = $sA$7$shrunk&255;
    $1807 = $Pixel$7 & $16;
    $1808 = ($1807|0)==($17|0);
    L120: do {
     if (!($1808)) {
      switch ($30|0) {
      case 1:  {
       $1809 = HEAP32[$31>>2]|0;
       $1810 = $1809 & 255;
       $1811 = $sR$7 >>> $1810;
       $1812 = HEAP32[$32>>2]|0;
       $1813 = $1812 & 255;
       $1814 = $1811 << $1813;
       $1815 = $1809 >>> 8;
       $1816 = $1815 & 255;
       $1817 = $sG$7 >>> $1816;
       $1818 = $1812 >>> 8;
       $1819 = $1818 & 255;
       $1820 = $1817 << $1819;
       $1821 = $1820 | $1814;
       $1822 = $1809 >>> 16;
       $1823 = $1822 & 255;
       $1824 = $sB$7 >>> $1823;
       $1825 = $1812 >>> 16;
       $1826 = $1825 & 255;
       $1827 = $1824 << $1826;
       $1828 = $1821 | $1827;
       $1829 = $1809 >>> 24;
       $1830 = $sA$7 >>> $1829;
       $1831 = $1812 >>> 24;
       $1832 = $1830 << $1831;
       $1833 = $1828 | $1832;
       $1834 = $1833&255;
       HEAP8[$dst$8>>0] = $1834;
       break L120;
       break;
      }
      case 2:  {
       $1835 = HEAP32[$31>>2]|0;
       $1836 = $1835 & 255;
       $1837 = $sR$7 >>> $1836;
       $1838 = HEAP32[$32>>2]|0;
       $1839 = $1838 & 255;
       $1840 = $1837 << $1839;
       $1841 = $1835 >>> 8;
       $1842 = $1841 & 255;
       $1843 = $sG$7 >>> $1842;
       $1844 = $1838 >>> 8;
       $1845 = $1844 & 255;
       $1846 = $1843 << $1845;
       $1847 = $1846 | $1840;
       $1848 = $1835 >>> 16;
       $1849 = $1848 & 255;
       $1850 = $sB$7 >>> $1849;
       $1851 = $1838 >>> 16;
       $1852 = $1851 & 255;
       $1853 = $1850 << $1852;
       $1854 = $1847 | $1853;
       $1855 = $1835 >>> 24;
       $1856 = $sA$7 >>> $1855;
       $1857 = $1838 >>> 24;
       $1858 = $1856 << $1857;
       $1859 = $1854 | $1858;
       $1860 = $1859&65535;
       HEAP16[$dst$8>>1] = $1860;
       break L120;
       break;
      }
      case 3:  {
       $1861 = HEAP8[$32>>0]|0;
       $div42 = ($1861&255) >>> 3;
       $1862 = $div42&255;
       $1863 = (($dst$8) + ($1862)|0);
       HEAP8[$1863>>0] = $sR$7$shrunk;
       $1864 = HEAP8[$33>>0]|0;
       $div43 = ($1864&255) >>> 3;
       $1865 = $div43&255;
       $1866 = (($dst$8) + ($1865)|0);
       HEAP8[$1866>>0] = $sG$7$shrunk;
       $1867 = HEAP8[$34>>0]|0;
       $div44 = ($1867&255) >>> 3;
       $1868 = $div44&255;
       $1869 = (($dst$8) + ($1868)|0);
       HEAP8[$1869>>0] = $sB$7$shrunk;
       break L120;
       break;
      }
      case 4:  {
       $1870 = HEAP32[$31>>2]|0;
       $1871 = $1870 & 255;
       $1872 = $sR$7 >>> $1871;
       $1873 = HEAP32[$32>>2]|0;
       $1874 = $1873 & 255;
       $1875 = $1872 << $1874;
       $1876 = $1870 >>> 8;
       $1877 = $1876 & 255;
       $1878 = $sG$7 >>> $1877;
       $1879 = $1873 >>> 8;
       $1880 = $1879 & 255;
       $1881 = $1878 << $1880;
       $1882 = $1881 | $1875;
       $1883 = $1870 >>> 16;
       $1884 = $1883 & 255;
       $1885 = $sB$7 >>> $1884;
       $1886 = $1873 >>> 16;
       $1887 = $1886 & 255;
       $1888 = $1885 << $1887;
       $1889 = $1882 | $1888;
       $1890 = $1870 >>> 24;
       $1891 = $sA$7 >>> $1890;
       $1892 = $1873 >>> 24;
       $1893 = $1891 << $1892;
       $1894 = $1889 | $1893;
       HEAP32[$dst$8>>2] = $1894;
       break L120;
       break;
      }
      default: {
       break L120;
      }
      }
     }
    } while(0);
    $1895 = (($dst$8) + ($30)|0);
    $1896 = (($src$8) + ($29)|0);
    $1897 = (($n$7) + -1)|0;
    $1898 = ($n$7|0)>(1);
    if ($1898) {
     $dst$1 = $1895;$n$0 = $1897;$src$1 = $1896;
     label = 4;
     continue;
    } else {
     break;
    }
   }
  }
  $1899 = (($1896) + ($5)|0);
  $1900 = (($1895) + ($7)|0);
  $1901 = ($40|0)==(0);
  if ($1901) {
   label = 102;
   break;
  } else {
   $$in = $40;$dst$048 = $1900;$src$049 = $1899;
  }
 }
 if ((label|0) == 100) {
  // unreachable;
 }
 else if ((label|0) == 102) {
  return;
 }
}
function _SDL_sscanf($text,$fmt,$varargs) {
 $text = $text|0;
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $0 = 0, $ap = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ap = sp;
 HEAP32[$ap>>2] = $varargs;
 $0 = (_vsscanf($text,$fmt,$ap)|0);
 STACKTOP = sp;return ($0|0);
}
function _SDL_Has3DNow() {
 var $$lobit = 0, $$pre$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $altivec$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $altivec$i$i = sp;
 $0 = HEAP32[33822]|0;
 $1 = ($0|0)==(-1);
 if ($1) {
  HEAP32[33822] = 0;
  HEAP32[$altivec$i$i>>2] = 0;
  $2 = HEAP32[$altivec$i$i>>2]|0;
  $3 = ($2|0)==(0);
  $$pre$i = HEAP32[33822]|0;
  if ($3) {
   $6 = $$pre$i;
  } else {
   $4 = $$pre$i | 2;
   HEAP32[33822] = $4;
   $6 = $4;
  }
 } else {
  $6 = $0;
 }
 $5 = $6 >>> 3;
 $$lobit = $5 & 1;
 STACKTOP = sp;return ($$lobit|0);
}
function _SDL_HasSSE() {
 var $$lobit = 0, $$pre$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $altivec$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $altivec$i$i = sp;
 $0 = HEAP32[33822]|0;
 $1 = ($0|0)==(-1);
 if ($1) {
  HEAP32[33822] = 0;
  HEAP32[$altivec$i$i>>2] = 0;
  $2 = HEAP32[$altivec$i$i>>2]|0;
  $3 = ($2|0)==(0);
  $$pre$i = HEAP32[33822]|0;
  if ($3) {
   $6 = $$pre$i;
  } else {
   $4 = $$pre$i | 2;
   HEAP32[33822] = $4;
   $6 = $4;
  }
 } else {
  $6 = $0;
 }
 $5 = $6 >>> 4;
 $$lobit = $5 & 1;
 STACKTOP = sp;return ($$lobit|0);
}
function _SDL_HasSSE2() {
 var $$lobit = 0, $$pre$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $altivec$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $altivec$i$i = sp;
 $0 = HEAP32[33822]|0;
 $1 = ($0|0)==(-1);
 if ($1) {
  HEAP32[33822] = 0;
  HEAP32[$altivec$i$i>>2] = 0;
  $2 = HEAP32[$altivec$i$i>>2]|0;
  $3 = ($2|0)==(0);
  $$pre$i = HEAP32[33822]|0;
  if ($3) {
   $6 = $$pre$i;
  } else {
   $4 = $$pre$i | 2;
   HEAP32[33822] = $4;
   $6 = $4;
  }
 } else {
  $6 = $0;
 }
 $5 = $6 >>> 5;
 $$lobit = $5 & 1;
 STACKTOP = sp;return ($$lobit|0);
}
function _SDL_HasAltiVec() {
 var $$lobit = 0, $$pre$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $altivec$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $altivec$i$i = sp;
 $0 = HEAP32[33822]|0;
 $1 = ($0|0)==(-1);
 if ($1) {
  HEAP32[33822] = 0;
  HEAP32[$altivec$i$i>>2] = 0;
  $2 = HEAP32[$altivec$i$i>>2]|0;
  $3 = ($2|0)==(0);
  $$pre$i = HEAP32[33822]|0;
  if ($3) {
   $6 = $$pre$i;
  } else {
   $4 = $$pre$i | 2;
   HEAP32[33822] = $4;
   $6 = $4;
  }
 } else {
  $6 = $0;
 }
 $5 = $6 >>> 1;
 $$lobit = $5 & 1;
 STACKTOP = sp;return ($$lobit|0);
}
function _SDL_Blit_RGB888_RGB888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0, $posx$010 = 0;
 var $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$46 = $$pre;$51 = $12;$52 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $48 = $51;
  } else {
   $$in = $25;$dst$013 = $52;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    HEAP32[$dst$013>>2] = $40;
    $41 = (($posx$2) + ($11))|0;
    $42 = ((($dst$013)) + 4|0);
    $43 = ($26|0)==(0);
    if ($43) {
     break;
    } else {
     $$in = $26;$dst$013 = $42;$posx$010 = $41;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $48 = $$pre25;
  }
  $44 = HEAP32[$15>>2]|0;
  $45 = (($46) + ($44)|0);
  $47 = (($48) + -1)|0;
  HEAP32[$3>>2] = $47;
  $49 = ($48|0)==(0);
  if ($49) {
   $$lcssa = $45;
   break;
  }
  $50 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$46 = $45;$51 = $47;$52 = $45;$posy$016 = $50;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGB888_RGB888_Blend($info) {
 $info = $info|0;
 var $$$us11 = 0, $$1$us13 = 0, $$3$us12 = 0, $$lcssa = 0, $$lcssa31 = 0, $$pre = 0, $$pre21 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $9 = 0, $cond = 0, $dst$05 = 0, $dst$05$us = 0, $dst$05$us9 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0;
 var $src$04 = 0, $src$04$us = 0, $src$04$us10 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre21 = HEAP32[$6>>2]|0;
 $cond = ($10|0)==(64);
 $77 = $$pre;$80 = $$pre21;$84 = $$pre21;$85 = $$pre;
 while(1) {
  $11 = HEAP32[$7>>2]|0;
  $12 = (($11) + -1)|0;
  $13 = ($11|0)==(0);
  if (!($13)) {
   switch ($10|0) {
   case 16:  {
    $19 = $12;$dst$05$us = $84;$src$04$us = $85;
    while(1) {
     $14 = HEAP32[$src$04$us>>2]|0;
     $15 = $14 & 16777215;
     HEAP32[$dst$05$us>>2] = $15;
     $16 = ((($src$04$us)) + 4|0);
     $17 = ((($dst$05$us)) + 4|0);
     $18 = (($19) + -1)|0;
     $20 = ($19|0)==(0);
     if ($20) {
      break;
     } else {
      $19 = $18;$dst$05$us = $17;$src$04$us = $16;
     }
    }
    break;
   }
   case 32:  {
    $46 = $12;$dst$05$us9 = $84;$src$04$us10 = $85;
    while(1) {
     $21 = HEAP32[$src$04$us10>>2]|0;
     $22 = $21 >>> 16;
     $23 = $22 & 255;
     $24 = $21 >>> 8;
     $25 = $24 & 255;
     $26 = $21 & 255;
     $27 = HEAP32[$dst$05$us9>>2]|0;
     $28 = $27 >>> 16;
     $29 = $28 & 255;
     $30 = $27 >>> 8;
     $31 = $30 & 255;
     $32 = $27 & 255;
     $33 = (($29) + ($23))|0;
     $34 = ($33>>>0)>(255);
     $$$us11 = $34 ? 255 : $33;
     $35 = (($31) + ($25))|0;
     $36 = ($35>>>0)>(255);
     $$3$us12 = $36 ? 255 : $35;
     $37 = (($32) + ($26))|0;
     $38 = ($37>>>0)>(255);
     $$1$us13 = $38 ? 255 : $37;
     $39 = $$$us11 << 16;
     $40 = $$3$us12 << 8;
     $41 = $40 | $39;
     $42 = $41 | $$1$us13;
     HEAP32[$dst$05$us9>>2] = $42;
     $43 = ((($src$04$us10)) + 4|0);
     $44 = ((($dst$05$us9)) + 4|0);
     $45 = (($46) + -1)|0;
     $47 = ($46|0)==(0);
     if ($47) {
      break;
     } else {
      $46 = $45;$dst$05$us9 = $44;$src$04$us10 = $43;
     }
    }
    break;
   }
   default: {
    $73 = $12;$dst$05 = $84;$src$04 = $85;
    while(1) {
     $48 = HEAP32[$src$04>>2]|0;
     $49 = HEAP32[$dst$05>>2]|0;
     $50 = $49 >>> 16;
     $51 = $50 & 255;
     $52 = $49 >>> 8;
     $53 = $52 & 255;
     $54 = $49 & 255;
     if ($cond) {
      $55 = $48 & 255;
      $56 = $48 >>> 8;
      $57 = $56 & 255;
      $58 = $48 >>> 16;
      $59 = $58 & 255;
      $60 = Math_imul($51, $59)|0;
      $61 = (($60>>>0) / 255)&-1;
      $62 = Math_imul($53, $57)|0;
      $63 = (($62>>>0) / 255)&-1;
      $64 = Math_imul($54, $55)|0;
      $65 = (($64>>>0) / 255)&-1;
      $dstB$0 = $65;$dstG$0 = $63;$dstR$0 = $61;
     } else {
      $dstB$0 = $54;$dstG$0 = $53;$dstR$0 = $51;
     }
     $66 = $dstR$0 << 16;
     $67 = $dstG$0 << 8;
     $68 = $67 | $66;
     $69 = $68 | $dstB$0;
     HEAP32[$dst$05>>2] = $69;
     $70 = ((($src$04)) + 4|0);
     $71 = ((($dst$05)) + 4|0);
     $72 = (($73) + -1)|0;
     $74 = ($73|0)==(0);
     if ($74) {
      break;
     } else {
      $73 = $72;$dst$05 = $71;$src$04 = $70;
     }
    }
   }
   }
  }
  $75 = HEAP32[$8>>2]|0;
  $76 = (($77) + ($75)|0);
  $78 = HEAP32[$9>>2]|0;
  $79 = (($80) + ($78)|0);
  $81 = HEAP32[$2>>2]|0;
  $82 = (($81) + -1)|0;
  HEAP32[$2>>2] = $82;
  $83 = ($81|0)==(0);
  if ($83) {
   $$lcssa = $76;$$lcssa31 = $79;
   break;
  } else {
   $77 = $76;$80 = $79;$84 = $79;$85 = $76;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa31;
 return;
}
function _SDL_Blit_RGB888_RGB888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$lcssa = 0, $$pre = 0, $$pre30 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dst$016$us = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $posx$014 = 0, $posx$014$us = 0, $posx$2 = 0, $posx$2$us = 0;
 var $posy$020 = 0, $posy$1$lcssa = 0, $src$015 = 0, $src$015$us = 0, $src$1 = 0, $src$1$us = 0, $srcx$013 = 0, $srcx$013$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 112;
 $19 = ((($info)) + 12|0);
 $20 = ($18|0)==(16);
 $$pre = HEAP32[$16>>2]|0;
 $104 = $$pre;$29 = $12;$99 = $$pre;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $21 = ($posy$020|0)>(65535);
  if ($21) {
   $22 = (($posy$020) + -65536)|0;
   $23 = $22 >>> 16;
   $24 = $23 << 16;
   $25 = (($srcy$019) + 1)|0;
   $26 = (($22) - ($24))|0;
   $27 = (($25) + ($23))|0;
   $posy$1$lcssa = $26;$srcy$1$lcssa = $27;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $28 = (($29) + -1)|0;
  $30 = ($29|0)==(0);
  L9: do {
   if (!($30)) {
    if ($20) {
     $49 = $28;$dst$016$us = $104;$posx$014$us = 65536;$src$015$us = 0;$srcx$013$us = -1;
     while(1) {
      $31 = ($posx$014$us|0)>(65535);
      if ($31) {
       $32 = (($posx$014$us) + -65536)|0;
       $33 = $32 >>> 16;
       $34 = $33 << 16;
       $35 = (($srcx$013$us) + 1)|0;
       $36 = (($32) - ($34))|0;
       $37 = (($35) + ($33))|0;
       $38 = HEAP32[$info>>2]|0;
       $39 = HEAP32[$19>>2]|0;
       $40 = Math_imul($39, $srcy$1$lcssa)|0;
       $41 = (($38) + ($40)|0);
       $42 = $37 << 2;
       $43 = (($41) + ($42)|0);
       $posx$2$us = $36;$src$1$us = $43;$srcx$2$us = $37;
      } else {
       $posx$2$us = $posx$014$us;$src$1$us = $src$015$us;$srcx$2$us = $srcx$013$us;
      }
      $44 = HEAP32[$src$1$us>>2]|0;
      $45 = $44 & 16777215;
      HEAP32[$dst$016$us>>2] = $45;
      $46 = (($posx$2$us) + ($13))|0;
      $47 = ((($dst$016$us)) + 4|0);
      $48 = (($49) + -1)|0;
      $50 = ($49|0)==(0);
      if ($50) {
       break L9;
      } else {
       $49 = $48;$dst$016$us = $47;$posx$014$us = $46;$src$015$us = $src$1$us;$srcx$013$us = $srcx$2$us;
      }
     }
    } else {
     $95 = $28;$dst$016 = $104;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
    }
    while(1) {
     $51 = ($posx$014|0)>(65535);
     if ($51) {
      $52 = (($posx$014) + -65536)|0;
      $53 = $52 >>> 16;
      $54 = $53 << 16;
      $55 = (($srcx$013) + 1)|0;
      $56 = (($52) - ($54))|0;
      $57 = (($55) + ($53))|0;
      $58 = HEAP32[$info>>2]|0;
      $59 = HEAP32[$19>>2]|0;
      $60 = Math_imul($59, $srcy$1$lcssa)|0;
      $61 = (($58) + ($60)|0);
      $62 = $57 << 2;
      $63 = (($61) + ($62)|0);
      $posx$2 = $56;$src$1 = $63;$srcx$2 = $57;
     } else {
      $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
     }
     $64 = HEAP32[$src$1>>2]|0;
     $65 = $64 >>> 16;
     $66 = $65 & 255;
     $67 = $64 >>> 8;
     $68 = $67 & 255;
     $69 = $64 & 255;
     $70 = HEAP32[$dst$016>>2]|0;
     $71 = $70 >>> 16;
     $72 = $71 & 255;
     $73 = $70 >>> 8;
     $74 = $73 & 255;
     $75 = $70 & 255;
     switch ($18|0) {
     case 64:  {
      $82 = Math_imul($72, $66)|0;
      $83 = (($82>>>0) / 255)&-1;
      $84 = Math_imul($74, $68)|0;
      $85 = (($84>>>0) / 255)&-1;
      $86 = Math_imul($75, $69)|0;
      $87 = (($86>>>0) / 255)&-1;
      $dstB$0 = $87;$dstG$0 = $85;$dstR$0 = $83;
      break;
     }
     case 32:  {
      $76 = (($72) + ($66))|0;
      $77 = ($76>>>0)>(255);
      $$ = $77 ? 255 : $76;
      $78 = (($74) + ($68))|0;
      $79 = ($78>>>0)>(255);
      $$3 = $79 ? 255 : $78;
      $80 = (($75) + ($69))|0;
      $81 = ($80>>>0)>(255);
      $$1 = $81 ? 255 : $80;
      $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
      break;
     }
     default: {
      $dstB$0 = $75;$dstG$0 = $74;$dstR$0 = $72;
     }
     }
     $88 = $dstR$0 << 16;
     $89 = $dstG$0 << 8;
     $90 = $89 | $88;
     $91 = $90 | $dstB$0;
     HEAP32[$dst$016>>2] = $91;
     $92 = (($posx$2) + ($13))|0;
     $93 = ((($dst$016)) + 4|0);
     $94 = (($95) + -1)|0;
     $96 = ($95|0)==(0);
     if ($96) {
      break L9;
     } else {
      $95 = $94;$dst$016 = $93;$posx$014 = $92;$src$015 = $src$1;$srcx$013 = $srcx$2;
     }
    }
   }
  } while(0);
  $97 = HEAP32[$17>>2]|0;
  $98 = (($99) + ($97)|0);
  $100 = HEAP32[$5>>2]|0;
  $101 = (($100) + -1)|0;
  HEAP32[$5>>2] = $101;
  $102 = ($100|0)==(0);
  if ($102) {
   $$lcssa = $98;
   break;
  }
  $103 = (($posy$1$lcssa) + ($7))|0;
  $$pre30 = HEAP32[$11>>2]|0;
  $104 = $98;$29 = $$pre30;$99 = $98;$posy$020 = $103;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGB888_RGB888_Modulate($info) {
 $info = $info|0;
 var $$in = 0, $$in13 = 0, $$lcssa = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$pre16 = 0, $$pre17 = 0, $$pre18 = 0, $$pre19 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $8 = 0, $9 = 0, $dst$02 = 0, $dst$02$us$us = 0, $src$01 = 0, $src$01$us$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = ((($info)) + 20|0);
 $14 = ((($info)) + 24|0);
 $15 = ((($info)) + 12|0);
 $16 = ((($info)) + 32|0);
 $17 = $1 & 1;
 $18 = ($17|0)==(0);
 $$pre17 = HEAP32[$info>>2]|0;
 $$pre18 = HEAP32[$13>>2]|0;
 if ($18) {
  $23 = $$pre17;$26 = $$pre18;$67 = $11;$68 = $$pre18;$69 = $$pre17;
  while(1) {
   $19 = HEAP32[$14>>2]|0;
   $20 = ($19|0)==(0);
   if ($20) {
    $28 = $67;
   } else {
    $$in13 = $19;$dst$02$us$us = $68;$src$01$us$us = $69;
    while(1) {
     $30 = (($$in13) + -1)|0;
     $31 = HEAP32[$src$01$us$us>>2]|0;
     $32 = $31 & 16777215;
     HEAP32[$dst$02$us$us>>2] = $32;
     $33 = ((($src$01$us$us)) + 4|0);
     $34 = ((($dst$02$us$us)) + 4|0);
     $35 = ($30|0)==(0);
     if ($35) {
      break;
     } else {
      $$in13 = $30;$dst$02$us$us = $34;$src$01$us$us = $33;
     }
    }
    $$pre19 = HEAP32[$9>>2]|0;
    $28 = $$pre19;
   }
   $21 = HEAP32[$15>>2]|0;
   $22 = (($23) + ($21)|0);
   $24 = HEAP32[$16>>2]|0;
   $25 = (($26) + ($24)|0);
   $27 = (($28) + -1)|0;
   HEAP32[$9>>2] = $27;
   $29 = ($28|0)==(0);
   if ($29) {
    $$lcssa = $22;$$lcssa30 = $25;
    break;
   } else {
    $23 = $22;$26 = $25;$67 = $27;$68 = $25;$69 = $22;
   }
  }
  HEAP32[$info>>2] = $$lcssa;
  HEAP32[$13>>2] = $$lcssa30;
  return;
 } else {
  $60 = $$pre17;$63 = $$pre18;$70 = $11;$71 = $$pre18;$72 = $$pre17;
  while(1) {
   $36 = HEAP32[$14>>2]|0;
   $37 = ($36|0)==(0);
   if ($37) {
    $65 = $70;
   } else {
    $$in = $36;$dst$02 = $71;$src$01 = $72;
    while(1) {
     $38 = (($$in) + -1)|0;
     $39 = HEAP32[$src$01>>2]|0;
     $40 = $39 >>> 16;
     $41 = $40 & 255;
     $42 = $39 >>> 8;
     $43 = $42 & 255;
     $44 = $39 & 255;
     $45 = Math_imul($41, $4)|0;
     $46 = (($45>>>0) / 255)&-1;
     $47 = Math_imul($43, $6)|0;
     $48 = (($47>>>0) / 255)&-1;
     $49 = Math_imul($44, $8)|0;
     $50 = (($49>>>0) / 255)&-1;
     $51 = $46 << 16;
     $52 = $48 << 8;
     $53 = $52 | $51;
     $54 = $53 | $50;
     HEAP32[$dst$02>>2] = $54;
     $55 = ((($src$01)) + 4|0);
     $56 = ((($dst$02)) + 4|0);
     $57 = ($38|0)==(0);
     if ($57) {
      break;
     } else {
      $$in = $38;$dst$02 = $56;$src$01 = $55;
     }
    }
    $$pre16 = HEAP32[$9>>2]|0;
    $65 = $$pre16;
   }
   $58 = HEAP32[$15>>2]|0;
   $59 = (($60) + ($58)|0);
   $61 = HEAP32[$16>>2]|0;
   $62 = (($63) + ($61)|0);
   $64 = (($65) + -1)|0;
   HEAP32[$9>>2] = $64;
   $66 = ($65|0)==(0);
   if ($66) {
    $$lcssa31 = $59;$$lcssa32 = $62;
    break;
   } else {
    $60 = $59;$63 = $62;$70 = $64;$71 = $62;$72 = $59;
   }
  }
  HEAP32[$info>>2] = $$lcssa31;
  HEAP32[$13>>2] = $$lcssa32;
  return;
 }
}
function _SDL_Blit_RGB888_RGB888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $dst$013 = 0, $dst$013$us = 0, $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$017 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0, $src$1$us = 0, $srcx$010 = 0, $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$016 = 0;
 var $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = ((($info)) + 20|0);
 $24 = ((($info)) + 32|0);
 $25 = $1 & 1;
 $26 = ($25|0)==(0);
 $27 = ((($info)) + 12|0);
 $$pre = HEAP32[$23>>2]|0;
 $36 = $19;$94 = $$pre;$99 = $$pre;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $28 = ($posy$017|0)>(65535);
  if ($28) {
   $29 = (($posy$017) + -65536)|0;
   $30 = $29 >>> 16;
   $31 = $30 << 16;
   $32 = (($srcy$016) + 1)|0;
   $33 = (($29) - ($31))|0;
   $34 = (($32) + ($30))|0;
   $posy$1$lcssa = $33;$srcy$1$lcssa = $34;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $35 = (($36) + -1)|0;
  $37 = ($36|0)==(0);
  if (!($37)) {
   if ($26) {
    $56 = $35;$dst$013$us = $99;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $38 = ($posx$011$us|0)>(65535);
     if ($38) {
      $39 = (($posx$011$us) + -65536)|0;
      $40 = $39 >>> 16;
      $41 = $40 << 16;
      $42 = (($srcx$010$us) + 1)|0;
      $43 = (($39) - ($41))|0;
      $44 = (($42) + ($40))|0;
      $45 = HEAP32[$info>>2]|0;
      $46 = HEAP32[$27>>2]|0;
      $47 = Math_imul($46, $srcy$1$lcssa)|0;
      $48 = (($45) + ($47)|0);
      $49 = $44 << 2;
      $50 = (($48) + ($49)|0);
      $posx$2$us = $43;$src$1$us = $50;$srcx$2$us = $44;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $51 = HEAP32[$src$1$us>>2]|0;
     $52 = $51 & 16777215;
     HEAP32[$dst$013$us>>2] = $52;
     $53 = (($posx$2$us) + ($20))|0;
     $54 = ((($dst$013$us)) + 4|0);
     $55 = (($56) + -1)|0;
     $57 = ($56|0)==(0);
     if ($57) {
      break;
     } else {
      $56 = $55;$dst$013$us = $54;$posx$011$us = $53;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $90 = $35;$dst$013 = $99;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $58 = ($posx$011|0)>(65535);
     if ($58) {
      $59 = (($posx$011) + -65536)|0;
      $60 = $59 >>> 16;
      $61 = $60 << 16;
      $62 = (($srcx$010) + 1)|0;
      $63 = (($59) - ($61))|0;
      $64 = (($62) + ($60))|0;
      $65 = HEAP32[$info>>2]|0;
      $66 = HEAP32[$27>>2]|0;
      $67 = Math_imul($66, $srcy$1$lcssa)|0;
      $68 = (($65) + ($67)|0);
      $69 = $64 << 2;
      $70 = (($68) + ($69)|0);
      $posx$2 = $63;$src$1 = $70;$srcx$2 = $64;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $71 = HEAP32[$src$1>>2]|0;
     $72 = $71 >>> 16;
     $73 = $72 & 255;
     $74 = $71 >>> 8;
     $75 = $74 & 255;
     $76 = $71 & 255;
     $77 = Math_imul($73, $4)|0;
     $78 = (($77>>>0) / 255)&-1;
     $79 = Math_imul($75, $6)|0;
     $80 = (($79>>>0) / 255)&-1;
     $81 = Math_imul($76, $8)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = $78 << 16;
     $84 = $80 << 8;
     $85 = $82 | $84;
     $86 = $85 | $83;
     HEAP32[$dst$013>>2] = $86;
     $87 = (($posx$2) + ($20))|0;
     $88 = ((($dst$013)) + 4|0);
     $89 = (($90) + -1)|0;
     $91 = ($90|0)==(0);
     if ($91) {
      break;
     } else {
      $90 = $89;$dst$013 = $88;$posx$011 = $87;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $92 = HEAP32[$24>>2]|0;
  $93 = (($94) + ($92)|0);
  $95 = HEAP32[$12>>2]|0;
  $96 = (($95) + -1)|0;
  HEAP32[$12>>2] = $96;
  $97 = ($95|0)==(0);
  if ($97) {
   $$lcssa = $93;
   break;
  }
  $98 = (($posy$1$lcssa) + ($14))|0;
  $$pre27 = HEAP32[$18>>2]|0;
  $36 = $$pre27;$94 = $93;$99 = $93;$posy$017 = $98;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$23>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGB888_RGB888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$$us = 0, $$1 = 0, $$1$us = 0, $$3 = 0, $$3$us = 0, $$4 = 0, $$lcssa = 0, $$lcssa18 = 0, $$pre = 0, $$pre11 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$06 = 0, $dst$06$us = 0, $dstB$0 = 0, $dstB$0$us = 0, $dstG$0 = 0, $dstG$0$us = 0;
 var $dstR$0 = 0, $dstR$0$us = 0, $or$cond = 0, $src$05 = 0, $src$05$us = 0, $srcB$1 = 0, $srcB$1$us = 0, $srcG$1 = 0, $srcG$1$us = 0, $srcR$1 = 0, $srcR$1$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = $3 >>> 24;
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $$4 = $21 ? 255 : $13;
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = ($$4|0)!=(255);
 $or$cond = $23 & $24;
 $25 = $1 & 112;
 $26 = $$4 ^ 255;
 $$pre = HEAP32[$info>>2]|0;
 $$pre11 = HEAP32[$14>>2]|0;
 $134 = $$pre;$137 = $$pre11;$141 = $$pre11;$142 = $$pre;
 while(1) {
  $27 = HEAP32[$15>>2]|0;
  $28 = (($27) + -1)|0;
  $29 = ($27|0)==(0);
  L6: do {
   if (!($29)) {
    if ($19) {
     $76 = $28;$dst$06$us = $141;$src$05$us = $142;
     while(1) {
      $30 = HEAP32[$src$05$us>>2]|0;
      $31 = $30 >>> 16;
      $32 = $31 & 255;
      $33 = $30 >>> 8;
      $34 = $33 & 255;
      $35 = $30 & 255;
      $36 = HEAP32[$dst$06$us>>2]|0;
      $37 = $36 >>> 16;
      $38 = $37 & 255;
      $39 = $36 >>> 8;
      $40 = $39 & 255;
      $41 = $36 & 255;
      if ($or$cond) {
       $42 = Math_imul($32, $$4)|0;
       $43 = (($42>>>0) / 255)&-1;
       $44 = Math_imul($34, $$4)|0;
       $45 = (($44>>>0) / 255)&-1;
       $46 = Math_imul($35, $$4)|0;
       $47 = (($46>>>0) / 255)&-1;
       $srcB$1$us = $47;$srcG$1$us = $45;$srcR$1$us = $43;
      } else {
       $srcB$1$us = $35;$srcG$1$us = $34;$srcR$1$us = $32;
      }
      switch ($25|0) {
      case 16:  {
       $60 = Math_imul($38, $26)|0;
       $61 = (($60>>>0) / 255)&-1;
       $62 = (($srcR$1$us) + ($61))|0;
       $63 = Math_imul($40, $26)|0;
       $64 = (($63>>>0) / 255)&-1;
       $65 = (($srcG$1$us) + ($64))|0;
       $66 = Math_imul($41, $26)|0;
       $67 = (($66>>>0) / 255)&-1;
       $68 = (($srcB$1$us) + ($67))|0;
       $dstB$0$us = $68;$dstG$0$us = $65;$dstR$0$us = $62;
       break;
      }
      case 32:  {
       $54 = (($srcR$1$us) + ($38))|0;
       $55 = ($54>>>0)>(255);
       $$$us = $55 ? 255 : $54;
       $56 = (($srcG$1$us) + ($40))|0;
       $57 = ($56>>>0)>(255);
       $$3$us = $57 ? 255 : $56;
       $58 = (($srcB$1$us) + ($41))|0;
       $59 = ($58>>>0)>(255);
       $$1$us = $59 ? 255 : $58;
       $dstB$0$us = $$1$us;$dstG$0$us = $$3$us;$dstR$0$us = $$$us;
       break;
      }
      case 64:  {
       $48 = Math_imul($srcR$1$us, $38)|0;
       $49 = (($48>>>0) / 255)&-1;
       $50 = Math_imul($srcG$1$us, $40)|0;
       $51 = (($50>>>0) / 255)&-1;
       $52 = Math_imul($srcB$1$us, $41)|0;
       $53 = (($52>>>0) / 255)&-1;
       $dstB$0$us = $53;$dstG$0$us = $51;$dstR$0$us = $49;
       break;
      }
      default: {
       $dstB$0$us = $41;$dstG$0$us = $40;$dstR$0$us = $38;
      }
      }
      $69 = $dstR$0$us << 16;
      $70 = $dstG$0$us << 8;
      $71 = $70 | $69;
      $72 = $71 | $dstB$0$us;
      HEAP32[$dst$06$us>>2] = $72;
      $73 = ((($src$05$us)) + 4|0);
      $74 = ((($dst$06$us)) + 4|0);
      $75 = (($76) + -1)|0;
      $77 = ($76|0)==(0);
      if ($77) {
       break L6;
      } else {
       $76 = $75;$dst$06$us = $74;$src$05$us = $73;
      }
     }
    } else {
     $130 = $28;$dst$06 = $141;$src$05 = $142;
     while(1) {
      $78 = HEAP32[$src$05>>2]|0;
      $79 = $78 >>> 16;
      $80 = $79 & 255;
      $81 = $78 >>> 8;
      $82 = $81 & 255;
      $83 = $78 & 255;
      $84 = HEAP32[$dst$06>>2]|0;
      $85 = $84 >>> 16;
      $86 = $85 & 255;
      $87 = $84 >>> 8;
      $88 = $87 & 255;
      $89 = $84 & 255;
      $90 = Math_imul($80, $4)|0;
      $91 = (($90>>>0) / 255)&-1;
      $92 = Math_imul($82, $6)|0;
      $93 = (($92>>>0) / 255)&-1;
      $94 = Math_imul($83, $8)|0;
      $95 = (($94>>>0) / 255)&-1;
      if ($or$cond) {
       $96 = Math_imul($91, $$4)|0;
       $97 = (($96>>>0) / 255)&-1;
       $98 = Math_imul($93, $$4)|0;
       $99 = (($98>>>0) / 255)&-1;
       $100 = Math_imul($95, $$4)|0;
       $101 = (($100>>>0) / 255)&-1;
       $srcB$1 = $101;$srcG$1 = $99;$srcR$1 = $97;
      } else {
       $srcB$1 = $95;$srcG$1 = $93;$srcR$1 = $91;
      }
      switch ($25|0) {
      case 16:  {
       $102 = Math_imul($86, $26)|0;
       $103 = (($102>>>0) / 255)&-1;
       $104 = (($srcR$1) + ($103))|0;
       $105 = Math_imul($88, $26)|0;
       $106 = (($105>>>0) / 255)&-1;
       $107 = (($srcG$1) + ($106))|0;
       $108 = Math_imul($89, $26)|0;
       $109 = (($108>>>0) / 255)&-1;
       $110 = (($srcB$1) + ($109))|0;
       $dstB$0 = $110;$dstG$0 = $107;$dstR$0 = $104;
       break;
      }
      case 32:  {
       $111 = (($srcR$1) + ($86))|0;
       $112 = ($111>>>0)>(255);
       $$ = $112 ? 255 : $111;
       $113 = (($srcG$1) + ($88))|0;
       $114 = ($113>>>0)>(255);
       $$3 = $114 ? 255 : $113;
       $115 = (($srcB$1) + ($89))|0;
       $116 = ($115>>>0)>(255);
       $$1 = $116 ? 255 : $115;
       $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
       break;
      }
      case 64:  {
       $117 = Math_imul($srcR$1, $86)|0;
       $118 = (($117>>>0) / 255)&-1;
       $119 = Math_imul($srcG$1, $88)|0;
       $120 = (($119>>>0) / 255)&-1;
       $121 = Math_imul($srcB$1, $89)|0;
       $122 = (($121>>>0) / 255)&-1;
       $dstB$0 = $122;$dstG$0 = $120;$dstR$0 = $118;
       break;
      }
      default: {
       $dstB$0 = $89;$dstG$0 = $88;$dstR$0 = $86;
      }
      }
      $123 = $dstR$0 << 16;
      $124 = $dstG$0 << 8;
      $125 = $124 | $123;
      $126 = $125 | $dstB$0;
      HEAP32[$dst$06>>2] = $126;
      $127 = ((($src$05)) + 4|0);
      $128 = ((($dst$06)) + 4|0);
      $129 = (($130) + -1)|0;
      $131 = ($130|0)==(0);
      if ($131) {
       break L6;
      } else {
       $130 = $129;$dst$06 = $128;$src$05 = $127;
      }
     }
    }
   }
  } while(0);
  $132 = HEAP32[$16>>2]|0;
  $133 = (($134) + ($132)|0);
  $135 = HEAP32[$17>>2]|0;
  $136 = (($137) + ($135)|0);
  $138 = HEAP32[$9>>2]|0;
  $139 = (($138) + -1)|0;
  HEAP32[$9>>2] = $139;
  $140 = ($138|0)==(0);
  if ($140) {
   $$lcssa = $133;$$lcssa18 = $136;
   break;
  } else {
   $134 = $133;$137 = $136;$141 = $136;$142 = $133;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa18;
 return;
}
function _SDL_Blit_RGB888_RGB888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$4 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre28 = 0, $$pre29 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$017 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$015 = 0, $posx$2 = 0, $posy$021 = 0, $posy$1$lcssa = 0, $src$016 = 0, $src$1 = 0;
 var $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$014 = 0, $srcx$2 = 0, $srcy$020 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = $3 >>> 24;
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $$4 = $29 ? 255 : $23;
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = ($$4|0)!=(255);
 $or$cond = $31 & $32;
 $33 = $1 & 112;
 $34 = $$4 ^ 255;
 $35 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $113 = $$pre;$118 = $21;$119 = $$pre;$44 = $19;$posy$021 = 0;$srcy$020 = 0;
 while(1) {
  $36 = ($posy$021|0)>(65535);
  if ($36) {
   $37 = (($posy$021) + -65536)|0;
   $38 = $37 >>> 16;
   $39 = $38 << 16;
   $40 = (($srcy$020) + 1)|0;
   $41 = (($37) - ($39))|0;
   $42 = (($40) + ($38))|0;
   $posy$1$lcssa = $41;$srcy$1$lcssa = $42;
  } else {
   $posy$1$lcssa = $posy$021;$srcy$1$lcssa = $srcy$020;
  }
  $43 = ($44|0)==(0);
  if ($43) {
   $115 = $118;
  } else {
   $$in = $44;$dst$017 = $119;$posx$015 = 65536;$src$016 = 0;$srcx$014 = -1;
   while(1) {
    $45 = (($$in) + -1)|0;
    $46 = ($posx$015|0)>(65535);
    if ($46) {
     $47 = (($posx$015) + -65536)|0;
     $48 = $47 >>> 16;
     $49 = $48 << 16;
     $50 = (($srcx$014) + 1)|0;
     $51 = (($47) - ($49))|0;
     $52 = (($50) + ($48))|0;
     $53 = HEAP32[$info>>2]|0;
     $54 = HEAP32[$35>>2]|0;
     $55 = Math_imul($54, $srcy$1$lcssa)|0;
     $56 = (($53) + ($55)|0);
     $57 = $52 << 2;
     $58 = (($56) + ($57)|0);
     $posx$2 = $51;$src$1 = $58;$srcx$2 = $52;
    } else {
     $posx$2 = $posx$015;$src$1 = $src$016;$srcx$2 = $srcx$014;
    }
    $59 = HEAP32[$src$1>>2]|0;
    $60 = $59 >>> 16;
    $61 = $60 & 255;
    $62 = $59 >>> 8;
    $63 = $62 & 255;
    $64 = $59 & 255;
    $65 = HEAP32[$dst$017>>2]|0;
    $66 = $65 >>> 16;
    $67 = $66 & 255;
    $68 = $65 >>> 8;
    $69 = $68 & 255;
    $70 = $65 & 255;
    if ($27) {
     $srcB$0 = $64;$srcG$0 = $63;$srcR$0 = $61;
    } else {
     $71 = Math_imul($61, $4)|0;
     $72 = (($71>>>0) / 255)&-1;
     $73 = Math_imul($63, $6)|0;
     $74 = (($73>>>0) / 255)&-1;
     $75 = Math_imul($64, $8)|0;
     $76 = (($75>>>0) / 255)&-1;
     $srcB$0 = $76;$srcG$0 = $74;$srcR$0 = $72;
    }
    if ($or$cond) {
     $77 = Math_imul($srcR$0, $$4)|0;
     $78 = (($77>>>0) / 255)&-1;
     $79 = Math_imul($srcG$0, $$4)|0;
     $80 = (($79>>>0) / 255)&-1;
     $81 = Math_imul($srcB$0, $$4)|0;
     $82 = (($81>>>0) / 255)&-1;
     $srcB$1 = $82;$srcG$1 = $80;$srcR$1 = $78;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($33|0) {
    case 16:  {
     $83 = Math_imul($67, $34)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = (($srcR$1) + ($84))|0;
     $86 = Math_imul($69, $34)|0;
     $87 = (($86>>>0) / 255)&-1;
     $88 = (($srcG$1) + ($87))|0;
     $89 = Math_imul($70, $34)|0;
     $90 = (($89>>>0) / 255)&-1;
     $91 = (($srcB$1) + ($90))|0;
     $dstB$0 = $91;$dstG$0 = $88;$dstR$0 = $85;
     break;
    }
    case 32:  {
     $92 = (($srcR$1) + ($67))|0;
     $93 = ($92>>>0)>(255);
     $$ = $93 ? 255 : $92;
     $94 = (($srcG$1) + ($69))|0;
     $95 = ($94>>>0)>(255);
     $$3 = $95 ? 255 : $94;
     $96 = (($srcB$1) + ($70))|0;
     $97 = ($96>>>0)>(255);
     $$1 = $97 ? 255 : $96;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $98 = Math_imul($srcR$1, $67)|0;
     $99 = (($98>>>0) / 255)&-1;
     $100 = Math_imul($srcG$1, $69)|0;
     $101 = (($100>>>0) / 255)&-1;
     $102 = Math_imul($srcB$1, $70)|0;
     $103 = (($102>>>0) / 255)&-1;
     $dstB$0 = $103;$dstG$0 = $101;$dstR$0 = $99;
     break;
    }
    default: {
     $dstB$0 = $70;$dstG$0 = $69;$dstR$0 = $67;
    }
    }
    $104 = $dstR$0 << 16;
    $105 = $dstG$0 << 8;
    $106 = $105 | $104;
    $107 = $106 | $dstB$0;
    HEAP32[$dst$017>>2] = $107;
    $108 = (($posx$2) + ($20))|0;
    $109 = ((($dst$017)) + 4|0);
    $110 = ($45|0)==(0);
    if ($110) {
     break;
    } else {
     $$in = $45;$dst$017 = $109;$posx$015 = $108;$src$016 = $src$1;$srcx$014 = $srcx$2;
    }
   }
   $$pre29 = HEAP32[$12>>2]|0;
   $115 = $$pre29;
  }
  $111 = HEAP32[$25>>2]|0;
  $112 = (($113) + ($111)|0);
  $114 = (($115) + -1)|0;
  HEAP32[$12>>2] = $114;
  $116 = ($115|0)==(0);
  if ($116) {
   $$lcssa = $112;
   break;
  }
  $117 = (($posy$1$lcssa) + ($14))|0;
  $$pre28 = HEAP32[$18>>2]|0;
  $113 = $112;$118 = $114;$119 = $112;$44 = $$pre28;$posy$021 = $117;$srcy$020 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGB888_BGR888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0, $posx$010 = 0, $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$53 = $$pre;$58 = $12;$59 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $55 = $58;
  } else {
   $$in = $25;$dst$013 = $59;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    $41 = $40 >>> 16;
    $42 = $41 & 255;
    $43 = $40 & 65280;
    $44 = $40 << 16;
    $45 = $44 & 16711680;
    $46 = $45 | $43;
    $47 = $46 | $42;
    HEAP32[$dst$013>>2] = $47;
    $48 = (($posx$2) + ($11))|0;
    $49 = ((($dst$013)) + 4|0);
    $50 = ($26|0)==(0);
    if ($50) {
     break;
    } else {
     $$in = $26;$dst$013 = $49;$posx$010 = $48;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $55 = $$pre25;
  }
  $51 = HEAP32[$15>>2]|0;
  $52 = (($53) + ($51)|0);
  $54 = (($55) + -1)|0;
  HEAP32[$3>>2] = $54;
  $56 = ($55|0)==(0);
  if ($56) {
   $$lcssa = $52;
   break;
  }
  $57 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$53 = $52;$58 = $54;$59 = $52;$posy$016 = $57;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGB888_BGR888_Blend($info) {
 $info = $info|0;
 var $$$us11 = 0, $$1$us13 = 0, $$3$us12 = 0, $$lcssa = 0, $$lcssa31 = 0, $$pre = 0, $$pre21 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $cond = 0;
 var $dst$05 = 0, $dst$05$us = 0, $dst$05$us9 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $src$04 = 0, $src$04$us = 0, $src$04$us10 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre21 = HEAP32[$6>>2]|0;
 $cond = ($10|0)==(64);
 $83 = $$pre;$86 = $$pre21;$90 = $$pre21;$91 = $$pre;
 while(1) {
  $11 = HEAP32[$7>>2]|0;
  $12 = (($11) + -1)|0;
  $13 = ($11|0)==(0);
  if (!($13)) {
   switch ($10|0) {
   case 16:  {
    $25 = $12;$dst$05$us = $90;$src$04$us = $91;
    while(1) {
     $14 = HEAP32[$src$04$us>>2]|0;
     $15 = $14 >>> 16;
     $16 = $15 & 255;
     $17 = $14 & 65280;
     $18 = $14 << 16;
     $19 = $18 & 16711680;
     $20 = $17 | $16;
     $21 = $20 | $19;
     HEAP32[$dst$05$us>>2] = $21;
     $22 = ((($src$04$us)) + 4|0);
     $23 = ((($dst$05$us)) + 4|0);
     $24 = (($25) + -1)|0;
     $26 = ($25|0)==(0);
     if ($26) {
      break;
     } else {
      $25 = $24;$dst$05$us = $23;$src$04$us = $22;
     }
    }
    break;
   }
   case 32:  {
    $52 = $12;$dst$05$us9 = $90;$src$04$us10 = $91;
    while(1) {
     $27 = HEAP32[$src$04$us10>>2]|0;
     $28 = $27 >>> 16;
     $29 = $28 & 255;
     $30 = $27 >>> 8;
     $31 = $30 & 255;
     $32 = $27 & 255;
     $33 = HEAP32[$dst$05$us9>>2]|0;
     $34 = $33 >>> 16;
     $35 = $34 & 255;
     $36 = $33 >>> 8;
     $37 = $36 & 255;
     $38 = $33 & 255;
     $39 = (($29) + ($38))|0;
     $40 = ($39>>>0)>(255);
     $$$us11 = $40 ? 255 : $39;
     $41 = (($37) + ($31))|0;
     $42 = ($41>>>0)>(255);
     $$3$us12 = $42 ? 255 : $41;
     $43 = (($35) + ($32))|0;
     $44 = ($43>>>0)>(255);
     $$1$us13 = $44 ? 255 : $43;
     $45 = $$1$us13 << 16;
     $46 = $$3$us12 << 8;
     $47 = $46 | $$$us11;
     $48 = $47 | $45;
     HEAP32[$dst$05$us9>>2] = $48;
     $49 = ((($src$04$us10)) + 4|0);
     $50 = ((($dst$05$us9)) + 4|0);
     $51 = (($52) + -1)|0;
     $53 = ($52|0)==(0);
     if ($53) {
      break;
     } else {
      $52 = $51;$dst$05$us9 = $50;$src$04$us10 = $49;
     }
    }
    break;
   }
   default: {
    $79 = $12;$dst$05 = $90;$src$04 = $91;
    while(1) {
     $54 = HEAP32[$src$04>>2]|0;
     $55 = HEAP32[$dst$05>>2]|0;
     $56 = $55 >>> 16;
     $57 = $56 & 255;
     $58 = $55 >>> 8;
     $59 = $58 & 255;
     $60 = $55 & 255;
     if ($cond) {
      $61 = $54 & 255;
      $62 = $54 >>> 8;
      $63 = $62 & 255;
      $64 = $54 >>> 16;
      $65 = $64 & 255;
      $66 = Math_imul($65, $60)|0;
      $67 = (($66>>>0) / 255)&-1;
      $68 = Math_imul($59, $63)|0;
      $69 = (($68>>>0) / 255)&-1;
      $70 = Math_imul($57, $61)|0;
      $71 = (($70>>>0) / 255)&-1;
      $dstB$0 = $71;$dstG$0 = $69;$dstR$0 = $67;
     } else {
      $dstB$0 = $57;$dstG$0 = $59;$dstR$0 = $60;
     }
     $72 = $dstB$0 << 16;
     $73 = $dstG$0 << 8;
     $74 = $73 | $dstR$0;
     $75 = $74 | $72;
     HEAP32[$dst$05>>2] = $75;
     $76 = ((($src$04)) + 4|0);
     $77 = ((($dst$05)) + 4|0);
     $78 = (($79) + -1)|0;
     $80 = ($79|0)==(0);
     if ($80) {
      break;
     } else {
      $79 = $78;$dst$05 = $77;$src$04 = $76;
     }
    }
   }
   }
  }
  $81 = HEAP32[$8>>2]|0;
  $82 = (($83) + ($81)|0);
  $84 = HEAP32[$9>>2]|0;
  $85 = (($86) + ($84)|0);
  $87 = HEAP32[$2>>2]|0;
  $88 = (($87) + -1)|0;
  HEAP32[$2>>2] = $88;
  $89 = ($87|0)==(0);
  if ($89) {
   $$lcssa = $82;$$lcssa31 = $85;
   break;
  } else {
   $83 = $82;$86 = $85;$90 = $85;$91 = $82;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa31;
 return;
}
function _SDL_Blit_RGB888_BGR888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$lcssa = 0, $$pre = 0, $$pre30 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dst$016$us = 0, $dstB$0 = 0;
 var $dstG$0 = 0, $dstR$0 = 0, $posx$014 = 0, $posx$014$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$020 = 0, $posy$1$lcssa = 0, $src$015 = 0, $src$015$us = 0, $src$1 = 0, $src$1$us = 0, $srcx$013 = 0, $srcx$013$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 112;
 $19 = ((($info)) + 12|0);
 $20 = ($18|0)==(16);
 $$pre = HEAP32[$16>>2]|0;
 $105 = $$pre;$110 = $$pre;$29 = $12;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $21 = ($posy$020|0)>(65535);
  if ($21) {
   $22 = (($posy$020) + -65536)|0;
   $23 = $22 >>> 16;
   $24 = $23 << 16;
   $25 = (($srcy$019) + 1)|0;
   $26 = (($22) - ($24))|0;
   $27 = (($25) + ($23))|0;
   $posy$1$lcssa = $26;$srcy$1$lcssa = $27;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $28 = (($29) + -1)|0;
  $30 = ($29|0)==(0);
  L9: do {
   if (!($30)) {
    if ($20) {
     $55 = $28;$dst$016$us = $110;$posx$014$us = 65536;$src$015$us = 0;$srcx$013$us = -1;
     while(1) {
      $31 = ($posx$014$us|0)>(65535);
      if ($31) {
       $32 = (($posx$014$us) + -65536)|0;
       $33 = $32 >>> 16;
       $34 = $33 << 16;
       $35 = (($srcx$013$us) + 1)|0;
       $36 = (($32) - ($34))|0;
       $37 = (($35) + ($33))|0;
       $38 = HEAP32[$info>>2]|0;
       $39 = HEAP32[$19>>2]|0;
       $40 = Math_imul($39, $srcy$1$lcssa)|0;
       $41 = (($38) + ($40)|0);
       $42 = $37 << 2;
       $43 = (($41) + ($42)|0);
       $posx$2$us = $36;$src$1$us = $43;$srcx$2$us = $37;
      } else {
       $posx$2$us = $posx$014$us;$src$1$us = $src$015$us;$srcx$2$us = $srcx$013$us;
      }
      $44 = HEAP32[$src$1$us>>2]|0;
      $45 = $44 >>> 16;
      $46 = $45 & 255;
      $47 = $44 & 65280;
      $48 = $44 << 16;
      $49 = $48 & 16711680;
      $50 = $47 | $46;
      $51 = $50 | $49;
      HEAP32[$dst$016$us>>2] = $51;
      $52 = (($posx$2$us) + ($13))|0;
      $53 = ((($dst$016$us)) + 4|0);
      $54 = (($55) + -1)|0;
      $56 = ($55|0)==(0);
      if ($56) {
       break L9;
      } else {
       $55 = $54;$dst$016$us = $53;$posx$014$us = $52;$src$015$us = $src$1$us;$srcx$013$us = $srcx$2$us;
      }
     }
    } else {
     $101 = $28;$dst$016 = $110;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
    }
    while(1) {
     $57 = ($posx$014|0)>(65535);
     if ($57) {
      $58 = (($posx$014) + -65536)|0;
      $59 = $58 >>> 16;
      $60 = $59 << 16;
      $61 = (($srcx$013) + 1)|0;
      $62 = (($58) - ($60))|0;
      $63 = (($61) + ($59))|0;
      $64 = HEAP32[$info>>2]|0;
      $65 = HEAP32[$19>>2]|0;
      $66 = Math_imul($65, $srcy$1$lcssa)|0;
      $67 = (($64) + ($66)|0);
      $68 = $63 << 2;
      $69 = (($67) + ($68)|0);
      $posx$2 = $62;$src$1 = $69;$srcx$2 = $63;
     } else {
      $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
     }
     $70 = HEAP32[$src$1>>2]|0;
     $71 = $70 >>> 16;
     $72 = $71 & 255;
     $73 = $70 >>> 8;
     $74 = $73 & 255;
     $75 = $70 & 255;
     $76 = HEAP32[$dst$016>>2]|0;
     $77 = $76 >>> 16;
     $78 = $77 & 255;
     $79 = $76 >>> 8;
     $80 = $79 & 255;
     $81 = $76 & 255;
     switch ($18|0) {
     case 64:  {
      $88 = Math_imul($72, $81)|0;
      $89 = (($88>>>0) / 255)&-1;
      $90 = Math_imul($80, $74)|0;
      $91 = (($90>>>0) / 255)&-1;
      $92 = Math_imul($78, $75)|0;
      $93 = (($92>>>0) / 255)&-1;
      $dstB$0 = $93;$dstG$0 = $91;$dstR$0 = $89;
      break;
     }
     case 32:  {
      $82 = (($72) + ($81))|0;
      $83 = ($82>>>0)>(255);
      $$ = $83 ? 255 : $82;
      $84 = (($80) + ($74))|0;
      $85 = ($84>>>0)>(255);
      $$3 = $85 ? 255 : $84;
      $86 = (($78) + ($75))|0;
      $87 = ($86>>>0)>(255);
      $$1 = $87 ? 255 : $86;
      $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
      break;
     }
     default: {
      $dstB$0 = $78;$dstG$0 = $80;$dstR$0 = $81;
     }
     }
     $94 = $dstB$0 << 16;
     $95 = $dstG$0 << 8;
     $96 = $95 | $dstR$0;
     $97 = $96 | $94;
     HEAP32[$dst$016>>2] = $97;
     $98 = (($posx$2) + ($13))|0;
     $99 = ((($dst$016)) + 4|0);
     $100 = (($101) + -1)|0;
     $102 = ($101|0)==(0);
     if ($102) {
      break L9;
     } else {
      $101 = $100;$dst$016 = $99;$posx$014 = $98;$src$015 = $src$1;$srcx$013 = $srcx$2;
     }
    }
   }
  } while(0);
  $103 = HEAP32[$17>>2]|0;
  $104 = (($105) + ($103)|0);
  $106 = HEAP32[$5>>2]|0;
  $107 = (($106) + -1)|0;
  HEAP32[$5>>2] = $107;
  $108 = ($106|0)==(0);
  if ($108) {
   $$lcssa = $104;
   break;
  }
  $109 = (($posy$1$lcssa) + ($7))|0;
  $$pre30 = HEAP32[$11>>2]|0;
  $105 = $104;$110 = $104;$29 = $$pre30;$posy$020 = $109;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGB888_BGR888_Modulate($info) {
 $info = $info|0;
 var $$in = 0, $$in13 = 0, $$lcssa = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$pre16 = 0, $$pre17 = 0, $$pre18 = 0, $$pre19 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $dst$02 = 0, $dst$02$us$us = 0, $src$01 = 0, $src$01$us$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = ((($info)) + 20|0);
 $14 = ((($info)) + 24|0);
 $15 = ((($info)) + 12|0);
 $16 = ((($info)) + 32|0);
 $17 = $1 & 1;
 $18 = ($17|0)==(0);
 $$pre17 = HEAP32[$info>>2]|0;
 $$pre18 = HEAP32[$13>>2]|0;
 if ($18) {
  $23 = $$pre17;$26 = $$pre18;$73 = $11;$74 = $$pre18;$75 = $$pre17;
  while(1) {
   $19 = HEAP32[$14>>2]|0;
   $20 = ($19|0)==(0);
   if ($20) {
    $28 = $73;
   } else {
    $$in13 = $19;$dst$02$us$us = $74;$src$01$us$us = $75;
    while(1) {
     $30 = (($$in13) + -1)|0;
     $31 = HEAP32[$src$01$us$us>>2]|0;
     $32 = $31 >>> 16;
     $33 = $32 & 255;
     $34 = $31 & 65280;
     $35 = $31 << 16;
     $36 = $35 & 16711680;
     $37 = $34 | $33;
     $38 = $37 | $36;
     HEAP32[$dst$02$us$us>>2] = $38;
     $39 = ((($src$01$us$us)) + 4|0);
     $40 = ((($dst$02$us$us)) + 4|0);
     $41 = ($30|0)==(0);
     if ($41) {
      break;
     } else {
      $$in13 = $30;$dst$02$us$us = $40;$src$01$us$us = $39;
     }
    }
    $$pre19 = HEAP32[$9>>2]|0;
    $28 = $$pre19;
   }
   $21 = HEAP32[$15>>2]|0;
   $22 = (($23) + ($21)|0);
   $24 = HEAP32[$16>>2]|0;
   $25 = (($26) + ($24)|0);
   $27 = (($28) + -1)|0;
   HEAP32[$9>>2] = $27;
   $29 = ($28|0)==(0);
   if ($29) {
    $$lcssa = $22;$$lcssa30 = $25;
    break;
   } else {
    $23 = $22;$26 = $25;$73 = $27;$74 = $25;$75 = $22;
   }
  }
  HEAP32[$info>>2] = $$lcssa;
  HEAP32[$13>>2] = $$lcssa30;
  return;
 } else {
  $66 = $$pre17;$69 = $$pre18;$76 = $11;$77 = $$pre18;$78 = $$pre17;
  while(1) {
   $42 = HEAP32[$14>>2]|0;
   $43 = ($42|0)==(0);
   if ($43) {
    $71 = $76;
   } else {
    $$in = $42;$dst$02 = $77;$src$01 = $78;
    while(1) {
     $44 = (($$in) + -1)|0;
     $45 = HEAP32[$src$01>>2]|0;
     $46 = $45 >>> 16;
     $47 = $46 & 255;
     $48 = $45 >>> 8;
     $49 = $48 & 255;
     $50 = $45 & 255;
     $51 = Math_imul($47, $4)|0;
     $52 = (($51>>>0) / 255)&-1;
     $53 = Math_imul($49, $6)|0;
     $54 = (($53>>>0) / 255)&-1;
     $55 = Math_imul($50, $8)|0;
     $56 = (($55>>>0) / 255)&-1;
     $57 = $56 << 16;
     $58 = $54 << 8;
     $59 = $58 | $52;
     $60 = $59 | $57;
     HEAP32[$dst$02>>2] = $60;
     $61 = ((($src$01)) + 4|0);
     $62 = ((($dst$02)) + 4|0);
     $63 = ($44|0)==(0);
     if ($63) {
      break;
     } else {
      $$in = $44;$dst$02 = $62;$src$01 = $61;
     }
    }
    $$pre16 = HEAP32[$9>>2]|0;
    $71 = $$pre16;
   }
   $64 = HEAP32[$15>>2]|0;
   $65 = (($66) + ($64)|0);
   $67 = HEAP32[$16>>2]|0;
   $68 = (($69) + ($67)|0);
   $70 = (($71) + -1)|0;
   HEAP32[$9>>2] = $70;
   $72 = ($71|0)==(0);
   if ($72) {
    $$lcssa31 = $65;$$lcssa32 = $68;
    break;
   } else {
    $66 = $65;$69 = $68;$76 = $70;$77 = $68;$78 = $65;
   }
  }
  HEAP32[$info>>2] = $$lcssa31;
  HEAP32[$13>>2] = $$lcssa32;
  return;
 }
}
function _SDL_Blit_RGB888_BGR888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$013 = 0, $dst$013$us = 0, $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$017 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0;
 var $src$1$us = 0, $srcx$010 = 0, $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$016 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = ((($info)) + 20|0);
 $24 = ((($info)) + 32|0);
 $25 = $1 & 1;
 $26 = ($25|0)==(0);
 $27 = ((($info)) + 12|0);
 $$pre = HEAP32[$23>>2]|0;
 $100 = $$pre;$105 = $$pre;$36 = $19;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $28 = ($posy$017|0)>(65535);
  if ($28) {
   $29 = (($posy$017) + -65536)|0;
   $30 = $29 >>> 16;
   $31 = $30 << 16;
   $32 = (($srcy$016) + 1)|0;
   $33 = (($29) - ($31))|0;
   $34 = (($32) + ($30))|0;
   $posy$1$lcssa = $33;$srcy$1$lcssa = $34;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $35 = (($36) + -1)|0;
  $37 = ($36|0)==(0);
  if (!($37)) {
   if ($26) {
    $62 = $35;$dst$013$us = $105;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $38 = ($posx$011$us|0)>(65535);
     if ($38) {
      $39 = (($posx$011$us) + -65536)|0;
      $40 = $39 >>> 16;
      $41 = $40 << 16;
      $42 = (($srcx$010$us) + 1)|0;
      $43 = (($39) - ($41))|0;
      $44 = (($42) + ($40))|0;
      $45 = HEAP32[$info>>2]|0;
      $46 = HEAP32[$27>>2]|0;
      $47 = Math_imul($46, $srcy$1$lcssa)|0;
      $48 = (($45) + ($47)|0);
      $49 = $44 << 2;
      $50 = (($48) + ($49)|0);
      $posx$2$us = $43;$src$1$us = $50;$srcx$2$us = $44;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $51 = HEAP32[$src$1$us>>2]|0;
     $52 = $51 >>> 16;
     $53 = $52 & 255;
     $54 = $51 & 65280;
     $55 = $51 << 16;
     $56 = $55 & 16711680;
     $57 = $56 | $54;
     $58 = $57 | $53;
     HEAP32[$dst$013$us>>2] = $58;
     $59 = (($posx$2$us) + ($20))|0;
     $60 = ((($dst$013$us)) + 4|0);
     $61 = (($62) + -1)|0;
     $63 = ($62|0)==(0);
     if ($63) {
      break;
     } else {
      $62 = $61;$dst$013$us = $60;$posx$011$us = $59;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $96 = $35;$dst$013 = $105;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $64 = ($posx$011|0)>(65535);
     if ($64) {
      $65 = (($posx$011) + -65536)|0;
      $66 = $65 >>> 16;
      $67 = $66 << 16;
      $68 = (($srcx$010) + 1)|0;
      $69 = (($65) - ($67))|0;
      $70 = (($68) + ($66))|0;
      $71 = HEAP32[$info>>2]|0;
      $72 = HEAP32[$27>>2]|0;
      $73 = Math_imul($72, $srcy$1$lcssa)|0;
      $74 = (($71) + ($73)|0);
      $75 = $70 << 2;
      $76 = (($74) + ($75)|0);
      $posx$2 = $69;$src$1 = $76;$srcx$2 = $70;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $77 = HEAP32[$src$1>>2]|0;
     $78 = $77 >>> 16;
     $79 = $78 & 255;
     $80 = $77 >>> 8;
     $81 = $80 & 255;
     $82 = $77 & 255;
     $83 = Math_imul($79, $4)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = Math_imul($81, $6)|0;
     $86 = (($85>>>0) / 255)&-1;
     $87 = Math_imul($82, $8)|0;
     $88 = (($87>>>0) / 255)&-1;
     $89 = $88 << 16;
     $90 = $86 << 8;
     $91 = $89 | $90;
     $92 = $91 | $84;
     HEAP32[$dst$013>>2] = $92;
     $93 = (($posx$2) + ($20))|0;
     $94 = ((($dst$013)) + 4|0);
     $95 = (($96) + -1)|0;
     $97 = ($96|0)==(0);
     if ($97) {
      break;
     } else {
      $96 = $95;$dst$013 = $94;$posx$011 = $93;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $98 = HEAP32[$24>>2]|0;
  $99 = (($100) + ($98)|0);
  $101 = HEAP32[$12>>2]|0;
  $102 = (($101) + -1)|0;
  HEAP32[$12>>2] = $102;
  $103 = ($101|0)==(0);
  if ($103) {
   $$lcssa = $99;
   break;
  }
  $104 = (($posy$1$lcssa) + ($14))|0;
  $$pre27 = HEAP32[$18>>2]|0;
  $100 = $99;$105 = $99;$36 = $$pre27;$posy$017 = $104;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$23>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGB888_BGR888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$$us = 0, $$1 = 0, $$1$us = 0, $$3 = 0, $$3$us = 0, $$4 = 0, $$lcssa = 0, $$lcssa18 = 0, $$pre = 0, $$pre11 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$06 = 0, $dst$06$us = 0, $dstB$0 = 0, $dstB$0$us = 0, $dstG$0 = 0, $dstG$0$us = 0;
 var $dstR$0 = 0, $dstR$0$us = 0, $or$cond = 0, $src$05 = 0, $src$05$us = 0, $srcB$1 = 0, $srcB$1$us = 0, $srcG$1 = 0, $srcG$1$us = 0, $srcR$1 = 0, $srcR$1$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = $3 >>> 24;
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $$4 = $21 ? 255 : $13;
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = ($$4|0)!=(255);
 $or$cond = $23 & $24;
 $25 = $1 & 112;
 $26 = $$4 ^ 255;
 $$pre = HEAP32[$info>>2]|0;
 $$pre11 = HEAP32[$14>>2]|0;
 $134 = $$pre;$137 = $$pre11;$141 = $$pre11;$142 = $$pre;
 while(1) {
  $27 = HEAP32[$15>>2]|0;
  $28 = (($27) + -1)|0;
  $29 = ($27|0)==(0);
  L6: do {
   if (!($29)) {
    if ($19) {
     $76 = $28;$dst$06$us = $141;$src$05$us = $142;
     while(1) {
      $30 = HEAP32[$src$05$us>>2]|0;
      $31 = $30 >>> 16;
      $32 = $31 & 255;
      $33 = $30 >>> 8;
      $34 = $33 & 255;
      $35 = $30 & 255;
      $36 = HEAP32[$dst$06$us>>2]|0;
      $37 = $36 >>> 16;
      $38 = $37 & 255;
      $39 = $36 >>> 8;
      $40 = $39 & 255;
      $41 = $36 & 255;
      if ($or$cond) {
       $42 = Math_imul($32, $$4)|0;
       $43 = (($42>>>0) / 255)&-1;
       $44 = Math_imul($34, $$4)|0;
       $45 = (($44>>>0) / 255)&-1;
       $46 = Math_imul($35, $$4)|0;
       $47 = (($46>>>0) / 255)&-1;
       $srcB$1$us = $47;$srcG$1$us = $45;$srcR$1$us = $43;
      } else {
       $srcB$1$us = $35;$srcG$1$us = $34;$srcR$1$us = $32;
      }
      switch ($25|0) {
      case 16:  {
       $60 = Math_imul($41, $26)|0;
       $61 = (($60>>>0) / 255)&-1;
       $62 = (($srcR$1$us) + ($61))|0;
       $63 = Math_imul($40, $26)|0;
       $64 = (($63>>>0) / 255)&-1;
       $65 = (($srcG$1$us) + ($64))|0;
       $66 = Math_imul($38, $26)|0;
       $67 = (($66>>>0) / 255)&-1;
       $68 = (($srcB$1$us) + ($67))|0;
       $dstB$0$us = $68;$dstG$0$us = $65;$dstR$0$us = $62;
       break;
      }
      case 32:  {
       $54 = (($srcR$1$us) + ($41))|0;
       $55 = ($54>>>0)>(255);
       $$$us = $55 ? 255 : $54;
       $56 = (($srcG$1$us) + ($40))|0;
       $57 = ($56>>>0)>(255);
       $$3$us = $57 ? 255 : $56;
       $58 = (($srcB$1$us) + ($38))|0;
       $59 = ($58>>>0)>(255);
       $$1$us = $59 ? 255 : $58;
       $dstB$0$us = $$1$us;$dstG$0$us = $$3$us;$dstR$0$us = $$$us;
       break;
      }
      case 64:  {
       $48 = Math_imul($srcR$1$us, $41)|0;
       $49 = (($48>>>0) / 255)&-1;
       $50 = Math_imul($srcG$1$us, $40)|0;
       $51 = (($50>>>0) / 255)&-1;
       $52 = Math_imul($srcB$1$us, $38)|0;
       $53 = (($52>>>0) / 255)&-1;
       $dstB$0$us = $53;$dstG$0$us = $51;$dstR$0$us = $49;
       break;
      }
      default: {
       $dstB$0$us = $38;$dstG$0$us = $40;$dstR$0$us = $41;
      }
      }
      $69 = $dstB$0$us << 16;
      $70 = $dstG$0$us << 8;
      $71 = $70 | $dstR$0$us;
      $72 = $71 | $69;
      HEAP32[$dst$06$us>>2] = $72;
      $73 = ((($src$05$us)) + 4|0);
      $74 = ((($dst$06$us)) + 4|0);
      $75 = (($76) + -1)|0;
      $77 = ($76|0)==(0);
      if ($77) {
       break L6;
      } else {
       $76 = $75;$dst$06$us = $74;$src$05$us = $73;
      }
     }
    } else {
     $130 = $28;$dst$06 = $141;$src$05 = $142;
     while(1) {
      $78 = HEAP32[$src$05>>2]|0;
      $79 = $78 >>> 16;
      $80 = $79 & 255;
      $81 = $78 >>> 8;
      $82 = $81 & 255;
      $83 = $78 & 255;
      $84 = HEAP32[$dst$06>>2]|0;
      $85 = $84 >>> 16;
      $86 = $85 & 255;
      $87 = $84 >>> 8;
      $88 = $87 & 255;
      $89 = $84 & 255;
      $90 = Math_imul($80, $4)|0;
      $91 = (($90>>>0) / 255)&-1;
      $92 = Math_imul($82, $6)|0;
      $93 = (($92>>>0) / 255)&-1;
      $94 = Math_imul($83, $8)|0;
      $95 = (($94>>>0) / 255)&-1;
      if ($or$cond) {
       $96 = Math_imul($91, $$4)|0;
       $97 = (($96>>>0) / 255)&-1;
       $98 = Math_imul($93, $$4)|0;
       $99 = (($98>>>0) / 255)&-1;
       $100 = Math_imul($95, $$4)|0;
       $101 = (($100>>>0) / 255)&-1;
       $srcB$1 = $101;$srcG$1 = $99;$srcR$1 = $97;
      } else {
       $srcB$1 = $95;$srcG$1 = $93;$srcR$1 = $91;
      }
      switch ($25|0) {
      case 16:  {
       $102 = Math_imul($89, $26)|0;
       $103 = (($102>>>0) / 255)&-1;
       $104 = (($srcR$1) + ($103))|0;
       $105 = Math_imul($88, $26)|0;
       $106 = (($105>>>0) / 255)&-1;
       $107 = (($srcG$1) + ($106))|0;
       $108 = Math_imul($86, $26)|0;
       $109 = (($108>>>0) / 255)&-1;
       $110 = (($srcB$1) + ($109))|0;
       $dstB$0 = $110;$dstG$0 = $107;$dstR$0 = $104;
       break;
      }
      case 32:  {
       $111 = (($srcR$1) + ($89))|0;
       $112 = ($111>>>0)>(255);
       $$ = $112 ? 255 : $111;
       $113 = (($srcG$1) + ($88))|0;
       $114 = ($113>>>0)>(255);
       $$3 = $114 ? 255 : $113;
       $115 = (($srcB$1) + ($86))|0;
       $116 = ($115>>>0)>(255);
       $$1 = $116 ? 255 : $115;
       $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
       break;
      }
      case 64:  {
       $117 = Math_imul($srcR$1, $89)|0;
       $118 = (($117>>>0) / 255)&-1;
       $119 = Math_imul($srcG$1, $88)|0;
       $120 = (($119>>>0) / 255)&-1;
       $121 = Math_imul($srcB$1, $86)|0;
       $122 = (($121>>>0) / 255)&-1;
       $dstB$0 = $122;$dstG$0 = $120;$dstR$0 = $118;
       break;
      }
      default: {
       $dstB$0 = $86;$dstG$0 = $88;$dstR$0 = $89;
      }
      }
      $123 = $dstB$0 << 16;
      $124 = $dstG$0 << 8;
      $125 = $124 | $dstR$0;
      $126 = $125 | $123;
      HEAP32[$dst$06>>2] = $126;
      $127 = ((($src$05)) + 4|0);
      $128 = ((($dst$06)) + 4|0);
      $129 = (($130) + -1)|0;
      $131 = ($130|0)==(0);
      if ($131) {
       break L6;
      } else {
       $130 = $129;$dst$06 = $128;$src$05 = $127;
      }
     }
    }
   }
  } while(0);
  $132 = HEAP32[$16>>2]|0;
  $133 = (($134) + ($132)|0);
  $135 = HEAP32[$17>>2]|0;
  $136 = (($137) + ($135)|0);
  $138 = HEAP32[$9>>2]|0;
  $139 = (($138) + -1)|0;
  HEAP32[$9>>2] = $139;
  $140 = ($138|0)==(0);
  if ($140) {
   $$lcssa = $133;$$lcssa18 = $136;
   break;
  } else {
   $134 = $133;$137 = $136;$141 = $136;$142 = $133;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa18;
 return;
}
function _SDL_Blit_RGB888_BGR888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$4 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre28 = 0, $$pre29 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$017 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$015 = 0, $posx$2 = 0, $posy$021 = 0, $posy$1$lcssa = 0, $src$016 = 0, $src$1 = 0;
 var $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$014 = 0, $srcx$2 = 0, $srcy$020 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = $3 >>> 24;
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $$4 = $29 ? 255 : $23;
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = ($$4|0)!=(255);
 $or$cond = $31 & $32;
 $33 = $1 & 112;
 $34 = $$4 ^ 255;
 $35 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $113 = $$pre;$118 = $21;$119 = $$pre;$44 = $19;$posy$021 = 0;$srcy$020 = 0;
 while(1) {
  $36 = ($posy$021|0)>(65535);
  if ($36) {
   $37 = (($posy$021) + -65536)|0;
   $38 = $37 >>> 16;
   $39 = $38 << 16;
   $40 = (($srcy$020) + 1)|0;
   $41 = (($37) - ($39))|0;
   $42 = (($40) + ($38))|0;
   $posy$1$lcssa = $41;$srcy$1$lcssa = $42;
  } else {
   $posy$1$lcssa = $posy$021;$srcy$1$lcssa = $srcy$020;
  }
  $43 = ($44|0)==(0);
  if ($43) {
   $115 = $118;
  } else {
   $$in = $44;$dst$017 = $119;$posx$015 = 65536;$src$016 = 0;$srcx$014 = -1;
   while(1) {
    $45 = (($$in) + -1)|0;
    $46 = ($posx$015|0)>(65535);
    if ($46) {
     $47 = (($posx$015) + -65536)|0;
     $48 = $47 >>> 16;
     $49 = $48 << 16;
     $50 = (($srcx$014) + 1)|0;
     $51 = (($47) - ($49))|0;
     $52 = (($50) + ($48))|0;
     $53 = HEAP32[$info>>2]|0;
     $54 = HEAP32[$35>>2]|0;
     $55 = Math_imul($54, $srcy$1$lcssa)|0;
     $56 = (($53) + ($55)|0);
     $57 = $52 << 2;
     $58 = (($56) + ($57)|0);
     $posx$2 = $51;$src$1 = $58;$srcx$2 = $52;
    } else {
     $posx$2 = $posx$015;$src$1 = $src$016;$srcx$2 = $srcx$014;
    }
    $59 = HEAP32[$src$1>>2]|0;
    $60 = $59 >>> 16;
    $61 = $60 & 255;
    $62 = $59 >>> 8;
    $63 = $62 & 255;
    $64 = $59 & 255;
    $65 = HEAP32[$dst$017>>2]|0;
    $66 = $65 >>> 16;
    $67 = $66 & 255;
    $68 = $65 >>> 8;
    $69 = $68 & 255;
    $70 = $65 & 255;
    if ($27) {
     $srcB$0 = $64;$srcG$0 = $63;$srcR$0 = $61;
    } else {
     $71 = Math_imul($61, $4)|0;
     $72 = (($71>>>0) / 255)&-1;
     $73 = Math_imul($63, $6)|0;
     $74 = (($73>>>0) / 255)&-1;
     $75 = Math_imul($64, $8)|0;
     $76 = (($75>>>0) / 255)&-1;
     $srcB$0 = $76;$srcG$0 = $74;$srcR$0 = $72;
    }
    if ($or$cond) {
     $77 = Math_imul($srcR$0, $$4)|0;
     $78 = (($77>>>0) / 255)&-1;
     $79 = Math_imul($srcG$0, $$4)|0;
     $80 = (($79>>>0) / 255)&-1;
     $81 = Math_imul($srcB$0, $$4)|0;
     $82 = (($81>>>0) / 255)&-1;
     $srcB$1 = $82;$srcG$1 = $80;$srcR$1 = $78;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($33|0) {
    case 16:  {
     $83 = Math_imul($70, $34)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = (($srcR$1) + ($84))|0;
     $86 = Math_imul($69, $34)|0;
     $87 = (($86>>>0) / 255)&-1;
     $88 = (($srcG$1) + ($87))|0;
     $89 = Math_imul($67, $34)|0;
     $90 = (($89>>>0) / 255)&-1;
     $91 = (($srcB$1) + ($90))|0;
     $dstB$0 = $91;$dstG$0 = $88;$dstR$0 = $85;
     break;
    }
    case 32:  {
     $92 = (($srcR$1) + ($70))|0;
     $93 = ($92>>>0)>(255);
     $$ = $93 ? 255 : $92;
     $94 = (($srcG$1) + ($69))|0;
     $95 = ($94>>>0)>(255);
     $$3 = $95 ? 255 : $94;
     $96 = (($srcB$1) + ($67))|0;
     $97 = ($96>>>0)>(255);
     $$1 = $97 ? 255 : $96;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $98 = Math_imul($srcR$1, $70)|0;
     $99 = (($98>>>0) / 255)&-1;
     $100 = Math_imul($srcG$1, $69)|0;
     $101 = (($100>>>0) / 255)&-1;
     $102 = Math_imul($srcB$1, $67)|0;
     $103 = (($102>>>0) / 255)&-1;
     $dstB$0 = $103;$dstG$0 = $101;$dstR$0 = $99;
     break;
    }
    default: {
     $dstB$0 = $67;$dstG$0 = $69;$dstR$0 = $70;
    }
    }
    $104 = $dstB$0 << 16;
    $105 = $dstG$0 << 8;
    $106 = $105 | $dstR$0;
    $107 = $106 | $104;
    HEAP32[$dst$017>>2] = $107;
    $108 = (($posx$2) + ($20))|0;
    $109 = ((($dst$017)) + 4|0);
    $110 = ($45|0)==(0);
    if ($110) {
     break;
    } else {
     $$in = $45;$dst$017 = $109;$posx$015 = $108;$src$016 = $src$1;$srcx$014 = $srcx$2;
    }
   }
   $$pre29 = HEAP32[$12>>2]|0;
   $115 = $$pre29;
  }
  $111 = HEAP32[$25>>2]|0;
  $112 = (($113) + ($111)|0);
  $114 = (($115) + -1)|0;
  HEAP32[$12>>2] = $114;
  $116 = ($115|0)==(0);
  if ($116) {
   $$lcssa = $112;
   break;
  }
  $117 = (($posy$1$lcssa) + ($14))|0;
  $$pre28 = HEAP32[$18>>2]|0;
  $113 = $112;$118 = $114;$119 = $112;$44 = $$pre28;$posy$021 = $117;$srcy$020 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGB888_ARGB8888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0;
 var $posx$010 = 0, $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$47 = $$pre;$52 = $12;$53 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $49 = $52;
  } else {
   $$in = $25;$dst$013 = $53;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    $41 = $40 | -16777216;
    HEAP32[$dst$013>>2] = $41;
    $42 = (($posx$2) + ($11))|0;
    $43 = ((($dst$013)) + 4|0);
    $44 = ($26|0)==(0);
    if ($44) {
     break;
    } else {
     $$in = $26;$dst$013 = $43;$posx$010 = $42;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $49 = $$pre25;
  }
  $45 = HEAP32[$15>>2]|0;
  $46 = (($47) + ($45)|0);
  $48 = (($49) + -1)|0;
  HEAP32[$3>>2] = $48;
  $50 = ($49|0)==(0);
  if ($50) {
   $$lcssa = $46;
   break;
  }
  $51 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$47 = $46;$52 = $48;$53 = $46;$posy$016 = $51;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGB888_ARGB8888_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$lcssa = 0, $$lcssa17 = 0, $$pre = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$05 = 0, $dst$05$us = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0;
 var $src$04 = 0, $src$04$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 112;
 $11 = ($10|0)==(16);
 $$pre = HEAP32[$info>>2]|0;
 $$pre10 = HEAP32[$6>>2]|0;
 $59 = $$pre;$62 = $$pre10;$66 = $$pre10;$67 = $$pre;
 while(1) {
  $12 = HEAP32[$7>>2]|0;
  $13 = (($12) + -1)|0;
  $14 = ($12|0)==(0);
  L6: do {
   if (!($14)) {
    if ($11) {
     $20 = $13;$dst$05$us = $66;$src$04$us = $67;
     while(1) {
      $15 = HEAP32[$src$04$us>>2]|0;
      $16 = $15 | -16777216;
      HEAP32[$dst$05$us>>2] = $16;
      $17 = ((($src$04$us)) + 4|0);
      $18 = ((($dst$05$us)) + 4|0);
      $19 = (($20) + -1)|0;
      $21 = ($20|0)==(0);
      if ($21) {
       break L6;
      } else {
       $20 = $19;$dst$05$us = $18;$src$04$us = $17;
      }
     }
    } else {
     $55 = $13;$dst$05 = $66;$src$04 = $67;
    }
    while(1) {
     $22 = HEAP32[$src$04>>2]|0;
     $23 = $22 >>> 16;
     $24 = $23 & 255;
     $25 = $22 >>> 8;
     $26 = $25 & 255;
     $27 = $22 & 255;
     $28 = HEAP32[$dst$05>>2]|0;
     $29 = $28 >>> 16;
     $30 = $29 & 255;
     $31 = $28 >>> 8;
     $32 = $31 & 255;
     $33 = $28 & 255;
     $34 = $28 & -16777216;
     switch ($10|0) {
     case 64:  {
      $41 = Math_imul($30, $24)|0;
      $42 = (($41>>>0) / 255)&-1;
      $43 = Math_imul($32, $26)|0;
      $44 = (($43>>>0) / 255)&-1;
      $45 = Math_imul($33, $27)|0;
      $46 = (($45>>>0) / 255)&-1;
      $dstB$0 = $46;$dstG$0 = $44;$dstR$0 = $42;
      break;
     }
     case 32:  {
      $35 = (($30) + ($24))|0;
      $36 = ($35>>>0)>(255);
      $$ = $36 ? 255 : $35;
      $37 = (($32) + ($26))|0;
      $38 = ($37>>>0)>(255);
      $$3 = $38 ? 255 : $37;
      $39 = (($33) + ($27))|0;
      $40 = ($39>>>0)>(255);
      $$1 = $40 ? 255 : $39;
      $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
      break;
     }
     default: {
      $dstB$0 = $33;$dstG$0 = $32;$dstR$0 = $30;
     }
     }
     $47 = $dstR$0 << 16;
     $48 = $dstG$0 << 8;
     $49 = $48 | $47;
     $50 = $49 | $dstB$0;
     $51 = $50 | $34;
     HEAP32[$dst$05>>2] = $51;
     $52 = ((($src$04)) + 4|0);
     $53 = ((($dst$05)) + 4|0);
     $54 = (($55) + -1)|0;
     $56 = ($55|0)==(0);
     if ($56) {
      break L6;
     } else {
      $55 = $54;$dst$05 = $53;$src$04 = $52;
     }
    }
   }
  } while(0);
  $57 = HEAP32[$8>>2]|0;
  $58 = (($59) + ($57)|0);
  $60 = HEAP32[$9>>2]|0;
  $61 = (($62) + ($60)|0);
  $63 = HEAP32[$2>>2]|0;
  $64 = (($63) + -1)|0;
  HEAP32[$2>>2] = $64;
  $65 = ($63|0)==(0);
  if ($65) {
   $$lcssa = $58;$$lcssa17 = $61;
   break;
  } else {
   $59 = $58;$62 = $61;$66 = $61;$67 = $58;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa17;
 return;
}
function _SDL_Blit_RGB888_ARGB8888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$lcssa = 0, $$pre = 0, $$pre30 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dst$016$us = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $posx$014 = 0, $posx$014$us = 0;
 var $posx$2 = 0, $posx$2$us = 0, $posy$020 = 0, $posy$1$lcssa = 0, $src$015 = 0, $src$015$us = 0, $src$1 = 0, $src$1$us = 0, $srcx$013 = 0, $srcx$013$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 112;
 $19 = ((($info)) + 12|0);
 $20 = ($18|0)==(16);
 $$pre = HEAP32[$16>>2]|0;
 $101 = $$pre;$106 = $$pre;$29 = $12;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $21 = ($posy$020|0)>(65535);
  if ($21) {
   $22 = (($posy$020) + -65536)|0;
   $23 = $22 >>> 16;
   $24 = $23 << 16;
   $25 = (($srcy$019) + 1)|0;
   $26 = (($22) - ($24))|0;
   $27 = (($25) + ($23))|0;
   $posy$1$lcssa = $26;$srcy$1$lcssa = $27;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $28 = (($29) + -1)|0;
  $30 = ($29|0)==(0);
  L9: do {
   if (!($30)) {
    if ($20) {
     $49 = $28;$dst$016$us = $106;$posx$014$us = 65536;$src$015$us = 0;$srcx$013$us = -1;
     while(1) {
      $31 = ($posx$014$us|0)>(65535);
      if ($31) {
       $32 = (($posx$014$us) + -65536)|0;
       $33 = $32 >>> 16;
       $34 = $33 << 16;
       $35 = (($srcx$013$us) + 1)|0;
       $36 = (($32) - ($34))|0;
       $37 = (($35) + ($33))|0;
       $38 = HEAP32[$info>>2]|0;
       $39 = HEAP32[$19>>2]|0;
       $40 = Math_imul($39, $srcy$1$lcssa)|0;
       $41 = (($38) + ($40)|0);
       $42 = $37 << 2;
       $43 = (($41) + ($42)|0);
       $posx$2$us = $36;$src$1$us = $43;$srcx$2$us = $37;
      } else {
       $posx$2$us = $posx$014$us;$src$1$us = $src$015$us;$srcx$2$us = $srcx$013$us;
      }
      $44 = HEAP32[$src$1$us>>2]|0;
      $45 = $44 | -16777216;
      HEAP32[$dst$016$us>>2] = $45;
      $46 = (($posx$2$us) + ($13))|0;
      $47 = ((($dst$016$us)) + 4|0);
      $48 = (($49) + -1)|0;
      $50 = ($49|0)==(0);
      if ($50) {
       break L9;
      } else {
       $49 = $48;$dst$016$us = $47;$posx$014$us = $46;$src$015$us = $src$1$us;$srcx$013$us = $srcx$2$us;
      }
     }
    } else {
     $97 = $28;$dst$016 = $106;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
    }
    while(1) {
     $51 = ($posx$014|0)>(65535);
     if ($51) {
      $52 = (($posx$014) + -65536)|0;
      $53 = $52 >>> 16;
      $54 = $53 << 16;
      $55 = (($srcx$013) + 1)|0;
      $56 = (($52) - ($54))|0;
      $57 = (($55) + ($53))|0;
      $58 = HEAP32[$info>>2]|0;
      $59 = HEAP32[$19>>2]|0;
      $60 = Math_imul($59, $srcy$1$lcssa)|0;
      $61 = (($58) + ($60)|0);
      $62 = $57 << 2;
      $63 = (($61) + ($62)|0);
      $posx$2 = $56;$src$1 = $63;$srcx$2 = $57;
     } else {
      $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
     }
     $64 = HEAP32[$src$1>>2]|0;
     $65 = $64 >>> 16;
     $66 = $65 & 255;
     $67 = $64 >>> 8;
     $68 = $67 & 255;
     $69 = $64 & 255;
     $70 = HEAP32[$dst$016>>2]|0;
     $71 = $70 >>> 16;
     $72 = $71 & 255;
     $73 = $70 >>> 8;
     $74 = $73 & 255;
     $75 = $70 & 255;
     $76 = $70 & -16777216;
     switch ($18|0) {
     case 64:  {
      $83 = Math_imul($72, $66)|0;
      $84 = (($83>>>0) / 255)&-1;
      $85 = Math_imul($74, $68)|0;
      $86 = (($85>>>0) / 255)&-1;
      $87 = Math_imul($75, $69)|0;
      $88 = (($87>>>0) / 255)&-1;
      $dstB$0 = $88;$dstG$0 = $86;$dstR$0 = $84;
      break;
     }
     case 32:  {
      $77 = (($72) + ($66))|0;
      $78 = ($77>>>0)>(255);
      $$ = $78 ? 255 : $77;
      $79 = (($74) + ($68))|0;
      $80 = ($79>>>0)>(255);
      $$3 = $80 ? 255 : $79;
      $81 = (($75) + ($69))|0;
      $82 = ($81>>>0)>(255);
      $$1 = $82 ? 255 : $81;
      $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
      break;
     }
     default: {
      $dstB$0 = $75;$dstG$0 = $74;$dstR$0 = $72;
     }
     }
     $89 = $dstR$0 << 16;
     $90 = $dstG$0 << 8;
     $91 = $90 | $89;
     $92 = $91 | $dstB$0;
     $93 = $92 | $76;
     HEAP32[$dst$016>>2] = $93;
     $94 = (($posx$2) + ($13))|0;
     $95 = ((($dst$016)) + 4|0);
     $96 = (($97) + -1)|0;
     $98 = ($97|0)==(0);
     if ($98) {
      break L9;
     } else {
      $97 = $96;$dst$016 = $95;$posx$014 = $94;$src$015 = $src$1;$srcx$013 = $srcx$2;
     }
    }
   }
  } while(0);
  $99 = HEAP32[$17>>2]|0;
  $100 = (($101) + ($99)|0);
  $102 = HEAP32[$5>>2]|0;
  $103 = (($102) + -1)|0;
  HEAP32[$5>>2] = $103;
  $104 = ($102|0)==(0);
  if ($104) {
   $$lcssa = $100;
   break;
  }
  $105 = (($posy$1$lcssa) + ($7))|0;
  $$pre30 = HEAP32[$11>>2]|0;
  $101 = $100;$106 = $100;$29 = $$pre30;$posy$020 = $105;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGB888_ARGB8888_Modulate($info) {
 $info = $info|0;
 var $$in = 0, $$in13 = 0, $$lcssa = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$pre16 = 0, $$pre17 = 0, $$pre18 = 0, $$pre19 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $9 = 0, $dst$02 = 0, $dst$02$us$us = 0, $src$01 = 0, $src$01$us$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = $3 & -16777216;
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $22 = $21 ? -16777216 : $13;
 $$pre17 = HEAP32[$info>>2]|0;
 $$pre18 = HEAP32[$14>>2]|0;
 if ($19) {
  $27 = $$pre17;$30 = $$pre18;$77 = $11;$78 = $$pre18;$79 = $$pre17;
  while(1) {
   $23 = HEAP32[$15>>2]|0;
   $24 = ($23|0)==(0);
   if ($24) {
    $32 = $77;
   } else {
    $$in13 = $23;$dst$02$us$us = $78;$src$01$us$us = $79;
    while(1) {
     $34 = (($$in13) + -1)|0;
     $35 = HEAP32[$src$01$us$us>>2]|0;
     $36 = $35 & 16711680;
     $37 = $35 & 65280;
     $38 = $35 & 255;
     $39 = $36 | $22;
     $40 = $39 | $37;
     $41 = $40 | $38;
     HEAP32[$dst$02$us$us>>2] = $41;
     $42 = ((($src$01$us$us)) + 4|0);
     $43 = ((($dst$02$us$us)) + 4|0);
     $44 = ($34|0)==(0);
     if ($44) {
      break;
     } else {
      $$in13 = $34;$dst$02$us$us = $43;$src$01$us$us = $42;
     }
    }
    $$pre19 = HEAP32[$9>>2]|0;
    $32 = $$pre19;
   }
   $25 = HEAP32[$16>>2]|0;
   $26 = (($27) + ($25)|0);
   $28 = HEAP32[$17>>2]|0;
   $29 = (($30) + ($28)|0);
   $31 = (($32) + -1)|0;
   HEAP32[$9>>2] = $31;
   $33 = ($32|0)==(0);
   if ($33) {
    $$lcssa = $26;$$lcssa30 = $29;
    break;
   } else {
    $27 = $26;$30 = $29;$77 = $31;$78 = $29;$79 = $26;
   }
  }
  HEAP32[$info>>2] = $$lcssa;
  HEAP32[$14>>2] = $$lcssa30;
  return;
 } else {
  $70 = $$pre17;$73 = $$pre18;$80 = $11;$81 = $$pre18;$82 = $$pre17;
  while(1) {
   $45 = HEAP32[$15>>2]|0;
   $46 = ($45|0)==(0);
   if ($46) {
    $75 = $80;
   } else {
    $$in = $45;$dst$02 = $81;$src$01 = $82;
    while(1) {
     $47 = (($$in) + -1)|0;
     $48 = HEAP32[$src$01>>2]|0;
     $49 = $48 >>> 16;
     $50 = $49 & 255;
     $51 = $48 >>> 8;
     $52 = $51 & 255;
     $53 = $48 & 255;
     $54 = Math_imul($50, $4)|0;
     $55 = (($54>>>0) / 255)&-1;
     $56 = Math_imul($52, $6)|0;
     $57 = (($56>>>0) / 255)&-1;
     $58 = Math_imul($53, $8)|0;
     $59 = (($58>>>0) / 255)&-1;
     $60 = $55 << 16;
     $61 = $60 | $22;
     $62 = $57 << 8;
     $63 = $61 | $62;
     $64 = $63 | $59;
     HEAP32[$dst$02>>2] = $64;
     $65 = ((($src$01)) + 4|0);
     $66 = ((($dst$02)) + 4|0);
     $67 = ($47|0)==(0);
     if ($67) {
      break;
     } else {
      $$in = $47;$dst$02 = $66;$src$01 = $65;
     }
    }
    $$pre16 = HEAP32[$9>>2]|0;
    $75 = $$pre16;
   }
   $68 = HEAP32[$16>>2]|0;
   $69 = (($70) + ($68)|0);
   $71 = HEAP32[$17>>2]|0;
   $72 = (($73) + ($71)|0);
   $74 = (($75) + -1)|0;
   HEAP32[$9>>2] = $74;
   $76 = ($75|0)==(0);
   if ($76) {
    $$lcssa31 = $69;$$lcssa32 = $72;
    break;
   } else {
    $70 = $69;$73 = $72;$80 = $74;$81 = $72;$82 = $69;
   }
  }
  HEAP32[$info>>2] = $$lcssa31;
  HEAP32[$14>>2] = $$lcssa32;
  return;
 }
}
function _SDL_Blit_RGB888_ARGB8888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$013 = 0, $dst$013$us = 0, $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$017 = 0;
 var $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0, $src$1$us = 0, $srcx$010 = 0, $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$016 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = $3 & -16777216;
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = $29 ? -16777216 : $23;
 $31 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $104 = $$pre;$109 = $$pre;$40 = $19;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $32 = ($posy$017|0)>(65535);
  if ($32) {
   $33 = (($posy$017) + -65536)|0;
   $34 = $33 >>> 16;
   $35 = $34 << 16;
   $36 = (($srcy$016) + 1)|0;
   $37 = (($33) - ($35))|0;
   $38 = (($36) + ($34))|0;
   $posy$1$lcssa = $37;$srcy$1$lcssa = $38;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $39 = (($40) + -1)|0;
  $41 = ($40|0)==(0);
  if (!($41)) {
   if ($27) {
    $65 = $39;$dst$013$us = $109;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $42 = ($posx$011$us|0)>(65535);
     if ($42) {
      $43 = (($posx$011$us) + -65536)|0;
      $44 = $43 >>> 16;
      $45 = $44 << 16;
      $46 = (($srcx$010$us) + 1)|0;
      $47 = (($43) - ($45))|0;
      $48 = (($46) + ($44))|0;
      $49 = HEAP32[$info>>2]|0;
      $50 = HEAP32[$31>>2]|0;
      $51 = Math_imul($50, $srcy$1$lcssa)|0;
      $52 = (($49) + ($51)|0);
      $53 = $48 << 2;
      $54 = (($52) + ($53)|0);
      $posx$2$us = $47;$src$1$us = $54;$srcx$2$us = $48;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $55 = HEAP32[$src$1$us>>2]|0;
     $56 = $55 & 16711680;
     $57 = $55 & 65280;
     $58 = $55 & 255;
     $59 = $57 | $30;
     $60 = $59 | $58;
     $61 = $60 | $56;
     HEAP32[$dst$013$us>>2] = $61;
     $62 = (($posx$2$us) + ($20))|0;
     $63 = ((($dst$013$us)) + 4|0);
     $64 = (($65) + -1)|0;
     $66 = ($65|0)==(0);
     if ($66) {
      break;
     } else {
      $65 = $64;$dst$013$us = $63;$posx$011$us = $62;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $100 = $39;$dst$013 = $109;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $67 = ($posx$011|0)>(65535);
     if ($67) {
      $68 = (($posx$011) + -65536)|0;
      $69 = $68 >>> 16;
      $70 = $69 << 16;
      $71 = (($srcx$010) + 1)|0;
      $72 = (($68) - ($70))|0;
      $73 = (($71) + ($69))|0;
      $74 = HEAP32[$info>>2]|0;
      $75 = HEAP32[$31>>2]|0;
      $76 = Math_imul($75, $srcy$1$lcssa)|0;
      $77 = (($74) + ($76)|0);
      $78 = $73 << 2;
      $79 = (($77) + ($78)|0);
      $posx$2 = $72;$src$1 = $79;$srcx$2 = $73;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $80 = HEAP32[$src$1>>2]|0;
     $81 = $80 >>> 16;
     $82 = $81 & 255;
     $83 = $80 >>> 8;
     $84 = $83 & 255;
     $85 = $80 & 255;
     $86 = Math_imul($82, $4)|0;
     $87 = (($86>>>0) / 255)&-1;
     $88 = Math_imul($84, $6)|0;
     $89 = (($88>>>0) / 255)&-1;
     $90 = Math_imul($85, $8)|0;
     $91 = (($90>>>0) / 255)&-1;
     $92 = $87 << 16;
     $93 = $89 << 8;
     $94 = $93 | $30;
     $95 = $94 | $91;
     $96 = $95 | $92;
     HEAP32[$dst$013>>2] = $96;
     $97 = (($posx$2) + ($20))|0;
     $98 = ((($dst$013)) + 4|0);
     $99 = (($100) + -1)|0;
     $101 = ($100|0)==(0);
     if ($101) {
      break;
     } else {
      $100 = $99;$dst$013 = $98;$posx$011 = $97;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $102 = HEAP32[$25>>2]|0;
  $103 = (($104) + ($102)|0);
  $105 = HEAP32[$12>>2]|0;
  $106 = (($105) + -1)|0;
  HEAP32[$12>>2] = $106;
  $107 = ($105|0)==(0);
  if ($107) {
   $$lcssa = $103;
   break;
  }
  $108 = (($posy$1$lcssa) + ($14))|0;
  $$pre27 = HEAP32[$18>>2]|0;
  $104 = $103;$109 = $103;$40 = $$pre27;$posy$017 = $108;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGB888_ARGB8888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$4 = 0, $$in = 0, $$lcssa = 0, $$lcssa15 = 0, $$pre = 0, $$pre10 = 0, $$pre11 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$06 = 0, $dstA$0 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $src$05 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0;
 var $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = $3 >>> 24;
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $$4 = $21 ? 255 : $13;
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = ($$4|0)!=(255);
 $or$cond = $23 & $24;
 $25 = $1 & 112;
 $26 = $$4 ^ 255;
 $$pre = HEAP32[$info>>2]|0;
 $$pre10 = HEAP32[$14>>2]|0;
 $90 = $$pre;$93 = $$pre10;$97 = $11;$98 = $$pre10;$99 = $$pre;
 while(1) {
  $27 = HEAP32[$15>>2]|0;
  $28 = ($27|0)==(0);
  if ($28) {
   $95 = $97;
  } else {
   $$in = $27;$dst$06 = $98;$src$05 = $99;
   while(1) {
    $29 = (($$in) + -1)|0;
    $30 = HEAP32[$src$05>>2]|0;
    $31 = $30 >>> 16;
    $32 = $31 & 255;
    $33 = $30 >>> 8;
    $34 = $33 & 255;
    $35 = $30 & 255;
    $36 = HEAP32[$dst$06>>2]|0;
    $37 = $36 >>> 16;
    $38 = $37 & 255;
    $39 = $36 >>> 8;
    $40 = $39 & 255;
    $41 = $36 & 255;
    $42 = $36 >>> 24;
    if ($19) {
     $srcB$0 = $35;$srcG$0 = $34;$srcR$0 = $32;
    } else {
     $43 = Math_imul($32, $4)|0;
     $44 = (($43>>>0) / 255)&-1;
     $45 = Math_imul($34, $6)|0;
     $46 = (($45>>>0) / 255)&-1;
     $47 = Math_imul($35, $8)|0;
     $48 = (($47>>>0) / 255)&-1;
     $srcB$0 = $48;$srcG$0 = $46;$srcR$0 = $44;
    }
    if ($or$cond) {
     $49 = Math_imul($srcR$0, $$4)|0;
     $50 = (($49>>>0) / 255)&-1;
     $51 = Math_imul($srcG$0, $$4)|0;
     $52 = (($51>>>0) / 255)&-1;
     $53 = Math_imul($srcB$0, $$4)|0;
     $54 = (($53>>>0) / 255)&-1;
     $srcB$1 = $54;$srcG$1 = $52;$srcR$1 = $50;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($25|0) {
    case 16:  {
     $55 = Math_imul($38, $26)|0;
     $56 = (($55>>>0) / 255)&-1;
     $57 = (($srcR$1) + ($56))|0;
     $58 = Math_imul($40, $26)|0;
     $59 = (($58>>>0) / 255)&-1;
     $60 = (($srcG$1) + ($59))|0;
     $61 = Math_imul($41, $26)|0;
     $62 = (($61>>>0) / 255)&-1;
     $63 = (($srcB$1) + ($62))|0;
     $64 = Math_imul($42, $26)|0;
     $65 = (($64>>>0) / 255)&-1;
     $66 = (($65) + ($$4))|0;
     $dstA$0 = $66;$dstB$0 = $63;$dstG$0 = $60;$dstR$0 = $57;
     break;
    }
    case 32:  {
     $67 = (($srcR$1) + ($38))|0;
     $68 = ($67>>>0)>(255);
     $$ = $68 ? 255 : $67;
     $69 = (($srcG$1) + ($40))|0;
     $70 = ($69>>>0)>(255);
     $$3 = $70 ? 255 : $69;
     $71 = (($srcB$1) + ($41))|0;
     $72 = ($71>>>0)>(255);
     $$1 = $72 ? 255 : $71;
     $dstA$0 = $42;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $73 = Math_imul($srcR$1, $38)|0;
     $74 = (($73>>>0) / 255)&-1;
     $75 = Math_imul($srcG$1, $40)|0;
     $76 = (($75>>>0) / 255)&-1;
     $77 = Math_imul($srcB$1, $41)|0;
     $78 = (($77>>>0) / 255)&-1;
     $dstA$0 = $42;$dstB$0 = $78;$dstG$0 = $76;$dstR$0 = $74;
     break;
    }
    default: {
     $dstA$0 = $42;$dstB$0 = $41;$dstG$0 = $40;$dstR$0 = $38;
    }
    }
    $79 = $dstA$0 << 24;
    $80 = $dstR$0 << 16;
    $81 = $dstG$0 << 8;
    $82 = $81 | $80;
    $83 = $82 | $dstB$0;
    $84 = $83 | $79;
    HEAP32[$dst$06>>2] = $84;
    $85 = ((($src$05)) + 4|0);
    $86 = ((($dst$06)) + 4|0);
    $87 = ($29|0)==(0);
    if ($87) {
     break;
    } else {
     $$in = $29;$dst$06 = $86;$src$05 = $85;
    }
   }
   $$pre11 = HEAP32[$9>>2]|0;
   $95 = $$pre11;
  }
  $88 = HEAP32[$16>>2]|0;
  $89 = (($90) + ($88)|0);
  $91 = HEAP32[$17>>2]|0;
  $92 = (($93) + ($91)|0);
  $94 = (($95) + -1)|0;
  HEAP32[$9>>2] = $94;
  $96 = ($95|0)==(0);
  if ($96) {
   $$lcssa = $89;$$lcssa15 = $92;
   break;
  } else {
   $90 = $89;$93 = $92;$97 = $94;$98 = $92;$99 = $89;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa15;
 return;
}
function _SDL_Blit_RGB888_ARGB8888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$4 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre28 = 0, $$pre29 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$017 = 0, $dstA$0 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0;
 var $or$cond = 0, $posx$015 = 0, $posx$2 = 0, $posy$021 = 0, $posy$1$lcssa = 0, $src$016 = 0, $src$1 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$014 = 0, $srcx$2 = 0, $srcy$020 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = $3 >>> 24;
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $$4 = $29 ? 255 : $23;
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = ($$4|0)!=(255);
 $or$cond = $31 & $32;
 $33 = $1 & 112;
 $34 = $$4 ^ 255;
 $35 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $119 = $$pre;$124 = $21;$125 = $$pre;$44 = $19;$posy$021 = 0;$srcy$020 = 0;
 while(1) {
  $36 = ($posy$021|0)>(65535);
  if ($36) {
   $37 = (($posy$021) + -65536)|0;
   $38 = $37 >>> 16;
   $39 = $38 << 16;
   $40 = (($srcy$020) + 1)|0;
   $41 = (($37) - ($39))|0;
   $42 = (($40) + ($38))|0;
   $posy$1$lcssa = $41;$srcy$1$lcssa = $42;
  } else {
   $posy$1$lcssa = $posy$021;$srcy$1$lcssa = $srcy$020;
  }
  $43 = ($44|0)==(0);
  if ($43) {
   $121 = $124;
  } else {
   $$in = $44;$dst$017 = $125;$posx$015 = 65536;$src$016 = 0;$srcx$014 = -1;
   while(1) {
    $45 = (($$in) + -1)|0;
    $46 = ($posx$015|0)>(65535);
    if ($46) {
     $47 = (($posx$015) + -65536)|0;
     $48 = $47 >>> 16;
     $49 = $48 << 16;
     $50 = (($srcx$014) + 1)|0;
     $51 = (($47) - ($49))|0;
     $52 = (($50) + ($48))|0;
     $53 = HEAP32[$info>>2]|0;
     $54 = HEAP32[$35>>2]|0;
     $55 = Math_imul($54, $srcy$1$lcssa)|0;
     $56 = (($53) + ($55)|0);
     $57 = $52 << 2;
     $58 = (($56) + ($57)|0);
     $posx$2 = $51;$src$1 = $58;$srcx$2 = $52;
    } else {
     $posx$2 = $posx$015;$src$1 = $src$016;$srcx$2 = $srcx$014;
    }
    $59 = HEAP32[$src$1>>2]|0;
    $60 = $59 >>> 16;
    $61 = $60 & 255;
    $62 = $59 >>> 8;
    $63 = $62 & 255;
    $64 = $59 & 255;
    $65 = HEAP32[$dst$017>>2]|0;
    $66 = $65 >>> 16;
    $67 = $66 & 255;
    $68 = $65 >>> 8;
    $69 = $68 & 255;
    $70 = $65 & 255;
    $71 = $65 >>> 24;
    if ($27) {
     $srcB$0 = $64;$srcG$0 = $63;$srcR$0 = $61;
    } else {
     $72 = Math_imul($61, $4)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($63, $6)|0;
     $75 = (($74>>>0) / 255)&-1;
     $76 = Math_imul($64, $8)|0;
     $77 = (($76>>>0) / 255)&-1;
     $srcB$0 = $77;$srcG$0 = $75;$srcR$0 = $73;
    }
    if ($or$cond) {
     $78 = Math_imul($srcR$0, $$4)|0;
     $79 = (($78>>>0) / 255)&-1;
     $80 = Math_imul($srcG$0, $$4)|0;
     $81 = (($80>>>0) / 255)&-1;
     $82 = Math_imul($srcB$0, $$4)|0;
     $83 = (($82>>>0) / 255)&-1;
     $srcB$1 = $83;$srcG$1 = $81;$srcR$1 = $79;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($33|0) {
    case 16:  {
     $84 = Math_imul($67, $34)|0;
     $85 = (($84>>>0) / 255)&-1;
     $86 = (($srcR$1) + ($85))|0;
     $87 = Math_imul($69, $34)|0;
     $88 = (($87>>>0) / 255)&-1;
     $89 = (($srcG$1) + ($88))|0;
     $90 = Math_imul($70, $34)|0;
     $91 = (($90>>>0) / 255)&-1;
     $92 = (($srcB$1) + ($91))|0;
     $93 = Math_imul($71, $34)|0;
     $94 = (($93>>>0) / 255)&-1;
     $95 = (($94) + ($$4))|0;
     $dstA$0 = $95;$dstB$0 = $92;$dstG$0 = $89;$dstR$0 = $86;
     break;
    }
    case 32:  {
     $96 = (($srcR$1) + ($67))|0;
     $97 = ($96>>>0)>(255);
     $$ = $97 ? 255 : $96;
     $98 = (($srcG$1) + ($69))|0;
     $99 = ($98>>>0)>(255);
     $$3 = $99 ? 255 : $98;
     $100 = (($srcB$1) + ($70))|0;
     $101 = ($100>>>0)>(255);
     $$1 = $101 ? 255 : $100;
     $dstA$0 = $71;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $102 = Math_imul($srcR$1, $67)|0;
     $103 = (($102>>>0) / 255)&-1;
     $104 = Math_imul($srcG$1, $69)|0;
     $105 = (($104>>>0) / 255)&-1;
     $106 = Math_imul($srcB$1, $70)|0;
     $107 = (($106>>>0) / 255)&-1;
     $dstA$0 = $71;$dstB$0 = $107;$dstG$0 = $105;$dstR$0 = $103;
     break;
    }
    default: {
     $dstA$0 = $71;$dstB$0 = $70;$dstG$0 = $69;$dstR$0 = $67;
    }
    }
    $108 = $dstA$0 << 24;
    $109 = $dstR$0 << 16;
    $110 = $dstG$0 << 8;
    $111 = $110 | $109;
    $112 = $111 | $dstB$0;
    $113 = $112 | $108;
    HEAP32[$dst$017>>2] = $113;
    $114 = (($posx$2) + ($20))|0;
    $115 = ((($dst$017)) + 4|0);
    $116 = ($45|0)==(0);
    if ($116) {
     break;
    } else {
     $$in = $45;$dst$017 = $115;$posx$015 = $114;$src$016 = $src$1;$srcx$014 = $srcx$2;
    }
   }
   $$pre29 = HEAP32[$12>>2]|0;
   $121 = $$pre29;
  }
  $117 = HEAP32[$25>>2]|0;
  $118 = (($119) + ($117)|0);
  $120 = (($121) + -1)|0;
  HEAP32[$12>>2] = $120;
  $122 = ($121|0)==(0);
  if ($122) {
   $$lcssa = $118;
   break;
  }
  $123 = (($posy$1$lcssa) + ($14))|0;
  $$pre28 = HEAP32[$18>>2]|0;
  $119 = $118;$124 = $120;$125 = $118;$44 = $$pre28;$posy$021 = $123;$srcy$020 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGR888_RGB888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0, $posx$010 = 0, $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$53 = $$pre;$58 = $12;$59 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $55 = $58;
  } else {
   $$in = $25;$dst$013 = $59;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    $41 = $40 >>> 16;
    $42 = $41 & 255;
    $43 = $40 & 65280;
    $44 = $40 << 16;
    $45 = $44 & 16711680;
    $46 = $45 | $43;
    $47 = $46 | $42;
    HEAP32[$dst$013>>2] = $47;
    $48 = (($posx$2) + ($11))|0;
    $49 = ((($dst$013)) + 4|0);
    $50 = ($26|0)==(0);
    if ($50) {
     break;
    } else {
     $$in = $26;$dst$013 = $49;$posx$010 = $48;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $55 = $$pre25;
  }
  $51 = HEAP32[$15>>2]|0;
  $52 = (($53) + ($51)|0);
  $54 = (($55) + -1)|0;
  HEAP32[$3>>2] = $54;
  $56 = ($55|0)==(0);
  if ($56) {
   $$lcssa = $52;
   break;
  }
  $57 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$53 = $52;$58 = $54;$59 = $52;$posy$016 = $57;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGR888_RGB888_Blend($info) {
 $info = $info|0;
 var $$$us11 = 0, $$1$us13 = 0, $$3$us12 = 0, $$lcssa = 0, $$lcssa31 = 0, $$pre = 0, $$pre21 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $cond = 0;
 var $dst$05 = 0, $dst$05$us = 0, $dst$05$us9 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $src$04 = 0, $src$04$us = 0, $src$04$us10 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre21 = HEAP32[$6>>2]|0;
 $cond = ($10|0)==(64);
 $83 = $$pre;$86 = $$pre21;$90 = $$pre21;$91 = $$pre;
 while(1) {
  $11 = HEAP32[$7>>2]|0;
  $12 = (($11) + -1)|0;
  $13 = ($11|0)==(0);
  if (!($13)) {
   switch ($10|0) {
   case 16:  {
    $25 = $12;$dst$05$us = $90;$src$04$us = $91;
    while(1) {
     $14 = HEAP32[$src$04$us>>2]|0;
     $15 = $14 >>> 16;
     $16 = $15 & 255;
     $17 = $14 & 65280;
     $18 = $14 << 16;
     $19 = $18 & 16711680;
     $20 = $17 | $19;
     $21 = $20 | $16;
     HEAP32[$dst$05$us>>2] = $21;
     $22 = ((($src$04$us)) + 4|0);
     $23 = ((($dst$05$us)) + 4|0);
     $24 = (($25) + -1)|0;
     $26 = ($25|0)==(0);
     if ($26) {
      break;
     } else {
      $25 = $24;$dst$05$us = $23;$src$04$us = $22;
     }
    }
    break;
   }
   case 32:  {
    $52 = $12;$dst$05$us9 = $90;$src$04$us10 = $91;
    while(1) {
     $27 = HEAP32[$src$04$us10>>2]|0;
     $28 = $27 >>> 16;
     $29 = $28 & 255;
     $30 = $27 >>> 8;
     $31 = $30 & 255;
     $32 = $27 & 255;
     $33 = HEAP32[$dst$05$us9>>2]|0;
     $34 = $33 >>> 16;
     $35 = $34 & 255;
     $36 = $33 >>> 8;
     $37 = $36 & 255;
     $38 = $33 & 255;
     $39 = (($35) + ($32))|0;
     $40 = ($39>>>0)>(255);
     $$$us11 = $40 ? 255 : $39;
     $41 = (($37) + ($31))|0;
     $42 = ($41>>>0)>(255);
     $$3$us12 = $42 ? 255 : $41;
     $43 = (($29) + ($38))|0;
     $44 = ($43>>>0)>(255);
     $$1$us13 = $44 ? 255 : $43;
     $45 = $$$us11 << 16;
     $46 = $$3$us12 << 8;
     $47 = $46 | $45;
     $48 = $47 | $$1$us13;
     HEAP32[$dst$05$us9>>2] = $48;
     $49 = ((($src$04$us10)) + 4|0);
     $50 = ((($dst$05$us9)) + 4|0);
     $51 = (($52) + -1)|0;
     $53 = ($52|0)==(0);
     if ($53) {
      break;
     } else {
      $52 = $51;$dst$05$us9 = $50;$src$04$us10 = $49;
     }
    }
    break;
   }
   default: {
    $79 = $12;$dst$05 = $90;$src$04 = $91;
    while(1) {
     $54 = HEAP32[$src$04>>2]|0;
     $55 = HEAP32[$dst$05>>2]|0;
     $56 = $55 >>> 16;
     $57 = $56 & 255;
     $58 = $55 >>> 8;
     $59 = $58 & 255;
     $60 = $55 & 255;
     if ($cond) {
      $61 = $54 & 255;
      $62 = $54 >>> 8;
      $63 = $62 & 255;
      $64 = $54 >>> 16;
      $65 = $64 & 255;
      $66 = Math_imul($57, $61)|0;
      $67 = (($66>>>0) / 255)&-1;
      $68 = Math_imul($59, $63)|0;
      $69 = (($68>>>0) / 255)&-1;
      $70 = Math_imul($65, $60)|0;
      $71 = (($70>>>0) / 255)&-1;
      $dstB$0 = $71;$dstG$0 = $69;$dstR$0 = $67;
     } else {
      $dstB$0 = $60;$dstG$0 = $59;$dstR$0 = $57;
     }
     $72 = $dstR$0 << 16;
     $73 = $dstG$0 << 8;
     $74 = $73 | $72;
     $75 = $74 | $dstB$0;
     HEAP32[$dst$05>>2] = $75;
     $76 = ((($src$04)) + 4|0);
     $77 = ((($dst$05)) + 4|0);
     $78 = (($79) + -1)|0;
     $80 = ($79|0)==(0);
     if ($80) {
      break;
     } else {
      $79 = $78;$dst$05 = $77;$src$04 = $76;
     }
    }
   }
   }
  }
  $81 = HEAP32[$8>>2]|0;
  $82 = (($83) + ($81)|0);
  $84 = HEAP32[$9>>2]|0;
  $85 = (($86) + ($84)|0);
  $87 = HEAP32[$2>>2]|0;
  $88 = (($87) + -1)|0;
  HEAP32[$2>>2] = $88;
  $89 = ($87|0)==(0);
  if ($89) {
   $$lcssa = $82;$$lcssa31 = $85;
   break;
  } else {
   $83 = $82;$86 = $85;$90 = $85;$91 = $82;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa31;
 return;
}
function _SDL_Blit_BGR888_RGB888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$lcssa = 0, $$pre = 0, $$pre30 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dst$016$us = 0, $dstB$0 = 0;
 var $dstG$0 = 0, $dstR$0 = 0, $posx$014 = 0, $posx$014$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$020 = 0, $posy$1$lcssa = 0, $src$015 = 0, $src$015$us = 0, $src$1 = 0, $src$1$us = 0, $srcx$013 = 0, $srcx$013$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 112;
 $19 = ((($info)) + 12|0);
 $20 = ($18|0)==(16);
 $$pre = HEAP32[$16>>2]|0;
 $105 = $$pre;$110 = $$pre;$29 = $12;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $21 = ($posy$020|0)>(65535);
  if ($21) {
   $22 = (($posy$020) + -65536)|0;
   $23 = $22 >>> 16;
   $24 = $23 << 16;
   $25 = (($srcy$019) + 1)|0;
   $26 = (($22) - ($24))|0;
   $27 = (($25) + ($23))|0;
   $posy$1$lcssa = $26;$srcy$1$lcssa = $27;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $28 = (($29) + -1)|0;
  $30 = ($29|0)==(0);
  L9: do {
   if (!($30)) {
    if ($20) {
     $55 = $28;$dst$016$us = $110;$posx$014$us = 65536;$src$015$us = 0;$srcx$013$us = -1;
     while(1) {
      $31 = ($posx$014$us|0)>(65535);
      if ($31) {
       $32 = (($posx$014$us) + -65536)|0;
       $33 = $32 >>> 16;
       $34 = $33 << 16;
       $35 = (($srcx$013$us) + 1)|0;
       $36 = (($32) - ($34))|0;
       $37 = (($35) + ($33))|0;
       $38 = HEAP32[$info>>2]|0;
       $39 = HEAP32[$19>>2]|0;
       $40 = Math_imul($39, $srcy$1$lcssa)|0;
       $41 = (($38) + ($40)|0);
       $42 = $37 << 2;
       $43 = (($41) + ($42)|0);
       $posx$2$us = $36;$src$1$us = $43;$srcx$2$us = $37;
      } else {
       $posx$2$us = $posx$014$us;$src$1$us = $src$015$us;$srcx$2$us = $srcx$013$us;
      }
      $44 = HEAP32[$src$1$us>>2]|0;
      $45 = $44 >>> 16;
      $46 = $45 & 255;
      $47 = $44 & 65280;
      $48 = $44 << 16;
      $49 = $48 & 16711680;
      $50 = $47 | $49;
      $51 = $50 | $46;
      HEAP32[$dst$016$us>>2] = $51;
      $52 = (($posx$2$us) + ($13))|0;
      $53 = ((($dst$016$us)) + 4|0);
      $54 = (($55) + -1)|0;
      $56 = ($55|0)==(0);
      if ($56) {
       break L9;
      } else {
       $55 = $54;$dst$016$us = $53;$posx$014$us = $52;$src$015$us = $src$1$us;$srcx$013$us = $srcx$2$us;
      }
     }
    } else {
     $101 = $28;$dst$016 = $110;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
    }
    while(1) {
     $57 = ($posx$014|0)>(65535);
     if ($57) {
      $58 = (($posx$014) + -65536)|0;
      $59 = $58 >>> 16;
      $60 = $59 << 16;
      $61 = (($srcx$013) + 1)|0;
      $62 = (($58) - ($60))|0;
      $63 = (($61) + ($59))|0;
      $64 = HEAP32[$info>>2]|0;
      $65 = HEAP32[$19>>2]|0;
      $66 = Math_imul($65, $srcy$1$lcssa)|0;
      $67 = (($64) + ($66)|0);
      $68 = $63 << 2;
      $69 = (($67) + ($68)|0);
      $posx$2 = $62;$src$1 = $69;$srcx$2 = $63;
     } else {
      $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
     }
     $70 = HEAP32[$src$1>>2]|0;
     $71 = $70 >>> 16;
     $72 = $71 & 255;
     $73 = $70 >>> 8;
     $74 = $73 & 255;
     $75 = $70 & 255;
     $76 = HEAP32[$dst$016>>2]|0;
     $77 = $76 >>> 16;
     $78 = $77 & 255;
     $79 = $76 >>> 8;
     $80 = $79 & 255;
     $81 = $76 & 255;
     switch ($18|0) {
     case 64:  {
      $88 = Math_imul($78, $75)|0;
      $89 = (($88>>>0) / 255)&-1;
      $90 = Math_imul($80, $74)|0;
      $91 = (($90>>>0) / 255)&-1;
      $92 = Math_imul($72, $81)|0;
      $93 = (($92>>>0) / 255)&-1;
      $dstB$0 = $93;$dstG$0 = $91;$dstR$0 = $89;
      break;
     }
     case 32:  {
      $82 = (($78) + ($75))|0;
      $83 = ($82>>>0)>(255);
      $$ = $83 ? 255 : $82;
      $84 = (($80) + ($74))|0;
      $85 = ($84>>>0)>(255);
      $$3 = $85 ? 255 : $84;
      $86 = (($72) + ($81))|0;
      $87 = ($86>>>0)>(255);
      $$1 = $87 ? 255 : $86;
      $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
      break;
     }
     default: {
      $dstB$0 = $81;$dstG$0 = $80;$dstR$0 = $78;
     }
     }
     $94 = $dstR$0 << 16;
     $95 = $dstG$0 << 8;
     $96 = $95 | $94;
     $97 = $96 | $dstB$0;
     HEAP32[$dst$016>>2] = $97;
     $98 = (($posx$2) + ($13))|0;
     $99 = ((($dst$016)) + 4|0);
     $100 = (($101) + -1)|0;
     $102 = ($101|0)==(0);
     if ($102) {
      break L9;
     } else {
      $101 = $100;$dst$016 = $99;$posx$014 = $98;$src$015 = $src$1;$srcx$013 = $srcx$2;
     }
    }
   }
  } while(0);
  $103 = HEAP32[$17>>2]|0;
  $104 = (($105) + ($103)|0);
  $106 = HEAP32[$5>>2]|0;
  $107 = (($106) + -1)|0;
  HEAP32[$5>>2] = $107;
  $108 = ($106|0)==(0);
  if ($108) {
   $$lcssa = $104;
   break;
  }
  $109 = (($posy$1$lcssa) + ($7))|0;
  $$pre30 = HEAP32[$11>>2]|0;
  $105 = $104;$110 = $104;$29 = $$pre30;$posy$020 = $109;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGR888_RGB888_Modulate($info) {
 $info = $info|0;
 var $$in = 0, $$in13 = 0, $$lcssa = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$pre16 = 0, $$pre17 = 0, $$pre18 = 0, $$pre19 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $dst$02 = 0, $dst$02$us$us = 0, $src$01 = 0, $src$01$us$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = ((($info)) + 20|0);
 $14 = ((($info)) + 24|0);
 $15 = ((($info)) + 12|0);
 $16 = ((($info)) + 32|0);
 $17 = $1 & 1;
 $18 = ($17|0)==(0);
 $$pre17 = HEAP32[$info>>2]|0;
 $$pre18 = HEAP32[$13>>2]|0;
 if ($18) {
  $23 = $$pre17;$26 = $$pre18;$73 = $11;$74 = $$pre18;$75 = $$pre17;
  while(1) {
   $19 = HEAP32[$14>>2]|0;
   $20 = ($19|0)==(0);
   if ($20) {
    $28 = $73;
   } else {
    $$in13 = $19;$dst$02$us$us = $74;$src$01$us$us = $75;
    while(1) {
     $30 = (($$in13) + -1)|0;
     $31 = HEAP32[$src$01$us$us>>2]|0;
     $32 = $31 >>> 16;
     $33 = $32 & 255;
     $34 = $31 & 65280;
     $35 = $31 << 16;
     $36 = $35 & 16711680;
     $37 = $34 | $36;
     $38 = $37 | $33;
     HEAP32[$dst$02$us$us>>2] = $38;
     $39 = ((($src$01$us$us)) + 4|0);
     $40 = ((($dst$02$us$us)) + 4|0);
     $41 = ($30|0)==(0);
     if ($41) {
      break;
     } else {
      $$in13 = $30;$dst$02$us$us = $40;$src$01$us$us = $39;
     }
    }
    $$pre19 = HEAP32[$9>>2]|0;
    $28 = $$pre19;
   }
   $21 = HEAP32[$15>>2]|0;
   $22 = (($23) + ($21)|0);
   $24 = HEAP32[$16>>2]|0;
   $25 = (($26) + ($24)|0);
   $27 = (($28) + -1)|0;
   HEAP32[$9>>2] = $27;
   $29 = ($28|0)==(0);
   if ($29) {
    $$lcssa = $22;$$lcssa30 = $25;
    break;
   } else {
    $23 = $22;$26 = $25;$73 = $27;$74 = $25;$75 = $22;
   }
  }
  HEAP32[$info>>2] = $$lcssa;
  HEAP32[$13>>2] = $$lcssa30;
  return;
 } else {
  $66 = $$pre17;$69 = $$pre18;$76 = $11;$77 = $$pre18;$78 = $$pre17;
  while(1) {
   $42 = HEAP32[$14>>2]|0;
   $43 = ($42|0)==(0);
   if ($43) {
    $71 = $76;
   } else {
    $$in = $42;$dst$02 = $77;$src$01 = $78;
    while(1) {
     $44 = (($$in) + -1)|0;
     $45 = HEAP32[$src$01>>2]|0;
     $46 = $45 >>> 16;
     $47 = $46 & 255;
     $48 = $45 >>> 8;
     $49 = $48 & 255;
     $50 = $45 & 255;
     $51 = Math_imul($50, $4)|0;
     $52 = (($51>>>0) / 255)&-1;
     $53 = Math_imul($49, $6)|0;
     $54 = (($53>>>0) / 255)&-1;
     $55 = Math_imul($47, $8)|0;
     $56 = (($55>>>0) / 255)&-1;
     $57 = $52 << 16;
     $58 = $54 << 8;
     $59 = $58 | $57;
     $60 = $59 | $56;
     HEAP32[$dst$02>>2] = $60;
     $61 = ((($src$01)) + 4|0);
     $62 = ((($dst$02)) + 4|0);
     $63 = ($44|0)==(0);
     if ($63) {
      break;
     } else {
      $$in = $44;$dst$02 = $62;$src$01 = $61;
     }
    }
    $$pre16 = HEAP32[$9>>2]|0;
    $71 = $$pre16;
   }
   $64 = HEAP32[$15>>2]|0;
   $65 = (($66) + ($64)|0);
   $67 = HEAP32[$16>>2]|0;
   $68 = (($69) + ($67)|0);
   $70 = (($71) + -1)|0;
   HEAP32[$9>>2] = $70;
   $72 = ($71|0)==(0);
   if ($72) {
    $$lcssa31 = $65;$$lcssa32 = $68;
    break;
   } else {
    $66 = $65;$69 = $68;$76 = $70;$77 = $68;$78 = $65;
   }
  }
  HEAP32[$info>>2] = $$lcssa31;
  HEAP32[$13>>2] = $$lcssa32;
  return;
 }
}
function _SDL_Blit_BGR888_RGB888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$013 = 0, $dst$013$us = 0, $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$017 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0;
 var $src$1$us = 0, $srcx$010 = 0, $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$016 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = ((($info)) + 20|0);
 $24 = ((($info)) + 32|0);
 $25 = $1 & 1;
 $26 = ($25|0)==(0);
 $27 = ((($info)) + 12|0);
 $$pre = HEAP32[$23>>2]|0;
 $100 = $$pre;$105 = $$pre;$36 = $19;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $28 = ($posy$017|0)>(65535);
  if ($28) {
   $29 = (($posy$017) + -65536)|0;
   $30 = $29 >>> 16;
   $31 = $30 << 16;
   $32 = (($srcy$016) + 1)|0;
   $33 = (($29) - ($31))|0;
   $34 = (($32) + ($30))|0;
   $posy$1$lcssa = $33;$srcy$1$lcssa = $34;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $35 = (($36) + -1)|0;
  $37 = ($36|0)==(0);
  if (!($37)) {
   if ($26) {
    $62 = $35;$dst$013$us = $105;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $38 = ($posx$011$us|0)>(65535);
     if ($38) {
      $39 = (($posx$011$us) + -65536)|0;
      $40 = $39 >>> 16;
      $41 = $40 << 16;
      $42 = (($srcx$010$us) + 1)|0;
      $43 = (($39) - ($41))|0;
      $44 = (($42) + ($40))|0;
      $45 = HEAP32[$info>>2]|0;
      $46 = HEAP32[$27>>2]|0;
      $47 = Math_imul($46, $srcy$1$lcssa)|0;
      $48 = (($45) + ($47)|0);
      $49 = $44 << 2;
      $50 = (($48) + ($49)|0);
      $posx$2$us = $43;$src$1$us = $50;$srcx$2$us = $44;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $51 = HEAP32[$src$1$us>>2]|0;
     $52 = $51 >>> 16;
     $53 = $52 & 255;
     $54 = $51 & 65280;
     $55 = $51 << 16;
     $56 = $55 & 16711680;
     $57 = $53 | $54;
     $58 = $57 | $56;
     HEAP32[$dst$013$us>>2] = $58;
     $59 = (($posx$2$us) + ($20))|0;
     $60 = ((($dst$013$us)) + 4|0);
     $61 = (($62) + -1)|0;
     $63 = ($62|0)==(0);
     if ($63) {
      break;
     } else {
      $62 = $61;$dst$013$us = $60;$posx$011$us = $59;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $96 = $35;$dst$013 = $105;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $64 = ($posx$011|0)>(65535);
     if ($64) {
      $65 = (($posx$011) + -65536)|0;
      $66 = $65 >>> 16;
      $67 = $66 << 16;
      $68 = (($srcx$010) + 1)|0;
      $69 = (($65) - ($67))|0;
      $70 = (($68) + ($66))|0;
      $71 = HEAP32[$info>>2]|0;
      $72 = HEAP32[$27>>2]|0;
      $73 = Math_imul($72, $srcy$1$lcssa)|0;
      $74 = (($71) + ($73)|0);
      $75 = $70 << 2;
      $76 = (($74) + ($75)|0);
      $posx$2 = $69;$src$1 = $76;$srcx$2 = $70;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $77 = HEAP32[$src$1>>2]|0;
     $78 = $77 >>> 16;
     $79 = $78 & 255;
     $80 = $77 >>> 8;
     $81 = $80 & 255;
     $82 = $77 & 255;
     $83 = Math_imul($82, $4)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = Math_imul($81, $6)|0;
     $86 = (($85>>>0) / 255)&-1;
     $87 = Math_imul($79, $8)|0;
     $88 = (($87>>>0) / 255)&-1;
     $89 = $84 << 16;
     $90 = $86 << 8;
     $91 = $88 | $90;
     $92 = $91 | $89;
     HEAP32[$dst$013>>2] = $92;
     $93 = (($posx$2) + ($20))|0;
     $94 = ((($dst$013)) + 4|0);
     $95 = (($96) + -1)|0;
     $97 = ($96|0)==(0);
     if ($97) {
      break;
     } else {
      $96 = $95;$dst$013 = $94;$posx$011 = $93;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $98 = HEAP32[$24>>2]|0;
  $99 = (($100) + ($98)|0);
  $101 = HEAP32[$12>>2]|0;
  $102 = (($101) + -1)|0;
  HEAP32[$12>>2] = $102;
  $103 = ($101|0)==(0);
  if ($103) {
   $$lcssa = $99;
   break;
  }
  $104 = (($posy$1$lcssa) + ($14))|0;
  $$pre27 = HEAP32[$18>>2]|0;
  $100 = $99;$105 = $99;$36 = $$pre27;$posy$017 = $104;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$23>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGR888_RGB888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$$us = 0, $$1 = 0, $$1$us = 0, $$3 = 0, $$3$us = 0, $$4 = 0, $$lcssa = 0, $$lcssa18 = 0, $$pre = 0, $$pre11 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$06 = 0, $dst$06$us = 0, $dstB$0 = 0, $dstB$0$us = 0, $dstG$0 = 0, $dstG$0$us = 0;
 var $dstR$0 = 0, $dstR$0$us = 0, $or$cond = 0, $src$05 = 0, $src$05$us = 0, $srcB$1 = 0, $srcB$1$us = 0, $srcG$1 = 0, $srcG$1$us = 0, $srcR$1 = 0, $srcR$1$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = $3 >>> 24;
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $$4 = $21 ? 255 : $13;
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = ($$4|0)!=(255);
 $or$cond = $23 & $24;
 $25 = $1 & 112;
 $26 = $$4 ^ 255;
 $$pre = HEAP32[$info>>2]|0;
 $$pre11 = HEAP32[$14>>2]|0;
 $134 = $$pre;$137 = $$pre11;$141 = $$pre11;$142 = $$pre;
 while(1) {
  $27 = HEAP32[$15>>2]|0;
  $28 = (($27) + -1)|0;
  $29 = ($27|0)==(0);
  L6: do {
   if (!($29)) {
    if ($19) {
     $76 = $28;$dst$06$us = $141;$src$05$us = $142;
     while(1) {
      $30 = HEAP32[$src$05$us>>2]|0;
      $31 = $30 >>> 16;
      $32 = $31 & 255;
      $33 = $30 >>> 8;
      $34 = $33 & 255;
      $35 = $30 & 255;
      $36 = HEAP32[$dst$06$us>>2]|0;
      $37 = $36 >>> 16;
      $38 = $37 & 255;
      $39 = $36 >>> 8;
      $40 = $39 & 255;
      $41 = $36 & 255;
      if ($or$cond) {
       $42 = Math_imul($35, $$4)|0;
       $43 = (($42>>>0) / 255)&-1;
       $44 = Math_imul($34, $$4)|0;
       $45 = (($44>>>0) / 255)&-1;
       $46 = Math_imul($32, $$4)|0;
       $47 = (($46>>>0) / 255)&-1;
       $srcB$1$us = $47;$srcG$1$us = $45;$srcR$1$us = $43;
      } else {
       $srcB$1$us = $32;$srcG$1$us = $34;$srcR$1$us = $35;
      }
      switch ($25|0) {
      case 16:  {
       $60 = Math_imul($38, $26)|0;
       $61 = (($60>>>0) / 255)&-1;
       $62 = (($srcR$1$us) + ($61))|0;
       $63 = Math_imul($40, $26)|0;
       $64 = (($63>>>0) / 255)&-1;
       $65 = (($srcG$1$us) + ($64))|0;
       $66 = Math_imul($41, $26)|0;
       $67 = (($66>>>0) / 255)&-1;
       $68 = (($srcB$1$us) + ($67))|0;
       $dstB$0$us = $68;$dstG$0$us = $65;$dstR$0$us = $62;
       break;
      }
      case 32:  {
       $54 = (($srcR$1$us) + ($38))|0;
       $55 = ($54>>>0)>(255);
       $$$us = $55 ? 255 : $54;
       $56 = (($srcG$1$us) + ($40))|0;
       $57 = ($56>>>0)>(255);
       $$3$us = $57 ? 255 : $56;
       $58 = (($srcB$1$us) + ($41))|0;
       $59 = ($58>>>0)>(255);
       $$1$us = $59 ? 255 : $58;
       $dstB$0$us = $$1$us;$dstG$0$us = $$3$us;$dstR$0$us = $$$us;
       break;
      }
      case 64:  {
       $48 = Math_imul($srcR$1$us, $38)|0;
       $49 = (($48>>>0) / 255)&-1;
       $50 = Math_imul($srcG$1$us, $40)|0;
       $51 = (($50>>>0) / 255)&-1;
       $52 = Math_imul($srcB$1$us, $41)|0;
       $53 = (($52>>>0) / 255)&-1;
       $dstB$0$us = $53;$dstG$0$us = $51;$dstR$0$us = $49;
       break;
      }
      default: {
       $dstB$0$us = $41;$dstG$0$us = $40;$dstR$0$us = $38;
      }
      }
      $69 = $dstR$0$us << 16;
      $70 = $dstG$0$us << 8;
      $71 = $70 | $69;
      $72 = $71 | $dstB$0$us;
      HEAP32[$dst$06$us>>2] = $72;
      $73 = ((($src$05$us)) + 4|0);
      $74 = ((($dst$06$us)) + 4|0);
      $75 = (($76) + -1)|0;
      $77 = ($76|0)==(0);
      if ($77) {
       break L6;
      } else {
       $76 = $75;$dst$06$us = $74;$src$05$us = $73;
      }
     }
    } else {
     $130 = $28;$dst$06 = $141;$src$05 = $142;
     while(1) {
      $78 = HEAP32[$src$05>>2]|0;
      $79 = $78 >>> 16;
      $80 = $79 & 255;
      $81 = $78 >>> 8;
      $82 = $81 & 255;
      $83 = $78 & 255;
      $84 = HEAP32[$dst$06>>2]|0;
      $85 = $84 >>> 16;
      $86 = $85 & 255;
      $87 = $84 >>> 8;
      $88 = $87 & 255;
      $89 = $84 & 255;
      $90 = Math_imul($83, $4)|0;
      $91 = (($90>>>0) / 255)&-1;
      $92 = Math_imul($82, $6)|0;
      $93 = (($92>>>0) / 255)&-1;
      $94 = Math_imul($80, $8)|0;
      $95 = (($94>>>0) / 255)&-1;
      if ($or$cond) {
       $96 = Math_imul($91, $$4)|0;
       $97 = (($96>>>0) / 255)&-1;
       $98 = Math_imul($93, $$4)|0;
       $99 = (($98>>>0) / 255)&-1;
       $100 = Math_imul($95, $$4)|0;
       $101 = (($100>>>0) / 255)&-1;
       $srcB$1 = $101;$srcG$1 = $99;$srcR$1 = $97;
      } else {
       $srcB$1 = $95;$srcG$1 = $93;$srcR$1 = $91;
      }
      switch ($25|0) {
      case 16:  {
       $102 = Math_imul($86, $26)|0;
       $103 = (($102>>>0) / 255)&-1;
       $104 = (($srcR$1) + ($103))|0;
       $105 = Math_imul($88, $26)|0;
       $106 = (($105>>>0) / 255)&-1;
       $107 = (($srcG$1) + ($106))|0;
       $108 = Math_imul($89, $26)|0;
       $109 = (($108>>>0) / 255)&-1;
       $110 = (($srcB$1) + ($109))|0;
       $dstB$0 = $110;$dstG$0 = $107;$dstR$0 = $104;
       break;
      }
      case 32:  {
       $111 = (($srcR$1) + ($86))|0;
       $112 = ($111>>>0)>(255);
       $$ = $112 ? 255 : $111;
       $113 = (($srcG$1) + ($88))|0;
       $114 = ($113>>>0)>(255);
       $$3 = $114 ? 255 : $113;
       $115 = (($srcB$1) + ($89))|0;
       $116 = ($115>>>0)>(255);
       $$1 = $116 ? 255 : $115;
       $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
       break;
      }
      case 64:  {
       $117 = Math_imul($srcR$1, $86)|0;
       $118 = (($117>>>0) / 255)&-1;
       $119 = Math_imul($srcG$1, $88)|0;
       $120 = (($119>>>0) / 255)&-1;
       $121 = Math_imul($srcB$1, $89)|0;
       $122 = (($121>>>0) / 255)&-1;
       $dstB$0 = $122;$dstG$0 = $120;$dstR$0 = $118;
       break;
      }
      default: {
       $dstB$0 = $89;$dstG$0 = $88;$dstR$0 = $86;
      }
      }
      $123 = $dstR$0 << 16;
      $124 = $dstG$0 << 8;
      $125 = $124 | $123;
      $126 = $125 | $dstB$0;
      HEAP32[$dst$06>>2] = $126;
      $127 = ((($src$05)) + 4|0);
      $128 = ((($dst$06)) + 4|0);
      $129 = (($130) + -1)|0;
      $131 = ($130|0)==(0);
      if ($131) {
       break L6;
      } else {
       $130 = $129;$dst$06 = $128;$src$05 = $127;
      }
     }
    }
   }
  } while(0);
  $132 = HEAP32[$16>>2]|0;
  $133 = (($134) + ($132)|0);
  $135 = HEAP32[$17>>2]|0;
  $136 = (($137) + ($135)|0);
  $138 = HEAP32[$9>>2]|0;
  $139 = (($138) + -1)|0;
  HEAP32[$9>>2] = $139;
  $140 = ($138|0)==(0);
  if ($140) {
   $$lcssa = $133;$$lcssa18 = $136;
   break;
  } else {
   $134 = $133;$137 = $136;$141 = $136;$142 = $133;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa18;
 return;
}
function _SDL_Blit_BGR888_RGB888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$4 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre28 = 0, $$pre29 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$017 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$015 = 0, $posx$2 = 0, $posy$021 = 0, $posy$1$lcssa = 0, $src$016 = 0, $src$1 = 0;
 var $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$014 = 0, $srcx$2 = 0, $srcy$020 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = $3 >>> 24;
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $$4 = $29 ? 255 : $23;
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = ($$4|0)!=(255);
 $or$cond = $31 & $32;
 $33 = $1 & 112;
 $34 = $$4 ^ 255;
 $35 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $113 = $$pre;$118 = $21;$119 = $$pre;$44 = $19;$posy$021 = 0;$srcy$020 = 0;
 while(1) {
  $36 = ($posy$021|0)>(65535);
  if ($36) {
   $37 = (($posy$021) + -65536)|0;
   $38 = $37 >>> 16;
   $39 = $38 << 16;
   $40 = (($srcy$020) + 1)|0;
   $41 = (($37) - ($39))|0;
   $42 = (($40) + ($38))|0;
   $posy$1$lcssa = $41;$srcy$1$lcssa = $42;
  } else {
   $posy$1$lcssa = $posy$021;$srcy$1$lcssa = $srcy$020;
  }
  $43 = ($44|0)==(0);
  if ($43) {
   $115 = $118;
  } else {
   $$in = $44;$dst$017 = $119;$posx$015 = 65536;$src$016 = 0;$srcx$014 = -1;
   while(1) {
    $45 = (($$in) + -1)|0;
    $46 = ($posx$015|0)>(65535);
    if ($46) {
     $47 = (($posx$015) + -65536)|0;
     $48 = $47 >>> 16;
     $49 = $48 << 16;
     $50 = (($srcx$014) + 1)|0;
     $51 = (($47) - ($49))|0;
     $52 = (($50) + ($48))|0;
     $53 = HEAP32[$info>>2]|0;
     $54 = HEAP32[$35>>2]|0;
     $55 = Math_imul($54, $srcy$1$lcssa)|0;
     $56 = (($53) + ($55)|0);
     $57 = $52 << 2;
     $58 = (($56) + ($57)|0);
     $posx$2 = $51;$src$1 = $58;$srcx$2 = $52;
    } else {
     $posx$2 = $posx$015;$src$1 = $src$016;$srcx$2 = $srcx$014;
    }
    $59 = HEAP32[$src$1>>2]|0;
    $60 = $59 >>> 16;
    $61 = $60 & 255;
    $62 = $59 >>> 8;
    $63 = $62 & 255;
    $64 = $59 & 255;
    $65 = HEAP32[$dst$017>>2]|0;
    $66 = $65 >>> 16;
    $67 = $66 & 255;
    $68 = $65 >>> 8;
    $69 = $68 & 255;
    $70 = $65 & 255;
    if ($27) {
     $srcB$0 = $61;$srcG$0 = $63;$srcR$0 = $64;
    } else {
     $71 = Math_imul($64, $4)|0;
     $72 = (($71>>>0) / 255)&-1;
     $73 = Math_imul($63, $6)|0;
     $74 = (($73>>>0) / 255)&-1;
     $75 = Math_imul($61, $8)|0;
     $76 = (($75>>>0) / 255)&-1;
     $srcB$0 = $76;$srcG$0 = $74;$srcR$0 = $72;
    }
    if ($or$cond) {
     $77 = Math_imul($srcR$0, $$4)|0;
     $78 = (($77>>>0) / 255)&-1;
     $79 = Math_imul($srcG$0, $$4)|0;
     $80 = (($79>>>0) / 255)&-1;
     $81 = Math_imul($srcB$0, $$4)|0;
     $82 = (($81>>>0) / 255)&-1;
     $srcB$1 = $82;$srcG$1 = $80;$srcR$1 = $78;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($33|0) {
    case 16:  {
     $83 = Math_imul($67, $34)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = (($srcR$1) + ($84))|0;
     $86 = Math_imul($69, $34)|0;
     $87 = (($86>>>0) / 255)&-1;
     $88 = (($srcG$1) + ($87))|0;
     $89 = Math_imul($70, $34)|0;
     $90 = (($89>>>0) / 255)&-1;
     $91 = (($srcB$1) + ($90))|0;
     $dstB$0 = $91;$dstG$0 = $88;$dstR$0 = $85;
     break;
    }
    case 32:  {
     $92 = (($srcR$1) + ($67))|0;
     $93 = ($92>>>0)>(255);
     $$ = $93 ? 255 : $92;
     $94 = (($srcG$1) + ($69))|0;
     $95 = ($94>>>0)>(255);
     $$3 = $95 ? 255 : $94;
     $96 = (($srcB$1) + ($70))|0;
     $97 = ($96>>>0)>(255);
     $$1 = $97 ? 255 : $96;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $98 = Math_imul($srcR$1, $67)|0;
     $99 = (($98>>>0) / 255)&-1;
     $100 = Math_imul($srcG$1, $69)|0;
     $101 = (($100>>>0) / 255)&-1;
     $102 = Math_imul($srcB$1, $70)|0;
     $103 = (($102>>>0) / 255)&-1;
     $dstB$0 = $103;$dstG$0 = $101;$dstR$0 = $99;
     break;
    }
    default: {
     $dstB$0 = $70;$dstG$0 = $69;$dstR$0 = $67;
    }
    }
    $104 = $dstR$0 << 16;
    $105 = $dstG$0 << 8;
    $106 = $105 | $104;
    $107 = $106 | $dstB$0;
    HEAP32[$dst$017>>2] = $107;
    $108 = (($posx$2) + ($20))|0;
    $109 = ((($dst$017)) + 4|0);
    $110 = ($45|0)==(0);
    if ($110) {
     break;
    } else {
     $$in = $45;$dst$017 = $109;$posx$015 = $108;$src$016 = $src$1;$srcx$014 = $srcx$2;
    }
   }
   $$pre29 = HEAP32[$12>>2]|0;
   $115 = $$pre29;
  }
  $111 = HEAP32[$25>>2]|0;
  $112 = (($113) + ($111)|0);
  $114 = (($115) + -1)|0;
  HEAP32[$12>>2] = $114;
  $116 = ($115|0)==(0);
  if ($116) {
   $$lcssa = $112;
   break;
  }
  $117 = (($posy$1$lcssa) + ($14))|0;
  $$pre28 = HEAP32[$18>>2]|0;
  $113 = $112;$118 = $114;$119 = $112;$44 = $$pre28;$posy$021 = $117;$srcy$020 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGR888_BGR888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0, $posx$010 = 0;
 var $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$46 = $$pre;$51 = $12;$52 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $48 = $51;
  } else {
   $$in = $25;$dst$013 = $52;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    HEAP32[$dst$013>>2] = $40;
    $41 = (($posx$2) + ($11))|0;
    $42 = ((($dst$013)) + 4|0);
    $43 = ($26|0)==(0);
    if ($43) {
     break;
    } else {
     $$in = $26;$dst$013 = $42;$posx$010 = $41;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $48 = $$pre25;
  }
  $44 = HEAP32[$15>>2]|0;
  $45 = (($46) + ($44)|0);
  $47 = (($48) + -1)|0;
  HEAP32[$3>>2] = $47;
  $49 = ($48|0)==(0);
  if ($49) {
   $$lcssa = $45;
   break;
  }
  $50 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$46 = $45;$51 = $47;$52 = $45;$posy$016 = $50;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGR888_BGR888_Blend($info) {
 $info = $info|0;
 var $$$us11 = 0, $$1$us13 = 0, $$3$us12 = 0, $$lcssa = 0, $$lcssa31 = 0, $$pre = 0, $$pre21 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $9 = 0, $cond = 0, $dst$05 = 0, $dst$05$us = 0, $dst$05$us9 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0;
 var $src$04 = 0, $src$04$us = 0, $src$04$us10 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre21 = HEAP32[$6>>2]|0;
 $cond = ($10|0)==(64);
 $77 = $$pre;$80 = $$pre21;$84 = $$pre21;$85 = $$pre;
 while(1) {
  $11 = HEAP32[$7>>2]|0;
  $12 = (($11) + -1)|0;
  $13 = ($11|0)==(0);
  if (!($13)) {
   switch ($10|0) {
   case 16:  {
    $19 = $12;$dst$05$us = $84;$src$04$us = $85;
    while(1) {
     $14 = HEAP32[$src$04$us>>2]|0;
     $15 = $14 & 16777215;
     HEAP32[$dst$05$us>>2] = $15;
     $16 = ((($src$04$us)) + 4|0);
     $17 = ((($dst$05$us)) + 4|0);
     $18 = (($19) + -1)|0;
     $20 = ($19|0)==(0);
     if ($20) {
      break;
     } else {
      $19 = $18;$dst$05$us = $17;$src$04$us = $16;
     }
    }
    break;
   }
   case 32:  {
    $46 = $12;$dst$05$us9 = $84;$src$04$us10 = $85;
    while(1) {
     $21 = HEAP32[$src$04$us10>>2]|0;
     $22 = $21 >>> 16;
     $23 = $22 & 255;
     $24 = $21 >>> 8;
     $25 = $24 & 255;
     $26 = $21 & 255;
     $27 = HEAP32[$dst$05$us9>>2]|0;
     $28 = $27 >>> 16;
     $29 = $28 & 255;
     $30 = $27 >>> 8;
     $31 = $30 & 255;
     $32 = $27 & 255;
     $33 = (($32) + ($26))|0;
     $34 = ($33>>>0)>(255);
     $$$us11 = $34 ? 255 : $33;
     $35 = (($31) + ($25))|0;
     $36 = ($35>>>0)>(255);
     $$3$us12 = $36 ? 255 : $35;
     $37 = (($29) + ($23))|0;
     $38 = ($37>>>0)>(255);
     $$1$us13 = $38 ? 255 : $37;
     $39 = $$1$us13 << 16;
     $40 = $$3$us12 << 8;
     $41 = $40 | $$$us11;
     $42 = $41 | $39;
     HEAP32[$dst$05$us9>>2] = $42;
     $43 = ((($src$04$us10)) + 4|0);
     $44 = ((($dst$05$us9)) + 4|0);
     $45 = (($46) + -1)|0;
     $47 = ($46|0)==(0);
     if ($47) {
      break;
     } else {
      $46 = $45;$dst$05$us9 = $44;$src$04$us10 = $43;
     }
    }
    break;
   }
   default: {
    $73 = $12;$dst$05 = $84;$src$04 = $85;
    while(1) {
     $48 = HEAP32[$src$04>>2]|0;
     $49 = HEAP32[$dst$05>>2]|0;
     $50 = $49 >>> 16;
     $51 = $50 & 255;
     $52 = $49 >>> 8;
     $53 = $52 & 255;
     $54 = $49 & 255;
     if ($cond) {
      $55 = $48 & 255;
      $56 = $48 >>> 8;
      $57 = $56 & 255;
      $58 = $48 >>> 16;
      $59 = $58 & 255;
      $60 = Math_imul($54, $55)|0;
      $61 = (($60>>>0) / 255)&-1;
      $62 = Math_imul($53, $57)|0;
      $63 = (($62>>>0) / 255)&-1;
      $64 = Math_imul($51, $59)|0;
      $65 = (($64>>>0) / 255)&-1;
      $dstB$0 = $65;$dstG$0 = $63;$dstR$0 = $61;
     } else {
      $dstB$0 = $51;$dstG$0 = $53;$dstR$0 = $54;
     }
     $66 = $dstB$0 << 16;
     $67 = $dstG$0 << 8;
     $68 = $67 | $dstR$0;
     $69 = $68 | $66;
     HEAP32[$dst$05>>2] = $69;
     $70 = ((($src$04)) + 4|0);
     $71 = ((($dst$05)) + 4|0);
     $72 = (($73) + -1)|0;
     $74 = ($73|0)==(0);
     if ($74) {
      break;
     } else {
      $73 = $72;$dst$05 = $71;$src$04 = $70;
     }
    }
   }
   }
  }
  $75 = HEAP32[$8>>2]|0;
  $76 = (($77) + ($75)|0);
  $78 = HEAP32[$9>>2]|0;
  $79 = (($80) + ($78)|0);
  $81 = HEAP32[$2>>2]|0;
  $82 = (($81) + -1)|0;
  HEAP32[$2>>2] = $82;
  $83 = ($81|0)==(0);
  if ($83) {
   $$lcssa = $76;$$lcssa31 = $79;
   break;
  } else {
   $77 = $76;$80 = $79;$84 = $79;$85 = $76;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa31;
 return;
}
function _SDL_Blit_BGR888_BGR888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$lcssa = 0, $$pre = 0, $$pre30 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dst$016$us = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $posx$014 = 0, $posx$014$us = 0, $posx$2 = 0, $posx$2$us = 0;
 var $posy$020 = 0, $posy$1$lcssa = 0, $src$015 = 0, $src$015$us = 0, $src$1 = 0, $src$1$us = 0, $srcx$013 = 0, $srcx$013$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 112;
 $19 = ((($info)) + 12|0);
 $20 = ($18|0)==(16);
 $$pre = HEAP32[$16>>2]|0;
 $104 = $$pre;$29 = $12;$99 = $$pre;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $21 = ($posy$020|0)>(65535);
  if ($21) {
   $22 = (($posy$020) + -65536)|0;
   $23 = $22 >>> 16;
   $24 = $23 << 16;
   $25 = (($srcy$019) + 1)|0;
   $26 = (($22) - ($24))|0;
   $27 = (($25) + ($23))|0;
   $posy$1$lcssa = $26;$srcy$1$lcssa = $27;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $28 = (($29) + -1)|0;
  $30 = ($29|0)==(0);
  L9: do {
   if (!($30)) {
    if ($20) {
     $49 = $28;$dst$016$us = $104;$posx$014$us = 65536;$src$015$us = 0;$srcx$013$us = -1;
     while(1) {
      $31 = ($posx$014$us|0)>(65535);
      if ($31) {
       $32 = (($posx$014$us) + -65536)|0;
       $33 = $32 >>> 16;
       $34 = $33 << 16;
       $35 = (($srcx$013$us) + 1)|0;
       $36 = (($32) - ($34))|0;
       $37 = (($35) + ($33))|0;
       $38 = HEAP32[$info>>2]|0;
       $39 = HEAP32[$19>>2]|0;
       $40 = Math_imul($39, $srcy$1$lcssa)|0;
       $41 = (($38) + ($40)|0);
       $42 = $37 << 2;
       $43 = (($41) + ($42)|0);
       $posx$2$us = $36;$src$1$us = $43;$srcx$2$us = $37;
      } else {
       $posx$2$us = $posx$014$us;$src$1$us = $src$015$us;$srcx$2$us = $srcx$013$us;
      }
      $44 = HEAP32[$src$1$us>>2]|0;
      $45 = $44 & 16777215;
      HEAP32[$dst$016$us>>2] = $45;
      $46 = (($posx$2$us) + ($13))|0;
      $47 = ((($dst$016$us)) + 4|0);
      $48 = (($49) + -1)|0;
      $50 = ($49|0)==(0);
      if ($50) {
       break L9;
      } else {
       $49 = $48;$dst$016$us = $47;$posx$014$us = $46;$src$015$us = $src$1$us;$srcx$013$us = $srcx$2$us;
      }
     }
    } else {
     $95 = $28;$dst$016 = $104;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
    }
    while(1) {
     $51 = ($posx$014|0)>(65535);
     if ($51) {
      $52 = (($posx$014) + -65536)|0;
      $53 = $52 >>> 16;
      $54 = $53 << 16;
      $55 = (($srcx$013) + 1)|0;
      $56 = (($52) - ($54))|0;
      $57 = (($55) + ($53))|0;
      $58 = HEAP32[$info>>2]|0;
      $59 = HEAP32[$19>>2]|0;
      $60 = Math_imul($59, $srcy$1$lcssa)|0;
      $61 = (($58) + ($60)|0);
      $62 = $57 << 2;
      $63 = (($61) + ($62)|0);
      $posx$2 = $56;$src$1 = $63;$srcx$2 = $57;
     } else {
      $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
     }
     $64 = HEAP32[$src$1>>2]|0;
     $65 = $64 >>> 16;
     $66 = $65 & 255;
     $67 = $64 >>> 8;
     $68 = $67 & 255;
     $69 = $64 & 255;
     $70 = HEAP32[$dst$016>>2]|0;
     $71 = $70 >>> 16;
     $72 = $71 & 255;
     $73 = $70 >>> 8;
     $74 = $73 & 255;
     $75 = $70 & 255;
     switch ($18|0) {
     case 64:  {
      $82 = Math_imul($75, $69)|0;
      $83 = (($82>>>0) / 255)&-1;
      $84 = Math_imul($74, $68)|0;
      $85 = (($84>>>0) / 255)&-1;
      $86 = Math_imul($72, $66)|0;
      $87 = (($86>>>0) / 255)&-1;
      $dstB$0 = $87;$dstG$0 = $85;$dstR$0 = $83;
      break;
     }
     case 32:  {
      $76 = (($75) + ($69))|0;
      $77 = ($76>>>0)>(255);
      $$ = $77 ? 255 : $76;
      $78 = (($74) + ($68))|0;
      $79 = ($78>>>0)>(255);
      $$3 = $79 ? 255 : $78;
      $80 = (($72) + ($66))|0;
      $81 = ($80>>>0)>(255);
      $$1 = $81 ? 255 : $80;
      $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
      break;
     }
     default: {
      $dstB$0 = $72;$dstG$0 = $74;$dstR$0 = $75;
     }
     }
     $88 = $dstB$0 << 16;
     $89 = $dstG$0 << 8;
     $90 = $89 | $dstR$0;
     $91 = $90 | $88;
     HEAP32[$dst$016>>2] = $91;
     $92 = (($posx$2) + ($13))|0;
     $93 = ((($dst$016)) + 4|0);
     $94 = (($95) + -1)|0;
     $96 = ($95|0)==(0);
     if ($96) {
      break L9;
     } else {
      $95 = $94;$dst$016 = $93;$posx$014 = $92;$src$015 = $src$1;$srcx$013 = $srcx$2;
     }
    }
   }
  } while(0);
  $97 = HEAP32[$17>>2]|0;
  $98 = (($99) + ($97)|0);
  $100 = HEAP32[$5>>2]|0;
  $101 = (($100) + -1)|0;
  HEAP32[$5>>2] = $101;
  $102 = ($100|0)==(0);
  if ($102) {
   $$lcssa = $98;
   break;
  }
  $103 = (($posy$1$lcssa) + ($7))|0;
  $$pre30 = HEAP32[$11>>2]|0;
  $104 = $98;$29 = $$pre30;$99 = $98;$posy$020 = $103;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGR888_BGR888_Modulate($info) {
 $info = $info|0;
 var $$in = 0, $$in13 = 0, $$lcssa = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$pre16 = 0, $$pre17 = 0, $$pre18 = 0, $$pre19 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $8 = 0, $9 = 0, $dst$02 = 0, $dst$02$us$us = 0, $src$01 = 0, $src$01$us$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = ((($info)) + 20|0);
 $14 = ((($info)) + 24|0);
 $15 = ((($info)) + 12|0);
 $16 = ((($info)) + 32|0);
 $17 = $1 & 1;
 $18 = ($17|0)==(0);
 $$pre17 = HEAP32[$info>>2]|0;
 $$pre18 = HEAP32[$13>>2]|0;
 if ($18) {
  $23 = $$pre17;$26 = $$pre18;$67 = $11;$68 = $$pre18;$69 = $$pre17;
  while(1) {
   $19 = HEAP32[$14>>2]|0;
   $20 = ($19|0)==(0);
   if ($20) {
    $28 = $67;
   } else {
    $$in13 = $19;$dst$02$us$us = $68;$src$01$us$us = $69;
    while(1) {
     $30 = (($$in13) + -1)|0;
     $31 = HEAP32[$src$01$us$us>>2]|0;
     $32 = $31 & 16777215;
     HEAP32[$dst$02$us$us>>2] = $32;
     $33 = ((($src$01$us$us)) + 4|0);
     $34 = ((($dst$02$us$us)) + 4|0);
     $35 = ($30|0)==(0);
     if ($35) {
      break;
     } else {
      $$in13 = $30;$dst$02$us$us = $34;$src$01$us$us = $33;
     }
    }
    $$pre19 = HEAP32[$9>>2]|0;
    $28 = $$pre19;
   }
   $21 = HEAP32[$15>>2]|0;
   $22 = (($23) + ($21)|0);
   $24 = HEAP32[$16>>2]|0;
   $25 = (($26) + ($24)|0);
   $27 = (($28) + -1)|0;
   HEAP32[$9>>2] = $27;
   $29 = ($28|0)==(0);
   if ($29) {
    $$lcssa = $22;$$lcssa30 = $25;
    break;
   } else {
    $23 = $22;$26 = $25;$67 = $27;$68 = $25;$69 = $22;
   }
  }
  HEAP32[$info>>2] = $$lcssa;
  HEAP32[$13>>2] = $$lcssa30;
  return;
 } else {
  $60 = $$pre17;$63 = $$pre18;$70 = $11;$71 = $$pre18;$72 = $$pre17;
  while(1) {
   $36 = HEAP32[$14>>2]|0;
   $37 = ($36|0)==(0);
   if ($37) {
    $65 = $70;
   } else {
    $$in = $36;$dst$02 = $71;$src$01 = $72;
    while(1) {
     $38 = (($$in) + -1)|0;
     $39 = HEAP32[$src$01>>2]|0;
     $40 = $39 >>> 16;
     $41 = $40 & 255;
     $42 = $39 >>> 8;
     $43 = $42 & 255;
     $44 = $39 & 255;
     $45 = Math_imul($44, $4)|0;
     $46 = (($45>>>0) / 255)&-1;
     $47 = Math_imul($43, $6)|0;
     $48 = (($47>>>0) / 255)&-1;
     $49 = Math_imul($41, $8)|0;
     $50 = (($49>>>0) / 255)&-1;
     $51 = $50 << 16;
     $52 = $48 << 8;
     $53 = $52 | $46;
     $54 = $53 | $51;
     HEAP32[$dst$02>>2] = $54;
     $55 = ((($src$01)) + 4|0);
     $56 = ((($dst$02)) + 4|0);
     $57 = ($38|0)==(0);
     if ($57) {
      break;
     } else {
      $$in = $38;$dst$02 = $56;$src$01 = $55;
     }
    }
    $$pre16 = HEAP32[$9>>2]|0;
    $65 = $$pre16;
   }
   $58 = HEAP32[$15>>2]|0;
   $59 = (($60) + ($58)|0);
   $61 = HEAP32[$16>>2]|0;
   $62 = (($63) + ($61)|0);
   $64 = (($65) + -1)|0;
   HEAP32[$9>>2] = $64;
   $66 = ($65|0)==(0);
   if ($66) {
    $$lcssa31 = $59;$$lcssa32 = $62;
    break;
   } else {
    $60 = $59;$63 = $62;$70 = $64;$71 = $62;$72 = $59;
   }
  }
  HEAP32[$info>>2] = $$lcssa31;
  HEAP32[$13>>2] = $$lcssa32;
  return;
 }
}
function _SDL_Blit_BGR888_BGR888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $dst$013 = 0, $dst$013$us = 0, $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$017 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0, $src$1$us = 0, $srcx$010 = 0, $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$016 = 0;
 var $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = ((($info)) + 20|0);
 $24 = ((($info)) + 32|0);
 $25 = $1 & 1;
 $26 = ($25|0)==(0);
 $27 = ((($info)) + 12|0);
 $$pre = HEAP32[$23>>2]|0;
 $36 = $19;$94 = $$pre;$99 = $$pre;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $28 = ($posy$017|0)>(65535);
  if ($28) {
   $29 = (($posy$017) + -65536)|0;
   $30 = $29 >>> 16;
   $31 = $30 << 16;
   $32 = (($srcy$016) + 1)|0;
   $33 = (($29) - ($31))|0;
   $34 = (($32) + ($30))|0;
   $posy$1$lcssa = $33;$srcy$1$lcssa = $34;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $35 = (($36) + -1)|0;
  $37 = ($36|0)==(0);
  if (!($37)) {
   if ($26) {
    $56 = $35;$dst$013$us = $99;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $38 = ($posx$011$us|0)>(65535);
     if ($38) {
      $39 = (($posx$011$us) + -65536)|0;
      $40 = $39 >>> 16;
      $41 = $40 << 16;
      $42 = (($srcx$010$us) + 1)|0;
      $43 = (($39) - ($41))|0;
      $44 = (($42) + ($40))|0;
      $45 = HEAP32[$info>>2]|0;
      $46 = HEAP32[$27>>2]|0;
      $47 = Math_imul($46, $srcy$1$lcssa)|0;
      $48 = (($45) + ($47)|0);
      $49 = $44 << 2;
      $50 = (($48) + ($49)|0);
      $posx$2$us = $43;$src$1$us = $50;$srcx$2$us = $44;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $51 = HEAP32[$src$1$us>>2]|0;
     $52 = $51 & 16777215;
     HEAP32[$dst$013$us>>2] = $52;
     $53 = (($posx$2$us) + ($20))|0;
     $54 = ((($dst$013$us)) + 4|0);
     $55 = (($56) + -1)|0;
     $57 = ($56|0)==(0);
     if ($57) {
      break;
     } else {
      $56 = $55;$dst$013$us = $54;$posx$011$us = $53;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $90 = $35;$dst$013 = $99;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $58 = ($posx$011|0)>(65535);
     if ($58) {
      $59 = (($posx$011) + -65536)|0;
      $60 = $59 >>> 16;
      $61 = $60 << 16;
      $62 = (($srcx$010) + 1)|0;
      $63 = (($59) - ($61))|0;
      $64 = (($62) + ($60))|0;
      $65 = HEAP32[$info>>2]|0;
      $66 = HEAP32[$27>>2]|0;
      $67 = Math_imul($66, $srcy$1$lcssa)|0;
      $68 = (($65) + ($67)|0);
      $69 = $64 << 2;
      $70 = (($68) + ($69)|0);
      $posx$2 = $63;$src$1 = $70;$srcx$2 = $64;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $71 = HEAP32[$src$1>>2]|0;
     $72 = $71 >>> 16;
     $73 = $72 & 255;
     $74 = $71 >>> 8;
     $75 = $74 & 255;
     $76 = $71 & 255;
     $77 = Math_imul($76, $4)|0;
     $78 = (($77>>>0) / 255)&-1;
     $79 = Math_imul($75, $6)|0;
     $80 = (($79>>>0) / 255)&-1;
     $81 = Math_imul($73, $8)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = $82 << 16;
     $84 = $80 << 8;
     $85 = $83 | $84;
     $86 = $85 | $78;
     HEAP32[$dst$013>>2] = $86;
     $87 = (($posx$2) + ($20))|0;
     $88 = ((($dst$013)) + 4|0);
     $89 = (($90) + -1)|0;
     $91 = ($90|0)==(0);
     if ($91) {
      break;
     } else {
      $90 = $89;$dst$013 = $88;$posx$011 = $87;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $92 = HEAP32[$24>>2]|0;
  $93 = (($94) + ($92)|0);
  $95 = HEAP32[$12>>2]|0;
  $96 = (($95) + -1)|0;
  HEAP32[$12>>2] = $96;
  $97 = ($95|0)==(0);
  if ($97) {
   $$lcssa = $93;
   break;
  }
  $98 = (($posy$1$lcssa) + ($14))|0;
  $$pre27 = HEAP32[$18>>2]|0;
  $36 = $$pre27;$94 = $93;$99 = $93;$posy$017 = $98;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$23>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGR888_BGR888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$$us = 0, $$1 = 0, $$1$us = 0, $$3 = 0, $$3$us = 0, $$4 = 0, $$lcssa = 0, $$lcssa18 = 0, $$pre = 0, $$pre11 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$06 = 0, $dst$06$us = 0, $dstB$0 = 0, $dstB$0$us = 0, $dstG$0 = 0, $dstG$0$us = 0;
 var $dstR$0 = 0, $dstR$0$us = 0, $or$cond = 0, $src$05 = 0, $src$05$us = 0, $srcB$1 = 0, $srcB$1$us = 0, $srcG$1 = 0, $srcG$1$us = 0, $srcR$1 = 0, $srcR$1$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = $3 >>> 24;
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $$4 = $21 ? 255 : $13;
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = ($$4|0)!=(255);
 $or$cond = $23 & $24;
 $25 = $1 & 112;
 $26 = $$4 ^ 255;
 $$pre = HEAP32[$info>>2]|0;
 $$pre11 = HEAP32[$14>>2]|0;
 $134 = $$pre;$137 = $$pre11;$141 = $$pre11;$142 = $$pre;
 while(1) {
  $27 = HEAP32[$15>>2]|0;
  $28 = (($27) + -1)|0;
  $29 = ($27|0)==(0);
  L6: do {
   if (!($29)) {
    if ($19) {
     $76 = $28;$dst$06$us = $141;$src$05$us = $142;
     while(1) {
      $30 = HEAP32[$src$05$us>>2]|0;
      $31 = $30 >>> 16;
      $32 = $31 & 255;
      $33 = $30 >>> 8;
      $34 = $33 & 255;
      $35 = $30 & 255;
      $36 = HEAP32[$dst$06$us>>2]|0;
      $37 = $36 >>> 16;
      $38 = $37 & 255;
      $39 = $36 >>> 8;
      $40 = $39 & 255;
      $41 = $36 & 255;
      if ($or$cond) {
       $42 = Math_imul($35, $$4)|0;
       $43 = (($42>>>0) / 255)&-1;
       $44 = Math_imul($34, $$4)|0;
       $45 = (($44>>>0) / 255)&-1;
       $46 = Math_imul($32, $$4)|0;
       $47 = (($46>>>0) / 255)&-1;
       $srcB$1$us = $47;$srcG$1$us = $45;$srcR$1$us = $43;
      } else {
       $srcB$1$us = $32;$srcG$1$us = $34;$srcR$1$us = $35;
      }
      switch ($25|0) {
      case 16:  {
       $60 = Math_imul($41, $26)|0;
       $61 = (($60>>>0) / 255)&-1;
       $62 = (($srcR$1$us) + ($61))|0;
       $63 = Math_imul($40, $26)|0;
       $64 = (($63>>>0) / 255)&-1;
       $65 = (($srcG$1$us) + ($64))|0;
       $66 = Math_imul($38, $26)|0;
       $67 = (($66>>>0) / 255)&-1;
       $68 = (($srcB$1$us) + ($67))|0;
       $dstB$0$us = $68;$dstG$0$us = $65;$dstR$0$us = $62;
       break;
      }
      case 32:  {
       $54 = (($srcR$1$us) + ($41))|0;
       $55 = ($54>>>0)>(255);
       $$$us = $55 ? 255 : $54;
       $56 = (($srcG$1$us) + ($40))|0;
       $57 = ($56>>>0)>(255);
       $$3$us = $57 ? 255 : $56;
       $58 = (($srcB$1$us) + ($38))|0;
       $59 = ($58>>>0)>(255);
       $$1$us = $59 ? 255 : $58;
       $dstB$0$us = $$1$us;$dstG$0$us = $$3$us;$dstR$0$us = $$$us;
       break;
      }
      case 64:  {
       $48 = Math_imul($srcR$1$us, $41)|0;
       $49 = (($48>>>0) / 255)&-1;
       $50 = Math_imul($srcG$1$us, $40)|0;
       $51 = (($50>>>0) / 255)&-1;
       $52 = Math_imul($srcB$1$us, $38)|0;
       $53 = (($52>>>0) / 255)&-1;
       $dstB$0$us = $53;$dstG$0$us = $51;$dstR$0$us = $49;
       break;
      }
      default: {
       $dstB$0$us = $38;$dstG$0$us = $40;$dstR$0$us = $41;
      }
      }
      $69 = $dstB$0$us << 16;
      $70 = $dstG$0$us << 8;
      $71 = $70 | $dstR$0$us;
      $72 = $71 | $69;
      HEAP32[$dst$06$us>>2] = $72;
      $73 = ((($src$05$us)) + 4|0);
      $74 = ((($dst$06$us)) + 4|0);
      $75 = (($76) + -1)|0;
      $77 = ($76|0)==(0);
      if ($77) {
       break L6;
      } else {
       $76 = $75;$dst$06$us = $74;$src$05$us = $73;
      }
     }
    } else {
     $130 = $28;$dst$06 = $141;$src$05 = $142;
     while(1) {
      $78 = HEAP32[$src$05>>2]|0;
      $79 = $78 >>> 16;
      $80 = $79 & 255;
      $81 = $78 >>> 8;
      $82 = $81 & 255;
      $83 = $78 & 255;
      $84 = HEAP32[$dst$06>>2]|0;
      $85 = $84 >>> 16;
      $86 = $85 & 255;
      $87 = $84 >>> 8;
      $88 = $87 & 255;
      $89 = $84 & 255;
      $90 = Math_imul($83, $4)|0;
      $91 = (($90>>>0) / 255)&-1;
      $92 = Math_imul($82, $6)|0;
      $93 = (($92>>>0) / 255)&-1;
      $94 = Math_imul($80, $8)|0;
      $95 = (($94>>>0) / 255)&-1;
      if ($or$cond) {
       $96 = Math_imul($91, $$4)|0;
       $97 = (($96>>>0) / 255)&-1;
       $98 = Math_imul($93, $$4)|0;
       $99 = (($98>>>0) / 255)&-1;
       $100 = Math_imul($95, $$4)|0;
       $101 = (($100>>>0) / 255)&-1;
       $srcB$1 = $101;$srcG$1 = $99;$srcR$1 = $97;
      } else {
       $srcB$1 = $95;$srcG$1 = $93;$srcR$1 = $91;
      }
      switch ($25|0) {
      case 16:  {
       $102 = Math_imul($89, $26)|0;
       $103 = (($102>>>0) / 255)&-1;
       $104 = (($srcR$1) + ($103))|0;
       $105 = Math_imul($88, $26)|0;
       $106 = (($105>>>0) / 255)&-1;
       $107 = (($srcG$1) + ($106))|0;
       $108 = Math_imul($86, $26)|0;
       $109 = (($108>>>0) / 255)&-1;
       $110 = (($srcB$1) + ($109))|0;
       $dstB$0 = $110;$dstG$0 = $107;$dstR$0 = $104;
       break;
      }
      case 32:  {
       $111 = (($srcR$1) + ($89))|0;
       $112 = ($111>>>0)>(255);
       $$ = $112 ? 255 : $111;
       $113 = (($srcG$1) + ($88))|0;
       $114 = ($113>>>0)>(255);
       $$3 = $114 ? 255 : $113;
       $115 = (($srcB$1) + ($86))|0;
       $116 = ($115>>>0)>(255);
       $$1 = $116 ? 255 : $115;
       $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
       break;
      }
      case 64:  {
       $117 = Math_imul($srcR$1, $89)|0;
       $118 = (($117>>>0) / 255)&-1;
       $119 = Math_imul($srcG$1, $88)|0;
       $120 = (($119>>>0) / 255)&-1;
       $121 = Math_imul($srcB$1, $86)|0;
       $122 = (($121>>>0) / 255)&-1;
       $dstB$0 = $122;$dstG$0 = $120;$dstR$0 = $118;
       break;
      }
      default: {
       $dstB$0 = $86;$dstG$0 = $88;$dstR$0 = $89;
      }
      }
      $123 = $dstB$0 << 16;
      $124 = $dstG$0 << 8;
      $125 = $124 | $dstR$0;
      $126 = $125 | $123;
      HEAP32[$dst$06>>2] = $126;
      $127 = ((($src$05)) + 4|0);
      $128 = ((($dst$06)) + 4|0);
      $129 = (($130) + -1)|0;
      $131 = ($130|0)==(0);
      if ($131) {
       break L6;
      } else {
       $130 = $129;$dst$06 = $128;$src$05 = $127;
      }
     }
    }
   }
  } while(0);
  $132 = HEAP32[$16>>2]|0;
  $133 = (($134) + ($132)|0);
  $135 = HEAP32[$17>>2]|0;
  $136 = (($137) + ($135)|0);
  $138 = HEAP32[$9>>2]|0;
  $139 = (($138) + -1)|0;
  HEAP32[$9>>2] = $139;
  $140 = ($138|0)==(0);
  if ($140) {
   $$lcssa = $133;$$lcssa18 = $136;
   break;
  } else {
   $134 = $133;$137 = $136;$141 = $136;$142 = $133;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa18;
 return;
}
function _SDL_Blit_BGR888_BGR888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$4 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre28 = 0, $$pre29 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$017 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$015 = 0, $posx$2 = 0, $posy$021 = 0, $posy$1$lcssa = 0, $src$016 = 0, $src$1 = 0;
 var $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$014 = 0, $srcx$2 = 0, $srcy$020 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = $3 >>> 24;
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $$4 = $29 ? 255 : $23;
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = ($$4|0)!=(255);
 $or$cond = $31 & $32;
 $33 = $1 & 112;
 $34 = $$4 ^ 255;
 $35 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $113 = $$pre;$118 = $21;$119 = $$pre;$44 = $19;$posy$021 = 0;$srcy$020 = 0;
 while(1) {
  $36 = ($posy$021|0)>(65535);
  if ($36) {
   $37 = (($posy$021) + -65536)|0;
   $38 = $37 >>> 16;
   $39 = $38 << 16;
   $40 = (($srcy$020) + 1)|0;
   $41 = (($37) - ($39))|0;
   $42 = (($40) + ($38))|0;
   $posy$1$lcssa = $41;$srcy$1$lcssa = $42;
  } else {
   $posy$1$lcssa = $posy$021;$srcy$1$lcssa = $srcy$020;
  }
  $43 = ($44|0)==(0);
  if ($43) {
   $115 = $118;
  } else {
   $$in = $44;$dst$017 = $119;$posx$015 = 65536;$src$016 = 0;$srcx$014 = -1;
   while(1) {
    $45 = (($$in) + -1)|0;
    $46 = ($posx$015|0)>(65535);
    if ($46) {
     $47 = (($posx$015) + -65536)|0;
     $48 = $47 >>> 16;
     $49 = $48 << 16;
     $50 = (($srcx$014) + 1)|0;
     $51 = (($47) - ($49))|0;
     $52 = (($50) + ($48))|0;
     $53 = HEAP32[$info>>2]|0;
     $54 = HEAP32[$35>>2]|0;
     $55 = Math_imul($54, $srcy$1$lcssa)|0;
     $56 = (($53) + ($55)|0);
     $57 = $52 << 2;
     $58 = (($56) + ($57)|0);
     $posx$2 = $51;$src$1 = $58;$srcx$2 = $52;
    } else {
     $posx$2 = $posx$015;$src$1 = $src$016;$srcx$2 = $srcx$014;
    }
    $59 = HEAP32[$src$1>>2]|0;
    $60 = $59 >>> 16;
    $61 = $60 & 255;
    $62 = $59 >>> 8;
    $63 = $62 & 255;
    $64 = $59 & 255;
    $65 = HEAP32[$dst$017>>2]|0;
    $66 = $65 >>> 16;
    $67 = $66 & 255;
    $68 = $65 >>> 8;
    $69 = $68 & 255;
    $70 = $65 & 255;
    if ($27) {
     $srcB$0 = $61;$srcG$0 = $63;$srcR$0 = $64;
    } else {
     $71 = Math_imul($64, $4)|0;
     $72 = (($71>>>0) / 255)&-1;
     $73 = Math_imul($63, $6)|0;
     $74 = (($73>>>0) / 255)&-1;
     $75 = Math_imul($61, $8)|0;
     $76 = (($75>>>0) / 255)&-1;
     $srcB$0 = $76;$srcG$0 = $74;$srcR$0 = $72;
    }
    if ($or$cond) {
     $77 = Math_imul($srcR$0, $$4)|0;
     $78 = (($77>>>0) / 255)&-1;
     $79 = Math_imul($srcG$0, $$4)|0;
     $80 = (($79>>>0) / 255)&-1;
     $81 = Math_imul($srcB$0, $$4)|0;
     $82 = (($81>>>0) / 255)&-1;
     $srcB$1 = $82;$srcG$1 = $80;$srcR$1 = $78;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($33|0) {
    case 16:  {
     $83 = Math_imul($70, $34)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = (($srcR$1) + ($84))|0;
     $86 = Math_imul($69, $34)|0;
     $87 = (($86>>>0) / 255)&-1;
     $88 = (($srcG$1) + ($87))|0;
     $89 = Math_imul($67, $34)|0;
     $90 = (($89>>>0) / 255)&-1;
     $91 = (($srcB$1) + ($90))|0;
     $dstB$0 = $91;$dstG$0 = $88;$dstR$0 = $85;
     break;
    }
    case 32:  {
     $92 = (($srcR$1) + ($70))|0;
     $93 = ($92>>>0)>(255);
     $$ = $93 ? 255 : $92;
     $94 = (($srcG$1) + ($69))|0;
     $95 = ($94>>>0)>(255);
     $$3 = $95 ? 255 : $94;
     $96 = (($srcB$1) + ($67))|0;
     $97 = ($96>>>0)>(255);
     $$1 = $97 ? 255 : $96;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $98 = Math_imul($srcR$1, $70)|0;
     $99 = (($98>>>0) / 255)&-1;
     $100 = Math_imul($srcG$1, $69)|0;
     $101 = (($100>>>0) / 255)&-1;
     $102 = Math_imul($srcB$1, $67)|0;
     $103 = (($102>>>0) / 255)&-1;
     $dstB$0 = $103;$dstG$0 = $101;$dstR$0 = $99;
     break;
    }
    default: {
     $dstB$0 = $67;$dstG$0 = $69;$dstR$0 = $70;
    }
    }
    $104 = $dstB$0 << 16;
    $105 = $dstG$0 << 8;
    $106 = $105 | $dstR$0;
    $107 = $106 | $104;
    HEAP32[$dst$017>>2] = $107;
    $108 = (($posx$2) + ($20))|0;
    $109 = ((($dst$017)) + 4|0);
    $110 = ($45|0)==(0);
    if ($110) {
     break;
    } else {
     $$in = $45;$dst$017 = $109;$posx$015 = $108;$src$016 = $src$1;$srcx$014 = $srcx$2;
    }
   }
   $$pre29 = HEAP32[$12>>2]|0;
   $115 = $$pre29;
  }
  $111 = HEAP32[$25>>2]|0;
  $112 = (($113) + ($111)|0);
  $114 = (($115) + -1)|0;
  HEAP32[$12>>2] = $114;
  $116 = ($115|0)==(0);
  if ($116) {
   $$lcssa = $112;
   break;
  }
  $117 = (($posy$1$lcssa) + ($14))|0;
  $$pre28 = HEAP32[$18>>2]|0;
  $113 = $112;$118 = $114;$119 = $112;$44 = $$pre28;$posy$021 = $117;$srcy$020 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGR888_ARGB8888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0, $posx$010 = 0, $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$53 = $$pre;$58 = $12;$59 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $55 = $58;
  } else {
   $$in = $25;$dst$013 = $59;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    $41 = $40 >>> 16;
    $42 = $41 & 255;
    $43 = $40 & 65280;
    $44 = $40 << 16;
    $45 = $43 | $44;
    $46 = $45 | $42;
    $47 = $46 | -16777216;
    HEAP32[$dst$013>>2] = $47;
    $48 = (($posx$2) + ($11))|0;
    $49 = ((($dst$013)) + 4|0);
    $50 = ($26|0)==(0);
    if ($50) {
     break;
    } else {
     $$in = $26;$dst$013 = $49;$posx$010 = $48;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $55 = $$pre25;
  }
  $51 = HEAP32[$15>>2]|0;
  $52 = (($53) + ($51)|0);
  $54 = (($55) + -1)|0;
  HEAP32[$3>>2] = $54;
  $56 = ($55|0)==(0);
  if ($56) {
   $$lcssa = $52;
   break;
  }
  $57 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$53 = $52;$58 = $54;$59 = $52;$posy$016 = $57;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGR888_ARGB8888_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$lcssa = 0, $$lcssa17 = 0, $$pre = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $8 = 0;
 var $9 = 0, $dst$05 = 0, $dst$05$us = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $src$04 = 0, $src$04$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 112;
 $11 = ($10|0)==(16);
 $$pre = HEAP32[$info>>2]|0;
 $$pre10 = HEAP32[$6>>2]|0;
 $65 = $$pre;$68 = $$pre10;$72 = $$pre10;$73 = $$pre;
 while(1) {
  $12 = HEAP32[$7>>2]|0;
  $13 = (($12) + -1)|0;
  $14 = ($12|0)==(0);
  L6: do {
   if (!($14)) {
    if ($11) {
     $26 = $13;$dst$05$us = $72;$src$04$us = $73;
     while(1) {
      $15 = HEAP32[$src$04$us>>2]|0;
      $16 = $15 >>> 16;
      $17 = $16 & 255;
      $18 = $15 & 65280;
      $19 = $15 << 16;
      $20 = $18 | $19;
      $21 = $20 | $17;
      $22 = $21 | -16777216;
      HEAP32[$dst$05$us>>2] = $22;
      $23 = ((($src$04$us)) + 4|0);
      $24 = ((($dst$05$us)) + 4|0);
      $25 = (($26) + -1)|0;
      $27 = ($26|0)==(0);
      if ($27) {
       break L6;
      } else {
       $26 = $25;$dst$05$us = $24;$src$04$us = $23;
      }
     }
    } else {
     $61 = $13;$dst$05 = $72;$src$04 = $73;
    }
    while(1) {
     $28 = HEAP32[$src$04>>2]|0;
     $29 = $28 >>> 16;
     $30 = $29 & 255;
     $31 = $28 >>> 8;
     $32 = $31 & 255;
     $33 = $28 & 255;
     $34 = HEAP32[$dst$05>>2]|0;
     $35 = $34 >>> 16;
     $36 = $35 & 255;
     $37 = $34 >>> 8;
     $38 = $37 & 255;
     $39 = $34 & 255;
     $40 = $34 & -16777216;
     switch ($10|0) {
     case 64:  {
      $47 = Math_imul($36, $33)|0;
      $48 = (($47>>>0) / 255)&-1;
      $49 = Math_imul($38, $32)|0;
      $50 = (($49>>>0) / 255)&-1;
      $51 = Math_imul($30, $39)|0;
      $52 = (($51>>>0) / 255)&-1;
      $dstB$0 = $52;$dstG$0 = $50;$dstR$0 = $48;
      break;
     }
     case 32:  {
      $41 = (($36) + ($33))|0;
      $42 = ($41>>>0)>(255);
      $$ = $42 ? 255 : $41;
      $43 = (($38) + ($32))|0;
      $44 = ($43>>>0)>(255);
      $$3 = $44 ? 255 : $43;
      $45 = (($30) + ($39))|0;
      $46 = ($45>>>0)>(255);
      $$1 = $46 ? 255 : $45;
      $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
      break;
     }
     default: {
      $dstB$0 = $39;$dstG$0 = $38;$dstR$0 = $36;
     }
     }
     $53 = $dstR$0 << 16;
     $54 = $dstG$0 << 8;
     $55 = $54 | $53;
     $56 = $55 | $dstB$0;
     $57 = $56 | $40;
     HEAP32[$dst$05>>2] = $57;
     $58 = ((($src$04)) + 4|0);
     $59 = ((($dst$05)) + 4|0);
     $60 = (($61) + -1)|0;
     $62 = ($61|0)==(0);
     if ($62) {
      break L6;
     } else {
      $61 = $60;$dst$05 = $59;$src$04 = $58;
     }
    }
   }
  } while(0);
  $63 = HEAP32[$8>>2]|0;
  $64 = (($65) + ($63)|0);
  $66 = HEAP32[$9>>2]|0;
  $67 = (($68) + ($66)|0);
  $69 = HEAP32[$2>>2]|0;
  $70 = (($69) + -1)|0;
  HEAP32[$2>>2] = $70;
  $71 = ($69|0)==(0);
  if ($71) {
   $$lcssa = $64;$$lcssa17 = $67;
   break;
  } else {
   $65 = $64;$68 = $67;$72 = $67;$73 = $64;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa17;
 return;
}
function _SDL_Blit_BGR888_ARGB8888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$lcssa = 0, $$pre = 0, $$pre30 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0;
 var $dst$016$us = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $posx$014 = 0, $posx$014$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$020 = 0, $posy$1$lcssa = 0, $src$015 = 0, $src$015$us = 0, $src$1 = 0, $src$1$us = 0, $srcx$013 = 0, $srcx$013$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$019 = 0, $srcy$1$lcssa = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 112;
 $19 = ((($info)) + 12|0);
 $20 = ($18|0)==(16);
 $$pre = HEAP32[$16>>2]|0;
 $107 = $$pre;$112 = $$pre;$29 = $12;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $21 = ($posy$020|0)>(65535);
  if ($21) {
   $22 = (($posy$020) + -65536)|0;
   $23 = $22 >>> 16;
   $24 = $23 << 16;
   $25 = (($srcy$019) + 1)|0;
   $26 = (($22) - ($24))|0;
   $27 = (($25) + ($23))|0;
   $posy$1$lcssa = $26;$srcy$1$lcssa = $27;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $28 = (($29) + -1)|0;
  $30 = ($29|0)==(0);
  L9: do {
   if (!($30)) {
    if ($20) {
     $55 = $28;$dst$016$us = $112;$posx$014$us = 65536;$src$015$us = 0;$srcx$013$us = -1;
     while(1) {
      $31 = ($posx$014$us|0)>(65535);
      if ($31) {
       $32 = (($posx$014$us) + -65536)|0;
       $33 = $32 >>> 16;
       $34 = $33 << 16;
       $35 = (($srcx$013$us) + 1)|0;
       $36 = (($32) - ($34))|0;
       $37 = (($35) + ($33))|0;
       $38 = HEAP32[$info>>2]|0;
       $39 = HEAP32[$19>>2]|0;
       $40 = Math_imul($39, $srcy$1$lcssa)|0;
       $41 = (($38) + ($40)|0);
       $42 = $37 << 2;
       $43 = (($41) + ($42)|0);
       $posx$2$us = $36;$src$1$us = $43;$srcx$2$us = $37;
      } else {
       $posx$2$us = $posx$014$us;$src$1$us = $src$015$us;$srcx$2$us = $srcx$013$us;
      }
      $44 = HEAP32[$src$1$us>>2]|0;
      $45 = $44 >>> 16;
      $46 = $45 & 255;
      $47 = $44 & 65280;
      $48 = $44 << 16;
      $49 = $47 | $48;
      $50 = $49 | $46;
      $51 = $50 | -16777216;
      HEAP32[$dst$016$us>>2] = $51;
      $52 = (($posx$2$us) + ($13))|0;
      $53 = ((($dst$016$us)) + 4|0);
      $54 = (($55) + -1)|0;
      $56 = ($55|0)==(0);
      if ($56) {
       break L9;
      } else {
       $55 = $54;$dst$016$us = $53;$posx$014$us = $52;$src$015$us = $src$1$us;$srcx$013$us = $srcx$2$us;
      }
     }
    } else {
     $103 = $28;$dst$016 = $112;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
    }
    while(1) {
     $57 = ($posx$014|0)>(65535);
     if ($57) {
      $58 = (($posx$014) + -65536)|0;
      $59 = $58 >>> 16;
      $60 = $59 << 16;
      $61 = (($srcx$013) + 1)|0;
      $62 = (($58) - ($60))|0;
      $63 = (($61) + ($59))|0;
      $64 = HEAP32[$info>>2]|0;
      $65 = HEAP32[$19>>2]|0;
      $66 = Math_imul($65, $srcy$1$lcssa)|0;
      $67 = (($64) + ($66)|0);
      $68 = $63 << 2;
      $69 = (($67) + ($68)|0);
      $posx$2 = $62;$src$1 = $69;$srcx$2 = $63;
     } else {
      $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
     }
     $70 = HEAP32[$src$1>>2]|0;
     $71 = $70 >>> 16;
     $72 = $71 & 255;
     $73 = $70 >>> 8;
     $74 = $73 & 255;
     $75 = $70 & 255;
     $76 = HEAP32[$dst$016>>2]|0;
     $77 = $76 >>> 16;
     $78 = $77 & 255;
     $79 = $76 >>> 8;
     $80 = $79 & 255;
     $81 = $76 & 255;
     $82 = $76 & -16777216;
     switch ($18|0) {
     case 64:  {
      $89 = Math_imul($78, $75)|0;
      $90 = (($89>>>0) / 255)&-1;
      $91 = Math_imul($80, $74)|0;
      $92 = (($91>>>0) / 255)&-1;
      $93 = Math_imul($72, $81)|0;
      $94 = (($93>>>0) / 255)&-1;
      $dstB$0 = $94;$dstG$0 = $92;$dstR$0 = $90;
      break;
     }
     case 32:  {
      $83 = (($78) + ($75))|0;
      $84 = ($83>>>0)>(255);
      $$ = $84 ? 255 : $83;
      $85 = (($80) + ($74))|0;
      $86 = ($85>>>0)>(255);
      $$3 = $86 ? 255 : $85;
      $87 = (($72) + ($81))|0;
      $88 = ($87>>>0)>(255);
      $$1 = $88 ? 255 : $87;
      $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
      break;
     }
     default: {
      $dstB$0 = $81;$dstG$0 = $80;$dstR$0 = $78;
     }
     }
     $95 = $dstR$0 << 16;
     $96 = $dstG$0 << 8;
     $97 = $96 | $95;
     $98 = $97 | $dstB$0;
     $99 = $98 | $82;
     HEAP32[$dst$016>>2] = $99;
     $100 = (($posx$2) + ($13))|0;
     $101 = ((($dst$016)) + 4|0);
     $102 = (($103) + -1)|0;
     $104 = ($103|0)==(0);
     if ($104) {
      break L9;
     } else {
      $103 = $102;$dst$016 = $101;$posx$014 = $100;$src$015 = $src$1;$srcx$013 = $srcx$2;
     }
    }
   }
  } while(0);
  $105 = HEAP32[$17>>2]|0;
  $106 = (($107) + ($105)|0);
  $108 = HEAP32[$5>>2]|0;
  $109 = (($108) + -1)|0;
  HEAP32[$5>>2] = $109;
  $110 = ($108|0)==(0);
  if ($110) {
   $$lcssa = $106;
   break;
  }
  $111 = (($posy$1$lcssa) + ($7))|0;
  $$pre30 = HEAP32[$11>>2]|0;
  $107 = $106;$112 = $106;$29 = $$pre30;$posy$020 = $111;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGR888_ARGB8888_Modulate($info) {
 $info = $info|0;
 var $$in = 0, $$in13 = 0, $$lcssa = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$pre16 = 0, $$pre17 = 0, $$pre18 = 0, $$pre19 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $9 = 0, $dst$02 = 0, $dst$02$us$us = 0, $src$01 = 0, $src$01$us$us = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = $3 & -16777216;
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $22 = $21 ? -16777216 : $13;
 $$pre17 = HEAP32[$info>>2]|0;
 $$pre18 = HEAP32[$14>>2]|0;
 if ($19) {
  $27 = $$pre17;$30 = $$pre18;$79 = $11;$80 = $$pre18;$81 = $$pre17;
  while(1) {
   $23 = HEAP32[$15>>2]|0;
   $24 = ($23|0)==(0);
   if ($24) {
    $32 = $79;
   } else {
    $$in13 = $23;$dst$02$us$us = $80;$src$01$us$us = $81;
    while(1) {
     $34 = (($$in13) + -1)|0;
     $35 = HEAP32[$src$01$us$us>>2]|0;
     $36 = $35 >>> 16;
     $37 = $36 & 255;
     $38 = $35 & 65280;
     $39 = $35 << 16;
     $40 = $39 & 16711680;
     $41 = $40 | $22;
     $42 = $41 | $38;
     $43 = $42 | $37;
     HEAP32[$dst$02$us$us>>2] = $43;
     $44 = ((($src$01$us$us)) + 4|0);
     $45 = ((($dst$02$us$us)) + 4|0);
     $46 = ($34|0)==(0);
     if ($46) {
      break;
     } else {
      $$in13 = $34;$dst$02$us$us = $45;$src$01$us$us = $44;
     }
    }
    $$pre19 = HEAP32[$9>>2]|0;
    $32 = $$pre19;
   }
   $25 = HEAP32[$16>>2]|0;
   $26 = (($27) + ($25)|0);
   $28 = HEAP32[$17>>2]|0;
   $29 = (($30) + ($28)|0);
   $31 = (($32) + -1)|0;
   HEAP32[$9>>2] = $31;
   $33 = ($32|0)==(0);
   if ($33) {
    $$lcssa = $26;$$lcssa30 = $29;
    break;
   } else {
    $27 = $26;$30 = $29;$79 = $31;$80 = $29;$81 = $26;
   }
  }
  HEAP32[$info>>2] = $$lcssa;
  HEAP32[$14>>2] = $$lcssa30;
  return;
 } else {
  $72 = $$pre17;$75 = $$pre18;$82 = $11;$83 = $$pre18;$84 = $$pre17;
  while(1) {
   $47 = HEAP32[$15>>2]|0;
   $48 = ($47|0)==(0);
   if ($48) {
    $77 = $82;
   } else {
    $$in = $47;$dst$02 = $83;$src$01 = $84;
    while(1) {
     $49 = (($$in) + -1)|0;
     $50 = HEAP32[$src$01>>2]|0;
     $51 = $50 >>> 16;
     $52 = $51 & 255;
     $53 = $50 >>> 8;
     $54 = $53 & 255;
     $55 = $50 & 255;
     $56 = Math_imul($55, $4)|0;
     $57 = (($56>>>0) / 255)&-1;
     $58 = Math_imul($54, $6)|0;
     $59 = (($58>>>0) / 255)&-1;
     $60 = Math_imul($52, $8)|0;
     $61 = (($60>>>0) / 255)&-1;
     $62 = $57 << 16;
     $63 = $62 | $22;
     $64 = $59 << 8;
     $65 = $63 | $64;
     $66 = $65 | $61;
     HEAP32[$dst$02>>2] = $66;
     $67 = ((($src$01)) + 4|0);
     $68 = ((($dst$02)) + 4|0);
     $69 = ($49|0)==(0);
     if ($69) {
      break;
     } else {
      $$in = $49;$dst$02 = $68;$src$01 = $67;
     }
    }
    $$pre16 = HEAP32[$9>>2]|0;
    $77 = $$pre16;
   }
   $70 = HEAP32[$16>>2]|0;
   $71 = (($72) + ($70)|0);
   $73 = HEAP32[$17>>2]|0;
   $74 = (($75) + ($73)|0);
   $76 = (($77) + -1)|0;
   HEAP32[$9>>2] = $76;
   $78 = ($77|0)==(0);
   if ($78) {
    $$lcssa31 = $71;$$lcssa32 = $74;
    break;
   } else {
    $72 = $71;$75 = $74;$82 = $76;$83 = $74;$84 = $71;
   }
  }
  HEAP32[$info>>2] = $$lcssa31;
  HEAP32[$14>>2] = $$lcssa32;
  return;
 }
}
function _SDL_Blit_BGR888_ARGB8888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $12 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$013 = 0, $dst$013$us = 0, $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0;
 var $posx$2$us = 0, $posy$017 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0, $src$1$us = 0, $srcx$010 = 0, $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$016 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = $3 & -16777216;
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = $29 ? -16777216 : $23;
 $31 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $106 = $$pre;$111 = $$pre;$40 = $19;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $32 = ($posy$017|0)>(65535);
  if ($32) {
   $33 = (($posy$017) + -65536)|0;
   $34 = $33 >>> 16;
   $35 = $34 << 16;
   $36 = (($srcy$016) + 1)|0;
   $37 = (($33) - ($35))|0;
   $38 = (($36) + ($34))|0;
   $posy$1$lcssa = $37;$srcy$1$lcssa = $38;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $39 = (($40) + -1)|0;
  $41 = ($40|0)==(0);
  if (!($41)) {
   if ($27) {
    $67 = $39;$dst$013$us = $111;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $42 = ($posx$011$us|0)>(65535);
     if ($42) {
      $43 = (($posx$011$us) + -65536)|0;
      $44 = $43 >>> 16;
      $45 = $44 << 16;
      $46 = (($srcx$010$us) + 1)|0;
      $47 = (($43) - ($45))|0;
      $48 = (($46) + ($44))|0;
      $49 = HEAP32[$info>>2]|0;
      $50 = HEAP32[$31>>2]|0;
      $51 = Math_imul($50, $srcy$1$lcssa)|0;
      $52 = (($49) + ($51)|0);
      $53 = $48 << 2;
      $54 = (($52) + ($53)|0);
      $posx$2$us = $47;$src$1$us = $54;$srcx$2$us = $48;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $55 = HEAP32[$src$1$us>>2]|0;
     $56 = $55 >>> 16;
     $57 = $56 & 255;
     $58 = $55 & 65280;
     $59 = $55 << 16;
     $60 = $59 & 16711680;
     $61 = $58 | $30;
     $62 = $61 | $57;
     $63 = $62 | $60;
     HEAP32[$dst$013$us>>2] = $63;
     $64 = (($posx$2$us) + ($20))|0;
     $65 = ((($dst$013$us)) + 4|0);
     $66 = (($67) + -1)|0;
     $68 = ($67|0)==(0);
     if ($68) {
      break;
     } else {
      $67 = $66;$dst$013$us = $65;$posx$011$us = $64;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $102 = $39;$dst$013 = $111;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $69 = ($posx$011|0)>(65535);
     if ($69) {
      $70 = (($posx$011) + -65536)|0;
      $71 = $70 >>> 16;
      $72 = $71 << 16;
      $73 = (($srcx$010) + 1)|0;
      $74 = (($70) - ($72))|0;
      $75 = (($73) + ($71))|0;
      $76 = HEAP32[$info>>2]|0;
      $77 = HEAP32[$31>>2]|0;
      $78 = Math_imul($77, $srcy$1$lcssa)|0;
      $79 = (($76) + ($78)|0);
      $80 = $75 << 2;
      $81 = (($79) + ($80)|0);
      $posx$2 = $74;$src$1 = $81;$srcx$2 = $75;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $82 = HEAP32[$src$1>>2]|0;
     $83 = $82 >>> 16;
     $84 = $83 & 255;
     $85 = $82 >>> 8;
     $86 = $85 & 255;
     $87 = $82 & 255;
     $88 = Math_imul($87, $4)|0;
     $89 = (($88>>>0) / 255)&-1;
     $90 = Math_imul($86, $6)|0;
     $91 = (($90>>>0) / 255)&-1;
     $92 = Math_imul($84, $8)|0;
     $93 = (($92>>>0) / 255)&-1;
     $94 = $89 << 16;
     $95 = $91 << 8;
     $96 = $95 | $30;
     $97 = $96 | $93;
     $98 = $97 | $94;
     HEAP32[$dst$013>>2] = $98;
     $99 = (($posx$2) + ($20))|0;
     $100 = ((($dst$013)) + 4|0);
     $101 = (($102) + -1)|0;
     $103 = ($102|0)==(0);
     if ($103) {
      break;
     } else {
      $102 = $101;$dst$013 = $100;$posx$011 = $99;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $104 = HEAP32[$25>>2]|0;
  $105 = (($106) + ($104)|0);
  $107 = HEAP32[$12>>2]|0;
  $108 = (($107) + -1)|0;
  HEAP32[$12>>2] = $108;
  $109 = ($107|0)==(0);
  if ($109) {
   $$lcssa = $105;
   break;
  }
  $110 = (($posy$1$lcssa) + ($14))|0;
  $$pre27 = HEAP32[$18>>2]|0;
  $106 = $105;$111 = $105;$40 = $$pre27;$posy$017 = $110;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGR888_ARGB8888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$4 = 0, $$in = 0, $$lcssa = 0, $$lcssa15 = 0, $$pre = 0, $$pre10 = 0, $$pre11 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$06 = 0, $dstA$0 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $src$05 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0;
 var $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = $3 >>> 24;
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $$4 = $21 ? 255 : $13;
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = ($$4|0)!=(255);
 $or$cond = $23 & $24;
 $25 = $1 & 112;
 $26 = $$4 ^ 255;
 $$pre = HEAP32[$info>>2]|0;
 $$pre10 = HEAP32[$14>>2]|0;
 $90 = $$pre;$93 = $$pre10;$97 = $11;$98 = $$pre10;$99 = $$pre;
 while(1) {
  $27 = HEAP32[$15>>2]|0;
  $28 = ($27|0)==(0);
  if ($28) {
   $95 = $97;
  } else {
   $$in = $27;$dst$06 = $98;$src$05 = $99;
   while(1) {
    $29 = (($$in) + -1)|0;
    $30 = HEAP32[$src$05>>2]|0;
    $31 = $30 >>> 16;
    $32 = $31 & 255;
    $33 = $30 >>> 8;
    $34 = $33 & 255;
    $35 = $30 & 255;
    $36 = HEAP32[$dst$06>>2]|0;
    $37 = $36 >>> 16;
    $38 = $37 & 255;
    $39 = $36 >>> 8;
    $40 = $39 & 255;
    $41 = $36 & 255;
    $42 = $36 >>> 24;
    if ($19) {
     $srcB$0 = $32;$srcG$0 = $34;$srcR$0 = $35;
    } else {
     $43 = Math_imul($35, $4)|0;
     $44 = (($43>>>0) / 255)&-1;
     $45 = Math_imul($34, $6)|0;
     $46 = (($45>>>0) / 255)&-1;
     $47 = Math_imul($32, $8)|0;
     $48 = (($47>>>0) / 255)&-1;
     $srcB$0 = $48;$srcG$0 = $46;$srcR$0 = $44;
    }
    if ($or$cond) {
     $49 = Math_imul($srcR$0, $$4)|0;
     $50 = (($49>>>0) / 255)&-1;
     $51 = Math_imul($srcG$0, $$4)|0;
     $52 = (($51>>>0) / 255)&-1;
     $53 = Math_imul($srcB$0, $$4)|0;
     $54 = (($53>>>0) / 255)&-1;
     $srcB$1 = $54;$srcG$1 = $52;$srcR$1 = $50;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($25|0) {
    case 16:  {
     $55 = Math_imul($38, $26)|0;
     $56 = (($55>>>0) / 255)&-1;
     $57 = (($srcR$1) + ($56))|0;
     $58 = Math_imul($40, $26)|0;
     $59 = (($58>>>0) / 255)&-1;
     $60 = (($srcG$1) + ($59))|0;
     $61 = Math_imul($41, $26)|0;
     $62 = (($61>>>0) / 255)&-1;
     $63 = (($srcB$1) + ($62))|0;
     $64 = Math_imul($42, $26)|0;
     $65 = (($64>>>0) / 255)&-1;
     $66 = (($65) + ($$4))|0;
     $dstA$0 = $66;$dstB$0 = $63;$dstG$0 = $60;$dstR$0 = $57;
     break;
    }
    case 32:  {
     $67 = (($srcR$1) + ($38))|0;
     $68 = ($67>>>0)>(255);
     $$ = $68 ? 255 : $67;
     $69 = (($srcG$1) + ($40))|0;
     $70 = ($69>>>0)>(255);
     $$3 = $70 ? 255 : $69;
     $71 = (($srcB$1) + ($41))|0;
     $72 = ($71>>>0)>(255);
     $$1 = $72 ? 255 : $71;
     $dstA$0 = $42;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $73 = Math_imul($srcR$1, $38)|0;
     $74 = (($73>>>0) / 255)&-1;
     $75 = Math_imul($srcG$1, $40)|0;
     $76 = (($75>>>0) / 255)&-1;
     $77 = Math_imul($srcB$1, $41)|0;
     $78 = (($77>>>0) / 255)&-1;
     $dstA$0 = $42;$dstB$0 = $78;$dstG$0 = $76;$dstR$0 = $74;
     break;
    }
    default: {
     $dstA$0 = $42;$dstB$0 = $41;$dstG$0 = $40;$dstR$0 = $38;
    }
    }
    $79 = $dstA$0 << 24;
    $80 = $dstR$0 << 16;
    $81 = $dstG$0 << 8;
    $82 = $81 | $80;
    $83 = $82 | $dstB$0;
    $84 = $83 | $79;
    HEAP32[$dst$06>>2] = $84;
    $85 = ((($src$05)) + 4|0);
    $86 = ((($dst$06)) + 4|0);
    $87 = ($29|0)==(0);
    if ($87) {
     break;
    } else {
     $$in = $29;$dst$06 = $86;$src$05 = $85;
    }
   }
   $$pre11 = HEAP32[$9>>2]|0;
   $95 = $$pre11;
  }
  $88 = HEAP32[$16>>2]|0;
  $89 = (($90) + ($88)|0);
  $91 = HEAP32[$17>>2]|0;
  $92 = (($93) + ($91)|0);
  $94 = (($95) + -1)|0;
  HEAP32[$9>>2] = $94;
  $96 = ($95|0)==(0);
  if ($96) {
   $$lcssa = $89;$$lcssa15 = $92;
   break;
  } else {
   $90 = $89;$93 = $92;$97 = $94;$98 = $92;$99 = $89;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa15;
 return;
}
function _SDL_Blit_BGR888_ARGB8888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$4 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre28 = 0, $$pre29 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$017 = 0, $dstA$0 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0;
 var $or$cond = 0, $posx$015 = 0, $posx$2 = 0, $posy$021 = 0, $posy$1$lcssa = 0, $src$016 = 0, $src$1 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$014 = 0, $srcx$2 = 0, $srcy$020 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = $3 >>> 24;
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $$4 = $29 ? 255 : $23;
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = ($$4|0)!=(255);
 $or$cond = $31 & $32;
 $33 = $1 & 112;
 $34 = $$4 ^ 255;
 $35 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $119 = $$pre;$124 = $21;$125 = $$pre;$44 = $19;$posy$021 = 0;$srcy$020 = 0;
 while(1) {
  $36 = ($posy$021|0)>(65535);
  if ($36) {
   $37 = (($posy$021) + -65536)|0;
   $38 = $37 >>> 16;
   $39 = $38 << 16;
   $40 = (($srcy$020) + 1)|0;
   $41 = (($37) - ($39))|0;
   $42 = (($40) + ($38))|0;
   $posy$1$lcssa = $41;$srcy$1$lcssa = $42;
  } else {
   $posy$1$lcssa = $posy$021;$srcy$1$lcssa = $srcy$020;
  }
  $43 = ($44|0)==(0);
  if ($43) {
   $121 = $124;
  } else {
   $$in = $44;$dst$017 = $125;$posx$015 = 65536;$src$016 = 0;$srcx$014 = -1;
   while(1) {
    $45 = (($$in) + -1)|0;
    $46 = ($posx$015|0)>(65535);
    if ($46) {
     $47 = (($posx$015) + -65536)|0;
     $48 = $47 >>> 16;
     $49 = $48 << 16;
     $50 = (($srcx$014) + 1)|0;
     $51 = (($47) - ($49))|0;
     $52 = (($50) + ($48))|0;
     $53 = HEAP32[$info>>2]|0;
     $54 = HEAP32[$35>>2]|0;
     $55 = Math_imul($54, $srcy$1$lcssa)|0;
     $56 = (($53) + ($55)|0);
     $57 = $52 << 2;
     $58 = (($56) + ($57)|0);
     $posx$2 = $51;$src$1 = $58;$srcx$2 = $52;
    } else {
     $posx$2 = $posx$015;$src$1 = $src$016;$srcx$2 = $srcx$014;
    }
    $59 = HEAP32[$src$1>>2]|0;
    $60 = $59 >>> 16;
    $61 = $60 & 255;
    $62 = $59 >>> 8;
    $63 = $62 & 255;
    $64 = $59 & 255;
    $65 = HEAP32[$dst$017>>2]|0;
    $66 = $65 >>> 16;
    $67 = $66 & 255;
    $68 = $65 >>> 8;
    $69 = $68 & 255;
    $70 = $65 & 255;
    $71 = $65 >>> 24;
    if ($27) {
     $srcB$0 = $61;$srcG$0 = $63;$srcR$0 = $64;
    } else {
     $72 = Math_imul($64, $4)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($63, $6)|0;
     $75 = (($74>>>0) / 255)&-1;
     $76 = Math_imul($61, $8)|0;
     $77 = (($76>>>0) / 255)&-1;
     $srcB$0 = $77;$srcG$0 = $75;$srcR$0 = $73;
    }
    if ($or$cond) {
     $78 = Math_imul($srcR$0, $$4)|0;
     $79 = (($78>>>0) / 255)&-1;
     $80 = Math_imul($srcG$0, $$4)|0;
     $81 = (($80>>>0) / 255)&-1;
     $82 = Math_imul($srcB$0, $$4)|0;
     $83 = (($82>>>0) / 255)&-1;
     $srcB$1 = $83;$srcG$1 = $81;$srcR$1 = $79;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($33|0) {
    case 16:  {
     $84 = Math_imul($67, $34)|0;
     $85 = (($84>>>0) / 255)&-1;
     $86 = (($srcR$1) + ($85))|0;
     $87 = Math_imul($69, $34)|0;
     $88 = (($87>>>0) / 255)&-1;
     $89 = (($srcG$1) + ($88))|0;
     $90 = Math_imul($70, $34)|0;
     $91 = (($90>>>0) / 255)&-1;
     $92 = (($srcB$1) + ($91))|0;
     $93 = Math_imul($71, $34)|0;
     $94 = (($93>>>0) / 255)&-1;
     $95 = (($94) + ($$4))|0;
     $dstA$0 = $95;$dstB$0 = $92;$dstG$0 = $89;$dstR$0 = $86;
     break;
    }
    case 32:  {
     $96 = (($srcR$1) + ($67))|0;
     $97 = ($96>>>0)>(255);
     $$ = $97 ? 255 : $96;
     $98 = (($srcG$1) + ($69))|0;
     $99 = ($98>>>0)>(255);
     $$3 = $99 ? 255 : $98;
     $100 = (($srcB$1) + ($70))|0;
     $101 = ($100>>>0)>(255);
     $$1 = $101 ? 255 : $100;
     $dstA$0 = $71;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $102 = Math_imul($srcR$1, $67)|0;
     $103 = (($102>>>0) / 255)&-1;
     $104 = Math_imul($srcG$1, $69)|0;
     $105 = (($104>>>0) / 255)&-1;
     $106 = Math_imul($srcB$1, $70)|0;
     $107 = (($106>>>0) / 255)&-1;
     $dstA$0 = $71;$dstB$0 = $107;$dstG$0 = $105;$dstR$0 = $103;
     break;
    }
    default: {
     $dstA$0 = $71;$dstB$0 = $70;$dstG$0 = $69;$dstR$0 = $67;
    }
    }
    $108 = $dstA$0 << 24;
    $109 = $dstR$0 << 16;
    $110 = $dstG$0 << 8;
    $111 = $110 | $109;
    $112 = $111 | $dstB$0;
    $113 = $112 | $108;
    HEAP32[$dst$017>>2] = $113;
    $114 = (($posx$2) + ($20))|0;
    $115 = ((($dst$017)) + 4|0);
    $116 = ($45|0)==(0);
    if ($116) {
     break;
    } else {
     $$in = $45;$dst$017 = $115;$posx$015 = $114;$src$016 = $src$1;$srcx$014 = $srcx$2;
    }
   }
   $$pre29 = HEAP32[$12>>2]|0;
   $121 = $$pre29;
  }
  $117 = HEAP32[$25>>2]|0;
  $118 = (($119) + ($117)|0);
  $120 = (($121) + -1)|0;
  HEAP32[$12>>2] = $120;
  $122 = ($121|0)==(0);
  if ($122) {
   $$lcssa = $118;
   break;
  }
  $123 = (($posy$1$lcssa) + ($14))|0;
  $$pre28 = HEAP32[$18>>2]|0;
  $119 = $118;$124 = $120;$125 = $118;$44 = $$pre28;$posy$021 = $123;$srcy$020 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ARGB8888_RGB888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0;
 var $posx$010 = 0, $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$47 = $$pre;$52 = $12;$53 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $49 = $52;
  } else {
   $$in = $25;$dst$013 = $53;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    $41 = $40 & 16777215;
    HEAP32[$dst$013>>2] = $41;
    $42 = (($posx$2) + ($11))|0;
    $43 = ((($dst$013)) + 4|0);
    $44 = ($26|0)==(0);
    if ($44) {
     break;
    } else {
     $$in = $26;$dst$013 = $43;$posx$010 = $42;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $49 = $$pre25;
  }
  $45 = HEAP32[$15>>2]|0;
  $46 = (($47) + ($45)|0);
  $48 = (($49) + -1)|0;
  HEAP32[$3>>2] = $48;
  $50 = ($49|0)==(0);
  if ($50) {
   $$lcssa = $46;
   break;
  }
  $51 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$47 = $46;$52 = $48;$53 = $46;$posy$016 = $51;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ARGB8888_RGB888_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$$us = 0, $$1 = 0, $$1$us = 0, $$3 = 0, $$3$us = 0, $$lcssa = 0, $$lcssa17 = 0, $$pre = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$05 = 0, $dst$05$us = 0, $dstB$0 = 0, $dstB$0$us = 0, $dstG$0 = 0, $dstG$0$us = 0, $dstR$0 = 0, $dstR$0$us = 0;
 var $src$04 = 0, $src$04$us = 0, $srcB$0$us = 0, $srcG$0$us = 0, $srcR$0$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 48;
 $11 = ($10|0)==(0);
 $12 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre10 = HEAP32[$6>>2]|0;
 $113 = $$pre;$116 = $$pre10;$120 = $$pre10;$121 = $$pre;
 while(1) {
  $13 = HEAP32[$7>>2]|0;
  $14 = (($13) + -1)|0;
  $15 = ($13|0)==(0);
  L6: do {
   if (!($15)) {
    if ($11) {
     $109 = $14;$dst$05 = $120;$src$04 = $121;
     while(1) {
      $67 = HEAP32[$src$04>>2]|0;
      $68 = $67 >>> 16;
      $69 = $68 & 255;
      $70 = $67 >>> 8;
      $71 = $70 & 255;
      $72 = $67 & 255;
      $73 = HEAP32[$dst$05>>2]|0;
      $74 = $73 >>> 16;
      $75 = $74 & 255;
      $76 = $73 >>> 8;
      $77 = $76 & 255;
      $78 = $73 & 255;
      switch ($12|0) {
      case 16:  {
       $79 = $67 >>> 24;
       $80 = $79 ^ 255;
       $81 = Math_imul($75, $80)|0;
       $82 = (($81>>>0) / 255)&-1;
       $83 = (($69) + ($82))|0;
       $84 = Math_imul($77, $80)|0;
       $85 = (($84>>>0) / 255)&-1;
       $86 = (($71) + ($85))|0;
       $87 = Math_imul($80, $78)|0;
       $88 = (($87>>>0) / 255)&-1;
       $89 = (($72) + ($88))|0;
       $dstB$0 = $89;$dstG$0 = $86;$dstR$0 = $83;
       break;
      }
      case 32:  {
       $90 = (($69) + ($75))|0;
       $91 = ($90>>>0)>(255);
       $$ = $91 ? 255 : $90;
       $92 = (($71) + ($77))|0;
       $93 = ($92>>>0)>(255);
       $$3 = $93 ? 255 : $92;
       $94 = (($72) + ($78))|0;
       $95 = ($94>>>0)>(255);
       $$1 = $95 ? 255 : $94;
       $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
       break;
      }
      case 64:  {
       $96 = Math_imul($69, $75)|0;
       $97 = (($96>>>0) / 255)&-1;
       $98 = Math_imul($71, $77)|0;
       $99 = (($98>>>0) / 255)&-1;
       $100 = Math_imul($72, $78)|0;
       $101 = (($100>>>0) / 255)&-1;
       $dstB$0 = $101;$dstG$0 = $99;$dstR$0 = $97;
       break;
      }
      default: {
       $dstB$0 = $78;$dstG$0 = $77;$dstR$0 = $75;
      }
      }
      $102 = $dstR$0 << 16;
      $103 = $dstG$0 << 8;
      $104 = $103 | $102;
      $105 = $104 | $dstB$0;
      HEAP32[$dst$05>>2] = $105;
      $106 = ((($src$04)) + 4|0);
      $107 = ((($dst$05)) + 4|0);
      $108 = (($109) + -1)|0;
      $110 = ($109|0)==(0);
      if ($110) {
       break L6;
      } else {
       $109 = $108;$dst$05 = $107;$src$04 = $106;
      }
     }
    } else {
     $65 = $14;$dst$05$us = $120;$src$04$us = $121;
    }
    while(1) {
     $16 = HEAP32[$src$04$us>>2]|0;
     $17 = $16 >>> 16;
     $18 = $17 & 255;
     $19 = $16 >>> 8;
     $20 = $19 & 255;
     $21 = $16 & 255;
     $22 = $16 >>> 24;
     $23 = HEAP32[$dst$05$us>>2]|0;
     $24 = $23 >>> 16;
     $25 = $24 & 255;
     $26 = $23 >>> 8;
     $27 = $26 & 255;
     $28 = $23 & 255;
     $29 = ($22|0)==(255);
     if ($29) {
      $srcB$0$us = $21;$srcG$0$us = $20;$srcR$0$us = $18;
     } else {
      $30 = Math_imul($18, $22)|0;
      $31 = (($30>>>0) / 255)&-1;
      $32 = Math_imul($20, $22)|0;
      $33 = (($32>>>0) / 255)&-1;
      $34 = Math_imul($21, $22)|0;
      $35 = (($34>>>0) / 255)&-1;
      $srcB$0$us = $35;$srcG$0$us = $33;$srcR$0$us = $31;
     }
     switch ($12|0) {
     case 16:  {
      $48 = $22 ^ 255;
      $49 = Math_imul($25, $48)|0;
      $50 = (($49>>>0) / 255)&-1;
      $51 = (($srcR$0$us) + ($50))|0;
      $52 = Math_imul($27, $48)|0;
      $53 = (($52>>>0) / 255)&-1;
      $54 = (($srcG$0$us) + ($53))|0;
      $55 = Math_imul($48, $28)|0;
      $56 = (($55>>>0) / 255)&-1;
      $57 = (($srcB$0$us) + ($56))|0;
      $dstB$0$us = $57;$dstG$0$us = $54;$dstR$0$us = $51;
      break;
     }
     case 32:  {
      $42 = (($srcR$0$us) + ($25))|0;
      $43 = ($42>>>0)>(255);
      $$$us = $43 ? 255 : $42;
      $44 = (($srcG$0$us) + ($27))|0;
      $45 = ($44>>>0)>(255);
      $$3$us = $45 ? 255 : $44;
      $46 = (($srcB$0$us) + ($28))|0;
      $47 = ($46>>>0)>(255);
      $$1$us = $47 ? 255 : $46;
      $dstB$0$us = $$1$us;$dstG$0$us = $$3$us;$dstR$0$us = $$$us;
      break;
     }
     case 64:  {
      $36 = Math_imul($srcR$0$us, $25)|0;
      $37 = (($36>>>0) / 255)&-1;
      $38 = Math_imul($srcG$0$us, $27)|0;
      $39 = (($38>>>0) / 255)&-1;
      $40 = Math_imul($srcB$0$us, $28)|0;
      $41 = (($40>>>0) / 255)&-1;
      $dstB$0$us = $41;$dstG$0$us = $39;$dstR$0$us = $37;
      break;
     }
     default: {
      $dstB$0$us = $28;$dstG$0$us = $27;$dstR$0$us = $25;
     }
     }
     $58 = $dstR$0$us << 16;
     $59 = $dstG$0$us << 8;
     $60 = $59 | $58;
     $61 = $60 | $dstB$0$us;
     HEAP32[$dst$05$us>>2] = $61;
     $62 = ((($src$04$us)) + 4|0);
     $63 = ((($dst$05$us)) + 4|0);
     $64 = (($65) + -1)|0;
     $66 = ($65|0)==(0);
     if ($66) {
      break L6;
     } else {
      $65 = $64;$dst$05$us = $63;$src$04$us = $62;
     }
    }
   }
  } while(0);
  $111 = HEAP32[$8>>2]|0;
  $112 = (($113) + ($111)|0);
  $114 = HEAP32[$9>>2]|0;
  $115 = (($116) + ($114)|0);
  $117 = HEAP32[$2>>2]|0;
  $118 = (($117) + -1)|0;
  HEAP32[$2>>2] = $118;
  $119 = ($117|0)==(0);
  if ($119) {
   $$lcssa = $112;$$lcssa17 = $115;
   break;
  } else {
   $113 = $112;$116 = $115;$120 = $115;$121 = $112;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa17;
 return;
}
function _SDL_Blit_ARGB8888_RGB888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0;
 var $src$015 = 0, $src$1 = 0, $srcB$0 = 0, $srcG$0 = 0, $srcR$0 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 48;
 $19 = ($18|0)!=(0);
 $20 = $1 & 112;
 $21 = ((($info)) + 12|0);
 $$pre = HEAP32[$16>>2]|0;
 $101 = $14;$102 = $$pre;$30 = $12;$96 = $$pre;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $22 = ($posy$020|0)>(65535);
  if ($22) {
   $23 = (($posy$020) + -65536)|0;
   $24 = $23 >>> 16;
   $25 = $24 << 16;
   $26 = (($srcy$019) + 1)|0;
   $27 = (($23) - ($25))|0;
   $28 = (($26) + ($24))|0;
   $posy$1$lcssa = $27;$srcy$1$lcssa = $28;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $29 = ($30|0)==(0);
  if ($29) {
   $98 = $101;
  } else {
   $$in = $30;$dst$016 = $102;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $31 = (($$in) + -1)|0;
    $32 = ($posx$014|0)>(65535);
    if ($32) {
     $33 = (($posx$014) + -65536)|0;
     $34 = $33 >>> 16;
     $35 = $34 << 16;
     $36 = (($srcx$013) + 1)|0;
     $37 = (($33) - ($35))|0;
     $38 = (($36) + ($34))|0;
     $39 = HEAP32[$info>>2]|0;
     $40 = HEAP32[$21>>2]|0;
     $41 = Math_imul($40, $srcy$1$lcssa)|0;
     $42 = (($39) + ($41)|0);
     $43 = $38 << 2;
     $44 = (($42) + ($43)|0);
     $posx$2 = $37;$src$1 = $44;$srcx$2 = $38;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $45 = HEAP32[$src$1>>2]|0;
    $46 = $45 >>> 16;
    $47 = $46 & 255;
    $48 = $45 >>> 8;
    $49 = $48 & 255;
    $50 = $45 & 255;
    $51 = $45 >>> 24;
    $52 = HEAP32[$dst$016>>2]|0;
    $53 = $52 >>> 16;
    $54 = $53 & 255;
    $55 = $52 >>> 8;
    $56 = $55 & 255;
    $57 = $52 & 255;
    $58 = ($51|0)!=(255);
    $or$cond = $19 & $58;
    if ($or$cond) {
     $59 = Math_imul($47, $51)|0;
     $60 = (($59>>>0) / 255)&-1;
     $61 = Math_imul($49, $51)|0;
     $62 = (($61>>>0) / 255)&-1;
     $63 = Math_imul($50, $51)|0;
     $64 = (($63>>>0) / 255)&-1;
     $srcB$0 = $64;$srcG$0 = $62;$srcR$0 = $60;
    } else {
     $srcB$0 = $50;$srcG$0 = $49;$srcR$0 = $47;
    }
    switch ($20|0) {
    case 16:  {
     $65 = $51 ^ 255;
     $66 = Math_imul($54, $65)|0;
     $67 = (($66>>>0) / 255)&-1;
     $68 = (($srcR$0) + ($67))|0;
     $69 = Math_imul($56, $65)|0;
     $70 = (($69>>>0) / 255)&-1;
     $71 = (($srcG$0) + ($70))|0;
     $72 = Math_imul($65, $57)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = (($srcB$0) + ($73))|0;
     $dstB$0 = $74;$dstG$0 = $71;$dstR$0 = $68;
     break;
    }
    case 32:  {
     $75 = (($srcR$0) + ($54))|0;
     $76 = ($75>>>0)>(255);
     $$ = $76 ? 255 : $75;
     $77 = (($srcG$0) + ($56))|0;
     $78 = ($77>>>0)>(255);
     $$3 = $78 ? 255 : $77;
     $79 = (($srcB$0) + ($57))|0;
     $80 = ($79>>>0)>(255);
     $$1 = $80 ? 255 : $79;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $81 = Math_imul($srcR$0, $54)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = Math_imul($srcG$0, $56)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = Math_imul($srcB$0, $57)|0;
     $86 = (($85>>>0) / 255)&-1;
     $dstB$0 = $86;$dstG$0 = $84;$dstR$0 = $82;
     break;
    }
    default: {
     $dstB$0 = $57;$dstG$0 = $56;$dstR$0 = $54;
    }
    }
    $87 = $dstR$0 << 16;
    $88 = $dstG$0 << 8;
    $89 = $88 | $87;
    $90 = $89 | $dstB$0;
    HEAP32[$dst$016>>2] = $90;
    $91 = (($posx$2) + ($13))|0;
    $92 = ((($dst$016)) + 4|0);
    $93 = ($31|0)==(0);
    if ($93) {
     break;
    } else {
     $$in = $31;$dst$016 = $92;$posx$014 = $91;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$5>>2]|0;
   $98 = $$pre28;
  }
  $94 = HEAP32[$17>>2]|0;
  $95 = (($96) + ($94)|0);
  $97 = (($98) + -1)|0;
  HEAP32[$5>>2] = $97;
  $99 = ($98|0)==(0);
  if ($99) {
   $$lcssa = $95;
   break;
  }
  $100 = (($posy$1$lcssa) + ($7))|0;
  $$pre27 = HEAP32[$11>>2]|0;
  $101 = $97;$102 = $95;$30 = $$pre27;$96 = $95;$posy$020 = $100;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ARGB8888_RGB888_Modulate($info) {
 $info = $info|0;
 var $$in = 0, $$in13 = 0, $$lcssa = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$pre16 = 0, $$pre17 = 0, $$pre18 = 0, $$pre19 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $8 = 0, $9 = 0, $dst$02 = 0, $dst$02$us$us = 0, $src$01 = 0, $src$01$us$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = ((($info)) + 20|0);
 $14 = ((($info)) + 24|0);
 $15 = ((($info)) + 12|0);
 $16 = ((($info)) + 32|0);
 $17 = $1 & 1;
 $18 = ($17|0)==(0);
 $$pre17 = HEAP32[$info>>2]|0;
 $$pre18 = HEAP32[$13>>2]|0;
 if ($18) {
  $23 = $$pre17;$26 = $$pre18;$67 = $11;$68 = $$pre18;$69 = $$pre17;
  while(1) {
   $19 = HEAP32[$14>>2]|0;
   $20 = ($19|0)==(0);
   if ($20) {
    $28 = $67;
   } else {
    $$in13 = $19;$dst$02$us$us = $68;$src$01$us$us = $69;
    while(1) {
     $30 = (($$in13) + -1)|0;
     $31 = HEAP32[$src$01$us$us>>2]|0;
     $32 = $31 & 16777215;
     HEAP32[$dst$02$us$us>>2] = $32;
     $33 = ((($src$01$us$us)) + 4|0);
     $34 = ((($dst$02$us$us)) + 4|0);
     $35 = ($30|0)==(0);
     if ($35) {
      break;
     } else {
      $$in13 = $30;$dst$02$us$us = $34;$src$01$us$us = $33;
     }
    }
    $$pre19 = HEAP32[$9>>2]|0;
    $28 = $$pre19;
   }
   $21 = HEAP32[$15>>2]|0;
   $22 = (($23) + ($21)|0);
   $24 = HEAP32[$16>>2]|0;
   $25 = (($26) + ($24)|0);
   $27 = (($28) + -1)|0;
   HEAP32[$9>>2] = $27;
   $29 = ($28|0)==(0);
   if ($29) {
    $$lcssa = $22;$$lcssa30 = $25;
    break;
   } else {
    $23 = $22;$26 = $25;$67 = $27;$68 = $25;$69 = $22;
   }
  }
  HEAP32[$info>>2] = $$lcssa;
  HEAP32[$13>>2] = $$lcssa30;
  return;
 } else {
  $60 = $$pre17;$63 = $$pre18;$70 = $11;$71 = $$pre18;$72 = $$pre17;
  while(1) {
   $36 = HEAP32[$14>>2]|0;
   $37 = ($36|0)==(0);
   if ($37) {
    $65 = $70;
   } else {
    $$in = $36;$dst$02 = $71;$src$01 = $72;
    while(1) {
     $38 = (($$in) + -1)|0;
     $39 = HEAP32[$src$01>>2]|0;
     $40 = $39 >>> 16;
     $41 = $40 & 255;
     $42 = $39 >>> 8;
     $43 = $42 & 255;
     $44 = $39 & 255;
     $45 = Math_imul($41, $4)|0;
     $46 = (($45>>>0) / 255)&-1;
     $47 = Math_imul($43, $6)|0;
     $48 = (($47>>>0) / 255)&-1;
     $49 = Math_imul($44, $8)|0;
     $50 = (($49>>>0) / 255)&-1;
     $51 = $46 << 16;
     $52 = $48 << 8;
     $53 = $52 | $51;
     $54 = $53 | $50;
     HEAP32[$dst$02>>2] = $54;
     $55 = ((($src$01)) + 4|0);
     $56 = ((($dst$02)) + 4|0);
     $57 = ($38|0)==(0);
     if ($57) {
      break;
     } else {
      $$in = $38;$dst$02 = $56;$src$01 = $55;
     }
    }
    $$pre16 = HEAP32[$9>>2]|0;
    $65 = $$pre16;
   }
   $58 = HEAP32[$15>>2]|0;
   $59 = (($60) + ($58)|0);
   $61 = HEAP32[$16>>2]|0;
   $62 = (($63) + ($61)|0);
   $64 = (($65) + -1)|0;
   HEAP32[$9>>2] = $64;
   $66 = ($65|0)==(0);
   if ($66) {
    $$lcssa31 = $59;$$lcssa32 = $62;
    break;
   } else {
    $60 = $59;$63 = $62;$70 = $64;$71 = $62;$72 = $59;
   }
  }
  HEAP32[$info>>2] = $$lcssa31;
  HEAP32[$13>>2] = $$lcssa32;
  return;
 }
}
function _SDL_Blit_ARGB8888_RGB888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $dst$013 = 0, $dst$013$us = 0, $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$017 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0, $src$1$us = 0, $srcx$010 = 0, $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$016 = 0;
 var $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = ((($info)) + 20|0);
 $24 = ((($info)) + 32|0);
 $25 = $1 & 1;
 $26 = ($25|0)==(0);
 $27 = ((($info)) + 12|0);
 $$pre = HEAP32[$23>>2]|0;
 $36 = $19;$94 = $$pre;$99 = $$pre;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $28 = ($posy$017|0)>(65535);
  if ($28) {
   $29 = (($posy$017) + -65536)|0;
   $30 = $29 >>> 16;
   $31 = $30 << 16;
   $32 = (($srcy$016) + 1)|0;
   $33 = (($29) - ($31))|0;
   $34 = (($32) + ($30))|0;
   $posy$1$lcssa = $33;$srcy$1$lcssa = $34;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $35 = (($36) + -1)|0;
  $37 = ($36|0)==(0);
  if (!($37)) {
   if ($26) {
    $56 = $35;$dst$013$us = $99;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $38 = ($posx$011$us|0)>(65535);
     if ($38) {
      $39 = (($posx$011$us) + -65536)|0;
      $40 = $39 >>> 16;
      $41 = $40 << 16;
      $42 = (($srcx$010$us) + 1)|0;
      $43 = (($39) - ($41))|0;
      $44 = (($42) + ($40))|0;
      $45 = HEAP32[$info>>2]|0;
      $46 = HEAP32[$27>>2]|0;
      $47 = Math_imul($46, $srcy$1$lcssa)|0;
      $48 = (($45) + ($47)|0);
      $49 = $44 << 2;
      $50 = (($48) + ($49)|0);
      $posx$2$us = $43;$src$1$us = $50;$srcx$2$us = $44;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $51 = HEAP32[$src$1$us>>2]|0;
     $52 = $51 & 16777215;
     HEAP32[$dst$013$us>>2] = $52;
     $53 = (($posx$2$us) + ($20))|0;
     $54 = ((($dst$013$us)) + 4|0);
     $55 = (($56) + -1)|0;
     $57 = ($56|0)==(0);
     if ($57) {
      break;
     } else {
      $56 = $55;$dst$013$us = $54;$posx$011$us = $53;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $90 = $35;$dst$013 = $99;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $58 = ($posx$011|0)>(65535);
     if ($58) {
      $59 = (($posx$011) + -65536)|0;
      $60 = $59 >>> 16;
      $61 = $60 << 16;
      $62 = (($srcx$010) + 1)|0;
      $63 = (($59) - ($61))|0;
      $64 = (($62) + ($60))|0;
      $65 = HEAP32[$info>>2]|0;
      $66 = HEAP32[$27>>2]|0;
      $67 = Math_imul($66, $srcy$1$lcssa)|0;
      $68 = (($65) + ($67)|0);
      $69 = $64 << 2;
      $70 = (($68) + ($69)|0);
      $posx$2 = $63;$src$1 = $70;$srcx$2 = $64;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $71 = HEAP32[$src$1>>2]|0;
     $72 = $71 >>> 16;
     $73 = $72 & 255;
     $74 = $71 >>> 8;
     $75 = $74 & 255;
     $76 = $71 & 255;
     $77 = Math_imul($73, $4)|0;
     $78 = (($77>>>0) / 255)&-1;
     $79 = Math_imul($75, $6)|0;
     $80 = (($79>>>0) / 255)&-1;
     $81 = Math_imul($76, $8)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = $78 << 16;
     $84 = $80 << 8;
     $85 = $82 | $84;
     $86 = $85 | $83;
     HEAP32[$dst$013>>2] = $86;
     $87 = (($posx$2) + ($20))|0;
     $88 = ((($dst$013)) + 4|0);
     $89 = (($90) + -1)|0;
     $91 = ($90|0)==(0);
     if ($91) {
      break;
     } else {
      $90 = $89;$dst$013 = $88;$posx$011 = $87;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $92 = HEAP32[$24>>2]|0;
  $93 = (($94) + ($92)|0);
  $95 = HEAP32[$12>>2]|0;
  $96 = (($95) + -1)|0;
  HEAP32[$12>>2] = $96;
  $97 = ($95|0)==(0);
  if ($97) {
   $$lcssa = $93;
   break;
  }
  $98 = (($posy$1$lcssa) + ($14))|0;
  $$pre27 = HEAP32[$18>>2]|0;
  $36 = $$pre27;$94 = $93;$99 = $93;$posy$017 = $98;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$23>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ARGB8888_RGB888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$lcssa14 = 0, $$pre = 0, $$pre10 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $dst$05 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $src$04 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + -1)|0;
 HEAP32[$10>>2] = $12;
 $13 = ($11|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre9 = HEAP32[$14>>2]|0;
 $87 = $$pre;$90 = $$pre9;$94 = $12;$95 = $$pre9;$96 = $$pre;
 while(1) {
  $25 = HEAP32[$15>>2]|0;
  $26 = ($25|0)==(0);
  if ($26) {
   $92 = $94;
  } else {
   $$in = $25;$dst$05 = $95;$src$04 = $96;
   while(1) {
    $27 = (($$in) + -1)|0;
    $28 = HEAP32[$src$04>>2]|0;
    $29 = $28 >>> 16;
    $30 = $29 & 255;
    $31 = $28 >>> 8;
    $32 = $31 & 255;
    $33 = $28 & 255;
    $34 = $28 >>> 24;
    $35 = HEAP32[$dst$05>>2]|0;
    $36 = $35 >>> 16;
    $37 = $36 & 255;
    $38 = $35 >>> 8;
    $39 = $38 & 255;
    $40 = $35 & 255;
    if ($19) {
     $srcB$0 = $33;$srcG$0 = $32;$srcR$0 = $30;
    } else {
     $41 = Math_imul($30, $4)|0;
     $42 = (($41>>>0) / 255)&-1;
     $43 = Math_imul($32, $6)|0;
     $44 = (($43>>>0) / 255)&-1;
     $45 = Math_imul($33, $8)|0;
     $46 = (($45>>>0) / 255)&-1;
     $srcB$0 = $46;$srcG$0 = $44;$srcR$0 = $42;
    }
    if ($21) {
     $srcA$0 = $34;
    } else {
     $47 = Math_imul($34, $9)|0;
     $48 = (($47>>>0) / 255)&-1;
     $srcA$0 = $48;
    }
    $49 = ($srcA$0>>>0)<(255);
    $or$cond = $23 & $49;
    if ($or$cond) {
     $50 = Math_imul($srcA$0, $srcR$0)|0;
     $51 = (($50>>>0) / 255)&-1;
     $52 = Math_imul($srcA$0, $srcG$0)|0;
     $53 = (($52>>>0) / 255)&-1;
     $54 = Math_imul($srcA$0, $srcB$0)|0;
     $55 = (($54>>>0) / 255)&-1;
     $srcB$1 = $55;$srcG$1 = $53;$srcR$1 = $51;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($24|0) {
    case 16:  {
     $56 = (255 - ($srcA$0))|0;
     $57 = Math_imul($56, $37)|0;
     $58 = (($57>>>0) / 255)&-1;
     $59 = (($srcR$1) + ($58))|0;
     $60 = Math_imul($56, $39)|0;
     $61 = (($60>>>0) / 255)&-1;
     $62 = (($srcG$1) + ($61))|0;
     $63 = Math_imul($56, $40)|0;
     $64 = (($63>>>0) / 255)&-1;
     $65 = (($srcB$1) + ($64))|0;
     $dstB$0 = $65;$dstG$0 = $62;$dstR$0 = $59;
     break;
    }
    case 32:  {
     $66 = (($srcR$1) + ($37))|0;
     $67 = ($66>>>0)>(255);
     $$ = $67 ? 255 : $66;
     $68 = (($srcG$1) + ($39))|0;
     $69 = ($68>>>0)>(255);
     $$3 = $69 ? 255 : $68;
     $70 = (($srcB$1) + ($40))|0;
     $71 = ($70>>>0)>(255);
     $$1 = $71 ? 255 : $70;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $72 = Math_imul($srcR$1, $37)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($srcG$1, $39)|0;
     $75 = (($74>>>0) / 255)&-1;
     $76 = Math_imul($srcB$1, $40)|0;
     $77 = (($76>>>0) / 255)&-1;
     $dstB$0 = $77;$dstG$0 = $75;$dstR$0 = $73;
     break;
    }
    default: {
     $dstB$0 = $40;$dstG$0 = $39;$dstR$0 = $37;
    }
    }
    $78 = $dstR$0 << 16;
    $79 = $dstG$0 << 8;
    $80 = $79 | $78;
    $81 = $80 | $dstB$0;
    HEAP32[$dst$05>>2] = $81;
    $82 = ((($src$04)) + 4|0);
    $83 = ((($dst$05)) + 4|0);
    $84 = ($27|0)==(0);
    if ($84) {
     break;
    } else {
     $$in = $27;$dst$05 = $83;$src$04 = $82;
    }
   }
   $$pre10 = HEAP32[$10>>2]|0;
   $92 = $$pre10;
  }
  $85 = HEAP32[$16>>2]|0;
  $86 = (($87) + ($85)|0);
  $88 = HEAP32[$17>>2]|0;
  $89 = (($90) + ($88)|0);
  $91 = (($92) + -1)|0;
  HEAP32[$10>>2] = $91;
  $93 = ($92|0)==(0);
  if ($93) {
   $$lcssa = $86;$$lcssa14 = $89;
   break;
  } else {
   $87 = $86;$90 = $89;$94 = $91;$95 = $89;$96 = $86;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa14;
 return;
}
function _SDL_Blit_ARGB8888_RGB888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0;
 var $src$015 = 0, $src$1 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 << 16;
 $13 = ((($info)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12|0) / ($14|0))&-1;
 $16 = ((($info)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 << 16;
 $19 = ((($info)) + 24|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($18|0) / ($20|0))&-1;
 $22 = (($14) + -1)|0;
 HEAP32[$13>>2] = $22;
 $23 = ($14|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = $1 & 112;
 $33 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $116 = $$pre;$121 = $22;$122 = $$pre;$42 = $20;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $34 = ($posy$020|0)>(65535);
  if ($34) {
   $35 = (($posy$020) + -65536)|0;
   $36 = $35 >>> 16;
   $37 = $36 << 16;
   $38 = (($srcy$019) + 1)|0;
   $39 = (($35) - ($37))|0;
   $40 = (($38) + ($36))|0;
   $posy$1$lcssa = $39;$srcy$1$lcssa = $40;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $41 = ($42|0)==(0);
  if ($41) {
   $118 = $121;
  } else {
   $$in = $42;$dst$016 = $122;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $43 = (($$in) + -1)|0;
    $44 = ($posx$014|0)>(65535);
    if ($44) {
     $45 = (($posx$014) + -65536)|0;
     $46 = $45 >>> 16;
     $47 = $46 << 16;
     $48 = (($srcx$013) + 1)|0;
     $49 = (($45) - ($47))|0;
     $50 = (($48) + ($46))|0;
     $51 = HEAP32[$info>>2]|0;
     $52 = HEAP32[$33>>2]|0;
     $53 = Math_imul($52, $srcy$1$lcssa)|0;
     $54 = (($51) + ($53)|0);
     $55 = $50 << 2;
     $56 = (($54) + ($55)|0);
     $posx$2 = $49;$src$1 = $56;$srcx$2 = $50;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $57 = HEAP32[$src$1>>2]|0;
    $58 = $57 >>> 16;
    $59 = $58 & 255;
    $60 = $57 >>> 8;
    $61 = $60 & 255;
    $62 = $57 & 255;
    $63 = $57 >>> 24;
    $64 = HEAP32[$dst$016>>2]|0;
    $65 = $64 >>> 16;
    $66 = $65 & 255;
    $67 = $64 >>> 8;
    $68 = $67 & 255;
    $69 = $64 & 255;
    if ($27) {
     $srcB$0 = $62;$srcG$0 = $61;$srcR$0 = $59;
    } else {
     $70 = Math_imul($59, $4)|0;
     $71 = (($70>>>0) / 255)&-1;
     $72 = Math_imul($61, $6)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($62, $8)|0;
     $75 = (($74>>>0) / 255)&-1;
     $srcB$0 = $75;$srcG$0 = $73;$srcR$0 = $71;
    }
    if ($29) {
     $srcA$0 = $63;
    } else {
     $76 = Math_imul($63, $9)|0;
     $77 = (($76>>>0) / 255)&-1;
     $srcA$0 = $77;
    }
    $78 = ($srcA$0>>>0)<(255);
    $or$cond = $31 & $78;
    if ($or$cond) {
     $79 = Math_imul($srcA$0, $srcR$0)|0;
     $80 = (($79>>>0) / 255)&-1;
     $81 = Math_imul($srcA$0, $srcG$0)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = Math_imul($srcA$0, $srcB$0)|0;
     $84 = (($83>>>0) / 255)&-1;
     $srcB$1 = $84;$srcG$1 = $82;$srcR$1 = $80;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($32|0) {
    case 16:  {
     $85 = (255 - ($srcA$0))|0;
     $86 = Math_imul($85, $66)|0;
     $87 = (($86>>>0) / 255)&-1;
     $88 = (($srcR$1) + ($87))|0;
     $89 = Math_imul($85, $68)|0;
     $90 = (($89>>>0) / 255)&-1;
     $91 = (($srcG$1) + ($90))|0;
     $92 = Math_imul($85, $69)|0;
     $93 = (($92>>>0) / 255)&-1;
     $94 = (($srcB$1) + ($93))|0;
     $dstB$0 = $94;$dstG$0 = $91;$dstR$0 = $88;
     break;
    }
    case 32:  {
     $95 = (($srcR$1) + ($66))|0;
     $96 = ($95>>>0)>(255);
     $$ = $96 ? 255 : $95;
     $97 = (($srcG$1) + ($68))|0;
     $98 = ($97>>>0)>(255);
     $$3 = $98 ? 255 : $97;
     $99 = (($srcB$1) + ($69))|0;
     $100 = ($99>>>0)>(255);
     $$1 = $100 ? 255 : $99;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $101 = Math_imul($srcR$1, $66)|0;
     $102 = (($101>>>0) / 255)&-1;
     $103 = Math_imul($srcG$1, $68)|0;
     $104 = (($103>>>0) / 255)&-1;
     $105 = Math_imul($srcB$1, $69)|0;
     $106 = (($105>>>0) / 255)&-1;
     $dstB$0 = $106;$dstG$0 = $104;$dstR$0 = $102;
     break;
    }
    default: {
     $dstB$0 = $69;$dstG$0 = $68;$dstR$0 = $66;
    }
    }
    $107 = $dstR$0 << 16;
    $108 = $dstG$0 << 8;
    $109 = $108 | $107;
    $110 = $109 | $dstB$0;
    HEAP32[$dst$016>>2] = $110;
    $111 = (($posx$2) + ($21))|0;
    $112 = ((($dst$016)) + 4|0);
    $113 = ($43|0)==(0);
    if ($113) {
     break;
    } else {
     $$in = $43;$dst$016 = $112;$posx$014 = $111;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$13>>2]|0;
   $118 = $$pre28;
  }
  $114 = HEAP32[$25>>2]|0;
  $115 = (($116) + ($114)|0);
  $117 = (($118) + -1)|0;
  HEAP32[$13>>2] = $117;
  $119 = ($118|0)==(0);
  if ($119) {
   $$lcssa = $115;
   break;
  }
  $120 = (($posy$1$lcssa) + ($15))|0;
  $$pre27 = HEAP32[$19>>2]|0;
  $116 = $115;$121 = $117;$122 = $115;$42 = $$pre27;$posy$020 = $120;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ARGB8888_BGR888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0, $posx$010 = 0, $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$53 = $$pre;$58 = $12;$59 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $55 = $58;
  } else {
   $$in = $25;$dst$013 = $59;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    $41 = $40 >>> 16;
    $42 = $41 & 255;
    $43 = $40 & 65280;
    $44 = $40 << 16;
    $45 = $44 & 16711680;
    $46 = $45 | $43;
    $47 = $46 | $42;
    HEAP32[$dst$013>>2] = $47;
    $48 = (($posx$2) + ($11))|0;
    $49 = ((($dst$013)) + 4|0);
    $50 = ($26|0)==(0);
    if ($50) {
     break;
    } else {
     $$in = $26;$dst$013 = $49;$posx$010 = $48;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $55 = $$pre25;
  }
  $51 = HEAP32[$15>>2]|0;
  $52 = (($53) + ($51)|0);
  $54 = (($55) + -1)|0;
  HEAP32[$3>>2] = $54;
  $56 = ($55|0)==(0);
  if ($56) {
   $$lcssa = $52;
   break;
  }
  $57 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$53 = $52;$58 = $54;$59 = $52;$posy$016 = $57;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ARGB8888_BGR888_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$$us = 0, $$1 = 0, $$1$us = 0, $$3 = 0, $$3$us = 0, $$lcssa = 0, $$lcssa17 = 0, $$pre = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$05 = 0, $dst$05$us = 0, $dstB$0 = 0, $dstB$0$us = 0, $dstG$0 = 0, $dstG$0$us = 0, $dstR$0 = 0, $dstR$0$us = 0;
 var $src$04 = 0, $src$04$us = 0, $srcB$0$us = 0, $srcG$0$us = 0, $srcR$0$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 48;
 $11 = ($10|0)==(0);
 $12 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre10 = HEAP32[$6>>2]|0;
 $113 = $$pre;$116 = $$pre10;$120 = $$pre10;$121 = $$pre;
 while(1) {
  $13 = HEAP32[$7>>2]|0;
  $14 = (($13) + -1)|0;
  $15 = ($13|0)==(0);
  L6: do {
   if (!($15)) {
    if ($11) {
     $109 = $14;$dst$05 = $120;$src$04 = $121;
     while(1) {
      $67 = HEAP32[$src$04>>2]|0;
      $68 = $67 >>> 16;
      $69 = $68 & 255;
      $70 = $67 >>> 8;
      $71 = $70 & 255;
      $72 = $67 & 255;
      $73 = HEAP32[$dst$05>>2]|0;
      $74 = $73 >>> 16;
      $75 = $74 & 255;
      $76 = $73 >>> 8;
      $77 = $76 & 255;
      $78 = $73 & 255;
      switch ($12|0) {
      case 16:  {
       $79 = $67 >>> 24;
       $80 = $79 ^ 255;
       $81 = Math_imul($80, $78)|0;
       $82 = (($81>>>0) / 255)&-1;
       $83 = (($69) + ($82))|0;
       $84 = Math_imul($77, $80)|0;
       $85 = (($84>>>0) / 255)&-1;
       $86 = (($71) + ($85))|0;
       $87 = Math_imul($75, $80)|0;
       $88 = (($87>>>0) / 255)&-1;
       $89 = (($72) + ($88))|0;
       $dstB$0 = $89;$dstG$0 = $86;$dstR$0 = $83;
       break;
      }
      case 32:  {
       $90 = (($69) + ($78))|0;
       $91 = ($90>>>0)>(255);
       $$ = $91 ? 255 : $90;
       $92 = (($71) + ($77))|0;
       $93 = ($92>>>0)>(255);
       $$3 = $93 ? 255 : $92;
       $94 = (($72) + ($75))|0;
       $95 = ($94>>>0)>(255);
       $$1 = $95 ? 255 : $94;
       $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
       break;
      }
      case 64:  {
       $96 = Math_imul($69, $78)|0;
       $97 = (($96>>>0) / 255)&-1;
       $98 = Math_imul($71, $77)|0;
       $99 = (($98>>>0) / 255)&-1;
       $100 = Math_imul($72, $75)|0;
       $101 = (($100>>>0) / 255)&-1;
       $dstB$0 = $101;$dstG$0 = $99;$dstR$0 = $97;
       break;
      }
      default: {
       $dstB$0 = $75;$dstG$0 = $77;$dstR$0 = $78;
      }
      }
      $102 = $dstB$0 << 16;
      $103 = $dstG$0 << 8;
      $104 = $103 | $dstR$0;
      $105 = $104 | $102;
      HEAP32[$dst$05>>2] = $105;
      $106 = ((($src$04)) + 4|0);
      $107 = ((($dst$05)) + 4|0);
      $108 = (($109) + -1)|0;
      $110 = ($109|0)==(0);
      if ($110) {
       break L6;
      } else {
       $109 = $108;$dst$05 = $107;$src$04 = $106;
      }
     }
    } else {
     $65 = $14;$dst$05$us = $120;$src$04$us = $121;
    }
    while(1) {
     $16 = HEAP32[$src$04$us>>2]|0;
     $17 = $16 >>> 16;
     $18 = $17 & 255;
     $19 = $16 >>> 8;
     $20 = $19 & 255;
     $21 = $16 & 255;
     $22 = $16 >>> 24;
     $23 = HEAP32[$dst$05$us>>2]|0;
     $24 = $23 >>> 16;
     $25 = $24 & 255;
     $26 = $23 >>> 8;
     $27 = $26 & 255;
     $28 = $23 & 255;
     $29 = ($22|0)==(255);
     if ($29) {
      $srcB$0$us = $21;$srcG$0$us = $20;$srcR$0$us = $18;
     } else {
      $30 = Math_imul($18, $22)|0;
      $31 = (($30>>>0) / 255)&-1;
      $32 = Math_imul($20, $22)|0;
      $33 = (($32>>>0) / 255)&-1;
      $34 = Math_imul($21, $22)|0;
      $35 = (($34>>>0) / 255)&-1;
      $srcB$0$us = $35;$srcG$0$us = $33;$srcR$0$us = $31;
     }
     switch ($12|0) {
     case 16:  {
      $48 = $22 ^ 255;
      $49 = Math_imul($48, $28)|0;
      $50 = (($49>>>0) / 255)&-1;
      $51 = (($srcR$0$us) + ($50))|0;
      $52 = Math_imul($27, $48)|0;
      $53 = (($52>>>0) / 255)&-1;
      $54 = (($srcG$0$us) + ($53))|0;
      $55 = Math_imul($25, $48)|0;
      $56 = (($55>>>0) / 255)&-1;
      $57 = (($srcB$0$us) + ($56))|0;
      $dstB$0$us = $57;$dstG$0$us = $54;$dstR$0$us = $51;
      break;
     }
     case 32:  {
      $42 = (($srcR$0$us) + ($28))|0;
      $43 = ($42>>>0)>(255);
      $$$us = $43 ? 255 : $42;
      $44 = (($srcG$0$us) + ($27))|0;
      $45 = ($44>>>0)>(255);
      $$3$us = $45 ? 255 : $44;
      $46 = (($srcB$0$us) + ($25))|0;
      $47 = ($46>>>0)>(255);
      $$1$us = $47 ? 255 : $46;
      $dstB$0$us = $$1$us;$dstG$0$us = $$3$us;$dstR$0$us = $$$us;
      break;
     }
     case 64:  {
      $36 = Math_imul($srcR$0$us, $28)|0;
      $37 = (($36>>>0) / 255)&-1;
      $38 = Math_imul($srcG$0$us, $27)|0;
      $39 = (($38>>>0) / 255)&-1;
      $40 = Math_imul($srcB$0$us, $25)|0;
      $41 = (($40>>>0) / 255)&-1;
      $dstB$0$us = $41;$dstG$0$us = $39;$dstR$0$us = $37;
      break;
     }
     default: {
      $dstB$0$us = $25;$dstG$0$us = $27;$dstR$0$us = $28;
     }
     }
     $58 = $dstB$0$us << 16;
     $59 = $dstG$0$us << 8;
     $60 = $59 | $dstR$0$us;
     $61 = $60 | $58;
     HEAP32[$dst$05$us>>2] = $61;
     $62 = ((($src$04$us)) + 4|0);
     $63 = ((($dst$05$us)) + 4|0);
     $64 = (($65) + -1)|0;
     $66 = ($65|0)==(0);
     if ($66) {
      break L6;
     } else {
      $65 = $64;$dst$05$us = $63;$src$04$us = $62;
     }
    }
   }
  } while(0);
  $111 = HEAP32[$8>>2]|0;
  $112 = (($113) + ($111)|0);
  $114 = HEAP32[$9>>2]|0;
  $115 = (($116) + ($114)|0);
  $117 = HEAP32[$2>>2]|0;
  $118 = (($117) + -1)|0;
  HEAP32[$2>>2] = $118;
  $119 = ($117|0)==(0);
  if ($119) {
   $$lcssa = $112;$$lcssa17 = $115;
   break;
  } else {
   $113 = $112;$116 = $115;$120 = $115;$121 = $112;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa17;
 return;
}
function _SDL_Blit_ARGB8888_BGR888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0;
 var $src$015 = 0, $src$1 = 0, $srcB$0 = 0, $srcG$0 = 0, $srcR$0 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 48;
 $19 = ($18|0)!=(0);
 $20 = $1 & 112;
 $21 = ((($info)) + 12|0);
 $$pre = HEAP32[$16>>2]|0;
 $101 = $14;$102 = $$pre;$30 = $12;$96 = $$pre;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $22 = ($posy$020|0)>(65535);
  if ($22) {
   $23 = (($posy$020) + -65536)|0;
   $24 = $23 >>> 16;
   $25 = $24 << 16;
   $26 = (($srcy$019) + 1)|0;
   $27 = (($23) - ($25))|0;
   $28 = (($26) + ($24))|0;
   $posy$1$lcssa = $27;$srcy$1$lcssa = $28;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $29 = ($30|0)==(0);
  if ($29) {
   $98 = $101;
  } else {
   $$in = $30;$dst$016 = $102;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $31 = (($$in) + -1)|0;
    $32 = ($posx$014|0)>(65535);
    if ($32) {
     $33 = (($posx$014) + -65536)|0;
     $34 = $33 >>> 16;
     $35 = $34 << 16;
     $36 = (($srcx$013) + 1)|0;
     $37 = (($33) - ($35))|0;
     $38 = (($36) + ($34))|0;
     $39 = HEAP32[$info>>2]|0;
     $40 = HEAP32[$21>>2]|0;
     $41 = Math_imul($40, $srcy$1$lcssa)|0;
     $42 = (($39) + ($41)|0);
     $43 = $38 << 2;
     $44 = (($42) + ($43)|0);
     $posx$2 = $37;$src$1 = $44;$srcx$2 = $38;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $45 = HEAP32[$src$1>>2]|0;
    $46 = $45 >>> 16;
    $47 = $46 & 255;
    $48 = $45 >>> 8;
    $49 = $48 & 255;
    $50 = $45 & 255;
    $51 = $45 >>> 24;
    $52 = HEAP32[$dst$016>>2]|0;
    $53 = $52 >>> 16;
    $54 = $53 & 255;
    $55 = $52 >>> 8;
    $56 = $55 & 255;
    $57 = $52 & 255;
    $58 = ($51|0)!=(255);
    $or$cond = $19 & $58;
    if ($or$cond) {
     $59 = Math_imul($47, $51)|0;
     $60 = (($59>>>0) / 255)&-1;
     $61 = Math_imul($49, $51)|0;
     $62 = (($61>>>0) / 255)&-1;
     $63 = Math_imul($50, $51)|0;
     $64 = (($63>>>0) / 255)&-1;
     $srcB$0 = $64;$srcG$0 = $62;$srcR$0 = $60;
    } else {
     $srcB$0 = $50;$srcG$0 = $49;$srcR$0 = $47;
    }
    switch ($20|0) {
    case 16:  {
     $65 = $51 ^ 255;
     $66 = Math_imul($65, $57)|0;
     $67 = (($66>>>0) / 255)&-1;
     $68 = (($srcR$0) + ($67))|0;
     $69 = Math_imul($56, $65)|0;
     $70 = (($69>>>0) / 255)&-1;
     $71 = (($srcG$0) + ($70))|0;
     $72 = Math_imul($54, $65)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = (($srcB$0) + ($73))|0;
     $dstB$0 = $74;$dstG$0 = $71;$dstR$0 = $68;
     break;
    }
    case 32:  {
     $75 = (($srcR$0) + ($57))|0;
     $76 = ($75>>>0)>(255);
     $$ = $76 ? 255 : $75;
     $77 = (($srcG$0) + ($56))|0;
     $78 = ($77>>>0)>(255);
     $$3 = $78 ? 255 : $77;
     $79 = (($srcB$0) + ($54))|0;
     $80 = ($79>>>0)>(255);
     $$1 = $80 ? 255 : $79;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $81 = Math_imul($srcR$0, $57)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = Math_imul($srcG$0, $56)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = Math_imul($srcB$0, $54)|0;
     $86 = (($85>>>0) / 255)&-1;
     $dstB$0 = $86;$dstG$0 = $84;$dstR$0 = $82;
     break;
    }
    default: {
     $dstB$0 = $54;$dstG$0 = $56;$dstR$0 = $57;
    }
    }
    $87 = $dstB$0 << 16;
    $88 = $dstG$0 << 8;
    $89 = $88 | $dstR$0;
    $90 = $89 | $87;
    HEAP32[$dst$016>>2] = $90;
    $91 = (($posx$2) + ($13))|0;
    $92 = ((($dst$016)) + 4|0);
    $93 = ($31|0)==(0);
    if ($93) {
     break;
    } else {
     $$in = $31;$dst$016 = $92;$posx$014 = $91;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$5>>2]|0;
   $98 = $$pre28;
  }
  $94 = HEAP32[$17>>2]|0;
  $95 = (($96) + ($94)|0);
  $97 = (($98) + -1)|0;
  HEAP32[$5>>2] = $97;
  $99 = ($98|0)==(0);
  if ($99) {
   $$lcssa = $95;
   break;
  }
  $100 = (($posy$1$lcssa) + ($7))|0;
  $$pre27 = HEAP32[$11>>2]|0;
  $101 = $97;$102 = $95;$30 = $$pre27;$96 = $95;$posy$020 = $100;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ARGB8888_BGR888_Modulate($info) {
 $info = $info|0;
 var $$in = 0, $$in13 = 0, $$lcssa = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$pre16 = 0, $$pre17 = 0, $$pre18 = 0, $$pre19 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $dst$02 = 0, $dst$02$us$us = 0, $src$01 = 0, $src$01$us$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = ((($info)) + 20|0);
 $14 = ((($info)) + 24|0);
 $15 = ((($info)) + 12|0);
 $16 = ((($info)) + 32|0);
 $17 = $1 & 1;
 $18 = ($17|0)==(0);
 $$pre17 = HEAP32[$info>>2]|0;
 $$pre18 = HEAP32[$13>>2]|0;
 if ($18) {
  $23 = $$pre17;$26 = $$pre18;$73 = $11;$74 = $$pre18;$75 = $$pre17;
  while(1) {
   $19 = HEAP32[$14>>2]|0;
   $20 = ($19|0)==(0);
   if ($20) {
    $28 = $73;
   } else {
    $$in13 = $19;$dst$02$us$us = $74;$src$01$us$us = $75;
    while(1) {
     $30 = (($$in13) + -1)|0;
     $31 = HEAP32[$src$01$us$us>>2]|0;
     $32 = $31 >>> 16;
     $33 = $32 & 255;
     $34 = $31 & 65280;
     $35 = $31 << 16;
     $36 = $35 & 16711680;
     $37 = $34 | $33;
     $38 = $37 | $36;
     HEAP32[$dst$02$us$us>>2] = $38;
     $39 = ((($src$01$us$us)) + 4|0);
     $40 = ((($dst$02$us$us)) + 4|0);
     $41 = ($30|0)==(0);
     if ($41) {
      break;
     } else {
      $$in13 = $30;$dst$02$us$us = $40;$src$01$us$us = $39;
     }
    }
    $$pre19 = HEAP32[$9>>2]|0;
    $28 = $$pre19;
   }
   $21 = HEAP32[$15>>2]|0;
   $22 = (($23) + ($21)|0);
   $24 = HEAP32[$16>>2]|0;
   $25 = (($26) + ($24)|0);
   $27 = (($28) + -1)|0;
   HEAP32[$9>>2] = $27;
   $29 = ($28|0)==(0);
   if ($29) {
    $$lcssa = $22;$$lcssa30 = $25;
    break;
   } else {
    $23 = $22;$26 = $25;$73 = $27;$74 = $25;$75 = $22;
   }
  }
  HEAP32[$info>>2] = $$lcssa;
  HEAP32[$13>>2] = $$lcssa30;
  return;
 } else {
  $66 = $$pre17;$69 = $$pre18;$76 = $11;$77 = $$pre18;$78 = $$pre17;
  while(1) {
   $42 = HEAP32[$14>>2]|0;
   $43 = ($42|0)==(0);
   if ($43) {
    $71 = $76;
   } else {
    $$in = $42;$dst$02 = $77;$src$01 = $78;
    while(1) {
     $44 = (($$in) + -1)|0;
     $45 = HEAP32[$src$01>>2]|0;
     $46 = $45 >>> 16;
     $47 = $46 & 255;
     $48 = $45 >>> 8;
     $49 = $48 & 255;
     $50 = $45 & 255;
     $51 = Math_imul($47, $4)|0;
     $52 = (($51>>>0) / 255)&-1;
     $53 = Math_imul($49, $6)|0;
     $54 = (($53>>>0) / 255)&-1;
     $55 = Math_imul($50, $8)|0;
     $56 = (($55>>>0) / 255)&-1;
     $57 = $56 << 16;
     $58 = $54 << 8;
     $59 = $58 | $52;
     $60 = $59 | $57;
     HEAP32[$dst$02>>2] = $60;
     $61 = ((($src$01)) + 4|0);
     $62 = ((($dst$02)) + 4|0);
     $63 = ($44|0)==(0);
     if ($63) {
      break;
     } else {
      $$in = $44;$dst$02 = $62;$src$01 = $61;
     }
    }
    $$pre16 = HEAP32[$9>>2]|0;
    $71 = $$pre16;
   }
   $64 = HEAP32[$15>>2]|0;
   $65 = (($66) + ($64)|0);
   $67 = HEAP32[$16>>2]|0;
   $68 = (($69) + ($67)|0);
   $70 = (($71) + -1)|0;
   HEAP32[$9>>2] = $70;
   $72 = ($71|0)==(0);
   if ($72) {
    $$lcssa31 = $65;$$lcssa32 = $68;
    break;
   } else {
    $66 = $65;$69 = $68;$76 = $70;$77 = $68;$78 = $65;
   }
  }
  HEAP32[$info>>2] = $$lcssa31;
  HEAP32[$13>>2] = $$lcssa32;
  return;
 }
}
function _SDL_Blit_ARGB8888_BGR888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$013 = 0, $dst$013$us = 0, $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$017 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0;
 var $src$1$us = 0, $srcx$010 = 0, $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$016 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = ((($info)) + 20|0);
 $24 = ((($info)) + 32|0);
 $25 = $1 & 1;
 $26 = ($25|0)==(0);
 $27 = ((($info)) + 12|0);
 $$pre = HEAP32[$23>>2]|0;
 $100 = $$pre;$105 = $$pre;$36 = $19;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $28 = ($posy$017|0)>(65535);
  if ($28) {
   $29 = (($posy$017) + -65536)|0;
   $30 = $29 >>> 16;
   $31 = $30 << 16;
   $32 = (($srcy$016) + 1)|0;
   $33 = (($29) - ($31))|0;
   $34 = (($32) + ($30))|0;
   $posy$1$lcssa = $33;$srcy$1$lcssa = $34;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $35 = (($36) + -1)|0;
  $37 = ($36|0)==(0);
  if (!($37)) {
   if ($26) {
    $62 = $35;$dst$013$us = $105;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $38 = ($posx$011$us|0)>(65535);
     if ($38) {
      $39 = (($posx$011$us) + -65536)|0;
      $40 = $39 >>> 16;
      $41 = $40 << 16;
      $42 = (($srcx$010$us) + 1)|0;
      $43 = (($39) - ($41))|0;
      $44 = (($42) + ($40))|0;
      $45 = HEAP32[$info>>2]|0;
      $46 = HEAP32[$27>>2]|0;
      $47 = Math_imul($46, $srcy$1$lcssa)|0;
      $48 = (($45) + ($47)|0);
      $49 = $44 << 2;
      $50 = (($48) + ($49)|0);
      $posx$2$us = $43;$src$1$us = $50;$srcx$2$us = $44;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $51 = HEAP32[$src$1$us>>2]|0;
     $52 = $51 >>> 16;
     $53 = $52 & 255;
     $54 = $51 & 65280;
     $55 = $51 << 16;
     $56 = $55 & 16711680;
     $57 = $56 | $54;
     $58 = $57 | $53;
     HEAP32[$dst$013$us>>2] = $58;
     $59 = (($posx$2$us) + ($20))|0;
     $60 = ((($dst$013$us)) + 4|0);
     $61 = (($62) + -1)|0;
     $63 = ($62|0)==(0);
     if ($63) {
      break;
     } else {
      $62 = $61;$dst$013$us = $60;$posx$011$us = $59;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $96 = $35;$dst$013 = $105;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $64 = ($posx$011|0)>(65535);
     if ($64) {
      $65 = (($posx$011) + -65536)|0;
      $66 = $65 >>> 16;
      $67 = $66 << 16;
      $68 = (($srcx$010) + 1)|0;
      $69 = (($65) - ($67))|0;
      $70 = (($68) + ($66))|0;
      $71 = HEAP32[$info>>2]|0;
      $72 = HEAP32[$27>>2]|0;
      $73 = Math_imul($72, $srcy$1$lcssa)|0;
      $74 = (($71) + ($73)|0);
      $75 = $70 << 2;
      $76 = (($74) + ($75)|0);
      $posx$2 = $69;$src$1 = $76;$srcx$2 = $70;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $77 = HEAP32[$src$1>>2]|0;
     $78 = $77 >>> 16;
     $79 = $78 & 255;
     $80 = $77 >>> 8;
     $81 = $80 & 255;
     $82 = $77 & 255;
     $83 = Math_imul($79, $4)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = Math_imul($81, $6)|0;
     $86 = (($85>>>0) / 255)&-1;
     $87 = Math_imul($82, $8)|0;
     $88 = (($87>>>0) / 255)&-1;
     $89 = $88 << 16;
     $90 = $86 << 8;
     $91 = $89 | $90;
     $92 = $91 | $84;
     HEAP32[$dst$013>>2] = $92;
     $93 = (($posx$2) + ($20))|0;
     $94 = ((($dst$013)) + 4|0);
     $95 = (($96) + -1)|0;
     $97 = ($96|0)==(0);
     if ($97) {
      break;
     } else {
      $96 = $95;$dst$013 = $94;$posx$011 = $93;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $98 = HEAP32[$24>>2]|0;
  $99 = (($100) + ($98)|0);
  $101 = HEAP32[$12>>2]|0;
  $102 = (($101) + -1)|0;
  HEAP32[$12>>2] = $102;
  $103 = ($101|0)==(0);
  if ($103) {
   $$lcssa = $99;
   break;
  }
  $104 = (($posy$1$lcssa) + ($14))|0;
  $$pre27 = HEAP32[$18>>2]|0;
  $100 = $99;$105 = $99;$36 = $$pre27;$posy$017 = $104;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$23>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ARGB8888_BGR888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$lcssa14 = 0, $$pre = 0, $$pre10 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $dst$05 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $src$04 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + -1)|0;
 HEAP32[$10>>2] = $12;
 $13 = ($11|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre9 = HEAP32[$14>>2]|0;
 $87 = $$pre;$90 = $$pre9;$94 = $12;$95 = $$pre9;$96 = $$pre;
 while(1) {
  $25 = HEAP32[$15>>2]|0;
  $26 = ($25|0)==(0);
  if ($26) {
   $92 = $94;
  } else {
   $$in = $25;$dst$05 = $95;$src$04 = $96;
   while(1) {
    $27 = (($$in) + -1)|0;
    $28 = HEAP32[$src$04>>2]|0;
    $29 = $28 >>> 16;
    $30 = $29 & 255;
    $31 = $28 >>> 8;
    $32 = $31 & 255;
    $33 = $28 & 255;
    $34 = $28 >>> 24;
    $35 = HEAP32[$dst$05>>2]|0;
    $36 = $35 >>> 16;
    $37 = $36 & 255;
    $38 = $35 >>> 8;
    $39 = $38 & 255;
    $40 = $35 & 255;
    if ($19) {
     $srcB$0 = $33;$srcG$0 = $32;$srcR$0 = $30;
    } else {
     $41 = Math_imul($30, $4)|0;
     $42 = (($41>>>0) / 255)&-1;
     $43 = Math_imul($32, $6)|0;
     $44 = (($43>>>0) / 255)&-1;
     $45 = Math_imul($33, $8)|0;
     $46 = (($45>>>0) / 255)&-1;
     $srcB$0 = $46;$srcG$0 = $44;$srcR$0 = $42;
    }
    if ($21) {
     $srcA$0 = $34;
    } else {
     $47 = Math_imul($34, $9)|0;
     $48 = (($47>>>0) / 255)&-1;
     $srcA$0 = $48;
    }
    $49 = ($srcA$0>>>0)<(255);
    $or$cond = $23 & $49;
    if ($or$cond) {
     $50 = Math_imul($srcA$0, $srcR$0)|0;
     $51 = (($50>>>0) / 255)&-1;
     $52 = Math_imul($srcA$0, $srcG$0)|0;
     $53 = (($52>>>0) / 255)&-1;
     $54 = Math_imul($srcA$0, $srcB$0)|0;
     $55 = (($54>>>0) / 255)&-1;
     $srcB$1 = $55;$srcG$1 = $53;$srcR$1 = $51;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($24|0) {
    case 16:  {
     $56 = (255 - ($srcA$0))|0;
     $57 = Math_imul($56, $40)|0;
     $58 = (($57>>>0) / 255)&-1;
     $59 = (($srcR$1) + ($58))|0;
     $60 = Math_imul($56, $39)|0;
     $61 = (($60>>>0) / 255)&-1;
     $62 = (($srcG$1) + ($61))|0;
     $63 = Math_imul($56, $37)|0;
     $64 = (($63>>>0) / 255)&-1;
     $65 = (($srcB$1) + ($64))|0;
     $dstB$0 = $65;$dstG$0 = $62;$dstR$0 = $59;
     break;
    }
    case 32:  {
     $66 = (($srcR$1) + ($40))|0;
     $67 = ($66>>>0)>(255);
     $$ = $67 ? 255 : $66;
     $68 = (($srcG$1) + ($39))|0;
     $69 = ($68>>>0)>(255);
     $$3 = $69 ? 255 : $68;
     $70 = (($srcB$1) + ($37))|0;
     $71 = ($70>>>0)>(255);
     $$1 = $71 ? 255 : $70;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $72 = Math_imul($srcR$1, $40)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($srcG$1, $39)|0;
     $75 = (($74>>>0) / 255)&-1;
     $76 = Math_imul($srcB$1, $37)|0;
     $77 = (($76>>>0) / 255)&-1;
     $dstB$0 = $77;$dstG$0 = $75;$dstR$0 = $73;
     break;
    }
    default: {
     $dstB$0 = $37;$dstG$0 = $39;$dstR$0 = $40;
    }
    }
    $78 = $dstB$0 << 16;
    $79 = $dstG$0 << 8;
    $80 = $79 | $dstR$0;
    $81 = $80 | $78;
    HEAP32[$dst$05>>2] = $81;
    $82 = ((($src$04)) + 4|0);
    $83 = ((($dst$05)) + 4|0);
    $84 = ($27|0)==(0);
    if ($84) {
     break;
    } else {
     $$in = $27;$dst$05 = $83;$src$04 = $82;
    }
   }
   $$pre10 = HEAP32[$10>>2]|0;
   $92 = $$pre10;
  }
  $85 = HEAP32[$16>>2]|0;
  $86 = (($87) + ($85)|0);
  $88 = HEAP32[$17>>2]|0;
  $89 = (($90) + ($88)|0);
  $91 = (($92) + -1)|0;
  HEAP32[$10>>2] = $91;
  $93 = ($92|0)==(0);
  if ($93) {
   $$lcssa = $86;$$lcssa14 = $89;
   break;
  } else {
   $87 = $86;$90 = $89;$94 = $91;$95 = $89;$96 = $86;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa14;
 return;
}
function _SDL_Blit_ARGB8888_BGR888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0;
 var $src$015 = 0, $src$1 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 << 16;
 $13 = ((($info)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12|0) / ($14|0))&-1;
 $16 = ((($info)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 << 16;
 $19 = ((($info)) + 24|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($18|0) / ($20|0))&-1;
 $22 = (($14) + -1)|0;
 HEAP32[$13>>2] = $22;
 $23 = ($14|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = $1 & 112;
 $33 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $116 = $$pre;$121 = $22;$122 = $$pre;$42 = $20;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $34 = ($posy$020|0)>(65535);
  if ($34) {
   $35 = (($posy$020) + -65536)|0;
   $36 = $35 >>> 16;
   $37 = $36 << 16;
   $38 = (($srcy$019) + 1)|0;
   $39 = (($35) - ($37))|0;
   $40 = (($38) + ($36))|0;
   $posy$1$lcssa = $39;$srcy$1$lcssa = $40;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $41 = ($42|0)==(0);
  if ($41) {
   $118 = $121;
  } else {
   $$in = $42;$dst$016 = $122;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $43 = (($$in) + -1)|0;
    $44 = ($posx$014|0)>(65535);
    if ($44) {
     $45 = (($posx$014) + -65536)|0;
     $46 = $45 >>> 16;
     $47 = $46 << 16;
     $48 = (($srcx$013) + 1)|0;
     $49 = (($45) - ($47))|0;
     $50 = (($48) + ($46))|0;
     $51 = HEAP32[$info>>2]|0;
     $52 = HEAP32[$33>>2]|0;
     $53 = Math_imul($52, $srcy$1$lcssa)|0;
     $54 = (($51) + ($53)|0);
     $55 = $50 << 2;
     $56 = (($54) + ($55)|0);
     $posx$2 = $49;$src$1 = $56;$srcx$2 = $50;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $57 = HEAP32[$src$1>>2]|0;
    $58 = $57 >>> 16;
    $59 = $58 & 255;
    $60 = $57 >>> 8;
    $61 = $60 & 255;
    $62 = $57 & 255;
    $63 = $57 >>> 24;
    $64 = HEAP32[$dst$016>>2]|0;
    $65 = $64 >>> 16;
    $66 = $65 & 255;
    $67 = $64 >>> 8;
    $68 = $67 & 255;
    $69 = $64 & 255;
    if ($27) {
     $srcB$0 = $62;$srcG$0 = $61;$srcR$0 = $59;
    } else {
     $70 = Math_imul($59, $4)|0;
     $71 = (($70>>>0) / 255)&-1;
     $72 = Math_imul($61, $6)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($62, $8)|0;
     $75 = (($74>>>0) / 255)&-1;
     $srcB$0 = $75;$srcG$0 = $73;$srcR$0 = $71;
    }
    if ($29) {
     $srcA$0 = $63;
    } else {
     $76 = Math_imul($63, $9)|0;
     $77 = (($76>>>0) / 255)&-1;
     $srcA$0 = $77;
    }
    $78 = ($srcA$0>>>0)<(255);
    $or$cond = $31 & $78;
    if ($or$cond) {
     $79 = Math_imul($srcA$0, $srcR$0)|0;
     $80 = (($79>>>0) / 255)&-1;
     $81 = Math_imul($srcA$0, $srcG$0)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = Math_imul($srcA$0, $srcB$0)|0;
     $84 = (($83>>>0) / 255)&-1;
     $srcB$1 = $84;$srcG$1 = $82;$srcR$1 = $80;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($32|0) {
    case 16:  {
     $85 = (255 - ($srcA$0))|0;
     $86 = Math_imul($85, $69)|0;
     $87 = (($86>>>0) / 255)&-1;
     $88 = (($srcR$1) + ($87))|0;
     $89 = Math_imul($85, $68)|0;
     $90 = (($89>>>0) / 255)&-1;
     $91 = (($srcG$1) + ($90))|0;
     $92 = Math_imul($85, $66)|0;
     $93 = (($92>>>0) / 255)&-1;
     $94 = (($srcB$1) + ($93))|0;
     $dstB$0 = $94;$dstG$0 = $91;$dstR$0 = $88;
     break;
    }
    case 32:  {
     $95 = (($srcR$1) + ($69))|0;
     $96 = ($95>>>0)>(255);
     $$ = $96 ? 255 : $95;
     $97 = (($srcG$1) + ($68))|0;
     $98 = ($97>>>0)>(255);
     $$3 = $98 ? 255 : $97;
     $99 = (($srcB$1) + ($66))|0;
     $100 = ($99>>>0)>(255);
     $$1 = $100 ? 255 : $99;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $101 = Math_imul($srcR$1, $69)|0;
     $102 = (($101>>>0) / 255)&-1;
     $103 = Math_imul($srcG$1, $68)|0;
     $104 = (($103>>>0) / 255)&-1;
     $105 = Math_imul($srcB$1, $66)|0;
     $106 = (($105>>>0) / 255)&-1;
     $dstB$0 = $106;$dstG$0 = $104;$dstR$0 = $102;
     break;
    }
    default: {
     $dstB$0 = $66;$dstG$0 = $68;$dstR$0 = $69;
    }
    }
    $107 = $dstB$0 << 16;
    $108 = $dstG$0 << 8;
    $109 = $108 | $dstR$0;
    $110 = $109 | $107;
    HEAP32[$dst$016>>2] = $110;
    $111 = (($posx$2) + ($21))|0;
    $112 = ((($dst$016)) + 4|0);
    $113 = ($43|0)==(0);
    if ($113) {
     break;
    } else {
     $$in = $43;$dst$016 = $112;$posx$014 = $111;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$13>>2]|0;
   $118 = $$pre28;
  }
  $114 = HEAP32[$25>>2]|0;
  $115 = (($116) + ($114)|0);
  $117 = (($118) + -1)|0;
  HEAP32[$13>>2] = $117;
  $119 = ($118|0)==(0);
  if ($119) {
   $$lcssa = $115;
   break;
  }
  $120 = (($posy$1$lcssa) + ($15))|0;
  $$pre27 = HEAP32[$19>>2]|0;
  $116 = $115;$121 = $117;$122 = $115;$42 = $$pre27;$posy$020 = $120;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ARGB8888_ARGB8888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0, $posx$010 = 0;
 var $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$46 = $$pre;$51 = $12;$52 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $48 = $51;
  } else {
   $$in = $25;$dst$013 = $52;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    HEAP32[$dst$013>>2] = $40;
    $41 = (($posx$2) + ($11))|0;
    $42 = ((($dst$013)) + 4|0);
    $43 = ($26|0)==(0);
    if ($43) {
     break;
    } else {
     $$in = $26;$dst$013 = $42;$posx$010 = $41;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $48 = $$pre25;
  }
  $44 = HEAP32[$15>>2]|0;
  $45 = (($46) + ($44)|0);
  $47 = (($48) + -1)|0;
  HEAP32[$3>>2] = $47;
  $49 = ($48|0)==(0);
  if ($49) {
   $$lcssa = $45;
   break;
  }
  $50 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$46 = $45;$51 = $47;$52 = $45;$posy$016 = $50;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ARGB8888_ARGB8888_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$$us = 0, $$1 = 0, $$1$us = 0, $$3 = 0, $$3$us = 0, $$lcssa = 0, $$lcssa17 = 0, $$pre = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$05 = 0, $dst$05$us = 0, $dstA$0 = 0, $dstA$0$us = 0, $dstB$0 = 0, $dstB$0$us = 0, $dstG$0 = 0, $dstG$0$us = 0, $dstR$0 = 0, $dstR$0$us = 0, $src$04 = 0, $src$04$us = 0, $srcB$0$us = 0, $srcG$0$us = 0, $srcR$0$us = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 48;
 $11 = ($10|0)==(0);
 $12 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre10 = HEAP32[$6>>2]|0;
 $125 = $$pre;$128 = $$pre10;$132 = $$pre10;$133 = $$pre;
 while(1) {
  $13 = HEAP32[$7>>2]|0;
  $14 = (($13) + -1)|0;
  $15 = ($13|0)==(0);
  L6: do {
   if (!($15)) {
    if ($11) {
     $121 = $14;$dst$05 = $132;$src$04 = $133;
     while(1) {
      $73 = HEAP32[$src$04>>2]|0;
      $74 = $73 >>> 16;
      $75 = $74 & 255;
      $76 = $73 >>> 8;
      $77 = $76 & 255;
      $78 = $73 & 255;
      $79 = $73 >>> 24;
      $80 = HEAP32[$dst$05>>2]|0;
      $81 = $80 >>> 16;
      $82 = $81 & 255;
      $83 = $80 >>> 8;
      $84 = $83 & 255;
      $85 = $80 & 255;
      $86 = $80 >>> 24;
      switch ($12|0) {
      case 16:  {
       $87 = $79 ^ 255;
       $88 = Math_imul($82, $87)|0;
       $89 = (($88>>>0) / 255)&-1;
       $90 = (($75) + ($89))|0;
       $91 = Math_imul($84, $87)|0;
       $92 = (($91>>>0) / 255)&-1;
       $93 = (($77) + ($92))|0;
       $94 = Math_imul($87, $85)|0;
       $95 = (($94>>>0) / 255)&-1;
       $96 = (($78) + ($95))|0;
       $97 = Math_imul($87, $86)|0;
       $98 = (($97>>>0) / 255)&-1;
       $99 = (($98) + ($79))|0;
       $dstA$0 = $99;$dstB$0 = $96;$dstG$0 = $93;$dstR$0 = $90;
       break;
      }
      case 32:  {
       $100 = (($75) + ($82))|0;
       $101 = ($100>>>0)>(255);
       $$ = $101 ? 255 : $100;
       $102 = (($77) + ($84))|0;
       $103 = ($102>>>0)>(255);
       $$3 = $103 ? 255 : $102;
       $104 = (($78) + ($85))|0;
       $105 = ($104>>>0)>(255);
       $$1 = $105 ? 255 : $104;
       $dstA$0 = $86;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
       break;
      }
      case 64:  {
       $106 = Math_imul($75, $82)|0;
       $107 = (($106>>>0) / 255)&-1;
       $108 = Math_imul($77, $84)|0;
       $109 = (($108>>>0) / 255)&-1;
       $110 = Math_imul($78, $85)|0;
       $111 = (($110>>>0) / 255)&-1;
       $dstA$0 = $86;$dstB$0 = $111;$dstG$0 = $109;$dstR$0 = $107;
       break;
      }
      default: {
       $dstA$0 = $86;$dstB$0 = $85;$dstG$0 = $84;$dstR$0 = $82;
      }
      }
      $112 = $dstA$0 << 24;
      $113 = $dstR$0 << 16;
      $114 = $dstG$0 << 8;
      $115 = $114 | $113;
      $116 = $115 | $dstB$0;
      $117 = $116 | $112;
      HEAP32[$dst$05>>2] = $117;
      $118 = ((($src$04)) + 4|0);
      $119 = ((($dst$05)) + 4|0);
      $120 = (($121) + -1)|0;
      $122 = ($121|0)==(0);
      if ($122) {
       break L6;
      } else {
       $121 = $120;$dst$05 = $119;$src$04 = $118;
      }
     }
    } else {
     $71 = $14;$dst$05$us = $132;$src$04$us = $133;
    }
    while(1) {
     $16 = HEAP32[$src$04$us>>2]|0;
     $17 = $16 >>> 16;
     $18 = $17 & 255;
     $19 = $16 >>> 8;
     $20 = $19 & 255;
     $21 = $16 & 255;
     $22 = $16 >>> 24;
     $23 = HEAP32[$dst$05$us>>2]|0;
     $24 = $23 >>> 16;
     $25 = $24 & 255;
     $26 = $23 >>> 8;
     $27 = $26 & 255;
     $28 = $23 & 255;
     $29 = $23 >>> 24;
     $30 = ($22|0)==(255);
     if ($30) {
      $srcB$0$us = $21;$srcG$0$us = $20;$srcR$0$us = $18;
     } else {
      $31 = Math_imul($18, $22)|0;
      $32 = (($31>>>0) / 255)&-1;
      $33 = Math_imul($20, $22)|0;
      $34 = (($33>>>0) / 255)&-1;
      $35 = Math_imul($21, $22)|0;
      $36 = (($35>>>0) / 255)&-1;
      $srcB$0$us = $36;$srcG$0$us = $34;$srcR$0$us = $32;
     }
     switch ($12|0) {
     case 16:  {
      $49 = $22 ^ 255;
      $50 = Math_imul($25, $49)|0;
      $51 = (($50>>>0) / 255)&-1;
      $52 = (($srcR$0$us) + ($51))|0;
      $53 = Math_imul($27, $49)|0;
      $54 = (($53>>>0) / 255)&-1;
      $55 = (($srcG$0$us) + ($54))|0;
      $56 = Math_imul($49, $28)|0;
      $57 = (($56>>>0) / 255)&-1;
      $58 = (($srcB$0$us) + ($57))|0;
      $59 = Math_imul($49, $29)|0;
      $60 = (($59>>>0) / 255)&-1;
      $61 = (($60) + ($22))|0;
      $dstA$0$us = $61;$dstB$0$us = $58;$dstG$0$us = $55;$dstR$0$us = $52;
      break;
     }
     case 32:  {
      $43 = (($srcR$0$us) + ($25))|0;
      $44 = ($43>>>0)>(255);
      $$$us = $44 ? 255 : $43;
      $45 = (($srcG$0$us) + ($27))|0;
      $46 = ($45>>>0)>(255);
      $$3$us = $46 ? 255 : $45;
      $47 = (($srcB$0$us) + ($28))|0;
      $48 = ($47>>>0)>(255);
      $$1$us = $48 ? 255 : $47;
      $dstA$0$us = $29;$dstB$0$us = $$1$us;$dstG$0$us = $$3$us;$dstR$0$us = $$$us;
      break;
     }
     case 64:  {
      $37 = Math_imul($srcR$0$us, $25)|0;
      $38 = (($37>>>0) / 255)&-1;
      $39 = Math_imul($srcG$0$us, $27)|0;
      $40 = (($39>>>0) / 255)&-1;
      $41 = Math_imul($srcB$0$us, $28)|0;
      $42 = (($41>>>0) / 255)&-1;
      $dstA$0$us = $29;$dstB$0$us = $42;$dstG$0$us = $40;$dstR$0$us = $38;
      break;
     }
     default: {
      $dstA$0$us = $29;$dstB$0$us = $28;$dstG$0$us = $27;$dstR$0$us = $25;
     }
     }
     $62 = $dstA$0$us << 24;
     $63 = $dstR$0$us << 16;
     $64 = $dstG$0$us << 8;
     $65 = $64 | $63;
     $66 = $65 | $dstB$0$us;
     $67 = $66 | $62;
     HEAP32[$dst$05$us>>2] = $67;
     $68 = ((($src$04$us)) + 4|0);
     $69 = ((($dst$05$us)) + 4|0);
     $70 = (($71) + -1)|0;
     $72 = ($71|0)==(0);
     if ($72) {
      break L6;
     } else {
      $71 = $70;$dst$05$us = $69;$src$04$us = $68;
     }
    }
   }
  } while(0);
  $123 = HEAP32[$8>>2]|0;
  $124 = (($125) + ($123)|0);
  $126 = HEAP32[$9>>2]|0;
  $127 = (($128) + ($126)|0);
  $129 = HEAP32[$2>>2]|0;
  $130 = (($129) + -1)|0;
  HEAP32[$2>>2] = $130;
  $131 = ($129|0)==(0);
  if ($131) {
   $$lcssa = $124;$$lcssa17 = $127;
   break;
  } else {
   $125 = $124;$128 = $127;$132 = $127;$133 = $124;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa17;
 return;
}
function _SDL_Blit_ARGB8888_ARGB8888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstA$0 = 0, $dstB$0 = 0;
 var $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0, $src$015 = 0, $src$1 = 0, $srcB$0 = 0, $srcG$0 = 0, $srcR$0 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 48;
 $19 = ($18|0)!=(0);
 $20 = $1 & 112;
 $21 = ((($info)) + 12|0);
 $$pre = HEAP32[$16>>2]|0;
 $102 = $$pre;$107 = $14;$108 = $$pre;$30 = $12;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $22 = ($posy$020|0)>(65535);
  if ($22) {
   $23 = (($posy$020) + -65536)|0;
   $24 = $23 >>> 16;
   $25 = $24 << 16;
   $26 = (($srcy$019) + 1)|0;
   $27 = (($23) - ($25))|0;
   $28 = (($26) + ($24))|0;
   $posy$1$lcssa = $27;$srcy$1$lcssa = $28;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $29 = ($30|0)==(0);
  if ($29) {
   $104 = $107;
  } else {
   $$in = $30;$dst$016 = $108;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $31 = (($$in) + -1)|0;
    $32 = ($posx$014|0)>(65535);
    if ($32) {
     $33 = (($posx$014) + -65536)|0;
     $34 = $33 >>> 16;
     $35 = $34 << 16;
     $36 = (($srcx$013) + 1)|0;
     $37 = (($33) - ($35))|0;
     $38 = (($36) + ($34))|0;
     $39 = HEAP32[$info>>2]|0;
     $40 = HEAP32[$21>>2]|0;
     $41 = Math_imul($40, $srcy$1$lcssa)|0;
     $42 = (($39) + ($41)|0);
     $43 = $38 << 2;
     $44 = (($42) + ($43)|0);
     $posx$2 = $37;$src$1 = $44;$srcx$2 = $38;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $45 = HEAP32[$src$1>>2]|0;
    $46 = $45 >>> 16;
    $47 = $46 & 255;
    $48 = $45 >>> 8;
    $49 = $48 & 255;
    $50 = $45 & 255;
    $51 = $45 >>> 24;
    $52 = HEAP32[$dst$016>>2]|0;
    $53 = $52 >>> 16;
    $54 = $53 & 255;
    $55 = $52 >>> 8;
    $56 = $55 & 255;
    $57 = $52 & 255;
    $58 = $52 >>> 24;
    $59 = ($51|0)!=(255);
    $or$cond = $19 & $59;
    if ($or$cond) {
     $60 = Math_imul($47, $51)|0;
     $61 = (($60>>>0) / 255)&-1;
     $62 = Math_imul($49, $51)|0;
     $63 = (($62>>>0) / 255)&-1;
     $64 = Math_imul($50, $51)|0;
     $65 = (($64>>>0) / 255)&-1;
     $srcB$0 = $65;$srcG$0 = $63;$srcR$0 = $61;
    } else {
     $srcB$0 = $50;$srcG$0 = $49;$srcR$0 = $47;
    }
    switch ($20|0) {
    case 16:  {
     $66 = $51 ^ 255;
     $67 = Math_imul($54, $66)|0;
     $68 = (($67>>>0) / 255)&-1;
     $69 = (($srcR$0) + ($68))|0;
     $70 = Math_imul($56, $66)|0;
     $71 = (($70>>>0) / 255)&-1;
     $72 = (($srcG$0) + ($71))|0;
     $73 = Math_imul($66, $57)|0;
     $74 = (($73>>>0) / 255)&-1;
     $75 = (($srcB$0) + ($74))|0;
     $76 = Math_imul($66, $58)|0;
     $77 = (($76>>>0) / 255)&-1;
     $78 = (($77) + ($51))|0;
     $dstA$0 = $78;$dstB$0 = $75;$dstG$0 = $72;$dstR$0 = $69;
     break;
    }
    case 32:  {
     $79 = (($srcR$0) + ($54))|0;
     $80 = ($79>>>0)>(255);
     $$ = $80 ? 255 : $79;
     $81 = (($srcG$0) + ($56))|0;
     $82 = ($81>>>0)>(255);
     $$3 = $82 ? 255 : $81;
     $83 = (($srcB$0) + ($57))|0;
     $84 = ($83>>>0)>(255);
     $$1 = $84 ? 255 : $83;
     $dstA$0 = $58;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $85 = Math_imul($srcR$0, $54)|0;
     $86 = (($85>>>0) / 255)&-1;
     $87 = Math_imul($srcG$0, $56)|0;
     $88 = (($87>>>0) / 255)&-1;
     $89 = Math_imul($srcB$0, $57)|0;
     $90 = (($89>>>0) / 255)&-1;
     $dstA$0 = $58;$dstB$0 = $90;$dstG$0 = $88;$dstR$0 = $86;
     break;
    }
    default: {
     $dstA$0 = $58;$dstB$0 = $57;$dstG$0 = $56;$dstR$0 = $54;
    }
    }
    $91 = $dstA$0 << 24;
    $92 = $dstR$0 << 16;
    $93 = $dstG$0 << 8;
    $94 = $93 | $92;
    $95 = $94 | $dstB$0;
    $96 = $95 | $91;
    HEAP32[$dst$016>>2] = $96;
    $97 = (($posx$2) + ($13))|0;
    $98 = ((($dst$016)) + 4|0);
    $99 = ($31|0)==(0);
    if ($99) {
     break;
    } else {
     $$in = $31;$dst$016 = $98;$posx$014 = $97;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$5>>2]|0;
   $104 = $$pre28;
  }
  $100 = HEAP32[$17>>2]|0;
  $101 = (($102) + ($100)|0);
  $103 = (($104) + -1)|0;
  HEAP32[$5>>2] = $103;
  $105 = ($104|0)==(0);
  if ($105) {
   $$lcssa = $101;
   break;
  }
  $106 = (($posy$1$lcssa) + ($7))|0;
  $$pre27 = HEAP32[$11>>2]|0;
  $102 = $101;$107 = $103;$108 = $101;$30 = $$pre27;$posy$020 = $106;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ARGB8888_ARGB8888_Modulate($info) {
 $info = $info|0;
 var $$lcssa = 0, $$lcssa24 = 0, $$pre = 0, $$pre14 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $A$0$us = 0, $dst$02 = 0, $dst$02$us = 0, $dst$02$us6 = 0, $src$01 = 0, $src$01$us = 0, $src$01$us7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + -1)|0;
 HEAP32[$10>>2] = $12;
 $13 = ($11|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $$pre = HEAP32[$info>>2]|0;
 $$pre14 = HEAP32[$14>>2]|0;
 $88 = $$pre;$91 = $$pre14;$95 = $$pre14;$96 = $$pre;
 while(1) {
  $22 = HEAP32[$15>>2]|0;
  $23 = (($22) + -1)|0;
  $24 = ($22|0)==(0);
  if (!($24)) {
   if ($19) {
    $35 = $23;$dst$02$us = $95;$src$01$us = $96;
    while(1) {
     $25 = HEAP32[$src$01$us>>2]|0;
     $26 = $25 >>> 24;
     if ($21) {
      $A$0$us = $26;
     } else {
      $27 = Math_imul($26, $9)|0;
      $28 = (($27>>>0) / 255)&-1;
      $A$0$us = $28;
     }
     $29 = $A$0$us << 24;
     $30 = $25 & 16777215;
     $31 = $30 | $29;
     HEAP32[$dst$02$us>>2] = $31;
     $32 = ((($src$01$us)) + 4|0);
     $33 = ((($dst$02$us)) + 4|0);
     $34 = (($35) + -1)|0;
     $36 = ($35|0)==(0);
     if ($36) {
      break;
     } else {
      $35 = $34;$dst$02$us = $33;$src$01$us = $32;
     }
    }
   } else {
    if ($21) {
     $58 = $23;$dst$02$us6 = $95;$src$01$us7 = $96;
     while(1) {
      $37 = HEAP32[$src$01$us7>>2]|0;
      $38 = $37 >>> 16;
      $39 = $38 & 255;
      $40 = $37 >>> 8;
      $41 = $40 & 255;
      $42 = $37 & 255;
      $43 = $37 & -16777216;
      $44 = Math_imul($39, $4)|0;
      $45 = (($44>>>0) / 255)&-1;
      $46 = Math_imul($41, $6)|0;
      $47 = (($46>>>0) / 255)&-1;
      $48 = Math_imul($42, $8)|0;
      $49 = (($48>>>0) / 255)&-1;
      $50 = $45 << 16;
      $51 = $47 << 8;
      $52 = $51 | $50;
      $53 = $52 | $49;
      $54 = $53 | $43;
      HEAP32[$dst$02$us6>>2] = $54;
      $55 = ((($src$01$us7)) + 4|0);
      $56 = ((($dst$02$us6)) + 4|0);
      $57 = (($58) + -1)|0;
      $59 = ($58|0)==(0);
      if ($59) {
       break;
      } else {
       $58 = $57;$dst$02$us6 = $56;$src$01$us7 = $55;
      }
     }
    } else {
     $84 = $23;$dst$02 = $95;$src$01 = $96;
     while(1) {
      $60 = HEAP32[$src$01>>2]|0;
      $61 = $60 >>> 16;
      $62 = $61 & 255;
      $63 = $60 >>> 8;
      $64 = $63 & 255;
      $65 = $60 & 255;
      $66 = $60 >>> 24;
      $67 = Math_imul($62, $4)|0;
      $68 = (($67>>>0) / 255)&-1;
      $69 = Math_imul($64, $6)|0;
      $70 = (($69>>>0) / 255)&-1;
      $71 = Math_imul($65, $8)|0;
      $72 = (($71>>>0) / 255)&-1;
      $73 = Math_imul($66, $9)|0;
      $74 = (($73>>>0) / 255)&-1;
      $75 = $74 << 24;
      $76 = $68 << 16;
      $77 = $70 << 8;
      $78 = $77 | $76;
      $79 = $78 | $72;
      $80 = $79 | $75;
      HEAP32[$dst$02>>2] = $80;
      $81 = ((($src$01)) + 4|0);
      $82 = ((($dst$02)) + 4|0);
      $83 = (($84) + -1)|0;
      $85 = ($84|0)==(0);
      if ($85) {
       break;
      } else {
       $84 = $83;$dst$02 = $82;$src$01 = $81;
      }
     }
    }
   }
  }
  $86 = HEAP32[$16>>2]|0;
  $87 = (($88) + ($86)|0);
  $89 = HEAP32[$17>>2]|0;
  $90 = (($91) + ($89)|0);
  $92 = HEAP32[$10>>2]|0;
  $93 = (($92) + -1)|0;
  HEAP32[$10>>2] = $93;
  $94 = ($92|0)==(0);
  if ($94) {
   $$lcssa = $87;$$lcssa24 = $90;
   break;
  } else {
   $88 = $87;$91 = $90;$95 = $90;$96 = $87;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa24;
 return;
}
function _SDL_Blit_ARGB8888_ARGB8888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $A$0 = 0, $A$0$us = 0, $dst$013 = 0, $dst$013$us = 0;
 var $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$017 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0, $src$1$us = 0, $srcx$010 = 0, $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$016 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 << 16;
 $13 = ((($info)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12|0) / ($14|0))&-1;
 $16 = ((($info)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 << 16;
 $19 = ((($info)) + 24|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($18|0) / ($20|0))&-1;
 $22 = (($14) + -1)|0;
 HEAP32[$13>>2] = $22;
 $23 = ($14|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $107 = $$pre;$112 = $$pre;$39 = $20;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $31 = ($posy$017|0)>(65535);
  if ($31) {
   $32 = (($posy$017) + -65536)|0;
   $33 = $32 >>> 16;
   $34 = $33 << 16;
   $35 = (($srcy$016) + 1)|0;
   $36 = (($32) - ($34))|0;
   $37 = (($35) + ($33))|0;
   $posy$1$lcssa = $36;$srcy$1$lcssa = $37;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $38 = (($39) + -1)|0;
  $40 = ($39|0)==(0);
  if (!($40)) {
   if ($27) {
    $64 = $38;$dst$013$us = $112;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $41 = ($posx$011$us|0)>(65535);
     if ($41) {
      $42 = (($posx$011$us) + -65536)|0;
      $43 = $42 >>> 16;
      $44 = $43 << 16;
      $45 = (($srcx$010$us) + 1)|0;
      $46 = (($42) - ($44))|0;
      $47 = (($45) + ($43))|0;
      $48 = HEAP32[$info>>2]|0;
      $49 = HEAP32[$30>>2]|0;
      $50 = Math_imul($49, $srcy$1$lcssa)|0;
      $51 = (($48) + ($50)|0);
      $52 = $47 << 2;
      $53 = (($51) + ($52)|0);
      $posx$2$us = $46;$src$1$us = $53;$srcx$2$us = $47;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $54 = HEAP32[$src$1$us>>2]|0;
     $55 = $54 >>> 24;
     if ($29) {
      $A$0$us = $55;
     } else {
      $56 = Math_imul($55, $9)|0;
      $57 = (($56>>>0) / 255)&-1;
      $A$0$us = $57;
     }
     $58 = $A$0$us << 24;
     $59 = $54 & 16777215;
     $60 = $59 | $58;
     HEAP32[$dst$013$us>>2] = $60;
     $61 = (($posx$2$us) + ($21))|0;
     $62 = ((($dst$013$us)) + 4|0);
     $63 = (($64) + -1)|0;
     $65 = ($64|0)==(0);
     if ($65) {
      break;
     } else {
      $64 = $63;$dst$013$us = $62;$posx$011$us = $61;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $103 = $38;$dst$013 = $112;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $66 = ($posx$011|0)>(65535);
     if ($66) {
      $67 = (($posx$011) + -65536)|0;
      $68 = $67 >>> 16;
      $69 = $68 << 16;
      $70 = (($srcx$010) + 1)|0;
      $71 = (($67) - ($69))|0;
      $72 = (($70) + ($68))|0;
      $73 = HEAP32[$info>>2]|0;
      $74 = HEAP32[$30>>2]|0;
      $75 = Math_imul($74, $srcy$1$lcssa)|0;
      $76 = (($73) + ($75)|0);
      $77 = $72 << 2;
      $78 = (($76) + ($77)|0);
      $posx$2 = $71;$src$1 = $78;$srcx$2 = $72;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $79 = HEAP32[$src$1>>2]|0;
     $80 = $79 >>> 16;
     $81 = $80 & 255;
     $82 = $79 >>> 8;
     $83 = $82 & 255;
     $84 = $79 & 255;
     $85 = $79 >>> 24;
     $86 = Math_imul($81, $4)|0;
     $87 = (($86>>>0) / 255)&-1;
     $88 = Math_imul($83, $6)|0;
     $89 = (($88>>>0) / 255)&-1;
     $90 = Math_imul($84, $8)|0;
     $91 = (($90>>>0) / 255)&-1;
     if ($29) {
      $A$0 = $85;
     } else {
      $92 = Math_imul($85, $9)|0;
      $93 = (($92>>>0) / 255)&-1;
      $A$0 = $93;
     }
     $94 = $A$0 << 24;
     $95 = $87 << 16;
     $96 = $89 << 8;
     $97 = $91 | $96;
     $98 = $97 | $95;
     $99 = $98 | $94;
     HEAP32[$dst$013>>2] = $99;
     $100 = (($posx$2) + ($21))|0;
     $101 = ((($dst$013)) + 4|0);
     $102 = (($103) + -1)|0;
     $104 = ($103|0)==(0);
     if ($104) {
      break;
     } else {
      $103 = $102;$dst$013 = $101;$posx$011 = $100;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $105 = HEAP32[$25>>2]|0;
  $106 = (($107) + ($105)|0);
  $108 = HEAP32[$13>>2]|0;
  $109 = (($108) + -1)|0;
  HEAP32[$13>>2] = $109;
  $110 = ($108|0)==(0);
  if ($110) {
   $$lcssa = $106;
   break;
  }
  $111 = (($posy$1$lcssa) + ($15))|0;
  $$pre27 = HEAP32[$19>>2]|0;
  $107 = $106;$112 = $106;$39 = $$pre27;$posy$017 = $111;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ARGB8888_ARGB8888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$lcssa14 = 0, $$pre = 0, $$pre10 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$05 = 0, $dstA$0 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $src$04 = 0, $srcA$0 = 0;
 var $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + -1)|0;
 HEAP32[$10>>2] = $12;
 $13 = ($11|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre9 = HEAP32[$14>>2]|0;
 $100 = $12;$101 = $$pre9;$102 = $$pre;$93 = $$pre;$96 = $$pre9;
 while(1) {
  $25 = HEAP32[$15>>2]|0;
  $26 = ($25|0)==(0);
  if ($26) {
   $98 = $100;
  } else {
   $$in = $25;$dst$05 = $101;$src$04 = $102;
   while(1) {
    $27 = (($$in) + -1)|0;
    $28 = HEAP32[$src$04>>2]|0;
    $29 = $28 >>> 16;
    $30 = $29 & 255;
    $31 = $28 >>> 8;
    $32 = $31 & 255;
    $33 = $28 & 255;
    $34 = $28 >>> 24;
    $35 = HEAP32[$dst$05>>2]|0;
    $36 = $35 >>> 16;
    $37 = $36 & 255;
    $38 = $35 >>> 8;
    $39 = $38 & 255;
    $40 = $35 & 255;
    $41 = $35 >>> 24;
    if ($19) {
     $srcB$0 = $33;$srcG$0 = $32;$srcR$0 = $30;
    } else {
     $42 = Math_imul($30, $4)|0;
     $43 = (($42>>>0) / 255)&-1;
     $44 = Math_imul($32, $6)|0;
     $45 = (($44>>>0) / 255)&-1;
     $46 = Math_imul($33, $8)|0;
     $47 = (($46>>>0) / 255)&-1;
     $srcB$0 = $47;$srcG$0 = $45;$srcR$0 = $43;
    }
    if ($21) {
     $srcA$0 = $34;
    } else {
     $48 = Math_imul($34, $9)|0;
     $49 = (($48>>>0) / 255)&-1;
     $srcA$0 = $49;
    }
    $50 = ($srcA$0>>>0)<(255);
    $or$cond = $23 & $50;
    if ($or$cond) {
     $51 = Math_imul($srcA$0, $srcR$0)|0;
     $52 = (($51>>>0) / 255)&-1;
     $53 = Math_imul($srcA$0, $srcG$0)|0;
     $54 = (($53>>>0) / 255)&-1;
     $55 = Math_imul($srcA$0, $srcB$0)|0;
     $56 = (($55>>>0) / 255)&-1;
     $srcB$1 = $56;$srcG$1 = $54;$srcR$1 = $52;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($24|0) {
    case 16:  {
     $57 = (255 - ($srcA$0))|0;
     $58 = Math_imul($57, $37)|0;
     $59 = (($58>>>0) / 255)&-1;
     $60 = (($srcR$1) + ($59))|0;
     $61 = Math_imul($57, $39)|0;
     $62 = (($61>>>0) / 255)&-1;
     $63 = (($srcG$1) + ($62))|0;
     $64 = Math_imul($57, $40)|0;
     $65 = (($64>>>0) / 255)&-1;
     $66 = (($srcB$1) + ($65))|0;
     $67 = Math_imul($57, $41)|0;
     $68 = (($67>>>0) / 255)&-1;
     $69 = (($68) + ($srcA$0))|0;
     $dstA$0 = $69;$dstB$0 = $66;$dstG$0 = $63;$dstR$0 = $60;
     break;
    }
    case 32:  {
     $70 = (($srcR$1) + ($37))|0;
     $71 = ($70>>>0)>(255);
     $$ = $71 ? 255 : $70;
     $72 = (($srcG$1) + ($39))|0;
     $73 = ($72>>>0)>(255);
     $$3 = $73 ? 255 : $72;
     $74 = (($srcB$1) + ($40))|0;
     $75 = ($74>>>0)>(255);
     $$1 = $75 ? 255 : $74;
     $dstA$0 = $41;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $76 = Math_imul($srcR$1, $37)|0;
     $77 = (($76>>>0) / 255)&-1;
     $78 = Math_imul($srcG$1, $39)|0;
     $79 = (($78>>>0) / 255)&-1;
     $80 = Math_imul($srcB$1, $40)|0;
     $81 = (($80>>>0) / 255)&-1;
     $dstA$0 = $41;$dstB$0 = $81;$dstG$0 = $79;$dstR$0 = $77;
     break;
    }
    default: {
     $dstA$0 = $41;$dstB$0 = $40;$dstG$0 = $39;$dstR$0 = $37;
    }
    }
    $82 = $dstA$0 << 24;
    $83 = $dstR$0 << 16;
    $84 = $dstG$0 << 8;
    $85 = $84 | $83;
    $86 = $85 | $dstB$0;
    $87 = $86 | $82;
    HEAP32[$dst$05>>2] = $87;
    $88 = ((($src$04)) + 4|0);
    $89 = ((($dst$05)) + 4|0);
    $90 = ($27|0)==(0);
    if ($90) {
     break;
    } else {
     $$in = $27;$dst$05 = $89;$src$04 = $88;
    }
   }
   $$pre10 = HEAP32[$10>>2]|0;
   $98 = $$pre10;
  }
  $91 = HEAP32[$16>>2]|0;
  $92 = (($93) + ($91)|0);
  $94 = HEAP32[$17>>2]|0;
  $95 = (($96) + ($94)|0);
  $97 = (($98) + -1)|0;
  HEAP32[$10>>2] = $97;
  $99 = ($98|0)==(0);
  if ($99) {
   $$lcssa = $92;$$lcssa14 = $95;
   break;
  } else {
   $100 = $97;$101 = $95;$102 = $92;$93 = $92;$96 = $95;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa14;
 return;
}
function _SDL_Blit_ARGB8888_ARGB8888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstA$0 = 0, $dstB$0 = 0;
 var $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0, $src$015 = 0, $src$1 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 << 16;
 $13 = ((($info)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12|0) / ($14|0))&-1;
 $16 = ((($info)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 << 16;
 $19 = ((($info)) + 24|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($18|0) / ($20|0))&-1;
 $22 = (($14) + -1)|0;
 HEAP32[$13>>2] = $22;
 $23 = ($14|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = $1 & 112;
 $33 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $122 = $$pre;$127 = $22;$128 = $$pre;$42 = $20;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $34 = ($posy$020|0)>(65535);
  if ($34) {
   $35 = (($posy$020) + -65536)|0;
   $36 = $35 >>> 16;
   $37 = $36 << 16;
   $38 = (($srcy$019) + 1)|0;
   $39 = (($35) - ($37))|0;
   $40 = (($38) + ($36))|0;
   $posy$1$lcssa = $39;$srcy$1$lcssa = $40;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $41 = ($42|0)==(0);
  if ($41) {
   $124 = $127;
  } else {
   $$in = $42;$dst$016 = $128;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $43 = (($$in) + -1)|0;
    $44 = ($posx$014|0)>(65535);
    if ($44) {
     $45 = (($posx$014) + -65536)|0;
     $46 = $45 >>> 16;
     $47 = $46 << 16;
     $48 = (($srcx$013) + 1)|0;
     $49 = (($45) - ($47))|0;
     $50 = (($48) + ($46))|0;
     $51 = HEAP32[$info>>2]|0;
     $52 = HEAP32[$33>>2]|0;
     $53 = Math_imul($52, $srcy$1$lcssa)|0;
     $54 = (($51) + ($53)|0);
     $55 = $50 << 2;
     $56 = (($54) + ($55)|0);
     $posx$2 = $49;$src$1 = $56;$srcx$2 = $50;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $57 = HEAP32[$src$1>>2]|0;
    $58 = $57 >>> 16;
    $59 = $58 & 255;
    $60 = $57 >>> 8;
    $61 = $60 & 255;
    $62 = $57 & 255;
    $63 = $57 >>> 24;
    $64 = HEAP32[$dst$016>>2]|0;
    $65 = $64 >>> 16;
    $66 = $65 & 255;
    $67 = $64 >>> 8;
    $68 = $67 & 255;
    $69 = $64 & 255;
    $70 = $64 >>> 24;
    if ($27) {
     $srcB$0 = $62;$srcG$0 = $61;$srcR$0 = $59;
    } else {
     $71 = Math_imul($59, $4)|0;
     $72 = (($71>>>0) / 255)&-1;
     $73 = Math_imul($61, $6)|0;
     $74 = (($73>>>0) / 255)&-1;
     $75 = Math_imul($62, $8)|0;
     $76 = (($75>>>0) / 255)&-1;
     $srcB$0 = $76;$srcG$0 = $74;$srcR$0 = $72;
    }
    if ($29) {
     $srcA$0 = $63;
    } else {
     $77 = Math_imul($63, $9)|0;
     $78 = (($77>>>0) / 255)&-1;
     $srcA$0 = $78;
    }
    $79 = ($srcA$0>>>0)<(255);
    $or$cond = $31 & $79;
    if ($or$cond) {
     $80 = Math_imul($srcA$0, $srcR$0)|0;
     $81 = (($80>>>0) / 255)&-1;
     $82 = Math_imul($srcA$0, $srcG$0)|0;
     $83 = (($82>>>0) / 255)&-1;
     $84 = Math_imul($srcA$0, $srcB$0)|0;
     $85 = (($84>>>0) / 255)&-1;
     $srcB$1 = $85;$srcG$1 = $83;$srcR$1 = $81;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($32|0) {
    case 16:  {
     $86 = (255 - ($srcA$0))|0;
     $87 = Math_imul($86, $66)|0;
     $88 = (($87>>>0) / 255)&-1;
     $89 = (($srcR$1) + ($88))|0;
     $90 = Math_imul($86, $68)|0;
     $91 = (($90>>>0) / 255)&-1;
     $92 = (($srcG$1) + ($91))|0;
     $93 = Math_imul($86, $69)|0;
     $94 = (($93>>>0) / 255)&-1;
     $95 = (($srcB$1) + ($94))|0;
     $96 = Math_imul($86, $70)|0;
     $97 = (($96>>>0) / 255)&-1;
     $98 = (($97) + ($srcA$0))|0;
     $dstA$0 = $98;$dstB$0 = $95;$dstG$0 = $92;$dstR$0 = $89;
     break;
    }
    case 32:  {
     $99 = (($srcR$1) + ($66))|0;
     $100 = ($99>>>0)>(255);
     $$ = $100 ? 255 : $99;
     $101 = (($srcG$1) + ($68))|0;
     $102 = ($101>>>0)>(255);
     $$3 = $102 ? 255 : $101;
     $103 = (($srcB$1) + ($69))|0;
     $104 = ($103>>>0)>(255);
     $$1 = $104 ? 255 : $103;
     $dstA$0 = $70;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $105 = Math_imul($srcR$1, $66)|0;
     $106 = (($105>>>0) / 255)&-1;
     $107 = Math_imul($srcG$1, $68)|0;
     $108 = (($107>>>0) / 255)&-1;
     $109 = Math_imul($srcB$1, $69)|0;
     $110 = (($109>>>0) / 255)&-1;
     $dstA$0 = $70;$dstB$0 = $110;$dstG$0 = $108;$dstR$0 = $106;
     break;
    }
    default: {
     $dstA$0 = $70;$dstB$0 = $69;$dstG$0 = $68;$dstR$0 = $66;
    }
    }
    $111 = $dstA$0 << 24;
    $112 = $dstR$0 << 16;
    $113 = $dstG$0 << 8;
    $114 = $113 | $112;
    $115 = $114 | $dstB$0;
    $116 = $115 | $111;
    HEAP32[$dst$016>>2] = $116;
    $117 = (($posx$2) + ($21))|0;
    $118 = ((($dst$016)) + 4|0);
    $119 = ($43|0)==(0);
    if ($119) {
     break;
    } else {
     $$in = $43;$dst$016 = $118;$posx$014 = $117;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$13>>2]|0;
   $124 = $$pre28;
  }
  $120 = HEAP32[$25>>2]|0;
  $121 = (($122) + ($120)|0);
  $123 = (($124) + -1)|0;
  HEAP32[$13>>2] = $123;
  $125 = ($124|0)==(0);
  if ($125) {
   $$lcssa = $121;
   break;
  }
  $126 = (($posy$1$lcssa) + ($15))|0;
  $$pre27 = HEAP32[$19>>2]|0;
  $122 = $121;$127 = $123;$128 = $121;$42 = $$pre27;$posy$020 = $126;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGBA8888_RGB888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0, $posx$010 = 0, $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$53 = $$pre;$58 = $12;$59 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $55 = $58;
  } else {
   $$in = $25;$dst$013 = $59;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    $41 = $40 >>> 24;
    $42 = $40 >>> 8;
    $43 = $42 & 255;
    $44 = $41 << 16;
    $45 = $42 & 65280;
    $46 = $44 | $45;
    $47 = $46 | $43;
    HEAP32[$dst$013>>2] = $47;
    $48 = (($posx$2) + ($11))|0;
    $49 = ((($dst$013)) + 4|0);
    $50 = ($26|0)==(0);
    if ($50) {
     break;
    } else {
     $$in = $26;$dst$013 = $49;$posx$010 = $48;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $55 = $$pre25;
  }
  $51 = HEAP32[$15>>2]|0;
  $52 = (($53) + ($51)|0);
  $54 = (($55) + -1)|0;
  HEAP32[$3>>2] = $54;
  $56 = ($55|0)==(0);
  if ($56) {
   $$lcssa = $52;
   break;
  }
  $57 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$53 = $52;$58 = $54;$59 = $52;$posy$016 = $57;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGBA8888_RGB888_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$$us = 0, $$1 = 0, $$1$us = 0, $$3 = 0, $$3$us = 0, $$lcssa = 0, $$lcssa17 = 0, $$pre = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$05 = 0, $dst$05$us = 0, $dstB$0 = 0, $dstB$0$us = 0, $dstG$0 = 0, $dstG$0$us = 0, $dstR$0 = 0, $dstR$0$us = 0;
 var $src$04 = 0, $src$04$us = 0, $srcB$0$us = 0, $srcG$0$us = 0, $srcR$0$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 48;
 $11 = ($10|0)==(0);
 $12 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre10 = HEAP32[$6>>2]|0;
 $113 = $$pre;$116 = $$pre10;$120 = $$pre10;$121 = $$pre;
 while(1) {
  $13 = HEAP32[$7>>2]|0;
  $14 = (($13) + -1)|0;
  $15 = ($13|0)==(0);
  L6: do {
   if (!($15)) {
    if ($11) {
     $109 = $14;$dst$05 = $120;$src$04 = $121;
     while(1) {
      $67 = HEAP32[$src$04>>2]|0;
      $68 = $67 >>> 24;
      $69 = $67 >>> 16;
      $70 = $69 & 255;
      $71 = $67 >>> 8;
      $72 = $71 & 255;
      $73 = HEAP32[$dst$05>>2]|0;
      $74 = $73 >>> 16;
      $75 = $74 & 255;
      $76 = $73 >>> 8;
      $77 = $76 & 255;
      $78 = $73 & 255;
      switch ($12|0) {
      case 16:  {
       $79 = $67 & 255;
       $80 = $79 ^ 255;
       $81 = Math_imul($75, $80)|0;
       $82 = (($81>>>0) / 255)&-1;
       $83 = (($68) + ($82))|0;
       $84 = Math_imul($77, $80)|0;
       $85 = (($84>>>0) / 255)&-1;
       $86 = (($70) + ($85))|0;
       $87 = Math_imul($80, $78)|0;
       $88 = (($87>>>0) / 255)&-1;
       $89 = (($72) + ($88))|0;
       $dstB$0 = $89;$dstG$0 = $86;$dstR$0 = $83;
       break;
      }
      case 32:  {
       $90 = (($68) + ($75))|0;
       $91 = ($90>>>0)>(255);
       $$ = $91 ? 255 : $90;
       $92 = (($70) + ($77))|0;
       $93 = ($92>>>0)>(255);
       $$3 = $93 ? 255 : $92;
       $94 = (($72) + ($78))|0;
       $95 = ($94>>>0)>(255);
       $$1 = $95 ? 255 : $94;
       $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
       break;
      }
      case 64:  {
       $96 = Math_imul($68, $75)|0;
       $97 = (($96>>>0) / 255)&-1;
       $98 = Math_imul($70, $77)|0;
       $99 = (($98>>>0) / 255)&-1;
       $100 = Math_imul($72, $78)|0;
       $101 = (($100>>>0) / 255)&-1;
       $dstB$0 = $101;$dstG$0 = $99;$dstR$0 = $97;
       break;
      }
      default: {
       $dstB$0 = $78;$dstG$0 = $77;$dstR$0 = $75;
      }
      }
      $102 = $dstR$0 << 16;
      $103 = $dstG$0 << 8;
      $104 = $103 | $102;
      $105 = $104 | $dstB$0;
      HEAP32[$dst$05>>2] = $105;
      $106 = ((($src$04)) + 4|0);
      $107 = ((($dst$05)) + 4|0);
      $108 = (($109) + -1)|0;
      $110 = ($109|0)==(0);
      if ($110) {
       break L6;
      } else {
       $109 = $108;$dst$05 = $107;$src$04 = $106;
      }
     }
    } else {
     $65 = $14;$dst$05$us = $120;$src$04$us = $121;
    }
    while(1) {
     $16 = HEAP32[$src$04$us>>2]|0;
     $17 = $16 >>> 24;
     $18 = $16 >>> 16;
     $19 = $18 & 255;
     $20 = $16 >>> 8;
     $21 = $20 & 255;
     $22 = $16 & 255;
     $23 = HEAP32[$dst$05$us>>2]|0;
     $24 = $23 >>> 16;
     $25 = $24 & 255;
     $26 = $23 >>> 8;
     $27 = $26 & 255;
     $28 = $23 & 255;
     $29 = ($22|0)==(255);
     if ($29) {
      $srcB$0$us = $21;$srcG$0$us = $19;$srcR$0$us = $17;
     } else {
      $30 = Math_imul($17, $22)|0;
      $31 = (($30>>>0) / 255)&-1;
      $32 = Math_imul($19, $22)|0;
      $33 = (($32>>>0) / 255)&-1;
      $34 = Math_imul($21, $22)|0;
      $35 = (($34>>>0) / 255)&-1;
      $srcB$0$us = $35;$srcG$0$us = $33;$srcR$0$us = $31;
     }
     switch ($12|0) {
     case 16:  {
      $48 = $22 ^ 255;
      $49 = Math_imul($25, $48)|0;
      $50 = (($49>>>0) / 255)&-1;
      $51 = (($srcR$0$us) + ($50))|0;
      $52 = Math_imul($27, $48)|0;
      $53 = (($52>>>0) / 255)&-1;
      $54 = (($srcG$0$us) + ($53))|0;
      $55 = Math_imul($48, $28)|0;
      $56 = (($55>>>0) / 255)&-1;
      $57 = (($srcB$0$us) + ($56))|0;
      $dstB$0$us = $57;$dstG$0$us = $54;$dstR$0$us = $51;
      break;
     }
     case 32:  {
      $42 = (($srcR$0$us) + ($25))|0;
      $43 = ($42>>>0)>(255);
      $$$us = $43 ? 255 : $42;
      $44 = (($srcG$0$us) + ($27))|0;
      $45 = ($44>>>0)>(255);
      $$3$us = $45 ? 255 : $44;
      $46 = (($srcB$0$us) + ($28))|0;
      $47 = ($46>>>0)>(255);
      $$1$us = $47 ? 255 : $46;
      $dstB$0$us = $$1$us;$dstG$0$us = $$3$us;$dstR$0$us = $$$us;
      break;
     }
     case 64:  {
      $36 = Math_imul($srcR$0$us, $25)|0;
      $37 = (($36>>>0) / 255)&-1;
      $38 = Math_imul($srcG$0$us, $27)|0;
      $39 = (($38>>>0) / 255)&-1;
      $40 = Math_imul($srcB$0$us, $28)|0;
      $41 = (($40>>>0) / 255)&-1;
      $dstB$0$us = $41;$dstG$0$us = $39;$dstR$0$us = $37;
      break;
     }
     default: {
      $dstB$0$us = $28;$dstG$0$us = $27;$dstR$0$us = $25;
     }
     }
     $58 = $dstR$0$us << 16;
     $59 = $dstG$0$us << 8;
     $60 = $59 | $58;
     $61 = $60 | $dstB$0$us;
     HEAP32[$dst$05$us>>2] = $61;
     $62 = ((($src$04$us)) + 4|0);
     $63 = ((($dst$05$us)) + 4|0);
     $64 = (($65) + -1)|0;
     $66 = ($65|0)==(0);
     if ($66) {
      break L6;
     } else {
      $65 = $64;$dst$05$us = $63;$src$04$us = $62;
     }
    }
   }
  } while(0);
  $111 = HEAP32[$8>>2]|0;
  $112 = (($113) + ($111)|0);
  $114 = HEAP32[$9>>2]|0;
  $115 = (($116) + ($114)|0);
  $117 = HEAP32[$2>>2]|0;
  $118 = (($117) + -1)|0;
  HEAP32[$2>>2] = $118;
  $119 = ($117|0)==(0);
  if ($119) {
   $$lcssa = $112;$$lcssa17 = $115;
   break;
  } else {
   $113 = $112;$116 = $115;$120 = $115;$121 = $112;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa17;
 return;
}
function _SDL_Blit_RGBA8888_RGB888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0;
 var $src$015 = 0, $src$1 = 0, $srcB$0 = 0, $srcG$0 = 0, $srcR$0 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 48;
 $19 = ($18|0)!=(0);
 $20 = $1 & 112;
 $21 = ((($info)) + 12|0);
 $$pre = HEAP32[$16>>2]|0;
 $101 = $14;$102 = $$pre;$30 = $12;$96 = $$pre;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $22 = ($posy$020|0)>(65535);
  if ($22) {
   $23 = (($posy$020) + -65536)|0;
   $24 = $23 >>> 16;
   $25 = $24 << 16;
   $26 = (($srcy$019) + 1)|0;
   $27 = (($23) - ($25))|0;
   $28 = (($26) + ($24))|0;
   $posy$1$lcssa = $27;$srcy$1$lcssa = $28;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $29 = ($30|0)==(0);
  if ($29) {
   $98 = $101;
  } else {
   $$in = $30;$dst$016 = $102;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $31 = (($$in) + -1)|0;
    $32 = ($posx$014|0)>(65535);
    if ($32) {
     $33 = (($posx$014) + -65536)|0;
     $34 = $33 >>> 16;
     $35 = $34 << 16;
     $36 = (($srcx$013) + 1)|0;
     $37 = (($33) - ($35))|0;
     $38 = (($36) + ($34))|0;
     $39 = HEAP32[$info>>2]|0;
     $40 = HEAP32[$21>>2]|0;
     $41 = Math_imul($40, $srcy$1$lcssa)|0;
     $42 = (($39) + ($41)|0);
     $43 = $38 << 2;
     $44 = (($42) + ($43)|0);
     $posx$2 = $37;$src$1 = $44;$srcx$2 = $38;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $45 = HEAP32[$src$1>>2]|0;
    $46 = $45 >>> 24;
    $47 = $45 >>> 16;
    $48 = $47 & 255;
    $49 = $45 >>> 8;
    $50 = $49 & 255;
    $51 = $45 & 255;
    $52 = HEAP32[$dst$016>>2]|0;
    $53 = $52 >>> 16;
    $54 = $53 & 255;
    $55 = $52 >>> 8;
    $56 = $55 & 255;
    $57 = $52 & 255;
    $58 = ($51|0)!=(255);
    $or$cond = $19 & $58;
    if ($or$cond) {
     $59 = Math_imul($46, $51)|0;
     $60 = (($59>>>0) / 255)&-1;
     $61 = Math_imul($48, $51)|0;
     $62 = (($61>>>0) / 255)&-1;
     $63 = Math_imul($50, $51)|0;
     $64 = (($63>>>0) / 255)&-1;
     $srcB$0 = $64;$srcG$0 = $62;$srcR$0 = $60;
    } else {
     $srcB$0 = $50;$srcG$0 = $48;$srcR$0 = $46;
    }
    switch ($20|0) {
    case 16:  {
     $65 = $51 ^ 255;
     $66 = Math_imul($54, $65)|0;
     $67 = (($66>>>0) / 255)&-1;
     $68 = (($srcR$0) + ($67))|0;
     $69 = Math_imul($56, $65)|0;
     $70 = (($69>>>0) / 255)&-1;
     $71 = (($srcG$0) + ($70))|0;
     $72 = Math_imul($65, $57)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = (($srcB$0) + ($73))|0;
     $dstB$0 = $74;$dstG$0 = $71;$dstR$0 = $68;
     break;
    }
    case 32:  {
     $75 = (($srcR$0) + ($54))|0;
     $76 = ($75>>>0)>(255);
     $$ = $76 ? 255 : $75;
     $77 = (($srcG$0) + ($56))|0;
     $78 = ($77>>>0)>(255);
     $$3 = $78 ? 255 : $77;
     $79 = (($srcB$0) + ($57))|0;
     $80 = ($79>>>0)>(255);
     $$1 = $80 ? 255 : $79;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $81 = Math_imul($srcR$0, $54)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = Math_imul($srcG$0, $56)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = Math_imul($srcB$0, $57)|0;
     $86 = (($85>>>0) / 255)&-1;
     $dstB$0 = $86;$dstG$0 = $84;$dstR$0 = $82;
     break;
    }
    default: {
     $dstB$0 = $57;$dstG$0 = $56;$dstR$0 = $54;
    }
    }
    $87 = $dstR$0 << 16;
    $88 = $dstG$0 << 8;
    $89 = $88 | $87;
    $90 = $89 | $dstB$0;
    HEAP32[$dst$016>>2] = $90;
    $91 = (($posx$2) + ($13))|0;
    $92 = ((($dst$016)) + 4|0);
    $93 = ($31|0)==(0);
    if ($93) {
     break;
    } else {
     $$in = $31;$dst$016 = $92;$posx$014 = $91;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$5>>2]|0;
   $98 = $$pre28;
  }
  $94 = HEAP32[$17>>2]|0;
  $95 = (($96) + ($94)|0);
  $97 = (($98) + -1)|0;
  HEAP32[$5>>2] = $97;
  $99 = ($98|0)==(0);
  if ($99) {
   $$lcssa = $95;
   break;
  }
  $100 = (($posy$1$lcssa) + ($7))|0;
  $$pre27 = HEAP32[$11>>2]|0;
  $101 = $97;$102 = $95;$30 = $$pre27;$96 = $95;$posy$020 = $100;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGBA8888_RGB888_Modulate($info) {
 $info = $info|0;
 var $$in = 0, $$in13 = 0, $$lcssa = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$pre16 = 0, $$pre17 = 0, $$pre18 = 0, $$pre19 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $dst$02 = 0, $dst$02$us$us = 0, $src$01 = 0, $src$01$us$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = ((($info)) + 20|0);
 $14 = ((($info)) + 24|0);
 $15 = ((($info)) + 12|0);
 $16 = ((($info)) + 32|0);
 $17 = $1 & 1;
 $18 = ($17|0)==(0);
 $$pre17 = HEAP32[$info>>2]|0;
 $$pre18 = HEAP32[$13>>2]|0;
 if ($18) {
  $23 = $$pre17;$26 = $$pre18;$73 = $11;$74 = $$pre18;$75 = $$pre17;
  while(1) {
   $19 = HEAP32[$14>>2]|0;
   $20 = ($19|0)==(0);
   if ($20) {
    $28 = $73;
   } else {
    $$in13 = $19;$dst$02$us$us = $74;$src$01$us$us = $75;
    while(1) {
     $30 = (($$in13) + -1)|0;
     $31 = HEAP32[$src$01$us$us>>2]|0;
     $32 = $31 >>> 24;
     $33 = $31 >>> 8;
     $34 = $33 & 255;
     $35 = $32 << 16;
     $36 = $33 & 65280;
     $37 = $36 | $35;
     $38 = $37 | $34;
     HEAP32[$dst$02$us$us>>2] = $38;
     $39 = ((($src$01$us$us)) + 4|0);
     $40 = ((($dst$02$us$us)) + 4|0);
     $41 = ($30|0)==(0);
     if ($41) {
      break;
     } else {
      $$in13 = $30;$dst$02$us$us = $40;$src$01$us$us = $39;
     }
    }
    $$pre19 = HEAP32[$9>>2]|0;
    $28 = $$pre19;
   }
   $21 = HEAP32[$15>>2]|0;
   $22 = (($23) + ($21)|0);
   $24 = HEAP32[$16>>2]|0;
   $25 = (($26) + ($24)|0);
   $27 = (($28) + -1)|0;
   HEAP32[$9>>2] = $27;
   $29 = ($28|0)==(0);
   if ($29) {
    $$lcssa = $22;$$lcssa30 = $25;
    break;
   } else {
    $23 = $22;$26 = $25;$73 = $27;$74 = $25;$75 = $22;
   }
  }
  HEAP32[$info>>2] = $$lcssa;
  HEAP32[$13>>2] = $$lcssa30;
  return;
 } else {
  $66 = $$pre17;$69 = $$pre18;$76 = $11;$77 = $$pre18;$78 = $$pre17;
  while(1) {
   $42 = HEAP32[$14>>2]|0;
   $43 = ($42|0)==(0);
   if ($43) {
    $71 = $76;
   } else {
    $$in = $42;$dst$02 = $77;$src$01 = $78;
    while(1) {
     $44 = (($$in) + -1)|0;
     $45 = HEAP32[$src$01>>2]|0;
     $46 = $45 >>> 24;
     $47 = $45 >>> 16;
     $48 = $47 & 255;
     $49 = $45 >>> 8;
     $50 = $49 & 255;
     $51 = Math_imul($46, $4)|0;
     $52 = (($51>>>0) / 255)&-1;
     $53 = Math_imul($48, $6)|0;
     $54 = (($53>>>0) / 255)&-1;
     $55 = Math_imul($50, $8)|0;
     $56 = (($55>>>0) / 255)&-1;
     $57 = $52 << 16;
     $58 = $54 << 8;
     $59 = $58 | $57;
     $60 = $59 | $56;
     HEAP32[$dst$02>>2] = $60;
     $61 = ((($src$01)) + 4|0);
     $62 = ((($dst$02)) + 4|0);
     $63 = ($44|0)==(0);
     if ($63) {
      break;
     } else {
      $$in = $44;$dst$02 = $62;$src$01 = $61;
     }
    }
    $$pre16 = HEAP32[$9>>2]|0;
    $71 = $$pre16;
   }
   $64 = HEAP32[$15>>2]|0;
   $65 = (($66) + ($64)|0);
   $67 = HEAP32[$16>>2]|0;
   $68 = (($69) + ($67)|0);
   $70 = (($71) + -1)|0;
   HEAP32[$9>>2] = $70;
   $72 = ($71|0)==(0);
   if ($72) {
    $$lcssa31 = $65;$$lcssa32 = $68;
    break;
   } else {
    $66 = $65;$69 = $68;$76 = $70;$77 = $68;$78 = $65;
   }
  }
  HEAP32[$info>>2] = $$lcssa31;
  HEAP32[$13>>2] = $$lcssa32;
  return;
 }
}
function _SDL_Blit_RGBA8888_RGB888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$013 = 0, $dst$013$us = 0, $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$017 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0, $src$1$us = 0, $srcx$010 = 0;
 var $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$016 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = ((($info)) + 20|0);
 $24 = ((($info)) + 32|0);
 $25 = $1 & 1;
 $26 = ($25|0)==(0);
 $27 = ((($info)) + 12|0);
 $$pre = HEAP32[$23>>2]|0;
 $103 = $$pre;$36 = $19;$98 = $$pre;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $28 = ($posy$017|0)>(65535);
  if ($28) {
   $29 = (($posy$017) + -65536)|0;
   $30 = $29 >>> 16;
   $31 = $30 << 16;
   $32 = (($srcy$016) + 1)|0;
   $33 = (($29) - ($31))|0;
   $34 = (($32) + ($30))|0;
   $posy$1$lcssa = $33;$srcy$1$lcssa = $34;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $35 = (($36) + -1)|0;
  $37 = ($36|0)==(0);
  if (!($37)) {
   if ($26) {
    $60 = $35;$dst$013$us = $103;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $38 = ($posx$011$us|0)>(65535);
     if ($38) {
      $39 = (($posx$011$us) + -65536)|0;
      $40 = $39 >>> 16;
      $41 = $40 << 16;
      $42 = (($srcx$010$us) + 1)|0;
      $43 = (($39) - ($41))|0;
      $44 = (($42) + ($40))|0;
      $45 = HEAP32[$info>>2]|0;
      $46 = HEAP32[$27>>2]|0;
      $47 = Math_imul($46, $srcy$1$lcssa)|0;
      $48 = (($45) + ($47)|0);
      $49 = $44 << 2;
      $50 = (($48) + ($49)|0);
      $posx$2$us = $43;$src$1$us = $50;$srcx$2$us = $44;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $51 = HEAP32[$src$1$us>>2]|0;
     $52 = $51 >>> 24;
     $53 = $51 >>> 8;
     $54 = $52 << 16;
     $55 = $53 & 65535;
     $56 = $55 | $54;
     HEAP32[$dst$013$us>>2] = $56;
     $57 = (($posx$2$us) + ($20))|0;
     $58 = ((($dst$013$us)) + 4|0);
     $59 = (($60) + -1)|0;
     $61 = ($60|0)==(0);
     if ($61) {
      break;
     } else {
      $60 = $59;$dst$013$us = $58;$posx$011$us = $57;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $94 = $35;$dst$013 = $103;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $62 = ($posx$011|0)>(65535);
     if ($62) {
      $63 = (($posx$011) + -65536)|0;
      $64 = $63 >>> 16;
      $65 = $64 << 16;
      $66 = (($srcx$010) + 1)|0;
      $67 = (($63) - ($65))|0;
      $68 = (($66) + ($64))|0;
      $69 = HEAP32[$info>>2]|0;
      $70 = HEAP32[$27>>2]|0;
      $71 = Math_imul($70, $srcy$1$lcssa)|0;
      $72 = (($69) + ($71)|0);
      $73 = $68 << 2;
      $74 = (($72) + ($73)|0);
      $posx$2 = $67;$src$1 = $74;$srcx$2 = $68;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $75 = HEAP32[$src$1>>2]|0;
     $76 = $75 >>> 24;
     $77 = $75 >>> 16;
     $78 = $77 & 255;
     $79 = $75 >>> 8;
     $80 = $79 & 255;
     $81 = Math_imul($76, $4)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = Math_imul($78, $6)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = Math_imul($80, $8)|0;
     $86 = (($85>>>0) / 255)&-1;
     $87 = $82 << 16;
     $88 = $84 << 8;
     $89 = $86 | $88;
     $90 = $89 | $87;
     HEAP32[$dst$013>>2] = $90;
     $91 = (($posx$2) + ($20))|0;
     $92 = ((($dst$013)) + 4|0);
     $93 = (($94) + -1)|0;
     $95 = ($94|0)==(0);
     if ($95) {
      break;
     } else {
      $94 = $93;$dst$013 = $92;$posx$011 = $91;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $96 = HEAP32[$24>>2]|0;
  $97 = (($98) + ($96)|0);
  $99 = HEAP32[$12>>2]|0;
  $100 = (($99) + -1)|0;
  HEAP32[$12>>2] = $100;
  $101 = ($99|0)==(0);
  if ($101) {
   $$lcssa = $97;
   break;
  }
  $102 = (($posy$1$lcssa) + ($14))|0;
  $$pre27 = HEAP32[$18>>2]|0;
  $103 = $97;$36 = $$pre27;$98 = $97;$posy$017 = $102;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$23>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGBA8888_RGB888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$lcssa14 = 0, $$pre = 0, $$pre10 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $dst$05 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $src$04 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + -1)|0;
 HEAP32[$10>>2] = $12;
 $13 = ($11|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre9 = HEAP32[$14>>2]|0;
 $87 = $$pre;$90 = $$pre9;$94 = $12;$95 = $$pre9;$96 = $$pre;
 while(1) {
  $25 = HEAP32[$15>>2]|0;
  $26 = ($25|0)==(0);
  if ($26) {
   $92 = $94;
  } else {
   $$in = $25;$dst$05 = $95;$src$04 = $96;
   while(1) {
    $27 = (($$in) + -1)|0;
    $28 = HEAP32[$src$04>>2]|0;
    $29 = $28 >>> 24;
    $30 = $28 >>> 16;
    $31 = $30 & 255;
    $32 = $28 >>> 8;
    $33 = $32 & 255;
    $34 = $28 & 255;
    $35 = HEAP32[$dst$05>>2]|0;
    $36 = $35 >>> 16;
    $37 = $36 & 255;
    $38 = $35 >>> 8;
    $39 = $38 & 255;
    $40 = $35 & 255;
    if ($19) {
     $srcB$0 = $33;$srcG$0 = $31;$srcR$0 = $29;
    } else {
     $41 = Math_imul($29, $4)|0;
     $42 = (($41>>>0) / 255)&-1;
     $43 = Math_imul($31, $6)|0;
     $44 = (($43>>>0) / 255)&-1;
     $45 = Math_imul($33, $8)|0;
     $46 = (($45>>>0) / 255)&-1;
     $srcB$0 = $46;$srcG$0 = $44;$srcR$0 = $42;
    }
    if ($21) {
     $srcA$0 = $34;
    } else {
     $47 = Math_imul($34, $9)|0;
     $48 = (($47>>>0) / 255)&-1;
     $srcA$0 = $48;
    }
    $49 = ($srcA$0>>>0)<(255);
    $or$cond = $23 & $49;
    if ($or$cond) {
     $50 = Math_imul($srcA$0, $srcR$0)|0;
     $51 = (($50>>>0) / 255)&-1;
     $52 = Math_imul($srcA$0, $srcG$0)|0;
     $53 = (($52>>>0) / 255)&-1;
     $54 = Math_imul($srcA$0, $srcB$0)|0;
     $55 = (($54>>>0) / 255)&-1;
     $srcB$1 = $55;$srcG$1 = $53;$srcR$1 = $51;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($24|0) {
    case 16:  {
     $56 = (255 - ($srcA$0))|0;
     $57 = Math_imul($56, $37)|0;
     $58 = (($57>>>0) / 255)&-1;
     $59 = (($srcR$1) + ($58))|0;
     $60 = Math_imul($56, $39)|0;
     $61 = (($60>>>0) / 255)&-1;
     $62 = (($srcG$1) + ($61))|0;
     $63 = Math_imul($56, $40)|0;
     $64 = (($63>>>0) / 255)&-1;
     $65 = (($srcB$1) + ($64))|0;
     $dstB$0 = $65;$dstG$0 = $62;$dstR$0 = $59;
     break;
    }
    case 32:  {
     $66 = (($srcR$1) + ($37))|0;
     $67 = ($66>>>0)>(255);
     $$ = $67 ? 255 : $66;
     $68 = (($srcG$1) + ($39))|0;
     $69 = ($68>>>0)>(255);
     $$3 = $69 ? 255 : $68;
     $70 = (($srcB$1) + ($40))|0;
     $71 = ($70>>>0)>(255);
     $$1 = $71 ? 255 : $70;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $72 = Math_imul($srcR$1, $37)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($srcG$1, $39)|0;
     $75 = (($74>>>0) / 255)&-1;
     $76 = Math_imul($srcB$1, $40)|0;
     $77 = (($76>>>0) / 255)&-1;
     $dstB$0 = $77;$dstG$0 = $75;$dstR$0 = $73;
     break;
    }
    default: {
     $dstB$0 = $40;$dstG$0 = $39;$dstR$0 = $37;
    }
    }
    $78 = $dstR$0 << 16;
    $79 = $dstG$0 << 8;
    $80 = $79 | $78;
    $81 = $80 | $dstB$0;
    HEAP32[$dst$05>>2] = $81;
    $82 = ((($src$04)) + 4|0);
    $83 = ((($dst$05)) + 4|0);
    $84 = ($27|0)==(0);
    if ($84) {
     break;
    } else {
     $$in = $27;$dst$05 = $83;$src$04 = $82;
    }
   }
   $$pre10 = HEAP32[$10>>2]|0;
   $92 = $$pre10;
  }
  $85 = HEAP32[$16>>2]|0;
  $86 = (($87) + ($85)|0);
  $88 = HEAP32[$17>>2]|0;
  $89 = (($90) + ($88)|0);
  $91 = (($92) + -1)|0;
  HEAP32[$10>>2] = $91;
  $93 = ($92|0)==(0);
  if ($93) {
   $$lcssa = $86;$$lcssa14 = $89;
   break;
  } else {
   $87 = $86;$90 = $89;$94 = $91;$95 = $89;$96 = $86;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa14;
 return;
}
function _SDL_Blit_RGBA8888_RGB888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0;
 var $src$015 = 0, $src$1 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 << 16;
 $13 = ((($info)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12|0) / ($14|0))&-1;
 $16 = ((($info)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 << 16;
 $19 = ((($info)) + 24|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($18|0) / ($20|0))&-1;
 $22 = (($14) + -1)|0;
 HEAP32[$13>>2] = $22;
 $23 = ($14|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = $1 & 112;
 $33 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $116 = $$pre;$121 = $22;$122 = $$pre;$42 = $20;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $34 = ($posy$020|0)>(65535);
  if ($34) {
   $35 = (($posy$020) + -65536)|0;
   $36 = $35 >>> 16;
   $37 = $36 << 16;
   $38 = (($srcy$019) + 1)|0;
   $39 = (($35) - ($37))|0;
   $40 = (($38) + ($36))|0;
   $posy$1$lcssa = $39;$srcy$1$lcssa = $40;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $41 = ($42|0)==(0);
  if ($41) {
   $118 = $121;
  } else {
   $$in = $42;$dst$016 = $122;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $43 = (($$in) + -1)|0;
    $44 = ($posx$014|0)>(65535);
    if ($44) {
     $45 = (($posx$014) + -65536)|0;
     $46 = $45 >>> 16;
     $47 = $46 << 16;
     $48 = (($srcx$013) + 1)|0;
     $49 = (($45) - ($47))|0;
     $50 = (($48) + ($46))|0;
     $51 = HEAP32[$info>>2]|0;
     $52 = HEAP32[$33>>2]|0;
     $53 = Math_imul($52, $srcy$1$lcssa)|0;
     $54 = (($51) + ($53)|0);
     $55 = $50 << 2;
     $56 = (($54) + ($55)|0);
     $posx$2 = $49;$src$1 = $56;$srcx$2 = $50;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $57 = HEAP32[$src$1>>2]|0;
    $58 = $57 >>> 24;
    $59 = $57 >>> 16;
    $60 = $59 & 255;
    $61 = $57 >>> 8;
    $62 = $61 & 255;
    $63 = $57 & 255;
    $64 = HEAP32[$dst$016>>2]|0;
    $65 = $64 >>> 16;
    $66 = $65 & 255;
    $67 = $64 >>> 8;
    $68 = $67 & 255;
    $69 = $64 & 255;
    if ($27) {
     $srcB$0 = $62;$srcG$0 = $60;$srcR$0 = $58;
    } else {
     $70 = Math_imul($58, $4)|0;
     $71 = (($70>>>0) / 255)&-1;
     $72 = Math_imul($60, $6)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($62, $8)|0;
     $75 = (($74>>>0) / 255)&-1;
     $srcB$0 = $75;$srcG$0 = $73;$srcR$0 = $71;
    }
    if ($29) {
     $srcA$0 = $63;
    } else {
     $76 = Math_imul($63, $9)|0;
     $77 = (($76>>>0) / 255)&-1;
     $srcA$0 = $77;
    }
    $78 = ($srcA$0>>>0)<(255);
    $or$cond = $31 & $78;
    if ($or$cond) {
     $79 = Math_imul($srcA$0, $srcR$0)|0;
     $80 = (($79>>>0) / 255)&-1;
     $81 = Math_imul($srcA$0, $srcG$0)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = Math_imul($srcA$0, $srcB$0)|0;
     $84 = (($83>>>0) / 255)&-1;
     $srcB$1 = $84;$srcG$1 = $82;$srcR$1 = $80;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($32|0) {
    case 16:  {
     $85 = (255 - ($srcA$0))|0;
     $86 = Math_imul($85, $66)|0;
     $87 = (($86>>>0) / 255)&-1;
     $88 = (($srcR$1) + ($87))|0;
     $89 = Math_imul($85, $68)|0;
     $90 = (($89>>>0) / 255)&-1;
     $91 = (($srcG$1) + ($90))|0;
     $92 = Math_imul($85, $69)|0;
     $93 = (($92>>>0) / 255)&-1;
     $94 = (($srcB$1) + ($93))|0;
     $dstB$0 = $94;$dstG$0 = $91;$dstR$0 = $88;
     break;
    }
    case 32:  {
     $95 = (($srcR$1) + ($66))|0;
     $96 = ($95>>>0)>(255);
     $$ = $96 ? 255 : $95;
     $97 = (($srcG$1) + ($68))|0;
     $98 = ($97>>>0)>(255);
     $$3 = $98 ? 255 : $97;
     $99 = (($srcB$1) + ($69))|0;
     $100 = ($99>>>0)>(255);
     $$1 = $100 ? 255 : $99;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $101 = Math_imul($srcR$1, $66)|0;
     $102 = (($101>>>0) / 255)&-1;
     $103 = Math_imul($srcG$1, $68)|0;
     $104 = (($103>>>0) / 255)&-1;
     $105 = Math_imul($srcB$1, $69)|0;
     $106 = (($105>>>0) / 255)&-1;
     $dstB$0 = $106;$dstG$0 = $104;$dstR$0 = $102;
     break;
    }
    default: {
     $dstB$0 = $69;$dstG$0 = $68;$dstR$0 = $66;
    }
    }
    $107 = $dstR$0 << 16;
    $108 = $dstG$0 << 8;
    $109 = $108 | $107;
    $110 = $109 | $dstB$0;
    HEAP32[$dst$016>>2] = $110;
    $111 = (($posx$2) + ($21))|0;
    $112 = ((($dst$016)) + 4|0);
    $113 = ($43|0)==(0);
    if ($113) {
     break;
    } else {
     $$in = $43;$dst$016 = $112;$posx$014 = $111;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$13>>2]|0;
   $118 = $$pre28;
  }
  $114 = HEAP32[$25>>2]|0;
  $115 = (($116) + ($114)|0);
  $117 = (($118) + -1)|0;
  HEAP32[$13>>2] = $117;
  $119 = ($118|0)==(0);
  if ($119) {
   $$lcssa = $115;
   break;
  }
  $120 = (($posy$1$lcssa) + ($15))|0;
  $$pre27 = HEAP32[$19>>2]|0;
  $116 = $115;$121 = $117;$122 = $115;$42 = $$pre27;$posy$020 = $120;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGBA8888_BGR888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0, $posx$010 = 0, $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$53 = $$pre;$58 = $12;$59 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $55 = $58;
  } else {
   $$in = $25;$dst$013 = $59;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    $41 = $40 >>> 24;
    $42 = $40 << 8;
    $43 = $42 & 16711680;
    $44 = $40 >>> 8;
    $45 = $44 & 65280;
    $46 = $45 | $41;
    $47 = $46 | $43;
    HEAP32[$dst$013>>2] = $47;
    $48 = (($posx$2) + ($11))|0;
    $49 = ((($dst$013)) + 4|0);
    $50 = ($26|0)==(0);
    if ($50) {
     break;
    } else {
     $$in = $26;$dst$013 = $49;$posx$010 = $48;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $55 = $$pre25;
  }
  $51 = HEAP32[$15>>2]|0;
  $52 = (($53) + ($51)|0);
  $54 = (($55) + -1)|0;
  HEAP32[$3>>2] = $54;
  $56 = ($55|0)==(0);
  if ($56) {
   $$lcssa = $52;
   break;
  }
  $57 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$53 = $52;$58 = $54;$59 = $52;$posy$016 = $57;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGBA8888_BGR888_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$$us = 0, $$1 = 0, $$1$us = 0, $$3 = 0, $$3$us = 0, $$lcssa = 0, $$lcssa17 = 0, $$pre = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$05 = 0, $dst$05$us = 0, $dstB$0 = 0, $dstB$0$us = 0, $dstG$0 = 0, $dstG$0$us = 0, $dstR$0 = 0, $dstR$0$us = 0;
 var $src$04 = 0, $src$04$us = 0, $srcB$0$us = 0, $srcG$0$us = 0, $srcR$0$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 48;
 $11 = ($10|0)==(0);
 $12 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre10 = HEAP32[$6>>2]|0;
 $113 = $$pre;$116 = $$pre10;$120 = $$pre10;$121 = $$pre;
 while(1) {
  $13 = HEAP32[$7>>2]|0;
  $14 = (($13) + -1)|0;
  $15 = ($13|0)==(0);
  L6: do {
   if (!($15)) {
    if ($11) {
     $109 = $14;$dst$05 = $120;$src$04 = $121;
     while(1) {
      $67 = HEAP32[$src$04>>2]|0;
      $68 = $67 >>> 24;
      $69 = $67 >>> 16;
      $70 = $69 & 255;
      $71 = $67 >>> 8;
      $72 = $71 & 255;
      $73 = HEAP32[$dst$05>>2]|0;
      $74 = $73 >>> 16;
      $75 = $74 & 255;
      $76 = $73 >>> 8;
      $77 = $76 & 255;
      $78 = $73 & 255;
      switch ($12|0) {
      case 16:  {
       $79 = $67 & 255;
       $80 = $79 ^ 255;
       $81 = Math_imul($80, $78)|0;
       $82 = (($81>>>0) / 255)&-1;
       $83 = (($68) + ($82))|0;
       $84 = Math_imul($77, $80)|0;
       $85 = (($84>>>0) / 255)&-1;
       $86 = (($70) + ($85))|0;
       $87 = Math_imul($75, $80)|0;
       $88 = (($87>>>0) / 255)&-1;
       $89 = (($72) + ($88))|0;
       $dstB$0 = $89;$dstG$0 = $86;$dstR$0 = $83;
       break;
      }
      case 32:  {
       $90 = (($68) + ($78))|0;
       $91 = ($90>>>0)>(255);
       $$ = $91 ? 255 : $90;
       $92 = (($70) + ($77))|0;
       $93 = ($92>>>0)>(255);
       $$3 = $93 ? 255 : $92;
       $94 = (($72) + ($75))|0;
       $95 = ($94>>>0)>(255);
       $$1 = $95 ? 255 : $94;
       $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
       break;
      }
      case 64:  {
       $96 = Math_imul($68, $78)|0;
       $97 = (($96>>>0) / 255)&-1;
       $98 = Math_imul($70, $77)|0;
       $99 = (($98>>>0) / 255)&-1;
       $100 = Math_imul($72, $75)|0;
       $101 = (($100>>>0) / 255)&-1;
       $dstB$0 = $101;$dstG$0 = $99;$dstR$0 = $97;
       break;
      }
      default: {
       $dstB$0 = $75;$dstG$0 = $77;$dstR$0 = $78;
      }
      }
      $102 = $dstB$0 << 16;
      $103 = $dstG$0 << 8;
      $104 = $103 | $dstR$0;
      $105 = $104 | $102;
      HEAP32[$dst$05>>2] = $105;
      $106 = ((($src$04)) + 4|0);
      $107 = ((($dst$05)) + 4|0);
      $108 = (($109) + -1)|0;
      $110 = ($109|0)==(0);
      if ($110) {
       break L6;
      } else {
       $109 = $108;$dst$05 = $107;$src$04 = $106;
      }
     }
    } else {
     $65 = $14;$dst$05$us = $120;$src$04$us = $121;
    }
    while(1) {
     $16 = HEAP32[$src$04$us>>2]|0;
     $17 = $16 >>> 24;
     $18 = $16 >>> 16;
     $19 = $18 & 255;
     $20 = $16 >>> 8;
     $21 = $20 & 255;
     $22 = $16 & 255;
     $23 = HEAP32[$dst$05$us>>2]|0;
     $24 = $23 >>> 16;
     $25 = $24 & 255;
     $26 = $23 >>> 8;
     $27 = $26 & 255;
     $28 = $23 & 255;
     $29 = ($22|0)==(255);
     if ($29) {
      $srcB$0$us = $21;$srcG$0$us = $19;$srcR$0$us = $17;
     } else {
      $30 = Math_imul($17, $22)|0;
      $31 = (($30>>>0) / 255)&-1;
      $32 = Math_imul($19, $22)|0;
      $33 = (($32>>>0) / 255)&-1;
      $34 = Math_imul($21, $22)|0;
      $35 = (($34>>>0) / 255)&-1;
      $srcB$0$us = $35;$srcG$0$us = $33;$srcR$0$us = $31;
     }
     switch ($12|0) {
     case 16:  {
      $48 = $22 ^ 255;
      $49 = Math_imul($48, $28)|0;
      $50 = (($49>>>0) / 255)&-1;
      $51 = (($srcR$0$us) + ($50))|0;
      $52 = Math_imul($27, $48)|0;
      $53 = (($52>>>0) / 255)&-1;
      $54 = (($srcG$0$us) + ($53))|0;
      $55 = Math_imul($25, $48)|0;
      $56 = (($55>>>0) / 255)&-1;
      $57 = (($srcB$0$us) + ($56))|0;
      $dstB$0$us = $57;$dstG$0$us = $54;$dstR$0$us = $51;
      break;
     }
     case 32:  {
      $42 = (($srcR$0$us) + ($28))|0;
      $43 = ($42>>>0)>(255);
      $$$us = $43 ? 255 : $42;
      $44 = (($srcG$0$us) + ($27))|0;
      $45 = ($44>>>0)>(255);
      $$3$us = $45 ? 255 : $44;
      $46 = (($srcB$0$us) + ($25))|0;
      $47 = ($46>>>0)>(255);
      $$1$us = $47 ? 255 : $46;
      $dstB$0$us = $$1$us;$dstG$0$us = $$3$us;$dstR$0$us = $$$us;
      break;
     }
     case 64:  {
      $36 = Math_imul($srcR$0$us, $28)|0;
      $37 = (($36>>>0) / 255)&-1;
      $38 = Math_imul($srcG$0$us, $27)|0;
      $39 = (($38>>>0) / 255)&-1;
      $40 = Math_imul($srcB$0$us, $25)|0;
      $41 = (($40>>>0) / 255)&-1;
      $dstB$0$us = $41;$dstG$0$us = $39;$dstR$0$us = $37;
      break;
     }
     default: {
      $dstB$0$us = $25;$dstG$0$us = $27;$dstR$0$us = $28;
     }
     }
     $58 = $dstB$0$us << 16;
     $59 = $dstG$0$us << 8;
     $60 = $59 | $dstR$0$us;
     $61 = $60 | $58;
     HEAP32[$dst$05$us>>2] = $61;
     $62 = ((($src$04$us)) + 4|0);
     $63 = ((($dst$05$us)) + 4|0);
     $64 = (($65) + -1)|0;
     $66 = ($65|0)==(0);
     if ($66) {
      break L6;
     } else {
      $65 = $64;$dst$05$us = $63;$src$04$us = $62;
     }
    }
   }
  } while(0);
  $111 = HEAP32[$8>>2]|0;
  $112 = (($113) + ($111)|0);
  $114 = HEAP32[$9>>2]|0;
  $115 = (($116) + ($114)|0);
  $117 = HEAP32[$2>>2]|0;
  $118 = (($117) + -1)|0;
  HEAP32[$2>>2] = $118;
  $119 = ($117|0)==(0);
  if ($119) {
   $$lcssa = $112;$$lcssa17 = $115;
   break;
  } else {
   $113 = $112;$116 = $115;$120 = $115;$121 = $112;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa17;
 return;
}
function _SDL_Blit_RGBA8888_BGR888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0;
 var $src$015 = 0, $src$1 = 0, $srcB$0 = 0, $srcG$0 = 0, $srcR$0 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 48;
 $19 = ($18|0)!=(0);
 $20 = $1 & 112;
 $21 = ((($info)) + 12|0);
 $$pre = HEAP32[$16>>2]|0;
 $101 = $14;$102 = $$pre;$30 = $12;$96 = $$pre;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $22 = ($posy$020|0)>(65535);
  if ($22) {
   $23 = (($posy$020) + -65536)|0;
   $24 = $23 >>> 16;
   $25 = $24 << 16;
   $26 = (($srcy$019) + 1)|0;
   $27 = (($23) - ($25))|0;
   $28 = (($26) + ($24))|0;
   $posy$1$lcssa = $27;$srcy$1$lcssa = $28;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $29 = ($30|0)==(0);
  if ($29) {
   $98 = $101;
  } else {
   $$in = $30;$dst$016 = $102;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $31 = (($$in) + -1)|0;
    $32 = ($posx$014|0)>(65535);
    if ($32) {
     $33 = (($posx$014) + -65536)|0;
     $34 = $33 >>> 16;
     $35 = $34 << 16;
     $36 = (($srcx$013) + 1)|0;
     $37 = (($33) - ($35))|0;
     $38 = (($36) + ($34))|0;
     $39 = HEAP32[$info>>2]|0;
     $40 = HEAP32[$21>>2]|0;
     $41 = Math_imul($40, $srcy$1$lcssa)|0;
     $42 = (($39) + ($41)|0);
     $43 = $38 << 2;
     $44 = (($42) + ($43)|0);
     $posx$2 = $37;$src$1 = $44;$srcx$2 = $38;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $45 = HEAP32[$src$1>>2]|0;
    $46 = $45 >>> 24;
    $47 = $45 >>> 16;
    $48 = $47 & 255;
    $49 = $45 >>> 8;
    $50 = $49 & 255;
    $51 = $45 & 255;
    $52 = HEAP32[$dst$016>>2]|0;
    $53 = $52 >>> 16;
    $54 = $53 & 255;
    $55 = $52 >>> 8;
    $56 = $55 & 255;
    $57 = $52 & 255;
    $58 = ($51|0)!=(255);
    $or$cond = $19 & $58;
    if ($or$cond) {
     $59 = Math_imul($46, $51)|0;
     $60 = (($59>>>0) / 255)&-1;
     $61 = Math_imul($48, $51)|0;
     $62 = (($61>>>0) / 255)&-1;
     $63 = Math_imul($50, $51)|0;
     $64 = (($63>>>0) / 255)&-1;
     $srcB$0 = $64;$srcG$0 = $62;$srcR$0 = $60;
    } else {
     $srcB$0 = $50;$srcG$0 = $48;$srcR$0 = $46;
    }
    switch ($20|0) {
    case 16:  {
     $65 = $51 ^ 255;
     $66 = Math_imul($65, $57)|0;
     $67 = (($66>>>0) / 255)&-1;
     $68 = (($srcR$0) + ($67))|0;
     $69 = Math_imul($56, $65)|0;
     $70 = (($69>>>0) / 255)&-1;
     $71 = (($srcG$0) + ($70))|0;
     $72 = Math_imul($54, $65)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = (($srcB$0) + ($73))|0;
     $dstB$0 = $74;$dstG$0 = $71;$dstR$0 = $68;
     break;
    }
    case 32:  {
     $75 = (($srcR$0) + ($57))|0;
     $76 = ($75>>>0)>(255);
     $$ = $76 ? 255 : $75;
     $77 = (($srcG$0) + ($56))|0;
     $78 = ($77>>>0)>(255);
     $$3 = $78 ? 255 : $77;
     $79 = (($srcB$0) + ($54))|0;
     $80 = ($79>>>0)>(255);
     $$1 = $80 ? 255 : $79;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $81 = Math_imul($srcR$0, $57)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = Math_imul($srcG$0, $56)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = Math_imul($srcB$0, $54)|0;
     $86 = (($85>>>0) / 255)&-1;
     $dstB$0 = $86;$dstG$0 = $84;$dstR$0 = $82;
     break;
    }
    default: {
     $dstB$0 = $54;$dstG$0 = $56;$dstR$0 = $57;
    }
    }
    $87 = $dstB$0 << 16;
    $88 = $dstG$0 << 8;
    $89 = $88 | $dstR$0;
    $90 = $89 | $87;
    HEAP32[$dst$016>>2] = $90;
    $91 = (($posx$2) + ($13))|0;
    $92 = ((($dst$016)) + 4|0);
    $93 = ($31|0)==(0);
    if ($93) {
     break;
    } else {
     $$in = $31;$dst$016 = $92;$posx$014 = $91;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$5>>2]|0;
   $98 = $$pre28;
  }
  $94 = HEAP32[$17>>2]|0;
  $95 = (($96) + ($94)|0);
  $97 = (($98) + -1)|0;
  HEAP32[$5>>2] = $97;
  $99 = ($98|0)==(0);
  if ($99) {
   $$lcssa = $95;
   break;
  }
  $100 = (($posy$1$lcssa) + ($7))|0;
  $$pre27 = HEAP32[$11>>2]|0;
  $101 = $97;$102 = $95;$30 = $$pre27;$96 = $95;$posy$020 = $100;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGBA8888_BGR888_Modulate($info) {
 $info = $info|0;
 var $$in = 0, $$in13 = 0, $$lcssa = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$pre16 = 0, $$pre17 = 0, $$pre18 = 0, $$pre19 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $dst$02 = 0, $dst$02$us$us = 0, $src$01 = 0, $src$01$us$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = ((($info)) + 20|0);
 $14 = ((($info)) + 24|0);
 $15 = ((($info)) + 12|0);
 $16 = ((($info)) + 32|0);
 $17 = $1 & 1;
 $18 = ($17|0)==(0);
 $$pre17 = HEAP32[$info>>2]|0;
 $$pre18 = HEAP32[$13>>2]|0;
 if ($18) {
  $23 = $$pre17;$26 = $$pre18;$73 = $11;$74 = $$pre18;$75 = $$pre17;
  while(1) {
   $19 = HEAP32[$14>>2]|0;
   $20 = ($19|0)==(0);
   if ($20) {
    $28 = $73;
   } else {
    $$in13 = $19;$dst$02$us$us = $74;$src$01$us$us = $75;
    while(1) {
     $30 = (($$in13) + -1)|0;
     $31 = HEAP32[$src$01$us$us>>2]|0;
     $32 = $31 >>> 24;
     $33 = $31 << 8;
     $34 = $33 & 16711680;
     $35 = $31 >>> 8;
     $36 = $35 & 65280;
     $37 = $36 | $32;
     $38 = $37 | $34;
     HEAP32[$dst$02$us$us>>2] = $38;
     $39 = ((($src$01$us$us)) + 4|0);
     $40 = ((($dst$02$us$us)) + 4|0);
     $41 = ($30|0)==(0);
     if ($41) {
      break;
     } else {
      $$in13 = $30;$dst$02$us$us = $40;$src$01$us$us = $39;
     }
    }
    $$pre19 = HEAP32[$9>>2]|0;
    $28 = $$pre19;
   }
   $21 = HEAP32[$15>>2]|0;
   $22 = (($23) + ($21)|0);
   $24 = HEAP32[$16>>2]|0;
   $25 = (($26) + ($24)|0);
   $27 = (($28) + -1)|0;
   HEAP32[$9>>2] = $27;
   $29 = ($28|0)==(0);
   if ($29) {
    $$lcssa = $22;$$lcssa30 = $25;
    break;
   } else {
    $23 = $22;$26 = $25;$73 = $27;$74 = $25;$75 = $22;
   }
  }
  HEAP32[$info>>2] = $$lcssa;
  HEAP32[$13>>2] = $$lcssa30;
  return;
 } else {
  $66 = $$pre17;$69 = $$pre18;$76 = $11;$77 = $$pre18;$78 = $$pre17;
  while(1) {
   $42 = HEAP32[$14>>2]|0;
   $43 = ($42|0)==(0);
   if ($43) {
    $71 = $76;
   } else {
    $$in = $42;$dst$02 = $77;$src$01 = $78;
    while(1) {
     $44 = (($$in) + -1)|0;
     $45 = HEAP32[$src$01>>2]|0;
     $46 = $45 >>> 24;
     $47 = $45 >>> 16;
     $48 = $47 & 255;
     $49 = $45 >>> 8;
     $50 = $49 & 255;
     $51 = Math_imul($46, $4)|0;
     $52 = (($51>>>0) / 255)&-1;
     $53 = Math_imul($48, $6)|0;
     $54 = (($53>>>0) / 255)&-1;
     $55 = Math_imul($50, $8)|0;
     $56 = (($55>>>0) / 255)&-1;
     $57 = $56 << 16;
     $58 = $54 << 8;
     $59 = $58 | $52;
     $60 = $59 | $57;
     HEAP32[$dst$02>>2] = $60;
     $61 = ((($src$01)) + 4|0);
     $62 = ((($dst$02)) + 4|0);
     $63 = ($44|0)==(0);
     if ($63) {
      break;
     } else {
      $$in = $44;$dst$02 = $62;$src$01 = $61;
     }
    }
    $$pre16 = HEAP32[$9>>2]|0;
    $71 = $$pre16;
   }
   $64 = HEAP32[$15>>2]|0;
   $65 = (($66) + ($64)|0);
   $67 = HEAP32[$16>>2]|0;
   $68 = (($69) + ($67)|0);
   $70 = (($71) + -1)|0;
   HEAP32[$9>>2] = $70;
   $72 = ($71|0)==(0);
   if ($72) {
    $$lcssa31 = $65;$$lcssa32 = $68;
    break;
   } else {
    $66 = $65;$69 = $68;$76 = $70;$77 = $68;$78 = $65;
   }
  }
  HEAP32[$info>>2] = $$lcssa31;
  HEAP32[$13>>2] = $$lcssa32;
  return;
 }
}
function _SDL_Blit_RGBA8888_BGR888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$013 = 0, $dst$013$us = 0, $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$017 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0;
 var $src$1$us = 0, $srcx$010 = 0, $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$016 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = ((($info)) + 20|0);
 $24 = ((($info)) + 32|0);
 $25 = $1 & 1;
 $26 = ($25|0)==(0);
 $27 = ((($info)) + 12|0);
 $$pre = HEAP32[$23>>2]|0;
 $100 = $$pre;$105 = $$pre;$36 = $19;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $28 = ($posy$017|0)>(65535);
  if ($28) {
   $29 = (($posy$017) + -65536)|0;
   $30 = $29 >>> 16;
   $31 = $30 << 16;
   $32 = (($srcy$016) + 1)|0;
   $33 = (($29) - ($31))|0;
   $34 = (($32) + ($30))|0;
   $posy$1$lcssa = $33;$srcy$1$lcssa = $34;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $35 = (($36) + -1)|0;
  $37 = ($36|0)==(0);
  if (!($37)) {
   if ($26) {
    $62 = $35;$dst$013$us = $105;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $38 = ($posx$011$us|0)>(65535);
     if ($38) {
      $39 = (($posx$011$us) + -65536)|0;
      $40 = $39 >>> 16;
      $41 = $40 << 16;
      $42 = (($srcx$010$us) + 1)|0;
      $43 = (($39) - ($41))|0;
      $44 = (($42) + ($40))|0;
      $45 = HEAP32[$info>>2]|0;
      $46 = HEAP32[$27>>2]|0;
      $47 = Math_imul($46, $srcy$1$lcssa)|0;
      $48 = (($45) + ($47)|0);
      $49 = $44 << 2;
      $50 = (($48) + ($49)|0);
      $posx$2$us = $43;$src$1$us = $50;$srcx$2$us = $44;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $51 = HEAP32[$src$1$us>>2]|0;
     $52 = $51 >>> 24;
     $53 = $51 << 8;
     $54 = $53 & 16711680;
     $55 = $51 >>> 8;
     $56 = $55 & 65280;
     $57 = $54 | $56;
     $58 = $57 | $52;
     HEAP32[$dst$013$us>>2] = $58;
     $59 = (($posx$2$us) + ($20))|0;
     $60 = ((($dst$013$us)) + 4|0);
     $61 = (($62) + -1)|0;
     $63 = ($62|0)==(0);
     if ($63) {
      break;
     } else {
      $62 = $61;$dst$013$us = $60;$posx$011$us = $59;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $96 = $35;$dst$013 = $105;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $64 = ($posx$011|0)>(65535);
     if ($64) {
      $65 = (($posx$011) + -65536)|0;
      $66 = $65 >>> 16;
      $67 = $66 << 16;
      $68 = (($srcx$010) + 1)|0;
      $69 = (($65) - ($67))|0;
      $70 = (($68) + ($66))|0;
      $71 = HEAP32[$info>>2]|0;
      $72 = HEAP32[$27>>2]|0;
      $73 = Math_imul($72, $srcy$1$lcssa)|0;
      $74 = (($71) + ($73)|0);
      $75 = $70 << 2;
      $76 = (($74) + ($75)|0);
      $posx$2 = $69;$src$1 = $76;$srcx$2 = $70;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $77 = HEAP32[$src$1>>2]|0;
     $78 = $77 >>> 24;
     $79 = $77 >>> 16;
     $80 = $79 & 255;
     $81 = $77 >>> 8;
     $82 = $81 & 255;
     $83 = Math_imul($78, $4)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = Math_imul($80, $6)|0;
     $86 = (($85>>>0) / 255)&-1;
     $87 = Math_imul($82, $8)|0;
     $88 = (($87>>>0) / 255)&-1;
     $89 = $88 << 16;
     $90 = $86 << 8;
     $91 = $89 | $90;
     $92 = $91 | $84;
     HEAP32[$dst$013>>2] = $92;
     $93 = (($posx$2) + ($20))|0;
     $94 = ((($dst$013)) + 4|0);
     $95 = (($96) + -1)|0;
     $97 = ($96|0)==(0);
     if ($97) {
      break;
     } else {
      $96 = $95;$dst$013 = $94;$posx$011 = $93;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $98 = HEAP32[$24>>2]|0;
  $99 = (($100) + ($98)|0);
  $101 = HEAP32[$12>>2]|0;
  $102 = (($101) + -1)|0;
  HEAP32[$12>>2] = $102;
  $103 = ($101|0)==(0);
  if ($103) {
   $$lcssa = $99;
   break;
  }
  $104 = (($posy$1$lcssa) + ($14))|0;
  $$pre27 = HEAP32[$18>>2]|0;
  $100 = $99;$105 = $99;$36 = $$pre27;$posy$017 = $104;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$23>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGBA8888_BGR888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$lcssa14 = 0, $$pre = 0, $$pre10 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $dst$05 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $src$04 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + -1)|0;
 HEAP32[$10>>2] = $12;
 $13 = ($11|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre9 = HEAP32[$14>>2]|0;
 $87 = $$pre;$90 = $$pre9;$94 = $12;$95 = $$pre9;$96 = $$pre;
 while(1) {
  $25 = HEAP32[$15>>2]|0;
  $26 = ($25|0)==(0);
  if ($26) {
   $92 = $94;
  } else {
   $$in = $25;$dst$05 = $95;$src$04 = $96;
   while(1) {
    $27 = (($$in) + -1)|0;
    $28 = HEAP32[$src$04>>2]|0;
    $29 = $28 >>> 24;
    $30 = $28 >>> 16;
    $31 = $30 & 255;
    $32 = $28 >>> 8;
    $33 = $32 & 255;
    $34 = $28 & 255;
    $35 = HEAP32[$dst$05>>2]|0;
    $36 = $35 >>> 16;
    $37 = $36 & 255;
    $38 = $35 >>> 8;
    $39 = $38 & 255;
    $40 = $35 & 255;
    if ($19) {
     $srcB$0 = $33;$srcG$0 = $31;$srcR$0 = $29;
    } else {
     $41 = Math_imul($29, $4)|0;
     $42 = (($41>>>0) / 255)&-1;
     $43 = Math_imul($31, $6)|0;
     $44 = (($43>>>0) / 255)&-1;
     $45 = Math_imul($33, $8)|0;
     $46 = (($45>>>0) / 255)&-1;
     $srcB$0 = $46;$srcG$0 = $44;$srcR$0 = $42;
    }
    if ($21) {
     $srcA$0 = $34;
    } else {
     $47 = Math_imul($34, $9)|0;
     $48 = (($47>>>0) / 255)&-1;
     $srcA$0 = $48;
    }
    $49 = ($srcA$0>>>0)<(255);
    $or$cond = $23 & $49;
    if ($or$cond) {
     $50 = Math_imul($srcA$0, $srcR$0)|0;
     $51 = (($50>>>0) / 255)&-1;
     $52 = Math_imul($srcA$0, $srcG$0)|0;
     $53 = (($52>>>0) / 255)&-1;
     $54 = Math_imul($srcA$0, $srcB$0)|0;
     $55 = (($54>>>0) / 255)&-1;
     $srcB$1 = $55;$srcG$1 = $53;$srcR$1 = $51;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($24|0) {
    case 16:  {
     $56 = (255 - ($srcA$0))|0;
     $57 = Math_imul($56, $40)|0;
     $58 = (($57>>>0) / 255)&-1;
     $59 = (($srcR$1) + ($58))|0;
     $60 = Math_imul($56, $39)|0;
     $61 = (($60>>>0) / 255)&-1;
     $62 = (($srcG$1) + ($61))|0;
     $63 = Math_imul($56, $37)|0;
     $64 = (($63>>>0) / 255)&-1;
     $65 = (($srcB$1) + ($64))|0;
     $dstB$0 = $65;$dstG$0 = $62;$dstR$0 = $59;
     break;
    }
    case 32:  {
     $66 = (($srcR$1) + ($40))|0;
     $67 = ($66>>>0)>(255);
     $$ = $67 ? 255 : $66;
     $68 = (($srcG$1) + ($39))|0;
     $69 = ($68>>>0)>(255);
     $$3 = $69 ? 255 : $68;
     $70 = (($srcB$1) + ($37))|0;
     $71 = ($70>>>0)>(255);
     $$1 = $71 ? 255 : $70;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $72 = Math_imul($srcR$1, $40)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($srcG$1, $39)|0;
     $75 = (($74>>>0) / 255)&-1;
     $76 = Math_imul($srcB$1, $37)|0;
     $77 = (($76>>>0) / 255)&-1;
     $dstB$0 = $77;$dstG$0 = $75;$dstR$0 = $73;
     break;
    }
    default: {
     $dstB$0 = $37;$dstG$0 = $39;$dstR$0 = $40;
    }
    }
    $78 = $dstB$0 << 16;
    $79 = $dstG$0 << 8;
    $80 = $79 | $dstR$0;
    $81 = $80 | $78;
    HEAP32[$dst$05>>2] = $81;
    $82 = ((($src$04)) + 4|0);
    $83 = ((($dst$05)) + 4|0);
    $84 = ($27|0)==(0);
    if ($84) {
     break;
    } else {
     $$in = $27;$dst$05 = $83;$src$04 = $82;
    }
   }
   $$pre10 = HEAP32[$10>>2]|0;
   $92 = $$pre10;
  }
  $85 = HEAP32[$16>>2]|0;
  $86 = (($87) + ($85)|0);
  $88 = HEAP32[$17>>2]|0;
  $89 = (($90) + ($88)|0);
  $91 = (($92) + -1)|0;
  HEAP32[$10>>2] = $91;
  $93 = ($92|0)==(0);
  if ($93) {
   $$lcssa = $86;$$lcssa14 = $89;
   break;
  } else {
   $87 = $86;$90 = $89;$94 = $91;$95 = $89;$96 = $86;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa14;
 return;
}
function _SDL_Blit_RGBA8888_BGR888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0;
 var $src$015 = 0, $src$1 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 << 16;
 $13 = ((($info)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12|0) / ($14|0))&-1;
 $16 = ((($info)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 << 16;
 $19 = ((($info)) + 24|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($18|0) / ($20|0))&-1;
 $22 = (($14) + -1)|0;
 HEAP32[$13>>2] = $22;
 $23 = ($14|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = $1 & 112;
 $33 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $116 = $$pre;$121 = $22;$122 = $$pre;$42 = $20;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $34 = ($posy$020|0)>(65535);
  if ($34) {
   $35 = (($posy$020) + -65536)|0;
   $36 = $35 >>> 16;
   $37 = $36 << 16;
   $38 = (($srcy$019) + 1)|0;
   $39 = (($35) - ($37))|0;
   $40 = (($38) + ($36))|0;
   $posy$1$lcssa = $39;$srcy$1$lcssa = $40;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $41 = ($42|0)==(0);
  if ($41) {
   $118 = $121;
  } else {
   $$in = $42;$dst$016 = $122;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $43 = (($$in) + -1)|0;
    $44 = ($posx$014|0)>(65535);
    if ($44) {
     $45 = (($posx$014) + -65536)|0;
     $46 = $45 >>> 16;
     $47 = $46 << 16;
     $48 = (($srcx$013) + 1)|0;
     $49 = (($45) - ($47))|0;
     $50 = (($48) + ($46))|0;
     $51 = HEAP32[$info>>2]|0;
     $52 = HEAP32[$33>>2]|0;
     $53 = Math_imul($52, $srcy$1$lcssa)|0;
     $54 = (($51) + ($53)|0);
     $55 = $50 << 2;
     $56 = (($54) + ($55)|0);
     $posx$2 = $49;$src$1 = $56;$srcx$2 = $50;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $57 = HEAP32[$src$1>>2]|0;
    $58 = $57 >>> 24;
    $59 = $57 >>> 16;
    $60 = $59 & 255;
    $61 = $57 >>> 8;
    $62 = $61 & 255;
    $63 = $57 & 255;
    $64 = HEAP32[$dst$016>>2]|0;
    $65 = $64 >>> 16;
    $66 = $65 & 255;
    $67 = $64 >>> 8;
    $68 = $67 & 255;
    $69 = $64 & 255;
    if ($27) {
     $srcB$0 = $62;$srcG$0 = $60;$srcR$0 = $58;
    } else {
     $70 = Math_imul($58, $4)|0;
     $71 = (($70>>>0) / 255)&-1;
     $72 = Math_imul($60, $6)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($62, $8)|0;
     $75 = (($74>>>0) / 255)&-1;
     $srcB$0 = $75;$srcG$0 = $73;$srcR$0 = $71;
    }
    if ($29) {
     $srcA$0 = $63;
    } else {
     $76 = Math_imul($63, $9)|0;
     $77 = (($76>>>0) / 255)&-1;
     $srcA$0 = $77;
    }
    $78 = ($srcA$0>>>0)<(255);
    $or$cond = $31 & $78;
    if ($or$cond) {
     $79 = Math_imul($srcA$0, $srcR$0)|0;
     $80 = (($79>>>0) / 255)&-1;
     $81 = Math_imul($srcA$0, $srcG$0)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = Math_imul($srcA$0, $srcB$0)|0;
     $84 = (($83>>>0) / 255)&-1;
     $srcB$1 = $84;$srcG$1 = $82;$srcR$1 = $80;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($32|0) {
    case 16:  {
     $85 = (255 - ($srcA$0))|0;
     $86 = Math_imul($85, $69)|0;
     $87 = (($86>>>0) / 255)&-1;
     $88 = (($srcR$1) + ($87))|0;
     $89 = Math_imul($85, $68)|0;
     $90 = (($89>>>0) / 255)&-1;
     $91 = (($srcG$1) + ($90))|0;
     $92 = Math_imul($85, $66)|0;
     $93 = (($92>>>0) / 255)&-1;
     $94 = (($srcB$1) + ($93))|0;
     $dstB$0 = $94;$dstG$0 = $91;$dstR$0 = $88;
     break;
    }
    case 32:  {
     $95 = (($srcR$1) + ($69))|0;
     $96 = ($95>>>0)>(255);
     $$ = $96 ? 255 : $95;
     $97 = (($srcG$1) + ($68))|0;
     $98 = ($97>>>0)>(255);
     $$3 = $98 ? 255 : $97;
     $99 = (($srcB$1) + ($66))|0;
     $100 = ($99>>>0)>(255);
     $$1 = $100 ? 255 : $99;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $101 = Math_imul($srcR$1, $69)|0;
     $102 = (($101>>>0) / 255)&-1;
     $103 = Math_imul($srcG$1, $68)|0;
     $104 = (($103>>>0) / 255)&-1;
     $105 = Math_imul($srcB$1, $66)|0;
     $106 = (($105>>>0) / 255)&-1;
     $dstB$0 = $106;$dstG$0 = $104;$dstR$0 = $102;
     break;
    }
    default: {
     $dstB$0 = $66;$dstG$0 = $68;$dstR$0 = $69;
    }
    }
    $107 = $dstB$0 << 16;
    $108 = $dstG$0 << 8;
    $109 = $108 | $dstR$0;
    $110 = $109 | $107;
    HEAP32[$dst$016>>2] = $110;
    $111 = (($posx$2) + ($21))|0;
    $112 = ((($dst$016)) + 4|0);
    $113 = ($43|0)==(0);
    if ($113) {
     break;
    } else {
     $$in = $43;$dst$016 = $112;$posx$014 = $111;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$13>>2]|0;
   $118 = $$pre28;
  }
  $114 = HEAP32[$25>>2]|0;
  $115 = (($116) + ($114)|0);
  $117 = (($118) + -1)|0;
  HEAP32[$13>>2] = $117;
  $119 = ($118|0)==(0);
  if ($119) {
   $$lcssa = $115;
   break;
  }
  $120 = (($posy$1$lcssa) + ($15))|0;
  $$pre27 = HEAP32[$19>>2]|0;
  $116 = $115;$121 = $117;$122 = $115;$42 = $$pre27;$posy$020 = $120;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGBA8888_ARGB8888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0, $posx$010 = 0, $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$55 = $$pre;$60 = $12;$61 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $57 = $60;
  } else {
   $$in = $25;$dst$013 = $61;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    $41 = $40 >>> 24;
    $42 = $40 >>> 8;
    $43 = $42 & 255;
    $44 = $40 << 24;
    $45 = $41 << 16;
    $46 = $45 | $44;
    $47 = $42 & 65280;
    $48 = $46 | $47;
    $49 = $48 | $43;
    HEAP32[$dst$013>>2] = $49;
    $50 = (($posx$2) + ($11))|0;
    $51 = ((($dst$013)) + 4|0);
    $52 = ($26|0)==(0);
    if ($52) {
     break;
    } else {
     $$in = $26;$dst$013 = $51;$posx$010 = $50;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $57 = $$pre25;
  }
  $53 = HEAP32[$15>>2]|0;
  $54 = (($55) + ($53)|0);
  $56 = (($57) + -1)|0;
  HEAP32[$3>>2] = $56;
  $58 = ($57|0)==(0);
  if ($58) {
   $$lcssa = $54;
   break;
  }
  $59 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$55 = $54;$60 = $56;$61 = $54;$posy$016 = $59;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGBA8888_ARGB8888_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$$us = 0, $$1 = 0, $$1$us = 0, $$3 = 0, $$3$us = 0, $$lcssa = 0, $$lcssa17 = 0, $$pre = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$05 = 0, $dst$05$us = 0, $dstA$0 = 0, $dstA$0$us = 0, $dstB$0 = 0, $dstB$0$us = 0, $dstG$0 = 0, $dstG$0$us = 0, $dstR$0 = 0, $dstR$0$us = 0, $src$04 = 0, $src$04$us = 0, $srcB$0$us = 0, $srcG$0$us = 0, $srcR$0$us = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 48;
 $11 = ($10|0)==(0);
 $12 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre10 = HEAP32[$6>>2]|0;
 $125 = $$pre;$128 = $$pre10;$132 = $$pre10;$133 = $$pre;
 while(1) {
  $13 = HEAP32[$7>>2]|0;
  $14 = (($13) + -1)|0;
  $15 = ($13|0)==(0);
  L6: do {
   if (!($15)) {
    if ($11) {
     $121 = $14;$dst$05 = $132;$src$04 = $133;
     while(1) {
      $73 = HEAP32[$src$04>>2]|0;
      $74 = $73 >>> 24;
      $75 = $73 >>> 16;
      $76 = $75 & 255;
      $77 = $73 >>> 8;
      $78 = $77 & 255;
      $79 = $73 & 255;
      $80 = HEAP32[$dst$05>>2]|0;
      $81 = $80 >>> 16;
      $82 = $81 & 255;
      $83 = $80 >>> 8;
      $84 = $83 & 255;
      $85 = $80 & 255;
      $86 = $80 >>> 24;
      switch ($12|0) {
      case 16:  {
       $87 = $79 ^ 255;
       $88 = Math_imul($82, $87)|0;
       $89 = (($88>>>0) / 255)&-1;
       $90 = (($74) + ($89))|0;
       $91 = Math_imul($84, $87)|0;
       $92 = (($91>>>0) / 255)&-1;
       $93 = (($76) + ($92))|0;
       $94 = Math_imul($87, $85)|0;
       $95 = (($94>>>0) / 255)&-1;
       $96 = (($78) + ($95))|0;
       $97 = Math_imul($87, $86)|0;
       $98 = (($97>>>0) / 255)&-1;
       $99 = (($98) + ($79))|0;
       $dstA$0 = $99;$dstB$0 = $96;$dstG$0 = $93;$dstR$0 = $90;
       break;
      }
      case 32:  {
       $100 = (($74) + ($82))|0;
       $101 = ($100>>>0)>(255);
       $$ = $101 ? 255 : $100;
       $102 = (($76) + ($84))|0;
       $103 = ($102>>>0)>(255);
       $$3 = $103 ? 255 : $102;
       $104 = (($78) + ($85))|0;
       $105 = ($104>>>0)>(255);
       $$1 = $105 ? 255 : $104;
       $dstA$0 = $86;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
       break;
      }
      case 64:  {
       $106 = Math_imul($74, $82)|0;
       $107 = (($106>>>0) / 255)&-1;
       $108 = Math_imul($76, $84)|0;
       $109 = (($108>>>0) / 255)&-1;
       $110 = Math_imul($78, $85)|0;
       $111 = (($110>>>0) / 255)&-1;
       $dstA$0 = $86;$dstB$0 = $111;$dstG$0 = $109;$dstR$0 = $107;
       break;
      }
      default: {
       $dstA$0 = $86;$dstB$0 = $85;$dstG$0 = $84;$dstR$0 = $82;
      }
      }
      $112 = $dstA$0 << 24;
      $113 = $dstR$0 << 16;
      $114 = $dstG$0 << 8;
      $115 = $114 | $113;
      $116 = $115 | $dstB$0;
      $117 = $116 | $112;
      HEAP32[$dst$05>>2] = $117;
      $118 = ((($src$04)) + 4|0);
      $119 = ((($dst$05)) + 4|0);
      $120 = (($121) + -1)|0;
      $122 = ($121|0)==(0);
      if ($122) {
       break L6;
      } else {
       $121 = $120;$dst$05 = $119;$src$04 = $118;
      }
     }
    } else {
     $71 = $14;$dst$05$us = $132;$src$04$us = $133;
    }
    while(1) {
     $16 = HEAP32[$src$04$us>>2]|0;
     $17 = $16 >>> 24;
     $18 = $16 >>> 16;
     $19 = $18 & 255;
     $20 = $16 >>> 8;
     $21 = $20 & 255;
     $22 = $16 & 255;
     $23 = HEAP32[$dst$05$us>>2]|0;
     $24 = $23 >>> 16;
     $25 = $24 & 255;
     $26 = $23 >>> 8;
     $27 = $26 & 255;
     $28 = $23 & 255;
     $29 = $23 >>> 24;
     $30 = ($22|0)==(255);
     if ($30) {
      $srcB$0$us = $21;$srcG$0$us = $19;$srcR$0$us = $17;
     } else {
      $31 = Math_imul($17, $22)|0;
      $32 = (($31>>>0) / 255)&-1;
      $33 = Math_imul($19, $22)|0;
      $34 = (($33>>>0) / 255)&-1;
      $35 = Math_imul($21, $22)|0;
      $36 = (($35>>>0) / 255)&-1;
      $srcB$0$us = $36;$srcG$0$us = $34;$srcR$0$us = $32;
     }
     switch ($12|0) {
     case 16:  {
      $49 = $22 ^ 255;
      $50 = Math_imul($25, $49)|0;
      $51 = (($50>>>0) / 255)&-1;
      $52 = (($srcR$0$us) + ($51))|0;
      $53 = Math_imul($27, $49)|0;
      $54 = (($53>>>0) / 255)&-1;
      $55 = (($srcG$0$us) + ($54))|0;
      $56 = Math_imul($49, $28)|0;
      $57 = (($56>>>0) / 255)&-1;
      $58 = (($srcB$0$us) + ($57))|0;
      $59 = Math_imul($49, $29)|0;
      $60 = (($59>>>0) / 255)&-1;
      $61 = (($60) + ($22))|0;
      $dstA$0$us = $61;$dstB$0$us = $58;$dstG$0$us = $55;$dstR$0$us = $52;
      break;
     }
     case 32:  {
      $43 = (($srcR$0$us) + ($25))|0;
      $44 = ($43>>>0)>(255);
      $$$us = $44 ? 255 : $43;
      $45 = (($srcG$0$us) + ($27))|0;
      $46 = ($45>>>0)>(255);
      $$3$us = $46 ? 255 : $45;
      $47 = (($srcB$0$us) + ($28))|0;
      $48 = ($47>>>0)>(255);
      $$1$us = $48 ? 255 : $47;
      $dstA$0$us = $29;$dstB$0$us = $$1$us;$dstG$0$us = $$3$us;$dstR$0$us = $$$us;
      break;
     }
     case 64:  {
      $37 = Math_imul($srcR$0$us, $25)|0;
      $38 = (($37>>>0) / 255)&-1;
      $39 = Math_imul($srcG$0$us, $27)|0;
      $40 = (($39>>>0) / 255)&-1;
      $41 = Math_imul($srcB$0$us, $28)|0;
      $42 = (($41>>>0) / 255)&-1;
      $dstA$0$us = $29;$dstB$0$us = $42;$dstG$0$us = $40;$dstR$0$us = $38;
      break;
     }
     default: {
      $dstA$0$us = $29;$dstB$0$us = $28;$dstG$0$us = $27;$dstR$0$us = $25;
     }
     }
     $62 = $dstA$0$us << 24;
     $63 = $dstR$0$us << 16;
     $64 = $dstG$0$us << 8;
     $65 = $64 | $63;
     $66 = $65 | $dstB$0$us;
     $67 = $66 | $62;
     HEAP32[$dst$05$us>>2] = $67;
     $68 = ((($src$04$us)) + 4|0);
     $69 = ((($dst$05$us)) + 4|0);
     $70 = (($71) + -1)|0;
     $72 = ($71|0)==(0);
     if ($72) {
      break L6;
     } else {
      $71 = $70;$dst$05$us = $69;$src$04$us = $68;
     }
    }
   }
  } while(0);
  $123 = HEAP32[$8>>2]|0;
  $124 = (($125) + ($123)|0);
  $126 = HEAP32[$9>>2]|0;
  $127 = (($128) + ($126)|0);
  $129 = HEAP32[$2>>2]|0;
  $130 = (($129) + -1)|0;
  HEAP32[$2>>2] = $130;
  $131 = ($129|0)==(0);
  if ($131) {
   $$lcssa = $124;$$lcssa17 = $127;
   break;
  } else {
   $125 = $124;$128 = $127;$132 = $127;$133 = $124;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa17;
 return;
}
function _SDL_Blit_RGBA8888_ARGB8888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstA$0 = 0, $dstB$0 = 0;
 var $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0, $src$015 = 0, $src$1 = 0, $srcB$0 = 0, $srcG$0 = 0, $srcR$0 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 48;
 $19 = ($18|0)!=(0);
 $20 = $1 & 112;
 $21 = ((($info)) + 12|0);
 $$pre = HEAP32[$16>>2]|0;
 $102 = $$pre;$107 = $14;$108 = $$pre;$30 = $12;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $22 = ($posy$020|0)>(65535);
  if ($22) {
   $23 = (($posy$020) + -65536)|0;
   $24 = $23 >>> 16;
   $25 = $24 << 16;
   $26 = (($srcy$019) + 1)|0;
   $27 = (($23) - ($25))|0;
   $28 = (($26) + ($24))|0;
   $posy$1$lcssa = $27;$srcy$1$lcssa = $28;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $29 = ($30|0)==(0);
  if ($29) {
   $104 = $107;
  } else {
   $$in = $30;$dst$016 = $108;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $31 = (($$in) + -1)|0;
    $32 = ($posx$014|0)>(65535);
    if ($32) {
     $33 = (($posx$014) + -65536)|0;
     $34 = $33 >>> 16;
     $35 = $34 << 16;
     $36 = (($srcx$013) + 1)|0;
     $37 = (($33) - ($35))|0;
     $38 = (($36) + ($34))|0;
     $39 = HEAP32[$info>>2]|0;
     $40 = HEAP32[$21>>2]|0;
     $41 = Math_imul($40, $srcy$1$lcssa)|0;
     $42 = (($39) + ($41)|0);
     $43 = $38 << 2;
     $44 = (($42) + ($43)|0);
     $posx$2 = $37;$src$1 = $44;$srcx$2 = $38;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $45 = HEAP32[$src$1>>2]|0;
    $46 = $45 >>> 24;
    $47 = $45 >>> 16;
    $48 = $47 & 255;
    $49 = $45 >>> 8;
    $50 = $49 & 255;
    $51 = $45 & 255;
    $52 = HEAP32[$dst$016>>2]|0;
    $53 = $52 >>> 16;
    $54 = $53 & 255;
    $55 = $52 >>> 8;
    $56 = $55 & 255;
    $57 = $52 & 255;
    $58 = $52 >>> 24;
    $59 = ($51|0)!=(255);
    $or$cond = $19 & $59;
    if ($or$cond) {
     $60 = Math_imul($46, $51)|0;
     $61 = (($60>>>0) / 255)&-1;
     $62 = Math_imul($48, $51)|0;
     $63 = (($62>>>0) / 255)&-1;
     $64 = Math_imul($50, $51)|0;
     $65 = (($64>>>0) / 255)&-1;
     $srcB$0 = $65;$srcG$0 = $63;$srcR$0 = $61;
    } else {
     $srcB$0 = $50;$srcG$0 = $48;$srcR$0 = $46;
    }
    switch ($20|0) {
    case 16:  {
     $66 = $51 ^ 255;
     $67 = Math_imul($54, $66)|0;
     $68 = (($67>>>0) / 255)&-1;
     $69 = (($srcR$0) + ($68))|0;
     $70 = Math_imul($56, $66)|0;
     $71 = (($70>>>0) / 255)&-1;
     $72 = (($srcG$0) + ($71))|0;
     $73 = Math_imul($66, $57)|0;
     $74 = (($73>>>0) / 255)&-1;
     $75 = (($srcB$0) + ($74))|0;
     $76 = Math_imul($66, $58)|0;
     $77 = (($76>>>0) / 255)&-1;
     $78 = (($77) + ($51))|0;
     $dstA$0 = $78;$dstB$0 = $75;$dstG$0 = $72;$dstR$0 = $69;
     break;
    }
    case 32:  {
     $79 = (($srcR$0) + ($54))|0;
     $80 = ($79>>>0)>(255);
     $$ = $80 ? 255 : $79;
     $81 = (($srcG$0) + ($56))|0;
     $82 = ($81>>>0)>(255);
     $$3 = $82 ? 255 : $81;
     $83 = (($srcB$0) + ($57))|0;
     $84 = ($83>>>0)>(255);
     $$1 = $84 ? 255 : $83;
     $dstA$0 = $58;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $85 = Math_imul($srcR$0, $54)|0;
     $86 = (($85>>>0) / 255)&-1;
     $87 = Math_imul($srcG$0, $56)|0;
     $88 = (($87>>>0) / 255)&-1;
     $89 = Math_imul($srcB$0, $57)|0;
     $90 = (($89>>>0) / 255)&-1;
     $dstA$0 = $58;$dstB$0 = $90;$dstG$0 = $88;$dstR$0 = $86;
     break;
    }
    default: {
     $dstA$0 = $58;$dstB$0 = $57;$dstG$0 = $56;$dstR$0 = $54;
    }
    }
    $91 = $dstA$0 << 24;
    $92 = $dstR$0 << 16;
    $93 = $dstG$0 << 8;
    $94 = $93 | $92;
    $95 = $94 | $dstB$0;
    $96 = $95 | $91;
    HEAP32[$dst$016>>2] = $96;
    $97 = (($posx$2) + ($13))|0;
    $98 = ((($dst$016)) + 4|0);
    $99 = ($31|0)==(0);
    if ($99) {
     break;
    } else {
     $$in = $31;$dst$016 = $98;$posx$014 = $97;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$5>>2]|0;
   $104 = $$pre28;
  }
  $100 = HEAP32[$17>>2]|0;
  $101 = (($102) + ($100)|0);
  $103 = (($104) + -1)|0;
  HEAP32[$5>>2] = $103;
  $105 = ($104|0)==(0);
  if ($105) {
   $$lcssa = $101;
   break;
  }
  $106 = (($posy$1$lcssa) + ($7))|0;
  $$pre27 = HEAP32[$11>>2]|0;
  $102 = $101;$107 = $103;$108 = $101;$30 = $$pre27;$posy$020 = $106;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGBA8888_ARGB8888_Modulate($info) {
 $info = $info|0;
 var $$lcssa = 0, $$lcssa24 = 0, $$pre = 0, $$pre14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $A$0$us = 0, $dst$02 = 0, $dst$02$us = 0, $dst$02$us6 = 0, $src$01 = 0, $src$01$us = 0, $src$01$us7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + -1)|0;
 HEAP32[$10>>2] = $12;
 $13 = ($11|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $$pre = HEAP32[$info>>2]|0;
 $$pre14 = HEAP32[$14>>2]|0;
 $101 = $$pre14;$102 = $$pre;$94 = $$pre;$97 = $$pre14;
 while(1) {
  $22 = HEAP32[$15>>2]|0;
  $23 = (($22) + -1)|0;
  $24 = ($22|0)==(0);
  if (!($24)) {
   if ($19) {
    $41 = $23;$dst$02$us = $101;$src$01$us = $102;
    while(1) {
     $25 = HEAP32[$src$01$us>>2]|0;
     $26 = $25 >>> 24;
     $27 = $25 >>> 8;
     $28 = $27 & 255;
     $29 = $25 & 255;
     if ($21) {
      $A$0$us = $29;
     } else {
      $30 = Math_imul($29, $9)|0;
      $31 = (($30>>>0) / 255)&-1;
      $A$0$us = $31;
     }
     $32 = $A$0$us << 24;
     $33 = $26 << 16;
     $34 = $27 & 65280;
     $35 = $34 | $33;
     $36 = $35 | $28;
     $37 = $36 | $32;
     HEAP32[$dst$02$us>>2] = $37;
     $38 = ((($src$01$us)) + 4|0);
     $39 = ((($dst$02$us)) + 4|0);
     $40 = (($41) + -1)|0;
     $42 = ($41|0)==(0);
     if ($42) {
      break;
     } else {
      $41 = $40;$dst$02$us = $39;$src$01$us = $38;
     }
    }
   } else {
    if ($21) {
     $64 = $23;$dst$02$us6 = $101;$src$01$us7 = $102;
     while(1) {
      $43 = HEAP32[$src$01$us7>>2]|0;
      $44 = $43 >>> 24;
      $45 = $43 >>> 16;
      $46 = $45 & 255;
      $47 = $43 >>> 8;
      $48 = $47 & 255;
      $49 = Math_imul($44, $4)|0;
      $50 = (($49>>>0) / 255)&-1;
      $51 = Math_imul($46, $6)|0;
      $52 = (($51>>>0) / 255)&-1;
      $53 = Math_imul($48, $8)|0;
      $54 = (($53>>>0) / 255)&-1;
      $55 = $43 << 24;
      $56 = $50 << 16;
      $57 = $52 << 8;
      $58 = $57 | $56;
      $59 = $58 | $54;
      $60 = $59 | $55;
      HEAP32[$dst$02$us6>>2] = $60;
      $61 = ((($src$01$us7)) + 4|0);
      $62 = ((($dst$02$us6)) + 4|0);
      $63 = (($64) + -1)|0;
      $65 = ($64|0)==(0);
      if ($65) {
       break;
      } else {
       $64 = $63;$dst$02$us6 = $62;$src$01$us7 = $61;
      }
     }
    } else {
     $90 = $23;$dst$02 = $101;$src$01 = $102;
     while(1) {
      $66 = HEAP32[$src$01>>2]|0;
      $67 = $66 >>> 24;
      $68 = $66 >>> 16;
      $69 = $68 & 255;
      $70 = $66 >>> 8;
      $71 = $70 & 255;
      $72 = $66 & 255;
      $73 = Math_imul($67, $4)|0;
      $74 = (($73>>>0) / 255)&-1;
      $75 = Math_imul($69, $6)|0;
      $76 = (($75>>>0) / 255)&-1;
      $77 = Math_imul($71, $8)|0;
      $78 = (($77>>>0) / 255)&-1;
      $79 = Math_imul($72, $9)|0;
      $80 = (($79>>>0) / 255)&-1;
      $81 = $80 << 24;
      $82 = $74 << 16;
      $83 = $76 << 8;
      $84 = $83 | $82;
      $85 = $84 | $78;
      $86 = $85 | $81;
      HEAP32[$dst$02>>2] = $86;
      $87 = ((($src$01)) + 4|0);
      $88 = ((($dst$02)) + 4|0);
      $89 = (($90) + -1)|0;
      $91 = ($90|0)==(0);
      if ($91) {
       break;
      } else {
       $90 = $89;$dst$02 = $88;$src$01 = $87;
      }
     }
    }
   }
  }
  $92 = HEAP32[$16>>2]|0;
  $93 = (($94) + ($92)|0);
  $95 = HEAP32[$17>>2]|0;
  $96 = (($97) + ($95)|0);
  $98 = HEAP32[$10>>2]|0;
  $99 = (($98) + -1)|0;
  HEAP32[$10>>2] = $99;
  $100 = ($98|0)==(0);
  if ($100) {
   $$lcssa = $93;$$lcssa24 = $96;
   break;
  } else {
   $101 = $96;$102 = $93;$94 = $93;$97 = $96;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa24;
 return;
}
function _SDL_Blit_RGBA8888_ARGB8888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $A$0 = 0, $A$0$us = 0, $dst$013 = 0, $dst$013$us = 0, $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$017 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0, $src$1$us = 0, $srcx$010 = 0, $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$016 = 0, $srcy$1$lcssa = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 << 16;
 $13 = ((($info)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12|0) / ($14|0))&-1;
 $16 = ((($info)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 << 16;
 $19 = ((($info)) + 24|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($18|0) / ($20|0))&-1;
 $22 = (($14) + -1)|0;
 HEAP32[$13>>2] = $22;
 $23 = ($14|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $111 = $$pre;$116 = $$pre;$39 = $20;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $31 = ($posy$017|0)>(65535);
  if ($31) {
   $32 = (($posy$017) + -65536)|0;
   $33 = $32 >>> 16;
   $34 = $33 << 16;
   $35 = (($srcy$016) + 1)|0;
   $36 = (($32) - ($34))|0;
   $37 = (($35) + ($33))|0;
   $posy$1$lcssa = $36;$srcy$1$lcssa = $37;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $38 = (($39) + -1)|0;
  $40 = ($39|0)==(0);
  if (!($40)) {
   if ($27) {
    $68 = $38;$dst$013$us = $116;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $41 = ($posx$011$us|0)>(65535);
     if ($41) {
      $42 = (($posx$011$us) + -65536)|0;
      $43 = $42 >>> 16;
      $44 = $43 << 16;
      $45 = (($srcx$010$us) + 1)|0;
      $46 = (($42) - ($44))|0;
      $47 = (($45) + ($43))|0;
      $48 = HEAP32[$info>>2]|0;
      $49 = HEAP32[$30>>2]|0;
      $50 = Math_imul($49, $srcy$1$lcssa)|0;
      $51 = (($48) + ($50)|0);
      $52 = $47 << 2;
      $53 = (($51) + ($52)|0);
      $posx$2$us = $46;$src$1$us = $53;$srcx$2$us = $47;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $54 = HEAP32[$src$1$us>>2]|0;
     $55 = $54 >>> 24;
     $56 = $54 >>> 8;
     $57 = $54 & 255;
     if ($29) {
      $A$0$us = $57;
     } else {
      $58 = Math_imul($57, $9)|0;
      $59 = (($58>>>0) / 255)&-1;
      $A$0$us = $59;
     }
     $60 = $A$0$us << 24;
     $61 = $55 << 16;
     $62 = $56 & 65535;
     $63 = $62 | $61;
     $64 = $63 | $60;
     HEAP32[$dst$013$us>>2] = $64;
     $65 = (($posx$2$us) + ($21))|0;
     $66 = ((($dst$013$us)) + 4|0);
     $67 = (($68) + -1)|0;
     $69 = ($68|0)==(0);
     if ($69) {
      break;
     } else {
      $68 = $67;$dst$013$us = $66;$posx$011$us = $65;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $107 = $38;$dst$013 = $116;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $70 = ($posx$011|0)>(65535);
     if ($70) {
      $71 = (($posx$011) + -65536)|0;
      $72 = $71 >>> 16;
      $73 = $72 << 16;
      $74 = (($srcx$010) + 1)|0;
      $75 = (($71) - ($73))|0;
      $76 = (($74) + ($72))|0;
      $77 = HEAP32[$info>>2]|0;
      $78 = HEAP32[$30>>2]|0;
      $79 = Math_imul($78, $srcy$1$lcssa)|0;
      $80 = (($77) + ($79)|0);
      $81 = $76 << 2;
      $82 = (($80) + ($81)|0);
      $posx$2 = $75;$src$1 = $82;$srcx$2 = $76;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $83 = HEAP32[$src$1>>2]|0;
     $84 = $83 >>> 24;
     $85 = $83 >>> 16;
     $86 = $85 & 255;
     $87 = $83 >>> 8;
     $88 = $87 & 255;
     $89 = $83 & 255;
     $90 = Math_imul($84, $4)|0;
     $91 = (($90>>>0) / 255)&-1;
     $92 = Math_imul($86, $6)|0;
     $93 = (($92>>>0) / 255)&-1;
     $94 = Math_imul($88, $8)|0;
     $95 = (($94>>>0) / 255)&-1;
     if ($29) {
      $A$0 = $89;
     } else {
      $96 = Math_imul($89, $9)|0;
      $97 = (($96>>>0) / 255)&-1;
      $A$0 = $97;
     }
     $98 = $A$0 << 24;
     $99 = $91 << 16;
     $100 = $93 << 8;
     $101 = $95 | $100;
     $102 = $101 | $99;
     $103 = $102 | $98;
     HEAP32[$dst$013>>2] = $103;
     $104 = (($posx$2) + ($21))|0;
     $105 = ((($dst$013)) + 4|0);
     $106 = (($107) + -1)|0;
     $108 = ($107|0)==(0);
     if ($108) {
      break;
     } else {
      $107 = $106;$dst$013 = $105;$posx$011 = $104;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $109 = HEAP32[$25>>2]|0;
  $110 = (($111) + ($109)|0);
  $112 = HEAP32[$13>>2]|0;
  $113 = (($112) + -1)|0;
  HEAP32[$13>>2] = $113;
  $114 = ($112|0)==(0);
  if ($114) {
   $$lcssa = $110;
   break;
  }
  $115 = (($posy$1$lcssa) + ($15))|0;
  $$pre27 = HEAP32[$19>>2]|0;
  $111 = $110;$116 = $110;$39 = $$pre27;$posy$017 = $115;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_RGBA8888_ARGB8888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$lcssa14 = 0, $$pre = 0, $$pre10 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$05 = 0, $dstA$0 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $src$04 = 0, $srcA$0 = 0;
 var $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + -1)|0;
 HEAP32[$10>>2] = $12;
 $13 = ($11|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre9 = HEAP32[$14>>2]|0;
 $100 = $12;$101 = $$pre9;$102 = $$pre;$93 = $$pre;$96 = $$pre9;
 while(1) {
  $25 = HEAP32[$15>>2]|0;
  $26 = ($25|0)==(0);
  if ($26) {
   $98 = $100;
  } else {
   $$in = $25;$dst$05 = $101;$src$04 = $102;
   while(1) {
    $27 = (($$in) + -1)|0;
    $28 = HEAP32[$src$04>>2]|0;
    $29 = $28 >>> 24;
    $30 = $28 >>> 16;
    $31 = $30 & 255;
    $32 = $28 >>> 8;
    $33 = $32 & 255;
    $34 = $28 & 255;
    $35 = HEAP32[$dst$05>>2]|0;
    $36 = $35 >>> 16;
    $37 = $36 & 255;
    $38 = $35 >>> 8;
    $39 = $38 & 255;
    $40 = $35 & 255;
    $41 = $35 >>> 24;
    if ($19) {
     $srcB$0 = $33;$srcG$0 = $31;$srcR$0 = $29;
    } else {
     $42 = Math_imul($29, $4)|0;
     $43 = (($42>>>0) / 255)&-1;
     $44 = Math_imul($31, $6)|0;
     $45 = (($44>>>0) / 255)&-1;
     $46 = Math_imul($33, $8)|0;
     $47 = (($46>>>0) / 255)&-1;
     $srcB$0 = $47;$srcG$0 = $45;$srcR$0 = $43;
    }
    if ($21) {
     $srcA$0 = $34;
    } else {
     $48 = Math_imul($34, $9)|0;
     $49 = (($48>>>0) / 255)&-1;
     $srcA$0 = $49;
    }
    $50 = ($srcA$0>>>0)<(255);
    $or$cond = $23 & $50;
    if ($or$cond) {
     $51 = Math_imul($srcA$0, $srcR$0)|0;
     $52 = (($51>>>0) / 255)&-1;
     $53 = Math_imul($srcA$0, $srcG$0)|0;
     $54 = (($53>>>0) / 255)&-1;
     $55 = Math_imul($srcA$0, $srcB$0)|0;
     $56 = (($55>>>0) / 255)&-1;
     $srcB$1 = $56;$srcG$1 = $54;$srcR$1 = $52;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($24|0) {
    case 16:  {
     $57 = (255 - ($srcA$0))|0;
     $58 = Math_imul($57, $37)|0;
     $59 = (($58>>>0) / 255)&-1;
     $60 = (($srcR$1) + ($59))|0;
     $61 = Math_imul($57, $39)|0;
     $62 = (($61>>>0) / 255)&-1;
     $63 = (($srcG$1) + ($62))|0;
     $64 = Math_imul($57, $40)|0;
     $65 = (($64>>>0) / 255)&-1;
     $66 = (($srcB$1) + ($65))|0;
     $67 = Math_imul($57, $41)|0;
     $68 = (($67>>>0) / 255)&-1;
     $69 = (($68) + ($srcA$0))|0;
     $dstA$0 = $69;$dstB$0 = $66;$dstG$0 = $63;$dstR$0 = $60;
     break;
    }
    case 32:  {
     $70 = (($srcR$1) + ($37))|0;
     $71 = ($70>>>0)>(255);
     $$ = $71 ? 255 : $70;
     $72 = (($srcG$1) + ($39))|0;
     $73 = ($72>>>0)>(255);
     $$3 = $73 ? 255 : $72;
     $74 = (($srcB$1) + ($40))|0;
     $75 = ($74>>>0)>(255);
     $$1 = $75 ? 255 : $74;
     $dstA$0 = $41;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $76 = Math_imul($srcR$1, $37)|0;
     $77 = (($76>>>0) / 255)&-1;
     $78 = Math_imul($srcG$1, $39)|0;
     $79 = (($78>>>0) / 255)&-1;
     $80 = Math_imul($srcB$1, $40)|0;
     $81 = (($80>>>0) / 255)&-1;
     $dstA$0 = $41;$dstB$0 = $81;$dstG$0 = $79;$dstR$0 = $77;
     break;
    }
    default: {
     $dstA$0 = $41;$dstB$0 = $40;$dstG$0 = $39;$dstR$0 = $37;
    }
    }
    $82 = $dstA$0 << 24;
    $83 = $dstR$0 << 16;
    $84 = $dstG$0 << 8;
    $85 = $84 | $83;
    $86 = $85 | $dstB$0;
    $87 = $86 | $82;
    HEAP32[$dst$05>>2] = $87;
    $88 = ((($src$04)) + 4|0);
    $89 = ((($dst$05)) + 4|0);
    $90 = ($27|0)==(0);
    if ($90) {
     break;
    } else {
     $$in = $27;$dst$05 = $89;$src$04 = $88;
    }
   }
   $$pre10 = HEAP32[$10>>2]|0;
   $98 = $$pre10;
  }
  $91 = HEAP32[$16>>2]|0;
  $92 = (($93) + ($91)|0);
  $94 = HEAP32[$17>>2]|0;
  $95 = (($96) + ($94)|0);
  $97 = (($98) + -1)|0;
  HEAP32[$10>>2] = $97;
  $99 = ($98|0)==(0);
  if ($99) {
   $$lcssa = $92;$$lcssa14 = $95;
   break;
  } else {
   $100 = $97;$101 = $95;$102 = $92;$93 = $92;$96 = $95;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa14;
 return;
}
function _SDL_Blit_RGBA8888_ARGB8888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstA$0 = 0, $dstB$0 = 0;
 var $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0, $src$015 = 0, $src$1 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 << 16;
 $13 = ((($info)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12|0) / ($14|0))&-1;
 $16 = ((($info)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 << 16;
 $19 = ((($info)) + 24|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($18|0) / ($20|0))&-1;
 $22 = (($14) + -1)|0;
 HEAP32[$13>>2] = $22;
 $23 = ($14|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = $1 & 112;
 $33 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $122 = $$pre;$127 = $22;$128 = $$pre;$42 = $20;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $34 = ($posy$020|0)>(65535);
  if ($34) {
   $35 = (($posy$020) + -65536)|0;
   $36 = $35 >>> 16;
   $37 = $36 << 16;
   $38 = (($srcy$019) + 1)|0;
   $39 = (($35) - ($37))|0;
   $40 = (($38) + ($36))|0;
   $posy$1$lcssa = $39;$srcy$1$lcssa = $40;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $41 = ($42|0)==(0);
  if ($41) {
   $124 = $127;
  } else {
   $$in = $42;$dst$016 = $128;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $43 = (($$in) + -1)|0;
    $44 = ($posx$014|0)>(65535);
    if ($44) {
     $45 = (($posx$014) + -65536)|0;
     $46 = $45 >>> 16;
     $47 = $46 << 16;
     $48 = (($srcx$013) + 1)|0;
     $49 = (($45) - ($47))|0;
     $50 = (($48) + ($46))|0;
     $51 = HEAP32[$info>>2]|0;
     $52 = HEAP32[$33>>2]|0;
     $53 = Math_imul($52, $srcy$1$lcssa)|0;
     $54 = (($51) + ($53)|0);
     $55 = $50 << 2;
     $56 = (($54) + ($55)|0);
     $posx$2 = $49;$src$1 = $56;$srcx$2 = $50;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $57 = HEAP32[$src$1>>2]|0;
    $58 = $57 >>> 24;
    $59 = $57 >>> 16;
    $60 = $59 & 255;
    $61 = $57 >>> 8;
    $62 = $61 & 255;
    $63 = $57 & 255;
    $64 = HEAP32[$dst$016>>2]|0;
    $65 = $64 >>> 16;
    $66 = $65 & 255;
    $67 = $64 >>> 8;
    $68 = $67 & 255;
    $69 = $64 & 255;
    $70 = $64 >>> 24;
    if ($27) {
     $srcB$0 = $62;$srcG$0 = $60;$srcR$0 = $58;
    } else {
     $71 = Math_imul($58, $4)|0;
     $72 = (($71>>>0) / 255)&-1;
     $73 = Math_imul($60, $6)|0;
     $74 = (($73>>>0) / 255)&-1;
     $75 = Math_imul($62, $8)|0;
     $76 = (($75>>>0) / 255)&-1;
     $srcB$0 = $76;$srcG$0 = $74;$srcR$0 = $72;
    }
    if ($29) {
     $srcA$0 = $63;
    } else {
     $77 = Math_imul($63, $9)|0;
     $78 = (($77>>>0) / 255)&-1;
     $srcA$0 = $78;
    }
    $79 = ($srcA$0>>>0)<(255);
    $or$cond = $31 & $79;
    if ($or$cond) {
     $80 = Math_imul($srcA$0, $srcR$0)|0;
     $81 = (($80>>>0) / 255)&-1;
     $82 = Math_imul($srcA$0, $srcG$0)|0;
     $83 = (($82>>>0) / 255)&-1;
     $84 = Math_imul($srcA$0, $srcB$0)|0;
     $85 = (($84>>>0) / 255)&-1;
     $srcB$1 = $85;$srcG$1 = $83;$srcR$1 = $81;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($32|0) {
    case 16:  {
     $86 = (255 - ($srcA$0))|0;
     $87 = Math_imul($86, $66)|0;
     $88 = (($87>>>0) / 255)&-1;
     $89 = (($srcR$1) + ($88))|0;
     $90 = Math_imul($86, $68)|0;
     $91 = (($90>>>0) / 255)&-1;
     $92 = (($srcG$1) + ($91))|0;
     $93 = Math_imul($86, $69)|0;
     $94 = (($93>>>0) / 255)&-1;
     $95 = (($srcB$1) + ($94))|0;
     $96 = Math_imul($86, $70)|0;
     $97 = (($96>>>0) / 255)&-1;
     $98 = (($97) + ($srcA$0))|0;
     $dstA$0 = $98;$dstB$0 = $95;$dstG$0 = $92;$dstR$0 = $89;
     break;
    }
    case 32:  {
     $99 = (($srcR$1) + ($66))|0;
     $100 = ($99>>>0)>(255);
     $$ = $100 ? 255 : $99;
     $101 = (($srcG$1) + ($68))|0;
     $102 = ($101>>>0)>(255);
     $$3 = $102 ? 255 : $101;
     $103 = (($srcB$1) + ($69))|0;
     $104 = ($103>>>0)>(255);
     $$1 = $104 ? 255 : $103;
     $dstA$0 = $70;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $105 = Math_imul($srcR$1, $66)|0;
     $106 = (($105>>>0) / 255)&-1;
     $107 = Math_imul($srcG$1, $68)|0;
     $108 = (($107>>>0) / 255)&-1;
     $109 = Math_imul($srcB$1, $69)|0;
     $110 = (($109>>>0) / 255)&-1;
     $dstA$0 = $70;$dstB$0 = $110;$dstG$0 = $108;$dstR$0 = $106;
     break;
    }
    default: {
     $dstA$0 = $70;$dstB$0 = $69;$dstG$0 = $68;$dstR$0 = $66;
    }
    }
    $111 = $dstA$0 << 24;
    $112 = $dstR$0 << 16;
    $113 = $dstG$0 << 8;
    $114 = $113 | $112;
    $115 = $114 | $dstB$0;
    $116 = $115 | $111;
    HEAP32[$dst$016>>2] = $116;
    $117 = (($posx$2) + ($21))|0;
    $118 = ((($dst$016)) + 4|0);
    $119 = ($43|0)==(0);
    if ($119) {
     break;
    } else {
     $$in = $43;$dst$016 = $118;$posx$014 = $117;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$13>>2]|0;
   $124 = $$pre28;
  }
  $120 = HEAP32[$25>>2]|0;
  $121 = (($122) + ($120)|0);
  $123 = (($124) + -1)|0;
  HEAP32[$13>>2] = $123;
  $125 = ($124|0)==(0);
  if ($125) {
   $$lcssa = $121;
   break;
  }
  $126 = (($posy$1$lcssa) + ($15))|0;
  $$pre27 = HEAP32[$19>>2]|0;
  $122 = $121;$127 = $123;$128 = $121;$42 = $$pre27;$posy$020 = $126;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ABGR8888_RGB888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0, $posx$010 = 0, $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$53 = $$pre;$58 = $12;$59 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $55 = $58;
  } else {
   $$in = $25;$dst$013 = $59;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    $41 = $40 >>> 16;
    $42 = $41 & 255;
    $43 = $40 & 65280;
    $44 = $40 << 16;
    $45 = $44 & 16711680;
    $46 = $45 | $43;
    $47 = $46 | $42;
    HEAP32[$dst$013>>2] = $47;
    $48 = (($posx$2) + ($11))|0;
    $49 = ((($dst$013)) + 4|0);
    $50 = ($26|0)==(0);
    if ($50) {
     break;
    } else {
     $$in = $26;$dst$013 = $49;$posx$010 = $48;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $55 = $$pre25;
  }
  $51 = HEAP32[$15>>2]|0;
  $52 = (($53) + ($51)|0);
  $54 = (($55) + -1)|0;
  HEAP32[$3>>2] = $54;
  $56 = ($55|0)==(0);
  if ($56) {
   $$lcssa = $52;
   break;
  }
  $57 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$53 = $52;$58 = $54;$59 = $52;$posy$016 = $57;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ABGR8888_RGB888_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$$us = 0, $$1 = 0, $$1$us = 0, $$3 = 0, $$3$us = 0, $$lcssa = 0, $$lcssa17 = 0, $$pre = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$05 = 0, $dst$05$us = 0, $dstB$0 = 0, $dstB$0$us = 0, $dstG$0 = 0, $dstG$0$us = 0, $dstR$0 = 0, $dstR$0$us = 0;
 var $src$04 = 0, $src$04$us = 0, $srcB$0$us = 0, $srcG$0$us = 0, $srcR$0$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 48;
 $11 = ($10|0)==(0);
 $12 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre10 = HEAP32[$6>>2]|0;
 $113 = $$pre;$116 = $$pre10;$120 = $$pre10;$121 = $$pre;
 while(1) {
  $13 = HEAP32[$7>>2]|0;
  $14 = (($13) + -1)|0;
  $15 = ($13|0)==(0);
  L6: do {
   if (!($15)) {
    if ($11) {
     $109 = $14;$dst$05 = $120;$src$04 = $121;
     while(1) {
      $67 = HEAP32[$src$04>>2]|0;
      $68 = $67 >>> 16;
      $69 = $68 & 255;
      $70 = $67 >>> 8;
      $71 = $70 & 255;
      $72 = $67 & 255;
      $73 = HEAP32[$dst$05>>2]|0;
      $74 = $73 >>> 16;
      $75 = $74 & 255;
      $76 = $73 >>> 8;
      $77 = $76 & 255;
      $78 = $73 & 255;
      switch ($12|0) {
      case 16:  {
       $79 = $67 >>> 24;
       $80 = $79 ^ 255;
       $81 = Math_imul($75, $80)|0;
       $82 = (($81>>>0) / 255)&-1;
       $83 = (($72) + ($82))|0;
       $84 = Math_imul($77, $80)|0;
       $85 = (($84>>>0) / 255)&-1;
       $86 = (($71) + ($85))|0;
       $87 = Math_imul($80, $78)|0;
       $88 = (($87>>>0) / 255)&-1;
       $89 = (($69) + ($88))|0;
       $dstB$0 = $89;$dstG$0 = $86;$dstR$0 = $83;
       break;
      }
      case 32:  {
       $90 = (($72) + ($75))|0;
       $91 = ($90>>>0)>(255);
       $$ = $91 ? 255 : $90;
       $92 = (($71) + ($77))|0;
       $93 = ($92>>>0)>(255);
       $$3 = $93 ? 255 : $92;
       $94 = (($69) + ($78))|0;
       $95 = ($94>>>0)>(255);
       $$1 = $95 ? 255 : $94;
       $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
       break;
      }
      case 64:  {
       $96 = Math_imul($72, $75)|0;
       $97 = (($96>>>0) / 255)&-1;
       $98 = Math_imul($71, $77)|0;
       $99 = (($98>>>0) / 255)&-1;
       $100 = Math_imul($69, $78)|0;
       $101 = (($100>>>0) / 255)&-1;
       $dstB$0 = $101;$dstG$0 = $99;$dstR$0 = $97;
       break;
      }
      default: {
       $dstB$0 = $78;$dstG$0 = $77;$dstR$0 = $75;
      }
      }
      $102 = $dstR$0 << 16;
      $103 = $dstG$0 << 8;
      $104 = $103 | $102;
      $105 = $104 | $dstB$0;
      HEAP32[$dst$05>>2] = $105;
      $106 = ((($src$04)) + 4|0);
      $107 = ((($dst$05)) + 4|0);
      $108 = (($109) + -1)|0;
      $110 = ($109|0)==(0);
      if ($110) {
       break L6;
      } else {
       $109 = $108;$dst$05 = $107;$src$04 = $106;
      }
     }
    } else {
     $65 = $14;$dst$05$us = $120;$src$04$us = $121;
    }
    while(1) {
     $16 = HEAP32[$src$04$us>>2]|0;
     $17 = $16 >>> 16;
     $18 = $17 & 255;
     $19 = $16 >>> 8;
     $20 = $19 & 255;
     $21 = $16 & 255;
     $22 = $16 >>> 24;
     $23 = HEAP32[$dst$05$us>>2]|0;
     $24 = $23 >>> 16;
     $25 = $24 & 255;
     $26 = $23 >>> 8;
     $27 = $26 & 255;
     $28 = $23 & 255;
     $29 = ($22|0)==(255);
     if ($29) {
      $srcB$0$us = $18;$srcG$0$us = $20;$srcR$0$us = $21;
     } else {
      $30 = Math_imul($21, $22)|0;
      $31 = (($30>>>0) / 255)&-1;
      $32 = Math_imul($20, $22)|0;
      $33 = (($32>>>0) / 255)&-1;
      $34 = Math_imul($18, $22)|0;
      $35 = (($34>>>0) / 255)&-1;
      $srcB$0$us = $35;$srcG$0$us = $33;$srcR$0$us = $31;
     }
     switch ($12|0) {
     case 16:  {
      $48 = $22 ^ 255;
      $49 = Math_imul($25, $48)|0;
      $50 = (($49>>>0) / 255)&-1;
      $51 = (($srcR$0$us) + ($50))|0;
      $52 = Math_imul($27, $48)|0;
      $53 = (($52>>>0) / 255)&-1;
      $54 = (($srcG$0$us) + ($53))|0;
      $55 = Math_imul($48, $28)|0;
      $56 = (($55>>>0) / 255)&-1;
      $57 = (($srcB$0$us) + ($56))|0;
      $dstB$0$us = $57;$dstG$0$us = $54;$dstR$0$us = $51;
      break;
     }
     case 32:  {
      $42 = (($srcR$0$us) + ($25))|0;
      $43 = ($42>>>0)>(255);
      $$$us = $43 ? 255 : $42;
      $44 = (($srcG$0$us) + ($27))|0;
      $45 = ($44>>>0)>(255);
      $$3$us = $45 ? 255 : $44;
      $46 = (($srcB$0$us) + ($28))|0;
      $47 = ($46>>>0)>(255);
      $$1$us = $47 ? 255 : $46;
      $dstB$0$us = $$1$us;$dstG$0$us = $$3$us;$dstR$0$us = $$$us;
      break;
     }
     case 64:  {
      $36 = Math_imul($srcR$0$us, $25)|0;
      $37 = (($36>>>0) / 255)&-1;
      $38 = Math_imul($srcG$0$us, $27)|0;
      $39 = (($38>>>0) / 255)&-1;
      $40 = Math_imul($srcB$0$us, $28)|0;
      $41 = (($40>>>0) / 255)&-1;
      $dstB$0$us = $41;$dstG$0$us = $39;$dstR$0$us = $37;
      break;
     }
     default: {
      $dstB$0$us = $28;$dstG$0$us = $27;$dstR$0$us = $25;
     }
     }
     $58 = $dstR$0$us << 16;
     $59 = $dstG$0$us << 8;
     $60 = $59 | $58;
     $61 = $60 | $dstB$0$us;
     HEAP32[$dst$05$us>>2] = $61;
     $62 = ((($src$04$us)) + 4|0);
     $63 = ((($dst$05$us)) + 4|0);
     $64 = (($65) + -1)|0;
     $66 = ($65|0)==(0);
     if ($66) {
      break L6;
     } else {
      $65 = $64;$dst$05$us = $63;$src$04$us = $62;
     }
    }
   }
  } while(0);
  $111 = HEAP32[$8>>2]|0;
  $112 = (($113) + ($111)|0);
  $114 = HEAP32[$9>>2]|0;
  $115 = (($116) + ($114)|0);
  $117 = HEAP32[$2>>2]|0;
  $118 = (($117) + -1)|0;
  HEAP32[$2>>2] = $118;
  $119 = ($117|0)==(0);
  if ($119) {
   $$lcssa = $112;$$lcssa17 = $115;
   break;
  } else {
   $113 = $112;$116 = $115;$120 = $115;$121 = $112;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa17;
 return;
}
function _SDL_Blit_ABGR8888_RGB888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0;
 var $src$015 = 0, $src$1 = 0, $srcB$0 = 0, $srcG$0 = 0, $srcR$0 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 48;
 $19 = ($18|0)!=(0);
 $20 = $1 & 112;
 $21 = ((($info)) + 12|0);
 $$pre = HEAP32[$16>>2]|0;
 $101 = $14;$102 = $$pre;$30 = $12;$96 = $$pre;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $22 = ($posy$020|0)>(65535);
  if ($22) {
   $23 = (($posy$020) + -65536)|0;
   $24 = $23 >>> 16;
   $25 = $24 << 16;
   $26 = (($srcy$019) + 1)|0;
   $27 = (($23) - ($25))|0;
   $28 = (($26) + ($24))|0;
   $posy$1$lcssa = $27;$srcy$1$lcssa = $28;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $29 = ($30|0)==(0);
  if ($29) {
   $98 = $101;
  } else {
   $$in = $30;$dst$016 = $102;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $31 = (($$in) + -1)|0;
    $32 = ($posx$014|0)>(65535);
    if ($32) {
     $33 = (($posx$014) + -65536)|0;
     $34 = $33 >>> 16;
     $35 = $34 << 16;
     $36 = (($srcx$013) + 1)|0;
     $37 = (($33) - ($35))|0;
     $38 = (($36) + ($34))|0;
     $39 = HEAP32[$info>>2]|0;
     $40 = HEAP32[$21>>2]|0;
     $41 = Math_imul($40, $srcy$1$lcssa)|0;
     $42 = (($39) + ($41)|0);
     $43 = $38 << 2;
     $44 = (($42) + ($43)|0);
     $posx$2 = $37;$src$1 = $44;$srcx$2 = $38;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $45 = HEAP32[$src$1>>2]|0;
    $46 = $45 >>> 16;
    $47 = $46 & 255;
    $48 = $45 >>> 8;
    $49 = $48 & 255;
    $50 = $45 & 255;
    $51 = $45 >>> 24;
    $52 = HEAP32[$dst$016>>2]|0;
    $53 = $52 >>> 16;
    $54 = $53 & 255;
    $55 = $52 >>> 8;
    $56 = $55 & 255;
    $57 = $52 & 255;
    $58 = ($51|0)!=(255);
    $or$cond = $19 & $58;
    if ($or$cond) {
     $59 = Math_imul($50, $51)|0;
     $60 = (($59>>>0) / 255)&-1;
     $61 = Math_imul($49, $51)|0;
     $62 = (($61>>>0) / 255)&-1;
     $63 = Math_imul($47, $51)|0;
     $64 = (($63>>>0) / 255)&-1;
     $srcB$0 = $64;$srcG$0 = $62;$srcR$0 = $60;
    } else {
     $srcB$0 = $47;$srcG$0 = $49;$srcR$0 = $50;
    }
    switch ($20|0) {
    case 16:  {
     $65 = $51 ^ 255;
     $66 = Math_imul($54, $65)|0;
     $67 = (($66>>>0) / 255)&-1;
     $68 = (($srcR$0) + ($67))|0;
     $69 = Math_imul($56, $65)|0;
     $70 = (($69>>>0) / 255)&-1;
     $71 = (($srcG$0) + ($70))|0;
     $72 = Math_imul($65, $57)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = (($srcB$0) + ($73))|0;
     $dstB$0 = $74;$dstG$0 = $71;$dstR$0 = $68;
     break;
    }
    case 32:  {
     $75 = (($srcR$0) + ($54))|0;
     $76 = ($75>>>0)>(255);
     $$ = $76 ? 255 : $75;
     $77 = (($srcG$0) + ($56))|0;
     $78 = ($77>>>0)>(255);
     $$3 = $78 ? 255 : $77;
     $79 = (($srcB$0) + ($57))|0;
     $80 = ($79>>>0)>(255);
     $$1 = $80 ? 255 : $79;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $81 = Math_imul($srcR$0, $54)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = Math_imul($srcG$0, $56)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = Math_imul($srcB$0, $57)|0;
     $86 = (($85>>>0) / 255)&-1;
     $dstB$0 = $86;$dstG$0 = $84;$dstR$0 = $82;
     break;
    }
    default: {
     $dstB$0 = $57;$dstG$0 = $56;$dstR$0 = $54;
    }
    }
    $87 = $dstR$0 << 16;
    $88 = $dstG$0 << 8;
    $89 = $88 | $87;
    $90 = $89 | $dstB$0;
    HEAP32[$dst$016>>2] = $90;
    $91 = (($posx$2) + ($13))|0;
    $92 = ((($dst$016)) + 4|0);
    $93 = ($31|0)==(0);
    if ($93) {
     break;
    } else {
     $$in = $31;$dst$016 = $92;$posx$014 = $91;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$5>>2]|0;
   $98 = $$pre28;
  }
  $94 = HEAP32[$17>>2]|0;
  $95 = (($96) + ($94)|0);
  $97 = (($98) + -1)|0;
  HEAP32[$5>>2] = $97;
  $99 = ($98|0)==(0);
  if ($99) {
   $$lcssa = $95;
   break;
  }
  $100 = (($posy$1$lcssa) + ($7))|0;
  $$pre27 = HEAP32[$11>>2]|0;
  $101 = $97;$102 = $95;$30 = $$pre27;$96 = $95;$posy$020 = $100;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ABGR8888_RGB888_Modulate($info) {
 $info = $info|0;
 var $$in = 0, $$in13 = 0, $$lcssa = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$pre16 = 0, $$pre17 = 0, $$pre18 = 0, $$pre19 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $dst$02 = 0, $dst$02$us$us = 0, $src$01 = 0, $src$01$us$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = ((($info)) + 20|0);
 $14 = ((($info)) + 24|0);
 $15 = ((($info)) + 12|0);
 $16 = ((($info)) + 32|0);
 $17 = $1 & 1;
 $18 = ($17|0)==(0);
 $$pre17 = HEAP32[$info>>2]|0;
 $$pre18 = HEAP32[$13>>2]|0;
 if ($18) {
  $23 = $$pre17;$26 = $$pre18;$73 = $11;$74 = $$pre18;$75 = $$pre17;
  while(1) {
   $19 = HEAP32[$14>>2]|0;
   $20 = ($19|0)==(0);
   if ($20) {
    $28 = $73;
   } else {
    $$in13 = $19;$dst$02$us$us = $74;$src$01$us$us = $75;
    while(1) {
     $30 = (($$in13) + -1)|0;
     $31 = HEAP32[$src$01$us$us>>2]|0;
     $32 = $31 >>> 16;
     $33 = $32 & 255;
     $34 = $31 & 65280;
     $35 = $31 << 16;
     $36 = $35 & 16711680;
     $37 = $34 | $36;
     $38 = $37 | $33;
     HEAP32[$dst$02$us$us>>2] = $38;
     $39 = ((($src$01$us$us)) + 4|0);
     $40 = ((($dst$02$us$us)) + 4|0);
     $41 = ($30|0)==(0);
     if ($41) {
      break;
     } else {
      $$in13 = $30;$dst$02$us$us = $40;$src$01$us$us = $39;
     }
    }
    $$pre19 = HEAP32[$9>>2]|0;
    $28 = $$pre19;
   }
   $21 = HEAP32[$15>>2]|0;
   $22 = (($23) + ($21)|0);
   $24 = HEAP32[$16>>2]|0;
   $25 = (($26) + ($24)|0);
   $27 = (($28) + -1)|0;
   HEAP32[$9>>2] = $27;
   $29 = ($28|0)==(0);
   if ($29) {
    $$lcssa = $22;$$lcssa30 = $25;
    break;
   } else {
    $23 = $22;$26 = $25;$73 = $27;$74 = $25;$75 = $22;
   }
  }
  HEAP32[$info>>2] = $$lcssa;
  HEAP32[$13>>2] = $$lcssa30;
  return;
 } else {
  $66 = $$pre17;$69 = $$pre18;$76 = $11;$77 = $$pre18;$78 = $$pre17;
  while(1) {
   $42 = HEAP32[$14>>2]|0;
   $43 = ($42|0)==(0);
   if ($43) {
    $71 = $76;
   } else {
    $$in = $42;$dst$02 = $77;$src$01 = $78;
    while(1) {
     $44 = (($$in) + -1)|0;
     $45 = HEAP32[$src$01>>2]|0;
     $46 = $45 >>> 16;
     $47 = $46 & 255;
     $48 = $45 >>> 8;
     $49 = $48 & 255;
     $50 = $45 & 255;
     $51 = Math_imul($50, $4)|0;
     $52 = (($51>>>0) / 255)&-1;
     $53 = Math_imul($49, $6)|0;
     $54 = (($53>>>0) / 255)&-1;
     $55 = Math_imul($47, $8)|0;
     $56 = (($55>>>0) / 255)&-1;
     $57 = $52 << 16;
     $58 = $54 << 8;
     $59 = $58 | $57;
     $60 = $59 | $56;
     HEAP32[$dst$02>>2] = $60;
     $61 = ((($src$01)) + 4|0);
     $62 = ((($dst$02)) + 4|0);
     $63 = ($44|0)==(0);
     if ($63) {
      break;
     } else {
      $$in = $44;$dst$02 = $62;$src$01 = $61;
     }
    }
    $$pre16 = HEAP32[$9>>2]|0;
    $71 = $$pre16;
   }
   $64 = HEAP32[$15>>2]|0;
   $65 = (($66) + ($64)|0);
   $67 = HEAP32[$16>>2]|0;
   $68 = (($69) + ($67)|0);
   $70 = (($71) + -1)|0;
   HEAP32[$9>>2] = $70;
   $72 = ($71|0)==(0);
   if ($72) {
    $$lcssa31 = $65;$$lcssa32 = $68;
    break;
   } else {
    $66 = $65;$69 = $68;$76 = $70;$77 = $68;$78 = $65;
   }
  }
  HEAP32[$info>>2] = $$lcssa31;
  HEAP32[$13>>2] = $$lcssa32;
  return;
 }
}
function _SDL_Blit_ABGR8888_RGB888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$013 = 0, $dst$013$us = 0, $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$017 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0;
 var $src$1$us = 0, $srcx$010 = 0, $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$016 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = ((($info)) + 20|0);
 $24 = ((($info)) + 32|0);
 $25 = $1 & 1;
 $26 = ($25|0)==(0);
 $27 = ((($info)) + 12|0);
 $$pre = HEAP32[$23>>2]|0;
 $100 = $$pre;$105 = $$pre;$36 = $19;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $28 = ($posy$017|0)>(65535);
  if ($28) {
   $29 = (($posy$017) + -65536)|0;
   $30 = $29 >>> 16;
   $31 = $30 << 16;
   $32 = (($srcy$016) + 1)|0;
   $33 = (($29) - ($31))|0;
   $34 = (($32) + ($30))|0;
   $posy$1$lcssa = $33;$srcy$1$lcssa = $34;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $35 = (($36) + -1)|0;
  $37 = ($36|0)==(0);
  if (!($37)) {
   if ($26) {
    $62 = $35;$dst$013$us = $105;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $38 = ($posx$011$us|0)>(65535);
     if ($38) {
      $39 = (($posx$011$us) + -65536)|0;
      $40 = $39 >>> 16;
      $41 = $40 << 16;
      $42 = (($srcx$010$us) + 1)|0;
      $43 = (($39) - ($41))|0;
      $44 = (($42) + ($40))|0;
      $45 = HEAP32[$info>>2]|0;
      $46 = HEAP32[$27>>2]|0;
      $47 = Math_imul($46, $srcy$1$lcssa)|0;
      $48 = (($45) + ($47)|0);
      $49 = $44 << 2;
      $50 = (($48) + ($49)|0);
      $posx$2$us = $43;$src$1$us = $50;$srcx$2$us = $44;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $51 = HEAP32[$src$1$us>>2]|0;
     $52 = $51 >>> 16;
     $53 = $52 & 255;
     $54 = $51 & 65280;
     $55 = $51 << 16;
     $56 = $55 & 16711680;
     $57 = $53 | $54;
     $58 = $57 | $56;
     HEAP32[$dst$013$us>>2] = $58;
     $59 = (($posx$2$us) + ($20))|0;
     $60 = ((($dst$013$us)) + 4|0);
     $61 = (($62) + -1)|0;
     $63 = ($62|0)==(0);
     if ($63) {
      break;
     } else {
      $62 = $61;$dst$013$us = $60;$posx$011$us = $59;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $96 = $35;$dst$013 = $105;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $64 = ($posx$011|0)>(65535);
     if ($64) {
      $65 = (($posx$011) + -65536)|0;
      $66 = $65 >>> 16;
      $67 = $66 << 16;
      $68 = (($srcx$010) + 1)|0;
      $69 = (($65) - ($67))|0;
      $70 = (($68) + ($66))|0;
      $71 = HEAP32[$info>>2]|0;
      $72 = HEAP32[$27>>2]|0;
      $73 = Math_imul($72, $srcy$1$lcssa)|0;
      $74 = (($71) + ($73)|0);
      $75 = $70 << 2;
      $76 = (($74) + ($75)|0);
      $posx$2 = $69;$src$1 = $76;$srcx$2 = $70;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $77 = HEAP32[$src$1>>2]|0;
     $78 = $77 >>> 16;
     $79 = $78 & 255;
     $80 = $77 >>> 8;
     $81 = $80 & 255;
     $82 = $77 & 255;
     $83 = Math_imul($82, $4)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = Math_imul($81, $6)|0;
     $86 = (($85>>>0) / 255)&-1;
     $87 = Math_imul($79, $8)|0;
     $88 = (($87>>>0) / 255)&-1;
     $89 = $84 << 16;
     $90 = $86 << 8;
     $91 = $88 | $90;
     $92 = $91 | $89;
     HEAP32[$dst$013>>2] = $92;
     $93 = (($posx$2) + ($20))|0;
     $94 = ((($dst$013)) + 4|0);
     $95 = (($96) + -1)|0;
     $97 = ($96|0)==(0);
     if ($97) {
      break;
     } else {
      $96 = $95;$dst$013 = $94;$posx$011 = $93;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $98 = HEAP32[$24>>2]|0;
  $99 = (($100) + ($98)|0);
  $101 = HEAP32[$12>>2]|0;
  $102 = (($101) + -1)|0;
  HEAP32[$12>>2] = $102;
  $103 = ($101|0)==(0);
  if ($103) {
   $$lcssa = $99;
   break;
  }
  $104 = (($posy$1$lcssa) + ($14))|0;
  $$pre27 = HEAP32[$18>>2]|0;
  $100 = $99;$105 = $99;$36 = $$pre27;$posy$017 = $104;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$23>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ABGR8888_RGB888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$lcssa14 = 0, $$pre = 0, $$pre10 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $dst$05 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $src$04 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + -1)|0;
 HEAP32[$10>>2] = $12;
 $13 = ($11|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre9 = HEAP32[$14>>2]|0;
 $87 = $$pre;$90 = $$pre9;$94 = $12;$95 = $$pre9;$96 = $$pre;
 while(1) {
  $25 = HEAP32[$15>>2]|0;
  $26 = ($25|0)==(0);
  if ($26) {
   $92 = $94;
  } else {
   $$in = $25;$dst$05 = $95;$src$04 = $96;
   while(1) {
    $27 = (($$in) + -1)|0;
    $28 = HEAP32[$src$04>>2]|0;
    $29 = $28 >>> 16;
    $30 = $29 & 255;
    $31 = $28 >>> 8;
    $32 = $31 & 255;
    $33 = $28 & 255;
    $34 = $28 >>> 24;
    $35 = HEAP32[$dst$05>>2]|0;
    $36 = $35 >>> 16;
    $37 = $36 & 255;
    $38 = $35 >>> 8;
    $39 = $38 & 255;
    $40 = $35 & 255;
    if ($19) {
     $srcB$0 = $30;$srcG$0 = $32;$srcR$0 = $33;
    } else {
     $41 = Math_imul($33, $4)|0;
     $42 = (($41>>>0) / 255)&-1;
     $43 = Math_imul($32, $6)|0;
     $44 = (($43>>>0) / 255)&-1;
     $45 = Math_imul($30, $8)|0;
     $46 = (($45>>>0) / 255)&-1;
     $srcB$0 = $46;$srcG$0 = $44;$srcR$0 = $42;
    }
    if ($21) {
     $srcA$0 = $34;
    } else {
     $47 = Math_imul($34, $9)|0;
     $48 = (($47>>>0) / 255)&-1;
     $srcA$0 = $48;
    }
    $49 = ($srcA$0>>>0)<(255);
    $or$cond = $23 & $49;
    if ($or$cond) {
     $50 = Math_imul($srcA$0, $srcR$0)|0;
     $51 = (($50>>>0) / 255)&-1;
     $52 = Math_imul($srcA$0, $srcG$0)|0;
     $53 = (($52>>>0) / 255)&-1;
     $54 = Math_imul($srcA$0, $srcB$0)|0;
     $55 = (($54>>>0) / 255)&-1;
     $srcB$1 = $55;$srcG$1 = $53;$srcR$1 = $51;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($24|0) {
    case 16:  {
     $56 = (255 - ($srcA$0))|0;
     $57 = Math_imul($56, $37)|0;
     $58 = (($57>>>0) / 255)&-1;
     $59 = (($srcR$1) + ($58))|0;
     $60 = Math_imul($56, $39)|0;
     $61 = (($60>>>0) / 255)&-1;
     $62 = (($srcG$1) + ($61))|0;
     $63 = Math_imul($56, $40)|0;
     $64 = (($63>>>0) / 255)&-1;
     $65 = (($srcB$1) + ($64))|0;
     $dstB$0 = $65;$dstG$0 = $62;$dstR$0 = $59;
     break;
    }
    case 32:  {
     $66 = (($srcR$1) + ($37))|0;
     $67 = ($66>>>0)>(255);
     $$ = $67 ? 255 : $66;
     $68 = (($srcG$1) + ($39))|0;
     $69 = ($68>>>0)>(255);
     $$3 = $69 ? 255 : $68;
     $70 = (($srcB$1) + ($40))|0;
     $71 = ($70>>>0)>(255);
     $$1 = $71 ? 255 : $70;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $72 = Math_imul($srcR$1, $37)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($srcG$1, $39)|0;
     $75 = (($74>>>0) / 255)&-1;
     $76 = Math_imul($srcB$1, $40)|0;
     $77 = (($76>>>0) / 255)&-1;
     $dstB$0 = $77;$dstG$0 = $75;$dstR$0 = $73;
     break;
    }
    default: {
     $dstB$0 = $40;$dstG$0 = $39;$dstR$0 = $37;
    }
    }
    $78 = $dstR$0 << 16;
    $79 = $dstG$0 << 8;
    $80 = $79 | $78;
    $81 = $80 | $dstB$0;
    HEAP32[$dst$05>>2] = $81;
    $82 = ((($src$04)) + 4|0);
    $83 = ((($dst$05)) + 4|0);
    $84 = ($27|0)==(0);
    if ($84) {
     break;
    } else {
     $$in = $27;$dst$05 = $83;$src$04 = $82;
    }
   }
   $$pre10 = HEAP32[$10>>2]|0;
   $92 = $$pre10;
  }
  $85 = HEAP32[$16>>2]|0;
  $86 = (($87) + ($85)|0);
  $88 = HEAP32[$17>>2]|0;
  $89 = (($90) + ($88)|0);
  $91 = (($92) + -1)|0;
  HEAP32[$10>>2] = $91;
  $93 = ($92|0)==(0);
  if ($93) {
   $$lcssa = $86;$$lcssa14 = $89;
   break;
  } else {
   $87 = $86;$90 = $89;$94 = $91;$95 = $89;$96 = $86;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa14;
 return;
}
function _SDL_Blit_ABGR8888_RGB888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0;
 var $src$015 = 0, $src$1 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 << 16;
 $13 = ((($info)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12|0) / ($14|0))&-1;
 $16 = ((($info)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 << 16;
 $19 = ((($info)) + 24|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($18|0) / ($20|0))&-1;
 $22 = (($14) + -1)|0;
 HEAP32[$13>>2] = $22;
 $23 = ($14|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = $1 & 112;
 $33 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $116 = $$pre;$121 = $22;$122 = $$pre;$42 = $20;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $34 = ($posy$020|0)>(65535);
  if ($34) {
   $35 = (($posy$020) + -65536)|0;
   $36 = $35 >>> 16;
   $37 = $36 << 16;
   $38 = (($srcy$019) + 1)|0;
   $39 = (($35) - ($37))|0;
   $40 = (($38) + ($36))|0;
   $posy$1$lcssa = $39;$srcy$1$lcssa = $40;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $41 = ($42|0)==(0);
  if ($41) {
   $118 = $121;
  } else {
   $$in = $42;$dst$016 = $122;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $43 = (($$in) + -1)|0;
    $44 = ($posx$014|0)>(65535);
    if ($44) {
     $45 = (($posx$014) + -65536)|0;
     $46 = $45 >>> 16;
     $47 = $46 << 16;
     $48 = (($srcx$013) + 1)|0;
     $49 = (($45) - ($47))|0;
     $50 = (($48) + ($46))|0;
     $51 = HEAP32[$info>>2]|0;
     $52 = HEAP32[$33>>2]|0;
     $53 = Math_imul($52, $srcy$1$lcssa)|0;
     $54 = (($51) + ($53)|0);
     $55 = $50 << 2;
     $56 = (($54) + ($55)|0);
     $posx$2 = $49;$src$1 = $56;$srcx$2 = $50;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $57 = HEAP32[$src$1>>2]|0;
    $58 = $57 >>> 16;
    $59 = $58 & 255;
    $60 = $57 >>> 8;
    $61 = $60 & 255;
    $62 = $57 & 255;
    $63 = $57 >>> 24;
    $64 = HEAP32[$dst$016>>2]|0;
    $65 = $64 >>> 16;
    $66 = $65 & 255;
    $67 = $64 >>> 8;
    $68 = $67 & 255;
    $69 = $64 & 255;
    if ($27) {
     $srcB$0 = $59;$srcG$0 = $61;$srcR$0 = $62;
    } else {
     $70 = Math_imul($62, $4)|0;
     $71 = (($70>>>0) / 255)&-1;
     $72 = Math_imul($61, $6)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($59, $8)|0;
     $75 = (($74>>>0) / 255)&-1;
     $srcB$0 = $75;$srcG$0 = $73;$srcR$0 = $71;
    }
    if ($29) {
     $srcA$0 = $63;
    } else {
     $76 = Math_imul($63, $9)|0;
     $77 = (($76>>>0) / 255)&-1;
     $srcA$0 = $77;
    }
    $78 = ($srcA$0>>>0)<(255);
    $or$cond = $31 & $78;
    if ($or$cond) {
     $79 = Math_imul($srcA$0, $srcR$0)|0;
     $80 = (($79>>>0) / 255)&-1;
     $81 = Math_imul($srcA$0, $srcG$0)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = Math_imul($srcA$0, $srcB$0)|0;
     $84 = (($83>>>0) / 255)&-1;
     $srcB$1 = $84;$srcG$1 = $82;$srcR$1 = $80;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($32|0) {
    case 16:  {
     $85 = (255 - ($srcA$0))|0;
     $86 = Math_imul($85, $66)|0;
     $87 = (($86>>>0) / 255)&-1;
     $88 = (($srcR$1) + ($87))|0;
     $89 = Math_imul($85, $68)|0;
     $90 = (($89>>>0) / 255)&-1;
     $91 = (($srcG$1) + ($90))|0;
     $92 = Math_imul($85, $69)|0;
     $93 = (($92>>>0) / 255)&-1;
     $94 = (($srcB$1) + ($93))|0;
     $dstB$0 = $94;$dstG$0 = $91;$dstR$0 = $88;
     break;
    }
    case 32:  {
     $95 = (($srcR$1) + ($66))|0;
     $96 = ($95>>>0)>(255);
     $$ = $96 ? 255 : $95;
     $97 = (($srcG$1) + ($68))|0;
     $98 = ($97>>>0)>(255);
     $$3 = $98 ? 255 : $97;
     $99 = (($srcB$1) + ($69))|0;
     $100 = ($99>>>0)>(255);
     $$1 = $100 ? 255 : $99;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $101 = Math_imul($srcR$1, $66)|0;
     $102 = (($101>>>0) / 255)&-1;
     $103 = Math_imul($srcG$1, $68)|0;
     $104 = (($103>>>0) / 255)&-1;
     $105 = Math_imul($srcB$1, $69)|0;
     $106 = (($105>>>0) / 255)&-1;
     $dstB$0 = $106;$dstG$0 = $104;$dstR$0 = $102;
     break;
    }
    default: {
     $dstB$0 = $69;$dstG$0 = $68;$dstR$0 = $66;
    }
    }
    $107 = $dstR$0 << 16;
    $108 = $dstG$0 << 8;
    $109 = $108 | $107;
    $110 = $109 | $dstB$0;
    HEAP32[$dst$016>>2] = $110;
    $111 = (($posx$2) + ($21))|0;
    $112 = ((($dst$016)) + 4|0);
    $113 = ($43|0)==(0);
    if ($113) {
     break;
    } else {
     $$in = $43;$dst$016 = $112;$posx$014 = $111;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$13>>2]|0;
   $118 = $$pre28;
  }
  $114 = HEAP32[$25>>2]|0;
  $115 = (($116) + ($114)|0);
  $117 = (($118) + -1)|0;
  HEAP32[$13>>2] = $117;
  $119 = ($118|0)==(0);
  if ($119) {
   $$lcssa = $115;
   break;
  }
  $120 = (($posy$1$lcssa) + ($15))|0;
  $$pre27 = HEAP32[$19>>2]|0;
  $116 = $115;$121 = $117;$122 = $115;$42 = $$pre27;$posy$020 = $120;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ABGR8888_BGR888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0;
 var $posx$010 = 0, $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$47 = $$pre;$52 = $12;$53 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $49 = $52;
  } else {
   $$in = $25;$dst$013 = $53;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    $41 = $40 & 16777215;
    HEAP32[$dst$013>>2] = $41;
    $42 = (($posx$2) + ($11))|0;
    $43 = ((($dst$013)) + 4|0);
    $44 = ($26|0)==(0);
    if ($44) {
     break;
    } else {
     $$in = $26;$dst$013 = $43;$posx$010 = $42;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $49 = $$pre25;
  }
  $45 = HEAP32[$15>>2]|0;
  $46 = (($47) + ($45)|0);
  $48 = (($49) + -1)|0;
  HEAP32[$3>>2] = $48;
  $50 = ($49|0)==(0);
  if ($50) {
   $$lcssa = $46;
   break;
  }
  $51 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$47 = $46;$52 = $48;$53 = $46;$posy$016 = $51;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ABGR8888_BGR888_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$$us = 0, $$1 = 0, $$1$us = 0, $$3 = 0, $$3$us = 0, $$lcssa = 0, $$lcssa17 = 0, $$pre = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$05 = 0, $dst$05$us = 0, $dstB$0 = 0, $dstB$0$us = 0, $dstG$0 = 0, $dstG$0$us = 0, $dstR$0 = 0, $dstR$0$us = 0;
 var $src$04 = 0, $src$04$us = 0, $srcB$0$us = 0, $srcG$0$us = 0, $srcR$0$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 48;
 $11 = ($10|0)==(0);
 $12 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre10 = HEAP32[$6>>2]|0;
 $113 = $$pre;$116 = $$pre10;$120 = $$pre10;$121 = $$pre;
 while(1) {
  $13 = HEAP32[$7>>2]|0;
  $14 = (($13) + -1)|0;
  $15 = ($13|0)==(0);
  L6: do {
   if (!($15)) {
    if ($11) {
     $109 = $14;$dst$05 = $120;$src$04 = $121;
     while(1) {
      $67 = HEAP32[$src$04>>2]|0;
      $68 = $67 >>> 16;
      $69 = $68 & 255;
      $70 = $67 >>> 8;
      $71 = $70 & 255;
      $72 = $67 & 255;
      $73 = HEAP32[$dst$05>>2]|0;
      $74 = $73 >>> 16;
      $75 = $74 & 255;
      $76 = $73 >>> 8;
      $77 = $76 & 255;
      $78 = $73 & 255;
      switch ($12|0) {
      case 16:  {
       $79 = $67 >>> 24;
       $80 = $79 ^ 255;
       $81 = Math_imul($80, $78)|0;
       $82 = (($81>>>0) / 255)&-1;
       $83 = (($72) + ($82))|0;
       $84 = Math_imul($77, $80)|0;
       $85 = (($84>>>0) / 255)&-1;
       $86 = (($71) + ($85))|0;
       $87 = Math_imul($75, $80)|0;
       $88 = (($87>>>0) / 255)&-1;
       $89 = (($69) + ($88))|0;
       $dstB$0 = $89;$dstG$0 = $86;$dstR$0 = $83;
       break;
      }
      case 32:  {
       $90 = (($72) + ($78))|0;
       $91 = ($90>>>0)>(255);
       $$ = $91 ? 255 : $90;
       $92 = (($71) + ($77))|0;
       $93 = ($92>>>0)>(255);
       $$3 = $93 ? 255 : $92;
       $94 = (($69) + ($75))|0;
       $95 = ($94>>>0)>(255);
       $$1 = $95 ? 255 : $94;
       $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
       break;
      }
      case 64:  {
       $96 = Math_imul($72, $78)|0;
       $97 = (($96>>>0) / 255)&-1;
       $98 = Math_imul($71, $77)|0;
       $99 = (($98>>>0) / 255)&-1;
       $100 = Math_imul($69, $75)|0;
       $101 = (($100>>>0) / 255)&-1;
       $dstB$0 = $101;$dstG$0 = $99;$dstR$0 = $97;
       break;
      }
      default: {
       $dstB$0 = $75;$dstG$0 = $77;$dstR$0 = $78;
      }
      }
      $102 = $dstB$0 << 16;
      $103 = $dstG$0 << 8;
      $104 = $103 | $dstR$0;
      $105 = $104 | $102;
      HEAP32[$dst$05>>2] = $105;
      $106 = ((($src$04)) + 4|0);
      $107 = ((($dst$05)) + 4|0);
      $108 = (($109) + -1)|0;
      $110 = ($109|0)==(0);
      if ($110) {
       break L6;
      } else {
       $109 = $108;$dst$05 = $107;$src$04 = $106;
      }
     }
    } else {
     $65 = $14;$dst$05$us = $120;$src$04$us = $121;
    }
    while(1) {
     $16 = HEAP32[$src$04$us>>2]|0;
     $17 = $16 >>> 16;
     $18 = $17 & 255;
     $19 = $16 >>> 8;
     $20 = $19 & 255;
     $21 = $16 & 255;
     $22 = $16 >>> 24;
     $23 = HEAP32[$dst$05$us>>2]|0;
     $24 = $23 >>> 16;
     $25 = $24 & 255;
     $26 = $23 >>> 8;
     $27 = $26 & 255;
     $28 = $23 & 255;
     $29 = ($22|0)==(255);
     if ($29) {
      $srcB$0$us = $18;$srcG$0$us = $20;$srcR$0$us = $21;
     } else {
      $30 = Math_imul($21, $22)|0;
      $31 = (($30>>>0) / 255)&-1;
      $32 = Math_imul($20, $22)|0;
      $33 = (($32>>>0) / 255)&-1;
      $34 = Math_imul($18, $22)|0;
      $35 = (($34>>>0) / 255)&-1;
      $srcB$0$us = $35;$srcG$0$us = $33;$srcR$0$us = $31;
     }
     switch ($12|0) {
     case 16:  {
      $48 = $22 ^ 255;
      $49 = Math_imul($48, $28)|0;
      $50 = (($49>>>0) / 255)&-1;
      $51 = (($srcR$0$us) + ($50))|0;
      $52 = Math_imul($27, $48)|0;
      $53 = (($52>>>0) / 255)&-1;
      $54 = (($srcG$0$us) + ($53))|0;
      $55 = Math_imul($25, $48)|0;
      $56 = (($55>>>0) / 255)&-1;
      $57 = (($srcB$0$us) + ($56))|0;
      $dstB$0$us = $57;$dstG$0$us = $54;$dstR$0$us = $51;
      break;
     }
     case 32:  {
      $42 = (($srcR$0$us) + ($28))|0;
      $43 = ($42>>>0)>(255);
      $$$us = $43 ? 255 : $42;
      $44 = (($srcG$0$us) + ($27))|0;
      $45 = ($44>>>0)>(255);
      $$3$us = $45 ? 255 : $44;
      $46 = (($srcB$0$us) + ($25))|0;
      $47 = ($46>>>0)>(255);
      $$1$us = $47 ? 255 : $46;
      $dstB$0$us = $$1$us;$dstG$0$us = $$3$us;$dstR$0$us = $$$us;
      break;
     }
     case 64:  {
      $36 = Math_imul($srcR$0$us, $28)|0;
      $37 = (($36>>>0) / 255)&-1;
      $38 = Math_imul($srcG$0$us, $27)|0;
      $39 = (($38>>>0) / 255)&-1;
      $40 = Math_imul($srcB$0$us, $25)|0;
      $41 = (($40>>>0) / 255)&-1;
      $dstB$0$us = $41;$dstG$0$us = $39;$dstR$0$us = $37;
      break;
     }
     default: {
      $dstB$0$us = $25;$dstG$0$us = $27;$dstR$0$us = $28;
     }
     }
     $58 = $dstB$0$us << 16;
     $59 = $dstG$0$us << 8;
     $60 = $59 | $dstR$0$us;
     $61 = $60 | $58;
     HEAP32[$dst$05$us>>2] = $61;
     $62 = ((($src$04$us)) + 4|0);
     $63 = ((($dst$05$us)) + 4|0);
     $64 = (($65) + -1)|0;
     $66 = ($65|0)==(0);
     if ($66) {
      break L6;
     } else {
      $65 = $64;$dst$05$us = $63;$src$04$us = $62;
     }
    }
   }
  } while(0);
  $111 = HEAP32[$8>>2]|0;
  $112 = (($113) + ($111)|0);
  $114 = HEAP32[$9>>2]|0;
  $115 = (($116) + ($114)|0);
  $117 = HEAP32[$2>>2]|0;
  $118 = (($117) + -1)|0;
  HEAP32[$2>>2] = $118;
  $119 = ($117|0)==(0);
  if ($119) {
   $$lcssa = $112;$$lcssa17 = $115;
   break;
  } else {
   $113 = $112;$116 = $115;$120 = $115;$121 = $112;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa17;
 return;
}
function _SDL_Blit_ABGR8888_BGR888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0;
 var $src$015 = 0, $src$1 = 0, $srcB$0 = 0, $srcG$0 = 0, $srcR$0 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 48;
 $19 = ($18|0)!=(0);
 $20 = $1 & 112;
 $21 = ((($info)) + 12|0);
 $$pre = HEAP32[$16>>2]|0;
 $101 = $14;$102 = $$pre;$30 = $12;$96 = $$pre;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $22 = ($posy$020|0)>(65535);
  if ($22) {
   $23 = (($posy$020) + -65536)|0;
   $24 = $23 >>> 16;
   $25 = $24 << 16;
   $26 = (($srcy$019) + 1)|0;
   $27 = (($23) - ($25))|0;
   $28 = (($26) + ($24))|0;
   $posy$1$lcssa = $27;$srcy$1$lcssa = $28;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $29 = ($30|0)==(0);
  if ($29) {
   $98 = $101;
  } else {
   $$in = $30;$dst$016 = $102;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $31 = (($$in) + -1)|0;
    $32 = ($posx$014|0)>(65535);
    if ($32) {
     $33 = (($posx$014) + -65536)|0;
     $34 = $33 >>> 16;
     $35 = $34 << 16;
     $36 = (($srcx$013) + 1)|0;
     $37 = (($33) - ($35))|0;
     $38 = (($36) + ($34))|0;
     $39 = HEAP32[$info>>2]|0;
     $40 = HEAP32[$21>>2]|0;
     $41 = Math_imul($40, $srcy$1$lcssa)|0;
     $42 = (($39) + ($41)|0);
     $43 = $38 << 2;
     $44 = (($42) + ($43)|0);
     $posx$2 = $37;$src$1 = $44;$srcx$2 = $38;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $45 = HEAP32[$src$1>>2]|0;
    $46 = $45 >>> 16;
    $47 = $46 & 255;
    $48 = $45 >>> 8;
    $49 = $48 & 255;
    $50 = $45 & 255;
    $51 = $45 >>> 24;
    $52 = HEAP32[$dst$016>>2]|0;
    $53 = $52 >>> 16;
    $54 = $53 & 255;
    $55 = $52 >>> 8;
    $56 = $55 & 255;
    $57 = $52 & 255;
    $58 = ($51|0)!=(255);
    $or$cond = $19 & $58;
    if ($or$cond) {
     $59 = Math_imul($50, $51)|0;
     $60 = (($59>>>0) / 255)&-1;
     $61 = Math_imul($49, $51)|0;
     $62 = (($61>>>0) / 255)&-1;
     $63 = Math_imul($47, $51)|0;
     $64 = (($63>>>0) / 255)&-1;
     $srcB$0 = $64;$srcG$0 = $62;$srcR$0 = $60;
    } else {
     $srcB$0 = $47;$srcG$0 = $49;$srcR$0 = $50;
    }
    switch ($20|0) {
    case 16:  {
     $65 = $51 ^ 255;
     $66 = Math_imul($65, $57)|0;
     $67 = (($66>>>0) / 255)&-1;
     $68 = (($srcR$0) + ($67))|0;
     $69 = Math_imul($56, $65)|0;
     $70 = (($69>>>0) / 255)&-1;
     $71 = (($srcG$0) + ($70))|0;
     $72 = Math_imul($54, $65)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = (($srcB$0) + ($73))|0;
     $dstB$0 = $74;$dstG$0 = $71;$dstR$0 = $68;
     break;
    }
    case 32:  {
     $75 = (($srcR$0) + ($57))|0;
     $76 = ($75>>>0)>(255);
     $$ = $76 ? 255 : $75;
     $77 = (($srcG$0) + ($56))|0;
     $78 = ($77>>>0)>(255);
     $$3 = $78 ? 255 : $77;
     $79 = (($srcB$0) + ($54))|0;
     $80 = ($79>>>0)>(255);
     $$1 = $80 ? 255 : $79;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $81 = Math_imul($srcR$0, $57)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = Math_imul($srcG$0, $56)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = Math_imul($srcB$0, $54)|0;
     $86 = (($85>>>0) / 255)&-1;
     $dstB$0 = $86;$dstG$0 = $84;$dstR$0 = $82;
     break;
    }
    default: {
     $dstB$0 = $54;$dstG$0 = $56;$dstR$0 = $57;
    }
    }
    $87 = $dstB$0 << 16;
    $88 = $dstG$0 << 8;
    $89 = $88 | $dstR$0;
    $90 = $89 | $87;
    HEAP32[$dst$016>>2] = $90;
    $91 = (($posx$2) + ($13))|0;
    $92 = ((($dst$016)) + 4|0);
    $93 = ($31|0)==(0);
    if ($93) {
     break;
    } else {
     $$in = $31;$dst$016 = $92;$posx$014 = $91;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$5>>2]|0;
   $98 = $$pre28;
  }
  $94 = HEAP32[$17>>2]|0;
  $95 = (($96) + ($94)|0);
  $97 = (($98) + -1)|0;
  HEAP32[$5>>2] = $97;
  $99 = ($98|0)==(0);
  if ($99) {
   $$lcssa = $95;
   break;
  }
  $100 = (($posy$1$lcssa) + ($7))|0;
  $$pre27 = HEAP32[$11>>2]|0;
  $101 = $97;$102 = $95;$30 = $$pre27;$96 = $95;$posy$020 = $100;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ABGR8888_BGR888_Modulate($info) {
 $info = $info|0;
 var $$in = 0, $$in13 = 0, $$lcssa = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$pre16 = 0, $$pre17 = 0, $$pre18 = 0, $$pre19 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $8 = 0, $9 = 0, $dst$02 = 0, $dst$02$us$us = 0, $src$01 = 0, $src$01$us$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = ((($info)) + 20|0);
 $14 = ((($info)) + 24|0);
 $15 = ((($info)) + 12|0);
 $16 = ((($info)) + 32|0);
 $17 = $1 & 1;
 $18 = ($17|0)==(0);
 $$pre17 = HEAP32[$info>>2]|0;
 $$pre18 = HEAP32[$13>>2]|0;
 if ($18) {
  $23 = $$pre17;$26 = $$pre18;$67 = $11;$68 = $$pre18;$69 = $$pre17;
  while(1) {
   $19 = HEAP32[$14>>2]|0;
   $20 = ($19|0)==(0);
   if ($20) {
    $28 = $67;
   } else {
    $$in13 = $19;$dst$02$us$us = $68;$src$01$us$us = $69;
    while(1) {
     $30 = (($$in13) + -1)|0;
     $31 = HEAP32[$src$01$us$us>>2]|0;
     $32 = $31 & 16777215;
     HEAP32[$dst$02$us$us>>2] = $32;
     $33 = ((($src$01$us$us)) + 4|0);
     $34 = ((($dst$02$us$us)) + 4|0);
     $35 = ($30|0)==(0);
     if ($35) {
      break;
     } else {
      $$in13 = $30;$dst$02$us$us = $34;$src$01$us$us = $33;
     }
    }
    $$pre19 = HEAP32[$9>>2]|0;
    $28 = $$pre19;
   }
   $21 = HEAP32[$15>>2]|0;
   $22 = (($23) + ($21)|0);
   $24 = HEAP32[$16>>2]|0;
   $25 = (($26) + ($24)|0);
   $27 = (($28) + -1)|0;
   HEAP32[$9>>2] = $27;
   $29 = ($28|0)==(0);
   if ($29) {
    $$lcssa = $22;$$lcssa30 = $25;
    break;
   } else {
    $23 = $22;$26 = $25;$67 = $27;$68 = $25;$69 = $22;
   }
  }
  HEAP32[$info>>2] = $$lcssa;
  HEAP32[$13>>2] = $$lcssa30;
  return;
 } else {
  $60 = $$pre17;$63 = $$pre18;$70 = $11;$71 = $$pre18;$72 = $$pre17;
  while(1) {
   $36 = HEAP32[$14>>2]|0;
   $37 = ($36|0)==(0);
   if ($37) {
    $65 = $70;
   } else {
    $$in = $36;$dst$02 = $71;$src$01 = $72;
    while(1) {
     $38 = (($$in) + -1)|0;
     $39 = HEAP32[$src$01>>2]|0;
     $40 = $39 >>> 16;
     $41 = $40 & 255;
     $42 = $39 >>> 8;
     $43 = $42 & 255;
     $44 = $39 & 255;
     $45 = Math_imul($44, $4)|0;
     $46 = (($45>>>0) / 255)&-1;
     $47 = Math_imul($43, $6)|0;
     $48 = (($47>>>0) / 255)&-1;
     $49 = Math_imul($41, $8)|0;
     $50 = (($49>>>0) / 255)&-1;
     $51 = $50 << 16;
     $52 = $48 << 8;
     $53 = $52 | $46;
     $54 = $53 | $51;
     HEAP32[$dst$02>>2] = $54;
     $55 = ((($src$01)) + 4|0);
     $56 = ((($dst$02)) + 4|0);
     $57 = ($38|0)==(0);
     if ($57) {
      break;
     } else {
      $$in = $38;$dst$02 = $56;$src$01 = $55;
     }
    }
    $$pre16 = HEAP32[$9>>2]|0;
    $65 = $$pre16;
   }
   $58 = HEAP32[$15>>2]|0;
   $59 = (($60) + ($58)|0);
   $61 = HEAP32[$16>>2]|0;
   $62 = (($63) + ($61)|0);
   $64 = (($65) + -1)|0;
   HEAP32[$9>>2] = $64;
   $66 = ($65|0)==(0);
   if ($66) {
    $$lcssa31 = $59;$$lcssa32 = $62;
    break;
   } else {
    $60 = $59;$63 = $62;$70 = $64;$71 = $62;$72 = $59;
   }
  }
  HEAP32[$info>>2] = $$lcssa31;
  HEAP32[$13>>2] = $$lcssa32;
  return;
 }
}
function _SDL_Blit_ABGR8888_BGR888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $dst$013 = 0, $dst$013$us = 0, $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$017 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0, $src$1$us = 0, $srcx$010 = 0, $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$016 = 0;
 var $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = ((($info)) + 20|0);
 $24 = ((($info)) + 32|0);
 $25 = $1 & 1;
 $26 = ($25|0)==(0);
 $27 = ((($info)) + 12|0);
 $$pre = HEAP32[$23>>2]|0;
 $36 = $19;$94 = $$pre;$99 = $$pre;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $28 = ($posy$017|0)>(65535);
  if ($28) {
   $29 = (($posy$017) + -65536)|0;
   $30 = $29 >>> 16;
   $31 = $30 << 16;
   $32 = (($srcy$016) + 1)|0;
   $33 = (($29) - ($31))|0;
   $34 = (($32) + ($30))|0;
   $posy$1$lcssa = $33;$srcy$1$lcssa = $34;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $35 = (($36) + -1)|0;
  $37 = ($36|0)==(0);
  if (!($37)) {
   if ($26) {
    $56 = $35;$dst$013$us = $99;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $38 = ($posx$011$us|0)>(65535);
     if ($38) {
      $39 = (($posx$011$us) + -65536)|0;
      $40 = $39 >>> 16;
      $41 = $40 << 16;
      $42 = (($srcx$010$us) + 1)|0;
      $43 = (($39) - ($41))|0;
      $44 = (($42) + ($40))|0;
      $45 = HEAP32[$info>>2]|0;
      $46 = HEAP32[$27>>2]|0;
      $47 = Math_imul($46, $srcy$1$lcssa)|0;
      $48 = (($45) + ($47)|0);
      $49 = $44 << 2;
      $50 = (($48) + ($49)|0);
      $posx$2$us = $43;$src$1$us = $50;$srcx$2$us = $44;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $51 = HEAP32[$src$1$us>>2]|0;
     $52 = $51 & 16777215;
     HEAP32[$dst$013$us>>2] = $52;
     $53 = (($posx$2$us) + ($20))|0;
     $54 = ((($dst$013$us)) + 4|0);
     $55 = (($56) + -1)|0;
     $57 = ($56|0)==(0);
     if ($57) {
      break;
     } else {
      $56 = $55;$dst$013$us = $54;$posx$011$us = $53;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $90 = $35;$dst$013 = $99;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $58 = ($posx$011|0)>(65535);
     if ($58) {
      $59 = (($posx$011) + -65536)|0;
      $60 = $59 >>> 16;
      $61 = $60 << 16;
      $62 = (($srcx$010) + 1)|0;
      $63 = (($59) - ($61))|0;
      $64 = (($62) + ($60))|0;
      $65 = HEAP32[$info>>2]|0;
      $66 = HEAP32[$27>>2]|0;
      $67 = Math_imul($66, $srcy$1$lcssa)|0;
      $68 = (($65) + ($67)|0);
      $69 = $64 << 2;
      $70 = (($68) + ($69)|0);
      $posx$2 = $63;$src$1 = $70;$srcx$2 = $64;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $71 = HEAP32[$src$1>>2]|0;
     $72 = $71 >>> 16;
     $73 = $72 & 255;
     $74 = $71 >>> 8;
     $75 = $74 & 255;
     $76 = $71 & 255;
     $77 = Math_imul($76, $4)|0;
     $78 = (($77>>>0) / 255)&-1;
     $79 = Math_imul($75, $6)|0;
     $80 = (($79>>>0) / 255)&-1;
     $81 = Math_imul($73, $8)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = $82 << 16;
     $84 = $80 << 8;
     $85 = $83 | $84;
     $86 = $85 | $78;
     HEAP32[$dst$013>>2] = $86;
     $87 = (($posx$2) + ($20))|0;
     $88 = ((($dst$013)) + 4|0);
     $89 = (($90) + -1)|0;
     $91 = ($90|0)==(0);
     if ($91) {
      break;
     } else {
      $90 = $89;$dst$013 = $88;$posx$011 = $87;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $92 = HEAP32[$24>>2]|0;
  $93 = (($94) + ($92)|0);
  $95 = HEAP32[$12>>2]|0;
  $96 = (($95) + -1)|0;
  HEAP32[$12>>2] = $96;
  $97 = ($95|0)==(0);
  if ($97) {
   $$lcssa = $93;
   break;
  }
  $98 = (($posy$1$lcssa) + ($14))|0;
  $$pre27 = HEAP32[$18>>2]|0;
  $36 = $$pre27;$94 = $93;$99 = $93;$posy$017 = $98;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$23>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ABGR8888_BGR888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$lcssa14 = 0, $$pre = 0, $$pre10 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $dst$05 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $src$04 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + -1)|0;
 HEAP32[$10>>2] = $12;
 $13 = ($11|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre9 = HEAP32[$14>>2]|0;
 $87 = $$pre;$90 = $$pre9;$94 = $12;$95 = $$pre9;$96 = $$pre;
 while(1) {
  $25 = HEAP32[$15>>2]|0;
  $26 = ($25|0)==(0);
  if ($26) {
   $92 = $94;
  } else {
   $$in = $25;$dst$05 = $95;$src$04 = $96;
   while(1) {
    $27 = (($$in) + -1)|0;
    $28 = HEAP32[$src$04>>2]|0;
    $29 = $28 >>> 16;
    $30 = $29 & 255;
    $31 = $28 >>> 8;
    $32 = $31 & 255;
    $33 = $28 & 255;
    $34 = $28 >>> 24;
    $35 = HEAP32[$dst$05>>2]|0;
    $36 = $35 >>> 16;
    $37 = $36 & 255;
    $38 = $35 >>> 8;
    $39 = $38 & 255;
    $40 = $35 & 255;
    if ($19) {
     $srcB$0 = $30;$srcG$0 = $32;$srcR$0 = $33;
    } else {
     $41 = Math_imul($33, $4)|0;
     $42 = (($41>>>0) / 255)&-1;
     $43 = Math_imul($32, $6)|0;
     $44 = (($43>>>0) / 255)&-1;
     $45 = Math_imul($30, $8)|0;
     $46 = (($45>>>0) / 255)&-1;
     $srcB$0 = $46;$srcG$0 = $44;$srcR$0 = $42;
    }
    if ($21) {
     $srcA$0 = $34;
    } else {
     $47 = Math_imul($34, $9)|0;
     $48 = (($47>>>0) / 255)&-1;
     $srcA$0 = $48;
    }
    $49 = ($srcA$0>>>0)<(255);
    $or$cond = $23 & $49;
    if ($or$cond) {
     $50 = Math_imul($srcA$0, $srcR$0)|0;
     $51 = (($50>>>0) / 255)&-1;
     $52 = Math_imul($srcA$0, $srcG$0)|0;
     $53 = (($52>>>0) / 255)&-1;
     $54 = Math_imul($srcA$0, $srcB$0)|0;
     $55 = (($54>>>0) / 255)&-1;
     $srcB$1 = $55;$srcG$1 = $53;$srcR$1 = $51;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($24|0) {
    case 16:  {
     $56 = (255 - ($srcA$0))|0;
     $57 = Math_imul($56, $40)|0;
     $58 = (($57>>>0) / 255)&-1;
     $59 = (($srcR$1) + ($58))|0;
     $60 = Math_imul($56, $39)|0;
     $61 = (($60>>>0) / 255)&-1;
     $62 = (($srcG$1) + ($61))|0;
     $63 = Math_imul($56, $37)|0;
     $64 = (($63>>>0) / 255)&-1;
     $65 = (($srcB$1) + ($64))|0;
     $dstB$0 = $65;$dstG$0 = $62;$dstR$0 = $59;
     break;
    }
    case 32:  {
     $66 = (($srcR$1) + ($40))|0;
     $67 = ($66>>>0)>(255);
     $$ = $67 ? 255 : $66;
     $68 = (($srcG$1) + ($39))|0;
     $69 = ($68>>>0)>(255);
     $$3 = $69 ? 255 : $68;
     $70 = (($srcB$1) + ($37))|0;
     $71 = ($70>>>0)>(255);
     $$1 = $71 ? 255 : $70;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $72 = Math_imul($srcR$1, $40)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($srcG$1, $39)|0;
     $75 = (($74>>>0) / 255)&-1;
     $76 = Math_imul($srcB$1, $37)|0;
     $77 = (($76>>>0) / 255)&-1;
     $dstB$0 = $77;$dstG$0 = $75;$dstR$0 = $73;
     break;
    }
    default: {
     $dstB$0 = $37;$dstG$0 = $39;$dstR$0 = $40;
    }
    }
    $78 = $dstB$0 << 16;
    $79 = $dstG$0 << 8;
    $80 = $79 | $dstR$0;
    $81 = $80 | $78;
    HEAP32[$dst$05>>2] = $81;
    $82 = ((($src$04)) + 4|0);
    $83 = ((($dst$05)) + 4|0);
    $84 = ($27|0)==(0);
    if ($84) {
     break;
    } else {
     $$in = $27;$dst$05 = $83;$src$04 = $82;
    }
   }
   $$pre10 = HEAP32[$10>>2]|0;
   $92 = $$pre10;
  }
  $85 = HEAP32[$16>>2]|0;
  $86 = (($87) + ($85)|0);
  $88 = HEAP32[$17>>2]|0;
  $89 = (($90) + ($88)|0);
  $91 = (($92) + -1)|0;
  HEAP32[$10>>2] = $91;
  $93 = ($92|0)==(0);
  if ($93) {
   $$lcssa = $86;$$lcssa14 = $89;
   break;
  } else {
   $87 = $86;$90 = $89;$94 = $91;$95 = $89;$96 = $86;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa14;
 return;
}
function _SDL_Blit_ABGR8888_BGR888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0;
 var $src$015 = 0, $src$1 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 << 16;
 $13 = ((($info)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12|0) / ($14|0))&-1;
 $16 = ((($info)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 << 16;
 $19 = ((($info)) + 24|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($18|0) / ($20|0))&-1;
 $22 = (($14) + -1)|0;
 HEAP32[$13>>2] = $22;
 $23 = ($14|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = $1 & 112;
 $33 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $116 = $$pre;$121 = $22;$122 = $$pre;$42 = $20;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $34 = ($posy$020|0)>(65535);
  if ($34) {
   $35 = (($posy$020) + -65536)|0;
   $36 = $35 >>> 16;
   $37 = $36 << 16;
   $38 = (($srcy$019) + 1)|0;
   $39 = (($35) - ($37))|0;
   $40 = (($38) + ($36))|0;
   $posy$1$lcssa = $39;$srcy$1$lcssa = $40;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $41 = ($42|0)==(0);
  if ($41) {
   $118 = $121;
  } else {
   $$in = $42;$dst$016 = $122;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $43 = (($$in) + -1)|0;
    $44 = ($posx$014|0)>(65535);
    if ($44) {
     $45 = (($posx$014) + -65536)|0;
     $46 = $45 >>> 16;
     $47 = $46 << 16;
     $48 = (($srcx$013) + 1)|0;
     $49 = (($45) - ($47))|0;
     $50 = (($48) + ($46))|0;
     $51 = HEAP32[$info>>2]|0;
     $52 = HEAP32[$33>>2]|0;
     $53 = Math_imul($52, $srcy$1$lcssa)|0;
     $54 = (($51) + ($53)|0);
     $55 = $50 << 2;
     $56 = (($54) + ($55)|0);
     $posx$2 = $49;$src$1 = $56;$srcx$2 = $50;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $57 = HEAP32[$src$1>>2]|0;
    $58 = $57 >>> 16;
    $59 = $58 & 255;
    $60 = $57 >>> 8;
    $61 = $60 & 255;
    $62 = $57 & 255;
    $63 = $57 >>> 24;
    $64 = HEAP32[$dst$016>>2]|0;
    $65 = $64 >>> 16;
    $66 = $65 & 255;
    $67 = $64 >>> 8;
    $68 = $67 & 255;
    $69 = $64 & 255;
    if ($27) {
     $srcB$0 = $59;$srcG$0 = $61;$srcR$0 = $62;
    } else {
     $70 = Math_imul($62, $4)|0;
     $71 = (($70>>>0) / 255)&-1;
     $72 = Math_imul($61, $6)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($59, $8)|0;
     $75 = (($74>>>0) / 255)&-1;
     $srcB$0 = $75;$srcG$0 = $73;$srcR$0 = $71;
    }
    if ($29) {
     $srcA$0 = $63;
    } else {
     $76 = Math_imul($63, $9)|0;
     $77 = (($76>>>0) / 255)&-1;
     $srcA$0 = $77;
    }
    $78 = ($srcA$0>>>0)<(255);
    $or$cond = $31 & $78;
    if ($or$cond) {
     $79 = Math_imul($srcA$0, $srcR$0)|0;
     $80 = (($79>>>0) / 255)&-1;
     $81 = Math_imul($srcA$0, $srcG$0)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = Math_imul($srcA$0, $srcB$0)|0;
     $84 = (($83>>>0) / 255)&-1;
     $srcB$1 = $84;$srcG$1 = $82;$srcR$1 = $80;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($32|0) {
    case 16:  {
     $85 = (255 - ($srcA$0))|0;
     $86 = Math_imul($85, $69)|0;
     $87 = (($86>>>0) / 255)&-1;
     $88 = (($srcR$1) + ($87))|0;
     $89 = Math_imul($85, $68)|0;
     $90 = (($89>>>0) / 255)&-1;
     $91 = (($srcG$1) + ($90))|0;
     $92 = Math_imul($85, $66)|0;
     $93 = (($92>>>0) / 255)&-1;
     $94 = (($srcB$1) + ($93))|0;
     $dstB$0 = $94;$dstG$0 = $91;$dstR$0 = $88;
     break;
    }
    case 32:  {
     $95 = (($srcR$1) + ($69))|0;
     $96 = ($95>>>0)>(255);
     $$ = $96 ? 255 : $95;
     $97 = (($srcG$1) + ($68))|0;
     $98 = ($97>>>0)>(255);
     $$3 = $98 ? 255 : $97;
     $99 = (($srcB$1) + ($66))|0;
     $100 = ($99>>>0)>(255);
     $$1 = $100 ? 255 : $99;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $101 = Math_imul($srcR$1, $69)|0;
     $102 = (($101>>>0) / 255)&-1;
     $103 = Math_imul($srcG$1, $68)|0;
     $104 = (($103>>>0) / 255)&-1;
     $105 = Math_imul($srcB$1, $66)|0;
     $106 = (($105>>>0) / 255)&-1;
     $dstB$0 = $106;$dstG$0 = $104;$dstR$0 = $102;
     break;
    }
    default: {
     $dstB$0 = $66;$dstG$0 = $68;$dstR$0 = $69;
    }
    }
    $107 = $dstB$0 << 16;
    $108 = $dstG$0 << 8;
    $109 = $108 | $dstR$0;
    $110 = $109 | $107;
    HEAP32[$dst$016>>2] = $110;
    $111 = (($posx$2) + ($21))|0;
    $112 = ((($dst$016)) + 4|0);
    $113 = ($43|0)==(0);
    if ($113) {
     break;
    } else {
     $$in = $43;$dst$016 = $112;$posx$014 = $111;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$13>>2]|0;
   $118 = $$pre28;
  }
  $114 = HEAP32[$25>>2]|0;
  $115 = (($116) + ($114)|0);
  $117 = (($118) + -1)|0;
  HEAP32[$13>>2] = $117;
  $119 = ($118|0)==(0);
  if ($119) {
   $$lcssa = $115;
   break;
  }
  $120 = (($posy$1$lcssa) + ($15))|0;
  $$pre27 = HEAP32[$19>>2]|0;
  $116 = $115;$121 = $117;$122 = $115;$42 = $$pre27;$posy$020 = $120;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ABGR8888_ARGB8888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0, $posx$010 = 0, $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$53 = $$pre;$58 = $12;$59 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $55 = $58;
  } else {
   $$in = $25;$dst$013 = $59;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    $41 = $40 >>> 16;
    $42 = $41 & 255;
    $43 = $40 << 16;
    $44 = $43 & 16711680;
    $45 = $40 & -16711936;
    $46 = $45 | $44;
    $47 = $46 | $42;
    HEAP32[$dst$013>>2] = $47;
    $48 = (($posx$2) + ($11))|0;
    $49 = ((($dst$013)) + 4|0);
    $50 = ($26|0)==(0);
    if ($50) {
     break;
    } else {
     $$in = $26;$dst$013 = $49;$posx$010 = $48;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $55 = $$pre25;
  }
  $51 = HEAP32[$15>>2]|0;
  $52 = (($53) + ($51)|0);
  $54 = (($55) + -1)|0;
  HEAP32[$3>>2] = $54;
  $56 = ($55|0)==(0);
  if ($56) {
   $$lcssa = $52;
   break;
  }
  $57 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$53 = $52;$58 = $54;$59 = $52;$posy$016 = $57;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ABGR8888_ARGB8888_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$$us = 0, $$1 = 0, $$1$us = 0, $$3 = 0, $$3$us = 0, $$lcssa = 0, $$lcssa17 = 0, $$pre = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$05 = 0, $dst$05$us = 0, $dstA$0 = 0, $dstA$0$us = 0, $dstB$0 = 0, $dstB$0$us = 0, $dstG$0 = 0, $dstG$0$us = 0, $dstR$0 = 0, $dstR$0$us = 0, $src$04 = 0, $src$04$us = 0, $srcB$0$us = 0, $srcG$0$us = 0, $srcR$0$us = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 48;
 $11 = ($10|0)==(0);
 $12 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre10 = HEAP32[$6>>2]|0;
 $125 = $$pre;$128 = $$pre10;$132 = $$pre10;$133 = $$pre;
 while(1) {
  $13 = HEAP32[$7>>2]|0;
  $14 = (($13) + -1)|0;
  $15 = ($13|0)==(0);
  L6: do {
   if (!($15)) {
    if ($11) {
     $121 = $14;$dst$05 = $132;$src$04 = $133;
     while(1) {
      $73 = HEAP32[$src$04>>2]|0;
      $74 = $73 >>> 16;
      $75 = $74 & 255;
      $76 = $73 >>> 8;
      $77 = $76 & 255;
      $78 = $73 & 255;
      $79 = $73 >>> 24;
      $80 = HEAP32[$dst$05>>2]|0;
      $81 = $80 >>> 16;
      $82 = $81 & 255;
      $83 = $80 >>> 8;
      $84 = $83 & 255;
      $85 = $80 & 255;
      $86 = $80 >>> 24;
      switch ($12|0) {
      case 16:  {
       $87 = $79 ^ 255;
       $88 = Math_imul($82, $87)|0;
       $89 = (($88>>>0) / 255)&-1;
       $90 = (($78) + ($89))|0;
       $91 = Math_imul($84, $87)|0;
       $92 = (($91>>>0) / 255)&-1;
       $93 = (($77) + ($92))|0;
       $94 = Math_imul($87, $85)|0;
       $95 = (($94>>>0) / 255)&-1;
       $96 = (($75) + ($95))|0;
       $97 = Math_imul($87, $86)|0;
       $98 = (($97>>>0) / 255)&-1;
       $99 = (($98) + ($79))|0;
       $dstA$0 = $99;$dstB$0 = $96;$dstG$0 = $93;$dstR$0 = $90;
       break;
      }
      case 32:  {
       $100 = (($78) + ($82))|0;
       $101 = ($100>>>0)>(255);
       $$ = $101 ? 255 : $100;
       $102 = (($77) + ($84))|0;
       $103 = ($102>>>0)>(255);
       $$3 = $103 ? 255 : $102;
       $104 = (($75) + ($85))|0;
       $105 = ($104>>>0)>(255);
       $$1 = $105 ? 255 : $104;
       $dstA$0 = $86;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
       break;
      }
      case 64:  {
       $106 = Math_imul($78, $82)|0;
       $107 = (($106>>>0) / 255)&-1;
       $108 = Math_imul($77, $84)|0;
       $109 = (($108>>>0) / 255)&-1;
       $110 = Math_imul($75, $85)|0;
       $111 = (($110>>>0) / 255)&-1;
       $dstA$0 = $86;$dstB$0 = $111;$dstG$0 = $109;$dstR$0 = $107;
       break;
      }
      default: {
       $dstA$0 = $86;$dstB$0 = $85;$dstG$0 = $84;$dstR$0 = $82;
      }
      }
      $112 = $dstA$0 << 24;
      $113 = $dstR$0 << 16;
      $114 = $dstG$0 << 8;
      $115 = $114 | $113;
      $116 = $115 | $dstB$0;
      $117 = $116 | $112;
      HEAP32[$dst$05>>2] = $117;
      $118 = ((($src$04)) + 4|0);
      $119 = ((($dst$05)) + 4|0);
      $120 = (($121) + -1)|0;
      $122 = ($121|0)==(0);
      if ($122) {
       break L6;
      } else {
       $121 = $120;$dst$05 = $119;$src$04 = $118;
      }
     }
    } else {
     $71 = $14;$dst$05$us = $132;$src$04$us = $133;
    }
    while(1) {
     $16 = HEAP32[$src$04$us>>2]|0;
     $17 = $16 >>> 16;
     $18 = $17 & 255;
     $19 = $16 >>> 8;
     $20 = $19 & 255;
     $21 = $16 & 255;
     $22 = $16 >>> 24;
     $23 = HEAP32[$dst$05$us>>2]|0;
     $24 = $23 >>> 16;
     $25 = $24 & 255;
     $26 = $23 >>> 8;
     $27 = $26 & 255;
     $28 = $23 & 255;
     $29 = $23 >>> 24;
     $30 = ($22|0)==(255);
     if ($30) {
      $srcB$0$us = $18;$srcG$0$us = $20;$srcR$0$us = $21;
     } else {
      $31 = Math_imul($21, $22)|0;
      $32 = (($31>>>0) / 255)&-1;
      $33 = Math_imul($20, $22)|0;
      $34 = (($33>>>0) / 255)&-1;
      $35 = Math_imul($18, $22)|0;
      $36 = (($35>>>0) / 255)&-1;
      $srcB$0$us = $36;$srcG$0$us = $34;$srcR$0$us = $32;
     }
     switch ($12|0) {
     case 16:  {
      $49 = $22 ^ 255;
      $50 = Math_imul($25, $49)|0;
      $51 = (($50>>>0) / 255)&-1;
      $52 = (($srcR$0$us) + ($51))|0;
      $53 = Math_imul($27, $49)|0;
      $54 = (($53>>>0) / 255)&-1;
      $55 = (($srcG$0$us) + ($54))|0;
      $56 = Math_imul($49, $28)|0;
      $57 = (($56>>>0) / 255)&-1;
      $58 = (($srcB$0$us) + ($57))|0;
      $59 = Math_imul($49, $29)|0;
      $60 = (($59>>>0) / 255)&-1;
      $61 = (($60) + ($22))|0;
      $dstA$0$us = $61;$dstB$0$us = $58;$dstG$0$us = $55;$dstR$0$us = $52;
      break;
     }
     case 32:  {
      $43 = (($srcR$0$us) + ($25))|0;
      $44 = ($43>>>0)>(255);
      $$$us = $44 ? 255 : $43;
      $45 = (($srcG$0$us) + ($27))|0;
      $46 = ($45>>>0)>(255);
      $$3$us = $46 ? 255 : $45;
      $47 = (($srcB$0$us) + ($28))|0;
      $48 = ($47>>>0)>(255);
      $$1$us = $48 ? 255 : $47;
      $dstA$0$us = $29;$dstB$0$us = $$1$us;$dstG$0$us = $$3$us;$dstR$0$us = $$$us;
      break;
     }
     case 64:  {
      $37 = Math_imul($srcR$0$us, $25)|0;
      $38 = (($37>>>0) / 255)&-1;
      $39 = Math_imul($srcG$0$us, $27)|0;
      $40 = (($39>>>0) / 255)&-1;
      $41 = Math_imul($srcB$0$us, $28)|0;
      $42 = (($41>>>0) / 255)&-1;
      $dstA$0$us = $29;$dstB$0$us = $42;$dstG$0$us = $40;$dstR$0$us = $38;
      break;
     }
     default: {
      $dstA$0$us = $29;$dstB$0$us = $28;$dstG$0$us = $27;$dstR$0$us = $25;
     }
     }
     $62 = $dstA$0$us << 24;
     $63 = $dstR$0$us << 16;
     $64 = $dstG$0$us << 8;
     $65 = $64 | $63;
     $66 = $65 | $dstB$0$us;
     $67 = $66 | $62;
     HEAP32[$dst$05$us>>2] = $67;
     $68 = ((($src$04$us)) + 4|0);
     $69 = ((($dst$05$us)) + 4|0);
     $70 = (($71) + -1)|0;
     $72 = ($71|0)==(0);
     if ($72) {
      break L6;
     } else {
      $71 = $70;$dst$05$us = $69;$src$04$us = $68;
     }
    }
   }
  } while(0);
  $123 = HEAP32[$8>>2]|0;
  $124 = (($125) + ($123)|0);
  $126 = HEAP32[$9>>2]|0;
  $127 = (($128) + ($126)|0);
  $129 = HEAP32[$2>>2]|0;
  $130 = (($129) + -1)|0;
  HEAP32[$2>>2] = $130;
  $131 = ($129|0)==(0);
  if ($131) {
   $$lcssa = $124;$$lcssa17 = $127;
   break;
  } else {
   $125 = $124;$128 = $127;$132 = $127;$133 = $124;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa17;
 return;
}
function _SDL_Blit_ABGR8888_ARGB8888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstA$0 = 0, $dstB$0 = 0;
 var $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0, $src$015 = 0, $src$1 = 0, $srcB$0 = 0, $srcG$0 = 0, $srcR$0 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 48;
 $19 = ($18|0)!=(0);
 $20 = $1 & 112;
 $21 = ((($info)) + 12|0);
 $$pre = HEAP32[$16>>2]|0;
 $102 = $$pre;$107 = $14;$108 = $$pre;$30 = $12;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $22 = ($posy$020|0)>(65535);
  if ($22) {
   $23 = (($posy$020) + -65536)|0;
   $24 = $23 >>> 16;
   $25 = $24 << 16;
   $26 = (($srcy$019) + 1)|0;
   $27 = (($23) - ($25))|0;
   $28 = (($26) + ($24))|0;
   $posy$1$lcssa = $27;$srcy$1$lcssa = $28;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $29 = ($30|0)==(0);
  if ($29) {
   $104 = $107;
  } else {
   $$in = $30;$dst$016 = $108;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $31 = (($$in) + -1)|0;
    $32 = ($posx$014|0)>(65535);
    if ($32) {
     $33 = (($posx$014) + -65536)|0;
     $34 = $33 >>> 16;
     $35 = $34 << 16;
     $36 = (($srcx$013) + 1)|0;
     $37 = (($33) - ($35))|0;
     $38 = (($36) + ($34))|0;
     $39 = HEAP32[$info>>2]|0;
     $40 = HEAP32[$21>>2]|0;
     $41 = Math_imul($40, $srcy$1$lcssa)|0;
     $42 = (($39) + ($41)|0);
     $43 = $38 << 2;
     $44 = (($42) + ($43)|0);
     $posx$2 = $37;$src$1 = $44;$srcx$2 = $38;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $45 = HEAP32[$src$1>>2]|0;
    $46 = $45 >>> 16;
    $47 = $46 & 255;
    $48 = $45 >>> 8;
    $49 = $48 & 255;
    $50 = $45 & 255;
    $51 = $45 >>> 24;
    $52 = HEAP32[$dst$016>>2]|0;
    $53 = $52 >>> 16;
    $54 = $53 & 255;
    $55 = $52 >>> 8;
    $56 = $55 & 255;
    $57 = $52 & 255;
    $58 = $52 >>> 24;
    $59 = ($51|0)!=(255);
    $or$cond = $19 & $59;
    if ($or$cond) {
     $60 = Math_imul($50, $51)|0;
     $61 = (($60>>>0) / 255)&-1;
     $62 = Math_imul($49, $51)|0;
     $63 = (($62>>>0) / 255)&-1;
     $64 = Math_imul($47, $51)|0;
     $65 = (($64>>>0) / 255)&-1;
     $srcB$0 = $65;$srcG$0 = $63;$srcR$0 = $61;
    } else {
     $srcB$0 = $47;$srcG$0 = $49;$srcR$0 = $50;
    }
    switch ($20|0) {
    case 16:  {
     $66 = $51 ^ 255;
     $67 = Math_imul($54, $66)|0;
     $68 = (($67>>>0) / 255)&-1;
     $69 = (($srcR$0) + ($68))|0;
     $70 = Math_imul($56, $66)|0;
     $71 = (($70>>>0) / 255)&-1;
     $72 = (($srcG$0) + ($71))|0;
     $73 = Math_imul($66, $57)|0;
     $74 = (($73>>>0) / 255)&-1;
     $75 = (($srcB$0) + ($74))|0;
     $76 = Math_imul($66, $58)|0;
     $77 = (($76>>>0) / 255)&-1;
     $78 = (($77) + ($51))|0;
     $dstA$0 = $78;$dstB$0 = $75;$dstG$0 = $72;$dstR$0 = $69;
     break;
    }
    case 32:  {
     $79 = (($srcR$0) + ($54))|0;
     $80 = ($79>>>0)>(255);
     $$ = $80 ? 255 : $79;
     $81 = (($srcG$0) + ($56))|0;
     $82 = ($81>>>0)>(255);
     $$3 = $82 ? 255 : $81;
     $83 = (($srcB$0) + ($57))|0;
     $84 = ($83>>>0)>(255);
     $$1 = $84 ? 255 : $83;
     $dstA$0 = $58;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $85 = Math_imul($srcR$0, $54)|0;
     $86 = (($85>>>0) / 255)&-1;
     $87 = Math_imul($srcG$0, $56)|0;
     $88 = (($87>>>0) / 255)&-1;
     $89 = Math_imul($srcB$0, $57)|0;
     $90 = (($89>>>0) / 255)&-1;
     $dstA$0 = $58;$dstB$0 = $90;$dstG$0 = $88;$dstR$0 = $86;
     break;
    }
    default: {
     $dstA$0 = $58;$dstB$0 = $57;$dstG$0 = $56;$dstR$0 = $54;
    }
    }
    $91 = $dstA$0 << 24;
    $92 = $dstR$0 << 16;
    $93 = $dstG$0 << 8;
    $94 = $93 | $92;
    $95 = $94 | $dstB$0;
    $96 = $95 | $91;
    HEAP32[$dst$016>>2] = $96;
    $97 = (($posx$2) + ($13))|0;
    $98 = ((($dst$016)) + 4|0);
    $99 = ($31|0)==(0);
    if ($99) {
     break;
    } else {
     $$in = $31;$dst$016 = $98;$posx$014 = $97;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$5>>2]|0;
   $104 = $$pre28;
  }
  $100 = HEAP32[$17>>2]|0;
  $101 = (($102) + ($100)|0);
  $103 = (($104) + -1)|0;
  HEAP32[$5>>2] = $103;
  $105 = ($104|0)==(0);
  if ($105) {
   $$lcssa = $101;
   break;
  }
  $106 = (($posy$1$lcssa) + ($7))|0;
  $$pre27 = HEAP32[$11>>2]|0;
  $102 = $101;$107 = $103;$108 = $101;$30 = $$pre27;$posy$020 = $106;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ABGR8888_ARGB8888_Modulate($info) {
 $info = $info|0;
 var $$lcssa = 0, $$lcssa24 = 0, $$pre = 0, $$pre14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $A$0$us = 0, $dst$02 = 0, $dst$02$us = 0, $dst$02$us6 = 0, $src$01 = 0, $src$01$us = 0, $src$01$us7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + -1)|0;
 HEAP32[$10>>2] = $12;
 $13 = ($11|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $$pre = HEAP32[$info>>2]|0;
 $$pre14 = HEAP32[$14>>2]|0;
 $101 = $$pre14;$102 = $$pre;$94 = $$pre;$97 = $$pre14;
 while(1) {
  $22 = HEAP32[$15>>2]|0;
  $23 = (($22) + -1)|0;
  $24 = ($22|0)==(0);
  if (!($24)) {
   if ($19) {
    $41 = $23;$dst$02$us = $101;$src$01$us = $102;
    while(1) {
     $25 = HEAP32[$src$01$us>>2]|0;
     $26 = $25 >>> 16;
     $27 = $26 & 255;
     $28 = $25 & 65280;
     $29 = $25 >>> 24;
     if ($21) {
      $A$0$us = $29;
     } else {
      $30 = Math_imul($29, $9)|0;
      $31 = (($30>>>0) / 255)&-1;
      $A$0$us = $31;
     }
     $32 = $A$0$us << 24;
     $33 = $25 << 16;
     $34 = $33 & 16711680;
     $35 = $28 | $34;
     $36 = $35 | $27;
     $37 = $36 | $32;
     HEAP32[$dst$02$us>>2] = $37;
     $38 = ((($src$01$us)) + 4|0);
     $39 = ((($dst$02$us)) + 4|0);
     $40 = (($41) + -1)|0;
     $42 = ($41|0)==(0);
     if ($42) {
      break;
     } else {
      $41 = $40;$dst$02$us = $39;$src$01$us = $38;
     }
    }
   } else {
    if ($21) {
     $64 = $23;$dst$02$us6 = $101;$src$01$us7 = $102;
     while(1) {
      $43 = HEAP32[$src$01$us7>>2]|0;
      $44 = $43 >>> 16;
      $45 = $44 & 255;
      $46 = $43 >>> 8;
      $47 = $46 & 255;
      $48 = $43 & 255;
      $49 = $43 & -16777216;
      $50 = Math_imul($48, $4)|0;
      $51 = (($50>>>0) / 255)&-1;
      $52 = Math_imul($47, $6)|0;
      $53 = (($52>>>0) / 255)&-1;
      $54 = Math_imul($45, $8)|0;
      $55 = (($54>>>0) / 255)&-1;
      $56 = $51 << 16;
      $57 = $53 << 8;
      $58 = $57 | $56;
      $59 = $58 | $55;
      $60 = $59 | $49;
      HEAP32[$dst$02$us6>>2] = $60;
      $61 = ((($src$01$us7)) + 4|0);
      $62 = ((($dst$02$us6)) + 4|0);
      $63 = (($64) + -1)|0;
      $65 = ($64|0)==(0);
      if ($65) {
       break;
      } else {
       $64 = $63;$dst$02$us6 = $62;$src$01$us7 = $61;
      }
     }
    } else {
     $90 = $23;$dst$02 = $101;$src$01 = $102;
     while(1) {
      $66 = HEAP32[$src$01>>2]|0;
      $67 = $66 >>> 16;
      $68 = $67 & 255;
      $69 = $66 >>> 8;
      $70 = $69 & 255;
      $71 = $66 & 255;
      $72 = $66 >>> 24;
      $73 = Math_imul($71, $4)|0;
      $74 = (($73>>>0) / 255)&-1;
      $75 = Math_imul($70, $6)|0;
      $76 = (($75>>>0) / 255)&-1;
      $77 = Math_imul($68, $8)|0;
      $78 = (($77>>>0) / 255)&-1;
      $79 = Math_imul($72, $9)|0;
      $80 = (($79>>>0) / 255)&-1;
      $81 = $80 << 24;
      $82 = $74 << 16;
      $83 = $76 << 8;
      $84 = $83 | $82;
      $85 = $84 | $78;
      $86 = $85 | $81;
      HEAP32[$dst$02>>2] = $86;
      $87 = ((($src$01)) + 4|0);
      $88 = ((($dst$02)) + 4|0);
      $89 = (($90) + -1)|0;
      $91 = ($90|0)==(0);
      if ($91) {
       break;
      } else {
       $90 = $89;$dst$02 = $88;$src$01 = $87;
      }
     }
    }
   }
  }
  $92 = HEAP32[$16>>2]|0;
  $93 = (($94) + ($92)|0);
  $95 = HEAP32[$17>>2]|0;
  $96 = (($97) + ($95)|0);
  $98 = HEAP32[$10>>2]|0;
  $99 = (($98) + -1)|0;
  HEAP32[$10>>2] = $99;
  $100 = ($98|0)==(0);
  if ($100) {
   $$lcssa = $93;$$lcssa24 = $96;
   break;
  } else {
   $101 = $96;$102 = $93;$94 = $93;$97 = $96;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa24;
 return;
}
function _SDL_Blit_ABGR8888_ARGB8888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $A$0 = 0, $A$0$us = 0, $dst$013 = 0, $dst$013$us = 0, $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$017 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0, $src$1$us = 0, $srcx$010 = 0, $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0;
 var $srcy$016 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 << 16;
 $13 = ((($info)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12|0) / ($14|0))&-1;
 $16 = ((($info)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 << 16;
 $19 = ((($info)) + 24|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($18|0) / ($20|0))&-1;
 $22 = (($14) + -1)|0;
 HEAP32[$13>>2] = $22;
 $23 = ($14|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $113 = $$pre;$118 = $$pre;$39 = $20;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $31 = ($posy$017|0)>(65535);
  if ($31) {
   $32 = (($posy$017) + -65536)|0;
   $33 = $32 >>> 16;
   $34 = $33 << 16;
   $35 = (($srcy$016) + 1)|0;
   $36 = (($32) - ($34))|0;
   $37 = (($35) + ($33))|0;
   $posy$1$lcssa = $36;$srcy$1$lcssa = $37;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $38 = (($39) + -1)|0;
  $40 = ($39|0)==(0);
  if (!($40)) {
   if ($27) {
    $70 = $38;$dst$013$us = $118;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $41 = ($posx$011$us|0)>(65535);
     if ($41) {
      $42 = (($posx$011$us) + -65536)|0;
      $43 = $42 >>> 16;
      $44 = $43 << 16;
      $45 = (($srcx$010$us) + 1)|0;
      $46 = (($42) - ($44))|0;
      $47 = (($45) + ($43))|0;
      $48 = HEAP32[$info>>2]|0;
      $49 = HEAP32[$30>>2]|0;
      $50 = Math_imul($49, $srcy$1$lcssa)|0;
      $51 = (($48) + ($50)|0);
      $52 = $47 << 2;
      $53 = (($51) + ($52)|0);
      $posx$2$us = $46;$src$1$us = $53;$srcx$2$us = $47;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $54 = HEAP32[$src$1$us>>2]|0;
     $55 = $54 >>> 16;
     $56 = $55 & 255;
     $57 = $54 & 65280;
     $58 = $54 >>> 24;
     if ($29) {
      $A$0$us = $58;
     } else {
      $59 = Math_imul($58, $9)|0;
      $60 = (($59>>>0) / 255)&-1;
      $A$0$us = $60;
     }
     $61 = $A$0$us << 24;
     $62 = $54 << 16;
     $63 = $62 & 16711680;
     $64 = $56 | $57;
     $65 = $64 | $63;
     $66 = $65 | $61;
     HEAP32[$dst$013$us>>2] = $66;
     $67 = (($posx$2$us) + ($21))|0;
     $68 = ((($dst$013$us)) + 4|0);
     $69 = (($70) + -1)|0;
     $71 = ($70|0)==(0);
     if ($71) {
      break;
     } else {
      $70 = $69;$dst$013$us = $68;$posx$011$us = $67;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $109 = $38;$dst$013 = $118;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $72 = ($posx$011|0)>(65535);
     if ($72) {
      $73 = (($posx$011) + -65536)|0;
      $74 = $73 >>> 16;
      $75 = $74 << 16;
      $76 = (($srcx$010) + 1)|0;
      $77 = (($73) - ($75))|0;
      $78 = (($76) + ($74))|0;
      $79 = HEAP32[$info>>2]|0;
      $80 = HEAP32[$30>>2]|0;
      $81 = Math_imul($80, $srcy$1$lcssa)|0;
      $82 = (($79) + ($81)|0);
      $83 = $78 << 2;
      $84 = (($82) + ($83)|0);
      $posx$2 = $77;$src$1 = $84;$srcx$2 = $78;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $85 = HEAP32[$src$1>>2]|0;
     $86 = $85 >>> 16;
     $87 = $86 & 255;
     $88 = $85 >>> 8;
     $89 = $88 & 255;
     $90 = $85 & 255;
     $91 = $85 >>> 24;
     $92 = Math_imul($90, $4)|0;
     $93 = (($92>>>0) / 255)&-1;
     $94 = Math_imul($89, $6)|0;
     $95 = (($94>>>0) / 255)&-1;
     $96 = Math_imul($87, $8)|0;
     $97 = (($96>>>0) / 255)&-1;
     if ($29) {
      $A$0 = $91;
     } else {
      $98 = Math_imul($91, $9)|0;
      $99 = (($98>>>0) / 255)&-1;
      $A$0 = $99;
     }
     $100 = $A$0 << 24;
     $101 = $93 << 16;
     $102 = $95 << 8;
     $103 = $97 | $102;
     $104 = $103 | $101;
     $105 = $104 | $100;
     HEAP32[$dst$013>>2] = $105;
     $106 = (($posx$2) + ($21))|0;
     $107 = ((($dst$013)) + 4|0);
     $108 = (($109) + -1)|0;
     $110 = ($109|0)==(0);
     if ($110) {
      break;
     } else {
      $109 = $108;$dst$013 = $107;$posx$011 = $106;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $111 = HEAP32[$25>>2]|0;
  $112 = (($113) + ($111)|0);
  $114 = HEAP32[$13>>2]|0;
  $115 = (($114) + -1)|0;
  HEAP32[$13>>2] = $115;
  $116 = ($114|0)==(0);
  if ($116) {
   $$lcssa = $112;
   break;
  }
  $117 = (($posy$1$lcssa) + ($15))|0;
  $$pre27 = HEAP32[$19>>2]|0;
  $113 = $112;$118 = $112;$39 = $$pre27;$posy$017 = $117;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_ABGR8888_ARGB8888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$lcssa14 = 0, $$pre = 0, $$pre10 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$05 = 0, $dstA$0 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $src$04 = 0, $srcA$0 = 0;
 var $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + -1)|0;
 HEAP32[$10>>2] = $12;
 $13 = ($11|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre9 = HEAP32[$14>>2]|0;
 $100 = $12;$101 = $$pre9;$102 = $$pre;$93 = $$pre;$96 = $$pre9;
 while(1) {
  $25 = HEAP32[$15>>2]|0;
  $26 = ($25|0)==(0);
  if ($26) {
   $98 = $100;
  } else {
   $$in = $25;$dst$05 = $101;$src$04 = $102;
   while(1) {
    $27 = (($$in) + -1)|0;
    $28 = HEAP32[$src$04>>2]|0;
    $29 = $28 >>> 16;
    $30 = $29 & 255;
    $31 = $28 >>> 8;
    $32 = $31 & 255;
    $33 = $28 & 255;
    $34 = $28 >>> 24;
    $35 = HEAP32[$dst$05>>2]|0;
    $36 = $35 >>> 16;
    $37 = $36 & 255;
    $38 = $35 >>> 8;
    $39 = $38 & 255;
    $40 = $35 & 255;
    $41 = $35 >>> 24;
    if ($19) {
     $srcB$0 = $30;$srcG$0 = $32;$srcR$0 = $33;
    } else {
     $42 = Math_imul($33, $4)|0;
     $43 = (($42>>>0) / 255)&-1;
     $44 = Math_imul($32, $6)|0;
     $45 = (($44>>>0) / 255)&-1;
     $46 = Math_imul($30, $8)|0;
     $47 = (($46>>>0) / 255)&-1;
     $srcB$0 = $47;$srcG$0 = $45;$srcR$0 = $43;
    }
    if ($21) {
     $srcA$0 = $34;
    } else {
     $48 = Math_imul($34, $9)|0;
     $49 = (($48>>>0) / 255)&-1;
     $srcA$0 = $49;
    }
    $50 = ($srcA$0>>>0)<(255);
    $or$cond = $23 & $50;
    if ($or$cond) {
     $51 = Math_imul($srcA$0, $srcR$0)|0;
     $52 = (($51>>>0) / 255)&-1;
     $53 = Math_imul($srcA$0, $srcG$0)|0;
     $54 = (($53>>>0) / 255)&-1;
     $55 = Math_imul($srcA$0, $srcB$0)|0;
     $56 = (($55>>>0) / 255)&-1;
     $srcB$1 = $56;$srcG$1 = $54;$srcR$1 = $52;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($24|0) {
    case 16:  {
     $57 = (255 - ($srcA$0))|0;
     $58 = Math_imul($57, $37)|0;
     $59 = (($58>>>0) / 255)&-1;
     $60 = (($srcR$1) + ($59))|0;
     $61 = Math_imul($57, $39)|0;
     $62 = (($61>>>0) / 255)&-1;
     $63 = (($srcG$1) + ($62))|0;
     $64 = Math_imul($57, $40)|0;
     $65 = (($64>>>0) / 255)&-1;
     $66 = (($srcB$1) + ($65))|0;
     $67 = Math_imul($57, $41)|0;
     $68 = (($67>>>0) / 255)&-1;
     $69 = (($68) + ($srcA$0))|0;
     $dstA$0 = $69;$dstB$0 = $66;$dstG$0 = $63;$dstR$0 = $60;
     break;
    }
    case 32:  {
     $70 = (($srcR$1) + ($37))|0;
     $71 = ($70>>>0)>(255);
     $$ = $71 ? 255 : $70;
     $72 = (($srcG$1) + ($39))|0;
     $73 = ($72>>>0)>(255);
     $$3 = $73 ? 255 : $72;
     $74 = (($srcB$1) + ($40))|0;
     $75 = ($74>>>0)>(255);
     $$1 = $75 ? 255 : $74;
     $dstA$0 = $41;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $76 = Math_imul($srcR$1, $37)|0;
     $77 = (($76>>>0) / 255)&-1;
     $78 = Math_imul($srcG$1, $39)|0;
     $79 = (($78>>>0) / 255)&-1;
     $80 = Math_imul($srcB$1, $40)|0;
     $81 = (($80>>>0) / 255)&-1;
     $dstA$0 = $41;$dstB$0 = $81;$dstG$0 = $79;$dstR$0 = $77;
     break;
    }
    default: {
     $dstA$0 = $41;$dstB$0 = $40;$dstG$0 = $39;$dstR$0 = $37;
    }
    }
    $82 = $dstA$0 << 24;
    $83 = $dstR$0 << 16;
    $84 = $dstG$0 << 8;
    $85 = $84 | $83;
    $86 = $85 | $dstB$0;
    $87 = $86 | $82;
    HEAP32[$dst$05>>2] = $87;
    $88 = ((($src$04)) + 4|0);
    $89 = ((($dst$05)) + 4|0);
    $90 = ($27|0)==(0);
    if ($90) {
     break;
    } else {
     $$in = $27;$dst$05 = $89;$src$04 = $88;
    }
   }
   $$pre10 = HEAP32[$10>>2]|0;
   $98 = $$pre10;
  }
  $91 = HEAP32[$16>>2]|0;
  $92 = (($93) + ($91)|0);
  $94 = HEAP32[$17>>2]|0;
  $95 = (($96) + ($94)|0);
  $97 = (($98) + -1)|0;
  HEAP32[$10>>2] = $97;
  $99 = ($98|0)==(0);
  if ($99) {
   $$lcssa = $92;$$lcssa14 = $95;
   break;
  } else {
   $100 = $97;$101 = $95;$102 = $92;$93 = $92;$96 = $95;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa14;
 return;
}
function _SDL_Blit_ABGR8888_ARGB8888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstA$0 = 0, $dstB$0 = 0;
 var $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0, $src$015 = 0, $src$1 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 << 16;
 $13 = ((($info)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12|0) / ($14|0))&-1;
 $16 = ((($info)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 << 16;
 $19 = ((($info)) + 24|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($18|0) / ($20|0))&-1;
 $22 = (($14) + -1)|0;
 HEAP32[$13>>2] = $22;
 $23 = ($14|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = $1 & 112;
 $33 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $122 = $$pre;$127 = $22;$128 = $$pre;$42 = $20;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $34 = ($posy$020|0)>(65535);
  if ($34) {
   $35 = (($posy$020) + -65536)|0;
   $36 = $35 >>> 16;
   $37 = $36 << 16;
   $38 = (($srcy$019) + 1)|0;
   $39 = (($35) - ($37))|0;
   $40 = (($38) + ($36))|0;
   $posy$1$lcssa = $39;$srcy$1$lcssa = $40;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $41 = ($42|0)==(0);
  if ($41) {
   $124 = $127;
  } else {
   $$in = $42;$dst$016 = $128;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $43 = (($$in) + -1)|0;
    $44 = ($posx$014|0)>(65535);
    if ($44) {
     $45 = (($posx$014) + -65536)|0;
     $46 = $45 >>> 16;
     $47 = $46 << 16;
     $48 = (($srcx$013) + 1)|0;
     $49 = (($45) - ($47))|0;
     $50 = (($48) + ($46))|0;
     $51 = HEAP32[$info>>2]|0;
     $52 = HEAP32[$33>>2]|0;
     $53 = Math_imul($52, $srcy$1$lcssa)|0;
     $54 = (($51) + ($53)|0);
     $55 = $50 << 2;
     $56 = (($54) + ($55)|0);
     $posx$2 = $49;$src$1 = $56;$srcx$2 = $50;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $57 = HEAP32[$src$1>>2]|0;
    $58 = $57 >>> 16;
    $59 = $58 & 255;
    $60 = $57 >>> 8;
    $61 = $60 & 255;
    $62 = $57 & 255;
    $63 = $57 >>> 24;
    $64 = HEAP32[$dst$016>>2]|0;
    $65 = $64 >>> 16;
    $66 = $65 & 255;
    $67 = $64 >>> 8;
    $68 = $67 & 255;
    $69 = $64 & 255;
    $70 = $64 >>> 24;
    if ($27) {
     $srcB$0 = $59;$srcG$0 = $61;$srcR$0 = $62;
    } else {
     $71 = Math_imul($62, $4)|0;
     $72 = (($71>>>0) / 255)&-1;
     $73 = Math_imul($61, $6)|0;
     $74 = (($73>>>0) / 255)&-1;
     $75 = Math_imul($59, $8)|0;
     $76 = (($75>>>0) / 255)&-1;
     $srcB$0 = $76;$srcG$0 = $74;$srcR$0 = $72;
    }
    if ($29) {
     $srcA$0 = $63;
    } else {
     $77 = Math_imul($63, $9)|0;
     $78 = (($77>>>0) / 255)&-1;
     $srcA$0 = $78;
    }
    $79 = ($srcA$0>>>0)<(255);
    $or$cond = $31 & $79;
    if ($or$cond) {
     $80 = Math_imul($srcA$0, $srcR$0)|0;
     $81 = (($80>>>0) / 255)&-1;
     $82 = Math_imul($srcA$0, $srcG$0)|0;
     $83 = (($82>>>0) / 255)&-1;
     $84 = Math_imul($srcA$0, $srcB$0)|0;
     $85 = (($84>>>0) / 255)&-1;
     $srcB$1 = $85;$srcG$1 = $83;$srcR$1 = $81;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($32|0) {
    case 16:  {
     $86 = (255 - ($srcA$0))|0;
     $87 = Math_imul($86, $66)|0;
     $88 = (($87>>>0) / 255)&-1;
     $89 = (($srcR$1) + ($88))|0;
     $90 = Math_imul($86, $68)|0;
     $91 = (($90>>>0) / 255)&-1;
     $92 = (($srcG$1) + ($91))|0;
     $93 = Math_imul($86, $69)|0;
     $94 = (($93>>>0) / 255)&-1;
     $95 = (($srcB$1) + ($94))|0;
     $96 = Math_imul($86, $70)|0;
     $97 = (($96>>>0) / 255)&-1;
     $98 = (($97) + ($srcA$0))|0;
     $dstA$0 = $98;$dstB$0 = $95;$dstG$0 = $92;$dstR$0 = $89;
     break;
    }
    case 32:  {
     $99 = (($srcR$1) + ($66))|0;
     $100 = ($99>>>0)>(255);
     $$ = $100 ? 255 : $99;
     $101 = (($srcG$1) + ($68))|0;
     $102 = ($101>>>0)>(255);
     $$3 = $102 ? 255 : $101;
     $103 = (($srcB$1) + ($69))|0;
     $104 = ($103>>>0)>(255);
     $$1 = $104 ? 255 : $103;
     $dstA$0 = $70;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $105 = Math_imul($srcR$1, $66)|0;
     $106 = (($105>>>0) / 255)&-1;
     $107 = Math_imul($srcG$1, $68)|0;
     $108 = (($107>>>0) / 255)&-1;
     $109 = Math_imul($srcB$1, $69)|0;
     $110 = (($109>>>0) / 255)&-1;
     $dstA$0 = $70;$dstB$0 = $110;$dstG$0 = $108;$dstR$0 = $106;
     break;
    }
    default: {
     $dstA$0 = $70;$dstB$0 = $69;$dstG$0 = $68;$dstR$0 = $66;
    }
    }
    $111 = $dstA$0 << 24;
    $112 = $dstR$0 << 16;
    $113 = $dstG$0 << 8;
    $114 = $113 | $112;
    $115 = $114 | $dstB$0;
    $116 = $115 | $111;
    HEAP32[$dst$016>>2] = $116;
    $117 = (($posx$2) + ($21))|0;
    $118 = ((($dst$016)) + 4|0);
    $119 = ($43|0)==(0);
    if ($119) {
     break;
    } else {
     $$in = $43;$dst$016 = $118;$posx$014 = $117;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$13>>2]|0;
   $124 = $$pre28;
  }
  $120 = HEAP32[$25>>2]|0;
  $121 = (($122) + ($120)|0);
  $123 = (($124) + -1)|0;
  HEAP32[$13>>2] = $123;
  $125 = ($124|0)==(0);
  if ($125) {
   $$lcssa = $121;
   break;
  }
  $126 = (($posy$1$lcssa) + ($15))|0;
  $$pre27 = HEAP32[$19>>2]|0;
  $122 = $121;$127 = $123;$128 = $121;$42 = $$pre27;$posy$020 = $126;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGRA8888_RGB888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0, $posx$010 = 0, $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$53 = $$pre;$58 = $12;$59 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $55 = $58;
  } else {
   $$in = $25;$dst$013 = $59;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    $41 = $40 >>> 24;
    $42 = $40 << 8;
    $43 = $42 & 16711680;
    $44 = $40 >>> 8;
    $45 = $44 & 65280;
    $46 = $45 | $41;
    $47 = $46 | $43;
    HEAP32[$dst$013>>2] = $47;
    $48 = (($posx$2) + ($11))|0;
    $49 = ((($dst$013)) + 4|0);
    $50 = ($26|0)==(0);
    if ($50) {
     break;
    } else {
     $$in = $26;$dst$013 = $49;$posx$010 = $48;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $55 = $$pre25;
  }
  $51 = HEAP32[$15>>2]|0;
  $52 = (($53) + ($51)|0);
  $54 = (($55) + -1)|0;
  HEAP32[$3>>2] = $54;
  $56 = ($55|0)==(0);
  if ($56) {
   $$lcssa = $52;
   break;
  }
  $57 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$53 = $52;$58 = $54;$59 = $52;$posy$016 = $57;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGRA8888_RGB888_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$$us = 0, $$1 = 0, $$1$us = 0, $$3 = 0, $$3$us = 0, $$lcssa = 0, $$lcssa17 = 0, $$pre = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$05 = 0, $dst$05$us = 0, $dstB$0 = 0, $dstB$0$us = 0, $dstG$0 = 0, $dstG$0$us = 0, $dstR$0 = 0, $dstR$0$us = 0;
 var $src$04 = 0, $src$04$us = 0, $srcB$0$us = 0, $srcG$0$us = 0, $srcR$0$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 48;
 $11 = ($10|0)==(0);
 $12 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre10 = HEAP32[$6>>2]|0;
 $113 = $$pre;$116 = $$pre10;$120 = $$pre10;$121 = $$pre;
 while(1) {
  $13 = HEAP32[$7>>2]|0;
  $14 = (($13) + -1)|0;
  $15 = ($13|0)==(0);
  L6: do {
   if (!($15)) {
    if ($11) {
     $109 = $14;$dst$05 = $120;$src$04 = $121;
     while(1) {
      $67 = HEAP32[$src$04>>2]|0;
      $68 = $67 >>> 24;
      $69 = $67 >>> 16;
      $70 = $69 & 255;
      $71 = $67 >>> 8;
      $72 = $71 & 255;
      $73 = HEAP32[$dst$05>>2]|0;
      $74 = $73 >>> 16;
      $75 = $74 & 255;
      $76 = $73 >>> 8;
      $77 = $76 & 255;
      $78 = $73 & 255;
      switch ($12|0) {
      case 16:  {
       $79 = $67 & 255;
       $80 = $79 ^ 255;
       $81 = Math_imul($75, $80)|0;
       $82 = (($81>>>0) / 255)&-1;
       $83 = (($72) + ($82))|0;
       $84 = Math_imul($77, $80)|0;
       $85 = (($84>>>0) / 255)&-1;
       $86 = (($70) + ($85))|0;
       $87 = Math_imul($80, $78)|0;
       $88 = (($87>>>0) / 255)&-1;
       $89 = (($68) + ($88))|0;
       $dstB$0 = $89;$dstG$0 = $86;$dstR$0 = $83;
       break;
      }
      case 32:  {
       $90 = (($72) + ($75))|0;
       $91 = ($90>>>0)>(255);
       $$ = $91 ? 255 : $90;
       $92 = (($70) + ($77))|0;
       $93 = ($92>>>0)>(255);
       $$3 = $93 ? 255 : $92;
       $94 = (($68) + ($78))|0;
       $95 = ($94>>>0)>(255);
       $$1 = $95 ? 255 : $94;
       $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
       break;
      }
      case 64:  {
       $96 = Math_imul($72, $75)|0;
       $97 = (($96>>>0) / 255)&-1;
       $98 = Math_imul($70, $77)|0;
       $99 = (($98>>>0) / 255)&-1;
       $100 = Math_imul($68, $78)|0;
       $101 = (($100>>>0) / 255)&-1;
       $dstB$0 = $101;$dstG$0 = $99;$dstR$0 = $97;
       break;
      }
      default: {
       $dstB$0 = $78;$dstG$0 = $77;$dstR$0 = $75;
      }
      }
      $102 = $dstR$0 << 16;
      $103 = $dstG$0 << 8;
      $104 = $103 | $102;
      $105 = $104 | $dstB$0;
      HEAP32[$dst$05>>2] = $105;
      $106 = ((($src$04)) + 4|0);
      $107 = ((($dst$05)) + 4|0);
      $108 = (($109) + -1)|0;
      $110 = ($109|0)==(0);
      if ($110) {
       break L6;
      } else {
       $109 = $108;$dst$05 = $107;$src$04 = $106;
      }
     }
    } else {
     $65 = $14;$dst$05$us = $120;$src$04$us = $121;
    }
    while(1) {
     $16 = HEAP32[$src$04$us>>2]|0;
     $17 = $16 >>> 24;
     $18 = $16 >>> 16;
     $19 = $18 & 255;
     $20 = $16 >>> 8;
     $21 = $20 & 255;
     $22 = $16 & 255;
     $23 = HEAP32[$dst$05$us>>2]|0;
     $24 = $23 >>> 16;
     $25 = $24 & 255;
     $26 = $23 >>> 8;
     $27 = $26 & 255;
     $28 = $23 & 255;
     $29 = ($22|0)==(255);
     if ($29) {
      $srcB$0$us = $17;$srcG$0$us = $19;$srcR$0$us = $21;
     } else {
      $30 = Math_imul($21, $22)|0;
      $31 = (($30>>>0) / 255)&-1;
      $32 = Math_imul($19, $22)|0;
      $33 = (($32>>>0) / 255)&-1;
      $34 = Math_imul($17, $22)|0;
      $35 = (($34>>>0) / 255)&-1;
      $srcB$0$us = $35;$srcG$0$us = $33;$srcR$0$us = $31;
     }
     switch ($12|0) {
     case 16:  {
      $48 = $22 ^ 255;
      $49 = Math_imul($25, $48)|0;
      $50 = (($49>>>0) / 255)&-1;
      $51 = (($srcR$0$us) + ($50))|0;
      $52 = Math_imul($27, $48)|0;
      $53 = (($52>>>0) / 255)&-1;
      $54 = (($srcG$0$us) + ($53))|0;
      $55 = Math_imul($48, $28)|0;
      $56 = (($55>>>0) / 255)&-1;
      $57 = (($srcB$0$us) + ($56))|0;
      $dstB$0$us = $57;$dstG$0$us = $54;$dstR$0$us = $51;
      break;
     }
     case 32:  {
      $42 = (($srcR$0$us) + ($25))|0;
      $43 = ($42>>>0)>(255);
      $$$us = $43 ? 255 : $42;
      $44 = (($srcG$0$us) + ($27))|0;
      $45 = ($44>>>0)>(255);
      $$3$us = $45 ? 255 : $44;
      $46 = (($srcB$0$us) + ($28))|0;
      $47 = ($46>>>0)>(255);
      $$1$us = $47 ? 255 : $46;
      $dstB$0$us = $$1$us;$dstG$0$us = $$3$us;$dstR$0$us = $$$us;
      break;
     }
     case 64:  {
      $36 = Math_imul($srcR$0$us, $25)|0;
      $37 = (($36>>>0) / 255)&-1;
      $38 = Math_imul($srcG$0$us, $27)|0;
      $39 = (($38>>>0) / 255)&-1;
      $40 = Math_imul($srcB$0$us, $28)|0;
      $41 = (($40>>>0) / 255)&-1;
      $dstB$0$us = $41;$dstG$0$us = $39;$dstR$0$us = $37;
      break;
     }
     default: {
      $dstB$0$us = $28;$dstG$0$us = $27;$dstR$0$us = $25;
     }
     }
     $58 = $dstR$0$us << 16;
     $59 = $dstG$0$us << 8;
     $60 = $59 | $58;
     $61 = $60 | $dstB$0$us;
     HEAP32[$dst$05$us>>2] = $61;
     $62 = ((($src$04$us)) + 4|0);
     $63 = ((($dst$05$us)) + 4|0);
     $64 = (($65) + -1)|0;
     $66 = ($65|0)==(0);
     if ($66) {
      break L6;
     } else {
      $65 = $64;$dst$05$us = $63;$src$04$us = $62;
     }
    }
   }
  } while(0);
  $111 = HEAP32[$8>>2]|0;
  $112 = (($113) + ($111)|0);
  $114 = HEAP32[$9>>2]|0;
  $115 = (($116) + ($114)|0);
  $117 = HEAP32[$2>>2]|0;
  $118 = (($117) + -1)|0;
  HEAP32[$2>>2] = $118;
  $119 = ($117|0)==(0);
  if ($119) {
   $$lcssa = $112;$$lcssa17 = $115;
   break;
  } else {
   $113 = $112;$116 = $115;$120 = $115;$121 = $112;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa17;
 return;
}
function _SDL_Blit_BGRA8888_RGB888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0;
 var $src$015 = 0, $src$1 = 0, $srcB$0 = 0, $srcG$0 = 0, $srcR$0 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 48;
 $19 = ($18|0)!=(0);
 $20 = $1 & 112;
 $21 = ((($info)) + 12|0);
 $$pre = HEAP32[$16>>2]|0;
 $101 = $14;$102 = $$pre;$30 = $12;$96 = $$pre;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $22 = ($posy$020|0)>(65535);
  if ($22) {
   $23 = (($posy$020) + -65536)|0;
   $24 = $23 >>> 16;
   $25 = $24 << 16;
   $26 = (($srcy$019) + 1)|0;
   $27 = (($23) - ($25))|0;
   $28 = (($26) + ($24))|0;
   $posy$1$lcssa = $27;$srcy$1$lcssa = $28;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $29 = ($30|0)==(0);
  if ($29) {
   $98 = $101;
  } else {
   $$in = $30;$dst$016 = $102;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $31 = (($$in) + -1)|0;
    $32 = ($posx$014|0)>(65535);
    if ($32) {
     $33 = (($posx$014) + -65536)|0;
     $34 = $33 >>> 16;
     $35 = $34 << 16;
     $36 = (($srcx$013) + 1)|0;
     $37 = (($33) - ($35))|0;
     $38 = (($36) + ($34))|0;
     $39 = HEAP32[$info>>2]|0;
     $40 = HEAP32[$21>>2]|0;
     $41 = Math_imul($40, $srcy$1$lcssa)|0;
     $42 = (($39) + ($41)|0);
     $43 = $38 << 2;
     $44 = (($42) + ($43)|0);
     $posx$2 = $37;$src$1 = $44;$srcx$2 = $38;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $45 = HEAP32[$src$1>>2]|0;
    $46 = $45 >>> 24;
    $47 = $45 >>> 16;
    $48 = $47 & 255;
    $49 = $45 >>> 8;
    $50 = $49 & 255;
    $51 = $45 & 255;
    $52 = HEAP32[$dst$016>>2]|0;
    $53 = $52 >>> 16;
    $54 = $53 & 255;
    $55 = $52 >>> 8;
    $56 = $55 & 255;
    $57 = $52 & 255;
    $58 = ($51|0)!=(255);
    $or$cond = $19 & $58;
    if ($or$cond) {
     $59 = Math_imul($50, $51)|0;
     $60 = (($59>>>0) / 255)&-1;
     $61 = Math_imul($48, $51)|0;
     $62 = (($61>>>0) / 255)&-1;
     $63 = Math_imul($46, $51)|0;
     $64 = (($63>>>0) / 255)&-1;
     $srcB$0 = $64;$srcG$0 = $62;$srcR$0 = $60;
    } else {
     $srcB$0 = $46;$srcG$0 = $48;$srcR$0 = $50;
    }
    switch ($20|0) {
    case 16:  {
     $65 = $51 ^ 255;
     $66 = Math_imul($54, $65)|0;
     $67 = (($66>>>0) / 255)&-1;
     $68 = (($srcR$0) + ($67))|0;
     $69 = Math_imul($56, $65)|0;
     $70 = (($69>>>0) / 255)&-1;
     $71 = (($srcG$0) + ($70))|0;
     $72 = Math_imul($65, $57)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = (($srcB$0) + ($73))|0;
     $dstB$0 = $74;$dstG$0 = $71;$dstR$0 = $68;
     break;
    }
    case 32:  {
     $75 = (($srcR$0) + ($54))|0;
     $76 = ($75>>>0)>(255);
     $$ = $76 ? 255 : $75;
     $77 = (($srcG$0) + ($56))|0;
     $78 = ($77>>>0)>(255);
     $$3 = $78 ? 255 : $77;
     $79 = (($srcB$0) + ($57))|0;
     $80 = ($79>>>0)>(255);
     $$1 = $80 ? 255 : $79;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $81 = Math_imul($srcR$0, $54)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = Math_imul($srcG$0, $56)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = Math_imul($srcB$0, $57)|0;
     $86 = (($85>>>0) / 255)&-1;
     $dstB$0 = $86;$dstG$0 = $84;$dstR$0 = $82;
     break;
    }
    default: {
     $dstB$0 = $57;$dstG$0 = $56;$dstR$0 = $54;
    }
    }
    $87 = $dstR$0 << 16;
    $88 = $dstG$0 << 8;
    $89 = $88 | $87;
    $90 = $89 | $dstB$0;
    HEAP32[$dst$016>>2] = $90;
    $91 = (($posx$2) + ($13))|0;
    $92 = ((($dst$016)) + 4|0);
    $93 = ($31|0)==(0);
    if ($93) {
     break;
    } else {
     $$in = $31;$dst$016 = $92;$posx$014 = $91;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$5>>2]|0;
   $98 = $$pre28;
  }
  $94 = HEAP32[$17>>2]|0;
  $95 = (($96) + ($94)|0);
  $97 = (($98) + -1)|0;
  HEAP32[$5>>2] = $97;
  $99 = ($98|0)==(0);
  if ($99) {
   $$lcssa = $95;
   break;
  }
  $100 = (($posy$1$lcssa) + ($7))|0;
  $$pre27 = HEAP32[$11>>2]|0;
  $101 = $97;$102 = $95;$30 = $$pre27;$96 = $95;$posy$020 = $100;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGRA8888_RGB888_Modulate($info) {
 $info = $info|0;
 var $$in = 0, $$in13 = 0, $$lcssa = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$pre16 = 0, $$pre17 = 0, $$pre18 = 0, $$pre19 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $dst$02 = 0, $dst$02$us$us = 0, $src$01 = 0, $src$01$us$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = ((($info)) + 20|0);
 $14 = ((($info)) + 24|0);
 $15 = ((($info)) + 12|0);
 $16 = ((($info)) + 32|0);
 $17 = $1 & 1;
 $18 = ($17|0)==(0);
 $$pre17 = HEAP32[$info>>2]|0;
 $$pre18 = HEAP32[$13>>2]|0;
 if ($18) {
  $23 = $$pre17;$26 = $$pre18;$73 = $11;$74 = $$pre18;$75 = $$pre17;
  while(1) {
   $19 = HEAP32[$14>>2]|0;
   $20 = ($19|0)==(0);
   if ($20) {
    $28 = $73;
   } else {
    $$in13 = $19;$dst$02$us$us = $74;$src$01$us$us = $75;
    while(1) {
     $30 = (($$in13) + -1)|0;
     $31 = HEAP32[$src$01$us$us>>2]|0;
     $32 = $31 >>> 24;
     $33 = $31 << 8;
     $34 = $33 & 16711680;
     $35 = $31 >>> 8;
     $36 = $35 & 65280;
     $37 = $36 | $34;
     $38 = $37 | $32;
     HEAP32[$dst$02$us$us>>2] = $38;
     $39 = ((($src$01$us$us)) + 4|0);
     $40 = ((($dst$02$us$us)) + 4|0);
     $41 = ($30|0)==(0);
     if ($41) {
      break;
     } else {
      $$in13 = $30;$dst$02$us$us = $40;$src$01$us$us = $39;
     }
    }
    $$pre19 = HEAP32[$9>>2]|0;
    $28 = $$pre19;
   }
   $21 = HEAP32[$15>>2]|0;
   $22 = (($23) + ($21)|0);
   $24 = HEAP32[$16>>2]|0;
   $25 = (($26) + ($24)|0);
   $27 = (($28) + -1)|0;
   HEAP32[$9>>2] = $27;
   $29 = ($28|0)==(0);
   if ($29) {
    $$lcssa = $22;$$lcssa30 = $25;
    break;
   } else {
    $23 = $22;$26 = $25;$73 = $27;$74 = $25;$75 = $22;
   }
  }
  HEAP32[$info>>2] = $$lcssa;
  HEAP32[$13>>2] = $$lcssa30;
  return;
 } else {
  $66 = $$pre17;$69 = $$pre18;$76 = $11;$77 = $$pre18;$78 = $$pre17;
  while(1) {
   $42 = HEAP32[$14>>2]|0;
   $43 = ($42|0)==(0);
   if ($43) {
    $71 = $76;
   } else {
    $$in = $42;$dst$02 = $77;$src$01 = $78;
    while(1) {
     $44 = (($$in) + -1)|0;
     $45 = HEAP32[$src$01>>2]|0;
     $46 = $45 >>> 24;
     $47 = $45 >>> 16;
     $48 = $47 & 255;
     $49 = $45 >>> 8;
     $50 = $49 & 255;
     $51 = Math_imul($50, $4)|0;
     $52 = (($51>>>0) / 255)&-1;
     $53 = Math_imul($48, $6)|0;
     $54 = (($53>>>0) / 255)&-1;
     $55 = Math_imul($46, $8)|0;
     $56 = (($55>>>0) / 255)&-1;
     $57 = $52 << 16;
     $58 = $54 << 8;
     $59 = $58 | $57;
     $60 = $59 | $56;
     HEAP32[$dst$02>>2] = $60;
     $61 = ((($src$01)) + 4|0);
     $62 = ((($dst$02)) + 4|0);
     $63 = ($44|0)==(0);
     if ($63) {
      break;
     } else {
      $$in = $44;$dst$02 = $62;$src$01 = $61;
     }
    }
    $$pre16 = HEAP32[$9>>2]|0;
    $71 = $$pre16;
   }
   $64 = HEAP32[$15>>2]|0;
   $65 = (($66) + ($64)|0);
   $67 = HEAP32[$16>>2]|0;
   $68 = (($69) + ($67)|0);
   $70 = (($71) + -1)|0;
   HEAP32[$9>>2] = $70;
   $72 = ($71|0)==(0);
   if ($72) {
    $$lcssa31 = $65;$$lcssa32 = $68;
    break;
   } else {
    $66 = $65;$69 = $68;$76 = $70;$77 = $68;$78 = $65;
   }
  }
  HEAP32[$info>>2] = $$lcssa31;
  HEAP32[$13>>2] = $$lcssa32;
  return;
 }
}
function _SDL_Blit_BGRA8888_RGB888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$013 = 0, $dst$013$us = 0, $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$017 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0;
 var $src$1$us = 0, $srcx$010 = 0, $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$016 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = ((($info)) + 20|0);
 $24 = ((($info)) + 32|0);
 $25 = $1 & 1;
 $26 = ($25|0)==(0);
 $27 = ((($info)) + 12|0);
 $$pre = HEAP32[$23>>2]|0;
 $100 = $$pre;$105 = $$pre;$36 = $19;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $28 = ($posy$017|0)>(65535);
  if ($28) {
   $29 = (($posy$017) + -65536)|0;
   $30 = $29 >>> 16;
   $31 = $30 << 16;
   $32 = (($srcy$016) + 1)|0;
   $33 = (($29) - ($31))|0;
   $34 = (($32) + ($30))|0;
   $posy$1$lcssa = $33;$srcy$1$lcssa = $34;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $35 = (($36) + -1)|0;
  $37 = ($36|0)==(0);
  if (!($37)) {
   if ($26) {
    $62 = $35;$dst$013$us = $105;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $38 = ($posx$011$us|0)>(65535);
     if ($38) {
      $39 = (($posx$011$us) + -65536)|0;
      $40 = $39 >>> 16;
      $41 = $40 << 16;
      $42 = (($srcx$010$us) + 1)|0;
      $43 = (($39) - ($41))|0;
      $44 = (($42) + ($40))|0;
      $45 = HEAP32[$info>>2]|0;
      $46 = HEAP32[$27>>2]|0;
      $47 = Math_imul($46, $srcy$1$lcssa)|0;
      $48 = (($45) + ($47)|0);
      $49 = $44 << 2;
      $50 = (($48) + ($49)|0);
      $posx$2$us = $43;$src$1$us = $50;$srcx$2$us = $44;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $51 = HEAP32[$src$1$us>>2]|0;
     $52 = $51 >>> 24;
     $53 = $51 << 8;
     $54 = $53 & 16711680;
     $55 = $51 >>> 8;
     $56 = $55 & 65280;
     $57 = $52 | $56;
     $58 = $57 | $54;
     HEAP32[$dst$013$us>>2] = $58;
     $59 = (($posx$2$us) + ($20))|0;
     $60 = ((($dst$013$us)) + 4|0);
     $61 = (($62) + -1)|0;
     $63 = ($62|0)==(0);
     if ($63) {
      break;
     } else {
      $62 = $61;$dst$013$us = $60;$posx$011$us = $59;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $96 = $35;$dst$013 = $105;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $64 = ($posx$011|0)>(65535);
     if ($64) {
      $65 = (($posx$011) + -65536)|0;
      $66 = $65 >>> 16;
      $67 = $66 << 16;
      $68 = (($srcx$010) + 1)|0;
      $69 = (($65) - ($67))|0;
      $70 = (($68) + ($66))|0;
      $71 = HEAP32[$info>>2]|0;
      $72 = HEAP32[$27>>2]|0;
      $73 = Math_imul($72, $srcy$1$lcssa)|0;
      $74 = (($71) + ($73)|0);
      $75 = $70 << 2;
      $76 = (($74) + ($75)|0);
      $posx$2 = $69;$src$1 = $76;$srcx$2 = $70;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $77 = HEAP32[$src$1>>2]|0;
     $78 = $77 >>> 24;
     $79 = $77 >>> 16;
     $80 = $79 & 255;
     $81 = $77 >>> 8;
     $82 = $81 & 255;
     $83 = Math_imul($82, $4)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = Math_imul($80, $6)|0;
     $86 = (($85>>>0) / 255)&-1;
     $87 = Math_imul($78, $8)|0;
     $88 = (($87>>>0) / 255)&-1;
     $89 = $84 << 16;
     $90 = $86 << 8;
     $91 = $88 | $90;
     $92 = $91 | $89;
     HEAP32[$dst$013>>2] = $92;
     $93 = (($posx$2) + ($20))|0;
     $94 = ((($dst$013)) + 4|0);
     $95 = (($96) + -1)|0;
     $97 = ($96|0)==(0);
     if ($97) {
      break;
     } else {
      $96 = $95;$dst$013 = $94;$posx$011 = $93;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $98 = HEAP32[$24>>2]|0;
  $99 = (($100) + ($98)|0);
  $101 = HEAP32[$12>>2]|0;
  $102 = (($101) + -1)|0;
  HEAP32[$12>>2] = $102;
  $103 = ($101|0)==(0);
  if ($103) {
   $$lcssa = $99;
   break;
  }
  $104 = (($posy$1$lcssa) + ($14))|0;
  $$pre27 = HEAP32[$18>>2]|0;
  $100 = $99;$105 = $99;$36 = $$pre27;$posy$017 = $104;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$23>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGRA8888_RGB888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$lcssa14 = 0, $$pre = 0, $$pre10 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $dst$05 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $src$04 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + -1)|0;
 HEAP32[$10>>2] = $12;
 $13 = ($11|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre9 = HEAP32[$14>>2]|0;
 $87 = $$pre;$90 = $$pre9;$94 = $12;$95 = $$pre9;$96 = $$pre;
 while(1) {
  $25 = HEAP32[$15>>2]|0;
  $26 = ($25|0)==(0);
  if ($26) {
   $92 = $94;
  } else {
   $$in = $25;$dst$05 = $95;$src$04 = $96;
   while(1) {
    $27 = (($$in) + -1)|0;
    $28 = HEAP32[$src$04>>2]|0;
    $29 = $28 >>> 24;
    $30 = $28 >>> 16;
    $31 = $30 & 255;
    $32 = $28 >>> 8;
    $33 = $32 & 255;
    $34 = $28 & 255;
    $35 = HEAP32[$dst$05>>2]|0;
    $36 = $35 >>> 16;
    $37 = $36 & 255;
    $38 = $35 >>> 8;
    $39 = $38 & 255;
    $40 = $35 & 255;
    if ($19) {
     $srcB$0 = $29;$srcG$0 = $31;$srcR$0 = $33;
    } else {
     $41 = Math_imul($33, $4)|0;
     $42 = (($41>>>0) / 255)&-1;
     $43 = Math_imul($31, $6)|0;
     $44 = (($43>>>0) / 255)&-1;
     $45 = Math_imul($29, $8)|0;
     $46 = (($45>>>0) / 255)&-1;
     $srcB$0 = $46;$srcG$0 = $44;$srcR$0 = $42;
    }
    if ($21) {
     $srcA$0 = $34;
    } else {
     $47 = Math_imul($34, $9)|0;
     $48 = (($47>>>0) / 255)&-1;
     $srcA$0 = $48;
    }
    $49 = ($srcA$0>>>0)<(255);
    $or$cond = $23 & $49;
    if ($or$cond) {
     $50 = Math_imul($srcA$0, $srcR$0)|0;
     $51 = (($50>>>0) / 255)&-1;
     $52 = Math_imul($srcA$0, $srcG$0)|0;
     $53 = (($52>>>0) / 255)&-1;
     $54 = Math_imul($srcA$0, $srcB$0)|0;
     $55 = (($54>>>0) / 255)&-1;
     $srcB$1 = $55;$srcG$1 = $53;$srcR$1 = $51;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($24|0) {
    case 16:  {
     $56 = (255 - ($srcA$0))|0;
     $57 = Math_imul($56, $37)|0;
     $58 = (($57>>>0) / 255)&-1;
     $59 = (($srcR$1) + ($58))|0;
     $60 = Math_imul($56, $39)|0;
     $61 = (($60>>>0) / 255)&-1;
     $62 = (($srcG$1) + ($61))|0;
     $63 = Math_imul($56, $40)|0;
     $64 = (($63>>>0) / 255)&-1;
     $65 = (($srcB$1) + ($64))|0;
     $dstB$0 = $65;$dstG$0 = $62;$dstR$0 = $59;
     break;
    }
    case 32:  {
     $66 = (($srcR$1) + ($37))|0;
     $67 = ($66>>>0)>(255);
     $$ = $67 ? 255 : $66;
     $68 = (($srcG$1) + ($39))|0;
     $69 = ($68>>>0)>(255);
     $$3 = $69 ? 255 : $68;
     $70 = (($srcB$1) + ($40))|0;
     $71 = ($70>>>0)>(255);
     $$1 = $71 ? 255 : $70;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $72 = Math_imul($srcR$1, $37)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($srcG$1, $39)|0;
     $75 = (($74>>>0) / 255)&-1;
     $76 = Math_imul($srcB$1, $40)|0;
     $77 = (($76>>>0) / 255)&-1;
     $dstB$0 = $77;$dstG$0 = $75;$dstR$0 = $73;
     break;
    }
    default: {
     $dstB$0 = $40;$dstG$0 = $39;$dstR$0 = $37;
    }
    }
    $78 = $dstR$0 << 16;
    $79 = $dstG$0 << 8;
    $80 = $79 | $78;
    $81 = $80 | $dstB$0;
    HEAP32[$dst$05>>2] = $81;
    $82 = ((($src$04)) + 4|0);
    $83 = ((($dst$05)) + 4|0);
    $84 = ($27|0)==(0);
    if ($84) {
     break;
    } else {
     $$in = $27;$dst$05 = $83;$src$04 = $82;
    }
   }
   $$pre10 = HEAP32[$10>>2]|0;
   $92 = $$pre10;
  }
  $85 = HEAP32[$16>>2]|0;
  $86 = (($87) + ($85)|0);
  $88 = HEAP32[$17>>2]|0;
  $89 = (($90) + ($88)|0);
  $91 = (($92) + -1)|0;
  HEAP32[$10>>2] = $91;
  $93 = ($92|0)==(0);
  if ($93) {
   $$lcssa = $86;$$lcssa14 = $89;
   break;
  } else {
   $87 = $86;$90 = $89;$94 = $91;$95 = $89;$96 = $86;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa14;
 return;
}
function _SDL_Blit_BGRA8888_RGB888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0;
 var $src$015 = 0, $src$1 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 << 16;
 $13 = ((($info)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12|0) / ($14|0))&-1;
 $16 = ((($info)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 << 16;
 $19 = ((($info)) + 24|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($18|0) / ($20|0))&-1;
 $22 = (($14) + -1)|0;
 HEAP32[$13>>2] = $22;
 $23 = ($14|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = $1 & 112;
 $33 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $116 = $$pre;$121 = $22;$122 = $$pre;$42 = $20;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $34 = ($posy$020|0)>(65535);
  if ($34) {
   $35 = (($posy$020) + -65536)|0;
   $36 = $35 >>> 16;
   $37 = $36 << 16;
   $38 = (($srcy$019) + 1)|0;
   $39 = (($35) - ($37))|0;
   $40 = (($38) + ($36))|0;
   $posy$1$lcssa = $39;$srcy$1$lcssa = $40;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $41 = ($42|0)==(0);
  if ($41) {
   $118 = $121;
  } else {
   $$in = $42;$dst$016 = $122;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $43 = (($$in) + -1)|0;
    $44 = ($posx$014|0)>(65535);
    if ($44) {
     $45 = (($posx$014) + -65536)|0;
     $46 = $45 >>> 16;
     $47 = $46 << 16;
     $48 = (($srcx$013) + 1)|0;
     $49 = (($45) - ($47))|0;
     $50 = (($48) + ($46))|0;
     $51 = HEAP32[$info>>2]|0;
     $52 = HEAP32[$33>>2]|0;
     $53 = Math_imul($52, $srcy$1$lcssa)|0;
     $54 = (($51) + ($53)|0);
     $55 = $50 << 2;
     $56 = (($54) + ($55)|0);
     $posx$2 = $49;$src$1 = $56;$srcx$2 = $50;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $57 = HEAP32[$src$1>>2]|0;
    $58 = $57 >>> 24;
    $59 = $57 >>> 16;
    $60 = $59 & 255;
    $61 = $57 >>> 8;
    $62 = $61 & 255;
    $63 = $57 & 255;
    $64 = HEAP32[$dst$016>>2]|0;
    $65 = $64 >>> 16;
    $66 = $65 & 255;
    $67 = $64 >>> 8;
    $68 = $67 & 255;
    $69 = $64 & 255;
    if ($27) {
     $srcB$0 = $58;$srcG$0 = $60;$srcR$0 = $62;
    } else {
     $70 = Math_imul($62, $4)|0;
     $71 = (($70>>>0) / 255)&-1;
     $72 = Math_imul($60, $6)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($58, $8)|0;
     $75 = (($74>>>0) / 255)&-1;
     $srcB$0 = $75;$srcG$0 = $73;$srcR$0 = $71;
    }
    if ($29) {
     $srcA$0 = $63;
    } else {
     $76 = Math_imul($63, $9)|0;
     $77 = (($76>>>0) / 255)&-1;
     $srcA$0 = $77;
    }
    $78 = ($srcA$0>>>0)<(255);
    $or$cond = $31 & $78;
    if ($or$cond) {
     $79 = Math_imul($srcA$0, $srcR$0)|0;
     $80 = (($79>>>0) / 255)&-1;
     $81 = Math_imul($srcA$0, $srcG$0)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = Math_imul($srcA$0, $srcB$0)|0;
     $84 = (($83>>>0) / 255)&-1;
     $srcB$1 = $84;$srcG$1 = $82;$srcR$1 = $80;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($32|0) {
    case 16:  {
     $85 = (255 - ($srcA$0))|0;
     $86 = Math_imul($85, $66)|0;
     $87 = (($86>>>0) / 255)&-1;
     $88 = (($srcR$1) + ($87))|0;
     $89 = Math_imul($85, $68)|0;
     $90 = (($89>>>0) / 255)&-1;
     $91 = (($srcG$1) + ($90))|0;
     $92 = Math_imul($85, $69)|0;
     $93 = (($92>>>0) / 255)&-1;
     $94 = (($srcB$1) + ($93))|0;
     $dstB$0 = $94;$dstG$0 = $91;$dstR$0 = $88;
     break;
    }
    case 32:  {
     $95 = (($srcR$1) + ($66))|0;
     $96 = ($95>>>0)>(255);
     $$ = $96 ? 255 : $95;
     $97 = (($srcG$1) + ($68))|0;
     $98 = ($97>>>0)>(255);
     $$3 = $98 ? 255 : $97;
     $99 = (($srcB$1) + ($69))|0;
     $100 = ($99>>>0)>(255);
     $$1 = $100 ? 255 : $99;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $101 = Math_imul($srcR$1, $66)|0;
     $102 = (($101>>>0) / 255)&-1;
     $103 = Math_imul($srcG$1, $68)|0;
     $104 = (($103>>>0) / 255)&-1;
     $105 = Math_imul($srcB$1, $69)|0;
     $106 = (($105>>>0) / 255)&-1;
     $dstB$0 = $106;$dstG$0 = $104;$dstR$0 = $102;
     break;
    }
    default: {
     $dstB$0 = $69;$dstG$0 = $68;$dstR$0 = $66;
    }
    }
    $107 = $dstR$0 << 16;
    $108 = $dstG$0 << 8;
    $109 = $108 | $107;
    $110 = $109 | $dstB$0;
    HEAP32[$dst$016>>2] = $110;
    $111 = (($posx$2) + ($21))|0;
    $112 = ((($dst$016)) + 4|0);
    $113 = ($43|0)==(0);
    if ($113) {
     break;
    } else {
     $$in = $43;$dst$016 = $112;$posx$014 = $111;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$13>>2]|0;
   $118 = $$pre28;
  }
  $114 = HEAP32[$25>>2]|0;
  $115 = (($116) + ($114)|0);
  $117 = (($118) + -1)|0;
  HEAP32[$13>>2] = $117;
  $119 = ($118|0)==(0);
  if ($119) {
   $$lcssa = $115;
   break;
  }
  $120 = (($posy$1$lcssa) + ($15))|0;
  $$pre27 = HEAP32[$19>>2]|0;
  $116 = $115;$121 = $117;$122 = $115;$42 = $$pre27;$posy$020 = $120;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGRA8888_BGR888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0, $posx$010 = 0, $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$53 = $$pre;$58 = $12;$59 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $55 = $58;
  } else {
   $$in = $25;$dst$013 = $59;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    $41 = $40 >>> 24;
    $42 = $40 >>> 8;
    $43 = $42 & 255;
    $44 = $41 << 16;
    $45 = $42 & 65280;
    $46 = $44 | $45;
    $47 = $46 | $43;
    HEAP32[$dst$013>>2] = $47;
    $48 = (($posx$2) + ($11))|0;
    $49 = ((($dst$013)) + 4|0);
    $50 = ($26|0)==(0);
    if ($50) {
     break;
    } else {
     $$in = $26;$dst$013 = $49;$posx$010 = $48;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $55 = $$pre25;
  }
  $51 = HEAP32[$15>>2]|0;
  $52 = (($53) + ($51)|0);
  $54 = (($55) + -1)|0;
  HEAP32[$3>>2] = $54;
  $56 = ($55|0)==(0);
  if ($56) {
   $$lcssa = $52;
   break;
  }
  $57 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$53 = $52;$58 = $54;$59 = $52;$posy$016 = $57;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGRA8888_BGR888_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$$us = 0, $$1 = 0, $$1$us = 0, $$3 = 0, $$3$us = 0, $$lcssa = 0, $$lcssa17 = 0, $$pre = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$05 = 0, $dst$05$us = 0, $dstB$0 = 0, $dstB$0$us = 0, $dstG$0 = 0, $dstG$0$us = 0, $dstR$0 = 0, $dstR$0$us = 0;
 var $src$04 = 0, $src$04$us = 0, $srcB$0$us = 0, $srcG$0$us = 0, $srcR$0$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 48;
 $11 = ($10|0)==(0);
 $12 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre10 = HEAP32[$6>>2]|0;
 $113 = $$pre;$116 = $$pre10;$120 = $$pre10;$121 = $$pre;
 while(1) {
  $13 = HEAP32[$7>>2]|0;
  $14 = (($13) + -1)|0;
  $15 = ($13|0)==(0);
  L6: do {
   if (!($15)) {
    if ($11) {
     $109 = $14;$dst$05 = $120;$src$04 = $121;
     while(1) {
      $67 = HEAP32[$src$04>>2]|0;
      $68 = $67 >>> 24;
      $69 = $67 >>> 16;
      $70 = $69 & 255;
      $71 = $67 >>> 8;
      $72 = $71 & 255;
      $73 = HEAP32[$dst$05>>2]|0;
      $74 = $73 >>> 16;
      $75 = $74 & 255;
      $76 = $73 >>> 8;
      $77 = $76 & 255;
      $78 = $73 & 255;
      switch ($12|0) {
      case 16:  {
       $79 = $67 & 255;
       $80 = $79 ^ 255;
       $81 = Math_imul($80, $78)|0;
       $82 = (($81>>>0) / 255)&-1;
       $83 = (($72) + ($82))|0;
       $84 = Math_imul($77, $80)|0;
       $85 = (($84>>>0) / 255)&-1;
       $86 = (($70) + ($85))|0;
       $87 = Math_imul($75, $80)|0;
       $88 = (($87>>>0) / 255)&-1;
       $89 = (($68) + ($88))|0;
       $dstB$0 = $89;$dstG$0 = $86;$dstR$0 = $83;
       break;
      }
      case 32:  {
       $90 = (($72) + ($78))|0;
       $91 = ($90>>>0)>(255);
       $$ = $91 ? 255 : $90;
       $92 = (($70) + ($77))|0;
       $93 = ($92>>>0)>(255);
       $$3 = $93 ? 255 : $92;
       $94 = (($68) + ($75))|0;
       $95 = ($94>>>0)>(255);
       $$1 = $95 ? 255 : $94;
       $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
       break;
      }
      case 64:  {
       $96 = Math_imul($72, $78)|0;
       $97 = (($96>>>0) / 255)&-1;
       $98 = Math_imul($70, $77)|0;
       $99 = (($98>>>0) / 255)&-1;
       $100 = Math_imul($68, $75)|0;
       $101 = (($100>>>0) / 255)&-1;
       $dstB$0 = $101;$dstG$0 = $99;$dstR$0 = $97;
       break;
      }
      default: {
       $dstB$0 = $75;$dstG$0 = $77;$dstR$0 = $78;
      }
      }
      $102 = $dstB$0 << 16;
      $103 = $dstG$0 << 8;
      $104 = $103 | $dstR$0;
      $105 = $104 | $102;
      HEAP32[$dst$05>>2] = $105;
      $106 = ((($src$04)) + 4|0);
      $107 = ((($dst$05)) + 4|0);
      $108 = (($109) + -1)|0;
      $110 = ($109|0)==(0);
      if ($110) {
       break L6;
      } else {
       $109 = $108;$dst$05 = $107;$src$04 = $106;
      }
     }
    } else {
     $65 = $14;$dst$05$us = $120;$src$04$us = $121;
    }
    while(1) {
     $16 = HEAP32[$src$04$us>>2]|0;
     $17 = $16 >>> 24;
     $18 = $16 >>> 16;
     $19 = $18 & 255;
     $20 = $16 >>> 8;
     $21 = $20 & 255;
     $22 = $16 & 255;
     $23 = HEAP32[$dst$05$us>>2]|0;
     $24 = $23 >>> 16;
     $25 = $24 & 255;
     $26 = $23 >>> 8;
     $27 = $26 & 255;
     $28 = $23 & 255;
     $29 = ($22|0)==(255);
     if ($29) {
      $srcB$0$us = $17;$srcG$0$us = $19;$srcR$0$us = $21;
     } else {
      $30 = Math_imul($21, $22)|0;
      $31 = (($30>>>0) / 255)&-1;
      $32 = Math_imul($19, $22)|0;
      $33 = (($32>>>0) / 255)&-1;
      $34 = Math_imul($17, $22)|0;
      $35 = (($34>>>0) / 255)&-1;
      $srcB$0$us = $35;$srcG$0$us = $33;$srcR$0$us = $31;
     }
     switch ($12|0) {
     case 16:  {
      $48 = $22 ^ 255;
      $49 = Math_imul($48, $28)|0;
      $50 = (($49>>>0) / 255)&-1;
      $51 = (($srcR$0$us) + ($50))|0;
      $52 = Math_imul($27, $48)|0;
      $53 = (($52>>>0) / 255)&-1;
      $54 = (($srcG$0$us) + ($53))|0;
      $55 = Math_imul($25, $48)|0;
      $56 = (($55>>>0) / 255)&-1;
      $57 = (($srcB$0$us) + ($56))|0;
      $dstB$0$us = $57;$dstG$0$us = $54;$dstR$0$us = $51;
      break;
     }
     case 32:  {
      $42 = (($srcR$0$us) + ($28))|0;
      $43 = ($42>>>0)>(255);
      $$$us = $43 ? 255 : $42;
      $44 = (($srcG$0$us) + ($27))|0;
      $45 = ($44>>>0)>(255);
      $$3$us = $45 ? 255 : $44;
      $46 = (($srcB$0$us) + ($25))|0;
      $47 = ($46>>>0)>(255);
      $$1$us = $47 ? 255 : $46;
      $dstB$0$us = $$1$us;$dstG$0$us = $$3$us;$dstR$0$us = $$$us;
      break;
     }
     case 64:  {
      $36 = Math_imul($srcR$0$us, $28)|0;
      $37 = (($36>>>0) / 255)&-1;
      $38 = Math_imul($srcG$0$us, $27)|0;
      $39 = (($38>>>0) / 255)&-1;
      $40 = Math_imul($srcB$0$us, $25)|0;
      $41 = (($40>>>0) / 255)&-1;
      $dstB$0$us = $41;$dstG$0$us = $39;$dstR$0$us = $37;
      break;
     }
     default: {
      $dstB$0$us = $25;$dstG$0$us = $27;$dstR$0$us = $28;
     }
     }
     $58 = $dstB$0$us << 16;
     $59 = $dstG$0$us << 8;
     $60 = $59 | $dstR$0$us;
     $61 = $60 | $58;
     HEAP32[$dst$05$us>>2] = $61;
     $62 = ((($src$04$us)) + 4|0);
     $63 = ((($dst$05$us)) + 4|0);
     $64 = (($65) + -1)|0;
     $66 = ($65|0)==(0);
     if ($66) {
      break L6;
     } else {
      $65 = $64;$dst$05$us = $63;$src$04$us = $62;
     }
    }
   }
  } while(0);
  $111 = HEAP32[$8>>2]|0;
  $112 = (($113) + ($111)|0);
  $114 = HEAP32[$9>>2]|0;
  $115 = (($116) + ($114)|0);
  $117 = HEAP32[$2>>2]|0;
  $118 = (($117) + -1)|0;
  HEAP32[$2>>2] = $118;
  $119 = ($117|0)==(0);
  if ($119) {
   $$lcssa = $112;$$lcssa17 = $115;
   break;
  } else {
   $113 = $112;$116 = $115;$120 = $115;$121 = $112;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa17;
 return;
}
function _SDL_Blit_BGRA8888_BGR888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0;
 var $src$015 = 0, $src$1 = 0, $srcB$0 = 0, $srcG$0 = 0, $srcR$0 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 48;
 $19 = ($18|0)!=(0);
 $20 = $1 & 112;
 $21 = ((($info)) + 12|0);
 $$pre = HEAP32[$16>>2]|0;
 $101 = $14;$102 = $$pre;$30 = $12;$96 = $$pre;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $22 = ($posy$020|0)>(65535);
  if ($22) {
   $23 = (($posy$020) + -65536)|0;
   $24 = $23 >>> 16;
   $25 = $24 << 16;
   $26 = (($srcy$019) + 1)|0;
   $27 = (($23) - ($25))|0;
   $28 = (($26) + ($24))|0;
   $posy$1$lcssa = $27;$srcy$1$lcssa = $28;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $29 = ($30|0)==(0);
  if ($29) {
   $98 = $101;
  } else {
   $$in = $30;$dst$016 = $102;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $31 = (($$in) + -1)|0;
    $32 = ($posx$014|0)>(65535);
    if ($32) {
     $33 = (($posx$014) + -65536)|0;
     $34 = $33 >>> 16;
     $35 = $34 << 16;
     $36 = (($srcx$013) + 1)|0;
     $37 = (($33) - ($35))|0;
     $38 = (($36) + ($34))|0;
     $39 = HEAP32[$info>>2]|0;
     $40 = HEAP32[$21>>2]|0;
     $41 = Math_imul($40, $srcy$1$lcssa)|0;
     $42 = (($39) + ($41)|0);
     $43 = $38 << 2;
     $44 = (($42) + ($43)|0);
     $posx$2 = $37;$src$1 = $44;$srcx$2 = $38;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $45 = HEAP32[$src$1>>2]|0;
    $46 = $45 >>> 24;
    $47 = $45 >>> 16;
    $48 = $47 & 255;
    $49 = $45 >>> 8;
    $50 = $49 & 255;
    $51 = $45 & 255;
    $52 = HEAP32[$dst$016>>2]|0;
    $53 = $52 >>> 16;
    $54 = $53 & 255;
    $55 = $52 >>> 8;
    $56 = $55 & 255;
    $57 = $52 & 255;
    $58 = ($51|0)!=(255);
    $or$cond = $19 & $58;
    if ($or$cond) {
     $59 = Math_imul($50, $51)|0;
     $60 = (($59>>>0) / 255)&-1;
     $61 = Math_imul($48, $51)|0;
     $62 = (($61>>>0) / 255)&-1;
     $63 = Math_imul($46, $51)|0;
     $64 = (($63>>>0) / 255)&-1;
     $srcB$0 = $64;$srcG$0 = $62;$srcR$0 = $60;
    } else {
     $srcB$0 = $46;$srcG$0 = $48;$srcR$0 = $50;
    }
    switch ($20|0) {
    case 16:  {
     $65 = $51 ^ 255;
     $66 = Math_imul($65, $57)|0;
     $67 = (($66>>>0) / 255)&-1;
     $68 = (($srcR$0) + ($67))|0;
     $69 = Math_imul($56, $65)|0;
     $70 = (($69>>>0) / 255)&-1;
     $71 = (($srcG$0) + ($70))|0;
     $72 = Math_imul($54, $65)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = (($srcB$0) + ($73))|0;
     $dstB$0 = $74;$dstG$0 = $71;$dstR$0 = $68;
     break;
    }
    case 32:  {
     $75 = (($srcR$0) + ($57))|0;
     $76 = ($75>>>0)>(255);
     $$ = $76 ? 255 : $75;
     $77 = (($srcG$0) + ($56))|0;
     $78 = ($77>>>0)>(255);
     $$3 = $78 ? 255 : $77;
     $79 = (($srcB$0) + ($54))|0;
     $80 = ($79>>>0)>(255);
     $$1 = $80 ? 255 : $79;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $81 = Math_imul($srcR$0, $57)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = Math_imul($srcG$0, $56)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = Math_imul($srcB$0, $54)|0;
     $86 = (($85>>>0) / 255)&-1;
     $dstB$0 = $86;$dstG$0 = $84;$dstR$0 = $82;
     break;
    }
    default: {
     $dstB$0 = $54;$dstG$0 = $56;$dstR$0 = $57;
    }
    }
    $87 = $dstB$0 << 16;
    $88 = $dstG$0 << 8;
    $89 = $88 | $dstR$0;
    $90 = $89 | $87;
    HEAP32[$dst$016>>2] = $90;
    $91 = (($posx$2) + ($13))|0;
    $92 = ((($dst$016)) + 4|0);
    $93 = ($31|0)==(0);
    if ($93) {
     break;
    } else {
     $$in = $31;$dst$016 = $92;$posx$014 = $91;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$5>>2]|0;
   $98 = $$pre28;
  }
  $94 = HEAP32[$17>>2]|0;
  $95 = (($96) + ($94)|0);
  $97 = (($98) + -1)|0;
  HEAP32[$5>>2] = $97;
  $99 = ($98|0)==(0);
  if ($99) {
   $$lcssa = $95;
   break;
  }
  $100 = (($posy$1$lcssa) + ($7))|0;
  $$pre27 = HEAP32[$11>>2]|0;
  $101 = $97;$102 = $95;$30 = $$pre27;$96 = $95;$posy$020 = $100;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGRA8888_BGR888_Modulate($info) {
 $info = $info|0;
 var $$in = 0, $$in13 = 0, $$lcssa = 0, $$lcssa30 = 0, $$lcssa31 = 0, $$lcssa32 = 0, $$pre16 = 0, $$pre17 = 0, $$pre18 = 0, $$pre19 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $8 = 0, $9 = 0, $dst$02 = 0, $dst$02$us$us = 0, $src$01 = 0, $src$01$us$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + -1)|0;
 HEAP32[$9>>2] = $11;
 $12 = ($10|0)==(0);
 if ($12) {
  return;
 }
 $13 = ((($info)) + 20|0);
 $14 = ((($info)) + 24|0);
 $15 = ((($info)) + 12|0);
 $16 = ((($info)) + 32|0);
 $17 = $1 & 1;
 $18 = ($17|0)==(0);
 $$pre17 = HEAP32[$info>>2]|0;
 $$pre18 = HEAP32[$13>>2]|0;
 if ($18) {
  $23 = $$pre17;$26 = $$pre18;$71 = $11;$72 = $$pre18;$73 = $$pre17;
  while(1) {
   $19 = HEAP32[$14>>2]|0;
   $20 = ($19|0)==(0);
   if ($20) {
    $28 = $71;
   } else {
    $$in13 = $19;$dst$02$us$us = $72;$src$01$us$us = $73;
    while(1) {
     $30 = (($$in13) + -1)|0;
     $31 = HEAP32[$src$01$us$us>>2]|0;
     $32 = $31 >>> 24;
     $33 = $31 >>> 8;
     $34 = $32 << 16;
     $35 = $33 & 65535;
     $36 = $35 | $34;
     HEAP32[$dst$02$us$us>>2] = $36;
     $37 = ((($src$01$us$us)) + 4|0);
     $38 = ((($dst$02$us$us)) + 4|0);
     $39 = ($30|0)==(0);
     if ($39) {
      break;
     } else {
      $$in13 = $30;$dst$02$us$us = $38;$src$01$us$us = $37;
     }
    }
    $$pre19 = HEAP32[$9>>2]|0;
    $28 = $$pre19;
   }
   $21 = HEAP32[$15>>2]|0;
   $22 = (($23) + ($21)|0);
   $24 = HEAP32[$16>>2]|0;
   $25 = (($26) + ($24)|0);
   $27 = (($28) + -1)|0;
   HEAP32[$9>>2] = $27;
   $29 = ($28|0)==(0);
   if ($29) {
    $$lcssa = $22;$$lcssa30 = $25;
    break;
   } else {
    $23 = $22;$26 = $25;$71 = $27;$72 = $25;$73 = $22;
   }
  }
  HEAP32[$info>>2] = $$lcssa;
  HEAP32[$13>>2] = $$lcssa30;
  return;
 } else {
  $64 = $$pre17;$67 = $$pre18;$74 = $11;$75 = $$pre18;$76 = $$pre17;
  while(1) {
   $40 = HEAP32[$14>>2]|0;
   $41 = ($40|0)==(0);
   if ($41) {
    $69 = $74;
   } else {
    $$in = $40;$dst$02 = $75;$src$01 = $76;
    while(1) {
     $42 = (($$in) + -1)|0;
     $43 = HEAP32[$src$01>>2]|0;
     $44 = $43 >>> 24;
     $45 = $43 >>> 16;
     $46 = $45 & 255;
     $47 = $43 >>> 8;
     $48 = $47 & 255;
     $49 = Math_imul($48, $4)|0;
     $50 = (($49>>>0) / 255)&-1;
     $51 = Math_imul($46, $6)|0;
     $52 = (($51>>>0) / 255)&-1;
     $53 = Math_imul($44, $8)|0;
     $54 = (($53>>>0) / 255)&-1;
     $55 = $54 << 16;
     $56 = $52 << 8;
     $57 = $56 | $50;
     $58 = $57 | $55;
     HEAP32[$dst$02>>2] = $58;
     $59 = ((($src$01)) + 4|0);
     $60 = ((($dst$02)) + 4|0);
     $61 = ($42|0)==(0);
     if ($61) {
      break;
     } else {
      $$in = $42;$dst$02 = $60;$src$01 = $59;
     }
    }
    $$pre16 = HEAP32[$9>>2]|0;
    $69 = $$pre16;
   }
   $62 = HEAP32[$15>>2]|0;
   $63 = (($64) + ($62)|0);
   $65 = HEAP32[$16>>2]|0;
   $66 = (($67) + ($65)|0);
   $68 = (($69) + -1)|0;
   HEAP32[$9>>2] = $68;
   $70 = ($69|0)==(0);
   if ($70) {
    $$lcssa31 = $63;$$lcssa32 = $66;
    break;
   } else {
    $64 = $63;$67 = $66;$74 = $68;$75 = $66;$76 = $63;
   }
  }
  HEAP32[$info>>2] = $$lcssa31;
  HEAP32[$13>>2] = $$lcssa32;
  return;
 }
}
function _SDL_Blit_BGRA8888_BGR888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$013 = 0, $dst$013$us = 0, $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$017 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0;
 var $src$1$us = 0, $srcx$010 = 0, $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0, $srcy$016 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = ((($info)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $10 << 16;
 $12 = ((($info)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (($11|0) / ($13|0))&-1;
 $15 = ((($info)) + 4|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = $16 << 16;
 $18 = ((($info)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($17|0) / ($19|0))&-1;
 $21 = (($13) + -1)|0;
 HEAP32[$12>>2] = $21;
 $22 = ($13|0)==(0);
 if ($22) {
  return;
 }
 $23 = ((($info)) + 20|0);
 $24 = ((($info)) + 32|0);
 $25 = $1 & 1;
 $26 = ($25|0)==(0);
 $27 = ((($info)) + 12|0);
 $$pre = HEAP32[$23>>2]|0;
 $100 = $$pre;$105 = $$pre;$36 = $19;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $28 = ($posy$017|0)>(65535);
  if ($28) {
   $29 = (($posy$017) + -65536)|0;
   $30 = $29 >>> 16;
   $31 = $30 << 16;
   $32 = (($srcy$016) + 1)|0;
   $33 = (($29) - ($31))|0;
   $34 = (($32) + ($30))|0;
   $posy$1$lcssa = $33;$srcy$1$lcssa = $34;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $35 = (($36) + -1)|0;
  $37 = ($36|0)==(0);
  if (!($37)) {
   if ($26) {
    $62 = $35;$dst$013$us = $105;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $38 = ($posx$011$us|0)>(65535);
     if ($38) {
      $39 = (($posx$011$us) + -65536)|0;
      $40 = $39 >>> 16;
      $41 = $40 << 16;
      $42 = (($srcx$010$us) + 1)|0;
      $43 = (($39) - ($41))|0;
      $44 = (($42) + ($40))|0;
      $45 = HEAP32[$info>>2]|0;
      $46 = HEAP32[$27>>2]|0;
      $47 = Math_imul($46, $srcy$1$lcssa)|0;
      $48 = (($45) + ($47)|0);
      $49 = $44 << 2;
      $50 = (($48) + ($49)|0);
      $posx$2$us = $43;$src$1$us = $50;$srcx$2$us = $44;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $51 = HEAP32[$src$1$us>>2]|0;
     $52 = $51 >>> 24;
     $53 = $51 >>> 8;
     $54 = $53 & 255;
     $55 = $52 << 16;
     $56 = $53 & 65280;
     $57 = $55 | $56;
     $58 = $57 | $54;
     HEAP32[$dst$013$us>>2] = $58;
     $59 = (($posx$2$us) + ($20))|0;
     $60 = ((($dst$013$us)) + 4|0);
     $61 = (($62) + -1)|0;
     $63 = ($62|0)==(0);
     if ($63) {
      break;
     } else {
      $62 = $61;$dst$013$us = $60;$posx$011$us = $59;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $96 = $35;$dst$013 = $105;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $64 = ($posx$011|0)>(65535);
     if ($64) {
      $65 = (($posx$011) + -65536)|0;
      $66 = $65 >>> 16;
      $67 = $66 << 16;
      $68 = (($srcx$010) + 1)|0;
      $69 = (($65) - ($67))|0;
      $70 = (($68) + ($66))|0;
      $71 = HEAP32[$info>>2]|0;
      $72 = HEAP32[$27>>2]|0;
      $73 = Math_imul($72, $srcy$1$lcssa)|0;
      $74 = (($71) + ($73)|0);
      $75 = $70 << 2;
      $76 = (($74) + ($75)|0);
      $posx$2 = $69;$src$1 = $76;$srcx$2 = $70;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $77 = HEAP32[$src$1>>2]|0;
     $78 = $77 >>> 24;
     $79 = $77 >>> 16;
     $80 = $79 & 255;
     $81 = $77 >>> 8;
     $82 = $81 & 255;
     $83 = Math_imul($82, $4)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = Math_imul($80, $6)|0;
     $86 = (($85>>>0) / 255)&-1;
     $87 = Math_imul($78, $8)|0;
     $88 = (($87>>>0) / 255)&-1;
     $89 = $88 << 16;
     $90 = $86 << 8;
     $91 = $89 | $90;
     $92 = $91 | $84;
     HEAP32[$dst$013>>2] = $92;
     $93 = (($posx$2) + ($20))|0;
     $94 = ((($dst$013)) + 4|0);
     $95 = (($96) + -1)|0;
     $97 = ($96|0)==(0);
     if ($97) {
      break;
     } else {
      $96 = $95;$dst$013 = $94;$posx$011 = $93;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $98 = HEAP32[$24>>2]|0;
  $99 = (($100) + ($98)|0);
  $101 = HEAP32[$12>>2]|0;
  $102 = (($101) + -1)|0;
  HEAP32[$12>>2] = $102;
  $103 = ($101|0)==(0);
  if ($103) {
   $$lcssa = $99;
   break;
  }
  $104 = (($posy$1$lcssa) + ($14))|0;
  $$pre27 = HEAP32[$18>>2]|0;
  $100 = $99;$105 = $99;$36 = $$pre27;$posy$017 = $104;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$23>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGRA8888_BGR888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$lcssa14 = 0, $$pre = 0, $$pre10 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $dst$05 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $src$04 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + -1)|0;
 HEAP32[$10>>2] = $12;
 $13 = ($11|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre9 = HEAP32[$14>>2]|0;
 $87 = $$pre;$90 = $$pre9;$94 = $12;$95 = $$pre9;$96 = $$pre;
 while(1) {
  $25 = HEAP32[$15>>2]|0;
  $26 = ($25|0)==(0);
  if ($26) {
   $92 = $94;
  } else {
   $$in = $25;$dst$05 = $95;$src$04 = $96;
   while(1) {
    $27 = (($$in) + -1)|0;
    $28 = HEAP32[$src$04>>2]|0;
    $29 = $28 >>> 24;
    $30 = $28 >>> 16;
    $31 = $30 & 255;
    $32 = $28 >>> 8;
    $33 = $32 & 255;
    $34 = $28 & 255;
    $35 = HEAP32[$dst$05>>2]|0;
    $36 = $35 >>> 16;
    $37 = $36 & 255;
    $38 = $35 >>> 8;
    $39 = $38 & 255;
    $40 = $35 & 255;
    if ($19) {
     $srcB$0 = $29;$srcG$0 = $31;$srcR$0 = $33;
    } else {
     $41 = Math_imul($33, $4)|0;
     $42 = (($41>>>0) / 255)&-1;
     $43 = Math_imul($31, $6)|0;
     $44 = (($43>>>0) / 255)&-1;
     $45 = Math_imul($29, $8)|0;
     $46 = (($45>>>0) / 255)&-1;
     $srcB$0 = $46;$srcG$0 = $44;$srcR$0 = $42;
    }
    if ($21) {
     $srcA$0 = $34;
    } else {
     $47 = Math_imul($34, $9)|0;
     $48 = (($47>>>0) / 255)&-1;
     $srcA$0 = $48;
    }
    $49 = ($srcA$0>>>0)<(255);
    $or$cond = $23 & $49;
    if ($or$cond) {
     $50 = Math_imul($srcA$0, $srcR$0)|0;
     $51 = (($50>>>0) / 255)&-1;
     $52 = Math_imul($srcA$0, $srcG$0)|0;
     $53 = (($52>>>0) / 255)&-1;
     $54 = Math_imul($srcA$0, $srcB$0)|0;
     $55 = (($54>>>0) / 255)&-1;
     $srcB$1 = $55;$srcG$1 = $53;$srcR$1 = $51;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($24|0) {
    case 16:  {
     $56 = (255 - ($srcA$0))|0;
     $57 = Math_imul($56, $40)|0;
     $58 = (($57>>>0) / 255)&-1;
     $59 = (($srcR$1) + ($58))|0;
     $60 = Math_imul($56, $39)|0;
     $61 = (($60>>>0) / 255)&-1;
     $62 = (($srcG$1) + ($61))|0;
     $63 = Math_imul($56, $37)|0;
     $64 = (($63>>>0) / 255)&-1;
     $65 = (($srcB$1) + ($64))|0;
     $dstB$0 = $65;$dstG$0 = $62;$dstR$0 = $59;
     break;
    }
    case 32:  {
     $66 = (($srcR$1) + ($40))|0;
     $67 = ($66>>>0)>(255);
     $$ = $67 ? 255 : $66;
     $68 = (($srcG$1) + ($39))|0;
     $69 = ($68>>>0)>(255);
     $$3 = $69 ? 255 : $68;
     $70 = (($srcB$1) + ($37))|0;
     $71 = ($70>>>0)>(255);
     $$1 = $71 ? 255 : $70;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $72 = Math_imul($srcR$1, $40)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($srcG$1, $39)|0;
     $75 = (($74>>>0) / 255)&-1;
     $76 = Math_imul($srcB$1, $37)|0;
     $77 = (($76>>>0) / 255)&-1;
     $dstB$0 = $77;$dstG$0 = $75;$dstR$0 = $73;
     break;
    }
    default: {
     $dstB$0 = $37;$dstG$0 = $39;$dstR$0 = $40;
    }
    }
    $78 = $dstB$0 << 16;
    $79 = $dstG$0 << 8;
    $80 = $79 | $dstR$0;
    $81 = $80 | $78;
    HEAP32[$dst$05>>2] = $81;
    $82 = ((($src$04)) + 4|0);
    $83 = ((($dst$05)) + 4|0);
    $84 = ($27|0)==(0);
    if ($84) {
     break;
    } else {
     $$in = $27;$dst$05 = $83;$src$04 = $82;
    }
   }
   $$pre10 = HEAP32[$10>>2]|0;
   $92 = $$pre10;
  }
  $85 = HEAP32[$16>>2]|0;
  $86 = (($87) + ($85)|0);
  $88 = HEAP32[$17>>2]|0;
  $89 = (($90) + ($88)|0);
  $91 = (($92) + -1)|0;
  HEAP32[$10>>2] = $91;
  $93 = ($92|0)==(0);
  if ($93) {
   $$lcssa = $86;$$lcssa14 = $89;
   break;
  } else {
   $87 = $86;$90 = $89;$94 = $91;$95 = $89;$96 = $86;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa14;
 return;
}
function _SDL_Blit_BGRA8888_BGR888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0;
 var $src$015 = 0, $src$1 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 << 16;
 $13 = ((($info)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12|0) / ($14|0))&-1;
 $16 = ((($info)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 << 16;
 $19 = ((($info)) + 24|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($18|0) / ($20|0))&-1;
 $22 = (($14) + -1)|0;
 HEAP32[$13>>2] = $22;
 $23 = ($14|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = $1 & 112;
 $33 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $116 = $$pre;$121 = $22;$122 = $$pre;$42 = $20;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $34 = ($posy$020|0)>(65535);
  if ($34) {
   $35 = (($posy$020) + -65536)|0;
   $36 = $35 >>> 16;
   $37 = $36 << 16;
   $38 = (($srcy$019) + 1)|0;
   $39 = (($35) - ($37))|0;
   $40 = (($38) + ($36))|0;
   $posy$1$lcssa = $39;$srcy$1$lcssa = $40;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $41 = ($42|0)==(0);
  if ($41) {
   $118 = $121;
  } else {
   $$in = $42;$dst$016 = $122;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $43 = (($$in) + -1)|0;
    $44 = ($posx$014|0)>(65535);
    if ($44) {
     $45 = (($posx$014) + -65536)|0;
     $46 = $45 >>> 16;
     $47 = $46 << 16;
     $48 = (($srcx$013) + 1)|0;
     $49 = (($45) - ($47))|0;
     $50 = (($48) + ($46))|0;
     $51 = HEAP32[$info>>2]|0;
     $52 = HEAP32[$33>>2]|0;
     $53 = Math_imul($52, $srcy$1$lcssa)|0;
     $54 = (($51) + ($53)|0);
     $55 = $50 << 2;
     $56 = (($54) + ($55)|0);
     $posx$2 = $49;$src$1 = $56;$srcx$2 = $50;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $57 = HEAP32[$src$1>>2]|0;
    $58 = $57 >>> 24;
    $59 = $57 >>> 16;
    $60 = $59 & 255;
    $61 = $57 >>> 8;
    $62 = $61 & 255;
    $63 = $57 & 255;
    $64 = HEAP32[$dst$016>>2]|0;
    $65 = $64 >>> 16;
    $66 = $65 & 255;
    $67 = $64 >>> 8;
    $68 = $67 & 255;
    $69 = $64 & 255;
    if ($27) {
     $srcB$0 = $58;$srcG$0 = $60;$srcR$0 = $62;
    } else {
     $70 = Math_imul($62, $4)|0;
     $71 = (($70>>>0) / 255)&-1;
     $72 = Math_imul($60, $6)|0;
     $73 = (($72>>>0) / 255)&-1;
     $74 = Math_imul($58, $8)|0;
     $75 = (($74>>>0) / 255)&-1;
     $srcB$0 = $75;$srcG$0 = $73;$srcR$0 = $71;
    }
    if ($29) {
     $srcA$0 = $63;
    } else {
     $76 = Math_imul($63, $9)|0;
     $77 = (($76>>>0) / 255)&-1;
     $srcA$0 = $77;
    }
    $78 = ($srcA$0>>>0)<(255);
    $or$cond = $31 & $78;
    if ($or$cond) {
     $79 = Math_imul($srcA$0, $srcR$0)|0;
     $80 = (($79>>>0) / 255)&-1;
     $81 = Math_imul($srcA$0, $srcG$0)|0;
     $82 = (($81>>>0) / 255)&-1;
     $83 = Math_imul($srcA$0, $srcB$0)|0;
     $84 = (($83>>>0) / 255)&-1;
     $srcB$1 = $84;$srcG$1 = $82;$srcR$1 = $80;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($32|0) {
    case 16:  {
     $85 = (255 - ($srcA$0))|0;
     $86 = Math_imul($85, $69)|0;
     $87 = (($86>>>0) / 255)&-1;
     $88 = (($srcR$1) + ($87))|0;
     $89 = Math_imul($85, $68)|0;
     $90 = (($89>>>0) / 255)&-1;
     $91 = (($srcG$1) + ($90))|0;
     $92 = Math_imul($85, $66)|0;
     $93 = (($92>>>0) / 255)&-1;
     $94 = (($srcB$1) + ($93))|0;
     $dstB$0 = $94;$dstG$0 = $91;$dstR$0 = $88;
     break;
    }
    case 32:  {
     $95 = (($srcR$1) + ($69))|0;
     $96 = ($95>>>0)>(255);
     $$ = $96 ? 255 : $95;
     $97 = (($srcG$1) + ($68))|0;
     $98 = ($97>>>0)>(255);
     $$3 = $98 ? 255 : $97;
     $99 = (($srcB$1) + ($66))|0;
     $100 = ($99>>>0)>(255);
     $$1 = $100 ? 255 : $99;
     $dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $101 = Math_imul($srcR$1, $69)|0;
     $102 = (($101>>>0) / 255)&-1;
     $103 = Math_imul($srcG$1, $68)|0;
     $104 = (($103>>>0) / 255)&-1;
     $105 = Math_imul($srcB$1, $66)|0;
     $106 = (($105>>>0) / 255)&-1;
     $dstB$0 = $106;$dstG$0 = $104;$dstR$0 = $102;
     break;
    }
    default: {
     $dstB$0 = $66;$dstG$0 = $68;$dstR$0 = $69;
    }
    }
    $107 = $dstB$0 << 16;
    $108 = $dstG$0 << 8;
    $109 = $108 | $dstR$0;
    $110 = $109 | $107;
    HEAP32[$dst$016>>2] = $110;
    $111 = (($posx$2) + ($21))|0;
    $112 = ((($dst$016)) + 4|0);
    $113 = ($43|0)==(0);
    if ($113) {
     break;
    } else {
     $$in = $43;$dst$016 = $112;$posx$014 = $111;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$13>>2]|0;
   $118 = $$pre28;
  }
  $114 = HEAP32[$25>>2]|0;
  $115 = (($116) + ($114)|0);
  $117 = (($118) + -1)|0;
  HEAP32[$13>>2] = $117;
  $119 = ($118|0)==(0);
  if ($119) {
   $$lcssa = $115;
   break;
  }
  $120 = (($posy$1$lcssa) + ($15))|0;
  $$pre27 = HEAP32[$19>>2]|0;
  $116 = $115;$121 = $117;$122 = $115;$42 = $$pre27;$posy$020 = $120;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGRA8888_ARGB8888_Scale($info) {
 $info = $info|0;
 var $$in = 0, $$lcssa = 0, $$pre = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$013 = 0;
 var $posx$010 = 0, $posx$2 = 0, $posy$016 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$1 = 0, $srcx$011 = 0, $srcx$2 = 0, $srcy$017 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 << 16;
 $3 = ((($info)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2|0) / ($4|0))&-1;
 $6 = ((($info)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 << 16;
 $9 = ((($info)) + 24|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($8|0) / ($10|0))&-1;
 $12 = (($4) + -1)|0;
 HEAP32[$3>>2] = $12;
 $13 = ($4|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 32|0);
 $16 = ((($info)) + 12|0);
 $$pre = HEAP32[$14>>2]|0;
 $25 = $10;$47 = $$pre;$52 = $12;$53 = $$pre;$posy$016 = 0;$srcy$017 = 0;
 while(1) {
  $17 = ($posy$016|0)>(65535);
  if ($17) {
   $18 = (($posy$016) + -65536)|0;
   $19 = $18 >>> 16;
   $20 = $19 << 16;
   $21 = (($srcy$017) + 1)|0;
   $22 = (($18) - ($20))|0;
   $23 = (($21) + ($19))|0;
   $posy$1$lcssa = $22;$srcy$1$lcssa = $23;
  } else {
   $posy$1$lcssa = $posy$016;$srcy$1$lcssa = $srcy$017;
  }
  $24 = ($25|0)==(0);
  if ($24) {
   $49 = $52;
  } else {
   $$in = $25;$dst$013 = $53;$posx$010 = 65536;$src$012 = 0;$srcx$011 = -1;
   while(1) {
    $26 = (($$in) + -1)|0;
    $27 = ($posx$010|0)>(65535);
    if ($27) {
     $28 = (($posx$010) + -65536)|0;
     $29 = $28 >>> 16;
     $30 = $29 << 16;
     $31 = (($srcx$011) + 1)|0;
     $32 = (($28) - ($30))|0;
     $33 = (($31) + ($29))|0;
     $34 = HEAP32[$info>>2]|0;
     $35 = HEAP32[$16>>2]|0;
     $36 = Math_imul($35, $srcy$1$lcssa)|0;
     $37 = (($34) + ($36)|0);
     $38 = $33 << 2;
     $39 = (($37) + ($38)|0);
     $posx$2 = $32;$src$1 = $39;$srcx$2 = $33;
    } else {
     $posx$2 = $posx$010;$src$1 = $src$012;$srcx$2 = $srcx$011;
    }
    $40 = HEAP32[$src$1>>2]|0;
    $41 = (_llvm_bswap_i32(($40|0))|0);
    HEAP32[$dst$013>>2] = $41;
    $42 = (($posx$2) + ($11))|0;
    $43 = ((($dst$013)) + 4|0);
    $44 = ($26|0)==(0);
    if ($44) {
     break;
    } else {
     $$in = $26;$dst$013 = $43;$posx$010 = $42;$src$012 = $src$1;$srcx$011 = $srcx$2;
    }
   }
   $$pre25 = HEAP32[$3>>2]|0;
   $49 = $$pre25;
  }
  $45 = HEAP32[$15>>2]|0;
  $46 = (($47) + ($45)|0);
  $48 = (($49) + -1)|0;
  HEAP32[$3>>2] = $48;
  $50 = ($49|0)==(0);
  if ($50) {
   $$lcssa = $46;
   break;
  }
  $51 = (($posy$1$lcssa) + ($5))|0;
  $$pre24 = HEAP32[$9>>2]|0;
  $25 = $$pre24;$47 = $46;$52 = $48;$53 = $46;$posy$016 = $51;$srcy$017 = $srcy$1$lcssa;
 }
 HEAP32[$14>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGRA8888_ARGB8888_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$$us = 0, $$1 = 0, $$1$us = 0, $$3 = 0, $$3$us = 0, $$lcssa = 0, $$lcssa17 = 0, $$pre = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$05 = 0, $dst$05$us = 0, $dstA$0 = 0, $dstA$0$us = 0, $dstB$0 = 0, $dstB$0$us = 0, $dstG$0 = 0, $dstG$0$us = 0, $dstR$0 = 0, $dstR$0$us = 0, $src$04 = 0, $src$04$us = 0, $srcB$0$us = 0, $srcG$0$us = 0, $srcR$0$us = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + -1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($3|0)==(0);
 if ($5) {
  return;
 }
 $6 = ((($info)) + 20|0);
 $7 = ((($info)) + 24|0);
 $8 = ((($info)) + 12|0);
 $9 = ((($info)) + 32|0);
 $10 = $1 & 48;
 $11 = ($10|0)==(0);
 $12 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre10 = HEAP32[$6>>2]|0;
 $125 = $$pre;$128 = $$pre10;$132 = $$pre10;$133 = $$pre;
 while(1) {
  $13 = HEAP32[$7>>2]|0;
  $14 = (($13) + -1)|0;
  $15 = ($13|0)==(0);
  L6: do {
   if (!($15)) {
    if ($11) {
     $121 = $14;$dst$05 = $132;$src$04 = $133;
     while(1) {
      $73 = HEAP32[$src$04>>2]|0;
      $74 = $73 >>> 24;
      $75 = $73 >>> 16;
      $76 = $75 & 255;
      $77 = $73 >>> 8;
      $78 = $77 & 255;
      $79 = $73 & 255;
      $80 = HEAP32[$dst$05>>2]|0;
      $81 = $80 >>> 16;
      $82 = $81 & 255;
      $83 = $80 >>> 8;
      $84 = $83 & 255;
      $85 = $80 & 255;
      $86 = $80 >>> 24;
      switch ($12|0) {
      case 16:  {
       $87 = $79 ^ 255;
       $88 = Math_imul($82, $87)|0;
       $89 = (($88>>>0) / 255)&-1;
       $90 = (($78) + ($89))|0;
       $91 = Math_imul($84, $87)|0;
       $92 = (($91>>>0) / 255)&-1;
       $93 = (($76) + ($92))|0;
       $94 = Math_imul($87, $85)|0;
       $95 = (($94>>>0) / 255)&-1;
       $96 = (($74) + ($95))|0;
       $97 = Math_imul($87, $86)|0;
       $98 = (($97>>>0) / 255)&-1;
       $99 = (($98) + ($79))|0;
       $dstA$0 = $99;$dstB$0 = $96;$dstG$0 = $93;$dstR$0 = $90;
       break;
      }
      case 32:  {
       $100 = (($78) + ($82))|0;
       $101 = ($100>>>0)>(255);
       $$ = $101 ? 255 : $100;
       $102 = (($76) + ($84))|0;
       $103 = ($102>>>0)>(255);
       $$3 = $103 ? 255 : $102;
       $104 = (($74) + ($85))|0;
       $105 = ($104>>>0)>(255);
       $$1 = $105 ? 255 : $104;
       $dstA$0 = $86;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
       break;
      }
      case 64:  {
       $106 = Math_imul($78, $82)|0;
       $107 = (($106>>>0) / 255)&-1;
       $108 = Math_imul($76, $84)|0;
       $109 = (($108>>>0) / 255)&-1;
       $110 = Math_imul($74, $85)|0;
       $111 = (($110>>>0) / 255)&-1;
       $dstA$0 = $86;$dstB$0 = $111;$dstG$0 = $109;$dstR$0 = $107;
       break;
      }
      default: {
       $dstA$0 = $86;$dstB$0 = $85;$dstG$0 = $84;$dstR$0 = $82;
      }
      }
      $112 = $dstA$0 << 24;
      $113 = $dstR$0 << 16;
      $114 = $dstG$0 << 8;
      $115 = $114 | $113;
      $116 = $115 | $dstB$0;
      $117 = $116 | $112;
      HEAP32[$dst$05>>2] = $117;
      $118 = ((($src$04)) + 4|0);
      $119 = ((($dst$05)) + 4|0);
      $120 = (($121) + -1)|0;
      $122 = ($121|0)==(0);
      if ($122) {
       break L6;
      } else {
       $121 = $120;$dst$05 = $119;$src$04 = $118;
      }
     }
    } else {
     $71 = $14;$dst$05$us = $132;$src$04$us = $133;
    }
    while(1) {
     $16 = HEAP32[$src$04$us>>2]|0;
     $17 = $16 >>> 24;
     $18 = $16 >>> 16;
     $19 = $18 & 255;
     $20 = $16 >>> 8;
     $21 = $20 & 255;
     $22 = $16 & 255;
     $23 = HEAP32[$dst$05$us>>2]|0;
     $24 = $23 >>> 16;
     $25 = $24 & 255;
     $26 = $23 >>> 8;
     $27 = $26 & 255;
     $28 = $23 & 255;
     $29 = $23 >>> 24;
     $30 = ($22|0)==(255);
     if ($30) {
      $srcB$0$us = $17;$srcG$0$us = $19;$srcR$0$us = $21;
     } else {
      $31 = Math_imul($21, $22)|0;
      $32 = (($31>>>0) / 255)&-1;
      $33 = Math_imul($19, $22)|0;
      $34 = (($33>>>0) / 255)&-1;
      $35 = Math_imul($17, $22)|0;
      $36 = (($35>>>0) / 255)&-1;
      $srcB$0$us = $36;$srcG$0$us = $34;$srcR$0$us = $32;
     }
     switch ($12|0) {
     case 16:  {
      $49 = $22 ^ 255;
      $50 = Math_imul($25, $49)|0;
      $51 = (($50>>>0) / 255)&-1;
      $52 = (($srcR$0$us) + ($51))|0;
      $53 = Math_imul($27, $49)|0;
      $54 = (($53>>>0) / 255)&-1;
      $55 = (($srcG$0$us) + ($54))|0;
      $56 = Math_imul($49, $28)|0;
      $57 = (($56>>>0) / 255)&-1;
      $58 = (($srcB$0$us) + ($57))|0;
      $59 = Math_imul($49, $29)|0;
      $60 = (($59>>>0) / 255)&-1;
      $61 = (($60) + ($22))|0;
      $dstA$0$us = $61;$dstB$0$us = $58;$dstG$0$us = $55;$dstR$0$us = $52;
      break;
     }
     case 32:  {
      $43 = (($srcR$0$us) + ($25))|0;
      $44 = ($43>>>0)>(255);
      $$$us = $44 ? 255 : $43;
      $45 = (($srcG$0$us) + ($27))|0;
      $46 = ($45>>>0)>(255);
      $$3$us = $46 ? 255 : $45;
      $47 = (($srcB$0$us) + ($28))|0;
      $48 = ($47>>>0)>(255);
      $$1$us = $48 ? 255 : $47;
      $dstA$0$us = $29;$dstB$0$us = $$1$us;$dstG$0$us = $$3$us;$dstR$0$us = $$$us;
      break;
     }
     case 64:  {
      $37 = Math_imul($srcR$0$us, $25)|0;
      $38 = (($37>>>0) / 255)&-1;
      $39 = Math_imul($srcG$0$us, $27)|0;
      $40 = (($39>>>0) / 255)&-1;
      $41 = Math_imul($srcB$0$us, $28)|0;
      $42 = (($41>>>0) / 255)&-1;
      $dstA$0$us = $29;$dstB$0$us = $42;$dstG$0$us = $40;$dstR$0$us = $38;
      break;
     }
     default: {
      $dstA$0$us = $29;$dstB$0$us = $28;$dstG$0$us = $27;$dstR$0$us = $25;
     }
     }
     $62 = $dstA$0$us << 24;
     $63 = $dstR$0$us << 16;
     $64 = $dstG$0$us << 8;
     $65 = $64 | $63;
     $66 = $65 | $dstB$0$us;
     $67 = $66 | $62;
     HEAP32[$dst$05$us>>2] = $67;
     $68 = ((($src$04$us)) + 4|0);
     $69 = ((($dst$05$us)) + 4|0);
     $70 = (($71) + -1)|0;
     $72 = ($71|0)==(0);
     if ($72) {
      break L6;
     } else {
      $71 = $70;$dst$05$us = $69;$src$04$us = $68;
     }
    }
   }
  } while(0);
  $123 = HEAP32[$8>>2]|0;
  $124 = (($125) + ($123)|0);
  $126 = HEAP32[$9>>2]|0;
  $127 = (($128) + ($126)|0);
  $129 = HEAP32[$2>>2]|0;
  $130 = (($129) + -1)|0;
  HEAP32[$2>>2] = $130;
  $131 = ($129|0)==(0);
  if ($131) {
   $$lcssa = $124;$$lcssa17 = $127;
   break;
  } else {
   $125 = $124;$128 = $127;$132 = $127;$133 = $124;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$6>>2] = $$lcssa17;
 return;
}
function _SDL_Blit_BGRA8888_ARGB8888_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstA$0 = 0, $dstB$0 = 0;
 var $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0, $src$015 = 0, $src$1 = 0, $srcB$0 = 0, $srcG$0 = 0, $srcR$0 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 << 16;
 $5 = ((($info)) + 28|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4|0) / ($6|0))&-1;
 $8 = ((($info)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 << 16;
 $11 = ((($info)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($10|0) / ($12|0))&-1;
 $14 = (($6) + -1)|0;
 HEAP32[$5>>2] = $14;
 $15 = ($6|0)==(0);
 if ($15) {
  return;
 }
 $16 = ((($info)) + 20|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 48;
 $19 = ($18|0)!=(0);
 $20 = $1 & 112;
 $21 = ((($info)) + 12|0);
 $$pre = HEAP32[$16>>2]|0;
 $102 = $$pre;$107 = $14;$108 = $$pre;$30 = $12;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $22 = ($posy$020|0)>(65535);
  if ($22) {
   $23 = (($posy$020) + -65536)|0;
   $24 = $23 >>> 16;
   $25 = $24 << 16;
   $26 = (($srcy$019) + 1)|0;
   $27 = (($23) - ($25))|0;
   $28 = (($26) + ($24))|0;
   $posy$1$lcssa = $27;$srcy$1$lcssa = $28;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $29 = ($30|0)==(0);
  if ($29) {
   $104 = $107;
  } else {
   $$in = $30;$dst$016 = $108;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $31 = (($$in) + -1)|0;
    $32 = ($posx$014|0)>(65535);
    if ($32) {
     $33 = (($posx$014) + -65536)|0;
     $34 = $33 >>> 16;
     $35 = $34 << 16;
     $36 = (($srcx$013) + 1)|0;
     $37 = (($33) - ($35))|0;
     $38 = (($36) + ($34))|0;
     $39 = HEAP32[$info>>2]|0;
     $40 = HEAP32[$21>>2]|0;
     $41 = Math_imul($40, $srcy$1$lcssa)|0;
     $42 = (($39) + ($41)|0);
     $43 = $38 << 2;
     $44 = (($42) + ($43)|0);
     $posx$2 = $37;$src$1 = $44;$srcx$2 = $38;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $45 = HEAP32[$src$1>>2]|0;
    $46 = $45 >>> 24;
    $47 = $45 >>> 16;
    $48 = $47 & 255;
    $49 = $45 >>> 8;
    $50 = $49 & 255;
    $51 = $45 & 255;
    $52 = HEAP32[$dst$016>>2]|0;
    $53 = $52 >>> 16;
    $54 = $53 & 255;
    $55 = $52 >>> 8;
    $56 = $55 & 255;
    $57 = $52 & 255;
    $58 = $52 >>> 24;
    $59 = ($51|0)!=(255);
    $or$cond = $19 & $59;
    if ($or$cond) {
     $60 = Math_imul($50, $51)|0;
     $61 = (($60>>>0) / 255)&-1;
     $62 = Math_imul($48, $51)|0;
     $63 = (($62>>>0) / 255)&-1;
     $64 = Math_imul($46, $51)|0;
     $65 = (($64>>>0) / 255)&-1;
     $srcB$0 = $65;$srcG$0 = $63;$srcR$0 = $61;
    } else {
     $srcB$0 = $46;$srcG$0 = $48;$srcR$0 = $50;
    }
    switch ($20|0) {
    case 16:  {
     $66 = $51 ^ 255;
     $67 = Math_imul($54, $66)|0;
     $68 = (($67>>>0) / 255)&-1;
     $69 = (($srcR$0) + ($68))|0;
     $70 = Math_imul($56, $66)|0;
     $71 = (($70>>>0) / 255)&-1;
     $72 = (($srcG$0) + ($71))|0;
     $73 = Math_imul($66, $57)|0;
     $74 = (($73>>>0) / 255)&-1;
     $75 = (($srcB$0) + ($74))|0;
     $76 = Math_imul($66, $58)|0;
     $77 = (($76>>>0) / 255)&-1;
     $78 = (($77) + ($51))|0;
     $dstA$0 = $78;$dstB$0 = $75;$dstG$0 = $72;$dstR$0 = $69;
     break;
    }
    case 32:  {
     $79 = (($srcR$0) + ($54))|0;
     $80 = ($79>>>0)>(255);
     $$ = $80 ? 255 : $79;
     $81 = (($srcG$0) + ($56))|0;
     $82 = ($81>>>0)>(255);
     $$3 = $82 ? 255 : $81;
     $83 = (($srcB$0) + ($57))|0;
     $84 = ($83>>>0)>(255);
     $$1 = $84 ? 255 : $83;
     $dstA$0 = $58;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $85 = Math_imul($srcR$0, $54)|0;
     $86 = (($85>>>0) / 255)&-1;
     $87 = Math_imul($srcG$0, $56)|0;
     $88 = (($87>>>0) / 255)&-1;
     $89 = Math_imul($srcB$0, $57)|0;
     $90 = (($89>>>0) / 255)&-1;
     $dstA$0 = $58;$dstB$0 = $90;$dstG$0 = $88;$dstR$0 = $86;
     break;
    }
    default: {
     $dstA$0 = $58;$dstB$0 = $57;$dstG$0 = $56;$dstR$0 = $54;
    }
    }
    $91 = $dstA$0 << 24;
    $92 = $dstR$0 << 16;
    $93 = $dstG$0 << 8;
    $94 = $93 | $92;
    $95 = $94 | $dstB$0;
    $96 = $95 | $91;
    HEAP32[$dst$016>>2] = $96;
    $97 = (($posx$2) + ($13))|0;
    $98 = ((($dst$016)) + 4|0);
    $99 = ($31|0)==(0);
    if ($99) {
     break;
    } else {
     $$in = $31;$dst$016 = $98;$posx$014 = $97;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$5>>2]|0;
   $104 = $$pre28;
  }
  $100 = HEAP32[$17>>2]|0;
  $101 = (($102) + ($100)|0);
  $103 = (($104) + -1)|0;
  HEAP32[$5>>2] = $103;
  $105 = ($104|0)==(0);
  if ($105) {
   $$lcssa = $101;
   break;
  }
  $106 = (($posy$1$lcssa) + ($7))|0;
  $$pre27 = HEAP32[$11>>2]|0;
  $102 = $101;$107 = $103;$108 = $101;$30 = $$pre27;$posy$020 = $106;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$16>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGRA8888_ARGB8888_Modulate($info) {
 $info = $info|0;
 var $$lcssa = 0, $$lcssa24 = 0, $$pre = 0, $$pre14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $A$0$us = 0, $dst$02 = 0, $dst$02$us = 0, $dst$02$us6 = 0, $src$01 = 0, $src$01$us = 0, $src$01$us7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + -1)|0;
 HEAP32[$10>>2] = $12;
 $13 = ($11|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $$pre = HEAP32[$info>>2]|0;
 $$pre14 = HEAP32[$14>>2]|0;
 $101 = $$pre14;$102 = $$pre;$94 = $$pre;$97 = $$pre14;
 while(1) {
  $22 = HEAP32[$15>>2]|0;
  $23 = (($22) + -1)|0;
  $24 = ($22|0)==(0);
  if (!($24)) {
   if ($19) {
    $41 = $23;$dst$02$us = $101;$src$01$us = $102;
    while(1) {
     $25 = HEAP32[$src$01$us>>2]|0;
     $26 = $25 >>> 24;
     $27 = $25 & 255;
     if ($21) {
      $A$0$us = $27;
     } else {
      $28 = Math_imul($27, $9)|0;
      $29 = (($28>>>0) / 255)&-1;
      $A$0$us = $29;
     }
     $30 = $A$0$us << 24;
     $31 = $25 << 8;
     $32 = $31 & 16711680;
     $33 = $25 >>> 8;
     $34 = $33 & 65280;
     $35 = $34 | $32;
     $36 = $35 | $26;
     $37 = $36 | $30;
     HEAP32[$dst$02$us>>2] = $37;
     $38 = ((($src$01$us)) + 4|0);
     $39 = ((($dst$02$us)) + 4|0);
     $40 = (($41) + -1)|0;
     $42 = ($41|0)==(0);
     if ($42) {
      break;
     } else {
      $41 = $40;$dst$02$us = $39;$src$01$us = $38;
     }
    }
   } else {
    if ($21) {
     $64 = $23;$dst$02$us6 = $101;$src$01$us7 = $102;
     while(1) {
      $43 = HEAP32[$src$01$us7>>2]|0;
      $44 = $43 >>> 24;
      $45 = $43 >>> 16;
      $46 = $45 & 255;
      $47 = $43 >>> 8;
      $48 = $47 & 255;
      $49 = Math_imul($48, $4)|0;
      $50 = (($49>>>0) / 255)&-1;
      $51 = Math_imul($46, $6)|0;
      $52 = (($51>>>0) / 255)&-1;
      $53 = Math_imul($44, $8)|0;
      $54 = (($53>>>0) / 255)&-1;
      $55 = $43 << 24;
      $56 = $50 << 16;
      $57 = $52 << 8;
      $58 = $57 | $56;
      $59 = $58 | $54;
      $60 = $59 | $55;
      HEAP32[$dst$02$us6>>2] = $60;
      $61 = ((($src$01$us7)) + 4|0);
      $62 = ((($dst$02$us6)) + 4|0);
      $63 = (($64) + -1)|0;
      $65 = ($64|0)==(0);
      if ($65) {
       break;
      } else {
       $64 = $63;$dst$02$us6 = $62;$src$01$us7 = $61;
      }
     }
    } else {
     $90 = $23;$dst$02 = $101;$src$01 = $102;
     while(1) {
      $66 = HEAP32[$src$01>>2]|0;
      $67 = $66 >>> 24;
      $68 = $66 >>> 16;
      $69 = $68 & 255;
      $70 = $66 >>> 8;
      $71 = $70 & 255;
      $72 = $66 & 255;
      $73 = Math_imul($71, $4)|0;
      $74 = (($73>>>0) / 255)&-1;
      $75 = Math_imul($69, $6)|0;
      $76 = (($75>>>0) / 255)&-1;
      $77 = Math_imul($67, $8)|0;
      $78 = (($77>>>0) / 255)&-1;
      $79 = Math_imul($72, $9)|0;
      $80 = (($79>>>0) / 255)&-1;
      $81 = $80 << 24;
      $82 = $74 << 16;
      $83 = $76 << 8;
      $84 = $83 | $82;
      $85 = $84 | $78;
      $86 = $85 | $81;
      HEAP32[$dst$02>>2] = $86;
      $87 = ((($src$01)) + 4|0);
      $88 = ((($dst$02)) + 4|0);
      $89 = (($90) + -1)|0;
      $91 = ($90|0)==(0);
      if ($91) {
       break;
      } else {
       $90 = $89;$dst$02 = $88;$src$01 = $87;
      }
     }
    }
   }
  }
  $92 = HEAP32[$16>>2]|0;
  $93 = (($94) + ($92)|0);
  $95 = HEAP32[$17>>2]|0;
  $96 = (($97) + ($95)|0);
  $98 = HEAP32[$10>>2]|0;
  $99 = (($98) + -1)|0;
  HEAP32[$10>>2] = $99;
  $100 = ($98|0)==(0);
  if ($100) {
   $$lcssa = $93;$$lcssa24 = $96;
   break;
  } else {
   $101 = $96;$102 = $93;$94 = $93;$97 = $96;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa24;
 return;
}
function _SDL_Blit_BGRA8888_ARGB8888_Modulate_Scale($info) {
 $info = $info|0;
 var $$lcssa = 0, $$pre = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $A$0 = 0, $A$0$us = 0, $dst$013 = 0, $dst$013$us = 0, $posx$011 = 0, $posx$011$us = 0, $posx$2 = 0, $posx$2$us = 0, $posy$017 = 0, $posy$1$lcssa = 0, $src$012 = 0, $src$012$us = 0, $src$1 = 0, $src$1$us = 0, $srcx$010 = 0, $srcx$010$us = 0, $srcx$2 = 0, $srcx$2$us = 0;
 var $srcy$016 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 << 16;
 $13 = ((($info)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12|0) / ($14|0))&-1;
 $16 = ((($info)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 << 16;
 $19 = ((($info)) + 24|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($18|0) / ($20|0))&-1;
 $22 = (($14) + -1)|0;
 HEAP32[$13>>2] = $22;
 $23 = ($14|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $113 = $$pre;$118 = $$pre;$39 = $20;$posy$017 = 0;$srcy$016 = 0;
 while(1) {
  $31 = ($posy$017|0)>(65535);
  if ($31) {
   $32 = (($posy$017) + -65536)|0;
   $33 = $32 >>> 16;
   $34 = $33 << 16;
   $35 = (($srcy$016) + 1)|0;
   $36 = (($32) - ($34))|0;
   $37 = (($35) + ($33))|0;
   $posy$1$lcssa = $36;$srcy$1$lcssa = $37;
  } else {
   $posy$1$lcssa = $posy$017;$srcy$1$lcssa = $srcy$016;
  }
  $38 = (($39) + -1)|0;
  $40 = ($39|0)==(0);
  if (!($40)) {
   if ($27) {
    $70 = $38;$dst$013$us = $118;$posx$011$us = 65536;$src$012$us = 0;$srcx$010$us = -1;
    while(1) {
     $41 = ($posx$011$us|0)>(65535);
     if ($41) {
      $42 = (($posx$011$us) + -65536)|0;
      $43 = $42 >>> 16;
      $44 = $43 << 16;
      $45 = (($srcx$010$us) + 1)|0;
      $46 = (($42) - ($44))|0;
      $47 = (($45) + ($43))|0;
      $48 = HEAP32[$info>>2]|0;
      $49 = HEAP32[$30>>2]|0;
      $50 = Math_imul($49, $srcy$1$lcssa)|0;
      $51 = (($48) + ($50)|0);
      $52 = $47 << 2;
      $53 = (($51) + ($52)|0);
      $posx$2$us = $46;$src$1$us = $53;$srcx$2$us = $47;
     } else {
      $posx$2$us = $posx$011$us;$src$1$us = $src$012$us;$srcx$2$us = $srcx$010$us;
     }
     $54 = HEAP32[$src$1$us>>2]|0;
     $55 = $54 >>> 24;
     $56 = $54 & 255;
     if ($29) {
      $A$0$us = $56;
     } else {
      $57 = Math_imul($56, $9)|0;
      $58 = (($57>>>0) / 255)&-1;
      $A$0$us = $58;
     }
     $59 = $A$0$us << 24;
     $60 = $54 << 8;
     $61 = $60 & 16711680;
     $62 = $54 >>> 8;
     $63 = $62 & 65280;
     $64 = $55 | $63;
     $65 = $64 | $61;
     $66 = $65 | $59;
     HEAP32[$dst$013$us>>2] = $66;
     $67 = (($posx$2$us) + ($21))|0;
     $68 = ((($dst$013$us)) + 4|0);
     $69 = (($70) + -1)|0;
     $71 = ($70|0)==(0);
     if ($71) {
      break;
     } else {
      $70 = $69;$dst$013$us = $68;$posx$011$us = $67;$src$012$us = $src$1$us;$srcx$010$us = $srcx$2$us;
     }
    }
   } else {
    $109 = $38;$dst$013 = $118;$posx$011 = 65536;$src$012 = 0;$srcx$010 = -1;
    while(1) {
     $72 = ($posx$011|0)>(65535);
     if ($72) {
      $73 = (($posx$011) + -65536)|0;
      $74 = $73 >>> 16;
      $75 = $74 << 16;
      $76 = (($srcx$010) + 1)|0;
      $77 = (($73) - ($75))|0;
      $78 = (($76) + ($74))|0;
      $79 = HEAP32[$info>>2]|0;
      $80 = HEAP32[$30>>2]|0;
      $81 = Math_imul($80, $srcy$1$lcssa)|0;
      $82 = (($79) + ($81)|0);
      $83 = $78 << 2;
      $84 = (($82) + ($83)|0);
      $posx$2 = $77;$src$1 = $84;$srcx$2 = $78;
     } else {
      $posx$2 = $posx$011;$src$1 = $src$012;$srcx$2 = $srcx$010;
     }
     $85 = HEAP32[$src$1>>2]|0;
     $86 = $85 >>> 24;
     $87 = $85 >>> 16;
     $88 = $87 & 255;
     $89 = $85 >>> 8;
     $90 = $89 & 255;
     $91 = $85 & 255;
     $92 = Math_imul($90, $4)|0;
     $93 = (($92>>>0) / 255)&-1;
     $94 = Math_imul($88, $6)|0;
     $95 = (($94>>>0) / 255)&-1;
     $96 = Math_imul($86, $8)|0;
     $97 = (($96>>>0) / 255)&-1;
     if ($29) {
      $A$0 = $91;
     } else {
      $98 = Math_imul($91, $9)|0;
      $99 = (($98>>>0) / 255)&-1;
      $A$0 = $99;
     }
     $100 = $A$0 << 24;
     $101 = $93 << 16;
     $102 = $95 << 8;
     $103 = $97 | $102;
     $104 = $103 | $101;
     $105 = $104 | $100;
     HEAP32[$dst$013>>2] = $105;
     $106 = (($posx$2) + ($21))|0;
     $107 = ((($dst$013)) + 4|0);
     $108 = (($109) + -1)|0;
     $110 = ($109|0)==(0);
     if ($110) {
      break;
     } else {
      $109 = $108;$dst$013 = $107;$posx$011 = $106;$src$012 = $src$1;$srcx$010 = $srcx$2;
     }
    }
   }
  }
  $111 = HEAP32[$25>>2]|0;
  $112 = (($113) + ($111)|0);
  $114 = HEAP32[$13>>2]|0;
  $115 = (($114) + -1)|0;
  HEAP32[$13>>2] = $115;
  $116 = ($114|0)==(0);
  if ($116) {
   $$lcssa = $112;
   break;
  }
  $117 = (($posy$1$lcssa) + ($15))|0;
  $$pre27 = HEAP32[$19>>2]|0;
  $113 = $112;$118 = $112;$39 = $$pre27;$posy$017 = $117;$srcy$016 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_Blit_BGRA8888_ARGB8888_Modulate_Blend($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$lcssa14 = 0, $$pre = 0, $$pre10 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$05 = 0, $dstA$0 = 0, $dstB$0 = 0, $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $src$04 = 0, $srcA$0 = 0;
 var $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + -1)|0;
 HEAP32[$10>>2] = $12;
 $13 = ($11|0)==(0);
 if ($13) {
  return;
 }
 $14 = ((($info)) + 20|0);
 $15 = ((($info)) + 24|0);
 $16 = ((($info)) + 12|0);
 $17 = ((($info)) + 32|0);
 $18 = $1 & 1;
 $19 = ($18|0)==(0);
 $20 = $1 & 2;
 $21 = ($20|0)==(0);
 $22 = $1 & 48;
 $23 = ($22|0)!=(0);
 $24 = $1 & 112;
 $$pre = HEAP32[$info>>2]|0;
 $$pre9 = HEAP32[$14>>2]|0;
 $100 = $12;$101 = $$pre9;$102 = $$pre;$93 = $$pre;$96 = $$pre9;
 while(1) {
  $25 = HEAP32[$15>>2]|0;
  $26 = ($25|0)==(0);
  if ($26) {
   $98 = $100;
  } else {
   $$in = $25;$dst$05 = $101;$src$04 = $102;
   while(1) {
    $27 = (($$in) + -1)|0;
    $28 = HEAP32[$src$04>>2]|0;
    $29 = $28 >>> 24;
    $30 = $28 >>> 16;
    $31 = $30 & 255;
    $32 = $28 >>> 8;
    $33 = $32 & 255;
    $34 = $28 & 255;
    $35 = HEAP32[$dst$05>>2]|0;
    $36 = $35 >>> 16;
    $37 = $36 & 255;
    $38 = $35 >>> 8;
    $39 = $38 & 255;
    $40 = $35 & 255;
    $41 = $35 >>> 24;
    if ($19) {
     $srcB$0 = $29;$srcG$0 = $31;$srcR$0 = $33;
    } else {
     $42 = Math_imul($33, $4)|0;
     $43 = (($42>>>0) / 255)&-1;
     $44 = Math_imul($31, $6)|0;
     $45 = (($44>>>0) / 255)&-1;
     $46 = Math_imul($29, $8)|0;
     $47 = (($46>>>0) / 255)&-1;
     $srcB$0 = $47;$srcG$0 = $45;$srcR$0 = $43;
    }
    if ($21) {
     $srcA$0 = $34;
    } else {
     $48 = Math_imul($34, $9)|0;
     $49 = (($48>>>0) / 255)&-1;
     $srcA$0 = $49;
    }
    $50 = ($srcA$0>>>0)<(255);
    $or$cond = $23 & $50;
    if ($or$cond) {
     $51 = Math_imul($srcA$0, $srcR$0)|0;
     $52 = (($51>>>0) / 255)&-1;
     $53 = Math_imul($srcA$0, $srcG$0)|0;
     $54 = (($53>>>0) / 255)&-1;
     $55 = Math_imul($srcA$0, $srcB$0)|0;
     $56 = (($55>>>0) / 255)&-1;
     $srcB$1 = $56;$srcG$1 = $54;$srcR$1 = $52;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($24|0) {
    case 16:  {
     $57 = (255 - ($srcA$0))|0;
     $58 = Math_imul($57, $37)|0;
     $59 = (($58>>>0) / 255)&-1;
     $60 = (($srcR$1) + ($59))|0;
     $61 = Math_imul($57, $39)|0;
     $62 = (($61>>>0) / 255)&-1;
     $63 = (($srcG$1) + ($62))|0;
     $64 = Math_imul($57, $40)|0;
     $65 = (($64>>>0) / 255)&-1;
     $66 = (($srcB$1) + ($65))|0;
     $67 = Math_imul($57, $41)|0;
     $68 = (($67>>>0) / 255)&-1;
     $69 = (($68) + ($srcA$0))|0;
     $dstA$0 = $69;$dstB$0 = $66;$dstG$0 = $63;$dstR$0 = $60;
     break;
    }
    case 32:  {
     $70 = (($srcR$1) + ($37))|0;
     $71 = ($70>>>0)>(255);
     $$ = $71 ? 255 : $70;
     $72 = (($srcG$1) + ($39))|0;
     $73 = ($72>>>0)>(255);
     $$3 = $73 ? 255 : $72;
     $74 = (($srcB$1) + ($40))|0;
     $75 = ($74>>>0)>(255);
     $$1 = $75 ? 255 : $74;
     $dstA$0 = $41;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $76 = Math_imul($srcR$1, $37)|0;
     $77 = (($76>>>0) / 255)&-1;
     $78 = Math_imul($srcG$1, $39)|0;
     $79 = (($78>>>0) / 255)&-1;
     $80 = Math_imul($srcB$1, $40)|0;
     $81 = (($80>>>0) / 255)&-1;
     $dstA$0 = $41;$dstB$0 = $81;$dstG$0 = $79;$dstR$0 = $77;
     break;
    }
    default: {
     $dstA$0 = $41;$dstB$0 = $40;$dstG$0 = $39;$dstR$0 = $37;
    }
    }
    $82 = $dstA$0 << 24;
    $83 = $dstR$0 << 16;
    $84 = $dstG$0 << 8;
    $85 = $84 | $83;
    $86 = $85 | $dstB$0;
    $87 = $86 | $82;
    HEAP32[$dst$05>>2] = $87;
    $88 = ((($src$04)) + 4|0);
    $89 = ((($dst$05)) + 4|0);
    $90 = ($27|0)==(0);
    if ($90) {
     break;
    } else {
     $$in = $27;$dst$05 = $89;$src$04 = $88;
    }
   }
   $$pre10 = HEAP32[$10>>2]|0;
   $98 = $$pre10;
  }
  $91 = HEAP32[$16>>2]|0;
  $92 = (($93) + ($91)|0);
  $94 = HEAP32[$17>>2]|0;
  $95 = (($96) + ($94)|0);
  $97 = (($98) + -1)|0;
  HEAP32[$10>>2] = $97;
  $99 = ($98|0)==(0);
  if ($99) {
   $$lcssa = $92;$$lcssa14 = $95;
   break;
  } else {
   $100 = $97;$101 = $95;$102 = $92;$93 = $92;$96 = $95;
  }
 }
 HEAP32[$info>>2] = $$lcssa;
 HEAP32[$14>>2] = $$lcssa14;
 return;
}
function _SDL_Blit_BGRA8888_ARGB8888_Modulate_Blend_Scale($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$lcssa = 0, $$pre = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$016 = 0, $dstA$0 = 0, $dstB$0 = 0;
 var $dstG$0 = 0, $dstR$0 = 0, $or$cond = 0, $posx$014 = 0, $posx$2 = 0, $posy$020 = 0, $posy$1$lcssa = 0, $src$015 = 0, $src$1 = 0, $srcA$0 = 0, $srcB$0 = 0, $srcB$1 = 0, $srcG$0 = 0, $srcG$1 = 0, $srcR$0 = 0, $srcR$1 = 0, $srcx$013 = 0, $srcx$2 = 0, $srcy$019 = 0, $srcy$1$lcssa = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $11 << 16;
 $13 = ((($info)) + 28|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12|0) / ($14|0))&-1;
 $16 = ((($info)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 << 16;
 $19 = ((($info)) + 24|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (($18|0) / ($20|0))&-1;
 $22 = (($14) + -1)|0;
 HEAP32[$13>>2] = $22;
 $23 = ($14|0)==(0);
 if ($23) {
  return;
 }
 $24 = ((($info)) + 20|0);
 $25 = ((($info)) + 32|0);
 $26 = $1 & 1;
 $27 = ($26|0)==(0);
 $28 = $1 & 2;
 $29 = ($28|0)==(0);
 $30 = $1 & 48;
 $31 = ($30|0)!=(0);
 $32 = $1 & 112;
 $33 = ((($info)) + 12|0);
 $$pre = HEAP32[$24>>2]|0;
 $122 = $$pre;$127 = $22;$128 = $$pre;$42 = $20;$posy$020 = 0;$srcy$019 = 0;
 while(1) {
  $34 = ($posy$020|0)>(65535);
  if ($34) {
   $35 = (($posy$020) + -65536)|0;
   $36 = $35 >>> 16;
   $37 = $36 << 16;
   $38 = (($srcy$019) + 1)|0;
   $39 = (($35) - ($37))|0;
   $40 = (($38) + ($36))|0;
   $posy$1$lcssa = $39;$srcy$1$lcssa = $40;
  } else {
   $posy$1$lcssa = $posy$020;$srcy$1$lcssa = $srcy$019;
  }
  $41 = ($42|0)==(0);
  if ($41) {
   $124 = $127;
  } else {
   $$in = $42;$dst$016 = $128;$posx$014 = 65536;$src$015 = 0;$srcx$013 = -1;
   while(1) {
    $43 = (($$in) + -1)|0;
    $44 = ($posx$014|0)>(65535);
    if ($44) {
     $45 = (($posx$014) + -65536)|0;
     $46 = $45 >>> 16;
     $47 = $46 << 16;
     $48 = (($srcx$013) + 1)|0;
     $49 = (($45) - ($47))|0;
     $50 = (($48) + ($46))|0;
     $51 = HEAP32[$info>>2]|0;
     $52 = HEAP32[$33>>2]|0;
     $53 = Math_imul($52, $srcy$1$lcssa)|0;
     $54 = (($51) + ($53)|0);
     $55 = $50 << 2;
     $56 = (($54) + ($55)|0);
     $posx$2 = $49;$src$1 = $56;$srcx$2 = $50;
    } else {
     $posx$2 = $posx$014;$src$1 = $src$015;$srcx$2 = $srcx$013;
    }
    $57 = HEAP32[$src$1>>2]|0;
    $58 = $57 >>> 24;
    $59 = $57 >>> 16;
    $60 = $59 & 255;
    $61 = $57 >>> 8;
    $62 = $61 & 255;
    $63 = $57 & 255;
    $64 = HEAP32[$dst$016>>2]|0;
    $65 = $64 >>> 16;
    $66 = $65 & 255;
    $67 = $64 >>> 8;
    $68 = $67 & 255;
    $69 = $64 & 255;
    $70 = $64 >>> 24;
    if ($27) {
     $srcB$0 = $58;$srcG$0 = $60;$srcR$0 = $62;
    } else {
     $71 = Math_imul($62, $4)|0;
     $72 = (($71>>>0) / 255)&-1;
     $73 = Math_imul($60, $6)|0;
     $74 = (($73>>>0) / 255)&-1;
     $75 = Math_imul($58, $8)|0;
     $76 = (($75>>>0) / 255)&-1;
     $srcB$0 = $76;$srcG$0 = $74;$srcR$0 = $72;
    }
    if ($29) {
     $srcA$0 = $63;
    } else {
     $77 = Math_imul($63, $9)|0;
     $78 = (($77>>>0) / 255)&-1;
     $srcA$0 = $78;
    }
    $79 = ($srcA$0>>>0)<(255);
    $or$cond = $31 & $79;
    if ($or$cond) {
     $80 = Math_imul($srcA$0, $srcR$0)|0;
     $81 = (($80>>>0) / 255)&-1;
     $82 = Math_imul($srcA$0, $srcG$0)|0;
     $83 = (($82>>>0) / 255)&-1;
     $84 = Math_imul($srcA$0, $srcB$0)|0;
     $85 = (($84>>>0) / 255)&-1;
     $srcB$1 = $85;$srcG$1 = $83;$srcR$1 = $81;
    } else {
     $srcB$1 = $srcB$0;$srcG$1 = $srcG$0;$srcR$1 = $srcR$0;
    }
    switch ($32|0) {
    case 16:  {
     $86 = (255 - ($srcA$0))|0;
     $87 = Math_imul($86, $66)|0;
     $88 = (($87>>>0) / 255)&-1;
     $89 = (($srcR$1) + ($88))|0;
     $90 = Math_imul($86, $68)|0;
     $91 = (($90>>>0) / 255)&-1;
     $92 = (($srcG$1) + ($91))|0;
     $93 = Math_imul($86, $69)|0;
     $94 = (($93>>>0) / 255)&-1;
     $95 = (($srcB$1) + ($94))|0;
     $96 = Math_imul($86, $70)|0;
     $97 = (($96>>>0) / 255)&-1;
     $98 = (($97) + ($srcA$0))|0;
     $dstA$0 = $98;$dstB$0 = $95;$dstG$0 = $92;$dstR$0 = $89;
     break;
    }
    case 32:  {
     $99 = (($srcR$1) + ($66))|0;
     $100 = ($99>>>0)>(255);
     $$ = $100 ? 255 : $99;
     $101 = (($srcG$1) + ($68))|0;
     $102 = ($101>>>0)>(255);
     $$3 = $102 ? 255 : $101;
     $103 = (($srcB$1) + ($69))|0;
     $104 = ($103>>>0)>(255);
     $$1 = $104 ? 255 : $103;
     $dstA$0 = $70;$dstB$0 = $$1;$dstG$0 = $$3;$dstR$0 = $$;
     break;
    }
    case 64:  {
     $105 = Math_imul($srcR$1, $66)|0;
     $106 = (($105>>>0) / 255)&-1;
     $107 = Math_imul($srcG$1, $68)|0;
     $108 = (($107>>>0) / 255)&-1;
     $109 = Math_imul($srcB$1, $69)|0;
     $110 = (($109>>>0) / 255)&-1;
     $dstA$0 = $70;$dstB$0 = $110;$dstG$0 = $108;$dstR$0 = $106;
     break;
    }
    default: {
     $dstA$0 = $70;$dstB$0 = $69;$dstG$0 = $68;$dstR$0 = $66;
    }
    }
    $111 = $dstA$0 << 24;
    $112 = $dstR$0 << 16;
    $113 = $dstG$0 << 8;
    $114 = $113 | $112;
    $115 = $114 | $dstB$0;
    $116 = $115 | $111;
    HEAP32[$dst$016>>2] = $116;
    $117 = (($posx$2) + ($21))|0;
    $118 = ((($dst$016)) + 4|0);
    $119 = ($43|0)==(0);
    if ($119) {
     break;
    } else {
     $$in = $43;$dst$016 = $118;$posx$014 = $117;$src$015 = $src$1;$srcx$013 = $srcx$2;
    }
   }
   $$pre28 = HEAP32[$13>>2]|0;
   $124 = $$pre28;
  }
  $120 = HEAP32[$25>>2]|0;
  $121 = (($122) + ($120)|0);
  $123 = (($124) + -1)|0;
  HEAP32[$13>>2] = $123;
  $125 = ($124|0)==(0);
  if ($125) {
   $$lcssa = $121;
   break;
  }
  $126 = (($posy$1$lcssa) + ($15))|0;
  $$pre27 = HEAP32[$19>>2]|0;
  $122 = $121;$127 = $123;$128 = $121;$42 = $$pre27;$posy$020 = $126;$srcy$019 = $srcy$1$lcssa;
 }
 HEAP32[$24>>2] = $$lcssa;
 return;
}
function _SDL_BlitCopy($info) {
 $info = $info|0;
 var $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$03 = 0, $dst$16 = 0, $h$04 = 0, $h$17 = 0, $src$05 = 0, $src$18 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 44|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 9|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = Math_imul($6, $1)|0;
 $8 = ((($info)) + 28|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = HEAP32[$info>>2]|0;
 $11 = ((($info)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($info)) + 12|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($info)) + 32|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ($10>>>0)<($12>>>0);
 $18 = Math_imul($14, $9)|0;
 $19 = (($10) + ($18)|0);
 $20 = ($12>>>0)<($19>>>0);
 $21 = Math_imul($16, $9)|0;
 $22 = (($12) + ($21)|0);
 $23 = ($10>>>0)<($22>>>0);
 $$sink = $17 ? $20 : $23;
 $24 = ($9|0)==(0);
 if ($$sink) {
  if ($24) {
   return;
  } else {
   $dst$03 = $12;$h$04 = $9;$src$05 = $10;
  }
  while(1) {
   $25 = (($h$04) + -1)|0;
   (_SDL_memmove($dst$03,$src$05,$7)|0);
   $26 = (($src$05) + ($14)|0);
   $27 = (($dst$03) + ($16)|0);
   $28 = ($25|0)==(0);
   if ($28) {
    break;
   } else {
    $dst$03 = $27;$h$04 = $25;$src$05 = $26;
   }
  }
  return;
 } else {
  if ($24) {
   return;
  } else {
   $dst$16 = $12;$h$17 = $9;$src$18 = $10;
  }
  while(1) {
   $29 = (($h$17) + -1)|0;
   (_SDL_memcpy($dst$16,$src$18,$7)|0);
   $30 = (($src$18) + ($14)|0);
   $31 = (($dst$16) + ($16)|0);
   $32 = ($29|0)==(0);
   if ($32) {
    break;
   } else {
    $dst$16 = $31;$h$17 = $29;$src$18 = $30;
   }
  }
  return;
 }
}
function _SDL_memmove($dst,$src,$len) {
 $dst = $dst|0;
 $src = $src|0;
 $len = $len|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _memmove(($dst|0),($src|0),($len|0))|0;
 return ($dst|0);
}
function _SDL_Blit_Slow($info) {
 $info = $info|0;
 var $$ = 0, $$1 = 0, $$3 = 0, $$in = 0, $$pre = 0, $$pre44 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0;
 var $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0;
 var $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0;
 var $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0;
 var $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0;
 var $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0;
 var $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0;
 var $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0;
 var $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0;
 var $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0;
 var $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0;
 var $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0;
 var $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0;
 var $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0;
 var $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0;
 var $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0;
 var $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0;
 var $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0;
 var $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0;
 var $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0;
 var $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0;
 var $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0;
 var $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0;
 var $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0;
 var $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0;
 var $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0;
 var $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0;
 var $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0;
 var $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0;
 var $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0;
 var $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0;
 var $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0;
 var $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0;
 var $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $div = 0, $div10 = 0, $div12 = 0, $div13 = 0, $div15 = 0, $div16 = 0, $div18 = 0, $div19 = 0, $div4 = 0, $div5 = 0, $div6 = 0, $div7 = 0, $div8 = 0;
 var $div9 = 0, $dst$033 = 0, $dstA$0 = 0, $dstA$0$shrunk = 0, $dstA$1 = 0, $dstB$1 = 0, $dstB$1$shrunk = 0, $dstB$2 = 0, $dstG$1 = 0, $dstG$1$shrunk = 0, $dstG$2 = 0, $dstR$1 = 0, $dstR$1$shrunk = 0, $dstR$2 = 0, $or$cond = 0, $posx$030 = 0, $posx$2 = 0, $posy$036 = 0, $posy$1$lcssa = 0, $src$032 = 0;
 var $src$1 = 0, $srcA$0 = 0, $srcA$0$shrunk = 0, $srcA$1 = 0, $srcB$1 = 0, $srcB$1$shrunk = 0, $srcB$2 = 0, $srcB$3 = 0, $srcG$1 = 0, $srcG$1$shrunk = 0, $srcG$2 = 0, $srcG$3 = 0, $srcR$1 = 0, $srcR$1$shrunk = 0, $srcR$2 = 0, $srcR$3 = 0, $srcpixel$1 = 0, $srcpixel$2 = 0, $srcx$031 = 0, $srcx$2 = 0;
 var $srcy$037 = 0, $srcy$1$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($info)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 255;
 $5 = $3 >>> 8;
 $6 = $5 & 255;
 $7 = $3 >>> 16;
 $8 = $7 & 255;
 $9 = $3 >>> 24;
 $10 = ((($info)) + 40|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($info)) + 44|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($11)) + 9|0);
 $15 = HEAP8[$14>>0]|0;
 $16 = $15&255;
 $17 = ((($13)) + 9|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = $18&255;
 $20 = ((($info)) + 8|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = $21 << 16;
 $23 = ((($info)) + 28|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = (($22|0) / ($24|0))&-1;
 $26 = ((($info)) + 4|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = $27 << 16;
 $29 = ((($info)) + 24|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = (($28|0) / ($30|0))&-1;
 $32 = (($24) + -1)|0;
 HEAP32[$23>>2] = $32;
 $33 = ($24|0)==(0);
 if ($33) {
  return;
 }
 $34 = ((($info)) + 20|0);
 $35 = ((($info)) + 32|0);
 $36 = ((($11)) + 24|0);
 $37 = ((($11)) + 12|0);
 $38 = ((($11)) + 32|0);
 $39 = ((($11)) + 28|0);
 $40 = ((($11)) + 16|0);
 $41 = ((($11)) + 20|0);
 $42 = $1 & 256;
 $43 = ($42|0)==(0);
 $44 = ((($13)) + 24|0);
 $45 = ((($13)) + 12|0);
 $46 = ((($13)) + 32|0);
 $47 = ((($13)) + 28|0);
 $48 = ((($13)) + 16|0);
 $49 = ((($13)) + 33|0);
 $50 = ((($13)) + 20|0);
 $51 = ((($13)) + 34|0);
 $52 = $1 & 1;
 $53 = ($52|0)==(0);
 $54 = $1 & 2;
 $55 = ($54|0)==(0);
 $56 = $1 & 48;
 $57 = ($56|0)!=(0);
 $58 = $1 & 112;
 $59 = ($15<<24>>24)==(3);
 $60 = ((($info)) + 56|0);
 $61 = ((($info)) + 12|0);
 $$pre = HEAP32[$34>>2]|0;
 $70 = $30;$820 = $$pre;$posy$036 = 0;$srcy$037 = 0;
 while(1) {
  $62 = ($posy$036|0)>(65535);
  if ($62) {
   $63 = (($posy$036) + -65536)|0;
   $64 = $63 >>> 16;
   $65 = $64 << 16;
   $66 = (($srcy$037) + 1)|0;
   $67 = (($63) - ($65))|0;
   $68 = (($66) + ($64))|0;
   $posy$1$lcssa = $67;$srcy$1$lcssa = $68;
  } else {
   $posy$1$lcssa = $posy$036;$srcy$1$lcssa = $srcy$037;
  }
  $69 = ($70|0)==(0);
  if (!($69)) {
   $$in = $70;$dst$033 = $820;$posx$030 = 65536;$src$032 = 0;$srcx$031 = -1;
   while(1) {
    $71 = (($$in) + -1)|0;
    $72 = ($posx$030|0)>(65535);
    if ($72) {
     $73 = (($posx$030) + -65536)|0;
     $74 = $73 >>> 16;
     $75 = $74 << 16;
     $76 = (($srcx$031) + 1)|0;
     $77 = (($73) - ($75))|0;
     $78 = (($76) + ($74))|0;
     $79 = HEAP32[$info>>2]|0;
     $80 = HEAP32[$61>>2]|0;
     $81 = Math_imul($80, $srcy$1$lcssa)|0;
     $82 = (($79) + ($81)|0);
     $83 = Math_imul($78, $16)|0;
     $84 = (($82) + ($83)|0);
     $posx$2 = $77;$src$1 = $84;$srcx$2 = $78;
    } else {
     $posx$2 = $posx$030;$src$1 = $src$032;$srcx$2 = $srcx$031;
    }
    $85 = HEAP32[$36>>2]|0;
    $86 = ($85|0)==(0);
    L15: do {
     if ($86) {
      switch ($16|0) {
      case 1:  {
       $226 = HEAP8[$src$1>>0]|0;
       $227 = $226&255;
       $228 = HEAP32[$37>>2]|0;
       $229 = $227 & $228;
       $230 = HEAP32[$38>>2]|0;
       $231 = $230 & 255;
       $232 = $229 >>> $231;
       $233 = HEAP32[$39>>2]|0;
       $234 = $233 & 255;
       $235 = (125924 + ($234<<2)|0);
       $236 = HEAP32[$235>>2]|0;
       $237 = (($236) + ($232)|0);
       $238 = HEAP8[$237>>0]|0;
       $239 = HEAP32[$40>>2]|0;
       $240 = $239 & $227;
       $241 = $230 >>> 8;
       $242 = $241 & 255;
       $243 = $240 >>> $242;
       $244 = $233 >>> 8;
       $245 = $244 & 255;
       $246 = (125924 + ($245<<2)|0);
       $247 = HEAP32[$246>>2]|0;
       $248 = (($247) + ($243)|0);
       $249 = HEAP8[$248>>0]|0;
       $250 = HEAP32[$41>>2]|0;
       $251 = $250 & $227;
       $252 = $230 >>> 16;
       $253 = $252 & 255;
       $254 = $251 >>> $253;
       $255 = $233 >>> 16;
       $256 = $255 & 255;
       $257 = (125924 + ($256<<2)|0);
       $258 = HEAP32[$257>>2]|0;
       $259 = (($258) + ($254)|0);
       $260 = HEAP8[$259>>0]|0;
       $srcA$0$shrunk = -1;$srcB$1$shrunk = $260;$srcG$1$shrunk = $249;$srcR$1$shrunk = $238;$srcpixel$1 = $227;
       break L15;
       break;
      }
      case 2:  {
       $261 = HEAP16[$src$1>>1]|0;
       $262 = $261&65535;
       $263 = HEAP32[$37>>2]|0;
       $264 = $262 & $263;
       $265 = HEAP32[$38>>2]|0;
       $266 = $265 & 255;
       $267 = $264 >>> $266;
       $268 = HEAP32[$39>>2]|0;
       $269 = $268 & 255;
       $270 = (125924 + ($269<<2)|0);
       $271 = HEAP32[$270>>2]|0;
       $272 = (($271) + ($267)|0);
       $273 = HEAP8[$272>>0]|0;
       $274 = HEAP32[$40>>2]|0;
       $275 = $274 & $262;
       $276 = $265 >>> 8;
       $277 = $276 & 255;
       $278 = $275 >>> $277;
       $279 = $268 >>> 8;
       $280 = $279 & 255;
       $281 = (125924 + ($280<<2)|0);
       $282 = HEAP32[$281>>2]|0;
       $283 = (($282) + ($278)|0);
       $284 = HEAP8[$283>>0]|0;
       $285 = HEAP32[$41>>2]|0;
       $286 = $285 & $262;
       $287 = $265 >>> 16;
       $288 = $287 & 255;
       $289 = $286 >>> $288;
       $290 = $268 >>> 16;
       $291 = $290 & 255;
       $292 = (125924 + ($291<<2)|0);
       $293 = HEAP32[$292>>2]|0;
       $294 = (($293) + ($289)|0);
       $295 = HEAP8[$294>>0]|0;
       $srcA$0$shrunk = -1;$srcB$1$shrunk = $295;$srcG$1$shrunk = $284;$srcR$1$shrunk = $273;$srcpixel$1 = $262;
       break L15;
       break;
      }
      case 3:  {
       $296 = HEAP32[$38>>2]|0;
       $div15 = $296 >>> 3;
       $297 = $div15 & 31;
       $298 = (($src$1) + ($297)|0);
       $299 = HEAP8[$298>>0]|0;
       $300 = $296 >>> 11;
       $div16 = $300 & 31;
       $301 = (($src$1) + ($div16)|0);
       $302 = HEAP8[$301>>0]|0;
       $303 = $296 >>> 19;
       $304 = $303 & 31;
       $305 = (($src$1) + ($304)|0);
       $306 = HEAP8[$305>>0]|0;
       $srcA$0$shrunk = -1;$srcB$1$shrunk = $306;$srcG$1$shrunk = $302;$srcR$1$shrunk = $299;$srcpixel$1 = 0;
       break L15;
       break;
      }
      case 4:  {
       $307 = HEAP32[$src$1>>2]|0;
       $308 = HEAP32[$37>>2]|0;
       $309 = $308 & $307;
       $310 = HEAP32[$38>>2]|0;
       $311 = $310 & 255;
       $312 = $309 >>> $311;
       $313 = HEAP32[$39>>2]|0;
       $314 = $313 & 255;
       $315 = (125924 + ($314<<2)|0);
       $316 = HEAP32[$315>>2]|0;
       $317 = (($316) + ($312)|0);
       $318 = HEAP8[$317>>0]|0;
       $319 = HEAP32[$40>>2]|0;
       $320 = $319 & $307;
       $321 = $310 >>> 8;
       $322 = $321 & 255;
       $323 = $320 >>> $322;
       $324 = $313 >>> 8;
       $325 = $324 & 255;
       $326 = (125924 + ($325<<2)|0);
       $327 = HEAP32[$326>>2]|0;
       $328 = (($327) + ($323)|0);
       $329 = HEAP8[$328>>0]|0;
       $330 = HEAP32[$41>>2]|0;
       $331 = $330 & $307;
       $332 = $310 >>> 16;
       $333 = $332 & 255;
       $334 = $331 >>> $333;
       $335 = $313 >>> 16;
       $336 = $335 & 255;
       $337 = (125924 + ($336<<2)|0);
       $338 = HEAP32[$337>>2]|0;
       $339 = (($338) + ($334)|0);
       $340 = HEAP8[$339>>0]|0;
       $srcA$0$shrunk = -1;$srcB$1$shrunk = $340;$srcG$1$shrunk = $329;$srcR$1$shrunk = $318;$srcpixel$1 = $307;
       break L15;
       break;
      }
      default: {
       $srcA$0$shrunk = -1;$srcB$1$shrunk = 0;$srcG$1$shrunk = 0;$srcR$1$shrunk = 0;$srcpixel$1 = 0;
       break L15;
      }
      }
     } else {
      switch ($16|0) {
      case 1:  {
       $87 = HEAP8[$src$1>>0]|0;
       $88 = $87&255;
       $89 = HEAP32[$37>>2]|0;
       $90 = $88 & $89;
       $91 = HEAP32[$38>>2]|0;
       $92 = $91 & 255;
       $93 = $90 >>> $92;
       $94 = HEAP32[$39>>2]|0;
       $95 = $94 & 255;
       $96 = (125924 + ($95<<2)|0);
       $97 = HEAP32[$96>>2]|0;
       $98 = (($97) + ($93)|0);
       $99 = HEAP8[$98>>0]|0;
       $100 = HEAP32[$40>>2]|0;
       $101 = $100 & $88;
       $102 = $91 >>> 8;
       $103 = $102 & 255;
       $104 = $101 >>> $103;
       $105 = $94 >>> 8;
       $106 = $105 & 255;
       $107 = (125924 + ($106<<2)|0);
       $108 = HEAP32[$107>>2]|0;
       $109 = (($108) + ($104)|0);
       $110 = HEAP8[$109>>0]|0;
       $111 = HEAP32[$41>>2]|0;
       $112 = $111 & $88;
       $113 = $91 >>> 16;
       $114 = $113 & 255;
       $115 = $112 >>> $114;
       $116 = $94 >>> 16;
       $117 = $116 & 255;
       $118 = (125924 + ($117<<2)|0);
       $119 = HEAP32[$118>>2]|0;
       $120 = (($119) + ($115)|0);
       $121 = HEAP8[$120>>0]|0;
       $122 = $88 & $85;
       $123 = $91 >>> 24;
       $124 = $122 >>> $123;
       $125 = $94 >>> 24;
       $126 = (125924 + ($125<<2)|0);
       $127 = HEAP32[$126>>2]|0;
       $128 = (($127) + ($124)|0);
       $129 = HEAP8[$128>>0]|0;
       $srcA$0$shrunk = $129;$srcB$1$shrunk = $121;$srcG$1$shrunk = $110;$srcR$1$shrunk = $99;$srcpixel$1 = $88;
       break L15;
       break;
      }
      case 2:  {
       $130 = HEAP16[$src$1>>1]|0;
       $131 = $130&65535;
       $132 = HEAP32[$37>>2]|0;
       $133 = $131 & $132;
       $134 = HEAP32[$38>>2]|0;
       $135 = $134 & 255;
       $136 = $133 >>> $135;
       $137 = HEAP32[$39>>2]|0;
       $138 = $137 & 255;
       $139 = (125924 + ($138<<2)|0);
       $140 = HEAP32[$139>>2]|0;
       $141 = (($140) + ($136)|0);
       $142 = HEAP8[$141>>0]|0;
       $143 = HEAP32[$40>>2]|0;
       $144 = $143 & $131;
       $145 = $134 >>> 8;
       $146 = $145 & 255;
       $147 = $144 >>> $146;
       $148 = $137 >>> 8;
       $149 = $148 & 255;
       $150 = (125924 + ($149<<2)|0);
       $151 = HEAP32[$150>>2]|0;
       $152 = (($151) + ($147)|0);
       $153 = HEAP8[$152>>0]|0;
       $154 = HEAP32[$41>>2]|0;
       $155 = $154 & $131;
       $156 = $134 >>> 16;
       $157 = $156 & 255;
       $158 = $155 >>> $157;
       $159 = $137 >>> 16;
       $160 = $159 & 255;
       $161 = (125924 + ($160<<2)|0);
       $162 = HEAP32[$161>>2]|0;
       $163 = (($162) + ($158)|0);
       $164 = HEAP8[$163>>0]|0;
       $165 = $131 & $85;
       $166 = $134 >>> 24;
       $167 = $165 >>> $166;
       $168 = $137 >>> 24;
       $169 = (125924 + ($168<<2)|0);
       $170 = HEAP32[$169>>2]|0;
       $171 = (($170) + ($167)|0);
       $172 = HEAP8[$171>>0]|0;
       $srcA$0$shrunk = $172;$srcB$1$shrunk = $164;$srcG$1$shrunk = $153;$srcR$1$shrunk = $142;$srcpixel$1 = $131;
       break L15;
       break;
      }
      case 3:  {
       $173 = HEAP32[$38>>2]|0;
       $div18 = $173 >>> 3;
       $174 = $div18 & 31;
       $175 = (($src$1) + ($174)|0);
       $176 = HEAP8[$175>>0]|0;
       $177 = $173 >>> 11;
       $div19 = $177 & 31;
       $178 = (($src$1) + ($div19)|0);
       $179 = HEAP8[$178>>0]|0;
       $180 = $173 >>> 19;
       $181 = $180 & 31;
       $182 = (($src$1) + ($181)|0);
       $183 = HEAP8[$182>>0]|0;
       $srcA$0$shrunk = -1;$srcB$1$shrunk = $183;$srcG$1$shrunk = $179;$srcR$1$shrunk = $176;$srcpixel$1 = 0;
       break L15;
       break;
      }
      case 4:  {
       $184 = HEAP32[$src$1>>2]|0;
       $185 = HEAP32[$37>>2]|0;
       $186 = $185 & $184;
       $187 = HEAP32[$38>>2]|0;
       $188 = $187 & 255;
       $189 = $186 >>> $188;
       $190 = HEAP32[$39>>2]|0;
       $191 = $190 & 255;
       $192 = (125924 + ($191<<2)|0);
       $193 = HEAP32[$192>>2]|0;
       $194 = (($193) + ($189)|0);
       $195 = HEAP8[$194>>0]|0;
       $196 = HEAP32[$40>>2]|0;
       $197 = $196 & $184;
       $198 = $187 >>> 8;
       $199 = $198 & 255;
       $200 = $197 >>> $199;
       $201 = $190 >>> 8;
       $202 = $201 & 255;
       $203 = (125924 + ($202<<2)|0);
       $204 = HEAP32[$203>>2]|0;
       $205 = (($204) + ($200)|0);
       $206 = HEAP8[$205>>0]|0;
       $207 = HEAP32[$41>>2]|0;
       $208 = $207 & $184;
       $209 = $187 >>> 16;
       $210 = $209 & 255;
       $211 = $208 >>> $210;
       $212 = $190 >>> 16;
       $213 = $212 & 255;
       $214 = (125924 + ($213<<2)|0);
       $215 = HEAP32[$214>>2]|0;
       $216 = (($215) + ($211)|0);
       $217 = HEAP8[$216>>0]|0;
       $218 = $184 & $85;
       $219 = $187 >>> 24;
       $220 = $218 >>> $219;
       $221 = $190 >>> 24;
       $222 = (125924 + ($221<<2)|0);
       $223 = HEAP32[$222>>2]|0;
       $224 = (($223) + ($220)|0);
       $225 = HEAP8[$224>>0]|0;
       $srcA$0$shrunk = $225;$srcB$1$shrunk = $217;$srcG$1$shrunk = $206;$srcR$1$shrunk = $195;$srcpixel$1 = $184;
       break L15;
       break;
      }
      default: {
       $srcA$0$shrunk = 0;$srcB$1$shrunk = 0;$srcG$1$shrunk = 0;$srcR$1$shrunk = 0;$srcpixel$1 = 0;
       break L15;
      }
      }
     }
    } while(0);
    $srcR$1 = $srcR$1$shrunk&255;
    $srcG$1 = $srcG$1$shrunk&255;
    $srcB$1 = $srcB$1$shrunk&255;
    $srcA$0 = $srcA$0$shrunk&255;
    if ($43) {
     label = 24;
    } else {
     if ($59) {
      $341 = HEAP32[$38>>2]|0;
      $342 = $341 & 255;
      $343 = $srcR$1 << $342;
      $344 = $341 >>> 8;
      $345 = $344 & 255;
      $346 = $srcG$1 << $345;
      $347 = $346 | $343;
      $348 = $341 >>> 16;
      $349 = $348 & 255;
      $350 = $srcB$1 << $349;
      $351 = $347 | $350;
      $srcpixel$2 = $351;
     } else {
      $srcpixel$2 = $srcpixel$1;
     }
     $352 = HEAP32[$60>>2]|0;
     $353 = ($srcpixel$2|0)==($352|0);
     if (!($353)) {
      label = 24;
     }
    }
    L34: do {
     if ((label|0) == 24) {
      label = 0;
      $357 = HEAP32[$44>>2]|0;
      $358 = ($357|0)==(0);
      L36: do {
       if ($358) {
        switch ($19|0) {
        case 1:  {
         $498 = HEAP8[$dst$033>>0]|0;
         $499 = $498&255;
         $500 = HEAP32[$45>>2]|0;
         $501 = $499 & $500;
         $502 = HEAP32[$46>>2]|0;
         $503 = $502 & 255;
         $504 = $501 >>> $503;
         $505 = HEAP32[$47>>2]|0;
         $506 = $505 & 255;
         $507 = (125924 + ($506<<2)|0);
         $508 = HEAP32[$507>>2]|0;
         $509 = (($508) + ($504)|0);
         $510 = HEAP8[$509>>0]|0;
         $511 = HEAP32[$48>>2]|0;
         $512 = $511 & $499;
         $513 = $502 >>> 8;
         $514 = $513 & 255;
         $515 = $512 >>> $514;
         $516 = $505 >>> 8;
         $517 = $516 & 255;
         $518 = (125924 + ($517<<2)|0);
         $519 = HEAP32[$518>>2]|0;
         $520 = (($519) + ($515)|0);
         $521 = HEAP8[$520>>0]|0;
         $522 = HEAP32[$50>>2]|0;
         $523 = $522 & $499;
         $524 = $502 >>> 16;
         $525 = $524 & 255;
         $526 = $523 >>> $525;
         $527 = $505 >>> 16;
         $528 = $527 & 255;
         $529 = (125924 + ($528<<2)|0);
         $530 = HEAP32[$529>>2]|0;
         $531 = (($530) + ($526)|0);
         $532 = HEAP8[$531>>0]|0;
         $dstA$0$shrunk = -1;$dstB$1$shrunk = $532;$dstG$1$shrunk = $521;$dstR$1$shrunk = $510;
         break L36;
         break;
        }
        case 2:  {
         $533 = HEAP16[$dst$033>>1]|0;
         $534 = $533&65535;
         $535 = HEAP32[$45>>2]|0;
         $536 = $534 & $535;
         $537 = HEAP32[$46>>2]|0;
         $538 = $537 & 255;
         $539 = $536 >>> $538;
         $540 = HEAP32[$47>>2]|0;
         $541 = $540 & 255;
         $542 = (125924 + ($541<<2)|0);
         $543 = HEAP32[$542>>2]|0;
         $544 = (($543) + ($539)|0);
         $545 = HEAP8[$544>>0]|0;
         $546 = HEAP32[$48>>2]|0;
         $547 = $546 & $534;
         $548 = $537 >>> 8;
         $549 = $548 & 255;
         $550 = $547 >>> $549;
         $551 = $540 >>> 8;
         $552 = $551 & 255;
         $553 = (125924 + ($552<<2)|0);
         $554 = HEAP32[$553>>2]|0;
         $555 = (($554) + ($550)|0);
         $556 = HEAP8[$555>>0]|0;
         $557 = HEAP32[$50>>2]|0;
         $558 = $557 & $534;
         $559 = $537 >>> 16;
         $560 = $559 & 255;
         $561 = $558 >>> $560;
         $562 = $540 >>> 16;
         $563 = $562 & 255;
         $564 = (125924 + ($563<<2)|0);
         $565 = HEAP32[$564>>2]|0;
         $566 = (($565) + ($561)|0);
         $567 = HEAP8[$566>>0]|0;
         $dstA$0$shrunk = -1;$dstB$1$shrunk = $567;$dstG$1$shrunk = $556;$dstR$1$shrunk = $545;
         break L36;
         break;
        }
        case 3:  {
         $568 = HEAP32[$46>>2]|0;
         $div9 = $568 >>> 3;
         $569 = $div9 & 31;
         $570 = (($dst$033) + ($569)|0);
         $571 = HEAP8[$570>>0]|0;
         $572 = $568 >>> 11;
         $div10 = $572 & 31;
         $573 = (($dst$033) + ($div10)|0);
         $574 = HEAP8[$573>>0]|0;
         $575 = $568 >>> 19;
         $576 = $575 & 31;
         $577 = (($dst$033) + ($576)|0);
         $578 = HEAP8[$577>>0]|0;
         $dstA$0$shrunk = -1;$dstB$1$shrunk = $578;$dstG$1$shrunk = $574;$dstR$1$shrunk = $571;
         break L36;
         break;
        }
        case 4:  {
         $579 = HEAP32[$dst$033>>2]|0;
         $580 = HEAP32[$45>>2]|0;
         $581 = $580 & $579;
         $582 = HEAP32[$46>>2]|0;
         $583 = $582 & 255;
         $584 = $581 >>> $583;
         $585 = HEAP32[$47>>2]|0;
         $586 = $585 & 255;
         $587 = (125924 + ($586<<2)|0);
         $588 = HEAP32[$587>>2]|0;
         $589 = (($588) + ($584)|0);
         $590 = HEAP8[$589>>0]|0;
         $591 = HEAP32[$48>>2]|0;
         $592 = $591 & $579;
         $593 = $582 >>> 8;
         $594 = $593 & 255;
         $595 = $592 >>> $594;
         $596 = $585 >>> 8;
         $597 = $596 & 255;
         $598 = (125924 + ($597<<2)|0);
         $599 = HEAP32[$598>>2]|0;
         $600 = (($599) + ($595)|0);
         $601 = HEAP8[$600>>0]|0;
         $602 = HEAP32[$50>>2]|0;
         $603 = $602 & $579;
         $604 = $582 >>> 16;
         $605 = $604 & 255;
         $606 = $603 >>> $605;
         $607 = $585 >>> 16;
         $608 = $607 & 255;
         $609 = (125924 + ($608<<2)|0);
         $610 = HEAP32[$609>>2]|0;
         $611 = (($610) + ($606)|0);
         $612 = HEAP8[$611>>0]|0;
         $dstA$0$shrunk = -1;$dstB$1$shrunk = $612;$dstG$1$shrunk = $601;$dstR$1$shrunk = $590;
         break L36;
         break;
        }
        default: {
         $dstA$0$shrunk = -1;$dstB$1$shrunk = 0;$dstG$1$shrunk = 0;$dstR$1$shrunk = 0;
         break L36;
        }
        }
       } else {
        switch ($19|0) {
        case 1:  {
         $359 = HEAP8[$dst$033>>0]|0;
         $360 = $359&255;
         $361 = HEAP32[$45>>2]|0;
         $362 = $360 & $361;
         $363 = HEAP32[$46>>2]|0;
         $364 = $363 & 255;
         $365 = $362 >>> $364;
         $366 = HEAP32[$47>>2]|0;
         $367 = $366 & 255;
         $368 = (125924 + ($367<<2)|0);
         $369 = HEAP32[$368>>2]|0;
         $370 = (($369) + ($365)|0);
         $371 = HEAP8[$370>>0]|0;
         $372 = HEAP32[$48>>2]|0;
         $373 = $372 & $360;
         $374 = $363 >>> 8;
         $375 = $374 & 255;
         $376 = $373 >>> $375;
         $377 = $366 >>> 8;
         $378 = $377 & 255;
         $379 = (125924 + ($378<<2)|0);
         $380 = HEAP32[$379>>2]|0;
         $381 = (($380) + ($376)|0);
         $382 = HEAP8[$381>>0]|0;
         $383 = HEAP32[$50>>2]|0;
         $384 = $383 & $360;
         $385 = $363 >>> 16;
         $386 = $385 & 255;
         $387 = $384 >>> $386;
         $388 = $366 >>> 16;
         $389 = $388 & 255;
         $390 = (125924 + ($389<<2)|0);
         $391 = HEAP32[$390>>2]|0;
         $392 = (($391) + ($387)|0);
         $393 = HEAP8[$392>>0]|0;
         $394 = $360 & $357;
         $395 = $363 >>> 24;
         $396 = $394 >>> $395;
         $397 = $366 >>> 24;
         $398 = (125924 + ($397<<2)|0);
         $399 = HEAP32[$398>>2]|0;
         $400 = (($399) + ($396)|0);
         $401 = HEAP8[$400>>0]|0;
         $dstA$0$shrunk = $401;$dstB$1$shrunk = $393;$dstG$1$shrunk = $382;$dstR$1$shrunk = $371;
         break L36;
         break;
        }
        case 2:  {
         $402 = HEAP16[$dst$033>>1]|0;
         $403 = $402&65535;
         $404 = HEAP32[$45>>2]|0;
         $405 = $403 & $404;
         $406 = HEAP32[$46>>2]|0;
         $407 = $406 & 255;
         $408 = $405 >>> $407;
         $409 = HEAP32[$47>>2]|0;
         $410 = $409 & 255;
         $411 = (125924 + ($410<<2)|0);
         $412 = HEAP32[$411>>2]|0;
         $413 = (($412) + ($408)|0);
         $414 = HEAP8[$413>>0]|0;
         $415 = HEAP32[$48>>2]|0;
         $416 = $415 & $403;
         $417 = $406 >>> 8;
         $418 = $417 & 255;
         $419 = $416 >>> $418;
         $420 = $409 >>> 8;
         $421 = $420 & 255;
         $422 = (125924 + ($421<<2)|0);
         $423 = HEAP32[$422>>2]|0;
         $424 = (($423) + ($419)|0);
         $425 = HEAP8[$424>>0]|0;
         $426 = HEAP32[$50>>2]|0;
         $427 = $426 & $403;
         $428 = $406 >>> 16;
         $429 = $428 & 255;
         $430 = $427 >>> $429;
         $431 = $409 >>> 16;
         $432 = $431 & 255;
         $433 = (125924 + ($432<<2)|0);
         $434 = HEAP32[$433>>2]|0;
         $435 = (($434) + ($430)|0);
         $436 = HEAP8[$435>>0]|0;
         $437 = $403 & $357;
         $438 = $406 >>> 24;
         $439 = $437 >>> $438;
         $440 = $409 >>> 24;
         $441 = (125924 + ($440<<2)|0);
         $442 = HEAP32[$441>>2]|0;
         $443 = (($442) + ($439)|0);
         $444 = HEAP8[$443>>0]|0;
         $dstA$0$shrunk = $444;$dstB$1$shrunk = $436;$dstG$1$shrunk = $425;$dstR$1$shrunk = $414;
         break L36;
         break;
        }
        case 3:  {
         $445 = HEAP32[$46>>2]|0;
         $div12 = $445 >>> 3;
         $446 = $div12 & 31;
         $447 = (($dst$033) + ($446)|0);
         $448 = HEAP8[$447>>0]|0;
         $449 = $445 >>> 11;
         $div13 = $449 & 31;
         $450 = (($dst$033) + ($div13)|0);
         $451 = HEAP8[$450>>0]|0;
         $452 = $445 >>> 19;
         $453 = $452 & 31;
         $454 = (($dst$033) + ($453)|0);
         $455 = HEAP8[$454>>0]|0;
         $dstA$0$shrunk = -1;$dstB$1$shrunk = $455;$dstG$1$shrunk = $451;$dstR$1$shrunk = $448;
         break L36;
         break;
        }
        case 4:  {
         $456 = HEAP32[$dst$033>>2]|0;
         $457 = HEAP32[$45>>2]|0;
         $458 = $457 & $456;
         $459 = HEAP32[$46>>2]|0;
         $460 = $459 & 255;
         $461 = $458 >>> $460;
         $462 = HEAP32[$47>>2]|0;
         $463 = $462 & 255;
         $464 = (125924 + ($463<<2)|0);
         $465 = HEAP32[$464>>2]|0;
         $466 = (($465) + ($461)|0);
         $467 = HEAP8[$466>>0]|0;
         $468 = HEAP32[$48>>2]|0;
         $469 = $468 & $456;
         $470 = $459 >>> 8;
         $471 = $470 & 255;
         $472 = $469 >>> $471;
         $473 = $462 >>> 8;
         $474 = $473 & 255;
         $475 = (125924 + ($474<<2)|0);
         $476 = HEAP32[$475>>2]|0;
         $477 = (($476) + ($472)|0);
         $478 = HEAP8[$477>>0]|0;
         $479 = HEAP32[$50>>2]|0;
         $480 = $479 & $456;
         $481 = $459 >>> 16;
         $482 = $481 & 255;
         $483 = $480 >>> $482;
         $484 = $462 >>> 16;
         $485 = $484 & 255;
         $486 = (125924 + ($485<<2)|0);
         $487 = HEAP32[$486>>2]|0;
         $488 = (($487) + ($483)|0);
         $489 = HEAP8[$488>>0]|0;
         $490 = $456 & $357;
         $491 = $459 >>> 24;
         $492 = $490 >>> $491;
         $493 = $462 >>> 24;
         $494 = (125924 + ($493<<2)|0);
         $495 = HEAP32[$494>>2]|0;
         $496 = (($495) + ($492)|0);
         $497 = HEAP8[$496>>0]|0;
         $dstA$0$shrunk = $497;$dstB$1$shrunk = $489;$dstG$1$shrunk = $478;$dstR$1$shrunk = $467;
         break L36;
         break;
        }
        default: {
         $dstA$0$shrunk = 0;$dstB$1$shrunk = 0;$dstG$1$shrunk = 0;$dstR$1$shrunk = 0;
         break L36;
        }
        }
       }
      } while(0);
      $dstR$1 = $dstR$1$shrunk&255;
      $dstG$1 = $dstG$1$shrunk&255;
      $dstB$1 = $dstB$1$shrunk&255;
      $dstA$0 = $dstA$0$shrunk&255;
      if ($53) {
       $srcB$2 = $srcB$1;$srcG$2 = $srcG$1;$srcR$2 = $srcR$1;
      } else {
       $613 = Math_imul($srcR$1, $4)|0;
       $614 = (($613>>>0) / 255)&-1;
       $615 = Math_imul($srcG$1, $6)|0;
       $616 = (($615>>>0) / 255)&-1;
       $617 = Math_imul($srcB$1, $8)|0;
       $618 = (($617>>>0) / 255)&-1;
       $srcB$2 = $618;$srcG$2 = $616;$srcR$2 = $614;
      }
      if ($55) {
       $srcA$1 = $srcA$0;
      } else {
       $619 = Math_imul($srcA$0, $9)|0;
       $620 = (($619>>>0) / 255)&-1;
       $srcA$1 = $620;
      }
      $621 = ($srcA$1>>>0)<(255);
      $or$cond = $57 & $621;
      if ($or$cond) {
       $622 = Math_imul($srcA$1, $srcR$2)|0;
       $623 = (($622>>>0) / 255)&-1;
       $624 = Math_imul($srcA$1, $srcG$2)|0;
       $625 = (($624>>>0) / 255)&-1;
       $626 = Math_imul($srcA$1, $srcB$2)|0;
       $627 = (($626>>>0) / 255)&-1;
       $srcB$3 = $627;$srcG$3 = $625;$srcR$3 = $623;
      } else {
       $srcB$3 = $srcB$2;$srcG$3 = $srcG$2;$srcR$3 = $srcR$2;
      }
      switch ($58|0) {
      case 0:  {
       $dstA$1 = $srcA$1;$dstB$2 = $srcB$3;$dstG$2 = $srcG$3;$dstR$2 = $srcR$3;
       break;
      }
      case 16:  {
       $628 = (255 - ($srcA$1))|0;
       $629 = Math_imul($628, $dstR$1)|0;
       $630 = (($629>>>0) / 255)&-1;
       $631 = (($srcR$3) + ($630))|0;
       $632 = Math_imul($628, $dstG$1)|0;
       $633 = (($632>>>0) / 255)&-1;
       $634 = (($srcG$3) + ($633))|0;
       $635 = Math_imul($628, $dstB$1)|0;
       $636 = (($635>>>0) / 255)&-1;
       $637 = (($srcB$3) + ($636))|0;
       $dstA$1 = $dstA$0;$dstB$2 = $637;$dstG$2 = $634;$dstR$2 = $631;
       break;
      }
      case 32:  {
       $638 = (($srcR$3) + ($dstR$1))|0;
       $639 = ($638>>>0)>(255);
       $$ = $639 ? 255 : $638;
       $640 = (($srcG$3) + ($dstG$1))|0;
       $641 = ($640>>>0)>(255);
       $$3 = $641 ? 255 : $640;
       $642 = (($srcB$3) + ($dstB$1))|0;
       $643 = ($642>>>0)>(255);
       $$1 = $643 ? 255 : $642;
       $dstA$1 = $dstA$0;$dstB$2 = $$1;$dstG$2 = $$3;$dstR$2 = $$;
       break;
      }
      case 64:  {
       $644 = Math_imul($srcR$3, $dstR$1)|0;
       $645 = (($644>>>0) / 255)&-1;
       $646 = Math_imul($srcG$3, $dstG$1)|0;
       $647 = (($646>>>0) / 255)&-1;
       $648 = Math_imul($srcB$3, $dstB$1)|0;
       $649 = (($648>>>0) / 255)&-1;
       $dstA$1 = $dstA$0;$dstB$2 = $649;$dstG$2 = $647;$dstR$2 = $645;
       break;
      }
      default: {
       $dstA$1 = $dstA$0;$dstB$2 = $dstB$1;$dstG$2 = $dstG$1;$dstR$2 = $dstR$1;
      }
      }
      if ($358) {
       switch ($19|0) {
       case 1:  {
        $739 = HEAP32[$47>>2]|0;
        $740 = $739 & 255;
        $741 = $dstR$2 >>> $740;
        $742 = HEAP32[$46>>2]|0;
        $743 = $742 & 255;
        $744 = $741 << $743;
        $745 = $739 >>> 8;
        $746 = $745 & 255;
        $747 = $dstG$2 >>> $746;
        $748 = $742 >>> 8;
        $749 = $748 & 255;
        $750 = $747 << $749;
        $751 = $739 >>> 16;
        $752 = $751 & 255;
        $753 = $dstB$2 >>> $752;
        $754 = $742 >>> 16;
        $755 = $754 & 255;
        $756 = $753 << $755;
        $757 = $744 | $750;
        $758 = $757 | $756;
        $759 = $758&255;
        HEAP8[$dst$033>>0] = $759;
        break L34;
        break;
       }
       case 2:  {
        $760 = HEAP32[$47>>2]|0;
        $761 = $760 & 255;
        $762 = $dstR$2 >>> $761;
        $763 = HEAP32[$46>>2]|0;
        $764 = $763 & 255;
        $765 = $762 << $764;
        $766 = $760 >>> 8;
        $767 = $766 & 255;
        $768 = $dstG$2 >>> $767;
        $769 = $763 >>> 8;
        $770 = $769 & 255;
        $771 = $768 << $770;
        $772 = $760 >>> 16;
        $773 = $772 & 255;
        $774 = $dstB$2 >>> $773;
        $775 = $763 >>> 16;
        $776 = $775 & 255;
        $777 = $774 << $776;
        $778 = $765 | $771;
        $779 = $778 | $777;
        $780 = $779&65535;
        HEAP16[$dst$033>>1] = $780;
        break L34;
        break;
       }
       case 3:  {
        $781 = $dstR$2&255;
        $782 = HEAP8[$46>>0]|0;
        $div = ($782&255) >>> 3;
        $783 = $div&255;
        $784 = (($dst$033) + ($783)|0);
        HEAP8[$784>>0] = $781;
        $785 = $dstG$2&255;
        $786 = HEAP8[$49>>0]|0;
        $div4 = ($786&255) >>> 3;
        $787 = $div4&255;
        $788 = (($dst$033) + ($787)|0);
        HEAP8[$788>>0] = $785;
        $789 = $dstB$2&255;
        $790 = HEAP8[$51>>0]|0;
        $div5 = ($790&255) >>> 3;
        $791 = $div5&255;
        $792 = (($dst$033) + ($791)|0);
        HEAP8[$792>>0] = $789;
        break L34;
        break;
       }
       case 4:  {
        $793 = HEAP32[$47>>2]|0;
        $794 = $793 & 255;
        $795 = $dstR$2 >>> $794;
        $796 = HEAP32[$46>>2]|0;
        $797 = $796 & 255;
        $798 = $795 << $797;
        $799 = $793 >>> 8;
        $800 = $799 & 255;
        $801 = $dstG$2 >>> $800;
        $802 = $796 >>> 8;
        $803 = $802 & 255;
        $804 = $801 << $803;
        $805 = $793 >>> 16;
        $806 = $805 & 255;
        $807 = $dstB$2 >>> $806;
        $808 = $796 >>> 16;
        $809 = $808 & 255;
        $810 = $807 << $809;
        $811 = $798 | $804;
        $812 = $811 | $810;
        HEAP32[$dst$033>>2] = $812;
        break L34;
        break;
       }
       default: {
        break L34;
       }
       }
      } else {
       switch ($19|0) {
       case 1:  {
        $650 = HEAP32[$47>>2]|0;
        $651 = $650 & 255;
        $652 = $dstR$2 >>> $651;
        $653 = HEAP32[$46>>2]|0;
        $654 = $653 & 255;
        $655 = $652 << $654;
        $656 = $650 >>> 8;
        $657 = $656 & 255;
        $658 = $dstG$2 >>> $657;
        $659 = $653 >>> 8;
        $660 = $659 & 255;
        $661 = $658 << $660;
        $662 = $661 | $655;
        $663 = $650 >>> 16;
        $664 = $663 & 255;
        $665 = $dstB$2 >>> $664;
        $666 = $653 >>> 16;
        $667 = $666 & 255;
        $668 = $665 << $667;
        $669 = $662 | $668;
        $670 = $650 >>> 24;
        $671 = $dstA$1 >>> $670;
        $672 = $653 >>> 24;
        $673 = $671 << $672;
        $674 = $669 | $673;
        $675 = $674&255;
        HEAP8[$dst$033>>0] = $675;
        break L34;
        break;
       }
       case 2:  {
        $676 = HEAP32[$47>>2]|0;
        $677 = $676 & 255;
        $678 = $dstR$2 >>> $677;
        $679 = HEAP32[$46>>2]|0;
        $680 = $679 & 255;
        $681 = $678 << $680;
        $682 = $676 >>> 8;
        $683 = $682 & 255;
        $684 = $dstG$2 >>> $683;
        $685 = $679 >>> 8;
        $686 = $685 & 255;
        $687 = $684 << $686;
        $688 = $687 | $681;
        $689 = $676 >>> 16;
        $690 = $689 & 255;
        $691 = $dstB$2 >>> $690;
        $692 = $679 >>> 16;
        $693 = $692 & 255;
        $694 = $691 << $693;
        $695 = $688 | $694;
        $696 = $676 >>> 24;
        $697 = $dstA$1 >>> $696;
        $698 = $679 >>> 24;
        $699 = $697 << $698;
        $700 = $695 | $699;
        $701 = $700&65535;
        HEAP16[$dst$033>>1] = $701;
        break L34;
        break;
       }
       case 3:  {
        $702 = $dstR$2&255;
        $703 = HEAP8[$46>>0]|0;
        $div6 = ($703&255) >>> 3;
        $704 = $div6&255;
        $705 = (($dst$033) + ($704)|0);
        HEAP8[$705>>0] = $702;
        $706 = $dstG$2&255;
        $707 = HEAP8[$49>>0]|0;
        $div7 = ($707&255) >>> 3;
        $708 = $div7&255;
        $709 = (($dst$033) + ($708)|0);
        HEAP8[$709>>0] = $706;
        $710 = $dstB$2&255;
        $711 = HEAP8[$51>>0]|0;
        $div8 = ($711&255) >>> 3;
        $712 = $div8&255;
        $713 = (($dst$033) + ($712)|0);
        HEAP8[$713>>0] = $710;
        break L34;
        break;
       }
       case 4:  {
        $714 = HEAP32[$47>>2]|0;
        $715 = $714 & 255;
        $716 = $dstR$2 >>> $715;
        $717 = HEAP32[$46>>2]|0;
        $718 = $717 & 255;
        $719 = $716 << $718;
        $720 = $714 >>> 8;
        $721 = $720 & 255;
        $722 = $dstG$2 >>> $721;
        $723 = $717 >>> 8;
        $724 = $723 & 255;
        $725 = $722 << $724;
        $726 = $725 | $719;
        $727 = $714 >>> 16;
        $728 = $727 & 255;
        $729 = $dstB$2 >>> $728;
        $730 = $717 >>> 16;
        $731 = $730 & 255;
        $732 = $729 << $731;
        $733 = $726 | $732;
        $734 = $714 >>> 24;
        $735 = $dstA$1 >>> $734;
        $736 = $717 >>> 24;
        $737 = $735 << $736;
        $738 = $733 | $737;
        HEAP32[$dst$033>>2] = $738;
        break L34;
        break;
       }
       default: {
        break L34;
       }
       }
      }
     }
    } while(0);
    $354 = (($posx$2) + ($31))|0;
    $355 = (($dst$033) + ($19)|0);
    $356 = ($71|0)==(0);
    if ($356) {
     break;
    } else {
     $$in = $71;$dst$033 = $355;$posx$030 = $354;$src$032 = $src$1;$srcx$031 = $srcx$2;
    }
   }
  }
  $813 = HEAP32[$35>>2]|0;
  $814 = HEAP32[$34>>2]|0;
  $815 = (($814) + ($813)|0);
  HEAP32[$34>>2] = $815;
  $816 = HEAP32[$23>>2]|0;
  $817 = (($816) + -1)|0;
  HEAP32[$23>>2] = $817;
  $818 = ($816|0)==(0);
  if ($818) {
   break;
  }
  $819 = (($posy$1$lcssa) + ($25))|0;
  $$pre44 = HEAP32[$29>>2]|0;
  $70 = $$pre44;$820 = $815;$posy$036 = $819;$srcy$037 = $srcy$1$lcssa;
 }
 return;
}
function _GLES2_RenderPresent($renderer) {
 $renderer = $renderer|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[300738]|0;
 $3 = HEAP32[$1>>2]|0;
 $4 = ($2|0)==($3|0);
 if ($4) {
  $12 = $1;
  label = 4;
 } else {
  $5 = ((($1)) + 268|0);
  HEAP32[$5>>2] = 0;
  $6 = ((($renderer)) + 188|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (_SDL_GL_MakeCurrent($7,$3)|0);
  $9 = ($8|0)<(0);
  if (!($9)) {
   $10 = HEAP32[$1>>2]|0;
   HEAP32[300738] = $10;
   (_GLES2_UpdateViewport($renderer)|0);
   $$pre$i = HEAP32[$0>>2]|0;
   $12 = $$pre$i;
   label = 4;
  }
 }
 if ((label|0) == 4) {
  $11 = ((($12)) + 4|0);
  $13 = HEAP32[$11>>2]|0;
  $14 = ($13|0)==(0);
  if (!($14)) {
   $15 = ((($12)) + 108|0);
   while(1) {
    $16 = HEAP32[$15>>2]|0;
    $17 = (FUNCTION_TABLE_i[$16 & 8191]()|0);
    $18 = ($17|0)==(0);
    if ($18) {
     break;
    }
   }
  }
 }
 $19 = ((($renderer)) + 188|0);
 $20 = HEAP32[$19>>2]|0;
 _SDL_GL_SwapWindow($20);
 return;
}
function _SDL_GL_SwapWindow($window) {
 $window = $window|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0;
 var $vararg_buffer5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ((($window)) + 48|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = $7 & 2;
   $9 = ($8|0)==(0);
   if ($9) {
    (_SDL_SetError(597590,$vararg_buffer3)|0);
    STACKTOP = sp;return;
   }
   $10 = ((($0)) + 636|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = (_SDL_TLSGet($11)|0);
   $13 = ($12|0)==($window|0);
   if ($13) {
    $14 = HEAP32[300441]|0;
    $15 = ((($14)) + 172|0);
    $16 = HEAP32[$15>>2]|0;
    FUNCTION_TABLE_vii[$16 & 8191]($14,$window);
    STACKTOP = sp;return;
   } else {
    (_SDL_SetError(603261,$vararg_buffer5)|0);
    STACKTOP = sp;return;
   }
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 STACKTOP = sp;return;
}
function _GLES2_DestroyTexture($renderer,$texture) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($texture)) + 68|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[300738]|0;
 $5 = HEAP32[$1>>2]|0;
 $6 = ($4|0)==($5|0);
 if ($6) {
  $14 = $1;
  label = 4;
 } else {
  $7 = ((($1)) + 268|0);
  HEAP32[$7>>2] = 0;
  $8 = ((($renderer)) + 188|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = (_SDL_GL_MakeCurrent($9,$5)|0);
  $11 = ($10|0)<(0);
  if (!($11)) {
   $12 = HEAP32[$1>>2]|0;
   HEAP32[300738] = $12;
   (_GLES2_UpdateViewport($renderer)|0);
   $$pre$i = HEAP32[$0>>2]|0;
   $14 = $$pre$i;
   label = 4;
  }
 }
 if ((label|0) == 4) {
  $13 = ((($14)) + 4|0);
  $15 = HEAP32[$13>>2]|0;
  $16 = ($15|0)==(0);
  if (!($16)) {
   $17 = ((($14)) + 108|0);
   while(1) {
    $18 = HEAP32[$17>>2]|0;
    $19 = (FUNCTION_TABLE_i[$18 & 8191]()|0);
    $20 = ($19|0)==(0);
    if ($20) {
     break;
    }
   }
  }
 }
 $21 = ($3|0)==(0|0);
 if ($21) {
  return;
 }
 $22 = ((($1)) + 64|0);
 $23 = HEAP32[$22>>2]|0;
 FUNCTION_TABLE_vii[$23 & 8191](1,$3);
 $24 = ((($3)) + 32|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = ($25|0)==(0);
 if (!($26)) {
  $27 = HEAP32[$22>>2]|0;
  FUNCTION_TABLE_vii[$27 & 8191](1,$24);
 }
 $28 = ((($3)) + 36|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = ($29|0)==(0);
 if (!($30)) {
  $31 = HEAP32[$22>>2]|0;
  FUNCTION_TABLE_vii[$31 & 8191](1,$28);
 }
 $32 = ((($3)) + 16|0);
 $33 = HEAP32[$32>>2]|0;
 _SDL_free($33);
 _SDL_free($3);
 HEAP32[$2>>2] = 0;
 return;
}
function _GLES2_BindTexture($renderer,$texture,$texw,$texh) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 $texw = $texw|0;
 $texh = $texh|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($texture)) + 68|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[300738]|0;
 $5 = HEAP32[$1>>2]|0;
 $6 = ($4|0)==($5|0);
 if ($6) {
  $14 = $1;
  label = 4;
 } else {
  $7 = ((($1)) + 268|0);
  HEAP32[$7>>2] = 0;
  $8 = ((($renderer)) + 188|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = (_SDL_GL_MakeCurrent($9,$5)|0);
  $11 = ($10|0)<(0);
  if (!($11)) {
   $12 = HEAP32[$1>>2]|0;
   HEAP32[300738] = $12;
   (_GLES2_UpdateViewport($renderer)|0);
   $$pre$i = HEAP32[$0>>2]|0;
   $14 = $$pre$i;
   label = 4;
  }
 }
 if ((label|0) == 4) {
  $13 = ((($14)) + 4|0);
  $15 = HEAP32[$13>>2]|0;
  $16 = ($15|0)==(0);
  if (!($16)) {
   $17 = ((($14)) + 108|0);
   while(1) {
    $18 = HEAP32[$17>>2]|0;
    $19 = (FUNCTION_TABLE_i[$18 & 8191]()|0);
    $20 = ($19|0)==(0);
    if ($20) {
     break;
    }
   }
  }
 }
 $21 = ((($1)) + 28|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ((($3)) + 4|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = HEAP32[$3>>2]|0;
 FUNCTION_TABLE_vii[$22 & 8191]($24,$25);
 $26 = ($texw|0)==(0|0);
 if (!($26)) {
  HEAPF32[$texw>>2] = 1.0;
 }
 $27 = ($texh|0)==(0|0);
 if ($27) {
  return 0;
 }
 HEAPF32[$texh>>2] = 1.0;
 return 0;
}
function _GLES2_UnbindTexture($renderer,$texture) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($texture)) + 68|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[300738]|0;
 $5 = HEAP32[$1>>2]|0;
 $6 = ($4|0)==($5|0);
 if ($6) {
  $14 = $1;
  label = 4;
 } else {
  $7 = ((($1)) + 268|0);
  HEAP32[$7>>2] = 0;
  $8 = ((($renderer)) + 188|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = (_SDL_GL_MakeCurrent($9,$5)|0);
  $11 = ($10|0)<(0);
  if (!($11)) {
   $12 = HEAP32[$1>>2]|0;
   HEAP32[300738] = $12;
   (_GLES2_UpdateViewport($renderer)|0);
   $$pre$i = HEAP32[$0>>2]|0;
   $14 = $$pre$i;
   label = 4;
  }
 }
 if ((label|0) == 4) {
  $13 = ((($14)) + 4|0);
  $15 = HEAP32[$13>>2]|0;
  $16 = ($15|0)==(0);
  if (!($16)) {
   $17 = ((($14)) + 108|0);
   while(1) {
    $18 = HEAP32[$17>>2]|0;
    $19 = (FUNCTION_TABLE_i[$18 & 8191]()|0);
    $20 = ($19|0)==(0);
    if ($20) {
     break;
    }
   }
  }
 }
 $21 = ((($1)) + 28|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ((($3)) + 4|0);
 $24 = HEAP32[$23>>2]|0;
 FUNCTION_TABLE_vii[$22 & 8191]($24,0);
 return 0;
}
function _SW_CreateRenderer($window,$flags) {
 $window = $window|0;
 $flags = $flags|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GetWindowSurface($window)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  $$0 = 0;
  return ($$0|0);
 }
 $2 = (_SW_CreateRendererForSurface($0)|0);
 $$0 = $2;
 return ($$0|0);
}
function _SDL_GetWindowSurface($window) {
 $window = $window|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $Amask$i = 0, $Bmask$i = 0, $Gmask$i = 0, $Rmask$i = 0, $bpp$i = 0, $format$i = 0, $pitch$i = 0, $pixels$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $format$i = sp + 40|0;
 $pixels$i = sp + 36|0;
 $pitch$i = sp + 32|0;
 $bpp$i = sp + 28|0;
 $Rmask$i = sp + 24|0;
 $Gmask$i = sp + 20|0;
 $Bmask$i = sp + 16|0;
 $Amask$i = sp + 12|0;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ((($window)) + 108|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0);
   $9 = ((($window)) + 104|0);
   $10 = HEAP32[$9>>2]|0;
   if (!($8)) {
    $$0 = $10;
    STACKTOP = sp;return ($$0|0);
   }
   $11 = ($10|0)==(0|0);
   if ($11) {
    $15 = $0;
   } else {
    $12 = HEAP32[$10>>2]|0;
    $13 = $12 & -5;
    HEAP32[$10>>2] = $13;
    _SDL_FreeSurface($10);
    $$pre = HEAP32[300441]|0;
    $15 = $$pre;
   }
   $14 = ((($15)) + 108|0);
   $16 = HEAP32[$14>>2]|0;
   $17 = ($16|0)==(0|0);
   if ($17) {
    $38 = 0;
   } else {
    $18 = ((($15)) + 112|0);
    $19 = HEAP32[$18>>2]|0;
    $20 = ($19|0)==(0|0);
    if ($20) {
     $38 = 0;
    } else {
     $21 = (FUNCTION_TABLE_iiiiii[$16 & 8191]($15,$window,$format$i,$pixels$i,$pitch$i)|0);
     $22 = ($21|0)<(0);
     if ($22) {
      $38 = 0;
     } else {
      $23 = HEAP32[$format$i>>2]|0;
      $24 = (_SDL_PixelFormatEnumToMasks($23,$bpp$i,$Rmask$i,$Gmask$i,$Bmask$i,$Amask$i)|0);
      $25 = ($24|0)==(0);
      if ($25) {
       $38 = 0;
      } else {
       $26 = HEAP32[$pixels$i>>2]|0;
       $27 = ((($window)) + 24|0);
       $28 = HEAP32[$27>>2]|0;
       $29 = ((($window)) + 28|0);
       $30 = HEAP32[$29>>2]|0;
       $31 = HEAP32[$bpp$i>>2]|0;
       $32 = HEAP32[$pitch$i>>2]|0;
       $33 = HEAP32[$Rmask$i>>2]|0;
       $34 = HEAP32[$Gmask$i>>2]|0;
       $35 = HEAP32[$Bmask$i>>2]|0;
       $36 = HEAP32[$Amask$i>>2]|0;
       $37 = (_SDL_CreateRGBSurfaceFrom($26,$28,$30,$31,$32,$33,$34,$35,$36)|0);
       $38 = $37;
      }
     }
    }
   }
   HEAP32[$9>>2] = $38;
   $39 = ($38|0)==(0|0);
   if ($39) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   HEAP32[$6>>2] = 1;
   $40 = HEAP32[$38>>2]|0;
   $41 = $40 | 4;
   HEAP32[$38>>2] = $41;
   $$0 = $38;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_CreateRGBSurfaceFrom($pixels,$width,$height,$depth,$pitch,$Rmask,$Gmask,$Bmask,$Amask) {
 $pixels = $pixels|0;
 $width = $width|0;
 $height = $height|0;
 $depth = $depth|0;
 $pitch = $pitch|0;
 $Rmask = $Rmask|0;
 $Gmask = $Gmask|0;
 $Bmask = $Bmask|0;
 $Amask = $Amask|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_CreateRGBSurface(0,0,0,$depth,$Rmask,$Gmask,$Bmask,$Amask)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  return ($0|0);
 }
 $2 = HEAP32[$0>>2]|0;
 $3 = $2 | 1;
 HEAP32[$0>>2] = $3;
 $4 = ((($0)) + 20|0);
 HEAP32[$4>>2] = $pixels;
 $5 = ((($0)) + 8|0);
 HEAP32[$5>>2] = $width;
 $6 = ((($0)) + 12|0);
 HEAP32[$6>>2] = $height;
 $7 = ((($0)) + 16|0);
 HEAP32[$7>>2] = $pitch;
 $8 = ((($0)) + 36|0);
 HEAP32[$8>>2] = 0;
 $9 = ((($0)) + 40|0);
 HEAP32[$9>>2] = 0;
 $10 = ((($0)) + 44|0);
 HEAP32[$10>>2] = $width;
 $11 = ((($0)) + 48|0);
 HEAP32[$11>>2] = $height;
 return ($0|0);
}
function _SDL_CreateRGBSurface($flags,$width,$height,$depth,$Rmask,$Gmask,$Bmask,$Amask) {
 $flags = $flags|0;
 $width = $width|0;
 $height = $height|0;
 $depth = $depth|0;
 $Rmask = $Rmask|0;
 $Gmask = $Gmask|0;
 $Bmask = $Bmask|0;
 $Amask = $Amask|0;
 var $$1 = 0, $$mask = 0, $$off = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $switch = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_SDL_MasksToPixelFormatEnum($depth,$Rmask,$Gmask,$Bmask,$Amask)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  (_SDL_SetError(603356,$vararg_buffer)|0);
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $2 = (_SDL_calloc(1,60)|0);
 $3 = ($2|0)==(0|0);
 if ($3) {
  (_SDL_Error(0)|0);
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $4 = (_SDL_AllocFormat($0)|0);
 $5 = ((($2)) + 4|0);
 HEAP32[$5>>2] = $4;
 $6 = ($4|0)==(0|0);
 if ($6) {
  _SDL_FreeSurface($2);
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $7 = ((($2)) + 8|0);
 HEAP32[$7>>2] = $width;
 $8 = ((($2)) + 12|0);
 HEAP32[$8>>2] = $height;
 $9 = (_SDL_CalculatePitch($2)|0);
 $10 = ((($2)) + 16|0);
 HEAP32[$10>>2] = $9;
 $11 = HEAP32[$7>>2]|0;
 $12 = HEAP32[$8>>2]|0;
 $13 = ((($2)) + 36|0);
 HEAP32[$13>>2] = 0;
 $14 = ((($2)) + 40|0);
 HEAP32[$14>>2] = 0;
 $15 = ((($2)) + 44|0);
 HEAP32[$15>>2] = $11;
 $16 = ((($2)) + 48|0);
 HEAP32[$16>>2] = $12;
 $17 = HEAP32[$5>>2]|0;
 $18 = HEAP32[$17>>2]|0;
 $19 = ($18|0)==(0);
 $$mask = $18 & -268435456;
 $20 = ($$mask|0)==(268435456);
 $or$cond = $19 | $20;
 if ($or$cond) {
  $21 = $18 >>> 24;
  $22 = $21 & 15;
  $$off = (($22) + -1)|0;
  $switch = ($$off>>>0)<(3);
  if ($switch) {
   $23 = ((($17)) + 8|0);
   $24 = HEAP8[$23>>0]|0;
   $25 = $24&255;
   $26 = 1 << $25;
   $27 = (_SDL_AllocPalette($26)|0);
   $28 = ($27|0)==(0|0);
   if ($28) {
    _SDL_FreeSurface($2);
    $$1 = 0;
    STACKTOP = sp;return ($$1|0);
   }
   $29 = HEAP32[$27>>2]|0;
   $30 = ($29|0)==(2);
   if ($30) {
    $31 = ((($27)) + 4|0);
    $32 = HEAP32[$31>>2]|0;
    HEAP8[$32>>0] = -1;
    $33 = ((($32)) + 1|0);
    HEAP8[$33>>0] = -1;
    $34 = ((($32)) + 2|0);
    HEAP8[$34>>0] = -1;
    $35 = ((($32)) + 4|0);
    HEAP8[$35>>0] = 0;
    $36 = ((($32)) + 5|0);
    HEAP8[$36>>0] = 0;
    $37 = ((($32)) + 6|0);
    HEAP8[$37>>0] = 0;
   }
   $38 = HEAP32[$5>>2]|0;
   $39 = (_SDL_SetPixelFormatPalette($38,$27)|0);
   $40 = ($39|0)<(0);
   if (!($40)) {
    $41 = ((($2)) + 52|0);
    $42 = HEAP32[$41>>2]|0;
    _SDL_InvalidateMap($42);
   }
   _SDL_FreePalette($27);
   $$pre = HEAP32[$7>>2]|0;
   $44 = $$pre;
  } else {
   $44 = $11;
  }
 } else {
  $44 = $11;
 }
 $43 = ($44|0)==(0);
 do {
  if (!($43)) {
   $45 = HEAP32[$8>>2]|0;
   $46 = ($45|0)==(0);
   if (!($46)) {
    $47 = HEAP32[$10>>2]|0;
    $48 = Math_imul($47, $45)|0;
    $49 = (_SDL_malloc($48)|0);
    $50 = ((($2)) + 20|0);
    HEAP32[$50>>2] = $49;
    $51 = ($49|0)==(0|0);
    if (!($51)) {
     $52 = HEAP32[$8>>2]|0;
     $53 = HEAP32[$10>>2]|0;
     $54 = Math_imul($53, $52)|0;
     (_SDL_memset($49,0,$54)|0);
     break;
    }
    _SDL_FreeSurface($2);
    (_SDL_Error(0)|0);
    $$1 = 0;
    STACKTOP = sp;return ($$1|0);
   }
  }
 } while(0);
 $55 = (_SDL_AllocBlitMap()|0);
 $56 = ((($2)) + 52|0);
 HEAP32[$56>>2] = $55;
 $57 = ($55|0)==(0|0);
 if ($57) {
  _SDL_FreeSurface($2);
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $58 = ($Amask|0)==(0);
 if (!($58)) {
  $59 = ((($55)) + 68|0);
  $60 = HEAP32[$59>>2]|0;
  $61 = $60 & -113;
  $62 = $61 | 16;
  HEAP32[$59>>2] = $62;
  $63 = ($62|0)==($60|0);
  if (!($63)) {
   _SDL_InvalidateMap($55);
  }
 }
 $64 = ((($2)) + 56|0);
 HEAP32[$64>>2] = 1;
 $$1 = $2;
 STACKTOP = sp;return ($$1|0);
}
function _SDL_MasksToPixelFormatEnum($bpp,$Rmask,$Gmask,$Bmask,$Amask) {
 $bpp = $bpp|0;
 $Rmask = $Rmask|0;
 $Gmask = $Gmask|0;
 $Bmask = $Bmask|0;
 $Amask = $Amask|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cond = 0, $or$cond = 0, $or$cond101 = 0, $or$cond103 = 0;
 var $or$cond105 = 0, $or$cond107 = 0, $or$cond11 = 0, $or$cond113 = 0, $or$cond119 = 0, $or$cond125 = 0, $or$cond13 = 0, $or$cond131 = 0, $or$cond133 = 0, $or$cond135 = 0, $or$cond137 = 0, $or$cond15 = 0, $or$cond17 = 0, $or$cond19 = 0, $or$cond21 = 0, $or$cond23 = 0, $or$cond25 = 0, $or$cond27 = 0, $or$cond29 = 0, $or$cond3 = 0;
 var $or$cond31 = 0, $or$cond33 = 0, $or$cond35 = 0, $or$cond37 = 0, $or$cond39 = 0, $or$cond41 = 0, $or$cond43 = 0, $or$cond45 = 0, $or$cond47 = 0, $or$cond5 = 0, $or$cond53 = 0, $or$cond55 = 0, $or$cond57 = 0, $or$cond59 = 0, $or$cond65 = 0, $or$cond67 = 0, $or$cond69 = 0, $or$cond7 = 0, $or$cond71 = 0, $or$cond73 = 0;
 var $or$cond75 = 0, $or$cond77 = 0, $or$cond79 = 0, $or$cond81 = 0, $or$cond83 = 0, $or$cond85 = 0, $or$cond87 = 0, $or$cond89 = 0, $or$cond9 = 0, $or$cond91 = 0, $or$cond93 = 0, $or$cond95 = 0, $or$cond97 = 0, $or$cond99 = 0, $switch$split12D = 0, $switch$split2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 L1: do {
  switch ($bpp|0) {
  case 4:  {
   $$0 = 304088064;
   return ($$0|0);
   break;
  }
  case 8:  {
   $0 = ($Rmask|0)==(0);
   if ($0) {
    $$0 = 318769153;
    return ($$0|0);
   }
   $1 = ($Rmask|0)==(224);
   $2 = ($Gmask|0)==(28);
   $or$cond = $1 & $2;
   $3 = ($Bmask|0)==(3);
   $or$cond3 = $or$cond & $3;
   $4 = ($Amask|0)==(0);
   $or$cond5 = $or$cond3 & $4;
   if ($or$cond5) {
    $$0 = 336660481;
    return ($$0|0);
   }
   break;
  }
  case 12:  {
   $5 = ($Rmask|0)==(0);
   if ($5) {
    $$0 = 353504258;
    return ($$0|0);
   }
   $6 = ($Rmask|0)==(3840);
   $7 = ($Gmask|0)==(240);
   $or$cond7 = $6 & $7;
   $8 = ($Bmask|0)==(15);
   $or$cond9 = $or$cond7 & $8;
   $9 = ($Amask|0)==(0);
   $or$cond11 = $or$cond9 & $9;
   if ($or$cond11) {
    $$0 = 353504258;
    return ($$0|0);
   }
   break;
  }
  case 15:  {
   $cond = ($Rmask|0)==(0);
   if ($cond) {
    $$0 = 353570562;
    return ($$0|0);
   } else {
    label = 9;
   }
   break;
  }
  case 16:  {
   $10 = ($Rmask|0)==(0);
   if ($10) {
    $$0 = 353701890;
    return ($$0|0);
   } else {
    label = 9;
   }
   break;
  }
  case 24:  {
   $switch$split2D = ($Rmask|0)<(255);
   if ($switch$split2D) {
    switch ($Rmask|0) {
    case 0:  {
     $$0 = 390076419;
     break;
    }
    default: {
     label = 24;
     break L1;
    }
    }
    return ($$0|0);
   }
   $switch$split12D = ($Rmask|0)<(16711680);
   if (!($switch$split12D)) {
    switch ($Rmask|0) {
    case 16711680:  {
     $$0 = 390076419;
     break;
    }
    default: {
     label = 24;
     break L1;
    }
    }
    return ($$0|0);
   }
   switch ($Rmask|0) {
   case 255:  {
    break;
   }
   default: {
    label = 24;
    break L1;
   }
   }
   $$0 = 386930691;
   return ($$0|0);
   break;
  }
  case 32:  {
   $37 = ($Rmask|0)==(0);
   if ($37) {
    $$0 = 370546692;
    return ($$0|0);
   } else {
    label = 24;
   }
   break;
  }
  case 1:  {
   $$0 = 287310080;
   return ($$0|0);
   break;
  }
  default: {
  }
  }
 } while(0);
 if ((label|0) == 9) {
  $11 = ($Rmask|0)==(31744);
  $12 = ($Gmask|0)==(992);
  $or$cond13 = $11 & $12;
  $13 = ($Bmask|0)==(31);
  $or$cond15 = $or$cond13 & $13;
  $14 = ($Amask|0)==(0);
  $or$cond17 = $or$cond15 & $14;
  if ($or$cond17) {
   $$0 = 353570562;
   return ($$0|0);
  }
  $15 = ($Rmask|0)==(31);
  $or$cond19 = $15 & $12;
  $16 = ($Bmask|0)==(31744);
  $or$cond21 = $or$cond19 & $16;
  $or$cond23 = $or$cond21 & $14;
  if ($or$cond23) {
   $$0 = 357764866;
   return ($$0|0);
  }
  $17 = ($Rmask|0)==(3840);
  $18 = ($Gmask|0)==(240);
  $or$cond25 = $17 & $18;
  $19 = ($Bmask|0)==(15);
  $or$cond27 = $or$cond25 & $19;
  $20 = ($Amask|0)==(61440);
  $or$cond29 = $or$cond27 & $20;
  if ($or$cond29) {
   $$0 = 355602434;
   return ($$0|0);
  }
  $21 = ($Rmask|0)==(61440);
  $22 = ($Gmask|0)==(3840);
  $or$cond31 = $21 & $22;
  $23 = ($Bmask|0)==(240);
  $or$cond33 = $or$cond31 & $23;
  $24 = ($Amask|0)==(15);
  $or$cond35 = $or$cond33 & $24;
  if ($or$cond35) {
   $$0 = 356651010;
   return ($$0|0);
  }
  $25 = ($Rmask|0)==(15);
  $or$cond37 = $25 & $18;
  $26 = ($Bmask|0)==(3840);
  $or$cond39 = $or$cond37 & $26;
  $or$cond41 = $or$cond39 & $20;
  if ($or$cond41) {
   $$0 = 359796738;
   return ($$0|0);
  }
  $27 = ($Rmask|0)==(240);
  $or$cond43 = $27 & $22;
  $28 = ($Bmask|0)==(61440);
  $or$cond45 = $or$cond43 & $28;
  $or$cond47 = $or$cond45 & $24;
  if ($or$cond47) {
   $$0 = 360845314;
   return ($$0|0);
  }
  $29 = ($Amask|0)==(32768);
  $or$cond53 = $or$cond15 & $29;
  if ($or$cond53) {
   $$0 = 355667970;
   return ($$0|0);
  }
  $30 = ($Rmask|0)==(63488);
  $31 = ($Gmask|0)==(1984);
  $or$cond55 = $30 & $31;
  $32 = ($Bmask|0)==(62);
  $or$cond57 = $or$cond55 & $32;
  $33 = ($Amask|0)==(1);
  $or$cond59 = $or$cond57 & $33;
  if ($or$cond59) {
   $$0 = 356782082;
   return ($$0|0);
  }
  $or$cond65 = $or$cond21 & $29;
  if ($or$cond65) {
   $$0 = 359862274;
   return ($$0|0);
  }
  $34 = ($Rmask|0)==(62);
  $or$cond67 = $34 & $31;
  $35 = ($Bmask|0)==(63488);
  $or$cond69 = $or$cond67 & $35;
  $or$cond71 = $or$cond69 & $33;
  if ($or$cond71) {
   $$0 = 360976386;
   return ($$0|0);
  }
  $36 = ($Gmask|0)==(2016);
  $or$cond73 = $30 & $36;
  $or$cond75 = $or$cond73 & $13;
  $or$cond77 = $or$cond75 & $14;
  if ($or$cond77) {
   $$0 = 353701890;
   return ($$0|0);
  }
  $or$cond79 = $15 & $36;
  $or$cond81 = $or$cond79 & $35;
  $or$cond83 = $or$cond81 & $14;
  if ($or$cond83) {
   $$0 = 357896194;
   return ($$0|0);
  }
 }
 else if ((label|0) == 24) {
  $38 = ($Rmask|0)==(16711680);
  $39 = ($Gmask|0)==(65280);
  $or$cond85 = $38 & $39;
  $40 = ($Bmask|0)==(255);
  $or$cond87 = $or$cond85 & $40;
  $41 = ($Amask|0)==(0);
  $or$cond89 = $or$cond87 & $41;
  if ($or$cond89) {
   $$0 = 370546692;
   return ($$0|0);
  }
  $42 = ($Rmask|0)==(-16777216);
  $43 = ($Gmask|0)==(16711680);
  $or$cond91 = $42 & $43;
  $44 = ($Bmask|0)==(65280);
  $or$cond93 = $or$cond91 & $44;
  $or$cond95 = $or$cond93 & $41;
  if ($or$cond95) {
   $$0 = 371595268;
   return ($$0|0);
  }
  $45 = ($Rmask|0)==(255);
  $or$cond97 = $45 & $39;
  $46 = ($Bmask|0)==(16711680);
  $or$cond99 = $or$cond97 & $46;
  $or$cond101 = $or$cond99 & $41;
  if ($or$cond101) {
   $$0 = 374740996;
   return ($$0|0);
  }
  $47 = ($Rmask|0)==(65280);
  $or$cond103 = $47 & $43;
  $48 = ($Bmask|0)==(-16777216);
  $or$cond105 = $or$cond103 & $48;
  $or$cond107 = $or$cond105 & $41;
  if ($or$cond107) {
   $$0 = 375789572;
   return ($$0|0);
  }
  $49 = ($Amask|0)==(-16777216);
  $or$cond113 = $or$cond87 & $49;
  if ($or$cond113) {
   $$0 = 372645892;
   return ($$0|0);
  }
  $50 = ($Amask|0)==(255);
  $or$cond119 = $or$cond93 & $50;
  if ($or$cond119) {
   $$0 = 373694468;
   return ($$0|0);
  }
  $or$cond125 = $or$cond99 & $49;
  if ($or$cond125) {
   $$0 = 376840196;
   return ($$0|0);
  }
  $or$cond131 = $or$cond105 & $50;
  if ($or$cond131) {
   $$0 = 377888772;
   return ($$0|0);
  }
  $51 = ($Rmask|0)==(1072693248);
  $52 = ($Gmask|0)==(1047552);
  $or$cond133 = $51 & $52;
  $53 = ($Bmask|0)==(1023);
  $or$cond135 = $or$cond133 & $53;
  $54 = ($Amask|0)==(-1073741824);
  $or$cond137 = $or$cond135 & $54;
  if ($or$cond137) {
   $$0 = 372711428;
   return ($$0|0);
  }
 }
 $$0 = 0;
 return ($$0|0);
}
function _SDL_AllocFormat($pixel_format) {
 $pixel_format = $pixel_format|0;
 var $$0 = 0, $$mask = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $format$0 = 0, $format$03 = 0, $format$04 = 0, $format$04$lcssa = 0, $or$cond = 0, $switch = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $format$03 = HEAP32[300737]|0;
 $0 = ($format$03|0)==(0|0);
 L1: do {
  if (!($0)) {
   $format$04 = $format$03;
   while(1) {
    $1 = HEAP32[$format$04>>2]|0;
    $2 = ($1|0)==($pixel_format|0);
    if ($2) {
     $format$04$lcssa = $format$04;
     break;
    }
    $6 = ((($format$04)) + 40|0);
    $format$0 = HEAP32[$6>>2]|0;
    $7 = ($format$0|0)==(0|0);
    if ($7) {
     break L1;
    } else {
     $format$04 = $format$0;
    }
   }
   $3 = ((($format$04$lcssa)) + 36|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = (($4) + 1)|0;
   HEAP32[$3>>2] = $5;
   $$0 = $format$04$lcssa;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 $8 = (_SDL_malloc(44)|0);
 $9 = ($8|0)==(0|0);
 if ($9) {
  (_SDL_Error(0)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $10 = (_SDL_InitFormat($8,$pixel_format)|0);
 $11 = ($10|0)<(0);
 if ($11) {
  _SDL_free($8);
  HEAP32[$vararg_buffer>>2] = 597583;
  (_SDL_SetError(611267,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $12 = ($pixel_format|0)==(0);
 $$mask = $pixel_format & -268435456;
 $13 = ($$mask|0)==(268435456);
 $or$cond = $12 | $13;
 if ($or$cond) {
  $14 = $pixel_format >>> 24;
  $15 = $14 & 15;
  $$off = (($15) + -1)|0;
  $switch = ($$off>>>0)<(3);
  if ($switch) {
   $$0 = $8;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $16 = HEAP32[300737]|0;
 $17 = ((($8)) + 40|0);
 HEAP32[$17>>2] = $16;
 HEAP32[300737] = $8;
 $$0 = $8;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_CalculatePitch($surface) {
 $surface = $surface|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $pitch$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($surface)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($surface)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 9|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $7 = Math_imul($6, $1)|0;
 $8 = ((($3)) + 8|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 switch ($10|0) {
 case 1:  {
  $11 = (($7) + 7)|0;
  $12 = (($11|0) / 8)&-1;
  $pitch$0 = $12;
  break;
 }
 case 4:  {
  $13 = (($7) + 1)|0;
  $14 = (($13|0) / 2)&-1;
  $pitch$0 = $14;
  break;
 }
 default: {
  $pitch$0 = $7;
 }
 }
 $15 = (($pitch$0) + 3)|0;
 $16 = $15 & -4;
 return ($16|0);
}
function _SDL_AllocPalette($ncolors) {
 $ncolors = $ncolors|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($ncolors|0)<(1);
 if ($0) {
  HEAP32[$vararg_buffer>>2] = 603377;
  (_SDL_SetError(611267,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = (_SDL_malloc(16)|0);
 $2 = ($1|0)==(0|0);
 if ($2) {
  (_SDL_Error(0)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = $ncolors << 2;
 $4 = (_SDL_malloc($3)|0);
 $5 = ((($1)) + 4|0);
 HEAP32[$5>>2] = $4;
 $6 = ($4|0)==(0|0);
 if ($6) {
  _SDL_free($1);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  HEAP32[$1>>2] = $ncolors;
  $7 = ((($1)) + 8|0);
  HEAP32[$7>>2] = 1;
  $8 = ((($1)) + 12|0);
  HEAP32[$8>>2] = 1;
  (_SDL_memset($4,255,$3)|0);
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_FreePalette($palette) {
 $palette = $palette|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($palette|0)==(0|0);
 if ($0) {
  HEAP32[$vararg_buffer>>2] = 603385;
  (_SDL_SetError(611267,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $1 = ((($palette)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($2) + -1)|0;
 HEAP32[$1>>2] = $3;
 $4 = ($2|0)>(1);
 if ($4) {
  STACKTOP = sp;return;
 }
 $5 = ((($palette)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 _SDL_free($6);
 _SDL_free($palette);
 STACKTOP = sp;return;
}
function _SDL_AllocBlitMap() {
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_calloc(1,88)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_Error(0)|0);
  $$0 = 0;
  return ($$0|0);
 } else {
  $2 = ((($0)) + 76|0);
  HEAP32[$2>>2] = -1;
  $$0 = $0;
  return ($$0|0);
 }
 return (0)|0;
}
function _SW_CreateRendererForSurface($surface) {
 $surface = $surface|0;
 var $$0 = 0, $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($surface|0)==(0|0);
 if ($0) {
  (_SDL_SetError(603393,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = (_SDL_calloc(1,320)|0);
 $2 = ($1|0)==(0|0);
 if ($2) {
  (_SDL_Error(0)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = (_SDL_calloc(1,8)|0);
 $4 = ($3|0)==(0|0);
 if ($4) {
  $5 = ((($1)) + 316|0);
  $6 = HEAP32[$5>>2]|0;
  _SDL_free($6);
  _SDL_free($1);
  (_SDL_Error(0)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$3>>2] = $surface;
 $7 = ((($3)) + 4|0);
 HEAP32[$7>>2] = $surface;
 $8 = ((($1)) + 4|0);
 HEAP32[$8>>2] = 4520;
 $9 = ((($1)) + 8|0);
 HEAP32[$9>>2] = 4521;
 $10 = ((($1)) + 12|0);
 HEAP32[$10>>2] = 4522;
 $11 = ((($1)) + 16|0);
 HEAP32[$11>>2] = 4523;
 $12 = ((($1)) + 20|0);
 HEAP32[$12>>2] = 4524;
 $13 = ((($1)) + 24|0);
 HEAP32[$13>>2] = 4525;
 $14 = ((($1)) + 28|0);
 HEAP32[$14>>2] = 4526;
 $15 = ((($1)) + 36|0);
 HEAP32[$15>>2] = 4527;
 $16 = ((($1)) + 40|0);
 HEAP32[$16>>2] = 4528;
 $17 = ((($1)) + 44|0);
 HEAP32[$17>>2] = 4529;
 $18 = ((($1)) + 48|0);
 HEAP32[$18>>2] = 4530;
 $19 = ((($1)) + 52|0);
 HEAP32[$19>>2] = 4531;
 $20 = ((($1)) + 56|0);
 HEAP32[$20>>2] = 4532;
 $21 = ((($1)) + 60|0);
 HEAP32[$21>>2] = 4533;
 $22 = ((($1)) + 64|0);
 HEAP32[$22>>2] = 4534;
 $23 = ((($1)) + 68|0);
 HEAP32[$23>>2] = 4535;
 $24 = ((($1)) + 72|0);
 HEAP32[$24>>2] = 4536;
 $25 = ((($1)) + 76|0);
 HEAP32[$25>>2] = 4537;
 $26 = ((($1)) + 80|0);
 HEAP32[$26>>2] = 4538;
 $27 = ((($1)) + 84|0);
 HEAP32[$27>>2] = 4539;
 $28 = ((($1)) + 88|0);
 HEAP32[$28>>2] = 4540;
 $29 = ((($1)) + 92|0);
 HEAP32[$29>>2] = 4541;
 $30 = ((($1)) + 104|0);
 dest=$30; src=(137840); stop=dest+84|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 $31 = ((($1)) + 316|0);
 HEAP32[$31>>2] = $3;
 $32 = HEAP32[$3>>2]|0;
 $33 = ($32|0)==(0|0);
 if (!($33)) {
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $34 = HEAP32[$7>>2]|0;
 HEAP32[$3>>2] = $34;
 $35 = ($34|0)==(0);
 if (!($35)) {
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $36 = ((($1)) + 188|0);
 $37 = HEAP32[$36>>2]|0;
 $38 = (_SDL_GetWindowSurface($37)|0);
 $39 = ($38|0)==(0|0);
 if ($39) {
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$7>>2] = $38;
 HEAP32[$3>>2] = $38;
 $40 = HEAP32[$31>>2]|0;
 $41 = HEAP32[$40>>2]|0;
 $42 = ($41|0)==(0|0);
 if ($42) {
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $43 = ((($1)) + 212|0);
 (_SDL_SetClipRect($41,$43)|0);
 $$pre = HEAP32[$31>>2]|0;
 $$pre1 = HEAP32[$$pre>>2]|0;
 $44 = ($$pre1|0)==(0|0);
 if ($44) {
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $45 = ((($1)) + 276|0);
 $46 = HEAP32[$45>>2]|0;
 $47 = ($46|0)==(0);
 if ($47) {
  (_SDL_SetClipRect($$pre1,0)|0);
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 } else {
  $48 = ((($1)) + 244|0);
  (_SDL_SetClipRect($$pre1,$48)|0);
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SW_WindowEvent($renderer,$event) {
 $renderer = $renderer|0;
 $event = $event|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($event)) + 12|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = ($3<<24>>24)==(6);
 if (!($4)) {
  return;
 }
 HEAP32[$1>>2] = 0;
 $5 = ((($1)) + 4|0);
 HEAP32[$5>>2] = 0;
 return;
}
function _SW_GetOutputSize($renderer,$w,$h) {
 $renderer = $renderer|0;
 $w = $w|0;
 $h = $h|0;
 var $$0 = 0, $$pr = 0, $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $4 = ((($1)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  HEAP32[$1>>2] = $5;
  $6 = ($5|0)==(0);
  $7 = $5;
  if ($6) {
   $8 = ((($renderer)) + 188|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (_SDL_GetWindowSurface($9)|0);
   $11 = ($10|0)==(0|0);
   do {
    if (!($11)) {
     HEAP32[$4>>2] = $10;
     HEAP32[$1>>2] = $10;
     $12 = HEAP32[$0>>2]|0;
     $13 = HEAP32[$12>>2]|0;
     $14 = ($13|0)==(0|0);
     if (!($14)) {
      $15 = ((($renderer)) + 212|0);
      (_SDL_SetClipRect($13,$15)|0);
      $$pre = HEAP32[$0>>2]|0;
      $$pre1 = HEAP32[$$pre>>2]|0;
      $16 = ($$pre1|0)==(0|0);
      if (!($16)) {
       $17 = ((($renderer)) + 276|0);
       $18 = HEAP32[$17>>2]|0;
       $19 = ($18|0)==(0);
       if ($19) {
        (_SDL_SetClipRect($$pre1,0)|0);
        break;
       } else {
        $20 = ((($renderer)) + 244|0);
        (_SDL_SetClipRect($$pre1,$20)|0);
        break;
       }
      }
     }
    }
   } while(0);
   $$pr = HEAP32[$1>>2]|0;
   $22 = $$pr;
  } else {
   $22 = $7;
  }
  $21 = ($22|0)==(0|0);
  if ($21) {
   (_SDL_SetError(603432,$vararg_buffer)|0);
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  } else {
   $25 = $22;
  }
 } else {
  $25 = $2;
 }
 $23 = ($w|0)==(0|0);
 if (!($23)) {
  $24 = ((($25)) + 8|0);
  $26 = HEAP32[$24>>2]|0;
  HEAP32[$w>>2] = $26;
 }
 $27 = ($h|0)==(0|0);
 if ($27) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $28 = ((($25)) + 12|0);
 $29 = HEAP32[$28>>2]|0;
 HEAP32[$h>>2] = $29;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SetClipRect($surface,$rect) {
 $surface = $surface|0;
 $rect = $rect|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $full_rect = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $full_rect = sp;
 $0 = ($surface|0)==(0|0);
 if ($0) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$full_rect>>2] = 0;
 $1 = ((($full_rect)) + 4|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($surface)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($full_rect)) + 8|0);
 HEAP32[$4>>2] = $3;
 $5 = ((($surface)) + 12|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($full_rect)) + 12|0);
 HEAP32[$7>>2] = $6;
 $8 = ($rect|0)==(0|0);
 $9 = ((($surface)) + 36|0);
 if ($8) {
  ;HEAP32[$9>>2]=HEAP32[$full_rect>>2]|0;HEAP32[$9+4>>2]=HEAP32[$full_rect+4>>2]|0;HEAP32[$9+8>>2]=HEAP32[$full_rect+8>>2]|0;HEAP32[$9+12>>2]=HEAP32[$full_rect+12>>2]|0;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 } else {
  $10 = (_SDL_IntersectRect($rect,$full_rect,$9)|0);
  $$0 = $10;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SW_CreateTexture($renderer,$texture) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $Amask = 0, $Bmask = 0;
 var $Gmask = 0, $Rmask = 0, $bpp = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $bpp = sp + 20|0;
 $Rmask = sp + 16|0;
 $Gmask = sp + 12|0;
 $Bmask = sp + 8|0;
 $Amask = sp + 4|0;
 $0 = ((($texture)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_SDL_PixelFormatEnumToMasks($1,$bpp,$Rmask,$Gmask,$Bmask,$Amask)|0);
 $3 = ($2|0)==(0);
 if ($3) {
  $4 = (_SDL_SetError(603481,$vararg_buffer)|0);
  $$0 = $4;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = ((($texture)) + 12|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($texture)) + 16|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = HEAP32[$bpp>>2]|0;
 $10 = HEAP32[$Rmask>>2]|0;
 $11 = HEAP32[$Gmask>>2]|0;
 $12 = HEAP32[$Bmask>>2]|0;
 $13 = HEAP32[$Amask>>2]|0;
 $14 = (_SDL_CreateRGBSurface(0,$6,$8,$9,$10,$11,$12,$13)|0);
 $15 = ((($texture)) + 68|0);
 HEAP32[$15>>2] = $14;
 $16 = ((($texture)) + 28|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17&65535;
 $19 = $17&255;
 $20 = ($18&65535) >>> 8;
 $21 = $20&255;
 $22 = $17 >>> 16;
 $23 = $22&255;
 (_SDL_SetSurfaceColorMod($14,$19,$21,$23)|0);
 $24 = HEAP32[$15>>2]|0;
 $25 = ((($texture)) + 31|0);
 $26 = HEAP8[$25>>0]|0;
 (_SDL_SetSurfaceAlphaMod($24,$26)|0);
 $27 = HEAP32[$15>>2]|0;
 $28 = ((($texture)) + 24|0);
 $29 = HEAP32[$28>>2]|0;
 (_SDL_SetSurfaceBlendMode($27,$29)|0);
 $30 = ((($texture)) + 8|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = ($31|0)==(0);
 if ($32) {
  $33 = HEAP32[$15>>2]|0;
  (_SDL_SetSurfaceRLE($33,1)|0);
 }
 $34 = HEAP32[$15>>2]|0;
 $35 = ($34|0)==(0|0);
 $$ = $35 << 31 >> 31;
 $$0 = $$;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SetSurfaceColorMod($surface,$r,$g,$b) {
 $surface = $surface|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond1 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($surface|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ((($surface)) + 52|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 76|0);
 HEAP8[$3>>0] = $r;
 $4 = ((($2)) + 77|0);
 HEAP8[$4>>0] = $g;
 $5 = ((($2)) + 78|0);
 HEAP8[$5>>0] = $b;
 $6 = ((($2)) + 68|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $g & $r;
 $9 = ($8<<24>>24)==(-1);
 $10 = ($b<<24>>24)==(-1);
 $or$cond1 = $9 & $10;
 $11 = $7 | 1;
 $12 = $7 & -2;
 $storemerge = $or$cond1 ? $12 : $11;
 HEAP32[$6>>2] = $storemerge;
 $13 = ($storemerge|0)==($7|0);
 if ($13) {
  $$0 = 0;
  return ($$0|0);
 }
 _SDL_InvalidateMap($2);
 $$0 = 0;
 return ($$0|0);
}
function _SDL_SetSurfaceAlphaMod($surface,$alpha) {
 $surface = $surface|0;
 $alpha = $alpha|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($surface|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ((($surface)) + 52|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 79|0);
 HEAP8[$3>>0] = $alpha;
 $4 = ((($2)) + 68|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($alpha<<24>>24)==(-1);
 $7 = $5 & -3;
 $8 = $5 | 2;
 $storemerge = $6 ? $7 : $8;
 HEAP32[$4>>2] = $storemerge;
 $9 = ($storemerge|0)==($5|0);
 if ($9) {
  $$0 = 0;
  return ($$0|0);
 }
 _SDL_InvalidateMap($2);
 $$0 = 0;
 return ($$0|0);
}
function _SDL_SetSurfaceBlendMode($surface,$blendMode) {
 $surface = $surface|0;
 $blendMode = $blendMode|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $status$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($surface|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ((($surface)) + 52|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 68|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 & -113;
 HEAP32[$3>>2] = $5;
 switch ($blendMode|0) {
 case 0:  {
  $11 = $5;$12 = $2;$status$0 = 0;
  break;
 }
 case 1:  {
  $6 = $5 | 16;
  HEAP32[$3>>2] = $6;
  $11 = $6;$12 = $2;$status$0 = 0;
  break;
 }
 case 2:  {
  $7 = $5 | 32;
  HEAP32[$3>>2] = $7;
  $11 = $7;$12 = $2;$status$0 = 0;
  break;
 }
 case 4:  {
  $8 = $5 | 64;
  HEAP32[$3>>2] = $8;
  $11 = $8;$12 = $2;$status$0 = 0;
  break;
 }
 default: {
  $9 = (_SDL_Error(4)|0);
  $$pre = HEAP32[$1>>2]|0;
  $$phi$trans$insert = ((($$pre)) + 68|0);
  $$pre1 = HEAP32[$$phi$trans$insert>>2]|0;
  $11 = $$pre1;$12 = $$pre;$status$0 = $9;
 }
 }
 $10 = ($11|0)==($4|0);
 if ($10) {
  $$0 = $status$0;
  return ($$0|0);
 }
 _SDL_InvalidateMap($12);
 $$0 = $status$0;
 return ($$0|0);
}
function _SDL_SetSurfaceRLE($surface,$flag) {
 $surface = $surface|0;
 $flag = $flag|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($surface|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ((($surface)) + 52|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 68|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($flag|0)==(0);
 $6 = $4 & -4097;
 $7 = $4 | 4096;
 $storemerge = $5 ? $6 : $7;
 HEAP32[$3>>2] = $storemerge;
 $8 = ($storemerge|0)==($4|0);
 if ($8) {
  $$0 = 0;
  return ($$0|0);
 }
 _SDL_InvalidateMap($2);
 $$0 = 0;
 return ($$0|0);
}
function _SW_SetTextureColorMod($renderer,$texture) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 var $$in = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($texture)) + 68|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($texture)) + 28|0);
 $3 = HEAP16[$2>>1]|0;
 $4 = HEAP32[$2>>2]|0;
 $5 = $4&65535;
 $6 = $4&255;
 $7 = ($5&65535) >>> 8;
 $8 = $7&255;
 $9 = $8 & $6;
 $10 = $4 >>> 16;
 $11 = $10&255;
 $12 = $9 & $11;
 $13 = ($12<<24>>24)==(-1);
 $14 = ($3&65535) >>> 8;
 $15 = $14&255;
 if ($13) {
  $$in = $4;$22 = $15;$23 = $11;
  $21 = $$in&255;
  $24 = (_SDL_SetSurfaceColorMod($1,$21,$22,$23)|0);
  return ($24|0);
 }
 (_SDL_SetSurfaceRLE($1,0)|0);
 $$pre = HEAP32[$2>>2]|0;
 $16 = $$pre&65535;
 $17 = ($16&65535) >>> 8;
 $18 = $17&255;
 $19 = $$pre >>> 16;
 $20 = $19&255;
 $$in = $$pre;$22 = $18;$23 = $20;
 $21 = $$in&255;
 $24 = (_SDL_SetSurfaceColorMod($1,$21,$22,$23)|0);
 return ($24|0);
}
function _SW_SetTextureAlphaMod($renderer,$texture) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($texture)) + 68|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($texture)) + 31|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = ($3<<24>>24)==(-1);
 if ($4) {
  $10 = -1;
 } else {
  $5 = ((($1)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ((($6)) + 24|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8|0)==(0);
  if ($9) {
   $10 = $3;
  } else {
   (_SDL_SetSurfaceRLE($1,0)|0);
   $$pre = HEAP8[$2>>0]|0;
   $10 = $$pre;
  }
 }
 $11 = (_SDL_SetSurfaceAlphaMod($1,$10)|0);
 return ($11|0);
}
function _SW_SetTextureBlendMode($renderer,$texture) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 var $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($texture)) + 68|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($texture)) + 24|0);
 $3 = HEAP32[$2>>2]|0;
 switch ($3|0) {
 case 4: case 2:  {
  (_SDL_SetSurfaceRLE($1,0)|0);
  $$pre = HEAP32[$2>>2]|0;
  $4 = $$pre;
  break;
 }
 default: {
  $4 = $3;
 }
 }
 $5 = (_SDL_SetSurfaceBlendMode($1,$4)|0);
 return ($5|0);
}
function _SW_UpdateTexture($renderer,$texture,$rect,$pixels,$pitch) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 $rect = $rect|0;
 $pixels = $pixels|0;
 $pitch = $pitch|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst$01 = 0, $row$02 = 0, $src$03 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($texture)) + 68|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = $2 & 2;
 $4 = ($3|0)==(0);
 if (!($4)) {
  (_SDL_LockSurface($1)|0);
 }
 $5 = ((($1)) + 16|0);
 $6 = ((($1)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($7)) + 9|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 $11 = ((($rect)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = Math_imul($12, $10)|0;
 $14 = ((($rect)) + 12|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)>(0);
 if ($16) {
  $17 = ((($1)) + 20|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = HEAP32[$5>>2]|0;
  $20 = ((($rect)) + 4|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = Math_imul($19, $21)|0;
  $23 = (($18) + ($22)|0);
  $24 = HEAP32[$rect>>2]|0;
  $25 = Math_imul($10, $24)|0;
  $26 = (($23) + ($25)|0);
  $dst$01 = $26;$row$02 = 0;$src$03 = $pixels;
  while(1) {
   (_SDL_memcpy($dst$01,$src$03,$13)|0);
   $27 = (($src$03) + ($pitch)|0);
   $28 = HEAP32[$5>>2]|0;
   $29 = (($dst$01) + ($28)|0);
   $30 = (($row$02) + 1)|0;
   $31 = HEAP32[$14>>2]|0;
   $32 = ($30|0)<($31|0);
   if ($32) {
    $dst$01 = $29;$row$02 = $30;$src$03 = $27;
   } else {
    break;
   }
  }
 }
 $33 = HEAP32[$1>>2]|0;
 $34 = $33 & 2;
 $35 = ($34|0)==(0);
 if ($35) {
  return 0;
 }
 _SDL_UnlockSurface($1);
 return 0;
}
function _SW_LockTexture($renderer,$texture,$rect,$pixels,$pitch) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 $rect = $rect|0;
 $pixels = $pixels|0;
 $pitch = $pitch|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($texture)) + 68|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 20|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($rect)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($1)) + 16|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = Math_imul($7, $5)|0;
 $9 = (($3) + ($8)|0);
 $10 = HEAP32[$rect>>2]|0;
 $11 = ((($1)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($12)) + 9|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = $14&255;
 $16 = Math_imul($15, $10)|0;
 $17 = (($9) + ($16)|0);
 HEAP32[$pixels>>2] = $17;
 HEAP32[$pitch>>2] = $7;
 return 0;
}
function _SW_UnlockTexture($renderer,$texture) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _SW_SetRenderTarget($renderer,$texture) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($texture|0)==(0|0);
 if ($2) {
  $5 = ((($1)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  HEAP32[$1>>2] = $6;
  return 0;
 } else {
  $3 = ((($texture)) + 68|0);
  $4 = HEAP32[$3>>2]|0;
  HEAP32[$1>>2] = $4;
  return 0;
 }
 return (0)|0;
}
function _SW_UpdateViewport($renderer) {
 $renderer = $renderer|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  return 0;
 }
 $4 = ((($renderer)) + 212|0);
 (_SDL_SetClipRect($2,$4)|0);
 return 0;
}
function _SW_UpdateClipRect($renderer) {
 $renderer = $renderer|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  return 0;
 }
 $4 = ((($renderer)) + 276|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0);
 if ($6) {
  (_SDL_SetClipRect($2,0)|0);
  return 0;
 } else {
  $7 = ((($renderer)) + 244|0);
  (_SDL_SetClipRect($2,$7)|0);
  return 0;
 }
 return (0)|0;
}
function _SW_RenderClear($renderer) {
 $renderer = $renderer|0;
 var $$0 = 0, $$pr = 0, $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $clip_rect = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $clip_rect = sp;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $4 = ((($1)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  HEAP32[$1>>2] = $5;
  $6 = ($5|0)==(0);
  $7 = $5;
  if ($6) {
   $8 = ((($renderer)) + 188|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (_SDL_GetWindowSurface($9)|0);
   $11 = ($10|0)==(0|0);
   do {
    if (!($11)) {
     HEAP32[$4>>2] = $10;
     HEAP32[$1>>2] = $10;
     $12 = HEAP32[$0>>2]|0;
     $13 = HEAP32[$12>>2]|0;
     $14 = ($13|0)==(0|0);
     if (!($14)) {
      $15 = ((($renderer)) + 212|0);
      (_SDL_SetClipRect($13,$15)|0);
      $$pre = HEAP32[$0>>2]|0;
      $$pre1 = HEAP32[$$pre>>2]|0;
      $16 = ($$pre1|0)==(0|0);
      if (!($16)) {
       $17 = ((($renderer)) + 276|0);
       $18 = HEAP32[$17>>2]|0;
       $19 = ($18|0)==(0);
       if ($19) {
        (_SDL_SetClipRect($$pre1,0)|0);
        break;
       } else {
        $20 = ((($renderer)) + 244|0);
        (_SDL_SetClipRect($$pre1,$20)|0);
        break;
       }
      }
     }
    }
   } while(0);
   $$pr = HEAP32[$1>>2]|0;
   $22 = $$pr;
  } else {
   $22 = $7;
  }
  $21 = ($22|0)==(0|0);
  if ($21) {
   $$0 = -1;$39 = $clip_rect;
   STACKTOP = sp;return ($$0|0);
  } else {
   $24 = $22;$38 = $clip_rect;
  }
 } else {
  $24 = $2;$38 = $clip_rect;
 }
 $23 = ((($24)) + 4|0);
 $25 = HEAP32[$23>>2]|0;
 $26 = ((($renderer)) + 308|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = $27&65535;
 $29 = $27&255;
 $30 = ($28&65535) >>> 8;
 $31 = $30&255;
 $32 = $27 >>> 16;
 $33 = $32&255;
 $34 = $27 >>> 24;
 $35 = $34&255;
 $36 = (_SDL_MapRGBA($25,$29,$31,$33,$35)|0);
 $37 = ((($24)) + 36|0);
 ;HEAP32[$38>>2]=HEAP32[$37>>2]|0;HEAP32[$38+4>>2]=HEAP32[$37+4>>2]|0;HEAP32[$38+8>>2]=HEAP32[$37+8>>2]|0;HEAP32[$38+12>>2]=HEAP32[$37+12>>2]|0;
 (_SDL_SetClipRect($24,0)|0);
 (_SDL_FillRect($24,0,$36)|0);
 (_SDL_SetClipRect($24,$clip_rect)|0);
 $$0 = 0;$39 = $38;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_MapRGBA($format,$r,$g,$b,$a) {
 $format = $format|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $8 = 0, $9 = 0, $i$02$i = 0, $pixel$03$i = 0, $pixel$1$i = 0, $pixel$2$i = 0, $smallest$01$i = 0;
 var $smallest$1$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($format)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $3 = $r&255;
  $4 = ((($format)) + 28|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = $5 & 255;
  $7 = $3 >>> $6;
  $8 = ((($format)) + 32|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = $9 & 255;
  $11 = $7 << $10;
  $12 = $g&255;
  $13 = $5 >>> 8;
  $14 = $13 & 255;
  $15 = $12 >>> $14;
  $16 = $9 >>> 8;
  $17 = $16 & 255;
  $18 = $15 << $17;
  $19 = $18 | $11;
  $20 = $b&255;
  $21 = $5 >>> 16;
  $22 = $21 & 255;
  $23 = $20 >>> $22;
  $24 = $9 >>> 16;
  $25 = $24 & 255;
  $26 = $23 << $25;
  $27 = $19 | $26;
  $28 = $a&255;
  $29 = $5 >>> 24;
  $30 = $28 >>> $29;
  $31 = $9 >>> 24;
  $32 = $30 << $31;
  $33 = ((($format)) + 24|0);
  $34 = HEAP32[$33>>2]|0;
  $35 = $32 & $34;
  $36 = $27 | $35;
  $$0 = $36;
  return ($$0|0);
 }
 $37 = HEAP32[$1>>2]|0;
 $38 = ($37|0)>(0);
 L5: do {
  if ($38) {
   $39 = ((($1)) + 4|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = $r&255;
   $42 = $g&255;
   $43 = $b&255;
   $44 = $a&255;
   $i$02$i = 0;$pixel$03$i = 0;$smallest$01$i = -1;
   while(1) {
    $45 = (($40) + ($i$02$i<<2)|0);
    $46 = HEAP8[$45>>0]|0;
    $47 = $46&255;
    $48 = (($47) - ($41))|0;
    $49 = (((($40) + ($i$02$i<<2)|0)) + 1|0);
    $50 = HEAP8[$49>>0]|0;
    $51 = $50&255;
    $52 = (($51) - ($42))|0;
    $53 = (((($40) + ($i$02$i<<2)|0)) + 2|0);
    $54 = HEAP8[$53>>0]|0;
    $55 = $54&255;
    $56 = (($55) - ($43))|0;
    $57 = (((($40) + ($i$02$i<<2)|0)) + 3|0);
    $58 = HEAP8[$57>>0]|0;
    $59 = $58&255;
    $60 = (($59) - ($44))|0;
    $61 = Math_imul($48, $48)|0;
    $62 = Math_imul($52, $52)|0;
    $63 = (($62) + ($61))|0;
    $64 = Math_imul($56, $56)|0;
    $65 = (($63) + ($64))|0;
    $66 = Math_imul($60, $60)|0;
    $67 = (($65) + ($66))|0;
    $68 = ($67>>>0)<($smallest$01$i>>>0);
    if ($68) {
     $69 = $i$02$i&255;
     $70 = ($67|0)==(0);
     if ($70) {
      $pixel$2$i = $69;
      break L5;
     } else {
      $pixel$1$i = $69;$smallest$1$i = $67;
     }
    } else {
     $pixel$1$i = $pixel$03$i;$smallest$1$i = $smallest$01$i;
    }
    $71 = (($i$02$i) + 1)|0;
    $72 = ($71|0)<($37|0);
    if ($72) {
     $i$02$i = $71;$pixel$03$i = $pixel$1$i;$smallest$01$i = $smallest$1$i;
    } else {
     $pixel$2$i = $pixel$1$i;
     break;
    }
   }
  } else {
   $pixel$2$i = 0;
  }
 } while(0);
 $73 = $pixel$2$i&255;
 $$0 = $73;
 return ($$0|0);
}
function _SW_RenderDrawPoints($renderer,$points,$count) {
 $renderer = $renderer|0;
 $points = $points|0;
 $count = $count|0;
 var $$0 = 0, $$alloca_mul = 0, $$pr = 0, $$pre = 0, $$pre7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0.0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $8 = 0;
 var $9 = 0, $exitcond = 0, $exitcond6 = 0, $i$03 = 0, $i$12 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $4 = ((($1)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  HEAP32[$1>>2] = $5;
  $6 = ($5|0)==(0);
  $7 = $5;
  if ($6) {
   $8 = ((($renderer)) + 188|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (_SDL_GetWindowSurface($9)|0);
   $11 = ($10|0)==(0|0);
   do {
    if (!($11)) {
     HEAP32[$4>>2] = $10;
     HEAP32[$1>>2] = $10;
     $12 = HEAP32[$0>>2]|0;
     $13 = HEAP32[$12>>2]|0;
     $14 = ($13|0)==(0|0);
     if (!($14)) {
      $15 = ((($renderer)) + 212|0);
      (_SDL_SetClipRect($13,$15)|0);
      $$pre = HEAP32[$0>>2]|0;
      $$pre7 = HEAP32[$$pre>>2]|0;
      $16 = ($$pre7|0)==(0|0);
      if (!($16)) {
       $17 = ((($renderer)) + 276|0);
       $18 = HEAP32[$17>>2]|0;
       $19 = ($18|0)==(0);
       if ($19) {
        (_SDL_SetClipRect($$pre7,0)|0);
        break;
       } else {
        $20 = ((($renderer)) + 244|0);
        (_SDL_SetClipRect($$pre7,$20)|0);
        break;
       }
      }
     }
    }
   } while(0);
   $$pr = HEAP32[$1>>2]|0;
   $22 = $$pr;
  } else {
   $22 = $7;
  }
  $21 = ($22|0)==(0|0);
  if ($21) {
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  } else {
   $71 = $22;
  }
 } else {
  $71 = $2;
 }
 $23 = $count << 3;
 $$alloca_mul = $23;
 $24 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $25 = ((($renderer)) + 212|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ($26|0)==(0);
 if ($27) {
  $28 = ((($renderer)) + 216|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = ($29|0)==(0);
  if ($30) {
   $31 = ($count|0)>(0);
   if ($31) {
    $i$12 = 0;
    while(1) {
     $48 = (($points) + ($i$12<<3)|0);
     $49 = +HEAPF32[$48>>2];
     $50 = (~~(($49)));
     $51 = (($24) + ($i$12<<3)|0);
     HEAP32[$51>>2] = $50;
     $52 = (((($points) + ($i$12<<3)|0)) + 4|0);
     $53 = +HEAPF32[$52>>2];
     $54 = (~~(($53)));
     $55 = (((($24) + ($i$12<<3)|0)) + 4|0);
     HEAP32[$55>>2] = $54;
     $56 = (($i$12) + 1)|0;
     $exitcond = ($56|0)==($count|0);
     if ($exitcond) {
      break;
     } else {
      $i$12 = $56;
     }
    }
   }
  } else {
   label = 14;
  }
 } else {
  label = 14;
 }
 if ((label|0) == 14) {
  $32 = ($count|0)>(0);
  if ($32) {
   $33 = ((($renderer)) + 216|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = (+($26|0));
   $36 = (+($34|0));
   $i$03 = 0;
   while(1) {
    $37 = (($points) + ($i$03<<3)|0);
    $38 = +HEAPF32[$37>>2];
    $39 = $35 + $38;
    $40 = (~~(($39)));
    $41 = (($24) + ($i$03<<3)|0);
    HEAP32[$41>>2] = $40;
    $42 = (((($points) + ($i$03<<3)|0)) + 4|0);
    $43 = +HEAPF32[$42>>2];
    $44 = $36 + $43;
    $45 = (~~(($44)));
    $46 = (((($24) + ($i$03<<3)|0)) + 4|0);
    HEAP32[$46>>2] = $45;
    $47 = (($i$03) + 1)|0;
    $exitcond6 = ($47|0)==($count|0);
    if ($exitcond6) {
     break;
    } else {
     $i$03 = $47;
    }
   }
  }
 }
 $57 = ((($renderer)) + 312|0);
 $58 = HEAP32[$57>>2]|0;
 $59 = ($58|0)==(0);
 $60 = ((($renderer)) + 308|0);
 $61 = HEAP32[$60>>2]|0;
 $62 = $61&65535;
 $63 = $61&255;
 $64 = ($62&65535) >>> 8;
 $65 = $64&255;
 $66 = $61 >>> 16;
 $67 = $66&255;
 $68 = $61 >>> 24;
 $69 = $68&255;
 if ($59) {
  $70 = ((($71)) + 4|0);
  $72 = HEAP32[$70>>2]|0;
  $73 = (_SDL_MapRGBA($72,$63,$65,$67,$69)|0);
  $74 = (_SDL_DrawPoints($71,$24,$count,$73)|0);
  $$0 = $74;
  STACKTOP = sp;return ($$0|0);
 } else {
  $75 = (_SDL_BlendPoints($71,$24,$count,$58,$63,$65,$67,$69)|0);
  $$0 = $75;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_DrawPoints($dst,$points,$count,$color) {
 $dst = $dst|0;
 $points = $points|0;
 $count = $count|0;
 $color = $color|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$03 = 0, $notlhs = 0, $notrhs = 0, $or$cond$not = 0, $or$cond1 = 0, $or$cond1$not = 0;
 var $or$cond2 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($dst|0)==(0|0);
 if ($0) {
  $1 = (_SDL_SetError(603923,$vararg_buffer)|0);
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ((($dst)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 8|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ($5&255)<(8);
 if ($6) {
  $7 = (_SDL_SetError(603504,$vararg_buffer1)|0);
  $$0 = $7;
  STACKTOP = sp;return ($$0|0);
 }
 $8 = ((($dst)) + 36|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($dst)) + 44|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + ($9))|0;
 $13 = ((($dst)) + 40|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($dst)) + 48|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = (($16) + ($14))|0;
 $18 = ($count|0)>(0);
 if (!($18)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $19 = $color&255;
 $20 = ((($dst)) + 20|0);
 $21 = ((($dst)) + 16|0);
 $22 = $color&65535;
 $i$03 = 0;
 L12: while(1) {
  $23 = (($points) + ($i$03<<3)|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = (((($points) + ($i$03<<3)|0)) + 4|0);
  $26 = HEAP32[$25>>2]|0;
  $notlhs = ($24|0)<($9|0);
  $notrhs = ($24|0)>=($12|0);
  $or$cond$not = $notrhs | $notlhs;
  $27 = ($26|0)<($14|0);
  $or$cond1 = $or$cond$not | $27;
  $or$cond1$not = $or$cond1 ^ 1;
  $28 = ($26|0)<($17|0);
  $or$cond2 = $28 & $or$cond1$not;
  L14: do {
   if ($or$cond2) {
    $29 = HEAP32[$2>>2]|0;
    $30 = ((($29)) + 9|0);
    $31 = HEAP8[$30>>0]|0;
    $32 = $31&255;
    switch ($32|0) {
    case 3:  {
     break L12;
     break;
    }
    case 1:  {
     $33 = HEAP32[$20>>2]|0;
     $34 = HEAP32[$21>>2]|0;
     $35 = Math_imul($34, $26)|0;
     $36 = (($33) + ($35)|0);
     $37 = (($36) + ($24)|0);
     HEAP8[$37>>0] = $19;
     break L14;
     break;
    }
    case 2:  {
     $38 = HEAP32[$20>>2]|0;
     $39 = HEAP32[$21>>2]|0;
     $40 = Math_imul($39, $26)|0;
     $41 = (($38) + ($40)|0);
     $42 = $24 << 1;
     $43 = (($41) + ($42)|0);
     HEAP16[$43>>1] = $22;
     break L14;
     break;
    }
    case 4:  {
     $45 = HEAP32[$20>>2]|0;
     $46 = HEAP32[$21>>2]|0;
     $47 = Math_imul($46, $26)|0;
     $48 = (($45) + ($47)|0);
     $49 = $24 << 2;
     $50 = (($48) + ($49)|0);
     HEAP32[$50>>2] = $color;
     break L14;
     break;
    }
    default: {
     break L14;
    }
    }
   }
  } while(0);
  $51 = (($i$03) + 1)|0;
  $52 = ($51|0)<($count|0);
  if ($52) {
   $i$03 = $51;
  } else {
   $$0 = 0;
   label = 14;
   break;
  }
 }
 if ((label|0) == 14) {
  STACKTOP = sp;return ($$0|0);
 }
 $44 = (_SDL_Error(4)|0);
 $$0 = $44;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_BlendPoints($dst,$points,$count,$blendMode,$r,$g,$b,$a) {
 $dst = $dst|0;
 $points = $points|0;
 $count = $count|0;
 $blendMode = $blendMode|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $$0 = 0, $$04 = 0, $$05 = 0, $$06 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $SDL_BlendPoint_RGB$SDL_BlendPoint_RGBA = 0, $SDL_BlendPoint_RGB888$SDL_BlendPoint_ARGB8888 = 0;
 var $blendMode$off = 0, $cond = 0, $cond1 = 0, $cond2 = 0, $exitcond = 0, $func$1 = 0, $i$012 = 0, $notlhs = 0, $notrhs = 0, $or$cond$not = 0, $or$cond7 = 0, $or$cond7$not = 0, $or$cond8 = 0, $status$013 = 0, $status$1 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($dst|0)==(0|0);
 if ($0) {
  $1 = (_SDL_SetError(603923,$vararg_buffer)|0);
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ((($dst)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 8|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ($5&255)<(8);
 if ($6) {
  $7 = (_SDL_SetError(603549,$vararg_buffer1)|0);
  $$0 = $7;
  STACKTOP = sp;return ($$0|0);
 }
 $blendMode$off = (($blendMode) + -1)|0;
 $8 = ($blendMode$off>>>0)<(2);
 if ($8) {
  $9 = $r&255;
  $10 = $a&255;
  $11 = Math_imul($10, $9)|0;
  $12 = (($11>>>0) / 255)&-1;
  $13 = $12&255;
  $14 = $g&255;
  $15 = Math_imul($10, $14)|0;
  $16 = (($15>>>0) / 255)&-1;
  $17 = $16&255;
  $18 = $b&255;
  $19 = Math_imul($10, $18)|0;
  $20 = (($19>>>0) / 255)&-1;
  $21 = $20&255;
  $$04 = $13;$$05 = $17;$$06 = $21;
 } else {
  $$04 = $r;$$05 = $g;$$06 = $b;
 }
 $22 = $5&255;
 switch ($22|0) {
 case 15:  {
  $23 = ((($3)) + 12|0);
  $24 = HEAP32[$23>>2]|0;
  $cond2 = ($24|0)==(31744);
  if ($cond2) {
   $func$1 = 4546;
  } else {
   label = 12;
  }
  break;
 }
 case 16:  {
  $25 = ((($3)) + 12|0);
  $26 = HEAP32[$25>>2]|0;
  $cond1 = ($26|0)==(63488);
  if ($cond1) {
   $func$1 = 4547;
  } else {
   label = 12;
  }
  break;
 }
 case 32:  {
  $27 = ((($3)) + 12|0);
  $28 = HEAP32[$27>>2]|0;
  $cond = ($28|0)==(16711680);
  if ($cond) {
   $29 = ((($3)) + 24|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = ($30|0)==(0);
   $SDL_BlendPoint_RGB888$SDL_BlendPoint_ARGB8888 = $31 ? 4542 : 4543;
   $func$1 = $SDL_BlendPoint_RGB888$SDL_BlendPoint_ARGB8888;
  } else {
   label = 12;
  }
  break;
 }
 default: {
  label = 12;
 }
 }
 if ((label|0) == 12) {
  $32 = ((($3)) + 24|0);
  $33 = HEAP32[$32>>2]|0;
  $34 = ($33|0)==(0);
  $SDL_BlendPoint_RGB$SDL_BlendPoint_RGBA = $34 ? 4544 : 4545;
  $func$1 = $SDL_BlendPoint_RGB$SDL_BlendPoint_RGBA;
 }
 $35 = ((($dst)) + 36|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = ((($dst)) + 44|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = (($38) + ($36))|0;
 $40 = ((($dst)) + 40|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = ((($dst)) + 48|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = (($43) + ($41))|0;
 $45 = ($count|0)>(0);
 if ($45) {
  $i$012 = 0;$status$013 = 0;
 } else {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 while(1) {
  $46 = (($points) + ($i$012<<3)|0);
  $47 = HEAP32[$46>>2]|0;
  $48 = (((($points) + ($i$012<<3)|0)) + 4|0);
  $49 = HEAP32[$48>>2]|0;
  $notlhs = ($47|0)<($36|0);
  $notrhs = ($47|0)>=($39|0);
  $or$cond$not = $notrhs | $notlhs;
  $50 = ($49|0)<($41|0);
  $or$cond7 = $or$cond$not | $50;
  $or$cond7$not = $or$cond7 ^ 1;
  $51 = ($49|0)<($44|0);
  $or$cond8 = $51 & $or$cond7$not;
  if ($or$cond8) {
   $52 = (FUNCTION_TABLE_iiiiiiiii[$func$1 & 8191]($dst,$47,$49,$blendMode,$$04,$$05,$$06,$a)|0);
   $status$1 = $52;
  } else {
   $status$1 = $status$013;
  }
  $53 = (($i$012) + 1)|0;
  $exitcond = ($53|0)==($count|0);
  if ($exitcond) {
   $$0 = $status$1;
   break;
  } else {
   $i$012 = $53;$status$013 = $status$1;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _SDL_BlendPoint_RGB888($dst,$x,$y,$blendMode,$r,$g,$b,$a) {
 $dst = $dst|0;
 $x = $x|0;
 $y = $y|0;
 $blendMode = $blendMode|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $$ = 0, $$1 = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $a&255;
 $1 = $0 ^ 255;
 switch ($blendMode|0) {
 case 1:  {
  $2 = ((($dst)) + 20|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($dst)) + 16|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = Math_imul($5, $y)|0;
  $7 = (($3) + ($6)|0);
  $8 = $x << 2;
  $9 = (($7) + ($8)|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = $10 >>> 16;
  $12 = $11 & 255;
  $13 = $10 >>> 8;
  $14 = $13 & 255;
  $15 = $10 & 255;
  $16 = Math_imul($12, $1)|0;
  $17 = (($16>>>0) / 255)&-1;
  $18 = $r&255;
  $19 = (($17) + ($18))|0;
  $20 = Math_imul($14, $1)|0;
  $21 = (($20>>>0) / 255)&-1;
  $22 = $g&255;
  $23 = (($21) + ($22))|0;
  $24 = Math_imul($15, $1)|0;
  $25 = (($24>>>0) / 255)&-1;
  $26 = $b&255;
  $27 = (($25) + ($26))|0;
  $28 = $19 << 16;
  $29 = $23 << 8;
  $30 = $29 | $27;
  $31 = $30 | $28;
  HEAP32[$9>>2] = $31;
  return 0;
  break;
 }
 case 2:  {
  $32 = ((($dst)) + 20|0);
  $33 = HEAP32[$32>>2]|0;
  $34 = ((($dst)) + 16|0);
  $35 = HEAP32[$34>>2]|0;
  $36 = Math_imul($35, $y)|0;
  $37 = (($33) + ($36)|0);
  $38 = $x << 2;
  $39 = (($37) + ($38)|0);
  $40 = HEAP32[$39>>2]|0;
  $41 = $40 >>> 16;
  $42 = $41 & 255;
  $43 = $40 >>> 8;
  $44 = $43 & 255;
  $45 = $40 & 255;
  $46 = $r&255;
  $47 = (($42) + ($46))|0;
  $48 = ($47>>>0)>(255);
  $$1 = $48 ? 255 : $47;
  $49 = $g&255;
  $50 = (($44) + ($49))|0;
  $51 = ($50>>>0)>(255);
  $$ = $51 ? 255 : $50;
  $52 = $b&255;
  $53 = (($45) + ($52))|0;
  $54 = ($53>>>0)>(255);
  $$2 = $54 ? 255 : $53;
  $55 = $$1 << 16;
  $56 = $$ << 8;
  $57 = $56 | $$2;
  $58 = $57 | $55;
  HEAP32[$39>>2] = $58;
  return 0;
  break;
 }
 case 4:  {
  $59 = ((($dst)) + 20|0);
  $60 = HEAP32[$59>>2]|0;
  $61 = ((($dst)) + 16|0);
  $62 = HEAP32[$61>>2]|0;
  $63 = Math_imul($62, $y)|0;
  $64 = (($60) + ($63)|0);
  $65 = $x << 2;
  $66 = (($64) + ($65)|0);
  $67 = HEAP32[$66>>2]|0;
  $68 = $67 >>> 16;
  $69 = $68 & 255;
  $70 = $67 >>> 8;
  $71 = $70 & 255;
  $72 = $67 & 255;
  $73 = $r&255;
  $74 = Math_imul($69, $73)|0;
  $75 = (($74>>>0) / 255)&-1;
  $76 = $g&255;
  $77 = Math_imul($71, $76)|0;
  $78 = (($77>>>0) / 255)&-1;
  $79 = $b&255;
  $80 = Math_imul($72, $79)|0;
  $81 = (($80>>>0) / 255)&-1;
  $82 = $75 << 16;
  $83 = $78 << 8;
  $84 = $83 | $81;
  $85 = $84 | $82;
  HEAP32[$66>>2] = $85;
  return 0;
  break;
 }
 default: {
  $86 = ((($dst)) + 20|0);
  $87 = HEAP32[$86>>2]|0;
  $88 = ((($dst)) + 16|0);
  $89 = HEAP32[$88>>2]|0;
  $90 = Math_imul($89, $y)|0;
  $91 = (($87) + ($90)|0);
  $92 = $x << 2;
  $93 = (($91) + ($92)|0);
  $94 = $r&255;
  $95 = $g&255;
  $96 = $b&255;
  $97 = $94 << 16;
  $98 = $95 << 8;
  $99 = $98 | $97;
  $100 = $99 | $96;
  HEAP32[$93>>2] = $100;
  return 0;
 }
 }
 return (0)|0;
}
function _SDL_BlendPoint_ARGB8888($dst,$x,$y,$blendMode,$r,$g,$b,$a) {
 $dst = $dst|0;
 $x = $x|0;
 $y = $y|0;
 $blendMode = $blendMode|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $$ = 0, $$1 = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $a&255;
 $1 = $0 ^ 255;
 switch ($blendMode|0) {
 case 1:  {
  $2 = ((($dst)) + 20|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($dst)) + 16|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = Math_imul($5, $y)|0;
  $7 = (($3) + ($6)|0);
  $8 = $x << 2;
  $9 = (($7) + ($8)|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = $10 >>> 16;
  $12 = $11 & 255;
  $13 = $10 >>> 8;
  $14 = $13 & 255;
  $15 = $10 & 255;
  $16 = $10 >>> 24;
  $17 = Math_imul($12, $1)|0;
  $18 = (($17>>>0) / 255)&-1;
  $19 = $r&255;
  $20 = (($18) + ($19))|0;
  $21 = Math_imul($14, $1)|0;
  $22 = (($21>>>0) / 255)&-1;
  $23 = $g&255;
  $24 = (($22) + ($23))|0;
  $25 = Math_imul($15, $1)|0;
  $26 = (($25>>>0) / 255)&-1;
  $27 = $b&255;
  $28 = (($26) + ($27))|0;
  $29 = Math_imul($16, $1)|0;
  $30 = (($29>>>0) / 255)&-1;
  $31 = (($30) + ($0))|0;
  $32 = $31 << 24;
  $33 = $20 << 16;
  $34 = $24 << 8;
  $35 = $32 | $28;
  $36 = $35 | $33;
  $37 = $36 | $34;
  HEAP32[$9>>2] = $37;
  return 0;
  break;
 }
 case 2:  {
  $38 = ((($dst)) + 20|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = ((($dst)) + 16|0);
  $41 = HEAP32[$40>>2]|0;
  $42 = Math_imul($41, $y)|0;
  $43 = (($39) + ($42)|0);
  $44 = $x << 2;
  $45 = (($43) + ($44)|0);
  $46 = HEAP32[$45>>2]|0;
  $47 = $46 >>> 16;
  $48 = $47 & 255;
  $49 = $46 >>> 8;
  $50 = $49 & 255;
  $51 = $46 & 255;
  $52 = $46 & -16777216;
  $53 = $r&255;
  $54 = (($48) + ($53))|0;
  $55 = ($54>>>0)>(255);
  $$1 = $55 ? 255 : $54;
  $56 = $g&255;
  $57 = (($50) + ($56))|0;
  $58 = ($57>>>0)>(255);
  $$ = $58 ? 255 : $57;
  $59 = $b&255;
  $60 = (($51) + ($59))|0;
  $61 = ($60>>>0)>(255);
  $$2 = $61 ? 255 : $60;
  $62 = $$1 << 16;
  $63 = $$ << 8;
  $64 = $$2 | $52;
  $65 = $64 | $62;
  $66 = $65 | $63;
  HEAP32[$45>>2] = $66;
  return 0;
  break;
 }
 case 4:  {
  $67 = ((($dst)) + 20|0);
  $68 = HEAP32[$67>>2]|0;
  $69 = ((($dst)) + 16|0);
  $70 = HEAP32[$69>>2]|0;
  $71 = Math_imul($70, $y)|0;
  $72 = (($68) + ($71)|0);
  $73 = $x << 2;
  $74 = (($72) + ($73)|0);
  $75 = HEAP32[$74>>2]|0;
  $76 = $75 >>> 16;
  $77 = $76 & 255;
  $78 = $75 >>> 8;
  $79 = $78 & 255;
  $80 = $75 & 255;
  $81 = $75 & -16777216;
  $82 = $r&255;
  $83 = Math_imul($77, $82)|0;
  $84 = (($83>>>0) / 255)&-1;
  $85 = $g&255;
  $86 = Math_imul($79, $85)|0;
  $87 = (($86>>>0) / 255)&-1;
  $88 = $b&255;
  $89 = Math_imul($80, $88)|0;
  $90 = (($89>>>0) / 255)&-1;
  $91 = $84 << 16;
  $92 = $87 << 8;
  $93 = $90 | $81;
  $94 = $93 | $91;
  $95 = $94 | $92;
  HEAP32[$74>>2] = $95;
  return 0;
  break;
 }
 default: {
  $96 = ((($dst)) + 20|0);
  $97 = HEAP32[$96>>2]|0;
  $98 = ((($dst)) + 16|0);
  $99 = HEAP32[$98>>2]|0;
  $100 = Math_imul($99, $y)|0;
  $101 = (($97) + ($100)|0);
  $102 = $x << 2;
  $103 = (($101) + ($102)|0);
  $104 = $r&255;
  $105 = $g&255;
  $106 = $b&255;
  $107 = $0 << 24;
  $108 = $104 << 16;
  $109 = $105 << 8;
  $110 = $109 | $108;
  $111 = $110 | $106;
  $112 = $111 | $107;
  HEAP32[$103>>2] = $112;
  return 0;
 }
 }
 return (0)|0;
}
function _SDL_BlendPoint_RGB($dst,$x,$y,$blendMode,$r,$g,$b,$a) {
 $dst = $dst|0;
 $x = $x|0;
 $y = $y|0;
 $blendMode = $blendMode|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $$ = 0, $$0 = 0, $$1 = 0, $$2 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0;
 var $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0;
 var $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0;
 var $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0;
 var $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0;
 var $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0;
 var $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0;
 var $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0;
 var $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0;
 var $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0;
 var $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0;
 var $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0;
 var $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0;
 var $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0;
 var $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0;
 var $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0;
 var $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0;
 var $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0;
 var $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0;
 var $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($dst)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $a&255;
 $3 = $2 ^ 255;
 $4 = ((($1)) + 9|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 switch ($6|0) {
 case 2:  {
  switch ($blendMode|0) {
  case 1:  {
   $7 = ((($dst)) + 20|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ((($dst)) + 16|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = Math_imul($10, $y)|0;
   $12 = (($8) + ($11)|0);
   $13 = $x << 1;
   $14 = (($12) + ($13)|0);
   $15 = HEAP16[$14>>1]|0;
   $16 = $15&65535;
   $17 = ((($1)) + 12|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = $16 & $18;
   $20 = ((($1)) + 32|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = $21 & 255;
   $23 = $19 >>> $22;
   $24 = ((($1)) + 28|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = $25 & 255;
   $27 = (125924 + ($26<<2)|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = (($28) + ($23)|0);
   $30 = HEAP8[$29>>0]|0;
   $31 = $30&255;
   $32 = ((($1)) + 16|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = $33 & $16;
   $35 = $21 >>> 8;
   $36 = $35 & 255;
   $37 = $34 >>> $36;
   $38 = $25 >>> 8;
   $39 = $38 & 255;
   $40 = (125924 + ($39<<2)|0);
   $41 = HEAP32[$40>>2]|0;
   $42 = (($41) + ($37)|0);
   $43 = HEAP8[$42>>0]|0;
   $44 = $43&255;
   $45 = ((($1)) + 20|0);
   $46 = HEAP32[$45>>2]|0;
   $47 = $46 & $16;
   $48 = $21 >>> 16;
   $49 = $48 & 255;
   $50 = $47 >>> $49;
   $51 = $25 >>> 16;
   $52 = $51 & 255;
   $53 = (125924 + ($52<<2)|0);
   $54 = HEAP32[$53>>2]|0;
   $55 = (($54) + ($50)|0);
   $56 = HEAP8[$55>>0]|0;
   $57 = $56&255;
   $58 = Math_imul($31, $3)|0;
   $59 = (($58>>>0) / 255)&-1;
   $60 = $r&255;
   $61 = (($59) + ($60))|0;
   $62 = Math_imul($44, $3)|0;
   $63 = (($62>>>0) / 255)&-1;
   $64 = $g&255;
   $65 = (($63) + ($64))|0;
   $66 = Math_imul($57, $3)|0;
   $67 = (($66>>>0) / 255)&-1;
   $68 = $b&255;
   $69 = (($67) + ($68))|0;
   $70 = $61 >>> $26;
   $71 = $70 << $22;
   $72 = $65 >>> $39;
   $73 = $72 << $36;
   $74 = $73 | $71;
   $75 = $69 >>> $52;
   $76 = $75 << $49;
   $77 = ((($1)) + 24|0);
   $78 = HEAP32[$77>>2]|0;
   $79 = $74 | $78;
   $80 = $79 | $76;
   $81 = $80&65535;
   HEAP16[$14>>1] = $81;
   $$0 = 0;
   return ($$0|0);
   break;
  }
  case 2:  {
   $82 = ((($dst)) + 20|0);
   $83 = HEAP32[$82>>2]|0;
   $84 = ((($dst)) + 16|0);
   $85 = HEAP32[$84>>2]|0;
   $86 = Math_imul($85, $y)|0;
   $87 = (($83) + ($86)|0);
   $88 = $x << 1;
   $89 = (($87) + ($88)|0);
   $90 = HEAP16[$89>>1]|0;
   $91 = $90&65535;
   $92 = ((($1)) + 12|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = $91 & $93;
   $95 = ((($1)) + 32|0);
   $96 = HEAP32[$95>>2]|0;
   $97 = $96 & 255;
   $98 = $94 >>> $97;
   $99 = ((($1)) + 28|0);
   $100 = HEAP32[$99>>2]|0;
   $101 = $100 & 255;
   $102 = (125924 + ($101<<2)|0);
   $103 = HEAP32[$102>>2]|0;
   $104 = (($103) + ($98)|0);
   $105 = HEAP8[$104>>0]|0;
   $106 = $105&255;
   $107 = ((($1)) + 16|0);
   $108 = HEAP32[$107>>2]|0;
   $109 = $108 & $91;
   $110 = $96 >>> 8;
   $111 = $110 & 255;
   $112 = $109 >>> $111;
   $113 = $100 >>> 8;
   $114 = $113 & 255;
   $115 = (125924 + ($114<<2)|0);
   $116 = HEAP32[$115>>2]|0;
   $117 = (($116) + ($112)|0);
   $118 = HEAP8[$117>>0]|0;
   $119 = $118&255;
   $120 = ((($1)) + 20|0);
   $121 = HEAP32[$120>>2]|0;
   $122 = $121 & $91;
   $123 = $96 >>> 16;
   $124 = $123 & 255;
   $125 = $122 >>> $124;
   $126 = $100 >>> 16;
   $127 = $126 & 255;
   $128 = (125924 + ($127<<2)|0);
   $129 = HEAP32[$128>>2]|0;
   $130 = (($129) + ($125)|0);
   $131 = HEAP8[$130>>0]|0;
   $132 = $131&255;
   $133 = $r&255;
   $134 = (($106) + ($133))|0;
   $135 = ($134>>>0)>(255);
   $$2 = $135 ? 255 : $134;
   $136 = $g&255;
   $137 = (($119) + ($136))|0;
   $138 = ($137>>>0)>(255);
   $$ = $138 ? 255 : $137;
   $139 = $b&255;
   $140 = (($132) + ($139))|0;
   $141 = ($140>>>0)>(255);
   $$4 = $141 ? 255 : $140;
   $142 = $$2 >>> $101;
   $143 = $142 << $97;
   $144 = $$ >>> $114;
   $145 = $144 << $111;
   $146 = $145 | $143;
   $147 = $$4 >>> $127;
   $148 = $147 << $124;
   $149 = ((($1)) + 24|0);
   $150 = HEAP32[$149>>2]|0;
   $151 = $146 | $150;
   $152 = $151 | $148;
   $153 = $152&65535;
   HEAP16[$89>>1] = $153;
   $$0 = 0;
   return ($$0|0);
   break;
  }
  case 4:  {
   $154 = ((($dst)) + 20|0);
   $155 = HEAP32[$154>>2]|0;
   $156 = ((($dst)) + 16|0);
   $157 = HEAP32[$156>>2]|0;
   $158 = Math_imul($157, $y)|0;
   $159 = (($155) + ($158)|0);
   $160 = $x << 1;
   $161 = (($159) + ($160)|0);
   $162 = HEAP16[$161>>1]|0;
   $163 = $162&65535;
   $164 = ((($1)) + 12|0);
   $165 = HEAP32[$164>>2]|0;
   $166 = $163 & $165;
   $167 = ((($1)) + 32|0);
   $168 = HEAP32[$167>>2]|0;
   $169 = $168 & 255;
   $170 = $166 >>> $169;
   $171 = ((($1)) + 28|0);
   $172 = HEAP32[$171>>2]|0;
   $173 = $172 & 255;
   $174 = (125924 + ($173<<2)|0);
   $175 = HEAP32[$174>>2]|0;
   $176 = (($175) + ($170)|0);
   $177 = HEAP8[$176>>0]|0;
   $178 = $177&255;
   $179 = ((($1)) + 16|0);
   $180 = HEAP32[$179>>2]|0;
   $181 = $180 & $163;
   $182 = $168 >>> 8;
   $183 = $182 & 255;
   $184 = $181 >>> $183;
   $185 = $172 >>> 8;
   $186 = $185 & 255;
   $187 = (125924 + ($186<<2)|0);
   $188 = HEAP32[$187>>2]|0;
   $189 = (($188) + ($184)|0);
   $190 = HEAP8[$189>>0]|0;
   $191 = $190&255;
   $192 = ((($1)) + 20|0);
   $193 = HEAP32[$192>>2]|0;
   $194 = $193 & $163;
   $195 = $168 >>> 16;
   $196 = $195 & 255;
   $197 = $194 >>> $196;
   $198 = $172 >>> 16;
   $199 = $198 & 255;
   $200 = (125924 + ($199<<2)|0);
   $201 = HEAP32[$200>>2]|0;
   $202 = (($201) + ($197)|0);
   $203 = HEAP8[$202>>0]|0;
   $204 = $203&255;
   $205 = $r&255;
   $206 = Math_imul($178, $205)|0;
   $207 = (($206>>>0) / 255)&-1;
   $208 = $g&255;
   $209 = Math_imul($191, $208)|0;
   $210 = (($209>>>0) / 255)&-1;
   $211 = $b&255;
   $212 = Math_imul($204, $211)|0;
   $213 = (($212>>>0) / 255)&-1;
   $214 = $207 >>> $173;
   $215 = $214 << $169;
   $216 = $210 >>> $186;
   $217 = $216 << $183;
   $218 = $217 | $215;
   $219 = $213 >>> $199;
   $220 = $219 << $196;
   $221 = ((($1)) + 24|0);
   $222 = HEAP32[$221>>2]|0;
   $223 = $218 | $222;
   $224 = $223 | $220;
   $225 = $224&65535;
   HEAP16[$161>>1] = $225;
   $$0 = 0;
   return ($$0|0);
   break;
  }
  default: {
   $226 = ((($dst)) + 20|0);
   $227 = HEAP32[$226>>2]|0;
   $228 = ((($dst)) + 16|0);
   $229 = HEAP32[$228>>2]|0;
   $230 = Math_imul($229, $y)|0;
   $231 = (($227) + ($230)|0);
   $232 = $x << 1;
   $233 = (($231) + ($232)|0);
   $234 = $r&255;
   $235 = $g&255;
   $236 = $b&255;
   $237 = ((($1)) + 28|0);
   $238 = HEAP32[$237>>2]|0;
   $239 = $238 & 255;
   $240 = $234 >>> $239;
   $241 = ((($1)) + 32|0);
   $242 = HEAP32[$241>>2]|0;
   $243 = $242 & 255;
   $244 = $240 << $243;
   $245 = $238 >>> 8;
   $246 = $245 & 255;
   $247 = $235 >>> $246;
   $248 = $242 >>> 8;
   $249 = $248 & 255;
   $250 = $247 << $249;
   $251 = $250 | $244;
   $252 = $238 >>> 16;
   $253 = $252 & 255;
   $254 = $236 >>> $253;
   $255 = $242 >>> 16;
   $256 = $255 & 255;
   $257 = $254 << $256;
   $258 = ((($1)) + 24|0);
   $259 = HEAP32[$258>>2]|0;
   $260 = $251 | $259;
   $261 = $260 | $257;
   $262 = $261&65535;
   HEAP16[$233>>1] = $262;
   $$0 = 0;
   return ($$0|0);
  }
  }
  break;
 }
 case 4:  {
  switch ($blendMode|0) {
  case 1:  {
   $263 = ((($dst)) + 20|0);
   $264 = HEAP32[$263>>2]|0;
   $265 = ((($dst)) + 16|0);
   $266 = HEAP32[$265>>2]|0;
   $267 = Math_imul($266, $y)|0;
   $268 = (($264) + ($267)|0);
   $269 = $x << 2;
   $270 = (($268) + ($269)|0);
   $271 = HEAP32[$270>>2]|0;
   $272 = ((($1)) + 12|0);
   $273 = HEAP32[$272>>2]|0;
   $274 = $273 & $271;
   $275 = ((($1)) + 32|0);
   $276 = HEAP32[$275>>2]|0;
   $277 = $276 & 255;
   $278 = $274 >>> $277;
   $279 = ((($1)) + 28|0);
   $280 = HEAP32[$279>>2]|0;
   $281 = $280 & 255;
   $282 = (125924 + ($281<<2)|0);
   $283 = HEAP32[$282>>2]|0;
   $284 = (($283) + ($278)|0);
   $285 = HEAP8[$284>>0]|0;
   $286 = $285&255;
   $287 = ((($1)) + 16|0);
   $288 = HEAP32[$287>>2]|0;
   $289 = $288 & $271;
   $290 = $276 >>> 8;
   $291 = $290 & 255;
   $292 = $289 >>> $291;
   $293 = $280 >>> 8;
   $294 = $293 & 255;
   $295 = (125924 + ($294<<2)|0);
   $296 = HEAP32[$295>>2]|0;
   $297 = (($296) + ($292)|0);
   $298 = HEAP8[$297>>0]|0;
   $299 = $298&255;
   $300 = ((($1)) + 20|0);
   $301 = HEAP32[$300>>2]|0;
   $302 = $301 & $271;
   $303 = $276 >>> 16;
   $304 = $303 & 255;
   $305 = $302 >>> $304;
   $306 = $280 >>> 16;
   $307 = $306 & 255;
   $308 = (125924 + ($307<<2)|0);
   $309 = HEAP32[$308>>2]|0;
   $310 = (($309) + ($305)|0);
   $311 = HEAP8[$310>>0]|0;
   $312 = $311&255;
   $313 = Math_imul($286, $3)|0;
   $314 = (($313>>>0) / 255)&-1;
   $315 = $r&255;
   $316 = (($314) + ($315))|0;
   $317 = Math_imul($299, $3)|0;
   $318 = (($317>>>0) / 255)&-1;
   $319 = $g&255;
   $320 = (($318) + ($319))|0;
   $321 = Math_imul($312, $3)|0;
   $322 = (($321>>>0) / 255)&-1;
   $323 = $b&255;
   $324 = (($322) + ($323))|0;
   $325 = $316 >>> $281;
   $326 = $325 << $277;
   $327 = $320 >>> $294;
   $328 = $327 << $291;
   $329 = $328 | $326;
   $330 = $324 >>> $307;
   $331 = $330 << $304;
   $332 = ((($1)) + 24|0);
   $333 = HEAP32[$332>>2]|0;
   $334 = $329 | $333;
   $335 = $334 | $331;
   HEAP32[$270>>2] = $335;
   $$0 = 0;
   return ($$0|0);
   break;
  }
  case 2:  {
   $336 = ((($dst)) + 20|0);
   $337 = HEAP32[$336>>2]|0;
   $338 = ((($dst)) + 16|0);
   $339 = HEAP32[$338>>2]|0;
   $340 = Math_imul($339, $y)|0;
   $341 = (($337) + ($340)|0);
   $342 = $x << 2;
   $343 = (($341) + ($342)|0);
   $344 = HEAP32[$343>>2]|0;
   $345 = ((($1)) + 12|0);
   $346 = HEAP32[$345>>2]|0;
   $347 = $346 & $344;
   $348 = ((($1)) + 32|0);
   $349 = HEAP32[$348>>2]|0;
   $350 = $349 & 255;
   $351 = $347 >>> $350;
   $352 = ((($1)) + 28|0);
   $353 = HEAP32[$352>>2]|0;
   $354 = $353 & 255;
   $355 = (125924 + ($354<<2)|0);
   $356 = HEAP32[$355>>2]|0;
   $357 = (($356) + ($351)|0);
   $358 = HEAP8[$357>>0]|0;
   $359 = $358&255;
   $360 = ((($1)) + 16|0);
   $361 = HEAP32[$360>>2]|0;
   $362 = $361 & $344;
   $363 = $349 >>> 8;
   $364 = $363 & 255;
   $365 = $362 >>> $364;
   $366 = $353 >>> 8;
   $367 = $366 & 255;
   $368 = (125924 + ($367<<2)|0);
   $369 = HEAP32[$368>>2]|0;
   $370 = (($369) + ($365)|0);
   $371 = HEAP8[$370>>0]|0;
   $372 = $371&255;
   $373 = ((($1)) + 20|0);
   $374 = HEAP32[$373>>2]|0;
   $375 = $374 & $344;
   $376 = $349 >>> 16;
   $377 = $376 & 255;
   $378 = $375 >>> $377;
   $379 = $353 >>> 16;
   $380 = $379 & 255;
   $381 = (125924 + ($380<<2)|0);
   $382 = HEAP32[$381>>2]|0;
   $383 = (($382) + ($378)|0);
   $384 = HEAP8[$383>>0]|0;
   $385 = $384&255;
   $386 = $r&255;
   $387 = (($359) + ($386))|0;
   $388 = ($387>>>0)>(255);
   $$3 = $388 ? 255 : $387;
   $389 = $g&255;
   $390 = (($372) + ($389))|0;
   $391 = ($390>>>0)>(255);
   $$1 = $391 ? 255 : $390;
   $392 = $b&255;
   $393 = (($385) + ($392))|0;
   $394 = ($393>>>0)>(255);
   $$5 = $394 ? 255 : $393;
   $395 = $$3 >>> $354;
   $396 = $395 << $350;
   $397 = $$1 >>> $367;
   $398 = $397 << $364;
   $399 = $398 | $396;
   $400 = $$5 >>> $380;
   $401 = $400 << $377;
   $402 = ((($1)) + 24|0);
   $403 = HEAP32[$402>>2]|0;
   $404 = $399 | $403;
   $405 = $404 | $401;
   HEAP32[$343>>2] = $405;
   $$0 = 0;
   return ($$0|0);
   break;
  }
  case 4:  {
   $406 = ((($dst)) + 20|0);
   $407 = HEAP32[$406>>2]|0;
   $408 = ((($dst)) + 16|0);
   $409 = HEAP32[$408>>2]|0;
   $410 = Math_imul($409, $y)|0;
   $411 = (($407) + ($410)|0);
   $412 = $x << 2;
   $413 = (($411) + ($412)|0);
   $414 = HEAP32[$413>>2]|0;
   $415 = ((($1)) + 12|0);
   $416 = HEAP32[$415>>2]|0;
   $417 = $416 & $414;
   $418 = ((($1)) + 32|0);
   $419 = HEAP32[$418>>2]|0;
   $420 = $419 & 255;
   $421 = $417 >>> $420;
   $422 = ((($1)) + 28|0);
   $423 = HEAP32[$422>>2]|0;
   $424 = $423 & 255;
   $425 = (125924 + ($424<<2)|0);
   $426 = HEAP32[$425>>2]|0;
   $427 = (($426) + ($421)|0);
   $428 = HEAP8[$427>>0]|0;
   $429 = $428&255;
   $430 = ((($1)) + 16|0);
   $431 = HEAP32[$430>>2]|0;
   $432 = $431 & $414;
   $433 = $419 >>> 8;
   $434 = $433 & 255;
   $435 = $432 >>> $434;
   $436 = $423 >>> 8;
   $437 = $436 & 255;
   $438 = (125924 + ($437<<2)|0);
   $439 = HEAP32[$438>>2]|0;
   $440 = (($439) + ($435)|0);
   $441 = HEAP8[$440>>0]|0;
   $442 = $441&255;
   $443 = ((($1)) + 20|0);
   $444 = HEAP32[$443>>2]|0;
   $445 = $444 & $414;
   $446 = $419 >>> 16;
   $447 = $446 & 255;
   $448 = $445 >>> $447;
   $449 = $423 >>> 16;
   $450 = $449 & 255;
   $451 = (125924 + ($450<<2)|0);
   $452 = HEAP32[$451>>2]|0;
   $453 = (($452) + ($448)|0);
   $454 = HEAP8[$453>>0]|0;
   $455 = $454&255;
   $456 = $r&255;
   $457 = Math_imul($429, $456)|0;
   $458 = (($457>>>0) / 255)&-1;
   $459 = $g&255;
   $460 = Math_imul($442, $459)|0;
   $461 = (($460>>>0) / 255)&-1;
   $462 = $b&255;
   $463 = Math_imul($455, $462)|0;
   $464 = (($463>>>0) / 255)&-1;
   $465 = $458 >>> $424;
   $466 = $465 << $420;
   $467 = $461 >>> $437;
   $468 = $467 << $434;
   $469 = $468 | $466;
   $470 = $464 >>> $450;
   $471 = $470 << $447;
   $472 = ((($1)) + 24|0);
   $473 = HEAP32[$472>>2]|0;
   $474 = $469 | $473;
   $475 = $474 | $471;
   HEAP32[$413>>2] = $475;
   $$0 = 0;
   return ($$0|0);
   break;
  }
  default: {
   $476 = ((($dst)) + 20|0);
   $477 = HEAP32[$476>>2]|0;
   $478 = ((($dst)) + 16|0);
   $479 = HEAP32[$478>>2]|0;
   $480 = Math_imul($479, $y)|0;
   $481 = (($477) + ($480)|0);
   $482 = $x << 2;
   $483 = (($481) + ($482)|0);
   $484 = $r&255;
   $485 = $g&255;
   $486 = $b&255;
   $487 = ((($1)) + 28|0);
   $488 = HEAP32[$487>>2]|0;
   $489 = $488 & 255;
   $490 = $484 >>> $489;
   $491 = ((($1)) + 32|0);
   $492 = HEAP32[$491>>2]|0;
   $493 = $492 & 255;
   $494 = $490 << $493;
   $495 = $488 >>> 8;
   $496 = $495 & 255;
   $497 = $485 >>> $496;
   $498 = $492 >>> 8;
   $499 = $498 & 255;
   $500 = $497 << $499;
   $501 = $500 | $494;
   $502 = $488 >>> 16;
   $503 = $502 & 255;
   $504 = $486 >>> $503;
   $505 = $492 >>> 16;
   $506 = $505 & 255;
   $507 = $504 << $506;
   $508 = ((($1)) + 24|0);
   $509 = HEAP32[$508>>2]|0;
   $510 = $501 | $509;
   $511 = $510 | $507;
   HEAP32[$483>>2] = $511;
   $$0 = 0;
   return ($$0|0);
  }
  }
  break;
 }
 default: {
  $512 = (_SDL_Error(4)|0);
  $$0 = $512;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function _SDL_BlendPoint_RGBA($dst,$x,$y,$blendMode,$r,$g,$b,$a) {
 $dst = $dst|0;
 $x = $x|0;
 $y = $y|0;
 $blendMode = $blendMode|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $$ = 0, $$0 = 0, $$1 = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0;
 var $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0;
 var $293 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cond = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($dst)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $a&255;
 $3 = $2 ^ 255;
 $4 = ((($1)) + 9|0);
 $5 = HEAP8[$4>>0]|0;
 $cond = ($5<<24>>24)==(4);
 if (!($cond)) {
  $293 = (_SDL_Error(4)|0);
  $$0 = $293;
  return ($$0|0);
 }
 switch ($blendMode|0) {
 case 1:  {
  $6 = ((($dst)) + 20|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ((($dst)) + 16|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = Math_imul($9, $y)|0;
  $11 = (($7) + ($10)|0);
  $12 = $x << 2;
  $13 = (($11) + ($12)|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ((($1)) + 12|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = $16 & $14;
  $18 = ((($1)) + 32|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = $19 & 255;
  $21 = $17 >>> $20;
  $22 = ((($1)) + 28|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = $23 & 255;
  $25 = (125924 + ($24<<2)|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = (($26) + ($21)|0);
  $28 = HEAP8[$27>>0]|0;
  $29 = $28&255;
  $30 = ((($1)) + 16|0);
  $31 = HEAP32[$30>>2]|0;
  $32 = $31 & $14;
  $33 = $19 >>> 8;
  $34 = $33 & 255;
  $35 = $32 >>> $34;
  $36 = $23 >>> 8;
  $37 = $36 & 255;
  $38 = (125924 + ($37<<2)|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = (($39) + ($35)|0);
  $41 = HEAP8[$40>>0]|0;
  $42 = $41&255;
  $43 = ((($1)) + 20|0);
  $44 = HEAP32[$43>>2]|0;
  $45 = $44 & $14;
  $46 = $19 >>> 16;
  $47 = $46 & 255;
  $48 = $45 >>> $47;
  $49 = $23 >>> 16;
  $50 = $49 & 255;
  $51 = (125924 + ($50<<2)|0);
  $52 = HEAP32[$51>>2]|0;
  $53 = (($52) + ($48)|0);
  $54 = HEAP8[$53>>0]|0;
  $55 = $54&255;
  $56 = ((($1)) + 24|0);
  $57 = HEAP32[$56>>2]|0;
  $58 = $57 & $14;
  $59 = $19 >>> 24;
  $60 = $58 >>> $59;
  $61 = $23 >>> 24;
  $62 = (125924 + ($61<<2)|0);
  $63 = HEAP32[$62>>2]|0;
  $64 = (($63) + ($60)|0);
  $65 = HEAP8[$64>>0]|0;
  $66 = $65&255;
  $67 = Math_imul($29, $3)|0;
  $68 = (($67>>>0) / 255)&-1;
  $69 = $r&255;
  $70 = (($68) + ($69))|0;
  $71 = Math_imul($42, $3)|0;
  $72 = (($71>>>0) / 255)&-1;
  $73 = $g&255;
  $74 = (($72) + ($73))|0;
  $75 = Math_imul($55, $3)|0;
  $76 = (($75>>>0) / 255)&-1;
  $77 = $b&255;
  $78 = (($76) + ($77))|0;
  $79 = Math_imul($66, $3)|0;
  $80 = (($79>>>0) / 255)&-1;
  $81 = (($80) + ($2))|0;
  $82 = $70 >>> $24;
  $83 = $82 << $20;
  $84 = $74 >>> $37;
  $85 = $84 << $34;
  $86 = $85 | $83;
  $87 = $78 >>> $50;
  $88 = $87 << $47;
  $89 = $86 | $88;
  $90 = $81 >>> $61;
  $91 = $90 << $59;
  $92 = $89 | $91;
  HEAP32[$13>>2] = $92;
  $$0 = 0;
  return ($$0|0);
  break;
 }
 case 2:  {
  $93 = ((($dst)) + 20|0);
  $94 = HEAP32[$93>>2]|0;
  $95 = ((($dst)) + 16|0);
  $96 = HEAP32[$95>>2]|0;
  $97 = Math_imul($96, $y)|0;
  $98 = (($94) + ($97)|0);
  $99 = $x << 2;
  $100 = (($98) + ($99)|0);
  $101 = HEAP32[$100>>2]|0;
  $102 = ((($1)) + 12|0);
  $103 = HEAP32[$102>>2]|0;
  $104 = $103 & $101;
  $105 = ((($1)) + 32|0);
  $106 = HEAP32[$105>>2]|0;
  $107 = $106 & 255;
  $108 = $104 >>> $107;
  $109 = ((($1)) + 28|0);
  $110 = HEAP32[$109>>2]|0;
  $111 = $110 & 255;
  $112 = (125924 + ($111<<2)|0);
  $113 = HEAP32[$112>>2]|0;
  $114 = (($113) + ($108)|0);
  $115 = HEAP8[$114>>0]|0;
  $116 = $115&255;
  $117 = ((($1)) + 16|0);
  $118 = HEAP32[$117>>2]|0;
  $119 = $118 & $101;
  $120 = $106 >>> 8;
  $121 = $120 & 255;
  $122 = $119 >>> $121;
  $123 = $110 >>> 8;
  $124 = $123 & 255;
  $125 = (125924 + ($124<<2)|0);
  $126 = HEAP32[$125>>2]|0;
  $127 = (($126) + ($122)|0);
  $128 = HEAP8[$127>>0]|0;
  $129 = $128&255;
  $130 = ((($1)) + 20|0);
  $131 = HEAP32[$130>>2]|0;
  $132 = $131 & $101;
  $133 = $106 >>> 16;
  $134 = $133 & 255;
  $135 = $132 >>> $134;
  $136 = $110 >>> 16;
  $137 = $136 & 255;
  $138 = (125924 + ($137<<2)|0);
  $139 = HEAP32[$138>>2]|0;
  $140 = (($139) + ($135)|0);
  $141 = HEAP8[$140>>0]|0;
  $142 = $141&255;
  $143 = ((($1)) + 24|0);
  $144 = HEAP32[$143>>2]|0;
  $145 = $144 & $101;
  $146 = $106 >>> 24;
  $147 = $145 >>> $146;
  $148 = $110 >>> 24;
  $149 = (125924 + ($148<<2)|0);
  $150 = HEAP32[$149>>2]|0;
  $151 = (($150) + ($147)|0);
  $152 = HEAP8[$151>>0]|0;
  $153 = $152&255;
  $154 = $r&255;
  $155 = (($116) + ($154))|0;
  $156 = ($155>>>0)>(255);
  $$1 = $156 ? 255 : $155;
  $157 = $g&255;
  $158 = (($129) + ($157))|0;
  $159 = ($158>>>0)>(255);
  $$ = $159 ? 255 : $158;
  $160 = $b&255;
  $161 = (($142) + ($160))|0;
  $162 = ($161>>>0)>(255);
  $$2 = $162 ? 255 : $161;
  $163 = $$1 >>> $111;
  $164 = $163 << $107;
  $165 = $$ >>> $124;
  $166 = $165 << $121;
  $167 = $166 | $164;
  $168 = $$2 >>> $137;
  $169 = $168 << $134;
  $170 = $167 | $169;
  $171 = $153 >>> $148;
  $172 = $171 << $146;
  $173 = $170 | $172;
  HEAP32[$100>>2] = $173;
  $$0 = 0;
  return ($$0|0);
  break;
 }
 case 4:  {
  $174 = ((($dst)) + 20|0);
  $175 = HEAP32[$174>>2]|0;
  $176 = ((($dst)) + 16|0);
  $177 = HEAP32[$176>>2]|0;
  $178 = Math_imul($177, $y)|0;
  $179 = (($175) + ($178)|0);
  $180 = $x << 2;
  $181 = (($179) + ($180)|0);
  $182 = HEAP32[$181>>2]|0;
  $183 = ((($1)) + 12|0);
  $184 = HEAP32[$183>>2]|0;
  $185 = $184 & $182;
  $186 = ((($1)) + 32|0);
  $187 = HEAP32[$186>>2]|0;
  $188 = $187 & 255;
  $189 = $185 >>> $188;
  $190 = ((($1)) + 28|0);
  $191 = HEAP32[$190>>2]|0;
  $192 = $191 & 255;
  $193 = (125924 + ($192<<2)|0);
  $194 = HEAP32[$193>>2]|0;
  $195 = (($194) + ($189)|0);
  $196 = HEAP8[$195>>0]|0;
  $197 = $196&255;
  $198 = ((($1)) + 16|0);
  $199 = HEAP32[$198>>2]|0;
  $200 = $199 & $182;
  $201 = $187 >>> 8;
  $202 = $201 & 255;
  $203 = $200 >>> $202;
  $204 = $191 >>> 8;
  $205 = $204 & 255;
  $206 = (125924 + ($205<<2)|0);
  $207 = HEAP32[$206>>2]|0;
  $208 = (($207) + ($203)|0);
  $209 = HEAP8[$208>>0]|0;
  $210 = $209&255;
  $211 = ((($1)) + 20|0);
  $212 = HEAP32[$211>>2]|0;
  $213 = $212 & $182;
  $214 = $187 >>> 16;
  $215 = $214 & 255;
  $216 = $213 >>> $215;
  $217 = $191 >>> 16;
  $218 = $217 & 255;
  $219 = (125924 + ($218<<2)|0);
  $220 = HEAP32[$219>>2]|0;
  $221 = (($220) + ($216)|0);
  $222 = HEAP8[$221>>0]|0;
  $223 = $222&255;
  $224 = ((($1)) + 24|0);
  $225 = HEAP32[$224>>2]|0;
  $226 = $225 & $182;
  $227 = $187 >>> 24;
  $228 = $226 >>> $227;
  $229 = $191 >>> 24;
  $230 = (125924 + ($229<<2)|0);
  $231 = HEAP32[$230>>2]|0;
  $232 = (($231) + ($228)|0);
  $233 = HEAP8[$232>>0]|0;
  $234 = $233&255;
  $235 = $r&255;
  $236 = Math_imul($197, $235)|0;
  $237 = (($236>>>0) / 255)&-1;
  $238 = $g&255;
  $239 = Math_imul($210, $238)|0;
  $240 = (($239>>>0) / 255)&-1;
  $241 = $b&255;
  $242 = Math_imul($223, $241)|0;
  $243 = (($242>>>0) / 255)&-1;
  $244 = $237 >>> $192;
  $245 = $244 << $188;
  $246 = $240 >>> $205;
  $247 = $246 << $202;
  $248 = $247 | $245;
  $249 = $243 >>> $218;
  $250 = $249 << $215;
  $251 = $248 | $250;
  $252 = $234 >>> $229;
  $253 = $252 << $227;
  $254 = $251 | $253;
  HEAP32[$181>>2] = $254;
  $$0 = 0;
  return ($$0|0);
  break;
 }
 default: {
  $255 = ((($dst)) + 20|0);
  $256 = HEAP32[$255>>2]|0;
  $257 = ((($dst)) + 16|0);
  $258 = HEAP32[$257>>2]|0;
  $259 = Math_imul($258, $y)|0;
  $260 = (($256) + ($259)|0);
  $261 = $x << 2;
  $262 = (($260) + ($261)|0);
  $263 = $r&255;
  $264 = $g&255;
  $265 = $b&255;
  $266 = ((($1)) + 28|0);
  $267 = HEAP32[$266>>2]|0;
  $268 = $267 & 255;
  $269 = $263 >>> $268;
  $270 = ((($1)) + 32|0);
  $271 = HEAP32[$270>>2]|0;
  $272 = $271 & 255;
  $273 = $269 << $272;
  $274 = $267 >>> 8;
  $275 = $274 & 255;
  $276 = $264 >>> $275;
  $277 = $271 >>> 8;
  $278 = $277 & 255;
  $279 = $276 << $278;
  $280 = $279 | $273;
  $281 = $267 >>> 16;
  $282 = $281 & 255;
  $283 = $265 >>> $282;
  $284 = $271 >>> 16;
  $285 = $284 & 255;
  $286 = $283 << $285;
  $287 = $280 | $286;
  $288 = $267 >>> 24;
  $289 = $2 >>> $288;
  $290 = $271 >>> 24;
  $291 = $289 << $290;
  $292 = $287 | $291;
  HEAP32[$262>>2] = $292;
  $$0 = 0;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function _SDL_BlendPoint_RGB555($dst,$x,$y,$blendMode,$r,$g,$b,$a) {
 $dst = $dst|0;
 $x = $x|0;
 $y = $y|0;
 $blendMode = $blendMode|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $$ = 0, $$1 = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $a&255;
 $1 = $0 ^ 255;
 switch ($blendMode|0) {
 case 1:  {
  $2 = ((($dst)) + 20|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($dst)) + 16|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = Math_imul($5, $y)|0;
  $7 = (($3) + ($6)|0);
  $8 = $x << 1;
  $9 = (($7) + ($8)|0);
  $10 = HEAP16[$9>>1]|0;
  $11 = $10&65535;
  $12 = $11 >>> 10;
  $13 = $12 & 31;
  $14 = HEAP32[(125936)>>2]|0;
  $15 = (($14) + ($13)|0);
  $16 = HEAP8[$15>>0]|0;
  $17 = $16&255;
  $18 = $11 >>> 5;
  $19 = $18 & 31;
  $20 = (($14) + ($19)|0);
  $21 = HEAP8[$20>>0]|0;
  $22 = $21&255;
  $23 = $11 & 31;
  $24 = (($14) + ($23)|0);
  $25 = HEAP8[$24>>0]|0;
  $26 = $25&255;
  $27 = Math_imul($17, $1)|0;
  $28 = (($27>>>0) / 255)&-1;
  $29 = $r&255;
  $30 = (($28) + ($29))|0;
  $31 = Math_imul($22, $1)|0;
  $32 = (($31>>>0) / 255)&-1;
  $33 = $g&255;
  $34 = (($32) + ($33))|0;
  $35 = Math_imul($26, $1)|0;
  $36 = (($35>>>0) / 255)&-1;
  $37 = $b&255;
  $38 = (($36) + ($37))|0;
  $39 = $30 >>> 3;
  $40 = $39 << 10;
  $41 = $34 >>> 3;
  $42 = $41 << 5;
  $43 = $42 | $40;
  $44 = $38 >>> 3;
  $45 = $43 | $44;
  $46 = $45&65535;
  HEAP16[$9>>1] = $46;
  return 0;
  break;
 }
 case 2:  {
  $47 = ((($dst)) + 20|0);
  $48 = HEAP32[$47>>2]|0;
  $49 = ((($dst)) + 16|0);
  $50 = HEAP32[$49>>2]|0;
  $51 = Math_imul($50, $y)|0;
  $52 = (($48) + ($51)|0);
  $53 = $x << 1;
  $54 = (($52) + ($53)|0);
  $55 = HEAP16[$54>>1]|0;
  $56 = $55&65535;
  $57 = $56 >>> 10;
  $58 = $57 & 31;
  $59 = HEAP32[(125936)>>2]|0;
  $60 = (($59) + ($58)|0);
  $61 = HEAP8[$60>>0]|0;
  $62 = $61&255;
  $63 = $56 >>> 5;
  $64 = $63 & 31;
  $65 = (($59) + ($64)|0);
  $66 = HEAP8[$65>>0]|0;
  $67 = $66&255;
  $68 = $56 & 31;
  $69 = (($59) + ($68)|0);
  $70 = HEAP8[$69>>0]|0;
  $71 = $70&255;
  $72 = $r&255;
  $73 = (($62) + ($72))|0;
  $74 = ($73>>>0)>(255);
  $$1 = $74 ? 255 : $73;
  $75 = $g&255;
  $76 = (($67) + ($75))|0;
  $77 = ($76>>>0)>(255);
  $$ = $77 ? 255 : $76;
  $78 = $b&255;
  $79 = (($71) + ($78))|0;
  $80 = ($79>>>0)>(255);
  $$2 = $80 ? 255 : $79;
  $81 = $$1 >>> 3;
  $82 = $81 << 10;
  $83 = $$ >>> 3;
  $84 = $83 << 5;
  $85 = $84 | $82;
  $86 = $$2 >>> 3;
  $87 = $85 | $86;
  $88 = $87&65535;
  HEAP16[$54>>1] = $88;
  return 0;
  break;
 }
 case 4:  {
  $89 = ((($dst)) + 20|0);
  $90 = HEAP32[$89>>2]|0;
  $91 = ((($dst)) + 16|0);
  $92 = HEAP32[$91>>2]|0;
  $93 = Math_imul($92, $y)|0;
  $94 = (($90) + ($93)|0);
  $95 = $x << 1;
  $96 = (($94) + ($95)|0);
  $97 = HEAP16[$96>>1]|0;
  $98 = $97&65535;
  $99 = $98 >>> 10;
  $100 = $99 & 31;
  $101 = HEAP32[(125936)>>2]|0;
  $102 = (($101) + ($100)|0);
  $103 = HEAP8[$102>>0]|0;
  $104 = $103&255;
  $105 = $98 >>> 5;
  $106 = $105 & 31;
  $107 = (($101) + ($106)|0);
  $108 = HEAP8[$107>>0]|0;
  $109 = $108&255;
  $110 = $98 & 31;
  $111 = (($101) + ($110)|0);
  $112 = HEAP8[$111>>0]|0;
  $113 = $112&255;
  $114 = $r&255;
  $115 = Math_imul($104, $114)|0;
  $116 = (($115>>>0) / 255)&-1;
  $117 = $g&255;
  $118 = Math_imul($109, $117)|0;
  $119 = (($118>>>0) / 255)&-1;
  $120 = $b&255;
  $121 = Math_imul($113, $120)|0;
  $122 = (($121>>>0) / 255)&-1;
  $123 = $116 >>> 3;
  $124 = $123 << 10;
  $125 = $119 >>> 3;
  $126 = $125 << 5;
  $127 = $126 | $124;
  $128 = $122 >>> 3;
  $129 = $127 | $128;
  $130 = $129&65535;
  HEAP16[$96>>1] = $130;
  return 0;
  break;
 }
 default: {
  $131 = ((($dst)) + 20|0);
  $132 = HEAP32[$131>>2]|0;
  $133 = ((($dst)) + 16|0);
  $134 = HEAP32[$133>>2]|0;
  $135 = Math_imul($134, $y)|0;
  $136 = (($132) + ($135)|0);
  $137 = $x << 1;
  $138 = (($136) + ($137)|0);
  $139 = $r&255;
  $140 = $g&255;
  $141 = $b&255;
  $142 = $139 >>> 3;
  $143 = $142 << 10;
  $144 = $140 >>> 3;
  $145 = $144 << 5;
  $146 = $145 | $143;
  $147 = $141 >>> 3;
  $148 = $146 | $147;
  $149 = $148&65535;
  HEAP16[$138>>1] = $149;
  return 0;
 }
 }
 return (0)|0;
}
function _SDL_BlendPoint_RGB565($dst,$x,$y,$blendMode,$r,$g,$b,$a) {
 $dst = $dst|0;
 $x = $x|0;
 $y = $y|0;
 $blendMode = $blendMode|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $$ = 0, $$1 = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $a&255;
 $1 = $0 ^ 255;
 switch ($blendMode|0) {
 case 1:  {
  $2 = ((($dst)) + 20|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($dst)) + 16|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = Math_imul($5, $y)|0;
  $7 = (($3) + ($6)|0);
  $8 = $x << 1;
  $9 = (($7) + ($8)|0);
  $10 = HEAP16[$9>>1]|0;
  $11 = $10&65535;
  $12 = $11 >>> 11;
  $13 = HEAP32[(125936)>>2]|0;
  $14 = (($13) + ($12)|0);
  $15 = HEAP8[$14>>0]|0;
  $16 = $15&255;
  $17 = $11 >>> 5;
  $18 = $17 & 63;
  $19 = HEAP32[(125932)>>2]|0;
  $20 = (($19) + ($18)|0);
  $21 = HEAP8[$20>>0]|0;
  $22 = $21&255;
  $23 = $11 & 31;
  $24 = (($13) + ($23)|0);
  $25 = HEAP8[$24>>0]|0;
  $26 = $25&255;
  $27 = Math_imul($16, $1)|0;
  $28 = (($27>>>0) / 255)&-1;
  $29 = $r&255;
  $30 = (($28) + ($29))|0;
  $31 = Math_imul($22, $1)|0;
  $32 = (($31>>>0) / 255)&-1;
  $33 = $g&255;
  $34 = (($32) + ($33))|0;
  $35 = Math_imul($26, $1)|0;
  $36 = (($35>>>0) / 255)&-1;
  $37 = $b&255;
  $38 = (($36) + ($37))|0;
  $39 = $30 >>> 3;
  $40 = $39 << 11;
  $41 = $34 >>> 2;
  $42 = $41 << 5;
  $43 = $42 | $40;
  $44 = $38 >>> 3;
  $45 = $43 | $44;
  $46 = $45&65535;
  HEAP16[$9>>1] = $46;
  return 0;
  break;
 }
 case 2:  {
  $47 = ((($dst)) + 20|0);
  $48 = HEAP32[$47>>2]|0;
  $49 = ((($dst)) + 16|0);
  $50 = HEAP32[$49>>2]|0;
  $51 = Math_imul($50, $y)|0;
  $52 = (($48) + ($51)|0);
  $53 = $x << 1;
  $54 = (($52) + ($53)|0);
  $55 = HEAP16[$54>>1]|0;
  $56 = $55&65535;
  $57 = $56 >>> 11;
  $58 = HEAP32[(125936)>>2]|0;
  $59 = (($58) + ($57)|0);
  $60 = HEAP8[$59>>0]|0;
  $61 = $60&255;
  $62 = $56 >>> 5;
  $63 = $62 & 63;
  $64 = HEAP32[(125932)>>2]|0;
  $65 = (($64) + ($63)|0);
  $66 = HEAP8[$65>>0]|0;
  $67 = $66&255;
  $68 = $56 & 31;
  $69 = (($58) + ($68)|0);
  $70 = HEAP8[$69>>0]|0;
  $71 = $70&255;
  $72 = $r&255;
  $73 = (($61) + ($72))|0;
  $74 = ($73>>>0)>(255);
  $$1 = $74 ? 255 : $73;
  $75 = $g&255;
  $76 = (($67) + ($75))|0;
  $77 = ($76>>>0)>(255);
  $$ = $77 ? 255 : $76;
  $78 = $b&255;
  $79 = (($71) + ($78))|0;
  $80 = ($79>>>0)>(255);
  $$2 = $80 ? 255 : $79;
  $81 = $$1 >>> 3;
  $82 = $81 << 11;
  $83 = $$ >>> 2;
  $84 = $83 << 5;
  $85 = $84 | $82;
  $86 = $$2 >>> 3;
  $87 = $85 | $86;
  $88 = $87&65535;
  HEAP16[$54>>1] = $88;
  return 0;
  break;
 }
 case 4:  {
  $89 = ((($dst)) + 20|0);
  $90 = HEAP32[$89>>2]|0;
  $91 = ((($dst)) + 16|0);
  $92 = HEAP32[$91>>2]|0;
  $93 = Math_imul($92, $y)|0;
  $94 = (($90) + ($93)|0);
  $95 = $x << 1;
  $96 = (($94) + ($95)|0);
  $97 = HEAP16[$96>>1]|0;
  $98 = $97&65535;
  $99 = $98 >>> 11;
  $100 = HEAP32[(125936)>>2]|0;
  $101 = (($100) + ($99)|0);
  $102 = HEAP8[$101>>0]|0;
  $103 = $102&255;
  $104 = $98 >>> 5;
  $105 = $104 & 63;
  $106 = HEAP32[(125932)>>2]|0;
  $107 = (($106) + ($105)|0);
  $108 = HEAP8[$107>>0]|0;
  $109 = $108&255;
  $110 = $98 & 31;
  $111 = (($100) + ($110)|0);
  $112 = HEAP8[$111>>0]|0;
  $113 = $112&255;
  $114 = $r&255;
  $115 = Math_imul($103, $114)|0;
  $116 = (($115>>>0) / 255)&-1;
  $117 = $g&255;
  $118 = Math_imul($109, $117)|0;
  $119 = (($118>>>0) / 255)&-1;
  $120 = $b&255;
  $121 = Math_imul($113, $120)|0;
  $122 = (($121>>>0) / 255)&-1;
  $123 = $116 >>> 3;
  $124 = $123 << 11;
  $125 = $119 >>> 2;
  $126 = $125 << 5;
  $127 = $126 | $124;
  $128 = $122 >>> 3;
  $129 = $127 | $128;
  $130 = $129&65535;
  HEAP16[$96>>1] = $130;
  return 0;
  break;
 }
 default: {
  $131 = ((($dst)) + 20|0);
  $132 = HEAP32[$131>>2]|0;
  $133 = ((($dst)) + 16|0);
  $134 = HEAP32[$133>>2]|0;
  $135 = Math_imul($134, $y)|0;
  $136 = (($132) + ($135)|0);
  $137 = $x << 1;
  $138 = (($136) + ($137)|0);
  $139 = $r&255;
  $140 = $g&255;
  $141 = $b&255;
  $142 = $139 >>> 3;
  $143 = $142 << 11;
  $144 = $140 >>> 2;
  $145 = $144 << 5;
  $146 = $145 | $143;
  $147 = $141 >>> 3;
  $148 = $146 | $147;
  $149 = $148&65535;
  HEAP16[$138>>1] = $149;
  return 0;
 }
 }
 return (0)|0;
}
function _SW_RenderDrawLines($renderer,$points,$count) {
 $renderer = $renderer|0;
 $points = $points|0;
 $count = $count|0;
 var $$0 = 0, $$alloca_mul = 0, $$pr = 0, $$pre = 0, $$pre7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0.0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $8 = 0;
 var $9 = 0, $exitcond = 0, $exitcond6 = 0, $i$03 = 0, $i$12 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $4 = ((($1)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  HEAP32[$1>>2] = $5;
  $6 = ($5|0)==(0);
  $7 = $5;
  if ($6) {
   $8 = ((($renderer)) + 188|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (_SDL_GetWindowSurface($9)|0);
   $11 = ($10|0)==(0|0);
   do {
    if (!($11)) {
     HEAP32[$4>>2] = $10;
     HEAP32[$1>>2] = $10;
     $12 = HEAP32[$0>>2]|0;
     $13 = HEAP32[$12>>2]|0;
     $14 = ($13|0)==(0|0);
     if (!($14)) {
      $15 = ((($renderer)) + 212|0);
      (_SDL_SetClipRect($13,$15)|0);
      $$pre = HEAP32[$0>>2]|0;
      $$pre7 = HEAP32[$$pre>>2]|0;
      $16 = ($$pre7|0)==(0|0);
      if (!($16)) {
       $17 = ((($renderer)) + 276|0);
       $18 = HEAP32[$17>>2]|0;
       $19 = ($18|0)==(0);
       if ($19) {
        (_SDL_SetClipRect($$pre7,0)|0);
        break;
       } else {
        $20 = ((($renderer)) + 244|0);
        (_SDL_SetClipRect($$pre7,$20)|0);
        break;
       }
      }
     }
    }
   } while(0);
   $$pr = HEAP32[$1>>2]|0;
   $22 = $$pr;
  } else {
   $22 = $7;
  }
  $21 = ($22|0)==(0|0);
  if ($21) {
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  } else {
   $71 = $22;
  }
 } else {
  $71 = $2;
 }
 $23 = $count << 3;
 $$alloca_mul = $23;
 $24 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $25 = ((($renderer)) + 212|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ($26|0)==(0);
 if ($27) {
  $28 = ((($renderer)) + 216|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = ($29|0)==(0);
  if ($30) {
   $31 = ($count|0)>(0);
   if ($31) {
    $i$12 = 0;
    while(1) {
     $48 = (($points) + ($i$12<<3)|0);
     $49 = +HEAPF32[$48>>2];
     $50 = (~~(($49)));
     $51 = (($24) + ($i$12<<3)|0);
     HEAP32[$51>>2] = $50;
     $52 = (((($points) + ($i$12<<3)|0)) + 4|0);
     $53 = +HEAPF32[$52>>2];
     $54 = (~~(($53)));
     $55 = (((($24) + ($i$12<<3)|0)) + 4|0);
     HEAP32[$55>>2] = $54;
     $56 = (($i$12) + 1)|0;
     $exitcond = ($56|0)==($count|0);
     if ($exitcond) {
      break;
     } else {
      $i$12 = $56;
     }
    }
   }
  } else {
   label = 14;
  }
 } else {
  label = 14;
 }
 if ((label|0) == 14) {
  $32 = ($count|0)>(0);
  if ($32) {
   $33 = ((($renderer)) + 216|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = (+($26|0));
   $36 = (+($34|0));
   $i$03 = 0;
   while(1) {
    $37 = (($points) + ($i$03<<3)|0);
    $38 = +HEAPF32[$37>>2];
    $39 = $35 + $38;
    $40 = (~~(($39)));
    $41 = (($24) + ($i$03<<3)|0);
    HEAP32[$41>>2] = $40;
    $42 = (((($points) + ($i$03<<3)|0)) + 4|0);
    $43 = +HEAPF32[$42>>2];
    $44 = $36 + $43;
    $45 = (~~(($44)));
    $46 = (((($24) + ($i$03<<3)|0)) + 4|0);
    HEAP32[$46>>2] = $45;
    $47 = (($i$03) + 1)|0;
    $exitcond6 = ($47|0)==($count|0);
    if ($exitcond6) {
     break;
    } else {
     $i$03 = $47;
    }
   }
  }
 }
 $57 = ((($renderer)) + 312|0);
 $58 = HEAP32[$57>>2]|0;
 $59 = ($58|0)==(0);
 $60 = ((($renderer)) + 308|0);
 $61 = HEAP32[$60>>2]|0;
 $62 = $61&65535;
 $63 = $61&255;
 $64 = ($62&65535) >>> 8;
 $65 = $64&255;
 $66 = $61 >>> 16;
 $67 = $66&255;
 $68 = $61 >>> 24;
 $69 = $68&255;
 if ($59) {
  $70 = ((($71)) + 4|0);
  $72 = HEAP32[$70>>2]|0;
  $73 = (_SDL_MapRGBA($72,$63,$65,$67,$69)|0);
  $74 = (_SDL_DrawLines($71,$24,$count,$73)|0);
  $$0 = $74;
  STACKTOP = sp;return ($$0|0);
 } else {
  $75 = (_SDL_BlendLines($71,$24,$count,$58,$63,$65,$67,$69)|0);
  $$0 = $75;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_DrawLines($dst,$points,$count,$color) {
 $dst = $dst|0;
 $points = $points|0;
 $count = $count|0;
 $color = $color|0;
 var $$0 = 0, $$0$i$ph = 0, $$phi$trans$insert = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$02 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $x1 = 0, $x2 = 0, $y1 = 0, $y2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $x1 = sp + 24|0;
 $y1 = sp + 20|0;
 $x2 = sp + 16|0;
 $y2 = sp + 12|0;
 $0 = ($dst|0)==(0|0);
 if ($0) {
  $1 = (_SDL_SetError(603595,$vararg_buffer)|0);
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ((($dst)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 9|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 switch ($6|0) {
 case 1:  {
  $7 = ((($3)) + 8|0);
  $8 = HEAP8[$7>>0]|0;
  $9 = ($8&255)<(8);
  if ($9) {
   label = 8;
  } else {
   $$0$i$ph = 4549;
  }
  break;
 }
 case 2:  {
  $$0$i$ph = 4548;
  break;
 }
 case 4:  {
  $$0$i$ph = 4550;
  break;
 }
 default: {
  label = 8;
 }
 }
 if ((label|0) == 8) {
  $12 = (_SDL_SetError(603644,$vararg_buffer1)|0);
  $$0 = $12;
  STACKTOP = sp;return ($$0|0);
 }
 $10 = ($count|0)>(1);
 if ($10) {
  $11 = ((($dst)) + 36|0);
  $i$02 = 1;
  while(1) {
   $13 = (($i$02) + -1)|0;
   $14 = (($points) + ($13<<3)|0);
   $15 = HEAP32[$14>>2]|0;
   HEAP32[$x1>>2] = $15;
   $16 = (((($points) + ($13<<3)|0)) + 4|0);
   $17 = HEAP32[$16>>2]|0;
   HEAP32[$y1>>2] = $17;
   $18 = (($points) + ($i$02<<3)|0);
   $19 = HEAP32[$18>>2]|0;
   HEAP32[$x2>>2] = $19;
   $20 = (((($points) + ($i$02<<3)|0)) + 4|0);
   $21 = HEAP32[$20>>2]|0;
   HEAP32[$y2>>2] = $21;
   $22 = (_SDL_IntersectRectAndLine($11,$x1,$y1,$x2,$y2)|0);
   $23 = ($22|0)==(0);
   if (!($23)) {
    $24 = HEAP32[$x2>>2]|0;
    $25 = HEAP32[$18>>2]|0;
    $26 = ($24|0)==($25|0);
    $27 = HEAP32[$y2>>2]|0;
    if ($26) {
     $28 = HEAP32[$20>>2]|0;
     $29 = ($27|0)!=($28|0);
     $31 = $29;
    } else {
     $31 = 1;
    }
    $30 = $31&1;
    $32 = HEAP32[$x1>>2]|0;
    $33 = HEAP32[$y1>>2]|0;
    FUNCTION_TABLE_viiiiiii[$$0$i$ph & 8191]($dst,$32,$33,$24,$27,$color,$30);
   }
   $34 = (($i$02) + 1)|0;
   $exitcond = ($34|0)==($count|0);
   if ($exitcond) {
    break;
   } else {
    $i$02 = $34;
   }
  }
 }
 $35 = HEAP32[$points>>2]|0;
 $36 = (($count) + -1)|0;
 $37 = (($points) + ($36<<3)|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = ($35|0)==($38|0);
 if ($39) {
  $40 = ((($points)) + 4|0);
  $41 = HEAP32[$40>>2]|0;
  $42 = (((($points) + ($36<<3)|0)) + 4|0);
  $43 = HEAP32[$42>>2]|0;
  $44 = ($41|0)==($43|0);
  if ($44) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  } else {
   $45 = $43;
  }
 } else {
  $$phi$trans$insert = (((($points) + ($36<<3)|0)) + 4|0);
  $$pre4 = HEAP32[$$phi$trans$insert>>2]|0;
  $45 = $$pre4;
 }
 (_SDL_DrawPoint($dst,$38,$45,$color)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_DrawLine2($dst,$x1,$y1,$x2,$y2,$color,$draw_end) {
 $dst = $dst|0;
 $x1 = $x1|0;
 $y1 = $y1|0;
 $x2 = $x2|0;
 $y2 = $y2|0;
 $color = $color|0;
 $draw_end = $draw_end|0;
 var $$ = 0, $$1 = 0, $$2 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$xinc1$0 = 0, $$xinc1$0$$3 = 0, $$xinc115$0 = 0, $$xinc115$0$$4 = 0, $$xinc128$0 = 0, $$xinc128$0$$5 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0;
 var $196 = 0, $197 = 0, $198 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $_a = 0, $_b = 0, $_g = 0, $_r = 0, $d$0 = 0, $d$129 = 0, $d$2 = 0, $d11$0 = 0, $d11$134 = 0;
 var $d11$2 = 0, $d24$0 = 0, $d24$124 = 0, $d24$2 = 0, $dinc1$0 = 0, $dinc1$0$dinc2$0 = 0, $dinc112$0 = 0, $dinc112$0$dinc213$0 = 0, $dinc125$0 = 0, $dinc125$0$dinc226$0 = 0, $dinc2$0 = 0, $dinc2$0$in = 0, $dinc213$0 = 0, $dinc213$0$in = 0, $dinc226$0 = 0, $dinc226$0$in = 0, $i$028 = 0, $i20$023 = 0, $i7$033 = 0, $length$0$ph = 0;
 var $length$016 = 0, $length1$0$ph = 0, $length1$018 = 0, $length4$0 = 0, $length4$0$ = 0, $length4$121 = 0, $not$ = 0, $not$6 = 0, $not$7 = 0, $not$8 = 0, $numpixels$0 = 0, $numpixels$0$in = 0, $numpixels$0$in$numpixels$0 = 0, $numpixels10$0 = 0, $numpixels10$0$in = 0, $numpixels10$0$in$numpixels10$0 = 0, $numpixels23$0 = 0, $numpixels23$0$in = 0, $numpixels23$0$in$numpixels23$0 = 0, $pitch5$0 = 0;
 var $pitch5$0$v = 0, $pitch5$1 = 0, $pitch5$1$v = 0, $pitch5$2 = 0, $pixel$1$ph = 0, $pixel$115 = 0, $pixel3$1$ph = 0, $pixel3$117 = 0, $pixel6$1 = 0, $pixel6$220 = 0, $switch$split2D = 0, $x$030 = 0, $x$1 = 0, $x14$035 = 0, $x14$1 = 0, $x27$025 = 0, $x27$1 = 0, $xinc1$0 = 0, $xinc115$0 = 0, $xinc128$0 = 0;
 var $y$031 = 0, $y$1 = 0, $y17$036 = 0, $y17$1 = 0, $y30$026 = 0, $y30$1 = 0, $yinc1$0 = 0, $yinc1$1 = 0, $yinc1$1$yinc2$1 = 0, $yinc118$0 = 0, $yinc118$1 = 0, $yinc118$1$yinc219$1 = 0, $yinc131$0 = 0, $yinc131$1 = 0, $yinc131$1$yinc232$1 = 0, $yinc2$1 = 0, $yinc219$1 = 0, $yinc232$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $_r = sp + 3|0;
 $_g = sp + 2|0;
 $_b = sp + 1|0;
 $_a = sp;
 $0 = ($y2|0)==($y1|0);
 if ($0) {
  $1 = ((($dst)) + 16|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ((($dst)) + 4|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ((($4)) + 9|0);
  $6 = HEAP8[$5>>0]|0;
  $7 = $6&255;
  $8 = (($2|0) / ($7|0))&-1;
  $9 = ($x2|0)<($x1|0);
  $10 = ((($dst)) + 20|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = Math_imul($8, $y2)|0;
  $13 = (($11) + ($12<<1)|0);
  if ($9) {
   $19 = (($13) + ($x2<<1)|0);
   $20 = ($draw_end|0)!=(0);
   $21 = ((($19)) + 2|0);
   $$ = $20 ? $19 : $21;
   $22 = (($x1) - ($x2))|0;
   $23 = $20&1;
   $24 = (($22) + ($23))|0;
   $length$0$ph = $24;$pixel$1$ph = $$;
  } else {
   $14 = (($13) + ($x1<<1)|0);
   $15 = ($draw_end|0)!=(0);
   $16 = (($x2) - ($x1))|0;
   $17 = $15&1;
   $18 = (($16) + ($17))|0;
   $length$0$ph = $18;$pixel$1$ph = $14;
  }
  $25 = ($length$0$ph|0)==(0);
  if ($25) {
   STACKTOP = sp;return;
  }
  $26 = $color&65535;
  $length$016 = $length$0$ph;$pixel$115 = $pixel$1$ph;
  while(1) {
   $27 = (($length$016) + -1)|0;
   HEAP16[$pixel$115>>1] = $26;
   $28 = ((($pixel$115)) + 2|0);
   $29 = ($27|0)==(0);
   if ($29) {
    break;
   } else {
    $length$016 = $27;$pixel$115 = $28;
   }
  }
  STACKTOP = sp;return;
 }
 $30 = ($x2|0)==($x1|0);
 if ($30) {
  $31 = ((($dst)) + 16|0);
  $32 = HEAP32[$31>>2]|0;
  $33 = ((($dst)) + 4|0);
  $34 = HEAP32[$33>>2]|0;
  $35 = ((($34)) + 9|0);
  $36 = HEAP8[$35>>0]|0;
  $37 = $36&255;
  $38 = (($32|0) / ($37|0))&-1;
  $39 = ($y2|0)<($y1|0);
  $40 = ((($dst)) + 20|0);
  $41 = HEAP32[$40>>2]|0;
  if ($39) {
   $49 = Math_imul($38, $y2)|0;
   $50 = (($41) + ($49<<1)|0);
   $51 = (($50) + ($x2<<1)|0);
   $52 = ($draw_end|0)!=(0);
   $53 = (($51) + ($38<<1)|0);
   $$1 = $52 ? $51 : $53;
   $54 = (($y1) - ($y2))|0;
   $55 = $52&1;
   $56 = (($54) + ($55))|0;
   $length1$0$ph = $56;$pixel3$1$ph = $$1;
  } else {
   $42 = Math_imul($38, $y1)|0;
   $43 = (($41) + ($42<<1)|0);
   $44 = (($43) + ($x2<<1)|0);
   $45 = ($draw_end|0)!=(0);
   $46 = (($y2) - ($y1))|0;
   $47 = $45&1;
   $48 = (($46) + ($47))|0;
   $length1$0$ph = $48;$pixel3$1$ph = $44;
  }
  $57 = ($length1$0$ph|0)==(0);
  if ($57) {
   STACKTOP = sp;return;
  }
  $58 = $color&65535;
  $length1$018 = $length1$0$ph;$pixel3$117 = $pixel3$1$ph;
  while(1) {
   $59 = (($length1$018) + -1)|0;
   HEAP16[$pixel3$117>>1] = $58;
   $60 = (($pixel3$117) + ($38<<1)|0);
   $61 = ($59|0)==(0);
   if ($61) {
    break;
   } else {
    $length1$018 = $59;$pixel3$117 = $60;
   }
  }
  STACKTOP = sp;return;
 }
 $62 = (($x1) - ($x2))|0;
 $63 = ($62|0)<(0);
 $64 = (0 - ($62))|0;
 $65 = $63 ? $64 : $62;
 $66 = (($y1) - ($y2))|0;
 $67 = ($66|0)<(0);
 $68 = (0 - ($66))|0;
 $69 = $67 ? $68 : $66;
 $70 = ($65|0)==($69|0);
 if ($70) {
  $71 = ((($dst)) + 16|0);
  $72 = HEAP32[$71>>2]|0;
  $73 = ((($dst)) + 4|0);
  $74 = HEAP32[$73>>2]|0;
  $75 = ((($74)) + 9|0);
  $76 = HEAP8[$75>>0]|0;
  $77 = $76&255;
  $78 = (($72|0) / ($77|0))&-1;
  $79 = ($y2|0)<($y1|0);
  $80 = ((($dst)) + 20|0);
  $81 = HEAP32[$80>>2]|0;
  if ($79) {
   $87 = Math_imul($78, $y2)|0;
   $88 = (($81) + ($87<<1)|0);
   $89 = (($88) + ($x2<<1)|0);
   $90 = ($x2|0)>($x1|0);
   $pitch5$1$v = $90 ? -1 : 1;
   $pitch5$1 = (($78) + ($pitch5$1$v))|0;
   $91 = ($draw_end|0)==(0);
   $92 = (($89) + ($pitch5$1<<1)|0);
   $$2 = $91 ? $92 : $89;
   $length4$0 = $66;$pitch5$2 = $pitch5$1;$pixel6$1 = $$2;
  } else {
   $82 = Math_imul($78, $y1)|0;
   $83 = (($81) + ($82<<1)|0);
   $84 = (($83) + ($x1<<1)|0);
   $85 = ($x2|0)<($x1|0);
   $pitch5$0$v = $85 ? -1 : 1;
   $pitch5$0 = (($78) + ($pitch5$0$v))|0;
   $86 = (($y2) - ($y1))|0;
   $length4$0 = $86;$pitch5$2 = $pitch5$0;$pixel6$1 = $84;
  }
  $not$8 = ($draw_end|0)!=(0);
  $93 = $not$8&1;
  $length4$0$ = (($length4$0) + ($93))|0;
  $94 = ($length4$0$|0)==(0);
  if ($94) {
   STACKTOP = sp;return;
  }
  $95 = $color&65535;
  $length4$121 = $length4$0$;$pixel6$220 = $pixel6$1;
  while(1) {
   $96 = (($length4$121) + -1)|0;
   HEAP16[$pixel6$220>>1] = $95;
   $97 = (($pixel6$220) + ($pitch5$2<<1)|0);
   $98 = ($96|0)==(0);
   if ($98) {
    break;
   } else {
    $length4$121 = $96;$pixel6$220 = $97;
   }
  }
  STACKTOP = sp;return;
 }
 $99 = ((($dst)) + 4|0);
 $100 = HEAP32[$99>>2]|0;
 _SDL_GetRGBA($color,$100,$_r,$_g,$_b,$_a);
 $101 = ((($100)) + 12|0);
 $102 = HEAP32[$101>>2]|0;
 $switch$split2D = ($102|0)<(63488);
 L40: do {
  if ($switch$split2D) {
   switch ($102|0) {
   case 31744:  {
    break;
   }
   default: {
    label = 35;
    break L40;
   }
   }
   $103 = (($x2) - ($x1))|0;
   $104 = ($103|0)<(0);
   $105 = (0 - ($103))|0;
   $106 = $104 ? $105 : $103;
   $107 = (($y2) - ($y1))|0;
   $108 = ($107|0)<(0);
   $109 = (0 - ($107))|0;
   $110 = $108 ? $109 : $107;
   $111 = ($106|0)<($110|0);
   if ($111) {
    $115 = $106 << 1;
    $116 = (($115) - ($110))|0;
    $117 = (($106) - ($110))|0;
    $d$0 = $116;$dinc1$0 = $115;$dinc2$0$in = $117;$numpixels$0$in = $110;$xinc1$0 = 0;$yinc1$0 = 1;
   } else {
    $112 = $110 << 1;
    $113 = (($112) - ($106))|0;
    $114 = (($110) - ($106))|0;
    $d$0 = $113;$dinc1$0 = $112;$dinc2$0$in = $114;$numpixels$0$in = $106;$xinc1$0 = 1;$yinc1$0 = 0;
   }
   $dinc2$0 = $dinc2$0$in << 1;
   $118 = ($x2|0)<($x1|0);
   $119 = (0 - ($xinc1$0))|0;
   $$xinc1$0 = $118 ? $119 : $xinc1$0;
   $$3 = $118 ? -1 : 1;
   $120 = ($y2|0)<($y1|0);
   $121 = (0 - ($yinc1$0))|0;
   $yinc1$1 = $120 ? $121 : $yinc1$0;
   $yinc2$1 = $120 ? -1 : 1;
   $not$6 = ($draw_end|0)!=(0);
   $numpixels$0 = $not$6&1;
   $numpixels$0$in$numpixels$0 = (($numpixels$0$in) + ($numpixels$0))|0;
   $122 = ($numpixels$0$in$numpixels$0|0)>(0);
   if ($122) {
    $123 = $color&65535;
    $124 = ((($dst)) + 20|0);
    $125 = HEAP32[$124>>2]|0;
    $126 = ((($dst)) + 16|0);
    $127 = HEAP32[$126>>2]|0;
    $d$129 = $d$0;$i$028 = 0;$x$030 = $x1;$y$031 = $y1;
    while(1) {
     $128 = Math_imul($127, $y$031)|0;
     $129 = (($125) + ($128)|0);
     $130 = $x$030 << 1;
     $131 = (($129) + ($130)|0);
     HEAP16[$131>>1] = $123;
     $132 = ($d$129|0)<(0);
     $dinc1$0$dinc2$0 = $132 ? $dinc1$0 : $dinc2$0;
     $$xinc1$0$$3 = $132 ? $$xinc1$0 : $$3;
     $yinc1$1$yinc2$1 = $132 ? $yinc1$1 : $yinc2$1;
     $y$1 = (($y$031) + ($yinc1$1$yinc2$1))|0;
     $x$1 = (($$xinc1$0$$3) + ($x$030))|0;
     $d$2 = (($dinc1$0$dinc2$0) + ($d$129))|0;
     $133 = (($i$028) + 1)|0;
     $134 = ($133|0)<($numpixels$0$in$numpixels$0|0);
     if ($134) {
      $d$129 = $d$2;$i$028 = $133;$x$030 = $x$1;$y$031 = $y$1;
     } else {
      break;
     }
    }
   }
  } else {
   switch ($102|0) {
   case 63488:  {
    break;
   }
   default: {
    label = 35;
    break L40;
   }
   }
   $135 = (($x2) - ($x1))|0;
   $136 = ($135|0)<(0);
   $137 = (0 - ($135))|0;
   $138 = $136 ? $137 : $135;
   $139 = (($y2) - ($y1))|0;
   $140 = ($139|0)<(0);
   $141 = (0 - ($139))|0;
   $142 = $140 ? $141 : $139;
   $143 = ($138|0)<($142|0);
   if ($143) {
    $147 = $138 << 1;
    $148 = (($147) - ($142))|0;
    $149 = (($138) - ($142))|0;
    $d11$0 = $148;$dinc112$0 = $147;$dinc213$0$in = $149;$numpixels10$0$in = $142;$xinc115$0 = 0;$yinc118$0 = 1;
   } else {
    $144 = $142 << 1;
    $145 = (($144) - ($138))|0;
    $146 = (($142) - ($138))|0;
    $d11$0 = $145;$dinc112$0 = $144;$dinc213$0$in = $146;$numpixels10$0$in = $138;$xinc115$0 = 1;$yinc118$0 = 0;
   }
   $dinc213$0 = $dinc213$0$in << 1;
   $150 = ($x2|0)<($x1|0);
   $151 = (0 - ($xinc115$0))|0;
   $$xinc115$0 = $150 ? $151 : $xinc115$0;
   $$4 = $150 ? -1 : 1;
   $152 = ($y2|0)<($y1|0);
   $153 = (0 - ($yinc118$0))|0;
   $yinc118$1 = $152 ? $153 : $yinc118$0;
   $yinc219$1 = $152 ? -1 : 1;
   $not$ = ($draw_end|0)!=(0);
   $numpixels10$0 = $not$&1;
   $numpixels10$0$in$numpixels10$0 = (($numpixels10$0$in) + ($numpixels10$0))|0;
   $154 = ($numpixels10$0$in$numpixels10$0|0)>(0);
   if ($154) {
    $155 = $color&65535;
    $156 = ((($dst)) + 20|0);
    $157 = HEAP32[$156>>2]|0;
    $158 = ((($dst)) + 16|0);
    $159 = HEAP32[$158>>2]|0;
    $d11$134 = $d11$0;$i7$033 = 0;$x14$035 = $x1;$y17$036 = $y1;
    while(1) {
     $160 = Math_imul($159, $y17$036)|0;
     $161 = (($157) + ($160)|0);
     $162 = $x14$035 << 1;
     $163 = (($161) + ($162)|0);
     HEAP16[$163>>1] = $155;
     $164 = ($d11$134|0)<(0);
     $dinc112$0$dinc213$0 = $164 ? $dinc112$0 : $dinc213$0;
     $$xinc115$0$$4 = $164 ? $$xinc115$0 : $$4;
     $yinc118$1$yinc219$1 = $164 ? $yinc118$1 : $yinc219$1;
     $y17$1 = (($y17$036) + ($yinc118$1$yinc219$1))|0;
     $x14$1 = (($$xinc115$0$$4) + ($x14$035))|0;
     $d11$2 = (($dinc112$0$dinc213$0) + ($d11$134))|0;
     $165 = (($i7$033) + 1)|0;
     $166 = ($165|0)<($numpixels10$0$in$numpixels10$0|0);
     if ($166) {
      $d11$134 = $d11$2;$i7$033 = $165;$x14$035 = $x14$1;$y17$036 = $y17$1;
     } else {
      break;
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 35) {
  $167 = (($x2) - ($x1))|0;
  $168 = ($167|0)<(0);
  $169 = (0 - ($167))|0;
  $170 = $168 ? $169 : $167;
  $171 = (($y2) - ($y1))|0;
  $172 = ($171|0)<(0);
  $173 = (0 - ($171))|0;
  $174 = $172 ? $173 : $171;
  $175 = ($170|0)<($174|0);
  if ($175) {
   $179 = $170 << 1;
   $180 = (($179) - ($174))|0;
   $181 = (($170) - ($174))|0;
   $d24$0 = $180;$dinc125$0 = $179;$dinc226$0$in = $181;$numpixels23$0$in = $174;$xinc128$0 = 0;$yinc131$0 = 1;
  } else {
   $176 = $174 << 1;
   $177 = (($176) - ($170))|0;
   $178 = (($174) - ($170))|0;
   $d24$0 = $177;$dinc125$0 = $176;$dinc226$0$in = $178;$numpixels23$0$in = $170;$xinc128$0 = 1;$yinc131$0 = 0;
  }
  $dinc226$0 = $dinc226$0$in << 1;
  $182 = ($x2|0)<($x1|0);
  $183 = (0 - ($xinc128$0))|0;
  $$xinc128$0 = $182 ? $183 : $xinc128$0;
  $$5 = $182 ? -1 : 1;
  $184 = ($y2|0)<($y1|0);
  $185 = (0 - ($yinc131$0))|0;
  $yinc131$1 = $184 ? $185 : $yinc131$0;
  $yinc232$1 = $184 ? -1 : 1;
  $not$7 = ($draw_end|0)!=(0);
  $numpixels23$0 = $not$7&1;
  $numpixels23$0$in$numpixels23$0 = (($numpixels23$0$in) + ($numpixels23$0))|0;
  $186 = ($numpixels23$0$in$numpixels23$0|0)>(0);
  if ($186) {
   $187 = $color&65535;
   $188 = ((($dst)) + 20|0);
   $189 = HEAP32[$188>>2]|0;
   $190 = ((($dst)) + 16|0);
   $191 = HEAP32[$190>>2]|0;
   $d24$124 = $d24$0;$i20$023 = 0;$x27$025 = $x1;$y30$026 = $y1;
   while(1) {
    $192 = Math_imul($191, $y30$026)|0;
    $193 = (($189) + ($192)|0);
    $194 = $x27$025 << 1;
    $195 = (($193) + ($194)|0);
    HEAP16[$195>>1] = $187;
    $196 = ($d24$124|0)<(0);
    $dinc125$0$dinc226$0 = $196 ? $dinc125$0 : $dinc226$0;
    $$xinc128$0$$5 = $196 ? $$xinc128$0 : $$5;
    $yinc131$1$yinc232$1 = $196 ? $yinc131$1 : $yinc232$1;
    $y30$1 = (($y30$026) + ($yinc131$1$yinc232$1))|0;
    $x27$1 = (($$xinc128$0$$5) + ($x27$025))|0;
    $d24$2 = (($dinc125$0$dinc226$0) + ($d24$124))|0;
    $197 = (($i20$023) + 1)|0;
    $198 = ($197|0)<($numpixels23$0$in$numpixels23$0|0);
    if ($198) {
     $d24$124 = $d24$2;$i20$023 = $197;$x27$025 = $x27$1;$y30$026 = $y30$1;
    } else {
     break;
    }
   }
  }
 }
 STACKTOP = sp;return;
}
function _SDL_GetRGBA($pixel,$format,$r,$g,$b,$a) {
 $pixel = $pixel|0;
 $format = $format|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($format)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $3 = ((($format)) + 12|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = $4 & $pixel;
  $6 = ((($format)) + 32|0);
  $7 = HEAP8[$6>>0]|0;
  $8 = $7&255;
  $9 = $5 >>> $8;
  $10 = ((($format)) + 28|0);
  $11 = HEAP8[$10>>0]|0;
  $12 = $11&255;
  $13 = (125924 + ($12<<2)|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = (($14) + ($9)|0);
  $16 = HEAP8[$15>>0]|0;
  HEAP8[$r>>0] = $16;
  $17 = ((($format)) + 16|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = $18 & $pixel;
  $20 = ((($format)) + 33|0);
  $21 = HEAP8[$20>>0]|0;
  $22 = $21&255;
  $23 = $19 >>> $22;
  $24 = ((($format)) + 29|0);
  $25 = HEAP8[$24>>0]|0;
  $26 = $25&255;
  $27 = (125924 + ($26<<2)|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = (($28) + ($23)|0);
  $30 = HEAP8[$29>>0]|0;
  HEAP8[$g>>0] = $30;
  $31 = ((($format)) + 20|0);
  $32 = HEAP32[$31>>2]|0;
  $33 = $32 & $pixel;
  $34 = ((($format)) + 34|0);
  $35 = HEAP8[$34>>0]|0;
  $36 = $35&255;
  $37 = $33 >>> $36;
  $38 = ((($format)) + 30|0);
  $39 = HEAP8[$38>>0]|0;
  $40 = $39&255;
  $41 = (125924 + ($40<<2)|0);
  $42 = HEAP32[$41>>2]|0;
  $43 = (($42) + ($37)|0);
  $44 = HEAP8[$43>>0]|0;
  HEAP8[$b>>0] = $44;
  $45 = ((($format)) + 24|0);
  $46 = HEAP32[$45>>2]|0;
  $47 = $46 & $pixel;
  $48 = ((($format)) + 35|0);
  $49 = HEAP8[$48>>0]|0;
  $50 = $49&255;
  $51 = $47 >>> $50;
  $52 = ((($format)) + 31|0);
  $53 = HEAP8[$52>>0]|0;
  $54 = $53&255;
  $55 = (125924 + ($54<<2)|0);
  $56 = HEAP32[$55>>2]|0;
  $57 = (($56) + ($51)|0);
  $58 = HEAP8[$57>>0]|0;
  HEAP8[$a>>0] = $58;
  return;
 }
 $59 = HEAP32[$1>>2]|0;
 $60 = ($59>>>0)>($pixel>>>0);
 if ($60) {
  $61 = ((($1)) + 4|0);
  $62 = HEAP32[$61>>2]|0;
  $63 = (($62) + ($pixel<<2)|0);
  $64 = HEAP8[$63>>0]|0;
  HEAP8[$r>>0] = $64;
  $65 = HEAP32[$0>>2]|0;
  $66 = ((($65)) + 4|0);
  $67 = HEAP32[$66>>2]|0;
  $68 = (((($67) + ($pixel<<2)|0)) + 1|0);
  $69 = HEAP8[$68>>0]|0;
  HEAP8[$g>>0] = $69;
  $70 = HEAP32[$0>>2]|0;
  $71 = ((($70)) + 4|0);
  $72 = HEAP32[$71>>2]|0;
  $73 = (((($72) + ($pixel<<2)|0)) + 2|0);
  $74 = HEAP8[$73>>0]|0;
  HEAP8[$b>>0] = $74;
  $75 = HEAP32[$0>>2]|0;
  $76 = ((($75)) + 4|0);
  $77 = HEAP32[$76>>2]|0;
  $78 = (((($77) + ($pixel<<2)|0)) + 3|0);
  $79 = HEAP8[$78>>0]|0;
  HEAP8[$a>>0] = $79;
  return;
 } else {
  HEAP8[$a>>0] = 0;
  HEAP8[$b>>0] = 0;
  HEAP8[$g>>0] = 0;
  HEAP8[$r>>0] = 0;
  return;
 }
}
function _SDL_DrawLine1($dst,$x1,$y1,$x2,$y2,$color,$draw_end) {
 $dst = $dst|0;
 $x1 = $x1|0;
 $y1 = $y1|0;
 $x2 = $x2|0;
 $y2 = $y2|0;
 $color = $color|0;
 $draw_end = $draw_end|0;
 var $$ = 0, $$1 = 0, $$2 = 0, $$3 = 0, $$xinc1$0 = 0, $$xinc1$0$$3 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $d$0 = 0, $d$113 = 0, $d$2 = 0, $dinc1$0 = 0, $dinc1$0$dinc2$0 = 0, $dinc2$0 = 0, $dinc2$0$in = 0, $i$012 = 0, $length$0 = 0;
 var $length1$0$ph = 0, $length1$07 = 0, $length4$0 = 0, $length4$0$ = 0, $length4$19 = 0, $not$ = 0, $not$4 = 0, $numpixels$0 = 0, $numpixels$0$in = 0, $numpixels$0$in$numpixels$0 = 0, $pitch5$0 = 0, $pitch5$0$v = 0, $pitch5$1 = 0, $pitch5$1$v = 0, $pitch5$2 = 0, $pixel$1 = 0, $pixel3$1$ph = 0, $pixel3$18 = 0, $pixel6$1 = 0, $pixel6$210 = 0;
 var $x$014 = 0, $x$1 = 0, $xinc1$0 = 0, $y$015 = 0, $y$1 = 0, $yinc1$0 = 0, $yinc1$1 = 0, $yinc1$1$yinc2$1 = 0, $yinc2$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($y2|0)==($y1|0);
 if ($0) {
  $1 = ((($dst)) + 16|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ((($dst)) + 4|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ((($4)) + 9|0);
  $6 = HEAP8[$5>>0]|0;
  $7 = $6&255;
  $8 = (($2|0) / ($7|0))&-1;
  $9 = ($x1|0)>($x2|0);
  $10 = ((($dst)) + 20|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = Math_imul($8, $y2)|0;
  $13 = (($11) + ($12)|0);
  if ($9) {
   $19 = (($13) + ($x2)|0);
   $20 = ($draw_end|0)!=(0);
   $21 = ((($19)) + 1|0);
   $$ = $20 ? $19 : $21;
   $22 = (($x1) - ($x2))|0;
   $23 = $20&1;
   $24 = (($22) + ($23))|0;
   $length$0 = $24;$pixel$1 = $$;
  } else {
   $14 = (($13) + ($x1)|0);
   $15 = ($draw_end|0)!=(0);
   $16 = (($x2) - ($x1))|0;
   $17 = $15&1;
   $18 = (($16) + ($17))|0;
   $length$0 = $18;$pixel$1 = $14;
  }
  (_SDL_memset($pixel$1,$color,$length$0)|0);
  return;
 }
 $25 = ($x1|0)==($x2|0);
 if ($25) {
  $26 = ((($dst)) + 16|0);
  $27 = HEAP32[$26>>2]|0;
  $28 = ((($dst)) + 4|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = ((($29)) + 9|0);
  $31 = HEAP8[$30>>0]|0;
  $32 = $31&255;
  $33 = (($27|0) / ($32|0))&-1;
  $34 = ($y2|0)<($y1|0);
  $35 = ((($dst)) + 20|0);
  $36 = HEAP32[$35>>2]|0;
  if ($34) {
   $44 = Math_imul($33, $y2)|0;
   $45 = (($36) + ($44)|0);
   $46 = (($45) + ($x1)|0);
   $47 = ($draw_end|0)!=(0);
   $48 = (($46) + ($33)|0);
   $$1 = $47 ? $46 : $48;
   $49 = (($y1) - ($y2))|0;
   $50 = $47&1;
   $51 = (($49) + ($50))|0;
   $length1$0$ph = $51;$pixel3$1$ph = $$1;
  } else {
   $37 = Math_imul($33, $y1)|0;
   $38 = (($36) + ($37)|0);
   $39 = (($38) + ($x1)|0);
   $40 = ($draw_end|0)!=(0);
   $41 = (($y2) - ($y1))|0;
   $42 = $40&1;
   $43 = (($41) + ($42))|0;
   $length1$0$ph = $43;$pixel3$1$ph = $39;
  }
  $52 = ($length1$0$ph|0)==(0);
  if ($52) {
   return;
  }
  $53 = $color&255;
  $length1$07 = $length1$0$ph;$pixel3$18 = $pixel3$1$ph;
  while(1) {
   $54 = (($length1$07) + -1)|0;
   HEAP8[$pixel3$18>>0] = $53;
   $55 = (($pixel3$18) + ($33)|0);
   $56 = ($54|0)==(0);
   if ($56) {
    break;
   } else {
    $length1$07 = $54;$pixel3$18 = $55;
   }
  }
  return;
 }
 $57 = (($x1) - ($x2))|0;
 $58 = ($57|0)<(0);
 $59 = (0 - ($57))|0;
 $60 = $58 ? $59 : $57;
 $61 = (($y1) - ($y2))|0;
 $62 = ($61|0)<(0);
 $63 = (0 - ($61))|0;
 $64 = $62 ? $63 : $61;
 $65 = ($60|0)==($64|0);
 if ($65) {
  $66 = ((($dst)) + 16|0);
  $67 = HEAP32[$66>>2]|0;
  $68 = ((($dst)) + 4|0);
  $69 = HEAP32[$68>>2]|0;
  $70 = ((($69)) + 9|0);
  $71 = HEAP8[$70>>0]|0;
  $72 = $71&255;
  $73 = (($67|0) / ($72|0))&-1;
  $74 = ($y2|0)<($y1|0);
  $75 = ((($dst)) + 20|0);
  $76 = HEAP32[$75>>2]|0;
  if ($74) {
   $82 = Math_imul($73, $y2)|0;
   $83 = (($76) + ($82)|0);
   $84 = (($83) + ($x2)|0);
   $85 = ($x2|0)>($x1|0);
   $pitch5$1$v = $85 ? -1 : 1;
   $pitch5$1 = (($73) + ($pitch5$1$v))|0;
   $86 = ($draw_end|0)==(0);
   $87 = (($84) + ($pitch5$1)|0);
   $$2 = $86 ? $87 : $84;
   $length4$0 = $61;$pitch5$2 = $pitch5$1;$pixel6$1 = $$2;
  } else {
   $77 = Math_imul($73, $y1)|0;
   $78 = (($76) + ($77)|0);
   $79 = (($78) + ($x1)|0);
   $80 = ($x1|0)>($x2|0);
   $pitch5$0$v = $80 ? -1 : 1;
   $pitch5$0 = (($73) + ($pitch5$0$v))|0;
   $81 = (($y2) - ($y1))|0;
   $length4$0 = $81;$pitch5$2 = $pitch5$0;$pixel6$1 = $79;
  }
  $not$4 = ($draw_end|0)!=(0);
  $88 = $not$4&1;
  $length4$0$ = (($length4$0) + ($88))|0;
  $89 = ($length4$0$|0)==(0);
  if ($89) {
   return;
  }
  $90 = $color&255;
  $length4$19 = $length4$0$;$pixel6$210 = $pixel6$1;
  while(1) {
   $91 = (($length4$19) + -1)|0;
   HEAP8[$pixel6$210>>0] = $90;
   $92 = (($pixel6$210) + ($pitch5$2)|0);
   $93 = ($91|0)==(0);
   if ($93) {
    break;
   } else {
    $length4$19 = $91;$pixel6$210 = $92;
   }
  }
  return;
 } else {
  $94 = (($x2) - ($x1))|0;
  $95 = ($94|0)<(0);
  $96 = (0 - ($94))|0;
  $97 = $95 ? $96 : $94;
  $98 = (($y2) - ($y1))|0;
  $99 = ($98|0)<(0);
  $100 = (0 - ($98))|0;
  $101 = $99 ? $100 : $98;
  $102 = ($97|0)<($101|0);
  if ($102) {
   $106 = $97 << 1;
   $107 = (($106) - ($101))|0;
   $108 = (($97) - ($101))|0;
   $d$0 = $107;$dinc1$0 = $106;$dinc2$0$in = $108;$numpixels$0$in = $101;$xinc1$0 = 0;$yinc1$0 = 1;
  } else {
   $103 = $101 << 1;
   $104 = (($103) - ($97))|0;
   $105 = (($101) - ($97))|0;
   $d$0 = $104;$dinc1$0 = $103;$dinc2$0$in = $105;$numpixels$0$in = $97;$xinc1$0 = 1;$yinc1$0 = 0;
  }
  $dinc2$0 = $dinc2$0$in << 1;
  $109 = ($x1|0)>($x2|0);
  $110 = (0 - ($xinc1$0))|0;
  $$xinc1$0 = $109 ? $110 : $xinc1$0;
  $$3 = $109 ? -1 : 1;
  $111 = ($y2|0)<($y1|0);
  $112 = (0 - ($yinc1$0))|0;
  $yinc1$1 = $111 ? $112 : $yinc1$0;
  $yinc2$1 = $111 ? -1 : 1;
  $not$ = ($draw_end|0)!=(0);
  $numpixels$0 = $not$&1;
  $numpixels$0$in$numpixels$0 = (($numpixels$0$in) + ($numpixels$0))|0;
  $113 = ($numpixels$0$in$numpixels$0|0)>(0);
  if (!($113)) {
   return;
  }
  $114 = $color&255;
  $115 = ((($dst)) + 20|0);
  $116 = ((($dst)) + 16|0);
  $d$113 = $d$0;$i$012 = 0;$x$014 = $x1;$y$015 = $y1;
  while(1) {
   $117 = HEAP32[$115>>2]|0;
   $118 = HEAP32[$116>>2]|0;
   $119 = Math_imul($118, $y$015)|0;
   $120 = (($117) + ($119)|0);
   $121 = (($120) + ($x$014)|0);
   HEAP8[$121>>0] = $114;
   $122 = ($d$113|0)<(0);
   $dinc1$0$dinc2$0 = $122 ? $dinc1$0 : $dinc2$0;
   $$xinc1$0$$3 = $122 ? $$xinc1$0 : $$3;
   $yinc1$1$yinc2$1 = $122 ? $yinc1$1 : $yinc2$1;
   $y$1 = (($y$015) + ($yinc1$1$yinc2$1))|0;
   $x$1 = (($$xinc1$0$$3) + ($x$014))|0;
   $d$2 = (($dinc1$0$dinc2$0) + ($d$113))|0;
   $123 = (($i$012) + 1)|0;
   $124 = ($123|0)<($numpixels$0$in$numpixels$0|0);
   if ($124) {
    $d$113 = $d$2;$i$012 = $123;$x$014 = $x$1;$y$015 = $y$1;
   } else {
    break;
   }
  }
  return;
 }
}
function _SDL_DrawLine4($dst,$x1,$y1,$x2,$y2,$color,$draw_end) {
 $dst = $dst|0;
 $x1 = $x1|0;
 $y1 = $y1|0;
 $x2 = $x2|0;
 $y2 = $y2|0;
 $color = $color|0;
 $draw_end = $draw_end|0;
 var $$ = 0, $$1 = 0, $$2 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$xinc1$0 = 0, $$xinc1$0$$3 = 0, $$xinc115$0 = 0, $$xinc115$0$$4 = 0, $$xinc128$0 = 0, $$xinc128$0$$5 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $_a = 0, $_b = 0, $_g = 0, $_r = 0, $d$0 = 0, $d$124 = 0, $d$2 = 0, $d11$0 = 0, $d11$129 = 0, $d11$2 = 0, $d24$0 = 0, $d24$134 = 0, $d24$2 = 0, $dinc1$0 = 0, $dinc1$0$dinc2$0 = 0, $dinc112$0 = 0, $dinc112$0$dinc213$0 = 0, $dinc125$0 = 0, $dinc125$0$dinc226$0 = 0, $dinc2$0 = 0;
 var $dinc2$0$in = 0, $dinc213$0 = 0, $dinc213$0$in = 0, $dinc226$0 = 0, $dinc226$0$in = 0, $i$023 = 0, $i20$033 = 0, $i7$028 = 0, $length$0$ph = 0, $length$016 = 0, $length1$0$ph = 0, $length1$018 = 0, $length4$0 = 0, $length4$0$ = 0, $length4$121 = 0, $not$ = 0, $not$6 = 0, $not$7 = 0, $not$8 = 0, $numpixels$0 = 0;
 var $numpixels$0$in = 0, $numpixels$0$in$numpixels$0 = 0, $numpixels10$0 = 0, $numpixels10$0$in = 0, $numpixels10$0$in$numpixels10$0 = 0, $numpixels23$0 = 0, $numpixels23$0$in = 0, $numpixels23$0$in$numpixels23$0 = 0, $pitch5$0 = 0, $pitch5$0$v = 0, $pitch5$1 = 0, $pitch5$1$v = 0, $pitch5$2 = 0, $pixel$1$ph = 0, $pixel$115 = 0, $pixel3$1$ph = 0, $pixel3$117 = 0, $pixel6$1 = 0, $pixel6$220 = 0, $x$025 = 0;
 var $x$1 = 0, $x14$030 = 0, $x14$1 = 0, $x27$035 = 0, $x27$1 = 0, $xinc1$0 = 0, $xinc115$0 = 0, $xinc128$0 = 0, $y$026 = 0, $y$1 = 0, $y17$031 = 0, $y17$1 = 0, $y30$036 = 0, $y30$1 = 0, $yinc1$0 = 0, $yinc1$1 = 0, $yinc1$1$yinc2$1 = 0, $yinc118$0 = 0, $yinc118$1 = 0, $yinc118$1$yinc219$1 = 0;
 var $yinc131$0 = 0, $yinc131$1 = 0, $yinc131$1$yinc232$1 = 0, $yinc2$1 = 0, $yinc219$1 = 0, $yinc232$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $_r = sp + 3|0;
 $_g = sp + 2|0;
 $_b = sp + 1|0;
 $_a = sp;
 $0 = ($y2|0)==($y1|0);
 if ($0) {
  $1 = ((($dst)) + 16|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ((($dst)) + 4|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ((($4)) + 9|0);
  $6 = HEAP8[$5>>0]|0;
  $7 = $6&255;
  $8 = (($2|0) / ($7|0))&-1;
  $9 = ($x2|0)<($x1|0);
  $10 = ((($dst)) + 20|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = Math_imul($8, $y2)|0;
  $13 = (($11) + ($12<<2)|0);
  if ($9) {
   $19 = (($13) + ($x2<<2)|0);
   $20 = ($draw_end|0)!=(0);
   $21 = ((($19)) + 4|0);
   $$ = $20 ? $19 : $21;
   $22 = (($x1) - ($x2))|0;
   $23 = $20&1;
   $24 = (($22) + ($23))|0;
   $length$0$ph = $24;$pixel$1$ph = $$;
  } else {
   $14 = (($13) + ($x1<<2)|0);
   $15 = ($draw_end|0)!=(0);
   $16 = (($x2) - ($x1))|0;
   $17 = $15&1;
   $18 = (($16) + ($17))|0;
   $length$0$ph = $18;$pixel$1$ph = $14;
  }
  $25 = ($length$0$ph|0)==(0);
  if ($25) {
   STACKTOP = sp;return;
  } else {
   $length$016 = $length$0$ph;$pixel$115 = $pixel$1$ph;
  }
  while(1) {
   $26 = (($length$016) + -1)|0;
   HEAP32[$pixel$115>>2] = $color;
   $27 = ((($pixel$115)) + 4|0);
   $28 = ($26|0)==(0);
   if ($28) {
    break;
   } else {
    $length$016 = $26;$pixel$115 = $27;
   }
  }
  STACKTOP = sp;return;
 }
 $29 = ($x2|0)==($x1|0);
 if ($29) {
  $30 = ((($dst)) + 16|0);
  $31 = HEAP32[$30>>2]|0;
  $32 = ((($dst)) + 4|0);
  $33 = HEAP32[$32>>2]|0;
  $34 = ((($33)) + 9|0);
  $35 = HEAP8[$34>>0]|0;
  $36 = $35&255;
  $37 = (($31|0) / ($36|0))&-1;
  $38 = ($y2|0)<($y1|0);
  $39 = ((($dst)) + 20|0);
  $40 = HEAP32[$39>>2]|0;
  if ($38) {
   $48 = Math_imul($37, $y2)|0;
   $49 = (($40) + ($48<<2)|0);
   $50 = (($49) + ($x2<<2)|0);
   $51 = ($draw_end|0)!=(0);
   $52 = (($50) + ($37<<2)|0);
   $$1 = $51 ? $50 : $52;
   $53 = (($y1) - ($y2))|0;
   $54 = $51&1;
   $55 = (($53) + ($54))|0;
   $length1$0$ph = $55;$pixel3$1$ph = $$1;
  } else {
   $41 = Math_imul($37, $y1)|0;
   $42 = (($40) + ($41<<2)|0);
   $43 = (($42) + ($x2<<2)|0);
   $44 = ($draw_end|0)!=(0);
   $45 = (($y2) - ($y1))|0;
   $46 = $44&1;
   $47 = (($45) + ($46))|0;
   $length1$0$ph = $47;$pixel3$1$ph = $43;
  }
  $56 = ($length1$0$ph|0)==(0);
  if ($56) {
   STACKTOP = sp;return;
  } else {
   $length1$018 = $length1$0$ph;$pixel3$117 = $pixel3$1$ph;
  }
  while(1) {
   $57 = (($length1$018) + -1)|0;
   HEAP32[$pixel3$117>>2] = $color;
   $58 = (($pixel3$117) + ($37<<2)|0);
   $59 = ($57|0)==(0);
   if ($59) {
    break;
   } else {
    $length1$018 = $57;$pixel3$117 = $58;
   }
  }
  STACKTOP = sp;return;
 }
 $60 = (($x1) - ($x2))|0;
 $61 = ($60|0)<(0);
 $62 = (0 - ($60))|0;
 $63 = $61 ? $62 : $60;
 $64 = (($y1) - ($y2))|0;
 $65 = ($64|0)<(0);
 $66 = (0 - ($64))|0;
 $67 = $65 ? $66 : $64;
 $68 = ($63|0)==($67|0);
 if ($68) {
  $69 = ((($dst)) + 16|0);
  $70 = HEAP32[$69>>2]|0;
  $71 = ((($dst)) + 4|0);
  $72 = HEAP32[$71>>2]|0;
  $73 = ((($72)) + 9|0);
  $74 = HEAP8[$73>>0]|0;
  $75 = $74&255;
  $76 = (($70|0) / ($75|0))&-1;
  $77 = ($y2|0)<($y1|0);
  $78 = ((($dst)) + 20|0);
  $79 = HEAP32[$78>>2]|0;
  if ($77) {
   $85 = Math_imul($76, $y2)|0;
   $86 = (($79) + ($85<<2)|0);
   $87 = (($86) + ($x2<<2)|0);
   $88 = ($x2|0)>($x1|0);
   $pitch5$1$v = $88 ? -1 : 1;
   $pitch5$1 = (($76) + ($pitch5$1$v))|0;
   $89 = ($draw_end|0)==(0);
   $90 = (($87) + ($pitch5$1<<2)|0);
   $$2 = $89 ? $90 : $87;
   $length4$0 = $64;$pitch5$2 = $pitch5$1;$pixel6$1 = $$2;
  } else {
   $80 = Math_imul($76, $y1)|0;
   $81 = (($79) + ($80<<2)|0);
   $82 = (($81) + ($x1<<2)|0);
   $83 = ($x2|0)<($x1|0);
   $pitch5$0$v = $83 ? -1 : 1;
   $pitch5$0 = (($76) + ($pitch5$0$v))|0;
   $84 = (($y2) - ($y1))|0;
   $length4$0 = $84;$pitch5$2 = $pitch5$0;$pixel6$1 = $82;
  }
  $not$8 = ($draw_end|0)!=(0);
  $91 = $not$8&1;
  $length4$0$ = (($length4$0) + ($91))|0;
  $92 = ($length4$0$|0)==(0);
  if ($92) {
   STACKTOP = sp;return;
  } else {
   $length4$121 = $length4$0$;$pixel6$220 = $pixel6$1;
  }
  while(1) {
   $93 = (($length4$121) + -1)|0;
   HEAP32[$pixel6$220>>2] = $color;
   $94 = (($pixel6$220) + ($pitch5$2<<2)|0);
   $95 = ($93|0)==(0);
   if ($95) {
    break;
   } else {
    $length4$121 = $93;$pixel6$220 = $94;
   }
  }
  STACKTOP = sp;return;
 }
 $96 = ((($dst)) + 4|0);
 $97 = HEAP32[$96>>2]|0;
 _SDL_GetRGBA($color,$97,$_r,$_g,$_b,$_a);
 $98 = ((($97)) + 12|0);
 $99 = HEAP32[$98>>2]|0;
 $100 = ($99|0)==(16711680);
 do {
  if ($100) {
   $101 = ((($97)) + 24|0);
   $102 = HEAP32[$101>>2]|0;
   $103 = ($102|0)==(0);
   $104 = (($x2) - ($x1))|0;
   $105 = ($104|0)<(0);
   $106 = (0 - ($104))|0;
   $107 = $105 ? $106 : $104;
   $108 = (($y2) - ($y1))|0;
   $109 = ($108|0)<(0);
   $110 = (0 - ($108))|0;
   $111 = $109 ? $110 : $108;
   $112 = ($107|0)<($111|0);
   if ($103) {
    if ($112) {
     $116 = $107 << 1;
     $117 = (($116) - ($111))|0;
     $118 = (($107) - ($111))|0;
     $d$0 = $117;$dinc1$0 = $116;$dinc2$0$in = $118;$numpixels$0$in = $111;$xinc1$0 = 0;$yinc1$0 = 1;
    } else {
     $113 = $111 << 1;
     $114 = (($113) - ($107))|0;
     $115 = (($111) - ($107))|0;
     $d$0 = $114;$dinc1$0 = $113;$dinc2$0$in = $115;$numpixels$0$in = $107;$xinc1$0 = 1;$yinc1$0 = 0;
    }
    $dinc2$0 = $dinc2$0$in << 1;
    $119 = ($x2|0)<($x1|0);
    $120 = (0 - ($xinc1$0))|0;
    $$xinc1$0 = $119 ? $120 : $xinc1$0;
    $$3 = $119 ? -1 : 1;
    $121 = ($y2|0)<($y1|0);
    $122 = (0 - ($yinc1$0))|0;
    $yinc1$1 = $121 ? $122 : $yinc1$0;
    $yinc2$1 = $121 ? -1 : 1;
    $not$7 = ($draw_end|0)!=(0);
    $numpixels$0 = $not$7&1;
    $numpixels$0$in$numpixels$0 = (($numpixels$0$in) + ($numpixels$0))|0;
    $123 = ($numpixels$0$in$numpixels$0|0)>(0);
    if (!($123)) {
     break;
    }
    $124 = ((($dst)) + 20|0);
    $125 = HEAP32[$124>>2]|0;
    $126 = ((($dst)) + 16|0);
    $d$124 = $d$0;$i$023 = 0;$x$025 = $x1;$y$026 = $y1;
    while(1) {
     $127 = HEAP32[$126>>2]|0;
     $128 = Math_imul($127, $y$026)|0;
     $129 = (($125) + ($128)|0);
     $130 = $x$025 << 2;
     $131 = (($129) + ($130)|0);
     HEAP32[$131>>2] = $color;
     $132 = ($d$124|0)<(0);
     $dinc1$0$dinc2$0 = $132 ? $dinc1$0 : $dinc2$0;
     $$xinc1$0$$3 = $132 ? $$xinc1$0 : $$3;
     $yinc1$1$yinc2$1 = $132 ? $yinc1$1 : $yinc2$1;
     $y$1 = (($y$026) + ($yinc1$1$yinc2$1))|0;
     $x$1 = (($$xinc1$0$$3) + ($x$025))|0;
     $d$2 = (($dinc1$0$dinc2$0) + ($d$124))|0;
     $133 = (($i$023) + 1)|0;
     $134 = ($133|0)<($numpixels$0$in$numpixels$0|0);
     if ($134) {
      $d$124 = $d$2;$i$023 = $133;$x$025 = $x$1;$y$026 = $y$1;
     } else {
      break;
     }
    }
   } else {
    if ($112) {
     $138 = $107 << 1;
     $139 = (($138) - ($111))|0;
     $140 = (($107) - ($111))|0;
     $d11$0 = $139;$dinc112$0 = $138;$dinc213$0$in = $140;$numpixels10$0$in = $111;$xinc115$0 = 0;$yinc118$0 = 1;
    } else {
     $135 = $111 << 1;
     $136 = (($135) - ($107))|0;
     $137 = (($111) - ($107))|0;
     $d11$0 = $136;$dinc112$0 = $135;$dinc213$0$in = $137;$numpixels10$0$in = $107;$xinc115$0 = 1;$yinc118$0 = 0;
    }
    $dinc213$0 = $dinc213$0$in << 1;
    $141 = ($x2|0)<($x1|0);
    $142 = (0 - ($xinc115$0))|0;
    $$xinc115$0 = $141 ? $142 : $xinc115$0;
    $$4 = $141 ? -1 : 1;
    $143 = ($y2|0)<($y1|0);
    $144 = (0 - ($yinc118$0))|0;
    $yinc118$1 = $143 ? $144 : $yinc118$0;
    $yinc219$1 = $143 ? -1 : 1;
    $not$6 = ($draw_end|0)!=(0);
    $numpixels10$0 = $not$6&1;
    $numpixels10$0$in$numpixels10$0 = (($numpixels10$0$in) + ($numpixels10$0))|0;
    $145 = ($numpixels10$0$in$numpixels10$0|0)>(0);
    if (!($145)) {
     break;
    }
    $146 = ((($dst)) + 20|0);
    $147 = HEAP32[$146>>2]|0;
    $148 = ((($dst)) + 16|0);
    $d11$129 = $d11$0;$i7$028 = 0;$x14$030 = $x1;$y17$031 = $y1;
    while(1) {
     $149 = HEAP32[$148>>2]|0;
     $150 = Math_imul($149, $y17$031)|0;
     $151 = (($147) + ($150)|0);
     $152 = $x14$030 << 2;
     $153 = (($151) + ($152)|0);
     HEAP32[$153>>2] = $color;
     $154 = ($d11$129|0)<(0);
     $dinc112$0$dinc213$0 = $154 ? $dinc112$0 : $dinc213$0;
     $$xinc115$0$$4 = $154 ? $$xinc115$0 : $$4;
     $yinc118$1$yinc219$1 = $154 ? $yinc118$1 : $yinc219$1;
     $y17$1 = (($y17$031) + ($yinc118$1$yinc219$1))|0;
     $x14$1 = (($$xinc115$0$$4) + ($x14$030))|0;
     $d11$2 = (($dinc112$0$dinc213$0) + ($d11$129))|0;
     $155 = (($i7$028) + 1)|0;
     $156 = ($155|0)<($numpixels10$0$in$numpixels10$0|0);
     if ($156) {
      $d11$129 = $d11$2;$i7$028 = $155;$x14$030 = $x14$1;$y17$031 = $y17$1;
     } else {
      break;
     }
    }
   }
  } else {
   $157 = (($x2) - ($x1))|0;
   $158 = ($157|0)<(0);
   $159 = (0 - ($157))|0;
   $160 = $158 ? $159 : $157;
   $161 = (($y2) - ($y1))|0;
   $162 = ($161|0)<(0);
   $163 = (0 - ($161))|0;
   $164 = $162 ? $163 : $161;
   $165 = ($160|0)<($164|0);
   if ($165) {
    $169 = $160 << 1;
    $170 = (($169) - ($164))|0;
    $171 = (($160) - ($164))|0;
    $d24$0 = $170;$dinc125$0 = $169;$dinc226$0$in = $171;$numpixels23$0$in = $164;$xinc128$0 = 0;$yinc131$0 = 1;
   } else {
    $166 = $164 << 1;
    $167 = (($166) - ($160))|0;
    $168 = (($164) - ($160))|0;
    $d24$0 = $167;$dinc125$0 = $166;$dinc226$0$in = $168;$numpixels23$0$in = $160;$xinc128$0 = 1;$yinc131$0 = 0;
   }
   $dinc226$0 = $dinc226$0$in << 1;
   $172 = ($x2|0)<($x1|0);
   $173 = (0 - ($xinc128$0))|0;
   $$xinc128$0 = $172 ? $173 : $xinc128$0;
   $$5 = $172 ? -1 : 1;
   $174 = ($y2|0)<($y1|0);
   $175 = (0 - ($yinc131$0))|0;
   $yinc131$1 = $174 ? $175 : $yinc131$0;
   $yinc232$1 = $174 ? -1 : 1;
   $not$ = ($draw_end|0)!=(0);
   $numpixels23$0 = $not$&1;
   $numpixels23$0$in$numpixels23$0 = (($numpixels23$0$in) + ($numpixels23$0))|0;
   $176 = ($numpixels23$0$in$numpixels23$0|0)>(0);
   if ($176) {
    $177 = ((($dst)) + 20|0);
    $178 = HEAP32[$177>>2]|0;
    $179 = ((($dst)) + 16|0);
    $d24$134 = $d24$0;$i20$033 = 0;$x27$035 = $x1;$y30$036 = $y1;
    while(1) {
     $180 = HEAP32[$179>>2]|0;
     $181 = Math_imul($180, $y30$036)|0;
     $182 = (($178) + ($181)|0);
     $183 = $x27$035 << 2;
     $184 = (($182) + ($183)|0);
     HEAP32[$184>>2] = $color;
     $185 = ($d24$134|0)<(0);
     $dinc125$0$dinc226$0 = $185 ? $dinc125$0 : $dinc226$0;
     $$xinc128$0$$5 = $185 ? $$xinc128$0 : $$5;
     $yinc131$1$yinc232$1 = $185 ? $yinc131$1 : $yinc232$1;
     $y30$1 = (($y30$036) + ($yinc131$1$yinc232$1))|0;
     $x27$1 = (($$xinc128$0$$5) + ($x27$035))|0;
     $d24$2 = (($dinc125$0$dinc226$0) + ($d24$134))|0;
     $186 = (($i20$033) + 1)|0;
     $187 = ($186|0)<($numpixels23$0$in$numpixels23$0|0);
     if ($187) {
      $d24$134 = $d24$2;$i20$033 = $186;$x27$035 = $x27$1;$y30$036 = $y30$1;
     } else {
      break;
     }
    }
   }
  }
 } while(0);
 STACKTOP = sp;return;
}
function _SDL_IntersectRectAndLine($rect,$X1,$Y1,$X2,$Y2) {
 $rect = $rect|0;
 $X1 = $X1|0;
 $Y1 = $Y1|0;
 $X2 = $X2|0;
 $Y2 = $Y2|0;
 var $$$$i23 = 0, $$$i$us = 0, $$$i13 = 0, $$$i18 = 0, $$$i23 = 0, $$0 = 0, $$in = 0, $$in$lcssa = 0, $$lcssa = 0, $$lcssa172 = 0, $$lcssa175 = 0, $$lcssa178 = 0, $$not10 = 0, $$not28 = 0, $$not29 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $code$0$$i$us = 0, $code$0$$i15 = 0, $code$0$$i20 = 0, $code$0$$i25 = 0, $code$0$i$us = 0, $code$0$i14 = 0, $code$0$i19 = 0, $not$$i$us = 0, $not$$i12 = 0;
 var $not$$i17 = 0, $notlhs = 0, $notrhs = 0, $or$cond = 0, $or$cond$not = 0, $or$cond1 = 0, $or$cond11 = 0, $or$cond2$not = 0, $or$cond27 = 0, $or$cond27$not = 0, $or$cond3 = 0, $or$cond4$not = 0, $or$cond5 = 0, $or$cond5$not = 0, $or$cond6 = 0, $or$cond7 = 0, $or$cond8$not = 0, $or$cond9 = 0, $outcode1$0$ph$be = 0, $outcode2$0$be$us = 0;
 var $outcode2$0$ph = 0, $outcode2$048$us = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer10 = 0, $vararg_buffer4 = 0, $vararg_buffer7 = 0, $x$0$ph61 = 0, $x$0$ph61$lcssa = 0, $x$047$us = 0, $x$1 = 0, $x$2$us = 0, $x1$0$ph$lcssa43 = 0, $x1$0$ph59 = 0, $x1$0$ph59$lcssa = 0, $x2$0$lcssa = 0, $x2$044$us = 0, $y$0$ph60 = 0, $y$0$ph60$lcssa = 0, $y$046$us = 0;
 var $y$1 = 0, $y$2$us = 0, $y1$0$ph$lcssa42 = 0, $y1$0$ph56 = 0, $y1$0$ph56$lcssa = 0, $y2$0$lcssa = 0, $y2$045$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer10 = sp + 32|0;
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($rect|0)==(0|0);
 if ($0) {
  HEAP32[$vararg_buffer>>2] = 603688;
  (_SDL_SetError(611267,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ($X1|0)==(0|0);
 if ($1) {
  HEAP32[$vararg_buffer1>>2] = 603693;
  (_SDL_SetError(611267,$vararg_buffer1)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($Y1|0)==(0|0);
 if ($2) {
  HEAP32[$vararg_buffer4>>2] = 603696;
  (_SDL_SetError(611267,$vararg_buffer4)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ($X2|0)==(0|0);
 if ($3) {
  HEAP32[$vararg_buffer7>>2] = 603699;
  (_SDL_SetError(611267,$vararg_buffer7)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ($Y2|0)==(0|0);
 if ($4) {
  HEAP32[$vararg_buffer10>>2] = 603702;
  (_SDL_SetError(611267,$vararg_buffer10)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = ((($rect)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)<(1);
 if ($7) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $8 = ((($rect)) + 12|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)<(1);
 if ($10) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $11 = HEAP32[$X1>>2]|0;
 $12 = HEAP32[$Y1>>2]|0;
 $13 = HEAP32[$X2>>2]|0;
 $14 = HEAP32[$Y2>>2]|0;
 $15 = HEAP32[$rect>>2]|0;
 $16 = ((($rect)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = (($6) + ($15))|0;
 $19 = (($18) + -1)|0;
 $20 = (($9) + ($17))|0;
 $21 = (($20) + -1)|0;
 $notlhs = ($11|0)<($15|0);
 $notrhs = ($11|0)>=($18|0);
 $or$cond$not = $notlhs | $notrhs;
 $22 = ($13|0)<($15|0);
 $or$cond1 = $22 | $or$cond$not;
 $23 = ($13|0)<($18|0);
 $$not28 = $23 ^ 1;
 $or$cond2$not = $or$cond1 | $$not28;
 $24 = ($12|0)<($17|0);
 $or$cond3 = $24 | $or$cond2$not;
 $$not29 = ($12|0)>=($20|0);
 $or$cond4$not = $$not29 | $or$cond3;
 $25 = ($14|0)<($17|0);
 $or$cond5 = $25 | $or$cond4$not;
 $or$cond5$not = $or$cond5 ^ 1;
 $26 = ($14|0)<($20|0);
 $or$cond6 = $26 & $or$cond5$not;
 if ($or$cond6) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 $or$cond7 = $notlhs & $22;
 if ($or$cond7) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $or$cond8$not = $notrhs & $$not28;
 $or$cond9 = $24 & $25;
 $or$cond27 = $or$cond9 | $or$cond8$not;
 $or$cond27$not = $or$cond27 ^ 1;
 $$not10 = $$not29 ^ 1;
 $or$cond11 = $26 | $$not10;
 $or$cond = $or$cond11 & $or$cond27$not;
 if (!($or$cond)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $27 = ($12|0)==($14|0);
 if ($27) {
  if ($notlhs) {
   HEAP32[$X1>>2] = $15;
  } else {
   if ($notrhs) {
    HEAP32[$X1>>2] = $19;
   }
  }
  if ($22) {
   HEAP32[$X2>>2] = $15;
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
  }
  if ($23) {
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
  }
  HEAP32[$X2>>2] = $19;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 $28 = ($11|0)==($13|0);
 if ($28) {
  if ($24) {
   HEAP32[$Y1>>2] = $17;
  } else {
   if ($$not29) {
    HEAP32[$Y1>>2] = $21;
   }
  }
  if ($25) {
   HEAP32[$Y2>>2] = $17;
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
  }
  if ($26) {
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
  }
  HEAP32[$Y2>>2] = $21;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 $$$i23 = $$not29&1;
 $$$$i23 = $24 ? 2 : $$$i23;
 if ($notlhs) {
  $29 = $$$$i23 | 4;
  $35 = $29;
 } else {
  $30 = ($18|0)>($11|0);
  $31 = $$$$i23 | 8;
  $code$0$$i25 = $30 ? $$$$i23 : $31;
  $35 = $code$0$$i25;
 }
 $not$$i17 = ($20|0)<=($14|0);
 $$$i18 = $not$$i17&1;
 $code$0$i19 = $25 ? 2 : $$$i18;
 $32 = $code$0$i19 | 4;
 $33 = $code$0$i19 | 8;
 $code$0$$i20 = $23 ? $code$0$i19 : $33;
 $outcode2$0$ph = $22 ? $32 : $code$0$$i20;
 $34 = $outcode2$0$ph | $35;
 $36 = ($34|0)==(0);
 L72: do {
  if ($36) {
   $x1$0$ph$lcssa43 = $11;$x2$0$lcssa = $13;$y1$0$ph$lcssa42 = $12;$y2$0$lcssa = $14;
  } else {
   $$in = $35;$x$0$ph61 = 0;$x1$0$ph59 = $11;$y$0$ph60 = 0;$y1$0$ph56 = $12;
   while(1) {
    $37 = ($$in|0)==(0);
    $38 = (($19) - ($x1$0$ph59))|0;
    $39 = (($15) - ($x1$0$ph59))|0;
    $40 = (($21) - ($y1$0$ph56))|0;
    $41 = (($17) - ($y1$0$ph56))|0;
    if ($37) {
     $$in$lcssa = $$in;$$lcssa = $38;$$lcssa172 = $39;$$lcssa175 = $40;$$lcssa178 = $41;$x$0$ph61$lcssa = $x$0$ph61;$x1$0$ph59$lcssa = $x1$0$ph59;$y$0$ph60$lcssa = $y$0$ph60;$y1$0$ph56$lcssa = $y1$0$ph56;
     break;
    }
    $79 = $outcode2$0$ph & $$in;
    $80 = ($79|0)==(0);
    if (!($80)) {
     $$0 = 0;
     label = 67;
     break;
    }
    $81 = $$in & 2;
    $82 = ($81|0)==(0);
    do {
     if ($82) {
      $88 = $$in & 1;
      $89 = ($88|0)==(0);
      if (!($89)) {
       $90 = (($13) - ($x1$0$ph59))|0;
       $91 = Math_imul($90, $40)|0;
       $92 = (($14) - ($y1$0$ph56))|0;
       $93 = (($91|0) / ($92|0))&-1;
       $94 = (($93) + ($x1$0$ph59))|0;
       $x$1 = $94;$y$1 = $21;
       break;
      }
      $95 = $$in & 4;
      $96 = ($95|0)==(0);
      if (!($96)) {
       $97 = (($14) - ($y1$0$ph56))|0;
       $98 = Math_imul($39, $97)|0;
       $99 = (($13) - ($x1$0$ph59))|0;
       $100 = (($98|0) / ($99|0))&-1;
       $101 = (($100) + ($y1$0$ph56))|0;
       $x$1 = $15;$y$1 = $101;
       break;
      }
      $102 = $$in & 8;
      $103 = ($102|0)==(0);
      if ($103) {
       $x$1 = $x$0$ph61;$y$1 = $y$0$ph60;
       break;
      }
      $104 = (($14) - ($y1$0$ph56))|0;
      $105 = Math_imul($38, $104)|0;
      $106 = (($13) - ($x1$0$ph59))|0;
      $107 = (($105|0) / ($106|0))&-1;
      $108 = (($107) + ($y1$0$ph56))|0;
      $x$1 = $19;$y$1 = $108;
     } else {
      $83 = (($13) - ($x1$0$ph59))|0;
      $84 = Math_imul($83, $41)|0;
      $85 = (($14) - ($y1$0$ph56))|0;
      $86 = (($84|0) / ($85|0))&-1;
      $87 = (($86) + ($x1$0$ph59))|0;
      $x$1 = $87;$y$1 = $17;
     }
    } while(0);
    $109 = ($17|0)>($y$1|0);
    $not$$i12 = ($20|0)<=($y$1|0);
    $$$i13 = $not$$i12&1;
    $code$0$i14 = $109 ? 2 : $$$i13;
    $110 = ($15|0)>($x$1|0);
    if ($110) {
     $111 = $code$0$i14 | 4;
     $outcode1$0$ph$be = $111;
    } else {
     $112 = ($18|0)>($x$1|0);
     $113 = $code$0$i14 | 8;
     $code$0$$i15 = $112 ? $code$0$i14 : $113;
     $outcode1$0$ph$be = $code$0$$i15;
    }
    $114 = $outcode2$0$ph | $outcode1$0$ph$be;
    $115 = ($114|0)==(0);
    if ($115) {
     $x1$0$ph$lcssa43 = $x$1;$x2$0$lcssa = $13;$y1$0$ph$lcssa42 = $y$1;$y2$0$lcssa = $14;
     break L72;
    } else {
     $$in = $outcode1$0$ph$be;$x$0$ph61 = $x$1;$x1$0$ph59 = $x$1;$y$0$ph60 = $y$1;$y1$0$ph56 = $y$1;
    }
   }
   if ((label|0) == 67) {
    STACKTOP = sp;return ($$0|0);
   }
   $outcode2$048$us = $outcode2$0$ph;$x$047$us = $x$0$ph61$lcssa;$x2$044$us = $13;$y$046$us = $y$0$ph60$lcssa;$y2$045$us = $14;
   while(1) {
    $42 = $outcode2$048$us & $$in$lcssa;
    $43 = ($42|0)==(0);
    if (!($43)) {
     $$0 = 0;
     break;
    }
    $44 = $outcode2$048$us & 2;
    $45 = ($44|0)==(0);
    do {
     if ($45) {
      $51 = $outcode2$048$us & 1;
      $52 = ($51|0)==(0);
      if (!($52)) {
       $53 = (($x2$044$us) - ($x1$0$ph59$lcssa))|0;
       $54 = Math_imul($53, $$lcssa175)|0;
       $55 = (($y2$045$us) - ($y1$0$ph56$lcssa))|0;
       $56 = (($54|0) / ($55|0))&-1;
       $57 = (($56) + ($x1$0$ph59$lcssa))|0;
       $x$2$us = $57;$y$2$us = $21;
       break;
      }
      $58 = $outcode2$048$us & 4;
      $59 = ($58|0)==(0);
      if (!($59)) {
       $60 = (($y2$045$us) - ($y1$0$ph56$lcssa))|0;
       $61 = Math_imul($$lcssa172, $60)|0;
       $62 = (($x2$044$us) - ($x1$0$ph59$lcssa))|0;
       $63 = (($61|0) / ($62|0))&-1;
       $64 = (($63) + ($y1$0$ph56$lcssa))|0;
       $x$2$us = $15;$y$2$us = $64;
       break;
      }
      $65 = $outcode2$048$us & 8;
      $66 = ($65|0)==(0);
      if ($66) {
       $x$2$us = $x$047$us;$y$2$us = $y$046$us;
       break;
      }
      $67 = (($y2$045$us) - ($y1$0$ph56$lcssa))|0;
      $68 = Math_imul($$lcssa, $67)|0;
      $69 = (($x2$044$us) - ($x1$0$ph59$lcssa))|0;
      $70 = (($68|0) / ($69|0))&-1;
      $71 = (($70) + ($y1$0$ph56$lcssa))|0;
      $x$2$us = $19;$y$2$us = $71;
     } else {
      $46 = (($x2$044$us) - ($x1$0$ph59$lcssa))|0;
      $47 = Math_imul($46, $$lcssa178)|0;
      $48 = (($y2$045$us) - ($y1$0$ph56$lcssa))|0;
      $49 = (($47|0) / ($48|0))&-1;
      $50 = (($49) + ($x1$0$ph59$lcssa))|0;
      $x$2$us = $50;$y$2$us = $17;
     }
    } while(0);
    $72 = ($17|0)>($y$2$us|0);
    $not$$i$us = ($20|0)<=($y$2$us|0);
    $$$i$us = $not$$i$us&1;
    $code$0$i$us = $72 ? 2 : $$$i$us;
    $73 = ($15|0)>($x$2$us|0);
    if ($73) {
     $76 = $code$0$i$us | 4;
     $outcode2$0$be$us = $76;
    } else {
     $74 = ($18|0)>($x$2$us|0);
     $75 = $code$0$i$us | 8;
     $code$0$$i$us = $74 ? $code$0$i$us : $75;
     $outcode2$0$be$us = $code$0$$i$us;
    }
    $77 = $outcode2$0$be$us | $$in$lcssa;
    $78 = ($77|0)==(0);
    if ($78) {
     $x1$0$ph$lcssa43 = $x1$0$ph59$lcssa;$x2$0$lcssa = $x$2$us;$y1$0$ph$lcssa42 = $y1$0$ph56$lcssa;$y2$0$lcssa = $y$2$us;
     break L72;
    } else {
     $outcode2$048$us = $outcode2$0$be$us;$x$047$us = $x$2$us;$x2$044$us = $x$2$us;$y$046$us = $y$2$us;$y2$045$us = $y$2$us;
    }
   }
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 HEAP32[$X1>>2] = $x1$0$ph$lcssa43;
 HEAP32[$Y1>>2] = $y1$0$ph$lcssa42;
 HEAP32[$X2>>2] = $x2$0$lcssa;
 HEAP32[$Y2>>2] = $y2$0$lcssa;
 $$0 = 1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_DrawPoint($dst,$x,$y,$color) {
 $dst = $dst|0;
 $x = $x|0;
 $y = $y|0;
 $color = $color|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($dst|0)==(0|0);
 L1: do {
  if ($0) {
   $1 = (_SDL_SetError(603923,$vararg_buffer)|0);
   $$0 = $1;
  } else {
   $2 = ((($dst)) + 4|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = ((($3)) + 8|0);
   $5 = HEAP16[$4>>1]|0;
   $6 = $5&255;
   $7 = ($6&255)<(8);
   $8 = ($5&65535) >>> 8;
   $9 = $8&65535;
   if ($7) {
    $10 = (_SDL_SetError(603705,$vararg_buffer1)|0);
    $$0 = $10;
    break;
   }
   $11 = ((($dst)) + 36|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($12|0)>($x|0);
   if ($13) {
    $$0 = 0;
   } else {
    $14 = ((($dst)) + 40|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = ($15|0)>($y|0);
    if ($16) {
     $$0 = 0;
    } else {
     $17 = ((($dst)) + 44|0);
     $18 = HEAP32[$17>>2]|0;
     $19 = (($18) + ($12))|0;
     $20 = ($19|0)>($x|0);
     if ($20) {
      $21 = ((($dst)) + 48|0);
      $22 = HEAP32[$21>>2]|0;
      $23 = (($22) + ($15))|0;
      $24 = ($23|0)>($y|0);
      if ($24) {
       switch ($9|0) {
       case 1:  {
        $25 = $color&255;
        $26 = ((($dst)) + 20|0);
        $27 = HEAP32[$26>>2]|0;
        $28 = ((($dst)) + 16|0);
        $29 = HEAP32[$28>>2]|0;
        $30 = Math_imul($29, $y)|0;
        $31 = (($27) + ($30)|0);
        $32 = (($31) + ($x)|0);
        HEAP8[$32>>0] = $25;
        $$0 = 0;
        break L1;
        break;
       }
       case 2:  {
        $33 = $color&65535;
        $34 = ((($dst)) + 20|0);
        $35 = HEAP32[$34>>2]|0;
        $36 = ((($dst)) + 16|0);
        $37 = HEAP32[$36>>2]|0;
        $38 = Math_imul($37, $y)|0;
        $39 = (($35) + ($38)|0);
        $40 = $x << 1;
        $41 = (($39) + ($40)|0);
        HEAP16[$41>>1] = $33;
        $$0 = 0;
        break L1;
        break;
       }
       case 3:  {
        $42 = (_SDL_Error(4)|0);
        $$0 = $42;
        break L1;
        break;
       }
       case 4:  {
        $43 = ((($dst)) + 20|0);
        $44 = HEAP32[$43>>2]|0;
        $45 = ((($dst)) + 16|0);
        $46 = HEAP32[$45>>2]|0;
        $47 = Math_imul($46, $y)|0;
        $48 = (($44) + ($47)|0);
        $49 = $x << 2;
        $50 = (($48) + ($49)|0);
        HEAP32[$50>>2] = $color;
        $$0 = 0;
        break L1;
        break;
       }
       default: {
        $$0 = 0;
        break L1;
       }
       }
      } else {
       $$0 = 0;
      }
     } else {
      $$0 = 0;
     }
    }
   }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _SDL_BlendLines($dst,$points,$count,$blendMode,$r,$g,$b,$a) {
 $dst = $dst|0;
 $points = $points|0;
 $count = $count|0;
 $blendMode = $blendMode|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $$0 = 0, $$0$i1$ph = 0, $$phi$trans$insert = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $SDL_BlendLine_ARGB8888$SDL_BlendLine_RGB888$i = 0, $SDL_BlendLine_RGB555$switch$select$i = 0, $SDL_BlendLine_RGBA4$SDL_BlendLine_RGB4$i = 0, $exitcond = 0, $i$02 = 0;
 var $switch$select$i = 0, $switch$selectcmp$i = 0, $switch$selectcmp1$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $x1 = 0, $x2 = 0, $y1 = 0, $y2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $x1 = sp + 24|0;
 $y1 = sp + 20|0;
 $x2 = sp + 16|0;
 $y2 = sp + 12|0;
 $0 = ($dst|0)==(0|0);
 if ($0) {
  $1 = (_SDL_SetError(603749,$vararg_buffer)|0);
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ((($dst)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 9|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 L5: do {
  switch ($6|0) {
  case 2:  {
   $7 = ((($3)) + 12|0);
   $8 = HEAP32[$7>>2]|0;
   $switch$selectcmp$i = ($8|0)==(63488);
   $switch$select$i = $switch$selectcmp$i ? 4551 : 4552;
   $switch$selectcmp1$i = ($8|0)==(31744);
   $SDL_BlendLine_RGB555$switch$select$i = $switch$selectcmp1$i ? 4553 : $switch$select$i;
   $$0$i1$ph = $SDL_BlendLine_RGB555$switch$select$i;
   break;
  }
  case 4:  {
   $11 = ((($3)) + 12|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($12|0)==(16711680);
   $14 = ((($3)) + 24|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)!=(0);
   if ($13) {
    $SDL_BlendLine_ARGB8888$SDL_BlendLine_RGB888$i = $16 ? 4554 : 4555;
    $$0$i1$ph = $SDL_BlendLine_ARGB8888$SDL_BlendLine_RGB888$i;
    break L5;
   } else {
    $SDL_BlendLine_RGBA4$SDL_BlendLine_RGB4$i = $16 ? 4556 : 4557;
    $$0$i1$ph = $SDL_BlendLine_RGBA4$SDL_BlendLine_RGB4$i;
    break L5;
   }
   break;
  }
  default: {
   $17 = (_SDL_SetError(603799,$vararg_buffer1)|0);
   $$0 = $17;
   STACKTOP = sp;return ($$0|0);
  }
  }
 } while(0);
 $9 = ($count|0)>(1);
 if ($9) {
  $10 = ((($dst)) + 36|0);
  $i$02 = 1;
  while(1) {
   $18 = (($i$02) + -1)|0;
   $19 = (($points) + ($18<<3)|0);
   $20 = HEAP32[$19>>2]|0;
   HEAP32[$x1>>2] = $20;
   $21 = (((($points) + ($18<<3)|0)) + 4|0);
   $22 = HEAP32[$21>>2]|0;
   HEAP32[$y1>>2] = $22;
   $23 = (($points) + ($i$02<<3)|0);
   $24 = HEAP32[$23>>2]|0;
   HEAP32[$x2>>2] = $24;
   $25 = (((($points) + ($i$02<<3)|0)) + 4|0);
   $26 = HEAP32[$25>>2]|0;
   HEAP32[$y2>>2] = $26;
   $27 = (_SDL_IntersectRectAndLine($10,$x1,$y1,$x2,$y2)|0);
   $28 = ($27|0)==(0);
   if (!($28)) {
    $29 = HEAP32[$x2>>2]|0;
    $30 = HEAP32[$23>>2]|0;
    $31 = ($29|0)==($30|0);
    $32 = HEAP32[$y2>>2]|0;
    if ($31) {
     $33 = HEAP32[$25>>2]|0;
     $34 = ($32|0)!=($33|0);
     $36 = $34;
    } else {
     $36 = 1;
    }
    $35 = $36&1;
    $37 = HEAP32[$x1>>2]|0;
    $38 = HEAP32[$y1>>2]|0;
    FUNCTION_TABLE_viiiiiiiiiii[$$0$i1$ph & 8191]($dst,$37,$38,$29,$32,$blendMode,$r,$g,$b,$a,$35);
   }
   $39 = (($i$02) + 1)|0;
   $exitcond = ($39|0)==($count|0);
   if ($exitcond) {
    break;
   } else {
    $i$02 = $39;
   }
  }
 }
 $40 = HEAP32[$points>>2]|0;
 $41 = (($count) + -1)|0;
 $42 = (($points) + ($41<<3)|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = ($40|0)==($43|0);
 if ($44) {
  $45 = ((($points)) + 4|0);
  $46 = HEAP32[$45>>2]|0;
  $47 = (((($points) + ($41<<3)|0)) + 4|0);
  $48 = HEAP32[$47>>2]|0;
  $49 = ($46|0)==($48|0);
  if ($49) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  } else {
   $50 = $48;
  }
 } else {
  $$phi$trans$insert = (((($points) + ($41<<3)|0)) + 4|0);
  $$pre3 = HEAP32[$$phi$trans$insert>>2]|0;
  $50 = $$pre3;
 }
 (_SDL_BlendPoint($dst,$43,$50,$blendMode,$r,$g,$b,$a)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_BlendLine_RGB565($dst,$x1,$y1,$x2,$y2,$blendMode,$_r,$_g,$_b,$_a,$draw_end) {
 $dst = $dst|0;
 $x1 = $x1|0;
 $y1 = $y1|0;
 $x2 = $x2|0;
 $y2 = $y2|0;
 $blendMode = $blendMode|0;
 $_r = $_r|0;
 $_g = $_g|0;
 $_b = $_b|0;
 $_a = $_a|0;
 $draw_end = $draw_end|0;
 var $$ = 0, $$10 = 0, $$11 = 0, $$12 = 0, $$13 = 0, $$14 = 0, $$15 = 0, $$16 = 0, $$17 = 0, $$18 = 0, $$19 = 0, $$2 = 0, $$20 = 0, $$21 = 0, $$22 = 0, $$23 = 0, $$24 = 0, $$25 = 0, $$26 = 0, $$27 = 0;
 var $$28 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$6 = 0, $$7 = 0, $$8 = 0, $$9 = 0, $$pre = 0, $$pre128 = 0, $$pre129 = 0, $$pre130 = 0, $$pre131 = 0, $$pre132 = 0, $$pre133 = 0, $$pre134 = 0, $$pre135 = 0, $$pre136 = 0, $$pre137 = 0, $$pre138 = 0;
 var $$pre139 = 0, $$pre140 = 0, $$pre141 = 0, $$pre142 = 0, $$pre143 = 0, $$pre144 = 0, $$pre145 = 0, $$pre146 = 0, $$pre147 = 0, $$pre148 = 0, $$pre149 = 0, $$pre150 = 0, $$pre151 = 0, $$pre152 = 0, $$pre153 = 0, $$pre154 = 0, $$pre155 = 0, $$pre156 = 0, $$xinc1$0 = 0, $$xinc1$0$$25 = 0;
 var $$xinc1109$0 = 0, $$xinc1109$0$$27 = 0, $$xinc1127$0 = 0, $$xinc1127$0$$28 = 0, $$xinc191$0 = 0, $$xinc191$0$$26 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0;
 var $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0;
 var $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0;
 var $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0;
 var $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0;
 var $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0;
 var $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0;
 var $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0;
 var $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0;
 var $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0;
 var $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0;
 var $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0;
 var $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0;
 var $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0;
 var $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0;
 var $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0;
 var $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0;
 var $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0;
 var $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0;
 var $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0;
 var $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0;
 var $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0;
 var $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0;
 var $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0;
 var $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0;
 var $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0;
 var $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0;
 var $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0;
 var $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0;
 var $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0;
 var $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0;
 var $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0;
 var $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0;
 var $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0;
 var $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0;
 var $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0;
 var $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $a$0 = 0, $b$0 = 0, $blendMode$off = 0, $d$0 = 0, $d$199 = 0, $d$2 = 0, $d105$0 = 0, $d105$1109 = 0, $d105$2 = 0, $d123$0 = 0, $d123$194 = 0, $d123$2 = 0, $d87$0 = 0, $d87$1104 = 0, $d87$2 = 0, $dinc1$0 = 0, $dinc1$0$dinc2$0 = 0, $dinc1106$0 = 0;
 var $dinc1106$0$dinc2107$0 = 0, $dinc1124$0 = 0, $dinc1124$0$dinc2125$0 = 0, $dinc188$0 = 0, $dinc188$0$dinc289$0 = 0, $dinc2$0 = 0, $dinc2$0$in = 0, $dinc2107$0 = 0, $dinc2107$0$in = 0, $dinc2125$0 = 0, $dinc2125$0$in = 0, $dinc289$0 = 0, $dinc289$0$in = 0, $g$0 = 0, $i$098 = 0, $i101$0108 = 0, $i119$093 = 0, $i83$0103 = 0, $length$0$ph = 0, $length$061 = 0;
 var $length1$0$ph = 0, $length1$064 = 0, $length15$0$ph = 0, $length15$059 = 0, $length22$0$ph = 0, $length22$073 = 0, $length29$0$ph = 0, $length29$076 = 0, $length36$0$ph = 0, $length36$079 = 0, $length43$0$ph = 0, $length43$070 = 0, $length50$0 = 0, $length50$0$ = 0, $length50$185 = 0, $length57$0 = 0, $length57$0$ = 0, $length57$188 = 0, $length64$0 = 0, $length64$0$ = 0;
 var $length64$191 = 0, $length71$0 = 0, $length71$0$ = 0, $length71$182 = 0, $length8$0$ph = 0, $length8$067 = 0, $not$ = 0, $not$29 = 0, $not$30 = 0, $not$31 = 0, $not$32 = 0, $not$33 = 0, $not$34 = 0, $not$35 = 0, $numpixels$0 = 0, $numpixels$0$in = 0, $numpixels$0$in$numpixels$0 = 0, $numpixels104$0 = 0, $numpixels104$0$in = 0, $numpixels104$0$in$numpixels104$0 = 0;
 var $numpixels122$0 = 0, $numpixels122$0$in = 0, $numpixels122$0$in$numpixels122$0 = 0, $numpixels86$0 = 0, $numpixels86$0$in = 0, $numpixels86$0$in$numpixels86$0 = 0, $pitch51$0 = 0, $pitch51$0$v = 0, $pitch51$1 = 0, $pitch51$1$v = 0, $pitch51$2 = 0, $pitch58$0 = 0, $pitch58$0$v = 0, $pitch58$1 = 0, $pitch58$1$v = 0, $pitch58$2 = 0, $pitch65$0 = 0, $pitch65$0$v = 0, $pitch65$1 = 0, $pitch65$1$v = 0;
 var $pitch65$2 = 0, $pitch72$0 = 0, $pitch72$0$v = 0, $pitch72$1 = 0, $pitch72$1$v = 0, $pitch72$2 = 0, $pixel$1$ph = 0, $pixel$160 = 0, $pixel10$1$ph = 0, $pixel10$166 = 0, $pixel17$1$ph = 0, $pixel17$158 = 0, $pixel24$1$ph = 0, $pixel24$172 = 0, $pixel3$1$ph = 0, $pixel3$163 = 0, $pixel31$1$ph = 0, $pixel31$175 = 0, $pixel38$1$ph = 0, $pixel38$178 = 0;
 var $pixel45$1$ph = 0, $pixel45$169 = 0, $pixel52$1 = 0, $pixel52$284 = 0, $pixel59$1 = 0, $pixel59$287 = 0, $pixel66$1 = 0, $pixel66$290 = 0, $pixel73$1 = 0, $pixel73$281 = 0, $r$0 = 0, $x$0100 = 0, $x$1 = 0, $x108$0110 = 0, $x108$1 = 0, $x126$095 = 0, $x126$1 = 0, $x90$0105 = 0, $x90$1 = 0, $xinc1$0 = 0;
 var $xinc1109$0 = 0, $xinc1127$0 = 0, $xinc191$0 = 0, $y$0101 = 0, $y$1 = 0, $y111$0111 = 0, $y111$1 = 0, $y129$096 = 0, $y129$1 = 0, $y93$0106 = 0, $y93$1 = 0, $yinc1$0 = 0, $yinc1$1 = 0, $yinc1$1$yinc2$1 = 0, $yinc1112$0 = 0, $yinc1112$1 = 0, $yinc1112$1$yinc2113$1 = 0, $yinc1130$0 = 0, $yinc1130$1 = 0, $yinc1130$1$yinc2131$1 = 0;
 var $yinc194$0 = 0, $yinc194$1 = 0, $yinc194$1$yinc295$1 = 0, $yinc2$1 = 0, $yinc2113$1 = 0, $yinc2131$1 = 0, $yinc295$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $blendMode$off = (($blendMode) + -1)|0;
 $0 = ($blendMode$off>>>0)<(2);
 $1 = $_r&255;
 if ($0) {
  $2 = $_a&255;
  $3 = Math_imul($2, $1)|0;
  $4 = (($3>>>0) / 255)&-1;
  $5 = $_g&255;
  $6 = Math_imul($2, $5)|0;
  $7 = (($6>>>0) / 255)&-1;
  $8 = $_b&255;
  $9 = Math_imul($2, $8)|0;
  $10 = (($9>>>0) / 255)&-1;
  $a$0 = $2;$b$0 = $10;$g$0 = $7;$r$0 = $4;
 } else {
  $11 = $_g&255;
  $12 = $_b&255;
  $13 = $_a&255;
  $a$0 = $13;$b$0 = $12;$g$0 = $11;$r$0 = $1;
 }
 $14 = $a$0 ^ 255;
 $15 = ($y2|0)==($y1|0);
 if ($15) {
  switch ($blendMode|0) {
  case 1:  {
   $16 = ((($dst)) + 16|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($dst)) + 4|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ((($19)) + 9|0);
   $21 = HEAP8[$20>>0]|0;
   $22 = $21&255;
   $23 = (($17|0) / ($22|0))&-1;
   $24 = ($x2|0)<($x1|0);
   $25 = ((($dst)) + 20|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = Math_imul($23, $y2)|0;
   $28 = (($26) + ($27<<1)|0);
   if ($24) {
    $34 = (($28) + ($x2<<1)|0);
    $35 = ($draw_end|0)!=(0);
    $36 = ((($34)) + 2|0);
    $$13 = $35 ? $34 : $36;
    $37 = (($x1) - ($x2))|0;
    $38 = $35&1;
    $39 = (($37) + ($38))|0;
    $length$0$ph = $39;$pixel$1$ph = $$13;
   } else {
    $29 = (($28) + ($x1<<1)|0);
    $30 = ($draw_end|0)!=(0);
    $31 = (($x2) - ($x1))|0;
    $32 = $30&1;
    $33 = (($31) + ($32))|0;
    $length$0$ph = $33;$pixel$1$ph = $29;
   }
   $40 = ($length$0$ph|0)==(0);
   if ($40) {
    return;
   }
   $$pre155 = HEAP32[(125936)>>2]|0;
   $$pre156 = HEAP32[(125932)>>2]|0;
   $length$061 = $length$0$ph;$pixel$160 = $pixel$1$ph;
   while(1) {
    $41 = (($length$061) + -1)|0;
    $42 = HEAP16[$pixel$160>>1]|0;
    $43 = $42&65535;
    $44 = $43 >>> 11;
    $45 = (($$pre155) + ($44)|0);
    $46 = HEAP8[$45>>0]|0;
    $47 = $46&255;
    $48 = $43 >>> 5;
    $49 = $48 & 63;
    $50 = (($$pre156) + ($49)|0);
    $51 = HEAP8[$50>>0]|0;
    $52 = $51&255;
    $53 = $43 & 31;
    $54 = (($$pre155) + ($53)|0);
    $55 = HEAP8[$54>>0]|0;
    $56 = $55&255;
    $57 = Math_imul($47, $14)|0;
    $58 = (($57>>>0) / 255)&-1;
    $59 = (($58) + ($r$0))|0;
    $60 = Math_imul($52, $14)|0;
    $61 = (($60>>>0) / 255)&-1;
    $62 = (($61) + ($g$0))|0;
    $63 = Math_imul($56, $14)|0;
    $64 = (($63>>>0) / 255)&-1;
    $65 = (($64) + ($b$0))|0;
    $66 = $59 >>> 3;
    $67 = $66 << 11;
    $68 = $62 >>> 2;
    $69 = $68 << 5;
    $70 = $69 | $67;
    $71 = $65 >>> 3;
    $72 = $70 | $71;
    $73 = $72&65535;
    HEAP16[$pixel$160>>1] = $73;
    $74 = ((($pixel$160)) + 2|0);
    $75 = ($41|0)==(0);
    if ($75) {
     break;
    } else {
     $length$061 = $41;$pixel$160 = $74;
    }
   }
   return;
   break;
  }
  case 2:  {
   $76 = ((($dst)) + 16|0);
   $77 = HEAP32[$76>>2]|0;
   $78 = ((($dst)) + 4|0);
   $79 = HEAP32[$78>>2]|0;
   $80 = ((($79)) + 9|0);
   $81 = HEAP8[$80>>0]|0;
   $82 = $81&255;
   $83 = (($77|0) / ($82|0))&-1;
   $84 = ($x2|0)<($x1|0);
   $85 = ((($dst)) + 20|0);
   $86 = HEAP32[$85>>2]|0;
   $87 = Math_imul($83, $y2)|0;
   $88 = (($86) + ($87<<1)|0);
   if ($84) {
    $94 = (($88) + ($x2<<1)|0);
    $95 = ($draw_end|0)!=(0);
    $96 = ((($94)) + 2|0);
    $$14 = $95 ? $94 : $96;
    $97 = (($x1) - ($x2))|0;
    $98 = $95&1;
    $99 = (($97) + ($98))|0;
    $length1$0$ph = $99;$pixel3$1$ph = $$14;
   } else {
    $89 = (($88) + ($x1<<1)|0);
    $90 = ($draw_end|0)!=(0);
    $91 = (($x2) - ($x1))|0;
    $92 = $90&1;
    $93 = (($91) + ($92))|0;
    $length1$0$ph = $93;$pixel3$1$ph = $89;
   }
   $100 = ($length1$0$ph|0)==(0);
   if ($100) {
    return;
   }
   $$pre153 = HEAP32[(125936)>>2]|0;
   $$pre154 = HEAP32[(125932)>>2]|0;
   $length1$064 = $length1$0$ph;$pixel3$163 = $pixel3$1$ph;
   while(1) {
    $101 = (($length1$064) + -1)|0;
    $102 = HEAP16[$pixel3$163>>1]|0;
    $103 = $102&65535;
    $104 = $103 >>> 11;
    $105 = (($$pre153) + ($104)|0);
    $106 = HEAP8[$105>>0]|0;
    $107 = $106&255;
    $108 = $103 >>> 5;
    $109 = $108 & 63;
    $110 = (($$pre154) + ($109)|0);
    $111 = HEAP8[$110>>0]|0;
    $112 = $111&255;
    $113 = $103 & 31;
    $114 = (($$pre153) + ($113)|0);
    $115 = HEAP8[$114>>0]|0;
    $116 = $115&255;
    $117 = (($107) + ($r$0))|0;
    $118 = ($117>>>0)>(255);
    $$ = $118 ? 255 : $117;
    $119 = (($112) + ($g$0))|0;
    $120 = ($119>>>0)>(255);
    $$8 = $120 ? 255 : $119;
    $121 = (($116) + ($b$0))|0;
    $122 = ($121>>>0)>(255);
    $$2 = $122 ? 255 : $121;
    $123 = $$ >>> 3;
    $124 = $123 << 11;
    $125 = $$8 >>> 2;
    $126 = $125 << 5;
    $127 = $126 | $124;
    $128 = $$2 >>> 3;
    $129 = $127 | $128;
    $130 = $129&65535;
    HEAP16[$pixel3$163>>1] = $130;
    $131 = ((($pixel3$163)) + 2|0);
    $132 = ($101|0)==(0);
    if ($132) {
     break;
    } else {
     $length1$064 = $101;$pixel3$163 = $131;
    }
   }
   return;
   break;
  }
  case 4:  {
   $133 = ((($dst)) + 16|0);
   $134 = HEAP32[$133>>2]|0;
   $135 = ((($dst)) + 4|0);
   $136 = HEAP32[$135>>2]|0;
   $137 = ((($136)) + 9|0);
   $138 = HEAP8[$137>>0]|0;
   $139 = $138&255;
   $140 = (($134|0) / ($139|0))&-1;
   $141 = ($x2|0)<($x1|0);
   $142 = ((($dst)) + 20|0);
   $143 = HEAP32[$142>>2]|0;
   $144 = Math_imul($140, $y2)|0;
   $145 = (($143) + ($144<<1)|0);
   if ($141) {
    $151 = (($145) + ($x2<<1)|0);
    $152 = ($draw_end|0)!=(0);
    $153 = ((($151)) + 2|0);
    $$15 = $152 ? $151 : $153;
    $154 = (($x1) - ($x2))|0;
    $155 = $152&1;
    $156 = (($154) + ($155))|0;
    $length8$0$ph = $156;$pixel10$1$ph = $$15;
   } else {
    $146 = (($145) + ($x1<<1)|0);
    $147 = ($draw_end|0)!=(0);
    $148 = (($x2) - ($x1))|0;
    $149 = $147&1;
    $150 = (($148) + ($149))|0;
    $length8$0$ph = $150;$pixel10$1$ph = $146;
   }
   $157 = ($length8$0$ph|0)==(0);
   if ($157) {
    return;
   }
   $$pre151 = HEAP32[(125936)>>2]|0;
   $$pre152 = HEAP32[(125932)>>2]|0;
   $length8$067 = $length8$0$ph;$pixel10$166 = $pixel10$1$ph;
   while(1) {
    $158 = (($length8$067) + -1)|0;
    $159 = HEAP16[$pixel10$166>>1]|0;
    $160 = $159&65535;
    $161 = $160 >>> 11;
    $162 = (($$pre151) + ($161)|0);
    $163 = HEAP8[$162>>0]|0;
    $164 = $163&255;
    $165 = $160 >>> 5;
    $166 = $165 & 63;
    $167 = (($$pre152) + ($166)|0);
    $168 = HEAP8[$167>>0]|0;
    $169 = $168&255;
    $170 = $160 & 31;
    $171 = (($$pre151) + ($170)|0);
    $172 = HEAP8[$171>>0]|0;
    $173 = $172&255;
    $174 = Math_imul($164, $r$0)|0;
    $175 = (($174>>>0) / 255)&-1;
    $176 = Math_imul($169, $g$0)|0;
    $177 = (($176>>>0) / 255)&-1;
    $178 = Math_imul($173, $b$0)|0;
    $179 = (($178>>>0) / 255)&-1;
    $180 = $175 >>> 3;
    $181 = $180 << 11;
    $182 = $177 >>> 2;
    $183 = $182 << 5;
    $184 = $183 | $181;
    $185 = $179 >>> 3;
    $186 = $184 | $185;
    $187 = $186&65535;
    HEAP16[$pixel10$166>>1] = $187;
    $188 = ((($pixel10$166)) + 2|0);
    $189 = ($158|0)==(0);
    if ($189) {
     break;
    } else {
     $length8$067 = $158;$pixel10$166 = $188;
    }
   }
   return;
   break;
  }
  default: {
   $190 = ((($dst)) + 16|0);
   $191 = HEAP32[$190>>2]|0;
   $192 = ((($dst)) + 4|0);
   $193 = HEAP32[$192>>2]|0;
   $194 = ((($193)) + 9|0);
   $195 = HEAP8[$194>>0]|0;
   $196 = $195&255;
   $197 = (($191|0) / ($196|0))&-1;
   $198 = ($x2|0)<($x1|0);
   $199 = ((($dst)) + 20|0);
   $200 = HEAP32[$199>>2]|0;
   $201 = Math_imul($197, $y2)|0;
   $202 = (($200) + ($201<<1)|0);
   if ($198) {
    $208 = (($202) + ($x2<<1)|0);
    $209 = ($draw_end|0)!=(0);
    $210 = ((($208)) + 2|0);
    $$16 = $209 ? $208 : $210;
    $211 = (($x1) - ($x2))|0;
    $212 = $209&1;
    $213 = (($211) + ($212))|0;
    $length15$0$ph = $213;$pixel17$1$ph = $$16;
   } else {
    $203 = (($202) + ($x1<<1)|0);
    $204 = ($draw_end|0)!=(0);
    $205 = (($x2) - ($x1))|0;
    $206 = $204&1;
    $207 = (($205) + ($206))|0;
    $length15$0$ph = $207;$pixel17$1$ph = $203;
   }
   $214 = ($length15$0$ph|0)==(0);
   if ($214) {
    return;
   }
   $215 = $r$0 >>> 3;
   $216 = $215 << 11;
   $217 = $g$0 >>> 2;
   $218 = $217 << 5;
   $219 = $b$0 >>> 3;
   $220 = $218 | $219;
   $221 = $220 | $216;
   $222 = $221&65535;
   $length15$059 = $length15$0$ph;$pixel17$158 = $pixel17$1$ph;
   while(1) {
    $223 = (($length15$059) + -1)|0;
    HEAP16[$pixel17$158>>1] = $222;
    $224 = ((($pixel17$158)) + 2|0);
    $225 = ($223|0)==(0);
    if ($225) {
     break;
    } else {
     $length15$059 = $223;$pixel17$158 = $224;
    }
   }
   return;
  }
  }
 }
 $226 = ($x2|0)==($x1|0);
 if ($226) {
  switch ($blendMode|0) {
  case 1:  {
   $227 = ((($dst)) + 16|0);
   $228 = HEAP32[$227>>2]|0;
   $229 = ((($dst)) + 4|0);
   $230 = HEAP32[$229>>2]|0;
   $231 = ((($230)) + 9|0);
   $232 = HEAP8[$231>>0]|0;
   $233 = $232&255;
   $234 = (($228|0) / ($233|0))&-1;
   $235 = ($y2|0)<($y1|0);
   $236 = ((($dst)) + 20|0);
   $237 = HEAP32[$236>>2]|0;
   if ($235) {
    $245 = Math_imul($234, $y2)|0;
    $246 = (($237) + ($245<<1)|0);
    $247 = (($246) + ($x2<<1)|0);
    $248 = ($draw_end|0)!=(0);
    $249 = (($247) + ($234<<1)|0);
    $$17 = $248 ? $247 : $249;
    $250 = (($y1) - ($y2))|0;
    $251 = $248&1;
    $252 = (($250) + ($251))|0;
    $length22$0$ph = $252;$pixel24$1$ph = $$17;
   } else {
    $238 = Math_imul($234, $y1)|0;
    $239 = (($237) + ($238<<1)|0);
    $240 = (($239) + ($x2<<1)|0);
    $241 = ($draw_end|0)!=(0);
    $242 = (($y2) - ($y1))|0;
    $243 = $241&1;
    $244 = (($242) + ($243))|0;
    $length22$0$ph = $244;$pixel24$1$ph = $240;
   }
   $253 = ($length22$0$ph|0)==(0);
   if ($253) {
    return;
   }
   $$pre149 = HEAP32[(125936)>>2]|0;
   $$pre150 = HEAP32[(125932)>>2]|0;
   $length22$073 = $length22$0$ph;$pixel24$172 = $pixel24$1$ph;
   while(1) {
    $254 = (($length22$073) + -1)|0;
    $255 = HEAP16[$pixel24$172>>1]|0;
    $256 = $255&65535;
    $257 = $256 >>> 11;
    $258 = (($$pre149) + ($257)|0);
    $259 = HEAP8[$258>>0]|0;
    $260 = $259&255;
    $261 = $256 >>> 5;
    $262 = $261 & 63;
    $263 = (($$pre150) + ($262)|0);
    $264 = HEAP8[$263>>0]|0;
    $265 = $264&255;
    $266 = $256 & 31;
    $267 = (($$pre149) + ($266)|0);
    $268 = HEAP8[$267>>0]|0;
    $269 = $268&255;
    $270 = Math_imul($260, $14)|0;
    $271 = (($270>>>0) / 255)&-1;
    $272 = (($271) + ($r$0))|0;
    $273 = Math_imul($265, $14)|0;
    $274 = (($273>>>0) / 255)&-1;
    $275 = (($274) + ($g$0))|0;
    $276 = Math_imul($269, $14)|0;
    $277 = (($276>>>0) / 255)&-1;
    $278 = (($277) + ($b$0))|0;
    $279 = $272 >>> 3;
    $280 = $279 << 11;
    $281 = $275 >>> 2;
    $282 = $281 << 5;
    $283 = $282 | $280;
    $284 = $278 >>> 3;
    $285 = $283 | $284;
    $286 = $285&65535;
    HEAP16[$pixel24$172>>1] = $286;
    $287 = (($pixel24$172) + ($234<<1)|0);
    $288 = ($254|0)==(0);
    if ($288) {
     break;
    } else {
     $length22$073 = $254;$pixel24$172 = $287;
    }
   }
   return;
   break;
  }
  case 2:  {
   $289 = ((($dst)) + 16|0);
   $290 = HEAP32[$289>>2]|0;
   $291 = ((($dst)) + 4|0);
   $292 = HEAP32[$291>>2]|0;
   $293 = ((($292)) + 9|0);
   $294 = HEAP8[$293>>0]|0;
   $295 = $294&255;
   $296 = (($290|0) / ($295|0))&-1;
   $297 = ($y2|0)<($y1|0);
   $298 = ((($dst)) + 20|0);
   $299 = HEAP32[$298>>2]|0;
   if ($297) {
    $307 = Math_imul($296, $y2)|0;
    $308 = (($299) + ($307<<1)|0);
    $309 = (($308) + ($x2<<1)|0);
    $310 = ($draw_end|0)!=(0);
    $311 = (($309) + ($296<<1)|0);
    $$18 = $310 ? $309 : $311;
    $312 = (($y1) - ($y2))|0;
    $313 = $310&1;
    $314 = (($312) + ($313))|0;
    $length29$0$ph = $314;$pixel31$1$ph = $$18;
   } else {
    $300 = Math_imul($296, $y1)|0;
    $301 = (($299) + ($300<<1)|0);
    $302 = (($301) + ($x2<<1)|0);
    $303 = ($draw_end|0)!=(0);
    $304 = (($y2) - ($y1))|0;
    $305 = $303&1;
    $306 = (($304) + ($305))|0;
    $length29$0$ph = $306;$pixel31$1$ph = $302;
   }
   $315 = ($length29$0$ph|0)==(0);
   if ($315) {
    return;
   }
   $$pre147 = HEAP32[(125936)>>2]|0;
   $$pre148 = HEAP32[(125932)>>2]|0;
   $length29$076 = $length29$0$ph;$pixel31$175 = $pixel31$1$ph;
   while(1) {
    $316 = (($length29$076) + -1)|0;
    $317 = HEAP16[$pixel31$175>>1]|0;
    $318 = $317&65535;
    $319 = $318 >>> 11;
    $320 = (($$pre147) + ($319)|0);
    $321 = HEAP8[$320>>0]|0;
    $322 = $321&255;
    $323 = $318 >>> 5;
    $324 = $323 & 63;
    $325 = (($$pre148) + ($324)|0);
    $326 = HEAP8[$325>>0]|0;
    $327 = $326&255;
    $328 = $318 & 31;
    $329 = (($$pre147) + ($328)|0);
    $330 = HEAP8[$329>>0]|0;
    $331 = $330&255;
    $332 = (($322) + ($r$0))|0;
    $333 = ($332>>>0)>(255);
    $$3 = $333 ? 255 : $332;
    $334 = (($327) + ($g$0))|0;
    $335 = ($334>>>0)>(255);
    $$9 = $335 ? 255 : $334;
    $336 = (($331) + ($b$0))|0;
    $337 = ($336>>>0)>(255);
    $$4 = $337 ? 255 : $336;
    $338 = $$3 >>> 3;
    $339 = $338 << 11;
    $340 = $$9 >>> 2;
    $341 = $340 << 5;
    $342 = $341 | $339;
    $343 = $$4 >>> 3;
    $344 = $342 | $343;
    $345 = $344&65535;
    HEAP16[$pixel31$175>>1] = $345;
    $346 = (($pixel31$175) + ($296<<1)|0);
    $347 = ($316|0)==(0);
    if ($347) {
     break;
    } else {
     $length29$076 = $316;$pixel31$175 = $346;
    }
   }
   return;
   break;
  }
  case 4:  {
   $348 = ((($dst)) + 16|0);
   $349 = HEAP32[$348>>2]|0;
   $350 = ((($dst)) + 4|0);
   $351 = HEAP32[$350>>2]|0;
   $352 = ((($351)) + 9|0);
   $353 = HEAP8[$352>>0]|0;
   $354 = $353&255;
   $355 = (($349|0) / ($354|0))&-1;
   $356 = ($y2|0)<($y1|0);
   $357 = ((($dst)) + 20|0);
   $358 = HEAP32[$357>>2]|0;
   if ($356) {
    $366 = Math_imul($355, $y2)|0;
    $367 = (($358) + ($366<<1)|0);
    $368 = (($367) + ($x2<<1)|0);
    $369 = ($draw_end|0)!=(0);
    $370 = (($368) + ($355<<1)|0);
    $$19 = $369 ? $368 : $370;
    $371 = (($y1) - ($y2))|0;
    $372 = $369&1;
    $373 = (($371) + ($372))|0;
    $length36$0$ph = $373;$pixel38$1$ph = $$19;
   } else {
    $359 = Math_imul($355, $y1)|0;
    $360 = (($358) + ($359<<1)|0);
    $361 = (($360) + ($x2<<1)|0);
    $362 = ($draw_end|0)!=(0);
    $363 = (($y2) - ($y1))|0;
    $364 = $362&1;
    $365 = (($363) + ($364))|0;
    $length36$0$ph = $365;$pixel38$1$ph = $361;
   }
   $374 = ($length36$0$ph|0)==(0);
   if ($374) {
    return;
   }
   $$pre145 = HEAP32[(125936)>>2]|0;
   $$pre146 = HEAP32[(125932)>>2]|0;
   $length36$079 = $length36$0$ph;$pixel38$178 = $pixel38$1$ph;
   while(1) {
    $375 = (($length36$079) + -1)|0;
    $376 = HEAP16[$pixel38$178>>1]|0;
    $377 = $376&65535;
    $378 = $377 >>> 11;
    $379 = (($$pre145) + ($378)|0);
    $380 = HEAP8[$379>>0]|0;
    $381 = $380&255;
    $382 = $377 >>> 5;
    $383 = $382 & 63;
    $384 = (($$pre146) + ($383)|0);
    $385 = HEAP8[$384>>0]|0;
    $386 = $385&255;
    $387 = $377 & 31;
    $388 = (($$pre145) + ($387)|0);
    $389 = HEAP8[$388>>0]|0;
    $390 = $389&255;
    $391 = Math_imul($381, $r$0)|0;
    $392 = (($391>>>0) / 255)&-1;
    $393 = Math_imul($386, $g$0)|0;
    $394 = (($393>>>0) / 255)&-1;
    $395 = Math_imul($390, $b$0)|0;
    $396 = (($395>>>0) / 255)&-1;
    $397 = $392 >>> 3;
    $398 = $397 << 11;
    $399 = $394 >>> 2;
    $400 = $399 << 5;
    $401 = $400 | $398;
    $402 = $396 >>> 3;
    $403 = $401 | $402;
    $404 = $403&65535;
    HEAP16[$pixel38$178>>1] = $404;
    $405 = (($pixel38$178) + ($355<<1)|0);
    $406 = ($375|0)==(0);
    if ($406) {
     break;
    } else {
     $length36$079 = $375;$pixel38$178 = $405;
    }
   }
   return;
   break;
  }
  default: {
   $407 = ((($dst)) + 16|0);
   $408 = HEAP32[$407>>2]|0;
   $409 = ((($dst)) + 4|0);
   $410 = HEAP32[$409>>2]|0;
   $411 = ((($410)) + 9|0);
   $412 = HEAP8[$411>>0]|0;
   $413 = $412&255;
   $414 = (($408|0) / ($413|0))&-1;
   $415 = ($y2|0)<($y1|0);
   $416 = ((($dst)) + 20|0);
   $417 = HEAP32[$416>>2]|0;
   if ($415) {
    $425 = Math_imul($414, $y2)|0;
    $426 = (($417) + ($425<<1)|0);
    $427 = (($426) + ($x2<<1)|0);
    $428 = ($draw_end|0)!=(0);
    $429 = (($427) + ($414<<1)|0);
    $$20 = $428 ? $427 : $429;
    $430 = (($y1) - ($y2))|0;
    $431 = $428&1;
    $432 = (($430) + ($431))|0;
    $length43$0$ph = $432;$pixel45$1$ph = $$20;
   } else {
    $418 = Math_imul($414, $y1)|0;
    $419 = (($417) + ($418<<1)|0);
    $420 = (($419) + ($x2<<1)|0);
    $421 = ($draw_end|0)!=(0);
    $422 = (($y2) - ($y1))|0;
    $423 = $421&1;
    $424 = (($422) + ($423))|0;
    $length43$0$ph = $424;$pixel45$1$ph = $420;
   }
   $433 = ($length43$0$ph|0)==(0);
   if ($433) {
    return;
   }
   $434 = $r$0 >>> 3;
   $435 = $434 << 11;
   $436 = $g$0 >>> 2;
   $437 = $436 << 5;
   $438 = $b$0 >>> 3;
   $439 = $437 | $438;
   $440 = $439 | $435;
   $441 = $440&65535;
   $length43$070 = $length43$0$ph;$pixel45$169 = $pixel45$1$ph;
   while(1) {
    $442 = (($length43$070) + -1)|0;
    HEAP16[$pixel45$169>>1] = $441;
    $443 = (($pixel45$169) + ($414<<1)|0);
    $444 = ($442|0)==(0);
    if ($444) {
     break;
    } else {
     $length43$070 = $442;$pixel45$169 = $443;
    }
   }
   return;
  }
  }
 }
 $445 = (($x1) - ($x2))|0;
 $446 = ($445|0)<(0);
 $447 = (0 - ($445))|0;
 $448 = $446 ? $447 : $445;
 $449 = (($y1) - ($y2))|0;
 $450 = ($449|0)<(0);
 $451 = (0 - ($449))|0;
 $452 = $450 ? $451 : $449;
 $453 = ($448|0)==($452|0);
 if ($453) {
  switch ($blendMode|0) {
  case 1:  {
   $454 = ((($dst)) + 16|0);
   $455 = HEAP32[$454>>2]|0;
   $456 = ((($dst)) + 4|0);
   $457 = HEAP32[$456>>2]|0;
   $458 = ((($457)) + 9|0);
   $459 = HEAP8[$458>>0]|0;
   $460 = $459&255;
   $461 = (($455|0) / ($460|0))&-1;
   $462 = ($y2|0)<($y1|0);
   $463 = ((($dst)) + 20|0);
   $464 = HEAP32[$463>>2]|0;
   if ($462) {
    $470 = Math_imul($461, $y2)|0;
    $471 = (($464) + ($470<<1)|0);
    $472 = (($471) + ($x2<<1)|0);
    $473 = ($x2|0)>($x1|0);
    $pitch51$1$v = $473 ? -1 : 1;
    $pitch51$1 = (($461) + ($pitch51$1$v))|0;
    $474 = ($draw_end|0)==(0);
    $475 = (($472) + ($pitch51$1<<1)|0);
    $$21 = $474 ? $475 : $472;
    $length50$0 = $449;$pitch51$2 = $pitch51$1;$pixel52$1 = $$21;
   } else {
    $465 = Math_imul($461, $y1)|0;
    $466 = (($464) + ($465<<1)|0);
    $467 = (($466) + ($x1<<1)|0);
    $468 = ($x2|0)<($x1|0);
    $pitch51$0$v = $468 ? -1 : 1;
    $pitch51$0 = (($461) + ($pitch51$0$v))|0;
    $469 = (($y2) - ($y1))|0;
    $length50$0 = $469;$pitch51$2 = $pitch51$0;$pixel52$1 = $467;
   }
   $not$34 = ($draw_end|0)!=(0);
   $476 = $not$34&1;
   $length50$0$ = (($length50$0) + ($476))|0;
   $477 = ($length50$0$|0)==(0);
   if ($477) {
    return;
   }
   $$pre143 = HEAP32[(125936)>>2]|0;
   $$pre144 = HEAP32[(125932)>>2]|0;
   $length50$185 = $length50$0$;$pixel52$284 = $pixel52$1;
   while(1) {
    $478 = (($length50$185) + -1)|0;
    $479 = HEAP16[$pixel52$284>>1]|0;
    $480 = $479&65535;
    $481 = $480 >>> 11;
    $482 = (($$pre143) + ($481)|0);
    $483 = HEAP8[$482>>0]|0;
    $484 = $483&255;
    $485 = $480 >>> 5;
    $486 = $485 & 63;
    $487 = (($$pre144) + ($486)|0);
    $488 = HEAP8[$487>>0]|0;
    $489 = $488&255;
    $490 = $480 & 31;
    $491 = (($$pre143) + ($490)|0);
    $492 = HEAP8[$491>>0]|0;
    $493 = $492&255;
    $494 = Math_imul($484, $14)|0;
    $495 = (($494>>>0) / 255)&-1;
    $496 = (($495) + ($r$0))|0;
    $497 = Math_imul($489, $14)|0;
    $498 = (($497>>>0) / 255)&-1;
    $499 = (($498) + ($g$0))|0;
    $500 = Math_imul($493, $14)|0;
    $501 = (($500>>>0) / 255)&-1;
    $502 = (($501) + ($b$0))|0;
    $503 = $496 >>> 3;
    $504 = $503 << 11;
    $505 = $499 >>> 2;
    $506 = $505 << 5;
    $507 = $506 | $504;
    $508 = $502 >>> 3;
    $509 = $507 | $508;
    $510 = $509&65535;
    HEAP16[$pixel52$284>>1] = $510;
    $511 = (($pixel52$284) + ($pitch51$2<<1)|0);
    $512 = ($478|0)==(0);
    if ($512) {
     break;
    } else {
     $length50$185 = $478;$pixel52$284 = $511;
    }
   }
   return;
   break;
  }
  case 2:  {
   $513 = ((($dst)) + 16|0);
   $514 = HEAP32[$513>>2]|0;
   $515 = ((($dst)) + 4|0);
   $516 = HEAP32[$515>>2]|0;
   $517 = ((($516)) + 9|0);
   $518 = HEAP8[$517>>0]|0;
   $519 = $518&255;
   $520 = (($514|0) / ($519|0))&-1;
   $521 = ($y2|0)<($y1|0);
   $522 = ((($dst)) + 20|0);
   $523 = HEAP32[$522>>2]|0;
   if ($521) {
    $529 = Math_imul($520, $y2)|0;
    $530 = (($523) + ($529<<1)|0);
    $531 = (($530) + ($x2<<1)|0);
    $532 = ($x2|0)>($x1|0);
    $pitch58$1$v = $532 ? -1 : 1;
    $pitch58$1 = (($520) + ($pitch58$1$v))|0;
    $533 = ($draw_end|0)==(0);
    $534 = (($531) + ($pitch58$1<<1)|0);
    $$22 = $533 ? $534 : $531;
    $length57$0 = $449;$pitch58$2 = $pitch58$1;$pixel59$1 = $$22;
   } else {
    $524 = Math_imul($520, $y1)|0;
    $525 = (($523) + ($524<<1)|0);
    $526 = (($525) + ($x1<<1)|0);
    $527 = ($x2|0)<($x1|0);
    $pitch58$0$v = $527 ? -1 : 1;
    $pitch58$0 = (($520) + ($pitch58$0$v))|0;
    $528 = (($y2) - ($y1))|0;
    $length57$0 = $528;$pitch58$2 = $pitch58$0;$pixel59$1 = $526;
   }
   $not$33 = ($draw_end|0)!=(0);
   $535 = $not$33&1;
   $length57$0$ = (($length57$0) + ($535))|0;
   $536 = ($length57$0$|0)==(0);
   if ($536) {
    return;
   }
   $$pre141 = HEAP32[(125936)>>2]|0;
   $$pre142 = HEAP32[(125932)>>2]|0;
   $length57$188 = $length57$0$;$pixel59$287 = $pixel59$1;
   while(1) {
    $537 = (($length57$188) + -1)|0;
    $538 = HEAP16[$pixel59$287>>1]|0;
    $539 = $538&65535;
    $540 = $539 >>> 11;
    $541 = (($$pre141) + ($540)|0);
    $542 = HEAP8[$541>>0]|0;
    $543 = $542&255;
    $544 = $539 >>> 5;
    $545 = $544 & 63;
    $546 = (($$pre142) + ($545)|0);
    $547 = HEAP8[$546>>0]|0;
    $548 = $547&255;
    $549 = $539 & 31;
    $550 = (($$pre141) + ($549)|0);
    $551 = HEAP8[$550>>0]|0;
    $552 = $551&255;
    $553 = (($543) + ($r$0))|0;
    $554 = ($553>>>0)>(255);
    $$5 = $554 ? 255 : $553;
    $555 = (($548) + ($g$0))|0;
    $556 = ($555>>>0)>(255);
    $$10 = $556 ? 255 : $555;
    $557 = (($552) + ($b$0))|0;
    $558 = ($557>>>0)>(255);
    $$6 = $558 ? 255 : $557;
    $559 = $$5 >>> 3;
    $560 = $559 << 11;
    $561 = $$10 >>> 2;
    $562 = $561 << 5;
    $563 = $562 | $560;
    $564 = $$6 >>> 3;
    $565 = $563 | $564;
    $566 = $565&65535;
    HEAP16[$pixel59$287>>1] = $566;
    $567 = (($pixel59$287) + ($pitch58$2<<1)|0);
    $568 = ($537|0)==(0);
    if ($568) {
     break;
    } else {
     $length57$188 = $537;$pixel59$287 = $567;
    }
   }
   return;
   break;
  }
  case 4:  {
   $569 = ((($dst)) + 16|0);
   $570 = HEAP32[$569>>2]|0;
   $571 = ((($dst)) + 4|0);
   $572 = HEAP32[$571>>2]|0;
   $573 = ((($572)) + 9|0);
   $574 = HEAP8[$573>>0]|0;
   $575 = $574&255;
   $576 = (($570|0) / ($575|0))&-1;
   $577 = ($y2|0)<($y1|0);
   $578 = ((($dst)) + 20|0);
   $579 = HEAP32[$578>>2]|0;
   if ($577) {
    $585 = Math_imul($576, $y2)|0;
    $586 = (($579) + ($585<<1)|0);
    $587 = (($586) + ($x2<<1)|0);
    $588 = ($x2|0)>($x1|0);
    $pitch65$1$v = $588 ? -1 : 1;
    $pitch65$1 = (($576) + ($pitch65$1$v))|0;
    $589 = ($draw_end|0)==(0);
    $590 = (($587) + ($pitch65$1<<1)|0);
    $$23 = $589 ? $590 : $587;
    $length64$0 = $449;$pitch65$2 = $pitch65$1;$pixel66$1 = $$23;
   } else {
    $580 = Math_imul($576, $y1)|0;
    $581 = (($579) + ($580<<1)|0);
    $582 = (($581) + ($x1<<1)|0);
    $583 = ($x2|0)<($x1|0);
    $pitch65$0$v = $583 ? -1 : 1;
    $pitch65$0 = (($576) + ($pitch65$0$v))|0;
    $584 = (($y2) - ($y1))|0;
    $length64$0 = $584;$pitch65$2 = $pitch65$0;$pixel66$1 = $582;
   }
   $not$32 = ($draw_end|0)!=(0);
   $591 = $not$32&1;
   $length64$0$ = (($length64$0) + ($591))|0;
   $592 = ($length64$0$|0)==(0);
   if ($592) {
    return;
   }
   $$pre139 = HEAP32[(125936)>>2]|0;
   $$pre140 = HEAP32[(125932)>>2]|0;
   $length64$191 = $length64$0$;$pixel66$290 = $pixel66$1;
   while(1) {
    $593 = (($length64$191) + -1)|0;
    $594 = HEAP16[$pixel66$290>>1]|0;
    $595 = $594&65535;
    $596 = $595 >>> 11;
    $597 = (($$pre139) + ($596)|0);
    $598 = HEAP8[$597>>0]|0;
    $599 = $598&255;
    $600 = $595 >>> 5;
    $601 = $600 & 63;
    $602 = (($$pre140) + ($601)|0);
    $603 = HEAP8[$602>>0]|0;
    $604 = $603&255;
    $605 = $595 & 31;
    $606 = (($$pre139) + ($605)|0);
    $607 = HEAP8[$606>>0]|0;
    $608 = $607&255;
    $609 = Math_imul($599, $r$0)|0;
    $610 = (($609>>>0) / 255)&-1;
    $611 = Math_imul($604, $g$0)|0;
    $612 = (($611>>>0) / 255)&-1;
    $613 = Math_imul($608, $b$0)|0;
    $614 = (($613>>>0) / 255)&-1;
    $615 = $610 >>> 3;
    $616 = $615 << 11;
    $617 = $612 >>> 2;
    $618 = $617 << 5;
    $619 = $618 | $616;
    $620 = $614 >>> 3;
    $621 = $619 | $620;
    $622 = $621&65535;
    HEAP16[$pixel66$290>>1] = $622;
    $623 = (($pixel66$290) + ($pitch65$2<<1)|0);
    $624 = ($593|0)==(0);
    if ($624) {
     break;
    } else {
     $length64$191 = $593;$pixel66$290 = $623;
    }
   }
   return;
   break;
  }
  default: {
   $625 = ((($dst)) + 16|0);
   $626 = HEAP32[$625>>2]|0;
   $627 = ((($dst)) + 4|0);
   $628 = HEAP32[$627>>2]|0;
   $629 = ((($628)) + 9|0);
   $630 = HEAP8[$629>>0]|0;
   $631 = $630&255;
   $632 = (($626|0) / ($631|0))&-1;
   $633 = ($y2|0)<($y1|0);
   $634 = ((($dst)) + 20|0);
   $635 = HEAP32[$634>>2]|0;
   if ($633) {
    $641 = Math_imul($632, $y2)|0;
    $642 = (($635) + ($641<<1)|0);
    $643 = (($642) + ($x2<<1)|0);
    $644 = ($x2|0)>($x1|0);
    $pitch72$1$v = $644 ? -1 : 1;
    $pitch72$1 = (($632) + ($pitch72$1$v))|0;
    $645 = ($draw_end|0)==(0);
    $646 = (($643) + ($pitch72$1<<1)|0);
    $$24 = $645 ? $646 : $643;
    $length71$0 = $449;$pitch72$2 = $pitch72$1;$pixel73$1 = $$24;
   } else {
    $636 = Math_imul($632, $y1)|0;
    $637 = (($635) + ($636<<1)|0);
    $638 = (($637) + ($x1<<1)|0);
    $639 = ($x2|0)<($x1|0);
    $pitch72$0$v = $639 ? -1 : 1;
    $pitch72$0 = (($632) + ($pitch72$0$v))|0;
    $640 = (($y2) - ($y1))|0;
    $length71$0 = $640;$pitch72$2 = $pitch72$0;$pixel73$1 = $638;
   }
   $not$35 = ($draw_end|0)!=(0);
   $647 = $not$35&1;
   $length71$0$ = (($length71$0) + ($647))|0;
   $648 = ($length71$0$|0)==(0);
   if ($648) {
    return;
   }
   $649 = $r$0 >>> 3;
   $650 = $649 << 11;
   $651 = $g$0 >>> 2;
   $652 = $651 << 5;
   $653 = $b$0 >>> 3;
   $654 = $652 | $653;
   $655 = $654 | $650;
   $656 = $655&65535;
   $length71$182 = $length71$0$;$pixel73$281 = $pixel73$1;
   while(1) {
    $657 = (($length71$182) + -1)|0;
    HEAP16[$pixel73$281>>1] = $656;
    $658 = (($pixel73$281) + ($pitch72$2<<1)|0);
    $659 = ($657|0)==(0);
    if ($659) {
     break;
    } else {
     $length71$182 = $657;$pixel73$281 = $658;
    }
   }
   return;
  }
  }
 } else {
  switch ($blendMode|0) {
  case 1:  {
   $660 = (($x2) - ($x1))|0;
   $661 = ($660|0)<(0);
   $662 = (0 - ($660))|0;
   $663 = $661 ? $662 : $660;
   $664 = (($y2) - ($y1))|0;
   $665 = ($664|0)<(0);
   $666 = (0 - ($664))|0;
   $667 = $665 ? $666 : $664;
   $668 = ($663|0)<($667|0);
   if ($668) {
    $672 = $663 << 1;
    $673 = (($672) - ($667))|0;
    $674 = (($663) - ($667))|0;
    $d$0 = $673;$dinc1$0 = $672;$dinc2$0$in = $674;$numpixels$0$in = $667;$xinc1$0 = 0;$yinc1$0 = 1;
   } else {
    $669 = $667 << 1;
    $670 = (($669) - ($663))|0;
    $671 = (($667) - ($663))|0;
    $d$0 = $670;$dinc1$0 = $669;$dinc2$0$in = $671;$numpixels$0$in = $663;$xinc1$0 = 1;$yinc1$0 = 0;
   }
   $dinc2$0 = $dinc2$0$in << 1;
   $675 = ($x2|0)<($x1|0);
   $676 = (0 - ($xinc1$0))|0;
   $$xinc1$0 = $675 ? $676 : $xinc1$0;
   $$25 = $675 ? -1 : 1;
   $677 = ($y2|0)<($y1|0);
   $678 = (0 - ($yinc1$0))|0;
   $yinc1$1 = $677 ? $678 : $yinc1$0;
   $yinc2$1 = $677 ? -1 : 1;
   $not$30 = ($draw_end|0)!=(0);
   $numpixels$0 = $not$30&1;
   $numpixels$0$in$numpixels$0 = (($numpixels$0$in) + ($numpixels$0))|0;
   $679 = ($numpixels$0$in$numpixels$0|0)>(0);
   if (!($679)) {
    return;
   }
   $680 = ((($dst)) + 20|0);
   $681 = ((($dst)) + 16|0);
   $$pre135 = HEAP32[$680>>2]|0;
   $$pre136 = HEAP32[$681>>2]|0;
   $$pre137 = HEAP32[(125936)>>2]|0;
   $$pre138 = HEAP32[(125932)>>2]|0;
   $d$199 = $d$0;$i$098 = 0;$x$0100 = $x1;$y$0101 = $y1;
   while(1) {
    $682 = Math_imul($$pre136, $y$0101)|0;
    $683 = (($$pre135) + ($682)|0);
    $684 = $x$0100 << 1;
    $685 = (($683) + ($684)|0);
    $686 = HEAP16[$685>>1]|0;
    $687 = $686&65535;
    $688 = $687 >>> 11;
    $689 = (($$pre137) + ($688)|0);
    $690 = HEAP8[$689>>0]|0;
    $691 = $690&255;
    $692 = $687 >>> 5;
    $693 = $692 & 63;
    $694 = (($$pre138) + ($693)|0);
    $695 = HEAP8[$694>>0]|0;
    $696 = $695&255;
    $697 = $687 & 31;
    $698 = (($$pre137) + ($697)|0);
    $699 = HEAP8[$698>>0]|0;
    $700 = $699&255;
    $701 = Math_imul($691, $14)|0;
    $702 = (($701>>>0) / 255)&-1;
    $703 = (($702) + ($r$0))|0;
    $704 = Math_imul($696, $14)|0;
    $705 = (($704>>>0) / 255)&-1;
    $706 = (($705) + ($g$0))|0;
    $707 = Math_imul($700, $14)|0;
    $708 = (($707>>>0) / 255)&-1;
    $709 = (($708) + ($b$0))|0;
    $710 = $703 >>> 3;
    $711 = $710 << 11;
    $712 = $706 >>> 2;
    $713 = $712 << 5;
    $714 = $713 | $711;
    $715 = $709 >>> 3;
    $716 = $714 | $715;
    $717 = $716&65535;
    HEAP16[$685>>1] = $717;
    $718 = ($d$199|0)<(0);
    $dinc1$0$dinc2$0 = $718 ? $dinc1$0 : $dinc2$0;
    $$xinc1$0$$25 = $718 ? $$xinc1$0 : $$25;
    $yinc1$1$yinc2$1 = $718 ? $yinc1$1 : $yinc2$1;
    $y$1 = (($y$0101) + ($yinc1$1$yinc2$1))|0;
    $x$1 = (($$xinc1$0$$25) + ($x$0100))|0;
    $d$2 = (($dinc1$0$dinc2$0) + ($d$199))|0;
    $719 = (($i$098) + 1)|0;
    $720 = ($719|0)<($numpixels$0$in$numpixels$0|0);
    if ($720) {
     $d$199 = $d$2;$i$098 = $719;$x$0100 = $x$1;$y$0101 = $y$1;
    } else {
     break;
    }
   }
   return;
   break;
  }
  case 2:  {
   $721 = (($x2) - ($x1))|0;
   $722 = ($721|0)<(0);
   $723 = (0 - ($721))|0;
   $724 = $722 ? $723 : $721;
   $725 = (($y2) - ($y1))|0;
   $726 = ($725|0)<(0);
   $727 = (0 - ($725))|0;
   $728 = $726 ? $727 : $725;
   $729 = ($724|0)<($728|0);
   if ($729) {
    $733 = $724 << 1;
    $734 = (($733) - ($728))|0;
    $735 = (($724) - ($728))|0;
    $d87$0 = $734;$dinc188$0 = $733;$dinc289$0$in = $735;$numpixels86$0$in = $728;$xinc191$0 = 0;$yinc194$0 = 1;
   } else {
    $730 = $728 << 1;
    $731 = (($730) - ($724))|0;
    $732 = (($728) - ($724))|0;
    $d87$0 = $731;$dinc188$0 = $730;$dinc289$0$in = $732;$numpixels86$0$in = $724;$xinc191$0 = 1;$yinc194$0 = 0;
   }
   $dinc289$0 = $dinc289$0$in << 1;
   $736 = ($x2|0)<($x1|0);
   $737 = (0 - ($xinc191$0))|0;
   $$xinc191$0 = $736 ? $737 : $xinc191$0;
   $$26 = $736 ? -1 : 1;
   $738 = ($y2|0)<($y1|0);
   $739 = (0 - ($yinc194$0))|0;
   $yinc194$1 = $738 ? $739 : $yinc194$0;
   $yinc295$1 = $738 ? -1 : 1;
   $not$29 = ($draw_end|0)!=(0);
   $numpixels86$0 = $not$29&1;
   $numpixels86$0$in$numpixels86$0 = (($numpixels86$0$in) + ($numpixels86$0))|0;
   $740 = ($numpixels86$0$in$numpixels86$0|0)>(0);
   if (!($740)) {
    return;
   }
   $741 = ((($dst)) + 20|0);
   $742 = ((($dst)) + 16|0);
   $$pre131 = HEAP32[$741>>2]|0;
   $$pre132 = HEAP32[$742>>2]|0;
   $$pre133 = HEAP32[(125936)>>2]|0;
   $$pre134 = HEAP32[(125932)>>2]|0;
   $d87$1104 = $d87$0;$i83$0103 = 0;$x90$0105 = $x1;$y93$0106 = $y1;
   while(1) {
    $743 = Math_imul($$pre132, $y93$0106)|0;
    $744 = (($$pre131) + ($743)|0);
    $745 = $x90$0105 << 1;
    $746 = (($744) + ($745)|0);
    $747 = HEAP16[$746>>1]|0;
    $748 = $747&65535;
    $749 = $748 >>> 11;
    $750 = (($$pre133) + ($749)|0);
    $751 = HEAP8[$750>>0]|0;
    $752 = $751&255;
    $753 = $748 >>> 5;
    $754 = $753 & 63;
    $755 = (($$pre134) + ($754)|0);
    $756 = HEAP8[$755>>0]|0;
    $757 = $756&255;
    $758 = $748 & 31;
    $759 = (($$pre133) + ($758)|0);
    $760 = HEAP8[$759>>0]|0;
    $761 = $760&255;
    $762 = (($752) + ($r$0))|0;
    $763 = ($762>>>0)>(255);
    $$11 = $763 ? 255 : $762;
    $764 = (($757) + ($g$0))|0;
    $765 = ($764>>>0)>(255);
    $$7 = $765 ? 255 : $764;
    $766 = (($761) + ($b$0))|0;
    $767 = ($766>>>0)>(255);
    $$12 = $767 ? 255 : $766;
    $768 = $$11 >>> 3;
    $769 = $768 << 11;
    $770 = $$7 >>> 2;
    $771 = $770 << 5;
    $772 = $771 | $769;
    $773 = $$12 >>> 3;
    $774 = $772 | $773;
    $775 = $774&65535;
    HEAP16[$746>>1] = $775;
    $776 = ($d87$1104|0)<(0);
    $dinc188$0$dinc289$0 = $776 ? $dinc188$0 : $dinc289$0;
    $$xinc191$0$$26 = $776 ? $$xinc191$0 : $$26;
    $yinc194$1$yinc295$1 = $776 ? $yinc194$1 : $yinc295$1;
    $y93$1 = (($y93$0106) + ($yinc194$1$yinc295$1))|0;
    $x90$1 = (($$xinc191$0$$26) + ($x90$0105))|0;
    $d87$2 = (($dinc188$0$dinc289$0) + ($d87$1104))|0;
    $777 = (($i83$0103) + 1)|0;
    $778 = ($777|0)<($numpixels86$0$in$numpixels86$0|0);
    if ($778) {
     $d87$1104 = $d87$2;$i83$0103 = $777;$x90$0105 = $x90$1;$y93$0106 = $y93$1;
    } else {
     break;
    }
   }
   return;
   break;
  }
  case 4:  {
   $779 = (($x2) - ($x1))|0;
   $780 = ($779|0)<(0);
   $781 = (0 - ($779))|0;
   $782 = $780 ? $781 : $779;
   $783 = (($y2) - ($y1))|0;
   $784 = ($783|0)<(0);
   $785 = (0 - ($783))|0;
   $786 = $784 ? $785 : $783;
   $787 = ($782|0)<($786|0);
   if ($787) {
    $791 = $782 << 1;
    $792 = (($791) - ($786))|0;
    $793 = (($782) - ($786))|0;
    $d105$0 = $792;$dinc1106$0 = $791;$dinc2107$0$in = $793;$numpixels104$0$in = $786;$xinc1109$0 = 0;$yinc1112$0 = 1;
   } else {
    $788 = $786 << 1;
    $789 = (($788) - ($782))|0;
    $790 = (($786) - ($782))|0;
    $d105$0 = $789;$dinc1106$0 = $788;$dinc2107$0$in = $790;$numpixels104$0$in = $782;$xinc1109$0 = 1;$yinc1112$0 = 0;
   }
   $dinc2107$0 = $dinc2107$0$in << 1;
   $794 = ($x2|0)<($x1|0);
   $795 = (0 - ($xinc1109$0))|0;
   $$xinc1109$0 = $794 ? $795 : $xinc1109$0;
   $$27 = $794 ? -1 : 1;
   $796 = ($y2|0)<($y1|0);
   $797 = (0 - ($yinc1112$0))|0;
   $yinc1112$1 = $796 ? $797 : $yinc1112$0;
   $yinc2113$1 = $796 ? -1 : 1;
   $not$ = ($draw_end|0)!=(0);
   $numpixels104$0 = $not$&1;
   $numpixels104$0$in$numpixels104$0 = (($numpixels104$0$in) + ($numpixels104$0))|0;
   $798 = ($numpixels104$0$in$numpixels104$0|0)>(0);
   if (!($798)) {
    return;
   }
   $799 = ((($dst)) + 20|0);
   $800 = ((($dst)) + 16|0);
   $$pre = HEAP32[$799>>2]|0;
   $$pre128 = HEAP32[$800>>2]|0;
   $$pre129 = HEAP32[(125936)>>2]|0;
   $$pre130 = HEAP32[(125932)>>2]|0;
   $d105$1109 = $d105$0;$i101$0108 = 0;$x108$0110 = $x1;$y111$0111 = $y1;
   while(1) {
    $801 = Math_imul($$pre128, $y111$0111)|0;
    $802 = (($$pre) + ($801)|0);
    $803 = $x108$0110 << 1;
    $804 = (($802) + ($803)|0);
    $805 = HEAP16[$804>>1]|0;
    $806 = $805&65535;
    $807 = $806 >>> 11;
    $808 = (($$pre129) + ($807)|0);
    $809 = HEAP8[$808>>0]|0;
    $810 = $809&255;
    $811 = $806 >>> 5;
    $812 = $811 & 63;
    $813 = (($$pre130) + ($812)|0);
    $814 = HEAP8[$813>>0]|0;
    $815 = $814&255;
    $816 = $806 & 31;
    $817 = (($$pre129) + ($816)|0);
    $818 = HEAP8[$817>>0]|0;
    $819 = $818&255;
    $820 = Math_imul($810, $r$0)|0;
    $821 = (($820>>>0) / 255)&-1;
    $822 = Math_imul($815, $g$0)|0;
    $823 = (($822>>>0) / 255)&-1;
    $824 = Math_imul($819, $b$0)|0;
    $825 = (($824>>>0) / 255)&-1;
    $826 = $821 >>> 3;
    $827 = $826 << 11;
    $828 = $823 >>> 2;
    $829 = $828 << 5;
    $830 = $829 | $827;
    $831 = $825 >>> 3;
    $832 = $830 | $831;
    $833 = $832&65535;
    HEAP16[$804>>1] = $833;
    $834 = ($d105$1109|0)<(0);
    $dinc1106$0$dinc2107$0 = $834 ? $dinc1106$0 : $dinc2107$0;
    $$xinc1109$0$$27 = $834 ? $$xinc1109$0 : $$27;
    $yinc1112$1$yinc2113$1 = $834 ? $yinc1112$1 : $yinc2113$1;
    $y111$1 = (($y111$0111) + ($yinc1112$1$yinc2113$1))|0;
    $x108$1 = (($$xinc1109$0$$27) + ($x108$0110))|0;
    $d105$2 = (($dinc1106$0$dinc2107$0) + ($d105$1109))|0;
    $835 = (($i101$0108) + 1)|0;
    $836 = ($835|0)<($numpixels104$0$in$numpixels104$0|0);
    if ($836) {
     $d105$1109 = $d105$2;$i101$0108 = $835;$x108$0110 = $x108$1;$y111$0111 = $y111$1;
    } else {
     break;
    }
   }
   return;
   break;
  }
  default: {
   $837 = (($x2) - ($x1))|0;
   $838 = ($837|0)<(0);
   $839 = (0 - ($837))|0;
   $840 = $838 ? $839 : $837;
   $841 = (($y2) - ($y1))|0;
   $842 = ($841|0)<(0);
   $843 = (0 - ($841))|0;
   $844 = $842 ? $843 : $841;
   $845 = ($840|0)<($844|0);
   if ($845) {
    $849 = $840 << 1;
    $850 = (($849) - ($844))|0;
    $851 = (($840) - ($844))|0;
    $d123$0 = $850;$dinc1124$0 = $849;$dinc2125$0$in = $851;$numpixels122$0$in = $844;$xinc1127$0 = 0;$yinc1130$0 = 1;
   } else {
    $846 = $844 << 1;
    $847 = (($846) - ($840))|0;
    $848 = (($844) - ($840))|0;
    $d123$0 = $847;$dinc1124$0 = $846;$dinc2125$0$in = $848;$numpixels122$0$in = $840;$xinc1127$0 = 1;$yinc1130$0 = 0;
   }
   $dinc2125$0 = $dinc2125$0$in << 1;
   $852 = ($x2|0)<($x1|0);
   $853 = (0 - ($xinc1127$0))|0;
   $$xinc1127$0 = $852 ? $853 : $xinc1127$0;
   $$28 = $852 ? -1 : 1;
   $854 = ($y2|0)<($y1|0);
   $855 = (0 - ($yinc1130$0))|0;
   $yinc1130$1 = $854 ? $855 : $yinc1130$0;
   $yinc2131$1 = $854 ? -1 : 1;
   $not$31 = ($draw_end|0)!=(0);
   $numpixels122$0 = $not$31&1;
   $numpixels122$0$in$numpixels122$0 = (($numpixels122$0$in) + ($numpixels122$0))|0;
   $856 = ($numpixels122$0$in$numpixels122$0|0)>(0);
   if (!($856)) {
    return;
   }
   $857 = ((($dst)) + 20|0);
   $858 = HEAP32[$857>>2]|0;
   $859 = ((($dst)) + 16|0);
   $860 = HEAP32[$859>>2]|0;
   $861 = $r$0 >>> 3;
   $862 = $861 << 11;
   $863 = $g$0 >>> 2;
   $864 = $863 << 5;
   $865 = $b$0 >>> 3;
   $866 = $864 | $865;
   $867 = $866 | $862;
   $868 = $867&65535;
   $d123$194 = $d123$0;$i119$093 = 0;$x126$095 = $x1;$y129$096 = $y1;
   while(1) {
    $869 = Math_imul($860, $y129$096)|0;
    $870 = (($858) + ($869)|0);
    $871 = $x126$095 << 1;
    $872 = (($870) + ($871)|0);
    HEAP16[$872>>1] = $868;
    $873 = ($d123$194|0)<(0);
    $dinc1124$0$dinc2125$0 = $873 ? $dinc1124$0 : $dinc2125$0;
    $$xinc1127$0$$28 = $873 ? $$xinc1127$0 : $$28;
    $yinc1130$1$yinc2131$1 = $873 ? $yinc1130$1 : $yinc2131$1;
    $y129$1 = (($y129$096) + ($yinc1130$1$yinc2131$1))|0;
    $x126$1 = (($$xinc1127$0$$28) + ($x126$095))|0;
    $d123$2 = (($dinc1124$0$dinc2125$0) + ($d123$194))|0;
    $874 = (($i119$093) + 1)|0;
    $875 = ($874|0)<($numpixels122$0$in$numpixels122$0|0);
    if ($875) {
     $d123$194 = $d123$2;$i119$093 = $874;$x126$095 = $x126$1;$y129$096 = $y129$1;
    } else {
     break;
    }
   }
   return;
  }
  }
 }
}
function _SDL_BlendLine_RGB2($dst,$x1,$y1,$x2,$y2,$blendMode,$_r,$_g,$_b,$_a,$draw_end) {
 $dst = $dst|0;
 $x1 = $x1|0;
 $y1 = $y1|0;
 $x2 = $x2|0;
 $y2 = $y2|0;
 $blendMode = $blendMode|0;
 $_r = $_r|0;
 $_g = $_g|0;
 $_b = $_b|0;
 $_a = $_a|0;
 $draw_end = $draw_end|0;
 var $$ = 0, $$10 = 0, $$11 = 0, $$12 = 0, $$13 = 0, $$14 = 0, $$15 = 0, $$16 = 0, $$17 = 0, $$18 = 0, $$19 = 0, $$2 = 0, $$20 = 0, $$21 = 0, $$22 = 0, $$23 = 0, $$24 = 0, $$25 = 0, $$26 = 0, $$27 = 0;
 var $$28 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$6 = 0, $$7 = 0, $$8 = 0, $$9 = 0, $$phi$trans$insert = 0, $$phi$trans$insert132 = 0, $$phi$trans$insert136 = 0, $$phi$trans$insert139 = 0, $$phi$trans$insert140 = 0, $$phi$trans$insert148 = 0, $$phi$trans$insert149 = 0, $$phi$trans$insert153 = 0, $$phi$trans$insert156 = 0, $$phi$trans$insert157 = 0, $$phi$trans$insert165 = 0, $$phi$trans$insert166 = 0;
 var $$phi$trans$insert170 = 0, $$phi$trans$insert173 = 0, $$phi$trans$insert174 = 0, $$phi$trans$insert180 = 0, $$phi$trans$insert181 = 0, $$phi$trans$insert185 = 0, $$phi$trans$insert188 = 0, $$phi$trans$insert189 = 0, $$phi$trans$insert195 = 0, $$phi$trans$insert196 = 0, $$phi$trans$insert200 = 0, $$phi$trans$insert203 = 0, $$phi$trans$insert204 = 0, $$phi$trans$insert210 = 0, $$phi$trans$insert211 = 0, $$phi$trans$insert215 = 0, $$phi$trans$insert218 = 0, $$phi$trans$insert219 = 0, $$phi$trans$insert225 = 0, $$phi$trans$insert226 = 0;
 var $$phi$trans$insert230 = 0, $$phi$trans$insert233 = 0, $$phi$trans$insert234 = 0, $$phi$trans$insert240 = 0, $$phi$trans$insert241 = 0, $$phi$trans$insert245 = 0, $$phi$trans$insert248 = 0, $$phi$trans$insert249 = 0, $$phi$trans$insert255 = 0, $$phi$trans$insert256 = 0, $$phi$trans$insert260 = 0, $$phi$trans$insert263 = 0, $$phi$trans$insert264 = 0, $$phi$trans$insert270 = 0, $$phi$trans$insert271 = 0, $$phi$trans$insert275 = 0, $$phi$trans$insert278 = 0, $$phi$trans$insert279 = 0, $$phi$trans$insert285 = 0, $$phi$trans$insert286 = 0;
 var $$phi$trans$insert290 = 0, $$phi$trans$insert293 = 0, $$phi$trans$insert294 = 0, $$phi$trans$insert300 = 0, $$phi$trans$insert301 = 0, $$phi$trans$insert305 = 0, $$phi$trans$insert308 = 0, $$phi$trans$insert309 = 0, $$pre = 0, $$pre128 = 0, $$pre129 = 0, $$pre130 = 0, $$pre131 = 0, $$pre133 = 0, $$pre134 = 0, $$pre137 = 0, $$pre138 = 0, $$pre141 = 0, $$pre142 = 0, $$pre143 = 0;
 var $$pre144 = 0, $$pre145 = 0, $$pre146 = 0, $$pre147 = 0, $$pre150 = 0, $$pre151 = 0, $$pre154 = 0, $$pre155 = 0, $$pre158 = 0, $$pre159 = 0, $$pre160 = 0, $$pre161 = 0, $$pre162 = 0, $$pre163 = 0, $$pre164 = 0, $$pre167 = 0, $$pre168 = 0, $$pre171 = 0, $$pre172 = 0, $$pre175 = 0;
 var $$pre176 = 0, $$pre177 = 0, $$pre178 = 0, $$pre179 = 0, $$pre182 = 0, $$pre183 = 0, $$pre186 = 0, $$pre187 = 0, $$pre190 = 0, $$pre191 = 0, $$pre192 = 0, $$pre193 = 0, $$pre194 = 0, $$pre197 = 0, $$pre198 = 0, $$pre201 = 0, $$pre202 = 0, $$pre205 = 0, $$pre206 = 0, $$pre207 = 0;
 var $$pre208 = 0, $$pre209 = 0, $$pre212 = 0, $$pre213 = 0, $$pre216 = 0, $$pre217 = 0, $$pre220 = 0, $$pre221 = 0, $$pre222 = 0, $$pre223 = 0, $$pre224 = 0, $$pre227 = 0, $$pre228 = 0, $$pre231 = 0, $$pre232 = 0, $$pre235 = 0, $$pre236 = 0, $$pre237 = 0, $$pre238 = 0, $$pre239 = 0;
 var $$pre242 = 0, $$pre243 = 0, $$pre246 = 0, $$pre247 = 0, $$pre250 = 0, $$pre251 = 0, $$pre252 = 0, $$pre253 = 0, $$pre254 = 0, $$pre257 = 0, $$pre258 = 0, $$pre261 = 0, $$pre262 = 0, $$pre265 = 0, $$pre266 = 0, $$pre267 = 0, $$pre268 = 0, $$pre269 = 0, $$pre272 = 0, $$pre273 = 0;
 var $$pre276 = 0, $$pre277 = 0, $$pre280 = 0, $$pre281 = 0, $$pre282 = 0, $$pre283 = 0, $$pre284 = 0, $$pre287 = 0, $$pre288 = 0, $$pre291 = 0, $$pre292 = 0, $$pre295 = 0, $$pre296 = 0, $$pre297 = 0, $$pre298 = 0, $$pre299 = 0, $$pre302 = 0, $$pre303 = 0, $$pre306 = 0, $$pre307 = 0;
 var $$pre310 = 0, $$pre311 = 0, $$xinc1$0 = 0, $$xinc1$0$$25 = 0, $$xinc1109$0 = 0, $$xinc1109$0$$27 = 0, $$xinc1127$0 = 0, $$xinc1127$0$$28 = 0, $$xinc191$0 = 0, $$xinc191$0$$26 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0;
 var $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0;
 var $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0;
 var $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0;
 var $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0;
 var $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0;
 var $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0;
 var $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0;
 var $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0;
 var $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0;
 var $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0;
 var $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0;
 var $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0;
 var $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0;
 var $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0;
 var $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0;
 var $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0;
 var $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0;
 var $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0;
 var $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0;
 var $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0;
 var $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0;
 var $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0;
 var $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0;
 var $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0;
 var $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0;
 var $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0;
 var $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0;
 var $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0;
 var $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0;
 var $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0;
 var $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0;
 var $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0;
 var $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0;
 var $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0;
 var $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0;
 var $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0;
 var $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0;
 var $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0;
 var $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0;
 var $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0;
 var $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0;
 var $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0;
 var $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0;
 var $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0;
 var $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0;
 var $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0;
 var $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0;
 var $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0;
 var $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0;
 var $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0;
 var $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0;
 var $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0;
 var $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $a$0 = 0, $b$0 = 0, $blendMode$off = 0, $d$0 = 0, $d$199 = 0, $d$2 = 0, $d105$0 = 0, $d105$1109 = 0;
 var $d105$2 = 0, $d123$0 = 0, $d123$194 = 0, $d123$2 = 0, $d87$0 = 0, $d87$1104 = 0, $d87$2 = 0, $dinc1$0 = 0, $dinc1$0$dinc2$0 = 0, $dinc1106$0 = 0, $dinc1106$0$dinc2107$0 = 0, $dinc1124$0 = 0, $dinc1124$0$dinc2125$0 = 0, $dinc188$0 = 0, $dinc188$0$dinc289$0 = 0, $dinc2$0 = 0, $dinc2$0$in = 0, $dinc2107$0 = 0, $dinc2107$0$in = 0, $dinc2125$0 = 0;
 var $dinc2125$0$in = 0, $dinc289$0 = 0, $dinc289$0$in = 0, $g$0 = 0, $i$098 = 0, $i101$0108 = 0, $i119$093 = 0, $i83$0103 = 0, $length$0$ph = 0, $length$061 = 0, $length1$0$ph = 0, $length1$064 = 0, $length15$0$ph = 0, $length15$059 = 0, $length22$0$ph = 0, $length22$073 = 0, $length29$0$ph = 0, $length29$076 = 0, $length36$0$ph = 0, $length36$079 = 0;
 var $length43$0$ph = 0, $length43$070 = 0, $length50$0 = 0, $length50$0$ = 0, $length50$185 = 0, $length57$0 = 0, $length57$0$ = 0, $length57$188 = 0, $length64$0 = 0, $length64$0$ = 0, $length64$191 = 0, $length71$0 = 0, $length71$0$ = 0, $length71$182 = 0, $length8$0$ph = 0, $length8$067 = 0, $not$ = 0, $not$29 = 0, $not$30 = 0, $not$31 = 0;
 var $not$32 = 0, $not$33 = 0, $not$34 = 0, $not$35 = 0, $numpixels$0 = 0, $numpixels$0$in = 0, $numpixels$0$in$numpixels$0 = 0, $numpixels104$0 = 0, $numpixels104$0$in = 0, $numpixels104$0$in$numpixels104$0 = 0, $numpixels122$0 = 0, $numpixels122$0$in = 0, $numpixels122$0$in$numpixels122$0 = 0, $numpixels86$0 = 0, $numpixels86$0$in = 0, $numpixels86$0$in$numpixels86$0 = 0, $pitch51$0 = 0, $pitch51$0$v = 0, $pitch51$1 = 0, $pitch51$1$v = 0;
 var $pitch51$2 = 0, $pitch58$0 = 0, $pitch58$0$v = 0, $pitch58$1 = 0, $pitch58$1$v = 0, $pitch58$2 = 0, $pitch65$0 = 0, $pitch65$0$v = 0, $pitch65$1 = 0, $pitch65$1$v = 0, $pitch65$2 = 0, $pitch72$0 = 0, $pitch72$0$v = 0, $pitch72$1 = 0, $pitch72$1$v = 0, $pitch72$2 = 0, $pixel$1$ph = 0, $pixel$160 = 0, $pixel10$1$ph = 0, $pixel10$166 = 0;
 var $pixel17$1$ph = 0, $pixel17$158 = 0, $pixel24$1$ph = 0, $pixel24$172 = 0, $pixel3$1$ph = 0, $pixel3$163 = 0, $pixel31$1$ph = 0, $pixel31$175 = 0, $pixel38$1$ph = 0, $pixel38$178 = 0, $pixel45$1$ph = 0, $pixel45$169 = 0, $pixel52$1 = 0, $pixel52$284 = 0, $pixel59$1 = 0, $pixel59$287 = 0, $pixel66$1 = 0, $pixel66$290 = 0, $pixel73$1 = 0, $pixel73$281 = 0;
 var $r$0 = 0, $x$0100 = 0, $x$1 = 0, $x108$0110 = 0, $x108$1 = 0, $x126$095 = 0, $x126$1 = 0, $x90$0105 = 0, $x90$1 = 0, $xinc1$0 = 0, $xinc1109$0 = 0, $xinc1127$0 = 0, $xinc191$0 = 0, $y$0101 = 0, $y$1 = 0, $y111$0111 = 0, $y111$1 = 0, $y129$096 = 0, $y129$1 = 0, $y93$0106 = 0;
 var $y93$1 = 0, $yinc1$0 = 0, $yinc1$1 = 0, $yinc1$1$yinc2$1 = 0, $yinc1112$0 = 0, $yinc1112$1 = 0, $yinc1112$1$yinc2113$1 = 0, $yinc1130$0 = 0, $yinc1130$1 = 0, $yinc1130$1$yinc2131$1 = 0, $yinc194$0 = 0, $yinc194$1 = 0, $yinc194$1$yinc295$1 = 0, $yinc2$1 = 0, $yinc2113$1 = 0, $yinc2131$1 = 0, $yinc295$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($dst)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $blendMode$off = (($blendMode) + -1)|0;
 $2 = ($blendMode$off>>>0)<(2);
 $3 = $_r&255;
 if ($2) {
  $4 = $_a&255;
  $5 = Math_imul($4, $3)|0;
  $6 = (($5>>>0) / 255)&-1;
  $7 = $_g&255;
  $8 = Math_imul($4, $7)|0;
  $9 = (($8>>>0) / 255)&-1;
  $10 = $_b&255;
  $11 = Math_imul($4, $10)|0;
  $12 = (($11>>>0) / 255)&-1;
  $a$0 = $4;$b$0 = $12;$g$0 = $9;$r$0 = $6;
 } else {
  $13 = $_g&255;
  $14 = $_b&255;
  $15 = $_a&255;
  $a$0 = $15;$b$0 = $14;$g$0 = $13;$r$0 = $3;
 }
 $16 = $a$0 ^ 255;
 $17 = ($y2|0)==($y1|0);
 if ($17) {
  switch ($blendMode|0) {
  case 1:  {
   $18 = ((($dst)) + 16|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ((($1)) + 9|0);
   $21 = HEAP8[$20>>0]|0;
   $22 = $21&255;
   $23 = (($19|0) / ($22|0))&-1;
   $24 = ($x2|0)<($x1|0);
   $25 = ((($dst)) + 20|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = Math_imul($23, $y2)|0;
   $28 = (($26) + ($27<<1)|0);
   if ($24) {
    $34 = (($28) + ($x2<<1)|0);
    $35 = ($draw_end|0)!=(0);
    $36 = ((($34)) + 2|0);
    $$13 = $35 ? $34 : $36;
    $37 = (($x1) - ($x2))|0;
    $38 = $35&1;
    $39 = (($37) + ($38))|0;
    $length$0$ph = $39;$pixel$1$ph = $$13;
   } else {
    $29 = (($28) + ($x1<<1)|0);
    $30 = ($draw_end|0)!=(0);
    $31 = (($x2) - ($x1))|0;
    $32 = $30&1;
    $33 = (($31) + ($32))|0;
    $length$0$ph = $33;$pixel$1$ph = $29;
   }
   $40 = ($length$0$ph|0)==(0);
   if ($40) {
    return;
   }
   $41 = ((($1)) + 12|0);
   $42 = ((($1)) + 32|0);
   $43 = ((($1)) + 28|0);
   $44 = ((($1)) + 16|0);
   $45 = ((($1)) + 20|0);
   $46 = ((($1)) + 24|0);
   $$pre297 = HEAP32[$41>>2]|0;
   $$pre298 = HEAP32[$42>>2]|0;
   $$pre299 = HEAP32[$43>>2]|0;
   $$phi$trans$insert300 = $$pre299 & 255;
   $$phi$trans$insert301 = (125924 + ($$phi$trans$insert300<<2)|0);
   $$pre302 = HEAP32[$$phi$trans$insert301>>2]|0;
   $$pre303 = HEAP32[$44>>2]|0;
   $47 = $$pre298 >>> 8;
   $48 = $47 & 255;
   $49 = $$pre299 >>> 8;
   $50 = $49 & 255;
   $$phi$trans$insert305 = (125924 + ($50<<2)|0);
   $$pre306 = HEAP32[$$phi$trans$insert305>>2]|0;
   $$pre307 = HEAP32[$45>>2]|0;
   $51 = $$pre298 >>> 16;
   $52 = $$pre299 >>> 16;
   $$phi$trans$insert308 = $52 & 255;
   $$phi$trans$insert309 = (125924 + ($$phi$trans$insert308<<2)|0);
   $$pre310 = HEAP32[$$phi$trans$insert309>>2]|0;
   $$pre311 = HEAP32[$46>>2]|0;
   $53 = $$pre298 & 255;
   $54 = $51 & 255;
   $length$061 = $length$0$ph;$pixel$160 = $pixel$1$ph;
   while(1) {
    $55 = (($length$061) + -1)|0;
    $56 = HEAP16[$pixel$160>>1]|0;
    $57 = $56&65535;
    $58 = $57 & $$pre297;
    $59 = $58 >>> $53;
    $60 = (($$pre302) + ($59)|0);
    $61 = HEAP8[$60>>0]|0;
    $62 = $61&255;
    $63 = $$pre303 & $57;
    $64 = $63 >>> $48;
    $65 = (($$pre306) + ($64)|0);
    $66 = HEAP8[$65>>0]|0;
    $67 = $66&255;
    $68 = $$pre307 & $57;
    $69 = $68 >>> $54;
    $70 = (($$pre310) + ($69)|0);
    $71 = HEAP8[$70>>0]|0;
    $72 = $71&255;
    $73 = Math_imul($62, $16)|0;
    $74 = (($73>>>0) / 255)&-1;
    $75 = (($74) + ($r$0))|0;
    $76 = Math_imul($67, $16)|0;
    $77 = (($76>>>0) / 255)&-1;
    $78 = (($77) + ($g$0))|0;
    $79 = Math_imul($72, $16)|0;
    $80 = (($79>>>0) / 255)&-1;
    $81 = (($80) + ($b$0))|0;
    $82 = $75 >>> $$phi$trans$insert300;
    $83 = $82 << $53;
    $84 = $78 >>> $50;
    $85 = $84 << $48;
    $86 = $85 | $83;
    $87 = $81 >>> $$phi$trans$insert308;
    $88 = $87 << $54;
    $89 = $86 | $$pre311;
    $90 = $89 | $88;
    $91 = $90&65535;
    HEAP16[$pixel$160>>1] = $91;
    $92 = ((($pixel$160)) + 2|0);
    $93 = ($55|0)==(0);
    if ($93) {
     break;
    } else {
     $length$061 = $55;$pixel$160 = $92;
    }
   }
   return;
   break;
  }
  case 2:  {
   $94 = ((($dst)) + 16|0);
   $95 = HEAP32[$94>>2]|0;
   $96 = ((($1)) + 9|0);
   $97 = HEAP8[$96>>0]|0;
   $98 = $97&255;
   $99 = (($95|0) / ($98|0))&-1;
   $100 = ($x2|0)<($x1|0);
   $101 = ((($dst)) + 20|0);
   $102 = HEAP32[$101>>2]|0;
   $103 = Math_imul($99, $y2)|0;
   $104 = (($102) + ($103<<1)|0);
   if ($100) {
    $110 = (($104) + ($x2<<1)|0);
    $111 = ($draw_end|0)!=(0);
    $112 = ((($110)) + 2|0);
    $$14 = $111 ? $110 : $112;
    $113 = (($x1) - ($x2))|0;
    $114 = $111&1;
    $115 = (($113) + ($114))|0;
    $length1$0$ph = $115;$pixel3$1$ph = $$14;
   } else {
    $105 = (($104) + ($x1<<1)|0);
    $106 = ($draw_end|0)!=(0);
    $107 = (($x2) - ($x1))|0;
    $108 = $106&1;
    $109 = (($107) + ($108))|0;
    $length1$0$ph = $109;$pixel3$1$ph = $105;
   }
   $116 = ($length1$0$ph|0)==(0);
   if ($116) {
    return;
   }
   $117 = ((($1)) + 12|0);
   $118 = ((($1)) + 32|0);
   $119 = ((($1)) + 28|0);
   $120 = ((($1)) + 16|0);
   $121 = ((($1)) + 20|0);
   $122 = ((($1)) + 24|0);
   $$pre282 = HEAP32[$117>>2]|0;
   $$pre283 = HEAP32[$118>>2]|0;
   $$pre284 = HEAP32[$119>>2]|0;
   $$phi$trans$insert285 = $$pre284 & 255;
   $$phi$trans$insert286 = (125924 + ($$phi$trans$insert285<<2)|0);
   $$pre287 = HEAP32[$$phi$trans$insert286>>2]|0;
   $$pre288 = HEAP32[$120>>2]|0;
   $123 = $$pre283 >>> 8;
   $124 = $123 & 255;
   $125 = $$pre284 >>> 8;
   $126 = $125 & 255;
   $$phi$trans$insert290 = (125924 + ($126<<2)|0);
   $$pre291 = HEAP32[$$phi$trans$insert290>>2]|0;
   $$pre292 = HEAP32[$121>>2]|0;
   $127 = $$pre283 >>> 16;
   $128 = $$pre284 >>> 16;
   $$phi$trans$insert293 = $128 & 255;
   $$phi$trans$insert294 = (125924 + ($$phi$trans$insert293<<2)|0);
   $$pre295 = HEAP32[$$phi$trans$insert294>>2]|0;
   $$pre296 = HEAP32[$122>>2]|0;
   $129 = $$pre283 & 255;
   $130 = $127 & 255;
   $length1$064 = $length1$0$ph;$pixel3$163 = $pixel3$1$ph;
   while(1) {
    $131 = (($length1$064) + -1)|0;
    $132 = HEAP16[$pixel3$163>>1]|0;
    $133 = $132&65535;
    $134 = $133 & $$pre282;
    $135 = $134 >>> $129;
    $136 = (($$pre287) + ($135)|0);
    $137 = HEAP8[$136>>0]|0;
    $138 = $137&255;
    $139 = $$pre288 & $133;
    $140 = $139 >>> $124;
    $141 = (($$pre291) + ($140)|0);
    $142 = HEAP8[$141>>0]|0;
    $143 = $142&255;
    $144 = $$pre292 & $133;
    $145 = $144 >>> $130;
    $146 = (($$pre295) + ($145)|0);
    $147 = HEAP8[$146>>0]|0;
    $148 = $147&255;
    $149 = (($138) + ($r$0))|0;
    $150 = ($149>>>0)>(255);
    $$ = $150 ? 255 : $149;
    $151 = (($143) + ($g$0))|0;
    $152 = ($151>>>0)>(255);
    $$8 = $152 ? 255 : $151;
    $153 = (($148) + ($b$0))|0;
    $154 = ($153>>>0)>(255);
    $$2 = $154 ? 255 : $153;
    $155 = $$ >>> $$phi$trans$insert285;
    $156 = $155 << $129;
    $157 = $$8 >>> $126;
    $158 = $157 << $124;
    $159 = $158 | $156;
    $160 = $$2 >>> $$phi$trans$insert293;
    $161 = $160 << $130;
    $162 = $159 | $$pre296;
    $163 = $162 | $161;
    $164 = $163&65535;
    HEAP16[$pixel3$163>>1] = $164;
    $165 = ((($pixel3$163)) + 2|0);
    $166 = ($131|0)==(0);
    if ($166) {
     break;
    } else {
     $length1$064 = $131;$pixel3$163 = $165;
    }
   }
   return;
   break;
  }
  case 4:  {
   $167 = ((($dst)) + 16|0);
   $168 = HEAP32[$167>>2]|0;
   $169 = ((($1)) + 9|0);
   $170 = HEAP8[$169>>0]|0;
   $171 = $170&255;
   $172 = (($168|0) / ($171|0))&-1;
   $173 = ($x2|0)<($x1|0);
   $174 = ((($dst)) + 20|0);
   $175 = HEAP32[$174>>2]|0;
   $176 = Math_imul($172, $y2)|0;
   $177 = (($175) + ($176<<1)|0);
   if ($173) {
    $183 = (($177) + ($x2<<1)|0);
    $184 = ($draw_end|0)!=(0);
    $185 = ((($183)) + 2|0);
    $$15 = $184 ? $183 : $185;
    $186 = (($x1) - ($x2))|0;
    $187 = $184&1;
    $188 = (($186) + ($187))|0;
    $length8$0$ph = $188;$pixel10$1$ph = $$15;
   } else {
    $178 = (($177) + ($x1<<1)|0);
    $179 = ($draw_end|0)!=(0);
    $180 = (($x2) - ($x1))|0;
    $181 = $179&1;
    $182 = (($180) + ($181))|0;
    $length8$0$ph = $182;$pixel10$1$ph = $178;
   }
   $189 = ($length8$0$ph|0)==(0);
   if ($189) {
    return;
   }
   $190 = ((($1)) + 12|0);
   $191 = ((($1)) + 32|0);
   $192 = ((($1)) + 28|0);
   $193 = ((($1)) + 16|0);
   $194 = ((($1)) + 20|0);
   $195 = ((($1)) + 24|0);
   $$pre267 = HEAP32[$190>>2]|0;
   $$pre268 = HEAP32[$191>>2]|0;
   $$pre269 = HEAP32[$192>>2]|0;
   $$phi$trans$insert270 = $$pre269 & 255;
   $$phi$trans$insert271 = (125924 + ($$phi$trans$insert270<<2)|0);
   $$pre272 = HEAP32[$$phi$trans$insert271>>2]|0;
   $$pre273 = HEAP32[$193>>2]|0;
   $196 = $$pre268 >>> 8;
   $197 = $196 & 255;
   $198 = $$pre269 >>> 8;
   $199 = $198 & 255;
   $$phi$trans$insert275 = (125924 + ($199<<2)|0);
   $$pre276 = HEAP32[$$phi$trans$insert275>>2]|0;
   $$pre277 = HEAP32[$194>>2]|0;
   $200 = $$pre268 >>> 16;
   $201 = $$pre269 >>> 16;
   $$phi$trans$insert278 = $201 & 255;
   $$phi$trans$insert279 = (125924 + ($$phi$trans$insert278<<2)|0);
   $$pre280 = HEAP32[$$phi$trans$insert279>>2]|0;
   $$pre281 = HEAP32[$195>>2]|0;
   $202 = $$pre268 & 255;
   $203 = $200 & 255;
   $length8$067 = $length8$0$ph;$pixel10$166 = $pixel10$1$ph;
   while(1) {
    $204 = (($length8$067) + -1)|0;
    $205 = HEAP16[$pixel10$166>>1]|0;
    $206 = $205&65535;
    $207 = $206 & $$pre267;
    $208 = $207 >>> $202;
    $209 = (($$pre272) + ($208)|0);
    $210 = HEAP8[$209>>0]|0;
    $211 = $210&255;
    $212 = $$pre273 & $206;
    $213 = $212 >>> $197;
    $214 = (($$pre276) + ($213)|0);
    $215 = HEAP8[$214>>0]|0;
    $216 = $215&255;
    $217 = $$pre277 & $206;
    $218 = $217 >>> $203;
    $219 = (($$pre280) + ($218)|0);
    $220 = HEAP8[$219>>0]|0;
    $221 = $220&255;
    $222 = Math_imul($211, $r$0)|0;
    $223 = (($222>>>0) / 255)&-1;
    $224 = Math_imul($216, $g$0)|0;
    $225 = (($224>>>0) / 255)&-1;
    $226 = Math_imul($221, $b$0)|0;
    $227 = (($226>>>0) / 255)&-1;
    $228 = $223 >>> $$phi$trans$insert270;
    $229 = $228 << $202;
    $230 = $225 >>> $199;
    $231 = $230 << $197;
    $232 = $231 | $229;
    $233 = $227 >>> $$phi$trans$insert278;
    $234 = $233 << $203;
    $235 = $232 | $$pre281;
    $236 = $235 | $234;
    $237 = $236&65535;
    HEAP16[$pixel10$166>>1] = $237;
    $238 = ((($pixel10$166)) + 2|0);
    $239 = ($204|0)==(0);
    if ($239) {
     break;
    } else {
     $length8$067 = $204;$pixel10$166 = $238;
    }
   }
   return;
   break;
  }
  default: {
   $240 = ((($dst)) + 16|0);
   $241 = HEAP32[$240>>2]|0;
   $242 = ((($1)) + 9|0);
   $243 = HEAP8[$242>>0]|0;
   $244 = $243&255;
   $245 = (($241|0) / ($244|0))&-1;
   $246 = ($x2|0)<($x1|0);
   $247 = ((($dst)) + 20|0);
   $248 = HEAP32[$247>>2]|0;
   $249 = Math_imul($245, $y2)|0;
   $250 = (($248) + ($249<<1)|0);
   if ($246) {
    $256 = (($250) + ($x2<<1)|0);
    $257 = ($draw_end|0)!=(0);
    $258 = ((($256)) + 2|0);
    $$16 = $257 ? $256 : $258;
    $259 = (($x1) - ($x2))|0;
    $260 = $257&1;
    $261 = (($259) + ($260))|0;
    $length15$0$ph = $261;$pixel17$1$ph = $$16;
   } else {
    $251 = (($250) + ($x1<<1)|0);
    $252 = ($draw_end|0)!=(0);
    $253 = (($x2) - ($x1))|0;
    $254 = $252&1;
    $255 = (($253) + ($254))|0;
    $length15$0$ph = $255;$pixel17$1$ph = $251;
   }
   $262 = ($length15$0$ph|0)==(0);
   if ($262) {
    return;
   }
   $263 = ((($1)) + 28|0);
   $264 = HEAP32[$263>>2]|0;
   $265 = $264 & 255;
   $266 = $r$0 >>> $265;
   $267 = ((($1)) + 32|0);
   $268 = HEAP32[$267>>2]|0;
   $269 = $268 & 255;
   $270 = $266 << $269;
   $271 = $264 >>> 8;
   $272 = $271 & 255;
   $273 = $g$0 >>> $272;
   $274 = $268 >>> 8;
   $275 = $274 & 255;
   $276 = $273 << $275;
   $277 = $276 | $270;
   $278 = $264 >>> 16;
   $279 = $278 & 255;
   $280 = $b$0 >>> $279;
   $281 = $268 >>> 16;
   $282 = $281 & 255;
   $283 = $280 << $282;
   $284 = ((($1)) + 24|0);
   $285 = HEAP32[$284>>2]|0;
   $286 = $277 | $285;
   $287 = $286 | $283;
   $288 = $287&65535;
   $length15$059 = $length15$0$ph;$pixel17$158 = $pixel17$1$ph;
   while(1) {
    $289 = (($length15$059) + -1)|0;
    HEAP16[$pixel17$158>>1] = $288;
    $290 = ((($pixel17$158)) + 2|0);
    $291 = ($289|0)==(0);
    if ($291) {
     break;
    } else {
     $length15$059 = $289;$pixel17$158 = $290;
    }
   }
   return;
  }
  }
 }
 $292 = ($x2|0)==($x1|0);
 if ($292) {
  switch ($blendMode|0) {
  case 1:  {
   $293 = ((($dst)) + 16|0);
   $294 = HEAP32[$293>>2]|0;
   $295 = ((($1)) + 9|0);
   $296 = HEAP8[$295>>0]|0;
   $297 = $296&255;
   $298 = (($294|0) / ($297|0))&-1;
   $299 = ($y2|0)<($y1|0);
   $300 = ((($dst)) + 20|0);
   $301 = HEAP32[$300>>2]|0;
   if ($299) {
    $309 = Math_imul($298, $y2)|0;
    $310 = (($301) + ($309<<1)|0);
    $311 = (($310) + ($x2<<1)|0);
    $312 = ($draw_end|0)!=(0);
    $313 = (($311) + ($298<<1)|0);
    $$17 = $312 ? $311 : $313;
    $314 = (($y1) - ($y2))|0;
    $315 = $312&1;
    $316 = (($314) + ($315))|0;
    $length22$0$ph = $316;$pixel24$1$ph = $$17;
   } else {
    $302 = Math_imul($298, $y1)|0;
    $303 = (($301) + ($302<<1)|0);
    $304 = (($303) + ($x2<<1)|0);
    $305 = ($draw_end|0)!=(0);
    $306 = (($y2) - ($y1))|0;
    $307 = $305&1;
    $308 = (($306) + ($307))|0;
    $length22$0$ph = $308;$pixel24$1$ph = $304;
   }
   $317 = ($length22$0$ph|0)==(0);
   if ($317) {
    return;
   }
   $318 = ((($1)) + 12|0);
   $319 = ((($1)) + 32|0);
   $320 = ((($1)) + 28|0);
   $321 = ((($1)) + 16|0);
   $322 = ((($1)) + 20|0);
   $323 = ((($1)) + 24|0);
   $$pre252 = HEAP32[$318>>2]|0;
   $$pre253 = HEAP32[$319>>2]|0;
   $$pre254 = HEAP32[$320>>2]|0;
   $$phi$trans$insert255 = $$pre254 & 255;
   $$phi$trans$insert256 = (125924 + ($$phi$trans$insert255<<2)|0);
   $$pre257 = HEAP32[$$phi$trans$insert256>>2]|0;
   $$pre258 = HEAP32[$321>>2]|0;
   $324 = $$pre253 >>> 8;
   $325 = $324 & 255;
   $326 = $$pre254 >>> 8;
   $327 = $326 & 255;
   $$phi$trans$insert260 = (125924 + ($327<<2)|0);
   $$pre261 = HEAP32[$$phi$trans$insert260>>2]|0;
   $$pre262 = HEAP32[$322>>2]|0;
   $328 = $$pre253 >>> 16;
   $329 = $$pre254 >>> 16;
   $$phi$trans$insert263 = $329 & 255;
   $$phi$trans$insert264 = (125924 + ($$phi$trans$insert263<<2)|0);
   $$pre265 = HEAP32[$$phi$trans$insert264>>2]|0;
   $$pre266 = HEAP32[$323>>2]|0;
   $330 = $$pre253 & 255;
   $331 = $328 & 255;
   $length22$073 = $length22$0$ph;$pixel24$172 = $pixel24$1$ph;
   while(1) {
    $332 = (($length22$073) + -1)|0;
    $333 = HEAP16[$pixel24$172>>1]|0;
    $334 = $333&65535;
    $335 = $334 & $$pre252;
    $336 = $335 >>> $330;
    $337 = (($$pre257) + ($336)|0);
    $338 = HEAP8[$337>>0]|0;
    $339 = $338&255;
    $340 = $$pre258 & $334;
    $341 = $340 >>> $325;
    $342 = (($$pre261) + ($341)|0);
    $343 = HEAP8[$342>>0]|0;
    $344 = $343&255;
    $345 = $$pre262 & $334;
    $346 = $345 >>> $331;
    $347 = (($$pre265) + ($346)|0);
    $348 = HEAP8[$347>>0]|0;
    $349 = $348&255;
    $350 = Math_imul($339, $16)|0;
    $351 = (($350>>>0) / 255)&-1;
    $352 = (($351) + ($r$0))|0;
    $353 = Math_imul($344, $16)|0;
    $354 = (($353>>>0) / 255)&-1;
    $355 = (($354) + ($g$0))|0;
    $356 = Math_imul($349, $16)|0;
    $357 = (($356>>>0) / 255)&-1;
    $358 = (($357) + ($b$0))|0;
    $359 = $352 >>> $$phi$trans$insert255;
    $360 = $359 << $330;
    $361 = $355 >>> $327;
    $362 = $361 << $325;
    $363 = $362 | $360;
    $364 = $358 >>> $$phi$trans$insert263;
    $365 = $364 << $331;
    $366 = $363 | $$pre266;
    $367 = $366 | $365;
    $368 = $367&65535;
    HEAP16[$pixel24$172>>1] = $368;
    $369 = (($pixel24$172) + ($298<<1)|0);
    $370 = ($332|0)==(0);
    if ($370) {
     break;
    } else {
     $length22$073 = $332;$pixel24$172 = $369;
    }
   }
   return;
   break;
  }
  case 2:  {
   $371 = ((($dst)) + 16|0);
   $372 = HEAP32[$371>>2]|0;
   $373 = ((($1)) + 9|0);
   $374 = HEAP8[$373>>0]|0;
   $375 = $374&255;
   $376 = (($372|0) / ($375|0))&-1;
   $377 = ($y2|0)<($y1|0);
   $378 = ((($dst)) + 20|0);
   $379 = HEAP32[$378>>2]|0;
   if ($377) {
    $387 = Math_imul($376, $y2)|0;
    $388 = (($379) + ($387<<1)|0);
    $389 = (($388) + ($x2<<1)|0);
    $390 = ($draw_end|0)!=(0);
    $391 = (($389) + ($376<<1)|0);
    $$18 = $390 ? $389 : $391;
    $392 = (($y1) - ($y2))|0;
    $393 = $390&1;
    $394 = (($392) + ($393))|0;
    $length29$0$ph = $394;$pixel31$1$ph = $$18;
   } else {
    $380 = Math_imul($376, $y1)|0;
    $381 = (($379) + ($380<<1)|0);
    $382 = (($381) + ($x2<<1)|0);
    $383 = ($draw_end|0)!=(0);
    $384 = (($y2) - ($y1))|0;
    $385 = $383&1;
    $386 = (($384) + ($385))|0;
    $length29$0$ph = $386;$pixel31$1$ph = $382;
   }
   $395 = ($length29$0$ph|0)==(0);
   if ($395) {
    return;
   }
   $396 = ((($1)) + 12|0);
   $397 = ((($1)) + 32|0);
   $398 = ((($1)) + 28|0);
   $399 = ((($1)) + 16|0);
   $400 = ((($1)) + 20|0);
   $401 = ((($1)) + 24|0);
   $$pre237 = HEAP32[$396>>2]|0;
   $$pre238 = HEAP32[$397>>2]|0;
   $$pre239 = HEAP32[$398>>2]|0;
   $$phi$trans$insert240 = $$pre239 & 255;
   $$phi$trans$insert241 = (125924 + ($$phi$trans$insert240<<2)|0);
   $$pre242 = HEAP32[$$phi$trans$insert241>>2]|0;
   $$pre243 = HEAP32[$399>>2]|0;
   $402 = $$pre238 >>> 8;
   $403 = $402 & 255;
   $404 = $$pre239 >>> 8;
   $405 = $404 & 255;
   $$phi$trans$insert245 = (125924 + ($405<<2)|0);
   $$pre246 = HEAP32[$$phi$trans$insert245>>2]|0;
   $$pre247 = HEAP32[$400>>2]|0;
   $406 = $$pre238 >>> 16;
   $407 = $$pre239 >>> 16;
   $$phi$trans$insert248 = $407 & 255;
   $$phi$trans$insert249 = (125924 + ($$phi$trans$insert248<<2)|0);
   $$pre250 = HEAP32[$$phi$trans$insert249>>2]|0;
   $$pre251 = HEAP32[$401>>2]|0;
   $408 = $$pre238 & 255;
   $409 = $406 & 255;
   $length29$076 = $length29$0$ph;$pixel31$175 = $pixel31$1$ph;
   while(1) {
    $410 = (($length29$076) + -1)|0;
    $411 = HEAP16[$pixel31$175>>1]|0;
    $412 = $411&65535;
    $413 = $412 & $$pre237;
    $414 = $413 >>> $408;
    $415 = (($$pre242) + ($414)|0);
    $416 = HEAP8[$415>>0]|0;
    $417 = $416&255;
    $418 = $$pre243 & $412;
    $419 = $418 >>> $403;
    $420 = (($$pre246) + ($419)|0);
    $421 = HEAP8[$420>>0]|0;
    $422 = $421&255;
    $423 = $$pre247 & $412;
    $424 = $423 >>> $409;
    $425 = (($$pre250) + ($424)|0);
    $426 = HEAP8[$425>>0]|0;
    $427 = $426&255;
    $428 = (($417) + ($r$0))|0;
    $429 = ($428>>>0)>(255);
    $$3 = $429 ? 255 : $428;
    $430 = (($422) + ($g$0))|0;
    $431 = ($430>>>0)>(255);
    $$9 = $431 ? 255 : $430;
    $432 = (($427) + ($b$0))|0;
    $433 = ($432>>>0)>(255);
    $$4 = $433 ? 255 : $432;
    $434 = $$3 >>> $$phi$trans$insert240;
    $435 = $434 << $408;
    $436 = $$9 >>> $405;
    $437 = $436 << $403;
    $438 = $437 | $435;
    $439 = $$4 >>> $$phi$trans$insert248;
    $440 = $439 << $409;
    $441 = $438 | $$pre251;
    $442 = $441 | $440;
    $443 = $442&65535;
    HEAP16[$pixel31$175>>1] = $443;
    $444 = (($pixel31$175) + ($376<<1)|0);
    $445 = ($410|0)==(0);
    if ($445) {
     break;
    } else {
     $length29$076 = $410;$pixel31$175 = $444;
    }
   }
   return;
   break;
  }
  case 4:  {
   $446 = ((($dst)) + 16|0);
   $447 = HEAP32[$446>>2]|0;
   $448 = ((($1)) + 9|0);
   $449 = HEAP8[$448>>0]|0;
   $450 = $449&255;
   $451 = (($447|0) / ($450|0))&-1;
   $452 = ($y2|0)<($y1|0);
   $453 = ((($dst)) + 20|0);
   $454 = HEAP32[$453>>2]|0;
   if ($452) {
    $462 = Math_imul($451, $y2)|0;
    $463 = (($454) + ($462<<1)|0);
    $464 = (($463) + ($x2<<1)|0);
    $465 = ($draw_end|0)!=(0);
    $466 = (($464) + ($451<<1)|0);
    $$19 = $465 ? $464 : $466;
    $467 = (($y1) - ($y2))|0;
    $468 = $465&1;
    $469 = (($467) + ($468))|0;
    $length36$0$ph = $469;$pixel38$1$ph = $$19;
   } else {
    $455 = Math_imul($451, $y1)|0;
    $456 = (($454) + ($455<<1)|0);
    $457 = (($456) + ($x2<<1)|0);
    $458 = ($draw_end|0)!=(0);
    $459 = (($y2) - ($y1))|0;
    $460 = $458&1;
    $461 = (($459) + ($460))|0;
    $length36$0$ph = $461;$pixel38$1$ph = $457;
   }
   $470 = ($length36$0$ph|0)==(0);
   if ($470) {
    return;
   }
   $471 = ((($1)) + 12|0);
   $472 = ((($1)) + 32|0);
   $473 = ((($1)) + 28|0);
   $474 = ((($1)) + 16|0);
   $475 = ((($1)) + 20|0);
   $476 = ((($1)) + 24|0);
   $$pre222 = HEAP32[$471>>2]|0;
   $$pre223 = HEAP32[$472>>2]|0;
   $$pre224 = HEAP32[$473>>2]|0;
   $$phi$trans$insert225 = $$pre224 & 255;
   $$phi$trans$insert226 = (125924 + ($$phi$trans$insert225<<2)|0);
   $$pre227 = HEAP32[$$phi$trans$insert226>>2]|0;
   $$pre228 = HEAP32[$474>>2]|0;
   $477 = $$pre223 >>> 8;
   $478 = $477 & 255;
   $479 = $$pre224 >>> 8;
   $480 = $479 & 255;
   $$phi$trans$insert230 = (125924 + ($480<<2)|0);
   $$pre231 = HEAP32[$$phi$trans$insert230>>2]|0;
   $$pre232 = HEAP32[$475>>2]|0;
   $481 = $$pre223 >>> 16;
   $482 = $$pre224 >>> 16;
   $$phi$trans$insert233 = $482 & 255;
   $$phi$trans$insert234 = (125924 + ($$phi$trans$insert233<<2)|0);
   $$pre235 = HEAP32[$$phi$trans$insert234>>2]|0;
   $$pre236 = HEAP32[$476>>2]|0;
   $483 = $$pre223 & 255;
   $484 = $481 & 255;
   $length36$079 = $length36$0$ph;$pixel38$178 = $pixel38$1$ph;
   while(1) {
    $485 = (($length36$079) + -1)|0;
    $486 = HEAP16[$pixel38$178>>1]|0;
    $487 = $486&65535;
    $488 = $487 & $$pre222;
    $489 = $488 >>> $483;
    $490 = (($$pre227) + ($489)|0);
    $491 = HEAP8[$490>>0]|0;
    $492 = $491&255;
    $493 = $$pre228 & $487;
    $494 = $493 >>> $478;
    $495 = (($$pre231) + ($494)|0);
    $496 = HEAP8[$495>>0]|0;
    $497 = $496&255;
    $498 = $$pre232 & $487;
    $499 = $498 >>> $484;
    $500 = (($$pre235) + ($499)|0);
    $501 = HEAP8[$500>>0]|0;
    $502 = $501&255;
    $503 = Math_imul($492, $r$0)|0;
    $504 = (($503>>>0) / 255)&-1;
    $505 = Math_imul($497, $g$0)|0;
    $506 = (($505>>>0) / 255)&-1;
    $507 = Math_imul($502, $b$0)|0;
    $508 = (($507>>>0) / 255)&-1;
    $509 = $504 >>> $$phi$trans$insert225;
    $510 = $509 << $483;
    $511 = $506 >>> $480;
    $512 = $511 << $478;
    $513 = $512 | $510;
    $514 = $508 >>> $$phi$trans$insert233;
    $515 = $514 << $484;
    $516 = $513 | $$pre236;
    $517 = $516 | $515;
    $518 = $517&65535;
    HEAP16[$pixel38$178>>1] = $518;
    $519 = (($pixel38$178) + ($451<<1)|0);
    $520 = ($485|0)==(0);
    if ($520) {
     break;
    } else {
     $length36$079 = $485;$pixel38$178 = $519;
    }
   }
   return;
   break;
  }
  default: {
   $521 = ((($dst)) + 16|0);
   $522 = HEAP32[$521>>2]|0;
   $523 = ((($1)) + 9|0);
   $524 = HEAP8[$523>>0]|0;
   $525 = $524&255;
   $526 = (($522|0) / ($525|0))&-1;
   $527 = ($y2|0)<($y1|0);
   $528 = ((($dst)) + 20|0);
   $529 = HEAP32[$528>>2]|0;
   if ($527) {
    $537 = Math_imul($526, $y2)|0;
    $538 = (($529) + ($537<<1)|0);
    $539 = (($538) + ($x2<<1)|0);
    $540 = ($draw_end|0)!=(0);
    $541 = (($539) + ($526<<1)|0);
    $$20 = $540 ? $539 : $541;
    $542 = (($y1) - ($y2))|0;
    $543 = $540&1;
    $544 = (($542) + ($543))|0;
    $length43$0$ph = $544;$pixel45$1$ph = $$20;
   } else {
    $530 = Math_imul($526, $y1)|0;
    $531 = (($529) + ($530<<1)|0);
    $532 = (($531) + ($x2<<1)|0);
    $533 = ($draw_end|0)!=(0);
    $534 = (($y2) - ($y1))|0;
    $535 = $533&1;
    $536 = (($534) + ($535))|0;
    $length43$0$ph = $536;$pixel45$1$ph = $532;
   }
   $545 = ($length43$0$ph|0)==(0);
   if ($545) {
    return;
   }
   $546 = ((($1)) + 28|0);
   $547 = HEAP32[$546>>2]|0;
   $548 = $547 & 255;
   $549 = $r$0 >>> $548;
   $550 = ((($1)) + 32|0);
   $551 = HEAP32[$550>>2]|0;
   $552 = $551 & 255;
   $553 = $549 << $552;
   $554 = $547 >>> 8;
   $555 = $554 & 255;
   $556 = $g$0 >>> $555;
   $557 = $551 >>> 8;
   $558 = $557 & 255;
   $559 = $556 << $558;
   $560 = $559 | $553;
   $561 = $547 >>> 16;
   $562 = $561 & 255;
   $563 = $b$0 >>> $562;
   $564 = $551 >>> 16;
   $565 = $564 & 255;
   $566 = $563 << $565;
   $567 = ((($1)) + 24|0);
   $568 = HEAP32[$567>>2]|0;
   $569 = $560 | $568;
   $570 = $569 | $566;
   $571 = $570&65535;
   $length43$070 = $length43$0$ph;$pixel45$169 = $pixel45$1$ph;
   while(1) {
    $572 = (($length43$070) + -1)|0;
    HEAP16[$pixel45$169>>1] = $571;
    $573 = (($pixel45$169) + ($526<<1)|0);
    $574 = ($572|0)==(0);
    if ($574) {
     break;
    } else {
     $length43$070 = $572;$pixel45$169 = $573;
    }
   }
   return;
  }
  }
 }
 $575 = (($x1) - ($x2))|0;
 $576 = ($575|0)<(0);
 $577 = (0 - ($575))|0;
 $578 = $576 ? $577 : $575;
 $579 = (($y1) - ($y2))|0;
 $580 = ($579|0)<(0);
 $581 = (0 - ($579))|0;
 $582 = $580 ? $581 : $579;
 $583 = ($578|0)==($582|0);
 if ($583) {
  switch ($blendMode|0) {
  case 1:  {
   $584 = ((($dst)) + 16|0);
   $585 = HEAP32[$584>>2]|0;
   $586 = ((($1)) + 9|0);
   $587 = HEAP8[$586>>0]|0;
   $588 = $587&255;
   $589 = (($585|0) / ($588|0))&-1;
   $590 = ($y2|0)<($y1|0);
   $591 = ((($dst)) + 20|0);
   $592 = HEAP32[$591>>2]|0;
   if ($590) {
    $598 = Math_imul($589, $y2)|0;
    $599 = (($592) + ($598<<1)|0);
    $600 = (($599) + ($x2<<1)|0);
    $601 = ($x2|0)>($x1|0);
    $pitch51$1$v = $601 ? -1 : 1;
    $pitch51$1 = (($589) + ($pitch51$1$v))|0;
    $602 = ($draw_end|0)==(0);
    $603 = (($600) + ($pitch51$1<<1)|0);
    $$21 = $602 ? $603 : $600;
    $length50$0 = $579;$pitch51$2 = $pitch51$1;$pixel52$1 = $$21;
   } else {
    $593 = Math_imul($589, $y1)|0;
    $594 = (($592) + ($593<<1)|0);
    $595 = (($594) + ($x1<<1)|0);
    $596 = ($x2|0)<($x1|0);
    $pitch51$0$v = $596 ? -1 : 1;
    $pitch51$0 = (($589) + ($pitch51$0$v))|0;
    $597 = (($y2) - ($y1))|0;
    $length50$0 = $597;$pitch51$2 = $pitch51$0;$pixel52$1 = $595;
   }
   $not$34 = ($draw_end|0)!=(0);
   $604 = $not$34&1;
   $length50$0$ = (($length50$0) + ($604))|0;
   $605 = ($length50$0$|0)==(0);
   if ($605) {
    return;
   }
   $606 = ((($1)) + 12|0);
   $607 = ((($1)) + 32|0);
   $608 = ((($1)) + 28|0);
   $609 = ((($1)) + 16|0);
   $610 = ((($1)) + 20|0);
   $611 = ((($1)) + 24|0);
   $$pre207 = HEAP32[$606>>2]|0;
   $$pre208 = HEAP32[$607>>2]|0;
   $$pre209 = HEAP32[$608>>2]|0;
   $$phi$trans$insert210 = $$pre209 & 255;
   $$phi$trans$insert211 = (125924 + ($$phi$trans$insert210<<2)|0);
   $$pre212 = HEAP32[$$phi$trans$insert211>>2]|0;
   $$pre213 = HEAP32[$609>>2]|0;
   $612 = $$pre208 >>> 8;
   $613 = $612 & 255;
   $614 = $$pre209 >>> 8;
   $615 = $614 & 255;
   $$phi$trans$insert215 = (125924 + ($615<<2)|0);
   $$pre216 = HEAP32[$$phi$trans$insert215>>2]|0;
   $$pre217 = HEAP32[$610>>2]|0;
   $616 = $$pre208 >>> 16;
   $617 = $$pre209 >>> 16;
   $$phi$trans$insert218 = $617 & 255;
   $$phi$trans$insert219 = (125924 + ($$phi$trans$insert218<<2)|0);
   $$pre220 = HEAP32[$$phi$trans$insert219>>2]|0;
   $$pre221 = HEAP32[$611>>2]|0;
   $618 = $$pre208 & 255;
   $619 = $616 & 255;
   $length50$185 = $length50$0$;$pixel52$284 = $pixel52$1;
   while(1) {
    $620 = (($length50$185) + -1)|0;
    $621 = HEAP16[$pixel52$284>>1]|0;
    $622 = $621&65535;
    $623 = $622 & $$pre207;
    $624 = $623 >>> $618;
    $625 = (($$pre212) + ($624)|0);
    $626 = HEAP8[$625>>0]|0;
    $627 = $626&255;
    $628 = $$pre213 & $622;
    $629 = $628 >>> $613;
    $630 = (($$pre216) + ($629)|0);
    $631 = HEAP8[$630>>0]|0;
    $632 = $631&255;
    $633 = $$pre217 & $622;
    $634 = $633 >>> $619;
    $635 = (($$pre220) + ($634)|0);
    $636 = HEAP8[$635>>0]|0;
    $637 = $636&255;
    $638 = Math_imul($627, $16)|0;
    $639 = (($638>>>0) / 255)&-1;
    $640 = (($639) + ($r$0))|0;
    $641 = Math_imul($632, $16)|0;
    $642 = (($641>>>0) / 255)&-1;
    $643 = (($642) + ($g$0))|0;
    $644 = Math_imul($637, $16)|0;
    $645 = (($644>>>0) / 255)&-1;
    $646 = (($645) + ($b$0))|0;
    $647 = $640 >>> $$phi$trans$insert210;
    $648 = $647 << $618;
    $649 = $643 >>> $615;
    $650 = $649 << $613;
    $651 = $650 | $648;
    $652 = $646 >>> $$phi$trans$insert218;
    $653 = $652 << $619;
    $654 = $651 | $$pre221;
    $655 = $654 | $653;
    $656 = $655&65535;
    HEAP16[$pixel52$284>>1] = $656;
    $657 = (($pixel52$284) + ($pitch51$2<<1)|0);
    $658 = ($620|0)==(0);
    if ($658) {
     break;
    } else {
     $length50$185 = $620;$pixel52$284 = $657;
    }
   }
   return;
   break;
  }
  case 2:  {
   $659 = ((($dst)) + 16|0);
   $660 = HEAP32[$659>>2]|0;
   $661 = ((($1)) + 9|0);
   $662 = HEAP8[$661>>0]|0;
   $663 = $662&255;
   $664 = (($660|0) / ($663|0))&-1;
   $665 = ($y2|0)<($y1|0);
   $666 = ((($dst)) + 20|0);
   $667 = HEAP32[$666>>2]|0;
   if ($665) {
    $673 = Math_imul($664, $y2)|0;
    $674 = (($667) + ($673<<1)|0);
    $675 = (($674) + ($x2<<1)|0);
    $676 = ($x2|0)>($x1|0);
    $pitch58$1$v = $676 ? -1 : 1;
    $pitch58$1 = (($664) + ($pitch58$1$v))|0;
    $677 = ($draw_end|0)==(0);
    $678 = (($675) + ($pitch58$1<<1)|0);
    $$22 = $677 ? $678 : $675;
    $length57$0 = $579;$pitch58$2 = $pitch58$1;$pixel59$1 = $$22;
   } else {
    $668 = Math_imul($664, $y1)|0;
    $669 = (($667) + ($668<<1)|0);
    $670 = (($669) + ($x1<<1)|0);
    $671 = ($x2|0)<($x1|0);
    $pitch58$0$v = $671 ? -1 : 1;
    $pitch58$0 = (($664) + ($pitch58$0$v))|0;
    $672 = (($y2) - ($y1))|0;
    $length57$0 = $672;$pitch58$2 = $pitch58$0;$pixel59$1 = $670;
   }
   $not$33 = ($draw_end|0)!=(0);
   $679 = $not$33&1;
   $length57$0$ = (($length57$0) + ($679))|0;
   $680 = ($length57$0$|0)==(0);
   if ($680) {
    return;
   }
   $681 = ((($1)) + 12|0);
   $682 = ((($1)) + 32|0);
   $683 = ((($1)) + 28|0);
   $684 = ((($1)) + 16|0);
   $685 = ((($1)) + 20|0);
   $686 = ((($1)) + 24|0);
   $$pre192 = HEAP32[$681>>2]|0;
   $$pre193 = HEAP32[$682>>2]|0;
   $$pre194 = HEAP32[$683>>2]|0;
   $$phi$trans$insert195 = $$pre194 & 255;
   $$phi$trans$insert196 = (125924 + ($$phi$trans$insert195<<2)|0);
   $$pre197 = HEAP32[$$phi$trans$insert196>>2]|0;
   $$pre198 = HEAP32[$684>>2]|0;
   $687 = $$pre193 >>> 8;
   $688 = $687 & 255;
   $689 = $$pre194 >>> 8;
   $690 = $689 & 255;
   $$phi$trans$insert200 = (125924 + ($690<<2)|0);
   $$pre201 = HEAP32[$$phi$trans$insert200>>2]|0;
   $$pre202 = HEAP32[$685>>2]|0;
   $691 = $$pre193 >>> 16;
   $692 = $$pre194 >>> 16;
   $$phi$trans$insert203 = $692 & 255;
   $$phi$trans$insert204 = (125924 + ($$phi$trans$insert203<<2)|0);
   $$pre205 = HEAP32[$$phi$trans$insert204>>2]|0;
   $$pre206 = HEAP32[$686>>2]|0;
   $693 = $$pre193 & 255;
   $694 = $691 & 255;
   $length57$188 = $length57$0$;$pixel59$287 = $pixel59$1;
   while(1) {
    $695 = (($length57$188) + -1)|0;
    $696 = HEAP16[$pixel59$287>>1]|0;
    $697 = $696&65535;
    $698 = $697 & $$pre192;
    $699 = $698 >>> $693;
    $700 = (($$pre197) + ($699)|0);
    $701 = HEAP8[$700>>0]|0;
    $702 = $701&255;
    $703 = $$pre198 & $697;
    $704 = $703 >>> $688;
    $705 = (($$pre201) + ($704)|0);
    $706 = HEAP8[$705>>0]|0;
    $707 = $706&255;
    $708 = $$pre202 & $697;
    $709 = $708 >>> $694;
    $710 = (($$pre205) + ($709)|0);
    $711 = HEAP8[$710>>0]|0;
    $712 = $711&255;
    $713 = (($702) + ($r$0))|0;
    $714 = ($713>>>0)>(255);
    $$5 = $714 ? 255 : $713;
    $715 = (($707) + ($g$0))|0;
    $716 = ($715>>>0)>(255);
    $$10 = $716 ? 255 : $715;
    $717 = (($712) + ($b$0))|0;
    $718 = ($717>>>0)>(255);
    $$6 = $718 ? 255 : $717;
    $719 = $$5 >>> $$phi$trans$insert195;
    $720 = $719 << $693;
    $721 = $$10 >>> $690;
    $722 = $721 << $688;
    $723 = $722 | $720;
    $724 = $$6 >>> $$phi$trans$insert203;
    $725 = $724 << $694;
    $726 = $723 | $$pre206;
    $727 = $726 | $725;
    $728 = $727&65535;
    HEAP16[$pixel59$287>>1] = $728;
    $729 = (($pixel59$287) + ($pitch58$2<<1)|0);
    $730 = ($695|0)==(0);
    if ($730) {
     break;
    } else {
     $length57$188 = $695;$pixel59$287 = $729;
    }
   }
   return;
   break;
  }
  case 4:  {
   $731 = ((($dst)) + 16|0);
   $732 = HEAP32[$731>>2]|0;
   $733 = ((($1)) + 9|0);
   $734 = HEAP8[$733>>0]|0;
   $735 = $734&255;
   $736 = (($732|0) / ($735|0))&-1;
   $737 = ($y2|0)<($y1|0);
   $738 = ((($dst)) + 20|0);
   $739 = HEAP32[$738>>2]|0;
   if ($737) {
    $745 = Math_imul($736, $y2)|0;
    $746 = (($739) + ($745<<1)|0);
    $747 = (($746) + ($x2<<1)|0);
    $748 = ($x2|0)>($x1|0);
    $pitch65$1$v = $748 ? -1 : 1;
    $pitch65$1 = (($736) + ($pitch65$1$v))|0;
    $749 = ($draw_end|0)==(0);
    $750 = (($747) + ($pitch65$1<<1)|0);
    $$23 = $749 ? $750 : $747;
    $length64$0 = $579;$pitch65$2 = $pitch65$1;$pixel66$1 = $$23;
   } else {
    $740 = Math_imul($736, $y1)|0;
    $741 = (($739) + ($740<<1)|0);
    $742 = (($741) + ($x1<<1)|0);
    $743 = ($x2|0)<($x1|0);
    $pitch65$0$v = $743 ? -1 : 1;
    $pitch65$0 = (($736) + ($pitch65$0$v))|0;
    $744 = (($y2) - ($y1))|0;
    $length64$0 = $744;$pitch65$2 = $pitch65$0;$pixel66$1 = $742;
   }
   $not$32 = ($draw_end|0)!=(0);
   $751 = $not$32&1;
   $length64$0$ = (($length64$0) + ($751))|0;
   $752 = ($length64$0$|0)==(0);
   if ($752) {
    return;
   }
   $753 = ((($1)) + 12|0);
   $754 = ((($1)) + 32|0);
   $755 = ((($1)) + 28|0);
   $756 = ((($1)) + 16|0);
   $757 = ((($1)) + 20|0);
   $758 = ((($1)) + 24|0);
   $$pre177 = HEAP32[$753>>2]|0;
   $$pre178 = HEAP32[$754>>2]|0;
   $$pre179 = HEAP32[$755>>2]|0;
   $$phi$trans$insert180 = $$pre179 & 255;
   $$phi$trans$insert181 = (125924 + ($$phi$trans$insert180<<2)|0);
   $$pre182 = HEAP32[$$phi$trans$insert181>>2]|0;
   $$pre183 = HEAP32[$756>>2]|0;
   $759 = $$pre178 >>> 8;
   $760 = $759 & 255;
   $761 = $$pre179 >>> 8;
   $762 = $761 & 255;
   $$phi$trans$insert185 = (125924 + ($762<<2)|0);
   $$pre186 = HEAP32[$$phi$trans$insert185>>2]|0;
   $$pre187 = HEAP32[$757>>2]|0;
   $763 = $$pre178 >>> 16;
   $764 = $$pre179 >>> 16;
   $$phi$trans$insert188 = $764 & 255;
   $$phi$trans$insert189 = (125924 + ($$phi$trans$insert188<<2)|0);
   $$pre190 = HEAP32[$$phi$trans$insert189>>2]|0;
   $$pre191 = HEAP32[$758>>2]|0;
   $765 = $$pre178 & 255;
   $766 = $763 & 255;
   $length64$191 = $length64$0$;$pixel66$290 = $pixel66$1;
   while(1) {
    $767 = (($length64$191) + -1)|0;
    $768 = HEAP16[$pixel66$290>>1]|0;
    $769 = $768&65535;
    $770 = $769 & $$pre177;
    $771 = $770 >>> $765;
    $772 = (($$pre182) + ($771)|0);
    $773 = HEAP8[$772>>0]|0;
    $774 = $773&255;
    $775 = $$pre183 & $769;
    $776 = $775 >>> $760;
    $777 = (($$pre186) + ($776)|0);
    $778 = HEAP8[$777>>0]|0;
    $779 = $778&255;
    $780 = $$pre187 & $769;
    $781 = $780 >>> $766;
    $782 = (($$pre190) + ($781)|0);
    $783 = HEAP8[$782>>0]|0;
    $784 = $783&255;
    $785 = Math_imul($774, $r$0)|0;
    $786 = (($785>>>0) / 255)&-1;
    $787 = Math_imul($779, $g$0)|0;
    $788 = (($787>>>0) / 255)&-1;
    $789 = Math_imul($784, $b$0)|0;
    $790 = (($789>>>0) / 255)&-1;
    $791 = $786 >>> $$phi$trans$insert180;
    $792 = $791 << $765;
    $793 = $788 >>> $762;
    $794 = $793 << $760;
    $795 = $794 | $792;
    $796 = $790 >>> $$phi$trans$insert188;
    $797 = $796 << $766;
    $798 = $795 | $$pre191;
    $799 = $798 | $797;
    $800 = $799&65535;
    HEAP16[$pixel66$290>>1] = $800;
    $801 = (($pixel66$290) + ($pitch65$2<<1)|0);
    $802 = ($767|0)==(0);
    if ($802) {
     break;
    } else {
     $length64$191 = $767;$pixel66$290 = $801;
    }
   }
   return;
   break;
  }
  default: {
   $803 = ((($dst)) + 16|0);
   $804 = HEAP32[$803>>2]|0;
   $805 = ((($1)) + 9|0);
   $806 = HEAP8[$805>>0]|0;
   $807 = $806&255;
   $808 = (($804|0) / ($807|0))&-1;
   $809 = ($y2|0)<($y1|0);
   $810 = ((($dst)) + 20|0);
   $811 = HEAP32[$810>>2]|0;
   if ($809) {
    $817 = Math_imul($808, $y2)|0;
    $818 = (($811) + ($817<<1)|0);
    $819 = (($818) + ($x2<<1)|0);
    $820 = ($x2|0)>($x1|0);
    $pitch72$1$v = $820 ? -1 : 1;
    $pitch72$1 = (($808) + ($pitch72$1$v))|0;
    $821 = ($draw_end|0)==(0);
    $822 = (($819) + ($pitch72$1<<1)|0);
    $$24 = $821 ? $822 : $819;
    $length71$0 = $579;$pitch72$2 = $pitch72$1;$pixel73$1 = $$24;
   } else {
    $812 = Math_imul($808, $y1)|0;
    $813 = (($811) + ($812<<1)|0);
    $814 = (($813) + ($x1<<1)|0);
    $815 = ($x2|0)<($x1|0);
    $pitch72$0$v = $815 ? -1 : 1;
    $pitch72$0 = (($808) + ($pitch72$0$v))|0;
    $816 = (($y2) - ($y1))|0;
    $length71$0 = $816;$pitch72$2 = $pitch72$0;$pixel73$1 = $814;
   }
   $not$35 = ($draw_end|0)!=(0);
   $823 = $not$35&1;
   $length71$0$ = (($length71$0) + ($823))|0;
   $824 = ($length71$0$|0)==(0);
   if ($824) {
    return;
   }
   $825 = ((($1)) + 28|0);
   $826 = HEAP32[$825>>2]|0;
   $827 = $826 & 255;
   $828 = $r$0 >>> $827;
   $829 = ((($1)) + 32|0);
   $830 = HEAP32[$829>>2]|0;
   $831 = $830 & 255;
   $832 = $828 << $831;
   $833 = $826 >>> 8;
   $834 = $833 & 255;
   $835 = $g$0 >>> $834;
   $836 = $830 >>> 8;
   $837 = $836 & 255;
   $838 = $835 << $837;
   $839 = $838 | $832;
   $840 = $826 >>> 16;
   $841 = $840 & 255;
   $842 = $b$0 >>> $841;
   $843 = $830 >>> 16;
   $844 = $843 & 255;
   $845 = $842 << $844;
   $846 = ((($1)) + 24|0);
   $847 = HEAP32[$846>>2]|0;
   $848 = $839 | $847;
   $849 = $848 | $845;
   $850 = $849&65535;
   $length71$182 = $length71$0$;$pixel73$281 = $pixel73$1;
   while(1) {
    $851 = (($length71$182) + -1)|0;
    HEAP16[$pixel73$281>>1] = $850;
    $852 = (($pixel73$281) + ($pitch72$2<<1)|0);
    $853 = ($851|0)==(0);
    if ($853) {
     break;
    } else {
     $length71$182 = $851;$pixel73$281 = $852;
    }
   }
   return;
  }
  }
 } else {
  switch ($blendMode|0) {
  case 1:  {
   $854 = (($x2) - ($x1))|0;
   $855 = ($854|0)<(0);
   $856 = (0 - ($854))|0;
   $857 = $855 ? $856 : $854;
   $858 = (($y2) - ($y1))|0;
   $859 = ($858|0)<(0);
   $860 = (0 - ($858))|0;
   $861 = $859 ? $860 : $858;
   $862 = ($857|0)<($861|0);
   if ($862) {
    $866 = $857 << 1;
    $867 = (($866) - ($861))|0;
    $868 = (($857) - ($861))|0;
    $d$0 = $867;$dinc1$0 = $866;$dinc2$0$in = $868;$numpixels$0$in = $861;$xinc1$0 = 0;$yinc1$0 = 1;
   } else {
    $863 = $861 << 1;
    $864 = (($863) - ($857))|0;
    $865 = (($861) - ($857))|0;
    $d$0 = $864;$dinc1$0 = $863;$dinc2$0$in = $865;$numpixels$0$in = $857;$xinc1$0 = 1;$yinc1$0 = 0;
   }
   $dinc2$0 = $dinc2$0$in << 1;
   $869 = ($x2|0)<($x1|0);
   $870 = (0 - ($xinc1$0))|0;
   $$xinc1$0 = $869 ? $870 : $xinc1$0;
   $$25 = $869 ? -1 : 1;
   $871 = ($y2|0)<($y1|0);
   $872 = (0 - ($yinc1$0))|0;
   $yinc1$1 = $871 ? $872 : $yinc1$0;
   $yinc2$1 = $871 ? -1 : 1;
   $not$30 = ($draw_end|0)!=(0);
   $numpixels$0 = $not$30&1;
   $numpixels$0$in$numpixels$0 = (($numpixels$0$in) + ($numpixels$0))|0;
   $873 = ($numpixels$0$in$numpixels$0|0)>(0);
   if (!($873)) {
    return;
   }
   $874 = ((($dst)) + 20|0);
   $875 = ((($dst)) + 16|0);
   $876 = ((($1)) + 12|0);
   $877 = ((($1)) + 32|0);
   $878 = ((($1)) + 28|0);
   $879 = ((($1)) + 16|0);
   $880 = ((($1)) + 20|0);
   $881 = ((($1)) + 24|0);
   $$pre160 = HEAP32[$874>>2]|0;
   $$pre161 = HEAP32[$875>>2]|0;
   $$pre162 = HEAP32[$876>>2]|0;
   $$pre163 = HEAP32[$877>>2]|0;
   $$pre164 = HEAP32[$878>>2]|0;
   $$phi$trans$insert165 = $$pre164 & 255;
   $$phi$trans$insert166 = (125924 + ($$phi$trans$insert165<<2)|0);
   $$pre167 = HEAP32[$$phi$trans$insert166>>2]|0;
   $$pre168 = HEAP32[$879>>2]|0;
   $882 = $$pre163 >>> 8;
   $883 = $882 & 255;
   $884 = $$pre164 >>> 8;
   $885 = $884 & 255;
   $$phi$trans$insert170 = (125924 + ($885<<2)|0);
   $$pre171 = HEAP32[$$phi$trans$insert170>>2]|0;
   $$pre172 = HEAP32[$880>>2]|0;
   $886 = $$pre163 >>> 16;
   $887 = $$pre164 >>> 16;
   $$phi$trans$insert173 = $887 & 255;
   $$phi$trans$insert174 = (125924 + ($$phi$trans$insert173<<2)|0);
   $$pre175 = HEAP32[$$phi$trans$insert174>>2]|0;
   $$pre176 = HEAP32[$881>>2]|0;
   $888 = $$pre163 & 255;
   $889 = $886 & 255;
   $d$199 = $d$0;$i$098 = 0;$x$0100 = $x1;$y$0101 = $y1;
   while(1) {
    $890 = Math_imul($$pre161, $y$0101)|0;
    $891 = (($$pre160) + ($890)|0);
    $892 = $x$0100 << 1;
    $893 = (($891) + ($892)|0);
    $894 = HEAP16[$893>>1]|0;
    $895 = $894&65535;
    $896 = $895 & $$pre162;
    $897 = $896 >>> $888;
    $898 = (($$pre167) + ($897)|0);
    $899 = HEAP8[$898>>0]|0;
    $900 = $899&255;
    $901 = $$pre168 & $895;
    $902 = $901 >>> $883;
    $903 = (($$pre171) + ($902)|0);
    $904 = HEAP8[$903>>0]|0;
    $905 = $904&255;
    $906 = $$pre172 & $895;
    $907 = $906 >>> $889;
    $908 = (($$pre175) + ($907)|0);
    $909 = HEAP8[$908>>0]|0;
    $910 = $909&255;
    $911 = Math_imul($900, $16)|0;
    $912 = (($911>>>0) / 255)&-1;
    $913 = (($912) + ($r$0))|0;
    $914 = Math_imul($905, $16)|0;
    $915 = (($914>>>0) / 255)&-1;
    $916 = (($915) + ($g$0))|0;
    $917 = Math_imul($910, $16)|0;
    $918 = (($917>>>0) / 255)&-1;
    $919 = (($918) + ($b$0))|0;
    $920 = $913 >>> $$phi$trans$insert165;
    $921 = $920 << $888;
    $922 = $916 >>> $885;
    $923 = $922 << $883;
    $924 = $923 | $921;
    $925 = $919 >>> $$phi$trans$insert173;
    $926 = $925 << $889;
    $927 = $924 | $$pre176;
    $928 = $927 | $926;
    $929 = $928&65535;
    HEAP16[$893>>1] = $929;
    $930 = ($d$199|0)<(0);
    $dinc1$0$dinc2$0 = $930 ? $dinc1$0 : $dinc2$0;
    $$xinc1$0$$25 = $930 ? $$xinc1$0 : $$25;
    $yinc1$1$yinc2$1 = $930 ? $yinc1$1 : $yinc2$1;
    $y$1 = (($y$0101) + ($yinc1$1$yinc2$1))|0;
    $x$1 = (($$xinc1$0$$25) + ($x$0100))|0;
    $d$2 = (($dinc1$0$dinc2$0) + ($d$199))|0;
    $931 = (($i$098) + 1)|0;
    $932 = ($931|0)<($numpixels$0$in$numpixels$0|0);
    if ($932) {
     $d$199 = $d$2;$i$098 = $931;$x$0100 = $x$1;$y$0101 = $y$1;
    } else {
     break;
    }
   }
   return;
   break;
  }
  case 2:  {
   $933 = (($x2) - ($x1))|0;
   $934 = ($933|0)<(0);
   $935 = (0 - ($933))|0;
   $936 = $934 ? $935 : $933;
   $937 = (($y2) - ($y1))|0;
   $938 = ($937|0)<(0);
   $939 = (0 - ($937))|0;
   $940 = $938 ? $939 : $937;
   $941 = ($936|0)<($940|0);
   if ($941) {
    $945 = $936 << 1;
    $946 = (($945) - ($940))|0;
    $947 = (($936) - ($940))|0;
    $d87$0 = $946;$dinc188$0 = $945;$dinc289$0$in = $947;$numpixels86$0$in = $940;$xinc191$0 = 0;$yinc194$0 = 1;
   } else {
    $942 = $940 << 1;
    $943 = (($942) - ($936))|0;
    $944 = (($940) - ($936))|0;
    $d87$0 = $943;$dinc188$0 = $942;$dinc289$0$in = $944;$numpixels86$0$in = $936;$xinc191$0 = 1;$yinc194$0 = 0;
   }
   $dinc289$0 = $dinc289$0$in << 1;
   $948 = ($x2|0)<($x1|0);
   $949 = (0 - ($xinc191$0))|0;
   $$xinc191$0 = $948 ? $949 : $xinc191$0;
   $$26 = $948 ? -1 : 1;
   $950 = ($y2|0)<($y1|0);
   $951 = (0 - ($yinc194$0))|0;
   $yinc194$1 = $950 ? $951 : $yinc194$0;
   $yinc295$1 = $950 ? -1 : 1;
   $not$29 = ($draw_end|0)!=(0);
   $numpixels86$0 = $not$29&1;
   $numpixels86$0$in$numpixels86$0 = (($numpixels86$0$in) + ($numpixels86$0))|0;
   $952 = ($numpixels86$0$in$numpixels86$0|0)>(0);
   if (!($952)) {
    return;
   }
   $953 = ((($dst)) + 20|0);
   $954 = ((($dst)) + 16|0);
   $955 = ((($1)) + 12|0);
   $956 = ((($1)) + 32|0);
   $957 = ((($1)) + 28|0);
   $958 = ((($1)) + 16|0);
   $959 = ((($1)) + 20|0);
   $960 = ((($1)) + 24|0);
   $$pre143 = HEAP32[$953>>2]|0;
   $$pre144 = HEAP32[$954>>2]|0;
   $$pre145 = HEAP32[$955>>2]|0;
   $$pre146 = HEAP32[$956>>2]|0;
   $$pre147 = HEAP32[$957>>2]|0;
   $$phi$trans$insert148 = $$pre147 & 255;
   $$phi$trans$insert149 = (125924 + ($$phi$trans$insert148<<2)|0);
   $$pre150 = HEAP32[$$phi$trans$insert149>>2]|0;
   $$pre151 = HEAP32[$958>>2]|0;
   $961 = $$pre146 >>> 8;
   $962 = $961 & 255;
   $963 = $$pre147 >>> 8;
   $964 = $963 & 255;
   $$phi$trans$insert153 = (125924 + ($964<<2)|0);
   $$pre154 = HEAP32[$$phi$trans$insert153>>2]|0;
   $$pre155 = HEAP32[$959>>2]|0;
   $965 = $$pre146 >>> 16;
   $966 = $$pre147 >>> 16;
   $$phi$trans$insert156 = $966 & 255;
   $$phi$trans$insert157 = (125924 + ($$phi$trans$insert156<<2)|0);
   $$pre158 = HEAP32[$$phi$trans$insert157>>2]|0;
   $$pre159 = HEAP32[$960>>2]|0;
   $967 = $$pre146 & 255;
   $968 = $965 & 255;
   $d87$1104 = $d87$0;$i83$0103 = 0;$x90$0105 = $x1;$y93$0106 = $y1;
   while(1) {
    $969 = Math_imul($$pre144, $y93$0106)|0;
    $970 = (($$pre143) + ($969)|0);
    $971 = $x90$0105 << 1;
    $972 = (($970) + ($971)|0);
    $973 = HEAP16[$972>>1]|0;
    $974 = $973&65535;
    $975 = $974 & $$pre145;
    $976 = $975 >>> $967;
    $977 = (($$pre150) + ($976)|0);
    $978 = HEAP8[$977>>0]|0;
    $979 = $978&255;
    $980 = $$pre151 & $974;
    $981 = $980 >>> $962;
    $982 = (($$pre154) + ($981)|0);
    $983 = HEAP8[$982>>0]|0;
    $984 = $983&255;
    $985 = $$pre155 & $974;
    $986 = $985 >>> $968;
    $987 = (($$pre158) + ($986)|0);
    $988 = HEAP8[$987>>0]|0;
    $989 = $988&255;
    $990 = (($979) + ($r$0))|0;
    $991 = ($990>>>0)>(255);
    $$11 = $991 ? 255 : $990;
    $992 = (($984) + ($g$0))|0;
    $993 = ($992>>>0)>(255);
    $$7 = $993 ? 255 : $992;
    $994 = (($989) + ($b$0))|0;
    $995 = ($994>>>0)>(255);
    $$12 = $995 ? 255 : $994;
    $996 = $$11 >>> $$phi$trans$insert148;
    $997 = $996 << $967;
    $998 = $$7 >>> $964;
    $999 = $998 << $962;
    $1000 = $999 | $997;
    $1001 = $$12 >>> $$phi$trans$insert156;
    $1002 = $1001 << $968;
    $1003 = $1000 | $$pre159;
    $1004 = $1003 | $1002;
    $1005 = $1004&65535;
    HEAP16[$972>>1] = $1005;
    $1006 = ($d87$1104|0)<(0);
    $dinc188$0$dinc289$0 = $1006 ? $dinc188$0 : $dinc289$0;
    $$xinc191$0$$26 = $1006 ? $$xinc191$0 : $$26;
    $yinc194$1$yinc295$1 = $1006 ? $yinc194$1 : $yinc295$1;
    $y93$1 = (($y93$0106) + ($yinc194$1$yinc295$1))|0;
    $x90$1 = (($$xinc191$0$$26) + ($x90$0105))|0;
    $d87$2 = (($dinc188$0$dinc289$0) + ($d87$1104))|0;
    $1007 = (($i83$0103) + 1)|0;
    $1008 = ($1007|0)<($numpixels86$0$in$numpixels86$0|0);
    if ($1008) {
     $d87$1104 = $d87$2;$i83$0103 = $1007;$x90$0105 = $x90$1;$y93$0106 = $y93$1;
    } else {
     break;
    }
   }
   return;
   break;
  }
  case 4:  {
   $1009 = (($x2) - ($x1))|0;
   $1010 = ($1009|0)<(0);
   $1011 = (0 - ($1009))|0;
   $1012 = $1010 ? $1011 : $1009;
   $1013 = (($y2) - ($y1))|0;
   $1014 = ($1013|0)<(0);
   $1015 = (0 - ($1013))|0;
   $1016 = $1014 ? $1015 : $1013;
   $1017 = ($1012|0)<($1016|0);
   if ($1017) {
    $1021 = $1012 << 1;
    $1022 = (($1021) - ($1016))|0;
    $1023 = (($1012) - ($1016))|0;
    $d105$0 = $1022;$dinc1106$0 = $1021;$dinc2107$0$in = $1023;$numpixels104$0$in = $1016;$xinc1109$0 = 0;$yinc1112$0 = 1;
   } else {
    $1018 = $1016 << 1;
    $1019 = (($1018) - ($1012))|0;
    $1020 = (($1016) - ($1012))|0;
    $d105$0 = $1019;$dinc1106$0 = $1018;$dinc2107$0$in = $1020;$numpixels104$0$in = $1012;$xinc1109$0 = 1;$yinc1112$0 = 0;
   }
   $dinc2107$0 = $dinc2107$0$in << 1;
   $1024 = ($x2|0)<($x1|0);
   $1025 = (0 - ($xinc1109$0))|0;
   $$xinc1109$0 = $1024 ? $1025 : $xinc1109$0;
   $$27 = $1024 ? -1 : 1;
   $1026 = ($y2|0)<($y1|0);
   $1027 = (0 - ($yinc1112$0))|0;
   $yinc1112$1 = $1026 ? $1027 : $yinc1112$0;
   $yinc2113$1 = $1026 ? -1 : 1;
   $not$ = ($draw_end|0)!=(0);
   $numpixels104$0 = $not$&1;
   $numpixels104$0$in$numpixels104$0 = (($numpixels104$0$in) + ($numpixels104$0))|0;
   $1028 = ($numpixels104$0$in$numpixels104$0|0)>(0);
   if (!($1028)) {
    return;
   }
   $1029 = ((($dst)) + 20|0);
   $1030 = ((($dst)) + 16|0);
   $1031 = ((($1)) + 12|0);
   $1032 = ((($1)) + 32|0);
   $1033 = ((($1)) + 28|0);
   $1034 = ((($1)) + 16|0);
   $1035 = ((($1)) + 20|0);
   $1036 = ((($1)) + 24|0);
   $$pre = HEAP32[$1029>>2]|0;
   $$pre128 = HEAP32[$1030>>2]|0;
   $$pre129 = HEAP32[$1031>>2]|0;
   $$pre130 = HEAP32[$1032>>2]|0;
   $$pre131 = HEAP32[$1033>>2]|0;
   $$phi$trans$insert = $$pre131 & 255;
   $$phi$trans$insert132 = (125924 + ($$phi$trans$insert<<2)|0);
   $$pre133 = HEAP32[$$phi$trans$insert132>>2]|0;
   $$pre134 = HEAP32[$1034>>2]|0;
   $1037 = $$pre130 >>> 8;
   $1038 = $1037 & 255;
   $1039 = $$pre131 >>> 8;
   $1040 = $1039 & 255;
   $$phi$trans$insert136 = (125924 + ($1040<<2)|0);
   $$pre137 = HEAP32[$$phi$trans$insert136>>2]|0;
   $$pre138 = HEAP32[$1035>>2]|0;
   $1041 = $$pre130 >>> 16;
   $1042 = $$pre131 >>> 16;
   $$phi$trans$insert139 = $1042 & 255;
   $$phi$trans$insert140 = (125924 + ($$phi$trans$insert139<<2)|0);
   $$pre141 = HEAP32[$$phi$trans$insert140>>2]|0;
   $$pre142 = HEAP32[$1036>>2]|0;
   $1043 = $$pre130 & 255;
   $1044 = $1041 & 255;
   $d105$1109 = $d105$0;$i101$0108 = 0;$x108$0110 = $x1;$y111$0111 = $y1;
   while(1) {
    $1045 = Math_imul($$pre128, $y111$0111)|0;
    $1046 = (($$pre) + ($1045)|0);
    $1047 = $x108$0110 << 1;
    $1048 = (($1046) + ($1047)|0);
    $1049 = HEAP16[$1048>>1]|0;
    $1050 = $1049&65535;
    $1051 = $1050 & $$pre129;
    $1052 = $1051 >>> $1043;
    $1053 = (($$pre133) + ($1052)|0);
    $1054 = HEAP8[$1053>>0]|0;
    $1055 = $1054&255;
    $1056 = $$pre134 & $1050;
    $1057 = $1056 >>> $1038;
    $1058 = (($$pre137) + ($1057)|0);
    $1059 = HEAP8[$1058>>0]|0;
    $1060 = $1059&255;
    $1061 = $$pre138 & $1050;
    $1062 = $1061 >>> $1044;
    $1063 = (($$pre141) + ($1062)|0);
    $1064 = HEAP8[$1063>>0]|0;
    $1065 = $1064&255;
    $1066 = Math_imul($1055, $r$0)|0;
    $1067 = (($1066>>>0) / 255)&-1;
    $1068 = Math_imul($1060, $g$0)|0;
    $1069 = (($1068>>>0) / 255)&-1;
    $1070 = Math_imul($1065, $b$0)|0;
    $1071 = (($1070>>>0) / 255)&-1;
    $1072 = $1067 >>> $$phi$trans$insert;
    $1073 = $1072 << $1043;
    $1074 = $1069 >>> $1040;
    $1075 = $1074 << $1038;
    $1076 = $1075 | $1073;
    $1077 = $1071 >>> $$phi$trans$insert139;
    $1078 = $1077 << $1044;
    $1079 = $1076 | $$pre142;
    $1080 = $1079 | $1078;
    $1081 = $1080&65535;
    HEAP16[$1048>>1] = $1081;
    $1082 = ($d105$1109|0)<(0);
    $dinc1106$0$dinc2107$0 = $1082 ? $dinc1106$0 : $dinc2107$0;
    $$xinc1109$0$$27 = $1082 ? $$xinc1109$0 : $$27;
    $yinc1112$1$yinc2113$1 = $1082 ? $yinc1112$1 : $yinc2113$1;
    $y111$1 = (($y111$0111) + ($yinc1112$1$yinc2113$1))|0;
    $x108$1 = (($$xinc1109$0$$27) + ($x108$0110))|0;
    $d105$2 = (($dinc1106$0$dinc2107$0) + ($d105$1109))|0;
    $1083 = (($i101$0108) + 1)|0;
    $1084 = ($1083|0)<($numpixels104$0$in$numpixels104$0|0);
    if ($1084) {
     $d105$1109 = $d105$2;$i101$0108 = $1083;$x108$0110 = $x108$1;$y111$0111 = $y111$1;
    } else {
     break;
    }
   }
   return;
   break;
  }
  default: {
   $1085 = (($x2) - ($x1))|0;
   $1086 = ($1085|0)<(0);
   $1087 = (0 - ($1085))|0;
   $1088 = $1086 ? $1087 : $1085;
   $1089 = (($y2) - ($y1))|0;
   $1090 = ($1089|0)<(0);
   $1091 = (0 - ($1089))|0;
   $1092 = $1090 ? $1091 : $1089;
   $1093 = ($1088|0)<($1092|0);
   if ($1093) {
    $1097 = $1088 << 1;
    $1098 = (($1097) - ($1092))|0;
    $1099 = (($1088) - ($1092))|0;
    $d123$0 = $1098;$dinc1124$0 = $1097;$dinc2125$0$in = $1099;$numpixels122$0$in = $1092;$xinc1127$0 = 0;$yinc1130$0 = 1;
   } else {
    $1094 = $1092 << 1;
    $1095 = (($1094) - ($1088))|0;
    $1096 = (($1092) - ($1088))|0;
    $d123$0 = $1095;$dinc1124$0 = $1094;$dinc2125$0$in = $1096;$numpixels122$0$in = $1088;$xinc1127$0 = 1;$yinc1130$0 = 0;
   }
   $dinc2125$0 = $dinc2125$0$in << 1;
   $1100 = ($x2|0)<($x1|0);
   $1101 = (0 - ($xinc1127$0))|0;
   $$xinc1127$0 = $1100 ? $1101 : $xinc1127$0;
   $$28 = $1100 ? -1 : 1;
   $1102 = ($y2|0)<($y1|0);
   $1103 = (0 - ($yinc1130$0))|0;
   $yinc1130$1 = $1102 ? $1103 : $yinc1130$0;
   $yinc2131$1 = $1102 ? -1 : 1;
   $not$31 = ($draw_end|0)!=(0);
   $numpixels122$0 = $not$31&1;
   $numpixels122$0$in$numpixels122$0 = (($numpixels122$0$in) + ($numpixels122$0))|0;
   $1104 = ($numpixels122$0$in$numpixels122$0|0)>(0);
   if (!($1104)) {
    return;
   }
   $1105 = ((($dst)) + 20|0);
   $1106 = HEAP32[$1105>>2]|0;
   $1107 = ((($dst)) + 16|0);
   $1108 = HEAP32[$1107>>2]|0;
   $1109 = ((($1)) + 28|0);
   $1110 = HEAP32[$1109>>2]|0;
   $1111 = $1110 & 255;
   $1112 = $r$0 >>> $1111;
   $1113 = ((($1)) + 32|0);
   $1114 = HEAP32[$1113>>2]|0;
   $1115 = $1114 & 255;
   $1116 = $1112 << $1115;
   $1117 = $1110 >>> 8;
   $1118 = $1117 & 255;
   $1119 = $g$0 >>> $1118;
   $1120 = $1114 >>> 8;
   $1121 = $1120 & 255;
   $1122 = $1119 << $1121;
   $1123 = $1122 | $1116;
   $1124 = $1110 >>> 16;
   $1125 = $1124 & 255;
   $1126 = $b$0 >>> $1125;
   $1127 = $1114 >>> 16;
   $1128 = $1127 & 255;
   $1129 = $1126 << $1128;
   $1130 = ((($1)) + 24|0);
   $1131 = HEAP32[$1130>>2]|0;
   $1132 = $1123 | $1131;
   $1133 = $1132 | $1129;
   $1134 = $1133&65535;
   $d123$194 = $d123$0;$i119$093 = 0;$x126$095 = $x1;$y129$096 = $y1;
   while(1) {
    $1135 = Math_imul($1108, $y129$096)|0;
    $1136 = (($1106) + ($1135)|0);
    $1137 = $x126$095 << 1;
    $1138 = (($1136) + ($1137)|0);
    HEAP16[$1138>>1] = $1134;
    $1139 = ($d123$194|0)<(0);
    $dinc1124$0$dinc2125$0 = $1139 ? $dinc1124$0 : $dinc2125$0;
    $$xinc1127$0$$28 = $1139 ? $$xinc1127$0 : $$28;
    $yinc1130$1$yinc2131$1 = $1139 ? $yinc1130$1 : $yinc2131$1;
    $y129$1 = (($y129$096) + ($yinc1130$1$yinc2131$1))|0;
    $x126$1 = (($$xinc1127$0$$28) + ($x126$095))|0;
    $d123$2 = (($dinc1124$0$dinc2125$0) + ($d123$194))|0;
    $1140 = (($i119$093) + 1)|0;
    $1141 = ($1140|0)<($numpixels122$0$in$numpixels122$0|0);
    if ($1141) {
     $d123$194 = $d123$2;$i119$093 = $1140;$x126$095 = $x126$1;$y129$096 = $y129$1;
    } else {
     break;
    }
   }
   return;
  }
  }
 }
}
function _SDL_BlendLine_RGB555($dst,$x1,$y1,$x2,$y2,$blendMode,$_r,$_g,$_b,$_a,$draw_end) {
 $dst = $dst|0;
 $x1 = $x1|0;
 $y1 = $y1|0;
 $x2 = $x2|0;
 $y2 = $y2|0;
 $blendMode = $blendMode|0;
 $_r = $_r|0;
 $_g = $_g|0;
 $_b = $_b|0;
 $_a = $_a|0;
 $draw_end = $draw_end|0;
 var $$ = 0, $$10 = 0, $$11 = 0, $$12 = 0, $$13 = 0, $$14 = 0, $$15 = 0, $$16 = 0, $$17 = 0, $$18 = 0, $$19 = 0, $$2 = 0, $$20 = 0, $$21 = 0, $$22 = 0, $$23 = 0, $$24 = 0, $$25 = 0, $$26 = 0, $$27 = 0;
 var $$28 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$6 = 0, $$7 = 0, $$8 = 0, $$9 = 0, $$pre = 0, $$pre128 = 0, $$pre129 = 0, $$pre130 = 0, $$pre131 = 0, $$pre132 = 0, $$pre133 = 0, $$pre134 = 0, $$pre135 = 0, $$pre136 = 0, $$pre137 = 0, $$pre138 = 0;
 var $$pre139 = 0, $$pre140 = 0, $$pre141 = 0, $$pre142 = 0, $$pre143 = 0, $$pre144 = 0, $$xinc1$0 = 0, $$xinc1$0$$25 = 0, $$xinc1109$0 = 0, $$xinc1109$0$$27 = 0, $$xinc1127$0 = 0, $$xinc1127$0$$28 = 0, $$xinc191$0 = 0, $$xinc191$0$$26 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0;
 var $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0;
 var $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0;
 var $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0;
 var $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0;
 var $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0;
 var $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0;
 var $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0;
 var $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0;
 var $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0;
 var $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0;
 var $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0;
 var $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0;
 var $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0;
 var $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0;
 var $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0;
 var $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0;
 var $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0;
 var $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0;
 var $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0;
 var $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0;
 var $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0;
 var $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0;
 var $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0;
 var $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0;
 var $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0;
 var $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0;
 var $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0;
 var $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0;
 var $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0;
 var $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0;
 var $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0;
 var $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0;
 var $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0;
 var $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0;
 var $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0;
 var $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0;
 var $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0;
 var $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0;
 var $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $a$0 = 0, $b$0 = 0, $blendMode$off = 0, $d$0 = 0, $d$199 = 0, $d$2 = 0, $d105$0 = 0, $d105$1109 = 0, $d105$2 = 0, $d123$0 = 0, $d123$194 = 0, $d123$2 = 0, $d87$0 = 0, $d87$1104 = 0, $d87$2 = 0, $dinc1$0 = 0, $dinc1$0$dinc2$0 = 0, $dinc1106$0 = 0;
 var $dinc1106$0$dinc2107$0 = 0, $dinc1124$0 = 0, $dinc1124$0$dinc2125$0 = 0, $dinc188$0 = 0, $dinc188$0$dinc289$0 = 0, $dinc2$0 = 0, $dinc2$0$in = 0, $dinc2107$0 = 0, $dinc2107$0$in = 0, $dinc2125$0 = 0, $dinc2125$0$in = 0, $dinc289$0 = 0, $dinc289$0$in = 0, $g$0 = 0, $i$098 = 0, $i101$0108 = 0, $i119$093 = 0, $i83$0103 = 0, $length$0$ph = 0, $length$061 = 0;
 var $length1$0$ph = 0, $length1$064 = 0, $length15$0$ph = 0, $length15$059 = 0, $length22$0$ph = 0, $length22$073 = 0, $length29$0$ph = 0, $length29$076 = 0, $length36$0$ph = 0, $length36$079 = 0, $length43$0$ph = 0, $length43$070 = 0, $length50$0 = 0, $length50$0$ = 0, $length50$185 = 0, $length57$0 = 0, $length57$0$ = 0, $length57$188 = 0, $length64$0 = 0, $length64$0$ = 0;
 var $length64$191 = 0, $length71$0 = 0, $length71$0$ = 0, $length71$182 = 0, $length8$0$ph = 0, $length8$067 = 0, $not$ = 0, $not$29 = 0, $not$30 = 0, $not$31 = 0, $not$32 = 0, $not$33 = 0, $not$34 = 0, $not$35 = 0, $numpixels$0 = 0, $numpixels$0$in = 0, $numpixels$0$in$numpixels$0 = 0, $numpixels104$0 = 0, $numpixels104$0$in = 0, $numpixels104$0$in$numpixels104$0 = 0;
 var $numpixels122$0 = 0, $numpixels122$0$in = 0, $numpixels122$0$in$numpixels122$0 = 0, $numpixels86$0 = 0, $numpixels86$0$in = 0, $numpixels86$0$in$numpixels86$0 = 0, $pitch51$0 = 0, $pitch51$0$v = 0, $pitch51$1 = 0, $pitch51$1$v = 0, $pitch51$2 = 0, $pitch58$0 = 0, $pitch58$0$v = 0, $pitch58$1 = 0, $pitch58$1$v = 0, $pitch58$2 = 0, $pitch65$0 = 0, $pitch65$0$v = 0, $pitch65$1 = 0, $pitch65$1$v = 0;
 var $pitch65$2 = 0, $pitch72$0 = 0, $pitch72$0$v = 0, $pitch72$1 = 0, $pitch72$1$v = 0, $pitch72$2 = 0, $pixel$1$ph = 0, $pixel$160 = 0, $pixel10$1$ph = 0, $pixel10$166 = 0, $pixel17$1$ph = 0, $pixel17$158 = 0, $pixel24$1$ph = 0, $pixel24$172 = 0, $pixel3$1$ph = 0, $pixel3$163 = 0, $pixel31$1$ph = 0, $pixel31$175 = 0, $pixel38$1$ph = 0, $pixel38$178 = 0;
 var $pixel45$1$ph = 0, $pixel45$169 = 0, $pixel52$1 = 0, $pixel52$284 = 0, $pixel59$1 = 0, $pixel59$287 = 0, $pixel66$1 = 0, $pixel66$290 = 0, $pixel73$1 = 0, $pixel73$281 = 0, $r$0 = 0, $x$0100 = 0, $x$1 = 0, $x108$0110 = 0, $x108$1 = 0, $x126$095 = 0, $x126$1 = 0, $x90$0105 = 0, $x90$1 = 0, $xinc1$0 = 0;
 var $xinc1109$0 = 0, $xinc1127$0 = 0, $xinc191$0 = 0, $y$0101 = 0, $y$1 = 0, $y111$0111 = 0, $y111$1 = 0, $y129$096 = 0, $y129$1 = 0, $y93$0106 = 0, $y93$1 = 0, $yinc1$0 = 0, $yinc1$1 = 0, $yinc1$1$yinc2$1 = 0, $yinc1112$0 = 0, $yinc1112$1 = 0, $yinc1112$1$yinc2113$1 = 0, $yinc1130$0 = 0, $yinc1130$1 = 0, $yinc1130$1$yinc2131$1 = 0;
 var $yinc194$0 = 0, $yinc194$1 = 0, $yinc194$1$yinc295$1 = 0, $yinc2$1 = 0, $yinc2113$1 = 0, $yinc2131$1 = 0, $yinc295$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $blendMode$off = (($blendMode) + -1)|0;
 $0 = ($blendMode$off>>>0)<(2);
 $1 = $_r&255;
 if ($0) {
  $2 = $_a&255;
  $3 = Math_imul($2, $1)|0;
  $4 = (($3>>>0) / 255)&-1;
  $5 = $_g&255;
  $6 = Math_imul($2, $5)|0;
  $7 = (($6>>>0) / 255)&-1;
  $8 = $_b&255;
  $9 = Math_imul($2, $8)|0;
  $10 = (($9>>>0) / 255)&-1;
  $a$0 = $2;$b$0 = $10;$g$0 = $7;$r$0 = $4;
 } else {
  $11 = $_g&255;
  $12 = $_b&255;
  $13 = $_a&255;
  $a$0 = $13;$b$0 = $12;$g$0 = $11;$r$0 = $1;
 }
 $14 = $a$0 ^ 255;
 $15 = ($y2|0)==($y1|0);
 if ($15) {
  switch ($blendMode|0) {
  case 1:  {
   $16 = ((($dst)) + 16|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($dst)) + 4|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ((($19)) + 9|0);
   $21 = HEAP8[$20>>0]|0;
   $22 = $21&255;
   $23 = (($17|0) / ($22|0))&-1;
   $24 = ($x2|0)<($x1|0);
   $25 = ((($dst)) + 20|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = Math_imul($23, $y2)|0;
   $28 = (($26) + ($27<<1)|0);
   if ($24) {
    $34 = (($28) + ($x2<<1)|0);
    $35 = ($draw_end|0)!=(0);
    $36 = ((($34)) + 2|0);
    $$13 = $35 ? $34 : $36;
    $37 = (($x1) - ($x2))|0;
    $38 = $35&1;
    $39 = (($37) + ($38))|0;
    $length$0$ph = $39;$pixel$1$ph = $$13;
   } else {
    $29 = (($28) + ($x1<<1)|0);
    $30 = ($draw_end|0)!=(0);
    $31 = (($x2) - ($x1))|0;
    $32 = $30&1;
    $33 = (($31) + ($32))|0;
    $length$0$ph = $33;$pixel$1$ph = $29;
   }
   $40 = ($length$0$ph|0)==(0);
   if ($40) {
    return;
   }
   $$pre144 = HEAP32[(125936)>>2]|0;
   $length$061 = $length$0$ph;$pixel$160 = $pixel$1$ph;
   while(1) {
    $41 = (($length$061) + -1)|0;
    $42 = HEAP16[$pixel$160>>1]|0;
    $43 = $42&65535;
    $44 = $43 >>> 10;
    $45 = $44 & 31;
    $46 = (($$pre144) + ($45)|0);
    $47 = HEAP8[$46>>0]|0;
    $48 = $47&255;
    $49 = $43 >>> 5;
    $50 = $49 & 31;
    $51 = (($$pre144) + ($50)|0);
    $52 = HEAP8[$51>>0]|0;
    $53 = $52&255;
    $54 = $43 & 31;
    $55 = (($$pre144) + ($54)|0);
    $56 = HEAP8[$55>>0]|0;
    $57 = $56&255;
    $58 = Math_imul($48, $14)|0;
    $59 = (($58>>>0) / 255)&-1;
    $60 = (($59) + ($r$0))|0;
    $61 = Math_imul($53, $14)|0;
    $62 = (($61>>>0) / 255)&-1;
    $63 = (($62) + ($g$0))|0;
    $64 = Math_imul($57, $14)|0;
    $65 = (($64>>>0) / 255)&-1;
    $66 = (($65) + ($b$0))|0;
    $67 = $60 >>> 3;
    $68 = $67 << 10;
    $69 = $63 >>> 3;
    $70 = $69 << 5;
    $71 = $70 | $68;
    $72 = $66 >>> 3;
    $73 = $71 | $72;
    $74 = $73&65535;
    HEAP16[$pixel$160>>1] = $74;
    $75 = ((($pixel$160)) + 2|0);
    $76 = ($41|0)==(0);
    if ($76) {
     break;
    } else {
     $length$061 = $41;$pixel$160 = $75;
    }
   }
   return;
   break;
  }
  case 2:  {
   $77 = ((($dst)) + 16|0);
   $78 = HEAP32[$77>>2]|0;
   $79 = ((($dst)) + 4|0);
   $80 = HEAP32[$79>>2]|0;
   $81 = ((($80)) + 9|0);
   $82 = HEAP8[$81>>0]|0;
   $83 = $82&255;
   $84 = (($78|0) / ($83|0))&-1;
   $85 = ($x2|0)<($x1|0);
   $86 = ((($dst)) + 20|0);
   $87 = HEAP32[$86>>2]|0;
   $88 = Math_imul($84, $y2)|0;
   $89 = (($87) + ($88<<1)|0);
   if ($85) {
    $95 = (($89) + ($x2<<1)|0);
    $96 = ($draw_end|0)!=(0);
    $97 = ((($95)) + 2|0);
    $$14 = $96 ? $95 : $97;
    $98 = (($x1) - ($x2))|0;
    $99 = $96&1;
    $100 = (($98) + ($99))|0;
    $length1$0$ph = $100;$pixel3$1$ph = $$14;
   } else {
    $90 = (($89) + ($x1<<1)|0);
    $91 = ($draw_end|0)!=(0);
    $92 = (($x2) - ($x1))|0;
    $93 = $91&1;
    $94 = (($92) + ($93))|0;
    $length1$0$ph = $94;$pixel3$1$ph = $90;
   }
   $101 = ($length1$0$ph|0)==(0);
   if ($101) {
    return;
   }
   $$pre143 = HEAP32[(125936)>>2]|0;
   $length1$064 = $length1$0$ph;$pixel3$163 = $pixel3$1$ph;
   while(1) {
    $102 = (($length1$064) + -1)|0;
    $103 = HEAP16[$pixel3$163>>1]|0;
    $104 = $103&65535;
    $105 = $104 >>> 10;
    $106 = $105 & 31;
    $107 = (($$pre143) + ($106)|0);
    $108 = HEAP8[$107>>0]|0;
    $109 = $108&255;
    $110 = $104 >>> 5;
    $111 = $110 & 31;
    $112 = (($$pre143) + ($111)|0);
    $113 = HEAP8[$112>>0]|0;
    $114 = $113&255;
    $115 = $104 & 31;
    $116 = (($$pre143) + ($115)|0);
    $117 = HEAP8[$116>>0]|0;
    $118 = $117&255;
    $119 = (($109) + ($r$0))|0;
    $120 = ($119>>>0)>(255);
    $$ = $120 ? 255 : $119;
    $121 = (($114) + ($g$0))|0;
    $122 = ($121>>>0)>(255);
    $$8 = $122 ? 255 : $121;
    $123 = (($118) + ($b$0))|0;
    $124 = ($123>>>0)>(255);
    $$2 = $124 ? 255 : $123;
    $125 = $$ >>> 3;
    $126 = $125 << 10;
    $127 = $$8 >>> 3;
    $128 = $127 << 5;
    $129 = $128 | $126;
    $130 = $$2 >>> 3;
    $131 = $129 | $130;
    $132 = $131&65535;
    HEAP16[$pixel3$163>>1] = $132;
    $133 = ((($pixel3$163)) + 2|0);
    $134 = ($102|0)==(0);
    if ($134) {
     break;
    } else {
     $length1$064 = $102;$pixel3$163 = $133;
    }
   }
   return;
   break;
  }
  case 4:  {
   $135 = ((($dst)) + 16|0);
   $136 = HEAP32[$135>>2]|0;
   $137 = ((($dst)) + 4|0);
   $138 = HEAP32[$137>>2]|0;
   $139 = ((($138)) + 9|0);
   $140 = HEAP8[$139>>0]|0;
   $141 = $140&255;
   $142 = (($136|0) / ($141|0))&-1;
   $143 = ($x2|0)<($x1|0);
   $144 = ((($dst)) + 20|0);
   $145 = HEAP32[$144>>2]|0;
   $146 = Math_imul($142, $y2)|0;
   $147 = (($145) + ($146<<1)|0);
   if ($143) {
    $153 = (($147) + ($x2<<1)|0);
    $154 = ($draw_end|0)!=(0);
    $155 = ((($153)) + 2|0);
    $$15 = $154 ? $153 : $155;
    $156 = (($x1) - ($x2))|0;
    $157 = $154&1;
    $158 = (($156) + ($157))|0;
    $length8$0$ph = $158;$pixel10$1$ph = $$15;
   } else {
    $148 = (($147) + ($x1<<1)|0);
    $149 = ($draw_end|0)!=(0);
    $150 = (($x2) - ($x1))|0;
    $151 = $149&1;
    $152 = (($150) + ($151))|0;
    $length8$0$ph = $152;$pixel10$1$ph = $148;
   }
   $159 = ($length8$0$ph|0)==(0);
   if ($159) {
    return;
   }
   $$pre142 = HEAP32[(125936)>>2]|0;
   $length8$067 = $length8$0$ph;$pixel10$166 = $pixel10$1$ph;
   while(1) {
    $160 = (($length8$067) + -1)|0;
    $161 = HEAP16[$pixel10$166>>1]|0;
    $162 = $161&65535;
    $163 = $162 >>> 10;
    $164 = $163 & 31;
    $165 = (($$pre142) + ($164)|0);
    $166 = HEAP8[$165>>0]|0;
    $167 = $166&255;
    $168 = $162 >>> 5;
    $169 = $168 & 31;
    $170 = (($$pre142) + ($169)|0);
    $171 = HEAP8[$170>>0]|0;
    $172 = $171&255;
    $173 = $162 & 31;
    $174 = (($$pre142) + ($173)|0);
    $175 = HEAP8[$174>>0]|0;
    $176 = $175&255;
    $177 = Math_imul($167, $r$0)|0;
    $178 = (($177>>>0) / 255)&-1;
    $179 = Math_imul($172, $g$0)|0;
    $180 = (($179>>>0) / 255)&-1;
    $181 = Math_imul($176, $b$0)|0;
    $182 = (($181>>>0) / 255)&-1;
    $183 = $178 >>> 3;
    $184 = $183 << 10;
    $185 = $180 >>> 3;
    $186 = $185 << 5;
    $187 = $186 | $184;
    $188 = $182 >>> 3;
    $189 = $187 | $188;
    $190 = $189&65535;
    HEAP16[$pixel10$166>>1] = $190;
    $191 = ((($pixel10$166)) + 2|0);
    $192 = ($160|0)==(0);
    if ($192) {
     break;
    } else {
     $length8$067 = $160;$pixel10$166 = $191;
    }
   }
   return;
   break;
  }
  default: {
   $193 = ((($dst)) + 16|0);
   $194 = HEAP32[$193>>2]|0;
   $195 = ((($dst)) + 4|0);
   $196 = HEAP32[$195>>2]|0;
   $197 = ((($196)) + 9|0);
   $198 = HEAP8[$197>>0]|0;
   $199 = $198&255;
   $200 = (($194|0) / ($199|0))&-1;
   $201 = ($x2|0)<($x1|0);
   $202 = ((($dst)) + 20|0);
   $203 = HEAP32[$202>>2]|0;
   $204 = Math_imul($200, $y2)|0;
   $205 = (($203) + ($204<<1)|0);
   if ($201) {
    $211 = (($205) + ($x2<<1)|0);
    $212 = ($draw_end|0)!=(0);
    $213 = ((($211)) + 2|0);
    $$16 = $212 ? $211 : $213;
    $214 = (($x1) - ($x2))|0;
    $215 = $212&1;
    $216 = (($214) + ($215))|0;
    $length15$0$ph = $216;$pixel17$1$ph = $$16;
   } else {
    $206 = (($205) + ($x1<<1)|0);
    $207 = ($draw_end|0)!=(0);
    $208 = (($x2) - ($x1))|0;
    $209 = $207&1;
    $210 = (($208) + ($209))|0;
    $length15$0$ph = $210;$pixel17$1$ph = $206;
   }
   $217 = ($length15$0$ph|0)==(0);
   if ($217) {
    return;
   }
   $218 = $r$0 >>> 3;
   $219 = $218 << 10;
   $220 = $g$0 >>> 3;
   $221 = $220 << 5;
   $222 = $b$0 >>> 3;
   $223 = $221 | $222;
   $224 = $223 | $219;
   $225 = $224&65535;
   $length15$059 = $length15$0$ph;$pixel17$158 = $pixel17$1$ph;
   while(1) {
    $226 = (($length15$059) + -1)|0;
    HEAP16[$pixel17$158>>1] = $225;
    $227 = ((($pixel17$158)) + 2|0);
    $228 = ($226|0)==(0);
    if ($228) {
     break;
    } else {
     $length15$059 = $226;$pixel17$158 = $227;
    }
   }
   return;
  }
  }
 }
 $229 = ($x2|0)==($x1|0);
 if ($229) {
  switch ($blendMode|0) {
  case 1:  {
   $230 = ((($dst)) + 16|0);
   $231 = HEAP32[$230>>2]|0;
   $232 = ((($dst)) + 4|0);
   $233 = HEAP32[$232>>2]|0;
   $234 = ((($233)) + 9|0);
   $235 = HEAP8[$234>>0]|0;
   $236 = $235&255;
   $237 = (($231|0) / ($236|0))&-1;
   $238 = ($y2|0)<($y1|0);
   $239 = ((($dst)) + 20|0);
   $240 = HEAP32[$239>>2]|0;
   if ($238) {
    $248 = Math_imul($237, $y2)|0;
    $249 = (($240) + ($248<<1)|0);
    $250 = (($249) + ($x2<<1)|0);
    $251 = ($draw_end|0)!=(0);
    $252 = (($250) + ($237<<1)|0);
    $$17 = $251 ? $250 : $252;
    $253 = (($y1) - ($y2))|0;
    $254 = $251&1;
    $255 = (($253) + ($254))|0;
    $length22$0$ph = $255;$pixel24$1$ph = $$17;
   } else {
    $241 = Math_imul($237, $y1)|0;
    $242 = (($240) + ($241<<1)|0);
    $243 = (($242) + ($x2<<1)|0);
    $244 = ($draw_end|0)!=(0);
    $245 = (($y2) - ($y1))|0;
    $246 = $244&1;
    $247 = (($245) + ($246))|0;
    $length22$0$ph = $247;$pixel24$1$ph = $243;
   }
   $256 = ($length22$0$ph|0)==(0);
   if ($256) {
    return;
   }
   $$pre141 = HEAP32[(125936)>>2]|0;
   $length22$073 = $length22$0$ph;$pixel24$172 = $pixel24$1$ph;
   while(1) {
    $257 = (($length22$073) + -1)|0;
    $258 = HEAP16[$pixel24$172>>1]|0;
    $259 = $258&65535;
    $260 = $259 >>> 10;
    $261 = $260 & 31;
    $262 = (($$pre141) + ($261)|0);
    $263 = HEAP8[$262>>0]|0;
    $264 = $263&255;
    $265 = $259 >>> 5;
    $266 = $265 & 31;
    $267 = (($$pre141) + ($266)|0);
    $268 = HEAP8[$267>>0]|0;
    $269 = $268&255;
    $270 = $259 & 31;
    $271 = (($$pre141) + ($270)|0);
    $272 = HEAP8[$271>>0]|0;
    $273 = $272&255;
    $274 = Math_imul($264, $14)|0;
    $275 = (($274>>>0) / 255)&-1;
    $276 = (($275) + ($r$0))|0;
    $277 = Math_imul($269, $14)|0;
    $278 = (($277>>>0) / 255)&-1;
    $279 = (($278) + ($g$0))|0;
    $280 = Math_imul($273, $14)|0;
    $281 = (($280>>>0) / 255)&-1;
    $282 = (($281) + ($b$0))|0;
    $283 = $276 >>> 3;
    $284 = $283 << 10;
    $285 = $279 >>> 3;
    $286 = $285 << 5;
    $287 = $286 | $284;
    $288 = $282 >>> 3;
    $289 = $287 | $288;
    $290 = $289&65535;
    HEAP16[$pixel24$172>>1] = $290;
    $291 = (($pixel24$172) + ($237<<1)|0);
    $292 = ($257|0)==(0);
    if ($292) {
     break;
    } else {
     $length22$073 = $257;$pixel24$172 = $291;
    }
   }
   return;
   break;
  }
  case 2:  {
   $293 = ((($dst)) + 16|0);
   $294 = HEAP32[$293>>2]|0;
   $295 = ((($dst)) + 4|0);
   $296 = HEAP32[$295>>2]|0;
   $297 = ((($296)) + 9|0);
   $298 = HEAP8[$297>>0]|0;
   $299 = $298&255;
   $300 = (($294|0) / ($299|0))&-1;
   $301 = ($y2|0)<($y1|0);
   $302 = ((($dst)) + 20|0);
   $303 = HEAP32[$302>>2]|0;
   if ($301) {
    $311 = Math_imul($300, $y2)|0;
    $312 = (($303) + ($311<<1)|0);
    $313 = (($312) + ($x2<<1)|0);
    $314 = ($draw_end|0)!=(0);
    $315 = (($313) + ($300<<1)|0);
    $$18 = $314 ? $313 : $315;
    $316 = (($y1) - ($y2))|0;
    $317 = $314&1;
    $318 = (($316) + ($317))|0;
    $length29$0$ph = $318;$pixel31$1$ph = $$18;
   } else {
    $304 = Math_imul($300, $y1)|0;
    $305 = (($303) + ($304<<1)|0);
    $306 = (($305) + ($x2<<1)|0);
    $307 = ($draw_end|0)!=(0);
    $308 = (($y2) - ($y1))|0;
    $309 = $307&1;
    $310 = (($308) + ($309))|0;
    $length29$0$ph = $310;$pixel31$1$ph = $306;
   }
   $319 = ($length29$0$ph|0)==(0);
   if ($319) {
    return;
   }
   $$pre140 = HEAP32[(125936)>>2]|0;
   $length29$076 = $length29$0$ph;$pixel31$175 = $pixel31$1$ph;
   while(1) {
    $320 = (($length29$076) + -1)|0;
    $321 = HEAP16[$pixel31$175>>1]|0;
    $322 = $321&65535;
    $323 = $322 >>> 10;
    $324 = $323 & 31;
    $325 = (($$pre140) + ($324)|0);
    $326 = HEAP8[$325>>0]|0;
    $327 = $326&255;
    $328 = $322 >>> 5;
    $329 = $328 & 31;
    $330 = (($$pre140) + ($329)|0);
    $331 = HEAP8[$330>>0]|0;
    $332 = $331&255;
    $333 = $322 & 31;
    $334 = (($$pre140) + ($333)|0);
    $335 = HEAP8[$334>>0]|0;
    $336 = $335&255;
    $337 = (($327) + ($r$0))|0;
    $338 = ($337>>>0)>(255);
    $$3 = $338 ? 255 : $337;
    $339 = (($332) + ($g$0))|0;
    $340 = ($339>>>0)>(255);
    $$9 = $340 ? 255 : $339;
    $341 = (($336) + ($b$0))|0;
    $342 = ($341>>>0)>(255);
    $$4 = $342 ? 255 : $341;
    $343 = $$3 >>> 3;
    $344 = $343 << 10;
    $345 = $$9 >>> 3;
    $346 = $345 << 5;
    $347 = $346 | $344;
    $348 = $$4 >>> 3;
    $349 = $347 | $348;
    $350 = $349&65535;
    HEAP16[$pixel31$175>>1] = $350;
    $351 = (($pixel31$175) + ($300<<1)|0);
    $352 = ($320|0)==(0);
    if ($352) {
     break;
    } else {
     $length29$076 = $320;$pixel31$175 = $351;
    }
   }
   return;
   break;
  }
  case 4:  {
   $353 = ((($dst)) + 16|0);
   $354 = HEAP32[$353>>2]|0;
   $355 = ((($dst)) + 4|0);
   $356 = HEAP32[$355>>2]|0;
   $357 = ((($356)) + 9|0);
   $358 = HEAP8[$357>>0]|0;
   $359 = $358&255;
   $360 = (($354|0) / ($359|0))&-1;
   $361 = ($y2|0)<($y1|0);
   $362 = ((($dst)) + 20|0);
   $363 = HEAP32[$362>>2]|0;
   if ($361) {
    $371 = Math_imul($360, $y2)|0;
    $372 = (($363) + ($371<<1)|0);
    $373 = (($372) + ($x2<<1)|0);
    $374 = ($draw_end|0)!=(0);
    $375 = (($373) + ($360<<1)|0);
    $$19 = $374 ? $373 : $375;
    $376 = (($y1) - ($y2))|0;
    $377 = $374&1;
    $378 = (($376) + ($377))|0;
    $length36$0$ph = $378;$pixel38$1$ph = $$19;
   } else {
    $364 = Math_imul($360, $y1)|0;
    $365 = (($363) + ($364<<1)|0);
    $366 = (($365) + ($x2<<1)|0);
    $367 = ($draw_end|0)!=(0);
    $368 = (($y2) - ($y1))|0;
    $369 = $367&1;
    $370 = (($368) + ($369))|0;
    $length36$0$ph = $370;$pixel38$1$ph = $366;
   }
   $379 = ($length36$0$ph|0)==(0);
   if ($379) {
    return;
   }
   $$pre139 = HEAP32[(125936)>>2]|0;
   $length36$079 = $length36$0$ph;$pixel38$178 = $pixel38$1$ph;
   while(1) {
    $380 = (($length36$079) + -1)|0;
    $381 = HEAP16[$pixel38$178>>1]|0;
    $382 = $381&65535;
    $383 = $382 >>> 10;
    $384 = $383 & 31;
    $385 = (($$pre139) + ($384)|0);
    $386 = HEAP8[$385>>0]|0;
    $387 = $386&255;
    $388 = $382 >>> 5;
    $389 = $388 & 31;
    $390 = (($$pre139) + ($389)|0);
    $391 = HEAP8[$390>>0]|0;
    $392 = $391&255;
    $393 = $382 & 31;
    $394 = (($$pre139) + ($393)|0);
    $395 = HEAP8[$394>>0]|0;
    $396 = $395&255;
    $397 = Math_imul($387, $r$0)|0;
    $398 = (($397>>>0) / 255)&-1;
    $399 = Math_imul($392, $g$0)|0;
    $400 = (($399>>>0) / 255)&-1;
    $401 = Math_imul($396, $b$0)|0;
    $402 = (($401>>>0) / 255)&-1;
    $403 = $398 >>> 3;
    $404 = $403 << 10;
    $405 = $400 >>> 3;
    $406 = $405 << 5;
    $407 = $406 | $404;
    $408 = $402 >>> 3;
    $409 = $407 | $408;
    $410 = $409&65535;
    HEAP16[$pixel38$178>>1] = $410;
    $411 = (($pixel38$178) + ($360<<1)|0);
    $412 = ($380|0)==(0);
    if ($412) {
     break;
    } else {
     $length36$079 = $380;$pixel38$178 = $411;
    }
   }
   return;
   break;
  }
  default: {
   $413 = ((($dst)) + 16|0);
   $414 = HEAP32[$413>>2]|0;
   $415 = ((($dst)) + 4|0);
   $416 = HEAP32[$415>>2]|0;
   $417 = ((($416)) + 9|0);
   $418 = HEAP8[$417>>0]|0;
   $419 = $418&255;
   $420 = (($414|0) / ($419|0))&-1;
   $421 = ($y2|0)<($y1|0);
   $422 = ((($dst)) + 20|0);
   $423 = HEAP32[$422>>2]|0;
   if ($421) {
    $431 = Math_imul($420, $y2)|0;
    $432 = (($423) + ($431<<1)|0);
    $433 = (($432) + ($x2<<1)|0);
    $434 = ($draw_end|0)!=(0);
    $435 = (($433) + ($420<<1)|0);
    $$20 = $434 ? $433 : $435;
    $436 = (($y1) - ($y2))|0;
    $437 = $434&1;
    $438 = (($436) + ($437))|0;
    $length43$0$ph = $438;$pixel45$1$ph = $$20;
   } else {
    $424 = Math_imul($420, $y1)|0;
    $425 = (($423) + ($424<<1)|0);
    $426 = (($425) + ($x2<<1)|0);
    $427 = ($draw_end|0)!=(0);
    $428 = (($y2) - ($y1))|0;
    $429 = $427&1;
    $430 = (($428) + ($429))|0;
    $length43$0$ph = $430;$pixel45$1$ph = $426;
   }
   $439 = ($length43$0$ph|0)==(0);
   if ($439) {
    return;
   }
   $440 = $r$0 >>> 3;
   $441 = $440 << 10;
   $442 = $g$0 >>> 3;
   $443 = $442 << 5;
   $444 = $b$0 >>> 3;
   $445 = $443 | $444;
   $446 = $445 | $441;
   $447 = $446&65535;
   $length43$070 = $length43$0$ph;$pixel45$169 = $pixel45$1$ph;
   while(1) {
    $448 = (($length43$070) + -1)|0;
    HEAP16[$pixel45$169>>1] = $447;
    $449 = (($pixel45$169) + ($420<<1)|0);
    $450 = ($448|0)==(0);
    if ($450) {
     break;
    } else {
     $length43$070 = $448;$pixel45$169 = $449;
    }
   }
   return;
  }
  }
 }
 $451 = (($x1) - ($x2))|0;
 $452 = ($451|0)<(0);
 $453 = (0 - ($451))|0;
 $454 = $452 ? $453 : $451;
 $455 = (($y1) - ($y2))|0;
 $456 = ($455|0)<(0);
 $457 = (0 - ($455))|0;
 $458 = $456 ? $457 : $455;
 $459 = ($454|0)==($458|0);
 if ($459) {
  switch ($blendMode|0) {
  case 1:  {
   $460 = ((($dst)) + 16|0);
   $461 = HEAP32[$460>>2]|0;
   $462 = ((($dst)) + 4|0);
   $463 = HEAP32[$462>>2]|0;
   $464 = ((($463)) + 9|0);
   $465 = HEAP8[$464>>0]|0;
   $466 = $465&255;
   $467 = (($461|0) / ($466|0))&-1;
   $468 = ($y2|0)<($y1|0);
   $469 = ((($dst)) + 20|0);
   $470 = HEAP32[$469>>2]|0;
   if ($468) {
    $476 = Math_imul($467, $y2)|0;
    $477 = (($470) + ($476<<1)|0);
    $478 = (($477) + ($x2<<1)|0);
    $479 = ($x2|0)>($x1|0);
    $pitch51$1$v = $479 ? -1 : 1;
    $pitch51$1 = (($467) + ($pitch51$1$v))|0;
    $480 = ($draw_end|0)==(0);
    $481 = (($478) + ($pitch51$1<<1)|0);
    $$21 = $480 ? $481 : $478;
    $length50$0 = $455;$pitch51$2 = $pitch51$1;$pixel52$1 = $$21;
   } else {
    $471 = Math_imul($467, $y1)|0;
    $472 = (($470) + ($471<<1)|0);
    $473 = (($472) + ($x1<<1)|0);
    $474 = ($x2|0)<($x1|0);
    $pitch51$0$v = $474 ? -1 : 1;
    $pitch51$0 = (($467) + ($pitch51$0$v))|0;
    $475 = (($y2) - ($y1))|0;
    $length50$0 = $475;$pitch51$2 = $pitch51$0;$pixel52$1 = $473;
   }
   $not$34 = ($draw_end|0)!=(0);
   $482 = $not$34&1;
   $length50$0$ = (($length50$0) + ($482))|0;
   $483 = ($length50$0$|0)==(0);
   if ($483) {
    return;
   }
   $$pre138 = HEAP32[(125936)>>2]|0;
   $length50$185 = $length50$0$;$pixel52$284 = $pixel52$1;
   while(1) {
    $484 = (($length50$185) + -1)|0;
    $485 = HEAP16[$pixel52$284>>1]|0;
    $486 = $485&65535;
    $487 = $486 >>> 10;
    $488 = $487 & 31;
    $489 = (($$pre138) + ($488)|0);
    $490 = HEAP8[$489>>0]|0;
    $491 = $490&255;
    $492 = $486 >>> 5;
    $493 = $492 & 31;
    $494 = (($$pre138) + ($493)|0);
    $495 = HEAP8[$494>>0]|0;
    $496 = $495&255;
    $497 = $486 & 31;
    $498 = (($$pre138) + ($497)|0);
    $499 = HEAP8[$498>>0]|0;
    $500 = $499&255;
    $501 = Math_imul($491, $14)|0;
    $502 = (($501>>>0) / 255)&-1;
    $503 = (($502) + ($r$0))|0;
    $504 = Math_imul($496, $14)|0;
    $505 = (($504>>>0) / 255)&-1;
    $506 = (($505) + ($g$0))|0;
    $507 = Math_imul($500, $14)|0;
    $508 = (($507>>>0) / 255)&-1;
    $509 = (($508) + ($b$0))|0;
    $510 = $503 >>> 3;
    $511 = $510 << 10;
    $512 = $506 >>> 3;
    $513 = $512 << 5;
    $514 = $513 | $511;
    $515 = $509 >>> 3;
    $516 = $514 | $515;
    $517 = $516&65535;
    HEAP16[$pixel52$284>>1] = $517;
    $518 = (($pixel52$284) + ($pitch51$2<<1)|0);
    $519 = ($484|0)==(0);
    if ($519) {
     break;
    } else {
     $length50$185 = $484;$pixel52$284 = $518;
    }
   }
   return;
   break;
  }
  case 2:  {
   $520 = ((($dst)) + 16|0);
   $521 = HEAP32[$520>>2]|0;
   $522 = ((($dst)) + 4|0);
   $523 = HEAP32[$522>>2]|0;
   $524 = ((($523)) + 9|0);
   $525 = HEAP8[$524>>0]|0;
   $526 = $525&255;
   $527 = (($521|0) / ($526|0))&-1;
   $528 = ($y2|0)<($y1|0);
   $529 = ((($dst)) + 20|0);
   $530 = HEAP32[$529>>2]|0;
   if ($528) {
    $536 = Math_imul($527, $y2)|0;
    $537 = (($530) + ($536<<1)|0);
    $538 = (($537) + ($x2<<1)|0);
    $539 = ($x2|0)>($x1|0);
    $pitch58$1$v = $539 ? -1 : 1;
    $pitch58$1 = (($527) + ($pitch58$1$v))|0;
    $540 = ($draw_end|0)==(0);
    $541 = (($538) + ($pitch58$1<<1)|0);
    $$22 = $540 ? $541 : $538;
    $length57$0 = $455;$pitch58$2 = $pitch58$1;$pixel59$1 = $$22;
   } else {
    $531 = Math_imul($527, $y1)|0;
    $532 = (($530) + ($531<<1)|0);
    $533 = (($532) + ($x1<<1)|0);
    $534 = ($x2|0)<($x1|0);
    $pitch58$0$v = $534 ? -1 : 1;
    $pitch58$0 = (($527) + ($pitch58$0$v))|0;
    $535 = (($y2) - ($y1))|0;
    $length57$0 = $535;$pitch58$2 = $pitch58$0;$pixel59$1 = $533;
   }
   $not$33 = ($draw_end|0)!=(0);
   $542 = $not$33&1;
   $length57$0$ = (($length57$0) + ($542))|0;
   $543 = ($length57$0$|0)==(0);
   if ($543) {
    return;
   }
   $$pre137 = HEAP32[(125936)>>2]|0;
   $length57$188 = $length57$0$;$pixel59$287 = $pixel59$1;
   while(1) {
    $544 = (($length57$188) + -1)|0;
    $545 = HEAP16[$pixel59$287>>1]|0;
    $546 = $545&65535;
    $547 = $546 >>> 10;
    $548 = $547 & 31;
    $549 = (($$pre137) + ($548)|0);
    $550 = HEAP8[$549>>0]|0;
    $551 = $550&255;
    $552 = $546 >>> 5;
    $553 = $552 & 31;
    $554 = (($$pre137) + ($553)|0);
    $555 = HEAP8[$554>>0]|0;
    $556 = $555&255;
    $557 = $546 & 31;
    $558 = (($$pre137) + ($557)|0);
    $559 = HEAP8[$558>>0]|0;
    $560 = $559&255;
    $561 = (($551) + ($r$0))|0;
    $562 = ($561>>>0)>(255);
    $$5 = $562 ? 255 : $561;
    $563 = (($556) + ($g$0))|0;
    $564 = ($563>>>0)>(255);
    $$10 = $564 ? 255 : $563;
    $565 = (($560) + ($b$0))|0;
    $566 = ($565>>>0)>(255);
    $$6 = $566 ? 255 : $565;
    $567 = $$5 >>> 3;
    $568 = $567 << 10;
    $569 = $$10 >>> 3;
    $570 = $569 << 5;
    $571 = $570 | $568;
    $572 = $$6 >>> 3;
    $573 = $571 | $572;
    $574 = $573&65535;
    HEAP16[$pixel59$287>>1] = $574;
    $575 = (($pixel59$287) + ($pitch58$2<<1)|0);
    $576 = ($544|0)==(0);
    if ($576) {
     break;
    } else {
     $length57$188 = $544;$pixel59$287 = $575;
    }
   }
   return;
   break;
  }
  case 4:  {
   $577 = ((($dst)) + 16|0);
   $578 = HEAP32[$577>>2]|0;
   $579 = ((($dst)) + 4|0);
   $580 = HEAP32[$579>>2]|0;
   $581 = ((($580)) + 9|0);
   $582 = HEAP8[$581>>0]|0;
   $583 = $582&255;
   $584 = (($578|0) / ($583|0))&-1;
   $585 = ($y2|0)<($y1|0);
   $586 = ((($dst)) + 20|0);
   $587 = HEAP32[$586>>2]|0;
   if ($585) {
    $593 = Math_imul($584, $y2)|0;
    $594 = (($587) + ($593<<1)|0);
    $595 = (($594) + ($x2<<1)|0);
    $596 = ($x2|0)>($x1|0);
    $pitch65$1$v = $596 ? -1 : 1;
    $pitch65$1 = (($584) + ($pitch65$1$v))|0;
    $597 = ($draw_end|0)==(0);
    $598 = (($595) + ($pitch65$1<<1)|0);
    $$23 = $597 ? $598 : $595;
    $length64$0 = $455;$pitch65$2 = $pitch65$1;$pixel66$1 = $$23;
   } else {
    $588 = Math_imul($584, $y1)|0;
    $589 = (($587) + ($588<<1)|0);
    $590 = (($589) + ($x1<<1)|0);
    $591 = ($x2|0)<($x1|0);
    $pitch65$0$v = $591 ? -1 : 1;
    $pitch65$0 = (($584) + ($pitch65$0$v))|0;
    $592 = (($y2) - ($y1))|0;
    $length64$0 = $592;$pitch65$2 = $pitch65$0;$pixel66$1 = $590;
   }
   $not$32 = ($draw_end|0)!=(0);
   $599 = $not$32&1;
   $length64$0$ = (($length64$0) + ($599))|0;
   $600 = ($length64$0$|0)==(0);
   if ($600) {
    return;
   }
   $$pre136 = HEAP32[(125936)>>2]|0;
   $length64$191 = $length64$0$;$pixel66$290 = $pixel66$1;
   while(1) {
    $601 = (($length64$191) + -1)|0;
    $602 = HEAP16[$pixel66$290>>1]|0;
    $603 = $602&65535;
    $604 = $603 >>> 10;
    $605 = $604 & 31;
    $606 = (($$pre136) + ($605)|0);
    $607 = HEAP8[$606>>0]|0;
    $608 = $607&255;
    $609 = $603 >>> 5;
    $610 = $609 & 31;
    $611 = (($$pre136) + ($610)|0);
    $612 = HEAP8[$611>>0]|0;
    $613 = $612&255;
    $614 = $603 & 31;
    $615 = (($$pre136) + ($614)|0);
    $616 = HEAP8[$615>>0]|0;
    $617 = $616&255;
    $618 = Math_imul($608, $r$0)|0;
    $619 = (($618>>>0) / 255)&-1;
    $620 = Math_imul($613, $g$0)|0;
    $621 = (($620>>>0) / 255)&-1;
    $622 = Math_imul($617, $b$0)|0;
    $623 = (($622>>>0) / 255)&-1;
    $624 = $619 >>> 3;
    $625 = $624 << 10;
    $626 = $621 >>> 3;
    $627 = $626 << 5;
    $628 = $627 | $625;
    $629 = $623 >>> 3;
    $630 = $628 | $629;
    $631 = $630&65535;
    HEAP16[$pixel66$290>>1] = $631;
    $632 = (($pixel66$290) + ($pitch65$2<<1)|0);
    $633 = ($601|0)==(0);
    if ($633) {
     break;
    } else {
     $length64$191 = $601;$pixel66$290 = $632;
    }
   }
   return;
   break;
  }
  default: {
   $634 = ((($dst)) + 16|0);
   $635 = HEAP32[$634>>2]|0;
   $636 = ((($dst)) + 4|0);
   $637 = HEAP32[$636>>2]|0;
   $638 = ((($637)) + 9|0);
   $639 = HEAP8[$638>>0]|0;
   $640 = $639&255;
   $641 = (($635|0) / ($640|0))&-1;
   $642 = ($y2|0)<($y1|0);
   $643 = ((($dst)) + 20|0);
   $644 = HEAP32[$643>>2]|0;
   if ($642) {
    $650 = Math_imul($641, $y2)|0;
    $651 = (($644) + ($650<<1)|0);
    $652 = (($651) + ($x2<<1)|0);
    $653 = ($x2|0)>($x1|0);
    $pitch72$1$v = $653 ? -1 : 1;
    $pitch72$1 = (($641) + ($pitch72$1$v))|0;
    $654 = ($draw_end|0)==(0);
    $655 = (($652) + ($pitch72$1<<1)|0);
    $$24 = $654 ? $655 : $652;
    $length71$0 = $455;$pitch72$2 = $pitch72$1;$pixel73$1 = $$24;
   } else {
    $645 = Math_imul($641, $y1)|0;
    $646 = (($644) + ($645<<1)|0);
    $647 = (($646) + ($x1<<1)|0);
    $648 = ($x2|0)<($x1|0);
    $pitch72$0$v = $648 ? -1 : 1;
    $pitch72$0 = (($641) + ($pitch72$0$v))|0;
    $649 = (($y2) - ($y1))|0;
    $length71$0 = $649;$pitch72$2 = $pitch72$0;$pixel73$1 = $647;
   }
   $not$35 = ($draw_end|0)!=(0);
   $656 = $not$35&1;
   $length71$0$ = (($length71$0) + ($656))|0;
   $657 = ($length71$0$|0)==(0);
   if ($657) {
    return;
   }
   $658 = $r$0 >>> 3;
   $659 = $658 << 10;
   $660 = $g$0 >>> 3;
   $661 = $660 << 5;
   $662 = $b$0 >>> 3;
   $663 = $661 | $662;
   $664 = $663 | $659;
   $665 = $664&65535;
   $length71$182 = $length71$0$;$pixel73$281 = $pixel73$1;
   while(1) {
    $666 = (($length71$182) + -1)|0;
    HEAP16[$pixel73$281>>1] = $665;
    $667 = (($pixel73$281) + ($pitch72$2<<1)|0);
    $668 = ($666|0)==(0);
    if ($668) {
     break;
    } else {
     $length71$182 = $666;$pixel73$281 = $667;
    }
   }
   return;
  }
  }
 } else {
  switch ($blendMode|0) {
  case 1:  {
   $669 = (($x2) - ($x1))|0;
   $670 = ($669|0)<(0);
   $671 = (0 - ($669))|0;
   $672 = $670 ? $671 : $669;
   $673 = (($y2) - ($y1))|0;
   $674 = ($673|0)<(0);
   $675 = (0 - ($673))|0;
   $676 = $674 ? $675 : $673;
   $677 = ($672|0)<($676|0);
   if ($677) {
    $681 = $672 << 1;
    $682 = (($681) - ($676))|0;
    $683 = (($672) - ($676))|0;
    $d$0 = $682;$dinc1$0 = $681;$dinc2$0$in = $683;$numpixels$0$in = $676;$xinc1$0 = 0;$yinc1$0 = 1;
   } else {
    $678 = $676 << 1;
    $679 = (($678) - ($672))|0;
    $680 = (($676) - ($672))|0;
    $d$0 = $679;$dinc1$0 = $678;$dinc2$0$in = $680;$numpixels$0$in = $672;$xinc1$0 = 1;$yinc1$0 = 0;
   }
   $dinc2$0 = $dinc2$0$in << 1;
   $684 = ($x2|0)<($x1|0);
   $685 = (0 - ($xinc1$0))|0;
   $$xinc1$0 = $684 ? $685 : $xinc1$0;
   $$25 = $684 ? -1 : 1;
   $686 = ($y2|0)<($y1|0);
   $687 = (0 - ($yinc1$0))|0;
   $yinc1$1 = $686 ? $687 : $yinc1$0;
   $yinc2$1 = $686 ? -1 : 1;
   $not$30 = ($draw_end|0)!=(0);
   $numpixels$0 = $not$30&1;
   $numpixels$0$in$numpixels$0 = (($numpixels$0$in) + ($numpixels$0))|0;
   $688 = ($numpixels$0$in$numpixels$0|0)>(0);
   if (!($688)) {
    return;
   }
   $689 = ((($dst)) + 20|0);
   $690 = ((($dst)) + 16|0);
   $$pre133 = HEAP32[$689>>2]|0;
   $$pre134 = HEAP32[$690>>2]|0;
   $$pre135 = HEAP32[(125936)>>2]|0;
   $d$199 = $d$0;$i$098 = 0;$x$0100 = $x1;$y$0101 = $y1;
   while(1) {
    $691 = Math_imul($$pre134, $y$0101)|0;
    $692 = (($$pre133) + ($691)|0);
    $693 = $x$0100 << 1;
    $694 = (($692) + ($693)|0);
    $695 = HEAP16[$694>>1]|0;
    $696 = $695&65535;
    $697 = $696 >>> 10;
    $698 = $697 & 31;
    $699 = (($$pre135) + ($698)|0);
    $700 = HEAP8[$699>>0]|0;
    $701 = $700&255;
    $702 = $696 >>> 5;
    $703 = $702 & 31;
    $704 = (($$pre135) + ($703)|0);
    $705 = HEAP8[$704>>0]|0;
    $706 = $705&255;
    $707 = $696 & 31;
    $708 = (($$pre135) + ($707)|0);
    $709 = HEAP8[$708>>0]|0;
    $710 = $709&255;
    $711 = Math_imul($701, $14)|0;
    $712 = (($711>>>0) / 255)&-1;
    $713 = (($712) + ($r$0))|0;
    $714 = Math_imul($706, $14)|0;
    $715 = (($714>>>0) / 255)&-1;
    $716 = (($715) + ($g$0))|0;
    $717 = Math_imul($710, $14)|0;
    $718 = (($717>>>0) / 255)&-1;
    $719 = (($718) + ($b$0))|0;
    $720 = $713 >>> 3;
    $721 = $720 << 10;
    $722 = $716 >>> 3;
    $723 = $722 << 5;
    $724 = $723 | $721;
    $725 = $719 >>> 3;
    $726 = $724 | $725;
    $727 = $726&65535;
    HEAP16[$694>>1] = $727;
    $728 = ($d$199|0)<(0);
    $dinc1$0$dinc2$0 = $728 ? $dinc1$0 : $dinc2$0;
    $$xinc1$0$$25 = $728 ? $$xinc1$0 : $$25;
    $yinc1$1$yinc2$1 = $728 ? $yinc1$1 : $yinc2$1;
    $y$1 = (($y$0101) + ($yinc1$1$yinc2$1))|0;
    $x$1 = (($$xinc1$0$$25) + ($x$0100))|0;
    $d$2 = (($dinc1$0$dinc2$0) + ($d$199))|0;
    $729 = (($i$098) + 1)|0;
    $730 = ($729|0)<($numpixels$0$in$numpixels$0|0);
    if ($730) {
     $d$199 = $d$2;$i$098 = $729;$x$0100 = $x$1;$y$0101 = $y$1;
    } else {
     break;
    }
   }
   return;
   break;
  }
  case 2:  {
   $731 = (($x2) - ($x1))|0;
   $732 = ($731|0)<(0);
   $733 = (0 - ($731))|0;
   $734 = $732 ? $733 : $731;
   $735 = (($y2) - ($y1))|0;
   $736 = ($735|0)<(0);
   $737 = (0 - ($735))|0;
   $738 = $736 ? $737 : $735;
   $739 = ($734|0)<($738|0);
   if ($739) {
    $743 = $734 << 1;
    $744 = (($743) - ($738))|0;
    $745 = (($734) - ($738))|0;
    $d87$0 = $744;$dinc188$0 = $743;$dinc289$0$in = $745;$numpixels86$0$in = $738;$xinc191$0 = 0;$yinc194$0 = 1;
   } else {
    $740 = $738 << 1;
    $741 = (($740) - ($734))|0;
    $742 = (($738) - ($734))|0;
    $d87$0 = $741;$dinc188$0 = $740;$dinc289$0$in = $742;$numpixels86$0$in = $734;$xinc191$0 = 1;$yinc194$0 = 0;
   }
   $dinc289$0 = $dinc289$0$in << 1;
   $746 = ($x2|0)<($x1|0);
   $747 = (0 - ($xinc191$0))|0;
   $$xinc191$0 = $746 ? $747 : $xinc191$0;
   $$26 = $746 ? -1 : 1;
   $748 = ($y2|0)<($y1|0);
   $749 = (0 - ($yinc194$0))|0;
   $yinc194$1 = $748 ? $749 : $yinc194$0;
   $yinc295$1 = $748 ? -1 : 1;
   $not$29 = ($draw_end|0)!=(0);
   $numpixels86$0 = $not$29&1;
   $numpixels86$0$in$numpixels86$0 = (($numpixels86$0$in) + ($numpixels86$0))|0;
   $750 = ($numpixels86$0$in$numpixels86$0|0)>(0);
   if (!($750)) {
    return;
   }
   $751 = ((($dst)) + 20|0);
   $752 = ((($dst)) + 16|0);
   $$pre130 = HEAP32[$751>>2]|0;
   $$pre131 = HEAP32[$752>>2]|0;
   $$pre132 = HEAP32[(125936)>>2]|0;
   $d87$1104 = $d87$0;$i83$0103 = 0;$x90$0105 = $x1;$y93$0106 = $y1;
   while(1) {
    $753 = Math_imul($$pre131, $y93$0106)|0;
    $754 = (($$pre130) + ($753)|0);
    $755 = $x90$0105 << 1;
    $756 = (($754) + ($755)|0);
    $757 = HEAP16[$756>>1]|0;
    $758 = $757&65535;
    $759 = $758 >>> 10;
    $760 = $759 & 31;
    $761 = (($$pre132) + ($760)|0);
    $762 = HEAP8[$761>>0]|0;
    $763 = $762&255;
    $764 = $758 >>> 5;
    $765 = $764 & 31;
    $766 = (($$pre132) + ($765)|0);
    $767 = HEAP8[$766>>0]|0;
    $768 = $767&255;
    $769 = $758 & 31;
    $770 = (($$pre132) + ($769)|0);
    $771 = HEAP8[$770>>0]|0;
    $772 = $771&255;
    $773 = (($763) + ($r$0))|0;
    $774 = ($773>>>0)>(255);
    $$11 = $774 ? 255 : $773;
    $775 = (($768) + ($g$0))|0;
    $776 = ($775>>>0)>(255);
    $$7 = $776 ? 255 : $775;
    $777 = (($772) + ($b$0))|0;
    $778 = ($777>>>0)>(255);
    $$12 = $778 ? 255 : $777;
    $779 = $$11 >>> 3;
    $780 = $779 << 10;
    $781 = $$7 >>> 3;
    $782 = $781 << 5;
    $783 = $782 | $780;
    $784 = $$12 >>> 3;
    $785 = $783 | $784;
    $786 = $785&65535;
    HEAP16[$756>>1] = $786;
    $787 = ($d87$1104|0)<(0);
    $dinc188$0$dinc289$0 = $787 ? $dinc188$0 : $dinc289$0;
    $$xinc191$0$$26 = $787 ? $$xinc191$0 : $$26;
    $yinc194$1$yinc295$1 = $787 ? $yinc194$1 : $yinc295$1;
    $y93$1 = (($y93$0106) + ($yinc194$1$yinc295$1))|0;
    $x90$1 = (($$xinc191$0$$26) + ($x90$0105))|0;
    $d87$2 = (($dinc188$0$dinc289$0) + ($d87$1104))|0;
    $788 = (($i83$0103) + 1)|0;
    $789 = ($788|0)<($numpixels86$0$in$numpixels86$0|0);
    if ($789) {
     $d87$1104 = $d87$2;$i83$0103 = $788;$x90$0105 = $x90$1;$y93$0106 = $y93$1;
    } else {
     break;
    }
   }
   return;
   break;
  }
  case 4:  {
   $790 = (($x2) - ($x1))|0;
   $791 = ($790|0)<(0);
   $792 = (0 - ($790))|0;
   $793 = $791 ? $792 : $790;
   $794 = (($y2) - ($y1))|0;
   $795 = ($794|0)<(0);
   $796 = (0 - ($794))|0;
   $797 = $795 ? $796 : $794;
   $798 = ($793|0)<($797|0);
   if ($798) {
    $802 = $793 << 1;
    $803 = (($802) - ($797))|0;
    $804 = (($793) - ($797))|0;
    $d105$0 = $803;$dinc1106$0 = $802;$dinc2107$0$in = $804;$numpixels104$0$in = $797;$xinc1109$0 = 0;$yinc1112$0 = 1;
   } else {
    $799 = $797 << 1;
    $800 = (($799) - ($793))|0;
    $801 = (($797) - ($793))|0;
    $d105$0 = $800;$dinc1106$0 = $799;$dinc2107$0$in = $801;$numpixels104$0$in = $793;$xinc1109$0 = 1;$yinc1112$0 = 0;
   }
   $dinc2107$0 = $dinc2107$0$in << 1;
   $805 = ($x2|0)<($x1|0);
   $806 = (0 - ($xinc1109$0))|0;
   $$xinc1109$0 = $805 ? $806 : $xinc1109$0;
   $$27 = $805 ? -1 : 1;
   $807 = ($y2|0)<($y1|0);
   $808 = (0 - ($yinc1112$0))|0;
   $yinc1112$1 = $807 ? $808 : $yinc1112$0;
   $yinc2113$1 = $807 ? -1 : 1;
   $not$ = ($draw_end|0)!=(0);
   $numpixels104$0 = $not$&1;
   $numpixels104$0$in$numpixels104$0 = (($numpixels104$0$in) + ($numpixels104$0))|0;
   $809 = ($numpixels104$0$in$numpixels104$0|0)>(0);
   if (!($809)) {
    return;
   }
   $810 = ((($dst)) + 20|0);
   $811 = ((($dst)) + 16|0);
   $$pre = HEAP32[$810>>2]|0;
   $$pre128 = HEAP32[$811>>2]|0;
   $$pre129 = HEAP32[(125936)>>2]|0;
   $d105$1109 = $d105$0;$i101$0108 = 0;$x108$0110 = $x1;$y111$0111 = $y1;
   while(1) {
    $812 = Math_imul($$pre128, $y111$0111)|0;
    $813 = (($$pre) + ($812)|0);
    $814 = $x108$0110 << 1;
    $815 = (($813) + ($814)|0);
    $816 = HEAP16[$815>>1]|0;
    $817 = $816&65535;
    $818 = $817 >>> 10;
    $819 = $818 & 31;
    $820 = (($$pre129) + ($819)|0);
    $821 = HEAP8[$820>>0]|0;
    $822 = $821&255;
    $823 = $817 >>> 5;
    $824 = $823 & 31;
    $825 = (($$pre129) + ($824)|0);
    $826 = HEAP8[$825>>0]|0;
    $827 = $826&255;
    $828 = $817 & 31;
    $829 = (($$pre129) + ($828)|0);
    $830 = HEAP8[$829>>0]|0;
    $831 = $830&255;
    $832 = Math_imul($822, $r$0)|0;
    $833 = (($832>>>0) / 255)&-1;
    $834 = Math_imul($827, $g$0)|0;
    $835 = (($834>>>0) / 255)&-1;
    $836 = Math_imul($831, $b$0)|0;
    $837 = (($836>>>0) / 255)&-1;
    $838 = $833 >>> 3;
    $839 = $838 << 10;
    $840 = $835 >>> 3;
    $841 = $840 << 5;
    $842 = $841 | $839;
    $843 = $837 >>> 3;
    $844 = $842 | $843;
    $845 = $844&65535;
    HEAP16[$815>>1] = $845;
    $846 = ($d105$1109|0)<(0);
    $dinc1106$0$dinc2107$0 = $846 ? $dinc1106$0 : $dinc2107$0;
    $$xinc1109$0$$27 = $846 ? $$xinc1109$0 : $$27;
    $yinc1112$1$yinc2113$1 = $846 ? $yinc1112$1 : $yinc2113$1;
    $y111$1 = (($y111$0111) + ($yinc1112$1$yinc2113$1))|0;
    $x108$1 = (($$xinc1109$0$$27) + ($x108$0110))|0;
    $d105$2 = (($dinc1106$0$dinc2107$0) + ($d105$1109))|0;
    $847 = (($i101$0108) + 1)|0;
    $848 = ($847|0)<($numpixels104$0$in$numpixels104$0|0);
    if ($848) {
     $d105$1109 = $d105$2;$i101$0108 = $847;$x108$0110 = $x108$1;$y111$0111 = $y111$1;
    } else {
     break;
    }
   }
   return;
   break;
  }
  default: {
   $849 = (($x2) - ($x1))|0;
   $850 = ($849|0)<(0);
   $851 = (0 - ($849))|0;
   $852 = $850 ? $851 : $849;
   $853 = (($y2) - ($y1))|0;
   $854 = ($853|0)<(0);
   $855 = (0 - ($853))|0;
   $856 = $854 ? $855 : $853;
   $857 = ($852|0)<($856|0);
   if ($857) {
    $861 = $852 << 1;
    $862 = (($861) - ($856))|0;
    $863 = (($852) - ($856))|0;
    $d123$0 = $862;$dinc1124$0 = $861;$dinc2125$0$in = $863;$numpixels122$0$in = $856;$xinc1127$0 = 0;$yinc1130$0 = 1;
   } else {
    $858 = $856 << 1;
    $859 = (($858) - ($852))|0;
    $860 = (($856) - ($852))|0;
    $d123$0 = $859;$dinc1124$0 = $858;$dinc2125$0$in = $860;$numpixels122$0$in = $852;$xinc1127$0 = 1;$yinc1130$0 = 0;
   }
   $dinc2125$0 = $dinc2125$0$in << 1;
   $864 = ($x2|0)<($x1|0);
   $865 = (0 - ($xinc1127$0))|0;
   $$xinc1127$0 = $864 ? $865 : $xinc1127$0;
   $$28 = $864 ? -1 : 1;
   $866 = ($y2|0)<($y1|0);
   $867 = (0 - ($yinc1130$0))|0;
   $yinc1130$1 = $866 ? $867 : $yinc1130$0;
   $yinc2131$1 = $866 ? -1 : 1;
   $not$31 = ($draw_end|0)!=(0);
   $numpixels122$0 = $not$31&1;
   $numpixels122$0$in$numpixels122$0 = (($numpixels122$0$in) + ($numpixels122$0))|0;
   $868 = ($numpixels122$0$in$numpixels122$0|0)>(0);
   if (!($868)) {
    return;
   }
   $869 = ((($dst)) + 20|0);
   $870 = HEAP32[$869>>2]|0;
   $871 = ((($dst)) + 16|0);
   $872 = HEAP32[$871>>2]|0;
   $873 = $r$0 >>> 3;
   $874 = $873 << 10;
   $875 = $g$0 >>> 3;
   $876 = $875 << 5;
   $877 = $b$0 >>> 3;
   $878 = $876 | $877;
   $879 = $878 | $874;
   $880 = $879&65535;
   $d123$194 = $d123$0;$i119$093 = 0;$x126$095 = $x1;$y129$096 = $y1;
   while(1) {
    $881 = Math_imul($872, $y129$096)|0;
    $882 = (($870) + ($881)|0);
    $883 = $x126$095 << 1;
    $884 = (($882) + ($883)|0);
    HEAP16[$884>>1] = $880;
    $885 = ($d123$194|0)<(0);
    $dinc1124$0$dinc2125$0 = $885 ? $dinc1124$0 : $dinc2125$0;
    $$xinc1127$0$$28 = $885 ? $$xinc1127$0 : $$28;
    $yinc1130$1$yinc2131$1 = $885 ? $yinc1130$1 : $yinc2131$1;
    $y129$1 = (($y129$096) + ($yinc1130$1$yinc2131$1))|0;
    $x126$1 = (($$xinc1127$0$$28) + ($x126$095))|0;
    $d123$2 = (($dinc1124$0$dinc2125$0) + ($d123$194))|0;
    $886 = (($i119$093) + 1)|0;
    $887 = ($886|0)<($numpixels122$0$in$numpixels122$0|0);
    if ($887) {
     $d123$194 = $d123$2;$i119$093 = $886;$x126$095 = $x126$1;$y129$096 = $y129$1;
    } else {
     break;
    }
   }
   return;
  }
  }
 }
}
function _SDL_BlendLine_ARGB8888($dst,$x1,$y1,$x2,$y2,$blendMode,$_r,$_g,$_b,$_a,$draw_end) {
 $dst = $dst|0;
 $x1 = $x1|0;
 $y1 = $y1|0;
 $x2 = $x2|0;
 $y2 = $y2|0;
 $blendMode = $blendMode|0;
 $_r = $_r|0;
 $_g = $_g|0;
 $_b = $_b|0;
 $_a = $_a|0;
 $draw_end = $draw_end|0;
 var $$ = 0, $$10 = 0, $$11 = 0, $$12 = 0, $$13 = 0, $$14 = 0, $$15 = 0, $$16 = 0, $$17 = 0, $$18 = 0, $$19 = 0, $$2 = 0, $$20 = 0, $$21 = 0, $$22 = 0, $$23 = 0, $$24 = 0, $$25 = 0, $$26 = 0, $$27 = 0;
 var $$28 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$6 = 0, $$7 = 0, $$8 = 0, $$9 = 0, $$xinc1$0 = 0, $$xinc1$0$$25 = 0, $$xinc1109$0 = 0, $$xinc1109$0$$27 = 0, $$xinc1127$0 = 0, $$xinc1127$0$$28 = 0, $$xinc191$0 = 0, $$xinc191$0$$26 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0;
 var $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0;
 var $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0;
 var $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0;
 var $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0;
 var $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0;
 var $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0;
 var $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0;
 var $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0;
 var $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0;
 var $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0;
 var $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0;
 var $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0;
 var $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0;
 var $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0;
 var $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0;
 var $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0;
 var $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0;
 var $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0;
 var $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0;
 var $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0;
 var $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0;
 var $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0;
 var $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0;
 var $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0;
 var $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0;
 var $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0;
 var $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0;
 var $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0;
 var $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $a$0 = 0, $b$0 = 0, $blendMode$off = 0, $d$0 = 0, $d$199 = 0, $d$2 = 0;
 var $d105$0 = 0, $d105$1109 = 0, $d105$2 = 0, $d123$0 = 0, $d123$194 = 0, $d123$2 = 0, $d87$0 = 0, $d87$1104 = 0, $d87$2 = 0, $dinc1$0 = 0, $dinc1$0$dinc2$0 = 0, $dinc1106$0 = 0, $dinc1106$0$dinc2107$0 = 0, $dinc1124$0 = 0, $dinc1124$0$dinc2125$0 = 0, $dinc188$0 = 0, $dinc188$0$dinc289$0 = 0, $dinc2$0 = 0, $dinc2$0$in = 0, $dinc2107$0 = 0;
 var $dinc2107$0$in = 0, $dinc2125$0 = 0, $dinc2125$0$in = 0, $dinc289$0 = 0, $dinc289$0$in = 0, $g$0 = 0, $i$098 = 0, $i101$0108 = 0, $i119$093 = 0, $i83$0103 = 0, $length$0$ph = 0, $length$061 = 0, $length1$0$ph = 0, $length1$064 = 0, $length15$0$ph = 0, $length15$059 = 0, $length22$0$ph = 0, $length22$073 = 0, $length29$0$ph = 0, $length29$076 = 0;
 var $length36$0$ph = 0, $length36$079 = 0, $length43$0$ph = 0, $length43$070 = 0, $length50$0 = 0, $length50$0$ = 0, $length50$185 = 0, $length57$0 = 0, $length57$0$ = 0, $length57$188 = 0, $length64$0 = 0, $length64$0$ = 0, $length64$191 = 0, $length71$0 = 0, $length71$0$ = 0, $length71$181 = 0, $length8$0$ph = 0, $length8$067 = 0, $not$ = 0, $not$29 = 0;
 var $not$30 = 0, $not$31 = 0, $not$32 = 0, $not$33 = 0, $not$34 = 0, $not$35 = 0, $numpixels$0 = 0, $numpixels$0$in = 0, $numpixels$0$in$numpixels$0 = 0, $numpixels104$0 = 0, $numpixels104$0$in = 0, $numpixels104$0$in$numpixels104$0 = 0, $numpixels122$0 = 0, $numpixels122$0$in = 0, $numpixels122$0$in$numpixels122$0 = 0, $numpixels86$0 = 0, $numpixels86$0$in = 0, $numpixels86$0$in$numpixels86$0 = 0, $pitch51$0 = 0, $pitch51$0$v = 0;
 var $pitch51$1 = 0, $pitch51$1$v = 0, $pitch51$2 = 0, $pitch58$0 = 0, $pitch58$0$v = 0, $pitch58$1 = 0, $pitch58$1$v = 0, $pitch58$2 = 0, $pitch65$0 = 0, $pitch65$0$v = 0, $pitch65$1 = 0, $pitch65$1$v = 0, $pitch65$2 = 0, $pitch72$0 = 0, $pitch72$0$v = 0, $pitch72$1 = 0, $pitch72$1$v = 0, $pitch72$2 = 0, $pixel$1$ph = 0, $pixel$160 = 0;
 var $pixel10$1$ph = 0, $pixel10$166 = 0, $pixel17$1$ph = 0, $pixel17$158 = 0, $pixel24$1$ph = 0, $pixel24$172 = 0, $pixel3$1$ph = 0, $pixel3$163 = 0, $pixel31$1$ph = 0, $pixel31$175 = 0, $pixel38$1$ph = 0, $pixel38$178 = 0, $pixel45$1$ph = 0, $pixel45$169 = 0, $pixel52$1 = 0, $pixel52$284 = 0, $pixel59$1 = 0, $pixel59$287 = 0, $pixel66$1 = 0, $pixel66$290 = 0;
 var $pixel73$1 = 0, $pixel73$282 = 0, $r$0 = 0, $x$0100 = 0, $x$1 = 0, $x108$0110 = 0, $x108$1 = 0, $x126$095 = 0, $x126$1 = 0, $x90$0105 = 0, $x90$1 = 0, $xinc1$0 = 0, $xinc1109$0 = 0, $xinc1127$0 = 0, $xinc191$0 = 0, $y$0101 = 0, $y$1 = 0, $y111$0111 = 0, $y111$1 = 0, $y129$096 = 0;
 var $y129$1 = 0, $y93$0106 = 0, $y93$1 = 0, $yinc1$0 = 0, $yinc1$1 = 0, $yinc1$1$yinc2$1 = 0, $yinc1112$0 = 0, $yinc1112$1 = 0, $yinc1112$1$yinc2113$1 = 0, $yinc1130$0 = 0, $yinc1130$1 = 0, $yinc1130$1$yinc2131$1 = 0, $yinc194$0 = 0, $yinc194$1 = 0, $yinc194$1$yinc295$1 = 0, $yinc2$1 = 0, $yinc2113$1 = 0, $yinc2131$1 = 0, $yinc295$1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $blendMode$off = (($blendMode) + -1)|0;
 $0 = ($blendMode$off>>>0)<(2);
 $1 = $_r&255;
 if ($0) {
  $2 = $_a&255;
  $3 = Math_imul($2, $1)|0;
  $4 = (($3>>>0) / 255)&-1;
  $5 = $_g&255;
  $6 = Math_imul($2, $5)|0;
  $7 = (($6>>>0) / 255)&-1;
  $8 = $_b&255;
  $9 = Math_imul($2, $8)|0;
  $10 = (($9>>>0) / 255)&-1;
  $a$0 = $2;$b$0 = $10;$g$0 = $7;$r$0 = $4;
 } else {
  $11 = $_g&255;
  $12 = $_b&255;
  $13 = $_a&255;
  $a$0 = $13;$b$0 = $12;$g$0 = $11;$r$0 = $1;
 }
 $14 = $a$0 ^ 255;
 $15 = ($y2|0)==($y1|0);
 if ($15) {
  switch ($blendMode|0) {
  case 1:  {
   $16 = ((($dst)) + 16|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($dst)) + 4|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ((($19)) + 9|0);
   $21 = HEAP8[$20>>0]|0;
   $22 = $21&255;
   $23 = (($17|0) / ($22|0))&-1;
   $24 = ($x2|0)<($x1|0);
   $25 = ((($dst)) + 20|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = Math_imul($23, $y2)|0;
   $28 = (($26) + ($27<<2)|0);
   if ($24) {
    $34 = (($28) + ($x2<<2)|0);
    $35 = ($draw_end|0)!=(0);
    $36 = ((($34)) + 4|0);
    $$13 = $35 ? $34 : $36;
    $37 = (($x1) - ($x2))|0;
    $38 = $35&1;
    $39 = (($37) + ($38))|0;
    $length$0$ph = $39;$pixel$1$ph = $$13;
   } else {
    $29 = (($28) + ($x1<<2)|0);
    $30 = ($draw_end|0)!=(0);
    $31 = (($x2) - ($x1))|0;
    $32 = $30&1;
    $33 = (($31) + ($32))|0;
    $length$0$ph = $33;$pixel$1$ph = $29;
   }
   $40 = ($length$0$ph|0)==(0);
   if ($40) {
    return;
   } else {
    $length$061 = $length$0$ph;$pixel$160 = $pixel$1$ph;
   }
   while(1) {
    $41 = (($length$061) + -1)|0;
    $42 = HEAP32[$pixel$160>>2]|0;
    $43 = $42 >>> 16;
    $44 = $43 & 255;
    $45 = $42 >>> 8;
    $46 = $45 & 255;
    $47 = $42 & 255;
    $48 = $42 >>> 24;
    $49 = Math_imul($44, $14)|0;
    $50 = (($49>>>0) / 255)&-1;
    $51 = (($50) + ($r$0))|0;
    $52 = Math_imul($46, $14)|0;
    $53 = (($52>>>0) / 255)&-1;
    $54 = (($53) + ($g$0))|0;
    $55 = Math_imul($47, $14)|0;
    $56 = (($55>>>0) / 255)&-1;
    $57 = (($56) + ($b$0))|0;
    $58 = Math_imul($48, $14)|0;
    $59 = (($58>>>0) / 255)&-1;
    $60 = (($59) + ($a$0))|0;
    $61 = $60 << 24;
    $62 = $51 << 16;
    $63 = $54 << 8;
    $64 = $61 | $57;
    $65 = $64 | $62;
    $66 = $65 | $63;
    HEAP32[$pixel$160>>2] = $66;
    $67 = ((($pixel$160)) + 4|0);
    $68 = ($41|0)==(0);
    if ($68) {
     break;
    } else {
     $length$061 = $41;$pixel$160 = $67;
    }
   }
   return;
   break;
  }
  case 2:  {
   $69 = ((($dst)) + 16|0);
   $70 = HEAP32[$69>>2]|0;
   $71 = ((($dst)) + 4|0);
   $72 = HEAP32[$71>>2]|0;
   $73 = ((($72)) + 9|0);
   $74 = HEAP8[$73>>0]|0;
   $75 = $74&255;
   $76 = (($70|0) / ($75|0))&-1;
   $77 = ($x2|0)<($x1|0);
   $78 = ((($dst)) + 20|0);
   $79 = HEAP32[$78>>2]|0;
   $80 = Math_imul($76, $y2)|0;
   $81 = (($79) + ($80<<2)|0);
   if ($77) {
    $87 = (($81) + ($x2<<2)|0);
    $88 = ($draw_end|0)!=(0);
    $89 = ((($87)) + 4|0);
    $$14 = $88 ? $87 : $89;
    $90 = (($x1) - ($x2))|0;
    $91 = $88&1;
    $92 = (($90) + ($91))|0;
    $length1$0$ph = $92;$pixel3$1$ph = $$14;
   } else {
    $82 = (($81) + ($x1<<2)|0);
    $83 = ($draw_end|0)!=(0);
    $84 = (($x2) - ($x1))|0;
    $85 = $83&1;
    $86 = (($84) + ($85))|0;
    $length1$0$ph = $86;$pixel3$1$ph = $82;
   }
   $93 = ($length1$0$ph|0)==(0);
   if ($93) {
    return;
   } else {
    $length1$064 = $length1$0$ph;$pixel3$163 = $pixel3$1$ph;
   }
   while(1) {
    $94 = (($length1$064) + -1)|0;
    $95 = HEAP32[$pixel3$163>>2]|0;
    $96 = $95 >>> 16;
    $97 = $96 & 255;
    $98 = $95 >>> 8;
    $99 = $98 & 255;
    $100 = $95 & 255;
    $101 = $95 & -16777216;
    $102 = (($97) + ($r$0))|0;
    $103 = ($102>>>0)>(255);
    $$ = $103 ? 255 : $102;
    $104 = (($99) + ($g$0))|0;
    $105 = ($104>>>0)>(255);
    $$8 = $105 ? 255 : $104;
    $106 = (($100) + ($b$0))|0;
    $107 = ($106>>>0)>(255);
    $$2 = $107 ? 255 : $106;
    $108 = $$ << 16;
    $109 = $$8 << 8;
    $110 = $$2 | $101;
    $111 = $110 | $108;
    $112 = $111 | $109;
    HEAP32[$pixel3$163>>2] = $112;
    $113 = ((($pixel3$163)) + 4|0);
    $114 = ($94|0)==(0);
    if ($114) {
     break;
    } else {
     $length1$064 = $94;$pixel3$163 = $113;
    }
   }
   return;
   break;
  }
  case 4:  {
   $115 = ((($dst)) + 16|0);
   $116 = HEAP32[$115>>2]|0;
   $117 = ((($dst)) + 4|0);
   $118 = HEAP32[$117>>2]|0;
   $119 = ((($118)) + 9|0);
   $120 = HEAP8[$119>>0]|0;
   $121 = $120&255;
   $122 = (($116|0) / ($121|0))&-1;
   $123 = ($x2|0)<($x1|0);
   $124 = ((($dst)) + 20|0);
   $125 = HEAP32[$124>>2]|0;
   $126 = Math_imul($122, $y2)|0;
   $127 = (($125) + ($126<<2)|0);
   if ($123) {
    $133 = (($127) + ($x2<<2)|0);
    $134 = ($draw_end|0)!=(0);
    $135 = ((($133)) + 4|0);
    $$15 = $134 ? $133 : $135;
    $136 = (($x1) - ($x2))|0;
    $137 = $134&1;
    $138 = (($136) + ($137))|0;
    $length8$0$ph = $138;$pixel10$1$ph = $$15;
   } else {
    $128 = (($127) + ($x1<<2)|0);
    $129 = ($draw_end|0)!=(0);
    $130 = (($x2) - ($x1))|0;
    $131 = $129&1;
    $132 = (($130) + ($131))|0;
    $length8$0$ph = $132;$pixel10$1$ph = $128;
   }
   $139 = ($length8$0$ph|0)==(0);
   if ($139) {
    return;
   } else {
    $length8$067 = $length8$0$ph;$pixel10$166 = $pixel10$1$ph;
   }
   while(1) {
    $140 = (($length8$067) + -1)|0;
    $141 = HEAP32[$pixel10$166>>2]|0;
    $142 = $141 >>> 16;
    $143 = $142 & 255;
    $144 = $141 >>> 8;
    $145 = $144 & 255;
    $146 = $141 & 255;
    $147 = $141 & -16777216;
    $148 = Math_imul($143, $r$0)|0;
    $149 = (($148>>>0) / 255)&-1;
    $150 = Math_imul($145, $g$0)|0;
    $151 = (($150>>>0) / 255)&-1;
    $152 = Math_imul($146, $b$0)|0;
    $153 = (($152>>>0) / 255)&-1;
    $154 = $149 << 16;
    $155 = $151 << 8;
    $156 = $153 | $147;
    $157 = $156 | $154;
    $158 = $157 | $155;
    HEAP32[$pixel10$166>>2] = $158;
    $159 = ((($pixel10$166)) + 4|0);
    $160 = ($140|0)==(0);
    if ($160) {
     break;
    } else {
     $length8$067 = $140;$pixel10$166 = $159;
    }
   }
   return;
   break;
  }
  default: {
   $161 = ((($dst)) + 16|0);
   $162 = HEAP32[$161>>2]|0;
   $163 = ((($dst)) + 4|0);
   $164 = HEAP32[$163>>2]|0;
   $165 = ((($164)) + 9|0);
   $166 = HEAP8[$165>>0]|0;
   $167 = $166&255;
   $168 = (($162|0) / ($167|0))&-1;
   $169 = ($x2|0)<($x1|0);
   $170 = ((($dst)) + 20|0);
   $171 = HEAP32[$170>>2]|0;
   $172 = Math_imul($168, $y2)|0;
   $173 = (($171) + ($172<<2)|0);
   if ($169) {
    $179 = (($173) + ($x2<<2)|0);
    $180 = ($draw_end|0)!=(0);
    $181 = ((($179)) + 4|0);
    $$16 = $180 ? $179 : $181;
    $182 = (($x1) - ($x2))|0;
    $183 = $180&1;
    $184 = (($182) + ($183))|0;
    $length15$0$ph = $184;$pixel17$1$ph = $$16;
   } else {
    $174 = (($173) + ($x1<<2)|0);
    $175 = ($draw_end|0)!=(0);
    $176 = (($x2) - ($x1))|0;
    $177 = $175&1;
    $178 = (($176) + ($177))|0;
    $length15$0$ph = $178;$pixel17$1$ph = $174;
   }
   $185 = ($length15$0$ph|0)==(0);
   if ($185) {
    return;
   }
   $186 = $a$0 << 24;
   $187 = $r$0 << 16;
   $188 = $g$0 << 8;
   $189 = $b$0 | $186;
   $190 = $189 | $188;
   $191 = $190 | $187;
   $length15$059 = $length15$0$ph;$pixel17$158 = $pixel17$1$ph;
   while(1) {
    $192 = (($length15$059) + -1)|0;
    HEAP32[$pixel17$158>>2] = $191;
    $193 = ((($pixel17$158)) + 4|0);
    $194 = ($192|0)==(0);
    if ($194) {
     break;
    } else {
     $length15$059 = $192;$pixel17$158 = $193;
    }
   }
   return;
  }
  }
 }
 $195 = ($x2|0)==($x1|0);
 if ($195) {
  switch ($blendMode|0) {
  case 1:  {
   $196 = ((($dst)) + 16|0);
   $197 = HEAP32[$196>>2]|0;
   $198 = ((($dst)) + 4|0);
   $199 = HEAP32[$198>>2]|0;
   $200 = ((($199)) + 9|0);
   $201 = HEAP8[$200>>0]|0;
   $202 = $201&255;
   $203 = (($197|0) / ($202|0))&-1;
   $204 = ($y2|0)<($y1|0);
   $205 = ((($dst)) + 20|0);
   $206 = HEAP32[$205>>2]|0;
   if ($204) {
    $214 = Math_imul($203, $y2)|0;
    $215 = (($206) + ($214<<2)|0);
    $216 = (($215) + ($x2<<2)|0);
    $217 = ($draw_end|0)!=(0);
    $218 = (($216) + ($203<<2)|0);
    $$17 = $217 ? $216 : $218;
    $219 = (($y1) - ($y2))|0;
    $220 = $217&1;
    $221 = (($219) + ($220))|0;
    $length22$0$ph = $221;$pixel24$1$ph = $$17;
   } else {
    $207 = Math_imul($203, $y1)|0;
    $208 = (($206) + ($207<<2)|0);
    $209 = (($208) + ($x2<<2)|0);
    $210 = ($draw_end|0)!=(0);
    $211 = (($y2) - ($y1))|0;
    $212 = $210&1;
    $213 = (($211) + ($212))|0;
    $length22$0$ph = $213;$pixel24$1$ph = $209;
   }
   $222 = ($length22$0$ph|0)==(0);
   if ($222) {
    return;
   } else {
    $length22$073 = $length22$0$ph;$pixel24$172 = $pixel24$1$ph;
   }
   while(1) {
    $223 = (($length22$073) + -1)|0;
    $224 = HEAP32[$pixel24$172>>2]|0;
    $225 = $224 >>> 16;
    $226 = $225 & 255;
    $227 = $224 >>> 8;
    $228 = $227 & 255;
    $229 = $224 & 255;
    $230 = $224 >>> 24;
    $231 = Math_imul($226, $14)|0;
    $232 = (($231>>>0) / 255)&-1;
    $233 = (($232) + ($r$0))|0;
    $234 = Math_imul($228, $14)|0;
    $235 = (($234>>>0) / 255)&-1;
    $236 = (($235) + ($g$0))|0;
    $237 = Math_imul($229, $14)|0;
    $238 = (($237>>>0) / 255)&-1;
    $239 = (($238) + ($b$0))|0;
    $240 = Math_imul($230, $14)|0;
    $241 = (($240>>>0) / 255)&-1;
    $242 = (($241) + ($a$0))|0;
    $243 = $242 << 24;
    $244 = $233 << 16;
    $245 = $236 << 8;
    $246 = $243 | $239;
    $247 = $246 | $244;
    $248 = $247 | $245;
    HEAP32[$pixel24$172>>2] = $248;
    $249 = (($pixel24$172) + ($203<<2)|0);
    $250 = ($223|0)==(0);
    if ($250) {
     break;
    } else {
     $length22$073 = $223;$pixel24$172 = $249;
    }
   }
   return;
   break;
  }
  case 2:  {
   $251 = ((($dst)) + 16|0);
   $252 = HEAP32[$251>>2]|0;
   $253 = ((($dst)) + 4|0);
   $254 = HEAP32[$253>>2]|0;
   $255 = ((($254)) + 9|0);
   $256 = HEAP8[$255>>0]|0;
   $257 = $256&255;
   $258 = (($252|0) / ($257|0))&-1;
   $259 = ($y2|0)<($y1|0);
   $260 = ((($dst)) + 20|0);
   $261 = HEAP32[$260>>2]|0;
   if ($259) {
    $269 = Math_imul($258, $y2)|0;
    $270 = (($261) + ($269<<2)|0);
    $271 = (($270) + ($x2<<2)|0);
    $272 = ($draw_end|0)!=(0);
    $273 = (($271) + ($258<<2)|0);
    $$18 = $272 ? $271 : $273;
    $274 = (($y1) - ($y2))|0;
    $275 = $272&1;
    $276 = (($274) + ($275))|0;
    $length29$0$ph = $276;$pixel31$1$ph = $$18;
   } else {
    $262 = Math_imul($258, $y1)|0;
    $263 = (($261) + ($262<<2)|0);
    $264 = (($263) + ($x2<<2)|0);
    $265 = ($draw_end|0)!=(0);
    $266 = (($y2) - ($y1))|0;
    $267 = $265&1;
    $268 = (($266) + ($267))|0;
    $length29$0$ph = $268;$pixel31$1$ph = $264;
   }
   $277 = ($length29$0$ph|0)==(0);
   if ($277) {
    return;
   } else {
    $length29$076 = $length29$0$ph;$pixel31$175 = $pixel31$1$ph;
   }
   while(1) {
    $278 = (($length29$076) + -1)|0;
    $279 = HEAP32[$pixel31$175>>2]|0;
    $280 = $279 >>> 16;
    $281 = $280 & 255;
    $282 = $279 >>> 8;
    $283 = $282 & 255;
    $284 = $279 & 255;
    $285 = $279 & -16777216;
    $286 = (($281) + ($r$0))|0;
    $287 = ($286>>>0)>(255);
    $$3 = $287 ? 255 : $286;
    $288 = (($283) + ($g$0))|0;
    $289 = ($288>>>0)>(255);
    $$9 = $289 ? 255 : $288;
    $290 = (($284) + ($b$0))|0;
    $291 = ($290>>>0)>(255);
    $$4 = $291 ? 255 : $290;
    $292 = $$3 << 16;
    $293 = $$9 << 8;
    $294 = $$4 | $285;
    $295 = $294 | $292;
    $296 = $295 | $293;
    HEAP32[$pixel31$175>>2] = $296;
    $297 = (($pixel31$175) + ($258<<2)|0);
    $298 = ($278|0)==(0);
    if ($298) {
     break;
    } else {
     $length29$076 = $278;$pixel31$175 = $297;
    }
   }
   return;
   break;
  }
  case 4:  {
   $299 = ((($dst)) + 16|0);
   $300 = HEAP32[$299>>2]|0;
   $301 = ((($dst)) + 4|0);
   $302 = HEAP32[$301>>2]|0;
   $303 = ((($302)) + 9|0);
   $304 = HEAP8[$303>>0]|0;
   $305 = $304&255;
   $306 = (($300|0) / ($305|0))&-1;
   $307 = ($y2|0)<($y1|0);
   $308 = ((($dst)) + 20|0);
   $309 = HEAP32[$308>>2]|0;
   if ($307) {
    $317 = Math_imul($306, $y2)|0;
    $318 = (($309) + ($317<<2)|0);
    $319 = (($318) + ($x2<<2)|0);
    $320 = ($draw_end|0)!=(0);
    $321 = (($319) + ($306<<2)|0);
    $$19 = $320 ? $319 : $321;
    $322 = (($y1) - ($y2))|0;
    $323 = $320&1;
    $324 = (($322) + ($323))|0;
    $length36$0$ph = $324;$pixel38$1$ph = $$19;
   } else {
    $310 = Math_imul($306, $y1)|0;
    $311 = (($309) + ($310<<2)|0);
    $312 = (($311) + ($x2<<2)|0);
    $313 = ($draw_end|0)!=(0);
    $314 = (($y2) - ($y1))|0;
    $315 = $313&1;
    $316 = (($314) + ($315))|0;
    $length36$0$ph = $316;$pixel38$1$ph = $312;
   }
   $325 = ($length36$0$ph|0)==(0);
   if ($325) {
    return;
   } else {
    $length36$079 = $length36$0$ph;$pixel38$178 = $pixel38$1$ph;
   }
   while(1) {
    $326 = (($length36$079) + -1)|0;
    $327 = HEAP32[$pixel38$178>>2]|0;
    $328 = $327 >>> 16;
    $329 = $328 & 255;
    $330 = $327 >>> 8;
    $331 = $330 & 255;
    $332 = $327 & 255;
    $333 = $327 & -16777216;
    $334 = Math_imul($329, $r$0)|0;
    $335 = (($334>>>0) / 255)&-1;
    $336 = Math_imul($331, $g$0)|0;
    $337 = (($336>>>0) / 255)&-1;
    $338 = Math_imul($332, $b$0)|0;
    $339 = (($338>>>0) / 255)&-1;
    $340 = $335 << 16;
    $341 = $337 << 8;
    $342 = $339 | $333;
    $343 = $342 | $340;
    $344 = $343 | $341;
    HEAP32[$pixel38$178>>2] = $344;
    $345 = (($pixel38$178) + ($306<<2)|0);
    $346 = ($326|0)==(0);
    if ($346) {
     break;
    } else {
     $length36$079 = $326;$pixel38$178 = $345;
    }
   }
   return;
   break;
  }
  default: {
   $347 = ((($dst)) + 16|0);
   $348 = HEAP32[$347>>2]|0;
   $349 = ((($dst)) + 4|0);
   $350 = HEAP32[$349>>2]|0;
   $351 = ((($350)) + 9|0);
   $352 = HEAP8[$351>>0]|0;
   $353 = $352&255;
   $354 = (($348|0) / ($353|0))&-1;
   $355 = ($y2|0)<($y1|0);
   $356 = ((($dst)) + 20|0);
   $357 = HEAP32[$356>>2]|0;
   if ($355) {
    $365 = Math_imul($354, $y2)|0;
    $366 = (($357) + ($365<<2)|0);
    $367 = (($366) + ($x2<<2)|0);
    $368 = ($draw_end|0)!=(0);
    $369 = (($367) + ($354<<2)|0);
    $$20 = $368 ? $367 : $369;
    $370 = (($y1) - ($y2))|0;
    $371 = $368&1;
    $372 = (($370) + ($371))|0;
    $length43$0$ph = $372;$pixel45$1$ph = $$20;
   } else {
    $358 = Math_imul($354, $y1)|0;
    $359 = (($357) + ($358<<2)|0);
    $360 = (($359) + ($x2<<2)|0);
    $361 = ($draw_end|0)!=(0);
    $362 = (($y2) - ($y1))|0;
    $363 = $361&1;
    $364 = (($362) + ($363))|0;
    $length43$0$ph = $364;$pixel45$1$ph = $360;
   }
   $373 = ($length43$0$ph|0)==(0);
   if ($373) {
    return;
   }
   $374 = $a$0 << 24;
   $375 = $r$0 << 16;
   $376 = $g$0 << 8;
   $377 = $b$0 | $374;
   $378 = $377 | $376;
   $379 = $378 | $375;
   $length43$070 = $length43$0$ph;$pixel45$169 = $pixel45$1$ph;
   while(1) {
    $380 = (($length43$070) + -1)|0;
    HEAP32[$pixel45$169>>2] = $379;
    $381 = (($pixel45$169) + ($354<<2)|0);
    $382 = ($380|0)==(0);
    if ($382) {
     break;
    } else {
     $length43$070 = $380;$pixel45$169 = $381;
    }
   }
   return;
  }
  }
 }
 $383 = (($x1) - ($x2))|0;
 $384 = ($383|0)<(0);
 $385 = (0 - ($383))|0;
 $386 = $384 ? $385 : $383;
 $387 = (($y1) - ($y2))|0;
 $388 = ($387|0)<(0);
 $389 = (0 - ($387))|0;
 $390 = $388 ? $389 : $387;
 $391 = ($386|0)==($390|0);
 if ($391) {
  switch ($blendMode|0) {
  case 1:  {
   $392 = ((($dst)) + 16|0);
   $393 = HEAP32[$392>>2]|0;
   $394 = ((($dst)) + 4|0);
   $395 = HEAP32[$394>>2]|0;
   $396 = ((($395)) + 9|0);
   $397 = HEAP8[$396>>0]|0;
   $398 = $397&255;
   $399 = (($393|0) / ($398|0))&-1;
   $400 = ($y2|0)<($y1|0);
   $401 = ((($dst)) + 20|0);
   $402 = HEAP32[$401>>2]|0;
   if ($400) {
    $408 = Math_imul($399, $y2)|0;
    $409 = (($402) + ($408<<2)|0);
    $410 = (($409) + ($x2<<2)|0);
    $411 = ($x2|0)>($x1|0);
    $pitch51$1$v = $411 ? -1 : 1;
    $pitch51$1 = (($399) + ($pitch51$1$v))|0;
    $412 = ($draw_end|0)==(0);
    $413 = (($410) + ($pitch51$1<<2)|0);
    $$21 = $412 ? $413 : $410;
    $length50$0 = $387;$pitch51$2 = $pitch51$1;$pixel52$1 = $$21;
   } else {
    $403 = Math_imul($399, $y1)|0;
    $404 = (($402) + ($403<<2)|0);
    $405 = (($404) + ($x1<<2)|0);
    $406 = ($x2|0)<($x1|0);
    $pitch51$0$v = $406 ? -1 : 1;
    $pitch51$0 = (($399) + ($pitch51$0$v))|0;
    $407 = (($y2) - ($y1))|0;
    $length50$0 = $407;$pitch51$2 = $pitch51$0;$pixel52$1 = $405;
   }
   $not$34 = ($draw_end|0)!=(0);
   $414 = $not$34&1;
   $length50$0$ = (($length50$0) + ($414))|0;
   $415 = ($length50$0$|0)==(0);
   if ($415) {
    return;
   } else {
    $length50$185 = $length50$0$;$pixel52$284 = $pixel52$1;
   }
   while(1) {
    $416 = (($length50$185) + -1)|0;
    $417 = HEAP32[$pixel52$284>>2]|0;
    $418 = $417 >>> 16;
    $419 = $418 & 255;
    $420 = $417 >>> 8;
    $421 = $420 & 255;
    $422 = $417 & 255;
    $423 = $417 >>> 24;
    $424 = Math_imul($419, $14)|0;
    $425 = (($424>>>0) / 255)&-1;
    $426 = (($425) + ($r$0))|0;
    $427 = Math_imul($421, $14)|0;
    $428 = (($427>>>0) / 255)&-1;
    $429 = (($428) + ($g$0))|0;
    $430 = Math_imul($422, $14)|0;
    $431 = (($430>>>0) / 255)&-1;
    $432 = (($431) + ($b$0))|0;
    $433 = Math_imul($423, $14)|0;
    $434 = (($433>>>0) / 255)&-1;
    $435 = (($434) + ($a$0))|0;
    $436 = $435 << 24;
    $437 = $426 << 16;
    $438 = $429 << 8;
    $439 = $436 | $432;
    $440 = $439 | $437;
    $441 = $440 | $438;
    HEAP32[$pixel52$284>>2] = $441;
    $442 = (($pixel52$284) + ($pitch51$2<<2)|0);
    $443 = ($416|0)==(0);
    if ($443) {
     break;
    } else {
     $length50$185 = $416;$pixel52$284 = $442;
    }
   }
   return;
   break;
  }
  case 2:  {
   $444 = ((($dst)) + 16|0);
   $445 = HEAP32[$444>>2]|0;
   $446 = ((($dst)) + 4|0);
   $447 = HEAP32[$446>>2]|0;
   $448 = ((($447)) + 9|0);
   $449 = HEAP8[$448>>0]|0;
   $450 = $449&255;
   $451 = (($445|0) / ($450|0))&-1;
   $452 = ($y2|0)<($y1|0);
   $453 = ((($dst)) + 20|0);
   $454 = HEAP32[$453>>2]|0;
   if ($452) {
    $460 = Math_imul($451, $y2)|0;
    $461 = (($454) + ($460<<2)|0);
    $462 = (($461) + ($x2<<2)|0);
    $463 = ($x2|0)>($x1|0);
    $pitch58$1$v = $463 ? -1 : 1;
    $pitch58$1 = (($451) + ($pitch58$1$v))|0;
    $464 = ($draw_end|0)==(0);
    $465 = (($462) + ($pitch58$1<<2)|0);
    $$22 = $464 ? $465 : $462;
    $length57$0 = $387;$pitch58$2 = $pitch58$1;$pixel59$1 = $$22;
   } else {
    $455 = Math_imul($451, $y1)|0;
    $456 = (($454) + ($455<<2)|0);
    $457 = (($456) + ($x1<<2)|0);
    $458 = ($x2|0)<($x1|0);
    $pitch58$0$v = $458 ? -1 : 1;
    $pitch58$0 = (($451) + ($pitch58$0$v))|0;
    $459 = (($y2) - ($y1))|0;
    $length57$0 = $459;$pitch58$2 = $pitch58$0;$pixel59$1 = $457;
   }
   $not$33 = ($draw_end|0)!=(0);
   $466 = $not$33&1;
   $length57$0$ = (($length57$0) + ($466))|0;
   $467 = ($length57$0$|0)==(0);
   if ($467) {
    return;
   } else {
    $length57$188 = $length57$0$;$pixel59$287 = $pixel59$1;
   }
   while(1) {
    $468 = (($length57$188) + -1)|0;
    $469 = HEAP32[$pixel59$287>>2]|0;
    $470 = $469 >>> 16;
    $471 = $470 & 255;
    $472 = $469 >>> 8;
    $473 = $472 & 255;
    $474 = $469 & 255;
    $475 = $469 & -16777216;
    $476 = (($471) + ($r$0))|0;
    $477 = ($476>>>0)>(255);
    $$5 = $477 ? 255 : $476;
    $478 = (($473) + ($g$0))|0;
    $479 = ($478>>>0)>(255);
    $$10 = $479 ? 255 : $478;
    $480 = (($474) + ($b$0))|0;
    $481 = ($480>>>0)>(255);
    $$6 = $481 ? 255 : $480;
    $482 = $$5 << 16;
    $483 = $$10 << 8;
    $484 = $$6 | $475;
    $485 = $484 | $482;
    $486 = $485 | $483;
    HEAP32[$pixel59$287>>2] = $486;
    $487 = (($pixel59$287) + ($pitch58$2<<2)|0);
    $488 = ($468|0)==(0);
    if ($488) {
     break;
    } else {
     $length57$188 = $468;$pixel59$287 = $487;
    }
   }
   return;
   break;
  }
  case 4:  {
   $489 = ((($dst)) + 16|0);
   $490 = HEAP32[$489>>2]|0;
   $491 = ((($dst)) + 4|0);
   $492 = HEAP32[$491>>2]|0;
   $493 = ((($492)) + 9|0);
   $494 = HEAP8[$493>>0]|0;
   $495 = $494&255;
   $496 = (($490|0) / ($495|0))&-1;
   $497 = ($y2|0)<($y1|0);
   $498 = ((($dst)) + 20|0);
   $499 = HEAP32[$498>>2]|0;
   if ($497) {
    $505 = Math_imul($496, $y2)|0;
    $506 = (($499) + ($505<<2)|0);
    $507 = (($506) + ($x2<<2)|0);
    $508 = ($x2|0)>($x1|0);
    $pitch65$1$v = $508 ? -1 : 1;
    $pitch65$1 = (($496) + ($pitch65$1$v))|0;
    $509 = ($draw_end|0)==(0);
    $510 = (($507) + ($pitch65$1<<2)|0);
    $$23 = $509 ? $510 : $507;
    $length64$0 = $387;$pitch65$2 = $pitch65$1;$pixel66$1 = $$23;
   } else {
    $500 = Math_imul($496, $y1)|0;
    $501 = (($499) + ($500<<2)|0);
    $502 = (($501) + ($x1<<2)|0);
    $503 = ($x2|0)<($x1|0);
    $pitch65$0$v = $503 ? -1 : 1;
    $pitch65$0 = (($496) + ($pitch65$0$v))|0;
    $504 = (($y2) - ($y1))|0;
    $length64$0 = $504;$pitch65$2 = $pitch65$0;$pixel66$1 = $502;
   }
   $not$32 = ($draw_end|0)!=(0);
   $511 = $not$32&1;
   $length64$0$ = (($length64$0) + ($511))|0;
   $512 = ($length64$0$|0)==(0);
   if ($512) {
    return;
   } else {
    $length64$191 = $length64$0$;$pixel66$290 = $pixel66$1;
   }
   while(1) {
    $513 = (($length64$191) + -1)|0;
    $514 = HEAP32[$pixel66$290>>2]|0;
    $515 = $514 >>> 16;
    $516 = $515 & 255;
    $517 = $514 >>> 8;
    $518 = $517 & 255;
    $519 = $514 & 255;
    $520 = $514 & -16777216;
    $521 = Math_imul($516, $r$0)|0;
    $522 = (($521>>>0) / 255)&-1;
    $523 = Math_imul($518, $g$0)|0;
    $524 = (($523>>>0) / 255)&-1;
    $525 = Math_imul($519, $b$0)|0;
    $526 = (($525>>>0) / 255)&-1;
    $527 = $522 << 16;
    $528 = $524 << 8;
    $529 = $526 | $520;
    $530 = $529 | $527;
    $531 = $530 | $528;
    HEAP32[$pixel66$290>>2] = $531;
    $532 = (($pixel66$290) + ($pitch65$2<<2)|0);
    $533 = ($513|0)==(0);
    if ($533) {
     break;
    } else {
     $length64$191 = $513;$pixel66$290 = $532;
    }
   }
   return;
   break;
  }
  default: {
   $534 = ((($dst)) + 16|0);
   $535 = HEAP32[$534>>2]|0;
   $536 = ((($dst)) + 4|0);
   $537 = HEAP32[$536>>2]|0;
   $538 = ((($537)) + 9|0);
   $539 = HEAP8[$538>>0]|0;
   $540 = $539&255;
   $541 = (($535|0) / ($540|0))&-1;
   $542 = ($y2|0)<($y1|0);
   $543 = ((($dst)) + 20|0);
   $544 = HEAP32[$543>>2]|0;
   if ($542) {
    $550 = Math_imul($541, $y2)|0;
    $551 = (($544) + ($550<<2)|0);
    $552 = (($551) + ($x2<<2)|0);
    $553 = ($x2|0)>($x1|0);
    $pitch72$1$v = $553 ? -1 : 1;
    $pitch72$1 = (($541) + ($pitch72$1$v))|0;
    $554 = ($draw_end|0)==(0);
    $555 = (($552) + ($pitch72$1<<2)|0);
    $$24 = $554 ? $555 : $552;
    $length71$0 = $387;$pitch72$2 = $pitch72$1;$pixel73$1 = $$24;
   } else {
    $545 = Math_imul($541, $y1)|0;
    $546 = (($544) + ($545<<2)|0);
    $547 = (($546) + ($x1<<2)|0);
    $548 = ($x2|0)<($x1|0);
    $pitch72$0$v = $548 ? -1 : 1;
    $pitch72$0 = (($541) + ($pitch72$0$v))|0;
    $549 = (($y2) - ($y1))|0;
    $length71$0 = $549;$pitch72$2 = $pitch72$0;$pixel73$1 = $547;
   }
   $not$35 = ($draw_end|0)!=(0);
   $556 = $not$35&1;
   $length71$0$ = (($length71$0) + ($556))|0;
   $557 = ($length71$0$|0)==(0);
   if ($557) {
    return;
   }
   $558 = $a$0 << 24;
   $559 = $r$0 << 16;
   $560 = $g$0 << 8;
   $561 = $b$0 | $558;
   $562 = $561 | $560;
   $563 = $562 | $559;
   $length71$181 = $length71$0$;$pixel73$282 = $pixel73$1;
   while(1) {
    $564 = (($length71$181) + -1)|0;
    HEAP32[$pixel73$282>>2] = $563;
    $565 = (($pixel73$282) + ($pitch72$2<<2)|0);
    $566 = ($564|0)==(0);
    if ($566) {
     break;
    } else {
     $length71$181 = $564;$pixel73$282 = $565;
    }
   }
   return;
  }
  }
 }
 switch ($blendMode|0) {
 case 1:  {
  $567 = (($x2) - ($x1))|0;
  $568 = ($567|0)<(0);
  $569 = (0 - ($567))|0;
  $570 = $568 ? $569 : $567;
  $571 = (($y2) - ($y1))|0;
  $572 = ($571|0)<(0);
  $573 = (0 - ($571))|0;
  $574 = $572 ? $573 : $571;
  $575 = ($570|0)<($574|0);
  if ($575) {
   $579 = $570 << 1;
   $580 = (($579) - ($574))|0;
   $581 = (($570) - ($574))|0;
   $d$0 = $580;$dinc1$0 = $579;$dinc2$0$in = $581;$numpixels$0$in = $574;$xinc1$0 = 0;$yinc1$0 = 1;
  } else {
   $576 = $574 << 1;
   $577 = (($576) - ($570))|0;
   $578 = (($574) - ($570))|0;
   $d$0 = $577;$dinc1$0 = $576;$dinc2$0$in = $578;$numpixels$0$in = $570;$xinc1$0 = 1;$yinc1$0 = 0;
  }
  $dinc2$0 = $dinc2$0$in << 1;
  $582 = ($x2|0)<($x1|0);
  $583 = (0 - ($xinc1$0))|0;
  $$xinc1$0 = $582 ? $583 : $xinc1$0;
  $$25 = $582 ? -1 : 1;
  $584 = ($y2|0)<($y1|0);
  $585 = (0 - ($yinc1$0))|0;
  $yinc1$1 = $584 ? $585 : $yinc1$0;
  $yinc2$1 = $584 ? -1 : 1;
  $not$30 = ($draw_end|0)!=(0);
  $numpixels$0 = $not$30&1;
  $numpixels$0$in$numpixels$0 = (($numpixels$0$in) + ($numpixels$0))|0;
  $586 = ($numpixels$0$in$numpixels$0|0)>(0);
  if (!($586)) {
   return;
  }
  $587 = ((($dst)) + 20|0);
  $588 = HEAP32[$587>>2]|0;
  $589 = ((($dst)) + 16|0);
  $d$199 = $d$0;$i$098 = 0;$x$0100 = $x1;$y$0101 = $y1;
  while(1) {
   $590 = HEAP32[$589>>2]|0;
   $591 = Math_imul($590, $y$0101)|0;
   $592 = (($588) + ($591)|0);
   $593 = $x$0100 << 2;
   $594 = (($592) + ($593)|0);
   $595 = HEAP32[$594>>2]|0;
   $596 = $595 >>> 16;
   $597 = $596 & 255;
   $598 = $595 >>> 8;
   $599 = $598 & 255;
   $600 = $595 & 255;
   $601 = $595 >>> 24;
   $602 = Math_imul($597, $14)|0;
   $603 = (($602>>>0) / 255)&-1;
   $604 = (($603) + ($r$0))|0;
   $605 = Math_imul($599, $14)|0;
   $606 = (($605>>>0) / 255)&-1;
   $607 = (($606) + ($g$0))|0;
   $608 = Math_imul($600, $14)|0;
   $609 = (($608>>>0) / 255)&-1;
   $610 = (($609) + ($b$0))|0;
   $611 = Math_imul($601, $14)|0;
   $612 = (($611>>>0) / 255)&-1;
   $613 = (($612) + ($a$0))|0;
   $614 = $613 << 24;
   $615 = $604 << 16;
   $616 = $607 << 8;
   $617 = $614 | $610;
   $618 = $617 | $615;
   $619 = $618 | $616;
   HEAP32[$594>>2] = $619;
   $620 = ($d$199|0)<(0);
   $dinc1$0$dinc2$0 = $620 ? $dinc1$0 : $dinc2$0;
   $$xinc1$0$$25 = $620 ? $$xinc1$0 : $$25;
   $yinc1$1$yinc2$1 = $620 ? $yinc1$1 : $yinc2$1;
   $y$1 = (($y$0101) + ($yinc1$1$yinc2$1))|0;
   $x$1 = (($$xinc1$0$$25) + ($x$0100))|0;
   $d$2 = (($dinc1$0$dinc2$0) + ($d$199))|0;
   $621 = (($i$098) + 1)|0;
   $622 = ($621|0)<($numpixels$0$in$numpixels$0|0);
   if ($622) {
    $d$199 = $d$2;$i$098 = $621;$x$0100 = $x$1;$y$0101 = $y$1;
   } else {
    break;
   }
  }
  return;
  break;
 }
 case 2:  {
  $623 = (($x2) - ($x1))|0;
  $624 = ($623|0)<(0);
  $625 = (0 - ($623))|0;
  $626 = $624 ? $625 : $623;
  $627 = (($y2) - ($y1))|0;
  $628 = ($627|0)<(0);
  $629 = (0 - ($627))|0;
  $630 = $628 ? $629 : $627;
  $631 = ($626|0)<($630|0);
  if ($631) {
   $635 = $626 << 1;
   $636 = (($635) - ($630))|0;
   $637 = (($626) - ($630))|0;
   $d87$0 = $636;$dinc188$0 = $635;$dinc289$0$in = $637;$numpixels86$0$in = $630;$xinc191$0 = 0;$yinc194$0 = 1;
  } else {
   $632 = $630 << 1;
   $633 = (($632) - ($626))|0;
   $634 = (($630) - ($626))|0;
   $d87$0 = $633;$dinc188$0 = $632;$dinc289$0$in = $634;$numpixels86$0$in = $626;$xinc191$0 = 1;$yinc194$0 = 0;
  }
  $dinc289$0 = $dinc289$0$in << 1;
  $638 = ($x2|0)<($x1|0);
  $639 = (0 - ($xinc191$0))|0;
  $$xinc191$0 = $638 ? $639 : $xinc191$0;
  $$26 = $638 ? -1 : 1;
  $640 = ($y2|0)<($y1|0);
  $641 = (0 - ($yinc194$0))|0;
  $yinc194$1 = $640 ? $641 : $yinc194$0;
  $yinc295$1 = $640 ? -1 : 1;
  $not$29 = ($draw_end|0)!=(0);
  $numpixels86$0 = $not$29&1;
  $numpixels86$0$in$numpixels86$0 = (($numpixels86$0$in) + ($numpixels86$0))|0;
  $642 = ($numpixels86$0$in$numpixels86$0|0)>(0);
  if (!($642)) {
   return;
  }
  $643 = ((($dst)) + 20|0);
  $644 = HEAP32[$643>>2]|0;
  $645 = ((($dst)) + 16|0);
  $d87$1104 = $d87$0;$i83$0103 = 0;$x90$0105 = $x1;$y93$0106 = $y1;
  while(1) {
   $646 = HEAP32[$645>>2]|0;
   $647 = Math_imul($646, $y93$0106)|0;
   $648 = (($644) + ($647)|0);
   $649 = $x90$0105 << 2;
   $650 = (($648) + ($649)|0);
   $651 = HEAP32[$650>>2]|0;
   $652 = $651 >>> 16;
   $653 = $652 & 255;
   $654 = $651 >>> 8;
   $655 = $654 & 255;
   $656 = $651 & 255;
   $657 = $651 & -16777216;
   $658 = (($653) + ($r$0))|0;
   $659 = ($658>>>0)>(255);
   $$11 = $659 ? 255 : $658;
   $660 = (($655) + ($g$0))|0;
   $661 = ($660>>>0)>(255);
   $$7 = $661 ? 255 : $660;
   $662 = (($656) + ($b$0))|0;
   $663 = ($662>>>0)>(255);
   $$12 = $663 ? 255 : $662;
   $664 = $$11 << 16;
   $665 = $$7 << 8;
   $666 = $$12 | $657;
   $667 = $666 | $664;
   $668 = $667 | $665;
   HEAP32[$650>>2] = $668;
   $669 = ($d87$1104|0)<(0);
   $dinc188$0$dinc289$0 = $669 ? $dinc188$0 : $dinc289$0;
   $$xinc191$0$$26 = $669 ? $$xinc191$0 : $$26;
   $yinc194$1$yinc295$1 = $669 ? $yinc194$1 : $yinc295$1;
   $y93$1 = (($y93$0106) + ($yinc194$1$yinc295$1))|0;
   $x90$1 = (($$xinc191$0$$26) + ($x90$0105))|0;
   $d87$2 = (($dinc188$0$dinc289$0) + ($d87$1104))|0;
   $670 = (($i83$0103) + 1)|0;
   $671 = ($670|0)<($numpixels86$0$in$numpixels86$0|0);
   if ($671) {
    $d87$1104 = $d87$2;$i83$0103 = $670;$x90$0105 = $x90$1;$y93$0106 = $y93$1;
   } else {
    break;
   }
  }
  return;
  break;
 }
 case 4:  {
  $672 = (($x2) - ($x1))|0;
  $673 = ($672|0)<(0);
  $674 = (0 - ($672))|0;
  $675 = $673 ? $674 : $672;
  $676 = (($y2) - ($y1))|0;
  $677 = ($676|0)<(0);
  $678 = (0 - ($676))|0;
  $679 = $677 ? $678 : $676;
  $680 = ($675|0)<($679|0);
  if ($680) {
   $684 = $675 << 1;
   $685 = (($684) - ($679))|0;
   $686 = (($675) - ($679))|0;
   $d105$0 = $685;$dinc1106$0 = $684;$dinc2107$0$in = $686;$numpixels104$0$in = $679;$xinc1109$0 = 0;$yinc1112$0 = 1;
  } else {
   $681 = $679 << 1;
   $682 = (($681) - ($675))|0;
   $683 = (($679) - ($675))|0;
   $d105$0 = $682;$dinc1106$0 = $681;$dinc2107$0$in = $683;$numpixels104$0$in = $675;$xinc1109$0 = 1;$yinc1112$0 = 0;
  }
  $dinc2107$0 = $dinc2107$0$in << 1;
  $687 = ($x2|0)<($x1|0);
  $688 = (0 - ($xinc1109$0))|0;
  $$xinc1109$0 = $687 ? $688 : $xinc1109$0;
  $$27 = $687 ? -1 : 1;
  $689 = ($y2|0)<($y1|0);
  $690 = (0 - ($yinc1112$0))|0;
  $yinc1112$1 = $689 ? $690 : $yinc1112$0;
  $yinc2113$1 = $689 ? -1 : 1;
  $not$ = ($draw_end|0)!=(0);
  $numpixels104$0 = $not$&1;
  $numpixels104$0$in$numpixels104$0 = (($numpixels104$0$in) + ($numpixels104$0))|0;
  $691 = ($numpixels104$0$in$numpixels104$0|0)>(0);
  if (!($691)) {
   return;
  }
  $692 = ((($dst)) + 20|0);
  $693 = HEAP32[$692>>2]|0;
  $694 = ((($dst)) + 16|0);
  $d105$1109 = $d105$0;$i101$0108 = 0;$x108$0110 = $x1;$y111$0111 = $y1;
  while(1) {
   $695 = HEAP32[$694>>2]|0;
   $696 = Math_imul($695, $y111$0111)|0;
   $697 = (($693) + ($696)|0);
   $698 = $x108$0110 << 2;
   $699 = (($697) + ($698)|0);
   $700 = HEAP32[$699>>2]|0;
   $701 = $700 >>> 16;
   $702 = $701 & 255;
   $703 = $700 >>> 8;
   $704 = $703 & 255;
   $705 = $700 & 255;
   $706 = $700 & -16777216;
   $707 = Math_imul($702, $r$0)|0;
   $708 = (($707>>>0) / 255)&-1;
   $709 = Math_imul($704, $g$0)|0;
   $710 = (($709>>>0) / 255)&-1;
   $711 = Math_imul($705, $b$0)|0;
   $712 = (($711>>>0) / 255)&-1;
   $713 = $708 << 16;
   $714 = $710 << 8;
   $715 = $712 | $706;
   $716 = $715 | $713;
   $717 = $716 | $714;
   HEAP32[$699>>2] = $717;
   $718 = ($d105$1109|0)<(0);
   $dinc1106$0$dinc2107$0 = $718 ? $dinc1106$0 : $dinc2107$0;
   $$xinc1109$0$$27 = $718 ? $$xinc1109$0 : $$27;
   $yinc1112$1$yinc2113$1 = $718 ? $yinc1112$1 : $yinc2113$1;
   $y111$1 = (($y111$0111) + ($yinc1112$1$yinc2113$1))|0;
   $x108$1 = (($$xinc1109$0$$27) + ($x108$0110))|0;
   $d105$2 = (($dinc1106$0$dinc2107$0) + ($d105$1109))|0;
   $719 = (($i101$0108) + 1)|0;
   $720 = ($719|0)<($numpixels104$0$in$numpixels104$0|0);
   if ($720) {
    $d105$1109 = $d105$2;$i101$0108 = $719;$x108$0110 = $x108$1;$y111$0111 = $y111$1;
   } else {
    break;
   }
  }
  return;
  break;
 }
 default: {
  $721 = (($x2) - ($x1))|0;
  $722 = ($721|0)<(0);
  $723 = (0 - ($721))|0;
  $724 = $722 ? $723 : $721;
  $725 = (($y2) - ($y1))|0;
  $726 = ($725|0)<(0);
  $727 = (0 - ($725))|0;
  $728 = $726 ? $727 : $725;
  $729 = ($724|0)<($728|0);
  if ($729) {
   $733 = $724 << 1;
   $734 = (($733) - ($728))|0;
   $735 = (($724) - ($728))|0;
   $d123$0 = $734;$dinc1124$0 = $733;$dinc2125$0$in = $735;$numpixels122$0$in = $728;$xinc1127$0 = 0;$yinc1130$0 = 1;
  } else {
   $730 = $728 << 1;
   $731 = (($730) - ($724))|0;
   $732 = (($728) - ($724))|0;
   $d123$0 = $731;$dinc1124$0 = $730;$dinc2125$0$in = $732;$numpixels122$0$in = $724;$xinc1127$0 = 1;$yinc1130$0 = 0;
  }
  $dinc2125$0 = $dinc2125$0$in << 1;
  $736 = ($x2|0)<($x1|0);
  $737 = (0 - ($xinc1127$0))|0;
  $$xinc1127$0 = $736 ? $737 : $xinc1127$0;
  $$28 = $736 ? -1 : 1;
  $738 = ($y2|0)<($y1|0);
  $739 = (0 - ($yinc1130$0))|0;
  $yinc1130$1 = $738 ? $739 : $yinc1130$0;
  $yinc2131$1 = $738 ? -1 : 1;
  $not$31 = ($draw_end|0)!=(0);
  $numpixels122$0 = $not$31&1;
  $numpixels122$0$in$numpixels122$0 = (($numpixels122$0$in) + ($numpixels122$0))|0;
  $740 = ($numpixels122$0$in$numpixels122$0|0)>(0);
  if (!($740)) {
   return;
  }
  $741 = ((($dst)) + 20|0);
  $742 = HEAP32[$741>>2]|0;
  $743 = ((($dst)) + 16|0);
  $744 = $a$0 << 24;
  $745 = $r$0 << 16;
  $746 = $g$0 << 8;
  $747 = $b$0 | $744;
  $748 = $747 | $746;
  $749 = $748 | $745;
  $d123$194 = $d123$0;$i119$093 = 0;$x126$095 = $x1;$y129$096 = $y1;
  while(1) {
   $750 = HEAP32[$743>>2]|0;
   $751 = Math_imul($750, $y129$096)|0;
   $752 = (($742) + ($751)|0);
   $753 = $x126$095 << 2;
   $754 = (($752) + ($753)|0);
   HEAP32[$754>>2] = $749;
   $755 = ($d123$194|0)<(0);
   $dinc1124$0$dinc2125$0 = $755 ? $dinc1124$0 : $dinc2125$0;
   $$xinc1127$0$$28 = $755 ? $$xinc1127$0 : $$28;
   $yinc1130$1$yinc2131$1 = $755 ? $yinc1130$1 : $yinc2131$1;
   $y129$1 = (($y129$096) + ($yinc1130$1$yinc2131$1))|0;
   $x126$1 = (($$xinc1127$0$$28) + ($x126$095))|0;
   $d123$2 = (($dinc1124$0$dinc2125$0) + ($d123$194))|0;
   $756 = (($i119$093) + 1)|0;
   $757 = ($756|0)<($numpixels122$0$in$numpixels122$0|0);
   if ($757) {
    $d123$194 = $d123$2;$i119$093 = $756;$x126$095 = $x126$1;$y129$096 = $y129$1;
   } else {
    break;
   }
  }
  return;
 }
 }
}
function _SDL_BlendLine_RGB888($dst,$x1,$y1,$x2,$y2,$blendMode,$_r,$_g,$_b,$_a,$draw_end) {
 $dst = $dst|0;
 $x1 = $x1|0;
 $y1 = $y1|0;
 $x2 = $x2|0;
 $y2 = $y2|0;
 $blendMode = $blendMode|0;
 $_r = $_r|0;
 $_g = $_g|0;
 $_b = $_b|0;
 $_a = $_a|0;
 $draw_end = $draw_end|0;
 var $$ = 0, $$10 = 0, $$11 = 0, $$12 = 0, $$13 = 0, $$14 = 0, $$15 = 0, $$16 = 0, $$17 = 0, $$18 = 0, $$19 = 0, $$2 = 0, $$20 = 0, $$21 = 0, $$22 = 0, $$23 = 0, $$24 = 0, $$25 = 0, $$26 = 0, $$27 = 0;
 var $$28 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$6 = 0, $$7 = 0, $$8 = 0, $$9 = 0, $$xinc1$0 = 0, $$xinc1$0$$25 = 0, $$xinc1109$0 = 0, $$xinc1109$0$$27 = 0, $$xinc1127$0 = 0, $$xinc1127$0$$28 = 0, $$xinc191$0 = 0, $$xinc191$0$$26 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0;
 var $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0;
 var $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0;
 var $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0;
 var $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0;
 var $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0;
 var $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0;
 var $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0;
 var $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0;
 var $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0;
 var $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0;
 var $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0;
 var $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0;
 var $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0;
 var $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0;
 var $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0;
 var $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0;
 var $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0;
 var $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0;
 var $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0;
 var $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0;
 var $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0;
 var $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0;
 var $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0;
 var $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0;
 var $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0;
 var $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $a$0 = 0, $b$0 = 0, $blendMode$off = 0, $d$0 = 0, $d$199 = 0, $d$2 = 0, $d105$0 = 0, $d105$1109 = 0, $d105$2 = 0, $d123$0 = 0, $d123$194 = 0, $d123$2 = 0, $d87$0 = 0, $d87$1104 = 0;
 var $d87$2 = 0, $dinc1$0 = 0, $dinc1$0$dinc2$0 = 0, $dinc1106$0 = 0, $dinc1106$0$dinc2107$0 = 0, $dinc1124$0 = 0, $dinc1124$0$dinc2125$0 = 0, $dinc188$0 = 0, $dinc188$0$dinc289$0 = 0, $dinc2$0 = 0, $dinc2$0$in = 0, $dinc2107$0 = 0, $dinc2107$0$in = 0, $dinc2125$0 = 0, $dinc2125$0$in = 0, $dinc289$0 = 0, $dinc289$0$in = 0, $g$0 = 0, $i$098 = 0, $i101$0108 = 0;
 var $i119$093 = 0, $i83$0103 = 0, $length$0$ph = 0, $length$061 = 0, $length1$0$ph = 0, $length1$064 = 0, $length15$0$ph = 0, $length15$059 = 0, $length22$0$ph = 0, $length22$073 = 0, $length29$0$ph = 0, $length29$076 = 0, $length36$0$ph = 0, $length36$079 = 0, $length43$0$ph = 0, $length43$070 = 0, $length50$0 = 0, $length50$0$ = 0, $length50$185 = 0, $length57$0 = 0;
 var $length57$0$ = 0, $length57$188 = 0, $length64$0 = 0, $length64$0$ = 0, $length64$191 = 0, $length71$0 = 0, $length71$0$ = 0, $length71$182 = 0, $length8$0$ph = 0, $length8$067 = 0, $not$ = 0, $not$29 = 0, $not$30 = 0, $not$31 = 0, $not$32 = 0, $not$33 = 0, $not$34 = 0, $not$35 = 0, $numpixels$0 = 0, $numpixels$0$in = 0;
 var $numpixels$0$in$numpixels$0 = 0, $numpixels104$0 = 0, $numpixels104$0$in = 0, $numpixels104$0$in$numpixels104$0 = 0, $numpixels122$0 = 0, $numpixels122$0$in = 0, $numpixels122$0$in$numpixels122$0 = 0, $numpixels86$0 = 0, $numpixels86$0$in = 0, $numpixels86$0$in$numpixels86$0 = 0, $pitch51$0 = 0, $pitch51$0$v = 0, $pitch51$1 = 0, $pitch51$1$v = 0, $pitch51$2 = 0, $pitch58$0 = 0, $pitch58$0$v = 0, $pitch58$1 = 0, $pitch58$1$v = 0, $pitch58$2 = 0;
 var $pitch65$0 = 0, $pitch65$0$v = 0, $pitch65$1 = 0, $pitch65$1$v = 0, $pitch65$2 = 0, $pitch72$0 = 0, $pitch72$0$v = 0, $pitch72$1 = 0, $pitch72$1$v = 0, $pitch72$2 = 0, $pixel$1$ph = 0, $pixel$160 = 0, $pixel10$1$ph = 0, $pixel10$166 = 0, $pixel17$1$ph = 0, $pixel17$158 = 0, $pixel24$1$ph = 0, $pixel24$172 = 0, $pixel3$1$ph = 0, $pixel3$163 = 0;
 var $pixel31$1$ph = 0, $pixel31$175 = 0, $pixel38$1$ph = 0, $pixel38$178 = 0, $pixel45$1$ph = 0, $pixel45$169 = 0, $pixel52$1 = 0, $pixel52$284 = 0, $pixel59$1 = 0, $pixel59$287 = 0, $pixel66$1 = 0, $pixel66$290 = 0, $pixel73$1 = 0, $pixel73$281 = 0, $r$0 = 0, $x$0100 = 0, $x$1 = 0, $x108$0110 = 0, $x108$1 = 0, $x126$095 = 0;
 var $x126$1 = 0, $x90$0105 = 0, $x90$1 = 0, $xinc1$0 = 0, $xinc1109$0 = 0, $xinc1127$0 = 0, $xinc191$0 = 0, $y$0101 = 0, $y$1 = 0, $y111$0111 = 0, $y111$1 = 0, $y129$096 = 0, $y129$1 = 0, $y93$0106 = 0, $y93$1 = 0, $yinc1$0 = 0, $yinc1$1 = 0, $yinc1$1$yinc2$1 = 0, $yinc1112$0 = 0, $yinc1112$1 = 0;
 var $yinc1112$1$yinc2113$1 = 0, $yinc1130$0 = 0, $yinc1130$1 = 0, $yinc1130$1$yinc2131$1 = 0, $yinc194$0 = 0, $yinc194$1 = 0, $yinc194$1$yinc295$1 = 0, $yinc2$1 = 0, $yinc2113$1 = 0, $yinc2131$1 = 0, $yinc295$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $blendMode$off = (($blendMode) + -1)|0;
 $0 = ($blendMode$off>>>0)<(2);
 $1 = $_r&255;
 if ($0) {
  $2 = $_a&255;
  $3 = Math_imul($2, $1)|0;
  $4 = (($3>>>0) / 255)&-1;
  $5 = $_g&255;
  $6 = Math_imul($2, $5)|0;
  $7 = (($6>>>0) / 255)&-1;
  $8 = $_b&255;
  $9 = Math_imul($2, $8)|0;
  $10 = (($9>>>0) / 255)&-1;
  $a$0 = $2;$b$0 = $10;$g$0 = $7;$r$0 = $4;
 } else {
  $11 = $_g&255;
  $12 = $_b&255;
  $13 = $_a&255;
  $a$0 = $13;$b$0 = $12;$g$0 = $11;$r$0 = $1;
 }
 $14 = $a$0 ^ 255;
 $15 = ($y2|0)==($y1|0);
 if ($15) {
  switch ($blendMode|0) {
  case 1:  {
   $16 = ((($dst)) + 16|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($dst)) + 4|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ((($19)) + 9|0);
   $21 = HEAP8[$20>>0]|0;
   $22 = $21&255;
   $23 = (($17|0) / ($22|0))&-1;
   $24 = ($x2|0)<($x1|0);
   $25 = ((($dst)) + 20|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = Math_imul($23, $y2)|0;
   $28 = (($26) + ($27<<2)|0);
   if ($24) {
    $34 = (($28) + ($x2<<2)|0);
    $35 = ($draw_end|0)!=(0);
    $36 = ((($34)) + 4|0);
    $$13 = $35 ? $34 : $36;
    $37 = (($x1) - ($x2))|0;
    $38 = $35&1;
    $39 = (($37) + ($38))|0;
    $length$0$ph = $39;$pixel$1$ph = $$13;
   } else {
    $29 = (($28) + ($x1<<2)|0);
    $30 = ($draw_end|0)!=(0);
    $31 = (($x2) - ($x1))|0;
    $32 = $30&1;
    $33 = (($31) + ($32))|0;
    $length$0$ph = $33;$pixel$1$ph = $29;
   }
   $40 = ($length$0$ph|0)==(0);
   if ($40) {
    return;
   } else {
    $length$061 = $length$0$ph;$pixel$160 = $pixel$1$ph;
   }
   while(1) {
    $41 = (($length$061) + -1)|0;
    $42 = HEAP32[$pixel$160>>2]|0;
    $43 = $42 >>> 16;
    $44 = $43 & 255;
    $45 = $42 >>> 8;
    $46 = $45 & 255;
    $47 = $42 & 255;
    $48 = Math_imul($44, $14)|0;
    $49 = (($48>>>0) / 255)&-1;
    $50 = (($49) + ($r$0))|0;
    $51 = Math_imul($46, $14)|0;
    $52 = (($51>>>0) / 255)&-1;
    $53 = (($52) + ($g$0))|0;
    $54 = Math_imul($47, $14)|0;
    $55 = (($54>>>0) / 255)&-1;
    $56 = (($55) + ($b$0))|0;
    $57 = $50 << 16;
    $58 = $53 << 8;
    $59 = $58 | $56;
    $60 = $59 | $57;
    HEAP32[$pixel$160>>2] = $60;
    $61 = ((($pixel$160)) + 4|0);
    $62 = ($41|0)==(0);
    if ($62) {
     break;
    } else {
     $length$061 = $41;$pixel$160 = $61;
    }
   }
   return;
   break;
  }
  case 2:  {
   $63 = ((($dst)) + 16|0);
   $64 = HEAP32[$63>>2]|0;
   $65 = ((($dst)) + 4|0);
   $66 = HEAP32[$65>>2]|0;
   $67 = ((($66)) + 9|0);
   $68 = HEAP8[$67>>0]|0;
   $69 = $68&255;
   $70 = (($64|0) / ($69|0))&-1;
   $71 = ($x2|0)<($x1|0);
   $72 = ((($dst)) + 20|0);
   $73 = HEAP32[$72>>2]|0;
   $74 = Math_imul($70, $y2)|0;
   $75 = (($73) + ($74<<2)|0);
   if ($71) {
    $81 = (($75) + ($x2<<2)|0);
    $82 = ($draw_end|0)!=(0);
    $83 = ((($81)) + 4|0);
    $$14 = $82 ? $81 : $83;
    $84 = (($x1) - ($x2))|0;
    $85 = $82&1;
    $86 = (($84) + ($85))|0;
    $length1$0$ph = $86;$pixel3$1$ph = $$14;
   } else {
    $76 = (($75) + ($x1<<2)|0);
    $77 = ($draw_end|0)!=(0);
    $78 = (($x2) - ($x1))|0;
    $79 = $77&1;
    $80 = (($78) + ($79))|0;
    $length1$0$ph = $80;$pixel3$1$ph = $76;
   }
   $87 = ($length1$0$ph|0)==(0);
   if ($87) {
    return;
   } else {
    $length1$064 = $length1$0$ph;$pixel3$163 = $pixel3$1$ph;
   }
   while(1) {
    $88 = (($length1$064) + -1)|0;
    $89 = HEAP32[$pixel3$163>>2]|0;
    $90 = $89 >>> 16;
    $91 = $90 & 255;
    $92 = $89 >>> 8;
    $93 = $92 & 255;
    $94 = $89 & 255;
    $95 = (($91) + ($r$0))|0;
    $96 = ($95>>>0)>(255);
    $$ = $96 ? 255 : $95;
    $97 = (($93) + ($g$0))|0;
    $98 = ($97>>>0)>(255);
    $$8 = $98 ? 255 : $97;
    $99 = (($94) + ($b$0))|0;
    $100 = ($99>>>0)>(255);
    $$2 = $100 ? 255 : $99;
    $101 = $$ << 16;
    $102 = $$8 << 8;
    $103 = $102 | $$2;
    $104 = $103 | $101;
    HEAP32[$pixel3$163>>2] = $104;
    $105 = ((($pixel3$163)) + 4|0);
    $106 = ($88|0)==(0);
    if ($106) {
     break;
    } else {
     $length1$064 = $88;$pixel3$163 = $105;
    }
   }
   return;
   break;
  }
  case 4:  {
   $107 = ((($dst)) + 16|0);
   $108 = HEAP32[$107>>2]|0;
   $109 = ((($dst)) + 4|0);
   $110 = HEAP32[$109>>2]|0;
   $111 = ((($110)) + 9|0);
   $112 = HEAP8[$111>>0]|0;
   $113 = $112&255;
   $114 = (($108|0) / ($113|0))&-1;
   $115 = ($x2|0)<($x1|0);
   $116 = ((($dst)) + 20|0);
   $117 = HEAP32[$116>>2]|0;
   $118 = Math_imul($114, $y2)|0;
   $119 = (($117) + ($118<<2)|0);
   if ($115) {
    $125 = (($119) + ($x2<<2)|0);
    $126 = ($draw_end|0)!=(0);
    $127 = ((($125)) + 4|0);
    $$15 = $126 ? $125 : $127;
    $128 = (($x1) - ($x2))|0;
    $129 = $126&1;
    $130 = (($128) + ($129))|0;
    $length8$0$ph = $130;$pixel10$1$ph = $$15;
   } else {
    $120 = (($119) + ($x1<<2)|0);
    $121 = ($draw_end|0)!=(0);
    $122 = (($x2) - ($x1))|0;
    $123 = $121&1;
    $124 = (($122) + ($123))|0;
    $length8$0$ph = $124;$pixel10$1$ph = $120;
   }
   $131 = ($length8$0$ph|0)==(0);
   if ($131) {
    return;
   } else {
    $length8$067 = $length8$0$ph;$pixel10$166 = $pixel10$1$ph;
   }
   while(1) {
    $132 = (($length8$067) + -1)|0;
    $133 = HEAP32[$pixel10$166>>2]|0;
    $134 = $133 >>> 16;
    $135 = $134 & 255;
    $136 = $133 >>> 8;
    $137 = $136 & 255;
    $138 = $133 & 255;
    $139 = Math_imul($135, $r$0)|0;
    $140 = (($139>>>0) / 255)&-1;
    $141 = Math_imul($137, $g$0)|0;
    $142 = (($141>>>0) / 255)&-1;
    $143 = Math_imul($138, $b$0)|0;
    $144 = (($143>>>0) / 255)&-1;
    $145 = $140 << 16;
    $146 = $142 << 8;
    $147 = $146 | $144;
    $148 = $147 | $145;
    HEAP32[$pixel10$166>>2] = $148;
    $149 = ((($pixel10$166)) + 4|0);
    $150 = ($132|0)==(0);
    if ($150) {
     break;
    } else {
     $length8$067 = $132;$pixel10$166 = $149;
    }
   }
   return;
   break;
  }
  default: {
   $151 = ((($dst)) + 16|0);
   $152 = HEAP32[$151>>2]|0;
   $153 = ((($dst)) + 4|0);
   $154 = HEAP32[$153>>2]|0;
   $155 = ((($154)) + 9|0);
   $156 = HEAP8[$155>>0]|0;
   $157 = $156&255;
   $158 = (($152|0) / ($157|0))&-1;
   $159 = ($x2|0)<($x1|0);
   $160 = ((($dst)) + 20|0);
   $161 = HEAP32[$160>>2]|0;
   $162 = Math_imul($158, $y2)|0;
   $163 = (($161) + ($162<<2)|0);
   if ($159) {
    $169 = (($163) + ($x2<<2)|0);
    $170 = ($draw_end|0)!=(0);
    $171 = ((($169)) + 4|0);
    $$16 = $170 ? $169 : $171;
    $172 = (($x1) - ($x2))|0;
    $173 = $170&1;
    $174 = (($172) + ($173))|0;
    $length15$0$ph = $174;$pixel17$1$ph = $$16;
   } else {
    $164 = (($163) + ($x1<<2)|0);
    $165 = ($draw_end|0)!=(0);
    $166 = (($x2) - ($x1))|0;
    $167 = $165&1;
    $168 = (($166) + ($167))|0;
    $length15$0$ph = $168;$pixel17$1$ph = $164;
   }
   $175 = ($length15$0$ph|0)==(0);
   if ($175) {
    return;
   }
   $176 = $r$0 << 16;
   $177 = $g$0 << 8;
   $178 = $177 | $b$0;
   $179 = $178 | $176;
   $length15$059 = $length15$0$ph;$pixel17$158 = $pixel17$1$ph;
   while(1) {
    $180 = (($length15$059) + -1)|0;
    HEAP32[$pixel17$158>>2] = $179;
    $181 = ((($pixel17$158)) + 4|0);
    $182 = ($180|0)==(0);
    if ($182) {
     break;
    } else {
     $length15$059 = $180;$pixel17$158 = $181;
    }
   }
   return;
  }
  }
 }
 $183 = ($x2|0)==($x1|0);
 if ($183) {
  switch ($blendMode|0) {
  case 1:  {
   $184 = ((($dst)) + 16|0);
   $185 = HEAP32[$184>>2]|0;
   $186 = ((($dst)) + 4|0);
   $187 = HEAP32[$186>>2]|0;
   $188 = ((($187)) + 9|0);
   $189 = HEAP8[$188>>0]|0;
   $190 = $189&255;
   $191 = (($185|0) / ($190|0))&-1;
   $192 = ($y2|0)<($y1|0);
   $193 = ((($dst)) + 20|0);
   $194 = HEAP32[$193>>2]|0;
   if ($192) {
    $202 = Math_imul($191, $y2)|0;
    $203 = (($194) + ($202<<2)|0);
    $204 = (($203) + ($x2<<2)|0);
    $205 = ($draw_end|0)!=(0);
    $206 = (($204) + ($191<<2)|0);
    $$17 = $205 ? $204 : $206;
    $207 = (($y1) - ($y2))|0;
    $208 = $205&1;
    $209 = (($207) + ($208))|0;
    $length22$0$ph = $209;$pixel24$1$ph = $$17;
   } else {
    $195 = Math_imul($191, $y1)|0;
    $196 = (($194) + ($195<<2)|0);
    $197 = (($196) + ($x2<<2)|0);
    $198 = ($draw_end|0)!=(0);
    $199 = (($y2) - ($y1))|0;
    $200 = $198&1;
    $201 = (($199) + ($200))|0;
    $length22$0$ph = $201;$pixel24$1$ph = $197;
   }
   $210 = ($length22$0$ph|0)==(0);
   if ($210) {
    return;
   } else {
    $length22$073 = $length22$0$ph;$pixel24$172 = $pixel24$1$ph;
   }
   while(1) {
    $211 = (($length22$073) + -1)|0;
    $212 = HEAP32[$pixel24$172>>2]|0;
    $213 = $212 >>> 16;
    $214 = $213 & 255;
    $215 = $212 >>> 8;
    $216 = $215 & 255;
    $217 = $212 & 255;
    $218 = Math_imul($214, $14)|0;
    $219 = (($218>>>0) / 255)&-1;
    $220 = (($219) + ($r$0))|0;
    $221 = Math_imul($216, $14)|0;
    $222 = (($221>>>0) / 255)&-1;
    $223 = (($222) + ($g$0))|0;
    $224 = Math_imul($217, $14)|0;
    $225 = (($224>>>0) / 255)&-1;
    $226 = (($225) + ($b$0))|0;
    $227 = $220 << 16;
    $228 = $223 << 8;
    $229 = $228 | $226;
    $230 = $229 | $227;
    HEAP32[$pixel24$172>>2] = $230;
    $231 = (($pixel24$172) + ($191<<2)|0);
    $232 = ($211|0)==(0);
    if ($232) {
     break;
    } else {
     $length22$073 = $211;$pixel24$172 = $231;
    }
   }
   return;
   break;
  }
  case 2:  {
   $233 = ((($dst)) + 16|0);
   $234 = HEAP32[$233>>2]|0;
   $235 = ((($dst)) + 4|0);
   $236 = HEAP32[$235>>2]|0;
   $237 = ((($236)) + 9|0);
   $238 = HEAP8[$237>>0]|0;
   $239 = $238&255;
   $240 = (($234|0) / ($239|0))&-1;
   $241 = ($y2|0)<($y1|0);
   $242 = ((($dst)) + 20|0);
   $243 = HEAP32[$242>>2]|0;
   if ($241) {
    $251 = Math_imul($240, $y2)|0;
    $252 = (($243) + ($251<<2)|0);
    $253 = (($252) + ($x2<<2)|0);
    $254 = ($draw_end|0)!=(0);
    $255 = (($253) + ($240<<2)|0);
    $$18 = $254 ? $253 : $255;
    $256 = (($y1) - ($y2))|0;
    $257 = $254&1;
    $258 = (($256) + ($257))|0;
    $length29$0$ph = $258;$pixel31$1$ph = $$18;
   } else {
    $244 = Math_imul($240, $y1)|0;
    $245 = (($243) + ($244<<2)|0);
    $246 = (($245) + ($x2<<2)|0);
    $247 = ($draw_end|0)!=(0);
    $248 = (($y2) - ($y1))|0;
    $249 = $247&1;
    $250 = (($248) + ($249))|0;
    $length29$0$ph = $250;$pixel31$1$ph = $246;
   }
   $259 = ($length29$0$ph|0)==(0);
   if ($259) {
    return;
   } else {
    $length29$076 = $length29$0$ph;$pixel31$175 = $pixel31$1$ph;
   }
   while(1) {
    $260 = (($length29$076) + -1)|0;
    $261 = HEAP32[$pixel31$175>>2]|0;
    $262 = $261 >>> 16;
    $263 = $262 & 255;
    $264 = $261 >>> 8;
    $265 = $264 & 255;
    $266 = $261 & 255;
    $267 = (($263) + ($r$0))|0;
    $268 = ($267>>>0)>(255);
    $$3 = $268 ? 255 : $267;
    $269 = (($265) + ($g$0))|0;
    $270 = ($269>>>0)>(255);
    $$9 = $270 ? 255 : $269;
    $271 = (($266) + ($b$0))|0;
    $272 = ($271>>>0)>(255);
    $$4 = $272 ? 255 : $271;
    $273 = $$3 << 16;
    $274 = $$9 << 8;
    $275 = $274 | $$4;
    $276 = $275 | $273;
    HEAP32[$pixel31$175>>2] = $276;
    $277 = (($pixel31$175) + ($240<<2)|0);
    $278 = ($260|0)==(0);
    if ($278) {
     break;
    } else {
     $length29$076 = $260;$pixel31$175 = $277;
    }
   }
   return;
   break;
  }
  case 4:  {
   $279 = ((($dst)) + 16|0);
   $280 = HEAP32[$279>>2]|0;
   $281 = ((($dst)) + 4|0);
   $282 = HEAP32[$281>>2]|0;
   $283 = ((($282)) + 9|0);
   $284 = HEAP8[$283>>0]|0;
   $285 = $284&255;
   $286 = (($280|0) / ($285|0))&-1;
   $287 = ($y2|0)<($y1|0);
   $288 = ((($dst)) + 20|0);
   $289 = HEAP32[$288>>2]|0;
   if ($287) {
    $297 = Math_imul($286, $y2)|0;
    $298 = (($289) + ($297<<2)|0);
    $299 = (($298) + ($x2<<2)|0);
    $300 = ($draw_end|0)!=(0);
    $301 = (($299) + ($286<<2)|0);
    $$19 = $300 ? $299 : $301;
    $302 = (($y1) - ($y2))|0;
    $303 = $300&1;
    $304 = (($302) + ($303))|0;
    $length36$0$ph = $304;$pixel38$1$ph = $$19;
   } else {
    $290 = Math_imul($286, $y1)|0;
    $291 = (($289) + ($290<<2)|0);
    $292 = (($291) + ($x2<<2)|0);
    $293 = ($draw_end|0)!=(0);
    $294 = (($y2) - ($y1))|0;
    $295 = $293&1;
    $296 = (($294) + ($295))|0;
    $length36$0$ph = $296;$pixel38$1$ph = $292;
   }
   $305 = ($length36$0$ph|0)==(0);
   if ($305) {
    return;
   } else {
    $length36$079 = $length36$0$ph;$pixel38$178 = $pixel38$1$ph;
   }
   while(1) {
    $306 = (($length36$079) + -1)|0;
    $307 = HEAP32[$pixel38$178>>2]|0;
    $308 = $307 >>> 16;
    $309 = $308 & 255;
    $310 = $307 >>> 8;
    $311 = $310 & 255;
    $312 = $307 & 255;
    $313 = Math_imul($309, $r$0)|0;
    $314 = (($313>>>0) / 255)&-1;
    $315 = Math_imul($311, $g$0)|0;
    $316 = (($315>>>0) / 255)&-1;
    $317 = Math_imul($312, $b$0)|0;
    $318 = (($317>>>0) / 255)&-1;
    $319 = $314 << 16;
    $320 = $316 << 8;
    $321 = $320 | $318;
    $322 = $321 | $319;
    HEAP32[$pixel38$178>>2] = $322;
    $323 = (($pixel38$178) + ($286<<2)|0);
    $324 = ($306|0)==(0);
    if ($324) {
     break;
    } else {
     $length36$079 = $306;$pixel38$178 = $323;
    }
   }
   return;
   break;
  }
  default: {
   $325 = ((($dst)) + 16|0);
   $326 = HEAP32[$325>>2]|0;
   $327 = ((($dst)) + 4|0);
   $328 = HEAP32[$327>>2]|0;
   $329 = ((($328)) + 9|0);
   $330 = HEAP8[$329>>0]|0;
   $331 = $330&255;
   $332 = (($326|0) / ($331|0))&-1;
   $333 = ($y2|0)<($y1|0);
   $334 = ((($dst)) + 20|0);
   $335 = HEAP32[$334>>2]|0;
   if ($333) {
    $343 = Math_imul($332, $y2)|0;
    $344 = (($335) + ($343<<2)|0);
    $345 = (($344) + ($x2<<2)|0);
    $346 = ($draw_end|0)!=(0);
    $347 = (($345) + ($332<<2)|0);
    $$20 = $346 ? $345 : $347;
    $348 = (($y1) - ($y2))|0;
    $349 = $346&1;
    $350 = (($348) + ($349))|0;
    $length43$0$ph = $350;$pixel45$1$ph = $$20;
   } else {
    $336 = Math_imul($332, $y1)|0;
    $337 = (($335) + ($336<<2)|0);
    $338 = (($337) + ($x2<<2)|0);
    $339 = ($draw_end|0)!=(0);
    $340 = (($y2) - ($y1))|0;
    $341 = $339&1;
    $342 = (($340) + ($341))|0;
    $length43$0$ph = $342;$pixel45$1$ph = $338;
   }
   $351 = ($length43$0$ph|0)==(0);
   if ($351) {
    return;
   }
   $352 = $r$0 << 16;
   $353 = $g$0 << 8;
   $354 = $353 | $b$0;
   $355 = $354 | $352;
   $length43$070 = $length43$0$ph;$pixel45$169 = $pixel45$1$ph;
   while(1) {
    $356 = (($length43$070) + -1)|0;
    HEAP32[$pixel45$169>>2] = $355;
    $357 = (($pixel45$169) + ($332<<2)|0);
    $358 = ($356|0)==(0);
    if ($358) {
     break;
    } else {
     $length43$070 = $356;$pixel45$169 = $357;
    }
   }
   return;
  }
  }
 }
 $359 = (($x1) - ($x2))|0;
 $360 = ($359|0)<(0);
 $361 = (0 - ($359))|0;
 $362 = $360 ? $361 : $359;
 $363 = (($y1) - ($y2))|0;
 $364 = ($363|0)<(0);
 $365 = (0 - ($363))|0;
 $366 = $364 ? $365 : $363;
 $367 = ($362|0)==($366|0);
 if ($367) {
  switch ($blendMode|0) {
  case 1:  {
   $368 = ((($dst)) + 16|0);
   $369 = HEAP32[$368>>2]|0;
   $370 = ((($dst)) + 4|0);
   $371 = HEAP32[$370>>2]|0;
   $372 = ((($371)) + 9|0);
   $373 = HEAP8[$372>>0]|0;
   $374 = $373&255;
   $375 = (($369|0) / ($374|0))&-1;
   $376 = ($y2|0)<($y1|0);
   $377 = ((($dst)) + 20|0);
   $378 = HEAP32[$377>>2]|0;
   if ($376) {
    $384 = Math_imul($375, $y2)|0;
    $385 = (($378) + ($384<<2)|0);
    $386 = (($385) + ($x2<<2)|0);
    $387 = ($x2|0)>($x1|0);
    $pitch51$1$v = $387 ? -1 : 1;
    $pitch51$1 = (($375) + ($pitch51$1$v))|0;
    $388 = ($draw_end|0)==(0);
    $389 = (($386) + ($pitch51$1<<2)|0);
    $$21 = $388 ? $389 : $386;
    $length50$0 = $363;$pitch51$2 = $pitch51$1;$pixel52$1 = $$21;
   } else {
    $379 = Math_imul($375, $y1)|0;
    $380 = (($378) + ($379<<2)|0);
    $381 = (($380) + ($x1<<2)|0);
    $382 = ($x2|0)<($x1|0);
    $pitch51$0$v = $382 ? -1 : 1;
    $pitch51$0 = (($375) + ($pitch51$0$v))|0;
    $383 = (($y2) - ($y1))|0;
    $length50$0 = $383;$pitch51$2 = $pitch51$0;$pixel52$1 = $381;
   }
   $not$34 = ($draw_end|0)!=(0);
   $390 = $not$34&1;
   $length50$0$ = (($length50$0) + ($390))|0;
   $391 = ($length50$0$|0)==(0);
   if ($391) {
    return;
   } else {
    $length50$185 = $length50$0$;$pixel52$284 = $pixel52$1;
   }
   while(1) {
    $392 = (($length50$185) + -1)|0;
    $393 = HEAP32[$pixel52$284>>2]|0;
    $394 = $393 >>> 16;
    $395 = $394 & 255;
    $396 = $393 >>> 8;
    $397 = $396 & 255;
    $398 = $393 & 255;
    $399 = Math_imul($395, $14)|0;
    $400 = (($399>>>0) / 255)&-1;
    $401 = (($400) + ($r$0))|0;
    $402 = Math_imul($397, $14)|0;
    $403 = (($402>>>0) / 255)&-1;
    $404 = (($403) + ($g$0))|0;
    $405 = Math_imul($398, $14)|0;
    $406 = (($405>>>0) / 255)&-1;
    $407 = (($406) + ($b$0))|0;
    $408 = $401 << 16;
    $409 = $404 << 8;
    $410 = $409 | $407;
    $411 = $410 | $408;
    HEAP32[$pixel52$284>>2] = $411;
    $412 = (($pixel52$284) + ($pitch51$2<<2)|0);
    $413 = ($392|0)==(0);
    if ($413) {
     break;
    } else {
     $length50$185 = $392;$pixel52$284 = $412;
    }
   }
   return;
   break;
  }
  case 2:  {
   $414 = ((($dst)) + 16|0);
   $415 = HEAP32[$414>>2]|0;
   $416 = ((($dst)) + 4|0);
   $417 = HEAP32[$416>>2]|0;
   $418 = ((($417)) + 9|0);
   $419 = HEAP8[$418>>0]|0;
   $420 = $419&255;
   $421 = (($415|0) / ($420|0))&-1;
   $422 = ($y2|0)<($y1|0);
   $423 = ((($dst)) + 20|0);
   $424 = HEAP32[$423>>2]|0;
   if ($422) {
    $430 = Math_imul($421, $y2)|0;
    $431 = (($424) + ($430<<2)|0);
    $432 = (($431) + ($x2<<2)|0);
    $433 = ($x2|0)>($x1|0);
    $pitch58$1$v = $433 ? -1 : 1;
    $pitch58$1 = (($421) + ($pitch58$1$v))|0;
    $434 = ($draw_end|0)==(0);
    $435 = (($432) + ($pitch58$1<<2)|0);
    $$22 = $434 ? $435 : $432;
    $length57$0 = $363;$pitch58$2 = $pitch58$1;$pixel59$1 = $$22;
   } else {
    $425 = Math_imul($421, $y1)|0;
    $426 = (($424) + ($425<<2)|0);
    $427 = (($426) + ($x1<<2)|0);
    $428 = ($x2|0)<($x1|0);
    $pitch58$0$v = $428 ? -1 : 1;
    $pitch58$0 = (($421) + ($pitch58$0$v))|0;
    $429 = (($y2) - ($y1))|0;
    $length57$0 = $429;$pitch58$2 = $pitch58$0;$pixel59$1 = $427;
   }
   $not$33 = ($draw_end|0)!=(0);
   $436 = $not$33&1;
   $length57$0$ = (($length57$0) + ($436))|0;
   $437 = ($length57$0$|0)==(0);
   if ($437) {
    return;
   } else {
    $length57$188 = $length57$0$;$pixel59$287 = $pixel59$1;
   }
   while(1) {
    $438 = (($length57$188) + -1)|0;
    $439 = HEAP32[$pixel59$287>>2]|0;
    $440 = $439 >>> 16;
    $441 = $440 & 255;
    $442 = $439 >>> 8;
    $443 = $442 & 255;
    $444 = $439 & 255;
    $445 = (($441) + ($r$0))|0;
    $446 = ($445>>>0)>(255);
    $$5 = $446 ? 255 : $445;
    $447 = (($443) + ($g$0))|0;
    $448 = ($447>>>0)>(255);
    $$10 = $448 ? 255 : $447;
    $449 = (($444) + ($b$0))|0;
    $450 = ($449>>>0)>(255);
    $$6 = $450 ? 255 : $449;
    $451 = $$5 << 16;
    $452 = $$10 << 8;
    $453 = $452 | $$6;
    $454 = $453 | $451;
    HEAP32[$pixel59$287>>2] = $454;
    $455 = (($pixel59$287) + ($pitch58$2<<2)|0);
    $456 = ($438|0)==(0);
    if ($456) {
     break;
    } else {
     $length57$188 = $438;$pixel59$287 = $455;
    }
   }
   return;
   break;
  }
  case 4:  {
   $457 = ((($dst)) + 16|0);
   $458 = HEAP32[$457>>2]|0;
   $459 = ((($dst)) + 4|0);
   $460 = HEAP32[$459>>2]|0;
   $461 = ((($460)) + 9|0);
   $462 = HEAP8[$461>>0]|0;
   $463 = $462&255;
   $464 = (($458|0) / ($463|0))&-1;
   $465 = ($y2|0)<($y1|0);
   $466 = ((($dst)) + 20|0);
   $467 = HEAP32[$466>>2]|0;
   if ($465) {
    $473 = Math_imul($464, $y2)|0;
    $474 = (($467) + ($473<<2)|0);
    $475 = (($474) + ($x2<<2)|0);
    $476 = ($x2|0)>($x1|0);
    $pitch65$1$v = $476 ? -1 : 1;
    $pitch65$1 = (($464) + ($pitch65$1$v))|0;
    $477 = ($draw_end|0)==(0);
    $478 = (($475) + ($pitch65$1<<2)|0);
    $$23 = $477 ? $478 : $475;
    $length64$0 = $363;$pitch65$2 = $pitch65$1;$pixel66$1 = $$23;
   } else {
    $468 = Math_imul($464, $y1)|0;
    $469 = (($467) + ($468<<2)|0);
    $470 = (($469) + ($x1<<2)|0);
    $471 = ($x2|0)<($x1|0);
    $pitch65$0$v = $471 ? -1 : 1;
    $pitch65$0 = (($464) + ($pitch65$0$v))|0;
    $472 = (($y2) - ($y1))|0;
    $length64$0 = $472;$pitch65$2 = $pitch65$0;$pixel66$1 = $470;
   }
   $not$32 = ($draw_end|0)!=(0);
   $479 = $not$32&1;
   $length64$0$ = (($length64$0) + ($479))|0;
   $480 = ($length64$0$|0)==(0);
   if ($480) {
    return;
   } else {
    $length64$191 = $length64$0$;$pixel66$290 = $pixel66$1;
   }
   while(1) {
    $481 = (($length64$191) + -1)|0;
    $482 = HEAP32[$pixel66$290>>2]|0;
    $483 = $482 >>> 16;
    $484 = $483 & 255;
    $485 = $482 >>> 8;
    $486 = $485 & 255;
    $487 = $482 & 255;
    $488 = Math_imul($484, $r$0)|0;
    $489 = (($488>>>0) / 255)&-1;
    $490 = Math_imul($486, $g$0)|0;
    $491 = (($490>>>0) / 255)&-1;
    $492 = Math_imul($487, $b$0)|0;
    $493 = (($492>>>0) / 255)&-1;
    $494 = $489 << 16;
    $495 = $491 << 8;
    $496 = $495 | $493;
    $497 = $496 | $494;
    HEAP32[$pixel66$290>>2] = $497;
    $498 = (($pixel66$290) + ($pitch65$2<<2)|0);
    $499 = ($481|0)==(0);
    if ($499) {
     break;
    } else {
     $length64$191 = $481;$pixel66$290 = $498;
    }
   }
   return;
   break;
  }
  default: {
   $500 = ((($dst)) + 16|0);
   $501 = HEAP32[$500>>2]|0;
   $502 = ((($dst)) + 4|0);
   $503 = HEAP32[$502>>2]|0;
   $504 = ((($503)) + 9|0);
   $505 = HEAP8[$504>>0]|0;
   $506 = $505&255;
   $507 = (($501|0) / ($506|0))&-1;
   $508 = ($y2|0)<($y1|0);
   $509 = ((($dst)) + 20|0);
   $510 = HEAP32[$509>>2]|0;
   if ($508) {
    $516 = Math_imul($507, $y2)|0;
    $517 = (($510) + ($516<<2)|0);
    $518 = (($517) + ($x2<<2)|0);
    $519 = ($x2|0)>($x1|0);
    $pitch72$1$v = $519 ? -1 : 1;
    $pitch72$1 = (($507) + ($pitch72$1$v))|0;
    $520 = ($draw_end|0)==(0);
    $521 = (($518) + ($pitch72$1<<2)|0);
    $$24 = $520 ? $521 : $518;
    $length71$0 = $363;$pitch72$2 = $pitch72$1;$pixel73$1 = $$24;
   } else {
    $511 = Math_imul($507, $y1)|0;
    $512 = (($510) + ($511<<2)|0);
    $513 = (($512) + ($x1<<2)|0);
    $514 = ($x2|0)<($x1|0);
    $pitch72$0$v = $514 ? -1 : 1;
    $pitch72$0 = (($507) + ($pitch72$0$v))|0;
    $515 = (($y2) - ($y1))|0;
    $length71$0 = $515;$pitch72$2 = $pitch72$0;$pixel73$1 = $513;
   }
   $not$35 = ($draw_end|0)!=(0);
   $522 = $not$35&1;
   $length71$0$ = (($length71$0) + ($522))|0;
   $523 = ($length71$0$|0)==(0);
   if ($523) {
    return;
   }
   $524 = $r$0 << 16;
   $525 = $g$0 << 8;
   $526 = $525 | $b$0;
   $527 = $526 | $524;
   $length71$182 = $length71$0$;$pixel73$281 = $pixel73$1;
   while(1) {
    $528 = (($length71$182) + -1)|0;
    HEAP32[$pixel73$281>>2] = $527;
    $529 = (($pixel73$281) + ($pitch72$2<<2)|0);
    $530 = ($528|0)==(0);
    if ($530) {
     break;
    } else {
     $length71$182 = $528;$pixel73$281 = $529;
    }
   }
   return;
  }
  }
 }
 switch ($blendMode|0) {
 case 1:  {
  $531 = (($x2) - ($x1))|0;
  $532 = ($531|0)<(0);
  $533 = (0 - ($531))|0;
  $534 = $532 ? $533 : $531;
  $535 = (($y2) - ($y1))|0;
  $536 = ($535|0)<(0);
  $537 = (0 - ($535))|0;
  $538 = $536 ? $537 : $535;
  $539 = ($534|0)<($538|0);
  if ($539) {
   $543 = $534 << 1;
   $544 = (($543) - ($538))|0;
   $545 = (($534) - ($538))|0;
   $d$0 = $544;$dinc1$0 = $543;$dinc2$0$in = $545;$numpixels$0$in = $538;$xinc1$0 = 0;$yinc1$0 = 1;
  } else {
   $540 = $538 << 1;
   $541 = (($540) - ($534))|0;
   $542 = (($538) - ($534))|0;
   $d$0 = $541;$dinc1$0 = $540;$dinc2$0$in = $542;$numpixels$0$in = $534;$xinc1$0 = 1;$yinc1$0 = 0;
  }
  $dinc2$0 = $dinc2$0$in << 1;
  $546 = ($x2|0)<($x1|0);
  $547 = (0 - ($xinc1$0))|0;
  $$xinc1$0 = $546 ? $547 : $xinc1$0;
  $$25 = $546 ? -1 : 1;
  $548 = ($y2|0)<($y1|0);
  $549 = (0 - ($yinc1$0))|0;
  $yinc1$1 = $548 ? $549 : $yinc1$0;
  $yinc2$1 = $548 ? -1 : 1;
  $not$30 = ($draw_end|0)!=(0);
  $numpixels$0 = $not$30&1;
  $numpixels$0$in$numpixels$0 = (($numpixels$0$in) + ($numpixels$0))|0;
  $550 = ($numpixels$0$in$numpixels$0|0)>(0);
  if (!($550)) {
   return;
  }
  $551 = ((($dst)) + 20|0);
  $552 = HEAP32[$551>>2]|0;
  $553 = ((($dst)) + 16|0);
  $d$199 = $d$0;$i$098 = 0;$x$0100 = $x1;$y$0101 = $y1;
  while(1) {
   $554 = HEAP32[$553>>2]|0;
   $555 = Math_imul($554, $y$0101)|0;
   $556 = (($552) + ($555)|0);
   $557 = $x$0100 << 2;
   $558 = (($556) + ($557)|0);
   $559 = HEAP32[$558>>2]|0;
   $560 = $559 >>> 16;
   $561 = $560 & 255;
   $562 = $559 >>> 8;
   $563 = $562 & 255;
   $564 = $559 & 255;
   $565 = Math_imul($561, $14)|0;
   $566 = (($565>>>0) / 255)&-1;
   $567 = (($566) + ($r$0))|0;
   $568 = Math_imul($563, $14)|0;
   $569 = (($568>>>0) / 255)&-1;
   $570 = (($569) + ($g$0))|0;
   $571 = Math_imul($564, $14)|0;
   $572 = (($571>>>0) / 255)&-1;
   $573 = (($572) + ($b$0))|0;
   $574 = $567 << 16;
   $575 = $570 << 8;
   $576 = $575 | $573;
   $577 = $576 | $574;
   HEAP32[$558>>2] = $577;
   $578 = ($d$199|0)<(0);
   $dinc1$0$dinc2$0 = $578 ? $dinc1$0 : $dinc2$0;
   $$xinc1$0$$25 = $578 ? $$xinc1$0 : $$25;
   $yinc1$1$yinc2$1 = $578 ? $yinc1$1 : $yinc2$1;
   $y$1 = (($y$0101) + ($yinc1$1$yinc2$1))|0;
   $x$1 = (($$xinc1$0$$25) + ($x$0100))|0;
   $d$2 = (($dinc1$0$dinc2$0) + ($d$199))|0;
   $579 = (($i$098) + 1)|0;
   $580 = ($579|0)<($numpixels$0$in$numpixels$0|0);
   if ($580) {
    $d$199 = $d$2;$i$098 = $579;$x$0100 = $x$1;$y$0101 = $y$1;
   } else {
    break;
   }
  }
  return;
  break;
 }
 case 2:  {
  $581 = (($x2) - ($x1))|0;
  $582 = ($581|0)<(0);
  $583 = (0 - ($581))|0;
  $584 = $582 ? $583 : $581;
  $585 = (($y2) - ($y1))|0;
  $586 = ($585|0)<(0);
  $587 = (0 - ($585))|0;
  $588 = $586 ? $587 : $585;
  $589 = ($584|0)<($588|0);
  if ($589) {
   $593 = $584 << 1;
   $594 = (($593) - ($588))|0;
   $595 = (($584) - ($588))|0;
   $d87$0 = $594;$dinc188$0 = $593;$dinc289$0$in = $595;$numpixels86$0$in = $588;$xinc191$0 = 0;$yinc194$0 = 1;
  } else {
   $590 = $588 << 1;
   $591 = (($590) - ($584))|0;
   $592 = (($588) - ($584))|0;
   $d87$0 = $591;$dinc188$0 = $590;$dinc289$0$in = $592;$numpixels86$0$in = $584;$xinc191$0 = 1;$yinc194$0 = 0;
  }
  $dinc289$0 = $dinc289$0$in << 1;
  $596 = ($x2|0)<($x1|0);
  $597 = (0 - ($xinc191$0))|0;
  $$xinc191$0 = $596 ? $597 : $xinc191$0;
  $$26 = $596 ? -1 : 1;
  $598 = ($y2|0)<($y1|0);
  $599 = (0 - ($yinc194$0))|0;
  $yinc194$1 = $598 ? $599 : $yinc194$0;
  $yinc295$1 = $598 ? -1 : 1;
  $not$29 = ($draw_end|0)!=(0);
  $numpixels86$0 = $not$29&1;
  $numpixels86$0$in$numpixels86$0 = (($numpixels86$0$in) + ($numpixels86$0))|0;
  $600 = ($numpixels86$0$in$numpixels86$0|0)>(0);
  if (!($600)) {
   return;
  }
  $601 = ((($dst)) + 20|0);
  $602 = HEAP32[$601>>2]|0;
  $603 = ((($dst)) + 16|0);
  $d87$1104 = $d87$0;$i83$0103 = 0;$x90$0105 = $x1;$y93$0106 = $y1;
  while(1) {
   $604 = HEAP32[$603>>2]|0;
   $605 = Math_imul($604, $y93$0106)|0;
   $606 = (($602) + ($605)|0);
   $607 = $x90$0105 << 2;
   $608 = (($606) + ($607)|0);
   $609 = HEAP32[$608>>2]|0;
   $610 = $609 >>> 16;
   $611 = $610 & 255;
   $612 = $609 >>> 8;
   $613 = $612 & 255;
   $614 = $609 & 255;
   $615 = (($611) + ($r$0))|0;
   $616 = ($615>>>0)>(255);
   $$11 = $616 ? 255 : $615;
   $617 = (($613) + ($g$0))|0;
   $618 = ($617>>>0)>(255);
   $$7 = $618 ? 255 : $617;
   $619 = (($614) + ($b$0))|0;
   $620 = ($619>>>0)>(255);
   $$12 = $620 ? 255 : $619;
   $621 = $$11 << 16;
   $622 = $$7 << 8;
   $623 = $622 | $$12;
   $624 = $623 | $621;
   HEAP32[$608>>2] = $624;
   $625 = ($d87$1104|0)<(0);
   $dinc188$0$dinc289$0 = $625 ? $dinc188$0 : $dinc289$0;
   $$xinc191$0$$26 = $625 ? $$xinc191$0 : $$26;
   $yinc194$1$yinc295$1 = $625 ? $yinc194$1 : $yinc295$1;
   $y93$1 = (($y93$0106) + ($yinc194$1$yinc295$1))|0;
   $x90$1 = (($$xinc191$0$$26) + ($x90$0105))|0;
   $d87$2 = (($dinc188$0$dinc289$0) + ($d87$1104))|0;
   $626 = (($i83$0103) + 1)|0;
   $627 = ($626|0)<($numpixels86$0$in$numpixels86$0|0);
   if ($627) {
    $d87$1104 = $d87$2;$i83$0103 = $626;$x90$0105 = $x90$1;$y93$0106 = $y93$1;
   } else {
    break;
   }
  }
  return;
  break;
 }
 case 4:  {
  $628 = (($x2) - ($x1))|0;
  $629 = ($628|0)<(0);
  $630 = (0 - ($628))|0;
  $631 = $629 ? $630 : $628;
  $632 = (($y2) - ($y1))|0;
  $633 = ($632|0)<(0);
  $634 = (0 - ($632))|0;
  $635 = $633 ? $634 : $632;
  $636 = ($631|0)<($635|0);
  if ($636) {
   $640 = $631 << 1;
   $641 = (($640) - ($635))|0;
   $642 = (($631) - ($635))|0;
   $d105$0 = $641;$dinc1106$0 = $640;$dinc2107$0$in = $642;$numpixels104$0$in = $635;$xinc1109$0 = 0;$yinc1112$0 = 1;
  } else {
   $637 = $635 << 1;
   $638 = (($637) - ($631))|0;
   $639 = (($635) - ($631))|0;
   $d105$0 = $638;$dinc1106$0 = $637;$dinc2107$0$in = $639;$numpixels104$0$in = $631;$xinc1109$0 = 1;$yinc1112$0 = 0;
  }
  $dinc2107$0 = $dinc2107$0$in << 1;
  $643 = ($x2|0)<($x1|0);
  $644 = (0 - ($xinc1109$0))|0;
  $$xinc1109$0 = $643 ? $644 : $xinc1109$0;
  $$27 = $643 ? -1 : 1;
  $645 = ($y2|0)<($y1|0);
  $646 = (0 - ($yinc1112$0))|0;
  $yinc1112$1 = $645 ? $646 : $yinc1112$0;
  $yinc2113$1 = $645 ? -1 : 1;
  $not$ = ($draw_end|0)!=(0);
  $numpixels104$0 = $not$&1;
  $numpixels104$0$in$numpixels104$0 = (($numpixels104$0$in) + ($numpixels104$0))|0;
  $647 = ($numpixels104$0$in$numpixels104$0|0)>(0);
  if (!($647)) {
   return;
  }
  $648 = ((($dst)) + 20|0);
  $649 = HEAP32[$648>>2]|0;
  $650 = ((($dst)) + 16|0);
  $d105$1109 = $d105$0;$i101$0108 = 0;$x108$0110 = $x1;$y111$0111 = $y1;
  while(1) {
   $651 = HEAP32[$650>>2]|0;
   $652 = Math_imul($651, $y111$0111)|0;
   $653 = (($649) + ($652)|0);
   $654 = $x108$0110 << 2;
   $655 = (($653) + ($654)|0);
   $656 = HEAP32[$655>>2]|0;
   $657 = $656 >>> 16;
   $658 = $657 & 255;
   $659 = $656 >>> 8;
   $660 = $659 & 255;
   $661 = $656 & 255;
   $662 = Math_imul($658, $r$0)|0;
   $663 = (($662>>>0) / 255)&-1;
   $664 = Math_imul($660, $g$0)|0;
   $665 = (($664>>>0) / 255)&-1;
   $666 = Math_imul($661, $b$0)|0;
   $667 = (($666>>>0) / 255)&-1;
   $668 = $663 << 16;
   $669 = $665 << 8;
   $670 = $669 | $667;
   $671 = $670 | $668;
   HEAP32[$655>>2] = $671;
   $672 = ($d105$1109|0)<(0);
   $dinc1106$0$dinc2107$0 = $672 ? $dinc1106$0 : $dinc2107$0;
   $$xinc1109$0$$27 = $672 ? $$xinc1109$0 : $$27;
   $yinc1112$1$yinc2113$1 = $672 ? $yinc1112$1 : $yinc2113$1;
   $y111$1 = (($y111$0111) + ($yinc1112$1$yinc2113$1))|0;
   $x108$1 = (($$xinc1109$0$$27) + ($x108$0110))|0;
   $d105$2 = (($dinc1106$0$dinc2107$0) + ($d105$1109))|0;
   $673 = (($i101$0108) + 1)|0;
   $674 = ($673|0)<($numpixels104$0$in$numpixels104$0|0);
   if ($674) {
    $d105$1109 = $d105$2;$i101$0108 = $673;$x108$0110 = $x108$1;$y111$0111 = $y111$1;
   } else {
    break;
   }
  }
  return;
  break;
 }
 default: {
  $675 = (($x2) - ($x1))|0;
  $676 = ($675|0)<(0);
  $677 = (0 - ($675))|0;
  $678 = $676 ? $677 : $675;
  $679 = (($y2) - ($y1))|0;
  $680 = ($679|0)<(0);
  $681 = (0 - ($679))|0;
  $682 = $680 ? $681 : $679;
  $683 = ($678|0)<($682|0);
  if ($683) {
   $687 = $678 << 1;
   $688 = (($687) - ($682))|0;
   $689 = (($678) - ($682))|0;
   $d123$0 = $688;$dinc1124$0 = $687;$dinc2125$0$in = $689;$numpixels122$0$in = $682;$xinc1127$0 = 0;$yinc1130$0 = 1;
  } else {
   $684 = $682 << 1;
   $685 = (($684) - ($678))|0;
   $686 = (($682) - ($678))|0;
   $d123$0 = $685;$dinc1124$0 = $684;$dinc2125$0$in = $686;$numpixels122$0$in = $678;$xinc1127$0 = 1;$yinc1130$0 = 0;
  }
  $dinc2125$0 = $dinc2125$0$in << 1;
  $690 = ($x2|0)<($x1|0);
  $691 = (0 - ($xinc1127$0))|0;
  $$xinc1127$0 = $690 ? $691 : $xinc1127$0;
  $$28 = $690 ? -1 : 1;
  $692 = ($y2|0)<($y1|0);
  $693 = (0 - ($yinc1130$0))|0;
  $yinc1130$1 = $692 ? $693 : $yinc1130$0;
  $yinc2131$1 = $692 ? -1 : 1;
  $not$31 = ($draw_end|0)!=(0);
  $numpixels122$0 = $not$31&1;
  $numpixels122$0$in$numpixels122$0 = (($numpixels122$0$in) + ($numpixels122$0))|0;
  $694 = ($numpixels122$0$in$numpixels122$0|0)>(0);
  if (!($694)) {
   return;
  }
  $695 = ((($dst)) + 20|0);
  $696 = HEAP32[$695>>2]|0;
  $697 = ((($dst)) + 16|0);
  $698 = $r$0 << 16;
  $699 = $g$0 << 8;
  $700 = $699 | $b$0;
  $701 = $700 | $698;
  $d123$194 = $d123$0;$i119$093 = 0;$x126$095 = $x1;$y129$096 = $y1;
  while(1) {
   $702 = HEAP32[$697>>2]|0;
   $703 = Math_imul($702, $y129$096)|0;
   $704 = (($696) + ($703)|0);
   $705 = $x126$095 << 2;
   $706 = (($704) + ($705)|0);
   HEAP32[$706>>2] = $701;
   $707 = ($d123$194|0)<(0);
   $dinc1124$0$dinc2125$0 = $707 ? $dinc1124$0 : $dinc2125$0;
   $$xinc1127$0$$28 = $707 ? $$xinc1127$0 : $$28;
   $yinc1130$1$yinc2131$1 = $707 ? $yinc1130$1 : $yinc2131$1;
   $y129$1 = (($y129$096) + ($yinc1130$1$yinc2131$1))|0;
   $x126$1 = (($$xinc1127$0$$28) + ($x126$095))|0;
   $d123$2 = (($dinc1124$0$dinc2125$0) + ($d123$194))|0;
   $708 = (($i119$093) + 1)|0;
   $709 = ($708|0)<($numpixels122$0$in$numpixels122$0|0);
   if ($709) {
    $d123$194 = $d123$2;$i119$093 = $708;$x126$095 = $x126$1;$y129$096 = $y129$1;
   } else {
    break;
   }
  }
  return;
 }
 }
}
function _SDL_BlendLine_RGBA4($dst,$x1,$y1,$x2,$y2,$blendMode,$_r,$_g,$_b,$_a,$draw_end) {
 $dst = $dst|0;
 $x1 = $x1|0;
 $y1 = $y1|0;
 $x2 = $x2|0;
 $y2 = $y2|0;
 $blendMode = $blendMode|0;
 $_r = $_r|0;
 $_g = $_g|0;
 $_b = $_b|0;
 $_a = $_a|0;
 $draw_end = $draw_end|0;
 var $$ = 0, $$10 = 0, $$11 = 0, $$12 = 0, $$13 = 0, $$14 = 0, $$15 = 0, $$16 = 0, $$17 = 0, $$18 = 0, $$19 = 0, $$2 = 0, $$20 = 0, $$21 = 0, $$22 = 0, $$23 = 0, $$24 = 0, $$25 = 0, $$26 = 0, $$27 = 0;
 var $$28 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$6 = 0, $$7 = 0, $$8 = 0, $$9 = 0, $$phi$trans$insert = 0, $$phi$trans$insert130 = 0, $$phi$trans$insert133 = 0, $$phi$trans$insert135 = 0, $$phi$trans$insert136 = 0, $$phi$trans$insert139 = 0, $$phi$trans$insert144 = 0, $$phi$trans$insert145 = 0, $$phi$trans$insert148 = 0, $$phi$trans$insert150 = 0, $$phi$trans$insert151 = 0, $$phi$trans$insert154 = 0;
 var $$phi$trans$insert159 = 0, $$phi$trans$insert160 = 0, $$phi$trans$insert163 = 0, $$phi$trans$insert165 = 0, $$phi$trans$insert166 = 0, $$phi$trans$insert169 = 0, $$phi$trans$insert173 = 0, $$phi$trans$insert174 = 0, $$phi$trans$insert177 = 0, $$phi$trans$insert179 = 0, $$phi$trans$insert180 = 0, $$phi$trans$insert183 = 0, $$phi$trans$insert187 = 0, $$phi$trans$insert188 = 0, $$phi$trans$insert191 = 0, $$phi$trans$insert193 = 0, $$phi$trans$insert194 = 0, $$phi$trans$insert197 = 0, $$phi$trans$insert201 = 0, $$phi$trans$insert202 = 0;
 var $$phi$trans$insert205 = 0, $$phi$trans$insert207 = 0, $$phi$trans$insert208 = 0, $$phi$trans$insert211 = 0, $$phi$trans$insert215 = 0, $$phi$trans$insert216 = 0, $$phi$trans$insert219 = 0, $$phi$trans$insert221 = 0, $$phi$trans$insert222 = 0, $$phi$trans$insert225 = 0, $$phi$trans$insert229 = 0, $$phi$trans$insert230 = 0, $$phi$trans$insert233 = 0, $$phi$trans$insert235 = 0, $$phi$trans$insert236 = 0, $$phi$trans$insert239 = 0, $$phi$trans$insert243 = 0, $$phi$trans$insert244 = 0, $$phi$trans$insert247 = 0, $$phi$trans$insert249 = 0;
 var $$phi$trans$insert250 = 0, $$phi$trans$insert253 = 0, $$phi$trans$insert257 = 0, $$phi$trans$insert258 = 0, $$phi$trans$insert261 = 0, $$phi$trans$insert263 = 0, $$phi$trans$insert264 = 0, $$phi$trans$insert267 = 0, $$phi$trans$insert271 = 0, $$phi$trans$insert272 = 0, $$phi$trans$insert275 = 0, $$phi$trans$insert277 = 0, $$phi$trans$insert278 = 0, $$phi$trans$insert281 = 0, $$phi$trans$insert285 = 0, $$phi$trans$insert286 = 0, $$phi$trans$insert289 = 0, $$phi$trans$insert291 = 0, $$phi$trans$insert292 = 0, $$phi$trans$insert295 = 0;
 var $$pre = 0, $$pre131 = 0, $$pre134 = 0, $$pre137 = 0, $$pre140 = 0, $$pre141 = 0, $$pre146 = 0, $$pre149 = 0, $$pre152 = 0, $$pre155 = 0, $$pre156 = 0, $$pre161 = 0, $$pre164 = 0, $$pre167 = 0, $$pre170 = 0, $$pre175 = 0, $$pre178 = 0, $$pre181 = 0, $$pre184 = 0, $$pre189 = 0;
 var $$pre192 = 0, $$pre195 = 0, $$pre198 = 0, $$pre203 = 0, $$pre206 = 0, $$pre209 = 0, $$pre212 = 0, $$pre217 = 0, $$pre220 = 0, $$pre223 = 0, $$pre226 = 0, $$pre231 = 0, $$pre234 = 0, $$pre237 = 0, $$pre240 = 0, $$pre245 = 0, $$pre248 = 0, $$pre251 = 0, $$pre254 = 0, $$pre259 = 0;
 var $$pre262 = 0, $$pre265 = 0, $$pre268 = 0, $$pre273 = 0, $$pre276 = 0, $$pre279 = 0, $$pre282 = 0, $$pre287 = 0, $$pre290 = 0, $$pre293 = 0, $$pre296 = 0, $$xinc1$0 = 0, $$xinc1$0$$25 = 0, $$xinc1109$0 = 0, $$xinc1109$0$$27 = 0, $$xinc1127$0 = 0, $$xinc1127$0$$28 = 0, $$xinc191$0 = 0, $$xinc191$0$$26 = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0;
 var $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0;
 var $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0;
 var $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0;
 var $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0, $1088 = 0;
 var $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0;
 var $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0;
 var $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0;
 var $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0, $116 = 0;
 var $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0, $1178 = 0;
 var $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0, $1196 = 0;
 var $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0, $1213 = 0;
 var $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0, $1231 = 0;
 var $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0, $125 = 0;
 var $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0, $1267 = 0, $1268 = 0;
 var $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0, $1285 = 0, $1286 = 0;
 var $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0, $1302 = 0, $1303 = 0;
 var $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0;
 var $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0;
 var $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0;
 var $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0;
 var $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0;
 var $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0;
 var $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0;
 var $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0;
 var $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0;
 var $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0;
 var $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0;
 var $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0;
 var $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0;
 var $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0;
 var $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0;
 var $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0;
 var $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0;
 var $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0;
 var $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0;
 var $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0;
 var $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0;
 var $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0;
 var $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0;
 var $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0;
 var $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0;
 var $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0;
 var $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0;
 var $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0;
 var $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0;
 var $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0;
 var $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0;
 var $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0;
 var $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0;
 var $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0;
 var $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0;
 var $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0;
 var $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0;
 var $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0;
 var $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0;
 var $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0;
 var $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0;
 var $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0;
 var $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $a$0 = 0, $b$0 = 0, $blendMode$off = 0, $d$0 = 0;
 var $d$199 = 0, $d$2 = 0, $d105$0 = 0, $d105$1109 = 0, $d105$2 = 0, $d123$0 = 0, $d123$194 = 0, $d123$2 = 0, $d87$0 = 0, $d87$1104 = 0, $d87$2 = 0, $dinc1$0 = 0, $dinc1$0$dinc2$0 = 0, $dinc1106$0 = 0, $dinc1106$0$dinc2107$0 = 0, $dinc1124$0 = 0, $dinc1124$0$dinc2125$0 = 0, $dinc188$0 = 0, $dinc188$0$dinc289$0 = 0, $dinc2$0 = 0;
 var $dinc2$0$in = 0, $dinc2107$0 = 0, $dinc2107$0$in = 0, $dinc2125$0 = 0, $dinc2125$0$in = 0, $dinc289$0 = 0, $dinc289$0$in = 0, $g$0 = 0, $i$098 = 0, $i101$0108 = 0, $i119$093 = 0, $i83$0103 = 0, $length$0$ph = 0, $length$061 = 0, $length1$0$ph = 0, $length1$064 = 0, $length15$0$ph = 0, $length15$059 = 0, $length22$0$ph = 0, $length22$073 = 0;
 var $length29$0$ph = 0, $length29$076 = 0, $length36$0$ph = 0, $length36$079 = 0, $length43$0$ph = 0, $length43$070 = 0, $length50$0 = 0, $length50$0$ = 0, $length50$185 = 0, $length57$0 = 0, $length57$0$ = 0, $length57$188 = 0, $length64$0 = 0, $length64$0$ = 0, $length64$191 = 0, $length71$0 = 0, $length71$0$ = 0, $length71$181 = 0, $length8$0$ph = 0, $length8$067 = 0;
 var $not$ = 0, $not$29 = 0, $not$30 = 0, $not$31 = 0, $not$32 = 0, $not$33 = 0, $not$34 = 0, $not$35 = 0, $numpixels$0 = 0, $numpixels$0$in = 0, $numpixels$0$in$numpixels$0 = 0, $numpixels104$0 = 0, $numpixels104$0$in = 0, $numpixels104$0$in$numpixels104$0 = 0, $numpixels122$0 = 0, $numpixels122$0$in = 0, $numpixels122$0$in$numpixels122$0 = 0, $numpixels86$0 = 0, $numpixels86$0$in = 0, $numpixels86$0$in$numpixels86$0 = 0;
 var $pitch51$0 = 0, $pitch51$0$v = 0, $pitch51$1 = 0, $pitch51$1$v = 0, $pitch51$2 = 0, $pitch58$0 = 0, $pitch58$0$v = 0, $pitch58$1 = 0, $pitch58$1$v = 0, $pitch58$2 = 0, $pitch65$0 = 0, $pitch65$0$v = 0, $pitch65$1 = 0, $pitch65$1$v = 0, $pitch65$2 = 0, $pitch72$0 = 0, $pitch72$0$v = 0, $pitch72$1 = 0, $pitch72$1$v = 0, $pitch72$2 = 0;
 var $pixel$1$ph = 0, $pixel$160 = 0, $pixel10$1$ph = 0, $pixel10$166 = 0, $pixel17$1$ph = 0, $pixel17$158 = 0, $pixel24$1$ph = 0, $pixel24$172 = 0, $pixel3$1$ph = 0, $pixel3$163 = 0, $pixel31$1$ph = 0, $pixel31$175 = 0, $pixel38$1$ph = 0, $pixel38$178 = 0, $pixel45$1$ph = 0, $pixel45$169 = 0, $pixel52$1 = 0, $pixel52$284 = 0, $pixel59$1 = 0, $pixel59$287 = 0;
 var $pixel66$1 = 0, $pixel66$290 = 0, $pixel73$1 = 0, $pixel73$282 = 0, $r$0 = 0, $x$0100 = 0, $x$1 = 0, $x108$0110 = 0, $x108$1 = 0, $x126$095 = 0, $x126$1 = 0, $x90$0105 = 0, $x90$1 = 0, $xinc1$0 = 0, $xinc1109$0 = 0, $xinc1127$0 = 0, $xinc191$0 = 0, $y$0101 = 0, $y$1 = 0, $y111$0111 = 0;
 var $y111$1 = 0, $y129$096 = 0, $y129$1 = 0, $y93$0106 = 0, $y93$1 = 0, $yinc1$0 = 0, $yinc1$1 = 0, $yinc1$1$yinc2$1 = 0, $yinc1112$0 = 0, $yinc1112$1 = 0, $yinc1112$1$yinc2113$1 = 0, $yinc1130$0 = 0, $yinc1130$1 = 0, $yinc1130$1$yinc2131$1 = 0, $yinc194$0 = 0, $yinc194$1 = 0, $yinc194$1$yinc295$1 = 0, $yinc2$1 = 0, $yinc2113$1 = 0, $yinc2131$1 = 0;
 var $yinc295$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($dst)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $blendMode$off = (($blendMode) + -1)|0;
 $2 = ($blendMode$off>>>0)<(2);
 $3 = $_r&255;
 if ($2) {
  $4 = $_a&255;
  $5 = Math_imul($4, $3)|0;
  $6 = (($5>>>0) / 255)&-1;
  $7 = $_g&255;
  $8 = Math_imul($4, $7)|0;
  $9 = (($8>>>0) / 255)&-1;
  $10 = $_b&255;
  $11 = Math_imul($4, $10)|0;
  $12 = (($11>>>0) / 255)&-1;
  $a$0 = $4;$b$0 = $12;$g$0 = $9;$r$0 = $6;
 } else {
  $13 = $_g&255;
  $14 = $_b&255;
  $15 = $_a&255;
  $a$0 = $15;$b$0 = $14;$g$0 = $13;$r$0 = $3;
 }
 $16 = $a$0 ^ 255;
 $17 = ($y2|0)==($y1|0);
 if ($17) {
  switch ($blendMode|0) {
  case 1:  {
   $18 = ((($dst)) + 16|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ((($1)) + 9|0);
   $21 = HEAP8[$20>>0]|0;
   $22 = $21&255;
   $23 = (($19|0) / ($22|0))&-1;
   $24 = ($x2|0)<($x1|0);
   $25 = ((($dst)) + 20|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = Math_imul($23, $y2)|0;
   $28 = (($26) + ($27<<2)|0);
   if ($24) {
    $34 = (($28) + ($x2<<2)|0);
    $35 = ($draw_end|0)!=(0);
    $36 = ((($34)) + 4|0);
    $$13 = $35 ? $34 : $36;
    $37 = (($x1) - ($x2))|0;
    $38 = $35&1;
    $39 = (($37) + ($38))|0;
    $length$0$ph = $39;$pixel$1$ph = $$13;
   } else {
    $29 = (($28) + ($x1<<2)|0);
    $30 = ($draw_end|0)!=(0);
    $31 = (($x2) - ($x1))|0;
    $32 = $30&1;
    $33 = (($31) + ($32))|0;
    $length$0$ph = $33;$pixel$1$ph = $29;
   }
   $40 = ($length$0$ph|0)==(0);
   if ($40) {
    return;
   }
   $41 = ((($1)) + 12|0);
   $42 = ((($1)) + 32|0);
   $43 = ((($1)) + 28|0);
   $44 = ((($1)) + 16|0);
   $45 = ((($1)) + 20|0);
   $46 = ((($1)) + 24|0);
   $47 = HEAP32[$42>>2]|0;
   $48 = HEAP32[$43>>2]|0;
   $$phi$trans$insert285 = $48 & 255;
   $$phi$trans$insert286 = (125924 + ($$phi$trans$insert285<<2)|0);
   $$pre287 = HEAP32[$$phi$trans$insert286>>2]|0;
   $49 = $47 >>> 8;
   $50 = $49 & 255;
   $51 = $48 >>> 8;
   $52 = $51 & 255;
   $$phi$trans$insert289 = (125924 + ($52<<2)|0);
   $$pre290 = HEAP32[$$phi$trans$insert289>>2]|0;
   $53 = $47 >>> 16;
   $54 = $48 >>> 16;
   $$phi$trans$insert291 = $54 & 255;
   $$phi$trans$insert292 = (125924 + ($$phi$trans$insert291<<2)|0);
   $$pre293 = HEAP32[$$phi$trans$insert292>>2]|0;
   $55 = $47 >>> 24;
   $56 = $48 >>> 24;
   $$phi$trans$insert295 = (125924 + ($56<<2)|0);
   $$pre296 = HEAP32[$$phi$trans$insert295>>2]|0;
   $57 = $47 & 255;
   $58 = $53 & 255;
   $length$061 = $length$0$ph;$pixel$160 = $pixel$1$ph;
   while(1) {
    $59 = (($length$061) + -1)|0;
    $60 = HEAP32[$pixel$160>>2]|0;
    $61 = HEAP32[$41>>2]|0;
    $62 = $61 & $60;
    $63 = $62 >>> $57;
    $64 = (($$pre287) + ($63)|0);
    $65 = HEAP8[$64>>0]|0;
    $66 = $65&255;
    $67 = HEAP32[$44>>2]|0;
    $68 = $67 & $60;
    $69 = $68 >>> $50;
    $70 = (($$pre290) + ($69)|0);
    $71 = HEAP8[$70>>0]|0;
    $72 = $71&255;
    $73 = HEAP32[$45>>2]|0;
    $74 = $73 & $60;
    $75 = $74 >>> $58;
    $76 = (($$pre293) + ($75)|0);
    $77 = HEAP8[$76>>0]|0;
    $78 = $77&255;
    $79 = HEAP32[$46>>2]|0;
    $80 = $79 & $60;
    $81 = $80 >>> $55;
    $82 = (($$pre296) + ($81)|0);
    $83 = HEAP8[$82>>0]|0;
    $84 = $83&255;
    $85 = Math_imul($66, $16)|0;
    $86 = (($85>>>0) / 255)&-1;
    $87 = (($86) + ($r$0))|0;
    $88 = Math_imul($72, $16)|0;
    $89 = (($88>>>0) / 255)&-1;
    $90 = (($89) + ($g$0))|0;
    $91 = Math_imul($78, $16)|0;
    $92 = (($91>>>0) / 255)&-1;
    $93 = (($92) + ($b$0))|0;
    $94 = Math_imul($84, $16)|0;
    $95 = (($94>>>0) / 255)&-1;
    $96 = (($95) + ($a$0))|0;
    $97 = $87 >>> $$phi$trans$insert285;
    $98 = $97 << $57;
    $99 = $90 >>> $52;
    $100 = $99 << $50;
    $101 = $100 | $98;
    $102 = $93 >>> $$phi$trans$insert291;
    $103 = $102 << $58;
    $104 = $101 | $103;
    $105 = $96 >>> $56;
    $106 = $105 << $55;
    $107 = $104 | $106;
    HEAP32[$pixel$160>>2] = $107;
    $108 = ((($pixel$160)) + 4|0);
    $109 = ($59|0)==(0);
    if ($109) {
     break;
    } else {
     $length$061 = $59;$pixel$160 = $108;
    }
   }
   return;
   break;
  }
  case 2:  {
   $110 = ((($dst)) + 16|0);
   $111 = HEAP32[$110>>2]|0;
   $112 = ((($1)) + 9|0);
   $113 = HEAP8[$112>>0]|0;
   $114 = $113&255;
   $115 = (($111|0) / ($114|0))&-1;
   $116 = ($x2|0)<($x1|0);
   $117 = ((($dst)) + 20|0);
   $118 = HEAP32[$117>>2]|0;
   $119 = Math_imul($115, $y2)|0;
   $120 = (($118) + ($119<<2)|0);
   if ($116) {
    $126 = (($120) + ($x2<<2)|0);
    $127 = ($draw_end|0)!=(0);
    $128 = ((($126)) + 4|0);
    $$14 = $127 ? $126 : $128;
    $129 = (($x1) - ($x2))|0;
    $130 = $127&1;
    $131 = (($129) + ($130))|0;
    $length1$0$ph = $131;$pixel3$1$ph = $$14;
   } else {
    $121 = (($120) + ($x1<<2)|0);
    $122 = ($draw_end|0)!=(0);
    $123 = (($x2) - ($x1))|0;
    $124 = $122&1;
    $125 = (($123) + ($124))|0;
    $length1$0$ph = $125;$pixel3$1$ph = $121;
   }
   $132 = ($length1$0$ph|0)==(0);
   if ($132) {
    return;
   }
   $133 = ((($1)) + 12|0);
   $134 = ((($1)) + 32|0);
   $135 = ((($1)) + 28|0);
   $136 = ((($1)) + 16|0);
   $137 = ((($1)) + 20|0);
   $138 = ((($1)) + 24|0);
   $139 = HEAP32[$134>>2]|0;
   $140 = HEAP32[$135>>2]|0;
   $$phi$trans$insert271 = $140 & 255;
   $$phi$trans$insert272 = (125924 + ($$phi$trans$insert271<<2)|0);
   $$pre273 = HEAP32[$$phi$trans$insert272>>2]|0;
   $141 = $139 >>> 8;
   $142 = $141 & 255;
   $143 = $140 >>> 8;
   $144 = $143 & 255;
   $$phi$trans$insert275 = (125924 + ($144<<2)|0);
   $$pre276 = HEAP32[$$phi$trans$insert275>>2]|0;
   $145 = $139 >>> 16;
   $146 = $140 >>> 16;
   $$phi$trans$insert277 = $146 & 255;
   $$phi$trans$insert278 = (125924 + ($$phi$trans$insert277<<2)|0);
   $$pre279 = HEAP32[$$phi$trans$insert278>>2]|0;
   $147 = $139 >>> 24;
   $148 = $140 >>> 24;
   $$phi$trans$insert281 = (125924 + ($148<<2)|0);
   $$pre282 = HEAP32[$$phi$trans$insert281>>2]|0;
   $149 = $139 & 255;
   $150 = $145 & 255;
   $length1$064 = $length1$0$ph;$pixel3$163 = $pixel3$1$ph;
   while(1) {
    $151 = (($length1$064) + -1)|0;
    $152 = HEAP32[$pixel3$163>>2]|0;
    $153 = HEAP32[$133>>2]|0;
    $154 = $153 & $152;
    $155 = $154 >>> $149;
    $156 = (($$pre273) + ($155)|0);
    $157 = HEAP8[$156>>0]|0;
    $158 = $157&255;
    $159 = HEAP32[$136>>2]|0;
    $160 = $159 & $152;
    $161 = $160 >>> $142;
    $162 = (($$pre276) + ($161)|0);
    $163 = HEAP8[$162>>0]|0;
    $164 = $163&255;
    $165 = HEAP32[$137>>2]|0;
    $166 = $165 & $152;
    $167 = $166 >>> $150;
    $168 = (($$pre279) + ($167)|0);
    $169 = HEAP8[$168>>0]|0;
    $170 = $169&255;
    $171 = HEAP32[$138>>2]|0;
    $172 = $171 & $152;
    $173 = $172 >>> $147;
    $174 = (($$pre282) + ($173)|0);
    $175 = HEAP8[$174>>0]|0;
    $176 = $175&255;
    $177 = (($158) + ($r$0))|0;
    $178 = ($177>>>0)>(255);
    $$ = $178 ? 255 : $177;
    $179 = (($164) + ($g$0))|0;
    $180 = ($179>>>0)>(255);
    $$8 = $180 ? 255 : $179;
    $181 = (($170) + ($b$0))|0;
    $182 = ($181>>>0)>(255);
    $$2 = $182 ? 255 : $181;
    $183 = $$ >>> $$phi$trans$insert271;
    $184 = $183 << $149;
    $185 = $$8 >>> $144;
    $186 = $185 << $142;
    $187 = $186 | $184;
    $188 = $$2 >>> $$phi$trans$insert277;
    $189 = $188 << $150;
    $190 = $187 | $189;
    $191 = $176 >>> $148;
    $192 = $191 << $147;
    $193 = $190 | $192;
    HEAP32[$pixel3$163>>2] = $193;
    $194 = ((($pixel3$163)) + 4|0);
    $195 = ($151|0)==(0);
    if ($195) {
     break;
    } else {
     $length1$064 = $151;$pixel3$163 = $194;
    }
   }
   return;
   break;
  }
  case 4:  {
   $196 = ((($dst)) + 16|0);
   $197 = HEAP32[$196>>2]|0;
   $198 = ((($1)) + 9|0);
   $199 = HEAP8[$198>>0]|0;
   $200 = $199&255;
   $201 = (($197|0) / ($200|0))&-1;
   $202 = ($x2|0)<($x1|0);
   $203 = ((($dst)) + 20|0);
   $204 = HEAP32[$203>>2]|0;
   $205 = Math_imul($201, $y2)|0;
   $206 = (($204) + ($205<<2)|0);
   if ($202) {
    $212 = (($206) + ($x2<<2)|0);
    $213 = ($draw_end|0)!=(0);
    $214 = ((($212)) + 4|0);
    $$15 = $213 ? $212 : $214;
    $215 = (($x1) - ($x2))|0;
    $216 = $213&1;
    $217 = (($215) + ($216))|0;
    $length8$0$ph = $217;$pixel10$1$ph = $$15;
   } else {
    $207 = (($206) + ($x1<<2)|0);
    $208 = ($draw_end|0)!=(0);
    $209 = (($x2) - ($x1))|0;
    $210 = $208&1;
    $211 = (($209) + ($210))|0;
    $length8$0$ph = $211;$pixel10$1$ph = $207;
   }
   $218 = ($length8$0$ph|0)==(0);
   if ($218) {
    return;
   }
   $219 = ((($1)) + 12|0);
   $220 = ((($1)) + 32|0);
   $221 = ((($1)) + 28|0);
   $222 = ((($1)) + 16|0);
   $223 = ((($1)) + 20|0);
   $224 = ((($1)) + 24|0);
   $225 = HEAP32[$220>>2]|0;
   $226 = HEAP32[$221>>2]|0;
   $$phi$trans$insert257 = $226 & 255;
   $$phi$trans$insert258 = (125924 + ($$phi$trans$insert257<<2)|0);
   $$pre259 = HEAP32[$$phi$trans$insert258>>2]|0;
   $227 = $225 >>> 8;
   $228 = $227 & 255;
   $229 = $226 >>> 8;
   $230 = $229 & 255;
   $$phi$trans$insert261 = (125924 + ($230<<2)|0);
   $$pre262 = HEAP32[$$phi$trans$insert261>>2]|0;
   $231 = $225 >>> 16;
   $232 = $226 >>> 16;
   $$phi$trans$insert263 = $232 & 255;
   $$phi$trans$insert264 = (125924 + ($$phi$trans$insert263<<2)|0);
   $$pre265 = HEAP32[$$phi$trans$insert264>>2]|0;
   $233 = $225 >>> 24;
   $234 = $226 >>> 24;
   $$phi$trans$insert267 = (125924 + ($234<<2)|0);
   $$pre268 = HEAP32[$$phi$trans$insert267>>2]|0;
   $235 = $225 & 255;
   $236 = $231 & 255;
   $length8$067 = $length8$0$ph;$pixel10$166 = $pixel10$1$ph;
   while(1) {
    $237 = (($length8$067) + -1)|0;
    $238 = HEAP32[$pixel10$166>>2]|0;
    $239 = HEAP32[$219>>2]|0;
    $240 = $239 & $238;
    $241 = $240 >>> $235;
    $242 = (($$pre259) + ($241)|0);
    $243 = HEAP8[$242>>0]|0;
    $244 = $243&255;
    $245 = HEAP32[$222>>2]|0;
    $246 = $245 & $238;
    $247 = $246 >>> $228;
    $248 = (($$pre262) + ($247)|0);
    $249 = HEAP8[$248>>0]|0;
    $250 = $249&255;
    $251 = HEAP32[$223>>2]|0;
    $252 = $251 & $238;
    $253 = $252 >>> $236;
    $254 = (($$pre265) + ($253)|0);
    $255 = HEAP8[$254>>0]|0;
    $256 = $255&255;
    $257 = HEAP32[$224>>2]|0;
    $258 = $257 & $238;
    $259 = $258 >>> $233;
    $260 = (($$pre268) + ($259)|0);
    $261 = HEAP8[$260>>0]|0;
    $262 = $261&255;
    $263 = Math_imul($244, $r$0)|0;
    $264 = (($263>>>0) / 255)&-1;
    $265 = Math_imul($250, $g$0)|0;
    $266 = (($265>>>0) / 255)&-1;
    $267 = Math_imul($256, $b$0)|0;
    $268 = (($267>>>0) / 255)&-1;
    $269 = $264 >>> $$phi$trans$insert257;
    $270 = $269 << $235;
    $271 = $266 >>> $230;
    $272 = $271 << $228;
    $273 = $272 | $270;
    $274 = $268 >>> $$phi$trans$insert263;
    $275 = $274 << $236;
    $276 = $273 | $275;
    $277 = $262 >>> $234;
    $278 = $277 << $233;
    $279 = $276 | $278;
    HEAP32[$pixel10$166>>2] = $279;
    $280 = ((($pixel10$166)) + 4|0);
    $281 = ($237|0)==(0);
    if ($281) {
     break;
    } else {
     $length8$067 = $237;$pixel10$166 = $280;
    }
   }
   return;
   break;
  }
  default: {
   $282 = ((($dst)) + 16|0);
   $283 = HEAP32[$282>>2]|0;
   $284 = ((($1)) + 9|0);
   $285 = HEAP8[$284>>0]|0;
   $286 = $285&255;
   $287 = (($283|0) / ($286|0))&-1;
   $288 = ($x2|0)<($x1|0);
   $289 = ((($dst)) + 20|0);
   $290 = HEAP32[$289>>2]|0;
   $291 = Math_imul($287, $y2)|0;
   $292 = (($290) + ($291<<2)|0);
   if ($288) {
    $298 = (($292) + ($x2<<2)|0);
    $299 = ($draw_end|0)!=(0);
    $300 = ((($298)) + 4|0);
    $$16 = $299 ? $298 : $300;
    $301 = (($x1) - ($x2))|0;
    $302 = $299&1;
    $303 = (($301) + ($302))|0;
    $length15$0$ph = $303;$pixel17$1$ph = $$16;
   } else {
    $293 = (($292) + ($x1<<2)|0);
    $294 = ($draw_end|0)!=(0);
    $295 = (($x2) - ($x1))|0;
    $296 = $294&1;
    $297 = (($295) + ($296))|0;
    $length15$0$ph = $297;$pixel17$1$ph = $293;
   }
   $304 = ($length15$0$ph|0)==(0);
   if ($304) {
    return;
   }
   $305 = ((($1)) + 28|0);
   $306 = HEAP32[$305>>2]|0;
   $307 = $306 & 255;
   $308 = $r$0 >>> $307;
   $309 = ((($1)) + 32|0);
   $310 = HEAP32[$309>>2]|0;
   $311 = $310 & 255;
   $312 = $308 << $311;
   $313 = $306 >>> 8;
   $314 = $313 & 255;
   $315 = $g$0 >>> $314;
   $316 = $310 >>> 8;
   $317 = $316 & 255;
   $318 = $315 << $317;
   $319 = $318 | $312;
   $320 = $306 >>> 16;
   $321 = $320 & 255;
   $322 = $b$0 >>> $321;
   $323 = $310 >>> 16;
   $324 = $323 & 255;
   $325 = $322 << $324;
   $326 = $319 | $325;
   $327 = $306 >>> 24;
   $328 = $a$0 >>> $327;
   $329 = $310 >>> 24;
   $330 = $328 << $329;
   $331 = $326 | $330;
   $length15$059 = $length15$0$ph;$pixel17$158 = $pixel17$1$ph;
   while(1) {
    $332 = (($length15$059) + -1)|0;
    HEAP32[$pixel17$158>>2] = $331;
    $333 = ((($pixel17$158)) + 4|0);
    $334 = ($332|0)==(0);
    if ($334) {
     break;
    } else {
     $length15$059 = $332;$pixel17$158 = $333;
    }
   }
   return;
  }
  }
 }
 $335 = ($x2|0)==($x1|0);
 if ($335) {
  switch ($blendMode|0) {
  case 1:  {
   $336 = ((($dst)) + 16|0);
   $337 = HEAP32[$336>>2]|0;
   $338 = ((($1)) + 9|0);
   $339 = HEAP8[$338>>0]|0;
   $340 = $339&255;
   $341 = (($337|0) / ($340|0))&-1;
   $342 = ($y2|0)<($y1|0);
   $343 = ((($dst)) + 20|0);
   $344 = HEAP32[$343>>2]|0;
   if ($342) {
    $352 = Math_imul($341, $y2)|0;
    $353 = (($344) + ($352<<2)|0);
    $354 = (($353) + ($x2<<2)|0);
    $355 = ($draw_end|0)!=(0);
    $356 = (($354) + ($341<<2)|0);
    $$17 = $355 ? $354 : $356;
    $357 = (($y1) - ($y2))|0;
    $358 = $355&1;
    $359 = (($357) + ($358))|0;
    $length22$0$ph = $359;$pixel24$1$ph = $$17;
   } else {
    $345 = Math_imul($341, $y1)|0;
    $346 = (($344) + ($345<<2)|0);
    $347 = (($346) + ($x2<<2)|0);
    $348 = ($draw_end|0)!=(0);
    $349 = (($y2) - ($y1))|0;
    $350 = $348&1;
    $351 = (($349) + ($350))|0;
    $length22$0$ph = $351;$pixel24$1$ph = $347;
   }
   $360 = ($length22$0$ph|0)==(0);
   if ($360) {
    return;
   }
   $361 = ((($1)) + 12|0);
   $362 = ((($1)) + 32|0);
   $363 = ((($1)) + 28|0);
   $364 = ((($1)) + 16|0);
   $365 = ((($1)) + 20|0);
   $366 = ((($1)) + 24|0);
   $367 = HEAP32[$362>>2]|0;
   $368 = HEAP32[$363>>2]|0;
   $$phi$trans$insert243 = $368 & 255;
   $$phi$trans$insert244 = (125924 + ($$phi$trans$insert243<<2)|0);
   $$pre245 = HEAP32[$$phi$trans$insert244>>2]|0;
   $369 = $367 >>> 8;
   $370 = $369 & 255;
   $371 = $368 >>> 8;
   $372 = $371 & 255;
   $$phi$trans$insert247 = (125924 + ($372<<2)|0);
   $$pre248 = HEAP32[$$phi$trans$insert247>>2]|0;
   $373 = $367 >>> 16;
   $374 = $368 >>> 16;
   $$phi$trans$insert249 = $374 & 255;
   $$phi$trans$insert250 = (125924 + ($$phi$trans$insert249<<2)|0);
   $$pre251 = HEAP32[$$phi$trans$insert250>>2]|0;
   $375 = $367 >>> 24;
   $376 = $368 >>> 24;
   $$phi$trans$insert253 = (125924 + ($376<<2)|0);
   $$pre254 = HEAP32[$$phi$trans$insert253>>2]|0;
   $377 = $367 & 255;
   $378 = $373 & 255;
   $length22$073 = $length22$0$ph;$pixel24$172 = $pixel24$1$ph;
   while(1) {
    $379 = (($length22$073) + -1)|0;
    $380 = HEAP32[$pixel24$172>>2]|0;
    $381 = HEAP32[$361>>2]|0;
    $382 = $381 & $380;
    $383 = $382 >>> $377;
    $384 = (($$pre245) + ($383)|0);
    $385 = HEAP8[$384>>0]|0;
    $386 = $385&255;
    $387 = HEAP32[$364>>2]|0;
    $388 = $387 & $380;
    $389 = $388 >>> $370;
    $390 = (($$pre248) + ($389)|0);
    $391 = HEAP8[$390>>0]|0;
    $392 = $391&255;
    $393 = HEAP32[$365>>2]|0;
    $394 = $393 & $380;
    $395 = $394 >>> $378;
    $396 = (($$pre251) + ($395)|0);
    $397 = HEAP8[$396>>0]|0;
    $398 = $397&255;
    $399 = HEAP32[$366>>2]|0;
    $400 = $399 & $380;
    $401 = $400 >>> $375;
    $402 = (($$pre254) + ($401)|0);
    $403 = HEAP8[$402>>0]|0;
    $404 = $403&255;
    $405 = Math_imul($386, $16)|0;
    $406 = (($405>>>0) / 255)&-1;
    $407 = (($406) + ($r$0))|0;
    $408 = Math_imul($392, $16)|0;
    $409 = (($408>>>0) / 255)&-1;
    $410 = (($409) + ($g$0))|0;
    $411 = Math_imul($398, $16)|0;
    $412 = (($411>>>0) / 255)&-1;
    $413 = (($412) + ($b$0))|0;
    $414 = Math_imul($404, $16)|0;
    $415 = (($414>>>0) / 255)&-1;
    $416 = (($415) + ($a$0))|0;
    $417 = $407 >>> $$phi$trans$insert243;
    $418 = $417 << $377;
    $419 = $410 >>> $372;
    $420 = $419 << $370;
    $421 = $420 | $418;
    $422 = $413 >>> $$phi$trans$insert249;
    $423 = $422 << $378;
    $424 = $421 | $423;
    $425 = $416 >>> $376;
    $426 = $425 << $375;
    $427 = $424 | $426;
    HEAP32[$pixel24$172>>2] = $427;
    $428 = (($pixel24$172) + ($341<<2)|0);
    $429 = ($379|0)==(0);
    if ($429) {
     break;
    } else {
     $length22$073 = $379;$pixel24$172 = $428;
    }
   }
   return;
   break;
  }
  case 2:  {
   $430 = ((($dst)) + 16|0);
   $431 = HEAP32[$430>>2]|0;
   $432 = ((($1)) + 9|0);
   $433 = HEAP8[$432>>0]|0;
   $434 = $433&255;
   $435 = (($431|0) / ($434|0))&-1;
   $436 = ($y2|0)<($y1|0);
   $437 = ((($dst)) + 20|0);
   $438 = HEAP32[$437>>2]|0;
   if ($436) {
    $446 = Math_imul($435, $y2)|0;
    $447 = (($438) + ($446<<2)|0);
    $448 = (($447) + ($x2<<2)|0);
    $449 = ($draw_end|0)!=(0);
    $450 = (($448) + ($435<<2)|0);
    $$18 = $449 ? $448 : $450;
    $451 = (($y1) - ($y2))|0;
    $452 = $449&1;
    $453 = (($451) + ($452))|0;
    $length29$0$ph = $453;$pixel31$1$ph = $$18;
   } else {
    $439 = Math_imul($435, $y1)|0;
    $440 = (($438) + ($439<<2)|0);
    $441 = (($440) + ($x2<<2)|0);
    $442 = ($draw_end|0)!=(0);
    $443 = (($y2) - ($y1))|0;
    $444 = $442&1;
    $445 = (($443) + ($444))|0;
    $length29$0$ph = $445;$pixel31$1$ph = $441;
   }
   $454 = ($length29$0$ph|0)==(0);
   if ($454) {
    return;
   }
   $455 = ((($1)) + 12|0);
   $456 = ((($1)) + 32|0);
   $457 = ((($1)) + 28|0);
   $458 = ((($1)) + 16|0);
   $459 = ((($1)) + 20|0);
   $460 = ((($1)) + 24|0);
   $461 = HEAP32[$456>>2]|0;
   $462 = HEAP32[$457>>2]|0;
   $$phi$trans$insert229 = $462 & 255;
   $$phi$trans$insert230 = (125924 + ($$phi$trans$insert229<<2)|0);
   $$pre231 = HEAP32[$$phi$trans$insert230>>2]|0;
   $463 = $461 >>> 8;
   $464 = $463 & 255;
   $465 = $462 >>> 8;
   $466 = $465 & 255;
   $$phi$trans$insert233 = (125924 + ($466<<2)|0);
   $$pre234 = HEAP32[$$phi$trans$insert233>>2]|0;
   $467 = $461 >>> 16;
   $468 = $462 >>> 16;
   $$phi$trans$insert235 = $468 & 255;
   $$phi$trans$insert236 = (125924 + ($$phi$trans$insert235<<2)|0);
   $$pre237 = HEAP32[$$phi$trans$insert236>>2]|0;
   $469 = $461 >>> 24;
   $470 = $462 >>> 24;
   $$phi$trans$insert239 = (125924 + ($470<<2)|0);
   $$pre240 = HEAP32[$$phi$trans$insert239>>2]|0;
   $471 = $461 & 255;
   $472 = $467 & 255;
   $length29$076 = $length29$0$ph;$pixel31$175 = $pixel31$1$ph;
   while(1) {
    $473 = (($length29$076) + -1)|0;
    $474 = HEAP32[$pixel31$175>>2]|0;
    $475 = HEAP32[$455>>2]|0;
    $476 = $475 & $474;
    $477 = $476 >>> $471;
    $478 = (($$pre231) + ($477)|0);
    $479 = HEAP8[$478>>0]|0;
    $480 = $479&255;
    $481 = HEAP32[$458>>2]|0;
    $482 = $481 & $474;
    $483 = $482 >>> $464;
    $484 = (($$pre234) + ($483)|0);
    $485 = HEAP8[$484>>0]|0;
    $486 = $485&255;
    $487 = HEAP32[$459>>2]|0;
    $488 = $487 & $474;
    $489 = $488 >>> $472;
    $490 = (($$pre237) + ($489)|0);
    $491 = HEAP8[$490>>0]|0;
    $492 = $491&255;
    $493 = HEAP32[$460>>2]|0;
    $494 = $493 & $474;
    $495 = $494 >>> $469;
    $496 = (($$pre240) + ($495)|0);
    $497 = HEAP8[$496>>0]|0;
    $498 = $497&255;
    $499 = (($480) + ($r$0))|0;
    $500 = ($499>>>0)>(255);
    $$3 = $500 ? 255 : $499;
    $501 = (($486) + ($g$0))|0;
    $502 = ($501>>>0)>(255);
    $$9 = $502 ? 255 : $501;
    $503 = (($492) + ($b$0))|0;
    $504 = ($503>>>0)>(255);
    $$4 = $504 ? 255 : $503;
    $505 = $$3 >>> $$phi$trans$insert229;
    $506 = $505 << $471;
    $507 = $$9 >>> $466;
    $508 = $507 << $464;
    $509 = $508 | $506;
    $510 = $$4 >>> $$phi$trans$insert235;
    $511 = $510 << $472;
    $512 = $509 | $511;
    $513 = $498 >>> $470;
    $514 = $513 << $469;
    $515 = $512 | $514;
    HEAP32[$pixel31$175>>2] = $515;
    $516 = (($pixel31$175) + ($435<<2)|0);
    $517 = ($473|0)==(0);
    if ($517) {
     break;
    } else {
     $length29$076 = $473;$pixel31$175 = $516;
    }
   }
   return;
   break;
  }
  case 4:  {
   $518 = ((($dst)) + 16|0);
   $519 = HEAP32[$518>>2]|0;
   $520 = ((($1)) + 9|0);
   $521 = HEAP8[$520>>0]|0;
   $522 = $521&255;
   $523 = (($519|0) / ($522|0))&-1;
   $524 = ($y2|0)<($y1|0);
   $525 = ((($dst)) + 20|0);
   $526 = HEAP32[$525>>2]|0;
   if ($524) {
    $534 = Math_imul($523, $y2)|0;
    $535 = (($526) + ($534<<2)|0);
    $536 = (($535) + ($x2<<2)|0);
    $537 = ($draw_end|0)!=(0);
    $538 = (($536) + ($523<<2)|0);
    $$19 = $537 ? $536 : $538;
    $539 = (($y1) - ($y2))|0;
    $540 = $537&1;
    $541 = (($539) + ($540))|0;
    $length36$0$ph = $541;$pixel38$1$ph = $$19;
   } else {
    $527 = Math_imul($523, $y1)|0;
    $528 = (($526) + ($527<<2)|0);
    $529 = (($528) + ($x2<<2)|0);
    $530 = ($draw_end|0)!=(0);
    $531 = (($y2) - ($y1))|0;
    $532 = $530&1;
    $533 = (($531) + ($532))|0;
    $length36$0$ph = $533;$pixel38$1$ph = $529;
   }
   $542 = ($length36$0$ph|0)==(0);
   if ($542) {
    return;
   }
   $543 = ((($1)) + 12|0);
   $544 = ((($1)) + 32|0);
   $545 = ((($1)) + 28|0);
   $546 = ((($1)) + 16|0);
   $547 = ((($1)) + 20|0);
   $548 = ((($1)) + 24|0);
   $549 = HEAP32[$544>>2]|0;
   $550 = HEAP32[$545>>2]|0;
   $$phi$trans$insert215 = $550 & 255;
   $$phi$trans$insert216 = (125924 + ($$phi$trans$insert215<<2)|0);
   $$pre217 = HEAP32[$$phi$trans$insert216>>2]|0;
   $551 = $549 >>> 8;
   $552 = $551 & 255;
   $553 = $550 >>> 8;
   $554 = $553 & 255;
   $$phi$trans$insert219 = (125924 + ($554<<2)|0);
   $$pre220 = HEAP32[$$phi$trans$insert219>>2]|0;
   $555 = $549 >>> 16;
   $556 = $550 >>> 16;
   $$phi$trans$insert221 = $556 & 255;
   $$phi$trans$insert222 = (125924 + ($$phi$trans$insert221<<2)|0);
   $$pre223 = HEAP32[$$phi$trans$insert222>>2]|0;
   $557 = $549 >>> 24;
   $558 = $550 >>> 24;
   $$phi$trans$insert225 = (125924 + ($558<<2)|0);
   $$pre226 = HEAP32[$$phi$trans$insert225>>2]|0;
   $559 = $549 & 255;
   $560 = $555 & 255;
   $length36$079 = $length36$0$ph;$pixel38$178 = $pixel38$1$ph;
   while(1) {
    $561 = (($length36$079) + -1)|0;
    $562 = HEAP32[$pixel38$178>>2]|0;
    $563 = HEAP32[$543>>2]|0;
    $564 = $563 & $562;
    $565 = $564 >>> $559;
    $566 = (($$pre217) + ($565)|0);
    $567 = HEAP8[$566>>0]|0;
    $568 = $567&255;
    $569 = HEAP32[$546>>2]|0;
    $570 = $569 & $562;
    $571 = $570 >>> $552;
    $572 = (($$pre220) + ($571)|0);
    $573 = HEAP8[$572>>0]|0;
    $574 = $573&255;
    $575 = HEAP32[$547>>2]|0;
    $576 = $575 & $562;
    $577 = $576 >>> $560;
    $578 = (($$pre223) + ($577)|0);
    $579 = HEAP8[$578>>0]|0;
    $580 = $579&255;
    $581 = HEAP32[$548>>2]|0;
    $582 = $581 & $562;
    $583 = $582 >>> $557;
    $584 = (($$pre226) + ($583)|0);
    $585 = HEAP8[$584>>0]|0;
    $586 = $585&255;
    $587 = Math_imul($568, $r$0)|0;
    $588 = (($587>>>0) / 255)&-1;
    $589 = Math_imul($574, $g$0)|0;
    $590 = (($589>>>0) / 255)&-1;
    $591 = Math_imul($580, $b$0)|0;
    $592 = (($591>>>0) / 255)&-1;
    $593 = $588 >>> $$phi$trans$insert215;
    $594 = $593 << $559;
    $595 = $590 >>> $554;
    $596 = $595 << $552;
    $597 = $596 | $594;
    $598 = $592 >>> $$phi$trans$insert221;
    $599 = $598 << $560;
    $600 = $597 | $599;
    $601 = $586 >>> $558;
    $602 = $601 << $557;
    $603 = $600 | $602;
    HEAP32[$pixel38$178>>2] = $603;
    $604 = (($pixel38$178) + ($523<<2)|0);
    $605 = ($561|0)==(0);
    if ($605) {
     break;
    } else {
     $length36$079 = $561;$pixel38$178 = $604;
    }
   }
   return;
   break;
  }
  default: {
   $606 = ((($dst)) + 16|0);
   $607 = HEAP32[$606>>2]|0;
   $608 = ((($1)) + 9|0);
   $609 = HEAP8[$608>>0]|0;
   $610 = $609&255;
   $611 = (($607|0) / ($610|0))&-1;
   $612 = ($y2|0)<($y1|0);
   $613 = ((($dst)) + 20|0);
   $614 = HEAP32[$613>>2]|0;
   if ($612) {
    $622 = Math_imul($611, $y2)|0;
    $623 = (($614) + ($622<<2)|0);
    $624 = (($623) + ($x2<<2)|0);
    $625 = ($draw_end|0)!=(0);
    $626 = (($624) + ($611<<2)|0);
    $$20 = $625 ? $624 : $626;
    $627 = (($y1) - ($y2))|0;
    $628 = $625&1;
    $629 = (($627) + ($628))|0;
    $length43$0$ph = $629;$pixel45$1$ph = $$20;
   } else {
    $615 = Math_imul($611, $y1)|0;
    $616 = (($614) + ($615<<2)|0);
    $617 = (($616) + ($x2<<2)|0);
    $618 = ($draw_end|0)!=(0);
    $619 = (($y2) - ($y1))|0;
    $620 = $618&1;
    $621 = (($619) + ($620))|0;
    $length43$0$ph = $621;$pixel45$1$ph = $617;
   }
   $630 = ($length43$0$ph|0)==(0);
   if ($630) {
    return;
   }
   $631 = ((($1)) + 28|0);
   $632 = HEAP32[$631>>2]|0;
   $633 = $632 & 255;
   $634 = $r$0 >>> $633;
   $635 = ((($1)) + 32|0);
   $636 = HEAP32[$635>>2]|0;
   $637 = $636 & 255;
   $638 = $634 << $637;
   $639 = $632 >>> 8;
   $640 = $639 & 255;
   $641 = $g$0 >>> $640;
   $642 = $636 >>> 8;
   $643 = $642 & 255;
   $644 = $641 << $643;
   $645 = $644 | $638;
   $646 = $632 >>> 16;
   $647 = $646 & 255;
   $648 = $b$0 >>> $647;
   $649 = $636 >>> 16;
   $650 = $649 & 255;
   $651 = $648 << $650;
   $652 = $645 | $651;
   $653 = $632 >>> 24;
   $654 = $a$0 >>> $653;
   $655 = $636 >>> 24;
   $656 = $654 << $655;
   $657 = $652 | $656;
   $length43$070 = $length43$0$ph;$pixel45$169 = $pixel45$1$ph;
   while(1) {
    $658 = (($length43$070) + -1)|0;
    HEAP32[$pixel45$169>>2] = $657;
    $659 = (($pixel45$169) + ($611<<2)|0);
    $660 = ($658|0)==(0);
    if ($660) {
     break;
    } else {
     $length43$070 = $658;$pixel45$169 = $659;
    }
   }
   return;
  }
  }
 }
 $661 = (($x1) - ($x2))|0;
 $662 = ($661|0)<(0);
 $663 = (0 - ($661))|0;
 $664 = $662 ? $663 : $661;
 $665 = (($y1) - ($y2))|0;
 $666 = ($665|0)<(0);
 $667 = (0 - ($665))|0;
 $668 = $666 ? $667 : $665;
 $669 = ($664|0)==($668|0);
 if ($669) {
  switch ($blendMode|0) {
  case 1:  {
   $670 = ((($dst)) + 16|0);
   $671 = HEAP32[$670>>2]|0;
   $672 = ((($1)) + 9|0);
   $673 = HEAP8[$672>>0]|0;
   $674 = $673&255;
   $675 = (($671|0) / ($674|0))&-1;
   $676 = ($y2|0)<($y1|0);
   $677 = ((($dst)) + 20|0);
   $678 = HEAP32[$677>>2]|0;
   if ($676) {
    $684 = Math_imul($675, $y2)|0;
    $685 = (($678) + ($684<<2)|0);
    $686 = (($685) + ($x2<<2)|0);
    $687 = ($x2|0)>($x1|0);
    $pitch51$1$v = $687 ? -1 : 1;
    $pitch51$1 = (($675) + ($pitch51$1$v))|0;
    $688 = ($draw_end|0)==(0);
    $689 = (($686) + ($pitch51$1<<2)|0);
    $$21 = $688 ? $689 : $686;
    $length50$0 = $665;$pitch51$2 = $pitch51$1;$pixel52$1 = $$21;
   } else {
    $679 = Math_imul($675, $y1)|0;
    $680 = (($678) + ($679<<2)|0);
    $681 = (($680) + ($x1<<2)|0);
    $682 = ($x2|0)<($x1|0);
    $pitch51$0$v = $682 ? -1 : 1;
    $pitch51$0 = (($675) + ($pitch51$0$v))|0;
    $683 = (($y2) - ($y1))|0;
    $length50$0 = $683;$pitch51$2 = $pitch51$0;$pixel52$1 = $681;
   }
   $not$34 = ($draw_end|0)!=(0);
   $690 = $not$34&1;
   $length50$0$ = (($length50$0) + ($690))|0;
   $691 = ($length50$0$|0)==(0);
   if ($691) {
    return;
   }
   $692 = ((($1)) + 12|0);
   $693 = ((($1)) + 32|0);
   $694 = ((($1)) + 28|0);
   $695 = ((($1)) + 16|0);
   $696 = ((($1)) + 20|0);
   $697 = ((($1)) + 24|0);
   $698 = HEAP32[$693>>2]|0;
   $699 = HEAP32[$694>>2]|0;
   $$phi$trans$insert201 = $699 & 255;
   $$phi$trans$insert202 = (125924 + ($$phi$trans$insert201<<2)|0);
   $$pre203 = HEAP32[$$phi$trans$insert202>>2]|0;
   $700 = $698 >>> 8;
   $701 = $700 & 255;
   $702 = $699 >>> 8;
   $703 = $702 & 255;
   $$phi$trans$insert205 = (125924 + ($703<<2)|0);
   $$pre206 = HEAP32[$$phi$trans$insert205>>2]|0;
   $704 = $698 >>> 16;
   $705 = $699 >>> 16;
   $$phi$trans$insert207 = $705 & 255;
   $$phi$trans$insert208 = (125924 + ($$phi$trans$insert207<<2)|0);
   $$pre209 = HEAP32[$$phi$trans$insert208>>2]|0;
   $706 = $698 >>> 24;
   $707 = $699 >>> 24;
   $$phi$trans$insert211 = (125924 + ($707<<2)|0);
   $$pre212 = HEAP32[$$phi$trans$insert211>>2]|0;
   $708 = $698 & 255;
   $709 = $704 & 255;
   $length50$185 = $length50$0$;$pixel52$284 = $pixel52$1;
   while(1) {
    $710 = (($length50$185) + -1)|0;
    $711 = HEAP32[$pixel52$284>>2]|0;
    $712 = HEAP32[$692>>2]|0;
    $713 = $712 & $711;
    $714 = $713 >>> $708;
    $715 = (($$pre203) + ($714)|0);
    $716 = HEAP8[$715>>0]|0;
    $717 = $716&255;
    $718 = HEAP32[$695>>2]|0;
    $719 = $718 & $711;
    $720 = $719 >>> $701;
    $721 = (($$pre206) + ($720)|0);
    $722 = HEAP8[$721>>0]|0;
    $723 = $722&255;
    $724 = HEAP32[$696>>2]|0;
    $725 = $724 & $711;
    $726 = $725 >>> $709;
    $727 = (($$pre209) + ($726)|0);
    $728 = HEAP8[$727>>0]|0;
    $729 = $728&255;
    $730 = HEAP32[$697>>2]|0;
    $731 = $730 & $711;
    $732 = $731 >>> $706;
    $733 = (($$pre212) + ($732)|0);
    $734 = HEAP8[$733>>0]|0;
    $735 = $734&255;
    $736 = Math_imul($717, $16)|0;
    $737 = (($736>>>0) / 255)&-1;
    $738 = (($737) + ($r$0))|0;
    $739 = Math_imul($723, $16)|0;
    $740 = (($739>>>0) / 255)&-1;
    $741 = (($740) + ($g$0))|0;
    $742 = Math_imul($729, $16)|0;
    $743 = (($742>>>0) / 255)&-1;
    $744 = (($743) + ($b$0))|0;
    $745 = Math_imul($735, $16)|0;
    $746 = (($745>>>0) / 255)&-1;
    $747 = (($746) + ($a$0))|0;
    $748 = $738 >>> $$phi$trans$insert201;
    $749 = $748 << $708;
    $750 = $741 >>> $703;
    $751 = $750 << $701;
    $752 = $751 | $749;
    $753 = $744 >>> $$phi$trans$insert207;
    $754 = $753 << $709;
    $755 = $752 | $754;
    $756 = $747 >>> $707;
    $757 = $756 << $706;
    $758 = $755 | $757;
    HEAP32[$pixel52$284>>2] = $758;
    $759 = (($pixel52$284) + ($pitch51$2<<2)|0);
    $760 = ($710|0)==(0);
    if ($760) {
     break;
    } else {
     $length50$185 = $710;$pixel52$284 = $759;
    }
   }
   return;
   break;
  }
  case 2:  {
   $761 = ((($dst)) + 16|0);
   $762 = HEAP32[$761>>2]|0;
   $763 = ((($1)) + 9|0);
   $764 = HEAP8[$763>>0]|0;
   $765 = $764&255;
   $766 = (($762|0) / ($765|0))&-1;
   $767 = ($y2|0)<($y1|0);
   $768 = ((($dst)) + 20|0);
   $769 = HEAP32[$768>>2]|0;
   if ($767) {
    $775 = Math_imul($766, $y2)|0;
    $776 = (($769) + ($775<<2)|0);
    $777 = (($776) + ($x2<<2)|0);
    $778 = ($x2|0)>($x1|0);
    $pitch58$1$v = $778 ? -1 : 1;
    $pitch58$1 = (($766) + ($pitch58$1$v))|0;
    $779 = ($draw_end|0)==(0);
    $780 = (($777) + ($pitch58$1<<2)|0);
    $$22 = $779 ? $780 : $777;
    $length57$0 = $665;$pitch58$2 = $pitch58$1;$pixel59$1 = $$22;
   } else {
    $770 = Math_imul($766, $y1)|0;
    $771 = (($769) + ($770<<2)|0);
    $772 = (($771) + ($x1<<2)|0);
    $773 = ($x2|0)<($x1|0);
    $pitch58$0$v = $773 ? -1 : 1;
    $pitch58$0 = (($766) + ($pitch58$0$v))|0;
    $774 = (($y2) - ($y1))|0;
    $length57$0 = $774;$pitch58$2 = $pitch58$0;$pixel59$1 = $772;
   }
   $not$33 = ($draw_end|0)!=(0);
   $781 = $not$33&1;
   $length57$0$ = (($length57$0) + ($781))|0;
   $782 = ($length57$0$|0)==(0);
   if ($782) {
    return;
   }
   $783 = ((($1)) + 12|0);
   $784 = ((($1)) + 32|0);
   $785 = ((($1)) + 28|0);
   $786 = ((($1)) + 16|0);
   $787 = ((($1)) + 20|0);
   $788 = ((($1)) + 24|0);
   $789 = HEAP32[$784>>2]|0;
   $790 = HEAP32[$785>>2]|0;
   $$phi$trans$insert187 = $790 & 255;
   $$phi$trans$insert188 = (125924 + ($$phi$trans$insert187<<2)|0);
   $$pre189 = HEAP32[$$phi$trans$insert188>>2]|0;
   $791 = $789 >>> 8;
   $792 = $791 & 255;
   $793 = $790 >>> 8;
   $794 = $793 & 255;
   $$phi$trans$insert191 = (125924 + ($794<<2)|0);
   $$pre192 = HEAP32[$$phi$trans$insert191>>2]|0;
   $795 = $789 >>> 16;
   $796 = $790 >>> 16;
   $$phi$trans$insert193 = $796 & 255;
   $$phi$trans$insert194 = (125924 + ($$phi$trans$insert193<<2)|0);
   $$pre195 = HEAP32[$$phi$trans$insert194>>2]|0;
   $797 = $789 >>> 24;
   $798 = $790 >>> 24;
   $$phi$trans$insert197 = (125924 + ($798<<2)|0);
   $$pre198 = HEAP32[$$phi$trans$insert197>>2]|0;
   $799 = $789 & 255;
   $800 = $795 & 255;
   $length57$188 = $length57$0$;$pixel59$287 = $pixel59$1;
   while(1) {
    $801 = (($length57$188) + -1)|0;
    $802 = HEAP32[$pixel59$287>>2]|0;
    $803 = HEAP32[$783>>2]|0;
    $804 = $803 & $802;
    $805 = $804 >>> $799;
    $806 = (($$pre189) + ($805)|0);
    $807 = HEAP8[$806>>0]|0;
    $808 = $807&255;
    $809 = HEAP32[$786>>2]|0;
    $810 = $809 & $802;
    $811 = $810 >>> $792;
    $812 = (($$pre192) + ($811)|0);
    $813 = HEAP8[$812>>0]|0;
    $814 = $813&255;
    $815 = HEAP32[$787>>2]|0;
    $816 = $815 & $802;
    $817 = $816 >>> $800;
    $818 = (($$pre195) + ($817)|0);
    $819 = HEAP8[$818>>0]|0;
    $820 = $819&255;
    $821 = HEAP32[$788>>2]|0;
    $822 = $821 & $802;
    $823 = $822 >>> $797;
    $824 = (($$pre198) + ($823)|0);
    $825 = HEAP8[$824>>0]|0;
    $826 = $825&255;
    $827 = (($808) + ($r$0))|0;
    $828 = ($827>>>0)>(255);
    $$5 = $828 ? 255 : $827;
    $829 = (($814) + ($g$0))|0;
    $830 = ($829>>>0)>(255);
    $$10 = $830 ? 255 : $829;
    $831 = (($820) + ($b$0))|0;
    $832 = ($831>>>0)>(255);
    $$6 = $832 ? 255 : $831;
    $833 = $$5 >>> $$phi$trans$insert187;
    $834 = $833 << $799;
    $835 = $$10 >>> $794;
    $836 = $835 << $792;
    $837 = $836 | $834;
    $838 = $$6 >>> $$phi$trans$insert193;
    $839 = $838 << $800;
    $840 = $837 | $839;
    $841 = $826 >>> $798;
    $842 = $841 << $797;
    $843 = $840 | $842;
    HEAP32[$pixel59$287>>2] = $843;
    $844 = (($pixel59$287) + ($pitch58$2<<2)|0);
    $845 = ($801|0)==(0);
    if ($845) {
     break;
    } else {
     $length57$188 = $801;$pixel59$287 = $844;
    }
   }
   return;
   break;
  }
  case 4:  {
   $846 = ((($dst)) + 16|0);
   $847 = HEAP32[$846>>2]|0;
   $848 = ((($1)) + 9|0);
   $849 = HEAP8[$848>>0]|0;
   $850 = $849&255;
   $851 = (($847|0) / ($850|0))&-1;
   $852 = ($y2|0)<($y1|0);
   $853 = ((($dst)) + 20|0);
   $854 = HEAP32[$853>>2]|0;
   if ($852) {
    $860 = Math_imul($851, $y2)|0;
    $861 = (($854) + ($860<<2)|0);
    $862 = (($861) + ($x2<<2)|0);
    $863 = ($x2|0)>($x1|0);
    $pitch65$1$v = $863 ? -1 : 1;
    $pitch65$1 = (($851) + ($pitch65$1$v))|0;
    $864 = ($draw_end|0)==(0);
    $865 = (($862) + ($pitch65$1<<2)|0);
    $$23 = $864 ? $865 : $862;
    $length64$0 = $665;$pitch65$2 = $pitch65$1;$pixel66$1 = $$23;
   } else {
    $855 = Math_imul($851, $y1)|0;
    $856 = (($854) + ($855<<2)|0);
    $857 = (($856) + ($x1<<2)|0);
    $858 = ($x2|0)<($x1|0);
    $pitch65$0$v = $858 ? -1 : 1;
    $pitch65$0 = (($851) + ($pitch65$0$v))|0;
    $859 = (($y2) - ($y1))|0;
    $length64$0 = $859;$pitch65$2 = $pitch65$0;$pixel66$1 = $857;
   }
   $not$32 = ($draw_end|0)!=(0);
   $866 = $not$32&1;
   $length64$0$ = (($length64$0) + ($866))|0;
   $867 = ($length64$0$|0)==(0);
   if ($867) {
    return;
   }
   $868 = ((($1)) + 12|0);
   $869 = ((($1)) + 32|0);
   $870 = ((($1)) + 28|0);
   $871 = ((($1)) + 16|0);
   $872 = ((($1)) + 20|0);
   $873 = ((($1)) + 24|0);
   $874 = HEAP32[$869>>2]|0;
   $875 = HEAP32[$870>>2]|0;
   $$phi$trans$insert173 = $875 & 255;
   $$phi$trans$insert174 = (125924 + ($$phi$trans$insert173<<2)|0);
   $$pre175 = HEAP32[$$phi$trans$insert174>>2]|0;
   $876 = $874 >>> 8;
   $877 = $876 & 255;
   $878 = $875 >>> 8;
   $879 = $878 & 255;
   $$phi$trans$insert177 = (125924 + ($879<<2)|0);
   $$pre178 = HEAP32[$$phi$trans$insert177>>2]|0;
   $880 = $874 >>> 16;
   $881 = $875 >>> 16;
   $$phi$trans$insert179 = $881 & 255;
   $$phi$trans$insert180 = (125924 + ($$phi$trans$insert179<<2)|0);
   $$pre181 = HEAP32[$$phi$trans$insert180>>2]|0;
   $882 = $874 >>> 24;
   $883 = $875 >>> 24;
   $$phi$trans$insert183 = (125924 + ($883<<2)|0);
   $$pre184 = HEAP32[$$phi$trans$insert183>>2]|0;
   $884 = $874 & 255;
   $885 = $880 & 255;
   $length64$191 = $length64$0$;$pixel66$290 = $pixel66$1;
   while(1) {
    $886 = (($length64$191) + -1)|0;
    $887 = HEAP32[$pixel66$290>>2]|0;
    $888 = HEAP32[$868>>2]|0;
    $889 = $888 & $887;
    $890 = $889 >>> $884;
    $891 = (($$pre175) + ($890)|0);
    $892 = HEAP8[$891>>0]|0;
    $893 = $892&255;
    $894 = HEAP32[$871>>2]|0;
    $895 = $894 & $887;
    $896 = $895 >>> $877;
    $897 = (($$pre178) + ($896)|0);
    $898 = HEAP8[$897>>0]|0;
    $899 = $898&255;
    $900 = HEAP32[$872>>2]|0;
    $901 = $900 & $887;
    $902 = $901 >>> $885;
    $903 = (($$pre181) + ($902)|0);
    $904 = HEAP8[$903>>0]|0;
    $905 = $904&255;
    $906 = HEAP32[$873>>2]|0;
    $907 = $906 & $887;
    $908 = $907 >>> $882;
    $909 = (($$pre184) + ($908)|0);
    $910 = HEAP8[$909>>0]|0;
    $911 = $910&255;
    $912 = Math_imul($893, $r$0)|0;
    $913 = (($912>>>0) / 255)&-1;
    $914 = Math_imul($899, $g$0)|0;
    $915 = (($914>>>0) / 255)&-1;
    $916 = Math_imul($905, $b$0)|0;
    $917 = (($916>>>0) / 255)&-1;
    $918 = $913 >>> $$phi$trans$insert173;
    $919 = $918 << $884;
    $920 = $915 >>> $879;
    $921 = $920 << $877;
    $922 = $921 | $919;
    $923 = $917 >>> $$phi$trans$insert179;
    $924 = $923 << $885;
    $925 = $922 | $924;
    $926 = $911 >>> $883;
    $927 = $926 << $882;
    $928 = $925 | $927;
    HEAP32[$pixel66$290>>2] = $928;
    $929 = (($pixel66$290) + ($pitch65$2<<2)|0);
    $930 = ($886|0)==(0);
    if ($930) {
     break;
    } else {
     $length64$191 = $886;$pixel66$290 = $929;
    }
   }
   return;
   break;
  }
  default: {
   $931 = ((($dst)) + 16|0);
   $932 = HEAP32[$931>>2]|0;
   $933 = ((($1)) + 9|0);
   $934 = HEAP8[$933>>0]|0;
   $935 = $934&255;
   $936 = (($932|0) / ($935|0))&-1;
   $937 = ($y2|0)<($y1|0);
   $938 = ((($dst)) + 20|0);
   $939 = HEAP32[$938>>2]|0;
   if ($937) {
    $945 = Math_imul($936, $y2)|0;
    $946 = (($939) + ($945<<2)|0);
    $947 = (($946) + ($x2<<2)|0);
    $948 = ($x2|0)>($x1|0);
    $pitch72$1$v = $948 ? -1 : 1;
    $pitch72$1 = (($936) + ($pitch72$1$v))|0;
    $949 = ($draw_end|0)==(0);
    $950 = (($947) + ($pitch72$1<<2)|0);
    $$24 = $949 ? $950 : $947;
    $length71$0 = $665;$pitch72$2 = $pitch72$1;$pixel73$1 = $$24;
   } else {
    $940 = Math_imul($936, $y1)|0;
    $941 = (($939) + ($940<<2)|0);
    $942 = (($941) + ($x1<<2)|0);
    $943 = ($x2|0)<($x1|0);
    $pitch72$0$v = $943 ? -1 : 1;
    $pitch72$0 = (($936) + ($pitch72$0$v))|0;
    $944 = (($y2) - ($y1))|0;
    $length71$0 = $944;$pitch72$2 = $pitch72$0;$pixel73$1 = $942;
   }
   $not$35 = ($draw_end|0)!=(0);
   $951 = $not$35&1;
   $length71$0$ = (($length71$0) + ($951))|0;
   $952 = ($length71$0$|0)==(0);
   if ($952) {
    return;
   }
   $953 = ((($1)) + 28|0);
   $954 = HEAP32[$953>>2]|0;
   $955 = $954 & 255;
   $956 = $r$0 >>> $955;
   $957 = ((($1)) + 32|0);
   $958 = HEAP32[$957>>2]|0;
   $959 = $958 & 255;
   $960 = $956 << $959;
   $961 = $954 >>> 8;
   $962 = $961 & 255;
   $963 = $g$0 >>> $962;
   $964 = $958 >>> 8;
   $965 = $964 & 255;
   $966 = $963 << $965;
   $967 = $966 | $960;
   $968 = $954 >>> 16;
   $969 = $968 & 255;
   $970 = $b$0 >>> $969;
   $971 = $958 >>> 16;
   $972 = $971 & 255;
   $973 = $970 << $972;
   $974 = $967 | $973;
   $975 = $954 >>> 24;
   $976 = $a$0 >>> $975;
   $977 = $958 >>> 24;
   $978 = $976 << $977;
   $979 = $974 | $978;
   $length71$181 = $length71$0$;$pixel73$282 = $pixel73$1;
   while(1) {
    $980 = (($length71$181) + -1)|0;
    HEAP32[$pixel73$282>>2] = $979;
    $981 = (($pixel73$282) + ($pitch72$2<<2)|0);
    $982 = ($980|0)==(0);
    if ($982) {
     break;
    } else {
     $length71$181 = $980;$pixel73$282 = $981;
    }
   }
   return;
  }
  }
 } else {
  switch ($blendMode|0) {
  case 1:  {
   $983 = (($x2) - ($x1))|0;
   $984 = ($983|0)<(0);
   $985 = (0 - ($983))|0;
   $986 = $984 ? $985 : $983;
   $987 = (($y2) - ($y1))|0;
   $988 = ($987|0)<(0);
   $989 = (0 - ($987))|0;
   $990 = $988 ? $989 : $987;
   $991 = ($986|0)<($990|0);
   if ($991) {
    $995 = $986 << 1;
    $996 = (($995) - ($990))|0;
    $997 = (($986) - ($990))|0;
    $d$0 = $996;$dinc1$0 = $995;$dinc2$0$in = $997;$numpixels$0$in = $990;$xinc1$0 = 0;$yinc1$0 = 1;
   } else {
    $992 = $990 << 1;
    $993 = (($992) - ($986))|0;
    $994 = (($990) - ($986))|0;
    $d$0 = $993;$dinc1$0 = $992;$dinc2$0$in = $994;$numpixels$0$in = $986;$xinc1$0 = 1;$yinc1$0 = 0;
   }
   $dinc2$0 = $dinc2$0$in << 1;
   $998 = ($x2|0)<($x1|0);
   $999 = (0 - ($xinc1$0))|0;
   $$xinc1$0 = $998 ? $999 : $xinc1$0;
   $$25 = $998 ? -1 : 1;
   $1000 = ($y2|0)<($y1|0);
   $1001 = (0 - ($yinc1$0))|0;
   $yinc1$1 = $1000 ? $1001 : $yinc1$0;
   $yinc2$1 = $1000 ? -1 : 1;
   $not$30 = ($draw_end|0)!=(0);
   $numpixels$0 = $not$30&1;
   $numpixels$0$in$numpixels$0 = (($numpixels$0$in) + ($numpixels$0))|0;
   $1002 = ($numpixels$0$in$numpixels$0|0)>(0);
   if (!($1002)) {
    return;
   }
   $1003 = ((($dst)) + 20|0);
   $1004 = ((($dst)) + 16|0);
   $1005 = ((($1)) + 12|0);
   $1006 = ((($1)) + 32|0);
   $1007 = ((($1)) + 28|0);
   $1008 = ((($1)) + 16|0);
   $1009 = ((($1)) + 20|0);
   $1010 = ((($1)) + 24|0);
   $$pre156 = HEAP32[$1003>>2]|0;
   $1011 = HEAP32[$1006>>2]|0;
   $1012 = HEAP32[$1007>>2]|0;
   $$phi$trans$insert159 = $1012 & 255;
   $$phi$trans$insert160 = (125924 + ($$phi$trans$insert159<<2)|0);
   $$pre161 = HEAP32[$$phi$trans$insert160>>2]|0;
   $1013 = $1011 >>> 8;
   $1014 = $1013 & 255;
   $1015 = $1012 >>> 8;
   $1016 = $1015 & 255;
   $$phi$trans$insert163 = (125924 + ($1016<<2)|0);
   $$pre164 = HEAP32[$$phi$trans$insert163>>2]|0;
   $1017 = $1011 >>> 16;
   $1018 = $1012 >>> 16;
   $$phi$trans$insert165 = $1018 & 255;
   $$phi$trans$insert166 = (125924 + ($$phi$trans$insert165<<2)|0);
   $$pre167 = HEAP32[$$phi$trans$insert166>>2]|0;
   $1019 = $1011 >>> 24;
   $1020 = $1012 >>> 24;
   $$phi$trans$insert169 = (125924 + ($1020<<2)|0);
   $$pre170 = HEAP32[$$phi$trans$insert169>>2]|0;
   $1021 = $1011 & 255;
   $1022 = $1017 & 255;
   $d$199 = $d$0;$i$098 = 0;$x$0100 = $x1;$y$0101 = $y1;
   while(1) {
    $1023 = HEAP32[$1004>>2]|0;
    $1024 = Math_imul($1023, $y$0101)|0;
    $1025 = (($$pre156) + ($1024)|0);
    $1026 = $x$0100 << 2;
    $1027 = (($1025) + ($1026)|0);
    $1028 = HEAP32[$1027>>2]|0;
    $1029 = HEAP32[$1005>>2]|0;
    $1030 = $1029 & $1028;
    $1031 = $1030 >>> $1021;
    $1032 = (($$pre161) + ($1031)|0);
    $1033 = HEAP8[$1032>>0]|0;
    $1034 = $1033&255;
    $1035 = HEAP32[$1008>>2]|0;
    $1036 = $1035 & $1028;
    $1037 = $1036 >>> $1014;
    $1038 = (($$pre164) + ($1037)|0);
    $1039 = HEAP8[$1038>>0]|0;
    $1040 = $1039&255;
    $1041 = HEAP32[$1009>>2]|0;
    $1042 = $1041 & $1028;
    $1043 = $1042 >>> $1022;
    $1044 = (($$pre167) + ($1043)|0);
    $1045 = HEAP8[$1044>>0]|0;
    $1046 = $1045&255;
    $1047 = HEAP32[$1010>>2]|0;
    $1048 = $1047 & $1028;
    $1049 = $1048 >>> $1019;
    $1050 = (($$pre170) + ($1049)|0);
    $1051 = HEAP8[$1050>>0]|0;
    $1052 = $1051&255;
    $1053 = Math_imul($1034, $16)|0;
    $1054 = (($1053>>>0) / 255)&-1;
    $1055 = (($1054) + ($r$0))|0;
    $1056 = Math_imul($1040, $16)|0;
    $1057 = (($1056>>>0) / 255)&-1;
    $1058 = (($1057) + ($g$0))|0;
    $1059 = Math_imul($1046, $16)|0;
    $1060 = (($1059>>>0) / 255)&-1;
    $1061 = (($1060) + ($b$0))|0;
    $1062 = Math_imul($1052, $16)|0;
    $1063 = (($1062>>>0) / 255)&-1;
    $1064 = (($1063) + ($a$0))|0;
    $1065 = $1055 >>> $$phi$trans$insert159;
    $1066 = $1065 << $1021;
    $1067 = $1058 >>> $1016;
    $1068 = $1067 << $1014;
    $1069 = $1068 | $1066;
    $1070 = $1061 >>> $$phi$trans$insert165;
    $1071 = $1070 << $1022;
    $1072 = $1069 | $1071;
    $1073 = $1064 >>> $1020;
    $1074 = $1073 << $1019;
    $1075 = $1072 | $1074;
    HEAP32[$1027>>2] = $1075;
    $1076 = ($d$199|0)<(0);
    $dinc1$0$dinc2$0 = $1076 ? $dinc1$0 : $dinc2$0;
    $$xinc1$0$$25 = $1076 ? $$xinc1$0 : $$25;
    $yinc1$1$yinc2$1 = $1076 ? $yinc1$1 : $yinc2$1;
    $y$1 = (($y$0101) + ($yinc1$1$yinc2$1))|0;
    $x$1 = (($$xinc1$0$$25) + ($x$0100))|0;
    $d$2 = (($dinc1$0$dinc2$0) + ($d$199))|0;
    $1077 = (($i$098) + 1)|0;
    $1078 = ($1077|0)<($numpixels$0$in$numpixels$0|0);
    if ($1078) {
     $d$199 = $d$2;$i$098 = $1077;$x$0100 = $x$1;$y$0101 = $y$1;
    } else {
     break;
    }
   }
   return;
   break;
  }
  case 2:  {
   $1079 = (($x2) - ($x1))|0;
   $1080 = ($1079|0)<(0);
   $1081 = (0 - ($1079))|0;
   $1082 = $1080 ? $1081 : $1079;
   $1083 = (($y2) - ($y1))|0;
   $1084 = ($1083|0)<(0);
   $1085 = (0 - ($1083))|0;
   $1086 = $1084 ? $1085 : $1083;
   $1087 = ($1082|0)<($1086|0);
   if ($1087) {
    $1091 = $1082 << 1;
    $1092 = (($1091) - ($1086))|0;
    $1093 = (($1082) - ($1086))|0;
    $d87$0 = $1092;$dinc188$0 = $1091;$dinc289$0$in = $1093;$numpixels86$0$in = $1086;$xinc191$0 = 0;$yinc194$0 = 1;
   } else {
    $1088 = $1086 << 1;
    $1089 = (($1088) - ($1082))|0;
    $1090 = (($1086) - ($1082))|0;
    $d87$0 = $1089;$dinc188$0 = $1088;$dinc289$0$in = $1090;$numpixels86$0$in = $1082;$xinc191$0 = 1;$yinc194$0 = 0;
   }
   $dinc289$0 = $dinc289$0$in << 1;
   $1094 = ($x2|0)<($x1|0);
   $1095 = (0 - ($xinc191$0))|0;
   $$xinc191$0 = $1094 ? $1095 : $xinc191$0;
   $$26 = $1094 ? -1 : 1;
   $1096 = ($y2|0)<($y1|0);
   $1097 = (0 - ($yinc194$0))|0;
   $yinc194$1 = $1096 ? $1097 : $yinc194$0;
   $yinc295$1 = $1096 ? -1 : 1;
   $not$29 = ($draw_end|0)!=(0);
   $numpixels86$0 = $not$29&1;
   $numpixels86$0$in$numpixels86$0 = (($numpixels86$0$in) + ($numpixels86$0))|0;
   $1098 = ($numpixels86$0$in$numpixels86$0|0)>(0);
   if (!($1098)) {
    return;
   }
   $1099 = ((($dst)) + 20|0);
   $1100 = ((($dst)) + 16|0);
   $1101 = ((($1)) + 12|0);
   $1102 = ((($1)) + 32|0);
   $1103 = ((($1)) + 28|0);
   $1104 = ((($1)) + 16|0);
   $1105 = ((($1)) + 20|0);
   $1106 = ((($1)) + 24|0);
   $$pre141 = HEAP32[$1099>>2]|0;
   $1107 = HEAP32[$1102>>2]|0;
   $1108 = HEAP32[$1103>>2]|0;
   $$phi$trans$insert144 = $1108 & 255;
   $$phi$trans$insert145 = (125924 + ($$phi$trans$insert144<<2)|0);
   $$pre146 = HEAP32[$$phi$trans$insert145>>2]|0;
   $1109 = $1107 >>> 8;
   $1110 = $1109 & 255;
   $1111 = $1108 >>> 8;
   $1112 = $1111 & 255;
   $$phi$trans$insert148 = (125924 + ($1112<<2)|0);
   $$pre149 = HEAP32[$$phi$trans$insert148>>2]|0;
   $1113 = $1107 >>> 16;
   $1114 = $1108 >>> 16;
   $$phi$trans$insert150 = $1114 & 255;
   $$phi$trans$insert151 = (125924 + ($$phi$trans$insert150<<2)|0);
   $$pre152 = HEAP32[$$phi$trans$insert151>>2]|0;
   $1115 = $1107 >>> 24;
   $1116 = $1108 >>> 24;
   $$phi$trans$insert154 = (125924 + ($1116<<2)|0);
   $$pre155 = HEAP32[$$phi$trans$insert154>>2]|0;
   $1117 = $1107 & 255;
   $1118 = $1113 & 255;
   $d87$1104 = $d87$0;$i83$0103 = 0;$x90$0105 = $x1;$y93$0106 = $y1;
   while(1) {
    $1119 = HEAP32[$1100>>2]|0;
    $1120 = Math_imul($1119, $y93$0106)|0;
    $1121 = (($$pre141) + ($1120)|0);
    $1122 = $x90$0105 << 2;
    $1123 = (($1121) + ($1122)|0);
    $1124 = HEAP32[$1123>>2]|0;
    $1125 = HEAP32[$1101>>2]|0;
    $1126 = $1125 & $1124;
    $1127 = $1126 >>> $1117;
    $1128 = (($$pre146) + ($1127)|0);
    $1129 = HEAP8[$1128>>0]|0;
    $1130 = $1129&255;
    $1131 = HEAP32[$1104>>2]|0;
    $1132 = $1131 & $1124;
    $1133 = $1132 >>> $1110;
    $1134 = (($$pre149) + ($1133)|0);
    $1135 = HEAP8[$1134>>0]|0;
    $1136 = $1135&255;
    $1137 = HEAP32[$1105>>2]|0;
    $1138 = $1137 & $1124;
    $1139 = $1138 >>> $1118;
    $1140 = (($$pre152) + ($1139)|0);
    $1141 = HEAP8[$1140>>0]|0;
    $1142 = $1141&255;
    $1143 = HEAP32[$1106>>2]|0;
    $1144 = $1143 & $1124;
    $1145 = $1144 >>> $1115;
    $1146 = (($$pre155) + ($1145)|0);
    $1147 = HEAP8[$1146>>0]|0;
    $1148 = $1147&255;
    $1149 = (($1130) + ($r$0))|0;
    $1150 = ($1149>>>0)>(255);
    $$11 = $1150 ? 255 : $1149;
    $1151 = (($1136) + ($g$0))|0;
    $1152 = ($1151>>>0)>(255);
    $$7 = $1152 ? 255 : $1151;
    $1153 = (($1142) + ($b$0))|0;
    $1154 = ($1153>>>0)>(255);
    $$12 = $1154 ? 255 : $1153;
    $1155 = $$11 >>> $$phi$trans$insert144;
    $1156 = $1155 << $1117;
    $1157 = $$7 >>> $1112;
    $1158 = $1157 << $1110;
    $1159 = $1158 | $1156;
    $1160 = $$12 >>> $$phi$trans$insert150;
    $1161 = $1160 << $1118;
    $1162 = $1159 | $1161;
    $1163 = $1148 >>> $1116;
    $1164 = $1163 << $1115;
    $1165 = $1162 | $1164;
    HEAP32[$1123>>2] = $1165;
    $1166 = ($d87$1104|0)<(0);
    $dinc188$0$dinc289$0 = $1166 ? $dinc188$0 : $dinc289$0;
    $$xinc191$0$$26 = $1166 ? $$xinc191$0 : $$26;
    $yinc194$1$yinc295$1 = $1166 ? $yinc194$1 : $yinc295$1;
    $y93$1 = (($y93$0106) + ($yinc194$1$yinc295$1))|0;
    $x90$1 = (($$xinc191$0$$26) + ($x90$0105))|0;
    $d87$2 = (($dinc188$0$dinc289$0) + ($d87$1104))|0;
    $1167 = (($i83$0103) + 1)|0;
    $1168 = ($1167|0)<($numpixels86$0$in$numpixels86$0|0);
    if ($1168) {
     $d87$1104 = $d87$2;$i83$0103 = $1167;$x90$0105 = $x90$1;$y93$0106 = $y93$1;
    } else {
     break;
    }
   }
   return;
   break;
  }
  case 4:  {
   $1169 = (($x2) - ($x1))|0;
   $1170 = ($1169|0)<(0);
   $1171 = (0 - ($1169))|0;
   $1172 = $1170 ? $1171 : $1169;
   $1173 = (($y2) - ($y1))|0;
   $1174 = ($1173|0)<(0);
   $1175 = (0 - ($1173))|0;
   $1176 = $1174 ? $1175 : $1173;
   $1177 = ($1172|0)<($1176|0);
   if ($1177) {
    $1181 = $1172 << 1;
    $1182 = (($1181) - ($1176))|0;
    $1183 = (($1172) - ($1176))|0;
    $d105$0 = $1182;$dinc1106$0 = $1181;$dinc2107$0$in = $1183;$numpixels104$0$in = $1176;$xinc1109$0 = 0;$yinc1112$0 = 1;
   } else {
    $1178 = $1176 << 1;
    $1179 = (($1178) - ($1172))|0;
    $1180 = (($1176) - ($1172))|0;
    $d105$0 = $1179;$dinc1106$0 = $1178;$dinc2107$0$in = $1180;$numpixels104$0$in = $1172;$xinc1109$0 = 1;$yinc1112$0 = 0;
   }
   $dinc2107$0 = $dinc2107$0$in << 1;
   $1184 = ($x2|0)<($x1|0);
   $1185 = (0 - ($xinc1109$0))|0;
   $$xinc1109$0 = $1184 ? $1185 : $xinc1109$0;
   $$27 = $1184 ? -1 : 1;
   $1186 = ($y2|0)<($y1|0);
   $1187 = (0 - ($yinc1112$0))|0;
   $yinc1112$1 = $1186 ? $1187 : $yinc1112$0;
   $yinc2113$1 = $1186 ? -1 : 1;
   $not$ = ($draw_end|0)!=(0);
   $numpixels104$0 = $not$&1;
   $numpixels104$0$in$numpixels104$0 = (($numpixels104$0$in) + ($numpixels104$0))|0;
   $1188 = ($numpixels104$0$in$numpixels104$0|0)>(0);
   if (!($1188)) {
    return;
   }
   $1189 = ((($dst)) + 20|0);
   $1190 = ((($dst)) + 16|0);
   $1191 = ((($1)) + 12|0);
   $1192 = ((($1)) + 32|0);
   $1193 = ((($1)) + 28|0);
   $1194 = ((($1)) + 16|0);
   $1195 = ((($1)) + 20|0);
   $1196 = ((($1)) + 24|0);
   $$pre = HEAP32[$1189>>2]|0;
   $1197 = HEAP32[$1192>>2]|0;
   $1198 = HEAP32[$1193>>2]|0;
   $$phi$trans$insert = $1198 & 255;
   $$phi$trans$insert130 = (125924 + ($$phi$trans$insert<<2)|0);
   $$pre131 = HEAP32[$$phi$trans$insert130>>2]|0;
   $1199 = $1197 >>> 8;
   $1200 = $1199 & 255;
   $1201 = $1198 >>> 8;
   $1202 = $1201 & 255;
   $$phi$trans$insert133 = (125924 + ($1202<<2)|0);
   $$pre134 = HEAP32[$$phi$trans$insert133>>2]|0;
   $1203 = $1197 >>> 16;
   $1204 = $1198 >>> 16;
   $$phi$trans$insert135 = $1204 & 255;
   $$phi$trans$insert136 = (125924 + ($$phi$trans$insert135<<2)|0);
   $$pre137 = HEAP32[$$phi$trans$insert136>>2]|0;
   $1205 = $1197 >>> 24;
   $1206 = $1198 >>> 24;
   $$phi$trans$insert139 = (125924 + ($1206<<2)|0);
   $$pre140 = HEAP32[$$phi$trans$insert139>>2]|0;
   $1207 = $1197 & 255;
   $1208 = $1203 & 255;
   $d105$1109 = $d105$0;$i101$0108 = 0;$x108$0110 = $x1;$y111$0111 = $y1;
   while(1) {
    $1209 = HEAP32[$1190>>2]|0;
    $1210 = Math_imul($1209, $y111$0111)|0;
    $1211 = (($$pre) + ($1210)|0);
    $1212 = $x108$0110 << 2;
    $1213 = (($1211) + ($1212)|0);
    $1214 = HEAP32[$1213>>2]|0;
    $1215 = HEAP32[$1191>>2]|0;
    $1216 = $1215 & $1214;
    $1217 = $1216 >>> $1207;
    $1218 = (($$pre131) + ($1217)|0);
    $1219 = HEAP8[$1218>>0]|0;
    $1220 = $1219&255;
    $1221 = HEAP32[$1194>>2]|0;
    $1222 = $1221 & $1214;
    $1223 = $1222 >>> $1200;
    $1224 = (($$pre134) + ($1223)|0);
    $1225 = HEAP8[$1224>>0]|0;
    $1226 = $1225&255;
    $1227 = HEAP32[$1195>>2]|0;
    $1228 = $1227 & $1214;
    $1229 = $1228 >>> $1208;
    $1230 = (($$pre137) + ($1229)|0);
    $1231 = HEAP8[$1230>>0]|0;
    $1232 = $1231&255;
    $1233 = HEAP32[$1196>>2]|0;
    $1234 = $1233 & $1214;
    $1235 = $1234 >>> $1205;
    $1236 = (($$pre140) + ($1235)|0);
    $1237 = HEAP8[$1236>>0]|0;
    $1238 = $1237&255;
    $1239 = Math_imul($1220, $r$0)|0;
    $1240 = (($1239>>>0) / 255)&-1;
    $1241 = Math_imul($1226, $g$0)|0;
    $1242 = (($1241>>>0) / 255)&-1;
    $1243 = Math_imul($1232, $b$0)|0;
    $1244 = (($1243>>>0) / 255)&-1;
    $1245 = $1240 >>> $$phi$trans$insert;
    $1246 = $1245 << $1207;
    $1247 = $1242 >>> $1202;
    $1248 = $1247 << $1200;
    $1249 = $1248 | $1246;
    $1250 = $1244 >>> $$phi$trans$insert135;
    $1251 = $1250 << $1208;
    $1252 = $1249 | $1251;
    $1253 = $1238 >>> $1206;
    $1254 = $1253 << $1205;
    $1255 = $1252 | $1254;
    HEAP32[$1213>>2] = $1255;
    $1256 = ($d105$1109|0)<(0);
    $dinc1106$0$dinc2107$0 = $1256 ? $dinc1106$0 : $dinc2107$0;
    $$xinc1109$0$$27 = $1256 ? $$xinc1109$0 : $$27;
    $yinc1112$1$yinc2113$1 = $1256 ? $yinc1112$1 : $yinc2113$1;
    $y111$1 = (($y111$0111) + ($yinc1112$1$yinc2113$1))|0;
    $x108$1 = (($$xinc1109$0$$27) + ($x108$0110))|0;
    $d105$2 = (($dinc1106$0$dinc2107$0) + ($d105$1109))|0;
    $1257 = (($i101$0108) + 1)|0;
    $1258 = ($1257|0)<($numpixels104$0$in$numpixels104$0|0);
    if ($1258) {
     $d105$1109 = $d105$2;$i101$0108 = $1257;$x108$0110 = $x108$1;$y111$0111 = $y111$1;
    } else {
     break;
    }
   }
   return;
   break;
  }
  default: {
   $1259 = (($x2) - ($x1))|0;
   $1260 = ($1259|0)<(0);
   $1261 = (0 - ($1259))|0;
   $1262 = $1260 ? $1261 : $1259;
   $1263 = (($y2) - ($y1))|0;
   $1264 = ($1263|0)<(0);
   $1265 = (0 - ($1263))|0;
   $1266 = $1264 ? $1265 : $1263;
   $1267 = ($1262|0)<($1266|0);
   if ($1267) {
    $1271 = $1262 << 1;
    $1272 = (($1271) - ($1266))|0;
    $1273 = (($1262) - ($1266))|0;
    $d123$0 = $1272;$dinc1124$0 = $1271;$dinc2125$0$in = $1273;$numpixels122$0$in = $1266;$xinc1127$0 = 0;$yinc1130$0 = 1;
   } else {
    $1268 = $1266 << 1;
    $1269 = (($1268) - ($1262))|0;
    $1270 = (($1266) - ($1262))|0;
    $d123$0 = $1269;$dinc1124$0 = $1268;$dinc2125$0$in = $1270;$numpixels122$0$in = $1262;$xinc1127$0 = 1;$yinc1130$0 = 0;
   }
   $dinc2125$0 = $dinc2125$0$in << 1;
   $1274 = ($x2|0)<($x1|0);
   $1275 = (0 - ($xinc1127$0))|0;
   $$xinc1127$0 = $1274 ? $1275 : $xinc1127$0;
   $$28 = $1274 ? -1 : 1;
   $1276 = ($y2|0)<($y1|0);
   $1277 = (0 - ($yinc1130$0))|0;
   $yinc1130$1 = $1276 ? $1277 : $yinc1130$0;
   $yinc2131$1 = $1276 ? -1 : 1;
   $not$31 = ($draw_end|0)!=(0);
   $numpixels122$0 = $not$31&1;
   $numpixels122$0$in$numpixels122$0 = (($numpixels122$0$in) + ($numpixels122$0))|0;
   $1278 = ($numpixels122$0$in$numpixels122$0|0)>(0);
   if (!($1278)) {
    return;
   }
   $1279 = ((($dst)) + 20|0);
   $1280 = HEAP32[$1279>>2]|0;
   $1281 = ((($dst)) + 16|0);
   $1282 = ((($1)) + 28|0);
   $1283 = HEAP32[$1282>>2]|0;
   $1284 = $1283 & 255;
   $1285 = $r$0 >>> $1284;
   $1286 = ((($1)) + 32|0);
   $1287 = HEAP32[$1286>>2]|0;
   $1288 = $1287 & 255;
   $1289 = $1285 << $1288;
   $1290 = $1283 >>> 8;
   $1291 = $1290 & 255;
   $1292 = $g$0 >>> $1291;
   $1293 = $1287 >>> 8;
   $1294 = $1293 & 255;
   $1295 = $1292 << $1294;
   $1296 = $1295 | $1289;
   $1297 = $1283 >>> 16;
   $1298 = $1297 & 255;
   $1299 = $b$0 >>> $1298;
   $1300 = $1287 >>> 16;
   $1301 = $1300 & 255;
   $1302 = $1299 << $1301;
   $1303 = $1296 | $1302;
   $1304 = $1283 >>> 24;
   $1305 = $a$0 >>> $1304;
   $1306 = $1287 >>> 24;
   $1307 = $1305 << $1306;
   $1308 = $1303 | $1307;
   $d123$194 = $d123$0;$i119$093 = 0;$x126$095 = $x1;$y129$096 = $y1;
   while(1) {
    $1309 = HEAP32[$1281>>2]|0;
    $1310 = Math_imul($1309, $y129$096)|0;
    $1311 = (($1280) + ($1310)|0);
    $1312 = $x126$095 << 2;
    $1313 = (($1311) + ($1312)|0);
    HEAP32[$1313>>2] = $1308;
    $1314 = ($d123$194|0)<(0);
    $dinc1124$0$dinc2125$0 = $1314 ? $dinc1124$0 : $dinc2125$0;
    $$xinc1127$0$$28 = $1314 ? $$xinc1127$0 : $$28;
    $yinc1130$1$yinc2131$1 = $1314 ? $yinc1130$1 : $yinc2131$1;
    $y129$1 = (($y129$096) + ($yinc1130$1$yinc2131$1))|0;
    $x126$1 = (($$xinc1127$0$$28) + ($x126$095))|0;
    $d123$2 = (($dinc1124$0$dinc2125$0) + ($d123$194))|0;
    $1315 = (($i119$093) + 1)|0;
    $1316 = ($1315|0)<($numpixels122$0$in$numpixels122$0|0);
    if ($1316) {
     $d123$194 = $d123$2;$i119$093 = $1315;$x126$095 = $x126$1;$y129$096 = $y129$1;
    } else {
     break;
    }
   }
   return;
  }
  }
 }
}
function _SDL_BlendLine_RGB4($dst,$x1,$y1,$x2,$y2,$blendMode,$_r,$_g,$_b,$_a,$draw_end) {
 $dst = $dst|0;
 $x1 = $x1|0;
 $y1 = $y1|0;
 $x2 = $x2|0;
 $y2 = $y2|0;
 $blendMode = $blendMode|0;
 $_r = $_r|0;
 $_g = $_g|0;
 $_b = $_b|0;
 $_a = $_a|0;
 $draw_end = $draw_end|0;
 var $$ = 0, $$10 = 0, $$11 = 0, $$12 = 0, $$13 = 0, $$14 = 0, $$15 = 0, $$16 = 0, $$17 = 0, $$18 = 0, $$19 = 0, $$2 = 0, $$20 = 0, $$21 = 0, $$22 = 0, $$23 = 0, $$24 = 0, $$25 = 0, $$26 = 0, $$27 = 0;
 var $$28 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$6 = 0, $$7 = 0, $$8 = 0, $$9 = 0, $$phi$trans$insert = 0, $$phi$trans$insert130 = 0, $$phi$trans$insert133 = 0, $$phi$trans$insert135 = 0, $$phi$trans$insert136 = 0, $$phi$trans$insert141 = 0, $$phi$trans$insert142 = 0, $$phi$trans$insert145 = 0, $$phi$trans$insert147 = 0, $$phi$trans$insert148 = 0, $$phi$trans$insert153 = 0, $$phi$trans$insert154 = 0;
 var $$phi$trans$insert157 = 0, $$phi$trans$insert159 = 0, $$phi$trans$insert160 = 0, $$phi$trans$insert164 = 0, $$phi$trans$insert165 = 0, $$phi$trans$insert168 = 0, $$phi$trans$insert170 = 0, $$phi$trans$insert171 = 0, $$phi$trans$insert175 = 0, $$phi$trans$insert176 = 0, $$phi$trans$insert179 = 0, $$phi$trans$insert181 = 0, $$phi$trans$insert182 = 0, $$phi$trans$insert186 = 0, $$phi$trans$insert187 = 0, $$phi$trans$insert190 = 0, $$phi$trans$insert192 = 0, $$phi$trans$insert193 = 0, $$phi$trans$insert197 = 0, $$phi$trans$insert198 = 0;
 var $$phi$trans$insert201 = 0, $$phi$trans$insert203 = 0, $$phi$trans$insert204 = 0, $$phi$trans$insert208 = 0, $$phi$trans$insert209 = 0, $$phi$trans$insert212 = 0, $$phi$trans$insert214 = 0, $$phi$trans$insert215 = 0, $$phi$trans$insert219 = 0, $$phi$trans$insert220 = 0, $$phi$trans$insert223 = 0, $$phi$trans$insert225 = 0, $$phi$trans$insert226 = 0, $$phi$trans$insert230 = 0, $$phi$trans$insert231 = 0, $$phi$trans$insert234 = 0, $$phi$trans$insert236 = 0, $$phi$trans$insert237 = 0, $$phi$trans$insert241 = 0, $$phi$trans$insert242 = 0;
 var $$phi$trans$insert245 = 0, $$phi$trans$insert247 = 0, $$phi$trans$insert248 = 0, $$phi$trans$insert252 = 0, $$phi$trans$insert253 = 0, $$phi$trans$insert256 = 0, $$phi$trans$insert258 = 0, $$phi$trans$insert259 = 0, $$pre = 0, $$pre128 = 0, $$pre129 = 0, $$pre131 = 0, $$pre134 = 0, $$pre137 = 0, $$pre138 = 0, $$pre139 = 0, $$pre140 = 0, $$pre143 = 0, $$pre146 = 0, $$pre149 = 0;
 var $$pre150 = 0, $$pre151 = 0, $$pre152 = 0, $$pre155 = 0, $$pre158 = 0, $$pre161 = 0, $$pre162 = 0, $$pre163 = 0, $$pre166 = 0, $$pre169 = 0, $$pre172 = 0, $$pre173 = 0, $$pre174 = 0, $$pre177 = 0, $$pre180 = 0, $$pre183 = 0, $$pre184 = 0, $$pre185 = 0, $$pre188 = 0, $$pre191 = 0;
 var $$pre194 = 0, $$pre195 = 0, $$pre196 = 0, $$pre199 = 0, $$pre202 = 0, $$pre205 = 0, $$pre206 = 0, $$pre207 = 0, $$pre210 = 0, $$pre213 = 0, $$pre216 = 0, $$pre217 = 0, $$pre218 = 0, $$pre221 = 0, $$pre224 = 0, $$pre227 = 0, $$pre228 = 0, $$pre229 = 0, $$pre232 = 0, $$pre235 = 0;
 var $$pre238 = 0, $$pre239 = 0, $$pre240 = 0, $$pre243 = 0, $$pre246 = 0, $$pre249 = 0, $$pre250 = 0, $$pre251 = 0, $$pre254 = 0, $$pre257 = 0, $$pre260 = 0, $$xinc1$0 = 0, $$xinc1$0$$25 = 0, $$xinc1109$0 = 0, $$xinc1109$0$$27 = 0, $$xinc1127$0 = 0, $$xinc1127$0$$28 = 0, $$xinc191$0 = 0, $$xinc191$0$$26 = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0;
 var $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0;
 var $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0;
 var $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0;
 var $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0, $1088 = 0;
 var $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0;
 var $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0;
 var $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0;
 var $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0, $116 = 0;
 var $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0;
 var $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0;
 var $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0;
 var $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0;
 var $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0;
 var $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0;
 var $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0;
 var $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0;
 var $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0;
 var $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0;
 var $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0;
 var $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0;
 var $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0;
 var $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0;
 var $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0;
 var $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0;
 var $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0;
 var $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0;
 var $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0;
 var $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0;
 var $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0;
 var $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0;
 var $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0;
 var $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0;
 var $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0;
 var $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0;
 var $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0;
 var $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0;
 var $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0;
 var $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0;
 var $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0;
 var $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0;
 var $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0;
 var $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0;
 var $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0;
 var $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0;
 var $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0;
 var $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0;
 var $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0;
 var $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0;
 var $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0;
 var $996 = 0, $997 = 0, $998 = 0, $999 = 0, $a$0 = 0, $b$0 = 0, $blendMode$off = 0, $d$0 = 0, $d$199 = 0, $d$2 = 0, $d105$0 = 0, $d105$1109 = 0, $d105$2 = 0, $d123$0 = 0, $d123$194 = 0, $d123$2 = 0, $d87$0 = 0, $d87$1104 = 0, $d87$2 = 0, $dinc1$0 = 0;
 var $dinc1$0$dinc2$0 = 0, $dinc1106$0 = 0, $dinc1106$0$dinc2107$0 = 0, $dinc1124$0 = 0, $dinc1124$0$dinc2125$0 = 0, $dinc188$0 = 0, $dinc188$0$dinc289$0 = 0, $dinc2$0 = 0, $dinc2$0$in = 0, $dinc2107$0 = 0, $dinc2107$0$in = 0, $dinc2125$0 = 0, $dinc2125$0$in = 0, $dinc289$0 = 0, $dinc289$0$in = 0, $g$0 = 0, $i$098 = 0, $i101$0108 = 0, $i119$093 = 0, $i83$0103 = 0;
 var $length$0$ph = 0, $length$061 = 0, $length1$0$ph = 0, $length1$064 = 0, $length15$0$ph = 0, $length15$059 = 0, $length22$0$ph = 0, $length22$073 = 0, $length29$0$ph = 0, $length29$076 = 0, $length36$0$ph = 0, $length36$079 = 0, $length43$0$ph = 0, $length43$070 = 0, $length50$0 = 0, $length50$0$ = 0, $length50$185 = 0, $length57$0 = 0, $length57$0$ = 0, $length57$188 = 0;
 var $length64$0 = 0, $length64$0$ = 0, $length64$191 = 0, $length71$0 = 0, $length71$0$ = 0, $length71$182 = 0, $length8$0$ph = 0, $length8$067 = 0, $not$ = 0, $not$29 = 0, $not$30 = 0, $not$31 = 0, $not$32 = 0, $not$33 = 0, $not$34 = 0, $not$35 = 0, $numpixels$0 = 0, $numpixels$0$in = 0, $numpixels$0$in$numpixels$0 = 0, $numpixels104$0 = 0;
 var $numpixels104$0$in = 0, $numpixels104$0$in$numpixels104$0 = 0, $numpixels122$0 = 0, $numpixels122$0$in = 0, $numpixels122$0$in$numpixels122$0 = 0, $numpixels86$0 = 0, $numpixels86$0$in = 0, $numpixels86$0$in$numpixels86$0 = 0, $pitch51$0 = 0, $pitch51$0$v = 0, $pitch51$1 = 0, $pitch51$1$v = 0, $pitch51$2 = 0, $pitch58$0 = 0, $pitch58$0$v = 0, $pitch58$1 = 0, $pitch58$1$v = 0, $pitch58$2 = 0, $pitch65$0 = 0, $pitch65$0$v = 0;
 var $pitch65$1 = 0, $pitch65$1$v = 0, $pitch65$2 = 0, $pitch72$0 = 0, $pitch72$0$v = 0, $pitch72$1 = 0, $pitch72$1$v = 0, $pitch72$2 = 0, $pixel$1$ph = 0, $pixel$160 = 0, $pixel10$1$ph = 0, $pixel10$166 = 0, $pixel17$1$ph = 0, $pixel17$158 = 0, $pixel24$1$ph = 0, $pixel24$172 = 0, $pixel3$1$ph = 0, $pixel3$163 = 0, $pixel31$1$ph = 0, $pixel31$175 = 0;
 var $pixel38$1$ph = 0, $pixel38$178 = 0, $pixel45$1$ph = 0, $pixel45$169 = 0, $pixel52$1 = 0, $pixel52$284 = 0, $pixel59$1 = 0, $pixel59$287 = 0, $pixel66$1 = 0, $pixel66$290 = 0, $pixel73$1 = 0, $pixel73$281 = 0, $r$0 = 0, $x$0100 = 0, $x$1 = 0, $x108$0110 = 0, $x108$1 = 0, $x126$095 = 0, $x126$1 = 0, $x90$0105 = 0;
 var $x90$1 = 0, $xinc1$0 = 0, $xinc1109$0 = 0, $xinc1127$0 = 0, $xinc191$0 = 0, $y$0101 = 0, $y$1 = 0, $y111$0111 = 0, $y111$1 = 0, $y129$096 = 0, $y129$1 = 0, $y93$0106 = 0, $y93$1 = 0, $yinc1$0 = 0, $yinc1$1 = 0, $yinc1$1$yinc2$1 = 0, $yinc1112$0 = 0, $yinc1112$1 = 0, $yinc1112$1$yinc2113$1 = 0, $yinc1130$0 = 0;
 var $yinc1130$1 = 0, $yinc1130$1$yinc2131$1 = 0, $yinc194$0 = 0, $yinc194$1 = 0, $yinc194$1$yinc295$1 = 0, $yinc2$1 = 0, $yinc2113$1 = 0, $yinc2131$1 = 0, $yinc295$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($dst)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $blendMode$off = (($blendMode) + -1)|0;
 $2 = ($blendMode$off>>>0)<(2);
 $3 = $_r&255;
 if ($2) {
  $4 = $_a&255;
  $5 = Math_imul($4, $3)|0;
  $6 = (($5>>>0) / 255)&-1;
  $7 = $_g&255;
  $8 = Math_imul($4, $7)|0;
  $9 = (($8>>>0) / 255)&-1;
  $10 = $_b&255;
  $11 = Math_imul($4, $10)|0;
  $12 = (($11>>>0) / 255)&-1;
  $a$0 = $4;$b$0 = $12;$g$0 = $9;$r$0 = $6;
 } else {
  $13 = $_g&255;
  $14 = $_b&255;
  $15 = $_a&255;
  $a$0 = $15;$b$0 = $14;$g$0 = $13;$r$0 = $3;
 }
 $16 = $a$0 ^ 255;
 $17 = ($y2|0)==($y1|0);
 if ($17) {
  switch ($blendMode|0) {
  case 1:  {
   $18 = ((($dst)) + 16|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ((($1)) + 9|0);
   $21 = HEAP8[$20>>0]|0;
   $22 = $21&255;
   $23 = (($19|0) / ($22|0))&-1;
   $24 = ($x2|0)<($x1|0);
   $25 = ((($dst)) + 20|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = Math_imul($23, $y2)|0;
   $28 = (($26) + ($27<<2)|0);
   if ($24) {
    $34 = (($28) + ($x2<<2)|0);
    $35 = ($draw_end|0)!=(0);
    $36 = ((($34)) + 4|0);
    $$13 = $35 ? $34 : $36;
    $37 = (($x1) - ($x2))|0;
    $38 = $35&1;
    $39 = (($37) + ($38))|0;
    $length$0$ph = $39;$pixel$1$ph = $$13;
   } else {
    $29 = (($28) + ($x1<<2)|0);
    $30 = ($draw_end|0)!=(0);
    $31 = (($x2) - ($x1))|0;
    $32 = $30&1;
    $33 = (($31) + ($32))|0;
    $length$0$ph = $33;$pixel$1$ph = $29;
   }
   $40 = ($length$0$ph|0)==(0);
   if ($40) {
    return;
   }
   $41 = ((($1)) + 12|0);
   $42 = ((($1)) + 32|0);
   $43 = ((($1)) + 28|0);
   $44 = ((($1)) + 16|0);
   $45 = ((($1)) + 20|0);
   $46 = ((($1)) + 24|0);
   $$pre250 = HEAP32[$42>>2]|0;
   $$pre251 = HEAP32[$43>>2]|0;
   $$phi$trans$insert252 = $$pre251 & 255;
   $$phi$trans$insert253 = (125924 + ($$phi$trans$insert252<<2)|0);
   $$pre254 = HEAP32[$$phi$trans$insert253>>2]|0;
   $47 = $$pre250 >>> 8;
   $48 = $47 & 255;
   $49 = $$pre251 >>> 8;
   $50 = $49 & 255;
   $$phi$trans$insert256 = (125924 + ($50<<2)|0);
   $$pre257 = HEAP32[$$phi$trans$insert256>>2]|0;
   $51 = $$pre250 >>> 16;
   $52 = $$pre251 >>> 16;
   $$phi$trans$insert258 = $52 & 255;
   $$phi$trans$insert259 = (125924 + ($$phi$trans$insert258<<2)|0);
   $$pre260 = HEAP32[$$phi$trans$insert259>>2]|0;
   $53 = $$pre250 & 255;
   $54 = $51 & 255;
   $length$061 = $length$0$ph;$pixel$160 = $pixel$1$ph;
   while(1) {
    $55 = (($length$061) + -1)|0;
    $56 = HEAP32[$pixel$160>>2]|0;
    $57 = HEAP32[$41>>2]|0;
    $58 = $57 & $56;
    $59 = $58 >>> $53;
    $60 = (($$pre254) + ($59)|0);
    $61 = HEAP8[$60>>0]|0;
    $62 = $61&255;
    $63 = HEAP32[$44>>2]|0;
    $64 = $63 & $56;
    $65 = $64 >>> $48;
    $66 = (($$pre257) + ($65)|0);
    $67 = HEAP8[$66>>0]|0;
    $68 = $67&255;
    $69 = HEAP32[$45>>2]|0;
    $70 = $69 & $56;
    $71 = $70 >>> $54;
    $72 = (($$pre260) + ($71)|0);
    $73 = HEAP8[$72>>0]|0;
    $74 = $73&255;
    $75 = Math_imul($62, $16)|0;
    $76 = (($75>>>0) / 255)&-1;
    $77 = (($76) + ($r$0))|0;
    $78 = Math_imul($68, $16)|0;
    $79 = (($78>>>0) / 255)&-1;
    $80 = (($79) + ($g$0))|0;
    $81 = Math_imul($74, $16)|0;
    $82 = (($81>>>0) / 255)&-1;
    $83 = (($82) + ($b$0))|0;
    $84 = $77 >>> $$phi$trans$insert252;
    $85 = $84 << $53;
    $86 = $80 >>> $50;
    $87 = $86 << $48;
    $88 = $87 | $85;
    $89 = $83 >>> $$phi$trans$insert258;
    $90 = $89 << $54;
    $91 = HEAP32[$46>>2]|0;
    $92 = $88 | $91;
    $93 = $92 | $90;
    HEAP32[$pixel$160>>2] = $93;
    $94 = ((($pixel$160)) + 4|0);
    $95 = ($55|0)==(0);
    if ($95) {
     break;
    } else {
     $length$061 = $55;$pixel$160 = $94;
    }
   }
   return;
   break;
  }
  case 2:  {
   $96 = ((($dst)) + 16|0);
   $97 = HEAP32[$96>>2]|0;
   $98 = ((($1)) + 9|0);
   $99 = HEAP8[$98>>0]|0;
   $100 = $99&255;
   $101 = (($97|0) / ($100|0))&-1;
   $102 = ($x2|0)<($x1|0);
   $103 = ((($dst)) + 20|0);
   $104 = HEAP32[$103>>2]|0;
   $105 = Math_imul($101, $y2)|0;
   $106 = (($104) + ($105<<2)|0);
   if ($102) {
    $112 = (($106) + ($x2<<2)|0);
    $113 = ($draw_end|0)!=(0);
    $114 = ((($112)) + 4|0);
    $$14 = $113 ? $112 : $114;
    $115 = (($x1) - ($x2))|0;
    $116 = $113&1;
    $117 = (($115) + ($116))|0;
    $length1$0$ph = $117;$pixel3$1$ph = $$14;
   } else {
    $107 = (($106) + ($x1<<2)|0);
    $108 = ($draw_end|0)!=(0);
    $109 = (($x2) - ($x1))|0;
    $110 = $108&1;
    $111 = (($109) + ($110))|0;
    $length1$0$ph = $111;$pixel3$1$ph = $107;
   }
   $118 = ($length1$0$ph|0)==(0);
   if ($118) {
    return;
   }
   $119 = ((($1)) + 12|0);
   $120 = ((($1)) + 32|0);
   $121 = ((($1)) + 28|0);
   $122 = ((($1)) + 16|0);
   $123 = ((($1)) + 20|0);
   $124 = ((($1)) + 24|0);
   $$pre239 = HEAP32[$120>>2]|0;
   $$pre240 = HEAP32[$121>>2]|0;
   $$phi$trans$insert241 = $$pre240 & 255;
   $$phi$trans$insert242 = (125924 + ($$phi$trans$insert241<<2)|0);
   $$pre243 = HEAP32[$$phi$trans$insert242>>2]|0;
   $125 = $$pre239 >>> 8;
   $126 = $125 & 255;
   $127 = $$pre240 >>> 8;
   $128 = $127 & 255;
   $$phi$trans$insert245 = (125924 + ($128<<2)|0);
   $$pre246 = HEAP32[$$phi$trans$insert245>>2]|0;
   $129 = $$pre239 >>> 16;
   $130 = $$pre240 >>> 16;
   $$phi$trans$insert247 = $130 & 255;
   $$phi$trans$insert248 = (125924 + ($$phi$trans$insert247<<2)|0);
   $$pre249 = HEAP32[$$phi$trans$insert248>>2]|0;
   $131 = $$pre239 & 255;
   $132 = $129 & 255;
   $length1$064 = $length1$0$ph;$pixel3$163 = $pixel3$1$ph;
   while(1) {
    $133 = (($length1$064) + -1)|0;
    $134 = HEAP32[$pixel3$163>>2]|0;
    $135 = HEAP32[$119>>2]|0;
    $136 = $135 & $134;
    $137 = $136 >>> $131;
    $138 = (($$pre243) + ($137)|0);
    $139 = HEAP8[$138>>0]|0;
    $140 = $139&255;
    $141 = HEAP32[$122>>2]|0;
    $142 = $141 & $134;
    $143 = $142 >>> $126;
    $144 = (($$pre246) + ($143)|0);
    $145 = HEAP8[$144>>0]|0;
    $146 = $145&255;
    $147 = HEAP32[$123>>2]|0;
    $148 = $147 & $134;
    $149 = $148 >>> $132;
    $150 = (($$pre249) + ($149)|0);
    $151 = HEAP8[$150>>0]|0;
    $152 = $151&255;
    $153 = (($140) + ($r$0))|0;
    $154 = ($153>>>0)>(255);
    $$ = $154 ? 255 : $153;
    $155 = (($146) + ($g$0))|0;
    $156 = ($155>>>0)>(255);
    $$8 = $156 ? 255 : $155;
    $157 = (($152) + ($b$0))|0;
    $158 = ($157>>>0)>(255);
    $$2 = $158 ? 255 : $157;
    $159 = $$ >>> $$phi$trans$insert241;
    $160 = $159 << $131;
    $161 = $$8 >>> $128;
    $162 = $161 << $126;
    $163 = $162 | $160;
    $164 = $$2 >>> $$phi$trans$insert247;
    $165 = $164 << $132;
    $166 = HEAP32[$124>>2]|0;
    $167 = $163 | $166;
    $168 = $167 | $165;
    HEAP32[$pixel3$163>>2] = $168;
    $169 = ((($pixel3$163)) + 4|0);
    $170 = ($133|0)==(0);
    if ($170) {
     break;
    } else {
     $length1$064 = $133;$pixel3$163 = $169;
    }
   }
   return;
   break;
  }
  case 4:  {
   $171 = ((($dst)) + 16|0);
   $172 = HEAP32[$171>>2]|0;
   $173 = ((($1)) + 9|0);
   $174 = HEAP8[$173>>0]|0;
   $175 = $174&255;
   $176 = (($172|0) / ($175|0))&-1;
   $177 = ($x2|0)<($x1|0);
   $178 = ((($dst)) + 20|0);
   $179 = HEAP32[$178>>2]|0;
   $180 = Math_imul($176, $y2)|0;
   $181 = (($179) + ($180<<2)|0);
   if ($177) {
    $187 = (($181) + ($x2<<2)|0);
    $188 = ($draw_end|0)!=(0);
    $189 = ((($187)) + 4|0);
    $$15 = $188 ? $187 : $189;
    $190 = (($x1) - ($x2))|0;
    $191 = $188&1;
    $192 = (($190) + ($191))|0;
    $length8$0$ph = $192;$pixel10$1$ph = $$15;
   } else {
    $182 = (($181) + ($x1<<2)|0);
    $183 = ($draw_end|0)!=(0);
    $184 = (($x2) - ($x1))|0;
    $185 = $183&1;
    $186 = (($184) + ($185))|0;
    $length8$0$ph = $186;$pixel10$1$ph = $182;
   }
   $193 = ($length8$0$ph|0)==(0);
   if ($193) {
    return;
   }
   $194 = ((($1)) + 12|0);
   $195 = ((($1)) + 32|0);
   $196 = ((($1)) + 28|0);
   $197 = ((($1)) + 16|0);
   $198 = ((($1)) + 20|0);
   $199 = ((($1)) + 24|0);
   $$pre228 = HEAP32[$195>>2]|0;
   $$pre229 = HEAP32[$196>>2]|0;
   $$phi$trans$insert230 = $$pre229 & 255;
   $$phi$trans$insert231 = (125924 + ($$phi$trans$insert230<<2)|0);
   $$pre232 = HEAP32[$$phi$trans$insert231>>2]|0;
   $200 = $$pre228 >>> 8;
   $201 = $200 & 255;
   $202 = $$pre229 >>> 8;
   $203 = $202 & 255;
   $$phi$trans$insert234 = (125924 + ($203<<2)|0);
   $$pre235 = HEAP32[$$phi$trans$insert234>>2]|0;
   $204 = $$pre228 >>> 16;
   $205 = $$pre229 >>> 16;
   $$phi$trans$insert236 = $205 & 255;
   $$phi$trans$insert237 = (125924 + ($$phi$trans$insert236<<2)|0);
   $$pre238 = HEAP32[$$phi$trans$insert237>>2]|0;
   $206 = $$pre228 & 255;
   $207 = $204 & 255;
   $length8$067 = $length8$0$ph;$pixel10$166 = $pixel10$1$ph;
   while(1) {
    $208 = (($length8$067) + -1)|0;
    $209 = HEAP32[$pixel10$166>>2]|0;
    $210 = HEAP32[$194>>2]|0;
    $211 = $210 & $209;
    $212 = $211 >>> $206;
    $213 = (($$pre232) + ($212)|0);
    $214 = HEAP8[$213>>0]|0;
    $215 = $214&255;
    $216 = HEAP32[$197>>2]|0;
    $217 = $216 & $209;
    $218 = $217 >>> $201;
    $219 = (($$pre235) + ($218)|0);
    $220 = HEAP8[$219>>0]|0;
    $221 = $220&255;
    $222 = HEAP32[$198>>2]|0;
    $223 = $222 & $209;
    $224 = $223 >>> $207;
    $225 = (($$pre238) + ($224)|0);
    $226 = HEAP8[$225>>0]|0;
    $227 = $226&255;
    $228 = Math_imul($215, $r$0)|0;
    $229 = (($228>>>0) / 255)&-1;
    $230 = Math_imul($221, $g$0)|0;
    $231 = (($230>>>0) / 255)&-1;
    $232 = Math_imul($227, $b$0)|0;
    $233 = (($232>>>0) / 255)&-1;
    $234 = $229 >>> $$phi$trans$insert230;
    $235 = $234 << $206;
    $236 = $231 >>> $203;
    $237 = $236 << $201;
    $238 = $237 | $235;
    $239 = $233 >>> $$phi$trans$insert236;
    $240 = $239 << $207;
    $241 = HEAP32[$199>>2]|0;
    $242 = $238 | $241;
    $243 = $242 | $240;
    HEAP32[$pixel10$166>>2] = $243;
    $244 = ((($pixel10$166)) + 4|0);
    $245 = ($208|0)==(0);
    if ($245) {
     break;
    } else {
     $length8$067 = $208;$pixel10$166 = $244;
    }
   }
   return;
   break;
  }
  default: {
   $246 = ((($dst)) + 16|0);
   $247 = HEAP32[$246>>2]|0;
   $248 = ((($1)) + 9|0);
   $249 = HEAP8[$248>>0]|0;
   $250 = $249&255;
   $251 = (($247|0) / ($250|0))&-1;
   $252 = ($x2|0)<($x1|0);
   $253 = ((($dst)) + 20|0);
   $254 = HEAP32[$253>>2]|0;
   $255 = Math_imul($251, $y2)|0;
   $256 = (($254) + ($255<<2)|0);
   if ($252) {
    $262 = (($256) + ($x2<<2)|0);
    $263 = ($draw_end|0)!=(0);
    $264 = ((($262)) + 4|0);
    $$16 = $263 ? $262 : $264;
    $265 = (($x1) - ($x2))|0;
    $266 = $263&1;
    $267 = (($265) + ($266))|0;
    $length15$0$ph = $267;$pixel17$1$ph = $$16;
   } else {
    $257 = (($256) + ($x1<<2)|0);
    $258 = ($draw_end|0)!=(0);
    $259 = (($x2) - ($x1))|0;
    $260 = $258&1;
    $261 = (($259) + ($260))|0;
    $length15$0$ph = $261;$pixel17$1$ph = $257;
   }
   $268 = ($length15$0$ph|0)==(0);
   if ($268) {
    return;
   }
   $269 = ((($1)) + 28|0);
   $270 = HEAP32[$269>>2]|0;
   $271 = $270 & 255;
   $272 = $r$0 >>> $271;
   $273 = ((($1)) + 32|0);
   $274 = HEAP32[$273>>2]|0;
   $275 = $274 & 255;
   $276 = $272 << $275;
   $277 = $270 >>> 8;
   $278 = $277 & 255;
   $279 = $g$0 >>> $278;
   $280 = $274 >>> 8;
   $281 = $280 & 255;
   $282 = $279 << $281;
   $283 = $282 | $276;
   $284 = $270 >>> 16;
   $285 = $284 & 255;
   $286 = $b$0 >>> $285;
   $287 = $274 >>> 16;
   $288 = $287 & 255;
   $289 = $286 << $288;
   $290 = ((($1)) + 24|0);
   $length15$059 = $length15$0$ph;$pixel17$158 = $pixel17$1$ph;
   while(1) {
    $291 = (($length15$059) + -1)|0;
    $292 = HEAP32[$290>>2]|0;
    $293 = $283 | $292;
    $294 = $293 | $289;
    HEAP32[$pixel17$158>>2] = $294;
    $295 = ((($pixel17$158)) + 4|0);
    $296 = ($291|0)==(0);
    if ($296) {
     break;
    } else {
     $length15$059 = $291;$pixel17$158 = $295;
    }
   }
   return;
  }
  }
 }
 $297 = ($x2|0)==($x1|0);
 if ($297) {
  switch ($blendMode|0) {
  case 1:  {
   $298 = ((($dst)) + 16|0);
   $299 = HEAP32[$298>>2]|0;
   $300 = ((($1)) + 9|0);
   $301 = HEAP8[$300>>0]|0;
   $302 = $301&255;
   $303 = (($299|0) / ($302|0))&-1;
   $304 = ($y2|0)<($y1|0);
   $305 = ((($dst)) + 20|0);
   $306 = HEAP32[$305>>2]|0;
   if ($304) {
    $314 = Math_imul($303, $y2)|0;
    $315 = (($306) + ($314<<2)|0);
    $316 = (($315) + ($x2<<2)|0);
    $317 = ($draw_end|0)!=(0);
    $318 = (($316) + ($303<<2)|0);
    $$17 = $317 ? $316 : $318;
    $319 = (($y1) - ($y2))|0;
    $320 = $317&1;
    $321 = (($319) + ($320))|0;
    $length22$0$ph = $321;$pixel24$1$ph = $$17;
   } else {
    $307 = Math_imul($303, $y1)|0;
    $308 = (($306) + ($307<<2)|0);
    $309 = (($308) + ($x2<<2)|0);
    $310 = ($draw_end|0)!=(0);
    $311 = (($y2) - ($y1))|0;
    $312 = $310&1;
    $313 = (($311) + ($312))|0;
    $length22$0$ph = $313;$pixel24$1$ph = $309;
   }
   $322 = ($length22$0$ph|0)==(0);
   if ($322) {
    return;
   }
   $323 = ((($1)) + 12|0);
   $324 = ((($1)) + 32|0);
   $325 = ((($1)) + 28|0);
   $326 = ((($1)) + 16|0);
   $327 = ((($1)) + 20|0);
   $328 = ((($1)) + 24|0);
   $$pre217 = HEAP32[$324>>2]|0;
   $$pre218 = HEAP32[$325>>2]|0;
   $$phi$trans$insert219 = $$pre218 & 255;
   $$phi$trans$insert220 = (125924 + ($$phi$trans$insert219<<2)|0);
   $$pre221 = HEAP32[$$phi$trans$insert220>>2]|0;
   $329 = $$pre217 >>> 8;
   $330 = $329 & 255;
   $331 = $$pre218 >>> 8;
   $332 = $331 & 255;
   $$phi$trans$insert223 = (125924 + ($332<<2)|0);
   $$pre224 = HEAP32[$$phi$trans$insert223>>2]|0;
   $333 = $$pre217 >>> 16;
   $334 = $$pre218 >>> 16;
   $$phi$trans$insert225 = $334 & 255;
   $$phi$trans$insert226 = (125924 + ($$phi$trans$insert225<<2)|0);
   $$pre227 = HEAP32[$$phi$trans$insert226>>2]|0;
   $335 = $$pre217 & 255;
   $336 = $333 & 255;
   $length22$073 = $length22$0$ph;$pixel24$172 = $pixel24$1$ph;
   while(1) {
    $337 = (($length22$073) + -1)|0;
    $338 = HEAP32[$pixel24$172>>2]|0;
    $339 = HEAP32[$323>>2]|0;
    $340 = $339 & $338;
    $341 = $340 >>> $335;
    $342 = (($$pre221) + ($341)|0);
    $343 = HEAP8[$342>>0]|0;
    $344 = $343&255;
    $345 = HEAP32[$326>>2]|0;
    $346 = $345 & $338;
    $347 = $346 >>> $330;
    $348 = (($$pre224) + ($347)|0);
    $349 = HEAP8[$348>>0]|0;
    $350 = $349&255;
    $351 = HEAP32[$327>>2]|0;
    $352 = $351 & $338;
    $353 = $352 >>> $336;
    $354 = (($$pre227) + ($353)|0);
    $355 = HEAP8[$354>>0]|0;
    $356 = $355&255;
    $357 = Math_imul($344, $16)|0;
    $358 = (($357>>>0) / 255)&-1;
    $359 = (($358) + ($r$0))|0;
    $360 = Math_imul($350, $16)|0;
    $361 = (($360>>>0) / 255)&-1;
    $362 = (($361) + ($g$0))|0;
    $363 = Math_imul($356, $16)|0;
    $364 = (($363>>>0) / 255)&-1;
    $365 = (($364) + ($b$0))|0;
    $366 = $359 >>> $$phi$trans$insert219;
    $367 = $366 << $335;
    $368 = $362 >>> $332;
    $369 = $368 << $330;
    $370 = $369 | $367;
    $371 = $365 >>> $$phi$trans$insert225;
    $372 = $371 << $336;
    $373 = HEAP32[$328>>2]|0;
    $374 = $370 | $373;
    $375 = $374 | $372;
    HEAP32[$pixel24$172>>2] = $375;
    $376 = (($pixel24$172) + ($303<<2)|0);
    $377 = ($337|0)==(0);
    if ($377) {
     break;
    } else {
     $length22$073 = $337;$pixel24$172 = $376;
    }
   }
   return;
   break;
  }
  case 2:  {
   $378 = ((($dst)) + 16|0);
   $379 = HEAP32[$378>>2]|0;
   $380 = ((($1)) + 9|0);
   $381 = HEAP8[$380>>0]|0;
   $382 = $381&255;
   $383 = (($379|0) / ($382|0))&-1;
   $384 = ($y2|0)<($y1|0);
   $385 = ((($dst)) + 20|0);
   $386 = HEAP32[$385>>2]|0;
   if ($384) {
    $394 = Math_imul($383, $y2)|0;
    $395 = (($386) + ($394<<2)|0);
    $396 = (($395) + ($x2<<2)|0);
    $397 = ($draw_end|0)!=(0);
    $398 = (($396) + ($383<<2)|0);
    $$18 = $397 ? $396 : $398;
    $399 = (($y1) - ($y2))|0;
    $400 = $397&1;
    $401 = (($399) + ($400))|0;
    $length29$0$ph = $401;$pixel31$1$ph = $$18;
   } else {
    $387 = Math_imul($383, $y1)|0;
    $388 = (($386) + ($387<<2)|0);
    $389 = (($388) + ($x2<<2)|0);
    $390 = ($draw_end|0)!=(0);
    $391 = (($y2) - ($y1))|0;
    $392 = $390&1;
    $393 = (($391) + ($392))|0;
    $length29$0$ph = $393;$pixel31$1$ph = $389;
   }
   $402 = ($length29$0$ph|0)==(0);
   if ($402) {
    return;
   }
   $403 = ((($1)) + 12|0);
   $404 = ((($1)) + 32|0);
   $405 = ((($1)) + 28|0);
   $406 = ((($1)) + 16|0);
   $407 = ((($1)) + 20|0);
   $408 = ((($1)) + 24|0);
   $$pre206 = HEAP32[$404>>2]|0;
   $$pre207 = HEAP32[$405>>2]|0;
   $$phi$trans$insert208 = $$pre207 & 255;
   $$phi$trans$insert209 = (125924 + ($$phi$trans$insert208<<2)|0);
   $$pre210 = HEAP32[$$phi$trans$insert209>>2]|0;
   $409 = $$pre206 >>> 8;
   $410 = $409 & 255;
   $411 = $$pre207 >>> 8;
   $412 = $411 & 255;
   $$phi$trans$insert212 = (125924 + ($412<<2)|0);
   $$pre213 = HEAP32[$$phi$trans$insert212>>2]|0;
   $413 = $$pre206 >>> 16;
   $414 = $$pre207 >>> 16;
   $$phi$trans$insert214 = $414 & 255;
   $$phi$trans$insert215 = (125924 + ($$phi$trans$insert214<<2)|0);
   $$pre216 = HEAP32[$$phi$trans$insert215>>2]|0;
   $415 = $$pre206 & 255;
   $416 = $413 & 255;
   $length29$076 = $length29$0$ph;$pixel31$175 = $pixel31$1$ph;
   while(1) {
    $417 = (($length29$076) + -1)|0;
    $418 = HEAP32[$pixel31$175>>2]|0;
    $419 = HEAP32[$403>>2]|0;
    $420 = $419 & $418;
    $421 = $420 >>> $415;
    $422 = (($$pre210) + ($421)|0);
    $423 = HEAP8[$422>>0]|0;
    $424 = $423&255;
    $425 = HEAP32[$406>>2]|0;
    $426 = $425 & $418;
    $427 = $426 >>> $410;
    $428 = (($$pre213) + ($427)|0);
    $429 = HEAP8[$428>>0]|0;
    $430 = $429&255;
    $431 = HEAP32[$407>>2]|0;
    $432 = $431 & $418;
    $433 = $432 >>> $416;
    $434 = (($$pre216) + ($433)|0);
    $435 = HEAP8[$434>>0]|0;
    $436 = $435&255;
    $437 = (($424) + ($r$0))|0;
    $438 = ($437>>>0)>(255);
    $$3 = $438 ? 255 : $437;
    $439 = (($430) + ($g$0))|0;
    $440 = ($439>>>0)>(255);
    $$9 = $440 ? 255 : $439;
    $441 = (($436) + ($b$0))|0;
    $442 = ($441>>>0)>(255);
    $$4 = $442 ? 255 : $441;
    $443 = $$3 >>> $$phi$trans$insert208;
    $444 = $443 << $415;
    $445 = $$9 >>> $412;
    $446 = $445 << $410;
    $447 = $446 | $444;
    $448 = $$4 >>> $$phi$trans$insert214;
    $449 = $448 << $416;
    $450 = HEAP32[$408>>2]|0;
    $451 = $447 | $450;
    $452 = $451 | $449;
    HEAP32[$pixel31$175>>2] = $452;
    $453 = (($pixel31$175) + ($383<<2)|0);
    $454 = ($417|0)==(0);
    if ($454) {
     break;
    } else {
     $length29$076 = $417;$pixel31$175 = $453;
    }
   }
   return;
   break;
  }
  case 4:  {
   $455 = ((($dst)) + 16|0);
   $456 = HEAP32[$455>>2]|0;
   $457 = ((($1)) + 9|0);
   $458 = HEAP8[$457>>0]|0;
   $459 = $458&255;
   $460 = (($456|0) / ($459|0))&-1;
   $461 = ($y2|0)<($y1|0);
   $462 = ((($dst)) + 20|0);
   $463 = HEAP32[$462>>2]|0;
   if ($461) {
    $471 = Math_imul($460, $y2)|0;
    $472 = (($463) + ($471<<2)|0);
    $473 = (($472) + ($x2<<2)|0);
    $474 = ($draw_end|0)!=(0);
    $475 = (($473) + ($460<<2)|0);
    $$19 = $474 ? $473 : $475;
    $476 = (($y1) - ($y2))|0;
    $477 = $474&1;
    $478 = (($476) + ($477))|0;
    $length36$0$ph = $478;$pixel38$1$ph = $$19;
   } else {
    $464 = Math_imul($460, $y1)|0;
    $465 = (($463) + ($464<<2)|0);
    $466 = (($465) + ($x2<<2)|0);
    $467 = ($draw_end|0)!=(0);
    $468 = (($y2) - ($y1))|0;
    $469 = $467&1;
    $470 = (($468) + ($469))|0;
    $length36$0$ph = $470;$pixel38$1$ph = $466;
   }
   $479 = ($length36$0$ph|0)==(0);
   if ($479) {
    return;
   }
   $480 = ((($1)) + 12|0);
   $481 = ((($1)) + 32|0);
   $482 = ((($1)) + 28|0);
   $483 = ((($1)) + 16|0);
   $484 = ((($1)) + 20|0);
   $485 = ((($1)) + 24|0);
   $$pre195 = HEAP32[$481>>2]|0;
   $$pre196 = HEAP32[$482>>2]|0;
   $$phi$trans$insert197 = $$pre196 & 255;
   $$phi$trans$insert198 = (125924 + ($$phi$trans$insert197<<2)|0);
   $$pre199 = HEAP32[$$phi$trans$insert198>>2]|0;
   $486 = $$pre195 >>> 8;
   $487 = $486 & 255;
   $488 = $$pre196 >>> 8;
   $489 = $488 & 255;
   $$phi$trans$insert201 = (125924 + ($489<<2)|0);
   $$pre202 = HEAP32[$$phi$trans$insert201>>2]|0;
   $490 = $$pre195 >>> 16;
   $491 = $$pre196 >>> 16;
   $$phi$trans$insert203 = $491 & 255;
   $$phi$trans$insert204 = (125924 + ($$phi$trans$insert203<<2)|0);
   $$pre205 = HEAP32[$$phi$trans$insert204>>2]|0;
   $492 = $$pre195 & 255;
   $493 = $490 & 255;
   $length36$079 = $length36$0$ph;$pixel38$178 = $pixel38$1$ph;
   while(1) {
    $494 = (($length36$079) + -1)|0;
    $495 = HEAP32[$pixel38$178>>2]|0;
    $496 = HEAP32[$480>>2]|0;
    $497 = $496 & $495;
    $498 = $497 >>> $492;
    $499 = (($$pre199) + ($498)|0);
    $500 = HEAP8[$499>>0]|0;
    $501 = $500&255;
    $502 = HEAP32[$483>>2]|0;
    $503 = $502 & $495;
    $504 = $503 >>> $487;
    $505 = (($$pre202) + ($504)|0);
    $506 = HEAP8[$505>>0]|0;
    $507 = $506&255;
    $508 = HEAP32[$484>>2]|0;
    $509 = $508 & $495;
    $510 = $509 >>> $493;
    $511 = (($$pre205) + ($510)|0);
    $512 = HEAP8[$511>>0]|0;
    $513 = $512&255;
    $514 = Math_imul($501, $r$0)|0;
    $515 = (($514>>>0) / 255)&-1;
    $516 = Math_imul($507, $g$0)|0;
    $517 = (($516>>>0) / 255)&-1;
    $518 = Math_imul($513, $b$0)|0;
    $519 = (($518>>>0) / 255)&-1;
    $520 = $515 >>> $$phi$trans$insert197;
    $521 = $520 << $492;
    $522 = $517 >>> $489;
    $523 = $522 << $487;
    $524 = $523 | $521;
    $525 = $519 >>> $$phi$trans$insert203;
    $526 = $525 << $493;
    $527 = HEAP32[$485>>2]|0;
    $528 = $524 | $527;
    $529 = $528 | $526;
    HEAP32[$pixel38$178>>2] = $529;
    $530 = (($pixel38$178) + ($460<<2)|0);
    $531 = ($494|0)==(0);
    if ($531) {
     break;
    } else {
     $length36$079 = $494;$pixel38$178 = $530;
    }
   }
   return;
   break;
  }
  default: {
   $532 = ((($dst)) + 16|0);
   $533 = HEAP32[$532>>2]|0;
   $534 = ((($1)) + 9|0);
   $535 = HEAP8[$534>>0]|0;
   $536 = $535&255;
   $537 = (($533|0) / ($536|0))&-1;
   $538 = ($y2|0)<($y1|0);
   $539 = ((($dst)) + 20|0);
   $540 = HEAP32[$539>>2]|0;
   if ($538) {
    $548 = Math_imul($537, $y2)|0;
    $549 = (($540) + ($548<<2)|0);
    $550 = (($549) + ($x2<<2)|0);
    $551 = ($draw_end|0)!=(0);
    $552 = (($550) + ($537<<2)|0);
    $$20 = $551 ? $550 : $552;
    $553 = (($y1) - ($y2))|0;
    $554 = $551&1;
    $555 = (($553) + ($554))|0;
    $length43$0$ph = $555;$pixel45$1$ph = $$20;
   } else {
    $541 = Math_imul($537, $y1)|0;
    $542 = (($540) + ($541<<2)|0);
    $543 = (($542) + ($x2<<2)|0);
    $544 = ($draw_end|0)!=(0);
    $545 = (($y2) - ($y1))|0;
    $546 = $544&1;
    $547 = (($545) + ($546))|0;
    $length43$0$ph = $547;$pixel45$1$ph = $543;
   }
   $556 = ($length43$0$ph|0)==(0);
   if ($556) {
    return;
   }
   $557 = ((($1)) + 28|0);
   $558 = HEAP32[$557>>2]|0;
   $559 = $558 & 255;
   $560 = $r$0 >>> $559;
   $561 = ((($1)) + 32|0);
   $562 = HEAP32[$561>>2]|0;
   $563 = $562 & 255;
   $564 = $560 << $563;
   $565 = $558 >>> 8;
   $566 = $565 & 255;
   $567 = $g$0 >>> $566;
   $568 = $562 >>> 8;
   $569 = $568 & 255;
   $570 = $567 << $569;
   $571 = $570 | $564;
   $572 = $558 >>> 16;
   $573 = $572 & 255;
   $574 = $b$0 >>> $573;
   $575 = $562 >>> 16;
   $576 = $575 & 255;
   $577 = $574 << $576;
   $578 = ((($1)) + 24|0);
   $length43$070 = $length43$0$ph;$pixel45$169 = $pixel45$1$ph;
   while(1) {
    $579 = (($length43$070) + -1)|0;
    $580 = HEAP32[$578>>2]|0;
    $581 = $571 | $580;
    $582 = $581 | $577;
    HEAP32[$pixel45$169>>2] = $582;
    $583 = (($pixel45$169) + ($537<<2)|0);
    $584 = ($579|0)==(0);
    if ($584) {
     break;
    } else {
     $length43$070 = $579;$pixel45$169 = $583;
    }
   }
   return;
  }
  }
 }
 $585 = (($x1) - ($x2))|0;
 $586 = ($585|0)<(0);
 $587 = (0 - ($585))|0;
 $588 = $586 ? $587 : $585;
 $589 = (($y1) - ($y2))|0;
 $590 = ($589|0)<(0);
 $591 = (0 - ($589))|0;
 $592 = $590 ? $591 : $589;
 $593 = ($588|0)==($592|0);
 if ($593) {
  switch ($blendMode|0) {
  case 1:  {
   $594 = ((($dst)) + 16|0);
   $595 = HEAP32[$594>>2]|0;
   $596 = ((($1)) + 9|0);
   $597 = HEAP8[$596>>0]|0;
   $598 = $597&255;
   $599 = (($595|0) / ($598|0))&-1;
   $600 = ($y2|0)<($y1|0);
   $601 = ((($dst)) + 20|0);
   $602 = HEAP32[$601>>2]|0;
   if ($600) {
    $608 = Math_imul($599, $y2)|0;
    $609 = (($602) + ($608<<2)|0);
    $610 = (($609) + ($x2<<2)|0);
    $611 = ($x2|0)>($x1|0);
    $pitch51$1$v = $611 ? -1 : 1;
    $pitch51$1 = (($599) + ($pitch51$1$v))|0;
    $612 = ($draw_end|0)==(0);
    $613 = (($610) + ($pitch51$1<<2)|0);
    $$21 = $612 ? $613 : $610;
    $length50$0 = $589;$pitch51$2 = $pitch51$1;$pixel52$1 = $$21;
   } else {
    $603 = Math_imul($599, $y1)|0;
    $604 = (($602) + ($603<<2)|0);
    $605 = (($604) + ($x1<<2)|0);
    $606 = ($x2|0)<($x1|0);
    $pitch51$0$v = $606 ? -1 : 1;
    $pitch51$0 = (($599) + ($pitch51$0$v))|0;
    $607 = (($y2) - ($y1))|0;
    $length50$0 = $607;$pitch51$2 = $pitch51$0;$pixel52$1 = $605;
   }
   $not$34 = ($draw_end|0)!=(0);
   $614 = $not$34&1;
   $length50$0$ = (($length50$0) + ($614))|0;
   $615 = ($length50$0$|0)==(0);
   if ($615) {
    return;
   }
   $616 = ((($1)) + 12|0);
   $617 = ((($1)) + 32|0);
   $618 = ((($1)) + 28|0);
   $619 = ((($1)) + 16|0);
   $620 = ((($1)) + 20|0);
   $621 = ((($1)) + 24|0);
   $$pre184 = HEAP32[$617>>2]|0;
   $$pre185 = HEAP32[$618>>2]|0;
   $$phi$trans$insert186 = $$pre185 & 255;
   $$phi$trans$insert187 = (125924 + ($$phi$trans$insert186<<2)|0);
   $$pre188 = HEAP32[$$phi$trans$insert187>>2]|0;
   $622 = $$pre184 >>> 8;
   $623 = $622 & 255;
   $624 = $$pre185 >>> 8;
   $625 = $624 & 255;
   $$phi$trans$insert190 = (125924 + ($625<<2)|0);
   $$pre191 = HEAP32[$$phi$trans$insert190>>2]|0;
   $626 = $$pre184 >>> 16;
   $627 = $$pre185 >>> 16;
   $$phi$trans$insert192 = $627 & 255;
   $$phi$trans$insert193 = (125924 + ($$phi$trans$insert192<<2)|0);
   $$pre194 = HEAP32[$$phi$trans$insert193>>2]|0;
   $628 = $$pre184 & 255;
   $629 = $626 & 255;
   $length50$185 = $length50$0$;$pixel52$284 = $pixel52$1;
   while(1) {
    $630 = (($length50$185) + -1)|0;
    $631 = HEAP32[$pixel52$284>>2]|0;
    $632 = HEAP32[$616>>2]|0;
    $633 = $632 & $631;
    $634 = $633 >>> $628;
    $635 = (($$pre188) + ($634)|0);
    $636 = HEAP8[$635>>0]|0;
    $637 = $636&255;
    $638 = HEAP32[$619>>2]|0;
    $639 = $638 & $631;
    $640 = $639 >>> $623;
    $641 = (($$pre191) + ($640)|0);
    $642 = HEAP8[$641>>0]|0;
    $643 = $642&255;
    $644 = HEAP32[$620>>2]|0;
    $645 = $644 & $631;
    $646 = $645 >>> $629;
    $647 = (($$pre194) + ($646)|0);
    $648 = HEAP8[$647>>0]|0;
    $649 = $648&255;
    $650 = Math_imul($637, $16)|0;
    $651 = (($650>>>0) / 255)&-1;
    $652 = (($651) + ($r$0))|0;
    $653 = Math_imul($643, $16)|0;
    $654 = (($653>>>0) / 255)&-1;
    $655 = (($654) + ($g$0))|0;
    $656 = Math_imul($649, $16)|0;
    $657 = (($656>>>0) / 255)&-1;
    $658 = (($657) + ($b$0))|0;
    $659 = $652 >>> $$phi$trans$insert186;
    $660 = $659 << $628;
    $661 = $655 >>> $625;
    $662 = $661 << $623;
    $663 = $662 | $660;
    $664 = $658 >>> $$phi$trans$insert192;
    $665 = $664 << $629;
    $666 = HEAP32[$621>>2]|0;
    $667 = $663 | $666;
    $668 = $667 | $665;
    HEAP32[$pixel52$284>>2] = $668;
    $669 = (($pixel52$284) + ($pitch51$2<<2)|0);
    $670 = ($630|0)==(0);
    if ($670) {
     break;
    } else {
     $length50$185 = $630;$pixel52$284 = $669;
    }
   }
   return;
   break;
  }
  case 2:  {
   $671 = ((($dst)) + 16|0);
   $672 = HEAP32[$671>>2]|0;
   $673 = ((($1)) + 9|0);
   $674 = HEAP8[$673>>0]|0;
   $675 = $674&255;
   $676 = (($672|0) / ($675|0))&-1;
   $677 = ($y2|0)<($y1|0);
   $678 = ((($dst)) + 20|0);
   $679 = HEAP32[$678>>2]|0;
   if ($677) {
    $685 = Math_imul($676, $y2)|0;
    $686 = (($679) + ($685<<2)|0);
    $687 = (($686) + ($x2<<2)|0);
    $688 = ($x2|0)>($x1|0);
    $pitch58$1$v = $688 ? -1 : 1;
    $pitch58$1 = (($676) + ($pitch58$1$v))|0;
    $689 = ($draw_end|0)==(0);
    $690 = (($687) + ($pitch58$1<<2)|0);
    $$22 = $689 ? $690 : $687;
    $length57$0 = $589;$pitch58$2 = $pitch58$1;$pixel59$1 = $$22;
   } else {
    $680 = Math_imul($676, $y1)|0;
    $681 = (($679) + ($680<<2)|0);
    $682 = (($681) + ($x1<<2)|0);
    $683 = ($x2|0)<($x1|0);
    $pitch58$0$v = $683 ? -1 : 1;
    $pitch58$0 = (($676) + ($pitch58$0$v))|0;
    $684 = (($y2) - ($y1))|0;
    $length57$0 = $684;$pitch58$2 = $pitch58$0;$pixel59$1 = $682;
   }
   $not$33 = ($draw_end|0)!=(0);
   $691 = $not$33&1;
   $length57$0$ = (($length57$0) + ($691))|0;
   $692 = ($length57$0$|0)==(0);
   if ($692) {
    return;
   }
   $693 = ((($1)) + 12|0);
   $694 = ((($1)) + 32|0);
   $695 = ((($1)) + 28|0);
   $696 = ((($1)) + 16|0);
   $697 = ((($1)) + 20|0);
   $698 = ((($1)) + 24|0);
   $$pre173 = HEAP32[$694>>2]|0;
   $$pre174 = HEAP32[$695>>2]|0;
   $$phi$trans$insert175 = $$pre174 & 255;
   $$phi$trans$insert176 = (125924 + ($$phi$trans$insert175<<2)|0);
   $$pre177 = HEAP32[$$phi$trans$insert176>>2]|0;
   $699 = $$pre173 >>> 8;
   $700 = $699 & 255;
   $701 = $$pre174 >>> 8;
   $702 = $701 & 255;
   $$phi$trans$insert179 = (125924 + ($702<<2)|0);
   $$pre180 = HEAP32[$$phi$trans$insert179>>2]|0;
   $703 = $$pre173 >>> 16;
   $704 = $$pre174 >>> 16;
   $$phi$trans$insert181 = $704 & 255;
   $$phi$trans$insert182 = (125924 + ($$phi$trans$insert181<<2)|0);
   $$pre183 = HEAP32[$$phi$trans$insert182>>2]|0;
   $705 = $$pre173 & 255;
   $706 = $703 & 255;
   $length57$188 = $length57$0$;$pixel59$287 = $pixel59$1;
   while(1) {
    $707 = (($length57$188) + -1)|0;
    $708 = HEAP32[$pixel59$287>>2]|0;
    $709 = HEAP32[$693>>2]|0;
    $710 = $709 & $708;
    $711 = $710 >>> $705;
    $712 = (($$pre177) + ($711)|0);
    $713 = HEAP8[$712>>0]|0;
    $714 = $713&255;
    $715 = HEAP32[$696>>2]|0;
    $716 = $715 & $708;
    $717 = $716 >>> $700;
    $718 = (($$pre180) + ($717)|0);
    $719 = HEAP8[$718>>0]|0;
    $720 = $719&255;
    $721 = HEAP32[$697>>2]|0;
    $722 = $721 & $708;
    $723 = $722 >>> $706;
    $724 = (($$pre183) + ($723)|0);
    $725 = HEAP8[$724>>0]|0;
    $726 = $725&255;
    $727 = (($714) + ($r$0))|0;
    $728 = ($727>>>0)>(255);
    $$5 = $728 ? 255 : $727;
    $729 = (($720) + ($g$0))|0;
    $730 = ($729>>>0)>(255);
    $$10 = $730 ? 255 : $729;
    $731 = (($726) + ($b$0))|0;
    $732 = ($731>>>0)>(255);
    $$6 = $732 ? 255 : $731;
    $733 = $$5 >>> $$phi$trans$insert175;
    $734 = $733 << $705;
    $735 = $$10 >>> $702;
    $736 = $735 << $700;
    $737 = $736 | $734;
    $738 = $$6 >>> $$phi$trans$insert181;
    $739 = $738 << $706;
    $740 = HEAP32[$698>>2]|0;
    $741 = $737 | $740;
    $742 = $741 | $739;
    HEAP32[$pixel59$287>>2] = $742;
    $743 = (($pixel59$287) + ($pitch58$2<<2)|0);
    $744 = ($707|0)==(0);
    if ($744) {
     break;
    } else {
     $length57$188 = $707;$pixel59$287 = $743;
    }
   }
   return;
   break;
  }
  case 4:  {
   $745 = ((($dst)) + 16|0);
   $746 = HEAP32[$745>>2]|0;
   $747 = ((($1)) + 9|0);
   $748 = HEAP8[$747>>0]|0;
   $749 = $748&255;
   $750 = (($746|0) / ($749|0))&-1;
   $751 = ($y2|0)<($y1|0);
   $752 = ((($dst)) + 20|0);
   $753 = HEAP32[$752>>2]|0;
   if ($751) {
    $759 = Math_imul($750, $y2)|0;
    $760 = (($753) + ($759<<2)|0);
    $761 = (($760) + ($x2<<2)|0);
    $762 = ($x2|0)>($x1|0);
    $pitch65$1$v = $762 ? -1 : 1;
    $pitch65$1 = (($750) + ($pitch65$1$v))|0;
    $763 = ($draw_end|0)==(0);
    $764 = (($761) + ($pitch65$1<<2)|0);
    $$23 = $763 ? $764 : $761;
    $length64$0 = $589;$pitch65$2 = $pitch65$1;$pixel66$1 = $$23;
   } else {
    $754 = Math_imul($750, $y1)|0;
    $755 = (($753) + ($754<<2)|0);
    $756 = (($755) + ($x1<<2)|0);
    $757 = ($x2|0)<($x1|0);
    $pitch65$0$v = $757 ? -1 : 1;
    $pitch65$0 = (($750) + ($pitch65$0$v))|0;
    $758 = (($y2) - ($y1))|0;
    $length64$0 = $758;$pitch65$2 = $pitch65$0;$pixel66$1 = $756;
   }
   $not$32 = ($draw_end|0)!=(0);
   $765 = $not$32&1;
   $length64$0$ = (($length64$0) + ($765))|0;
   $766 = ($length64$0$|0)==(0);
   if ($766) {
    return;
   }
   $767 = ((($1)) + 12|0);
   $768 = ((($1)) + 32|0);
   $769 = ((($1)) + 28|0);
   $770 = ((($1)) + 16|0);
   $771 = ((($1)) + 20|0);
   $772 = ((($1)) + 24|0);
   $$pre162 = HEAP32[$768>>2]|0;
   $$pre163 = HEAP32[$769>>2]|0;
   $$phi$trans$insert164 = $$pre163 & 255;
   $$phi$trans$insert165 = (125924 + ($$phi$trans$insert164<<2)|0);
   $$pre166 = HEAP32[$$phi$trans$insert165>>2]|0;
   $773 = $$pre162 >>> 8;
   $774 = $773 & 255;
   $775 = $$pre163 >>> 8;
   $776 = $775 & 255;
   $$phi$trans$insert168 = (125924 + ($776<<2)|0);
   $$pre169 = HEAP32[$$phi$trans$insert168>>2]|0;
   $777 = $$pre162 >>> 16;
   $778 = $$pre163 >>> 16;
   $$phi$trans$insert170 = $778 & 255;
   $$phi$trans$insert171 = (125924 + ($$phi$trans$insert170<<2)|0);
   $$pre172 = HEAP32[$$phi$trans$insert171>>2]|0;
   $779 = $$pre162 & 255;
   $780 = $777 & 255;
   $length64$191 = $length64$0$;$pixel66$290 = $pixel66$1;
   while(1) {
    $781 = (($length64$191) + -1)|0;
    $782 = HEAP32[$pixel66$290>>2]|0;
    $783 = HEAP32[$767>>2]|0;
    $784 = $783 & $782;
    $785 = $784 >>> $779;
    $786 = (($$pre166) + ($785)|0);
    $787 = HEAP8[$786>>0]|0;
    $788 = $787&255;
    $789 = HEAP32[$770>>2]|0;
    $790 = $789 & $782;
    $791 = $790 >>> $774;
    $792 = (($$pre169) + ($791)|0);
    $793 = HEAP8[$792>>0]|0;
    $794 = $793&255;
    $795 = HEAP32[$771>>2]|0;
    $796 = $795 & $782;
    $797 = $796 >>> $780;
    $798 = (($$pre172) + ($797)|0);
    $799 = HEAP8[$798>>0]|0;
    $800 = $799&255;
    $801 = Math_imul($788, $r$0)|0;
    $802 = (($801>>>0) / 255)&-1;
    $803 = Math_imul($794, $g$0)|0;
    $804 = (($803>>>0) / 255)&-1;
    $805 = Math_imul($800, $b$0)|0;
    $806 = (($805>>>0) / 255)&-1;
    $807 = $802 >>> $$phi$trans$insert164;
    $808 = $807 << $779;
    $809 = $804 >>> $776;
    $810 = $809 << $774;
    $811 = $810 | $808;
    $812 = $806 >>> $$phi$trans$insert170;
    $813 = $812 << $780;
    $814 = HEAP32[$772>>2]|0;
    $815 = $811 | $814;
    $816 = $815 | $813;
    HEAP32[$pixel66$290>>2] = $816;
    $817 = (($pixel66$290) + ($pitch65$2<<2)|0);
    $818 = ($781|0)==(0);
    if ($818) {
     break;
    } else {
     $length64$191 = $781;$pixel66$290 = $817;
    }
   }
   return;
   break;
  }
  default: {
   $819 = ((($dst)) + 16|0);
   $820 = HEAP32[$819>>2]|0;
   $821 = ((($1)) + 9|0);
   $822 = HEAP8[$821>>0]|0;
   $823 = $822&255;
   $824 = (($820|0) / ($823|0))&-1;
   $825 = ($y2|0)<($y1|0);
   $826 = ((($dst)) + 20|0);
   $827 = HEAP32[$826>>2]|0;
   if ($825) {
    $833 = Math_imul($824, $y2)|0;
    $834 = (($827) + ($833<<2)|0);
    $835 = (($834) + ($x2<<2)|0);
    $836 = ($x2|0)>($x1|0);
    $pitch72$1$v = $836 ? -1 : 1;
    $pitch72$1 = (($824) + ($pitch72$1$v))|0;
    $837 = ($draw_end|0)==(0);
    $838 = (($835) + ($pitch72$1<<2)|0);
    $$24 = $837 ? $838 : $835;
    $length71$0 = $589;$pitch72$2 = $pitch72$1;$pixel73$1 = $$24;
   } else {
    $828 = Math_imul($824, $y1)|0;
    $829 = (($827) + ($828<<2)|0);
    $830 = (($829) + ($x1<<2)|0);
    $831 = ($x2|0)<($x1|0);
    $pitch72$0$v = $831 ? -1 : 1;
    $pitch72$0 = (($824) + ($pitch72$0$v))|0;
    $832 = (($y2) - ($y1))|0;
    $length71$0 = $832;$pitch72$2 = $pitch72$0;$pixel73$1 = $830;
   }
   $not$35 = ($draw_end|0)!=(0);
   $839 = $not$35&1;
   $length71$0$ = (($length71$0) + ($839))|0;
   $840 = ($length71$0$|0)==(0);
   if ($840) {
    return;
   }
   $841 = ((($1)) + 28|0);
   $842 = HEAP32[$841>>2]|0;
   $843 = $842 & 255;
   $844 = $r$0 >>> $843;
   $845 = ((($1)) + 32|0);
   $846 = HEAP32[$845>>2]|0;
   $847 = $846 & 255;
   $848 = $844 << $847;
   $849 = $842 >>> 8;
   $850 = $849 & 255;
   $851 = $g$0 >>> $850;
   $852 = $846 >>> 8;
   $853 = $852 & 255;
   $854 = $851 << $853;
   $855 = $854 | $848;
   $856 = $842 >>> 16;
   $857 = $856 & 255;
   $858 = $b$0 >>> $857;
   $859 = $846 >>> 16;
   $860 = $859 & 255;
   $861 = $858 << $860;
   $862 = ((($1)) + 24|0);
   $length71$182 = $length71$0$;$pixel73$281 = $pixel73$1;
   while(1) {
    $863 = (($length71$182) + -1)|0;
    $864 = HEAP32[$862>>2]|0;
    $865 = $855 | $864;
    $866 = $865 | $861;
    HEAP32[$pixel73$281>>2] = $866;
    $867 = (($pixel73$281) + ($pitch72$2<<2)|0);
    $868 = ($863|0)==(0);
    if ($868) {
     break;
    } else {
     $length71$182 = $863;$pixel73$281 = $867;
    }
   }
   return;
  }
  }
 } else {
  switch ($blendMode|0) {
  case 1:  {
   $869 = (($x2) - ($x1))|0;
   $870 = ($869|0)<(0);
   $871 = (0 - ($869))|0;
   $872 = $870 ? $871 : $869;
   $873 = (($y2) - ($y1))|0;
   $874 = ($873|0)<(0);
   $875 = (0 - ($873))|0;
   $876 = $874 ? $875 : $873;
   $877 = ($872|0)<($876|0);
   if ($877) {
    $881 = $872 << 1;
    $882 = (($881) - ($876))|0;
    $883 = (($872) - ($876))|0;
    $d$0 = $882;$dinc1$0 = $881;$dinc2$0$in = $883;$numpixels$0$in = $876;$xinc1$0 = 0;$yinc1$0 = 1;
   } else {
    $878 = $876 << 1;
    $879 = (($878) - ($872))|0;
    $880 = (($876) - ($872))|0;
    $d$0 = $879;$dinc1$0 = $878;$dinc2$0$in = $880;$numpixels$0$in = $872;$xinc1$0 = 1;$yinc1$0 = 0;
   }
   $dinc2$0 = $dinc2$0$in << 1;
   $884 = ($x2|0)<($x1|0);
   $885 = (0 - ($xinc1$0))|0;
   $$xinc1$0 = $884 ? $885 : $xinc1$0;
   $$25 = $884 ? -1 : 1;
   $886 = ($y2|0)<($y1|0);
   $887 = (0 - ($yinc1$0))|0;
   $yinc1$1 = $886 ? $887 : $yinc1$0;
   $yinc2$1 = $886 ? -1 : 1;
   $not$30 = ($draw_end|0)!=(0);
   $numpixels$0 = $not$30&1;
   $numpixels$0$in$numpixels$0 = (($numpixels$0$in) + ($numpixels$0))|0;
   $888 = ($numpixels$0$in$numpixels$0|0)>(0);
   if (!($888)) {
    return;
   }
   $889 = ((($dst)) + 20|0);
   $890 = ((($dst)) + 16|0);
   $891 = ((($1)) + 12|0);
   $892 = ((($1)) + 32|0);
   $893 = ((($1)) + 28|0);
   $894 = ((($1)) + 16|0);
   $895 = ((($1)) + 20|0);
   $896 = ((($1)) + 24|0);
   $$pre150 = HEAP32[$889>>2]|0;
   $$pre151 = HEAP32[$892>>2]|0;
   $$pre152 = HEAP32[$893>>2]|0;
   $$phi$trans$insert153 = $$pre152 & 255;
   $$phi$trans$insert154 = (125924 + ($$phi$trans$insert153<<2)|0);
   $$pre155 = HEAP32[$$phi$trans$insert154>>2]|0;
   $897 = $$pre151 >>> 8;
   $898 = $897 & 255;
   $899 = $$pre152 >>> 8;
   $900 = $899 & 255;
   $$phi$trans$insert157 = (125924 + ($900<<2)|0);
   $$pre158 = HEAP32[$$phi$trans$insert157>>2]|0;
   $901 = $$pre151 >>> 16;
   $902 = $$pre152 >>> 16;
   $$phi$trans$insert159 = $902 & 255;
   $$phi$trans$insert160 = (125924 + ($$phi$trans$insert159<<2)|0);
   $$pre161 = HEAP32[$$phi$trans$insert160>>2]|0;
   $903 = $$pre151 & 255;
   $904 = $901 & 255;
   $d$199 = $d$0;$i$098 = 0;$x$0100 = $x1;$y$0101 = $y1;
   while(1) {
    $905 = HEAP32[$890>>2]|0;
    $906 = Math_imul($905, $y$0101)|0;
    $907 = (($$pre150) + ($906)|0);
    $908 = $x$0100 << 2;
    $909 = (($907) + ($908)|0);
    $910 = HEAP32[$909>>2]|0;
    $911 = HEAP32[$891>>2]|0;
    $912 = $911 & $910;
    $913 = $912 >>> $903;
    $914 = (($$pre155) + ($913)|0);
    $915 = HEAP8[$914>>0]|0;
    $916 = $915&255;
    $917 = HEAP32[$894>>2]|0;
    $918 = $917 & $910;
    $919 = $918 >>> $898;
    $920 = (($$pre158) + ($919)|0);
    $921 = HEAP8[$920>>0]|0;
    $922 = $921&255;
    $923 = HEAP32[$895>>2]|0;
    $924 = $923 & $910;
    $925 = $924 >>> $904;
    $926 = (($$pre161) + ($925)|0);
    $927 = HEAP8[$926>>0]|0;
    $928 = $927&255;
    $929 = Math_imul($916, $16)|0;
    $930 = (($929>>>0) / 255)&-1;
    $931 = (($930) + ($r$0))|0;
    $932 = Math_imul($922, $16)|0;
    $933 = (($932>>>0) / 255)&-1;
    $934 = (($933) + ($g$0))|0;
    $935 = Math_imul($928, $16)|0;
    $936 = (($935>>>0) / 255)&-1;
    $937 = (($936) + ($b$0))|0;
    $938 = $931 >>> $$phi$trans$insert153;
    $939 = $938 << $903;
    $940 = $934 >>> $900;
    $941 = $940 << $898;
    $942 = $941 | $939;
    $943 = $937 >>> $$phi$trans$insert159;
    $944 = $943 << $904;
    $945 = HEAP32[$896>>2]|0;
    $946 = $942 | $945;
    $947 = $946 | $944;
    HEAP32[$909>>2] = $947;
    $948 = ($d$199|0)<(0);
    $dinc1$0$dinc2$0 = $948 ? $dinc1$0 : $dinc2$0;
    $$xinc1$0$$25 = $948 ? $$xinc1$0 : $$25;
    $yinc1$1$yinc2$1 = $948 ? $yinc1$1 : $yinc2$1;
    $y$1 = (($y$0101) + ($yinc1$1$yinc2$1))|0;
    $x$1 = (($$xinc1$0$$25) + ($x$0100))|0;
    $d$2 = (($dinc1$0$dinc2$0) + ($d$199))|0;
    $949 = (($i$098) + 1)|0;
    $950 = ($949|0)<($numpixels$0$in$numpixels$0|0);
    if ($950) {
     $d$199 = $d$2;$i$098 = $949;$x$0100 = $x$1;$y$0101 = $y$1;
    } else {
     break;
    }
   }
   return;
   break;
  }
  case 2:  {
   $951 = (($x2) - ($x1))|0;
   $952 = ($951|0)<(0);
   $953 = (0 - ($951))|0;
   $954 = $952 ? $953 : $951;
   $955 = (($y2) - ($y1))|0;
   $956 = ($955|0)<(0);
   $957 = (0 - ($955))|0;
   $958 = $956 ? $957 : $955;
   $959 = ($954|0)<($958|0);
   if ($959) {
    $963 = $954 << 1;
    $964 = (($963) - ($958))|0;
    $965 = (($954) - ($958))|0;
    $d87$0 = $964;$dinc188$0 = $963;$dinc289$0$in = $965;$numpixels86$0$in = $958;$xinc191$0 = 0;$yinc194$0 = 1;
   } else {
    $960 = $958 << 1;
    $961 = (($960) - ($954))|0;
    $962 = (($958) - ($954))|0;
    $d87$0 = $961;$dinc188$0 = $960;$dinc289$0$in = $962;$numpixels86$0$in = $954;$xinc191$0 = 1;$yinc194$0 = 0;
   }
   $dinc289$0 = $dinc289$0$in << 1;
   $966 = ($x2|0)<($x1|0);
   $967 = (0 - ($xinc191$0))|0;
   $$xinc191$0 = $966 ? $967 : $xinc191$0;
   $$26 = $966 ? -1 : 1;
   $968 = ($y2|0)<($y1|0);
   $969 = (0 - ($yinc194$0))|0;
   $yinc194$1 = $968 ? $969 : $yinc194$0;
   $yinc295$1 = $968 ? -1 : 1;
   $not$29 = ($draw_end|0)!=(0);
   $numpixels86$0 = $not$29&1;
   $numpixels86$0$in$numpixels86$0 = (($numpixels86$0$in) + ($numpixels86$0))|0;
   $970 = ($numpixels86$0$in$numpixels86$0|0)>(0);
   if (!($970)) {
    return;
   }
   $971 = ((($dst)) + 20|0);
   $972 = ((($dst)) + 16|0);
   $973 = ((($1)) + 12|0);
   $974 = ((($1)) + 32|0);
   $975 = ((($1)) + 28|0);
   $976 = ((($1)) + 16|0);
   $977 = ((($1)) + 20|0);
   $978 = ((($1)) + 24|0);
   $$pre138 = HEAP32[$971>>2]|0;
   $$pre139 = HEAP32[$974>>2]|0;
   $$pre140 = HEAP32[$975>>2]|0;
   $$phi$trans$insert141 = $$pre140 & 255;
   $$phi$trans$insert142 = (125924 + ($$phi$trans$insert141<<2)|0);
   $$pre143 = HEAP32[$$phi$trans$insert142>>2]|0;
   $979 = $$pre139 >>> 8;
   $980 = $979 & 255;
   $981 = $$pre140 >>> 8;
   $982 = $981 & 255;
   $$phi$trans$insert145 = (125924 + ($982<<2)|0);
   $$pre146 = HEAP32[$$phi$trans$insert145>>2]|0;
   $983 = $$pre139 >>> 16;
   $984 = $$pre140 >>> 16;
   $$phi$trans$insert147 = $984 & 255;
   $$phi$trans$insert148 = (125924 + ($$phi$trans$insert147<<2)|0);
   $$pre149 = HEAP32[$$phi$trans$insert148>>2]|0;
   $985 = $$pre139 & 255;
   $986 = $983 & 255;
   $d87$1104 = $d87$0;$i83$0103 = 0;$x90$0105 = $x1;$y93$0106 = $y1;
   while(1) {
    $987 = HEAP32[$972>>2]|0;
    $988 = Math_imul($987, $y93$0106)|0;
    $989 = (($$pre138) + ($988)|0);
    $990 = $x90$0105 << 2;
    $991 = (($989) + ($990)|0);
    $992 = HEAP32[$991>>2]|0;
    $993 = HEAP32[$973>>2]|0;
    $994 = $993 & $992;
    $995 = $994 >>> $985;
    $996 = (($$pre143) + ($995)|0);
    $997 = HEAP8[$996>>0]|0;
    $998 = $997&255;
    $999 = HEAP32[$976>>2]|0;
    $1000 = $999 & $992;
    $1001 = $1000 >>> $980;
    $1002 = (($$pre146) + ($1001)|0);
    $1003 = HEAP8[$1002>>0]|0;
    $1004 = $1003&255;
    $1005 = HEAP32[$977>>2]|0;
    $1006 = $1005 & $992;
    $1007 = $1006 >>> $986;
    $1008 = (($$pre149) + ($1007)|0);
    $1009 = HEAP8[$1008>>0]|0;
    $1010 = $1009&255;
    $1011 = (($998) + ($r$0))|0;
    $1012 = ($1011>>>0)>(255);
    $$11 = $1012 ? 255 : $1011;
    $1013 = (($1004) + ($g$0))|0;
    $1014 = ($1013>>>0)>(255);
    $$7 = $1014 ? 255 : $1013;
    $1015 = (($1010) + ($b$0))|0;
    $1016 = ($1015>>>0)>(255);
    $$12 = $1016 ? 255 : $1015;
    $1017 = $$11 >>> $$phi$trans$insert141;
    $1018 = $1017 << $985;
    $1019 = $$7 >>> $982;
    $1020 = $1019 << $980;
    $1021 = $1020 | $1018;
    $1022 = $$12 >>> $$phi$trans$insert147;
    $1023 = $1022 << $986;
    $1024 = HEAP32[$978>>2]|0;
    $1025 = $1021 | $1024;
    $1026 = $1025 | $1023;
    HEAP32[$991>>2] = $1026;
    $1027 = ($d87$1104|0)<(0);
    $dinc188$0$dinc289$0 = $1027 ? $dinc188$0 : $dinc289$0;
    $$xinc191$0$$26 = $1027 ? $$xinc191$0 : $$26;
    $yinc194$1$yinc295$1 = $1027 ? $yinc194$1 : $yinc295$1;
    $y93$1 = (($y93$0106) + ($yinc194$1$yinc295$1))|0;
    $x90$1 = (($$xinc191$0$$26) + ($x90$0105))|0;
    $d87$2 = (($dinc188$0$dinc289$0) + ($d87$1104))|0;
    $1028 = (($i83$0103) + 1)|0;
    $1029 = ($1028|0)<($numpixels86$0$in$numpixels86$0|0);
    if ($1029) {
     $d87$1104 = $d87$2;$i83$0103 = $1028;$x90$0105 = $x90$1;$y93$0106 = $y93$1;
    } else {
     break;
    }
   }
   return;
   break;
  }
  case 4:  {
   $1030 = (($x2) - ($x1))|0;
   $1031 = ($1030|0)<(0);
   $1032 = (0 - ($1030))|0;
   $1033 = $1031 ? $1032 : $1030;
   $1034 = (($y2) - ($y1))|0;
   $1035 = ($1034|0)<(0);
   $1036 = (0 - ($1034))|0;
   $1037 = $1035 ? $1036 : $1034;
   $1038 = ($1033|0)<($1037|0);
   if ($1038) {
    $1042 = $1033 << 1;
    $1043 = (($1042) - ($1037))|0;
    $1044 = (($1033) - ($1037))|0;
    $d105$0 = $1043;$dinc1106$0 = $1042;$dinc2107$0$in = $1044;$numpixels104$0$in = $1037;$xinc1109$0 = 0;$yinc1112$0 = 1;
   } else {
    $1039 = $1037 << 1;
    $1040 = (($1039) - ($1033))|0;
    $1041 = (($1037) - ($1033))|0;
    $d105$0 = $1040;$dinc1106$0 = $1039;$dinc2107$0$in = $1041;$numpixels104$0$in = $1033;$xinc1109$0 = 1;$yinc1112$0 = 0;
   }
   $dinc2107$0 = $dinc2107$0$in << 1;
   $1045 = ($x2|0)<($x1|0);
   $1046 = (0 - ($xinc1109$0))|0;
   $$xinc1109$0 = $1045 ? $1046 : $xinc1109$0;
   $$27 = $1045 ? -1 : 1;
   $1047 = ($y2|0)<($y1|0);
   $1048 = (0 - ($yinc1112$0))|0;
   $yinc1112$1 = $1047 ? $1048 : $yinc1112$0;
   $yinc2113$1 = $1047 ? -1 : 1;
   $not$ = ($draw_end|0)!=(0);
   $numpixels104$0 = $not$&1;
   $numpixels104$0$in$numpixels104$0 = (($numpixels104$0$in) + ($numpixels104$0))|0;
   $1049 = ($numpixels104$0$in$numpixels104$0|0)>(0);
   if (!($1049)) {
    return;
   }
   $1050 = ((($dst)) + 20|0);
   $1051 = ((($dst)) + 16|0);
   $1052 = ((($1)) + 12|0);
   $1053 = ((($1)) + 32|0);
   $1054 = ((($1)) + 28|0);
   $1055 = ((($1)) + 16|0);
   $1056 = ((($1)) + 20|0);
   $1057 = ((($1)) + 24|0);
   $$pre = HEAP32[$1050>>2]|0;
   $$pre128 = HEAP32[$1053>>2]|0;
   $$pre129 = HEAP32[$1054>>2]|0;
   $$phi$trans$insert = $$pre129 & 255;
   $$phi$trans$insert130 = (125924 + ($$phi$trans$insert<<2)|0);
   $$pre131 = HEAP32[$$phi$trans$insert130>>2]|0;
   $1058 = $$pre128 >>> 8;
   $1059 = $1058 & 255;
   $1060 = $$pre129 >>> 8;
   $1061 = $1060 & 255;
   $$phi$trans$insert133 = (125924 + ($1061<<2)|0);
   $$pre134 = HEAP32[$$phi$trans$insert133>>2]|0;
   $1062 = $$pre128 >>> 16;
   $1063 = $$pre129 >>> 16;
   $$phi$trans$insert135 = $1063 & 255;
   $$phi$trans$insert136 = (125924 + ($$phi$trans$insert135<<2)|0);
   $$pre137 = HEAP32[$$phi$trans$insert136>>2]|0;
   $1064 = $$pre128 & 255;
   $1065 = $1062 & 255;
   $d105$1109 = $d105$0;$i101$0108 = 0;$x108$0110 = $x1;$y111$0111 = $y1;
   while(1) {
    $1066 = HEAP32[$1051>>2]|0;
    $1067 = Math_imul($1066, $y111$0111)|0;
    $1068 = (($$pre) + ($1067)|0);
    $1069 = $x108$0110 << 2;
    $1070 = (($1068) + ($1069)|0);
    $1071 = HEAP32[$1070>>2]|0;
    $1072 = HEAP32[$1052>>2]|0;
    $1073 = $1072 & $1071;
    $1074 = $1073 >>> $1064;
    $1075 = (($$pre131) + ($1074)|0);
    $1076 = HEAP8[$1075>>0]|0;
    $1077 = $1076&255;
    $1078 = HEAP32[$1055>>2]|0;
    $1079 = $1078 & $1071;
    $1080 = $1079 >>> $1059;
    $1081 = (($$pre134) + ($1080)|0);
    $1082 = HEAP8[$1081>>0]|0;
    $1083 = $1082&255;
    $1084 = HEAP32[$1056>>2]|0;
    $1085 = $1084 & $1071;
    $1086 = $1085 >>> $1065;
    $1087 = (($$pre137) + ($1086)|0);
    $1088 = HEAP8[$1087>>0]|0;
    $1089 = $1088&255;
    $1090 = Math_imul($1077, $r$0)|0;
    $1091 = (($1090>>>0) / 255)&-1;
    $1092 = Math_imul($1083, $g$0)|0;
    $1093 = (($1092>>>0) / 255)&-1;
    $1094 = Math_imul($1089, $b$0)|0;
    $1095 = (($1094>>>0) / 255)&-1;
    $1096 = $1091 >>> $$phi$trans$insert;
    $1097 = $1096 << $1064;
    $1098 = $1093 >>> $1061;
    $1099 = $1098 << $1059;
    $1100 = $1099 | $1097;
    $1101 = $1095 >>> $$phi$trans$insert135;
    $1102 = $1101 << $1065;
    $1103 = HEAP32[$1057>>2]|0;
    $1104 = $1100 | $1103;
    $1105 = $1104 | $1102;
    HEAP32[$1070>>2] = $1105;
    $1106 = ($d105$1109|0)<(0);
    $dinc1106$0$dinc2107$0 = $1106 ? $dinc1106$0 : $dinc2107$0;
    $$xinc1109$0$$27 = $1106 ? $$xinc1109$0 : $$27;
    $yinc1112$1$yinc2113$1 = $1106 ? $yinc1112$1 : $yinc2113$1;
    $y111$1 = (($y111$0111) + ($yinc1112$1$yinc2113$1))|0;
    $x108$1 = (($$xinc1109$0$$27) + ($x108$0110))|0;
    $d105$2 = (($dinc1106$0$dinc2107$0) + ($d105$1109))|0;
    $1107 = (($i101$0108) + 1)|0;
    $1108 = ($1107|0)<($numpixels104$0$in$numpixels104$0|0);
    if ($1108) {
     $d105$1109 = $d105$2;$i101$0108 = $1107;$x108$0110 = $x108$1;$y111$0111 = $y111$1;
    } else {
     break;
    }
   }
   return;
   break;
  }
  default: {
   $1109 = (($x2) - ($x1))|0;
   $1110 = ($1109|0)<(0);
   $1111 = (0 - ($1109))|0;
   $1112 = $1110 ? $1111 : $1109;
   $1113 = (($y2) - ($y1))|0;
   $1114 = ($1113|0)<(0);
   $1115 = (0 - ($1113))|0;
   $1116 = $1114 ? $1115 : $1113;
   $1117 = ($1112|0)<($1116|0);
   if ($1117) {
    $1121 = $1112 << 1;
    $1122 = (($1121) - ($1116))|0;
    $1123 = (($1112) - ($1116))|0;
    $d123$0 = $1122;$dinc1124$0 = $1121;$dinc2125$0$in = $1123;$numpixels122$0$in = $1116;$xinc1127$0 = 0;$yinc1130$0 = 1;
   } else {
    $1118 = $1116 << 1;
    $1119 = (($1118) - ($1112))|0;
    $1120 = (($1116) - ($1112))|0;
    $d123$0 = $1119;$dinc1124$0 = $1118;$dinc2125$0$in = $1120;$numpixels122$0$in = $1112;$xinc1127$0 = 1;$yinc1130$0 = 0;
   }
   $dinc2125$0 = $dinc2125$0$in << 1;
   $1124 = ($x2|0)<($x1|0);
   $1125 = (0 - ($xinc1127$0))|0;
   $$xinc1127$0 = $1124 ? $1125 : $xinc1127$0;
   $$28 = $1124 ? -1 : 1;
   $1126 = ($y2|0)<($y1|0);
   $1127 = (0 - ($yinc1130$0))|0;
   $yinc1130$1 = $1126 ? $1127 : $yinc1130$0;
   $yinc2131$1 = $1126 ? -1 : 1;
   $not$31 = ($draw_end|0)!=(0);
   $numpixels122$0 = $not$31&1;
   $numpixels122$0$in$numpixels122$0 = (($numpixels122$0$in) + ($numpixels122$0))|0;
   $1128 = ($numpixels122$0$in$numpixels122$0|0)>(0);
   if (!($1128)) {
    return;
   }
   $1129 = ((($dst)) + 20|0);
   $1130 = HEAP32[$1129>>2]|0;
   $1131 = ((($dst)) + 16|0);
   $1132 = ((($1)) + 28|0);
   $1133 = HEAP32[$1132>>2]|0;
   $1134 = $1133 & 255;
   $1135 = $r$0 >>> $1134;
   $1136 = ((($1)) + 32|0);
   $1137 = HEAP32[$1136>>2]|0;
   $1138 = $1137 & 255;
   $1139 = $1135 << $1138;
   $1140 = $1133 >>> 8;
   $1141 = $1140 & 255;
   $1142 = $g$0 >>> $1141;
   $1143 = $1137 >>> 8;
   $1144 = $1143 & 255;
   $1145 = $1142 << $1144;
   $1146 = $1145 | $1139;
   $1147 = $1133 >>> 16;
   $1148 = $1147 & 255;
   $1149 = $b$0 >>> $1148;
   $1150 = $1137 >>> 16;
   $1151 = $1150 & 255;
   $1152 = $1149 << $1151;
   $1153 = ((($1)) + 24|0);
   $d123$194 = $d123$0;$i119$093 = 0;$x126$095 = $x1;$y129$096 = $y1;
   while(1) {
    $1154 = HEAP32[$1131>>2]|0;
    $1155 = Math_imul($1154, $y129$096)|0;
    $1156 = (($1130) + ($1155)|0);
    $1157 = $x126$095 << 2;
    $1158 = (($1156) + ($1157)|0);
    $1159 = HEAP32[$1153>>2]|0;
    $1160 = $1146 | $1159;
    $1161 = $1160 | $1152;
    HEAP32[$1158>>2] = $1161;
    $1162 = ($d123$194|0)<(0);
    $dinc1124$0$dinc2125$0 = $1162 ? $dinc1124$0 : $dinc2125$0;
    $$xinc1127$0$$28 = $1162 ? $$xinc1127$0 : $$28;
    $yinc1130$1$yinc2131$1 = $1162 ? $yinc1130$1 : $yinc2131$1;
    $y129$1 = (($y129$096) + ($yinc1130$1$yinc2131$1))|0;
    $x126$1 = (($$xinc1127$0$$28) + ($x126$095))|0;
    $d123$2 = (($dinc1124$0$dinc2125$0) + ($d123$194))|0;
    $1163 = (($i119$093) + 1)|0;
    $1164 = ($1163|0)<($numpixels122$0$in$numpixels122$0|0);
    if ($1164) {
     $d123$194 = $d123$2;$i119$093 = $1163;$x126$095 = $x126$1;$y129$096 = $y129$1;
    } else {
     break;
    }
   }
   return;
  }
  }
 }
}
function _SDL_BlendPoint($dst,$x,$y,$blendMode,$r,$g,$b,$a) {
 $dst = $dst|0;
 $x = $x|0;
 $y = $y|0;
 $blendMode = $blendMode|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $$0 = 0, $$04 = 0, $$05 = 0, $$06 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $blendMode$off = 0, $cond = 0, $cond1 = 0, $cond2 = 0, $vararg_buffer = 0;
 var $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($dst|0)==(0|0);
 if ($0) {
  $1 = (_SDL_SetError(603923,$vararg_buffer)|0);
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ((($dst)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 8|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ($5&255)<(8);
 if ($6) {
  $7 = (_SDL_SetError(603844,$vararg_buffer1)|0);
  $$0 = $7;
  STACKTOP = sp;return ($$0|0);
 }
 $8 = ((($dst)) + 36|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)>($x|0);
 if ($10) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $11 = ((($dst)) + 40|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)>($y|0);
 if ($13) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $14 = ((($dst)) + 44|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = (($15) + ($9))|0;
 $17 = ($16|0)>($x|0);
 if (!($17)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $18 = ((($dst)) + 48|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (($19) + ($12))|0;
 $21 = ($20|0)>($y|0);
 if (!($21)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $blendMode$off = (($blendMode) + -1)|0;
 $22 = ($blendMode$off>>>0)<(2);
 if ($22) {
  $23 = $r&255;
  $24 = $a&255;
  $25 = Math_imul($24, $23)|0;
  $26 = (($25>>>0) / 255)&-1;
  $27 = $26&255;
  $28 = $g&255;
  $29 = Math_imul($24, $28)|0;
  $30 = (($29>>>0) / 255)&-1;
  $31 = $30&255;
  $32 = $b&255;
  $33 = Math_imul($24, $32)|0;
  $34 = (($33>>>0) / 255)&-1;
  $35 = $34&255;
  $$04 = $35;$$05 = $31;$$06 = $27;
 } else {
  $$04 = $b;$$05 = $g;$$06 = $r;
 }
 $36 = $5&255;
 switch ($36|0) {
 case 15:  {
  $37 = ((($3)) + 12|0);
  $38 = HEAP32[$37>>2]|0;
  $cond2 = ($38|0)==(31744);
  if ($cond2) {
   (_SDL_BlendPoint_RGB555($dst,$x,$y,$blendMode,$$06,$$05,$$04,$a)|0);
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
  break;
 }
 case 16:  {
  $39 = ((($3)) + 12|0);
  $40 = HEAP32[$39>>2]|0;
  $cond1 = ($40|0)==(63488);
  if ($cond1) {
   (_SDL_BlendPoint_RGB565($dst,$x,$y,$blendMode,$$06,$$05,$$04,$a)|0);
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
  break;
 }
 case 32:  {
  $41 = ((($3)) + 12|0);
  $42 = HEAP32[$41>>2]|0;
  $cond = ($42|0)==(16711680);
  if ($cond) {
   $43 = ((($3)) + 24|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = ($44|0)==(0);
   if ($45) {
    (_SDL_BlendPoint_RGB888($dst,$x,$y,$blendMode,$$06,$$05,$$04,$a)|0);
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   } else {
    (_SDL_BlendPoint_ARGB8888($dst,$x,$y,$blendMode,$$06,$$05,$$04,$a)|0);
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
  }
  break;
 }
 default: {
 }
 }
 $46 = ((($3)) + 24|0);
 $47 = HEAP32[$46>>2]|0;
 $48 = ($47|0)==(0);
 if ($48) {
  $49 = (_SDL_BlendPoint_RGB($dst,$x,$y,$blendMode,$$06,$$05,$$04,$a)|0);
  $$0 = $49;
  STACKTOP = sp;return ($$0|0);
 } else {
  $50 = (_SDL_BlendPoint_RGBA($dst,$x,$y,$blendMode,$$06,$$05,$$04,$a)|0);
  $$0 = $50;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SW_RenderFillRects($renderer,$rects,$count) {
 $renderer = $renderer|0;
 $rects = $rects|0;
 $count = $count|0;
 var $$ = 0, $$0 = 0, $$1 = 0, $$alloca_mul = 0, $$pr = 0, $$pre = 0, $$pre8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0;
 var $39 = 0.0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0, $47 = 0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0.0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0, $62 = 0, $63 = 0, $64 = 0.0, $65 = 0, $66 = 0, $67 = 0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0.0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $exitcond = 0, $exitcond7 = 0, $i$04 = 0, $i$13 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $4 = ((($1)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  HEAP32[$1>>2] = $5;
  $6 = ($5|0)==(0);
  $7 = $5;
  if ($6) {
   $8 = ((($renderer)) + 188|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (_SDL_GetWindowSurface($9)|0);
   $11 = ($10|0)==(0|0);
   do {
    if (!($11)) {
     HEAP32[$4>>2] = $10;
     HEAP32[$1>>2] = $10;
     $12 = HEAP32[$0>>2]|0;
     $13 = HEAP32[$12>>2]|0;
     $14 = ($13|0)==(0|0);
     if (!($14)) {
      $15 = ((($renderer)) + 212|0);
      (_SDL_SetClipRect($13,$15)|0);
      $$pre = HEAP32[$0>>2]|0;
      $$pre8 = HEAP32[$$pre>>2]|0;
      $16 = ($$pre8|0)==(0|0);
      if (!($16)) {
       $17 = ((($renderer)) + 276|0);
       $18 = HEAP32[$17>>2]|0;
       $19 = ($18|0)==(0);
       if ($19) {
        (_SDL_SetClipRect($$pre8,0)|0);
        break;
       } else {
        $20 = ((($renderer)) + 244|0);
        (_SDL_SetClipRect($$pre8,$20)|0);
        break;
       }
      }
     }
    }
   } while(0);
   $$pr = HEAP32[$1>>2]|0;
   $22 = $$pr;
  } else {
   $22 = $7;
  }
  $21 = ($22|0)==(0|0);
  if ($21) {
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  } else {
   $93 = $22;
  }
 } else {
  $93 = $2;
 }
 $23 = $count << 4;
 $$alloca_mul = $23;
 $24 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 $25 = ((($renderer)) + 212|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ($26|0)==(0);
 if ($27) {
  $28 = ((($renderer)) + 216|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = ($29|0)==(0);
  if ($30) {
   $31 = ($count|0)>(0);
   if ($31) {
    $i$13 = 0;
    while(1) {
     $59 = (($rects) + ($i$13<<4)|0);
     $60 = +HEAPF32[$59>>2];
     $61 = (~~(($60)));
     $62 = (($24) + ($i$13<<4)|0);
     HEAP32[$62>>2] = $61;
     $63 = (((($rects) + ($i$13<<4)|0)) + 4|0);
     $64 = +HEAPF32[$63>>2];
     $65 = (~~(($64)));
     $66 = (((($24) + ($i$13<<4)|0)) + 4|0);
     HEAP32[$66>>2] = $65;
     $67 = (((($rects) + ($i$13<<4)|0)) + 8|0);
     $68 = +HEAPF32[$67>>2];
     $69 = (~~(($68)));
     $70 = ($69|0)>(1);
     $$1 = $70 ? $69 : 1;
     $71 = (((($24) + ($i$13<<4)|0)) + 8|0);
     HEAP32[$71>>2] = $$1;
     $72 = (((($rects) + ($i$13<<4)|0)) + 12|0);
     $73 = +HEAPF32[$72>>2];
     $74 = (~~(($73)));
     $75 = ($74|0)>(1);
     $76 = $75 ? $74 : 1;
     $77 = (((($24) + ($i$13<<4)|0)) + 12|0);
     HEAP32[$77>>2] = $76;
     $78 = (($i$13) + 1)|0;
     $exitcond = ($78|0)==($count|0);
     if ($exitcond) {
      break;
     } else {
      $i$13 = $78;
     }
    }
   }
  } else {
   label = 14;
  }
 } else {
  label = 14;
 }
 if ((label|0) == 14) {
  $32 = ($count|0)>(0);
  if ($32) {
   $33 = ((($renderer)) + 216|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = (+($26|0));
   $36 = (+($34|0));
   $i$04 = 0;
   while(1) {
    $37 = (($rects) + ($i$04<<4)|0);
    $38 = +HEAPF32[$37>>2];
    $39 = $35 + $38;
    $40 = (~~(($39)));
    $41 = (($24) + ($i$04<<4)|0);
    HEAP32[$41>>2] = $40;
    $42 = (((($rects) + ($i$04<<4)|0)) + 4|0);
    $43 = +HEAPF32[$42>>2];
    $44 = $36 + $43;
    $45 = (~~(($44)));
    $46 = (((($24) + ($i$04<<4)|0)) + 4|0);
    HEAP32[$46>>2] = $45;
    $47 = (((($rects) + ($i$04<<4)|0)) + 8|0);
    $48 = +HEAPF32[$47>>2];
    $49 = (~~(($48)));
    $50 = ($49|0)>(1);
    $$ = $50 ? $49 : 1;
    $51 = (((($24) + ($i$04<<4)|0)) + 8|0);
    HEAP32[$51>>2] = $$;
    $52 = (((($rects) + ($i$04<<4)|0)) + 12|0);
    $53 = +HEAPF32[$52>>2];
    $54 = (~~(($53)));
    $55 = ($54|0)>(1);
    $56 = $55 ? $54 : 1;
    $57 = (((($24) + ($i$04<<4)|0)) + 12|0);
    HEAP32[$57>>2] = $56;
    $58 = (($i$04) + 1)|0;
    $exitcond7 = ($58|0)==($count|0);
    if ($exitcond7) {
     break;
    } else {
     $i$04 = $58;
    }
   }
  }
 }
 $79 = ((($renderer)) + 312|0);
 $80 = HEAP32[$79>>2]|0;
 $81 = ($80|0)==(0);
 $82 = ((($renderer)) + 308|0);
 $83 = HEAP32[$82>>2]|0;
 $84 = $83&65535;
 $85 = $83&255;
 $86 = ($84&65535) >>> 8;
 $87 = $86&255;
 $88 = $83 >>> 16;
 $89 = $88&255;
 $90 = $83 >>> 24;
 $91 = $90&255;
 if ($81) {
  $92 = ((($93)) + 4|0);
  $94 = HEAP32[$92>>2]|0;
  $95 = (_SDL_MapRGBA($94,$85,$87,$89,$91)|0);
  $96 = (_SDL_FillRects($93,$24,$count,$95)|0);
  $$0 = $96;
  STACKTOP = sp;return ($$0|0);
 } else {
  $97 = (_SDL_BlendFillRects($93,$24,$count,$80,$85,$87,$89,$91)|0);
  $$0 = $97;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_FillRects($dst,$rects,$count,$color) {
 $dst = $dst|0;
 $rects = $rects|0;
 $count = $count|0;
 $color = $color|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $exitcond = 0, $i$01 = 0, $status$02 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($rects|0)==(0|0);
 if ($0) {
  $2 = (_SDL_SetError(603889,$vararg_buffer)|0);
  $$0 = $2;
 } else {
  $1 = ($count|0)>(0);
  if ($1) {
   $i$01 = 0;$status$02 = 0;
   while(1) {
    $3 = (($rects) + ($i$01<<4)|0);
    $4 = (_SDL_FillRect($dst,$3,$color)|0);
    $5 = (($4) + ($status$02))|0;
    $6 = (($i$01) + 1)|0;
    $exitcond = ($6|0)==($count|0);
    if ($exitcond) {
     $$0 = $5;
     break;
    } else {
     $i$01 = $6;$status$02 = $5;
    }
   }
  } else {
   $$0 = 0;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _SDL_BlendFillRects($dst,$rects,$count,$blendMode,$r,$g,$b,$a) {
 $dst = $dst|0;
 $rects = $rects|0;
 $count = $count|0;
 $blendMode = $blendMode|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $$0 = 0, $$04 = 0, $$05 = 0, $$06 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $SDL_BlendFillRect_RGB$SDL_BlendFillRect_RGBA = 0, $SDL_BlendFillRect_RGB888$SDL_BlendFillRect_ARGB8888 = 0, $blendMode$off = 0, $cond = 0, $cond1 = 0, $cond2 = 0, $exitcond = 0, $func$1 = 0, $i$010 = 0, $rect = 0, $status$011 = 0, $status$1 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $rect = sp + 16|0;
 $0 = ($dst|0)==(0|0);
 if ($0) {
  $1 = (_SDL_SetError(603923,$vararg_buffer)|0);
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ((($dst)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 8|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ($5&255)<(8);
 if ($6) {
  $7 = (_SDL_SetError(603955,$vararg_buffer1)|0);
  $$0 = $7;
  STACKTOP = sp;return ($$0|0);
 }
 $blendMode$off = (($blendMode) + -1)|0;
 $8 = ($blendMode$off>>>0)<(2);
 if ($8) {
  $9 = $r&255;
  $10 = $a&255;
  $11 = Math_imul($10, $9)|0;
  $12 = (($11>>>0) / 255)&-1;
  $13 = $12&255;
  $14 = $g&255;
  $15 = Math_imul($10, $14)|0;
  $16 = (($15>>>0) / 255)&-1;
  $17 = $16&255;
  $18 = $b&255;
  $19 = Math_imul($10, $18)|0;
  $20 = (($19>>>0) / 255)&-1;
  $21 = $20&255;
  $$04 = $21;$$05 = $17;$$06 = $13;
 } else {
  $$04 = $b;$$05 = $g;$$06 = $r;
 }
 $22 = $5&255;
 switch ($22|0) {
 case 15:  {
  $23 = ((($3)) + 12|0);
  $24 = HEAP32[$23>>2]|0;
  $cond2 = ($24|0)==(31744);
  if ($cond2) {
   $func$1 = 4562;
  } else {
   label = 12;
  }
  break;
 }
 case 16:  {
  $25 = ((($3)) + 12|0);
  $26 = HEAP32[$25>>2]|0;
  $cond1 = ($26|0)==(63488);
  if ($cond1) {
   $func$1 = 4563;
  } else {
   label = 12;
  }
  break;
 }
 case 32:  {
  $27 = ((($3)) + 12|0);
  $28 = HEAP32[$27>>2]|0;
  $cond = ($28|0)==(16711680);
  if ($cond) {
   $29 = ((($3)) + 24|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = ($30|0)==(0);
   $SDL_BlendFillRect_RGB888$SDL_BlendFillRect_ARGB8888 = $31 ? 4558 : 4559;
   $func$1 = $SDL_BlendFillRect_RGB888$SDL_BlendFillRect_ARGB8888;
  } else {
   label = 12;
  }
  break;
 }
 default: {
  label = 12;
 }
 }
 if ((label|0) == 12) {
  $32 = ((($3)) + 24|0);
  $33 = HEAP32[$32>>2]|0;
  $34 = ($33|0)==(0);
  $SDL_BlendFillRect_RGB$SDL_BlendFillRect_RGBA = $34 ? 4560 : 4561;
  $func$1 = $SDL_BlendFillRect_RGB$SDL_BlendFillRect_RGBA;
 }
 $35 = ($count|0)>(0);
 if (!($35)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $36 = ((($dst)) + 36|0);
 $i$010 = 0;$status$011 = 0;
 while(1) {
  $37 = (($rects) + ($i$010<<4)|0);
  $38 = (_SDL_IntersectRect($37,$36,$rect)|0);
  $39 = ($38|0)==(0);
  if ($39) {
   $status$1 = $status$011;
  } else {
   $40 = (FUNCTION_TABLE_iiiiiiii[$func$1 & 8191]($dst,$rect,$blendMode,$$06,$$05,$$04,$a)|0);
   $status$1 = $40;
  }
  $41 = (($i$010) + 1)|0;
  $exitcond = ($41|0)==($count|0);
  if ($exitcond) {
   $$0 = $status$1;
   break;
  } else {
   $i$010 = $41;$status$011 = $status$1;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _SDL_BlendFillRect_RGB888($dst,$rect,$blendMode,$r,$g,$b,$a) {
 $dst = $dst|0;
 $rect = $rect|0;
 $blendMode = $blendMode|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $$ = 0, $$1 = 0, $$10 = 0, $$11 = 0, $$2 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$6 = 0, $$7 = 0, $$8 = 0, $$9 = 0, $$in = 0, $$in46 = 0, $$in47 = 0, $$lcssa = 0, $$lcssa73 = 0, $$lcssa74 = 0, $$lcssa99 = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0;
 var $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0;
 var $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0;
 var $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0;
 var $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0;
 var $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0;
 var $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0;
 var $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0;
 var $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n18$0 = 0, $n18$1 = 0, $n18$2 = 0, $n18$3 = 0, $n40$0 = 0, $n40$1 = 0, $n40$2 = 0, $n40$3 = 0, $n62$0$us = 0, $n62$1$us2578 = 0, $n62$2$us3883 = 0, $n62$375 = 0, $pixel$016 = 0, $pixel$1 = 0, $pixel$2 = 0;
 var $pixel$3 = 0, $pixel$4 = 0, $pixel17$018 = 0, $pixel17$1 = 0, $pixel17$2 = 0, $pixel17$3 = 0, $pixel17$4 = 0, $pixel39$020 = 0, $pixel39$1 = 0, $pixel39$2 = 0, $pixel39$3 = 0, $pixel39$4 = 0, $pixel61$015 = 0, $pixel61$015$us = 0, $pixel61$015$us22 = 0, $pixel61$015$us33 = 0, $pixel61$1$us = 0, $pixel61$2$us2679 = 0, $pixel61$3$us3984 = 0, $pixel61$476 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $a&255;
 $1 = $0 ^ 255;
 switch ($blendMode|0) {
 case 1:  {
  $2 = ((($rect)) + 8|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($rect)) + 12|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ((($dst)) + 16|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ((($dst)) + 4|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($9)) + 9|0);
  $11 = HEAP8[$10>>0]|0;
  $12 = $11&255;
  $13 = (($7|0) / ($12|0))&-1;
  $14 = (($13) - ($3))|0;
  $15 = ($5|0)==(0);
  if ($15) {
   return 0;
  }
  $16 = ((($dst)) + 20|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = ((($rect)) + 4|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = Math_imul($19, $13)|0;
  $21 = (($17) + ($20<<2)|0);
  $22 = HEAP32[$rect>>2]|0;
  $23 = (($21) + ($22<<2)|0);
  $24 = (($3) + 3)|0;
  $25 = (($24|0) / 4)&-1;
  $26 = $3 & 3;
  $27 = $r&255;
  $28 = $g&255;
  $29 = $b&255;
  $$in47 = $5;$pixel$016 = $23;
  L41: while(1) {
   $30 = (($$in47) + -1)|0;
   switch ($26|0) {
   case 0:  {
    $n$0 = $25;$pixel$1 = $pixel$016;
    label = 5;
    break;
   }
   case 3:  {
    $n$1 = $25;$pixel$2 = $pixel$016;
    label = 6;
    break;
   }
   case 2:  {
    $n$2 = $25;$pixel$3 = $pixel$016;
    label = 7;
    break;
   }
   case 1:  {
    $n$3 = $25;$pixel$4 = $pixel$016;
    label = 8;
    break;
   }
   default: {
    label = 9;
    break L41;
   }
   }
   while(1) {
    if ((label|0) == 5) {
     label = 0;
     $31 = HEAP32[$pixel$1>>2]|0;
     $32 = $31 >>> 16;
     $33 = $32 & 255;
     $34 = $31 >>> 8;
     $35 = $34 & 255;
     $36 = $31 & 255;
     $37 = Math_imul($33, $1)|0;
     $38 = (($37>>>0) / 255)&-1;
     $39 = (($38) + ($27))|0;
     $40 = Math_imul($35, $1)|0;
     $41 = (($40>>>0) / 255)&-1;
     $42 = (($41) + ($28))|0;
     $43 = Math_imul($36, $1)|0;
     $44 = (($43>>>0) / 255)&-1;
     $45 = (($44) + ($29))|0;
     $46 = $39 << 16;
     $47 = $42 << 8;
     $48 = $47 | $45;
     $49 = $48 | $46;
     HEAP32[$pixel$1>>2] = $49;
     $50 = ((($pixel$1)) + 4|0);
     $n$1 = $n$0;$pixel$2 = $50;
     label = 6;
     continue;
    }
    else if ((label|0) == 6) {
     label = 0;
     $51 = HEAP32[$pixel$2>>2]|0;
     $52 = $51 >>> 16;
     $53 = $52 & 255;
     $54 = $51 >>> 8;
     $55 = $54 & 255;
     $56 = $51 & 255;
     $57 = Math_imul($53, $1)|0;
     $58 = (($57>>>0) / 255)&-1;
     $59 = (($58) + ($27))|0;
     $60 = Math_imul($55, $1)|0;
     $61 = (($60>>>0) / 255)&-1;
     $62 = (($61) + ($28))|0;
     $63 = Math_imul($56, $1)|0;
     $64 = (($63>>>0) / 255)&-1;
     $65 = (($64) + ($29))|0;
     $66 = $59 << 16;
     $67 = $62 << 8;
     $68 = $67 | $65;
     $69 = $68 | $66;
     HEAP32[$pixel$2>>2] = $69;
     $70 = ((($pixel$2)) + 4|0);
     $n$2 = $n$1;$pixel$3 = $70;
     label = 7;
     continue;
    }
    else if ((label|0) == 7) {
     label = 0;
     $71 = HEAP32[$pixel$3>>2]|0;
     $72 = $71 >>> 16;
     $73 = $72 & 255;
     $74 = $71 >>> 8;
     $75 = $74 & 255;
     $76 = $71 & 255;
     $77 = Math_imul($73, $1)|0;
     $78 = (($77>>>0) / 255)&-1;
     $79 = (($78) + ($27))|0;
     $80 = Math_imul($75, $1)|0;
     $81 = (($80>>>0) / 255)&-1;
     $82 = (($81) + ($28))|0;
     $83 = Math_imul($76, $1)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = (($84) + ($29))|0;
     $86 = $79 << 16;
     $87 = $82 << 8;
     $88 = $87 | $85;
     $89 = $88 | $86;
     HEAP32[$pixel$3>>2] = $89;
     $90 = ((($pixel$3)) + 4|0);
     $n$3 = $n$2;$pixel$4 = $90;
     label = 8;
     continue;
    }
    else if ((label|0) == 8) {
     label = 0;
     $91 = HEAP32[$pixel$4>>2]|0;
     $92 = $91 >>> 16;
     $93 = $92 & 255;
     $94 = $91 >>> 8;
     $95 = $94 & 255;
     $96 = $91 & 255;
     $97 = Math_imul($93, $1)|0;
     $98 = (($97>>>0) / 255)&-1;
     $99 = (($98) + ($27))|0;
     $100 = Math_imul($95, $1)|0;
     $101 = (($100>>>0) / 255)&-1;
     $102 = (($101) + ($28))|0;
     $103 = Math_imul($96, $1)|0;
     $104 = (($103>>>0) / 255)&-1;
     $105 = (($104) + ($29))|0;
     $106 = $99 << 16;
     $107 = $102 << 8;
     $108 = $107 | $105;
     $109 = $108 | $106;
     HEAP32[$pixel$4>>2] = $109;
     $110 = ((($pixel$4)) + 4|0);
     $111 = (($n$3) + -1)|0;
     $112 = ($n$3|0)>(1);
     if ($112) {
      $n$0 = $111;$pixel$1 = $110;
      label = 5;
      continue;
     } else {
      break;
     }
    }
   }
   $113 = (($110) + ($14<<2)|0);
   $114 = ($30|0)==(0);
   if ($114) {
    label = 46;
    break;
   } else {
    $$in47 = $30;$pixel$016 = $113;
   }
  }
  if ((label|0) == 9) {
   // unreachable;
  }
  else if ((label|0) == 46) {
   return 0;
  }
  break;
 }
 case 2:  {
  $115 = ((($rect)) + 8|0);
  $116 = HEAP32[$115>>2]|0;
  $117 = ((($rect)) + 12|0);
  $118 = HEAP32[$117>>2]|0;
  $119 = ((($dst)) + 16|0);
  $120 = HEAP32[$119>>2]|0;
  $121 = ((($dst)) + 4|0);
  $122 = HEAP32[$121>>2]|0;
  $123 = ((($122)) + 9|0);
  $124 = HEAP8[$123>>0]|0;
  $125 = $124&255;
  $126 = (($120|0) / ($125|0))&-1;
  $127 = (($126) - ($116))|0;
  $128 = ($118|0)==(0);
  if ($128) {
   return 0;
  }
  $129 = ((($dst)) + 20|0);
  $130 = HEAP32[$129>>2]|0;
  $131 = ((($rect)) + 4|0);
  $132 = HEAP32[$131>>2]|0;
  $133 = Math_imul($132, $126)|0;
  $134 = (($130) + ($133<<2)|0);
  $135 = HEAP32[$rect>>2]|0;
  $136 = (($134) + ($135<<2)|0);
  $137 = (($116) + 3)|0;
  $138 = (($137|0) / 4)&-1;
  $139 = $116 & 3;
  $140 = $r&255;
  $141 = $g&255;
  $142 = $b&255;
  $$in46 = $118;$pixel17$018 = $136;
  L57: while(1) {
   $143 = (($$in46) + -1)|0;
   switch ($139|0) {
   case 0:  {
    $n18$0 = $138;$pixel17$1 = $pixel17$018;
    label = 14;
    break;
   }
   case 3:  {
    $n18$1 = $138;$pixel17$2 = $pixel17$018;
    label = 15;
    break;
   }
   case 2:  {
    $n18$2 = $138;$pixel17$3 = $pixel17$018;
    label = 16;
    break;
   }
   case 1:  {
    $n18$3 = $138;$pixel17$4 = $pixel17$018;
    label = 17;
    break;
   }
   default: {
    label = 18;
    break L57;
   }
   }
   while(1) {
    if ((label|0) == 14) {
     label = 0;
     $144 = HEAP32[$pixel17$1>>2]|0;
     $145 = $144 >>> 16;
     $146 = $145 & 255;
     $147 = $144 >>> 8;
     $148 = $147 & 255;
     $149 = $144 & 255;
     $150 = (($146) + ($140))|0;
     $151 = ($150>>>0)>(255);
     $$ = $151 ? 255 : $150;
     $152 = (($148) + ($141))|0;
     $153 = ($152>>>0)>(255);
     $$8 = $153 ? 255 : $152;
     $154 = (($149) + ($142))|0;
     $155 = ($154>>>0)>(255);
     $$1 = $155 ? 255 : $154;
     $156 = $$ << 16;
     $157 = $$8 << 8;
     $158 = $157 | $$1;
     $159 = $158 | $156;
     HEAP32[$pixel17$1>>2] = $159;
     $160 = ((($pixel17$1)) + 4|0);
     $n18$1 = $n18$0;$pixel17$2 = $160;
     label = 15;
     continue;
    }
    else if ((label|0) == 15) {
     label = 0;
     $161 = HEAP32[$pixel17$2>>2]|0;
     $162 = $161 >>> 16;
     $163 = $162 & 255;
     $164 = $161 >>> 8;
     $165 = $164 & 255;
     $166 = $161 & 255;
     $167 = (($163) + ($140))|0;
     $168 = ($167>>>0)>(255);
     $$2 = $168 ? 255 : $167;
     $169 = (($165) + ($141))|0;
     $170 = ($169>>>0)>(255);
     $$9 = $170 ? 255 : $169;
     $171 = (($166) + ($142))|0;
     $172 = ($171>>>0)>(255);
     $$3 = $172 ? 255 : $171;
     $173 = $$2 << 16;
     $174 = $$9 << 8;
     $175 = $174 | $$3;
     $176 = $175 | $173;
     HEAP32[$pixel17$2>>2] = $176;
     $177 = ((($pixel17$2)) + 4|0);
     $n18$2 = $n18$1;$pixel17$3 = $177;
     label = 16;
     continue;
    }
    else if ((label|0) == 16) {
     label = 0;
     $178 = HEAP32[$pixel17$3>>2]|0;
     $179 = $178 >>> 16;
     $180 = $179 & 255;
     $181 = $178 >>> 8;
     $182 = $181 & 255;
     $183 = $178 & 255;
     $184 = (($180) + ($140))|0;
     $185 = ($184>>>0)>(255);
     $$4 = $185 ? 255 : $184;
     $186 = (($182) + ($141))|0;
     $187 = ($186>>>0)>(255);
     $$10 = $187 ? 255 : $186;
     $188 = (($183) + ($142))|0;
     $189 = ($188>>>0)>(255);
     $$5 = $189 ? 255 : $188;
     $190 = $$4 << 16;
     $191 = $$10 << 8;
     $192 = $191 | $$5;
     $193 = $192 | $190;
     HEAP32[$pixel17$3>>2] = $193;
     $194 = ((($pixel17$3)) + 4|0);
     $n18$3 = $n18$2;$pixel17$4 = $194;
     label = 17;
     continue;
    }
    else if ((label|0) == 17) {
     label = 0;
     $195 = HEAP32[$pixel17$4>>2]|0;
     $196 = $195 >>> 16;
     $197 = $196 & 255;
     $198 = $195 >>> 8;
     $199 = $198 & 255;
     $200 = $195 & 255;
     $201 = (($197) + ($140))|0;
     $202 = ($201>>>0)>(255);
     $$6 = $202 ? 255 : $201;
     $203 = (($199) + ($141))|0;
     $204 = ($203>>>0)>(255);
     $$11 = $204 ? 255 : $203;
     $205 = (($200) + ($142))|0;
     $206 = ($205>>>0)>(255);
     $$7 = $206 ? 255 : $205;
     $207 = $$6 << 16;
     $208 = $$11 << 8;
     $209 = $208 | $$7;
     $210 = $209 | $207;
     HEAP32[$pixel17$4>>2] = $210;
     $211 = ((($pixel17$4)) + 4|0);
     $212 = (($n18$3) + -1)|0;
     $213 = ($n18$3|0)>(1);
     if ($213) {
      $n18$0 = $212;$pixel17$1 = $211;
      label = 14;
      continue;
     } else {
      break;
     }
    }
   }
   $214 = (($211) + ($127<<2)|0);
   $215 = ($143|0)==(0);
   if ($215) {
    label = 46;
    break;
   } else {
    $$in46 = $143;$pixel17$018 = $214;
   }
  }
  if ((label|0) == 18) {
   // unreachable;
  }
  else if ((label|0) == 46) {
   return 0;
  }
  break;
 }
 case 4:  {
  $216 = ((($rect)) + 8|0);
  $217 = HEAP32[$216>>2]|0;
  $218 = ((($rect)) + 12|0);
  $219 = HEAP32[$218>>2]|0;
  $220 = ((($dst)) + 16|0);
  $221 = HEAP32[$220>>2]|0;
  $222 = ((($dst)) + 4|0);
  $223 = HEAP32[$222>>2]|0;
  $224 = ((($223)) + 9|0);
  $225 = HEAP8[$224>>0]|0;
  $226 = $225&255;
  $227 = (($221|0) / ($226|0))&-1;
  $228 = (($227) - ($217))|0;
  $229 = ($219|0)==(0);
  if ($229) {
   return 0;
  }
  $230 = ((($dst)) + 20|0);
  $231 = HEAP32[$230>>2]|0;
  $232 = ((($rect)) + 4|0);
  $233 = HEAP32[$232>>2]|0;
  $234 = Math_imul($233, $227)|0;
  $235 = (($231) + ($234<<2)|0);
  $236 = HEAP32[$rect>>2]|0;
  $237 = (($235) + ($236<<2)|0);
  $238 = (($217) + 3)|0;
  $239 = (($238|0) / 4)&-1;
  $240 = $217 & 3;
  $241 = $r&255;
  $242 = $g&255;
  $243 = $b&255;
  $$in = $219;$pixel39$020 = $237;
  L73: while(1) {
   $244 = (($$in) + -1)|0;
   switch ($240|0) {
   case 0:  {
    $n40$0 = $239;$pixel39$1 = $pixel39$020;
    label = 23;
    break;
   }
   case 3:  {
    $n40$1 = $239;$pixel39$2 = $pixel39$020;
    label = 24;
    break;
   }
   case 2:  {
    $n40$2 = $239;$pixel39$3 = $pixel39$020;
    label = 25;
    break;
   }
   case 1:  {
    $n40$3 = $239;$pixel39$4 = $pixel39$020;
    label = 26;
    break;
   }
   default: {
    label = 27;
    break L73;
   }
   }
   while(1) {
    if ((label|0) == 23) {
     label = 0;
     $245 = HEAP32[$pixel39$1>>2]|0;
     $246 = $245 >>> 16;
     $247 = $246 & 255;
     $248 = $245 >>> 8;
     $249 = $248 & 255;
     $250 = $245 & 255;
     $251 = Math_imul($247, $241)|0;
     $252 = (($251>>>0) / 255)&-1;
     $253 = Math_imul($249, $242)|0;
     $254 = (($253>>>0) / 255)&-1;
     $255 = Math_imul($250, $243)|0;
     $256 = (($255>>>0) / 255)&-1;
     $257 = $252 << 16;
     $258 = $254 << 8;
     $259 = $258 | $256;
     $260 = $259 | $257;
     HEAP32[$pixel39$1>>2] = $260;
     $261 = ((($pixel39$1)) + 4|0);
     $n40$1 = $n40$0;$pixel39$2 = $261;
     label = 24;
     continue;
    }
    else if ((label|0) == 24) {
     label = 0;
     $262 = HEAP32[$pixel39$2>>2]|0;
     $263 = $262 >>> 16;
     $264 = $263 & 255;
     $265 = $262 >>> 8;
     $266 = $265 & 255;
     $267 = $262 & 255;
     $268 = Math_imul($264, $241)|0;
     $269 = (($268>>>0) / 255)&-1;
     $270 = Math_imul($266, $242)|0;
     $271 = (($270>>>0) / 255)&-1;
     $272 = Math_imul($267, $243)|0;
     $273 = (($272>>>0) / 255)&-1;
     $274 = $269 << 16;
     $275 = $271 << 8;
     $276 = $275 | $273;
     $277 = $276 | $274;
     HEAP32[$pixel39$2>>2] = $277;
     $278 = ((($pixel39$2)) + 4|0);
     $n40$2 = $n40$1;$pixel39$3 = $278;
     label = 25;
     continue;
    }
    else if ((label|0) == 25) {
     label = 0;
     $279 = HEAP32[$pixel39$3>>2]|0;
     $280 = $279 >>> 16;
     $281 = $280 & 255;
     $282 = $279 >>> 8;
     $283 = $282 & 255;
     $284 = $279 & 255;
     $285 = Math_imul($281, $241)|0;
     $286 = (($285>>>0) / 255)&-1;
     $287 = Math_imul($283, $242)|0;
     $288 = (($287>>>0) / 255)&-1;
     $289 = Math_imul($284, $243)|0;
     $290 = (($289>>>0) / 255)&-1;
     $291 = $286 << 16;
     $292 = $288 << 8;
     $293 = $292 | $290;
     $294 = $293 | $291;
     HEAP32[$pixel39$3>>2] = $294;
     $295 = ((($pixel39$3)) + 4|0);
     $n40$3 = $n40$2;$pixel39$4 = $295;
     label = 26;
     continue;
    }
    else if ((label|0) == 26) {
     label = 0;
     $296 = HEAP32[$pixel39$4>>2]|0;
     $297 = $296 >>> 16;
     $298 = $297 & 255;
     $299 = $296 >>> 8;
     $300 = $299 & 255;
     $301 = $296 & 255;
     $302 = Math_imul($298, $241)|0;
     $303 = (($302>>>0) / 255)&-1;
     $304 = Math_imul($300, $242)|0;
     $305 = (($304>>>0) / 255)&-1;
     $306 = Math_imul($301, $243)|0;
     $307 = (($306>>>0) / 255)&-1;
     $308 = $303 << 16;
     $309 = $305 << 8;
     $310 = $309 | $307;
     $311 = $310 | $308;
     HEAP32[$pixel39$4>>2] = $311;
     $312 = ((($pixel39$4)) + 4|0);
     $313 = (($n40$3) + -1)|0;
     $314 = ($n40$3|0)>(1);
     if ($314) {
      $n40$0 = $313;$pixel39$1 = $312;
      label = 23;
      continue;
     } else {
      break;
     }
    }
   }
   $315 = (($312) + ($228<<2)|0);
   $316 = ($244|0)==(0);
   if ($316) {
    label = 46;
    break;
   } else {
    $$in = $244;$pixel39$020 = $315;
   }
  }
  if ((label|0) == 27) {
   // unreachable;
  }
  else if ((label|0) == 46) {
   return 0;
  }
  break;
 }
 default: {
  $317 = ((($rect)) + 8|0);
  $318 = HEAP32[$317>>2]|0;
  $319 = ((($rect)) + 12|0);
  $320 = HEAP32[$319>>2]|0;
  $321 = ((($dst)) + 16|0);
  $322 = HEAP32[$321>>2]|0;
  $323 = ((($dst)) + 4|0);
  $324 = HEAP32[$323>>2]|0;
  $325 = ((($324)) + 9|0);
  $326 = HEAP8[$325>>0]|0;
  $327 = $326&255;
  $328 = (($322|0) / ($327|0))&-1;
  $329 = (($328) - ($318))|0;
  $330 = ((($dst)) + 20|0);
  $331 = HEAP32[$330>>2]|0;
  $332 = ((($rect)) + 4|0);
  $333 = HEAP32[$332>>2]|0;
  $334 = Math_imul($333, $328)|0;
  $335 = (($331) + ($334<<2)|0);
  $336 = HEAP32[$rect>>2]|0;
  $337 = (($335) + ($336<<2)|0);
  $338 = (($320) + -1)|0;
  $339 = ($320|0)==(0);
  if ($339) {
   return 0;
  }
  $340 = (($318) + 3)|0;
  $341 = (($340|0) / 4)&-1;
  $342 = $318 & 3;
  $343 = $r&255;
  $344 = $g&255;
  $345 = $b&255;
  $346 = $343 << 16;
  $347 = $344 << 8;
  $348 = $347 | $346;
  $349 = $348 | $345;
  switch ($342|0) {
  case 0:  {
   $361 = $338;$pixel61$015$us = $337;
   while(1) {
    $n62$0$us = $341;$pixel61$1$us = $pixel61$015$us;
    while(1) {
     HEAP32[$pixel61$1$us>>2] = $349;
     $353 = ((($pixel61$1$us)) + 4|0);
     HEAP32[$353>>2] = $349;
     $354 = ((($pixel61$1$us)) + 8|0);
     HEAP32[$354>>2] = $349;
     $355 = ((($pixel61$1$us)) + 12|0);
     HEAP32[$355>>2] = $349;
     $356 = ((($pixel61$1$us)) + 16|0);
     $357 = (($n62$0$us) + -1)|0;
     $358 = ($n62$0$us|0)>(1);
     if ($358) {
      $n62$0$us = $357;$pixel61$1$us = $356;
     } else {
      $$lcssa99 = $356;
      break;
     }
    }
    $359 = (($$lcssa99) + ($329<<2)|0);
    $360 = (($361) + -1)|0;
    $362 = ($361|0)==(0);
    if ($362) {
     break;
    } else {
     $361 = $360;$pixel61$015$us = $359;
    }
   }
   return 0;
   break;
  }
  case 3:  {
   $351 = ($340|0)>(7);
   $375 = $338;$pixel61$015$us22 = $337;
   while(1) {
    HEAP32[$pixel61$015$us22>>2] = $349;
    $363 = ((($pixel61$015$us22)) + 4|0);
    HEAP32[$363>>2] = $349;
    $364 = ((($pixel61$015$us22)) + 8|0);
    HEAP32[$364>>2] = $349;
    $365 = ((($pixel61$015$us22)) + 12|0);
    if ($351) {
     $367 = $365;$n62$1$us2578 = $341;$pixel61$2$us2679 = $pixel61$015$us22;
     while(1) {
      $366 = (($n62$1$us2578) + -1)|0;
      HEAP32[$367>>2] = $349;
      $368 = ((($pixel61$2$us2679)) + 16|0);
      HEAP32[$368>>2] = $349;
      $369 = ((($pixel61$2$us2679)) + 20|0);
      HEAP32[$369>>2] = $349;
      $370 = ((($pixel61$2$us2679)) + 24|0);
      HEAP32[$370>>2] = $349;
      $371 = ((($pixel61$2$us2679)) + 28|0);
      $372 = ($366|0)>(1);
      if ($372) {
       $367 = $371;$n62$1$us2578 = $366;$pixel61$2$us2679 = $368;
      } else {
       $$lcssa73 = $371;
       break;
      }
     }
    } else {
     $$lcssa73 = $365;
    }
    $373 = (($$lcssa73) + ($329<<2)|0);
    $374 = (($375) + -1)|0;
    $376 = ($375|0)==(0);
    if ($376) {
     break;
    } else {
     $375 = $374;$pixel61$015$us22 = $373;
    }
   }
   return 0;
   break;
  }
  case 2:  {
   $350 = ($340|0)>(7);
   $388 = $338;$pixel61$015$us33 = $337;
   while(1) {
    HEAP32[$pixel61$015$us33>>2] = $349;
    $377 = ((($pixel61$015$us33)) + 4|0);
    HEAP32[$377>>2] = $349;
    $378 = ((($pixel61$015$us33)) + 8|0);
    if ($350) {
     $380 = $378;$n62$2$us3883 = $341;$pixel61$3$us3984 = $pixel61$015$us33;
     while(1) {
      $379 = (($n62$2$us3883) + -1)|0;
      HEAP32[$380>>2] = $349;
      $381 = ((($pixel61$3$us3984)) + 12|0);
      HEAP32[$381>>2] = $349;
      $382 = ((($pixel61$3$us3984)) + 16|0);
      HEAP32[$382>>2] = $349;
      $383 = ((($pixel61$3$us3984)) + 20|0);
      HEAP32[$383>>2] = $349;
      $384 = ((($pixel61$3$us3984)) + 24|0);
      $385 = ($379|0)>(1);
      if ($385) {
       $380 = $384;$n62$2$us3883 = $379;$pixel61$3$us3984 = $382;
      } else {
       $$lcssa74 = $384;
       break;
      }
     }
    } else {
     $$lcssa74 = $378;
    }
    $386 = (($$lcssa74) + ($329<<2)|0);
    $387 = (($388) + -1)|0;
    $389 = ($388|0)==(0);
    if ($389) {
     break;
    } else {
     $388 = $387;$pixel61$015$us33 = $386;
    }
   }
   return 0;
   break;
  }
  default: {
   $352 = ($340|0)>(7);
   $400 = $338;$pixel61$015 = $337;
   while(1) {
    HEAP32[$pixel61$015>>2] = $349;
    $390 = ((($pixel61$015)) + 4|0);
    if ($352) {
     $392 = $390;$n62$375 = $341;$pixel61$476 = $pixel61$015;
     while(1) {
      $391 = (($n62$375) + -1)|0;
      HEAP32[$392>>2] = $349;
      $393 = ((($pixel61$476)) + 8|0);
      HEAP32[$393>>2] = $349;
      $394 = ((($pixel61$476)) + 12|0);
      HEAP32[$394>>2] = $349;
      $395 = ((($pixel61$476)) + 16|0);
      HEAP32[$395>>2] = $349;
      $396 = ((($pixel61$476)) + 20|0);
      $397 = ($391|0)>(1);
      if ($397) {
       $392 = $396;$n62$375 = $391;$pixel61$476 = $395;
      } else {
       $$lcssa = $396;
       break;
      }
     }
    } else {
     $$lcssa = $390;
    }
    $398 = (($$lcssa) + ($329<<2)|0);
    $399 = (($400) + -1)|0;
    $401 = ($400|0)==(0);
    if ($401) {
     break;
    } else {
     $400 = $399;$pixel61$015 = $398;
    }
   }
   return 0;
  }
  }
 }
 }
 return (0)|0;
}
function _SDL_BlendFillRect_ARGB8888($dst,$rect,$blendMode,$r,$g,$b,$a) {
 $dst = $dst|0;
 $rect = $rect|0;
 $blendMode = $blendMode|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $$ = 0, $$1 = 0, $$10 = 0, $$11 = 0, $$2 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$6 = 0, $$7 = 0, $$8 = 0, $$9 = 0, $$in = 0, $$in50 = 0, $$in51 = 0, $$lcssa = 0, $$lcssa103 = 0, $$lcssa77 = 0, $$lcssa78 = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0;
 var $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0;
 var $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0;
 var $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0;
 var $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0;
 var $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0;
 var $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0;
 var $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0;
 var $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0;
 var $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0;
 var $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0;
 var $441 = 0, $442 = 0, $443 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n18$0 = 0, $n18$1 = 0, $n18$2 = 0, $n18$3 = 0, $n40$0 = 0, $n40$1 = 0, $n40$2 = 0, $n40$3 = 0, $n62$0$us = 0, $n62$1$us3083 = 0, $n62$2$us4388 = 0, $n62$380 = 0, $pixel$016 = 0;
 var $pixel$1 = 0, $pixel$2 = 0, $pixel$3 = 0, $pixel$4 = 0, $pixel17$020 = 0, $pixel17$1 = 0, $pixel17$2 = 0, $pixel17$3 = 0, $pixel17$4 = 0, $pixel39$024 = 0, $pixel39$1 = 0, $pixel39$2 = 0, $pixel39$3 = 0, $pixel39$4 = 0, $pixel61$015 = 0, $pixel61$015$us = 0, $pixel61$015$us26 = 0, $pixel61$015$us37 = 0, $pixel61$1$us = 0, $pixel61$2$us2982 = 0;
 var $pixel61$3$us4287 = 0, $pixel61$479 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $a&255;
 $1 = $0 ^ 255;
 switch ($blendMode|0) {
 case 1:  {
  $2 = ((($rect)) + 8|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($rect)) + 12|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ((($dst)) + 16|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ((($dst)) + 4|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($9)) + 9|0);
  $11 = HEAP8[$10>>0]|0;
  $12 = $11&255;
  $13 = (($7|0) / ($12|0))&-1;
  $14 = (($13) - ($3))|0;
  $15 = ($5|0)==(0);
  if ($15) {
   return 0;
  }
  $16 = ((($dst)) + 20|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = ((($rect)) + 4|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = Math_imul($19, $13)|0;
  $21 = (($17) + ($20<<2)|0);
  $22 = HEAP32[$rect>>2]|0;
  $23 = (($21) + ($22<<2)|0);
  $24 = (($3) + 3)|0;
  $25 = (($24|0) / 4)&-1;
  $26 = $3 & 3;
  $27 = $r&255;
  $28 = $g&255;
  $29 = $b&255;
  $$in51 = $5;$pixel$016 = $23;
  L41: while(1) {
   $30 = (($$in51) + -1)|0;
   switch ($26|0) {
   case 0:  {
    $n$0 = $25;$pixel$1 = $pixel$016;
    label = 5;
    break;
   }
   case 3:  {
    $n$1 = $25;$pixel$2 = $pixel$016;
    label = 6;
    break;
   }
   case 2:  {
    $n$2 = $25;$pixel$3 = $pixel$016;
    label = 7;
    break;
   }
   case 1:  {
    $n$3 = $25;$pixel$4 = $pixel$016;
    label = 8;
    break;
   }
   default: {
    label = 9;
    break L41;
   }
   }
   while(1) {
    if ((label|0) == 5) {
     label = 0;
     $31 = HEAP32[$pixel$1>>2]|0;
     $32 = $31 >>> 16;
     $33 = $32 & 255;
     $34 = $31 >>> 8;
     $35 = $34 & 255;
     $36 = $31 & 255;
     $37 = $31 >>> 24;
     $38 = Math_imul($33, $1)|0;
     $39 = (($38>>>0) / 255)&-1;
     $40 = (($39) + ($27))|0;
     $41 = Math_imul($35, $1)|0;
     $42 = (($41>>>0) / 255)&-1;
     $43 = (($42) + ($28))|0;
     $44 = Math_imul($36, $1)|0;
     $45 = (($44>>>0) / 255)&-1;
     $46 = (($45) + ($29))|0;
     $47 = Math_imul($37, $1)|0;
     $48 = (($47>>>0) / 255)&-1;
     $49 = (($48) + ($0))|0;
     $50 = $49 << 24;
     $51 = $40 << 16;
     $52 = $43 << 8;
     $53 = $50 | $46;
     $54 = $53 | $51;
     $55 = $54 | $52;
     HEAP32[$pixel$1>>2] = $55;
     $56 = ((($pixel$1)) + 4|0);
     $n$1 = $n$0;$pixel$2 = $56;
     label = 6;
     continue;
    }
    else if ((label|0) == 6) {
     label = 0;
     $57 = HEAP32[$pixel$2>>2]|0;
     $58 = $57 >>> 16;
     $59 = $58 & 255;
     $60 = $57 >>> 8;
     $61 = $60 & 255;
     $62 = $57 & 255;
     $63 = $57 >>> 24;
     $64 = Math_imul($59, $1)|0;
     $65 = (($64>>>0) / 255)&-1;
     $66 = (($65) + ($27))|0;
     $67 = Math_imul($61, $1)|0;
     $68 = (($67>>>0) / 255)&-1;
     $69 = (($68) + ($28))|0;
     $70 = Math_imul($62, $1)|0;
     $71 = (($70>>>0) / 255)&-1;
     $72 = (($71) + ($29))|0;
     $73 = Math_imul($63, $1)|0;
     $74 = (($73>>>0) / 255)&-1;
     $75 = (($74) + ($0))|0;
     $76 = $75 << 24;
     $77 = $66 << 16;
     $78 = $69 << 8;
     $79 = $76 | $72;
     $80 = $79 | $77;
     $81 = $80 | $78;
     HEAP32[$pixel$2>>2] = $81;
     $82 = ((($pixel$2)) + 4|0);
     $n$2 = $n$1;$pixel$3 = $82;
     label = 7;
     continue;
    }
    else if ((label|0) == 7) {
     label = 0;
     $83 = HEAP32[$pixel$3>>2]|0;
     $84 = $83 >>> 16;
     $85 = $84 & 255;
     $86 = $83 >>> 8;
     $87 = $86 & 255;
     $88 = $83 & 255;
     $89 = $83 >>> 24;
     $90 = Math_imul($85, $1)|0;
     $91 = (($90>>>0) / 255)&-1;
     $92 = (($91) + ($27))|0;
     $93 = Math_imul($87, $1)|0;
     $94 = (($93>>>0) / 255)&-1;
     $95 = (($94) + ($28))|0;
     $96 = Math_imul($88, $1)|0;
     $97 = (($96>>>0) / 255)&-1;
     $98 = (($97) + ($29))|0;
     $99 = Math_imul($89, $1)|0;
     $100 = (($99>>>0) / 255)&-1;
     $101 = (($100) + ($0))|0;
     $102 = $101 << 24;
     $103 = $92 << 16;
     $104 = $95 << 8;
     $105 = $102 | $98;
     $106 = $105 | $103;
     $107 = $106 | $104;
     HEAP32[$pixel$3>>2] = $107;
     $108 = ((($pixel$3)) + 4|0);
     $n$3 = $n$2;$pixel$4 = $108;
     label = 8;
     continue;
    }
    else if ((label|0) == 8) {
     label = 0;
     $109 = HEAP32[$pixel$4>>2]|0;
     $110 = $109 >>> 16;
     $111 = $110 & 255;
     $112 = $109 >>> 8;
     $113 = $112 & 255;
     $114 = $109 & 255;
     $115 = $109 >>> 24;
     $116 = Math_imul($111, $1)|0;
     $117 = (($116>>>0) / 255)&-1;
     $118 = (($117) + ($27))|0;
     $119 = Math_imul($113, $1)|0;
     $120 = (($119>>>0) / 255)&-1;
     $121 = (($120) + ($28))|0;
     $122 = Math_imul($114, $1)|0;
     $123 = (($122>>>0) / 255)&-1;
     $124 = (($123) + ($29))|0;
     $125 = Math_imul($115, $1)|0;
     $126 = (($125>>>0) / 255)&-1;
     $127 = (($126) + ($0))|0;
     $128 = $127 << 24;
     $129 = $118 << 16;
     $130 = $121 << 8;
     $131 = $128 | $124;
     $132 = $131 | $129;
     $133 = $132 | $130;
     HEAP32[$pixel$4>>2] = $133;
     $134 = ((($pixel$4)) + 4|0);
     $135 = (($n$3) + -1)|0;
     $136 = ($n$3|0)>(1);
     if ($136) {
      $n$0 = $135;$pixel$1 = $134;
      label = 5;
      continue;
     } else {
      break;
     }
    }
   }
   $137 = (($134) + ($14<<2)|0);
   $138 = ($30|0)==(0);
   if ($138) {
    label = 46;
    break;
   } else {
    $$in51 = $30;$pixel$016 = $137;
   }
  }
  if ((label|0) == 9) {
   // unreachable;
  }
  else if ((label|0) == 46) {
   return 0;
  }
  break;
 }
 case 2:  {
  $139 = ((($rect)) + 8|0);
  $140 = HEAP32[$139>>2]|0;
  $141 = ((($rect)) + 12|0);
  $142 = HEAP32[$141>>2]|0;
  $143 = ((($dst)) + 16|0);
  $144 = HEAP32[$143>>2]|0;
  $145 = ((($dst)) + 4|0);
  $146 = HEAP32[$145>>2]|0;
  $147 = ((($146)) + 9|0);
  $148 = HEAP8[$147>>0]|0;
  $149 = $148&255;
  $150 = (($144|0) / ($149|0))&-1;
  $151 = (($150) - ($140))|0;
  $152 = ($142|0)==(0);
  if ($152) {
   return 0;
  }
  $153 = ((($dst)) + 20|0);
  $154 = HEAP32[$153>>2]|0;
  $155 = ((($rect)) + 4|0);
  $156 = HEAP32[$155>>2]|0;
  $157 = Math_imul($156, $150)|0;
  $158 = (($154) + ($157<<2)|0);
  $159 = HEAP32[$rect>>2]|0;
  $160 = (($158) + ($159<<2)|0);
  $161 = (($140) + 3)|0;
  $162 = (($161|0) / 4)&-1;
  $163 = $140 & 3;
  $164 = $r&255;
  $165 = $g&255;
  $166 = $b&255;
  $$in50 = $142;$pixel17$020 = $160;
  L57: while(1) {
   $167 = (($$in50) + -1)|0;
   switch ($163|0) {
   case 0:  {
    $n18$0 = $162;$pixel17$1 = $pixel17$020;
    label = 14;
    break;
   }
   case 3:  {
    $n18$1 = $162;$pixel17$2 = $pixel17$020;
    label = 15;
    break;
   }
   case 2:  {
    $n18$2 = $162;$pixel17$3 = $pixel17$020;
    label = 16;
    break;
   }
   case 1:  {
    $n18$3 = $162;$pixel17$4 = $pixel17$020;
    label = 17;
    break;
   }
   default: {
    label = 18;
    break L57;
   }
   }
   while(1) {
    if ((label|0) == 14) {
     label = 0;
     $168 = HEAP32[$pixel17$1>>2]|0;
     $169 = $168 >>> 16;
     $170 = $169 & 255;
     $171 = $168 >>> 8;
     $172 = $171 & 255;
     $173 = $168 & 255;
     $174 = $168 & -16777216;
     $175 = (($170) + ($164))|0;
     $176 = ($175>>>0)>(255);
     $$ = $176 ? 255 : $175;
     $177 = (($172) + ($165))|0;
     $178 = ($177>>>0)>(255);
     $$8 = $178 ? 255 : $177;
     $179 = (($173) + ($166))|0;
     $180 = ($179>>>0)>(255);
     $$1 = $180 ? 255 : $179;
     $181 = $$ << 16;
     $182 = $$8 << 8;
     $183 = $$1 | $174;
     $184 = $183 | $181;
     $185 = $184 | $182;
     HEAP32[$pixel17$1>>2] = $185;
     $186 = ((($pixel17$1)) + 4|0);
     $n18$1 = $n18$0;$pixel17$2 = $186;
     label = 15;
     continue;
    }
    else if ((label|0) == 15) {
     label = 0;
     $187 = HEAP32[$pixel17$2>>2]|0;
     $188 = $187 >>> 16;
     $189 = $188 & 255;
     $190 = $187 >>> 8;
     $191 = $190 & 255;
     $192 = $187 & 255;
     $193 = $187 & -16777216;
     $194 = (($189) + ($164))|0;
     $195 = ($194>>>0)>(255);
     $$2 = $195 ? 255 : $194;
     $196 = (($191) + ($165))|0;
     $197 = ($196>>>0)>(255);
     $$9 = $197 ? 255 : $196;
     $198 = (($192) + ($166))|0;
     $199 = ($198>>>0)>(255);
     $$3 = $199 ? 255 : $198;
     $200 = $$2 << 16;
     $201 = $$9 << 8;
     $202 = $$3 | $193;
     $203 = $202 | $200;
     $204 = $203 | $201;
     HEAP32[$pixel17$2>>2] = $204;
     $205 = ((($pixel17$2)) + 4|0);
     $n18$2 = $n18$1;$pixel17$3 = $205;
     label = 16;
     continue;
    }
    else if ((label|0) == 16) {
     label = 0;
     $206 = HEAP32[$pixel17$3>>2]|0;
     $207 = $206 >>> 16;
     $208 = $207 & 255;
     $209 = $206 >>> 8;
     $210 = $209 & 255;
     $211 = $206 & 255;
     $212 = $206 & -16777216;
     $213 = (($208) + ($164))|0;
     $214 = ($213>>>0)>(255);
     $$4 = $214 ? 255 : $213;
     $215 = (($210) + ($165))|0;
     $216 = ($215>>>0)>(255);
     $$10 = $216 ? 255 : $215;
     $217 = (($211) + ($166))|0;
     $218 = ($217>>>0)>(255);
     $$5 = $218 ? 255 : $217;
     $219 = $$4 << 16;
     $220 = $$10 << 8;
     $221 = $$5 | $212;
     $222 = $221 | $219;
     $223 = $222 | $220;
     HEAP32[$pixel17$3>>2] = $223;
     $224 = ((($pixel17$3)) + 4|0);
     $n18$3 = $n18$2;$pixel17$4 = $224;
     label = 17;
     continue;
    }
    else if ((label|0) == 17) {
     label = 0;
     $225 = HEAP32[$pixel17$4>>2]|0;
     $226 = $225 >>> 16;
     $227 = $226 & 255;
     $228 = $225 >>> 8;
     $229 = $228 & 255;
     $230 = $225 & 255;
     $231 = $225 & -16777216;
     $232 = (($227) + ($164))|0;
     $233 = ($232>>>0)>(255);
     $$6 = $233 ? 255 : $232;
     $234 = (($229) + ($165))|0;
     $235 = ($234>>>0)>(255);
     $$11 = $235 ? 255 : $234;
     $236 = (($230) + ($166))|0;
     $237 = ($236>>>0)>(255);
     $$7 = $237 ? 255 : $236;
     $238 = $$6 << 16;
     $239 = $$11 << 8;
     $240 = $$7 | $231;
     $241 = $240 | $238;
     $242 = $241 | $239;
     HEAP32[$pixel17$4>>2] = $242;
     $243 = ((($pixel17$4)) + 4|0);
     $244 = (($n18$3) + -1)|0;
     $245 = ($n18$3|0)>(1);
     if ($245) {
      $n18$0 = $244;$pixel17$1 = $243;
      label = 14;
      continue;
     } else {
      break;
     }
    }
   }
   $246 = (($243) + ($151<<2)|0);
   $247 = ($167|0)==(0);
   if ($247) {
    label = 46;
    break;
   } else {
    $$in50 = $167;$pixel17$020 = $246;
   }
  }
  if ((label|0) == 18) {
   // unreachable;
  }
  else if ((label|0) == 46) {
   return 0;
  }
  break;
 }
 case 4:  {
  $248 = ((($rect)) + 8|0);
  $249 = HEAP32[$248>>2]|0;
  $250 = ((($rect)) + 12|0);
  $251 = HEAP32[$250>>2]|0;
  $252 = ((($dst)) + 16|0);
  $253 = HEAP32[$252>>2]|0;
  $254 = ((($dst)) + 4|0);
  $255 = HEAP32[$254>>2]|0;
  $256 = ((($255)) + 9|0);
  $257 = HEAP8[$256>>0]|0;
  $258 = $257&255;
  $259 = (($253|0) / ($258|0))&-1;
  $260 = (($259) - ($249))|0;
  $261 = ($251|0)==(0);
  if ($261) {
   return 0;
  }
  $262 = ((($dst)) + 20|0);
  $263 = HEAP32[$262>>2]|0;
  $264 = ((($rect)) + 4|0);
  $265 = HEAP32[$264>>2]|0;
  $266 = Math_imul($265, $259)|0;
  $267 = (($263) + ($266<<2)|0);
  $268 = HEAP32[$rect>>2]|0;
  $269 = (($267) + ($268<<2)|0);
  $270 = (($249) + 3)|0;
  $271 = (($270|0) / 4)&-1;
  $272 = $249 & 3;
  $273 = $r&255;
  $274 = $g&255;
  $275 = $b&255;
  $$in = $251;$pixel39$024 = $269;
  L73: while(1) {
   $276 = (($$in) + -1)|0;
   switch ($272|0) {
   case 0:  {
    $n40$0 = $271;$pixel39$1 = $pixel39$024;
    label = 23;
    break;
   }
   case 3:  {
    $n40$1 = $271;$pixel39$2 = $pixel39$024;
    label = 24;
    break;
   }
   case 2:  {
    $n40$2 = $271;$pixel39$3 = $pixel39$024;
    label = 25;
    break;
   }
   case 1:  {
    $n40$3 = $271;$pixel39$4 = $pixel39$024;
    label = 26;
    break;
   }
   default: {
    label = 27;
    break L73;
   }
   }
   while(1) {
    if ((label|0) == 23) {
     label = 0;
     $277 = HEAP32[$pixel39$1>>2]|0;
     $278 = $277 >>> 16;
     $279 = $278 & 255;
     $280 = $277 >>> 8;
     $281 = $280 & 255;
     $282 = $277 & 255;
     $283 = $277 & -16777216;
     $284 = Math_imul($279, $273)|0;
     $285 = (($284>>>0) / 255)&-1;
     $286 = Math_imul($281, $274)|0;
     $287 = (($286>>>0) / 255)&-1;
     $288 = Math_imul($282, $275)|0;
     $289 = (($288>>>0) / 255)&-1;
     $290 = $285 << 16;
     $291 = $287 << 8;
     $292 = $289 | $283;
     $293 = $292 | $290;
     $294 = $293 | $291;
     HEAP32[$pixel39$1>>2] = $294;
     $295 = ((($pixel39$1)) + 4|0);
     $n40$1 = $n40$0;$pixel39$2 = $295;
     label = 24;
     continue;
    }
    else if ((label|0) == 24) {
     label = 0;
     $296 = HEAP32[$pixel39$2>>2]|0;
     $297 = $296 >>> 16;
     $298 = $297 & 255;
     $299 = $296 >>> 8;
     $300 = $299 & 255;
     $301 = $296 & 255;
     $302 = $296 & -16777216;
     $303 = Math_imul($298, $273)|0;
     $304 = (($303>>>0) / 255)&-1;
     $305 = Math_imul($300, $274)|0;
     $306 = (($305>>>0) / 255)&-1;
     $307 = Math_imul($301, $275)|0;
     $308 = (($307>>>0) / 255)&-1;
     $309 = $304 << 16;
     $310 = $306 << 8;
     $311 = $308 | $302;
     $312 = $311 | $309;
     $313 = $312 | $310;
     HEAP32[$pixel39$2>>2] = $313;
     $314 = ((($pixel39$2)) + 4|0);
     $n40$2 = $n40$1;$pixel39$3 = $314;
     label = 25;
     continue;
    }
    else if ((label|0) == 25) {
     label = 0;
     $315 = HEAP32[$pixel39$3>>2]|0;
     $316 = $315 >>> 16;
     $317 = $316 & 255;
     $318 = $315 >>> 8;
     $319 = $318 & 255;
     $320 = $315 & 255;
     $321 = $315 & -16777216;
     $322 = Math_imul($317, $273)|0;
     $323 = (($322>>>0) / 255)&-1;
     $324 = Math_imul($319, $274)|0;
     $325 = (($324>>>0) / 255)&-1;
     $326 = Math_imul($320, $275)|0;
     $327 = (($326>>>0) / 255)&-1;
     $328 = $323 << 16;
     $329 = $325 << 8;
     $330 = $327 | $321;
     $331 = $330 | $328;
     $332 = $331 | $329;
     HEAP32[$pixel39$3>>2] = $332;
     $333 = ((($pixel39$3)) + 4|0);
     $n40$3 = $n40$2;$pixel39$4 = $333;
     label = 26;
     continue;
    }
    else if ((label|0) == 26) {
     label = 0;
     $334 = HEAP32[$pixel39$4>>2]|0;
     $335 = $334 >>> 16;
     $336 = $335 & 255;
     $337 = $334 >>> 8;
     $338 = $337 & 255;
     $339 = $334 & 255;
     $340 = $334 & -16777216;
     $341 = Math_imul($336, $273)|0;
     $342 = (($341>>>0) / 255)&-1;
     $343 = Math_imul($338, $274)|0;
     $344 = (($343>>>0) / 255)&-1;
     $345 = Math_imul($339, $275)|0;
     $346 = (($345>>>0) / 255)&-1;
     $347 = $342 << 16;
     $348 = $344 << 8;
     $349 = $346 | $340;
     $350 = $349 | $347;
     $351 = $350 | $348;
     HEAP32[$pixel39$4>>2] = $351;
     $352 = ((($pixel39$4)) + 4|0);
     $353 = (($n40$3) + -1)|0;
     $354 = ($n40$3|0)>(1);
     if ($354) {
      $n40$0 = $353;$pixel39$1 = $352;
      label = 23;
      continue;
     } else {
      break;
     }
    }
   }
   $355 = (($352) + ($260<<2)|0);
   $356 = ($276|0)==(0);
   if ($356) {
    label = 46;
    break;
   } else {
    $$in = $276;$pixel39$024 = $355;
   }
  }
  if ((label|0) == 27) {
   // unreachable;
  }
  else if ((label|0) == 46) {
   return 0;
  }
  break;
 }
 default: {
  $357 = ((($rect)) + 8|0);
  $358 = HEAP32[$357>>2]|0;
  $359 = ((($rect)) + 12|0);
  $360 = HEAP32[$359>>2]|0;
  $361 = ((($dst)) + 16|0);
  $362 = HEAP32[$361>>2]|0;
  $363 = ((($dst)) + 4|0);
  $364 = HEAP32[$363>>2]|0;
  $365 = ((($364)) + 9|0);
  $366 = HEAP8[$365>>0]|0;
  $367 = $366&255;
  $368 = (($362|0) / ($367|0))&-1;
  $369 = (($368) - ($358))|0;
  $370 = ((($dst)) + 20|0);
  $371 = HEAP32[$370>>2]|0;
  $372 = ((($rect)) + 4|0);
  $373 = HEAP32[$372>>2]|0;
  $374 = Math_imul($373, $368)|0;
  $375 = (($371) + ($374<<2)|0);
  $376 = HEAP32[$rect>>2]|0;
  $377 = (($375) + ($376<<2)|0);
  $378 = (($360) + -1)|0;
  $379 = ($360|0)==(0);
  if ($379) {
   return 0;
  }
  $380 = (($358) + 3)|0;
  $381 = (($380|0) / 4)&-1;
  $382 = $358 & 3;
  $383 = $r&255;
  $384 = $g&255;
  $385 = $b&255;
  $386 = $0 << 24;
  $387 = $383 << 16;
  $388 = $384 << 8;
  $389 = $388 | $387;
  $390 = $389 | $385;
  $391 = $390 | $386;
  switch ($382|0) {
  case 0:  {
   $403 = $378;$pixel61$015$us = $377;
   while(1) {
    $n62$0$us = $381;$pixel61$1$us = $pixel61$015$us;
    while(1) {
     HEAP32[$pixel61$1$us>>2] = $391;
     $395 = ((($pixel61$1$us)) + 4|0);
     HEAP32[$395>>2] = $391;
     $396 = ((($pixel61$1$us)) + 8|0);
     HEAP32[$396>>2] = $391;
     $397 = ((($pixel61$1$us)) + 12|0);
     HEAP32[$397>>2] = $391;
     $398 = ((($pixel61$1$us)) + 16|0);
     $399 = (($n62$0$us) + -1)|0;
     $400 = ($n62$0$us|0)>(1);
     if ($400) {
      $n62$0$us = $399;$pixel61$1$us = $398;
     } else {
      $$lcssa103 = $398;
      break;
     }
    }
    $401 = (($$lcssa103) + ($369<<2)|0);
    $402 = (($403) + -1)|0;
    $404 = ($403|0)==(0);
    if ($404) {
     break;
    } else {
     $403 = $402;$pixel61$015$us = $401;
    }
   }
   return 0;
   break;
  }
  case 3:  {
   $393 = ($380|0)>(7);
   $417 = $378;$pixel61$015$us26 = $377;
   while(1) {
    HEAP32[$pixel61$015$us26>>2] = $391;
    $405 = ((($pixel61$015$us26)) + 4|0);
    HEAP32[$405>>2] = $391;
    $406 = ((($pixel61$015$us26)) + 8|0);
    HEAP32[$406>>2] = $391;
    $407 = ((($pixel61$015$us26)) + 12|0);
    if ($393) {
     $409 = $407;$n62$1$us3083 = $381;$pixel61$2$us2982 = $pixel61$015$us26;
     while(1) {
      $408 = (($n62$1$us3083) + -1)|0;
      HEAP32[$409>>2] = $391;
      $410 = ((($pixel61$2$us2982)) + 16|0);
      HEAP32[$410>>2] = $391;
      $411 = ((($pixel61$2$us2982)) + 20|0);
      HEAP32[$411>>2] = $391;
      $412 = ((($pixel61$2$us2982)) + 24|0);
      HEAP32[$412>>2] = $391;
      $413 = ((($pixel61$2$us2982)) + 28|0);
      $414 = ($408|0)>(1);
      if ($414) {
       $409 = $413;$n62$1$us3083 = $408;$pixel61$2$us2982 = $410;
      } else {
       $$lcssa77 = $413;
       break;
      }
     }
    } else {
     $$lcssa77 = $407;
    }
    $415 = (($$lcssa77) + ($369<<2)|0);
    $416 = (($417) + -1)|0;
    $418 = ($417|0)==(0);
    if ($418) {
     break;
    } else {
     $417 = $416;$pixel61$015$us26 = $415;
    }
   }
   return 0;
   break;
  }
  case 2:  {
   $392 = ($380|0)>(7);
   $430 = $378;$pixel61$015$us37 = $377;
   while(1) {
    HEAP32[$pixel61$015$us37>>2] = $391;
    $419 = ((($pixel61$015$us37)) + 4|0);
    HEAP32[$419>>2] = $391;
    $420 = ((($pixel61$015$us37)) + 8|0);
    if ($392) {
     $422 = $420;$n62$2$us4388 = $381;$pixel61$3$us4287 = $pixel61$015$us37;
     while(1) {
      $421 = (($n62$2$us4388) + -1)|0;
      HEAP32[$422>>2] = $391;
      $423 = ((($pixel61$3$us4287)) + 12|0);
      HEAP32[$423>>2] = $391;
      $424 = ((($pixel61$3$us4287)) + 16|0);
      HEAP32[$424>>2] = $391;
      $425 = ((($pixel61$3$us4287)) + 20|0);
      HEAP32[$425>>2] = $391;
      $426 = ((($pixel61$3$us4287)) + 24|0);
      $427 = ($421|0)>(1);
      if ($427) {
       $422 = $426;$n62$2$us4388 = $421;$pixel61$3$us4287 = $424;
      } else {
       $$lcssa78 = $426;
       break;
      }
     }
    } else {
     $$lcssa78 = $420;
    }
    $428 = (($$lcssa78) + ($369<<2)|0);
    $429 = (($430) + -1)|0;
    $431 = ($430|0)==(0);
    if ($431) {
     break;
    } else {
     $430 = $429;$pixel61$015$us37 = $428;
    }
   }
   return 0;
   break;
  }
  default: {
   $394 = ($380|0)>(7);
   $442 = $378;$pixel61$015 = $377;
   while(1) {
    HEAP32[$pixel61$015>>2] = $391;
    $432 = ((($pixel61$015)) + 4|0);
    if ($394) {
     $434 = $432;$n62$380 = $381;$pixel61$479 = $pixel61$015;
     while(1) {
      $433 = (($n62$380) + -1)|0;
      HEAP32[$434>>2] = $391;
      $435 = ((($pixel61$479)) + 8|0);
      HEAP32[$435>>2] = $391;
      $436 = ((($pixel61$479)) + 12|0);
      HEAP32[$436>>2] = $391;
      $437 = ((($pixel61$479)) + 16|0);
      HEAP32[$437>>2] = $391;
      $438 = ((($pixel61$479)) + 20|0);
      $439 = ($433|0)>(1);
      if ($439) {
       $434 = $438;$n62$380 = $433;$pixel61$479 = $437;
      } else {
       $$lcssa = $438;
       break;
      }
     }
    } else {
     $$lcssa = $432;
    }
    $440 = (($$lcssa) + ($369<<2)|0);
    $441 = (($442) + -1)|0;
    $443 = ($442|0)==(0);
    if ($443) {
     break;
    } else {
     $442 = $441;$pixel61$015 = $440;
    }
   }
   return 0;
  }
  }
 }
 }
 return (0)|0;
}
function _SDL_BlendFillRect_RGB($dst,$rect,$blendMode,$r,$g,$b,$a) {
 $dst = $dst|0;
 $rect = $rect|0;
 $blendMode = $blendMode|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $$ = 0, $$0 = 0, $$1 = 0, $$10 = 0, $$11 = 0, $$12 = 0, $$13 = 0, $$14 = 0, $$15 = 0, $$16 = 0, $$17 = 0, $$18 = 0, $$19 = 0, $$2 = 0, $$20 = 0, $$21 = 0, $$22 = 0, $$23 = 0, $$3 = 0, $$4 = 0;
 var $$5 = 0, $$6 = 0, $$7 = 0, $$8 = 0, $$9 = 0, $$in = 0, $$in46 = 0, $$in47 = 0, $$in48 = 0, $$in49 = 0, $$in50 = 0, $$in51 = 0, $$in52 = 0, $$phi$trans$insert = 0, $$phi$trans$insert102 = 0, $$phi$trans$insert103 = 0, $$phi$trans$insert105 = 0, $$phi$trans$insert106 = 0, $$phi$trans$insert108 = 0, $$phi$trans$insert109 = 0;
 var $$phi$trans$insert136 = 0, $$phi$trans$insert137 = 0, $$phi$trans$insert140 = 0, $$phi$trans$insert141 = 0, $$phi$trans$insert144 = 0, $$phi$trans$insert145 = 0, $$phi$trans$insert173 = 0, $$phi$trans$insert174 = 0, $$phi$trans$insert177 = 0, $$phi$trans$insert178 = 0, $$phi$trans$insert181 = 0, $$phi$trans$insert182 = 0, $$phi$trans$insert210 = 0, $$phi$trans$insert211 = 0, $$phi$trans$insert214 = 0, $$phi$trans$insert215 = 0, $$phi$trans$insert218 = 0, $$phi$trans$insert219 = 0, $$phi$trans$insert61 = 0, $$phi$trans$insert63 = 0;
 var $$phi$trans$insert64 = 0, $$phi$trans$insert66 = 0, $$phi$trans$insert67 = 0, $$phi$trans$insert81 = 0, $$phi$trans$insert82 = 0, $$phi$trans$insert84 = 0, $$phi$trans$insert85 = 0, $$phi$trans$insert87 = 0, $$phi$trans$insert88 = 0, $$pre104 = 0, $$pre107 = 0, $$pre110 = 0, $$pre118 = 0, $$pre119 = 0, $$pre130 = 0, $$pre131 = 0, $$pre133 = 0, $$pre138 = 0, $$pre139 = 0, $$pre142 = 0;
 var $$pre143 = 0, $$pre146 = 0, $$pre147 = 0, $$pre149 = 0, $$pre152 = 0, $$pre153 = 0, $$pre154 = 0, $$pre156 = 0, $$pre159 = 0, $$pre160 = 0, $$pre161 = 0, $$pre163 = 0, $$pre164 = 0, $$pre165 = 0, $$pre166 = 0, $$pre167 = 0, $$pre168 = 0, $$pre170 = 0, $$pre175 = 0, $$pre176 = 0;
 var $$pre179 = 0, $$pre180 = 0, $$pre183 = 0, $$pre184 = 0, $$pre186 = 0, $$pre189 = 0, $$pre190 = 0, $$pre191 = 0, $$pre193 = 0, $$pre196 = 0, $$pre197 = 0, $$pre198 = 0, $$pre200 = 0, $$pre201 = 0, $$pre202 = 0, $$pre203 = 0, $$pre204 = 0, $$pre205 = 0, $$pre207 = 0, $$pre212 = 0;
 var $$pre213 = 0, $$pre216 = 0, $$pre217 = 0, $$pre220 = 0, $$pre221 = 0, $$pre223 = 0, $$pre226 = 0, $$pre227 = 0, $$pre228 = 0, $$pre230 = 0, $$pre233 = 0, $$pre234 = 0, $$pre235 = 0, $$pre237 = 0, $$pre238 = 0, $$pre239 = 0, $$pre240 = 0, $$pre241 = 0, $$pre242 = 0, $$pre246 = 0;
 var $$pre250 = 0, $$pre254 = 0, $$pre256 = 0, $$pre257 = 0, $$pre258 = 0, $$pre62 = 0, $$pre65 = 0, $$pre68 = 0, $$pre76 = 0, $$pre77 = 0, $$pre83 = 0, $$pre86 = 0, $$pre89 = 0, $$pre97 = 0, $$pre98 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0;
 var $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0;
 var $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0;
 var $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0;
 var $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0;
 var $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0;
 var $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0;
 var $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0;
 var $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0;
 var $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0;
 var $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0;
 var $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0;
 var $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0;
 var $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0;
 var $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0;
 var $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0, $1267 = 0, $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0;
 var $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0, $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0, $1285 = 0, $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0;
 var $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0, $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0, $1302 = 0, $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0;
 var $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0, $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0, $1320 = 0, $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0;
 var $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0, $1330 = 0, $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0, $1339 = 0, $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0;
 var $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0, $1349 = 0, $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0, $1357 = 0, $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0;
 var $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0, $1367 = 0, $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0, $1375 = 0, $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0;
 var $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0, $1385 = 0, $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0, $1393 = 0, $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0;
 var $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0, $1402 = 0, $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0, $1410 = 0, $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0;
 var $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0, $1420 = 0, $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0, $1429 = 0, $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0;
 var $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0, $1439 = 0, $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0, $1447 = 0, $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0;
 var $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0, $1457 = 0, $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0, $1465 = 0, $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0;
 var $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0, $1475 = 0, $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0, $1483 = 0, $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0;
 var $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0, $1493 = 0, $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0, $1500 = 0, $1501 = 0, $1502 = 0, $1503 = 0, $1504 = 0, $1505 = 0;
 var $1506 = 0, $1507 = 0, $1508 = 0, $1509 = 0, $151 = 0, $1510 = 0, $1511 = 0, $1512 = 0, $1513 = 0, $1514 = 0, $1515 = 0, $1516 = 0, $1517 = 0, $1518 = 0, $1519 = 0, $152 = 0, $1520 = 0, $1521 = 0, $1522 = 0, $1523 = 0;
 var $1524 = 0, $1525 = 0, $1526 = 0, $1527 = 0, $1528 = 0, $1529 = 0, $153 = 0, $1530 = 0, $1531 = 0, $1532 = 0, $1533 = 0, $1534 = 0, $1535 = 0, $1536 = 0, $1537 = 0, $1538 = 0, $1539 = 0, $154 = 0, $1540 = 0, $1541 = 0;
 var $1542 = 0, $1543 = 0, $1544 = 0, $1545 = 0, $1546 = 0, $1547 = 0, $1548 = 0, $1549 = 0, $155 = 0, $1550 = 0, $1551 = 0, $1552 = 0, $1553 = 0, $1554 = 0, $1555 = 0, $1556 = 0, $1557 = 0, $1558 = 0, $1559 = 0, $156 = 0;
 var $1560 = 0, $1561 = 0, $1562 = 0, $1563 = 0, $1564 = 0, $1565 = 0, $1566 = 0, $1567 = 0, $1568 = 0, $1569 = 0, $157 = 0, $1570 = 0, $1571 = 0, $1572 = 0, $1573 = 0, $1574 = 0, $1575 = 0, $1576 = 0, $1577 = 0, $1578 = 0;
 var $1579 = 0, $158 = 0, $1580 = 0, $1581 = 0, $1582 = 0, $1583 = 0, $1584 = 0, $1585 = 0, $1586 = 0, $1587 = 0, $1588 = 0, $1589 = 0, $159 = 0, $1590 = 0, $1591 = 0, $1592 = 0, $1593 = 0, $1594 = 0, $1595 = 0, $1596 = 0;
 var $1597 = 0, $1598 = 0, $1599 = 0, $16 = 0, $160 = 0, $1600 = 0, $1601 = 0, $1602 = 0, $1603 = 0, $1604 = 0, $1605 = 0, $1606 = 0, $1607 = 0, $1608 = 0, $1609 = 0, $161 = 0, $1610 = 0, $1611 = 0, $1612 = 0, $1613 = 0;
 var $1614 = 0, $1615 = 0, $1616 = 0, $1617 = 0, $1618 = 0, $1619 = 0, $162 = 0, $1620 = 0, $1621 = 0, $1622 = 0, $1623 = 0, $1624 = 0, $1625 = 0, $1626 = 0, $1627 = 0, $1628 = 0, $1629 = 0, $163 = 0, $1630 = 0, $1631 = 0;
 var $1632 = 0, $1633 = 0, $1634 = 0, $1635 = 0, $1636 = 0, $1637 = 0, $1638 = 0, $1639 = 0, $164 = 0, $1640 = 0, $1641 = 0, $1642 = 0, $1643 = 0, $1644 = 0, $1645 = 0, $1646 = 0, $1647 = 0, $1648 = 0, $1649 = 0, $165 = 0;
 var $1650 = 0, $1651 = 0, $1652 = 0, $1653 = 0, $1654 = 0, $1655 = 0, $1656 = 0, $1657 = 0, $1658 = 0, $1659 = 0, $166 = 0, $1660 = 0, $1661 = 0, $1662 = 0, $1663 = 0, $1664 = 0, $1665 = 0, $1666 = 0, $1667 = 0, $1668 = 0;
 var $1669 = 0, $167 = 0, $1670 = 0, $1671 = 0, $1672 = 0, $1673 = 0, $1674 = 0, $1675 = 0, $1676 = 0, $1677 = 0, $1678 = 0, $1679 = 0, $168 = 0, $1680 = 0, $1681 = 0, $1682 = 0, $1683 = 0, $1684 = 0, $1685 = 0, $1686 = 0;
 var $1687 = 0, $1688 = 0, $1689 = 0, $169 = 0, $1690 = 0, $1691 = 0, $1692 = 0, $1693 = 0, $1694 = 0, $1695 = 0, $1696 = 0, $1697 = 0, $1698 = 0, $1699 = 0, $17 = 0, $170 = 0, $1700 = 0, $1701 = 0, $1702 = 0, $1703 = 0;
 var $1704 = 0, $1705 = 0, $1706 = 0, $1707 = 0, $1708 = 0, $1709 = 0, $171 = 0, $1710 = 0, $1711 = 0, $1712 = 0, $1713 = 0, $1714 = 0, $1715 = 0, $1716 = 0, $1717 = 0, $1718 = 0, $1719 = 0, $172 = 0, $1720 = 0, $1721 = 0;
 var $1722 = 0, $1723 = 0, $1724 = 0, $1725 = 0, $1726 = 0, $1727 = 0, $1728 = 0, $1729 = 0, $173 = 0, $1730 = 0, $1731 = 0, $1732 = 0, $1733 = 0, $1734 = 0, $1735 = 0, $1736 = 0, $1737 = 0, $1738 = 0, $1739 = 0, $174 = 0;
 var $1740 = 0, $1741 = 0, $1742 = 0, $1743 = 0, $1744 = 0, $1745 = 0, $1746 = 0, $1747 = 0, $1748 = 0, $1749 = 0, $175 = 0, $1750 = 0, $1751 = 0, $1752 = 0, $1753 = 0, $1754 = 0, $1755 = 0, $1756 = 0, $1757 = 0, $1758 = 0;
 var $1759 = 0, $176 = 0, $1760 = 0, $1761 = 0, $1762 = 0, $1763 = 0, $1764 = 0, $1765 = 0, $1766 = 0, $1767 = 0, $1768 = 0, $1769 = 0, $177 = 0, $1770 = 0, $1771 = 0, $1772 = 0, $1773 = 0, $1774 = 0, $1775 = 0, $1776 = 0;
 var $1777 = 0, $1778 = 0, $1779 = 0, $178 = 0, $1780 = 0, $1781 = 0, $1782 = 0, $1783 = 0, $1784 = 0, $1785 = 0, $1786 = 0, $1787 = 0, $1788 = 0, $1789 = 0, $179 = 0, $1790 = 0, $1791 = 0, $1792 = 0, $1793 = 0, $1794 = 0;
 var $1795 = 0, $1796 = 0, $1797 = 0, $1798 = 0, $1799 = 0, $18 = 0, $180 = 0, $1800 = 0, $1801 = 0, $1802 = 0, $1803 = 0, $1804 = 0, $1805 = 0, $1806 = 0, $1807 = 0, $1808 = 0, $1809 = 0, $181 = 0, $1810 = 0, $1811 = 0;
 var $1812 = 0, $1813 = 0, $1814 = 0, $1815 = 0, $1816 = 0, $1817 = 0, $1818 = 0, $1819 = 0, $182 = 0, $1820 = 0, $1821 = 0, $1822 = 0, $1823 = 0, $1824 = 0, $1825 = 0, $1826 = 0, $1827 = 0, $1828 = 0, $1829 = 0, $183 = 0;
 var $1830 = 0, $1831 = 0, $1832 = 0, $1833 = 0, $1834 = 0, $1835 = 0, $1836 = 0, $1837 = 0, $1838 = 0, $1839 = 0, $184 = 0, $1840 = 0, $1841 = 0, $1842 = 0, $1843 = 0, $1844 = 0, $1845 = 0, $1846 = 0, $1847 = 0, $1848 = 0;
 var $1849 = 0, $185 = 0, $1850 = 0, $1851 = 0, $1852 = 0, $1853 = 0, $1854 = 0, $1855 = 0, $1856 = 0, $1857 = 0, $1858 = 0, $1859 = 0, $186 = 0, $1860 = 0, $1861 = 0, $1862 = 0, $1863 = 0, $1864 = 0, $1865 = 0, $1866 = 0;
 var $1867 = 0, $1868 = 0, $1869 = 0, $187 = 0, $1870 = 0, $1871 = 0, $1872 = 0, $1873 = 0, $1874 = 0, $1875 = 0, $1876 = 0, $1877 = 0, $1878 = 0, $1879 = 0, $188 = 0, $1880 = 0, $1881 = 0, $1882 = 0, $1883 = 0, $1884 = 0;
 var $1885 = 0, $1886 = 0, $1887 = 0, $1888 = 0, $1889 = 0, $189 = 0, $1890 = 0, $1891 = 0, $1892 = 0, $1893 = 0, $1894 = 0, $1895 = 0, $1896 = 0, $1897 = 0, $1898 = 0, $1899 = 0, $19 = 0, $190 = 0, $1900 = 0, $1901 = 0;
 var $1902 = 0, $1903 = 0, $1904 = 0, $1905 = 0, $1906 = 0, $1907 = 0, $1908 = 0, $1909 = 0, $191 = 0, $1910 = 0, $1911 = 0, $1912 = 0, $1913 = 0, $1914 = 0, $1915 = 0, $1916 = 0, $1917 = 0, $1918 = 0, $1919 = 0, $192 = 0;
 var $1920 = 0, $1921 = 0, $1922 = 0, $1923 = 0, $1924 = 0, $1925 = 0, $1926 = 0, $1927 = 0, $1928 = 0, $1929 = 0, $193 = 0, $1930 = 0, $1931 = 0, $1932 = 0, $1933 = 0, $1934 = 0, $1935 = 0, $1936 = 0, $1937 = 0, $1938 = 0;
 var $1939 = 0, $194 = 0, $1940 = 0, $1941 = 0, $1942 = 0, $1943 = 0, $1944 = 0, $1945 = 0, $1946 = 0, $1947 = 0, $1948 = 0, $1949 = 0, $195 = 0, $1950 = 0, $1951 = 0, $1952 = 0, $1953 = 0, $1954 = 0, $1955 = 0, $1956 = 0;
 var $1957 = 0, $1958 = 0, $1959 = 0, $196 = 0, $1960 = 0, $1961 = 0, $1962 = 0, $1963 = 0, $1964 = 0, $1965 = 0, $1966 = 0, $1967 = 0, $1968 = 0, $1969 = 0, $197 = 0, $1970 = 0, $1971 = 0, $1972 = 0, $1973 = 0, $1974 = 0;
 var $1975 = 0, $1976 = 0, $1977 = 0, $1978 = 0, $1979 = 0, $198 = 0, $1980 = 0, $1981 = 0, $1982 = 0, $1983 = 0, $1984 = 0, $1985 = 0, $1986 = 0, $1987 = 0, $1988 = 0, $1989 = 0, $199 = 0, $1990 = 0, $1991 = 0, $1992 = 0;
 var $1993 = 0, $1994 = 0, $1995 = 0, $1996 = 0, $1997 = 0, $1998 = 0, $1999 = 0, $2 = 0, $20 = 0, $200 = 0, $2000 = 0, $2001 = 0, $2002 = 0, $2003 = 0, $2004 = 0, $2005 = 0, $2006 = 0, $2007 = 0, $2008 = 0, $2009 = 0;
 var $201 = 0, $2010 = 0, $2011 = 0, $2012 = 0, $2013 = 0, $2014 = 0, $2015 = 0, $2016 = 0, $2017 = 0, $2018 = 0, $2019 = 0, $202 = 0, $2020 = 0, $2021 = 0, $2022 = 0, $2023 = 0, $2024 = 0, $2025 = 0, $2026 = 0, $2027 = 0;
 var $2028 = 0, $2029 = 0, $203 = 0, $2030 = 0, $2031 = 0, $2032 = 0, $2033 = 0, $2034 = 0, $2035 = 0, $2036 = 0, $2037 = 0, $2038 = 0, $2039 = 0, $204 = 0, $2040 = 0, $2041 = 0, $2042 = 0, $2043 = 0, $2044 = 0, $2045 = 0;
 var $2046 = 0, $2047 = 0, $2048 = 0, $2049 = 0, $205 = 0, $2050 = 0, $2051 = 0, $2052 = 0, $2053 = 0, $2054 = 0, $2055 = 0, $2056 = 0, $2057 = 0, $2058 = 0, $2059 = 0, $206 = 0, $2060 = 0, $2061 = 0, $2062 = 0, $2063 = 0;
 var $2064 = 0, $2065 = 0, $2066 = 0, $2067 = 0, $2068 = 0, $2069 = 0, $207 = 0, $2070 = 0, $2071 = 0, $2072 = 0, $2073 = 0, $2074 = 0, $2075 = 0, $2076 = 0, $2077 = 0, $2078 = 0, $2079 = 0, $208 = 0, $2080 = 0, $2081 = 0;
 var $2082 = 0, $2083 = 0, $2084 = 0, $2085 = 0, $2086 = 0, $2087 = 0, $2088 = 0, $2089 = 0, $209 = 0, $2090 = 0, $2091 = 0, $2092 = 0, $2093 = 0, $2094 = 0, $2095 = 0, $2096 = 0, $2097 = 0, $2098 = 0, $2099 = 0, $21 = 0;
 var $210 = 0, $2100 = 0, $2101 = 0, $2102 = 0, $2103 = 0, $2104 = 0, $2105 = 0, $2106 = 0, $2107 = 0, $2108 = 0, $2109 = 0, $211 = 0, $2110 = 0, $2111 = 0, $2112 = 0, $2113 = 0, $2114 = 0, $2115 = 0, $2116 = 0, $2117 = 0;
 var $2118 = 0, $2119 = 0, $212 = 0, $2120 = 0, $2121 = 0, $2122 = 0, $2123 = 0, $2124 = 0, $2125 = 0, $2126 = 0, $2127 = 0, $2128 = 0, $2129 = 0, $213 = 0, $2130 = 0, $2131 = 0, $2132 = 0, $2133 = 0, $2134 = 0, $2135 = 0;
 var $2136 = 0, $2137 = 0, $2138 = 0, $2139 = 0, $214 = 0, $2140 = 0, $2141 = 0, $2142 = 0, $2143 = 0, $2144 = 0, $2145 = 0, $2146 = 0, $2147 = 0, $2148 = 0, $2149 = 0, $215 = 0, $2150 = 0, $2151 = 0, $2152 = 0, $2153 = 0;
 var $2154 = 0, $2155 = 0, $2156 = 0, $2157 = 0, $2158 = 0, $2159 = 0, $216 = 0, $2160 = 0, $2161 = 0, $2162 = 0, $2163 = 0, $2164 = 0, $2165 = 0, $2166 = 0, $2167 = 0, $2168 = 0, $2169 = 0, $217 = 0, $2170 = 0, $2171 = 0;
 var $2172 = 0, $2173 = 0, $2174 = 0, $2175 = 0, $2176 = 0, $2177 = 0, $2178 = 0, $2179 = 0, $218 = 0, $2180 = 0, $2181 = 0, $2182 = 0, $2183 = 0, $2184 = 0, $2185 = 0, $2186 = 0, $2187 = 0, $2188 = 0, $2189 = 0, $219 = 0;
 var $2190 = 0, $2191 = 0, $2192 = 0, $2193 = 0, $2194 = 0, $2195 = 0, $2196 = 0, $2197 = 0, $2198 = 0, $2199 = 0, $22 = 0, $220 = 0, $2200 = 0, $2201 = 0, $2202 = 0, $2203 = 0, $2204 = 0, $2205 = 0, $2206 = 0, $2207 = 0;
 var $2208 = 0, $2209 = 0, $221 = 0, $2210 = 0, $2211 = 0, $2212 = 0, $2213 = 0, $2214 = 0, $2215 = 0, $2216 = 0, $2217 = 0, $2218 = 0, $2219 = 0, $222 = 0, $2220 = 0, $2221 = 0, $2222 = 0, $2223 = 0, $2224 = 0, $2225 = 0;
 var $2226 = 0, $2227 = 0, $2228 = 0, $2229 = 0, $223 = 0, $2230 = 0, $2231 = 0, $2232 = 0, $2233 = 0, $2234 = 0, $2235 = 0, $2236 = 0, $2237 = 0, $2238 = 0, $2239 = 0, $224 = 0, $2240 = 0, $2241 = 0, $2242 = 0, $2243 = 0;
 var $2244 = 0, $2245 = 0, $2246 = 0, $2247 = 0, $2248 = 0, $2249 = 0, $225 = 0, $2250 = 0, $2251 = 0, $2252 = 0, $2253 = 0, $2254 = 0, $2255 = 0, $2256 = 0, $2257 = 0, $2258 = 0, $2259 = 0, $226 = 0, $2260 = 0, $2261 = 0;
 var $2262 = 0, $2263 = 0, $2264 = 0, $2265 = 0, $2266 = 0, $2267 = 0, $2268 = 0, $2269 = 0, $227 = 0, $2270 = 0, $2271 = 0, $2272 = 0, $2273 = 0, $2274 = 0, $2275 = 0, $2276 = 0, $2277 = 0, $2278 = 0, $2279 = 0, $228 = 0;
 var $2280 = 0, $2281 = 0, $2282 = 0, $2283 = 0, $2284 = 0, $2285 = 0, $2286 = 0, $2287 = 0, $2288 = 0, $2289 = 0, $229 = 0, $2290 = 0, $2291 = 0, $2292 = 0, $2293 = 0, $2294 = 0, $2295 = 0, $2296 = 0, $2297 = 0, $2298 = 0;
 var $2299 = 0, $23 = 0, $230 = 0, $2300 = 0, $2301 = 0, $2302 = 0, $2303 = 0, $2304 = 0, $2305 = 0, $2306 = 0, $2307 = 0, $2308 = 0, $2309 = 0, $231 = 0, $2310 = 0, $2311 = 0, $2312 = 0, $2313 = 0, $2314 = 0, $2315 = 0;
 var $2316 = 0, $2317 = 0, $2318 = 0, $2319 = 0, $232 = 0, $2320 = 0, $2321 = 0, $2322 = 0, $2323 = 0, $2324 = 0, $2325 = 0, $2326 = 0, $2327 = 0, $2328 = 0, $2329 = 0, $233 = 0, $2330 = 0, $2331 = 0, $2332 = 0, $2333 = 0;
 var $2334 = 0, $2335 = 0, $2336 = 0, $2337 = 0, $2338 = 0, $2339 = 0, $234 = 0, $2340 = 0, $2341 = 0, $2342 = 0, $2343 = 0, $2344 = 0, $2345 = 0, $2346 = 0, $2347 = 0, $2348 = 0, $2349 = 0, $235 = 0, $2350 = 0, $2351 = 0;
 var $2352 = 0, $2353 = 0, $2354 = 0, $2355 = 0, $2356 = 0, $2357 = 0, $2358 = 0, $2359 = 0, $236 = 0, $2360 = 0, $2361 = 0, $2362 = 0, $2363 = 0, $2364 = 0, $2365 = 0, $2366 = 0, $2367 = 0, $2368 = 0, $2369 = 0, $237 = 0;
 var $2370 = 0, $2371 = 0, $2372 = 0, $2373 = 0, $2374 = 0, $2375 = 0, $2376 = 0, $2377 = 0, $2378 = 0, $2379 = 0, $238 = 0, $2380 = 0, $2381 = 0, $2382 = 0, $2383 = 0, $2384 = 0, $2385 = 0, $2386 = 0, $2387 = 0, $2388 = 0;
 var $2389 = 0, $239 = 0, $2390 = 0, $2391 = 0, $2392 = 0, $2393 = 0, $2394 = 0, $2395 = 0, $2396 = 0, $2397 = 0, $2398 = 0, $2399 = 0, $24 = 0, $240 = 0, $2400 = 0, $2401 = 0, $2402 = 0, $2403 = 0, $2404 = 0, $2405 = 0;
 var $2406 = 0, $2407 = 0, $2408 = 0, $2409 = 0, $241 = 0, $2410 = 0, $2411 = 0, $2412 = 0, $2413 = 0, $2414 = 0, $2415 = 0, $2416 = 0, $2417 = 0, $2418 = 0, $2419 = 0, $242 = 0, $2420 = 0, $2421 = 0, $2422 = 0, $2423 = 0;
 var $2424 = 0, $2425 = 0, $2426 = 0, $2427 = 0, $2428 = 0, $2429 = 0, $243 = 0, $2430 = 0, $2431 = 0, $2432 = 0, $2433 = 0, $2434 = 0, $2435 = 0, $2436 = 0, $2437 = 0, $2438 = 0, $2439 = 0, $244 = 0, $2440 = 0, $2441 = 0;
 var $2442 = 0, $2443 = 0, $2444 = 0, $2445 = 0, $2446 = 0, $2447 = 0, $2448 = 0, $2449 = 0, $245 = 0, $2450 = 0, $2451 = 0, $2452 = 0, $2453 = 0, $2454 = 0, $2455 = 0, $2456 = 0, $2457 = 0, $2458 = 0, $2459 = 0, $246 = 0;
 var $2460 = 0, $2461 = 0, $2462 = 0, $2463 = 0, $2464 = 0, $2465 = 0, $2466 = 0, $2467 = 0, $2468 = 0, $2469 = 0, $247 = 0, $2470 = 0, $2471 = 0, $2472 = 0, $2473 = 0, $2474 = 0, $2475 = 0, $2476 = 0, $2477 = 0, $2478 = 0;
 var $2479 = 0, $248 = 0, $2480 = 0, $2481 = 0, $2482 = 0, $2483 = 0, $2484 = 0, $2485 = 0, $2486 = 0, $2487 = 0, $2488 = 0, $2489 = 0, $249 = 0, $2490 = 0, $2491 = 0, $2492 = 0, $2493 = 0, $2494 = 0, $2495 = 0, $2496 = 0;
 var $2497 = 0, $2498 = 0, $2499 = 0, $25 = 0, $250 = 0, $2500 = 0, $2501 = 0, $2502 = 0, $2503 = 0, $2504 = 0, $2505 = 0, $2506 = 0, $2507 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0;
 var $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0;
 var $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0;
 var $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0;
 var $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0;
 var $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0;
 var $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0;
 var $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0;
 var $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0;
 var $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0;
 var $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0;
 var $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0;
 var $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0;
 var $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0;
 var $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0;
 var $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0;
 var $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0;
 var $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0;
 var $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0;
 var $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0;
 var $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0;
 var $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0;
 var $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0;
 var $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0;
 var $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0;
 var $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0;
 var $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0;
 var $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0;
 var $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0;
 var $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0;
 var $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0;
 var $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0;
 var $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0;
 var $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0;
 var $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0;
 var $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0;
 var $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0;
 var $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0;
 var $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0;
 var $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0;
 var $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0;
 var $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0;
 var $997 = 0, $998 = 0, $999 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n106$0 = 0, $n106$1 = 0, $n106$2 = 0, $n106$3 = 0, $n128$0 = 0, $n128$1 = 0, $n128$2 = 0, $n128$3 = 0, $n150$0 = 0, $n150$1 = 0, $n150$2 = 0, $n150$3 = 0, $n18$0 = 0;
 var $n18$1 = 0, $n18$2 = 0, $n18$3 = 0, $n40$0 = 0, $n40$1 = 0, $n40$2 = 0, $n40$3 = 0, $n62$0 = 0, $n62$1 = 0, $n62$2 = 0, $n62$3 = 0, $n84$0 = 0, $n84$1 = 0, $n84$2 = 0, $n84$3 = 0, $pixel$032 = 0, $pixel$1 = 0, $pixel$2 = 0, $pixel$3 = 0, $pixel$4 = 0;
 var $pixel105$042 = 0, $pixel105$1 = 0, $pixel105$2 = 0, $pixel105$3 = 0, $pixel105$4 = 0, $pixel127$044 = 0, $pixel127$1 = 0, $pixel127$2 = 0, $pixel127$3 = 0, $pixel127$4 = 0, $pixel149$038 = 0, $pixel149$1 = 0, $pixel149$2 = 0, $pixel149$3 = 0, $pixel149$4 = 0, $pixel17$034 = 0, $pixel17$1 = 0, $pixel17$2 = 0, $pixel17$3 = 0, $pixel17$4 = 0;
 var $pixel39$036 = 0, $pixel39$1 = 0, $pixel39$2 = 0, $pixel39$3 = 0, $pixel39$4 = 0, $pixel61$031 = 0, $pixel61$1 = 0, $pixel61$2 = 0, $pixel61$3 = 0, $pixel61$4 = 0, $pixel83$040 = 0, $pixel83$1 = 0, $pixel83$2 = 0, $pixel83$3 = 0, $pixel83$4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($dst)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $a&255;
 $3 = $2 ^ 255;
 $4 = ((($1)) + 9|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 switch ($6|0) {
 case 2:  {
  switch ($blendMode|0) {
  case 1:  {
   $7 = ((($rect)) + 8|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ((($rect)) + 12|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ((($dst)) + 16|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = (($12|0) / 2)&-1;
   $14 = (($13) - ($8))|0;
   $15 = ($10|0)==(0);
   if ($15) {
    $$0 = 0;
    return ($$0|0);
   }
   $16 = ((($dst)) + 20|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($rect)) + 4|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = Math_imul($19, $13)|0;
   $21 = (($17) + ($20<<1)|0);
   $22 = HEAP32[$rect>>2]|0;
   $23 = (($21) + ($22<<1)|0);
   $24 = (($8) + 3)|0;
   $25 = (($24|0) / 4)&-1;
   $26 = $8 & 3;
   $27 = ((($1)) + 12|0);
   $28 = ((($1)) + 32|0);
   $29 = ((($1)) + 28|0);
   $30 = ((($1)) + 16|0);
   $31 = ((($1)) + 20|0);
   $32 = $r&255;
   $33 = $g&255;
   $34 = $b&255;
   $35 = ((($1)) + 24|0);
   $$in51 = $10;$pixel$032 = $23;
   L31: while(1) {
    $36 = (($$in51) + -1)|0;
    switch ($26|0) {
    case 0:  {
     $$pre207 = HEAP32[$27>>2]|0;
     $93 = HEAP16[$28>>1]|0;
     $94 = HEAP32[$28>>2]|0;
     $95 = $94&65535;
     $96 = $94&255;
     $97 = HEAP16[$29>>1]|0;
     $98 = HEAP32[$29>>2]|0;
     $99 = $98&65535;
     $100 = $98&255;
     $$phi$trans$insert210 = $98 & 255;
     $$phi$trans$insert211 = (125924 + ($$phi$trans$insert210<<2)|0);
     $$pre212 = HEAP32[$$phi$trans$insert211>>2]|0;
     $$pre213 = HEAP32[$30>>2]|0;
     $101 = ($95&65535) >>> 8;
     $102 = $101&255;
     $103 = ($99&65535) >>> 8;
     $104 = $103&255;
     $$phi$trans$insert214 = $103&65535;
     $$phi$trans$insert215 = (125924 + ($$phi$trans$insert214<<2)|0);
     $$pre216 = HEAP32[$$phi$trans$insert215>>2]|0;
     $$pre217 = HEAP32[$31>>2]|0;
     $105 = $94 >>> 16;
     $106 = $105&255;
     $107 = $98 >>> 16;
     $108 = $107&255;
     $$phi$trans$insert218 = $107 & 255;
     $$phi$trans$insert219 = (125924 + ($$phi$trans$insert218<<2)|0);
     $$pre220 = HEAP32[$$phi$trans$insert219>>2]|0;
     $$pre221 = HEAP32[$35>>2]|0;
     $109 = $94&255;
     $110 = $98&255;
     $111 = ($93&65535) >>> 8;
     $112 = $111&255;
     $113 = ($97&65535) >>> 8;
     $114 = $113&255;
     $118 = $$pre207;$120 = $96;$123 = $100;$125 = $$pre212;$129 = $$pre213;$131 = $102;$134 = $104;$136 = $$pre216;$140 = $$pre217;$142 = $106;$145 = $108;$147 = $$pre220;$167 = $$pre221;$2420 = $114;$2421 = $112;$2422 = $110;$2423 = $109;$n$0 = $25;$pixel$1 = $pixel$032;
     label = 10;
     break;
    }
    case 3:  {
     $$pre223 = HEAP32[$27>>2]|0;
     $71 = HEAP16[$28>>1]|0;
     $72 = HEAP32[$28>>2]|0;
     $73 = $72&65535;
     $74 = $72&255;
     $75 = HEAP16[$29>>1]|0;
     $76 = HEAP32[$29>>2]|0;
     $77 = $76&65535;
     $78 = $76&255;
     $$pre226 = HEAP32[$30>>2]|0;
     $79 = ($73&65535) >>> 8;
     $80 = $79&255;
     $81 = ($77&65535) >>> 8;
     $82 = $81&255;
     $$pre227 = HEAP32[$31>>2]|0;
     $83 = $72 >>> 16;
     $84 = $83&255;
     $85 = $76 >>> 16;
     $86 = $85&255;
     $$pre228 = HEAP32[$35>>2]|0;
     $87 = $72&255;
     $88 = $76&255;
     $89 = ($71&65535) >>> 8;
     $90 = $89&255;
     $91 = ($75&65535) >>> 8;
     $92 = $91&255;
     $174 = $$pre223;$176 = $74;$179 = $78;$186 = $$pre226;$188 = $80;$191 = $82;$198 = $$pre227;$200 = $84;$203 = $86;$226 = $$pre228;$2416 = $92;$2417 = $90;$2418 = $88;$2419 = $87;$n$1 = $25;$pixel$2 = $pixel$032;
     label = 11;
     break;
    }
    case 2:  {
     $$pre230 = HEAP32[$27>>2]|0;
     $49 = HEAP16[$28>>1]|0;
     $50 = HEAP32[$28>>2]|0;
     $51 = $50&65535;
     $52 = $50&255;
     $53 = HEAP16[$29>>1]|0;
     $54 = HEAP32[$29>>2]|0;
     $55 = $54&65535;
     $56 = $54&255;
     $$pre233 = HEAP32[$30>>2]|0;
     $57 = ($51&65535) >>> 8;
     $58 = $57&255;
     $59 = ($55&65535) >>> 8;
     $60 = $59&255;
     $$pre234 = HEAP32[$31>>2]|0;
     $61 = $50 >>> 16;
     $62 = $61&255;
     $63 = $54 >>> 16;
     $64 = $63&255;
     $$pre235 = HEAP32[$35>>2]|0;
     $65 = $50&255;
     $66 = $54&255;
     $67 = ($49&65535) >>> 8;
     $68 = $67&255;
     $69 = ($53&65535) >>> 8;
     $70 = $69&255;
     $233 = $$pre230;$235 = $52;$238 = $56;$2412 = $70;$2413 = $68;$2414 = $66;$2415 = $65;$245 = $$pre233;$247 = $58;$250 = $60;$257 = $$pre234;$259 = $62;$262 = $64;$285 = $$pre235;$n$2 = $25;$pixel$3 = $pixel$032;
     label = 12;
     break;
    }
    case 1:  {
     $$pre237 = HEAP32[$27>>2]|0;
     $$pre238 = HEAP32[$28>>2]|0;
     $37 = $$pre238&65535;
     $38 = $$pre238&255;
     $$pre239 = HEAP32[$29>>2]|0;
     $39 = $$pre239&65535;
     $40 = $$pre239&255;
     $$pre240 = HEAP32[$30>>2]|0;
     $41 = ($37&65535) >>> 8;
     $42 = $41&255;
     $43 = ($39&65535) >>> 8;
     $44 = $43&255;
     $$pre241 = HEAP32[$31>>2]|0;
     $45 = $$pre238 >>> 16;
     $46 = $45&255;
     $47 = $$pre239 >>> 16;
     $48 = $47&255;
     $$pre242 = HEAP32[$35>>2]|0;
     $292 = $$pre237;$294 = $38;$297 = $40;$304 = $$pre240;$306 = $42;$309 = $44;$316 = $$pre241;$318 = $46;$321 = $48;$344 = $$pre242;$n$3 = $25;$pixel$4 = $pixel$032;
     label = 13;
     break;
    }
    default: {
     label = 14;
     break L31;
    }
    }
    while(1) {
     if ((label|0) == 10) {
      label = 0;
      $115 = HEAP16[$pixel$1>>1]|0;
      $116 = $115&65535;
      $117 = $116 & $118;
      $119 = $120&255;
      $121 = $117 >>> $119;
      $122 = $123&255;
      $124 = (($125) + ($121)|0);
      $126 = HEAP8[$124>>0]|0;
      $127 = $126&255;
      $128 = $129 & $116;
      $130 = $131&255;
      $132 = $128 >>> $130;
      $133 = $134&255;
      $135 = (($136) + ($132)|0);
      $137 = HEAP8[$135>>0]|0;
      $138 = $137&255;
      $139 = $140 & $116;
      $141 = $142&255;
      $143 = $139 >>> $141;
      $144 = $145&255;
      $146 = (($147) + ($143)|0);
      $148 = HEAP8[$146>>0]|0;
      $149 = $148&255;
      $150 = Math_imul($127, $3)|0;
      $151 = (($150>>>0) / 255)&-1;
      $152 = (($151) + ($32))|0;
      $153 = Math_imul($138, $3)|0;
      $154 = (($153>>>0) / 255)&-1;
      $155 = (($154) + ($33))|0;
      $156 = Math_imul($149, $3)|0;
      $157 = (($156>>>0) / 255)&-1;
      $158 = (($157) + ($34))|0;
      $159 = $152 >>> $122;
      $160 = $159 << $119;
      $161 = $155 >>> $133;
      $162 = $161 << $130;
      $163 = $162 | $160;
      $164 = $158 >>> $144;
      $165 = $164 << $141;
      $166 = $163 | $167;
      $168 = $166 | $165;
      $169 = $168&65535;
      HEAP16[$pixel$1>>1] = $169;
      $170 = ((($pixel$1)) + 2|0);
      $174 = $118;$176 = $2423;$179 = $2422;$186 = $129;$188 = $2421;$191 = $2420;$198 = $140;$200 = $142;$203 = $145;$226 = $167;$2416 = $2420;$2417 = $2421;$2418 = $2422;$2419 = $2423;$n$1 = $n$0;$pixel$2 = $170;
      label = 11;
      continue;
     }
     else if ((label|0) == 11) {
      label = 0;
      $171 = HEAP16[$pixel$2>>1]|0;
      $172 = $171&65535;
      $173 = $172 & $174;
      $175 = $176&255;
      $177 = $173 >>> $175;
      $178 = $179&255;
      $180 = (125924 + ($178<<2)|0);
      $181 = HEAP32[$180>>2]|0;
      $182 = (($181) + ($177)|0);
      $183 = HEAP8[$182>>0]|0;
      $184 = $183&255;
      $185 = $186 & $172;
      $187 = $188&255;
      $189 = $185 >>> $187;
      $190 = $191&255;
      $192 = (125924 + ($190<<2)|0);
      $193 = HEAP32[$192>>2]|0;
      $194 = (($193) + ($189)|0);
      $195 = HEAP8[$194>>0]|0;
      $196 = $195&255;
      $197 = $198 & $172;
      $199 = $200&255;
      $201 = $197 >>> $199;
      $202 = $203&255;
      $204 = (125924 + ($202<<2)|0);
      $205 = HEAP32[$204>>2]|0;
      $206 = (($205) + ($201)|0);
      $207 = HEAP8[$206>>0]|0;
      $208 = $207&255;
      $209 = Math_imul($184, $3)|0;
      $210 = (($209>>>0) / 255)&-1;
      $211 = (($210) + ($32))|0;
      $212 = Math_imul($196, $3)|0;
      $213 = (($212>>>0) / 255)&-1;
      $214 = (($213) + ($33))|0;
      $215 = Math_imul($208, $3)|0;
      $216 = (($215>>>0) / 255)&-1;
      $217 = (($216) + ($34))|0;
      $218 = $211 >>> $178;
      $219 = $218 << $175;
      $220 = $214 >>> $190;
      $221 = $220 << $187;
      $222 = $221 | $219;
      $223 = $217 >>> $202;
      $224 = $223 << $199;
      $225 = $222 | $226;
      $227 = $225 | $224;
      $228 = $227&65535;
      HEAP16[$pixel$2>>1] = $228;
      $229 = ((($pixel$2)) + 2|0);
      $233 = $174;$235 = $2419;$238 = $2418;$2412 = $2416;$2413 = $2417;$2414 = $2418;$2415 = $2419;$245 = $186;$247 = $2417;$250 = $2416;$257 = $198;$259 = $200;$262 = $203;$285 = $226;$n$2 = $n$1;$pixel$3 = $229;
      label = 12;
      continue;
     }
     else if ((label|0) == 12) {
      label = 0;
      $230 = HEAP16[$pixel$3>>1]|0;
      $231 = $230&65535;
      $232 = $231 & $233;
      $234 = $235&255;
      $236 = $232 >>> $234;
      $237 = $238&255;
      $239 = (125924 + ($237<<2)|0);
      $240 = HEAP32[$239>>2]|0;
      $241 = (($240) + ($236)|0);
      $242 = HEAP8[$241>>0]|0;
      $243 = $242&255;
      $244 = $245 & $231;
      $246 = $247&255;
      $248 = $244 >>> $246;
      $249 = $250&255;
      $251 = (125924 + ($249<<2)|0);
      $252 = HEAP32[$251>>2]|0;
      $253 = (($252) + ($248)|0);
      $254 = HEAP8[$253>>0]|0;
      $255 = $254&255;
      $256 = $257 & $231;
      $258 = $259&255;
      $260 = $256 >>> $258;
      $261 = $262&255;
      $263 = (125924 + ($261<<2)|0);
      $264 = HEAP32[$263>>2]|0;
      $265 = (($264) + ($260)|0);
      $266 = HEAP8[$265>>0]|0;
      $267 = $266&255;
      $268 = Math_imul($243, $3)|0;
      $269 = (($268>>>0) / 255)&-1;
      $270 = (($269) + ($32))|0;
      $271 = Math_imul($255, $3)|0;
      $272 = (($271>>>0) / 255)&-1;
      $273 = (($272) + ($33))|0;
      $274 = Math_imul($267, $3)|0;
      $275 = (($274>>>0) / 255)&-1;
      $276 = (($275) + ($34))|0;
      $277 = $270 >>> $237;
      $278 = $277 << $234;
      $279 = $273 >>> $249;
      $280 = $279 << $246;
      $281 = $280 | $278;
      $282 = $276 >>> $261;
      $283 = $282 << $258;
      $284 = $281 | $285;
      $286 = $284 | $283;
      $287 = $286&65535;
      HEAP16[$pixel$3>>1] = $287;
      $288 = ((($pixel$3)) + 2|0);
      $292 = $233;$294 = $2415;$297 = $2414;$304 = $245;$306 = $2413;$309 = $2412;$316 = $257;$318 = $259;$321 = $262;$344 = $285;$n$3 = $n$2;$pixel$4 = $288;
      label = 13;
      continue;
     }
     else if ((label|0) == 13) {
      label = 0;
      $289 = HEAP16[$pixel$4>>1]|0;
      $290 = $289&65535;
      $291 = $290 & $292;
      $293 = $294&255;
      $295 = $291 >>> $293;
      $296 = $297&255;
      $298 = (125924 + ($296<<2)|0);
      $299 = HEAP32[$298>>2]|0;
      $300 = (($299) + ($295)|0);
      $301 = HEAP8[$300>>0]|0;
      $302 = $301&255;
      $303 = $304 & $290;
      $305 = $306&255;
      $307 = $303 >>> $305;
      $308 = $309&255;
      $310 = (125924 + ($308<<2)|0);
      $311 = HEAP32[$310>>2]|0;
      $312 = (($311) + ($307)|0);
      $313 = HEAP8[$312>>0]|0;
      $314 = $313&255;
      $315 = $316 & $290;
      $317 = $318&255;
      $319 = $315 >>> $317;
      $320 = $321&255;
      $322 = (125924 + ($320<<2)|0);
      $323 = HEAP32[$322>>2]|0;
      $324 = (($323) + ($319)|0);
      $325 = HEAP8[$324>>0]|0;
      $326 = $325&255;
      $327 = Math_imul($302, $3)|0;
      $328 = (($327>>>0) / 255)&-1;
      $329 = (($328) + ($32))|0;
      $330 = Math_imul($314, $3)|0;
      $331 = (($330>>>0) / 255)&-1;
      $332 = (($331) + ($33))|0;
      $333 = Math_imul($326, $3)|0;
      $334 = (($333>>>0) / 255)&-1;
      $335 = (($334) + ($34))|0;
      $336 = $329 >>> $296;
      $337 = $336 << $293;
      $338 = $332 >>> $308;
      $339 = $338 << $305;
      $340 = $339 | $337;
      $341 = $335 >>> $320;
      $342 = $341 << $317;
      $343 = $340 | $344;
      $345 = $343 | $342;
      $346 = $345&65535;
      HEAP16[$pixel$4>>1] = $346;
      $347 = ((($pixel$4)) + 2|0);
      $348 = (($n$3) + -1)|0;
      $349 = ($n$3|0)>(1);
      if ($349) {
       $118 = $292;$120 = $294;$123 = $297;$125 = $299;$129 = $304;$131 = $306;$134 = $309;$136 = $311;$140 = $316;$142 = $318;$145 = $321;$147 = $323;$167 = $344;$2420 = $309;$2421 = $306;$2422 = $297;$2423 = $294;$n$0 = $348;$pixel$1 = $347;
       label = 10;
       continue;
      } else {
       break;
      }
     }
    }
    $350 = (($347) + ($14<<1)|0);
    $351 = ($36|0)==(0);
    if ($351) {
     $$0 = 0;
     label = 109;
     break;
    } else {
     $$in51 = $36;$pixel$032 = $350;
    }
   }
   if ((label|0) == 14) {
    // unreachable;
   }
   else if ((label|0) == 109) {
    return ($$0|0);
   }
   break;
  }
  case 2:  {
   $352 = ((($rect)) + 8|0);
   $353 = HEAP32[$352>>2]|0;
   $354 = ((($rect)) + 12|0);
   $355 = HEAP32[$354>>2]|0;
   $356 = ((($dst)) + 16|0);
   $357 = HEAP32[$356>>2]|0;
   $358 = (($357|0) / 2)&-1;
   $359 = (($358) - ($353))|0;
   $360 = ($355|0)==(0);
   if ($360) {
    $$0 = 0;
    return ($$0|0);
   }
   $361 = ((($dst)) + 20|0);
   $362 = HEAP32[$361>>2]|0;
   $363 = ((($rect)) + 4|0);
   $364 = HEAP32[$363>>2]|0;
   $365 = Math_imul($364, $358)|0;
   $366 = (($362) + ($365<<1)|0);
   $367 = HEAP32[$rect>>2]|0;
   $368 = (($366) + ($367<<1)|0);
   $369 = (($353) + 3)|0;
   $370 = (($369|0) / 4)&-1;
   $371 = $353 & 3;
   $372 = ((($1)) + 12|0);
   $373 = ((($1)) + 32|0);
   $374 = ((($1)) + 28|0);
   $375 = ((($1)) + 16|0);
   $376 = ((($1)) + 20|0);
   $377 = $r&255;
   $378 = $g&255;
   $379 = $b&255;
   $380 = ((($1)) + 24|0);
   $$in50 = $355;$pixel17$034 = $368;
   L52: while(1) {
    $381 = (($$in50) + -1)|0;
    switch ($371|0) {
    case 0:  {
     $$pre170 = HEAP32[$372>>2]|0;
     $438 = HEAP16[$373>>1]|0;
     $439 = HEAP32[$373>>2]|0;
     $440 = $439&65535;
     $441 = $439&255;
     $442 = HEAP16[$374>>1]|0;
     $443 = HEAP32[$374>>2]|0;
     $444 = $443&65535;
     $445 = $443&255;
     $$phi$trans$insert173 = $443 & 255;
     $$phi$trans$insert174 = (125924 + ($$phi$trans$insert173<<2)|0);
     $$pre175 = HEAP32[$$phi$trans$insert174>>2]|0;
     $$pre176 = HEAP32[$375>>2]|0;
     $446 = ($440&65535) >>> 8;
     $447 = $446&255;
     $448 = ($444&65535) >>> 8;
     $449 = $448&255;
     $$phi$trans$insert177 = $448&65535;
     $$phi$trans$insert178 = (125924 + ($$phi$trans$insert177<<2)|0);
     $$pre179 = HEAP32[$$phi$trans$insert178>>2]|0;
     $$pre180 = HEAP32[$376>>2]|0;
     $450 = $439 >>> 16;
     $451 = $450&255;
     $452 = $443 >>> 16;
     $453 = $452&255;
     $$phi$trans$insert181 = $452 & 255;
     $$phi$trans$insert182 = (125924 + ($$phi$trans$insert181<<2)|0);
     $$pre183 = HEAP32[$$phi$trans$insert182>>2]|0;
     $$pre184 = HEAP32[$380>>2]|0;
     $454 = $439&255;
     $455 = $443&255;
     $456 = ($438&65535) >>> 8;
     $457 = $456&255;
     $458 = ($442&65535) >>> 8;
     $459 = $458&255;
     $2432 = $459;$2433 = $457;$2434 = $455;$2435 = $454;$463 = $$pre170;$465 = $441;$468 = $445;$470 = $$pre175;$474 = $$pre176;$476 = $447;$479 = $449;$481 = $$pre179;$485 = $$pre180;$487 = $451;$490 = $453;$492 = $$pre183;$509 = $$pre184;$n18$0 = $370;$pixel17$1 = $pixel17$034;
     label = 23;
     break;
    }
    case 3:  {
     $$pre186 = HEAP32[$372>>2]|0;
     $416 = HEAP16[$373>>1]|0;
     $417 = HEAP32[$373>>2]|0;
     $418 = $417&65535;
     $419 = $417&255;
     $420 = HEAP16[$374>>1]|0;
     $421 = HEAP32[$374>>2]|0;
     $422 = $421&65535;
     $423 = $421&255;
     $$pre189 = HEAP32[$375>>2]|0;
     $424 = ($418&65535) >>> 8;
     $425 = $424&255;
     $426 = ($422&65535) >>> 8;
     $427 = $426&255;
     $$pre190 = HEAP32[$376>>2]|0;
     $428 = $417 >>> 16;
     $429 = $428&255;
     $430 = $421 >>> 16;
     $431 = $430&255;
     $$pre191 = HEAP32[$380>>2]|0;
     $432 = $417&255;
     $433 = $421&255;
     $434 = ($416&65535) >>> 8;
     $435 = $434&255;
     $436 = ($420&65535) >>> 8;
     $437 = $436&255;
     $2428 = $437;$2429 = $435;$2430 = $433;$2431 = $432;$516 = $$pre186;$518 = $419;$521 = $423;$528 = $$pre189;$530 = $425;$533 = $427;$540 = $$pre190;$542 = $429;$545 = $431;$565 = $$pre191;$n18$1 = $370;$pixel17$2 = $pixel17$034;
     label = 24;
     break;
    }
    case 2:  {
     $$pre193 = HEAP32[$372>>2]|0;
     $394 = HEAP16[$373>>1]|0;
     $395 = HEAP32[$373>>2]|0;
     $396 = $395&65535;
     $397 = $395&255;
     $398 = HEAP16[$374>>1]|0;
     $399 = HEAP32[$374>>2]|0;
     $400 = $399&65535;
     $401 = $399&255;
     $$pre196 = HEAP32[$375>>2]|0;
     $402 = ($396&65535) >>> 8;
     $403 = $402&255;
     $404 = ($400&65535) >>> 8;
     $405 = $404&255;
     $$pre197 = HEAP32[$376>>2]|0;
     $406 = $395 >>> 16;
     $407 = $406&255;
     $408 = $399 >>> 16;
     $409 = $408&255;
     $$pre198 = HEAP32[$380>>2]|0;
     $410 = $395&255;
     $411 = $399&255;
     $412 = ($394&65535) >>> 8;
     $413 = $412&255;
     $414 = ($398&65535) >>> 8;
     $415 = $414&255;
     $2424 = $415;$2425 = $413;$2426 = $411;$2427 = $410;$572 = $$pre193;$574 = $397;$577 = $401;$584 = $$pre196;$586 = $403;$589 = $405;$596 = $$pre197;$598 = $407;$601 = $409;$621 = $$pre198;$n18$2 = $370;$pixel17$3 = $pixel17$034;
     label = 25;
     break;
    }
    case 1:  {
     $$pre200 = HEAP32[$372>>2]|0;
     $$pre201 = HEAP32[$373>>2]|0;
     $382 = $$pre201&65535;
     $383 = $$pre201&255;
     $$pre202 = HEAP32[$374>>2]|0;
     $384 = $$pre202&65535;
     $385 = $$pre202&255;
     $$pre203 = HEAP32[$375>>2]|0;
     $386 = ($382&65535) >>> 8;
     $387 = $386&255;
     $388 = ($384&65535) >>> 8;
     $389 = $388&255;
     $$pre204 = HEAP32[$376>>2]|0;
     $390 = $$pre201 >>> 16;
     $391 = $390&255;
     $392 = $$pre202 >>> 16;
     $393 = $392&255;
     $$pre205 = HEAP32[$380>>2]|0;
     $628 = $$pre200;$630 = $383;$633 = $385;$640 = $$pre203;$642 = $387;$645 = $389;$652 = $$pre204;$654 = $391;$657 = $393;$677 = $$pre205;$n18$3 = $370;$pixel17$4 = $pixel17$034;
     label = 26;
     break;
    }
    default: {
     label = 27;
     break L52;
    }
    }
    while(1) {
     if ((label|0) == 23) {
      label = 0;
      $460 = HEAP16[$pixel17$1>>1]|0;
      $461 = $460&65535;
      $462 = $461 & $463;
      $464 = $465&255;
      $466 = $462 >>> $464;
      $467 = $468&255;
      $469 = (($470) + ($466)|0);
      $471 = HEAP8[$469>>0]|0;
      $472 = $471&255;
      $473 = $474 & $461;
      $475 = $476&255;
      $477 = $473 >>> $475;
      $478 = $479&255;
      $480 = (($481) + ($477)|0);
      $482 = HEAP8[$480>>0]|0;
      $483 = $482&255;
      $484 = $485 & $461;
      $486 = $487&255;
      $488 = $484 >>> $486;
      $489 = $490&255;
      $491 = (($492) + ($488)|0);
      $493 = HEAP8[$491>>0]|0;
      $494 = $493&255;
      $495 = (($472) + ($377))|0;
      $496 = ($495>>>0)>(255);
      $$ = $496 ? 255 : $495;
      $497 = (($483) + ($378))|0;
      $498 = ($497>>>0)>(255);
      $$16 = $498 ? 255 : $497;
      $499 = (($494) + ($379))|0;
      $500 = ($499>>>0)>(255);
      $$1 = $500 ? 255 : $499;
      $501 = $$ >>> $467;
      $502 = $501 << $464;
      $503 = $$16 >>> $478;
      $504 = $503 << $475;
      $505 = $504 | $502;
      $506 = $$1 >>> $489;
      $507 = $506 << $486;
      $508 = $505 | $509;
      $510 = $508 | $507;
      $511 = $510&65535;
      HEAP16[$pixel17$1>>1] = $511;
      $512 = ((($pixel17$1)) + 2|0);
      $2428 = $2432;$2429 = $2433;$2430 = $2434;$2431 = $2435;$516 = $463;$518 = $2435;$521 = $2434;$528 = $474;$530 = $2433;$533 = $2432;$540 = $485;$542 = $487;$545 = $490;$565 = $509;$n18$1 = $n18$0;$pixel17$2 = $512;
      label = 24;
      continue;
     }
     else if ((label|0) == 24) {
      label = 0;
      $513 = HEAP16[$pixel17$2>>1]|0;
      $514 = $513&65535;
      $515 = $514 & $516;
      $517 = $518&255;
      $519 = $515 >>> $517;
      $520 = $521&255;
      $522 = (125924 + ($520<<2)|0);
      $523 = HEAP32[$522>>2]|0;
      $524 = (($523) + ($519)|0);
      $525 = HEAP8[$524>>0]|0;
      $526 = $525&255;
      $527 = $528 & $514;
      $529 = $530&255;
      $531 = $527 >>> $529;
      $532 = $533&255;
      $534 = (125924 + ($532<<2)|0);
      $535 = HEAP32[$534>>2]|0;
      $536 = (($535) + ($531)|0);
      $537 = HEAP8[$536>>0]|0;
      $538 = $537&255;
      $539 = $540 & $514;
      $541 = $542&255;
      $543 = $539 >>> $541;
      $544 = $545&255;
      $546 = (125924 + ($544<<2)|0);
      $547 = HEAP32[$546>>2]|0;
      $548 = (($547) + ($543)|0);
      $549 = HEAP8[$548>>0]|0;
      $550 = $549&255;
      $551 = (($526) + ($377))|0;
      $552 = ($551>>>0)>(255);
      $$2 = $552 ? 255 : $551;
      $553 = (($538) + ($378))|0;
      $554 = ($553>>>0)>(255);
      $$17 = $554 ? 255 : $553;
      $555 = (($550) + ($379))|0;
      $556 = ($555>>>0)>(255);
      $$3 = $556 ? 255 : $555;
      $557 = $$2 >>> $520;
      $558 = $557 << $517;
      $559 = $$17 >>> $532;
      $560 = $559 << $529;
      $561 = $560 | $558;
      $562 = $$3 >>> $544;
      $563 = $562 << $541;
      $564 = $561 | $565;
      $566 = $564 | $563;
      $567 = $566&65535;
      HEAP16[$pixel17$2>>1] = $567;
      $568 = ((($pixel17$2)) + 2|0);
      $2424 = $2428;$2425 = $2429;$2426 = $2430;$2427 = $2431;$572 = $516;$574 = $2431;$577 = $2430;$584 = $528;$586 = $2429;$589 = $2428;$596 = $540;$598 = $542;$601 = $545;$621 = $565;$n18$2 = $n18$1;$pixel17$3 = $568;
      label = 25;
      continue;
     }
     else if ((label|0) == 25) {
      label = 0;
      $569 = HEAP16[$pixel17$3>>1]|0;
      $570 = $569&65535;
      $571 = $570 & $572;
      $573 = $574&255;
      $575 = $571 >>> $573;
      $576 = $577&255;
      $578 = (125924 + ($576<<2)|0);
      $579 = HEAP32[$578>>2]|0;
      $580 = (($579) + ($575)|0);
      $581 = HEAP8[$580>>0]|0;
      $582 = $581&255;
      $583 = $584 & $570;
      $585 = $586&255;
      $587 = $583 >>> $585;
      $588 = $589&255;
      $590 = (125924 + ($588<<2)|0);
      $591 = HEAP32[$590>>2]|0;
      $592 = (($591) + ($587)|0);
      $593 = HEAP8[$592>>0]|0;
      $594 = $593&255;
      $595 = $596 & $570;
      $597 = $598&255;
      $599 = $595 >>> $597;
      $600 = $601&255;
      $602 = (125924 + ($600<<2)|0);
      $603 = HEAP32[$602>>2]|0;
      $604 = (($603) + ($599)|0);
      $605 = HEAP8[$604>>0]|0;
      $606 = $605&255;
      $607 = (($582) + ($377))|0;
      $608 = ($607>>>0)>(255);
      $$4 = $608 ? 255 : $607;
      $609 = (($594) + ($378))|0;
      $610 = ($609>>>0)>(255);
      $$18 = $610 ? 255 : $609;
      $611 = (($606) + ($379))|0;
      $612 = ($611>>>0)>(255);
      $$5 = $612 ? 255 : $611;
      $613 = $$4 >>> $576;
      $614 = $613 << $573;
      $615 = $$18 >>> $588;
      $616 = $615 << $585;
      $617 = $616 | $614;
      $618 = $$5 >>> $600;
      $619 = $618 << $597;
      $620 = $617 | $621;
      $622 = $620 | $619;
      $623 = $622&65535;
      HEAP16[$pixel17$3>>1] = $623;
      $624 = ((($pixel17$3)) + 2|0);
      $628 = $572;$630 = $2427;$633 = $2426;$640 = $584;$642 = $2425;$645 = $2424;$652 = $596;$654 = $598;$657 = $601;$677 = $621;$n18$3 = $n18$2;$pixel17$4 = $624;
      label = 26;
      continue;
     }
     else if ((label|0) == 26) {
      label = 0;
      $625 = HEAP16[$pixel17$4>>1]|0;
      $626 = $625&65535;
      $627 = $626 & $628;
      $629 = $630&255;
      $631 = $627 >>> $629;
      $632 = $633&255;
      $634 = (125924 + ($632<<2)|0);
      $635 = HEAP32[$634>>2]|0;
      $636 = (($635) + ($631)|0);
      $637 = HEAP8[$636>>0]|0;
      $638 = $637&255;
      $639 = $640 & $626;
      $641 = $642&255;
      $643 = $639 >>> $641;
      $644 = $645&255;
      $646 = (125924 + ($644<<2)|0);
      $647 = HEAP32[$646>>2]|0;
      $648 = (($647) + ($643)|0);
      $649 = HEAP8[$648>>0]|0;
      $650 = $649&255;
      $651 = $652 & $626;
      $653 = $654&255;
      $655 = $651 >>> $653;
      $656 = $657&255;
      $658 = (125924 + ($656<<2)|0);
      $659 = HEAP32[$658>>2]|0;
      $660 = (($659) + ($655)|0);
      $661 = HEAP8[$660>>0]|0;
      $662 = $661&255;
      $663 = (($638) + ($377))|0;
      $664 = ($663>>>0)>(255);
      $$6 = $664 ? 255 : $663;
      $665 = (($650) + ($378))|0;
      $666 = ($665>>>0)>(255);
      $$19 = $666 ? 255 : $665;
      $667 = (($662) + ($379))|0;
      $668 = ($667>>>0)>(255);
      $$7 = $668 ? 255 : $667;
      $669 = $$6 >>> $632;
      $670 = $669 << $629;
      $671 = $$19 >>> $644;
      $672 = $671 << $641;
      $673 = $672 | $670;
      $674 = $$7 >>> $656;
      $675 = $674 << $653;
      $676 = $673 | $677;
      $678 = $676 | $675;
      $679 = $678&65535;
      HEAP16[$pixel17$4>>1] = $679;
      $680 = ((($pixel17$4)) + 2|0);
      $681 = (($n18$3) + -1)|0;
      $682 = ($n18$3|0)>(1);
      if ($682) {
       $2432 = $645;$2433 = $642;$2434 = $633;$2435 = $630;$463 = $628;$465 = $630;$468 = $633;$470 = $635;$474 = $640;$476 = $642;$479 = $645;$481 = $647;$485 = $652;$487 = $654;$490 = $657;$492 = $659;$509 = $677;$n18$0 = $681;$pixel17$1 = $680;
       label = 23;
       continue;
      } else {
       break;
      }
     }
    }
    $683 = (($680) + ($359<<1)|0);
    $684 = ($381|0)==(0);
    if ($684) {
     $$0 = 0;
     label = 109;
     break;
    } else {
     $$in50 = $381;$pixel17$034 = $683;
    }
   }
   if ((label|0) == 27) {
    // unreachable;
   }
   else if ((label|0) == 109) {
    return ($$0|0);
   }
   break;
  }
  case 4:  {
   $685 = ((($rect)) + 8|0);
   $686 = HEAP32[$685>>2]|0;
   $687 = ((($rect)) + 12|0);
   $688 = HEAP32[$687>>2]|0;
   $689 = ((($dst)) + 16|0);
   $690 = HEAP32[$689>>2]|0;
   $691 = (($690|0) / 2)&-1;
   $692 = (($691) - ($686))|0;
   $693 = ($688|0)==(0);
   if ($693) {
    $$0 = 0;
    return ($$0|0);
   }
   $694 = ((($dst)) + 20|0);
   $695 = HEAP32[$694>>2]|0;
   $696 = ((($rect)) + 4|0);
   $697 = HEAP32[$696>>2]|0;
   $698 = Math_imul($697, $691)|0;
   $699 = (($695) + ($698<<1)|0);
   $700 = HEAP32[$rect>>2]|0;
   $701 = (($699) + ($700<<1)|0);
   $702 = (($686) + 3)|0;
   $703 = (($702|0) / 4)&-1;
   $704 = $686 & 3;
   $705 = ((($1)) + 12|0);
   $706 = ((($1)) + 32|0);
   $707 = ((($1)) + 28|0);
   $708 = ((($1)) + 16|0);
   $709 = ((($1)) + 20|0);
   $710 = $r&255;
   $711 = $g&255;
   $712 = $b&255;
   $713 = ((($1)) + 24|0);
   $$in49 = $688;$pixel39$036 = $701;
   L73: while(1) {
    $714 = (($$in49) + -1)|0;
    switch ($704|0) {
    case 0:  {
     $$pre133 = HEAP32[$705>>2]|0;
     $771 = HEAP16[$706>>1]|0;
     $772 = HEAP32[$706>>2]|0;
     $773 = $772&65535;
     $774 = $772&255;
     $775 = HEAP16[$707>>1]|0;
     $776 = HEAP32[$707>>2]|0;
     $777 = $776&65535;
     $778 = $776&255;
     $$phi$trans$insert136 = $776 & 255;
     $$phi$trans$insert137 = (125924 + ($$phi$trans$insert136<<2)|0);
     $$pre138 = HEAP32[$$phi$trans$insert137>>2]|0;
     $$pre139 = HEAP32[$708>>2]|0;
     $779 = ($773&65535) >>> 8;
     $780 = $779&255;
     $781 = ($777&65535) >>> 8;
     $782 = $781&255;
     $$phi$trans$insert140 = $781&65535;
     $$phi$trans$insert141 = (125924 + ($$phi$trans$insert140<<2)|0);
     $$pre142 = HEAP32[$$phi$trans$insert141>>2]|0;
     $$pre143 = HEAP32[$709>>2]|0;
     $783 = $772 >>> 16;
     $784 = $783&255;
     $785 = $776 >>> 16;
     $786 = $785&255;
     $$phi$trans$insert144 = $785 & 255;
     $$phi$trans$insert145 = (125924 + ($$phi$trans$insert144<<2)|0);
     $$pre146 = HEAP32[$$phi$trans$insert145>>2]|0;
     $$pre147 = HEAP32[$713>>2]|0;
     $787 = $772&255;
     $788 = $776&255;
     $789 = ($771&65535) >>> 8;
     $790 = $789&255;
     $791 = ($775&65535) >>> 8;
     $792 = $791&255;
     $2444 = $792;$2445 = $790;$2446 = $788;$2447 = $787;$796 = $$pre133;$798 = $774;$801 = $778;$803 = $$pre138;$807 = $$pre139;$809 = $780;$812 = $782;$814 = $$pre142;$818 = $$pre143;$820 = $784;$823 = $786;$825 = $$pre146;$842 = $$pre147;$n40$0 = $703;$pixel39$1 = $pixel39$036;
     label = 36;
     break;
    }
    case 3:  {
     $$pre149 = HEAP32[$705>>2]|0;
     $749 = HEAP16[$706>>1]|0;
     $750 = HEAP32[$706>>2]|0;
     $751 = $750&65535;
     $752 = $750&255;
     $753 = HEAP16[$707>>1]|0;
     $754 = HEAP32[$707>>2]|0;
     $755 = $754&65535;
     $756 = $754&255;
     $$pre152 = HEAP32[$708>>2]|0;
     $757 = ($751&65535) >>> 8;
     $758 = $757&255;
     $759 = ($755&65535) >>> 8;
     $760 = $759&255;
     $$pre153 = HEAP32[$709>>2]|0;
     $761 = $750 >>> 16;
     $762 = $761&255;
     $763 = $754 >>> 16;
     $764 = $763&255;
     $$pre154 = HEAP32[$713>>2]|0;
     $765 = $750&255;
     $766 = $754&255;
     $767 = ($749&65535) >>> 8;
     $768 = $767&255;
     $769 = ($753&65535) >>> 8;
     $770 = $769&255;
     $2440 = $770;$2441 = $768;$2442 = $766;$2443 = $765;$849 = $$pre149;$851 = $752;$854 = $756;$861 = $$pre152;$863 = $758;$866 = $760;$873 = $$pre153;$875 = $762;$878 = $764;$898 = $$pre154;$n40$1 = $703;$pixel39$2 = $pixel39$036;
     label = 37;
     break;
    }
    case 2:  {
     $$pre156 = HEAP32[$705>>2]|0;
     $727 = HEAP16[$706>>1]|0;
     $728 = HEAP32[$706>>2]|0;
     $729 = $728&65535;
     $730 = $728&255;
     $731 = HEAP16[$707>>1]|0;
     $732 = HEAP32[$707>>2]|0;
     $733 = $732&65535;
     $734 = $732&255;
     $$pre159 = HEAP32[$708>>2]|0;
     $735 = ($729&65535) >>> 8;
     $736 = $735&255;
     $737 = ($733&65535) >>> 8;
     $738 = $737&255;
     $$pre160 = HEAP32[$709>>2]|0;
     $739 = $728 >>> 16;
     $740 = $739&255;
     $741 = $732 >>> 16;
     $742 = $741&255;
     $$pre161 = HEAP32[$713>>2]|0;
     $743 = $728&255;
     $744 = $732&255;
     $745 = ($727&65535) >>> 8;
     $746 = $745&255;
     $747 = ($731&65535) >>> 8;
     $748 = $747&255;
     $2436 = $748;$2437 = $746;$2438 = $744;$2439 = $743;$905 = $$pre156;$907 = $730;$910 = $734;$917 = $$pre159;$919 = $736;$922 = $738;$929 = $$pre160;$931 = $740;$934 = $742;$954 = $$pre161;$n40$2 = $703;$pixel39$3 = $pixel39$036;
     label = 38;
     break;
    }
    case 1:  {
     $$pre163 = HEAP32[$705>>2]|0;
     $$pre164 = HEAP32[$706>>2]|0;
     $715 = $$pre164&65535;
     $716 = $$pre164&255;
     $$pre165 = HEAP32[$707>>2]|0;
     $717 = $$pre165&65535;
     $718 = $$pre165&255;
     $$pre166 = HEAP32[$708>>2]|0;
     $719 = ($715&65535) >>> 8;
     $720 = $719&255;
     $721 = ($717&65535) >>> 8;
     $722 = $721&255;
     $$pre167 = HEAP32[$709>>2]|0;
     $723 = $$pre164 >>> 16;
     $724 = $723&255;
     $725 = $$pre165 >>> 16;
     $726 = $725&255;
     $$pre168 = HEAP32[$713>>2]|0;
     $1010 = $$pre168;$961 = $$pre163;$963 = $716;$966 = $718;$973 = $$pre166;$975 = $720;$978 = $722;$985 = $$pre167;$987 = $724;$990 = $726;$n40$3 = $703;$pixel39$4 = $pixel39$036;
     label = 39;
     break;
    }
    default: {
     label = 40;
     break L73;
    }
    }
    while(1) {
     if ((label|0) == 36) {
      label = 0;
      $793 = HEAP16[$pixel39$1>>1]|0;
      $794 = $793&65535;
      $795 = $794 & $796;
      $797 = $798&255;
      $799 = $795 >>> $797;
      $800 = $801&255;
      $802 = (($803) + ($799)|0);
      $804 = HEAP8[$802>>0]|0;
      $805 = $804&255;
      $806 = $807 & $794;
      $808 = $809&255;
      $810 = $806 >>> $808;
      $811 = $812&255;
      $813 = (($814) + ($810)|0);
      $815 = HEAP8[$813>>0]|0;
      $816 = $815&255;
      $817 = $818 & $794;
      $819 = $820&255;
      $821 = $817 >>> $819;
      $822 = $823&255;
      $824 = (($825) + ($821)|0);
      $826 = HEAP8[$824>>0]|0;
      $827 = $826&255;
      $828 = Math_imul($805, $710)|0;
      $829 = (($828>>>0) / 255)&-1;
      $830 = Math_imul($816, $711)|0;
      $831 = (($830>>>0) / 255)&-1;
      $832 = Math_imul($827, $712)|0;
      $833 = (($832>>>0) / 255)&-1;
      $834 = $829 >>> $800;
      $835 = $834 << $797;
      $836 = $831 >>> $811;
      $837 = $836 << $808;
      $838 = $837 | $835;
      $839 = $833 >>> $822;
      $840 = $839 << $819;
      $841 = $838 | $842;
      $843 = $841 | $840;
      $844 = $843&65535;
      HEAP16[$pixel39$1>>1] = $844;
      $845 = ((($pixel39$1)) + 2|0);
      $2440 = $2444;$2441 = $2445;$2442 = $2446;$2443 = $2447;$849 = $796;$851 = $2447;$854 = $2446;$861 = $807;$863 = $2445;$866 = $2444;$873 = $818;$875 = $820;$878 = $823;$898 = $842;$n40$1 = $n40$0;$pixel39$2 = $845;
      label = 37;
      continue;
     }
     else if ((label|0) == 37) {
      label = 0;
      $846 = HEAP16[$pixel39$2>>1]|0;
      $847 = $846&65535;
      $848 = $847 & $849;
      $850 = $851&255;
      $852 = $848 >>> $850;
      $853 = $854&255;
      $855 = (125924 + ($853<<2)|0);
      $856 = HEAP32[$855>>2]|0;
      $857 = (($856) + ($852)|0);
      $858 = HEAP8[$857>>0]|0;
      $859 = $858&255;
      $860 = $861 & $847;
      $862 = $863&255;
      $864 = $860 >>> $862;
      $865 = $866&255;
      $867 = (125924 + ($865<<2)|0);
      $868 = HEAP32[$867>>2]|0;
      $869 = (($868) + ($864)|0);
      $870 = HEAP8[$869>>0]|0;
      $871 = $870&255;
      $872 = $873 & $847;
      $874 = $875&255;
      $876 = $872 >>> $874;
      $877 = $878&255;
      $879 = (125924 + ($877<<2)|0);
      $880 = HEAP32[$879>>2]|0;
      $881 = (($880) + ($876)|0);
      $882 = HEAP8[$881>>0]|0;
      $883 = $882&255;
      $884 = Math_imul($859, $710)|0;
      $885 = (($884>>>0) / 255)&-1;
      $886 = Math_imul($871, $711)|0;
      $887 = (($886>>>0) / 255)&-1;
      $888 = Math_imul($883, $712)|0;
      $889 = (($888>>>0) / 255)&-1;
      $890 = $885 >>> $853;
      $891 = $890 << $850;
      $892 = $887 >>> $865;
      $893 = $892 << $862;
      $894 = $893 | $891;
      $895 = $889 >>> $877;
      $896 = $895 << $874;
      $897 = $894 | $898;
      $899 = $897 | $896;
      $900 = $899&65535;
      HEAP16[$pixel39$2>>1] = $900;
      $901 = ((($pixel39$2)) + 2|0);
      $2436 = $2440;$2437 = $2441;$2438 = $2442;$2439 = $2443;$905 = $849;$907 = $2443;$910 = $2442;$917 = $861;$919 = $2441;$922 = $2440;$929 = $873;$931 = $875;$934 = $878;$954 = $898;$n40$2 = $n40$1;$pixel39$3 = $901;
      label = 38;
      continue;
     }
     else if ((label|0) == 38) {
      label = 0;
      $902 = HEAP16[$pixel39$3>>1]|0;
      $903 = $902&65535;
      $904 = $903 & $905;
      $906 = $907&255;
      $908 = $904 >>> $906;
      $909 = $910&255;
      $911 = (125924 + ($909<<2)|0);
      $912 = HEAP32[$911>>2]|0;
      $913 = (($912) + ($908)|0);
      $914 = HEAP8[$913>>0]|0;
      $915 = $914&255;
      $916 = $917 & $903;
      $918 = $919&255;
      $920 = $916 >>> $918;
      $921 = $922&255;
      $923 = (125924 + ($921<<2)|0);
      $924 = HEAP32[$923>>2]|0;
      $925 = (($924) + ($920)|0);
      $926 = HEAP8[$925>>0]|0;
      $927 = $926&255;
      $928 = $929 & $903;
      $930 = $931&255;
      $932 = $928 >>> $930;
      $933 = $934&255;
      $935 = (125924 + ($933<<2)|0);
      $936 = HEAP32[$935>>2]|0;
      $937 = (($936) + ($932)|0);
      $938 = HEAP8[$937>>0]|0;
      $939 = $938&255;
      $940 = Math_imul($915, $710)|0;
      $941 = (($940>>>0) / 255)&-1;
      $942 = Math_imul($927, $711)|0;
      $943 = (($942>>>0) / 255)&-1;
      $944 = Math_imul($939, $712)|0;
      $945 = (($944>>>0) / 255)&-1;
      $946 = $941 >>> $909;
      $947 = $946 << $906;
      $948 = $943 >>> $921;
      $949 = $948 << $918;
      $950 = $949 | $947;
      $951 = $945 >>> $933;
      $952 = $951 << $930;
      $953 = $950 | $954;
      $955 = $953 | $952;
      $956 = $955&65535;
      HEAP16[$pixel39$3>>1] = $956;
      $957 = ((($pixel39$3)) + 2|0);
      $1010 = $954;$961 = $905;$963 = $2439;$966 = $2438;$973 = $917;$975 = $2437;$978 = $2436;$985 = $929;$987 = $931;$990 = $934;$n40$3 = $n40$2;$pixel39$4 = $957;
      label = 39;
      continue;
     }
     else if ((label|0) == 39) {
      label = 0;
      $958 = HEAP16[$pixel39$4>>1]|0;
      $959 = $958&65535;
      $960 = $959 & $961;
      $962 = $963&255;
      $964 = $960 >>> $962;
      $965 = $966&255;
      $967 = (125924 + ($965<<2)|0);
      $968 = HEAP32[$967>>2]|0;
      $969 = (($968) + ($964)|0);
      $970 = HEAP8[$969>>0]|0;
      $971 = $970&255;
      $972 = $973 & $959;
      $974 = $975&255;
      $976 = $972 >>> $974;
      $977 = $978&255;
      $979 = (125924 + ($977<<2)|0);
      $980 = HEAP32[$979>>2]|0;
      $981 = (($980) + ($976)|0);
      $982 = HEAP8[$981>>0]|0;
      $983 = $982&255;
      $984 = $985 & $959;
      $986 = $987&255;
      $988 = $984 >>> $986;
      $989 = $990&255;
      $991 = (125924 + ($989<<2)|0);
      $992 = HEAP32[$991>>2]|0;
      $993 = (($992) + ($988)|0);
      $994 = HEAP8[$993>>0]|0;
      $995 = $994&255;
      $996 = Math_imul($971, $710)|0;
      $997 = (($996>>>0) / 255)&-1;
      $998 = Math_imul($983, $711)|0;
      $999 = (($998>>>0) / 255)&-1;
      $1000 = Math_imul($995, $712)|0;
      $1001 = (($1000>>>0) / 255)&-1;
      $1002 = $997 >>> $965;
      $1003 = $1002 << $962;
      $1004 = $999 >>> $977;
      $1005 = $1004 << $974;
      $1006 = $1005 | $1003;
      $1007 = $1001 >>> $989;
      $1008 = $1007 << $986;
      $1009 = $1006 | $1010;
      $1011 = $1009 | $1008;
      $1012 = $1011&65535;
      HEAP16[$pixel39$4>>1] = $1012;
      $1013 = ((($pixel39$4)) + 2|0);
      $1014 = (($n40$3) + -1)|0;
      $1015 = ($n40$3|0)>(1);
      if ($1015) {
       $2444 = $978;$2445 = $975;$2446 = $966;$2447 = $963;$796 = $961;$798 = $963;$801 = $966;$803 = $968;$807 = $973;$809 = $975;$812 = $978;$814 = $980;$818 = $985;$820 = $987;$823 = $990;$825 = $992;$842 = $1010;$n40$0 = $1014;$pixel39$1 = $1013;
       label = 36;
       continue;
      } else {
       break;
      }
     }
    }
    $1016 = (($1013) + ($692<<1)|0);
    $1017 = ($714|0)==(0);
    if ($1017) {
     $$0 = 0;
     label = 109;
     break;
    } else {
     $$in49 = $714;$pixel39$036 = $1016;
    }
   }
   if ((label|0) == 40) {
    // unreachable;
   }
   else if ((label|0) == 109) {
    return ($$0|0);
   }
   break;
  }
  default: {
   $1018 = ((($rect)) + 8|0);
   $1019 = HEAP32[$1018>>2]|0;
   $1020 = ((($rect)) + 12|0);
   $1021 = HEAP32[$1020>>2]|0;
   $1022 = ((($dst)) + 16|0);
   $1023 = HEAP32[$1022>>2]|0;
   $1024 = (($1023|0) / 2)&-1;
   $1025 = (($1024) - ($1019))|0;
   $1026 = ($1021|0)==(0);
   if ($1026) {
    $$0 = 0;
    return ($$0|0);
   }
   $1027 = ((($dst)) + 20|0);
   $1028 = HEAP32[$1027>>2]|0;
   $1029 = ((($rect)) + 4|0);
   $1030 = HEAP32[$1029>>2]|0;
   $1031 = Math_imul($1030, $1024)|0;
   $1032 = (($1028) + ($1031<<1)|0);
   $1033 = HEAP32[$rect>>2]|0;
   $1034 = (($1032) + ($1033<<1)|0);
   $1035 = (($1019) + 3)|0;
   $1036 = (($1035|0) / 4)&-1;
   $1037 = $1019 & 3;
   $1038 = $r&255;
   $1039 = $g&255;
   $1040 = $b&255;
   $1041 = ((($1)) + 28|0);
   $1042 = ((($1)) + 32|0);
   $1043 = ((($1)) + 24|0);
   $$in52 = $1021;$pixel61$031 = $1034;
   L10: while(1) {
    $1044 = (($$in52) + -1)|0;
    switch ($1037|0) {
    case 0:  {
     $1101 = HEAP16[$1041>>1]|0;
     $1102 = HEAP32[$1041>>2]|0;
     $1103 = $1102&65535;
     $1104 = $1102&255;
     $1105 = HEAP16[$1042>>1]|0;
     $1106 = HEAP32[$1042>>2]|0;
     $1107 = $1106&65535;
     $1108 = $1106&255;
     $1109 = ($1103&65535) >>> 8;
     $1110 = $1109&255;
     $1111 = ($1107&65535) >>> 8;
     $1112 = $1111&255;
     $1113 = $1102 >>> 16;
     $1114 = $1113&255;
     $1115 = $1106 >>> 16;
     $1116 = $1115&255;
     $$pre246 = HEAP32[$1043>>2]|0;
     $1117 = $1102&255;
     $1118 = $1106&255;
     $1119 = ($1101&65535) >>> 8;
     $1120 = $1119&255;
     $1121 = ($1105&65535) >>> 8;
     $1122 = $1121&255;
     $1124 = $1104;$1127 = $1108;$1130 = $1110;$1133 = $1112;$1137 = $1114;$1140 = $1116;$1143 = $$pre246;$2456 = $1122;$2457 = $1120;$2458 = $1118;$2459 = $1117;$n62$0 = $1036;$pixel61$1 = $pixel61$031;
     label = 49;
     break;
    }
    case 3:  {
     $1079 = HEAP16[$1041>>1]|0;
     $1080 = HEAP32[$1041>>2]|0;
     $1081 = $1080&65535;
     $1082 = $1080&255;
     $1083 = HEAP16[$1042>>1]|0;
     $1084 = HEAP32[$1042>>2]|0;
     $1085 = $1084&65535;
     $1086 = $1084&255;
     $1087 = ($1081&65535) >>> 8;
     $1088 = $1087&255;
     $1089 = ($1085&65535) >>> 8;
     $1090 = $1089&255;
     $1091 = $1080 >>> 16;
     $1092 = $1091&255;
     $1093 = $1084 >>> 16;
     $1094 = $1093&255;
     $$pre250 = HEAP32[$1043>>2]|0;
     $1095 = $1080&255;
     $1096 = $1084&255;
     $1097 = ($1079&65535) >>> 8;
     $1098 = $1097&255;
     $1099 = ($1083&65535) >>> 8;
     $1100 = $1099&255;
     $1148 = $1082;$1151 = $1086;$1154 = $1088;$1157 = $1090;$1161 = $1092;$1164 = $1094;$1167 = $$pre250;$2452 = $1100;$2453 = $1098;$2454 = $1096;$2455 = $1095;$n62$1 = $1036;$pixel61$2 = $pixel61$031;
     label = 50;
     break;
    }
    case 2:  {
     $1057 = HEAP16[$1041>>1]|0;
     $1058 = HEAP32[$1041>>2]|0;
     $1059 = $1058&65535;
     $1060 = $1058&255;
     $1061 = HEAP16[$1042>>1]|0;
     $1062 = HEAP32[$1042>>2]|0;
     $1063 = $1062&65535;
     $1064 = $1062&255;
     $1065 = ($1059&65535) >>> 8;
     $1066 = $1065&255;
     $1067 = ($1063&65535) >>> 8;
     $1068 = $1067&255;
     $1069 = $1058 >>> 16;
     $1070 = $1069&255;
     $1071 = $1062 >>> 16;
     $1072 = $1071&255;
     $$pre254 = HEAP32[$1043>>2]|0;
     $1073 = $1058&255;
     $1074 = $1062&255;
     $1075 = ($1057&65535) >>> 8;
     $1076 = $1075&255;
     $1077 = ($1061&65535) >>> 8;
     $1078 = $1077&255;
     $1172 = $1060;$1175 = $1064;$1178 = $1066;$1181 = $1068;$1185 = $1070;$1188 = $1072;$1191 = $$pre254;$2448 = $1078;$2449 = $1076;$2450 = $1074;$2451 = $1073;$n62$2 = $1036;$pixel61$3 = $pixel61$031;
     label = 51;
     break;
    }
    case 1:  {
     $$pre256 = HEAP32[$1041>>2]|0;
     $1045 = $$pre256&65535;
     $1046 = $$pre256&255;
     $$pre257 = HEAP32[$1042>>2]|0;
     $1047 = $$pre257&65535;
     $1048 = $$pre257&255;
     $1049 = ($1045&65535) >>> 8;
     $1050 = $1049&255;
     $1051 = ($1047&65535) >>> 8;
     $1052 = $1051&255;
     $1053 = $$pre256 >>> 16;
     $1054 = $1053&255;
     $1055 = $$pre257 >>> 16;
     $1056 = $1055&255;
     $$pre258 = HEAP32[$1043>>2]|0;
     $1196 = $1046;$1199 = $1048;$1202 = $1050;$1205 = $1052;$1209 = $1054;$1212 = $1056;$1215 = $$pre258;$n62$3 = $1036;$pixel61$4 = $pixel61$031;
     label = 52;
     break;
    }
    default: {
     label = 53;
     break L10;
    }
    }
    while(1) {
     if ((label|0) == 49) {
      label = 0;
      $1123 = $1124&255;
      $1125 = $1038 >>> $1123;
      $1126 = $1127&255;
      $1128 = $1125 << $1126;
      $1129 = $1130&255;
      $1131 = $1039 >>> $1129;
      $1132 = $1133&255;
      $1134 = $1131 << $1132;
      $1135 = $1134 | $1128;
      $1136 = $1137&255;
      $1138 = $1040 >>> $1136;
      $1139 = $1140&255;
      $1141 = $1138 << $1139;
      $1142 = $1135 | $1143;
      $1144 = $1142 | $1141;
      $1145 = $1144&65535;
      HEAP16[$pixel61$1>>1] = $1145;
      $1146 = ((($pixel61$1)) + 2|0);
      $1148 = $2459;$1151 = $2458;$1154 = $2457;$1157 = $2456;$1161 = $1137;$1164 = $1140;$1167 = $1143;$2452 = $2456;$2453 = $2457;$2454 = $2458;$2455 = $2459;$n62$1 = $n62$0;$pixel61$2 = $1146;
      label = 50;
      continue;
     }
     else if ((label|0) == 50) {
      label = 0;
      $1147 = $1148&255;
      $1149 = $1038 >>> $1147;
      $1150 = $1151&255;
      $1152 = $1149 << $1150;
      $1153 = $1154&255;
      $1155 = $1039 >>> $1153;
      $1156 = $1157&255;
      $1158 = $1155 << $1156;
      $1159 = $1158 | $1152;
      $1160 = $1161&255;
      $1162 = $1040 >>> $1160;
      $1163 = $1164&255;
      $1165 = $1162 << $1163;
      $1166 = $1159 | $1167;
      $1168 = $1166 | $1165;
      $1169 = $1168&65535;
      HEAP16[$pixel61$2>>1] = $1169;
      $1170 = ((($pixel61$2)) + 2|0);
      $1172 = $2455;$1175 = $2454;$1178 = $2453;$1181 = $2452;$1185 = $1161;$1188 = $1164;$1191 = $1167;$2448 = $2452;$2449 = $2453;$2450 = $2454;$2451 = $2455;$n62$2 = $n62$1;$pixel61$3 = $1170;
      label = 51;
      continue;
     }
     else if ((label|0) == 51) {
      label = 0;
      $1171 = $1172&255;
      $1173 = $1038 >>> $1171;
      $1174 = $1175&255;
      $1176 = $1173 << $1174;
      $1177 = $1178&255;
      $1179 = $1039 >>> $1177;
      $1180 = $1181&255;
      $1182 = $1179 << $1180;
      $1183 = $1182 | $1176;
      $1184 = $1185&255;
      $1186 = $1040 >>> $1184;
      $1187 = $1188&255;
      $1189 = $1186 << $1187;
      $1190 = $1183 | $1191;
      $1192 = $1190 | $1189;
      $1193 = $1192&65535;
      HEAP16[$pixel61$3>>1] = $1193;
      $1194 = ((($pixel61$3)) + 2|0);
      $1196 = $2451;$1199 = $2450;$1202 = $2449;$1205 = $2448;$1209 = $1185;$1212 = $1188;$1215 = $1191;$n62$3 = $n62$2;$pixel61$4 = $1194;
      label = 52;
      continue;
     }
     else if ((label|0) == 52) {
      label = 0;
      $1195 = $1196&255;
      $1197 = $1038 >>> $1195;
      $1198 = $1199&255;
      $1200 = $1197 << $1198;
      $1201 = $1202&255;
      $1203 = $1039 >>> $1201;
      $1204 = $1205&255;
      $1206 = $1203 << $1204;
      $1207 = $1206 | $1200;
      $1208 = $1209&255;
      $1210 = $1040 >>> $1208;
      $1211 = $1212&255;
      $1213 = $1210 << $1211;
      $1214 = $1207 | $1215;
      $1216 = $1214 | $1213;
      $1217 = $1216&65535;
      HEAP16[$pixel61$4>>1] = $1217;
      $1218 = ((($pixel61$4)) + 2|0);
      $1219 = (($n62$3) + -1)|0;
      $1220 = ($n62$3|0)>(1);
      if ($1220) {
       $1124 = $1196;$1127 = $1199;$1130 = $1202;$1133 = $1205;$1137 = $1209;$1140 = $1212;$1143 = $1215;$2456 = $1205;$2457 = $1202;$2458 = $1199;$2459 = $1196;$n62$0 = $1219;$pixel61$1 = $1218;
       label = 49;
       continue;
      } else {
       break;
      }
     }
    }
    $1221 = (($1218) + ($1025<<1)|0);
    $1222 = ($1044|0)==(0);
    if ($1222) {
     $$0 = 0;
     label = 109;
     break;
    } else {
     $$in52 = $1044;$pixel61$031 = $1221;
    }
   }
   if ((label|0) == 53) {
    // unreachable;
   }
   else if ((label|0) == 109) {
    return ($$0|0);
   }
  }
  }
  break;
 }
 case 4:  {
  switch ($blendMode|0) {
  case 1:  {
   $1223 = ((($rect)) + 8|0);
   $1224 = HEAP32[$1223>>2]|0;
   $1225 = ((($rect)) + 12|0);
   $1226 = HEAP32[$1225>>2]|0;
   $1227 = ((($dst)) + 16|0);
   $1228 = HEAP32[$1227>>2]|0;
   $1229 = (($1228|0) / 4)&-1;
   $1230 = (($1229) - ($1224))|0;
   $1231 = ($1226|0)==(0);
   if ($1231) {
    $$0 = 0;
    return ($$0|0);
   }
   $1232 = ((($dst)) + 20|0);
   $1233 = HEAP32[$1232>>2]|0;
   $1234 = ((($rect)) + 4|0);
   $1235 = HEAP32[$1234>>2]|0;
   $1236 = Math_imul($1235, $1229)|0;
   $1237 = (($1233) + ($1236<<2)|0);
   $1238 = HEAP32[$rect>>2]|0;
   $1239 = (($1237) + ($1238<<2)|0);
   $1240 = (($1224) + 3)|0;
   $1241 = (($1240|0) / 4)&-1;
   $1242 = $1224 & 3;
   $1243 = ((($1)) + 12|0);
   $1244 = ((($1)) + 32|0);
   $1245 = ((($1)) + 28|0);
   $1246 = ((($1)) + 16|0);
   $1247 = ((($1)) + 20|0);
   $1248 = $r&255;
   $1249 = $g&255;
   $1250 = $b&255;
   $1251 = ((($1)) + 24|0);
   $$in47 = $1226;$pixel83$040 = $1239;
   L117: while(1) {
    $1252 = (($$in47) + -1)|0;
    switch ($1242|0) {
    case 0:  {
     $1309 = HEAP16[$1244>>1]|0;
     $1310 = HEAP32[$1244>>2]|0;
     $1311 = $1310&65535;
     $1312 = $1310&255;
     $1313 = HEAP16[$1245>>1]|0;
     $1314 = HEAP32[$1245>>2]|0;
     $1315 = $1314&65535;
     $1316 = $1314&255;
     $$phi$trans$insert102 = $1314 & 255;
     $$phi$trans$insert103 = (125924 + ($$phi$trans$insert102<<2)|0);
     $$pre104 = HEAP32[$$phi$trans$insert103>>2]|0;
     $1317 = ($1311&65535) >>> 8;
     $1318 = $1317&255;
     $1319 = ($1315&65535) >>> 8;
     $1320 = $1319&255;
     $$phi$trans$insert105 = $1319&65535;
     $$phi$trans$insert106 = (125924 + ($$phi$trans$insert105<<2)|0);
     $$pre107 = HEAP32[$$phi$trans$insert106>>2]|0;
     $1321 = $1310 >>> 16;
     $1322 = $1321&255;
     $1323 = $1314 >>> 16;
     $1324 = $1323&255;
     $$phi$trans$insert108 = $1323 & 255;
     $$phi$trans$insert109 = (125924 + ($$phi$trans$insert108<<2)|0);
     $$pre110 = HEAP32[$$phi$trans$insert109>>2]|0;
     $1325 = $1310&255;
     $1326 = $1314&255;
     $1327 = ($1309&65535) >>> 8;
     $1328 = $1327&255;
     $1329 = ($1313&65535) >>> 8;
     $1330 = $1329&255;
     $1335 = $1312;$1338 = $1316;$1340 = $$pre104;$1346 = $1318;$1349 = $1320;$1351 = $$pre107;$1357 = $1322;$1360 = $1324;$1362 = $$pre110;$2468 = $1330;$2469 = $1328;$2470 = $1326;$2471 = $1325;$n84$0 = $1241;$pixel83$1 = $pixel83$040;
     label = 63;
     break;
    }
    case 3:  {
     $1287 = HEAP16[$1244>>1]|0;
     $1288 = HEAP32[$1244>>2]|0;
     $1289 = $1288&65535;
     $1290 = $1288&255;
     $1291 = HEAP16[$1245>>1]|0;
     $1292 = HEAP32[$1245>>2]|0;
     $1293 = $1292&65535;
     $1294 = $1292&255;
     $1295 = ($1289&65535) >>> 8;
     $1296 = $1295&255;
     $1297 = ($1293&65535) >>> 8;
     $1298 = $1297&255;
     $1299 = $1288 >>> 16;
     $1300 = $1299&255;
     $1301 = $1292 >>> 16;
     $1302 = $1301&255;
     $1303 = $1288&255;
     $1304 = $1292&255;
     $1305 = ($1287&65535) >>> 8;
     $1306 = $1305&255;
     $1307 = ($1291&65535) >>> 8;
     $1308 = $1307&255;
     $1389 = $1290;$1392 = $1294;$1401 = $1296;$1404 = $1298;$1413 = $1300;$1416 = $1302;$2464 = $1308;$2465 = $1306;$2466 = $1304;$2467 = $1303;$n84$1 = $1241;$pixel83$2 = $pixel83$040;
     label = 64;
     break;
    }
    case 2:  {
     $1265 = HEAP16[$1244>>1]|0;
     $1266 = HEAP32[$1244>>2]|0;
     $1267 = $1266&65535;
     $1268 = $1266&255;
     $1269 = HEAP16[$1245>>1]|0;
     $1270 = HEAP32[$1245>>2]|0;
     $1271 = $1270&65535;
     $1272 = $1270&255;
     $1273 = ($1267&65535) >>> 8;
     $1274 = $1273&255;
     $1275 = ($1271&65535) >>> 8;
     $1276 = $1275&255;
     $1277 = $1266 >>> 16;
     $1278 = $1277&255;
     $1279 = $1270 >>> 16;
     $1280 = $1279&255;
     $1281 = $1266&255;
     $1282 = $1270&255;
     $1283 = ($1265&65535) >>> 8;
     $1284 = $1283&255;
     $1285 = ($1269&65535) >>> 8;
     $1286 = $1285&255;
     $1446 = $1268;$1449 = $1272;$1458 = $1274;$1461 = $1276;$1470 = $1278;$1473 = $1280;$2460 = $1286;$2461 = $1284;$2462 = $1282;$2463 = $1281;$n84$2 = $1241;$pixel83$3 = $pixel83$040;
     label = 65;
     break;
    }
    case 1:  {
     $$pre118 = HEAP32[$1244>>2]|0;
     $1253 = $$pre118&65535;
     $1254 = $$pre118&255;
     $$pre119 = HEAP32[$1245>>2]|0;
     $1255 = $$pre119&65535;
     $1256 = $$pre119&255;
     $1257 = ($1253&65535) >>> 8;
     $1258 = $1257&255;
     $1259 = ($1255&65535) >>> 8;
     $1260 = $1259&255;
     $1261 = $$pre118 >>> 16;
     $1262 = $1261&255;
     $1263 = $$pre119 >>> 16;
     $1264 = $1263&255;
     $1503 = $1254;$1506 = $1256;$1515 = $1258;$1518 = $1260;$1527 = $1262;$1530 = $1264;$n84$3 = $1241;$pixel83$4 = $pixel83$040;
     label = 66;
     break;
    }
    default: {
     label = 67;
     break L117;
    }
    }
    while(1) {
     if ((label|0) == 63) {
      label = 0;
      $1331 = HEAP32[$pixel83$1>>2]|0;
      $1332 = HEAP32[$1243>>2]|0;
      $1333 = $1332 & $1331;
      $1334 = $1335&255;
      $1336 = $1333 >>> $1334;
      $1337 = $1338&255;
      $1339 = (($1340) + ($1336)|0);
      $1341 = HEAP8[$1339>>0]|0;
      $1342 = $1341&255;
      $1343 = HEAP32[$1246>>2]|0;
      $1344 = $1343 & $1331;
      $1345 = $1346&255;
      $1347 = $1344 >>> $1345;
      $1348 = $1349&255;
      $1350 = (($1351) + ($1347)|0);
      $1352 = HEAP8[$1350>>0]|0;
      $1353 = $1352&255;
      $1354 = HEAP32[$1247>>2]|0;
      $1355 = $1354 & $1331;
      $1356 = $1357&255;
      $1358 = $1355 >>> $1356;
      $1359 = $1360&255;
      $1361 = (($1362) + ($1358)|0);
      $1363 = HEAP8[$1361>>0]|0;
      $1364 = $1363&255;
      $1365 = Math_imul($1342, $3)|0;
      $1366 = (($1365>>>0) / 255)&-1;
      $1367 = (($1366) + ($1248))|0;
      $1368 = Math_imul($1353, $3)|0;
      $1369 = (($1368>>>0) / 255)&-1;
      $1370 = (($1369) + ($1249))|0;
      $1371 = Math_imul($1364, $3)|0;
      $1372 = (($1371>>>0) / 255)&-1;
      $1373 = (($1372) + ($1250))|0;
      $1374 = $1367 >>> $1337;
      $1375 = $1374 << $1334;
      $1376 = $1370 >>> $1348;
      $1377 = $1376 << $1345;
      $1378 = $1377 | $1375;
      $1379 = $1373 >>> $1359;
      $1380 = $1379 << $1356;
      $1381 = HEAP32[$1251>>2]|0;
      $1382 = $1378 | $1381;
      $1383 = $1382 | $1380;
      HEAP32[$pixel83$1>>2] = $1383;
      $1384 = ((($pixel83$1)) + 4|0);
      $1389 = $2471;$1392 = $2470;$1401 = $2469;$1404 = $2468;$1413 = $1357;$1416 = $1360;$2464 = $2468;$2465 = $2469;$2466 = $2470;$2467 = $2471;$n84$1 = $n84$0;$pixel83$2 = $1384;
      label = 64;
      continue;
     }
     else if ((label|0) == 64) {
      label = 0;
      $1385 = HEAP32[$pixel83$2>>2]|0;
      $1386 = HEAP32[$1243>>2]|0;
      $1387 = $1386 & $1385;
      $1388 = $1389&255;
      $1390 = $1387 >>> $1388;
      $1391 = $1392&255;
      $1393 = (125924 + ($1391<<2)|0);
      $1394 = HEAP32[$1393>>2]|0;
      $1395 = (($1394) + ($1390)|0);
      $1396 = HEAP8[$1395>>0]|0;
      $1397 = $1396&255;
      $1398 = HEAP32[$1246>>2]|0;
      $1399 = $1398 & $1385;
      $1400 = $1401&255;
      $1402 = $1399 >>> $1400;
      $1403 = $1404&255;
      $1405 = (125924 + ($1403<<2)|0);
      $1406 = HEAP32[$1405>>2]|0;
      $1407 = (($1406) + ($1402)|0);
      $1408 = HEAP8[$1407>>0]|0;
      $1409 = $1408&255;
      $1410 = HEAP32[$1247>>2]|0;
      $1411 = $1410 & $1385;
      $1412 = $1413&255;
      $1414 = $1411 >>> $1412;
      $1415 = $1416&255;
      $1417 = (125924 + ($1415<<2)|0);
      $1418 = HEAP32[$1417>>2]|0;
      $1419 = (($1418) + ($1414)|0);
      $1420 = HEAP8[$1419>>0]|0;
      $1421 = $1420&255;
      $1422 = Math_imul($1397, $3)|0;
      $1423 = (($1422>>>0) / 255)&-1;
      $1424 = (($1423) + ($1248))|0;
      $1425 = Math_imul($1409, $3)|0;
      $1426 = (($1425>>>0) / 255)&-1;
      $1427 = (($1426) + ($1249))|0;
      $1428 = Math_imul($1421, $3)|0;
      $1429 = (($1428>>>0) / 255)&-1;
      $1430 = (($1429) + ($1250))|0;
      $1431 = $1424 >>> $1391;
      $1432 = $1431 << $1388;
      $1433 = $1427 >>> $1403;
      $1434 = $1433 << $1400;
      $1435 = $1434 | $1432;
      $1436 = $1430 >>> $1415;
      $1437 = $1436 << $1412;
      $1438 = HEAP32[$1251>>2]|0;
      $1439 = $1435 | $1438;
      $1440 = $1439 | $1437;
      HEAP32[$pixel83$2>>2] = $1440;
      $1441 = ((($pixel83$2)) + 4|0);
      $1446 = $2467;$1449 = $2466;$1458 = $2465;$1461 = $2464;$1470 = $1413;$1473 = $1416;$2460 = $2464;$2461 = $2465;$2462 = $2466;$2463 = $2467;$n84$2 = $n84$1;$pixel83$3 = $1441;
      label = 65;
      continue;
     }
     else if ((label|0) == 65) {
      label = 0;
      $1442 = HEAP32[$pixel83$3>>2]|0;
      $1443 = HEAP32[$1243>>2]|0;
      $1444 = $1443 & $1442;
      $1445 = $1446&255;
      $1447 = $1444 >>> $1445;
      $1448 = $1449&255;
      $1450 = (125924 + ($1448<<2)|0);
      $1451 = HEAP32[$1450>>2]|0;
      $1452 = (($1451) + ($1447)|0);
      $1453 = HEAP8[$1452>>0]|0;
      $1454 = $1453&255;
      $1455 = HEAP32[$1246>>2]|0;
      $1456 = $1455 & $1442;
      $1457 = $1458&255;
      $1459 = $1456 >>> $1457;
      $1460 = $1461&255;
      $1462 = (125924 + ($1460<<2)|0);
      $1463 = HEAP32[$1462>>2]|0;
      $1464 = (($1463) + ($1459)|0);
      $1465 = HEAP8[$1464>>0]|0;
      $1466 = $1465&255;
      $1467 = HEAP32[$1247>>2]|0;
      $1468 = $1467 & $1442;
      $1469 = $1470&255;
      $1471 = $1468 >>> $1469;
      $1472 = $1473&255;
      $1474 = (125924 + ($1472<<2)|0);
      $1475 = HEAP32[$1474>>2]|0;
      $1476 = (($1475) + ($1471)|0);
      $1477 = HEAP8[$1476>>0]|0;
      $1478 = $1477&255;
      $1479 = Math_imul($1454, $3)|0;
      $1480 = (($1479>>>0) / 255)&-1;
      $1481 = (($1480) + ($1248))|0;
      $1482 = Math_imul($1466, $3)|0;
      $1483 = (($1482>>>0) / 255)&-1;
      $1484 = (($1483) + ($1249))|0;
      $1485 = Math_imul($1478, $3)|0;
      $1486 = (($1485>>>0) / 255)&-1;
      $1487 = (($1486) + ($1250))|0;
      $1488 = $1481 >>> $1448;
      $1489 = $1488 << $1445;
      $1490 = $1484 >>> $1460;
      $1491 = $1490 << $1457;
      $1492 = $1491 | $1489;
      $1493 = $1487 >>> $1472;
      $1494 = $1493 << $1469;
      $1495 = HEAP32[$1251>>2]|0;
      $1496 = $1492 | $1495;
      $1497 = $1496 | $1494;
      HEAP32[$pixel83$3>>2] = $1497;
      $1498 = ((($pixel83$3)) + 4|0);
      $1503 = $2463;$1506 = $2462;$1515 = $2461;$1518 = $2460;$1527 = $1470;$1530 = $1473;$n84$3 = $n84$2;$pixel83$4 = $1498;
      label = 66;
      continue;
     }
     else if ((label|0) == 66) {
      label = 0;
      $1499 = HEAP32[$pixel83$4>>2]|0;
      $1500 = HEAP32[$1243>>2]|0;
      $1501 = $1500 & $1499;
      $1502 = $1503&255;
      $1504 = $1501 >>> $1502;
      $1505 = $1506&255;
      $1507 = (125924 + ($1505<<2)|0);
      $1508 = HEAP32[$1507>>2]|0;
      $1509 = (($1508) + ($1504)|0);
      $1510 = HEAP8[$1509>>0]|0;
      $1511 = $1510&255;
      $1512 = HEAP32[$1246>>2]|0;
      $1513 = $1512 & $1499;
      $1514 = $1515&255;
      $1516 = $1513 >>> $1514;
      $1517 = $1518&255;
      $1519 = (125924 + ($1517<<2)|0);
      $1520 = HEAP32[$1519>>2]|0;
      $1521 = (($1520) + ($1516)|0);
      $1522 = HEAP8[$1521>>0]|0;
      $1523 = $1522&255;
      $1524 = HEAP32[$1247>>2]|0;
      $1525 = $1524 & $1499;
      $1526 = $1527&255;
      $1528 = $1525 >>> $1526;
      $1529 = $1530&255;
      $1531 = (125924 + ($1529<<2)|0);
      $1532 = HEAP32[$1531>>2]|0;
      $1533 = (($1532) + ($1528)|0);
      $1534 = HEAP8[$1533>>0]|0;
      $1535 = $1534&255;
      $1536 = Math_imul($1511, $3)|0;
      $1537 = (($1536>>>0) / 255)&-1;
      $1538 = (($1537) + ($1248))|0;
      $1539 = Math_imul($1523, $3)|0;
      $1540 = (($1539>>>0) / 255)&-1;
      $1541 = (($1540) + ($1249))|0;
      $1542 = Math_imul($1535, $3)|0;
      $1543 = (($1542>>>0) / 255)&-1;
      $1544 = (($1543) + ($1250))|0;
      $1545 = $1538 >>> $1505;
      $1546 = $1545 << $1502;
      $1547 = $1541 >>> $1517;
      $1548 = $1547 << $1514;
      $1549 = $1548 | $1546;
      $1550 = $1544 >>> $1529;
      $1551 = $1550 << $1526;
      $1552 = HEAP32[$1251>>2]|0;
      $1553 = $1549 | $1552;
      $1554 = $1553 | $1551;
      HEAP32[$pixel83$4>>2] = $1554;
      $1555 = ((($pixel83$4)) + 4|0);
      $1556 = (($n84$3) + -1)|0;
      $1557 = ($n84$3|0)>(1);
      if ($1557) {
       $1335 = $1503;$1338 = $1506;$1340 = $1508;$1346 = $1515;$1349 = $1518;$1351 = $1520;$1357 = $1527;$1360 = $1530;$1362 = $1532;$2468 = $1518;$2469 = $1515;$2470 = $1506;$2471 = $1503;$n84$0 = $1556;$pixel83$1 = $1555;
       label = 63;
       continue;
      } else {
       break;
      }
     }
    }
    $1558 = (($1555) + ($1230<<2)|0);
    $1559 = ($1252|0)==(0);
    if ($1559) {
     $$0 = 0;
     label = 109;
     break;
    } else {
     $$in47 = $1252;$pixel83$040 = $1558;
    }
   }
   if ((label|0) == 67) {
    // unreachable;
   }
   else if ((label|0) == 109) {
    return ($$0|0);
   }
   break;
  }
  case 2:  {
   $1560 = ((($rect)) + 8|0);
   $1561 = HEAP32[$1560>>2]|0;
   $1562 = ((($rect)) + 12|0);
   $1563 = HEAP32[$1562>>2]|0;
   $1564 = ((($dst)) + 16|0);
   $1565 = HEAP32[$1564>>2]|0;
   $1566 = (($1565|0) / 4)&-1;
   $1567 = (($1566) - ($1561))|0;
   $1568 = ($1563|0)==(0);
   if ($1568) {
    $$0 = 0;
    return ($$0|0);
   }
   $1569 = ((($dst)) + 20|0);
   $1570 = HEAP32[$1569>>2]|0;
   $1571 = ((($rect)) + 4|0);
   $1572 = HEAP32[$1571>>2]|0;
   $1573 = Math_imul($1572, $1566)|0;
   $1574 = (($1570) + ($1573<<2)|0);
   $1575 = HEAP32[$rect>>2]|0;
   $1576 = (($1574) + ($1575<<2)|0);
   $1577 = (($1561) + 3)|0;
   $1578 = (($1577|0) / 4)&-1;
   $1579 = $1561 & 3;
   $1580 = ((($1)) + 12|0);
   $1581 = ((($1)) + 32|0);
   $1582 = ((($1)) + 28|0);
   $1583 = ((($1)) + 16|0);
   $1584 = ((($1)) + 20|0);
   $1585 = $r&255;
   $1586 = $g&255;
   $1587 = $b&255;
   $1588 = ((($1)) + 24|0);
   $$in46 = $1563;$pixel105$042 = $1576;
   L138: while(1) {
    $1589 = (($$in46) + -1)|0;
    switch ($1579|0) {
    case 0:  {
     $1646 = HEAP16[$1581>>1]|0;
     $1647 = HEAP32[$1581>>2]|0;
     $1648 = $1647&65535;
     $1649 = $1647&255;
     $1650 = HEAP16[$1582>>1]|0;
     $1651 = HEAP32[$1582>>2]|0;
     $1652 = $1651&65535;
     $1653 = $1651&255;
     $$phi$trans$insert81 = $1651 & 255;
     $$phi$trans$insert82 = (125924 + ($$phi$trans$insert81<<2)|0);
     $$pre83 = HEAP32[$$phi$trans$insert82>>2]|0;
     $1654 = ($1648&65535) >>> 8;
     $1655 = $1654&255;
     $1656 = ($1652&65535) >>> 8;
     $1657 = $1656&255;
     $$phi$trans$insert84 = $1656&65535;
     $$phi$trans$insert85 = (125924 + ($$phi$trans$insert84<<2)|0);
     $$pre86 = HEAP32[$$phi$trans$insert85>>2]|0;
     $1658 = $1647 >>> 16;
     $1659 = $1658&255;
     $1660 = $1651 >>> 16;
     $1661 = $1660&255;
     $$phi$trans$insert87 = $1660 & 255;
     $$phi$trans$insert88 = (125924 + ($$phi$trans$insert87<<2)|0);
     $$pre89 = HEAP32[$$phi$trans$insert88>>2]|0;
     $1662 = $1647&255;
     $1663 = $1651&255;
     $1664 = ($1646&65535) >>> 8;
     $1665 = $1664&255;
     $1666 = ($1650&65535) >>> 8;
     $1667 = $1666&255;
     $1672 = $1649;$1675 = $1653;$1677 = $$pre83;$1683 = $1655;$1686 = $1657;$1688 = $$pre86;$1694 = $1659;$1697 = $1661;$1699 = $$pre89;$2480 = $1667;$2481 = $1665;$2482 = $1663;$2483 = $1662;$n106$0 = $1578;$pixel105$1 = $pixel105$042;
     label = 76;
     break;
    }
    case 3:  {
     $1624 = HEAP16[$1581>>1]|0;
     $1625 = HEAP32[$1581>>2]|0;
     $1626 = $1625&65535;
     $1627 = $1625&255;
     $1628 = HEAP16[$1582>>1]|0;
     $1629 = HEAP32[$1582>>2]|0;
     $1630 = $1629&65535;
     $1631 = $1629&255;
     $1632 = ($1626&65535) >>> 8;
     $1633 = $1632&255;
     $1634 = ($1630&65535) >>> 8;
     $1635 = $1634&255;
     $1636 = $1625 >>> 16;
     $1637 = $1636&255;
     $1638 = $1629 >>> 16;
     $1639 = $1638&255;
     $1640 = $1625&255;
     $1641 = $1629&255;
     $1642 = ($1624&65535) >>> 8;
     $1643 = $1642&255;
     $1644 = ($1628&65535) >>> 8;
     $1645 = $1644&255;
     $1723 = $1627;$1726 = $1631;$1735 = $1633;$1738 = $1635;$1747 = $1637;$1750 = $1639;$2476 = $1645;$2477 = $1643;$2478 = $1641;$2479 = $1640;$n106$1 = $1578;$pixel105$2 = $pixel105$042;
     label = 77;
     break;
    }
    case 2:  {
     $1602 = HEAP16[$1581>>1]|0;
     $1603 = HEAP32[$1581>>2]|0;
     $1604 = $1603&65535;
     $1605 = $1603&255;
     $1606 = HEAP16[$1582>>1]|0;
     $1607 = HEAP32[$1582>>2]|0;
     $1608 = $1607&65535;
     $1609 = $1607&255;
     $1610 = ($1604&65535) >>> 8;
     $1611 = $1610&255;
     $1612 = ($1608&65535) >>> 8;
     $1613 = $1612&255;
     $1614 = $1603 >>> 16;
     $1615 = $1614&255;
     $1616 = $1607 >>> 16;
     $1617 = $1616&255;
     $1618 = $1603&255;
     $1619 = $1607&255;
     $1620 = ($1602&65535) >>> 8;
     $1621 = $1620&255;
     $1622 = ($1606&65535) >>> 8;
     $1623 = $1622&255;
     $1777 = $1605;$1780 = $1609;$1789 = $1611;$1792 = $1613;$1801 = $1615;$1804 = $1617;$2472 = $1623;$2473 = $1621;$2474 = $1619;$2475 = $1618;$n106$2 = $1578;$pixel105$3 = $pixel105$042;
     label = 78;
     break;
    }
    case 1:  {
     $$pre97 = HEAP32[$1581>>2]|0;
     $1590 = $$pre97&65535;
     $1591 = $$pre97&255;
     $$pre98 = HEAP32[$1582>>2]|0;
     $1592 = $$pre98&65535;
     $1593 = $$pre98&255;
     $1594 = ($1590&65535) >>> 8;
     $1595 = $1594&255;
     $1596 = ($1592&65535) >>> 8;
     $1597 = $1596&255;
     $1598 = $$pre97 >>> 16;
     $1599 = $1598&255;
     $1600 = $$pre98 >>> 16;
     $1601 = $1600&255;
     $1831 = $1591;$1834 = $1593;$1843 = $1595;$1846 = $1597;$1855 = $1599;$1858 = $1601;$n106$3 = $1578;$pixel105$4 = $pixel105$042;
     label = 79;
     break;
    }
    default: {
     label = 80;
     break L138;
    }
    }
    while(1) {
     if ((label|0) == 76) {
      label = 0;
      $1668 = HEAP32[$pixel105$1>>2]|0;
      $1669 = HEAP32[$1580>>2]|0;
      $1670 = $1669 & $1668;
      $1671 = $1672&255;
      $1673 = $1670 >>> $1671;
      $1674 = $1675&255;
      $1676 = (($1677) + ($1673)|0);
      $1678 = HEAP8[$1676>>0]|0;
      $1679 = $1678&255;
      $1680 = HEAP32[$1583>>2]|0;
      $1681 = $1680 & $1668;
      $1682 = $1683&255;
      $1684 = $1681 >>> $1682;
      $1685 = $1686&255;
      $1687 = (($1688) + ($1684)|0);
      $1689 = HEAP8[$1687>>0]|0;
      $1690 = $1689&255;
      $1691 = HEAP32[$1584>>2]|0;
      $1692 = $1691 & $1668;
      $1693 = $1694&255;
      $1695 = $1692 >>> $1693;
      $1696 = $1697&255;
      $1698 = (($1699) + ($1695)|0);
      $1700 = HEAP8[$1698>>0]|0;
      $1701 = $1700&255;
      $1702 = (($1679) + ($1585))|0;
      $1703 = ($1702>>>0)>(255);
      $$8 = $1703 ? 255 : $1702;
      $1704 = (($1690) + ($1586))|0;
      $1705 = ($1704>>>0)>(255);
      $$20 = $1705 ? 255 : $1704;
      $1706 = (($1701) + ($1587))|0;
      $1707 = ($1706>>>0)>(255);
      $$9 = $1707 ? 255 : $1706;
      $1708 = $$8 >>> $1674;
      $1709 = $1708 << $1671;
      $1710 = $$20 >>> $1685;
      $1711 = $1710 << $1682;
      $1712 = $1711 | $1709;
      $1713 = $$9 >>> $1696;
      $1714 = $1713 << $1693;
      $1715 = HEAP32[$1588>>2]|0;
      $1716 = $1712 | $1715;
      $1717 = $1716 | $1714;
      HEAP32[$pixel105$1>>2] = $1717;
      $1718 = ((($pixel105$1)) + 4|0);
      $1723 = $2483;$1726 = $2482;$1735 = $2481;$1738 = $2480;$1747 = $1694;$1750 = $1697;$2476 = $2480;$2477 = $2481;$2478 = $2482;$2479 = $2483;$n106$1 = $n106$0;$pixel105$2 = $1718;
      label = 77;
      continue;
     }
     else if ((label|0) == 77) {
      label = 0;
      $1719 = HEAP32[$pixel105$2>>2]|0;
      $1720 = HEAP32[$1580>>2]|0;
      $1721 = $1720 & $1719;
      $1722 = $1723&255;
      $1724 = $1721 >>> $1722;
      $1725 = $1726&255;
      $1727 = (125924 + ($1725<<2)|0);
      $1728 = HEAP32[$1727>>2]|0;
      $1729 = (($1728) + ($1724)|0);
      $1730 = HEAP8[$1729>>0]|0;
      $1731 = $1730&255;
      $1732 = HEAP32[$1583>>2]|0;
      $1733 = $1732 & $1719;
      $1734 = $1735&255;
      $1736 = $1733 >>> $1734;
      $1737 = $1738&255;
      $1739 = (125924 + ($1737<<2)|0);
      $1740 = HEAP32[$1739>>2]|0;
      $1741 = (($1740) + ($1736)|0);
      $1742 = HEAP8[$1741>>0]|0;
      $1743 = $1742&255;
      $1744 = HEAP32[$1584>>2]|0;
      $1745 = $1744 & $1719;
      $1746 = $1747&255;
      $1748 = $1745 >>> $1746;
      $1749 = $1750&255;
      $1751 = (125924 + ($1749<<2)|0);
      $1752 = HEAP32[$1751>>2]|0;
      $1753 = (($1752) + ($1748)|0);
      $1754 = HEAP8[$1753>>0]|0;
      $1755 = $1754&255;
      $1756 = (($1731) + ($1585))|0;
      $1757 = ($1756>>>0)>(255);
      $$10 = $1757 ? 255 : $1756;
      $1758 = (($1743) + ($1586))|0;
      $1759 = ($1758>>>0)>(255);
      $$21 = $1759 ? 255 : $1758;
      $1760 = (($1755) + ($1587))|0;
      $1761 = ($1760>>>0)>(255);
      $$11 = $1761 ? 255 : $1760;
      $1762 = $$10 >>> $1725;
      $1763 = $1762 << $1722;
      $1764 = $$21 >>> $1737;
      $1765 = $1764 << $1734;
      $1766 = $1765 | $1763;
      $1767 = $$11 >>> $1749;
      $1768 = $1767 << $1746;
      $1769 = HEAP32[$1588>>2]|0;
      $1770 = $1766 | $1769;
      $1771 = $1770 | $1768;
      HEAP32[$pixel105$2>>2] = $1771;
      $1772 = ((($pixel105$2)) + 4|0);
      $1777 = $2479;$1780 = $2478;$1789 = $2477;$1792 = $2476;$1801 = $1747;$1804 = $1750;$2472 = $2476;$2473 = $2477;$2474 = $2478;$2475 = $2479;$n106$2 = $n106$1;$pixel105$3 = $1772;
      label = 78;
      continue;
     }
     else if ((label|0) == 78) {
      label = 0;
      $1773 = HEAP32[$pixel105$3>>2]|0;
      $1774 = HEAP32[$1580>>2]|0;
      $1775 = $1774 & $1773;
      $1776 = $1777&255;
      $1778 = $1775 >>> $1776;
      $1779 = $1780&255;
      $1781 = (125924 + ($1779<<2)|0);
      $1782 = HEAP32[$1781>>2]|0;
      $1783 = (($1782) + ($1778)|0);
      $1784 = HEAP8[$1783>>0]|0;
      $1785 = $1784&255;
      $1786 = HEAP32[$1583>>2]|0;
      $1787 = $1786 & $1773;
      $1788 = $1789&255;
      $1790 = $1787 >>> $1788;
      $1791 = $1792&255;
      $1793 = (125924 + ($1791<<2)|0);
      $1794 = HEAP32[$1793>>2]|0;
      $1795 = (($1794) + ($1790)|0);
      $1796 = HEAP8[$1795>>0]|0;
      $1797 = $1796&255;
      $1798 = HEAP32[$1584>>2]|0;
      $1799 = $1798 & $1773;
      $1800 = $1801&255;
      $1802 = $1799 >>> $1800;
      $1803 = $1804&255;
      $1805 = (125924 + ($1803<<2)|0);
      $1806 = HEAP32[$1805>>2]|0;
      $1807 = (($1806) + ($1802)|0);
      $1808 = HEAP8[$1807>>0]|0;
      $1809 = $1808&255;
      $1810 = (($1785) + ($1585))|0;
      $1811 = ($1810>>>0)>(255);
      $$12 = $1811 ? 255 : $1810;
      $1812 = (($1797) + ($1586))|0;
      $1813 = ($1812>>>0)>(255);
      $$22 = $1813 ? 255 : $1812;
      $1814 = (($1809) + ($1587))|0;
      $1815 = ($1814>>>0)>(255);
      $$13 = $1815 ? 255 : $1814;
      $1816 = $$12 >>> $1779;
      $1817 = $1816 << $1776;
      $1818 = $$22 >>> $1791;
      $1819 = $1818 << $1788;
      $1820 = $1819 | $1817;
      $1821 = $$13 >>> $1803;
      $1822 = $1821 << $1800;
      $1823 = HEAP32[$1588>>2]|0;
      $1824 = $1820 | $1823;
      $1825 = $1824 | $1822;
      HEAP32[$pixel105$3>>2] = $1825;
      $1826 = ((($pixel105$3)) + 4|0);
      $1831 = $2475;$1834 = $2474;$1843 = $2473;$1846 = $2472;$1855 = $1801;$1858 = $1804;$n106$3 = $n106$2;$pixel105$4 = $1826;
      label = 79;
      continue;
     }
     else if ((label|0) == 79) {
      label = 0;
      $1827 = HEAP32[$pixel105$4>>2]|0;
      $1828 = HEAP32[$1580>>2]|0;
      $1829 = $1828 & $1827;
      $1830 = $1831&255;
      $1832 = $1829 >>> $1830;
      $1833 = $1834&255;
      $1835 = (125924 + ($1833<<2)|0);
      $1836 = HEAP32[$1835>>2]|0;
      $1837 = (($1836) + ($1832)|0);
      $1838 = HEAP8[$1837>>0]|0;
      $1839 = $1838&255;
      $1840 = HEAP32[$1583>>2]|0;
      $1841 = $1840 & $1827;
      $1842 = $1843&255;
      $1844 = $1841 >>> $1842;
      $1845 = $1846&255;
      $1847 = (125924 + ($1845<<2)|0);
      $1848 = HEAP32[$1847>>2]|0;
      $1849 = (($1848) + ($1844)|0);
      $1850 = HEAP8[$1849>>0]|0;
      $1851 = $1850&255;
      $1852 = HEAP32[$1584>>2]|0;
      $1853 = $1852 & $1827;
      $1854 = $1855&255;
      $1856 = $1853 >>> $1854;
      $1857 = $1858&255;
      $1859 = (125924 + ($1857<<2)|0);
      $1860 = HEAP32[$1859>>2]|0;
      $1861 = (($1860) + ($1856)|0);
      $1862 = HEAP8[$1861>>0]|0;
      $1863 = $1862&255;
      $1864 = (($1839) + ($1585))|0;
      $1865 = ($1864>>>0)>(255);
      $$14 = $1865 ? 255 : $1864;
      $1866 = (($1851) + ($1586))|0;
      $1867 = ($1866>>>0)>(255);
      $$23 = $1867 ? 255 : $1866;
      $1868 = (($1863) + ($1587))|0;
      $1869 = ($1868>>>0)>(255);
      $$15 = $1869 ? 255 : $1868;
      $1870 = $$14 >>> $1833;
      $1871 = $1870 << $1830;
      $1872 = $$23 >>> $1845;
      $1873 = $1872 << $1842;
      $1874 = $1873 | $1871;
      $1875 = $$15 >>> $1857;
      $1876 = $1875 << $1854;
      $1877 = HEAP32[$1588>>2]|0;
      $1878 = $1874 | $1877;
      $1879 = $1878 | $1876;
      HEAP32[$pixel105$4>>2] = $1879;
      $1880 = ((($pixel105$4)) + 4|0);
      $1881 = (($n106$3) + -1)|0;
      $1882 = ($n106$3|0)>(1);
      if ($1882) {
       $1672 = $1831;$1675 = $1834;$1677 = $1836;$1683 = $1843;$1686 = $1846;$1688 = $1848;$1694 = $1855;$1697 = $1858;$1699 = $1860;$2480 = $1846;$2481 = $1843;$2482 = $1834;$2483 = $1831;$n106$0 = $1881;$pixel105$1 = $1880;
       label = 76;
       continue;
      } else {
       break;
      }
     }
    }
    $1883 = (($1880) + ($1567<<2)|0);
    $1884 = ($1589|0)==(0);
    if ($1884) {
     $$0 = 0;
     label = 109;
     break;
    } else {
     $$in46 = $1589;$pixel105$042 = $1883;
    }
   }
   if ((label|0) == 80) {
    // unreachable;
   }
   else if ((label|0) == 109) {
    return ($$0|0);
   }
   break;
  }
  case 4:  {
   $1885 = ((($rect)) + 8|0);
   $1886 = HEAP32[$1885>>2]|0;
   $1887 = ((($rect)) + 12|0);
   $1888 = HEAP32[$1887>>2]|0;
   $1889 = ((($dst)) + 16|0);
   $1890 = HEAP32[$1889>>2]|0;
   $1891 = (($1890|0) / 4)&-1;
   $1892 = (($1891) - ($1886))|0;
   $1893 = ($1888|0)==(0);
   if ($1893) {
    $$0 = 0;
    return ($$0|0);
   }
   $1894 = ((($dst)) + 20|0);
   $1895 = HEAP32[$1894>>2]|0;
   $1896 = ((($rect)) + 4|0);
   $1897 = HEAP32[$1896>>2]|0;
   $1898 = Math_imul($1897, $1891)|0;
   $1899 = (($1895) + ($1898<<2)|0);
   $1900 = HEAP32[$rect>>2]|0;
   $1901 = (($1899) + ($1900<<2)|0);
   $1902 = (($1886) + 3)|0;
   $1903 = (($1902|0) / 4)&-1;
   $1904 = $1886 & 3;
   $1905 = ((($1)) + 12|0);
   $1906 = ((($1)) + 32|0);
   $1907 = ((($1)) + 28|0);
   $1908 = ((($1)) + 16|0);
   $1909 = ((($1)) + 20|0);
   $1910 = $r&255;
   $1911 = $g&255;
   $1912 = $b&255;
   $1913 = ((($1)) + 24|0);
   $$in = $1888;$pixel127$044 = $1901;
   L159: while(1) {
    $1914 = (($$in) + -1)|0;
    switch ($1904|0) {
    case 0:  {
     $1971 = HEAP16[$1906>>1]|0;
     $1972 = HEAP32[$1906>>2]|0;
     $1973 = $1972&65535;
     $1974 = $1972&255;
     $1975 = HEAP16[$1907>>1]|0;
     $1976 = HEAP32[$1907>>2]|0;
     $1977 = $1976&65535;
     $1978 = $1976&255;
     $$phi$trans$insert = $1976 & 255;
     $$phi$trans$insert61 = (125924 + ($$phi$trans$insert<<2)|0);
     $$pre62 = HEAP32[$$phi$trans$insert61>>2]|0;
     $1979 = ($1973&65535) >>> 8;
     $1980 = $1979&255;
     $1981 = ($1977&65535) >>> 8;
     $1982 = $1981&255;
     $$phi$trans$insert63 = $1981&65535;
     $$phi$trans$insert64 = (125924 + ($$phi$trans$insert63<<2)|0);
     $$pre65 = HEAP32[$$phi$trans$insert64>>2]|0;
     $1983 = $1972 >>> 16;
     $1984 = $1983&255;
     $1985 = $1976 >>> 16;
     $1986 = $1985&255;
     $$phi$trans$insert66 = $1985 & 255;
     $$phi$trans$insert67 = (125924 + ($$phi$trans$insert66<<2)|0);
     $$pre68 = HEAP32[$$phi$trans$insert67>>2]|0;
     $1987 = $1972&255;
     $1988 = $1976&255;
     $1989 = ($1971&65535) >>> 8;
     $1990 = $1989&255;
     $1991 = ($1975&65535) >>> 8;
     $1992 = $1991&255;
     $1997 = $1974;$2000 = $1978;$2002 = $$pre62;$2008 = $1980;$2011 = $1982;$2013 = $$pre65;$2019 = $1984;$2022 = $1986;$2024 = $$pre68;$2492 = $1992;$2493 = $1990;$2494 = $1988;$2495 = $1987;$n128$0 = $1903;$pixel127$1 = $pixel127$044;
     label = 89;
     break;
    }
    case 3:  {
     $1949 = HEAP16[$1906>>1]|0;
     $1950 = HEAP32[$1906>>2]|0;
     $1951 = $1950&65535;
     $1952 = $1950&255;
     $1953 = HEAP16[$1907>>1]|0;
     $1954 = HEAP32[$1907>>2]|0;
     $1955 = $1954&65535;
     $1956 = $1954&255;
     $1957 = ($1951&65535) >>> 8;
     $1958 = $1957&255;
     $1959 = ($1955&65535) >>> 8;
     $1960 = $1959&255;
     $1961 = $1950 >>> 16;
     $1962 = $1961&255;
     $1963 = $1954 >>> 16;
     $1964 = $1963&255;
     $1965 = $1950&255;
     $1966 = $1954&255;
     $1967 = ($1949&65535) >>> 8;
     $1968 = $1967&255;
     $1969 = ($1953&65535) >>> 8;
     $1970 = $1969&255;
     $2048 = $1952;$2051 = $1956;$2060 = $1958;$2063 = $1960;$2072 = $1962;$2075 = $1964;$2488 = $1970;$2489 = $1968;$2490 = $1966;$2491 = $1965;$n128$1 = $1903;$pixel127$2 = $pixel127$044;
     label = 90;
     break;
    }
    case 2:  {
     $1927 = HEAP16[$1906>>1]|0;
     $1928 = HEAP32[$1906>>2]|0;
     $1929 = $1928&65535;
     $1930 = $1928&255;
     $1931 = HEAP16[$1907>>1]|0;
     $1932 = HEAP32[$1907>>2]|0;
     $1933 = $1932&65535;
     $1934 = $1932&255;
     $1935 = ($1929&65535) >>> 8;
     $1936 = $1935&255;
     $1937 = ($1933&65535) >>> 8;
     $1938 = $1937&255;
     $1939 = $1928 >>> 16;
     $1940 = $1939&255;
     $1941 = $1932 >>> 16;
     $1942 = $1941&255;
     $1943 = $1928&255;
     $1944 = $1932&255;
     $1945 = ($1927&65535) >>> 8;
     $1946 = $1945&255;
     $1947 = ($1931&65535) >>> 8;
     $1948 = $1947&255;
     $2102 = $1930;$2105 = $1934;$2114 = $1936;$2117 = $1938;$2126 = $1940;$2129 = $1942;$2484 = $1948;$2485 = $1946;$2486 = $1944;$2487 = $1943;$n128$2 = $1903;$pixel127$3 = $pixel127$044;
     label = 91;
     break;
    }
    case 1:  {
     $$pre76 = HEAP32[$1906>>2]|0;
     $1915 = $$pre76&65535;
     $1916 = $$pre76&255;
     $$pre77 = HEAP32[$1907>>2]|0;
     $1917 = $$pre77&65535;
     $1918 = $$pre77&255;
     $1919 = ($1915&65535) >>> 8;
     $1920 = $1919&255;
     $1921 = ($1917&65535) >>> 8;
     $1922 = $1921&255;
     $1923 = $$pre76 >>> 16;
     $1924 = $1923&255;
     $1925 = $$pre77 >>> 16;
     $1926 = $1925&255;
     $2156 = $1916;$2159 = $1918;$2168 = $1920;$2171 = $1922;$2180 = $1924;$2183 = $1926;$n128$3 = $1903;$pixel127$4 = $pixel127$044;
     label = 92;
     break;
    }
    default: {
     label = 93;
     break L159;
    }
    }
    while(1) {
     if ((label|0) == 89) {
      label = 0;
      $1993 = HEAP32[$pixel127$1>>2]|0;
      $1994 = HEAP32[$1905>>2]|0;
      $1995 = $1994 & $1993;
      $1996 = $1997&255;
      $1998 = $1995 >>> $1996;
      $1999 = $2000&255;
      $2001 = (($2002) + ($1998)|0);
      $2003 = HEAP8[$2001>>0]|0;
      $2004 = $2003&255;
      $2005 = HEAP32[$1908>>2]|0;
      $2006 = $2005 & $1993;
      $2007 = $2008&255;
      $2009 = $2006 >>> $2007;
      $2010 = $2011&255;
      $2012 = (($2013) + ($2009)|0);
      $2014 = HEAP8[$2012>>0]|0;
      $2015 = $2014&255;
      $2016 = HEAP32[$1909>>2]|0;
      $2017 = $2016 & $1993;
      $2018 = $2019&255;
      $2020 = $2017 >>> $2018;
      $2021 = $2022&255;
      $2023 = (($2024) + ($2020)|0);
      $2025 = HEAP8[$2023>>0]|0;
      $2026 = $2025&255;
      $2027 = Math_imul($2004, $1910)|0;
      $2028 = (($2027>>>0) / 255)&-1;
      $2029 = Math_imul($2015, $1911)|0;
      $2030 = (($2029>>>0) / 255)&-1;
      $2031 = Math_imul($2026, $1912)|0;
      $2032 = (($2031>>>0) / 255)&-1;
      $2033 = $2028 >>> $1999;
      $2034 = $2033 << $1996;
      $2035 = $2030 >>> $2010;
      $2036 = $2035 << $2007;
      $2037 = $2036 | $2034;
      $2038 = $2032 >>> $2021;
      $2039 = $2038 << $2018;
      $2040 = HEAP32[$1913>>2]|0;
      $2041 = $2037 | $2040;
      $2042 = $2041 | $2039;
      HEAP32[$pixel127$1>>2] = $2042;
      $2043 = ((($pixel127$1)) + 4|0);
      $2048 = $2495;$2051 = $2494;$2060 = $2493;$2063 = $2492;$2072 = $2019;$2075 = $2022;$2488 = $2492;$2489 = $2493;$2490 = $2494;$2491 = $2495;$n128$1 = $n128$0;$pixel127$2 = $2043;
      label = 90;
      continue;
     }
     else if ((label|0) == 90) {
      label = 0;
      $2044 = HEAP32[$pixel127$2>>2]|0;
      $2045 = HEAP32[$1905>>2]|0;
      $2046 = $2045 & $2044;
      $2047 = $2048&255;
      $2049 = $2046 >>> $2047;
      $2050 = $2051&255;
      $2052 = (125924 + ($2050<<2)|0);
      $2053 = HEAP32[$2052>>2]|0;
      $2054 = (($2053) + ($2049)|0);
      $2055 = HEAP8[$2054>>0]|0;
      $2056 = $2055&255;
      $2057 = HEAP32[$1908>>2]|0;
      $2058 = $2057 & $2044;
      $2059 = $2060&255;
      $2061 = $2058 >>> $2059;
      $2062 = $2063&255;
      $2064 = (125924 + ($2062<<2)|0);
      $2065 = HEAP32[$2064>>2]|0;
      $2066 = (($2065) + ($2061)|0);
      $2067 = HEAP8[$2066>>0]|0;
      $2068 = $2067&255;
      $2069 = HEAP32[$1909>>2]|0;
      $2070 = $2069 & $2044;
      $2071 = $2072&255;
      $2073 = $2070 >>> $2071;
      $2074 = $2075&255;
      $2076 = (125924 + ($2074<<2)|0);
      $2077 = HEAP32[$2076>>2]|0;
      $2078 = (($2077) + ($2073)|0);
      $2079 = HEAP8[$2078>>0]|0;
      $2080 = $2079&255;
      $2081 = Math_imul($2056, $1910)|0;
      $2082 = (($2081>>>0) / 255)&-1;
      $2083 = Math_imul($2068, $1911)|0;
      $2084 = (($2083>>>0) / 255)&-1;
      $2085 = Math_imul($2080, $1912)|0;
      $2086 = (($2085>>>0) / 255)&-1;
      $2087 = $2082 >>> $2050;
      $2088 = $2087 << $2047;
      $2089 = $2084 >>> $2062;
      $2090 = $2089 << $2059;
      $2091 = $2090 | $2088;
      $2092 = $2086 >>> $2074;
      $2093 = $2092 << $2071;
      $2094 = HEAP32[$1913>>2]|0;
      $2095 = $2091 | $2094;
      $2096 = $2095 | $2093;
      HEAP32[$pixel127$2>>2] = $2096;
      $2097 = ((($pixel127$2)) + 4|0);
      $2102 = $2491;$2105 = $2490;$2114 = $2489;$2117 = $2488;$2126 = $2072;$2129 = $2075;$2484 = $2488;$2485 = $2489;$2486 = $2490;$2487 = $2491;$n128$2 = $n128$1;$pixel127$3 = $2097;
      label = 91;
      continue;
     }
     else if ((label|0) == 91) {
      label = 0;
      $2098 = HEAP32[$pixel127$3>>2]|0;
      $2099 = HEAP32[$1905>>2]|0;
      $2100 = $2099 & $2098;
      $2101 = $2102&255;
      $2103 = $2100 >>> $2101;
      $2104 = $2105&255;
      $2106 = (125924 + ($2104<<2)|0);
      $2107 = HEAP32[$2106>>2]|0;
      $2108 = (($2107) + ($2103)|0);
      $2109 = HEAP8[$2108>>0]|0;
      $2110 = $2109&255;
      $2111 = HEAP32[$1908>>2]|0;
      $2112 = $2111 & $2098;
      $2113 = $2114&255;
      $2115 = $2112 >>> $2113;
      $2116 = $2117&255;
      $2118 = (125924 + ($2116<<2)|0);
      $2119 = HEAP32[$2118>>2]|0;
      $2120 = (($2119) + ($2115)|0);
      $2121 = HEAP8[$2120>>0]|0;
      $2122 = $2121&255;
      $2123 = HEAP32[$1909>>2]|0;
      $2124 = $2123 & $2098;
      $2125 = $2126&255;
      $2127 = $2124 >>> $2125;
      $2128 = $2129&255;
      $2130 = (125924 + ($2128<<2)|0);
      $2131 = HEAP32[$2130>>2]|0;
      $2132 = (($2131) + ($2127)|0);
      $2133 = HEAP8[$2132>>0]|0;
      $2134 = $2133&255;
      $2135 = Math_imul($2110, $1910)|0;
      $2136 = (($2135>>>0) / 255)&-1;
      $2137 = Math_imul($2122, $1911)|0;
      $2138 = (($2137>>>0) / 255)&-1;
      $2139 = Math_imul($2134, $1912)|0;
      $2140 = (($2139>>>0) / 255)&-1;
      $2141 = $2136 >>> $2104;
      $2142 = $2141 << $2101;
      $2143 = $2138 >>> $2116;
      $2144 = $2143 << $2113;
      $2145 = $2144 | $2142;
      $2146 = $2140 >>> $2128;
      $2147 = $2146 << $2125;
      $2148 = HEAP32[$1913>>2]|0;
      $2149 = $2145 | $2148;
      $2150 = $2149 | $2147;
      HEAP32[$pixel127$3>>2] = $2150;
      $2151 = ((($pixel127$3)) + 4|0);
      $2156 = $2487;$2159 = $2486;$2168 = $2485;$2171 = $2484;$2180 = $2126;$2183 = $2129;$n128$3 = $n128$2;$pixel127$4 = $2151;
      label = 92;
      continue;
     }
     else if ((label|0) == 92) {
      label = 0;
      $2152 = HEAP32[$pixel127$4>>2]|0;
      $2153 = HEAP32[$1905>>2]|0;
      $2154 = $2153 & $2152;
      $2155 = $2156&255;
      $2157 = $2154 >>> $2155;
      $2158 = $2159&255;
      $2160 = (125924 + ($2158<<2)|0);
      $2161 = HEAP32[$2160>>2]|0;
      $2162 = (($2161) + ($2157)|0);
      $2163 = HEAP8[$2162>>0]|0;
      $2164 = $2163&255;
      $2165 = HEAP32[$1908>>2]|0;
      $2166 = $2165 & $2152;
      $2167 = $2168&255;
      $2169 = $2166 >>> $2167;
      $2170 = $2171&255;
      $2172 = (125924 + ($2170<<2)|0);
      $2173 = HEAP32[$2172>>2]|0;
      $2174 = (($2173) + ($2169)|0);
      $2175 = HEAP8[$2174>>0]|0;
      $2176 = $2175&255;
      $2177 = HEAP32[$1909>>2]|0;
      $2178 = $2177 & $2152;
      $2179 = $2180&255;
      $2181 = $2178 >>> $2179;
      $2182 = $2183&255;
      $2184 = (125924 + ($2182<<2)|0);
      $2185 = HEAP32[$2184>>2]|0;
      $2186 = (($2185) + ($2181)|0);
      $2187 = HEAP8[$2186>>0]|0;
      $2188 = $2187&255;
      $2189 = Math_imul($2164, $1910)|0;
      $2190 = (($2189>>>0) / 255)&-1;
      $2191 = Math_imul($2176, $1911)|0;
      $2192 = (($2191>>>0) / 255)&-1;
      $2193 = Math_imul($2188, $1912)|0;
      $2194 = (($2193>>>0) / 255)&-1;
      $2195 = $2190 >>> $2158;
      $2196 = $2195 << $2155;
      $2197 = $2192 >>> $2170;
      $2198 = $2197 << $2167;
      $2199 = $2198 | $2196;
      $2200 = $2194 >>> $2182;
      $2201 = $2200 << $2179;
      $2202 = HEAP32[$1913>>2]|0;
      $2203 = $2199 | $2202;
      $2204 = $2203 | $2201;
      HEAP32[$pixel127$4>>2] = $2204;
      $2205 = ((($pixel127$4)) + 4|0);
      $2206 = (($n128$3) + -1)|0;
      $2207 = ($n128$3|0)>(1);
      if ($2207) {
       $1997 = $2156;$2000 = $2159;$2002 = $2161;$2008 = $2168;$2011 = $2171;$2013 = $2173;$2019 = $2180;$2022 = $2183;$2024 = $2185;$2492 = $2171;$2493 = $2168;$2494 = $2159;$2495 = $2156;$n128$0 = $2206;$pixel127$1 = $2205;
       label = 89;
       continue;
      } else {
       break;
      }
     }
    }
    $2208 = (($2205) + ($1892<<2)|0);
    $2209 = ($1914|0)==(0);
    if ($2209) {
     $$0 = 0;
     label = 109;
     break;
    } else {
     $$in = $1914;$pixel127$044 = $2208;
    }
   }
   if ((label|0) == 93) {
    // unreachable;
   }
   else if ((label|0) == 109) {
    return ($$0|0);
   }
   break;
  }
  default: {
   $2210 = ((($rect)) + 8|0);
   $2211 = HEAP32[$2210>>2]|0;
   $2212 = ((($rect)) + 12|0);
   $2213 = HEAP32[$2212>>2]|0;
   $2214 = ((($dst)) + 16|0);
   $2215 = HEAP32[$2214>>2]|0;
   $2216 = (($2215|0) / 4)&-1;
   $2217 = (($2216) - ($2211))|0;
   $2218 = ($2213|0)==(0);
   if ($2218) {
    $$0 = 0;
    return ($$0|0);
   }
   $2219 = ((($dst)) + 20|0);
   $2220 = HEAP32[$2219>>2]|0;
   $2221 = ((($rect)) + 4|0);
   $2222 = HEAP32[$2221>>2]|0;
   $2223 = Math_imul($2222, $2216)|0;
   $2224 = (($2220) + ($2223<<2)|0);
   $2225 = HEAP32[$rect>>2]|0;
   $2226 = (($2224) + ($2225<<2)|0);
   $2227 = (($2211) + 3)|0;
   $2228 = (($2227|0) / 4)&-1;
   $2229 = $2211 & 3;
   $2230 = $r&255;
   $2231 = $g&255;
   $2232 = $b&255;
   $2233 = ((($1)) + 28|0);
   $2234 = ((($1)) + 32|0);
   $2235 = ((($1)) + 24|0);
   $$in48 = $2213;$pixel149$038 = $2226;
   L96: while(1) {
    $2236 = (($$in48) + -1)|0;
    switch ($2229|0) {
    case 0:  {
     $2293 = HEAP16[$2233>>1]|0;
     $2294 = HEAP32[$2233>>2]|0;
     $2295 = $2294&65535;
     $2296 = $2294&255;
     $2297 = HEAP16[$2234>>1]|0;
     $2298 = HEAP32[$2234>>2]|0;
     $2299 = $2298&65535;
     $2300 = $2298&255;
     $2301 = ($2295&65535) >>> 8;
     $2302 = $2301&255;
     $2303 = ($2299&65535) >>> 8;
     $2304 = $2303&255;
     $2305 = $2294 >>> 16;
     $2306 = $2305&255;
     $2307 = $2298 >>> 16;
     $2308 = $2307&255;
     $2309 = $2294&255;
     $2310 = $2298&255;
     $2311 = ($2293&65535) >>> 8;
     $2312 = $2311&255;
     $2313 = ($2297&65535) >>> 8;
     $2314 = $2313&255;
     $2316 = $2296;$2319 = $2300;$2322 = $2302;$2325 = $2304;$2329 = $2306;$2332 = $2308;$2504 = $2314;$2505 = $2312;$2506 = $2310;$2507 = $2309;$n150$0 = $2228;$pixel149$1 = $pixel149$038;
     label = 102;
     break;
    }
    case 3:  {
     $2271 = HEAP16[$2233>>1]|0;
     $2272 = HEAP32[$2233>>2]|0;
     $2273 = $2272&65535;
     $2274 = $2272&255;
     $2275 = HEAP16[$2234>>1]|0;
     $2276 = HEAP32[$2234>>2]|0;
     $2277 = $2276&65535;
     $2278 = $2276&255;
     $2279 = ($2273&65535) >>> 8;
     $2280 = $2279&255;
     $2281 = ($2277&65535) >>> 8;
     $2282 = $2281&255;
     $2283 = $2272 >>> 16;
     $2284 = $2283&255;
     $2285 = $2276 >>> 16;
     $2286 = $2285&255;
     $2287 = $2272&255;
     $2288 = $2276&255;
     $2289 = ($2271&65535) >>> 8;
     $2290 = $2289&255;
     $2291 = ($2275&65535) >>> 8;
     $2292 = $2291&255;
     $2339 = $2274;$2342 = $2278;$2345 = $2280;$2348 = $2282;$2352 = $2284;$2355 = $2286;$2500 = $2292;$2501 = $2290;$2502 = $2288;$2503 = $2287;$n150$1 = $2228;$pixel149$2 = $pixel149$038;
     label = 103;
     break;
    }
    case 2:  {
     $2249 = HEAP16[$2233>>1]|0;
     $2250 = HEAP32[$2233>>2]|0;
     $2251 = $2250&65535;
     $2252 = $2250&255;
     $2253 = HEAP16[$2234>>1]|0;
     $2254 = HEAP32[$2234>>2]|0;
     $2255 = $2254&65535;
     $2256 = $2254&255;
     $2257 = ($2251&65535) >>> 8;
     $2258 = $2257&255;
     $2259 = ($2255&65535) >>> 8;
     $2260 = $2259&255;
     $2261 = $2250 >>> 16;
     $2262 = $2261&255;
     $2263 = $2254 >>> 16;
     $2264 = $2263&255;
     $2265 = $2250&255;
     $2266 = $2254&255;
     $2267 = ($2249&65535) >>> 8;
     $2268 = $2267&255;
     $2269 = ($2253&65535) >>> 8;
     $2270 = $2269&255;
     $2362 = $2252;$2365 = $2256;$2368 = $2258;$2371 = $2260;$2375 = $2262;$2378 = $2264;$2496 = $2270;$2497 = $2268;$2498 = $2266;$2499 = $2265;$n150$2 = $2228;$pixel149$3 = $pixel149$038;
     label = 104;
     break;
    }
    case 1:  {
     $$pre130 = HEAP32[$2233>>2]|0;
     $2237 = $$pre130&65535;
     $2238 = $$pre130&255;
     $$pre131 = HEAP32[$2234>>2]|0;
     $2239 = $$pre131&65535;
     $2240 = $$pre131&255;
     $2241 = ($2237&65535) >>> 8;
     $2242 = $2241&255;
     $2243 = ($2239&65535) >>> 8;
     $2244 = $2243&255;
     $2245 = $$pre130 >>> 16;
     $2246 = $2245&255;
     $2247 = $$pre131 >>> 16;
     $2248 = $2247&255;
     $2385 = $2238;$2388 = $2240;$2391 = $2242;$2394 = $2244;$2398 = $2246;$2401 = $2248;$n150$3 = $2228;$pixel149$4 = $pixel149$038;
     label = 105;
     break;
    }
    default: {
     label = 106;
     break L96;
    }
    }
    while(1) {
     if ((label|0) == 102) {
      label = 0;
      $2315 = $2316&255;
      $2317 = $2230 >>> $2315;
      $2318 = $2319&255;
      $2320 = $2317 << $2318;
      $2321 = $2322&255;
      $2323 = $2231 >>> $2321;
      $2324 = $2325&255;
      $2326 = $2323 << $2324;
      $2327 = $2326 | $2320;
      $2328 = $2329&255;
      $2330 = $2232 >>> $2328;
      $2331 = $2332&255;
      $2333 = $2330 << $2331;
      $2334 = HEAP32[$2235>>2]|0;
      $2335 = $2327 | $2334;
      $2336 = $2335 | $2333;
      HEAP32[$pixel149$1>>2] = $2336;
      $2337 = ((($pixel149$1)) + 4|0);
      $2339 = $2507;$2342 = $2506;$2345 = $2505;$2348 = $2504;$2352 = $2329;$2355 = $2332;$2500 = $2504;$2501 = $2505;$2502 = $2506;$2503 = $2507;$n150$1 = $n150$0;$pixel149$2 = $2337;
      label = 103;
      continue;
     }
     else if ((label|0) == 103) {
      label = 0;
      $2338 = $2339&255;
      $2340 = $2230 >>> $2338;
      $2341 = $2342&255;
      $2343 = $2340 << $2341;
      $2344 = $2345&255;
      $2346 = $2231 >>> $2344;
      $2347 = $2348&255;
      $2349 = $2346 << $2347;
      $2350 = $2349 | $2343;
      $2351 = $2352&255;
      $2353 = $2232 >>> $2351;
      $2354 = $2355&255;
      $2356 = $2353 << $2354;
      $2357 = HEAP32[$2235>>2]|0;
      $2358 = $2350 | $2357;
      $2359 = $2358 | $2356;
      HEAP32[$pixel149$2>>2] = $2359;
      $2360 = ((($pixel149$2)) + 4|0);
      $2362 = $2503;$2365 = $2502;$2368 = $2501;$2371 = $2500;$2375 = $2352;$2378 = $2355;$2496 = $2500;$2497 = $2501;$2498 = $2502;$2499 = $2503;$n150$2 = $n150$1;$pixel149$3 = $2360;
      label = 104;
      continue;
     }
     else if ((label|0) == 104) {
      label = 0;
      $2361 = $2362&255;
      $2363 = $2230 >>> $2361;
      $2364 = $2365&255;
      $2366 = $2363 << $2364;
      $2367 = $2368&255;
      $2369 = $2231 >>> $2367;
      $2370 = $2371&255;
      $2372 = $2369 << $2370;
      $2373 = $2372 | $2366;
      $2374 = $2375&255;
      $2376 = $2232 >>> $2374;
      $2377 = $2378&255;
      $2379 = $2376 << $2377;
      $2380 = HEAP32[$2235>>2]|0;
      $2381 = $2373 | $2380;
      $2382 = $2381 | $2379;
      HEAP32[$pixel149$3>>2] = $2382;
      $2383 = ((($pixel149$3)) + 4|0);
      $2385 = $2499;$2388 = $2498;$2391 = $2497;$2394 = $2496;$2398 = $2375;$2401 = $2378;$n150$3 = $n150$2;$pixel149$4 = $2383;
      label = 105;
      continue;
     }
     else if ((label|0) == 105) {
      label = 0;
      $2384 = $2385&255;
      $2386 = $2230 >>> $2384;
      $2387 = $2388&255;
      $2389 = $2386 << $2387;
      $2390 = $2391&255;
      $2392 = $2231 >>> $2390;
      $2393 = $2394&255;
      $2395 = $2392 << $2393;
      $2396 = $2395 | $2389;
      $2397 = $2398&255;
      $2399 = $2232 >>> $2397;
      $2400 = $2401&255;
      $2402 = $2399 << $2400;
      $2403 = HEAP32[$2235>>2]|0;
      $2404 = $2396 | $2403;
      $2405 = $2404 | $2402;
      HEAP32[$pixel149$4>>2] = $2405;
      $2406 = ((($pixel149$4)) + 4|0);
      $2407 = (($n150$3) + -1)|0;
      $2408 = ($n150$3|0)>(1);
      if ($2408) {
       $2316 = $2385;$2319 = $2388;$2322 = $2391;$2325 = $2394;$2329 = $2398;$2332 = $2401;$2504 = $2394;$2505 = $2391;$2506 = $2388;$2507 = $2385;$n150$0 = $2407;$pixel149$1 = $2406;
       label = 102;
       continue;
      } else {
       break;
      }
     }
    }
    $2409 = (($2406) + ($2217<<2)|0);
    $2410 = ($2236|0)==(0);
    if ($2410) {
     $$0 = 0;
     label = 109;
     break;
    } else {
     $$in48 = $2236;$pixel149$038 = $2409;
    }
   }
   if ((label|0) == 106) {
    // unreachable;
   }
   else if ((label|0) == 109) {
    return ($$0|0);
   }
  }
  }
  break;
 }
 default: {
  $2411 = (_SDL_Error(4)|0);
  $$0 = $2411;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function _SDL_BlendFillRect_RGBA($dst,$rect,$blendMode,$r,$g,$b,$a) {
 $dst = $dst|0;
 $rect = $rect|0;
 $blendMode = $blendMode|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $$ = 0, $$0 = 0, $$1 = 0, $$10 = 0, $$11 = 0, $$2 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$6 = 0, $$7 = 0, $$8 = 0, $$9 = 0, $$in = 0, $$in26 = 0, $$in27 = 0, $$in28 = 0, $$phi$trans$insert = 0, $$phi$trans$insert33 = 0, $$phi$trans$insert35 = 0;
 var $$phi$trans$insert36 = 0, $$phi$trans$insert38 = 0, $$phi$trans$insert39 = 0, $$phi$trans$insert42 = 0, $$phi$trans$insert56 = 0, $$phi$trans$insert57 = 0, $$phi$trans$insert59 = 0, $$phi$trans$insert60 = 0, $$phi$trans$insert62 = 0, $$phi$trans$insert63 = 0, $$phi$trans$insert66 = 0, $$phi$trans$insert80 = 0, $$phi$trans$insert81 = 0, $$phi$trans$insert83 = 0, $$phi$trans$insert84 = 0, $$phi$trans$insert86 = 0, $$phi$trans$insert87 = 0, $$phi$trans$insert90 = 0, $$pre34 = 0, $$pre37 = 0;
 var $$pre40 = 0, $$pre43 = 0, $$pre58 = 0, $$pre61 = 0, $$pre64 = 0, $$pre67 = 0, $$pre82 = 0, $$pre85 = 0, $$pre88 = 0, $$pre91 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0;
 var $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0;
 var $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0;
 var $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0;
 var $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0;
 var $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0;
 var $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0;
 var $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0;
 var $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0;
 var $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0;
 var $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0;
 var $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0, $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0;
 var $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0, $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0;
 var $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0, $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0;
 var $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0, $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0;
 var $1259 = 0, $126 = 0, $1260 = 0, $1261 = 0, $1262 = 0, $1263 = 0, $1264 = 0, $1265 = 0, $1266 = 0, $1267 = 0, $1268 = 0, $1269 = 0, $127 = 0, $1270 = 0, $1271 = 0, $1272 = 0, $1273 = 0, $1274 = 0, $1275 = 0, $1276 = 0;
 var $1277 = 0, $1278 = 0, $1279 = 0, $128 = 0, $1280 = 0, $1281 = 0, $1282 = 0, $1283 = 0, $1284 = 0, $1285 = 0, $1286 = 0, $1287 = 0, $1288 = 0, $1289 = 0, $129 = 0, $1290 = 0, $1291 = 0, $1292 = 0, $1293 = 0, $1294 = 0;
 var $1295 = 0, $1296 = 0, $1297 = 0, $1298 = 0, $1299 = 0, $13 = 0, $130 = 0, $1300 = 0, $1301 = 0, $1302 = 0, $1303 = 0, $1304 = 0, $1305 = 0, $1306 = 0, $1307 = 0, $1308 = 0, $1309 = 0, $131 = 0, $1310 = 0, $1311 = 0;
 var $1312 = 0, $1313 = 0, $1314 = 0, $1315 = 0, $1316 = 0, $1317 = 0, $1318 = 0, $1319 = 0, $132 = 0, $1320 = 0, $1321 = 0, $1322 = 0, $1323 = 0, $1324 = 0, $1325 = 0, $1326 = 0, $1327 = 0, $1328 = 0, $1329 = 0, $133 = 0;
 var $1330 = 0, $1331 = 0, $1332 = 0, $1333 = 0, $1334 = 0, $1335 = 0, $1336 = 0, $1337 = 0, $1338 = 0, $1339 = 0, $134 = 0, $1340 = 0, $1341 = 0, $1342 = 0, $1343 = 0, $1344 = 0, $1345 = 0, $1346 = 0, $1347 = 0, $1348 = 0;
 var $1349 = 0, $135 = 0, $1350 = 0, $1351 = 0, $1352 = 0, $1353 = 0, $1354 = 0, $1355 = 0, $1356 = 0, $1357 = 0, $1358 = 0, $1359 = 0, $136 = 0, $1360 = 0, $1361 = 0, $1362 = 0, $1363 = 0, $1364 = 0, $1365 = 0, $1366 = 0;
 var $1367 = 0, $1368 = 0, $1369 = 0, $137 = 0, $1370 = 0, $1371 = 0, $1372 = 0, $1373 = 0, $1374 = 0, $1375 = 0, $1376 = 0, $1377 = 0, $1378 = 0, $1379 = 0, $138 = 0, $1380 = 0, $1381 = 0, $1382 = 0, $1383 = 0, $1384 = 0;
 var $1385 = 0, $1386 = 0, $1387 = 0, $1388 = 0, $1389 = 0, $139 = 0, $1390 = 0, $1391 = 0, $1392 = 0, $1393 = 0, $1394 = 0, $1395 = 0, $1396 = 0, $1397 = 0, $1398 = 0, $1399 = 0, $14 = 0, $140 = 0, $1400 = 0, $1401 = 0;
 var $1402 = 0, $1403 = 0, $1404 = 0, $1405 = 0, $1406 = 0, $1407 = 0, $1408 = 0, $1409 = 0, $141 = 0, $1410 = 0, $1411 = 0, $1412 = 0, $1413 = 0, $1414 = 0, $1415 = 0, $1416 = 0, $1417 = 0, $1418 = 0, $1419 = 0, $142 = 0;
 var $1420 = 0, $1421 = 0, $1422 = 0, $1423 = 0, $1424 = 0, $1425 = 0, $1426 = 0, $1427 = 0, $1428 = 0, $1429 = 0, $143 = 0, $1430 = 0, $1431 = 0, $1432 = 0, $1433 = 0, $1434 = 0, $1435 = 0, $1436 = 0, $1437 = 0, $1438 = 0;
 var $1439 = 0, $144 = 0, $1440 = 0, $1441 = 0, $1442 = 0, $1443 = 0, $1444 = 0, $1445 = 0, $1446 = 0, $1447 = 0, $1448 = 0, $1449 = 0, $145 = 0, $1450 = 0, $1451 = 0, $1452 = 0, $1453 = 0, $1454 = 0, $1455 = 0, $1456 = 0;
 var $1457 = 0, $1458 = 0, $1459 = 0, $146 = 0, $1460 = 0, $1461 = 0, $1462 = 0, $1463 = 0, $1464 = 0, $1465 = 0, $1466 = 0, $1467 = 0, $1468 = 0, $1469 = 0, $147 = 0, $1470 = 0, $1471 = 0, $1472 = 0, $1473 = 0, $1474 = 0;
 var $1475 = 0, $1476 = 0, $1477 = 0, $1478 = 0, $1479 = 0, $148 = 0, $1480 = 0, $1481 = 0, $1482 = 0, $1483 = 0, $1484 = 0, $1485 = 0, $1486 = 0, $1487 = 0, $1488 = 0, $1489 = 0, $149 = 0, $1490 = 0, $1491 = 0, $1492 = 0;
 var $1493 = 0, $1494 = 0, $1495 = 0, $1496 = 0, $1497 = 0, $1498 = 0, $1499 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0;
 var $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0;
 var $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0;
 var $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0;
 var $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0;
 var $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0;
 var $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0;
 var $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0;
 var $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0;
 var $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0;
 var $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0;
 var $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0;
 var $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0;
 var $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0;
 var $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0;
 var $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0;
 var $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0;
 var $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0;
 var $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0;
 var $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0;
 var $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0;
 var $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0;
 var $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0;
 var $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0;
 var $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0;
 var $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0;
 var $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0;
 var $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0;
 var $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0;
 var $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0;
 var $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0;
 var $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0;
 var $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0;
 var $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0;
 var $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0;
 var $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0;
 var $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0;
 var $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0;
 var $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0;
 var $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0;
 var $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0;
 var $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0;
 var $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0;
 var $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0;
 var $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0;
 var $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $cond = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n18$0 = 0, $n18$1 = 0, $n18$2 = 0, $n18$3 = 0, $n40$0 = 0;
 var $n40$1 = 0, $n40$2 = 0, $n40$3 = 0, $n62$0 = 0, $n62$1 = 0, $n62$2 = 0, $n62$3 = 0, $pixel$016 = 0, $pixel$1 = 0, $pixel$2 = 0, $pixel$3 = 0, $pixel$4 = 0, $pixel17$020 = 0, $pixel17$1 = 0, $pixel17$2 = 0, $pixel17$3 = 0, $pixel17$4 = 0, $pixel39$024 = 0, $pixel39$1 = 0, $pixel39$2 = 0;
 var $pixel39$3 = 0, $pixel39$4 = 0, $pixel61$015 = 0, $pixel61$1 = 0, $pixel61$2 = 0, $pixel61$3 = 0, $pixel61$4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($dst)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $a&255;
 $3 = $2 ^ 255;
 $4 = ((($1)) + 9|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = $5&255;
 $cond = ($5<<24>>24)==(4);
 if (!($cond)) {
  $1451 = (_SDL_Error(4)|0);
  $$0 = $1451;
  return ($$0|0);
 }
 switch ($blendMode|0) {
 case 1:  {
  $7 = ((($rect)) + 8|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ((($rect)) + 12|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ((($dst)) + 16|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = (($12|0) / ($6|0))&-1;
  $14 = (($13) - ($8))|0;
  $15 = ($10|0)==(0);
  if ($15) {
   $$0 = 0;
   return ($$0|0);
  }
  $16 = ((($dst)) + 20|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = ((($rect)) + 4|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = Math_imul($19, $13)|0;
  $21 = (($17) + ($20<<2)|0);
  $22 = HEAP32[$rect>>2]|0;
  $23 = (($21) + ($22<<2)|0);
  $24 = (($8) + 3)|0;
  $25 = (($24|0) / 4)&-1;
  $26 = $8 & 3;
  $27 = ((($1)) + 12|0);
  $28 = ((($1)) + 32|0);
  $29 = ((($1)) + 28|0);
  $30 = ((($1)) + 16|0);
  $31 = ((($1)) + 20|0);
  $32 = ((($1)) + 24|0);
  $33 = $r&255;
  $34 = $g&255;
  $35 = $b&255;
  $$in27 = $10;$pixel$016 = $23;
  L31: while(1) {
   $36 = (($$in27) + -1)|0;
   switch ($26|0) {
   case 0:  {
    $107 = HEAP16[$28>>1]|0;
    $108 = HEAP32[$28>>2]|0;
    $109 = $108&65535;
    $110 = $108&255;
    $111 = HEAP16[$29>>1]|0;
    $112 = HEAP32[$29>>2]|0;
    $113 = $112&65535;
    $114 = $112&255;
    $$phi$trans$insert80 = $112 & 255;
    $$phi$trans$insert81 = (125924 + ($$phi$trans$insert80<<2)|0);
    $$pre82 = HEAP32[$$phi$trans$insert81>>2]|0;
    $115 = ($109&65535) >>> 8;
    $116 = $115&255;
    $117 = ($113&65535) >>> 8;
    $118 = $117&255;
    $$phi$trans$insert83 = $117&65535;
    $$phi$trans$insert84 = (125924 + ($$phi$trans$insert83<<2)|0);
    $$pre85 = HEAP32[$$phi$trans$insert84>>2]|0;
    $119 = $108 >>> 16;
    $120 = $119&255;
    $121 = $112 >>> 16;
    $122 = $121&255;
    $$phi$trans$insert86 = $121 & 255;
    $$phi$trans$insert87 = (125924 + ($$phi$trans$insert86<<2)|0);
    $$pre88 = HEAP32[$$phi$trans$insert87>>2]|0;
    $123 = $108 >>> 24;
    $124 = $123&255;
    $125 = $112 >>> 24;
    $126 = $125&255;
    $$phi$trans$insert90 = (125924 + ($125<<2)|0);
    $$pre91 = HEAP32[$$phi$trans$insert90>>2]|0;
    $127 = $108&255;
    $128 = $112&255;
    $129 = ($107&65535) >>> 8;
    $130 = $129&255;
    $131 = ($111&65535) >>> 8;
    $132 = $131&255;
    $137 = $110;$140 = $114;$142 = $$pre82;$1460 = $132;$1461 = $130;$1462 = $128;$1463 = $127;$148 = $116;$151 = $118;$153 = $$pre85;$159 = $120;$162 = $122;$164 = $$pre88;$170 = $124;$173 = $126;$175 = $$pre91;$n$0 = $25;$pixel$1 = $pixel$016;
    label = 10;
    break;
   }
   case 3:  {
    $81 = HEAP16[$28>>1]|0;
    $82 = HEAP32[$28>>2]|0;
    $83 = $82&65535;
    $84 = $82&255;
    $85 = HEAP16[$29>>1]|0;
    $86 = HEAP32[$29>>2]|0;
    $87 = $86&65535;
    $88 = $86&255;
    $89 = ($83&65535) >>> 8;
    $90 = $89&255;
    $91 = ($87&65535) >>> 8;
    $92 = $91&255;
    $93 = $82 >>> 16;
    $94 = $93&255;
    $95 = $86 >>> 16;
    $96 = $95&255;
    $97 = $82 >>> 24;
    $98 = $97&255;
    $99 = $86 >>> 24;
    $100 = $99&255;
    $101 = $82&255;
    $102 = $86&255;
    $103 = ($81&65535) >>> 8;
    $104 = $103&255;
    $105 = ($85&65535) >>> 8;
    $106 = $105&255;
    $1456 = $106;$1457 = $104;$1458 = $102;$1459 = $101;$206 = $84;$209 = $88;$218 = $90;$221 = $92;$230 = $94;$233 = $96;$242 = $98;$245 = $100;$n$1 = $25;$pixel$2 = $pixel$016;
    label = 11;
    break;
   }
   case 2:  {
    $55 = HEAP16[$28>>1]|0;
    $56 = HEAP32[$28>>2]|0;
    $57 = $56&65535;
    $58 = $56&255;
    $59 = HEAP16[$29>>1]|0;
    $60 = HEAP32[$29>>2]|0;
    $61 = $60&65535;
    $62 = $60&255;
    $63 = ($57&65535) >>> 8;
    $64 = $63&255;
    $65 = ($61&65535) >>> 8;
    $66 = $65&255;
    $67 = $56 >>> 16;
    $68 = $67&255;
    $69 = $60 >>> 16;
    $70 = $69&255;
    $71 = $56 >>> 24;
    $72 = $71&255;
    $73 = $60 >>> 24;
    $74 = $73&255;
    $75 = $56&255;
    $76 = $60&255;
    $77 = ($55&65535) >>> 8;
    $78 = $77&255;
    $79 = ($59&65535) >>> 8;
    $80 = $79&255;
    $1452 = $80;$1453 = $78;$1454 = $76;$1455 = $75;$279 = $58;$282 = $62;$291 = $64;$294 = $66;$303 = $68;$306 = $70;$315 = $72;$318 = $74;$n$2 = $25;$pixel$3 = $pixel$016;
    label = 12;
    break;
   }
   case 1:  {
    $37 = HEAP32[$28>>2]|0;
    $38 = $37&65535;
    $39 = $37&255;
    $40 = HEAP32[$29>>2]|0;
    $41 = $40&65535;
    $42 = $40&255;
    $43 = ($38&65535) >>> 8;
    $44 = $43&255;
    $45 = ($41&65535) >>> 8;
    $46 = $45&255;
    $47 = $37 >>> 16;
    $48 = $47&255;
    $49 = $40 >>> 16;
    $50 = $49&255;
    $51 = $37 >>> 24;
    $52 = $51&255;
    $53 = $40 >>> 24;
    $54 = $53&255;
    $352 = $39;$355 = $42;$364 = $44;$367 = $46;$376 = $48;$379 = $50;$388 = $52;$391 = $54;$n$3 = $25;$pixel$4 = $pixel$016;
    label = 13;
    break;
   }
   default: {
    label = 14;
    break L31;
   }
   }
   while(1) {
    if ((label|0) == 10) {
     label = 0;
     $133 = HEAP32[$pixel$1>>2]|0;
     $134 = HEAP32[$27>>2]|0;
     $135 = $134 & $133;
     $136 = $137&255;
     $138 = $135 >>> $136;
     $139 = $140&255;
     $141 = (($142) + ($138)|0);
     $143 = HEAP8[$141>>0]|0;
     $144 = $143&255;
     $145 = HEAP32[$30>>2]|0;
     $146 = $145 & $133;
     $147 = $148&255;
     $149 = $146 >>> $147;
     $150 = $151&255;
     $152 = (($153) + ($149)|0);
     $154 = HEAP8[$152>>0]|0;
     $155 = $154&255;
     $156 = HEAP32[$31>>2]|0;
     $157 = $156 & $133;
     $158 = $159&255;
     $160 = $157 >>> $158;
     $161 = $162&255;
     $163 = (($164) + ($160)|0);
     $165 = HEAP8[$163>>0]|0;
     $166 = $165&255;
     $167 = HEAP32[$32>>2]|0;
     $168 = $167 & $133;
     $169 = $170&255;
     $171 = $168 >>> $169;
     $172 = $173&255;
     $174 = (($175) + ($171)|0);
     $176 = HEAP8[$174>>0]|0;
     $177 = $176&255;
     $178 = Math_imul($144, $3)|0;
     $179 = (($178>>>0) / 255)&-1;
     $180 = (($179) + ($33))|0;
     $181 = Math_imul($155, $3)|0;
     $182 = (($181>>>0) / 255)&-1;
     $183 = (($182) + ($34))|0;
     $184 = Math_imul($166, $3)|0;
     $185 = (($184>>>0) / 255)&-1;
     $186 = (($185) + ($35))|0;
     $187 = Math_imul($177, $3)|0;
     $188 = (($187>>>0) / 255)&-1;
     $189 = (($188) + ($2))|0;
     $190 = $180 >>> $139;
     $191 = $190 << $136;
     $192 = $183 >>> $150;
     $193 = $192 << $147;
     $194 = $193 | $191;
     $195 = $186 >>> $161;
     $196 = $195 << $158;
     $197 = $194 | $196;
     $198 = $189 >>> $172;
     $199 = $198 << $169;
     $200 = $197 | $199;
     HEAP32[$pixel$1>>2] = $200;
     $201 = ((($pixel$1)) + 4|0);
     $1456 = $1460;$1457 = $1461;$1458 = $1462;$1459 = $1463;$206 = $1463;$209 = $1462;$218 = $1461;$221 = $1460;$230 = $159;$233 = $162;$242 = $170;$245 = $173;$n$1 = $n$0;$pixel$2 = $201;
     label = 11;
     continue;
    }
    else if ((label|0) == 11) {
     label = 0;
     $202 = HEAP32[$pixel$2>>2]|0;
     $203 = HEAP32[$27>>2]|0;
     $204 = $203 & $202;
     $205 = $206&255;
     $207 = $204 >>> $205;
     $208 = $209&255;
     $210 = (125924 + ($208<<2)|0);
     $211 = HEAP32[$210>>2]|0;
     $212 = (($211) + ($207)|0);
     $213 = HEAP8[$212>>0]|0;
     $214 = $213&255;
     $215 = HEAP32[$30>>2]|0;
     $216 = $215 & $202;
     $217 = $218&255;
     $219 = $216 >>> $217;
     $220 = $221&255;
     $222 = (125924 + ($220<<2)|0);
     $223 = HEAP32[$222>>2]|0;
     $224 = (($223) + ($219)|0);
     $225 = HEAP8[$224>>0]|0;
     $226 = $225&255;
     $227 = HEAP32[$31>>2]|0;
     $228 = $227 & $202;
     $229 = $230&255;
     $231 = $228 >>> $229;
     $232 = $233&255;
     $234 = (125924 + ($232<<2)|0);
     $235 = HEAP32[$234>>2]|0;
     $236 = (($235) + ($231)|0);
     $237 = HEAP8[$236>>0]|0;
     $238 = $237&255;
     $239 = HEAP32[$32>>2]|0;
     $240 = $239 & $202;
     $241 = $242&255;
     $243 = $240 >>> $241;
     $244 = $245&255;
     $246 = (125924 + ($244<<2)|0);
     $247 = HEAP32[$246>>2]|0;
     $248 = (($247) + ($243)|0);
     $249 = HEAP8[$248>>0]|0;
     $250 = $249&255;
     $251 = Math_imul($214, $3)|0;
     $252 = (($251>>>0) / 255)&-1;
     $253 = (($252) + ($33))|0;
     $254 = Math_imul($226, $3)|0;
     $255 = (($254>>>0) / 255)&-1;
     $256 = (($255) + ($34))|0;
     $257 = Math_imul($238, $3)|0;
     $258 = (($257>>>0) / 255)&-1;
     $259 = (($258) + ($35))|0;
     $260 = Math_imul($250, $3)|0;
     $261 = (($260>>>0) / 255)&-1;
     $262 = (($261) + ($2))|0;
     $263 = $253 >>> $208;
     $264 = $263 << $205;
     $265 = $256 >>> $220;
     $266 = $265 << $217;
     $267 = $266 | $264;
     $268 = $259 >>> $232;
     $269 = $268 << $229;
     $270 = $267 | $269;
     $271 = $262 >>> $244;
     $272 = $271 << $241;
     $273 = $270 | $272;
     HEAP32[$pixel$2>>2] = $273;
     $274 = ((($pixel$2)) + 4|0);
     $1452 = $1456;$1453 = $1457;$1454 = $1458;$1455 = $1459;$279 = $1459;$282 = $1458;$291 = $1457;$294 = $1456;$303 = $230;$306 = $233;$315 = $242;$318 = $245;$n$2 = $n$1;$pixel$3 = $274;
     label = 12;
     continue;
    }
    else if ((label|0) == 12) {
     label = 0;
     $275 = HEAP32[$pixel$3>>2]|0;
     $276 = HEAP32[$27>>2]|0;
     $277 = $276 & $275;
     $278 = $279&255;
     $280 = $277 >>> $278;
     $281 = $282&255;
     $283 = (125924 + ($281<<2)|0);
     $284 = HEAP32[$283>>2]|0;
     $285 = (($284) + ($280)|0);
     $286 = HEAP8[$285>>0]|0;
     $287 = $286&255;
     $288 = HEAP32[$30>>2]|0;
     $289 = $288 & $275;
     $290 = $291&255;
     $292 = $289 >>> $290;
     $293 = $294&255;
     $295 = (125924 + ($293<<2)|0);
     $296 = HEAP32[$295>>2]|0;
     $297 = (($296) + ($292)|0);
     $298 = HEAP8[$297>>0]|0;
     $299 = $298&255;
     $300 = HEAP32[$31>>2]|0;
     $301 = $300 & $275;
     $302 = $303&255;
     $304 = $301 >>> $302;
     $305 = $306&255;
     $307 = (125924 + ($305<<2)|0);
     $308 = HEAP32[$307>>2]|0;
     $309 = (($308) + ($304)|0);
     $310 = HEAP8[$309>>0]|0;
     $311 = $310&255;
     $312 = HEAP32[$32>>2]|0;
     $313 = $312 & $275;
     $314 = $315&255;
     $316 = $313 >>> $314;
     $317 = $318&255;
     $319 = (125924 + ($317<<2)|0);
     $320 = HEAP32[$319>>2]|0;
     $321 = (($320) + ($316)|0);
     $322 = HEAP8[$321>>0]|0;
     $323 = $322&255;
     $324 = Math_imul($287, $3)|0;
     $325 = (($324>>>0) / 255)&-1;
     $326 = (($325) + ($33))|0;
     $327 = Math_imul($299, $3)|0;
     $328 = (($327>>>0) / 255)&-1;
     $329 = (($328) + ($34))|0;
     $330 = Math_imul($311, $3)|0;
     $331 = (($330>>>0) / 255)&-1;
     $332 = (($331) + ($35))|0;
     $333 = Math_imul($323, $3)|0;
     $334 = (($333>>>0) / 255)&-1;
     $335 = (($334) + ($2))|0;
     $336 = $326 >>> $281;
     $337 = $336 << $278;
     $338 = $329 >>> $293;
     $339 = $338 << $290;
     $340 = $339 | $337;
     $341 = $332 >>> $305;
     $342 = $341 << $302;
     $343 = $340 | $342;
     $344 = $335 >>> $317;
     $345 = $344 << $314;
     $346 = $343 | $345;
     HEAP32[$pixel$3>>2] = $346;
     $347 = ((($pixel$3)) + 4|0);
     $352 = $1455;$355 = $1454;$364 = $1453;$367 = $1452;$376 = $303;$379 = $306;$388 = $315;$391 = $318;$n$3 = $n$2;$pixel$4 = $347;
     label = 13;
     continue;
    }
    else if ((label|0) == 13) {
     label = 0;
     $348 = HEAP32[$pixel$4>>2]|0;
     $349 = HEAP32[$27>>2]|0;
     $350 = $349 & $348;
     $351 = $352&255;
     $353 = $350 >>> $351;
     $354 = $355&255;
     $356 = (125924 + ($354<<2)|0);
     $357 = HEAP32[$356>>2]|0;
     $358 = (($357) + ($353)|0);
     $359 = HEAP8[$358>>0]|0;
     $360 = $359&255;
     $361 = HEAP32[$30>>2]|0;
     $362 = $361 & $348;
     $363 = $364&255;
     $365 = $362 >>> $363;
     $366 = $367&255;
     $368 = (125924 + ($366<<2)|0);
     $369 = HEAP32[$368>>2]|0;
     $370 = (($369) + ($365)|0);
     $371 = HEAP8[$370>>0]|0;
     $372 = $371&255;
     $373 = HEAP32[$31>>2]|0;
     $374 = $373 & $348;
     $375 = $376&255;
     $377 = $374 >>> $375;
     $378 = $379&255;
     $380 = (125924 + ($378<<2)|0);
     $381 = HEAP32[$380>>2]|0;
     $382 = (($381) + ($377)|0);
     $383 = HEAP8[$382>>0]|0;
     $384 = $383&255;
     $385 = HEAP32[$32>>2]|0;
     $386 = $385 & $348;
     $387 = $388&255;
     $389 = $386 >>> $387;
     $390 = $391&255;
     $392 = (125924 + ($390<<2)|0);
     $393 = HEAP32[$392>>2]|0;
     $394 = (($393) + ($389)|0);
     $395 = HEAP8[$394>>0]|0;
     $396 = $395&255;
     $397 = Math_imul($360, $3)|0;
     $398 = (($397>>>0) / 255)&-1;
     $399 = (($398) + ($33))|0;
     $400 = Math_imul($372, $3)|0;
     $401 = (($400>>>0) / 255)&-1;
     $402 = (($401) + ($34))|0;
     $403 = Math_imul($384, $3)|0;
     $404 = (($403>>>0) / 255)&-1;
     $405 = (($404) + ($35))|0;
     $406 = Math_imul($396, $3)|0;
     $407 = (($406>>>0) / 255)&-1;
     $408 = (($407) + ($2))|0;
     $409 = $399 >>> $354;
     $410 = $409 << $351;
     $411 = $402 >>> $366;
     $412 = $411 << $363;
     $413 = $412 | $410;
     $414 = $405 >>> $378;
     $415 = $414 << $375;
     $416 = $413 | $415;
     $417 = $408 >>> $390;
     $418 = $417 << $387;
     $419 = $416 | $418;
     HEAP32[$pixel$4>>2] = $419;
     $420 = ((($pixel$4)) + 4|0);
     $421 = (($n$3) + -1)|0;
     $422 = ($n$3|0)>(1);
     if ($422) {
      $137 = $352;$140 = $355;$142 = $357;$1460 = $367;$1461 = $364;$1462 = $355;$1463 = $352;$148 = $364;$151 = $367;$153 = $369;$159 = $376;$162 = $379;$164 = $381;$170 = $388;$173 = $391;$175 = $393;$n$0 = $421;$pixel$1 = $420;
      label = 10;
      continue;
     } else {
      break;
     }
    }
   }
   $423 = (($420) + ($14<<2)|0);
   $424 = ($36|0)==(0);
   if ($424) {
    $$0 = 0;
    label = 56;
    break;
   } else {
    $$in27 = $36;$pixel$016 = $423;
   }
  }
  if ((label|0) == 14) {
   // unreachable;
  }
  else if ((label|0) == 56) {
   return ($$0|0);
  }
  break;
 }
 case 2:  {
  $425 = ((($rect)) + 8|0);
  $426 = HEAP32[$425>>2]|0;
  $427 = ((($rect)) + 12|0);
  $428 = HEAP32[$427>>2]|0;
  $429 = ((($dst)) + 16|0);
  $430 = HEAP32[$429>>2]|0;
  $431 = (($430|0) / ($6|0))&-1;
  $432 = (($431) - ($426))|0;
  $433 = ($428|0)==(0);
  if ($433) {
   $$0 = 0;
   return ($$0|0);
  }
  $434 = ((($dst)) + 20|0);
  $435 = HEAP32[$434>>2]|0;
  $436 = ((($rect)) + 4|0);
  $437 = HEAP32[$436>>2]|0;
  $438 = Math_imul($437, $431)|0;
  $439 = (($435) + ($438<<2)|0);
  $440 = HEAP32[$rect>>2]|0;
  $441 = (($439) + ($440<<2)|0);
  $442 = (($426) + 3)|0;
  $443 = (($442|0) / 4)&-1;
  $444 = $426 & 3;
  $445 = ((($1)) + 12|0);
  $446 = ((($1)) + 32|0);
  $447 = ((($1)) + 28|0);
  $448 = ((($1)) + 16|0);
  $449 = ((($1)) + 20|0);
  $450 = ((($1)) + 24|0);
  $451 = $r&255;
  $452 = $g&255;
  $453 = $b&255;
  $$in26 = $428;$pixel17$020 = $441;
  L52: while(1) {
   $454 = (($$in26) + -1)|0;
   switch ($444|0) {
   case 0:  {
    $525 = HEAP16[$446>>1]|0;
    $526 = HEAP32[$446>>2]|0;
    $527 = $526&65535;
    $528 = $526&255;
    $529 = HEAP16[$447>>1]|0;
    $530 = HEAP32[$447>>2]|0;
    $531 = $530&65535;
    $532 = $530&255;
    $$phi$trans$insert56 = $530 & 255;
    $$phi$trans$insert57 = (125924 + ($$phi$trans$insert56<<2)|0);
    $$pre58 = HEAP32[$$phi$trans$insert57>>2]|0;
    $533 = ($527&65535) >>> 8;
    $534 = $533&255;
    $535 = ($531&65535) >>> 8;
    $536 = $535&255;
    $$phi$trans$insert59 = $535&65535;
    $$phi$trans$insert60 = (125924 + ($$phi$trans$insert59<<2)|0);
    $$pre61 = HEAP32[$$phi$trans$insert60>>2]|0;
    $537 = $526 >>> 16;
    $538 = $537&255;
    $539 = $530 >>> 16;
    $540 = $539&255;
    $$phi$trans$insert62 = $539 & 255;
    $$phi$trans$insert63 = (125924 + ($$phi$trans$insert62<<2)|0);
    $$pre64 = HEAP32[$$phi$trans$insert63>>2]|0;
    $541 = $526 >>> 24;
    $542 = $541&255;
    $543 = $530 >>> 24;
    $544 = $543&255;
    $$phi$trans$insert66 = (125924 + ($543<<2)|0);
    $$pre67 = HEAP32[$$phi$trans$insert66>>2]|0;
    $545 = $526&255;
    $546 = $530&255;
    $547 = ($525&65535) >>> 8;
    $548 = $547&255;
    $549 = ($529&65535) >>> 8;
    $550 = $549&255;
    $1472 = $550;$1473 = $548;$1474 = $546;$1475 = $545;$555 = $528;$558 = $532;$560 = $$pre58;$566 = $534;$569 = $536;$571 = $$pre61;$577 = $538;$580 = $540;$582 = $$pre64;$588 = $542;$591 = $544;$593 = $$pre67;$n18$0 = $443;$pixel17$1 = $pixel17$020;
    label = 23;
    break;
   }
   case 3:  {
    $499 = HEAP16[$446>>1]|0;
    $500 = HEAP32[$446>>2]|0;
    $501 = $500&65535;
    $502 = $500&255;
    $503 = HEAP16[$447>>1]|0;
    $504 = HEAP32[$447>>2]|0;
    $505 = $504&65535;
    $506 = $504&255;
    $507 = ($501&65535) >>> 8;
    $508 = $507&255;
    $509 = ($505&65535) >>> 8;
    $510 = $509&255;
    $511 = $500 >>> 16;
    $512 = $511&255;
    $513 = $504 >>> 16;
    $514 = $513&255;
    $515 = $500 >>> 24;
    $516 = $515&255;
    $517 = $504 >>> 24;
    $518 = $517&255;
    $519 = $500&255;
    $520 = $504&255;
    $521 = ($499&65535) >>> 8;
    $522 = $521&255;
    $523 = ($503&65535) >>> 8;
    $524 = $523&255;
    $1468 = $524;$1469 = $522;$1470 = $520;$1471 = $519;$618 = $502;$621 = $506;$630 = $508;$633 = $510;$642 = $512;$645 = $514;$654 = $516;$657 = $518;$n18$1 = $443;$pixel17$2 = $pixel17$020;
    label = 24;
    break;
   }
   case 2:  {
    $473 = HEAP16[$446>>1]|0;
    $474 = HEAP32[$446>>2]|0;
    $475 = $474&65535;
    $476 = $474&255;
    $477 = HEAP16[$447>>1]|0;
    $478 = HEAP32[$447>>2]|0;
    $479 = $478&65535;
    $480 = $478&255;
    $481 = ($475&65535) >>> 8;
    $482 = $481&255;
    $483 = ($479&65535) >>> 8;
    $484 = $483&255;
    $485 = $474 >>> 16;
    $486 = $485&255;
    $487 = $478 >>> 16;
    $488 = $487&255;
    $489 = $474 >>> 24;
    $490 = $489&255;
    $491 = $478 >>> 24;
    $492 = $491&255;
    $493 = $474&255;
    $494 = $478&255;
    $495 = ($473&65535) >>> 8;
    $496 = $495&255;
    $497 = ($477&65535) >>> 8;
    $498 = $497&255;
    $1464 = $498;$1465 = $496;$1466 = $494;$1467 = $493;$685 = $476;$688 = $480;$697 = $482;$700 = $484;$709 = $486;$712 = $488;$721 = $490;$724 = $492;$n18$2 = $443;$pixel17$3 = $pixel17$020;
    label = 25;
    break;
   }
   case 1:  {
    $455 = HEAP32[$446>>2]|0;
    $456 = $455&65535;
    $457 = $455&255;
    $458 = HEAP32[$447>>2]|0;
    $459 = $458&65535;
    $460 = $458&255;
    $461 = ($456&65535) >>> 8;
    $462 = $461&255;
    $463 = ($459&65535) >>> 8;
    $464 = $463&255;
    $465 = $455 >>> 16;
    $466 = $465&255;
    $467 = $458 >>> 16;
    $468 = $467&255;
    $469 = $455 >>> 24;
    $470 = $469&255;
    $471 = $458 >>> 24;
    $472 = $471&255;
    $752 = $457;$755 = $460;$764 = $462;$767 = $464;$776 = $466;$779 = $468;$788 = $470;$791 = $472;$n18$3 = $443;$pixel17$4 = $pixel17$020;
    label = 26;
    break;
   }
   default: {
    label = 27;
    break L52;
   }
   }
   while(1) {
    if ((label|0) == 23) {
     label = 0;
     $551 = HEAP32[$pixel17$1>>2]|0;
     $552 = HEAP32[$445>>2]|0;
     $553 = $552 & $551;
     $554 = $555&255;
     $556 = $553 >>> $554;
     $557 = $558&255;
     $559 = (($560) + ($556)|0);
     $561 = HEAP8[$559>>0]|0;
     $562 = $561&255;
     $563 = HEAP32[$448>>2]|0;
     $564 = $563 & $551;
     $565 = $566&255;
     $567 = $564 >>> $565;
     $568 = $569&255;
     $570 = (($571) + ($567)|0);
     $572 = HEAP8[$570>>0]|0;
     $573 = $572&255;
     $574 = HEAP32[$449>>2]|0;
     $575 = $574 & $551;
     $576 = $577&255;
     $578 = $575 >>> $576;
     $579 = $580&255;
     $581 = (($582) + ($578)|0);
     $583 = HEAP8[$581>>0]|0;
     $584 = $583&255;
     $585 = HEAP32[$450>>2]|0;
     $586 = $585 & $551;
     $587 = $588&255;
     $589 = $586 >>> $587;
     $590 = $591&255;
     $592 = (($593) + ($589)|0);
     $594 = HEAP8[$592>>0]|0;
     $595 = $594&255;
     $596 = (($562) + ($451))|0;
     $597 = ($596>>>0)>(255);
     $$ = $597 ? 255 : $596;
     $598 = (($573) + ($452))|0;
     $599 = ($598>>>0)>(255);
     $$8 = $599 ? 255 : $598;
     $600 = (($584) + ($453))|0;
     $601 = ($600>>>0)>(255);
     $$1 = $601 ? 255 : $600;
     $602 = $$ >>> $557;
     $603 = $602 << $554;
     $604 = $$8 >>> $568;
     $605 = $604 << $565;
     $606 = $605 | $603;
     $607 = $$1 >>> $579;
     $608 = $607 << $576;
     $609 = $606 | $608;
     $610 = $595 >>> $590;
     $611 = $610 << $587;
     $612 = $609 | $611;
     HEAP32[$pixel17$1>>2] = $612;
     $613 = ((($pixel17$1)) + 4|0);
     $1468 = $1472;$1469 = $1473;$1470 = $1474;$1471 = $1475;$618 = $1475;$621 = $1474;$630 = $1473;$633 = $1472;$642 = $577;$645 = $580;$654 = $588;$657 = $591;$n18$1 = $n18$0;$pixel17$2 = $613;
     label = 24;
     continue;
    }
    else if ((label|0) == 24) {
     label = 0;
     $614 = HEAP32[$pixel17$2>>2]|0;
     $615 = HEAP32[$445>>2]|0;
     $616 = $615 & $614;
     $617 = $618&255;
     $619 = $616 >>> $617;
     $620 = $621&255;
     $622 = (125924 + ($620<<2)|0);
     $623 = HEAP32[$622>>2]|0;
     $624 = (($623) + ($619)|0);
     $625 = HEAP8[$624>>0]|0;
     $626 = $625&255;
     $627 = HEAP32[$448>>2]|0;
     $628 = $627 & $614;
     $629 = $630&255;
     $631 = $628 >>> $629;
     $632 = $633&255;
     $634 = (125924 + ($632<<2)|0);
     $635 = HEAP32[$634>>2]|0;
     $636 = (($635) + ($631)|0);
     $637 = HEAP8[$636>>0]|0;
     $638 = $637&255;
     $639 = HEAP32[$449>>2]|0;
     $640 = $639 & $614;
     $641 = $642&255;
     $643 = $640 >>> $641;
     $644 = $645&255;
     $646 = (125924 + ($644<<2)|0);
     $647 = HEAP32[$646>>2]|0;
     $648 = (($647) + ($643)|0);
     $649 = HEAP8[$648>>0]|0;
     $650 = $649&255;
     $651 = HEAP32[$450>>2]|0;
     $652 = $651 & $614;
     $653 = $654&255;
     $655 = $652 >>> $653;
     $656 = $657&255;
     $658 = (125924 + ($656<<2)|0);
     $659 = HEAP32[$658>>2]|0;
     $660 = (($659) + ($655)|0);
     $661 = HEAP8[$660>>0]|0;
     $662 = $661&255;
     $663 = (($626) + ($451))|0;
     $664 = ($663>>>0)>(255);
     $$2 = $664 ? 255 : $663;
     $665 = (($638) + ($452))|0;
     $666 = ($665>>>0)>(255);
     $$9 = $666 ? 255 : $665;
     $667 = (($650) + ($453))|0;
     $668 = ($667>>>0)>(255);
     $$3 = $668 ? 255 : $667;
     $669 = $$2 >>> $620;
     $670 = $669 << $617;
     $671 = $$9 >>> $632;
     $672 = $671 << $629;
     $673 = $672 | $670;
     $674 = $$3 >>> $644;
     $675 = $674 << $641;
     $676 = $673 | $675;
     $677 = $662 >>> $656;
     $678 = $677 << $653;
     $679 = $676 | $678;
     HEAP32[$pixel17$2>>2] = $679;
     $680 = ((($pixel17$2)) + 4|0);
     $1464 = $1468;$1465 = $1469;$1466 = $1470;$1467 = $1471;$685 = $1471;$688 = $1470;$697 = $1469;$700 = $1468;$709 = $642;$712 = $645;$721 = $654;$724 = $657;$n18$2 = $n18$1;$pixel17$3 = $680;
     label = 25;
     continue;
    }
    else if ((label|0) == 25) {
     label = 0;
     $681 = HEAP32[$pixel17$3>>2]|0;
     $682 = HEAP32[$445>>2]|0;
     $683 = $682 & $681;
     $684 = $685&255;
     $686 = $683 >>> $684;
     $687 = $688&255;
     $689 = (125924 + ($687<<2)|0);
     $690 = HEAP32[$689>>2]|0;
     $691 = (($690) + ($686)|0);
     $692 = HEAP8[$691>>0]|0;
     $693 = $692&255;
     $694 = HEAP32[$448>>2]|0;
     $695 = $694 & $681;
     $696 = $697&255;
     $698 = $695 >>> $696;
     $699 = $700&255;
     $701 = (125924 + ($699<<2)|0);
     $702 = HEAP32[$701>>2]|0;
     $703 = (($702) + ($698)|0);
     $704 = HEAP8[$703>>0]|0;
     $705 = $704&255;
     $706 = HEAP32[$449>>2]|0;
     $707 = $706 & $681;
     $708 = $709&255;
     $710 = $707 >>> $708;
     $711 = $712&255;
     $713 = (125924 + ($711<<2)|0);
     $714 = HEAP32[$713>>2]|0;
     $715 = (($714) + ($710)|0);
     $716 = HEAP8[$715>>0]|0;
     $717 = $716&255;
     $718 = HEAP32[$450>>2]|0;
     $719 = $718 & $681;
     $720 = $721&255;
     $722 = $719 >>> $720;
     $723 = $724&255;
     $725 = (125924 + ($723<<2)|0);
     $726 = HEAP32[$725>>2]|0;
     $727 = (($726) + ($722)|0);
     $728 = HEAP8[$727>>0]|0;
     $729 = $728&255;
     $730 = (($693) + ($451))|0;
     $731 = ($730>>>0)>(255);
     $$4 = $731 ? 255 : $730;
     $732 = (($705) + ($452))|0;
     $733 = ($732>>>0)>(255);
     $$10 = $733 ? 255 : $732;
     $734 = (($717) + ($453))|0;
     $735 = ($734>>>0)>(255);
     $$5 = $735 ? 255 : $734;
     $736 = $$4 >>> $687;
     $737 = $736 << $684;
     $738 = $$10 >>> $699;
     $739 = $738 << $696;
     $740 = $739 | $737;
     $741 = $$5 >>> $711;
     $742 = $741 << $708;
     $743 = $740 | $742;
     $744 = $729 >>> $723;
     $745 = $744 << $720;
     $746 = $743 | $745;
     HEAP32[$pixel17$3>>2] = $746;
     $747 = ((($pixel17$3)) + 4|0);
     $752 = $1467;$755 = $1466;$764 = $1465;$767 = $1464;$776 = $709;$779 = $712;$788 = $721;$791 = $724;$n18$3 = $n18$2;$pixel17$4 = $747;
     label = 26;
     continue;
    }
    else if ((label|0) == 26) {
     label = 0;
     $748 = HEAP32[$pixel17$4>>2]|0;
     $749 = HEAP32[$445>>2]|0;
     $750 = $749 & $748;
     $751 = $752&255;
     $753 = $750 >>> $751;
     $754 = $755&255;
     $756 = (125924 + ($754<<2)|0);
     $757 = HEAP32[$756>>2]|0;
     $758 = (($757) + ($753)|0);
     $759 = HEAP8[$758>>0]|0;
     $760 = $759&255;
     $761 = HEAP32[$448>>2]|0;
     $762 = $761 & $748;
     $763 = $764&255;
     $765 = $762 >>> $763;
     $766 = $767&255;
     $768 = (125924 + ($766<<2)|0);
     $769 = HEAP32[$768>>2]|0;
     $770 = (($769) + ($765)|0);
     $771 = HEAP8[$770>>0]|0;
     $772 = $771&255;
     $773 = HEAP32[$449>>2]|0;
     $774 = $773 & $748;
     $775 = $776&255;
     $777 = $774 >>> $775;
     $778 = $779&255;
     $780 = (125924 + ($778<<2)|0);
     $781 = HEAP32[$780>>2]|0;
     $782 = (($781) + ($777)|0);
     $783 = HEAP8[$782>>0]|0;
     $784 = $783&255;
     $785 = HEAP32[$450>>2]|0;
     $786 = $785 & $748;
     $787 = $788&255;
     $789 = $786 >>> $787;
     $790 = $791&255;
     $792 = (125924 + ($790<<2)|0);
     $793 = HEAP32[$792>>2]|0;
     $794 = (($793) + ($789)|0);
     $795 = HEAP8[$794>>0]|0;
     $796 = $795&255;
     $797 = (($760) + ($451))|0;
     $798 = ($797>>>0)>(255);
     $$6 = $798 ? 255 : $797;
     $799 = (($772) + ($452))|0;
     $800 = ($799>>>0)>(255);
     $$11 = $800 ? 255 : $799;
     $801 = (($784) + ($453))|0;
     $802 = ($801>>>0)>(255);
     $$7 = $802 ? 255 : $801;
     $803 = $$6 >>> $754;
     $804 = $803 << $751;
     $805 = $$11 >>> $766;
     $806 = $805 << $763;
     $807 = $806 | $804;
     $808 = $$7 >>> $778;
     $809 = $808 << $775;
     $810 = $807 | $809;
     $811 = $796 >>> $790;
     $812 = $811 << $787;
     $813 = $810 | $812;
     HEAP32[$pixel17$4>>2] = $813;
     $814 = ((($pixel17$4)) + 4|0);
     $815 = (($n18$3) + -1)|0;
     $816 = ($n18$3|0)>(1);
     if ($816) {
      $1472 = $767;$1473 = $764;$1474 = $755;$1475 = $752;$555 = $752;$558 = $755;$560 = $757;$566 = $764;$569 = $767;$571 = $769;$577 = $776;$580 = $779;$582 = $781;$588 = $788;$591 = $791;$593 = $793;$n18$0 = $815;$pixel17$1 = $814;
      label = 23;
      continue;
     } else {
      break;
     }
    }
   }
   $817 = (($814) + ($432<<2)|0);
   $818 = ($454|0)==(0);
   if ($818) {
    $$0 = 0;
    label = 56;
    break;
   } else {
    $$in26 = $454;$pixel17$020 = $817;
   }
  }
  if ((label|0) == 27) {
   // unreachable;
  }
  else if ((label|0) == 56) {
   return ($$0|0);
  }
  break;
 }
 case 4:  {
  $819 = ((($rect)) + 8|0);
  $820 = HEAP32[$819>>2]|0;
  $821 = ((($rect)) + 12|0);
  $822 = HEAP32[$821>>2]|0;
  $823 = ((($dst)) + 16|0);
  $824 = HEAP32[$823>>2]|0;
  $825 = (($824|0) / ($6|0))&-1;
  $826 = (($825) - ($820))|0;
  $827 = ($822|0)==(0);
  if ($827) {
   $$0 = 0;
   return ($$0|0);
  }
  $828 = ((($dst)) + 20|0);
  $829 = HEAP32[$828>>2]|0;
  $830 = ((($rect)) + 4|0);
  $831 = HEAP32[$830>>2]|0;
  $832 = Math_imul($831, $825)|0;
  $833 = (($829) + ($832<<2)|0);
  $834 = HEAP32[$rect>>2]|0;
  $835 = (($833) + ($834<<2)|0);
  $836 = (($820) + 3)|0;
  $837 = (($836|0) / 4)&-1;
  $838 = $820 & 3;
  $839 = ((($1)) + 12|0);
  $840 = ((($1)) + 32|0);
  $841 = ((($1)) + 28|0);
  $842 = ((($1)) + 16|0);
  $843 = ((($1)) + 20|0);
  $844 = ((($1)) + 24|0);
  $845 = $r&255;
  $846 = $g&255;
  $847 = $b&255;
  $$in = $822;$pixel39$024 = $835;
  L73: while(1) {
   $848 = (($$in) + -1)|0;
   switch ($838|0) {
   case 0:  {
    $919 = HEAP16[$840>>1]|0;
    $920 = HEAP32[$840>>2]|0;
    $921 = $920&65535;
    $922 = $920&255;
    $923 = HEAP16[$841>>1]|0;
    $924 = HEAP32[$841>>2]|0;
    $925 = $924&65535;
    $926 = $924&255;
    $$phi$trans$insert = $924 & 255;
    $$phi$trans$insert33 = (125924 + ($$phi$trans$insert<<2)|0);
    $$pre34 = HEAP32[$$phi$trans$insert33>>2]|0;
    $927 = ($921&65535) >>> 8;
    $928 = $927&255;
    $929 = ($925&65535) >>> 8;
    $930 = $929&255;
    $$phi$trans$insert35 = $929&65535;
    $$phi$trans$insert36 = (125924 + ($$phi$trans$insert35<<2)|0);
    $$pre37 = HEAP32[$$phi$trans$insert36>>2]|0;
    $931 = $920 >>> 16;
    $932 = $931&255;
    $933 = $924 >>> 16;
    $934 = $933&255;
    $$phi$trans$insert38 = $933 & 255;
    $$phi$trans$insert39 = (125924 + ($$phi$trans$insert38<<2)|0);
    $$pre40 = HEAP32[$$phi$trans$insert39>>2]|0;
    $935 = $920 >>> 24;
    $936 = $935&255;
    $937 = $924 >>> 24;
    $938 = $937&255;
    $$phi$trans$insert42 = (125924 + ($937<<2)|0);
    $$pre43 = HEAP32[$$phi$trans$insert42>>2]|0;
    $939 = $920&255;
    $940 = $924&255;
    $941 = ($919&65535) >>> 8;
    $942 = $941&255;
    $943 = ($923&65535) >>> 8;
    $944 = $943&255;
    $1484 = $944;$1485 = $942;$1486 = $940;$1487 = $939;$949 = $922;$952 = $926;$954 = $$pre34;$960 = $928;$963 = $930;$965 = $$pre37;$971 = $932;$974 = $934;$976 = $$pre40;$982 = $936;$985 = $938;$987 = $$pre43;$n40$0 = $837;$pixel39$1 = $pixel39$024;
    label = 36;
    break;
   }
   case 3:  {
    $893 = HEAP16[$840>>1]|0;
    $894 = HEAP32[$840>>2]|0;
    $895 = $894&65535;
    $896 = $894&255;
    $897 = HEAP16[$841>>1]|0;
    $898 = HEAP32[$841>>2]|0;
    $899 = $898&65535;
    $900 = $898&255;
    $901 = ($895&65535) >>> 8;
    $902 = $901&255;
    $903 = ($899&65535) >>> 8;
    $904 = $903&255;
    $905 = $894 >>> 16;
    $906 = $905&255;
    $907 = $898 >>> 16;
    $908 = $907&255;
    $909 = $894 >>> 24;
    $910 = $909&255;
    $911 = $898 >>> 24;
    $912 = $911&255;
    $913 = $894&255;
    $914 = $898&255;
    $915 = ($893&65535) >>> 8;
    $916 = $915&255;
    $917 = ($897&65535) >>> 8;
    $918 = $917&255;
    $1012 = $896;$1015 = $900;$1024 = $902;$1027 = $904;$1036 = $906;$1039 = $908;$1048 = $910;$1051 = $912;$1480 = $918;$1481 = $916;$1482 = $914;$1483 = $913;$n40$1 = $837;$pixel39$2 = $pixel39$024;
    label = 37;
    break;
   }
   case 2:  {
    $867 = HEAP16[$840>>1]|0;
    $868 = HEAP32[$840>>2]|0;
    $869 = $868&65535;
    $870 = $868&255;
    $871 = HEAP16[$841>>1]|0;
    $872 = HEAP32[$841>>2]|0;
    $873 = $872&65535;
    $874 = $872&255;
    $875 = ($869&65535) >>> 8;
    $876 = $875&255;
    $877 = ($873&65535) >>> 8;
    $878 = $877&255;
    $879 = $868 >>> 16;
    $880 = $879&255;
    $881 = $872 >>> 16;
    $882 = $881&255;
    $883 = $868 >>> 24;
    $884 = $883&255;
    $885 = $872 >>> 24;
    $886 = $885&255;
    $887 = $868&255;
    $888 = $872&255;
    $889 = ($867&65535) >>> 8;
    $890 = $889&255;
    $891 = ($871&65535) >>> 8;
    $892 = $891&255;
    $1079 = $870;$1082 = $874;$1091 = $876;$1094 = $878;$1103 = $880;$1106 = $882;$1115 = $884;$1118 = $886;$1476 = $892;$1477 = $890;$1478 = $888;$1479 = $887;$n40$2 = $837;$pixel39$3 = $pixel39$024;
    label = 38;
    break;
   }
   case 1:  {
    $849 = HEAP32[$840>>2]|0;
    $850 = $849&65535;
    $851 = $849&255;
    $852 = HEAP32[$841>>2]|0;
    $853 = $852&65535;
    $854 = $852&255;
    $855 = ($850&65535) >>> 8;
    $856 = $855&255;
    $857 = ($853&65535) >>> 8;
    $858 = $857&255;
    $859 = $849 >>> 16;
    $860 = $859&255;
    $861 = $852 >>> 16;
    $862 = $861&255;
    $863 = $849 >>> 24;
    $864 = $863&255;
    $865 = $852 >>> 24;
    $866 = $865&255;
    $1146 = $851;$1149 = $854;$1158 = $856;$1161 = $858;$1170 = $860;$1173 = $862;$1182 = $864;$1185 = $866;$n40$3 = $837;$pixel39$4 = $pixel39$024;
    label = 39;
    break;
   }
   default: {
    label = 40;
    break L73;
   }
   }
   while(1) {
    if ((label|0) == 36) {
     label = 0;
     $945 = HEAP32[$pixel39$1>>2]|0;
     $946 = HEAP32[$839>>2]|0;
     $947 = $946 & $945;
     $948 = $949&255;
     $950 = $947 >>> $948;
     $951 = $952&255;
     $953 = (($954) + ($950)|0);
     $955 = HEAP8[$953>>0]|0;
     $956 = $955&255;
     $957 = HEAP32[$842>>2]|0;
     $958 = $957 & $945;
     $959 = $960&255;
     $961 = $958 >>> $959;
     $962 = $963&255;
     $964 = (($965) + ($961)|0);
     $966 = HEAP8[$964>>0]|0;
     $967 = $966&255;
     $968 = HEAP32[$843>>2]|0;
     $969 = $968 & $945;
     $970 = $971&255;
     $972 = $969 >>> $970;
     $973 = $974&255;
     $975 = (($976) + ($972)|0);
     $977 = HEAP8[$975>>0]|0;
     $978 = $977&255;
     $979 = HEAP32[$844>>2]|0;
     $980 = $979 & $945;
     $981 = $982&255;
     $983 = $980 >>> $981;
     $984 = $985&255;
     $986 = (($987) + ($983)|0);
     $988 = HEAP8[$986>>0]|0;
     $989 = $988&255;
     $990 = Math_imul($956, $845)|0;
     $991 = (($990>>>0) / 255)&-1;
     $992 = Math_imul($967, $846)|0;
     $993 = (($992>>>0) / 255)&-1;
     $994 = Math_imul($978, $847)|0;
     $995 = (($994>>>0) / 255)&-1;
     $996 = $991 >>> $951;
     $997 = $996 << $948;
     $998 = $993 >>> $962;
     $999 = $998 << $959;
     $1000 = $999 | $997;
     $1001 = $995 >>> $973;
     $1002 = $1001 << $970;
     $1003 = $1000 | $1002;
     $1004 = $989 >>> $984;
     $1005 = $1004 << $981;
     $1006 = $1003 | $1005;
     HEAP32[$pixel39$1>>2] = $1006;
     $1007 = ((($pixel39$1)) + 4|0);
     $1012 = $1487;$1015 = $1486;$1024 = $1485;$1027 = $1484;$1036 = $971;$1039 = $974;$1048 = $982;$1051 = $985;$1480 = $1484;$1481 = $1485;$1482 = $1486;$1483 = $1487;$n40$1 = $n40$0;$pixel39$2 = $1007;
     label = 37;
     continue;
    }
    else if ((label|0) == 37) {
     label = 0;
     $1008 = HEAP32[$pixel39$2>>2]|0;
     $1009 = HEAP32[$839>>2]|0;
     $1010 = $1009 & $1008;
     $1011 = $1012&255;
     $1013 = $1010 >>> $1011;
     $1014 = $1015&255;
     $1016 = (125924 + ($1014<<2)|0);
     $1017 = HEAP32[$1016>>2]|0;
     $1018 = (($1017) + ($1013)|0);
     $1019 = HEAP8[$1018>>0]|0;
     $1020 = $1019&255;
     $1021 = HEAP32[$842>>2]|0;
     $1022 = $1021 & $1008;
     $1023 = $1024&255;
     $1025 = $1022 >>> $1023;
     $1026 = $1027&255;
     $1028 = (125924 + ($1026<<2)|0);
     $1029 = HEAP32[$1028>>2]|0;
     $1030 = (($1029) + ($1025)|0);
     $1031 = HEAP8[$1030>>0]|0;
     $1032 = $1031&255;
     $1033 = HEAP32[$843>>2]|0;
     $1034 = $1033 & $1008;
     $1035 = $1036&255;
     $1037 = $1034 >>> $1035;
     $1038 = $1039&255;
     $1040 = (125924 + ($1038<<2)|0);
     $1041 = HEAP32[$1040>>2]|0;
     $1042 = (($1041) + ($1037)|0);
     $1043 = HEAP8[$1042>>0]|0;
     $1044 = $1043&255;
     $1045 = HEAP32[$844>>2]|0;
     $1046 = $1045 & $1008;
     $1047 = $1048&255;
     $1049 = $1046 >>> $1047;
     $1050 = $1051&255;
     $1052 = (125924 + ($1050<<2)|0);
     $1053 = HEAP32[$1052>>2]|0;
     $1054 = (($1053) + ($1049)|0);
     $1055 = HEAP8[$1054>>0]|0;
     $1056 = $1055&255;
     $1057 = Math_imul($1020, $845)|0;
     $1058 = (($1057>>>0) / 255)&-1;
     $1059 = Math_imul($1032, $846)|0;
     $1060 = (($1059>>>0) / 255)&-1;
     $1061 = Math_imul($1044, $847)|0;
     $1062 = (($1061>>>0) / 255)&-1;
     $1063 = $1058 >>> $1014;
     $1064 = $1063 << $1011;
     $1065 = $1060 >>> $1026;
     $1066 = $1065 << $1023;
     $1067 = $1066 | $1064;
     $1068 = $1062 >>> $1038;
     $1069 = $1068 << $1035;
     $1070 = $1067 | $1069;
     $1071 = $1056 >>> $1050;
     $1072 = $1071 << $1047;
     $1073 = $1070 | $1072;
     HEAP32[$pixel39$2>>2] = $1073;
     $1074 = ((($pixel39$2)) + 4|0);
     $1079 = $1483;$1082 = $1482;$1091 = $1481;$1094 = $1480;$1103 = $1036;$1106 = $1039;$1115 = $1048;$1118 = $1051;$1476 = $1480;$1477 = $1481;$1478 = $1482;$1479 = $1483;$n40$2 = $n40$1;$pixel39$3 = $1074;
     label = 38;
     continue;
    }
    else if ((label|0) == 38) {
     label = 0;
     $1075 = HEAP32[$pixel39$3>>2]|0;
     $1076 = HEAP32[$839>>2]|0;
     $1077 = $1076 & $1075;
     $1078 = $1079&255;
     $1080 = $1077 >>> $1078;
     $1081 = $1082&255;
     $1083 = (125924 + ($1081<<2)|0);
     $1084 = HEAP32[$1083>>2]|0;
     $1085 = (($1084) + ($1080)|0);
     $1086 = HEAP8[$1085>>0]|0;
     $1087 = $1086&255;
     $1088 = HEAP32[$842>>2]|0;
     $1089 = $1088 & $1075;
     $1090 = $1091&255;
     $1092 = $1089 >>> $1090;
     $1093 = $1094&255;
     $1095 = (125924 + ($1093<<2)|0);
     $1096 = HEAP32[$1095>>2]|0;
     $1097 = (($1096) + ($1092)|0);
     $1098 = HEAP8[$1097>>0]|0;
     $1099 = $1098&255;
     $1100 = HEAP32[$843>>2]|0;
     $1101 = $1100 & $1075;
     $1102 = $1103&255;
     $1104 = $1101 >>> $1102;
     $1105 = $1106&255;
     $1107 = (125924 + ($1105<<2)|0);
     $1108 = HEAP32[$1107>>2]|0;
     $1109 = (($1108) + ($1104)|0);
     $1110 = HEAP8[$1109>>0]|0;
     $1111 = $1110&255;
     $1112 = HEAP32[$844>>2]|0;
     $1113 = $1112 & $1075;
     $1114 = $1115&255;
     $1116 = $1113 >>> $1114;
     $1117 = $1118&255;
     $1119 = (125924 + ($1117<<2)|0);
     $1120 = HEAP32[$1119>>2]|0;
     $1121 = (($1120) + ($1116)|0);
     $1122 = HEAP8[$1121>>0]|0;
     $1123 = $1122&255;
     $1124 = Math_imul($1087, $845)|0;
     $1125 = (($1124>>>0) / 255)&-1;
     $1126 = Math_imul($1099, $846)|0;
     $1127 = (($1126>>>0) / 255)&-1;
     $1128 = Math_imul($1111, $847)|0;
     $1129 = (($1128>>>0) / 255)&-1;
     $1130 = $1125 >>> $1081;
     $1131 = $1130 << $1078;
     $1132 = $1127 >>> $1093;
     $1133 = $1132 << $1090;
     $1134 = $1133 | $1131;
     $1135 = $1129 >>> $1105;
     $1136 = $1135 << $1102;
     $1137 = $1134 | $1136;
     $1138 = $1123 >>> $1117;
     $1139 = $1138 << $1114;
     $1140 = $1137 | $1139;
     HEAP32[$pixel39$3>>2] = $1140;
     $1141 = ((($pixel39$3)) + 4|0);
     $1146 = $1479;$1149 = $1478;$1158 = $1477;$1161 = $1476;$1170 = $1103;$1173 = $1106;$1182 = $1115;$1185 = $1118;$n40$3 = $n40$2;$pixel39$4 = $1141;
     label = 39;
     continue;
    }
    else if ((label|0) == 39) {
     label = 0;
     $1142 = HEAP32[$pixel39$4>>2]|0;
     $1143 = HEAP32[$839>>2]|0;
     $1144 = $1143 & $1142;
     $1145 = $1146&255;
     $1147 = $1144 >>> $1145;
     $1148 = $1149&255;
     $1150 = (125924 + ($1148<<2)|0);
     $1151 = HEAP32[$1150>>2]|0;
     $1152 = (($1151) + ($1147)|0);
     $1153 = HEAP8[$1152>>0]|0;
     $1154 = $1153&255;
     $1155 = HEAP32[$842>>2]|0;
     $1156 = $1155 & $1142;
     $1157 = $1158&255;
     $1159 = $1156 >>> $1157;
     $1160 = $1161&255;
     $1162 = (125924 + ($1160<<2)|0);
     $1163 = HEAP32[$1162>>2]|0;
     $1164 = (($1163) + ($1159)|0);
     $1165 = HEAP8[$1164>>0]|0;
     $1166 = $1165&255;
     $1167 = HEAP32[$843>>2]|0;
     $1168 = $1167 & $1142;
     $1169 = $1170&255;
     $1171 = $1168 >>> $1169;
     $1172 = $1173&255;
     $1174 = (125924 + ($1172<<2)|0);
     $1175 = HEAP32[$1174>>2]|0;
     $1176 = (($1175) + ($1171)|0);
     $1177 = HEAP8[$1176>>0]|0;
     $1178 = $1177&255;
     $1179 = HEAP32[$844>>2]|0;
     $1180 = $1179 & $1142;
     $1181 = $1182&255;
     $1183 = $1180 >>> $1181;
     $1184 = $1185&255;
     $1186 = (125924 + ($1184<<2)|0);
     $1187 = HEAP32[$1186>>2]|0;
     $1188 = (($1187) + ($1183)|0);
     $1189 = HEAP8[$1188>>0]|0;
     $1190 = $1189&255;
     $1191 = Math_imul($1154, $845)|0;
     $1192 = (($1191>>>0) / 255)&-1;
     $1193 = Math_imul($1166, $846)|0;
     $1194 = (($1193>>>0) / 255)&-1;
     $1195 = Math_imul($1178, $847)|0;
     $1196 = (($1195>>>0) / 255)&-1;
     $1197 = $1192 >>> $1148;
     $1198 = $1197 << $1145;
     $1199 = $1194 >>> $1160;
     $1200 = $1199 << $1157;
     $1201 = $1200 | $1198;
     $1202 = $1196 >>> $1172;
     $1203 = $1202 << $1169;
     $1204 = $1201 | $1203;
     $1205 = $1190 >>> $1184;
     $1206 = $1205 << $1181;
     $1207 = $1204 | $1206;
     HEAP32[$pixel39$4>>2] = $1207;
     $1208 = ((($pixel39$4)) + 4|0);
     $1209 = (($n40$3) + -1)|0;
     $1210 = ($n40$3|0)>(1);
     if ($1210) {
      $1484 = $1161;$1485 = $1158;$1486 = $1149;$1487 = $1146;$949 = $1146;$952 = $1149;$954 = $1151;$960 = $1158;$963 = $1161;$965 = $1163;$971 = $1170;$974 = $1173;$976 = $1175;$982 = $1182;$985 = $1185;$987 = $1187;$n40$0 = $1209;$pixel39$1 = $1208;
      label = 36;
      continue;
     } else {
      break;
     }
    }
   }
   $1211 = (($1208) + ($826<<2)|0);
   $1212 = ($848|0)==(0);
   if ($1212) {
    $$0 = 0;
    label = 56;
    break;
   } else {
    $$in = $848;$pixel39$024 = $1211;
   }
  }
  if ((label|0) == 40) {
   // unreachable;
  }
  else if ((label|0) == 56) {
   return ($$0|0);
  }
  break;
 }
 default: {
  $1213 = ((($rect)) + 8|0);
  $1214 = HEAP32[$1213>>2]|0;
  $1215 = ((($rect)) + 12|0);
  $1216 = HEAP32[$1215>>2]|0;
  $1217 = ((($dst)) + 16|0);
  $1218 = HEAP32[$1217>>2]|0;
  $1219 = (($1218|0) / ($6|0))&-1;
  $1220 = (($1219) - ($1214))|0;
  $1221 = ($1216|0)==(0);
  if ($1221) {
   $$0 = 0;
   return ($$0|0);
  }
  $1222 = ((($dst)) + 20|0);
  $1223 = HEAP32[$1222>>2]|0;
  $1224 = ((($rect)) + 4|0);
  $1225 = HEAP32[$1224>>2]|0;
  $1226 = Math_imul($1225, $1219)|0;
  $1227 = (($1223) + ($1226<<2)|0);
  $1228 = HEAP32[$rect>>2]|0;
  $1229 = (($1227) + ($1228<<2)|0);
  $1230 = (($1214) + 3)|0;
  $1231 = (($1230|0) / 4)&-1;
  $1232 = $1214 & 3;
  $1233 = $r&255;
  $1234 = $g&255;
  $1235 = $b&255;
  $1236 = ((($1)) + 28|0);
  $1237 = ((($1)) + 32|0);
  $$in28 = $1216;$pixel61$015 = $1229;
  L10: while(1) {
   $1238 = (($$in28) + -1)|0;
   switch ($1232|0) {
   case 0:  {
    $1309 = HEAP16[$1236>>1]|0;
    $1310 = HEAP32[$1236>>2]|0;
    $1311 = $1310&65535;
    $1312 = $1310&255;
    $1313 = HEAP16[$1237>>1]|0;
    $1314 = HEAP32[$1237>>2]|0;
    $1315 = $1314&65535;
    $1316 = $1314&255;
    $1317 = ($1311&65535) >>> 8;
    $1318 = $1317&255;
    $1319 = ($1315&65535) >>> 8;
    $1320 = $1319&255;
    $1321 = $1310 >>> 16;
    $1322 = $1321&255;
    $1323 = $1314 >>> 16;
    $1324 = $1323&255;
    $1325 = $1310 >>> 24;
    $1326 = $1325&255;
    $1327 = $1314 >>> 24;
    $1328 = $1327&255;
    $1329 = $1310&255;
    $1330 = $1314&255;
    $1331 = ($1309&65535) >>> 8;
    $1332 = $1331&255;
    $1333 = ($1313&65535) >>> 8;
    $1334 = $1333&255;
    $1336 = $1312;$1339 = $1316;$1342 = $1318;$1345 = $1320;$1349 = $1322;$1352 = $1324;$1356 = $1326;$1359 = $1328;$1496 = $1334;$1497 = $1332;$1498 = $1330;$1499 = $1329;$n62$0 = $1231;$pixel61$1 = $pixel61$015;
    label = 49;
    break;
   }
   case 3:  {
    $1283 = HEAP16[$1236>>1]|0;
    $1284 = HEAP32[$1236>>2]|0;
    $1285 = $1284&65535;
    $1286 = $1284&255;
    $1287 = HEAP16[$1237>>1]|0;
    $1288 = HEAP32[$1237>>2]|0;
    $1289 = $1288&65535;
    $1290 = $1288&255;
    $1291 = ($1285&65535) >>> 8;
    $1292 = $1291&255;
    $1293 = ($1289&65535) >>> 8;
    $1294 = $1293&255;
    $1295 = $1284 >>> 16;
    $1296 = $1295&255;
    $1297 = $1288 >>> 16;
    $1298 = $1297&255;
    $1299 = $1284 >>> 24;
    $1300 = $1299&255;
    $1301 = $1288 >>> 24;
    $1302 = $1301&255;
    $1303 = $1284&255;
    $1304 = $1288&255;
    $1305 = ($1283&65535) >>> 8;
    $1306 = $1305&255;
    $1307 = ($1287&65535) >>> 8;
    $1308 = $1307&255;
    $1364 = $1286;$1367 = $1290;$1370 = $1292;$1373 = $1294;$1377 = $1296;$1380 = $1298;$1384 = $1300;$1387 = $1302;$1492 = $1308;$1493 = $1306;$1494 = $1304;$1495 = $1303;$n62$1 = $1231;$pixel61$2 = $pixel61$015;
    label = 50;
    break;
   }
   case 2:  {
    $1257 = HEAP16[$1236>>1]|0;
    $1258 = HEAP32[$1236>>2]|0;
    $1259 = $1258&65535;
    $1260 = $1258&255;
    $1261 = HEAP16[$1237>>1]|0;
    $1262 = HEAP32[$1237>>2]|0;
    $1263 = $1262&65535;
    $1264 = $1262&255;
    $1265 = ($1259&65535) >>> 8;
    $1266 = $1265&255;
    $1267 = ($1263&65535) >>> 8;
    $1268 = $1267&255;
    $1269 = $1258 >>> 16;
    $1270 = $1269&255;
    $1271 = $1262 >>> 16;
    $1272 = $1271&255;
    $1273 = $1258 >>> 24;
    $1274 = $1273&255;
    $1275 = $1262 >>> 24;
    $1276 = $1275&255;
    $1277 = $1258&255;
    $1278 = $1262&255;
    $1279 = ($1257&65535) >>> 8;
    $1280 = $1279&255;
    $1281 = ($1261&65535) >>> 8;
    $1282 = $1281&255;
    $1392 = $1260;$1395 = $1264;$1398 = $1266;$1401 = $1268;$1405 = $1270;$1408 = $1272;$1412 = $1274;$1415 = $1276;$1488 = $1282;$1489 = $1280;$1490 = $1278;$1491 = $1277;$n62$2 = $1231;$pixel61$3 = $pixel61$015;
    label = 51;
    break;
   }
   case 1:  {
    $1239 = HEAP32[$1236>>2]|0;
    $1240 = $1239&65535;
    $1241 = $1239&255;
    $1242 = HEAP32[$1237>>2]|0;
    $1243 = $1242&65535;
    $1244 = $1242&255;
    $1245 = ($1240&65535) >>> 8;
    $1246 = $1245&255;
    $1247 = ($1243&65535) >>> 8;
    $1248 = $1247&255;
    $1249 = $1239 >>> 16;
    $1250 = $1249&255;
    $1251 = $1242 >>> 16;
    $1252 = $1251&255;
    $1253 = $1239 >>> 24;
    $1254 = $1253&255;
    $1255 = $1242 >>> 24;
    $1256 = $1255&255;
    $1420 = $1241;$1423 = $1244;$1426 = $1246;$1429 = $1248;$1433 = $1250;$1436 = $1252;$1440 = $1254;$1443 = $1256;$n62$3 = $1231;$pixel61$4 = $pixel61$015;
    label = 52;
    break;
   }
   default: {
    label = 53;
    break L10;
   }
   }
   while(1) {
    if ((label|0) == 49) {
     label = 0;
     $1335 = $1336&255;
     $1337 = $1233 >>> $1335;
     $1338 = $1339&255;
     $1340 = $1337 << $1338;
     $1341 = $1342&255;
     $1343 = $1234 >>> $1341;
     $1344 = $1345&255;
     $1346 = $1343 << $1344;
     $1347 = $1346 | $1340;
     $1348 = $1349&255;
     $1350 = $1235 >>> $1348;
     $1351 = $1352&255;
     $1353 = $1350 << $1351;
     $1354 = $1347 | $1353;
     $1355 = $1356&255;
     $1357 = $2 >>> $1355;
     $1358 = $1359&255;
     $1360 = $1357 << $1358;
     $1361 = $1354 | $1360;
     HEAP32[$pixel61$1>>2] = $1361;
     $1362 = ((($pixel61$1)) + 4|0);
     $1364 = $1499;$1367 = $1498;$1370 = $1497;$1373 = $1496;$1377 = $1349;$1380 = $1352;$1384 = $1356;$1387 = $1359;$1492 = $1496;$1493 = $1497;$1494 = $1498;$1495 = $1499;$n62$1 = $n62$0;$pixel61$2 = $1362;
     label = 50;
     continue;
    }
    else if ((label|0) == 50) {
     label = 0;
     $1363 = $1364&255;
     $1365 = $1233 >>> $1363;
     $1366 = $1367&255;
     $1368 = $1365 << $1366;
     $1369 = $1370&255;
     $1371 = $1234 >>> $1369;
     $1372 = $1373&255;
     $1374 = $1371 << $1372;
     $1375 = $1374 | $1368;
     $1376 = $1377&255;
     $1378 = $1235 >>> $1376;
     $1379 = $1380&255;
     $1381 = $1378 << $1379;
     $1382 = $1375 | $1381;
     $1383 = $1384&255;
     $1385 = $2 >>> $1383;
     $1386 = $1387&255;
     $1388 = $1385 << $1386;
     $1389 = $1382 | $1388;
     HEAP32[$pixel61$2>>2] = $1389;
     $1390 = ((($pixel61$2)) + 4|0);
     $1392 = $1495;$1395 = $1494;$1398 = $1493;$1401 = $1492;$1405 = $1377;$1408 = $1380;$1412 = $1384;$1415 = $1387;$1488 = $1492;$1489 = $1493;$1490 = $1494;$1491 = $1495;$n62$2 = $n62$1;$pixel61$3 = $1390;
     label = 51;
     continue;
    }
    else if ((label|0) == 51) {
     label = 0;
     $1391 = $1392&255;
     $1393 = $1233 >>> $1391;
     $1394 = $1395&255;
     $1396 = $1393 << $1394;
     $1397 = $1398&255;
     $1399 = $1234 >>> $1397;
     $1400 = $1401&255;
     $1402 = $1399 << $1400;
     $1403 = $1402 | $1396;
     $1404 = $1405&255;
     $1406 = $1235 >>> $1404;
     $1407 = $1408&255;
     $1409 = $1406 << $1407;
     $1410 = $1403 | $1409;
     $1411 = $1412&255;
     $1413 = $2 >>> $1411;
     $1414 = $1415&255;
     $1416 = $1413 << $1414;
     $1417 = $1410 | $1416;
     HEAP32[$pixel61$3>>2] = $1417;
     $1418 = ((($pixel61$3)) + 4|0);
     $1420 = $1491;$1423 = $1490;$1426 = $1489;$1429 = $1488;$1433 = $1405;$1436 = $1408;$1440 = $1412;$1443 = $1415;$n62$3 = $n62$2;$pixel61$4 = $1418;
     label = 52;
     continue;
    }
    else if ((label|0) == 52) {
     label = 0;
     $1419 = $1420&255;
     $1421 = $1233 >>> $1419;
     $1422 = $1423&255;
     $1424 = $1421 << $1422;
     $1425 = $1426&255;
     $1427 = $1234 >>> $1425;
     $1428 = $1429&255;
     $1430 = $1427 << $1428;
     $1431 = $1430 | $1424;
     $1432 = $1433&255;
     $1434 = $1235 >>> $1432;
     $1435 = $1436&255;
     $1437 = $1434 << $1435;
     $1438 = $1431 | $1437;
     $1439 = $1440&255;
     $1441 = $2 >>> $1439;
     $1442 = $1443&255;
     $1444 = $1441 << $1442;
     $1445 = $1438 | $1444;
     HEAP32[$pixel61$4>>2] = $1445;
     $1446 = ((($pixel61$4)) + 4|0);
     $1447 = (($n62$3) + -1)|0;
     $1448 = ($n62$3|0)>(1);
     if ($1448) {
      $1336 = $1420;$1339 = $1423;$1342 = $1426;$1345 = $1429;$1349 = $1433;$1352 = $1436;$1356 = $1440;$1359 = $1443;$1496 = $1429;$1497 = $1426;$1498 = $1423;$1499 = $1420;$n62$0 = $1447;$pixel61$1 = $1446;
      label = 49;
      continue;
     } else {
      break;
     }
    }
   }
   $1449 = (($1446) + ($1220<<2)|0);
   $1450 = ($1238|0)==(0);
   if ($1450) {
    $$0 = 0;
    label = 56;
    break;
   } else {
    $$in28 = $1238;$pixel61$015 = $1449;
   }
  }
  if ((label|0) == 53) {
   // unreachable;
  }
  else if ((label|0) == 56) {
   return ($$0|0);
  }
 }
 }
 return (0)|0;
}
function _SDL_BlendFillRect_RGB555($dst,$rect,$blendMode,$r,$g,$b,$a) {
 $dst = $dst|0;
 $rect = $rect|0;
 $blendMode = $blendMode|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $$ = 0, $$1 = 0, $$10 = 0, $$11 = 0, $$2 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$6 = 0, $$7 = 0, $$8 = 0, $$9 = 0, $$in = 0, $$in46 = 0, $$in47 = 0, $$lcssa = 0, $$lcssa122 = 0, $$lcssa95 = 0, $$lcssa96 = 0, $$pre = 0;
 var $$pre58 = 0, $$pre60 = 0, $$pre62 = 0, $$pre64 = 0, $$pre66 = 0, $$pre68 = 0, $$pre70 = 0, $$pre72 = 0, $$pre74 = 0, $$pre76 = 0, $$pre78 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0;
 var $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0;
 var $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0;
 var $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0;
 var $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0;
 var $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0;
 var $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0;
 var $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0;
 var $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0;
 var $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0;
 var $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0;
 var $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0;
 var $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0;
 var $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0;
 var $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0;
 var $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0;
 var $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0;
 var $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0;
 var $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0;
 var $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0;
 var $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0;
 var $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0;
 var $n$3 = 0, $n18$0 = 0, $n18$1 = 0, $n18$2 = 0, $n18$3 = 0, $n40$0 = 0, $n40$1 = 0, $n40$2 = 0, $n40$3 = 0, $n62$0$us = 0, $n62$1$us25101 = 0, $n62$2$us38106 = 0, $n62$397 = 0, $pixel$016 = 0, $pixel$1 = 0, $pixel$2 = 0, $pixel$3 = 0, $pixel$4 = 0, $pixel17$018 = 0, $pixel17$1 = 0;
 var $pixel17$2 = 0, $pixel17$3 = 0, $pixel17$4 = 0, $pixel39$020 = 0, $pixel39$1 = 0, $pixel39$2 = 0, $pixel39$3 = 0, $pixel39$4 = 0, $pixel61$015 = 0, $pixel61$015$us = 0, $pixel61$015$us22 = 0, $pixel61$015$us33 = 0, $pixel61$1$us = 0, $pixel61$2$us26102 = 0, $pixel61$3$us39107 = 0, $pixel61$498 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $a&255;
 $1 = $0 ^ 255;
 switch ($blendMode|0) {
 case 1:  {
  $2 = ((($rect)) + 8|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($rect)) + 12|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ((($dst)) + 16|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ((($dst)) + 4|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($9)) + 9|0);
  $11 = HEAP8[$10>>0]|0;
  $12 = $11&255;
  $13 = (($7|0) / ($12|0))&-1;
  $14 = (($13) - ($3))|0;
  $15 = ($5|0)==(0);
  if ($15) {
   return 0;
  }
  $16 = ((($dst)) + 20|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = ((($rect)) + 4|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = Math_imul($19, $13)|0;
  $21 = (($17) + ($20<<1)|0);
  $22 = HEAP32[$rect>>2]|0;
  $23 = (($21) + ($22<<1)|0);
  $24 = (($3) + 3)|0;
  $25 = (($24|0) / 4)&-1;
  $26 = $3 & 3;
  $27 = $r&255;
  $28 = $g&255;
  $29 = $b&255;
  $$in47 = $5;$pixel$016 = $23;
  L41: while(1) {
   $30 = (($$in47) + -1)|0;
   switch ($26|0) {
   case 0:  {
    $$pre72 = HEAP32[(125936)>>2]|0;
    $36 = $$pre72;$n$0 = $25;$pixel$1 = $pixel$016;
    label = 9;
    break;
   }
   case 3:  {
    $$pre74 = HEAP32[(125936)>>2]|0;
    $71 = $$pre74;$n$1 = $25;$pixel$2 = $pixel$016;
    label = 10;
    break;
   }
   case 2:  {
    $$pre76 = HEAP32[(125936)>>2]|0;
    $106 = $$pre76;$n$2 = $25;$pixel$3 = $pixel$016;
    label = 11;
    break;
   }
   case 1:  {
    $$pre78 = HEAP32[(125936)>>2]|0;
    $141 = $$pre78;$n$3 = $25;$pixel$4 = $pixel$016;
    label = 12;
    break;
   }
   default: {
    label = 13;
    break L41;
   }
   }
   while(1) {
    if ((label|0) == 9) {
     label = 0;
     $31 = HEAP16[$pixel$1>>1]|0;
     $32 = $31&65535;
     $33 = $32 >>> 10;
     $34 = $33 & 31;
     $35 = (($36) + ($34)|0);
     $37 = HEAP8[$35>>0]|0;
     $38 = $37&255;
     $39 = $32 >>> 5;
     $40 = $39 & 31;
     $41 = (($36) + ($40)|0);
     $42 = HEAP8[$41>>0]|0;
     $43 = $42&255;
     $44 = $32 & 31;
     $45 = (($36) + ($44)|0);
     $46 = HEAP8[$45>>0]|0;
     $47 = $46&255;
     $48 = Math_imul($38, $1)|0;
     $49 = (($48>>>0) / 255)&-1;
     $50 = (($49) + ($27))|0;
     $51 = Math_imul($43, $1)|0;
     $52 = (($51>>>0) / 255)&-1;
     $53 = (($52) + ($28))|0;
     $54 = Math_imul($47, $1)|0;
     $55 = (($54>>>0) / 255)&-1;
     $56 = (($55) + ($29))|0;
     $57 = $50 >>> 3;
     $58 = $57 << 10;
     $59 = $53 >>> 3;
     $60 = $59 << 5;
     $61 = $60 | $58;
     $62 = $56 >>> 3;
     $63 = $61 | $62;
     $64 = $63&65535;
     HEAP16[$pixel$1>>1] = $64;
     $65 = ((($pixel$1)) + 2|0);
     $71 = $36;$n$1 = $n$0;$pixel$2 = $65;
     label = 10;
     continue;
    }
    else if ((label|0) == 10) {
     label = 0;
     $66 = HEAP16[$pixel$2>>1]|0;
     $67 = $66&65535;
     $68 = $67 >>> 10;
     $69 = $68 & 31;
     $70 = (($71) + ($69)|0);
     $72 = HEAP8[$70>>0]|0;
     $73 = $72&255;
     $74 = $67 >>> 5;
     $75 = $74 & 31;
     $76 = (($71) + ($75)|0);
     $77 = HEAP8[$76>>0]|0;
     $78 = $77&255;
     $79 = $67 & 31;
     $80 = (($71) + ($79)|0);
     $81 = HEAP8[$80>>0]|0;
     $82 = $81&255;
     $83 = Math_imul($73, $1)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = (($84) + ($27))|0;
     $86 = Math_imul($78, $1)|0;
     $87 = (($86>>>0) / 255)&-1;
     $88 = (($87) + ($28))|0;
     $89 = Math_imul($82, $1)|0;
     $90 = (($89>>>0) / 255)&-1;
     $91 = (($90) + ($29))|0;
     $92 = $85 >>> 3;
     $93 = $92 << 10;
     $94 = $88 >>> 3;
     $95 = $94 << 5;
     $96 = $95 | $93;
     $97 = $91 >>> 3;
     $98 = $96 | $97;
     $99 = $98&65535;
     HEAP16[$pixel$2>>1] = $99;
     $100 = ((($pixel$2)) + 2|0);
     $106 = $71;$n$2 = $n$1;$pixel$3 = $100;
     label = 11;
     continue;
    }
    else if ((label|0) == 11) {
     label = 0;
     $101 = HEAP16[$pixel$3>>1]|0;
     $102 = $101&65535;
     $103 = $102 >>> 10;
     $104 = $103 & 31;
     $105 = (($106) + ($104)|0);
     $107 = HEAP8[$105>>0]|0;
     $108 = $107&255;
     $109 = $102 >>> 5;
     $110 = $109 & 31;
     $111 = (($106) + ($110)|0);
     $112 = HEAP8[$111>>0]|0;
     $113 = $112&255;
     $114 = $102 & 31;
     $115 = (($106) + ($114)|0);
     $116 = HEAP8[$115>>0]|0;
     $117 = $116&255;
     $118 = Math_imul($108, $1)|0;
     $119 = (($118>>>0) / 255)&-1;
     $120 = (($119) + ($27))|0;
     $121 = Math_imul($113, $1)|0;
     $122 = (($121>>>0) / 255)&-1;
     $123 = (($122) + ($28))|0;
     $124 = Math_imul($117, $1)|0;
     $125 = (($124>>>0) / 255)&-1;
     $126 = (($125) + ($29))|0;
     $127 = $120 >>> 3;
     $128 = $127 << 10;
     $129 = $123 >>> 3;
     $130 = $129 << 5;
     $131 = $130 | $128;
     $132 = $126 >>> 3;
     $133 = $131 | $132;
     $134 = $133&65535;
     HEAP16[$pixel$3>>1] = $134;
     $135 = ((($pixel$3)) + 2|0);
     $141 = $106;$n$3 = $n$2;$pixel$4 = $135;
     label = 12;
     continue;
    }
    else if ((label|0) == 12) {
     label = 0;
     $136 = HEAP16[$pixel$4>>1]|0;
     $137 = $136&65535;
     $138 = $137 >>> 10;
     $139 = $138 & 31;
     $140 = (($141) + ($139)|0);
     $142 = HEAP8[$140>>0]|0;
     $143 = $142&255;
     $144 = $137 >>> 5;
     $145 = $144 & 31;
     $146 = (($141) + ($145)|0);
     $147 = HEAP8[$146>>0]|0;
     $148 = $147&255;
     $149 = $137 & 31;
     $150 = (($141) + ($149)|0);
     $151 = HEAP8[$150>>0]|0;
     $152 = $151&255;
     $153 = Math_imul($143, $1)|0;
     $154 = (($153>>>0) / 255)&-1;
     $155 = (($154) + ($27))|0;
     $156 = Math_imul($148, $1)|0;
     $157 = (($156>>>0) / 255)&-1;
     $158 = (($157) + ($28))|0;
     $159 = Math_imul($152, $1)|0;
     $160 = (($159>>>0) / 255)&-1;
     $161 = (($160) + ($29))|0;
     $162 = $155 >>> 3;
     $163 = $162 << 10;
     $164 = $158 >>> 3;
     $165 = $164 << 5;
     $166 = $165 | $163;
     $167 = $161 >>> 3;
     $168 = $166 | $167;
     $169 = $168&65535;
     HEAP16[$pixel$4>>1] = $169;
     $170 = ((($pixel$4)) + 2|0);
     $171 = (($n$3) + -1)|0;
     $172 = ($n$3|0)>(1);
     if ($172) {
      $36 = $141;$n$0 = $171;$pixel$1 = $170;
      label = 9;
      continue;
     } else {
      break;
     }
    }
   }
   $173 = (($170) + ($14<<1)|0);
   $174 = ($30|0)==(0);
   if ($174) {
    label = 58;
    break;
   } else {
    $$in47 = $30;$pixel$016 = $173;
   }
  }
  if ((label|0) == 13) {
   // unreachable;
  }
  else if ((label|0) == 58) {
   return 0;
  }
  break;
 }
 case 2:  {
  $175 = ((($rect)) + 8|0);
  $176 = HEAP32[$175>>2]|0;
  $177 = ((($rect)) + 12|0);
  $178 = HEAP32[$177>>2]|0;
  $179 = ((($dst)) + 16|0);
  $180 = HEAP32[$179>>2]|0;
  $181 = ((($dst)) + 4|0);
  $182 = HEAP32[$181>>2]|0;
  $183 = ((($182)) + 9|0);
  $184 = HEAP8[$183>>0]|0;
  $185 = $184&255;
  $186 = (($180|0) / ($185|0))&-1;
  $187 = (($186) - ($176))|0;
  $188 = ($178|0)==(0);
  if ($188) {
   return 0;
  }
  $189 = ((($dst)) + 20|0);
  $190 = HEAP32[$189>>2]|0;
  $191 = ((($rect)) + 4|0);
  $192 = HEAP32[$191>>2]|0;
  $193 = Math_imul($192, $186)|0;
  $194 = (($190) + ($193<<1)|0);
  $195 = HEAP32[$rect>>2]|0;
  $196 = (($194) + ($195<<1)|0);
  $197 = (($176) + 3)|0;
  $198 = (($197|0) / 4)&-1;
  $199 = $176 & 3;
  $200 = $r&255;
  $201 = $g&255;
  $202 = $b&255;
  $$in46 = $178;$pixel17$018 = $196;
  L62: while(1) {
   $203 = (($$in46) + -1)|0;
   switch ($199|0) {
   case 0:  {
    $$pre64 = HEAP32[(125936)>>2]|0;
    $209 = $$pre64;$n18$0 = $198;$pixel17$1 = $pixel17$018;
    label = 22;
    break;
   }
   case 3:  {
    $$pre66 = HEAP32[(125936)>>2]|0;
    $241 = $$pre66;$n18$1 = $198;$pixel17$2 = $pixel17$018;
    label = 23;
    break;
   }
   case 2:  {
    $$pre68 = HEAP32[(125936)>>2]|0;
    $273 = $$pre68;$n18$2 = $198;$pixel17$3 = $pixel17$018;
    label = 24;
    break;
   }
   case 1:  {
    $$pre70 = HEAP32[(125936)>>2]|0;
    $305 = $$pre70;$n18$3 = $198;$pixel17$4 = $pixel17$018;
    label = 25;
    break;
   }
   default: {
    label = 26;
    break L62;
   }
   }
   while(1) {
    if ((label|0) == 22) {
     label = 0;
     $204 = HEAP16[$pixel17$1>>1]|0;
     $205 = $204&65535;
     $206 = $205 >>> 10;
     $207 = $206 & 31;
     $208 = (($209) + ($207)|0);
     $210 = HEAP8[$208>>0]|0;
     $211 = $210&255;
     $212 = $205 >>> 5;
     $213 = $212 & 31;
     $214 = (($209) + ($213)|0);
     $215 = HEAP8[$214>>0]|0;
     $216 = $215&255;
     $217 = $205 & 31;
     $218 = (($209) + ($217)|0);
     $219 = HEAP8[$218>>0]|0;
     $220 = $219&255;
     $221 = (($211) + ($200))|0;
     $222 = ($221>>>0)>(255);
     $$ = $222 ? 255 : $221;
     $223 = (($216) + ($201))|0;
     $224 = ($223>>>0)>(255);
     $$8 = $224 ? 255 : $223;
     $225 = (($220) + ($202))|0;
     $226 = ($225>>>0)>(255);
     $$1 = $226 ? 255 : $225;
     $227 = $$ >>> 3;
     $228 = $227 << 10;
     $229 = $$8 >>> 3;
     $230 = $229 << 5;
     $231 = $230 | $228;
     $232 = $$1 >>> 3;
     $233 = $231 | $232;
     $234 = $233&65535;
     HEAP16[$pixel17$1>>1] = $234;
     $235 = ((($pixel17$1)) + 2|0);
     $241 = $209;$n18$1 = $n18$0;$pixel17$2 = $235;
     label = 23;
     continue;
    }
    else if ((label|0) == 23) {
     label = 0;
     $236 = HEAP16[$pixel17$2>>1]|0;
     $237 = $236&65535;
     $238 = $237 >>> 10;
     $239 = $238 & 31;
     $240 = (($241) + ($239)|0);
     $242 = HEAP8[$240>>0]|0;
     $243 = $242&255;
     $244 = $237 >>> 5;
     $245 = $244 & 31;
     $246 = (($241) + ($245)|0);
     $247 = HEAP8[$246>>0]|0;
     $248 = $247&255;
     $249 = $237 & 31;
     $250 = (($241) + ($249)|0);
     $251 = HEAP8[$250>>0]|0;
     $252 = $251&255;
     $253 = (($243) + ($200))|0;
     $254 = ($253>>>0)>(255);
     $$2 = $254 ? 255 : $253;
     $255 = (($248) + ($201))|0;
     $256 = ($255>>>0)>(255);
     $$9 = $256 ? 255 : $255;
     $257 = (($252) + ($202))|0;
     $258 = ($257>>>0)>(255);
     $$3 = $258 ? 255 : $257;
     $259 = $$2 >>> 3;
     $260 = $259 << 10;
     $261 = $$9 >>> 3;
     $262 = $261 << 5;
     $263 = $262 | $260;
     $264 = $$3 >>> 3;
     $265 = $263 | $264;
     $266 = $265&65535;
     HEAP16[$pixel17$2>>1] = $266;
     $267 = ((($pixel17$2)) + 2|0);
     $273 = $241;$n18$2 = $n18$1;$pixel17$3 = $267;
     label = 24;
     continue;
    }
    else if ((label|0) == 24) {
     label = 0;
     $268 = HEAP16[$pixel17$3>>1]|0;
     $269 = $268&65535;
     $270 = $269 >>> 10;
     $271 = $270 & 31;
     $272 = (($273) + ($271)|0);
     $274 = HEAP8[$272>>0]|0;
     $275 = $274&255;
     $276 = $269 >>> 5;
     $277 = $276 & 31;
     $278 = (($273) + ($277)|0);
     $279 = HEAP8[$278>>0]|0;
     $280 = $279&255;
     $281 = $269 & 31;
     $282 = (($273) + ($281)|0);
     $283 = HEAP8[$282>>0]|0;
     $284 = $283&255;
     $285 = (($275) + ($200))|0;
     $286 = ($285>>>0)>(255);
     $$4 = $286 ? 255 : $285;
     $287 = (($280) + ($201))|0;
     $288 = ($287>>>0)>(255);
     $$10 = $288 ? 255 : $287;
     $289 = (($284) + ($202))|0;
     $290 = ($289>>>0)>(255);
     $$5 = $290 ? 255 : $289;
     $291 = $$4 >>> 3;
     $292 = $291 << 10;
     $293 = $$10 >>> 3;
     $294 = $293 << 5;
     $295 = $294 | $292;
     $296 = $$5 >>> 3;
     $297 = $295 | $296;
     $298 = $297&65535;
     HEAP16[$pixel17$3>>1] = $298;
     $299 = ((($pixel17$3)) + 2|0);
     $305 = $273;$n18$3 = $n18$2;$pixel17$4 = $299;
     label = 25;
     continue;
    }
    else if ((label|0) == 25) {
     label = 0;
     $300 = HEAP16[$pixel17$4>>1]|0;
     $301 = $300&65535;
     $302 = $301 >>> 10;
     $303 = $302 & 31;
     $304 = (($305) + ($303)|0);
     $306 = HEAP8[$304>>0]|0;
     $307 = $306&255;
     $308 = $301 >>> 5;
     $309 = $308 & 31;
     $310 = (($305) + ($309)|0);
     $311 = HEAP8[$310>>0]|0;
     $312 = $311&255;
     $313 = $301 & 31;
     $314 = (($305) + ($313)|0);
     $315 = HEAP8[$314>>0]|0;
     $316 = $315&255;
     $317 = (($307) + ($200))|0;
     $318 = ($317>>>0)>(255);
     $$6 = $318 ? 255 : $317;
     $319 = (($312) + ($201))|0;
     $320 = ($319>>>0)>(255);
     $$11 = $320 ? 255 : $319;
     $321 = (($316) + ($202))|0;
     $322 = ($321>>>0)>(255);
     $$7 = $322 ? 255 : $321;
     $323 = $$6 >>> 3;
     $324 = $323 << 10;
     $325 = $$11 >>> 3;
     $326 = $325 << 5;
     $327 = $326 | $324;
     $328 = $$7 >>> 3;
     $329 = $327 | $328;
     $330 = $329&65535;
     HEAP16[$pixel17$4>>1] = $330;
     $331 = ((($pixel17$4)) + 2|0);
     $332 = (($n18$3) + -1)|0;
     $333 = ($n18$3|0)>(1);
     if ($333) {
      $209 = $305;$n18$0 = $332;$pixel17$1 = $331;
      label = 22;
      continue;
     } else {
      break;
     }
    }
   }
   $334 = (($331) + ($187<<1)|0);
   $335 = ($203|0)==(0);
   if ($335) {
    label = 58;
    break;
   } else {
    $$in46 = $203;$pixel17$018 = $334;
   }
  }
  if ((label|0) == 26) {
   // unreachable;
  }
  else if ((label|0) == 58) {
   return 0;
  }
  break;
 }
 case 4:  {
  $336 = ((($rect)) + 8|0);
  $337 = HEAP32[$336>>2]|0;
  $338 = ((($rect)) + 12|0);
  $339 = HEAP32[$338>>2]|0;
  $340 = ((($dst)) + 16|0);
  $341 = HEAP32[$340>>2]|0;
  $342 = ((($dst)) + 4|0);
  $343 = HEAP32[$342>>2]|0;
  $344 = ((($343)) + 9|0);
  $345 = HEAP8[$344>>0]|0;
  $346 = $345&255;
  $347 = (($341|0) / ($346|0))&-1;
  $348 = (($347) - ($337))|0;
  $349 = ($339|0)==(0);
  if ($349) {
   return 0;
  }
  $350 = ((($dst)) + 20|0);
  $351 = HEAP32[$350>>2]|0;
  $352 = ((($rect)) + 4|0);
  $353 = HEAP32[$352>>2]|0;
  $354 = Math_imul($353, $347)|0;
  $355 = (($351) + ($354<<1)|0);
  $356 = HEAP32[$rect>>2]|0;
  $357 = (($355) + ($356<<1)|0);
  $358 = (($337) + 3)|0;
  $359 = (($358|0) / 4)&-1;
  $360 = $337 & 3;
  $361 = $r&255;
  $362 = $g&255;
  $363 = $b&255;
  $$in = $339;$pixel39$020 = $357;
  L83: while(1) {
   $364 = (($$in) + -1)|0;
   switch ($360|0) {
   case 0:  {
    $$pre = HEAP32[(125936)>>2]|0;
    $370 = $$pre;$n40$0 = $359;$pixel39$1 = $pixel39$020;
    label = 35;
    break;
   }
   case 3:  {
    $$pre58 = HEAP32[(125936)>>2]|0;
    $402 = $$pre58;$n40$1 = $359;$pixel39$2 = $pixel39$020;
    label = 36;
    break;
   }
   case 2:  {
    $$pre60 = HEAP32[(125936)>>2]|0;
    $434 = $$pre60;$n40$2 = $359;$pixel39$3 = $pixel39$020;
    label = 37;
    break;
   }
   case 1:  {
    $$pre62 = HEAP32[(125936)>>2]|0;
    $466 = $$pre62;$n40$3 = $359;$pixel39$4 = $pixel39$020;
    label = 38;
    break;
   }
   default: {
    label = 39;
    break L83;
   }
   }
   while(1) {
    if ((label|0) == 35) {
     label = 0;
     $365 = HEAP16[$pixel39$1>>1]|0;
     $366 = $365&65535;
     $367 = $366 >>> 10;
     $368 = $367 & 31;
     $369 = (($370) + ($368)|0);
     $371 = HEAP8[$369>>0]|0;
     $372 = $371&255;
     $373 = $366 >>> 5;
     $374 = $373 & 31;
     $375 = (($370) + ($374)|0);
     $376 = HEAP8[$375>>0]|0;
     $377 = $376&255;
     $378 = $366 & 31;
     $379 = (($370) + ($378)|0);
     $380 = HEAP8[$379>>0]|0;
     $381 = $380&255;
     $382 = Math_imul($372, $361)|0;
     $383 = (($382>>>0) / 255)&-1;
     $384 = Math_imul($377, $362)|0;
     $385 = (($384>>>0) / 255)&-1;
     $386 = Math_imul($381, $363)|0;
     $387 = (($386>>>0) / 255)&-1;
     $388 = $383 >>> 3;
     $389 = $388 << 10;
     $390 = $385 >>> 3;
     $391 = $390 << 5;
     $392 = $391 | $389;
     $393 = $387 >>> 3;
     $394 = $392 | $393;
     $395 = $394&65535;
     HEAP16[$pixel39$1>>1] = $395;
     $396 = ((($pixel39$1)) + 2|0);
     $402 = $370;$n40$1 = $n40$0;$pixel39$2 = $396;
     label = 36;
     continue;
    }
    else if ((label|0) == 36) {
     label = 0;
     $397 = HEAP16[$pixel39$2>>1]|0;
     $398 = $397&65535;
     $399 = $398 >>> 10;
     $400 = $399 & 31;
     $401 = (($402) + ($400)|0);
     $403 = HEAP8[$401>>0]|0;
     $404 = $403&255;
     $405 = $398 >>> 5;
     $406 = $405 & 31;
     $407 = (($402) + ($406)|0);
     $408 = HEAP8[$407>>0]|0;
     $409 = $408&255;
     $410 = $398 & 31;
     $411 = (($402) + ($410)|0);
     $412 = HEAP8[$411>>0]|0;
     $413 = $412&255;
     $414 = Math_imul($404, $361)|0;
     $415 = (($414>>>0) / 255)&-1;
     $416 = Math_imul($409, $362)|0;
     $417 = (($416>>>0) / 255)&-1;
     $418 = Math_imul($413, $363)|0;
     $419 = (($418>>>0) / 255)&-1;
     $420 = $415 >>> 3;
     $421 = $420 << 10;
     $422 = $417 >>> 3;
     $423 = $422 << 5;
     $424 = $423 | $421;
     $425 = $419 >>> 3;
     $426 = $424 | $425;
     $427 = $426&65535;
     HEAP16[$pixel39$2>>1] = $427;
     $428 = ((($pixel39$2)) + 2|0);
     $434 = $402;$n40$2 = $n40$1;$pixel39$3 = $428;
     label = 37;
     continue;
    }
    else if ((label|0) == 37) {
     label = 0;
     $429 = HEAP16[$pixel39$3>>1]|0;
     $430 = $429&65535;
     $431 = $430 >>> 10;
     $432 = $431 & 31;
     $433 = (($434) + ($432)|0);
     $435 = HEAP8[$433>>0]|0;
     $436 = $435&255;
     $437 = $430 >>> 5;
     $438 = $437 & 31;
     $439 = (($434) + ($438)|0);
     $440 = HEAP8[$439>>0]|0;
     $441 = $440&255;
     $442 = $430 & 31;
     $443 = (($434) + ($442)|0);
     $444 = HEAP8[$443>>0]|0;
     $445 = $444&255;
     $446 = Math_imul($436, $361)|0;
     $447 = (($446>>>0) / 255)&-1;
     $448 = Math_imul($441, $362)|0;
     $449 = (($448>>>0) / 255)&-1;
     $450 = Math_imul($445, $363)|0;
     $451 = (($450>>>0) / 255)&-1;
     $452 = $447 >>> 3;
     $453 = $452 << 10;
     $454 = $449 >>> 3;
     $455 = $454 << 5;
     $456 = $455 | $453;
     $457 = $451 >>> 3;
     $458 = $456 | $457;
     $459 = $458&65535;
     HEAP16[$pixel39$3>>1] = $459;
     $460 = ((($pixel39$3)) + 2|0);
     $466 = $434;$n40$3 = $n40$2;$pixel39$4 = $460;
     label = 38;
     continue;
    }
    else if ((label|0) == 38) {
     label = 0;
     $461 = HEAP16[$pixel39$4>>1]|0;
     $462 = $461&65535;
     $463 = $462 >>> 10;
     $464 = $463 & 31;
     $465 = (($466) + ($464)|0);
     $467 = HEAP8[$465>>0]|0;
     $468 = $467&255;
     $469 = $462 >>> 5;
     $470 = $469 & 31;
     $471 = (($466) + ($470)|0);
     $472 = HEAP8[$471>>0]|0;
     $473 = $472&255;
     $474 = $462 & 31;
     $475 = (($466) + ($474)|0);
     $476 = HEAP8[$475>>0]|0;
     $477 = $476&255;
     $478 = Math_imul($468, $361)|0;
     $479 = (($478>>>0) / 255)&-1;
     $480 = Math_imul($473, $362)|0;
     $481 = (($480>>>0) / 255)&-1;
     $482 = Math_imul($477, $363)|0;
     $483 = (($482>>>0) / 255)&-1;
     $484 = $479 >>> 3;
     $485 = $484 << 10;
     $486 = $481 >>> 3;
     $487 = $486 << 5;
     $488 = $487 | $485;
     $489 = $483 >>> 3;
     $490 = $488 | $489;
     $491 = $490&65535;
     HEAP16[$pixel39$4>>1] = $491;
     $492 = ((($pixel39$4)) + 2|0);
     $493 = (($n40$3) + -1)|0;
     $494 = ($n40$3|0)>(1);
     if ($494) {
      $370 = $466;$n40$0 = $493;$pixel39$1 = $492;
      label = 35;
      continue;
     } else {
      break;
     }
    }
   }
   $495 = (($492) + ($348<<1)|0);
   $496 = ($364|0)==(0);
   if ($496) {
    label = 58;
    break;
   } else {
    $$in = $364;$pixel39$020 = $495;
   }
  }
  if ((label|0) == 39) {
   // unreachable;
  }
  else if ((label|0) == 58) {
   return 0;
  }
  break;
 }
 default: {
  $497 = ((($rect)) + 8|0);
  $498 = HEAP32[$497>>2]|0;
  $499 = ((($rect)) + 12|0);
  $500 = HEAP32[$499>>2]|0;
  $501 = ((($dst)) + 16|0);
  $502 = HEAP32[$501>>2]|0;
  $503 = ((($dst)) + 4|0);
  $504 = HEAP32[$503>>2]|0;
  $505 = ((($504)) + 9|0);
  $506 = HEAP8[$505>>0]|0;
  $507 = $506&255;
  $508 = (($502|0) / ($507|0))&-1;
  $509 = (($508) - ($498))|0;
  $510 = ((($dst)) + 20|0);
  $511 = HEAP32[$510>>2]|0;
  $512 = ((($rect)) + 4|0);
  $513 = HEAP32[$512>>2]|0;
  $514 = Math_imul($513, $508)|0;
  $515 = (($511) + ($514<<1)|0);
  $516 = HEAP32[$rect>>2]|0;
  $517 = (($515) + ($516<<1)|0);
  $518 = (($500) + -1)|0;
  $519 = ($500|0)==(0);
  if ($519) {
   return 0;
  }
  $520 = (($498) + 3)|0;
  $521 = (($520|0) / 4)&-1;
  $522 = $498 & 3;
  $523 = $r&255;
  $524 = $g&255;
  $525 = $b&255;
  $526 = $523 >>> 3;
  $527 = $526 << 10;
  $528 = $524 >>> 3;
  $529 = $528 << 5;
  $530 = $529 | $527;
  $531 = $525 >>> 3;
  $532 = $530 | $531;
  $533 = $532&65535;
  switch ($522|0) {
  case 0:  {
   $545 = $518;$pixel61$015$us = $517;
   while(1) {
    $n62$0$us = $521;$pixel61$1$us = $pixel61$015$us;
    while(1) {
     HEAP16[$pixel61$1$us>>1] = $533;
     $537 = ((($pixel61$1$us)) + 2|0);
     HEAP16[$537>>1] = $533;
     $538 = ((($pixel61$1$us)) + 4|0);
     HEAP16[$538>>1] = $533;
     $539 = ((($pixel61$1$us)) + 6|0);
     HEAP16[$539>>1] = $533;
     $540 = ((($pixel61$1$us)) + 8|0);
     $541 = (($n62$0$us) + -1)|0;
     $542 = ($n62$0$us|0)>(1);
     if ($542) {
      $n62$0$us = $541;$pixel61$1$us = $540;
     } else {
      $$lcssa122 = $540;
      break;
     }
    }
    $543 = (($$lcssa122) + ($509<<1)|0);
    $544 = (($545) + -1)|0;
    $546 = ($545|0)==(0);
    if ($546) {
     break;
    } else {
     $545 = $544;$pixel61$015$us = $543;
    }
   }
   return 0;
   break;
  }
  case 3:  {
   $535 = ($520|0)>(7);
   $559 = $518;$pixel61$015$us22 = $517;
   while(1) {
    HEAP16[$pixel61$015$us22>>1] = $533;
    $547 = ((($pixel61$015$us22)) + 2|0);
    HEAP16[$547>>1] = $533;
    $548 = ((($pixel61$015$us22)) + 4|0);
    HEAP16[$548>>1] = $533;
    $549 = ((($pixel61$015$us22)) + 6|0);
    if ($535) {
     $551 = $549;$n62$1$us25101 = $521;$pixel61$2$us26102 = $pixel61$015$us22;
     while(1) {
      $550 = (($n62$1$us25101) + -1)|0;
      HEAP16[$551>>1] = $533;
      $552 = ((($pixel61$2$us26102)) + 8|0);
      HEAP16[$552>>1] = $533;
      $553 = ((($pixel61$2$us26102)) + 10|0);
      HEAP16[$553>>1] = $533;
      $554 = ((($pixel61$2$us26102)) + 12|0);
      HEAP16[$554>>1] = $533;
      $555 = ((($pixel61$2$us26102)) + 14|0);
      $556 = ($550|0)>(1);
      if ($556) {
       $551 = $555;$n62$1$us25101 = $550;$pixel61$2$us26102 = $552;
      } else {
       $$lcssa95 = $555;
       break;
      }
     }
    } else {
     $$lcssa95 = $549;
    }
    $557 = (($$lcssa95) + ($509<<1)|0);
    $558 = (($559) + -1)|0;
    $560 = ($559|0)==(0);
    if ($560) {
     break;
    } else {
     $559 = $558;$pixel61$015$us22 = $557;
    }
   }
   return 0;
   break;
  }
  case 2:  {
   $534 = ($520|0)>(7);
   $572 = $518;$pixel61$015$us33 = $517;
   while(1) {
    HEAP16[$pixel61$015$us33>>1] = $533;
    $561 = ((($pixel61$015$us33)) + 2|0);
    HEAP16[$561>>1] = $533;
    $562 = ((($pixel61$015$us33)) + 4|0);
    if ($534) {
     $564 = $562;$n62$2$us38106 = $521;$pixel61$3$us39107 = $pixel61$015$us33;
     while(1) {
      $563 = (($n62$2$us38106) + -1)|0;
      HEAP16[$564>>1] = $533;
      $565 = ((($pixel61$3$us39107)) + 6|0);
      HEAP16[$565>>1] = $533;
      $566 = ((($pixel61$3$us39107)) + 8|0);
      HEAP16[$566>>1] = $533;
      $567 = ((($pixel61$3$us39107)) + 10|0);
      HEAP16[$567>>1] = $533;
      $568 = ((($pixel61$3$us39107)) + 12|0);
      $569 = ($563|0)>(1);
      if ($569) {
       $564 = $568;$n62$2$us38106 = $563;$pixel61$3$us39107 = $566;
      } else {
       $$lcssa96 = $568;
       break;
      }
     }
    } else {
     $$lcssa96 = $562;
    }
    $570 = (($$lcssa96) + ($509<<1)|0);
    $571 = (($572) + -1)|0;
    $573 = ($572|0)==(0);
    if ($573) {
     break;
    } else {
     $572 = $571;$pixel61$015$us33 = $570;
    }
   }
   return 0;
   break;
  }
  default: {
   $536 = ($520|0)>(7);
   $584 = $518;$pixel61$015 = $517;
   while(1) {
    HEAP16[$pixel61$015>>1] = $533;
    $574 = ((($pixel61$015)) + 2|0);
    if ($536) {
     $576 = $574;$n62$397 = $521;$pixel61$498 = $pixel61$015;
     while(1) {
      $575 = (($n62$397) + -1)|0;
      HEAP16[$576>>1] = $533;
      $577 = ((($pixel61$498)) + 4|0);
      HEAP16[$577>>1] = $533;
      $578 = ((($pixel61$498)) + 6|0);
      HEAP16[$578>>1] = $533;
      $579 = ((($pixel61$498)) + 8|0);
      HEAP16[$579>>1] = $533;
      $580 = ((($pixel61$498)) + 10|0);
      $581 = ($575|0)>(1);
      if ($581) {
       $576 = $580;$n62$397 = $575;$pixel61$498 = $579;
      } else {
       $$lcssa = $580;
       break;
      }
     }
    } else {
     $$lcssa = $574;
    }
    $582 = (($$lcssa) + ($509<<1)|0);
    $583 = (($584) + -1)|0;
    $585 = ($584|0)==(0);
    if ($585) {
     break;
    } else {
     $584 = $583;$pixel61$015 = $582;
    }
   }
   return 0;
  }
  }
 }
 }
 return (0)|0;
}
function _SDL_BlendFillRect_RGB565($dst,$rect,$blendMode,$r,$g,$b,$a) {
 $dst = $dst|0;
 $rect = $rect|0;
 $blendMode = $blendMode|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 $a = $a|0;
 var $$ = 0, $$1 = 0, $$10 = 0, $$11 = 0, $$2 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$6 = 0, $$7 = 0, $$8 = 0, $$9 = 0, $$in = 0, $$in46 = 0, $$in47 = 0, $$lcssa = 0, $$lcssa107 = 0, $$lcssa108 = 0, $$lcssa134 = 0, $$pre = 0;
 var $$pre57 = 0, $$pre59 = 0, $$pre60 = 0, $$pre62 = 0, $$pre63 = 0, $$pre65 = 0, $$pre66 = 0, $$pre68 = 0, $$pre69 = 0, $$pre71 = 0, $$pre72 = 0, $$pre74 = 0, $$pre75 = 0, $$pre77 = 0, $$pre78 = 0, $$pre80 = 0, $$pre81 = 0, $$pre83 = 0, $$pre84 = 0, $$pre86 = 0;
 var $$pre87 = 0, $$pre89 = 0, $$pre90 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0;
 var $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0;
 var $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0;
 var $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0;
 var $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0;
 var $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0;
 var $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0;
 var $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0;
 var $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0;
 var $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0;
 var $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0;
 var $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0;
 var $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0;
 var $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0;
 var $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0;
 var $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0;
 var $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0;
 var $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0;
 var $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0;
 var $582 = 0, $583 = 0, $584 = 0, $585 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $n$0 = 0, $n$1 = 0, $n$2 = 0, $n$3 = 0, $n18$0 = 0, $n18$1 = 0, $n18$2 = 0, $n18$3 = 0, $n40$0 = 0, $n40$1 = 0, $n40$2 = 0;
 var $n40$3 = 0, $n62$0$us = 0, $n62$1$us25113 = 0, $n62$2$us38118 = 0, $n62$3109 = 0, $pixel$016 = 0, $pixel$1 = 0, $pixel$2 = 0, $pixel$3 = 0, $pixel$4 = 0, $pixel17$018 = 0, $pixel17$1 = 0, $pixel17$2 = 0, $pixel17$3 = 0, $pixel17$4 = 0, $pixel39$020 = 0, $pixel39$1 = 0, $pixel39$2 = 0, $pixel39$3 = 0, $pixel39$4 = 0;
 var $pixel61$015 = 0, $pixel61$015$us = 0, $pixel61$015$us22 = 0, $pixel61$015$us33 = 0, $pixel61$1$us = 0, $pixel61$2$us26114 = 0, $pixel61$3$us39119 = 0, $pixel61$4110 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $a&255;
 $1 = $0 ^ 255;
 switch ($blendMode|0) {
 case 1:  {
  $2 = ((($rect)) + 8|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($rect)) + 12|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ((($dst)) + 16|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ((($dst)) + 4|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($9)) + 9|0);
  $11 = HEAP8[$10>>0]|0;
  $12 = $11&255;
  $13 = (($7|0) / ($12|0))&-1;
  $14 = (($13) - ($3))|0;
  $15 = ($5|0)==(0);
  if ($15) {
   return 0;
  }
  $16 = ((($dst)) + 20|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = ((($rect)) + 4|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = Math_imul($19, $13)|0;
  $21 = (($17) + ($20<<1)|0);
  $22 = HEAP32[$rect>>2]|0;
  $23 = (($21) + ($22<<1)|0);
  $24 = (($3) + 3)|0;
  $25 = (($24|0) / 4)&-1;
  $26 = $3 & 3;
  $27 = $r&255;
  $28 = $g&255;
  $29 = $b&255;
  $$in47 = $5;$pixel$016 = $23;
  L41: while(1) {
   $30 = (($$in47) + -1)|0;
   switch ($26|0) {
   case 0:  {
    $$pre80 = HEAP32[(125936)>>2]|0;
    $$pre81 = HEAP32[(125932)>>2]|0;
    $35 = $$pre80;$41 = $$pre81;$n$0 = $25;$pixel$1 = $pixel$016;
    label = 9;
    break;
   }
   case 3:  {
    $$pre83 = HEAP32[(125936)>>2]|0;
    $$pre84 = HEAP32[(125932)>>2]|0;
    $70 = $$pre83;$76 = $$pre84;$n$1 = $25;$pixel$2 = $pixel$016;
    label = 10;
    break;
   }
   case 2:  {
    $$pre86 = HEAP32[(125936)>>2]|0;
    $$pre87 = HEAP32[(125932)>>2]|0;
    $105 = $$pre86;$111 = $$pre87;$n$2 = $25;$pixel$3 = $pixel$016;
    label = 11;
    break;
   }
   case 1:  {
    $$pre89 = HEAP32[(125936)>>2]|0;
    $$pre90 = HEAP32[(125932)>>2]|0;
    $140 = $$pre89;$146 = $$pre90;$n$3 = $25;$pixel$4 = $pixel$016;
    label = 12;
    break;
   }
   default: {
    label = 13;
    break L41;
   }
   }
   while(1) {
    if ((label|0) == 9) {
     label = 0;
     $31 = HEAP16[$pixel$1>>1]|0;
     $32 = $31&65535;
     $33 = $32 >>> 11;
     $34 = (($35) + ($33)|0);
     $36 = HEAP8[$34>>0]|0;
     $37 = $36&255;
     $38 = $32 >>> 5;
     $39 = $38 & 63;
     $40 = (($41) + ($39)|0);
     $42 = HEAP8[$40>>0]|0;
     $43 = $42&255;
     $44 = $32 & 31;
     $45 = (($35) + ($44)|0);
     $46 = HEAP8[$45>>0]|0;
     $47 = $46&255;
     $48 = Math_imul($37, $1)|0;
     $49 = (($48>>>0) / 255)&-1;
     $50 = (($49) + ($27))|0;
     $51 = Math_imul($43, $1)|0;
     $52 = (($51>>>0) / 255)&-1;
     $53 = (($52) + ($28))|0;
     $54 = Math_imul($47, $1)|0;
     $55 = (($54>>>0) / 255)&-1;
     $56 = (($55) + ($29))|0;
     $57 = $50 >>> 3;
     $58 = $57 << 11;
     $59 = $53 >>> 2;
     $60 = $59 << 5;
     $61 = $60 | $58;
     $62 = $56 >>> 3;
     $63 = $61 | $62;
     $64 = $63&65535;
     HEAP16[$pixel$1>>1] = $64;
     $65 = ((($pixel$1)) + 2|0);
     $70 = $35;$76 = $41;$n$1 = $n$0;$pixel$2 = $65;
     label = 10;
     continue;
    }
    else if ((label|0) == 10) {
     label = 0;
     $66 = HEAP16[$pixel$2>>1]|0;
     $67 = $66&65535;
     $68 = $67 >>> 11;
     $69 = (($70) + ($68)|0);
     $71 = HEAP8[$69>>0]|0;
     $72 = $71&255;
     $73 = $67 >>> 5;
     $74 = $73 & 63;
     $75 = (($76) + ($74)|0);
     $77 = HEAP8[$75>>0]|0;
     $78 = $77&255;
     $79 = $67 & 31;
     $80 = (($70) + ($79)|0);
     $81 = HEAP8[$80>>0]|0;
     $82 = $81&255;
     $83 = Math_imul($72, $1)|0;
     $84 = (($83>>>0) / 255)&-1;
     $85 = (($84) + ($27))|0;
     $86 = Math_imul($78, $1)|0;
     $87 = (($86>>>0) / 255)&-1;
     $88 = (($87) + ($28))|0;
     $89 = Math_imul($82, $1)|0;
     $90 = (($89>>>0) / 255)&-1;
     $91 = (($90) + ($29))|0;
     $92 = $85 >>> 3;
     $93 = $92 << 11;
     $94 = $88 >>> 2;
     $95 = $94 << 5;
     $96 = $95 | $93;
     $97 = $91 >>> 3;
     $98 = $96 | $97;
     $99 = $98&65535;
     HEAP16[$pixel$2>>1] = $99;
     $100 = ((($pixel$2)) + 2|0);
     $105 = $70;$111 = $76;$n$2 = $n$1;$pixel$3 = $100;
     label = 11;
     continue;
    }
    else if ((label|0) == 11) {
     label = 0;
     $101 = HEAP16[$pixel$3>>1]|0;
     $102 = $101&65535;
     $103 = $102 >>> 11;
     $104 = (($105) + ($103)|0);
     $106 = HEAP8[$104>>0]|0;
     $107 = $106&255;
     $108 = $102 >>> 5;
     $109 = $108 & 63;
     $110 = (($111) + ($109)|0);
     $112 = HEAP8[$110>>0]|0;
     $113 = $112&255;
     $114 = $102 & 31;
     $115 = (($105) + ($114)|0);
     $116 = HEAP8[$115>>0]|0;
     $117 = $116&255;
     $118 = Math_imul($107, $1)|0;
     $119 = (($118>>>0) / 255)&-1;
     $120 = (($119) + ($27))|0;
     $121 = Math_imul($113, $1)|0;
     $122 = (($121>>>0) / 255)&-1;
     $123 = (($122) + ($28))|0;
     $124 = Math_imul($117, $1)|0;
     $125 = (($124>>>0) / 255)&-1;
     $126 = (($125) + ($29))|0;
     $127 = $120 >>> 3;
     $128 = $127 << 11;
     $129 = $123 >>> 2;
     $130 = $129 << 5;
     $131 = $130 | $128;
     $132 = $126 >>> 3;
     $133 = $131 | $132;
     $134 = $133&65535;
     HEAP16[$pixel$3>>1] = $134;
     $135 = ((($pixel$3)) + 2|0);
     $140 = $105;$146 = $111;$n$3 = $n$2;$pixel$4 = $135;
     label = 12;
     continue;
    }
    else if ((label|0) == 12) {
     label = 0;
     $136 = HEAP16[$pixel$4>>1]|0;
     $137 = $136&65535;
     $138 = $137 >>> 11;
     $139 = (($140) + ($138)|0);
     $141 = HEAP8[$139>>0]|0;
     $142 = $141&255;
     $143 = $137 >>> 5;
     $144 = $143 & 63;
     $145 = (($146) + ($144)|0);
     $147 = HEAP8[$145>>0]|0;
     $148 = $147&255;
     $149 = $137 & 31;
     $150 = (($140) + ($149)|0);
     $151 = HEAP8[$150>>0]|0;
     $152 = $151&255;
     $153 = Math_imul($142, $1)|0;
     $154 = (($153>>>0) / 255)&-1;
     $155 = (($154) + ($27))|0;
     $156 = Math_imul($148, $1)|0;
     $157 = (($156>>>0) / 255)&-1;
     $158 = (($157) + ($28))|0;
     $159 = Math_imul($152, $1)|0;
     $160 = (($159>>>0) / 255)&-1;
     $161 = (($160) + ($29))|0;
     $162 = $155 >>> 3;
     $163 = $162 << 11;
     $164 = $158 >>> 2;
     $165 = $164 << 5;
     $166 = $165 | $163;
     $167 = $161 >>> 3;
     $168 = $166 | $167;
     $169 = $168&65535;
     HEAP16[$pixel$4>>1] = $169;
     $170 = ((($pixel$4)) + 2|0);
     $171 = (($n$3) + -1)|0;
     $172 = ($n$3|0)>(1);
     if ($172) {
      $35 = $140;$41 = $146;$n$0 = $171;$pixel$1 = $170;
      label = 9;
      continue;
     } else {
      break;
     }
    }
   }
   $173 = (($170) + ($14<<1)|0);
   $174 = ($30|0)==(0);
   if ($174) {
    label = 58;
    break;
   } else {
    $$in47 = $30;$pixel$016 = $173;
   }
  }
  if ((label|0) == 13) {
   // unreachable;
  }
  else if ((label|0) == 58) {
   return 0;
  }
  break;
 }
 case 2:  {
  $175 = ((($rect)) + 8|0);
  $176 = HEAP32[$175>>2]|0;
  $177 = ((($rect)) + 12|0);
  $178 = HEAP32[$177>>2]|0;
  $179 = ((($dst)) + 16|0);
  $180 = HEAP32[$179>>2]|0;
  $181 = ((($dst)) + 4|0);
  $182 = HEAP32[$181>>2]|0;
  $183 = ((($182)) + 9|0);
  $184 = HEAP8[$183>>0]|0;
  $185 = $184&255;
  $186 = (($180|0) / ($185|0))&-1;
  $187 = (($186) - ($176))|0;
  $188 = ($178|0)==(0);
  if ($188) {
   return 0;
  }
  $189 = ((($dst)) + 20|0);
  $190 = HEAP32[$189>>2]|0;
  $191 = ((($rect)) + 4|0);
  $192 = HEAP32[$191>>2]|0;
  $193 = Math_imul($192, $186)|0;
  $194 = (($190) + ($193<<1)|0);
  $195 = HEAP32[$rect>>2]|0;
  $196 = (($194) + ($195<<1)|0);
  $197 = (($176) + 3)|0;
  $198 = (($197|0) / 4)&-1;
  $199 = $176 & 3;
  $200 = $r&255;
  $201 = $g&255;
  $202 = $b&255;
  $$in46 = $178;$pixel17$018 = $196;
  L62: while(1) {
   $203 = (($$in46) + -1)|0;
   switch ($199|0) {
   case 0:  {
    $$pre68 = HEAP32[(125936)>>2]|0;
    $$pre69 = HEAP32[(125932)>>2]|0;
    $208 = $$pre68;$214 = $$pre69;$n18$0 = $198;$pixel17$1 = $pixel17$018;
    label = 22;
    break;
   }
   case 3:  {
    $$pre71 = HEAP32[(125936)>>2]|0;
    $$pre72 = HEAP32[(125932)>>2]|0;
    $240 = $$pre71;$246 = $$pre72;$n18$1 = $198;$pixel17$2 = $pixel17$018;
    label = 23;
    break;
   }
   case 2:  {
    $$pre74 = HEAP32[(125936)>>2]|0;
    $$pre75 = HEAP32[(125932)>>2]|0;
    $272 = $$pre74;$278 = $$pre75;$n18$2 = $198;$pixel17$3 = $pixel17$018;
    label = 24;
    break;
   }
   case 1:  {
    $$pre77 = HEAP32[(125936)>>2]|0;
    $$pre78 = HEAP32[(125932)>>2]|0;
    $304 = $$pre77;$310 = $$pre78;$n18$3 = $198;$pixel17$4 = $pixel17$018;
    label = 25;
    break;
   }
   default: {
    label = 26;
    break L62;
   }
   }
   while(1) {
    if ((label|0) == 22) {
     label = 0;
     $204 = HEAP16[$pixel17$1>>1]|0;
     $205 = $204&65535;
     $206 = $205 >>> 11;
     $207 = (($208) + ($206)|0);
     $209 = HEAP8[$207>>0]|0;
     $210 = $209&255;
     $211 = $205 >>> 5;
     $212 = $211 & 63;
     $213 = (($214) + ($212)|0);
     $215 = HEAP8[$213>>0]|0;
     $216 = $215&255;
     $217 = $205 & 31;
     $218 = (($208) + ($217)|0);
     $219 = HEAP8[$218>>0]|0;
     $220 = $219&255;
     $221 = (($210) + ($200))|0;
     $222 = ($221>>>0)>(255);
     $$ = $222 ? 255 : $221;
     $223 = (($216) + ($201))|0;
     $224 = ($223>>>0)>(255);
     $$8 = $224 ? 255 : $223;
     $225 = (($220) + ($202))|0;
     $226 = ($225>>>0)>(255);
     $$1 = $226 ? 255 : $225;
     $227 = $$ >>> 3;
     $228 = $227 << 11;
     $229 = $$8 >>> 2;
     $230 = $229 << 5;
     $231 = $230 | $228;
     $232 = $$1 >>> 3;
     $233 = $231 | $232;
     $234 = $233&65535;
     HEAP16[$pixel17$1>>1] = $234;
     $235 = ((($pixel17$1)) + 2|0);
     $240 = $208;$246 = $214;$n18$1 = $n18$0;$pixel17$2 = $235;
     label = 23;
     continue;
    }
    else if ((label|0) == 23) {
     label = 0;
     $236 = HEAP16[$pixel17$2>>1]|0;
     $237 = $236&65535;
     $238 = $237 >>> 11;
     $239 = (($240) + ($238)|0);
     $241 = HEAP8[$239>>0]|0;
     $242 = $241&255;
     $243 = $237 >>> 5;
     $244 = $243 & 63;
     $245 = (($246) + ($244)|0);
     $247 = HEAP8[$245>>0]|0;
     $248 = $247&255;
     $249 = $237 & 31;
     $250 = (($240) + ($249)|0);
     $251 = HEAP8[$250>>0]|0;
     $252 = $251&255;
     $253 = (($242) + ($200))|0;
     $254 = ($253>>>0)>(255);
     $$2 = $254 ? 255 : $253;
     $255 = (($248) + ($201))|0;
     $256 = ($255>>>0)>(255);
     $$9 = $256 ? 255 : $255;
     $257 = (($252) + ($202))|0;
     $258 = ($257>>>0)>(255);
     $$3 = $258 ? 255 : $257;
     $259 = $$2 >>> 3;
     $260 = $259 << 11;
     $261 = $$9 >>> 2;
     $262 = $261 << 5;
     $263 = $262 | $260;
     $264 = $$3 >>> 3;
     $265 = $263 | $264;
     $266 = $265&65535;
     HEAP16[$pixel17$2>>1] = $266;
     $267 = ((($pixel17$2)) + 2|0);
     $272 = $240;$278 = $246;$n18$2 = $n18$1;$pixel17$3 = $267;
     label = 24;
     continue;
    }
    else if ((label|0) == 24) {
     label = 0;
     $268 = HEAP16[$pixel17$3>>1]|0;
     $269 = $268&65535;
     $270 = $269 >>> 11;
     $271 = (($272) + ($270)|0);
     $273 = HEAP8[$271>>0]|0;
     $274 = $273&255;
     $275 = $269 >>> 5;
     $276 = $275 & 63;
     $277 = (($278) + ($276)|0);
     $279 = HEAP8[$277>>0]|0;
     $280 = $279&255;
     $281 = $269 & 31;
     $282 = (($272) + ($281)|0);
     $283 = HEAP8[$282>>0]|0;
     $284 = $283&255;
     $285 = (($274) + ($200))|0;
     $286 = ($285>>>0)>(255);
     $$4 = $286 ? 255 : $285;
     $287 = (($280) + ($201))|0;
     $288 = ($287>>>0)>(255);
     $$10 = $288 ? 255 : $287;
     $289 = (($284) + ($202))|0;
     $290 = ($289>>>0)>(255);
     $$5 = $290 ? 255 : $289;
     $291 = $$4 >>> 3;
     $292 = $291 << 11;
     $293 = $$10 >>> 2;
     $294 = $293 << 5;
     $295 = $294 | $292;
     $296 = $$5 >>> 3;
     $297 = $295 | $296;
     $298 = $297&65535;
     HEAP16[$pixel17$3>>1] = $298;
     $299 = ((($pixel17$3)) + 2|0);
     $304 = $272;$310 = $278;$n18$3 = $n18$2;$pixel17$4 = $299;
     label = 25;
     continue;
    }
    else if ((label|0) == 25) {
     label = 0;
     $300 = HEAP16[$pixel17$4>>1]|0;
     $301 = $300&65535;
     $302 = $301 >>> 11;
     $303 = (($304) + ($302)|0);
     $305 = HEAP8[$303>>0]|0;
     $306 = $305&255;
     $307 = $301 >>> 5;
     $308 = $307 & 63;
     $309 = (($310) + ($308)|0);
     $311 = HEAP8[$309>>0]|0;
     $312 = $311&255;
     $313 = $301 & 31;
     $314 = (($304) + ($313)|0);
     $315 = HEAP8[$314>>0]|0;
     $316 = $315&255;
     $317 = (($306) + ($200))|0;
     $318 = ($317>>>0)>(255);
     $$6 = $318 ? 255 : $317;
     $319 = (($312) + ($201))|0;
     $320 = ($319>>>0)>(255);
     $$11 = $320 ? 255 : $319;
     $321 = (($316) + ($202))|0;
     $322 = ($321>>>0)>(255);
     $$7 = $322 ? 255 : $321;
     $323 = $$6 >>> 3;
     $324 = $323 << 11;
     $325 = $$11 >>> 2;
     $326 = $325 << 5;
     $327 = $326 | $324;
     $328 = $$7 >>> 3;
     $329 = $327 | $328;
     $330 = $329&65535;
     HEAP16[$pixel17$4>>1] = $330;
     $331 = ((($pixel17$4)) + 2|0);
     $332 = (($n18$3) + -1)|0;
     $333 = ($n18$3|0)>(1);
     if ($333) {
      $208 = $304;$214 = $310;$n18$0 = $332;$pixel17$1 = $331;
      label = 22;
      continue;
     } else {
      break;
     }
    }
   }
   $334 = (($331) + ($187<<1)|0);
   $335 = ($203|0)==(0);
   if ($335) {
    label = 58;
    break;
   } else {
    $$in46 = $203;$pixel17$018 = $334;
   }
  }
  if ((label|0) == 26) {
   // unreachable;
  }
  else if ((label|0) == 58) {
   return 0;
  }
  break;
 }
 case 4:  {
  $336 = ((($rect)) + 8|0);
  $337 = HEAP32[$336>>2]|0;
  $338 = ((($rect)) + 12|0);
  $339 = HEAP32[$338>>2]|0;
  $340 = ((($dst)) + 16|0);
  $341 = HEAP32[$340>>2]|0;
  $342 = ((($dst)) + 4|0);
  $343 = HEAP32[$342>>2]|0;
  $344 = ((($343)) + 9|0);
  $345 = HEAP8[$344>>0]|0;
  $346 = $345&255;
  $347 = (($341|0) / ($346|0))&-1;
  $348 = (($347) - ($337))|0;
  $349 = ($339|0)==(0);
  if ($349) {
   return 0;
  }
  $350 = ((($dst)) + 20|0);
  $351 = HEAP32[$350>>2]|0;
  $352 = ((($rect)) + 4|0);
  $353 = HEAP32[$352>>2]|0;
  $354 = Math_imul($353, $347)|0;
  $355 = (($351) + ($354<<1)|0);
  $356 = HEAP32[$rect>>2]|0;
  $357 = (($355) + ($356<<1)|0);
  $358 = (($337) + 3)|0;
  $359 = (($358|0) / 4)&-1;
  $360 = $337 & 3;
  $361 = $r&255;
  $362 = $g&255;
  $363 = $b&255;
  $$in = $339;$pixel39$020 = $357;
  L83: while(1) {
   $364 = (($$in) + -1)|0;
   switch ($360|0) {
   case 0:  {
    $$pre = HEAP32[(125936)>>2]|0;
    $$pre57 = HEAP32[(125932)>>2]|0;
    $369 = $$pre;$375 = $$pre57;$n40$0 = $359;$pixel39$1 = $pixel39$020;
    label = 35;
    break;
   }
   case 3:  {
    $$pre59 = HEAP32[(125936)>>2]|0;
    $$pre60 = HEAP32[(125932)>>2]|0;
    $401 = $$pre59;$407 = $$pre60;$n40$1 = $359;$pixel39$2 = $pixel39$020;
    label = 36;
    break;
   }
   case 2:  {
    $$pre62 = HEAP32[(125936)>>2]|0;
    $$pre63 = HEAP32[(125932)>>2]|0;
    $433 = $$pre62;$439 = $$pre63;$n40$2 = $359;$pixel39$3 = $pixel39$020;
    label = 37;
    break;
   }
   case 1:  {
    $$pre65 = HEAP32[(125936)>>2]|0;
    $$pre66 = HEAP32[(125932)>>2]|0;
    $465 = $$pre65;$471 = $$pre66;$n40$3 = $359;$pixel39$4 = $pixel39$020;
    label = 38;
    break;
   }
   default: {
    label = 39;
    break L83;
   }
   }
   while(1) {
    if ((label|0) == 35) {
     label = 0;
     $365 = HEAP16[$pixel39$1>>1]|0;
     $366 = $365&65535;
     $367 = $366 >>> 11;
     $368 = (($369) + ($367)|0);
     $370 = HEAP8[$368>>0]|0;
     $371 = $370&255;
     $372 = $366 >>> 5;
     $373 = $372 & 63;
     $374 = (($375) + ($373)|0);
     $376 = HEAP8[$374>>0]|0;
     $377 = $376&255;
     $378 = $366 & 31;
     $379 = (($369) + ($378)|0);
     $380 = HEAP8[$379>>0]|0;
     $381 = $380&255;
     $382 = Math_imul($371, $361)|0;
     $383 = (($382>>>0) / 255)&-1;
     $384 = Math_imul($377, $362)|0;
     $385 = (($384>>>0) / 255)&-1;
     $386 = Math_imul($381, $363)|0;
     $387 = (($386>>>0) / 255)&-1;
     $388 = $383 >>> 3;
     $389 = $388 << 11;
     $390 = $385 >>> 2;
     $391 = $390 << 5;
     $392 = $391 | $389;
     $393 = $387 >>> 3;
     $394 = $392 | $393;
     $395 = $394&65535;
     HEAP16[$pixel39$1>>1] = $395;
     $396 = ((($pixel39$1)) + 2|0);
     $401 = $369;$407 = $375;$n40$1 = $n40$0;$pixel39$2 = $396;
     label = 36;
     continue;
    }
    else if ((label|0) == 36) {
     label = 0;
     $397 = HEAP16[$pixel39$2>>1]|0;
     $398 = $397&65535;
     $399 = $398 >>> 11;
     $400 = (($401) + ($399)|0);
     $402 = HEAP8[$400>>0]|0;
     $403 = $402&255;
     $404 = $398 >>> 5;
     $405 = $404 & 63;
     $406 = (($407) + ($405)|0);
     $408 = HEAP8[$406>>0]|0;
     $409 = $408&255;
     $410 = $398 & 31;
     $411 = (($401) + ($410)|0);
     $412 = HEAP8[$411>>0]|0;
     $413 = $412&255;
     $414 = Math_imul($403, $361)|0;
     $415 = (($414>>>0) / 255)&-1;
     $416 = Math_imul($409, $362)|0;
     $417 = (($416>>>0) / 255)&-1;
     $418 = Math_imul($413, $363)|0;
     $419 = (($418>>>0) / 255)&-1;
     $420 = $415 >>> 3;
     $421 = $420 << 11;
     $422 = $417 >>> 2;
     $423 = $422 << 5;
     $424 = $423 | $421;
     $425 = $419 >>> 3;
     $426 = $424 | $425;
     $427 = $426&65535;
     HEAP16[$pixel39$2>>1] = $427;
     $428 = ((($pixel39$2)) + 2|0);
     $433 = $401;$439 = $407;$n40$2 = $n40$1;$pixel39$3 = $428;
     label = 37;
     continue;
    }
    else if ((label|0) == 37) {
     label = 0;
     $429 = HEAP16[$pixel39$3>>1]|0;
     $430 = $429&65535;
     $431 = $430 >>> 11;
     $432 = (($433) + ($431)|0);
     $434 = HEAP8[$432>>0]|0;
     $435 = $434&255;
     $436 = $430 >>> 5;
     $437 = $436 & 63;
     $438 = (($439) + ($437)|0);
     $440 = HEAP8[$438>>0]|0;
     $441 = $440&255;
     $442 = $430 & 31;
     $443 = (($433) + ($442)|0);
     $444 = HEAP8[$443>>0]|0;
     $445 = $444&255;
     $446 = Math_imul($435, $361)|0;
     $447 = (($446>>>0) / 255)&-1;
     $448 = Math_imul($441, $362)|0;
     $449 = (($448>>>0) / 255)&-1;
     $450 = Math_imul($445, $363)|0;
     $451 = (($450>>>0) / 255)&-1;
     $452 = $447 >>> 3;
     $453 = $452 << 11;
     $454 = $449 >>> 2;
     $455 = $454 << 5;
     $456 = $455 | $453;
     $457 = $451 >>> 3;
     $458 = $456 | $457;
     $459 = $458&65535;
     HEAP16[$pixel39$3>>1] = $459;
     $460 = ((($pixel39$3)) + 2|0);
     $465 = $433;$471 = $439;$n40$3 = $n40$2;$pixel39$4 = $460;
     label = 38;
     continue;
    }
    else if ((label|0) == 38) {
     label = 0;
     $461 = HEAP16[$pixel39$4>>1]|0;
     $462 = $461&65535;
     $463 = $462 >>> 11;
     $464 = (($465) + ($463)|0);
     $466 = HEAP8[$464>>0]|0;
     $467 = $466&255;
     $468 = $462 >>> 5;
     $469 = $468 & 63;
     $470 = (($471) + ($469)|0);
     $472 = HEAP8[$470>>0]|0;
     $473 = $472&255;
     $474 = $462 & 31;
     $475 = (($465) + ($474)|0);
     $476 = HEAP8[$475>>0]|0;
     $477 = $476&255;
     $478 = Math_imul($467, $361)|0;
     $479 = (($478>>>0) / 255)&-1;
     $480 = Math_imul($473, $362)|0;
     $481 = (($480>>>0) / 255)&-1;
     $482 = Math_imul($477, $363)|0;
     $483 = (($482>>>0) / 255)&-1;
     $484 = $479 >>> 3;
     $485 = $484 << 11;
     $486 = $481 >>> 2;
     $487 = $486 << 5;
     $488 = $487 | $485;
     $489 = $483 >>> 3;
     $490 = $488 | $489;
     $491 = $490&65535;
     HEAP16[$pixel39$4>>1] = $491;
     $492 = ((($pixel39$4)) + 2|0);
     $493 = (($n40$3) + -1)|0;
     $494 = ($n40$3|0)>(1);
     if ($494) {
      $369 = $465;$375 = $471;$n40$0 = $493;$pixel39$1 = $492;
      label = 35;
      continue;
     } else {
      break;
     }
    }
   }
   $495 = (($492) + ($348<<1)|0);
   $496 = ($364|0)==(0);
   if ($496) {
    label = 58;
    break;
   } else {
    $$in = $364;$pixel39$020 = $495;
   }
  }
  if ((label|0) == 39) {
   // unreachable;
  }
  else if ((label|0) == 58) {
   return 0;
  }
  break;
 }
 default: {
  $497 = ((($rect)) + 8|0);
  $498 = HEAP32[$497>>2]|0;
  $499 = ((($rect)) + 12|0);
  $500 = HEAP32[$499>>2]|0;
  $501 = ((($dst)) + 16|0);
  $502 = HEAP32[$501>>2]|0;
  $503 = ((($dst)) + 4|0);
  $504 = HEAP32[$503>>2]|0;
  $505 = ((($504)) + 9|0);
  $506 = HEAP8[$505>>0]|0;
  $507 = $506&255;
  $508 = (($502|0) / ($507|0))&-1;
  $509 = (($508) - ($498))|0;
  $510 = ((($dst)) + 20|0);
  $511 = HEAP32[$510>>2]|0;
  $512 = ((($rect)) + 4|0);
  $513 = HEAP32[$512>>2]|0;
  $514 = Math_imul($513, $508)|0;
  $515 = (($511) + ($514<<1)|0);
  $516 = HEAP32[$rect>>2]|0;
  $517 = (($515) + ($516<<1)|0);
  $518 = (($500) + -1)|0;
  $519 = ($500|0)==(0);
  if ($519) {
   return 0;
  }
  $520 = (($498) + 3)|0;
  $521 = (($520|0) / 4)&-1;
  $522 = $498 & 3;
  $523 = $r&255;
  $524 = $g&255;
  $525 = $b&255;
  $526 = $523 >>> 3;
  $527 = $526 << 11;
  $528 = $524 >>> 2;
  $529 = $528 << 5;
  $530 = $529 | $527;
  $531 = $525 >>> 3;
  $532 = $530 | $531;
  $533 = $532&65535;
  switch ($522|0) {
  case 0:  {
   $545 = $518;$pixel61$015$us = $517;
   while(1) {
    $n62$0$us = $521;$pixel61$1$us = $pixel61$015$us;
    while(1) {
     HEAP16[$pixel61$1$us>>1] = $533;
     $537 = ((($pixel61$1$us)) + 2|0);
     HEAP16[$537>>1] = $533;
     $538 = ((($pixel61$1$us)) + 4|0);
     HEAP16[$538>>1] = $533;
     $539 = ((($pixel61$1$us)) + 6|0);
     HEAP16[$539>>1] = $533;
     $540 = ((($pixel61$1$us)) + 8|0);
     $541 = (($n62$0$us) + -1)|0;
     $542 = ($n62$0$us|0)>(1);
     if ($542) {
      $n62$0$us = $541;$pixel61$1$us = $540;
     } else {
      $$lcssa134 = $540;
      break;
     }
    }
    $543 = (($$lcssa134) + ($509<<1)|0);
    $544 = (($545) + -1)|0;
    $546 = ($545|0)==(0);
    if ($546) {
     break;
    } else {
     $545 = $544;$pixel61$015$us = $543;
    }
   }
   return 0;
   break;
  }
  case 3:  {
   $535 = ($520|0)>(7);
   $559 = $518;$pixel61$015$us22 = $517;
   while(1) {
    HEAP16[$pixel61$015$us22>>1] = $533;
    $547 = ((($pixel61$015$us22)) + 2|0);
    HEAP16[$547>>1] = $533;
    $548 = ((($pixel61$015$us22)) + 4|0);
    HEAP16[$548>>1] = $533;
    $549 = ((($pixel61$015$us22)) + 6|0);
    if ($535) {
     $551 = $549;$n62$1$us25113 = $521;$pixel61$2$us26114 = $pixel61$015$us22;
     while(1) {
      $550 = (($n62$1$us25113) + -1)|0;
      HEAP16[$551>>1] = $533;
      $552 = ((($pixel61$2$us26114)) + 8|0);
      HEAP16[$552>>1] = $533;
      $553 = ((($pixel61$2$us26114)) + 10|0);
      HEAP16[$553>>1] = $533;
      $554 = ((($pixel61$2$us26114)) + 12|0);
      HEAP16[$554>>1] = $533;
      $555 = ((($pixel61$2$us26114)) + 14|0);
      $556 = ($550|0)>(1);
      if ($556) {
       $551 = $555;$n62$1$us25113 = $550;$pixel61$2$us26114 = $552;
      } else {
       $$lcssa107 = $555;
       break;
      }
     }
    } else {
     $$lcssa107 = $549;
    }
    $557 = (($$lcssa107) + ($509<<1)|0);
    $558 = (($559) + -1)|0;
    $560 = ($559|0)==(0);
    if ($560) {
     break;
    } else {
     $559 = $558;$pixel61$015$us22 = $557;
    }
   }
   return 0;
   break;
  }
  case 2:  {
   $534 = ($520|0)>(7);
   $572 = $518;$pixel61$015$us33 = $517;
   while(1) {
    HEAP16[$pixel61$015$us33>>1] = $533;
    $561 = ((($pixel61$015$us33)) + 2|0);
    HEAP16[$561>>1] = $533;
    $562 = ((($pixel61$015$us33)) + 4|0);
    if ($534) {
     $564 = $562;$n62$2$us38118 = $521;$pixel61$3$us39119 = $pixel61$015$us33;
     while(1) {
      $563 = (($n62$2$us38118) + -1)|0;
      HEAP16[$564>>1] = $533;
      $565 = ((($pixel61$3$us39119)) + 6|0);
      HEAP16[$565>>1] = $533;
      $566 = ((($pixel61$3$us39119)) + 8|0);
      HEAP16[$566>>1] = $533;
      $567 = ((($pixel61$3$us39119)) + 10|0);
      HEAP16[$567>>1] = $533;
      $568 = ((($pixel61$3$us39119)) + 12|0);
      $569 = ($563|0)>(1);
      if ($569) {
       $564 = $568;$n62$2$us38118 = $563;$pixel61$3$us39119 = $566;
      } else {
       $$lcssa108 = $568;
       break;
      }
     }
    } else {
     $$lcssa108 = $562;
    }
    $570 = (($$lcssa108) + ($509<<1)|0);
    $571 = (($572) + -1)|0;
    $573 = ($572|0)==(0);
    if ($573) {
     break;
    } else {
     $572 = $571;$pixel61$015$us33 = $570;
    }
   }
   return 0;
   break;
  }
  default: {
   $536 = ($520|0)>(7);
   $584 = $518;$pixel61$015 = $517;
   while(1) {
    HEAP16[$pixel61$015>>1] = $533;
    $574 = ((($pixel61$015)) + 2|0);
    if ($536) {
     $576 = $574;$n62$3109 = $521;$pixel61$4110 = $pixel61$015;
     while(1) {
      $575 = (($n62$3109) + -1)|0;
      HEAP16[$576>>1] = $533;
      $577 = ((($pixel61$4110)) + 4|0);
      HEAP16[$577>>1] = $533;
      $578 = ((($pixel61$4110)) + 6|0);
      HEAP16[$578>>1] = $533;
      $579 = ((($pixel61$4110)) + 8|0);
      HEAP16[$579>>1] = $533;
      $580 = ((($pixel61$4110)) + 10|0);
      $581 = ($575|0)>(1);
      if ($581) {
       $576 = $580;$n62$3109 = $575;$pixel61$4110 = $579;
      } else {
       $$lcssa = $580;
       break;
      }
     }
    } else {
     $$lcssa = $574;
    }
    $582 = (($$lcssa) + ($509<<1)|0);
    $583 = (($584) + -1)|0;
    $585 = ($584|0)==(0);
    if ($585) {
     break;
    } else {
     $584 = $583;$pixel61$015 = $582;
    }
   }
   return 0;
  }
  }
 }
 }
 return (0)|0;
}
function _SW_RenderCopy($renderer,$texture,$srcrect,$dstrect) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 $srcrect = $srcrect|0;
 $dstrect = $dstrect|0;
 var $$0 = 0, $$pr = 0, $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0, $37 = 0.0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0;
 var $41 = 0, $42 = 0, $43 = 0.0, $44 = 0, $45 = 0, $46 = 0.0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0, $53 = 0, $54 = 0.0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $final_rect = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $final_rect = sp;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $6 = ((($1)) + 4|0);
  $7 = HEAP32[$6>>2]|0;
  HEAP32[$1>>2] = $7;
  $8 = ($7|0)==(0);
  $9 = $7;
  if ($8) {
   $10 = ((($renderer)) + 188|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = (_SDL_GetWindowSurface($11)|0);
   $13 = ($12|0)==(0|0);
   do {
    if (!($13)) {
     HEAP32[$6>>2] = $12;
     HEAP32[$1>>2] = $12;
     $14 = HEAP32[$0>>2]|0;
     $15 = HEAP32[$14>>2]|0;
     $16 = ($15|0)==(0|0);
     if (!($16)) {
      $17 = ((($renderer)) + 212|0);
      (_SDL_SetClipRect($15,$17)|0);
      $$pre = HEAP32[$0>>2]|0;
      $$pre1 = HEAP32[$$pre>>2]|0;
      $18 = ($$pre1|0)==(0|0);
      if (!($18)) {
       $19 = ((($renderer)) + 276|0);
       $20 = HEAP32[$19>>2]|0;
       $21 = ($20|0)==(0);
       if ($21) {
        (_SDL_SetClipRect($$pre1,0)|0);
        break;
       } else {
        $22 = ((($renderer)) + 244|0);
        (_SDL_SetClipRect($$pre1,$22)|0);
        break;
       }
      }
     }
    }
   } while(0);
   $$pr = HEAP32[$1>>2]|0;
   $26 = $$pr;
  } else {
   $26 = $9;
  }
  $23 = ((($texture)) + 68|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = ($26|0)==(0|0);
  if ($25) {
   $$0 = -1;$68 = $final_rect;
   STACKTOP = sp;return ($$0|0);
  } else {
   $63 = $24;$64 = $26;$67 = $final_rect;
  }
 } else {
  $4 = ((($texture)) + 68|0);
  $5 = HEAP32[$4>>2]|0;
  $63 = $5;$64 = $2;$67 = $final_rect;
 }
 $27 = ((($renderer)) + 212|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ((($renderer)) + 216|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = $28 | $30;
 $32 = ($31|0)==(0);
 if ($32) {
  $43 = +HEAPF32[$dstrect>>2];
  $44 = (~~(($43)));
  HEAP32[$final_rect>>2] = $44;
  $45 = ((($dstrect)) + 4|0);
  $46 = +HEAPF32[$45>>2];
  $47 = (~~(($46)));
  $48 = ((($final_rect)) + 4|0);
  HEAP32[$48>>2] = $47;
 } else {
  $33 = (+($28|0));
  $34 = +HEAPF32[$dstrect>>2];
  $35 = $33 + $34;
  $36 = (~~(($35)));
  HEAP32[$final_rect>>2] = $36;
  $37 = (+($30|0));
  $38 = ((($dstrect)) + 4|0);
  $39 = +HEAPF32[$38>>2];
  $40 = $37 + $39;
  $41 = (~~(($40)));
  $42 = ((($final_rect)) + 4|0);
  HEAP32[$42>>2] = $41;
 }
 $49 = ((($dstrect)) + 8|0);
 $50 = +HEAPF32[$49>>2];
 $51 = (~~(($50)));
 $52 = ((($final_rect)) + 8|0);
 HEAP32[$52>>2] = $51;
 $53 = ((($dstrect)) + 12|0);
 $54 = +HEAPF32[$53>>2];
 $55 = (~~(($54)));
 $56 = ((($final_rect)) + 12|0);
 HEAP32[$56>>2] = $55;
 $57 = ((($srcrect)) + 8|0);
 $58 = HEAP32[$57>>2]|0;
 $59 = ($58|0)==($51|0);
 if ($59) {
  $60 = ((($srcrect)) + 12|0);
  $61 = HEAP32[$60>>2]|0;
  $62 = ($61|0)==($55|0);
  if ($62) {
   $65 = (_SDL_UpperBlit($63,$srcrect,$64,$final_rect)|0);
   $$0 = $65;$68 = $67;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_SDL_SetSurfaceRLE($64,0)|0);
 $66 = (_SDL_UpperBlitScaled($63,$srcrect,$64,$final_rect)|0);
 $$0 = $66;$68 = $67;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_UpperBlit($src,$srcrect,$dst,$dstrect) {
 $src = $src|0;
 $srcrect = $srcrect|0;
 $dst = $dst|0;
 $dstrect = $dstrect|0;
 var $$0 = 0, $$0$i = 0, $$04 = 0, $$h$0 = 0, $$h$3 = 0, $$pre = 0, $$pre$i = 0, $$pre8 = 0, $$w$0 = 0, $$w$3 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fulldst = 0, $h$0 = 0, $h$2 = 0, $h$3 = 0, $or$cond = 0;
 var $or$cond3 = 0, $sr = 0, $srcx$0 = 0, $srcx$1 = 0, $srcx$2 = 0, $srcy$0 = 0, $srcy$1 = 0, $srcy$2 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $w$0 = 0, $w$2 = 0, $w$3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $fulldst = sp + 32|0;
 $sr = sp + 16|0;
 $0 = ($src|0)!=(0|0);
 $1 = ($dst|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  $2 = (_SDL_SetError(604004,$vararg_buffer)|0);
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ((($src)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0);
 if ($5) {
  $6 = ((($dst)) + 28|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)==(0);
  if ($8) {
   $10 = ($dstrect|0)==(0|0);
   if ($10) {
    $11 = ((($fulldst)) + 4|0);
    HEAP32[$11>>2] = 0;
    HEAP32[$fulldst>>2] = 0;
    $12 = ((($dst)) + 8|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($fulldst)) + 8|0);
    HEAP32[$14>>2] = $13;
    $15 = ((($dst)) + 12|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($fulldst)) + 12|0);
    HEAP32[$17>>2] = $16;
    $$04 = $fulldst;
   } else {
    $$04 = $dstrect;
   }
   $18 = ($srcrect|0)==(0|0);
   if ($18) {
    $43 = ((($src)) + 8|0);
    $44 = HEAP32[$43>>2]|0;
    $45 = ((($src)) + 12|0);
    $46 = HEAP32[$45>>2]|0;
    $h$2 = $46;$srcx$1 = 0;$srcy$1 = 0;$w$2 = $44;
   } else {
    $19 = HEAP32[$srcrect>>2]|0;
    $20 = ((($srcrect)) + 8|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = ($19|0)<(0);
    if ($22) {
     $23 = (($21) + ($19))|0;
     $24 = HEAP32[$$04>>2]|0;
     $25 = (($24) - ($19))|0;
     HEAP32[$$04>>2] = $25;
     $srcx$0 = 0;$w$0 = $23;
    } else {
     $srcx$0 = $19;$w$0 = $21;
    }
    $26 = ((($src)) + 8|0);
    $27 = HEAP32[$26>>2]|0;
    $28 = (($27) - ($srcx$0))|0;
    $29 = ($28|0)<($w$0|0);
    $$w$0 = $29 ? $28 : $w$0;
    $30 = ((($srcrect)) + 4|0);
    $31 = HEAP32[$30>>2]|0;
    $32 = ((($srcrect)) + 12|0);
    $33 = HEAP32[$32>>2]|0;
    $34 = ($31|0)<(0);
    if ($34) {
     $35 = (($33) + ($31))|0;
     $36 = ((($$04)) + 4|0);
     $37 = HEAP32[$36>>2]|0;
     $38 = (($37) - ($31))|0;
     HEAP32[$36>>2] = $38;
     $h$0 = $35;$srcy$0 = 0;
    } else {
     $h$0 = $33;$srcy$0 = $31;
    }
    $39 = ((($src)) + 12|0);
    $40 = HEAP32[$39>>2]|0;
    $41 = (($40) - ($srcy$0))|0;
    $42 = ($41|0)<($h$0|0);
    $$h$0 = $42 ? $41 : $h$0;
    $h$2 = $$h$0;$srcx$1 = $srcx$0;$srcy$1 = $srcy$0;$w$2 = $$w$0;
   }
   $47 = ((($dst)) + 36|0);
   $48 = HEAP32[$47>>2]|0;
   $49 = HEAP32[$$04>>2]|0;
   $50 = (($48) - ($49))|0;
   $51 = ($50|0)>(0);
   if ($51) {
    $52 = (($w$2) - ($50))|0;
    HEAP32[$$04>>2] = $48;
    $53 = (($50) + ($srcx$1))|0;
    $$pre = HEAP32[$47>>2]|0;
    $55 = $48;$57 = $$pre;$srcx$2 = $53;$w$3 = $52;
   } else {
    $55 = $49;$57 = $48;$srcx$2 = $srcx$1;$w$3 = $w$2;
   }
   $54 = (($55) + ($w$3))|0;
   $56 = (($54) - ($57))|0;
   $58 = ((($dst)) + 44|0);
   $59 = HEAP32[$58>>2]|0;
   $60 = (($56) - ($59))|0;
   $61 = ($60|0)>(0);
   $62 = $61 ? $60 : 0;
   $$w$3 = (($w$3) - ($62))|0;
   $63 = ((($dst)) + 40|0);
   $64 = HEAP32[$63>>2]|0;
   $65 = ((($$04)) + 4|0);
   $66 = HEAP32[$65>>2]|0;
   $67 = (($64) - ($66))|0;
   $68 = ($67|0)>(0);
   if ($68) {
    $69 = (($h$2) - ($67))|0;
    HEAP32[$65>>2] = $64;
    $70 = (($67) + ($srcy$1))|0;
    $$pre8 = HEAP32[$63>>2]|0;
    $72 = $64;$74 = $$pre8;$h$3 = $69;$srcy$2 = $70;
   } else {
    $72 = $66;$74 = $64;$h$3 = $h$2;$srcy$2 = $srcy$1;
   }
   $71 = (($72) + ($h$3))|0;
   $73 = (($71) - ($74))|0;
   $75 = ((($dst)) + 48|0);
   $76 = HEAP32[$75>>2]|0;
   $77 = (($73) - ($76))|0;
   $78 = ($77|0)>(0);
   $79 = $78 ? $77 : 0;
   $$h$3 = (($h$3) - ($79))|0;
   $80 = ((($src)) + 52|0);
   $81 = HEAP32[$80>>2]|0;
   $82 = ((($81)) + 68|0);
   $83 = HEAP32[$82>>2]|0;
   $84 = $83 & 512;
   $85 = ($84|0)==(0);
   if (!($85)) {
    $86 = $83 & -513;
    HEAP32[$82>>2] = $86;
    _SDL_InvalidateMap($81);
   }
   $87 = ($$w$3|0)>(0);
   $88 = ($$h$3|0)>(0);
   $or$cond3 = $87 & $88;
   if (!($or$cond3)) {
    $123 = ((($$04)) + 12|0);
    HEAP32[$123>>2] = 0;
    $124 = ((($$04)) + 8|0);
    HEAP32[$124>>2] = 0;
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   HEAP32[$sr>>2] = $srcx$2;
   $89 = ((($sr)) + 4|0);
   HEAP32[$89>>2] = $srcy$2;
   $90 = ((($$04)) + 8|0);
   HEAP32[$90>>2] = $$w$3;
   $91 = ((($sr)) + 8|0);
   HEAP32[$91>>2] = $$w$3;
   $92 = ((($$04)) + 12|0);
   HEAP32[$92>>2] = $$h$3;
   $93 = ((($sr)) + 12|0);
   HEAP32[$93>>2] = $$h$3;
   $94 = HEAP32[$80>>2]|0;
   $95 = HEAP32[$94>>2]|0;
   $96 = ($95|0)==($dst|0);
   do {
    if ($96) {
     $97 = ((($dst)) + 4|0);
     $98 = HEAP32[$97>>2]|0;
     $99 = ((($98)) + 4|0);
     $100 = HEAP32[$99>>2]|0;
     $101 = ($100|0)==(0|0);
     if (!($101)) {
      $102 = ((($94)) + 80|0);
      $103 = HEAP32[$102>>2]|0;
      $104 = ((($100)) + 8|0);
      $105 = HEAP32[$104>>2]|0;
      $106 = ($103|0)==($105|0);
      if (!($106)) {
       label = 27;
       break;
      }
     }
     $107 = ((($src)) + 4|0);
     $108 = HEAP32[$107>>2]|0;
     $109 = ((($108)) + 4|0);
     $110 = HEAP32[$109>>2]|0;
     $111 = ($110|0)==(0|0);
     if ($111) {
      $120 = $94;
      label = 29;
     } else {
      $112 = ((($94)) + 84|0);
      $113 = HEAP32[$112>>2]|0;
      $114 = ((($110)) + 8|0);
      $115 = HEAP32[$114>>2]|0;
      $116 = ($113|0)==($115|0);
      if ($116) {
       $120 = $94;
       label = 29;
      } else {
       label = 27;
      }
     }
    } else {
     label = 27;
    }
   } while(0);
   if ((label|0) == 27) {
    $117 = (_SDL_MapSurface($src,$dst)|0);
    $118 = ($117|0)<(0);
    if ($118) {
     $$0$i = -1;
    } else {
     $$pre$i = HEAP32[$80>>2]|0;
     $120 = $$pre$i;
     label = 29;
    }
   }
   if ((label|0) == 29) {
    $119 = ((($120)) + 8|0);
    $121 = HEAP32[$119>>2]|0;
    $122 = (FUNCTION_TABLE_iiiii[$121 & 8191]($src,$sr,$dst,$$04)|0);
    $$0$i = $122;
   }
   $$0 = $$0$i;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $9 = (_SDL_SetError(604041,$vararg_buffer1)|0);
 $$0 = $9;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_UpperBlitScaled($src,$srcrect,$dst,$dstrect) {
 $src = $src|0;
 $srcrect = $srcrect|0;
 $dst = $dst|0;
 $dstrect = $dstrect|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0, $102 = 0.0, $103 = 0.0, $104 = 0.0, $105 = 0.0, $106 = 0, $107 = 0.0, $108 = 0, $109 = 0.0, $11 = 0, $110 = 0.0, $111 = 0, $112 = 0, $113 = 0.0, $114 = 0;
 var $115 = 0.0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0.0, $123 = 0.0, $124 = 0.0, $125 = 0.0, $126 = 0.0, $127 = 0, $128 = 0.0, $129 = 0.0, $13 = 0, $130 = 0, $131 = 0, $132 = 0.0;
 var $133 = 0.0, $134 = 0.0, $135 = 0, $136 = 0, $137 = 0.0, $138 = 0.0, $139 = 0.0, $14 = 0, $140 = 0, $141 = 0, $142 = 0.0, $143 = 0.0, $144 = 0, $145 = 0.0, $146 = 0.0, $147 = 0, $148 = 0, $149 = 0.0, $15 = 0, $150 = 0.0;
 var $151 = 0.0, $152 = 0, $153 = 0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0.0, $33 = 0, $34 = 0.0, $35 = 0, $36 = 0;
 var $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0, $43 = 0.0, $44 = 0, $45 = 0.0, $46 = 0, $47 = 0.0, $48 = 0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0;
 var $55 = 0, $56 = 0.0, $57 = 0.0, $58 = 0, $59 = 0, $6 = 0, $60 = 0.0, $61 = 0, $62 = 0.0, $63 = 0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0.0, $72 = 0.0;
 var $73 = 0, $74 = 0, $75 = 0.0, $76 = 0, $77 = 0.0, $78 = 0, $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0.0, $85 = 0, $86 = 0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $95 = 0, $96 = 0.0, $97 = 0.0, $98 = 0, $99 = 0, $dst_h$0 = 0, $dst_h$0$in = 0, $dst_w$0 = 0, $dst_w$0$in = 0, $dst_x0$0 = 0.0, $dst_x0$1 = 0.0, $dst_x0$2 = 0.0, $dst_x0$3 = 0.0, $dst_x1$0 = 0.0, $dst_x1$1 = 0.0, $dst_x1$2 = 0.0;
 var $dst_x1$3 = 0.0, $dst_y0$0 = 0.0, $dst_y0$1 = 0.0, $dst_y0$2 = 0.0, $dst_y0$3 = 0.0, $dst_y1$0 = 0.0, $dst_y1$1 = 0.0, $dst_y1$2 = 0.0, $final_dst = 0, $final_src = 0, $or$cond = 0, $or$cond2 = 0, $or$cond4 = 0, $or$cond6 = 0, $or$cond8 = 0, $src_h$0 = 0, $src_h$0$in = 0, $src_w$0 = 0, $src_w$0$in = 0, $src_x0$1 = 0.0;
 var $src_x0$2 = 0.0, $src_x1$0 = 0.0, $src_x1$1 = 0.0, $src_x1$2 = 0.0, $src_y0$1 = 0.0, $src_y0$2 = 0.0, $src_y1$0 = 0.0, $src_y1$1 = 0.0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $final_src = sp + 32|0;
 $final_dst = sp + 16|0;
 $0 = ($src|0)!=(0|0);
 $1 = ($dst|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  $2 = (_SDL_SetError(604081,$vararg_buffer)|0);
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ((($src)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0);
 if ($5) {
  $6 = ((($dst)) + 28|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)==(0);
  if ($8) {
   $10 = ($srcrect|0)==(0|0);
   $11 = ((($src)) + 8|0);
   $12 = ((($src)) + 12|0);
   $13 = ((($srcrect)) + 8|0);
   $14 = ((($srcrect)) + 12|0);
   $src_w$0$in = $10 ? $11 : $13;
   $src_h$0$in = $10 ? $12 : $14;
   $src_h$0 = HEAP32[$src_h$0$in>>2]|0;
   $src_w$0 = HEAP32[$src_w$0$in>>2]|0;
   $15 = ($dstrect|0)==(0|0);
   $16 = ((($dst)) + 8|0);
   $17 = ((($dst)) + 12|0);
   $18 = ((($dstrect)) + 8|0);
   $19 = ((($dstrect)) + 12|0);
   $dst_w$0$in = $15 ? $16 : $18;
   $dst_h$0$in = $15 ? $17 : $19;
   $dst_h$0 = HEAP32[$dst_h$0$in>>2]|0;
   $dst_w$0 = HEAP32[$dst_w$0$in>>2]|0;
   $20 = ($dst_w$0|0)==($src_w$0|0);
   $21 = ($dst_h$0|0)==($src_h$0|0);
   $or$cond2 = $21 & $20;
   if ($or$cond2) {
    $22 = (_SDL_UpperBlit($src,$srcrect,$dst,$dstrect)|0);
    $$0 = $22;
    STACKTOP = sp;return ($$0|0);
   }
   $23 = (+($dst_w$0|0));
   $24 = (+($src_w$0|0));
   $25 = $23 / $24;
   $26 = (+($dst_h$0|0));
   $27 = (+($src_h$0|0));
   $28 = $26 / $27;
   if ($15) {
    $29 = (($dst_w$0) + -1)|0;
    $30 = (+($29|0));
    $31 = (($dst_h$0) + -1)|0;
    $32 = (+($31|0));
    $dst_x0$0 = 0.0;$dst_x1$0 = $30;$dst_y0$0 = 0.0;$dst_y1$0 = $32;
   } else {
    $33 = HEAP32[$dstrect>>2]|0;
    $34 = (+($33|0));
    $35 = ((($dstrect)) + 4|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = (+($36|0));
    $38 = $23 + $34;
    $39 = $38 + -1.0;
    $40 = $26 + $37;
    $41 = $40 + -1.0;
    $dst_x0$0 = $34;$dst_x1$0 = $39;$dst_y0$0 = $37;$dst_y1$0 = $41;
   }
   if ($10) {
    $42 = (($src_w$0) + -1)|0;
    $43 = (+($42|0));
    $44 = (($src_h$0) + -1)|0;
    $45 = (+($44|0));
    $dst_x0$2 = $dst_x0$0;$dst_x1$2 = $dst_x1$0;$dst_y0$2 = $dst_y0$0;$dst_y1$1 = $dst_y1$0;$src_x0$1 = 0.0;$src_x1$1 = $43;$src_y0$1 = 0.0;$src_y1$0 = $45;
   } else {
    $46 = HEAP32[$srcrect>>2]|0;
    $47 = (+($46|0));
    $48 = ((($srcrect)) + 4|0);
    $49 = HEAP32[$48>>2]|0;
    $50 = (+($49|0));
    $51 = $24 + $47;
    $52 = $51 + -1.0;
    $53 = $27 + $50;
    $54 = $53 + -1.0;
    $55 = ($46|0)<(0);
    $56 = $25 * $47;
    $57 = $dst_x0$0 - $56;
    $dst_x0$1 = $55 ? $57 : $dst_x0$0;
    $58 = ($46|0)>(0);
    $59 = $58 ? $46 : 0;
    $60 = (+($59|0));
    $61 = HEAP32[$11>>2]|0;
    $62 = (+($61|0));
    $63 = !($52 >= $62);
    if ($63) {
     $dst_x1$1 = $dst_x1$0;$src_x1$0 = $52;
    } else {
     $64 = $52 - $62;
     $65 = $64 + 1.0;
     $66 = $25 * $65;
     $67 = $dst_x1$0 - $66;
     $68 = (($61) + -1)|0;
     $69 = (+($68|0));
     $dst_x1$1 = $67;$src_x1$0 = $69;
    }
    $70 = ($49|0)<(0);
    $71 = $28 * $50;
    $72 = $dst_y0$0 - $71;
    $73 = ($49|0)>(0);
    $74 = $73 ? $49 : 0;
    $75 = (+($74|0));
    $dst_y0$1 = $70 ? $72 : $dst_y0$0;
    $76 = HEAP32[$12>>2]|0;
    $77 = (+($76|0));
    $78 = !($54 >= $77);
    if ($78) {
     $dst_x0$2 = $dst_x0$1;$dst_x1$2 = $dst_x1$1;$dst_y0$2 = $dst_y0$1;$dst_y1$1 = $dst_y1$0;$src_x0$1 = $60;$src_x1$1 = $src_x1$0;$src_y0$1 = $75;$src_y1$0 = $54;
    } else {
     $79 = $54 - $77;
     $80 = $79 + 1.0;
     $81 = $28 * $80;
     $82 = $dst_y1$0 - $81;
     $83 = (($76) + -1)|0;
     $84 = (+($83|0));
     $dst_x0$2 = $dst_x0$1;$dst_x1$2 = $dst_x1$1;$dst_y0$2 = $dst_y0$1;$dst_y1$1 = $82;$src_x0$1 = $60;$src_x1$1 = $src_x1$0;$src_y0$1 = $75;$src_y1$0 = $84;
    }
   }
   $85 = ((($dst)) + 36|0);
   $86 = HEAP32[$85>>2]|0;
   $87 = (+($86|0));
   $88 = $dst_x0$2 - $87;
   $89 = $dst_x1$2 - $87;
   $90 = ((($dst)) + 40|0);
   $91 = HEAP32[$90>>2]|0;
   $92 = (+($91|0));
   $93 = $dst_y0$2 - $92;
   $94 = $dst_y1$1 - $92;
   $95 = $88 < 0.0;
   if ($95) {
    $96 = $88 / $25;
    $97 = $src_x0$1 - $96;
    $dst_x0$3 = 0.0;$src_x0$2 = $97;
   } else {
    $dst_x0$3 = $88;$src_x0$2 = $src_x0$1;
   }
   $98 = ((($dst)) + 44|0);
   $99 = HEAP32[$98>>2]|0;
   $100 = (+($99|0));
   $101 = !($89 >= $100);
   if ($101) {
    $dst_x1$3 = $89;$src_x1$2 = $src_x1$1;
   } else {
    $102 = $89 - $100;
    $103 = $102 + 1.0;
    $104 = $103 / $25;
    $105 = $src_x1$1 - $104;
    $106 = (($99) + -1)|0;
    $107 = (+($106|0));
    $dst_x1$3 = $107;$src_x1$2 = $105;
   }
   $108 = $93 < 0.0;
   if ($108) {
    $109 = $93 / $28;
    $110 = $src_y0$1 - $109;
    $dst_y0$3 = 0.0;$src_y0$2 = $110;
   } else {
    $dst_y0$3 = $93;$src_y0$2 = $src_y0$1;
   }
   $111 = ((($dst)) + 48|0);
   $112 = HEAP32[$111>>2]|0;
   $113 = (+($112|0));
   $114 = !($94 >= $113);
   if ($114) {
    $dst_y1$2 = $94;$src_y1$1 = $src_y1$0;
   } else {
    $115 = $94 - $113;
    $116 = $115 + 1.0;
    $117 = $116 / $28;
    $118 = $src_y1$0 - $117;
    $119 = (($112) + -1)|0;
    $120 = (+($119|0));
    $dst_y1$2 = $120;$src_y1$1 = $118;
   }
   $121 = $dst_x0$3 + $87;
   $122 = $dst_x1$3 + $87;
   $123 = $dst_y0$3 + $92;
   $124 = $dst_y1$2 + $92;
   $125 = $src_x0$2 + 0.5;
   $126 = (+_SDL_floor($125));
   $127 = (~~(($126)));
   HEAP32[$final_src>>2] = $127;
   $128 = $src_y0$2 + 0.5;
   $129 = (+_SDL_floor($128));
   $130 = (~~(($129)));
   $131 = ((($final_src)) + 4|0);
   HEAP32[$131>>2] = $130;
   $132 = $src_x1$2 - $src_x0$2;
   $133 = $132 + 1.5;
   $134 = (+_SDL_floor($133));
   $135 = (~~(($134)));
   $136 = ((($final_src)) + 8|0);
   HEAP32[$136>>2] = $135;
   $137 = $src_y1$1 - $src_y0$2;
   $138 = $137 + 1.5;
   $139 = (+_SDL_floor($138));
   $140 = (~~(($139)));
   $141 = ((($final_src)) + 12|0);
   HEAP32[$141>>2] = $140;
   $142 = $121 + 0.5;
   $143 = (+_SDL_floor($142));
   $144 = (~~(($143)));
   HEAP32[$final_dst>>2] = $144;
   $145 = $123 + 0.5;
   $146 = (+_SDL_floor($145));
   $147 = (~~(($146)));
   $148 = ((($final_dst)) + 4|0);
   HEAP32[$148>>2] = $147;
   $149 = $122 - $121;
   $150 = $149 + 1.5;
   $151 = (+_SDL_floor($150));
   $152 = (~~(($151)));
   $153 = ((($final_dst)) + 8|0);
   HEAP32[$153>>2] = $152;
   $154 = $124 - $123;
   $155 = $154 + 1.5;
   $156 = (+_SDL_floor($155));
   $157 = (~~(($156)));
   $158 = ((($final_dst)) + 12|0);
   HEAP32[$158>>2] = $157;
   $159 = ($152|0)<(0);
   if ($159) {
    HEAP32[$153>>2] = 0;
    $162 = 0;
   } else {
    $162 = $152;
   }
   $160 = ($157|0)<(0);
   if ($160) {
    HEAP32[$158>>2] = 0;
    $164 = 0;
   } else {
    $164 = $157;
   }
   if (!($15)) {
    ;HEAP32[$dstrect>>2]=HEAP32[$final_dst>>2]|0;HEAP32[$dstrect+4>>2]=HEAP32[$final_dst+4>>2]|0;HEAP32[$dstrect+8>>2]=HEAP32[$final_dst+8>>2]|0;HEAP32[$dstrect+12>>2]=HEAP32[$final_dst+12>>2]|0;
   }
   $161 = ($162|0)==(0);
   $163 = ($164|0)==(0);
   $or$cond4 = $161 | $163;
   $165 = ($135|0)<(1);
   $or$cond6 = $or$cond4 | $165;
   $166 = ($140|0)<(1);
   $or$cond8 = $or$cond6 | $166;
   if ($or$cond8) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   $167 = (_SDL_LowerBlitScaled($src,$final_src,$dst,$final_dst)|0);
   $$0 = $167;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $9 = (_SDL_SetError(604041,$vararg_buffer1)|0);
 $$0 = $9;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_floor($x) {
 $x = +$x;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+Math_floor((+$x)));
 return (+$0);
}
function _SDL_LowerBlitScaled($src,$srcrect,$dst,$dstrect) {
 $src = $src|0;
 $srcrect = $srcrect|0;
 $dst = $dst|0;
 $dstrect = $dstrect|0;
 var $$0 = 0, $$mask = 0, $$off = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$i = 0, $$pre2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $switch = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($src)) + 52|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 68|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 512;
 $5 = ($4|0)==(0);
 if ($5) {
  $6 = $3 | 512;
  HEAP32[$2>>2] = $6;
  _SDL_InvalidateMap($1);
  $$pre = HEAP32[$0>>2]|0;
  $$phi$trans$insert = ((($$pre)) + 68|0);
  $$pre2 = HEAP32[$$phi$trans$insert>>2]|0;
  $23 = $$pre;$8 = $$pre2;
 } else {
  $23 = $1;$8 = $3;
 }
 $7 = $8 & 371;
 $9 = ($7|0)==(0);
 do {
  if ($9) {
   $10 = ((($src)) + 4|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = HEAP32[$11>>2]|0;
   $13 = ((($dst)) + 4|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = HEAP32[$14>>2]|0;
   $16 = ($12|0)==($15|0);
   if ($16) {
    $17 = ($12|0)==(0);
    $$mask = $12 & -268435456;
    $18 = ($$mask|0)==(268435456);
    $or$cond = $17 | $18;
    if ($or$cond) {
     $19 = $12 >>> 24;
     $20 = $19 & 15;
     $$off = (($20) + -1)|0;
     $switch = ($$off>>>0)<(3);
     if ($switch) {
      break;
     }
    }
    $21 = (_SDL_SoftStretch($src,$srcrect,$dst,$dstrect)|0);
    $$0 = $21;
    return ($$0|0);
   }
  }
 } while(0);
 $22 = HEAP32[$23>>2]|0;
 $24 = ($22|0)==($dst|0);
 do {
  if ($24) {
   $25 = ((($dst)) + 4|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = ((($26)) + 4|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = ($28|0)==(0|0);
   if (!($29)) {
    $30 = ((($23)) + 80|0);
    $31 = HEAP32[$30>>2]|0;
    $32 = ((($28)) + 8|0);
    $33 = HEAP32[$32>>2]|0;
    $34 = ($31|0)==($33|0);
    if (!($34)) {
     label = 13;
     break;
    }
   }
   $35 = ((($src)) + 4|0);
   $36 = HEAP32[$35>>2]|0;
   $37 = ((($36)) + 4|0);
   $38 = HEAP32[$37>>2]|0;
   $39 = ($38|0)==(0|0);
   if ($39) {
    $48 = $23;
   } else {
    $40 = ((($23)) + 84|0);
    $41 = HEAP32[$40>>2]|0;
    $42 = ((($38)) + 8|0);
    $43 = HEAP32[$42>>2]|0;
    $44 = ($41|0)==($43|0);
    if ($44) {
     $48 = $23;
    } else {
     label = 13;
    }
   }
  } else {
   label = 13;
  }
 } while(0);
 do {
  if ((label|0) == 13) {
   $45 = (_SDL_MapSurface($src,$dst)|0);
   $46 = ($45|0)<(0);
   if ($46) {
    $$0 = -1;
    return ($$0|0);
   } else {
    $$pre$i = HEAP32[$0>>2]|0;
    $48 = $$pre$i;
    break;
   }
  }
 } while(0);
 $47 = ((($48)) + 8|0);
 $49 = HEAP32[$47>>2]|0;
 $50 = (FUNCTION_TABLE_iiiii[$49 & 8191]($src,$srcrect,$dst,$dstrect)|0);
 $$0 = $50;
 return ($$0|0);
}
function _SDL_SoftStretch($src,$srcrect,$dst,$dstrect) {
 $src = $src|0;
 $srcrect = $srcrect|0;
 $dst = $dst|0;
 $dstrect = $dstrect|0;
 var $$0 = 0, $$01 = 0, $$011$i = 0, $$016$i = 0, $$016$i25 = 0, $$016$i7 = 0, $$018$i = 0, $$02 = 0, $$09$i = 0, $$09$i22 = 0, $$09$i4 = 0, $$1$lcssa$i = 0, $$1$lcssa$i14 = 0, $$1$lcssa$i19 = 0, $$1$lcssa$i32 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $dst_locked$0 = 0, $dst_row$043 = 0, $full_dst = 0, $full_src = 0, $i$07$i = 0, $i$07$i24 = 0, $i$07$i6 = 0, $i$09$i = 0, $pixel$010$i = 0, $pixel$010$i21 = 0, $pixel$010$i3 = 0, $pixel$1$lcssa$i = 0, $pixel$1$lcssa$i13 = 0, $pixel$1$lcssa$i31 = 0, $pixel$sroa$0$012$i = 0, $pixel$sroa$0$1$lcssa$i = 0, $pixel$sroa$5$013$i = 0;
 var $pixel$sroa$5$1$lcssa$i = 0, $pixel$sroa$7$014$i = 0, $pixel$sroa$7$1$lcssa$i = 0, $pos$010$i = 0, $pos$041 = 0, $pos$08$i = 0, $pos$08$i23 = 0, $pos$08$i5 = 0, $pos$1$lcssa = 0, $pos$1$lcssa$i = 0, $pos$1$lcssa$i15 = 0, $pos$1$lcssa$i20 = 0, $pos$1$lcssa$i33 = 0, $scevgep$i = 0, $scevgep$i17 = 0, $scevgep$i27 = 0, $scevgep$i9 = 0, $scevgep13$i = 0, $scevgep13$i10 = 0, $scevgep13$i28 = 0;
 var $scevgep14$i = 0, $scevgep14$i11 = 0, $scevgep14$i29 = 0, $scevgep20$i = 0, $scevgep21$i = 0, $src_locked$0 = 0, $src_row$042 = 0, $src_row$1$lcssa = 0, $srcp$044 = 0, $srcp$1$lcssa = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $full_src = sp + 56|0;
 $full_dst = sp + 40|0;
 $0 = ((($dst)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 9|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 $5 = ((($src)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = HEAP32[$6>>2]|0;
 $8 = HEAP32[$1>>2]|0;
 $9 = ($7|0)==($8|0);
 if (!($9)) {
  $10 = (_SDL_SetError(604124,$vararg_buffer)|0);
  $$0 = $10;
  STACKTOP = sp;return ($$0|0);
 }
 $11 = ($srcrect|0)==(0|0);
 do {
  if ($11) {
   HEAP32[$full_src>>2] = 0;
   $30 = ((($full_src)) + 4|0);
   HEAP32[$30>>2] = 0;
   $31 = ((($src)) + 8|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = ((($full_src)) + 8|0);
   HEAP32[$33>>2] = $32;
   $34 = ((($src)) + 12|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = ((($full_src)) + 12|0);
   HEAP32[$36>>2] = $35;
   $$01 = $full_src;
  } else {
   $12 = HEAP32[$srcrect>>2]|0;
   $13 = ($12|0)<(0);
   if (!($13)) {
    $14 = ((($srcrect)) + 4|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = ($15|0)<(0);
    if (!($16)) {
     $17 = ((($srcrect)) + 8|0);
     $18 = HEAP32[$17>>2]|0;
     $19 = (($18) + ($12))|0;
     $20 = ((($src)) + 8|0);
     $21 = HEAP32[$20>>2]|0;
     $22 = ($19|0)>($21|0);
     if (!($22)) {
      $23 = ((($srcrect)) + 12|0);
      $24 = HEAP32[$23>>2]|0;
      $25 = (($24) + ($15))|0;
      $26 = ((($src)) + 12|0);
      $27 = HEAP32[$26>>2]|0;
      $28 = ($25|0)>($27|0);
      if (!($28)) {
       $$01 = $srcrect;
       break;
      }
     }
    }
   }
   $29 = (_SDL_SetError(604161,$vararg_buffer1)|0);
   $$0 = $29;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 $37 = ($dstrect|0)==(0|0);
 do {
  if ($37) {
   HEAP32[$full_dst>>2] = 0;
   $56 = ((($full_dst)) + 4|0);
   HEAP32[$56>>2] = 0;
   $57 = ((($dst)) + 8|0);
   $58 = HEAP32[$57>>2]|0;
   $59 = ((($full_dst)) + 8|0);
   HEAP32[$59>>2] = $58;
   $60 = ((($dst)) + 12|0);
   $61 = HEAP32[$60>>2]|0;
   $62 = ((($full_dst)) + 12|0);
   HEAP32[$62>>2] = $61;
   $$02 = $full_dst;
  } else {
   $38 = HEAP32[$dstrect>>2]|0;
   $39 = ($38|0)<(0);
   if (!($39)) {
    $40 = ((($dstrect)) + 4|0);
    $41 = HEAP32[$40>>2]|0;
    $42 = ($41|0)<(0);
    if (!($42)) {
     $43 = ((($dstrect)) + 8|0);
     $44 = HEAP32[$43>>2]|0;
     $45 = (($44) + ($38))|0;
     $46 = ((($dst)) + 8|0);
     $47 = HEAP32[$46>>2]|0;
     $48 = ($45|0)>($47|0);
     if (!($48)) {
      $49 = ((($dstrect)) + 12|0);
      $50 = HEAP32[$49>>2]|0;
      $51 = (($50) + ($41))|0;
      $52 = ((($dst)) + 12|0);
      $53 = HEAP32[$52>>2]|0;
      $54 = ($51|0)>($53|0);
      if (!($54)) {
       $$02 = $dstrect;
       break;
      }
     }
    }
   }
   $55 = (_SDL_SetError(604191,$vararg_buffer3)|0);
   $$0 = $55;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 $63 = HEAP32[$dst>>2]|0;
 $64 = $63 & 2;
 $65 = ($64|0)==(0);
 if ($65) {
  $dst_locked$0 = 0;
 } else {
  $66 = (_SDL_LockSurface($dst)|0);
  $67 = ($66|0)<(0);
  if ($67) {
   $68 = (_SDL_SetError(604226,$vararg_buffer5)|0);
   $$0 = $68;
   STACKTOP = sp;return ($$0|0);
  } else {
   $dst_locked$0 = 1;
  }
 }
 $69 = HEAP32[$src>>2]|0;
 $70 = $69 & 2;
 $71 = ($70|0)==(0);
 if ($71) {
  $src_locked$0 = 0;
 } else {
  $72 = (_SDL_LockSurface($src)|0);
  $73 = ($72|0)<(0);
  if ($73) {
   $74 = ($dst_locked$0|0)==(0);
   if (!($74)) {
    _SDL_UnlockSurface($dst);
   }
   $75 = (_SDL_SetError(604261,$vararg_buffer7)|0);
   $$0 = $75;
   STACKTOP = sp;return ($$0|0);
  } else {
   $src_locked$0 = 1;
  }
 }
 $76 = ((($$01)) + 12|0);
 $77 = HEAP32[$76>>2]|0;
 $78 = $77 << 16;
 $79 = ((($$02)) + 12|0);
 $80 = HEAP32[$79>>2]|0;
 $81 = (($78|0) / ($80|0))&-1;
 $82 = ((($$02)) + 4|0);
 $83 = HEAP32[$82>>2]|0;
 $84 = (($83) + ($80))|0;
 $85 = ($80|0)>(0);
 L38: do {
  if ($85) {
   $86 = ((($$01)) + 4|0);
   $87 = HEAP32[$86>>2]|0;
   $88 = ((($dst)) + 20|0);
   $89 = ((($dst)) + 16|0);
   $90 = ((($$01)) + 8|0);
   $91 = ((($$02)) + 8|0);
   $92 = ((($src)) + 20|0);
   $93 = ((($src)) + 16|0);
   $dst_row$043 = $83;$pos$041 = 65536;$src_row$042 = $87;$srcp$044 = 0;
   while(1) {
    $94 = HEAP32[$88>>2]|0;
    $95 = HEAP32[$89>>2]|0;
    $96 = Math_imul($95, $dst_row$043)|0;
    $97 = (($94) + ($96)|0);
    $98 = HEAP32[$$02>>2]|0;
    $99 = Math_imul($98, $4)|0;
    $100 = (($97) + ($99)|0);
    $101 = ($pos$041|0)>(65535);
    if ($101) {
     $102 = (($pos$041) + -65536)|0;
     $103 = $102 >>> 16;
     $104 = $103 << 16;
     $105 = (($src_row$042) + 1)|0;
     $106 = (($105) + ($103))|0;
     $107 = (($102) - ($104))|0;
     $108 = (($src_row$042) + ($103))|0;
     $109 = HEAP32[$92>>2]|0;
     $110 = HEAP32[$93>>2]|0;
     $111 = Math_imul($110, $108)|0;
     $112 = (($109) + ($111)|0);
     $113 = HEAP32[$$01>>2]|0;
     $114 = Math_imul($113, $4)|0;
     $115 = (($112) + ($114)|0);
     $pos$1$lcssa = $107;$src_row$1$lcssa = $106;$srcp$1$lcssa = $115;
    } else {
     $pos$1$lcssa = $pos$041;$src_row$1$lcssa = $src_row$042;$srcp$1$lcssa = $srcp$044;
    }
    switch ($4|0) {
    case 1:  {
     $116 = HEAP32[$90>>2]|0;
     $117 = HEAP32[$91>>2]|0;
     $118 = $116 << 16;
     $119 = (($118|0) / ($117|0))&-1;
     $120 = ($117|0)>(0);
     if ($120) {
      $$016$i = $100;$$09$i = $srcp$1$lcssa;$i$07$i = $117;$pixel$010$i = 0;$pos$08$i = 65536;
      while(1) {
       $121 = ($pos$08$i|0)>(65535);
       if ($121) {
        $122 = (($pos$08$i) + -65536)|0;
        $123 = $122 >>> 16;
        $124 = $123 << 16;
        $scevgep$i = ((($$09$i)) + 1|0);
        $scevgep13$i = (($scevgep$i) + ($123)|0);
        $125 = (($122) - ($124))|0;
        $scevgep14$i = (($$09$i) + ($123)|0);
        $126 = HEAP8[$scevgep14$i>>0]|0;
        $$1$lcssa$i = $scevgep13$i;$pixel$1$lcssa$i = $126;$pos$1$lcssa$i = $125;
       } else {
        $$1$lcssa$i = $$09$i;$pixel$1$lcssa$i = $pixel$010$i;$pos$1$lcssa$i = $pos$08$i;
       }
       $127 = ((($$016$i)) + 1|0);
       HEAP8[$$016$i>>0] = $pixel$1$lcssa$i;
       $128 = (($pos$1$lcssa$i) + ($119))|0;
       $129 = (($i$07$i) + -1)|0;
       $130 = ($i$07$i|0)>(1);
       if ($130) {
        $$016$i = $127;$$09$i = $$1$lcssa$i;$i$07$i = $129;$pixel$010$i = $pixel$1$lcssa$i;$pos$08$i = $128;
       } else {
        break;
       }
      }
     }
     break;
    }
    case 2:  {
     $131 = HEAP32[$90>>2]|0;
     $132 = HEAP32[$91>>2]|0;
     $133 = $131 << 16;
     $134 = (($133|0) / ($132|0))&-1;
     $135 = ($132|0)>(0);
     if ($135) {
      $$016$i25 = $100;$$09$i22 = $srcp$1$lcssa;$i$07$i24 = $132;$pixel$010$i21 = 0;$pos$08$i23 = 65536;
      while(1) {
       $136 = ($pos$08$i23|0)>(65535);
       if ($136) {
        $137 = (($pos$08$i23) + -65536)|0;
        $138 = $137 >>> 16;
        $139 = $138 << 16;
        $scevgep$i27 = ((($$09$i22)) + 2|0);
        $scevgep13$i28 = (($scevgep$i27) + ($138<<1)|0);
        $140 = (($137) - ($139))|0;
        $scevgep14$i29 = (($$09$i22) + ($138<<1)|0);
        $141 = HEAP16[$scevgep14$i29>>1]|0;
        $$1$lcssa$i32 = $scevgep13$i28;$pixel$1$lcssa$i31 = $141;$pos$1$lcssa$i33 = $140;
       } else {
        $$1$lcssa$i32 = $$09$i22;$pixel$1$lcssa$i31 = $pixel$010$i21;$pos$1$lcssa$i33 = $pos$08$i23;
       }
       $142 = ((($$016$i25)) + 2|0);
       HEAP16[$$016$i25>>1] = $pixel$1$lcssa$i31;
       $143 = (($pos$1$lcssa$i33) + ($134))|0;
       $144 = (($i$07$i24) + -1)|0;
       $145 = ($i$07$i24|0)>(1);
       if ($145) {
        $$016$i25 = $142;$$09$i22 = $$1$lcssa$i32;$i$07$i24 = $144;$pixel$010$i21 = $pixel$1$lcssa$i31;$pos$08$i23 = $143;
       } else {
        break;
       }
      }
     }
     break;
    }
    case 3:  {
     $146 = HEAP32[$90>>2]|0;
     $147 = HEAP32[$91>>2]|0;
     $148 = $146 << 16;
     $149 = (($148|0) / ($147|0))&-1;
     $150 = ($147|0)>(0);
     if ($150) {
      $$011$i = $srcp$1$lcssa;$$018$i = $100;$i$09$i = $147;$pixel$sroa$0$012$i = 0;$pixel$sroa$5$013$i = 0;$pixel$sroa$7$014$i = 0;$pos$010$i = 65536;
      while(1) {
       $151 = ($pos$010$i|0)>(65535);
       if ($151) {
        $152 = (($pos$010$i) + -65536)|0;
        $153 = $152 >>> 16;
        $154 = $153 << 16;
        $scevgep$i17 = ((($$011$i)) + 3|0);
        $155 = ($153*3)|0;
        $scevgep20$i = (($scevgep$i17) + ($155)|0);
        $156 = (($152) - ($154))|0;
        $scevgep21$i = (($$011$i) + ($155)|0);
        $157 = ((($scevgep21$i)) + 1|0);
        $158 = HEAP8[$scevgep21$i>>0]|0;
        $159 = ((($scevgep21$i)) + 2|0);
        $160 = HEAP8[$157>>0]|0;
        $161 = HEAP8[$159>>0]|0;
        $$1$lcssa$i19 = $scevgep20$i;$pixel$sroa$0$1$lcssa$i = $158;$pixel$sroa$5$1$lcssa$i = $160;$pixel$sroa$7$1$lcssa$i = $161;$pos$1$lcssa$i20 = $156;
       } else {
        $$1$lcssa$i19 = $$011$i;$pixel$sroa$0$1$lcssa$i = $pixel$sroa$0$012$i;$pixel$sroa$5$1$lcssa$i = $pixel$sroa$5$013$i;$pixel$sroa$7$1$lcssa$i = $pixel$sroa$7$014$i;$pos$1$lcssa$i20 = $pos$010$i;
       }
       $162 = ((($$018$i)) + 1|0);
       HEAP8[$$018$i>>0] = $pixel$sroa$0$1$lcssa$i;
       $163 = ((($$018$i)) + 2|0);
       HEAP8[$162>>0] = $pixel$sroa$5$1$lcssa$i;
       $164 = ((($$018$i)) + 3|0);
       HEAP8[$163>>0] = $pixel$sroa$7$1$lcssa$i;
       $165 = (($pos$1$lcssa$i20) + ($149))|0;
       $166 = (($i$09$i) + -1)|0;
       $167 = ($i$09$i|0)>(1);
       if ($167) {
        $$011$i = $$1$lcssa$i19;$$018$i = $164;$i$09$i = $166;$pixel$sroa$0$012$i = $pixel$sroa$0$1$lcssa$i;$pixel$sroa$5$013$i = $pixel$sroa$5$1$lcssa$i;$pixel$sroa$7$014$i = $pixel$sroa$7$1$lcssa$i;$pos$010$i = $165;
       } else {
        break;
       }
      }
     }
     break;
    }
    case 4:  {
     $168 = HEAP32[$90>>2]|0;
     $169 = HEAP32[$91>>2]|0;
     $170 = $168 << 16;
     $171 = (($170|0) / ($169|0))&-1;
     $172 = ($169|0)>(0);
     if ($172) {
      $$016$i7 = $100;$$09$i4 = $srcp$1$lcssa;$i$07$i6 = $169;$pixel$010$i3 = 0;$pos$08$i5 = 65536;
      while(1) {
       $173 = ($pos$08$i5|0)>(65535);
       if ($173) {
        $174 = (($pos$08$i5) + -65536)|0;
        $175 = $174 >>> 16;
        $176 = $175 << 16;
        $scevgep$i9 = ((($$09$i4)) + 4|0);
        $scevgep13$i10 = (($scevgep$i9) + ($175<<2)|0);
        $177 = (($174) - ($176))|0;
        $scevgep14$i11 = (($$09$i4) + ($175<<2)|0);
        $178 = HEAP32[$scevgep14$i11>>2]|0;
        $$1$lcssa$i14 = $scevgep13$i10;$pixel$1$lcssa$i13 = $178;$pos$1$lcssa$i15 = $177;
       } else {
        $$1$lcssa$i14 = $$09$i4;$pixel$1$lcssa$i13 = $pixel$010$i3;$pos$1$lcssa$i15 = $pos$08$i5;
       }
       $179 = ((($$016$i7)) + 4|0);
       HEAP32[$$016$i7>>2] = $pixel$1$lcssa$i13;
       $180 = (($pos$1$lcssa$i15) + ($171))|0;
       $181 = (($i$07$i6) + -1)|0;
       $182 = ($i$07$i6|0)>(1);
       if ($182) {
        $$016$i7 = $179;$$09$i4 = $$1$lcssa$i14;$i$07$i6 = $181;$pixel$010$i3 = $pixel$1$lcssa$i13;$pos$08$i5 = $180;
       } else {
        break;
       }
      }
     }
     break;
    }
    default: {
    }
    }
    $183 = (($pos$1$lcssa) + ($81))|0;
    $184 = (($dst_row$043) + 1)|0;
    $185 = ($184|0)<($84|0);
    if ($185) {
     $dst_row$043 = $184;$pos$041 = $183;$src_row$042 = $src_row$1$lcssa;$srcp$044 = $srcp$1$lcssa;
    } else {
     break L38;
    }
   }
  }
 } while(0);
 $186 = ($dst_locked$0|0)==(0);
 if (!($186)) {
  _SDL_UnlockSurface($dst);
 }
 $187 = ($src_locked$0|0)==(0);
 if ($187) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 _SDL_UnlockSurface($src);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SW_RenderCopyEx($renderer,$texture,$srcrect,$dstrect,$angle,$center,$flip) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 $srcrect = $srcrect|0;
 $dstrect = $dstrect|0;
 $angle = +$angle;
 $center = $center|0;
 $flip = $flip|0;
 var $$ = 0.0, $$$i = 0, $$0 = 0, $$1 = 0, $$pr = 0, $$pre = 0, $$pre10 = 0, $$pre11 = 0, $$pre7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0.0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0.0, $123 = 0.0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0.0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0.0, $132 = 0, $133 = 0, $134 = 0.0, $135 = 0.0, $136 = 0, $137 = 0.0, $138 = 0, $139 = 0.0, $14 = 0, $140 = 0.0, $141 = 0.0, $142 = 0.0, $143 = 0.0;
 var $144 = 0.0, $145 = 0.0, $146 = 0.0, $147 = 0.0, $148 = 0.0, $149 = 0.0, $15 = 0, $150 = 0.0, $151 = 0, $152 = 0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0.0, $157 = 0.0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0.0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0.0, $17 = 0, $170 = 0.0, $171 = 0.0, $172 = 0.0, $173 = 0, $174 = 0.0, $175 = 0, $176 = 0.0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0.0, $181 = 0, $182 = 0.0, $183 = 0, $184 = 0.0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0.0, $41 = 0, $42 = 0.0, $43 = 0, $44 = 0, $45 = 0.0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $alphaMod = 0;
 var $b = 0, $blendMode = 0, $cangle = 0, $cloneSource$0 = 0, $colorkey = 0, $dstheight = 0, $dstwidth = 0, $final_rect$sroa$0$0 = 0, $final_rect$sroa$0$0$in = 0.0, $final_rect$sroa$10$0 = 0, $final_rect$sroa$10$0$in = 0.0, $g = 0, $not$$i = 0, $r = 0, $retval$0 = 0, $retval$2 = 0, $sangle = 0, $surface_scaled$04 = 0, $surface_scaled$05 = 0, $tmp_rect = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tmp_rect = sp + 32|0;
 $dstwidth = sp + 28|0;
 $dstheight = sp + 24|0;
 $cangle = sp + 8|0;
 $sangle = sp;
 $colorkey = sp + 20|0;
 $blendMode = sp + 16|0;
 $alphaMod = sp + 51|0;
 $r = sp + 50|0;
 $g = sp + 49|0;
 $b = sp + 48|0;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $4 = ((($1)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  HEAP32[$1>>2] = $5;
  $6 = ($5|0)==(0);
  $7 = $5;
  if ($6) {
   $8 = ((($renderer)) + 188|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (_SDL_GetWindowSurface($9)|0);
   $11 = ($10|0)==(0|0);
   do {
    if (!($11)) {
     HEAP32[$4>>2] = $10;
     HEAP32[$1>>2] = $10;
     $12 = HEAP32[$0>>2]|0;
     $13 = HEAP32[$12>>2]|0;
     $14 = ($13|0)==(0|0);
     if (!($14)) {
      $15 = ((($renderer)) + 212|0);
      (_SDL_SetClipRect($13,$15)|0);
      $$pre = HEAP32[$0>>2]|0;
      $$pre7 = HEAP32[$$pre>>2]|0;
      $16 = ($$pre7|0)==(0|0);
      if (!($16)) {
       $17 = ((($renderer)) + 276|0);
       $18 = HEAP32[$17>>2]|0;
       $19 = ($18|0)==(0);
       if ($19) {
        (_SDL_SetClipRect($$pre7,0)|0);
        break;
       } else {
        $20 = ((($renderer)) + 244|0);
        (_SDL_SetClipRect($$pre7,$20)|0);
        break;
       }
      }
     }
    }
   } while(0);
   $$pr = HEAP32[$1>>2]|0;
   $24 = $$pr;
  } else {
   $24 = $7;
  }
 } else {
  $24 = $2;
 }
 $21 = ((($texture)) + 68|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ($24|0)==(0|0);
 if ($23) {
  $$1 = -1;
  STACKTOP = sp;return ($$1|0);
 }
 $25 = ((($renderer)) + 212|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ((($renderer)) + 216|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = $26 | $28;
 $30 = ($29|0)==(0);
 if ($30) {
  $38 = +HEAPF32[$dstrect>>2];
  $39 = ((($dstrect)) + 4|0);
  $40 = +HEAPF32[$39>>2];
  $final_rect$sroa$0$0$in = $38;$final_rect$sroa$10$0$in = $40;
 } else {
  $31 = (+($26|0));
  $32 = +HEAPF32[$dstrect>>2];
  $33 = $31 + $32;
  $34 = (+($28|0));
  $35 = ((($dstrect)) + 4|0);
  $36 = +HEAPF32[$35>>2];
  $37 = $34 + $36;
  $final_rect$sroa$0$0$in = $33;$final_rect$sroa$10$0$in = $37;
 }
 $final_rect$sroa$10$0 = (~~(($final_rect$sroa$10$0$in)));
 $final_rect$sroa$0$0 = (~~(($final_rect$sroa$0$0$in)));
 $41 = ((($dstrect)) + 8|0);
 $42 = +HEAPF32[$41>>2];
 $43 = (~~(($42)));
 $44 = ((($dstrect)) + 12|0);
 $45 = +HEAPF32[$44>>2];
 $46 = (~~(($45)));
 $47 = ((($tmp_rect)) + 4|0);
 $48 = ((($tmp_rect)) + 8|0);
 HEAP32[$48>>2] = $43;
 $49 = ((($tmp_rect)) + 12|0);
 HEAP32[$49>>2] = $46;
 HEAP32[$tmp_rect>>2] = 0;
 HEAP32[$47>>2] = 0;
 $50 = ((($srcrect)) + 8|0);
 $51 = HEAP32[$50>>2]|0;
 $52 = ($51|0)==($43|0);
 if ($52) {
  $53 = ((($srcrect)) + 12|0);
  $54 = HEAP32[$53>>2]|0;
  $55 = ($54|0)==($46|0);
  if ($55) {
   $56 = HEAP32[$srcrect>>2]|0;
   $57 = ($56|0)==(0);
   if ($57) {
    $58 = ((($srcrect)) + 4|0);
    $59 = HEAP32[$58>>2]|0;
    $60 = ($59|0)==(0);
    if ($60) {
     $107 = $43;$108 = $46;$surface_scaled$05 = $22;
     label = 32;
    } else {
     label = 18;
    }
   } else {
    label = 18;
   }
  } else {
   label = 18;
  }
 } else {
  label = 18;
 }
 if ((label|0) == 18) {
  $61 = ((($22)) + 4|0);
  $62 = HEAP32[$61>>2]|0;
  $63 = ((($62)) + 8|0);
  $64 = HEAP8[$63>>0]|0;
  $65 = $64&255;
  $66 = ((($62)) + 12|0);
  $67 = HEAP32[$66>>2]|0;
  $68 = ((($62)) + 16|0);
  $69 = HEAP32[$68>>2]|0;
  $70 = ((($62)) + 20|0);
  $71 = HEAP32[$70>>2]|0;
  $72 = ((($62)) + 24|0);
  $73 = HEAP32[$72>>2]|0;
  $74 = (_SDL_CreateRGBSurface(0,$43,$46,$65,$67,$69,$71,$73)|0);
  $75 = ($74|0)==(0|0);
  do {
   if ($75) {
    $$0 = 0;$retval$0 = 0;
   } else {
    $76 = (_SDL_GetColorKey($22,$colorkey)|0);
    $77 = ($76|0)==(0);
    if ($77) {
     $78 = HEAP32[$colorkey>>2]|0;
     (_SDL_SetColorKey($74,1,$78)|0);
     $cloneSource$0 = 1;
    } else {
     $cloneSource$0 = 0;
    }
    (_SDL_GetSurfaceAlphaMod($22,$alphaMod)|0);
    (_SDL_GetSurfaceBlendMode($22,$blendMode)|0);
    (_SDL_GetSurfaceColorMod($22,$r,$g,$b)|0);
    $79 = HEAP32[$blendMode>>2]|0;
    $80 = ($79|0)==(0);
    if ($80) {
     $81 = HEAP8[$alphaMod>>0]|0;
     $82 = HEAP8[$r>>0]|0;
     $83 = $82 & $81;
     $84 = HEAP8[$g>>0]|0;
     $85 = $83 & $84;
     $86 = HEAP8[$b>>0]|0;
     $87 = $85 & $86;
     $88 = ($87<<24>>24)!=(-1);
     $90 = $88;
    } else {
     $90 = 1;
    }
    $89 = $90&1;
    $91 = $89 | $cloneSource$0;
    $92 = ($91|0)==(0);
    if ($92) {
     $93 = (_SDL_UpperBlitScaled($22,$srcrect,$74,$tmp_rect)|0);
     $$0 = 1;$retval$0 = $93;
     break;
    }
    $94 = HEAP32[$61>>2]|0;
    $95 = HEAP32[$22>>2]|0;
    $96 = (_SDL_ConvertSurface($22,$94,$95)|0);
    $97 = ($96|0)==(0|0);
    if ($97) {
     _SDL_FreeSurface($74);
     $$0 = 0;$retval$0 = 0;
     break;
    }
    (_SDL_SetSurfaceAlphaMod($96,-1)|0);
    (_SDL_SetSurfaceBlendMode($96,0)|0);
    (_SDL_SetColorKey($96,0,0)|0);
    (_SDL_SetSurfaceColorMod($96,-1,-1,-1)|0);
    (_SDL_SetSurfaceRLE($96,0)|0);
    $98 = HEAP8[$alphaMod>>0]|0;
    (_SDL_SetSurfaceAlphaMod($74,$98)|0);
    $99 = HEAP32[$blendMode>>2]|0;
    (_SDL_SetSurfaceBlendMode($74,$99)|0);
    $100 = HEAP8[$r>>0]|0;
    $101 = HEAP8[$g>>0]|0;
    $102 = HEAP8[$b>>0]|0;
    (_SDL_SetSurfaceColorMod($74,$100,$101,$102)|0);
    $103 = (_SDL_UpperBlitScaled($96,$srcrect,$74,$tmp_rect)|0);
    $104 = ($96|0)==($22|0);
    if ($104) {
     $$0 = 1;$retval$0 = $103;
    } else {
     _SDL_FreeSurface($96);
     $$0 = 1;$retval$0 = $103;
    }
   }
  } while(0);
  if (!($$0)) {
   $$1 = -1;
   STACKTOP = sp;return ($$1|0);
  }
  $105 = ($retval$0|0)==(0);
  if ($105) {
   $$pre10 = HEAP32[$48>>2]|0;
   $$pre11 = HEAP32[$49>>2]|0;
   $107 = $$pre10;$108 = $$pre11;$surface_scaled$05 = $74;
   label = 32;
  } else {
   $retval$2 = $retval$0;$surface_scaled$04 = $74;
  }
 }
 if ((label|0) == 32) {
  $106 = -$angle;
  _SDLgfx_rotozoomSurfaceSizeTrig($107,$108,$106,$dstwidth,$dstheight,$cangle,$sangle);
  $109 = HEAP32[$dstwidth>>2]|0;
  $110 = (($109|0) / 2)&-1;
  $111 = HEAP32[$dstheight>>2]|0;
  $112 = (($111|0) / 2)&-1;
  $113 = (_SDL_GetHint(604329)|0);
  $114 = ($113|0)==(0|0);
  if ($114) {
   $124 = 0;
  } else {
   $115 = HEAP8[$113>>0]|0;
   $116 = ($115<<24>>24)==(48);
   if ($116) {
    $124 = 0;
   } else {
    $117 = (_SDL_strcasecmp($113,604354)|0);
    $not$$i = ($117|0)!=(0);
    $$$i = $not$$i&1;
    $124 = $$$i;
   }
  }
  $118 = $flip & 1;
  $119 = $flip & 2;
  $120 = HEAP32[$dstwidth>>2]|0;
  $121 = HEAP32[$dstheight>>2]|0;
  $122 = +HEAPF64[$cangle>>3];
  $123 = +HEAPF64[$sangle>>3];
  $125 = (_SDLgfx_rotateSurface($surface_scaled$05,$106,$110,$112,$124,$118,$119,$120,$121,$122,$123)|0);
  $126 = ($125|0)==(0|0);
  if ($126) {
   $retval$2 = 0;$surface_scaled$04 = $surface_scaled$05;
  } else {
   $127 = +HEAPF32[$center>>2];
   $128 = (~~(($127)));
   $129 = (($128) + ($final_rect$sroa$0$0))|0;
   $130 = ((($center)) + 4|0);
   $131 = +HEAPF32[$130>>2];
   $132 = (~~(($131)));
   $133 = (($132) + ($final_rect$sroa$10$0))|0;
   $134 = +HEAPF64[$sangle>>3];
   $135 = -$134;
   HEAPF64[$sangle>>3] = $135;
   $136 = (0 - ($128))|0;
   $137 = (+($136|0));
   $138 = (0 - ($132))|0;
   $139 = (+($138|0));
   $140 = +HEAPF64[$cangle>>3];
   $141 = $137 * $140;
   $142 = $139 * $135;
   $143 = $141 - $142;
   $144 = (+($129|0));
   $145 = $144 + $143;
   $146 = $137 * $135;
   $147 = $139 * $140;
   $148 = $146 + $147;
   $149 = (+($133|0));
   $150 = $149 + $148;
   $151 = (($43) + ($final_rect$sroa$0$0))|0;
   $152 = (($151) - ($129))|0;
   $153 = (+($152|0));
   $154 = $153 * $140;
   $155 = $154 - $142;
   $156 = $144 + $155;
   $157 = $153 * $135;
   $158 = $157 + $147;
   $159 = $149 + $158;
   $160 = (($46) + ($final_rect$sroa$10$0))|0;
   $161 = (($160) - ($133))|0;
   $162 = (+($161|0));
   $163 = $162 * $135;
   $164 = $141 - $163;
   $165 = $144 + $164;
   $166 = $140 * $162;
   $167 = $146 + $166;
   $168 = $149 + $167;
   $169 = $154 - $163;
   $170 = $144 + $169;
   $171 = $157 + $166;
   $172 = $149 + $171;
   $173 = $145 < $156;
   $174 = $173 ? $145 : $156;
   $175 = $165 < $170;
   $176 = $175 ? $165 : $170;
   $177 = $174 < $176;
   $$ = $177 ? $174 : $176;
   $178 = (~~(($$)));
   HEAP32[$tmp_rect>>2] = $178;
   $179 = $150 < $159;
   $180 = $179 ? $150 : $159;
   $181 = $168 < $172;
   $182 = $181 ? $168 : $172;
   $183 = $180 < $182;
   $184 = $183 ? $180 : $182;
   $185 = (~~(($184)));
   HEAP32[$47>>2] = $185;
   $186 = HEAP32[$dstwidth>>2]|0;
   HEAP32[$48>>2] = $186;
   $187 = HEAP32[$dstheight>>2]|0;
   HEAP32[$49>>2] = $187;
   $188 = (_SDL_UpperBlit($125,0,$24,$tmp_rect)|0);
   _SDL_FreeSurface($125);
   $retval$2 = $188;$surface_scaled$04 = $surface_scaled$05;
  }
 }
 $189 = ($surface_scaled$04|0)==($22|0);
 if ($189) {
  $$1 = $retval$2;
  STACKTOP = sp;return ($$1|0);
 }
 _SDL_FreeSurface($surface_scaled$04);
 $$1 = $retval$2;
 STACKTOP = sp;return ($$1|0);
}
function _SDL_GetColorKey($surface,$key) {
 $surface = $surface|0;
 $key = $key|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($surface|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ((($surface)) + 52|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 68|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 & 256;
 $6 = ($5|0)==(0);
 if ($6) {
  $$0 = -1;
  return ($$0|0);
 }
 $7 = ($key|0)==(0|0);
 if ($7) {
  $$0 = 0;
  return ($$0|0);
 }
 $8 = ((($2)) + 72|0);
 $9 = HEAP32[$8>>2]|0;
 HEAP32[$key>>2] = $9;
 $$0 = 0;
 return ($$0|0);
}
function _SDL_SetColorKey($surface,$flag,$key) {
 $surface = $surface|0;
 $flag = $flag|0;
 $key = $key|0;
 var $$ = 0, $$0 = 0, $$1 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($surface|0)==(0|0);
 if ($0) {
  HEAP32[$vararg_buffer>>2] = 604291;
  $1 = (_SDL_SetError(611267,$vararg_buffer)|0);
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ((($surface)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if (!($6)) {
  $7 = HEAP32[$5>>2]|0;
  $8 = ($7>>>0)>($key>>>0);
  if (!($8)) {
   HEAP32[$vararg_buffer1>>2] = 604299;
   $9 = (_SDL_SetError(611267,$vararg_buffer1)|0);
   $$0 = $9;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $10 = $flag & 2;
 $11 = ($10|0)==(0);
 $$pre = ((($surface)) + 52|0);
 if (!($11)) {
  $12 = HEAP32[$$pre>>2]|0;
  $13 = ((($12)) + 68|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = $14 | 4096;
  HEAP32[$13>>2] = $15;
  $16 = ($15|0)==($14|0);
  if (!($16)) {
   _SDL_InvalidateMap($12);
  }
 }
 $17 = HEAP32[$$pre>>2]|0;
 $18 = ((($17)) + 68|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($flag|0)==(0);
 if ($20) {
  $34 = HEAP32[$2>>2]|0;
  $35 = ((($34)) + 4|0);
  $36 = HEAP32[$35>>2]|0;
  $37 = ($36|0)==(0|0);
  if (!($37)) {
   $38 = ((($17)) + 72|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = ((($36)) + 4|0);
   $41 = HEAP32[$40>>2]|0;
   $42 = (((($41) + ($39<<2)|0)) + 3|0);
   HEAP8[$42>>0] = -1;
   $43 = ((($36)) + 8|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = (($44) + 1)|0;
   $46 = ($45|0)==(0);
   $$1 = $46 ? 1 : $45;
   HEAP32[$43>>2] = $$1;
  }
  $47 = $19 & -257;
  HEAP32[$18>>2] = $47;
  $49 = $47;
 } else {
  $21 = $19 | 256;
  HEAP32[$18>>2] = $21;
  $22 = ((($17)) + 72|0);
  HEAP32[$22>>2] = $key;
  $23 = HEAP32[$2>>2]|0;
  $24 = ((($23)) + 4|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = ($25|0)==(0|0);
  if ($26) {
   $49 = $21;
  } else {
   $27 = ((($25)) + 4|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = (((($28) + ($key<<2)|0)) + 3|0);
   HEAP8[$29>>0] = 0;
   $30 = ((($25)) + 8|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = (($31) + 1)|0;
   $33 = ($32|0)==(0);
   $$ = $33 ? 1 : $32;
   HEAP32[$30>>2] = $$;
   $49 = $21;
  }
 }
 $48 = ($49|0)==($19|0);
 if ($48) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 _SDL_InvalidateMap($17);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GetSurfaceAlphaMod($surface,$alpha) {
 $surface = $surface|0;
 $alpha = $alpha|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($surface|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ($alpha|0)==(0|0);
 if ($1) {
  $$0 = 0;
  return ($$0|0);
 }
 $2 = ((($surface)) + 52|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 79|0);
 $5 = HEAP8[$4>>0]|0;
 HEAP8[$alpha>>0] = $5;
 $$0 = 0;
 return ($$0|0);
}
function _SDL_GetSurfaceBlendMode($surface,$blendMode) {
 $surface = $surface|0;
 $blendMode = $blendMode|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($surface|0)==(0|0);
 L1: do {
  if ($0) {
   $$0 = -1;
  } else {
   $1 = ($blendMode|0)==(0|0);
   if ($1) {
    $$0 = 0;
   } else {
    $2 = ((($surface)) + 52|0);
    $3 = HEAP32[$2>>2]|0;
    $4 = ((($3)) + 68|0);
    $5 = HEAP32[$4>>2]|0;
    $6 = $5 & 112;
    switch ($6|0) {
    case 16:  {
     HEAP32[$blendMode>>2] = 1;
     $$0 = 0;
     break L1;
     break;
    }
    case 32:  {
     HEAP32[$blendMode>>2] = 2;
     $$0 = 0;
     break L1;
     break;
    }
    case 64:  {
     HEAP32[$blendMode>>2] = 4;
     $$0 = 0;
     break L1;
     break;
    }
    default: {
     HEAP32[$blendMode>>2] = 0;
     $$0 = 0;
     break L1;
    }
    }
   }
  }
 } while(0);
 return ($$0|0);
}
function _SDL_GetSurfaceColorMod($surface,$r,$g,$b) {
 $surface = $surface|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($surface|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = ($r|0)==(0|0);
 if (!($1)) {
  $2 = ((($surface)) + 52|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($3)) + 76|0);
  $5 = HEAP8[$4>>0]|0;
  HEAP8[$r>>0] = $5;
 }
 $6 = ($g|0)==(0|0);
 if (!($6)) {
  $7 = ((($surface)) + 52|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ((($8)) + 77|0);
  $10 = HEAP8[$9>>0]|0;
  HEAP8[$g>>0] = $10;
 }
 $11 = ($b|0)==(0|0);
 if ($11) {
  $$0 = 0;
  return ($$0|0);
 }
 $12 = ((($surface)) + 52|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($13)) + 78|0);
 $15 = HEAP8[$14>>0]|0;
 HEAP8[$b>>0] = $15;
 $$0 = 0;
 return ($$0|0);
}
function _SDL_ConvertSurface($surface,$format,$flags) {
 $surface = $surface|0;
 $format = $format|0;
 $flags = $flags|0;
 var $$$i = 0, $$$i11 = 0, $$1 = 0, $$1$i$i = 0, $$in$i = 0, $$in15$i = 0, $$in16$i = 0, $$in21$i = 0, $$old = 0, $$old4 = 0, $$phi$trans$insert = 0, $$pr$pre$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i14 = 0, $$pre20$i = 0, $$pre27 = 0, $$pre28 = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0;
 var $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0;
 var $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0;
 var $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0;
 var $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0;
 var $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bounds = 0, $full_rect$i = 0, $i$0$lcssa = 0;
 var $i$019 = 0, $keyA = 0, $keyB = 0, $keyG = 0, $keyR = 0, $row$05$i = 0, $row1$012$i = 0, $spot$04$i = 0, $spot2$08$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $full_rect$i = sp + 48|0;
 $bounds = sp + 32|0;
 $keyR = sp + 67|0;
 $keyG = sp + 66|0;
 $keyB = sp + 65|0;
 $keyA = sp + 64|0;
 $0 = ((($format)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$1>>2]|0;
  $4 = ($3|0)>(0);
  L3: do {
   if ($4) {
    $$phi$trans$insert = ((($1)) + 4|0);
    $$pre27 = HEAP32[$$phi$trans$insert>>2]|0;
    $i$019 = 0;
    while(1) {
     $5 = (($$pre27) + ($i$019<<2)|0);
     $6 = HEAP8[$5>>0]|0;
     $7 = ($6<<24>>24)==(-1);
     if (!($7)) {
      $i$0$lcssa = $i$019;
      break L3;
     }
     $8 = (((($$pre27) + ($i$019<<2)|0)) + 1|0);
     $9 = HEAP8[$8>>0]|0;
     $10 = ($9<<24>>24)==(-1);
     if (!($10)) {
      $i$0$lcssa = $i$019;
      break L3;
     }
     $11 = (((($$pre27) + ($i$019<<2)|0)) + 2|0);
     $12 = HEAP8[$11>>0]|0;
     $13 = ($12<<24>>24)==(-1);
     if (!($13)) {
      $i$0$lcssa = $i$019;
      break L3;
     }
     $14 = (($i$019) + 1)|0;
     $15 = ($14|0)<($3|0);
     if ($15) {
      $i$019 = $14;
     } else {
      $i$0$lcssa = $14;
      break;
     }
    }
   } else {
    $i$0$lcssa = 0;
   }
  } while(0);
  $16 = ($i$0$lcssa|0)==($3|0);
  if ($16) {
   (_SDL_SetError(604303,$vararg_buffer)|0);
   $$1 = 0;
   STACKTOP = sp;return ($$1|0);
  }
 }
 $17 = ((($surface)) + 8|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($surface)) + 12|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ((($format)) + 8|0);
 $22 = HEAP8[$21>>0]|0;
 $23 = $22&255;
 $24 = ((($format)) + 12|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = ((($format)) + 16|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ((($format)) + 20|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = ((($format)) + 24|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = (_SDL_CreateRGBSurface(0,$18,$20,$23,$25,$27,$29,$31)|0);
 $33 = ($32|0)==(0|0);
 if ($33) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 $34 = HEAP32[$0>>2]|0;
 $35 = ($34|0)==(0|0);
 if (!($35)) {
  $36 = ((($32)) + 4|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = ((($37)) + 4|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = ($39|0)==(0|0);
  if (!($40)) {
   $41 = ((($39)) + 4|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = ((($34)) + 4|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = HEAP32[$34>>2]|0;
   $46 = $45 << 2;
   (_SDL_memcpy($42,$44,$46)|0);
   $47 = HEAP32[$0>>2]|0;
   $48 = HEAP32[$47>>2]|0;
   $49 = HEAP32[$36>>2]|0;
   $50 = ((($49)) + 4|0);
   $51 = HEAP32[$50>>2]|0;
   HEAP32[$51>>2] = $48;
  }
 }
 $52 = ((($surface)) + 52|0);
 $53 = HEAP32[$52>>2]|0;
 $54 = ((($53)) + 68|0);
 $55 = HEAP32[$54>>2]|0;
 $56 = ((($53)) + 76|0);
 $57 = HEAP32[$56>>2]|0;
 $58 = $57&65535;
 $59 = $57&255;
 $60 = ($58&65535) >>> 8;
 $61 = $60&255;
 $62 = $57 >>> 16;
 $63 = $62&255;
 $64 = $57 >>> 24;
 $65 = $64&255;
 HEAP32[$56>>2] = -1;
 HEAP32[$54>>2] = 0;
 _SDL_InvalidateMap($53);
 HEAP32[$bounds>>2] = 0;
 $66 = ((($bounds)) + 4|0);
 HEAP32[$66>>2] = 0;
 $67 = HEAP32[$17>>2]|0;
 $68 = ((($bounds)) + 8|0);
 HEAP32[$68>>2] = $67;
 $69 = HEAP32[$19>>2]|0;
 $70 = ((($bounds)) + 12|0);
 HEAP32[$70>>2] = $69;
 $71 = HEAP32[$52>>2]|0;
 $72 = HEAP32[$71>>2]|0;
 $73 = ($72|0)==($32|0);
 do {
  if ($73) {
   $74 = ((($32)) + 4|0);
   $75 = HEAP32[$74>>2]|0;
   $76 = ((($75)) + 4|0);
   $77 = HEAP32[$76>>2]|0;
   $78 = ($77|0)==(0|0);
   if (!($78)) {
    $79 = ((($71)) + 80|0);
    $80 = HEAP32[$79>>2]|0;
    $81 = ((($77)) + 8|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = ($80|0)==($82|0);
    if (!($83)) {
     label = 19;
     break;
    }
   }
   $84 = ((($surface)) + 4|0);
   $85 = HEAP32[$84>>2]|0;
   $86 = ((($85)) + 4|0);
   $87 = HEAP32[$86>>2]|0;
   $88 = ($87|0)==(0|0);
   if ($88) {
    $97 = $71;
    label = 21;
   } else {
    $89 = ((($71)) + 84|0);
    $90 = HEAP32[$89>>2]|0;
    $91 = ((($87)) + 8|0);
    $92 = HEAP32[$91>>2]|0;
    $93 = ($90|0)==($92|0);
    if ($93) {
     $97 = $71;
     label = 21;
    } else {
     label = 19;
    }
   }
  } else {
   label = 19;
  }
 } while(0);
 if ((label|0) == 19) {
  $94 = (_SDL_MapSurface($surface,$32)|0);
  $95 = ($94|0)<(0);
  if (!($95)) {
   $$pre$i = HEAP32[$52>>2]|0;
   $97 = $$pre$i;
   label = 21;
  }
 }
 if ((label|0) == 21) {
  $96 = ((($97)) + 8|0);
  $98 = HEAP32[$96>>2]|0;
  (FUNCTION_TABLE_iiiii[$98 & 8191]($surface,$bounds,$32,$bounds)|0);
 }
 $99 = ((($32)) + 52|0);
 $100 = HEAP32[$99>>2]|0;
 $101 = ((($100)) + 76|0);
 HEAP8[$101>>0] = $59;
 $102 = ((($100)) + 77|0);
 HEAP8[$102>>0] = $61;
 $103 = ((($100)) + 78|0);
 HEAP8[$103>>0] = $63;
 $104 = ((($100)) + 79|0);
 HEAP8[$104>>0] = $65;
 $105 = $55 & -28945;
 $106 = ((($100)) + 68|0);
 HEAP32[$106>>2] = $105;
 $107 = HEAP32[$52>>2]|0;
 $108 = ((($107)) + 76|0);
 HEAP8[$108>>0] = $59;
 $109 = ((($107)) + 77|0);
 HEAP8[$109>>0] = $61;
 $110 = ((($107)) + 78|0);
 HEAP8[$110>>0] = $63;
 $111 = ((($107)) + 79|0);
 HEAP8[$111>>0] = $65;
 $112 = ((($107)) + 68|0);
 HEAP32[$112>>2] = $55;
 _SDL_InvalidateMap($107);
 $113 = $55 & 256;
 $114 = ($113|0)==(0);
 $$pre28 = ((($surface)) + 4|0);
 do {
  if (!($114)) {
   $115 = HEAP32[$$pre28>>2]|0;
   $116 = ((($115)) + 4|0);
   $117 = HEAP32[$116>>2]|0;
   $118 = ($117|0)==(0|0);
   if ($118) {
    $159 = $115;
   } else {
    $119 = HEAP32[$0>>2]|0;
    $120 = ($119|0)==(0|0);
    if (!($120)) {
     $121 = HEAP32[$117>>2]|0;
     $122 = HEAP32[$119>>2]|0;
     $123 = ($121|0)>($122|0);
     if (!($123)) {
      $124 = ((($117)) + 4|0);
      $125 = HEAP32[$124>>2]|0;
      $126 = ((($119)) + 4|0);
      $127 = HEAP32[$126>>2]|0;
      $128 = $121 << 2;
      $129 = (_SDL_memcmp($125,$127,$128)|0);
      $130 = ($129|0)==(0);
      if ($130) {
       $131 = HEAP32[$52>>2]|0;
       $132 = ((($131)) + 72|0);
       $133 = HEAP32[$132>>2]|0;
       $134 = ((($32)) + 4|0);
       $135 = HEAP32[$134>>2]|0;
       $136 = ((($135)) + 4|0);
       $137 = HEAP32[$136>>2]|0;
       $138 = ($137|0)==(0|0);
       if (!($138)) {
        $139 = HEAP32[$137>>2]|0;
        $140 = ($139>>>0)>($133>>>0);
        if (!($140)) {
         HEAP32[$vararg_buffer1>>2] = 604299;
         (_SDL_SetError(611267,$vararg_buffer1)|0);
         break;
        }
       }
       $141 = HEAP32[$99>>2]|0;
       $142 = ((($141)) + 68|0);
       $143 = HEAP32[$142>>2]|0;
       $144 = $143 | 256;
       HEAP32[$142>>2] = $144;
       $145 = ((($141)) + 72|0);
       HEAP32[$145>>2] = $133;
       if (!($138)) {
        $146 = ((($137)) + 4|0);
        $147 = HEAP32[$146>>2]|0;
        $148 = (((($147) + ($133<<2)|0)) + 3|0);
        HEAP8[$148>>0] = 0;
        $149 = ((($137)) + 8|0);
        $150 = HEAP32[$149>>2]|0;
        $151 = (($150) + 1)|0;
        $152 = ($151|0)==(0);
        $$$i = $152 ? 1 : $151;
        HEAP32[$149>>2] = $$$i;
       }
       $153 = ($144|0)==($143|0);
       if ($153) {
        break;
       }
       _SDL_InvalidateMap($141);
       break;
      }
     }
    }
    $154 = HEAP32[$30>>2]|0;
    $155 = ($154|0)==(0);
    if (!($155)) {
     break;
    }
    $$pre = HEAP32[$$pre28>>2]|0;
    $159 = $$pre;
   }
   $156 = HEAP32[$52>>2]|0;
   $157 = ((($156)) + 72|0);
   $158 = HEAP32[$157>>2]|0;
   _SDL_GetRGBA($158,$159,$keyR,$keyG,$keyB,$keyA);
   $160 = ((($32)) + 4|0);
   $161 = HEAP32[$160>>2]|0;
   $162 = HEAP8[$keyR>>0]|0;
   $163 = HEAP8[$keyG>>0]|0;
   $164 = HEAP8[$keyB>>0]|0;
   $165 = HEAP8[$keyA>>0]|0;
   $166 = (_SDL_MapRGBA($161,$162,$163,$164,$165)|0);
   $167 = HEAP32[$160>>2]|0;
   $168 = ((($167)) + 4|0);
   $169 = HEAP32[$168>>2]|0;
   $170 = ($169|0)==(0|0);
   if ($170) {
    label = 39;
   } else {
    $171 = HEAP32[$169>>2]|0;
    $172 = ($171>>>0)>($166>>>0);
    if ($172) {
     label = 39;
    } else {
     HEAP32[$vararg_buffer3>>2] = 604299;
     (_SDL_SetError(611267,$vararg_buffer3)|0);
    }
   }
   if ((label|0) == 39) {
    $173 = HEAP32[$99>>2]|0;
    $174 = ((($173)) + 68|0);
    $175 = HEAP32[$174>>2]|0;
    $176 = $175 | 256;
    HEAP32[$174>>2] = $176;
    $177 = ((($173)) + 72|0);
    HEAP32[$177>>2] = $166;
    if (!($170)) {
     $178 = ((($169)) + 4|0);
     $179 = HEAP32[$178>>2]|0;
     $180 = (((($179) + ($166<<2)|0)) + 3|0);
     HEAP8[$180>>0] = 0;
     $181 = ((($169)) + 8|0);
     $182 = HEAP32[$181>>2]|0;
     $183 = (($182) + 1)|0;
     $184 = ($183|0)==(0);
     $$$i11 = $184 ? 1 : $183;
     HEAP32[$181>>2] = $$$i11;
    }
    $185 = ($176|0)==($175|0);
    if (!($185)) {
     _SDL_InvalidateMap($173);
    }
   }
   $186 = HEAP32[$99>>2]|0;
   $187 = ((($186)) + 68|0);
   $188 = HEAP32[$187>>2]|0;
   $189 = $188 & 256;
   $190 = ($189|0)==(0);
   if (!($190)) {
    $191 = HEAP32[$160>>2]|0;
    $192 = ((($191)) + 24|0);
    $193 = HEAP32[$192>>2]|0;
    $194 = ($193|0)==(0);
    if (!($194)) {
     $195 = ((($32)) + 28|0);
     $196 = HEAP32[$195>>2]|0;
     $197 = ($196|0)==(0);
     if ($197) {
      $198 = HEAP32[$32>>2]|0;
      $199 = $198 & 2;
      $200 = ($199|0)==(0);
      if ($200) {
       $204 = 0;$206 = $191;
      } else {
       _SDL_UnRLESurface($32,1);
       $201 = HEAP32[$32>>2]|0;
       $202 = $201 | 2;
       HEAP32[$32>>2] = $202;
       $$pre$i$i = HEAP32[$195>>2]|0;
       $$pre$i14 = HEAP32[$160>>2]|0;
       $204 = $$pre$i$i;$206 = $$pre$i14;
      }
     } else {
      $204 = $196;$206 = $191;
     }
     $203 = (($204) + 1)|0;
     HEAP32[$195>>2] = $203;
     $205 = ((($206)) + 9|0);
     $207 = HEAP8[$205>>0]|0;
     $208 = $207&255;
     switch ($208|0) {
     case 2:  {
      $209 = HEAP32[$99>>2]|0;
      $210 = ((($209)) + 72|0);
      $211 = HEAP32[$210>>2]|0;
      $212 = ((($206)) + 24|0);
      $213 = HEAP32[$212>>2]|0;
      $214 = $213 & 65535;
      $215 = $214 ^ 65535;
      $216 = $215 & $211;
      $217 = ((($32)) + 12|0);
      $218 = HEAP32[$217>>2]|0;
      $219 = ($218|0)==(0);
      if ($219) {
       $267 = $203;
      } else {
       $220 = ((($32)) + 8|0);
       $221 = HEAP32[$220>>2]|0;
       $222 = ($221|0)==(0);
       $223 = ((($32)) + 16|0);
       $224 = HEAP32[$223>>2]|0;
       $225 = (($224|0) / 2)&-1;
       if ($222) {
        $267 = $203;
       } else {
        $226 = ((($32)) + 20|0);
        $227 = HEAP32[$226>>2]|0;
        $228 = $216&65535;
        $$in21$i = $218;$row$05$i = $227;
        while(1) {
         $$in16$i = $221;$spot$04$i = $row$05$i;
         while(1) {
          $229 = (($$in16$i) + -1)|0;
          $230 = HEAP16[$spot$04$i>>1]|0;
          $231 = $230&65535;
          $232 = $231 & $215;
          $233 = ($232|0)==($216|0);
          if ($233) {
           HEAP16[$spot$04$i>>1] = $228;
          }
          $234 = ((($spot$04$i)) + 2|0);
          $235 = ($229|0)==(0);
          if ($235) {
           break;
          } else {
           $$in16$i = $229;$spot$04$i = $234;
          }
         }
         $236 = (($$in21$i) + -1)|0;
         $237 = (($row$05$i) + ($225<<1)|0);
         $238 = ($236|0)==(0);
         if ($238) {
          $267 = $203;
          break;
         } else {
          $$in21$i = $236;$row$05$i = $237;
         }
        }
       }
      }
      break;
     }
     case 4:  {
      $239 = HEAP32[$99>>2]|0;
      $240 = ((($239)) + 72|0);
      $241 = HEAP32[$240>>2]|0;
      $242 = ((($206)) + 24|0);
      $243 = HEAP32[$242>>2]|0;
      $244 = $243 ^ -1;
      $245 = $241 & $244;
      $246 = ((($32)) + 12|0);
      $247 = HEAP32[$246>>2]|0;
      $248 = ($247|0)==(0);
      if ($248) {
       $267 = $203;
      } else {
       $249 = ((($32)) + 20|0);
       $250 = HEAP32[$249>>2]|0;
       $251 = ((($32)) + 8|0);
       $252 = ((($32)) + 16|0);
       $$in$i = $247;$row1$012$i = $250;
       while(1) {
        $253 = (($$in$i) + -1)|0;
        $254 = HEAP32[$251>>2]|0;
        $255 = ($254|0)==(0);
        if (!($255)) {
         $$in15$i = $254;$spot2$08$i = $row1$012$i;
         while(1) {
          $256 = (($$in15$i) + -1)|0;
          $257 = HEAP32[$spot2$08$i>>2]|0;
          $258 = $257 & $244;
          $259 = ($258|0)==($245|0);
          if ($259) {
           HEAP32[$spot2$08$i>>2] = $245;
          }
          $260 = ((($spot2$08$i)) + 4|0);
          $261 = ($256|0)==(0);
          if ($261) {
           break;
          } else {
           $$in15$i = $256;$spot2$08$i = $260;
          }
         }
        }
        $262 = HEAP32[$252>>2]|0;
        $263 = (($262|0) / 4)&-1;
        $264 = (($row1$012$i) + ($263<<2)|0);
        $265 = ($253|0)==(0);
        if ($265) {
         break;
        } else {
         $$in$i = $253;$row1$012$i = $264;
        }
       }
       $$pr$pre$i = HEAP32[$195>>2]|0;
       $267 = $$pr$pre$i;
      }
      break;
     }
     default: {
      $267 = $203;
     }
     }
     $266 = ($267|0)==(0);
     if ($266) {
      $275 = $206;
     } else {
      $268 = (($267) + -1)|0;
      HEAP32[$195>>2] = $268;
      $269 = ($267|0)>(1);
      if ($269) {
       $275 = $206;
      } else {
       $270 = HEAP32[$32>>2]|0;
       $271 = $270 & 2;
       $272 = ($271|0)==(0);
       if ($272) {
        $275 = $206;
       } else {
        $273 = $270 & -3;
        HEAP32[$32>>2] = $273;
        (_SDL_RLESurface($32)|0);
        $$pre20$i = HEAP32[$160>>2]|0;
        $275 = $$pre20$i;
       }
      }
     }
     $274 = ((($275)) + 4|0);
     $276 = HEAP32[$274>>2]|0;
     $277 = ($276|0)==(0|0);
     do {
      if ($277) {
       $278 = HEAP32[$99>>2]|0;
       $279 = ((($278)) + 68|0);
       $280 = HEAP32[$279>>2]|0;
       $296 = $280;$297 = $279;$299 = $278;
       label = 74;
      } else {
       $281 = HEAP32[$276>>2]|0;
       $282 = ($281|0)==(0);
       if ($282) {
        HEAP32[$vararg_buffer6>>2] = 604299;
        (_SDL_SetError(611267,$vararg_buffer6)|0);
        break;
       } else {
        $283 = HEAP32[$99>>2]|0;
        $284 = ((($283)) + 68|0);
        $285 = HEAP32[$284>>2]|0;
        $286 = ((($283)) + 72|0);
        $287 = HEAP32[$286>>2]|0;
        $288 = ((($276)) + 4|0);
        $289 = HEAP32[$288>>2]|0;
        $290 = (((($289) + ($287<<2)|0)) + 3|0);
        HEAP8[$290>>0] = -1;
        $291 = ((($276)) + 8|0);
        $292 = HEAP32[$291>>2]|0;
        $293 = (($292) + 1)|0;
        $294 = ($293|0)==(0);
        $$1$i$i = $294 ? 1 : $293;
        HEAP32[$291>>2] = $$1$i$i;
        $296 = $285;$297 = $284;$299 = $283;
        label = 74;
        break;
       }
      }
     } while(0);
     if ((label|0) == 74) {
      $295 = $296 & -257;
      HEAP32[$297>>2] = $295;
      $298 = ($295|0)==($296|0);
      if (!($298)) {
       _SDL_InvalidateMap($299);
      }
     }
     $300 = HEAP32[$99>>2]|0;
     $301 = ((($300)) + 68|0);
     $302 = HEAP32[$301>>2]|0;
     $303 = $302 & -113;
     $304 = $303 | 16;
     HEAP32[$301>>2] = $304;
     $305 = ($304|0)==($302|0);
     if (!($305)) {
      _SDL_InvalidateMap($300);
     }
    }
   }
  }
 } while(0);
 $306 = ((($surface)) + 36|0);
 HEAP32[$full_rect$i>>2] = 0;
 $307 = ((($full_rect$i)) + 4|0);
 HEAP32[$307>>2] = 0;
 $308 = ((($32)) + 8|0);
 $309 = HEAP32[$308>>2]|0;
 $310 = ((($full_rect$i)) + 8|0);
 HEAP32[$310>>2] = $309;
 $311 = ((($32)) + 12|0);
 $312 = HEAP32[$311>>2]|0;
 $313 = ((($full_rect$i)) + 12|0);
 HEAP32[$313>>2] = $312;
 $314 = ((($32)) + 36|0);
 (_SDL_IntersectRect($306,$full_rect$i,$314)|0);
 $315 = HEAP32[$$pre28>>2]|0;
 $316 = ((($315)) + 24|0);
 $317 = HEAP32[$316>>2]|0;
 $318 = ($317|0)==(0);
 if ($318) {
  $$old = $55 & 258;
  $$old4 = ($$old|0)==(0);
  if (!($$old4)) {
   label = 82;
  }
 } else {
  $319 = HEAP32[$30>>2]|0;
  $320 = $55 & 258;
  $321 = $319 | $320;
  $322 = ($321|0)==(0);
  if (!($322)) {
   label = 82;
  }
 }
 if ((label|0) == 82) {
  $323 = HEAP32[$99>>2]|0;
  $324 = ((($323)) + 68|0);
  $325 = HEAP32[$324>>2]|0;
  $326 = $325 & -113;
  $327 = $326 | 16;
  HEAP32[$324>>2] = $327;
  $328 = ($327|0)==($325|0);
  if (!($328)) {
   _SDL_InvalidateMap($323);
  }
 }
 $329 = $55 & 4096;
 $330 = $flags & 2;
 $331 = $329 | $330;
 $332 = ($331|0)==(0);
 if ($332) {
  $$1 = $32;
  STACKTOP = sp;return ($$1|0);
 }
 $333 = HEAP32[$99>>2]|0;
 $334 = ((($333)) + 68|0);
 $335 = HEAP32[$334>>2]|0;
 $336 = $335 | 4096;
 HEAP32[$334>>2] = $336;
 $337 = ($336|0)==($335|0);
 if ($337) {
  $$1 = $32;
  STACKTOP = sp;return ($$1|0);
 }
 _SDL_InvalidateMap($333);
 $$1 = $32;
 STACKTOP = sp;return ($$1|0);
}
function _SDLgfx_rotozoomSurfaceSizeTrig($width,$height,$angle,$dstwidth,$dstheight,$cangle,$sangle) {
 $width = $width|0;
 $height = $height|0;
 $angle = +$angle;
 $dstwidth = $dstwidth|0;
 $dstheight = $dstheight|0;
 $cangle = $cangle|0;
 $sangle = $sangle|0;
 var $$sink = 0.0, $$sink1 = 0.0, $$sink2 = 0.0, $$sink3 = 0.0, $$sink4 = 0.0, $$sink5 = 0.0, $$sink6 = 0.0, $$sink7 = 0.0, $0 = 0.0, $1 = 0.0, $10 = 0.0, $100 = 0.0, $101 = 0.0, $102 = 0, $103 = 0.0, $104 = 0.0, $105 = 0, $106 = 0.0, $107 = 0.0, $108 = 0;
 var $109 = 0.0, $11 = 0.0, $110 = 0.0, $111 = 0.0, $112 = 0.0, $113 = 0.0, $114 = 0.0, $115 = 0, $116 = 0, $117 = 0.0, $118 = 0.0, $119 = 0, $12 = 0.0, $120 = 0.0, $121 = 0.0, $122 = 0, $123 = 0.0, $124 = 0.0, $125 = 0, $126 = 0.0;
 var $127 = 0.0, $128 = 0.0, $129 = 0.0, $13 = 0.0, $130 = 0, $131 = 0.0, $132 = 0.0, $133 = 0.0, $134 = 0, $135 = 0.0, $136 = 0.0, $137 = 0, $138 = 0.0, $139 = 0.0, $14 = 0.0, $140 = 0, $141 = 0.0, $142 = 0.0, $143 = 0.0, $144 = 0.0;
 var $145 = 0.0, $146 = 0.0, $147 = 0, $148 = 0, $149 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0.0, $20 = 0.0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0.0;
 var $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0.0, $4 = 0.0, $40 = 0, $41 = 0.0, $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0.0;
 var $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0, $52 = 0.0, $53 = 0.0, $54 = 0, $55 = 0.0, $56 = 0.0, $57 = 0, $58 = 0.0, $59 = 0.0, $6 = 0.0, $60 = 0.0, $61 = 0.0, $62 = 0, $63 = 0.0, $64 = 0.0;
 var $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0.0, $69 = 0, $7 = 0.0, $70 = 0.0, $71 = 0.0, $72 = 0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0.0, $77 = 0.0, $78 = 0.0, $79 = 0, $8 = 0.0, $80 = 0.0, $81 = 0.0, $82 = 0.0;
 var $83 = 0.0, $84 = 0, $85 = 0.0, $86 = 0.0, $87 = 0.0, $88 = 0.0, $89 = 0, $9 = 0.0, $90 = 0.0, $91 = 0.0, $92 = 0, $93 = 0.0, $94 = 0.0, $95 = 0.0, $96 = 0.0, $97 = 0.0, $98 = 0, $99 = 0.0, $phitmp = 0, $phitmp1 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $angle * 0.017453292519943295;
 $1 = (+_SDL_sin($0));
 HEAPF64[$sangle>>3] = $1;
 $2 = (+_SDL_cos($0));
 HEAPF64[$cangle>>3] = $2;
 $3 = (($width|0) / 2)&-1;
 $4 = (+($3|0));
 $5 = (($height|0) / 2)&-1;
 $6 = (+($5|0));
 $7 = $4 * $2;
 $8 = $6 * $2;
 $9 = +HEAPF64[$sangle>>3];
 $10 = $4 * $9;
 $11 = $6 * $9;
 $12 = $7 + $11;
 $13 = (+_SDL_fabs($12));
 $14 = $7 - $11;
 $15 = (+_SDL_fabs($14));
 $16 = $13 > $15;
 $$sink = $16 ? $12 : $14;
 $17 = (+_SDL_fabs($$sink));
 $18 = -$7;
 $19 = $11 - $7;
 $20 = (+_SDL_fabs($19));
 $21 = $17 > $20;
 do {
  if ($21) {
   $22 = (+_SDL_fabs($12));
   $23 = (+_SDL_fabs($14));
   $24 = $22 > $23;
   if ($24) {
    $25 = (+_SDL_fabs($12));
    $31 = $25;
    break;
   } else {
    $26 = (+_SDL_fabs($14));
    $31 = $26;
    break;
   }
  } else {
   $27 = (+_SDL_fabs($19));
   $31 = $27;
  }
 } while(0);
 $28 = $18 - $11;
 $29 = (+_SDL_fabs($28));
 $30 = $31 > $29;
 do {
  if ($30) {
   $32 = (+_SDL_fabs($12));
   $33 = (+_SDL_fabs($14));
   $34 = $32 > $33;
   $$sink1 = $34 ? $12 : $14;
   $35 = (+_SDL_fabs($$sink1));
   $36 = (+_SDL_fabs($19));
   $37 = $35 > $36;
   if (!($37)) {
    $43 = (+_SDL_fabs($19));
    $45 = $43;
    break;
   }
   $38 = (+_SDL_fabs($12));
   $39 = (+_SDL_fabs($14));
   $40 = $38 > $39;
   if ($40) {
    $41 = (+_SDL_fabs($12));
    $45 = $41;
    break;
   } else {
    $42 = (+_SDL_fabs($14));
    $45 = $42;
    break;
   }
  } else {
   $44 = (+_SDL_fabs($28));
   $45 = $44;
  }
 } while(0);
 $46 = (+_SDL_ceil($45));
 $47 = (~~(($46)));
 $48 = ($47|0)>(1);
 if ($48) {
  $49 = (+_SDL_fabs($12));
  $50 = (+_SDL_fabs($14));
  $51 = $49 > $50;
  $$sink2 = $51 ? $12 : $14;
  $52 = (+_SDL_fabs($$sink2));
  $53 = (+_SDL_fabs($19));
  $54 = $52 > $53;
  do {
   if ($54) {
    $55 = (+_SDL_fabs($12));
    $56 = (+_SDL_fabs($14));
    $57 = $55 > $56;
    if ($57) {
     $58 = (+_SDL_fabs($12));
     $63 = $58;
     break;
    } else {
     $59 = (+_SDL_fabs($14));
     $63 = $59;
     break;
    }
   } else {
    $60 = (+_SDL_fabs($19));
    $63 = $60;
   }
  } while(0);
  $61 = (+_SDL_fabs($28));
  $62 = $63 > $61;
  do {
   if ($62) {
    $64 = (+_SDL_fabs($12));
    $65 = (+_SDL_fabs($14));
    $66 = $64 > $65;
    $$sink3 = $66 ? $12 : $14;
    $67 = (+_SDL_fabs($$sink3));
    $68 = (+_SDL_fabs($19));
    $69 = $67 > $68;
    if (!($69)) {
     $75 = (+_SDL_fabs($19));
     $77 = $75;
     break;
    }
    $70 = (+_SDL_fabs($12));
    $71 = (+_SDL_fabs($14));
    $72 = $70 > $71;
    if ($72) {
     $73 = (+_SDL_fabs($12));
     $77 = $73;
     break;
    } else {
     $74 = (+_SDL_fabs($14));
     $77 = $74;
     break;
    }
   } else {
    $76 = (+_SDL_fabs($28));
    $77 = $76;
   }
  } while(0);
  $78 = (+_SDL_ceil($77));
  $79 = (~~(($78)));
  $phitmp = $79 << 1;
  $148 = $phitmp;
 } else {
  $148 = 2;
 }
 $80 = $8 + $10;
 $81 = (+_SDL_fabs($80));
 $82 = $10 - $8;
 $83 = (+_SDL_fabs($82));
 $84 = $81 > $83;
 $$sink4 = $84 ? $80 : $82;
 $85 = (+_SDL_fabs($$sink4));
 $86 = -$10;
 $87 = $8 - $10;
 $88 = (+_SDL_fabs($87));
 $89 = $85 > $88;
 do {
  if ($89) {
   $90 = (+_SDL_fabs($80));
   $91 = (+_SDL_fabs($82));
   $92 = $90 > $91;
   if ($92) {
    $93 = (+_SDL_fabs($80));
    $99 = $93;
    break;
   } else {
    $94 = (+_SDL_fabs($82));
    $99 = $94;
    break;
   }
  } else {
   $95 = (+_SDL_fabs($87));
   $99 = $95;
  }
 } while(0);
 $96 = $86 - $8;
 $97 = (+_SDL_fabs($96));
 $98 = $99 > $97;
 do {
  if ($98) {
   $100 = (+_SDL_fabs($80));
   $101 = (+_SDL_fabs($82));
   $102 = $100 > $101;
   $$sink5 = $102 ? $80 : $82;
   $103 = (+_SDL_fabs($$sink5));
   $104 = (+_SDL_fabs($87));
   $105 = $103 > $104;
   if (!($105)) {
    $111 = (+_SDL_fabs($87));
    $113 = $111;
    break;
   }
   $106 = (+_SDL_fabs($80));
   $107 = (+_SDL_fabs($82));
   $108 = $106 > $107;
   if ($108) {
    $109 = (+_SDL_fabs($80));
    $113 = $109;
    break;
   } else {
    $110 = (+_SDL_fabs($82));
    $113 = $110;
    break;
   }
  } else {
   $112 = (+_SDL_fabs($96));
   $113 = $112;
  }
 } while(0);
 $114 = (+_SDL_ceil($113));
 $115 = (~~(($114)));
 $116 = ($115|0)>(1);
 if (!($116)) {
  $149 = 2;
  HEAP32[$dstwidth>>2] = $148;
  HEAP32[$dstheight>>2] = $149;
  return;
 }
 $117 = (+_SDL_fabs($80));
 $118 = (+_SDL_fabs($82));
 $119 = $117 > $118;
 $$sink6 = $119 ? $80 : $82;
 $120 = (+_SDL_fabs($$sink6));
 $121 = (+_SDL_fabs($87));
 $122 = $120 > $121;
 do {
  if ($122) {
   $123 = (+_SDL_fabs($80));
   $124 = (+_SDL_fabs($82));
   $125 = $123 > $124;
   if ($125) {
    $126 = (+_SDL_fabs($80));
    $131 = $126;
    break;
   } else {
    $127 = (+_SDL_fabs($82));
    $131 = $127;
    break;
   }
  } else {
   $128 = (+_SDL_fabs($87));
   $131 = $128;
  }
 } while(0);
 $129 = (+_SDL_fabs($96));
 $130 = $131 > $129;
 do {
  if ($130) {
   $132 = (+_SDL_fabs($80));
   $133 = (+_SDL_fabs($82));
   $134 = $132 > $133;
   $$sink7 = $134 ? $80 : $82;
   $135 = (+_SDL_fabs($$sink7));
   $136 = (+_SDL_fabs($87));
   $137 = $135 > $136;
   if (!($137)) {
    $143 = (+_SDL_fabs($87));
    $145 = $143;
    break;
   }
   $138 = (+_SDL_fabs($80));
   $139 = (+_SDL_fabs($82));
   $140 = $138 > $139;
   if ($140) {
    $141 = (+_SDL_fabs($80));
    $145 = $141;
    break;
   } else {
    $142 = (+_SDL_fabs($82));
    $145 = $142;
    break;
   }
  } else {
   $144 = (+_SDL_fabs($96));
   $145 = $144;
  }
 } while(0);
 $146 = (+_SDL_ceil($145));
 $147 = (~~(($146)));
 $phitmp1 = $147 << 1;
 $149 = $phitmp1;
 HEAP32[$dstwidth>>2] = $148;
 HEAP32[$dstheight>>2] = $149;
 return;
}
function _SDL_ceil($x) {
 $x = +$x;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+Math_ceil((+$x)));
 return (+$0);
}
function _SDLgfx_rotateSurface($src,$angle,$centerx,$centery,$smooth,$flipx,$flipy,$dstwidth,$dstheight,$cangle,$sangle) {
 $src = $src|0;
 $angle = +$angle;
 $centerx = $centerx|0;
 $centery = $centery|0;
 $smooth = $smooth|0;
 $flipx = $flipx|0;
 $flipy = $flipy|0;
 $dstwidth = $dstwidth|0;
 $dstheight = $dstheight|0;
 $cangle = +$cangle;
 $sangle = +$sangle;
 var $$$i = 0, $$0 = 0, $$1$$9$i = 0, $$1$i = 0, $$10$$2$i = 0, $$10$i = 0, $$11$$3$i = 0, $$11$i = 0, $$12$$4$i = 0, $$12$i = 0, $$13$$5$i = 0, $$13$i = 0, $$14$$6$i = 0, $$14$i = 0, $$15$$7$i = 0, $$15$i = 0, $$16$$8$i = 0, $$16$i = 0, $$2$$10$i = 0, $$2$i = 0;
 var $$3$$11$i = 0, $$3$i = 0, $$4$$12$i = 0, $$4$i = 0, $$5$$13$i = 0, $$5$i = 0, $$6$$14$i = 0, $$6$i = 0, $$7$$15$i = 0, $$7$i = 0, $$8$$16$i = 0, $$8$i = 0, $$9$$1$i = 0, $$9$i = 0, $$lcssa160 = 0, $$lcssa161 = 0, $$lcssa43 = 0, $$lcssa44 = 0, $$neg$i$neg = 0, $$neg1$i$neg = 0;
 var $$phi$trans$insert = 0, $$pre = 0, $$pre$i = 0, $$pre$i3 = 0, $$pre$phiZ2D = 0, $$pre26$i = 0, $$pre27$i = 0, $$pre28$i = 0, $$pre29$i = 0, $$pre52$i = 0, $$pre53$i = 0, $$pre54$i = 0, $$pre55$i = 0, $$pre56$i = 0, $$pre74 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0.0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0.0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0.0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0.0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0;
 var $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0;
 var $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0;
 var $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0;
 var $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0;
 var $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0;
 var $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0;
 var $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0;
 var $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0;
 var $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0;
 var $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0;
 var $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0;
 var $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0;
 var $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0;
 var $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0.0, $478 = 0.0, $479 = 0, $48 = 0, $480 = 0;
 var $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0.0, $56 = 0;
 var $57 = 0.0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $alphaMod = 0, $b = 0, $blendMode = 0, $cb = 0, $cg = 0, $colorKeyAvailable$0 = 0, $cr = 0, $dy$0$i = 0, $g = 0, $i$045 = 0, $is32bit$013 = 0, $key$i = 0, $key$i$i = 0;
 var $key$i1 = 0, $key$i2 = 0, $or$cond$i = 0, $pc$010$i = 0, $pc$036$i = 0, $pc$1$lcssa$i = 0, $pc$1$lcssa$i11 = 0, $pc$130$i = 0, $pc$17$i = 0, $pc$17$us$i = 0, $pc$17$us$us$i = 0, $pc$17$us15$i = 0, $pc$225$i = 0, $pc$3$lcssa$i = 0, $pc$323$i = 0, $pc$323$us$i = 0, $pc$323$us$us$i = 0, $pc$323$us40$i = 0, $r = 0, $rz_dst$0 = 0;
 var $rz_src$015 = 0, $rz_src$016 = 0, $sdx$029$i = 0, $sdx$05$i = 0, $sdx$05$us$i = 0, $sdx$05$us$us$i = 0, $sdx$05$us17$i = 0, $sdx$122$i = 0, $sdx$122$us$i = 0, $sdx$122$us$us$i = 0, $sdx$122$us41$i = 0, $sdy$028$i = 0, $sdy$06$i = 0, $sdy$06$us$i = 0, $sdy$06$us$us$i = 0, $sdy$06$us16$i = 0, $sdy$121$i = 0, $sdy$121$us$i = 0, $sdy$121$us$us$i = 0, $sdy$121$us42$i = 0;
 var $tmp = 0, $x$031$i = 0, $x$08$i = 0, $x$08$us$i = 0, $x$08$us$us$i = 0, $x$08$us14$i = 0, $x$124$i = 0, $x$124$us$i = 0, $x$124$us$us$i = 0, $x$124$us39$i = 0, $y$037$i = 0, $y$09$i = 0, $y$126$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $key$i$i = sp + 16|0;
 $key$i2 = sp + 12|0;
 $key$i1 = sp + 8|0;
 $key$i = sp + 4|0;
 $r = sp + 26|0;
 $g = sp + 25|0;
 $b = sp + 24|0;
 $blendMode = sp;
 $alphaMod = sp + 23|0;
 $cr = sp + 22|0;
 $cg = sp + 21|0;
 $cb = sp + 20|0;
 $0 = ($src|0)==(0|0);
 if ($0) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = HEAP32[$src>>2]|0;
 $2 = $1 & 1;
 $3 = ($2|0)==(0);
 if ($3) {
  $$pre74 = ((($src)) + 4|0);
  $$pre$phiZ2D = $$pre74;$colorKeyAvailable$0 = 0;
 } else {
  HEAP32[$key$i>>2] = 0;
  (_SDL_GetColorKey($src,$key$i)|0);
  $4 = HEAP32[$key$i>>2]|0;
  $5 = ((($src)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  _SDL_GetRGB($4,$6,$r,$g,$b);
  $$pre$phiZ2D = $5;$colorKeyAvailable$0 = 1;
 }
 $7 = HEAP32[$$pre$phiZ2D>>2]|0;
 $8 = ((($7)) + 8|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = ($9<<24>>24)==(32);
 switch ($9<<24>>24) {
 case 8: case 32:  {
  $17 = $sangle * 65536.0;
  $18 = $cangle * 65536.0;
  $19 = (($dstheight) + 2)|0;
  if ($10) {
   $21 = $7;$29 = $19;$477 = $18;$478 = $17;$rz_src$015 = $src;
   label = 8;
  } else {
   $31 = (_SDL_CreateRGBSurface(0,$dstwidth,$19,8,0,0,0,0)|0);
   $55 = $17;$57 = $18;$is32bit$013 = 0;$rz_dst$0 = $31;$rz_src$016 = $src;
  }
  break;
 }
 default: {
  $11 = (_SDL_MasksToPixelFormatEnum(32,255,65280,16711680,-16777216)|0);
  $12 = HEAP32[$src>>2]|0;
  $13 = (_SDL_ConvertSurfaceFormat($src,$11,$12)|0);
  $14 = $sangle * 65536.0;
  $15 = $cangle * 65536.0;
  $16 = (($dstheight) + 2)|0;
  $$phi$trans$insert = ((($13)) + 4|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $21 = $$pre;$29 = $16;$477 = $15;$478 = $14;$rz_src$015 = $13;
  label = 8;
 }
 }
 if ((label|0) == 8) {
  $20 = ((($21)) + 12|0);
  $22 = HEAP32[$20>>2]|0;
  $23 = ((($21)) + 16|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = ((($21)) + 20|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = ((($21)) + 24|0);
  $28 = HEAP32[$27>>2]|0;
  $30 = (_SDL_CreateRGBSurface(0,$dstwidth,$29,32,$22,$24,$26,$28)|0);
  $55 = $478;$57 = $477;$is32bit$013 = 1;$rz_dst$0 = $30;$rz_src$016 = $rz_src$015;
 }
 $32 = ($rz_dst$0|0)==(0|0);
 if ($32) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $33 = ((($rz_dst$0)) + 12|0);
 HEAP32[$33>>2] = $dstheight;
 $34 = ($colorKeyAvailable$0|0)==(1);
 if ($34) {
  $35 = ((($rz_dst$0)) + 4|0);
  $36 = HEAP32[$35>>2]|0;
  $37 = HEAP8[$r>>0]|0;
  $38 = HEAP8[$g>>0]|0;
  $39 = HEAP8[$b>>0]|0;
  $40 = (_SDL_MapRGB($36,$37,$38,$39)|0);
  (_SDL_FillRect($rz_dst$0,0,$40)|0);
 }
 $41 = HEAP32[$rz_src$016>>2]|0;
 $42 = $41 & 2;
 $43 = ($42|0)==(0);
 if (!($43)) {
  (_SDL_LockSurface($rz_src$016)|0);
 }
 if ($is32bit$013) {
  $54 = (~~(($55)));
  $56 = (~~(($57)));
  $58 = ((($rz_src$016)) + 8|0);
  $59 = HEAP32[$58>>2]|0;
  $60 = ((($rz_dst$0)) + 8|0);
  $61 = HEAP32[$60>>2]|0;
  $62 = (($59) - ($61))|0;
  $63 = $62 << 15;
  $64 = ((($rz_src$016)) + 12|0);
  $65 = HEAP32[$64>>2]|0;
  $66 = HEAP32[$33>>2]|0;
  $67 = (($65) - ($66))|0;
  $68 = $67 << 15;
  $69 = (65536 - ($56))|0;
  $70 = Math_imul($69, $centerx)|0;
  $71 = $centery << 16;
  $72 = Math_imul($54, $centerx)|0;
  $73 = (($71) - ($72))|0;
  $74 = (($59) + -1)|0;
  $75 = (($65) + -1)|0;
  $76 = ((($rz_dst$0)) + 20|0);
  $77 = HEAP32[$76>>2]|0;
  $78 = ((($rz_dst$0)) + 16|0);
  $79 = HEAP32[$78>>2]|0;
  $80 = $61 << 2;
  $81 = (($79) - ($80))|0;
  $82 = ($smooth|0)==(0);
  $83 = ($66|0)>(0);
  if ($82) {
   if ($83) {
    $90 = (($63) + ($70))|0;
    $91 = (($68) + ($73))|0;
    $92 = ($flipx|0)==(0);
    $93 = ($flipy|0)==(0);
    $94 = ((($rz_src$016)) + 20|0);
    $95 = ((($rz_src$016)) + 16|0);
    $243 = $61;$pc$225$i = $77;$y$126$i = 0;
    while(1) {
     $237 = (($centery) - ($y$126$i))|0;
     $238 = Math_imul($237, $54)|0;
     $239 = (($90) + ($238))|0;
     $240 = Math_imul($237, $56)|0;
     $241 = (($91) - ($240))|0;
     $242 = ($243|0)>(0);
     if ($242) {
      if ($92) {
       if ($93) {
        $483 = $243;$pc$323$us$us$i = $pc$225$i;$sdx$122$us$us$i = $239;$sdy$121$us$us$i = $241;$x$124$us$us$i = 0;
        while(1) {
         $244 = $sdx$122$us$us$i >> 16;
         $245 = $sdy$121$us$us$i >> 16;
         $246 = HEAP32[$58>>2]|0;
         $247 = ($244>>>0)<($246>>>0);
         do {
          if ($247) {
           $248 = HEAP32[$64>>2]|0;
           $249 = ($245>>>0)<($248>>>0);
           if (!($249)) {
            $261 = $483;
            break;
           }
           $250 = HEAP32[$94>>2]|0;
           $251 = HEAP32[$95>>2]|0;
           $252 = Math_imul($251, $245)|0;
           $253 = (($250) + ($252)|0);
           $254 = (($253) + ($244<<2)|0);
           $255 = HEAPU8[$254>>0]|(HEAPU8[$254+1>>0]<<8)|(HEAPU8[$254+2>>0]<<16)|(HEAPU8[$254+3>>0]<<24);
           HEAP8[$pc$323$us$us$i>>0]=$255&255;HEAP8[$pc$323$us$us$i+1>>0]=($255>>8)&255;HEAP8[$pc$323$us$us$i+2>>0]=($255>>16)&255;HEAP8[$pc$323$us$us$i+3>>0]=$255>>24;
           $$pre56$i = HEAP32[$60>>2]|0;
           $261 = $$pre56$i;
          } else {
           $261 = $483;
          }
         } while(0);
         $256 = (($sdx$122$us$us$i) + ($56))|0;
         $257 = (($sdy$121$us$us$i) + ($54))|0;
         $258 = ((($pc$323$us$us$i)) + 4|0);
         $259 = (($x$124$us$us$i) + 1)|0;
         $260 = ($259|0)<($261|0);
         if ($260) {
          $483 = $261;$pc$323$us$us$i = $258;$sdx$122$us$us$i = $256;$sdy$121$us$us$i = $257;$x$124$us$us$i = $259;
         } else {
          $482 = $261;$pc$3$lcssa$i = $258;
          break;
         }
        }
       } else {
        $484 = $243;$pc$323$us$i = $pc$225$i;$sdx$122$us$i = $239;$sdy$121$us$i = $241;$x$124$us$i = 0;
        while(1) {
         $262 = $sdx$122$us$i >> 16;
         $263 = $sdy$121$us$i >> 16;
         $264 = HEAP32[$58>>2]|0;
         $265 = ($262>>>0)<($264>>>0);
         do {
          if ($265) {
           $266 = HEAP32[$64>>2]|0;
           $267 = ($263>>>0)<($266>>>0);
           if (!($267)) {
            $280 = $484;
            break;
           }
           $268 = (($75) - ($263))|0;
           $269 = HEAP32[$94>>2]|0;
           $270 = HEAP32[$95>>2]|0;
           $271 = Math_imul($270, $268)|0;
           $272 = (($269) + ($271)|0);
           $273 = (($272) + ($262<<2)|0);
           $274 = HEAPU8[$273>>0]|(HEAPU8[$273+1>>0]<<8)|(HEAPU8[$273+2>>0]<<16)|(HEAPU8[$273+3>>0]<<24);
           HEAP8[$pc$323$us$i>>0]=$274&255;HEAP8[$pc$323$us$i+1>>0]=($274>>8)&255;HEAP8[$pc$323$us$i+2>>0]=($274>>16)&255;HEAP8[$pc$323$us$i+3>>0]=$274>>24;
           $$pre55$i = HEAP32[$60>>2]|0;
           $280 = $$pre55$i;
          } else {
           $280 = $484;
          }
         } while(0);
         $275 = (($sdx$122$us$i) + ($56))|0;
         $276 = (($sdy$121$us$i) + ($54))|0;
         $277 = ((($pc$323$us$i)) + 4|0);
         $278 = (($x$124$us$i) + 1)|0;
         $279 = ($278|0)<($280|0);
         if ($279) {
          $484 = $280;$pc$323$us$i = $277;$sdx$122$us$i = $275;$sdy$121$us$i = $276;$x$124$us$i = $278;
         } else {
          $482 = $280;$pc$3$lcssa$i = $277;
          break;
         }
        }
       }
      } else {
       if ($93) {
        $485 = $243;$pc$323$us40$i = $pc$225$i;$sdx$122$us41$i = $239;$sdy$121$us42$i = $241;$x$124$us39$i = 0;
        while(1) {
         $281 = $sdx$122$us41$i >> 16;
         $282 = $sdy$121$us42$i >> 16;
         $283 = HEAP32[$58>>2]|0;
         $284 = ($281>>>0)<($283>>>0);
         do {
          if ($284) {
           $285 = HEAP32[$64>>2]|0;
           $286 = ($282>>>0)<($285>>>0);
           if (!($286)) {
            $299 = $485;
            break;
           }
           $287 = (($74) - ($281))|0;
           $288 = HEAP32[$94>>2]|0;
           $289 = HEAP32[$95>>2]|0;
           $290 = Math_imul($289, $282)|0;
           $291 = (($288) + ($290)|0);
           $292 = (($291) + ($287<<2)|0);
           $293 = HEAPU8[$292>>0]|(HEAPU8[$292+1>>0]<<8)|(HEAPU8[$292+2>>0]<<16)|(HEAPU8[$292+3>>0]<<24);
           HEAP8[$pc$323$us40$i>>0]=$293&255;HEAP8[$pc$323$us40$i+1>>0]=($293>>8)&255;HEAP8[$pc$323$us40$i+2>>0]=($293>>16)&255;HEAP8[$pc$323$us40$i+3>>0]=$293>>24;
           $$pre54$i = HEAP32[$60>>2]|0;
           $299 = $$pre54$i;
          } else {
           $299 = $485;
          }
         } while(0);
         $294 = (($sdx$122$us41$i) + ($56))|0;
         $295 = (($sdy$121$us42$i) + ($54))|0;
         $296 = ((($pc$323$us40$i)) + 4|0);
         $297 = (($x$124$us39$i) + 1)|0;
         $298 = ($297|0)<($299|0);
         if ($298) {
          $485 = $299;$pc$323$us40$i = $296;$sdx$122$us41$i = $294;$sdy$121$us42$i = $295;$x$124$us39$i = $297;
         } else {
          $482 = $299;$pc$3$lcssa$i = $296;
          break;
         }
        }
       } else {
        $486 = $243;$pc$323$i = $pc$225$i;$sdx$122$i = $239;$sdy$121$i = $241;$x$124$i = 0;
        while(1) {
         $300 = $sdx$122$i >> 16;
         $301 = $sdy$121$i >> 16;
         $302 = HEAP32[$58>>2]|0;
         $303 = ($300>>>0)<($302>>>0);
         do {
          if ($303) {
           $304 = HEAP32[$64>>2]|0;
           $305 = ($301>>>0)<($304>>>0);
           if (!($305)) {
            $319 = $486;
            break;
           }
           $306 = (($74) - ($300))|0;
           $307 = (($75) - ($301))|0;
           $308 = HEAP32[$94>>2]|0;
           $309 = HEAP32[$95>>2]|0;
           $310 = Math_imul($309, $307)|0;
           $311 = (($308) + ($310)|0);
           $312 = (($311) + ($306<<2)|0);
           $313 = HEAPU8[$312>>0]|(HEAPU8[$312+1>>0]<<8)|(HEAPU8[$312+2>>0]<<16)|(HEAPU8[$312+3>>0]<<24);
           HEAP8[$pc$323$i>>0]=$313&255;HEAP8[$pc$323$i+1>>0]=($313>>8)&255;HEAP8[$pc$323$i+2>>0]=($313>>16)&255;HEAP8[$pc$323$i+3>>0]=$313>>24;
           $$pre53$i = HEAP32[$60>>2]|0;
           $319 = $$pre53$i;
          } else {
           $319 = $486;
          }
         } while(0);
         $314 = (($sdx$122$i) + ($56))|0;
         $315 = (($sdy$121$i) + ($54))|0;
         $316 = ((($pc$323$i)) + 4|0);
         $317 = (($x$124$i) + 1)|0;
         $318 = ($317|0)<($319|0);
         if ($318) {
          $486 = $319;$pc$323$i = $316;$sdx$122$i = $314;$sdy$121$i = $315;$x$124$i = $317;
         } else {
          $482 = $319;$pc$3$lcssa$i = $316;
          break;
         }
        }
       }
      }
     } else {
      $482 = $243;$pc$3$lcssa$i = $pc$225$i;
     }
     $320 = (($pc$3$lcssa$i) + ($81)|0);
     $321 = (($y$126$i) + 1)|0;
     $322 = HEAP32[$33>>2]|0;
     $323 = ($321|0)<($322|0);
     if ($323) {
      $243 = $482;$pc$225$i = $320;$y$126$i = $321;
     } else {
      break;
     }
    }
   }
  } else {
   if ($83) {
    $84 = (($63) + ($70))|0;
    $85 = (($68) + ($73))|0;
    $86 = ($flipx|0)!=(0);
    $87 = ($flipy|0)!=(0);
    $88 = ((($rz_src$016)) + 20|0);
    $89 = ((($rz_src$016)) + 16|0);
    $479 = $66;$98 = $61;$pc$036$i = $77;$y$037$i = 0;
    while(1) {
     $96 = (($centery) - ($y$037$i))|0;
     $97 = ($98|0)>(0);
     if ($97) {
      $99 = Math_imul($96, $56)|0;
      $100 = (($85) - ($99))|0;
      $101 = Math_imul($96, $54)|0;
      $102 = (($84) + ($101))|0;
      $481 = $98;$pc$130$i = $pc$036$i;$sdx$029$i = $102;$sdy$028$i = $100;$x$031$i = 0;
      while(1) {
       $103 = $sdx$029$i >> 16;
       $104 = $sdy$028$i >> 16;
       $105 = (($74) - ($103))|0;
       $$$i = $86 ? $105 : $103;
       $106 = (($75) - ($104))|0;
       $dy$0$i = $87 ? $106 : $104;
       $107 = ($$$i>>>0)<($74>>>0);
       $108 = ($dy$0$i>>>0)<($75>>>0);
       $or$cond$i = $107 & $108;
       if ($or$cond$i) {
        $109 = HEAP32[$88>>2]|0;
        $110 = HEAP32[$89>>2]|0;
        $111 = Math_imul($110, $dy$0$i)|0;
        $112 = (($109) + ($111)|0);
        $113 = (($112) + ($$$i<<2)|0);
        $114 = HEAP8[$113>>0]|0;
        $115 = (((($112) + ($$$i<<2)|0)) + 1|0);
        $116 = HEAP8[$115>>0]|0;
        $117 = (((($112) + ($$$i<<2)|0)) + 2|0);
        $118 = HEAP8[$117>>0]|0;
        $119 = (((($112) + ($$$i<<2)|0)) + 3|0);
        $120 = HEAP8[$119>>0]|0;
        $121 = ((($113)) + 4|0);
        $122 = HEAP8[$121>>0]|0;
        $123 = ((($121)) + 1|0);
        $124 = HEAP8[$123>>0]|0;
        $125 = ((($121)) + 2|0);
        $126 = HEAP8[$125>>0]|0;
        $127 = ((($121)) + 3|0);
        $128 = HEAP8[$127>>0]|0;
        $129 = (($110|0) / 4)&-1;
        $130 = (($121) + ($129<<2)|0);
        $131 = HEAP8[$130>>0]|0;
        $132 = ((($130)) + 1|0);
        $133 = HEAP8[$132>>0]|0;
        $134 = ((($130)) + 2|0);
        $135 = HEAP8[$134>>0]|0;
        $136 = ((($130)) + 3|0);
        $137 = HEAP8[$136>>0]|0;
        $138 = ((($130)) + -4|0);
        $139 = HEAP8[$138>>0]|0;
        $140 = ((($138)) + 1|0);
        $141 = HEAP8[$140>>0]|0;
        $142 = ((($138)) + 2|0);
        $143 = HEAP8[$142>>0]|0;
        $144 = ((($138)) + 3|0);
        $145 = HEAP8[$144>>0]|0;
        $$1$i = $86 ? $128 : $120;
        $$2$i = $86 ? $126 : $118;
        $$3$i = $86 ? $124 : $116;
        $$4$i = $86 ? $122 : $114;
        $$5$i = $86 ? $120 : $128;
        $$6$i = $86 ? $118 : $126;
        $$7$i = $86 ? $116 : $124;
        $$8$i = $86 ? $114 : $122;
        $$9$i = $86 ? $137 : $145;
        $$10$i = $86 ? $135 : $143;
        $$11$i = $86 ? $133 : $141;
        $$12$i = $86 ? $131 : $139;
        $$13$i = $86 ? $145 : $137;
        $$14$i = $86 ? $143 : $135;
        $$15$i = $86 ? $141 : $133;
        $$16$i = $86 ? $139 : $131;
        $$9$$1$i = $87 ? $$9$i : $$1$i;
        $$10$$2$i = $87 ? $$10$i : $$2$i;
        $$11$$3$i = $87 ? $$11$i : $$3$i;
        $$12$$4$i = $87 ? $$12$i : $$4$i;
        $$13$$5$i = $87 ? $$13$i : $$5$i;
        $$14$$6$i = $87 ? $$14$i : $$6$i;
        $$15$$7$i = $87 ? $$15$i : $$7$i;
        $$16$$8$i = $87 ? $$16$i : $$8$i;
        $$1$$9$i = $87 ? $$1$i : $$9$i;
        $$2$$10$i = $87 ? $$2$i : $$10$i;
        $$3$$11$i = $87 ? $$3$i : $$11$i;
        $$4$$12$i = $87 ? $$4$i : $$12$i;
        $$5$$13$i = $87 ? $$5$i : $$13$i;
        $$6$$14$i = $87 ? $$6$i : $$14$i;
        $$7$$15$i = $87 ? $$7$i : $$15$i;
        $$8$$16$i = $87 ? $$8$i : $$16$i;
        $146 = $sdx$029$i & 65535;
        $147 = $sdy$028$i & 65535;
        $148 = $$16$$8$i&255;
        $149 = $$12$$4$i&255;
        $150 = (($148) - ($149))|0;
        $151 = Math_imul($150, $146)|0;
        $152 = $151 >>> 16;
        $153 = (($152) + ($149))|0;
        $154 = $153 & 255;
        $155 = $$8$$16$i&255;
        $156 = $$4$$12$i&255;
        $157 = (($155) - ($156))|0;
        $158 = Math_imul($157, $146)|0;
        $159 = $158 >>> 16;
        $160 = (($159) + ($156))|0;
        $161 = $160 & 255;
        $162 = (($161) - ($154))|0;
        $163 = Math_imul($162, $147)|0;
        $164 = $163 >>> 16;
        $165 = (($164) + ($153))|0;
        $166 = $165&255;
        HEAP8[$pc$130$i>>0] = $166;
        $167 = $$15$$7$i&255;
        $168 = $$11$$3$i&255;
        $169 = (($167) - ($168))|0;
        $170 = Math_imul($169, $146)|0;
        $171 = $170 >>> 16;
        $172 = (($171) + ($168))|0;
        $173 = $172 & 255;
        $174 = $$7$$15$i&255;
        $175 = $$3$$11$i&255;
        $176 = (($174) - ($175))|0;
        $177 = Math_imul($176, $146)|0;
        $178 = $177 >>> 16;
        $179 = (($178) + ($175))|0;
        $180 = $179 & 255;
        $181 = (($180) - ($173))|0;
        $182 = Math_imul($181, $147)|0;
        $183 = $182 >>> 16;
        $184 = (($183) + ($172))|0;
        $185 = $184&255;
        $186 = ((($pc$130$i)) + 1|0);
        HEAP8[$186>>0] = $185;
        $187 = $$14$$6$i&255;
        $188 = $$10$$2$i&255;
        $189 = (($187) - ($188))|0;
        $190 = Math_imul($189, $146)|0;
        $191 = $190 >>> 16;
        $192 = (($191) + ($188))|0;
        $193 = $192 & 255;
        $194 = $$6$$14$i&255;
        $195 = $$2$$10$i&255;
        $196 = (($194) - ($195))|0;
        $197 = Math_imul($196, $146)|0;
        $198 = $197 >>> 16;
        $199 = (($198) + ($195))|0;
        $200 = $199 & 255;
        $201 = (($200) - ($193))|0;
        $202 = Math_imul($201, $147)|0;
        $203 = $202 >>> 16;
        $204 = (($203) + ($192))|0;
        $205 = $204&255;
        $206 = ((($pc$130$i)) + 2|0);
        HEAP8[$206>>0] = $205;
        $207 = $$13$$5$i&255;
        $208 = $$9$$1$i&255;
        $209 = (($207) - ($208))|0;
        $210 = Math_imul($209, $146)|0;
        $211 = $210 >>> 16;
        $212 = (($211) + ($208))|0;
        $213 = $212 & 255;
        $214 = $$5$$13$i&255;
        $215 = $$1$$9$i&255;
        $216 = (($214) - ($215))|0;
        $217 = Math_imul($216, $146)|0;
        $218 = $217 >>> 16;
        $219 = (($218) + ($215))|0;
        $220 = $219 & 255;
        $221 = (($220) - ($213))|0;
        $222 = Math_imul($221, $147)|0;
        $223 = $222 >>> 16;
        $224 = (($223) + ($212))|0;
        $225 = $224&255;
        $226 = ((($pc$130$i)) + 3|0);
        HEAP8[$226>>0] = $225;
        $$pre$i = HEAP32[$60>>2]|0;
        $232 = $$pre$i;
       } else {
        $232 = $481;
       }
       $227 = (($sdx$029$i) + ($56))|0;
       $228 = (($sdy$028$i) + ($54))|0;
       $229 = ((($pc$130$i)) + 4|0);
       $230 = (($x$031$i) + 1)|0;
       $231 = ($230|0)<($232|0);
       if ($231) {
        $481 = $232;$pc$130$i = $229;$sdx$029$i = $227;$sdy$028$i = $228;$x$031$i = $230;
       } else {
        $$lcssa160 = $232;$$lcssa161 = $229;
        break;
       }
      }
      $$pre52$i = HEAP32[$33>>2]|0;
      $236 = $$pre52$i;$480 = $$lcssa160;$pc$1$lcssa$i = $$lcssa161;
     } else {
      $236 = $479;$480 = $98;$pc$1$lcssa$i = $pc$036$i;
     }
     $233 = (($pc$1$lcssa$i) + ($81)|0);
     $234 = (($y$037$i) + 1)|0;
     $235 = ($234|0)<($236|0);
     if ($235) {
      $479 = $236;$98 = $480;$pc$036$i = $233;$y$037$i = $234;
     } else {
      break;
     }
    }
   }
  }
  HEAP32[$key$i2>>2] = 0;
  (_SDL_GetColorKey($rz_src$016,$key$i2)|0);
  $324 = HEAP32[$key$i2>>2]|0;
  (_SDL_SetColorKey($rz_dst$0,3,$324)|0);
 } else {
  $44 = ((($rz_src$016)) + 4|0);
  $45 = HEAP32[$44>>2]|0;
  $46 = ((($45)) + 4|0);
  $47 = HEAP32[$46>>2]|0;
  $48 = HEAP32[$47>>2]|0;
  $49 = ($48|0)>(0);
  $50 = ((($rz_dst$0)) + 4|0);
  $51 = HEAP32[$50>>2]|0;
  $52 = ((($51)) + 4|0);
  $53 = HEAP32[$52>>2]|0;
  if ($49) {
   $326 = $53;$329 = $47;$i$045 = 0;
   while(1) {
    $325 = ((($326)) + 4|0);
    $327 = HEAP32[$325>>2]|0;
    $328 = ((($329)) + 4|0);
    $330 = HEAP32[$328>>2]|0;
    $331 = (($327) + ($i$045<<2)|0);
    $332 = (($330) + ($i$045<<2)|0);
    $333 = HEAPU8[$332>>0]|(HEAPU8[$332+1>>0]<<8)|(HEAPU8[$332+2>>0]<<16)|(HEAPU8[$332+3>>0]<<24);
    HEAP8[$331>>0]=$333&255;HEAP8[$331+1>>0]=($333>>8)&255;HEAP8[$331+2>>0]=($333>>16)&255;HEAP8[$331+3>>0]=$333>>24;
    $334 = (($i$045) + 1)|0;
    $335 = HEAP32[$44>>2]|0;
    $336 = ((($335)) + 4|0);
    $337 = HEAP32[$336>>2]|0;
    $338 = HEAP32[$337>>2]|0;
    $339 = ($334|0)<($338|0);
    $340 = HEAP32[$50>>2]|0;
    $341 = ((($340)) + 4|0);
    $342 = HEAP32[$341>>2]|0;
    if ($339) {
     $326 = $342;$329 = $337;$i$045 = $334;
    } else {
     $$lcssa43 = $338;$$lcssa44 = $342;
     break;
    }
   }
  } else {
   $$lcssa43 = $48;$$lcssa44 = $53;
  }
  HEAP32[$$lcssa44>>2] = $$lcssa43;
  $343 = (~~(($55)));
  $344 = (~~(($57)));
  $345 = ((($rz_src$016)) + 8|0);
  $346 = HEAP32[$345>>2]|0;
  $347 = ((($rz_dst$0)) + 8|0);
  $348 = HEAP32[$347>>2]|0;
  $349 = ((($rz_src$016)) + 12|0);
  $350 = HEAP32[$349>>2]|0;
  $351 = HEAP32[$33>>2]|0;
  $352 = ((($rz_dst$0)) + 20|0);
  $353 = HEAP32[$352>>2]|0;
  $354 = ((($rz_dst$0)) + 16|0);
  $355 = HEAP32[$354>>2]|0;
  $356 = (($355) - ($348))|0;
  HEAP32[$key$i$i>>2] = 0;
  (_SDL_GetColorKey($rz_src$016,$key$i$i)|0);
  $357 = HEAP32[$key$i$i>>2]|0;
  $358 = $357 & 255;
  $359 = HEAP32[$354>>2]|0;
  $360 = HEAP32[$33>>2]|0;
  $361 = Math_imul($360, $359)|0;
  (_SDL_memset($353,$358,$361)|0);
  $362 = HEAP32[$33>>2]|0;
  $363 = ($362|0)>(0);
  if ($363) {
   $364 = $centery << 16;
   $365 = (65536 - ($344))|0;
   $366 = Math_imul($365, $centerx)|0;
   $367 = (($350) - ($351))|0;
   $368 = $367 << 15;
   $369 = (($346) - ($348))|0;
   $370 = $369 << 15;
   $371 = (($370) + ($366))|0;
   $372 = ($flipx|0)==(0);
   $373 = ($flipy|0)==(0);
   $374 = ((($rz_src$016)) + 20|0);
   $375 = ((($rz_src$016)) + 16|0);
   $$pre$i3 = HEAP32[$347>>2]|0;
   $$neg$i$neg = Math_imul($343, $centerx)|0;
   $376 = (($368) + ($364))|0;
   $382 = $$pre$i3;$pc$010$i = $353;$y$09$i = 0;
   while(1) {
    $377 = (($centery) - ($y$09$i))|0;
    $378 = Math_imul($377, $343)|0;
    $379 = (($371) + ($378))|0;
    $$neg1$i$neg = Math_imul($377, $344)|0;
    $tmp = (($$neg1$i$neg) + ($$neg$i$neg))|0;
    $380 = (($376) - ($tmp))|0;
    $381 = ($382|0)>(0);
    if ($381) {
     if ($372) {
      if ($373) {
       $488 = $382;$pc$17$us$us$i = $pc$010$i;$sdx$05$us$us$i = $379;$sdy$06$us$us$i = $380;$x$08$us$us$i = 0;
       while(1) {
        $383 = $sdx$05$us$us$i >> 16;
        $384 = $sdy$06$us$us$i >> 16;
        $385 = HEAP32[$345>>2]|0;
        $386 = ($383>>>0)<($385>>>0);
        do {
         if ($386) {
          $387 = HEAP32[$349>>2]|0;
          $388 = ($384>>>0)<($387>>>0);
          if (!($388)) {
           $400 = $488;
           break;
          }
          $389 = HEAP32[$374>>2]|0;
          $390 = HEAP32[$375>>2]|0;
          $391 = Math_imul($390, $384)|0;
          $392 = (($389) + ($391)|0);
          $393 = (($392) + ($383)|0);
          $394 = HEAP8[$393>>0]|0;
          HEAP8[$pc$17$us$us$i>>0] = $394;
          $$pre29$i = HEAP32[$347>>2]|0;
          $400 = $$pre29$i;
         } else {
          $400 = $488;
         }
        } while(0);
        $395 = (($sdx$05$us$us$i) + ($344))|0;
        $396 = (($sdy$06$us$us$i) + ($343))|0;
        $397 = ((($pc$17$us$us$i)) + 1|0);
        $398 = (($x$08$us$us$i) + 1)|0;
        $399 = ($398|0)<($400|0);
        if ($399) {
         $488 = $400;$pc$17$us$us$i = $397;$sdx$05$us$us$i = $395;$sdy$06$us$us$i = $396;$x$08$us$us$i = $398;
        } else {
         $487 = $400;$pc$1$lcssa$i11 = $397;
         break;
        }
       }
      } else {
       $489 = $382;$pc$17$us$i = $pc$010$i;$sdx$05$us$i = $379;$sdy$06$us$i = $380;$x$08$us$i = 0;
       while(1) {
        $401 = $sdx$05$us$i >> 16;
        $402 = $sdy$06$us$i >> 16;
        $403 = HEAP32[$345>>2]|0;
        $404 = ($401>>>0)<($403>>>0);
        do {
         if ($404) {
          $405 = HEAP32[$349>>2]|0;
          $406 = ($402>>>0)<($405>>>0);
          if (!($406)) {
           $420 = $489;
           break;
          }
          $407 = $402 ^ -1;
          $408 = (($405) + ($407))|0;
          $409 = HEAP32[$374>>2]|0;
          $410 = HEAP32[$375>>2]|0;
          $411 = Math_imul($410, $408)|0;
          $412 = (($409) + ($411)|0);
          $413 = (($412) + ($401)|0);
          $414 = HEAP8[$413>>0]|0;
          HEAP8[$pc$17$us$i>>0] = $414;
          $$pre28$i = HEAP32[$347>>2]|0;
          $420 = $$pre28$i;
         } else {
          $420 = $489;
         }
        } while(0);
        $415 = (($sdx$05$us$i) + ($344))|0;
        $416 = (($sdy$06$us$i) + ($343))|0;
        $417 = ((($pc$17$us$i)) + 1|0);
        $418 = (($x$08$us$i) + 1)|0;
        $419 = ($418|0)<($420|0);
        if ($419) {
         $489 = $420;$pc$17$us$i = $417;$sdx$05$us$i = $415;$sdy$06$us$i = $416;$x$08$us$i = $418;
        } else {
         $487 = $420;$pc$1$lcssa$i11 = $417;
         break;
        }
       }
      }
     } else {
      if ($373) {
       $490 = $382;$pc$17$us15$i = $pc$010$i;$sdx$05$us17$i = $379;$sdy$06$us16$i = $380;$x$08$us14$i = 0;
       while(1) {
        $421 = $sdx$05$us17$i >> 16;
        $422 = $sdy$06$us16$i >> 16;
        $423 = HEAP32[$345>>2]|0;
        $424 = ($421>>>0)<($423>>>0);
        do {
         if ($424) {
          $425 = HEAP32[$349>>2]|0;
          $426 = ($422>>>0)<($425>>>0);
          if (!($426)) {
           $440 = $490;
           break;
          }
          $427 = $421 ^ -1;
          $428 = (($423) + ($427))|0;
          $429 = HEAP32[$374>>2]|0;
          $430 = HEAP32[$375>>2]|0;
          $431 = Math_imul($430, $422)|0;
          $432 = (($429) + ($431)|0);
          $433 = (($432) + ($428)|0);
          $434 = HEAP8[$433>>0]|0;
          HEAP8[$pc$17$us15$i>>0] = $434;
          $$pre27$i = HEAP32[$347>>2]|0;
          $440 = $$pre27$i;
         } else {
          $440 = $490;
         }
        } while(0);
        $435 = (($sdx$05$us17$i) + ($344))|0;
        $436 = (($sdy$06$us16$i) + ($343))|0;
        $437 = ((($pc$17$us15$i)) + 1|0);
        $438 = (($x$08$us14$i) + 1)|0;
        $439 = ($438|0)<($440|0);
        if ($439) {
         $490 = $440;$pc$17$us15$i = $437;$sdx$05$us17$i = $435;$sdy$06$us16$i = $436;$x$08$us14$i = $438;
        } else {
         $487 = $440;$pc$1$lcssa$i11 = $437;
         break;
        }
       }
      } else {
       $491 = $382;$pc$17$i = $pc$010$i;$sdx$05$i = $379;$sdy$06$i = $380;$x$08$i = 0;
       while(1) {
        $441 = $sdx$05$i >> 16;
        $442 = $sdy$06$i >> 16;
        $443 = HEAP32[$345>>2]|0;
        $444 = ($441>>>0)<($443>>>0);
        do {
         if ($444) {
          $445 = HEAP32[$349>>2]|0;
          $446 = ($442>>>0)<($445>>>0);
          if (!($446)) {
           $462 = $491;
           break;
          }
          $447 = $441 ^ -1;
          $448 = (($443) + ($447))|0;
          $449 = $442 ^ -1;
          $450 = (($445) + ($449))|0;
          $451 = HEAP32[$374>>2]|0;
          $452 = HEAP32[$375>>2]|0;
          $453 = Math_imul($452, $450)|0;
          $454 = (($451) + ($453)|0);
          $455 = (($454) + ($448)|0);
          $456 = HEAP8[$455>>0]|0;
          HEAP8[$pc$17$i>>0] = $456;
          $$pre26$i = HEAP32[$347>>2]|0;
          $462 = $$pre26$i;
         } else {
          $462 = $491;
         }
        } while(0);
        $457 = (($sdx$05$i) + ($344))|0;
        $458 = (($sdy$06$i) + ($343))|0;
        $459 = ((($pc$17$i)) + 1|0);
        $460 = (($x$08$i) + 1)|0;
        $461 = ($460|0)<($462|0);
        if ($461) {
         $491 = $462;$pc$17$i = $459;$sdx$05$i = $457;$sdy$06$i = $458;$x$08$i = $460;
        } else {
         $487 = $462;$pc$1$lcssa$i11 = $459;
         break;
        }
       }
      }
     }
    } else {
     $487 = $382;$pc$1$lcssa$i11 = $pc$010$i;
    }
    $463 = (($pc$1$lcssa$i11) + ($356)|0);
    $464 = (($y$09$i) + 1)|0;
    $465 = HEAP32[$33>>2]|0;
    $466 = ($464|0)<($465|0);
    if ($466) {
     $382 = $487;$pc$010$i = $463;$y$09$i = $464;
    } else {
     break;
    }
   }
  }
  HEAP32[$key$i1>>2] = 0;
  (_SDL_GetColorKey($rz_src$016,$key$i1)|0);
  $467 = HEAP32[$key$i1>>2]|0;
  (_SDL_SetColorKey($rz_dst$0,3,$467)|0);
 }
 (_SDL_GetSurfaceAlphaMod($src,$alphaMod)|0);
 (_SDL_GetSurfaceBlendMode($src,$blendMode)|0);
 (_SDL_GetSurfaceColorMod($src,$cr,$cg,$cb)|0);
 $468 = HEAP8[$alphaMod>>0]|0;
 (_SDL_SetSurfaceAlphaMod($rz_dst$0,$468)|0);
 $469 = HEAP32[$blendMode>>2]|0;
 (_SDL_SetSurfaceBlendMode($rz_dst$0,$469)|0);
 $470 = HEAP8[$cr>>0]|0;
 $471 = HEAP8[$cg>>0]|0;
 $472 = HEAP8[$cb>>0]|0;
 (_SDL_SetSurfaceColorMod($rz_dst$0,$470,$471,$472)|0);
 $473 = HEAP32[$rz_src$016>>2]|0;
 $474 = $473 & 2;
 $475 = ($474|0)==(0);
 if (!($475)) {
  _SDL_UnlockSurface($rz_src$016);
 }
 $476 = ($rz_src$016|0)==($src|0);
 if ($476) {
  $$0 = $rz_dst$0;
  STACKTOP = sp;return ($$0|0);
 }
 _SDL_FreeSurface($rz_src$016);
 $$0 = $rz_dst$0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GetRGB($pixel,$format,$r,$g,$b) {
 $pixel = $pixel|0;
 $format = $format|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($format)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $3 = ((($format)) + 12|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = $4 & $pixel;
  $6 = ((($format)) + 32|0);
  $7 = HEAP8[$6>>0]|0;
  $8 = $7&255;
  $9 = $5 >>> $8;
  $10 = ((($format)) + 28|0);
  $11 = HEAP8[$10>>0]|0;
  $12 = $11&255;
  $13 = (125924 + ($12<<2)|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = (($14) + ($9)|0);
  $16 = HEAP8[$15>>0]|0;
  HEAP8[$r>>0] = $16;
  $17 = ((($format)) + 16|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = $18 & $pixel;
  $20 = ((($format)) + 33|0);
  $21 = HEAP8[$20>>0]|0;
  $22 = $21&255;
  $23 = $19 >>> $22;
  $24 = ((($format)) + 29|0);
  $25 = HEAP8[$24>>0]|0;
  $26 = $25&255;
  $27 = (125924 + ($26<<2)|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = (($28) + ($23)|0);
  $30 = HEAP8[$29>>0]|0;
  HEAP8[$g>>0] = $30;
  $31 = ((($format)) + 20|0);
  $32 = HEAP32[$31>>2]|0;
  $33 = $32 & $pixel;
  $34 = ((($format)) + 34|0);
  $35 = HEAP8[$34>>0]|0;
  $36 = $35&255;
  $37 = $33 >>> $36;
  $38 = ((($format)) + 30|0);
  $39 = HEAP8[$38>>0]|0;
  $40 = $39&255;
  $41 = (125924 + ($40<<2)|0);
  $42 = HEAP32[$41>>2]|0;
  $43 = (($42) + ($37)|0);
  $44 = HEAP8[$43>>0]|0;
  HEAP8[$b>>0] = $44;
  return;
 }
 $45 = HEAP32[$1>>2]|0;
 $46 = ($45>>>0)>($pixel>>>0);
 if ($46) {
  $47 = ((($1)) + 4|0);
  $48 = HEAP32[$47>>2]|0;
  $49 = (($48) + ($pixel<<2)|0);
  $50 = HEAP8[$49>>0]|0;
  HEAP8[$r>>0] = $50;
  $51 = HEAP32[$0>>2]|0;
  $52 = ((($51)) + 4|0);
  $53 = HEAP32[$52>>2]|0;
  $54 = (((($53) + ($pixel<<2)|0)) + 1|0);
  $55 = HEAP8[$54>>0]|0;
  HEAP8[$g>>0] = $55;
  $56 = HEAP32[$0>>2]|0;
  $57 = ((($56)) + 4|0);
  $58 = HEAP32[$57>>2]|0;
  $59 = (((($58) + ($pixel<<2)|0)) + 2|0);
  $60 = HEAP8[$59>>0]|0;
  HEAP8[$b>>0] = $60;
  return;
 } else {
  HEAP8[$b>>0] = 0;
  HEAP8[$g>>0] = 0;
  HEAP8[$r>>0] = 0;
  return;
 }
}
function _SDL_ConvertSurfaceFormat($surface,$pixel_format,$flags) {
 $surface = $surface|0;
 $pixel_format = $pixel_format|0;
 $flags = $flags|0;
 var $0 = 0, $1 = 0, $2 = 0, $convert$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_AllocFormat($pixel_format)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  $convert$0 = 0;
  return ($convert$0|0);
 }
 $2 = (_SDL_ConvertSurface($surface,$0,$flags)|0);
 _SDL_FreeFormat($0);
 $convert$0 = $2;
 return ($convert$0|0);
}
function _SDL_MapRGB($format,$r,$g,$b) {
 $format = $format|0;
 $r = $r|0;
 $g = $g|0;
 $b = $b|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, $i$02$i = 0, $pixel$03$i = 0, $pixel$1$i = 0, $pixel$2$i = 0, $smallest$01$i = 0, $smallest$1$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($format)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $3 = $r&255;
  $4 = ((($format)) + 28|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = $5 & 255;
  $7 = $3 >>> $6;
  $8 = ((($format)) + 32|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = $9 & 255;
  $11 = $7 << $10;
  $12 = $g&255;
  $13 = $5 >>> 8;
  $14 = $13 & 255;
  $15 = $12 >>> $14;
  $16 = $9 >>> 8;
  $17 = $16 & 255;
  $18 = $15 << $17;
  $19 = $18 | $11;
  $20 = $b&255;
  $21 = $5 >>> 16;
  $22 = $21 & 255;
  $23 = $20 >>> $22;
  $24 = $9 >>> 16;
  $25 = $24 & 255;
  $26 = $23 << $25;
  $27 = ((($format)) + 24|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = $19 | $28;
  $30 = $29 | $26;
  $$0 = $30;
  return ($$0|0);
 }
 $31 = HEAP32[$1>>2]|0;
 $32 = ($31|0)>(0);
 L5: do {
  if ($32) {
   $33 = ((($1)) + 4|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = $r&255;
   $36 = $g&255;
   $37 = $b&255;
   $i$02$i = 0;$pixel$03$i = 0;$smallest$01$i = -1;
   while(1) {
    $38 = (($34) + ($i$02$i<<2)|0);
    $39 = HEAP8[$38>>0]|0;
    $40 = $39&255;
    $41 = (($40) - ($35))|0;
    $42 = (((($34) + ($i$02$i<<2)|0)) + 1|0);
    $43 = HEAP8[$42>>0]|0;
    $44 = $43&255;
    $45 = (($44) - ($36))|0;
    $46 = (((($34) + ($i$02$i<<2)|0)) + 2|0);
    $47 = HEAP8[$46>>0]|0;
    $48 = $47&255;
    $49 = (($48) - ($37))|0;
    $50 = (((($34) + ($i$02$i<<2)|0)) + 3|0);
    $51 = HEAP8[$50>>0]|0;
    $52 = $51&255;
    $53 = (($52) + -255)|0;
    $54 = Math_imul($41, $41)|0;
    $55 = Math_imul($45, $45)|0;
    $56 = (($55) + ($54))|0;
    $57 = Math_imul($49, $49)|0;
    $58 = (($56) + ($57))|0;
    $59 = Math_imul($53, $53)|0;
    $60 = (($58) + ($59))|0;
    $61 = ($60>>>0)<($smallest$01$i>>>0);
    if ($61) {
     $62 = $i$02$i&255;
     $63 = ($60|0)==(0);
     if ($63) {
      $pixel$2$i = $62;
      break L5;
     } else {
      $pixel$1$i = $62;$smallest$1$i = $60;
     }
    } else {
     $pixel$1$i = $pixel$03$i;$smallest$1$i = $smallest$01$i;
    }
    $64 = (($i$02$i) + 1)|0;
    $65 = ($64|0)<($31|0);
    if ($65) {
     $i$02$i = $64;$pixel$03$i = $pixel$1$i;$smallest$01$i = $smallest$1$i;
    } else {
     $pixel$2$i = $pixel$1$i;
     break;
    }
   }
  } else {
   $pixel$2$i = 0;
  }
 } while(0);
 $66 = $pixel$2$i&255;
 $$0 = $66;
 return ($$0|0);
}
function _SW_RenderReadPixels($renderer,$rect,$format,$pixels,$pitch) {
 $renderer = $renderer|0;
 $rect = $rect|0;
 $format = $format|0;
 $pixels = $pixels|0;
 $pitch = $pitch|0;
 var $$0 = 0, $$01 = 0, $$pr = 0, $$pre = 0, $$pre2 = 0, $$pre7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $8 = 0, $9 = 0, $final_rect = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $final_rect = sp + 8|0;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $4 = ((($1)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  HEAP32[$1>>2] = $5;
  $6 = ($5|0)==(0);
  $7 = $5;
  if ($6) {
   $8 = ((($renderer)) + 188|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (_SDL_GetWindowSurface($9)|0);
   $11 = ($10|0)==(0|0);
   do {
    if (!($11)) {
     HEAP32[$4>>2] = $10;
     HEAP32[$1>>2] = $10;
     $12 = HEAP32[$0>>2]|0;
     $13 = HEAP32[$12>>2]|0;
     $14 = ($13|0)==(0|0);
     if (!($14)) {
      $15 = ((($renderer)) + 212|0);
      (_SDL_SetClipRect($13,$15)|0);
      $$pre = HEAP32[$0>>2]|0;
      $$pre2 = HEAP32[$$pre>>2]|0;
      $16 = ($$pre2|0)==(0|0);
      if (!($16)) {
       $17 = ((($renderer)) + 276|0);
       $18 = HEAP32[$17>>2]|0;
       $19 = ($18|0)==(0);
       if ($19) {
        (_SDL_SetClipRect($$pre2,0)|0);
        break;
       } else {
        $20 = ((($renderer)) + 244|0);
        (_SDL_SetClipRect($$pre2,$20)|0);
        break;
       }
      }
     }
    }
   } while(0);
   $$pr = HEAP32[$1>>2]|0;
   $22 = $$pr;
  } else {
   $22 = $7;
  }
  $21 = ($22|0)==(0|0);
  if ($21) {
   $$0 = -1;$75 = $final_rect;
   STACKTOP = sp;return ($$0|0);
  } else {
   $46 = $22;$74 = $final_rect;
  }
 } else {
  $46 = $2;$74 = $final_rect;
 }
 $23 = ((($renderer)) + 212|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($renderer)) + 216|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = $24 | $26;
 $28 = ($27|0)==(0);
 $$pre7 = HEAP32[$rect>>2]|0;
 if ($28) {
  $$01 = $rect;$41 = $$pre7;
 } else {
  $29 = (($$pre7) + ($24))|0;
  HEAP32[$final_rect>>2] = $29;
  $30 = ((($rect)) + 4|0);
  $31 = HEAP32[$30>>2]|0;
  $32 = (($31) + ($26))|0;
  $33 = ((($final_rect)) + 4|0);
  HEAP32[$33>>2] = $32;
  $34 = ((($rect)) + 8|0);
  $35 = HEAP32[$34>>2]|0;
  $36 = ((($final_rect)) + 8|0);
  HEAP32[$36>>2] = $35;
  $37 = ((($rect)) + 12|0);
  $38 = HEAP32[$37>>2]|0;
  $39 = ((($final_rect)) + 12|0);
  HEAP32[$39>>2] = $38;
  $$01 = $final_rect;$41 = $29;
 }
 $40 = ($41|0)<(0);
 if (!($40)) {
  $42 = ((($$01)) + 8|0);
  $43 = HEAP32[$42>>2]|0;
  $44 = (($43) + ($41))|0;
  $45 = ((($46)) + 8|0);
  $47 = HEAP32[$45>>2]|0;
  $48 = ($44|0)>($47|0);
  if (!($48)) {
   $49 = ((($$01)) + 4|0);
   $50 = HEAP32[$49>>2]|0;
   $51 = ($50|0)<(0);
   if (!($51)) {
    $52 = ((($$01)) + 12|0);
    $53 = HEAP32[$52>>2]|0;
    $54 = (($53) + ($50))|0;
    $55 = ((($46)) + 12|0);
    $56 = HEAP32[$55>>2]|0;
    $57 = ($54|0)>($56|0);
    if (!($57)) {
     $59 = ((($46)) + 4|0);
     $60 = HEAP32[$59>>2]|0;
     $61 = HEAP32[$60>>2]|0;
     $62 = ((($46)) + 20|0);
     $63 = HEAP32[$62>>2]|0;
     $64 = ((($46)) + 16|0);
     $65 = HEAP32[$64>>2]|0;
     $66 = Math_imul($65, $50)|0;
     $67 = (($63) + ($66)|0);
     $68 = ((($60)) + 9|0);
     $69 = HEAP8[$68>>0]|0;
     $70 = $69&255;
     $71 = Math_imul($70, $41)|0;
     $72 = (($67) + ($71)|0);
     $73 = (_SDL_ConvertPixels($43,$53,$61,$72,$65,$format,$pixels,$pitch)|0);
     $$0 = $73;$75 = $74;
     STACKTOP = sp;return ($$0|0);
    }
   }
  }
 }
 $58 = (_SDL_SetError(604362,$vararg_buffer)|0);
 $$0 = $58;$75 = $74;
 STACKTOP = sp;return ($$0|0);
}
function _SW_RenderPresent($renderer) {
 $renderer = $renderer|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($renderer)) + 188|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 (_SDL_UpdateWindowSurface($1)|0);
 return;
}
function _SDL_UpdateWindowSurface($window) {
 $window = $window|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $full_rect = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $full_rect = sp + 24|0;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   HEAP32[$full_rect>>2] = 0;
   $6 = ((($full_rect)) + 4|0);
   HEAP32[$6>>2] = 0;
   $7 = ((($window)) + 24|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ((($full_rect)) + 8|0);
   HEAP32[$9>>2] = $8;
   $10 = ((($window)) + 28|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = ((($full_rect)) + 12|0);
   HEAP32[$12>>2] = $11;
   $13 = ((($window)) + 108|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ($14|0)==(0);
   if ($15) {
    $16 = (_SDL_SetError(604402,$vararg_buffer3)|0);
    $$0 = $16;
    STACKTOP = sp;return ($$0|0);
   } else {
    $17 = ((($0)) + 112|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = (FUNCTION_TABLE_iiiii[$18 & 8191]($0,$window,$full_rect,1)|0);
    $$0 = $19;
    STACKTOP = sp;return ($$0|0);
   }
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SW_DestroyTexture($renderer,$texture) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($texture)) + 68|0);
 $1 = HEAP32[$0>>2]|0;
 _SDL_FreeSurface($1);
 return;
}
function _SW_DestroyRenderer($renderer) {
 $renderer = $renderer|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($renderer)) + 316|0);
 $1 = HEAP32[$0>>2]|0;
 _SDL_free($1);
 _SDL_free($renderer);
 return;
}
function _Emscripten_Available() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1;
}
function _Emscripten_CreateDevice($devindex) {
 $devindex = $devindex|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_calloc(1,664)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_Error(0)|0);
  $$0 = 0;
  return ($$0|0);
 } else {
  (_SDL_SetHint(604513,604546)|0);
  $2 = ((($0)) + 4|0);
  HEAP32[$2>>2] = 4564;
  $3 = ((($0)) + 8|0);
  HEAP32[$3>>2] = 4565;
  $4 = ((($0)) + 24|0);
  HEAP32[$4>>2] = 4566;
  $5 = ((($0)) + 180|0);
  HEAP32[$5>>2] = 4567;
  $6 = ((($0)) + 28|0);
  HEAP32[$6>>2] = 4568;
  $7 = ((($0)) + 36|0);
  HEAP32[$7>>2] = 4569;
  $8 = ((($0)) + 48|0);
  HEAP32[$8>>2] = 4570;
  $9 = ((($0)) + 104|0);
  HEAP32[$9>>2] = 4571;
  $10 = ((($0)) + 88|0);
  HEAP32[$10>>2] = 4572;
  $11 = ((($0)) + 108|0);
  HEAP32[$11>>2] = 4573;
  $12 = ((($0)) + 112|0);
  HEAP32[$12>>2] = 4574;
  $13 = ((($0)) + 116|0);
  HEAP32[$13>>2] = 4575;
  $14 = ((($0)) + 140|0);
  HEAP32[$14>>2] = 4576;
  $15 = ((($0)) + 144|0);
  HEAP32[$15>>2] = 4577;
  $16 = ((($0)) + 148|0);
  HEAP32[$16>>2] = 4578;
  $17 = ((($0)) + 152|0);
  HEAP32[$17>>2] = 4579;
  $18 = ((($0)) + 156|0);
  HEAP32[$18>>2] = 4580;
  $19 = ((($0)) + 164|0);
  HEAP32[$19>>2] = 4581;
  $20 = ((($0)) + 168|0);
  HEAP32[$20>>2] = 4582;
  $21 = ((($0)) + 172|0);
  HEAP32[$21>>2] = 4583;
  $22 = ((($0)) + 176|0);
  HEAP32[$22>>2] = 4584;
  $23 = ((($0)) + 160|0);
  HEAP32[$23>>2] = 4585;
  $24 = ((($0)) + 660|0);
  HEAP32[$24>>2] = 4586;
  $$0 = $0;
  return ($$0|0);
 }
 return (0)|0;
}
function _SDL_SetHint($name,$value) {
 $name = $name|0;
 $value = $value|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_SetHintWithPriority($name,$value,1)|0);
 return ($0|0);
}
function _SDL_SetHintWithPriority($name,$value,$priority) {
 $name = $name|0;
 $value = $value|0;
 $priority = $priority|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $entry$07 = 0, $hint$0 = 0, $hint$08 = 0, $hint$09 = 0, $hint$09$lcssa = 0, $or$cond = 0, $or$cond3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($name|0)!=(0|0);
 $1 = ($value|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  $$0 = 0;
  return ($$0|0);
 }
 $2 = (_SDL_getenv($name)|0);
 $3 = ($2|0)!=(0|0);
 $4 = ($priority>>>0)<(2);
 $or$cond3 = $4 & $3;
 if ($or$cond3) {
  $$0 = 0;
  return ($$0|0);
 }
 $hint$08 = HEAP32[300735]|0;
 $5 = ($hint$08|0)==(0|0);
 L7: do {
  if (!($5)) {
   $hint$09 = $hint$08;
   while(1) {
    $6 = HEAP32[$hint$09>>2]|0;
    $7 = (_SDL_strcmp($name,$6)|0);
    $8 = ($7|0)==(0);
    if ($8) {
     $hint$09$lcssa = $hint$09;
     break;
    }
    $29 = ((($hint$09)) + 16|0);
    $hint$0 = HEAP32[$29>>2]|0;
    $30 = ($hint$0|0)==(0|0);
    if ($30) {
     break L7;
    } else {
     $hint$09 = $hint$0;
    }
   }
   $9 = ((($hint$09$lcssa)) + 8|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($10>>>0)>($priority>>>0);
   if ($11) {
    $$0 = 0;
    return ($$0|0);
   }
   $12 = ((($hint$09$lcssa)) + 4|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = ($13|0)==(0|0);
   if ($14) {
    label = 8;
   } else {
    $15 = (_SDL_strcmp($13,$value)|0);
    $16 = ($15|0)==(0);
    if (!($16)) {
     label = 8;
    }
   }
   if ((label|0) == 8) {
    $17 = ((($hint$09$lcssa)) + 12|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = ($18|0)==(0|0);
    if (!($19)) {
     $entry$07 = $18;
     while(1) {
      $20 = ((($entry$07)) + 8|0);
      $21 = HEAP32[$20>>2]|0;
      $22 = HEAP32[$entry$07>>2]|0;
      $23 = ((($entry$07)) + 4|0);
      $24 = HEAP32[$23>>2]|0;
      $25 = HEAP32[$12>>2]|0;
      FUNCTION_TABLE_viiii[$22 & 8191]($24,$name,$25,$value);
      $26 = ($21|0)==(0|0);
      if ($26) {
       break;
      } else {
       $entry$07 = $21;
      }
     }
    }
    $27 = HEAP32[$12>>2]|0;
    _SDL_free($27);
    $28 = (_SDL_strdup($value)|0);
    HEAP32[$12>>2] = $28;
   }
   HEAP32[$9>>2] = $priority;
   $$0 = 1;
   return ($$0|0);
  }
 } while(0);
 $31 = (_SDL_malloc(20)|0);
 $32 = ($31|0)==(0|0);
 if ($32) {
  $$0 = 0;
  return ($$0|0);
 }
 $33 = (_SDL_strdup($name)|0);
 HEAP32[$31>>2] = $33;
 $34 = (_SDL_strdup($value)|0);
 $35 = ((($31)) + 4|0);
 HEAP32[$35>>2] = $34;
 $36 = ((($31)) + 8|0);
 HEAP32[$36>>2] = $priority;
 $37 = ((($31)) + 12|0);
 HEAP32[$37>>2] = 0;
 $38 = HEAP32[300735]|0;
 $39 = ((($31)) + 16|0);
 HEAP32[$39>>2] = $38;
 HEAP32[300735] = $31;
 $$0 = 1;
 return ($$0|0);
}
function _SDL_strdup($string) {
 $string = $string|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (___strdup($string)|0);
 return ($0|0);
}
function _Emscripten_VideoInit($_this) {
 $_this = $_this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $mode = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $mode = sp;
 HEAP32[$mode>>2] = 370546692;
 $0 = _emscripten_asm_const_i(1)|0;
 $1 = ((($mode)) + 4|0);
 HEAP32[$1>>2] = $0;
 $2 = _emscripten_asm_const_i(2)|0;
 $3 = ((($mode)) + 8|0);
 HEAP32[$3>>2] = $2;
 $4 = ((($mode)) + 12|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($mode)) + 16|0);
 HEAP32[$5>>2] = 0;
 $6 = (_SDL_AddBasicVideoDisplay($mode)|0);
 $7 = ($6|0)<(0);
 if ($7) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 (_SDL_memset($mode,0,20)|0);
 $8 = ((($_this)) + 244|0);
 $9 = HEAP32[$8>>2]|0;
 (_SDL_AddDisplayMode($9,$mode)|0);
 _Emscripten_InitMouse();
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_AddBasicVideoDisplay($desktop_mode) {
 $desktop_mode = $desktop_mode|0;
 var $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $display = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $display = sp;
 (_SDL_memset($display,0,68)|0);
 $0 = ($desktop_mode|0)==(0|0);
 $$pre = ((($display)) + 16|0);
 if (!($0)) {
  ;HEAP32[$$pre>>2]=HEAP32[$desktop_mode>>2]|0;HEAP32[$$pre+4>>2]=HEAP32[$desktop_mode+4>>2]|0;HEAP32[$$pre+8>>2]=HEAP32[$desktop_mode+8>>2]|0;HEAP32[$$pre+12>>2]=HEAP32[$desktop_mode+12>>2]|0;HEAP32[$$pre+16>>2]=HEAP32[$desktop_mode+16>>2]|0;
 }
 $1 = ((($display)) + 36|0);
 ;HEAP32[$1>>2]=HEAP32[$$pre>>2]|0;HEAP32[$1+4>>2]=HEAP32[$$pre+4>>2]|0;HEAP32[$1+8>>2]=HEAP32[$$pre+8>>2]|0;HEAP32[$1+12>>2]=HEAP32[$$pre+12>>2]|0;HEAP32[$1+16>>2]=HEAP32[$$pre+16>>2]|0;
 $2 = (_SDL_AddVideoDisplay($display)|0);
 STACKTOP = sp;return ($2|0);
}
function _SDL_AddVideoDisplay($display) {
 $display = $display|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $index$0 = 0, $name = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $name = sp;
 $0 = HEAP32[300441]|0;
 $1 = ((($0)) + 244|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($0)) + 240|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4*68)|0;
 $6 = (($5) + 68)|0;
 $7 = (_SDL_realloc($2,$6)|0);
 $8 = ($7|0)==(0|0);
 if ($8) {
  (_SDL_Error(0)|0);
  $index$0 = -1;
  STACKTOP = sp;return ($index$0|0);
 }
 $9 = HEAP32[300441]|0;
 $10 = ((($9)) + 240|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($11) + 1)|0;
 HEAP32[$10>>2] = $12;
 $13 = (($7) + (($11*68)|0)|0);
 dest=$13; src=$display; stop=dest+68|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 $14 = $9;
 $15 = (((($7) + (($11*68)|0)|0)) + 60|0);
 HEAP32[$15>>2] = $14;
 $16 = ((($9)) + 244|0);
 HEAP32[$16>>2] = $7;
 $17 = HEAP32[$display>>2]|0;
 $18 = ($17|0)==(0|0);
 if ($18) {
  (_SDL_itoa($11,$name,10)|0);
  $20 = (_SDL_strdup($name)|0);
  HEAP32[$13>>2] = $20;
  $index$0 = $11;
  STACKTOP = sp;return ($index$0|0);
 } else {
  $19 = (_SDL_strdup($17)|0);
  HEAP32[$13>>2] = $19;
  $index$0 = $11;
  STACKTOP = sp;return ($index$0|0);
 }
 return (0)|0;
}
function _SDL_itoa($value,$string,$radix) {
 $value = $value|0;
 $string = $string|0;
 $radix = $radix|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_ltoa($value,$string,$radix)|0);
 return ($0|0);
}
function _SDL_ltoa($value,$string,$radix) {
 $value = $value|0;
 $string = $string|0;
 $radix = $radix|0;
 var $$01$i = 0, $$01$i2 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $a$01$i$i = 0, $a$01$i$i9 = 0, $b$02$i$i = 0, $b$02$i$i8 = 0, $bufp$02$i = 0, $bufp$02$i1 = 0, $bufp$1$i4 = 0, $len$03$i$i = 0, $len$03$i$i7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($value|0)<(0);
 if ($0) {
  $1 = ((($string)) + 1|0);
  HEAP8[$string>>0] = 45;
  $2 = (0 - ($value))|0;
  $$01$i = $2;$bufp$02$i = $1;
  while(1) {
   $3 = (($$01$i>>>0) % ($radix>>>0))&-1;
   $4 = (604599 + ($3)|0);
   $5 = HEAP8[$4>>0]|0;
   $6 = ((($bufp$02$i)) + 1|0);
   HEAP8[$bufp$02$i>>0] = $5;
   $7 = (($$01$i>>>0) / ($radix>>>0))&-1;
   $8 = ($7|0)==(0);
   if ($8) {
    $$lcssa = $6;
    break;
   } else {
    $$01$i = $7;$bufp$02$i = $6;
   }
  }
  HEAP8[$$lcssa>>0] = 0;
  $9 = (_strlen($1)|0);
  $10 = $9 >>> 1;
  $11 = ($10|0)==(0);
  if ($11) {
   return ($string|0);
  }
  $12 = (($9) + -1)|0;
  $13 = (($1) + ($12)|0);
  $a$01$i$i = $1;$b$02$i$i = $13;$len$03$i$i = $10;
  while(1) {
   $14 = (($len$03$i$i) + -1)|0;
   $15 = HEAP8[$a$01$i$i>>0]|0;
   $16 = HEAP8[$b$02$i$i>>0]|0;
   $17 = ((($a$01$i$i)) + 1|0);
   HEAP8[$a$01$i$i>>0] = $16;
   $18 = ((($b$02$i$i)) + -1|0);
   HEAP8[$b$02$i$i>>0] = $15;
   $19 = ($14|0)==(0);
   if ($19) {
    break;
   } else {
    $a$01$i$i = $17;$b$02$i$i = $18;$len$03$i$i = $14;
   }
  }
  return ($string|0);
 }
 $20 = ($value|0)==(0);
 if ($20) {
  $27 = ((($string)) + 1|0);
  HEAP8[$string>>0] = 48;
  $bufp$1$i4 = $27;
 } else {
  $$01$i2 = $value;$bufp$02$i1 = $string;
  while(1) {
   $21 = (($$01$i2>>>0) % ($radix>>>0))&-1;
   $22 = (604599 + ($21)|0);
   $23 = HEAP8[$22>>0]|0;
   $24 = ((($bufp$02$i1)) + 1|0);
   HEAP8[$bufp$02$i1>>0] = $23;
   $25 = (($$01$i2>>>0) / ($radix>>>0))&-1;
   $26 = ($25|0)==(0);
   if ($26) {
    $bufp$1$i4 = $24;
    break;
   } else {
    $$01$i2 = $25;$bufp$02$i1 = $24;
   }
  }
 }
 HEAP8[$bufp$1$i4>>0] = 0;
 $28 = (_strlen($string)|0);
 $29 = $28 >>> 1;
 $30 = ($29|0)==(0);
 if ($30) {
  return ($string|0);
 }
 $31 = (($28) + -1)|0;
 $32 = (($string) + ($31)|0);
 $a$01$i$i9 = $string;$b$02$i$i8 = $32;$len$03$i$i7 = $29;
 while(1) {
  $33 = (($len$03$i$i7) + -1)|0;
  $34 = HEAP8[$a$01$i$i9>>0]|0;
  $35 = HEAP8[$b$02$i$i8>>0]|0;
  $36 = ((($a$01$i$i9)) + 1|0);
  HEAP8[$a$01$i$i9>>0] = $35;
  $37 = ((($b$02$i$i8)) + -1|0);
  HEAP8[$b$02$i$i8>>0] = $34;
  $38 = ($33|0)==(0);
  if ($38) {
   break;
  } else {
   $a$01$i$i9 = $36;$b$02$i$i8 = $37;$len$03$i$i7 = $33;
  }
 }
 return ($string|0);
}
function _SDL_AddDisplayMode($display,$mode) {
 $display = $display|0;
 $mode = $mode|0;
 var $$0 = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $modes$0 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($display)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($display)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)>(0);
 L1: do {
  if ($4) {
   $5 = ((($mode)) + 4|0);
   $6 = ((($mode)) + 8|0);
   $7 = ((($mode)) + 12|0);
   $i$01 = 0;
   L3: while(1) {
    $10 = (($1) + (($i$01*20)|0)|0);
    $11 = ($10|0)==($mode|0);
    if ($11) {
     $$0 = 0;
     label = 20;
     break;
    }
    $12 = HEAP32[$5>>2]|0;
    $13 = (((($1) + (($i$01*20)|0)|0)) + 4|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = ($14|0)==($12|0);
    do {
     if ($15) {
      $17 = HEAP32[$6>>2]|0;
      $18 = (((($1) + (($i$01*20)|0)|0)) + 8|0);
      $19 = HEAP32[$18>>2]|0;
      $20 = ($19|0)==($17|0);
      if (!($20)) {
       $21 = (($19) - ($17))|0;
       $42 = $21;
       break;
      }
      $22 = HEAP32[$mode>>2]|0;
      $23 = $22 >>> 8;
      $24 = $23 & 255;
      $25 = HEAP32[$10>>2]|0;
      $26 = $25 >>> 8;
      $27 = $26 & 255;
      $28 = ($27|0)==($24|0);
      if (!($28)) {
       $29 = (($27) - ($24))|0;
       $42 = $29;
       break;
      }
      $30 = $22 >>> 16;
      $31 = $30 & 15;
      $32 = $25 >>> 16;
      $33 = $32 & 15;
      $34 = ($33|0)==($31|0);
      if ($34) {
       $36 = HEAP32[$7>>2]|0;
       $37 = (((($1) + (($i$01*20)|0)|0)) + 12|0);
       $38 = HEAP32[$37>>2]|0;
       $39 = ($38|0)==($36|0);
       $40 = (($38) - ($36))|0;
       if ($39) {
        $$0 = 0;
        label = 20;
        break L3;
       } else {
        $42 = $40;
        break;
       }
      } else {
       $35 = (($33) - ($31))|0;
       $42 = $35;
       break;
      }
     } else {
      $16 = (($14) - ($12))|0;
      $42 = $16;
     }
    } while(0);
    $41 = ($42|0)==(0);
    $9 = (($i$01) + 1)|0;
    if ($41) {
     $$0 = 0;
     label = 20;
     break;
    }
    $8 = ($9|0)<($3|0);
    if ($8) {
     $i$01 = $9;
    } else {
     break L1;
    }
   }
   if ((label|0) == 20) {
    return ($$0|0);
   }
  }
 } while(0);
 $43 = ((($display)) + 4|0);
 $44 = HEAP32[$43>>2]|0;
 $45 = ($3|0)==($44|0);
 do {
  if ($45) {
   $46 = ($3*20)|0;
   $47 = (($46) + 640)|0;
   $48 = (_SDL_realloc($1,$47)|0);
   $49 = ($48|0)==(0|0);
   if ($49) {
    $$0 = 0;
    return ($$0|0);
   } else {
    HEAP32[$0>>2] = $48;
    $50 = HEAP32[$43>>2]|0;
    $51 = (($50) + 32)|0;
    HEAP32[$43>>2] = $51;
    $$pre$phiZ2D = $0;$modes$0 = $48;
    break;
   }
  } else {
   $$pre$phiZ2D = $0;$modes$0 = $1;
  }
 } while(0);
 $52 = (($modes$0) + (($3*20)|0)|0);
 ;HEAP32[$52>>2]=HEAP32[$mode>>2]|0;HEAP32[$52+4>>2]=HEAP32[$mode+4>>2]|0;HEAP32[$52+8>>2]=HEAP32[$mode+8>>2]|0;HEAP32[$52+12>>2]=HEAP32[$mode+12>>2]|0;HEAP32[$52+16>>2]=HEAP32[$mode+16>>2]|0;
 $53 = HEAP32[$2>>2]|0;
 $54 = (($53) + 1)|0;
 HEAP32[$2>>2] = $54;
 $55 = HEAP32[$$pre$phiZ2D>>2]|0;
 _SDL_qsort($55,$54,20,4476);
 $$0 = 1;
 return ($$0|0);
}
function _Emscripten_InitMouse() {
 var $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GetMouse()|0);
 HEAP32[$0>>2] = 4587;
 $1 = ((($0)) + 8|0);
 HEAP32[$1>>2] = 4588;
 $2 = ((($0)) + 16|0);
 HEAP32[$2>>2] = 4589;
 $3 = ((($0)) + 20|0);
 HEAP32[$3>>2] = 4590;
 $4 = ((($0)) + 4|0);
 HEAP32[$4>>2] = 4591;
 $5 = ((($0)) + 28|0);
 HEAP32[$5>>2] = 4592;
 $6 = (_SDL_calloc(1,8)|0);
 $7 = ($6|0)==(0|0);
 if ($7) {
  (_SDL_Error(0)|0);
  $$0$i = $6;
  _SDL_SetDefaultCursor($$0$i);
  return;
 }
 $8 = (_SDL_calloc(1,4)|0);
 $9 = ($8|0)==(0|0);
 if ($9) {
  (_SDL_Error(0)|0);
  _SDL_free($6);
  $$0$i = 0;
  _SDL_SetDefaultCursor($$0$i);
  return;
 } else {
  HEAP32[$8>>2] = 604635;
  $10 = ((($6)) + 4|0);
  HEAP32[$10>>2] = $8;
  $$0$i = $6;
  _SDL_SetDefaultCursor($$0$i);
  return;
 }
}
function _Emscripten_CreateCursor($sruface,$hot_x,$hot_y) {
 $sruface = $sruface|0;
 $hot_x = $hot_x|0;
 $hot_y = $hot_y|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_calloc(1,8)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_Error(0)|0);
  $$0$i = $0;
  return ($$0$i|0);
 }
 $2 = (_SDL_calloc(1,4)|0);
 $3 = ($2|0)==(0|0);
 if ($3) {
  (_SDL_Error(0)|0);
  _SDL_free($0);
  $$0$i = 0;
  return ($$0$i|0);
 } else {
  HEAP32[$2>>2] = 604635;
  $4 = ((($0)) + 4|0);
  HEAP32[$4>>2] = $2;
  $$0$i = $0;
  return ($$0$i|0);
 }
 return (0)|0;
}
function _Emscripten_ShowCursor($cursor) {
 $cursor = $cursor|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GetMouseFocus()|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  return 0;
 }
 $2 = ($cursor|0)==(0|0);
 if (!($2)) {
  $3 = ((($cursor)) + 4|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ($4|0)==(0|0);
  if (!($5)) {
   $6 = HEAP32[$4>>2]|0;
   $7 = ($6|0)==(0|0);
   if ($7) {
    return 0;
   }
   $8 = _emscripten_asm_const_ii(3, ($6|0))|0;
   return 0;
  }
 }
 _emscripten_asm_const_v(4);
 return 0;
}
function _Emscripten_FreeCursor($cursor) {
 $cursor = $cursor|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($cursor|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($cursor)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
  _SDL_free($2);
 }
 _SDL_free($cursor);
 return;
}
function _Emscripten_WarpMouse($window,$x,$y) {
 $window = $window|0;
 $x = $x|0;
 $y = $y|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 (_SDL_Error(4)|0);
 return;
}
function _Emscripten_CreateSystemCursor($id) {
 $id = $id|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $cursor_name$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 do {
  switch ($id|0) {
  case 0:  {
   $cursor_name$0 = 604635;
   label = 13;
   break;
  }
  case 1:  {
   $cursor_name$0 = 604908;
   label = 13;
   break;
  }
  case 2:  {
   $cursor_name$0 = 604903;
   label = 13;
   break;
  }
  case 3:  {
   $cursor_name$0 = 604893;
   label = 13;
   break;
  }
  case 4:  {
   $cursor_name$0 = 604884;
   label = 13;
   break;
  }
  case 5:  {
   $cursor_name$0 = 604872;
   label = 13;
   break;
  }
  case 6:  {
   $cursor_name$0 = 604860;
   label = 13;
   break;
  }
  case 7:  {
   $cursor_name$0 = 604850;
   label = 13;
   break;
  }
  case 8:  {
   $cursor_name$0 = 604840;
   label = 13;
   break;
  }
  case 9:  {
   $cursor_name$0 = 0;
   label = 13;
   break;
  }
  case 10:  {
   $cursor_name$0 = 604828;
   label = 13;
   break;
  }
  case 11:  {
   $cursor_name$0 = 604820;
   label = 13;
   break;
  }
  default: {
   $$0 = 0;
  }
  }
 } while(0);
 do {
  if ((label|0) == 13) {
   $0 = (_SDL_calloc(1,8)|0);
   $1 = ($0|0)==(0|0);
   if ($1) {
    (_SDL_Error(0)|0);
    $$0 = 0;
    break;
   }
   $2 = (_SDL_calloc(1,4)|0);
   $3 = ($2|0)==(0|0);
   if ($3) {
    (_SDL_Error(0)|0);
    _SDL_free($0);
    $$0 = 0;
    break;
   } else {
    HEAP32[$2>>2] = $cursor_name$0;
    $4 = ((($0)) + 4|0);
    HEAP32[$4>>2] = $2;
    $$0 = $0;
    break;
   }
  }
 } while(0);
 return ($$0|0);
}
function _Emscripten_SetRelativeMouseMode($enabled) {
 $enabled = $enabled|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($enabled|0)==(0);
 if ($0) {
  $3 = (_emscripten_exit_pointerlock()|0);
  $4 = ($3|0)>(-1);
  if ($4) {
   $$0 = 0;
   return ($$0|0);
  }
 } else {
  $1 = (_emscripten_request_pointerlock((0|0),1)|0);
  $2 = ($1|0)>(-1);
  if ($2) {
   $$0 = 0;
   return ($$0|0);
  }
 }
 $$0 = -1;
 return ($$0|0);
}
function _SDL_SetDefaultCursor($cursor) {
 $cursor = $cursor|0;
 var $$01$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $notlhs$i = 0, $notrhs$i = 0, $or$cond3$not$i = 0, $or$cond5$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(1201872)>>2] = $cursor;
 $0 = HEAP32[(1201876)>>2]|0;
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  return;
 }
 $2 = ($cursor|0)==(0|0);
 if ($2) {
  $$01$i = 0;
 } else {
  HEAP32[(1201876)>>2] = $cursor;
  $$01$i = $cursor;
 }
 $3 = HEAP32[(1201880)>>2]|0;
 $notlhs$i = ($$01$i|0)!=(0|0);
 $notrhs$i = ($3|0)!=(0);
 $or$cond3$not$i = $notlhs$i & $notrhs$i;
 $4 = HEAP32[(1201852)>>2]|0;
 $5 = ($4|0)==(0);
 $or$cond5$i = $or$cond3$not$i & $5;
 $6 = HEAP32[(1201784)>>2]|0;
 $7 = ($6|0)==(0|0);
 if ($or$cond5$i) {
  if ($7) {
   return;
  }
  (FUNCTION_TABLE_ii[$6 & 8191]($$01$i)|0);
  return;
 } else {
  if ($7) {
   return;
  }
  (FUNCTION_TABLE_ii[$6 & 8191](0)|0);
  return;
 }
}
function _Emscripten_VideoQuit($_this) {
 $_this = $_this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _Emscripten_FiniMouse();
 return;
}
function _Emscripten_FiniMouse() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GetMouse()|0);
 $1 = ((($0)) + 96|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
  $4 = ((($2)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ($5|0)==(0|0);
  if (!($6)) {
   _SDL_free($5);
  }
  _SDL_free($2);
 }
 HEAP32[$1>>2] = 0;
 HEAP32[$0>>2] = 0;
 $7 = ((($0)) + 8|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($0)) + 16|0);
 HEAP32[$8>>2] = 0;
 $9 = ((($0)) + 20|0);
 HEAP32[$9>>2] = 0;
 $10 = ((($0)) + 4|0);
 HEAP32[$10>>2] = 0;
 $11 = ((($0)) + 28|0);
 HEAP32[$11>>2] = 0;
 return;
}
function _Emscripten_SetDisplayMode($_this,$display,$mode) {
 $_this = $_this|0;
 $display = $display|0;
 $mode = $mode|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _Emscripten_PumpEvents($_this) {
 $_this = $_this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _Emscripten_CreateWindow($_this,$window) {
 $_this = $_this|0;
 $window = $window|0;
 var $$0 = 0, $$not = 0, $$not$demorgan = 0, $$pre = 0, $$pre$phi1Z2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0;
 var $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0, $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0;
 var $40 = 0.0, $41 = 0, $42 = 0, $43 = 0.0, $44 = 0, $45 = 0.0, $46 = 0, $47 = 0, $48 = 0, $49 = 0.0, $5 = 0, $50 = 0, $51 = 0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $7 = 0.0, $8 = 0.0, $9 = 0, $brmerge = 0, $css_h = 0, $css_w = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp + 16|0;
 $css_w = sp + 8|0;
 $css_h = sp;
 $0 = (_SDL_calloc(1,28)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  $2 = (_SDL_Error(0)|0);
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ((($window)) + 48|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 & 8192;
 $6 = ($5|0)==(0);
 if ($6) {
  $10 = ((($0)) + 12|0);
  HEAPF32[$10>>2] = 1.0;
  $$pre$phi1Z2D = $10;$15 = 1.0;
 } else {
  $7 = (+_emscripten_get_device_pixel_ratio());
  $8 = $7;
  $9 = ((($0)) + 12|0);
  HEAPF32[$9>>2] = $8;
  $$pre$phi1Z2D = $9;$15 = $8;
 }
 $11 = ((($window)) + 24|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (+($12|0));
 $14 = $13 * $15;
 $16 = $14;
 $17 = (+_SDL_floor($16));
 $18 = ((($window)) + 28|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (+($19|0));
 $21 = +HEAPF32[$$pre$phi1Z2D>>2];
 $22 = $20 * $21;
 $23 = $22;
 $24 = (+_SDL_floor($23));
 $25 = (~~(($17)));
 $26 = (~~(($24)));
 _emscripten_set_canvas_size(($25|0),($26|0));
 (_emscripten_get_element_css_size((0|0),($css_w|0),($css_h|0))|0);
 $27 = +HEAPF64[$css_w>>3];
 $28 = $27 != $17;
 $29 = +HEAPF64[$css_h>>3];
 $30 = $29 != $24;
 $31 = $28 | $30;
 $32 = $31&1;
 $33 = ((($0)) + 16|0);
 HEAP32[$33>>2] = $32;
 $34 = HEAP32[$3>>2]|0;
 $35 = $34 & 32;
 $36 = ($35|0)==(0);
 $$not$demorgan = $28 | $30;
 $$not = $$not$demorgan ^ 1;
 $brmerge = $36 | $$not;
 if ($brmerge) {
  $48 = $32;
 } else {
  $37 = +HEAPF32[$$pre$phi1Z2D>>2];
  $38 = $37;
  $39 = $27 * $38;
  $40 = $29 * $38;
  $41 = (~~(($39)));
  $42 = (~~(($40)));
  _emscripten_set_canvas_size(($41|0),($42|0));
  $43 = +HEAPF64[$css_w>>3];
  $44 = (~~(($43)));
  $45 = +HEAPF64[$css_h>>3];
  $46 = (~~(($45)));
  (_SDL_SendWindowEvent($window,5,$44,$46)|0);
  $$pre = HEAP32[$33>>2]|0;
  $48 = $$pre;
 }
 $47 = ($48|0)==(0);
 if ($47) {
  $49 = +HEAPF32[$$pre$phi1Z2D>>2];
  $50 = $49 != 1.0;
  if ($50) {
   $51 = HEAP32[$11>>2]|0;
   $52 = (+($51|0));
   $53 = HEAP32[$18>>2]|0;
   $54 = (+($53|0));
   (_emscripten_set_element_css_size((0|0),(+$52),(+$54))|0);
  }
 }
 $55 = HEAP32[$3>>2]|0;
 $56 = $55 & 2;
 $57 = ($56|0)==(0);
 if (!($57)) {
  $58 = ((($_this)) + 652|0);
  $59 = HEAP32[$58>>2]|0;
  $60 = ($59|0)==(0|0);
  if ($60) {
   $61 = (_SDL_GL_LoadLibrary(0)|0);
   $62 = ($61|0)<(0);
   if ($62) {
    $$0 = -1;
    STACKTOP = sp;return ($$0|0);
   }
  }
  $63 = (_SDL_EGL_CreateSurface($_this,0)|0);
  HEAP32[$0>>2] = $63;
  $64 = ($63|0)==(0|0);
  if ($64) {
   $65 = (_SDL_SetError(605038,$vararg_buffer)|0);
   $$0 = $65;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $66 = ((($0)) + 4|0);
 HEAP32[$66>>2] = $window;
 $67 = ((($window)) + 136|0);
 HEAP32[$67>>2] = $0;
 _SDL_SetMouseFocus($window);
 _SDL_SetKeyboardFocus($window);
 _Emscripten_RegisterEventHandlers($0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GL_LoadLibrary($path) {
 $path = $path|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  $2 = (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ((($0)) + 364|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0);
 do {
  if ($5) {
   $11 = ((($0)) + 140|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($12|0)==(0|0);
   if ($13) {
    $14 = (_SDL_SetError(595987,$vararg_buffer3)|0);
    $$0 = $14;
    STACKTOP = sp;return ($$0|0);
   }
   $15 = (FUNCTION_TABLE_iii[$12 & 8191]($0,$path)|0);
   $16 = ($15|0)==(0);
   $17 = HEAP32[300441]|0;
   if ($16) {
    $19 = $17;
   } else {
    $22 = ((($17)) + 148|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = ($23|0)==(0|0);
    if ($24) {
     $$0 = $15;
     STACKTOP = sp;return ($$0|0);
    }
    FUNCTION_TABLE_vi[$23 & 8191]($17);
    $$0 = $15;
    STACKTOP = sp;return ($$0|0);
   }
  } else {
   $6 = ($path|0)==(0|0);
   if ($6) {
    $19 = $0;
   } else {
    $7 = ((($0)) + 368|0);
    $8 = (_SDL_strcmp($path,$7)|0);
    $9 = ($8|0)==(0);
    if ($9) {
     $$pre = HEAP32[300441]|0;
     $19 = $$pre;
     break;
    }
    $10 = (_SDL_SetError(604913,$vararg_buffer1)|0);
    $$0 = $10;
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 $18 = ((($19)) + 364|0);
 $20 = HEAP32[$18>>2]|0;
 $21 = (($20) + 1)|0;
 HEAP32[$18>>2] = $21;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_EGL_CreateSurface($_this,$nw) {
 $_this = $_this|0;
 $nw = $nw|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_EGL_ChooseConfig($_this)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  $$0 = 0;
  return ($$0|0);
 }
 $2 = ((($_this)) + 652|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 48|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($3)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($3)) + 12|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (FUNCTION_TABLE_iiiii[$5 & 8191]($7,$9,$nw,0)|0);
 $$0 = $10;
 return ($$0|0);
}
function _SDL_EGL_ChooseConfig($_this) {
 $_this = $_this|0;
 var $$ = 0, $$$sink = 0, $$0 = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $12 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $attribs = 0, $best_bitdiff$010 = 0, $best_bitdiff$1 = 0, $bitdiff$0$lcssa = 0, $bitdiff$07 = 0;
 var $bitdiff$1 = 0, $configs = 0, $found_configs = 0, $i$0 = 0, $i$1 = 0, $i$2 = 0, $i$3 = 0, $i$4 = 0, $i$5 = 0, $i$68 = 0, $j$06 = 0, $or$cond = 0, $or$cond13 = 0, $or$cond3 = 0, $or$cond4 = 0, $switch = 0, $value = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 800|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $attribs = sp + 536|0;
 $found_configs = sp + 532|0;
 $value = sp + 528|0;
 $configs = sp + 16|0;
 HEAP32[$found_configs>>2] = 0;
 $0 = ((($_this)) + 652|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$attribs>>2] = 12324;
 $3 = ((($_this)) + 268|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($attribs)) + 4|0);
 HEAP32[$5>>2] = $4;
 $6 = ((($attribs)) + 8|0);
 HEAP32[$6>>2] = 12323;
 $7 = ((($_this)) + 272|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($attribs)) + 12|0);
 HEAP32[$9>>2] = $8;
 $10 = ((($attribs)) + 16|0);
 HEAP32[$10>>2] = 12322;
 $11 = ((($_this)) + 276|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($attribs)) + 20|0);
 HEAP32[$13>>2] = $12;
 $14 = ((($_this)) + 280|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)==(0);
 if ($16) {
  $i$0 = 6;
 } else {
  $17 = ((($attribs)) + 24|0);
  HEAP32[$17>>2] = 12321;
  $18 = ((($attribs)) + 28|0);
  HEAP32[$18>>2] = $15;
  $i$0 = 8;
 }
 $19 = ((($_this)) + 288|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ($20|0)==(0);
 if ($21) {
  $i$1 = $i$0;
 } else {
  $22 = $i$0 | 1;
  $23 = (($attribs) + ($i$0<<2)|0);
  HEAP32[$23>>2] = 12320;
  $24 = (($22) + 1)|0;
  $25 = (($attribs) + ($22<<2)|0);
  HEAP32[$25>>2] = $20;
  $i$1 = $24;
 }
 $26 = (($i$1) + 1)|0;
 $27 = (($attribs) + ($i$1<<2)|0);
 HEAP32[$27>>2] = 12325;
 $28 = ((($_this)) + 284|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = (($i$1) + 2)|0;
 $31 = (($attribs) + ($26<<2)|0);
 HEAP32[$31>>2] = $29;
 $32 = ((($_this)) + 292|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ($33|0)==(0);
 if ($34) {
  $i$2 = $30;
 } else {
  $35 = (($i$1) + 3)|0;
  $36 = (($attribs) + ($30<<2)|0);
  HEAP32[$36>>2] = 12326;
  $37 = (($i$1) + 4)|0;
  $38 = (($attribs) + ($35<<2)|0);
  HEAP32[$38>>2] = $33;
  $i$2 = $37;
 }
 $39 = ((($_this)) + 320|0);
 $40 = HEAP32[$39>>2]|0;
 $41 = ($40|0)==(0);
 if ($41) {
  $i$3 = $i$2;
 } else {
  $42 = (($i$2) + 1)|0;
  $43 = (($attribs) + ($i$2<<2)|0);
  HEAP32[$43>>2] = 12338;
  $44 = (($i$2) + 2)|0;
  $45 = (($attribs) + ($42<<2)|0);
  HEAP32[$45>>2] = $40;
  $i$3 = $44;
 }
 $46 = ((($_this)) + 324|0);
 $47 = HEAP32[$46>>2]|0;
 $48 = ($47|0)==(0);
 if ($48) {
  $i$4 = $i$3;
 } else {
  $49 = (($i$3) + 1)|0;
  $50 = (($attribs) + ($i$3<<2)|0);
  HEAP32[$50>>2] = 12337;
  $51 = (($i$3) + 2)|0;
  $52 = (($attribs) + ($49<<2)|0);
  HEAP32[$52>>2] = $47;
  $i$4 = $51;
 }
 $53 = ((($_this)) + 356|0);
 $54 = HEAP32[$53>>2]|0;
 $55 = ($54|0)==(0);
 if (!($55)) {
  $56 = (_SDL_SetError(604943,$vararg_buffer)|0);
  $$0 = $56;
  STACKTOP = sp;return ($$0|0);
 }
 $57 = (($i$4) + 1)|0;
 $58 = (($attribs) + ($i$4<<2)|0);
 HEAP32[$58>>2] = 12352;
 $59 = ((($_this)) + 344|0);
 $60 = HEAP32[$59>>2]|0;
 $61 = ($60|0)==(4);
 if ($61) {
  $62 = ((($_this)) + 332|0);
  $63 = HEAP32[$62>>2]|0;
  $64 = ($63|0)>(1);
  $$ = $64 ? 4 : 1;
  $65 = ((($1)) + 84|0);
  $66 = HEAP32[$65>>2]|0;
  (FUNCTION_TABLE_ii[$66 & 8191](12448)|0);
  $$$sink = $$;
 } else {
  $67 = ((($1)) + 84|0);
  $68 = HEAP32[$67>>2]|0;
  (FUNCTION_TABLE_ii[$68 & 8191](12450)|0);
  $$$sink = 8;
 }
 $69 = (($attribs) + ($57<<2)|0);
 HEAP32[$69>>2] = $$$sink;
 $i$5 = (($i$4) + 2)|0;
 $70 = (($attribs) + ($i$5<<2)|0);
 HEAP32[$70>>2] = 12344;
 $71 = HEAP32[$0>>2]|0;
 $72 = ((($71)) + 36|0);
 $73 = HEAP32[$72>>2]|0;
 $74 = ((($71)) + 8|0);
 $75 = HEAP32[$74>>2]|0;
 $76 = (FUNCTION_TABLE_iiiiii[$73 & 8191]($75,$attribs,$configs,128,$found_configs)|0);
 $77 = ($76|0)==(0);
 $78 = HEAP32[$found_configs>>2]|0;
 $79 = ($78|0)==(0);
 $or$cond = $77 | $79;
 if ($or$cond) {
  $81 = (_SDL_SetError(605004,$vararg_buffer1)|0);
  $$0 = $81;
  STACKTOP = sp;return ($$0|0);
 }
 $80 = ($78|0)>(0);
 if ($80) {
  $best_bitdiff$010 = -1;$i$68 = 0;
 } else {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 while(1) {
  $82 = (($configs) + ($i$68<<2)|0);
  $bitdiff$07 = 0;$j$06 = 0;
  while(1) {
   $83 = (($attribs) + ($j$06<<2)|0);
   $84 = HEAP32[$83>>2]|0;
   $85 = ($84|0)==(12344);
   if ($85) {
    $bitdiff$0$lcssa = $bitdiff$07;
    break;
   }
   $86 = $j$06 | 1;
   $87 = (($attribs) + ($86<<2)|0);
   $88 = HEAP32[$87>>2]|0;
   $89 = ($88|0)!=(-1);
   $$off = (($84) + -12321)|0;
   $switch = ($$off>>>0)<(6);
   $or$cond4 = $switch & $89;
   if ($or$cond4) {
    $90 = HEAP32[$0>>2]|0;
    $91 = ((($90)) + 72|0);
    $92 = HEAP32[$91>>2]|0;
    $93 = ((($90)) + 8|0);
    $94 = HEAP32[$93>>2]|0;
    $95 = HEAP32[$82>>2]|0;
    (FUNCTION_TABLE_iiiii[$92 & 8191]($94,$95,$84,$value)|0);
    $96 = HEAP32[$value>>2]|0;
    $97 = HEAP32[$87>>2]|0;
    $98 = (($96) + ($bitdiff$07))|0;
    $99 = (($98) - ($97))|0;
    $bitdiff$1 = $99;
   } else {
    $bitdiff$1 = $bitdiff$07;
   }
   $100 = (($j$06) + 2)|0;
   $101 = ($100>>>0)<(63);
   if ($101) {
    $bitdiff$07 = $bitdiff$1;$j$06 = $100;
   } else {
    $bitdiff$0$lcssa = $bitdiff$1;
    break;
   }
  }
  $102 = ($bitdiff$0$lcssa|0)<($best_bitdiff$010|0);
  $103 = ($best_bitdiff$010|0)==(-1);
  $or$cond3 = $103 | $102;
  if ($or$cond3) {
   $104 = HEAP32[$82>>2]|0;
   $105 = HEAP32[$0>>2]|0;
   $106 = ((($105)) + 12|0);
   HEAP32[$106>>2] = $104;
   $best_bitdiff$1 = $bitdiff$0$lcssa;
  } else {
   $best_bitdiff$1 = $best_bitdiff$010;
  }
  $107 = ($bitdiff$0$lcssa|0)!=(0);
  $108 = (($i$68) + 1)|0;
  $109 = HEAP32[$found_configs>>2]|0;
  $110 = ($108|0)<($109|0);
  $or$cond13 = $107 & $110;
  if ($or$cond13) {
   $best_bitdiff$010 = $best_bitdiff$1;$i$68 = $108;
  } else {
   $$0 = 0;
   break;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SetKeyboardFocus($window) {
 $window = $window|0;
 var $$pre = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i = 0, $or$cond = 0, $or$cond2 = 0, $scancode$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[300740]|0;
 $1 = ($0|0)==(0|0);
 $2 = ($window|0)!=(0|0);
 $or$cond = $2 | $1;
 if ($or$cond) {
  $8 = $0;
 } else {
  $scancode$01$i = 0;
  while(1) {
   $3 = (1202966 + ($scancode$01$i)|0);
   $4 = HEAP8[$3>>0]|0;
   $5 = ($4<<24>>24)==(1);
   if ($5) {
    (_SDL_SendKeyboardKey(0,$scancode$01$i)|0);
   }
   $6 = (($scancode$01$i) + 1)|0;
   $exitcond$i = ($6|0)==(512);
   if ($exitcond$i) {
    break;
   } else {
    $scancode$01$i = $6;
   }
  }
  $$pre = HEAP32[300740]|0;
  $8 = $$pre;
 }
 $7 = ($8|0)==(0|0);
 $9 = ($8|0)==($window|0);
 $or$cond2 = $7 | $9;
 if (!($or$cond2)) {
  $10 = ((($8)) + 48|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = $11 & 16384;
  $13 = ($12|0)==(0);
  if ($13) {
   $14 = $8;
  } else {
   (_SDL_CaptureMouse(0)|0);
   $$pre3 = HEAP32[300740]|0;
   $14 = $$pre3;
  }
  (_SDL_SendWindowEvent($14,13,0,0)|0);
  $15 = (_SDL_EventState(771,-1)|0);
  $16 = ($15<<24>>24)==(0);
  if (!($16)) {
   $17 = (_SDL_GetVideoDevice()|0);
   $18 = ($17|0)==(0|0);
   if (!($18)) {
    $19 = ((($17)) + 192|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = ($20|0)==(0|0);
    if (!($21)) {
     FUNCTION_TABLE_vi[$20 & 8191]($17);
    }
   }
  }
 }
 HEAP32[300740] = $window;
 if (!($2)) {
  return;
 }
 (_SDL_SendWindowEvent($window,12,0,0)|0);
 $22 = (_SDL_EventState(771,-1)|0);
 $23 = ($22<<24>>24)==(0);
 if ($23) {
  return;
 }
 $24 = (_SDL_GetVideoDevice()|0);
 $25 = ($24|0)==(0|0);
 if ($25) {
  return;
 }
 $26 = ((($24)) + 188|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ($27|0)==(0|0);
 if ($28) {
  return;
 }
 FUNCTION_TABLE_vi[$27 & 8191]($24);
 return;
}
function _SDL_SendKeyboardKey($state,$scancode) {
 $state = $state|0;
 $scancode = $scancode|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $event = 0, $modifier$0 = 0, $modstate$0 = 0, $or$cond = 0, $phitmp = 0, $type$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $event = sp;
 $0 = ($scancode|0)==(0);
 if ($0) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = $state&255;
 switch ($1|0) {
 case 1:  {
  $type$0 = 768;
  break;
 }
 case 0:  {
  $type$0 = 769;
  break;
 }
 default: {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 }
 $2 = ($state<<24>>24)==(0);
 $$phi$trans$insert = (1202966 + ($scancode)|0);
 $$pre = HEAP8[$$phi$trans$insert>>0]|0;
 $3 = ($$pre<<24>>24)!=(0);
 $phitmp = $3&1;
 $4 = $2 ? 0 : $phitmp;
 $5 = ($$pre<<24>>24)!=($state<<24>>24);
 $6 = ($4<<24>>24)!=(0);
 $or$cond = $6 | $5;
 if (!($or$cond)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP8[$$phi$trans$insert>>0] = $state;
 $7 = (1203480 + ($scancode<<2)|0);
 $8 = HEAP32[$7>>2]|0;
 switch ($8|0) {
 case 1073742048:  {
  $modifier$0 = 64;
  break;
 }
 case 1073742052:  {
  $modifier$0 = 128;
  break;
 }
 case 1073742049:  {
  $modifier$0 = 1;
  break;
 }
 case 1073742053:  {
  $modifier$0 = 2;
  break;
 }
 case 1073742050:  {
  $modifier$0 = 256;
  break;
 }
 case 1073742054:  {
  $modifier$0 = 512;
  break;
 }
 case 1073742051:  {
  $modifier$0 = 1024;
  break;
 }
 case 1073742055:  {
  $modifier$0 = 2048;
  break;
 }
 case 1073742081:  {
  $modifier$0 = 16384;
  break;
 }
 default: {
  $modifier$0 = 0;
 }
 }
 $9 = ($type$0|0)==(768);
 L22: do {
  if ($9) {
   $10 = HEAP16[(1202964)>>1]|0;
   switch ($8|0) {
   case 1073741907:  {
    $11 = $10&65535;
    $12 = $11 ^ 4096;
    $13 = $12&65535;
    HEAP16[(1202964)>>1] = $13;
    $modstate$0 = $10;
    break L22;
    break;
   }
   case 1073741881:  {
    $14 = $10&65535;
    $15 = $14 ^ 8192;
    $16 = $15&65535;
    HEAP16[(1202964)>>1] = $16;
    $modstate$0 = $10;
    break L22;
    break;
   }
   default: {
    $17 = $10&65535;
    $18 = $17 | $modifier$0;
    $19 = $18&65535;
    HEAP16[(1202964)>>1] = $19;
    $modstate$0 = $10;
    break L22;
   }
   }
  } else {
   $20 = $modifier$0 ^ 65535;
   $21 = HEAP16[(1202964)>>1]|0;
   $22 = $21&65535;
   $23 = $22 & $20;
   $24 = $23&65535;
   HEAP16[(1202964)>>1] = $24;
   $modstate$0 = $24;
  }
 } while(0);
 $25 = (_SDL_EventState($type$0,-1)|0);
 $26 = ($25<<24>>24)==(1);
 if (!($26)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$event>>2] = $type$0;
 $27 = ((($event)) + 12|0);
 HEAP8[$27>>0] = $state;
 $28 = ((($event)) + 13|0);
 HEAP8[$28>>0] = $4;
 $29 = ((($event)) + 16|0);
 HEAP32[$29>>2] = $scancode;
 $30 = ((($29)) + 4|0);
 HEAP32[$30>>2] = $8;
 $31 = ((($29)) + 8|0);
 HEAP16[$31>>1] = $modstate$0;
 $32 = HEAP32[300740]|0;
 $33 = ($32|0)==(0|0);
 if ($33) {
  $37 = 0;
 } else {
  $34 = ((($32)) + 4|0);
  $35 = HEAP32[$34>>2]|0;
  $37 = $35;
 }
 $36 = ((($event)) + 8|0);
 HEAP32[$36>>2] = $37;
 $38 = (_SDL_PushEvent($event)|0);
 $39 = ($38|0)>(0);
 $40 = $39&1;
 $$0 = $40;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_CaptureMouse($enabled) {
 $enabled = $enabled|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP32[(1201808)>>2]|0;
 $1 = ($0|0)==(0|0);
 do {
  if ($1) {
   $2 = (_SDL_Error(4)|0);
   $$0 = $2;
  } else {
   $3 = (_SDL_GetKeyboardFocus()|0);
   $4 = ($3|0)!=(0|0);
   if ($4) {
    $5 = ((($3)) + 48|0);
    $6 = HEAP32[$5>>2]|0;
    $7 = $6 & 16384;
    $8 = ($7|0)!=(0);
    $10 = $8;
   } else {
    $10 = 0;
   }
   $9 = $10&1;
   $11 = ($9|0)==($enabled|0);
   if ($11) {
    $$0 = 0;
   } else {
    $12 = ($enabled|0)==(0);
    if ($12) {
     $20 = HEAP32[(1201808)>>2]|0;
     $21 = (FUNCTION_TABLE_ii[$20 & 8191](0)|0);
     $22 = ($21|0)==(-1);
     if ($22) {
      $$0 = -1;
      break;
     }
     $23 = ((($3)) + 48|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = $24 & -16385;
     HEAP32[$23>>2] = $25;
     $$0 = 0;
     break;
    }
    if (!($4)) {
     $13 = (_SDL_SetError(605075,$vararg_buffer)|0);
     $$0 = $13;
     break;
    }
    $14 = HEAP32[(1201808)>>2]|0;
    $15 = (FUNCTION_TABLE_ii[$14 & 8191]($3)|0);
    $16 = ($15|0)==(-1);
    if ($16) {
     $$0 = -1;
    } else {
     $17 = ((($3)) + 48|0);
     $18 = HEAP32[$17>>2]|0;
     $19 = $18 | 16384;
     HEAP32[$17>>2] = $19;
     $$0 = 0;
    }
   }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GetKeyboardFocus() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[300740]|0;
 return ($0|0);
}
function _SDL_GetVideoDevice() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[300441]|0;
 return ($0|0);
}
function _Emscripten_RegisterEventHandlers($data) {
 $data = $data|0;
 var $$ = 0, $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 (_emscripten_set_mousemove_callback((605095|0),($data|0),0,(4593|0))|0);
 (_emscripten_set_mousedown_callback((605095|0),($data|0),0,(4594|0))|0);
 (_emscripten_set_mouseup_callback((605103|0),($data|0),0,(4594|0))|0);
 (_emscripten_set_mouseenter_callback((605095|0),($data|0),0,(4595|0))|0);
 (_emscripten_set_mouseleave_callback((605095|0),($data|0),0,(4595|0))|0);
 (_emscripten_set_wheel_callback((605095|0),($data|0),0,(4596|0))|0);
 (_emscripten_set_focus_callback((605113|0),($data|0),0,(4597|0))|0);
 (_emscripten_set_blur_callback((605113|0),($data|0),0,(4597|0))|0);
 (_emscripten_set_touchstart_callback((605095|0),($data|0),0,(4598|0))|0);
 (_emscripten_set_touchend_callback((605095|0),($data|0),0,(4598|0))|0);
 (_emscripten_set_touchmove_callback((605095|0),($data|0),0,(4598|0))|0);
 (_emscripten_set_touchcancel_callback((605095|0),($data|0),0,(4598|0))|0);
 $0 = (_SDL_GetHint(605142)|0);
 $1 = ($0|0)!=(0|0);
 $$ = $1 ? $0 : 605113;
 (_emscripten_set_keydown_callback(($$|0),($data|0),0,(4599|0))|0);
 (_emscripten_set_keyup_callback(($$|0),($data|0),0,(4599|0))|0);
 (_emscripten_set_keypress_callback(($$|0),($data|0),0,(4600|0))|0);
 (_emscripten_set_fullscreenchange_callback((605103|0),($data|0),0,(4601|0))|0);
 (_emscripten_set_resize_callback((605113|0),($data|0),0,(4602|0))|0);
 (_emscripten_set_visibilitychange_callback(($data|0),0,(4603|0))|0);
 return;
}
function _Emscripten_HandleMouseMove($eventType,$mouseEvent,$userData) {
 $eventType = $eventType|0;
 $mouseEvent = $mouseEvent|0;
 $userData = $userData|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0, $25 = 0.0, $26 = 0.0;
 var $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0.0, $35 = 0, $36 = 0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0.0, $40 = 0, $41 = 0, $42 = 0.0, $43 = 0.0, $44 = 0;
 var $45 = 0, $5 = 0.0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0.0, $client_h = 0, $client_w = 0, $mx$0 = 0, $my$0 = 0, $pointerlock_status = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 288|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $pointerlock_status = sp + 16|0;
 $client_w = sp + 8|0;
 $client_h = sp;
 (_emscripten_get_element_css_size((0|0),($client_w|0),($client_h|0))|0);
 $0 = ((($userData)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 24|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (+($3|0));
 $5 = +HEAPF64[$client_w>>3];
 $6 = $4 / $5;
 $7 = ((($1)) + 28|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (+($8|0));
 $10 = +HEAPF64[$client_h>>3];
 $11 = $9 / $10;
 $12 = (_emscripten_get_pointerlock_status(($pointerlock_status|0))|0);
 $13 = ($12|0)==(0);
 $14 = HEAP32[$pointerlock_status>>2]|0;
 $15 = $13 ? $14 : 0;
 $16 = ($15|0)==(0);
 if ($16) {
  $35 = ((($mouseEvent)) + 60|0);
  $36 = HEAP32[$35>>2]|0;
  $37 = (+($36|0));
  $38 = $6 * $37;
  $39 = (~~(($38)));
  $40 = ((($mouseEvent)) + 64|0);
  $41 = HEAP32[$40>>2]|0;
  $42 = (+($41|0));
  $43 = $11 * $42;
  $44 = (~~(($43)));
  $mx$0 = $39;$my$0 = $44;
  $45 = HEAP32[$0>>2]|0;
  (_SDL_SendMouseMotion($45,0,$15,$mx$0,$my$0)|0);
  STACKTOP = sp;return 0;
 } else {
  $17 = ((($mouseEvent)) + 44|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = (+($18|0));
  $20 = $6 * $19;
  $21 = +HEAPF64[78115];
  $22 = $21 + $20;
  $23 = ((($mouseEvent)) + 48|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = (+($24|0));
  $26 = $11 * $25;
  $27 = +HEAPF64[78116];
  $28 = $27 + $26;
  $29 = (~~(($22)));
  $30 = (+($29|0));
  $31 = $22 - $30;
  HEAPF64[78115] = $31;
  $32 = (~~(($28)));
  $33 = (+($32|0));
  $34 = $28 - $33;
  HEAPF64[78116] = $34;
  $mx$0 = $29;$my$0 = $32;
  $45 = HEAP32[$0>>2]|0;
  (_SDL_SendMouseMotion($45,0,$15,$mx$0,$my$0)|0);
  STACKTOP = sp;return 0;
 }
 return (0)|0;
}
function _SDL_SendMouseMotion($window,$mouseID,$relative,$x,$y) {
 $window = $window|0;
 $mouseID = $mouseID|0;
 $relative = $relative|0;
 $x = $x|0;
 $y = $y|0;
 var $$1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $not$ = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($window|0)==(0|0);
 $1 = ($relative|0)!=(0);
 $or$cond = $0 | $1;
 if (!($or$cond)) {
  $2 = (_SDL_UpdateMouseFocus($window,$x,$y)|0);
  $not$ = ($2|0)==(0);
  if ($not$) {
   $$1 = 0;
   return ($$1|0);
  }
 }
 $3 = (_SDL_PrivateSendMouseMotion($window,$mouseID,$relative,$x,$y)|0);
 $$1 = $3;
 return ($$1|0);
}
function _Emscripten_HandleMouseButton($eventType,$mouseEvent,$userData) {
 $eventType = $eventType|0;
 $mouseEvent = $mouseEvent|0;
 $userData = $userData|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $switch$idx$cast = 0, $switch$offset = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($mouseEvent)) + 40|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = ($1&65535)<(3);
 if (!($2)) {
  $$0 = 0;
  return ($$0|0);
 }
 $switch$idx$cast = $1&255;
 $switch$offset = (($switch$idx$cast) + 1)<<24>>24;
 $3 = ($eventType|0)==(5);
 $4 = $3&1;
 $5 = ((($userData)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $3&1;
 (_SDL_SendMouseButton($6,0,$7,$switch$offset)|0);
 $8 = (_SDL_EventState($4,-1)|0);
 $9 = ($8<<24>>24)==(1);
 $10 = $9&1;
 $$0 = $10;
 return ($$0|0);
}
function _SDL_SendMouseButton($window,$mouseID,$state,$button) {
 $window = $window|0;
 $mouseID = $mouseID|0;
 $state = $state|0;
 $button = $button|0;
 var $$0 = 0, $$1$i = 0, $$phi$trans$insert = 0, $$phi$trans$insert4 = 0, $$pre = 0, $$pre$i = 0, $$pre$phi913Z2D = 0, $$pre5 = 0, $$pre6$pre$phiZZ2D = 0, $$pre7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $9 = 0, $buttonstate$0 = 0, $click_count$0 = 0, $event = 0, $exitcond = 0, $i$01$i2 = 0, $or$cond = 0, $or$cond1 = 0, $posted$0 = 0, $type$0 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $event = sp;
 $0 = HEAP32[(1201848)>>2]|0;
 $1 = $button&255;
 $2 = HEAP32[(1201860)>>2]|0;
 $3 = ($1|0)<($2|0);
 if ($3) {
  label = 7;
 } else {
  $4 = (($1) + 1)|0;
  $5 = HEAP32[(1201864)>>2]|0;
  $6 = $4 << 4;
  $7 = (_SDL_realloc($5,$6)|0);
  $8 = ($7|0)==(0|0);
  if ($8) {
   $$1$i = 0;
  } else {
   HEAP32[(1201864)>>2] = $7;
   $9 = HEAP32[(1201860)>>2]|0;
   $10 = ($9|0)>($1|0);
   if (!($10)) {
    $11 = (($7) + ($9<<4)|0);
    (_SDL_memset($11,0,16)|0);
    $12 = ($9|0)<($1|0);
    if ($12) {
     $i$01$i2 = $9;
     while(1) {
      $13 = (($i$01$i2) + 1)|0;
      $$pre$i = HEAP32[(1201864)>>2]|0;
      $14 = (($$pre$i) + ($13<<4)|0);
      (_SDL_memset($14,0,16)|0);
      $exitcond = ($13|0)==($1|0);
      if ($exitcond) {
       break;
      } else {
       $i$01$i2 = $13;
      }
     }
    }
   }
   HEAP32[(1201860)>>2] = $4;
   label = 7;
  }
 }
 if ((label|0) == 7) {
  $15 = HEAP32[(1201864)>>2]|0;
  $16 = (($15) + ($1<<4)|0);
  $$1$i = $16;
 }
 $17 = $state&255;
 switch ($17|0) {
 case 1:  {
  $18 = (($1) + -1)|0;
  $19 = 1 << $18;
  $20 = $0 | $19;
  $buttonstate$0 = $20;$type$0 = 1025;
  break;
 }
 case 0:  {
  $21 = (($1) + -1)|0;
  $22 = 1 << $21;
  $23 = $22 ^ -1;
  $24 = $0 & $23;
  $buttonstate$0 = $24;$type$0 = 1026;
  break;
 }
 default: {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 }
 $25 = ($window|0)!=(0|0);
 $26 = ($state<<24>>24)==(1);
 $or$cond = $25 & $26;
 if ($or$cond) {
  $27 = HEAP32[(1201824)>>2]|0;
  $28 = HEAP32[(1201828)>>2]|0;
  (_SDL_UpdateMouseFocus($window,$27,$28)|0);
 }
 $29 = HEAP32[(1201848)>>2]|0;
 $30 = ($buttonstate$0|0)==($29|0);
 if ($30) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[(1201848)>>2] = $buttonstate$0;
 $31 = ($$1$i|0)==(0|0);
 do {
  if ($31) {
   $click_count$0 = 1;
  } else {
   if (!($26)) {
    $$phi$trans$insert4 = ((($$1$i)) + 12|0);
    $$pre5 = HEAP8[$$phi$trans$insert4>>0]|0;
    $click_count$0 = $$pre5;
    break;
   }
   $32 = (_SDL_GetTicks()|0);
   $33 = ((($$1$i)) + 8|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = HEAP32[34485]|0;
   $36 = (($34) - ($32))|0;
   $37 = (($36) + ($35))|0;
   $38 = ($37|0)<(1);
   if ($38) {
    $$pre6$pre$phiZZ2D = $$1$i;
    label = 21;
   } else {
    $39 = HEAP32[(1201824)>>2]|0;
    $40 = HEAP32[$$1$i>>2]|0;
    $41 = (($39) - ($40))|0;
    $42 = (_SDL_abs($41)|0);
    $43 = ($42|0)>(1);
    if ($43) {
     $$pre6$pre$phiZZ2D = $$1$i;
     label = 21;
    } else {
     $44 = HEAP32[(1201828)>>2]|0;
     $45 = ((($$1$i)) + 4|0);
     $46 = HEAP32[$45>>2]|0;
     $47 = (($44) - ($46))|0;
     $48 = (_SDL_abs($47)|0);
     $49 = ($48|0)>(1);
     if ($49) {
      $$pre6$pre$phiZZ2D = $$1$i;
      label = 21;
     } else {
      $$phi$trans$insert = ((($$1$i)) + 12|0);
      $$pre = HEAP8[$$phi$trans$insert>>0]|0;
      HEAP32[$33>>2] = $32;
      $53 = HEAP32[(1201824)>>2]|0;
      HEAP32[$$1$i>>2] = $53;
      $54 = HEAP32[(1201828)>>2]|0;
      HEAP32[$45>>2] = $54;
      $55 = ($$pre<<24>>24)==(-1);
      if ($55) {
       $click_count$0 = -1;
       break;
      } else {
       $$pre$phi913Z2D = $$phi$trans$insert;$57 = $$pre;
      }
     }
    }
   }
   if ((label|0) == 21) {
    $50 = ((($$1$i)) + 12|0);
    HEAP8[$50>>0] = 0;
    $$pre7 = ((($$1$i)) + 4|0);
    HEAP32[$33>>2] = $32;
    $51 = HEAP32[(1201824)>>2]|0;
    HEAP32[$$pre6$pre$phiZZ2D>>2] = $51;
    $52 = HEAP32[(1201828)>>2]|0;
    HEAP32[$$pre7>>2] = $52;
    $$pre$phi913Z2D = $50;$57 = 0;
   }
   $56 = (($57) + 1)<<24>>24;
   HEAP8[$$pre$phi913Z2D>>0] = $56;
   $click_count$0 = $56;
  }
 } while(0);
 $58 = (_SDL_EventState($type$0,-1)|0);
 $59 = ($58<<24>>24)==(1);
 if ($59) {
  HEAP32[$event>>2] = $type$0;
  $60 = HEAP32[(1201820)>>2]|0;
  $61 = ($60|0)==(0|0);
  if ($61) {
   $65 = 0;
  } else {
   $62 = ((($60)) + 4|0);
   $63 = HEAP32[$62>>2]|0;
   $65 = $63;
  }
  $64 = ((($event)) + 8|0);
  HEAP32[$64>>2] = $65;
  $66 = ((($event)) + 12|0);
  HEAP32[$66>>2] = $mouseID;
  $67 = ((($event)) + 17|0);
  HEAP8[$67>>0] = $state;
  $68 = ((($event)) + 16|0);
  HEAP8[$68>>0] = $button;
  $69 = ((($event)) + 18|0);
  HEAP8[$69>>0] = $click_count$0;
  $70 = HEAP32[(1201824)>>2]|0;
  $71 = ((($event)) + 20|0);
  HEAP32[$71>>2] = $70;
  $72 = HEAP32[(1201828)>>2]|0;
  $73 = ((($event)) + 24|0);
  HEAP32[$73>>2] = $72;
  $74 = (_SDL_PushEvent($event)|0);
  $75 = ($74|0)>(0);
  $76 = $75&1;
  $posted$0 = $76;
 } else {
  $posted$0 = 0;
 }
 $77 = ($state<<24>>24)==(0);
 $or$cond1 = $25 & $77;
 if (!($or$cond1)) {
  $$0 = $posted$0;
  STACKTOP = sp;return ($$0|0);
 }
 $78 = HEAP32[(1201824)>>2]|0;
 $79 = HEAP32[(1201828)>>2]|0;
 (_SDL_UpdateMouseFocus($window,$78,$79)|0);
 $$0 = $posted$0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_abs($x) {
 $x = $x|0;
 var $0 = 0, $ispos = 0, $neg = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $ispos = ($x|0)>(-1);
 $neg = (0 - ($x))|0;
 $0 = $ispos ? $x : $neg;
 return ($0|0);
}
function _Emscripten_HandleMouseFocus($eventType,$mouseEvent,$userData) {
 $eventType = $eventType|0;
 $mouseEvent = $mouseEvent|0;
 $userData = $userData|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0.0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0, $client_h = 0, $client_w = 0, $not$ = 0, $pointerlock_status = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 288|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $pointerlock_status = sp + 16|0;
 $client_w = sp + 8|0;
 $client_h = sp;
 $0 = ((($mouseEvent)) + 60|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($mouseEvent)) + 64|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (_emscripten_get_pointerlock_status(($pointerlock_status|0))|0);
 $5 = HEAP32[$pointerlock_status>>2]|0;
 $6 = ($5|0)==(0);
 $not$ = ($4|0)!=(0);
 $7 = $6 | $not$;
 if ($7) {
  (_emscripten_get_element_css_size((0|0),($client_w|0),($client_h|0))|0);
  $8 = (+($1|0));
  $9 = ((($userData)) + 4|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ((($10)) + 24|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = (+($12|0));
  $14 = +HEAPF64[$client_w>>3];
  $15 = $13 / $14;
  $16 = $8 * $15;
  $17 = (~~(($16)));
  $18 = (+($3|0));
  $19 = ((($10)) + 28|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = (+($20|0));
  $22 = +HEAPF64[$client_h>>3];
  $23 = $21 / $22;
  $24 = $18 * $23;
  $25 = (~~(($24)));
  (_SDL_SendMouseMotion($10,0,0,$17,$25)|0);
 }
 $26 = ($eventType|0)==(33);
 if (!($26)) {
  $29 = 0;
  _SDL_SetMouseFocus($29);
  $30 = (_SDL_EventState(512,-1)|0);
  $31 = ($30<<24>>24)==(1);
  $32 = $31&1;
  STACKTOP = sp;return ($32|0);
 }
 $27 = ((($userData)) + 4|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = $28;
 _SDL_SetMouseFocus($29);
 $30 = (_SDL_EventState(512,-1)|0);
 $31 = ($30<<24>>24)==(1);
 $32 = $31&1;
 STACKTOP = sp;return ($32|0);
}
function _Emscripten_HandleWheel($eventType,$wheelEvent,$userData) {
 $eventType = $eventType|0;
 $wheelEvent = $wheelEvent|0;
 $userData = $userData|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($userData)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($wheelEvent)) + 72|0);
 $3 = +HEAPF64[$2>>3];
 $4 = (~~(($3)));
 $5 = ((($wheelEvent)) + 80|0);
 $6 = +HEAPF64[$5>>3];
 $7 = -$6;
 $8 = (~~(($7)));
 (_SDL_SendMouseWheel($1,0,$4,$8,0)|0);
 $9 = (_SDL_EventState(1027,-1)|0);
 $10 = ($9<<24>>24)==(1);
 $11 = $10&1;
 return ($11|0);
}
function _SDL_SendMouseWheel($window,$mouseID,$x,$y,$direction) {
 $window = $window|0;
 $mouseID = $mouseID|0;
 $x = $x|0;
 $y = $y|0;
 $direction = $direction|0;
 var $$0 = 0, $$01$i$i = 0, $$pr$i = 0, $$val = 0, $$val2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $event = 0, $notlhs$i$i = 0, $notrhs$i$i = 0, $or$cond3$not$i$i = 0, $or$cond5$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $event = sp;
 $0 = ($window|0)==(0|0);
 do {
  if (!($0)) {
   $1 = HEAP32[(1201820)>>2]|0;
   $2 = ($1|0)==($window|0);
   if (!($2)) {
    $3 = ($1|0)==(0|0);
    if (!($3)) {
     (_SDL_SendWindowEvent($1,11,0,0)|0);
    }
    HEAP32[(1201820)>>2] = $window;
    (_SDL_SendWindowEvent($window,10,0,0)|0);
    $$pr$i = HEAP32[(1201820)>>2]|0;
    $4 = ($$pr$i|0)==(0|0);
    $$val = HEAP32[(1201872)>>2]|0;
    $$val2 = HEAP32[(1201876)>>2]|0;
    $$01$i$i = $4 ? $$val : $$val2;
    $5 = HEAP32[(1201880)>>2]|0;
    $notlhs$i$i = ($$01$i$i|0)!=(0|0);
    $notrhs$i$i = ($5|0)!=(0);
    $or$cond3$not$i$i = $notlhs$i$i & $notrhs$i$i;
    $6 = HEAP32[(1201852)>>2]|0;
    $7 = ($6|0)==(0);
    $or$cond5$i$i = $or$cond3$not$i$i & $7;
    $8 = HEAP32[(1201784)>>2]|0;
    $9 = ($8|0)==(0|0);
    if ($or$cond5$i$i) {
     if ($9) {
      break;
     }
     (FUNCTION_TABLE_ii[$8 & 8191]($$01$i$i)|0);
     break;
    } else {
     if ($9) {
      break;
     }
     (FUNCTION_TABLE_ii[$8 & 8191](0)|0);
     break;
    }
   }
  }
 } while(0);
 $10 = $y | $x;
 $11 = ($10|0)==(0);
 if ($11) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $12 = (_SDL_EventState(1027,-1)|0);
 $13 = ($12<<24>>24)==(1);
 if (!($13)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$event>>2] = 1027;
 $14 = HEAP32[(1201820)>>2]|0;
 $15 = ($14|0)==(0|0);
 if ($15) {
  $19 = 0;
 } else {
  $16 = ((($14)) + 4|0);
  $17 = HEAP32[$16>>2]|0;
  $19 = $17;
 }
 $18 = ((($event)) + 8|0);
 HEAP32[$18>>2] = $19;
 $20 = ((($event)) + 12|0);
 HEAP32[$20>>2] = $mouseID;
 $21 = ((($event)) + 16|0);
 HEAP32[$21>>2] = $x;
 $22 = ((($event)) + 20|0);
 HEAP32[$22>>2] = $y;
 $23 = ((($event)) + 24|0);
 HEAP32[$23>>2] = $direction;
 $24 = (_SDL_PushEvent($event)|0);
 $25 = ($24|0)>(0);
 $26 = $25&1;
 $$0 = $26;
 STACKTOP = sp;return ($$0|0);
}
function _Emscripten_HandleFocus($eventType,$wheelEvent,$userData) {
 $eventType = $eventType|0;
 $wheelEvent = $wheelEvent|0;
 $userData = $userData|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($eventType|0)==(12);
 if ($0) {
  _SDL_ResetKeyboard();
 }
 $1 = ((($userData)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($eventType|0)==(13);
 $4 = $3 ? 12 : 13;
 (_SDL_SendWindowEvent($2,$4,0,0)|0);
 $5 = (_SDL_EventState(512,-1)|0);
 $6 = ($5<<24>>24)==(1);
 $7 = $6&1;
 return ($7|0);
}
function _SDL_ResetKeyboard() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $exitcond = 0, $scancode$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $scancode$01 = 0;
 while(1) {
  $0 = (1202966 + ($scancode$01)|0);
  $1 = HEAP8[$0>>0]|0;
  $2 = ($1<<24>>24)==(1);
  if ($2) {
   (_SDL_SendKeyboardKey(0,$scancode$01)|0);
  }
  $3 = (($scancode$01) + 1)|0;
  $exitcond = ($3|0)==(512);
  if ($exitcond) {
   break;
  } else {
   $scancode$01 = $3;
  }
 }
 return;
}
function _Emscripten_HandleTouch($eventType,$touchEvent,$userData) {
 $eventType = $eventType|0;
 $touchEvent = $touchEvent|0;
 $userData = $userData|0;
 var $$0 = 0, $$preventDefault$0$us = 0, $$preventDefault$02 = 0, $$preventDefault$03 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0;
 var $23 = 0.0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0.0;
 var $41 = 0.0, $42 = 0.0, $43 = 0, $44 = 0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $client_h = 0, $client_w = 0, $cond = 0, $i$05 = 0, $i$05$us = 0, $preventDefault$0$lcssa = 0, $preventDefault$04 = 0, $preventDefault$04$us = 0, $preventDefault$2 = 0, $preventDefault$2$us = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $client_w = sp + 8|0;
 $client_h = sp;
 $0 = (_SDL_AddTouch(1,0,1282466)|0);
 $1 = ($0|0)<(0);
 if ($1) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 (_emscripten_get_element_css_size((0|0),($client_w|0),($client_h|0))|0);
 $2 = HEAP32[$touchEvent>>2]|0;
 $3 = ($2|0)>(0);
 L4: do {
  if ($3) {
   $4 = ($eventType|0)==(24);
   if ($4) {
    $i$05$us = 0;$preventDefault$04$us = 0;
    while(1) {
     $5 = (((((($touchEvent)) + 20|0) + (($i$05$us*52)|0)|0)) + 28|0);
     $6 = HEAP32[$5>>2]|0;
     $7 = ($6|0)==(0);
     if ($7) {
      $preventDefault$2$us = $preventDefault$04$us;
     } else {
      $8 = (((($touchEvent)) + 20|0) + (($i$05$us*52)|0)|0);
      $9 = HEAP32[$8>>2]|0;
      $10 = ($9|0)<(0);
      $11 = $10 << 31 >> 31;
      $12 = (((((($touchEvent)) + 20|0) + (($i$05$us*52)|0)|0)) + 44|0);
      $13 = HEAP32[$12>>2]|0;
      $14 = (+($13|0));
      $15 = +HEAPF64[$client_w>>3];
      $16 = $14 / $15;
      $17 = $16;
      $18 = (((((($touchEvent)) + 20|0) + (($i$05$us*52)|0)|0)) + 48|0);
      $19 = HEAP32[$18>>2]|0;
      $20 = (+($19|0));
      $21 = +HEAPF64[$client_h>>3];
      $22 = $20 / $21;
      $23 = $22;
      (_SDL_SendTouchMotion(1,0,$9,$11,$17,$23,1.0)|0);
      $24 = ($preventDefault$04$us|0)==(0);
      if ($24) {
       $25 = (_SDL_EventState(1794,-1)|0);
       $26 = ($25<<24>>24)==(1);
       $$preventDefault$0$us = $26&1;
       $preventDefault$2$us = $$preventDefault$0$us;
      } else {
       $preventDefault$2$us = $preventDefault$04$us;
      }
     }
     $27 = (($i$05$us) + 1)|0;
     $28 = HEAP32[$touchEvent>>2]|0;
     $29 = ($27|0)<($28|0);
     if ($29) {
      $i$05$us = $27;$preventDefault$04$us = $preventDefault$2$us;
     } else {
      $preventDefault$0$lcssa = $preventDefault$2$us;
      break L4;
     }
    }
   }
   $cond = ($eventType|0)==(22);
   $i$05 = 0;$preventDefault$04 = 0;
   while(1) {
    $30 = (((((($touchEvent)) + 20|0) + (($i$05*52)|0)|0)) + 28|0);
    $31 = HEAP32[$30>>2]|0;
    $32 = ($31|0)==(0);
    do {
     if ($32) {
      $preventDefault$2 = $preventDefault$04;
     } else {
      $33 = (((($touchEvent)) + 20|0) + (($i$05*52)|0)|0);
      $34 = HEAP32[$33>>2]|0;
      $35 = ($34|0)<(0);
      $36 = $35 << 31 >> 31;
      $37 = (((((($touchEvent)) + 20|0) + (($i$05*52)|0)|0)) + 44|0);
      $38 = HEAP32[$37>>2]|0;
      $39 = (+($38|0));
      $40 = +HEAPF64[$client_w>>3];
      $41 = $39 / $40;
      $42 = $41;
      $43 = (((((($touchEvent)) + 20|0) + (($i$05*52)|0)|0)) + 48|0);
      $44 = HEAP32[$43>>2]|0;
      $45 = (+($44|0));
      $46 = +HEAPF64[$client_h>>3];
      $47 = $45 / $46;
      $48 = $47;
      if ($cond) {
       (_SDL_SendTouch(1,0,$34,$36,1,$42,$48,1.0)|0);
       $49 = ($preventDefault$04|0)==(0);
       if (!($49)) {
        $preventDefault$2 = $preventDefault$04;
        break;
       }
       $50 = (_SDL_EventState(1792,-1)|0);
       $51 = ($50<<24>>24)==(1);
       $$preventDefault$02 = $51&1;
       $preventDefault$2 = $$preventDefault$02;
       break;
      } else {
       (_SDL_SendTouch(1,0,$34,$36,0,$42,$48,1.0)|0);
       $52 = ($preventDefault$04|0)==(0);
       if (!($52)) {
        $preventDefault$2 = $preventDefault$04;
        break;
       }
       $53 = (_SDL_EventState(1793,-1)|0);
       $54 = ($53<<24>>24)==(1);
       $$preventDefault$03 = $54&1;
       $preventDefault$2 = $$preventDefault$03;
       break;
      }
     }
    } while(0);
    $55 = (($i$05) + 1)|0;
    $56 = HEAP32[$touchEvent>>2]|0;
    $57 = ($55|0)<($56|0);
    if ($57) {
     $i$05 = $55;$preventDefault$04 = $preventDefault$2;
    } else {
     $preventDefault$0$lcssa = $preventDefault$2;
     break;
    }
   }
  } else {
   $preventDefault$0$lcssa = 0;
  }
 } while(0);
 $$0 = $preventDefault$0$lcssa;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_AddTouch($0,$1,$name) {
 $0 = $0|0;
 $1 = $1|0;
 $name = $name|0;
 var $$0 = 0, $$pre = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $index$01$i = 0, $index$01$i$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[301382]|0;
 $3 = ($2|0)>(0);
 L1: do {
  if ($3) {
   $4 = HEAP32[301383]|0;
   $index$01$i = 0;
   while(1) {
    $5 = (($4) + ($index$01$i<<2)|0);
    $6 = HEAP32[$5>>2]|0;
    $7 = $6;
    $8 = $7;
    $9 = HEAP32[$8>>2]|0;
    $10 = (($7) + 4)|0;
    $11 = $10;
    $12 = HEAP32[$11>>2]|0;
    $13 = ($9|0)==($0|0);
    $14 = ($12|0)==($1|0);
    $15 = $13 & $14;
    if ($15) {
     $index$01$i$lcssa = $index$01$i;
     break;
    }
    $16 = (($index$01$i) + 1)|0;
    $17 = ($16|0)<($2|0);
    if ($17) {
     $index$01$i = $16;
    } else {
     $21 = $4;
     break L1;
    }
   }
   $18 = ($index$01$i$lcssa|0)>(-1);
   if ($18) {
    $$0 = $index$01$i$lcssa;
    return ($$0|0);
   } else {
    $21 = $4;
   }
  } else {
   $$pre = HEAP32[301383]|0;
   $21 = $$pre;
  }
 } while(0);
 $19 = $2 << 2;
 $20 = (($19) + 4)|0;
 $22 = (_SDL_realloc($21,$20)|0);
 $23 = ($22|0)==(0|0);
 if ($23) {
  $24 = (_SDL_Error(0)|0);
  $$0 = $24;
  return ($$0|0);
 }
 HEAP32[301383] = $22;
 $25 = HEAP32[301382]|0;
 $26 = (_SDL_malloc(24)|0);
 $27 = HEAP32[301383]|0;
 $28 = (($27) + ($25<<2)|0);
 HEAP32[$28>>2] = $26;
 $29 = ($26|0)==(0|0);
 if ($29) {
  $30 = (_SDL_Error(0)|0);
  $$0 = $30;
  return ($$0|0);
 } else {
  $31 = HEAP32[301382]|0;
  $32 = (($31) + 1)|0;
  HEAP32[301382] = $32;
  $33 = $26;
  $34 = $33;
  HEAP32[$34>>2] = $0;
  $35 = (($33) + 4)|0;
  $36 = $35;
  HEAP32[$36>>2] = $1;
  $37 = ((($26)) + 8|0);
  HEAP32[$37>>2] = 0;
  $38 = ((($26)) + 12|0);
  HEAP32[$38>>2] = 0;
  $39 = ((($26)) + 16|0);
  HEAP32[$39>>2] = 0;
  (_SDL_GestureAddTouch($0,$1)|0);
  $$0 = $25;
  return ($$0|0);
 }
 return (0)|0;
}
function _SDL_GestureAddTouch($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[300439]|0;
 $3 = HEAP32[300437]|0;
 $4 = ($3*8232)|0;
 $5 = (($4) + 8232)|0;
 $6 = (_SDL_realloc($2,$5)|0);
 $7 = ($6|0)==(0|0);
 if ($7) {
  $8 = (_SDL_Error(0)|0);
  $$0 = $8;
  return ($$0|0);
 } else {
  HEAP32[300439] = $6;
  $9 = HEAP32[300437]|0;
  $10 = (($6) + (($9*8232)|0)|0);
  (_SDL_memset($10,0,8232)|0);
  $11 = HEAP32[300437]|0;
  $12 = HEAP32[300439]|0;
  $13 = (($12) + (($11*8232)|0)|0);
  $14 = $13;
  $15 = $14;
  HEAP32[$15>>2] = $0;
  $16 = (($14) + 4)|0;
  $17 = $16;
  HEAP32[$17>>2] = $1;
  $18 = (($11) + 1)|0;
  HEAP32[300437] = $18;
  $$0 = 0;
  return ($$0|0);
 }
 return (0)|0;
}
function _SDL_SendTouchMotion($0,$1,$2,$3,$x,$y,$pressure) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $x = +$x;
 $y = +$y;
 $pressure = +$pressure;
 var $$0 = 0, $$lcssa = 0, $$lcssa36 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0.0, $45 = 0.0;
 var $46 = 0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $8 = 0, $9 = 0, $event = 0, $index$01$i$i = 0, $index$01$i$i$lcssa = 0, $index$01$i$i5 = 0, $index$01$i$i5$lcssa = 0, $notlhs = 0;
 var $notrhs = 0, $or$cond = 0, $or$cond$i = 0, $or$cond$i6$not = 0, $or$cond3 = 0, $or$cond7 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp + 56|0;
 $event = sp;
 $4 = HEAP32[301382]|0;
 $5 = ($4|0)>(0);
 L1: do {
  if ($5) {
   $6 = HEAP32[301383]|0;
   $index$01$i$i = 0;
   while(1) {
    $7 = (($6) + ($index$01$i$i<<2)|0);
    $8 = HEAP32[$7>>2]|0;
    $9 = $8;
    $10 = $9;
    $11 = HEAP32[$10>>2]|0;
    $12 = (($9) + 4)|0;
    $13 = $12;
    $14 = HEAP32[$13>>2]|0;
    $15 = ($11|0)==($0|0);
    $16 = ($14|0)==($1|0);
    $17 = $15 & $16;
    if ($17) {
     $$lcssa36 = $8;$index$01$i$i$lcssa = $index$01$i$i;
     break;
    }
    $18 = (($index$01$i$i) + 1)|0;
    $19 = ($18|0)<($4|0);
    if ($19) {
     $index$01$i$i = $18;
    } else {
     break L1;
    }
   }
   $20 = ($index$01$i$i$lcssa|0)>(-1);
   $21 = ($index$01$i$i$lcssa|0)<($4|0);
   $or$cond$i = $20 & $21;
   if ($or$cond$i) {
    $22 = ($$lcssa36|0)==(0|0);
    if ($22) {
     $$0 = -1;
     STACKTOP = sp;return ($$0|0);
    }
    $23 = ((($$lcssa36)) + 8|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = ($24|0)>(0);
    L11: do {
     if ($25) {
      $26 = ((($$lcssa36)) + 16|0);
      $27 = HEAP32[$26>>2]|0;
      $index$01$i$i5 = 0;
      while(1) {
       $28 = (($27) + ($index$01$i$i5<<2)|0);
       $29 = HEAP32[$28>>2]|0;
       $30 = $29;
       $31 = $30;
       $32 = HEAP32[$31>>2]|0;
       $33 = (($30) + 4)|0;
       $34 = $33;
       $35 = HEAP32[$34>>2]|0;
       $36 = ($32|0)==($2|0);
       $37 = ($35|0)==($3|0);
       $38 = $36 & $37;
       if ($38) {
        $$lcssa = $29;$index$01$i$i5$lcssa = $index$01$i$i5;
        break;
       }
       $39 = (($index$01$i$i5) + 1)|0;
       $40 = ($39|0)<($24|0);
       if ($40) {
        $index$01$i$i5 = $39;
       } else {
        break L11;
       }
      }
      $notlhs = ($index$01$i$i5$lcssa|0)<(0);
      $notrhs = ($index$01$i$i5$lcssa|0)>=($24|0);
      $or$cond$i6$not = $notrhs | $notlhs;
      $41 = ($$lcssa|0)==(0|0);
      $or$cond7 = $or$cond$i6$not | $41;
      if (!($or$cond7)) {
       $43 = ((($$lcssa)) + 8|0);
       $44 = +HEAPF32[$43>>2];
       $45 = $x - $44;
       $46 = ((($$lcssa)) + 12|0);
       $47 = +HEAPF32[$46>>2];
       $48 = $y - $47;
       $49 = ((($$lcssa)) + 16|0);
       $50 = +HEAPF32[$49>>2];
       $51 = $pressure - $50;
       $52 = $45 != 0.0;
       $53 = $48 != 0.0;
       $or$cond = $52 | $53;
       $54 = $51 != 0.0;
       $or$cond3 = $or$cond | $54;
       if (!($or$cond3)) {
        $$0 = 0;
        STACKTOP = sp;return ($$0|0);
       }
       HEAPF32[$43>>2] = $x;
       HEAPF32[$46>>2] = $y;
       HEAPF32[$49>>2] = $pressure;
       $55 = (_SDL_EventState(1794,-1)|0);
       $56 = ($55<<24>>24)==(1);
       if (!($56)) {
        $$0 = 0;
        STACKTOP = sp;return ($$0|0);
       }
       HEAP32[$event>>2] = 1794;
       $57 = ((($event)) + 8|0);
       $58 = $57;
       $59 = $58;
       HEAP32[$59>>2] = $0;
       $60 = (($58) + 4)|0;
       $61 = $60;
       HEAP32[$61>>2] = $1;
       $62 = ((($event)) + 16|0);
       $63 = $62;
       $64 = $63;
       HEAP32[$64>>2] = $2;
       $65 = (($63) + 4)|0;
       $66 = $65;
       HEAP32[$66>>2] = $3;
       $67 = ((($event)) + 24|0);
       HEAPF32[$67>>2] = $x;
       $68 = ((($event)) + 28|0);
       HEAPF32[$68>>2] = $y;
       $69 = ((($event)) + 32|0);
       HEAPF32[$69>>2] = $45;
       $70 = ((($event)) + 36|0);
       HEAPF32[$70>>2] = $48;
       $71 = ((($event)) + 40|0);
       HEAPF32[$71>>2] = $pressure;
       $72 = (_SDL_PushEvent($event)|0);
       $73 = ($72|0)>(0);
       $74 = $73&1;
       $$0 = $74;
       STACKTOP = sp;return ($$0|0);
      }
     }
    } while(0);
    $42 = (_SDL_SendTouch($0,$1,$2,$3,1,$x,$y,$pressure)|0);
    $$0 = $42;
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 (_SDL_SetError(605121,$vararg_buffer)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SendTouch($0,$1,$2,$3,$down,$x,$y,$pressure) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $down = $down|0;
 $x = +$x;
 $y = +$y;
 $pressure = +$pressure;
 var $$0 = 0, $$1$i = 0, $$lcssa = 0, $$lcssa$$i = 0, $$lcssa60 = 0, $$lcssa63 = 0, $$lcssa66 = 0, $$pre$i = 0, $$pre2$i = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $event = 0, $event1 = 0, $index$01$i$i = 0, $index$01$i$i$lcssa = 0, $index$01$i$i2 = 0, $index$01$i$i2$lcssa = 0, $index$01$i$i5 = 0, $index$01$i$i5$lcssa = 0;
 var $or$cond$i = 0, $or$cond$i3 = 0, $posted$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp + 112|0;
 $event = sp + 56|0;
 $event1 = sp;
 $4 = HEAP32[301382]|0;
 $5 = ($4|0)>(0);
 L1: do {
  if ($5) {
   $6 = HEAP32[301383]|0;
   $index$01$i$i = 0;
   while(1) {
    $7 = (($6) + ($index$01$i$i<<2)|0);
    $8 = HEAP32[$7>>2]|0;
    $9 = $8;
    $10 = $9;
    $11 = HEAP32[$10>>2]|0;
    $12 = (($9) + 4)|0;
    $13 = $12;
    $14 = HEAP32[$13>>2]|0;
    $15 = ($11|0)==($0|0);
    $16 = ($14|0)==($1|0);
    $17 = $15 & $16;
    if ($17) {
     $$lcssa66 = $8;$index$01$i$i$lcssa = $index$01$i$i;
     break;
    }
    $18 = (($index$01$i$i) + 1)|0;
    $19 = ($18|0)<($4|0);
    if ($19) {
     $index$01$i$i = $18;
    } else {
     break L1;
    }
   }
   $20 = ($index$01$i$i$lcssa|0)>(-1);
   $21 = ($index$01$i$i$lcssa|0)<($4|0);
   $or$cond$i = $20 & $21;
   if ($or$cond$i) {
    $22 = ($$lcssa66|0)==(0|0);
    if ($22) {
     $$0 = -1;
     STACKTOP = sp;return ($$0|0);
    }
    $23 = ((($$lcssa66)) + 8|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = ($24|0)>(0);
    L11: do {
     if ($25) {
      $26 = ((($$lcssa66)) + 16|0);
      $27 = HEAP32[$26>>2]|0;
      $index$01$i$i2 = 0;
      while(1) {
       $28 = (($27) + ($index$01$i$i2<<2)|0);
       $29 = HEAP32[$28>>2]|0;
       $30 = $29;
       $31 = $30;
       $32 = HEAP32[$31>>2]|0;
       $33 = (($30) + 4)|0;
       $34 = $33;
       $35 = HEAP32[$34>>2]|0;
       $36 = ($32|0)==($2|0);
       $37 = ($35|0)==($3|0);
       $38 = $36 & $37;
       if ($38) {
        $$lcssa63 = $29;$index$01$i$i2$lcssa = $index$01$i$i2;
        break;
       }
       $39 = (($index$01$i$i2) + 1)|0;
       $40 = ($39|0)<($24|0);
       if ($40) {
        $index$01$i$i2 = $39;
       } else {
        $45 = 0;
        break L11;
       }
      }
      $41 = ($index$01$i$i2$lcssa|0)>(-1);
      $42 = ($index$01$i$i2$lcssa|0)<($24|0);
      $or$cond$i3 = $41 & $42;
      $$lcssa$$i = $or$cond$i3 ? $$lcssa63 : 0;
      $45 = $$lcssa$$i;
     } else {
      $45 = 0;
     }
    } while(0);
    $43 = ($down|0)==(0);
    $44 = ($45|0)!=(0|0);
    if ($43) {
     if (!($44)) {
      $$0 = 0;
      STACKTOP = sp;return ($$0|0);
     }
     $99 = (_SDL_EventState(1793,-1)|0);
     $100 = ($99<<24>>24)==(1);
     if ($100) {
      HEAP32[$event1>>2] = 1793;
      $101 = ((($event1)) + 8|0);
      $102 = $101;
      $103 = $102;
      HEAP32[$103>>2] = $0;
      $104 = (($102) + 4)|0;
      $105 = $104;
      HEAP32[$105>>2] = $1;
      $106 = ((($event1)) + 16|0);
      $107 = $106;
      $108 = $107;
      HEAP32[$108>>2] = $2;
      $109 = (($107) + 4)|0;
      $110 = $109;
      HEAP32[$110>>2] = $3;
      $111 = ((($45)) + 8|0);
      $112 = HEAP32[$111>>2]|0;
      $113 = ((($event1)) + 24|0);
      HEAP32[$113>>2] = $112;
      $114 = ((($45)) + 12|0);
      $115 = HEAP32[$114>>2]|0;
      $116 = ((($event1)) + 28|0);
      HEAP32[$116>>2] = $115;
      $117 = ((($event1)) + 32|0);
      HEAPF32[$117>>2] = 0.0;
      $118 = ((($event1)) + 36|0);
      HEAPF32[$118>>2] = 0.0;
      $119 = ((($event1)) + 40|0);
      HEAPF32[$119>>2] = $pressure;
      $120 = (_SDL_PushEvent($event1)|0);
      $121 = ($120|0)>(0);
      $122 = $121&1;
      $posted$0 = $122;
     } else {
      $posted$0 = 0;
     }
     $123 = HEAP32[$23>>2]|0;
     $124 = ($123|0)>(0);
     if (!($124)) {
      $$0 = $posted$0;
      STACKTOP = sp;return ($$0|0);
     }
     $125 = ((($$lcssa66)) + 16|0);
     $126 = HEAP32[$125>>2]|0;
     $index$01$i$i5 = 0;
     while(1) {
      $127 = (($126) + ($index$01$i$i5<<2)|0);
      $128 = HEAP32[$127>>2]|0;
      $129 = $128;
      $130 = $129;
      $131 = HEAP32[$130>>2]|0;
      $132 = (($129) + 4)|0;
      $133 = $132;
      $134 = HEAP32[$133>>2]|0;
      $135 = ($131|0)==($2|0);
      $136 = ($134|0)==($3|0);
      $137 = $135 & $136;
      if ($137) {
       $$lcssa = $127;$$lcssa60 = $128;$index$01$i$i5$lcssa = $index$01$i$i5;
       break;
      }
      $138 = (($index$01$i$i5) + 1)|0;
      $139 = ($138|0)<($123|0);
      if ($139) {
       $index$01$i$i5 = $138;
      } else {
       $$0 = $posted$0;
       label = 35;
       break;
      }
     }
     if ((label|0) == 35) {
      STACKTOP = sp;return ($$0|0);
     }
     $140 = ($index$01$i$i5$lcssa|0)<(0);
     if ($140) {
      $$0 = $posted$0;
      STACKTOP = sp;return ($$0|0);
     }
     $141 = $$lcssa60;
     $142 = (($123) + -1)|0;
     HEAP32[$23>>2] = $142;
     $143 = (($126) + ($142<<2)|0);
     $144 = HEAP32[$143>>2]|0;
     HEAP32[$$lcssa>>2] = $144;
     $145 = HEAP32[$125>>2]|0;
     $146 = (($145) + ($142<<2)|0);
     HEAP32[$146>>2] = $141;
     $$0 = $posted$0;
     STACKTOP = sp;return ($$0|0);
    }
    if ($44) {
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
    $46 = ((($$lcssa66)) + 12|0);
    $47 = HEAP32[$46>>2]|0;
    $48 = ($24|0)==($47|0);
    $49 = ((($$lcssa66)) + 16|0);
    L43: do {
     if ($48) {
      $50 = HEAP32[$49>>2]|0;
      $51 = $24 << 2;
      $52 = (($51) + 4)|0;
      $53 = (_SDL_realloc($50,$52)|0);
      $54 = ($53|0)==(0|0);
      do {
       if ($54) {
        $55 = (_SDL_Error(0)|0);
        $$1$i = $55;
       } else {
        HEAP32[$49>>2] = $53;
        $56 = (_SDL_malloc(24)|0);
        $57 = HEAP32[$46>>2]|0;
        $58 = HEAP32[$49>>2]|0;
        $59 = (($58) + ($57<<2)|0);
        HEAP32[$59>>2] = $56;
        $60 = HEAP32[$49>>2]|0;
        $61 = (($60) + ($57<<2)|0);
        $62 = HEAP32[$61>>2]|0;
        $63 = ($62|0)==(0|0);
        if ($63) {
         $64 = (_SDL_Error(0)|0);
         $$1$i = $64;
         break;
        } else {
         $65 = (($57) + 1)|0;
         HEAP32[$46>>2] = $65;
         $$pre$i = HEAP32[$23>>2]|0;
         $67 = $$pre$i;$69 = $60;
         label = 22;
         break L43;
        }
       }
      } while(0);
      $78 = ($$1$i|0)<(0);
      if ($78) {
       $$0 = 0;
       STACKTOP = sp;return ($$0|0);
      }
     } else {
      $$pre2$i = HEAP32[$49>>2]|0;
      $67 = $24;$69 = $$pre2$i;
      label = 22;
     }
    } while(0);
    if ((label|0) == 22) {
     $66 = (($67) + 1)|0;
     HEAP32[$23>>2] = $66;
     $68 = (($69) + ($67<<2)|0);
     $70 = HEAP32[$68>>2]|0;
     $71 = $70;
     $72 = $71;
     HEAP32[$72>>2] = $2;
     $73 = (($71) + 4)|0;
     $74 = $73;
     HEAP32[$74>>2] = $3;
     $75 = ((($70)) + 8|0);
     HEAPF32[$75>>2] = $x;
     $76 = ((($70)) + 12|0);
     HEAPF32[$76>>2] = $y;
     $77 = ((($70)) + 16|0);
     HEAPF32[$77>>2] = $pressure;
    }
    $79 = (_SDL_EventState(1792,-1)|0);
    $80 = ($79<<24>>24)==(1);
    if (!($80)) {
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
    HEAP32[$event>>2] = 1792;
    $81 = ((($event)) + 8|0);
    $82 = $81;
    $83 = $82;
    HEAP32[$83>>2] = $0;
    $84 = (($82) + 4)|0;
    $85 = $84;
    HEAP32[$85>>2] = $1;
    $86 = ((($event)) + 16|0);
    $87 = $86;
    $88 = $87;
    HEAP32[$88>>2] = $2;
    $89 = (($87) + 4)|0;
    $90 = $89;
    HEAP32[$90>>2] = $3;
    $91 = ((($event)) + 24|0);
    HEAPF32[$91>>2] = $x;
    $92 = ((($event)) + 28|0);
    HEAPF32[$92>>2] = $y;
    $93 = ((($event)) + 32|0);
    HEAPF32[$93>>2] = 0.0;
    $94 = ((($event)) + 36|0);
    HEAPF32[$94>>2] = 0.0;
    $95 = ((($event)) + 40|0);
    HEAPF32[$95>>2] = $pressure;
    $96 = (_SDL_PushEvent($event)|0);
    $97 = ($96|0)>(0);
    $98 = $97&1;
    $$0 = $98;
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 (_SDL_SetError(605121,$vararg_buffer)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _Emscripten_HandleKey($eventType,$keyEvent,$userData) {
 $eventType = $eventType|0;
 $keyEvent = $keyEvent|0;
 $userData = $userData|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $scancode$0 = 0, $switch = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($keyEvent)) + 156|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1>>>0)<(223);
 if ($2) {
  $3 = (137944 + ($1<<2)|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ($4|0)==(0);
  if (!($5)) {
   $6 = ((($keyEvent)) + 64|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(2);
   L4: do {
    if ($8) {
     switch ($4|0) {
     case 225:  {
      $scancode$0 = 229;
      break L4;
      break;
     }
     case 224:  {
      $scancode$0 = 228;
      break L4;
      break;
     }
     case 226:  {
      $scancode$0 = 230;
      break L4;
      break;
     }
     case 227:  {
      $scancode$0 = 231;
      break L4;
      break;
     }
     default: {
      $scancode$0 = $4;
      break L4;
     }
     }
    } else {
     $scancode$0 = $4;
    }
   } while(0);
   $9 = ($eventType|0)==(2);
   $10 = $9&1;
   (_SDL_SendKeyboardKey($10,$scancode$0)|0);
  }
 }
 $11 = ($eventType|0)==(2);
 $12 = $11 ? 768 : 769;
 $13 = (_SDL_EventState($12,-1)|0);
 $14 = ($13<<24>>24)==(1);
 $15 = $14&1;
 if (!($11)) {
  return ($15|0);
 }
 $16 = (_SDL_EventState(771,-1)|0);
 $17 = ($16<<24>>24)==(1);
 if ($17) {
  $18 = HEAP32[$0>>2]|0;
  $19 = $18 & -2;
  $switch = ($19|0)==(8);
  $$ = $switch ? $15 : 0;
  return ($$|0);
 } else {
  return ($15|0);
 }
 return (0)|0;
}
function _Emscripten_HandleKeyPress($eventType,$keyEvent,$userData) {
 $eventType = $eventType|0;
 $keyEvent = $keyEvent|0;
 $userData = $userData|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $text = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $text = sp;
 $0 = ((($keyEvent)) + 152|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1>>>0)<(128);
 do {
  if ($2) {
   $3 = $1&255;
   HEAP8[$text>>0] = $3;
   $4 = ((($text)) + 1|0);
   HEAP8[$4>>0] = 0;
  } else {
   $5 = ($1>>>0)<(2048);
   if ($5) {
    $6 = $1 >>> 6;
    $7 = $6 & 31;
    $8 = $7 | 192;
    $9 = $8&255;
    HEAP8[$text>>0] = $9;
    $10 = $1 & 63;
    $11 = $10 | 128;
    $12 = $11&255;
    $13 = ((($text)) + 1|0);
    HEAP8[$13>>0] = $12;
    $14 = ((($text)) + 2|0);
    HEAP8[$14>>0] = 0;
    break;
   }
   $15 = ($1>>>0)<(65536);
   if ($15) {
    $16 = $1 >>> 12;
    $17 = $16 & 15;
    $18 = $17 | 224;
    $19 = $18&255;
    HEAP8[$text>>0] = $19;
    $20 = $1 >>> 6;
    $21 = $20 & 63;
    $22 = $21 | 128;
    $23 = $22&255;
    $24 = ((($text)) + 1|0);
    HEAP8[$24>>0] = $23;
    $25 = $1 & 63;
    $26 = $25 | 128;
    $27 = $26&255;
    $28 = ((($text)) + 2|0);
    HEAP8[$28>>0] = $27;
    $29 = ((($text)) + 3|0);
    HEAP8[$29>>0] = 0;
    break;
   }
   $30 = ($1>>>0)<(1114112);
   if ($30) {
    $31 = $1 >>> 18;
    $32 = $31 | 240;
    $33 = $32&255;
    HEAP8[$text>>0] = $33;
    $34 = $1 >>> 12;
    $35 = $34 & 63;
    $36 = $35 | 128;
    $37 = $36&255;
    $38 = ((($text)) + 1|0);
    HEAP8[$38>>0] = $37;
    $39 = $1 >>> 6;
    $40 = $39 & 63;
    $41 = $40 | 128;
    $42 = $41&255;
    $43 = ((($text)) + 2|0);
    HEAP8[$43>>0] = $42;
    $44 = $1 & 63;
    $45 = $44 | 128;
    $46 = $45&255;
    $47 = ((($text)) + 3|0);
    HEAP8[$47>>0] = $46;
    $48 = ((($text)) + 4|0);
    HEAP8[$48>>0] = 0;
    break;
   } else {
    $49 = (_SDL_EventState(771,-1)|0);
    $50 = ($49<<24>>24)==(1);
    $51 = $50&1;
    STACKTOP = sp;return ($51|0);
   }
  }
 } while(0);
 (_SDL_SendKeyboardText($text)|0);
 $49 = (_SDL_EventState(771,-1)|0);
 $50 = ($49<<24>>24)==(1);
 $51 = $50&1;
 STACKTOP = sp;return ($51|0);
}
function _SDL_SendKeyboardText($text) {
 $text = $text|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $event = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $event = sp;
 $0 = HEAP8[$text>>0]|0;
 $1 = ($0&255)<(32);
 $2 = ($0<<24>>24)==(127);
 $or$cond = $1 | $2;
 if ($or$cond) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = (_SDL_EventState(771,-1)|0);
 $4 = ($3<<24>>24)==(1);
 if (!($4)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$event>>2] = 771;
 $5 = HEAP32[300740]|0;
 $6 = ($5|0)==(0|0);
 if ($6) {
  $10 = 0;
 } else {
  $7 = ((($5)) + 4|0);
  $8 = HEAP32[$7>>2]|0;
  $10 = $8;
 }
 $9 = ((($event)) + 8|0);
 HEAP32[$9>>2] = $10;
 $11 = ((($event)) + 12|0);
 (_SDL_utf8strlcpy($11,$text,32)|0);
 $12 = (_SDL_PushEvent($event)|0);
 $13 = ($12|0)>(0);
 $14 = $13&1;
 $$0 = $14;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_utf8strlcpy($dst,$src,$dst_bytes) {
 $dst = $dst|0;
 $src = $src|0;
 $dst_bytes = $dst_bytes|0;
 var $$i$0 = 0, $$off = 0, $$ph = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $bytes$0 = 0, $bytes$1 = 0, $c$off$i = 0, $i$03$lcssa = 0, $i$038 = 0, $notlhs = 0, $notrhs = 0, $or$cond = 0, $or$cond1$not = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strlen($src)|0);
 $1 = (($dst_bytes) + -1)|0;
 $2 = ($0>>>0)<($1>>>0);
 $3 = $2 ? $0 : $1;
 $4 = ($3|0)==(0);
 if ($4) {
  $bytes$1 = 0;
  $22 = (($dst) + ($bytes$1)|0);
  HEAP8[$22>>0] = 0;
  return ($bytes$1|0);
 }
 $5 = (($3) + -1)|0;
 $6 = (($src) + ($5)|0);
 $7 = HEAP8[$6>>0]|0;
 $$off = (($7) + 64)<<24>>24;
 $8 = ($$off&255)<(53);
 L4: do {
  if ($8) {
   $bytes$0 = $5;
  } else {
   $notlhs = ($7<<24>>24)>(-1);
   $notrhs = ($7&255)>(191);
   $or$cond1$not = $notrhs | $notlhs;
   $9 = ($5|0)==(0);
   $or$cond = $or$cond1$not | $9;
   if ($or$cond) {
    $bytes$0 = $3;
   } else {
    $10 = $7 & -32;
    $11 = ($10<<24>>24)==(-64);
    L7: do {
     if ($11) {
      $$ph = 2;$i$03$lcssa = $5;
     } else {
      $13 = $7;$i$038 = $5;
      while(1) {
       $12 = $13 & -16;
       $14 = ($12<<24>>24)==(-32);
       if ($14) {
        $$ph = 3;$i$03$lcssa = $i$038;
        break L7;
       }
       $c$off$i = (($13) + 16)<<24>>24;
       $15 = ($c$off$i&255)<(5);
       if ($15) {
        $$ph = 4;$i$03$lcssa = $i$038;
        break L7;
       }
       $18 = (($i$038) + -1)|0;
       $19 = ($18|0)==(0);
       if ($19) {
        $bytes$0 = $3;
        break L4;
       }
       $$phi$trans$insert = (($src) + ($18)|0);
       $$pre = HEAP8[$$phi$trans$insert>>0]|0;
       $20 = $$pre & -32;
       $21 = ($20<<24>>24)==(-64);
       if ($21) {
        $$ph = 2;$i$03$lcssa = $18;
        break;
       } else {
        $13 = $$pre;$i$038 = $18;
       }
      }
     }
    } while(0);
    $16 = (($3) - ($i$03$lcssa))|0;
    $17 = ($16|0)==($$ph|0);
    $$i$0 = $17 ? $3 : $i$03$lcssa;
    $bytes$0 = $$i$0;
   }
  }
 } while(0);
 _memcpy(($dst|0),($src|0),($bytes$0|0))|0;
 $bytes$1 = $bytes$0;
 $22 = (($dst) + ($bytes$1)|0);
 HEAP8[$22>>0] = 0;
 return ($bytes$1|0);
}
function _Emscripten_HandleFullscreenChange($eventType,$fullscreenChangeEvent,$userData) {
 $eventType = $eventType|0;
 $fullscreenChangeEvent = $fullscreenChangeEvent|0;
 $userData = $userData|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$fullscreenChangeEvent>>2]|0;
 $1 = ($0|0)==(0);
 if ($1) {
  $10 = ((($userData)) + 4|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ((($11)) + 48|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = $13 & -4098;
  HEAP32[$12>>2] = $14;
  return 0;
 } else {
  $2 = ((($userData)) + 20|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($userData)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ((($5)) + 48|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = $7 | $3;
  HEAP32[$2>>2] = 0;
  $9 = $8 | 1;
  HEAP32[$6>>2] = $9;
  return 0;
 }
 return (0)|0;
}
function _Emscripten_HandleResize($eventType,$uiEvent,$userData) {
 $eventType = $eventType|0;
 $uiEvent = $uiEvent|0;
 $userData = $userData|0;
 var $$pre = 0.0, $$pre2 = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0.0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0.0, $27 = 0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0, $33 = 0.0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0, $h = 0, $w = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $w = sp + 8|0;
 $h = sp;
 $0 = ((($userData)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 48|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 4129;
 $5 = ($4|0)==(32);
 if (!($5)) {
  STACKTOP = sp;return 0;
 }
 $6 = ((($1)) + 24|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (+($7|0));
 HEAPF64[$w>>3] = $8;
 $9 = ((($1)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (+($10|0));
 HEAPF64[$h>>3] = $11;
 $12 = ((($userData)) + 16|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13|0)==(0);
 if ($14) {
  $19 = $8;$22 = $11;
 } else {
  (_emscripten_get_element_css_size((0|0),($w|0),($h|0))|0);
  $$pre = +HEAPF64[$w>>3];
  $$pre2 = +HEAPF64[$h>>3];
  $19 = $$pre;$22 = $$pre2;
 }
 $15 = ((($userData)) + 12|0);
 $16 = +HEAPF32[$15>>2];
 $17 = $16;
 $18 = $19 * $17;
 $20 = (~~(($18)));
 $21 = $22 * $17;
 $23 = (~~(($21)));
 _emscripten_set_canvas_size(($20|0),($23|0));
 $24 = HEAP32[$12>>2]|0;
 $25 = ($24|0)==(0);
 if ($25) {
  $26 = +HEAPF32[$15>>2];
  $27 = $26 != 1.0;
  if ($27) {
   $28 = +HEAPF64[$w>>3];
   $29 = +HEAPF64[$h>>3];
   (_emscripten_set_element_css_size((0|0),(+$28),(+$29))|0);
  }
 }
 $30 = HEAP32[$0>>2]|0;
 $31 = +HEAPF64[$w>>3];
 $32 = (~~(($31)));
 $33 = +HEAPF64[$h>>3];
 $34 = (~~(($33)));
 (_SDL_SendWindowEvent($30,5,$32,$34)|0);
 STACKTOP = sp;return 0;
}
function _Emscripten_HandleVisibilityChange($eventType,$visEvent,$userData) {
 $eventType = $eventType|0;
 $visEvent = $visEvent|0;
 $userData = $userData|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($userData)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$visEvent>>2]|0;
 $3 = ($2|0)!=(0);
 $4 = $3 ? 2 : 1;
 (_SDL_SendWindowEvent($1,$4,0,0)|0);
 return 0;
}
function _Emscripten_SetWindowTitle($_this,$window) {
 $_this = $_this|0;
 $window = $window|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($window)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = _emscripten_asm_const_ii(5, ($1|0))|0;
 return;
}
function _Emscripten_SetWindowSize($_this,$window) {
 $_this = $_this|0;
 $window = $window|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0.0, $19 = 0, $2 = 0, $20 = 0, $21 = 0.0, $22 = 0, $23 = 0.0, $3 = 0, $4 = 0, $5 = 0.0;
 var $6 = 0, $7 = 0.0, $8 = 0.0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($window)) + 136|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 $3 = ((($window)) + 24|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (+($4|0));
 $6 = ((($1)) + 12|0);
 $7 = +HEAPF32[$6>>2];
 $8 = $5 * $7;
 $9 = (~~(($8)));
 $10 = ((($window)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (+($11|0));
 $13 = $7 * $12;
 $14 = (~~(($13)));
 _emscripten_set_canvas_size(($9|0),($14|0));
 $15 = ((($1)) + 16|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ($16|0)==(0);
 if (!($17)) {
  return;
 }
 $18 = +HEAPF32[$6>>2];
 $19 = $18 != 1.0;
 if (!($19)) {
  return;
 }
 $20 = HEAP32[$3>>2]|0;
 $21 = (+($20|0));
 $22 = HEAP32[$10>>2]|0;
 $23 = (+($22|0));
 (_emscripten_set_element_css_size((0|0),(+$21),(+$23))|0);
 return;
}
function _Emscripten_DestroyWindow($_this,$window) {
 $_this = $_this|0;
 $window = $window|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($window)) + 136|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 _Emscripten_UnregisterEventHandlers($1);
 $3 = HEAP32[$1>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  _SDL_EGL_DestroySurface($_this,$3);
  HEAP32[$1>>2] = 0;
 }
 $5 = HEAP32[$0>>2]|0;
 _SDL_free($5);
 HEAP32[$0>>2] = 0;
 return;
}
function _Emscripten_UnregisterEventHandlers($data) {
 $data = $data|0;
 var $$ = 0, $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 (_emscripten_set_mousemove_callback((605095|0),(0|0),0,(0|0))|0);
 (_emscripten_set_mousedown_callback((605095|0),(0|0),0,(0|0))|0);
 (_emscripten_set_mouseup_callback((605103|0),(0|0),0,(0|0))|0);
 (_emscripten_set_mouseenter_callback((605095|0),(0|0),0,(0|0))|0);
 (_emscripten_set_mouseleave_callback((605095|0),(0|0),0,(0|0))|0);
 (_emscripten_set_wheel_callback((605095|0),(0|0),0,(0|0))|0);
 (_emscripten_set_focus_callback((605113|0),(0|0),0,(0|0))|0);
 (_emscripten_set_blur_callback((605113|0),(0|0),0,(0|0))|0);
 (_emscripten_set_touchstart_callback((605095|0),(0|0),0,(0|0))|0);
 (_emscripten_set_touchend_callback((605095|0),(0|0),0,(0|0))|0);
 (_emscripten_set_touchmove_callback((605095|0),(0|0),0,(0|0))|0);
 (_emscripten_set_touchcancel_callback((605095|0),(0|0),0,(0|0))|0);
 $0 = (_SDL_GetHint(605142)|0);
 $1 = ($0|0)!=(0|0);
 $$ = $1 ? $0 : 605113;
 (_emscripten_set_keydown_callback(($$|0),(0|0),0,(0|0))|0);
 (_emscripten_set_keyup_callback(($$|0),(0|0),0,(0|0))|0);
 (_emscripten_set_keypress_callback(($$|0),(0|0),0,(0|0))|0);
 (_emscripten_set_fullscreenchange_callback((605103|0),(0|0),0,(0|0))|0);
 (_emscripten_set_resize_callback((605113|0),(0|0),0,(0|0))|0);
 (_emscripten_set_visibilitychange_callback((0|0),0,(0|0))|0);
 return;
}
function _SDL_EGL_DestroySurface($_this,$egl_surface) {
 $_this = $_this|0;
 $egl_surface = $egl_surface|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_this)) + 652|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)!=(0|0);
 $3 = ($egl_surface|0)!=(0|0);
 $or$cond = $3 & $2;
 if (!($or$cond)) {
  return;
 }
 $4 = ((($1)) + 52|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($1)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 (FUNCTION_TABLE_iii[$5 & 8191]($7,$egl_surface)|0);
 return;
}
function _Emscripten_SetWindowFullscreen($_this,$window,$display,$fullscreen) {
 $_this = $_this|0;
 $window = $window|0;
 $display = $display|0;
 $fullscreen = $fullscreen|0;
 var $$cmp = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $strategy = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $strategy = sp;
 $0 = ((($window)) + 136|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  STACKTOP = sp;return;
 }
 $3 = ($fullscreen|0)==(0);
 if ($3) {
  (_emscripten_exit_fullscreen()|0);
  STACKTOP = sp;return;
 }
 $4 = ((($window)) + 48|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $5 & 4097;
 $7 = ($6|0)==(4097);
 $8 = $7&1;
 $9 = $7 ? 1 : 2;
 HEAP32[$strategy>>2] = $9;
 do {
  if ($7) {
   $11 = $5 & 8192;
   $12 = ($11|0)==(0);
   $13 = ((($strategy)) + 4|0);
   if ($12) {
    HEAP32[$13>>2] = 1;
    break;
   } else {
    HEAP32[$13>>2] = 2;
    break;
   }
  } else {
   $10 = ((($strategy)) + 4|0);
   HEAP32[$10>>2] = 0;
  }
 } while(0);
 $14 = ((($strategy)) + 8|0);
 HEAP32[$14>>2] = 0;
 $15 = ((($strategy)) + 12|0);
 HEAP32[$15>>2] = 4604;
 $16 = ((($strategy)) + 16|0);
 HEAP32[$16>>2] = $1;
 $17 = ((($1)) + 20|0);
 HEAP32[$17>>2] = $6;
 $18 = ((($1)) + 24|0);
 HEAP32[$18>>2] = $8;
 $19 = (_emscripten_request_fullscreen_strategy((0|0),1,($strategy|0))|0);
 $$cmp = ($19>>>0)>(1);
 if ($$cmp) {
  $20 = HEAP32[$4>>2]|0;
  $21 = $20 & -4098;
  HEAP32[$4>>2] = $21;
 }
 STACKTOP = sp;return;
}
function _Emscripten_HandleCanvasResize($eventType,$reserved,$userData) {
 $eventType = $eventType|0;
 $reserved = $reserved|0;
 $userData = $userData|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0.0, $6 = 0, $7 = 0.0, $8 = 0, $css_h = 0, $css_w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $css_w = sp + 8|0;
 $css_h = sp;
 $0 = ((($userData)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  STACKTOP = sp;return 0;
 }
 (_emscripten_get_element_css_size((0|0),($css_w|0),($css_h|0))|0);
 $3 = ((($userData)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = +HEAPF64[$css_w>>3];
 $6 = (~~(($5)));
 $7 = +HEAPF64[$css_h>>3];
 $8 = (~~(($7)));
 (_SDL_SendWindowEvent($4,5,$6,$8)|0);
 STACKTOP = sp;return 0;
}
function _Emscripten_CreateWindowFramebuffer($_this,$window,$format,$pixels,$pitch) {
 $_this = $_this|0;
 $window = $window|0;
 $format = $format|0;
 $pixels = $pixels|0;
 $pitch = $pitch|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $Amask = 0, $Bmask = 0;
 var $Gmask = 0, $Rmask = 0, $bpp = 0, $h = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $w = sp + 24|0;
 $h = sp + 20|0;
 $bpp = sp + 16|0;
 $Rmask = sp + 12|0;
 $Gmask = sp + 8|0;
 $Bmask = sp + 4|0;
 $Amask = sp;
 $0 = ((($window)) + 136|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 _SDL_FreeSurface($3);
 (_SDL_PixelFormatEnumToMasks(374740996,$bpp,$Rmask,$Gmask,$Bmask,$Amask)|0);
 _SDL_GetWindowSize($window,$w,$h);
 $4 = HEAP32[$w>>2]|0;
 $5 = HEAP32[$h>>2]|0;
 $6 = HEAP32[$bpp>>2]|0;
 $7 = HEAP32[$Rmask>>2]|0;
 $8 = HEAP32[$Gmask>>2]|0;
 $9 = HEAP32[$Bmask>>2]|0;
 $10 = HEAP32[$Amask>>2]|0;
 $11 = (_SDL_CreateRGBSurface(0,$4,$5,$6,$7,$8,$9,$10)|0);
 $12 = ($11|0)==(0|0);
 if ($12) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$2>>2] = $11;
 HEAP32[$format>>2] = 374740996;
 $13 = ((($11)) + 20|0);
 $14 = HEAP32[$13>>2]|0;
 HEAP32[$pixels>>2] = $14;
 $15 = ((($11)) + 16|0);
 $16 = HEAP32[$15>>2]|0;
 HEAP32[$pitch>>2] = $16;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _Emscripten_UpdateWindowFramebuffer($_this,$window,$rects,$numrects) {
 $_this = $_this|0;
 $window = $window|0;
 $rects = $rects|0;
 $numrects = $numrects|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($window)) + 136|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  $5 = (_SDL_SetError(605304,$vararg_buffer)|0);
  $$0 = $5;
  STACKTOP = sp;return ($$0|0);
 } else {
  $6 = ((($3)) + 8|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ((($3)) + 12|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($3)) + 20|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = _emscripten_asm_const_iiii(6, ($7|0), ($9|0), ($11|0))|0;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _Emscripten_DestroyWindowFramebuffer($_this,$window) {
 $_this = $_this|0;
 $window = $window|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($window)) + 136|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 _SDL_FreeSurface($3);
 HEAP32[$2>>2] = 0;
 return;
}
function _Emscripten_GLES_LoadLibrary($_this,$path) {
 $_this = $_this|0;
 $path = $path|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_SDL_calloc(1,88)|0);
 $1 = ((($_this)) + 652|0);
 HEAP32[$1>>2] = $0;
 $2 = ($0|0)==(0|0);
 if ($2) {
  $3 = (_SDL_Error(0)|0);
  $$0 = $3;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ((($0)) + 20|0);
 HEAP32[$4>>2] = 4605;
 $5 = ((($0)) + 24|0);
 HEAP32[$5>>2] = 4606;
 $6 = ((($0)) + 28|0);
 HEAP32[$6>>2] = 4607;
 $7 = ((($0)) + 32|0);
 HEAP32[$7>>2] = 4608;
 $8 = ((($0)) + 36|0);
 HEAP32[$8>>2] = 4609;
 $9 = ((($0)) + 72|0);
 HEAP32[$9>>2] = 4610;
 $10 = ((($0)) + 40|0);
 HEAP32[$10>>2] = 4611;
 $11 = ((($0)) + 44|0);
 HEAP32[$11>>2] = 4612;
 $12 = ((($0)) + 48|0);
 HEAP32[$12>>2] = 4613;
 $13 = ((($0)) + 52|0);
 HEAP32[$13>>2] = 4614;
 $14 = ((($0)) + 56|0);
 HEAP32[$14>>2] = 4615;
 $15 = ((($0)) + 60|0);
 HEAP32[$15>>2] = 4616;
 $16 = ((($0)) + 64|0);
 HEAP32[$16>>2] = 4617;
 $17 = ((($0)) + 76|0);
 HEAP32[$17>>2] = 4618;
 $18 = ((($0)) + 80|0);
 HEAP32[$18>>2] = 4619;
 $19 = ((($0)) + 84|0);
 HEAP32[$19>>2] = 4620;
 $20 = ((($0)) + 20|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = (FUNCTION_TABLE_ii[$21 & 8191](0)|0);
 $23 = HEAP32[$1>>2]|0;
 $24 = ((($23)) + 8|0);
 HEAP32[$24>>2] = $22;
 $25 = ($22|0)==(0|0);
 if ($25) {
  $26 = (_SDL_SetError(606804,$vararg_buffer)|0);
  $$0 = $26;
  STACKTOP = sp;return ($$0|0);
 }
 $27 = ((($23)) + 24|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = (FUNCTION_TABLE_iiii[$28 & 8191]($22,0,0)|0);
 $30 = ($29|0)==(1);
 if (!($30)) {
  $31 = (_SDL_SetError(606830,$vararg_buffer1)|0);
  $$0 = $31;
  STACKTOP = sp;return ($$0|0);
 }
 $32 = ((($_this)) + 364|0);
 HEAP32[$32>>2] = 1;
 $33 = ($path|0)==(0|0);
 $34 = ((($_this)) + 368|0);
 if ($33) {
  HEAP8[$34>>0] = 0;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  (_SDL_strlcpy($34,$path,255)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_EGL_GetProcAddress($_this,$proc) {
 $_this = $_this|0;
 $proc = $proc|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_this)) + 652|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 32|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 do {
  if ($4) {
   $8 = $1;
  } else {
   $5 = (FUNCTION_TABLE_ii[$3 & 8191]($proc)|0);
   $6 = ($5|0)==(0|0);
   if ($6) {
    $$pre = HEAP32[$0>>2]|0;
    $8 = $$pre;
    break;
   } else {
    $$0 = $5;
    return ($$0|0);
   }
  }
 } while(0);
 $7 = HEAP32[$8>>2]|0;
 $9 = (_SDL_LoadFunction($7,$proc)|0);
 $10 = ($9|0)==(0|0);
 if (!($10)) {
  $$0 = $9;
  return ($$0|0);
 }
 $11 = (_SDL_strlen($proc)|0);
 $12 = ($11>>>0)<(1023);
 if (!($12)) {
  $$0 = 0;
  return ($$0|0);
 }
 HEAP8[1279301] = 95;
 (_SDL_strlcpy((1279302),$proc,1022)|0);
 $13 = HEAP32[$0>>2]|0;
 $14 = HEAP32[$13>>2]|0;
 $15 = (_SDL_LoadFunction($14,1279301)|0);
 $$0 = $15;
 return ($$0|0);
}
function _SDL_LoadFunction($handle,$name) {
 $handle = $handle|0;
 $name = $name|0;
 var $$alloca_mul = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $symbol$0 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_dlsym(($handle|0),($name|0))|0);
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  $symbol$0 = $0;
  STACKTOP = sp;return ($symbol$0|0);
 }
 $2 = (_SDL_strlen($name)|0);
 $3 = (($2) + 2)|0;
 $$alloca_mul = $3;
 $4 = STACKTOP; STACKTOP = STACKTOP + ((((1*$$alloca_mul)|0)+15)&-16)|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();;
 HEAP8[$4>>0] = 95;
 $5 = ((($4)) + 1|0);
 (_SDL_strlcpy($5,$name,$3)|0);
 $6 = (_dlsym(($handle|0),($4|0))|0);
 $7 = ($6|0)==(0|0);
 if (!($7)) {
  $symbol$0 = $6;
  STACKTOP = sp;return ($symbol$0|0);
 }
 $8 = (_dlerror()|0);
 HEAP32[$vararg_buffer>>2] = $name;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $8;
 (_SDL_SetError(606855,$vararg_buffer)|0);
 $symbol$0 = 0;
 STACKTOP = sp;return ($symbol$0|0);
}
function _SDL_EGL_UnloadLibrary($_this) {
 $_this = $_this|0;
 var $$in = 0, $$in1 = 0, $$in2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_this)) + 652|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 $3 = ((($1)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if ($5) {
  $$in = $1;
 } else {
  $6 = ((($1)) + 28|0);
  $7 = HEAP32[$6>>2]|0;
  (FUNCTION_TABLE_ii[$7 & 8191]($4)|0);
  $8 = HEAP32[$0>>2]|0;
  $9 = ((($8)) + 8|0);
  HEAP32[$9>>2] = 0;
  $$in = $8;
 }
 $10 = ((($$in)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0|0);
 if ($12) {
  $$in1 = $$in;
 } else {
  _SDL_UnloadObject($11);
  $13 = HEAP32[$0>>2]|0;
  $14 = ((($13)) + 4|0);
  HEAP32[$14>>2] = 0;
  $$in1 = $13;
 }
 $15 = HEAP32[$$in1>>2]|0;
 $16 = ($15|0)==(0|0);
 if ($16) {
  $$in2 = $$in1;
 } else {
  _SDL_UnloadObject($15);
  $17 = HEAP32[$0>>2]|0;
  HEAP32[$17>>2] = 0;
  $$in2 = $17;
 }
 _SDL_free($$in2);
 HEAP32[$0>>2] = 0;
 return;
}
function _SDL_UnloadObject($handle) {
 $handle = $handle|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($handle|0)==(0|0);
 if ($0) {
  return;
 }
 (_dlclose(($handle|0))|0);
 return;
}
function _Emscripten_GLES_CreateContext($_this,$window) {
 $_this = $_this|0;
 $window = $window|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($window)) + 136|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = (_SDL_EGL_CreateContext($_this,$2)|0);
 return ($3|0);
}
function _SDL_EGL_CreateContext($_this,$egl_surface) {
 $_this = $_this|0;
 $egl_surface = $egl_surface|0;
 var $$0 = 0, $$0$i = 0, $$phi$trans$insert$i = 0, $$phi$trans$insert3$i = 0, $$pre$i = 0, $$pre2$i = 0, $$pre4$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $attribs = 0, $or$cond = 0, $or$cond4 = 0, $or$cond6 = 0, $share_context$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $attribs = sp + 36|0;
 $0 = ((($_this)) + 344|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($_this)) + 332|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($_this)) + 336|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($1|0)==(4);
 $7 = ((($_this)) + 652|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 if ($9) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $10 = ((($_this)) + 348|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0);
 if ($12) {
  $share_context$0 = 0;
 } else {
  $13 = (_SDL_GL_GetCurrentContext()|0);
  $share_context$0 = $13;
 }
 $14 = ($3|0)<(3);
 $15 = ($5|0)==(0);
 $or$cond = $6 & $15;
 $or$cond4 = $14 | $or$cond;
 if ($or$cond4) {
  $16 = ((($_this)) + 340|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = ($17|0)==(0);
  $19 = $1 | 4;
  $20 = ($19|0)==(4);
  $or$cond6 = $20 & $18;
  if ($or$cond6) {
   if ($6) {
    HEAP32[$attribs>>2] = 12440;
    $21 = ($3|0)>(1);
    $22 = $21 ? $3 : 1;
    $23 = ((($attribs)) + 4|0);
    HEAP32[$23>>2] = $22;
    $24 = ((($attribs)) + 8|0);
    HEAP32[$24>>2] = 12344;
    $25 = HEAP32[$7>>2]|0;
    $26 = ((($25)) + 84|0);
    $27 = HEAP32[$26>>2]|0;
    (FUNCTION_TABLE_ii[$27 & 8191](12448)|0);
   } else {
    HEAP32[$attribs>>2] = 12344;
    $28 = HEAP32[$7>>2]|0;
    $29 = ((($28)) + 84|0);
    $30 = HEAP32[$29>>2]|0;
    (FUNCTION_TABLE_ii[$30 & 8191](12450)|0);
   }
   $31 = HEAP32[$7>>2]|0;
   $32 = ((($31)) + 40|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ((($31)) + 8|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = ((($31)) + 12|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = (FUNCTION_TABLE_iiiii[$33 & 8191]($35,$37,$share_context$0,$attribs)|0);
   $39 = ($38|0)==(0|0);
   if ($39) {
    (_SDL_SetError(606945,$vararg_buffer1)|0);
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   $40 = HEAP32[$7>>2]|0;
   $41 = ((($40)) + 16|0);
   HEAP32[$41>>2] = 0;
   $42 = ($40|0)==(0|0);
   do {
    if ($42) {
     $43 = (_SDL_SetError(606974,$vararg_buffer3)|0);
     $$0$i = $43;
    } else {
     $44 = ($egl_surface|0)==(0|0);
     $45 = ((($40)) + 56|0);
     $46 = HEAP32[$45>>2]|0;
     $47 = ((($40)) + 8|0);
     $48 = HEAP32[$47>>2]|0;
     if ($44) {
      (FUNCTION_TABLE_iiiii[$46 & 8191]($48,0,0,0)|0);
      $$0 = $38;
      STACKTOP = sp;return ($$0|0);
     }
     $49 = (FUNCTION_TABLE_iiiii[$46 & 8191]($48,$egl_surface,$egl_surface,$38)|0);
     $50 = ($49|0)==(0);
     if ($50) {
      $51 = (_SDL_SetError(606997,$vararg_buffer5)|0);
      $$0$i = $51;
      break;
     } else {
      $$0 = $38;
      STACKTOP = sp;return ($$0|0);
     }
    }
   } while(0);
   $52 = ($$0$i|0)<(0);
   if (!($52)) {
    $$0 = $38;
    STACKTOP = sp;return ($$0|0);
   }
   $53 = HEAP32[$7>>2]|0;
   $54 = ($53|0)==(0|0);
   if (!($54)) {
    $55 = ((($53)) + 56|0);
    $56 = HEAP32[$55>>2]|0;
    $57 = ((($53)) + 8|0);
    $58 = HEAP32[$57>>2]|0;
    (FUNCTION_TABLE_iiiii[$56 & 8191]($58,0,0,0)|0);
    $$pre$i = HEAP32[$7>>2]|0;
    $$phi$trans$insert$i = ((($$pre$i)) + 44|0);
    $$pre2$i = HEAP32[$$phi$trans$insert$i>>2]|0;
    $$phi$trans$insert3$i = ((($$pre$i)) + 8|0);
    $$pre4$i = HEAP32[$$phi$trans$insert3$i>>2]|0;
    (FUNCTION_TABLE_iii[$$pre2$i & 8191]($$pre4$i,$38)|0);
   }
   (_SDL_SetError(607032,$vararg_buffer7)|0);
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_SDL_SetError(606877,$vararg_buffer)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GL_GetCurrentContext() {
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $2 = ((($0)) + 640|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = (_SDL_TLSGet($3)|0);
  $$0 = $4;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _Emscripten_GLES_MakeCurrent($_this,$window,$context) {
 $_this = $_this|0;
 $window = $window|0;
 $context = $context|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($window|0)!=(0|0);
 $1 = ($context|0)!=(0|0);
 $or$cond = $0 & $1;
 if ($or$cond) {
  $2 = ((($window)) + 136|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = HEAP32[$3>>2]|0;
  $5 = (_SDL_EGL_MakeCurrent($_this,$4,$context)|0);
  $$0 = $5;
  return ($$0|0);
 } else {
  $6 = (_SDL_EGL_MakeCurrent($_this,0,0)|0);
  $$0 = $6;
  return ($$0|0);
 }
 return (0)|0;
}
function _SDL_EGL_MakeCurrent($_this,$egl_surface,$context) {
 $_this = $_this|0;
 $egl_surface = $egl_surface|0;
 $context = $context|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($_this)) + 652|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $3 = (_SDL_SetError(606974,$vararg_buffer)|0);
  $$0 = $3;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ($context|0)!=(0|0);
 $5 = ($egl_surface|0)!=(0|0);
 $or$cond = $5 & $4;
 $6 = ((($1)) + 56|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($1)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 if (!($or$cond)) {
  (FUNCTION_TABLE_iiiii[$7 & 8191]($9,0,0,0)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $10 = (FUNCTION_TABLE_iiiii[$7 & 8191]($9,$egl_surface,$egl_surface,$context)|0);
 $11 = ($10|0)==(0);
 if (!($11)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $12 = (_SDL_SetError(606997,$vararg_buffer1)|0);
 $$0 = $12;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_EGL_SetSwapInterval($_this,$interval) {
 $_this = $_this|0;
 $interval = $interval|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($_this)) + 652|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $3 = (_SDL_SetError(607067,$vararg_buffer)|0);
  $$0 = $3;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ((($1)) + 64|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($1)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (FUNCTION_TABLE_iii[$5 & 8191]($7,$interval)|0);
 $9 = ($8|0)==(1);
 if ($9) {
  $10 = HEAP32[$0>>2]|0;
  $11 = ((($10)) + 16|0);
  HEAP32[$11>>2] = $interval;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $12 = (_SDL_SetError(607087,$vararg_buffer1)|0);
  $$0 = $12;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_EGL_GetSwapInterval($_this) {
 $_this = $_this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($_this)) + 652|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $3 = (_SDL_SetError(607067,$vararg_buffer)|0);
  $$0 = $3;
  STACKTOP = sp;return ($$0|0);
 } else {
  $4 = ((($1)) + 16|0);
  $5 = HEAP32[$4>>2]|0;
  $$0 = $5;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _Emscripten_GLES_SwapWindow($_this,$window) {
 $_this = $_this|0;
 $window = $window|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($window)) + 136|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 _SDL_EGL_SwapBuffers($_this,$2);
 return;
}
function _SDL_EGL_SwapBuffers($_this,$egl_surface) {
 $_this = $_this|0;
 $egl_surface = $egl_surface|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_this)) + 652|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 60|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($1)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 (FUNCTION_TABLE_iii[$3 & 8191]($5,$egl_surface)|0);
 return;
}
function _Emscripten_GLES_DeleteContext($_this,$context) {
 $_this = $_this|0;
 $context = $context|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _glClear(17664);
 _SDL_EGL_DeleteContext($_this,$context);
 return;
}
function _SDL_EGL_DeleteContext($_this,$context) {
 $_this = $_this|0;
 $context = $context|0;
 var $$phi$trans$insert = 0, $$phi$trans$insert3 = 0, $$pre = 0, $$pre2 = 0, $$pre4 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_this)) + 652|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 $3 = ($context|0)==(0|0);
 $or$cond = $3 | $2;
 if ($or$cond) {
  return;
 }
 $4 = ((($1)) + 56|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($1)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 (FUNCTION_TABLE_iiiii[$5 & 8191]($7,0,0,0)|0);
 $$pre = HEAP32[$0>>2]|0;
 $$phi$trans$insert = ((($$pre)) + 44|0);
 $$pre2 = HEAP32[$$phi$trans$insert>>2]|0;
 $$phi$trans$insert3 = ((($$pre)) + 8|0);
 $$pre4 = HEAP32[$$phi$trans$insert3>>2]|0;
 (FUNCTION_TABLE_iii[$$pre2 & 8191]($$pre4,$context)|0);
 return;
}
function _Emscripten_GLES_GetDrawableSize($_this,$window,$w,$h) {
 $_this = $_this|0;
 $window = $window|0;
 $w = $w|0;
 $h = $h|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0.0, $15 = 0, $16 = 0.0, $17 = 0.0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0.0, $9 = 0.0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($window)) + 136|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 $3 = ($w|0)==(0|0);
 if (!($3)) {
  $4 = ((($window)) + 24|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = (+($5|0));
  $7 = ((($1)) + 12|0);
  $8 = +HEAPF32[$7>>2];
  $9 = $6 * $8;
  $10 = (~~(($9)));
  HEAP32[$w>>2] = $10;
 }
 $11 = ($h|0)==(0|0);
 if ($11) {
  return;
 }
 $12 = ((($window)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = (+($13|0));
 $15 = ((($1)) + 12|0);
 $16 = +HEAPF32[$15>>2];
 $17 = $14 * $16;
 $18 = (~~(($17)));
 HEAP32[$h>>2] = $18;
 return;
}
function _Emscripten_DeleteDevice($device) {
 $device = $device|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _SDL_free($device);
 return;
}
function _Emscripten_Init($impl) {
 $impl = $impl|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($impl)) + 4|0);
 HEAP32[$0>>2] = 4621;
 $1 = ((($impl)) + 32|0);
 HEAP32[$1>>2] = 4622;
 $2 = ((($impl)) + 64|0);
 HEAP32[$2>>2] = 1;
 $3 = ((($impl)) + 56|0);
 HEAP32[$3>>2] = 1;
 $4 = ((($impl)) + 52|0);
 HEAP32[$4>>2] = 1;
 $5 = _emscripten_asm_const_i(7)|0;
 $6 = ($5|0)==(0);
 if (!($6)) {
  STACKTOP = sp;return ($5|0);
 }
 (_SDL_SetError(608554,$vararg_buffer)|0);
 STACKTOP = sp;return ($5|0);
}
function _Emscripten_OpenDevice($this,$handle,$devname,$iscapture) {
 $this = $this|0;
 $handle = $handle|0;
 $devname = $devname|0;
 $iscapture = $iscapture|0;
 var $$0 = 0, $$lcssa = 0.0, $$pre = 0, $$pre10 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0;
 var $23 = 0.0, $24 = 0.0, $25 = 0.0, $26 = 0.0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cond = 0, $i$0216 = 0, $test_format$03 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ((($this)) + 4|0);
 $1 = ((($this)) + 8|0);
 $2 = HEAP16[$1>>1]|0;
 $3 = (_SDL_FirstAudioFormat($2)|0);
 $4 = ($3<<16>>16)==(0);
 L1: do {
  if (!($4)) {
   $test_format$03 = $3;
   while(1) {
    $cond = ($test_format$03<<16>>16)==(-32480);
    if ($cond) {
     break;
    }
    $5 = (_SDL_NextAudioFormat()|0);
    $6 = ($5<<16>>16)==(0);
    if ($6) {
     break L1;
    } else {
     $test_format$03 = $5;
    }
   }
   HEAP16[$1>>1] = -32480;
   (_SDL_NextAudioFormat()|0);
   $8 = (_SDL_malloc(20)|0);
   $9 = ((($this)) + 184|0);
   HEAP32[$9>>2] = $8;
   $10 = ($8|0)==(0|0);
   if ($10) {
    $11 = (_SDL_Error(0)|0);
    $$0 = $11;
    STACKTOP = sp;return ($$0|0);
   }
   (_SDL_memset($8,0,20)|0);
   $12 = _emscripten_asm_const_i(8)|0;
   $13 = ($12|0)<(0);
   if ($13) {
    $14 = (_SDL_SetError(607530,$vararg_buffer1)|0);
    $$0 = $14;
    STACKTOP = sp;return ($$0|0);
   }
   $15 = _emscripten_asm_const_i(9)|0;
   $16 = HEAP32[$0>>2]|0;
   $17 = ($16|0)==($15|0);
   $$pre10 = ((($this)) + 12|0);
   if (!($17)) {
    $18 = HEAP16[$$pre10>>1]|0;
    $19 = ($18<<16>>16)==(0);
    L16: do {
     if (!($19)) {
      $20 = (+($15|0));
      $21 = (+($18&65535));
      $22 = $21 / $20;
      $23 = (+($16|0));
      $24 = $22 * $23;
      $25 = $24;
      $26 = (+_SDL_floor($25));
      $27 = $26 == $25;
      if ($27) {
       $$lcssa = $25;
      } else {
       $28 = $18&65535;
       $i$0216 = $28;
       while(1) {
        $29 = ($i$0216|0)>(1);
        if (!($29)) {
         break L16;
        }
        $30 = (($i$0216) + -1)|0;
        $$pre = HEAP32[$0>>2]|0;
        $31 = (+($30|0));
        $32 = $31 / $20;
        $33 = (+($$pre|0));
        $34 = $32 * $33;
        $35 = $34;
        $36 = (+_SDL_floor($35));
        $37 = $36 == $35;
        if ($37) {
         $$lcssa = $35;
         break;
        } else {
         $i$0216 = $30;
        }
       }
      }
      $38 = (+_SDL_floor($$lcssa));
      $39 = (~~(($38))>>>0);
      $40 = HEAP32[$9>>2]|0;
      $41 = ((($40)) + 8|0);
      HEAP32[$41>>2] = $39;
     }
    } while(0);
    HEAP32[$0>>2] = $15;
   }
   _SDL_CalculateAudioSpec($0);
   $42 = ((($this)) + 10|0);
   $43 = HEAP8[$42>>0]|0;
   $44 = $43&255;
   $45 = HEAP16[$$pre10>>1]|0;
   $46 = $45&65535;
   $47 = _emscripten_asm_const_iiiii(10, ($44|0), ($46|0), (4623|0), ($this|0))|0;
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 $7 = (_SDL_SetError(607151,$vararg_buffer)|0);
 $$0 = $7;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_FirstAudioFormat($format) {
 $format = $format|0;
 var $$0$i = 0, $$ph = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[301384] = 0;
 $0 = ($format<<16>>16)==(8);
 if ($0) {
  $$ph = 0;
 } else {
  HEAP32[301384] = 1;
  $1 = ($format<<16>>16)==(-32760);
  if ($1) {
   $$ph = 1;
  } else {
   HEAP32[301384] = 2;
   $4 = ($format<<16>>16)==(-32752);
   if ($4) {
    $$ph = 2;
   } else {
    HEAP32[301384] = 3;
    $5 = ($format<<16>>16)==(-28656);
    if ($5) {
     $$ph = 3;
    } else {
     HEAP32[301384] = 4;
     $6 = ($format<<16>>16)==(16);
     if ($6) {
      $$ph = 4;
     } else {
      HEAP32[301384] = 5;
      $7 = ($format<<16>>16)==(4112);
      if ($7) {
       $$ph = 5;
      } else {
       HEAP32[301384] = 6;
       $8 = ($format<<16>>16)==(-32736);
       if ($8) {
        $$ph = 6;
       } else {
        HEAP32[301384] = 7;
        $9 = ($format<<16>>16)==(-28640);
        if ($9) {
         $$ph = 7;
        } else {
         HEAP32[301384] = 8;
         $10 = ($format<<16>>16)==(-32480);
         if ($10) {
          $$ph = 8;
         } else {
          HEAP32[301384] = 9;
          $11 = ($format<<16>>16)==(-28384);
          if ($11) {
           $$ph = 9;
          } else {
           HEAP32[301384] = 10;
           HEAP32[301385] = 0;
           $$0$i = 0;
           return ($$0$i|0);
          }
         }
        }
       }
      }
     }
    }
   }
  }
 }
 HEAP32[301385] = 1;
 $2 = (175262 + (($$ph*20)|0)|0);
 $3 = HEAP16[$2>>1]|0;
 $$0$i = $3;
 return ($$0$i|0);
}
function _SDL_NextAudioFormat() {
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[301384]|0;
 $1 = ($0|0)==(10);
 $2 = HEAP32[301385]|0;
 $3 = ($2|0)==(10);
 $or$cond = $1 | $3;
 if ($or$cond) {
  $$0 = 0;
  return ($$0|0);
 }
 $4 = (($2) + 1)|0;
 HEAP32[301385] = $4;
 $5 = ((175262 + (($0*20)|0)|0) + ($2<<1)|0);
 $6 = HEAP16[$5>>1]|0;
 $$0 = $6;
 return ($$0|0);
}
function _SDL_CalculateAudioSpec($spec) {
 $spec = $spec|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($spec)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1&65535;
 $cond = ($2<<16>>16)==(8);
 $3 = ((($spec)) + 7|0);
 $$ = $cond ? -128 : 0;
 HEAP8[$3>>0] = $$;
 $4 = $1 >>> 3;
 $5 = $4 & 31;
 $6 = ((($spec)) + 12|0);
 $7 = $1 >>> 16;
 $8 = $7 & 255;
 $9 = Math_imul($8, $5)|0;
 $10 = ((($spec)) + 8|0);
 $11 = HEAP16[$10>>1]|0;
 $12 = $11&65535;
 $13 = Math_imul($12, $9)|0;
 HEAP32[$6>>2] = $13;
 return;
}
function _HandleAudioProcess($this) {
 $this = $this|0;
 var $$ = 0, $$lcssa = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$i = 0, $$pre$i3 = 0, $$pre$phiZ2D = 0, $$pre1$i = 0, $$pre1$i2 = 0, $$pre1$pre$i = 0, $$pre1$pre$i1 = 0, $$pre7 = 0, $$pre9 = 0, $$pre9$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $buf$0 = 0, $byte_len$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 >>> 3;
 $3 = $2 & 31;
 $4 = ((($this)) + 28|0);
 $5 = ((($this)) + 32|0);
 $6 = HEAP16[$5>>1]|0;
 $7 = $6&65535;
 $8 = $7 >>> 3;
 $9 = $8 & 31;
 $10 = ((($this)) + 136|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0);
 $13 = $1 >>> 16;
 if ($12) {
  return;
 }
 $14 = ((($this)) + 144|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)==(0);
 if (!($16)) {
  return;
 }
 $17 = HEAP32[$4>>2]|0;
 $18 = ($17|0)==(0);
 $19 = ((($this)) + 184|0);
 $20 = HEAP32[$19>>2]|0;
 if ($18) {
  $134 = HEAP32[$20>>2]|0;
  $135 = ($134|0)==(0|0);
  if ($135) {
   $136 = ((($this)) + 16|0);
   $137 = HEAP32[$136>>2]|0;
   $138 = ((($20)) + 4|0);
   HEAP32[$138>>2] = $137;
   $139 = (_SDL_malloc($137)|0);
   $140 = HEAP32[$19>>2]|0;
   HEAP32[$140>>2] = $139;
   $146 = $140;$148 = $139;
  } else {
   $146 = $20;$148 = $134;
  }
  $141 = ((($this)) + 20|0);
  $142 = HEAP32[$141>>2]|0;
  $143 = ((($this)) + 24|0);
  $144 = HEAP32[$143>>2]|0;
  $145 = ((($146)) + 4|0);
  $147 = HEAP32[$145>>2]|0;
  FUNCTION_TABLE_viii[$142 & 8191]($144,$148,$147);
  $149 = HEAP32[$19>>2]|0;
  $150 = HEAP32[$149>>2]|0;
  $151 = ((($149)) + 4|0);
  $152 = HEAP32[$151>>2]|0;
  $buf$0 = $150;$byte_len$1 = $152;
 } else {
  $21 = ((($20)) + 8|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = ($22|0)==(0);
  if ($23) {
   $$phi$trans$insert = ((($this)) + 48|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0;
   $$pre$phiZ2D = $$phi$trans$insert;$34 = $$pre;
  } else {
   $24 = Math_imul($22, $9)|0;
   $25 = $13 & 255;
   $26 = Math_imul($24, $25)|0;
   $27 = ((($this)) + 48|0);
   HEAP32[$27>>2] = $26;
   $$pre$phiZ2D = $27;$34 = $26;
  }
  $28 = ((($this)) + 20|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = ((($this)) + 24|0);
  $31 = HEAP32[$30>>2]|0;
  $32 = ((($this)) + 44|0);
  $33 = HEAP32[$32>>2]|0;
  FUNCTION_TABLE_viii[$29 & 8191]($31,$33,$34);
  (_SDL_ConvertAudio($4)|0);
  $35 = HEAP32[$32>>2]|0;
  $36 = ((($this)) + 52|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = ((($this)) + 16|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = ($37|0)==($39|0);
  if ($40) {
   $buf$0 = $35;$byte_len$1 = $37;
  } else {
   $41 = HEAP32[$19>>2]|0;
   $42 = HEAP32[$41>>2]|0;
   $43 = ($42|0)==(0|0);
   if ($43) {
    $44 = ($39>>>0)>($37>>>0);
    $$ = $44 ? $39 : $37;
    $45 = $$ << 1;
    $46 = ((($41)) + 4|0);
    HEAP32[$46>>2] = $45;
    $47 = (_SDL_malloc($45)|0);
    $48 = HEAP32[$19>>2]|0;
    HEAP32[$48>>2] = $47;
    $$pre7 = HEAP32[$36>>2]|0;
    $50 = $48;$53 = $$pre7;$61 = $47;
   } else {
    $50 = $41;$53 = $37;$61 = $42;
   }
   $49 = ((($50)) + 12|0);
   $51 = HEAP32[$49>>2]|0;
   $52 = (($53) + ($51))|0;
   $54 = ((($50)) + 4|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = ($52>>>0)>($55>>>0);
   if ($56) {
    $57 = ((($50)) + 16|0);
    $58 = HEAP32[$57>>2]|0;
    $59 = ($51>>>0)>($58>>>0);
    if ($59) {
     $60 = (($61) + ($58)|0);
     $62 = (($55) - ($58))|0;
     (_SDL_memmove($61,$60,$62)|0);
     $63 = HEAP32[$19>>2]|0;
     $64 = ((($63)) + 12|0);
     $65 = HEAP32[$64>>2]|0;
     $66 = ((($63)) + 16|0);
     $67 = HEAP32[$66>>2]|0;
     $68 = (($65) - ($67))|0;
     HEAP32[$64>>2] = $68;
     $$pre1$pre$i = HEAP32[$36>>2]|0;
     $$pre9$pre = HEAP32[$63>>2]|0;
     $$pre$i = $68;$$pre1$i = $$pre1$pre$i;$$pre9 = $$pre9$pre;$70 = $63;
    } else {
     HEAP32[$49>>2] = 0;
     $$pre$i = 0;$$pre1$i = $53;$$pre9 = $61;$70 = $50;
    }
    $69 = ((($70)) + 16|0);
    HEAP32[$69>>2] = 0;
    $72 = $$pre9;$73 = $$pre$i;$75 = $$pre1$i;
   } else {
    $72 = $61;$73 = $51;$75 = $53;
   }
   $71 = (($72) + ($73)|0);
   $74 = HEAP32[$32>>2]|0;
   (_SDL_memcpy($71,$74,$75)|0);
   $76 = HEAP32[$36>>2]|0;
   $77 = HEAP32[$19>>2]|0;
   $78 = ((($77)) + 12|0);
   $79 = HEAP32[$78>>2]|0;
   $80 = (($79) + ($76))|0;
   HEAP32[$78>>2] = $80;
   $81 = ((($77)) + 16|0);
   $82 = HEAP32[$81>>2]|0;
   $83 = (($80) - ($82))|0;
   $84 = HEAP32[$38>>2]|0;
   $85 = ($83>>>0)<($84>>>0);
   if ($85) {
    while(1) {
     $86 = HEAP32[$28>>2]|0;
     $87 = HEAP32[$30>>2]|0;
     $88 = HEAP32[$32>>2]|0;
     $89 = HEAP32[$$pre$phiZ2D>>2]|0;
     FUNCTION_TABLE_viii[$86 & 8191]($87,$88,$89);
     (_SDL_ConvertAudio($4)|0);
     $90 = HEAP32[$19>>2]|0;
     $91 = ((($90)) + 12|0);
     $92 = HEAP32[$91>>2]|0;
     $93 = HEAP32[$36>>2]|0;
     $94 = (($93) + ($92))|0;
     $95 = ((($90)) + 4|0);
     $96 = HEAP32[$95>>2]|0;
     $97 = ($94>>>0)>($96>>>0);
     if ($97) {
      $98 = ((($90)) + 16|0);
      $99 = HEAP32[$98>>2]|0;
      $100 = ($92>>>0)>($99>>>0);
      if ($100) {
       $101 = HEAP32[$90>>2]|0;
       $102 = (($101) + ($99)|0);
       $103 = (($96) - ($99))|0;
       (_SDL_memmove($101,$102,$103)|0);
       $104 = HEAP32[$19>>2]|0;
       $105 = ((($104)) + 12|0);
       $106 = HEAP32[$105>>2]|0;
       $107 = ((($104)) + 16|0);
       $108 = HEAP32[$107>>2]|0;
       $109 = (($106) - ($108))|0;
       HEAP32[$105>>2] = $109;
       $$pre1$pre$i1 = HEAP32[$36>>2]|0;
       $$pre$i3 = $109;$$pre1$i2 = $$pre1$pre$i1;$111 = $104;
      } else {
       HEAP32[$91>>2] = 0;
       $$pre$i3 = 0;$$pre1$i2 = $93;$111 = $90;
      }
      $110 = ((($111)) + 16|0);
      HEAP32[$110>>2] = 0;
      $113 = $111;$115 = $$pre$i3;$117 = $$pre1$i2;
     } else {
      $113 = $90;$115 = $92;$117 = $93;
     }
     $112 = HEAP32[$113>>2]|0;
     $114 = (($112) + ($115)|0);
     $116 = HEAP32[$32>>2]|0;
     (_SDL_memcpy($114,$116,$117)|0);
     $118 = HEAP32[$36>>2]|0;
     $119 = HEAP32[$19>>2]|0;
     $120 = ((($119)) + 12|0);
     $121 = HEAP32[$120>>2]|0;
     $122 = (($121) + ($118))|0;
     HEAP32[$120>>2] = $122;
     $123 = ((($119)) + 16|0);
     $124 = HEAP32[$123>>2]|0;
     $125 = (($122) - ($124))|0;
     $126 = HEAP32[$38>>2]|0;
     $127 = ($125>>>0)<($126>>>0);
     if (!($127)) {
      $$lcssa = $126;$129 = $119;$132 = $124;
      break;
     }
    }
   } else {
    $$lcssa = $84;$129 = $77;$132 = $82;
   }
   $128 = HEAP32[$129>>2]|0;
   $130 = ((($129)) + 16|0);
   $131 = (($128) + ($132)|0);
   $133 = (($132) + ($$lcssa))|0;
   HEAP32[$130>>2] = $133;
   $buf$0 = $131;$byte_len$1 = $$lcssa;
  }
 }
 $153 = ($buf$0|0)==(0|0);
 if ($153) {
  return;
 }
 $154 = (($byte_len$1|0) / ($3|0))&-1;
 $155 = ((($this)) + 10|0);
 $156 = HEAP8[$155>>0]|0;
 $157 = $156&255;
 $158 = (($154|0) / ($157|0))&-1;
 $159 = _emscripten_asm_const_iii(11, ($buf$0|0), ($158|0))|0;
 return;
}
function _SDL_ConvertAudio($cvt) {
 $cvt = $cvt|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($cvt)) + 16|0);
 $1 = HEAPU8[$0>>0]|(HEAPU8[$0+1>>0]<<8)|(HEAPU8[$0+2>>0]<<16)|(HEAPU8[$0+3>>0]<<24);
 $2 = ($1|0)==(0|0);
 if ($2) {
  (_SDL_SetError(607929,$vararg_buffer)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ((($cvt)) + 20|0);
 $4 = HEAPU8[$3>>0]|(HEAPU8[$3+1>>0]<<8)|(HEAPU8[$3+2>>0]<<16)|(HEAPU8[$3+3>>0]<<24);
 $5 = ((($cvt)) + 24|0);
 HEAP8[$5>>0]=$4&255;HEAP8[$5+1>>0]=($4>>8)&255;HEAP8[$5+2>>0]=($4>>16)&255;HEAP8[$5+3>>0]=$4>>24;
 $6 = ((($cvt)) + 40|0);
 $7 = HEAPU8[$6>>0]|(HEAPU8[$6+1>>0]<<8)|(HEAPU8[$6+2>>0]<<16)|(HEAPU8[$6+3>>0]<<24);
 $8 = ($7|0)==(0|0);
 if ($8) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $9 = ((($cvt)) + 80|0);
 HEAP8[$9>>0]=0&255;HEAP8[$9+1>>0]=(0>>8)&255;HEAP8[$9+2>>0]=(0>>16)&255;HEAP8[$9+3>>0]=0>>24;
 $10 = ((($cvt)) + 4|0);
 $11 = HEAPU8[$10>>0]|(HEAPU8[$10+1>>0]<<8);
 FUNCTION_TABLE_vii[$7 & 8191]($cvt,$11);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _Emscripten_CloseDevice($this) {
 $this = $this|0;
 var $$in = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 184|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 $3 = HEAP32[$1>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  $$in = $1;
 } else {
  _SDL_free($3);
  $5 = HEAP32[$0>>2]|0;
  HEAP32[$5>>2] = 0;
  $$in = $5;
 }
 _SDL_free($$in);
 HEAP32[$0>>2] = 0;
 return;
}
function _DUMMY_Available() {
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_getenv(609252)|0);
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  $2 = (_SDL_strcmp($0,609000)|0);
  $3 = ($2|0)==(0);
  if ($3) {
   $$0 = 1;
   return ($$0|0);
  }
 }
 $$0 = 0;
 return ($$0|0);
}
function _DUMMY_CreateDevice($devindex) {
 $devindex = $devindex|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_calloc(1,664)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_Error(0)|0);
  $$0 = 0;
  return ($$0|0);
 } else {
  $2 = ((($0)) + 4|0);
  HEAP32[$2>>2] = 4624;
  $3 = ((($0)) + 8|0);
  HEAP32[$3>>2] = 4625;
  $4 = ((($0)) + 24|0);
  HEAP32[$4>>2] = 4626;
  $5 = ((($0)) + 180|0);
  HEAP32[$5>>2] = 4627;
  $6 = ((($0)) + 108|0);
  HEAP32[$6>>2] = 4628;
  $7 = ((($0)) + 112|0);
  HEAP32[$7>>2] = 4629;
  $8 = ((($0)) + 116|0);
  HEAP32[$8>>2] = 4630;
  $9 = ((($0)) + 660|0);
  HEAP32[$9>>2] = 4631;
  $$0 = $0;
  return ($$0|0);
 }
 return (0)|0;
}
function _DUMMY_VideoInit($_this) {
 $_this = $_this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $mode = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $mode = sp;
 HEAP32[$mode>>2] = 370546692;
 $0 = ((($mode)) + 4|0);
 HEAP32[$0>>2] = 1024;
 $1 = ((($mode)) + 8|0);
 HEAP32[$1>>2] = 768;
 $2 = ((($mode)) + 12|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($mode)) + 16|0);
 HEAP32[$3>>2] = 0;
 $4 = (_SDL_AddBasicVideoDisplay($mode)|0);
 $5 = ($4|0)<(0);
 if ($5) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 (_SDL_memset($mode,0,20)|0);
 $6 = ((($_this)) + 244|0);
 $7 = HEAP32[$6>>2]|0;
 (_SDL_AddDisplayMode($7,$mode)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _DUMMY_VideoQuit($_this) {
 $_this = $_this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _DUMMY_SetDisplayMode($_this,$display,$mode) {
 $_this = $_this|0;
 $display = $display|0;
 $mode = $mode|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _DUMMY_PumpEvents($_this) {
 $_this = $_this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _SDL_DUMMY_CreateWindowFramebuffer($_this,$window,$format,$pixels,$pitch) {
 $_this = $_this|0;
 $window = $window|0;
 $format = $format|0;
 $pixels = $pixels|0;
 $pitch = $pitch|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $Amask = 0, $Bmask = 0, $Gmask = 0, $Rmask = 0, $bpp = 0;
 var $h = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $w = sp + 24|0;
 $h = sp + 20|0;
 $bpp = sp + 16|0;
 $Rmask = sp + 12|0;
 $Gmask = sp + 8|0;
 $Bmask = sp + 4|0;
 $Amask = sp;
 $0 = (_SDL_GetWindowData($window,608604)|0);
 _SDL_FreeSurface($0);
 (_SDL_PixelFormatEnumToMasks(370546692,$bpp,$Rmask,$Gmask,$Bmask,$Amask)|0);
 _SDL_GetWindowSize($window,$w,$h);
 $1 = HEAP32[$w>>2]|0;
 $2 = HEAP32[$h>>2]|0;
 $3 = HEAP32[$bpp>>2]|0;
 $4 = HEAP32[$Rmask>>2]|0;
 $5 = HEAP32[$Gmask>>2]|0;
 $6 = HEAP32[$Bmask>>2]|0;
 $7 = HEAP32[$Amask>>2]|0;
 $8 = (_SDL_CreateRGBSurface(0,$1,$2,$3,$4,$5,$6,$7)|0);
 $9 = ($8|0)==(0|0);
 if ($9) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 (_SDL_SetWindowData($window,608604,$8)|0);
 HEAP32[$format>>2] = 370546692;
 $10 = ((($8)) + 20|0);
 $11 = HEAP32[$10>>2]|0;
 HEAP32[$pixels>>2] = $11;
 $12 = ((($8)) + 16|0);
 $13 = HEAP32[$12>>2]|0;
 HEAP32[$pitch>>2] = $13;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GetWindowData($window,$name) {
 $window = $window|0;
 $name = $name|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $data$0 = 0, $data$02 = 0, $data$03 = 0, $data$03$lcssa = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ($name|0)==(0|0);
   if (!($6)) {
    $7 = HEAP8[$name>>0]|0;
    $8 = ($7<<24>>24)==(0);
    if (!($8)) {
     $9 = ((($window)) + 132|0);
     $data$02 = HEAP32[$9>>2]|0;
     $10 = ($data$02|0)==(0|0);
     if ($10) {
      $$0 = 0;
      STACKTOP = sp;return ($$0|0);
     } else {
      $data$03 = $data$02;
     }
     while(1) {
      $11 = HEAP32[$data$03>>2]|0;
      $12 = ($11|0)==(0|0);
      if (!($12)) {
       $13 = (_SDL_strcmp($11,$name)|0);
       $14 = ($13|0)==(0);
       if ($14) {
        $data$03$lcssa = $data$03;
        break;
       }
      }
      $17 = ((($data$03)) + 8|0);
      $data$0 = HEAP32[$17>>2]|0;
      $18 = ($data$0|0)==(0|0);
      if ($18) {
       $$0 = 0;
       label = 14;
       break;
      } else {
       $data$03 = $data$0;
      }
     }
     if ((label|0) == 14) {
      STACKTOP = sp;return ($$0|0);
     }
     $15 = ((($data$03$lcssa)) + 4|0);
     $16 = HEAP32[$15>>2]|0;
     $$0 = $16;
     STACKTOP = sp;return ($$0|0);
    }
   }
   HEAP32[$vararg_buffer3>>2] = 610375;
   (_SDL_SetError(611267,$vararg_buffer3)|0);
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SetWindowData($window,$name,$userdata) {
 $window = $window|0;
 $name = $name|0;
 $userdata = $userdata|0;
 var $$0 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $data$0 = 0, $data$03 = 0, $data$05 = 0, $data$05$lcssa = 0, $prev$04 = 0, $prev$04$lcssa = 0, $prev$04$phi = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ($name|0)==(0|0);
   if (!($6)) {
    $7 = HEAP8[$name>>0]|0;
    $8 = ($7<<24>>24)==(0);
    if (!($8)) {
     $9 = ((($window)) + 132|0);
     $data$03 = HEAP32[$9>>2]|0;
     $10 = ($data$03|0)==(0|0);
     L11: do {
      if (!($10)) {
       $data$05 = $data$03;$prev$04 = 0;
       while(1) {
        $11 = HEAP32[$data$05>>2]|0;
        $12 = ($11|0)==(0|0);
        if (!($12)) {
         $13 = (_SDL_strcmp($11,$name)|0);
         $14 = ($13|0)==(0);
         if ($14) {
          $$lcssa = $data$05;$data$05$lcssa = $data$05;$prev$04$lcssa = $prev$04;
          break;
         }
        }
        $23 = ((($data$05)) + 8|0);
        $data$0 = HEAP32[$23>>2]|0;
        $24 = ($data$0|0)==(0|0);
        if ($24) {
         break L11;
        } else {
         $prev$04$phi = $data$05;$data$05 = $data$0;$prev$04 = $prev$04$phi;
        }
       }
       $15 = ((($data$05$lcssa)) + 4|0);
       $16 = HEAP32[$15>>2]|0;
       $17 = ($userdata|0)==(0|0);
       if (!($17)) {
        HEAP32[$15>>2] = $userdata;
        $$0 = $16;
        STACKTOP = sp;return ($$0|0);
       }
       $18 = ($prev$04$lcssa|0)==(0|0);
       $19 = ((($data$05$lcssa)) + 8|0);
       $20 = HEAP32[$19>>2]|0;
       if ($18) {
        HEAP32[$9>>2] = $20;
       } else {
        $21 = ((($prev$04$lcssa)) + 8|0);
        HEAP32[$21>>2] = $20;
       }
       $22 = HEAP32[$$lcssa>>2]|0;
       _SDL_free($22);
       _SDL_free($data$05$lcssa);
       $$0 = $16;
       STACKTOP = sp;return ($$0|0);
      }
     } while(0);
     $25 = ($userdata|0)==(0|0);
     if ($25) {
      $$0 = 0;
      STACKTOP = sp;return ($$0|0);
     }
     $26 = (_SDL_malloc(12)|0);
     $27 = (_SDL_strdup($name)|0);
     HEAP32[$26>>2] = $27;
     $28 = ((($26)) + 4|0);
     HEAP32[$28>>2] = $userdata;
     $29 = HEAP32[$9>>2]|0;
     $30 = ((($26)) + 8|0);
     HEAP32[$30>>2] = $29;
     HEAP32[$9>>2] = $26;
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
   }
   HEAP32[$vararg_buffer3>>2] = 610375;
   (_SDL_SetError(611267,$vararg_buffer3)|0);
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_DUMMY_UpdateWindowFramebuffer($_this,$window,$rects,$numrects) {
 $_this = $_this|0;
 $window = $window|0;
 $rects = $rects|0;
 $numrects = $numrects|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $file = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_ptr3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $file = sp + 16|0;
 $0 = (_SDL_GetWindowData($window,608604)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  $2 = (_SDL_SetError(608622,$vararg_buffer)|0);
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = (_SDL_getenv(608661)|0);
 $4 = ($3|0)==(0|0);
 if ($4) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = (_SDL_GetWindowID($window)|0);
 $6 = HEAP32[301386]|0;
 $7 = (($6) + 1)|0;
 HEAP32[301386] = $7;
 HEAP32[$vararg_buffer1>>2] = $5;
 $vararg_ptr3 = ((($vararg_buffer1)) + 4|0);
 HEAP32[$vararg_ptr3>>2] = $7;
 (_SDL_snprintf($file,128,608689,$vararg_buffer1)|0);
 $8 = (_SDL_RWFromFile($file,608916)|0);
 (_SDL_SaveBMP_RW($0,$8,1)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GetWindowID($window) {
 $window = $window|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ((($window)) + 4|0);
   $7 = HEAP32[$6>>2]|0;
   $$0 = $7;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_RWFromFile($file,$mode) {
 $file = $file|0;
 $mode = $mode|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $vararg_buffer = 0;
 var $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($file|0)==(0|0);
 if (!($0)) {
  $1 = HEAP8[$file>>0]|0;
  $2 = ($1<<24>>24)!=(0);
  $3 = ($mode|0)!=(0|0);
  $or$cond = $3 & $2;
  if ($or$cond) {
   $4 = HEAP8[$mode>>0]|0;
   $5 = ($4<<24>>24)==(0);
   if (!($5)) {
    $6 = (_fopen($file,$mode)|0);
    $7 = ($6|0)==(0|0);
    if ($7) {
     HEAP32[$vararg_buffer1>>2] = $file;
     (_SDL_SetError(608759,$vararg_buffer1)|0);
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
    $8 = (_SDL_malloc(36)|0);
    $9 = ($8|0)==(0|0);
    if ($9) {
     (_SDL_Error(0)|0);
     $$0 = $8;
     STACKTOP = sp;return ($$0|0);
    } else {
     $10 = ((($8)) + 20|0);
     HEAP32[$8>>2] = 4632;
     $11 = ((($8)) + 4|0);
     HEAP32[$11>>2] = 4633;
     $12 = ((($8)) + 8|0);
     HEAP32[$12>>2] = 4634;
     $13 = ((($8)) + 12|0);
     HEAP32[$13>>2] = 4635;
     $14 = ((($8)) + 16|0);
     HEAP32[$14>>2] = 4636;
     $15 = ((($8)) + 24|0);
     $16 = ((($8)) + 28|0);
     HEAP32[$16>>2] = $6;
     HEAP32[$15>>2] = 1;
     HEAP32[$10>>2] = 2;
     $$0 = $8;
     STACKTOP = sp;return ($$0|0);
    }
   }
  }
 }
 (_SDL_SetError(608712,$vararg_buffer)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _stdio_size($context) {
 $context = $context|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($context)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (FUNCTION_TABLE_iiiii[$1 & 8191]($context,0,0,1)|0);
 $3 = tempRet0;
 $4 = ($3|0)<(0);
 if ($4) {
  $10 = -1;$11 = -1;
  tempRet0 = ($10);
  return ($11|0);
 }
 $5 = HEAP32[$0>>2]|0;
 $6 = (FUNCTION_TABLE_iiiii[$5 & 8191]($context,0,0,2)|0);
 $7 = tempRet0;
 $8 = HEAP32[$0>>2]|0;
 (FUNCTION_TABLE_iiiii[$8 & 8191]($context,$2,$3,0)|0);
 $9 = tempRet0;
 $10 = $7;$11 = $6;
 tempRet0 = ($10);
 return ($11|0);
}
function _stdio_seek($context,$0,$1,$whence) {
 $context = $context|0;
 $0 = $0|0;
 $1 = $1|0;
 $whence = $whence|0;
 var $$0$in = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($context)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (___fseeko($3,$0,$whence)|0);
 $5 = ($4|0)==(0);
 if ($5) {
  $6 = HEAP32[$2>>2]|0;
  $7 = (___ftello($6)|0);
  $$0$in = $7;
 } else {
  $8 = (_SDL_Error(3)|0);
  $$0$in = $8;
 }
 $9 = ($$0$in|0)<(0);
 $10 = $9 << 31 >> 31;
 tempRet0 = ($10);
 return ($$0$in|0);
}
function _stdio_read($context,$ptr,$size,$maxnum) {
 $context = $context|0;
 $ptr = $ptr|0;
 $size = $size|0;
 $maxnum = $maxnum|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($context)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_fread($ptr,$size,$maxnum,$1)|0);
 $3 = ($2|0)==(0);
 if (!($3)) {
  return ($2|0);
 }
 $4 = HEAP32[$0>>2]|0;
 $5 = (_ferror($4)|0);
 $6 = ($5|0)==(0);
 if ($6) {
  return ($2|0);
 }
 (_SDL_Error(1)|0);
 return ($2|0);
}
function _stdio_write($context,$ptr,$size,$num) {
 $context = $context|0;
 $ptr = $ptr|0;
 $size = $size|0;
 $num = $num|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($context)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_fwrite($ptr,$size,$num,$1)|0);
 $3 = ($2|0)==(0);
 if (!($3)) {
  return ($2|0);
 }
 $4 = HEAP32[$0>>2]|0;
 $5 = (_ferror($4)|0);
 $6 = ($5|0)==(0);
 if ($6) {
  return ($2|0);
 }
 (_SDL_Error(2)|0);
 return ($2|0);
}
function _stdio_close($context) {
 $context = $context|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $status$0 = 0, $status$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($context|0)==(0|0);
 if ($0) {
  $status$1 = 0;
  return ($status$1|0);
 }
 $1 = ((($context)) + 24|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0);
 if ($3) {
  $status$0 = 0;
 } else {
  $4 = ((($context)) + 28|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = (_fclose($5)|0);
  $7 = ($6|0)==(0);
  if ($7) {
   $status$0 = 0;
  } else {
   $8 = (_SDL_Error(2)|0);
   $status$0 = $8;
  }
 }
 _SDL_free($context);
 $status$1 = $status$0;
 return ($status$1|0);
}
function _SDL_SaveBMP_RW($saveme,$dst,$freedst) {
 $saveme = $saveme|0;
 $dst = $dst|0;
 $freedst = $freedst|0;
 var $$ = 0, $$0 = 0, $$pre27 = 0, $$pre28 = 0, $$pre29 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $biClrUsed$0 = 0, $exitcond = 0, $exitcond26 = 0, $format = 0, $i$012 = 0, $i$18$us16 = 0, $magic = 0, $not$ = 0, $or$cond = 0;
 var $padbyte = 0, $save32bit$0 = 0, $surface$04 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $magic = sp + 56|0;
 $format = sp + 12|0;
 $padbyte = sp + 58|0;
 HEAP16[$magic>>1] = 19778;
 $0 = ($dst|0)==(0|0);
 if ($0) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ((($saveme)) + 52|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 68|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 & 256;
 $6 = ($5|0)==(0);
 $7 = ((($saveme)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 if ($6) {
  $9 = ((($8)) + 24|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ($10|0)==(0);
  if ($11) {
   $save32bit$0 = 0;
  } else {
   label = 4;
  }
 } else {
  label = 4;
 }
 if ((label|0) == 4) {
  $save32bit$0 = 1;
 }
 $12 = ((($8)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13|0)==(0|0);
 $or$cond = $save32bit$0 | $14;
 $15 = ((($8)) + 8|0);
 $16 = HEAP8[$15>>0]|0;
 $17 = $16&255;
 do {
  if ($or$cond) {
   $19 = ($16<<24>>24)==(24);
   if ($19) {
    $20 = ((($8)) + 12|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = ($21|0)==(16711680);
    if ($22) {
     $23 = ((($8)) + 16|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ($24|0)==(65280);
     if ($25) {
      $26 = ((($8)) + 20|0);
      $27 = HEAP32[$26>>2]|0;
      $28 = ($27|0)==(255);
      if ($28) {
       $surface$04 = $saveme;
       label = 18;
       break;
      }
     }
    }
   }
   if ($save32bit$0) {
    (_SDL_InitFormat($format,372645892)|0);
   } else {
    (_SDL_InitFormat($format,390076419)|0);
   }
   $29 = (_SDL_ConvertSurface($saveme,$format,0)|0);
   $30 = ($29|0)==(0|0);
   if ($30) {
    $31 = ((($format)) + 8|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = $32&255;
    HEAP32[$vararg_buffer1>>2] = $33;
    (_SDL_SetError(608807,$vararg_buffer1)|0);
    break;
   } else {
    $surface$04 = $29;
    label = 18;
    break;
   }
  } else {
   $18 = ($16<<24>>24)==(8);
   if ($18) {
    $surface$04 = $saveme;
    label = 18;
   } else {
    HEAP32[$vararg_buffer>>2] = $17;
    (_SDL_SetError(608776,$vararg_buffer)|0);
   }
  }
 } while(0);
 if ((label|0) == 18) {
  $34 = (_SDL_LockSurface($surface$04)|0);
  $35 = ($34|0)==(0);
  if ($35) {
   $36 = ((($surface$04)) + 8|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = ((($surface$04)) + 4|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = ((($39)) + 9|0);
   $41 = HEAP8[$40>>0]|0;
   $42 = $41&255;
   $43 = Math_imul($42, $37)|0;
   $44 = ((($dst)) + 4|0);
   $45 = HEAP32[$44>>2]|0;
   $46 = (FUNCTION_TABLE_iiiii[$45 & 8191]($dst,0,0,1)|0);
   $47 = tempRet0;
   _SDL_ClearError();
   $48 = ((($dst)) + 12|0);
   $49 = HEAP32[$48>>2]|0;
   (FUNCTION_TABLE_iiiii[$49 & 8191]($dst,$magic,2,1)|0);
   (_SDL_WriteLE32($dst,0)|0);
   (_SDL_WriteLE16($dst,0)|0);
   (_SDL_WriteLE16($dst,0)|0);
   (_SDL_WriteLE32($dst,0)|0);
   $50 = HEAP32[$36>>2]|0;
   $51 = ((($surface$04)) + 12|0);
   $52 = HEAP32[$51>>2]|0;
   $53 = HEAP32[$38>>2]|0;
   $54 = ((($53)) + 8|0);
   $55 = HEAP8[$54>>0]|0;
   $56 = $55&255;
   $57 = ((($surface$04)) + 16|0);
   $58 = HEAP32[$57>>2]|0;
   $59 = Math_imul($58, $52)|0;
   $60 = ((($53)) + 4|0);
   $61 = HEAP32[$60>>2]|0;
   $62 = ($61|0)==(0|0);
   if ($62) {
    $biClrUsed$0 = 0;
   } else {
    $63 = HEAP32[$61>>2]|0;
    $biClrUsed$0 = $63;
   }
   (_SDL_WriteLE32($dst,40)|0);
   (_SDL_WriteLE32($dst,$50)|0);
   (_SDL_WriteLE32($dst,$52)|0);
   (_SDL_WriteLE16($dst,1)|0);
   (_SDL_WriteLE16($dst,$56)|0);
   (_SDL_WriteLE32($dst,0)|0);
   (_SDL_WriteLE32($dst,$59)|0);
   (_SDL_WriteLE32($dst,0)|0);
   (_SDL_WriteLE32($dst,0)|0);
   (_SDL_WriteLE32($dst,$biClrUsed$0)|0);
   (_SDL_WriteLE32($dst,0)|0);
   $64 = HEAP32[$38>>2]|0;
   $65 = ((($64)) + 4|0);
   $66 = HEAP32[$65>>2]|0;
   $67 = ($66|0)==(0|0);
   if (!($67)) {
    $68 = ((($66)) + 4|0);
    $69 = HEAP32[$68>>2]|0;
    $70 = HEAP32[$66>>2]|0;
    $71 = ($70|0)>(0);
    if ($71) {
     $i$012 = 0;
     while(1) {
      $72 = HEAP32[$48>>2]|0;
      $73 = (((($69) + ($i$012<<2)|0)) + 2|0);
      (FUNCTION_TABLE_iiiii[$72 & 8191]($dst,$73,1,1)|0);
      $74 = HEAP32[$48>>2]|0;
      $75 = (((($69) + ($i$012<<2)|0)) + 1|0);
      (FUNCTION_TABLE_iiiii[$74 & 8191]($dst,$75,1,1)|0);
      $76 = HEAP32[$48>>2]|0;
      $77 = (($69) + ($i$012<<2)|0);
      (FUNCTION_TABLE_iiiii[$76 & 8191]($dst,$77,1,1)|0);
      $78 = HEAP32[$48>>2]|0;
      $79 = (((($69) + ($i$012<<2)|0)) + 3|0);
      (FUNCTION_TABLE_iiiii[$78 & 8191]($dst,$79,1,1)|0);
      $80 = (($i$012) + 1)|0;
      $exitcond26 = ($80|0)==($70|0);
      if ($exitcond26) {
       break;
      } else {
       $i$012 = $80;
      }
     }
    }
   }
   $81 = HEAP32[$44>>2]|0;
   $82 = (FUNCTION_TABLE_iiiii[$81 & 8191]($dst,0,0,1)|0);
   $83 = tempRet0;
   $84 = (_i64Subtract(($82|0),($83|0),($46|0),($47|0))|0);
   $85 = tempRet0;
   $86 = HEAP32[$44>>2]|0;
   $87 = (_i64Add(($46|0),($47|0),10,0)|0);
   $88 = tempRet0;
   (FUNCTION_TABLE_iiiii[$86 & 8191]($dst,$87,$88,0)|0);
   $89 = tempRet0;
   $90 = ($89|0)<(0);
   if ($90) {
    (_SDL_Error(3)|0);
   }
   (_SDL_WriteLE32($dst,$84)|0);
   $91 = HEAP32[$44>>2]|0;
   $92 = (_i64Add(($84|0),0,($46|0),($47|0))|0);
   $93 = tempRet0;
   (FUNCTION_TABLE_iiiii[$91 & 8191]($dst,$92,$93,0)|0);
   $94 = tempRet0;
   $95 = ($94|0)<(0);
   if ($95) {
    (_SDL_Error(3)|0);
   }
   $96 = ((($surface$04)) + 20|0);
   $97 = HEAP32[$96>>2]|0;
   $98 = HEAP32[$51>>2]|0;
   $99 = HEAP32[$57>>2]|0;
   $100 = Math_imul($99, $98)|0;
   $101 = (($97) + ($100)|0);
   $102 = (($43|0) % 4)&-1;
   $103 = ($102|0)==(0);
   $104 = (4 - ($102))|0;
   $$ = $103 ? 0 : $104;
   $105 = ($100|0)>(0);
   L42: do {
    if ($105) {
     $106 = ($$|0)==(0);
     L44: do {
      if ($106) {
       $107 = (0 - ($99))|0;
       $108 = (($101) + ($107)|0);
       $109 = HEAP32[$48>>2]|0;
       $110 = (FUNCTION_TABLE_iiiii[$109 & 8191]($dst,$108,1,$43)|0);
       $111 = ($110|0)==($43|0);
       if ($111) {
        $114 = $108;
        while(1) {
         $112 = HEAP32[$96>>2]|0;
         $113 = ($114>>>0)>($112>>>0);
         if (!($113)) {
          break L42;
         }
         $$pre29 = HEAP32[$57>>2]|0;
         $115 = (0 - ($$pre29))|0;
         $116 = (($114) + ($115)|0);
         $117 = HEAP32[$48>>2]|0;
         $118 = (FUNCTION_TABLE_iiiii[$117 & 8191]($dst,$116,1,$43)|0);
         $119 = ($118|0)==($43|0);
         if ($119) {
          $114 = $116;
         } else {
          break;
         }
        }
       }
      } else {
       $120 = ($$|0)>(0);
       $121 = (0 - ($99))|0;
       $122 = (($101) + ($121)|0);
       $123 = HEAP32[$48>>2]|0;
       $124 = (FUNCTION_TABLE_iiiii[$123 & 8191]($dst,$122,1,$43)|0);
       $125 = ($124|0)==($43|0);
       if (!($120)) {
        if ($125) {
         $138 = $122;
        } else {
         break;
        }
        while(1) {
         $136 = HEAP32[$96>>2]|0;
         $137 = ($138>>>0)>($136>>>0);
         if (!($137)) {
          break L42;
         }
         $$pre27 = HEAP32[$57>>2]|0;
         $139 = (0 - ($$pre27))|0;
         $140 = (($138) + ($139)|0);
         $141 = HEAP32[$48>>2]|0;
         $142 = (FUNCTION_TABLE_iiiii[$141 & 8191]($dst,$140,1,$43)|0);
         $143 = ($142|0)==($43|0);
         if ($143) {
          $138 = $140;
         } else {
          break L44;
         }
        }
       }
       if ($125) {
        $130 = $122;
        while(1) {
         HEAP8[$padbyte>>0] = 0;
         $i$18$us16 = 0;
         while(1) {
          $126 = HEAP32[$48>>2]|0;
          (FUNCTION_TABLE_iiiii[$126 & 8191]($dst,$padbyte,1,1)|0);
          $127 = (($i$18$us16) + 1)|0;
          $exitcond = ($127|0)==($$|0);
          if ($exitcond) {
           break;
          } else {
           $i$18$us16 = $127;
          }
         }
         $128 = HEAP32[$96>>2]|0;
         $129 = ($130>>>0)>($128>>>0);
         if (!($129)) {
          break L42;
         }
         $$pre28 = HEAP32[$57>>2]|0;
         $131 = (0 - ($$pre28))|0;
         $132 = (($130) + ($131)|0);
         $133 = HEAP32[$48>>2]|0;
         $134 = (FUNCTION_TABLE_iiiii[$133 & 8191]($dst,$132,1,$43)|0);
         $135 = ($134|0)==($43|0);
         if ($135) {
          $130 = $132;
         } else {
          break;
         }
        }
       }
      }
     } while(0);
     (_SDL_Error(2)|0);
    }
   } while(0);
   $144 = HEAP32[$44>>2]|0;
   $145 = (FUNCTION_TABLE_iiiii[$144 & 8191]($dst,0,0,1)|0);
   $146 = tempRet0;
   $147 = (_i64Subtract(($145|0),($146|0),($46|0),($47|0))|0);
   $148 = tempRet0;
   $149 = HEAP32[$44>>2]|0;
   $150 = (_i64Add(($46|0),($47|0),2,0)|0);
   $151 = tempRet0;
   (FUNCTION_TABLE_iiiii[$149 & 8191]($dst,$150,$151,0)|0);
   $152 = tempRet0;
   $153 = ($152|0)<(0);
   if ($153) {
    (_SDL_Error(3)|0);
   }
   (_SDL_WriteLE32($dst,$147)|0);
   $154 = HEAP32[$44>>2]|0;
   $155 = (_i64Add(($147|0),0,($46|0),($47|0))|0);
   $156 = tempRet0;
   (FUNCTION_TABLE_iiiii[$154 & 8191]($dst,$155,$156,0)|0);
   $157 = tempRet0;
   $158 = ($157|0)<(0);
   if ($158) {
    (_SDL_Error(3)|0);
   }
   _SDL_UnlockSurface($surface$04);
   $159 = ($surface$04|0)==($saveme|0);
   if (!($159)) {
    _SDL_FreeSurface($surface$04);
   }
  }
 }
 $160 = ($freedst|0)==(0);
 if (!($160)) {
  $161 = ((($dst)) + 16|0);
  $162 = HEAP32[$161>>2]|0;
  (FUNCTION_TABLE_ii[$162 & 8191]($dst)|0);
 }
 $163 = (_SDL_GetError()|0);
 $164 = (_SDL_strcmp($163,1282466)|0);
 $not$ = ($164|0)!=(0);
 $165 = $not$ << 31 >> 31;
 $$0 = $165;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_ClearError() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[156036] = 0;
 return;
}
function _SDL_WriteLE32($dst,$value) {
 $dst = $dst|0;
 $value = $value|0;
 var $0 = 0, $1 = 0, $2 = 0, $swapped = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $swapped = sp;
 HEAP32[$swapped>>2] = $value;
 $0 = ((($dst)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (FUNCTION_TABLE_iiiii[$1 & 8191]($dst,$swapped,4,1)|0);
 STACKTOP = sp;return ($2|0);
}
function _SDL_WriteLE16($dst,$value) {
 $dst = $dst|0;
 $value = $value|0;
 var $0 = 0, $1 = 0, $2 = 0, $swapped = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $swapped = sp;
 HEAP16[$swapped>>1] = $value;
 $0 = ((($dst)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (FUNCTION_TABLE_iiiii[$1 & 8191]($dst,$swapped,2,1)|0);
 STACKTOP = sp;return ($2|0);
}
function _SDL_DUMMY_DestroyWindowFramebuffer($_this,$window) {
 $_this = $_this|0;
 $window = $window|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_SetWindowData($window,608604,0)|0);
 _SDL_FreeSurface($0);
 return;
}
function _DUMMY_DeleteDevice($device) {
 $device = $device|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _SDL_free($device);
 return;
}
function _DISKAUD_Init($impl) {
 $impl = $impl|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($impl)) + 4|0);
 HEAP32[$0>>2] = 4637;
 $1 = ((($impl)) + 12|0);
 HEAP32[$1>>2] = 4638;
 $2 = ((($impl)) + 16|0);
 HEAP32[$2>>2] = 4639;
 $3 = ((($impl)) + 24|0);
 HEAP32[$3>>2] = 4640;
 $4 = ((($impl)) + 32|0);
 HEAP32[$4>>2] = 4641;
 HEAP32[$impl>>2] = 4642;
 $5 = ((($impl)) + 72|0);
 HEAP32[$5>>2] = 1;
 return 1;
}
function _DISKAUD_OpenDevice($this,$handle,$devname,$iscapture) {
 $this = $this|0;
 $handle = $handle|0;
 $devname = $devname|0;
 $iscapture = $iscapture|0;
 var $$$i = 0, $$0 = 0, $$0$i = 0, $$in$i = 0, $$in$i1 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($handle|0)!=(0|0);
 $1 = $0 ? 0 : $devname;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $3 = (_SDL_getenv(608866)|0);
  $4 = ($3|0)==(0|0);
  $$$i = $4 ? 608884 : $3;
  $$0$i = $$$i;
 } else {
  $$0$i = $1;
 }
 $5 = (_SDL_getenv(608897)|0);
 $6 = (_SDL_malloc(16)|0);
 $7 = ((($this)) + 184|0);
 HEAP32[$7>>2] = $6;
 $8 = ($6|0)==(0|0);
 if ($8) {
  $9 = (_SDL_Error(0)|0);
  $$0 = $9;
  STACKTOP = sp;return ($$0|0);
 }
 (_SDL_memset($6,0,16)|0);
 $10 = ((($this)) + 16|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = HEAP32[$7>>2]|0;
 $13 = ((($12)) + 8|0);
 HEAP32[$13>>2] = $11;
 $14 = ($5|0)==(0|0);
 if ($14) {
  $17 = $12;$18 = 150;
 } else {
  $15 = (_SDL_atoi($5)|0);
  $$pre = HEAP32[$7>>2]|0;
  $17 = $$pre;$18 = $15;
 }
 $16 = ((($17)) + 12|0);
 HEAP32[$16>>2] = $18;
 $19 = (_SDL_RWFromFile($$0$i,608916)|0);
 $20 = HEAP32[$7>>2]|0;
 HEAP32[$20>>2] = $19;
 $21 = ($19|0)==(0|0);
 if ($21) {
  $22 = ($20|0)==(0|0);
  if ($22) {
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  }
  $23 = ((($20)) + 4|0);
  $24 = HEAP32[$23>>2]|0;
  _SDL_free($24);
  $25 = HEAP32[$7>>2]|0;
  $26 = ((($25)) + 4|0);
  HEAP32[$26>>2] = 0;
  $27 = HEAP32[$25>>2]|0;
  $28 = ($27|0)==(0|0);
  if ($28) {
   $$in$i = $25;
  } else {
   $29 = ((($27)) + 16|0);
   $30 = HEAP32[$29>>2]|0;
   (FUNCTION_TABLE_ii[$30 & 8191]($27)|0);
   $31 = HEAP32[$7>>2]|0;
   HEAP32[$31>>2] = 0;
   $$in$i = $31;
  }
  _SDL_free($$in$i);
  HEAP32[$7>>2] = 0;
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $32 = ((($20)) + 8|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = (_SDL_malloc($33)|0);
 $35 = HEAP32[$7>>2]|0;
 $36 = ((($35)) + 4|0);
 HEAP32[$36>>2] = $34;
 $37 = ($34|0)==(0|0);
 if (!($37)) {
  $46 = ((($this)) + 11|0);
  $47 = HEAP8[$46>>0]|0;
  $48 = $47&255;
  $49 = HEAP32[$10>>2]|0;
  (_SDL_memset($34,$48,$49)|0);
  $50 = HEAP32[37544]|0;
  HEAP32[$vararg_buffer>>2] = $$0$i;
  (_fprintf($50,608919,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $38 = ($35|0)==(0|0);
 if ($38) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 _SDL_free(0);
 $39 = HEAP32[$7>>2]|0;
 $40 = ((($39)) + 4|0);
 HEAP32[$40>>2] = 0;
 $41 = HEAP32[$39>>2]|0;
 $42 = ($41|0)==(0|0);
 if ($42) {
  $$in$i1 = $39;
 } else {
  $43 = ((($41)) + 16|0);
  $44 = HEAP32[$43>>2]|0;
  (FUNCTION_TABLE_ii[$44 & 8191]($41)|0);
  $45 = HEAP32[$7>>2]|0;
  HEAP32[$45>>2] = 0;
  $$in$i1 = $45;
 }
 _SDL_free($$in$i1);
 HEAP32[$7>>2] = 0;
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_atoi($string) {
 $string = $string|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_atoi($string)|0);
 return ($0|0);
}
function _DISKAUD_WaitDevice($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 184|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 _SDL_Delay($3);
 return;
}
function _DISKAUD_PlayDevice($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 184|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($1)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($1)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (FUNCTION_TABLE_iiiii[$4 & 8191]($2,$6,1,$8)|0);
 $10 = HEAP32[$0>>2]|0;
 $11 = ((($10)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($9|0)==($12|0);
 if ($13) {
  return;
 }
 _SDL_OpenedAudioDeviceDisconnected($this);
 return;
}
function _SDL_OpenedAudioDeviceDisconnected($device) {
 $device = $device|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $event = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $event = sp;
 $0 = ((($device)) + 136|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  STACKTOP = sp;return;
 }
 $3 = HEAP32[(1205592)>>2]|0;
 FUNCTION_TABLE_vi[$3 & 8191]($device);
 HEAP32[$0>>2] = 0;
 $4 = HEAP32[(1205596)>>2]|0;
 FUNCTION_TABLE_vi[$4 & 8191]($device);
 $5 = (_SDL_EventState(4353,-1)|0);
 $6 = ($5<<24>>24)==(1);
 if (!($6)) {
  STACKTOP = sp;return;
 }
 (_SDL_memset($event,0,56)|0);
 HEAP32[$event>>2] = 4353;
 $7 = HEAP32[$device>>2]|0;
 $8 = ((($event)) + 8|0);
 HEAP32[$8>>2] = $7;
 $9 = ((($device)) + 132|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($10|0)!=(0);
 $12 = $11&1;
 $13 = ((($event)) + 12|0);
 HEAP8[$13>>0] = $12;
 (_SDL_PushEvent($event)|0);
 STACKTOP = sp;return;
}
function _DISKAUD_GetDeviceBuf($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 184|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 return ($3|0);
}
function _DISKAUD_CloseDevice($this) {
 $this = $this|0;
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 184|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 $3 = ((($1)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 _SDL_free($4);
 $5 = HEAP32[$0>>2]|0;
 $6 = ((($5)) + 4|0);
 HEAP32[$6>>2] = 0;
 $7 = HEAP32[$5>>2]|0;
 $8 = ($7|0)==(0|0);
 if ($8) {
  $$in = $5;
 } else {
  $9 = ((($7)) + 16|0);
  $10 = HEAP32[$9>>2]|0;
  (FUNCTION_TABLE_ii[$10 & 8191]($7)|0);
  $11 = HEAP32[$0>>2]|0;
  HEAP32[$11>>2] = 0;
  $$in = $11;
 }
 _SDL_free($$in);
 HEAP32[$0>>2] = 0;
 return;
}
function _DISKAUD_DetectDevices() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 _SDL_AddAudioDevice(0,610283,(1));
 return;
}
function _SDL_AddAudioDevice($iscapture,$name,$handle) {
 $iscapture = $iscapture|0;
 $name = $name|0;
 $handle = $handle|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $event = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $event = sp;
 $0 = ($iscapture|0)==(0);
 $1 = (_SDL_strlen($name)|0);
 $2 = (($1) + 9)|0;
 $3 = (_SDL_malloc($2)|0);
 $4 = ($3|0)==(0|0);
 do {
  if ($0) {
   if ($4) {
    STACKTOP = sp;return;
   } else {
    HEAP32[$3>>2] = $handle;
    $11 = ((($3)) + 8|0);
    $12 = (($1) + 1)|0;
    (_SDL_strlcpy($11,$name,$12)|0);
    $13 = HEAP32[(1205652)>>2]|0;
    $14 = ((($3)) + 4|0);
    HEAP32[$14>>2] = $13;
    HEAP32[(1205652)>>2] = $3;
    $15 = HEAP32[(1205644)>>2]|0;
    $16 = (($15) + 1)|0;
    HEAP32[(1205644)>>2] = $16;
    $18 = $15;
    break;
   }
  } else {
   if ($4) {
    STACKTOP = sp;return;
   } else {
    HEAP32[$3>>2] = $handle;
    $5 = ((($3)) + 8|0);
    $6 = (($1) + 1)|0;
    (_SDL_strlcpy($5,$name,$6)|0);
    $7 = HEAP32[(1205656)>>2]|0;
    $8 = ((($3)) + 4|0);
    HEAP32[$8>>2] = $7;
    HEAP32[(1205656)>>2] = $3;
    $9 = HEAP32[(1205648)>>2]|0;
    $10 = (($9) + 1)|0;
    HEAP32[(1205648)>>2] = $10;
    $18 = $9;
    break;
   }
  }
 } while(0);
 $17 = ($18|0)==(-1);
 if ($17) {
  STACKTOP = sp;return;
 }
 $19 = (_SDL_EventState(4352,-1)|0);
 $20 = ($19<<24>>24)==(1);
 if (!($20)) {
  STACKTOP = sp;return;
 }
 (_SDL_memset($event,0,56)|0);
 HEAP32[$event>>2] = 4352;
 $21 = ((($event)) + 8|0);
 HEAP32[$21>>2] = $18;
 $22 = $iscapture&255;
 $23 = ((($event)) + 12|0);
 HEAP8[$23>>0] = $22;
 (_SDL_PushEvent($event)|0);
 STACKTOP = sp;return;
}
function _DUMMYAUD_Init($impl) {
 $impl = $impl|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($impl)) + 4|0);
 HEAP32[$0>>2] = 4643;
 $1 = ((($impl)) + 64|0);
 HEAP32[$1>>2] = 1;
 return 1;
}
function _DUMMYAUD_OpenDevice($this,$handle,$devname,$iscapture) {
 $this = $this|0;
 $handle = $handle|0;
 $devname = $devname|0;
 $iscapture = $iscapture|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _SDL_InitSubSystem($flags) {
 $flags = $flags|0;
 var $$0$ = 0, $$01 = 0, $$pre = 0, $$pre10 = 0, $$pre2 = 0, $$pre4 = 0, $$pre6 = 0, $$pre8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _SDL_ClearError();
 _SDL_TicksInit();
 $0 = $flags >>> 4;
 $1 = $0 & 512;
 $2 = $1 | $flags;
 $3 = $2 & 544;
 $4 = ($3|0)==(0);
 $5 = $2 | 16384;
 $$0$ = $4 ? $2 : $5;
 $6 = $$0$ & 16384;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $8 = HEAP8[(1280339)>>0]|0;
  $9 = ($8<<24>>24)==(0);
  do {
   if ($9) {
    $10 = (_SDL_StartEventLoop()|0);
    $11 = ($10|0)<(0);
    if ($11) {
     $$01 = -1;
     STACKTOP = sp;return ($$01|0);
    } else {
     (_SDL_QuitInit()|0);
     $$pre = HEAP8[(1280339)>>0]|0;
     $13 = $$pre;
     break;
    }
   } else {
    $13 = $8;
   }
  } while(0);
  $12 = (($13) + 1)<<24>>24;
  HEAP8[(1280339)>>0] = $12;
 }
 $14 = $$0$ & 1;
 $15 = ($14|0)==(0);
 if (!($15)) {
  $16 = HEAP8[1280325]|0;
  $17 = ($16<<24>>24)==(0);
  do {
   if ($17) {
    $18 = (_SDL_TimerInit()|0);
    $19 = ($18|0)<(0);
    if ($19) {
     $$01 = -1;
     STACKTOP = sp;return ($$01|0);
    } else {
     $$pre2 = HEAP8[1280325]|0;
     $21 = $$pre2;
     break;
    }
   } else {
    $21 = $16;
   }
  } while(0);
  $20 = (($21) + 1)<<24>>24;
  HEAP8[1280325] = $20;
 }
 $22 = $$0$ & 32;
 $23 = ($22|0)==(0);
 if (!($23)) {
  $24 = HEAP8[(1280330)>>0]|0;
  $25 = ($24<<24>>24)==(0);
  do {
   if ($25) {
    $26 = (_SDL_VideoInit(0)|0);
    $27 = ($26|0)<(0);
    if ($27) {
     $$01 = -1;
     STACKTOP = sp;return ($$01|0);
    } else {
     $$pre4 = HEAP8[(1280330)>>0]|0;
     $29 = $$pre4;
     break;
    }
   } else {
    $29 = $24;
   }
  } while(0);
  $28 = (($29) + 1)<<24>>24;
  HEAP8[(1280330)>>0] = $28;
 }
 $30 = $$0$ & 16;
 $31 = ($30|0)==(0);
 if (!($31)) {
  $32 = HEAP8[(1280329)>>0]|0;
  $33 = ($32<<24>>24)==(0);
  do {
   if ($33) {
    $34 = (_SDL_AudioInit(0)|0);
    $35 = ($34|0)<(0);
    if ($35) {
     $$01 = -1;
     STACKTOP = sp;return ($$01|0);
    } else {
     $$pre6 = HEAP8[(1280329)>>0]|0;
     $37 = $$pre6;
     break;
    }
   } else {
    $37 = $32;
   }
  } while(0);
  $36 = (($37) + 1)<<24>>24;
  HEAP8[(1280329)>>0] = $36;
 }
 $38 = $$0$ & 512;
 $39 = ($38|0)==(0);
 if (!($39)) {
  $40 = HEAP8[(1280334)>>0]|0;
  $41 = ($40<<24>>24)==(0);
  do {
   if ($41) {
    $42 = (_SDL_JoystickInit()|0);
    $43 = ($42|0)<(0);
    if ($43) {
     $$01 = -1;
     STACKTOP = sp;return ($$01|0);
    } else {
     $$pre8 = HEAP8[(1280334)>>0]|0;
     $45 = $$pre8;
     break;
    }
   } else {
    $45 = $40;
   }
  } while(0);
  $44 = (($45) + 1)<<24>>24;
  HEAP8[(1280334)>>0] = $44;
 }
 $46 = $$0$ & 8192;
 $47 = ($46|0)==(0);
 if (!($47)) {
  $48 = HEAP8[(1280338)>>0]|0;
  $49 = ($48<<24>>24)==(0);
  do {
   if ($49) {
    $50 = (_SDL_GameControllerInit()|0);
    $51 = ($50|0)<(0);
    if ($51) {
     $$01 = -1;
     STACKTOP = sp;return ($$01|0);
    } else {
     $$pre10 = HEAP8[(1280338)>>0]|0;
     $53 = $$pre10;
     break;
    }
   } else {
    $53 = $48;
   }
  } while(0);
  $52 = (($53) + 1)<<24>>24;
  HEAP8[(1280338)>>0] = $52;
 }
 $54 = $$0$ & 4096;
 $55 = ($54|0)==(0);
 if ($55) {
  $$01 = 0;
  STACKTOP = sp;return ($$01|0);
 }
 $56 = (_SDL_SetError(611162,$vararg_buffer)|0);
 $$01 = $56;
 STACKTOP = sp;return ($$01|0);
}
function _SDL_TicksInit() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[300728]|0;
 $1 = ($0|0)==(0);
 if (!($1)) {
  return;
 }
 HEAP32[300728] = 1;
 $2 = (_clock_gettime(4,(1202916|0))|0);
 $3 = ($2|0)==(0);
 if ($3) {
  HEAP32[300731] = 1;
  return;
 } else {
  (_gettimeofday((1202928|0),(0|0))|0);
  return;
 }
}
function _SDL_StartEventLoop() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 (_SDL_EventState(771,0)|0);
 (_SDL_EventState(770,0)|0);
 (_SDL_EventState(513,0)|0);
 HEAP32[(125892)>>2] = 1;
 return 0;
}
function _SDL_QuitInit() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $action$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $action$i = sp;
 $0 = (_SDL_GetHint(609029)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  HEAP32[301415] = 0;
 } else {
  $2 = (_SDL_atoi($0)|0);
  $3 = ($2|0)==(1);
  $4 = $3&1;
  HEAP32[301415] = $4;
  if ($3) {
   STACKTOP = sp;return 0;
  }
 }
 (_sigaction(2,(0|0),($action$i|0))|0);
 $5 = HEAP32[$action$i>>2]|0;
 $6 = ($5|0)==(0|0);
 if ($6) {
  HEAP32[$action$i>>2] = 4644;
  (_sigaction(2,($action$i|0),(0|0))|0);
 }
 (_sigaction(15,(0|0),($action$i|0))|0);
 $7 = HEAP32[$action$i>>2]|0;
 $8 = ($7|0)==(0|0);
 if ($8) {
  HEAP32[$action$i>>2] = 4644;
  (_sigaction(15,($action$i|0),(0|0))|0);
 }
 STACKTOP = sp;return 0;
}
function _SDL_HandleSIG($sig) {
 $sig = $sig|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 (_signal(($sig|0),(4644|0))|0);
 HEAP32[300736] = 1;
 return;
}
function _SDL_TimerInit() {
 var $$1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(1205824)>>2]|0;
 $1 = ($0|0)==(0);
 if (!($1)) {
  $$1 = 0;
  return ($$1|0);
 }
 $2 = (_SDL_CreateMutex()|0);
 HEAP32[(1205676)>>2] = $2;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $$1 = -1;
  return ($$1|0);
 }
 $4 = (_SDL_CreateSemaphore(0)|0);
 HEAP32[(1205812)>>2] = $4;
 $5 = ($4|0)==(0|0);
 if ($5) {
  $6 = HEAP32[(1205676)>>2]|0;
  _SDL_DestroyMutex($6);
  $$1 = -1;
  return ($$1|0);
 }
 HEAP32[(1205824)>>2] = 1;
 $7 = (_SDL_CreateThread(4645,609086,1205664)|0);
 HEAP32[301416] = $7;
 $8 = ($7|0)==(0|0);
 if ($8) {
  _SDL_TimerQuit();
  $$1 = -1;
  return ($$1|0);
 } else {
  (_SDL_AtomicSet((1205668),1)|0);
  $$1 = 0;
  return ($$1|0);
 }
 return (0)|0;
}
function _SDL_CreateMutex() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $mutex$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_malloc(12)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_Error(0)|0);
  $mutex$0 = $0;
  return ($mutex$0|0);
 }
 $2 = (_SDL_CreateSemaphore(1)|0);
 $3 = ((($0)) + 8|0);
 HEAP32[$3>>2] = $2;
 HEAP32[$0>>2] = 0;
 $4 = ((($0)) + 4|0);
 HEAP32[$4>>2] = 0;
 $5 = ($2|0)==(0|0);
 if (!($5)) {
  $mutex$0 = $0;
  return ($mutex$0|0);
 }
 _SDL_free($0);
 $mutex$0 = 0;
 return ($mutex$0|0);
}
function _SDL_CreateSemaphore($initial_value) {
 $initial_value = $initial_value|0;
 var $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 (_SDL_SetError(609052,$vararg_buffer)|0);
 STACKTOP = sp;return (0|0);
}
function _SDL_DestroyMutex($mutex) {
 $mutex = $mutex|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mutex|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($mutex)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
 }
 _SDL_free($mutex);
 return;
}
function _SDL_DestroySemaphore($sem) {
 $sem = $sem|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _SDL_TimerThread($_data) {
 $_data = $_data|0;
 var $$freelist_head$1 = 0, $$lcssa = 0, $$lcssa103 = 0, $$lcssa99 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, $curr$0$i = 0, $curr$0$i5 = 0, $curr$0$lcssa11$i = 0, $curr$0$lcssa11$i8 = 0;
 var $curr$0$lcssa8$i = 0, $curr$0$lcssa8$i12 = 0, $curr$0$lcssa9$i = 0, $curr$0$lcssa9$i10 = 0, $curr$01$i = 0, $curr$01$i1 = 0, $curr$03$i = 0, $curr$03$i$lcssa = 0, $curr$03$i3 = 0, $curr$03$i3$lcssa = 0, $delay$0 = 0, $delay$1 = 0, $freelist_head$0 = 0, $freelist_head$1$ph23 = 0, $freelist_head$1$ph31 = 0, $freelist_head$1$ph31$lcssa = 0, $freelist_tail$0 = 0, $freelist_tail$1$ph25 = 0, $freelist_tail$1$ph32 = 0, $freelist_tail$1$ph32$lcssa = 0;
 var $pending$026 = 0, $prev$0$lcssa12$i = 0, $prev$0$lcssa12$i7 = 0, $prev$02$i = 0, $prev$02$i$lcssa = 0, $prev$02$i$phi = 0, $prev$02$i4 = 0, $prev$02$i4$lcssa = 0, $prev$02$i4$phi = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($_data)) + 160|0);
 $1 = ((($_data)) + 144|0);
 $2 = ((($_data)) + 152|0);
 $3 = ((($_data)) + 164|0);
 $4 = ((($_data)) + 148|0);
 $5 = ((($_data)) + 156|0);
 $freelist_head$0 = 0;$freelist_tail$0 = 0;
 while(1) {
  _SDL_AtomicLock($1);
  $6 = HEAP32[$2>>2]|0;
  HEAP32[$2>>2] = 0;
  $7 = ($freelist_head$0|0)==(0|0);
  if (!($7)) {
   $8 = HEAP32[$5>>2]|0;
   $9 = ((($freelist_tail$0)) + 24|0);
   HEAP32[$9>>2] = $8;
   HEAP32[$5>>2] = $freelist_head$0;
  }
  _SDL_AtomicUnlock($1);
  $10 = ($6|0)==(0|0);
  if (!($10)) {
   $pending$026 = $6;
   while(1) {
    $11 = ((($pending$026)) + 24|0);
    $12 = HEAP32[$11>>2]|0;
    $curr$01$i = HEAP32[$3>>2]|0;
    $13 = ($curr$01$i|0)==(0|0);
    do {
     if ($13) {
      $curr$0$lcssa9$i = 0;
      label = 11;
     } else {
      $14 = ((($pending$026)) + 16|0);
      $15 = HEAP32[$14>>2]|0;
      $curr$03$i = $curr$01$i;$prev$02$i = 0;
      while(1) {
       $16 = ((($curr$03$i)) + 16|0);
       $17 = HEAP32[$16>>2]|0;
       $18 = (($15) - ($17))|0;
       $19 = ($18|0)<(0);
       if ($19) {
        $curr$03$i$lcssa = $curr$03$i;$prev$02$i$lcssa = $prev$02$i;
        label = 9;
        break;
       }
       $20 = ((($curr$03$i)) + 24|0);
       $curr$0$i = HEAP32[$20>>2]|0;
       $21 = ($curr$0$i|0)==(0|0);
       if ($21) {
        $curr$0$lcssa11$i = 0;$prev$0$lcssa12$i = $curr$03$i;
        break;
       } else {
        $prev$02$i$phi = $curr$03$i;$curr$03$i = $curr$0$i;$prev$02$i = $prev$02$i$phi;
       }
      }
      if ((label|0) == 9) {
       label = 0;
       $22 = ($prev$02$i$lcssa|0)==(0|0);
       if ($22) {
        $curr$0$lcssa9$i = $curr$03$i$lcssa;
        label = 11;
        break;
       } else {
        $curr$0$lcssa11$i = $curr$03$i$lcssa;$prev$0$lcssa12$i = $prev$02$i$lcssa;
       }
      }
      $23 = ((($prev$0$lcssa12$i)) + 24|0);
      HEAP32[$23>>2] = $pending$026;
      $curr$0$lcssa8$i = $curr$0$lcssa11$i;
     }
    } while(0);
    if ((label|0) == 11) {
     label = 0;
     HEAP32[$3>>2] = $pending$026;
     $curr$0$lcssa8$i = $curr$0$lcssa9$i;
    }
    HEAP32[$11>>2] = $curr$0$lcssa8$i;
    $24 = ($12|0)==(0|0);
    if ($24) {
     break;
    } else {
     $pending$026 = $12;
    }
   }
  }
  $25 = HEAP32[$0>>2]|0;
  $26 = ($25|0)==(0);
  if ($26) {
   break;
  }
  $27 = (_SDL_GetTicks()|0);
  $28 = HEAP32[$3>>2]|0;
  $29 = ($28|0)==(0|0);
  L22: do {
   if ($29) {
    $delay$0 = -1;$freelist_head$1$ph23 = 0;$freelist_tail$1$ph25 = 0;
   } else {
    $71 = $28;$freelist_head$1$ph31 = 0;$freelist_tail$1$ph32 = 0;
    L23: while(1) {
     $31 = $71;
     while(1) {
      $30 = ((($31)) + 16|0);
      $32 = HEAP32[$30>>2]|0;
      $33 = (($27) - ($32))|0;
      $34 = ($33|0)<(0);
      if ($34) {
       $$lcssa = $32;$freelist_head$1$ph31$lcssa = $freelist_head$1$ph31;$freelist_tail$1$ph32$lcssa = $freelist_tail$1$ph32;
       break L23;
      }
      $36 = ((($31)) + 24|0);
      $37 = HEAP32[$36>>2]|0;
      HEAP32[$3>>2] = $37;
      $38 = ((($31)) + 20|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ($39|0)==(0);
      if (!($40)) {
       $$lcssa103 = $38;$$lcssa99 = $31;
       break;
      }
      $41 = ((($31)) + 4|0);
      $42 = HEAP32[$41>>2]|0;
      $43 = ((($31)) + 12|0);
      $44 = HEAP32[$43>>2]|0;
      $45 = ((($31)) + 8|0);
      $46 = HEAP32[$45>>2]|0;
      $47 = (FUNCTION_TABLE_iii[$42 & 8191]($44,$46)|0);
      $48 = ($47|0)==(0);
      if ($48) {
       $$lcssa103 = $38;$$lcssa99 = $31;
       break;
      }
      $49 = (($47) + ($27))|0;
      HEAP32[$30>>2] = $49;
      $curr$01$i1 = HEAP32[$3>>2]|0;
      $50 = ($curr$01$i1|0)==(0|0);
      do {
       if ($50) {
        $curr$0$lcssa9$i10 = 0;
        label = 25;
       } else {
        $curr$03$i3 = $curr$01$i1;$prev$02$i4 = 0;
        while(1) {
         $51 = ((($curr$03$i3)) + 16|0);
         $52 = HEAP32[$51>>2]|0;
         $53 = (($49) - ($52))|0;
         $54 = ($53|0)<(0);
         if ($54) {
          $curr$03$i3$lcssa = $curr$03$i3;$prev$02$i4$lcssa = $prev$02$i4;
          label = 23;
          break;
         }
         $55 = ((($curr$03$i3)) + 24|0);
         $curr$0$i5 = HEAP32[$55>>2]|0;
         $56 = ($curr$0$i5|0)==(0|0);
         if ($56) {
          $curr$0$lcssa11$i8 = 0;$prev$0$lcssa12$i7 = $curr$03$i3;
          break;
         } else {
          $prev$02$i4$phi = $curr$03$i3;$curr$03$i3 = $curr$0$i5;$prev$02$i4 = $prev$02$i4$phi;
         }
        }
        if ((label|0) == 23) {
         label = 0;
         $57 = ($prev$02$i4$lcssa|0)==(0|0);
         if ($57) {
          $curr$0$lcssa9$i10 = $curr$03$i3$lcssa;
          label = 25;
          break;
         } else {
          $curr$0$lcssa11$i8 = $curr$03$i3$lcssa;$prev$0$lcssa12$i7 = $prev$02$i4$lcssa;
         }
        }
        $58 = ((($prev$0$lcssa12$i7)) + 24|0);
        HEAP32[$58>>2] = $31;
        $60 = $curr$01$i1;$curr$0$lcssa8$i12 = $curr$0$lcssa11$i8;
       }
      } while(0);
      if ((label|0) == 25) {
       label = 0;
       HEAP32[$3>>2] = $31;
       $60 = $31;$curr$0$lcssa8$i12 = $curr$0$lcssa9$i10;
      }
      HEAP32[$36>>2] = $curr$0$lcssa8$i12;
      $59 = ($60|0)==(0|0);
      if ($59) {
       $delay$0 = -1;$freelist_head$1$ph23 = $freelist_head$1$ph31;$freelist_tail$1$ph25 = $freelist_tail$1$ph32;
       break L22;
      } else {
       $31 = $60;
      }
     }
     $61 = ($freelist_head$1$ph31|0)==(0|0);
     $$freelist_head$1 = $61 ? $$lcssa99 : $freelist_head$1$ph31;
     $62 = ($freelist_tail$1$ph32|0)==(0|0);
     if (!($62)) {
      $63 = ((($freelist_tail$1$ph32)) + 24|0);
      HEAP32[$63>>2] = $$lcssa99;
     }
     HEAP32[$$lcssa103>>2] = 1;
     $64 = HEAP32[$3>>2]|0;
     $65 = ($64|0)==(0|0);
     if ($65) {
      $delay$0 = -1;$freelist_head$1$ph23 = $$freelist_head$1;$freelist_tail$1$ph25 = $$lcssa99;
      break L22;
     } else {
      $71 = $64;$freelist_head$1$ph31 = $$freelist_head$1;$freelist_tail$1$ph32 = $$lcssa99;
     }
    }
    $35 = (($$lcssa) - ($27))|0;
    $delay$0 = $35;$freelist_head$1$ph23 = $freelist_head$1$ph31$lcssa;$freelist_tail$1$ph25 = $freelist_tail$1$ph32$lcssa;
   }
  } while(0);
  $66 = (_SDL_GetTicks()|0);
  $67 = (($66) - ($27))|0;
  $68 = ($delay$0>>>0)<($67>>>0);
  $69 = (($delay$0) - ($67))|0;
  $delay$1 = $68 ? 0 : $69;
  $70 = HEAP32[$4>>2]|0;
  (_SDL_SemWaitTimeout($70,$delay$1)|0);
  $freelist_head$0 = $freelist_head$1$ph23;$freelist_tail$0 = $freelist_tail$1$ph25;
 }
 return 0;
}
function _SDL_SemWaitTimeout($sem,$timeout) {
 $sem = $sem|0;
 $timeout = $timeout|0;
 var $0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_SDL_SetError(609052,$vararg_buffer)|0);
 STACKTOP = sp;return ($0|0);
}
function _SDL_CreateThread($fn,$name,$data) {
 $fn = $fn|0;
 $name = $name|0;
 $data = $data|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $thread$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_malloc(800)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_Error(0)|0);
  $$0 = 0;
  return ($$0|0);
 }
 (_SDL_memset($0,0,800)|0);
 $2 = ((($0)) + 8|0);
 HEAP32[$2>>2] = -1;
 $3 = ((($0)) + 12|0);
 (_SDL_AtomicSet($3,0)|0);
 $4 = ($name|0)==(0|0);
 if (!($4)) {
  $5 = (_SDL_strdup($name)|0);
  $6 = ((($0)) + 792|0);
  HEAP32[$6>>2] = $5;
  $7 = ($5|0)==(0|0);
  if ($7) {
   (_SDL_Error(0)|0);
   _SDL_free($0);
   $$0 = 0;
   return ($$0|0);
  }
 }
 $8 = (_SDL_malloc(16)|0);
 $9 = ($8|0)==(0|0);
 if ($9) {
  (_SDL_Error(0)|0);
  $10 = ((($0)) + 792|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($11|0)==(0|0);
  if (!($12)) {
   _SDL_free($11);
  }
  _SDL_free($0);
  $$0 = 0;
  return ($$0|0);
 }
 HEAP32[$8>>2] = $fn;
 $13 = ((($8)) + 4|0);
 HEAP32[$13>>2] = $data;
 $14 = ((($8)) + 8|0);
 HEAP32[$14>>2] = $0;
 $15 = (_SDL_CreateSemaphore(0)|0);
 $16 = ((($8)) + 12|0);
 HEAP32[$16>>2] = $15;
 $17 = ($15|0)==(0|0);
 if ($17) {
  $18 = ((($0)) + 792|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = ($19|0)==(0|0);
  if (!($20)) {
   _SDL_free($19);
  }
  _SDL_free($0);
  _SDL_free($8);
  $$0 = 0;
  return ($$0|0);
 }
 $21 = (_SDL_SYS_CreateThread($0,$8)|0);
 $22 = ($21|0)>(-1);
 if ($22) {
  $23 = HEAP32[$16>>2]|0;
  (_SDL_SemWait($23)|0);
  $thread$0 = $0;
 } else {
  $24 = ((($0)) + 792|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = ($25|0)==(0|0);
  if (!($26)) {
   _SDL_free($25);
  }
  _SDL_free($0);
  $thread$0 = 0;
 }
 _SDL_free($8);
 $$0 = $thread$0;
 return ($$0|0);
}
function _SDL_AtomicSet($a,$v) {
 $a = $a|0;
 $v = $v|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$a>>2]|0;HEAP32[$a>>2] = $v;
 return ($0|0);
}
function _SDL_SYS_CreateThread($thread,$args) {
 $thread = $thread|0;
 $args = $args|0;
 var $0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_SDL_SetError(609095,$vararg_buffer)|0);
 STACKTOP = sp;return ($0|0);
}
function _SDL_SemWait($sem) {
 $sem = $sem|0;
 var $0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_SDL_SetError(609052,$vararg_buffer)|0);
 STACKTOP = sp;return ($0|0);
}
function _SDL_TimerQuit() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(1205824)>>2]|0;
 $1 = ($0|0)==(0);
 if ($1) {
  return;
 }
 HEAP32[(1205824)>>2] = 0;
 $2 = HEAP32[301416]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
  $4 = HEAP32[(1205812)>>2]|0;
  (_SDL_SemPost($4)|0);
  $5 = HEAP32[301416]|0;
  _SDL_WaitThread($5,0);
  HEAP32[301416] = 0;
 }
 HEAP32[(1205812)>>2] = 0;
 $6 = HEAP32[(1205828)>>2]|0;
 $7 = ($6|0)==(0|0);
 if (!($7)) {
  $11 = $6;
  while(1) {
   $10 = ((($11)) + 24|0);
   $12 = HEAP32[$10>>2]|0;
   HEAP32[(1205828)>>2] = $12;
   _SDL_free($11);
   $13 = HEAP32[(1205828)>>2]|0;
   $14 = ($13|0)==(0|0);
   if ($14) {
    break;
   } else {
    $11 = $13;
   }
  }
 }
 $8 = HEAP32[(1205820)>>2]|0;
 $9 = ($8|0)==(0|0);
 if (!($9)) {
  while(1) {
   $17 = HEAP32[(1205820)>>2]|0;
   $18 = ((($17)) + 24|0);
   $19 = HEAP32[$18>>2]|0;
   HEAP32[(1205820)>>2] = $19;
   _SDL_free($17);
   $20 = HEAP32[(1205820)>>2]|0;
   $21 = ($20|0)==(0|0);
   if ($21) {
    break;
   }
  }
 }
 $15 = HEAP32[(1205672)>>2]|0;
 $16 = ($15|0)==(0|0);
 if (!($16)) {
  $23 = $15;
  while(1) {
   $22 = ((($23)) + 8|0);
   $24 = HEAP32[$22>>2]|0;
   HEAP32[(1205672)>>2] = $24;
   _SDL_free($23);
   $25 = HEAP32[(1205672)>>2]|0;
   $26 = ($25|0)==(0|0);
   if ($26) {
    break;
   } else {
    $23 = $25;
   }
  }
 }
 $27 = HEAP32[(1205676)>>2]|0;
 _SDL_DestroyMutex($27);
 HEAP32[(1205676)>>2] = 0;
 return;
}
function _SDL_SemPost($sem) {
 $sem = $sem|0;
 var $0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_SDL_SetError(609052,$vararg_buffer)|0);
 STACKTOP = sp;return ($0|0);
}
function _SDL_WaitThread($thread,$status) {
 $thread = $thread|0;
 $status = $status|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($thread|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ($status|0)==(0|0);
 if (!($1)) {
  $2 = ((($thread)) + 8|0);
  $3 = HEAP32[$2>>2]|0;
  HEAP32[$status>>2] = $3;
 }
 $4 = ((($thread)) + 792|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if (!($6)) {
  _SDL_free($5);
 }
 _SDL_free($thread);
 return;
}
function _SDL_SYS_WaitThread($thread) {
 $thread = $thread|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _SDL_VideoInit($driver_name) {
 $driver_name = $driver_name|0;
 var $$0 = 0, $$013$ph = 0, $$lcssa = 0, $$pr = 0, $$pr$i = 0, $$pr$pre$i = 0, $$pr6$i = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$016$lcssa = 0, $i$29 = 0, $not$$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $video$28 = 0, $window$0$i$i = 0, $window$01$i$i = 0, $window$02$i$i = 0, $window$02$i$i$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  _SDL_VideoQuit();
 }
 _SDL_TicksInit();
 $2 = (_SDL_InitSubSystem(16384)|0);
 $3 = ($2|0)<(0);
 if ($3) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = (_SDL_KeyboardInit()|0);
 $5 = ($4|0)<(0);
 if ($5) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $6 = (_SDL_MouseInit()|0);
 $7 = ($6|0)<(0);
 if ($7) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $8 = (_SDL_TouchInit()|0);
 $9 = ($8|0)<(0);
 if ($9) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $10 = ($driver_name|0)==(0|0);
 do {
  if ($10) {
   $11 = (_SDL_getenv(609252)|0);
   $12 = ($11|0)==(0|0);
   if ($12) {
    $13 = HEAP32[(137932)>>2]|0;
    $14 = (FUNCTION_TABLE_i[$13 & 8191]()|0);
    $15 = ($14|0)==(0);
    if (!($15)) {
     $27 = HEAP32[(137936)>>2]|0;
     $28 = (FUNCTION_TABLE_ii[$27 & 8191](0)|0);
     $29 = ($28|0)==(0|0);
     if (!($29)) {
      $i$29 = 0;$video$28 = $28;
      break;
     }
    }
    $30 = HEAP32[(138860)>>2]|0;
    $31 = (FUNCTION_TABLE_i[$30 & 8191]()|0);
    $32 = ($31|0)==(0);
    if (!($32)) {
     $133 = HEAP32[(138864)>>2]|0;
     $134 = (FUNCTION_TABLE_ii[$133 & 8191](0)|0);
     $135 = ($134|0)==(0|0);
     if (!($135)) {
      $i$29 = 1;$video$28 = $134;
      break;
     }
    }
    $39 = (_SDL_SetError(609285,$vararg_buffer1)|0);
    $$0 = $39;
    STACKTOP = sp;return ($$0|0);
   } else {
    $$013$ph = $11;
    label = 10;
   }
  } else {
   $$013$ph = $driver_name;
   label = 10;
  }
 } while(0);
 do {
  if ((label|0) == 10) {
   $16 = HEAP32[34481]|0;
   $17 = (_SDL_strlen($$013$ph)|0);
   $18 = (_SDL_strncasecmp($16,$$013$ph,$17)|0);
   $19 = ($18|0)==(0);
   if ($19) {
    $20 = HEAP32[(137932)>>2]|0;
    $21 = (FUNCTION_TABLE_i[$20 & 8191]()|0);
    $22 = ($21|0)==(0);
    if ($22) {
     label = 12;
    } else {
     $$lcssa = 140948;$i$016$lcssa = 0;
     label = 15;
    }
   } else {
    label = 12;
   }
   if ((label|0) == 12) {
    $23 = HEAP32[34713]|0;
    $24 = (_SDL_strlen($$013$ph)|0);
    $25 = (_SDL_strncasecmp($23,$$013$ph,$24)|0);
    $26 = ($25|0)==(0);
    if ($26) {
     $130 = HEAP32[(138860)>>2]|0;
     $131 = (FUNCTION_TABLE_i[$130 & 8191]()|0);
     $132 = ($131|0)==(0);
     if (!($132)) {
      $$lcssa = (140952);$i$016$lcssa = 1;
      label = 15;
     }
    }
   }
   if ((label|0) == 15) {
    $33 = HEAP32[$$lcssa>>2]|0;
    $34 = ((($33)) + 12|0);
    $35 = HEAP32[$34>>2]|0;
    $36 = (FUNCTION_TABLE_ii[$35 & 8191](0)|0);
    $37 = ($36|0)==(0|0);
    if (!($37)) {
     $i$29 = $i$016$lcssa;$video$28 = $36;
     break;
    }
   }
   HEAP32[$vararg_buffer>>2] = $$013$ph;
   $38 = (_SDL_SetError(609268,$vararg_buffer)|0);
   $$0 = $38;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 HEAP32[300441] = $video$28;
 $40 = (140948 + ($i$29<<2)|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = HEAP32[$41>>2]|0;
 HEAP32[$video$28>>2] = $42;
 $43 = HEAP32[300441]|0;
 $44 = ((($43)) + 260|0);
 HEAP32[$44>>2] = 1;
 $45 = ((($43)) + 364|0);
 HEAP32[$45>>2] = 0;
 $46 = ((($43)) + 624|0);
 HEAP32[$46>>2] = 0;
 $47 = ($43|0)==(0|0);
 if (!($47)) {
  $48 = ((($43)) + 268|0);
  HEAP32[$48>>2] = 3;
  $49 = ((($43)) + 272|0);
  HEAP32[$49>>2] = 3;
  $50 = ((($43)) + 276|0);
  HEAP32[$50>>2] = 2;
  $51 = ((($43)) + 280|0);
  HEAP32[$51>>2] = 0;
  $52 = ((($43)) + 288|0);
  HEAP32[$52>>2] = 0;
  $53 = ((($43)) + 284|0);
  HEAP32[$53>>2] = 16;
  $54 = ((($43)) + 292|0);
  HEAP32[$54>>2] = 0;
  $55 = ((($43)) + 296|0);
  HEAP32[$55>>2] = 1;
  $56 = ((($43)) + 300|0);
  $57 = ((($43)) + 360|0);
  ;HEAP32[$56>>2]=0|0;HEAP32[$56+4>>2]=0|0;HEAP32[$56+8>>2]=0|0;HEAP32[$56+12>>2]=0|0;HEAP32[$56+16>>2]=0|0;HEAP32[$56+20>>2]=0|0;HEAP32[$56+24>>2]=0|0;
  HEAP32[$57>>2] = 1;
  $58 = ((($43)) + 328|0);
  HEAP32[$58>>2] = -1;
  $59 = ((($43)) + 344|0);
  $60 = ((($43)) + 332|0);
  HEAP32[$60>>2] = 2;
  $61 = ((($43)) + 336|0);
  HEAP32[$61>>2] = 0;
  HEAP32[$59>>2] = 4;
  $62 = ((($43)) + 340|0);
  HEAP32[$62>>2] = 0;
  $63 = ((($43)) + 356|0);
  HEAP32[$63>>2] = 0;
  $64 = ((($43)) + 352|0);
  HEAP32[$64>>2] = 1;
  $65 = ((($43)) + 348|0);
  HEAP32[$65>>2] = 0;
 }
 $66 = (_SDL_TLSCreate()|0);
 $67 = HEAP32[300441]|0;
 $68 = ((($67)) + 636|0);
 HEAP32[$68>>2] = $66;
 $69 = (_SDL_TLSCreate()|0);
 $70 = HEAP32[300441]|0;
 $71 = ((($70)) + 640|0);
 HEAP32[$71>>2] = $69;
 $72 = ((($70)) + 4|0);
 $73 = HEAP32[$72>>2]|0;
 $74 = (FUNCTION_TABLE_ii[$73 & 8191]($70)|0);
 $75 = ($74|0)<(0);
 if ($75) {
  _SDL_VideoQuit();
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $76 = HEAP32[300441]|0;
 $77 = ((($76)) + 240|0);
 $78 = HEAP32[$77>>2]|0;
 $79 = ($78|0)==(0);
 if ($79) {
  _SDL_VideoQuit();
  $80 = (_SDL_SetError(609311,$vararg_buffer3)|0);
  $$0 = $80;
  STACKTOP = sp;return ($$0|0);
 }
 $81 = ((($76)) + 108|0);
 $82 = HEAP32[$81>>2]|0;
 $83 = ($82|0)==(0|0);
 do {
  if ($83) {
   $92 = $76;
   label = 30;
  } else {
   $84 = (_SDL_GetHint(609508)|0);
   $85 = ($84|0)==(0|0);
   if (!($85)) {
    $86 = (_SDL_strcasecmp($84,609353)|0);
    $87 = ($86|0)==(0);
    if ($87) {
     break;
    }
   }
   $88 = (_SDL_GetHint(609362)|0);
   $89 = ($88|0)==(0|0);
   if (!($89)) {
    $90 = HEAP8[$88>>0]|0;
    $not$$i = ($90<<24>>24)==(48);
    if (!($not$$i)) {
     $$pre = HEAP32[300441]|0;
     $92 = $$pre;
     label = 30;
    }
   }
  }
 } while(0);
 if ((label|0) == 30) {
  $91 = ((($92)) + 108|0);
  HEAP32[$91>>2] = 4646;
  $93 = ((($92)) + 112|0);
  HEAP32[$93>>2] = 4647;
  $94 = ((($92)) + 116|0);
  HEAP32[$94>>2] = 4648;
 }
 $95 = (_SDL_GetHint(610181)|0);
 $96 = ($95|0)==(0|0);
 if ($96) {
  label = 33;
 } else {
  $97 = (_SDL_atoi($95)|0);
  $98 = ($97|0)==(0);
  if ($98) {
   label = 33;
  } else {
   label = 37;
  }
 }
 if ((label|0) == 33) {
  $99 = HEAP32[300441]|0;
  $100 = ($99|0)==(0|0);
  if (!($100)) {
   $101 = ((($99)) + 236|0);
   $102 = HEAP32[$101>>2]|0;
   $103 = ($102|0)==(0);
   if ($103) {
    HEAP32[$101>>2] = 1;
    $104 = ((($99)) + 184|0);
    $105 = HEAP32[$104>>2]|0;
    $106 = ($105|0)==(0|0);
    if ($106) {
     label = 37;
    } else {
     FUNCTION_TABLE_vi[$105 & 8191]($99);
     label = 37;
    }
   } else {
    label = 37;
   }
  }
 }
 if ((label|0) == 37) {
  $$pr = HEAP32[300441]|0;
  $107 = ($$pr|0)==(0|0);
  if (!($107)) {
   $108 = ((($$pr)) + 200|0);
   $109 = HEAP32[$108>>2]|0;
   $110 = ($109|0)==(0|0);
   if (!($110)) {
    $111 = (FUNCTION_TABLE_ii[$109 & 8191]($$pr)|0);
    $112 = ($111|0)==(0);
    if (!($112)) {
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
   }
  }
 }
 (_SDL_EventState(771,1)|0);
 (_SDL_EventState(770,1)|0);
 $113 = HEAP32[300441]|0;
 $114 = ($113|0)==(0|0);
 if ($114) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $115 = ((($113)) + 248|0);
 $window$01$i$i = HEAP32[$115>>2]|0;
 $116 = ($window$01$i$i|0)==(0|0);
 do {
  if ($116) {
   $$pr6$i = $113;
  } else {
   $window$02$i$i = $window$01$i$i;
   while(1) {
    $117 = ((($window$02$i$i)) + 48|0);
    $118 = HEAP32[$117>>2]|0;
    $119 = $118 & 512;
    $120 = ($119|0)==(0);
    if (!($120)) {
     $window$02$i$i$lcssa = $window$02$i$i;
     label = 44;
     break;
    }
    $121 = ((($window$02$i$i)) + 144|0);
    $window$0$i$i = HEAP32[$121>>2]|0;
    $122 = ($window$0$i$i|0)==(0|0);
    if ($122) {
     $$pr$i = $113;
     break;
    } else {
     $window$02$i$i = $window$0$i$i;
    }
   }
   if ((label|0) == 44) {
    $123 = ((($113)) + 204|0);
    $124 = HEAP32[$123>>2]|0;
    $125 = ($124|0)==(0|0);
    if ($125) {
     $$pr6$i = $113;
     break;
    }
    FUNCTION_TABLE_vii[$124 & 8191]($113,$window$02$i$i$lcssa);
    $$pr$pre$i = HEAP32[300441]|0;
    $$pr$i = $$pr$pre$i;
   }
   $126 = ($$pr$i|0)==(0|0);
   if ($126) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   } else {
    $$pr6$i = $$pr$i;
   }
  }
 } while(0);
 $127 = ((($$pr6$i)) + 188|0);
 $128 = HEAP32[$127>>2]|0;
 $129 = ($128|0)==(0|0);
 if ($129) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 FUNCTION_TABLE_vi[$128 & 8191]($$pr6$i);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_VideoQuit() {
 var $$in = 0, $$lcssa = 0, $$lcssa2 = 0, $$lcssa3 = 0, $$lcssa46 = 0, $$pre = 0, $$pre20 = 0, $$pre21 = 0, $$pre22 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $9 = 0, $i$09 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  return;
 }
 _SDL_TouchQuit();
 _SDL_MouseQuit();
 _SDL_QuitSubSystem(16384);
 $2 = HEAP32[300441]|0;
 $3 = ((($2)) + 236|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0);
 if ($5) {
  $10 = $2;
 } else {
  HEAP32[$3>>2] = 0;
  $6 = ((($2)) + 184|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)==(0|0);
  if ($8) {
   $10 = $2;
  } else {
   FUNCTION_TABLE_vi[$7 & 8191]($2);
   $$pre = HEAP32[300441]|0;
   $10 = $$pre;
  }
 }
 $9 = ((($10)) + 248|0);
 $11 = HEAP32[$9>>2]|0;
 $12 = ($11|0)==(0|0);
 if ($12) {
  $$lcssa3 = $10;
 } else {
  $13 = $11;
  while(1) {
   _SDL_DestroyWindow($13);
   $14 = HEAP32[300441]|0;
   $15 = ((($14)) + 248|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = ($16|0)==(0|0);
   if ($17) {
    $$lcssa3 = $14;
    break;
   } else {
    $13 = $16;
   }
  }
 }
 $18 = ((($$lcssa3)) + 8|0);
 $19 = HEAP32[$18>>2]|0;
 FUNCTION_TABLE_vi[$19 & 8191]($$lcssa3);
 $20 = HEAP32[300441]|0;
 $21 = ((($20)) + 240|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ($22|0)>(0);
 $24 = ((($20)) + 244|0);
 $25 = HEAP32[$24>>2]|0;
 if ($23) {
  $27 = $25;$i$09 = 0;
  while(1) {
   $26 = (((($27) + (($i$09*68)|0)|0)) + 8|0);
   $28 = HEAP32[$26>>2]|0;
   $29 = ($28|0)==(0);
   $30 = (((($27) + (($i$09*68)|0)|0)) + 12|0);
   if ($29) {
    $$pre21 = HEAP32[$30>>2]|0;
    $38 = $$pre21;
   } else {
    $$pre20 = HEAP32[$30>>2]|0;
    $$in = $28;$33 = $$pre20;
    while(1) {
     $31 = (($$in) + -1)|0;
     $32 = (((($33) + (($31*20)|0)|0)) + 16|0);
     $34 = HEAP32[$32>>2]|0;
     _SDL_free($34);
     $35 = HEAP32[$30>>2]|0;
     $36 = (((($35) + (($31*20)|0)|0)) + 16|0);
     HEAP32[$36>>2] = 0;
     $37 = ($31|0)==(0);
     if ($37) {
      $$lcssa46 = $35;
      break;
     } else {
      $$in = $31;$33 = $35;
     }
    }
    $38 = $$lcssa46;
   }
   _SDL_free($38);
   HEAP32[$30>>2] = 0;
   $39 = (((($27) + (($i$09*68)|0)|0)) + 32|0);
   $40 = HEAP32[$39>>2]|0;
   _SDL_free($40);
   HEAP32[$39>>2] = 0;
   $41 = (((($27) + (($i$09*68)|0)|0)) + 64|0);
   $42 = HEAP32[$41>>2]|0;
   _SDL_free($42);
   HEAP32[$41>>2] = 0;
   $43 = (($i$09) + 1)|0;
   $44 = HEAP32[300441]|0;
   $45 = ((($44)) + 240|0);
   $46 = HEAP32[$45>>2]|0;
   $47 = ($43|0)<($46|0);
   $48 = ((($44)) + 244|0);
   $49 = HEAP32[$48>>2]|0;
   if ($47) {
    $27 = $49;$i$09 = $43;
   } else {
    $$lcssa2 = $49;$52 = $46;$54 = $44;
    break;
   }
  }
 } else {
  $$lcssa2 = $25;$52 = $22;$54 = $20;
 }
 $50 = ($$lcssa2|0)==(0|0);
 if ($50) {
  $76 = $54;
 } else {
  $51 = ($52|0)>(0);
  $53 = ((($54)) + 244|0);
  if ($51) {
   $55 = HEAP32[$$lcssa2>>2]|0;
   _SDL_free($55);
   $56 = HEAP32[300441]|0;
   $57 = ((($56)) + 240|0);
   $58 = HEAP32[$57>>2]|0;
   $59 = ($58|0)>(1);
   $60 = ((($56)) + 244|0);
   if ($59) {
    $61 = $60;$63 = 1;
    while(1) {
     $$pre22 = HEAP32[$61>>2]|0;
     $62 = (($$pre22) + (($63*68)|0)|0);
     $64 = HEAP32[$62>>2]|0;
     _SDL_free($64);
     $65 = (($63) + 1)|0;
     $66 = HEAP32[300441]|0;
     $67 = ((($66)) + 240|0);
     $68 = HEAP32[$67>>2]|0;
     $69 = ($65|0)<($68|0);
     $70 = ((($66)) + 244|0);
     if ($69) {
      $61 = $70;$63 = $65;
     } else {
      $$lcssa = $70;
      break;
     }
    }
   } else {
    $$lcssa = $60;
   }
  } else {
   $$lcssa = $53;
  }
  $71 = HEAP32[$$lcssa>>2]|0;
  _SDL_free($71);
  $72 = HEAP32[300441]|0;
  $73 = ((($72)) + 244|0);
  HEAP32[$73>>2] = 0;
  $74 = ((($72)) + 240|0);
  HEAP32[$74>>2] = 0;
  $76 = $72;
 }
 $75 = ((($76)) + 264|0);
 $77 = HEAP32[$75>>2]|0;
 _SDL_free($77);
 $78 = HEAP32[300441]|0;
 $79 = ((($78)) + 264|0);
 HEAP32[$79>>2] = 0;
 $80 = ((($78)) + 660|0);
 $81 = HEAP32[$80>>2]|0;
 FUNCTION_TABLE_vi[$81 & 8191]($78);
 HEAP32[300441] = 0;
 return;
}
function _SDL_TouchQuit() {
 var $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[301382]|0;
 $1 = ($0|0)==(0);
 if (!($1)) {
  $$in = $0;
  while(1) {
   $2 = (($$in) + -1)|0;
   $3 = HEAP32[301383]|0;
   $4 = (($3) + ($2<<2)|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = $5;
   $7 = $6;
   $8 = HEAP32[$7>>2]|0;
   $9 = (($6) + 4)|0;
   $10 = $9;
   $11 = HEAP32[$10>>2]|0;
   _SDL_DelTouch($8,$11);
   $12 = ($2|0)==(0);
   if ($12) {
    break;
   } else {
    $$in = $2;
   }
  }
 }
 $13 = HEAP32[301383]|0;
 _SDL_free($13);
 HEAP32[301383] = 0;
 return;
}
function _SDL_DelTouch($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$ph = 0, $$lcssa = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$07 = 0, $index$01$i = 0, $index$01$i$i = 0, $index$01$i$i$lcssa = 0, $or$cond$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $2 = HEAP32[301382]|0;
 $3 = ($2|0)>(0);
 L1: do {
  if ($3) {
   $4 = HEAP32[301383]|0;
   $index$01$i = 0;
   while(1) {
    $5 = (($4) + ($index$01$i<<2)|0);
    $6 = HEAP32[$5>>2]|0;
    $7 = $6;
    $8 = $7;
    $9 = HEAP32[$8>>2]|0;
    $10 = (($7) + 4)|0;
    $11 = $10;
    $12 = HEAP32[$11>>2]|0;
    $13 = ($9|0)==($0|0);
    $14 = ($12|0)==($1|0);
    $15 = $13 & $14;
    if ($15) {
     $$0$i$ph = $index$01$i;
     break;
    }
    $16 = (($index$01$i) + 1)|0;
    $17 = ($16|0)<($2|0);
    if ($17) {
     $index$01$i = $16;
    } else {
     $$0$i$ph = -1;
     break;
    }
   }
   $index$01$i$i = 0;
   while(1) {
    $18 = (($4) + ($index$01$i$i<<2)|0);
    $19 = HEAP32[$18>>2]|0;
    $20 = $19;
    $21 = $20;
    $22 = HEAP32[$21>>2]|0;
    $23 = (($20) + 4)|0;
    $24 = $23;
    $25 = HEAP32[$24>>2]|0;
    $26 = ($22|0)==($0|0);
    $27 = ($25|0)==($1|0);
    $28 = $26 & $27;
    if ($28) {
     $$lcssa = $19;$index$01$i$i$lcssa = $index$01$i$i;
     break;
    }
    $29 = (($index$01$i$i) + 1)|0;
    $30 = ($29|0)<($2|0);
    if ($30) {
     $index$01$i$i = $29;
    } else {
     break L1;
    }
   }
   $31 = ($index$01$i$i$lcssa|0)>(-1);
   $32 = ($index$01$i$i$lcssa|0)<($2|0);
   $or$cond$i = $31 & $32;
   if ($or$cond$i) {
    $33 = ($$lcssa|0)==(0|0);
    if ($33) {
     STACKTOP = sp;return;
    }
    $34 = ((($$lcssa)) + 12|0);
    $35 = HEAP32[$34>>2]|0;
    $36 = ($35|0)>(0);
    $37 = ((($$lcssa)) + 16|0);
    if ($36) {
     $i$07 = 0;
     while(1) {
      $38 = HEAP32[$37>>2]|0;
      $39 = (($38) + ($i$07<<2)|0);
      $40 = HEAP32[$39>>2]|0;
      _SDL_free($40);
      $41 = (($i$07) + 1)|0;
      $42 = HEAP32[$34>>2]|0;
      $43 = ($41|0)<($42|0);
      if ($43) {
       $i$07 = $41;
      } else {
       break;
      }
     }
    }
    $44 = HEAP32[$37>>2]|0;
    _SDL_free($44);
    _SDL_free($$lcssa);
    $45 = HEAP32[301382]|0;
    $46 = (($45) + -1)|0;
    HEAP32[301382] = $46;
    $47 = HEAP32[301383]|0;
    $48 = (($47) + ($46<<2)|0);
    $49 = HEAP32[$48>>2]|0;
    $50 = (($47) + ($$0$i$ph<<2)|0);
    HEAP32[$50>>2] = $49;
    STACKTOP = sp;return;
   }
  }
 } while(0);
 (_SDL_SetError(605121,$vararg_buffer)|0);
 STACKTOP = sp;return;
}
function _SDL_MouseQuit() {
 var $$$i = 0, $$01$i$i = 0, $$01$i$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $curr$0$i = 0, $curr$03$i = 0, $curr$05$i = 0, $cursor$05 = 0, $notlhs$i$i = 0;
 var $notlhs$i$i2 = 0, $notrhs$i$i = 0, $or$cond3$not$i$i = 0, $or$cond5$i$i = 0, $or$cond5$i$i3 = 0, $prev$04$i = 0, $prev$04$i$lcssa = 0, $prev$04$i$phi = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(1201808)>>2]|0;
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  $2 = (_SDL_GetKeyboardFocus()|0);
  $3 = ($2|0)==(0|0);
  if (!($3)) {
   $4 = ((($2)) + 48|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = $5 & 16384;
   $7 = ($6|0)==(0);
   if (!($7)) {
    $8 = HEAP32[(1201808)>>2]|0;
    $9 = (FUNCTION_TABLE_ii[$8 & 8191](0)|0);
    $10 = ($9|0)==(-1);
    if (!($10)) {
     $11 = HEAP32[$4>>2]|0;
     $12 = $11 & -16385;
     HEAP32[$4>>2] = $12;
    }
   }
  }
 }
 (_SDL_SetRelativeMouseMode(0)|0);
 $13 = HEAP32[(1201880)>>2]|0;
 HEAP32[(1201880)>>2] = 1;
 $14 = ($13|0)==(1);
 do {
  if (!($14)) {
   $15 = HEAP32[(1201820)>>2]|0;
   $16 = ($15|0)==(0|0);
   $17 = HEAP32[(1201876)>>2]|0;
   $18 = HEAP32[(1201872)>>2]|0;
   $$01$i$i = $16 ? $18 : $17;
   $notlhs$i$i = ($$01$i$i|0)!=(0|0);
   $19 = HEAP32[(1201852)>>2]|0;
   $20 = ($19|0)==(0);
   $or$cond5$i$i = $notlhs$i$i & $20;
   $21 = HEAP32[(1201784)>>2]|0;
   $22 = ($21|0)==(0|0);
   if ($or$cond5$i$i) {
    if ($22) {
     break;
    }
    (FUNCTION_TABLE_ii[$21 & 8191]($$01$i$i)|0);
    break;
   } else {
    if ($22) {
     break;
    }
    (FUNCTION_TABLE_ii[$21 & 8191](0)|0);
    break;
   }
  }
 } while(0);
 $23 = HEAP32[(1201868)>>2]|0;
 $24 = ($23|0)==(0|0);
 if (!($24)) {
  $cursor$05 = $23;
  while(1) {
   $25 = HEAP32[$cursor$05>>2]|0;
   $26 = HEAP32[(1201872)>>2]|0;
   $27 = ($26|0)==($cursor$05|0);
   L18: do {
    if (!($27)) {
     $29 = HEAP32[(1201876)>>2]|0;
     $30 = ($29|0)==($cursor$05|0);
     do {
      if ($30) {
       $31 = ($26|0)==(0|0);
       if ($31) {
        $32 = HEAP32[(1201820)>>2]|0;
        $33 = ($32|0)==(0|0);
        $$$i = $33 ? 0 : $cursor$05;
        $$01$i$i1 = $$$i;
       } else {
        HEAP32[(1201876)>>2] = $26;
        $$01$i$i1 = $26;
       }
       $34 = HEAP32[(1201880)>>2]|0;
       $notlhs$i$i2 = ($$01$i$i1|0)!=(0|0);
       $notrhs$i$i = ($34|0)!=(0);
       $or$cond3$not$i$i = $notlhs$i$i2 & $notrhs$i$i;
       $35 = HEAP32[(1201852)>>2]|0;
       $36 = ($35|0)==(0);
       $or$cond5$i$i3 = $or$cond3$not$i$i & $36;
       $37 = HEAP32[(1201784)>>2]|0;
       $38 = ($37|0)==(0|0);
       if ($or$cond5$i$i3) {
        if ($38) {
         break;
        }
        (FUNCTION_TABLE_ii[$37 & 8191]($$01$i$i1)|0);
        break;
       } else {
        if ($38) {
         break;
        }
        (FUNCTION_TABLE_ii[$37 & 8191](0)|0);
        break;
       }
      }
     } while(0);
     $curr$03$i = HEAP32[(1201868)>>2]|0;
     $39 = ($curr$03$i|0)==(0|0);
     if (!($39)) {
      $curr$05$i = $curr$03$i;$prev$04$i = 0;
      while(1) {
       $40 = ($curr$05$i|0)==($cursor$05|0);
       if ($40) {
        $prev$04$i$lcssa = $prev$04$i;
        break;
       }
       $curr$0$i = HEAP32[$curr$05$i>>2]|0;
       $45 = ($curr$0$i|0)==(0|0);
       if ($45) {
        break L18;
       } else {
        $prev$04$i$phi = $curr$05$i;$curr$05$i = $curr$0$i;$prev$04$i = $prev$04$i$phi;
       }
      }
      $41 = ($prev$04$i$lcssa|0)==(0|0);
      $42 = HEAP32[$cursor$05>>2]|0;
      if ($41) {
       HEAP32[(1201868)>>2] = $42;
      } else {
       HEAP32[$prev$04$i$lcssa>>2] = $42;
      }
      $43 = HEAP32[(1201792)>>2]|0;
      $44 = ($43|0)==(0|0);
      if (!($44)) {
       FUNCTION_TABLE_vi[$43 & 8191]($cursor$05);
      }
     }
    }
   } while(0);
   $28 = ($25|0)==(0|0);
   if ($28) {
    break;
   } else {
    $cursor$05 = $25;
   }
  }
 }
 $46 = HEAP32[(1201872)>>2]|0;
 $47 = ($46|0)==(0|0);
 if (!($47)) {
  $48 = HEAP32[(1201792)>>2]|0;
  $49 = ($48|0)==(0|0);
  if (!($49)) {
   FUNCTION_TABLE_vi[$48 & 8191]($46);
  }
 }
 $50 = HEAP32[(1201864)>>2]|0;
 $51 = ($50|0)==(0|0);
 if ($51) {
  (_SDL_memset(1201776,0,112)|0);
  return;
 }
 _SDL_free($50);
 (_SDL_memset(1201776,0,112)|0);
 return;
}
function _SDL_SetRelativeMouseMode($enabled) {
 $enabled = $enabled|0;
 var $$01$i = 0, $$01$i$i = 0, $$pr$i = 0, $$val = 0, $$val11 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $not$$i = 0, $not$$i6 = 0;
 var $not$$i8 = 0, $notlhs$i = 0, $notlhs$i$i = 0, $notrhs$i = 0, $notrhs$i$i = 0, $or$cond = 0, $or$cond3 = 0, $or$cond3$not$i = 0, $or$cond3$not$i$i = 0, $or$cond5 = 0, $or$cond5$i = 0, $or$cond5$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GetKeyboardFocus()|0);
 $1 = HEAP32[(1201852)>>2]|0;
 $2 = ($1|0)==($enabled|0);
 if ($2) {
  return 0;
 }
 $3 = ($enabled|0)!=(0);
 $4 = ($0|0)!=(0|0);
 $or$cond = $3 & $4;
 do {
  if ($or$cond) {
   $5 = HEAP32[(1201820)>>2]|0;
   $6 = ($5|0)==($0|0);
   do {
    if (!($6)) {
     $7 = ($5|0)==(0|0);
     if (!($7)) {
      (_SDL_SendWindowEvent($5,11,0,0)|0);
     }
     HEAP32[(1201820)>>2] = $0;
     (_SDL_SendWindowEvent($0,10,0,0)|0);
     $$pr$i = HEAP32[(1201820)>>2]|0;
     $8 = ($$pr$i|0)==(0|0);
     $$val = HEAP32[(1201872)>>2]|0;
     $$val11 = HEAP32[(1201876)>>2]|0;
     $$01$i$i = $8 ? $$val : $$val11;
     $9 = HEAP32[(1201880)>>2]|0;
     $notlhs$i$i = ($$01$i$i|0)!=(0|0);
     $notrhs$i$i = ($9|0)!=(0);
     $or$cond3$not$i$i = $notlhs$i$i & $notrhs$i$i;
     $10 = HEAP32[(1201852)>>2]|0;
     $11 = ($10|0)==(0);
     $or$cond5$i$i = $or$cond3$not$i$i & $11;
     $12 = HEAP32[(1201784)>>2]|0;
     $13 = ($12|0)==(0|0);
     if ($or$cond5$i$i) {
      if ($13) {
       break;
      }
      (FUNCTION_TABLE_ii[$12 & 8191]($$01$i$i)|0);
      break;
     } else {
      if ($13) {
       break;
      }
      (FUNCTION_TABLE_ii[$12 & 8191](0)|0);
      break;
     }
    }
   } while(0);
   $14 = ((($0)) + 24|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = (($15|0) / 2)&-1;
   $17 = ((($0)) + 28|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = (($18|0) / 2)&-1;
   $20 = HEAP32[(1201796)>>2]|0;
   $21 = ($20|0)==(0|0);
   if (!($21)) {
    FUNCTION_TABLE_viii[$20 & 8191]($0,$16,$19);
    break;
   }
   $22 = HEAP32[(1201816)>>2]|0;
   $23 = (_SDL_UpdateMouseFocus($0,$16,$19)|0);
   $not$$i = ($23|0)==(0);
   if (!($not$$i)) {
    (_SDL_PrivateSendMouseMotion($0,$22,0,$16,$19)|0);
   }
  }
 } while(0);
 $24 = HEAP32[(1201856)>>2]|0;
 $25 = ($24|0)==(0);
 $or$cond5 = $3 | $25;
 L22: do {
  if ($or$cond5) {
   do {
    if ($3) {
     $26 = HEAP32[(1201804)>>2]|0;
     $27 = ($26|0)==(0|0);
     if (!($27)) {
      $28 = (_SDL_GetHint(609138)|0);
      $29 = ($28|0)==(0|0);
      if ($29) {
       break;
      }
      $30 = HEAP8[$28>>0]|0;
      $not$$i6 = ($30<<24>>24)==(48);
      if ($not$$i6) {
       break;
      }
     }
     HEAP32[(1201856)>>2] = 1;
     break L22;
    }
   } while(0);
   $31 = HEAP32[(1201804)>>2]|0;
   $32 = (FUNCTION_TABLE_ii[$31 & 8191]($enabled)|0);
   $33 = ($32|0)<(0);
   $or$cond3 = $3 & $33;
   if ($or$cond3) {
    HEAP32[(1201856)>>2] = 1;
   }
  } else {
   HEAP32[(1201856)>>2] = 0;
  }
 } while(0);
 HEAP32[(1201852)>>2] = $enabled;
 $34 = HEAP32[(1201820)>>2]|0;
 $35 = ($34|0)==(0|0);
 do {
  if (!($35)) {
   _SDL_UpdateWindowGrab($34);
   if (!($3)) {
    $36 = HEAP32[(1201820)>>2]|0;
    $37 = HEAP32[(1201824)>>2]|0;
    $38 = HEAP32[(1201828)>>2]|0;
    $39 = ($36|0)==(0|0);
    if (!($39)) {
     $40 = HEAP32[(1201796)>>2]|0;
     $41 = ($40|0)==(0|0);
     if (!($41)) {
      FUNCTION_TABLE_viii[$40 & 8191]($36,$37,$38);
      break;
     }
     $42 = HEAP32[(1201816)>>2]|0;
     $43 = (_SDL_UpdateMouseFocus($36,$37,$38)|0);
     $not$$i8 = ($43|0)==(0);
     if (!($not$$i8)) {
      (_SDL_PrivateSendMouseMotion($36,$42,0,$37,$38)|0);
     }
    }
   }
  }
 } while(0);
 _SDL_FlushEvent(1024);
 $44 = HEAP32[(1201820)>>2]|0;
 $45 = ($44|0)==(0|0);
 $46 = HEAP32[(1201876)>>2]|0;
 $47 = HEAP32[(1201872)>>2]|0;
 $$01$i = $45 ? $47 : $46;
 $48 = HEAP32[(1201880)>>2]|0;
 $notlhs$i = ($$01$i|0)!=(0|0);
 $notrhs$i = ($48|0)!=(0);
 $or$cond3$not$i = $notlhs$i & $notrhs$i;
 $49 = HEAP32[(1201852)>>2]|0;
 $50 = ($49|0)==(0);
 $or$cond5$i = $or$cond3$not$i & $50;
 $51 = HEAP32[(1201784)>>2]|0;
 $52 = ($51|0)==(0|0);
 if ($or$cond5$i) {
  if ($52) {
   return 0;
  }
  (FUNCTION_TABLE_ii[$51 & 8191]($$01$i)|0);
  return 0;
 } else {
  if ($52) {
   return 0;
  }
  (FUNCTION_TABLE_ii[$51 & 8191](0)|0);
  return 0;
 }
 return (0)|0;
}
function _SDL_UpdateWindowGrab($window) {
 $window = $window|0;
 var $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $grabbed$01 = 0, $or$cond = 0, $or$cond7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GetMouse()|0);
 $1 = ((($0)) + 76|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($window)) + 48|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 & 256;
 $6 = $2 | $5;
 $7 = ($6|0)==(0);
 $8 = $4 & 512;
 $9 = ($8|0)==(0);
 $or$cond7 = $7 | $9;
 $10 = HEAP32[300441]|0;
 $11 = ((($10)) + 252|0);
 $12 = HEAP32[$11>>2]|0;
 if ($or$cond7) {
  $23 = ($12|0)==($window|0);
  if ($23) {
   HEAP32[$11>>2] = 0;
   $25 = $10;$grabbed$01 = 0;
  } else {
   $25 = $10;$grabbed$01 = 0;
  }
 } else {
  $13 = ($12|0)==(0|0);
  $14 = ($12|0)==($window|0);
  $or$cond = $13 | $14;
  if ($or$cond) {
   $22 = $10;
  } else {
   $15 = ((($12)) + 48|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = $16 & -257;
   HEAP32[$15>>2] = $17;
   $18 = ((($10)) + 100|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ($19|0)==(0|0);
   if ($20) {
    $22 = $10;
   } else {
    FUNCTION_TABLE_viii[$19 & 8191]($10,$12,0);
    $$pre3 = HEAP32[300441]|0;
    $22 = $$pre3;
   }
  }
  $21 = ((($22)) + 252|0);
  HEAP32[$21>>2] = $window;
  $25 = $22;$grabbed$01 = 1;
 }
 $24 = ((($25)) + 100|0);
 $26 = HEAP32[$24>>2]|0;
 $27 = ($26|0)==(0|0);
 if ($27) {
  return;
 }
 FUNCTION_TABLE_viii[$26 & 8191]($25,$window,$grabbed$01);
 return;
}
function _SDL_FlushEvent($type) {
 $type = $type|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $entry$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(125892)>>2]|0;
 $1 = ($0|0)==(0);
 if ($1) {
  return;
 }
 $2 = HEAP32[31472]|0;
 $3 = (_SDL_LockMutex($2)|0);
 $4 = ($3|0)==(0);
 if (!($4)) {
  return;
 }
 $5 = HEAP32[(125904)>>2]|0;
 $6 = ($5|0)==(0|0);
 if (!($6)) {
  $entry$01$i = $5;
  while(1) {
   $7 = ((($entry$01$i)) + 72|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = HEAP32[$entry$01$i>>2]|0;
   $10 = ($9|0)==($type|0);
   $11 = $8;
   if ($10) {
    $13 = ((($entry$01$i)) + 68|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = ($14|0)==(0|0);
    $16 = $14;
    if ($15) {
     $19 = $8;
    } else {
     $17 = ((($14)) + 72|0);
     HEAP32[$17>>2] = $11;
     $$pre$i = HEAP32[$7>>2]|0;
     $19 = $$pre$i;
    }
    $18 = ($19|0)==(0|0);
    $20 = $19;
    if (!($18)) {
     $21 = ((($19)) + 68|0);
     HEAP32[$21>>2] = $16;
    }
    $22 = HEAP32[(125904)>>2]|0;
    $23 = ($22|0)==($entry$01$i|0);
    if ($23) {
     HEAP32[(125904)>>2] = $20;
    }
    $24 = HEAP32[(125908)>>2]|0;
    $25 = ($24|0)==($entry$01$i|0);
    if ($25) {
     $26 = HEAP32[$13>>2]|0;
     HEAP32[(125908)>>2] = $26;
    }
    $27 = HEAP32[(125912)>>2]|0;
    HEAP32[$7>>2] = $27;
    HEAP32[(125912)>>2] = $entry$01$i;
    $28 = HEAP32[(125896)>>2]|0;
    $29 = (($28) + -1)|0;
    HEAP32[(125896)>>2] = $29;
   }
   $12 = ($8|0)==(0|0);
   if ($12) {
    break;
   } else {
    $entry$01$i = $8;
   }
  }
 }
 return;
}
function _SDL_KeyboardQuit() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _SDL_QuitSubSystem($flags) {
 $flags = $flags|0;
 var $$0 = 0, $$1 = 0, $$2 = 0, $$pr$pre = 0, $$pr16$pre = 0, $$pr1628 = 0, $$pr17$pre = 0, $$pr1729 = 0, $$pr18$pre = 0, $$pr1830 = 0, $$pr19$pre = 0, $$pr1931 = 0, $$pr20$pre = 0, $$pr2032 = 0, $$pr27 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $flags & 8192;
 $1 = ($0|0)==(0);
 do {
  if ($1) {
   $$0 = $flags;
  } else {
   $2 = $flags | 512;
   $3 = HEAP8[(1280338)>>0]|0;
   $4 = ($3<<24>>24)==(0);
   if ($4) {
    $$0 = $2;
   } else {
    $5 = ($3<<24>>24)==(1);
    $6 = HEAP32[301458]|0;
    $7 = $5&1;
    $8 = $6 | $7;
    $9 = ($8|0)==(0);
    if ($9) {
     $$pr27 = $3;
    } else {
     _SDL_GameControllerQuit();
     $$pr$pre = HEAP8[(1280338)>>0]|0;
     $10 = ($$pr$pre<<24>>24)==(0);
     if ($10) {
      $$0 = $2;
      break;
     } else {
      $$pr27 = $$pr$pre;
     }
    }
    $11 = (($$pr27) + -1)<<24>>24;
    HEAP8[(1280338)>>0] = $11;
    $$0 = $2;
   }
  }
 } while(0);
 $12 = $$0 & 512;
 $13 = ($12|0)==(0);
 do {
  if ($13) {
   $$1 = $$0;
  } else {
   $14 = $$0 | 16384;
   $15 = HEAP8[(1280334)>>0]|0;
   $16 = ($15<<24>>24)==(0);
   if ($16) {
    $$1 = $14;
   } else {
    $17 = ($15<<24>>24)==(1);
    $18 = HEAP32[301458]|0;
    $19 = $17&1;
    $20 = $18 | $19;
    $21 = ($20|0)==(0);
    if ($21) {
     $$pr1628 = $15;
    } else {
     _SDL_JoystickQuit();
     $$pr16$pre = HEAP8[(1280334)>>0]|0;
     $22 = ($$pr16$pre<<24>>24)==(0);
     if ($22) {
      $$1 = $14;
      break;
     } else {
      $$pr1628 = $$pr16$pre;
     }
    }
    $23 = (($$pr1628) + -1)<<24>>24;
    HEAP8[(1280334)>>0] = $23;
    $$1 = $14;
   }
  }
 } while(0);
 $24 = $$1 & 16;
 $25 = ($24|0)==(0);
 do {
  if (!($25)) {
   $26 = HEAP8[(1280329)>>0]|0;
   $27 = ($26<<24>>24)==(0);
   if (!($27)) {
    $28 = ($26<<24>>24)==(1);
    $29 = HEAP32[301458]|0;
    $30 = $28&1;
    $31 = $29 | $30;
    $32 = ($31|0)==(0);
    if ($32) {
     $$pr1729 = $26;
    } else {
     _SDL_AudioQuit();
     $$pr17$pre = HEAP8[(1280329)>>0]|0;
     $33 = ($$pr17$pre<<24>>24)==(0);
     if ($33) {
      break;
     } else {
      $$pr1729 = $$pr17$pre;
     }
    }
    $34 = (($$pr1729) + -1)<<24>>24;
    HEAP8[(1280329)>>0] = $34;
   }
  }
 } while(0);
 $35 = $$1 & 32;
 $36 = ($35|0)==(0);
 do {
  if ($36) {
   $$2 = $$1;
  } else {
   $37 = $$1 | 16384;
   $38 = HEAP8[(1280330)>>0]|0;
   $39 = ($38<<24>>24)==(0);
   if ($39) {
    $$2 = $37;
   } else {
    $40 = ($38<<24>>24)==(1);
    $41 = HEAP32[301458]|0;
    $42 = $40&1;
    $43 = $41 | $42;
    $44 = ($43|0)==(0);
    if ($44) {
     $$pr1830 = $38;
    } else {
     _SDL_VideoQuit();
     $$pr18$pre = HEAP8[(1280330)>>0]|0;
     $45 = ($$pr18$pre<<24>>24)==(0);
     if ($45) {
      $$2 = $37;
      break;
     } else {
      $$pr1830 = $$pr18$pre;
     }
    }
    $46 = (($$pr1830) + -1)<<24>>24;
    HEAP8[(1280330)>>0] = $46;
    $$2 = $37;
   }
  }
 } while(0);
 $47 = $$2 & 1;
 $48 = ($47|0)==(0);
 do {
  if (!($48)) {
   $49 = HEAP8[1280325]|0;
   $50 = ($49<<24>>24)==(0);
   if (!($50)) {
    $51 = ($49<<24>>24)==(1);
    $52 = HEAP32[301458]|0;
    $53 = $51&1;
    $54 = $52 | $53;
    $55 = ($54|0)==(0);
    if ($55) {
     $$pr1931 = $49;
    } else {
     _SDL_TimerQuit();
     $$pr19$pre = HEAP8[1280325]|0;
     $56 = ($$pr19$pre<<24>>24)==(0);
     if ($56) {
      break;
     } else {
      $$pr1931 = $$pr19$pre;
     }
    }
    $57 = (($$pr1931) + -1)<<24>>24;
    HEAP8[1280325] = $57;
   }
  }
 } while(0);
 $58 = $$2 & 16384;
 $59 = ($58|0)==(0);
 if ($59) {
  return;
 }
 $60 = HEAP8[(1280339)>>0]|0;
 $61 = ($60<<24>>24)==(0);
 if ($61) {
  return;
 }
 $62 = ($60<<24>>24)==(1);
 $63 = HEAP32[301458]|0;
 $64 = $62&1;
 $65 = $63 | $64;
 $66 = ($65|0)==(0);
 if ($66) {
  $$pr2032 = $60;
 } else {
  _SDL_QuitQuit();
  _SDL_StopEventLoop();
  $$pr20$pre = HEAP8[(1280339)>>0]|0;
  $67 = ($$pr20$pre<<24>>24)==(0);
  if ($67) {
   return;
  } else {
   $$pr2032 = $$pr20$pre;
  }
 }
 $68 = (($$pr2032) + -1)<<24>>24;
 HEAP8[(1280339)>>0] = $68;
 return;
}
function _SDL_GameControllerQuit() {
 var $$pr = 0, $$pr7 = 0, $$pr8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $gamecontrollerlist$0$i = 0, $gamecontrollerlist$03$i = 0, $gamecontrollerlist$05$i = 0, $gamecontrollerlistprev$04$i = 0, $gamecontrollerlistprev$04$i$lcssa = 0, $gamecontrollerlistprev$04$i$phi = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$pr7 = HEAP32[301459]|0;
 $0 = ($$pr7|0)==(0|0);
 if (!($0)) {
  $$pr8 = $$pr7;
  while(1) {
   $3 = ((($$pr8)) + 4|0);
   HEAP32[$3>>2] = 0;
   $4 = HEAP32[$$pr8>>2]|0;
   _SDL_JoystickClose($4);
   $gamecontrollerlist$03$i = HEAP32[301459]|0;
   $5 = ($gamecontrollerlist$03$i|0)==(0|0);
   L4: do {
    if (!($5)) {
     $gamecontrollerlist$05$i = $gamecontrollerlist$03$i;$gamecontrollerlistprev$04$i = 0;
     while(1) {
      $6 = ($gamecontrollerlist$05$i|0)==($$pr8|0);
      if ($6) {
       $gamecontrollerlistprev$04$i$lcssa = $gamecontrollerlistprev$04$i;
       break;
      }
      $11 = ((($gamecontrollerlist$05$i)) + 896|0);
      $gamecontrollerlist$0$i = HEAP32[$11>>2]|0;
      $12 = ($gamecontrollerlist$0$i|0)==(0|0);
      if ($12) {
       break L4;
      } else {
       $gamecontrollerlistprev$04$i$phi = $gamecontrollerlist$05$i;$gamecontrollerlist$05$i = $gamecontrollerlist$0$i;$gamecontrollerlistprev$04$i = $gamecontrollerlistprev$04$i$phi;
      }
     }
     $7 = ($gamecontrollerlistprev$04$i$lcssa|0)==(0|0);
     $8 = ((($$pr8)) + 896|0);
     $9 = HEAP32[$8>>2]|0;
     if ($7) {
      HEAP32[301459] = $9;
      break;
     } else {
      $10 = ((($gamecontrollerlistprev$04$i$lcssa)) + 896|0);
      HEAP32[$10>>2] = $9;
      break;
     }
    }
   } while(0);
   _SDL_free($$pr8);
   $$pr = HEAP32[301459]|0;
   $13 = ($$pr|0)==(0|0);
   if ($13) {
    break;
   } else {
    $$pr8 = $$pr;
   }
  }
 }
 $1 = HEAP32[301460]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  _SDL_DelEventWatch(4649,0);
  return;
 } else {
  $15 = $1;
 }
 while(1) {
  $14 = ((($15)) + 24|0);
  $16 = HEAP32[$14>>2]|0;
  HEAP32[301460] = $16;
  $17 = ((($15)) + 16|0);
  $18 = HEAP32[$17>>2]|0;
  _SDL_free($18);
  $19 = ((($15)) + 20|0);
  $20 = HEAP32[$19>>2]|0;
  _SDL_free($20);
  _SDL_free($15);
  $21 = HEAP32[301460]|0;
  $22 = ($21|0)==(0|0);
  if ($22) {
   break;
  } else {
   $15 = $21;
  }
 }
 _SDL_DelEventWatch(4649,0);
 return;
}
function _SDL_JoystickClose($joystick) {
 $joystick = $joystick|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $joysticklist$0 = 0, $joysticklist$04 = 0, $joysticklist$06 = 0, $joysticklistprev$05 = 0, $joysticklistprev$05$lcssa = 0, $joysticklistprev$05$phi = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($joystick|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($joystick)) + 44|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($2) + -1)|0;
 HEAP32[$1>>2] = $3;
 $4 = ($2|0)>(1);
 $5 = HEAP32[301461]|0;
 $6 = ($5|0)==($joystick|0);
 $or$cond = $4 | $6;
 if ($or$cond) {
  return;
 }
 $7 = ((($joystick)) + 40|0);
 HEAP32[$7>>2] = 0;
 $joysticklist$04 = HEAP32[301462]|0;
 $8 = ($joysticklist$04|0)==(0|0);
 L7: do {
  if (!($8)) {
   $joysticklist$06 = $joysticklist$04;$joysticklistprev$05 = 0;
   while(1) {
    $9 = ($joysticklist$06|0)==($joystick|0);
    if ($9) {
     $joysticklistprev$05$lcssa = $joysticklistprev$05;
     break;
    }
    $14 = ((($joysticklist$06)) + 56|0);
    $joysticklist$0 = HEAP32[$14>>2]|0;
    $15 = ($joysticklist$0|0)==(0|0);
    if ($15) {
     break L7;
    } else {
     $joysticklistprev$05$phi = $joysticklist$06;$joysticklist$06 = $joysticklist$0;$joysticklistprev$05 = $joysticklistprev$05$phi;
    }
   }
   $10 = ($joysticklistprev$05$lcssa|0)==(0|0);
   $11 = ((($joystick)) + 56|0);
   $12 = HEAP32[$11>>2]|0;
   if ($10) {
    HEAP32[301462] = $12;
    break;
   } else {
    $13 = ((($joysticklistprev$05$lcssa)) + 56|0);
    HEAP32[$13>>2] = $12;
    break;
   }
  }
 } while(0);
 $16 = ((($joystick)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 _SDL_free($17);
 $18 = ((($joystick)) + 12|0);
 $19 = HEAP32[$18>>2]|0;
 _SDL_free($19);
 $20 = ((($joystick)) + 20|0);
 $21 = HEAP32[$20>>2]|0;
 _SDL_free($21);
 $22 = ((($joystick)) + 28|0);
 $23 = HEAP32[$22>>2]|0;
 _SDL_free($23);
 $24 = ((($joystick)) + 36|0);
 $25 = HEAP32[$24>>2]|0;
 _SDL_free($25);
 _SDL_free($joystick);
 return;
}
function _SDL_SYS_JoystickClose($joystick) {
 $joystick = $joystick|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _SDL_GameControllerEventWatcher($userdata,$event) {
 $userdata = $userdata|0;
 $event = $event|0;
 var $$ = 0, $$lcssa = 0, $$lcssa109 = 0, $$lcssa112 = 0, $$mask = 0, $$mask71 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0;
 var $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $controllerlist$0 = 0, $controllerlist$049 = 0, $controllerlist$050 = 0, $controllerlist$050$lcssa = 0, $controllerlist1$0 = 0, $controllerlist1$051 = 0, $controllerlist1$052 = 0, $controllerlist1$052$lcssa = 0;
 var $controllerlist3$0 = 0, $controllerlist3$054 = 0, $controllerlist3$055 = 0, $controllerlist3$055$lcssa = 0, $controllerlist5$0 = 0, $controllerlist5$057 = 0, $controllerlist5$058 = 0, $deviceevent = 0, $deviceevent6 = 0, $event$i = 0, $event$i11 = 0, $event$i14 = 0, $event$i17 = 0, $event$i20 = 0, $event$i23 = 0, $event$i26 = 0, $event$i29 = 0, $event$i32 = 0, $event$i4 = 0, $event$i5 = 0;
 var $event$i8 = 0, $not$$i = 0, $switch = 0, $value$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 784|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $event$i32 = sp + 728|0;
 $event$i29 = sp + 672|0;
 $event$i26 = sp + 616|0;
 $event$i23 = sp + 560|0;
 $event$i20 = sp + 504|0;
 $event$i17 = sp + 448|0;
 $event$i14 = sp + 392|0;
 $event$i11 = sp + 336|0;
 $event$i8 = sp + 280|0;
 $event$i5 = sp + 224|0;
 $event$i4 = sp + 168|0;
 $event$i = sp + 112|0;
 $deviceevent = sp + 56|0;
 $deviceevent6 = sp;
 $0 = HEAP32[$event>>2]|0;
 switch ($0|0) {
 case 1536:  {
  $4 = ((($event)) + 12|0);
  $5 = HEAP8[$4>>0]|0;
  $6 = ($5&255)>(19);
  if ($6) {
   STACKTOP = sp;return 1;
  }
  $controllerlist$049 = HEAP32[301459]|0;
  $7 = ($controllerlist$049|0)==(0|0);
  if ($7) {
   STACKTOP = sp;return 1;
  }
  $8 = ((($event)) + 8|0);
  $9 = HEAP32[$8>>2]|0;
  $controllerlist$050 = $controllerlist$049;
  while(1) {
   $10 = HEAP32[$controllerlist$050>>2]|0;
   $11 = HEAP32[$10>>2]|0;
   $12 = ($11|0)==($9|0);
   if ($12) {
    $$lcssa = $controllerlist$050;$controllerlist$050$lcssa = $controllerlist$050;
    break;
   }
   $51 = ((($controllerlist$050)) + 896|0);
   $controllerlist$0 = HEAP32[$51>>2]|0;
   $52 = ($controllerlist$0|0)==(0|0);
   if ($52) {
    label = 18;
    break;
   } else {
    $controllerlist$050 = $controllerlist$0;
   }
  }
  if ((label|0) == 18) {
   STACKTOP = sp;return 1;
  }
  $13 = $5&255;
  $14 = (((($controllerlist$050$lcssa)) + 320|0) + ($13<<2)|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ($15|0)>(-1);
  if ($16) {
   $17 = ((($event)) + 16|0);
   $18 = HEAP16[$17>>1]|0;
   $19 = $15 & -2;
   $switch = ($19|0)==(4);
   if ($switch) {
    $20 = $18 << 16 >> 16;
    $21 = (($20|0) / 2)&-1;
    $22 = (($21) + 16384)|0;
    $23 = $22&65535;
    $value$0 = $23;
   } else {
    $value$0 = $18;
   }
   $24 = (_SDL_EventState(1616,-1)|0);
   $25 = ($24<<24>>24)==(1);
   if (!($25)) {
    STACKTOP = sp;return 1;
   }
   HEAP32[$event$i>>2] = 1616;
   $26 = HEAP32[$$lcssa>>2]|0;
   $27 = HEAP32[$26>>2]|0;
   $28 = ((($event$i)) + 8|0);
   HEAP32[$28>>2] = $27;
   $29 = $15&255;
   $30 = ((($event$i)) + 12|0);
   HEAP8[$30>>0] = $29;
   $31 = ((($event$i)) + 16|0);
   HEAP16[$31>>1] = $value$0;
   (_SDL_PushEvent($event$i)|0);
   STACKTOP = sp;return 1;
  } else {
   $32 = (((($controllerlist$050$lcssa)) + 560|0) + ($13<<2)|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ($33|0)>(-1);
   if (!($34)) {
    STACKTOP = sp;return 1;
   }
   $35 = ((($event)) + 16|0);
   $36 = HEAP16[$35>>1]|0;
   $37 = $36 << 16 >> 16;
   $38 = ($36<<16>>16)<(0);
   $39 = (0 - ($37))|0;
   $40 = $38 ? $39 : $37;
   $41 = ($40|0)>(16384);
   $$ = $41 ? 1617 : 1618;
   HEAP32[$event$i4>>2] = $$;
   $42 = (_SDL_EventState($$,-1)|0);
   $43 = ($42<<24>>24)==(1);
   if ($43) {
    $44 = $41&1;
    $45 = HEAP32[$$lcssa>>2]|0;
    $46 = HEAP32[$45>>2]|0;
    $47 = ((($event$i4)) + 8|0);
    HEAP32[$47>>2] = $46;
    $48 = $33&255;
    $49 = ((($event$i4)) + 12|0);
    HEAP8[$49>>0] = $48;
    $50 = ((($event$i4)) + 13|0);
    HEAP8[$50>>0] = $44;
    (_SDL_PushEvent($event$i4)|0);
   }
   STACKTOP = sp;return 1;
  }
  break;
 }
 case 1540: case 1539:  {
  $53 = ((($event)) + 12|0);
  $54 = HEAP16[$53>>1]|0;
  $55 = $54&255;
  $56 = ($55&255)>(19);
  $57 = ($54&65535) >>> 8;
  $58 = $57&255;
  if ($56) {
   STACKTOP = sp;return 1;
  }
  $controllerlist1$051 = HEAP32[301459]|0;
  $59 = ($controllerlist1$051|0)==(0|0);
  if ($59) {
   STACKTOP = sp;return 1;
  }
  $60 = ((($event)) + 8|0);
  $61 = HEAP32[$60>>2]|0;
  $controllerlist1$052 = $controllerlist1$051;
  while(1) {
   $62 = HEAP32[$controllerlist1$052>>2]|0;
   $63 = HEAP32[$62>>2]|0;
   $64 = ($63|0)==($61|0);
   if ($64) {
    $$lcssa109 = $controllerlist1$052;$controllerlist1$052$lcssa = $controllerlist1$052;
    break;
   }
   $92 = ((($controllerlist1$052)) + 896|0);
   $controllerlist1$0 = HEAP32[$92>>2]|0;
   $93 = ($controllerlist1$0|0)==(0|0);
   if ($93) {
    label = 18;
    break;
   } else {
    $controllerlist1$052 = $controllerlist1$0;
   }
  }
  if ((label|0) == 18) {
   STACKTOP = sp;return 1;
  }
  $$mask71 = $54 & 255;
  $65 = $$mask71&65535;
  $66 = (((($controllerlist1$052$lcssa)) + 480|0) + ($65<<2)|0);
  $67 = HEAP32[$66>>2]|0;
  $68 = ($67|0)>(-1);
  if (!($68)) {
   $79 = (((($controllerlist1$052$lcssa)) + 400|0) + ($65<<2)|0);
   $80 = HEAP32[$79>>2]|0;
   $81 = ($80|0)>(-1);
   if (!($81)) {
    STACKTOP = sp;return 1;
   }
   $82 = (_SDL_EventState(1616,-1)|0);
   $83 = ($82<<24>>24)==(1);
   if (!($83)) {
    STACKTOP = sp;return 1;
   }
   $84 = ($58<<24>>24)!=(0);
   $85 = $84 ? 32767 : 0;
   HEAP32[$event$i8>>2] = 1616;
   $86 = HEAP32[$$lcssa109>>2]|0;
   $87 = HEAP32[$86>>2]|0;
   $88 = ((($event$i8)) + 8|0);
   HEAP32[$88>>2] = $87;
   $89 = $80&255;
   $90 = ((($event$i8)) + 12|0);
   HEAP8[$90>>0] = $89;
   $91 = ((($event$i8)) + 16|0);
   HEAP16[$91>>1] = $85;
   (_SDL_PushEvent($event$i8)|0);
   STACKTOP = sp;return 1;
  }
  $69 = $57&65535;
  switch ($69|0) {
  case 1:  {
   HEAP32[$event$i5>>2] = 1617;
   $70 = 1617;
   label = 27;
   break;
  }
  case 0:  {
   HEAP32[$event$i5>>2] = 1618;
   $70 = 1618;
   label = 27;
   break;
  }
  default: {
  }
  }
  if ((label|0) == 27) {
   $71 = (_SDL_EventState($70,-1)|0);
   $72 = ($71<<24>>24)==(1);
   if ($72) {
    $73 = HEAP32[$$lcssa109>>2]|0;
    $74 = HEAP32[$73>>2]|0;
    $75 = ((($event$i5)) + 8|0);
    HEAP32[$75>>2] = $74;
    $76 = $67&255;
    $77 = ((($event$i5)) + 12|0);
    HEAP8[$77>>0] = $76;
    $78 = ((($event$i5)) + 13|0);
    HEAP8[$78>>0] = $58;
    (_SDL_PushEvent($event$i5)|0);
   }
  }
  STACKTOP = sp;return 1;
  break;
 }
 case 1538:  {
  $94 = ((($event)) + 12|0);
  $95 = HEAP16[$94>>1]|0;
  $96 = $95&255;
  $97 = ($96&255)>(3);
  $98 = ($95&65535) >>> 8;
  $99 = $98&255;
  if ($97) {
   STACKTOP = sp;return 1;
  }
  $controllerlist3$054 = HEAP32[301459]|0;
  $100 = ($controllerlist3$054|0)==(0|0);
  if ($100) {
   STACKTOP = sp;return 1;
  }
  $101 = ((($event)) + 8|0);
  $102 = HEAP32[$101>>2]|0;
  $controllerlist3$055 = $controllerlist3$054;
  while(1) {
   $103 = HEAP32[$controllerlist3$055>>2]|0;
   $104 = HEAP32[$103>>2]|0;
   $105 = ($104|0)==($102|0);
   if ($105) {
    $$lcssa112 = $controllerlist3$055;$controllerlist3$055$lcssa = $controllerlist3$055;
    break;
   }
   $233 = ((($controllerlist3$055)) + 896|0);
   $controllerlist3$0 = HEAP32[$233>>2]|0;
   $234 = ($controllerlist3$0|0)==(0|0);
   if ($234) {
    label = 18;
    break;
   } else {
    $controllerlist3$055 = $controllerlist3$0;
   }
  }
  if ((label|0) == 18) {
   STACKTOP = sp;return 1;
  }
  $$mask = $95 & 255;
  $106 = $$mask&65535;
  $107 = (((($controllerlist3$055$lcssa)) + 8|0) + ($106)|0);
  $108 = HEAP8[$107>>0]|0;
  $109 = ((($event)) + 13|0);
  $110 = $99 & $108;
  $111 = $110 ^ $108;
  $112 = $106 << 4;
  $113 = $111&255;
  $114 = $113 & 4;
  $115 = ($114|0)==(0);
  if (!($115)) {
   $116 = $112 | 4;
   $117 = (((($controllerlist3$055$lcssa)) + 640|0) + ($116<<2)|0);
   $118 = HEAP32[$117>>2]|0;
   $119 = ($118|0)==(-1);
   if (!($119)) {
    HEAP32[$event$i11>>2] = 1618;
    $120 = (_SDL_EventState(1618,-1)|0);
    $121 = ($120<<24>>24)==(1);
    if ($121) {
     $122 = HEAP32[$$lcssa112>>2]|0;
     $123 = HEAP32[$122>>2]|0;
     $124 = ((($event$i11)) + 8|0);
     HEAP32[$124>>2] = $123;
     $125 = $118&255;
     $126 = ((($event$i11)) + 12|0);
     HEAP8[$126>>0] = $125;
     $127 = ((($event$i11)) + 13|0);
     HEAP8[$127>>0] = 0;
     (_SDL_PushEvent($event$i11)|0);
    }
   }
  }
  $128 = $113 & 1;
  $129 = ($128|0)==(0);
  if (!($129)) {
   $130 = $112 | 1;
   $131 = (((($controllerlist3$055$lcssa)) + 640|0) + ($130<<2)|0);
   $132 = HEAP32[$131>>2]|0;
   $133 = ($132|0)==(-1);
   if (!($133)) {
    HEAP32[$event$i14>>2] = 1618;
    $134 = (_SDL_EventState(1618,-1)|0);
    $135 = ($134<<24>>24)==(1);
    if ($135) {
     $136 = HEAP32[$$lcssa112>>2]|0;
     $137 = HEAP32[$136>>2]|0;
     $138 = ((($event$i14)) + 8|0);
     HEAP32[$138>>2] = $137;
     $139 = $132&255;
     $140 = ((($event$i14)) + 12|0);
     HEAP8[$140>>0] = $139;
     $141 = ((($event$i14)) + 13|0);
     HEAP8[$141>>0] = 0;
     (_SDL_PushEvent($event$i14)|0);
    }
   }
  }
  $142 = $113 & 8;
  $143 = ($142|0)==(0);
  if (!($143)) {
   $144 = $112 | 8;
   $145 = (((($controllerlist3$055$lcssa)) + 640|0) + ($144<<2)|0);
   $146 = HEAP32[$145>>2]|0;
   $147 = ($146|0)==(-1);
   if (!($147)) {
    HEAP32[$event$i17>>2] = 1618;
    $148 = (_SDL_EventState(1618,-1)|0);
    $149 = ($148<<24>>24)==(1);
    if ($149) {
     $150 = HEAP32[$$lcssa112>>2]|0;
     $151 = HEAP32[$150>>2]|0;
     $152 = ((($event$i17)) + 8|0);
     HEAP32[$152>>2] = $151;
     $153 = $146&255;
     $154 = ((($event$i17)) + 12|0);
     HEAP8[$154>>0] = $153;
     $155 = ((($event$i17)) + 13|0);
     HEAP8[$155>>0] = 0;
     (_SDL_PushEvent($event$i17)|0);
    }
   }
  }
  $156 = $113 & 2;
  $157 = ($156|0)==(0);
  if (!($157)) {
   $158 = $112 | 2;
   $159 = (((($controllerlist3$055$lcssa)) + 640|0) + ($158<<2)|0);
   $160 = HEAP32[$159>>2]|0;
   $161 = ($160|0)==(-1);
   if (!($161)) {
    HEAP32[$event$i20>>2] = 1618;
    $162 = (_SDL_EventState(1618,-1)|0);
    $163 = ($162<<24>>24)==(1);
    if ($163) {
     $164 = HEAP32[$$lcssa112>>2]|0;
     $165 = HEAP32[$164>>2]|0;
     $166 = ((($event$i20)) + 8|0);
     HEAP32[$166>>2] = $165;
     $167 = $160&255;
     $168 = ((($event$i20)) + 12|0);
     HEAP8[$168>>0] = $167;
     $169 = ((($event$i20)) + 13|0);
     HEAP8[$169>>0] = 0;
     (_SDL_PushEvent($event$i20)|0);
    }
   }
  }
  $170 = HEAP8[$109>>0]|0;
  $171 = $170 ^ $110;
  $172 = $171&255;
  $173 = $172 & 4;
  $174 = ($173|0)==(0);
  if (!($174)) {
   $175 = $112 | 4;
   $176 = (((($controllerlist3$055$lcssa)) + 640|0) + ($175<<2)|0);
   $177 = HEAP32[$176>>2]|0;
   $178 = ($177|0)==(-1);
   if (!($178)) {
    HEAP32[$event$i23>>2] = 1617;
    $179 = (_SDL_EventState(1617,-1)|0);
    $180 = ($179<<24>>24)==(1);
    if ($180) {
     $181 = HEAP32[$$lcssa112>>2]|0;
     $182 = HEAP32[$181>>2]|0;
     $183 = ((($event$i23)) + 8|0);
     HEAP32[$183>>2] = $182;
     $184 = $177&255;
     $185 = ((($event$i23)) + 12|0);
     HEAP8[$185>>0] = $184;
     $186 = ((($event$i23)) + 13|0);
     HEAP8[$186>>0] = 1;
     (_SDL_PushEvent($event$i23)|0);
    }
   }
  }
  $187 = $172 & 1;
  $188 = ($187|0)==(0);
  if (!($188)) {
   $189 = $112 | 1;
   $190 = (((($controllerlist3$055$lcssa)) + 640|0) + ($189<<2)|0);
   $191 = HEAP32[$190>>2]|0;
   $192 = ($191|0)==(-1);
   if (!($192)) {
    HEAP32[$event$i26>>2] = 1617;
    $193 = (_SDL_EventState(1617,-1)|0);
    $194 = ($193<<24>>24)==(1);
    if ($194) {
     $195 = HEAP32[$$lcssa112>>2]|0;
     $196 = HEAP32[$195>>2]|0;
     $197 = ((($event$i26)) + 8|0);
     HEAP32[$197>>2] = $196;
     $198 = $191&255;
     $199 = ((($event$i26)) + 12|0);
     HEAP8[$199>>0] = $198;
     $200 = ((($event$i26)) + 13|0);
     HEAP8[$200>>0] = 1;
     (_SDL_PushEvent($event$i26)|0);
    }
   }
  }
  $201 = $172 & 8;
  $202 = ($201|0)==(0);
  if (!($202)) {
   $203 = $112 | 8;
   $204 = (((($controllerlist3$055$lcssa)) + 640|0) + ($203<<2)|0);
   $205 = HEAP32[$204>>2]|0;
   $206 = ($205|0)==(-1);
   if (!($206)) {
    HEAP32[$event$i29>>2] = 1617;
    $207 = (_SDL_EventState(1617,-1)|0);
    $208 = ($207<<24>>24)==(1);
    if ($208) {
     $209 = HEAP32[$$lcssa112>>2]|0;
     $210 = HEAP32[$209>>2]|0;
     $211 = ((($event$i29)) + 8|0);
     HEAP32[$211>>2] = $210;
     $212 = $205&255;
     $213 = ((($event$i29)) + 12|0);
     HEAP8[$213>>0] = $212;
     $214 = ((($event$i29)) + 13|0);
     HEAP8[$214>>0] = 1;
     (_SDL_PushEvent($event$i29)|0);
    }
   }
  }
  $215 = $172 & 2;
  $216 = ($215|0)==(0);
  if (!($216)) {
   $217 = $112 | 2;
   $218 = (((($controllerlist3$055$lcssa)) + 640|0) + ($217<<2)|0);
   $219 = HEAP32[$218>>2]|0;
   $220 = ($219|0)==(-1);
   if (!($220)) {
    HEAP32[$event$i32>>2] = 1617;
    $221 = (_SDL_EventState(1617,-1)|0);
    $222 = ($221<<24>>24)==(1);
    if ($222) {
     $223 = HEAP32[$$lcssa112>>2]|0;
     $224 = HEAP32[$223>>2]|0;
     $225 = ((($event$i32)) + 8|0);
     HEAP32[$225>>2] = $224;
     $226 = $219&255;
     $227 = ((($event$i32)) + 12|0);
     HEAP8[$227>>0] = $226;
     $228 = ((($event$i32)) + 13|0);
     HEAP8[$228>>0] = 1;
     (_SDL_PushEvent($event$i32)|0);
    }
   }
  }
  $229 = HEAP8[$109>>0]|0;
  $230 = HEAP8[$94>>0]|0;
  $231 = $230&255;
  $232 = (((($controllerlist3$055$lcssa)) + 8|0) + ($231)|0);
  HEAP8[$232>>0] = $229;
  STACKTOP = sp;return 1;
  break;
 }
 case 1541:  {
  $235 = ((($event)) + 8|0);
  $236 = HEAP32[$235>>2]|0;
  $237 = (_SDL_PrivateGetControllerMapping($236)|0);
  $not$$i = ($237|0)==(0|0);
  if ($not$$i) {
   STACKTOP = sp;return 1;
  }
  HEAP32[$deviceevent>>2] = 1619;
  $238 = HEAP32[$235>>2]|0;
  $239 = ((($deviceevent)) + 8|0);
  HEAP32[$239>>2] = $238;
  (_SDL_PushEvent($deviceevent)|0);
  STACKTOP = sp;return 1;
  break;
 }
 case 1542:  {
  $controllerlist5$057 = HEAP32[301459]|0;
  $1 = ($controllerlist5$057|0)==(0|0);
  if ($1) {
   STACKTOP = sp;return 1;
  }
  $2 = ((($event)) + 8|0);
  $3 = HEAP32[$2>>2]|0;
  $controllerlist5$058 = $controllerlist5$057;
  while(1) {
   $242 = HEAP32[$controllerlist5$058>>2]|0;
   $243 = HEAP32[$242>>2]|0;
   $244 = ($243|0)==($3|0);
   if ($244) {
    break;
   }
   $240 = ((($controllerlist5$058)) + 896|0);
   $controllerlist5$0 = HEAP32[$240>>2]|0;
   $241 = ($controllerlist5$0|0)==(0|0);
   if ($241) {
    label = 18;
    break;
   } else {
    $controllerlist5$058 = $controllerlist5$0;
   }
  }
  if ((label|0) == 18) {
   STACKTOP = sp;return 1;
  }
  HEAP32[$deviceevent6>>2] = 1620;
  $245 = ((($deviceevent6)) + 8|0);
  HEAP32[$245>>2] = $3;
  (_SDL_PushEvent($deviceevent6)|0);
  STACKTOP = sp;return 1;
  break;
 }
 default: {
  STACKTOP = sp;return 1;
 }
 }
 return (0)|0;
}
function _SDL_PrivateGetControllerMapping($device_index) {
 $device_index = $device_index|0;
 var $$ = 0, $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $jGUID = 0, $mapping$2 = 0;
 var $or$cond = 0, $pSupportedController$0$i = 0, $pSupportedController$01$i = 0, $pSupportedController$02$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $jGUID = sp;
 _SDL_JoystickGetDeviceGUID($jGUID,$device_index);
 $pSupportedController$01$i = HEAP32[301460]|0;
 $0 = ($pSupportedController$01$i|0)==(0|0);
 L1: do {
  if ($0) {
   $$0$i = 0;
  } else {
   $pSupportedController$02$i = $pSupportedController$01$i;
   while(1) {
    $1 = (_SDL_memcmp($jGUID,$pSupportedController$02$i,16)|0);
    $2 = ($1|0)==(0);
    if ($2) {
     $$0$i = $pSupportedController$02$i;
     break L1;
    }
    $3 = ((($pSupportedController$02$i)) + 24|0);
    $pSupportedController$0$i = HEAP32[$3>>2]|0;
    $4 = ($pSupportedController$0$i|0)==(0|0);
    if ($4) {
     $$0$i = 0;
     break;
    } else {
     $pSupportedController$02$i = $pSupportedController$0$i;
    }
   }
  }
 } while(0);
 $5 = ($$0$i|0)==(0|0);
 $6 = HEAP32[301465]|0;
 $7 = ($6|0)!=(0|0);
 $or$cond = $5 & $7;
 $$ = $or$cond ? $6 : $$0$i;
 $8 = ($$|0)==(0|0);
 if (!($8)) {
  $mapping$2 = $$;
  STACKTOP = sp;return ($mapping$2|0);
 }
 $9 = (_SDL_JoystickNameForIndex($device_index)|0);
 $10 = ($9|0)==(0|0);
 if ($10) {
  $mapping$2 = 0;
  STACKTOP = sp;return ($mapping$2|0);
 }
 $11 = (_SDL_strstr($9,609167)|0);
 $12 = ($11|0)==(0|0);
 if ($12) {
  $13 = (_SDL_strstr($9,609172)|0);
  $14 = ($13|0)==(0|0);
  if ($14) {
   $mapping$2 = 0;
   STACKTOP = sp;return ($mapping$2|0);
  }
 }
 $15 = HEAP32[301466]|0;
 $mapping$2 = $15;
 STACKTOP = sp;return ($mapping$2|0);
}
function _SDL_JoystickGetDeviceGUID($agg$result,$device_index) {
 $agg$result = $agg$result|0;
 $device_index = $device_index|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $emptyGUID1 = 0, $vararg_buffer = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $emptyGUID1 = sp + 8|0;
 $0 = ($device_index|0)<(0);
 if (!($0)) {
  $1 = (_SDL_SYS_NumJoysticks()|0);
  $2 = ($1|0)>($device_index|0);
  if ($2) {
   _SDL_SYS_JoystickGetDeviceGUID($agg$result,$device_index);
   STACKTOP = sp;return;
  }
 }
 $3 = (_SDL_SYS_NumJoysticks()|0);
 HEAP32[$vararg_buffer>>2] = $3;
 (_SDL_SetError(611965,$vararg_buffer)|0);
 (_SDL_memset($emptyGUID1,0,16)|0);
 dest=$agg$result; src=$emptyGUID1; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
 STACKTOP = sp;return;
}
function _SDL_SYS_NumJoysticks() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[301463]|0;
 return ($0|0);
}
function _SDL_SYS_JoystickGetDeviceGUID($agg$result,$device_index) {
 $agg$result = $agg$result|0;
 $device_index = $device_index|0;
 var $$02$i$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $guid1 = 0, $item$0$i$i = 0, $item$0$lcssa$i$i = 0, $item$01$i$i = 0, $item$03$i$i = 0, dest = 0, label = 0, sp = 0, src = 0;
 var stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $guid1 = sp;
 $item$01$i$i = HEAP32[301464]|0;
 $0 = ($device_index|0)>(0);
 if ($0) {
  $$02$i$i = $device_index;$item$03$i$i = $item$01$i$i;
  while(1) {
   $1 = (($$02$i$i) + -1)|0;
   $2 = ((($item$03$i$i)) + 1320|0);
   $item$0$i$i = HEAP32[$2>>2]|0;
   $3 = ($$02$i$i|0)>(1);
   if ($3) {
    $$02$i$i = $1;$item$03$i$i = $item$0$i$i;
   } else {
    $item$0$lcssa$i$i = $item$0$i$i;
    break;
   }
  }
 } else {
  $item$0$lcssa$i$i = $item$01$i$i;
 }
 $4 = ((($item$0$lcssa$i$i)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 (_SDL_memset($guid1,0,16)|0);
 $6 = (_SDL_strlen($5)|0);
 $7 = ($6>>>0)>(16);
 if ($7) {
  $9 = 16;
 } else {
  $8 = (_SDL_strlen($5)|0);
  $9 = $8;
 }
 (_SDL_memcpy($guid1,$5,$9)|0);
 dest=$agg$result; src=$guid1; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
 STACKTOP = sp;return;
}
function _SDL_JoystickNameForIndex($device_index) {
 $device_index = $device_index|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($device_index|0)<(0);
 if (!($0)) {
  $1 = (_SDL_SYS_NumJoysticks()|0);
  $2 = ($1|0)>($device_index|0);
  if ($2) {
   $4 = (_SDL_SYS_JoystickNameForDeviceIndex($device_index)|0);
   $$0 = $4;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $3 = (_SDL_SYS_NumJoysticks()|0);
 HEAP32[$vararg_buffer>>2] = $3;
 (_SDL_SetError(611965,$vararg_buffer)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SYS_JoystickNameForDeviceIndex($device_index) {
 $device_index = $device_index|0;
 var $$02$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $item$0$i = 0, $item$0$lcssa$i = 0, $item$01$i = 0, $item$03$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $item$01$i = HEAP32[301464]|0;
 $0 = ($device_index|0)>(0);
 if ($0) {
  $$02$i = $device_index;$item$03$i = $item$01$i;
  while(1) {
   $1 = (($$02$i) + -1)|0;
   $2 = ((($item$03$i)) + 1320|0);
   $item$0$i = HEAP32[$2>>2]|0;
   $3 = ($$02$i|0)>(1);
   if ($3) {
    $$02$i = $1;$item$03$i = $item$0$i;
   } else {
    $item$0$lcssa$i = $item$0$i;
    break;
   }
  }
 } else {
  $item$0$lcssa$i = $item$01$i;
 }
 $4 = ((($item$0$lcssa$i)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 return ($5|0);
}
function _SDL_strstr($haystack,$needle) {
 $haystack = $haystack|0;
 $needle = $needle|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strstr($haystack,$needle)|0);
 return ($0|0);
}
function _SDL_DelEventWatch($filter,$userdata) {
 $filter = $filter|0;
 $userdata = $userdata|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $curr$0 = 0, $curr$03 = 0, $curr$05 = 0, $curr$05$lcssa = 0, $prev$04 = 0, $prev$04$lcssa = 0, $prev$04$phi = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $curr$03 = HEAP32[300734]|0;
 $0 = ($curr$03|0)==(0|0);
 if ($0) {
  return;
 } else {
  $curr$05 = $curr$03;$prev$04 = 0;
 }
 while(1) {
  $1 = HEAP32[$curr$05>>2]|0;
  $2 = ($1|0)==($filter|0);
  if ($2) {
   $3 = ((($curr$05)) + 4|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==($userdata|0);
   if ($5) {
    $curr$05$lcssa = $curr$05;$prev$04$lcssa = $prev$04;
    break;
   }
  }
  $10 = ((($curr$05)) + 8|0);
  $curr$0 = HEAP32[$10>>2]|0;
  $11 = ($curr$0|0)==(0|0);
  if ($11) {
   label = 9;
   break;
  } else {
   $prev$04$phi = $curr$05;$curr$05 = $curr$0;$prev$04 = $prev$04$phi;
  }
 }
 if ((label|0) == 9) {
  return;
 }
 $6 = ($prev$04$lcssa|0)==(0|0);
 $7 = ((($curr$05$lcssa)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 if ($6) {
  HEAP32[300734] = $8;
 } else {
  $9 = ((($prev$04$lcssa)) + 8|0);
  HEAP32[$9>>2] = $8;
 }
 _SDL_free($curr$05$lcssa);
 return;
}
function _SDL_JoystickQuit() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[301462]|0;
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  $3 = $0;
  while(1) {
   $2 = ((($3)) + 44|0);
   HEAP32[$2>>2] = 1;
   _SDL_JoystickClose($3);
   $4 = HEAP32[301462]|0;
   $5 = ($4|0)==(0|0);
   if ($5) {
    break;
   } else {
    $3 = $4;
   }
  }
 }
 _SDL_SYS_JoystickQuit();
 _SDL_QuitSubSystem(16384);
 return;
}
function _SDL_SYS_JoystickQuit() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $item$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[301464]|0;
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  $item$01 = $0;
  while(1) {
   $2 = ((($item$01)) + 1320|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = ((($item$01)) + 8|0);
   $5 = HEAP32[$4>>2]|0;
   _SDL_free($5);
   $6 = ((($item$01)) + 4|0);
   $7 = HEAP32[$6>>2]|0;
   _SDL_free($7);
   _SDL_free($item$01);
   $8 = ($3|0)==(0|0);
   if ($8) {
    break;
   } else {
    $item$01 = $3;
   }
  }
 }
 HEAP32[301467] = 0;
 HEAP32[301464] = 0;
 HEAP32[301463] = 0;
 HEAP32[301468] = 0;
 (_emscripten_set_gamepadconnected_callback((0|0),0,(0|0))|0);
 (_emscripten_set_gamepaddisconnected_callback((0|0),0,(0|0))|0);
 return;
}
function _SDL_AudioQuit() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $item$01$i = 0, $item$01$i1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[301387]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  return;
 }
 $2 = HEAP32[301469]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
  _close_audio_device($2);
  HEAP32[301469] = 0;
 }
 $4 = HEAP32[(1205880)>>2]|0;
 $5 = ($4|0)==(0|0);
 if (!($5)) {
  _close_audio_device($4);
  HEAP32[(1205880)>>2] = 0;
 }
 $22 = HEAP32[(1205884)>>2]|0;
 $23 = ($22|0)==(0|0);
 if (!($23)) {
  _close_audio_device($22);
  HEAP32[(1205884)>>2] = 0;
 }
 $24 = HEAP32[(1205888)>>2]|0;
 $25 = ($24|0)==(0|0);
 if (!($25)) {
  _close_audio_device($24);
  HEAP32[(1205888)>>2] = 0;
 }
 $26 = HEAP32[(1205892)>>2]|0;
 $27 = ($26|0)==(0|0);
 if (!($27)) {
  _close_audio_device($26);
  HEAP32[(1205892)>>2] = 0;
 }
 $28 = HEAP32[(1205896)>>2]|0;
 $29 = ($28|0)==(0|0);
 if (!($29)) {
  _close_audio_device($28);
  HEAP32[(1205896)>>2] = 0;
 }
 $30 = HEAP32[(1205900)>>2]|0;
 $31 = ($30|0)==(0|0);
 if (!($31)) {
  _close_audio_device($30);
  HEAP32[(1205900)>>2] = 0;
 }
 $32 = HEAP32[(1205904)>>2]|0;
 $33 = ($32|0)==(0|0);
 if (!($33)) {
  _close_audio_device($32);
  HEAP32[(1205904)>>2] = 0;
 }
 $34 = HEAP32[(1205908)>>2]|0;
 $35 = ($34|0)==(0|0);
 if (!($35)) {
  _close_audio_device($34);
  HEAP32[(1205908)>>2] = 0;
 }
 $36 = HEAP32[(1205912)>>2]|0;
 $37 = ($36|0)==(0|0);
 if (!($37)) {
  _close_audio_device($36);
  HEAP32[(1205912)>>2] = 0;
 }
 $38 = HEAP32[(1205916)>>2]|0;
 $39 = ($38|0)==(0|0);
 if (!($39)) {
  _close_audio_device($38);
  HEAP32[(1205916)>>2] = 0;
 }
 $40 = HEAP32[(1205920)>>2]|0;
 $41 = ($40|0)==(0|0);
 if (!($41)) {
  _close_audio_device($40);
  HEAP32[(1205920)>>2] = 0;
 }
 $42 = HEAP32[(1205924)>>2]|0;
 $43 = ($42|0)==(0|0);
 if (!($43)) {
  _close_audio_device($42);
  HEAP32[(1205924)>>2] = 0;
 }
 $44 = HEAP32[(1205928)>>2]|0;
 $45 = ($44|0)==(0|0);
 if (!($45)) {
  _close_audio_device($44);
  HEAP32[(1205928)>>2] = 0;
 }
 $46 = HEAP32[(1205932)>>2]|0;
 $47 = ($46|0)==(0|0);
 if (!($47)) {
  _close_audio_device($46);
  HEAP32[(1205932)>>2] = 0;
 }
 $48 = HEAP32[(1205936)>>2]|0;
 $49 = ($48|0)==(0|0);
 if (!($49)) {
  _close_audio_device($48);
  HEAP32[(1205936)>>2] = 0;
 }
 $50 = HEAP32[(1205652)>>2]|0;
 $51 = ($50|0)==(0|0);
 if (!($51)) {
  $item$01$i = $50;
  while(1) {
   $6 = ((($item$01$i)) + 4|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = HEAP32[$item$01$i>>2]|0;
   $9 = ($8|0)==(0|0);
   if (!($9)) {
    $10 = HEAP32[(1205600)>>2]|0;
    FUNCTION_TABLE_vi[$10 & 8191]($8);
   }
   _SDL_free($item$01$i);
   $11 = ($7|0)==(0|0);
   if ($11) {
    break;
   } else {
    $item$01$i = $7;
   }
  }
 }
 HEAP32[(1205652)>>2] = 0;
 HEAP32[(1205644)>>2] = 0;
 $12 = HEAP32[(1205656)>>2]|0;
 $13 = ($12|0)==(0|0);
 if (!($13)) {
  $item$01$i1 = $12;
  while(1) {
   $14 = ((($item$01$i1)) + 4|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = HEAP32[$item$01$i1>>2]|0;
   $17 = ($16|0)==(0|0);
   if (!($17)) {
    $18 = HEAP32[(1205600)>>2]|0;
    FUNCTION_TABLE_vi[$18 & 8191]($16);
   }
   _SDL_free($item$01$i1);
   $19 = ($15|0)==(0|0);
   if ($19) {
    break;
   } else {
    $item$01$i1 = $15;
   }
  }
 }
 HEAP32[(1205656)>>2] = 0;
 HEAP32[(1205648)>>2] = 0;
 $20 = HEAP32[(1205604)>>2]|0;
 FUNCTION_TABLE_v[$20 & 8191]();
 $21 = HEAP32[(1205632)>>2]|0;
 _SDL_DestroyMutex($21);
 (_SDL_memset(1205548,0,112)|0);
 (_SDL_memset(1205876,0,64)|0);
 return;
}
function _close_audio_device($device) {
 $device = $device|0;
 var $$01$i = 0, $$01$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($device)) + 136|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($device)) + 140|0);
 HEAP32[$1>>2] = 1;
 $2 = ((($device)) + 160|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  _SDL_WaitThread($3,0);
 }
 $5 = ((($device)) + 156|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0|0);
 if (!($7)) {
  _SDL_DestroyMutex($6);
 }
 $8 = ((($device)) + 152|0);
 $9 = HEAP32[$8>>2]|0;
 _SDL_free($9);
 $10 = ((($device)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0);
 if (!($12)) {
  $13 = ((($device)) + 44|0);
  $14 = HEAP32[$13>>2]|0;
  _SDL_free($14);
 }
 $15 = ((($device)) + 148|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ($16|0)==(0);
 if (!($17)) {
  $18 = HEAP32[(1205588)>>2]|0;
  FUNCTION_TABLE_vi[$18 & 8191]($device);
  HEAP32[$15>>2] = 0;
 }
 $19 = ((($device)) + 168|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ($20|0)==(0|0);
 if (!($21)) {
  $$01$i = $20;
  while(1) {
   $22 = ((($$01$i)) + 8200|0);
   $23 = HEAP32[$22>>2]|0;
   _SDL_free($$01$i);
   $24 = ($23|0)==(0|0);
   if ($24) {
    break;
   } else {
    $$01$i = $23;
   }
  }
 }
 $25 = ((($device)) + 176|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ($26|0)==(0|0);
 if ($27) {
  _SDL_free($device);
  return;
 } else {
  $$01$i1 = $26;
 }
 while(1) {
  $28 = ((($$01$i1)) + 8200|0);
  $29 = HEAP32[$28>>2]|0;
  _SDL_free($$01$i1);
  $30 = ($29|0)==(0|0);
  if ($30) {
   break;
  } else {
   $$01$i1 = $29;
  }
 }
 _SDL_free($device);
 return;
}
function _SDL_QuitQuit() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $action$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $action$i = sp;
 $0 = HEAP32[301415]|0;
 $1 = ($0|0)==(0);
 if (!($1)) {
  STACKTOP = sp;return;
 }
 (_sigaction(2,(0|0),($action$i|0))|0);
 $2 = HEAP32[$action$i>>2]|0;
 $3 = ($2|0)==(4644|0);
 if ($3) {
  HEAP32[$action$i>>2] = 0;
  (_sigaction(2,($action$i|0),(0|0))|0);
 }
 (_sigaction(15,(0|0),($action$i|0))|0);
 $4 = HEAP32[$action$i>>2]|0;
 $5 = ($4|0)==(4644|0);
 if ($5) {
  HEAP32[$action$i>>2] = 0;
  (_sigaction(15,($action$i|0),(0|0))|0);
 }
 STACKTOP = sp;return;
}
function _SDL_StopEventLoop() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $entry$014 = 0, $entry$110 = 0, $exitcond = 0, $i$01 = 0, $vararg_buffer = 0, $wmmsg$06 = 0, $wmmsg$12 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_SDL_GetHint(609178)|0);
 $1 = HEAP32[31472]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
 }
 HEAP32[(125892)>>2] = 0;
 $3 = ($0|0)==(0|0);
 if (!($3)) {
  $4 = (_SDL_atoi($0)|0);
  $5 = ($4|0)==(0);
  if (!($5)) {
   $6 = HEAP32[(125900)>>2]|0;
   HEAP32[$vararg_buffer>>2] = $6;
   _SDL_Log(609205,$vararg_buffer);
  }
 }
 $7 = HEAP32[(125904)>>2]|0;
 $8 = ($7|0)==(0|0);
 if (!($8)) {
  $entry$014 = $7;
  while(1) {
   $9 = ((($entry$014)) + 72|0);
   $10 = HEAP32[$9>>2]|0;
   _SDL_free($entry$014);
   $11 = ($10|0)==(0|0);
   if ($11) {
    break;
   } else {
    $entry$014 = $10;
   }
  }
 }
 $12 = HEAP32[(125912)>>2]|0;
 $13 = ($12|0)==(0|0);
 if (!($13)) {
  $entry$110 = $12;
  while(1) {
   $14 = ((($entry$110)) + 72|0);
   $15 = HEAP32[$14>>2]|0;
   _SDL_free($entry$110);
   $16 = ($15|0)==(0|0);
   if ($16) {
    break;
   } else {
    $entry$110 = $15;
   }
  }
 }
 $17 = HEAP32[(125916)>>2]|0;
 $18 = ($17|0)==(0|0);
 if (!($18)) {
  $wmmsg$06 = $17;
  while(1) {
   $19 = ((($wmmsg$06)) + 12|0);
   $20 = HEAP32[$19>>2]|0;
   _SDL_free($wmmsg$06);
   $21 = ($20|0)==(0|0);
   if ($21) {
    break;
   } else {
    $wmmsg$06 = $20;
   }
  }
 }
 $22 = HEAP32[(125920)>>2]|0;
 $23 = ($22|0)==(0|0);
 if (!($23)) {
  $wmmsg$12 = $22;
  while(1) {
   $24 = ((($wmmsg$12)) + 12|0);
   $25 = HEAP32[$24>>2]|0;
   _SDL_free($wmmsg$12);
   $26 = ($25|0)==(0|0);
   if ($26) {
    break;
   } else {
    $wmmsg$12 = $25;
   }
  }
 }
 HEAP32[(125896)>>2] = 0;
 HEAP32[(125900)>>2] = 0;
 ;HEAP32[(125904)>>2]=0|0;HEAP32[(125904)+4>>2]=0|0;HEAP32[(125904)+8>>2]=0|0;HEAP32[(125904)+12>>2]=0|0;HEAP32[(125904)+16>>2]=0|0;
 $i$01 = 0;
 while(1) {
  $29 = (1201888 + ($i$01<<2)|0);
  $30 = HEAP32[$29>>2]|0;
  _SDL_free($30);
  HEAP32[$29>>2] = 0;
  $31 = (($i$01) + 1)|0;
  $exitcond = ($31|0)==(256);
  if ($exitcond) {
   break;
  } else {
   $i$01 = $31;
  }
 }
 $27 = HEAP32[300734]|0;
 $28 = ($27|0)==(0|0);
 if (!($28)) {
  $33 = $27;
  while(1) {
   $32 = ((($33)) + 8|0);
   $34 = HEAP32[$32>>2]|0;
   HEAP32[300734] = $34;
   _SDL_free($33);
   $35 = HEAP32[300734]|0;
   $36 = ($35|0)==(0|0);
   if ($36) {
    break;
   } else {
    $33 = $35;
   }
  }
 }
 HEAP32[300435] = 0;
 $37 = HEAP32[31472]|0;
 $38 = ($37|0)==(0|0);
 if ($38) {
  STACKTOP = sp;return;
 }
 $39 = HEAP32[31472]|0;
 _SDL_DestroyMutex($39);
 HEAP32[31472] = 0;
 STACKTOP = sp;return;
}
function _SDL_Log($fmt,$varargs) {
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $ap = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ap = sp;
 HEAP32[$ap>>2] = $varargs;
 _SDL_LogMessageV(0,3,$fmt,$ap);
 STACKTOP = sp;return;
}
function _SDL_DestroyWindow($window) {
 $window = $window|0;
 var $$0$i = 0, $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ((($window)) + 116|0);
   HEAP32[$6>>2] = 1;
   $7 = ((($window)) + 48|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = $8 & 4;
   $10 = ($9|0)==(0);
   if (!($10)) {
    $11 = ((($window)) + 112|0);
    HEAP32[$11>>2] = 1;
    (_SDL_UpdateFullscreenMode($window,0)|0);
    $12 = HEAP32[300441]|0;
    $13 = ((($12)) + 64|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = ($14|0)==(0|0);
    if (!($15)) {
     FUNCTION_TABLE_vii[$14 & 8191]($12,$window);
    }
    HEAP32[$11>>2] = 0;
    (_SDL_SendWindowEvent($window,2,0,0)|0);
   }
   $16 = (_SDL_GetKeyboardFocus()|0);
   $17 = ($16|0)==($window|0);
   if ($17) {
    _SDL_SetKeyboardFocus(0);
   }
   $18 = (_SDL_GetMouseFocus()|0);
   $19 = ($18|0)==($window|0);
   if ($19) {
    _SDL_SetMouseFocus(0);
   }
   $20 = ((($window)) + 48|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = $21 & 2;
   $23 = ($22|0)==(0);
   if (!($23)) {
    $24 = HEAP32[300441]|0;
    $25 = ((($24)) + 628|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ($26|0)==($window|0);
    if ($27) {
     (_SDL_GL_MakeCurrent($window,0)|0);
    }
   }
   $28 = ((($window)) + 104|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = ($29|0)==(0|0);
   if (!($30)) {
    $31 = HEAP32[$29>>2]|0;
    $32 = $31 & -5;
    HEAP32[$29>>2] = $32;
    _SDL_FreeSurface($29);
   }
   $33 = HEAP32[300441]|0;
   $34 = ((($33)) + 116|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = ($35|0)==(0|0);
   if ($36) {
    $38 = $33;
   } else {
    FUNCTION_TABLE_vii[$35 & 8191]($33,$window);
    $$pre = HEAP32[300441]|0;
    $38 = $$pre;
   }
   $37 = ((($38)) + 104|0);
   $39 = HEAP32[$37>>2]|0;
   $40 = ($39|0)==(0|0);
   if (!($40)) {
    FUNCTION_TABLE_vii[$39 & 8191]($38,$window);
   }
   $41 = HEAP32[$20>>2]|0;
   $42 = $41 & 2;
   $43 = ($42|0)==(0);
   do {
    if (!($43)) {
     $44 = HEAP32[300441]|0;
     $45 = ($44|0)==(0|0);
     if ($45) {
      (_SDL_SetError(595917,$vararg_buffer3)|0);
      break;
     }
     $46 = ((($44)) + 364|0);
     $47 = HEAP32[$46>>2]|0;
     $48 = ($47|0)>(0);
     if ($48) {
      $49 = (($47) + -1)|0;
      HEAP32[$46>>2] = $49;
      $50 = ($47|0)>(1);
      if (!($50)) {
       $51 = ((($44)) + 148|0);
       $52 = HEAP32[$51>>2]|0;
       $53 = ($52|0)==(0|0);
       if (!($53)) {
        FUNCTION_TABLE_vi[$52 & 8191]($44);
       }
      }
     }
    }
   } while(0);
   $54 = (_SDL_GetWindowDisplayIndex($window)|0);
   $55 = ($54|0)>(-1);
   if ($55) {
    $56 = HEAP32[300441]|0;
    $57 = ((($56)) + 244|0);
    $58 = HEAP32[$57>>2]|0;
    $59 = (($58) + (($54*68)|0)|0);
    $$0$i = $59;
   } else {
    $$0$i = 0;
   }
   $60 = ((($$0$i)) + 56|0);
   $61 = HEAP32[$60>>2]|0;
   $62 = ($61|0)==($window|0);
   if ($62) {
    HEAP32[$60>>2] = 0;
   }
   HEAP32[$window>>2] = 0;
   $63 = ((($window)) + 8|0);
   $64 = HEAP32[$63>>2]|0;
   _SDL_free($64);
   $65 = ((($window)) + 12|0);
   $66 = HEAP32[$65>>2]|0;
   _SDL_FreeSurface($66);
   $67 = ((($window)) + 96|0);
   $68 = HEAP32[$67>>2]|0;
   _SDL_free($68);
   $69 = ((($window)) + 132|0);
   $70 = HEAP32[$69>>2]|0;
   $71 = ($70|0)==(0|0);
   if (!($71)) {
    $73 = $70;
    while(1) {
     $72 = ((($73)) + 8|0);
     $74 = HEAP32[$72>>2]|0;
     HEAP32[$69>>2] = $74;
     $75 = HEAP32[$73>>2]|0;
     _SDL_free($75);
     _SDL_free($73);
     $76 = HEAP32[$69>>2]|0;
     $77 = ($76|0)==(0|0);
     if ($77) {
      break;
     } else {
      $73 = $76;
     }
    }
   }
   $78 = ((($window)) + 144|0);
   $79 = HEAP32[$78>>2]|0;
   $80 = ($79|0)==(0|0);
   $81 = $79;
   $$pre1 = ((($window)) + 140|0);
   if (!($80)) {
    $82 = HEAP32[$$pre1>>2]|0;
    $83 = ((($79)) + 140|0);
    HEAP32[$83>>2] = $82;
   }
   $84 = HEAP32[$$pre1>>2]|0;
   $85 = ($84|0)==(0|0);
   if ($85) {
    $87 = HEAP32[300441]|0;
    $88 = ((($87)) + 248|0);
    HEAP32[$88>>2] = $81;
   } else {
    $86 = ((($84)) + 144|0);
    HEAP32[$86>>2] = $81;
   }
   _SDL_free($window);
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 STACKTOP = sp;return;
}
function _SDL_KeyboardInit() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 (_SDL_memcpy((1203480),138900,2048)|0);
 return 0;
}
function _SDL_MouseInit() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(1201880)>>2] = 1;
 return 0;
}
function _SDL_TouchInit() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _SDL_strncasecmp($str1,$str2,$maxlen) {
 $str1 = $str1|0;
 $str2 = $str2|0;
 $maxlen = $maxlen|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strncasecmp($str1,$str2,$maxlen)|0);
 return ($0|0);
}
function _SDL_TLSCreate() {
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_AtomicAdd(1205940,1)|0);
 $1 = (($0) + 1)|0;
 return ($1|0);
}
function _SDL_AtomicAdd($a,$v) {
 $a = $a|0;
 $v = $v|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$a>>2]|0;HEAP32[$a>>2] = (($0+$v)|0);
 return ($0|0);
}
function _SDL_CreateWindowTexture($unused,$window,$format,$pixels,$pitch) {
 $unused = $unused|0;
 $window = $window|0;
 $format = $format|0;
 $pixels = $pixels|0;
 $pitch = $pitch|0;
 var $$36 = 0, $$lcssa = 0, $$lcssa59 = 0, $$mask = 0, $$mask9 = 0, $$off17 = 0, $$off19 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $data$0$i = 0, $data$02$i = 0, $data$03$i = 0, $data$03$i$lcssa = 0, $data$1$in = 0, $i$039 = 0, $i$039$lcssa = 0, $i$137 = 0, $i5$036 = 0, $info = 0, $info1 = 0, $info4 = 0, $or$cond = 0;
 var $or$cond16 = 0, $phitmp = 0, $renderer$626 = 0, $switch = 0, $switch$split2D = 0, $switch18 = 0, $switch20 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 272|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $info = sp + 188|0;
 $info1 = sp + 104|0;
 $info4 = sp + 20|0;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 L1: do {
  if ($1) {
   (_SDL_SetError(595917,$vararg_buffer)|0);
   label = 11;
  } else {
   $2 = ($window|0)==(0|0);
   if (!($2)) {
    $3 = HEAP32[$window>>2]|0;
    $4 = ((($0)) + 256|0);
    $5 = ($3|0)==($4|0);
    if ($5) {
     $6 = ((($window)) + 132|0);
     $data$02$i = HEAP32[$6>>2]|0;
     $7 = ($data$02$i|0)==(0|0);
     if ($7) {
      label = 11;
      break;
     } else {
      $data$03$i = $data$02$i;
     }
     while(1) {
      $8 = HEAP32[$data$03$i>>2]|0;
      $9 = ($8|0)==(0|0);
      if (!($9)) {
       $10 = (_SDL_strcmp($8,609391)|0);
       $11 = ($10|0)==(0);
       if ($11) {
        $data$03$i$lcssa = $data$03$i;
        break;
       }
      }
      $12 = ((($data$03$i)) + 8|0);
      $data$0$i = HEAP32[$12>>2]|0;
      $13 = ($data$0$i|0)==(0|0);
      if ($13) {
       label = 11;
       break L1;
      } else {
       $data$03$i = $data$0$i;
      }
     }
     $14 = ((($data$03$i$lcssa)) + 4|0);
     $15 = HEAP32[$14>>2]|0;
     $16 = ($15|0)==(0|0);
     if ($16) {
      label = 11;
      break;
     } else {
      $data$1$in = $15;
      break;
     }
    }
   }
   (_SDL_SetError(609414,$vararg_buffer1)|0);
   label = 11;
  }
 } while(0);
 do {
  if ((label|0) == 11) {
   $17 = (_SDL_GetHint(609362)|0);
   $18 = ($17|0)==(0|0);
   L16: do {
    if ($18) {
     label = 19;
    } else {
     $19 = HEAP8[$17>>0]|0;
     $20 = $19 & -2;
     $switch = ($20<<24>>24)==(48);
     if ($switch) {
      label = 19;
     } else {
      $21 = (_SDL_strcasecmp($17,609353)|0);
      $22 = ($21|0)==(0);
      if ($22) {
       label = 19;
      } else {
       $23 = (_SDL_GetNumRenderDrivers()|0);
       $24 = ($23|0)>(0);
       if ($24) {
        $i$039 = 0;
        while(1) {
         (_SDL_GetRenderDriverInfo($i$039,$info)|0);
         $25 = HEAP32[$info>>2]|0;
         $26 = (_SDL_strcasecmp($25,$17)|0);
         $27 = ($26|0)==(0);
         if ($27) {
          $i$039$lcssa = $i$039;
          break;
         }
         $28 = (($i$039) + 1)|0;
         $29 = (_SDL_GetNumRenderDrivers()|0);
         $30 = ($28|0)<($29|0);
         if ($30) {
          $i$039 = $28;
         } else {
          label = 19;
          break L16;
         }
        }
        $31 = (_SDL_CreateRenderer($window,$i$039$lcssa,0)|0);
        $32 = ($31|0)==(0|0);
        if ($32) {
         label = 19;
        } else {
         $renderer$626 = $31;
        }
       } else {
        label = 19;
       }
      }
     }
    }
   } while(0);
   L25: do {
    if ((label|0) == 19) {
     $33 = (_SDL_GetNumRenderDrivers()|0);
     $34 = ($33|0)>(0);
     L27: do {
      if ($34) {
       $i$137 = 0;
       while(1) {
        (_SDL_GetRenderDriverInfo($i$137,$info1)|0);
        $35 = HEAP32[$info1>>2]|0;
        $36 = (_SDL_strcmp($35,609353)|0);
        $37 = ($36|0)==(0);
        if (!($37)) {
         $38 = (_SDL_CreateRenderer($window,$i$137,0)|0);
         $39 = ($38|0)==(0|0);
         if (!($39)) {
          $$lcssa59 = $38;
          break;
         }
        }
        $40 = (($i$137) + 1)|0;
        $41 = (_SDL_GetNumRenderDrivers()|0);
        $42 = ($40|0)<($41|0);
        if ($42) {
         $i$137 = $40;
        } else {
         break L27;
        }
       }
       $renderer$626 = $$lcssa59;
       break L25;
      }
     } while(0);
     $43 = (_SDL_SetError(609720,$vararg_buffer3)|0);
     $$36 = $43;
     STACKTOP = sp;return ($$36|0);
    }
   } while(0);
   $44 = (_SDL_calloc(1,20)|0);
   $45 = ($44|0)==(0|0);
   if (!($45)) {
    (_SDL_SetWindowData($window,609391,$44)|0);
    HEAP32[$44>>2] = $renderer$626;
    $data$1$in = $44;
    break;
   }
   _SDL_DestroyRenderer($renderer$626);
   $46 = (_SDL_Error(0)|0);
   $$36 = $46;
   STACKTOP = sp;return ($$36|0);
  }
 } while(0);
 $47 = ((($data$1$in)) + 4|0);
 $48 = HEAP32[$47>>2]|0;
 $49 = ($48|0)==(0|0);
 if (!($49)) {
  _SDL_DestroyTexture($48);
  HEAP32[$47>>2] = 0;
 }
 $50 = ((($data$1$in)) + 8|0);
 $51 = HEAP32[$50>>2]|0;
 _SDL_free($51);
 HEAP32[$50>>2] = 0;
 $52 = HEAP32[$data$1$in>>2]|0;
 $53 = (_SDL_GetRendererInfo($52,$info4)|0);
 $54 = ($53|0)<(0);
 if ($54) {
  $$36 = -1;
  STACKTOP = sp;return ($$36|0);
 }
 $55 = ((($info4)) + 12|0);
 $56 = HEAP32[$55>>2]|0;
 HEAP32[$format>>2] = $56;
 $57 = ((($info4)) + 8|0);
 $58 = HEAP32[$57>>2]|0;
 $59 = ($58|0)==(0);
 L50: do {
  if ($59) {
   $76 = $56;
  } else {
   $61 = $56;$i5$036 = 0;
   L51: while(1) {
    $60 = ($61|0)==(0);
    $$mask9 = $61 & -268435456;
    $62 = ($$mask9|0)==(268435456);
    $or$cond = $60 | $62;
    L53: do {
     if ($or$cond) {
      $63 = $61 >>> 24;
      $64 = $63 & 15;
      $$off17 = (($64) + -4)|0;
      $switch18 = ($$off17>>>0)<(3);
      if ($switch18) {
       $65 = $61 >>> 20;
       $66 = $65 & 15;
       switch ($66|0) {
       case 8: case 7: case 4: case 3:  {
        break L53;
        break;
       }
       default: {
       }
       }
      }
      $$off19 = (($64) + -7)|0;
      $switch20 = ($$off19>>>0)<(5);
      if (!($switch20)) {
       $$lcssa = $61;
       break L51;
      }
      $67 = $61 >>> 20;
      $68 = $67 & 15;
      switch ($68|0) {
      case 5: case 6: case 2: case 3:  {
       break;
      }
      default: {
       $$lcssa = $61;
       break L51;
      }
      }
     }
    } while(0);
    $69 = (($i5$036) + 1)|0;
    $70 = ($69>>>0)<($58>>>0);
    if (!($70)) {
     $76 = $56;
     break L50;
    }
    $$phi$trans$insert = (((($info4)) + 12|0) + ($69<<2)|0);
    $$pre = HEAP32[$$phi$trans$insert>>2]|0;
    $61 = $$pre;$i5$036 = $69;
   }
   HEAP32[$format>>2] = $$lcssa;
   $76 = $$lcssa;
  }
 } while(0);
 $71 = HEAP32[$data$1$in>>2]|0;
 $72 = ((($window)) + 24|0);
 $73 = HEAP32[$72>>2]|0;
 $74 = ((($window)) + 28|0);
 $75 = HEAP32[$74>>2]|0;
 $77 = (_SDL_CreateTexture($71,$76,1,$73,$75)|0);
 HEAP32[$47>>2] = $77;
 $78 = ($77|0)==(0|0);
 if ($78) {
  $$36 = -1;
  STACKTOP = sp;return ($$36|0);
 }
 $79 = HEAP32[$format>>2]|0;
 $80 = ($79|0)==(0);
 $$mask = $79 & -268435456;
 $81 = ($$mask|0)==(268435456);
 $or$cond16 = $80 | $81;
 L66: do {
  if ($or$cond16) {
   $83 = $79 & 255;
   $85 = $83;
  } else {
   $switch$split2D = ($79|0)<(1498831189);
   if ($switch$split2D) {
    switch ($79|0) {
    case 844715353:  {
     $85 = 2;
     break L66;
     break;
    }
    default: {
    }
    }
   } else {
    switch ($79|0) {
    case 1498831189:  {
     $85 = 2;
     break L66;
     break;
    }
    default: {
    }
    }
   }
   $82 = ($79|0)==(1431918169);
   $phitmp = $82 ? 2 : 1;
   $85 = $phitmp;
  }
 } while(0);
 $84 = ((($data$1$in)) + 16|0);
 HEAP32[$84>>2] = $85;
 $86 = HEAP32[$72>>2]|0;
 $87 = Math_imul($86, $85)|0;
 $88 = (($87) + 3)|0;
 $89 = $88 & -4;
 $90 = ((($data$1$in)) + 12|0);
 HEAP32[$90>>2] = $89;
 $91 = HEAP32[$74>>2]|0;
 $92 = Math_imul($89, $91)|0;
 $93 = (_SDL_malloc($92)|0);
 HEAP32[$50>>2] = $93;
 $94 = ($93|0)==(0|0);
 if ($94) {
  $95 = (_SDL_Error(0)|0);
  $$36 = $95;
  STACKTOP = sp;return ($$36|0);
 } else {
  HEAP32[$pixels>>2] = $93;
  $96 = HEAP32[$90>>2]|0;
  HEAP32[$pitch>>2] = $96;
  $97 = HEAP32[$data$1$in>>2]|0;
  (_SDL_RenderSetViewport($97,0)|0);
  $$36 = 0;
  STACKTOP = sp;return ($$36|0);
 }
 return (0)|0;
}
function _SDL_GetNumRenderDrivers() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 2;
}
function _SDL_GetRenderDriverInfo($index,$info) {
 $index = $index|0;
 $info = $info|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($index>>>0)>(1);
 if ($0) {
  HEAP32[$vararg_buffer>>2] = 1;
  $1 = (_SDL_SetError(612320,$vararg_buffer)|0);
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 } else {
  $2 = (140960 + ($index<<2)|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($3)) + 4|0);
  dest=$info; src=$4; stop=dest+84|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_CreateRenderer($window,$index,$flags) {
 $window = $window|0;
 $index = $index|0;
 $flags = $flags|0;
 var $$ = 0, $$0 = 0, $$05 = 0, $$lcssa18 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $not$ = 0, $renderer$712 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($window|0)==(0|0);
 if ($0) {
  (_SDL_SetError(609414,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = (_SDL_GetWindowData($window,609429)|0);
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  (_SDL_SetError(609451,$vararg_buffer1)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = (_SDL_GetHint(609491)|0);
 $4 = ($3|0)==(0|0);
 do {
  if ($4) {
   $$05 = $flags;
  } else {
   $5 = HEAP8[$3>>0]|0;
   $6 = ($5<<24>>24)==(48);
   if ($6) {
    $7 = $flags & -5;
    $$05 = $7;
    break;
   } else {
    $8 = $flags | 4;
    $$05 = $8;
    break;
   }
  }
 } while(0);
 $9 = ($index|0)<(0);
 L15: do {
  if ($9) {
   $10 = (_SDL_GetHint(609508)|0);
   $11 = ($10|0)==(0|0);
   do {
    if (!($11)) {
     $12 = HEAP32[(125760)>>2]|0;
     $13 = (_SDL_strcasecmp($10,$12)|0);
     $14 = ($13|0)==(0);
     if ($14) {
      $$lcssa18 = 125756;
     } else {
      $15 = HEAP32[(137840)>>2]|0;
      $16 = (_SDL_strcasecmp($10,$15)|0);
      $17 = ($16|0)==(0);
      if ($17) {
       $$lcssa18 = 137836;
      } else {
       break;
      }
     }
     $18 = HEAP32[$$lcssa18>>2]|0;
     $19 = (FUNCTION_TABLE_iii[$18 & 8191]($window,$$05)|0);
     $20 = ($19|0)==(0|0);
     if (!($20)) {
      $renderer$712 = $19;
      break L15;
     }
    }
   } while(0);
   $21 = HEAP32[(125764)>>2]|0;
   $22 = $21 & $$05;
   $23 = ($22|0)==($$05|0);
   if ($23) {
    $24 = HEAP32[31439]|0;
    $25 = (FUNCTION_TABLE_iii[$24 & 8191]($window,$$05)|0);
    $26 = ($25|0)==(0|0);
    if (!($26)) {
     $renderer$712 = $25;
     break;
    }
   }
   $27 = HEAP32[(137844)>>2]|0;
   $28 = $27 & $$05;
   $29 = ($28|0)==($$05|0);
   if ($29) {
    $44 = HEAP32[34459]|0;
    $45 = (FUNCTION_TABLE_iii[$44 & 8191]($window,$$05)|0);
    $46 = ($45|0)==(0|0);
    if (!($46)) {
     $renderer$712 = $45;
     break;
    }
   }
   (_SDL_SetError(609526,$vararg_buffer3)|0);
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  } else {
   $30 = ($index|0)>(1);
   if ($30) {
    HEAP32[$vararg_buffer5>>2] = 1;
    (_SDL_SetError(609563,$vararg_buffer5)|0);
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   } else {
    $31 = (140960 + ($index<<2)|0);
    $32 = HEAP32[$31>>2]|0;
    $33 = HEAP32[$32>>2]|0;
    $34 = (FUNCTION_TABLE_iii[$33 & 8191]($window,$$05)|0);
    $35 = ($34|0)==(0|0);
    if ($35) {
     $$0 = 0;
    } else {
     $renderer$712 = $34;
     break;
    }
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 HEAP32[$renderer$712>>2] = 1279299;
 $36 = ((($renderer$712)) + 188|0);
 HEAP32[$36>>2] = $window;
 $37 = ((($renderer$712)) + 284|0);
 HEAPF32[$37>>2] = 1.0;
 $38 = ((($renderer$712)) + 288|0);
 HEAPF32[$38>>2] = 1.0;
 $39 = (_SDL_GetWindowFlags($window)|0);
 $40 = $39 & 72;
 $41 = ((($renderer$712)) + 192|0);
 $not$ = ($40|0)!=(0);
 $$ = $not$&1;
 HEAP32[$41>>2] = $$;
 (_SDL_SetWindowData($window,609429,$renderer$712)|0);
 (_SDL_RenderSetViewport($renderer$712,0)|0);
 _SDL_AddEventWatch(4650,$renderer$712);
 $42 = ((($renderer$712)) + 104|0);
 $43 = HEAP32[$42>>2]|0;
 HEAP32[$vararg_buffer7>>2] = $43;
 _SDL_LogInfo(6,609699,$vararg_buffer7);
 $$0 = $renderer$712;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_RenderSetViewport($renderer,$rect) {
 $renderer = $renderer|0;
 $rect = $rect|0;
 var $$0 = 0, $$0$i = 0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0.0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0.0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($renderer|0)==(0|0);
 if (!($0)) {
  $1 = HEAP32[$renderer>>2]|0;
  $2 = ($1|0)==(1279299|0);
  if ($2) {
   $3 = ($rect|0)==(0|0);
   L4: do {
    if ($3) {
     $41 = ((($renderer)) + 212|0);
     HEAP32[$41>>2] = 0;
     $42 = ((($renderer)) + 216|0);
     HEAP32[$42>>2] = 0;
     $43 = ((($renderer)) + 220|0);
     $44 = ((($renderer)) + 224|0);
     $45 = ((($renderer)) + 304|0);
     $46 = HEAP32[$45>>2]|0;
     $47 = ($46|0)==(0|0);
     if (!($47)) {
      $48 = HEAP32[$46>>2]|0;
      $49 = ($48|0)==(1279300|0);
      if ($49) {
       $50 = ((($46)) + 12|0);
       $51 = HEAP32[$50>>2]|0;
       HEAP32[$43>>2] = $51;
       $52 = ((($46)) + 16|0);
       $53 = HEAP32[$52>>2]|0;
       HEAP32[$44>>2] = $53;
       break;
      }
      (_SDL_SetError(597654,$vararg_buffer1)|0);
      $$0 = -1;
      STACKTOP = sp;return ($$0|0);
     }
     $54 = ((($renderer)) + 8|0);
     $55 = HEAP32[$54>>2]|0;
     $56 = ($55|0)==(0|0);
     do {
      if ($56) {
       $58 = ((($renderer)) + 188|0);
       $59 = HEAP32[$58>>2]|0;
       $60 = ($59|0)==(0|0);
       if ($60) {
        $61 = (_SDL_SetError(597670,$vararg_buffer3)|0);
        $$0$i = $61;
        break;
       } else {
        _SDL_GetWindowSize($59,$43,$44);
        break L4;
       }
      } else {
       $57 = (FUNCTION_TABLE_iiii[$55 & 8191]($renderer,$43,$44)|0);
       $$0$i = $57;
      }
     } while(0);
     $62 = ($$0$i|0)<(0);
     if ($62) {
      $$0 = -1;
      STACKTOP = sp;return ($$0|0);
     }
    } else {
     $4 = HEAP32[$rect>>2]|0;
     $5 = (+($4|0));
     $6 = ((($renderer)) + 284|0);
     $7 = +HEAPF32[$6>>2];
     $8 = $5 * $7;
     $9 = $8;
     $10 = (+_SDL_floor($9));
     $11 = (~~(($10)));
     $12 = ((($renderer)) + 212|0);
     HEAP32[$12>>2] = $11;
     $13 = ((($rect)) + 4|0);
     $14 = HEAP32[$13>>2]|0;
     $15 = (+($14|0));
     $16 = ((($renderer)) + 288|0);
     $17 = +HEAPF32[$16>>2];
     $18 = $15 * $17;
     $19 = $18;
     $20 = (+_SDL_floor($19));
     $21 = (~~(($20)));
     $22 = ((($renderer)) + 216|0);
     HEAP32[$22>>2] = $21;
     $23 = ((($rect)) + 8|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = (+($24|0));
     $26 = +HEAPF32[$6>>2];
     $27 = $25 * $26;
     $28 = $27;
     $29 = (+_SDL_ceil($28));
     $30 = (~~(($29)));
     $31 = ((($renderer)) + 220|0);
     HEAP32[$31>>2] = $30;
     $32 = ((($rect)) + 12|0);
     $33 = HEAP32[$32>>2]|0;
     $34 = (+($33|0));
     $35 = +HEAPF32[$16>>2];
     $36 = $34 * $35;
     $37 = $36;
     $38 = (+_SDL_ceil($37));
     $39 = (~~(($38)));
     $40 = ((($renderer)) + 224|0);
     HEAP32[$40>>2] = $39;
    }
   } while(0);
   $63 = ((($renderer)) + 48|0);
   $64 = HEAP32[$63>>2]|0;
   $65 = (FUNCTION_TABLE_ii[$64 & 8191]($renderer)|0);
   $$0 = $65;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_SDL_SetError(597637,$vararg_buffer)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_RendererEventWatch($userdata,$event) {
 $userdata = $userdata|0;
 $event = $event|0;
 var $$ = 0, $$1 = 0, $$2 = 0, $$3 = 0, $$lobit$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0.0, $107 = 0, $108 = 0.0, $109 = 0.0, $11 = 0, $110 = 0;
 var $111 = 0.0, $112 = 0, $113 = 0.0, $114 = 0.0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0, $122 = 0, $123 = 0, $124 = 0.0, $125 = 0.0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0.0, $132 = 0.0, $133 = 0, $134 = 0, $135 = 0, $136 = 0.0, $137 = 0.0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0.0, $16 = 0, $160 = 0, $161 = 0.0, $162 = 0.0, $163 = 0, $164 = 0.0, $165 = 0;
 var $166 = 0.0, $167 = 0.0, $168 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $h = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $w = sp + 4|0;
 $h = sp;
 $0 = HEAP32[$event>>2]|0;
 switch ($0|0) {
 case 512:  {
  $1 = ((($event)) + 8|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = (_SDL_GetWindowFromID($2)|0);
  $4 = ((($userdata)) + 188|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ($3|0)==($5|0);
  if (!($6)) {
   STACKTOP = sp;return 0;
  }
  $7 = ((($userdata)) + 4|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8|0)==(0|0);
  if (!($9)) {
   FUNCTION_TABLE_vii[$8 & 8191]($userdata,$event);
  }
  $10 = ((($event)) + 12|0);
  $11 = HEAP8[$10>>0]|0;
  switch ($11<<24>>24) {
  case 6:  {
   $12 = ((($userdata)) + 304|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = ($13|0)!=(0|0);
   do {
    if ($14) {
     $15 = ((($userdata)) + 44|0);
     $16 = HEAP32[$15>>2]|0;
     $17 = ($16|0)==(0|0);
     if (!($17)) {
      $18 = ((($userdata)) + 108|0);
      $19 = HEAP32[$18>>2]|0;
      $$lobit$i$i = $19 & 8;
      $20 = ($$lobit$i$i|0)==(0);
      if (!($20)) {
       HEAP32[$12>>2] = 0;
       $21 = (FUNCTION_TABLE_iii[$16 & 8191]($userdata,0)|0);
       $22 = ($21|0)<(0);
       if ($22) {
        break;
       }
       $23 = ((($userdata)) + 212|0);
       $24 = ((($userdata)) + 228|0);
       ;HEAP32[$23>>2]=HEAP32[$24>>2]|0;HEAP32[$23+4>>2]=HEAP32[$24+4>>2]|0;HEAP32[$23+8>>2]=HEAP32[$24+8>>2]|0;HEAP32[$23+12>>2]=HEAP32[$24+12>>2]|0;
       $25 = ((($userdata)) + 244|0);
       $26 = ((($userdata)) + 260|0);
       ;HEAP32[$25>>2]=HEAP32[$26>>2]|0;HEAP32[$25+4>>2]=HEAP32[$26+4>>2]|0;HEAP32[$25+8>>2]=HEAP32[$26+8>>2]|0;HEAP32[$25+12>>2]=HEAP32[$26+12>>2]|0;
       $27 = ((($userdata)) + 280|0);
       $28 = HEAP32[$27>>2]|0;
       $29 = ((($userdata)) + 276|0);
       HEAP32[$29>>2] = $28;
       $30 = ((($userdata)) + 284|0);
       $31 = ((($userdata)) + 292|0);
       $32 = $31;
       $33 = $32;
       $34 = HEAP32[$33>>2]|0;
       $35 = (($32) + 4)|0;
       $36 = $35;
       $37 = HEAP32[$36>>2]|0;
       $38 = $30;
       $39 = $38;
       HEAP32[$39>>2] = $34;
       $40 = (($38) + 4)|0;
       $41 = $40;
       HEAP32[$41>>2] = $37;
       $42 = ((($userdata)) + 204|0);
       $43 = HEAP32[$42>>2]|0;
       $44 = ((($userdata)) + 208|0);
       $45 = HEAP32[$44>>2]|0;
       $46 = ((($userdata)) + 196|0);
       HEAP32[$46>>2] = $43;
       $47 = ((($userdata)) + 200|0);
       HEAP32[$47>>2] = $45;
       $48 = ((($userdata)) + 48|0);
       $49 = HEAP32[$48>>2]|0;
       $50 = (FUNCTION_TABLE_ii[$49 & 8191]($userdata)|0);
       $51 = ($50|0)<(0);
       if ($51) {
        break;
       }
       $52 = ((($userdata)) + 52|0);
       $53 = HEAP32[$52>>2]|0;
       (FUNCTION_TABLE_ii[$53 & 8191]($userdata)|0);
       break;
      }
     }
     (_SDL_Error(4)|0);
    }
   } while(0);
   $54 = ((($userdata)) + 196|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = ($55|0)==(0);
   if ($56) {
    $57 = ((($userdata)) + 8|0);
    $58 = HEAP32[$57>>2]|0;
    $59 = ($58|0)==(0|0);
    if ($59) {
     $60 = HEAP32[$4>>2]|0;
     _SDL_GetWindowSize($60,$w,$h);
    } else {
     (FUNCTION_TABLE_iiii[$58 & 8191]($userdata,$w,$h)|0);
    }
    $61 = HEAP32[$12>>2]|0;
    $62 = ($61|0)==(0|0);
    if ($62) {
     $69 = ((($userdata)) + 212|0);
     HEAP32[$69>>2] = 0;
     $70 = ((($userdata)) + 216|0);
     HEAP32[$70>>2] = 0;
     $71 = HEAP32[$w>>2]|0;
     $72 = ((($userdata)) + 220|0);
     HEAP32[$72>>2] = $71;
     $73 = HEAP32[$h>>2]|0;
     $74 = ((($userdata)) + 224|0);
     HEAP32[$74>>2] = $73;
     $75 = ((($userdata)) + 48|0);
     $76 = HEAP32[$75>>2]|0;
     (FUNCTION_TABLE_ii[$76 & 8191]($userdata)|0);
    } else {
     $63 = ((($userdata)) + 228|0);
     HEAP32[$63>>2] = 0;
     $64 = ((($userdata)) + 232|0);
     HEAP32[$64>>2] = 0;
     $65 = HEAP32[$w>>2]|0;
     $66 = ((($userdata)) + 236|0);
     HEAP32[$66>>2] = $65;
     $67 = HEAP32[$h>>2]|0;
     $68 = ((($userdata)) + 240|0);
     HEAP32[$68>>2] = $67;
    }
   } else {
    (_UpdateLogicalSize($userdata)|0);
   }
   if (!($14)) {
    STACKTOP = sp;return 0;
   }
   (_SDL_SetRenderTarget($userdata,$13)|0);
   STACKTOP = sp;return 0;
   break;
  }
  case 2:  {
   $77 = ((($userdata)) + 192|0);
   HEAP32[$77>>2] = 1;
   STACKTOP = sp;return 0;
   break;
  }
  case 1:  {
   $78 = (_SDL_GetWindowFlags($3)|0);
   $79 = $78 & 64;
   $80 = ($79|0)==(0);
   if (!($80)) {
    STACKTOP = sp;return 0;
   }
   $81 = ((($userdata)) + 192|0);
   HEAP32[$81>>2] = 0;
   STACKTOP = sp;return 0;
   break;
  }
  case 7:  {
   $82 = ((($userdata)) + 192|0);
   HEAP32[$82>>2] = 1;
   STACKTOP = sp;return 0;
   break;
  }
  case 8: case 9:  {
   $83 = (_SDL_GetWindowFlags($3)|0);
   $84 = $83 & 8;
   $85 = ($84|0)==(0);
   if (!($85)) {
    STACKTOP = sp;return 0;
   }
   $86 = ((($userdata)) + 192|0);
   HEAP32[$86>>2] = 0;
   STACKTOP = sp;return 0;
   break;
  }
  default: {
   STACKTOP = sp;return 0;
  }
  }
  break;
 }
 case 1024:  {
  $87 = ((($event)) + 8|0);
  $88 = HEAP32[$87>>2]|0;
  $89 = (_SDL_GetWindowFromID($88)|0);
  $90 = ((($userdata)) + 196|0);
  $91 = HEAP32[$90>>2]|0;
  $92 = ($91|0)==(0);
  if ($92) {
   STACKTOP = sp;return 0;
  }
  $93 = ((($userdata)) + 188|0);
  $94 = HEAP32[$93>>2]|0;
  $95 = ($89|0)==($94|0);
  if (!($95)) {
   STACKTOP = sp;return 0;
  }
  $96 = ((($userdata)) + 212|0);
  $97 = HEAP32[$96>>2]|0;
  $98 = ((($event)) + 20|0);
  $99 = HEAP32[$98>>2]|0;
  $100 = (($99) - ($97))|0;
  $101 = ((($userdata)) + 216|0);
  $102 = HEAP32[$101>>2]|0;
  $103 = ((($event)) + 24|0);
  $104 = HEAP32[$103>>2]|0;
  $105 = (($104) - ($102))|0;
  $106 = (+($100|0));
  $107 = ((($userdata)) + 284|0);
  $108 = +HEAPF32[$107>>2];
  $109 = $106 / $108;
  $110 = (~~(($109)));
  HEAP32[$98>>2] = $110;
  $111 = (+($105|0));
  $112 = ((($userdata)) + 288|0);
  $113 = +HEAPF32[$112>>2];
  $114 = $111 / $113;
  $115 = (~~(($114)));
  HEAP32[$103>>2] = $115;
  $116 = ((($event)) + 28|0);
  $117 = HEAP32[$116>>2]|0;
  $118 = ($117|0)>(0);
  if ($118) {
   $119 = (+($117|0));
   $120 = $119 / $108;
   $121 = (~~(($120)));
   $122 = ($121|0)<(1);
   $$ = $122 ? 1 : $121;
   HEAP32[$116>>2] = $$;
  } else {
   $123 = ($117|0)<(0);
   if ($123) {
    $124 = (+($117|0));
    $125 = $124 / $108;
    $126 = (~~(($125)));
    $127 = ($126|0)>(-1);
    $$1 = $127 ? -1 : $126;
    HEAP32[$116>>2] = $$1;
   }
  }
  $128 = ((($event)) + 32|0);
  $129 = HEAP32[$128>>2]|0;
  $130 = ($129|0)>(0);
  if ($130) {
   $131 = (+($129|0));
   $132 = $131 / $113;
   $133 = (~~(($132)));
   $134 = ($133|0)<(1);
   $$2 = $134 ? 1 : $133;
   HEAP32[$128>>2] = $$2;
   STACKTOP = sp;return 0;
  }
  $135 = ($129|0)<(0);
  if (!($135)) {
   STACKTOP = sp;return 0;
  }
  $136 = (+($129|0));
  $137 = $136 / $113;
  $138 = (~~(($137)));
  $139 = ($138|0)>(-1);
  $$3 = $139 ? -1 : $138;
  HEAP32[$128>>2] = $$3;
  STACKTOP = sp;return 0;
  break;
 }
 case 1026: case 1025:  {
  $140 = ((($event)) + 8|0);
  $141 = HEAP32[$140>>2]|0;
  $142 = (_SDL_GetWindowFromID($141)|0);
  $143 = ((($userdata)) + 196|0);
  $144 = HEAP32[$143>>2]|0;
  $145 = ($144|0)==(0);
  if ($145) {
   STACKTOP = sp;return 0;
  }
  $146 = ((($userdata)) + 188|0);
  $147 = HEAP32[$146>>2]|0;
  $148 = ($142|0)==($147|0);
  if (!($148)) {
   STACKTOP = sp;return 0;
  }
  $149 = ((($userdata)) + 212|0);
  $150 = HEAP32[$149>>2]|0;
  $151 = ((($event)) + 20|0);
  $152 = HEAP32[$151>>2]|0;
  $153 = (($152) - ($150))|0;
  $154 = ((($userdata)) + 216|0);
  $155 = HEAP32[$154>>2]|0;
  $156 = ((($event)) + 24|0);
  $157 = HEAP32[$156>>2]|0;
  $158 = (($157) - ($155))|0;
  $159 = (+($153|0));
  $160 = ((($userdata)) + 284|0);
  $161 = +HEAPF32[$160>>2];
  $162 = $159 / $161;
  $163 = (~~(($162)));
  HEAP32[$151>>2] = $163;
  $164 = (+($158|0));
  $165 = ((($userdata)) + 288|0);
  $166 = +HEAPF32[$165>>2];
  $167 = $164 / $166;
  $168 = (~~(($167)));
  HEAP32[$156>>2] = $168;
  STACKTOP = sp;return 0;
  break;
 }
 default: {
  STACKTOP = sp;return 0;
 }
 }
 return (0)|0;
}
function _SDL_GetWindowFromID($id) {
 $id = $id|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $window$0 = 0, $window$01 = 0, $window$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  $$0 = 0;
  return ($$0|0);
 }
 $2 = ((($0)) + 248|0);
 $window$01 = HEAP32[$2>>2]|0;
 $3 = ($window$01|0)==(0|0);
 if ($3) {
  $$0 = 0;
  return ($$0|0);
 } else {
  $window$02 = $window$01;
 }
 while(1) {
  $4 = ((($window$02)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ($5|0)==($id|0);
  if ($6) {
   $$0 = $window$02;
   label = 5;
   break;
  }
  $7 = ((($window$02)) + 144|0);
  $window$0 = HEAP32[$7>>2]|0;
  $8 = ($window$0|0)==(0|0);
  if ($8) {
   $$0 = 0;
   label = 5;
   break;
  } else {
   $window$02 = $window$0;
  }
 }
 if ((label|0) == 5) {
  return ($$0|0);
 }
 return (0)|0;
}
function _UpdateLogicalSize($renderer) {
 $renderer = $renderer|0;
 var $$0 = 0, $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0.0, $24 = 0;
 var $25 = 0, $26 = 0.0, $27 = 0.0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0, $42 = 0.0;
 var $43 = 0, $44 = 0.0, $45 = 0.0, $46 = 0, $47 = 0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0, $53 = 0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0.0, $66 = 0, $67 = 0.0, $68 = 0.0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0.0, $73 = 0.0, $74 = 0.0, $75 = 0.0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $9 = 0, $h = 0, $scale$0 = 0.0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, $viewport = 0, $w = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $w = sp + 60|0;
 $h = sp + 56|0;
 $viewport = sp + 40|0;
 HEAP32[$w>>2] = 1;
 HEAP32[$h>>2] = 1;
 $0 = ($renderer|0)==(0|0);
 if (!($0)) {
  $1 = HEAP32[$renderer>>2]|0;
  $2 = ($1|0)==(1279299|0);
  if ($2) {
   $3 = ((($renderer)) + 304|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==(0|0);
   L4: do {
    if ($5) {
     $12 = ((($renderer)) + 8|0);
     $13 = HEAP32[$12>>2]|0;
     $14 = ($13|0)==(0|0);
     do {
      if ($14) {
       $16 = ((($renderer)) + 188|0);
       $17 = HEAP32[$16>>2]|0;
       $18 = ($17|0)==(0|0);
       if ($18) {
        $19 = (_SDL_SetError(597670,$vararg_buffer3)|0);
        $$0$i = $19;
        break;
       } else {
        _SDL_GetWindowSize($17,$w,$h);
        break L4;
       }
      } else {
       $15 = (FUNCTION_TABLE_iiii[$13 & 8191]($renderer,$w,$h)|0);
       $$0$i = $15;
      }
     } while(0);
     $20 = ($$0$i|0)<(0);
     if ($20) {
      $$0 = -1;
      STACKTOP = sp;return ($$0|0);
     }
    } else {
     $6 = HEAP32[$4>>2]|0;
     $7 = ($6|0)==(1279300|0);
     if ($7) {
      $8 = ((($4)) + 12|0);
      $9 = HEAP32[$8>>2]|0;
      HEAP32[$w>>2] = $9;
      $10 = ((($4)) + 16|0);
      $11 = HEAP32[$10>>2]|0;
      HEAP32[$h>>2] = $11;
      break;
     }
     (_SDL_SetError(597654,$vararg_buffer1)|0);
     $$0 = -1;
     STACKTOP = sp;return ($$0|0);
    }
   } while(0);
   $21 = ((($renderer)) + 196|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = (+($22|0));
   $24 = ((($renderer)) + 200|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = (+($25|0));
   $27 = $23 / $26;
   $28 = HEAP32[$w>>2]|0;
   $29 = (+($28|0));
   $30 = HEAP32[$h>>2]|0;
   $31 = (+($30|0));
   $32 = $29 / $31;
   $33 = HEAP32[$renderer>>2]|0;
   $34 = ($33|0)==(1279299|0);
   if ($34) {
    $35 = ((($renderer)) + 284|0);
    HEAPF32[$35>>2] = 1.0;
    $36 = ((($renderer)) + 288|0);
    HEAPF32[$36>>2] = 1.0;
   } else {
    (_SDL_SetError(597637,$vararg_buffer5)|0);
   }
   $37 = $27 - $32;
   $38 = $37;
   $39 = (+_SDL_fabs($38));
   $40 = $39 < 1.0E-4;
   do {
    if ($40) {
     $41 = HEAP32[$w>>2]|0;
     $42 = (+($41|0));
     $43 = HEAP32[$21>>2]|0;
     $44 = (+($43|0));
     $45 = $42 / $44;
     (_SDL_RenderSetViewport($renderer,0)|0);
     $scale$0 = $45;
    } else {
     $46 = $27 > $32;
     if ($46) {
      $47 = HEAP32[$w>>2]|0;
      $48 = (+($47|0));
      $49 = HEAP32[$21>>2]|0;
      $50 = (+($49|0));
      $51 = $48 / $50;
      HEAP32[$viewport>>2] = 0;
      $52 = ((($viewport)) + 8|0);
      HEAP32[$52>>2] = $47;
      $53 = HEAP32[$24>>2]|0;
      $54 = (+($53|0));
      $55 = $51 * $54;
      $56 = $55;
      $57 = (+_SDL_ceil($56));
      $58 = (~~(($57)));
      $59 = ((($viewport)) + 12|0);
      HEAP32[$59>>2] = $58;
      $60 = HEAP32[$h>>2]|0;
      $61 = (($60) - ($58))|0;
      $62 = (($61|0) / 2)&-1;
      $63 = ((($viewport)) + 4|0);
      HEAP32[$63>>2] = $62;
      (_SDL_RenderSetViewport($renderer,$viewport)|0);
      $scale$0 = $51;
      break;
     } else {
      $64 = HEAP32[$h>>2]|0;
      $65 = (+($64|0));
      $66 = HEAP32[$24>>2]|0;
      $67 = (+($66|0));
      $68 = $65 / $67;
      $69 = ((($viewport)) + 4|0);
      HEAP32[$69>>2] = 0;
      $70 = ((($viewport)) + 12|0);
      HEAP32[$70>>2] = $64;
      $71 = HEAP32[$21>>2]|0;
      $72 = (+($71|0));
      $73 = $68 * $72;
      $74 = $73;
      $75 = (+_SDL_ceil($74));
      $76 = (~~(($75)));
      $77 = ((($viewport)) + 8|0);
      HEAP32[$77>>2] = $76;
      $78 = HEAP32[$w>>2]|0;
      $79 = (($78) - ($76))|0;
      $80 = (($79|0) / 2)&-1;
      HEAP32[$viewport>>2] = $80;
      (_SDL_RenderSetViewport($renderer,$viewport)|0);
      $scale$0 = $68;
      break;
     }
    }
   } while(0);
   $81 = HEAP32[$renderer>>2]|0;
   $82 = ($81|0)==(1279299|0);
   if ($82) {
    $83 = ((($renderer)) + 284|0);
    HEAPF32[$83>>2] = $scale$0;
    $84 = ((($renderer)) + 288|0);
    HEAPF32[$84>>2] = $scale$0;
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   } else {
    (_SDL_SetError(597637,$vararg_buffer7)|0);
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
  }
 }
 (_SDL_SetError(597637,$vararg_buffer)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SetRenderTarget($renderer,$texture) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 var $$01 = 0, $$02 = 0, $$lobit = 0, $$lobit$i = 0, $$sink = 0, $$sink4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $texture$ = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($renderer|0)==(0|0);
 if (!($0)) {
  $1 = ((($renderer)) + 44|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ($2|0)==(0|0);
  if (!($3)) {
   $4 = ((($renderer)) + 108|0);
   $5 = HEAP32[$4>>2]|0;
   $$lobit$i = $5 & 8;
   $6 = ($$lobit$i|0)==(0);
   if (!($6)) {
    $8 = ((($renderer)) + 304|0);
    $9 = HEAP32[$8>>2]|0;
    $10 = ($9|0)==($texture|0);
    if ($10) {
     $$01 = 0;
     STACKTOP = sp;return ($$01|0);
    }
    $11 = ($texture|0)==(0|0);
    do {
     if ($11) {
      $$02 = 0;$95 = 0;
     } else {
      $12 = HEAP32[$texture>>2]|0;
      $13 = ($12|0)==(1279300|0);
      if (!($13)) {
       (_SDL_SetError(597654,$vararg_buffer)|0);
       $$01 = -1;
       STACKTOP = sp;return ($$01|0);
      }
      $14 = ((($texture)) + 32|0);
      $15 = HEAP32[$14>>2]|0;
      $16 = ($15|0)==($renderer|0);
      if (!($16)) {
       $17 = (_SDL_SetError(609606,$vararg_buffer1)|0);
       $$01 = $17;
       STACKTOP = sp;return ($$01|0);
      }
      $18 = ((($texture)) + 8|0);
      $19 = HEAP32[$18>>2]|0;
      $20 = ($19|0)==(2);
      if ($20) {
       $22 = ((($texture)) + 36|0);
       $23 = HEAP32[$22>>2]|0;
       $24 = ($23|0)==(0|0);
       $texture$ = $24 ? $texture : $23;
       $25 = ($texture$|0)!=(0|0);
       $26 = ($9|0)==(0|0);
       if (!($26)) {
        $$02 = $texture$;$95 = $25;
        break;
       }
       $27 = ((($renderer)) + 228|0);
       $28 = ((($renderer)) + 212|0);
       ;HEAP32[$27>>2]=HEAP32[$28>>2]|0;HEAP32[$27+4>>2]=HEAP32[$28+4>>2]|0;HEAP32[$27+8>>2]=HEAP32[$28+8>>2]|0;HEAP32[$27+12>>2]=HEAP32[$28+12>>2]|0;
       $29 = ((($renderer)) + 260|0);
       $30 = ((($renderer)) + 244|0);
       ;HEAP32[$29>>2]=HEAP32[$30>>2]|0;HEAP32[$29+4>>2]=HEAP32[$30+4>>2]|0;HEAP32[$29+8>>2]=HEAP32[$30+8>>2]|0;HEAP32[$29+12>>2]=HEAP32[$30+12>>2]|0;
       $31 = ((($renderer)) + 276|0);
       $32 = HEAP32[$31>>2]|0;
       $33 = ((($renderer)) + 280|0);
       HEAP32[$33>>2] = $32;
       $34 = ((($renderer)) + 292|0);
       $35 = ((($renderer)) + 284|0);
       $36 = $35;
       $37 = $36;
       $38 = HEAP32[$37>>2]|0;
       $39 = (($36) + 4)|0;
       $40 = $39;
       $41 = HEAP32[$40>>2]|0;
       $42 = $34;
       $43 = $42;
       HEAP32[$43>>2] = $38;
       $44 = (($42) + 4)|0;
       $45 = $44;
       HEAP32[$45>>2] = $41;
       $46 = ((($renderer)) + 196|0);
       $47 = HEAP32[$46>>2]|0;
       $48 = ((($renderer)) + 204|0);
       HEAP32[$48>>2] = $47;
       $49 = ((($renderer)) + 200|0);
       $50 = HEAP32[$49>>2]|0;
       $51 = ((($renderer)) + 208|0);
       HEAP32[$51>>2] = $50;
       $$02 = $texture$;$95 = $25;
       break;
      } else {
       $21 = (_SDL_SetError(609649,$vararg_buffer3)|0);
       $$01 = $21;
       STACKTOP = sp;return ($$01|0);
      }
     }
    } while(0);
    HEAP32[$8>>2] = $$02;
    $52 = (FUNCTION_TABLE_iii[$2 & 8191]($renderer,$$02)|0);
    $53 = ($52|0)<(0);
    if ($53) {
     $$01 = -1;
     STACKTOP = sp;return ($$01|0);
    }
    $54 = ((($renderer)) + 212|0);
    if ($95) {
     HEAP32[$54>>2] = 0;
     $55 = ((($renderer)) + 216|0);
     HEAP32[$55>>2] = 0;
     $56 = ((($$02)) + 12|0);
     $57 = HEAP32[$56>>2]|0;
     $58 = ((($renderer)) + 220|0);
     HEAP32[$58>>2] = $57;
     $59 = ((($$02)) + 16|0);
     $60 = HEAP32[$59>>2]|0;
     $61 = ((($renderer)) + 224|0);
     HEAP32[$61>>2] = $60;
     $62 = ((($renderer)) + 284|0);
     HEAPF32[$62>>2] = 1.0;
     $63 = ((($renderer)) + 288|0);
     HEAPF32[$63>>2] = 1.0;
     $$sink = $60;$$sink4 = $57;
    } else {
     $64 = ((($renderer)) + 228|0);
     ;HEAP32[$54>>2]=HEAP32[$64>>2]|0;HEAP32[$54+4>>2]=HEAP32[$64+4>>2]|0;HEAP32[$54+8>>2]=HEAP32[$64+8>>2]|0;HEAP32[$54+12>>2]=HEAP32[$64+12>>2]|0;
     $65 = ((($renderer)) + 244|0);
     $66 = ((($renderer)) + 260|0);
     ;HEAP32[$65>>2]=HEAP32[$66>>2]|0;HEAP32[$65+4>>2]=HEAP32[$66+4>>2]|0;HEAP32[$65+8>>2]=HEAP32[$66+8>>2]|0;HEAP32[$65+12>>2]=HEAP32[$66+12>>2]|0;
     $67 = ((($renderer)) + 280|0);
     $68 = HEAP32[$67>>2]|0;
     $69 = ((($renderer)) + 276|0);
     HEAP32[$69>>2] = $68;
     $70 = ((($renderer)) + 284|0);
     $71 = ((($renderer)) + 292|0);
     $72 = $71;
     $73 = $72;
     $74 = HEAP32[$73>>2]|0;
     $75 = (($72) + 4)|0;
     $76 = $75;
     $77 = HEAP32[$76>>2]|0;
     $78 = $70;
     $79 = $78;
     HEAP32[$79>>2] = $74;
     $80 = (($78) + 4)|0;
     $81 = $80;
     HEAP32[$81>>2] = $77;
     $82 = ((($renderer)) + 204|0);
     $83 = HEAP32[$82>>2]|0;
     $84 = ((($renderer)) + 208|0);
     $85 = HEAP32[$84>>2]|0;
     $$sink = $85;$$sink4 = $83;
    }
    $86 = ((($renderer)) + 196|0);
    HEAP32[$86>>2] = $$sink4;
    $87 = ((($renderer)) + 200|0);
    HEAP32[$87>>2] = $$sink;
    $88 = ((($renderer)) + 48|0);
    $89 = HEAP32[$88>>2]|0;
    $90 = (FUNCTION_TABLE_ii[$89 & 8191]($renderer)|0);
    $91 = ($90|0)<(0);
    if ($91) {
     $$01 = -1;
     STACKTOP = sp;return ($$01|0);
    }
    $92 = ((($renderer)) + 52|0);
    $93 = HEAP32[$92>>2]|0;
    $94 = (FUNCTION_TABLE_ii[$93 & 8191]($renderer)|0);
    $$lobit = $94 >> 31;
    $$01 = $$lobit;
    STACKTOP = sp;return ($$01|0);
   }
  }
 }
 $7 = (_SDL_Error(4)|0);
 $$01 = $7;
 STACKTOP = sp;return ($$01|0);
}
function _SDL_AddEventWatch($filter,$userdata) {
 $filter = $filter|0;
 $userdata = $userdata|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $tail$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_malloc(12)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  return;
 }
 HEAP32[$0>>2] = $filter;
 $2 = ((($0)) + 4|0);
 HEAP32[$2>>2] = $userdata;
 $3 = ((($0)) + 8|0);
 HEAP32[$3>>2] = 0;
 $4 = HEAP32[300734]|0;
 $5 = ($4|0)==(0|0);
 if ($5) {
  HEAP32[300734] = $0;
  return;
 } else {
  $tail$0 = $4;
 }
 while(1) {
  $6 = ((($tail$0)) + 8|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)==(0|0);
  if ($8) {
   $$lcssa = $6;
   break;
  } else {
   $tail$0 = $7;
  }
 }
 HEAP32[$$lcssa>>2] = $0;
 return;
}
function _SDL_LogInfo($category,$fmt,$varargs) {
 $category = $category|0;
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $ap = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ap = sp;
 HEAP32[$ap>>2] = $varargs;
 _SDL_LogMessageV($category,3,$fmt,$ap);
 STACKTOP = sp;return;
}
function _SDL_DestroyRenderer($renderer) {
 $renderer = $renderer|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($renderer|0)==(0|0);
 if (!($0)) {
  $1 = HEAP32[$renderer>>2]|0;
  $2 = ($1|0)==(1279299|0);
  if ($2) {
   _SDL_DelEventWatch(4650,$renderer);
   $3 = ((($renderer)) + 300|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==(0|0);
   if (!($5)) {
    $6 = $4;
    while(1) {
     _SDL_DestroyTexture($6);
     $7 = HEAP32[$3>>2]|0;
     $8 = ($7|0)==(0|0);
     if ($8) {
      break;
     } else {
      $6 = $7;
     }
    }
   }
   $9 = ((($renderer)) + 188|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($10|0)==(0|0);
   if (!($11)) {
    (_SDL_SetWindowData($10,609429,0)|0);
   }
   HEAP32[$renderer>>2] = 0;
   $12 = ((($renderer)) + 92|0);
   $13 = HEAP32[$12>>2]|0;
   FUNCTION_TABLE_vi[$13 & 8191]($renderer);
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(597637,$vararg_buffer)|0);
 STACKTOP = sp;return;
}
function _SDL_DestroyTexture($texture) {
 $texture = $texture|0;
 var $$lobit$i$i = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($texture|0)==(0|0);
 if (!($0)) {
  $1 = HEAP32[$texture>>2]|0;
  $2 = ($1|0)==(1279300|0);
  if ($2) {
   $3 = ((($texture)) + 32|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ((($4)) + 304|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($6|0)==($texture|0);
   do {
    if ($7) {
     $8 = ($4|0)==(0|0);
     if (!($8)) {
      $9 = ((($4)) + 44|0);
      $10 = HEAP32[$9>>2]|0;
      $11 = ($10|0)==(0|0);
      if (!($11)) {
       $12 = ((($4)) + 108|0);
       $13 = HEAP32[$12>>2]|0;
       $$lobit$i$i = $13 & 8;
       $14 = ($$lobit$i$i|0)==(0);
       if (!($14)) {
        HEAP32[$5>>2] = 0;
        $15 = (FUNCTION_TABLE_iii[$10 & 8191]($4,0)|0);
        $16 = ($15|0)<(0);
        if ($16) {
         break;
        }
        $17 = ((($4)) + 212|0);
        $18 = ((($4)) + 228|0);
        ;HEAP32[$17>>2]=HEAP32[$18>>2]|0;HEAP32[$17+4>>2]=HEAP32[$18+4>>2]|0;HEAP32[$17+8>>2]=HEAP32[$18+8>>2]|0;HEAP32[$17+12>>2]=HEAP32[$18+12>>2]|0;
        $19 = ((($4)) + 244|0);
        $20 = ((($4)) + 260|0);
        ;HEAP32[$19>>2]=HEAP32[$20>>2]|0;HEAP32[$19+4>>2]=HEAP32[$20+4>>2]|0;HEAP32[$19+8>>2]=HEAP32[$20+8>>2]|0;HEAP32[$19+12>>2]=HEAP32[$20+12>>2]|0;
        $21 = ((($4)) + 280|0);
        $22 = HEAP32[$21>>2]|0;
        $23 = ((($4)) + 276|0);
        HEAP32[$23>>2] = $22;
        $24 = ((($4)) + 284|0);
        $25 = ((($4)) + 292|0);
        $26 = $25;
        $27 = $26;
        $28 = HEAP32[$27>>2]|0;
        $29 = (($26) + 4)|0;
        $30 = $29;
        $31 = HEAP32[$30>>2]|0;
        $32 = $24;
        $33 = $32;
        HEAP32[$33>>2] = $28;
        $34 = (($32) + 4)|0;
        $35 = $34;
        HEAP32[$35>>2] = $31;
        $36 = ((($4)) + 204|0);
        $37 = HEAP32[$36>>2]|0;
        $38 = ((($4)) + 208|0);
        $39 = HEAP32[$38>>2]|0;
        $40 = ((($4)) + 196|0);
        HEAP32[$40>>2] = $37;
        $41 = ((($4)) + 200|0);
        HEAP32[$41>>2] = $39;
        $42 = ((($4)) + 48|0);
        $43 = HEAP32[$42>>2]|0;
        $44 = (FUNCTION_TABLE_ii[$43 & 8191]($4)|0);
        $45 = ($44|0)<(0);
        if ($45) {
         break;
        }
        $46 = ((($4)) + 52|0);
        $47 = HEAP32[$46>>2]|0;
        (FUNCTION_TABLE_ii[$47 & 8191]($4)|0);
        break;
       }
      }
     }
     (_SDL_Error(4)|0);
    }
   } while(0);
   HEAP32[$texture>>2] = 0;
   $48 = ((($texture)) + 76|0);
   $49 = HEAP32[$48>>2]|0;
   $50 = ($49|0)==(0|0);
   $51 = $49;
   $$pre = ((($texture)) + 72|0);
   if (!($50)) {
    $52 = HEAP32[$$pre>>2]|0;
    $53 = ((($49)) + 72|0);
    HEAP32[$53>>2] = $52;
   }
   $54 = HEAP32[$$pre>>2]|0;
   $55 = ($54|0)==(0|0);
   if ($55) {
    $57 = ((($4)) + 300|0);
    HEAP32[$57>>2] = $51;
   } else {
    $56 = ((($54)) + 76|0);
    HEAP32[$56>>2] = $51;
   }
   $58 = ((($texture)) + 36|0);
   $59 = HEAP32[$58>>2]|0;
   $60 = ($59|0)==(0|0);
   if (!($60)) {
    _SDL_DestroyTexture($59);
   }
   $61 = ((($texture)) + 40|0);
   $62 = HEAP32[$61>>2]|0;
   $63 = ($62|0)==(0|0);
   if (!($63)) {
    _SDL_SW_DestroyYUVTexture($62);
   }
   $64 = ((($texture)) + 44|0);
   $65 = HEAP32[$64>>2]|0;
   _SDL_free($65);
   $66 = ((($4)) + 88|0);
   $67 = HEAP32[$66>>2]|0;
   FUNCTION_TABLE_vii[$67 & 8191]($4,$texture);
   _SDL_free($texture);
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(597654,$vararg_buffer)|0);
 STACKTOP = sp;return;
}
function _SDL_SW_DestroyYUVTexture($swdata) {
 $swdata = $swdata|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($swdata|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($swdata)) + 16|0);
 $2 = HEAP32[$1>>2]|0;
 _SDL_free($2);
 $3 = ((($swdata)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 _SDL_free($4);
 $5 = ((($swdata)) + 24|0);
 $6 = HEAP32[$5>>2]|0;
 _SDL_free($6);
 $7 = ((($swdata)) + 56|0);
 $8 = HEAP32[$7>>2]|0;
 _SDL_FreeSurface($8);
 $9 = ((($swdata)) + 60|0);
 $10 = HEAP32[$9>>2]|0;
 _SDL_FreeSurface($10);
 _SDL_free($swdata);
 return;
}
function _SDL_GetRendererInfo($renderer,$info) {
 $renderer = $renderer|0;
 $info = $info|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $vararg_buffer = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($renderer|0)==(0|0);
 if (!($0)) {
  $1 = HEAP32[$renderer>>2]|0;
  $2 = ($1|0)==(1279299|0);
  if ($2) {
   $3 = ((($renderer)) + 104|0);
   dest=$info; src=$3; stop=dest+84|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_SDL_SetError(597637,$vararg_buffer)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_CreateTexture($renderer,$format,$access,$w,$h) {
 $renderer = $renderer|0;
 $format = $format|0;
 $access = $access|0;
 $w = $w|0;
 $h = $h|0;
 var $$0 = 0, $$06 = 0, $$1$i = 0, $$clear = 0, $$clear10 = 0, $$clear11 = 0, $$mask3$i = 0, $$mask3$us$i = 0, $$mask7 = 0, $$mask9 = 0, $$not = 0, $$off = 0, $$off$i = 0, $$off21$i = 0, $$off25$i = 0, $$off25$us$i = 0, $$off34$i = 0, $$off34$us$i = 0, $$phi$trans$insert = 0, $$pre = 0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$01$i = 0, $i$046$i = 0, $i$145$i = 0, $i$145$us$i = 0, $or$cond11 = 0, $or$cond14 = 0, $or$cond15 = 0, $or$cond15$i = 0, $or$cond15$us$i = 0, $or$cond17 = 0, $or$cond3 = 0, $phitmp = 0, $switch = 0;
 var $switch$cast$i = 0, $switch$cast$i$clear = 0, $switch$cast38$i = 0, $switch$cast38$i$clear = 0, $switch$cast38$us$i = 0, $switch$cast38$us$i$clear = 0, $switch$downshift$i = 0, $switch$downshift40$i = 0, $switch$downshift40$us$i = 0, $switch$i = 0, $switch$lobit$i = 0, $switch$lobit33$i = 0, $switch$lobit33$us$i = 0, $switch$masked$i = 0, $switch$masked41$i = 0, $switch$maskindex$i = 0, $switch$maskindex31$i = 0, $switch$maskindex31$us$i = 0, $switch$shifted$i = 0, $switch$shifted32$i = 0;
 var $switch$shifted32$us$i = 0, $switch$split2D = 0, $switch$tableidx$i = 0, $switch$tableidx24$i = 0, $switch$tableidx28$i = 0, $switch$tableidx28$us$i = 0, $switch$tableidx37$i = 0, $switch$tableidx37$us$i = 0, $switch22$i = 0, $switch26$i = 0, $switch26$us$i = 0, $switch35$i = 0, $switch35$us$i = 0, $tmp$us$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, $vararg_ptr9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($renderer|0)==(0|0);
 if (!($0)) {
  $1 = HEAP32[$renderer>>2]|0;
  $2 = ($1|0)==(1279299|0);
  if ($2) {
   $3 = ($format|0)==(0);
   if ($3) {
    $4 = ((($renderer)) + 116|0);
    $5 = HEAP32[$4>>2]|0;
    $$06 = $5;
   } else {
    $$06 = $format;
   }
   $$not = ($$06|0)==(0);
   $$mask9 = $$06 & -268435456;
   $6 = ($$mask9|0)==(268435456);
   $or$cond11 = $$not | $6;
   if ($or$cond11) {
    $7 = $$06 & 255;
    $8 = ($7|0)==(0);
    if ($8) {
     (_SDL_SetError(609764,$vararg_buffer1)|0);
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
    $9 = $$06 >>> 24;
    $10 = $9 & 15;
    $$off = (($10) + -1)|0;
    $switch = ($$off>>>0)<(3);
    if ($switch) {
     (_SDL_SetError(609787,$vararg_buffer3)|0);
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
   }
   $11 = ($w|0)<(1);
   $12 = ($h|0)<(1);
   $or$cond3 = $11 | $12;
   if ($or$cond3) {
    (_SDL_SetError(609825,$vararg_buffer5)|0);
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   $13 = ((($renderer)) + 180|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ($14|0)!=(0);
   $16 = ($14|0)<($w|0);
   $or$cond14 = $15 & $16;
   $$phi$trans$insert = ((($renderer)) + 184|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0;
   if (!($or$cond14)) {
    $17 = ($$pre|0)!=(0);
    $18 = ($$pre|0)<($h|0);
    $or$cond15 = $17 & $18;
    if (!($or$cond15)) {
     $19 = (_SDL_calloc(1,80)|0);
     $20 = ($19|0)==(0|0);
     if ($20) {
      (_SDL_Error(0)|0);
      $$0 = 0;
      STACKTOP = sp;return ($$0|0);
     }
     HEAP32[$19>>2] = 1279300;
     $21 = ((($19)) + 4|0);
     HEAP32[$21>>2] = $$06;
     $22 = ((($19)) + 8|0);
     HEAP32[$22>>2] = $access;
     $23 = ((($19)) + 12|0);
     HEAP32[$23>>2] = $w;
     $24 = ((($19)) + 16|0);
     HEAP32[$24>>2] = $h;
     $25 = ((($19)) + 28|0);
     $26 = ((($19)) + 32|0);
     HEAP32[$25>>2] = -1;
     HEAP32[$26>>2] = $renderer;
     $27 = ((($renderer)) + 300|0);
     $28 = HEAP32[$27>>2]|0;
     $29 = ((($19)) + 76|0);
     HEAP32[$29>>2] = $28;
     $30 = ($28|0)==(0);
     if (!($30)) {
      $31 = $28;
      $32 = ((($31)) + 72|0);
      HEAP32[$32>>2] = $19;
     }
     HEAP32[$27>>2] = $19;
     $33 = ((($renderer)) + 112|0);
     $34 = HEAP32[$33>>2]|0;
     $35 = ($34|0)==(0);
     L30: do {
      if (!($35)) {
       $i$01$i = 0;
       while(1) {
        $38 = (((($renderer)) + 116|0) + ($i$01$i<<2)|0);
        $39 = HEAP32[$38>>2]|0;
        $40 = ($39|0)==($$06|0);
        $37 = (($i$01$i) + 1)|0;
        if ($40) {
         break;
        }
        $36 = ($37>>>0)<($34>>>0);
        if ($36) {
         $i$01$i = $37;
        } else {
         break L30;
        }
       }
       $41 = ((($renderer)) + 12|0);
       $42 = HEAP32[$41>>2]|0;
       $43 = (FUNCTION_TABLE_iii[$42 & 8191]($renderer,$19)|0);
       $44 = ($43|0)<(0);
       if (!($44)) {
        $$0 = $19;
        STACKTOP = sp;return ($$0|0);
       }
       _SDL_DestroyTexture($19);
       $$0 = 0;
       STACKTOP = sp;return ($$0|0);
      }
     } while(0);
     L40: do {
      if ($or$cond11) {
       $50 = $$06 >>> 24;
       $51 = $50 & 15;
       $$off$i = (($51) + -4)|0;
       $switch$i = ($$off$i>>>0)<(3);
       if ($switch$i) {
        $52 = $$06 >>> 20;
        $53 = $52 & 15;
        $switch$tableidx$i = (($53) + -3)|0;
        $54 = ($switch$tableidx$i>>>0)<(6);
        if ($54) {
         $switch$maskindex$i = $switch$tableidx$i&255;
         $switch$shifted$i = (51&255) >>> $switch$maskindex$i;
         $58 = $switch$shifted$i & 1;
         $switch$lobit$i = ($58<<24>>24)==(0);
         if ($switch$lobit$i) {
          label = 31;
         } else {
          $126 = 1;
         }
        } else {
         label = 31;
        }
       } else {
        label = 31;
       }
       if ((label|0) == 31) {
        $$off21$i = (($51) + -7)|0;
        $switch22$i = ($$off21$i>>>0)<(5);
        if ($switch22$i) {
         $55 = $$06 >>> 20;
         $56 = $55 & 15;
         $switch$tableidx24$i = (($56) + -2)|0;
         $57 = ($switch$tableidx24$i>>>0)<(5);
         if ($57) {
          $switch$cast$i = $switch$tableidx24$i&255;
          $switch$cast$i$clear = $switch$cast$i & 31;
          $switch$downshift$i = (27&255) >>> $switch$cast$i$clear;
          $59 = $switch$downshift$i & 1;
          $$clear = $59 & 31;
          $switch$masked$i = ($$clear<<24>>24)!=(0);
          $126 = $switch$masked$i;
         } else {
          $126 = 0;
         }
        } else {
         $126 = 0;
        }
       }
       if ($35) {
        label = 53;
       } else {
        if ($126) {
         $i$145$us$i = 0;
         while(1) {
          $60 = (((($renderer)) + 116|0) + ($i$145$us$i<<2)|0);
          $61 = HEAP32[$60>>2]|0;
          $62 = ($61|0)==(0);
          $$mask3$us$i = $61 & -268435456;
          $63 = ($$mask3$us$i|0)==(268435456);
          $or$cond15$us$i = $62 | $63;
          do {
           if ($or$cond15$us$i) {
            $64 = $61 >>> 24;
            $65 = $64 & 15;
            $$off25$us$i = (($65) + -4)|0;
            $switch26$us$i = ($$off25$us$i>>>0)<(3);
            do {
             if ($switch26$us$i) {
              $66 = $61 >>> 20;
              $67 = $66 & 15;
              $switch$tableidx28$us$i = (($67) + -3)|0;
              $68 = ($switch$tableidx28$us$i>>>0)<(6);
              if (!($68)) {
               break;
              }
              $switch$maskindex31$us$i = $switch$tableidx28$us$i&255;
              $switch$shifted32$us$i = (51&255) >>> $switch$maskindex31$us$i;
              $69 = $switch$shifted32$us$i & 1;
              $switch$lobit33$us$i = ($69<<24>>24)==(0);
              if (!($switch$lobit33$us$i)) {
               $$1$i = $61;
               break L40;
              }
             }
            } while(0);
            $$off34$us$i = (($65) + -7)|0;
            $switch35$us$i = ($$off34$us$i>>>0)<(5);
            if (!($switch35$us$i)) {
             break;
            }
            $70 = $61 >>> 20;
            $71 = $70 & 15;
            $switch$tableidx37$us$i = (($71) + -2)|0;
            $72 = ($switch$tableidx37$us$i>>>0)<(5);
            if (!($72)) {
             break;
            }
            $switch$cast38$us$i = $switch$tableidx37$us$i&255;
            $switch$cast38$us$i$clear = $switch$cast38$us$i & 31;
            $switch$downshift40$us$i = (27&255) >>> $switch$cast38$us$i$clear;
            $73 = $switch$downshift40$us$i & 1;
            $$clear10 = $73 & 31;
            $tmp$us$i = ($$clear10<<24>>24)==(0);
            if (!($tmp$us$i)) {
             $$1$i = $61;
             break L40;
            }
           }
          } while(0);
          $74 = (($i$145$us$i) + 1)|0;
          $75 = ($74>>>0)<($34>>>0);
          if ($75) {
           $i$145$us$i = $74;
          } else {
           label = 53;
           break;
          }
         }
        } else {
         $i$145$i = 0;
         while(1) {
          $76 = (((($renderer)) + 116|0) + ($i$145$i<<2)|0);
          $77 = HEAP32[$76>>2]|0;
          $78 = ($77|0)==(0);
          $$mask3$i = $77 & -268435456;
          $79 = ($$mask3$i|0)==(268435456);
          $or$cond15$i = $78 | $79;
          L65: do {
           if ($or$cond15$i) {
            $80 = $77 >>> 24;
            $81 = $80 & 15;
            $$off25$i = (($81) + -4)|0;
            $switch26$i = ($$off25$i>>>0)<(3);
            do {
             if ($switch26$i) {
              $82 = $77 >>> 20;
              $83 = $82 & 15;
              $switch$tableidx28$i = (($83) + -3)|0;
              $84 = ($switch$tableidx28$i>>>0)<(6);
              if (!($84)) {
               break;
              }
              $switch$maskindex31$i = $switch$tableidx28$i&255;
              $switch$shifted32$i = (51&255) >>> $switch$maskindex31$i;
              $88 = $switch$shifted32$i & 1;
              $switch$lobit33$i = ($88<<24>>24)==(0);
              if (!($switch$lobit33$i)) {
               break L65;
              }
             }
            } while(0);
            $$off34$i = (($81) + -7)|0;
            $switch35$i = ($$off34$i>>>0)<(5);
            if (!($switch35$i)) {
             $$1$i = $77;
             break L40;
            }
            $85 = $77 >>> 20;
            $86 = $85 & 15;
            $switch$tableidx37$i = (($86) + -2)|0;
            $87 = ($switch$tableidx37$i>>>0)<(5);
            if (!($87)) {
             $$1$i = $77;
             break L40;
            }
            $switch$cast38$i = $switch$tableidx37$i&255;
            $switch$cast38$i$clear = $switch$cast38$i & 31;
            $switch$downshift40$i = (27&255) >>> $switch$cast38$i$clear;
            $89 = $switch$downshift40$i & 1;
            $$clear11 = $89 & 31;
            $switch$masked41$i = ($$clear11<<24>>24)==(0);
            if ($switch$masked41$i) {
             $$1$i = $77;
             break L40;
            }
           }
          } while(0);
          $90 = (($i$145$i) + 1)|0;
          $91 = ($90>>>0)<($34>>>0);
          if ($91) {
           $i$145$i = $90;
          } else {
           label = 53;
           break;
          }
         }
        }
       }
      } else {
       if ($35) {
        label = 53;
       } else {
        $i$046$i = 0;
        while(1) {
         $47 = (((($renderer)) + 116|0) + ($i$046$i<<2)|0);
         $48 = HEAP32[$47>>2]|0;
         $49 = ($48|0)==($$06|0);
         $46 = (($i$046$i) + 1)|0;
         if ($49) {
          $$1$i = $$06;
          break L40;
         }
         $45 = ($46>>>0)<($34>>>0);
         if ($45) {
          $i$046$i = $46;
         } else {
          label = 53;
          break;
         }
        }
       }
      }
     } while(0);
     if ((label|0) == 53) {
      $92 = ((($renderer)) + 116|0);
      $93 = HEAP32[$92>>2]|0;
      $$1$i = $93;
     }
     $94 = (_SDL_CreateTexture($renderer,$$1$i,$access,$w,$h)|0);
     $95 = ((($19)) + 36|0);
     HEAP32[$95>>2] = $94;
     $96 = ($94|0)==(0|0);
     $97 = $94;
     if ($96) {
      _SDL_DestroyTexture($19);
      $$0 = 0;
      STACKTOP = sp;return ($$0|0);
     }
     $98 = HEAP32[$29>>2]|0;
     $99 = ((($94)) + 76|0);
     HEAP32[$99>>2] = $98;
     $100 = ($98|0)==(0);
     if (!($100)) {
      $101 = $98;
      $102 = ((($101)) + 72|0);
      HEAP32[$102>>2] = $94;
     }
     $103 = ((($94)) + 72|0);
     $104 = HEAP32[$103>>2]|0;
     $105 = ((($19)) + 72|0);
     HEAP32[$105>>2] = $104;
     $106 = ($104|0)==(0|0);
     if (!($106)) {
      $107 = ((($104)) + 76|0);
      HEAP32[$107>>2] = $19;
     }
     HEAP32[$103>>2] = $19;
     HEAP32[$29>>2] = $97;
     HEAP32[$27>>2] = $19;
     $108 = HEAP32[$21>>2]|0;
     $109 = ($108|0)==(0);
     $$mask7 = $108 & -268435456;
     $110 = ($$mask7|0)==(268435456);
     $or$cond17 = $109 | $110;
     if (!($or$cond17)) {
      $111 = (_SDL_SW_CreateYUVTexture($$06,$w,$h)|0);
      $112 = ((($19)) + 40|0);
      HEAP32[$112>>2] = $111;
      $113 = ($111|0)==(0|0);
      if (!($113)) {
       $$0 = $19;
       STACKTOP = sp;return ($$0|0);
      }
      _SDL_DestroyTexture($19);
      $$0 = 0;
      STACKTOP = sp;return ($$0|0);
     }
     $114 = ($access|0)==(1);
     if (!($114)) {
      $$0 = $19;
      STACKTOP = sp;return ($$0|0);
     }
     L101: do {
      if ($or$cond11) {
       $116 = $$06 & 255;
       $118 = $116;
      } else {
       $switch$split2D = ($$06|0)<(1498831189);
       if ($switch$split2D) {
        switch ($$06|0) {
        case 844715353:  {
         $118 = 2;
         break L101;
         break;
        }
        default: {
        }
        }
       } else {
        switch ($$06|0) {
        case 1498831189:  {
         $118 = 2;
         break L101;
         break;
        }
        default: {
        }
        }
       }
       $115 = ($$06|0)==(1431918169);
       $phitmp = $115 ? 2 : 1;
       $118 = $phitmp;
      }
     } while(0);
     $117 = Math_imul($118, $w)|0;
     $119 = (($117) + 3)|0;
     $120 = $119 & -4;
     $121 = ((($19)) + 48|0);
     HEAP32[$121>>2] = $120;
     $122 = Math_imul($120, $h)|0;
     $123 = (_SDL_calloc(1,$122)|0);
     $124 = ((($19)) + 44|0);
     HEAP32[$124>>2] = $123;
     $125 = ($123|0)==(0|0);
     if (!($125)) {
      $$0 = $19;
      STACKTOP = sp;return ($$0|0);
     }
     _SDL_DestroyTexture($19);
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
   }
   HEAP32[$vararg_buffer7>>2] = $14;
   $vararg_ptr9 = ((($vararg_buffer7)) + 4|0);
   HEAP32[$vararg_ptr9>>2] = $$pre;
   (_SDL_SetError(609855,$vararg_buffer7)|0);
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_SDL_SetError(597637,$vararg_buffer)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SW_CreateYUVTexture($format,$w,$h) {
 $format = $format|0;
 $w = $w|0;
 $h = $h|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $div = 0, $exitcond = 0, $i$01 = 0, $or$cond = 0, $switch$split102D = 0, $switch$split12D = 0, $switch$split132D = 0, $switch$split162D = 0, $switch$split192D = 0, $switch$split2D = 0, $switch$split42D = 0, $switch$split72D = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $switch$split2D = ($format|0)<(1431918169);
 L1: do {
  if ($switch$split2D) {
   $switch$split42D = ($format|0)<(844715353);
   if ($switch$split42D) {
    switch ($format|0) {
    case 842094169:  {
     break L1;
     break;
    }
    default: {
     label = 2;
     break L1;
    }
    }
   } else {
    switch ($format|0) {
    case 844715353:  {
     break L1;
     break;
    }
    default: {
     label = 2;
     break L1;
    }
    }
   }
  } else {
   $switch$split72D = ($format|0)<(1448433993);
   if ($switch$split72D) {
    switch ($format|0) {
    case 1431918169:  {
     break L1;
     break;
    }
    default: {
     label = 2;
     break L1;
    }
    }
   }
   $switch$split162D = ($format|0)<(1498831189);
   if ($switch$split162D) {
    switch ($format|0) {
    case 1448433993:  {
     break L1;
     break;
    }
    default: {
     label = 2;
     break L1;
    }
    }
   } else {
    switch ($format|0) {
    case 1498831189:  {
     break L1;
     break;
    }
    default: {
     label = 2;
     break L1;
    }
    }
   }
  }
 } while(0);
 if ((label|0) == 2) {
  (_SDL_SetError(609895,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $0 = (_SDL_calloc(1,64)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_Error(0)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$0>>2] = $format;
 $2 = ((($0)) + 4|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($0)) + 8|0);
 HEAP32[$3>>2] = $w;
 $4 = ((($0)) + 12|0);
 HEAP32[$4>>2] = $h;
 $5 = $w << 1;
 $6 = Math_imul($5, $h)|0;
 $7 = (_SDL_malloc($6)|0);
 $8 = ((($0)) + 16|0);
 HEAP32[$8>>2] = $7;
 $9 = (_SDL_malloc(4096)|0);
 $10 = ((($0)) + 20|0);
 HEAP32[$10>>2] = $9;
 $11 = (_SDL_malloc(9216)|0);
 $12 = ((($0)) + 24|0);
 HEAP32[$12>>2] = $11;
 $13 = HEAP32[$8>>2]|0;
 $14 = ($13|0)==(0|0);
 $15 = $13;
 if (!($14)) {
  $16 = HEAP32[$10>>2]|0;
  $17 = ($16|0)==(0|0);
  $18 = ($11|0)==(0|0);
  $or$cond = $18 | $17;
  if (!($or$cond)) {
   $25 = ((($16)) + 1024|0);
   $26 = ((($16)) + 2048|0);
   $27 = ((($16)) + 3072|0);
   $i$01 = 0;
   while(1) {
    $28 = (($i$01) + -128)|0;
    $29 = (+($28|0));
    $30 = $29 * 1.4013377926421404;
    $31 = (~~(($30)));
    $32 = (($16) + ($i$01<<2)|0);
    HEAP32[$32>>2] = $31;
    $33 = $29 * -0.71360381861575184;
    $34 = (~~(($33)));
    $35 = (($25) + ($i$01<<2)|0);
    HEAP32[$35>>2] = $34;
    $36 = $29 * -0.34441087613293053;
    $37 = (~~(($36)));
    $38 = (($26) + ($i$01<<2)|0);
    HEAP32[$38>>2] = $37;
    $39 = $29 * 1.7734138972809665;
    $40 = (~~(($39)));
    $41 = (($27) + ($i$01<<2)|0);
    HEAP32[$41>>2] = $40;
    $42 = (($i$01) + 1)|0;
    $exitcond = ($42|0)==(256);
    if ($exitcond) {
     break;
    } else {
     $i$01 = $42;
    }
   }
   $switch$split12D = ($format|0)<(1431918169);
   L27: do {
    if ($switch$split12D) {
     $switch$split102D = ($format|0)<(844715353);
     if ($switch$split102D) {
      switch ($format|0) {
      case 842094169:  {
       label = 11;
       break L27;
       break;
      }
      default: {
       $$0 = $0;
      }
      }
      STACKTOP = sp;return ($$0|0);
     } else {
      switch ($format|0) {
      case 844715353:  {
       label = 12;
       break L27;
       break;
      }
      default: {
       $$0 = $0;
      }
      }
      STACKTOP = sp;return ($$0|0);
     }
    } else {
     $switch$split132D = ($format|0)<(1448433993);
     if ($switch$split132D) {
      switch ($format|0) {
      case 1431918169:  {
       label = 12;
       break L27;
       break;
      }
      default: {
       $$0 = $0;
      }
      }
      STACKTOP = sp;return ($$0|0);
     }
     $switch$split192D = ($format|0)<(1498831189);
     if ($switch$split192D) {
      switch ($format|0) {
      case 1448433993:  {
       label = 11;
       break L27;
       break;
      }
      default: {
       $$0 = $0;
      }
      }
      STACKTOP = sp;return ($$0|0);
     } else {
      switch ($format|0) {
      case 1498831189:  {
       label = 12;
       break L27;
       break;
      }
      default: {
       $$0 = $0;
      }
      }
      STACKTOP = sp;return ($$0|0);
     }
    }
   } while(0);
   if ((label|0) == 11) {
    $43 = $w&65535;
    $44 = ((($0)) + 36|0);
    HEAP16[$44>>1] = $43;
    $45 = $w >>> 1;
    $46 = $45 & 32767;
    $47 = $46&65535;
    $48 = ((($0)) + 38|0);
    HEAP16[$48>>1] = $47;
    $div = ($43&65535) >>> 1;
    $49 = ((($0)) + 40|0);
    HEAP16[$49>>1] = $div;
    $50 = ((($0)) + 44|0);
    HEAP32[$50>>2] = $13;
    $51 = $w & 65535;
    $52 = Math_imul($51, $h)|0;
    $53 = (($13) + ($52)|0);
    $54 = ((($0)) + 48|0);
    HEAP32[$54>>2] = $53;
    $55 = Math_imul($46, $h)|0;
    $56 = (($55|0) / 2)&-1;
    $57 = (($53) + ($56)|0);
    $58 = ((($0)) + 52|0);
    HEAP32[$58>>2] = $57;
    $$0 = $0;
    STACKTOP = sp;return ($$0|0);
   }
   else if ((label|0) == 12) {
    $59 = $5&65535;
    $60 = ((($0)) + 36|0);
    HEAP16[$60>>1] = $59;
    $61 = ((($0)) + 44|0);
    HEAP32[$61>>2] = $15;
    $$0 = $0;
    STACKTOP = sp;return ($$0|0);
   }
  }
 }
 _SDL_free($13);
 $19 = HEAP32[$10>>2]|0;
 _SDL_free($19);
 $20 = HEAP32[$12>>2]|0;
 _SDL_free($20);
 $21 = ((($0)) + 56|0);
 $22 = HEAP32[$21>>2]|0;
 _SDL_FreeSurface($22);
 $23 = ((($0)) + 60|0);
 $24 = HEAP32[$23>>2]|0;
 _SDL_FreeSurface($24);
 _SDL_free($0);
 (_SDL_Error(0)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_UpdateWindowTexture($unused,$window,$rects,$numrects) {
 $unused = $unused|0;
 $window = $window|0;
 $rects = $rects|0;
 $numrects = $numrects|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $data$0$i = 0, $data$02$i = 0, $data$03$i = 0, $data$03$i$lcssa = 0, $rect = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $rect = sp + 24|0;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 L1: do {
  if ($1) {
   (_SDL_SetError(595917,$vararg_buffer)|0);
  } else {
   $2 = ($window|0)==(0|0);
   if (!($2)) {
    $3 = HEAP32[$window>>2]|0;
    $4 = ((($0)) + 256|0);
    $5 = ($3|0)==($4|0);
    if ($5) {
     $6 = ((($window)) + 132|0);
     $data$02$i = HEAP32[$6>>2]|0;
     $7 = ($data$02$i|0)==(0|0);
     if ($7) {
      break;
     } else {
      $data$03$i = $data$02$i;
     }
     while(1) {
      $8 = HEAP32[$data$03$i>>2]|0;
      $9 = ($8|0)==(0|0);
      if (!($9)) {
       $10 = (_SDL_strcmp($8,609391)|0);
       $11 = ($10|0)==(0);
       if ($11) {
        $data$03$i$lcssa = $data$03$i;
        break;
       }
      }
      $12 = ((($data$03$i)) + 8|0);
      $data$0$i = HEAP32[$12>>2]|0;
      $13 = ($data$0$i|0)==(0|0);
      if ($13) {
       break L1;
      } else {
       $data$03$i = $data$0$i;
      }
     }
     $14 = ((($data$03$i$lcssa)) + 4|0);
     $15 = HEAP32[$14>>2]|0;
     $16 = ($15|0)==(0|0);
     if ($16) {
      break;
     }
     $17 = ((($15)) + 4|0);
     $18 = HEAP32[$17>>2]|0;
     $19 = ($18|0)==(0|0);
     if ($19) {
      break;
     }
     $21 = ((($window)) + 24|0);
     $22 = HEAP32[$21>>2]|0;
     $23 = ((($window)) + 28|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = (_SDL_GetSpanEnclosingRect($22,$24,$numrects,$rects,$rect)|0);
     $26 = ($25|0)==(0);
     if ($26) {
      $$0 = 0;
      STACKTOP = sp;return ($$0|0);
     }
     $27 = ((($15)) + 8|0);
     $28 = HEAP32[$27>>2]|0;
     $29 = ((($rect)) + 4|0);
     $30 = HEAP32[$29>>2]|0;
     $31 = ((($15)) + 12|0);
     $32 = HEAP32[$31>>2]|0;
     $33 = Math_imul($32, $30)|0;
     $34 = (($28) + ($33)|0);
     $35 = HEAP32[$rect>>2]|0;
     $36 = ((($15)) + 16|0);
     $37 = HEAP32[$36>>2]|0;
     $38 = Math_imul($37, $35)|0;
     $39 = (($34) + ($38)|0);
     $40 = HEAP32[$17>>2]|0;
     $41 = (_SDL_UpdateTexture($40,$rect,$39,$32)|0);
     $42 = ($41|0)<(0);
     if ($42) {
      $$0 = -1;
      STACKTOP = sp;return ($$0|0);
     }
     $43 = HEAP32[$15>>2]|0;
     $44 = HEAP32[$17>>2]|0;
     $45 = (_SDL_RenderCopy($43,$44,0,0)|0);
     $46 = ($45|0)<(0);
     if ($46) {
      $$0 = -1;
      STACKTOP = sp;return ($$0|0);
     }
     $47 = HEAP32[$15>>2]|0;
     _SDL_RenderPresent($47);
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
   }
   (_SDL_SetError(609414,$vararg_buffer1)|0);
  }
 } while(0);
 $20 = (_SDL_SetError(609918,$vararg_buffer3)|0);
 $$0 = $20;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GetSpanEnclosingRect($width,$height,$numrects,$rects,$span) {
 $width = $width|0;
 $height = $height|0;
 $numrects = $numrects|0;
 $rects = $rects|0;
 $span = $span|0;
 var $$0 = 0, $$span_y1$0 = 0, $$span_y2$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$03 = 0, $span_y1$01 = 0, $span_y1$1 = 0, $span_y1$1$lcssa = 0, $span_y2$02 = 0, $span_y2$1 = 0, $span_y2$1$lcssa = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer10 = 0, $vararg_buffer4 = 0, $vararg_buffer7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer10 = sp + 32|0;
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($width|0)<(1);
 if ($0) {
  HEAP32[$vararg_buffer>>2] = 609941;
  (_SDL_SetError(611267,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ($height|0)<(1);
 if ($1) {
  HEAP32[$vararg_buffer1>>2] = 609947;
  (_SDL_SetError(611267,$vararg_buffer1)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($rects|0)==(0|0);
 if ($2) {
  HEAP32[$vararg_buffer4>>2] = 609954;
  (_SDL_SetError(611267,$vararg_buffer4)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ($span|0)==(0|0);
 if ($3) {
  HEAP32[$vararg_buffer7>>2] = 609960;
  (_SDL_SetError(611267,$vararg_buffer7)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ($numrects|0)<(1);
 if ($4) {
  HEAP32[$vararg_buffer10>>2] = 609965;
  (_SDL_SetError(611267,$vararg_buffer10)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $i$03 = 0;$span_y1$01 = $height;$span_y2$02 = 0;
 }
 while(1) {
  $5 = (((($rects) + ($i$03<<4)|0)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (((($rects) + ($i$03<<4)|0)) + 12|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = (($8) + ($6))|0;
  $10 = ($6|0)<(0);
  $11 = ($6|0)<($span_y1$01|0);
  $$span_y1$0 = $11 ? $6 : $span_y1$01;
  $span_y1$1 = $10 ? 0 : $$span_y1$0;
  $12 = ($9|0)>($height|0);
  $13 = ($9|0)>($span_y2$02|0);
  $$span_y2$0 = $13 ? $9 : $span_y2$02;
  $span_y2$1 = $12 ? $height : $$span_y2$0;
  $14 = (($i$03) + 1)|0;
  $exitcond = ($14|0)==($numrects|0);
  if ($exitcond) {
   $span_y1$1$lcssa = $span_y1$1;$span_y2$1$lcssa = $span_y2$1;
   break;
  } else {
   $i$03 = $14;$span_y1$01 = $span_y1$1;$span_y2$02 = $span_y2$1;
  }
 }
 $15 = ($span_y2$1$lcssa|0)>($span_y1$1$lcssa|0);
 if (!($15)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$span>>2] = 0;
 $16 = ((($span)) + 4|0);
 HEAP32[$16>>2] = $span_y1$1$lcssa;
 $17 = ((($span)) + 8|0);
 HEAP32[$17>>2] = $width;
 $18 = (($span_y2$1$lcssa) - ($span_y1$1$lcssa))|0;
 $19 = ((($span)) + 12|0);
 HEAP32[$19>>2] = $18;
 $$0 = 1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_UpdateTexture($texture,$rect,$pixels,$pitch) {
 $texture = $texture|0;
 $rect = $rect|0;
 $pixels = $pixels|0;
 $pitch = $pitch|0;
 var $$0 = 0, $$01 = 0, $$2$i = 0, $$mask$i = 0, $$mask$i5 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $full_rect = 0, $full_rect$i = 0, $native_pitch$i = 0, $native_pitch$i3 = 0, $native_pixels$i = 0, $native_pixels$i2 = 0, $or$cond$i = 0, $or$cond$i6 = 0, $phitmp$i = 0, $phitmp$i7 = 0;
 var $switch$split2D = 0, $switch$split62D = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $native_pixels$i2 = sp + 68|0;
 $native_pitch$i3 = sp + 64|0;
 $full_rect$i = sp + 48|0;
 $native_pixels$i = sp + 44|0;
 $native_pitch$i = sp + 40|0;
 $full_rect = sp + 24|0;
 $0 = ($texture|0)==(0|0);
 if (!($0)) {
  $1 = HEAP32[$texture>>2]|0;
  $2 = ($1|0)==(1279300|0);
  if ($2) {
   $3 = ($pixels|0)==(0|0);
   if ($3) {
    HEAP32[$vararg_buffer1>>2] = 609974;
    $4 = (_SDL_SetError(611267,$vararg_buffer1)|0);
    $$0 = $4;
    STACKTOP = sp;return ($$0|0);
   }
   $5 = ($pitch|0)==(0);
   if ($5) {
    HEAP32[$vararg_buffer3>>2] = 609981;
    $6 = (_SDL_SetError(611267,$vararg_buffer3)|0);
    $$0 = $6;
    STACKTOP = sp;return ($$0|0);
   }
   $7 = ($rect|0)==(0|0);
   if ($7) {
    HEAP32[$full_rect>>2] = 0;
    $8 = ((($full_rect)) + 4|0);
    HEAP32[$8>>2] = 0;
    $9 = ((($texture)) + 12|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = ((($full_rect)) + 8|0);
    HEAP32[$11>>2] = $10;
    $12 = ((($texture)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($full_rect)) + 12|0);
    HEAP32[$14>>2] = $13;
    $$01 = $full_rect;$17 = $10;
   } else {
    $$phi$trans$insert = ((($rect)) + 8|0);
    $$pre = HEAP32[$$phi$trans$insert>>2]|0;
    $$01 = $rect;$17 = $$pre;
   }
   $15 = ((($$01)) + 8|0);
   $16 = ($17|0)==(0);
   if ($16) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   $18 = ((($$01)) + 12|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ($19|0)==(0);
   if ($20) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   $21 = ((($texture)) + 40|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = ($22|0)==(0|0);
   $24 = ((($texture)) + 36|0);
   $25 = HEAP32[$24>>2]|0;
   if (!($23)) {
    $26 = (_SDL_SW_UpdateYUVTexture($22,$$01,$pixels,$pitch)|0);
    $27 = ($26|0)<(0);
    do {
     if ($27) {
      $$2$i = -1;
     } else {
      HEAP32[$full_rect$i>>2] = 0;
      $28 = ((($full_rect$i)) + 4|0);
      HEAP32[$28>>2] = 0;
      $29 = ((($texture)) + 12|0);
      $30 = HEAP32[$29>>2]|0;
      $31 = ((($full_rect$i)) + 8|0);
      HEAP32[$31>>2] = $30;
      $32 = ((($texture)) + 16|0);
      $33 = HEAP32[$32>>2]|0;
      $34 = ((($full_rect$i)) + 12|0);
      HEAP32[$34>>2] = $33;
      $35 = ((($texture)) + 8|0);
      $36 = HEAP32[$35>>2]|0;
      $37 = ($36|0)==(1);
      if ($37) {
       $38 = (_SDL_LockTexture($25,$full_rect$i,$native_pixels$i,$native_pitch$i)|0);
       $39 = ($38|0)<(0);
       if ($39) {
        $$2$i = -1;
        break;
       } else {
        $40 = HEAP32[$21>>2]|0;
        $41 = ((($25)) + 4|0);
        $42 = HEAP32[$41>>2]|0;
        $43 = HEAP32[$31>>2]|0;
        $44 = HEAP32[$34>>2]|0;
        $45 = HEAP32[$native_pixels$i>>2]|0;
        $46 = HEAP32[$native_pitch$i>>2]|0;
        (_SDL_SW_CopyYUVToRGB($40,$full_rect$i,$42,$43,$44,$45,$46)|0);
        _SDL_UnlockTexture($25);
        $$2$i = 0;
        break;
       }
      }
      $47 = ((($25)) + 4|0);
      $48 = HEAP32[$47>>2]|0;
      $49 = ($48|0)==(0);
      $$mask$i = $48 & -268435456;
      $50 = ($$mask$i|0)==(268435456);
      $or$cond$i = $49 | $50;
      L32: do {
       if ($or$cond$i) {
        $52 = $48 & 255;
        $54 = $52;
       } else {
        $switch$split2D = ($48|0)<(1498831189);
        if ($switch$split2D) {
         switch ($48|0) {
         case 844715353:  {
          $54 = 2;
          break L32;
          break;
         }
         default: {
         }
         }
        } else {
         switch ($48|0) {
         case 1498831189:  {
          $54 = 2;
          break L32;
          break;
         }
         default: {
         }
         }
        }
        $51 = ($48|0)==(1431918169);
        $phitmp$i = $51 ? 2 : 1;
        $54 = $phitmp$i;
       }
      } while(0);
      $53 = Math_imul($54, $30)|0;
      $55 = (($53) + 3)|0;
      $56 = $55 & -4;
      $57 = Math_imul($56, $33)|0;
      $58 = (_SDL_malloc($57)|0);
      $59 = ($58|0)==(0|0);
      if ($59) {
       $64 = (_SDL_Error(0)|0);
       $$2$i = $64;
       break;
      } else {
       $60 = HEAP32[$21>>2]|0;
       $61 = HEAP32[$47>>2]|0;
       $62 = HEAP32[$31>>2]|0;
       $63 = HEAP32[$34>>2]|0;
       (_SDL_SW_CopyYUVToRGB($60,$full_rect$i,$61,$62,$63,$58,$56)|0);
       (_SDL_UpdateTexture($25,$full_rect$i,$58,$56)|0);
       _SDL_free($58);
       $$2$i = 0;
       break;
      }
     }
    } while(0);
    $$0 = $$2$i;
    STACKTOP = sp;return ($$0|0);
   }
   $65 = ($25|0)==(0|0);
   if ($65) {
    $98 = ((($texture)) + 32|0);
    $99 = HEAP32[$98>>2]|0;
    $100 = ((($99)) + 28|0);
    $101 = HEAP32[$100>>2]|0;
    $102 = (FUNCTION_TABLE_iiiiii[$101 & 8191]($99,$texture,$$01,$pixels,$pitch)|0);
    $$0 = $102;
    STACKTOP = sp;return ($$0|0);
   }
   $66 = ((($texture)) + 8|0);
   $67 = HEAP32[$66>>2]|0;
   $68 = ($67|0)==(1);
   if ($68) {
    $69 = (_SDL_LockTexture($25,$$01,$native_pixels$i2,$native_pitch$i3)|0);
    $70 = ($69|0)<(0);
    if ($70) {
     $$0 = -1;
     STACKTOP = sp;return ($$0|0);
    } else {
     $71 = HEAP32[$15>>2]|0;
     $72 = HEAP32[$18>>2]|0;
     $73 = ((($texture)) + 4|0);
     $74 = HEAP32[$73>>2]|0;
     $75 = ((($25)) + 4|0);
     $76 = HEAP32[$75>>2]|0;
     $77 = HEAP32[$native_pixels$i2>>2]|0;
     $78 = HEAP32[$native_pitch$i3>>2]|0;
     (_SDL_ConvertPixels($71,$72,$74,$pixels,$pitch,$76,$77,$78)|0);
     _SDL_UnlockTexture($25);
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
   }
   $79 = ((($25)) + 4|0);
   $80 = HEAP32[$79>>2]|0;
   $81 = ($80|0)==(0);
   $$mask$i5 = $80 & -268435456;
   $82 = ($$mask$i5|0)==(268435456);
   $or$cond$i6 = $81 | $82;
   L58: do {
    if ($or$cond$i6) {
     $84 = $80 & 255;
     $86 = $84;
    } else {
     $switch$split62D = ($80|0)<(1498831189);
     if ($switch$split62D) {
      switch ($80|0) {
      case 844715353:  {
       $86 = 2;
       break L58;
       break;
      }
      default: {
      }
      }
     } else {
      switch ($80|0) {
      case 1498831189:  {
       $86 = 2;
       break L58;
       break;
      }
      default: {
      }
      }
     }
     $83 = ($80|0)==(1431918169);
     $phitmp$i7 = $83 ? 2 : 1;
     $86 = $phitmp$i7;
    }
   } while(0);
   $85 = Math_imul($86, $17)|0;
   $87 = (($85) + 3)|0;
   $88 = $87 & -4;
   $89 = Math_imul($88, $19)|0;
   $90 = (_SDL_malloc($89)|0);
   $91 = ($90|0)==(0|0);
   if ($91) {
    $97 = (_SDL_Error(0)|0);
    $$0 = $97;
    STACKTOP = sp;return ($$0|0);
   } else {
    $92 = HEAP32[$15>>2]|0;
    $93 = HEAP32[$18>>2]|0;
    $94 = ((($texture)) + 4|0);
    $95 = HEAP32[$94>>2]|0;
    $96 = HEAP32[$79>>2]|0;
    (_SDL_ConvertPixels($92,$93,$95,$pixels,$pitch,$96,$90,$88)|0);
    (_SDL_UpdateTexture($25,$$01,$90,$88)|0);
    _SDL_free($90);
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
  }
 }
 (_SDL_SetError(597654,$vararg_buffer)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SW_UpdateYUVTexture($swdata,$rect,$pixels,$pitch) {
 $swdata = $swdata|0;
 $rect = $rect|0;
 $pixels = $pixels|0;
 $pitch = $pitch|0;
 var $$lcssa = 0, $$lcssa45 = 0, $$lcssa46 = 0, $$lcssa47 = 0, $$pre28 = 0, $$pre29 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dst$011 = 0, $dst$17 = 0, $dst$24 = 0, $dst2$020 = 0, $row$012 = 0, $row$18 = 0, $row$25 = 0, $row3$021 = 0, $src$013 = 0, $src$19 = 0, $src$26 = 0, $src1$019 = 0, $switch$split12D = 0, $switch$split2D = 0;
 var $switch$split42D = 0, $switch$split72D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$swdata>>2]|0;
 $switch$split2D = ($0|0)<(1431918169);
 L1: do {
  if ($switch$split2D) {
   $switch$split12D = ($0|0)<(844715353);
   if ($switch$split12D) {
    switch ($0|0) {
    case 842094169:  {
     break L1;
     break;
    }
    default: {
    }
    }
    return 0;
   } else {
    switch ($0|0) {
    case 844715353:  {
     label = 15;
     break L1;
     break;
    }
    default: {
    }
    }
    return 0;
   }
  } else {
   $switch$split42D = ($0|0)<(1448433993);
   if ($switch$split42D) {
    switch ($0|0) {
    case 1431918169:  {
     label = 15;
     break L1;
     break;
    }
    default: {
    }
    }
    return 0;
   }
   $switch$split72D = ($0|0)<(1498831189);
   if ($switch$split72D) {
    switch ($0|0) {
    case 1448433993:  {
     break L1;
     break;
    }
    default: {
    }
    }
    return 0;
   } else {
    switch ($0|0) {
    case 1498831189:  {
     label = 15;
     break L1;
     break;
    }
    default: {
    }
    }
    return 0;
   }
  }
 } while(0);
 if ((label|0) == 15) {
  $95 = ((($swdata)) + 36|0);
  $96 = ((($rect)) + 8|0);
  $97 = HEAP32[$96>>2]|0;
  $98 = $97 << 1;
  $99 = ((($rect)) + 12|0);
  $100 = HEAP32[$99>>2]|0;
  $101 = ($100|0)>(0);
  if (!($101)) {
   return 0;
  }
  $102 = ((($swdata)) + 44|0);
  $103 = HEAP32[$102>>2]|0;
  $104 = HEAP16[$95>>1]|0;
  $105 = $104&65535;
  $106 = ((($rect)) + 4|0);
  $107 = HEAP32[$106>>2]|0;
  $108 = Math_imul($105, $107)|0;
  $109 = (($103) + ($108)|0);
  $110 = HEAP32[$rect>>2]|0;
  $111 = $110 << 1;
  $112 = (($109) + ($111)|0);
  $dst2$020 = $112;$row3$021 = 0;$src1$019 = $pixels;
  while(1) {
   (_SDL_memcpy($dst2$020,$src1$019,$98)|0);
   $113 = (($src1$019) + ($pitch)|0);
   $114 = HEAP16[$95>>1]|0;
   $115 = $114&65535;
   $116 = (($dst2$020) + ($115)|0);
   $117 = (($row3$021) + 1)|0;
   $118 = HEAP32[$99>>2]|0;
   $119 = ($117|0)<($118|0);
   if ($119) {
    $dst2$020 = $116;$row3$021 = $117;$src1$019 = $113;
   } else {
    break;
   }
  }
  return 0;
 }
 $1 = HEAP32[$rect>>2]|0;
 $2 = ((($rect)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $1 | $3;
 $5 = ($4|0)==(0);
 if ($5) {
  $6 = ((($rect)) + 8|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ((($swdata)) + 8|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($7|0)==($9|0);
  if ($10) {
   $11 = ((($rect)) + 12|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ((($swdata)) + 12|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ($12|0)==($14|0);
   if ($15) {
    $16 = ((($swdata)) + 16|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = Math_imul($12, $7)|0;
    $19 = (($18|0) / 2)&-1;
    $20 = (($19) + ($18))|0;
    (_SDL_memcpy($17,$pixels,$20)|0);
    return 0;
   } else {
    $31 = 0;
   }
  } else {
   $31 = 0;
  }
 } else {
  $31 = $3;
 }
 $21 = ((($swdata)) + 16|0);
 $22 = ((($swdata)) + 8|0);
 $23 = ((($rect)) + 8|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($rect)) + 12|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ($26|0)>(0);
 if (!($27)) {
  return 0;
 }
 $28 = HEAP32[$22>>2]|0;
 $29 = HEAP32[$21>>2]|0;
 $30 = Math_imul($28, $31)|0;
 $32 = (($29) + ($30)|0);
 $33 = (($32) + ($1)|0);
 $dst$011 = $33;$row$012 = 0;$src$013 = $pixels;
 while(1) {
  (_SDL_memcpy($dst$011,$src$013,$24)|0);
  $34 = (($src$013) + ($pitch)|0);
  $35 = HEAP32[$22>>2]|0;
  $36 = (($dst$011) + ($35)|0);
  $37 = (($row$012) + 1)|0;
  $38 = HEAP32[$25>>2]|0;
  $39 = ($37|0)<($38|0);
  if ($39) {
   $dst$011 = $36;$row$012 = $37;$src$013 = $34;
  } else {
   $$lcssa46 = $35;$$lcssa47 = $38;
   break;
  }
 }
 $$pre28 = HEAP32[$23>>2]|0;
 $40 = ((($swdata)) + 12|0);
 $41 = (($$pre28|0) / 2)&-1;
 $42 = ($$lcssa47|0)>(1);
 if (!($42)) {
  return 0;
 }
 $43 = HEAP32[$21>>2]|0;
 $44 = HEAP32[$40>>2]|0;
 $45 = Math_imul($$lcssa46, $44)|0;
 $46 = (($43) + ($45)|0);
 $47 = HEAP32[$2>>2]|0;
 $48 = (($47|0) / 2)&-1;
 $49 = Math_imul($48, $$lcssa46)|0;
 $50 = (($49|0) / 2)&-1;
 $51 = HEAP32[$rect>>2]|0;
 $52 = (($51|0) / 2)&-1;
 $53 = (($50) + ($52))|0;
 $54 = (($46) + ($53)|0);
 $55 = Math_imul($$lcssa47, $pitch)|0;
 $56 = (($pixels) + ($55)|0);
 $57 = (($pitch|0) / 2)&-1;
 $dst$17 = $54;$row$18 = 0;$src$19 = $56;
 while(1) {
  (_SDL_memcpy($dst$17,$src$19,$41)|0);
  $58 = (($src$19) + ($57)|0);
  $59 = HEAP32[$22>>2]|0;
  $60 = (($59|0) / 2)&-1;
  $61 = (($dst$17) + ($60)|0);
  $62 = (($row$18) + 1)|0;
  $63 = HEAP32[$25>>2]|0;
  $64 = (($63|0) / 2)&-1;
  $65 = ($62|0)<($64|0);
  if ($65) {
   $dst$17 = $61;$row$18 = $62;$src$19 = $58;
  } else {
   $$lcssa = $59;$$lcssa45 = $63;
   break;
  }
 }
 $$pre29 = HEAP32[$23>>2]|0;
 $66 = Math_imul($$lcssa45, $pitch)|0;
 $67 = HEAP32[$40>>2]|0;
 $68 = Math_imul($$lcssa, $67)|0;
 $69 = (($$pre29|0) / 2)&-1;
 $70 = ($$lcssa45|0)>(1);
 if (!($70)) {
  return 0;
 }
 $71 = HEAP32[$21>>2]|0;
 $72 = (($71) + ($68)|0);
 $73 = (($68|0) / 4)&-1;
 $74 = (($72) + ($73)|0);
 $75 = HEAP32[$2>>2]|0;
 $76 = (($75|0) / 2)&-1;
 $77 = Math_imul($76, $$lcssa)|0;
 $78 = (($77|0) / 2)&-1;
 $79 = HEAP32[$rect>>2]|0;
 $80 = (($79|0) / 2)&-1;
 $81 = (($78) + ($80))|0;
 $82 = (($74) + ($81)|0);
 $83 = (($pixels) + ($66)|0);
 $84 = (($66|0) / 4)&-1;
 $85 = (($83) + ($84)|0);
 $86 = (($pitch|0) / 2)&-1;
 $dst$24 = $82;$row$25 = 0;$src$26 = $85;
 while(1) {
  (_SDL_memcpy($dst$24,$src$26,$69)|0);
  $87 = (($src$26) + ($86)|0);
  $88 = HEAP32[$22>>2]|0;
  $89 = (($88|0) / 2)&-1;
  $90 = (($dst$24) + ($89)|0);
  $91 = (($row$25) + 1)|0;
  $92 = HEAP32[$25>>2]|0;
  $93 = (($92|0) / 2)&-1;
  $94 = ($91|0)<($93|0);
  if ($94) {
   $dst$24 = $90;$row$25 = $91;$src$26 = $87;
  } else {
   break;
  }
 }
 return 0;
}
function _SDL_LockTexture($texture,$rect,$pixels,$pitch) {
 $texture = $texture|0;
 $rect = $rect|0;
 $pixels = $pixels|0;
 $pitch = $pitch|0;
 var $$0 = 0, $$01 = 0, $$mask$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $full_rect = 0, $or$cond$i = 0, $phitmp$i = 0, $switch$split2D = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $full_rect = sp + 16|0;
 $0 = ($texture|0)==(0|0);
 if (!($0)) {
  $1 = HEAP32[$texture>>2]|0;
  $2 = ($1|0)==(1279300|0);
  if ($2) {
   $3 = ((($texture)) + 8|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==(1);
   if (!($5)) {
    $6 = (_SDL_SetError(609987,$vararg_buffer1)|0);
    $$0 = $6;
    STACKTOP = sp;return ($$0|0);
   }
   $7 = ($rect|0)==(0|0);
   if ($7) {
    HEAP32[$full_rect>>2] = 0;
    $8 = ((($full_rect)) + 4|0);
    HEAP32[$8>>2] = 0;
    $9 = ((($texture)) + 12|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = ((($full_rect)) + 8|0);
    HEAP32[$11>>2] = $10;
    $12 = ((($texture)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($full_rect)) + 12|0);
    HEAP32[$14>>2] = $13;
    $$01 = $full_rect;
   } else {
    $$01 = $rect;
   }
   $15 = ((($texture)) + 40|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = ($16|0)==(0|0);
   if (!($17)) {
    $18 = (_SDL_SW_LockYUVTexture($16,$$01,$pixels,$pitch)|0);
    $$0 = $18;
    STACKTOP = sp;return ($$0|0);
   }
   $19 = ((($texture)) + 36|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = ($20|0)==(0|0);
   if ($21) {
    $41 = ((($texture)) + 32|0);
    $42 = HEAP32[$41>>2]|0;
    $43 = ((($42)) + 36|0);
    $44 = HEAP32[$43>>2]|0;
    $45 = (FUNCTION_TABLE_iiiiii[$44 & 8191]($42,$texture,$$01,$pixels,$pitch)|0);
    $$0 = $45;
    STACKTOP = sp;return ($$0|0);
   }
   $22 = ((($texture)) + 52|0);
   ;HEAP32[$22>>2]=HEAP32[$$01>>2]|0;HEAP32[$22+4>>2]=HEAP32[$$01+4>>2]|0;HEAP32[$22+8>>2]=HEAP32[$$01+8>>2]|0;HEAP32[$22+12>>2]=HEAP32[$$01+12>>2]|0;
   $23 = ((($texture)) + 44|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = ((($$01)) + 4|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = ((($texture)) + 48|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = Math_imul($28, $26)|0;
   $30 = (($24) + ($29)|0);
   $31 = HEAP32[$$01>>2]|0;
   $32 = ((($texture)) + 4|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ($33|0)==(0);
   $$mask$i = $33 & -268435456;
   $35 = ($$mask$i|0)==(268435456);
   $or$cond$i = $34 | $35;
   L19: do {
    if ($or$cond$i) {
     $37 = $33 & 255;
     $39 = $37;
    } else {
     $switch$split2D = ($33|0)<(1498831189);
     if ($switch$split2D) {
      switch ($33|0) {
      case 844715353:  {
       $39 = 2;
       break L19;
       break;
      }
      default: {
      }
      }
     } else {
      switch ($33|0) {
      case 1498831189:  {
       $39 = 2;
       break L19;
       break;
      }
      default: {
      }
      }
     }
     $36 = ($33|0)==(1431918169);
     $phitmp$i = $36 ? 2 : 1;
     $39 = $phitmp$i;
    }
   } while(0);
   $38 = Math_imul($39, $31)|0;
   $40 = (($30) + ($38)|0);
   HEAP32[$pixels>>2] = $40;
   HEAP32[$pitch>>2] = $28;
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_SDL_SetError(597654,$vararg_buffer)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SW_LockYUVTexture($swdata,$rect,$pixels,$pitch) {
 $swdata = $swdata|0;
 $rect = $rect|0;
 $pixels = $pixels|0;
 $pitch = $pitch|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $phitmp = 0, $storemerge = 0, $switch$split2D = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP32[$swdata>>2]|0;
 $switch$split2D = ($0|0)<(1448433993);
 if ($switch$split2D) {
  switch ($0|0) {
  case 842094169:  {
   label = 2;
   break;
  }
  default: {
   label = 10;
  }
  }
 } else {
  switch ($0|0) {
  case 1448433993:  {
   label = 2;
   break;
  }
  default: {
   label = 10;
  }
  }
 }
 do {
  if ((label|0) == 2) {
   $1 = ($rect|0)==(0|0);
   if ($1) {
    $2 = ((($swdata)) + 44|0);
    $3 = HEAP32[$2>>2]|0;
    $storemerge = $3;
    break;
   }
   $4 = HEAP32[$rect>>2]|0;
   $5 = ($4|0)==(0);
   if ($5) {
    $6 = ((($rect)) + 4|0);
    $7 = HEAP32[$6>>2]|0;
    $8 = ($7|0)==(0);
    if ($8) {
     $9 = ((($rect)) + 8|0);
     $10 = HEAP32[$9>>2]|0;
     $11 = ((($swdata)) + 8|0);
     $12 = HEAP32[$11>>2]|0;
     $13 = ($10|0)==($12|0);
     if ($13) {
      $14 = ((($rect)) + 12|0);
      $15 = HEAP32[$14>>2]|0;
      $16 = ((($swdata)) + 12|0);
      $17 = HEAP32[$16>>2]|0;
      $18 = ($15|0)==($17|0);
      if ($18) {
       $19 = ((($swdata)) + 44|0);
       $20 = HEAP32[$19>>2]|0;
       $29 = 0;$31 = $20;$33 = 0;
       label = 12;
       break;
      }
     }
    }
   }
   $21 = (_SDL_SetError(610032,$vararg_buffer)|0);
   $$0 = $21;
   STACKTOP = sp;return ($$0|0);
  }
  else if ((label|0) == 10) {
   $22 = ($rect|0)==(0|0);
   $23 = ((($swdata)) + 44|0);
   $24 = HEAP32[$23>>2]|0;
   if ($22) {
    $storemerge = $24;
   } else {
    $$phi$trans$insert = ((($rect)) + 4|0);
    $$pre = HEAP32[$$phi$trans$insert>>2]|0;
    $$pre3 = HEAP32[$rect>>2]|0;
    $phitmp = $$pre3 << 1;
    $29 = $$pre;$31 = $24;$33 = $phitmp;
    label = 12;
   }
  }
 } while(0);
 if ((label|0) == 12) {
  $25 = ((($swdata)) + 36|0);
  $26 = HEAP16[$25>>1]|0;
  $27 = $26&65535;
  $28 = Math_imul($27, $29)|0;
  $30 = (($31) + ($28)|0);
  $32 = (($30) + ($33)|0);
  $storemerge = $32;
 }
 HEAP32[$pixels>>2] = $storemerge;
 $34 = ((($swdata)) + 36|0);
 $35 = HEAP16[$34>>1]|0;
 $36 = $35&65535;
 HEAP32[$pitch>>2] = $36;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SW_CopyYUVToRGB($swdata,$srcrect,$target_format,$w,$h,$pixels,$pitch) {
 $swdata = $swdata|0;
 $srcrect = $srcrect|0;
 $target_format = $target_format|0;
 $w = $w|0;
 $h = $h|0;
 $pixels = $pixels|0;
 $pitch = $pitch|0;
 var $$$i$i = 0, $$$i49$i = 0, $$$i55$i = 0, $$0 = 0, $$03 = 0, $$05 = 0, $$1 = 0, $$14 = 0, $$16 = 0, $$mask = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0;
 var $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0;
 var $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $Amask = 0, $Amask$i = 0, $Bmask = 0, $Bmask$i = 0, $Cb$0 = 0, $Cr$0 = 0, $Gmask = 0, $Gmask$i = 0, $Rmask = 0, $Rmask$i = 0, $a$tr2$i$i = 0, $a$tr2$i47$i = 0, $a$tr2$i53$i = 0;
 var $accumulator$tr$lcssa$i$i = 0, $accumulator$tr$lcssa$i51$i = 0, $accumulator$tr$lcssa$i57$i = 0, $accumulator$tr1$i$i = 0, $accumulator$tr1$i48$i = 0, $accumulator$tr1$i54$i = 0, $bpp = 0, $bpp$i = 0, $exitcond$i = 0, $exitcond92$i = 0, $exitcond93$i = 0, $i$091$i = 0, $i$190$i = 0, $i$289$i = 0, $lum$0 = 0, $or$cond = 0, $or$cond$i = 0, $or$cond14 = 0, $or$cond8 = 0, $phitmp = 0;
 var $rect = 0, $scale_2x$011 = 0, $stretch$09 = 0, $switch$split112D = 0, $switch$split142D = 0, $switch$split172D = 0, $switch$split202D = 0, $switch$split232D = 0, $switch$split262D = 0, $switch$split292D = 0, $switch$split2D = 0, $switch$split322D = 0, $switch$split352D = 0, $switch$split382D = 0, $switch$split412D = 0, $switch$split442D = 0, $switch$split472D = 0, $switch$split52D = 0, $switch$split82D = 0, $vararg_buffer = 0;
 var $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $bpp$i = sp + 76|0;
 $Rmask$i = sp + 72|0;
 $Gmask$i = sp + 68|0;
 $Bmask$i = sp + 64|0;
 $Amask$i = sp + 60|0;
 $bpp = sp + 56|0;
 $Rmask = sp + 52|0;
 $Gmask = sp + 48|0;
 $Bmask = sp + 44|0;
 $Amask = sp + 40|0;
 $rect = sp + 24|0;
 $0 = ($target_format|0)==(0);
 $$mask = $target_format & -268435456;
 $1 = ($$mask|0)==(268435456);
 $or$cond8 = $0 | $1;
 L1: do {
  if ($or$cond8) {
   $2 = $target_format & 255;
   $3 = ($2|0)==(0);
   if ($3) {
    $4 = (_SDL_SetError(610087,$vararg_buffer)|0);
    $$1 = $4;
    STACKTOP = sp;return ($$1|0);
   } else {
    $196 = $2;
   }
  } else {
   $switch$split2D = ($target_format|0)<(1431918169);
   L6: do {
    if ($switch$split2D) {
     switch ($target_format|0) {
     case 844715353:  {
      break;
     }
     default: {
      $196 = 1;
      break L1;
     }
     }
    } else {
     $switch$split202D = ($target_format|0)<(1498831189);
     if ($switch$split202D) {
      switch ($target_format|0) {
      case 1431918169:  {
       break L6;
       break;
      }
      default: {
       $196 = 1;
       break L1;
      }
      }
     } else {
      switch ($target_format|0) {
      case 1498831189:  {
       break L6;
       break;
      }
      default: {
       $196 = 1;
       break L1;
      }
      }
     }
    }
   } while(0);
   $196 = 2;
  }
 } while(0);
 $5 = ((($swdata)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==($target_format|0);
 do {
  if (!($7)) {
   $8 = (_SDL_PixelFormatEnumToMasks($target_format,$bpp$i,$Rmask$i,$Gmask$i,$Bmask$i,$Amask$i)|0);
   $9 = ($8|0)==(0);
   $10 = HEAP32[$bpp$i>>2]|0;
   $11 = ($10|0)<(15);
   $or$cond$i = $9 | $11;
   if ($or$cond$i) {
    $12 = (_SDL_SetError(610115,$vararg_buffer1)|0);
    $phitmp = ($12|0)<(0);
    if ($phitmp) {
     $$1 = -1;
    } else {
     break;
    }
    STACKTOP = sp;return ($$1|0);
   }
   HEAP32[$5>>2] = $target_format;
   $13 = ((($swdata)) + 24|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ((($14)) + 3072|0);
   $16 = ((($14)) + 6144|0);
   $i$091$i = 0;
   while(1) {
    $17 = HEAP32[$Rmask$i>>2]|0;
    $18 = ($17|0)==(0);
    if ($18) {
     $accumulator$tr$lcssa$i$i = 0;
    } else {
     $a$tr2$i$i = $17;$accumulator$tr1$i$i = 0;
     while(1) {
      $19 = $a$tr2$i$i & 1;
      $20 = $a$tr2$i$i >>> 1;
      $$$i$i = (($19) + ($accumulator$tr1$i$i))|0;
      $21 = ($20|0)==(0);
      if ($21) {
       $accumulator$tr$lcssa$i$i = $$$i$i;
       break;
      } else {
       $a$tr2$i$i = $20;$accumulator$tr1$i$i = $$$i$i;
      }
     }
    }
    $22 = (8 - ($accumulator$tr$lcssa$i$i))|0;
    $23 = $i$091$i >> $22;
    $24 = (($i$091$i) + 256)|0;
    $25 = (($14) + ($24<<2)|0);
    HEAP32[$25>>2] = $23;
    $26 = HEAP32[$Rmask$i>>2]|0;
    $27 = (_free_bits_at_bottom($26)|0);
    $28 = $23 << $27;
    HEAP32[$25>>2] = $28;
    $29 = HEAP32[$Amask$i>>2]|0;
    $30 = $29 | $28;
    HEAP32[$25>>2] = $30;
    $31 = HEAP32[$Gmask$i>>2]|0;
    $32 = ($31|0)==(0);
    if ($32) {
     $accumulator$tr$lcssa$i51$i = 0;
    } else {
     $a$tr2$i47$i = $31;$accumulator$tr1$i48$i = 0;
     while(1) {
      $33 = $a$tr2$i47$i & 1;
      $34 = $a$tr2$i47$i >>> 1;
      $$$i49$i = (($33) + ($accumulator$tr1$i48$i))|0;
      $35 = ($34|0)==(0);
      if ($35) {
       $accumulator$tr$lcssa$i51$i = $$$i49$i;
       break;
      } else {
       $a$tr2$i47$i = $34;$accumulator$tr1$i48$i = $$$i49$i;
      }
     }
    }
    $36 = (8 - ($accumulator$tr$lcssa$i51$i))|0;
    $37 = $i$091$i >> $36;
    $38 = (($15) + ($24<<2)|0);
    HEAP32[$38>>2] = $37;
    $39 = HEAP32[$Gmask$i>>2]|0;
    $40 = (_free_bits_at_bottom($39)|0);
    $41 = $37 << $40;
    HEAP32[$38>>2] = $41;
    $42 = HEAP32[$Amask$i>>2]|0;
    $43 = $42 | $41;
    HEAP32[$38>>2] = $43;
    $44 = HEAP32[$Bmask$i>>2]|0;
    $45 = ($44|0)==(0);
    if ($45) {
     $accumulator$tr$lcssa$i57$i = 0;
    } else {
     $a$tr2$i53$i = $44;$accumulator$tr1$i54$i = 0;
     while(1) {
      $46 = $a$tr2$i53$i & 1;
      $47 = $a$tr2$i53$i >>> 1;
      $$$i55$i = (($46) + ($accumulator$tr1$i54$i))|0;
      $48 = ($47|0)==(0);
      if ($48) {
       $accumulator$tr$lcssa$i57$i = $$$i55$i;
       break;
      } else {
       $a$tr2$i53$i = $47;$accumulator$tr1$i54$i = $$$i55$i;
      }
     }
    }
    $49 = (8 - ($accumulator$tr$lcssa$i57$i))|0;
    $50 = $i$091$i >> $49;
    $51 = (($16) + ($24<<2)|0);
    HEAP32[$51>>2] = $50;
    $52 = HEAP32[$Bmask$i>>2]|0;
    $53 = (_free_bits_at_bottom($52)|0);
    $54 = $50 << $53;
    HEAP32[$51>>2] = $54;
    $55 = HEAP32[$Amask$i>>2]|0;
    $56 = $55 | $54;
    HEAP32[$51>>2] = $56;
    $57 = (($i$091$i) + 1)|0;
    $exitcond93$i = ($57|0)==(256);
    if ($exitcond93$i) {
     break;
    } else {
     $i$091$i = $57;
    }
   }
   L35: do {
    if ($or$cond8) {
     $64 = $target_format & 255;
     $65 = ($64|0)==(2);
     if ($65) {
      $i$190$i = 0;
      label = 21;
     }
    } else {
     $switch$split52D = ($target_format|0)<(1431918169);
     if ($switch$split52D) {
      switch ($target_format|0) {
      case 844715353:  {
       $i$190$i = 0;
       label = 21;
       break L35;
       break;
      }
      default: {
       break L35;
      }
      }
     }
     $switch$split232D = ($target_format|0)<(1498831189);
     if ($switch$split232D) {
      switch ($target_format|0) {
      case 1431918169:  {
       $i$190$i = 0;
       label = 21;
       break L35;
       break;
      }
      default: {
       break L35;
      }
      }
     } else {
      switch ($target_format|0) {
      case 1498831189:  {
       $i$190$i = 0;
       label = 21;
       break L35;
       break;
      }
      default: {
       break L35;
      }
      }
     }
    }
   } while(0);
   if ((label|0) == 21) {
    while(1) {
     label = 0;
     $66 = (($i$190$i) + 256)|0;
     $67 = (($14) + ($66<<2)|0);
     $68 = HEAP32[$67>>2]|0;
     $69 = $68 << 16;
     $70 = $69 | $68;
     HEAP32[$67>>2] = $70;
     $71 = (($15) + ($66<<2)|0);
     $72 = HEAP32[$71>>2]|0;
     $73 = $72 << 16;
     $74 = $73 | $72;
     HEAP32[$71>>2] = $74;
     $75 = (($16) + ($66<<2)|0);
     $76 = HEAP32[$75>>2]|0;
     $77 = $76 << 16;
     $78 = $77 | $76;
     HEAP32[$75>>2] = $78;
     $79 = (($i$190$i) + 1)|0;
     $exitcond92$i = ($79|0)==(256);
     if ($exitcond92$i) {
      break;
     } else {
      $i$190$i = $79;
      label = 21;
     }
    }
   }
   $58 = ((($14)) + 1024|0);
   $59 = ((($14)) + 2044|0);
   $60 = ((($14)) + 4096|0);
   $61 = ((($14)) + 5116|0);
   $62 = ((($14)) + 7168|0);
   $63 = ((($14)) + 8188|0);
   $i$289$i = 0;
   while(1) {
    $80 = HEAP32[$58>>2]|0;
    $81 = (($14) + ($i$289$i<<2)|0);
    HEAP32[$81>>2] = $80;
    $82 = HEAP32[$59>>2]|0;
    $83 = (($i$289$i) + 512)|0;
    $84 = (($14) + ($83<<2)|0);
    HEAP32[$84>>2] = $82;
    $85 = HEAP32[$60>>2]|0;
    $86 = (($15) + ($i$289$i<<2)|0);
    HEAP32[$86>>2] = $85;
    $87 = HEAP32[$61>>2]|0;
    $88 = (($15) + ($83<<2)|0);
    HEAP32[$88>>2] = $87;
    $89 = HEAP32[$62>>2]|0;
    $90 = (($16) + ($i$289$i<<2)|0);
    HEAP32[$90>>2] = $89;
    $91 = HEAP32[$63>>2]|0;
    $92 = (($16) + ($83<<2)|0);
    HEAP32[$92>>2] = $91;
    $93 = (($i$289$i) + 1)|0;
    $exitcond$i = ($93|0)==(256);
    if ($exitcond$i) {
     break;
    } else {
     $i$289$i = $93;
    }
   }
   $94 = HEAP32[$swdata>>2]|0;
   $switch$split82D = ($94|0)<(1431918169);
   L51: do {
    if ($switch$split82D) {
     $switch$split262D = ($94|0)<(844715353);
     if ($switch$split262D) {
      switch ($94|0) {
      case 842094169:  {
       label = 24;
       break L51;
       break;
      }
      default: {
       break L51;
      }
      }
     } else {
      switch ($94|0) {
      case 844715353:  {
       label = 31;
       break L51;
       break;
      }
      default: {
       break L51;
      }
      }
     }
    } else {
     $switch$split292D = ($94|0)<(1448433993);
     if ($switch$split292D) {
      switch ($94|0) {
      case 1431918169:  {
       label = 31;
       break L51;
       break;
      }
      default: {
       break L51;
      }
      }
     }
     $switch$split442D = ($94|0)<(1498831189);
     if ($switch$split442D) {
      switch ($94|0) {
      case 1448433993:  {
       label = 24;
       break L51;
       break;
      }
      default: {
       break L51;
      }
      }
     } else {
      switch ($94|0) {
      case 1498831189:  {
       label = 31;
       break L51;
       break;
      }
      default: {
       break L51;
      }
      }
     }
    }
   } while(0);
   L63: do {
    if ((label|0) == 24) {
     if ($or$cond8) {
      $95 = $target_format & 255;
      switch ($95|0) {
      case 2:  {
       $96 = ((($swdata)) + 28|0);
       HEAP32[$96>>2] = 4651;
       $97 = ((($swdata)) + 32|0);
       HEAP32[$97>>2] = 4652;
       break L63;
       break;
      }
      case 3:  {
       $100 = ((($swdata)) + 28|0);
       HEAP32[$100>>2] = 4653;
       $101 = ((($swdata)) + 32|0);
       HEAP32[$101>>2] = 4654;
       break L63;
       break;
      }
      case 4:  {
       $102 = ((($swdata)) + 28|0);
       HEAP32[$102>>2] = 4655;
       $103 = ((($swdata)) + 32|0);
       HEAP32[$103>>2] = 4656;
       break L63;
       break;
      }
      default: {
       break L63;
      }
      }
     }
     $switch$split112D = ($target_format|0)<(1431918169);
     L72: do {
      if ($switch$split112D) {
       switch ($target_format|0) {
       case 844715353:  {
        break;
       }
       default: {
        break L63;
       }
       }
      } else {
       $switch$split322D = ($target_format|0)<(1498831189);
       if ($switch$split322D) {
        switch ($target_format|0) {
        case 1431918169:  {
         break L72;
         break;
        }
        default: {
         break L63;
        }
        }
       } else {
        switch ($target_format|0) {
        case 1498831189:  {
         break L72;
         break;
        }
        default: {
         break L63;
        }
        }
       }
      }
     } while(0);
     $98 = ((($swdata)) + 28|0);
     HEAP32[$98>>2] = 4651;
     $99 = ((($swdata)) + 32|0);
     HEAP32[$99>>2] = 4652;
    }
    else if ((label|0) == 31) {
     if ($or$cond8) {
      $104 = $target_format & 255;
      switch ($104|0) {
      case 2:  {
       $105 = ((($swdata)) + 28|0);
       HEAP32[$105>>2] = 4657;
       $106 = ((($swdata)) + 32|0);
       HEAP32[$106>>2] = 4658;
       break L63;
       break;
      }
      case 3:  {
       $109 = ((($swdata)) + 28|0);
       HEAP32[$109>>2] = 4659;
       $110 = ((($swdata)) + 32|0);
       HEAP32[$110>>2] = 4660;
       break L63;
       break;
      }
      case 4:  {
       $111 = ((($swdata)) + 28|0);
       HEAP32[$111>>2] = 4661;
       $112 = ((($swdata)) + 32|0);
       HEAP32[$112>>2] = 4662;
       break L63;
       break;
      }
      default: {
       break L63;
      }
      }
     }
     $switch$split142D = ($target_format|0)<(1431918169);
     L87: do {
      if ($switch$split142D) {
       switch ($target_format|0) {
       case 844715353:  {
        break;
       }
       default: {
        break L63;
       }
       }
      } else {
       $switch$split352D = ($target_format|0)<(1498831189);
       if ($switch$split352D) {
        switch ($target_format|0) {
        case 1431918169:  {
         break L87;
         break;
        }
        default: {
         break L63;
        }
        }
       } else {
        switch ($target_format|0) {
        case 1498831189:  {
         break L87;
         break;
        }
        default: {
         break L63;
        }
        }
       }
      }
     } while(0);
     $107 = ((($swdata)) + 28|0);
     HEAP32[$107>>2] = 4657;
     $108 = ((($swdata)) + 32|0);
     HEAP32[$108>>2] = 4658;
    }
   } while(0);
   $113 = ((($swdata)) + 60|0);
   $114 = HEAP32[$113>>2]|0;
   _SDL_FreeSurface($114);
   HEAP32[$113>>2] = 0;
  }
 } while(0);
 $115 = HEAP32[$srcrect>>2]|0;
 $116 = ($115|0)==(0);
 if ($116) {
  $117 = ((($srcrect)) + 4|0);
  $118 = HEAP32[$117>>2]|0;
  $119 = ($118|0)==(0);
  if ($119) {
   $120 = ((($srcrect)) + 8|0);
   $121 = HEAP32[$120>>2]|0;
   $122 = ((($swdata)) + 8|0);
   $123 = HEAP32[$122>>2]|0;
   $124 = ($121|0)<($123|0);
   if ($124) {
    label = 45;
   } else {
    $125 = ((($srcrect)) + 12|0);
    $126 = HEAP32[$125>>2]|0;
    $127 = ((($swdata)) + 12|0);
    $128 = HEAP32[$127>>2]|0;
    $129 = ($126|0)<($128|0);
    if ($129) {
     label = 45;
    } else {
     $130 = ($121|0)==($w|0);
     $131 = ($126|0)==($h|0);
     $or$cond = $130 & $131;
     if ($or$cond) {
      $$14 = $pitch;$$16 = $pixels;$scale_2x$011 = 0;$stretch$09 = 0;
     } else {
      $132 = $121 << 1;
      $133 = ($132|0)==($w|0);
      $134 = $126 << 1;
      $135 = ($134|0)==($h|0);
      $or$cond14 = $133 & $135;
      if ($or$cond14) {
       $$14 = $pitch;$$16 = $pixels;$scale_2x$011 = 1;$stretch$09 = 0;
      } else {
       label = 45;
      }
     }
    }
   }
  } else {
   label = 45;
  }
 } else {
  label = 45;
 }
 if ((label|0) == 45) {
  $136 = ((($swdata)) + 60|0);
  $137 = HEAP32[$136>>2]|0;
  $138 = ($137|0)==(0|0);
  if ($138) {
   (_SDL_PixelFormatEnumToMasks($target_format,$bpp,$Rmask,$Gmask,$Bmask,$Amask)|0);
   $143 = HEAP32[$bpp>>2]|0;
   $144 = HEAP32[$Rmask>>2]|0;
   $145 = HEAP32[$Gmask>>2]|0;
   $146 = HEAP32[$Bmask>>2]|0;
   $147 = HEAP32[$Amask>>2]|0;
   $148 = (_SDL_CreateRGBSurfaceFrom($pixels,$w,$h,$143,$pitch,$144,$145,$146,$147)|0);
   HEAP32[$136>>2] = $148;
   $149 = ($148|0)==(0|0);
   if ($149) {
    $$0 = 0;$$03 = $pitch;$$05 = $pixels;
   } else {
    label = 48;
   }
  } else {
   $139 = ((($137)) + 8|0);
   HEAP32[$139>>2] = $w;
   $140 = ((($137)) + 12|0);
   HEAP32[$140>>2] = $h;
   $141 = ((($137)) + 20|0);
   HEAP32[$141>>2] = $pixels;
   $142 = ((($137)) + 16|0);
   HEAP32[$142>>2] = $pitch;
   label = 48;
  }
  do {
   if ((label|0) == 48) {
    $150 = ((($swdata)) + 56|0);
    $151 = HEAP32[$150>>2]|0;
    $152 = ($151|0)==(0|0);
    if ($152) {
     (_SDL_PixelFormatEnumToMasks($target_format,$bpp,$Rmask,$Gmask,$Bmask,$Amask)|0);
     $153 = ((($swdata)) + 8|0);
     $154 = HEAP32[$153>>2]|0;
     $155 = ((($swdata)) + 12|0);
     $156 = HEAP32[$155>>2]|0;
     $157 = HEAP32[$bpp>>2]|0;
     $158 = HEAP32[$Rmask>>2]|0;
     $159 = HEAP32[$Gmask>>2]|0;
     $160 = HEAP32[$Bmask>>2]|0;
     $161 = HEAP32[$Amask>>2]|0;
     $162 = (_SDL_CreateRGBSurface(0,$154,$156,$157,$158,$159,$160,$161)|0);
     HEAP32[$150>>2] = $162;
     $163 = ($162|0)==(0|0);
     if ($163) {
      $$0 = 0;$$03 = $pitch;$$05 = $pixels;
      break;
     } else {
      $165 = $162;
     }
    } else {
     $165 = $151;
    }
    $164 = ((($165)) + 20|0);
    $166 = HEAP32[$164>>2]|0;
    $167 = ((($165)) + 16|0);
    $168 = HEAP32[$167>>2]|0;
    $$0 = 1;$$03 = $168;$$05 = $166;
   }
  } while(0);
  if ($$0) {
   $$14 = $$03;$$16 = $$05;$scale_2x$011 = 0;$stretch$09 = 1;
  } else {
   $$1 = -1;
   STACKTOP = sp;return ($$1|0);
  }
 }
 $169 = HEAP32[$swdata>>2]|0;
 $switch$split172D = ($169|0)<(1431918169);
 L115: do {
  if ($switch$split172D) {
   $switch$split382D = ($169|0)<(844715353);
   if ($switch$split382D) {
    switch ($169|0) {
    case 842094169:  {
     break;
    }
    default: {
     label = 58;
     break L115;
    }
    }
    $170 = ((($swdata)) + 44|0);
    $171 = HEAP32[$170>>2]|0;
    $172 = ((($swdata)) + 48|0);
    $173 = HEAP32[$172>>2]|0;
    $174 = ((($swdata)) + 52|0);
    $175 = HEAP32[$174>>2]|0;
    $Cb$0 = $175;$Cr$0 = $173;$lum$0 = $171;
    break;
   } else {
    switch ($169|0) {
    case 844715353:  {
     break;
    }
    default: {
     label = 58;
     break L115;
    }
    }
    $182 = ((($swdata)) + 44|0);
    $183 = HEAP32[$182>>2]|0;
    $184 = ((($183)) + 3|0);
    $185 = ((($183)) + 1|0);
    $Cb$0 = $185;$Cr$0 = $184;$lum$0 = $183;
    break;
   }
  } else {
   $switch$split412D = ($169|0)<(1448433993);
   if ($switch$split412D) {
    switch ($169|0) {
    case 1431918169:  {
     break;
    }
    default: {
     label = 58;
     break L115;
    }
    }
    $190 = ((($swdata)) + 44|0);
    $191 = HEAP32[$190>>2]|0;
    $192 = ((($191)) + 1|0);
    $193 = ((($191)) + 3|0);
    $Cb$0 = $193;$Cr$0 = $192;$lum$0 = $191;
    break;
   }
   $switch$split472D = ($169|0)<(1498831189);
   if ($switch$split472D) {
    switch ($169|0) {
    case 1448433993:  {
     break;
    }
    default: {
     label = 58;
     break L115;
    }
    }
    $176 = ((($swdata)) + 44|0);
    $177 = HEAP32[$176>>2]|0;
    $178 = ((($swdata)) + 52|0);
    $179 = HEAP32[$178>>2]|0;
    $180 = ((($swdata)) + 48|0);
    $181 = HEAP32[$180>>2]|0;
    $Cb$0 = $181;$Cr$0 = $179;$lum$0 = $177;
    break;
   } else {
    switch ($169|0) {
    case 1498831189:  {
     break;
    }
    default: {
     label = 58;
     break L115;
    }
    }
    $186 = ((($swdata)) + 44|0);
    $187 = HEAP32[$186>>2]|0;
    $188 = ((($187)) + 1|0);
    $189 = ((($187)) + 2|0);
    $Cb$0 = $187;$Cr$0 = $189;$lum$0 = $188;
    break;
   }
  }
 } while(0);
 if ((label|0) == 58) {
  $194 = (_SDL_SetError(610150,$vararg_buffer3)|0);
  $$1 = $194;
  STACKTOP = sp;return ($$1|0);
 }
 $195 = (($$14|0) / ($196|0))&-1;
 $197 = ($scale_2x$011|0)==(0);
 $198 = ((($swdata)) + 8|0);
 $199 = HEAP32[$198>>2]|0;
 $200 = ((($swdata)) + 20|0);
 $201 = HEAP32[$200>>2]|0;
 $202 = ((($swdata)) + 24|0);
 $203 = HEAP32[$202>>2]|0;
 $204 = ((($swdata)) + 12|0);
 $205 = HEAP32[$204>>2]|0;
 if ($197) {
  $210 = (($195) - ($199))|0;
  $211 = ((($swdata)) + 28|0);
  $212 = HEAP32[$211>>2]|0;
  FUNCTION_TABLE_viiiiiiiii[$212 & 8191]($201,$203,$lum$0,$Cr$0,$Cb$0,$$16,$205,$199,$210);
 } else {
  $206 = $199 << 1;
  $207 = (($195) - ($206))|0;
  $208 = ((($swdata)) + 32|0);
  $209 = HEAP32[$208>>2]|0;
  FUNCTION_TABLE_viiiiiiiii[$209 & 8191]($201,$203,$lum$0,$Cr$0,$Cb$0,$$16,$205,$199,$207);
 }
 if (!($stretch$09)) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 }
 ;HEAP32[$rect>>2]=HEAP32[$srcrect>>2]|0;HEAP32[$rect+4>>2]=HEAP32[$srcrect+4>>2]|0;HEAP32[$rect+8>>2]=HEAP32[$srcrect+8>>2]|0;HEAP32[$rect+12>>2]=HEAP32[$srcrect+12>>2]|0;
 $213 = ((($swdata)) + 56|0);
 $214 = HEAP32[$213>>2]|0;
 $215 = ((($swdata)) + 60|0);
 $216 = HEAP32[$215>>2]|0;
 (_SDL_SoftStretch($214,$rect,$216,0)|0);
 $$1 = 0;
 STACKTOP = sp;return ($$1|0);
}
function _free_bits_at_bottom($a) {
 $a = $a|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($a|0)==(0);
 if ($0) {
  $$0 = 32;
  return ($$0|0);
 }
 $1 = $a & 1;
 $2 = ($1|0)==(0);
 if ($2) {
  $3 = $a >>> 1;
  $4 = (_free_bits_at_bottom($3)|0);
  $5 = (($4) + 1)|0;
  return ($5|0);
 } else {
  $$0 = 0;
  return ($$0|0);
 }
 return (0)|0;
}
function _Color16DitherYV12Mod1X($colortab,$rgb_2_pix,$lum,$cr,$cb,$out,$rows,$cols,$mod) {
 $colortab = $colortab|0;
 $rgb_2_pix = $rgb_2_pix|0;
 $lum = $lum|0;
 $cr = $cr|0;
 $cb = $cb|0;
 $out = $out|0;
 $rows = $rows|0;
 $cols = $cols|0;
 $mod = $mod|0;
 var $$0121 = 0, $$022 = 0, $$0320 = 0, $$111 = 0, $$1210 = 0, $$149 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cols$off = 0, $factor = 0, $lum2$0 = 0, $lum2$16 = 0, $or$cond = 0, $row1$019 = 0, $row1$18 = 0, $row2$017 = 0, $row2$15 = 0, $rows$off = 0, $scevgep = 0, $scevgep33 = 0, $scevgep34 = 0;
 var $scevgep35 = 0, $scevgep36 = 0, $x$07 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($cols|0) / 2)&-1;
 $factor = $mod << 1;
 $1 = (($factor) + ($cols))|0;
 $rows$off = (($rows) + 1)|0;
 $2 = ($rows$off>>>0)<(3);
 $cols$off = (($cols) + 1)|0;
 $3 = ($cols$off>>>0)<(3);
 $or$cond = $2 | $3;
 if ($or$cond) {
  return;
 }
 $4 = (($rows|0) / 2)&-1;
 $5 = (($out) + ($cols<<1)|0);
 $6 = (($5) + ($mod<<1)|0);
 $7 = $0 << 1;
 $$0121 = $cr;$$022 = $lum;$$0320 = $cb;$$in = $4;$row1$019 = $out;$row2$017 = $6;
 while(1) {
  $8 = (($$in) + -1)|0;
  $lum2$0 = (($$022) + ($cols)|0);
  $scevgep = (($row2$017) + ($7<<1)|0);
  $scevgep33 = (($row1$019) + ($7<<1)|0);
  $scevgep35 = (($$0320) + ($0)|0);
  $$111 = $$022;$$1210 = $$0121;$$149 = $$0320;$lum2$16 = $lum2$0;$row1$18 = $row1$019;$row2$15 = $row2$017;$x$07 = $0;
  while(1) {
   $9 = (($x$07) + -1)|0;
   $10 = HEAP8[$$1210>>0]|0;
   $11 = $10&255;
   $12 = (($colortab) + ($11<<2)|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = (($13) + 256)|0;
   $15 = $11 | 256;
   $16 = (($colortab) + ($15<<2)|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = (($17) + 1024)|0;
   $19 = HEAP8[$$149>>0]|0;
   $20 = $19&255;
   $21 = $20 | 512;
   $22 = (($colortab) + ($21<<2)|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = (($18) + ($23))|0;
   $25 = $20 | 768;
   $26 = (($colortab) + ($25<<2)|0);
   $27 = HEAP32[$26>>2]|0;
   $28 = (($27) + 1792)|0;
   $29 = ((($$1210)) + 1|0);
   $30 = ((($$149)) + 1|0);
   $31 = ((($$111)) + 1|0);
   $32 = HEAP8[$$111>>0]|0;
   $33 = $32&255;
   $34 = (($33) + ($14))|0;
   $35 = (($rgb_2_pix) + ($34<<2)|0);
   $36 = HEAP32[$35>>2]|0;
   $37 = (($33) + ($24))|0;
   $38 = (($rgb_2_pix) + ($37<<2)|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = $39 | $36;
   $41 = (($33) + ($28))|0;
   $42 = (($rgb_2_pix) + ($41<<2)|0);
   $43 = HEAP32[$42>>2]|0;
   $44 = $40 | $43;
   $45 = $44&65535;
   $46 = ((($row1$18)) + 2|0);
   HEAP16[$row1$18>>1] = $45;
   $47 = ((($$111)) + 2|0);
   $48 = HEAP8[$31>>0]|0;
   $49 = $48&255;
   $50 = (($49) + ($14))|0;
   $51 = (($rgb_2_pix) + ($50<<2)|0);
   $52 = HEAP32[$51>>2]|0;
   $53 = (($49) + ($24))|0;
   $54 = (($rgb_2_pix) + ($53<<2)|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = $55 | $52;
   $57 = (($49) + ($28))|0;
   $58 = (($rgb_2_pix) + ($57<<2)|0);
   $59 = HEAP32[$58>>2]|0;
   $60 = $56 | $59;
   $61 = $60&65535;
   $62 = ((($row1$18)) + 4|0);
   HEAP16[$46>>1] = $61;
   $63 = ((($lum2$16)) + 1|0);
   $64 = HEAP8[$lum2$16>>0]|0;
   $65 = $64&255;
   $66 = (($65) + ($14))|0;
   $67 = (($rgb_2_pix) + ($66<<2)|0);
   $68 = HEAP32[$67>>2]|0;
   $69 = (($65) + ($24))|0;
   $70 = (($rgb_2_pix) + ($69<<2)|0);
   $71 = HEAP32[$70>>2]|0;
   $72 = $71 | $68;
   $73 = (($65) + ($28))|0;
   $74 = (($rgb_2_pix) + ($73<<2)|0);
   $75 = HEAP32[$74>>2]|0;
   $76 = $72 | $75;
   $77 = $76&65535;
   $78 = ((($row2$15)) + 2|0);
   HEAP16[$row2$15>>1] = $77;
   $79 = ((($lum2$16)) + 2|0);
   $80 = HEAP8[$63>>0]|0;
   $81 = $80&255;
   $82 = (($81) + ($14))|0;
   $83 = (($rgb_2_pix) + ($82<<2)|0);
   $84 = HEAP32[$83>>2]|0;
   $85 = (($81) + ($24))|0;
   $86 = (($rgb_2_pix) + ($85<<2)|0);
   $87 = HEAP32[$86>>2]|0;
   $88 = $87 | $84;
   $89 = (($81) + ($28))|0;
   $90 = (($rgb_2_pix) + ($89<<2)|0);
   $91 = HEAP32[$90>>2]|0;
   $92 = $88 | $91;
   $93 = $92&65535;
   $94 = ((($row2$15)) + 4|0);
   HEAP16[$78>>1] = $93;
   $95 = ($9|0)==(0);
   if ($95) {
    break;
   } else {
    $$111 = $47;$$1210 = $29;$$149 = $30;$lum2$16 = $79;$row1$18 = $62;$row2$15 = $94;$x$07 = $9;
   }
  }
  $scevgep34 = (($$022) + ($7)|0);
  $96 = (($scevgep34) + ($cols)|0);
  $scevgep36 = (($$0121) + ($0)|0);
  $97 = (($scevgep33) + ($1<<1)|0);
  $98 = (($scevgep) + ($1<<1)|0);
  $99 = ($8|0)==(0);
  if ($99) {
   break;
  } else {
   $$0121 = $scevgep36;$$022 = $96;$$0320 = $scevgep35;$$in = $8;$row1$019 = $97;$row2$017 = $98;
  }
 }
 return;
}
function _Color16DitherYV12Mod2X($colortab,$rgb_2_pix,$lum,$cr,$cb,$out,$rows,$cols,$mod) {
 $colortab = $colortab|0;
 $rgb_2_pix = $rgb_2_pix|0;
 $lum = $lum|0;
 $cr = $cr|0;
 $cb = $cb|0;
 $out = $out|0;
 $rows = $rows|0;
 $cols = $cols|0;
 $mod = $mod|0;
 var $$0121 = 0, $$022 = 0, $$0320 = 0, $$111 = 0, $$1210 = 0, $$149 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cols$off = 0, $lum2$0 = 0, $lum2$16 = 0, $or$cond = 0, $row1$019 = 0, $row1$18 = 0, $row2$017 = 0, $row2$15 = 0, $rows$off = 0, $scevgep = 0;
 var $scevgep33 = 0, $scevgep34 = 0, $scevgep35 = 0, $scevgep36 = 0, $x$07 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($mod|0) / 2)&-1;
 $1 = (($0) + ($cols))|0;
 $2 = (($cols|0) / 2)&-1;
 $3 = ($1*3)|0;
 $4 = (($3) + ($0))|0;
 $rows$off = (($rows) + 1)|0;
 $5 = ($rows$off>>>0)<(3);
 $cols$off = (($cols) + 1)|0;
 $6 = ($cols$off>>>0)<(3);
 $or$cond = $5 | $6;
 if ($or$cond) {
  return;
 }
 $7 = (($rows|0) / 2)&-1;
 $8 = $1 << 1;
 $9 = (($out) + ($8<<2)|0);
 $10 = $2 << 1;
 $$0121 = $cr;$$022 = $lum;$$0320 = $cb;$$in = $7;$row1$019 = $out;$row2$017 = $9;
 while(1) {
  $11 = (($$in) + -1)|0;
  $lum2$0 = (($$022) + ($cols)|0);
  $scevgep = (($row2$017) + ($10<<2)|0);
  $scevgep33 = (($row1$019) + ($10<<2)|0);
  $scevgep35 = (($$0320) + ($2)|0);
  $$111 = $$022;$$1210 = $$0121;$$149 = $$0320;$lum2$16 = $lum2$0;$row1$18 = $row1$019;$row2$15 = $row2$017;$x$07 = $2;
  while(1) {
   $12 = (($x$07) + -1)|0;
   $13 = HEAP8[$$1210>>0]|0;
   $14 = $13&255;
   $15 = (($colortab) + ($14<<2)|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = (($16) + 256)|0;
   $18 = $14 | 256;
   $19 = (($colortab) + ($18<<2)|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = (($20) + 1024)|0;
   $22 = HEAP8[$$149>>0]|0;
   $23 = $22&255;
   $24 = $23 | 512;
   $25 = (($colortab) + ($24<<2)|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = (($21) + ($26))|0;
   $28 = $23 | 768;
   $29 = (($colortab) + ($28<<2)|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = (($30) + 1792)|0;
   $32 = ((($$1210)) + 1|0);
   $33 = ((($$149)) + 1|0);
   $34 = ((($$111)) + 1|0);
   $35 = HEAP8[$$111>>0]|0;
   $36 = $35&255;
   $37 = (($36) + ($17))|0;
   $38 = (($rgb_2_pix) + ($37<<2)|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = (($36) + ($27))|0;
   $41 = (($rgb_2_pix) + ($40<<2)|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = $42 | $39;
   $44 = (($36) + ($31))|0;
   $45 = (($rgb_2_pix) + ($44<<2)|0);
   $46 = HEAP32[$45>>2]|0;
   $47 = $43 | $46;
   $48 = (($row1$18) + ($1<<2)|0);
   HEAP32[$48>>2] = $47;
   HEAP32[$row1$18>>2] = $47;
   $49 = ((($row1$18)) + 4|0);
   $50 = ((($$111)) + 2|0);
   $51 = HEAP8[$34>>0]|0;
   $52 = $51&255;
   $53 = (($52) + ($17))|0;
   $54 = (($rgb_2_pix) + ($53<<2)|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = (($52) + ($27))|0;
   $57 = (($rgb_2_pix) + ($56<<2)|0);
   $58 = HEAP32[$57>>2]|0;
   $59 = $58 | $55;
   $60 = (($52) + ($31))|0;
   $61 = (($rgb_2_pix) + ($60<<2)|0);
   $62 = HEAP32[$61>>2]|0;
   $63 = $59 | $62;
   $64 = (($49) + ($1<<2)|0);
   HEAP32[$64>>2] = $63;
   HEAP32[$49>>2] = $63;
   $65 = ((($row1$18)) + 8|0);
   $66 = ((($lum2$16)) + 1|0);
   $67 = HEAP8[$lum2$16>>0]|0;
   $68 = $67&255;
   $69 = (($68) + ($17))|0;
   $70 = (($rgb_2_pix) + ($69<<2)|0);
   $71 = HEAP32[$70>>2]|0;
   $72 = (($68) + ($27))|0;
   $73 = (($rgb_2_pix) + ($72<<2)|0);
   $74 = HEAP32[$73>>2]|0;
   $75 = $74 | $71;
   $76 = (($68) + ($31))|0;
   $77 = (($rgb_2_pix) + ($76<<2)|0);
   $78 = HEAP32[$77>>2]|0;
   $79 = $75 | $78;
   $80 = (($row2$15) + ($1<<2)|0);
   HEAP32[$80>>2] = $79;
   HEAP32[$row2$15>>2] = $79;
   $81 = ((($row2$15)) + 4|0);
   $82 = ((($lum2$16)) + 2|0);
   $83 = HEAP8[$66>>0]|0;
   $84 = $83&255;
   $85 = (($84) + ($17))|0;
   $86 = (($rgb_2_pix) + ($85<<2)|0);
   $87 = HEAP32[$86>>2]|0;
   $88 = (($84) + ($27))|0;
   $89 = (($rgb_2_pix) + ($88<<2)|0);
   $90 = HEAP32[$89>>2]|0;
   $91 = $90 | $87;
   $92 = (($84) + ($31))|0;
   $93 = (($rgb_2_pix) + ($92<<2)|0);
   $94 = HEAP32[$93>>2]|0;
   $95 = $91 | $94;
   $96 = (($81) + ($1<<2)|0);
   HEAP32[$96>>2] = $95;
   HEAP32[$81>>2] = $95;
   $97 = ((($row2$15)) + 8|0);
   $98 = ($12|0)==(0);
   if ($98) {
    break;
   } else {
    $$111 = $50;$$1210 = $32;$$149 = $33;$lum2$16 = $82;$row1$18 = $65;$row2$15 = $97;$x$07 = $12;
   }
  }
  $scevgep34 = (($$022) + ($10)|0);
  $99 = (($scevgep34) + ($cols)|0);
  $scevgep36 = (($$0121) + ($2)|0);
  $100 = (($scevgep33) + ($4<<2)|0);
  $101 = (($scevgep) + ($4<<2)|0);
  $102 = ($11|0)==(0);
  if ($102) {
   break;
  } else {
   $$0121 = $scevgep36;$$022 = $99;$$0320 = $scevgep35;$$in = $11;$row1$019 = $100;$row2$017 = $101;
  }
 }
 return;
}
function _Color24DitherYV12Mod1X($colortab,$rgb_2_pix,$lum,$cr,$cb,$out,$rows,$cols,$mod) {
 $colortab = $colortab|0;
 $rgb_2_pix = $rgb_2_pix|0;
 $lum = $lum|0;
 $cr = $cr|0;
 $cb = $cb|0;
 $out = $out|0;
 $rows = $rows|0;
 $cols = $cols|0;
 $mod = $mod|0;
 var $$0121 = 0, $$022 = 0, $$0320 = 0, $$1$lcssa = 0, $$111 = 0, $$12$lcssa = 0, $$1210 = 0, $$14$lcssa = 0, $$149 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cols$off = 0;
 var $factor = 0, $lum$pn18 = 0, $lum2$0 = 0, $lum2$1$lcssa = 0, $lum2$16 = 0, $row1$019 = 0, $row1$1$lcssa = 0, $row1$18 = 0, $row2$017 = 0, $row2$1$lcssa = 0, $row2$15 = 0, $rows$off = 0, $scevgep = 0, $scevgep31 = 0, $scevgep32 = 0, $scevgep33 = 0, $scevgep34 = 0, $scevgep35 = 0, $x$07 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = (($cols|0) / 2)&-1;
 $factor = $mod << 1;
 $1 = (($factor) + ($cols))|0;
 $2 = ($1*3)|0;
 $rows$off = (($rows) + 1)|0;
 $3 = ($rows$off>>>0)<(3);
 if ($3) {
  return;
 }
 $4 = (($rows|0) / 2)&-1;
 $5 = ($cols*3)|0;
 $6 = (($out) + ($5)|0);
 $7 = ($mod*3)|0;
 $8 = (($6) + ($7)|0);
 $cols$off = (($cols) + 1)|0;
 $9 = ($cols$off>>>0)<(3);
 $10 = ($0*6)|0;
 $11 = $0 << 1;
 $12 = (($11) + ($cols))|0;
 $$0121 = $cr;$$022 = $lum;$$0320 = $cb;$$in = $4;$lum$pn18 = $lum;$row1$019 = $out;$row2$017 = $8;
 while(1) {
  $13 = (($$in) + -1)|0;
  $lum2$0 = (($lum$pn18) + ($cols)|0);
  if ($9) {
   $$1$lcssa = $$022;$$12$lcssa = $$0121;$$14$lcssa = $$0320;$lum2$1$lcssa = $lum2$0;$row1$1$lcssa = $row1$019;$row2$1$lcssa = $row2$017;
  } else {
   $scevgep = (($row2$017) + ($10)|0);
   $scevgep32 = (($row1$019) + ($10)|0);
   $scevgep34 = (($$0320) + ($0)|0);
   $$111 = $$022;$$1210 = $$0121;$$149 = $$0320;$lum2$16 = $lum2$0;$row1$18 = $row1$019;$row2$15 = $row2$017;$x$07 = $0;
   while(1) {
    $14 = (($x$07) + -1)|0;
    $15 = HEAP8[$$1210>>0]|0;
    $16 = $15&255;
    $17 = (($colortab) + ($16<<2)|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = (($18) + 256)|0;
    $20 = $16 | 256;
    $21 = (($colortab) + ($20<<2)|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = (($22) + 1024)|0;
    $24 = HEAP8[$$149>>0]|0;
    $25 = $24&255;
    $26 = $25 | 512;
    $27 = (($colortab) + ($26<<2)|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = (($23) + ($28))|0;
    $30 = $25 | 768;
    $31 = (($colortab) + ($30<<2)|0);
    $32 = HEAP32[$31>>2]|0;
    $33 = (($32) + 1792)|0;
    $34 = ((($$1210)) + 1|0);
    $35 = ((($$149)) + 1|0);
    $36 = ((($$111)) + 1|0);
    $37 = HEAP8[$$111>>0]|0;
    $38 = $37&255;
    $39 = (($38) + ($19))|0;
    $40 = (($rgb_2_pix) + ($39<<2)|0);
    $41 = HEAP32[$40>>2]|0;
    $42 = (($38) + ($29))|0;
    $43 = (($rgb_2_pix) + ($42<<2)|0);
    $44 = HEAP32[$43>>2]|0;
    $45 = $44 | $41;
    $46 = (($38) + ($33))|0;
    $47 = (($rgb_2_pix) + ($46<<2)|0);
    $48 = HEAP32[$47>>2]|0;
    $49 = $45 | $48;
    $50 = $49&255;
    $51 = ((($row1$18)) + 1|0);
    HEAP8[$row1$18>>0] = $50;
    $52 = $49 >>> 8;
    $53 = $52&255;
    $54 = ((($row1$18)) + 2|0);
    HEAP8[$51>>0] = $53;
    $55 = $49 >>> 16;
    $56 = $55&255;
    $57 = ((($row1$18)) + 3|0);
    HEAP8[$54>>0] = $56;
    $58 = ((($$111)) + 2|0);
    $59 = HEAP8[$36>>0]|0;
    $60 = $59&255;
    $61 = (($60) + ($19))|0;
    $62 = (($rgb_2_pix) + ($61<<2)|0);
    $63 = HEAP32[$62>>2]|0;
    $64 = (($60) + ($29))|0;
    $65 = (($rgb_2_pix) + ($64<<2)|0);
    $66 = HEAP32[$65>>2]|0;
    $67 = $66 | $63;
    $68 = (($60) + ($33))|0;
    $69 = (($rgb_2_pix) + ($68<<2)|0);
    $70 = HEAP32[$69>>2]|0;
    $71 = $67 | $70;
    $72 = $71&255;
    $73 = ((($row1$18)) + 4|0);
    HEAP8[$57>>0] = $72;
    $74 = $71 >>> 8;
    $75 = $74&255;
    $76 = ((($row1$18)) + 5|0);
    HEAP8[$73>>0] = $75;
    $77 = $71 >>> 16;
    $78 = $77&255;
    $79 = ((($row1$18)) + 6|0);
    HEAP8[$76>>0] = $78;
    $80 = ((($lum2$16)) + 1|0);
    $81 = HEAP8[$lum2$16>>0]|0;
    $82 = $81&255;
    $83 = (($82) + ($19))|0;
    $84 = (($rgb_2_pix) + ($83<<2)|0);
    $85 = HEAP32[$84>>2]|0;
    $86 = (($82) + ($29))|0;
    $87 = (($rgb_2_pix) + ($86<<2)|0);
    $88 = HEAP32[$87>>2]|0;
    $89 = $88 | $85;
    $90 = (($82) + ($33))|0;
    $91 = (($rgb_2_pix) + ($90<<2)|0);
    $92 = HEAP32[$91>>2]|0;
    $93 = $89 | $92;
    $94 = $93&255;
    $95 = ((($row2$15)) + 1|0);
    HEAP8[$row2$15>>0] = $94;
    $96 = $93 >>> 8;
    $97 = $96&255;
    $98 = ((($row2$15)) + 2|0);
    HEAP8[$95>>0] = $97;
    $99 = $93 >>> 16;
    $100 = $99&255;
    $101 = ((($row2$15)) + 3|0);
    HEAP8[$98>>0] = $100;
    $102 = ((($lum2$16)) + 2|0);
    $103 = HEAP8[$80>>0]|0;
    $104 = $103&255;
    $105 = (($104) + ($19))|0;
    $106 = (($rgb_2_pix) + ($105<<2)|0);
    $107 = HEAP32[$106>>2]|0;
    $108 = (($104) + ($29))|0;
    $109 = (($rgb_2_pix) + ($108<<2)|0);
    $110 = HEAP32[$109>>2]|0;
    $111 = $110 | $107;
    $112 = (($104) + ($33))|0;
    $113 = (($rgb_2_pix) + ($112<<2)|0);
    $114 = HEAP32[$113>>2]|0;
    $115 = $111 | $114;
    $116 = $115&255;
    $117 = ((($row2$15)) + 4|0);
    HEAP8[$101>>0] = $116;
    $118 = $115 >>> 8;
    $119 = $118&255;
    $120 = ((($row2$15)) + 5|0);
    HEAP8[$117>>0] = $119;
    $121 = $115 >>> 16;
    $122 = $121&255;
    $123 = ((($row2$15)) + 6|0);
    HEAP8[$120>>0] = $122;
    $124 = ($14|0)==(0);
    if ($124) {
     break;
    } else {
     $$111 = $58;$$1210 = $34;$$149 = $35;$lum2$16 = $102;$row1$18 = $79;$row2$15 = $123;$x$07 = $14;
    }
   }
   $scevgep31 = (($lum$pn18) + ($12)|0);
   $scevgep33 = (($$022) + ($11)|0);
   $scevgep35 = (($$0121) + ($0)|0);
   $$1$lcssa = $scevgep33;$$12$lcssa = $scevgep35;$$14$lcssa = $scevgep34;$lum2$1$lcssa = $scevgep31;$row1$1$lcssa = $scevgep32;$row2$1$lcssa = $scevgep;
  }
  $125 = (($$1$lcssa) + ($cols)|0);
  $126 = (($row1$1$lcssa) + ($2)|0);
  $127 = (($row2$1$lcssa) + ($2)|0);
  $128 = ($13|0)==(0);
  if ($128) {
   break;
  } else {
   $$0121 = $$12$lcssa;$$022 = $125;$$0320 = $$14$lcssa;$$in = $13;$lum$pn18 = $lum2$1$lcssa;$row1$019 = $126;$row2$017 = $127;
  }
 }
 return;
}
function _Color24DitherYV12Mod2X($colortab,$rgb_2_pix,$lum,$cr,$cb,$out,$rows,$cols,$mod) {
 $colortab = $colortab|0;
 $rgb_2_pix = $rgb_2_pix|0;
 $lum = $lum|0;
 $cr = $cr|0;
 $cb = $cb|0;
 $out = $out|0;
 $rows = $rows|0;
 $cols = $cols|0;
 $mod = $mod|0;
 var $$0121 = 0, $$022 = 0, $$0320 = 0, $$1$lcssa = 0, $$111 = 0, $$12$lcssa = 0, $$1210 = 0, $$14$lcssa = 0, $$149 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $cols$off = 0, $lum$pn18 = 0, $lum2$0 = 0, $lum2$1$lcssa = 0, $lum2$16 = 0, $row1$019 = 0, $row1$1$lcssa = 0, $row1$18 = 0, $row2$017 = 0, $row2$1$lcssa = 0, $row2$15 = 0, $rows$off = 0, $scevgep = 0, $scevgep31 = 0, $scevgep32 = 0, $scevgep33 = 0, $scevgep34 = 0, $scevgep35 = 0;
 var $x$07 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $cols << 1;
 $1 = (($0) + ($mod))|0;
 $2 = ($1*3)|0;
 $3 = (($cols|0) / 2)&-1;
 $4 = ($1*9)|0;
 $5 = ($mod*3)|0;
 $6 = (($4) + ($5))|0;
 $rows$off = (($rows) + 1)|0;
 $7 = ($rows$off>>>0)<(3);
 if ($7) {
  return;
 }
 $8 = (($rows|0) / 2)&-1;
 $9 = ($1*6)|0;
 $10 = (($out) + ($9)|0);
 $cols$off = (($cols) + 1)|0;
 $11 = ($cols$off>>>0)<(3);
 $12 = (($2) + 3)|0;
 $13 = (($2) + 4)|0;
 $14 = (($2) + 1)|0;
 $15 = (($2) + 5)|0;
 $16 = (($2) + 2)|0;
 $17 = ($3*12)|0;
 $18 = $3 << 1;
 $19 = (($18) + ($cols))|0;
 $$0121 = $cr;$$022 = $lum;$$0320 = $cb;$$in = $8;$lum$pn18 = $lum;$row1$019 = $out;$row2$017 = $10;
 while(1) {
  $20 = (($$in) + -1)|0;
  $lum2$0 = (($lum$pn18) + ($cols)|0);
  if ($11) {
   $$1$lcssa = $$022;$$12$lcssa = $$0121;$$14$lcssa = $$0320;$lum2$1$lcssa = $lum2$0;$row1$1$lcssa = $row1$019;$row2$1$lcssa = $row2$017;
  } else {
   $scevgep = (($row2$017) + ($17)|0);
   $scevgep32 = (($row1$019) + ($17)|0);
   $scevgep34 = (($$0320) + ($3)|0);
   $$111 = $$022;$$1210 = $$0121;$$149 = $$0320;$lum2$16 = $lum2$0;$row1$18 = $row1$019;$row2$15 = $row2$017;$x$07 = $3;
   while(1) {
    $21 = (($x$07) + -1)|0;
    $22 = HEAP8[$$1210>>0]|0;
    $23 = $22&255;
    $24 = (($colortab) + ($23<<2)|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = (($25) + 256)|0;
    $27 = $23 | 256;
    $28 = (($colortab) + ($27<<2)|0);
    $29 = HEAP32[$28>>2]|0;
    $30 = (($29) + 1024)|0;
    $31 = HEAP8[$$149>>0]|0;
    $32 = $31&255;
    $33 = $32 | 512;
    $34 = (($colortab) + ($33<<2)|0);
    $35 = HEAP32[$34>>2]|0;
    $36 = (($30) + ($35))|0;
    $37 = $32 | 768;
    $38 = (($colortab) + ($37<<2)|0);
    $39 = HEAP32[$38>>2]|0;
    $40 = (($39) + 1792)|0;
    $41 = ((($$1210)) + 1|0);
    $42 = ((($$149)) + 1|0);
    $43 = ((($$111)) + 1|0);
    $44 = HEAP8[$$111>>0]|0;
    $45 = $44&255;
    $46 = (($45) + ($26))|0;
    $47 = (($rgb_2_pix) + ($46<<2)|0);
    $48 = HEAP32[$47>>2]|0;
    $49 = (($45) + ($36))|0;
    $50 = (($rgb_2_pix) + ($49<<2)|0);
    $51 = HEAP32[$50>>2]|0;
    $52 = $51 | $48;
    $53 = (($45) + ($40))|0;
    $54 = (($rgb_2_pix) + ($53<<2)|0);
    $55 = HEAP32[$54>>2]|0;
    $56 = $52 | $55;
    $57 = $56&255;
    $58 = (($row1$18) + ($12)|0);
    HEAP8[$58>>0] = $57;
    $59 = (($row1$18) + ($2)|0);
    HEAP8[$59>>0] = $57;
    $60 = ((($row1$18)) + 3|0);
    HEAP8[$60>>0] = $57;
    HEAP8[$row1$18>>0] = $57;
    $61 = $56 >>> 8;
    $62 = $61&255;
    $63 = (($row1$18) + ($13)|0);
    HEAP8[$63>>0] = $62;
    $64 = (($row1$18) + ($14)|0);
    HEAP8[$64>>0] = $62;
    $65 = ((($row1$18)) + 4|0);
    HEAP8[$65>>0] = $62;
    $66 = ((($row1$18)) + 1|0);
    HEAP8[$66>>0] = $62;
    $67 = $56 >>> 16;
    $68 = $67&255;
    $69 = (($row1$18) + ($15)|0);
    HEAP8[$69>>0] = $68;
    $70 = (($row1$18) + ($16)|0);
    HEAP8[$70>>0] = $68;
    $71 = ((($row1$18)) + 5|0);
    HEAP8[$71>>0] = $68;
    $72 = ((($row1$18)) + 2|0);
    HEAP8[$72>>0] = $68;
    $73 = ((($row1$18)) + 6|0);
    $74 = ((($$111)) + 2|0);
    $75 = HEAP8[$43>>0]|0;
    $76 = $75&255;
    $77 = (($76) + ($26))|0;
    $78 = (($rgb_2_pix) + ($77<<2)|0);
    $79 = HEAP32[$78>>2]|0;
    $80 = (($76) + ($36))|0;
    $81 = (($rgb_2_pix) + ($80<<2)|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = $82 | $79;
    $84 = (($76) + ($40))|0;
    $85 = (($rgb_2_pix) + ($84<<2)|0);
    $86 = HEAP32[$85>>2]|0;
    $87 = $83 | $86;
    $88 = $87&255;
    $89 = (($73) + ($12)|0);
    HEAP8[$89>>0] = $88;
    $90 = (($73) + ($2)|0);
    HEAP8[$90>>0] = $88;
    $91 = ((($row1$18)) + 9|0);
    HEAP8[$91>>0] = $88;
    HEAP8[$73>>0] = $88;
    $92 = $87 >>> 8;
    $93 = $92&255;
    $94 = (($73) + ($13)|0);
    HEAP8[$94>>0] = $93;
    $95 = (($73) + ($14)|0);
    HEAP8[$95>>0] = $93;
    $96 = ((($row1$18)) + 10|0);
    HEAP8[$96>>0] = $93;
    $97 = ((($row1$18)) + 7|0);
    HEAP8[$97>>0] = $93;
    $98 = $87 >>> 16;
    $99 = $98&255;
    $100 = (($73) + ($15)|0);
    HEAP8[$100>>0] = $99;
    $101 = (($73) + ($16)|0);
    HEAP8[$101>>0] = $99;
    $102 = ((($row1$18)) + 11|0);
    HEAP8[$102>>0] = $99;
    $103 = ((($row1$18)) + 8|0);
    HEAP8[$103>>0] = $99;
    $104 = ((($row1$18)) + 12|0);
    $105 = ((($lum2$16)) + 1|0);
    $106 = HEAP8[$lum2$16>>0]|0;
    $107 = $106&255;
    $108 = (($107) + ($26))|0;
    $109 = (($rgb_2_pix) + ($108<<2)|0);
    $110 = HEAP32[$109>>2]|0;
    $111 = (($107) + ($36))|0;
    $112 = (($rgb_2_pix) + ($111<<2)|0);
    $113 = HEAP32[$112>>2]|0;
    $114 = $113 | $110;
    $115 = (($107) + ($40))|0;
    $116 = (($rgb_2_pix) + ($115<<2)|0);
    $117 = HEAP32[$116>>2]|0;
    $118 = $114 | $117;
    $119 = $118&255;
    $120 = (($row2$15) + ($12)|0);
    HEAP8[$120>>0] = $119;
    $121 = (($row2$15) + ($2)|0);
    HEAP8[$121>>0] = $119;
    $122 = ((($row2$15)) + 3|0);
    HEAP8[$122>>0] = $119;
    HEAP8[$row2$15>>0] = $119;
    $123 = $118 >>> 8;
    $124 = $123&255;
    $125 = (($row2$15) + ($13)|0);
    HEAP8[$125>>0] = $124;
    $126 = (($row2$15) + ($14)|0);
    HEAP8[$126>>0] = $124;
    $127 = ((($row2$15)) + 4|0);
    HEAP8[$127>>0] = $124;
    $128 = ((($row2$15)) + 1|0);
    HEAP8[$128>>0] = $124;
    $129 = $118 >>> 16;
    $130 = $129&255;
    $131 = (($row2$15) + ($15)|0);
    HEAP8[$131>>0] = $130;
    $132 = (($row2$15) + ($16)|0);
    HEAP8[$132>>0] = $130;
    $133 = ((($row2$15)) + 5|0);
    HEAP8[$133>>0] = $130;
    $134 = ((($row2$15)) + 2|0);
    HEAP8[$134>>0] = $130;
    $135 = ((($row2$15)) + 6|0);
    $136 = ((($lum2$16)) + 2|0);
    $137 = HEAP8[$105>>0]|0;
    $138 = $137&255;
    $139 = (($138) + ($26))|0;
    $140 = (($rgb_2_pix) + ($139<<2)|0);
    $141 = HEAP32[$140>>2]|0;
    $142 = (($138) + ($36))|0;
    $143 = (($rgb_2_pix) + ($142<<2)|0);
    $144 = HEAP32[$143>>2]|0;
    $145 = $144 | $141;
    $146 = (($138) + ($40))|0;
    $147 = (($rgb_2_pix) + ($146<<2)|0);
    $148 = HEAP32[$147>>2]|0;
    $149 = $145 | $148;
    $150 = $149&255;
    $151 = (($135) + ($12)|0);
    HEAP8[$151>>0] = $150;
    $152 = (($135) + ($2)|0);
    HEAP8[$152>>0] = $150;
    $153 = ((($row2$15)) + 9|0);
    HEAP8[$153>>0] = $150;
    HEAP8[$135>>0] = $150;
    $154 = $149 >>> 8;
    $155 = $154&255;
    $156 = (($135) + ($13)|0);
    HEAP8[$156>>0] = $155;
    $157 = (($135) + ($14)|0);
    HEAP8[$157>>0] = $155;
    $158 = ((($row2$15)) + 10|0);
    HEAP8[$158>>0] = $155;
    $159 = ((($row2$15)) + 7|0);
    HEAP8[$159>>0] = $155;
    $160 = $149 >>> 16;
    $161 = $160&255;
    $162 = (($135) + ($15)|0);
    HEAP8[$162>>0] = $161;
    $163 = (($135) + ($16)|0);
    HEAP8[$163>>0] = $161;
    $164 = ((($row2$15)) + 11|0);
    HEAP8[$164>>0] = $161;
    $165 = ((($row2$15)) + 8|0);
    HEAP8[$165>>0] = $161;
    $166 = ((($row2$15)) + 12|0);
    $167 = ($21|0)==(0);
    if ($167) {
     break;
    } else {
     $$111 = $74;$$1210 = $41;$$149 = $42;$lum2$16 = $136;$row1$18 = $104;$row2$15 = $166;$x$07 = $21;
    }
   }
   $scevgep31 = (($lum$pn18) + ($19)|0);
   $scevgep33 = (($$022) + ($18)|0);
   $scevgep35 = (($$0121) + ($3)|0);
   $$1$lcssa = $scevgep33;$$12$lcssa = $scevgep35;$$14$lcssa = $scevgep34;$lum2$1$lcssa = $scevgep31;$row1$1$lcssa = $scevgep32;$row2$1$lcssa = $scevgep;
  }
  $168 = (($$1$lcssa) + ($cols)|0);
  $169 = (($row1$1$lcssa) + ($6)|0);
  $170 = (($row2$1$lcssa) + ($6)|0);
  $171 = ($20|0)==(0);
  if ($171) {
   break;
  } else {
   $$0121 = $$12$lcssa;$$022 = $168;$$0320 = $$14$lcssa;$$in = $20;$lum$pn18 = $lum2$1$lcssa;$row1$019 = $169;$row2$017 = $170;
  }
 }
 return;
}
function _Color32DitherYV12Mod1X($colortab,$rgb_2_pix,$lum,$cr,$cb,$out,$rows,$cols,$mod) {
 $colortab = $colortab|0;
 $rgb_2_pix = $rgb_2_pix|0;
 $lum = $lum|0;
 $cr = $cr|0;
 $cb = $cb|0;
 $out = $out|0;
 $rows = $rows|0;
 $cols = $cols|0;
 $mod = $mod|0;
 var $$0121 = 0, $$022 = 0, $$0320 = 0, $$111 = 0, $$1210 = 0, $$149 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $cols$off = 0, $factor = 0, $lum2$0 = 0, $lum2$16 = 0, $or$cond = 0, $row1$019 = 0, $row1$18 = 0, $row2$017 = 0, $row2$15 = 0, $rows$off = 0, $scevgep = 0, $scevgep33 = 0, $scevgep34 = 0, $scevgep35 = 0, $scevgep36 = 0, $x$07 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = (($cols|0) / 2)&-1;
 $factor = $mod << 1;
 $1 = (($factor) + ($cols))|0;
 $rows$off = (($rows) + 1)|0;
 $2 = ($rows$off>>>0)<(3);
 $cols$off = (($cols) + 1)|0;
 $3 = ($cols$off>>>0)<(3);
 $or$cond = $2 | $3;
 if ($or$cond) {
  return;
 }
 $4 = (($rows|0) / 2)&-1;
 $5 = (($out) + ($cols<<2)|0);
 $6 = (($5) + ($mod<<2)|0);
 $7 = $0 << 1;
 $$0121 = $cr;$$022 = $lum;$$0320 = $cb;$$in = $4;$row1$019 = $out;$row2$017 = $6;
 while(1) {
  $8 = (($$in) + -1)|0;
  $lum2$0 = (($$022) + ($cols)|0);
  $scevgep = (($row2$017) + ($7<<2)|0);
  $scevgep33 = (($row1$019) + ($7<<2)|0);
  $scevgep35 = (($$0320) + ($0)|0);
  $$111 = $$022;$$1210 = $$0121;$$149 = $$0320;$lum2$16 = $lum2$0;$row1$18 = $row1$019;$row2$15 = $row2$017;$x$07 = $0;
  while(1) {
   $9 = (($x$07) + -1)|0;
   $10 = HEAP8[$$1210>>0]|0;
   $11 = $10&255;
   $12 = (($colortab) + ($11<<2)|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = (($13) + 256)|0;
   $15 = $11 | 256;
   $16 = (($colortab) + ($15<<2)|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = (($17) + 1024)|0;
   $19 = HEAP8[$$149>>0]|0;
   $20 = $19&255;
   $21 = $20 | 512;
   $22 = (($colortab) + ($21<<2)|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = (($18) + ($23))|0;
   $25 = $20 | 768;
   $26 = (($colortab) + ($25<<2)|0);
   $27 = HEAP32[$26>>2]|0;
   $28 = (($27) + 1792)|0;
   $29 = ((($$1210)) + 1|0);
   $30 = ((($$149)) + 1|0);
   $31 = ((($$111)) + 1|0);
   $32 = HEAP8[$$111>>0]|0;
   $33 = $32&255;
   $34 = (($33) + ($14))|0;
   $35 = (($rgb_2_pix) + ($34<<2)|0);
   $36 = HEAP32[$35>>2]|0;
   $37 = (($33) + ($24))|0;
   $38 = (($rgb_2_pix) + ($37<<2)|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = $39 | $36;
   $41 = (($33) + ($28))|0;
   $42 = (($rgb_2_pix) + ($41<<2)|0);
   $43 = HEAP32[$42>>2]|0;
   $44 = $40 | $43;
   $45 = ((($row1$18)) + 4|0);
   HEAP32[$row1$18>>2] = $44;
   $46 = ((($$111)) + 2|0);
   $47 = HEAP8[$31>>0]|0;
   $48 = $47&255;
   $49 = (($48) + ($14))|0;
   $50 = (($rgb_2_pix) + ($49<<2)|0);
   $51 = HEAP32[$50>>2]|0;
   $52 = (($48) + ($24))|0;
   $53 = (($rgb_2_pix) + ($52<<2)|0);
   $54 = HEAP32[$53>>2]|0;
   $55 = $54 | $51;
   $56 = (($48) + ($28))|0;
   $57 = (($rgb_2_pix) + ($56<<2)|0);
   $58 = HEAP32[$57>>2]|0;
   $59 = $55 | $58;
   $60 = ((($row1$18)) + 8|0);
   HEAP32[$45>>2] = $59;
   $61 = ((($lum2$16)) + 1|0);
   $62 = HEAP8[$lum2$16>>0]|0;
   $63 = $62&255;
   $64 = (($63) + ($14))|0;
   $65 = (($rgb_2_pix) + ($64<<2)|0);
   $66 = HEAP32[$65>>2]|0;
   $67 = (($63) + ($24))|0;
   $68 = (($rgb_2_pix) + ($67<<2)|0);
   $69 = HEAP32[$68>>2]|0;
   $70 = $69 | $66;
   $71 = (($63) + ($28))|0;
   $72 = (($rgb_2_pix) + ($71<<2)|0);
   $73 = HEAP32[$72>>2]|0;
   $74 = $70 | $73;
   $75 = ((($row2$15)) + 4|0);
   HEAP32[$row2$15>>2] = $74;
   $76 = ((($lum2$16)) + 2|0);
   $77 = HEAP8[$61>>0]|0;
   $78 = $77&255;
   $79 = (($78) + ($14))|0;
   $80 = (($rgb_2_pix) + ($79<<2)|0);
   $81 = HEAP32[$80>>2]|0;
   $82 = (($78) + ($24))|0;
   $83 = (($rgb_2_pix) + ($82<<2)|0);
   $84 = HEAP32[$83>>2]|0;
   $85 = $84 | $81;
   $86 = (($78) + ($28))|0;
   $87 = (($rgb_2_pix) + ($86<<2)|0);
   $88 = HEAP32[$87>>2]|0;
   $89 = $85 | $88;
   $90 = ((($row2$15)) + 8|0);
   HEAP32[$75>>2] = $89;
   $91 = ($9|0)==(0);
   if ($91) {
    break;
   } else {
    $$111 = $46;$$1210 = $29;$$149 = $30;$lum2$16 = $76;$row1$18 = $60;$row2$15 = $90;$x$07 = $9;
   }
  }
  $scevgep34 = (($$022) + ($7)|0);
  $92 = (($scevgep34) + ($cols)|0);
  $scevgep36 = (($$0121) + ($0)|0);
  $93 = (($scevgep33) + ($1<<2)|0);
  $94 = (($scevgep) + ($1<<2)|0);
  $95 = ($8|0)==(0);
  if ($95) {
   break;
  } else {
   $$0121 = $scevgep36;$$022 = $92;$$0320 = $scevgep35;$$in = $8;$row1$019 = $93;$row2$017 = $94;
  }
 }
 return;
}
function _Color32DitherYV12Mod2X($colortab,$rgb_2_pix,$lum,$cr,$cb,$out,$rows,$cols,$mod) {
 $colortab = $colortab|0;
 $rgb_2_pix = $rgb_2_pix|0;
 $lum = $lum|0;
 $cr = $cr|0;
 $cb = $cb|0;
 $out = $out|0;
 $rows = $rows|0;
 $cols = $cols|0;
 $mod = $mod|0;
 var $$0121 = 0, $$022 = 0, $$0320 = 0, $$1$lcssa = 0, $$111 = 0, $$12$lcssa = 0, $$1210 = 0, $$14$lcssa = 0, $$149 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cols$off = 0, $lum$pn18 = 0, $lum2$0 = 0, $lum2$1$lcssa = 0, $lum2$16 = 0, $row1$019 = 0, $row1$1$lcssa = 0, $row1$18 = 0, $row2$017 = 0, $row2$1$lcssa = 0, $row2$15 = 0, $rows$off = 0, $scevgep = 0, $scevgep31 = 0, $scevgep32 = 0, $scevgep33 = 0;
 var $scevgep34 = 0, $scevgep35 = 0, $x$07 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $cols << 1;
 $1 = (($0) + ($mod))|0;
 $2 = (($cols|0) / 2)&-1;
 $3 = ($1*3)|0;
 $4 = (($3) + ($mod))|0;
 $rows$off = (($rows) + 1)|0;
 $5 = ($rows$off>>>0)<(3);
 if ($5) {
  return;
 }
 $6 = (($rows|0) / 2)&-1;
 $7 = $1 << 1;
 $8 = (($out) + ($7<<2)|0);
 $cols$off = (($cols) + 1)|0;
 $9 = ($cols$off>>>0)<(3);
 $10 = (($1) + 1)|0;
 $11 = $2 << 2;
 $12 = $2 << 1;
 $13 = (($12) + ($cols))|0;
 $$0121 = $cr;$$022 = $lum;$$0320 = $cb;$$in = $6;$lum$pn18 = $lum;$row1$019 = $out;$row2$017 = $8;
 while(1) {
  $14 = (($$in) + -1)|0;
  $lum2$0 = (($lum$pn18) + ($cols)|0);
  if ($9) {
   $$1$lcssa = $$022;$$12$lcssa = $$0121;$$14$lcssa = $$0320;$lum2$1$lcssa = $lum2$0;$row1$1$lcssa = $row1$019;$row2$1$lcssa = $row2$017;
  } else {
   $scevgep = (($row2$017) + ($11<<2)|0);
   $scevgep32 = (($row1$019) + ($11<<2)|0);
   $scevgep34 = (($$0320) + ($2)|0);
   $$111 = $$022;$$1210 = $$0121;$$149 = $$0320;$lum2$16 = $lum2$0;$row1$18 = $row1$019;$row2$15 = $row2$017;$x$07 = $2;
   while(1) {
    $15 = (($x$07) + -1)|0;
    $16 = HEAP8[$$1210>>0]|0;
    $17 = $16&255;
    $18 = (($colortab) + ($17<<2)|0);
    $19 = HEAP32[$18>>2]|0;
    $20 = (($19) + 256)|0;
    $21 = $17 | 256;
    $22 = (($colortab) + ($21<<2)|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = (($23) + 1024)|0;
    $25 = HEAP8[$$149>>0]|0;
    $26 = $25&255;
    $27 = $26 | 512;
    $28 = (($colortab) + ($27<<2)|0);
    $29 = HEAP32[$28>>2]|0;
    $30 = (($24) + ($29))|0;
    $31 = $26 | 768;
    $32 = (($colortab) + ($31<<2)|0);
    $33 = HEAP32[$32>>2]|0;
    $34 = (($33) + 1792)|0;
    $35 = ((($$1210)) + 1|0);
    $36 = ((($$149)) + 1|0);
    $37 = ((($$111)) + 1|0);
    $38 = HEAP8[$$111>>0]|0;
    $39 = $38&255;
    $40 = (($39) + ($20))|0;
    $41 = (($rgb_2_pix) + ($40<<2)|0);
    $42 = HEAP32[$41>>2]|0;
    $43 = (($39) + ($30))|0;
    $44 = (($rgb_2_pix) + ($43<<2)|0);
    $45 = HEAP32[$44>>2]|0;
    $46 = $45 | $42;
    $47 = (($39) + ($34))|0;
    $48 = (($rgb_2_pix) + ($47<<2)|0);
    $49 = HEAP32[$48>>2]|0;
    $50 = $46 | $49;
    $51 = (($row1$18) + ($10<<2)|0);
    HEAP32[$51>>2] = $50;
    $52 = (($row1$18) + ($1<<2)|0);
    HEAP32[$52>>2] = $50;
    $53 = ((($row1$18)) + 4|0);
    HEAP32[$53>>2] = $50;
    HEAP32[$row1$18>>2] = $50;
    $54 = ((($row1$18)) + 8|0);
    $55 = ((($$111)) + 2|0);
    $56 = HEAP8[$37>>0]|0;
    $57 = $56&255;
    $58 = (($57) + ($20))|0;
    $59 = (($rgb_2_pix) + ($58<<2)|0);
    $60 = HEAP32[$59>>2]|0;
    $61 = (($57) + ($30))|0;
    $62 = (($rgb_2_pix) + ($61<<2)|0);
    $63 = HEAP32[$62>>2]|0;
    $64 = $63 | $60;
    $65 = (($57) + ($34))|0;
    $66 = (($rgb_2_pix) + ($65<<2)|0);
    $67 = HEAP32[$66>>2]|0;
    $68 = $64 | $67;
    $69 = (($54) + ($10<<2)|0);
    HEAP32[$69>>2] = $68;
    $70 = (($54) + ($1<<2)|0);
    HEAP32[$70>>2] = $68;
    $71 = ((($row1$18)) + 12|0);
    HEAP32[$71>>2] = $68;
    HEAP32[$54>>2] = $68;
    $72 = ((($row1$18)) + 16|0);
    $73 = ((($lum2$16)) + 1|0);
    $74 = HEAP8[$lum2$16>>0]|0;
    $75 = $74&255;
    $76 = (($75) + ($20))|0;
    $77 = (($rgb_2_pix) + ($76<<2)|0);
    $78 = HEAP32[$77>>2]|0;
    $79 = (($75) + ($30))|0;
    $80 = (($rgb_2_pix) + ($79<<2)|0);
    $81 = HEAP32[$80>>2]|0;
    $82 = $81 | $78;
    $83 = (($75) + ($34))|0;
    $84 = (($rgb_2_pix) + ($83<<2)|0);
    $85 = HEAP32[$84>>2]|0;
    $86 = $82 | $85;
    $87 = (($row2$15) + ($10<<2)|0);
    HEAP32[$87>>2] = $86;
    $88 = (($row2$15) + ($1<<2)|0);
    HEAP32[$88>>2] = $86;
    $89 = ((($row2$15)) + 4|0);
    HEAP32[$89>>2] = $86;
    HEAP32[$row2$15>>2] = $86;
    $90 = ((($row2$15)) + 8|0);
    $91 = ((($lum2$16)) + 2|0);
    $92 = HEAP8[$73>>0]|0;
    $93 = $92&255;
    $94 = (($93) + ($20))|0;
    $95 = (($rgb_2_pix) + ($94<<2)|0);
    $96 = HEAP32[$95>>2]|0;
    $97 = (($93) + ($30))|0;
    $98 = (($rgb_2_pix) + ($97<<2)|0);
    $99 = HEAP32[$98>>2]|0;
    $100 = $99 | $96;
    $101 = (($93) + ($34))|0;
    $102 = (($rgb_2_pix) + ($101<<2)|0);
    $103 = HEAP32[$102>>2]|0;
    $104 = $100 | $103;
    $105 = (($90) + ($10<<2)|0);
    HEAP32[$105>>2] = $104;
    $106 = (($90) + ($1<<2)|0);
    HEAP32[$106>>2] = $104;
    $107 = ((($row2$15)) + 12|0);
    HEAP32[$107>>2] = $104;
    HEAP32[$90>>2] = $104;
    $108 = ((($row2$15)) + 16|0);
    $109 = ($15|0)==(0);
    if ($109) {
     break;
    } else {
     $$111 = $55;$$1210 = $35;$$149 = $36;$lum2$16 = $91;$row1$18 = $72;$row2$15 = $108;$x$07 = $15;
    }
   }
   $scevgep31 = (($lum$pn18) + ($13)|0);
   $scevgep33 = (($$022) + ($12)|0);
   $scevgep35 = (($$0121) + ($2)|0);
   $$1$lcssa = $scevgep33;$$12$lcssa = $scevgep35;$$14$lcssa = $scevgep34;$lum2$1$lcssa = $scevgep31;$row1$1$lcssa = $scevgep32;$row2$1$lcssa = $scevgep;
  }
  $110 = (($$1$lcssa) + ($cols)|0);
  $111 = (($row1$1$lcssa) + ($4<<2)|0);
  $112 = (($row2$1$lcssa) + ($4<<2)|0);
  $113 = ($14|0)==(0);
  if ($113) {
   break;
  } else {
   $$0121 = $$12$lcssa;$$022 = $110;$$0320 = $$14$lcssa;$$in = $14;$lum$pn18 = $lum2$1$lcssa;$row1$019 = $111;$row2$017 = $112;
  }
 }
 return;
}
function _Color16DitherYUY2Mod1X($colortab,$rgb_2_pix,$lum,$cr,$cb,$out,$rows,$cols,$mod) {
 $colortab = $colortab|0;
 $rgb_2_pix = $rgb_2_pix|0;
 $lum = $lum|0;
 $cr = $cr|0;
 $cb = $cb|0;
 $out = $out|0;
 $rows = $rows|0;
 $cols = $cols|0;
 $mod = $mod|0;
 var $$0115 = 0, $$016 = 0, $$0314 = 0, $$128 = 0, $$147 = 0, $$19 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $7 = 0, $8 = 0, $9 = 0, $cols$off = 0, $or$cond = 0, $row$013 = 0, $row$16 = 0, $scevgep = 0, $scevgep22 = 0, $scevgep23 = 0, $scevgep24 = 0, $x$05 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = (($cols|0) / 2)&-1;
 $1 = ($rows|0)==(0);
 $cols$off = (($cols) + 1)|0;
 $2 = ($cols$off>>>0)<(3);
 $or$cond = $1 | $2;
 if ($or$cond) {
  return;
 }
 $3 = $0 << 1;
 $4 = $0 << 2;
 $$0115 = $cr;$$016 = $lum;$$0314 = $cb;$$in = $rows;$row$013 = $out;
 while(1) {
  $5 = (($$in) + -1)|0;
  $scevgep = (($row$013) + ($3<<1)|0);
  $scevgep23 = (($$0314) + ($4)|0);
  $$128 = $$0115;$$147 = $$0314;$$19 = $$016;$row$16 = $row$013;$x$05 = $0;
  while(1) {
   $6 = (($x$05) + -1)|0;
   $7 = HEAP8[$$128>>0]|0;
   $8 = $7&255;
   $9 = (($colortab) + ($8<<2)|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = (($10) + 256)|0;
   $12 = $8 | 256;
   $13 = (($colortab) + ($12<<2)|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = (($14) + 1024)|0;
   $16 = HEAP8[$$147>>0]|0;
   $17 = $16&255;
   $18 = $17 | 512;
   $19 = (($colortab) + ($18<<2)|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = (($15) + ($20))|0;
   $22 = $17 | 768;
   $23 = (($colortab) + ($22<<2)|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = (($24) + 1792)|0;
   $26 = ((($$128)) + 4|0);
   $27 = ((($$147)) + 4|0);
   $28 = HEAP8[$$19>>0]|0;
   $29 = $28&255;
   $30 = ((($$19)) + 2|0);
   $31 = (($29) + ($11))|0;
   $32 = (($rgb_2_pix) + ($31<<2)|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = (($29) + ($21))|0;
   $35 = (($rgb_2_pix) + ($34<<2)|0);
   $36 = HEAP32[$35>>2]|0;
   $37 = $36 | $33;
   $38 = (($29) + ($25))|0;
   $39 = (($rgb_2_pix) + ($38<<2)|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = $37 | $40;
   $42 = $41&65535;
   $43 = ((($row$16)) + 2|0);
   HEAP16[$row$16>>1] = $42;
   $44 = HEAP8[$30>>0]|0;
   $45 = $44&255;
   $46 = ((($$19)) + 4|0);
   $47 = (($45) + ($11))|0;
   $48 = (($rgb_2_pix) + ($47<<2)|0);
   $49 = HEAP32[$48>>2]|0;
   $50 = (($45) + ($21))|0;
   $51 = (($rgb_2_pix) + ($50<<2)|0);
   $52 = HEAP32[$51>>2]|0;
   $53 = $52 | $49;
   $54 = (($45) + ($25))|0;
   $55 = (($rgb_2_pix) + ($54<<2)|0);
   $56 = HEAP32[$55>>2]|0;
   $57 = $53 | $56;
   $58 = $57&65535;
   $59 = ((($row$16)) + 4|0);
   HEAP16[$43>>1] = $58;
   $60 = ($6|0)==(0);
   if ($60) {
    break;
   } else {
    $$128 = $26;$$147 = $27;$$19 = $46;$row$16 = $59;$x$05 = $6;
   }
  }
  $scevgep22 = (($$016) + ($4)|0);
  $scevgep24 = (($$0115) + ($4)|0);
  $61 = (($scevgep) + ($mod<<1)|0);
  $62 = ($5|0)==(0);
  if ($62) {
   break;
  } else {
   $$0115 = $scevgep24;$$016 = $scevgep22;$$0314 = $scevgep23;$$in = $5;$row$013 = $61;
  }
 }
 return;
}
function _Color16DitherYUY2Mod2X($colortab,$rgb_2_pix,$lum,$cr,$cb,$out,$rows,$cols,$mod) {
 $colortab = $colortab|0;
 $rgb_2_pix = $rgb_2_pix|0;
 $lum = $lum|0;
 $cr = $cr|0;
 $cb = $cb|0;
 $out = $out|0;
 $rows = $rows|0;
 $cols = $cols|0;
 $mod = $mod|0;
 var $$0115 = 0, $$016 = 0, $$0314 = 0, $$128 = 0, $$147 = 0, $$19 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $cols$off = 0, $or$cond = 0, $row$013 = 0, $row$16 = 0, $scevgep = 0, $scevgep22 = 0, $scevgep23 = 0, $scevgep24 = 0;
 var $x$05 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($mod|0) / 2)&-1;
 $1 = (($0) + ($cols))|0;
 $2 = (($cols|0) / 2)&-1;
 $3 = ($rows|0)==(0);
 $cols$off = (($cols) + 1)|0;
 $4 = ($cols$off>>>0)<(3);
 $or$cond = $3 | $4;
 if ($or$cond) {
  return;
 }
 $5 = $2 << 1;
 $6 = $2 << 2;
 $$0115 = $cr;$$016 = $lum;$$0314 = $cb;$$in = $rows;$row$013 = $out;
 while(1) {
  $7 = (($$in) + -1)|0;
  $scevgep = (($row$013) + ($5<<2)|0);
  $scevgep23 = (($$0314) + ($6)|0);
  $$128 = $$0115;$$147 = $$0314;$$19 = $$016;$row$16 = $row$013;$x$05 = $2;
  while(1) {
   $8 = (($x$05) + -1)|0;
   $9 = HEAP8[$$128>>0]|0;
   $10 = $9&255;
   $11 = (($colortab) + ($10<<2)|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = (($12) + 256)|0;
   $14 = $10 | 256;
   $15 = (($colortab) + ($14<<2)|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = (($16) + 1024)|0;
   $18 = HEAP8[$$147>>0]|0;
   $19 = $18&255;
   $20 = $19 | 512;
   $21 = (($colortab) + ($20<<2)|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = (($17) + ($22))|0;
   $24 = $19 | 768;
   $25 = (($colortab) + ($24<<2)|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = (($26) + 1792)|0;
   $28 = ((($$128)) + 4|0);
   $29 = ((($$147)) + 4|0);
   $30 = HEAP8[$$19>>0]|0;
   $31 = $30&255;
   $32 = ((($$19)) + 2|0);
   $33 = (($31) + ($13))|0;
   $34 = (($rgb_2_pix) + ($33<<2)|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = (($31) + ($23))|0;
   $37 = (($rgb_2_pix) + ($36<<2)|0);
   $38 = HEAP32[$37>>2]|0;
   $39 = $38 | $35;
   $40 = (($31) + ($27))|0;
   $41 = (($rgb_2_pix) + ($40<<2)|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = $39 | $42;
   $44 = (($row$16) + ($1<<2)|0);
   HEAP32[$44>>2] = $43;
   HEAP32[$row$16>>2] = $43;
   $45 = ((($row$16)) + 4|0);
   $46 = HEAP8[$32>>0]|0;
   $47 = $46&255;
   $48 = ((($$19)) + 4|0);
   $49 = (($47) + ($13))|0;
   $50 = (($rgb_2_pix) + ($49<<2)|0);
   $51 = HEAP32[$50>>2]|0;
   $52 = (($47) + ($23))|0;
   $53 = (($rgb_2_pix) + ($52<<2)|0);
   $54 = HEAP32[$53>>2]|0;
   $55 = $54 | $51;
   $56 = (($47) + ($27))|0;
   $57 = (($rgb_2_pix) + ($56<<2)|0);
   $58 = HEAP32[$57>>2]|0;
   $59 = $55 | $58;
   $60 = (($45) + ($1<<2)|0);
   HEAP32[$60>>2] = $59;
   HEAP32[$45>>2] = $59;
   $61 = ((($row$16)) + 8|0);
   $62 = ($8|0)==(0);
   if ($62) {
    break;
   } else {
    $$128 = $28;$$147 = $29;$$19 = $48;$row$16 = $61;$x$05 = $8;
   }
  }
  $scevgep22 = (($$016) + ($6)|0);
  $scevgep24 = (($$0115) + ($6)|0);
  $63 = (($scevgep) + ($1<<2)|0);
  $64 = ($7|0)==(0);
  if ($64) {
   break;
  } else {
   $$0115 = $scevgep24;$$016 = $scevgep22;$$0314 = $scevgep23;$$in = $7;$row$013 = $63;
  }
 }
 return;
}
function _Color24DitherYUY2Mod1X($colortab,$rgb_2_pix,$lum,$cr,$cb,$out,$rows,$cols,$mod) {
 $colortab = $colortab|0;
 $rgb_2_pix = $rgb_2_pix|0;
 $lum = $lum|0;
 $cr = $cr|0;
 $cb = $cb|0;
 $out = $out|0;
 $rows = $rows|0;
 $cols = $cols|0;
 $mod = $mod|0;
 var $$0115 = 0, $$016 = 0, $$0314 = 0, $$128 = 0, $$147 = 0, $$19 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $8 = 0, $9 = 0, $cols$off = 0, $or$cond = 0, $row$013 = 0, $row$16 = 0, $scevgep = 0, $scevgep22 = 0, $scevgep23 = 0, $scevgep24 = 0, $x$05 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($cols|0) / 2)&-1;
 $1 = ($mod*3)|0;
 $2 = ($rows|0)==(0);
 $cols$off = (($cols) + 1)|0;
 $3 = ($cols$off>>>0)<(3);
 $or$cond = $2 | $3;
 if ($or$cond) {
  return;
 }
 $4 = ($0*6)|0;
 $5 = $0 << 2;
 $$0115 = $cr;$$016 = $lum;$$0314 = $cb;$$in = $rows;$row$013 = $out;
 while(1) {
  $6 = (($$in) + -1)|0;
  $scevgep = (($row$013) + ($4)|0);
  $scevgep23 = (($$0314) + ($5)|0);
  $$128 = $$0115;$$147 = $$0314;$$19 = $$016;$row$16 = $row$013;$x$05 = $0;
  while(1) {
   $7 = (($x$05) + -1)|0;
   $8 = HEAP8[$$128>>0]|0;
   $9 = $8&255;
   $10 = (($colortab) + ($9<<2)|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = (($11) + 256)|0;
   $13 = $9 | 256;
   $14 = (($colortab) + ($13<<2)|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = (($15) + 1024)|0;
   $17 = HEAP8[$$147>>0]|0;
   $18 = $17&255;
   $19 = $18 | 512;
   $20 = (($colortab) + ($19<<2)|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = (($16) + ($21))|0;
   $23 = $18 | 768;
   $24 = (($colortab) + ($23<<2)|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = (($25) + 1792)|0;
   $27 = ((($$128)) + 4|0);
   $28 = ((($$147)) + 4|0);
   $29 = HEAP8[$$19>>0]|0;
   $30 = $29&255;
   $31 = ((($$19)) + 2|0);
   $32 = (($30) + ($12))|0;
   $33 = (($rgb_2_pix) + ($32<<2)|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = (($30) + ($22))|0;
   $36 = (($rgb_2_pix) + ($35<<2)|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = $37 | $34;
   $39 = (($30) + ($26))|0;
   $40 = (($rgb_2_pix) + ($39<<2)|0);
   $41 = HEAP32[$40>>2]|0;
   $42 = $38 | $41;
   $43 = $42&255;
   $44 = ((($row$16)) + 1|0);
   HEAP8[$row$16>>0] = $43;
   $45 = $42 >>> 8;
   $46 = $45&255;
   $47 = ((($row$16)) + 2|0);
   HEAP8[$44>>0] = $46;
   $48 = $42 >>> 16;
   $49 = $48&255;
   $50 = ((($row$16)) + 3|0);
   HEAP8[$47>>0] = $49;
   $51 = HEAP8[$31>>0]|0;
   $52 = $51&255;
   $53 = ((($$19)) + 4|0);
   $54 = (($52) + ($12))|0;
   $55 = (($rgb_2_pix) + ($54<<2)|0);
   $56 = HEAP32[$55>>2]|0;
   $57 = (($52) + ($22))|0;
   $58 = (($rgb_2_pix) + ($57<<2)|0);
   $59 = HEAP32[$58>>2]|0;
   $60 = $59 | $56;
   $61 = (($52) + ($26))|0;
   $62 = (($rgb_2_pix) + ($61<<2)|0);
   $63 = HEAP32[$62>>2]|0;
   $64 = $60 | $63;
   $65 = $64&255;
   $66 = ((($row$16)) + 4|0);
   HEAP8[$50>>0] = $65;
   $67 = $64 >>> 8;
   $68 = $67&255;
   $69 = ((($row$16)) + 5|0);
   HEAP8[$66>>0] = $68;
   $70 = $64 >>> 16;
   $71 = $70&255;
   $72 = ((($row$16)) + 6|0);
   HEAP8[$69>>0] = $71;
   $73 = ($7|0)==(0);
   if ($73) {
    break;
   } else {
    $$128 = $27;$$147 = $28;$$19 = $53;$row$16 = $72;$x$05 = $7;
   }
  }
  $scevgep22 = (($$016) + ($5)|0);
  $scevgep24 = (($$0115) + ($5)|0);
  $74 = (($scevgep) + ($1)|0);
  $75 = ($6|0)==(0);
  if ($75) {
   break;
  } else {
   $$0115 = $scevgep24;$$016 = $scevgep22;$$0314 = $scevgep23;$$in = $6;$row$013 = $74;
  }
 }
 return;
}
function _Color24DitherYUY2Mod2X($colortab,$rgb_2_pix,$lum,$cr,$cb,$out,$rows,$cols,$mod) {
 $colortab = $colortab|0;
 $rgb_2_pix = $rgb_2_pix|0;
 $lum = $lum|0;
 $cr = $cr|0;
 $cb = $cb|0;
 $out = $out|0;
 $rows = $rows|0;
 $cols = $cols|0;
 $mod = $mod|0;
 var $$0115 = 0, $$016 = 0, $$0314 = 0, $$128 = 0, $$147 = 0, $$19 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cols$off = 0, $row$013 = 0, $row$16 = 0, $scevgep = 0, $scevgep22 = 0, $scevgep23 = 0, $scevgep24 = 0, $x$05 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $cols << 1;
 $1 = (($0) + ($mod))|0;
 $2 = ($1*3)|0;
 $3 = (($cols|0) / 2)&-1;
 $4 = ($rows|0)==(0);
 if ($4) {
  return;
 }
 $cols$off = (($cols) + 1)|0;
 $5 = ($cols$off>>>0)<(3);
 $6 = (($2) + 3)|0;
 $7 = (($2) + 4)|0;
 $8 = (($2) + 1)|0;
 $9 = (($2) + 5)|0;
 $10 = (($2) + 2)|0;
 if ($5) {
  return;
 }
 $11 = ($3*12)|0;
 $12 = $3 << 2;
 $$0115 = $cr;$$016 = $lum;$$0314 = $cb;$$in = $rows;$row$013 = $out;
 while(1) {
  $13 = (($$in) + -1)|0;
  $scevgep = (($row$013) + ($11)|0);
  $scevgep23 = (($$0314) + ($12)|0);
  $$128 = $$0115;$$147 = $$0314;$$19 = $$016;$row$16 = $row$013;$x$05 = $3;
  while(1) {
   $14 = (($x$05) + -1)|0;
   $15 = HEAP8[$$128>>0]|0;
   $16 = $15&255;
   $17 = (($colortab) + ($16<<2)|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = (($18) + 256)|0;
   $20 = $16 | 256;
   $21 = (($colortab) + ($20<<2)|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = (($22) + 1024)|0;
   $24 = HEAP8[$$147>>0]|0;
   $25 = $24&255;
   $26 = $25 | 512;
   $27 = (($colortab) + ($26<<2)|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = (($23) + ($28))|0;
   $30 = $25 | 768;
   $31 = (($colortab) + ($30<<2)|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = (($32) + 1792)|0;
   $34 = ((($$128)) + 4|0);
   $35 = ((($$147)) + 4|0);
   $36 = HEAP8[$$19>>0]|0;
   $37 = $36&255;
   $38 = ((($$19)) + 2|0);
   $39 = (($37) + ($19))|0;
   $40 = (($rgb_2_pix) + ($39<<2)|0);
   $41 = HEAP32[$40>>2]|0;
   $42 = (($37) + ($29))|0;
   $43 = (($rgb_2_pix) + ($42<<2)|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = $44 | $41;
   $46 = (($37) + ($33))|0;
   $47 = (($rgb_2_pix) + ($46<<2)|0);
   $48 = HEAP32[$47>>2]|0;
   $49 = $45 | $48;
   $50 = $49&255;
   $51 = (($row$16) + ($6)|0);
   HEAP8[$51>>0] = $50;
   $52 = (($row$16) + ($2)|0);
   HEAP8[$52>>0] = $50;
   $53 = ((($row$16)) + 3|0);
   HEAP8[$53>>0] = $50;
   HEAP8[$row$16>>0] = $50;
   $54 = $49 >>> 8;
   $55 = $54&255;
   $56 = (($row$16) + ($7)|0);
   HEAP8[$56>>0] = $55;
   $57 = (($row$16) + ($8)|0);
   HEAP8[$57>>0] = $55;
   $58 = ((($row$16)) + 4|0);
   HEAP8[$58>>0] = $55;
   $59 = ((($row$16)) + 1|0);
   HEAP8[$59>>0] = $55;
   $60 = $49 >>> 16;
   $61 = $60&255;
   $62 = (($row$16) + ($9)|0);
   HEAP8[$62>>0] = $61;
   $63 = (($row$16) + ($10)|0);
   HEAP8[$63>>0] = $61;
   $64 = ((($row$16)) + 5|0);
   HEAP8[$64>>0] = $61;
   $65 = ((($row$16)) + 2|0);
   HEAP8[$65>>0] = $61;
   $66 = ((($row$16)) + 6|0);
   $67 = HEAP8[$38>>0]|0;
   $68 = $67&255;
   $69 = ((($$19)) + 4|0);
   $70 = (($68) + ($19))|0;
   $71 = (($rgb_2_pix) + ($70<<2)|0);
   $72 = HEAP32[$71>>2]|0;
   $73 = (($68) + ($29))|0;
   $74 = (($rgb_2_pix) + ($73<<2)|0);
   $75 = HEAP32[$74>>2]|0;
   $76 = $75 | $72;
   $77 = (($68) + ($33))|0;
   $78 = (($rgb_2_pix) + ($77<<2)|0);
   $79 = HEAP32[$78>>2]|0;
   $80 = $76 | $79;
   $81 = $80&255;
   $82 = (($66) + ($6)|0);
   HEAP8[$82>>0] = $81;
   $83 = (($66) + ($2)|0);
   HEAP8[$83>>0] = $81;
   $84 = ((($row$16)) + 9|0);
   HEAP8[$84>>0] = $81;
   HEAP8[$66>>0] = $81;
   $85 = $80 >>> 8;
   $86 = $85&255;
   $87 = (($66) + ($7)|0);
   HEAP8[$87>>0] = $86;
   $88 = (($66) + ($8)|0);
   HEAP8[$88>>0] = $86;
   $89 = ((($row$16)) + 10|0);
   HEAP8[$89>>0] = $86;
   $90 = ((($row$16)) + 7|0);
   HEAP8[$90>>0] = $86;
   $91 = $80 >>> 16;
   $92 = $91&255;
   $93 = (($66) + ($9)|0);
   HEAP8[$93>>0] = $92;
   $94 = (($66) + ($10)|0);
   HEAP8[$94>>0] = $92;
   $95 = ((($row$16)) + 11|0);
   HEAP8[$95>>0] = $92;
   $96 = ((($row$16)) + 8|0);
   HEAP8[$96>>0] = $92;
   $97 = ((($row$16)) + 12|0);
   $98 = ($14|0)==(0);
   if ($98) {
    break;
   } else {
    $$128 = $34;$$147 = $35;$$19 = $69;$row$16 = $97;$x$05 = $14;
   }
  }
  $scevgep22 = (($$016) + ($12)|0);
  $scevgep24 = (($$0115) + ($12)|0);
  $99 = (($scevgep) + ($2)|0);
  $100 = ($13|0)==(0);
  if ($100) {
   break;
  } else {
   $$0115 = $scevgep24;$$016 = $scevgep22;$$0314 = $scevgep23;$$in = $13;$row$013 = $99;
  }
 }
 return;
}
function _Color32DitherYUY2Mod1X($colortab,$rgb_2_pix,$lum,$cr,$cb,$out,$rows,$cols,$mod) {
 $colortab = $colortab|0;
 $rgb_2_pix = $rgb_2_pix|0;
 $lum = $lum|0;
 $cr = $cr|0;
 $cb = $cb|0;
 $out = $out|0;
 $rows = $rows|0;
 $cols = $cols|0;
 $mod = $mod|0;
 var $$0115 = 0, $$016 = 0, $$0314 = 0, $$128 = 0, $$147 = 0, $$19 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0, $9 = 0, $cols$off = 0, $or$cond = 0, $row$013 = 0, $row$16 = 0, $scevgep = 0, $scevgep22 = 0, $scevgep23 = 0, $scevgep24 = 0, $x$05 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($cols|0) / 2)&-1;
 $1 = ($rows|0)==(0);
 $cols$off = (($cols) + 1)|0;
 $2 = ($cols$off>>>0)<(3);
 $or$cond = $1 | $2;
 if ($or$cond) {
  return;
 }
 $3 = $0 << 1;
 $4 = $0 << 2;
 $$0115 = $cr;$$016 = $lum;$$0314 = $cb;$$in = $rows;$row$013 = $out;
 while(1) {
  $5 = (($$in) + -1)|0;
  $scevgep = (($row$013) + ($3<<2)|0);
  $scevgep23 = (($$0314) + ($4)|0);
  $$128 = $$0115;$$147 = $$0314;$$19 = $$016;$row$16 = $row$013;$x$05 = $0;
  while(1) {
   $6 = (($x$05) + -1)|0;
   $7 = HEAP8[$$128>>0]|0;
   $8 = $7&255;
   $9 = (($colortab) + ($8<<2)|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = (($10) + 256)|0;
   $12 = $8 | 256;
   $13 = (($colortab) + ($12<<2)|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = (($14) + 1024)|0;
   $16 = HEAP8[$$147>>0]|0;
   $17 = $16&255;
   $18 = $17 | 512;
   $19 = (($colortab) + ($18<<2)|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = (($15) + ($20))|0;
   $22 = $17 | 768;
   $23 = (($colortab) + ($22<<2)|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = (($24) + 1792)|0;
   $26 = ((($$128)) + 4|0);
   $27 = ((($$147)) + 4|0);
   $28 = HEAP8[$$19>>0]|0;
   $29 = $28&255;
   $30 = ((($$19)) + 2|0);
   $31 = (($29) + ($11))|0;
   $32 = (($rgb_2_pix) + ($31<<2)|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = (($29) + ($21))|0;
   $35 = (($rgb_2_pix) + ($34<<2)|0);
   $36 = HEAP32[$35>>2]|0;
   $37 = $36 | $33;
   $38 = (($29) + ($25))|0;
   $39 = (($rgb_2_pix) + ($38<<2)|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = $37 | $40;
   $42 = ((($row$16)) + 4|0);
   HEAP32[$row$16>>2] = $41;
   $43 = HEAP8[$30>>0]|0;
   $44 = $43&255;
   $45 = ((($$19)) + 4|0);
   $46 = (($44) + ($11))|0;
   $47 = (($rgb_2_pix) + ($46<<2)|0);
   $48 = HEAP32[$47>>2]|0;
   $49 = (($44) + ($21))|0;
   $50 = (($rgb_2_pix) + ($49<<2)|0);
   $51 = HEAP32[$50>>2]|0;
   $52 = $51 | $48;
   $53 = (($44) + ($25))|0;
   $54 = (($rgb_2_pix) + ($53<<2)|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = $52 | $55;
   $57 = ((($row$16)) + 8|0);
   HEAP32[$42>>2] = $56;
   $58 = ($6|0)==(0);
   if ($58) {
    break;
   } else {
    $$128 = $26;$$147 = $27;$$19 = $45;$row$16 = $57;$x$05 = $6;
   }
  }
  $scevgep22 = (($$016) + ($4)|0);
  $scevgep24 = (($$0115) + ($4)|0);
  $59 = (($scevgep) + ($mod<<2)|0);
  $60 = ($5|0)==(0);
  if ($60) {
   break;
  } else {
   $$0115 = $scevgep24;$$016 = $scevgep22;$$0314 = $scevgep23;$$in = $5;$row$013 = $59;
  }
 }
 return;
}
function _Color32DitherYUY2Mod2X($colortab,$rgb_2_pix,$lum,$cr,$cb,$out,$rows,$cols,$mod) {
 $colortab = $colortab|0;
 $rgb_2_pix = $rgb_2_pix|0;
 $lum = $lum|0;
 $cr = $cr|0;
 $cb = $cb|0;
 $out = $out|0;
 $rows = $rows|0;
 $cols = $cols|0;
 $mod = $mod|0;
 var $$0115 = 0, $$016 = 0, $$0314 = 0, $$128 = 0, $$147 = 0, $$19 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $cols$off = 0, $row$013 = 0, $row$16 = 0, $scevgep = 0;
 var $scevgep22 = 0, $scevgep23 = 0, $scevgep24 = 0, $x$05 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $cols << 1;
 $1 = (($0) + ($mod))|0;
 $2 = (($cols|0) / 2)&-1;
 $3 = ($rows|0)==(0);
 if ($3) {
  return;
 }
 $cols$off = (($cols) + 1)|0;
 $4 = ($cols$off>>>0)<(3);
 $5 = (($1) + 1)|0;
 if ($4) {
  return;
 }
 $6 = $2 << 2;
 $$0115 = $cr;$$016 = $lum;$$0314 = $cb;$$in = $rows;$row$013 = $out;
 while(1) {
  $7 = (($$in) + -1)|0;
  $scevgep = (($row$013) + ($6<<2)|0);
  $scevgep23 = (($$0314) + ($6)|0);
  $$128 = $$0115;$$147 = $$0314;$$19 = $$016;$row$16 = $row$013;$x$05 = $2;
  while(1) {
   $8 = (($x$05) + -1)|0;
   $9 = HEAP8[$$128>>0]|0;
   $10 = $9&255;
   $11 = (($colortab) + ($10<<2)|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = (($12) + 256)|0;
   $14 = $10 | 256;
   $15 = (($colortab) + ($14<<2)|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = (($16) + 1024)|0;
   $18 = HEAP8[$$147>>0]|0;
   $19 = $18&255;
   $20 = $19 | 512;
   $21 = (($colortab) + ($20<<2)|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = (($17) + ($22))|0;
   $24 = $19 | 768;
   $25 = (($colortab) + ($24<<2)|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = (($26) + 1792)|0;
   $28 = ((($$128)) + 4|0);
   $29 = ((($$147)) + 4|0);
   $30 = HEAP8[$$19>>0]|0;
   $31 = $30&255;
   $32 = ((($$19)) + 2|0);
   $33 = (($31) + ($13))|0;
   $34 = (($rgb_2_pix) + ($33<<2)|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = (($31) + ($23))|0;
   $37 = (($rgb_2_pix) + ($36<<2)|0);
   $38 = HEAP32[$37>>2]|0;
   $39 = $38 | $35;
   $40 = (($31) + ($27))|0;
   $41 = (($rgb_2_pix) + ($40<<2)|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = $39 | $42;
   $44 = (($row$16) + ($5<<2)|0);
   HEAP32[$44>>2] = $43;
   $45 = (($row$16) + ($1<<2)|0);
   HEAP32[$45>>2] = $43;
   $46 = ((($row$16)) + 4|0);
   HEAP32[$46>>2] = $43;
   HEAP32[$row$16>>2] = $43;
   $47 = ((($row$16)) + 8|0);
   $48 = HEAP8[$32>>0]|0;
   $49 = $48&255;
   $50 = ((($$19)) + 4|0);
   $51 = (($49) + ($13))|0;
   $52 = (($rgb_2_pix) + ($51<<2)|0);
   $53 = HEAP32[$52>>2]|0;
   $54 = (($49) + ($23))|0;
   $55 = (($rgb_2_pix) + ($54<<2)|0);
   $56 = HEAP32[$55>>2]|0;
   $57 = $56 | $53;
   $58 = (($49) + ($27))|0;
   $59 = (($rgb_2_pix) + ($58<<2)|0);
   $60 = HEAP32[$59>>2]|0;
   $61 = $57 | $60;
   $62 = (($47) + ($5<<2)|0);
   HEAP32[$62>>2] = $61;
   $63 = (($47) + ($1<<2)|0);
   HEAP32[$63>>2] = $61;
   $64 = ((($row$16)) + 12|0);
   HEAP32[$64>>2] = $61;
   HEAP32[$47>>2] = $61;
   $65 = ((($row$16)) + 16|0);
   $66 = ($8|0)==(0);
   if ($66) {
    break;
   } else {
    $$128 = $28;$$147 = $29;$$19 = $50;$row$16 = $65;$x$05 = $8;
   }
  }
  $scevgep22 = (($$016) + ($6)|0);
  $scevgep24 = (($$0115) + ($6)|0);
  $67 = (($scevgep) + ($1<<2)|0);
  $68 = ($7|0)==(0);
  if ($68) {
   break;
  } else {
   $$0115 = $scevgep24;$$016 = $scevgep22;$$0314 = $scevgep23;$$in = $7;$row$013 = $67;
  }
 }
 return;
}
function _SDL_UnlockTexture($texture) {
 $texture = $texture|0;
 var $$mask$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $native_pitch$i = 0, $native_pitch$i2 = 0, $native_pixels$i = 0, $native_pixels$i1 = 0, $or$cond$i = 0, $phitmp$i = 0, $rect$i = 0, $switch$split2D = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $native_pixels$i1 = sp + 36|0;
 $native_pitch$i2 = sp + 32|0;
 $native_pixels$i = sp + 28|0;
 $native_pitch$i = sp + 24|0;
 $rect$i = sp + 8|0;
 $0 = ($texture|0)==(0|0);
 if (!($0)) {
  $1 = HEAP32[$texture>>2]|0;
  $2 = ($1|0)==(1279300|0);
  if ($2) {
   $3 = ((($texture)) + 8|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==(1);
   if (!($5)) {
    STACKTOP = sp;return;
   }
   $6 = ((($texture)) + 40|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0|0);
   $9 = ((($texture)) + 36|0);
   $10 = HEAP32[$9>>2]|0;
   if (!($8)) {
    HEAP32[$native_pixels$i>>2] = 0;
    HEAP32[$native_pitch$i>>2] = 0;
    HEAP32[$rect$i>>2] = 0;
    $11 = ((($rect$i)) + 4|0);
    HEAP32[$11>>2] = 0;
    $12 = ((($texture)) + 12|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($rect$i)) + 8|0);
    HEAP32[$14>>2] = $13;
    $15 = ((($texture)) + 16|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($rect$i)) + 12|0);
    HEAP32[$17>>2] = $16;
    $18 = (_SDL_LockTexture($10,$rect$i,$native_pixels$i,$native_pitch$i)|0);
    $19 = ($18|0)<(0);
    if (!($19)) {
     $20 = HEAP32[$6>>2]|0;
     $21 = ((($10)) + 4|0);
     $22 = HEAP32[$21>>2]|0;
     $23 = HEAP32[$14>>2]|0;
     $24 = HEAP32[$17>>2]|0;
     $25 = HEAP32[$native_pixels$i>>2]|0;
     $26 = HEAP32[$native_pitch$i>>2]|0;
     (_SDL_SW_CopyYUVToRGB($20,$rect$i,$22,$23,$24,$25,$26)|0);
     _SDL_UnlockTexture($10);
    }
    STACKTOP = sp;return;
   }
   $27 = ($10|0)==(0|0);
   if ($27) {
    $58 = ((($texture)) + 32|0);
    $59 = HEAP32[$58>>2]|0;
    $60 = ((($59)) + 40|0);
    $61 = HEAP32[$60>>2]|0;
    FUNCTION_TABLE_vii[$61 & 8191]($59,$texture);
    STACKTOP = sp;return;
   }
   HEAP32[$native_pixels$i1>>2] = 0;
   HEAP32[$native_pitch$i2>>2] = 0;
   $28 = ((($texture)) + 52|0);
   $29 = ((($texture)) + 44|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = ((($texture)) + 56|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = ((($texture)) + 48|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = Math_imul($34, $32)|0;
   $36 = (($30) + ($35)|0);
   $37 = HEAP32[$28>>2]|0;
   $38 = ((($texture)) + 4|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = ($39|0)==(0);
   $$mask$i = $39 & -268435456;
   $41 = ($$mask$i|0)==(268435456);
   $or$cond$i = $40 | $41;
   L18: do {
    if ($or$cond$i) {
     $43 = $39 & 255;
     $47 = $43;
    } else {
     $switch$split2D = ($39|0)<(1498831189);
     if ($switch$split2D) {
      switch ($39|0) {
      case 844715353:  {
       $47 = 2;
       break L18;
       break;
      }
      default: {
      }
      }
     } else {
      switch ($39|0) {
      case 1498831189:  {
       $47 = 2;
       break L18;
       break;
      }
      default: {
      }
      }
     }
     $42 = ($39|0)==(1431918169);
     $phitmp$i = $42 ? 2 : 1;
     $47 = $phitmp$i;
    }
   } while(0);
   $44 = (_SDL_LockTexture($10,$28,$native_pixels$i1,$native_pitch$i2)|0);
   $45 = ($44|0)<(0);
   if (!($45)) {
    $46 = Math_imul($47, $37)|0;
    $48 = (($36) + ($46)|0);
    $49 = ((($texture)) + 60|0);
    $50 = HEAP32[$49>>2]|0;
    $51 = ((($texture)) + 64|0);
    $52 = HEAP32[$51>>2]|0;
    $53 = HEAP32[$38>>2]|0;
    $54 = ((($10)) + 4|0);
    $55 = HEAP32[$54>>2]|0;
    $56 = HEAP32[$native_pixels$i1>>2]|0;
    $57 = HEAP32[$native_pitch$i2>>2]|0;
    (_SDL_ConvertPixels($50,$52,$53,$48,$34,$55,$56,$57)|0);
    _SDL_UnlockTexture($10);
   }
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(597654,$vararg_buffer)|0);
 STACKTOP = sp;return;
}
function _SDL_RenderCopy($renderer,$texture,$srcrect,$dstrect) {
 $renderer = $renderer|0;
 $texture = $texture|0;
 $srcrect = $srcrect|0;
 $dstrect = $dstrect|0;
 var $$0 = 0, $$pre = 0, $$pre$phi4Z2D = 0, $$pre$phiZ2D = 0, $$pre3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0.0, $23 = 0, $24 = 0.0, $25 = 0.0, $26 = 0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0.0, $32 = 0.0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0.0, $38 = 0.0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0.0, $44 = 0.0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0.0, $58 = 0;
 var $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0, $62 = 0.0, $63 = 0, $64 = 0.0, $65 = 0.0, $66 = 0, $67 = 0, $68 = 0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0, $73 = 0, $74 = 0.0, $75 = 0.0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $9 = 0, $frect = 0, $phitmp = 0, $real_dstrect = 0, $real_srcrect = 0, $texture$ = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $real_srcrect = sp + 64|0;
 $real_dstrect = sp + 48|0;
 $frect = sp + 32|0;
 ;HEAP32[$real_srcrect>>2]=0|0;HEAP32[$real_srcrect+4>>2]=0|0;HEAP32[$real_srcrect+8>>2]=0|0;HEAP32[$real_srcrect+12>>2]=0|0;
 ;HEAP32[$real_dstrect>>2]=0|0;HEAP32[$real_dstrect+4>>2]=0|0;HEAP32[$real_dstrect+8>>2]=0|0;HEAP32[$real_dstrect+12>>2]=0|0;
 $0 = ($renderer|0)==(0|0);
 if (!($0)) {
  $1 = HEAP32[$renderer>>2]|0;
  $2 = ($1|0)==(1279299|0);
  if ($2) {
   $3 = ($texture|0)==(0|0);
   if (!($3)) {
    $4 = HEAP32[$texture>>2]|0;
    $5 = ($4|0)==(1279300|0);
    if ($5) {
     $6 = ((($texture)) + 32|0);
     $7 = HEAP32[$6>>2]|0;
     $8 = ($7|0)==($renderer|0);
     if (!($8)) {
      $9 = (_SDL_SetError(609606,$vararg_buffer3)|0);
      $$0 = $9;
      STACKTOP = sp;return ($$0|0);
     }
     HEAP32[$real_srcrect>>2] = 0;
     $10 = ((($real_srcrect)) + 4|0);
     HEAP32[$10>>2] = 0;
     $11 = ((($texture)) + 12|0);
     $12 = HEAP32[$11>>2]|0;
     $13 = ((($real_srcrect)) + 8|0);
     HEAP32[$13>>2] = $12;
     $14 = ((($texture)) + 16|0);
     $15 = HEAP32[$14>>2]|0;
     $16 = ((($real_srcrect)) + 12|0);
     HEAP32[$16>>2] = $15;
     $17 = ($srcrect|0)==(0|0);
     if ($17) {
      label = 13;
     } else {
      $18 = (_SDL_IntersectRect($srcrect,$real_srcrect,$real_srcrect)|0);
      $19 = ($18|0)==(0);
      if ($19) {
       $$0 = 0;
       STACKTOP = sp;return ($$0|0);
      }
      $$pre = HEAP32[$renderer>>2]|0;
      $phitmp = ($$pre|0)==(1279299|0);
      if ($phitmp) {
       label = 13;
      } else {
       (_SDL_SetError(597637,$vararg_buffer5)|0);
       $$pre3 = ((($real_dstrect)) + 4|0);
       $$pre$phi4Z2D = $$pre3;$$pre$phiZ2D = $real_dstrect;
      }
     }
     if ((label|0) == 13) {
      $20 = ((($renderer)) + 212|0);
      $21 = HEAP32[$20>>2]|0;
      $22 = (+($21|0));
      $23 = ((($renderer)) + 284|0);
      $24 = +HEAPF32[$23>>2];
      $25 = $22 / $24;
      $26 = (~~(($25)));
      HEAP32[$real_dstrect>>2] = $26;
      $27 = ((($renderer)) + 216|0);
      $28 = HEAP32[$27>>2]|0;
      $29 = (+($28|0));
      $30 = ((($renderer)) + 288|0);
      $31 = +HEAPF32[$30>>2];
      $32 = $29 / $31;
      $33 = (~~(($32)));
      $34 = ((($real_dstrect)) + 4|0);
      HEAP32[$34>>2] = $33;
      $35 = ((($renderer)) + 220|0);
      $36 = HEAP32[$35>>2]|0;
      $37 = (+($36|0));
      $38 = $37 / $24;
      $39 = (~~(($38)));
      $40 = ((($real_dstrect)) + 8|0);
      HEAP32[$40>>2] = $39;
      $41 = ((($renderer)) + 224|0);
      $42 = HEAP32[$41>>2]|0;
      $43 = (+($42|0));
      $44 = $43 / $31;
      $45 = (~~(($44)));
      $46 = ((($real_dstrect)) + 12|0);
      HEAP32[$46>>2] = $45;
      $$pre$phi4Z2D = $34;$$pre$phiZ2D = $real_dstrect;
     }
     HEAP32[$$pre$phiZ2D>>2] = 0;
     HEAP32[$$pre$phi4Z2D>>2] = 0;
     $47 = ($dstrect|0)==(0|0);
     do {
      if (!($47)) {
       $48 = (_SDL_HasIntersection($dstrect,$real_dstrect)|0);
       $49 = ($48|0)==(0);
       if ($49) {
        $$0 = 0;
        STACKTOP = sp;return ($$0|0);
       } else {
        ;HEAP32[$real_dstrect>>2]=HEAP32[$dstrect>>2]|0;HEAP32[$real_dstrect+4>>2]=HEAP32[$dstrect+4>>2]|0;HEAP32[$real_dstrect+8>>2]=HEAP32[$dstrect+8>>2]|0;HEAP32[$real_dstrect+12>>2]=HEAP32[$dstrect+12>>2]|0;
        break;
       }
      }
     } while(0);
     $50 = ((($texture)) + 36|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = ((($renderer)) + 192|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ($53|0)==(0);
     if (!($54)) {
      $$0 = 0;
      STACKTOP = sp;return ($$0|0);
     }
     $55 = ($51|0)==(0|0);
     $texture$ = $55 ? $texture : $51;
     $56 = HEAP32[$$pre$phiZ2D>>2]|0;
     $57 = (+($56|0));
     $58 = ((($renderer)) + 284|0);
     $59 = +HEAPF32[$58>>2];
     $60 = $57 * $59;
     HEAPF32[$frect>>2] = $60;
     $61 = HEAP32[$$pre$phi4Z2D>>2]|0;
     $62 = (+($61|0));
     $63 = ((($renderer)) + 288|0);
     $64 = +HEAPF32[$63>>2];
     $65 = $62 * $64;
     $66 = ((($frect)) + 4|0);
     HEAPF32[$66>>2] = $65;
     $67 = ((($real_dstrect)) + 8|0);
     $68 = HEAP32[$67>>2]|0;
     $69 = (+($68|0));
     $70 = $69 * $59;
     $71 = ((($frect)) + 8|0);
     HEAPF32[$71>>2] = $70;
     $72 = ((($real_dstrect)) + 12|0);
     $73 = HEAP32[$72>>2]|0;
     $74 = (+($73|0));
     $75 = $74 * $64;
     $76 = ((($frect)) + 12|0);
     HEAPF32[$76>>2] = $75;
     $77 = ((($renderer)) + 72|0);
     $78 = HEAP32[$77>>2]|0;
     $79 = (FUNCTION_TABLE_iiiii[$78 & 8191]($renderer,$texture$,$real_srcrect,$frect)|0);
     $$0 = $79;
     STACKTOP = sp;return ($$0|0);
    }
   }
   (_SDL_SetError(597654,$vararg_buffer1)|0);
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_SDL_SetError(597637,$vararg_buffer)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_HasIntersection($A,$B) {
 $A = $A|0;
 $B = $B|0;
 var $$ = 0, $$0 = 0, $$1 = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $Amax$0 = 0, $Amax$1 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($A|0)==(0|0);
 do {
  if ($0) {
   HEAP32[$vararg_buffer>>2] = 596897;
   (_SDL_SetError(611267,$vararg_buffer)|0);
   $$0 = 0;
  } else {
   $1 = ($B|0)==(0|0);
   if ($1) {
    HEAP32[$vararg_buffer1>>2] = 596899;
    (_SDL_SetError(611267,$vararg_buffer1)|0);
    $$0 = 0;
    break;
   }
   $2 = ((($A)) + 8|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = ($3|0)<(1);
   if ($4) {
    $$0 = 0;
   } else {
    $5 = ((($A)) + 12|0);
    $6 = HEAP32[$5>>2]|0;
    $7 = ($6|0)<(1);
    if ($7) {
     $$0 = 0;
    } else {
     $8 = ((($B)) + 8|0);
     $9 = HEAP32[$8>>2]|0;
     $10 = ($9|0)<(1);
     if ($10) {
      $$0 = 0;
     } else {
      $11 = ((($B)) + 12|0);
      $12 = HEAP32[$11>>2]|0;
      $13 = ($12|0)<(1);
      if ($13) {
       $$0 = 0;
      } else {
       $14 = HEAP32[$A>>2]|0;
       $15 = (($3) + ($14))|0;
       $16 = HEAP32[$B>>2]|0;
       $17 = (($9) + ($16))|0;
       $18 = ($16|0)>($14|0);
       $$ = $18 ? $16 : $14;
       $19 = ($17|0)<($15|0);
       $Amax$0 = $19 ? $17 : $15;
       $20 = ($Amax$0|0)>($$|0);
       if ($20) {
        $21 = ((($A)) + 4|0);
        $22 = HEAP32[$21>>2]|0;
        $23 = (($6) + ($22))|0;
        $24 = ((($B)) + 4|0);
        $25 = HEAP32[$24>>2]|0;
        $26 = (($12) + ($25))|0;
        $27 = ($25|0)>($22|0);
        $$1 = $27 ? $25 : $22;
        $28 = ($26|0)<($23|0);
        $Amax$1 = $28 ? $26 : $23;
        $29 = ($Amax$1|0)>($$1|0);
        $$2 = $29&1;
        $$0 = $$2;
       } else {
        $$0 = 0;
       }
      }
     }
    }
   }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _SDL_RenderPresent($renderer) {
 $renderer = $renderer|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($renderer|0)==(0|0);
 if (!($0)) {
  $1 = HEAP32[$renderer>>2]|0;
  $2 = ($1|0)==(1279299|0);
  if ($2) {
   $3 = ((($renderer)) + 192|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==(0);
   if (!($5)) {
    STACKTOP = sp;return;
   }
   $6 = ((($renderer)) + 84|0);
   $7 = HEAP32[$6>>2]|0;
   FUNCTION_TABLE_vi[$7 & 8191]($renderer);
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(597637,$vararg_buffer)|0);
 STACKTOP = sp;return;
}
function _SDL_DestroyWindowTexture($unused,$window) {
 $unused = $unused|0;
 $window = $window|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_SetWindowData($window,609391,0)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  return;
 }
 $2 = ((($0)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  _SDL_DestroyTexture($3);
 }
 $5 = HEAP32[$0>>2]|0;
 $6 = ($5|0)==(0|0);
 if (!($6)) {
  _SDL_DestroyRenderer($5);
 }
 $7 = ((($0)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 _SDL_free($8);
 _SDL_free($0);
 return;
}
function _SDL_AudioInit($driver_name) {
 $driver_name = $driver_name|0;
 var $$01 = 0, $$0212 = 0, $$0213 = 0, $$in = 0, $$in37 = 0, $$lcssa = 0, $$lcssa56 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $7 = 0, $8 = 0, $9 = 0, $i$06$ph = 0, $i$06$ph$pn = 0, $i$06$ph$pn$lcssa = 0, $i$06$us$ph = 0, $i$06$us$ph$pn = 0, $i$06$us$ph$pn$lcssa = 0, $tried_to_init$0$lcssa5 = 0;
 var $tried_to_init$07$ph = 0, $tried_to_init$07$us$ph = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_SDL_WasInit(16)|0);
 $1 = ($0|0)==(0);
 if (!($1)) {
  _SDL_AudioQuit();
 }
 (_SDL_memset(1205548,0,112)|0);
 (_SDL_memset(1205876,0,64)|0);
 $2 = ($driver_name|0)==(0|0);
 L4: do {
  if ($2) {
   $3 = (_SDL_getenv(610209)|0);
   $4 = ($3|0)==(0|0);
   if ($4) {
    $i$06$us$ph = 0;$tried_to_init$07$us$ph = 0;
    while(1) {
     $19 = ($i$06$us$ph|0)==(3);
     if ($19) {
      $$0212 = $3;$65 = $4;$tried_to_init$0$lcssa5 = $tried_to_init$07$us$ph;
      label = 14;
      break L4;
     } else {
      $i$06$us$ph$pn = $i$06$us$ph;
     }
     while(1) {
      $$in37 = (140968 + ($i$06$us$ph$pn<<2)|0);
      $5 = HEAP32[$$in37>>2]|0;
      $6 = ((($5)) + 12|0);
      $7 = HEAP32[$6>>2]|0;
      $8 = ($7|0)==(0);
      if ($8) {
       $$lcssa = $5;$i$06$us$ph$pn$lcssa = $i$06$us$ph$pn;
       break;
      }
      $9 = (($i$06$us$ph$pn) + 1)|0;
      $10 = ($9|0)==(3);
      if ($10) {
       $$0212 = $3;$65 = $4;$tried_to_init$0$lcssa5 = $tried_to_init$07$us$ph;
       label = 14;
       break L4;
      } else {
       $i$06$us$ph$pn = $9;
      }
     }
     (_SDL_memset(1205548,0,112)|0);
     $11 = HEAP32[$$lcssa>>2]|0;
     HEAP32[301387] = $11;
     $12 = ((($$lcssa)) + 4|0);
     $13 = HEAP32[$12>>2]|0;
     HEAP32[(1205552)>>2] = $13;
     $14 = ((($$lcssa)) + 8|0);
     $15 = HEAP32[$14>>2]|0;
     $16 = (FUNCTION_TABLE_ii[$15 & 8191]((1205556))|0);
     $17 = (($i$06$us$ph$pn$lcssa) + 1)|0;
     $18 = ($16|0)==(0);
     if ($18) {
      $i$06$us$ph = $17;$tried_to_init$07$us$ph = 1;
     } else {
      break;
     }
    }
   } else {
    $$0213 = $3;
    label = 5;
   }
  } else {
   $$0213 = $driver_name;
   label = 5;
  }
 } while(0);
 L12: do {
  if ((label|0) == 5) {
   $i$06$ph = 0;$tried_to_init$07$ph = 0;
   while(1) {
    $35 = ($i$06$ph|0)==(3);
    if ($35) {
     $$0212 = $$0213;$65 = 0;$tried_to_init$0$lcssa5 = $tried_to_init$07$ph;
     label = 14;
     break L12;
    } else {
     $i$06$ph$pn = $i$06$ph;
    }
    while(1) {
     $$in = (140968 + ($i$06$ph$pn<<2)|0);
     $20 = HEAP32[$$in>>2]|0;
     $21 = HEAP32[$20>>2]|0;
     $22 = (_SDL_strlen($$0213)|0);
     $23 = (_SDL_strncasecmp($21,$$0213,$22)|0);
     $24 = ($23|0)==(0);
     if ($24) {
      $$lcssa56 = $20;$i$06$ph$pn$lcssa = $i$06$ph$pn;
      break;
     }
     $25 = (($i$06$ph$pn) + 1)|0;
     $26 = ($25|0)==(3);
     if ($26) {
      $$0212 = $$0213;$65 = 0;$tried_to_init$0$lcssa5 = $tried_to_init$07$ph;
      label = 14;
      break L12;
     } else {
      $i$06$ph$pn = $25;
     }
    }
    (_SDL_memset(1205548,0,112)|0);
    $27 = HEAP32[$$lcssa56>>2]|0;
    HEAP32[301387] = $27;
    $28 = ((($$lcssa56)) + 4|0);
    $29 = HEAP32[$28>>2]|0;
    HEAP32[(1205552)>>2] = $29;
    $30 = ((($$lcssa56)) + 8|0);
    $31 = HEAP32[$30>>2]|0;
    $32 = (FUNCTION_TABLE_ii[$31 & 8191]((1205556))|0);
    $33 = (($i$06$ph$pn$lcssa) + 1)|0;
    $34 = ($32|0)==(0);
    if ($34) {
     $i$06$ph = $33;$tried_to_init$07$ph = 1;
    } else {
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 14) {
  $36 = ($tried_to_init$0$lcssa5|0)==(0);
  do {
   if ($36) {
    if ($65) {
     (_SDL_SetError(610257,$vararg_buffer1)|0);
     break;
    } else {
     HEAP32[$vararg_buffer>>2] = $$0212;
     (_SDL_SetError(610225,$vararg_buffer)|0);
     break;
    }
   }
  } while(0);
  (_SDL_memset(1205548,0,112)|0);
  $$01 = -1;
  STACKTOP = sp;return ($$01|0);
 }
 $37 = (_SDL_CreateMutex()|0);
 HEAP32[(1205632)>>2] = $37;
 $38 = HEAP32[(1205556)>>2]|0;
 $39 = ($38|0)==(0|0);
 if ($39) {
  HEAP32[(1205556)>>2] = 4663;
  $64 = 4663;
 } else {
  $64 = $38;
 }
 $40 = HEAP32[(1205560)>>2]|0;
 $41 = ($40|0)==(0|0);
 if ($41) {
  HEAP32[(1205560)>>2] = 4664;
 }
 $42 = HEAP32[(1205564)>>2]|0;
 $43 = ($42|0)==(0|0);
 if ($43) {
  HEAP32[(1205564)>>2] = 4665;
 }
 $44 = HEAP32[(1205568)>>2]|0;
 $45 = ($44|0)==(0|0);
 if ($45) {
  HEAP32[(1205568)>>2] = 4666;
 }
 $46 = HEAP32[(1205572)>>2]|0;
 $47 = ($46|0)==(0|0);
 if ($47) {
  HEAP32[(1205572)>>2] = 4667;
 }
 $48 = HEAP32[(1205576)>>2]|0;
 $49 = ($48|0)==(0|0);
 if ($49) {
  HEAP32[(1205576)>>2] = 4668;
 }
 $50 = HEAP32[(1205580)>>2]|0;
 $51 = ($50|0)==(0|0);
 if ($51) {
  HEAP32[(1205580)>>2] = 4669;
 }
 $52 = HEAP32[(1205584)>>2]|0;
 $53 = ($52|0)==(0|0);
 if ($53) {
  HEAP32[(1205584)>>2] = 4670;
 }
 $54 = HEAP32[(1205588)>>2]|0;
 $55 = ($54|0)==(0|0);
 if ($55) {
  HEAP32[(1205588)>>2] = 4671;
 }
 $56 = HEAP32[(1205592)>>2]|0;
 $57 = ($56|0)==(0|0);
 if ($57) {
  HEAP32[(1205592)>>2] = 4672;
 }
 $58 = HEAP32[(1205596)>>2]|0;
 $59 = ($58|0)==(0|0);
 if ($59) {
  HEAP32[(1205596)>>2] = 4673;
 }
 $60 = HEAP32[(1205600)>>2]|0;
 $61 = ($60|0)==(0|0);
 if ($61) {
  HEAP32[(1205600)>>2] = 4674;
 }
 $62 = HEAP32[(1205604)>>2]|0;
 $63 = ($62|0)==(0|0);
 if ($63) {
  HEAP32[(1205604)>>2] = 4675;
 }
 FUNCTION_TABLE_v[$64 & 8191]();
 $$01 = 0;
 STACKTOP = sp;return ($$01|0);
}
function _SDL_WasInit($flags) {
 $flags = $flags|0;
 var $$11 = 0, $$flags = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$02 = 0, $initialized$0$lcssa = 0, $initialized$03 = 0, $initialized$1 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($flags|0)==(0);
 $$flags = $0 ? 29233 : $flags;
 $1 = (Math_clz32(($$flags|0))|0);
 $2 = ($1>>>0)>(31);
 if ($2) {
  $initialized$0$lcssa = 0;
  return ($initialized$0$lcssa|0);
 }
 $3 = (32 - ($1))|0;
 $$11 = $$flags;$i$02 = 0;$initialized$03 = 0;
 while(1) {
  $4 = $$11 & 1;
  $5 = ($4|0)==(0);
  if ($5) {
   $initialized$1 = $initialized$03;
  } else {
   $6 = (1280325 + ($i$02)|0);
   $7 = HEAP8[$6>>0]|0;
   $8 = ($7<<24>>24)==(0);
   if ($8) {
    $initialized$1 = $initialized$03;
   } else {
    $9 = 1 << $i$02;
    $10 = $initialized$03 | $9;
    $initialized$1 = $10;
   }
  }
  $11 = $$11 >>> 1;
  $12 = (($i$02) + 1)|0;
  $exitcond = ($12|0)==($3|0);
  if ($exitcond) {
   $initialized$0$lcssa = $initialized$1;
   break;
  } else {
   $$11 = $11;$i$02 = $12;$initialized$03 = $initialized$1;
  }
 }
 return ($initialized$0$lcssa|0);
}
function _SDL_AudioDetectDevices_Default() {
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _SDL_AddAudioDevice(0,610283,(1));
 $0 = HEAP32[(1205616)>>2]|0;
 $1 = ($0|0)==(0);
 if ($1) {
  return;
 }
 _SDL_AddAudioDevice(1,610310,(2));
 return;
}
function _SDL_AudioOpenDevice_Default($_this,$handle,$devname,$iscapture) {
 $_this = $_this|0;
 $handle = $handle|0;
 $devname = $devname|0;
 $iscapture = $iscapture|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_Error(4)|0);
 return ($0|0);
}
function _SDL_AudioThreadInit_Default($_this) {
 $_this = $_this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _SDL_AudioWaitDevice_Default($_this) {
 $_this = $_this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _SDL_AudioPlayDevice_Default($_this) {
 $_this = $_this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _SDL_AudioGetPendingBytes_Default($_this) {
 $_this = $_this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _SDL_AudioGetDeviceBuf_Default($_this) {
 $_this = $_this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (0|0);
}
function _SDL_AudioWaitDone_Default($_this) {
 $_this = $_this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _SDL_AudioCloseDevice_Default($_this) {
 $_this = $_this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _SDL_AudioLockDevice_Default($device) {
 $device = $device|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($device)) + 160|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = (_SDL_ThreadID()|0);
  $4 = ((($device)) + 164|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ($3|0)==($5|0);
  if ($6) {
   return;
  }
 }
 return;
}
function _SDL_AudioUnlockDevice_Default($device) {
 $device = $device|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($device)) + 160|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = (_SDL_ThreadID()|0);
  $4 = ((($device)) + 164|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ($3|0)==($5|0);
  if ($6) {
   return;
  }
 }
 return;
}
function _SDL_AudioFreeDeviceHandle_Default($handle) {
 $handle = $handle|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _SDL_AudioDeinitialize_Default() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _SDL_JoystickInit() {
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _SDL_AddHintCallback(610338,4676,0);
 $0 = (_SDL_InitSubSystem(16384)|0);
 $1 = ($0|0)<(0);
 if ($1) {
  $$0 = -1;
  return ($$0|0);
 }
 $2 = (_SDL_SYS_JoystickInit()|0);
 $3 = ($2|0)<(0);
 $$ = $3 ? $2 : 0;
 $$0 = $$;
 return ($$0|0);
}
function _SDL_JoystickAllowBackgroundEventsChanged($userdata,$name,$oldValue,$hint) {
 $userdata = $userdata|0;
 $name = $name|0;
 $oldValue = $oldValue|0;
 $hint = $hint|0;
 var $0 = 0, $1 = 0, $2 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($hint|0)==(0|0);
 if ($0) {
  label = 3;
 } else {
  $1 = HEAP8[$hint>>0]|0;
  $2 = ($1<<24>>24)==(49);
  if ($2) {
   $storemerge = 1;
  } else {
   label = 3;
  }
 }
 if ((label|0) == 3) {
  $storemerge = 0;
 }
 HEAP32[301486] = $storemerge;
 return;
}
function _SDL_AddHintCallback($name,$callback,$userdata) {
 $name = $name|0;
 $callback = $callback|0;
 $userdata = $userdata|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $entry$0$i = 0, $entry$05$i = 0, $entry$07$i = 0;
 var $entry$07$i$lcssa = 0, $hint$0 = 0, $hint$0$i = 0, $hint$0$i2 = 0, $hint$010 = 0, $hint$02$i = 0, $hint$03$i = 0, $hint$03$i$lcssa = 0, $hint$08$i = 0, $hint$09 = 0, $hint$09$i = 0, $hint$09$i$lcssa = 0, $hint$1 = 0, $prev$06$i = 0, $prev$06$i$lcssa = 0, $prev$06$i$phi = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($name|0)==(0|0);
 if (!($0)) {
  $1 = HEAP8[$name>>0]|0;
  $2 = ($1<<24>>24)==(0);
  if (!($2)) {
   $3 = ($callback|0)==(0|0);
   if ($3) {
    HEAP32[$vararg_buffer1>>2] = 610380;
    (_SDL_SetError(611267,$vararg_buffer1)|0);
    STACKTOP = sp;return;
   }
   $hint$08$i = HEAP32[300735]|0;
   $4 = ($hint$08$i|0)==(0|0);
   L8: do {
    if (!($4)) {
     $hint$09$i = $hint$08$i;
     while(1) {
      $5 = HEAP32[$hint$09$i>>2]|0;
      $6 = (_SDL_strcmp($name,$5)|0);
      $7 = ($6|0)==(0);
      if ($7) {
       $hint$09$i$lcssa = $hint$09$i;
       break;
      }
      $21 = ((($hint$09$i)) + 16|0);
      $hint$0$i = HEAP32[$21>>2]|0;
      $22 = ($hint$0$i|0)==(0|0);
      if ($22) {
       break L8;
      } else {
       $hint$09$i = $hint$0$i;
      }
     }
     $8 = ((($hint$09$i$lcssa)) + 12|0);
     $entry$05$i = HEAP32[$8>>2]|0;
     $9 = ($entry$05$i|0)==(0|0);
     if (!($9)) {
      $entry$07$i = $entry$05$i;$prev$06$i = 0;
      while(1) {
       $10 = HEAP32[$entry$07$i>>2]|0;
       $11 = ($10|0)==($callback|0);
       if ($11) {
        $12 = ((($entry$07$i)) + 4|0);
        $13 = HEAP32[$12>>2]|0;
        $14 = ($13|0)==($userdata|0);
        if ($14) {
         $entry$07$i$lcssa = $entry$07$i;$prev$06$i$lcssa = $prev$06$i;
         break;
        }
       }
       $19 = ((($entry$07$i)) + 8|0);
       $entry$0$i = HEAP32[$19>>2]|0;
       $20 = ($entry$0$i|0)==(0|0);
       if ($20) {
        break L8;
       } else {
        $prev$06$i$phi = $entry$07$i;$entry$07$i = $entry$0$i;$prev$06$i = $prev$06$i$phi;
       }
      }
      $15 = ($prev$06$i$lcssa|0)==(0|0);
      $16 = ((($entry$07$i$lcssa)) + 8|0);
      $17 = HEAP32[$16>>2]|0;
      if ($15) {
       HEAP32[$8>>2] = $17;
      } else {
       $18 = ((($prev$06$i$lcssa)) + 8|0);
       HEAP32[$18>>2] = $17;
      }
      _SDL_free($entry$07$i$lcssa);
     }
    }
   } while(0);
   $23 = (_SDL_malloc(12)|0);
   $24 = ($23|0)==(0|0);
   if ($24) {
    (_SDL_Error(0)|0);
    STACKTOP = sp;return;
   }
   HEAP32[$23>>2] = $callback;
   $25 = ((($23)) + 4|0);
   HEAP32[$25>>2] = $userdata;
   $hint$09 = HEAP32[300735]|0;
   $26 = ($hint$09|0)==(0|0);
   L28: do {
    if ($26) {
     label = 22;
    } else {
     $hint$010 = $hint$09;
     while(1) {
      $27 = HEAP32[$hint$010>>2]|0;
      $28 = (_SDL_strcmp($name,$27)|0);
      $29 = ($28|0)==(0);
      if ($29) {
       $hint$1 = $hint$010;
       break L28;
      }
      $30 = ((($hint$010)) + 16|0);
      $hint$0 = HEAP32[$30>>2]|0;
      $31 = ($hint$0|0)==(0|0);
      if ($31) {
       label = 22;
       break;
      } else {
       $hint$010 = $hint$0;
      }
     }
    }
   } while(0);
   do {
    if ((label|0) == 22) {
     $32 = (_SDL_malloc(20)|0);
     $33 = ($32|0)==(0|0);
     if (!($33)) {
      $34 = (_SDL_strdup($name)|0);
      HEAP32[$32>>2] = $34;
      $35 = ((($32)) + 4|0);
      HEAP32[$35>>2] = 0;
      $36 = ((($32)) + 8|0);
      HEAP32[$36>>2] = 0;
      $37 = ((($32)) + 12|0);
      HEAP32[$37>>2] = 0;
      $38 = HEAP32[300735]|0;
      $39 = ((($32)) + 16|0);
      HEAP32[$39>>2] = $38;
      HEAP32[300735] = $32;
      $hint$1 = $32;
      break;
     }
     (_SDL_Error(0)|0);
     _SDL_free($23);
     STACKTOP = sp;return;
    }
   } while(0);
   $40 = ((($hint$1)) + 12|0);
   $41 = HEAP32[$40>>2]|0;
   $42 = ((($23)) + 8|0);
   HEAP32[$42>>2] = $41;
   HEAP32[$40>>2] = $23;
   $43 = (_SDL_getenv($name)|0);
   $hint$02$i = HEAP32[300735]|0;
   $44 = ($hint$02$i|0)==(0|0);
   L39: do {
    if ($44) {
     $$0$i = $43;
    } else {
     $hint$03$i = $hint$02$i;
     while(1) {
      $45 = HEAP32[$hint$03$i>>2]|0;
      $46 = (_SDL_strcmp($name,$45)|0);
      $47 = ($46|0)==(0);
      if ($47) {
       $hint$03$i$lcssa = $hint$03$i;
       break;
      }
      $54 = ((($hint$03$i)) + 16|0);
      $hint$0$i2 = HEAP32[$54>>2]|0;
      $55 = ($hint$0$i2|0)==(0|0);
      if ($55) {
       $$0$i = $43;
       break L39;
      } else {
       $hint$03$i = $hint$0$i2;
      }
     }
     $48 = ($43|0)==(0|0);
     if (!($48)) {
      $49 = ((($hint$03$i$lcssa)) + 8|0);
      $50 = HEAP32[$49>>2]|0;
      $51 = ($50|0)==(2);
      if (!($51)) {
       $$0$i = $43;
       break;
      }
     }
     $52 = ((($hint$03$i$lcssa)) + 4|0);
     $53 = HEAP32[$52>>2]|0;
     $$0$i = $53;
    }
   } while(0);
   FUNCTION_TABLE_viiii[$callback & 8191]($userdata,$name,$$0$i,$$0$i);
   STACKTOP = sp;return;
  }
 }
 HEAP32[$vararg_buffer>>2] = 610375;
 (_SDL_SetError(611267,$vararg_buffer)|0);
 STACKTOP = sp;return;
}
function _SDL_SYS_JoystickInit() {
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $gamepadState = 0, $i$04 = 0, $item$01$i = 0, $item$01$i1 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1456|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 1448|0;
 $vararg_buffer1 = sp + 1440|0;
 $vararg_buffer = sp + 1432|0;
 $gamepadState = sp;
 HEAP32[301463] = 0;
 $0 = (_emscripten_get_num_gamepads()|0);
 $1 = ($0|0)==(-1);
 if ($1) {
  $2 = (_SDL_SetError(610389,$vararg_buffer)|0);
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ($0|0)>(0);
 if ($3) {
  $i$04 = 0;
  while(1) {
   $4 = (_emscripten_get_gamepad_status(($i$04|0),($gamepadState|0))|0);
   $5 = ($4|0)==(0);
   if ($5) {
    (_Emscripten_JoyStickConnected(0,$gamepadState,0)|0);
   }
   $6 = (($i$04) + 1)|0;
   $exitcond = ($6|0)==($0|0);
   if ($exitcond) {
    break;
   } else {
    $i$04 = $6;
   }
  }
 }
 $7 = (_emscripten_set_gamepadconnected_callback((0|0),0,(4677|0))|0);
 $8 = ($7|0)==(0);
 if (!($8)) {
  $9 = HEAP32[301464]|0;
  $10 = ($9|0)==(0|0);
  if (!($10)) {
   $item$01$i = $9;
   while(1) {
    $11 = ((($item$01$i)) + 1320|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = ((($item$01$i)) + 8|0);
    $14 = HEAP32[$13>>2]|0;
    _SDL_free($14);
    $15 = ((($item$01$i)) + 4|0);
    $16 = HEAP32[$15>>2]|0;
    _SDL_free($16);
    _SDL_free($item$01$i);
    $17 = ($12|0)==(0|0);
    if ($17) {
     break;
    } else {
     $item$01$i = $12;
    }
   }
  }
  HEAP32[301467] = 0;
  HEAP32[301464] = 0;
  HEAP32[301463] = 0;
  HEAP32[301468] = 0;
  (_emscripten_set_gamepadconnected_callback((0|0),0,(0|0))|0);
  (_emscripten_set_gamepaddisconnected_callback((0|0),0,(0|0))|0);
  $18 = (_SDL_SetError(610412,$vararg_buffer1)|0);
  $$0 = $18;
  STACKTOP = sp;return ($$0|0);
 }
 $19 = (_emscripten_set_gamepaddisconnected_callback((0|0),0,(4678|0))|0);
 $20 = ($19|0)==(0);
 if ($20) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $21 = HEAP32[301464]|0;
 $22 = ($21|0)==(0|0);
 if (!($22)) {
  $item$01$i1 = $21;
  while(1) {
   $23 = ((($item$01$i1)) + 1320|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = ((($item$01$i1)) + 8|0);
   $26 = HEAP32[$25>>2]|0;
   _SDL_free($26);
   $27 = ((($item$01$i1)) + 4|0);
   $28 = HEAP32[$27>>2]|0;
   _SDL_free($28);
   _SDL_free($item$01$i1);
   $29 = ($24|0)==(0|0);
   if ($29) {
    break;
   } else {
    $item$01$i1 = $24;
   }
  }
 }
 HEAP32[301467] = 0;
 HEAP32[301464] = 0;
 HEAP32[301463] = 0;
 HEAP32[301468] = 0;
 (_emscripten_set_gamepadconnected_callback((0|0),0,(0|0))|0);
 (_emscripten_set_gamepaddisconnected_callback((0|0),0,(0|0))|0);
 $30 = (_SDL_SetError(610451,$vararg_buffer3)|0);
 $$0 = $30;
 STACKTOP = sp;return ($$0|0);
}
function _Emscripten_JoyStickConnected($eventType,$gamepadEvent,$userData) {
 $eventType = $eventType|0;
 $gamepadEvent = $gamepadEvent|0;
 $userData = $userData|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0.0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, $event = 0, $i$02 = 0, $i$11 = 0, $item$0$i = 0, $item$01$i = 0, $item$02$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $event = sp;
 $0 = ((($gamepadEvent)) + 1300|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)<(0);
 L1: do {
  if (!($2)) {
   $item$01$i = HEAP32[301464]|0;
   $3 = ($item$01$i|0)==(0|0);
   if (!($3)) {
    $item$02$i = $item$01$i;
    while(1) {
     $4 = HEAP32[$item$02$i>>2]|0;
     $5 = ($4|0)==($1|0);
     if ($5) {
      break;
     }
     $6 = ((($item$02$i)) + 1320|0);
     $item$0$i = HEAP32[$6>>2]|0;
     $7 = ($item$0$i|0)==(0|0);
     if ($7) {
      break L1;
     } else {
      $item$02$i = $item$0$i;
     }
    }
    STACKTOP = sp;return 1;
   }
  }
 } while(0);
 $8 = (_SDL_malloc(1328)|0);
 $9 = ($8|0)==(0|0);
 do {
  if (!($9)) {
   (_SDL_memset($8,0,1328)|0);
   $10 = HEAP32[$0>>2]|0;
   HEAP32[$8>>2] = $10;
   $11 = ((($gamepadEvent)) + 1304|0);
   $12 = (_SDL_strdup($11)|0);
   $13 = ((($8)) + 4|0);
   HEAP32[$13>>2] = $12;
   $14 = ($12|0)==(0|0);
   if ($14) {
    _SDL_free($8);
    break;
   }
   $15 = ((($gamepadEvent)) + 1368|0);
   $16 = (_SDL_strdup($15)|0);
   $17 = ((($8)) + 8|0);
   HEAP32[$17>>2] = $16;
   $18 = ($16|0)==(0|0);
   if ($18) {
    $19 = HEAP32[$13>>2]|0;
    _SDL_free($19);
    _SDL_free($8);
    break;
   }
   $20 = ((($gamepadEvent)) + 8|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = ((($8)) + 24|0);
   HEAP32[$22>>2] = $21;
   $23 = ((($gamepadEvent)) + 12|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = ((($8)) + 20|0);
   HEAP32[$25>>2] = $24;
   $26 = HEAP32[301468]|0;
   $27 = (($26) + 1)|0;
   HEAP32[301468] = $27;
   $28 = ((($8)) + 12|0);
   HEAP32[$28>>2] = $26;
   $29 = +HEAPF64[$gamepadEvent>>3];
   $30 = ((($8)) + 32|0);
   HEAPF64[$30>>3] = $29;
   $31 = ($21|0)>(0);
   if ($31) {
    $32 = ((($8)) + 40|0);
    $i$02 = 0;
    while(1) {
     $37 = (((($gamepadEvent)) + 16|0) + ($i$02<<3)|0);
     $38 = +HEAPF64[$37>>3];
     $39 = (($32) + ($i$02<<3)|0);
     HEAPF64[$39>>3] = $38;
     $40 = (($i$02) + 1)|0;
     $41 = HEAP32[$22>>2]|0;
     $42 = ($40|0)<($41|0);
     if ($42) {
      $i$02 = $40;
     } else {
      break;
     }
    }
    $$pre = HEAP32[$25>>2]|0;
    $34 = $$pre;
   } else {
    $34 = $24;
   }
   $33 = ($34|0)>(0);
   if ($33) {
    $35 = ((($8)) + 552|0);
    $36 = ((($8)) + 1064|0);
    $i$11 = 0;
    while(1) {
     $43 = (((($gamepadEvent)) + 528|0) + ($i$11<<3)|0);
     $44 = +HEAPF64[$43>>3];
     $45 = (($35) + ($i$11<<3)|0);
     HEAPF64[$45>>3] = $44;
     $46 = (((($gamepadEvent)) + 1040|0) + ($i$11<<2)|0);
     $47 = HEAP32[$46>>2]|0;
     $48 = (($36) + ($i$11<<2)|0);
     HEAP32[$48>>2] = $47;
     $49 = (($i$11) + 1)|0;
     $50 = HEAP32[$25>>2]|0;
     $51 = ($49|0)<($50|0);
     if ($51) {
      $i$11 = $49;
     } else {
      break;
     }
    }
   }
   $52 = HEAP32[301467]|0;
   $53 = ($52|0)==(0|0);
   if ($53) {
    HEAP32[301467] = $8;
    HEAP32[301464] = $8;
   } else {
    $54 = ((($52)) + 1320|0);
    HEAP32[$54>>2] = $8;
    HEAP32[301467] = $8;
   }
   $55 = HEAP32[301463]|0;
   $56 = (($55) + 1)|0;
   HEAP32[301463] = $56;
   HEAP32[$event>>2] = 1541;
   $57 = (_SDL_EventState(1541,-1)|0);
   $58 = ($57<<24>>24)==(1);
   if ($58) {
    $59 = HEAP32[301463]|0;
    $60 = (($59) + -1)|0;
    $61 = ((($event)) + 8|0);
    HEAP32[$61>>2] = $60;
    $62 = HEAP32[300435]|0;
    $63 = ($62|0)==(0|0);
    if (!($63)) {
     $64 = HEAP32[300436]|0;
     $65 = (FUNCTION_TABLE_iii[$62 & 8191]($64,$event)|0);
     $66 = ($65|0)==(0);
     if ($66) {
      break;
     }
    }
    (_SDL_PushEvent($event)|0);
   }
  }
 } while(0);
 STACKTOP = sp;return 1;
}
function _Emscripten_JoyStickDisconnected($eventType,$gamepadEvent,$userData) {
 $eventType = $eventType|0;
 $gamepadEvent = $gamepadEvent|0;
 $userData = $userData|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cond = 0, $cond3 = 0, $event = 0, $item$04 = 0, $item$04$lcssa = 0, $prev$05 = 0, $prev$05$lcssa = 0;
 var $prev$05$phi = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $event = sp;
 $0 = HEAP32[301464]|0;
 $cond3 = ($0|0)==(0|0);
 if ($cond3) {
  STACKTOP = sp;return 1;
 }
 $1 = ((($gamepadEvent)) + 1300|0);
 $2 = HEAP32[$1>>2]|0;
 $item$04 = $0;$prev$05 = 0;
 while(1) {
  $3 = HEAP32[$item$04>>2]|0;
  $4 = ($3|0)==($2|0);
  if ($4) {
   $item$04$lcssa = $item$04;$prev$05$lcssa = $prev$05;
   break;
  }
  $5 = ((($item$04)) + 1320|0);
  $6 = HEAP32[$5>>2]|0;
  $cond = ($6|0)==(0|0);
  if ($cond) {
   label = 17;
   break;
  } else {
   $prev$05$phi = $item$04;$item$04 = $6;$prev$05 = $prev$05$phi;
  }
 }
 if ((label|0) == 17) {
  STACKTOP = sp;return 1;
 }
 $7 = ((($item$04$lcssa)) + 16|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 if (!($9)) {
  $10 = ((($8)) + 40|0);
  HEAP32[$10>>2] = 0;
 }
 $11 = ($prev$05$lcssa|0)==(0|0);
 $12 = ((($item$04$lcssa)) + 1320|0);
 $13 = HEAP32[$12>>2]|0;
 if ($11) {
  HEAP32[301464] = $13;
 } else {
  $14 = ((($prev$05$lcssa)) + 1320|0);
  HEAP32[$14>>2] = $13;
 }
 $15 = HEAP32[301467]|0;
 $16 = ($item$04$lcssa|0)==($15|0);
 if ($16) {
  HEAP32[301467] = $prev$05$lcssa;
 }
 $17 = HEAP32[301463]|0;
 $18 = (($17) + -1)|0;
 HEAP32[301463] = $18;
 HEAP32[$event>>2] = 1542;
 $19 = (_SDL_EventState(1542,-1)|0);
 $20 = ($19<<24>>24)==(1);
 do {
  if ($20) {
   $21 = ((($item$04$lcssa)) + 12|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = ((($event)) + 8|0);
   HEAP32[$23>>2] = $22;
   $24 = HEAP32[300435]|0;
   $25 = ($24|0)==(0|0);
   if (!($25)) {
    $26 = HEAP32[300436]|0;
    $27 = (FUNCTION_TABLE_iii[$24 & 8191]($26,$event)|0);
    $28 = ($27|0)==(0);
    if ($28) {
     break;
    }
   }
   (_SDL_PushEvent($event)|0);
  }
 } while(0);
 $29 = ((($item$04$lcssa)) + 4|0);
 $30 = HEAP32[$29>>2]|0;
 _SDL_free($30);
 $31 = ((($item$04$lcssa)) + 8|0);
 $32 = HEAP32[$31>>2]|0;
 _SDL_free($32);
 _SDL_free($item$04$lcssa);
 STACKTOP = sp;return 1;
}
function _SDL_GameControllerInit() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $deviceevent = 0, $i$11 = 0, $not$$i = 0, $pUserMappings$01$i$lcssa = 0, $pUserMappings$01$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $deviceevent = sp;
 HEAP32[301460] = 0;
 (_SDL_GameControllerAddMapping(610493)|0);
 $0 = (_SDL_GetHint(611137)|0);
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  $2 = HEAP8[$0>>0]|0;
  $3 = ($2<<24>>24)==(0);
  if (!($3)) {
   $4 = (_SDL_strlen($0)|0);
   $5 = (($4) + 1)|0;
   $6 = (_SDL_malloc($5)|0);
   (_SDL_memcpy($6,$0,$4)|0);
   $7 = (($6) + ($4)|0);
   HEAP8[$7>>0] = 0;
   $8 = ($6|0)==(0|0);
   if (!($8)) {
    $9 = (_SDL_strchr($6,10)|0);
    $10 = ($9|0)==(0|0);
    if ($10) {
     $pUserMappings$01$i$lcssa = $6;
    } else {
     $11 = $9;$pUserMappings$01$i2 = $6;
     while(1) {
      HEAP8[$11>>0] = 0;
      (_SDL_GameControllerAddMapping($pUserMappings$01$i2)|0);
      $12 = ((($11)) + 1|0);
      $13 = (_SDL_strchr($12,10)|0);
      $14 = ($13|0)==(0|0);
      if ($14) {
       $pUserMappings$01$i$lcssa = $12;
       break;
      } else {
       $11 = $13;$pUserMappings$01$i2 = $12;
      }
     }
    }
    (_SDL_GameControllerAddMapping($pUserMappings$01$i$lcssa)|0);
   }
   _SDL_free($6);
  }
 }
 _SDL_AddEventWatch(4649,0);
 $15 = (_SDL_NumJoysticks()|0);
 $16 = ($15|0)>(0);
 if (!($16)) {
  STACKTOP = sp;return 0;
 }
 $17 = ((($deviceevent)) + 8|0);
 $i$11 = 0;
 while(1) {
  $18 = (_SDL_PrivateGetControllerMapping($i$11)|0);
  $not$$i = ($18|0)==(0|0);
  if (!($not$$i)) {
   HEAP32[$deviceevent>>2] = 1619;
   HEAP32[$17>>2] = $i$11;
   (_SDL_PushEvent($deviceevent)|0);
  }
  $19 = (($i$11) + 1)|0;
  $20 = (_SDL_NumJoysticks()|0);
  $21 = ($19|0)<($20|0);
  if ($21) {
   $i$11 = $19;
  } else {
   break;
  }
 }
 STACKTOP = sp;return 0;
}
function _SDL_GameControllerAddMapping($mappingString) {
 $mappingString = $mappingString|0;
 var $$0 = 0, $$0$i = 0, $$byval_copy = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $event$i$i = 0, $existing$0 = 0, $gamecontrollerlist$0$i$i = 0, $gamecontrollerlist$01$i$i = 0, $gamecontrollerlist$02$i$i = 0, $jGUID$sroa$0 = 0, $jGUID1 = 0, $pSupportedController$0$i$i = 0, $pSupportedController$01$i$i = 0, $pSupportedController$02$i$i = 0, $pSupportedController$02$i$i$lcssa = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, $vararg_buffer7 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 160|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 136|0;
 $vararg_buffer7 = sp + 96|0;
 $vararg_buffer4 = sp + 88|0;
 $vararg_buffer1 = sp + 80|0;
 $vararg_buffer = sp + 72|0;
 $event$i$i = sp + 16|0;
 $jGUID1 = sp;
 $jGUID$sroa$0 = sp + 120|0;
 $0 = sp + 104|0;
 $1 = ($mappingString|0)==(0|0);
 if ($1) {
  HEAP32[$vararg_buffer>>2] = 610743;
  $2 = (_SDL_SetError(611267,$vararg_buffer)|0);
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = (_SDL_strchr($mappingString,44)|0);
 $4 = ($3|0)==(0|0);
 do {
  if (!($4)) {
   $5 = $3;
   $6 = $mappingString;
   $7 = (($5) - ($6))|0;
   $8 = (($7) + 1)|0;
   $9 = (_SDL_malloc($8)|0);
   $10 = ($9|0)==(0|0);
   if ($10) {
    (_SDL_Error(0)|0);
    break;
   }
   (_SDL_memcpy($9,$mappingString,$7)|0);
   $12 = (($9) + ($7)|0);
   HEAP8[$12>>0] = 0;
   $13 = (_SDL_strcasecmp($9,610785)|0);
   $14 = ($13|0)==(0);
   $15 = (_SDL_strcasecmp($9,610792)|0);
   $16 = ($15|0)==(0);
   _SDL_JoystickGetGUIDFromString($0,$9);
   dest=$jGUID$sroa$0; src=$0; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
   _SDL_free($9);
   dest=$jGUID1; src=$jGUID$sroa$0; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
   $17 = (_SDL_strchr($mappingString,44)|0);
   $18 = ($17|0)==(0|0);
   L10: do {
    if ($18) {
     label = 11;
    } else {
     $19 = ((($17)) + 1|0);
     $20 = (_SDL_strchr($19,44)|0);
     $21 = ($20|0)==(0|0);
     if ($21) {
      label = 11;
     } else {
      $22 = $20;
      $23 = $17;
      $24 = (($22) - ($23))|0;
      $25 = (_SDL_malloc($24)|0);
      $26 = ($25|0)==(0|0);
      if ($26) {
       (_SDL_Error(0)|0);
       label = 11;
       break;
      }
      (_SDL_memcpy($25,$19,$24)|0);
      $27 = (($24) + -1)|0;
      $28 = (($25) + ($27)|0);
      HEAP8[$28>>0] = 0;
      $29 = (_SDL_strchr($mappingString,44)|0);
      $30 = ($29|0)==(0|0);
      if (!($30)) {
       $31 = ((($29)) + 1|0);
       $32 = (_SDL_strchr($31,44)|0);
       $33 = ($32|0)==(0|0);
       if (!($33)) {
        $34 = ((($32)) + 1|0);
        $35 = (_SDL_strdup($34)|0);
        $36 = ($35|0)==(0|0);
        if (!($36)) {
         $pSupportedController$01$i$i = HEAP32[301460]|0;
         $37 = ($pSupportedController$01$i$i|0)==(0|0);
         L20: do {
          if ($37) {
           label = 25;
          } else {
           $pSupportedController$02$i$i = $pSupportedController$01$i$i;
           while(1) {
            $38 = (_SDL_memcmp($jGUID1,$pSupportedController$02$i$i,16)|0);
            $39 = ($38|0)==(0);
            if ($39) {
             $$lcssa = $pSupportedController$02$i$i;$pSupportedController$02$i$i$lcssa = $pSupportedController$02$i$i;
             break;
            }
            $40 = ((($pSupportedController$02$i$i)) + 24|0);
            $pSupportedController$0$i$i = HEAP32[$40>>2]|0;
            $41 = ($pSupportedController$0$i$i|0)==(0|0);
            if ($41) {
             label = 25;
             break L20;
            } else {
             $pSupportedController$02$i$i = $pSupportedController$0$i$i;
            }
           }
           $42 = ((($pSupportedController$02$i$i$lcssa)) + 16|0);
           $43 = HEAP32[$42>>2]|0;
           _SDL_free($43);
           HEAP32[$42>>2] = $25;
           $44 = ((($pSupportedController$02$i$i$lcssa)) + 20|0);
           $45 = HEAP32[$44>>2]|0;
           _SDL_free($45);
           HEAP32[$44>>2] = $35;
           $gamecontrollerlist$01$i$i = HEAP32[301459]|0;
           $46 = ($gamecontrollerlist$01$i$i|0)==(0|0);
           if ($46) {
            $$0 = 0;
            STACKTOP = sp;return ($$0|0);
           }
           $47 = ((($event$i$i)) + 8|0);
           $gamecontrollerlist$02$i$i = $gamecontrollerlist$01$i$i;
           while(1) {
            $48 = ((($gamecontrollerlist$02$i$i)) + 12|0);
            $49 = (_SDL_memcmp($48,$$lcssa,16)|0);
            $50 = ($49|0)==(0);
            if ($50) {
             HEAP32[$event$i$i>>2] = 1621;
             $51 = HEAP32[$gamecontrollerlist$02$i$i>>2]|0;
             $52 = HEAP32[$51>>2]|0;
             HEAP32[$47>>2] = $52;
             (_SDL_PushEvent($event$i$i)|0);
             $53 = HEAP32[$42>>2]|0;
             $54 = HEAP32[$44>>2]|0;
             dest=$$byval_copy; src=$pSupportedController$02$i$i$lcssa; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
             _SDL_PrivateLoadButtonMapping($48,$$byval_copy,$53,$54);
            }
            $55 = ((($gamecontrollerlist$02$i$i)) + 896|0);
            $gamecontrollerlist$0$i$i = HEAP32[$55>>2]|0;
            $56 = ($gamecontrollerlist$0$i$i|0)==(0|0);
            if ($56) {
             $$0$i = $pSupportedController$02$i$i$lcssa;$existing$0 = 1;
             break;
            } else {
             $gamecontrollerlist$02$i$i = $gamecontrollerlist$0$i$i;
            }
           }
          }
         } while(0);
         do {
          if ((label|0) == 25) {
           $57 = (_SDL_malloc(28)|0);
           $58 = ($57|0)==(0|0);
           if ($58) {
            _SDL_free($25);
            _SDL_free($35);
            (_SDL_Error(0)|0);
            break L10;
           } else {
            dest=$57; src=$jGUID1; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
            $59 = ((($57)) + 16|0);
            HEAP32[$59>>2] = $25;
            $60 = ((($57)) + 20|0);
            HEAP32[$60>>2] = $35;
            $61 = HEAP32[301460]|0;
            $62 = ((($57)) + 24|0);
            HEAP32[$62>>2] = $61;
            HEAP32[301460] = $57;
            $$0$i = $57;$existing$0 = 0;
            break;
           }
          }
         } while(0);
         $63 = ($$0$i|0)==(0|0);
         if ($63) {
          $$0 = -1;
          STACKTOP = sp;return ($$0|0);
         }
         $64 = ($existing$0|0)==(0);
         if (!($64)) {
          $$0 = 0;
          STACKTOP = sp;return ($$0|0);
         }
         if ($14) {
          HEAP32[301466] = $$0$i;
         }
         if (!($16)) {
          $$0 = 1;
          STACKTOP = sp;return ($$0|0);
         }
         HEAP32[301465] = $$0$i;
         $$0 = 1;
         STACKTOP = sp;return ($$0|0);
        }
       }
      }
      _SDL_free($25);
      HEAP32[$vararg_buffer7>>2] = $mappingString;
      (_SDL_SetError(610831,$vararg_buffer7)|0);
     }
    }
   } while(0);
   if ((label|0) == 11) {
    HEAP32[$vararg_buffer4>>2] = $mappingString;
    (_SDL_SetError(610803,$vararg_buffer4)|0);
   }
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 HEAP32[$vararg_buffer1>>2] = $mappingString;
 $11 = (_SDL_SetError(610757,$vararg_buffer1)|0);
 $$0 = $11;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_strchr($string,$c) {
 $string = $string|0;
 $c = $c|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strchr($string,$c)|0);
 return ($0|0);
}
function _SDL_JoystickGetGUIDFromString($agg$result,$pchGUID) {
 $agg$result = $agg$result|0;
 $pchGUID = $pchGUID|0;
 var $$ = 0, $$0$i = 0, $$0$i5 = 0, $$7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $c$off$i = 0, $c$off$i2 = 0, $c$off3$i = 0, $c$off3$i3 = 0;
 var $c$off4$i = 0, $c$off4$i4 = 0, $guid1 = 0, $i$09 = 0, $p$08 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $guid1 = sp;
 $0 = (_SDL_strlen($pchGUID)|0);
 $1 = $0 & -2;
 (_SDL_memset($guid1,0,16)|0);
 $2 = ($1|0)==(0);
 L1: do {
  if (!($2)) {
   $3 = $guid1;
   $i$09 = 0;$p$08 = $guid1;
   while(1) {
    $4 = $p$08;
    $5 = (($4) - ($3))|0;
    $6 = ($5|0)<(16);
    if (!($6)) {
     break L1;
    }
    $7 = (($pchGUID) + ($i$09)|0);
    $8 = HEAP8[$7>>0]|0;
    $9 = $8 << 24 >> 24;
    $c$off$i = (($8) + -48)<<24>>24;
    $10 = ($c$off$i&255)<(10);
    do {
     if ($10) {
      $11 = (($9) + 208)|0;
      $$0$i = $11;
     } else {
      $c$off3$i = (($8) + -65)<<24>>24;
      $12 = ($c$off3$i&255)<(6);
      if ($12) {
       $13 = (($9) + 201)|0;
       $$0$i = $13;
       break;
      } else {
       $c$off4$i = (($8) + -97)<<24>>24;
       $14 = ($c$off4$i&255)<(6);
       $15 = (($9) + 169)|0;
       $$ = $14 ? $15 : 0;
       $$0$i = $$;
       break;
      }
     }
    } while(0);
    $16 = $$0$i << 4;
    $17 = $i$09 | 1;
    $18 = (($pchGUID) + ($17)|0);
    $19 = HEAP8[$18>>0]|0;
    $20 = $19 << 24 >> 24;
    $c$off$i2 = (($19) + -48)<<24>>24;
    $21 = ($c$off$i2&255)<(10);
    do {
     if ($21) {
      $22 = (($20) + 208)|0;
      $$0$i5 = $22;
     } else {
      $c$off3$i3 = (($19) + -65)<<24>>24;
      $23 = ($c$off3$i3&255)<(6);
      if ($23) {
       $24 = (($20) + 201)|0;
       $$0$i5 = $24;
       break;
      } else {
       $c$off4$i4 = (($19) + -97)<<24>>24;
       $25 = ($c$off4$i4&255)<(6);
       $26 = (($20) + 169)|0;
       $$7 = $25 ? $26 : 0;
       $$0$i5 = $$7;
       break;
      }
     }
    } while(0);
    $27 = $$0$i5 | $16;
    $28 = $27&255;
    HEAP8[$p$08>>0] = $28;
    $29 = (($i$09) + 2)|0;
    $30 = ((($p$08)) + 1|0);
    $31 = ($29>>>0)<($1>>>0);
    if ($31) {
     $i$09 = $29;$p$08 = $30;
    } else {
     break;
    }
   }
  }
 } while(0);
 dest=$agg$result; src=$guid1; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
 STACKTOP = sp;return;
}
function _SDL_PrivateLoadButtonMapping($pMapping,$guid,$pchName,$pchMapping) {
 $pMapping = $pMapping|0;
 $guid = $guid|0;
 $pchName = $pchName|0;
 $pchMapping = $pchMapping|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $bGameButton$01$i = 0, $bGameButton$1$i = 0, $exitcond7 = 0, $i$02$i = 0, $i$1$i = 0, $j$24 = 0, $pchPos$03$i = 0, $scevgep = 0, $szGameButton$i = 0, $szJoystickButton$i = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $szGameButton$i = sp + 32|0;
 $szJoystickButton$i = sp + 12|0;
 dest=$pMapping; src=$guid; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
 $0 = ((($pMapping)) + 16|0);
 HEAP32[$0>>2] = $pchName;
 $1 = ((($pMapping)) + 20|0);
 $2 = ((($pMapping)) + 196|0);
 HEAP32[$2>>2] = -1;
 $3 = ((($pMapping)) + 204|0);
 HEAP32[$3>>2] = -1;
 $4 = ((($pMapping)) + 212|0);
 HEAP32[$4>>2] = -1;
 $5 = ((($pMapping)) + 220|0);
 HEAP32[$5>>2] = -1;
 $6 = ((($pMapping)) + 228|0);
 HEAP32[$6>>2] = -1;
 $7 = ((($pMapping)) + 236|0);
 HEAP32[$7>>2] = -1;
 $8 = ((($pMapping)) + 244|0);
 HEAP32[$8>>2] = -1;
 $9 = ((($pMapping)) + 252|0);
 HEAP32[$9>>2] = -1;
 $10 = ((($pMapping)) + 260|0);
 HEAP32[$10>>2] = -1;
 $11 = ((($pMapping)) + 268|0);
 HEAP32[$11>>2] = -1;
 $12 = ((($pMapping)) + 276|0);
 HEAP32[$12>>2] = -1;
 $13 = ((($pMapping)) + 284|0);
 HEAP32[$13>>2] = -1;
 $14 = ((($pMapping)) + 292|0);
 HEAP32[$14>>2] = -1;
 $15 = ((($pMapping)) + 300|0);
 HEAP32[$15>>2] = -1;
 _memset(($1|0),-1,172)|0;
 $j$24 = 0;
 while(1) {
  $16 = (((($pMapping)) + 308|0) + ($j$24<<2)|0);
  HEAP32[$16>>2] = -1;
  $17 = (((($pMapping)) + 388|0) + ($j$24<<2)|0);
  HEAP32[$17>>2] = -1;
  $18 = (((($pMapping)) + 468|0) + ($j$24<<2)|0);
  HEAP32[$18>>2] = -1;
  $19 = (((($pMapping)) + 548|0) + ($j$24<<2)|0);
  HEAP32[$19>>2] = -1;
  $20 = (($j$24) + 1)|0;
  $exitcond7 = ($20|0)==(20);
  if ($exitcond7) {
   break;
  } else {
   $j$24 = $20;
  }
 }
 $scevgep = ((($pMapping)) + 628|0);
 _memset(($scevgep|0),-1,256)|0;
 (_SDL_memset($szGameButton$i,0,20)|0);
 (_SDL_memset($szJoystickButton$i,0,20)|0);
 $21 = ($pchMapping|0)==(0|0);
 L4: do {
  if (!($21)) {
   $bGameButton$01$i = 1;$i$02$i = 0;$pchPos$03$i = $pchMapping;
   L5: while(1) {
    $22 = HEAP8[$pchPos$03$i>>0]|0;
    L7: do {
     switch ($22<<24>>24) {
     case 0:  {
      break L4;
      break;
     }
     case 58:  {
      $bGameButton$1$i = 0;$i$1$i = 0;
      break;
     }
     default: {
      switch ($22<<24>>24) {
      case 32:  {
       $bGameButton$1$i = $bGameButton$01$i;$i$1$i = $i$02$i;
       break L7;
       break;
      }
      case 44:  {
       _SDL_PrivateGameControllerParseButton($szGameButton$i,$szJoystickButton$i,$pMapping);
       (_SDL_memset($szGameButton$i,0,20)|0);
       (_SDL_memset($szJoystickButton$i,0,20)|0);
       $bGameButton$1$i = 1;$i$1$i = 0;
       break L7;
       break;
      }
      default: {
      }
      }
      $23 = ($bGameButton$01$i|0)==(0);
      $24 = ($i$02$i>>>0)>(19);
      if ($23) {
       if ($24) {
        label = 12;
        break L5;
       }
       $27 = (($szJoystickButton$i) + ($i$02$i)|0);
       HEAP8[$27>>0] = $22;
       $28 = (($i$02$i) + 1)|0;
       $bGameButton$1$i = 0;$i$1$i = $28;
       break L7;
      } else {
       if ($24) {
        label = 9;
        break L5;
       }
       $25 = (($szGameButton$i) + ($i$02$i)|0);
       HEAP8[$25>>0] = $22;
       $26 = (($i$02$i) + 1)|0;
       $bGameButton$1$i = $bGameButton$01$i;$i$1$i = $26;
       break L7;
      }
     }
     }
    } while(0);
    $29 = ((($pchPos$03$i)) + 1|0);
    $bGameButton$01$i = $bGameButton$1$i;$i$02$i = $i$1$i;$pchPos$03$i = $29;
   }
   if ((label|0) == 9) {
    HEAP32[$vararg_buffer>>2] = $szGameButton$i;
    (_SDL_SetError(611076,$vararg_buffer)|0);
    STACKTOP = sp;return;
   }
   else if ((label|0) == 12) {
    HEAP32[$vararg_buffer1>>2] = $szJoystickButton$i;
    (_SDL_SetError(611102,$vararg_buffer1)|0);
    STACKTOP = sp;return;
   }
  }
 } while(0);
 _SDL_PrivateGameControllerParseButton($szGameButton$i,$szJoystickButton$i,$pMapping);
 STACKTOP = sp;return;
}
function _SDL_PrivateGameControllerParseButton($szGameButton,$szJoystickButton,$pMapping) {
 $szGameButton = $szGameButton|0;
 $szJoystickButton = $szJoystickButton|0;
 $pMapping = $pMapping|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_SDL_GameControllerGetButtonFromString($szGameButton)|0);
 $1 = (_SDL_GameControllerGetAxisFromString($szGameButton)|0);
 $2 = ((($szJoystickButton)) + 1|0);
 $3 = (_SDL_atoi($2)|0);
 $4 = HEAP8[$szJoystickButton>>0]|0;
 switch ($4<<24>>24) {
 case 97:  {
  $5 = ($3|0)>(19);
  if ($5) {
   HEAP32[$vararg_buffer>>2] = $3;
   (_SDL_SetError(611000,$vararg_buffer)|0);
   STACKTOP = sp;return;
  }
  $6 = ($1|0)==(-1);
  if (!($6)) {
   $7 = (((($pMapping)) + 20|0) + ($1<<2)|0);
   HEAP32[$7>>2] = $3;
   $8 = (((($pMapping)) + 308|0) + ($3<<2)|0);
   HEAP32[$8>>2] = $1;
   STACKTOP = sp;return;
  }
  $9 = ($0|0)==(-1);
  if ($9) {
   STACKTOP = sp;return;
  }
  $10 = (((($pMapping)) + 128|0) + ($0<<2)|0);
  HEAP32[$10>>2] = $3;
  $11 = (((($pMapping)) + 548|0) + ($3<<2)|0);
  HEAP32[$11>>2] = $0;
  STACKTOP = sp;return;
  break;
 }
 case 98:  {
  $12 = ($3|0)>(19);
  if ($12) {
   HEAP32[$vararg_buffer1>>2] = $3;
   (_SDL_SetError(611025,$vararg_buffer1)|0);
   STACKTOP = sp;return;
  }
  $13 = ($0|0)==(-1);
  if (!($13)) {
   $14 = (((($pMapping)) + 68|0) + ($0<<2)|0);
   HEAP32[$14>>2] = $3;
   $15 = (((($pMapping)) + 468|0) + ($3<<2)|0);
   HEAP32[$15>>2] = $0;
   STACKTOP = sp;return;
  }
  $16 = ($1|0)==(-1);
  if ($16) {
   STACKTOP = sp;return;
  }
  $17 = (((($pMapping)) + 44|0) + ($1<<2)|0);
  HEAP32[$17>>2] = $3;
  $18 = (((($pMapping)) + 388|0) + ($3<<2)|0);
  HEAP32[$18>>2] = $1;
  STACKTOP = sp;return;
  break;
 }
 case 104:  {
  $19 = (_SDL_atoi($2)|0);
  $20 = ((($szJoystickButton)) + 3|0);
  $21 = (_SDL_atoi($20)|0);
  $22 = ($19|0)>(3);
  if ($22) {
   HEAP32[$vararg_buffer4>>2] = $3;
   (_SDL_SetError(611052,$vararg_buffer4)|0);
  }
  $23 = ($0|0)==(-1);
  if ($23) {
   STACKTOP = sp;return;
  }
  $24 = (((($pMapping)) + 188|0) + ($0<<3)|0);
  HEAP32[$24>>2] = $19;
  $25 = $21&255;
  $26 = (((((($pMapping)) + 188|0) + ($0<<3)|0)) + 4|0);
  HEAP8[$26>>0] = $25;
  $27 = $19 << 4;
  $28 = $27 | $21;
  $29 = (((($pMapping)) + 628|0) + ($28<<2)|0);
  HEAP32[$29>>2] = $0;
  STACKTOP = sp;return;
  break;
 }
 default: {
  STACKTOP = sp;return;
 }
 }
}
function _SDL_GameControllerGetButtonFromString($pchString) {
 $pchString = $pchString|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($pchString|0)==(0|0);
 if ($0) {
  $$0 = -1;
 } else {
  $1 = HEAP8[$pchString>>0]|0;
  $2 = ($1<<24>>24)==(0);
  if ($2) {
   $$0 = -1;
  } else {
   $3 = (_SDL_strcasecmp($pchString,610849)|0);
   $4 = ($3|0)==(0);
   if ($4) {
    $$0 = 0;
   } else {
    $5 = (_SDL_strcasecmp($pchString,610851)|0);
    $6 = ($5|0)==(0);
    if ($6) {
     $$0 = 1;
    } else {
     $7 = (_SDL_strcasecmp($pchString,610853)|0);
     $8 = ($7|0)==(0);
     if ($8) {
      $$0 = 2;
     } else {
      $9 = (_SDL_strcasecmp($pchString,610855)|0);
      $10 = ($9|0)==(0);
      if ($10) {
       $$0 = 3;
      } else {
       $11 = (_SDL_strcasecmp($pchString,610857)|0);
       $12 = ($11|0)==(0);
       if ($12) {
        $$0 = 4;
       } else {
        $13 = (_SDL_strcasecmp($pchString,610862)|0);
        $14 = ($13|0)==(0);
        if ($14) {
         $$0 = 5;
        } else {
         $15 = (_SDL_strcasecmp($pchString,610868)|0);
         $16 = ($15|0)==(0);
         if ($16) {
          $$0 = 6;
         } else {
          $17 = (_SDL_strcasecmp($pchString,610874)|0);
          $18 = ($17|0)==(0);
          if ($18) {
           $$0 = 7;
          } else {
           $19 = (_SDL_strcasecmp($pchString,610884)|0);
           $20 = ($19|0)==(0);
           if ($20) {
            $$0 = 8;
           } else {
            $21 = (_SDL_strcasecmp($pchString,610895)|0);
            $22 = ($21|0)==(0);
            if ($22) {
             $$0 = 9;
            } else {
             $23 = (_SDL_strcasecmp($pchString,610908)|0);
             $24 = ($23|0)==(0);
             if ($24) {
              $$0 = 10;
             } else {
              $25 = (_SDL_strcasecmp($pchString,610922)|0);
              $26 = ($25|0)==(0);
              if ($26) {
               $$0 = 11;
              } else {
               $27 = (_SDL_strcasecmp($pchString,610927)|0);
               $28 = ($27|0)==(0);
               if ($28) {
                $$0 = 12;
               } else {
                $29 = (_SDL_strcasecmp($pchString,610934)|0);
                $30 = ($29|0)==(0);
                if ($30) {
                 $$0 = 13;
                } else {
                 $31 = (_SDL_strcasecmp($pchString,610941)|0);
                 $32 = ($31|0)==(0);
                 $$ = $32 ? 14 : -1;
                 return ($$|0);
                }
               }
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    }
   }
  }
 }
 return ($$0|0);
}
function _SDL_GameControllerGetAxisFromString($pchString) {
 $pchString = $pchString|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($pchString|0)==(0|0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = HEAP8[$pchString>>0]|0;
 $2 = ($1<<24>>24)==(0);
 if ($2) {
  $$0 = -1;
  return ($$0|0);
 }
 $3 = (_SDL_strcasecmp($pchString,610949)|0);
 $4 = ($3|0)==(0);
 if ($4) {
  $$0 = 0;
  return ($$0|0);
 }
 $5 = (_SDL_strcasecmp($pchString,610955)|0);
 $6 = ($5|0)==(0);
 if ($6) {
  $$0 = 1;
  return ($$0|0);
 }
 $7 = (_SDL_strcasecmp($pchString,610961)|0);
 $8 = ($7|0)==(0);
 if ($8) {
  $$0 = 2;
  return ($$0|0);
 }
 $9 = (_SDL_strcasecmp($pchString,610968)|0);
 $10 = ($9|0)==(0);
 if ($10) {
  $$0 = 3;
  return ($$0|0);
 }
 $11 = (_SDL_strcasecmp($pchString,610975)|0);
 $12 = ($11|0)==(0);
 if ($12) {
  $$0 = 4;
  return ($$0|0);
 } else {
  $13 = (_SDL_strcasecmp($pchString,610987)|0);
  $14 = ($13|0)==(0);
  $$ = $14 ? 5 : -1;
  return ($$|0);
 }
 return (0)|0;
}
function _SDL_NumJoysticks() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_SYS_NumJoysticks()|0);
 return ($0|0);
}
function _SDL_MinimizeWindow($window) {
 $window = $window|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ((($window)) + 48|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = $7 & 64;
   $9 = ($8|0)==(0);
   if (!($9)) {
    STACKTOP = sp;return;
   }
   (_SDL_UpdateFullscreenMode($window,0)|0);
   $10 = HEAP32[300441]|0;
   $11 = ((($10)) + 76|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($12|0)==(0|0);
   if ($13) {
    STACKTOP = sp;return;
   }
   FUNCTION_TABLE_vii[$12 & 8191]($10,$window);
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 STACKTOP = sp;return;
}
function _SDL_ShowMessageBox($messageboxdata,$buttonid) {
 $messageboxdata = $messageboxdata|0;
 $buttonid = $buttonid|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dummybutton = 0, $dummybutton$buttonid = 0, $phitmp = 0, $phitmp4 = 0, $retval$02 = 0;
 var $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, $vararg_buffer9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer9 = sp + 40|0;
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $dummybutton = sp + 44|0;
 $0 = ($messageboxdata|0)==(0|0);
 if ($0) {
  HEAP32[$vararg_buffer>>2] = 611213;
  $1 = (_SDL_SetError(611267,$vararg_buffer)|0);
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = (_SDL_GetKeyboardFocus()|0);
 $3 = ($2|0)!=(0|0);
 do {
  if ($3) {
   $4 = HEAP32[300441]|0;
   $5 = ($4|0)==(0|0);
   if ($5) {
    (_SDL_SetError(595917,$vararg_buffer1)|0);
    $33 = 0;
    break;
   }
   $6 = HEAP32[$2>>2]|0;
   $7 = ((($4)) + 256|0);
   $8 = ($6|0)==($7|0);
   if ($8) {
    $9 = ((($2)) + 48|0);
    $10 = HEAP32[$9>>2]|0;
    $phitmp = $10 & 16384;
    $phitmp4 = ($phitmp|0)!=(0);
    $33 = $phitmp4;
    break;
   } else {
    (_SDL_SetError(609414,$vararg_buffer3)|0);
    $33 = 0;
    break;
   }
  } else {
   $33 = 0;
  }
 } while(0);
 $11 = (_SDL_GetRelativeMouseMode()|0);
 (_SDL_CaptureMouse(0)|0);
 (_SDL_SetRelativeMouseMode(0)|0);
 $12 = (_SDL_ShowCursor(1)|0);
 _SDL_ResetKeyboard();
 $13 = ($buttonid|0)==(0|0);
 $dummybutton$buttonid = $13 ? $dummybutton : $buttonid;
 $14 = HEAP32[300441]|0;
 $15 = ($14|0)==(0|0);
 if ($15) {
  label = 12;
 } else {
  $16 = ((($14)) + 228|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = ($17|0)==(0|0);
  if ($18) {
   label = 12;
  } else {
   $19 = (FUNCTION_TABLE_iiii[$17 & 8191]($14,$messageboxdata,$dummybutton$buttonid)|0);
   $20 = ($19|0)==(-1);
   if ($20) {
    label = 12;
   } else {
    $retval$02 = $19;
   }
  }
 }
 if ((label|0) == 12) {
  (_SDL_SetError(611228,$vararg_buffer5)|0);
  $retval$02 = -1;
 }
 if ($3) {
  $21 = HEAP32[300441]|0;
  $22 = ($21|0)==(0|0);
  do {
   if ($22) {
    (_SDL_SetError(595917,$vararg_buffer7)|0);
   } else {
    $23 = HEAP32[$2>>2]|0;
    $24 = ((($21)) + 256|0);
    $25 = ($23|0)==($24|0);
    if (!($25)) {
     (_SDL_SetError(609414,$vararg_buffer9)|0);
     break;
    }
    $26 = ((($2)) + 48|0);
    $27 = HEAP32[$26>>2]|0;
    $28 = $27 & 4;
    $29 = ($28|0)==(0);
    if (!($29)) {
     $30 = ((($21)) + 68|0);
     $31 = HEAP32[$30>>2]|0;
     $32 = ($31|0)==(0|0);
     if (!($32)) {
      FUNCTION_TABLE_vii[$31 & 8191]($21,$2);
     }
    }
   }
  } while(0);
  if ($33) {
   (_SDL_CaptureMouse(1)|0);
  }
 }
 (_SDL_ShowCursor($12)|0);
 (_SDL_SetRelativeMouseMode($11)|0);
 $$0 = $retval$02;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GetRelativeMouseMode() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(1201852)>>2]|0;
 return ($0|0);
}
function _SDL_ShowCursor($toggle) {
 $toggle = $toggle|0;
 var $$ = 0, $$01$i = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $not$ = 0, $notlhs$i = 0, $or$cond3$not$i = 0, $or$cond5$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(1201880)>>2]|0;
 $1 = ($toggle|0)>(-1);
 if (!($1)) {
  return ($0|0);
 }
 $not$ = ($toggle|0)!=(0);
 $$ = $not$&1;
 HEAP32[(1201880)>>2] = $$;
 $2 = ($$|0)==($0|0);
 if ($2) {
  return ($0|0);
 }
 $3 = HEAP32[(1201820)>>2]|0;
 $4 = ($3|0)==(0|0);
 $5 = HEAP32[(1201876)>>2]|0;
 $6 = HEAP32[(1201872)>>2]|0;
 $$01$i = $4 ? $6 : $5;
 $notlhs$i = ($$01$i|0)!=(0|0);
 $or$cond3$not$i = $notlhs$i & $not$;
 $7 = HEAP32[(1201852)>>2]|0;
 $8 = ($7|0)==(0);
 $or$cond5$i = $or$cond3$not$i & $8;
 $9 = HEAP32[(1201784)>>2]|0;
 $10 = ($9|0)==(0|0);
 if ($or$cond5$i) {
  if ($10) {
   return ($0|0);
  }
  (FUNCTION_TABLE_ii[$9 & 8191]($$01$i)|0);
  return ($0|0);
 } else {
  if ($10) {
   return ($0|0);
  }
  (FUNCTION_TABLE_ii[$9 & 8191](0)|0);
  return ($0|0);
 }
 return (0)|0;
}
function _SDL_GetVersion($ver) {
 $ver = $ver|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$ver>>0] = 2;
 $0 = ((($ver)) + 1|0);
 HEAP8[$0>>0] = 0;
 $1 = ((($ver)) + 2|0);
 HEAP8[$1>>0] = 4;
 return;
}
function _SDL_GetPlatform() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (611256|0);
}
function _SDL_GetCPUCount() {
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[301487]|0;
 $1 = ($0|0)==(0);
 if (!($1)) {
  $2 = $0;
  return ($2|0);
 }
 HEAP32[301487] = 1;
 $2 = 1;
 return ($2|0);
}
function _SDL_PumpEvents() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_GetVideoDevice()|0);
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  $2 = ((($0)) + 180|0);
  $3 = HEAP32[$2>>2]|0;
  FUNCTION_TABLE_vi[$3 & 8191]($0);
 }
 $4 = HEAP32[(1201912)>>2]|0;
 $5 = ($4|0)==(0|0);
 if (!($5)) {
  $6 = (_SDL_JoystickEventState(-1)|0);
  $7 = ($6|0)==(0);
  if ($7) {
   _SDL_SendPendingQuit();
   return;
  }
 }
 _SDL_JoystickUpdate();
 _SDL_SendPendingQuit();
 return;
}
function _SDL_JoystickEventState($state) {
 $state = $state|0;
 var $$1 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $cond = 0, $i$03 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $cond = ($state|0)==(-1);
 if ($cond) {
  $i$03 = 0;
 } else {
  (_SDL_EventState(1536,$state)|0);
  (_SDL_EventState(1537,$state)|0);
  (_SDL_EventState(1538,$state)|0);
  (_SDL_EventState(1539,$state)|0);
  (_SDL_EventState(1540,$state)|0);
  (_SDL_EventState(1541,$state)|0);
  (_SDL_EventState(1542,$state)|0);
  $$1 = $state;
  return ($$1|0);
 }
 while(1) {
  $0 = (140984 + ($i$03<<2)|0);
  $1 = HEAP32[$0>>2]|0;
  $2 = (_SDL_EventState($1,-1)|0);
  $3 = ($2<<24>>24)!=(1);
  $4 = (($i$03) + 1)|0;
  $5 = ($4>>>0)<(7);
  $or$cond = $3 & $5;
  if ($or$cond) {
   $i$03 = $4;
  } else {
   $$lcssa = $2;
   break;
  }
 }
 $6 = $$lcssa&255;
 $$1 = $6;
 return ($$1|0);
}
function _SDL_JoystickUpdate() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $event$i = 0, $event$i1 = 0, $event$i5 = 0, $i$011 = 0, $i$112 = 0, $i$214 = 0, $joystick$016 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $event$i5 = sp + 112|0;
 $event$i1 = sp + 56|0;
 $event$i = sp;
 $0 = HEAP32[301462]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  STACKTOP = sp;return;
 }
 $2 = ((($event$i5)) + 8|0);
 $3 = ((($event$i5)) + 12|0);
 $4 = ((($event$i5)) + 13|0);
 $5 = ((($event$i1)) + 8|0);
 $6 = ((($event$i1)) + 12|0);
 $7 = ((($event$i1)) + 13|0);
 $8 = ((($event$i)) + 8|0);
 $9 = ((($event$i)) + 12|0);
 $10 = ((($event$i)) + 16|0);
 $joystick$016 = $0;
 while(1) {
  $11 = ((($joystick$016)) + 56|0);
  $12 = HEAP32[$11>>2]|0;
  HEAP32[301461] = $joystick$016;
  _SDL_SYS_JoystickUpdate($joystick$016);
  $13 = ((($joystick$016)) + 48|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ($14|0)==(0);
  if (!($15)) {
   $16 = ((($joystick$016)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ($17|0)>(0);
   if ($18) {
    $19 = ((($joystick$016)) + 12|0);
    $27 = $17;$i$011 = 0;
    while(1) {
     $24 = $i$011&255;
     $25 = $i$011 & 255;
     $26 = ($25|0)<($27|0);
     if ($26) {
      $28 = HEAP32[$19>>2]|0;
      $29 = (($28) + ($25<<1)|0);
      $30 = HEAP16[$29>>1]|0;
      $31 = ($30<<16>>16)==(0);
      if (!($31)) {
       $32 = HEAP32[301486]|0;
       $33 = ($32|0)==(0);
       if ($33) {
        $34 = (_SDL_WasInit(32)|0);
        $35 = ($34|0)==(0);
        if (!($35)) {
        }
       }
       $36 = HEAP32[$19>>2]|0;
       $37 = (($36) + ($25<<1)|0);
       HEAP16[$37>>1] = 0;
       $38 = (_SDL_EventState(1536,-1)|0);
       $39 = ($38<<24>>24)==(1);
       if ($39) {
        HEAP32[$event$i>>2] = 1536;
        $40 = HEAP32[$joystick$016>>2]|0;
        HEAP32[$8>>2] = $40;
        HEAP8[$9>>0] = $24;
        HEAP16[$10>>1] = 0;
        (_SDL_PushEvent($event$i)|0);
       }
      }
     }
     $41 = (($i$011) + 1)|0;
     $42 = HEAP32[$16>>2]|0;
     $43 = ($41|0)<($42|0);
     if ($43) {
      $27 = $42;$i$011 = $41;
     } else {
      break;
     }
    }
   }
   $20 = ((($joystick$016)) + 32|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = ($21|0)>(0);
   if ($22) {
    $23 = ((($joystick$016)) + 36|0);
    $i$112 = 0;
    while(1) {
     $48 = $i$112&255;
     HEAP32[$event$i1>>2] = 1540;
     $49 = $i$112 & 255;
     $50 = HEAP32[$20>>2]|0;
     $51 = ($49|0)<($50|0);
     if ($51) {
      $52 = HEAP32[$23>>2]|0;
      $53 = (($52) + ($49)|0);
      $54 = HEAP8[$53>>0]|0;
      $55 = ($54<<24>>24)==(0);
      if (!($55)) {
       $56 = HEAP32[301486]|0;
       $57 = ($56|0)==(0);
       if ($57) {
        $58 = (_SDL_WasInit(32)|0);
        $59 = ($58|0)==(0);
        if (!($59)) {
        }
       }
       $60 = HEAP32[$23>>2]|0;
       $61 = (($60) + ($49)|0);
       HEAP8[$61>>0] = 0;
       $62 = HEAP32[$event$i1>>2]|0;
       $63 = (_SDL_EventState($62,-1)|0);
       $64 = ($63<<24>>24)==(1);
       if ($64) {
        $65 = HEAP32[$joystick$016>>2]|0;
        HEAP32[$5>>2] = $65;
        HEAP8[$6>>0] = $48;
        HEAP8[$7>>0] = 0;
        (_SDL_PushEvent($event$i1)|0);
       }
      }
     }
     $66 = (($i$112) + 1)|0;
     $67 = HEAP32[$20>>2]|0;
     $68 = ($66|0)<($67|0);
     if ($68) {
      $i$112 = $66;
     } else {
      break;
     }
    }
   }
   $44 = ((($joystick$016)) + 16|0);
   $45 = HEAP32[$44>>2]|0;
   $46 = ($45|0)>(0);
   if ($46) {
    $47 = ((($joystick$016)) + 20|0);
    $72 = $45;$i$214 = 0;
    while(1) {
     $69 = $i$214&255;
     $70 = $i$214 & 255;
     $71 = ($70|0)<($72|0);
     if ($71) {
      $73 = HEAP32[$47>>2]|0;
      $74 = (($73) + ($70)|0);
      $75 = HEAP8[$74>>0]|0;
      $76 = ($75<<24>>24)==(0);
      if (!($76)) {
       $77 = HEAP32[301486]|0;
       $78 = ($77|0)==(0);
       if ($78) {
        $79 = (_SDL_WasInit(32)|0);
        $80 = ($79|0)==(0);
        if (!($80)) {
        }
       }
       $81 = HEAP32[$47>>2]|0;
       $82 = (($81) + ($70)|0);
       HEAP8[$82>>0] = 0;
       $83 = (_SDL_EventState(1538,-1)|0);
       $84 = ($83<<24>>24)==(1);
       if ($84) {
        HEAP32[$event$i5>>2] = 1538;
        $85 = HEAP32[$joystick$016>>2]|0;
        HEAP32[$2>>2] = $85;
        HEAP8[$3>>0] = $69;
        HEAP8[$4>>0] = 0;
        (_SDL_PushEvent($event$i5)|0);
       }
      }
     }
     $86 = (($i$214) + 1)|0;
     $87 = HEAP32[$44>>2]|0;
     $88 = ($86|0)<($87|0);
     if ($88) {
      $72 = $87;$i$214 = $86;
     } else {
      break;
     }
    }
   }
   HEAP32[$13>>2] = 0;
  }
  HEAP32[301461] = 0;
  $89 = ((($joystick$016)) + 44|0);
  $90 = HEAP32[$89>>2]|0;
  $91 = ($90|0)<(1);
  if ($91) {
   _SDL_JoystickClose($joystick$016);
  }
  $92 = ($12|0)==(0|0);
  if ($92) {
   break;
  } else {
   $joystick$016 = $12;
  }
 }
 STACKTOP = sp;return;
}
function _SDL_SYS_JoystickUpdate($joystick) {
 $joystick = $joystick|0;
 var $$pre = 0, $$pre5 = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0.0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0.0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0.0;
 var $43 = 0, $44 = 0.0, $45 = 0, $46 = 0, $47 = 0, $48 = 0.0, $49 = 0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0.0, $gamepadState = 0, $i$03 = 0, $i$12 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1440|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $gamepadState = sp;
 $0 = ((($joystick)) + 40|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  STACKTOP = sp;return;
 }
 $3 = HEAP32[$1>>2]|0;
 $4 = (_emscripten_get_gamepad_status(($3|0),($gamepadState|0))|0);
 $5 = ($4|0)==(0);
 if (!($5)) {
  STACKTOP = sp;return;
 }
 $6 = +HEAPF64[$gamepadState>>3];
 $7 = $6 == 0.0;
 if (!($7)) {
  $8 = ((($1)) + 32|0);
  $9 = +HEAPF64[$8>>3];
  $10 = $6 != $9;
  if (!($10)) {
   STACKTOP = sp;return;
  }
 }
 $11 = ((($1)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)>(0);
 if ($13) {
  $14 = ((($1)) + 16|0);
  $i$03 = 0;
  while(1) {
   $19 = (((($1)) + 1064|0) + ($i$03<<2)|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = (((($gamepadState)) + 1040|0) + ($i$03<<2)|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = ($20|0)==($22|0);
   if ($23) {
    $31 = $20;
   } else {
    $24 = ($22|0)!=(0);
    $25 = HEAP32[$14>>2]|0;
    $26 = $i$03&255;
    $27 = $24&1;
    (_SDL_PrivateJoystickButton($25,$26,$27)|0);
    $$pre = HEAP32[$21>>2]|0;
    $31 = $$pre;
   }
   $28 = (((($gamepadState)) + 528|0) + ($i$03<<3)|0);
   $29 = +HEAPF64[$28>>3];
   $30 = (((($1)) + 552|0) + ($i$03<<3)|0);
   HEAPF64[$30>>3] = $29;
   HEAP32[$19>>2] = $31;
   $32 = (($i$03) + 1)|0;
   $33 = HEAP32[$11>>2]|0;
   $34 = ($32|0)<($33|0);
   if ($34) {
    $i$03 = $32;
   } else {
    break;
   }
  }
 }
 $15 = ((($1)) + 24|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ($16|0)>(0);
 if ($17) {
  $18 = ((($1)) + 16|0);
  $i$12 = 0;
  while(1) {
   $35 = (((($1)) + 40|0) + ($i$12<<3)|0);
   $36 = +HEAPF64[$35>>3];
   $37 = (((($gamepadState)) + 16|0) + ($i$12<<3)|0);
   $38 = +HEAPF64[$37>>3];
   $39 = $36 != $38;
   if ($39) {
    $40 = HEAP32[$18>>2]|0;
    $41 = $i$12&255;
    $42 = $38 * 32767.0;
    $43 = (~~(($42)));
    (_SDL_PrivateJoystickAxis($40,$41,$43)|0);
    $$pre5 = +HEAPF64[$37>>3];
    $44 = $$pre5;
   } else {
    $44 = $38;
   }
   HEAPF64[$35>>3] = $44;
   $45 = (($i$12) + 1)|0;
   $46 = HEAP32[$15>>2]|0;
   $47 = ($45|0)<($46|0);
   if ($47) {
    $i$12 = $45;
   } else {
    break;
   }
  }
 }
 $48 = +HEAPF64[$gamepadState>>3];
 $49 = ((($1)) + 32|0);
 HEAPF64[$49>>3] = $48;
 STACKTOP = sp;return;
}
function _SDL_PrivateJoystickButton($joystick,$button,$state) {
 $joystick = $joystick|0;
 $button = $button|0;
 $state = $state|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $event = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $event = sp;
 $0 = $state&255;
 switch ($0|0) {
 case 1:  {
  HEAP32[$event>>2] = 1539;
  break;
 }
 case 0:  {
  HEAP32[$event>>2] = 1540;
  break;
 }
 default: {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 }
 $1 = $button&255;
 $2 = ((($joystick)) + 32|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)<($3|0);
 if (!($4)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = ((($joystick)) + 36|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + ($1)|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = ($8<<24>>24)==($state<<24>>24);
 if ($9) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $10 = HEAP32[301486]|0;
 $11 = ($10|0)==(0);
 if ($11) {
  $12 = (_SDL_WasInit(32)|0);
  $13 = ($12|0)==(0);
  if (!($13)) {
   $14 = (_SDL_GetKeyboardFocus()|0);
   $15 = ($14|0)==(0|0);
   $16 = ($state<<24>>24)==(1);
   $or$cond = $16 & $15;
   if ($or$cond) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
  }
 }
 $17 = HEAP32[$5>>2]|0;
 $18 = (($17) + ($1)|0);
 HEAP8[$18>>0] = $state;
 $19 = HEAP32[$event>>2]|0;
 $20 = (_SDL_EventState($19,-1)|0);
 $21 = ($20<<24>>24)==(1);
 if (!($21)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $22 = HEAP32[$joystick>>2]|0;
 $23 = ((($event)) + 8|0);
 HEAP32[$23>>2] = $22;
 $24 = ((($event)) + 12|0);
 HEAP8[$24>>0] = $button;
 $25 = ((($event)) + 13|0);
 HEAP8[$25>>0] = $state;
 $26 = (_SDL_PushEvent($event)|0);
 $27 = ($26|0)==(1);
 $28 = $27&1;
 $$0 = $28;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_PrivateJoystickAxis($joystick,$axis,$value) {
 $joystick = $joystick|0;
 $axis = $axis|0;
 $value = $value|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $event = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $event = sp;
 $0 = $axis&255;
 $1 = ((($joystick)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($0|0)<($2|0);
 if (!($3)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ((($joystick)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($5) + ($0<<1)|0);
 $7 = HEAP16[$6>>1]|0;
 $8 = ($7<<16>>16)==($value<<16>>16);
 if ($8) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $9 = HEAP32[301486]|0;
 $10 = ($9|0)==(0);
 do {
  if ($10) {
   $11 = (_SDL_WasInit(32)|0);
   $12 = ($11|0)==(0);
   if (!($12)) {
    $13 = (_SDL_GetKeyboardFocus()|0);
    $14 = ($13|0)==(0|0);
    if ($14) {
     $15 = ($value<<16>>16)>(0);
     if ($15) {
      $16 = HEAP32[$4>>2]|0;
      $17 = (($16) + ($0<<1)|0);
      $18 = HEAP16[$17>>1]|0;
      $19 = ($18<<16>>16)>($value<<16>>16);
      if ($19) {
       break;
      } else {
       $$0 = 0;
      }
      STACKTOP = sp;return ($$0|0);
     }
     $20 = ($value<<16>>16)<(0);
     if ($20) {
      $21 = HEAP32[$4>>2]|0;
      $22 = (($21) + ($0<<1)|0);
      $23 = HEAP16[$22>>1]|0;
      $24 = ($23<<16>>16)<($value<<16>>16);
      if (!($24)) {
       $$0 = 0;
       STACKTOP = sp;return ($$0|0);
      }
     }
    }
   }
  }
 } while(0);
 $25 = HEAP32[$4>>2]|0;
 $26 = (($25) + ($0<<1)|0);
 HEAP16[$26>>1] = $value;
 $27 = (_SDL_EventState(1536,-1)|0);
 $28 = ($27<<24>>24)==(1);
 if (!($28)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$event>>2] = 1536;
 $29 = HEAP32[$joystick>>2]|0;
 $30 = ((($event)) + 8|0);
 HEAP32[$30>>2] = $29;
 $31 = ((($event)) + 12|0);
 HEAP8[$31>>0] = $axis;
 $32 = ((($event)) + 16|0);
 HEAP16[$32>>1] = $value;
 $33 = (_SDL_PushEvent($event)|0);
 $34 = ($33|0)==(1);
 $35 = $34&1;
 $$0 = $35;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SYS_JoystickDetect() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _SDL_SendPendingQuit() {
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[300736]|0;
 $1 = ($0|0)==(0);
 if ($1) {
  return;
 }
 HEAP32[300736] = 0;
 (_SDL_SendAppEvent(256)|0);
 return;
}
function _SDL_PollEvent($event) {
 $event = $event|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 L1: while(1) {
  $0 = (_SDL_GetVideoDevice()|0);
  $1 = ($0|0)==(0|0);
  if (!($1)) {
   $2 = ((($0)) + 180|0);
   $3 = HEAP32[$2>>2]|0;
   FUNCTION_TABLE_vi[$3 & 8191]($0);
  }
  $4 = HEAP32[(1201912)>>2]|0;
  $5 = ($4|0)==(0|0);
  if ($5) {
   label = 6;
  } else {
   $6 = (_SDL_JoystickEventState(-1)|0);
   $7 = ($6|0)==(0);
   if (!($7)) {
    label = 6;
   }
  }
  if ((label|0) == 6) {
   label = 0;
   _SDL_JoystickUpdate();
  }
  _SDL_SendPendingQuit();
  $8 = (_SDL_PeepEvents($event,1,2,0,65535)|0);
  switch ($8|0) {
  case 1:  {
   break L1;
   break;
  }
  case 0: case -1:  {
   $$0$i = 0;
   label = 9;
   break L1;
   break;
  }
  default: {
  }
  }
 }
 if ((label|0) == 9) {
  return ($$0$i|0);
 }
 $$0$i = 1;
 return ($$0$i|0);
}
function _SDL_WaitEvent($event) {
 $event = $event|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 L1: while(1) {
  $0 = (_SDL_GetVideoDevice()|0);
  $1 = ($0|0)==(0|0);
  if (!($1)) {
   $2 = ((($0)) + 180|0);
   $3 = HEAP32[$2>>2]|0;
   FUNCTION_TABLE_vi[$3 & 8191]($0);
  }
  $4 = HEAP32[(1201912)>>2]|0;
  $5 = ($4|0)==(0|0);
  if ($5) {
   label = 6;
  } else {
   $6 = (_SDL_JoystickEventState(-1)|0);
   $7 = ($6|0)==(0);
   if (!($7)) {
    label = 6;
   }
  }
  if ((label|0) == 6) {
   label = 0;
   _SDL_JoystickUpdate();
  }
  _SDL_SendPendingQuit();
  $8 = (_SDL_PeepEvents($event,1,2,0,65535)|0);
  switch ($8|0) {
  case 1:  {
   break L1;
   break;
  }
  case -1:  {
   $$0$i = 0;
   label = 10;
   break L1;
   break;
  }
  case 0:  {
   break;
  }
  default: {
   continue L1;
  }
  }
  _SDL_Delay(10);
 }
 if ((label|0) == 10) {
  return ($$0$i|0);
 }
 $$0$i = 1;
 return ($$0$i|0);
}
function _SDL_GetKeyboardState($numkeys) {
 $numkeys = $numkeys|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($numkeys|0)==(0|0);
 if (!($0)) {
  HEAP32[$numkeys>>2] = 512;
 }
 return ((1202966)|0);
}
function _SDL_GetKeyFromScancode($scancode) {
 $scancode = $scancode|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($scancode>>>0)>(511);
 if ($0) {
  HEAP32[$vararg_buffer>>2] = 611293;
  (_SDL_SetError(611267,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $1 = (1203480 + ($scancode<<2)|0);
  $2 = HEAP32[$1>>2]|0;
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_GetScancodeFromKey($key) {
 $key = $key|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $scancode$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $scancode$01 = 0;
 while(1) {
  $0 = (1203480 + ($scancode$01<<2)|0);
  $1 = HEAP32[$0>>2]|0;
  $2 = ($1|0)==($key|0);
  if ($2) {
   $$0 = $scancode$01;
   label = 4;
   break;
  }
  $3 = (($scancode$01) + 1)|0;
  $4 = ($3>>>0)<(512);
  if ($4) {
   $scancode$01 = $3;
  } else {
   $$0 = 0;
   label = 4;
   break;
  }
 }
 if ((label|0) == 4) {
  return ($$0|0);
 }
 return (0)|0;
}
function _SDL_SetCursor($cursor) {
 $cursor = $cursor|0;
 var $$01 = 0, $$not = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $found$0 = 0, $found$0$in = 0, $notlhs = 0, $notrhs = 0;
 var $or$cond = 0, $or$cond3$not = 0, $or$cond5 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($cursor|0)==(0|0);
 do {
  if ($0) {
   $5 = HEAP32[(1201820)>>2]|0;
   $6 = ($5|0)==(0|0);
   if ($6) {
    $8 = HEAP32[(1201872)>>2]|0;
    $$01 = $8;
    break;
   } else {
    $7 = HEAP32[(1201876)>>2]|0;
    $$01 = $7;
    break;
   }
  } else {
   $1 = HEAP32[(1201872)>>2]|0;
   $2 = ($1|0)==($cursor|0);
   if (!($2)) {
    $found$0$in = (1201868);
    while(1) {
     $found$0 = HEAP32[$found$0$in>>2]|0;
     $3 = ($found$0|0)!=(0|0);
     $$not = $3 ^ 1;
     $4 = ($found$0|0)==($cursor|0);
     $or$cond = $4 | $$not;
     if ($or$cond) {
      break;
     } else {
      $found$0$in = $found$0;
     }
    }
    if (!($3)) {
     (_SDL_SetError(611302,$vararg_buffer)|0);
     STACKTOP = sp;return;
    }
   }
   HEAP32[(1201876)>>2] = $cursor;
   $$01 = $cursor;
  }
 } while(0);
 $9 = HEAP32[(1201880)>>2]|0;
 $notlhs = ($$01|0)!=(0|0);
 $notrhs = ($9|0)!=(0);
 $or$cond3$not = $notlhs & $notrhs;
 $10 = HEAP32[(1201852)>>2]|0;
 $11 = ($10|0)==(0);
 $or$cond5 = $or$cond3$not & $11;
 $12 = HEAP32[(1201784)>>2]|0;
 $13 = ($12|0)==(0|0);
 if ($or$cond5) {
  if ($13) {
   STACKTOP = sp;return;
  }
  (FUNCTION_TABLE_ii[$12 & 8191]($$01)|0);
  STACKTOP = sp;return;
 } else {
  if ($13) {
   STACKTOP = sp;return;
  }
  (FUNCTION_TABLE_ii[$12 & 8191](0)|0);
  STACKTOP = sp;return;
 }
}
function _SDL_FreeCursor($cursor) {
 $cursor = $cursor|0;
 var $$ = 0, $$01$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $curr$0 = 0, $curr$03 = 0, $curr$05 = 0, $notlhs$i = 0, $notrhs$i = 0, $or$cond3$not$i = 0, $or$cond5$i = 0, $prev$04 = 0, $prev$04$lcssa = 0, $prev$04$phi = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($cursor|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = HEAP32[(1201872)>>2]|0;
 $2 = ($1|0)==($cursor|0);
 if ($2) {
  return;
 }
 $3 = HEAP32[(1201876)>>2]|0;
 $4 = ($3|0)==($cursor|0);
 do {
  if ($4) {
   $5 = ($1|0)==(0|0);
   if ($5) {
    $6 = HEAP32[(1201820)>>2]|0;
    $7 = ($6|0)==(0|0);
    $$ = $7 ? 0 : $cursor;
    $$01$i = $$;
   } else {
    HEAP32[(1201876)>>2] = $1;
    $$01$i = $1;
   }
   $8 = HEAP32[(1201880)>>2]|0;
   $notlhs$i = ($$01$i|0)!=(0|0);
   $notrhs$i = ($8|0)!=(0);
   $or$cond3$not$i = $notlhs$i & $notrhs$i;
   $9 = HEAP32[(1201852)>>2]|0;
   $10 = ($9|0)==(0);
   $or$cond5$i = $or$cond3$not$i & $10;
   $11 = HEAP32[(1201784)>>2]|0;
   $12 = ($11|0)==(0|0);
   if ($or$cond5$i) {
    if ($12) {
     break;
    }
    (FUNCTION_TABLE_ii[$11 & 8191]($$01$i)|0);
    break;
   } else {
    if ($12) {
     break;
    }
    (FUNCTION_TABLE_ii[$11 & 8191](0)|0);
    break;
   }
  }
 } while(0);
 $curr$03 = HEAP32[(1201868)>>2]|0;
 $13 = ($curr$03|0)==(0|0);
 if ($13) {
  return;
 } else {
  $curr$05 = $curr$03;$prev$04 = 0;
 }
 while(1) {
  $14 = ($curr$05|0)==($cursor|0);
  if ($14) {
   $prev$04$lcssa = $prev$04;
   break;
  }
  $curr$0 = HEAP32[$curr$05>>2]|0;
  $19 = ($curr$0|0)==(0|0);
  if ($19) {
   label = 20;
   break;
  } else {
   $prev$04$phi = $curr$05;$curr$05 = $curr$0;$prev$04 = $prev$04$phi;
  }
 }
 if ((label|0) == 20) {
  return;
 }
 $15 = ($prev$04$lcssa|0)==(0|0);
 $16 = HEAP32[$cursor>>2]|0;
 if ($15) {
  HEAP32[(1201868)>>2] = $16;
 } else {
  HEAP32[$prev$04$lcssa>>2] = $16;
 }
 $17 = HEAP32[(1201792)>>2]|0;
 $18 = ($17|0)==(0|0);
 if ($18) {
  return;
 }
 FUNCTION_TABLE_vi[$17 & 8191]($cursor);
 return;
}
function _SDL_CreateColorCursor($surface,$hot_x,$hot_y) {
 $surface = $surface|0;
 $hot_x = $hot_x|0;
 $hot_y = $hot_y|0;
 var $$0 = 0, $$02 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $temp$0 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($surface|0)==(0|0);
 if ($0) {
  (_SDL_SetError(611425,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = HEAP32[300444]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  (_SDL_SetError(611347,$vararg_buffer1)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = $hot_y | $hot_x;
 $4 = ($3|0)<(0);
 if (!($4)) {
  $5 = ((($surface)) + 8|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6|0)>($hot_x|0);
  if ($7) {
   $8 = ((($surface)) + 12|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = ($9|0)>($hot_y|0);
   if ($10) {
    $11 = ((($surface)) + 4|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = HEAP32[$12>>2]|0;
    $14 = ($13|0)==(372645892);
    do {
     if ($14) {
      $$02 = $surface;$17 = $1;$temp$0 = 0;
     } else {
      $15 = (_SDL_ConvertSurfaceFormat($surface,372645892,0)|0);
      $16 = ($15|0)==(0|0);
      if ($16) {
       $$0 = 0;
       STACKTOP = sp;return ($$0|0);
      } else {
       $$pre = HEAP32[300444]|0;
       $$02 = $15;$17 = $$pre;$temp$0 = $15;
       break;
      }
     }
    } while(0);
    $18 = (FUNCTION_TABLE_iiii[$17 & 8191]($$02,$hot_x,$hot_y)|0);
    $19 = ($18|0)==(0|0);
    if (!($19)) {
     $20 = HEAP32[(1201868)>>2]|0;
     HEAP32[$18>>2] = $20;
     HEAP32[(1201868)>>2] = $18;
    }
    _SDL_FreeSurface($temp$0);
    $$0 = $18;
    STACKTOP = sp;return ($$0|0);
   }
  }
 }
 (_SDL_SetError(611383,$vararg_buffer3)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_CreateSystemCursor($id) {
 $id = $id|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP32[(1201780)>>2]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(611452,$vararg_buffer)|0);
  $$0 = 0;
 } else {
  $2 = (FUNCTION_TABLE_ii[$0 & 8191]($id)|0);
  $3 = ($2|0)==(0|0);
  if ($3) {
   $$0 = 0;
  } else {
   $4 = HEAP32[(1201868)>>2]|0;
   HEAP32[$2>>2] = $4;
   HEAP32[(1201868)>>2] = $2;
   $$0 = $2;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GetDefaultCursor() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[(1201872)>>2]|0;
 return ($0|0);
}
function _SDL_SYS_NumHaptics() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _SDL_HapticIndex($haptic) {
 $haptic = $haptic|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $hapticlist$0$i = 0, $hapticlist$04$i = 0, $hapticlist$05$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($haptic|0)==(0|0);
 L1: do {
  if (!($0)) {
   $hapticlist$04$i = HEAP32[300440]|0;
   $1 = ($hapticlist$04$i|0)==(0|0);
   if (!($1)) {
    $hapticlist$05$i = $hapticlist$04$i;
    while(1) {
     $4 = ($hapticlist$05$i|0)==($haptic|0);
     if ($4) {
      break;
     }
     $2 = ((($hapticlist$05$i)) + 104|0);
     $hapticlist$0$i = HEAP32[$2>>2]|0;
     $3 = ($hapticlist$0$i|0)==(0|0);
     if ($3) {
      break L1;
     } else {
      $hapticlist$05$i = $hapticlist$0$i;
     }
    }
    $5 = HEAP8[$haptic>>0]|0;
    $6 = $5&255;
    $$0 = $6;
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 (_SDL_SetError(611586,$vararg_buffer)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_PrivateJoystickValid($joystick) {
 $joystick = $joystick|0;
 var $0 = 0, $valid$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($joystick|0)==(0|0);
 if ($0) {
  (_SDL_SetError(611627,$vararg_buffer)|0);
  $valid$0 = 0;
 } else {
  $valid$0 = 1;
 }
 STACKTOP = sp;return ($valid$0|0);
}
function _SDL_SYS_JoystickIsHaptic($joystick) {
 $joystick = $joystick|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _SDL_HapticOpenFromJoystick($joystick) {
 $joystick = $joystick|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hapticlist$0 = 0, $hapticlist$02 = 0, $hapticlist$03 = 0, $hapticlist$03$lcssa = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_SDL_SYS_NumHaptics()|0);
 $1 = ($0|0)<(1);
 if ($1) {
  $2 = (_SDL_SYS_NumHaptics()|0);
  HEAP32[$vararg_buffer>>2] = $2;
  (_SDL_SetError(611498,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = (_SDL_PrivateJoystickValid($joystick)|0);
 $4 = ($3|0)==(0);
 if ($4) {
  (_SDL_SetError(611659,$vararg_buffer1)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = (_SDL_SYS_JoystickIsHaptic($joystick)|0);
 $6 = ($5|0)<(1);
 if ($6) {
  (_SDL_SetError(611689,$vararg_buffer3)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $hapticlist$02 = HEAP32[300440]|0;
 $7 = ($hapticlist$02|0)==(0|0);
 L13: do {
  if (!($7)) {
   $hapticlist$03 = $hapticlist$02;
   while(1) {
    $8 = (_SDL_SYS_JoystickSameHaptic($hapticlist$03,$joystick)|0);
    $9 = ($8|0)==(0);
    if (!($9)) {
     $hapticlist$03$lcssa = $hapticlist$03;
     break;
    }
    $13 = ((($hapticlist$03)) + 104|0);
    $hapticlist$0 = HEAP32[$13>>2]|0;
    $14 = ($hapticlist$0|0)==(0|0);
    if ($14) {
     break L13;
    } else {
     $hapticlist$03 = $hapticlist$0;
    }
   }
   $10 = ((($hapticlist$03$lcssa)) + 28|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = (($11) + 1)|0;
   HEAP32[$10>>2] = $12;
   $$0 = $hapticlist$03$lcssa;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 $15 = (_SDL_malloc(108)|0);
 $16 = ($15|0)==(0|0);
 if ($16) {
  (_SDL_Error(0)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 (_SDL_memset($15,0,108)|0);
 $17 = ((($15)) + 32|0);
 HEAP32[$17>>2] = -1;
 $18 = (_SDL_SYS_HapticOpenFromJoystick($15,$joystick)|0);
 $19 = ($18|0)<(0);
 if ($19) {
  _SDL_free($15);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $20 = ((($15)) + 28|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = (($21) + 1)|0;
  HEAP32[$20>>2] = $22;
  $23 = HEAP32[300440]|0;
  $24 = ((($15)) + 104|0);
  HEAP32[$24>>2] = $23;
  HEAP32[300440] = $15;
  $$0 = $15;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_SYS_JoystickSameHaptic($haptic,$joystick) {
 $haptic = $haptic|0;
 $joystick = $joystick|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _SDL_SYS_HapticOpenFromJoystick($haptic,$joystick) {
 $haptic = $haptic|0;
 $joystick = $joystick|0;
 var $0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_SDL_SetError(611544,$vararg_buffer)|0);
 STACKTOP = sp;return ($0|0);
}
function _SDL_HapticClose($haptic) {
 $haptic = $haptic|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hapticlist$0 = 0, $hapticlist$0$i = 0, $hapticlist$0$i$i = 0, $hapticlist$03 = 0, $hapticlist$04$i = 0;
 var $hapticlist$04$i$i = 0, $hapticlist$05 = 0, $hapticlist$05$i = 0, $hapticlist$05$i$i = 0, $hapticlistprev$04 = 0, $hapticlistprev$04$lcssa = 0, $hapticlistprev$04$phi = 0, $i$06 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($haptic|0)==(0|0);
 L1: do {
  if (!($0)) {
   $hapticlist$04$i = HEAP32[300440]|0;
   $1 = ($hapticlist$04$i|0)==(0|0);
   if (!($1)) {
    $hapticlist$05$i = $hapticlist$04$i;
    while(1) {
     $4 = ($hapticlist$05$i|0)==($haptic|0);
     if ($4) {
      break;
     }
     $2 = ((($hapticlist$05$i)) + 104|0);
     $hapticlist$0$i = HEAP32[$2>>2]|0;
     $3 = ($hapticlist$0$i|0)==(0|0);
     if ($3) {
      break L1;
     } else {
      $hapticlist$05$i = $hapticlist$0$i;
     }
    }
    $5 = ((($haptic)) + 28|0);
    $6 = HEAP32[$5>>2]|0;
    $7 = (($6) + -1)|0;
    HEAP32[$5>>2] = $7;
    $8 = ($6|0)>(1);
    if ($8) {
     STACKTOP = sp;return;
    }
    $9 = ((($haptic)) + 8|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = ($10|0)>(0);
    if ($11) {
     $12 = ((($haptic)) + 4|0);
     $22 = $10;$i$06 = 0;
     while(1) {
      $13 = HEAP32[$12>>2]|0;
      $14 = (((($13) + (($i$06*72)|0)|0)) + 68|0);
      $15 = HEAP32[$14>>2]|0;
      $16 = ($15|0)==(0|0);
      L14: do {
       if (!($16)) {
        $hapticlist$04$i$i = HEAP32[300440]|0;
        $17 = ($hapticlist$04$i$i|0)==(0|0);
        L16: do {
         if (!($17)) {
          $hapticlist$05$i$i = $hapticlist$04$i$i;
          while(1) {
           $20 = ($hapticlist$05$i$i|0)==($haptic|0);
           if ($20) {
            break;
           }
           $18 = ((($hapticlist$05$i$i)) + 104|0);
           $hapticlist$0$i$i = HEAP32[$18>>2]|0;
           $19 = ($hapticlist$0$i$i|0)==(0|0);
           if ($19) {
            break L16;
           } else {
            $hapticlist$05$i$i = $hapticlist$0$i$i;
           }
          }
          $21 = ($22|0)>($i$06|0);
          if ($21) {
           $23 = (($13) + (($i$06*72)|0)|0);
           _SDL_SYS_HapticDestroyEffect($haptic,$23);
           break L14;
          } else {
           (_SDL_SetError(611729,$vararg_buffer3)|0);
           break L14;
          }
         }
        } while(0);
        (_SDL_SetError(611586,$vararg_buffer1)|0);
       }
      } while(0);
      $24 = (($i$06) + 1)|0;
      $25 = HEAP32[$9>>2]|0;
      $26 = ($24|0)<($25|0);
      if ($26) {
       $22 = $25;$i$06 = $24;
      } else {
       break;
      }
     }
    }
    $hapticlist$03 = HEAP32[300440]|0;
    $27 = ($hapticlist$03|0)==(0|0);
    L27: do {
     if (!($27)) {
      $hapticlist$05 = $hapticlist$03;$hapticlistprev$04 = 0;
      while(1) {
       $28 = ($hapticlist$05|0)==($haptic|0);
       if ($28) {
        $hapticlistprev$04$lcssa = $hapticlistprev$04;
        break;
       }
       $33 = ((($hapticlist$05)) + 104|0);
       $hapticlist$0 = HEAP32[$33>>2]|0;
       $34 = ($hapticlist$0|0)==(0|0);
       if ($34) {
        break L27;
       } else {
        $hapticlistprev$04$phi = $hapticlist$05;$hapticlist$05 = $hapticlist$0;$hapticlistprev$04 = $hapticlistprev$04$phi;
       }
      }
      $29 = ($hapticlistprev$04$lcssa|0)==(0|0);
      $30 = ((($haptic)) + 104|0);
      $31 = HEAP32[$30>>2]|0;
      if ($29) {
       HEAP32[300440] = $31;
       break;
      } else {
       $32 = ((($hapticlistprev$04$lcssa)) + 104|0);
       HEAP32[$32>>2] = $31;
       break;
      }
     }
    } while(0);
    _SDL_free($haptic);
    STACKTOP = sp;return;
   }
  }
 } while(0);
 (_SDL_SetError(611586,$vararg_buffer)|0);
 STACKTOP = sp;return;
}
function _SDL_SYS_HapticDestroyEffect($haptic,$effect) {
 $haptic = $haptic|0;
 $effect = $effect|0;
 var $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 (_SDL_SetError(611544,$vararg_buffer)|0);
 STACKTOP = sp;return;
}
function _SDL_SYS_HapticClose($haptic) {
 $haptic = $haptic|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _SDL_HapticDestroyEffect($haptic,$effect) {
 $haptic = $haptic|0;
 $effect = $effect|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hapticlist$0$i = 0, $hapticlist$04$i = 0, $hapticlist$05$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($haptic|0)==(0|0);
 L1: do {
  if (!($0)) {
   $hapticlist$04$i = HEAP32[300440]|0;
   $1 = ($hapticlist$04$i|0)==(0|0);
   if (!($1)) {
    $hapticlist$05$i = $hapticlist$04$i;
    while(1) {
     $4 = ($hapticlist$05$i|0)==($haptic|0);
     if ($4) {
      break;
     }
     $2 = ((($hapticlist$05$i)) + 104|0);
     $hapticlist$0$i = HEAP32[$2>>2]|0;
     $3 = ($hapticlist$0$i|0)==(0|0);
     if ($3) {
      break L1;
     } else {
      $hapticlist$05$i = $hapticlist$0$i;
     }
    }
    $5 = ($effect|0)<(0);
    if (!($5)) {
     $6 = ((($haptic)) + 8|0);
     $7 = HEAP32[$6>>2]|0;
     $8 = ($7|0)>($effect|0);
     if ($8) {
      $9 = ((($haptic)) + 4|0);
      $10 = HEAP32[$9>>2]|0;
      $11 = (((($10) + (($effect*72)|0)|0)) + 68|0);
      $12 = HEAP32[$11>>2]|0;
      $13 = ($12|0)==(0|0);
      if ($13) {
       STACKTOP = sp;return;
      }
      $14 = (($10) + (($effect*72)|0)|0);
      _SDL_SYS_HapticDestroyEffect($haptic,$14);
      STACKTOP = sp;return;
     }
    }
    (_SDL_SetError(611729,$vararg_buffer1)|0);
    STACKTOP = sp;return;
   }
  }
 } while(0);
 (_SDL_SetError(611586,$vararg_buffer)|0);
 STACKTOP = sp;return;
}
function _SDL_HapticQuery($haptic) {
 $haptic = $haptic|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $hapticlist$0$i = 0, $hapticlist$04$i = 0, $hapticlist$05$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($haptic|0)==(0|0);
 L1: do {
  if (!($0)) {
   $hapticlist$04$i = HEAP32[300440]|0;
   $1 = ($hapticlist$04$i|0)==(0|0);
   if (!($1)) {
    $hapticlist$05$i = $hapticlist$04$i;
    while(1) {
     $4 = ($hapticlist$05$i|0)==($haptic|0);
     if ($4) {
      break;
     }
     $2 = ((($hapticlist$05$i)) + 104|0);
     $hapticlist$0$i = HEAP32[$2>>2]|0;
     $3 = ($hapticlist$0$i|0)==(0|0);
     if ($3) {
      break L1;
     } else {
      $hapticlist$05$i = $hapticlist$0$i;
     }
    }
    $5 = ((($haptic)) + 16|0);
    $6 = HEAP32[$5>>2]|0;
    $$0 = $6;
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 (_SDL_SetError(611586,$vararg_buffer)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_HapticNumAxes($haptic) {
 $haptic = $haptic|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $hapticlist$0$i = 0, $hapticlist$04$i = 0, $hapticlist$05$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($haptic|0)==(0|0);
 L1: do {
  if (!($0)) {
   $hapticlist$04$i = HEAP32[300440]|0;
   $1 = ($hapticlist$04$i|0)==(0|0);
   if (!($1)) {
    $hapticlist$05$i = $hapticlist$04$i;
    while(1) {
     $4 = ($hapticlist$05$i|0)==($haptic|0);
     if ($4) {
      break;
     }
     $2 = ((($hapticlist$05$i)) + 104|0);
     $hapticlist$0$i = HEAP32[$2>>2]|0;
     $3 = ($hapticlist$0$i|0)==(0|0);
     if ($3) {
      break L1;
     } else {
      $hapticlist$05$i = $hapticlist$0$i;
     }
    }
    $5 = ((($haptic)) + 20|0);
    $6 = HEAP32[$5>>2]|0;
    $$0 = $6;
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 (_SDL_SetError(611586,$vararg_buffer)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_HapticNewEffect($haptic,$effect) {
 $haptic = $haptic|0;
 $effect = $effect|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hapticlist$0$i = 0, $hapticlist$0$i$i = 0, $hapticlist$04$i = 0, $hapticlist$05$i = 0, $hapticlist$05$i$i = 0, $i$03 = 0, $i$03$lcssa = 0, $not$$i = 0, $vararg_buffer = 0;
 var $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($haptic|0)==(0|0);
 L1: do {
  if (!($0)) {
   $hapticlist$04$i = HEAP32[300440]|0;
   $1 = ($hapticlist$04$i|0)==(0|0);
   if (!($1)) {
    $hapticlist$05$i = $hapticlist$04$i;
    while(1) {
     $4 = ($hapticlist$05$i|0)==($haptic|0);
     if ($4) {
      $hapticlist$05$i$i = $hapticlist$04$i;
      break;
     }
     $2 = ((($hapticlist$05$i)) + 104|0);
     $hapticlist$0$i = HEAP32[$2>>2]|0;
     $3 = ($hapticlist$0$i|0)==(0|0);
     if ($3) {
      break L1;
     } else {
      $hapticlist$05$i = $hapticlist$0$i;
     }
    }
    while(1) {
     $7 = ($hapticlist$05$i$i|0)==($haptic|0);
     if ($7) {
      label = 9;
      break;
     }
     $5 = ((($hapticlist$05$i$i)) + 104|0);
     $hapticlist$0$i$i = HEAP32[$5>>2]|0;
     $6 = ($hapticlist$0$i$i|0)==(0|0);
     if ($6) {
      label = 8;
      break;
     } else {
      $hapticlist$05$i$i = $hapticlist$0$i$i;
     }
    }
    if ((label|0) == 8) {
     (_SDL_SetError(611586,$vararg_buffer1)|0);
    }
    else if ((label|0) == 9) {
     $8 = ((($haptic)) + 16|0);
     $9 = HEAP32[$8>>2]|0;
     $10 = HEAP16[$effect>>1]|0;
     $11 = $10&65535;
     $12 = $11 & $9;
     $not$$i = ($12|0)==(0);
     if ($not$$i) {
      $18 = (_SDL_SetError(611764,$vararg_buffer3)|0);
      $$0 = $18;
      STACKTOP = sp;return ($$0|0);
     }
    }
    $13 = ((($haptic)) + 8|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = ($14|0)>(0);
    L15: do {
     if ($15) {
      $16 = ((($haptic)) + 4|0);
      $17 = HEAP32[$16>>2]|0;
      $i$03 = 0;
      while(1) {
       $19 = (((($17) + (($i$03*72)|0)|0)) + 68|0);
       $20 = HEAP32[$19>>2]|0;
       $21 = ($20|0)==(0|0);
       if ($21) {
        $i$03$lcssa = $i$03;
        break;
       }
       $27 = (($i$03) + 1)|0;
       $28 = ($27|0)<($14|0);
       if ($28) {
        $i$03 = $27;
       } else {
        break L15;
       }
      }
      $22 = (($17) + (($i$03$lcssa*72)|0)|0);
      $23 = (_SDL_SYS_HapticNewEffect($haptic,$22,$effect)|0);
      $24 = ($23|0)==(0);
      if (!($24)) {
       $$0 = -1;
       STACKTOP = sp;return ($$0|0);
      }
      $25 = HEAP32[$16>>2]|0;
      $26 = (($25) + (($i$03$lcssa*72)|0)|0);
      (_SDL_memcpy($26,$effect,68)|0);
      $$0 = $i$03$lcssa;
      STACKTOP = sp;return ($$0|0);
     }
    } while(0);
    $29 = (_SDL_SetError(611811,$vararg_buffer5)|0);
    $$0 = $29;
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 (_SDL_SetError(611586,$vararg_buffer)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SYS_HapticNewEffect($haptic,$effect,$base) {
 $haptic = $haptic|0;
 $effect = $effect|0;
 $base = $base|0;
 var $0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_SDL_SetError(611544,$vararg_buffer)|0);
 STACKTOP = sp;return ($0|0);
}
function _SDL_HapticUpdateEffect($haptic,$effect,$data) {
 $haptic = $haptic|0;
 $effect = $effect|0;
 $data = $data|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $hapticlist$0$i = 0, $hapticlist$04$i = 0, $hapticlist$05$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($haptic|0)==(0|0);
 L1: do {
  if (!($0)) {
   $hapticlist$04$i = HEAP32[300440]|0;
   $1 = ($hapticlist$04$i|0)==(0|0);
   if (!($1)) {
    $hapticlist$05$i = $hapticlist$04$i;
    while(1) {
     $4 = ($hapticlist$05$i|0)==($haptic|0);
     if ($4) {
      break;
     }
     $2 = ((($hapticlist$05$i)) + 104|0);
     $hapticlist$0$i = HEAP32[$2>>2]|0;
     $3 = ($hapticlist$0$i|0)==(0|0);
     if ($3) {
      break L1;
     } else {
      $hapticlist$05$i = $hapticlist$0$i;
     }
    }
    $5 = ($effect|0)<(0);
    if (!($5)) {
     $6 = ((($haptic)) + 8|0);
     $7 = HEAP32[$6>>2]|0;
     $8 = ($7|0)>($effect|0);
     if ($8) {
      $9 = HEAP16[$data>>1]|0;
      $10 = ((($haptic)) + 4|0);
      $11 = HEAP32[$10>>2]|0;
      $12 = (($11) + (($effect*72)|0)|0);
      $13 = HEAP16[$12>>1]|0;
      $14 = ($9<<16>>16)==($13<<16>>16);
      if (!($14)) {
       $15 = (_SDL_SetError(611850,$vararg_buffer3)|0);
       $$0 = $15;
       STACKTOP = sp;return ($$0|0);
      }
      $16 = (_SDL_SYS_HapticUpdateEffect($haptic,$12,$data)|0);
      $17 = ($16|0)<(0);
      if ($17) {
       $$0 = -1;
       STACKTOP = sp;return ($$0|0);
      }
      $18 = HEAP32[$10>>2]|0;
      $19 = (($18) + (($effect*72)|0)|0);
      (_SDL_memcpy($19,$data,68)|0);
      $$0 = 0;
      STACKTOP = sp;return ($$0|0);
     }
    }
    (_SDL_SetError(611729,$vararg_buffer1)|0);
    $$0 = -1;
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 (_SDL_SetError(611586,$vararg_buffer)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SYS_HapticUpdateEffect($haptic,$effect,$data) {
 $haptic = $haptic|0;
 $effect = $effect|0;
 $data = $data|0;
 var $0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_SDL_SetError(611544,$vararg_buffer)|0);
 STACKTOP = sp;return ($0|0);
}
function _SDL_HapticRunEffect($haptic,$effect,$iterations) {
 $haptic = $haptic|0;
 $effect = $effect|0;
 $iterations = $iterations|0;
 var $$0 = 0, $$lobit = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hapticlist$0$i = 0, $hapticlist$04$i = 0, $hapticlist$05$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($haptic|0)==(0|0);
 L1: do {
  if (!($0)) {
   $hapticlist$04$i = HEAP32[300440]|0;
   $1 = ($hapticlist$04$i|0)==(0|0);
   if (!($1)) {
    $hapticlist$05$i = $hapticlist$04$i;
    while(1) {
     $4 = ($hapticlist$05$i|0)==($haptic|0);
     if ($4) {
      break;
     }
     $2 = ((($hapticlist$05$i)) + 104|0);
     $hapticlist$0$i = HEAP32[$2>>2]|0;
     $3 = ($hapticlist$0$i|0)==(0|0);
     if ($3) {
      break L1;
     } else {
      $hapticlist$05$i = $hapticlist$0$i;
     }
    }
    $5 = ($effect|0)<(0);
    if (!($5)) {
     $6 = ((($haptic)) + 8|0);
     $7 = HEAP32[$6>>2]|0;
     $8 = ($7|0)>($effect|0);
     if ($8) {
      $9 = ((($haptic)) + 4|0);
      $10 = HEAP32[$9>>2]|0;
      $11 = (($10) + (($effect*72)|0)|0);
      $12 = (_SDL_SYS_HapticRunEffect($haptic,$11,$iterations)|0);
      $$lobit = $12 >> 31;
      $$0 = $$lobit;
      STACKTOP = sp;return ($$0|0);
     }
    }
    (_SDL_SetError(611729,$vararg_buffer1)|0);
    $$0 = -1;
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 (_SDL_SetError(611586,$vararg_buffer)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SYS_HapticRunEffect($haptic,$effect,$iterations) {
 $haptic = $haptic|0;
 $effect = $effect|0;
 $iterations = $iterations|0;
 var $0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_SDL_SetError(611544,$vararg_buffer)|0);
 STACKTOP = sp;return ($0|0);
}
function _SDL_HapticStopEffect($haptic,$effect) {
 $haptic = $haptic|0;
 $effect = $effect|0;
 var $$0 = 0, $$lobit = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hapticlist$0$i = 0, $hapticlist$04$i = 0, $hapticlist$05$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($haptic|0)==(0|0);
 L1: do {
  if (!($0)) {
   $hapticlist$04$i = HEAP32[300440]|0;
   $1 = ($hapticlist$04$i|0)==(0|0);
   if (!($1)) {
    $hapticlist$05$i = $hapticlist$04$i;
    while(1) {
     $4 = ($hapticlist$05$i|0)==($haptic|0);
     if ($4) {
      break;
     }
     $2 = ((($hapticlist$05$i)) + 104|0);
     $hapticlist$0$i = HEAP32[$2>>2]|0;
     $3 = ($hapticlist$0$i|0)==(0|0);
     if ($3) {
      break L1;
     } else {
      $hapticlist$05$i = $hapticlist$0$i;
     }
    }
    $5 = ($effect|0)<(0);
    if (!($5)) {
     $6 = ((($haptic)) + 8|0);
     $7 = HEAP32[$6>>2]|0;
     $8 = ($7|0)>($effect|0);
     if ($8) {
      $9 = ((($haptic)) + 4|0);
      $10 = HEAP32[$9>>2]|0;
      $11 = (($10) + (($effect*72)|0)|0);
      $12 = (_SDL_SYS_HapticStopEffect($haptic,$11)|0);
      $$lobit = $12 >> 31;
      $$0 = $$lobit;
      STACKTOP = sp;return ($$0|0);
     }
    }
    (_SDL_SetError(611729,$vararg_buffer1)|0);
    $$0 = -1;
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 (_SDL_SetError(611586,$vararg_buffer)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SYS_HapticStopEffect($haptic,$effect) {
 $haptic = $haptic|0;
 $effect = $effect|0;
 var $0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_SDL_SetError(611544,$vararg_buffer)|0);
 STACKTOP = sp;return ($0|0);
}
function _SDL_HapticGetEffectStatus($haptic,$effect) {
 $haptic = $haptic|0;
 $effect = $effect|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $hapticlist$0$i = 0;
 var $hapticlist$04$i = 0, $hapticlist$05$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($haptic|0)==(0|0);
 L1: do {
  if (!($0)) {
   $hapticlist$04$i = HEAP32[300440]|0;
   $1 = ($hapticlist$04$i|0)==(0|0);
   if (!($1)) {
    $hapticlist$05$i = $hapticlist$04$i;
    while(1) {
     $4 = ($hapticlist$05$i|0)==($haptic|0);
     if ($4) {
      break;
     }
     $2 = ((($hapticlist$05$i)) + 104|0);
     $hapticlist$0$i = HEAP32[$2>>2]|0;
     $3 = ($hapticlist$0$i|0)==(0|0);
     if ($3) {
      break L1;
     } else {
      $hapticlist$05$i = $hapticlist$0$i;
     }
    }
    $5 = ($effect|0)<(0);
    if (!($5)) {
     $6 = ((($haptic)) + 8|0);
     $7 = HEAP32[$6>>2]|0;
     $8 = ($7|0)>($effect|0);
     if ($8) {
      $9 = ((($haptic)) + 16|0);
      $10 = HEAP32[$9>>2]|0;
      $11 = $10 & 16384;
      $12 = ($11|0)==(0);
      if ($12) {
       $13 = (_SDL_SetError(611891,$vararg_buffer3)|0);
       $$0 = $13;
       STACKTOP = sp;return ($$0|0);
      } else {
       $14 = ((($haptic)) + 4|0);
       $15 = HEAP32[$14>>2]|0;
       $16 = (($15) + (($effect*72)|0)|0);
       $17 = (_SDL_SYS_HapticGetEffectStatus($haptic,$16)|0);
       $$0 = $17;
       STACKTOP = sp;return ($$0|0);
      }
     }
    }
    (_SDL_SetError(611729,$vararg_buffer1)|0);
    $$0 = -1;
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 (_SDL_SetError(611586,$vararg_buffer)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SYS_HapticGetEffectStatus($haptic,$effect) {
 $haptic = $haptic|0;
 $effect = $effect|0;
 var $0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_SDL_SetError(611544,$vararg_buffer)|0);
 STACKTOP = sp;return ($0|0);
}
function _SDL_IsGameController($device_index) {
 $device_index = $device_index|0;
 var $$ = 0, $0 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_PrivateGetControllerMapping($device_index)|0);
 $not$ = ($0|0)!=(0|0);
 $$ = $not$&1;
 return ($$|0);
}
function _SDL_GameControllerGetStringForAxis($axis) {
 $axis = $axis|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($axis>>>0)<(6);
 if (!($0)) {
  $$0 = 0;
  return ($$0|0);
 }
 $1 = (141012 + ($axis<<2)|0);
 $2 = HEAP32[$1>>2]|0;
 $$0 = $2;
 return ($$0|0);
}
function _SDL_GameControllerGetStringForButton($axis) {
 $axis = $axis|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($axis>>>0)<(15);
 if (!($0)) {
  $$0 = 0;
  return ($$0|0);
 }
 $1 = (141040 + ($axis<<2)|0);
 $2 = HEAP32[$1>>2]|0;
 $$0 = $2;
 return ($$0|0);
}
function _SDL_GameControllerMappingForGUID($guid) {
 $guid = $guid|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $guid$byval_copy = 0, $pSupportedController$0$i = 0, $pSupportedController$01$i = 0, $pSupportedController$02$i = 0, $pSupportedController$02$i$lcssa = 0, $pchGUID = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $guid$byval_copy = sp + 48|0;
 $vararg_buffer = sp;
 $pchGUID = sp + 12|0;
 $pSupportedController$01$i = HEAP32[301460]|0;
 $0 = ($pSupportedController$01$i|0)==(0|0);
 if ($0) {
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 } else {
  $pSupportedController$02$i = $pSupportedController$01$i;
 }
 while(1) {
  $1 = (_SDL_memcmp($guid,$pSupportedController$02$i,16)|0);
  $2 = ($1|0)==(0);
  if ($2) {
   $pSupportedController$02$i$lcssa = $pSupportedController$02$i;
   break;
  }
  $3 = ((($pSupportedController$02$i)) + 24|0);
  $pSupportedController$0$i = HEAP32[$3>>2]|0;
  $4 = ($pSupportedController$0$i|0)==(0|0);
  if ($4) {
   $$1 = 0;
   label = 7;
   break;
  } else {
   $pSupportedController$02$i = $pSupportedController$0$i;
  }
 }
 if ((label|0) == 7) {
  STACKTOP = sp;return ($$1|0);
 }
 dest=$guid$byval_copy; src=$guid; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
 _SDL_JoystickGetGUIDString($guid$byval_copy,$pchGUID,33);
 $5 = (_SDL_strlen($pchGUID)|0);
 $6 = ((($pSupportedController$02$i$lcssa)) + 16|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (_SDL_strlen($7)|0);
 $9 = ((($pSupportedController$02$i$lcssa)) + 20|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (_SDL_strlen($10)|0);
 $12 = (($5) + 3)|0;
 $13 = (($12) + ($8))|0;
 $14 = (($13) + ($11))|0;
 $15 = (_SDL_malloc($14)|0);
 $16 = ($15|0)==(0|0);
 if ($16) {
  (_SDL_Error(0)|0);
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 } else {
  $17 = HEAP32[$6>>2]|0;
  $18 = HEAP32[$9>>2]|0;
  HEAP32[$vararg_buffer>>2] = $pchGUID;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $17;
  $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
  HEAP32[$vararg_ptr2>>2] = $18;
  (_SDL_snprintf($15,$14,611956,$vararg_buffer)|0);
  $$1 = $15;
  STACKTOP = sp;return ($$1|0);
 }
 return (0)|0;
}
function _SDL_JoystickGetGUIDString($guid,$pszGUID,$cbGUID) {
 $guid = $guid|0;
 $pszGUID = $pszGUID|0;
 $cbGUID = $cbGUID|0;
 var $$0$lcssa = 0, $$04 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $i$03 = 0, $or$cond = 0, $or$cond2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($pszGUID|0)==(0|0);
 $1 = ($cbGUID|0)<(1);
 $or$cond = $0 | $1;
 if ($or$cond) {
  return;
 }
 $2 = (($cbGUID) + -1)|0;
 $3 = (($2|0) / 2)&-1;
 $4 = ($2|0)>(1);
 if ($4) {
  $$04 = $pszGUID;$i$03 = 0;
  while(1) {
   $5 = (($guid) + ($i$03)|0);
   $6 = HEAP8[$5>>0]|0;
   $7 = $6&255;
   $8 = $7 >>> 4;
   $9 = (611939 + ($8)|0);
   $10 = HEAP8[$9>>0]|0;
   $11 = ((($$04)) + 1|0);
   HEAP8[$$04>>0] = $10;
   $12 = $7 & 15;
   $13 = (611939 + ($12)|0);
   $14 = HEAP8[$13>>0]|0;
   $15 = ((($$04)) + 2|0);
   HEAP8[$11>>0] = $14;
   $16 = (($i$03) + 1)|0;
   $17 = ($16>>>0)<(16);
   $18 = ($16|0)<($3|0);
   $or$cond2 = $17 & $18;
   if ($or$cond2) {
    $$04 = $15;$i$03 = $16;
   } else {
    $$0$lcssa = $15;
    break;
   }
  }
 } else {
  $$0$lcssa = $pszGUID;
 }
 HEAP8[$$0$lcssa>>0] = 0;
 return;
}
function _SDL_GameControllerOpen($device_index) {
 $device_index = $device_index|0;
 var $$0 = 0, $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $gamecontrollerlist$0 = 0, $gamecontrollerlist$02 = 0, $gamecontrollerlist$03 = 0, $gamecontrollerlist$03$lcssa = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, dest = 0;
 var label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($device_index|0)<(0);
 if (!($0)) {
  $1 = (_SDL_NumJoysticks()|0);
  $2 = ($1|0)>($device_index|0);
  if ($2) {
   $gamecontrollerlist$02 = HEAP32[301459]|0;
   $3 = ($gamecontrollerlist$02|0)==(0|0);
   L4: do {
    if (!($3)) {
     $gamecontrollerlist$03 = $gamecontrollerlist$02;
     while(1) {
      $5 = (_SDL_SYS_GetInstanceIdOfDeviceIndex($device_index)|0);
      $6 = HEAP32[$gamecontrollerlist$03>>2]|0;
      $7 = HEAP32[$6>>2]|0;
      $8 = ($5|0)==($7|0);
      if ($8) {
       $gamecontrollerlist$03$lcssa = $gamecontrollerlist$03;
       break;
      }
      $12 = ((($gamecontrollerlist$03)) + 896|0);
      $gamecontrollerlist$0 = HEAP32[$12>>2]|0;
      $13 = ($gamecontrollerlist$0|0)==(0|0);
      if ($13) {
       break L4;
      } else {
       $gamecontrollerlist$03 = $gamecontrollerlist$0;
      }
     }
     $9 = ((($gamecontrollerlist$03$lcssa)) + 4|0);
     $10 = HEAP32[$9>>2]|0;
     $11 = (($10) + 1)|0;
     HEAP32[$9>>2] = $11;
     $$0 = $gamecontrollerlist$03$lcssa;
     STACKTOP = sp;return ($$0|0);
    }
   } while(0);
   $14 = (_SDL_PrivateGetControllerMapping($device_index)|0);
   $15 = ($14|0)==(0|0);
   if ($15) {
    HEAP32[$vararg_buffer1>>2] = $device_index;
    (_SDL_SetError(611998,$vararg_buffer1)|0);
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   $16 = (_SDL_malloc(900)|0);
   $17 = ($16|0)==(0|0);
   if ($17) {
    (_SDL_Error(0)|0);
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   (_SDL_memset($16,0,900)|0);
   $18 = (_SDL_JoystickOpen($device_index)|0);
   HEAP32[$16>>2] = $18;
   $19 = ($18|0)==(0|0);
   if ($19) {
    _SDL_free($16);
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   } else {
    $20 = ((($16)) + 12|0);
    $21 = ((($14)) + 16|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = ((($14)) + 20|0);
    $24 = HEAP32[$23>>2]|0;
    dest=$$byval_copy; src=$14; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
    _SDL_PrivateLoadButtonMapping($20,$$byval_copy,$22,$24);
    $25 = ((($16)) + 4|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = (($26) + 1)|0;
    HEAP32[$25>>2] = $27;
    $28 = HEAP32[301459]|0;
    $29 = ((($16)) + 896|0);
    HEAP32[$29>>2] = $28;
    HEAP32[301459] = $16;
    $30 = HEAP32[$16>>2]|0;
    _SDL_SYS_JoystickUpdate($30);
    $$0 = $16;
    STACKTOP = sp;return ($$0|0);
   }
  }
 }
 $4 = (_SDL_NumJoysticks()|0);
 HEAP32[$vararg_buffer>>2] = $4;
 (_SDL_SetError(611965,$vararg_buffer)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SYS_GetInstanceIdOfDeviceIndex($device_index) {
 $device_index = $device_index|0;
 var $$02$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $item$0$i = 0, $item$0$lcssa$i = 0, $item$01$i = 0, $item$03$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $item$01$i = HEAP32[301464]|0;
 $0 = ($device_index|0)>(0);
 if ($0) {
  $$02$i = $device_index;$item$03$i = $item$01$i;
  while(1) {
   $1 = (($$02$i) + -1)|0;
   $2 = ((($item$03$i)) + 1320|0);
   $item$0$i = HEAP32[$2>>2]|0;
   $3 = ($$02$i|0)>(1);
   if ($3) {
    $$02$i = $1;$item$03$i = $item$0$i;
   } else {
    $item$0$lcssa$i = $item$0$i;
    break;
   }
  }
 } else {
  $item$0$lcssa$i = $item$01$i;
 }
 $4 = ((($item$0$lcssa$i)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 return ($5|0);
}
function _SDL_JoystickOpen($device_index) {
 $device_index = $device_index|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $joysticklist$0 = 0, $joysticklist$02 = 0, $joysticklist$03 = 0, $joysticklist$03$lcssa = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($device_index|0)<(0);
 if (!($0)) {
  $1 = (_SDL_SYS_NumJoysticks()|0);
  $2 = ($1|0)>($device_index|0);
  if ($2) {
   $joysticklist$02 = HEAP32[301462]|0;
   $3 = ($joysticklist$02|0)==(0|0);
   L4: do {
    if (!($3)) {
     $joysticklist$03 = $joysticklist$02;
     while(1) {
      $5 = (_SDL_SYS_GetInstanceIdOfDeviceIndex($device_index)|0);
      $6 = HEAP32[$joysticklist$03>>2]|0;
      $7 = ($5|0)==($6|0);
      if ($7) {
       $joysticklist$03$lcssa = $joysticklist$03;
       break;
      }
      $11 = ((($joysticklist$03)) + 56|0);
      $joysticklist$0 = HEAP32[$11>>2]|0;
      $12 = ($joysticklist$0|0)==(0|0);
      if ($12) {
       break L4;
      } else {
       $joysticklist$03 = $joysticklist$0;
      }
     }
     $8 = ((($joysticklist$03$lcssa)) + 44|0);
     $9 = HEAP32[$8>>2]|0;
     $10 = (($9) + 1)|0;
     HEAP32[$8>>2] = $10;
     $$0 = $joysticklist$03$lcssa;
     STACKTOP = sp;return ($$0|0);
    }
   } while(0);
   $13 = (_SDL_malloc(60)|0);
   $14 = ($13|0)==(0|0);
   if ($14) {
    (_SDL_Error(0)|0);
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   (_SDL_memset($13,0,60)|0);
   $15 = (_SDL_SYS_JoystickOpen($13,$device_index)|0);
   $16 = ($15|0)<(0);
   if ($16) {
    _SDL_free($13);
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   $17 = (_SDL_SYS_JoystickNameForDeviceIndex($device_index)|0);
   $18 = ($17|0)==(0|0);
   if ($18) {
    $21 = ((($13)) + 4|0);
    HEAP32[$21>>2] = 0;
   } else {
    $19 = (_SDL_strdup($17)|0);
    $20 = ((($13)) + 4|0);
    HEAP32[$20>>2] = $19;
   }
   $22 = ((($13)) + 8|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = ($23|0)>(0);
   if ($24) {
    $25 = $23 << 1;
    $26 = (_SDL_malloc($25)|0);
    $27 = ((($13)) + 12|0);
    HEAP32[$27>>2] = $26;
   }
   $28 = ((($13)) + 16|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = ($29|0)>(0);
   if ($30) {
    $31 = (_SDL_malloc($29)|0);
    $32 = ((($13)) + 20|0);
    HEAP32[$32>>2] = $31;
   }
   $33 = ((($13)) + 24|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = ($34|0)>(0);
   if ($35) {
    $36 = $34 << 3;
    $37 = (_SDL_malloc($36)|0);
    $38 = ((($13)) + 28|0);
    HEAP32[$38>>2] = $37;
   }
   $39 = ((($13)) + 32|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = ($40|0)>(0);
   if ($41) {
    $42 = (_SDL_malloc($40)|0);
    $43 = ((($13)) + 36|0);
    HEAP32[$43>>2] = $42;
   }
   $44 = HEAP32[$22>>2]|0;
   $45 = ($44|0)>(0);
   if ($45) {
    $46 = ((($13)) + 12|0);
    $47 = HEAP32[$46>>2]|0;
    $48 = ($47|0)==(0|0);
    if (!($48)) {
     label = 25;
    }
   } else {
    label = 25;
   }
   do {
    if ((label|0) == 25) {
     $49 = HEAP32[$28>>2]|0;
     $50 = ($49|0)>(0);
     if ($50) {
      $51 = ((($13)) + 20|0);
      $52 = HEAP32[$51>>2]|0;
      $53 = ($52|0)==(0|0);
      if ($53) {
       break;
      }
     }
     $54 = HEAP32[$33>>2]|0;
     $55 = ($54|0)>(0);
     if ($55) {
      $56 = ((($13)) + 28|0);
      $57 = HEAP32[$56>>2]|0;
      $58 = ($57|0)==(0|0);
      if ($58) {
       break;
      }
     }
     $59 = HEAP32[$39>>2]|0;
     $60 = ($59|0)>(0);
     if ($60) {
      $61 = ((($13)) + 36|0);
      $62 = HEAP32[$61>>2]|0;
      $63 = ($62|0)==(0|0);
      if ($63) {
       break;
      }
     }
     $64 = ((($13)) + 12|0);
     $65 = HEAP32[$64>>2]|0;
     $66 = ($65|0)==(0|0);
     if (!($66)) {
      $67 = $44 << 1;
      (_SDL_memset($65,0,$67)|0);
     }
     $68 = ((($13)) + 20|0);
     $69 = HEAP32[$68>>2]|0;
     $70 = ($69|0)==(0|0);
     if (!($70)) {
      $71 = HEAP32[$28>>2]|0;
      (_SDL_memset($69,0,$71)|0);
     }
     $72 = ((($13)) + 28|0);
     $73 = HEAP32[$72>>2]|0;
     $74 = ($73|0)==(0|0);
     if (!($74)) {
      $75 = HEAP32[$33>>2]|0;
      $76 = $75 << 3;
      (_SDL_memset($73,0,$76)|0);
     }
     $77 = ((($13)) + 36|0);
     $78 = HEAP32[$77>>2]|0;
     $79 = ($78|0)==(0|0);
     if (!($79)) {
      $80 = HEAP32[$39>>2]|0;
      (_SDL_memset($78,0,$80)|0);
     }
     $81 = ((($13)) + 52|0);
     HEAP32[$81>>2] = -1;
     $82 = ((($13)) + 44|0);
     $83 = HEAP32[$82>>2]|0;
     $84 = (($83) + 1)|0;
     HEAP32[$82>>2] = $84;
     $85 = HEAP32[301462]|0;
     $86 = ((($13)) + 56|0);
     HEAP32[$86>>2] = $85;
     HEAP32[301462] = $13;
     _SDL_SYS_JoystickUpdate($13);
     $$0 = $13;
     STACKTOP = sp;return ($$0|0);
    }
   } while(0);
   (_SDL_Error(0)|0);
   _SDL_JoystickClose($13);
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $4 = (_SDL_SYS_NumJoysticks()|0);
 HEAP32[$vararg_buffer>>2] = $4;
 (_SDL_SetError(611965,$vararg_buffer)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SYS_JoystickOpen($joystick,$device_index) {
 $joystick = $joystick|0;
 $device_index = $device_index|0;
 var $$0 = 0, $$02$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $item$0$i = 0, $item$0$lcssa$i = 0, $item$01$i = 0, $item$03$i = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $item$01$i = HEAP32[301464]|0;
 $0 = ($device_index|0)>(0);
 if ($0) {
  $$02$i = $device_index;$item$03$i = $item$01$i;
  while(1) {
   $1 = (($$02$i) + -1)|0;
   $2 = ((($item$03$i)) + 1320|0);
   $item$0$i = HEAP32[$2>>2]|0;
   $3 = ($$02$i|0)>(1);
   if ($3) {
    $$02$i = $1;$item$03$i = $item$0$i;
   } else {
    $item$0$lcssa$i = $item$0$i;
    break;
   }
  }
 } else {
  $item$0$lcssa$i = $item$01$i;
 }
 $4 = ($item$0$lcssa$i|0)==(0|0);
 if ($4) {
  $5 = (_SDL_SetError(612036,$vararg_buffer)|0);
  $$0 = $5;
  STACKTOP = sp;return ($$0|0);
 }
 $6 = ((($item$0$lcssa$i)) + 16|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 if ($8) {
  $10 = ((($item$0$lcssa$i)) + 12|0);
  $11 = HEAP32[$10>>2]|0;
  HEAP32[$joystick>>2] = $11;
  $12 = ((($joystick)) + 40|0);
  HEAP32[$12>>2] = $item$0$lcssa$i;
  HEAP32[$6>>2] = $joystick;
  $13 = ((($joystick)) + 16|0);
  HEAP32[$13>>2] = 0;
  $14 = ((($joystick)) + 24|0);
  HEAP32[$14>>2] = 0;
  $15 = ((($item$0$lcssa$i)) + 20|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ((($joystick)) + 32|0);
  HEAP32[$17>>2] = $16;
  $18 = ((($item$0$lcssa$i)) + 24|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = ((($joystick)) + 8|0);
  HEAP32[$20>>2] = $19;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $9 = (_SDL_SetError(612051,$vararg_buffer1)|0);
  $$0 = $9;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_GameControllerGetAxis($gamecontroller,$axis) {
 $gamecontroller = $gamecontroller|0;
 $axis = $axis|0;
 var $$ = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $switch = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($gamecontroller|0)==(0|0);
 if ($0) {
  $$1 = 0;
  return ($$1|0);
 }
 $1 = (((($gamecontroller)) + 32|0) + ($axis<<2)|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)>(-1);
 if ($3) {
  $4 = HEAP32[$gamecontroller>>2]|0;
  $5 = (_SDL_JoystickGetAxis($4,$2)|0);
  $6 = $axis & -2;
  $switch = ($6|0)==(4);
  if (!($switch)) {
   $$1 = $5;
   return ($$1|0);
  }
  $7 = $5 << 16 >> 16;
  $8 = (($7|0) / 2)&-1;
  $9 = (($8) + 16384)|0;
  $10 = $9&65535;
  $$1 = $10;
  return ($$1|0);
 } else {
  $11 = (((($gamecontroller)) + 56|0) + ($axis<<2)|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ($12|0)>(-1);
  if (!($13)) {
   $$1 = 0;
   return ($$1|0);
  }
  $14 = HEAP32[$gamecontroller>>2]|0;
  $15 = (_SDL_JoystickGetButton($14,$12)|0);
  $16 = ($15<<24>>24)==(0);
  $$ = $16 ? 0 : 32767;
  $$1 = $$;
  return ($$1|0);
 }
 return (0)|0;
}
function _SDL_JoystickGetAxis($joystick,$axis) {
 $joystick = $joystick|0;
 $axis = $axis|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($joystick|0)==(0|0);
 if ($0) {
  (_SDL_SetError(611627,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ((($joystick)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)>($axis|0);
 if ($3) {
  $4 = ((($joystick)) + 12|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = (($5) + ($axis<<1)|0);
  $7 = HEAP16[$6>>1]|0;
  $$0 = $7;
  STACKTOP = sp;return ($$0|0);
 } else {
  HEAP32[$vararg_buffer1>>2] = $2;
  (_SDL_SetError(612075,$vararg_buffer1)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_JoystickGetButton($joystick,$button) {
 $joystick = $joystick|0;
 $button = $button|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($joystick|0)==(0|0);
 if ($0) {
  (_SDL_SetError(611627,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ((($joystick)) + 32|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)>($button|0);
 if ($3) {
  $4 = ((($joystick)) + 36|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = (($5) + ($button)|0);
  $7 = HEAP8[$6>>0]|0;
  $$0 = $7;
  STACKTOP = sp;return ($$0|0);
 } else {
  HEAP32[$vararg_buffer1>>2] = $2;
  (_SDL_SetError(612101,$vararg_buffer1)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_GameControllerGetButton($gamecontroller,$button) {
 $gamecontroller = $gamecontroller|0;
 $button = $button|0;
 var $$ = 0, $$2 = 0, $$3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($gamecontroller|0)==(0|0);
 if ($0) {
  $$2 = 0;
  return ($$2|0);
 }
 $1 = (((($gamecontroller)) + 80|0) + ($button<<2)|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)>(-1);
 if ($3) {
  $4 = HEAP32[$gamecontroller>>2]|0;
  $5 = (_SDL_JoystickGetButton($4,$2)|0);
  $$2 = $5;
  return ($$2|0);
 }
 $6 = (((($gamecontroller)) + 140|0) + ($button<<2)|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)>(-1);
 if ($8) {
  $9 = HEAP32[$gamecontroller>>2]|0;
  $10 = (_SDL_JoystickGetAxis($9,$7)|0);
  $11 = $10 << 16 >> 16;
  $12 = ($10<<16>>16)<(0);
  $13 = (0 - ($11))|0;
  $14 = $12 ? $13 : $11;
  $15 = ($14|0)>(16384);
  $$ = $15&1;
  $$2 = $$;
  return ($$2|0);
 }
 $16 = (((($gamecontroller)) + 200|0) + ($button<<3)|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ($17|0)>(-1);
 if (!($18)) {
  $$2 = 0;
  return ($$2|0);
 }
 $19 = HEAP32[$gamecontroller>>2]|0;
 $20 = (_SDL_JoystickGetHat($19,$17)|0);
 $21 = (((((($gamecontroller)) + 200|0) + ($button<<3)|0)) + 4|0);
 $22 = HEAP8[$21>>0]|0;
 $23 = $22 & $20;
 $not$ = ($23<<24>>24)!=(0);
 $$3 = $not$&1;
 $$2 = $$3;
 return ($$2|0);
}
function _SDL_JoystickGetHat($joystick,$hat) {
 $joystick = $joystick|0;
 $hat = $hat|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = ($joystick|0)==(0|0);
 if ($0) {
  (_SDL_SetError(611627,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ((($joystick)) + 16|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)>($hat|0);
 if ($3) {
  $4 = ((($joystick)) + 20|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = (($5) + ($hat)|0);
  $7 = HEAP8[$6>>0]|0;
  $$0 = $7;
  STACKTOP = sp;return ($$0|0);
 } else {
  HEAP32[$vararg_buffer1>>2] = $2;
  (_SDL_SetError(612130,$vararg_buffer1)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_JoystickGetAttached($joystick) {
 $joystick = $joystick|0;
 var $$0 = 0, $0 = 0, $1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($joystick|0)==(0|0);
 if ($0) {
  (_SDL_SetError(611627,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $1 = (_SDL_SYS_JoystickAttached($joystick)|0);
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_SYS_JoystickAttached($joystick) {
 $joystick = $joystick|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($joystick)) + 40|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)!=(0|0);
 $3 = $2&1;
 return ($3|0);
}
function _SDL_GameControllerName($gamecontroller) {
 $gamecontroller = $gamecontroller|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($gamecontroller|0)==(0|0);
 if ($0) {
  $$0 = 0;
  return ($$0|0);
 }
 $1 = ((($gamecontroller)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $$0 = $2;
 return ($$0|0);
}
function _SDL_GameControllerGetJoystick($gamecontroller) {
 $gamecontroller = $gamecontroller|0;
 var $$0 = 0, $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($gamecontroller|0)==(0|0);
 if ($0) {
  $$0 = 0;
  return ($$0|0);
 }
 $1 = HEAP32[$gamecontroller>>2]|0;
 $$0 = $1;
 return ($$0|0);
}
function _SDL_GameControllerClose($gamecontroller) {
 $gamecontroller = $gamecontroller|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $gamecontrollerlist$0 = 0, $gamecontrollerlist$03 = 0, $gamecontrollerlist$05 = 0, $gamecontrollerlistprev$04 = 0, $gamecontrollerlistprev$04$lcssa = 0, $gamecontrollerlistprev$04$phi = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($gamecontroller|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($gamecontroller)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($2) + -1)|0;
 HEAP32[$1>>2] = $3;
 $4 = ($2|0)>(1);
 if ($4) {
  return;
 }
 $5 = HEAP32[$gamecontroller>>2]|0;
 _SDL_JoystickClose($5);
 $gamecontrollerlist$03 = HEAP32[301459]|0;
 $6 = ($gamecontrollerlist$03|0)==(0|0);
 L7: do {
  if (!($6)) {
   $gamecontrollerlist$05 = $gamecontrollerlist$03;$gamecontrollerlistprev$04 = 0;
   while(1) {
    $7 = ($gamecontrollerlist$05|0)==($gamecontroller|0);
    if ($7) {
     $gamecontrollerlistprev$04$lcssa = $gamecontrollerlistprev$04;
     break;
    }
    $12 = ((($gamecontrollerlist$05)) + 896|0);
    $gamecontrollerlist$0 = HEAP32[$12>>2]|0;
    $13 = ($gamecontrollerlist$0|0)==(0|0);
    if ($13) {
     break L7;
    } else {
     $gamecontrollerlistprev$04$phi = $gamecontrollerlist$05;$gamecontrollerlist$05 = $gamecontrollerlist$0;$gamecontrollerlistprev$04 = $gamecontrollerlistprev$04$phi;
    }
   }
   $8 = ($gamecontrollerlistprev$04$lcssa|0)==(0|0);
   $9 = ((($gamecontroller)) + 896|0);
   $10 = HEAP32[$9>>2]|0;
   if ($8) {
    HEAP32[301459] = $10;
    break;
   } else {
    $11 = ((($gamecontrollerlistprev$04$lcssa)) + 896|0);
    HEAP32[$11>>2] = $10;
    break;
   }
  }
 } while(0);
 _SDL_free($gamecontroller);
 return;
}
function _SDL_GameControllerEventState($state) {
 $state = $state|0;
 var $$1 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $cond = 0, $i$03 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $cond = ($state|0)==(-1);
 if ($cond) {
  $i$03 = 0;
 } else {
  (_SDL_EventState(1616,$state)|0);
  (_SDL_EventState(1617,$state)|0);
  (_SDL_EventState(1618,$state)|0);
  (_SDL_EventState(1619,$state)|0);
  (_SDL_EventState(1620,$state)|0);
  (_SDL_EventState(1621,$state)|0);
  $$1 = $state;
  return ($$1|0);
 }
 while(1) {
  $0 = (141104 + ($i$03<<2)|0);
  $1 = HEAP32[$0>>2]|0;
  $2 = (_SDL_EventState($1,-1)|0);
  $3 = ($2<<24>>24)!=(1);
  $4 = (($i$03) + 1)|0;
  $5 = ($4>>>0)<(6);
  $or$cond = $3 & $5;
  if ($or$cond) {
   $i$03 = $4;
  } else {
   $$lcssa = $2;
   break;
  }
 }
 $6 = $$lcssa&255;
 $$1 = $6;
 return ($$1|0);
}
function _SDL_JoystickNumAxes($joystick) {
 $joystick = $joystick|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($joystick|0)==(0|0);
 if ($0) {
  (_SDL_SetError(611627,$vararg_buffer)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 } else {
  $1 = ((($joystick)) + 8|0);
  $2 = HEAP32[$1>>2]|0;
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_JoystickNumHats($joystick) {
 $joystick = $joystick|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($joystick|0)==(0|0);
 if ($0) {
  (_SDL_SetError(611627,$vararg_buffer)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 } else {
  $1 = ((($joystick)) + 16|0);
  $2 = HEAP32[$1>>2]|0;
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_JoystickNumButtons($joystick) {
 $joystick = $joystick|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($joystick|0)==(0|0);
 if ($0) {
  (_SDL_SetError(611627,$vararg_buffer)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 } else {
  $1 = ((($joystick)) + 32|0);
  $2 = HEAP32[$1>>2]|0;
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_JoystickInstanceID($joystick) {
 $joystick = $joystick|0;
 var $$0 = 0, $0 = 0, $1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($joystick|0)==(0|0);
 if ($0) {
  (_SDL_SetError(611627,$vararg_buffer)|0);
  $$0 = -1;
 } else {
  $1 = HEAP32[$joystick>>2]|0;
  $$0 = $1;
 }
 STACKTOP = sp;return ($$0|0);
}
function _SDL_JoystickName($joystick) {
 $joystick = $joystick|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($joystick|0)==(0|0);
 if ($0) {
  (_SDL_SetError(611627,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $1 = ((($joystick)) + 4|0);
  $2 = HEAP32[$1>>2]|0;
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_JoystickGetGUID($agg$result,$joystick) {
 $agg$result = $agg$result|0;
 $joystick = $joystick|0;
 var $0 = 0, $emptyGUID1 = 0, $vararg_buffer = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $emptyGUID1 = sp + 8|0;
 $0 = ($joystick|0)==(0|0);
 if ($0) {
  (_SDL_SetError(611627,$vararg_buffer)|0);
  (_SDL_memset($emptyGUID1,0,16)|0);
  dest=$agg$result; src=$emptyGUID1; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
  STACKTOP = sp;return;
 } else {
  _SDL_SYS_JoystickGetGUID($agg$result,$joystick);
  STACKTOP = sp;return;
 }
}
function _SDL_SYS_JoystickGetGUID($agg$result,$joystick) {
 $agg$result = $agg$result|0;
 $joystick = $joystick|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $guid1 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $guid1 = sp;
 $0 = ((($joystick)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 (_SDL_memset($guid1,0,16)|0);
 $2 = (_SDL_strlen($1)|0);
 $3 = ($2>>>0)>(16);
 if ($3) {
  $5 = 16;
 } else {
  $4 = (_SDL_strlen($1)|0);
  $5 = $4;
 }
 (_SDL_memcpy($guid1,$1,$5)|0);
 dest=$agg$result; src=$guid1; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
 STACKTOP = sp;return;
}
function _SDL_GetPowerInfo($seconds,$percent) {
 $seconds = $seconds|0;
 $percent = $percent|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $_percent = 0, $_seconds = 0, $_seconds$seconds = 0, $retval = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $_seconds = sp + 8|0;
 $_percent = sp + 4|0;
 $retval = sp;
 HEAP32[$retval>>2] = 0;
 $0 = ($seconds|0)==(0|0);
 $_seconds$seconds = $0 ? $_seconds : $seconds;
 $1 = ($percent|0)==(0|0);
 $$01 = $1 ? $_percent : $percent;
 $2 = (_SDL_GetPowerInfo_Emscripten($retval,$_seconds$seconds,$$01)|0);
 $3 = ($2|0)==(0);
 if ($3) {
  HEAP32[$_seconds$seconds>>2] = -1;
  HEAP32[$$01>>2] = -1;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $4 = HEAP32[$retval>>2]|0;
  $$0 = $4;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_GetPowerInfo_Emscripten($state,$seconds,$percent) {
 $state = $state|0;
 $seconds = $seconds|0;
 $percent = $percent|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0, $2 = 0, $3 = 0.0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0.0, $9 = 0, $batteryState = 0;
 var $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $batteryState = sp;
 $0 = (_emscripten_get_battery_status(($batteryState|0))|0);
 $1 = ($0|0)==(-1);
 if ($1) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ((($batteryState)) + 16|0);
 $3 = +HEAPF64[$2>>3];
 $4 = $3 != 1.0;
 $$phi$trans$insert = ((($batteryState)) + 24|0);
 $$pre = HEAP32[$$phi$trans$insert>>2]|0;
 $5 = ($$pre|0)==(0);
 if ($4) {
  if ($5) {
   $storemerge = 1;
  } else {
   label = 7;
  }
 } else {
  if ($5) {
   $storemerge = 1;
  } else {
   $6 = +HEAPF64[$batteryState>>3];
   $7 = $6 != 0.0;
   if ($7) {
    label = 7;
   } else {
    HEAP32[$state>>2] = 2;
    HEAP32[$seconds>>2] = -1;
    HEAP32[$percent>>2] = -1;
    $$0 = 1;
    STACKTOP = sp;return ($$0|0);
   }
  }
 }
 if ((label|0) == 7) {
  $8 = +HEAPF64[$batteryState>>3];
  $9 = $8 == 0.0;
  $10 = $9 ? 4 : 3;
  $storemerge = $10;
 }
 HEAP32[$state>>2] = $storemerge;
 $11 = ((($batteryState)) + 8|0);
 $12 = +HEAPF64[$11>>3];
 $13 = (~~(($12)));
 HEAP32[$seconds>>2] = $13;
 $14 = $3 * 100.0;
 $15 = (~~(($14)));
 HEAP32[$percent>>2] = $15;
 $$0 = 1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_CreateWindow($title,$x,$y,$w,$h,$flags) {
 $title = $title|0;
 $x = $x|0;
 $y = $y|0;
 $w = $w|0;
 $h = $h|0;
 $flags = $flags|0;
 var $$0 = 0, $$0$i = 0, $$0$i10 = 0, $$0$i11 = 0, $$03 = 0, $$04 = 0, $$flags = 0, $$phi$trans$insert = 0, $$phi$trans$insert$i = 0, $$pre = 0, $$pre$i = 0, $$pre13 = 0, $$pre14 = 0, $$w = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bounds = 0, $displayIndex$01$i = 0, $or$cond = 0, $switch$split112D = 0, $switch$split142D = 0, $switch$split172D = 0;
 var $switch$split2D = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, $vararg_buffer7 = 0, $vararg_buffer9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer9 = sp + 40|0;
 $vararg_buffer7 = sp + 32|0;
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $bounds = sp + 48|0;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  $2 = (_SDL_VideoInit(0)|0);
  $3 = ($2|0)<(0);
  if ($3) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $4 = ($w|0)<(1);
 $$w = $4 ? 1 : $w;
 $5 = ($h|0)<(1);
 $$03 = $5 ? 1 : $h;
 $6 = ($$w|0)>(16384);
 $7 = ($$03|0)>(16384);
 $or$cond = $6 | $7;
 if ($or$cond) {
  (_SDL_SetError(612156,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $8 = $flags & 2;
 $9 = ($8|0)==(0);
 L9: do {
  if (!($9)) {
   $10 = HEAP32[300441]|0;
   $11 = ((($10)) + 152|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($12|0)==(0|0);
   if ($13) {
    (_SDL_SetError(596263,$vararg_buffer1)|0);
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   $14 = ($10|0)==(0|0);
   L15: do {
    if ($14) {
     $15 = (_SDL_SetError(595917,$vararg_buffer3)|0);
     $$0$i = $15;
    } else {
     $16 = ((($10)) + 364|0);
     $17 = HEAP32[$16>>2]|0;
     $18 = ($17|0)==(0);
     do {
      if ($18) {
       $19 = ((($10)) + 140|0);
       $20 = HEAP32[$19>>2]|0;
       $21 = ($20|0)==(0|0);
       if ($21) {
        $22 = (_SDL_SetError(595987,$vararg_buffer5)|0);
        $$0$i = $22;
        break L15;
       }
       $23 = (FUNCTION_TABLE_iii[$20 & 8191]($10,0)|0);
       $24 = ($23|0)==(0);
       $25 = HEAP32[300441]|0;
       if ($24) {
        $$phi$trans$insert = ((($25)) + 364|0);
        $$pre = HEAP32[$$phi$trans$insert>>2]|0;
        $27 = $25;$29 = $$pre;
        break;
       }
       $30 = ((($25)) + 148|0);
       $31 = HEAP32[$30>>2]|0;
       $32 = ($31|0)==(0|0);
       if ($32) {
        $$0$i = $23;
        break L15;
       }
       FUNCTION_TABLE_vi[$31 & 8191]($25);
       $$0$i = $23;
       break L15;
      } else {
       $27 = $10;$29 = $17;
      }
     } while(0);
     $26 = ((($27)) + 364|0);
     $28 = (($29) + 1)|0;
     HEAP32[$26>>2] = $28;
     break L9;
    }
   } while(0);
   $33 = ($$0$i|0)<(0);
   if ($33) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
  }
 } while(0);
 $34 = $flags & 8192;
 $35 = ($34|0)==(0);
 if ($35) {
  $$04 = $flags;
 } else {
  $36 = (_SDL_GetHint(612177)|0);
  $37 = ($36|0)==(0|0);
  if ($37) {
   $$04 = $flags;
  } else {
   $38 = (_SDL_atoi($36)|0);
   $39 = ($38|0)>(0);
   $40 = $flags & -8193;
   $$flags = $39 ? $40 : $flags;
   $$04 = $$flags;
  }
 }
 $41 = (_SDL_calloc(1,148)|0);
 $42 = ($41|0)==(0|0);
 if ($42) {
  (_SDL_Error(0)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $43 = HEAP32[300441]|0;
 $44 = ((($43)) + 256|0);
 HEAP32[$41>>2] = $44;
 $45 = ((($43)) + 260|0);
 $46 = HEAP32[$45>>2]|0;
 $47 = (($46) + 1)|0;
 HEAP32[$45>>2] = $47;
 $48 = ((($41)) + 4|0);
 HEAP32[$48>>2] = $46;
 $49 = ((($41)) + 16|0);
 HEAP32[$49>>2] = $x;
 $50 = ((($41)) + 20|0);
 HEAP32[$50>>2] = $y;
 $51 = ((($41)) + 24|0);
 HEAP32[$51>>2] = $$w;
 $52 = ((($41)) + 28|0);
 HEAP32[$52>>2] = $$03;
 $53 = $x & -65536;
 $switch$split2D = ($53|0)<(805240832);
 if ($switch$split2D) {
  switch ($53|0) {
  case 536805376:  {
   label = 26;
   break;
  }
  default: {
   label = 25;
  }
  }
 } else {
  switch ($53|0) {
  case 805240832:  {
   label = 26;
   break;
  }
  default: {
   label = 25;
  }
  }
 }
 L42: do {
  if ((label|0) == 25) {
   $54 = $y & -65536;
   $switch$split112D = ($54|0)<(805240832);
   if ($switch$split112D) {
    switch ($54|0) {
    case 536805376:  {
     label = 26;
     break L42;
     break;
    }
    default: {
     $88 = $43;
     break L42;
    }
    }
   } else {
    switch ($54|0) {
    case 805240832:  {
     label = 26;
     break L42;
     break;
    }
    default: {
     $88 = $43;
     break L42;
    }
    }
   }
  }
 } while(0);
 if ((label|0) == 26) {
  $55 = (_SDL_GetWindowDisplayIndex($41)|0);
  $56 = ($55|0)>(-1);
  $57 = HEAP32[300441]|0;
  if ($56) {
   $58 = ((($57)) + 244|0);
   $59 = HEAP32[$58>>2]|0;
   $60 = (($59) + (($55*68)|0)|0);
   $$0$i10 = $60;
  } else {
   $$0$i10 = 0;
  }
  $61 = ((($57)) + 240|0);
  $62 = HEAP32[$61>>2]|0;
  $63 = ($62|0)>(0);
  L52: do {
   if ($63) {
    $$phi$trans$insert$i = ((($57)) + 244|0);
    $$pre$i = HEAP32[$$phi$trans$insert$i>>2]|0;
    $displayIndex$01$i = 0;
    while(1) {
     $64 = (($$pre$i) + (($displayIndex$01$i*68)|0)|0);
     $65 = ($64|0)==($$0$i10|0);
     if ($65) {
      $$0$i11 = $displayIndex$01$i;
      break L52;
     }
     $66 = (($displayIndex$01$i) + 1)|0;
     $67 = ($66|0)<($62|0);
     if ($67) {
      $displayIndex$01$i = $66;
     } else {
      $$0$i11 = 0;
      break;
     }
    }
   } else {
    $$0$i11 = 0;
   }
  } while(0);
  (_SDL_GetDisplayBounds($$0$i11,$bounds)|0);
  $switch$split142D = ($53|0)<(805240832);
  if ($switch$split142D) {
   switch ($53|0) {
   case 536805376:  {
    label = 33;
    break;
   }
   default: {
   }
   }
  } else {
   switch ($53|0) {
   case 805240832:  {
    label = 33;
    break;
   }
   default: {
   }
   }
  }
  if ((label|0) == 33) {
   $68 = HEAP32[$bounds>>2]|0;
   $69 = ((($bounds)) + 8|0);
   $70 = HEAP32[$69>>2]|0;
   $71 = (($70) - ($$w))|0;
   $72 = (($71|0) / 2)&-1;
   $73 = (($72) + ($68))|0;
   HEAP32[$49>>2] = $73;
  }
  $74 = $y & -65536;
  $switch$split172D = ($74|0)<(805240832);
  if ($switch$split172D) {
   switch ($74|0) {
   case 536805376:  {
    label = 35;
    break;
   }
   default: {
   }
   }
  } else {
   switch ($74|0) {
   case 805240832:  {
    label = 35;
    break;
   }
   default: {
   }
   }
  }
  if ((label|0) == 35) {
   $75 = ((($bounds)) + 4|0);
   $76 = HEAP32[$75>>2]|0;
   $77 = ((($bounds)) + 12|0);
   $78 = HEAP32[$77>>2]|0;
   $79 = (($78) - ($$03))|0;
   $80 = (($79|0) / 2)&-1;
   $81 = (($80) + ($76))|0;
   HEAP32[$50>>2] = $81;
  }
  $$pre13 = HEAP32[300441]|0;
  $88 = $$pre13;
 }
 $82 = $$04 & 8242;
 $83 = $82 | 8;
 $84 = ((($41)) + 48|0);
 HEAP32[$84>>2] = $83;
 $85 = ((($41)) + 52|0);
 HEAP32[$85>>2] = $83;
 $86 = ((($41)) + 92|0);
 HEAPF32[$86>>2] = 1.0;
 $87 = ((($88)) + 248|0);
 $89 = HEAP32[$87>>2]|0;
 $90 = ((($41)) + 144|0);
 HEAP32[$90>>2] = $89;
 $91 = ((($41)) + 116|0);
 HEAP32[$91>>2] = 0;
 $92 = ($89|0)==(0);
 if ($92) {
  $96 = $88;
 } else {
  $93 = $89;
  $94 = ((($93)) + 140|0);
  HEAP32[$94>>2] = $41;
  $$pre14 = HEAP32[300441]|0;
  $96 = $$pre14;
 }
 $95 = ((($96)) + 248|0);
 HEAP32[$95>>2] = $41;
 $97 = ((($96)) + 28|0);
 $98 = HEAP32[$97>>2]|0;
 $99 = ($98|0)==(0|0);
 if (!($99)) {
  $100 = (FUNCTION_TABLE_iii[$98 & 8191]($96,$41)|0);
  $101 = ($100|0)<(0);
  if ($101) {
   _SDL_DestroyWindow($41);
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $102 = ($title|0)==(0|0);
 do {
  if (!($102)) {
   $103 = HEAP32[300441]|0;
   $104 = ($103|0)==(0|0);
   if ($104) {
    (_SDL_SetError(595917,$vararg_buffer7)|0);
    break;
   }
   $105 = HEAP32[$41>>2]|0;
   $106 = ((($103)) + 256|0);
   $107 = ($105|0)==($106|0);
   if (!($107)) {
    (_SDL_SetError(609414,$vararg_buffer9)|0);
    break;
   }
   $108 = ((($41)) + 8|0);
   $109 = HEAP32[$108>>2]|0;
   $110 = ($109|0)==($title|0);
   if (!($110)) {
    _SDL_free($109);
    $111 = (_SDL_strdup($title)|0);
    HEAP32[$108>>2] = $111;
    $112 = HEAP32[300441]|0;
    $113 = ((($112)) + 36|0);
    $114 = HEAP32[$113>>2]|0;
    $115 = ($114|0)==(0|0);
    if (!($115)) {
     FUNCTION_TABLE_vii[$114 & 8191]($112,$41);
    }
   }
  }
 } while(0);
 _SDL_FinishWindowCreation($41,$$04);
 $116 = HEAP32[$84>>2]|0;
 $117 = $116 & 69;
 $118 = ($117|0)==(5);
 $119 = $118&1;
 (_SDL_UpdateFullscreenMode($41,$119)|0);
 $$0 = $41;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SetClipboardText($text) {
 $text = $text|0;
 var $$0 = 0, $$text = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_SDL_GetVideoDevice()|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  $2 = (_SDL_SetError(612204,$vararg_buffer)|0);
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ($text|0)==(0|0);
 $$text = $3 ? 1282466 : $text;
 $4 = ((($0)) + 216|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if ($6) {
  $8 = ((($0)) + 264|0);
  $9 = HEAP32[$8>>2]|0;
  _SDL_free($9);
  $10 = (_SDL_strdup($$text)|0);
  HEAP32[$8>>2] = $10;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $7 = (FUNCTION_TABLE_iii[$5 & 8191]($0,$$text)|0);
  $$0 = $7;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_GetClipboardText() {
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_SDL_GetVideoDevice()|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(612262,$vararg_buffer)|0);
  $2 = (_SDL_strdup(1282466)|0);
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 $3 = ((($0)) + 220|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if ($5) {
  $7 = ((($0)) + 264|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8|0)!=(0|0);
  $$ = $9 ? $8 : 1282466;
  $10 = (_SDL_strdup($$)|0);
  $$0 = $10;
  STACKTOP = sp;return ($$0|0);
 } else {
  $6 = (FUNCTION_TABLE_ii[$4 & 8191]($0)|0);
  $$0 = $6;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_SetWindowPosition($window,$x,$y) {
 $window = $window|0;
 $x = $x|0;
 $y = $y|0;
 var $$ = 0, $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $bounds = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $bounds = sp + 16|0;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = $x & -65536;
   $7 = ($6|0)==(805240832);
   $8 = $y & -65536;
   $9 = ($8|0)==(805240832);
   $or$cond = $7 | $9;
   if ($or$cond) {
    $10 = $x & 65535;
    $11 = ((($0)) + 240|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = ($10|0)>($12|0);
    $$ = $13 ? 0 : $10;
    (_SDL_memset($bounds,0,16)|0);
    (_SDL_GetDisplayBounds($$,$bounds)|0);
    if ($7) {
     $14 = HEAP32[$bounds>>2]|0;
     $15 = ((($bounds)) + 8|0);
     $16 = HEAP32[$15>>2]|0;
     $17 = ((($window)) + 24|0);
     $18 = HEAP32[$17>>2]|0;
     $19 = (($16) - ($18))|0;
     $20 = (($19|0) / 2)&-1;
     $21 = (($20) + ($14))|0;
     $$01 = $21;
    } else {
     $$01 = $x;
    }
    if ($9) {
     $22 = ((($bounds)) + 4|0);
     $23 = HEAP32[$22>>2]|0;
     $24 = ((($bounds)) + 12|0);
     $25 = HEAP32[$24>>2]|0;
     $26 = ((($window)) + 28|0);
     $27 = HEAP32[$26>>2]|0;
     $28 = (($25) - ($27))|0;
     $29 = (($28|0) / 2)&-1;
     $30 = (($29) + ($23))|0;
     $$0 = $30;
    } else {
     $$0 = $y;
    }
    $$1 = $$0;$$12 = $$01;
   } else {
    $$1 = $y;$$12 = $x;
   }
   $31 = ((($window)) + 48|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = $32 & 1;
   $34 = ($33|0)==(0);
   $35 = $$12 & -65536;
   $36 = ($35|0)==(536805376);
   if (!($34)) {
    if (!($36)) {
     $37 = ((($window)) + 56|0);
     HEAP32[$37>>2] = $$12;
    }
    $38 = $$1 & -65536;
    $39 = ($38|0)==(536805376);
    if ($39) {
     STACKTOP = sp;return;
    }
    $40 = ((($window)) + 60|0);
    HEAP32[$40>>2] = $$1;
    STACKTOP = sp;return;
   }
   if (!($36)) {
    $41 = ((($window)) + 16|0);
    HEAP32[$41>>2] = $$12;
   }
   $42 = $$1 & -65536;
   $43 = ($42|0)==(536805376);
   if (!($43)) {
    $44 = ((($window)) + 20|0);
    HEAP32[$44>>2] = $$1;
   }
   $45 = HEAP32[300441]|0;
   $46 = ((($45)) + 44|0);
   $47 = HEAP32[$46>>2]|0;
   $48 = ($47|0)==(0|0);
   if (!($48)) {
    FUNCTION_TABLE_vii[$47 & 8191]($45,$window);
   }
   (_SDL_SendWindowEvent($window,4,$$12,$$1)|0);
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 STACKTOP = sp;return;
}
function _SDL_DisableScreenSaver() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  return;
 }
 $2 = ((($0)) + 236|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0);
 if (!($4)) {
  return;
 }
 HEAP32[$2>>2] = 1;
 $5 = ((($0)) + 184|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0|0);
 if ($7) {
  return;
 }
 FUNCTION_TABLE_vi[$6 & 8191]($0);
 return;
}
function _SDL_HasScreenKeyboardSupport() {
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  $$0 = 0;
  return ($$0|0);
 }
 $2 = ((($0)) + 200|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  $$0 = 0;
  return ($$0|0);
 }
 $5 = (FUNCTION_TABLE_ii[$3 & 8191]($0)|0);
 $$0 = $5;
 return ($$0|0);
}
function _SDL_StartTextInput() {
 var $$pr = 0, $$pr$pre = 0, $$pr6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $window$0$i = 0, $window$01$i = 0, $window$02$i = 0, $window$02$i$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 (_SDL_EventState(771,1)|0);
 (_SDL_EventState(770,1)|0);
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  return;
 }
 $2 = ((($0)) + 248|0);
 $window$01$i = HEAP32[$2>>2]|0;
 $3 = ($window$01$i|0)==(0|0);
 do {
  if ($3) {
   $$pr6 = $0;
  } else {
   $window$02$i = $window$01$i;
   while(1) {
    $4 = ((($window$02$i)) + 48|0);
    $5 = HEAP32[$4>>2]|0;
    $6 = $5 & 512;
    $7 = ($6|0)==(0);
    if (!($7)) {
     $window$02$i$lcssa = $window$02$i;
     label = 5;
     break;
    }
    $8 = ((($window$02$i)) + 144|0);
    $window$0$i = HEAP32[$8>>2]|0;
    $9 = ($window$0$i|0)==(0|0);
    if ($9) {
     $$pr = $0;
     break;
    } else {
     $window$02$i = $window$0$i;
    }
   }
   if ((label|0) == 5) {
    $10 = ((($0)) + 204|0);
    $11 = HEAP32[$10>>2]|0;
    $12 = ($11|0)==(0|0);
    if ($12) {
     $$pr6 = $0;
     break;
    }
    FUNCTION_TABLE_vii[$11 & 8191]($0,$window$02$i$lcssa);
    $$pr$pre = HEAP32[300441]|0;
    $$pr = $$pr$pre;
   }
   $13 = ($$pr|0)==(0|0);
   if ($13) {
    return;
   } else {
    $$pr6 = $$pr;
   }
  }
 } while(0);
 $14 = ((($$pr6)) + 188|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)==(0|0);
 if ($16) {
  return;
 }
 FUNCTION_TABLE_vi[$15 & 8191]($$pr6);
 return;
}
function _SDL_GetCurrentVideoDriver() {
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $2 = HEAP32[$0>>2]|0;
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_GetNumVideoDisplays() {
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $2 = ((($0)) + 240|0);
  $3 = HEAP32[$2>>2]|0;
  $$0 = $3;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_GetDisplayName($displayIndex) {
 $displayIndex = $displayIndex|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($displayIndex|0)>(-1);
 $$phi$trans$insert = ((($0)) + 240|0);
 $$pre = HEAP32[$$phi$trans$insert>>2]|0;
 $3 = ($$pre|0)>($displayIndex|0);
 $or$cond = $2 & $3;
 if ($or$cond) {
  $5 = ((($0)) + 244|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (($6) + (($displayIndex*68)|0)|0);
  $8 = HEAP32[$7>>2]|0;
  $$0 = $8;
  STACKTOP = sp;return ($$0|0);
 } else {
  $4 = (($$pre) + -1)|0;
  HEAP32[$vararg_buffer1>>2] = $4;
  (_SDL_SetError(596297,$vararg_buffer1)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_GetNumDisplayModes($displayIndex) {
 $displayIndex = $displayIndex|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($displayIndex|0)>(-1);
 $$phi$trans$insert = ((($0)) + 240|0);
 $$pre = HEAP32[$$phi$trans$insert>>2]|0;
 $3 = ($$pre|0)>($displayIndex|0);
 $or$cond = $2 & $3;
 if (!($or$cond)) {
  $4 = (($$pre) + -1)|0;
  HEAP32[$vararg_buffer1>>2] = $4;
  (_SDL_SetError(596297,$vararg_buffer1)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = ((($0)) + 244|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + (($displayIndex*68)|0)|0);
 $8 = (((($6) + (($displayIndex*68)|0)|0)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)==(0);
 if (!($10)) {
  $$0 = $9;
  STACKTOP = sp;return ($$0|0);
 }
 $11 = ((($0)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)==(0|0);
 if ($13) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 FUNCTION_TABLE_vii[$12 & 8191]($0,$7);
 $14 = (((($6) + (($displayIndex*68)|0)|0)) + 12|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = HEAP32[$8>>2]|0;
 _SDL_qsort($15,$16,20,4476);
 $$pre$i = HEAP32[$8>>2]|0;
 $$0 = $$pre$i;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GetDisplayMode($displayIndex,$index,$mode) {
 $displayIndex = $displayIndex|0;
 $index = $index|0;
 $mode = $mode|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$phi$trans$insert4 = 0, $$pre = 0, $$pre$i = 0, $$pre$i1 = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($displayIndex|0)>(-1);
 $$phi$trans$insert4 = ((($0)) + 240|0);
 $$pre5 = HEAP32[$$phi$trans$insert4>>2]|0;
 $3 = ($$pre5|0)>($displayIndex|0);
 $or$cond = $2 & $3;
 if (!($or$cond)) {
  $4 = (($$pre5) + -1)|0;
  HEAP32[$vararg_buffer1>>2] = $4;
  (_SDL_SetError(596297,$vararg_buffer1)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = ((($0)) + 244|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($6) + (($displayIndex*68)|0)|0);
 $8 = ($index|0)<(0);
 $$phi$trans$insert = (((($6) + (($displayIndex*68)|0)|0)) + 8|0);
 $$pre = HEAP32[$$phi$trans$insert>>2]|0;
 if ($8) {
  $19 = $$pre;
 } else {
  $9 = ($$pre|0)==(0);
  if ($9) {
   $10 = ((($0)) + 20|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = ($11|0)==(0|0);
   if ($12) {
    $17 = 0;
   } else {
    FUNCTION_TABLE_vii[$11 & 8191]($0,$7);
    $13 = (((($6) + (($displayIndex*68)|0)|0)) + 12|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = HEAP32[$$phi$trans$insert>>2]|0;
    _SDL_qsort($14,$15,20,4476);
    $$pre$i = HEAP32[$$phi$trans$insert>>2]|0;
    $17 = $$pre$i;
   }
  } else {
   $17 = $$pre;
  }
  $16 = ($17|0)>($index|0);
  if ($16) {
   $30 = ($mode|0)==(0|0);
   if ($30) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   $31 = (((($6) + (($displayIndex*68)|0)|0)) + 12|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = (($32) + (($index*20)|0)|0);
   ;HEAP32[$mode>>2]=HEAP32[$33>>2]|0;HEAP32[$mode+4>>2]=HEAP32[$33+4>>2]|0;HEAP32[$mode+8>>2]=HEAP32[$33+8>>2]|0;HEAP32[$mode+12>>2]=HEAP32[$33+12>>2]|0;HEAP32[$mode+16>>2]=HEAP32[$33+16>>2]|0;
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  } else {
   $19 = $17;
  }
 }
 $18 = ($19|0)==(0);
 if ($18) {
  $20 = HEAP32[300441]|0;
  $21 = ((($20)) + 20|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = ($22|0)==(0|0);
  if ($23) {
   $28 = 0;
  } else {
   FUNCTION_TABLE_vii[$22 & 8191]($20,$7);
   $24 = (((($6) + (($displayIndex*68)|0)|0)) + 12|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = HEAP32[$$phi$trans$insert>>2]|0;
   _SDL_qsort($25,$26,20,4476);
   $$pre$i1 = HEAP32[$$phi$trans$insert>>2]|0;
   $28 = $$pre$i1;
  }
 } else {
  $28 = $19;
 }
 $27 = (($28) + -1)|0;
 HEAP32[$vararg_buffer3>>2] = $27;
 $29 = (_SDL_SetError(612320,$vararg_buffer3)|0);
 $$0 = $29;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GetDesktopDisplayMode($displayIndex,$mode) {
 $displayIndex = $displayIndex|0;
 $mode = $mode|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($displayIndex|0)>(-1);
 $$phi$trans$insert = ((($0)) + 240|0);
 $$pre = HEAP32[$$phi$trans$insert>>2]|0;
 $3 = ($$pre|0)>($displayIndex|0);
 $or$cond = $2 & $3;
 if (!($or$cond)) {
  $4 = (($$pre) + -1)|0;
  HEAP32[$vararg_buffer1>>2] = $4;
  (_SDL_SetError(596297,$vararg_buffer1)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = ($mode|0)==(0|0);
 if ($5) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $6 = ((($0)) + 244|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (((($7) + (($displayIndex*68)|0)|0)) + 16|0);
 ;HEAP32[$mode>>2]=HEAP32[$8>>2]|0;HEAP32[$mode+4>>2]=HEAP32[$8+4>>2]|0;HEAP32[$mode+8>>2]=HEAP32[$8+8>>2]|0;HEAP32[$mode+12>>2]=HEAP32[$8+12>>2]|0;HEAP32[$mode+16>>2]=HEAP32[$8+16>>2]|0;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GetCurrentDisplayMode($displayIndex,$mode) {
 $displayIndex = $displayIndex|0;
 $mode = $mode|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($displayIndex|0)>(-1);
 $$phi$trans$insert = ((($0)) + 240|0);
 $$pre = HEAP32[$$phi$trans$insert>>2]|0;
 $3 = ($$pre|0)>($displayIndex|0);
 $or$cond = $2 & $3;
 if (!($or$cond)) {
  $4 = (($$pre) + -1)|0;
  HEAP32[$vararg_buffer1>>2] = $4;
  (_SDL_SetError(596297,$vararg_buffer1)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $5 = ($mode|0)==(0|0);
 if ($5) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $6 = ((($0)) + 244|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (((($7) + (($displayIndex*68)|0)|0)) + 36|0);
 ;HEAP32[$mode>>2]=HEAP32[$8>>2]|0;HEAP32[$mode+4>>2]=HEAP32[$8+4>>2]|0;HEAP32[$mode+8>>2]=HEAP32[$8+8>>2]|0;HEAP32[$mode+12>>2]=HEAP32[$8+12>>2]|0;HEAP32[$mode+16>>2]=HEAP32[$8+16>>2]|0;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GetClosestDisplayMode($displayIndex,$mode,$closest) {
 $displayIndex = $displayIndex|0;
 $mode = $mode|0;
 $closest = $closest|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($displayIndex|0)>(-1);
 $$phi$trans$insert = ((($0)) + 240|0);
 $$pre = HEAP32[$$phi$trans$insert>>2]|0;
 $3 = ($$pre|0)>($displayIndex|0);
 $or$cond = $2 & $3;
 if ($or$cond) {
  $5 = ((($0)) + 244|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (($6) + (($displayIndex*68)|0)|0);
  $8 = (_SDL_GetClosestDisplayModeForDisplay($7,$mode,$closest)|0);
  $$0 = $8;
  STACKTOP = sp;return ($$0|0);
 } else {
  $4 = (($$pre) + -1)|0;
  HEAP32[$vararg_buffer1>>2] = $4;
  (_SDL_SetError(596297,$vararg_buffer1)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_SetWindowDisplayMode($window,$mode) {
 $window = $window|0;
 $mode = $mode|0;
 var $$0 = 0, $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $fullscreen_mode = 0, $or$cond3$not = 0, $or$cond5 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $fullscreen_mode = sp + 12|0;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ($mode|0)==(0|0);
   $7 = ((($window)) + 72|0);
   if ($6) {
    (_SDL_memset($7,0,20)|0);
   } else {
    ;HEAP32[$7>>2]=HEAP32[$mode>>2]|0;HEAP32[$7+4>>2]=HEAP32[$mode+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$mode+8>>2]|0;HEAP32[$7+12>>2]=HEAP32[$mode+12>>2]|0;HEAP32[$7+16>>2]=HEAP32[$mode+16>>2]|0;
   }
   $8 = ((($window)) + 48|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = $9 & 69;
   $or$cond3$not = ($10|0)!=(5);
   $11 = $9 & 4097;
   $12 = ($11|0)==(4097);
   $or$cond5 = $or$cond3$not | $12;
   if ($or$cond5) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   $13 = (_SDL_GetWindowDisplayMode($window,$fullscreen_mode)|0);
   $14 = ($13|0)==(0);
   if ($14) {
    $15 = (_SDL_GetWindowDisplayIndex($window)|0);
    $16 = ($15|0)>(-1);
    if ($16) {
     $17 = HEAP32[300441]|0;
     $18 = ((($17)) + 244|0);
     $19 = HEAP32[$18>>2]|0;
     $20 = (($19) + (($15*68)|0)|0);
     $$0$i = $20;
    } else {
     $$0$i = 0;
    }
    (_SDL_SetDisplayModeForDisplay($$0$i,$fullscreen_mode)|0);
   }
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_SetWindowTitle($window,$title) {
 $window = $window|0;
 $title = $title|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ((($window)) + 8|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==($title|0);
   if ($8) {
    STACKTOP = sp;return;
   }
   _SDL_free($7);
   $9 = ($title|0)!=(0|0);
   $10 = $9 ? $title : 1282466;
   $11 = (_SDL_strdup($10)|0);
   HEAP32[$6>>2] = $11;
   $12 = HEAP32[300441]|0;
   $13 = ((($12)) + 36|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ($14|0)==(0|0);
   if ($15) {
    STACKTOP = sp;return;
   }
   FUNCTION_TABLE_vii[$14 & 8191]($12,$window);
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 STACKTOP = sp;return;
}
function _SDL_SetWindowIcon($window,$icon) {
 $window = $window|0;
 $icon = $icon|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ($icon|0)==(0|0);
   if ($6) {
    STACKTOP = sp;return;
   }
   $7 = ((($window)) + 12|0);
   $8 = HEAP32[$7>>2]|0;
   _SDL_FreeSurface($8);
   $9 = (_SDL_ConvertSurfaceFormat($icon,372645892,0)|0);
   HEAP32[$7>>2] = $9;
   $10 = ($9|0)==(0|0);
   if ($10) {
    STACKTOP = sp;return;
   }
   $11 = HEAP32[300441]|0;
   $12 = ((($11)) + 40|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = ($13|0)==(0|0);
   if ($14) {
    STACKTOP = sp;return;
   }
   FUNCTION_TABLE_viii[$13 & 8191]($11,$window,$9);
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 STACKTOP = sp;return;
}
function _SDL_GetWindowPosition($window,$x,$y) {
 $window = $window|0;
 $x = $x|0;
 $y = $y|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $bounds = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $bounds = sp + 16|0;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ((($window)) + 48|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = $7 & 1;
   $9 = ($8|0)==(0);
   if ($9) {
    $17 = ($x|0)==(0|0);
    if (!($17)) {
     $18 = ((($window)) + 16|0);
     $19 = HEAP32[$18>>2]|0;
     HEAP32[$x>>2] = $19;
    }
    $20 = ($y|0)==(0|0);
    if ($20) {
     STACKTOP = sp;return;
    }
    $21 = ((($window)) + 20|0);
    $22 = HEAP32[$21>>2]|0;
    HEAP32[$y>>2] = $22;
    STACKTOP = sp;return;
   }
   $10 = ($x|0)!=(0|0);
   if ($10) {
    HEAP32[$x>>2] = 0;
   }
   $11 = ($y|0)!=(0|0);
   if ($11) {
    HEAP32[$y>>2] = 0;
   }
   $12 = (_SDL_GetWindowDisplayIndex($window)|0);
   $13 = ($12|0)>(-1);
   if (!($13)) {
    STACKTOP = sp;return;
   }
   (_SDL_memset($bounds,0,16)|0);
   (_SDL_GetDisplayBounds($12,$bounds)|0);
   if ($10) {
    $14 = HEAP32[$bounds>>2]|0;
    HEAP32[$x>>2] = $14;
   }
   if ($11) {
    $15 = ((($bounds)) + 4|0);
    $16 = HEAP32[$15>>2]|0;
    HEAP32[$y>>2] = $16;
   }
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 STACKTOP = sp;return;
}
function _SDL_SetWindowSize($window,$w,$h) {
 $window = $window|0;
 $w = $w|0;
 $h = $h|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond3 = 0, $or$cond4 = 0, $or$cond5 = 0, $or$cond7$not = 0, $or$cond9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ($w|0)<(1);
   if ($6) {
    HEAP32[$vararg_buffer3>>2] = 612357;
    (_SDL_SetError(611267,$vararg_buffer3)|0);
    STACKTOP = sp;return;
   }
   $7 = ($h|0)<(1);
   if ($7) {
    HEAP32[$vararg_buffer5>>2] = 612359;
    (_SDL_SetError(611267,$vararg_buffer5)|0);
    STACKTOP = sp;return;
   }
   $8 = ((($window)) + 32|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = ($9|0)!=(0);
   $11 = ($9|0)>($w|0);
   $or$cond = $10 & $11;
   $$01 = $or$cond ? $9 : $w;
   $12 = ((($window)) + 40|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = ($13|0)!=(0);
   $15 = ($$01|0)>($13|0);
   $or$cond3 = $14 & $15;
   $$12 = $or$cond3 ? $13 : $$01;
   $16 = ((($window)) + 36|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ($17|0)!=(0);
   $19 = ($17|0)>($h|0);
   $or$cond4 = $18 & $19;
   $$0 = $or$cond4 ? $17 : $h;
   $20 = ((($window)) + 44|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = ($21|0)!=(0);
   $23 = ($$0|0)>($21|0);
   $or$cond5 = $22 & $23;
   $$1 = $or$cond5 ? $21 : $$0;
   $24 = ((($window)) + 64|0);
   HEAP32[$24>>2] = $$12;
   $25 = ((($window)) + 68|0);
   HEAP32[$25>>2] = $$1;
   $26 = ((($window)) + 48|0);
   $27 = HEAP32[$26>>2]|0;
   $28 = $27 & 1;
   $29 = ($28|0)==(0);
   if (!($29)) {
    $30 = $27 & 68;
    $or$cond7$not = ($30|0)!=(4);
    $31 = $27 & 4097;
    $32 = ($31|0)==(4097);
    $or$cond9 = $or$cond7$not | $32;
    if ($or$cond9) {
     STACKTOP = sp;return;
    }
    $33 = ((($window)) + 52|0);
    HEAP32[$33>>2] = 0;
    (_SDL_UpdateFullscreenMode($window,1)|0);
    STACKTOP = sp;return;
   }
   $34 = ((($window)) + 24|0);
   HEAP32[$34>>2] = $$12;
   $35 = ((($window)) + 28|0);
   HEAP32[$35>>2] = $$1;
   $36 = ((($0)) + 48|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = ($37|0)==(0|0);
   if (!($38)) {
    FUNCTION_TABLE_vii[$37 & 8191]($0,$window);
    $$pre = HEAP32[$34>>2]|0;
    $39 = ($$pre|0)==($$12|0);
    if (!($39)) {
     STACKTOP = sp;return;
    }
   }
   $40 = HEAP32[$35>>2]|0;
   $41 = ($40|0)==($$1|0);
   if (!($41)) {
    STACKTOP = sp;return;
   }
   $42 = ((($window)) + 108|0);
   HEAP32[$42>>2] = 0;
   (_SDL_SendWindowEvent($window,6,$$12,$$1)|0);
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 STACKTOP = sp;return;
}
function _SDL_SetWindowMinimumSize($window,$min_w,$min_h) {
 $window = $window|0;
 $min_w = $min_w|0;
 $min_h = $min_h|0;
 var $$ = 0, $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ($min_w|0)<(1);
   if ($6) {
    HEAP32[$vararg_buffer3>>2] = 612361;
    (_SDL_SetError(611267,$vararg_buffer3)|0);
    STACKTOP = sp;return;
   }
   $7 = ($min_h|0)<(1);
   if ($7) {
    HEAP32[$vararg_buffer5>>2] = 612367;
    (_SDL_SetError(611267,$vararg_buffer5)|0);
    STACKTOP = sp;return;
   }
   $8 = ((($window)) + 48|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = $9 & 1;
   $11 = ($10|0)==(0);
   if (!($11)) {
    STACKTOP = sp;return;
   }
   $12 = ((($window)) + 32|0);
   HEAP32[$12>>2] = $min_w;
   $13 = ((($window)) + 36|0);
   HEAP32[$13>>2] = $min_h;
   $14 = ((($0)) + 52|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)==(0|0);
   if ($16) {
    $20 = $min_w;$24 = $min_h;
   } else {
    FUNCTION_TABLE_vii[$15 & 8191]($0,$window);
    $$pre = HEAP32[$12>>2]|0;
    $$pre1 = HEAP32[$13>>2]|0;
    $20 = $$pre;$24 = $$pre1;
   }
   $17 = ((($window)) + 24|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = ($18|0)>($20|0);
   $$ = $19 ? $18 : $20;
   $21 = ((($window)) + 28|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = ($22|0)>($24|0);
   $25 = $23 ? $22 : $24;
   _SDL_SetWindowSize($window,$$,$25);
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 STACKTOP = sp;return;
}
function _SDL_RaiseWindow($window) {
 $window = $window|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ((($window)) + 48|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = $7 & 4;
   $9 = ($8|0)==(0);
   if ($9) {
    STACKTOP = sp;return;
   }
   $10 = ((($0)) + 68|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = ($11|0)==(0|0);
   if ($12) {
    STACKTOP = sp;return;
   }
   FUNCTION_TABLE_vii[$11 & 8191]($0,$window);
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 STACKTOP = sp;return;
}
function _SDL_MaximizeWindow($window) {
 $window = $window|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  STACKTOP = sp;return;
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ((($window)) + 48|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = $7 & 128;
   $9 = ($8|0)==(0);
   if (!($9)) {
    STACKTOP = sp;return;
   }
   $10 = ((($0)) + 72|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = ($11|0)==(0|0);
   if ($12) {
    STACKTOP = sp;return;
   }
   FUNCTION_TABLE_vii[$11 & 8191]($0,$window);
   STACKTOP = sp;return;
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 STACKTOP = sp;return;
}
function _SDL_SetWindowFullscreen($window,$flags) {
 $window = $window|0;
 $flags = $flags|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = $flags & 4097;
   $7 = ((($window)) + 48|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = $8 & 4097;
   $10 = ($6|0)==($9|0);
   if ($10) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   $11 = $8 & -4098;
   $12 = $11 | $6;
   HEAP32[$7>>2] = $12;
   $13 = $flags & 1;
   $14 = ($13|0)==(0);
   $15 = $8 & 4;
   $16 = ($15|0)==(0);
   $or$cond = $14 | $16;
   if ($or$cond) {
    $20 = 0;
   } else {
    $17 = $8 & 64;
    $18 = ($17|0)==(0);
    $20 = $18;
   }
   $19 = $20&1;
   $21 = (_SDL_UpdateFullscreenMode($window,$19)|0);
   $22 = ($21|0)==(0);
   if ($22) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
   $23 = HEAP32[$7>>2]|0;
   $24 = $23 & -4098;
   $25 = $24 | $9;
   HEAP32[$7>>2] = $25;
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 $$0 = -1;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_GetWindowGrab($window) {
 $window = $window|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_SetError(595917,$vararg_buffer)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ($window|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$window>>2]|0;
  $4 = ((($0)) + 256|0);
  $5 = ($3|0)==($4|0);
  if ($5) {
   $6 = ((($0)) + 252|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==($window|0);
   $9 = $8&1;
   $$0 = $9;
   STACKTOP = sp;return ($$0|0);
  }
 }
 (_SDL_SetError(609414,$vararg_buffer1)|0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_IsScreenSaverEnabled() {
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  $$0 = 1;
  return ($$0|0);
 }
 $2 = ((($0)) + 236|0);
 $3 = HEAP32[$2>>2]|0;
 $not$ = ($3|0)==(0);
 $4 = $not$&1;
 $$0 = $4;
 return ($$0|0);
}
function _SDL_EnableScreenSaver() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  return;
 }
 $2 = ((($0)) + 236|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0);
 if ($4) {
  return;
 }
 HEAP32[$2>>2] = 0;
 $5 = ((($0)) + 184|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0|0);
 if ($7) {
  return;
 }
 FUNCTION_TABLE_vi[$6 & 8191]($0);
 return;
}
function _SDL_IsTextInputActive() {
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_EventState(771,-1)|0);
 $1 = ($0<<24>>24)==(1);
 $2 = $1&1;
 return ($2|0);
}
function _SDL_StopTextInput() {
 var $$pr$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $window$0$i = 0;
 var $window$01$i = 0, $window$02$i = 0, $window$02$i$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 L1: do {
  if (!($1)) {
   $2 = ((($0)) + 192|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = ($3|0)==(0|0);
   if ($4) {
    $7 = $0;
   } else {
    FUNCTION_TABLE_vi[$3 & 8191]($0);
    $$pr$pre = HEAP32[300441]|0;
    $5 = ($$pr$pre|0)==(0|0);
    if ($5) {
     break;
    } else {
     $7 = $$pr$pre;
    }
   }
   $6 = ((($7)) + 248|0);
   $window$01$i = HEAP32[$6>>2]|0;
   $8 = ($window$01$i|0)==(0|0);
   if (!($8)) {
    $window$02$i = $window$01$i;
    while(1) {
     $9 = ((($window$02$i)) + 48|0);
     $10 = HEAP32[$9>>2]|0;
     $11 = $10 & 512;
     $12 = ($11|0)==(0);
     if (!($12)) {
      $window$02$i$lcssa = $window$02$i;
      break;
     }
     $13 = ((($window$02$i)) + 144|0);
     $window$0$i = HEAP32[$13>>2]|0;
     $14 = ($window$0$i|0)==(0|0);
     if ($14) {
      break L1;
     } else {
      $window$02$i = $window$0$i;
     }
    }
    $15 = ((($7)) + 208|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ($16|0)==(0|0);
    if (!($17)) {
     FUNCTION_TABLE_vii[$16 & 8191]($7,$window$02$i$lcssa);
    }
   }
  }
 } while(0);
 (_SDL_EventState(771,0)|0);
 (_SDL_EventState(770,0)|0);
 return;
}
function _SDL_SetTextInputRect($rect) {
 $rect = $rect|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[300441]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  return;
 }
 $2 = ((($0)) + 196|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  return;
 }
 FUNCTION_TABLE_vii[$3 & 8191]($0,$rect);
 return;
}
function _SDL_ShowSimpleMessageBox($flags,$title,$message,$window) {
 $flags = $flags|0;
 $title = $title|0;
 $message = $message|0;
 $window = $window|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $button = 0, $data = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $data = sp + 12|0;
 $button = sp;
 (_SDL_memset($data,0,28)|0);
 HEAP32[$data>>2] = $flags;
 $0 = ((($data)) + 8|0);
 HEAP32[$0>>2] = $title;
 $1 = ((($data)) + 12|0);
 HEAP32[$1>>2] = $message;
 $2 = ((($data)) + 16|0);
 HEAP32[$2>>2] = 1;
 $3 = ((($data)) + 20|0);
 HEAP32[$3>>2] = $button;
 $4 = ((($data)) + 4|0);
 HEAP32[$4>>2] = $window;
 (_SDL_memset($button,0,12)|0);
 $5 = HEAP32[$button>>2]|0;
 $6 = $5 | 3;
 HEAP32[$button>>2] = $6;
 $7 = ((($button)) + 8|0);
 HEAP32[$7>>2] = 612373;
 $8 = (_SDL_ShowMessageBox($data,0)|0);
 STACKTOP = sp;return ($8|0);
}
function _SDL_CreateCond() {
 var $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cond$0 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_malloc(20)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  (_SDL_Error(0)|0);
  $cond$0 = $0;
  return ($cond$0|0);
 }
 $2 = (_SDL_CreateMutex()|0);
 HEAP32[$0>>2] = $2;
 $3 = (_SDL_CreateSemaphore(0)|0);
 $4 = ((($0)) + 12|0);
 HEAP32[$4>>2] = $3;
 $5 = (_SDL_CreateSemaphore(0)|0);
 $6 = ((($0)) + 16|0);
 HEAP32[$6>>2] = $5;
 $7 = ((($0)) + 8|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($0)) + 4|0);
 HEAP32[$8>>2] = 0;
 $9 = HEAP32[$0>>2]|0;
 $10 = ($9|0)==(0|0);
 $$pre = HEAP32[$4>>2]|0;
 $11 = ($$pre|0)==(0|0);
 if ($10) {
  if ($11) {
   $14 = $5;
  } else {
   label = 6;
  }
 } else {
  if ($11) {
   $14 = $5;
  } else {
   $12 = ($5|0)==(0|0);
   if ($12) {
    label = 6;
   } else {
    $cond$0 = $0;
    return ($cond$0|0);
   }
  }
 }
 if ((label|0) == 6) {
  $$pre1 = HEAP32[$6>>2]|0;
  $14 = $$pre1;
 }
 $13 = ($14|0)==(0|0);
 if (!($13)) {
 }
 $15 = HEAP32[$0>>2]|0;
 $16 = ($15|0)==(0|0);
 if (!($16)) {
  _SDL_DestroyMutex($15);
 }
 _SDL_free($0);
 $cond$0 = 0;
 return ($cond$0|0);
}
function _SDL_DestroyCond($cond) {
 $cond = $cond|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($cond|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($cond)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
 }
 $4 = ((($cond)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if (!($6)) {
 }
 $7 = HEAP32[$cond>>2]|0;
 $8 = ($7|0)==(0|0);
 if (!($8)) {
  _SDL_DestroyMutex($7);
 }
 _SDL_free($cond);
 return;
}
function _SDL_CondSignal($cond) {
 $cond = $cond|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($cond|0)==(0|0);
 if ($0) {
  $1 = (_SDL_SetError(612376,$vararg_buffer)|0);
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ((($cond)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($cond)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($3|0)>($5|0);
 if ($6) {
  $7 = (($5) + 1)|0;
  HEAP32[$4>>2] = $7;
  $8 = ((($cond)) + 12|0);
  $9 = HEAP32[$8>>2]|0;
  (_SDL_SemPost($9)|0);
  $10 = ((($cond)) + 16|0);
  $11 = HEAP32[$10>>2]|0;
  (_SDL_SemWait($11)|0);
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 } else {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function _SDL_CondBroadcast($cond) {
 $cond = $cond|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $exitcond4 = 0, $i$02 = 0, $i$11 = 0;
 var $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($cond|0)==(0|0);
 do {
  if ($0) {
   $1 = (_SDL_SetError(612376,$vararg_buffer)|0);
   $$0 = $1;
  } else {
   $2 = ((($cond)) + 4|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = ((($cond)) + 8|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = ($3|0)>($5|0);
   if (!($6)) {
    $$0 = 0;
    break;
   }
   $7 = (($3) - ($5))|0;
   HEAP32[$4>>2] = $3;
   $8 = ($7|0)>(0);
   if (!($8)) {
    $$0 = 0;
    break;
   }
   $9 = ((($cond)) + 12|0);
   $i$02 = 0;
   while(1) {
    $10 = HEAP32[$9>>2]|0;
    (_SDL_SemPost($10)|0);
    $11 = (($i$02) + 1)|0;
    $exitcond4 = ($11|0)==($7|0);
    if ($exitcond4) {
     break;
    } else {
     $i$02 = $11;
    }
   }
   if ($8) {
    $12 = ((($cond)) + 16|0);
    $i$11 = 0;
    while(1) {
     $13 = HEAP32[$12>>2]|0;
     (_SDL_SemWait($13)|0);
     $14 = (($i$11) + 1)|0;
     $exitcond = ($14|0)==($7|0);
     if ($exitcond) {
      $$0 = 0;
      break;
     } else {
      $i$11 = $14;
     }
    }
   } else {
    $$0 = 0;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _SDL_CondWaitTimeout($cond,$mutex,$ms) {
 $cond = $cond|0;
 $mutex = $mutex|0;
 $ms = $ms|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $retval$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ($cond|0)==(0|0);
 if ($0) {
  $1 = (_SDL_SetError(612376,$vararg_buffer)|0);
  $$0 = $1;
  STACKTOP = sp;return ($$0|0);
 }
 $2 = ((($cond)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + 1)|0;
 HEAP32[$2>>2] = $4;
 $5 = ($ms|0)==(-1);
 $6 = ((($cond)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 if ($5) {
  $8 = (_SDL_SemWait($7)|0);
  $retval$0 = $8;
 } else {
  $9 = (_SDL_SemWaitTimeout($7,$ms)|0);
  $retval$0 = $9;
 }
 $10 = ((($cond)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)>(0);
 if ($12) {
  $13 = ($retval$0|0)>(0);
  if ($13) {
   $14 = HEAP32[$6>>2]|0;
   (_SDL_SemWait($14)|0);
  }
  $15 = ((($cond)) + 16|0);
  $16 = HEAP32[$15>>2]|0;
  (_SDL_SemPost($16)|0);
  $17 = HEAP32[$10>>2]|0;
  $18 = (($17) + -1)|0;
  HEAP32[$10>>2] = $18;
 }
 $19 = HEAP32[$2>>2]|0;
 $20 = (($19) + -1)|0;
 HEAP32[$2>>2] = $20;
 $$0 = $retval$0;
 STACKTOP = sp;return ($$0|0);
}
function _SDL_CondWait($cond,$mutex) {
 $cond = $cond|0;
 $mutex = $mutex|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_SDL_CondWaitTimeout($cond,$mutex,-1)|0);
 return ($0|0);
}
function _compress($dest,$destLen,$source,$sourceLen) {
 $dest = $dest|0;
 $destLen = $destLen|0;
 $source = $source|0;
 $sourceLen = $sourceLen|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $stream$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $stream$i = sp;
 HEAP32[$stream$i>>2] = $source;
 $0 = ((($stream$i)) + 4|0);
 HEAP32[$0>>2] = $sourceLen;
 $1 = ((($stream$i)) + 12|0);
 HEAP32[$1>>2] = $dest;
 $2 = HEAP32[$destLen>>2]|0;
 $3 = ((($stream$i)) + 16|0);
 HEAP32[$3>>2] = $2;
 $4 = ((($stream$i)) + 32|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($stream$i)) + 36|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($stream$i)) + 40|0);
 HEAP32[$6>>2] = 0;
 $7 = (_deflateInit_($stream$i,-1,613577,56)|0);
 $8 = ($7|0)==(0);
 if (!($8)) {
  $$0$i = $7;
  STACKTOP = sp;return ($$0$i|0);
 }
 $9 = (_deflate($stream$i,4)|0);
 $10 = ($9|0)==(1);
 if ($10) {
  $13 = ((($stream$i)) + 20|0);
  $14 = HEAP32[$13>>2]|0;
  HEAP32[$destLen>>2] = $14;
  $15 = (_deflateEnd($stream$i)|0);
  $$0$i = $15;
  STACKTOP = sp;return ($$0$i|0);
 } else {
  (_deflateEnd($stream$i)|0);
  $11 = ($9|0)==(0);
  $12 = $11 ? -5 : $9;
  $$0$i = $12;
  STACKTOP = sp;return ($$0$i|0);
 }
 return (0)|0;
}
function _compressBound($sourceLen) {
 $sourceLen = $sourceLen|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $sourceLen >>> 12;
 $1 = $sourceLen >>> 14;
 $2 = $sourceLen >>> 25;
 $3 = (($sourceLen) + 13)|0;
 $4 = (($3) + ($0))|0;
 $5 = (($4) + ($1))|0;
 $6 = (($5) + ($2))|0;
 return ($6|0);
}
function _deflateInit_($strm,$level,$version,$stream_size) {
 $strm = $strm|0;
 $level = $level|0;
 $version = $version|0;
 $stream_size = $stream_size|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_deflateInit2_($strm,$level,8,15,8,0,$version,$stream_size)|0);
 return ($0|0);
}
function _deflateInit2_($strm,$level,$method,$windowBits,$memLevel,$strategy,$version,$stream_size) {
 $strm = $strm|0;
 $level = $level|0;
 $method = $method|0;
 $windowBits = $windowBits|0;
 $memLevel = $memLevel|0;
 $strategy = $strategy|0;
 $version = $version|0;
 $stream_size = $stream_size|0;
 var $$ = 0, $$$019 = 0, $$0 = 0, $$019 = 0, $$in$i = 0, $$in20$i = 0, $$in21$i = 0, $$level = 0, $$pre$i = 0, $$pre12$i = 0, $$pre13$i = 0, $$pre14$i = 0, $$windowBits = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $memLevel$off = 0, $or$cond = 0, $or$cond20 = 0, $or$cond5 = 0, $wrap$0 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ($version|0)==(0|0);
 if ($0) {
  $$0 = -6;
  return ($$0|0);
 }
 $1 = HEAP8[$version>>0]|0;
 $2 = ($1<<24>>24)!=(49);
 $3 = ($stream_size|0)!=(56);
 $or$cond = $3 | $2;
 if ($or$cond) {
  $$0 = -6;
  return ($$0|0);
 }
 $4 = ($strm|0)==(0|0);
 if ($4) {
  $$0 = -2;
  return ($$0|0);
 }
 $5 = ((($strm)) + 24|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($strm)) + 32|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 if ($8) {
  HEAP32[$6>>2] = 4679;
  $9 = ((($strm)) + 40|0);
  HEAP32[$9>>2] = 0;
  $30 = 4679;
 } else {
  $30 = $7;
 }
 $10 = ((($strm)) + 36|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0|0);
 if ($12) {
  HEAP32[$10>>2] = 4680;
 }
 $13 = ($level|0)==(-1);
 $$level = $13 ? 6 : $level;
 $14 = ($windowBits|0)<(0);
 if ($14) {
  $15 = (0 - ($windowBits))|0;
  $$019 = $15;$wrap$0 = 0;
 } else {
  $16 = ($windowBits|0)>(15);
  $17 = (($windowBits) + -16)|0;
  $$windowBits = $16 ? $17 : $windowBits;
  $$ = $16 ? 2 : 1;
  $$019 = $$windowBits;$wrap$0 = $$;
 }
 $memLevel$off = (($memLevel) + -1)|0;
 $18 = ($memLevel$off>>>0)>(8);
 $19 = ($method|0)!=(8);
 $or$cond5 = $19 | $18;
 $20 = $$019 & -8;
 $21 = ($20|0)!=(8);
 $22 = $or$cond5 | $21;
 $23 = ($$level>>>0)>(9);
 $24 = $23 | $22;
 $25 = ($strategy>>>0)>(4);
 $26 = $25 | $24;
 if ($26) {
  $$0 = -2;
  return ($$0|0);
 }
 $27 = ($$019|0)==(8);
 $$$019 = $27 ? 9 : $$019;
 $28 = ((($strm)) + 40|0);
 $29 = HEAP32[$28>>2]|0;
 $31 = (FUNCTION_TABLE_iiii[$30 & 8191]($29,1,5828)|0);
 $32 = ($31|0)==(0|0);
 if ($32) {
  $$0 = -4;
  return ($$0|0);
 }
 $33 = ((($strm)) + 28|0);
 HEAP32[$33>>2] = $31;
 HEAP32[$31>>2] = $strm;
 $34 = ((($31)) + 24|0);
 HEAP32[$34>>2] = $wrap$0;
 $35 = ((($31)) + 28|0);
 HEAP32[$35>>2] = 0;
 $36 = ((($31)) + 48|0);
 HEAP32[$36>>2] = $$$019;
 $37 = 1 << $$$019;
 $38 = ((($31)) + 44|0);
 HEAP32[$38>>2] = $37;
 $39 = (($37) + -1)|0;
 $40 = ((($31)) + 52|0);
 HEAP32[$40>>2] = $39;
 $41 = (($memLevel) + 7)|0;
 $42 = ((($31)) + 80|0);
 HEAP32[$42>>2] = $41;
 $43 = 1 << $41;
 $44 = ((($31)) + 76|0);
 HEAP32[$44>>2] = $43;
 $45 = (($43) + -1)|0;
 $46 = ((($31)) + 84|0);
 HEAP32[$46>>2] = $45;
 $47 = (($memLevel) + 9)|0;
 $48 = (($47>>>0) / 3)&-1;
 $49 = ((($31)) + 88|0);
 HEAP32[$49>>2] = $48;
 $50 = HEAP32[$6>>2]|0;
 $51 = HEAP32[$28>>2]|0;
 $52 = (FUNCTION_TABLE_iiii[$50 & 8191]($51,$37,2)|0);
 $53 = ((($31)) + 56|0);
 HEAP32[$53>>2] = $52;
 $54 = HEAP32[$6>>2]|0;
 $55 = HEAP32[$28>>2]|0;
 $56 = HEAP32[$38>>2]|0;
 $57 = (FUNCTION_TABLE_iiii[$54 & 8191]($55,$56,2)|0);
 $58 = ((($31)) + 64|0);
 HEAP32[$58>>2] = $57;
 $59 = HEAP32[$6>>2]|0;
 $60 = HEAP32[$28>>2]|0;
 $61 = HEAP32[$44>>2]|0;
 $62 = (FUNCTION_TABLE_iiii[$59 & 8191]($60,$61,2)|0);
 $63 = ((($31)) + 68|0);
 HEAP32[$63>>2] = $62;
 $64 = ((($31)) + 5824|0);
 HEAP32[$64>>2] = 0;
 $65 = (($memLevel) + 6)|0;
 $66 = 1 << $65;
 $67 = ((($31)) + 5788|0);
 HEAP32[$67>>2] = $66;
 $68 = HEAP32[$6>>2]|0;
 $69 = HEAP32[$28>>2]|0;
 $70 = (FUNCTION_TABLE_iiii[$68 & 8191]($69,$66,4)|0);
 $71 = ((($31)) + 8|0);
 HEAP32[$71>>2] = $70;
 $72 = HEAP32[$67>>2]|0;
 $73 = $72 << 2;
 $74 = ((($31)) + 12|0);
 HEAP32[$74>>2] = $73;
 $75 = HEAP32[$53>>2]|0;
 $76 = ($75|0)==(0|0);
 if (!($76)) {
  $77 = HEAP32[$58>>2]|0;
  $78 = ($77|0)==(0|0);
  if (!($78)) {
   $79 = HEAP32[$63>>2]|0;
   $80 = ($79|0)==(0|0);
   $81 = ($70|0)==(0|0);
   $or$cond20 = $80 | $81;
   if (!($or$cond20)) {
    $111 = $72 >>> 1;
    $112 = (($70) + ($111<<1)|0);
    $113 = ((($31)) + 5796|0);
    HEAP32[$113>>2] = $112;
    $114 = ($72*3)|0;
    $115 = (($70) + ($114)|0);
    $116 = ((($31)) + 5784|0);
    HEAP32[$116>>2] = $115;
    $117 = ((($31)) + 132|0);
    HEAP32[$117>>2] = $$level;
    $118 = ((($31)) + 136|0);
    HEAP32[$118>>2] = $strategy;
    $119 = ((($31)) + 36|0);
    HEAP8[$119>>0] = 8;
    $120 = (_deflateReset($strm)|0);
    $$0 = $120;
    return ($$0|0);
   }
  }
 }
 $82 = ((($31)) + 4|0);
 HEAP32[$82>>2] = 666;
 $83 = HEAP32[(141764)>>2]|0;
 HEAP32[$5>>2] = $83;
 $84 = HEAP32[$33>>2]|0;
 $85 = ($84|0)==(0|0);
 if ($85) {
  $$0 = -4;
  return ($$0|0);
 }
 $86 = ((($84)) + 4|0);
 $87 = HEAP32[$86>>2]|0;
 switch ($87|0) {
 case 42: case 69: case 73: case 91: case 103: case 113: case 666:  {
  break;
 }
 default: {
  $$0 = -4;
  return ($$0|0);
 }
 }
 $88 = ((($84)) + 8|0);
 $89 = HEAP32[$88>>2]|0;
 $90 = ($89|0)==(0|0);
 if ($90) {
  $$in$i = $84;
 } else {
  $91 = HEAP32[$10>>2]|0;
  $92 = HEAP32[$28>>2]|0;
  FUNCTION_TABLE_vii[$91 & 8191]($92,$89);
  $$pre$i = HEAP32[$33>>2]|0;
  $$in$i = $$pre$i;
 }
 $93 = ((($$in$i)) + 68|0);
 $94 = HEAP32[$93>>2]|0;
 $95 = ($94|0)==(0|0);
 if ($95) {
  $$in20$i = $$in$i;
 } else {
  $96 = HEAP32[$10>>2]|0;
  $97 = HEAP32[$28>>2]|0;
  FUNCTION_TABLE_vii[$96 & 8191]($97,$94);
  $$pre12$i = HEAP32[$33>>2]|0;
  $$in20$i = $$pre12$i;
 }
 $98 = ((($$in20$i)) + 64|0);
 $99 = HEAP32[$98>>2]|0;
 $100 = ($99|0)==(0|0);
 if ($100) {
  $$in21$i = $$in20$i;
 } else {
  $101 = HEAP32[$10>>2]|0;
  $102 = HEAP32[$28>>2]|0;
  FUNCTION_TABLE_vii[$101 & 8191]($102,$99);
  $$pre13$i = HEAP32[$33>>2]|0;
  $$in21$i = $$pre13$i;
 }
 $103 = ((($$in21$i)) + 56|0);
 $104 = HEAP32[$103>>2]|0;
 $105 = ($104|0)==(0|0);
 if ($105) {
  $110 = $$in21$i;
 } else {
  $106 = HEAP32[$10>>2]|0;
  $107 = HEAP32[$28>>2]|0;
  FUNCTION_TABLE_vii[$106 & 8191]($107,$104);
  $$pre14$i = HEAP32[$33>>2]|0;
  $110 = $$pre14$i;
 }
 $108 = HEAP32[$10>>2]|0;
 $109 = HEAP32[$28>>2]|0;
 FUNCTION_TABLE_vii[$108 & 8191]($109,$110);
 HEAP32[$33>>2] = 0;
 $$0 = -4;
 return ($$0|0);
}
function _deflateReset($strm) {
 $strm = $strm|0;
 var $$0$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($strm|0)==(0|0);
 if ($0) {
  $$0$i1 = -2;
  return ($$0$i1|0);
 }
 $1 = ((($strm)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $$0$i1 = -2;
  return ($$0$i1|0);
 }
 $4 = ((($strm)) + 32|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if ($6) {
  $$0$i1 = -2;
  return ($$0$i1|0);
 }
 $7 = ((($strm)) + 36|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 if ($9) {
  $$0$i1 = -2;
  return ($$0$i1|0);
 }
 $10 = ((($strm)) + 20|0);
 HEAP32[$10>>2] = 0;
 $11 = ((($strm)) + 8|0);
 HEAP32[$11>>2] = 0;
 $12 = ((($strm)) + 24|0);
 HEAP32[$12>>2] = 0;
 $13 = ((($strm)) + 44|0);
 HEAP32[$13>>2] = 2;
 $14 = ((($2)) + 20|0);
 HEAP32[$14>>2] = 0;
 $15 = ((($2)) + 8|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($2)) + 16|0);
 HEAP32[$17>>2] = $16;
 $18 = ((($2)) + 24|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($19|0)<(0);
 if ($20) {
  $21 = (0 - ($19))|0;
  HEAP32[$18>>2] = $21;
  $23 = $21;
 } else {
  $23 = $19;
 }
 $22 = ($23|0)!=(0);
 $24 = $22 ? 42 : 113;
 $25 = ((($2)) + 4|0);
 HEAP32[$25>>2] = $24;
 $26 = ($23|0)==(2);
 if ($26) {
  $27 = (_crc32(0,0,0)|0);
  $30 = $27;
 } else {
  $28 = (_adler32(0,0,0)|0);
  $30 = $28;
 }
 $29 = ((($strm)) + 48|0);
 HEAP32[$29>>2] = $30;
 $31 = ((($2)) + 40|0);
 HEAP32[$31>>2] = 0;
 __tr_init($2);
 $32 = HEAP32[$1>>2]|0;
 $33 = ((($32)) + 44|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = $34 << 1;
 $36 = ((($32)) + 60|0);
 HEAP32[$36>>2] = $35;
 $37 = ((($32)) + 76|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = (($38) + -1)|0;
 $40 = ((($32)) + 68|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = (($41) + ($39<<1)|0);
 HEAP16[$42>>1] = 0;
 $43 = $38 << 1;
 $44 = (($43) + -2)|0;
 _memset(($41|0),0,($44|0))|0;
 $45 = ((($32)) + 132|0);
 $46 = HEAP32[$45>>2]|0;
 $47 = (((141128 + (($46*12)|0)|0)) + 2|0);
 $48 = HEAP16[$47>>1]|0;
 $49 = $48&65535;
 $50 = ((($32)) + 128|0);
 HEAP32[$50>>2] = $49;
 $51 = (141128 + (($46*12)|0)|0);
 $52 = HEAP16[$51>>1]|0;
 $53 = $52&65535;
 $54 = ((($32)) + 140|0);
 HEAP32[$54>>2] = $53;
 $55 = (((141128 + (($46*12)|0)|0)) + 4|0);
 $56 = HEAP16[$55>>1]|0;
 $57 = $56&65535;
 $58 = ((($32)) + 144|0);
 HEAP32[$58>>2] = $57;
 $59 = (((141128 + (($46*12)|0)|0)) + 6|0);
 $60 = HEAP16[$59>>1]|0;
 $61 = $60&65535;
 $62 = ((($32)) + 124|0);
 HEAP32[$62>>2] = $61;
 $63 = ((($32)) + 108|0);
 HEAP32[$63>>2] = 0;
 $64 = ((($32)) + 92|0);
 HEAP32[$64>>2] = 0;
 $65 = ((($32)) + 116|0);
 HEAP32[$65>>2] = 0;
 $66 = ((($32)) + 5812|0);
 HEAP32[$66>>2] = 0;
 $67 = ((($32)) + 120|0);
 HEAP32[$67>>2] = 2;
 $68 = ((($32)) + 96|0);
 HEAP32[$68>>2] = 2;
 $69 = ((($32)) + 104|0);
 HEAP32[$69>>2] = 0;
 $70 = ((($32)) + 72|0);
 HEAP32[$70>>2] = 0;
 $$0$i1 = 0;
 return ($$0$i1|0);
}
function _deflate_stored($s,$flush) {
 $s = $s|0;
 $flush = $flush|0;
 var $$ = 0, $$$i = 0, $$$i3 = 0, $$$i5 = 0, $$$i7 = 0, $$0 = 0, $$2 = 0, $$pr = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0, $181 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($s)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + -5)|0;
 $3 = ($2>>>0)<(65535);
 $$ = $3 ? $2 : 65535;
 $4 = ((($s)) + 116|0);
 $5 = ((($s)) + 108|0);
 $6 = ((($s)) + 92|0);
 $7 = ((($s)) + 44|0);
 $8 = ((($s)) + 56|0);
 while(1) {
  $9 = HEAP32[$4>>2]|0;
  $10 = ($9>>>0)<(2);
  if ($10) {
   _fill_window($s);
   $11 = HEAP32[$4>>2]|0;
   $12 = $11 | $flush;
   $13 = ($12|0)==(0);
   if ($13) {
    $$0 = 0;
    label = 35;
    break;
   }
   $14 = ($11|0)==(0);
   if ($14) {
    label = 20;
    break;
   } else {
    $17 = $11;
   }
  } else {
   $17 = $9;
  }
  $15 = HEAP32[$5>>2]|0;
  $16 = (($15) + ($17))|0;
  HEAP32[$5>>2] = $16;
  HEAP32[$4>>2] = 0;
  $18 = HEAP32[$6>>2]|0;
  $19 = (($18) + ($$))|0;
  $20 = ($16|0)!=(0);
  $21 = ($16>>>0)<($19>>>0);
  $or$cond = $20 & $21;
  if ($or$cond) {
   $60 = $16;$61 = $18;
  } else {
   $22 = (($16) - ($19))|0;
   HEAP32[$4>>2] = $22;
   HEAP32[$5>>2] = $19;
   $23 = ($18|0)>(-1);
   if ($23) {
    $24 = HEAP32[$8>>2]|0;
    $25 = (($24) + ($18)|0);
    $26 = $25;
   } else {
    $26 = 0;
   }
   __tr_flush_block($s,$26,$$,0);
   $27 = HEAP32[$5>>2]|0;
   HEAP32[$6>>2] = $27;
   $28 = HEAP32[$s>>2]|0;
   $29 = ((($28)) + 28|0);
   $30 = HEAP32[$29>>2]|0;
   __tr_flush_bits($30);
   $31 = ((($30)) + 20|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = ((($28)) + 16|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = ($32>>>0)>($34>>>0);
   $$$i = $35 ? $34 : $32;
   $36 = ($$$i|0)==(0);
   if (!($36)) {
    $37 = ((($28)) + 12|0);
    $38 = HEAP32[$37>>2]|0;
    $39 = ((($30)) + 16|0);
    $40 = HEAP32[$39>>2]|0;
    _memcpy(($38|0),($40|0),($$$i|0))|0;
    $41 = HEAP32[$37>>2]|0;
    $42 = (($41) + ($$$i)|0);
    HEAP32[$37>>2] = $42;
    $43 = HEAP32[$39>>2]|0;
    $44 = (($43) + ($$$i)|0);
    HEAP32[$39>>2] = $44;
    $45 = ((($28)) + 20|0);
    $46 = HEAP32[$45>>2]|0;
    $47 = (($46) + ($$$i))|0;
    HEAP32[$45>>2] = $47;
    $48 = HEAP32[$33>>2]|0;
    $49 = (($48) - ($$$i))|0;
    HEAP32[$33>>2] = $49;
    $50 = HEAP32[$31>>2]|0;
    $51 = (($50) - ($$$i))|0;
    HEAP32[$31>>2] = $51;
    $52 = ($50|0)==($$$i|0);
    if ($52) {
     $53 = ((($30)) + 8|0);
     $54 = HEAP32[$53>>2]|0;
     HEAP32[$39>>2] = $54;
    }
   }
   $55 = HEAP32[$s>>2]|0;
   $56 = ((($55)) + 16|0);
   $57 = HEAP32[$56>>2]|0;
   $58 = ($57|0)==(0);
   if ($58) {
    $$0 = 0;
    label = 35;
    break;
   }
   $$pr = HEAP32[$6>>2]|0;
   $$pre = HEAP32[$5>>2]|0;
   $60 = $$pre;$61 = $$pr;
  }
  $59 = (($60) - ($61))|0;
  $62 = HEAP32[$7>>2]|0;
  $63 = (($62) + -262)|0;
  $64 = ($59>>>0)<($63>>>0);
  if ($64) {
   continue;
  }
  $65 = ($61|0)>(-1);
  if ($65) {
   $66 = HEAP32[$8>>2]|0;
   $67 = (($66) + ($61)|0);
   $68 = $67;
  } else {
   $68 = 0;
  }
  __tr_flush_block($s,$68,$59,0);
  $69 = HEAP32[$5>>2]|0;
  HEAP32[$6>>2] = $69;
  $70 = HEAP32[$s>>2]|0;
  $71 = ((($70)) + 28|0);
  $72 = HEAP32[$71>>2]|0;
  __tr_flush_bits($72);
  $73 = ((($72)) + 20|0);
  $74 = HEAP32[$73>>2]|0;
  $75 = ((($70)) + 16|0);
  $76 = HEAP32[$75>>2]|0;
  $77 = ($74>>>0)>($76>>>0);
  $$$i3 = $77 ? $76 : $74;
  $78 = ($$$i3|0)==(0);
  if (!($78)) {
   $79 = ((($70)) + 12|0);
   $80 = HEAP32[$79>>2]|0;
   $81 = ((($72)) + 16|0);
   $82 = HEAP32[$81>>2]|0;
   _memcpy(($80|0),($82|0),($$$i3|0))|0;
   $83 = HEAP32[$79>>2]|0;
   $84 = (($83) + ($$$i3)|0);
   HEAP32[$79>>2] = $84;
   $85 = HEAP32[$81>>2]|0;
   $86 = (($85) + ($$$i3)|0);
   HEAP32[$81>>2] = $86;
   $87 = ((($70)) + 20|0);
   $88 = HEAP32[$87>>2]|0;
   $89 = (($88) + ($$$i3))|0;
   HEAP32[$87>>2] = $89;
   $90 = HEAP32[$75>>2]|0;
   $91 = (($90) - ($$$i3))|0;
   HEAP32[$75>>2] = $91;
   $92 = HEAP32[$73>>2]|0;
   $93 = (($92) - ($$$i3))|0;
   HEAP32[$73>>2] = $93;
   $94 = ($92|0)==($$$i3|0);
   if ($94) {
    $95 = ((($72)) + 8|0);
    $96 = HEAP32[$95>>2]|0;
    HEAP32[$81>>2] = $96;
   }
  }
  $97 = HEAP32[$s>>2]|0;
  $98 = ((($97)) + 16|0);
  $99 = HEAP32[$98>>2]|0;
  $100 = ($99|0)==(0);
  if ($100) {
   $$0 = 0;
   label = 35;
   break;
  }
 }
 if ((label|0) == 20) {
  $101 = ((($s)) + 5812|0);
  HEAP32[$101>>2] = 0;
  $102 = ($flush|0)==(4);
  if ($102) {
   $103 = HEAP32[$6>>2]|0;
   $104 = ($103|0)>(-1);
   if ($104) {
    $105 = HEAP32[$8>>2]|0;
    $106 = (($105) + ($103)|0);
    $109 = $106;
   } else {
    $109 = 0;
   }
   $107 = HEAP32[$5>>2]|0;
   $108 = (($107) - ($103))|0;
   __tr_flush_block($s,$109,$108,1);
   $110 = HEAP32[$5>>2]|0;
   HEAP32[$6>>2] = $110;
   $111 = HEAP32[$s>>2]|0;
   $112 = ((($111)) + 28|0);
   $113 = HEAP32[$112>>2]|0;
   __tr_flush_bits($113);
   $114 = ((($113)) + 20|0);
   $115 = HEAP32[$114>>2]|0;
   $116 = ((($111)) + 16|0);
   $117 = HEAP32[$116>>2]|0;
   $118 = ($115>>>0)>($117>>>0);
   $$$i5 = $118 ? $117 : $115;
   $119 = ($$$i5|0)==(0);
   if (!($119)) {
    $120 = ((($111)) + 12|0);
    $121 = HEAP32[$120>>2]|0;
    $122 = ((($113)) + 16|0);
    $123 = HEAP32[$122>>2]|0;
    _memcpy(($121|0),($123|0),($$$i5|0))|0;
    $124 = HEAP32[$120>>2]|0;
    $125 = (($124) + ($$$i5)|0);
    HEAP32[$120>>2] = $125;
    $126 = HEAP32[$122>>2]|0;
    $127 = (($126) + ($$$i5)|0);
    HEAP32[$122>>2] = $127;
    $128 = ((($111)) + 20|0);
    $129 = HEAP32[$128>>2]|0;
    $130 = (($129) + ($$$i5))|0;
    HEAP32[$128>>2] = $130;
    $131 = HEAP32[$116>>2]|0;
    $132 = (($131) - ($$$i5))|0;
    HEAP32[$116>>2] = $132;
    $133 = HEAP32[$114>>2]|0;
    $134 = (($133) - ($$$i5))|0;
    HEAP32[$114>>2] = $134;
    $135 = ($133|0)==($$$i5|0);
    if ($135) {
     $136 = ((($113)) + 8|0);
     $137 = HEAP32[$136>>2]|0;
     HEAP32[$122>>2] = $137;
    }
   }
   $138 = HEAP32[$s>>2]|0;
   $139 = ((($138)) + 16|0);
   $140 = HEAP32[$139>>2]|0;
   $141 = ($140|0)==(0);
   $$2 = $141 ? 2 : 3;
   $$0 = $$2;
   return ($$0|0);
  }
  $142 = HEAP32[$5>>2]|0;
  $143 = HEAP32[$6>>2]|0;
  $144 = ($142|0)>($143|0);
  if ($144) {
   $145 = ($143|0)>(-1);
   if ($145) {
    $146 = HEAP32[$8>>2]|0;
    $147 = (($146) + ($143)|0);
    $149 = $147;
   } else {
    $149 = 0;
   }
   $148 = (($142) - ($143))|0;
   __tr_flush_block($s,$149,$148,0);
   $150 = HEAP32[$5>>2]|0;
   HEAP32[$6>>2] = $150;
   $151 = HEAP32[$s>>2]|0;
   $152 = ((($151)) + 28|0);
   $153 = HEAP32[$152>>2]|0;
   __tr_flush_bits($153);
   $154 = ((($153)) + 20|0);
   $155 = HEAP32[$154>>2]|0;
   $156 = ((($151)) + 16|0);
   $157 = HEAP32[$156>>2]|0;
   $158 = ($155>>>0)>($157>>>0);
   $$$i7 = $158 ? $157 : $155;
   $159 = ($$$i7|0)==(0);
   if (!($159)) {
    $160 = ((($151)) + 12|0);
    $161 = HEAP32[$160>>2]|0;
    $162 = ((($153)) + 16|0);
    $163 = HEAP32[$162>>2]|0;
    _memcpy(($161|0),($163|0),($$$i7|0))|0;
    $164 = HEAP32[$160>>2]|0;
    $165 = (($164) + ($$$i7)|0);
    HEAP32[$160>>2] = $165;
    $166 = HEAP32[$162>>2]|0;
    $167 = (($166) + ($$$i7)|0);
    HEAP32[$162>>2] = $167;
    $168 = ((($151)) + 20|0);
    $169 = HEAP32[$168>>2]|0;
    $170 = (($169) + ($$$i7))|0;
    HEAP32[$168>>2] = $170;
    $171 = HEAP32[$156>>2]|0;
    $172 = (($171) - ($$$i7))|0;
    HEAP32[$156>>2] = $172;
    $173 = HEAP32[$154>>2]|0;
    $174 = (($173) - ($$$i7))|0;
    HEAP32[$154>>2] = $174;
    $175 = ($173|0)==($$$i7|0);
    if ($175) {
     $176 = ((($153)) + 8|0);
     $177 = HEAP32[$176>>2]|0;
     HEAP32[$162>>2] = $177;
    }
   }
   $178 = HEAP32[$s>>2]|0;
   $179 = ((($178)) + 16|0);
   $180 = HEAP32[$179>>2]|0;
   $181 = ($180|0)==(0);
   if ($181) {
    $$0 = 0;
    return ($$0|0);
   }
  }
  $$0 = 1;
  return ($$0|0);
 }
 else if ((label|0) == 35) {
  return ($$0|0);
 }
 return (0)|0;
}
function _fill_window($s) {
 $s = $s|0;
 var $$ = 0, $$0$i = 0, $$3 = 0, $$pre = 0, $$pre8 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $more$0 = 0, $n$0 = 0, $n$1 = 0, $p$0 = 0, $p$1 = 0, $size$$i = 0, $str$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($s)) + 44|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($s)) + 60|0);
 $3 = ((($s)) + 116|0);
 $4 = ((($s)) + 108|0);
 $5 = (($1) + -262)|0;
 $6 = ((($s)) + 56|0);
 $7 = ((($s)) + 5812|0);
 $8 = ((($s)) + 72|0);
 $9 = ((($s)) + 88|0);
 $10 = ((($s)) + 84|0);
 $11 = ((($s)) + 68|0);
 $12 = ((($s)) + 52|0);
 $13 = ((($s)) + 64|0);
 $14 = ((($s)) + 112|0);
 $15 = ((($s)) + 92|0);
 $16 = ((($s)) + 76|0);
 $$pre = HEAP32[$3>>2]|0;
 $19 = $$pre;$23 = $1;
 while(1) {
  $17 = HEAP32[$2>>2]|0;
  $18 = (($17) - ($19))|0;
  $20 = HEAP32[$4>>2]|0;
  $21 = (($18) - ($20))|0;
  $22 = (($5) + ($23))|0;
  $24 = ($20>>>0)<($22>>>0);
  if ($24) {
   $63 = $20;$more$0 = $21;
  } else {
   $25 = HEAP32[$6>>2]|0;
   $26 = (($25) + ($1)|0);
   _memcpy(($25|0),($26|0),($1|0))|0;
   $27 = HEAP32[$14>>2]|0;
   $28 = (($27) - ($1))|0;
   HEAP32[$14>>2] = $28;
   $29 = HEAP32[$4>>2]|0;
   $30 = (($29) - ($1))|0;
   HEAP32[$4>>2] = $30;
   $31 = HEAP32[$15>>2]|0;
   $32 = (($31) - ($1))|0;
   HEAP32[$15>>2] = $32;
   $33 = HEAP32[$16>>2]|0;
   $34 = HEAP32[$11>>2]|0;
   $35 = (($34) + ($33<<1)|0);
   $n$0 = $33;$p$0 = $35;
   while(1) {
    $36 = ((($p$0)) + -2|0);
    $37 = HEAP16[$36>>1]|0;
    $38 = $37&65535;
    $39 = ($38>>>0)<($1>>>0);
    $40 = (($38) - ($1))|0;
    $41 = $40&65535;
    $42 = $39 ? 0 : $41;
    HEAP16[$36>>1] = $42;
    $43 = (($n$0) + -1)|0;
    $44 = ($43|0)==(0);
    if ($44) {
     break;
    } else {
     $n$0 = $43;$p$0 = $36;
    }
   }
   $45 = HEAP32[$13>>2]|0;
   $46 = (($45) + ($1<<1)|0);
   $n$1 = $1;$p$1 = $46;
   while(1) {
    $47 = ((($p$1)) + -2|0);
    $48 = HEAP16[$47>>1]|0;
    $49 = $48&65535;
    $50 = ($49>>>0)<($1>>>0);
    $51 = (($49) - ($1))|0;
    $52 = $51&65535;
    $53 = $50 ? 0 : $52;
    HEAP16[$47>>1] = $53;
    $54 = (($n$1) + -1)|0;
    $55 = ($54|0)==(0);
    if ($55) {
     break;
    } else {
     $n$1 = $54;$p$1 = $47;
    }
   }
   $56 = (($21) + ($1))|0;
   $63 = $30;$more$0 = $56;
  }
  $57 = HEAP32[$s>>2]|0;
  $58 = ((($57)) + 4|0);
  $59 = HEAP32[$58>>2]|0;
  $60 = ($59|0)==(0);
  if ($60) {
   break;
  }
  $61 = HEAP32[$6>>2]|0;
  $62 = (($61) + ($63)|0);
  $64 = HEAP32[$3>>2]|0;
  $65 = (($62) + ($64)|0);
  $66 = ($59>>>0)>($more$0>>>0);
  $size$$i = $66 ? $more$0 : $59;
  $67 = ($size$$i|0)==(0);
  if ($67) {
   $$0$i = 0;$86 = $64;
  } else {
   $68 = (($59) - ($size$$i))|0;
   HEAP32[$58>>2] = $68;
   $69 = HEAP32[$57>>2]|0;
   _memcpy(($65|0),($69|0),($size$$i|0))|0;
   $70 = ((($57)) + 28|0);
   $71 = HEAP32[$70>>2]|0;
   $72 = ((($71)) + 24|0);
   $73 = HEAP32[$72>>2]|0;
   switch ($73|0) {
   case 1:  {
    $74 = ((($57)) + 48|0);
    $75 = HEAP32[$74>>2]|0;
    $76 = (_adler32($75,$65,$size$$i)|0);
    HEAP32[$74>>2] = $76;
    break;
   }
   case 2:  {
    $77 = ((($57)) + 48|0);
    $78 = HEAP32[$77>>2]|0;
    $79 = (_crc32($78,$65,$size$$i)|0);
    HEAP32[$77>>2] = $79;
    break;
   }
   default: {
   }
   }
   $80 = HEAP32[$57>>2]|0;
   $81 = (($80) + ($size$$i)|0);
   HEAP32[$57>>2] = $81;
   $82 = ((($57)) + 8|0);
   $83 = HEAP32[$82>>2]|0;
   $84 = (($83) + ($size$$i))|0;
   HEAP32[$82>>2] = $84;
   $$pre9 = HEAP32[$3>>2]|0;
   $$0$i = $size$$i;$86 = $$pre9;
  }
  $85 = (($86) + ($$0$i))|0;
  HEAP32[$3>>2] = $85;
  $87 = HEAP32[$7>>2]|0;
  $88 = (($87) + ($85))|0;
  $89 = ($88>>>0)>(2);
  L20: do {
   if ($89) {
    $90 = HEAP32[$4>>2]|0;
    $91 = (($90) - ($87))|0;
    $92 = HEAP32[$6>>2]|0;
    $93 = (($92) + ($91)|0);
    $94 = HEAP8[$93>>0]|0;
    $95 = $94&255;
    HEAP32[$8>>2] = $95;
    $96 = HEAP32[$9>>2]|0;
    $97 = $95 << $96;
    $98 = (($91) + 1)|0;
    $99 = (($92) + ($98)|0);
    $100 = HEAP8[$99>>0]|0;
    $101 = $100&255;
    $102 = $101 ^ $97;
    $103 = HEAP32[$10>>2]|0;
    $104 = $102 & $103;
    HEAP32[$8>>2] = $104;
    $106 = $87;$108 = $104;$str$0 = $91;
    while(1) {
     $105 = ($106|0)==(0);
     if ($105) {
      break L20;
     }
     $107 = $108 << $96;
     $109 = (($str$0) + 2)|0;
     $110 = (($92) + ($109)|0);
     $111 = HEAP8[$110>>0]|0;
     $112 = $111&255;
     $113 = $112 ^ $107;
     $114 = $113 & $103;
     HEAP32[$8>>2] = $114;
     $115 = HEAP32[$11>>2]|0;
     $116 = (($115) + ($114<<1)|0);
     $117 = HEAP16[$116>>1]|0;
     $118 = HEAP32[$12>>2]|0;
     $119 = $118 & $str$0;
     $120 = HEAP32[$13>>2]|0;
     $121 = (($120) + ($119<<1)|0);
     HEAP16[$121>>1] = $117;
     $122 = $str$0&65535;
     HEAP16[$116>>1] = $122;
     $123 = (($str$0) + 1)|0;
     $124 = (($106) + -1)|0;
     HEAP32[$7>>2] = $124;
     $125 = (($85) + ($124))|0;
     $126 = ($125>>>0)<(3);
     if ($126) {
      break;
     } else {
      $106 = $124;$108 = $114;$str$0 = $123;
     }
    }
   }
  } while(0);
  $127 = ($85>>>0)<(262);
  if (!($127)) {
   break;
  }
  $128 = HEAP32[$s>>2]|0;
  $129 = ((($128)) + 4|0);
  $130 = HEAP32[$129>>2]|0;
  $131 = ($130|0)==(0);
  if ($131) {
   break;
  }
  $$pre8 = HEAP32[$0>>2]|0;
  $19 = $85;$23 = $$pre8;
 }
 $132 = ((($s)) + 5824|0);
 $133 = HEAP32[$132>>2]|0;
 $134 = HEAP32[$2>>2]|0;
 $135 = ($134>>>0)>($133>>>0);
 if (!($135)) {
  return;
 }
 $136 = HEAP32[$4>>2]|0;
 $137 = HEAP32[$3>>2]|0;
 $138 = (($137) + ($136))|0;
 $139 = ($133>>>0)<($138>>>0);
 if ($139) {
  $140 = (($134) - ($138))|0;
  $141 = ($140>>>0)>(258);
  $$ = $141 ? 258 : $140;
  $142 = HEAP32[$6>>2]|0;
  $143 = (($142) + ($138)|0);
  _memset(($143|0),0,($$|0))|0;
  $144 = (($$) + ($138))|0;
  HEAP32[$132>>2] = $144;
  return;
 }
 $145 = (($138) + 258)|0;
 $146 = ($145>>>0)>($133>>>0);
 if (!($146)) {
  return;
 }
 $147 = (($145) - ($133))|0;
 $148 = (($134) - ($133))|0;
 $149 = ($147>>>0)>($148>>>0);
 $$3 = $149 ? $148 : $147;
 $150 = HEAP32[$6>>2]|0;
 $151 = (($150) + ($133)|0);
 _memset(($151|0),0,($$3|0))|0;
 $152 = HEAP32[$132>>2]|0;
 $153 = (($152) + ($$3))|0;
 HEAP32[$132>>2] = $153;
 return;
}
function _deflate_fast($s,$flush) {
 $s = $s|0;
 $flush = $flush|0;
 var $$ = 0, $$$i = 0, $$$i4 = 0, $$$i6 = 0, $$0 = 0, $$3 = 0, $$lcssa = 0, $$pn = 0, $$pr = 0, $$pre = 0, $$pre10 = 0, $$pre11 = 0, $$pre12 = 0, $$pre13 = 0, $$pre14 = 0, $$pre15 = 0, $$pre9 = 0, $$sink = 0, $$sink$in = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0;
 var $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bflush$0 = 0, $or$cond = 0, $or$cond2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($s)) + 116|0);
 $1 = ($flush|0)==(0);
 $2 = ((($s)) + 72|0);
 $3 = ((($s)) + 88|0);
 $4 = ((($s)) + 108|0);
 $5 = ((($s)) + 56|0);
 $6 = ((($s)) + 84|0);
 $7 = ((($s)) + 68|0);
 $8 = ((($s)) + 52|0);
 $9 = ((($s)) + 64|0);
 $10 = ((($s)) + 44|0);
 $11 = ((($s)) + 96|0);
 $12 = ((($s)) + 112|0);
 $13 = ((($s)) + 5792|0);
 $14 = ((($s)) + 5796|0);
 $15 = ((($s)) + 5784|0);
 $16 = ((($s)) + 5788|0);
 $17 = ((($s)) + 128|0);
 $18 = ((($s)) + 92|0);
 while(1) {
  $19 = HEAP32[$0>>2]|0;
  $20 = ($19>>>0)<(262);
  if ($20) {
   _fill_window($s);
   $21 = HEAP32[$0>>2]|0;
   $22 = ($21>>>0)<(262);
   $or$cond = $1 & $22;
   if ($or$cond) {
    $$0 = 0;
    label = 39;
    break;
   }
   $23 = ($21|0)==(0);
   if ($23) {
    label = 24;
    break;
   }
   $24 = ($21>>>0)>(2);
   if ($24) {
    label = 6;
   } else {
    label = 9;
   }
  } else {
   label = 6;
  }
  if ((label|0) == 6) {
   label = 0;
   $25 = HEAP32[$2>>2]|0;
   $26 = HEAP32[$3>>2]|0;
   $27 = $25 << $26;
   $28 = HEAP32[$4>>2]|0;
   $29 = (($28) + 2)|0;
   $30 = HEAP32[$5>>2]|0;
   $31 = (($30) + ($29)|0);
   $32 = HEAP8[$31>>0]|0;
   $33 = $32&255;
   $34 = $33 ^ $27;
   $35 = HEAP32[$6>>2]|0;
   $36 = $34 & $35;
   HEAP32[$2>>2] = $36;
   $37 = HEAP32[$7>>2]|0;
   $38 = (($37) + ($36<<1)|0);
   $39 = HEAP16[$38>>1]|0;
   $40 = HEAP32[$8>>2]|0;
   $41 = $40 & $28;
   $42 = HEAP32[$9>>2]|0;
   $43 = (($42) + ($41<<1)|0);
   HEAP16[$43>>1] = $39;
   $44 = $39&65535;
   $45 = $28&65535;
   HEAP16[$38>>1] = $45;
   $46 = ($39<<16>>16)==(0);
   if ($46) {
    label = 9;
   } else {
    $47 = (($28) - ($44))|0;
    $48 = HEAP32[$10>>2]|0;
    $49 = (($48) + -262)|0;
    $50 = ($47>>>0)>($49>>>0);
    if ($50) {
     label = 9;
    } else {
     $51 = (_longest_match($s,$44)|0);
     HEAP32[$11>>2] = $51;
     $53 = $51;
    }
   }
  }
  if ((label|0) == 9) {
   label = 0;
   $$pr = HEAP32[$11>>2]|0;
   $53 = $$pr;
  }
  $52 = ($53>>>0)>(2);
  do {
   if ($52) {
    $54 = (($53) + 253)|0;
    $55 = $54&255;
    $56 = HEAP32[$4>>2]|0;
    $57 = HEAP32[$12>>2]|0;
    $58 = (($56) - ($57))|0;
    $59 = $58&65535;
    $60 = HEAP32[$13>>2]|0;
    $61 = HEAP32[$14>>2]|0;
    $62 = (($61) + ($60<<1)|0);
    HEAP16[$62>>1] = $59;
    $63 = (($60) + 1)|0;
    HEAP32[$13>>2] = $63;
    $64 = HEAP32[$15>>2]|0;
    $65 = (($64) + ($60)|0);
    HEAP8[$65>>0] = $55;
    $66 = (($58) + 65535)|0;
    $67 = $54 & 255;
    $68 = (613302 + ($67)|0);
    $69 = HEAP8[$68>>0]|0;
    $70 = $69&255;
    $71 = $70 | 256;
    $72 = (($71) + 1)|0;
    $73 = (((($s)) + 148|0) + ($72<<2)|0);
    $74 = HEAP16[$73>>1]|0;
    $75 = (($74) + 1)<<16>>16;
    HEAP16[$73>>1] = $75;
    $76 = $66 & 65535;
    $77 = ($76>>>0)<(256);
    $78 = $76 >>> 7;
    $79 = (($78) + 256)|0;
    $$pn = $77 ? $76 : $79;
    $$sink$in = (612790 + ($$pn)|0);
    $$sink = HEAP8[$$sink$in>>0]|0;
    $80 = $$sink&255;
    $81 = (((($s)) + 2440|0) + ($80<<2)|0);
    $82 = HEAP16[$81>>1]|0;
    $83 = (($82) + 1)<<16>>16;
    HEAP16[$81>>1] = $83;
    $84 = HEAP32[$13>>2]|0;
    $85 = HEAP32[$16>>2]|0;
    $86 = (($85) + -1)|0;
    $87 = ($84|0)==($86|0);
    $88 = $87&1;
    $89 = HEAP32[$11>>2]|0;
    $90 = HEAP32[$0>>2]|0;
    $91 = (($90) - ($89))|0;
    HEAP32[$0>>2] = $91;
    $92 = HEAP32[$17>>2]|0;
    $93 = ($89>>>0)<=($92>>>0);
    $94 = ($91>>>0)>(2);
    $or$cond2 = $94 & $93;
    if (!($or$cond2)) {
     $115 = HEAP32[$4>>2]|0;
     $116 = (($115) + ($89))|0;
     HEAP32[$4>>2] = $116;
     HEAP32[$11>>2] = 0;
     $117 = HEAP32[$5>>2]|0;
     $118 = (($117) + ($116)|0);
     $119 = HEAP8[$118>>0]|0;
     $120 = $119&255;
     HEAP32[$2>>2] = $120;
     $121 = HEAP32[$3>>2]|0;
     $122 = $120 << $121;
     $123 = (($116) + 1)|0;
     $124 = (($117) + ($123)|0);
     $125 = HEAP8[$124>>0]|0;
     $126 = $125&255;
     $127 = $126 ^ $122;
     $128 = HEAP32[$6>>2]|0;
     $129 = $127 & $128;
     HEAP32[$2>>2] = $129;
     $159 = $116;$bflush$0 = $88;
     break;
    }
    $95 = (($89) + -1)|0;
    HEAP32[$11>>2] = $95;
    $$pre = HEAP32[$4>>2]|0;
    $$pre9 = HEAP32[$2>>2]|0;
    $$pre10 = HEAP32[$3>>2]|0;
    $$pre11 = HEAP32[$5>>2]|0;
    $$pre12 = HEAP32[$6>>2]|0;
    $$pre13 = HEAP32[$7>>2]|0;
    $$pre14 = HEAP32[$8>>2]|0;
    $$pre15 = HEAP32[$9>>2]|0;
    $112 = $95;$97 = $$pre;$99 = $$pre9;
    while(1) {
     $96 = (($97) + 1)|0;
     HEAP32[$4>>2] = $96;
     $98 = $99 << $$pre10;
     $100 = (($97) + 3)|0;
     $101 = (($$pre11) + ($100)|0);
     $102 = HEAP8[$101>>0]|0;
     $103 = $102&255;
     $104 = $103 ^ $98;
     $105 = $104 & $$pre12;
     HEAP32[$2>>2] = $105;
     $106 = (($$pre13) + ($105<<1)|0);
     $107 = HEAP16[$106>>1]|0;
     $108 = $$pre14 & $96;
     $109 = (($$pre15) + ($108<<1)|0);
     HEAP16[$109>>1] = $107;
     $110 = $96&65535;
     HEAP16[$106>>1] = $110;
     $111 = (($112) + -1)|0;
     HEAP32[$11>>2] = $111;
     $113 = ($111|0)==(0);
     if ($113) {
      $$lcssa = $97;
      break;
     } else {
      $112 = $111;$97 = $96;$99 = $105;
     }
    }
    $114 = (($$lcssa) + 2)|0;
    HEAP32[$4>>2] = $114;
    $159 = $114;$bflush$0 = $88;
   } else {
    $130 = HEAP32[$4>>2]|0;
    $131 = HEAP32[$5>>2]|0;
    $132 = (($131) + ($130)|0);
    $133 = HEAP8[$132>>0]|0;
    $134 = HEAP32[$13>>2]|0;
    $135 = HEAP32[$14>>2]|0;
    $136 = (($135) + ($134<<1)|0);
    HEAP16[$136>>1] = 0;
    $137 = (($134) + 1)|0;
    HEAP32[$13>>2] = $137;
    $138 = HEAP32[$15>>2]|0;
    $139 = (($138) + ($134)|0);
    HEAP8[$139>>0] = $133;
    $140 = $133&255;
    $141 = (((($s)) + 148|0) + ($140<<2)|0);
    $142 = HEAP16[$141>>1]|0;
    $143 = (($142) + 1)<<16>>16;
    HEAP16[$141>>1] = $143;
    $144 = HEAP32[$13>>2]|0;
    $145 = HEAP32[$16>>2]|0;
    $146 = (($145) + -1)|0;
    $147 = ($144|0)==($146|0);
    $148 = $147&1;
    $149 = HEAP32[$0>>2]|0;
    $150 = (($149) + -1)|0;
    HEAP32[$0>>2] = $150;
    $151 = HEAP32[$4>>2]|0;
    $152 = (($151) + 1)|0;
    HEAP32[$4>>2] = $152;
    $159 = $152;$bflush$0 = $148;
   }
  } while(0);
  $153 = ($bflush$0|0)==(0);
  if ($153) {
   continue;
  }
  $154 = HEAP32[$18>>2]|0;
  $155 = ($154|0)>(-1);
  if ($155) {
   $156 = HEAP32[$5>>2]|0;
   $157 = (($156) + ($154)|0);
   $160 = $157;
  } else {
   $160 = 0;
  }
  $158 = (($159) - ($154))|0;
  __tr_flush_block($s,$160,$158,0);
  $161 = HEAP32[$4>>2]|0;
  HEAP32[$18>>2] = $161;
  $162 = HEAP32[$s>>2]|0;
  $163 = ((($162)) + 28|0);
  $164 = HEAP32[$163>>2]|0;
  __tr_flush_bits($164);
  $165 = ((($164)) + 20|0);
  $166 = HEAP32[$165>>2]|0;
  $167 = ((($162)) + 16|0);
  $168 = HEAP32[$167>>2]|0;
  $169 = ($166>>>0)>($168>>>0);
  $$$i = $169 ? $168 : $166;
  $170 = ($$$i|0)==(0);
  if (!($170)) {
   $171 = ((($162)) + 12|0);
   $172 = HEAP32[$171>>2]|0;
   $173 = ((($164)) + 16|0);
   $174 = HEAP32[$173>>2]|0;
   _memcpy(($172|0),($174|0),($$$i|0))|0;
   $175 = HEAP32[$171>>2]|0;
   $176 = (($175) + ($$$i)|0);
   HEAP32[$171>>2] = $176;
   $177 = HEAP32[$173>>2]|0;
   $178 = (($177) + ($$$i)|0);
   HEAP32[$173>>2] = $178;
   $179 = ((($162)) + 20|0);
   $180 = HEAP32[$179>>2]|0;
   $181 = (($180) + ($$$i))|0;
   HEAP32[$179>>2] = $181;
   $182 = HEAP32[$167>>2]|0;
   $183 = (($182) - ($$$i))|0;
   HEAP32[$167>>2] = $183;
   $184 = HEAP32[$165>>2]|0;
   $185 = (($184) - ($$$i))|0;
   HEAP32[$165>>2] = $185;
   $186 = ($184|0)==($$$i|0);
   if ($186) {
    $187 = ((($164)) + 8|0);
    $188 = HEAP32[$187>>2]|0;
    HEAP32[$173>>2] = $188;
   }
  }
  $189 = HEAP32[$s>>2]|0;
  $190 = ((($189)) + 16|0);
  $191 = HEAP32[$190>>2]|0;
  $192 = ($191|0)==(0);
  if ($192) {
   $$0 = 0;
   label = 39;
   break;
  }
 }
 if ((label|0) == 24) {
  $193 = HEAP32[$4>>2]|0;
  $194 = ($193>>>0)<(2);
  $$ = $194 ? $193 : 2;
  $195 = ((($s)) + 5812|0);
  HEAP32[$195>>2] = $$;
  $196 = ($flush|0)==(4);
  if ($196) {
   $197 = HEAP32[$18>>2]|0;
   $198 = ($197|0)>(-1);
   if ($198) {
    $199 = HEAP32[$5>>2]|0;
    $200 = (($199) + ($197)|0);
    $202 = $200;
   } else {
    $202 = 0;
   }
   $201 = (($193) - ($197))|0;
   __tr_flush_block($s,$202,$201,1);
   $203 = HEAP32[$4>>2]|0;
   HEAP32[$18>>2] = $203;
   $204 = HEAP32[$s>>2]|0;
   $205 = ((($204)) + 28|0);
   $206 = HEAP32[$205>>2]|0;
   __tr_flush_bits($206);
   $207 = ((($206)) + 20|0);
   $208 = HEAP32[$207>>2]|0;
   $209 = ((($204)) + 16|0);
   $210 = HEAP32[$209>>2]|0;
   $211 = ($208>>>0)>($210>>>0);
   $$$i4 = $211 ? $210 : $208;
   $212 = ($$$i4|0)==(0);
   if (!($212)) {
    $213 = ((($204)) + 12|0);
    $214 = HEAP32[$213>>2]|0;
    $215 = ((($206)) + 16|0);
    $216 = HEAP32[$215>>2]|0;
    _memcpy(($214|0),($216|0),($$$i4|0))|0;
    $217 = HEAP32[$213>>2]|0;
    $218 = (($217) + ($$$i4)|0);
    HEAP32[$213>>2] = $218;
    $219 = HEAP32[$215>>2]|0;
    $220 = (($219) + ($$$i4)|0);
    HEAP32[$215>>2] = $220;
    $221 = ((($204)) + 20|0);
    $222 = HEAP32[$221>>2]|0;
    $223 = (($222) + ($$$i4))|0;
    HEAP32[$221>>2] = $223;
    $224 = HEAP32[$209>>2]|0;
    $225 = (($224) - ($$$i4))|0;
    HEAP32[$209>>2] = $225;
    $226 = HEAP32[$207>>2]|0;
    $227 = (($226) - ($$$i4))|0;
    HEAP32[$207>>2] = $227;
    $228 = ($226|0)==($$$i4|0);
    if ($228) {
     $229 = ((($206)) + 8|0);
     $230 = HEAP32[$229>>2]|0;
     HEAP32[$215>>2] = $230;
    }
   }
   $231 = HEAP32[$s>>2]|0;
   $232 = ((($231)) + 16|0);
   $233 = HEAP32[$232>>2]|0;
   $234 = ($233|0)==(0);
   $$3 = $234 ? 2 : 3;
   $$0 = $$3;
   return ($$0|0);
  }
  $235 = HEAP32[$13>>2]|0;
  $236 = ($235|0)==(0);
  if (!($236)) {
   $237 = HEAP32[$18>>2]|0;
   $238 = ($237|0)>(-1);
   if ($238) {
    $239 = HEAP32[$5>>2]|0;
    $240 = (($239) + ($237)|0);
    $242 = $240;
   } else {
    $242 = 0;
   }
   $241 = (($193) - ($237))|0;
   __tr_flush_block($s,$242,$241,0);
   $243 = HEAP32[$4>>2]|0;
   HEAP32[$18>>2] = $243;
   $244 = HEAP32[$s>>2]|0;
   $245 = ((($244)) + 28|0);
   $246 = HEAP32[$245>>2]|0;
   __tr_flush_bits($246);
   $247 = ((($246)) + 20|0);
   $248 = HEAP32[$247>>2]|0;
   $249 = ((($244)) + 16|0);
   $250 = HEAP32[$249>>2]|0;
   $251 = ($248>>>0)>($250>>>0);
   $$$i6 = $251 ? $250 : $248;
   $252 = ($$$i6|0)==(0);
   if (!($252)) {
    $253 = ((($244)) + 12|0);
    $254 = HEAP32[$253>>2]|0;
    $255 = ((($246)) + 16|0);
    $256 = HEAP32[$255>>2]|0;
    _memcpy(($254|0),($256|0),($$$i6|0))|0;
    $257 = HEAP32[$253>>2]|0;
    $258 = (($257) + ($$$i6)|0);
    HEAP32[$253>>2] = $258;
    $259 = HEAP32[$255>>2]|0;
    $260 = (($259) + ($$$i6)|0);
    HEAP32[$255>>2] = $260;
    $261 = ((($244)) + 20|0);
    $262 = HEAP32[$261>>2]|0;
    $263 = (($262) + ($$$i6))|0;
    HEAP32[$261>>2] = $263;
    $264 = HEAP32[$249>>2]|0;
    $265 = (($264) - ($$$i6))|0;
    HEAP32[$249>>2] = $265;
    $266 = HEAP32[$247>>2]|0;
    $267 = (($266) - ($$$i6))|0;
    HEAP32[$247>>2] = $267;
    $268 = ($266|0)==($$$i6|0);
    if ($268) {
     $269 = ((($246)) + 8|0);
     $270 = HEAP32[$269>>2]|0;
     HEAP32[$255>>2] = $270;
    }
   }
   $271 = HEAP32[$s>>2]|0;
   $272 = ((($271)) + 16|0);
   $273 = HEAP32[$272>>2]|0;
   $274 = ($273|0)==(0);
   if ($274) {
    $$0 = 0;
    return ($$0|0);
   }
  }
  $$0 = 1;
  return ($$0|0);
 }
 else if ((label|0) == 39) {
  return ($$0|0);
 }
 return (0)|0;
}
function _longest_match($s,$cur_match) {
 $s = $s|0;
 $cur_match = $cur_match|0;
 var $$ = 0, $$09 = 0, $$10 = 0, $$best_len$2 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $best_len$0 = 0, $best_len$1 = 0, $best_len$2 = 0;
 var $chain_length$0 = 0, $chain_length$1 = 0, $or$cond = 0, $scan$0 = 0, $scan$1 = 0, $scan$2 = 0, $scan$3 = 0, $scan_end$0 = 0, $scan_end$1 = 0, $scan_end1$0 = 0, $scan_end1$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($s)) + 124|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($s)) + 56|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($s)) + 108|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($3) + ($5)|0);
 $7 = ((($s)) + 120|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($s)) + 144|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($s)) + 44|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($12) + -262)|0;
 $14 = ($5>>>0)>($13>>>0);
 $15 = (($5) - ($13))|0;
 $$ = $14 ? $15 : 0;
 $16 = ((($s)) + 64|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ((($s)) + 52|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ((($6)) + 258|0);
 $21 = (($8) + -1)|0;
 $22 = (($6) + ($21)|0);
 $23 = HEAP8[$22>>0]|0;
 $24 = (($6) + ($8)|0);
 $25 = HEAP8[$24>>0]|0;
 $26 = ((($s)) + 140|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ($8>>>0)<($27>>>0);
 $29 = $1 >>> 2;
 $chain_length$0 = $28 ? $1 : $29;
 $30 = ((($s)) + 116|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = ($10>>>0)>($31>>>0);
 $$10 = $32 ? $31 : $10;
 $33 = ((($s)) + 112|0);
 $34 = $20;
 $35 = ((($20)) + -258|0);
 $$09 = $cur_match;$best_len$0 = $8;$chain_length$1 = $chain_length$0;$scan$0 = $6;$scan_end$0 = $25;$scan_end1$0 = $23;
 while(1) {
  $36 = (($3) + ($$09)|0);
  $37 = (($36) + ($best_len$0)|0);
  $38 = HEAP8[$37>>0]|0;
  $39 = ($38<<24>>24)==($scan_end$0<<24>>24);
  if ($39) {
   $40 = (($best_len$0) + -1)|0;
   $41 = (($36) + ($40)|0);
   $42 = HEAP8[$41>>0]|0;
   $43 = ($42<<24>>24)==($scan_end1$0<<24>>24);
   if ($43) {
    $44 = HEAP8[$36>>0]|0;
    $45 = HEAP8[$scan$0>>0]|0;
    $46 = ($44<<24>>24)==($45<<24>>24);
    if ($46) {
     $47 = ((($36)) + 1|0);
     $48 = HEAP8[$47>>0]|0;
     $49 = ((($scan$0)) + 1|0);
     $50 = HEAP8[$49>>0]|0;
     $51 = ($48<<24>>24)==($50<<24>>24);
     if ($51) {
      $52 = ((($scan$0)) + 2|0);
      $56 = $47;$scan$1 = $52;
      while(1) {
       $53 = ((($scan$1)) + 1|0);
       $54 = HEAP8[$53>>0]|0;
       $55 = ((($56)) + 2|0);
       $57 = HEAP8[$55>>0]|0;
       $58 = ($54<<24>>24)==($57<<24>>24);
       if (!($58)) {
        $scan$2 = $53;
        break;
       }
       $59 = ((($scan$1)) + 2|0);
       $60 = HEAP8[$59>>0]|0;
       $61 = ((($56)) + 3|0);
       $62 = HEAP8[$61>>0]|0;
       $63 = ($60<<24>>24)==($62<<24>>24);
       if (!($63)) {
        $scan$2 = $59;
        break;
       }
       $64 = ((($scan$1)) + 3|0);
       $65 = HEAP8[$64>>0]|0;
       $66 = ((($56)) + 4|0);
       $67 = HEAP8[$66>>0]|0;
       $68 = ($65<<24>>24)==($67<<24>>24);
       if (!($68)) {
        $scan$2 = $64;
        break;
       }
       $69 = ((($scan$1)) + 4|0);
       $70 = HEAP8[$69>>0]|0;
       $71 = ((($56)) + 5|0);
       $72 = HEAP8[$71>>0]|0;
       $73 = ($70<<24>>24)==($72<<24>>24);
       if (!($73)) {
        $scan$2 = $69;
        break;
       }
       $74 = ((($scan$1)) + 5|0);
       $75 = HEAP8[$74>>0]|0;
       $76 = ((($56)) + 6|0);
       $77 = HEAP8[$76>>0]|0;
       $78 = ($75<<24>>24)==($77<<24>>24);
       if (!($78)) {
        $scan$2 = $74;
        break;
       }
       $79 = ((($scan$1)) + 6|0);
       $80 = HEAP8[$79>>0]|0;
       $81 = ((($56)) + 7|0);
       $82 = HEAP8[$81>>0]|0;
       $83 = ($80<<24>>24)==($82<<24>>24);
       if (!($83)) {
        $scan$2 = $79;
        break;
       }
       $84 = ((($scan$1)) + 7|0);
       $85 = HEAP8[$84>>0]|0;
       $86 = ((($56)) + 8|0);
       $87 = HEAP8[$86>>0]|0;
       $88 = ($85<<24>>24)==($87<<24>>24);
       if (!($88)) {
        $scan$2 = $84;
        break;
       }
       $89 = ((($scan$1)) + 8|0);
       $90 = HEAP8[$89>>0]|0;
       $91 = ((($56)) + 9|0);
       $92 = HEAP8[$91>>0]|0;
       $93 = ($90<<24>>24)==($92<<24>>24);
       $94 = ($89>>>0)<($20>>>0);
       $or$cond = $94 & $93;
       if ($or$cond) {
        $56 = $86;$scan$1 = $89;
       } else {
        $scan$2 = $89;
        break;
       }
      }
      $95 = $scan$2;
      $96 = (($95) - ($34))|0;
      $97 = (($96) + 258)|0;
      $98 = ($97|0)>($best_len$0|0);
      if ($98) {
       HEAP32[$33>>2] = $$09;
       $99 = ($97|0)<($$10|0);
       if (!($99)) {
        $best_len$2 = $97;
        label = 20;
        break;
       }
       $100 = (($96) + 257)|0;
       $101 = (($35) + ($100)|0);
       $102 = HEAP8[$101>>0]|0;
       $103 = (($35) + ($97)|0);
       $104 = HEAP8[$103>>0]|0;
       $best_len$1 = $97;$scan$3 = $35;$scan_end$1 = $104;$scan_end1$1 = $102;
      } else {
       $best_len$1 = $best_len$0;$scan$3 = $35;$scan_end$1 = $scan_end$0;$scan_end1$1 = $scan_end1$0;
      }
     } else {
      $best_len$1 = $best_len$0;$scan$3 = $scan$0;$scan_end$1 = $scan_end$0;$scan_end1$1 = $scan_end1$0;
     }
    } else {
     $best_len$1 = $best_len$0;$scan$3 = $scan$0;$scan_end$1 = $scan_end$0;$scan_end1$1 = $scan_end1$0;
    }
   } else {
    $best_len$1 = $best_len$0;$scan$3 = $scan$0;$scan_end$1 = $scan_end$0;$scan_end1$1 = $scan_end1$0;
   }
  } else {
   $best_len$1 = $best_len$0;$scan$3 = $scan$0;$scan_end$1 = $scan_end$0;$scan_end1$1 = $scan_end1$0;
  }
  $105 = $$09 & $19;
  $106 = (($17) + ($105<<1)|0);
  $107 = HEAP16[$106>>1]|0;
  $108 = $107&65535;
  $109 = ($108>>>0)>($$>>>0);
  if (!($109)) {
   $best_len$2 = $best_len$1;
   label = 20;
   break;
  }
  $110 = (($chain_length$1) + -1)|0;
  $111 = ($110|0)==(0);
  if ($111) {
   $best_len$2 = $best_len$1;
   label = 20;
   break;
  } else {
   $$09 = $108;$best_len$0 = $best_len$1;$chain_length$1 = $110;$scan$0 = $scan$3;$scan_end$0 = $scan_end$1;$scan_end1$0 = $scan_end1$1;
  }
 }
 if ((label|0) == 20) {
  $112 = ($best_len$2>>>0)>($31>>>0);
  $$best_len$2 = $112 ? $31 : $best_len$2;
  return ($$best_len$2|0);
 }
 return (0)|0;
}
function _deflate_slow($s,$flush) {
 $s = $s|0;
 $flush = $flush|0;
 var $$ = 0, $$$i = 0, $$$i4 = 0, $$$i6 = 0, $$$i8 = 0, $$2 = 0, $$3 = 0, $$lcssa = 0, $$lcssa21 = 0, $$neg12 = 0, $$pn = 0, $$pr = 0, $$pr10 = 0, $$pre = 0, $$sink = 0, $$sink$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0;
 var $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0;
 var $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0;
 var $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0;
 var $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0;
 var $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0;
 var $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, $or$cond16 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($s)) + 116|0);
 $1 = ($flush|0)==(0);
 $2 = ((($s)) + 72|0);
 $3 = ((($s)) + 88|0);
 $4 = ((($s)) + 108|0);
 $5 = ((($s)) + 56|0);
 $6 = ((($s)) + 84|0);
 $7 = ((($s)) + 68|0);
 $8 = ((($s)) + 52|0);
 $9 = ((($s)) + 64|0);
 $10 = ((($s)) + 96|0);
 $11 = ((($s)) + 120|0);
 $12 = ((($s)) + 112|0);
 $13 = ((($s)) + 100|0);
 $14 = ((($s)) + 5792|0);
 $15 = ((($s)) + 5796|0);
 $16 = ((($s)) + 5784|0);
 $17 = ((($s)) + 5788|0);
 $18 = ((($s)) + 104|0);
 $19 = ((($s)) + 92|0);
 $20 = ((($s)) + 128|0);
 $21 = ((($s)) + 44|0);
 $22 = ((($s)) + 136|0);
 L1: while(1) {
  $$pr = HEAP32[$0>>2]|0;
  $24 = $$pr;
  while(1) {
   $23 = ($24>>>0)<(262);
   if ($23) {
    _fill_window($s);
    $25 = HEAP32[$0>>2]|0;
    $26 = ($25>>>0)<(262);
    $or$cond = $1 & $26;
    if ($or$cond) {
     $$2 = 0;
     label = 55;
     break L1;
    }
    $27 = ($25|0)==(0);
    if ($27) {
     label = 38;
     break L1;
    }
    $28 = ($25>>>0)>(2);
    if ($28) {
     label = 8;
    } else {
     $29 = HEAP32[$10>>2]|0;
     HEAP32[$11>>2] = $29;
     $30 = HEAP32[$12>>2]|0;
     HEAP32[$13>>2] = $30;
     HEAP32[$10>>2] = 2;
     $348 = 2;
     label = 16;
    }
   } else {
    label = 8;
   }
   do {
    if ((label|0) == 8) {
     label = 0;
     $31 = HEAP32[$2>>2]|0;
     $32 = HEAP32[$3>>2]|0;
     $33 = $31 << $32;
     $34 = HEAP32[$4>>2]|0;
     $35 = (($34) + 2)|0;
     $36 = HEAP32[$5>>2]|0;
     $37 = (($36) + ($35)|0);
     $38 = HEAP8[$37>>0]|0;
     $39 = $38&255;
     $40 = $39 ^ $33;
     $41 = HEAP32[$6>>2]|0;
     $42 = $40 & $41;
     HEAP32[$2>>2] = $42;
     $43 = HEAP32[$7>>2]|0;
     $44 = (($43) + ($42<<1)|0);
     $45 = HEAP16[$44>>1]|0;
     $46 = HEAP32[$8>>2]|0;
     $47 = $46 & $34;
     $48 = HEAP32[$9>>2]|0;
     $49 = (($48) + ($47<<1)|0);
     HEAP16[$49>>1] = $45;
     $50 = $45&65535;
     $51 = $34&65535;
     HEAP16[$44>>1] = $51;
     $52 = HEAP32[$10>>2]|0;
     HEAP32[$11>>2] = $52;
     $53 = HEAP32[$12>>2]|0;
     HEAP32[$13>>2] = $53;
     HEAP32[$10>>2] = 2;
     $54 = ($45<<16>>16)==(0);
     if ($54) {
      $348 = 2;
      label = 16;
     } else {
      $55 = HEAP32[$20>>2]|0;
      $56 = ($52>>>0)<($55>>>0);
      if ($56) {
       $57 = HEAP32[$4>>2]|0;
       $58 = (($57) - ($50))|0;
       $59 = HEAP32[$21>>2]|0;
       $60 = (($59) + -262)|0;
       $61 = ($58>>>0)>($60>>>0);
       if ($61) {
        $348 = 2;
        label = 16;
       } else {
        $62 = (_longest_match($s,$50)|0);
        HEAP32[$10>>2] = $62;
        $63 = ($62>>>0)<(6);
        if ($63) {
         $64 = HEAP32[$22>>2]|0;
         $65 = ($64|0)==(1);
         if (!($65)) {
          $66 = ($62|0)==(3);
          if (!($66)) {
           $348 = $62;
           label = 16;
           break;
          }
          $67 = HEAP32[$4>>2]|0;
          $68 = HEAP32[$12>>2]|0;
          $69 = (($67) - ($68))|0;
          $70 = ($69>>>0)>(4096);
          if (!($70)) {
           $348 = 3;
           label = 16;
           break;
          }
         }
         HEAP32[$10>>2] = 2;
         $348 = 2;
         label = 16;
        } else {
         $348 = $62;
         label = 16;
        }
       }
      } else {
       $72 = $52;$74 = 2;
      }
     }
    }
   } while(0);
   if ((label|0) == 16) {
    label = 0;
    $$pr10 = HEAP32[$11>>2]|0;
    $72 = $$pr10;$74 = $348;
   }
   $71 = ($72>>>0)<(3);
   $73 = ($74>>>0)>($72>>>0);
   $or$cond16 = $71 | $73;
   if (!($or$cond16)) {
    $$lcssa = $72;
    break;
   }
   $181 = HEAP32[$18>>2]|0;
   $182 = ($181|0)==(0);
   if ($182) {
    HEAP32[$18>>2] = 1;
    $245 = HEAP32[$4>>2]|0;
    $246 = (($245) + 1)|0;
    HEAP32[$4>>2] = $246;
    $247 = HEAP32[$0>>2]|0;
    $248 = (($247) + -1)|0;
    HEAP32[$0>>2] = $248;
    $24 = $248;
    continue;
   }
   $183 = HEAP32[$4>>2]|0;
   $184 = (($183) + -1)|0;
   $185 = HEAP32[$5>>2]|0;
   $186 = (($185) + ($184)|0);
   $187 = HEAP8[$186>>0]|0;
   $188 = HEAP32[$14>>2]|0;
   $189 = HEAP32[$15>>2]|0;
   $190 = (($189) + ($188<<1)|0);
   HEAP16[$190>>1] = 0;
   $191 = (($188) + 1)|0;
   HEAP32[$14>>2] = $191;
   $192 = HEAP32[$16>>2]|0;
   $193 = (($192) + ($188)|0);
   HEAP8[$193>>0] = $187;
   $194 = $187&255;
   $195 = (((($s)) + 148|0) + ($194<<2)|0);
   $196 = HEAP16[$195>>1]|0;
   $197 = (($196) + 1)<<16>>16;
   HEAP16[$195>>1] = $197;
   $198 = HEAP32[$14>>2]|0;
   $199 = HEAP32[$17>>2]|0;
   $200 = (($199) + -1)|0;
   $201 = ($198|0)==($200|0);
   if ($201) {
    $202 = HEAP32[$19>>2]|0;
    $203 = ($202|0)>(-1);
    if ($203) {
     $204 = HEAP32[$5>>2]|0;
     $205 = (($204) + ($202)|0);
     $208 = $205;
    } else {
     $208 = 0;
    }
    $206 = HEAP32[$4>>2]|0;
    $207 = (($206) - ($202))|0;
    __tr_flush_block($s,$208,$207,0);
    $209 = HEAP32[$4>>2]|0;
    HEAP32[$19>>2] = $209;
    $210 = HEAP32[$s>>2]|0;
    $211 = ((($210)) + 28|0);
    $212 = HEAP32[$211>>2]|0;
    __tr_flush_bits($212);
    $213 = ((($212)) + 20|0);
    $214 = HEAP32[$213>>2]|0;
    $215 = ((($210)) + 16|0);
    $216 = HEAP32[$215>>2]|0;
    $217 = ($214>>>0)>($216>>>0);
    $$$i4 = $217 ? $216 : $214;
    $218 = ($$$i4|0)==(0);
    if (!($218)) {
     $219 = ((($210)) + 12|0);
     $220 = HEAP32[$219>>2]|0;
     $221 = ((($212)) + 16|0);
     $222 = HEAP32[$221>>2]|0;
     _memcpy(($220|0),($222|0),($$$i4|0))|0;
     $223 = HEAP32[$219>>2]|0;
     $224 = (($223) + ($$$i4)|0);
     HEAP32[$219>>2] = $224;
     $225 = HEAP32[$221>>2]|0;
     $226 = (($225) + ($$$i4)|0);
     HEAP32[$221>>2] = $226;
     $227 = ((($210)) + 20|0);
     $228 = HEAP32[$227>>2]|0;
     $229 = (($228) + ($$$i4))|0;
     HEAP32[$227>>2] = $229;
     $230 = HEAP32[$215>>2]|0;
     $231 = (($230) - ($$$i4))|0;
     HEAP32[$215>>2] = $231;
     $232 = HEAP32[$213>>2]|0;
     $233 = (($232) - ($$$i4))|0;
     HEAP32[$213>>2] = $233;
     $234 = ($232|0)==($$$i4|0);
     if ($234) {
      $235 = ((($212)) + 8|0);
      $236 = HEAP32[$235>>2]|0;
      HEAP32[$221>>2] = $236;
     }
    }
   }
   $237 = HEAP32[$4>>2]|0;
   $238 = (($237) + 1)|0;
   HEAP32[$4>>2] = $238;
   $239 = HEAP32[$0>>2]|0;
   $240 = (($239) + -1)|0;
   HEAP32[$0>>2] = $240;
   $241 = HEAP32[$s>>2]|0;
   $242 = ((($241)) + 16|0);
   $243 = HEAP32[$242>>2]|0;
   $244 = ($243|0)==(0);
   if ($244) {
    $$2 = 0;
    label = 55;
    break L1;
   } else {
    $24 = $240;
   }
  }
  $75 = HEAP32[$4>>2]|0;
  $76 = HEAP32[$0>>2]|0;
  $77 = (($75) + -3)|0;
  $78 = (($77) + ($76))|0;
  $79 = (($$lcssa) + 253)|0;
  $80 = $79&255;
  $81 = (($75) + 65535)|0;
  $82 = HEAP32[$13>>2]|0;
  $83 = (($81) - ($82))|0;
  $84 = $83&65535;
  $85 = HEAP32[$14>>2]|0;
  $86 = HEAP32[$15>>2]|0;
  $87 = (($86) + ($85<<1)|0);
  HEAP16[$87>>1] = $84;
  $88 = (($85) + 1)|0;
  HEAP32[$14>>2] = $88;
  $89 = HEAP32[$16>>2]|0;
  $90 = (($89) + ($85)|0);
  HEAP8[$90>>0] = $80;
  $91 = (($83) + 65535)|0;
  $92 = $79 & 255;
  $93 = (613302 + ($92)|0);
  $94 = HEAP8[$93>>0]|0;
  $95 = $94&255;
  $96 = $95 | 256;
  $97 = (($96) + 1)|0;
  $98 = (((($s)) + 148|0) + ($97<<2)|0);
  $99 = HEAP16[$98>>1]|0;
  $100 = (($99) + 1)<<16>>16;
  HEAP16[$98>>1] = $100;
  $101 = $91 & 65535;
  $102 = ($101>>>0)<(256);
  $103 = $101 >>> 7;
  $104 = (($103) + 256)|0;
  $$pn = $102 ? $101 : $104;
  $$sink$in = (612790 + ($$pn)|0);
  $$sink = HEAP8[$$sink$in>>0]|0;
  $105 = $$sink&255;
  $106 = (((($s)) + 2440|0) + ($105<<2)|0);
  $107 = HEAP16[$106>>1]|0;
  $108 = (($107) + 1)<<16>>16;
  HEAP16[$106>>1] = $108;
  $109 = HEAP32[$14>>2]|0;
  $110 = HEAP32[$17>>2]|0;
  $111 = (($110) + -1)|0;
  $112 = HEAP32[$11>>2]|0;
  $113 = HEAP32[$0>>2]|0;
  $$neg12 = (1 - ($112))|0;
  $114 = (($$neg12) + ($113))|0;
  HEAP32[$0>>2] = $114;
  $115 = (($112) + -2)|0;
  HEAP32[$11>>2] = $115;
  $$pre = HEAP32[$4>>2]|0;
  $117 = $$pre;$139 = $115;
  while(1) {
   $116 = (($117) + 1)|0;
   HEAP32[$4>>2] = $116;
   $118 = ($116>>>0)>($78>>>0);
   if (!($118)) {
    $119 = HEAP32[$2>>2]|0;
    $120 = HEAP32[$3>>2]|0;
    $121 = $119 << $120;
    $122 = (($117) + 3)|0;
    $123 = HEAP32[$5>>2]|0;
    $124 = (($123) + ($122)|0);
    $125 = HEAP8[$124>>0]|0;
    $126 = $125&255;
    $127 = $126 ^ $121;
    $128 = HEAP32[$6>>2]|0;
    $129 = $127 & $128;
    HEAP32[$2>>2] = $129;
    $130 = HEAP32[$7>>2]|0;
    $131 = (($130) + ($129<<1)|0);
    $132 = HEAP16[$131>>1]|0;
    $133 = HEAP32[$8>>2]|0;
    $134 = $133 & $116;
    $135 = HEAP32[$9>>2]|0;
    $136 = (($135) + ($134<<1)|0);
    HEAP16[$136>>1] = $132;
    $137 = $116&65535;
    HEAP16[$131>>1] = $137;
   }
   $138 = (($139) + -1)|0;
   HEAP32[$11>>2] = $138;
   $140 = ($138|0)==(0);
   if ($140) {
    $$lcssa21 = $117;
    break;
   } else {
    $117 = $116;$139 = $138;
   }
  }
  $141 = ($109|0)==($111|0);
  HEAP32[$18>>2] = 0;
  HEAP32[$10>>2] = 2;
  $142 = (($$lcssa21) + 2)|0;
  HEAP32[$4>>2] = $142;
  if (!($141)) {
   continue;
  }
  $143 = HEAP32[$19>>2]|0;
  $144 = ($143|0)>(-1);
  if ($144) {
   $145 = HEAP32[$5>>2]|0;
   $146 = (($145) + ($143)|0);
   $148 = $146;
  } else {
   $148 = 0;
  }
  $147 = (($142) - ($143))|0;
  __tr_flush_block($s,$148,$147,0);
  $149 = HEAP32[$4>>2]|0;
  HEAP32[$19>>2] = $149;
  $150 = HEAP32[$s>>2]|0;
  $151 = ((($150)) + 28|0);
  $152 = HEAP32[$151>>2]|0;
  __tr_flush_bits($152);
  $153 = ((($152)) + 20|0);
  $154 = HEAP32[$153>>2]|0;
  $155 = ((($150)) + 16|0);
  $156 = HEAP32[$155>>2]|0;
  $157 = ($154>>>0)>($156>>>0);
  $$$i = $157 ? $156 : $154;
  $158 = ($$$i|0)==(0);
  if (!($158)) {
   $159 = ((($150)) + 12|0);
   $160 = HEAP32[$159>>2]|0;
   $161 = ((($152)) + 16|0);
   $162 = HEAP32[$161>>2]|0;
   _memcpy(($160|0),($162|0),($$$i|0))|0;
   $163 = HEAP32[$159>>2]|0;
   $164 = (($163) + ($$$i)|0);
   HEAP32[$159>>2] = $164;
   $165 = HEAP32[$161>>2]|0;
   $166 = (($165) + ($$$i)|0);
   HEAP32[$161>>2] = $166;
   $167 = ((($150)) + 20|0);
   $168 = HEAP32[$167>>2]|0;
   $169 = (($168) + ($$$i))|0;
   HEAP32[$167>>2] = $169;
   $170 = HEAP32[$155>>2]|0;
   $171 = (($170) - ($$$i))|0;
   HEAP32[$155>>2] = $171;
   $172 = HEAP32[$153>>2]|0;
   $173 = (($172) - ($$$i))|0;
   HEAP32[$153>>2] = $173;
   $174 = ($172|0)==($$$i|0);
   if ($174) {
    $175 = ((($152)) + 8|0);
    $176 = HEAP32[$175>>2]|0;
    HEAP32[$161>>2] = $176;
   }
  }
  $177 = HEAP32[$s>>2]|0;
  $178 = ((($177)) + 16|0);
  $179 = HEAP32[$178>>2]|0;
  $180 = ($179|0)==(0);
  if ($180) {
   $$2 = 0;
   label = 55;
   break;
  }
 }
 if ((label|0) == 38) {
  $249 = HEAP32[$18>>2]|0;
  $250 = ($249|0)==(0);
  if (!($250)) {
   $251 = HEAP32[$4>>2]|0;
   $252 = (($251) + -1)|0;
   $253 = HEAP32[$5>>2]|0;
   $254 = (($253) + ($252)|0);
   $255 = HEAP8[$254>>0]|0;
   $256 = HEAP32[$14>>2]|0;
   $257 = HEAP32[$15>>2]|0;
   $258 = (($257) + ($256<<1)|0);
   HEAP16[$258>>1] = 0;
   $259 = (($256) + 1)|0;
   HEAP32[$14>>2] = $259;
   $260 = HEAP32[$16>>2]|0;
   $261 = (($260) + ($256)|0);
   HEAP8[$261>>0] = $255;
   $262 = $255&255;
   $263 = (((($s)) + 148|0) + ($262<<2)|0);
   $264 = HEAP16[$263>>1]|0;
   $265 = (($264) + 1)<<16>>16;
   HEAP16[$263>>1] = $265;
   HEAP32[$18>>2] = 0;
  }
  $266 = HEAP32[$4>>2]|0;
  $267 = ($266>>>0)<(2);
  $$ = $267 ? $266 : 2;
  $268 = ((($s)) + 5812|0);
  HEAP32[$268>>2] = $$;
  $269 = ($flush|0)==(4);
  if ($269) {
   $270 = HEAP32[$19>>2]|0;
   $271 = ($270|0)>(-1);
   if ($271) {
    $272 = HEAP32[$5>>2]|0;
    $273 = (($272) + ($270)|0);
    $275 = $273;
   } else {
    $275 = 0;
   }
   $274 = (($266) - ($270))|0;
   __tr_flush_block($s,$275,$274,1);
   $276 = HEAP32[$4>>2]|0;
   HEAP32[$19>>2] = $276;
   $277 = HEAP32[$s>>2]|0;
   $278 = ((($277)) + 28|0);
   $279 = HEAP32[$278>>2]|0;
   __tr_flush_bits($279);
   $280 = ((($279)) + 20|0);
   $281 = HEAP32[$280>>2]|0;
   $282 = ((($277)) + 16|0);
   $283 = HEAP32[$282>>2]|0;
   $284 = ($281>>>0)>($283>>>0);
   $$$i6 = $284 ? $283 : $281;
   $285 = ($$$i6|0)==(0);
   if (!($285)) {
    $286 = ((($277)) + 12|0);
    $287 = HEAP32[$286>>2]|0;
    $288 = ((($279)) + 16|0);
    $289 = HEAP32[$288>>2]|0;
    _memcpy(($287|0),($289|0),($$$i6|0))|0;
    $290 = HEAP32[$286>>2]|0;
    $291 = (($290) + ($$$i6)|0);
    HEAP32[$286>>2] = $291;
    $292 = HEAP32[$288>>2]|0;
    $293 = (($292) + ($$$i6)|0);
    HEAP32[$288>>2] = $293;
    $294 = ((($277)) + 20|0);
    $295 = HEAP32[$294>>2]|0;
    $296 = (($295) + ($$$i6))|0;
    HEAP32[$294>>2] = $296;
    $297 = HEAP32[$282>>2]|0;
    $298 = (($297) - ($$$i6))|0;
    HEAP32[$282>>2] = $298;
    $299 = HEAP32[$280>>2]|0;
    $300 = (($299) - ($$$i6))|0;
    HEAP32[$280>>2] = $300;
    $301 = ($299|0)==($$$i6|0);
    if ($301) {
     $302 = ((($279)) + 8|0);
     $303 = HEAP32[$302>>2]|0;
     HEAP32[$288>>2] = $303;
    }
   }
   $304 = HEAP32[$s>>2]|0;
   $305 = ((($304)) + 16|0);
   $306 = HEAP32[$305>>2]|0;
   $307 = ($306|0)==(0);
   $$3 = $307 ? 2 : 3;
   $$2 = $$3;
   return ($$2|0);
  }
  $308 = HEAP32[$14>>2]|0;
  $309 = ($308|0)==(0);
  if (!($309)) {
   $310 = HEAP32[$19>>2]|0;
   $311 = ($310|0)>(-1);
   if ($311) {
    $312 = HEAP32[$5>>2]|0;
    $313 = (($312) + ($310)|0);
    $315 = $313;
   } else {
    $315 = 0;
   }
   $314 = (($266) - ($310))|0;
   __tr_flush_block($s,$315,$314,0);
   $316 = HEAP32[$4>>2]|0;
   HEAP32[$19>>2] = $316;
   $317 = HEAP32[$s>>2]|0;
   $318 = ((($317)) + 28|0);
   $319 = HEAP32[$318>>2]|0;
   __tr_flush_bits($319);
   $320 = ((($319)) + 20|0);
   $321 = HEAP32[$320>>2]|0;
   $322 = ((($317)) + 16|0);
   $323 = HEAP32[$322>>2]|0;
   $324 = ($321>>>0)>($323>>>0);
   $$$i8 = $324 ? $323 : $321;
   $325 = ($$$i8|0)==(0);
   if (!($325)) {
    $326 = ((($317)) + 12|0);
    $327 = HEAP32[$326>>2]|0;
    $328 = ((($319)) + 16|0);
    $329 = HEAP32[$328>>2]|0;
    _memcpy(($327|0),($329|0),($$$i8|0))|0;
    $330 = HEAP32[$326>>2]|0;
    $331 = (($330) + ($$$i8)|0);
    HEAP32[$326>>2] = $331;
    $332 = HEAP32[$328>>2]|0;
    $333 = (($332) + ($$$i8)|0);
    HEAP32[$328>>2] = $333;
    $334 = ((($317)) + 20|0);
    $335 = HEAP32[$334>>2]|0;
    $336 = (($335) + ($$$i8))|0;
    HEAP32[$334>>2] = $336;
    $337 = HEAP32[$322>>2]|0;
    $338 = (($337) - ($$$i8))|0;
    HEAP32[$322>>2] = $338;
    $339 = HEAP32[$320>>2]|0;
    $340 = (($339) - ($$$i8))|0;
    HEAP32[$320>>2] = $340;
    $341 = ($339|0)==($$$i8|0);
    if ($341) {
     $342 = ((($319)) + 8|0);
     $343 = HEAP32[$342>>2]|0;
     HEAP32[$328>>2] = $343;
    }
   }
   $344 = HEAP32[$s>>2]|0;
   $345 = ((($344)) + 16|0);
   $346 = HEAP32[$345>>2]|0;
   $347 = ($346|0)==(0);
   if ($347) {
    $$2 = 0;
    return ($$2|0);
   }
  }
  $$2 = 1;
  return ($$2|0);
 }
 else if ((label|0) == 55) {
  return ($$2|0);
 }
 return (0)|0;
}
function _deflateEnd($strm) {
 $strm = $strm|0;
 var $$0 = 0, $$in = 0, $$in20 = 0, $$in21 = 0, $$pre = 0, $$pre$phi17Z2D = 0, $$pre$phiZ2D = 0, $$pre12 = 0, $$pre13 = 0, $$pre14 = 0, $$pre15 = 0, $$pre16 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($strm|0)==(0|0);
 if ($0) {
  $$0 = -2;
  return ($$0|0);
 }
 $1 = ((($strm)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $$0 = -2;
  return ($$0|0);
 }
 $4 = ((($2)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 switch ($5|0) {
 case 42: case 69: case 73: case 91: case 103: case 113: case 666:  {
  break;
 }
 default: {
  $$0 = -2;
  return ($$0|0);
 }
 }
 $6 = ((($2)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 if ($8) {
  $$in = $2;
 } else {
  $9 = ((($strm)) + 36|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ((($strm)) + 40|0);
  $12 = HEAP32[$11>>2]|0;
  FUNCTION_TABLE_vii[$10 & 8191]($12,$7);
  $$pre = HEAP32[$1>>2]|0;
  $$in = $$pre;
 }
 $13 = ((($$in)) + 68|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ($14|0)==(0|0);
 if ($15) {
  $$in20 = $$in;
 } else {
  $16 = ((($strm)) + 36|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = ((($strm)) + 40|0);
  $19 = HEAP32[$18>>2]|0;
  FUNCTION_TABLE_vii[$17 & 8191]($19,$14);
  $$pre12 = HEAP32[$1>>2]|0;
  $$in20 = $$pre12;
 }
 $20 = ((($$in20)) + 64|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ($21|0)==(0|0);
 if ($22) {
  $$in21 = $$in20;
 } else {
  $23 = ((($strm)) + 36|0);
  $24 = HEAP32[$23>>2]|0;
  $25 = ((($strm)) + 40|0);
  $26 = HEAP32[$25>>2]|0;
  FUNCTION_TABLE_vii[$24 & 8191]($26,$21);
  $$pre13 = HEAP32[$1>>2]|0;
  $$in21 = $$pre13;
 }
 $27 = ((($$in21)) + 56|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ($28|0)==(0|0);
 if ($29) {
  $$pre15 = ((($strm)) + 36|0);
  $$pre16 = ((($strm)) + 40|0);
  $$pre$phi17Z2D = $$pre16;$$pre$phiZ2D = $$pre15;$36 = $$in21;
 } else {
  $30 = ((($strm)) + 36|0);
  $31 = HEAP32[$30>>2]|0;
  $32 = ((($strm)) + 40|0);
  $33 = HEAP32[$32>>2]|0;
  FUNCTION_TABLE_vii[$31 & 8191]($33,$28);
  $$pre14 = HEAP32[$1>>2]|0;
  $$pre$phi17Z2D = $32;$$pre$phiZ2D = $30;$36 = $$pre14;
 }
 $34 = HEAP32[$$pre$phiZ2D>>2]|0;
 $35 = HEAP32[$$pre$phi17Z2D>>2]|0;
 FUNCTION_TABLE_vii[$34 & 8191]($35,$36);
 HEAP32[$1>>2] = 0;
 $37 = ($5|0)==(113);
 $38 = $37 ? -3 : 0;
 $$0 = $38;
 return ($$0|0);
}
function _deflate($strm,$flush) {
 $strm = $strm|0;
 $flush = $flush|0;
 var $$ = 0, $$$i = 0, $$$i$i = 0, $$$i$i35 = 0, $$$i1$i = 0, $$$i11$i = 0, $$$i13$i = 0, $$$i20 = 0, $$$i22 = 0, $$$i24 = 0, $$$i26 = 0, $$$i28 = 0, $$$i3$i = 0, $$$i30 = 0, $$$i32 = 0, $$$i34 = 0, $$1 = 0, $$10$i = 0, $$13 = 0, $$14 = 0;
 var $$15 = 0, $$lcssa = 0, $$phi$trans$insert = 0, $$pr = 0, $$pr39$pr = 0, $$pr39$pr$pre = 0, $$pr41 = 0, $$pr45$pr = 0, $$pr45$pr$pre = 0, $$pre = 0, $$pre$i = 0, $$pre$phi83Z2D = 0, $$pre$phi87Z2D = 0, $$pre$phi89Z2D = 0, $$pre$phi91Z2D = 0, $$pre64 = 0, $$pre66 = 0, $$pre67 = 0, $$pre68$pre = 0, $$pre69 = 0;
 var $$pre72 = 0, $$pre74 = 0, $$pre76 = 0, $$pre86 = 0, $$pre88 = 0, $$pre90 = 0, $$pre92 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0;
 var $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0;
 var $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0;
 var $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0;
 var $1063 = 0, $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0;
 var $1081 = 0, $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0;
 var $11 = 0, $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0;
 var $1117 = 0, $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0;
 var $1135 = 0, $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0;
 var $1153 = 0, $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0;
 var $1171 = 0, $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0, $1178 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0;
 var $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0;
 var $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0;
 var $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0;
 var $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0;
 var $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0;
 var $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0;
 var $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0;
 var $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0;
 var $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0;
 var $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0;
 var $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0;
 var $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0;
 var $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0;
 var $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0;
 var $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0;
 var $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0;
 var $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0;
 var $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0;
 var $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0;
 var $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0;
 var $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0;
 var $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0;
 var $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0;
 var $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0;
 var $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0;
 var $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0;
 var $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0;
 var $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0;
 var $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0;
 var $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0;
 var $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0;
 var $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0;
 var $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0;
 var $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0;
 var $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0;
 var $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0;
 var $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0;
 var $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0;
 var $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0;
 var $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0;
 var $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0;
 var $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0;
 var $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $beg$059 = 0, $beg$1 = 0, $beg$2 = 0, $beg1$0 = 0, $beg1$1 = 0, $beg1$2 = 0, $beg2$0 = 0, $beg2$1 = 0, $beg2$2 = 0, $bflush$0$i = 0, $brmerge = 0, $level_flags$0 = 0, $not$ = 0, $or$cond = 0;
 var $or$cond$i = 0, $or$cond16 = 0, $or$cond17 = 0, $or$cond5 = 0, $or$cond7 = 0, $or$cond9$i = 0, $or$cond97 = 0, $scan$1$i = 0, $val$0 = 0, $val3$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($strm|0)==(0|0);
 if ($0) {
  $$1 = -2;
  return ($$1|0);
 }
 $1 = ((($strm)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 $4 = ($flush>>>0)>(5);
 $5 = $4 | $3;
 if ($5) {
  $$1 = -2;
  return ($$1|0);
 }
 $6 = ((($strm)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 do {
  if (!($8)) {
   $9 = HEAP32[$strm>>2]|0;
   $10 = ($9|0)==(0|0);
   if ($10) {
    $11 = ((($strm)) + 4|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = ($12|0)==(0);
    if (!($13)) {
     break;
    }
   }
   $14 = ((($2)) + 4|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)==(666);
   $17 = ($flush|0)!=(4);
   $or$cond5 = $17 & $16;
   if (!($or$cond5)) {
    $20 = ((($strm)) + 16|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = ($21|0)==(0);
    if ($22) {
     $23 = HEAP32[(141768)>>2]|0;
     $24 = ((($strm)) + 24|0);
     HEAP32[$24>>2] = $23;
     $$1 = -5;
     return ($$1|0);
    }
    HEAP32[$2>>2] = $strm;
    $25 = ((($2)) + 40|0);
    $26 = HEAP32[$25>>2]|0;
    HEAP32[$25>>2] = $flush;
    $27 = ($15|0)==(42);
    do {
     if ($27) {
      $28 = ((($2)) + 24|0);
      $29 = HEAP32[$28>>2]|0;
      $30 = ($29|0)==(2);
      if (!($30)) {
       $202 = ((($2)) + 48|0);
       $203 = HEAP32[$202>>2]|0;
       $204 = $203 << 12;
       $205 = (($204) + -30720)|0;
       $206 = ((($2)) + 136|0);
       $207 = HEAP32[$206>>2]|0;
       $208 = ($207|0)>(1);
       if ($208) {
        $level_flags$0 = 0;
       } else {
        $209 = ((($2)) + 132|0);
        $210 = HEAP32[$209>>2]|0;
        $211 = ($210|0)<(2);
        if ($211) {
         $level_flags$0 = 0;
        } else {
         $212 = ($210|0)<(6);
         if ($212) {
          $level_flags$0 = 64;
         } else {
          $213 = ($210|0)==(6);
          $$14 = $213 ? 128 : 192;
          $level_flags$0 = $$14;
         }
        }
       }
       $214 = $level_flags$0 | $205;
       $215 = ((($2)) + 108|0);
       $216 = HEAP32[$215>>2]|0;
       $217 = ($216|0)==(0);
       $218 = $214 | 32;
       $219 = $217 ? $214 : $218;
       $220 = (($219>>>0) % 31)&-1;
       $$15 = $220 | $219;
       $221 = $$15 ^ 31;
       HEAP32[$14>>2] = 113;
       $222 = $219 >>> 8;
       $223 = $222&255;
       $224 = ((($2)) + 20|0);
       $225 = HEAP32[$224>>2]|0;
       $226 = (($225) + 1)|0;
       HEAP32[$224>>2] = $226;
       $227 = ((($2)) + 8|0);
       $228 = HEAP32[$227>>2]|0;
       $229 = (($228) + ($225)|0);
       HEAP8[$229>>0] = $223;
       $230 = $221&255;
       $231 = HEAP32[$224>>2]|0;
       $232 = (($231) + 1)|0;
       HEAP32[$224>>2] = $232;
       $233 = HEAP32[$227>>2]|0;
       $234 = (($233) + ($231)|0);
       HEAP8[$234>>0] = $230;
       $235 = HEAP32[$215>>2]|0;
       $236 = ($235|0)==(0);
       $$pre92 = ((($strm)) + 48|0);
       if (!($236)) {
        $237 = HEAP32[$$pre92>>2]|0;
        $238 = $237 >>> 16;
        $239 = $237 >>> 24;
        $240 = $239&255;
        $241 = HEAP32[$224>>2]|0;
        $242 = (($241) + 1)|0;
        HEAP32[$224>>2] = $242;
        $243 = HEAP32[$227>>2]|0;
        $244 = (($243) + ($241)|0);
        HEAP8[$244>>0] = $240;
        $245 = $238&255;
        $246 = HEAP32[$224>>2]|0;
        $247 = (($246) + 1)|0;
        HEAP32[$224>>2] = $247;
        $248 = HEAP32[$227>>2]|0;
        $249 = (($248) + ($246)|0);
        HEAP8[$249>>0] = $245;
        $250 = HEAP32[$$pre92>>2]|0;
        $251 = $250 >>> 8;
        $252 = $251&255;
        $253 = HEAP32[$224>>2]|0;
        $254 = (($253) + 1)|0;
        HEAP32[$224>>2] = $254;
        $255 = HEAP32[$227>>2]|0;
        $256 = (($255) + ($253)|0);
        HEAP8[$256>>0] = $252;
        $257 = $250&255;
        $258 = HEAP32[$224>>2]|0;
        $259 = (($258) + 1)|0;
        HEAP32[$224>>2] = $259;
        $260 = HEAP32[$227>>2]|0;
        $261 = (($260) + ($258)|0);
        HEAP8[$261>>0] = $257;
       }
       $262 = (_adler32(0,0,0)|0);
       HEAP32[$$pre92>>2] = $262;
       $$pr = HEAP32[$14>>2]|0;
       $264 = $$pr;
       label = 30;
       break;
      }
      $31 = (_crc32(0,0,0)|0);
      $32 = ((($strm)) + 48|0);
      HEAP32[$32>>2] = $31;
      $33 = ((($2)) + 20|0);
      $34 = HEAP32[$33>>2]|0;
      $35 = (($34) + 1)|0;
      HEAP32[$33>>2] = $35;
      $36 = ((($2)) + 8|0);
      $37 = HEAP32[$36>>2]|0;
      $38 = (($37) + ($34)|0);
      HEAP8[$38>>0] = 31;
      $39 = HEAP32[$33>>2]|0;
      $40 = (($39) + 1)|0;
      HEAP32[$33>>2] = $40;
      $41 = HEAP32[$36>>2]|0;
      $42 = (($41) + ($39)|0);
      HEAP8[$42>>0] = -117;
      $43 = HEAP32[$33>>2]|0;
      $44 = (($43) + 1)|0;
      HEAP32[$33>>2] = $44;
      $45 = HEAP32[$36>>2]|0;
      $46 = (($45) + ($43)|0);
      HEAP8[$46>>0] = 8;
      $47 = ((($2)) + 28|0);
      $48 = HEAP32[$47>>2]|0;
      $49 = ($48|0)==(0|0);
      if ($49) {
       $50 = HEAP32[$33>>2]|0;
       $51 = (($50) + 1)|0;
       HEAP32[$33>>2] = $51;
       $52 = HEAP32[$36>>2]|0;
       $53 = (($52) + ($50)|0);
       HEAP8[$53>>0] = 0;
       $54 = HEAP32[$33>>2]|0;
       $55 = (($54) + 1)|0;
       HEAP32[$33>>2] = $55;
       $56 = HEAP32[$36>>2]|0;
       $57 = (($56) + ($54)|0);
       HEAP8[$57>>0] = 0;
       $58 = HEAP32[$33>>2]|0;
       $59 = (($58) + 1)|0;
       HEAP32[$33>>2] = $59;
       $60 = HEAP32[$36>>2]|0;
       $61 = (($60) + ($58)|0);
       HEAP8[$61>>0] = 0;
       $62 = HEAP32[$33>>2]|0;
       $63 = (($62) + 1)|0;
       HEAP32[$33>>2] = $63;
       $64 = HEAP32[$36>>2]|0;
       $65 = (($64) + ($62)|0);
       HEAP8[$65>>0] = 0;
       $66 = HEAP32[$33>>2]|0;
       $67 = (($66) + 1)|0;
       HEAP32[$33>>2] = $67;
       $68 = HEAP32[$36>>2]|0;
       $69 = (($68) + ($66)|0);
       HEAP8[$69>>0] = 0;
       $70 = ((($2)) + 132|0);
       $71 = HEAP32[$70>>2]|0;
       $72 = ($71|0)==(9);
       if ($72) {
        $79 = 2;
       } else {
        $73 = ((($2)) + 136|0);
        $74 = HEAP32[$73>>2]|0;
        $75 = ($74|0)>(1);
        $76 = ($71|0)<(2);
        $$ = $76 | $75;
        $77 = $$ ? 4 : 0;
        $79 = $77;
       }
       $78 = $79&255;
       $80 = HEAP32[$33>>2]|0;
       $81 = (($80) + 1)|0;
       HEAP32[$33>>2] = $81;
       $82 = HEAP32[$36>>2]|0;
       $83 = (($82) + ($80)|0);
       HEAP8[$83>>0] = $78;
       $84 = HEAP32[$33>>2]|0;
       $85 = (($84) + 1)|0;
       HEAP32[$33>>2] = $85;
       $86 = HEAP32[$36>>2]|0;
       $87 = (($86) + ($84)|0);
       HEAP8[$87>>0] = 3;
       HEAP32[$14>>2] = 113;
       break;
      }
      $88 = HEAP32[$48>>2]|0;
      $89 = ($88|0)!=(0);
      $90 = $89&1;
      $91 = ((($48)) + 44|0);
      $92 = HEAP32[$91>>2]|0;
      $93 = ($92|0)!=(0);
      $94 = $93 ? 2 : 0;
      $95 = $94 | $90;
      $96 = ((($48)) + 16|0);
      $97 = HEAP32[$96>>2]|0;
      $98 = ($97|0)==(0|0);
      $99 = $98 ? 0 : 4;
      $100 = $95 | $99;
      $101 = ((($48)) + 28|0);
      $102 = HEAP32[$101>>2]|0;
      $103 = ($102|0)==(0|0);
      $104 = $103 ? 0 : 8;
      $105 = $100 | $104;
      $106 = ((($48)) + 36|0);
      $107 = HEAP32[$106>>2]|0;
      $108 = ($107|0)==(0|0);
      $109 = $108 ? 0 : 16;
      $110 = $105 | $109;
      $111 = $110&255;
      $112 = HEAP32[$33>>2]|0;
      $113 = (($112) + 1)|0;
      HEAP32[$33>>2] = $113;
      $114 = HEAP32[$36>>2]|0;
      $115 = (($114) + ($112)|0);
      HEAP8[$115>>0] = $111;
      $116 = HEAP32[$47>>2]|0;
      $117 = ((($116)) + 4|0);
      $118 = HEAP32[$117>>2]|0;
      $119 = $118&255;
      $120 = HEAP32[$33>>2]|0;
      $121 = (($120) + 1)|0;
      HEAP32[$33>>2] = $121;
      $122 = HEAP32[$36>>2]|0;
      $123 = (($122) + ($120)|0);
      HEAP8[$123>>0] = $119;
      $124 = HEAP32[$47>>2]|0;
      $125 = ((($124)) + 4|0);
      $126 = HEAP32[$125>>2]|0;
      $127 = $126 >>> 8;
      $128 = $127&255;
      $129 = HEAP32[$33>>2]|0;
      $130 = (($129) + 1)|0;
      HEAP32[$33>>2] = $130;
      $131 = HEAP32[$36>>2]|0;
      $132 = (($131) + ($129)|0);
      HEAP8[$132>>0] = $128;
      $133 = HEAP32[$47>>2]|0;
      $134 = ((($133)) + 4|0);
      $135 = HEAP32[$134>>2]|0;
      $136 = $135 >>> 16;
      $137 = $136&255;
      $138 = HEAP32[$33>>2]|0;
      $139 = (($138) + 1)|0;
      HEAP32[$33>>2] = $139;
      $140 = HEAP32[$36>>2]|0;
      $141 = (($140) + ($138)|0);
      HEAP8[$141>>0] = $137;
      $142 = HEAP32[$47>>2]|0;
      $143 = ((($142)) + 4|0);
      $144 = HEAP32[$143>>2]|0;
      $145 = $144 >>> 24;
      $146 = $145&255;
      $147 = HEAP32[$33>>2]|0;
      $148 = (($147) + 1)|0;
      HEAP32[$33>>2] = $148;
      $149 = HEAP32[$36>>2]|0;
      $150 = (($149) + ($147)|0);
      HEAP8[$150>>0] = $146;
      $151 = ((($2)) + 132|0);
      $152 = HEAP32[$151>>2]|0;
      $153 = ($152|0)==(9);
      if ($153) {
       $160 = 2;
      } else {
       $154 = ((($2)) + 136|0);
       $155 = HEAP32[$154>>2]|0;
       $156 = ($155|0)>(1);
       $157 = ($152|0)<(2);
       $$13 = $157 | $156;
       $158 = $$13 ? 4 : 0;
       $160 = $158;
      }
      $159 = $160&255;
      $161 = HEAP32[$33>>2]|0;
      $162 = (($161) + 1)|0;
      HEAP32[$33>>2] = $162;
      $163 = HEAP32[$36>>2]|0;
      $164 = (($163) + ($161)|0);
      HEAP8[$164>>0] = $159;
      $165 = HEAP32[$47>>2]|0;
      $166 = ((($165)) + 12|0);
      $167 = HEAP32[$166>>2]|0;
      $168 = $167&255;
      $169 = HEAP32[$33>>2]|0;
      $170 = (($169) + 1)|0;
      HEAP32[$33>>2] = $170;
      $171 = HEAP32[$36>>2]|0;
      $172 = (($171) + ($169)|0);
      HEAP8[$172>>0] = $168;
      $173 = HEAP32[$47>>2]|0;
      $174 = ((($173)) + 16|0);
      $175 = HEAP32[$174>>2]|0;
      $176 = ($175|0)==(0|0);
      if ($176) {
       $194 = $173;
      } else {
       $177 = ((($173)) + 20|0);
       $178 = HEAP32[$177>>2]|0;
       $179 = $178&255;
       $180 = HEAP32[$33>>2]|0;
       $181 = (($180) + 1)|0;
       HEAP32[$33>>2] = $181;
       $182 = HEAP32[$36>>2]|0;
       $183 = (($182) + ($180)|0);
       HEAP8[$183>>0] = $179;
       $184 = HEAP32[$47>>2]|0;
       $185 = ((($184)) + 20|0);
       $186 = HEAP32[$185>>2]|0;
       $187 = $186 >>> 8;
       $188 = $187&255;
       $189 = HEAP32[$33>>2]|0;
       $190 = (($189) + 1)|0;
       HEAP32[$33>>2] = $190;
       $191 = HEAP32[$36>>2]|0;
       $192 = (($191) + ($189)|0);
       HEAP8[$192>>0] = $188;
       $$pre = HEAP32[$47>>2]|0;
       $194 = $$pre;
      }
      $193 = ((($194)) + 44|0);
      $195 = HEAP32[$193>>2]|0;
      $196 = ($195|0)==(0);
      if (!($196)) {
       $197 = HEAP32[$32>>2]|0;
       $198 = HEAP32[$36>>2]|0;
       $199 = HEAP32[$33>>2]|0;
       $200 = (_crc32($197,$198,$199)|0);
       HEAP32[$32>>2] = $200;
      }
      $201 = ((($2)) + 32|0);
      HEAP32[$201>>2] = 0;
      HEAP32[$14>>2] = 69;
      $$pre$phi87Z2D = $47;
      label = 32;
     } else {
      $264 = $15;
      label = 30;
     }
    } while(0);
    if ((label|0) == 30) {
     $263 = ($264|0)==(69);
     if ($263) {
      $$pre86 = ((($2)) + 28|0);
      $$pre$phi87Z2D = $$pre86;
      label = 32;
     } else {
      $$pr39$pr = $264;
      label = 53;
     }
    }
    do {
     if ((label|0) == 32) {
      $265 = HEAP32[$$pre$phi87Z2D>>2]|0;
      $266 = ((($265)) + 16|0);
      $267 = HEAP32[$266>>2]|0;
      $268 = ($267|0)==(0|0);
      if ($268) {
       HEAP32[$14>>2] = 73;
       $$pre$phi83Z2D = $$pre$phi87Z2D;$356 = $265;
       label = 55;
       break;
      }
      $269 = ((($2)) + 20|0);
      $270 = HEAP32[$269>>2]|0;
      $271 = ((($2)) + 32|0);
      $272 = HEAP32[$271>>2]|0;
      $273 = ((($265)) + 20|0);
      $274 = HEAP32[$273>>2]|0;
      $275 = $274 & 65535;
      $276 = ($272>>>0)<($275>>>0);
      L53: do {
       if ($276) {
        $277 = ((($2)) + 12|0);
        $278 = ((($strm)) + 48|0);
        $279 = ((($2)) + 8|0);
        $280 = ((($strm)) + 20|0);
        $1178 = $272;$283 = $270;$285 = $265;$beg$059 = $270;
        while(1) {
         $281 = HEAP32[$277>>2]|0;
         $282 = ($283|0)==($281|0);
         if ($282) {
          $284 = ((($285)) + 44|0);
          $286 = HEAP32[$284>>2]|0;
          $287 = ($286|0)!=(0);
          $288 = ($283>>>0)>($beg$059>>>0);
          $or$cond = $288 & $287;
          if ($or$cond) {
           $289 = HEAP32[$278>>2]|0;
           $290 = HEAP32[$279>>2]|0;
           $291 = (($290) + ($beg$059)|0);
           $292 = (($283) - ($beg$059))|0;
           $293 = (_crc32($289,$291,$292)|0);
           HEAP32[$278>>2] = $293;
          }
          $294 = HEAP32[$1>>2]|0;
          __tr_flush_bits($294);
          $295 = ((($294)) + 20|0);
          $296 = HEAP32[$295>>2]|0;
          $297 = HEAP32[$20>>2]|0;
          $298 = ($296>>>0)>($297>>>0);
          $$$i22 = $298 ? $297 : $296;
          $299 = ($$$i22|0)==(0);
          if (!($299)) {
           $300 = HEAP32[$6>>2]|0;
           $301 = ((($294)) + 16|0);
           $302 = HEAP32[$301>>2]|0;
           _memcpy(($300|0),($302|0),($$$i22|0))|0;
           $303 = HEAP32[$6>>2]|0;
           $304 = (($303) + ($$$i22)|0);
           HEAP32[$6>>2] = $304;
           $305 = HEAP32[$301>>2]|0;
           $306 = (($305) + ($$$i22)|0);
           HEAP32[$301>>2] = $306;
           $307 = HEAP32[$280>>2]|0;
           $308 = (($307) + ($$$i22))|0;
           HEAP32[$280>>2] = $308;
           $309 = HEAP32[$20>>2]|0;
           $310 = (($309) - ($$$i22))|0;
           HEAP32[$20>>2] = $310;
           $311 = HEAP32[$295>>2]|0;
           $312 = (($311) - ($$$i22))|0;
           HEAP32[$295>>2] = $312;
           $313 = ($311|0)==($$$i22|0);
           if ($313) {
            $314 = ((($294)) + 8|0);
            $315 = HEAP32[$314>>2]|0;
            HEAP32[$301>>2] = $315;
           }
          }
          $316 = HEAP32[$269>>2]|0;
          $317 = HEAP32[$277>>2]|0;
          $318 = ($316|0)==($317|0);
          if ($318) {
           $$lcssa = $316;
           break;
          }
          $$pre66 = HEAP32[$271>>2]|0;
          $$pre67 = HEAP32[$$pre$phi87Z2D>>2]|0;
          $320 = $$pre67;$323 = $$pre66;$326 = $316;$beg$1 = $316;
         } else {
          $320 = $285;$323 = $1178;$326 = $283;$beg$1 = $beg$059;
         }
         $319 = ((($320)) + 16|0);
         $321 = HEAP32[$319>>2]|0;
         $322 = (($321) + ($323)|0);
         $324 = HEAP8[$322>>0]|0;
         $325 = (($326) + 1)|0;
         HEAP32[$269>>2] = $325;
         $327 = HEAP32[$279>>2]|0;
         $328 = (($327) + ($326)|0);
         HEAP8[$328>>0] = $324;
         $329 = HEAP32[$271>>2]|0;
         $330 = (($329) + 1)|0;
         HEAP32[$271>>2] = $330;
         $331 = HEAP32[$$pre$phi87Z2D>>2]|0;
         $332 = ((($331)) + 20|0);
         $333 = HEAP32[$332>>2]|0;
         $334 = $333 & 65535;
         $335 = ($330>>>0)<($334>>>0);
         if (!($335)) {
          $337 = $331;$beg$2 = $beg$1;
          break L53;
         }
         $$pre64 = HEAP32[$269>>2]|0;
         $1178 = $330;$283 = $$pre64;$285 = $331;$beg$059 = $beg$1;
        }
        $$pre68$pre = HEAP32[$$pre$phi87Z2D>>2]|0;
        $337 = $$pre68$pre;$beg$2 = $$lcssa;
       } else {
        $337 = $265;$beg$2 = $270;
       }
      } while(0);
      $336 = ((($337)) + 44|0);
      $338 = HEAP32[$336>>2]|0;
      $339 = ($338|0)==(0);
      if ($339) {
       $351 = $337;
      } else {
       $340 = HEAP32[$269>>2]|0;
       $341 = ($340>>>0)>($beg$2>>>0);
       if ($341) {
        $342 = ((($strm)) + 48|0);
        $343 = HEAP32[$342>>2]|0;
        $344 = ((($2)) + 8|0);
        $345 = HEAP32[$344>>2]|0;
        $346 = (($345) + ($beg$2)|0);
        $347 = (($340) - ($beg$2))|0;
        $348 = (_crc32($343,$346,$347)|0);
        HEAP32[$342>>2] = $348;
        $$pre69 = HEAP32[$$pre$phi87Z2D>>2]|0;
        $351 = $$pre69;
       } else {
        $351 = $337;
       }
      }
      $349 = HEAP32[$271>>2]|0;
      $350 = ((($351)) + 20|0);
      $352 = HEAP32[$350>>2]|0;
      $353 = ($349|0)==($352|0);
      if ($353) {
       HEAP32[$271>>2] = 0;
       HEAP32[$14>>2] = 73;
       $$pre$phi83Z2D = $$pre$phi87Z2D;$356 = $351;
       label = 55;
       break;
      } else {
       $$pr39$pr$pre = HEAP32[$14>>2]|0;
       $$pr39$pr = $$pr39$pr$pre;
       label = 53;
       break;
      }
     }
    } while(0);
    if ((label|0) == 53) {
     $354 = ($$pr39$pr|0)==(73);
     if ($354) {
      $$phi$trans$insert = ((($2)) + 28|0);
      $$pre72 = HEAP32[$$phi$trans$insert>>2]|0;
      $$pre$phi83Z2D = $$phi$trans$insert;$356 = $$pre72;
      label = 55;
     } else {
      $430 = $$pr39$pr;
      label = 73;
     }
    }
    do {
     if ((label|0) == 55) {
      $355 = ((($356)) + 28|0);
      $357 = HEAP32[$355>>2]|0;
      $358 = ($357|0)==(0|0);
      if ($358) {
       HEAP32[$14>>2] = 91;
       $$pre$phi89Z2D = $$pre$phi83Z2D;
       label = 75;
       break;
      }
      $359 = ((($2)) + 20|0);
      $360 = HEAP32[$359>>2]|0;
      $361 = ((($2)) + 12|0);
      $362 = ((($strm)) + 48|0);
      $363 = ((($2)) + 8|0);
      $364 = ((($strm)) + 20|0);
      $365 = ((($2)) + 32|0);
      $368 = $360;$beg1$0 = $360;
      while(1) {
       $366 = HEAP32[$361>>2]|0;
       $367 = ($368|0)==($366|0);
       if ($367) {
        $369 = HEAP32[$$pre$phi83Z2D>>2]|0;
        $370 = ((($369)) + 44|0);
        $371 = HEAP32[$370>>2]|0;
        $372 = ($371|0)!=(0);
        $373 = ($368>>>0)>($beg1$0>>>0);
        $or$cond16 = $373 & $372;
        if ($or$cond16) {
         $374 = HEAP32[$362>>2]|0;
         $375 = HEAP32[$363>>2]|0;
         $376 = (($375) + ($beg1$0)|0);
         $377 = (($368) - ($beg1$0))|0;
         $378 = (_crc32($374,$376,$377)|0);
         HEAP32[$362>>2] = $378;
        }
        $379 = HEAP32[$1>>2]|0;
        __tr_flush_bits($379);
        $380 = ((($379)) + 20|0);
        $381 = HEAP32[$380>>2]|0;
        $382 = HEAP32[$20>>2]|0;
        $383 = ($381>>>0)>($382>>>0);
        $$$i24 = $383 ? $382 : $381;
        $384 = ($$$i24|0)==(0);
        if (!($384)) {
         $385 = HEAP32[$6>>2]|0;
         $386 = ((($379)) + 16|0);
         $387 = HEAP32[$386>>2]|0;
         _memcpy(($385|0),($387|0),($$$i24|0))|0;
         $388 = HEAP32[$6>>2]|0;
         $389 = (($388) + ($$$i24)|0);
         HEAP32[$6>>2] = $389;
         $390 = HEAP32[$386>>2]|0;
         $391 = (($390) + ($$$i24)|0);
         HEAP32[$386>>2] = $391;
         $392 = HEAP32[$364>>2]|0;
         $393 = (($392) + ($$$i24))|0;
         HEAP32[$364>>2] = $393;
         $394 = HEAP32[$20>>2]|0;
         $395 = (($394) - ($$$i24))|0;
         HEAP32[$20>>2] = $395;
         $396 = HEAP32[$380>>2]|0;
         $397 = (($396) - ($$$i24))|0;
         HEAP32[$380>>2] = $397;
         $398 = ($396|0)==($$$i24|0);
         if ($398) {
          $399 = ((($379)) + 8|0);
          $400 = HEAP32[$399>>2]|0;
          HEAP32[$386>>2] = $400;
         }
        }
        $401 = HEAP32[$359>>2]|0;
        $402 = HEAP32[$361>>2]|0;
        $403 = ($401|0)==($402|0);
        if ($403) {
         $beg1$2 = $401;$val$0 = 1;
         break;
        } else {
         $413 = $401;$beg1$1 = $401;
        }
       } else {
        $413 = $368;$beg1$1 = $beg1$0;
       }
       $404 = HEAP32[$365>>2]|0;
       $405 = (($404) + 1)|0;
       HEAP32[$365>>2] = $405;
       $406 = HEAP32[$$pre$phi83Z2D>>2]|0;
       $407 = ((($406)) + 28|0);
       $408 = HEAP32[$407>>2]|0;
       $409 = (($408) + ($404)|0);
       $410 = HEAP8[$409>>0]|0;
       $411 = $410&255;
       $412 = (($413) + 1)|0;
       HEAP32[$359>>2] = $412;
       $414 = HEAP32[$363>>2]|0;
       $415 = (($414) + ($413)|0);
       HEAP8[$415>>0] = $410;
       $416 = ($410<<24>>24)==(0);
       if ($416) {
        $beg1$2 = $beg1$1;$val$0 = $411;
        break;
       }
       $$pre74 = HEAP32[$359>>2]|0;
       $368 = $$pre74;$beg1$0 = $beg1$1;
      }
      $417 = HEAP32[$$pre$phi83Z2D>>2]|0;
      $418 = ((($417)) + 44|0);
      $419 = HEAP32[$418>>2]|0;
      $420 = ($419|0)==(0);
      if (!($420)) {
       $421 = HEAP32[$359>>2]|0;
       $422 = ($421>>>0)>($beg1$2>>>0);
       if ($422) {
        $423 = HEAP32[$362>>2]|0;
        $424 = HEAP32[$363>>2]|0;
        $425 = (($424) + ($beg1$2)|0);
        $426 = (($421) - ($beg1$2))|0;
        $427 = (_crc32($423,$425,$426)|0);
        HEAP32[$362>>2] = $427;
       }
      }
      $428 = ($val$0|0)==(0);
      if ($428) {
       HEAP32[$365>>2] = 0;
       HEAP32[$14>>2] = 91;
       $$pre$phi89Z2D = $$pre$phi83Z2D;
       label = 75;
       break;
      } else {
       $$pr41 = HEAP32[$14>>2]|0;
       $430 = $$pr41;
       label = 73;
       break;
      }
     }
    } while(0);
    if ((label|0) == 73) {
     $429 = ($430|0)==(91);
     if ($429) {
      $$pre88 = ((($2)) + 28|0);
      $$pre$phi89Z2D = $$pre88;
      label = 75;
     } else {
      $$pr45$pr = $430;
      label = 93;
     }
    }
    do {
     if ((label|0) == 75) {
      $431 = HEAP32[$$pre$phi89Z2D>>2]|0;
      $432 = ((($431)) + 36|0);
      $433 = HEAP32[$432>>2]|0;
      $434 = ($433|0)==(0|0);
      if ($434) {
       HEAP32[$14>>2] = 103;
       $$pre$phi91Z2D = $$pre$phi89Z2D;
       label = 95;
       break;
      }
      $435 = ((($2)) + 20|0);
      $436 = HEAP32[$435>>2]|0;
      $437 = ((($2)) + 12|0);
      $438 = ((($strm)) + 48|0);
      $439 = ((($2)) + 8|0);
      $440 = ((($strm)) + 20|0);
      $441 = ((($2)) + 32|0);
      $444 = $436;$beg2$0 = $436;
      while(1) {
       $442 = HEAP32[$437>>2]|0;
       $443 = ($444|0)==($442|0);
       if ($443) {
        $445 = HEAP32[$$pre$phi89Z2D>>2]|0;
        $446 = ((($445)) + 44|0);
        $447 = HEAP32[$446>>2]|0;
        $448 = ($447|0)!=(0);
        $449 = ($444>>>0)>($beg2$0>>>0);
        $or$cond17 = $449 & $448;
        if ($or$cond17) {
         $450 = HEAP32[$438>>2]|0;
         $451 = HEAP32[$439>>2]|0;
         $452 = (($451) + ($beg2$0)|0);
         $453 = (($444) - ($beg2$0))|0;
         $454 = (_crc32($450,$452,$453)|0);
         HEAP32[$438>>2] = $454;
        }
        $455 = HEAP32[$1>>2]|0;
        __tr_flush_bits($455);
        $456 = ((($455)) + 20|0);
        $457 = HEAP32[$456>>2]|0;
        $458 = HEAP32[$20>>2]|0;
        $459 = ($457>>>0)>($458>>>0);
        $$$i26 = $459 ? $458 : $457;
        $460 = ($$$i26|0)==(0);
        if (!($460)) {
         $461 = HEAP32[$6>>2]|0;
         $462 = ((($455)) + 16|0);
         $463 = HEAP32[$462>>2]|0;
         _memcpy(($461|0),($463|0),($$$i26|0))|0;
         $464 = HEAP32[$6>>2]|0;
         $465 = (($464) + ($$$i26)|0);
         HEAP32[$6>>2] = $465;
         $466 = HEAP32[$462>>2]|0;
         $467 = (($466) + ($$$i26)|0);
         HEAP32[$462>>2] = $467;
         $468 = HEAP32[$440>>2]|0;
         $469 = (($468) + ($$$i26))|0;
         HEAP32[$440>>2] = $469;
         $470 = HEAP32[$20>>2]|0;
         $471 = (($470) - ($$$i26))|0;
         HEAP32[$20>>2] = $471;
         $472 = HEAP32[$456>>2]|0;
         $473 = (($472) - ($$$i26))|0;
         HEAP32[$456>>2] = $473;
         $474 = ($472|0)==($$$i26|0);
         if ($474) {
          $475 = ((($455)) + 8|0);
          $476 = HEAP32[$475>>2]|0;
          HEAP32[$462>>2] = $476;
         }
        }
        $477 = HEAP32[$435>>2]|0;
        $478 = HEAP32[$437>>2]|0;
        $479 = ($477|0)==($478|0);
        if ($479) {
         $beg2$2 = $477;$val3$0 = 1;
         break;
        } else {
         $489 = $477;$beg2$1 = $477;
        }
       } else {
        $489 = $444;$beg2$1 = $beg2$0;
       }
       $480 = HEAP32[$441>>2]|0;
       $481 = (($480) + 1)|0;
       HEAP32[$441>>2] = $481;
       $482 = HEAP32[$$pre$phi89Z2D>>2]|0;
       $483 = ((($482)) + 36|0);
       $484 = HEAP32[$483>>2]|0;
       $485 = (($484) + ($480)|0);
       $486 = HEAP8[$485>>0]|0;
       $487 = $486&255;
       $488 = (($489) + 1)|0;
       HEAP32[$435>>2] = $488;
       $490 = HEAP32[$439>>2]|0;
       $491 = (($490) + ($489)|0);
       HEAP8[$491>>0] = $486;
       $492 = ($486<<24>>24)==(0);
       if ($492) {
        $beg2$2 = $beg2$1;$val3$0 = $487;
        break;
       }
       $$pre76 = HEAP32[$435>>2]|0;
       $444 = $$pre76;$beg2$0 = $beg2$1;
      }
      $493 = HEAP32[$$pre$phi89Z2D>>2]|0;
      $494 = ((($493)) + 44|0);
      $495 = HEAP32[$494>>2]|0;
      $496 = ($495|0)==(0);
      if (!($496)) {
       $497 = HEAP32[$435>>2]|0;
       $498 = ($497>>>0)>($beg2$2>>>0);
       if ($498) {
        $499 = HEAP32[$438>>2]|0;
        $500 = HEAP32[$439>>2]|0;
        $501 = (($500) + ($beg2$2)|0);
        $502 = (($497) - ($beg2$2))|0;
        $503 = (_crc32($499,$501,$502)|0);
        HEAP32[$438>>2] = $503;
       }
      }
      $504 = ($val3$0|0)==(0);
      if ($504) {
       HEAP32[$14>>2] = 103;
       $$pre$phi91Z2D = $$pre$phi89Z2D;
       label = 95;
       break;
      } else {
       $$pr45$pr$pre = HEAP32[$14>>2]|0;
       $$pr45$pr = $$pr45$pr$pre;
       label = 93;
       break;
      }
     }
    } while(0);
    if ((label|0) == 93) {
     $505 = ($$pr45$pr|0)==(103);
     if ($505) {
      $$pre90 = ((($2)) + 28|0);
      $$pre$phi91Z2D = $$pre90;
      label = 95;
     }
    }
    do {
     if ((label|0) == 95) {
      $506 = HEAP32[$$pre$phi91Z2D>>2]|0;
      $507 = ((($506)) + 44|0);
      $508 = HEAP32[$507>>2]|0;
      $509 = ($508|0)==(0);
      if ($509) {
       HEAP32[$14>>2] = 113;
       break;
      }
      $510 = ((($2)) + 20|0);
      $511 = HEAP32[$510>>2]|0;
      $512 = (($511) + 2)|0;
      $513 = ((($2)) + 12|0);
      $514 = HEAP32[$513>>2]|0;
      $515 = ($512>>>0)>($514>>>0);
      if ($515) {
       $516 = HEAP32[$1>>2]|0;
       __tr_flush_bits($516);
       $517 = ((($516)) + 20|0);
       $518 = HEAP32[$517>>2]|0;
       $519 = HEAP32[$20>>2]|0;
       $520 = ($518>>>0)>($519>>>0);
       $$$i28 = $520 ? $519 : $518;
       $521 = ($$$i28|0)==(0);
       if (!($521)) {
        $522 = HEAP32[$6>>2]|0;
        $523 = ((($516)) + 16|0);
        $524 = HEAP32[$523>>2]|0;
        _memcpy(($522|0),($524|0),($$$i28|0))|0;
        $525 = HEAP32[$6>>2]|0;
        $526 = (($525) + ($$$i28)|0);
        HEAP32[$6>>2] = $526;
        $527 = HEAP32[$523>>2]|0;
        $528 = (($527) + ($$$i28)|0);
        HEAP32[$523>>2] = $528;
        $529 = ((($strm)) + 20|0);
        $530 = HEAP32[$529>>2]|0;
        $531 = (($530) + ($$$i28))|0;
        HEAP32[$529>>2] = $531;
        $532 = HEAP32[$20>>2]|0;
        $533 = (($532) - ($$$i28))|0;
        HEAP32[$20>>2] = $533;
        $534 = HEAP32[$517>>2]|0;
        $535 = (($534) - ($$$i28))|0;
        HEAP32[$517>>2] = $535;
        $536 = ($534|0)==($$$i28|0);
        if ($536) {
         $537 = ((($516)) + 8|0);
         $538 = HEAP32[$537>>2]|0;
         HEAP32[$523>>2] = $538;
        }
       }
      }
      $539 = HEAP32[$510>>2]|0;
      $540 = (($539) + 2)|0;
      $541 = HEAP32[$513>>2]|0;
      $542 = ($540>>>0)>($541>>>0);
      if (!($542)) {
       $543 = ((($strm)) + 48|0);
       $544 = HEAP32[$543>>2]|0;
       $545 = $544&255;
       $546 = (($539) + 1)|0;
       HEAP32[$510>>2] = $546;
       $547 = ((($2)) + 8|0);
       $548 = HEAP32[$547>>2]|0;
       $549 = (($548) + ($539)|0);
       HEAP8[$549>>0] = $545;
       $550 = HEAP32[$543>>2]|0;
       $551 = $550 >>> 8;
       $552 = $551&255;
       $553 = HEAP32[$510>>2]|0;
       $554 = (($553) + 1)|0;
       HEAP32[$510>>2] = $554;
       $555 = HEAP32[$547>>2]|0;
       $556 = (($555) + ($553)|0);
       HEAP8[$556>>0] = $552;
       $557 = (_crc32(0,0,0)|0);
       HEAP32[$543>>2] = $557;
       HEAP32[$14>>2] = 113;
      }
     }
    } while(0);
    $558 = ((($2)) + 20|0);
    $559 = HEAP32[$558>>2]|0;
    $560 = ($559|0)==(0);
    if ($560) {
     $586 = ((($strm)) + 4|0);
     $587 = HEAP32[$586>>2]|0;
     $588 = ($587|0)==(0);
     if ($588) {
      $589 = $flush << 1;
      $590 = ($flush|0)>(4);
      $591 = $590 ? 9 : 0;
      $592 = (($589) - ($591))|0;
      $593 = $26 << 1;
      $594 = ($26|0)>(4);
      $595 = $594 ? 9 : 0;
      $596 = (($593) - ($595))|0;
      $597 = ($592|0)<=($596|0);
      $or$cond7 = $17 & $597;
      if ($or$cond7) {
       $598 = HEAP32[(141768)>>2]|0;
       $599 = ((($strm)) + 24|0);
       HEAP32[$599>>2] = $598;
       $$1 = -5;
       return ($$1|0);
      }
     }
    } else {
     $561 = HEAP32[$1>>2]|0;
     __tr_flush_bits($561);
     $562 = ((($561)) + 20|0);
     $563 = HEAP32[$562>>2]|0;
     $564 = HEAP32[$20>>2]|0;
     $565 = ($563>>>0)>($564>>>0);
     $$$i30 = $565 ? $564 : $563;
     $566 = ($$$i30|0)==(0);
     if ($566) {
      $585 = $564;
     } else {
      $567 = HEAP32[$6>>2]|0;
      $568 = ((($561)) + 16|0);
      $569 = HEAP32[$568>>2]|0;
      _memcpy(($567|0),($569|0),($$$i30|0))|0;
      $570 = HEAP32[$6>>2]|0;
      $571 = (($570) + ($$$i30)|0);
      HEAP32[$6>>2] = $571;
      $572 = HEAP32[$568>>2]|0;
      $573 = (($572) + ($$$i30)|0);
      HEAP32[$568>>2] = $573;
      $574 = ((($strm)) + 20|0);
      $575 = HEAP32[$574>>2]|0;
      $576 = (($575) + ($$$i30))|0;
      HEAP32[$574>>2] = $576;
      $577 = HEAP32[$20>>2]|0;
      $578 = (($577) - ($$$i30))|0;
      HEAP32[$20>>2] = $578;
      $579 = HEAP32[$562>>2]|0;
      $580 = (($579) - ($$$i30))|0;
      HEAP32[$562>>2] = $580;
      $581 = ($579|0)==($$$i30|0);
      if ($581) {
       $582 = ((($561)) + 8|0);
       $583 = HEAP32[$582>>2]|0;
       HEAP32[$568>>2] = $583;
       $585 = $578;
      } else {
       $585 = $578;
      }
     }
     $584 = ($585|0)==(0);
     if ($584) {
      HEAP32[$25>>2] = -1;
      $$1 = 0;
      return ($$1|0);
     }
    }
    $600 = HEAP32[$14>>2]|0;
    $601 = ($600|0)==(666);
    $602 = ((($strm)) + 4|0);
    $603 = HEAP32[$602>>2]|0;
    $604 = ($603|0)==(0);
    if ($601) {
     if ($604) {
      label = 116;
     } else {
      $605 = HEAP32[(141768)>>2]|0;
      $606 = ((($strm)) + 24|0);
      HEAP32[$606>>2] = $605;
      $$1 = -5;
      return ($$1|0);
     }
    } else {
     if ($604) {
      label = 116;
     } else {
      label = 117;
     }
    }
    if ((label|0) == 116) {
     $607 = ((($2)) + 116|0);
     $608 = HEAP32[$607>>2]|0;
     $609 = ($608|0)==(0);
     $610 = ($flush|0)==(0);
     $brmerge = $610 | $601;
     $or$cond97 = $609 & $brmerge;
     if (!($or$cond97)) {
      label = 117;
     }
    }
    do {
     if ((label|0) == 117) {
      $611 = ((($2)) + 136|0);
      $612 = HEAP32[$611>>2]|0;
      L173: do {
       switch ($612|0) {
       case 2:  {
        $613 = ((($2)) + 116|0);
        $614 = ((($2)) + 96|0);
        $615 = ((($2)) + 108|0);
        $616 = ((($2)) + 56|0);
        $617 = ((($2)) + 5792|0);
        $618 = ((($2)) + 5796|0);
        $619 = ((($2)) + 5784|0);
        $620 = ((($2)) + 5788|0);
        $621 = ((($2)) + 92|0);
        while(1) {
         $622 = HEAP32[$613>>2]|0;
         $623 = ($622|0)==(0);
         if ($623) {
          _fill_window($2);
          $624 = HEAP32[$613>>2]|0;
          $625 = ($624|0)==(0);
          if ($625) {
           break;
          }
         }
         HEAP32[$614>>2] = 0;
         $627 = HEAP32[$615>>2]|0;
         $628 = HEAP32[$616>>2]|0;
         $629 = (($628) + ($627)|0);
         $630 = HEAP8[$629>>0]|0;
         $631 = HEAP32[$617>>2]|0;
         $632 = HEAP32[$618>>2]|0;
         $633 = (($632) + ($631<<1)|0);
         HEAP16[$633>>1] = 0;
         $634 = (($631) + 1)|0;
         HEAP32[$617>>2] = $634;
         $635 = HEAP32[$619>>2]|0;
         $636 = (($635) + ($631)|0);
         HEAP8[$636>>0] = $630;
         $637 = $630&255;
         $638 = (((($2)) + 148|0) + ($637<<2)|0);
         $639 = HEAP16[$638>>1]|0;
         $640 = (($639) + 1)<<16>>16;
         HEAP16[$638>>1] = $640;
         $641 = HEAP32[$617>>2]|0;
         $642 = HEAP32[$620>>2]|0;
         $643 = (($642) + -1)|0;
         $644 = ($641|0)==($643|0);
         $645 = HEAP32[$613>>2]|0;
         $646 = (($645) + -1)|0;
         HEAP32[$613>>2] = $646;
         $647 = HEAP32[$615>>2]|0;
         $648 = (($647) + 1)|0;
         HEAP32[$615>>2] = $648;
         if (!($644)) {
          continue;
         }
         $649 = HEAP32[$621>>2]|0;
         $650 = ($649|0)>(-1);
         if ($650) {
          $651 = HEAP32[$616>>2]|0;
          $652 = (($651) + ($649)|0);
          $654 = $652;
         } else {
          $654 = 0;
         }
         $653 = (($648) - ($649))|0;
         __tr_flush_block($2,$654,$653,0);
         $655 = HEAP32[$615>>2]|0;
         HEAP32[$621>>2] = $655;
         $656 = HEAP32[$2>>2]|0;
         $657 = ((($656)) + 28|0);
         $658 = HEAP32[$657>>2]|0;
         __tr_flush_bits($658);
         $659 = ((($658)) + 20|0);
         $660 = HEAP32[$659>>2]|0;
         $661 = ((($656)) + 16|0);
         $662 = HEAP32[$661>>2]|0;
         $663 = ($660>>>0)>($662>>>0);
         $$$i$i = $663 ? $662 : $660;
         $664 = ($$$i$i|0)==(0);
         if (!($664)) {
          $665 = ((($656)) + 12|0);
          $666 = HEAP32[$665>>2]|0;
          $667 = ((($658)) + 16|0);
          $668 = HEAP32[$667>>2]|0;
          _memcpy(($666|0),($668|0),($$$i$i|0))|0;
          $669 = HEAP32[$665>>2]|0;
          $670 = (($669) + ($$$i$i)|0);
          HEAP32[$665>>2] = $670;
          $671 = HEAP32[$667>>2]|0;
          $672 = (($671) + ($$$i$i)|0);
          HEAP32[$667>>2] = $672;
          $673 = ((($656)) + 20|0);
          $674 = HEAP32[$673>>2]|0;
          $675 = (($674) + ($$$i$i))|0;
          HEAP32[$673>>2] = $675;
          $676 = HEAP32[$661>>2]|0;
          $677 = (($676) - ($$$i$i))|0;
          HEAP32[$661>>2] = $677;
          $678 = HEAP32[$659>>2]|0;
          $679 = (($678) - ($$$i$i))|0;
          HEAP32[$659>>2] = $679;
          $680 = ($678|0)==($$$i$i|0);
          if ($680) {
           $681 = ((($658)) + 8|0);
           $682 = HEAP32[$681>>2]|0;
           HEAP32[$667>>2] = $682;
          }
         }
         $683 = HEAP32[$2>>2]|0;
         $684 = ((($683)) + 16|0);
         $685 = HEAP32[$684>>2]|0;
         $686 = ($685|0)==(0);
         if ($686) {
          label = 191;
          break L173;
         }
        }
        $626 = ($flush|0)==(0);
        if ($626) {
         label = 191;
        } else {
         $687 = ((($2)) + 5812|0);
         HEAP32[$687>>2] = 0;
         $688 = ($flush|0)==(4);
         if ($688) {
          $689 = HEAP32[$621>>2]|0;
          $690 = ($689|0)>(-1);
          if ($690) {
           $691 = HEAP32[$616>>2]|0;
           $692 = (($691) + ($689)|0);
           $695 = $692;
          } else {
           $695 = 0;
          }
          $693 = HEAP32[$615>>2]|0;
          $694 = (($693) - ($689))|0;
          __tr_flush_block($2,$695,$694,1);
          $696 = HEAP32[$615>>2]|0;
          HEAP32[$621>>2] = $696;
          $697 = HEAP32[$2>>2]|0;
          $698 = ((($697)) + 28|0);
          $699 = HEAP32[$698>>2]|0;
          __tr_flush_bits($699);
          $700 = ((($699)) + 20|0);
          $701 = HEAP32[$700>>2]|0;
          $702 = ((($697)) + 16|0);
          $703 = HEAP32[$702>>2]|0;
          $704 = ($701>>>0)>($703>>>0);
          $$$i1$i = $704 ? $703 : $701;
          $705 = ($$$i1$i|0)==(0);
          if (!($705)) {
           $706 = ((($697)) + 12|0);
           $707 = HEAP32[$706>>2]|0;
           $708 = ((($699)) + 16|0);
           $709 = HEAP32[$708>>2]|0;
           _memcpy(($707|0),($709|0),($$$i1$i|0))|0;
           $710 = HEAP32[$706>>2]|0;
           $711 = (($710) + ($$$i1$i)|0);
           HEAP32[$706>>2] = $711;
           $712 = HEAP32[$708>>2]|0;
           $713 = (($712) + ($$$i1$i)|0);
           HEAP32[$708>>2] = $713;
           $714 = ((($697)) + 20|0);
           $715 = HEAP32[$714>>2]|0;
           $716 = (($715) + ($$$i1$i))|0;
           HEAP32[$714>>2] = $716;
           $717 = HEAP32[$702>>2]|0;
           $718 = (($717) - ($$$i1$i))|0;
           HEAP32[$702>>2] = $718;
           $719 = HEAP32[$700>>2]|0;
           $720 = (($719) - ($$$i1$i))|0;
           HEAP32[$700>>2] = $720;
           $721 = ($719|0)==($$$i1$i|0);
           if ($721) {
            $722 = ((($699)) + 8|0);
            $723 = HEAP32[$722>>2]|0;
            HEAP32[$708>>2] = $723;
           }
          }
          $724 = HEAP32[$2>>2]|0;
          $725 = ((($724)) + 16|0);
          $726 = HEAP32[$725>>2]|0;
          $727 = ($726|0)==(0);
          $$$i32 = $727 ? 2 : 3;
          $1017 = $$$i32;
          label = 188;
          break L173;
         }
         $728 = HEAP32[$617>>2]|0;
         $729 = ($728|0)==(0);
         if (!($729)) {
          $730 = HEAP32[$621>>2]|0;
          $731 = ($730|0)>(-1);
          if ($731) {
           $732 = HEAP32[$616>>2]|0;
           $733 = (($732) + ($730)|0);
           $736 = $733;
          } else {
           $736 = 0;
          }
          $734 = HEAP32[$615>>2]|0;
          $735 = (($734) - ($730))|0;
          __tr_flush_block($2,$736,$735,0);
          $737 = HEAP32[$615>>2]|0;
          HEAP32[$621>>2] = $737;
          $738 = HEAP32[$2>>2]|0;
          $739 = ((($738)) + 28|0);
          $740 = HEAP32[$739>>2]|0;
          __tr_flush_bits($740);
          $741 = ((($740)) + 20|0);
          $742 = HEAP32[$741>>2]|0;
          $743 = ((($738)) + 16|0);
          $744 = HEAP32[$743>>2]|0;
          $745 = ($742>>>0)>($744>>>0);
          $$$i3$i = $745 ? $744 : $742;
          $746 = ($$$i3$i|0)==(0);
          if (!($746)) {
           $747 = ((($738)) + 12|0);
           $748 = HEAP32[$747>>2]|0;
           $749 = ((($740)) + 16|0);
           $750 = HEAP32[$749>>2]|0;
           _memcpy(($748|0),($750|0),($$$i3$i|0))|0;
           $751 = HEAP32[$747>>2]|0;
           $752 = (($751) + ($$$i3$i)|0);
           HEAP32[$747>>2] = $752;
           $753 = HEAP32[$749>>2]|0;
           $754 = (($753) + ($$$i3$i)|0);
           HEAP32[$749>>2] = $754;
           $755 = ((($738)) + 20|0);
           $756 = HEAP32[$755>>2]|0;
           $757 = (($756) + ($$$i3$i))|0;
           HEAP32[$755>>2] = $757;
           $758 = HEAP32[$743>>2]|0;
           $759 = (($758) - ($$$i3$i))|0;
           HEAP32[$743>>2] = $759;
           $760 = HEAP32[$741>>2]|0;
           $761 = (($760) - ($$$i3$i))|0;
           HEAP32[$741>>2] = $761;
           $762 = ($760|0)==($$$i3$i|0);
           if ($762) {
            $763 = ((($740)) + 8|0);
            $764 = HEAP32[$763>>2]|0;
            HEAP32[$749>>2] = $764;
           }
          }
          $765 = HEAP32[$2>>2]|0;
          $766 = ((($765)) + 16|0);
          $767 = HEAP32[$766>>2]|0;
          $768 = ($767|0)==(0);
          if ($768) {
           label = 191;
          }
         }
        }
        break;
       }
       case 3:  {
        $769 = ((($2)) + 116|0);
        $770 = ($flush|0)==(0);
        $771 = ((($2)) + 96|0);
        $772 = ((($2)) + 108|0);
        $773 = ((($2)) + 5792|0);
        $774 = ((($2)) + 5796|0);
        $775 = ((($2)) + 5784|0);
        $776 = ((($2)) + 5788|0);
        $777 = ((($2)) + 56|0);
        $778 = ((($2)) + 92|0);
        while(1) {
         $779 = HEAP32[$769>>2]|0;
         $780 = ($779>>>0)<(259);
         if ($780) {
          _fill_window($2);
          $781 = HEAP32[$769>>2]|0;
          $782 = ($781>>>0)<(259);
          $or$cond$i = $770 & $782;
          if ($or$cond$i) {
           label = 191;
           break L173;
          }
          $783 = ($781|0)==(0);
          if ($783) {
           break;
          }
          HEAP32[$771>>2] = 0;
          $784 = ($781>>>0)>(2);
          if ($784) {
           $832 = $781;
           label = 150;
          } else {
           $$pre$i = HEAP32[$772>>2]|0;
           $868 = $$pre$i;
           label = 165;
          }
         } else {
          HEAP32[$771>>2] = 0;
          $832 = $779;
          label = 150;
         }
         if ((label|0) == 150) {
          label = 0;
          $785 = HEAP32[$772>>2]|0;
          $786 = ($785|0)==(0);
          if ($786) {
           $868 = 0;
           label = 165;
          } else {
           $787 = HEAP32[$777>>2]|0;
           $788 = (($787) + ($785)|0);
           $789 = ((($788)) + -1|0);
           $790 = HEAP8[$789>>0]|0;
           $791 = ((($789)) + 1|0);
           $792 = HEAP8[$791>>0]|0;
           $793 = ($790<<24>>24)==($792<<24>>24);
           if ($793) {
            $794 = ((($791)) + 1|0);
            $795 = HEAP8[$794>>0]|0;
            $796 = ($790<<24>>24)==($795<<24>>24);
            if ($796) {
             $797 = ((($794)) + 1|0);
             $798 = HEAP8[$797>>0]|0;
             $799 = ($790<<24>>24)==($798<<24>>24);
             if ($799) {
              $800 = ((($788)) + 258|0);
              $802 = $794;
              while(1) {
               $801 = ((($802)) + 2|0);
               $803 = HEAP8[$801>>0]|0;
               $804 = ($790<<24>>24)==($803<<24>>24);
               if (!($804)) {
                $scan$1$i = $801;
                break;
               }
               $805 = ((($802)) + 3|0);
               $806 = HEAP8[$805>>0]|0;
               $807 = ($790<<24>>24)==($806<<24>>24);
               if (!($807)) {
                $scan$1$i = $805;
                break;
               }
               $808 = ((($802)) + 4|0);
               $809 = HEAP8[$808>>0]|0;
               $810 = ($790<<24>>24)==($809<<24>>24);
               if (!($810)) {
                $scan$1$i = $808;
                break;
               }
               $811 = ((($802)) + 5|0);
               $812 = HEAP8[$811>>0]|0;
               $813 = ($790<<24>>24)==($812<<24>>24);
               if (!($813)) {
                $scan$1$i = $811;
                break;
               }
               $814 = ((($802)) + 6|0);
               $815 = HEAP8[$814>>0]|0;
               $816 = ($790<<24>>24)==($815<<24>>24);
               if (!($816)) {
                $scan$1$i = $814;
                break;
               }
               $817 = ((($802)) + 7|0);
               $818 = HEAP8[$817>>0]|0;
               $819 = ($790<<24>>24)==($818<<24>>24);
               if (!($819)) {
                $scan$1$i = $817;
                break;
               }
               $820 = ((($802)) + 8|0);
               $821 = HEAP8[$820>>0]|0;
               $822 = ($790<<24>>24)==($821<<24>>24);
               if (!($822)) {
                $scan$1$i = $820;
                break;
               }
               $823 = ((($802)) + 9|0);
               $824 = HEAP8[$823>>0]|0;
               $825 = ($790<<24>>24)==($824<<24>>24);
               $826 = ($823>>>0)<($800>>>0);
               $or$cond9$i = $826 & $825;
               if ($or$cond9$i) {
                $802 = $820;
               } else {
                $scan$1$i = $823;
                break;
               }
              }
              $827 = $800;
              $828 = $scan$1$i;
              $829 = (($828) - ($827))|0;
              $830 = (($829) + 258)|0;
              $831 = ($830>>>0)>($832>>>0);
              $$$i34 = $831 ? $832 : $830;
              HEAP32[$771>>2] = $$$i34;
              $833 = ($$$i34>>>0)>(2);
              if ($833) {
               $834 = (($$$i34) + 253)|0;
               $835 = $834&255;
               $836 = HEAP32[$773>>2]|0;
               $837 = HEAP32[$774>>2]|0;
               $838 = (($837) + ($836<<1)|0);
               HEAP16[$838>>1] = 1;
               $839 = (($836) + 1)|0;
               HEAP32[$773>>2] = $839;
               $840 = HEAP32[$775>>2]|0;
               $841 = (($840) + ($836)|0);
               HEAP8[$841>>0] = $835;
               $842 = $834 & 255;
               $843 = (613302 + ($842)|0);
               $844 = HEAP8[$843>>0]|0;
               $845 = $844&255;
               $846 = $845 | 256;
               $847 = (($846) + 1)|0;
               $848 = (((($2)) + 148|0) + ($847<<2)|0);
               $849 = HEAP16[$848>>1]|0;
               $850 = (($849) + 1)<<16>>16;
               HEAP16[$848>>1] = $850;
               $851 = HEAP8[612790]|0;
               $852 = $851&255;
               $853 = (((($2)) + 2440|0) + ($852<<2)|0);
               $854 = HEAP16[$853>>1]|0;
               $855 = (($854) + 1)<<16>>16;
               HEAP16[$853>>1] = $855;
               $856 = HEAP32[$773>>2]|0;
               $857 = HEAP32[$776>>2]|0;
               $858 = (($857) + -1)|0;
               $859 = ($856|0)==($858|0);
               $860 = $859&1;
               $861 = HEAP32[$771>>2]|0;
               $862 = HEAP32[$769>>2]|0;
               $863 = (($862) - ($861))|0;
               HEAP32[$769>>2] = $863;
               $864 = HEAP32[$772>>2]|0;
               $865 = (($864) + ($861))|0;
               HEAP32[$772>>2] = $865;
               HEAP32[$771>>2] = 0;
               $895 = $865;$bflush$0$i = $860;
              } else {
               $868 = $785;
               label = 165;
              }
             } else {
              $868 = $785;
              label = 165;
             }
            } else {
             $868 = $785;
             label = 165;
            }
           } else {
            $868 = $785;
            label = 165;
           }
          }
         }
         if ((label|0) == 165) {
          label = 0;
          $866 = HEAP32[$777>>2]|0;
          $867 = (($866) + ($868)|0);
          $869 = HEAP8[$867>>0]|0;
          $870 = HEAP32[$773>>2]|0;
          $871 = HEAP32[$774>>2]|0;
          $872 = (($871) + ($870<<1)|0);
          HEAP16[$872>>1] = 0;
          $873 = (($870) + 1)|0;
          HEAP32[$773>>2] = $873;
          $874 = HEAP32[$775>>2]|0;
          $875 = (($874) + ($870)|0);
          HEAP8[$875>>0] = $869;
          $876 = $869&255;
          $877 = (((($2)) + 148|0) + ($876<<2)|0);
          $878 = HEAP16[$877>>1]|0;
          $879 = (($878) + 1)<<16>>16;
          HEAP16[$877>>1] = $879;
          $880 = HEAP32[$773>>2]|0;
          $881 = HEAP32[$776>>2]|0;
          $882 = (($881) + -1)|0;
          $883 = ($880|0)==($882|0);
          $884 = $883&1;
          $885 = HEAP32[$769>>2]|0;
          $886 = (($885) + -1)|0;
          HEAP32[$769>>2] = $886;
          $887 = HEAP32[$772>>2]|0;
          $888 = (($887) + 1)|0;
          HEAP32[$772>>2] = $888;
          $895 = $888;$bflush$0$i = $884;
         }
         $889 = ($bflush$0$i|0)==(0);
         if ($889) {
          continue;
         }
         $890 = HEAP32[$778>>2]|0;
         $891 = ($890|0)>(-1);
         if ($891) {
          $892 = HEAP32[$777>>2]|0;
          $893 = (($892) + ($890)|0);
          $896 = $893;
         } else {
          $896 = 0;
         }
         $894 = (($895) - ($890))|0;
         __tr_flush_block($2,$896,$894,0);
         $897 = HEAP32[$772>>2]|0;
         HEAP32[$778>>2] = $897;
         $898 = HEAP32[$2>>2]|0;
         $899 = ((($898)) + 28|0);
         $900 = HEAP32[$899>>2]|0;
         __tr_flush_bits($900);
         $901 = ((($900)) + 20|0);
         $902 = HEAP32[$901>>2]|0;
         $903 = ((($898)) + 16|0);
         $904 = HEAP32[$903>>2]|0;
         $905 = ($902>>>0)>($904>>>0);
         $$$i$i35 = $905 ? $904 : $902;
         $906 = ($$$i$i35|0)==(0);
         if (!($906)) {
          $907 = ((($898)) + 12|0);
          $908 = HEAP32[$907>>2]|0;
          $909 = ((($900)) + 16|0);
          $910 = HEAP32[$909>>2]|0;
          _memcpy(($908|0),($910|0),($$$i$i35|0))|0;
          $911 = HEAP32[$907>>2]|0;
          $912 = (($911) + ($$$i$i35)|0);
          HEAP32[$907>>2] = $912;
          $913 = HEAP32[$909>>2]|0;
          $914 = (($913) + ($$$i$i35)|0);
          HEAP32[$909>>2] = $914;
          $915 = ((($898)) + 20|0);
          $916 = HEAP32[$915>>2]|0;
          $917 = (($916) + ($$$i$i35))|0;
          HEAP32[$915>>2] = $917;
          $918 = HEAP32[$903>>2]|0;
          $919 = (($918) - ($$$i$i35))|0;
          HEAP32[$903>>2] = $919;
          $920 = HEAP32[$901>>2]|0;
          $921 = (($920) - ($$$i$i35))|0;
          HEAP32[$901>>2] = $921;
          $922 = ($920|0)==($$$i$i35|0);
          if ($922) {
           $923 = ((($900)) + 8|0);
           $924 = HEAP32[$923>>2]|0;
           HEAP32[$909>>2] = $924;
          }
         }
         $925 = HEAP32[$2>>2]|0;
         $926 = ((($925)) + 16|0);
         $927 = HEAP32[$926>>2]|0;
         $928 = ($927|0)==(0);
         if ($928) {
          label = 191;
          break L173;
         }
        }
        $929 = ((($2)) + 5812|0);
        HEAP32[$929>>2] = 0;
        $930 = ($flush|0)==(4);
        if ($930) {
         $931 = HEAP32[$778>>2]|0;
         $932 = ($931|0)>(-1);
         if ($932) {
          $933 = HEAP32[$777>>2]|0;
          $934 = (($933) + ($931)|0);
          $937 = $934;
         } else {
          $937 = 0;
         }
         $935 = HEAP32[$772>>2]|0;
         $936 = (($935) - ($931))|0;
         __tr_flush_block($2,$937,$936,1);
         $938 = HEAP32[$772>>2]|0;
         HEAP32[$778>>2] = $938;
         $939 = HEAP32[$2>>2]|0;
         $940 = ((($939)) + 28|0);
         $941 = HEAP32[$940>>2]|0;
         __tr_flush_bits($941);
         $942 = ((($941)) + 20|0);
         $943 = HEAP32[$942>>2]|0;
         $944 = ((($939)) + 16|0);
         $945 = HEAP32[$944>>2]|0;
         $946 = ($943>>>0)>($945>>>0);
         $$$i11$i = $946 ? $945 : $943;
         $947 = ($$$i11$i|0)==(0);
         if (!($947)) {
          $948 = ((($939)) + 12|0);
          $949 = HEAP32[$948>>2]|0;
          $950 = ((($941)) + 16|0);
          $951 = HEAP32[$950>>2]|0;
          _memcpy(($949|0),($951|0),($$$i11$i|0))|0;
          $952 = HEAP32[$948>>2]|0;
          $953 = (($952) + ($$$i11$i)|0);
          HEAP32[$948>>2] = $953;
          $954 = HEAP32[$950>>2]|0;
          $955 = (($954) + ($$$i11$i)|0);
          HEAP32[$950>>2] = $955;
          $956 = ((($939)) + 20|0);
          $957 = HEAP32[$956>>2]|0;
          $958 = (($957) + ($$$i11$i))|0;
          HEAP32[$956>>2] = $958;
          $959 = HEAP32[$944>>2]|0;
          $960 = (($959) - ($$$i11$i))|0;
          HEAP32[$944>>2] = $960;
          $961 = HEAP32[$942>>2]|0;
          $962 = (($961) - ($$$i11$i))|0;
          HEAP32[$942>>2] = $962;
          $963 = ($961|0)==($$$i11$i|0);
          if ($963) {
           $964 = ((($941)) + 8|0);
           $965 = HEAP32[$964>>2]|0;
           HEAP32[$950>>2] = $965;
          }
         }
         $966 = HEAP32[$2>>2]|0;
         $967 = ((($966)) + 16|0);
         $968 = HEAP32[$967>>2]|0;
         $969 = ($968|0)==(0);
         $$10$i = $969 ? 2 : 3;
         $1017 = $$10$i;
         label = 188;
         break L173;
        }
        $970 = HEAP32[$773>>2]|0;
        $971 = ($970|0)==(0);
        if (!($971)) {
         $972 = HEAP32[$778>>2]|0;
         $973 = ($972|0)>(-1);
         if ($973) {
          $974 = HEAP32[$777>>2]|0;
          $975 = (($974) + ($972)|0);
          $978 = $975;
         } else {
          $978 = 0;
         }
         $976 = HEAP32[$772>>2]|0;
         $977 = (($976) - ($972))|0;
         __tr_flush_block($2,$978,$977,0);
         $979 = HEAP32[$772>>2]|0;
         HEAP32[$778>>2] = $979;
         $980 = HEAP32[$2>>2]|0;
         $981 = ((($980)) + 28|0);
         $982 = HEAP32[$981>>2]|0;
         __tr_flush_bits($982);
         $983 = ((($982)) + 20|0);
         $984 = HEAP32[$983>>2]|0;
         $985 = ((($980)) + 16|0);
         $986 = HEAP32[$985>>2]|0;
         $987 = ($984>>>0)>($986>>>0);
         $$$i13$i = $987 ? $986 : $984;
         $988 = ($$$i13$i|0)==(0);
         if (!($988)) {
          $989 = ((($980)) + 12|0);
          $990 = HEAP32[$989>>2]|0;
          $991 = ((($982)) + 16|0);
          $992 = HEAP32[$991>>2]|0;
          _memcpy(($990|0),($992|0),($$$i13$i|0))|0;
          $993 = HEAP32[$989>>2]|0;
          $994 = (($993) + ($$$i13$i)|0);
          HEAP32[$989>>2] = $994;
          $995 = HEAP32[$991>>2]|0;
          $996 = (($995) + ($$$i13$i)|0);
          HEAP32[$991>>2] = $996;
          $997 = ((($980)) + 20|0);
          $998 = HEAP32[$997>>2]|0;
          $999 = (($998) + ($$$i13$i))|0;
          HEAP32[$997>>2] = $999;
          $1000 = HEAP32[$985>>2]|0;
          $1001 = (($1000) - ($$$i13$i))|0;
          HEAP32[$985>>2] = $1001;
          $1002 = HEAP32[$983>>2]|0;
          $1003 = (($1002) - ($$$i13$i))|0;
          HEAP32[$983>>2] = $1003;
          $1004 = ($1002|0)==($$$i13$i|0);
          if ($1004) {
           $1005 = ((($982)) + 8|0);
           $1006 = HEAP32[$1005>>2]|0;
           HEAP32[$991>>2] = $1006;
          }
         }
         $1007 = HEAP32[$2>>2]|0;
         $1008 = ((($1007)) + 16|0);
         $1009 = HEAP32[$1008>>2]|0;
         $1010 = ($1009|0)==(0);
         if ($1010) {
          label = 191;
         }
        }
        break;
       }
       default: {
        $1011 = ((($2)) + 132|0);
        $1012 = HEAP32[$1011>>2]|0;
        $1013 = (((141128 + (($1012*12)|0)|0)) + 8|0);
        $1014 = HEAP32[$1013>>2]|0;
        $1015 = (FUNCTION_TABLE_iii[$1014 & 8191]($2,$flush)|0);
        $1017 = $1015;
        label = 188;
       }
       }
      } while(0);
      if ((label|0) == 188) {
       $1016 = $1017 | 1;
       $1018 = ($1016|0)==(3);
       if ($1018) {
        HEAP32[$14>>2] = 666;
       }
       $1019 = $1017 | 2;
       $1020 = ($1019|0)==(2);
       if ($1020) {
        label = 191;
       } else {
        $1023 = ($1017|0)==(1);
        if (!($1023)) {
         break;
        }
       }
      }
      if ((label|0) == 191) {
       $1021 = HEAP32[$20>>2]|0;
       $1022 = ($1021|0)==(0);
       if (!($1022)) {
        $$1 = 0;
        return ($$1|0);
       }
       HEAP32[$25>>2] = -1;
       $$1 = 0;
       return ($$1|0);
      }
      switch ($flush|0) {
      case 1:  {
       __tr_align($2);
       break;
      }
      case 5:  {
       break;
      }
      default: {
       __tr_stored_block($2,0,0,0);
       $1024 = ($flush|0)==(3);
       if ($1024) {
        $1025 = ((($2)) + 76|0);
        $1026 = HEAP32[$1025>>2]|0;
        $1027 = (($1026) + -1)|0;
        $1028 = ((($2)) + 68|0);
        $1029 = HEAP32[$1028>>2]|0;
        $1030 = (($1029) + ($1027<<1)|0);
        HEAP16[$1030>>1] = 0;
        $1031 = $1026 << 1;
        $1032 = (($1031) + -2)|0;
        _memset(($1029|0),0,($1032|0))|0;
        $1033 = ((($2)) + 116|0);
        $1034 = HEAP32[$1033>>2]|0;
        $1035 = ($1034|0)==(0);
        if ($1035) {
         $1036 = ((($2)) + 108|0);
         HEAP32[$1036>>2] = 0;
         $1037 = ((($2)) + 92|0);
         HEAP32[$1037>>2] = 0;
         $1038 = ((($2)) + 5812|0);
         HEAP32[$1038>>2] = 0;
        }
       }
      }
      }
      $1039 = HEAP32[$1>>2]|0;
      __tr_flush_bits($1039);
      $1040 = ((($1039)) + 20|0);
      $1041 = HEAP32[$1040>>2]|0;
      $1042 = HEAP32[$20>>2]|0;
      $1043 = ($1041>>>0)>($1042>>>0);
      $$$i20 = $1043 ? $1042 : $1041;
      $1044 = ($$$i20|0)==(0);
      if ($1044) {
       $1063 = $1042;
      } else {
       $1045 = HEAP32[$6>>2]|0;
       $1046 = ((($1039)) + 16|0);
       $1047 = HEAP32[$1046>>2]|0;
       _memcpy(($1045|0),($1047|0),($$$i20|0))|0;
       $1048 = HEAP32[$6>>2]|0;
       $1049 = (($1048) + ($$$i20)|0);
       HEAP32[$6>>2] = $1049;
       $1050 = HEAP32[$1046>>2]|0;
       $1051 = (($1050) + ($$$i20)|0);
       HEAP32[$1046>>2] = $1051;
       $1052 = ((($strm)) + 20|0);
       $1053 = HEAP32[$1052>>2]|0;
       $1054 = (($1053) + ($$$i20))|0;
       HEAP32[$1052>>2] = $1054;
       $1055 = HEAP32[$20>>2]|0;
       $1056 = (($1055) - ($$$i20))|0;
       HEAP32[$20>>2] = $1056;
       $1057 = HEAP32[$1040>>2]|0;
       $1058 = (($1057) - ($$$i20))|0;
       HEAP32[$1040>>2] = $1058;
       $1059 = ($1057|0)==($$$i20|0);
       if ($1059) {
        $1060 = ((($1039)) + 8|0);
        $1061 = HEAP32[$1060>>2]|0;
        HEAP32[$1046>>2] = $1061;
        $1063 = $1056;
       } else {
        $1063 = $1056;
       }
      }
      $1062 = ($1063|0)==(0);
      if ($1062) {
       HEAP32[$25>>2] = -1;
       $$1 = 0;
       return ($$1|0);
      }
     }
    } while(0);
    if ($17) {
     $$1 = 0;
     return ($$1|0);
    }
    $1064 = ((($2)) + 24|0);
    $1065 = HEAP32[$1064>>2]|0;
    $1066 = ($1065|0)<(1);
    if ($1066) {
     $$1 = 1;
     return ($$1|0);
    }
    $1067 = ($1065|0)==(2);
    $1068 = ((($strm)) + 48|0);
    $1069 = HEAP32[$1068>>2]|0;
    if ($1067) {
     $1070 = $1069&255;
     $1071 = HEAP32[$558>>2]|0;
     $1072 = (($1071) + 1)|0;
     HEAP32[$558>>2] = $1072;
     $1073 = ((($2)) + 8|0);
     $1074 = HEAP32[$1073>>2]|0;
     $1075 = (($1074) + ($1071)|0);
     HEAP8[$1075>>0] = $1070;
     $1076 = HEAP32[$1068>>2]|0;
     $1077 = $1076 >>> 8;
     $1078 = $1077&255;
     $1079 = HEAP32[$558>>2]|0;
     $1080 = (($1079) + 1)|0;
     HEAP32[$558>>2] = $1080;
     $1081 = HEAP32[$1073>>2]|0;
     $1082 = (($1081) + ($1079)|0);
     HEAP8[$1082>>0] = $1078;
     $1083 = HEAP32[$1068>>2]|0;
     $1084 = $1083 >>> 16;
     $1085 = $1084&255;
     $1086 = HEAP32[$558>>2]|0;
     $1087 = (($1086) + 1)|0;
     HEAP32[$558>>2] = $1087;
     $1088 = HEAP32[$1073>>2]|0;
     $1089 = (($1088) + ($1086)|0);
     HEAP8[$1089>>0] = $1085;
     $1090 = HEAP32[$1068>>2]|0;
     $1091 = $1090 >>> 24;
     $1092 = $1091&255;
     $1093 = HEAP32[$558>>2]|0;
     $1094 = (($1093) + 1)|0;
     HEAP32[$558>>2] = $1094;
     $1095 = HEAP32[$1073>>2]|0;
     $1096 = (($1095) + ($1093)|0);
     HEAP8[$1096>>0] = $1092;
     $1097 = ((($strm)) + 8|0);
     $1098 = HEAP32[$1097>>2]|0;
     $1099 = $1098&255;
     $1100 = HEAP32[$558>>2]|0;
     $1101 = (($1100) + 1)|0;
     HEAP32[$558>>2] = $1101;
     $1102 = HEAP32[$1073>>2]|0;
     $1103 = (($1102) + ($1100)|0);
     HEAP8[$1103>>0] = $1099;
     $1104 = HEAP32[$1097>>2]|0;
     $1105 = $1104 >>> 8;
     $1106 = $1105&255;
     $1107 = HEAP32[$558>>2]|0;
     $1108 = (($1107) + 1)|0;
     HEAP32[$558>>2] = $1108;
     $1109 = HEAP32[$1073>>2]|0;
     $1110 = (($1109) + ($1107)|0);
     HEAP8[$1110>>0] = $1106;
     $1111 = HEAP32[$1097>>2]|0;
     $1112 = $1111 >>> 16;
     $1113 = $1112&255;
     $1114 = HEAP32[$558>>2]|0;
     $1115 = (($1114) + 1)|0;
     HEAP32[$558>>2] = $1115;
     $1116 = HEAP32[$1073>>2]|0;
     $1117 = (($1116) + ($1114)|0);
     HEAP8[$1117>>0] = $1113;
     $1118 = HEAP32[$1097>>2]|0;
     $1119 = $1118 >>> 24;
     $1120 = $1119&255;
     $1121 = HEAP32[$558>>2]|0;
     $1122 = (($1121) + 1)|0;
     HEAP32[$558>>2] = $1122;
     $1123 = HEAP32[$1073>>2]|0;
     $1124 = (($1123) + ($1121)|0);
     HEAP8[$1124>>0] = $1120;
    } else {
     $1125 = $1069 >>> 16;
     $1126 = $1069 >>> 24;
     $1127 = $1126&255;
     $1128 = HEAP32[$558>>2]|0;
     $1129 = (($1128) + 1)|0;
     HEAP32[$558>>2] = $1129;
     $1130 = ((($2)) + 8|0);
     $1131 = HEAP32[$1130>>2]|0;
     $1132 = (($1131) + ($1128)|0);
     HEAP8[$1132>>0] = $1127;
     $1133 = $1125&255;
     $1134 = HEAP32[$558>>2]|0;
     $1135 = (($1134) + 1)|0;
     HEAP32[$558>>2] = $1135;
     $1136 = HEAP32[$1130>>2]|0;
     $1137 = (($1136) + ($1134)|0);
     HEAP8[$1137>>0] = $1133;
     $1138 = HEAP32[$1068>>2]|0;
     $1139 = $1138 >>> 8;
     $1140 = $1139&255;
     $1141 = HEAP32[$558>>2]|0;
     $1142 = (($1141) + 1)|0;
     HEAP32[$558>>2] = $1142;
     $1143 = HEAP32[$1130>>2]|0;
     $1144 = (($1143) + ($1141)|0);
     HEAP8[$1144>>0] = $1140;
     $1145 = $1138&255;
     $1146 = HEAP32[$558>>2]|0;
     $1147 = (($1146) + 1)|0;
     HEAP32[$558>>2] = $1147;
     $1148 = HEAP32[$1130>>2]|0;
     $1149 = (($1148) + ($1146)|0);
     HEAP8[$1149>>0] = $1145;
    }
    $1150 = HEAP32[$1>>2]|0;
    __tr_flush_bits($1150);
    $1151 = ((($1150)) + 20|0);
    $1152 = HEAP32[$1151>>2]|0;
    $1153 = HEAP32[$20>>2]|0;
    $1154 = ($1152>>>0)>($1153>>>0);
    $$$i = $1154 ? $1153 : $1152;
    $1155 = ($$$i|0)==(0);
    if (!($1155)) {
     $1156 = HEAP32[$6>>2]|0;
     $1157 = ((($1150)) + 16|0);
     $1158 = HEAP32[$1157>>2]|0;
     _memcpy(($1156|0),($1158|0),($$$i|0))|0;
     $1159 = HEAP32[$6>>2]|0;
     $1160 = (($1159) + ($$$i)|0);
     HEAP32[$6>>2] = $1160;
     $1161 = HEAP32[$1157>>2]|0;
     $1162 = (($1161) + ($$$i)|0);
     HEAP32[$1157>>2] = $1162;
     $1163 = ((($strm)) + 20|0);
     $1164 = HEAP32[$1163>>2]|0;
     $1165 = (($1164) + ($$$i))|0;
     HEAP32[$1163>>2] = $1165;
     $1166 = HEAP32[$20>>2]|0;
     $1167 = (($1166) - ($$$i))|0;
     HEAP32[$20>>2] = $1167;
     $1168 = HEAP32[$1151>>2]|0;
     $1169 = (($1168) - ($$$i))|0;
     HEAP32[$1151>>2] = $1169;
     $1170 = ($1168|0)==($$$i|0);
     if ($1170) {
      $1171 = ((($1150)) + 8|0);
      $1172 = HEAP32[$1171>>2]|0;
      HEAP32[$1157>>2] = $1172;
     }
    }
    $1173 = HEAP32[$1064>>2]|0;
    $1174 = ($1173|0)>(0);
    if ($1174) {
     $1175 = (0 - ($1173))|0;
     HEAP32[$1064>>2] = $1175;
    }
    $1176 = HEAP32[$558>>2]|0;
    $not$ = ($1176|0)==(0);
    $1177 = $not$&1;
    $$1 = $1177;
    return ($$1|0);
   }
  }
 } while(0);
 $18 = HEAP32[(141756)>>2]|0;
 $19 = ((($strm)) + 24|0);
 HEAP32[$19>>2] = $18;
 $$1 = -2;
 return ($$1|0);
}
function _inflateReset2($strm,$windowBits) {
 $strm = $strm|0;
 $windowBits = $windowBits|0;
 var $$0 = 0, $$02 = 0, $$pre = 0, $$pre4 = 0, $$windowBits = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $wrap$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($strm|0)==(0|0);
 if ($0) {
  $$0 = -2;
  return ($$0|0);
 }
 $1 = ((($strm)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $$0 = -2;
  return ($$0|0);
 }
 $4 = ($windowBits|0)<(0);
 if ($4) {
  $5 = (0 - ($windowBits))|0;
  $$02 = $5;$wrap$0 = 0;
 } else {
  $6 = $windowBits >> 4;
  $7 = (($6) + 1)|0;
  $8 = ($windowBits|0)<(48);
  $9 = $windowBits & 15;
  $$windowBits = $8 ? $9 : $windowBits;
  $$02 = $$windowBits;$wrap$0 = $7;
 }
 $10 = ($$02|0)==(0);
 $11 = $$02 & -8;
 $12 = ($11|0)==(8);
 $or$cond = $10 | $12;
 if (!($or$cond)) {
  $$0 = -2;
  return ($$0|0);
 }
 $13 = ((($2)) + 52|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ($14|0)==(0|0);
 $$pre4 = ((($2)) + 36|0);
 if ($15) {
  label = 9;
 } else {
  $16 = HEAP32[$$pre4>>2]|0;
  $17 = ($16|0)==($$02|0);
  if ($17) {
   label = 9;
  } else {
   $19 = ((($strm)) + 36|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = ((($strm)) + 40|0);
   $22 = HEAP32[$21>>2]|0;
   FUNCTION_TABLE_vii[$20 & 8191]($22,$14);
   HEAP32[$13>>2] = 0;
   $$pre = HEAP32[$1>>2]|0;
   $23 = ((($2)) + 8|0);
   HEAP32[$23>>2] = $wrap$0;
   HEAP32[$$pre4>>2] = $$02;
   $24 = ($$pre|0)==(0|0);
   if ($24) {
    $$0 = -2;
    return ($$0|0);
   } else {
    $26 = $$pre;
   }
  }
 }
 if ((label|0) == 9) {
  $18 = ((($2)) + 8|0);
  HEAP32[$18>>2] = $wrap$0;
  HEAP32[$$pre4>>2] = $$02;
  $26 = $2;
 }
 $25 = ((($26)) + 40|0);
 HEAP32[$25>>2] = 0;
 $27 = ((($26)) + 44|0);
 HEAP32[$27>>2] = 0;
 $28 = ((($26)) + 48|0);
 HEAP32[$28>>2] = 0;
 $29 = ((($26)) + 28|0);
 HEAP32[$29>>2] = 0;
 $30 = ((($strm)) + 20|0);
 HEAP32[$30>>2] = 0;
 $31 = ((($strm)) + 8|0);
 HEAP32[$31>>2] = 0;
 $32 = ((($strm)) + 24|0);
 HEAP32[$32>>2] = 0;
 $33 = ((($26)) + 8|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = ($34|0)==(0);
 if (!($35)) {
  $36 = $34 & 1;
  $37 = ((($strm)) + 48|0);
  HEAP32[$37>>2] = $36;
 }
 HEAP32[$26>>2] = 0;
 $38 = ((($26)) + 4|0);
 HEAP32[$38>>2] = 0;
 $39 = ((($26)) + 12|0);
 HEAP32[$39>>2] = 0;
 $40 = ((($26)) + 20|0);
 HEAP32[$40>>2] = 32768;
 $41 = ((($26)) + 32|0);
 HEAP32[$41>>2] = 0;
 $42 = ((($26)) + 56|0);
 HEAP32[$42>>2] = 0;
 $43 = ((($26)) + 60|0);
 HEAP32[$43>>2] = 0;
 $44 = ((($26)) + 1328|0);
 $45 = ((($26)) + 108|0);
 HEAP32[$45>>2] = $44;
 $46 = ((($26)) + 80|0);
 HEAP32[$46>>2] = $44;
 $47 = ((($26)) + 76|0);
 HEAP32[$47>>2] = $44;
 $48 = ((($26)) + 7104|0);
 HEAP32[$48>>2] = 1;
 $49 = ((($26)) + 7108|0);
 HEAP32[$49>>2] = -1;
 $$0 = 0;
 return ($$0|0);
}
function _inflateInit2_($strm,$windowBits,$version,$stream_size) {
 $strm = $strm|0;
 $windowBits = $windowBits|0;
 $version = $version|0;
 $stream_size = $stream_size|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($version|0)==(0|0);
 if ($0) {
  $$0 = -6;
  return ($$0|0);
 }
 $1 = HEAP8[$version>>0]|0;
 $2 = ($1<<24>>24)!=(49);
 $3 = ($stream_size|0)!=(56);
 $or$cond = $3 | $2;
 if ($or$cond) {
  $$0 = -6;
  return ($$0|0);
 }
 $4 = ($strm|0)==(0|0);
 if ($4) {
  $$0 = -2;
  return ($$0|0);
 }
 $5 = ((($strm)) + 24|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($strm)) + 32|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 if ($8) {
  HEAP32[$6>>2] = 4679;
  $9 = ((($strm)) + 40|0);
  HEAP32[$9>>2] = 0;
  $15 = 4679;
 } else {
  $15 = $7;
 }
 $10 = ((($strm)) + 36|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0|0);
 if ($12) {
  HEAP32[$10>>2] = 4680;
 }
 $13 = ((($strm)) + 40|0);
 $14 = HEAP32[$13>>2]|0;
 $16 = (FUNCTION_TABLE_iiii[$15 & 8191]($14,1,7116)|0);
 $17 = ($16|0)==(0|0);
 if ($17) {
  $$0 = -4;
  return ($$0|0);
 }
 $18 = ((($strm)) + 28|0);
 HEAP32[$18>>2] = $16;
 $19 = ((($16)) + 52|0);
 HEAP32[$19>>2] = 0;
 $20 = (_inflateReset2($strm,$windowBits)|0);
 $21 = ($20|0)==(0);
 if ($21) {
  $$0 = 0;
  return ($$0|0);
 }
 $22 = HEAP32[$10>>2]|0;
 $23 = HEAP32[$13>>2]|0;
 FUNCTION_TABLE_vii[$22 & 8191]($23,$16);
 HEAP32[$18>>2] = 0;
 $$0 = $20;
 return ($$0|0);
}
function _inflateInit_($strm,$version,$stream_size) {
 $strm = $strm|0;
 $version = $version|0;
 $stream_size = $stream_size|0;
 var $$0$i = 0, $$pre4$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $or$cond$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($version|0)==(0|0);
 if ($0) {
  $$0$i = -6;
  return ($$0$i|0);
 }
 $1 = HEAP8[$version>>0]|0;
 $2 = ($1<<24>>24)!=(49);
 $3 = ($stream_size|0)!=(56);
 $or$cond$i = $3 | $2;
 if ($or$cond$i) {
  $$0$i = -6;
  return ($$0$i|0);
 }
 $4 = ($strm|0)==(0|0);
 if ($4) {
  $$0$i = -2;
  return ($$0$i|0);
 }
 $5 = ((($strm)) + 24|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($strm)) + 32|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 if ($8) {
  HEAP32[$6>>2] = 4679;
  $9 = ((($strm)) + 40|0);
  HEAP32[$9>>2] = 0;
  $15 = 4679;
 } else {
  $15 = $7;
 }
 $10 = ((($strm)) + 36|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0|0);
 if ($12) {
  HEAP32[$10>>2] = 4680;
 }
 $13 = ((($strm)) + 40|0);
 $14 = HEAP32[$13>>2]|0;
 $16 = (FUNCTION_TABLE_iiii[$15 & 8191]($14,1,7116)|0);
 $17 = ($16|0)==(0|0);
 if ($17) {
  $$0$i = -4;
  return ($$0$i|0);
 }
 $18 = ((($strm)) + 28|0);
 HEAP32[$18>>2] = $16;
 $19 = ((($16)) + 52|0);
 HEAP32[$19>>2] = 0;
 $$pre4$i = ((($16)) + 36|0);
 $20 = ((($16)) + 8|0);
 HEAP32[$20>>2] = 1;
 HEAP32[$$pre4$i>>2] = 15;
 $21 = ((($16)) + 40|0);
 HEAP32[$21>>2] = 0;
 $22 = ((($16)) + 44|0);
 HEAP32[$22>>2] = 0;
 $23 = ((($16)) + 48|0);
 HEAP32[$23>>2] = 0;
 $24 = ((($16)) + 28|0);
 HEAP32[$24>>2] = 0;
 $25 = ((($strm)) + 20|0);
 HEAP32[$25>>2] = 0;
 $26 = ((($strm)) + 8|0);
 HEAP32[$26>>2] = 0;
 HEAP32[$5>>2] = 0;
 $27 = ((($strm)) + 48|0);
 HEAP32[$27>>2] = 1;
 HEAP32[$16>>2] = 0;
 $28 = ((($16)) + 4|0);
 HEAP32[$28>>2] = 0;
 $29 = ((($16)) + 12|0);
 HEAP32[$29>>2] = 0;
 $30 = ((($16)) + 20|0);
 HEAP32[$30>>2] = 32768;
 $31 = ((($16)) + 32|0);
 HEAP32[$31>>2] = 0;
 $32 = ((($16)) + 56|0);
 HEAP32[$32>>2] = 0;
 $33 = ((($16)) + 60|0);
 HEAP32[$33>>2] = 0;
 $34 = ((($16)) + 1328|0);
 $35 = ((($16)) + 108|0);
 HEAP32[$35>>2] = $34;
 $36 = ((($16)) + 80|0);
 HEAP32[$36>>2] = $34;
 $37 = ((($16)) + 76|0);
 HEAP32[$37>>2] = $34;
 $38 = ((($16)) + 7104|0);
 HEAP32[$38>>2] = 1;
 $39 = ((($16)) + 7108|0);
 HEAP32[$39>>2] = -1;
 $$0$i = 0;
 return ($$0$i|0);
}
function _inflate($strm,$flush) {
 $strm = $strm|0;
 $flush = $flush|0;
 var $$0 = 0, $$copy$7 = 0, $$hold$47 = 0, $$lcssa101 = 0, $$lcssa101$in = 0, $$lcssa164 = 0, $$lcssa164$in = 0, $$lcssa166 = 0, $$lcssa168 = 0, $$lcssa168$in = 0, $$lcssa170 = 0, $$lcssa176 = 0, $$lcssa180 = 0, $$lcssa194 = 0, $$lcssa194$in = 0, $$lcssa196 = 0, $$lcssa198 = 0, $$lcssa198$in = 0, $$lcssa200 = 0, $$lcssa206 = 0;
 var $$lcssa210 = 0, $$lcssa2657 = 0, $$lcssa2658 = 0, $$lcssa2659 = 0, $$lcssa2712 = 0, $$lcssa2751 = 0, $$lcssa2752 = 0, $$lcssa2753 = 0, $$lcssa2754 = 0, $$lcssa2799 = 0, $$lcssa2803 = 0, $$lcssa2819 = 0, $$lcssa2821 = 0, $$lcssa783$pn = 0, $$lcssa95 = 0, $$ph = 0, $$pn = 0, $$pn1281 = 0, $$pn1283 = 0, $$pr21$pre = 0;
 var $$pre = 0, $$pre1261 = 0, $$pre1262 = 0, $$pre1263 = 0, $$pre1264 = 0, $$pre1266 = 0, $$pre1269 = 0, $$pre1271 = 0, $$pre1273 = 0, $$pre1275 = 0, $$pre1276 = 0, $$pre1278 = 0, $$pre1279 = 0, $$ret$8 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0;
 var $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0;
 var $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0;
 var $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0;
 var $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0;
 var $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0;
 var $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0;
 var $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0;
 var $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0;
 var $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0;
 var $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0;
 var $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0;
 var $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0;
 var $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0;
 var $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0;
 var $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0;
 var $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0;
 var $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0;
 var $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0;
 var $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0;
 var $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0;
 var $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0;
 var $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0;
 var $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0;
 var $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0;
 var $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0;
 var $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0;
 var $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0;
 var $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0;
 var $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0;
 var $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0;
 var $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0;
 var $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0;
 var $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0;
 var $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0;
 var $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0;
 var $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0;
 var $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0;
 var $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0;
 var $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0;
 var $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bits$0 = 0, $bits$0$be = 0, $bits$1$lcssa = 0, $bits$10 = 0, $bits$11 = 0, $bits$12$lcssa = 0;
 var $bits$12765 = 0, $bits$13 = 0, $bits$14709 = 0, $bits$15 = 0, $bits$15$lcssa = 0, $bits$16 = 0, $bits$17 = 0, $bits$1775 = 0, $bits$18$lcssa = 0, $bits$18$lcssa$lcssa3008 = 0, $bits$19$lcssa = 0, $bits$19698 = 0, $bits$2$lcssa = 0, $bits$20 = 0, $bits$21 = 0, $bits$22$lcssa = 0, $bits$22531 = 0, $bits$23$lcssa = 0, $bits$23$ph1295 = 0, $bits$23551 = 0;
 var $bits$24$lcssa = 0, $bits$25$be = 0, $bits$25$be$lcssa = 0, $bits$25$lcssa1296 = 0, $bits$25$ph = 0, $bits$25608 = 0, $bits$26$lcssa = 0, $bits$26564 = 0, $bits$27$lcssa = 0, $bits$2728 = 0, $bits$27587 = 0, $bits$28$lcssa = 0, $bits$28577 = 0, $bits$29$lcssa = 0, $bits$29597 = 0, $bits$30 = 0, $bits$30$lcssa = 0, $bits$32 = 0, $bits$33 = 0, $bits$34$lcssa = 0;
 var $bits$34619 = 0, $bits$35$lcssa = 0, $bits$35635 = 0, $bits$36 = 0, $bits$37 = 0, $bits$3738 = 0, $bits$38$lcssa = 0, $bits$38648 = 0, $bits$39 = 0, $bits$40$lcssa = 0, $bits$40$ph = 0, $bits$40659 = 0, $bits$41$lcssa = 0, $bits$41674 = 0, $bits$42 = 0, $bits$43 = 0, $bits$44$lcssa = 0, $bits$44687 = 0, $bits$45 = 0, $bits$46 = 0;
 var $bits$47$lcssa = 0, $bits$4747 = 0, $bits$47514 = 0, $bits$48 = 0, $bits$49 = 0, $bits$5 = 0, $bits$50$lcssa = 0, $bits$50521 = 0, $bits$51 = 0, $bits$53 = 0, $bits$53$ph = 0, $bits$6756 = 0, $bits$7 = 0, $bits$8 = 0, $bits$9 = 0, $copy$1 = 0, $copy$10 = 0, $copy$2 = 0, $copy$4 = 0, $copy$5 = 0;
 var $copy$7 = 0, $copy$8 = 0, $exitcond = 0, $flush$off = 0, $from$0 = 0, $from$1 = 0, $from$2 = 0, $have$0 = 0, $have$0$be = 0, $have$1$lcssa = 0, $have$10 = 0, $have$11 = 0, $have$12 = 0, $have$13 = 0, $have$14 = 0, $have$15 = 0, $have$16$lcssa = 0, $have$16763 = 0, $have$17 = 0, $have$1773 = 0;
 var $have$18$lcssa = 0, $have$18707 = 0, $have$19 = 0, $have$19$lcssa = 0, $have$2$lcssa = 0, $have$20 = 0, $have$21 = 0, $have$22$lcssa = 0, $have$22$lcssa$lcssa3010 = 0, $have$23$lcssa = 0, $have$23696 = 0, $have$24 = 0, $have$25 = 0, $have$25$ = 0, $have$26$lcssa = 0, $have$26529 = 0, $have$27$lcssa = 0, $have$27$ph1293 = 0, $have$2726 = 0, $have$27549 = 0;
 var $have$28$lcssa = 0, $have$29$be = 0, $have$29$be$lcssa = 0, $have$29$lcssa1298 = 0, $have$29$ph = 0, $have$29606 = 0, $have$3$lcssa = 0, $have$30$lcssa = 0, $have$30562 = 0, $have$31$lcssa = 0, $have$31$lcssa$lcssa = 0, $have$31585 = 0, $have$32$lcssa = 0, $have$32575 = 0, $have$33$lcssa = 0, $have$33595 = 0, $have$34 = 0, $have$34$lcssa = 0, $have$36 = 0, $have$37 = 0;
 var $have$3736 = 0, $have$38$lcssa = 0, $have$38617 = 0, $have$39$lcssa = 0, $have$39633 = 0, $have$4$lcssa = 0, $have$40 = 0, $have$41 = 0, $have$42$lcssa = 0, $have$42646 = 0, $have$43 = 0, $have$44$lcssa = 0, $have$44$ph = 0, $have$44657 = 0, $have$45$lcssa = 0, $have$45672 = 0, $have$46 = 0, $have$47 = 0, $have$4745 = 0, $have$48$lcssa = 0;
 var $have$48685 = 0, $have$49 = 0, $have$5 = 0, $have$50 = 0, $have$51$lcssa = 0, $have$51512 = 0, $have$52 = 0, $have$53 = 0, $have$54$lcssa = 0, $have$54519 = 0, $have$55 = 0, $have$57 = 0, $have$57$ph = 0, $have$6$lcssa = 0, $have$6754 = 0, $have$7 = 0, $have$8 = 0, $have$8$ = 0, $have$9 = 0, $hbuf = 0;
 var $here$sroa$0$0 = 0, $here$sroa$0$1 = 0, $here$sroa$18$0 = 0, $here$sroa$18$1 = 0, $here$sroa$42$0 = 0, $here$sroa$42$1 = 0, $hold$0 = 0, $hold$0$be = 0, $hold$1$lcssa = 0, $hold$10 = 0, $hold$11 = 0, $hold$12$lcssa = 0, $hold$12764 = 0, $hold$13 = 0, $hold$14$lcssa = 0, $hold$14708 = 0, $hold$15 = 0, $hold$15$lcssa = 0, $hold$16 = 0, $hold$17 = 0;
 var $hold$1774 = 0, $hold$18$lcssa = 0, $hold$18$lcssa$lcssa3009 = 0, $hold$19$lcssa = 0, $hold$19697 = 0, $hold$20 = 0, $hold$21 = 0, $hold$22$lcssa = 0, $hold$22530 = 0, $hold$23$lcssa = 0, $hold$23$ph1294 = 0, $hold$23550 = 0, $hold$24$lcssa = 0, $hold$25$be = 0, $hold$25$be$lcssa = 0, $hold$25$lcssa1297 = 0, $hold$25$ph = 0, $hold$25607 = 0, $hold$26$lcssa = 0, $hold$26563 = 0;
 var $hold$27$lcssa = 0, $hold$2727 = 0, $hold$27586 = 0, $hold$28$lcssa = 0, $hold$28576 = 0, $hold$29$lcssa = 0, $hold$29596 = 0, $hold$3$lcssa = 0, $hold$30 = 0, $hold$30$lcssa = 0, $hold$32 = 0, $hold$33 = 0, $hold$34$lcssa = 0, $hold$34618 = 0, $hold$35$lcssa = 0, $hold$35634 = 0, $hold$36 = 0, $hold$37 = 0, $hold$3737 = 0, $hold$38$lcssa = 0;
 var $hold$38647 = 0, $hold$39 = 0, $hold$4$lcssa = 0, $hold$40$lcssa = 0, $hold$40$ph = 0, $hold$40658 = 0, $hold$41$lcssa = 0, $hold$41673 = 0, $hold$42 = 0, $hold$43 = 0, $hold$44$lcssa = 0, $hold$44686 = 0, $hold$45 = 0, $hold$46 = 0, $hold$47$lcssa = 0, $hold$4746 = 0, $hold$47513 = 0, $hold$48 = 0, $hold$49 = 0, $hold$5 = 0;
 var $hold$50$lcssa = 0, $hold$50520 = 0, $hold$51 = 0, $hold$53 = 0, $hold$53$ph = 0, $hold$6$lcssa = 0, $hold$6755 = 0, $hold$7 = 0, $hold$8 = 0, $hold$9 = 0, $left$0 = 0, $left$0$be = 0, $left$0$copy$8 = 0, $left$0$lcssa2861 = 0, $left$0$lcssa2877 = 0, $left$0$lcssa2879 = 0, $len$0 = 0, $next$0 = 0, $next$0$be = 0, $next$1$lcssa = 0;
 var $next$10 = 0, $next$11 = 0, $next$12 = 0, $next$13 = 0, $next$14 = 0, $next$15 = 0, $next$16$lcssa = 0, $next$16762 = 0, $next$17 = 0, $next$1772 = 0, $next$18$lcssa = 0, $next$18706 = 0, $next$19 = 0, $next$19$lcssa = 0, $next$2$lcssa = 0, $next$20 = 0, $next$21 = 0, $next$22$lcssa = 0, $next$22$lcssa$lcssa3011 = 0, $next$23$lcssa = 0;
 var $next$23695 = 0, $next$24 = 0, $next$25 = 0, $next$26$lcssa = 0, $next$26528 = 0, $next$27$lcssa = 0, $next$27$ph1292 = 0, $next$2725 = 0, $next$27548 = 0, $next$28$lcssa = 0, $next$29$be = 0, $next$29$be$lcssa = 0, $next$29$lcssa1299 = 0, $next$29$ph = 0, $next$29605 = 0, $next$3$lcssa = 0, $next$30$lcssa = 0, $next$30561 = 0, $next$31$lcssa = 0, $next$31$lcssa$lcssa = 0;
 var $next$31584 = 0, $next$32$lcssa = 0, $next$32574 = 0, $next$33$lcssa = 0, $next$33594 = 0, $next$34 = 0, $next$34$lcssa = 0, $next$36 = 0, $next$37 = 0, $next$3735 = 0, $next$38$lcssa = 0, $next$38616 = 0, $next$39$lcssa = 0, $next$39632 = 0, $next$4$lcssa = 0, $next$40 = 0, $next$41 = 0, $next$42$lcssa = 0, $next$42645 = 0, $next$43 = 0;
 var $next$44$lcssa = 0, $next$44$ph = 0, $next$44656 = 0, $next$45$lcssa = 0, $next$45671 = 0, $next$46 = 0, $next$47 = 0, $next$4744 = 0, $next$48$lcssa = 0, $next$48684 = 0, $next$49 = 0, $next$5 = 0, $next$50 = 0, $next$51$lcssa = 0, $next$51511 = 0, $next$52 = 0, $next$53 = 0, $next$54$lcssa = 0, $next$54518 = 0, $next$55 = 0;
 var $next$57 = 0, $next$57$ph = 0, $next$6$lcssa = 0, $next$6753 = 0, $next$7 = 0, $next$8 = 0, $next$9 = 0, $or$cond = 0, $or$cond10 = 0, $or$cond12 = 0, $or$cond14 = 0, $or$cond16 = 0, $or$cond17 = 0, $or$cond1733 = 0, $or$cond18 = 0, $or$cond20 = 0, $or$cond6 = 0, $or$cond8 = 0, $out$0 = 0, $out$0$be = 0;
 var $out$0$lcssa2944 = 0, $out$0$lcssa2966 = 0, $out$1 = 0, $out$2 = 0, $out$2$lcssa = 0, $out$4 = 0, $put$0$be = 0, $put$1 = 0, $ret$0 = 0, $ret$0$be = 0, $ret$0$lcssa2995 = 0, $ret$1$ph = 0, $ret$2 = 0, $ret$3 = 0, $ret$4 = 0, $ret$5$ph = 0, $ret$6 = 0, $ret$7 = 0, $ret$8 = 0, $ret$8$ph = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $hbuf = sp;
 $0 = ($strm|0)==(0|0);
 if ($0) {
  $$0 = -2;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ((($strm)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $$0 = -2;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ((($strm)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if ($6) {
  $$0 = -2;
  STACKTOP = sp;return ($$0|0);
 }
 $7 = HEAP32[$strm>>2]|0;
 $8 = ($7|0)==(0|0);
 if ($8) {
  $9 = ((($strm)) + 4|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ($10|0)==(0);
  if (!($11)) {
   $$0 = -2;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $12 = HEAP32[$2>>2]|0;
 $13 = ($12|0)==(11);
 if ($13) {
  HEAP32[$2>>2] = 12;
  $898 = 12;
 } else {
  $898 = $12;
 }
 $14 = ((($strm)) + 16|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($strm)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = ((($2)) + 56|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ((($2)) + 60|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ((($2)) + 8|0);
 $23 = ((($2)) + 24|0);
 $24 = ((($hbuf)) + 1|0);
 $25 = ((($2)) + 16|0);
 $26 = ((($2)) + 32|0);
 $27 = ((($strm)) + 24|0);
 $28 = ((($2)) + 36|0);
 $29 = ((($2)) + 20|0);
 $30 = ((($strm)) + 48|0);
 $31 = ((($2)) + 64|0);
 $32 = ((($2)) + 12|0);
 $flush$off = (($flush) + -5)|0;
 $33 = ($flush$off>>>0)<(2);
 $34 = ((($2)) + 4|0);
 $35 = ((($2)) + 76|0);
 $36 = ((($2)) + 84|0);
 $37 = ((($2)) + 80|0);
 $38 = ((($2)) + 88|0);
 $39 = ($flush|0)==(6);
 $40 = ((($2)) + 7108|0);
 $41 = ((($2)) + 72|0);
 $42 = ((($2)) + 7112|0);
 $43 = ((($2)) + 68|0);
 $44 = ((($2)) + 44|0);
 $45 = ((($2)) + 7104|0);
 $46 = ((($2)) + 48|0);
 $47 = ((($2)) + 52|0);
 $48 = ((($2)) + 40|0);
 $49 = ((($strm)) + 20|0);
 $50 = ((($2)) + 28|0);
 $51 = ((($hbuf)) + 2|0);
 $52 = ((($hbuf)) + 3|0);
 $53 = ((($2)) + 104|0);
 $54 = ((($2)) + 96|0);
 $55 = ((($2)) + 100|0);
 $56 = ((($2)) + 624|0);
 $57 = ((($2)) + 1328|0);
 $58 = ((($2)) + 108|0);
 $59 = ((($2)) + 112|0);
 $60 = ((($2)) + 752|0);
 $61 = ((($2)) + 92|0);
 $380 = $5;$62 = $898;$bits$0 = $21;$have$0 = $17;$hold$0 = $19;$left$0 = $15;$next$0 = $7;$out$0 = $15;$ret$0 = 0;
 L17: while(1) {
  L19: do {
   switch ($62|0) {
   case 28:  {
    $$lcssa2799 = $380;$$ph = $left$0;$bits$53$ph = $bits$0;$have$57$ph = $have$0;$hold$53$ph = $hold$0;$next$57$ph = $next$0;$out$0$lcssa2944 = $out$0;$ret$8$ph = 1;
    label = 282;
    break L17;
    break;
   }
   case 29:  {
    $844 = $380;$845 = $left$0;$bits$53 = $bits$0;$have$57 = $have$0;$hold$53 = $hold$0;$next$57 = $next$0;$out$4 = $out$0;$ret$8 = -3;
    break L17;
    break;
   }
   case 30:  {
    label = 296;
    break L17;
    break;
   }
   case 0:  {
    $66 = HEAP32[$22>>2]|0;
    $67 = ($66|0)==(0);
    if ($67) {
     HEAP32[$2>>2] = 12;
     $bits$0$be = $bits$0;$have$0$be = $have$0;$hold$0$be = $hold$0;$left$0$be = $left$0;$next$0$be = $next$0;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$0;
     break L19;
    }
    $68 = ($bits$0>>>0)<(16);
    if ($68) {
     $bits$1775 = $bits$0;$have$1773 = $have$0;$hold$1774 = $hold$0;$next$1772 = $next$0;
     while(1) {
      $69 = ($have$1773|0)==(0);
      if ($69) {
       $844 = $380;$845 = $left$0;$bits$53 = $bits$1775;$have$57 = 0;$hold$53 = $hold$1774;$next$57 = $next$1772;$out$4 = $out$0;$ret$8 = $ret$0;
       break L17;
      }
      $70 = (($have$1773) + -1)|0;
      $71 = ((($next$1772)) + 1|0);
      $72 = HEAP8[$next$1772>>0]|0;
      $73 = $72&255;
      $74 = $73 << $bits$1775;
      $75 = (($74) + ($hold$1774))|0;
      $76 = (($bits$1775) + 8)|0;
      $77 = ($76>>>0)<(16);
      if ($77) {
       $bits$1775 = $76;$have$1773 = $70;$hold$1774 = $75;$next$1772 = $71;
      } else {
       $bits$1$lcssa = $76;$have$1$lcssa = $70;$hold$1$lcssa = $75;$next$1$lcssa = $71;
       break;
      }
     }
    } else {
     $bits$1$lcssa = $bits$0;$have$1$lcssa = $have$0;$hold$1$lcssa = $hold$0;$next$1$lcssa = $next$0;
    }
    $78 = $66 & 2;
    $79 = ($78|0)!=(0);
    $80 = ($hold$1$lcssa|0)==(35615);
    $or$cond = $80 & $79;
    if ($or$cond) {
     $81 = (_crc32(0,0,0)|0);
     HEAP32[$23>>2] = $81;
     HEAP8[$hbuf>>0] = 31;
     HEAP8[$24>>0] = -117;
     $82 = HEAP32[$23>>2]|0;
     $83 = (_crc32($82,$hbuf,2)|0);
     HEAP32[$23>>2] = $83;
     HEAP32[$2>>2] = 1;
     $bits$0$be = 0;$have$0$be = $have$1$lcssa;$hold$0$be = 0;$left$0$be = $left$0;$next$0$be = $next$1$lcssa;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$0;
     break L19;
    }
    HEAP32[$25>>2] = 0;
    $84 = HEAP32[$26>>2]|0;
    $85 = ($84|0)==(0|0);
    if (!($85)) {
     $86 = ((($84)) + 48|0);
     HEAP32[$86>>2] = -1;
    }
    $87 = $66 & 1;
    $88 = ($87|0)==(0);
    if (!($88)) {
     $89 = $hold$1$lcssa << 8;
     $90 = $89 & 65280;
     $91 = $hold$1$lcssa >>> 8;
     $92 = (($90) + ($91))|0;
     $93 = (($92>>>0) % 31)&-1;
     $94 = ($93|0)==(0);
     if ($94) {
      $95 = $hold$1$lcssa & 15;
      $96 = ($95|0)==(8);
      if (!($96)) {
       HEAP32[$27>>2] = 612432;
       HEAP32[$2>>2] = 29;
       $bits$0$be = $bits$1$lcssa;$have$0$be = $have$1$lcssa;$hold$0$be = $hold$1$lcssa;$left$0$be = $left$0;$next$0$be = $next$1$lcssa;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$0;
       break L19;
      }
      $97 = $hold$1$lcssa >>> 4;
      $98 = (($bits$1$lcssa) + -4)|0;
      $99 = $97 & 15;
      $100 = (($99) + 8)|0;
      $101 = HEAP32[$28>>2]|0;
      $102 = ($101|0)==(0);
      if ($102) {
       HEAP32[$28>>2] = $100;
      } else {
       $103 = ($100>>>0)>($101>>>0);
       if ($103) {
        HEAP32[$27>>2] = 612459;
        HEAP32[$2>>2] = 29;
        $bits$0$be = $98;$have$0$be = $have$1$lcssa;$hold$0$be = $97;$left$0$be = $left$0;$next$0$be = $next$1$lcssa;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$0;
        break L19;
       }
      }
      $104 = 1 << $100;
      HEAP32[$29>>2] = $104;
      $105 = (_adler32(0,0,0)|0);
      HEAP32[$23>>2] = $105;
      HEAP32[$30>>2] = $105;
      $106 = $hold$1$lcssa >>> 12;
      $107 = $106 & 2;
      $108 = $107 ^ 11;
      HEAP32[$2>>2] = $108;
      $bits$0$be = 0;$have$0$be = $have$1$lcssa;$hold$0$be = 0;$left$0$be = $left$0;$next$0$be = $next$1$lcssa;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$0;
      break L19;
     }
    }
    HEAP32[$27>>2] = 612409;
    HEAP32[$2>>2] = 29;
    $bits$0$be = $bits$1$lcssa;$have$0$be = $have$1$lcssa;$hold$0$be = $hold$1$lcssa;$left$0$be = $left$0;$next$0$be = $next$1$lcssa;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$0;
    break;
   }
   case 1:  {
    $65 = ($bits$0>>>0)<(16);
    if ($65) {
     $bits$2728 = $bits$0;$have$2726 = $have$0;$hold$2727 = $hold$0;$next$2725 = $next$0;
     while(1) {
      $109 = ($have$2726|0)==(0);
      if ($109) {
       $844 = $380;$845 = $left$0;$bits$53 = $bits$2728;$have$57 = 0;$hold$53 = $hold$2727;$next$57 = $next$2725;$out$4 = $out$0;$ret$8 = $ret$0;
       break L17;
      }
      $110 = (($have$2726) + -1)|0;
      $111 = ((($next$2725)) + 1|0);
      $112 = HEAP8[$next$2725>>0]|0;
      $113 = $112&255;
      $114 = $113 << $bits$2728;
      $115 = (($114) + ($hold$2727))|0;
      $116 = (($bits$2728) + 8)|0;
      $117 = ($116>>>0)<(16);
      if ($117) {
       $bits$2728 = $116;$have$2726 = $110;$hold$2727 = $115;$next$2725 = $111;
      } else {
       $118 = $115;$bits$2$lcssa = $116;$have$2$lcssa = $110;$next$2$lcssa = $111;
       break;
      }
     }
    } else {
     $118 = $hold$0;$bits$2$lcssa = $bits$0;$have$2$lcssa = $have$0;$next$2$lcssa = $next$0;
    }
    HEAP32[$25>>2] = $118;
    $119 = $118 & 255;
    $120 = ($119|0)==(8);
    if (!($120)) {
     HEAP32[$27>>2] = 612432;
     HEAP32[$2>>2] = 29;
     $bits$0$be = $bits$2$lcssa;$have$0$be = $have$2$lcssa;$hold$0$be = $118;$left$0$be = $left$0;$next$0$be = $next$2$lcssa;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$0;
     break L19;
    }
    $121 = $118 & 57344;
    $122 = ($121|0)==(0);
    if (!($122)) {
     HEAP32[$27>>2] = 612479;
     HEAP32[$2>>2] = 29;
     $bits$0$be = $bits$2$lcssa;$have$0$be = $have$2$lcssa;$hold$0$be = $118;$left$0$be = $left$0;$next$0$be = $next$2$lcssa;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$0;
     break L19;
    }
    $123 = HEAP32[$26>>2]|0;
    $124 = ($123|0)==(0|0);
    if (!($124)) {
     $125 = $118 >>> 8;
     $126 = $125 & 1;
     HEAP32[$123>>2] = $126;
    }
    $127 = $118 & 512;
    $128 = ($127|0)==(0);
    if (!($128)) {
     $129 = $118&255;
     HEAP8[$hbuf>>0] = $129;
     $130 = $118 >>> 8;
     $131 = $130&255;
     HEAP8[$24>>0] = $131;
     $132 = HEAP32[$23>>2]|0;
     $133 = (_crc32($132,$hbuf,2)|0);
     HEAP32[$23>>2] = $133;
    }
    HEAP32[$2>>2] = 2;
    $bits$3738 = 0;$have$3736 = $have$2$lcssa;$hold$3737 = 0;$next$3735 = $next$2$lcssa;
    label = 47;
    break;
   }
   case 2:  {
    $134 = ($bits$0>>>0)<(32);
    if ($134) {
     $bits$3738 = $bits$0;$have$3736 = $have$0;$hold$3737 = $hold$0;$next$3735 = $next$0;
     label = 47;
    } else {
     $have$3$lcssa = $have$0;$hold$3$lcssa = $hold$0;$next$3$lcssa = $next$0;
     label = 49;
    }
    break;
   }
   case 3:  {
    $159 = ($bits$0>>>0)<(16);
    if ($159) {
     $bits$4747 = $bits$0;$have$4745 = $have$0;$hold$4746 = $hold$0;$next$4744 = $next$0;
     label = 55;
    } else {
     $have$4$lcssa = $have$0;$hold$4$lcssa = $hold$0;$next$4$lcssa = $next$0;
     label = 57;
    }
    break;
   }
   case 4:  {
    $bits$5 = $bits$0;$have$5 = $have$0;$hold$5 = $hold$0;$next$5 = $next$0;
    label = 62;
    break;
   }
   case 5:  {
    $bits$8 = $bits$0;$have$8 = $have$0;$hold$8 = $hold$0;$next$8 = $next$0;
    label = 73;
    break;
   }
   case 6:  {
    $bits$9 = $bits$0;$have$11 = $have$0;$hold$9 = $hold$0;$next$11 = $next$0;
    label = 83;
    break;
   }
   case 7:  {
    $bits$10 = $bits$0;$have$13 = $have$0;$hold$10 = $hold$0;$next$13 = $next$0;
    label = 96;
    break;
   }
   case 8:  {
    $bits$11 = $bits$0;$have$15 = $have$0;$hold$11 = $hold$0;$next$15 = $next$0;
    label = 109;
    break;
   }
   case 9:  {
    $64 = ($bits$0>>>0)<(32);
    if ($64) {
     $bits$14709 = $bits$0;$have$18707 = $have$0;$hold$14708 = $hold$0;$next$18706 = $next$0;
     while(1) {
      $324 = ($have$18707|0)==(0);
      if ($324) {
       $844 = $380;$845 = $left$0;$bits$53 = $bits$14709;$have$57 = 0;$hold$53 = $hold$14708;$next$57 = $next$18706;$out$4 = $out$0;$ret$8 = $ret$0;
       break L17;
      }
      $325 = (($have$18707) + -1)|0;
      $326 = ((($next$18706)) + 1|0);
      $327 = HEAP8[$next$18706>>0]|0;
      $328 = $327&255;
      $329 = $328 << $bits$14709;
      $330 = (($329) + ($hold$14708))|0;
      $331 = (($bits$14709) + 8)|0;
      $332 = ($331>>>0)<(32);
      if ($332) {
       $bits$14709 = $331;$have$18707 = $325;$hold$14708 = $330;$next$18706 = $326;
      } else {
       $have$18$lcssa = $325;$hold$14$lcssa = $330;$next$18$lcssa = $326;
       break;
      }
     }
    } else {
     $have$18$lcssa = $have$0;$hold$14$lcssa = $hold$0;$next$18$lcssa = $next$0;
    }
    $333 = (_llvm_bswap_i32(($hold$14$lcssa|0))|0);
    HEAP32[$23>>2] = $333;
    HEAP32[$30>>2] = $333;
    HEAP32[$2>>2] = 10;
    $bits$15 = 0;$have$19 = $have$18$lcssa;$hold$15 = 0;$next$19 = $next$18$lcssa;
    label = 121;
    break;
   }
   case 10:  {
    $bits$15 = $bits$0;$have$19 = $have$0;$hold$15 = $hold$0;$next$19 = $next$0;
    label = 121;
    break;
   }
   case 11:  {
    $bits$16 = $bits$0;$have$20 = $have$0;$hold$16 = $hold$0;$next$20 = $next$0;
    label = 124;
    break;
   }
   case 12:  {
    $bits$17 = $bits$0;$have$21 = $have$0;$hold$17 = $hold$0;$next$21 = $next$0;
    label = 125;
    break;
   }
   case 13:  {
    $358 = $bits$0 & 7;
    $359 = $hold$0 >>> $358;
    $360 = (($bits$0) - ($358))|0;
    $361 = ($360>>>0)<(32);
    if ($361) {
     $bits$19698 = $360;$have$23696 = $have$0;$hold$19697 = $359;$next$23695 = $next$0;
     while(1) {
      $362 = ($have$23696|0)==(0);
      if ($362) {
       $844 = $380;$845 = $left$0;$bits$53 = $bits$19698;$have$57 = 0;$hold$53 = $hold$19697;$next$57 = $next$23695;$out$4 = $out$0;$ret$8 = $ret$0;
       break L17;
      }
      $363 = (($have$23696) + -1)|0;
      $364 = ((($next$23695)) + 1|0);
      $365 = HEAP8[$next$23695>>0]|0;
      $366 = $365&255;
      $367 = $366 << $bits$19698;
      $368 = (($367) + ($hold$19697))|0;
      $369 = (($bits$19698) + 8)|0;
      $370 = ($369>>>0)<(32);
      if ($370) {
       $bits$19698 = $369;$have$23696 = $363;$hold$19697 = $368;$next$23695 = $364;
      } else {
       $bits$19$lcssa = $369;$have$23$lcssa = $363;$hold$19$lcssa = $368;$next$23$lcssa = $364;
       break;
      }
     }
    } else {
     $bits$19$lcssa = $360;$have$23$lcssa = $have$0;$hold$19$lcssa = $359;$next$23$lcssa = $next$0;
    }
    $371 = $hold$19$lcssa & 65535;
    $372 = $hold$19$lcssa >>> 16;
    $373 = $372 ^ 65535;
    $374 = ($371|0)==($373|0);
    if ($374) {
     HEAP32[$31>>2] = $371;
     HEAP32[$2>>2] = 14;
     if ($39) {
      $$lcssa2799 = $380;$$ph = $left$0;$bits$53$ph = 0;$have$57$ph = $have$23$lcssa;$hold$53$ph = 0;$next$57$ph = $next$23$lcssa;$out$0$lcssa2944 = $out$0;$ret$8$ph = $ret$0;
      label = 282;
      break L17;
     } else {
      $bits$20 = 0;$have$24 = $have$23$lcssa;$hold$20 = 0;$next$24 = $next$23$lcssa;
      label = 144;
      break L19;
     }
    } else {
     HEAP32[$27>>2] = 612543;
     HEAP32[$2>>2] = 29;
     $bits$0$be = $bits$19$lcssa;$have$0$be = $have$23$lcssa;$hold$0$be = $hold$19$lcssa;$left$0$be = $left$0;$next$0$be = $next$23$lcssa;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$0;
     break L19;
    }
    break;
   }
   case 14:  {
    $bits$20 = $bits$0;$have$24 = $have$0;$hold$20 = $hold$0;$next$24 = $next$0;
    label = 144;
    break;
   }
   case 15:  {
    $bits$21 = $bits$0;$have$25 = $have$0;$hold$21 = $hold$0;$next$25 = $next$0;
    label = 145;
    break;
   }
   case 16:  {
    $63 = ($bits$0>>>0)<(14);
    if ($63) {
     $bits$22531 = $bits$0;$have$26529 = $have$0;$hold$22530 = $hold$0;$next$26528 = $next$0;
     while(1) {
      $387 = ($have$26529|0)==(0);
      if ($387) {
       $844 = $380;$845 = $left$0;$bits$53 = $bits$22531;$have$57 = 0;$hold$53 = $hold$22530;$next$57 = $next$26528;$out$4 = $out$0;$ret$8 = $ret$0;
       break L17;
      }
      $388 = (($have$26529) + -1)|0;
      $389 = ((($next$26528)) + 1|0);
      $390 = HEAP8[$next$26528>>0]|0;
      $391 = $390&255;
      $392 = $391 << $bits$22531;
      $393 = (($392) + ($hold$22530))|0;
      $394 = (($bits$22531) + 8)|0;
      $395 = ($394>>>0)<(14);
      if ($395) {
       $bits$22531 = $394;$have$26529 = $388;$hold$22530 = $393;$next$26528 = $389;
      } else {
       $bits$22$lcssa = $394;$have$26$lcssa = $388;$hold$22$lcssa = $393;$next$26$lcssa = $389;
       break;
      }
     }
    } else {
     $bits$22$lcssa = $bits$0;$have$26$lcssa = $have$0;$hold$22$lcssa = $hold$0;$next$26$lcssa = $next$0;
    }
    $396 = $hold$22$lcssa & 31;
    $397 = (($396) + 257)|0;
    HEAP32[$54>>2] = $397;
    $398 = $hold$22$lcssa >>> 5;
    $399 = $398 & 31;
    $400 = (($399) + 1)|0;
    HEAP32[$55>>2] = $400;
    $401 = $hold$22$lcssa >>> 10;
    $402 = $401 & 15;
    $403 = (($402) + 4)|0;
    HEAP32[$61>>2] = $403;
    $404 = $hold$22$lcssa >>> 14;
    $405 = (($bits$22$lcssa) + -14)|0;
    $406 = ($397>>>0)>(286);
    $407 = ($400>>>0)>(30);
    $or$cond1733 = $406 | $407;
    if ($or$cond1733) {
     HEAP32[$27>>2] = 612572;
     HEAP32[$2>>2] = 29;
     $bits$0$be = $405;$have$0$be = $have$26$lcssa;$hold$0$be = $404;$left$0$be = $left$0;$next$0$be = $next$26$lcssa;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$0;
     break L19;
    } else {
     HEAP32[$53>>2] = 0;
     HEAP32[$2>>2] = 17;
     $430 = $403;$899 = 0;$bits$23$ph1295 = $405;$have$27$ph1293 = $have$26$lcssa;$hold$23$ph1294 = $404;$next$27$ph1292 = $next$26$lcssa;
     label = 155;
     break L19;
    }
    break;
   }
   case 17:  {
    $$pre1262 = HEAP32[$53>>2]|0;
    $$pre1263 = HEAP32[$61>>2]|0;
    $408 = ($$pre1262>>>0)<($$pre1263>>>0);
    if ($408) {
     $430 = $$pre1263;$899 = $$pre1262;$bits$23$ph1295 = $bits$0;$have$27$ph1293 = $have$0;$hold$23$ph1294 = $hold$0;$next$27$ph1292 = $next$0;
     label = 155;
    } else {
     $$lcssa95 = $$pre1262;$bits$23$lcssa = $bits$0;$have$27$lcssa = $have$0;$hold$23$lcssa = $hold$0;$next$27$lcssa = $next$0;
     label = 156;
    }
    break;
   }
   case 18:  {
    $$pre1264 = HEAP32[$53>>2]|0;
    $443 = $$pre1264;$bits$25$ph = $bits$0;$have$29$ph = $have$0;$hold$25$ph = $hold$0;$next$29$ph = $next$0;$ret$1$ph = $ret$0;
    label = 166;
    break;
   }
   case 19:  {
    $bits$32 = $bits$0;$have$36 = $have$0;$hold$32 = $hold$0;$next$36 = $next$0;$ret$2 = $ret$0;
    label = 200;
    break;
   }
   case 20:  {
    $bits$33 = $bits$0;$have$37 = $have$0;$hold$33 = $hold$0;$next$37 = $next$0;$ret$3 = $ret$0;
    label = 201;
    break;
   }
   case 21:  {
    $$pre1269 = HEAP32[$41>>2]|0;
    $644 = $$pre1269;$bits$37 = $bits$0;$have$41 = $have$0;$hold$37 = $hold$0;$next$41 = $next$0;$ret$4 = $ret$0;
    label = 219;
    break;
   }
   case 22:  {
    $bits$40$ph = $bits$0;$have$44$ph = $have$0;$hold$40$ph = $hold$0;$next$44$ph = $next$0;$ret$5$ph = $ret$0;
    label = 226;
    break;
   }
   case 23:  {
    $$pre1275 = HEAP32[$41>>2]|0;
    $737 = $$pre1275;$bits$43 = $bits$0;$have$47 = $have$0;$hold$43 = $hold$0;$next$47 = $next$0;$ret$6 = $ret$0;
    label = 238;
    break;
   }
   case 24:  {
    $bits$46 = $bits$0;$have$50 = $have$0;$hold$46 = $hold$0;$next$50 = $next$0;$ret$7 = $ret$0;
    label = 244;
    break;
   }
   case 25:  {
    $792 = ($left$0|0)==(0);
    if ($792) {
     $$lcssa2799 = $380;$$ph = 0;$bits$53$ph = $bits$0;$have$57$ph = $have$0;$hold$53$ph = $hold$0;$next$57$ph = $next$0;$out$0$lcssa2944 = $out$0;$ret$8$ph = $ret$0;
     label = 282;
     break L17;
    }
    $793 = HEAP32[$31>>2]|0;
    $794 = $793&255;
    $795 = ((($380)) + 1|0);
    HEAP8[$380>>0] = $794;
    $796 = (($left$0) + -1)|0;
    HEAP32[$2>>2] = 20;
    $bits$0$be = $bits$0;$have$0$be = $have$0;$hold$0$be = $hold$0;$left$0$be = $796;$next$0$be = $next$0;$out$0$be = $out$0;$put$0$be = $795;$ret$0$be = $ret$0;
    break;
   }
   case 26:  {
    $797 = HEAP32[$22>>2]|0;
    $798 = ($797|0)==(0);
    if ($798) {
     $bits$48 = $bits$0;$have$52 = $have$0;$hold$48 = $hold$0;$next$52 = $next$0;$out$1 = $out$0;
    } else {
     $799 = ($bits$0>>>0)<(32);
     if ($799) {
      $bits$47514 = $bits$0;$have$51512 = $have$0;$hold$47513 = $hold$0;$next$51511 = $next$0;
      while(1) {
       $800 = ($have$51512|0)==(0);
       if ($800) {
        $844 = $380;$845 = $left$0;$bits$53 = $bits$47514;$have$57 = 0;$hold$53 = $hold$47513;$next$57 = $next$51511;$out$4 = $out$0;$ret$8 = $ret$0;
        break L17;
       }
       $801 = (($have$51512) + -1)|0;
       $802 = ((($next$51511)) + 1|0);
       $803 = HEAP8[$next$51511>>0]|0;
       $804 = $803&255;
       $805 = $804 << $bits$47514;
       $806 = (($805) + ($hold$47513))|0;
       $807 = (($bits$47514) + 8)|0;
       $808 = ($807>>>0)<(32);
       if ($808) {
        $bits$47514 = $807;$have$51512 = $801;$hold$47513 = $806;$next$51511 = $802;
       } else {
        $bits$47$lcssa = $807;$have$51$lcssa = $801;$hold$47$lcssa = $806;$next$51$lcssa = $802;
        break;
       }
      }
     } else {
      $bits$47$lcssa = $bits$0;$have$51$lcssa = $have$0;$hold$47$lcssa = $hold$0;$next$51$lcssa = $next$0;
     }
     $809 = (($out$0) - ($left$0))|0;
     $810 = HEAP32[$49>>2]|0;
     $811 = (($810) + ($809))|0;
     HEAP32[$49>>2] = $811;
     $812 = HEAP32[$50>>2]|0;
     $813 = (($812) + ($809))|0;
     HEAP32[$50>>2] = $813;
     $814 = ($out$0|0)==($left$0|0);
     if ($814) {
      $$pre1261 = HEAP32[$23>>2]|0;
      $827 = $$pre1261;
     } else {
      $815 = HEAP32[$25>>2]|0;
      $816 = ($815|0)==(0);
      $817 = HEAP32[$23>>2]|0;
      $818 = (0 - ($809))|0;
      $819 = (($380) + ($818)|0);
      if ($816) {
       $821 = (_adler32($817,$819,$809)|0);
       $822 = $821;
      } else {
       $820 = (_crc32($817,$819,$809)|0);
       $822 = $820;
      }
      HEAP32[$23>>2] = $822;
      HEAP32[$30>>2] = $822;
      $827 = $822;
     }
     $823 = HEAP32[$25>>2]|0;
     $824 = ($823|0)==(0);
     $825 = (_llvm_bswap_i32(($hold$47$lcssa|0))|0);
     $$hold$47 = $824 ? $825 : $hold$47$lcssa;
     $826 = ($$hold$47|0)==($827|0);
     if ($826) {
      $bits$48 = 0;$have$52 = $have$51$lcssa;$hold$48 = 0;$next$52 = $next$51$lcssa;$out$1 = $left$0;
     } else {
      HEAP32[$27>>2] = 612746;
      HEAP32[$2>>2] = 29;
      $bits$0$be = $bits$47$lcssa;$have$0$be = $have$51$lcssa;$hold$0$be = $hold$47$lcssa;$left$0$be = $left$0;$next$0$be = $next$51$lcssa;$out$0$be = $left$0;$put$0$be = $380;$ret$0$be = $ret$0;
      break L19;
     }
    }
    HEAP32[$2>>2] = 27;
    $bits$49 = $bits$48;$have$53 = $have$52;$hold$49 = $hold$48;$next$53 = $next$52;$out$2 = $out$1;
    label = 274;
    break;
   }
   case 27:  {
    $bits$49 = $bits$0;$have$53 = $have$0;$hold$49 = $hold$0;$next$53 = $next$0;$out$2 = $out$0;
    label = 274;
    break;
   }
   default: {
    $$0 = -2;
    label = 297;
    break L17;
   }
   }
  } while(0);
  if ((label|0) == 47) {
   while(1) {
    label = 0;
    $135 = ($have$3736|0)==(0);
    if ($135) {
     $844 = $380;$845 = $left$0;$bits$53 = $bits$3738;$have$57 = 0;$hold$53 = $hold$3737;$next$57 = $next$3735;$out$4 = $out$0;$ret$8 = $ret$0;
     break L17;
    }
    $136 = (($have$3736) + -1)|0;
    $137 = ((($next$3735)) + 1|0);
    $138 = HEAP8[$next$3735>>0]|0;
    $139 = $138&255;
    $140 = $139 << $bits$3738;
    $141 = (($140) + ($hold$3737))|0;
    $142 = (($bits$3738) + 8)|0;
    $143 = ($142>>>0)<(32);
    if ($143) {
     $bits$3738 = $142;$have$3736 = $136;$hold$3737 = $141;$next$3735 = $137;
     label = 47;
    } else {
     $have$3$lcssa = $136;$hold$3$lcssa = $141;$next$3$lcssa = $137;
     label = 49;
     break;
    }
   }
  }
  else if ((label|0) == 121) {
   label = 0;
   $334 = HEAP32[$32>>2]|0;
   $335 = ($334|0)==(0);
   if ($335) {
    $$lcssa2819 = $380;$bits$15$lcssa = $bits$15;$have$19$lcssa = $have$19;$hold$15$lcssa = $hold$15;$left$0$lcssa2877 = $left$0;$next$19$lcssa = $next$19;
    label = 122;
    break;
   }
   $336 = (_adler32(0,0,0)|0);
   HEAP32[$23>>2] = $336;
   HEAP32[$30>>2] = $336;
   HEAP32[$2>>2] = 11;
   $bits$16 = $bits$15;$have$20 = $have$19;$hold$16 = $hold$15;$next$20 = $next$19;
   label = 124;
  }
  else if ((label|0) == 144) {
   label = 0;
   HEAP32[$2>>2] = 15;
   $bits$21 = $bits$20;$have$25 = $have$24;$hold$21 = $hold$20;$next$25 = $next$24;
   label = 145;
  }
  else if ((label|0) == 155) {
   label = 0;
   $422 = $899;$bits$23551 = $bits$23$ph1295;$have$27549 = $have$27$ph1293;$hold$23550 = $hold$23$ph1294;$next$27548 = $next$27$ph1292;
   while(1) {
    $410 = ($bits$23551>>>0)<(3);
    if ($410) {
     $411 = ($have$27549|0)==(0);
     if ($411) {
      $844 = $380;$845 = $left$0;$bits$53 = $bits$23551;$have$57 = 0;$hold$53 = $hold$23550;$next$57 = $next$27548;$out$4 = $out$0;$ret$8 = $ret$0;
      break L17;
     }
     $412 = (($bits$23551) + 8)|0;
     $413 = HEAP8[$next$27548>>0]|0;
     $414 = $413&255;
     $415 = $414 << $bits$23551;
     $416 = (($415) + ($hold$23550))|0;
     $417 = ((($next$27548)) + 1|0);
     $418 = (($have$27549) + -1)|0;
     $bits$24$lcssa = $412;$have$28$lcssa = $418;$hold$24$lcssa = $416;$next$28$lcssa = $417;
    } else {
     $bits$24$lcssa = $bits$23551;$have$28$lcssa = $have$27549;$hold$24$lcssa = $hold$23550;$next$28$lcssa = $next$27548;
    }
    $419 = $hold$24$lcssa & 7;
    $420 = $419&65535;
    $421 = (($422) + 1)|0;
    HEAP32[$53>>2] = $421;
    $423 = (177638 + ($422<<1)|0);
    $424 = HEAP16[$423>>1]|0;
    $425 = $424&65535;
    $426 = (((($2)) + 112|0) + ($425<<1)|0);
    HEAP16[$426>>1] = $420;
    $427 = $hold$24$lcssa >>> 3;
    $428 = (($bits$24$lcssa) + -3)|0;
    $429 = ($421>>>0)<($430>>>0);
    if ($429) {
     $422 = $421;$bits$23551 = $428;$have$27549 = $have$28$lcssa;$hold$23550 = $427;$next$27548 = $next$28$lcssa;
    } else {
     $$lcssa95 = $421;$bits$23$lcssa = $428;$have$27$lcssa = $have$28$lcssa;$hold$23$lcssa = $427;$next$27$lcssa = $next$28$lcssa;
     label = 156;
     break;
    }
   }
  }
  else if ((label|0) == 274) {
   label = 0;
   $828 = HEAP32[$22>>2]|0;
   $829 = ($828|0)==(0);
   if ($829) {
    $$lcssa2803 = $380;$bits$51 = $bits$49;$have$55 = $have$53;$hold$51 = $hold$49;$left$0$lcssa2861 = $left$0;$next$55 = $next$53;$out$2$lcssa = $out$2;
    label = 281;
    break;
   }
   $830 = HEAP32[$25>>2]|0;
   $831 = ($830|0)==(0);
   if ($831) {
    $$lcssa2803 = $380;$bits$51 = $bits$49;$have$55 = $have$53;$hold$51 = $hold$49;$left$0$lcssa2861 = $left$0;$next$55 = $next$53;$out$2$lcssa = $out$2;
    label = 281;
    break;
   }
   $832 = ($bits$49>>>0)<(32);
   if ($832) {
    $bits$50521 = $bits$49;$have$54519 = $have$53;$hold$50520 = $hold$49;$next$54518 = $next$53;
    while(1) {
     $833 = ($have$54519|0)==(0);
     if ($833) {
      $844 = $380;$845 = $left$0;$bits$53 = $bits$50521;$have$57 = 0;$hold$53 = $hold$50520;$next$57 = $next$54518;$out$4 = $out$2;$ret$8 = $ret$0;
      break L17;
     }
     $834 = (($have$54519) + -1)|0;
     $835 = ((($next$54518)) + 1|0);
     $836 = HEAP8[$next$54518>>0]|0;
     $837 = $836&255;
     $838 = $837 << $bits$50521;
     $839 = (($838) + ($hold$50520))|0;
     $840 = (($bits$50521) + 8)|0;
     $841 = ($840>>>0)<(32);
     if ($841) {
      $bits$50521 = $840;$have$54519 = $834;$hold$50520 = $839;$next$54518 = $835;
     } else {
      $bits$50$lcssa = $840;$have$54$lcssa = $834;$hold$50$lcssa = $839;$next$54$lcssa = $835;
      break;
     }
    }
   } else {
    $bits$50$lcssa = $bits$49;$have$54$lcssa = $have$53;$hold$50$lcssa = $hold$49;$next$54$lcssa = $next$53;
   }
   $842 = HEAP32[$50>>2]|0;
   $843 = ($hold$50$lcssa|0)==($842|0);
   if ($843) {
    $$lcssa2803 = $380;$bits$51 = 0;$have$55 = $have$54$lcssa;$hold$51 = 0;$left$0$lcssa2861 = $left$0;$next$55 = $next$54$lcssa;$out$2$lcssa = $out$2;
    label = 281;
    break;
   }
   HEAP32[$27>>2] = 612767;
   HEAP32[$2>>2] = 29;
   $bits$0$be = $bits$50$lcssa;$have$0$be = $have$54$lcssa;$hold$0$be = $hold$50$lcssa;$left$0$be = $left$0;$next$0$be = $next$54$lcssa;$out$0$be = $out$2;$put$0$be = $380;$ret$0$be = $ret$0;
  }
  do {
   if ((label|0) == 49) {
    label = 0;
    $144 = HEAP32[$26>>2]|0;
    $145 = ($144|0)==(0|0);
    if (!($145)) {
     $146 = ((($144)) + 4|0);
     HEAP32[$146>>2] = $hold$3$lcssa;
    }
    $147 = HEAP32[$25>>2]|0;
    $148 = $147 & 512;
    $149 = ($148|0)==(0);
    if (!($149)) {
     $150 = $hold$3$lcssa&255;
     HEAP8[$hbuf>>0] = $150;
     $151 = $hold$3$lcssa >>> 8;
     $152 = $151&255;
     HEAP8[$24>>0] = $152;
     $153 = $hold$3$lcssa >>> 16;
     $154 = $153&255;
     HEAP8[$51>>0] = $154;
     $155 = $hold$3$lcssa >>> 24;
     $156 = $155&255;
     HEAP8[$52>>0] = $156;
     $157 = HEAP32[$23>>2]|0;
     $158 = (_crc32($157,$hbuf,4)|0);
     HEAP32[$23>>2] = $158;
    }
    HEAP32[$2>>2] = 3;
    $bits$4747 = 0;$have$4745 = $have$3$lcssa;$hold$4746 = 0;$next$4744 = $next$3$lcssa;
    label = 55;
   }
   else if ((label|0) == 124) {
    label = 0;
    if ($33) {
     $$lcssa2799 = $380;$$ph = $left$0;$bits$53$ph = $bits$16;$have$57$ph = $have$20;$hold$53$ph = $hold$16;$next$57$ph = $next$20;$out$0$lcssa2944 = $out$0;$ret$8$ph = $ret$0;
     label = 282;
     break L17;
    } else {
     $bits$17 = $bits$16;$have$21 = $have$20;$hold$17 = $hold$16;$next$21 = $next$20;
     label = 125;
    }
   }
   else if ((label|0) == 145) {
    label = 0;
    $375 = HEAP32[$31>>2]|0;
    $376 = ($375|0)==(0);
    if ($376) {
     HEAP32[$2>>2] = 11;
     $bits$0$be = $bits$21;$have$0$be = $have$25;$hold$0$be = $hold$21;$left$0$be = $left$0;$next$0$be = $next$25;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$0;
     break;
    }
    $377 = ($375>>>0)>($have$25>>>0);
    $have$25$ = $377 ? $have$25 : $375;
    $378 = ($have$25$>>>0)>($left$0>>>0);
    $copy$4 = $378 ? $left$0 : $have$25$;
    $379 = ($copy$4|0)==(0);
    if ($379) {
     $$lcssa2799 = $380;$$ph = $left$0;$bits$53$ph = $bits$21;$have$57$ph = $have$25;$hold$53$ph = $hold$21;$next$57$ph = $next$25;$out$0$lcssa2944 = $out$0;$ret$8$ph = $ret$0;
     label = 282;
     break L17;
    }
    _memcpy(($380|0),($next$25|0),($copy$4|0))|0;
    $381 = (($have$25) - ($copy$4))|0;
    $382 = (($next$25) + ($copy$4)|0);
    $383 = (($left$0) - ($copy$4))|0;
    $384 = (($380) + ($copy$4)|0);
    $385 = HEAP32[$31>>2]|0;
    $386 = (($385) - ($copy$4))|0;
    HEAP32[$31>>2] = $386;
    $bits$0$be = $bits$21;$have$0$be = $381;$hold$0$be = $hold$21;$left$0$be = $383;$next$0$be = $382;$out$0$be = $out$0;$put$0$be = $384;$ret$0$be = $ret$0;
   }
   else if ((label|0) == 156) {
    label = 0;
    $409 = ($$lcssa95>>>0)<(19);
    if ($409) {
     $432 = $$lcssa95;
     while(1) {
      $431 = (($432) + 1)|0;
      $433 = (177638 + ($432<<1)|0);
      $434 = HEAP16[$433>>1]|0;
      $435 = $434&65535;
      $436 = (((($2)) + 112|0) + ($435<<1)|0);
      HEAP16[$436>>1] = 0;
      $exitcond = ($431|0)==(19);
      if ($exitcond) {
       break;
      } else {
       $432 = $431;
      }
     }
     HEAP32[$53>>2] = 19;
    }
    HEAP32[$58>>2] = $57;
    HEAP32[$35>>2] = $57;
    HEAP32[$36>>2] = 7;
    $437 = (_inflate_table(0,$59,19,$58,$36,$60)|0);
    $438 = ($437|0)==(0);
    if ($438) {
     HEAP32[$53>>2] = 0;
     HEAP32[$2>>2] = 18;
     $443 = 0;$bits$25$ph = $bits$23$lcssa;$have$29$ph = $have$27$lcssa;$hold$25$ph = $hold$23$lcssa;$next$29$ph = $next$27$lcssa;$ret$1$ph = 0;
     label = 166;
     break;
    } else {
     HEAP32[$27>>2] = 612608;
     HEAP32[$2>>2] = 29;
     $bits$0$be = $bits$23$lcssa;$have$0$be = $have$27$lcssa;$hold$0$be = $hold$23$lcssa;$left$0$be = $left$0;$next$0$be = $next$27$lcssa;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $437;
     break;
    }
   }
  } while(0);
  L161: do {
   if ((label|0) == 55) {
    while(1) {
     label = 0;
     $160 = ($have$4745|0)==(0);
     if ($160) {
      $844 = $380;$845 = $left$0;$bits$53 = $bits$4747;$have$57 = 0;$hold$53 = $hold$4746;$next$57 = $next$4744;$out$4 = $out$0;$ret$8 = $ret$0;
      break L17;
     }
     $161 = (($have$4745) + -1)|0;
     $162 = ((($next$4744)) + 1|0);
     $163 = HEAP8[$next$4744>>0]|0;
     $164 = $163&255;
     $165 = $164 << $bits$4747;
     $166 = (($165) + ($hold$4746))|0;
     $167 = (($bits$4747) + 8)|0;
     $168 = ($167>>>0)<(16);
     if ($168) {
      $bits$4747 = $167;$have$4745 = $161;$hold$4746 = $166;$next$4744 = $162;
      label = 55;
     } else {
      $have$4$lcssa = $161;$hold$4$lcssa = $166;$next$4$lcssa = $162;
      label = 57;
      break;
     }
    }
   }
   else if ((label|0) == 125) {
    label = 0;
    $337 = HEAP32[$34>>2]|0;
    $338 = ($337|0)==(0);
    if (!($338)) {
     $340 = $bits$17 & 7;
     $341 = $hold$17 >>> $340;
     $342 = (($bits$17) - ($340))|0;
     HEAP32[$2>>2] = 26;
     $bits$0$be = $342;$have$0$be = $have$21;$hold$0$be = $341;$left$0$be = $left$0;$next$0$be = $next$21;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$0;
     break;
    }
    $339 = ($bits$17>>>0)<(3);
    if ($339) {
     $343 = ($have$21|0)==(0);
     if ($343) {
      $$lcssa2799 = $380;$$ph = $left$0;$bits$53$ph = $bits$17;$have$57$ph = 0;$hold$53$ph = $hold$17;$next$57$ph = $next$21;$out$0$lcssa2944 = $out$0;$ret$8$ph = $ret$0;
      label = 282;
      break L17;
     }
     $344 = (($bits$17) + 8)|0;
     $345 = HEAP8[$next$21>>0]|0;
     $346 = $345&255;
     $347 = $346 << $bits$17;
     $348 = (($347) + ($hold$17))|0;
     $349 = ((($next$21)) + 1|0);
     $350 = (($have$21) + -1)|0;
     $bits$18$lcssa = $344;$have$22$lcssa = $350;$hold$18$lcssa = $348;$next$22$lcssa = $349;
    } else {
     $bits$18$lcssa = $bits$17;$have$22$lcssa = $have$21;$hold$18$lcssa = $hold$17;$next$22$lcssa = $next$21;
    }
    $351 = $hold$18$lcssa & 1;
    HEAP32[$34>>2] = $351;
    $352 = $hold$18$lcssa >>> 1;
    $353 = $352 & 3;
    switch ($353|0) {
    case 0:  {
     HEAP32[$2>>2] = 13;
     break;
    }
    case 1:  {
     HEAP32[$35>>2] = 175462;
     HEAP32[$36>>2] = 9;
     HEAP32[$37>>2] = 177510;
     HEAP32[$38>>2] = 5;
     HEAP32[$2>>2] = 19;
     if ($39) {
      $$lcssa2821 = $380;$bits$18$lcssa$lcssa3008 = $bits$18$lcssa;$have$22$lcssa$lcssa3010 = $have$22$lcssa;$hold$18$lcssa$lcssa3009 = $hold$18$lcssa;$left$0$lcssa2879 = $left$0;$next$22$lcssa$lcssa3011 = $next$22$lcssa;$out$0$lcssa2966 = $out$0;$ret$0$lcssa2995 = $ret$0;
      label = 133;
      break L17;
     }
     break;
    }
    case 2:  {
     HEAP32[$2>>2] = 16;
     break;
    }
    case 3:  {
     HEAP32[$27>>2] = 612524;
     HEAP32[$2>>2] = 29;
     break;
    }
    default: {
     label = 136;
     break L17;
    }
    }
    $356 = $hold$18$lcssa >>> 3;
    $357 = (($bits$18$lcssa) + -3)|0;
    $bits$0$be = $357;$have$0$be = $have$22$lcssa;$hold$0$be = $356;$left$0$be = $left$0;$next$0$be = $next$22$lcssa;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$0;
   }
   else if ((label|0) == 166) {
    label = 0;
    $439 = HEAP32[$54>>2]|0;
    $440 = HEAP32[$55>>2]|0;
    $441 = (($440) + ($439))|0;
    $442 = ($443>>>0)<($441>>>0);
    do {
     if ($442) {
      $470 = $443;$535 = $440;$536 = $439;$bits$25608 = $bits$25$ph;$have$29606 = $have$29$ph;$hold$25607 = $hold$25$ph;$next$29605 = $next$29$ph;
      L181: while(1) {
       $444 = HEAP32[$36>>2]|0;
       $445 = 1 << $444;
       $446 = (($445) + -1)|0;
       $447 = $446 & $hold$25607;
       $448 = HEAP32[$35>>2]|0;
       $449 = (((($448) + ($447<<2)|0)) + 1|0);
       $450 = HEAP8[$449>>0]|0;
       $451 = $450&255;
       $452 = ($bits$25608>>>0)<($451>>>0);
       if ($452) {
        $bits$26564 = $bits$25608;$have$30562 = $have$29606;$hold$26563 = $hold$25607;$next$30561 = $next$29605;
        while(1) {
         $453 = ($have$30562|0)==(0);
         if ($453) {
          $844 = $380;$845 = $left$0;$bits$53 = $bits$26564;$have$57 = 0;$hold$53 = $hold$26563;$next$57 = $next$30561;$out$4 = $out$0;$ret$8 = $ret$1$ph;
          break L17;
         }
         $454 = (($have$30562) + -1)|0;
         $455 = ((($next$30561)) + 1|0);
         $456 = HEAP8[$next$30561>>0]|0;
         $457 = $456&255;
         $458 = $457 << $bits$26564;
         $459 = (($458) + ($hold$26563))|0;
         $460 = (($bits$26564) + 8)|0;
         $461 = $446 & $459;
         $462 = (((($448) + ($461<<2)|0)) + 1|0);
         $463 = HEAP8[$462>>0]|0;
         $464 = $463&255;
         $465 = ($460>>>0)<($464>>>0);
         if ($465) {
          $bits$26564 = $460;$have$30562 = $454;$hold$26563 = $459;$next$30561 = $455;
         } else {
          $$lcssa783$pn = $464;$$pn1283 = $461;$bits$26$lcssa = $460;$have$30$lcssa = $454;$hold$26$lcssa = $459;$next$30$lcssa = $455;
          break;
         }
        }
       } else {
        $$lcssa783$pn = $451;$$pn1283 = $447;$bits$26$lcssa = $bits$25608;$have$30$lcssa = $have$29606;$hold$26$lcssa = $hold$25607;$next$30$lcssa = $next$29605;
       }
       $$lcssa101$in = (((($448) + ($$pn1283<<2)|0)) + 2|0);
       $$lcssa101 = HEAP16[$$lcssa101$in>>1]|0;
       $466 = ($$lcssa101&65535)<(16);
       if ($466) {
        $467 = $hold$26$lcssa >>> $$lcssa783$pn;
        $468 = (($bits$26$lcssa) - ($$lcssa783$pn))|0;
        $469 = (($470) + 1)|0;
        HEAP32[$53>>2] = $469;
        $471 = (((($2)) + 112|0) + ($470<<1)|0);
        HEAP16[$471>>1] = $$lcssa101;
        $bits$25$be = $468;$have$29$be = $have$30$lcssa;$hold$25$be = $467;$next$29$be = $next$30$lcssa;
       } else {
        switch ($$lcssa101<<16>>16) {
        case 16:  {
         $479 = (($$lcssa783$pn) + 2)|0;
         $480 = ($bits$26$lcssa>>>0)<($479>>>0);
         if ($480) {
          $bits$27587 = $bits$26$lcssa;$have$31585 = $have$30$lcssa;$hold$27586 = $hold$26$lcssa;$next$31584 = $next$30$lcssa;
          while(1) {
           $483 = ($have$31585|0)==(0);
           if ($483) {
            $844 = $380;$845 = $left$0;$bits$53 = $bits$27587;$have$57 = 0;$hold$53 = $hold$27586;$next$57 = $next$31584;$out$4 = $out$0;$ret$8 = $ret$1$ph;
            break L17;
           }
           $484 = (($have$31585) + -1)|0;
           $485 = ((($next$31584)) + 1|0);
           $486 = HEAP8[$next$31584>>0]|0;
           $487 = $486&255;
           $488 = $487 << $bits$27587;
           $489 = (($488) + ($hold$27586))|0;
           $490 = (($bits$27587) + 8)|0;
           $491 = ($490>>>0)<($479>>>0);
           if ($491) {
            $bits$27587 = $490;$have$31585 = $484;$hold$27586 = $489;$next$31584 = $485;
           } else {
            $bits$27$lcssa = $490;$have$31$lcssa = $484;$hold$27$lcssa = $489;$next$31$lcssa = $485;
            break;
           }
          }
         } else {
          $bits$27$lcssa = $bits$26$lcssa;$have$31$lcssa = $have$30$lcssa;$hold$27$lcssa = $hold$26$lcssa;$next$31$lcssa = $next$30$lcssa;
         }
         $492 = $hold$27$lcssa >>> $$lcssa783$pn;
         $493 = (($bits$27$lcssa) - ($$lcssa783$pn))|0;
         $494 = ($470|0)==(0);
         if ($494) {
          $$lcssa2657 = $492;$$lcssa2658 = $493;$have$31$lcssa$lcssa = $have$31$lcssa;$next$31$lcssa$lcssa = $next$31$lcssa;
          label = 180;
          break L181;
         }
         $495 = (($470) + -1)|0;
         $496 = (((($2)) + 112|0) + ($495<<1)|0);
         $497 = HEAP16[$496>>1]|0;
         $498 = $497&65535;
         $499 = $492 & 3;
         $500 = (($499) + 3)|0;
         $501 = $492 >>> 2;
         $502 = (($493) + -2)|0;
         $bits$30 = $502;$copy$5 = $500;$have$34 = $have$31$lcssa;$hold$30 = $501;$len$0 = $498;$next$34 = $next$31$lcssa;
         break;
        }
        case 17:  {
         $477 = (($$lcssa783$pn) + 3)|0;
         $478 = ($bits$26$lcssa>>>0)<($477>>>0);
         if ($478) {
          $bits$28577 = $bits$26$lcssa;$have$32575 = $have$30$lcssa;$hold$28576 = $hold$26$lcssa;$next$32574 = $next$30$lcssa;
          while(1) {
           $503 = ($have$32575|0)==(0);
           if ($503) {
            $844 = $380;$845 = $left$0;$bits$53 = $bits$28577;$have$57 = 0;$hold$53 = $hold$28576;$next$57 = $next$32574;$out$4 = $out$0;$ret$8 = $ret$1$ph;
            break L17;
           }
           $504 = (($have$32575) + -1)|0;
           $505 = ((($next$32574)) + 1|0);
           $506 = HEAP8[$next$32574>>0]|0;
           $507 = $506&255;
           $508 = $507 << $bits$28577;
           $509 = (($508) + ($hold$28576))|0;
           $510 = (($bits$28577) + 8)|0;
           $511 = ($510>>>0)<($477>>>0);
           if ($511) {
            $bits$28577 = $510;$have$32575 = $504;$hold$28576 = $509;$next$32574 = $505;
           } else {
            $bits$28$lcssa = $510;$have$32$lcssa = $504;$hold$28$lcssa = $509;$next$32$lcssa = $505;
            break;
           }
          }
         } else {
          $bits$28$lcssa = $bits$26$lcssa;$have$32$lcssa = $have$30$lcssa;$hold$28$lcssa = $hold$26$lcssa;$next$32$lcssa = $next$30$lcssa;
         }
         $512 = $hold$28$lcssa >>> $$lcssa783$pn;
         $513 = $512 & 7;
         $514 = (($513) + 3)|0;
         $515 = $512 >>> 3;
         $516 = (-3 - ($$lcssa783$pn))|0;
         $517 = (($516) + ($bits$28$lcssa))|0;
         $bits$30 = $517;$copy$5 = $514;$have$34 = $have$32$lcssa;$hold$30 = $515;$len$0 = 0;$next$34 = $next$32$lcssa;
         break;
        }
        default: {
         $481 = (($$lcssa783$pn) + 7)|0;
         $482 = ($bits$26$lcssa>>>0)<($481>>>0);
         if ($482) {
          $bits$29597 = $bits$26$lcssa;$have$33595 = $have$30$lcssa;$hold$29596 = $hold$26$lcssa;$next$33594 = $next$30$lcssa;
          while(1) {
           $518 = ($have$33595|0)==(0);
           if ($518) {
            $844 = $380;$845 = $left$0;$bits$53 = $bits$29597;$have$57 = 0;$hold$53 = $hold$29596;$next$57 = $next$33594;$out$4 = $out$0;$ret$8 = $ret$1$ph;
            break L17;
           }
           $519 = (($have$33595) + -1)|0;
           $520 = ((($next$33594)) + 1|0);
           $521 = HEAP8[$next$33594>>0]|0;
           $522 = $521&255;
           $523 = $522 << $bits$29597;
           $524 = (($523) + ($hold$29596))|0;
           $525 = (($bits$29597) + 8)|0;
           $526 = ($525>>>0)<($481>>>0);
           if ($526) {
            $bits$29597 = $525;$have$33595 = $519;$hold$29596 = $524;$next$33594 = $520;
           } else {
            $bits$29$lcssa = $525;$have$33$lcssa = $519;$hold$29$lcssa = $524;$next$33$lcssa = $520;
            break;
           }
          }
         } else {
          $bits$29$lcssa = $bits$26$lcssa;$have$33$lcssa = $have$30$lcssa;$hold$29$lcssa = $hold$26$lcssa;$next$33$lcssa = $next$30$lcssa;
         }
         $527 = $hold$29$lcssa >>> $$lcssa783$pn;
         $528 = $527 & 127;
         $529 = (($528) + 11)|0;
         $530 = $527 >>> 7;
         $531 = (-7 - ($$lcssa783$pn))|0;
         $532 = (($531) + ($bits$29$lcssa))|0;
         $bits$30 = $532;$copy$5 = $529;$have$34 = $have$33$lcssa;$hold$30 = $530;$len$0 = 0;$next$34 = $next$33$lcssa;
        }
        }
        $533 = (($470) + ($copy$5))|0;
        $534 = (($535) + ($536))|0;
        $537 = ($533>>>0)>($534>>>0);
        if ($537) {
         $bits$30$lcssa = $bits$30;$have$34$lcssa = $have$34;$hold$30$lcssa = $hold$30;$next$34$lcssa = $next$34;
         label = 190;
         break;
        }
        $538 = $len$0&65535;
        $539 = (($copy$5) + -1)|0;
        $540 = (($470) + 1)|0;
        HEAP32[$53>>2] = $540;
        $541 = (((($2)) + 112|0) + ($470<<1)|0);
        HEAP16[$541>>1] = $538;
        $542 = ($539|0)==(0);
        if ($542) {
         $bits$25$be = $bits$30;$have$29$be = $have$34;$hold$25$be = $hold$30;$next$29$be = $next$34;
        } else {
         $544 = $539;
         while(1) {
          $$pre1266 = HEAP32[$53>>2]|0;
          $543 = (($544) + -1)|0;
          $545 = (($$pre1266) + 1)|0;
          HEAP32[$53>>2] = $545;
          $546 = (((($2)) + 112|0) + ($$pre1266<<1)|0);
          HEAP16[$546>>1] = $538;
          $547 = ($543|0)==(0);
          if ($547) {
           $bits$25$be = $bits$30;$have$29$be = $have$34;$hold$25$be = $hold$30;$next$29$be = $next$34;
           break;
          } else {
           $544 = $543;
          }
         }
        }
       }
       $472 = HEAP32[$53>>2]|0;
       $473 = HEAP32[$54>>2]|0;
       $474 = HEAP32[$55>>2]|0;
       $475 = (($474) + ($473))|0;
       $476 = ($472>>>0)<($475>>>0);
       if ($476) {
        $470 = $472;$535 = $474;$536 = $473;$bits$25608 = $bits$25$be;$have$29606 = $have$29$be;$hold$25607 = $hold$25$be;$next$29605 = $next$29$be;
       } else {
        $$lcssa2659 = $473;$bits$25$be$lcssa = $bits$25$be;$have$29$be$lcssa = $have$29$be;$hold$25$be$lcssa = $hold$25$be;$next$29$be$lcssa = $next$29$be;
        label = 192;
        break;
       }
      }
      if ((label|0) == 180) {
       label = 0;
       HEAP32[$27>>2] = 612633;
       HEAP32[$2>>2] = 29;
       $bits$0$be = $$lcssa2658;$have$0$be = $have$31$lcssa$lcssa;$hold$0$be = $$lcssa2657;$left$0$be = $left$0;$next$0$be = $next$31$lcssa$lcssa;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$1$ph;
       break L161;
      }
      else if ((label|0) == 190) {
       label = 0;
       HEAP32[$27>>2] = 612633;
       HEAP32[$2>>2] = 29;
       $bits$0$be = $bits$30$lcssa;$have$0$be = $have$34$lcssa;$hold$0$be = $hold$30$lcssa;$left$0$be = $left$0;$next$0$be = $next$34$lcssa;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$1$ph;
       break L161;
      }
      else if ((label|0) == 192) {
       label = 0;
       $$pr21$pre = HEAP32[$2>>2]|0;
       $548 = ($$pr21$pre|0)==(29);
       if ($548) {
        $bits$0$be = $bits$25$be$lcssa;$have$0$be = $have$29$be$lcssa;$hold$0$be = $hold$25$be$lcssa;$left$0$be = $left$0;$next$0$be = $next$29$be$lcssa;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$1$ph;
        break L161;
       } else {
        $551 = $$lcssa2659;$bits$25$lcssa1296 = $bits$25$be$lcssa;$have$29$lcssa1298 = $have$29$be$lcssa;$hold$25$lcssa1297 = $hold$25$be$lcssa;$next$29$lcssa1299 = $next$29$be$lcssa;
        break;
       }
      }
     } else {
      $551 = $439;$bits$25$lcssa1296 = $bits$25$ph;$have$29$lcssa1298 = $have$29$ph;$hold$25$lcssa1297 = $hold$25$ph;$next$29$lcssa1299 = $next$29$ph;
     }
    } while(0);
    $549 = HEAP16[$56>>1]|0;
    $550 = ($549<<16>>16)==(0);
    if ($550) {
     HEAP32[$27>>2] = 612659;
     HEAP32[$2>>2] = 29;
     $bits$0$be = $bits$25$lcssa1296;$have$0$be = $have$29$lcssa1298;$hold$0$be = $hold$25$lcssa1297;$left$0$be = $left$0;$next$0$be = $next$29$lcssa1299;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$1$ph;
     break;
    }
    HEAP32[$58>>2] = $57;
    HEAP32[$35>>2] = $57;
    HEAP32[$36>>2] = 9;
    $552 = (_inflate_table(1,$59,$551,$58,$36,$60)|0);
    $553 = ($552|0)==(0);
    if (!($553)) {
     HEAP32[$27>>2] = 612696;
     HEAP32[$2>>2] = 29;
     $bits$0$be = $bits$25$lcssa1296;$have$0$be = $have$29$lcssa1298;$hold$0$be = $hold$25$lcssa1297;$left$0$be = $left$0;$next$0$be = $next$29$lcssa1299;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $552;
     break;
    }
    $554 = HEAP32[$58>>2]|0;
    HEAP32[$37>>2] = $554;
    HEAP32[$38>>2] = 6;
    $555 = HEAP32[$54>>2]|0;
    $556 = (((($2)) + 112|0) + ($555<<1)|0);
    $557 = HEAP32[$55>>2]|0;
    $558 = (_inflate_table(2,$556,$557,$58,$38,$60)|0);
    $559 = ($558|0)==(0);
    if ($559) {
     HEAP32[$2>>2] = 19;
     if ($39) {
      $$lcssa2799 = $380;$$ph = $left$0;$bits$53$ph = $bits$25$lcssa1296;$have$57$ph = $have$29$lcssa1298;$hold$53$ph = $hold$25$lcssa1297;$next$57$ph = $next$29$lcssa1299;$out$0$lcssa2944 = $out$0;$ret$8$ph = 0;
      label = 282;
      break L17;
     } else {
      $bits$32 = $bits$25$lcssa1296;$have$36 = $have$29$lcssa1298;$hold$32 = $hold$25$lcssa1297;$next$36 = $next$29$lcssa1299;$ret$2 = 0;
      label = 200;
      break;
     }
    } else {
     HEAP32[$27>>2] = 612724;
     HEAP32[$2>>2] = 29;
     $bits$0$be = $bits$25$lcssa1296;$have$0$be = $have$29$lcssa1298;$hold$0$be = $hold$25$lcssa1297;$left$0$be = $left$0;$next$0$be = $next$29$lcssa1299;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $558;
     break;
    }
   }
  } while(0);
  if ((label|0) == 57) {
   label = 0;
   $169 = HEAP32[$26>>2]|0;
   $170 = ($169|0)==(0|0);
   if (!($170)) {
    $171 = $hold$4$lcssa & 255;
    $172 = ((($169)) + 8|0);
    HEAP32[$172>>2] = $171;
    $173 = $hold$4$lcssa >>> 8;
    $174 = ((($169)) + 12|0);
    HEAP32[$174>>2] = $173;
   }
   $175 = HEAP32[$25>>2]|0;
   $176 = $175 & 512;
   $177 = ($176|0)==(0);
   if (!($177)) {
    $178 = $hold$4$lcssa&255;
    HEAP8[$hbuf>>0] = $178;
    $179 = $hold$4$lcssa >>> 8;
    $180 = $179&255;
    HEAP8[$24>>0] = $180;
    $181 = HEAP32[$23>>2]|0;
    $182 = (_crc32($181,$hbuf,2)|0);
    HEAP32[$23>>2] = $182;
   }
   HEAP32[$2>>2] = 4;
   $bits$5 = 0;$have$5 = $have$4$lcssa;$hold$5 = 0;$next$5 = $next$4$lcssa;
   label = 62;
  }
  else if ((label|0) == 200) {
   label = 0;
   HEAP32[$2>>2] = 20;
   $bits$33 = $bits$32;$have$37 = $have$36;$hold$33 = $hold$32;$next$37 = $next$36;$ret$3 = $ret$2;
   label = 201;
  }
  do {
   if ((label|0) == 62) {
    label = 0;
    $183 = HEAP32[$25>>2]|0;
    $184 = $183 & 1024;
    $185 = ($184|0)==(0);
    if ($185) {
     $206 = HEAP32[$26>>2]|0;
     $207 = ($206|0)==(0|0);
     if ($207) {
      $bits$7 = $bits$5;$have$7 = $have$5;$hold$7 = $hold$5;$next$7 = $next$5;
     } else {
      $208 = ((($206)) + 16|0);
      HEAP32[$208>>2] = 0;
      $bits$7 = $bits$5;$have$7 = $have$5;$hold$7 = $hold$5;$next$7 = $next$5;
     }
    } else {
     $186 = ($bits$5>>>0)<(16);
     if ($186) {
      $bits$6756 = $bits$5;$have$6754 = $have$5;$hold$6755 = $hold$5;$next$6753 = $next$5;
      while(1) {
       $187 = ($have$6754|0)==(0);
       if ($187) {
        $844 = $380;$845 = $left$0;$bits$53 = $bits$6756;$have$57 = 0;$hold$53 = $hold$6755;$next$57 = $next$6753;$out$4 = $out$0;$ret$8 = $ret$0;
        break L17;
       }
       $188 = (($have$6754) + -1)|0;
       $189 = ((($next$6753)) + 1|0);
       $190 = HEAP8[$next$6753>>0]|0;
       $191 = $190&255;
       $192 = $191 << $bits$6756;
       $193 = (($192) + ($hold$6755))|0;
       $194 = (($bits$6756) + 8)|0;
       $195 = ($194>>>0)<(16);
       if ($195) {
        $bits$6756 = $194;$have$6754 = $188;$hold$6755 = $193;$next$6753 = $189;
       } else {
        $have$6$lcssa = $188;$hold$6$lcssa = $193;$next$6$lcssa = $189;
        break;
       }
      }
     } else {
      $have$6$lcssa = $have$5;$hold$6$lcssa = $hold$5;$next$6$lcssa = $next$5;
     }
     HEAP32[$31>>2] = $hold$6$lcssa;
     $196 = HEAP32[$26>>2]|0;
     $197 = ($196|0)==(0|0);
     if (!($197)) {
      $198 = ((($196)) + 20|0);
      HEAP32[$198>>2] = $hold$6$lcssa;
     }
     $199 = $183 & 512;
     $200 = ($199|0)==(0);
     if ($200) {
      $bits$7 = 0;$have$7 = $have$6$lcssa;$hold$7 = 0;$next$7 = $next$6$lcssa;
     } else {
      $201 = $hold$6$lcssa&255;
      HEAP8[$hbuf>>0] = $201;
      $202 = $hold$6$lcssa >>> 8;
      $203 = $202&255;
      HEAP8[$24>>0] = $203;
      $204 = HEAP32[$23>>2]|0;
      $205 = (_crc32($204,$hbuf,2)|0);
      HEAP32[$23>>2] = $205;
      $bits$7 = 0;$have$7 = $have$6$lcssa;$hold$7 = 0;$next$7 = $next$6$lcssa;
     }
    }
    HEAP32[$2>>2] = 5;
    $bits$8 = $bits$7;$have$8 = $have$7;$hold$8 = $hold$7;$next$8 = $next$7;
    label = 73;
   }
   else if ((label|0) == 201) {
    label = 0;
    $560 = ($have$37>>>0)>(5);
    $561 = ($left$0>>>0)>(257);
    $or$cond6 = $561 & $560;
    if ($or$cond6) {
     HEAP32[$4>>2] = $380;
     HEAP32[$14>>2] = $left$0;
     HEAP32[$strm>>2] = $next$37;
     HEAP32[$16>>2] = $have$37;
     HEAP32[$18>>2] = $hold$33;
     HEAP32[$20>>2] = $bits$33;
     _inflate_fast($strm,$out$0);
     $562 = HEAP32[$4>>2]|0;
     $563 = HEAP32[$14>>2]|0;
     $564 = HEAP32[$strm>>2]|0;
     $565 = HEAP32[$16>>2]|0;
     $566 = HEAP32[$18>>2]|0;
     $567 = HEAP32[$20>>2]|0;
     $568 = HEAP32[$2>>2]|0;
     $569 = ($568|0)==(11);
     if (!($569)) {
      $bits$0$be = $567;$have$0$be = $565;$hold$0$be = $566;$left$0$be = $563;$next$0$be = $564;$out$0$be = $out$0;$put$0$be = $562;$ret$0$be = $ret$3;
      break;
     }
     HEAP32[$40>>2] = -1;
     $bits$0$be = $567;$have$0$be = $565;$hold$0$be = $566;$left$0$be = $563;$next$0$be = $564;$out$0$be = $out$0;$put$0$be = $562;$ret$0$be = $ret$3;
     break;
    }
    HEAP32[$40>>2] = 0;
    $570 = HEAP32[$36>>2]|0;
    $571 = 1 << $570;
    $572 = (($571) + -1)|0;
    $573 = $572 & $hold$33;
    $574 = HEAP32[$35>>2]|0;
    $575 = (((($574) + ($573<<2)|0)) + 1|0);
    $576 = HEAP8[$575>>0]|0;
    $577 = $576&255;
    $578 = ($577>>>0)>($bits$33>>>0);
    if ($578) {
     $bits$34619 = $bits$33;$have$38617 = $have$37;$hold$34618 = $hold$33;$next$38616 = $next$37;
     while(1) {
      $579 = ($have$38617|0)==(0);
      if ($579) {
       $844 = $380;$845 = $left$0;$bits$53 = $bits$34619;$have$57 = 0;$hold$53 = $hold$34618;$next$57 = $next$38616;$out$4 = $out$0;$ret$8 = $ret$3;
       break L17;
      }
      $580 = (($have$38617) + -1)|0;
      $581 = ((($next$38616)) + 1|0);
      $582 = HEAP8[$next$38616>>0]|0;
      $583 = $582&255;
      $584 = $583 << $bits$34619;
      $585 = (($584) + ($hold$34618))|0;
      $586 = (($bits$34619) + 8)|0;
      $587 = $572 & $585;
      $588 = (((($574) + ($587<<2)|0)) + 1|0);
      $589 = HEAP8[$588>>0]|0;
      $590 = $589&255;
      $591 = ($590>>>0)>($586>>>0);
      if ($591) {
       $bits$34619 = $586;$have$38617 = $580;$hold$34618 = $585;$next$38616 = $581;
      } else {
       $$lcssa166 = $589;$$lcssa170 = $590;$$pn1281 = $587;$bits$34$lcssa = $586;$have$38$lcssa = $580;$hold$34$lcssa = $585;$next$38$lcssa = $581;
       break;
      }
     }
    } else {
     $$lcssa166 = $576;$$lcssa170 = $577;$$pn1281 = $573;$bits$34$lcssa = $bits$33;$have$38$lcssa = $have$37;$hold$34$lcssa = $hold$33;$next$38$lcssa = $next$37;
    }
    $$lcssa164$in = (($574) + ($$pn1281<<2)|0);
    $$lcssa168$in = (((($574) + ($$pn1281<<2)|0)) + 2|0);
    $$lcssa164 = HEAP8[$$lcssa164$in>>0]|0;
    $$lcssa168 = HEAP16[$$lcssa168$in>>1]|0;
    $592 = $$lcssa164&255;
    $593 = ($$lcssa164<<24>>24)!=(0);
    $594 = $592 & 240;
    $595 = ($594|0)==(0);
    $or$cond20 = $593 & $595;
    if ($or$cond20) {
     $596 = $$lcssa168&65535;
     $597 = (($$lcssa170) + ($592))|0;
     $598 = 1 << $597;
     $599 = (($598) + -1)|0;
     $600 = $hold$34$lcssa & $599;
     $601 = $600 >>> $$lcssa170;
     $602 = (($601) + ($596))|0;
     $603 = (((($574) + ($602<<2)|0)) + 1|0);
     $604 = HEAP8[$603>>0]|0;
     $605 = $604&255;
     $606 = (($605) + ($$lcssa170))|0;
     $607 = ($606>>>0)>($bits$34$lcssa>>>0);
     if ($607) {
      $bits$35635 = $bits$34$lcssa;$have$39633 = $have$38$lcssa;$hold$35634 = $hold$34$lcssa;$next$39632 = $next$38$lcssa;
      while(1) {
       $608 = ($have$39633|0)==(0);
       if ($608) {
        $844 = $380;$845 = $left$0;$bits$53 = $bits$35635;$have$57 = 0;$hold$53 = $hold$35634;$next$57 = $next$39632;$out$4 = $out$0;$ret$8 = $ret$3;
        break L17;
       }
       $609 = (($have$39633) + -1)|0;
       $610 = ((($next$39632)) + 1|0);
       $611 = HEAP8[$next$39632>>0]|0;
       $612 = $611&255;
       $613 = $612 << $bits$35635;
       $614 = (($613) + ($hold$35634))|0;
       $615 = (($bits$35635) + 8)|0;
       $616 = $614 & $599;
       $617 = $616 >>> $$lcssa170;
       $618 = (($617) + ($596))|0;
       $619 = (((($574) + ($618<<2)|0)) + 1|0);
       $620 = HEAP8[$619>>0]|0;
       $621 = $620&255;
       $622 = (($621) + ($$lcssa170))|0;
       $623 = ($622>>>0)>($615>>>0);
       if ($623) {
        $bits$35635 = $615;$have$39633 = $609;$hold$35634 = $614;$next$39632 = $610;
       } else {
        $$lcssa176 = $618;$$lcssa180 = $620;$bits$35$lcssa = $615;$have$39$lcssa = $609;$hold$35$lcssa = $614;$next$39$lcssa = $610;
        break;
       }
      }
     } else {
      $$lcssa176 = $602;$$lcssa180 = $604;$bits$35$lcssa = $bits$34$lcssa;$have$39$lcssa = $have$38$lcssa;$hold$35$lcssa = $hold$34$lcssa;$next$39$lcssa = $next$38$lcssa;
     }
     $624 = (((($574) + ($$lcssa176<<2)|0)) + 2|0);
     $625 = HEAP16[$624>>1]|0;
     $626 = (($574) + ($$lcssa176<<2)|0);
     $627 = HEAP8[$626>>0]|0;
     $628 = $hold$35$lcssa >>> $$lcssa170;
     $629 = (($bits$35$lcssa) - ($$lcssa170))|0;
     HEAP32[$40>>2] = $$lcssa170;
     $634 = $$lcssa170;$bits$36 = $629;$have$40 = $have$39$lcssa;$here$sroa$0$0 = $627;$here$sroa$18$0 = $$lcssa180;$here$sroa$42$0 = $625;$hold$36 = $628;$next$40 = $next$39$lcssa;
    } else {
     $634 = 0;$bits$36 = $bits$34$lcssa;$have$40 = $have$38$lcssa;$here$sroa$0$0 = $$lcssa164;$here$sroa$18$0 = $$lcssa166;$here$sroa$42$0 = $$lcssa168;$hold$36 = $hold$34$lcssa;$next$40 = $next$38$lcssa;
    }
    $630 = $here$sroa$18$0&255;
    $631 = $hold$36 >>> $630;
    $632 = (($bits$36) - ($630))|0;
    $633 = (($634) + ($630))|0;
    HEAP32[$40>>2] = $633;
    $635 = $here$sroa$42$0&65535;
    HEAP32[$31>>2] = $635;
    $636 = $here$sroa$0$0&255;
    $637 = ($here$sroa$0$0<<24>>24)==(0);
    if ($637) {
     HEAP32[$2>>2] = 25;
     $bits$0$be = $632;$have$0$be = $have$40;$hold$0$be = $631;$left$0$be = $left$0;$next$0$be = $next$40;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$3;
     break;
    }
    $638 = $636 & 32;
    $639 = ($638|0)==(0);
    if (!($639)) {
     HEAP32[$40>>2] = -1;
     HEAP32[$2>>2] = 11;
     $bits$0$be = $632;$have$0$be = $have$40;$hold$0$be = $631;$left$0$be = $left$0;$next$0$be = $next$40;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$3;
     break;
    }
    $640 = $636 & 64;
    $641 = ($640|0)==(0);
    if ($641) {
     $642 = $636 & 15;
     HEAP32[$41>>2] = $642;
     HEAP32[$2>>2] = 21;
     $644 = $642;$bits$37 = $632;$have$41 = $have$40;$hold$37 = $631;$next$41 = $next$40;$ret$4 = $ret$3;
     label = 219;
     break;
    } else {
     HEAP32[$27>>2] = 613751;
     HEAP32[$2>>2] = 29;
     $bits$0$be = $632;$have$0$be = $have$40;$hold$0$be = $631;$left$0$be = $left$0;$next$0$be = $next$40;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$3;
     break;
    }
   }
  } while(0);
  if ((label|0) == 73) {
   label = 0;
   $209 = HEAP32[$25>>2]|0;
   $210 = $209 & 1024;
   $211 = ($210|0)==(0);
   if ($211) {
    $have$10 = $have$8;$next$10 = $next$8;
   } else {
    $212 = HEAP32[$31>>2]|0;
    $213 = ($212>>>0)>($have$8>>>0);
    $have$8$ = $213 ? $have$8 : $212;
    $214 = ($have$8$|0)==(0);
    if ($214) {
     $240 = $212;$have$9 = $have$8;$next$9 = $next$8;
    } else {
     $215 = HEAP32[$26>>2]|0;
     $216 = ($215|0)==(0|0);
     if ($216) {
      $231 = $209;
     } else {
      $217 = ((($215)) + 16|0);
      $218 = HEAP32[$217>>2]|0;
      $219 = ($218|0)==(0|0);
      if ($219) {
       $231 = $209;
      } else {
       $220 = ((($215)) + 20|0);
       $221 = HEAP32[$220>>2]|0;
       $222 = (($221) - ($212))|0;
       $223 = (($218) + ($222)|0);
       $224 = (($222) + ($have$8$))|0;
       $225 = ((($215)) + 24|0);
       $226 = HEAP32[$225>>2]|0;
       $227 = ($224>>>0)>($226>>>0);
       $228 = (($226) - ($222))|0;
       $229 = $227 ? $228 : $have$8$;
       _memcpy(($223|0),($next$8|0),($229|0))|0;
       $$pre1276 = HEAP32[$25>>2]|0;
       $231 = $$pre1276;
      }
     }
     $230 = $231 & 512;
     $232 = ($230|0)==(0);
     if (!($232)) {
      $233 = HEAP32[$23>>2]|0;
      $234 = (_crc32($233,$next$8,$have$8$)|0);
      HEAP32[$23>>2] = $234;
     }
     $235 = (($have$8) - ($have$8$))|0;
     $236 = (($next$8) + ($have$8$)|0);
     $237 = HEAP32[$31>>2]|0;
     $238 = (($237) - ($have$8$))|0;
     HEAP32[$31>>2] = $238;
     $240 = $238;$have$9 = $235;$next$9 = $236;
    }
    $239 = ($240|0)==(0);
    if ($239) {
     $have$10 = $have$9;$next$10 = $next$9;
    } else {
     $$lcssa2799 = $380;$$ph = $left$0;$bits$53$ph = $bits$8;$have$57$ph = $have$9;$hold$53$ph = $hold$8;$next$57$ph = $next$9;$out$0$lcssa2944 = $out$0;$ret$8$ph = $ret$0;
     label = 282;
     break;
    }
   }
   HEAP32[$31>>2] = 0;
   HEAP32[$2>>2] = 6;
   $bits$9 = $bits$8;$have$11 = $have$10;$hold$9 = $hold$8;$next$11 = $next$10;
   label = 83;
  }
  else if ((label|0) == 219) {
   label = 0;
   $643 = ($644|0)==(0);
   if ($643) {
    $$pre1271 = HEAP32[$31>>2]|0;
    $664 = $$pre1271;$bits$39 = $bits$37;$have$43 = $have$41;$hold$39 = $hold$37;$next$43 = $next$41;
   } else {
    $645 = ($bits$37>>>0)<($644>>>0);
    if ($645) {
     $bits$38648 = $bits$37;$have$42646 = $have$41;$hold$38647 = $hold$37;$next$42645 = $next$41;
     while(1) {
      $646 = ($have$42646|0)==(0);
      if ($646) {
       $844 = $380;$845 = $left$0;$bits$53 = $bits$38648;$have$57 = 0;$hold$53 = $hold$38647;$next$57 = $next$42645;$out$4 = $out$0;$ret$8 = $ret$4;
       break L17;
      }
      $647 = (($have$42646) + -1)|0;
      $648 = ((($next$42645)) + 1|0);
      $649 = HEAP8[$next$42645>>0]|0;
      $650 = $649&255;
      $651 = $650 << $bits$38648;
      $652 = (($651) + ($hold$38647))|0;
      $653 = (($bits$38648) + 8)|0;
      $654 = ($653>>>0)<($644>>>0);
      if ($654) {
       $bits$38648 = $653;$have$42646 = $647;$hold$38647 = $652;$next$42645 = $648;
      } else {
       $bits$38$lcssa = $653;$have$42$lcssa = $647;$hold$38$lcssa = $652;$next$42$lcssa = $648;
       break;
      }
     }
    } else {
     $bits$38$lcssa = $bits$37;$have$42$lcssa = $have$41;$hold$38$lcssa = $hold$37;$next$42$lcssa = $next$41;
    }
    $655 = 1 << $644;
    $656 = (($655) + -1)|0;
    $657 = $656 & $hold$38$lcssa;
    $658 = HEAP32[$31>>2]|0;
    $659 = (($658) + ($657))|0;
    HEAP32[$31>>2] = $659;
    $660 = $hold$38$lcssa >>> $644;
    $661 = (($bits$38$lcssa) - ($644))|0;
    $662 = HEAP32[$40>>2]|0;
    $663 = (($662) + ($644))|0;
    HEAP32[$40>>2] = $663;
    $664 = $659;$bits$39 = $661;$have$43 = $have$42$lcssa;$hold$39 = $660;$next$43 = $next$42$lcssa;
   }
   HEAP32[$42>>2] = $664;
   HEAP32[$2>>2] = 22;
   $bits$40$ph = $bits$39;$have$44$ph = $have$43;$hold$40$ph = $hold$39;$next$44$ph = $next$43;$ret$5$ph = $ret$4;
   label = 226;
  }
  do {
   if ((label|0) == 83) {
    label = 0;
    $241 = HEAP32[$25>>2]|0;
    $242 = $241 & 2048;
    $243 = ($242|0)==(0);
    if ($243) {
     $268 = HEAP32[$26>>2]|0;
     $269 = ($268|0)==(0|0);
     if ($269) {
      $have$12 = $have$11;$next$12 = $next$11;
     } else {
      $270 = ((($268)) + 28|0);
      HEAP32[$270>>2] = 0;
      $have$12 = $have$11;$next$12 = $next$11;
     }
    } else {
     $244 = ($have$11|0)==(0);
     if ($244) {
      $$lcssa2799 = $380;$$ph = $left$0;$bits$53$ph = $bits$9;$have$57$ph = 0;$hold$53$ph = $hold$9;$next$57$ph = $next$11;$out$0$lcssa2944 = $out$0;$ret$8$ph = $ret$0;
      label = 282;
      break L17;
     } else {
      $copy$1 = 0;
     }
     while(1) {
      $245 = (($copy$1) + 1)|0;
      $246 = (($next$11) + ($copy$1)|0);
      $247 = HEAP8[$246>>0]|0;
      $248 = HEAP32[$26>>2]|0;
      $249 = ($248|0)==(0|0);
      if (!($249)) {
       $250 = ((($248)) + 28|0);
       $251 = HEAP32[$250>>2]|0;
       $252 = ($251|0)==(0|0);
       if (!($252)) {
        $253 = HEAP32[$31>>2]|0;
        $254 = ((($248)) + 32|0);
        $255 = HEAP32[$254>>2]|0;
        $256 = ($253>>>0)<($255>>>0);
        if ($256) {
         $257 = (($253) + 1)|0;
         HEAP32[$31>>2] = $257;
         $258 = (($251) + ($253)|0);
         HEAP8[$258>>0] = $247;
        }
       }
      }
      $259 = ($247<<24>>24)!=(0);
      $260 = ($have$11>>>0)>($245>>>0);
      $or$cond17 = $259 & $260;
      if ($or$cond17) {
       $copy$1 = $245;
      } else {
       $$lcssa2751 = $245;$$lcssa2752 = $259;
       break;
      }
     }
     $261 = HEAP32[$25>>2]|0;
     $262 = $261 & 512;
     $263 = ($262|0)==(0);
     if (!($263)) {
      $264 = HEAP32[$23>>2]|0;
      $265 = (_crc32($264,$next$11,$$lcssa2751)|0);
      HEAP32[$23>>2] = $265;
     }
     $266 = (($have$11) - ($$lcssa2751))|0;
     $267 = (($next$11) + ($$lcssa2751)|0);
     if ($$lcssa2752) {
      $$lcssa2799 = $380;$$ph = $left$0;$bits$53$ph = $bits$9;$have$57$ph = $266;$hold$53$ph = $hold$9;$next$57$ph = $267;$out$0$lcssa2944 = $out$0;$ret$8$ph = $ret$0;
      label = 282;
      break L17;
     } else {
      $have$12 = $266;$next$12 = $267;
     }
    }
    HEAP32[$31>>2] = 0;
    HEAP32[$2>>2] = 7;
    $bits$10 = $bits$9;$have$13 = $have$12;$hold$10 = $hold$9;$next$13 = $next$12;
    label = 96;
   }
   else if ((label|0) == 226) {
    label = 0;
    $665 = HEAP32[$38>>2]|0;
    $666 = 1 << $665;
    $667 = (($666) + -1)|0;
    $668 = $667 & $hold$40$ph;
    $669 = HEAP32[$37>>2]|0;
    $670 = (((($669) + ($668<<2)|0)) + 1|0);
    $671 = HEAP8[$670>>0]|0;
    $672 = $671&255;
    $673 = ($672>>>0)>($bits$40$ph>>>0);
    if ($673) {
     $bits$40659 = $bits$40$ph;$have$44657 = $have$44$ph;$hold$40658 = $hold$40$ph;$next$44656 = $next$44$ph;
     while(1) {
      $674 = ($have$44657|0)==(0);
      if ($674) {
       $844 = $380;$845 = $left$0;$bits$53 = $bits$40659;$have$57 = 0;$hold$53 = $hold$40658;$next$57 = $next$44656;$out$4 = $out$0;$ret$8 = $ret$5$ph;
       break L17;
      }
      $675 = (($have$44657) + -1)|0;
      $676 = ((($next$44656)) + 1|0);
      $677 = HEAP8[$next$44656>>0]|0;
      $678 = $677&255;
      $679 = $678 << $bits$40659;
      $680 = (($679) + ($hold$40658))|0;
      $681 = (($bits$40659) + 8)|0;
      $682 = $667 & $680;
      $683 = (((($669) + ($682<<2)|0)) + 1|0);
      $684 = HEAP8[$683>>0]|0;
      $685 = $684&255;
      $686 = ($685>>>0)>($681>>>0);
      if ($686) {
       $bits$40659 = $681;$have$44657 = $675;$hold$40658 = $680;$next$44656 = $676;
      } else {
       $$lcssa196 = $684;$$lcssa200 = $685;$$pn = $682;$bits$40$lcssa = $681;$have$44$lcssa = $675;$hold$40$lcssa = $680;$next$44$lcssa = $676;
       break;
      }
     }
    } else {
     $$lcssa196 = $671;$$lcssa200 = $672;$$pn = $668;$bits$40$lcssa = $bits$40$ph;$have$44$lcssa = $have$44$ph;$hold$40$lcssa = $hold$40$ph;$next$44$lcssa = $next$44$ph;
    }
    $$lcssa194$in = (($669) + ($$pn<<2)|0);
    $$lcssa198$in = (((($669) + ($$pn<<2)|0)) + 2|0);
    $$lcssa194 = HEAP8[$$lcssa194$in>>0]|0;
    $$lcssa198 = HEAP16[$$lcssa198$in>>1]|0;
    $687 = $$lcssa194&255;
    $688 = $687 & 240;
    $689 = ($688|0)==(0);
    if ($689) {
     $690 = $$lcssa198&65535;
     $691 = (($$lcssa200) + ($687))|0;
     $692 = 1 << $691;
     $693 = (($692) + -1)|0;
     $694 = $hold$40$lcssa & $693;
     $695 = $694 >>> $$lcssa200;
     $696 = (($695) + ($690))|0;
     $697 = (((($669) + ($696<<2)|0)) + 1|0);
     $698 = HEAP8[$697>>0]|0;
     $699 = $698&255;
     $700 = (($699) + ($$lcssa200))|0;
     $701 = ($700>>>0)>($bits$40$lcssa>>>0);
     if ($701) {
      $bits$41674 = $bits$40$lcssa;$have$45672 = $have$44$lcssa;$hold$41673 = $hold$40$lcssa;$next$45671 = $next$44$lcssa;
      while(1) {
       $702 = ($have$45672|0)==(0);
       if ($702) {
        $844 = $380;$845 = $left$0;$bits$53 = $bits$41674;$have$57 = 0;$hold$53 = $hold$41673;$next$57 = $next$45671;$out$4 = $out$0;$ret$8 = $ret$5$ph;
        break L17;
       }
       $703 = (($have$45672) + -1)|0;
       $704 = ((($next$45671)) + 1|0);
       $705 = HEAP8[$next$45671>>0]|0;
       $706 = $705&255;
       $707 = $706 << $bits$41674;
       $708 = (($707) + ($hold$41673))|0;
       $709 = (($bits$41674) + 8)|0;
       $710 = $708 & $693;
       $711 = $710 >>> $$lcssa200;
       $712 = (($711) + ($690))|0;
       $713 = (((($669) + ($712<<2)|0)) + 1|0);
       $714 = HEAP8[$713>>0]|0;
       $715 = $714&255;
       $716 = (($715) + ($$lcssa200))|0;
       $717 = ($716>>>0)>($709>>>0);
       if ($717) {
        $bits$41674 = $709;$have$45672 = $703;$hold$41673 = $708;$next$45671 = $704;
       } else {
        $$lcssa206 = $712;$$lcssa210 = $714;$bits$41$lcssa = $709;$have$45$lcssa = $703;$hold$41$lcssa = $708;$next$45$lcssa = $704;
        break;
       }
      }
     } else {
      $$lcssa206 = $696;$$lcssa210 = $698;$bits$41$lcssa = $bits$40$lcssa;$have$45$lcssa = $have$44$lcssa;$hold$41$lcssa = $hold$40$lcssa;$next$45$lcssa = $next$44$lcssa;
     }
     $718 = (((($669) + ($$lcssa206<<2)|0)) + 2|0);
     $719 = HEAP16[$718>>1]|0;
     $720 = (($669) + ($$lcssa206<<2)|0);
     $721 = HEAP8[$720>>0]|0;
     $722 = $hold$41$lcssa >>> $$lcssa200;
     $723 = (($bits$41$lcssa) - ($$lcssa200))|0;
     $724 = HEAP32[$40>>2]|0;
     $725 = (($724) + ($$lcssa200))|0;
     HEAP32[$40>>2] = $725;
     $730 = $725;$bits$42 = $723;$have$46 = $have$45$lcssa;$here$sroa$0$1 = $721;$here$sroa$18$1 = $$lcssa210;$here$sroa$42$1 = $719;$hold$42 = $722;$next$46 = $next$45$lcssa;
    } else {
     $$pre1273 = HEAP32[$40>>2]|0;
     $730 = $$pre1273;$bits$42 = $bits$40$lcssa;$have$46 = $have$44$lcssa;$here$sroa$0$1 = $$lcssa194;$here$sroa$18$1 = $$lcssa196;$here$sroa$42$1 = $$lcssa198;$hold$42 = $hold$40$lcssa;$next$46 = $next$44$lcssa;
    }
    $726 = $here$sroa$18$1&255;
    $727 = $hold$42 >>> $726;
    $728 = (($bits$42) - ($726))|0;
    $729 = (($730) + ($726))|0;
    HEAP32[$40>>2] = $729;
    $731 = $here$sroa$0$1&255;
    $732 = $731 & 64;
    $733 = ($732|0)==(0);
    if ($733) {
     $734 = $here$sroa$42$1&65535;
     HEAP32[$43>>2] = $734;
     $735 = $731 & 15;
     HEAP32[$41>>2] = $735;
     HEAP32[$2>>2] = 23;
     $737 = $735;$bits$43 = $728;$have$47 = $have$46;$hold$43 = $727;$next$47 = $next$46;$ret$6 = $ret$5$ph;
     label = 238;
     break;
    } else {
     HEAP32[$27>>2] = 613729;
     HEAP32[$2>>2] = 29;
     $bits$0$be = $728;$have$0$be = $have$46;$hold$0$be = $727;$left$0$be = $left$0;$next$0$be = $next$46;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$5$ph;
     break;
    }
   }
  } while(0);
  if ((label|0) == 96) {
   label = 0;
   $271 = HEAP32[$25>>2]|0;
   $272 = $271 & 4096;
   $273 = ($272|0)==(0);
   if ($273) {
    $298 = HEAP32[$26>>2]|0;
    $299 = ($298|0)==(0|0);
    if ($299) {
     $have$14 = $have$13;$next$14 = $next$13;
    } else {
     $300 = ((($298)) + 36|0);
     HEAP32[$300>>2] = 0;
     $have$14 = $have$13;$next$14 = $next$13;
    }
   } else {
    $274 = ($have$13|0)==(0);
    if ($274) {
     $$lcssa2799 = $380;$$ph = $left$0;$bits$53$ph = $bits$10;$have$57$ph = 0;$hold$53$ph = $hold$10;$next$57$ph = $next$13;$out$0$lcssa2944 = $out$0;$ret$8$ph = $ret$0;
     label = 282;
     break;
    } else {
     $copy$2 = 0;
    }
    while(1) {
     $275 = (($copy$2) + 1)|0;
     $276 = (($next$13) + ($copy$2)|0);
     $277 = HEAP8[$276>>0]|0;
     $278 = HEAP32[$26>>2]|0;
     $279 = ($278|0)==(0|0);
     if (!($279)) {
      $280 = ((($278)) + 36|0);
      $281 = HEAP32[$280>>2]|0;
      $282 = ($281|0)==(0|0);
      if (!($282)) {
       $283 = HEAP32[$31>>2]|0;
       $284 = ((($278)) + 40|0);
       $285 = HEAP32[$284>>2]|0;
       $286 = ($283>>>0)<($285>>>0);
       if ($286) {
        $287 = (($283) + 1)|0;
        HEAP32[$31>>2] = $287;
        $288 = (($281) + ($283)|0);
        HEAP8[$288>>0] = $277;
       }
      }
     }
     $289 = ($277<<24>>24)!=(0);
     $290 = ($have$13>>>0)>($275>>>0);
     $or$cond18 = $289 & $290;
     if ($or$cond18) {
      $copy$2 = $275;
     } else {
      $$lcssa2753 = $275;$$lcssa2754 = $289;
      break;
     }
    }
    $291 = HEAP32[$25>>2]|0;
    $292 = $291 & 512;
    $293 = ($292|0)==(0);
    if (!($293)) {
     $294 = HEAP32[$23>>2]|0;
     $295 = (_crc32($294,$next$13,$$lcssa2753)|0);
     HEAP32[$23>>2] = $295;
    }
    $296 = (($have$13) - ($$lcssa2753))|0;
    $297 = (($next$13) + ($$lcssa2753)|0);
    if ($$lcssa2754) {
     $$lcssa2799 = $380;$$ph = $left$0;$bits$53$ph = $bits$10;$have$57$ph = $296;$hold$53$ph = $hold$10;$next$57$ph = $297;$out$0$lcssa2944 = $out$0;$ret$8$ph = $ret$0;
     label = 282;
     break;
    } else {
     $have$14 = $296;$next$14 = $297;
    }
   }
   HEAP32[$2>>2] = 8;
   $bits$11 = $bits$10;$have$15 = $have$14;$hold$11 = $hold$10;$next$15 = $next$14;
   label = 109;
  }
  else if ((label|0) == 238) {
   label = 0;
   $736 = ($737|0)==(0);
   if ($736) {
    $bits$45 = $bits$43;$have$49 = $have$47;$hold$45 = $hold$43;$next$49 = $next$47;
   } else {
    $738 = ($bits$43>>>0)<($737>>>0);
    if ($738) {
     $bits$44687 = $bits$43;$have$48685 = $have$47;$hold$44686 = $hold$43;$next$48684 = $next$47;
     while(1) {
      $739 = ($have$48685|0)==(0);
      if ($739) {
       $844 = $380;$845 = $left$0;$bits$53 = $bits$44687;$have$57 = 0;$hold$53 = $hold$44686;$next$57 = $next$48684;$out$4 = $out$0;$ret$8 = $ret$6;
       break L17;
      }
      $740 = (($have$48685) + -1)|0;
      $741 = ((($next$48684)) + 1|0);
      $742 = HEAP8[$next$48684>>0]|0;
      $743 = $742&255;
      $744 = $743 << $bits$44687;
      $745 = (($744) + ($hold$44686))|0;
      $746 = (($bits$44687) + 8)|0;
      $747 = ($746>>>0)<($737>>>0);
      if ($747) {
       $bits$44687 = $746;$have$48685 = $740;$hold$44686 = $745;$next$48684 = $741;
      } else {
       $bits$44$lcssa = $746;$have$48$lcssa = $740;$hold$44$lcssa = $745;$next$48$lcssa = $741;
       break;
      }
     }
    } else {
     $bits$44$lcssa = $bits$43;$have$48$lcssa = $have$47;$hold$44$lcssa = $hold$43;$next$48$lcssa = $next$47;
    }
    $748 = 1 << $737;
    $749 = (($748) + -1)|0;
    $750 = $749 & $hold$44$lcssa;
    $751 = HEAP32[$43>>2]|0;
    $752 = (($751) + ($750))|0;
    HEAP32[$43>>2] = $752;
    $753 = $hold$44$lcssa >>> $737;
    $754 = (($bits$44$lcssa) - ($737))|0;
    $755 = HEAP32[$40>>2]|0;
    $756 = (($755) + ($737))|0;
    HEAP32[$40>>2] = $756;
    $bits$45 = $754;$have$49 = $have$48$lcssa;$hold$45 = $753;$next$49 = $next$48$lcssa;
   }
   HEAP32[$2>>2] = 24;
   $bits$46 = $bits$45;$have$50 = $have$49;$hold$46 = $hold$45;$next$50 = $next$49;$ret$7 = $ret$6;
   label = 244;
  }
  do {
   if ((label|0) == 109) {
    label = 0;
    $301 = HEAP32[$25>>2]|0;
    $302 = $301 & 512;
    $303 = ($302|0)==(0);
    if ($303) {
     $bits$13 = $bits$11;$have$17 = $have$15;$hold$13 = $hold$11;$next$17 = $next$15;
    } else {
     $304 = ($bits$11>>>0)<(16);
     if ($304) {
      $bits$12765 = $bits$11;$have$16763 = $have$15;$hold$12764 = $hold$11;$next$16762 = $next$15;
      while(1) {
       $305 = ($have$16763|0)==(0);
       if ($305) {
        $844 = $380;$845 = $left$0;$bits$53 = $bits$12765;$have$57 = 0;$hold$53 = $hold$12764;$next$57 = $next$16762;$out$4 = $out$0;$ret$8 = $ret$0;
        break L17;
       }
       $306 = (($have$16763) + -1)|0;
       $307 = ((($next$16762)) + 1|0);
       $308 = HEAP8[$next$16762>>0]|0;
       $309 = $308&255;
       $310 = $309 << $bits$12765;
       $311 = (($310) + ($hold$12764))|0;
       $312 = (($bits$12765) + 8)|0;
       $313 = ($312>>>0)<(16);
       if ($313) {
        $bits$12765 = $312;$have$16763 = $306;$hold$12764 = $311;$next$16762 = $307;
       } else {
        $bits$12$lcssa = $312;$have$16$lcssa = $306;$hold$12$lcssa = $311;$next$16$lcssa = $307;
        break;
       }
      }
     } else {
      $bits$12$lcssa = $bits$11;$have$16$lcssa = $have$15;$hold$12$lcssa = $hold$11;$next$16$lcssa = $next$15;
     }
     $314 = HEAP32[$23>>2]|0;
     $315 = $314 & 65535;
     $316 = ($hold$12$lcssa|0)==($315|0);
     if ($316) {
      $bits$13 = 0;$have$17 = $have$16$lcssa;$hold$13 = 0;$next$17 = $next$16$lcssa;
     } else {
      HEAP32[$27>>2] = 612504;
      HEAP32[$2>>2] = 29;
      $bits$0$be = $bits$12$lcssa;$have$0$be = $have$16$lcssa;$hold$0$be = $hold$12$lcssa;$left$0$be = $left$0;$next$0$be = $next$16$lcssa;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$0;
      break;
     }
    }
    $317 = HEAP32[$26>>2]|0;
    $318 = ($317|0)==(0|0);
    if (!($318)) {
     $319 = $301 >>> 9;
     $320 = $319 & 1;
     $321 = ((($317)) + 44|0);
     HEAP32[$321>>2] = $320;
     $322 = ((($317)) + 48|0);
     HEAP32[$322>>2] = 1;
    }
    $323 = (_crc32(0,0,0)|0);
    HEAP32[$23>>2] = $323;
    HEAP32[$30>>2] = $323;
    HEAP32[$2>>2] = 11;
    $bits$0$be = $bits$13;$have$0$be = $have$17;$hold$0$be = $hold$13;$left$0$be = $left$0;$next$0$be = $next$17;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$0;
   }
   else if ((label|0) == 244) {
    label = 0;
    $757 = ($left$0|0)==(0);
    if ($757) {
     $$lcssa2799 = $380;$$ph = 0;$bits$53$ph = $bits$46;$have$57$ph = $have$50;$hold$53$ph = $hold$46;$next$57$ph = $next$50;$out$0$lcssa2944 = $out$0;$ret$8$ph = $ret$7;
     label = 282;
     break L17;
    }
    $758 = (($out$0) - ($left$0))|0;
    $759 = HEAP32[$43>>2]|0;
    $760 = ($759>>>0)>($758>>>0);
    if ($760) {
     $761 = (($759) - ($758))|0;
     $762 = HEAP32[$44>>2]|0;
     $763 = ($761>>>0)>($762>>>0);
     if ($763) {
      $764 = HEAP32[$45>>2]|0;
      $765 = ($764|0)==(0);
      if (!($765)) {
       HEAP32[$27>>2] = 613699;
       HEAP32[$2>>2] = 29;
       $bits$0$be = $bits$46;$have$0$be = $have$50;$hold$0$be = $hold$46;$left$0$be = $left$0;$next$0$be = $next$50;$out$0$be = $out$0;$put$0$be = $380;$ret$0$be = $ret$7;
       break;
      }
     }
     $766 = HEAP32[$46>>2]|0;
     $767 = ($761>>>0)>($766>>>0);
     if ($767) {
      $768 = (($761) - ($766))|0;
      $769 = HEAP32[$47>>2]|0;
      $770 = HEAP32[$48>>2]|0;
      $771 = (($770) - ($768))|0;
      $772 = (($769) + ($771)|0);
      $copy$7 = $768;$from$0 = $772;
     } else {
      $773 = HEAP32[$47>>2]|0;
      $774 = (($766) - ($761))|0;
      $775 = (($773) + ($774)|0);
      $copy$7 = $761;$from$0 = $775;
     }
     $776 = HEAP32[$31>>2]|0;
     $777 = ($copy$7>>>0)>($776>>>0);
     $$copy$7 = $777 ? $776 : $copy$7;
     $783 = $776;$copy$8 = $$copy$7;$from$1 = $from$0;
    } else {
     $778 = (0 - ($759))|0;
     $779 = (($380) + ($778)|0);
     $780 = HEAP32[$31>>2]|0;
     $783 = $780;$copy$8 = $780;$from$1 = $779;
    }
    $781 = ($copy$8>>>0)>($left$0>>>0);
    $left$0$copy$8 = $781 ? $left$0 : $copy$8;
    $782 = (($783) - ($left$0$copy$8))|0;
    HEAP32[$31>>2] = $782;
    $copy$10 = $left$0$copy$8;$from$2 = $from$1;$put$1 = $380;
    while(1) {
     $784 = ((($from$2)) + 1|0);
     $785 = HEAP8[$from$2>>0]|0;
     $786 = ((($put$1)) + 1|0);
     HEAP8[$put$1>>0] = $785;
     $787 = (($copy$10) + -1)|0;
     $788 = ($787|0)==(0);
     if ($788) {
      $$lcssa2712 = $786;
      break;
     } else {
      $copy$10 = $787;$from$2 = $784;$put$1 = $786;
     }
    }
    $789 = (($left$0) - ($left$0$copy$8))|0;
    $790 = HEAP32[$31>>2]|0;
    $791 = ($790|0)==(0);
    if ($791) {
     HEAP32[$2>>2] = 20;
     $bits$0$be = $bits$46;$have$0$be = $have$50;$hold$0$be = $hold$46;$left$0$be = $789;$next$0$be = $next$50;$out$0$be = $out$0;$put$0$be = $$lcssa2712;$ret$0$be = $ret$7;
    } else {
     $bits$0$be = $bits$46;$have$0$be = $have$50;$hold$0$be = $hold$46;$left$0$be = $789;$next$0$be = $next$50;$out$0$be = $out$0;$put$0$be = $$lcssa2712;$ret$0$be = $ret$7;
    }
   }
  } while(0);
  $$pre = HEAP32[$2>>2]|0;
  $380 = $put$0$be;$62 = $$pre;$bits$0 = $bits$0$be;$have$0 = $have$0$be;$hold$0 = $hold$0$be;$left$0 = $left$0$be;$next$0 = $next$0$be;$out$0 = $out$0$be;$ret$0 = $ret$0$be;
 }
 if ((label|0) == 122) {
  HEAP32[$4>>2] = $$lcssa2819;
  HEAP32[$14>>2] = $left$0$lcssa2877;
  HEAP32[$strm>>2] = $next$19$lcssa;
  HEAP32[$16>>2] = $have$19$lcssa;
  HEAP32[$18>>2] = $hold$15$lcssa;
  HEAP32[$20>>2] = $bits$15$lcssa;
  $$0 = 2;
  STACKTOP = sp;return ($$0|0);
 }
 else if ((label|0) == 133) {
  $354 = $hold$18$lcssa$lcssa3009 >>> 3;
  $355 = (($bits$18$lcssa$lcssa3008) + -3)|0;
  $844 = $$lcssa2821;$845 = $left$0$lcssa2879;$bits$53 = $355;$have$57 = $have$22$lcssa$lcssa3010;$hold$53 = $354;$next$57 = $next$22$lcssa$lcssa3011;$out$4 = $out$0$lcssa2966;$ret$8 = $ret$0$lcssa2995;
 }
 else if ((label|0) == 136) {
  // unreachable;
 }
 else if ((label|0) == 281) {
  HEAP32[$2>>2] = 28;
  $844 = $$lcssa2803;$845 = $left$0$lcssa2861;$bits$53 = $bits$51;$have$57 = $have$55;$hold$53 = $hold$51;$next$57 = $next$55;$out$4 = $out$2$lcssa;$ret$8 = 1;
 }
 else if ((label|0) == 282) {
  $844 = $$lcssa2799;$845 = $$ph;$bits$53 = $bits$53$ph;$have$57 = $have$57$ph;$hold$53 = $hold$53$ph;$next$57 = $next$57$ph;$out$4 = $out$0$lcssa2944;$ret$8 = $ret$8$ph;
 }
 else if ((label|0) == 296) {
  $$0 = -4;
  STACKTOP = sp;return ($$0|0);
 }
 else if ((label|0) == 297) {
  STACKTOP = sp;return ($$0|0);
 }
 HEAP32[$4>>2] = $844;
 HEAP32[$14>>2] = $845;
 HEAP32[$strm>>2] = $next$57;
 HEAP32[$16>>2] = $have$57;
 HEAP32[$18>>2] = $hold$53;
 HEAP32[$20>>2] = $bits$53;
 $846 = HEAP32[$48>>2]|0;
 $847 = ($846|0)==(0);
 if ($847) {
  $848 = ($out$4|0)==($845|0);
  if ($848) {
   $857 = $have$57;$859 = $845;
  } else {
   $849 = HEAP32[$2>>2]|0;
   $850 = ($849>>>0)<(29);
   if ($850) {
    $851 = ($849>>>0)<(26);
    $852 = ($flush|0)!=(4);
    $or$cond8 = $852 | $851;
    if ($or$cond8) {
     label = 287;
    } else {
     $857 = $have$57;$859 = $845;
    }
   } else {
    $857 = $have$57;$859 = $845;
   }
  }
 } else {
  label = 287;
 }
 do {
  if ((label|0) == 287) {
   $853 = (($out$4) - ($845))|0;
   $854 = (_updatewindow($strm,$844,$853)|0);
   $855 = ($854|0)==(0);
   if ($855) {
    $$pre1278 = HEAP32[$16>>2]|0;
    $$pre1279 = HEAP32[$14>>2]|0;
    $857 = $$pre1278;$859 = $$pre1279;
    break;
   }
   HEAP32[$2>>2] = 30;
   $$0 = -4;
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 $856 = (($17) - ($857))|0;
 $858 = (($out$4) - ($859))|0;
 $860 = ((($strm)) + 8|0);
 $861 = HEAP32[$860>>2]|0;
 $862 = (($856) + ($861))|0;
 HEAP32[$860>>2] = $862;
 $863 = HEAP32[$49>>2]|0;
 $864 = (($863) + ($858))|0;
 HEAP32[$49>>2] = $864;
 $865 = HEAP32[$50>>2]|0;
 $866 = (($865) + ($858))|0;
 HEAP32[$50>>2] = $866;
 $867 = HEAP32[$22>>2]|0;
 $868 = ($867|0)!=(0);
 $869 = ($out$4|0)!=($859|0);
 $or$cond10 = $869 & $868;
 if ($or$cond10) {
  $870 = HEAP32[$25>>2]|0;
  $871 = ($870|0)==(0);
  $872 = HEAP32[$23>>2]|0;
  $873 = HEAP32[$4>>2]|0;
  $874 = (0 - ($858))|0;
  $875 = (($873) + ($874)|0);
  if ($871) {
   $877 = (_adler32($872,$875,$858)|0);
   $878 = $877;
  } else {
   $876 = (_crc32($872,$875,$858)|0);
   $878 = $876;
  }
  HEAP32[$23>>2] = $878;
  HEAP32[$30>>2] = $878;
 }
 $879 = HEAP32[$20>>2]|0;
 $880 = HEAP32[$34>>2]|0;
 $881 = ($880|0)!=(0);
 $882 = $881 ? 64 : 0;
 $883 = (($882) + ($879))|0;
 $884 = HEAP32[$2>>2]|0;
 $885 = ($884|0)==(11);
 $886 = $885 ? 128 : 0;
 $887 = (($883) + ($886))|0;
 $888 = ($884|0)==(19);
 $889 = ($884|0)==(14);
 $890 = $888 | $889;
 $891 = $890 ? 256 : 0;
 $892 = (($887) + ($891))|0;
 $893 = ((($strm)) + 44|0);
 HEAP32[$893>>2] = $892;
 $894 = ($17|0)==($857|0);
 $895 = ($out$4|0)==($859|0);
 $or$cond12 = $894 & $895;
 $896 = ($flush|0)==(4);
 $or$cond14 = $896 | $or$cond12;
 $897 = ($ret$8|0)==(0);
 $or$cond16 = $897 & $or$cond14;
 $$ret$8 = $or$cond16 ? -5 : $ret$8;
 $$0 = $$ret$8;
 STACKTOP = sp;return ($$0|0);
}
function _updatewindow($strm,$end,$copy) {
 $strm = $strm|0;
 $end = $end|0;
 $copy = $copy|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $copy$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($strm)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 52|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if ($4) {
  $5 = ((($strm)) + 32|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ((($strm)) + 40|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ((($1)) + 36|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = 1 << $10;
  $12 = (FUNCTION_TABLE_iiii[$6 & 8191]($8,$11,1)|0);
  HEAP32[$2>>2] = $12;
  $13 = ($12|0)==(0|0);
  if ($13) {
   $$0 = 1;
   return ($$0|0);
  } else {
   $26 = $12;
  }
 } else {
  $26 = $3;
 }
 $14 = ((($1)) + 40|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = ($15|0)==(0);
 if ($16) {
  $17 = ((($1)) + 36|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = 1 << $18;
  HEAP32[$14>>2] = $19;
  $20 = ((($1)) + 48|0);
  HEAP32[$20>>2] = 0;
  $21 = ((($1)) + 44|0);
  HEAP32[$21>>2] = 0;
  $23 = $19;
 } else {
  $23 = $15;
 }
 $22 = ($23>>>0)>($copy>>>0);
 if (!($22)) {
  $24 = (0 - ($23))|0;
  $25 = (($end) + ($24)|0);
  _memcpy(($26|0),($25|0),($23|0))|0;
  $27 = ((($1)) + 48|0);
  HEAP32[$27>>2] = 0;
  $28 = HEAP32[$14>>2]|0;
  $29 = ((($1)) + 44|0);
  HEAP32[$29>>2] = $28;
  $$0 = 0;
  return ($$0|0);
 }
 $30 = ((($1)) + 48|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = (($23) - ($31))|0;
 $33 = ($32>>>0)>($copy>>>0);
 $copy$ = $33 ? $copy : $32;
 $34 = (($26) + ($31)|0);
 $35 = (0 - ($copy))|0;
 $36 = (($end) + ($35)|0);
 _memcpy(($34|0),($36|0),($copy$|0))|0;
 $37 = (($copy) - ($copy$))|0;
 $38 = ($copy$|0)==($copy|0);
 if (!($38)) {
  $39 = HEAP32[$2>>2]|0;
  $40 = (0 - ($37))|0;
  $41 = (($end) + ($40)|0);
  _memcpy(($39|0),($41|0),($37|0))|0;
  HEAP32[$30>>2] = $37;
  $42 = HEAP32[$14>>2]|0;
  $43 = ((($1)) + 44|0);
  HEAP32[$43>>2] = $42;
  $$0 = 0;
  return ($$0|0);
 }
 $44 = HEAP32[$30>>2]|0;
 $45 = (($44) + ($copy))|0;
 $46 = HEAP32[$14>>2]|0;
 $47 = ($45|0)==($46|0);
 $$ = $47 ? 0 : $45;
 HEAP32[$30>>2] = $$;
 $48 = ((($1)) + 44|0);
 $49 = HEAP32[$48>>2]|0;
 $50 = ($49>>>0)<($46>>>0);
 if (!($50)) {
  $$0 = 0;
  return ($$0|0);
 }
 $51 = (($49) + ($copy))|0;
 HEAP32[$48>>2] = $51;
 $$0 = 0;
 return ($$0|0);
}
function _inflateEnd($strm) {
 $strm = $strm|0;
 var $$0 = 0, $$pre = 0, $$pre1 = 0, $$pre2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($strm|0)==(0|0);
 if ($0) {
  $$0 = -2;
  return ($$0|0);
 }
 $1 = ((($strm)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $$0 = -2;
  return ($$0|0);
 }
 $4 = ((($strm)) + 36|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if ($6) {
  $$0 = -2;
  return ($$0|0);
 }
 $7 = ((($2)) + 52|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 $$pre2 = ((($strm)) + 40|0);
 if ($9) {
  $12 = $5;$13 = $2;
 } else {
  $10 = HEAP32[$$pre2>>2]|0;
  FUNCTION_TABLE_vii[$5 & 8191]($10,$8);
  $$pre = HEAP32[$4>>2]|0;
  $$pre1 = HEAP32[$1>>2]|0;
  $12 = $$pre;$13 = $$pre1;
 }
 $11 = HEAP32[$$pre2>>2]|0;
 FUNCTION_TABLE_vii[$12 & 8191]($11,$13);
 HEAP32[$1>>2] = 0;
 $$0 = 0;
 return ($$0|0);
}
function _inflate_table($type,$lens,$codes,$table,$bits,$work) {
 $type = $type|0;
 $lens = $lens|0;
 $codes = $codes|0;
 $table = $table|0;
 $bits = $bits|0;
 $work = $work|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa121 = 0, $$lcssa122 = 0, $$ph = 0, $$ph21 = 0, $$ph21$in = 0, $$ph22 = 0, $$ph79 = 0, $$ph80 = 0, $$phi$trans$insert = 0, $$phi$trans$insert76 = 0, $$phi$trans$insert76$1 = 0, $$phi$trans$insert76$10 = 0, $$phi$trans$insert76$11 = 0, $$phi$trans$insert76$12 = 0, $$phi$trans$insert76$2 = 0, $$phi$trans$insert76$3 = 0, $$phi$trans$insert76$4 = 0, $$phi$trans$insert76$5 = 0;
 var $$phi$trans$insert76$6 = 0, $$phi$trans$insert76$7 = 0, $$phi$trans$insert76$8 = 0, $$phi$trans$insert76$9 = 0, $$pre = 0, $$pre77 = 0, $$pre77$1 = 0, $$pre77$10 = 0, $$pre77$11 = 0, $$pre77$12 = 0, $$pre77$2 = 0, $$pre77$3 = 0, $$pre77$4 = 0, $$pre77$5 = 0, $$pre77$6 = 0, $$pre77$7 = 0, $$pre77$8 = 0, $$pre77$9 = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0;
 var $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0;
 var $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0;
 var $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0;
 var $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0;
 var $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0;
 var $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $base$01320$ph = 0, $count = 0, $curr$0$ph = 0, $curr$1$lcssa = 0, $curr$139 = 0, $drop$0$ph = 0, $end$01518$ph = 0, $exitcond = 0, $exitcond73 = 0, $extra$01419$ph = 0, $fill$0 = 0;
 var $fold = 0, $fold108 = 0, $fold109 = 0, $fold110 = 0, $fold111 = 0, $fold112 = 0, $fold113 = 0, $fold114 = 0, $fold115 = 0, $fold116 = 0, $fold117 = 0, $here$sroa$0$0 = 0, $here$sroa$14$0 = 0, $huff$0 = 0, $huff$0$ph = 0, $huff$1 = 0, $huff$1$lcssa = 0, $huff$1$lcssa119 = 0, $incr$0 = 0, $incr$0$lcssa = 0;
 var $left$140 = 0, $len$3 = 0, $len$3$ph = 0, $len$4 = 0, $len$4$lcssa = 0, $low$0$ph = 0, $max$0$ = 0, $max$0$81 = 0, $max$0$84 = 0, $max$056$lcssa$ph = 0, $max$056$lcssa82 = 0, $min$0$lcssa = 0, $min$0$max$0$ = 0, $min$0$max$0$$drop$0 = 0, $min$049 = 0, $next$0$ph = 0, $next$0$ph24 = 0, $next$0$ph24$lcssa = 0, $offs = 0, $or$cond = 0;
 var $or$cond10 = 0, $or$cond5 = 0, $or$cond7 = 0, $or$cond9 = 0, $sym$057 = 0, $sym$143 = 0, $sym$2 = 0, $sym$2$ph = 0, $used$0$ph23 = 0, $used$0$ph23$lcssa = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $count = sp + 32|0;
 $offs = sp;
 dest=$count; stop=dest+32|0; do { HEAP16[dest>>1]=0|0; dest=dest+2|0; } while ((dest|0) < (stop|0));
 $0 = ($codes|0)==(0);
 if ($0) {
  $1 = HEAP32[$bits>>2]|0;
  $15 = $1;
  label = 5;
 } else {
  $sym$057 = 0;
  while(1) {
   $2 = (($lens) + ($sym$057<<1)|0);
   $3 = HEAP16[$2>>1]|0;
   $4 = $3&65535;
   $5 = (($count) + ($4<<1)|0);
   $6 = HEAP16[$5>>1]|0;
   $7 = (($6) + 1)<<16>>16;
   HEAP16[$5>>1] = $7;
   $8 = (($sym$057) + 1)|0;
   $exitcond73 = ($8|0)==($codes|0);
   if ($exitcond73) {
    break;
   } else {
    $sym$057 = $8;
   }
  }
  $$phi$trans$insert = ((($count)) + 30|0);
  $$pre = HEAP16[$$phi$trans$insert>>1]|0;
  $9 = HEAP32[$bits>>2]|0;
  $10 = ($$pre<<16>>16)==(0);
  if ($10) {
   $15 = $9;
   label = 5;
  } else {
   $$ph79 = $9;$$ph80 = $$pre;$max$056$lcssa$ph = 15;
   label = 7;
  }
 }
 do {
  if ((label|0) == 5) {
   $11 = ((($count)) + 28|0);
   $12 = HEAP16[$11>>1]|0;
   $13 = ($12<<16>>16)==(0);
   if ($13) {
    $264 = ((($count)) + 26|0);
    $265 = HEAP16[$264>>1]|0;
    $266 = ($265<<16>>16)==(0);
    if ($266) {
     $267 = ((($count)) + 24|0);
     $268 = HEAP16[$267>>1]|0;
     $269 = ($268<<16>>16)==(0);
     if ($269) {
      $270 = ((($count)) + 22|0);
      $271 = HEAP16[$270>>1]|0;
      $272 = ($271<<16>>16)==(0);
      if ($272) {
       $273 = ((($count)) + 20|0);
       $274 = HEAP16[$273>>1]|0;
       $275 = ($274<<16>>16)==(0);
       if ($275) {
        $276 = ((($count)) + 18|0);
        $277 = HEAP16[$276>>1]|0;
        $278 = ($277<<16>>16)==(0);
        if ($278) {
         $279 = ((($count)) + 16|0);
         $280 = HEAP16[$279>>1]|0;
         $281 = ($280<<16>>16)==(0);
         if ($281) {
          $282 = ((($count)) + 14|0);
          $283 = HEAP16[$282>>1]|0;
          $284 = ($283<<16>>16)==(0);
          if ($284) {
           $285 = ((($count)) + 12|0);
           $286 = HEAP16[$285>>1]|0;
           $287 = ($286<<16>>16)==(0);
           if ($287) {
            $288 = ((($count)) + 10|0);
            $289 = HEAP16[$288>>1]|0;
            $290 = ($289<<16>>16)==(0);
            if ($290) {
             $291 = ((($count)) + 8|0);
             $292 = HEAP16[$291>>1]|0;
             $293 = ($292<<16>>16)==(0);
             if ($293) {
              $294 = ((($count)) + 6|0);
              $295 = HEAP16[$294>>1]|0;
              $296 = ($295<<16>>16)==(0);
              if ($296) {
               $297 = ((($count)) + 4|0);
               $298 = HEAP16[$297>>1]|0;
               $299 = ($298<<16>>16)==(0);
               if ($299) {
                $300 = ((($count)) + 2|0);
                $301 = HEAP16[$300>>1]|0;
                $302 = ($301<<16>>16)==(0);
                if (!($302)) {
                 $14 = ($15>>>0)>(1);
                 $max$0$ = $14 ? 1 : $15;
                 $260 = 0;$max$0$84 = $max$0$;$max$056$lcssa82 = 1;$min$0$lcssa = 1;
                 break;
                }
                $303 = HEAP32[$table>>2]|0;
                $304 = ((($303)) + 4|0);
                HEAP32[$table>>2] = $304;
                HEAP8[$303>>0] = 64;
                $305 = ((($303)) + 1|0);
                HEAP8[$305>>0] = 1;
                $306 = ((($303)) + 2|0);
                HEAP16[$306>>1] = 0;
                $307 = HEAP32[$table>>2]|0;
                $308 = ((($307)) + 4|0);
                HEAP32[$table>>2] = $308;
                HEAP8[$307>>0] = 64;
                $309 = ((($307)) + 1|0);
                HEAP8[$309>>0] = 1;
                $310 = ((($307)) + 2|0);
                HEAP16[$310>>1] = 0;
                HEAP32[$bits>>2] = 1;
                $$0 = 0;
                STACKTOP = sp;return ($$0|0);
               } else {
                $$ph79 = $15;$$ph80 = 0;$max$056$lcssa$ph = 2;
                label = 7;
               }
              } else {
               $$ph79 = $15;$$ph80 = 0;$max$056$lcssa$ph = 3;
               label = 7;
              }
             } else {
              $$ph79 = $15;$$ph80 = 0;$max$056$lcssa$ph = 4;
              label = 7;
             }
            } else {
             $$ph79 = $15;$$ph80 = 0;$max$056$lcssa$ph = 5;
             label = 7;
            }
           } else {
            $$ph79 = $15;$$ph80 = 0;$max$056$lcssa$ph = 6;
            label = 7;
           }
          } else {
           $$ph79 = $15;$$ph80 = 0;$max$056$lcssa$ph = 7;
           label = 7;
          }
         } else {
          $$ph79 = $15;$$ph80 = 0;$max$056$lcssa$ph = 8;
          label = 7;
         }
        } else {
         $$ph79 = $15;$$ph80 = 0;$max$056$lcssa$ph = 9;
         label = 7;
        }
       } else {
        $$ph79 = $15;$$ph80 = 0;$max$056$lcssa$ph = 10;
        label = 7;
       }
      } else {
       $$ph79 = $15;$$ph80 = 0;$max$056$lcssa$ph = 11;
       label = 7;
      }
     } else {
      $$ph79 = $15;$$ph80 = 0;$max$056$lcssa$ph = 12;
      label = 7;
     }
    } else {
     $$ph79 = $15;$$ph80 = 0;$max$056$lcssa$ph = 13;
     label = 7;
    }
   } else {
    $$ph79 = $15;$$ph80 = 0;$max$056$lcssa$ph = 14;
    label = 7;
   }
  }
 } while(0);
 L25: do {
  if ((label|0) == 7) {
   $16 = ($$ph79>>>0)>($max$056$lcssa$ph>>>0);
   $max$0$81 = $16 ? $max$056$lcssa$ph : $$ph79;
   $min$049 = 1;
   while(1) {
    $17 = (($count) + ($min$049<<1)|0);
    $18 = HEAP16[$17>>1]|0;
    $19 = ($18<<16>>16)==(0);
    if (!($19)) {
     $260 = $$ph80;$max$0$84 = $max$0$81;$max$056$lcssa82 = $max$056$lcssa$ph;$min$0$lcssa = $min$049;
     break L25;
    }
    $20 = (($min$049) + 1)|0;
    $21 = ($20>>>0)<($max$056$lcssa$ph>>>0);
    if ($21) {
     $min$049 = $20;
    } else {
     $260 = $$ph80;$max$0$84 = $max$0$81;$max$056$lcssa82 = $max$056$lcssa$ph;$min$0$lcssa = $20;
     break;
    }
   }
  }
 } while(0);
 $22 = ($max$0$84>>>0)<($min$0$lcssa>>>0);
 $min$0$max$0$ = $22 ? $min$0$lcssa : $max$0$84;
 $23 = ((($count)) + 2|0);
 $24 = HEAP16[$23>>1]|0;
 $25 = $24&65535;
 $26 = (2 - ($25))|0;
 $27 = ($26|0)<(0);
 if ($27) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $28 = $26 << 1;
 $29 = ((($count)) + 4|0);
 $30 = HEAP16[$29>>1]|0;
 $31 = $30&65535;
 $32 = (($28) - ($31))|0;
 $33 = ($32|0)<(0);
 if ($33) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $186 = $32 << 1;
 $187 = ((($count)) + 6|0);
 $188 = HEAP16[$187>>1]|0;
 $189 = $188&65535;
 $190 = (($186) - ($189))|0;
 $191 = ($190|0)<(0);
 if ($191) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $192 = $190 << 1;
 $193 = ((($count)) + 8|0);
 $194 = HEAP16[$193>>1]|0;
 $195 = $194&65535;
 $196 = (($192) - ($195))|0;
 $197 = ($196|0)<(0);
 if ($197) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $198 = $196 << 1;
 $199 = ((($count)) + 10|0);
 $200 = HEAP16[$199>>1]|0;
 $201 = $200&65535;
 $202 = (($198) - ($201))|0;
 $203 = ($202|0)<(0);
 if ($203) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $204 = $202 << 1;
 $205 = ((($count)) + 12|0);
 $206 = HEAP16[$205>>1]|0;
 $207 = $206&65535;
 $208 = (($204) - ($207))|0;
 $209 = ($208|0)<(0);
 if ($209) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $210 = $208 << 1;
 $211 = ((($count)) + 14|0);
 $212 = HEAP16[$211>>1]|0;
 $213 = $212&65535;
 $214 = (($210) - ($213))|0;
 $215 = ($214|0)<(0);
 if ($215) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $216 = $214 << 1;
 $217 = ((($count)) + 16|0);
 $218 = HEAP16[$217>>1]|0;
 $219 = $218&65535;
 $220 = (($216) - ($219))|0;
 $221 = ($220|0)<(0);
 if ($221) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $222 = $220 << 1;
 $223 = ((($count)) + 18|0);
 $224 = HEAP16[$223>>1]|0;
 $225 = $224&65535;
 $226 = (($222) - ($225))|0;
 $227 = ($226|0)<(0);
 if ($227) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $228 = $226 << 1;
 $229 = ((($count)) + 20|0);
 $230 = HEAP16[$229>>1]|0;
 $231 = $230&65535;
 $232 = (($228) - ($231))|0;
 $233 = ($232|0)<(0);
 if ($233) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $234 = $232 << 1;
 $235 = ((($count)) + 22|0);
 $236 = HEAP16[$235>>1]|0;
 $237 = $236&65535;
 $238 = (($234) - ($237))|0;
 $239 = ($238|0)<(0);
 if ($239) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $240 = $238 << 1;
 $241 = ((($count)) + 24|0);
 $242 = HEAP16[$241>>1]|0;
 $243 = $242&65535;
 $244 = (($240) - ($243))|0;
 $245 = ($244|0)<(0);
 if ($245) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $246 = $244 << 1;
 $247 = ((($count)) + 26|0);
 $248 = HEAP16[$247>>1]|0;
 $249 = $248&65535;
 $250 = (($246) - ($249))|0;
 $251 = ($250|0)<(0);
 if ($251) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $252 = $250 << 1;
 $253 = ((($count)) + 28|0);
 $254 = HEAP16[$253>>1]|0;
 $255 = $254&65535;
 $256 = (($252) - ($255))|0;
 $257 = ($256|0)<(0);
 if ($257) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $258 = $256 << 1;
 $259 = $260&65535;
 $261 = (($258) - ($259))|0;
 $262 = ($261|0)<(0);
 if ($262) {
  $$0 = -1;
  STACKTOP = sp;return ($$0|0);
 }
 $263 = ($261|0)>(0);
 if ($263) {
  $34 = ($type|0)==(0);
  $35 = ($max$056$lcssa82|0)!=(1);
  $or$cond = $34 | $35;
  if ($or$cond) {
   $$0 = -1;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $36 = ((($offs)) + 2|0);
 HEAP16[$36>>1] = 0;
 $37 = ((($offs)) + 4|0);
 HEAP16[$37>>1] = $24;
 $38 = $24&65535;
 $$phi$trans$insert76 = ((($count)) + 4|0);
 $$pre77 = HEAP16[$$phi$trans$insert76>>1]|0;
 $39 = $$pre77&65535;
 $40 = (($39) + ($38))|0;
 $41 = $40&65535;
 $42 = ((($offs)) + 6|0);
 HEAP16[$42>>1] = $41;
 $$phi$trans$insert76$1 = ((($count)) + 6|0);
 $$pre77$1 = HEAP16[$$phi$trans$insert76$1>>1]|0;
 $43 = $$pre77$1&65535;
 $44 = (($43) + ($40))|0;
 $45 = $44&65535;
 $46 = ((($offs)) + 8|0);
 HEAP16[$46>>1] = $45;
 $$phi$trans$insert76$2 = ((($count)) + 8|0);
 $$pre77$2 = HEAP16[$$phi$trans$insert76$2>>1]|0;
 $fold = (($40) + ($43))|0;
 $47 = $$pre77$2&65535;
 $48 = (($47) + ($fold))|0;
 $49 = $48&65535;
 $50 = ((($offs)) + 10|0);
 HEAP16[$50>>1] = $49;
 $$phi$trans$insert76$3 = ((($count)) + 10|0);
 $$pre77$3 = HEAP16[$$phi$trans$insert76$3>>1]|0;
 $fold108 = (($fold) + ($47))|0;
 $51 = $$pre77$3&65535;
 $52 = (($51) + ($fold108))|0;
 $53 = $52&65535;
 $54 = ((($offs)) + 12|0);
 HEAP16[$54>>1] = $53;
 $$phi$trans$insert76$4 = ((($count)) + 12|0);
 $$pre77$4 = HEAP16[$$phi$trans$insert76$4>>1]|0;
 $fold109 = (($fold108) + ($51))|0;
 $55 = $$pre77$4&65535;
 $56 = (($55) + ($fold109))|0;
 $57 = $56&65535;
 $58 = ((($offs)) + 14|0);
 HEAP16[$58>>1] = $57;
 $$phi$trans$insert76$5 = ((($count)) + 14|0);
 $$pre77$5 = HEAP16[$$phi$trans$insert76$5>>1]|0;
 $fold110 = (($fold109) + ($55))|0;
 $59 = $$pre77$5&65535;
 $60 = (($59) + ($fold110))|0;
 $61 = $60&65535;
 $62 = ((($offs)) + 16|0);
 HEAP16[$62>>1] = $61;
 $$phi$trans$insert76$6 = ((($count)) + 16|0);
 $$pre77$6 = HEAP16[$$phi$trans$insert76$6>>1]|0;
 $fold111 = (($fold110) + ($59))|0;
 $63 = $$pre77$6&65535;
 $64 = (($63) + ($fold111))|0;
 $65 = $64&65535;
 $66 = ((($offs)) + 18|0);
 HEAP16[$66>>1] = $65;
 $$phi$trans$insert76$7 = ((($count)) + 18|0);
 $$pre77$7 = HEAP16[$$phi$trans$insert76$7>>1]|0;
 $fold112 = (($fold111) + ($63))|0;
 $67 = $$pre77$7&65535;
 $68 = (($67) + ($fold112))|0;
 $69 = $68&65535;
 $70 = ((($offs)) + 20|0);
 HEAP16[$70>>1] = $69;
 $$phi$trans$insert76$8 = ((($count)) + 20|0);
 $$pre77$8 = HEAP16[$$phi$trans$insert76$8>>1]|0;
 $fold113 = (($fold112) + ($67))|0;
 $71 = $$pre77$8&65535;
 $72 = (($71) + ($fold113))|0;
 $73 = $72&65535;
 $74 = ((($offs)) + 22|0);
 HEAP16[$74>>1] = $73;
 $$phi$trans$insert76$9 = ((($count)) + 22|0);
 $$pre77$9 = HEAP16[$$phi$trans$insert76$9>>1]|0;
 $fold114 = (($fold113) + ($71))|0;
 $75 = $$pre77$9&65535;
 $76 = (($75) + ($fold114))|0;
 $77 = $76&65535;
 $78 = ((($offs)) + 24|0);
 HEAP16[$78>>1] = $77;
 $$phi$trans$insert76$10 = ((($count)) + 24|0);
 $$pre77$10 = HEAP16[$$phi$trans$insert76$10>>1]|0;
 $fold115 = (($fold114) + ($75))|0;
 $79 = $$pre77$10&65535;
 $80 = (($79) + ($fold115))|0;
 $81 = $80&65535;
 $82 = ((($offs)) + 26|0);
 HEAP16[$82>>1] = $81;
 $$phi$trans$insert76$11 = ((($count)) + 26|0);
 $$pre77$11 = HEAP16[$$phi$trans$insert76$11>>1]|0;
 $fold116 = (($fold115) + ($79))|0;
 $83 = $$pre77$11&65535;
 $84 = (($83) + ($fold116))|0;
 $85 = $84&65535;
 $86 = ((($offs)) + 28|0);
 HEAP16[$86>>1] = $85;
 $$phi$trans$insert76$12 = ((($count)) + 28|0);
 $$pre77$12 = HEAP16[$$phi$trans$insert76$12>>1]|0;
 $fold117 = (($fold116) + ($83))|0;
 $87 = $$pre77$12&65535;
 $88 = (($87) + ($fold117))|0;
 $89 = $88&65535;
 $90 = ((($offs)) + 30|0);
 HEAP16[$90>>1] = $89;
 if (!($0)) {
  $sym$143 = 0;
  while(1) {
   $91 = (($lens) + ($sym$143<<1)|0);
   $92 = HEAP16[$91>>1]|0;
   $93 = ($92<<16>>16)==(0);
   if (!($93)) {
    $94 = $92&65535;
    $95 = $sym$143&65535;
    $96 = (($offs) + ($94<<1)|0);
    $97 = HEAP16[$96>>1]|0;
    $98 = (($97) + 1)<<16>>16;
    HEAP16[$96>>1] = $98;
    $99 = $97&65535;
    $100 = (($work) + ($99<<1)|0);
    HEAP16[$100>>1] = $95;
   }
   $101 = (($sym$143) + 1)|0;
   $exitcond = ($101|0)==($codes|0);
   if ($exitcond) {
    break;
   } else {
    $sym$143 = $101;
   }
  }
 }
 switch ($type|0) {
 case 0:  {
  $$ph = 0;$$ph22 = 0;$base$01320$ph = $work;$end$01518$ph = 19;$extra$01419$ph = $work;
  break;
 }
 case 1:  {
  $102 = ($min$0$max$0$>>>0)>(9);
  if ($102) {
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
  } else {
   $$ph = 0;$$ph22 = 1;$base$01320$ph = (177226);$end$01518$ph = 256;$extra$01419$ph = (177352);
  }
  break;
 }
 default: {
  $103 = ($min$0$max$0$>>>0)>(9);
  $104 = ($type|0)==(2);
  $or$cond5 = $104 & $103;
  if ($or$cond5) {
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
  } else {
   $$ph = $104;$$ph22 = 0;$base$01320$ph = 177676;$end$01518$ph = -1;$extra$01419$ph = 177802;
  }
 }
 }
 $$ph21$in = 1 << $min$0$max$0$;
 $next$0$ph = HEAP32[$table>>2]|0;
 $$ph21 = (($$ph21$in) + -1)|0;
 $105 = $min$0$max$0$&255;
 $curr$0$ph = $min$0$max$0$;$drop$0$ph = 0;$huff$0$ph = 0;$len$3$ph = $min$0$lcssa;$low$0$ph = -1;$next$0$ph24 = $next$0$ph;$sym$2$ph = 0;$used$0$ph23 = $$ph21$in;
 L93: while(1) {
  $106 = 1 << $curr$0$ph;
  $huff$0 = $huff$0$ph;$len$3 = $len$3$ph;$sym$2 = $sym$2$ph;
  while(1) {
   $107 = (($len$3) - ($drop$0$ph))|0;
   $108 = $107&255;
   $109 = (($work) + ($sym$2<<1)|0);
   $110 = HEAP16[$109>>1]|0;
   $111 = $110&65535;
   $112 = ($111|0)<($end$01518$ph|0);
   do {
    if ($112) {
     $here$sroa$0$0 = 0;$here$sroa$14$0 = $110;
    } else {
     $113 = ($111|0)>($end$01518$ph|0);
     if (!($113)) {
      $here$sroa$0$0 = 96;$here$sroa$14$0 = 0;
      break;
     }
     $114 = (($extra$01419$ph) + ($111<<1)|0);
     $115 = HEAP16[$114>>1]|0;
     $116 = $115&255;
     $117 = (($base$01320$ph) + ($111<<1)|0);
     $118 = HEAP16[$117>>1]|0;
     $here$sroa$0$0 = $116;$here$sroa$14$0 = $118;
    }
   } while(0);
   $119 = 1 << $107;
   $120 = $huff$0 >>> $drop$0$ph;
   $fill$0 = $106;
   while(1) {
    $121 = (($fill$0) - ($119))|0;
    $122 = (($121) + ($120))|0;
    $123 = (($next$0$ph24) + ($122<<2)|0);
    HEAP8[$123>>0] = $here$sroa$0$0;
    $124 = (((($next$0$ph24) + ($122<<2)|0)) + 1|0);
    HEAP8[$124>>0] = $108;
    $125 = (((($next$0$ph24) + ($122<<2)|0)) + 2|0);
    HEAP16[$125>>1] = $here$sroa$14$0;
    $126 = ($fill$0|0)==($119|0);
    if ($126) {
     break;
    } else {
     $fill$0 = $121;
    }
   }
   $127 = (($len$3) + -1)|0;
   $128 = 1 << $127;
   $incr$0 = $128;
   while(1) {
    $129 = $incr$0 & $huff$0;
    $130 = ($129|0)==(0);
    $131 = $incr$0 >>> 1;
    if ($130) {
     $incr$0$lcssa = $incr$0;
     break;
    } else {
     $incr$0 = $131;
    }
   }
   $132 = ($incr$0$lcssa|0)==(0);
   if ($132) {
    $huff$1 = 0;
   } else {
    $133 = (($incr$0$lcssa) + -1)|0;
    $134 = $133 & $huff$0;
    $135 = (($134) + ($incr$0$lcssa))|0;
    $huff$1 = $135;
   }
   $136 = (($sym$2) + 1)|0;
   $137 = (($count) + ($len$3<<1)|0);
   $138 = HEAP16[$137>>1]|0;
   $139 = (($138) + -1)<<16>>16;
   HEAP16[$137>>1] = $139;
   $140 = ($139<<16>>16)==(0);
   if ($140) {
    $141 = ($len$3|0)==($max$056$lcssa82|0);
    if ($141) {
     $$lcssa = $108;$huff$1$lcssa = $huff$1;$next$0$ph24$lcssa = $next$0$ph24;$used$0$ph23$lcssa = $used$0$ph23;
     break L93;
    }
    $142 = (($work) + ($136<<1)|0);
    $143 = HEAP16[$142>>1]|0;
    $144 = $143&65535;
    $145 = (($lens) + ($144<<1)|0);
    $146 = HEAP16[$145>>1]|0;
    $147 = $146&65535;
    $len$4 = $147;
   } else {
    $len$4 = $len$3;
   }
   $148 = ($len$4>>>0)>($min$0$max$0$>>>0);
   if (!($148)) {
    $huff$0 = $huff$1;$len$3 = $len$4;$sym$2 = $136;
    continue;
   }
   $149 = $huff$1 & $$ph21;
   $150 = ($149|0)==($low$0$ph|0);
   if ($150) {
    $huff$0 = $huff$1;$len$3 = $len$4;$sym$2 = $136;
   } else {
    $$lcssa121 = $136;$$lcssa122 = $149;$huff$1$lcssa119 = $huff$1;$len$4$lcssa = $len$4;
    break;
   }
  }
  $151 = ($drop$0$ph|0)==(0);
  $min$0$max$0$$drop$0 = $151 ? $min$0$max$0$ : $drop$0$ph;
  $152 = (($next$0$ph24) + ($106<<2)|0);
  $153 = (($len$4$lcssa) - ($min$0$max$0$$drop$0))|0;
  $154 = ($len$4$lcssa>>>0)<($max$056$lcssa82>>>0);
  L116: do {
   if ($154) {
    $155 = 1 << $153;
    $157 = $len$4$lcssa;$curr$139 = $153;$left$140 = $155;
    while(1) {
     $156 = (($count) + ($157<<1)|0);
     $158 = HEAP16[$156>>1]|0;
     $159 = $158&65535;
     $160 = (($left$140) - ($159))|0;
     $161 = ($160|0)<(1);
     if ($161) {
      $curr$1$lcssa = $curr$139;
      break L116;
     }
     $162 = (($curr$139) + 1)|0;
     $163 = $160 << 1;
     $164 = (($162) + ($min$0$max$0$$drop$0))|0;
     $165 = ($164>>>0)<($max$056$lcssa82>>>0);
     if ($165) {
      $157 = $164;$curr$139 = $162;$left$140 = $163;
     } else {
      $curr$1$lcssa = $162;
      break;
     }
    }
   } else {
    $curr$1$lcssa = $153;
   }
  } while(0);
  $166 = 1 << $curr$1$lcssa;
  $167 = (($166) + ($used$0$ph23))|0;
  $168 = ($167>>>0)>(852);
  $or$cond7 = $$ph22 & $168;
  $169 = ($167>>>0)>(592);
  $or$cond9 = $$ph & $169;
  $or$cond10 = $or$cond7 | $or$cond9;
  if ($or$cond10) {
   $$0 = 1;
   label = 45;
   break;
  }
  $170 = $curr$1$lcssa&255;
  $171 = HEAP32[$table>>2]|0;
  $172 = (($171) + ($$lcssa122<<2)|0);
  HEAP8[$172>>0] = $170;
  $173 = (((($171) + ($$lcssa122<<2)|0)) + 1|0);
  HEAP8[$173>>0] = $105;
  $174 = $152;
  $175 = $171;
  $176 = (($174) - ($175))|0;
  $177 = $176 >>> 2;
  $178 = $177&65535;
  $179 = (((($171) + ($$lcssa122<<2)|0)) + 2|0);
  HEAP16[$179>>1] = $178;
  $curr$0$ph = $curr$1$lcssa;$drop$0$ph = $min$0$max$0$$drop$0;$huff$0$ph = $huff$1$lcssa119;$len$3$ph = $len$4$lcssa;$low$0$ph = $$lcssa122;$next$0$ph24 = $152;$sym$2$ph = $$lcssa121;$used$0$ph23 = $167;
 }
 if ((label|0) == 45) {
  STACKTOP = sp;return ($$0|0);
 }
 $180 = ($huff$1$lcssa|0)==(0);
 if (!($180)) {
  $181 = (($next$0$ph24$lcssa) + ($huff$1$lcssa<<2)|0);
  HEAP8[$181>>0] = 64;
  $182 = (((($next$0$ph24$lcssa) + ($huff$1$lcssa<<2)|0)) + 1|0);
  HEAP8[$182>>0] = $$lcssa;
  $183 = (((($next$0$ph24$lcssa) + ($huff$1$lcssa<<2)|0)) + 2|0);
  HEAP16[$183>>1] = 0;
 }
 $184 = HEAP32[$table>>2]|0;
 $185 = (($184) + ($used$0$ph23$lcssa<<2)|0);
 HEAP32[$table>>2] = $185;
 HEAP32[$bits>>2] = $min$0$max$0$;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function __tr_init($s) {
 $s = $s|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($s)) + 148|0);
 $1 = ((($s)) + 2840|0);
 HEAP32[$1>>2] = $0;
 $2 = ((($s)) + 2848|0);
 HEAP32[$2>>2] = 141248;
 $3 = ((($s)) + 2440|0);
 $4 = ((($s)) + 2852|0);
 HEAP32[$4>>2] = $3;
 $5 = ((($s)) + 2860|0);
 HEAP32[$5>>2] = 141268;
 $6 = ((($s)) + 2684|0);
 $7 = ((($s)) + 2864|0);
 HEAP32[$7>>2] = $6;
 $8 = ((($s)) + 2872|0);
 HEAP32[$8>>2] = 141408;
 $9 = ((($s)) + 5816|0);
 HEAP16[$9>>1] = 0;
 $10 = ((($s)) + 5820|0);
 HEAP32[$10>>2] = 0;
 _init_block($s);
 return;
}
function _init_block($s) {
 $s = $s|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond6 = 0, $n$04 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $n$04 = 0;
 while(1) {
  $0 = (((($s)) + 148|0) + ($n$04<<2)|0);
  HEAP16[$0>>1] = 0;
  $1 = (($n$04) + 1)|0;
  $exitcond6 = ($1|0)==(286);
  if ($exitcond6) {
   break;
  } else {
   $n$04 = $1;
  }
 }
 $2 = ((($s)) + 2440|0);
 HEAP16[$2>>1] = 0;
 $3 = ((($s)) + 2444|0);
 HEAP16[$3>>1] = 0;
 $4 = ((($s)) + 2448|0);
 HEAP16[$4>>1] = 0;
 $5 = ((($s)) + 2452|0);
 HEAP16[$5>>1] = 0;
 $6 = ((($s)) + 2456|0);
 HEAP16[$6>>1] = 0;
 $7 = ((($s)) + 2460|0);
 HEAP16[$7>>1] = 0;
 $8 = ((($s)) + 2464|0);
 HEAP16[$8>>1] = 0;
 $9 = ((($s)) + 2468|0);
 HEAP16[$9>>1] = 0;
 $10 = ((($s)) + 2472|0);
 HEAP16[$10>>1] = 0;
 $11 = ((($s)) + 2476|0);
 HEAP16[$11>>1] = 0;
 $12 = ((($s)) + 2480|0);
 HEAP16[$12>>1] = 0;
 $13 = ((($s)) + 2484|0);
 HEAP16[$13>>1] = 0;
 $14 = ((($s)) + 2488|0);
 HEAP16[$14>>1] = 0;
 $15 = ((($s)) + 2492|0);
 HEAP16[$15>>1] = 0;
 $16 = ((($s)) + 2496|0);
 HEAP16[$16>>1] = 0;
 $17 = ((($s)) + 2500|0);
 HEAP16[$17>>1] = 0;
 $18 = ((($s)) + 2504|0);
 HEAP16[$18>>1] = 0;
 $19 = ((($s)) + 2508|0);
 HEAP16[$19>>1] = 0;
 $20 = ((($s)) + 2512|0);
 HEAP16[$20>>1] = 0;
 $21 = ((($s)) + 2516|0);
 HEAP16[$21>>1] = 0;
 $22 = ((($s)) + 2520|0);
 HEAP16[$22>>1] = 0;
 $23 = ((($s)) + 2524|0);
 HEAP16[$23>>1] = 0;
 $24 = ((($s)) + 2528|0);
 HEAP16[$24>>1] = 0;
 $25 = ((($s)) + 2532|0);
 HEAP16[$25>>1] = 0;
 $26 = ((($s)) + 2536|0);
 HEAP16[$26>>1] = 0;
 $27 = ((($s)) + 2540|0);
 HEAP16[$27>>1] = 0;
 $28 = ((($s)) + 2544|0);
 HEAP16[$28>>1] = 0;
 $29 = ((($s)) + 2548|0);
 HEAP16[$29>>1] = 0;
 $30 = ((($s)) + 2552|0);
 HEAP16[$30>>1] = 0;
 $31 = ((($s)) + 2556|0);
 HEAP16[$31>>1] = 0;
 $32 = ((($s)) + 2684|0);
 HEAP16[$32>>1] = 0;
 $33 = ((($s)) + 2688|0);
 HEAP16[$33>>1] = 0;
 $34 = ((($s)) + 2692|0);
 HEAP16[$34>>1] = 0;
 $35 = ((($s)) + 2696|0);
 HEAP16[$35>>1] = 0;
 $36 = ((($s)) + 2700|0);
 HEAP16[$36>>1] = 0;
 $37 = ((($s)) + 2704|0);
 HEAP16[$37>>1] = 0;
 $38 = ((($s)) + 2708|0);
 HEAP16[$38>>1] = 0;
 $39 = ((($s)) + 2712|0);
 HEAP16[$39>>1] = 0;
 $40 = ((($s)) + 2716|0);
 HEAP16[$40>>1] = 0;
 $41 = ((($s)) + 2720|0);
 HEAP16[$41>>1] = 0;
 $42 = ((($s)) + 2724|0);
 HEAP16[$42>>1] = 0;
 $43 = ((($s)) + 2728|0);
 HEAP16[$43>>1] = 0;
 $44 = ((($s)) + 2732|0);
 HEAP16[$44>>1] = 0;
 $45 = ((($s)) + 2736|0);
 HEAP16[$45>>1] = 0;
 $46 = ((($s)) + 2740|0);
 HEAP16[$46>>1] = 0;
 $47 = ((($s)) + 2744|0);
 HEAP16[$47>>1] = 0;
 $48 = ((($s)) + 2748|0);
 HEAP16[$48>>1] = 0;
 $49 = ((($s)) + 2752|0);
 HEAP16[$49>>1] = 0;
 $50 = ((($s)) + 2756|0);
 HEAP16[$50>>1] = 0;
 $51 = ((($s)) + 1172|0);
 HEAP16[$51>>1] = 1;
 $52 = ((($s)) + 5804|0);
 HEAP32[$52>>2] = 0;
 $53 = ((($s)) + 5800|0);
 HEAP32[$53>>2] = 0;
 $54 = ((($s)) + 5808|0);
 HEAP32[$54>>2] = 0;
 $55 = ((($s)) + 5792|0);
 HEAP32[$55>>2] = 0;
 return;
}
function __tr_stored_block($s,$buf,$stored_len,$last) {
 $s = $s|0;
 $buf = $buf|0;
 $stored_len = $stored_len|0;
 $last = $last|0;
 var $$011$i = 0, $$02$i = 0, $$off0 = 0, $$off0$in = 0, $$pre$phi$iZ2D = 0, $$pre$phi4$iZ2D = 0, $$pre3$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $9 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($s)) + 5820|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(13);
 $3 = $last & 65535;
 $4 = $3 << $1;
 $5 = ((($s)) + 5816|0);
 $6 = HEAP16[$5>>1]|0;
 $7 = $6&65535;
 $8 = $7 | $4;
 $9 = $8&65535;
 HEAP16[$5>>1] = $9;
 if ($2) {
  $10 = $8&255;
  $11 = ((($s)) + 20|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = (($12) + 1)|0;
  HEAP32[$11>>2] = $13;
  $14 = ((($s)) + 8|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = (($15) + ($12)|0);
  HEAP8[$16>>0] = $10;
  $17 = HEAP16[$5>>1]|0;
  $18 = ($17&65535) >>> 8;
  $19 = $18&255;
  $20 = HEAP32[$11>>2]|0;
  $21 = (($20) + 1)|0;
  HEAP32[$11>>2] = $21;
  $22 = HEAP32[$14>>2]|0;
  $23 = (($22) + ($20)|0);
  HEAP8[$23>>0] = $19;
  $24 = HEAP32[$0>>2]|0;
  $25 = (16 - ($24))|0;
  $26 = $3 >>> $25;
  $27 = $26&65535;
  HEAP16[$5>>1] = $27;
  $28 = (($24) + -13)|0;
  $$off0$in = $26;$storemerge = $28;
 } else {
  $29 = (($1) + 3)|0;
  $$off0$in = $8;$storemerge = $29;
 }
 $$off0 = $$off0$in&255;
 HEAP32[$0>>2] = $storemerge;
 $30 = ($storemerge|0)>(8);
 do {
  if ($30) {
   $31 = ((($s)) + 20|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = (($32) + 1)|0;
   HEAP32[$31>>2] = $33;
   $34 = ((($s)) + 8|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = (($35) + ($32)|0);
   HEAP8[$36>>0] = $$off0;
   $37 = HEAP16[$5>>1]|0;
   $38 = ($37&65535) >>> 8;
   $39 = $38&255;
   $40 = HEAP32[$31>>2]|0;
   $41 = (($40) + 1)|0;
   HEAP32[$31>>2] = $41;
   $42 = HEAP32[$34>>2]|0;
   $43 = (($42) + ($40)|0);
   HEAP8[$43>>0] = $39;
   $$pre$phi$iZ2D = $31;$$pre$phi4$iZ2D = $34;
  } else {
   $44 = ($storemerge|0)>(0);
   $45 = ((($s)) + 20|0);
   if ($44) {
    $46 = HEAP32[$45>>2]|0;
    $47 = (($46) + 1)|0;
    HEAP32[$45>>2] = $47;
    $48 = ((($s)) + 8|0);
    $49 = HEAP32[$48>>2]|0;
    $50 = (($49) + ($46)|0);
    HEAP8[$50>>0] = $$off0;
    $$pre$phi$iZ2D = $45;$$pre$phi4$iZ2D = $48;
    break;
   } else {
    $$pre3$i = ((($s)) + 8|0);
    $$pre$phi$iZ2D = $45;$$pre$phi4$iZ2D = $$pre3$i;
    break;
   }
  }
 } while(0);
 HEAP16[$5>>1] = 0;
 HEAP32[$0>>2] = 0;
 $51 = $stored_len&255;
 $52 = HEAP32[$$pre$phi$iZ2D>>2]|0;
 $53 = (($52) + 1)|0;
 HEAP32[$$pre$phi$iZ2D>>2] = $53;
 $54 = HEAP32[$$pre$phi4$iZ2D>>2]|0;
 $55 = (($54) + ($52)|0);
 HEAP8[$55>>0] = $51;
 $56 = $stored_len >>> 8;
 $57 = $56&255;
 $58 = HEAP32[$$pre$phi$iZ2D>>2]|0;
 $59 = (($58) + 1)|0;
 HEAP32[$$pre$phi$iZ2D>>2] = $59;
 $60 = HEAP32[$$pre$phi4$iZ2D>>2]|0;
 $61 = (($60) + ($58)|0);
 HEAP8[$61>>0] = $57;
 $62 = $stored_len & 65535;
 $63 = $62 ^ 65535;
 $64 = $63&255;
 $65 = HEAP32[$$pre$phi$iZ2D>>2]|0;
 $66 = (($65) + 1)|0;
 HEAP32[$$pre$phi$iZ2D>>2] = $66;
 $67 = HEAP32[$$pre$phi4$iZ2D>>2]|0;
 $68 = (($67) + ($65)|0);
 HEAP8[$68>>0] = $64;
 $69 = $63 >>> 8;
 $70 = $69&255;
 $71 = HEAP32[$$pre$phi$iZ2D>>2]|0;
 $72 = (($71) + 1)|0;
 HEAP32[$$pre$phi$iZ2D>>2] = $72;
 $73 = HEAP32[$$pre$phi4$iZ2D>>2]|0;
 $74 = (($73) + ($71)|0);
 HEAP8[$74>>0] = $70;
 $75 = ($stored_len|0)==(0);
 if ($75) {
  return;
 } else {
  $$011$i = $stored_len;$$02$i = $buf;
 }
 while(1) {
  $76 = (($$011$i) + -1)|0;
  $77 = ((($$02$i)) + 1|0);
  $78 = HEAP8[$$02$i>>0]|0;
  $79 = HEAP32[$$pre$phi$iZ2D>>2]|0;
  $80 = (($79) + 1)|0;
  HEAP32[$$pre$phi$iZ2D>>2] = $80;
  $81 = HEAP32[$$pre$phi4$iZ2D>>2]|0;
  $82 = (($81) + ($79)|0);
  HEAP8[$82>>0] = $78;
  $83 = ($76|0)==(0);
  if ($83) {
   break;
  } else {
   $$011$i = $76;$$02$i = $77;
  }
 }
 return;
}
function __tr_flush_bits($s) {
 $s = $s|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($s)) + 5820|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(16);
 if ($2) {
  $3 = ((($s)) + 5816|0);
  $4 = HEAP16[$3>>1]|0;
  $5 = $4&255;
  $6 = ((($s)) + 20|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (($7) + 1)|0;
  HEAP32[$6>>2] = $8;
  $9 = ((($s)) + 8|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = (($10) + ($7)|0);
  HEAP8[$11>>0] = $5;
  $12 = HEAP16[$3>>1]|0;
  $13 = ($12&65535) >>> 8;
  $14 = $13&255;
  $15 = HEAP32[$6>>2]|0;
  $16 = (($15) + 1)|0;
  HEAP32[$6>>2] = $16;
  $17 = HEAP32[$9>>2]|0;
  $18 = (($17) + ($15)|0);
  HEAP8[$18>>0] = $14;
  HEAP16[$3>>1] = 0;
  HEAP32[$0>>2] = 0;
  return;
 }
 $19 = ($1|0)>(7);
 if (!($19)) {
  return;
 }
 $20 = ((($s)) + 5816|0);
 $21 = HEAP16[$20>>1]|0;
 $22 = $21&255;
 $23 = ((($s)) + 20|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = (($24) + 1)|0;
 HEAP32[$23>>2] = $25;
 $26 = ((($s)) + 8|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = (($27) + ($24)|0);
 HEAP8[$28>>0] = $22;
 $29 = HEAP16[$20>>1]|0;
 $30 = ($29&65535) >>> 8;
 HEAP16[$20>>1] = $30;
 $31 = HEAP32[$0>>2]|0;
 $32 = (($31) + -8)|0;
 HEAP32[$0>>2] = $32;
 return;
}
function __tr_align($s) {
 $s = $s|0;
 var $$off0 = 0, $$off0$in = 0, $$off03 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $8 = 0, $9 = 0, $storemerge = 0, $storemerge1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($s)) + 5820|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(13);
 $3 = 2 << $1;
 $4 = ((($s)) + 5816|0);
 $5 = HEAP16[$4>>1]|0;
 $6 = $5&65535;
 $7 = $6 | $3;
 $8 = $7&65535;
 HEAP16[$4>>1] = $8;
 if ($2) {
  $9 = $7&255;
  $10 = ((($s)) + 20|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($11) + 1)|0;
  HEAP32[$10>>2] = $12;
  $13 = ((($s)) + 8|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = (($14) + ($11)|0);
  HEAP8[$15>>0] = $9;
  $16 = HEAP16[$4>>1]|0;
  $17 = ($16&65535) >>> 8;
  $18 = $17&255;
  $19 = HEAP32[$10>>2]|0;
  $20 = (($19) + 1)|0;
  HEAP32[$10>>2] = $20;
  $21 = HEAP32[$13>>2]|0;
  $22 = (($21) + ($19)|0);
  HEAP8[$22>>0] = $18;
  $23 = HEAP32[$0>>2]|0;
  $24 = (16 - ($23))|0;
  $25 = 2 >>> $24;
  $26 = $25&65535;
  HEAP16[$4>>1] = $26;
  $27 = (($23) + -13)|0;
  $$off0$in = $25;$storemerge = $27;
 } else {
  $28 = (($1) + 3)|0;
  $$off0$in = $7;$storemerge = $28;
 }
 $$off0 = $$off0$in&255;
 HEAP32[$0>>2] = $storemerge;
 $29 = ($storemerge|0)>(9);
 if ($29) {
  $30 = ((($s)) + 20|0);
  $31 = HEAP32[$30>>2]|0;
  $32 = (($31) + 1)|0;
  HEAP32[$30>>2] = $32;
  $33 = ((($s)) + 8|0);
  $34 = HEAP32[$33>>2]|0;
  $35 = (($34) + ($31)|0);
  HEAP8[$35>>0] = $$off0;
  $36 = HEAP16[$4>>1]|0;
  $37 = ($36&65535) >>> 8;
  $38 = $37&255;
  $39 = HEAP32[$30>>2]|0;
  $40 = (($39) + 1)|0;
  HEAP32[$30>>2] = $40;
  $41 = HEAP32[$33>>2]|0;
  $42 = (($41) + ($39)|0);
  HEAP8[$42>>0] = $38;
  HEAP16[$4>>1] = 0;
  $43 = HEAP32[$0>>2]|0;
  $44 = (($43) + -9)|0;
  $$off03 = 0;$storemerge1 = $44;
 } else {
  $45 = (($storemerge) + 7)|0;
  $$off03 = $$off0;$storemerge1 = $45;
 }
 HEAP32[$0>>2] = $storemerge1;
 $46 = ($storemerge1|0)==(16);
 if ($46) {
  $47 = ((($s)) + 20|0);
  $48 = HEAP32[$47>>2]|0;
  $49 = (($48) + 1)|0;
  HEAP32[$47>>2] = $49;
  $50 = ((($s)) + 8|0);
  $51 = HEAP32[$50>>2]|0;
  $52 = (($51) + ($48)|0);
  HEAP8[$52>>0] = $$off03;
  $53 = HEAP16[$4>>1]|0;
  $54 = ($53&65535) >>> 8;
  $55 = $54&255;
  $56 = HEAP32[$47>>2]|0;
  $57 = (($56) + 1)|0;
  HEAP32[$47>>2] = $57;
  $58 = HEAP32[$50>>2]|0;
  $59 = (($58) + ($56)|0);
  HEAP8[$59>>0] = $55;
  HEAP16[$4>>1] = 0;
  HEAP32[$0>>2] = 0;
  return;
 }
 $60 = ($storemerge1|0)>(7);
 if (!($60)) {
  return;
 }
 $61 = ((($s)) + 20|0);
 $62 = HEAP32[$61>>2]|0;
 $63 = (($62) + 1)|0;
 HEAP32[$61>>2] = $63;
 $64 = ((($s)) + 8|0);
 $65 = HEAP32[$64>>2]|0;
 $66 = (($65) + ($62)|0);
 HEAP8[$66>>0] = $$off03;
 $67 = HEAP16[$4>>1]|0;
 $68 = ($67&65535) >>> 8;
 HEAP16[$4>>1] = $68;
 $69 = HEAP32[$0>>2]|0;
 $70 = (($69) + -8)|0;
 HEAP32[$0>>2] = $70;
 return;
}
function __tr_flush_block($s,$buf,$stored_len,$last) {
 $s = $s|0;
 $buf = $buf|0;
 $stored_len = $stored_len|0;
 $last = $last|0;
 var $$ = 0, $$$i = 0, $$$i$i = 0, $$$i1$i = 0, $$0$i = 0, $$1$i$i = 0, $$1$i2$i = 0, $$2$i$i = 0, $$2$i12$i = 0, $$lcssa = 0, $$lcssa56 = 0, $$lcssa57 = 0, $$lcssa58 = 0, $$lcssa59 = 0, $$lcssa60 = 0, $$lcssa61 = 0, $$lcssa62 = 0, $$lcssa63 = 0, $$lcssa64 = 0, $$pre = 0;
 var $$pre$phi$iZ2D = 0, $$pre$phi25Z2D = 0, $$pre$phiZ2D = 0, $$pre24 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0;
 var $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0;
 var $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0;
 var $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0;
 var $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0;
 var $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0;
 var $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0;
 var $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $black_mask$03$i = 0, $count$0$i$i = 0, $count$0$i10$i = 0, $exitcond$i = 0, $max_blindex$0 = 0, $max_blindex$0$lcssa$i = 0, $max_count$1$ph$be$i$i = 0;
 var $max_count$1$ph$be$i13$i = 0, $max_count$1$ph$i$i = 0, $max_count$1$ph$i6$i = 0, $min_count$1$ph$be$i$i = 0, $min_count$1$ph$be$i14$i = 0, $min_count$1$ph$i$i = 0, $min_count$1$ph$i7$i = 0, $n$0$i$i = 0, $n$0$i9$i = 0, $n$0$ph$i$i = 0, $n$0$ph$i4$i = 0, $n$04$i = 0, $n$12$i = 0, $nextlen$0$ph$i$i = 0, $nextlen$0$ph$i5$i = 0, $opt_lenb$0 = 0, $or$cond = 0, $or$cond$i$i = 0, $or$cond$i11$i = 0, $or$cond3 = 0;
 var $prevlen$0$ph$i$i = 0, $prevlen$0$ph$i$i$phi = 0, $prevlen$0$ph$i3$i = 0, $prevlen$0$ph$i3$i$phi = 0, $rank$04$i = 0, $static_lenb$0 = 0, $storemerge = 0, $storemerge2 = 0, $storemerge2$i = 0, $storemerge3$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($s)) + 132|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(0);
 if ($2) {
  $3 = HEAP32[$s>>2]|0;
  $4 = ((($3)) + 44|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ($5|0)==(2);
  if ($6) {
   $black_mask$03$i = -201342849;$n$04$i = 0;
   while(1) {
    $7 = $black_mask$03$i & 1;
    $8 = ($7|0)==(0);
    if (!($8)) {
     $9 = (((($s)) + 148|0) + ($n$04$i<<2)|0);
     $10 = HEAP16[$9>>1]|0;
     $11 = ($10<<16>>16)==(0);
     if (!($11)) {
      $$0$i = 0;
      break;
     }
    }
    $12 = (($n$04$i) + 1)|0;
    $13 = $black_mask$03$i >>> 1;
    $14 = ($12|0)<(32);
    if ($14) {
     $black_mask$03$i = $13;$n$04$i = $12;
    } else {
     label = 6;
     break;
    }
   }
   L9: do {
    if ((label|0) == 6) {
     $15 = ((($s)) + 184|0);
     $16 = HEAP16[$15>>1]|0;
     $17 = ($16<<16>>16)==(0);
     if ($17) {
      $18 = ((($s)) + 188|0);
      $19 = HEAP16[$18>>1]|0;
      $20 = ($19<<16>>16)==(0);
      if ($20) {
       $21 = ((($s)) + 200|0);
       $22 = HEAP16[$21>>1]|0;
       $23 = ($22<<16>>16)==(0);
       if ($23) {
        $n$12$i = 32;
        while(1) {
         $26 = (((($s)) + 148|0) + ($n$12$i<<2)|0);
         $27 = HEAP16[$26>>1]|0;
         $28 = ($27<<16>>16)==(0);
         $25 = (($n$12$i) + 1)|0;
         if (!($28)) {
          $$0$i = 1;
          break L9;
         }
         $24 = ($25|0)<(256);
         if ($24) {
          $n$12$i = $25;
         } else {
          $$0$i = 0;
          break;
         }
        }
       } else {
        $$0$i = 1;
       }
      } else {
       $$0$i = 1;
      }
     } else {
      $$0$i = 1;
     }
    }
   } while(0);
   HEAP32[$4>>2] = $$0$i;
  }
  $29 = ((($s)) + 2840|0);
  _build_tree($s,$29);
  $30 = ((($s)) + 2852|0);
  _build_tree($s,$30);
  $31 = ((($s)) + 2844|0);
  $32 = HEAP32[$31>>2]|0;
  $33 = ((($s)) + 150|0);
  $34 = HEAP16[$33>>1]|0;
  $35 = $34&65535;
  $36 = ($34<<16>>16)==(0);
  $$$i$i = $36 ? 138 : 7;
  $$1$i$i = $36 ? 3 : 4;
  $37 = (($32) + 1)|0;
  $38 = (((((($s)) + 148|0) + ($37<<2)|0)) + 2|0);
  HEAP16[$38>>1] = -1;
  $39 = ((($s)) + 2752|0);
  $40 = ((($s)) + 2756|0);
  $41 = ((($s)) + 2748|0);
  $max_count$1$ph$i$i = $$$i$i;$min_count$1$ph$i$i = $$1$i$i;$n$0$ph$i$i = 0;$nextlen$0$ph$i$i = $35;$prevlen$0$ph$i$i = -1;
  L18: while(1) {
   $count$0$i$i = 0;$n$0$i$i = $n$0$ph$i$i;
   while(1) {
    $42 = ($n$0$i$i|0)>($32|0);
    if ($42) {
     break L18;
    }
    $43 = (($n$0$i$i) + 1)|0;
    $44 = (((((($s)) + 148|0) + ($43<<2)|0)) + 2|0);
    $45 = HEAP16[$44>>1]|0;
    $46 = $45&65535;
    $47 = (($count$0$i$i) + 1)|0;
    $48 = ($47|0)<($max_count$1$ph$i$i|0);
    $49 = ($nextlen$0$ph$i$i|0)==($46|0);
    $or$cond$i$i = $48 & $49;
    if ($or$cond$i$i) {
     $count$0$i$i = $47;$n$0$i$i = $43;
    } else {
     $$lcssa60 = $43;$$lcssa61 = $45;$$lcssa62 = $46;$$lcssa63 = $47;$$lcssa64 = $49;
     break;
    }
   }
   $50 = ($$lcssa63|0)<($min_count$1$ph$i$i|0);
   do {
    if ($50) {
     $51 = (((($s)) + 2684|0) + ($nextlen$0$ph$i$i<<2)|0);
     $52 = HEAP16[$51>>1]|0;
     $53 = $52&65535;
     $54 = (($53) + ($$lcssa63))|0;
     $55 = $54&65535;
     HEAP16[$51>>1] = $55;
    } else {
     $56 = ($nextlen$0$ph$i$i|0)==(0);
     if ($56) {
      $63 = ($$lcssa63|0)<(11);
      if ($63) {
       $64 = HEAP16[$39>>1]|0;
       $65 = (($64) + 1)<<16>>16;
       HEAP16[$39>>1] = $65;
       break;
      } else {
       $66 = HEAP16[$40>>1]|0;
       $67 = (($66) + 1)<<16>>16;
       HEAP16[$40>>1] = $67;
       break;
      }
     } else {
      $57 = ($nextlen$0$ph$i$i|0)==($prevlen$0$ph$i$i|0);
      if (!($57)) {
       $58 = (((($s)) + 2684|0) + ($nextlen$0$ph$i$i<<2)|0);
       $59 = HEAP16[$58>>1]|0;
       $60 = (($59) + 1)<<16>>16;
       HEAP16[$58>>1] = $60;
      }
      $61 = HEAP16[$41>>1]|0;
      $62 = (($61) + 1)<<16>>16;
      HEAP16[$41>>1] = $62;
      break;
     }
    }
   } while(0);
   $68 = ($$lcssa61<<16>>16)==(0);
   $$2$i$i = $$lcssa64 ? 6 : 7;
   $max_count$1$ph$be$i$i = $68 ? 138 : $$2$i$i;
   $69 = $68 | $$lcssa64;
   $min_count$1$ph$be$i$i = $69 ? 3 : 4;
   $prevlen$0$ph$i$i$phi = $nextlen$0$ph$i$i;$max_count$1$ph$i$i = $max_count$1$ph$be$i$i;$min_count$1$ph$i$i = $min_count$1$ph$be$i$i;$n$0$ph$i$i = $$lcssa60;$nextlen$0$ph$i$i = $$lcssa62;$prevlen$0$ph$i$i = $prevlen$0$ph$i$i$phi;
  }
  $70 = ((($s)) + 2856|0);
  $71 = HEAP32[$70>>2]|0;
  $72 = ((($s)) + 2442|0);
  $73 = HEAP16[$72>>1]|0;
  $74 = $73&65535;
  $75 = ($73<<16>>16)==(0);
  $$$i1$i = $75 ? 138 : 7;
  $$1$i2$i = $75 ? 3 : 4;
  $76 = (($71) + 1)|0;
  $77 = (((((($s)) + 2440|0) + ($76<<2)|0)) + 2|0);
  HEAP16[$77>>1] = -1;
  $max_count$1$ph$i6$i = $$$i1$i;$min_count$1$ph$i7$i = $$1$i2$i;$n$0$ph$i4$i = 0;$nextlen$0$ph$i5$i = $74;$prevlen$0$ph$i3$i = -1;
  L38: while(1) {
   $count$0$i10$i = 0;$n$0$i9$i = $n$0$ph$i4$i;
   while(1) {
    $78 = ($n$0$i9$i|0)>($71|0);
    if ($78) {
     break L38;
    }
    $79 = (($n$0$i9$i) + 1)|0;
    $80 = (((((($s)) + 2440|0) + ($79<<2)|0)) + 2|0);
    $81 = HEAP16[$80>>1]|0;
    $82 = $81&65535;
    $83 = (($count$0$i10$i) + 1)|0;
    $84 = ($83|0)<($max_count$1$ph$i6$i|0);
    $85 = ($nextlen$0$ph$i5$i|0)==($82|0);
    $or$cond$i11$i = $84 & $85;
    if ($or$cond$i11$i) {
     $count$0$i10$i = $83;$n$0$i9$i = $79;
    } else {
     $$lcssa = $79;$$lcssa56 = $81;$$lcssa57 = $82;$$lcssa58 = $83;$$lcssa59 = $85;
     break;
    }
   }
   $86 = ($$lcssa58|0)<($min_count$1$ph$i7$i|0);
   do {
    if ($86) {
     $87 = (((($s)) + 2684|0) + ($nextlen$0$ph$i5$i<<2)|0);
     $88 = HEAP16[$87>>1]|0;
     $89 = $88&65535;
     $90 = (($89) + ($$lcssa58))|0;
     $91 = $90&65535;
     HEAP16[$87>>1] = $91;
    } else {
     $92 = ($nextlen$0$ph$i5$i|0)==(0);
     if ($92) {
      $99 = ($$lcssa58|0)<(11);
      if ($99) {
       $100 = HEAP16[$39>>1]|0;
       $101 = (($100) + 1)<<16>>16;
       HEAP16[$39>>1] = $101;
       break;
      } else {
       $102 = HEAP16[$40>>1]|0;
       $103 = (($102) + 1)<<16>>16;
       HEAP16[$40>>1] = $103;
       break;
      }
     } else {
      $93 = ($nextlen$0$ph$i5$i|0)==($prevlen$0$ph$i3$i|0);
      if (!($93)) {
       $94 = (((($s)) + 2684|0) + ($nextlen$0$ph$i5$i<<2)|0);
       $95 = HEAP16[$94>>1]|0;
       $96 = (($95) + 1)<<16>>16;
       HEAP16[$94>>1] = $96;
      }
      $97 = HEAP16[$41>>1]|0;
      $98 = (($97) + 1)<<16>>16;
      HEAP16[$41>>1] = $98;
      break;
     }
    }
   } while(0);
   $104 = ($$lcssa56<<16>>16)==(0);
   $$2$i12$i = $$lcssa59 ? 6 : 7;
   $max_count$1$ph$be$i13$i = $104 ? 138 : $$2$i12$i;
   $105 = $104 | $$lcssa59;
   $min_count$1$ph$be$i14$i = $105 ? 3 : 4;
   $prevlen$0$ph$i3$i$phi = $nextlen$0$ph$i5$i;$max_count$1$ph$i6$i = $max_count$1$ph$be$i13$i;$min_count$1$ph$i7$i = $min_count$1$ph$be$i14$i;$n$0$ph$i4$i = $$lcssa;$nextlen$0$ph$i5$i = $$lcssa57;$prevlen$0$ph$i3$i = $prevlen$0$ph$i3$i$phi;
  }
  $106 = ((($s)) + 2864|0);
  _build_tree($s,$106);
  $107 = ((($s)) + 2746|0);
  $108 = HEAP16[$107>>1]|0;
  $109 = ($108<<16>>16)==(0);
  if ($109) {
   $110 = ((($s)) + 2690|0);
   $111 = HEAP16[$110>>1]|0;
   $112 = ($111<<16>>16)==(0);
   if ($112) {
    $113 = ((($s)) + 2742|0);
    $114 = HEAP16[$113>>1]|0;
    $115 = ($114<<16>>16)==(0);
    if ($115) {
     $116 = ((($s)) + 2694|0);
     $117 = HEAP16[$116>>1]|0;
     $118 = ($117<<16>>16)==(0);
     if ($118) {
      $119 = ((($s)) + 2738|0);
      $120 = HEAP16[$119>>1]|0;
      $121 = ($120<<16>>16)==(0);
      if ($121) {
       $122 = ((($s)) + 2698|0);
       $123 = HEAP16[$122>>1]|0;
       $124 = ($123<<16>>16)==(0);
       if ($124) {
        $125 = ((($s)) + 2734|0);
        $126 = HEAP16[$125>>1]|0;
        $127 = ($126<<16>>16)==(0);
        if ($127) {
         $128 = ((($s)) + 2702|0);
         $129 = HEAP16[$128>>1]|0;
         $130 = ($129<<16>>16)==(0);
         if ($130) {
          $131 = ((($s)) + 2730|0);
          $132 = HEAP16[$131>>1]|0;
          $133 = ($132<<16>>16)==(0);
          if ($133) {
           $134 = ((($s)) + 2706|0);
           $135 = HEAP16[$134>>1]|0;
           $136 = ($135<<16>>16)==(0);
           if ($136) {
            $137 = ((($s)) + 2726|0);
            $138 = HEAP16[$137>>1]|0;
            $139 = ($138<<16>>16)==(0);
            if ($139) {
             $140 = ((($s)) + 2710|0);
             $141 = HEAP16[$140>>1]|0;
             $142 = ($141<<16>>16)==(0);
             if ($142) {
              $143 = ((($s)) + 2722|0);
              $144 = HEAP16[$143>>1]|0;
              $145 = ($144<<16>>16)==(0);
              if ($145) {
               $146 = ((($s)) + 2714|0);
               $147 = HEAP16[$146>>1]|0;
               $148 = ($147<<16>>16)==(0);
               if ($148) {
                $149 = ((($s)) + 2718|0);
                $150 = HEAP16[$149>>1]|0;
                $151 = ($150<<16>>16)==(0);
                if ($151) {
                 $152 = ((($s)) + 2686|0);
                 $153 = HEAP16[$152>>1]|0;
                 $154 = ($153<<16>>16)==(0);
                 $$$i = $154 ? 2 : 3;
                 $max_blindex$0$lcssa$i = $$$i;
                } else {
                 $max_blindex$0$lcssa$i = 4;
                }
               } else {
                $max_blindex$0$lcssa$i = 5;
               }
              } else {
               $max_blindex$0$lcssa$i = 6;
              }
             } else {
              $max_blindex$0$lcssa$i = 7;
             }
            } else {
             $max_blindex$0$lcssa$i = 8;
            }
           } else {
            $max_blindex$0$lcssa$i = 9;
           }
          } else {
           $max_blindex$0$lcssa$i = 10;
          }
         } else {
          $max_blindex$0$lcssa$i = 11;
         }
        } else {
         $max_blindex$0$lcssa$i = 12;
        }
       } else {
        $max_blindex$0$lcssa$i = 13;
       }
      } else {
       $max_blindex$0$lcssa$i = 14;
      }
     } else {
      $max_blindex$0$lcssa$i = 15;
     }
    } else {
     $max_blindex$0$lcssa$i = 16;
    }
   } else {
    $max_blindex$0$lcssa$i = 17;
   }
  } else {
   $max_blindex$0$lcssa$i = 18;
  }
  $155 = ($max_blindex$0$lcssa$i*3)|0;
  $156 = (($155) + 17)|0;
  $157 = ((($s)) + 5800|0);
  $158 = HEAP32[$157>>2]|0;
  $159 = (($156) + ($158))|0;
  HEAP32[$157>>2] = $159;
  $160 = (($159) + 10)|0;
  $161 = $160 >>> 3;
  $162 = ((($s)) + 5804|0);
  $163 = HEAP32[$162>>2]|0;
  $164 = (($163) + 10)|0;
  $165 = $164 >>> 3;
  $166 = ($165>>>0)>($161>>>0);
  $$ = $166 ? $161 : $165;
  $max_blindex$0 = $max_blindex$0$lcssa$i;$opt_lenb$0 = $$;$static_lenb$0 = $165;
 } else {
  $167 = (($stored_len) + 5)|0;
  $max_blindex$0 = 0;$opt_lenb$0 = $167;$static_lenb$0 = $167;
 }
 $168 = (($stored_len) + 4)|0;
 $169 = ($168>>>0)<=($opt_lenb$0>>>0);
 $170 = ($buf|0)!=(0|0);
 $or$cond = $170 & $169;
 do {
  if ($or$cond) {
   __tr_stored_block($s,$buf,$stored_len,$last);
  } else {
   $171 = ((($s)) + 136|0);
   $172 = HEAP32[$171>>2]|0;
   $173 = ($172|0)==(4);
   $174 = ($static_lenb$0|0)==($opt_lenb$0|0);
   $or$cond3 = $174 | $173;
   $175 = ((($s)) + 5820|0);
   $176 = HEAP32[$175>>2]|0;
   $177 = ($176|0)>(13);
   if ($or$cond3) {
    $178 = (($last) + 2)|0;
    $179 = $178 & 65535;
    $180 = $179 << $176;
    $181 = ((($s)) + 5816|0);
    $182 = HEAP16[$181>>1]|0;
    $183 = $182&65535;
    $184 = $183 | $180;
    $185 = $184&65535;
    HEAP16[$181>>1] = $185;
    if ($177) {
     $186 = $184&255;
     $187 = ((($s)) + 20|0);
     $188 = HEAP32[$187>>2]|0;
     $189 = (($188) + 1)|0;
     HEAP32[$187>>2] = $189;
     $190 = ((($s)) + 8|0);
     $191 = HEAP32[$190>>2]|0;
     $192 = (($191) + ($188)|0);
     HEAP8[$192>>0] = $186;
     $193 = HEAP16[$181>>1]|0;
     $194 = ($193&65535) >>> 8;
     $195 = $194&255;
     $196 = HEAP32[$187>>2]|0;
     $197 = (($196) + 1)|0;
     HEAP32[$187>>2] = $197;
     $198 = HEAP32[$190>>2]|0;
     $199 = (($198) + ($196)|0);
     HEAP8[$199>>0] = $195;
     $200 = HEAP32[$175>>2]|0;
     $201 = (16 - ($200))|0;
     $202 = $179 >>> $201;
     $203 = $202&65535;
     HEAP16[$181>>1] = $203;
     $204 = (($200) + -13)|0;
     $storemerge2 = $204;
    } else {
     $205 = (($176) + 3)|0;
     $storemerge2 = $205;
    }
    HEAP32[$175>>2] = $storemerge2;
    _compress_block($s,177928,179080);
    break;
   }
   $206 = (($last) + 4)|0;
   $207 = $206 & 65535;
   $208 = $207 << $176;
   $209 = ((($s)) + 5816|0);
   $210 = HEAP16[$209>>1]|0;
   $211 = $210&65535;
   $212 = $211 | $208;
   $213 = $212&65535;
   HEAP16[$209>>1] = $213;
   if ($177) {
    $214 = $212&255;
    $215 = ((($s)) + 20|0);
    $216 = HEAP32[$215>>2]|0;
    $217 = (($216) + 1)|0;
    HEAP32[$215>>2] = $217;
    $218 = ((($s)) + 8|0);
    $219 = HEAP32[$218>>2]|0;
    $220 = (($219) + ($216)|0);
    HEAP8[$220>>0] = $214;
    $221 = HEAP16[$209>>1]|0;
    $222 = ($221&65535) >>> 8;
    $223 = $222&255;
    $224 = HEAP32[$215>>2]|0;
    $225 = (($224) + 1)|0;
    HEAP32[$215>>2] = $225;
    $226 = HEAP32[$218>>2]|0;
    $227 = (($226) + ($224)|0);
    HEAP8[$227>>0] = $223;
    $228 = HEAP32[$175>>2]|0;
    $229 = (16 - ($228))|0;
    $230 = $207 >>> $229;
    $231 = $230&65535;
    HEAP16[$209>>1] = $231;
    $232 = (($228) + -13)|0;
    $243 = $230;$storemerge = $232;
   } else {
    $233 = (($176) + 3)|0;
    $243 = $212;$storemerge = $233;
   }
   HEAP32[$175>>2] = $storemerge;
   $234 = ((($s)) + 2844|0);
   $235 = HEAP32[$234>>2]|0;
   $236 = ((($s)) + 2856|0);
   $237 = HEAP32[$236>>2]|0;
   $238 = ($storemerge|0)>(11);
   $239 = (($235) + 65280)|0;
   $240 = $239 & 65535;
   $241 = $240 << $storemerge;
   $242 = $243 & 65535;
   $244 = $242 | $241;
   $245 = $244&65535;
   HEAP16[$209>>1] = $245;
   if ($238) {
    $246 = $244&255;
    $247 = ((($s)) + 20|0);
    $248 = HEAP32[$247>>2]|0;
    $249 = (($248) + 1)|0;
    HEAP32[$247>>2] = $249;
    $250 = ((($s)) + 8|0);
    $251 = HEAP32[$250>>2]|0;
    $252 = (($251) + ($248)|0);
    HEAP8[$252>>0] = $246;
    $253 = HEAP16[$209>>1]|0;
    $254 = ($253&65535) >>> 8;
    $255 = $254&255;
    $256 = HEAP32[$247>>2]|0;
    $257 = (($256) + 1)|0;
    HEAP32[$247>>2] = $257;
    $258 = HEAP32[$250>>2]|0;
    $259 = (($258) + ($256)|0);
    HEAP8[$259>>0] = $255;
    $260 = HEAP32[$175>>2]|0;
    $261 = (16 - ($260))|0;
    $262 = $240 >>> $261;
    $263 = $262&65535;
    HEAP16[$209>>1] = $263;
    $264 = (($260) + -11)|0;
    $266 = $264;$271 = $262;
   } else {
    $265 = (($storemerge) + 5)|0;
    $266 = $265;$271 = $244;
   }
   HEAP32[$175>>2] = $266;
   $267 = ($266|0)>(11);
   $268 = $237 & 65535;
   $269 = $268 << $266;
   $270 = $271 & 65535;
   $272 = $269 | $270;
   $273 = $272&65535;
   HEAP16[$209>>1] = $273;
   if ($267) {
    $274 = $272&255;
    $275 = ((($s)) + 20|0);
    $276 = HEAP32[$275>>2]|0;
    $277 = (($276) + 1)|0;
    HEAP32[$275>>2] = $277;
    $278 = ((($s)) + 8|0);
    $279 = HEAP32[$278>>2]|0;
    $280 = (($279) + ($276)|0);
    HEAP8[$280>>0] = $274;
    $281 = HEAP16[$209>>1]|0;
    $282 = ($281&65535) >>> 8;
    $283 = $282&255;
    $284 = HEAP32[$275>>2]|0;
    $285 = (($284) + 1)|0;
    HEAP32[$275>>2] = $285;
    $286 = HEAP32[$278>>2]|0;
    $287 = (($286) + ($284)|0);
    HEAP8[$287>>0] = $283;
    $288 = HEAP32[$175>>2]|0;
    $289 = (16 - ($288))|0;
    $290 = $268 >>> $289;
    $291 = $290&65535;
    HEAP16[$209>>1] = $291;
    $292 = (($288) + -11)|0;
    $294 = $292;$300 = $290;
   } else {
    $293 = (($266) + 5)|0;
    $294 = $293;$300 = $272;
   }
   HEAP32[$175>>2] = $294;
   $295 = ($294|0)>(12);
   $296 = (($max_blindex$0) + 65533)|0;
   $297 = $296 & 65535;
   $298 = $297 << $294;
   $299 = $300 & 65535;
   $301 = $298 | $299;
   $302 = $301&65535;
   HEAP16[$209>>1] = $302;
   if ($295) {
    $303 = $301&255;
    $304 = ((($s)) + 20|0);
    $305 = HEAP32[$304>>2]|0;
    $306 = (($305) + 1)|0;
    HEAP32[$304>>2] = $306;
    $307 = ((($s)) + 8|0);
    $308 = HEAP32[$307>>2]|0;
    $309 = (($308) + ($305)|0);
    HEAP8[$309>>0] = $303;
    $310 = HEAP16[$209>>1]|0;
    $311 = ($310&65535) >>> 8;
    $312 = $311&255;
    $313 = HEAP32[$304>>2]|0;
    $314 = (($313) + 1)|0;
    HEAP32[$304>>2] = $314;
    $315 = HEAP32[$307>>2]|0;
    $316 = (($315) + ($313)|0);
    HEAP8[$316>>0] = $312;
    $317 = HEAP32[$175>>2]|0;
    $318 = (16 - ($317))|0;
    $319 = $297 >>> $318;
    $320 = $319&65535;
    HEAP16[$209>>1] = $320;
    $321 = (($317) + -12)|0;
    $$pre$phi25Z2D = $307;$$pre$phiZ2D = $304;$387 = $319;$storemerge2$i = $321;
   } else {
    $322 = (($294) + 4)|0;
    $$pre = ((($s)) + 20|0);
    $$pre24 = ((($s)) + 8|0);
    $$pre$phi25Z2D = $$pre24;$$pre$phiZ2D = $$pre;$387 = $301;$storemerge2$i = $322;
   }
   HEAP32[$175>>2] = $storemerge2$i;
   $324 = $storemerge2$i;$333 = $387;$rank$04$i = 0;
   while(1) {
    $323 = ($324|0)>(13);
    $325 = (613558 + ($rank$04$i)|0);
    $326 = HEAP8[$325>>0]|0;
    $327 = $326&255;
    $328 = (((((($s)) + 2684|0) + ($327<<2)|0)) + 2|0);
    $329 = HEAP16[$328>>1]|0;
    $330 = $329&65535;
    $331 = $330 << $324;
    $332 = $333 & 65535;
    $334 = $331 | $332;
    $335 = $334&65535;
    HEAP16[$209>>1] = $335;
    if ($323) {
     $336 = $334&255;
     $337 = HEAP32[$$pre$phiZ2D>>2]|0;
     $338 = (($337) + 1)|0;
     HEAP32[$$pre$phiZ2D>>2] = $338;
     $339 = HEAP32[$$pre$phi25Z2D>>2]|0;
     $340 = (($339) + ($337)|0);
     HEAP8[$340>>0] = $336;
     $341 = HEAP16[$209>>1]|0;
     $342 = ($341&65535) >>> 8;
     $343 = $342&255;
     $344 = HEAP32[$$pre$phiZ2D>>2]|0;
     $345 = (($344) + 1)|0;
     HEAP32[$$pre$phiZ2D>>2] = $345;
     $346 = HEAP32[$$pre$phi25Z2D>>2]|0;
     $347 = (($346) + ($344)|0);
     HEAP8[$347>>0] = $343;
     $348 = HEAP32[$175>>2]|0;
     $349 = (16 - ($348))|0;
     $350 = $330 >>> $349;
     $351 = $350&65535;
     HEAP16[$209>>1] = $351;
     $352 = (($348) + -13)|0;
     $388 = $350;$storemerge3$i = $352;
    } else {
     $353 = (($324) + 3)|0;
     $388 = $334;$storemerge3$i = $353;
    }
    HEAP32[$175>>2] = $storemerge3$i;
    $354 = (($rank$04$i) + 1)|0;
    $exitcond$i = ($rank$04$i|0)==($max_blindex$0|0);
    if ($exitcond$i) {
     break;
    } else {
     $324 = $storemerge3$i;$333 = $388;$rank$04$i = $354;
    }
   }
   $355 = ((($s)) + 148|0);
   _send_tree($s,$355,$235);
   $356 = ((($s)) + 2440|0);
   _send_tree($s,$356,$237);
   _compress_block($s,$355,$356);
  }
 } while(0);
 _init_block($s);
 $357 = ($last|0)==(0);
 if ($357) {
  return;
 }
 $358 = ((($s)) + 5820|0);
 $359 = HEAP32[$358>>2]|0;
 $360 = ($359|0)>(8);
 if ($360) {
  $361 = ((($s)) + 5816|0);
  $362 = HEAP16[$361>>1]|0;
  $363 = $362&255;
  $364 = ((($s)) + 20|0);
  $365 = HEAP32[$364>>2]|0;
  $366 = (($365) + 1)|0;
  HEAP32[$364>>2] = $366;
  $367 = ((($s)) + 8|0);
  $368 = HEAP32[$367>>2]|0;
  $369 = (($368) + ($365)|0);
  HEAP8[$369>>0] = $363;
  $370 = HEAP16[$361>>1]|0;
  $371 = ($370&65535) >>> 8;
  $372 = $371&255;
  $373 = HEAP32[$364>>2]|0;
  $374 = (($373) + 1)|0;
  HEAP32[$364>>2] = $374;
  $375 = HEAP32[$367>>2]|0;
  $376 = (($375) + ($373)|0);
  HEAP8[$376>>0] = $372;
  $$pre$phi$iZ2D = $361;
 } else {
  $377 = ($359|0)>(0);
  $378 = ((($s)) + 5816|0);
  if ($377) {
   $379 = HEAP16[$378>>1]|0;
   $380 = $379&255;
   $381 = ((($s)) + 20|0);
   $382 = HEAP32[$381>>2]|0;
   $383 = (($382) + 1)|0;
   HEAP32[$381>>2] = $383;
   $384 = ((($s)) + 8|0);
   $385 = HEAP32[$384>>2]|0;
   $386 = (($385) + ($382)|0);
   HEAP8[$386>>0] = $380;
   $$pre$phi$iZ2D = $378;
  } else {
   $$pre$phi$iZ2D = $378;
  }
 }
 HEAP16[$$pre$phi$iZ2D>>1] = 0;
 HEAP32[$358>>2] = 0;
 return;
}
function _build_tree($s,$desc) {
 $s = $s|0;
 $desc = $desc|0;
 var $$ = 0, $$$i = 0, $$$us = 0, $$$us$i = 0, $$0$i$i = 0, $$0$lcssa$i = 0, $$0$lcssa$i13 = 0, $$0$lcssa$i21 = 0, $$01$i$i = 0, $$02$i = 0, $$02$i10 = 0, $$02$i18 = 0, $$1 = 0, $$lcssa = 0, $$lcssa81 = 0, $$lcssa82 = 0, $$lcssa83 = 0, $$lcssa84 = 0, $$lcssa85 = 0, $$max_code$2 = 0;
 var $$max_code$2$us = 0, $$phi$trans$insert$i = 0, $$pre = 0, $$pre$i = 0, $$pre50 = 0, $$pre51$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0;
 var $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0;
 var $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0;
 var $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0;
 var $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0;
 var $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0;
 var $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0;
 var $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bits$03$i = 0, $bits$2$i = 0, $bits$2$in$i = 0;
 var $bits$2$in$i$lcssa = 0, $bits$36$i = 0, $code$04$i = 0, $desc$idx$val = 0, $desc$idx2$val = 0, $desc$idx3$val = 0, $exitcond = 0, $exitcond$i = 0, $exitcond$i5 = 0, $exitcond20$i = 0, $exitcond5$i = 0, $h$0$i = 0, $h$0$us$i = 0, $h$010$i = 0, $h$010$us$i = 0, $h$08$i = 0, $h$17$i = 0, $h$2$i = 0, $h$2$ph$i$lcssa = 0, $h$2$ph$i33 = 0;
 var $j$0$i = 0, $j$0$i12 = 0, $j$0$i20 = 0, $j$01$i = 0, $j$03$i = 0, $j$03$i17 = 0, $j$03$i9 = 0, $j$1$i = 0, $j$1$i11 = 0, $j$1$i19 = 0, $max_code$0$lcssa53 = 0, $max_code$039 = 0, $max_code$1 = 0, $max_code$1$lcssa = 0, $max_code$2$lcssa = 0, $max_code$235 = 0, $max_code$235$us = 0, $n$0$ph$i32 = 0, $n$01$i = 0, $n$038 = 0;
 var $n$134 = 0, $next_code$i = 0, $node$0 = 0, $overflow$0$$i = 0, $overflow$0$$us$i = 0, $overflow$0$lcssa$i = 0, $overflow$09$i = 0, $overflow$09$us$i = 0, $overflow$2$i = 0, $res$0$i$i = 0, $scevgep$i = 0, $xbits$0$i = 0, $xbits$0$us$i = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $next_code$i = sp;
 $0 = HEAP32[$desc>>2]|0;
 $1 = ((($desc)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($2)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($s)) + 5200|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($s)) + 5204|0);
 HEAP32[$7>>2] = 573;
 $8 = ($5|0)>(0);
 if ($8) {
  $max_code$039 = -1;$n$038 = 0;
  while(1) {
   $24 = (($0) + ($n$038<<2)|0);
   $25 = HEAP16[$24>>1]|0;
   $26 = ($25<<16>>16)==(0);
   if ($26) {
    $31 = (((($0) + ($n$038<<2)|0)) + 2|0);
    HEAP16[$31>>1] = 0;
    $max_code$1 = $max_code$039;
   } else {
    $27 = HEAP32[$6>>2]|0;
    $28 = (($27) + 1)|0;
    HEAP32[$6>>2] = $28;
    $29 = (((($s)) + 2908|0) + ($28<<2)|0);
    HEAP32[$29>>2] = $n$038;
    $30 = (((($s)) + 5208|0) + ($n$038)|0);
    HEAP8[$30>>0] = 0;
    $max_code$1 = $n$038;
   }
   $32 = (($n$038) + 1)|0;
   $exitcond = ($32|0)==($5|0);
   if ($exitcond) {
    $max_code$1$lcssa = $max_code$1;
    break;
   } else {
    $max_code$039 = $max_code$1;$n$038 = $32;
   }
  }
  $$pre = HEAP32[$6>>2]|0;
  $9 = ($$pre|0)<(2);
  if ($9) {
   $350 = $$pre;$max_code$0$lcssa53 = $max_code$1$lcssa;
   label = 3;
  } else {
   $51 = $$pre;$max_code$2$lcssa = $max_code$1$lcssa;
  }
 } else {
  $350 = 0;$max_code$0$lcssa53 = -1;
  label = 3;
 }
 if ((label|0) == 3) {
  $10 = ((($s)) + 5800|0);
  $11 = ($3|0)==(0|0);
  $12 = ((($s)) + 5804|0);
  if ($11) {
   $16 = $350;$max_code$235$us = $max_code$0$lcssa53;
   while(1) {
    $13 = ($max_code$235$us|0)<(2);
    $14 = (($max_code$235$us) + 1)|0;
    $$max_code$2$us = $13 ? $14 : $max_code$235$us;
    $$$us = $13 ? $14 : 0;
    $15 = (($16) + 1)|0;
    HEAP32[$6>>2] = $15;
    $17 = (((($s)) + 2908|0) + ($15<<2)|0);
    HEAP32[$17>>2] = $$$us;
    $18 = (($0) + ($$$us<<2)|0);
    HEAP16[$18>>1] = 1;
    $19 = (((($s)) + 5208|0) + ($$$us)|0);
    HEAP8[$19>>0] = 0;
    $20 = HEAP32[$10>>2]|0;
    $21 = (($20) + -1)|0;
    HEAP32[$10>>2] = $21;
    $22 = HEAP32[$6>>2]|0;
    $23 = ($22|0)<(2);
    if ($23) {
     $16 = $22;$max_code$235$us = $$max_code$2$us;
    } else {
     $51 = $22;$max_code$2$lcssa = $$max_code$2$us;
     break;
    }
   }
  } else {
   $36 = $350;$max_code$235 = $max_code$0$lcssa53;
   while(1) {
    $33 = ($max_code$235|0)<(2);
    $34 = (($max_code$235) + 1)|0;
    $$max_code$2 = $33 ? $34 : $max_code$235;
    $$ = $33 ? $34 : 0;
    $35 = (($36) + 1)|0;
    HEAP32[$6>>2] = $35;
    $37 = (((($s)) + 2908|0) + ($35<<2)|0);
    HEAP32[$37>>2] = $$;
    $38 = (($0) + ($$<<2)|0);
    HEAP16[$38>>1] = 1;
    $39 = (((($s)) + 5208|0) + ($$)|0);
    HEAP8[$39>>0] = 0;
    $40 = HEAP32[$10>>2]|0;
    $41 = (($40) + -1)|0;
    HEAP32[$10>>2] = $41;
    $42 = (((($3) + ($$<<2)|0)) + 2|0);
    $43 = HEAP16[$42>>1]|0;
    $44 = $43&65535;
    $45 = HEAP32[$12>>2]|0;
    $46 = (($45) - ($44))|0;
    HEAP32[$12>>2] = $46;
    $47 = HEAP32[$6>>2]|0;
    $48 = ($47|0)<(2);
    if ($48) {
     $36 = $47;$max_code$235 = $$max_code$2;
    } else {
     $51 = $47;$max_code$2$lcssa = $$max_code$2;
     break;
    }
   }
  }
 }
 $49 = ((($desc)) + 4|0);
 HEAP32[$49>>2] = $max_code$2$lcssa;
 $50 = (($51|0) / 2)&-1;
 $57 = $51;$n$134 = $50;
 while(1) {
  $53 = (((($s)) + 2908|0) + ($n$134<<2)|0);
  $54 = HEAP32[$53>>2]|0;
  $55 = (((($s)) + 5208|0) + ($54)|0);
  $j$01$i = $n$134 << 1;
  $56 = ($j$01$i|0)>($57|0);
  L19: do {
   if ($56) {
    $$0$lcssa$i = $n$134;
   } else {
    $58 = (($0) + ($54<<2)|0);
    $$02$i = $n$134;$60 = $57;$j$03$i = $j$01$i;
    while(1) {
     $59 = ($j$03$i|0)<($60|0);
     do {
      if ($59) {
       $61 = $j$03$i | 1;
       $62 = (((($s)) + 2908|0) + ($61<<2)|0);
       $63 = HEAP32[$62>>2]|0;
       $64 = (($0) + ($63<<2)|0);
       $65 = HEAP16[$64>>1]|0;
       $66 = (((($s)) + 2908|0) + ($j$03$i<<2)|0);
       $67 = HEAP32[$66>>2]|0;
       $68 = (($0) + ($67<<2)|0);
       $69 = HEAP16[$68>>1]|0;
       $70 = ($65&65535)<($69&65535);
       if (!($70)) {
        $71 = ($65<<16>>16)==($69<<16>>16);
        if (!($71)) {
         $j$1$i = $j$03$i;
         break;
        }
        $72 = (((($s)) + 5208|0) + ($63)|0);
        $73 = HEAP8[$72>>0]|0;
        $74 = (((($s)) + 5208|0) + ($67)|0);
        $75 = HEAP8[$74>>0]|0;
        $76 = ($73&255)>($75&255);
        if ($76) {
         $j$1$i = $j$03$i;
         break;
        }
       }
       $j$1$i = $61;
      } else {
       $j$1$i = $j$03$i;
      }
     } while(0);
     $77 = HEAP16[$58>>1]|0;
     $78 = (((($s)) + 2908|0) + ($j$1$i<<2)|0);
     $79 = HEAP32[$78>>2]|0;
     $80 = (($0) + ($79<<2)|0);
     $81 = HEAP16[$80>>1]|0;
     $82 = ($77&65535)<($81&65535);
     if ($82) {
      $$0$lcssa$i = $$02$i;
      break L19;
     }
     $83 = ($77<<16>>16)==($81<<16>>16);
     if ($83) {
      $84 = HEAP8[$55>>0]|0;
      $85 = (((($s)) + 5208|0) + ($79)|0);
      $86 = HEAP8[$85>>0]|0;
      $87 = ($84&255)>($86&255);
      if (!($87)) {
       $$0$lcssa$i = $$02$i;
       break L19;
      }
     }
     $88 = (((($s)) + 2908|0) + ($$02$i<<2)|0);
     HEAP32[$88>>2] = $79;
     $j$0$i = $j$1$i << 1;
     $89 = HEAP32[$6>>2]|0;
     $90 = ($j$0$i|0)>($89|0);
     if ($90) {
      $$0$lcssa$i = $j$1$i;
      break;
     } else {
      $$02$i = $j$1$i;$60 = $89;$j$03$i = $j$0$i;
     }
    }
   }
  } while(0);
  $91 = (((($s)) + 2908|0) + ($$0$lcssa$i<<2)|0);
  HEAP32[$91>>2] = $54;
  $92 = ($n$134|0)>(1);
  if (!($92)) {
   break;
  }
  $93 = (($n$134) + -1)|0;
  $$pre50 = HEAP32[$6>>2]|0;
  $57 = $$pre50;$n$134 = $93;
 }
 $$pre51$pre = HEAP32[$6>>2]|0;
 $52 = ((($s)) + 2912|0);
 $96 = $$pre51$pre;$node$0 = $5;
 while(1) {
  $94 = HEAP32[$52>>2]|0;
  $95 = (($96) + -1)|0;
  HEAP32[$6>>2] = $95;
  $97 = (((($s)) + 2908|0) + ($96<<2)|0);
  $98 = HEAP32[$97>>2]|0;
  HEAP32[$52>>2] = $98;
  $99 = (((($s)) + 5208|0) + ($98)|0);
  $100 = ($96|0)<(3);
  L39: do {
   if ($100) {
    $$0$lcssa$i21 = 1;
   } else {
    $101 = (($0) + ($98<<2)|0);
    $$02$i18 = 1;$103 = $95;$j$03$i17 = 2;
    while(1) {
     $102 = ($j$03$i17|0)<($103|0);
     do {
      if ($102) {
       $104 = $j$03$i17 | 1;
       $105 = (((($s)) + 2908|0) + ($104<<2)|0);
       $106 = HEAP32[$105>>2]|0;
       $107 = (($0) + ($106<<2)|0);
       $108 = HEAP16[$107>>1]|0;
       $109 = (((($s)) + 2908|0) + ($j$03$i17<<2)|0);
       $110 = HEAP32[$109>>2]|0;
       $111 = (($0) + ($110<<2)|0);
       $112 = HEAP16[$111>>1]|0;
       $113 = ($108&65535)<($112&65535);
       if (!($113)) {
        $114 = ($108<<16>>16)==($112<<16>>16);
        if (!($114)) {
         $j$1$i19 = $j$03$i17;
         break;
        }
        $115 = (((($s)) + 5208|0) + ($106)|0);
        $116 = HEAP8[$115>>0]|0;
        $117 = (((($s)) + 5208|0) + ($110)|0);
        $118 = HEAP8[$117>>0]|0;
        $119 = ($116&255)>($118&255);
        if ($119) {
         $j$1$i19 = $j$03$i17;
         break;
        }
       }
       $j$1$i19 = $104;
      } else {
       $j$1$i19 = $j$03$i17;
      }
     } while(0);
     $120 = HEAP16[$101>>1]|0;
     $121 = (((($s)) + 2908|0) + ($j$1$i19<<2)|0);
     $122 = HEAP32[$121>>2]|0;
     $123 = (($0) + ($122<<2)|0);
     $124 = HEAP16[$123>>1]|0;
     $125 = ($120&65535)<($124&65535);
     if ($125) {
      $$0$lcssa$i21 = $$02$i18;
      break L39;
     }
     $126 = ($120<<16>>16)==($124<<16>>16);
     if ($126) {
      $127 = HEAP8[$99>>0]|0;
      $128 = (((($s)) + 5208|0) + ($122)|0);
      $129 = HEAP8[$128>>0]|0;
      $130 = ($127&255)>($129&255);
      if (!($130)) {
       $$0$lcssa$i21 = $$02$i18;
       break L39;
      }
     }
     $131 = (((($s)) + 2908|0) + ($$02$i18<<2)|0);
     HEAP32[$131>>2] = $122;
     $j$0$i20 = $j$1$i19 << 1;
     $132 = HEAP32[$6>>2]|0;
     $133 = ($j$0$i20|0)>($132|0);
     if ($133) {
      $$0$lcssa$i21 = $j$1$i19;
      break;
     } else {
      $$02$i18 = $j$1$i19;$103 = $132;$j$03$i17 = $j$0$i20;
     }
    }
   }
  } while(0);
  $134 = (((($s)) + 2908|0) + ($$0$lcssa$i21<<2)|0);
  HEAP32[$134>>2] = $98;
  $135 = HEAP32[$52>>2]|0;
  $136 = HEAP32[$7>>2]|0;
  $137 = (($136) + -1)|0;
  HEAP32[$7>>2] = $137;
  $138 = (((($s)) + 2908|0) + ($137<<2)|0);
  HEAP32[$138>>2] = $94;
  $139 = HEAP32[$7>>2]|0;
  $140 = (($139) + -1)|0;
  HEAP32[$7>>2] = $140;
  $141 = (((($s)) + 2908|0) + ($140<<2)|0);
  HEAP32[$141>>2] = $135;
  $142 = (($0) + ($94<<2)|0);
  $143 = HEAP16[$142>>1]|0;
  $144 = $143&65535;
  $145 = (($0) + ($135<<2)|0);
  $146 = HEAP16[$145>>1]|0;
  $147 = $146&65535;
  $148 = (($147) + ($144))|0;
  $149 = $148&65535;
  $150 = (($0) + ($node$0<<2)|0);
  HEAP16[$150>>1] = $149;
  $151 = (((($s)) + 5208|0) + ($94)|0);
  $152 = HEAP8[$151>>0]|0;
  $153 = (((($s)) + 5208|0) + ($135)|0);
  $154 = HEAP8[$153>>0]|0;
  $155 = ($152&255)<($154&255);
  $$1 = $155 ? $154 : $152;
  $156 = $$1&255;
  $157 = (($156) + 1)|0;
  $158 = $157&255;
  $159 = (((($s)) + 5208|0) + ($node$0)|0);
  HEAP8[$159>>0] = $158;
  $160 = $node$0&65535;
  $161 = (((($0) + ($135<<2)|0)) + 2|0);
  HEAP16[$161>>1] = $160;
  $162 = (((($0) + ($94<<2)|0)) + 2|0);
  HEAP16[$162>>1] = $160;
  $163 = (($node$0) + 1)|0;
  HEAP32[$52>>2] = $node$0;
  $164 = HEAP32[$6>>2]|0;
  $165 = ($164|0)<(2);
  L55: do {
   if ($165) {
    $$0$lcssa$i13 = 1;
   } else {
    $$02$i10 = 1;$167 = $164;$j$03$i9 = 2;
    while(1) {
     $166 = ($j$03$i9|0)<($167|0);
     do {
      if ($166) {
       $168 = $j$03$i9 | 1;
       $169 = (((($s)) + 2908|0) + ($168<<2)|0);
       $170 = HEAP32[$169>>2]|0;
       $171 = (($0) + ($170<<2)|0);
       $172 = HEAP16[$171>>1]|0;
       $173 = (((($s)) + 2908|0) + ($j$03$i9<<2)|0);
       $174 = HEAP32[$173>>2]|0;
       $175 = (($0) + ($174<<2)|0);
       $176 = HEAP16[$175>>1]|0;
       $177 = ($172&65535)<($176&65535);
       if (!($177)) {
        $178 = ($172<<16>>16)==($176<<16>>16);
        if (!($178)) {
         $j$1$i11 = $j$03$i9;
         break;
        }
        $179 = (((($s)) + 5208|0) + ($170)|0);
        $180 = HEAP8[$179>>0]|0;
        $181 = (((($s)) + 5208|0) + ($174)|0);
        $182 = HEAP8[$181>>0]|0;
        $183 = ($180&255)>($182&255);
        if ($183) {
         $j$1$i11 = $j$03$i9;
         break;
        }
       }
       $j$1$i11 = $168;
      } else {
       $j$1$i11 = $j$03$i9;
      }
     } while(0);
     $184 = HEAP16[$150>>1]|0;
     $185 = (((($s)) + 2908|0) + ($j$1$i11<<2)|0);
     $186 = HEAP32[$185>>2]|0;
     $187 = (($0) + ($186<<2)|0);
     $188 = HEAP16[$187>>1]|0;
     $189 = ($184&65535)<($188&65535);
     if ($189) {
      $$0$lcssa$i13 = $$02$i10;
      break L55;
     }
     $190 = ($184<<16>>16)==($188<<16>>16);
     if ($190) {
      $191 = HEAP8[$159>>0]|0;
      $192 = (((($s)) + 5208|0) + ($186)|0);
      $193 = HEAP8[$192>>0]|0;
      $194 = ($191&255)>($193&255);
      if (!($194)) {
       $$0$lcssa$i13 = $$02$i10;
       break L55;
      }
     }
     $195 = (((($s)) + 2908|0) + ($$02$i10<<2)|0);
     HEAP32[$195>>2] = $186;
     $j$0$i12 = $j$1$i11 << 1;
     $196 = HEAP32[$6>>2]|0;
     $197 = ($j$0$i12|0)>($196|0);
     if ($197) {
      $$0$lcssa$i13 = $j$1$i11;
      break;
     } else {
      $$02$i10 = $j$1$i11;$167 = $196;$j$03$i9 = $j$0$i12;
     }
    }
   }
  } while(0);
  $198 = (((($s)) + 2908|0) + ($$0$lcssa$i13<<2)|0);
  HEAP32[$198>>2] = $node$0;
  $199 = HEAP32[$6>>2]|0;
  $200 = ($199|0)>(1);
  if ($200) {
   $96 = $199;$node$0 = $163;
  } else {
   break;
  }
 }
 $201 = HEAP32[$52>>2]|0;
 $202 = HEAP32[$7>>2]|0;
 $203 = (($202) + -1)|0;
 HEAP32[$7>>2] = $203;
 $204 = (((($s)) + 2908|0) + ($203<<2)|0);
 HEAP32[$204>>2] = $201;
 $desc$idx$val = HEAP32[$desc>>2]|0;
 $desc$idx2$val = HEAP32[$49>>2]|0;
 $desc$idx3$val = HEAP32[$1>>2]|0;
 $205 = HEAP32[$desc$idx3$val>>2]|0;
 $206 = ((($desc$idx3$val)) + 4|0);
 $207 = HEAP32[$206>>2]|0;
 $208 = ((($desc$idx3$val)) + 8|0);
 $209 = HEAP32[$208>>2]|0;
 $210 = ((($desc$idx3$val)) + 16|0);
 $211 = HEAP32[$210>>2]|0;
 $scevgep$i = ((($s)) + 2876|0);
 dest=$scevgep$i; stop=dest+32|0; do { HEAP16[dest>>1]=0|0; dest=dest+2|0; } while ((dest|0) < (stop|0));
 $212 = HEAP32[$7>>2]|0;
 $213 = (((($s)) + 2908|0) + ($212<<2)|0);
 $214 = HEAP32[$213>>2]|0;
 $215 = (((($desc$idx$val) + ($214<<2)|0)) + 2|0);
 HEAP16[$215>>1] = 0;
 $h$08$i = (($212) + 1)|0;
 $216 = ($h$08$i|0)<(573);
 L71: do {
  if ($216) {
   $217 = ((($s)) + 5800|0);
   $218 = ($205|0)==(0|0);
   $219 = ((($s)) + 5804|0);
   if ($218) {
    $h$010$us$i = $h$08$i;$overflow$09$us$i = 0;
    while(1) {
     $220 = (((($s)) + 2908|0) + ($h$010$us$i<<2)|0);
     $221 = HEAP32[$220>>2]|0;
     $222 = (((($desc$idx$val) + ($221<<2)|0)) + 2|0);
     $223 = HEAP16[$222>>1]|0;
     $224 = $223&65535;
     $225 = (((($desc$idx$val) + ($224<<2)|0)) + 2|0);
     $226 = HEAP16[$225>>1]|0;
     $227 = $226&65535;
     $228 = (($227) + 1)|0;
     $229 = ($227|0)<($211|0);
     $$$us$i = $229 ? $228 : $211;
     $230 = $229&1;
     $231 = $230 ^ 1;
     $overflow$0$$us$i = (($231) + ($overflow$09$us$i))|0;
     $232 = $$$us$i&65535;
     HEAP16[$222>>1] = $232;
     $233 = ($221|0)>($desc$idx2$val|0);
     if (!($233)) {
      $234 = (((($s)) + 2876|0) + ($$$us$i<<1)|0);
      $235 = HEAP16[$234>>1]|0;
      $236 = (($235) + 1)<<16>>16;
      HEAP16[$234>>1] = $236;
      $237 = ($221|0)<($209|0);
      if ($237) {
       $xbits$0$us$i = 0;
      } else {
       $238 = (($221) - ($209))|0;
       $239 = (($207) + ($238<<2)|0);
       $240 = HEAP32[$239>>2]|0;
       $xbits$0$us$i = $240;
      }
      $241 = (($desc$idx$val) + ($221<<2)|0);
      $242 = HEAP16[$241>>1]|0;
      $243 = $242&65535;
      $244 = (($xbits$0$us$i) + ($$$us$i))|0;
      $245 = Math_imul($243, $244)|0;
      $246 = HEAP32[$217>>2]|0;
      $247 = (($245) + ($246))|0;
      HEAP32[$217>>2] = $247;
     }
     $h$0$us$i = (($h$010$us$i) + 1)|0;
     $exitcond$i5 = ($h$0$us$i|0)==(573);
     if ($exitcond$i5) {
      $overflow$0$lcssa$i = $overflow$0$$us$i;
      break;
     } else {
      $h$010$us$i = $h$0$us$i;$overflow$09$us$i = $overflow$0$$us$i;
     }
    }
   } else {
    $h$010$i = $h$08$i;$overflow$09$i = 0;
    while(1) {
     $248 = (((($s)) + 2908|0) + ($h$010$i<<2)|0);
     $249 = HEAP32[$248>>2]|0;
     $250 = (((($desc$idx$val) + ($249<<2)|0)) + 2|0);
     $251 = HEAP16[$250>>1]|0;
     $252 = $251&65535;
     $253 = (((($desc$idx$val) + ($252<<2)|0)) + 2|0);
     $254 = HEAP16[$253>>1]|0;
     $255 = $254&65535;
     $256 = (($255) + 1)|0;
     $257 = ($255|0)<($211|0);
     $$$i = $257 ? $256 : $211;
     $258 = $257&1;
     $259 = $258 ^ 1;
     $overflow$0$$i = (($259) + ($overflow$09$i))|0;
     $260 = $$$i&65535;
     HEAP16[$250>>1] = $260;
     $261 = ($249|0)>($desc$idx2$val|0);
     if (!($261)) {
      $262 = (((($s)) + 2876|0) + ($$$i<<1)|0);
      $263 = HEAP16[$262>>1]|0;
      $264 = (($263) + 1)<<16>>16;
      HEAP16[$262>>1] = $264;
      $265 = ($249|0)<($209|0);
      if ($265) {
       $xbits$0$i = 0;
      } else {
       $266 = (($249) - ($209))|0;
       $267 = (($207) + ($266<<2)|0);
       $268 = HEAP32[$267>>2]|0;
       $xbits$0$i = $268;
      }
      $269 = (($desc$idx$val) + ($249<<2)|0);
      $270 = HEAP16[$269>>1]|0;
      $271 = $270&65535;
      $272 = (($xbits$0$i) + ($$$i))|0;
      $273 = Math_imul($271, $272)|0;
      $274 = HEAP32[$217>>2]|0;
      $275 = (($273) + ($274))|0;
      HEAP32[$217>>2] = $275;
      $276 = (((($205) + ($249<<2)|0)) + 2|0);
      $277 = HEAP16[$276>>1]|0;
      $278 = $277&65535;
      $279 = (($278) + ($xbits$0$i))|0;
      $280 = Math_imul($279, $271)|0;
      $281 = HEAP32[$219>>2]|0;
      $282 = (($280) + ($281))|0;
      HEAP32[$219>>2] = $282;
     }
     $h$0$i = (($h$010$i) + 1)|0;
     $exitcond20$i = ($h$0$i|0)==(573);
     if ($exitcond20$i) {
      $overflow$0$lcssa$i = $overflow$0$$i;
      break;
     } else {
      $h$010$i = $h$0$i;$overflow$09$i = $overflow$0$$i;
     }
    }
   }
   $283 = ($overflow$0$lcssa$i|0)==(0);
   if (!($283)) {
    $284 = (((($s)) + 2876|0) + ($211<<1)|0);
    $overflow$2$i = $overflow$0$lcssa$i;
    while(1) {
     $bits$2$in$i = $211;
     while(1) {
      $bits$2$i = (($bits$2$in$i) + -1)|0;
      $285 = (((($s)) + 2876|0) + ($bits$2$i<<1)|0);
      $286 = HEAP16[$285>>1]|0;
      $287 = ($286<<16>>16)==(0);
      if ($287) {
       $bits$2$in$i = $bits$2$i;
      } else {
       $$lcssa83 = $285;$$lcssa84 = $286;$bits$2$in$i$lcssa = $bits$2$in$i;
       break;
      }
     }
     $288 = (($$lcssa84) + -1)<<16>>16;
     HEAP16[$$lcssa83>>1] = $288;
     $289 = (((($s)) + 2876|0) + ($bits$2$in$i$lcssa<<1)|0);
     $290 = HEAP16[$289>>1]|0;
     $291 = $290&65535;
     $292 = (($291) + 2)|0;
     $293 = $292&65535;
     HEAP16[$289>>1] = $293;
     $294 = HEAP16[$284>>1]|0;
     $295 = (($294) + -1)<<16>>16;
     HEAP16[$284>>1] = $295;
     $296 = (($overflow$2$i) + -2)|0;
     $297 = ($overflow$2$i|0)>(2);
     if ($297) {
      $overflow$2$i = $296;
     } else {
      $$lcssa85 = $295;
      break;
     }
    }
    $298 = ($211|0)==(0);
    if (!($298)) {
     $301 = $$lcssa85;$bits$36$i = $211;$h$17$i = 573;
     while(1) {
      $299 = $bits$36$i&65535;
      $300 = ($301<<16>>16)==(0);
      if ($300) {
       $h$2$ph$i$lcssa = $h$17$i;
      } else {
       $302 = $301&65535;
       $h$2$ph$i33 = $h$17$i;$n$0$ph$i32 = $302;
       while(1) {
        $h$2$i = $h$2$ph$i33;
        while(1) {
         $303 = (($h$2$i) + -1)|0;
         $304 = (((($s)) + 2908|0) + ($303<<2)|0);
         $305 = HEAP32[$304>>2]|0;
         $306 = ($305|0)>($desc$idx2$val|0);
         if ($306) {
          $h$2$i = $303;
         } else {
          $$lcssa81 = $303;$$lcssa82 = $305;
          break;
         }
        }
        $307 = (((($desc$idx$val) + ($$lcssa82<<2)|0)) + 2|0);
        $308 = HEAP16[$307>>1]|0;
        $309 = $308&65535;
        $310 = ($bits$36$i|0)==($309|0);
        if (!($310)) {
         $311 = (($bits$36$i) - ($309))|0;
         $312 = (($desc$idx$val) + ($$lcssa82<<2)|0);
         $313 = HEAP16[$312>>1]|0;
         $314 = $313&65535;
         $315 = Math_imul($314, $311)|0;
         $316 = HEAP32[$217>>2]|0;
         $317 = (($315) + ($316))|0;
         HEAP32[$217>>2] = $317;
         HEAP16[$307>>1] = $299;
        }
        $318 = (($n$0$ph$i32) + -1)|0;
        $319 = ($318|0)==(0);
        if ($319) {
         $h$2$ph$i$lcssa = $$lcssa81;
         break;
        } else {
         $h$2$ph$i33 = $$lcssa81;$n$0$ph$i32 = $318;
        }
       }
      }
      $320 = (($bits$36$i) + -1)|0;
      $321 = ($320|0)==(0);
      if ($321) {
       break L71;
      }
      $$phi$trans$insert$i = (((($s)) + 2876|0) + ($320<<1)|0);
      $$pre$i = HEAP16[$$phi$trans$insert$i>>1]|0;
      $301 = $$pre$i;$bits$36$i = $320;$h$17$i = $h$2$ph$i$lcssa;
     }
    }
   }
  }
 } while(0);
 $bits$03$i = 1;$code$04$i = 0;
 while(1) {
  $323 = $code$04$i & 65534;
  $324 = (($bits$03$i) + -1)|0;
  $325 = (((($s)) + 2876|0) + ($324<<1)|0);
  $326 = HEAP16[$325>>1]|0;
  $327 = $326&65535;
  $328 = (($327) + ($323))|0;
  $329 = $328 << 1;
  $330 = $329&65535;
  $331 = (($next_code$i) + ($bits$03$i<<1)|0);
  HEAP16[$331>>1] = $330;
  $332 = (($bits$03$i) + 1)|0;
  $exitcond5$i = ($332|0)==(16);
  if ($exitcond5$i) {
   break;
  } else {
   $bits$03$i = $332;$code$04$i = $329;
  }
 }
 $322 = ($max_code$2$lcssa|0)<(0);
 if ($322) {
  STACKTOP = sp;return;
 } else {
  $n$01$i = 0;
 }
 while(1) {
  $333 = (((($0) + ($n$01$i<<2)|0)) + 2|0);
  $334 = HEAP16[$333>>1]|0;
  $335 = $334&65535;
  $336 = ($334<<16>>16)==(0);
  if (!($336)) {
   $337 = (($next_code$i) + ($335<<1)|0);
   $338 = HEAP16[$337>>1]|0;
   $339 = (($338) + 1)<<16>>16;
   HEAP16[$337>>1] = $339;
   $340 = $338&65535;
   $$0$i$i = $335;$$01$i$i = $340;$res$0$i$i = 0;
   while(1) {
    $341 = $$01$i$i & 1;
    $342 = $res$0$i$i | $341;
    $343 = $$01$i$i >>> 1;
    $344 = $342 << 1;
    $345 = (($$0$i$i) + -1)|0;
    $346 = ($$0$i$i|0)>(1);
    if ($346) {
     $$0$i$i = $345;$$01$i$i = $343;$res$0$i$i = $344;
    } else {
     $$lcssa = $342;
     break;
    }
   }
   $347 = $$lcssa&65535;
   $348 = (($0) + ($n$01$i<<2)|0);
   HEAP16[$348>>1] = $347;
  }
  $349 = (($n$01$i) + 1)|0;
  $exitcond$i = ($n$01$i|0)==($max_code$2$lcssa|0);
  if ($exitcond$i) {
   break;
  } else {
   $n$01$i = $349;
  }
 }
 STACKTOP = sp;return;
}
function _compress_block($s,$ltree,$dtree) {
 $s = $s|0;
 $ltree = $ltree|0;
 $dtree = $dtree|0;
 var $$phi$trans$insert = 0, $$phi$trans$insert6 = 0, $$pn = 0, $$pre = 0, $$pre$phi8Z2D = 0, $$pre$phiZ2D = 0, $$pre7 = 0, $$sink = 0, $$sink$in = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0;
 var $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0;
 var $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0;
 var $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $lx$0 = 0, $storemerge = 0, $storemerge3 = 0, $storemerge4 = 0, $storemerge5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($s)) + 5792|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  $$phi$trans$insert = ((($s)) + 5820|0);
  $$pre = HEAP32[$$phi$trans$insert>>2]|0;
  $$phi$trans$insert6 = ((($s)) + 5816|0);
  $$pre7 = HEAP16[$$phi$trans$insert6>>1]|0;
  $$pre$phi8Z2D = $$phi$trans$insert6;$$pre$phiZ2D = $$phi$trans$insert;$205 = $$pre;$211 = $$pre7;
 } else {
  $3 = ((($s)) + 5796|0);
  $4 = ((($s)) + 5784|0);
  $5 = ((($s)) + 5820|0);
  $6 = ((($s)) + 5816|0);
  $7 = ((($s)) + 20|0);
  $8 = ((($s)) + 8|0);
  $lx$0 = 0;
  while(1) {
   $9 = HEAP32[$3>>2]|0;
   $10 = (($9) + ($lx$0<<1)|0);
   $11 = HEAP16[$10>>1]|0;
   $12 = $11&65535;
   $13 = (($lx$0) + 1)|0;
   $14 = HEAP32[$4>>2]|0;
   $15 = (($14) + ($lx$0)|0);
   $16 = HEAP8[$15>>0]|0;
   $17 = $16&255;
   $18 = ($11<<16>>16)==(0);
   if ($18) {
    $19 = (((($ltree) + ($17<<2)|0)) + 2|0);
    $20 = HEAP16[$19>>1]|0;
    $21 = $20&65535;
    $22 = HEAP32[$5>>2]|0;
    $23 = (16 - ($21))|0;
    $24 = ($22|0)>($23|0);
    $25 = (($ltree) + ($17<<2)|0);
    $26 = HEAP16[$25>>1]|0;
    $27 = $26&65535;
    $28 = $27 << $22;
    $29 = HEAP16[$6>>1]|0;
    $30 = $29&65535;
    $31 = $30 | $28;
    $32 = $31&65535;
    HEAP16[$6>>1] = $32;
    if ($24) {
     $33 = $31&255;
     $34 = HEAP32[$7>>2]|0;
     $35 = (($34) + 1)|0;
     HEAP32[$7>>2] = $35;
     $36 = HEAP32[$8>>2]|0;
     $37 = (($36) + ($34)|0);
     HEAP8[$37>>0] = $33;
     $38 = HEAP16[$6>>1]|0;
     $39 = ($38&65535) >>> 8;
     $40 = $39&255;
     $41 = HEAP32[$7>>2]|0;
     $42 = (($41) + 1)|0;
     HEAP32[$7>>2] = $42;
     $43 = HEAP32[$8>>2]|0;
     $44 = (($43) + ($41)|0);
     HEAP8[$44>>0] = $40;
     $45 = HEAP32[$5>>2]|0;
     $46 = (16 - ($45))|0;
     $47 = $27 >>> $46;
     $48 = $47&65535;
     HEAP16[$6>>1] = $48;
     $49 = (($21) + -16)|0;
     $50 = (($49) + ($45))|0;
     $235 = $48;$storemerge5 = $50;
    } else {
     $51 = (($22) + ($21))|0;
     $235 = $32;$storemerge5 = $51;
    }
    HEAP32[$5>>2] = $storemerge5;
    $236 = $235;$237 = $storemerge5;
   } else {
    $52 = (613302 + ($17)|0);
    $53 = HEAP8[$52>>0]|0;
    $54 = $53&255;
    $55 = $54 | 256;
    $56 = (($55) + 1)|0;
    $57 = (((($ltree) + ($56<<2)|0)) + 2|0);
    $58 = HEAP16[$57>>1]|0;
    $59 = $58&65535;
    $60 = HEAP32[$5>>2]|0;
    $61 = (16 - ($59))|0;
    $62 = ($60|0)>($61|0);
    $63 = (($ltree) + ($56<<2)|0);
    $64 = HEAP16[$63>>1]|0;
    $65 = $64&65535;
    $66 = $65 << $60;
    $67 = HEAP16[$6>>1]|0;
    $68 = $67&65535;
    $69 = $68 | $66;
    $70 = $69&65535;
    HEAP16[$6>>1] = $70;
    if ($62) {
     $71 = $69&255;
     $72 = HEAP32[$7>>2]|0;
     $73 = (($72) + 1)|0;
     HEAP32[$7>>2] = $73;
     $74 = HEAP32[$8>>2]|0;
     $75 = (($74) + ($72)|0);
     HEAP8[$75>>0] = $71;
     $76 = HEAP16[$6>>1]|0;
     $77 = ($76&65535) >>> 8;
     $78 = $77&255;
     $79 = HEAP32[$7>>2]|0;
     $80 = (($79) + 1)|0;
     HEAP32[$7>>2] = $80;
     $81 = HEAP32[$8>>2]|0;
     $82 = (($81) + ($79)|0);
     HEAP8[$82>>0] = $78;
     $83 = HEAP32[$5>>2]|0;
     $84 = (16 - ($83))|0;
     $85 = $65 >>> $84;
     $86 = $85&65535;
     HEAP16[$6>>1] = $86;
     $87 = (($59) + -16)|0;
     $88 = (($87) + ($83))|0;
     $103 = $86;$90 = $88;
    } else {
     $89 = (($60) + ($59))|0;
     $103 = $70;$90 = $89;
    }
    HEAP32[$5>>2] = $90;
    $91 = (33388 + ($54<<2)|0);
    $92 = HEAP32[$91>>2]|0;
    $93 = (($54) + -8)|0;
    $94 = ($93>>>0)<(20);
    if ($94) {
     $95 = (141504 + ($54<<2)|0);
     $96 = HEAP32[$95>>2]|0;
     $97 = (($17) - ($96))|0;
     $98 = (16 - ($92))|0;
     $99 = ($90|0)>($98|0);
     $100 = $97 & 65535;
     $101 = $100 << $90;
     $102 = $103&65535;
     $104 = $101 | $102;
     $105 = $104&65535;
     HEAP16[$6>>1] = $105;
     if ($99) {
      $106 = $104&255;
      $107 = HEAP32[$7>>2]|0;
      $108 = (($107) + 1)|0;
      HEAP32[$7>>2] = $108;
      $109 = HEAP32[$8>>2]|0;
      $110 = (($109) + ($107)|0);
      HEAP8[$110>>0] = $106;
      $111 = HEAP16[$6>>1]|0;
      $112 = ($111&65535) >>> 8;
      $113 = $112&255;
      $114 = HEAP32[$7>>2]|0;
      $115 = (($114) + 1)|0;
      HEAP32[$7>>2] = $115;
      $116 = HEAP32[$8>>2]|0;
      $117 = (($116) + ($114)|0);
      HEAP8[$117>>0] = $113;
      $118 = HEAP32[$5>>2]|0;
      $119 = (16 - ($118))|0;
      $120 = $100 >>> $119;
      $121 = $120&65535;
      HEAP16[$6>>1] = $121;
      $122 = (($92) + -16)|0;
      $123 = (($122) + ($118))|0;
      $238 = $121;$storemerge4 = $123;
     } else {
      $124 = (($90) + ($92))|0;
      $238 = $105;$storemerge4 = $124;
     }
     HEAP32[$5>>2] = $storemerge4;
     $135 = $storemerge4;$141 = $238;
    } else {
     $135 = $90;$141 = $103;
    }
    $125 = (($12) + -1)|0;
    $126 = ($125>>>0)<(256);
    $127 = $125 >>> 7;
    $128 = (($127) + 256)|0;
    $$pn = $126 ? $125 : $128;
    $$sink$in = (612790 + ($$pn)|0);
    $$sink = HEAP8[$$sink$in>>0]|0;
    $129 = $$sink&255;
    $130 = (((($dtree) + ($129<<2)|0)) + 2|0);
    $131 = HEAP16[$130>>1]|0;
    $132 = $131&65535;
    $133 = (16 - ($132))|0;
    $134 = ($135|0)>($133|0);
    $136 = (($dtree) + ($129<<2)|0);
    $137 = HEAP16[$136>>1]|0;
    $138 = $137&65535;
    $139 = $138 << $135;
    $140 = $141&65535;
    $142 = $140 | $139;
    $143 = $142&65535;
    HEAP16[$6>>1] = $143;
    if ($134) {
     $144 = $142&255;
     $145 = HEAP32[$7>>2]|0;
     $146 = (($145) + 1)|0;
     HEAP32[$7>>2] = $146;
     $147 = HEAP32[$8>>2]|0;
     $148 = (($147) + ($145)|0);
     HEAP8[$148>>0] = $144;
     $149 = HEAP16[$6>>1]|0;
     $150 = ($149&65535) >>> 8;
     $151 = $150&255;
     $152 = HEAP32[$7>>2]|0;
     $153 = (($152) + 1)|0;
     HEAP32[$7>>2] = $153;
     $154 = HEAP32[$8>>2]|0;
     $155 = (($154) + ($152)|0);
     HEAP8[$155>>0] = $151;
     $156 = HEAP32[$5>>2]|0;
     $157 = (16 - ($156))|0;
     $158 = $138 >>> $157;
     $159 = $158&65535;
     HEAP16[$6>>1] = $159;
     $160 = (($132) + -16)|0;
     $161 = (($160) + ($156))|0;
     $163 = $161;$176 = $159;
    } else {
     $162 = (($135) + ($132))|0;
     $163 = $162;$176 = $143;
    }
    HEAP32[$5>>2] = $163;
    $164 = (141288 + ($129<<2)|0);
    $165 = HEAP32[$164>>2]|0;
    $166 = (($129) + -4)|0;
    $167 = ($166>>>0)<(26);
    if ($167) {
     $168 = (141620 + ($129<<2)|0);
     $169 = HEAP32[$168>>2]|0;
     $170 = (($125) - ($169))|0;
     $171 = (16 - ($165))|0;
     $172 = ($163|0)>($171|0);
     $173 = $170 & 65535;
     $174 = $173 << $163;
     $175 = $176&65535;
     $177 = $174 | $175;
     $178 = $177&65535;
     HEAP16[$6>>1] = $178;
     if ($172) {
      $179 = $177&255;
      $180 = HEAP32[$7>>2]|0;
      $181 = (($180) + 1)|0;
      HEAP32[$7>>2] = $181;
      $182 = HEAP32[$8>>2]|0;
      $183 = (($182) + ($180)|0);
      HEAP8[$183>>0] = $179;
      $184 = HEAP16[$6>>1]|0;
      $185 = ($184&65535) >>> 8;
      $186 = $185&255;
      $187 = HEAP32[$7>>2]|0;
      $188 = (($187) + 1)|0;
      HEAP32[$7>>2] = $188;
      $189 = HEAP32[$8>>2]|0;
      $190 = (($189) + ($187)|0);
      HEAP8[$190>>0] = $186;
      $191 = HEAP32[$5>>2]|0;
      $192 = (16 - ($191))|0;
      $193 = $173 >>> $192;
      $194 = $193&65535;
      HEAP16[$6>>1] = $194;
      $195 = (($165) + -16)|0;
      $196 = (($195) + ($191))|0;
      $239 = $194;$storemerge3 = $196;
     } else {
      $197 = (($163) + ($165))|0;
      $239 = $178;$storemerge3 = $197;
     }
     HEAP32[$5>>2] = $storemerge3;
     $236 = $239;$237 = $storemerge3;
    } else {
     $236 = $176;$237 = $163;
    }
   }
   $198 = HEAP32[$0>>2]|0;
   $199 = ($13>>>0)<($198>>>0);
   if ($199) {
    $lx$0 = $13;
   } else {
    $$pre$phi8Z2D = $6;$$pre$phiZ2D = $5;$205 = $237;$211 = $236;
    break;
   }
  }
 }
 $200 = ((($ltree)) + 1026|0);
 $201 = HEAP16[$200>>1]|0;
 $202 = $201&65535;
 $203 = (16 - ($202))|0;
 $204 = ($205|0)>($203|0);
 $206 = ((($ltree)) + 1024|0);
 $207 = HEAP16[$206>>1]|0;
 $208 = $207&65535;
 $209 = $208 << $205;
 $210 = $211&65535;
 $212 = $210 | $209;
 $213 = $212&65535;
 HEAP16[$$pre$phi8Z2D>>1] = $213;
 if ($204) {
  $214 = $212&255;
  $215 = ((($s)) + 20|0);
  $216 = HEAP32[$215>>2]|0;
  $217 = (($216) + 1)|0;
  HEAP32[$215>>2] = $217;
  $218 = ((($s)) + 8|0);
  $219 = HEAP32[$218>>2]|0;
  $220 = (($219) + ($216)|0);
  HEAP8[$220>>0] = $214;
  $221 = HEAP16[$$pre$phi8Z2D>>1]|0;
  $222 = ($221&65535) >>> 8;
  $223 = $222&255;
  $224 = HEAP32[$215>>2]|0;
  $225 = (($224) + 1)|0;
  HEAP32[$215>>2] = $225;
  $226 = HEAP32[$218>>2]|0;
  $227 = (($226) + ($224)|0);
  HEAP8[$227>>0] = $223;
  $228 = HEAP32[$$pre$phiZ2D>>2]|0;
  $229 = (16 - ($228))|0;
  $230 = $208 >>> $229;
  $231 = $230&65535;
  HEAP16[$$pre$phi8Z2D>>1] = $231;
  $232 = (($202) + -16)|0;
  $233 = (($232) + ($228))|0;
  $storemerge = $233;
  HEAP32[$$pre$phiZ2D>>2] = $storemerge;
  return;
 } else {
  $234 = (($205) + ($202))|0;
  $storemerge = $234;
  HEAP32[$$pre$phiZ2D>>2] = $storemerge;
  return;
 }
}
function _send_tree($s,$tree,$max_code) {
 $s = $s|0;
 $tree = $tree|0;
 $max_code = $max_code|0;
 var $$ = 0, $$10 = 0, $$8 = 0, $$lcssa = 0, $$lcssa42 = 0, $$lcssa43 = 0, $$lcssa44 = 0, $$lcssa45 = 0, $$pre = 0, $$pre23 = 0, $$pre24 = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0;
 var $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0;
 var $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0;
 var $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0;
 var $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0;
 var $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $count$0 = 0, $count$0$lcssa41 = 0, $count$1 = 0, $count$2 = 0, $max_count$1$ph = 0, $max_count$1$ph$be = 0, $min_count$1$ph = 0, $min_count$1$ph$be = 0, $n$0 = 0, $n$0$ph = 0, $nextlen$0$ph = 0, $or$cond = 0, $prevlen$0$ph = 0, $prevlen$0$ph$phi = 0, $storemerge1 = 0, $storemerge3 = 0;
 var $storemerge5 = 0, $storemerge6 = 0, $storemerge7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($tree)) + 2|0);
 $1 = HEAP16[$0>>1]|0;
 $2 = $1&65535;
 $3 = ($1<<16>>16)==(0);
 $$ = $3 ? 3 : 4;
 $$8 = $3 ? 138 : 7;
 $4 = ((($s)) + 2754|0);
 $5 = ((($s)) + 5820|0);
 $6 = ((($s)) + 2752|0);
 $7 = ((($s)) + 5816|0);
 $8 = ((($s)) + 20|0);
 $9 = ((($s)) + 8|0);
 $10 = ((($s)) + 2758|0);
 $11 = ((($s)) + 2756|0);
 $12 = ((($s)) + 2750|0);
 $13 = ((($s)) + 2748|0);
 $max_count$1$ph = $$8;$min_count$1$ph = $$;$n$0$ph = 0;$nextlen$0$ph = $2;$prevlen$0$ph = -1;
 L1: while(1) {
  $count$0 = 0;$n$0 = $n$0$ph;
  while(1) {
   $14 = ($n$0|0)>($max_code|0);
   if ($14) {
    break L1;
   }
   $15 = (($n$0) + 1)|0;
   $16 = (((($tree) + ($15<<2)|0)) + 2|0);
   $17 = HEAP16[$16>>1]|0;
   $18 = $17&65535;
   $19 = (($count$0) + 1)|0;
   $20 = ($19|0)<($max_count$1$ph|0);
   $21 = ($nextlen$0$ph|0)==($18|0);
   $or$cond = $20 & $21;
   if ($or$cond) {
    $count$0 = $19;$n$0 = $15;
   } else {
    $$lcssa = $15;$$lcssa42 = $17;$$lcssa43 = $18;$$lcssa44 = $19;$$lcssa45 = $21;$count$0$lcssa41 = $count$0;
    break;
   }
  }
  $22 = ($$lcssa44|0)<($min_count$1$ph|0);
  do {
   if ($22) {
    $23 = (((((($s)) + 2684|0) + ($nextlen$0$ph<<2)|0)) + 2|0);
    $24 = (((($s)) + 2684|0) + ($nextlen$0$ph<<2)|0);
    $$pre24 = HEAP32[$5>>2]|0;
    $$pre25 = HEAP16[$7>>1]|0;
    $29 = $$pre24;$34 = $$pre25;$count$1 = $$lcssa44;
    while(1) {
     $25 = HEAP16[$23>>1]|0;
     $26 = $25&65535;
     $27 = (16 - ($26))|0;
     $28 = ($29|0)>($27|0);
     $30 = HEAP16[$24>>1]|0;
     $31 = $30&65535;
     $32 = $31 << $29;
     $33 = $34&65535;
     $35 = $33 | $32;
     $36 = $35&65535;
     HEAP16[$7>>1] = $36;
     if ($28) {
      $37 = $35&255;
      $38 = HEAP32[$8>>2]|0;
      $39 = (($38) + 1)|0;
      HEAP32[$8>>2] = $39;
      $40 = HEAP32[$9>>2]|0;
      $41 = (($40) + ($38)|0);
      HEAP8[$41>>0] = $37;
      $42 = HEAP16[$7>>1]|0;
      $43 = ($42&65535) >>> 8;
      $44 = $43&255;
      $45 = HEAP32[$8>>2]|0;
      $46 = (($45) + 1)|0;
      HEAP32[$8>>2] = $46;
      $47 = HEAP32[$9>>2]|0;
      $48 = (($47) + ($45)|0);
      HEAP8[$48>>0] = $44;
      $49 = HEAP32[$5>>2]|0;
      $50 = (16 - ($49))|0;
      $51 = $31 >>> $50;
      $52 = $51&65535;
      HEAP16[$7>>1] = $52;
      $53 = (($26) + -16)|0;
      $54 = (($53) + ($49))|0;
      $270 = $52;$storemerge7 = $54;
     } else {
      $55 = (($29) + ($26))|0;
      $270 = $36;$storemerge7 = $55;
     }
     HEAP32[$5>>2] = $storemerge7;
     $56 = (($count$1) + -1)|0;
     $57 = ($56|0)==(0);
     if ($57) {
      break;
     } else {
      $29 = $storemerge7;$34 = $270;$count$1 = $56;
     }
    }
   } else {
    $58 = ($nextlen$0$ph|0)==(0);
    if (!($58)) {
     $59 = ($nextlen$0$ph|0)==($prevlen$0$ph|0);
     if ($59) {
      $$pre = HEAP32[$5>>2]|0;
      $$pre23 = HEAP16[$7>>1]|0;
      $102 = $$pre23;$97 = $$pre;$count$2 = $$lcssa44;
     } else {
      $60 = (((((($s)) + 2684|0) + ($nextlen$0$ph<<2)|0)) + 2|0);
      $61 = HEAP16[$60>>1]|0;
      $62 = $61&65535;
      $63 = HEAP32[$5>>2]|0;
      $64 = (16 - ($62))|0;
      $65 = ($63|0)>($64|0);
      $66 = (((($s)) + 2684|0) + ($nextlen$0$ph<<2)|0);
      $67 = HEAP16[$66>>1]|0;
      $68 = $67&65535;
      $69 = $68 << $63;
      $70 = HEAP16[$7>>1]|0;
      $71 = $70&65535;
      $72 = $71 | $69;
      $73 = $72&65535;
      HEAP16[$7>>1] = $73;
      if ($65) {
       $74 = $72&255;
       $75 = HEAP32[$8>>2]|0;
       $76 = (($75) + 1)|0;
       HEAP32[$8>>2] = $76;
       $77 = HEAP32[$9>>2]|0;
       $78 = (($77) + ($75)|0);
       HEAP8[$78>>0] = $74;
       $79 = HEAP16[$7>>1]|0;
       $80 = ($79&65535) >>> 8;
       $81 = $80&255;
       $82 = HEAP32[$8>>2]|0;
       $83 = (($82) + 1)|0;
       HEAP32[$8>>2] = $83;
       $84 = HEAP32[$9>>2]|0;
       $85 = (($84) + ($82)|0);
       HEAP8[$85>>0] = $81;
       $86 = HEAP32[$5>>2]|0;
       $87 = (16 - ($86))|0;
       $88 = $68 >>> $87;
       $89 = $88&65535;
       HEAP16[$7>>1] = $89;
       $90 = (($62) + -16)|0;
       $91 = (($90) + ($86))|0;
       $271 = $89;$storemerge6 = $91;
      } else {
       $92 = (($63) + ($62))|0;
       $271 = $73;$storemerge6 = $92;
      }
      HEAP32[$5>>2] = $storemerge6;
      $102 = $271;$97 = $storemerge6;$count$2 = $count$0$lcssa41;
     }
     $93 = HEAP16[$12>>1]|0;
     $94 = $93&65535;
     $95 = (16 - ($94))|0;
     $96 = ($97|0)>($95|0);
     $98 = HEAP16[$13>>1]|0;
     $99 = $98&65535;
     $100 = $99 << $97;
     $101 = $102&65535;
     $103 = $101 | $100;
     $104 = $103&65535;
     HEAP16[$7>>1] = $104;
     if ($96) {
      $105 = $103&255;
      $106 = HEAP32[$8>>2]|0;
      $107 = (($106) + 1)|0;
      HEAP32[$8>>2] = $107;
      $108 = HEAP32[$9>>2]|0;
      $109 = (($108) + ($106)|0);
      HEAP8[$109>>0] = $105;
      $110 = HEAP16[$7>>1]|0;
      $111 = ($110&65535) >>> 8;
      $112 = $111&255;
      $113 = HEAP32[$8>>2]|0;
      $114 = (($113) + 1)|0;
      HEAP32[$8>>2] = $114;
      $115 = HEAP32[$9>>2]|0;
      $116 = (($115) + ($113)|0);
      HEAP8[$116>>0] = $112;
      $117 = HEAP32[$5>>2]|0;
      $118 = (16 - ($117))|0;
      $119 = $99 >>> $118;
      $120 = $119&65535;
      HEAP16[$7>>1] = $120;
      $121 = (($94) + -16)|0;
      $122 = (($121) + ($117))|0;
      $124 = $122;$130 = $119;
     } else {
      $123 = (($97) + ($94))|0;
      $124 = $123;$130 = $103;
     }
     HEAP32[$5>>2] = $124;
     $125 = ($124|0)>(14);
     $126 = (($count$2) + 65533)|0;
     $127 = $126 & 65535;
     $128 = $127 << $124;
     $129 = $130 & 65535;
     $131 = $129 | $128;
     $132 = $131&65535;
     HEAP16[$7>>1] = $132;
     if ($125) {
      $133 = $131&255;
      $134 = HEAP32[$8>>2]|0;
      $135 = (($134) + 1)|0;
      HEAP32[$8>>2] = $135;
      $136 = HEAP32[$9>>2]|0;
      $137 = (($136) + ($134)|0);
      HEAP8[$137>>0] = $133;
      $138 = HEAP16[$7>>1]|0;
      $139 = ($138&65535) >>> 8;
      $140 = $139&255;
      $141 = HEAP32[$8>>2]|0;
      $142 = (($141) + 1)|0;
      HEAP32[$8>>2] = $142;
      $143 = HEAP32[$9>>2]|0;
      $144 = (($143) + ($141)|0);
      HEAP8[$144>>0] = $140;
      $145 = HEAP32[$5>>2]|0;
      $146 = (16 - ($145))|0;
      $147 = $127 >>> $146;
      $148 = $147&65535;
      HEAP16[$7>>1] = $148;
      $149 = (($145) + -14)|0;
      $storemerge5 = $149;
     } else {
      $150 = (($124) + 2)|0;
      $storemerge5 = $150;
     }
     HEAP32[$5>>2] = $storemerge5;
     break;
    }
    $151 = ($$lcssa44|0)<(11);
    if ($151) {
     $152 = HEAP16[$4>>1]|0;
     $153 = $152&65535;
     $154 = HEAP32[$5>>2]|0;
     $155 = (16 - ($153))|0;
     $156 = ($154|0)>($155|0);
     $157 = HEAP16[$6>>1]|0;
     $158 = $157&65535;
     $159 = $158 << $154;
     $160 = HEAP16[$7>>1]|0;
     $161 = $160&65535;
     $162 = $161 | $159;
     $163 = $162&65535;
     HEAP16[$7>>1] = $163;
     if ($156) {
      $164 = $162&255;
      $165 = HEAP32[$8>>2]|0;
      $166 = (($165) + 1)|0;
      HEAP32[$8>>2] = $166;
      $167 = HEAP32[$9>>2]|0;
      $168 = (($167) + ($165)|0);
      HEAP8[$168>>0] = $164;
      $169 = HEAP16[$7>>1]|0;
      $170 = ($169&65535) >>> 8;
      $171 = $170&255;
      $172 = HEAP32[$8>>2]|0;
      $173 = (($172) + 1)|0;
      HEAP32[$8>>2] = $173;
      $174 = HEAP32[$9>>2]|0;
      $175 = (($174) + ($172)|0);
      HEAP8[$175>>0] = $171;
      $176 = HEAP32[$5>>2]|0;
      $177 = (16 - ($176))|0;
      $178 = $158 >>> $177;
      $179 = $178&65535;
      HEAP16[$7>>1] = $179;
      $180 = (($153) + -16)|0;
      $181 = (($180) + ($176))|0;
      $183 = $181;$189 = $178;
     } else {
      $182 = (($154) + ($153))|0;
      $183 = $182;$189 = $162;
     }
     HEAP32[$5>>2] = $183;
     $184 = ($183|0)>(13);
     $185 = (($count$0$lcssa41) + 65534)|0;
     $186 = $185 & 65535;
     $187 = $186 << $183;
     $188 = $189 & 65535;
     $190 = $188 | $187;
     $191 = $190&65535;
     HEAP16[$7>>1] = $191;
     if ($184) {
      $192 = $190&255;
      $193 = HEAP32[$8>>2]|0;
      $194 = (($193) + 1)|0;
      HEAP32[$8>>2] = $194;
      $195 = HEAP32[$9>>2]|0;
      $196 = (($195) + ($193)|0);
      HEAP8[$196>>0] = $192;
      $197 = HEAP16[$7>>1]|0;
      $198 = ($197&65535) >>> 8;
      $199 = $198&255;
      $200 = HEAP32[$8>>2]|0;
      $201 = (($200) + 1)|0;
      HEAP32[$8>>2] = $201;
      $202 = HEAP32[$9>>2]|0;
      $203 = (($202) + ($200)|0);
      HEAP8[$203>>0] = $199;
      $204 = HEAP32[$5>>2]|0;
      $205 = (16 - ($204))|0;
      $206 = $186 >>> $205;
      $207 = $206&65535;
      HEAP16[$7>>1] = $207;
      $208 = (($204) + -13)|0;
      $storemerge3 = $208;
     } else {
      $209 = (($183) + 3)|0;
      $storemerge3 = $209;
     }
     HEAP32[$5>>2] = $storemerge3;
     break;
    } else {
     $210 = HEAP16[$10>>1]|0;
     $211 = $210&65535;
     $212 = HEAP32[$5>>2]|0;
     $213 = (16 - ($211))|0;
     $214 = ($212|0)>($213|0);
     $215 = HEAP16[$11>>1]|0;
     $216 = $215&65535;
     $217 = $216 << $212;
     $218 = HEAP16[$7>>1]|0;
     $219 = $218&65535;
     $220 = $219 | $217;
     $221 = $220&65535;
     HEAP16[$7>>1] = $221;
     if ($214) {
      $222 = $220&255;
      $223 = HEAP32[$8>>2]|0;
      $224 = (($223) + 1)|0;
      HEAP32[$8>>2] = $224;
      $225 = HEAP32[$9>>2]|0;
      $226 = (($225) + ($223)|0);
      HEAP8[$226>>0] = $222;
      $227 = HEAP16[$7>>1]|0;
      $228 = ($227&65535) >>> 8;
      $229 = $228&255;
      $230 = HEAP32[$8>>2]|0;
      $231 = (($230) + 1)|0;
      HEAP32[$8>>2] = $231;
      $232 = HEAP32[$9>>2]|0;
      $233 = (($232) + ($230)|0);
      HEAP8[$233>>0] = $229;
      $234 = HEAP32[$5>>2]|0;
      $235 = (16 - ($234))|0;
      $236 = $216 >>> $235;
      $237 = $236&65535;
      HEAP16[$7>>1] = $237;
      $238 = (($211) + -16)|0;
      $239 = (($238) + ($234))|0;
      $241 = $239;$247 = $236;
     } else {
      $240 = (($212) + ($211))|0;
      $241 = $240;$247 = $220;
     }
     HEAP32[$5>>2] = $241;
     $242 = ($241|0)>(9);
     $243 = (($count$0$lcssa41) + 65526)|0;
     $244 = $243 & 65535;
     $245 = $244 << $241;
     $246 = $247 & 65535;
     $248 = $246 | $245;
     $249 = $248&65535;
     HEAP16[$7>>1] = $249;
     if ($242) {
      $250 = $248&255;
      $251 = HEAP32[$8>>2]|0;
      $252 = (($251) + 1)|0;
      HEAP32[$8>>2] = $252;
      $253 = HEAP32[$9>>2]|0;
      $254 = (($253) + ($251)|0);
      HEAP8[$254>>0] = $250;
      $255 = HEAP16[$7>>1]|0;
      $256 = ($255&65535) >>> 8;
      $257 = $256&255;
      $258 = HEAP32[$8>>2]|0;
      $259 = (($258) + 1)|0;
      HEAP32[$8>>2] = $259;
      $260 = HEAP32[$9>>2]|0;
      $261 = (($260) + ($258)|0);
      HEAP8[$261>>0] = $257;
      $262 = HEAP32[$5>>2]|0;
      $263 = (16 - ($262))|0;
      $264 = $244 >>> $263;
      $265 = $264&65535;
      HEAP16[$7>>1] = $265;
      $266 = (($262) + -9)|0;
      $storemerge1 = $266;
     } else {
      $267 = (($241) + 7)|0;
      $storemerge1 = $267;
     }
     HEAP32[$5>>2] = $storemerge1;
     break;
    }
   }
  } while(0);
  $268 = ($$lcssa42<<16>>16)==(0);
  $$10 = $$lcssa45 ? 6 : 7;
  $269 = $268 | $$lcssa45;
  $min_count$1$ph$be = $269 ? 3 : 4;
  $max_count$1$ph$be = $268 ? 138 : $$10;
  $prevlen$0$ph$phi = $nextlen$0$ph;$max_count$1$ph = $max_count$1$ph$be;$min_count$1$ph = $min_count$1$ph$be;$n$0$ph = $$lcssa;$nextlen$0$ph = $$lcssa43;$prevlen$0$ph = $prevlen$0$ph$phi;
 }
 return;
}
function _uncompress($dest,$destLen,$source,$sourceLen) {
 $dest = $dest|0;
 $destLen = $destLen|0;
 $source = $source|0;
 $sourceLen = $sourceLen|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $stream = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $stream = sp;
 HEAP32[$stream>>2] = $source;
 $0 = ((($stream)) + 4|0);
 HEAP32[$0>>2] = $sourceLen;
 $1 = ((($stream)) + 12|0);
 HEAP32[$1>>2] = $dest;
 $2 = HEAP32[$destLen>>2]|0;
 $3 = ((($stream)) + 16|0);
 HEAP32[$3>>2] = $2;
 $4 = ((($stream)) + 32|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($stream)) + 36|0);
 HEAP32[$5>>2] = 0;
 $6 = (_inflateInit_($stream,613577,56)|0);
 $7 = ($6|0)==(0);
 if (!($7)) {
  $$0 = $6;
  STACKTOP = sp;return ($$0|0);
 }
 $8 = (_inflate($stream,4)|0);
 $9 = ($8|0)==(1);
 if ($9) {
  $12 = ((($stream)) + 20|0);
  $13 = HEAP32[$12>>2]|0;
  HEAP32[$destLen>>2] = $13;
  $14 = (_inflateEnd($stream)|0);
  $$0 = $14;
  STACKTOP = sp;return ($$0|0);
 }
 (_inflateEnd($stream)|0);
 switch ($8|0) {
 case -5:  {
  label = 4;
  break;
 }
 case 2:  {
  $$0 = -3;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 default: {
 }
 }
 if ((label|0) == 4) {
  $10 = HEAP32[$0>>2]|0;
  $11 = ($10|0)==(0);
  if ($11) {
   $$0 = -3;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $$0 = $8;
 STACKTOP = sp;return ($$0|0);
}
function _zcalloc($opaque,$items,$size) {
 $opaque = $opaque|0;
 $items = $items|0;
 $size = $size|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = Math_imul($size, $items)|0;
 $1 = (_malloc($0)|0);
 return ($1|0);
}
function _zcfree($opaque,$ptr) {
 $opaque = $opaque|0;
 $ptr = $ptr|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _free($ptr);
 return;
}
function _adler32($adler,$buf,$len) {
 $adler = $adler|0;
 $buf = $buf|0;
 $len = $len|0;
 var $$ = 0, $$$16 = 0, $$010 = 0, $$0117 = 0, $$018 = 0, $$1$lcssa6469 = 0, $$12$lcssa6568 = 0, $$1239 = 0, $$140 = 0, $$16$lcssa = 0, $$1616 = 0, $$23 = 0, $$231 = 0, $$323 = 0, $$3430 = 0, $$38$lcssa6667 = 0, $$3838 = 0, $$422 = 0, $$49 = 0, $$529 = 0;
 var $$6$lcssa = 0, $$621 = 0, $$7 = 0, $$lcssa108 = 0, $$lcssa109 = 0, $$lcssa110 = 0, $$lcssa111 = 0, $$lcssa112 = 0, $$lcssa113 = 0, $$lcssa114 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0;
 var $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0;
 var $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $n$0 = 0, $scevgep = 0, $scevgep61 = 0, $scevgep61$lcssa = 0, $sum2$0 = 0, $sum2$1$lcssa = 0, $sum2$119 = 0, $sum2$2$lcssa6370 = 0, $sum2$241 = 0, $sum2$3 = 0, $sum2$432 = 0, $sum2$5$lcssa = 0, $sum2$524 = 0, $sum2$6 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = $adler >>> 16;
 $1 = $adler & 65535;
 $2 = ($len|0)==(1);
 if ($2) {
  $3 = HEAP8[$buf>>0]|0;
  $4 = $3&255;
  $5 = (($4) + ($1))|0;
  $6 = ($5>>>0)>(65520);
  $7 = (($5) + -65521)|0;
  $$ = $6 ? $7 : $5;
  $8 = (($$) + ($0))|0;
  $9 = ($8>>>0)>(65520);
  $10 = (($8) + 15)|0;
  $sum2$0 = $9 ? $10 : $8;
  $11 = $sum2$0 << 16;
  $12 = $11 | $$;
  $$010 = $12;
  return ($$010|0);
 }
 $13 = ($buf|0)==(0|0);
 if ($13) {
  $$010 = 1;
  return ($$010|0);
 }
 $14 = ($len>>>0)<(16);
 if ($14) {
  $16 = ($len|0)==(0);
  if ($16) {
   $$16$lcssa = $1;$sum2$1$lcssa = $0;
  } else {
   $$0117 = $buf;$$018 = $len;$$1616 = $1;$sum2$119 = $0;
   while(1) {
    $17 = (($$018) + -1)|0;
    $18 = ((($$0117)) + 1|0);
    $19 = HEAP8[$$0117>>0]|0;
    $20 = $19&255;
    $21 = (($20) + ($$1616))|0;
    $22 = (($21) + ($sum2$119))|0;
    $23 = ($17|0)==(0);
    if ($23) {
     $$16$lcssa = $21;$sum2$1$lcssa = $22;
     break;
    } else {
     $$0117 = $18;$$018 = $17;$$1616 = $21;$sum2$119 = $22;
    }
   }
  }
  $24 = ($$16$lcssa>>>0)>(65520);
  $25 = (($$16$lcssa) + -65521)|0;
  $$$16 = $24 ? $25 : $$16$lcssa;
  $26 = (($sum2$1$lcssa>>>0) % 65521)&-1;
  $27 = $26 << 16;
  $28 = $27 | $$$16;
  $$010 = $28;
  return ($$010|0);
 }
 $15 = ($len>>>0)>(5551);
 if ($15) {
  $$1239 = $buf;$$140 = $len;$$3838 = $1;$sum2$241 = $0;
  while(1) {
   $29 = (($$140) + -5552)|0;
   $$23 = $$1239;$$49 = $$3838;$n$0 = 347;$sum2$3 = $sum2$241;
   while(1) {
    $30 = HEAP8[$$23>>0]|0;
    $31 = $30&255;
    $32 = (($31) + ($$49))|0;
    $33 = (($32) + ($sum2$3))|0;
    $34 = ((($$23)) + 1|0);
    $35 = HEAP8[$34>>0]|0;
    $36 = $35&255;
    $37 = (($32) + ($36))|0;
    $38 = (($33) + ($37))|0;
    $39 = ((($$23)) + 2|0);
    $40 = HEAP8[$39>>0]|0;
    $41 = $40&255;
    $42 = (($37) + ($41))|0;
    $43 = (($38) + ($42))|0;
    $44 = ((($$23)) + 3|0);
    $45 = HEAP8[$44>>0]|0;
    $46 = $45&255;
    $47 = (($42) + ($46))|0;
    $48 = (($43) + ($47))|0;
    $49 = ((($$23)) + 4|0);
    $50 = HEAP8[$49>>0]|0;
    $51 = $50&255;
    $52 = (($47) + ($51))|0;
    $53 = (($48) + ($52))|0;
    $54 = ((($$23)) + 5|0);
    $55 = HEAP8[$54>>0]|0;
    $56 = $55&255;
    $57 = (($52) + ($56))|0;
    $58 = (($53) + ($57))|0;
    $59 = ((($$23)) + 6|0);
    $60 = HEAP8[$59>>0]|0;
    $61 = $60&255;
    $62 = (($57) + ($61))|0;
    $63 = (($58) + ($62))|0;
    $64 = ((($$23)) + 7|0);
    $65 = HEAP8[$64>>0]|0;
    $66 = $65&255;
    $67 = (($62) + ($66))|0;
    $68 = (($63) + ($67))|0;
    $69 = ((($$23)) + 8|0);
    $70 = HEAP8[$69>>0]|0;
    $71 = $70&255;
    $72 = (($67) + ($71))|0;
    $73 = (($68) + ($72))|0;
    $74 = ((($$23)) + 9|0);
    $75 = HEAP8[$74>>0]|0;
    $76 = $75&255;
    $77 = (($72) + ($76))|0;
    $78 = (($73) + ($77))|0;
    $79 = ((($$23)) + 10|0);
    $80 = HEAP8[$79>>0]|0;
    $81 = $80&255;
    $82 = (($77) + ($81))|0;
    $83 = (($78) + ($82))|0;
    $84 = ((($$23)) + 11|0);
    $85 = HEAP8[$84>>0]|0;
    $86 = $85&255;
    $87 = (($82) + ($86))|0;
    $88 = (($83) + ($87))|0;
    $89 = ((($$23)) + 12|0);
    $90 = HEAP8[$89>>0]|0;
    $91 = $90&255;
    $92 = (($87) + ($91))|0;
    $93 = (($88) + ($92))|0;
    $94 = ((($$23)) + 13|0);
    $95 = HEAP8[$94>>0]|0;
    $96 = $95&255;
    $97 = (($92) + ($96))|0;
    $98 = (($93) + ($97))|0;
    $99 = ((($$23)) + 14|0);
    $100 = HEAP8[$99>>0]|0;
    $101 = $100&255;
    $102 = (($97) + ($101))|0;
    $103 = (($98) + ($102))|0;
    $104 = ((($$23)) + 15|0);
    $105 = HEAP8[$104>>0]|0;
    $106 = $105&255;
    $107 = (($102) + ($106))|0;
    $108 = (($103) + ($107))|0;
    $109 = ((($$23)) + 16|0);
    $110 = (($n$0) + -1)|0;
    $111 = ($110|0)==(0);
    if ($111) {
     $$lcssa110 = $107;$$lcssa111 = $108;
     break;
    } else {
     $$23 = $109;$$49 = $107;$n$0 = $110;$sum2$3 = $108;
    }
   }
   $scevgep61 = ((($$1239)) + 5552|0);
   $112 = (($$lcssa110>>>0) % 65521)&-1;
   $113 = (($$lcssa111>>>0) % 65521)&-1;
   $114 = ($29>>>0)>(5551);
   if ($114) {
    $$1239 = $scevgep61;$$140 = $29;$$3838 = $112;$sum2$241 = $113;
   } else {
    $$lcssa112 = $29;$$lcssa113 = $112;$$lcssa114 = $113;$scevgep61$lcssa = $scevgep61;
    break;
   }
  }
  $115 = ($$lcssa112|0)==(0);
  if ($115) {
   $$7 = $$lcssa113;$sum2$6 = $$lcssa114;
  } else {
   $116 = ($$lcssa112>>>0)>(15);
   if ($116) {
    $$1$lcssa6469 = $$lcssa112;$$12$lcssa6568 = $scevgep61$lcssa;$$38$lcssa6667 = $$lcssa113;$sum2$2$lcssa6370 = $$lcssa114;
    label = 14;
   } else {
    $$323 = $$lcssa112;$$422 = $scevgep61$lcssa;$$621 = $$lcssa113;$sum2$524 = $$lcssa114;
    label = 17;
   }
  }
 } else {
  $$1$lcssa6469 = $len;$$12$lcssa6568 = $buf;$$38$lcssa6667 = $1;$sum2$2$lcssa6370 = $0;
  label = 14;
 }
 if ((label|0) == 14) {
  $117 = (($$1$lcssa6469) + -16)|0;
  $118 = $117 & -16;
  $119 = (($118) + 16)|0;
  $scevgep = (($$12$lcssa6568) + ($119)|0);
  $$231 = $$1$lcssa6469;$$3430 = $$12$lcssa6568;$$529 = $$38$lcssa6667;$sum2$432 = $sum2$2$lcssa6370;
  while(1) {
   $122 = (($$231) + -16)|0;
   $123 = HEAP8[$$3430>>0]|0;
   $124 = $123&255;
   $125 = (($124) + ($$529))|0;
   $126 = (($125) + ($sum2$432))|0;
   $127 = ((($$3430)) + 1|0);
   $128 = HEAP8[$127>>0]|0;
   $129 = $128&255;
   $130 = (($125) + ($129))|0;
   $131 = (($126) + ($130))|0;
   $132 = ((($$3430)) + 2|0);
   $133 = HEAP8[$132>>0]|0;
   $134 = $133&255;
   $135 = (($130) + ($134))|0;
   $136 = (($131) + ($135))|0;
   $137 = ((($$3430)) + 3|0);
   $138 = HEAP8[$137>>0]|0;
   $139 = $138&255;
   $140 = (($135) + ($139))|0;
   $141 = (($136) + ($140))|0;
   $142 = ((($$3430)) + 4|0);
   $143 = HEAP8[$142>>0]|0;
   $144 = $143&255;
   $145 = (($140) + ($144))|0;
   $146 = (($141) + ($145))|0;
   $147 = ((($$3430)) + 5|0);
   $148 = HEAP8[$147>>0]|0;
   $149 = $148&255;
   $150 = (($145) + ($149))|0;
   $151 = (($146) + ($150))|0;
   $152 = ((($$3430)) + 6|0);
   $153 = HEAP8[$152>>0]|0;
   $154 = $153&255;
   $155 = (($150) + ($154))|0;
   $156 = (($151) + ($155))|0;
   $157 = ((($$3430)) + 7|0);
   $158 = HEAP8[$157>>0]|0;
   $159 = $158&255;
   $160 = (($155) + ($159))|0;
   $161 = (($156) + ($160))|0;
   $162 = ((($$3430)) + 8|0);
   $163 = HEAP8[$162>>0]|0;
   $164 = $163&255;
   $165 = (($160) + ($164))|0;
   $166 = (($161) + ($165))|0;
   $167 = ((($$3430)) + 9|0);
   $168 = HEAP8[$167>>0]|0;
   $169 = $168&255;
   $170 = (($165) + ($169))|0;
   $171 = (($166) + ($170))|0;
   $172 = ((($$3430)) + 10|0);
   $173 = HEAP8[$172>>0]|0;
   $174 = $173&255;
   $175 = (($170) + ($174))|0;
   $176 = (($171) + ($175))|0;
   $177 = ((($$3430)) + 11|0);
   $178 = HEAP8[$177>>0]|0;
   $179 = $178&255;
   $180 = (($175) + ($179))|0;
   $181 = (($176) + ($180))|0;
   $182 = ((($$3430)) + 12|0);
   $183 = HEAP8[$182>>0]|0;
   $184 = $183&255;
   $185 = (($180) + ($184))|0;
   $186 = (($181) + ($185))|0;
   $187 = ((($$3430)) + 13|0);
   $188 = HEAP8[$187>>0]|0;
   $189 = $188&255;
   $190 = (($185) + ($189))|0;
   $191 = (($186) + ($190))|0;
   $192 = ((($$3430)) + 14|0);
   $193 = HEAP8[$192>>0]|0;
   $194 = $193&255;
   $195 = (($190) + ($194))|0;
   $196 = (($191) + ($195))|0;
   $197 = ((($$3430)) + 15|0);
   $198 = HEAP8[$197>>0]|0;
   $199 = $198&255;
   $200 = (($195) + ($199))|0;
   $201 = (($196) + ($200))|0;
   $202 = ((($$3430)) + 16|0);
   $203 = ($122>>>0)>(15);
   if ($203) {
    $$231 = $122;$$3430 = $202;$$529 = $200;$sum2$432 = $201;
   } else {
    $$lcssa108 = $200;$$lcssa109 = $201;
    break;
   }
  }
  $120 = (($117) - ($118))|0;
  $121 = ($117|0)==($118|0);
  if ($121) {
   $$6$lcssa = $$lcssa108;$sum2$5$lcssa = $$lcssa109;
   label = 18;
  } else {
   $$323 = $120;$$422 = $scevgep;$$621 = $$lcssa108;$sum2$524 = $$lcssa109;
   label = 17;
  }
 }
 if ((label|0) == 17) {
  while(1) {
   label = 0;
   $204 = (($$323) + -1)|0;
   $205 = ((($$422)) + 1|0);
   $206 = HEAP8[$$422>>0]|0;
   $207 = $206&255;
   $208 = (($207) + ($$621))|0;
   $209 = (($208) + ($sum2$524))|0;
   $210 = ($204|0)==(0);
   if ($210) {
    $$6$lcssa = $208;$sum2$5$lcssa = $209;
    label = 18;
    break;
   } else {
    $$323 = $204;$$422 = $205;$$621 = $208;$sum2$524 = $209;
    label = 17;
   }
  }
 }
 if ((label|0) == 18) {
  $211 = (($$6$lcssa>>>0) % 65521)&-1;
  $212 = (($sum2$5$lcssa>>>0) % 65521)&-1;
  $$7 = $211;$sum2$6 = $212;
 }
 $213 = $sum2$6 << 16;
 $214 = $213 | $$7;
 $$010 = $214;
 return ($$010|0);
}
function _crc32($crc,$buf,$len) {
 $crc = $crc|0;
 $buf = $buf|0;
 $len = $len|0;
 var $$0116$i = 0, $$0116$i$lcssa = 0, $$017$i = 0, $$017$i$lcssa = 0, $$1 = 0, $$1$lcssa$i = 0, $$12$i = 0, $$19$i = 0, $$2$lcssa$i = 0, $$24$i = 0, $$3$i = 0, $$lcssa36 = 0, $$lcssa37 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0;
 var $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0;
 var $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0;
 var $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0;
 var $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $buf4$0$lcssa$i = 0, $buf4$011$i = 0, $buf4$1$lcssa$i = 0, $buf4$16$i = 0;
 var $c$0$masked$i = 0, $c$018$i = 0, $c$018$i$lcssa = 0, $c$1$lcssa$i = 0, $c$110$i = 0, $c$2$lcssa$i = 0, $c$25$i = 0, $c$3$i = 0, $c$3$masked$i = 0, $c$4$i = 0, $scevgep$i = 0, $scevgep34$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($buf|0)==(0|0);
 if ($0) {
  $$1 = 0;
  return ($$1|0);
 }
 $1 = $crc ^ -1;
 $2 = ($len|0)==(0);
 L4: do {
  if ($2) {
   $c$4$i = $1;
  } else {
   $$0116$i = $buf;$$017$i = $len;$c$018$i = $1;
   while(1) {
    $3 = $$0116$i;
    $4 = $3 & 3;
    $5 = ($4|0)==(0);
    if ($5) {
     $$0116$i$lcssa = $$0116$i;$$017$i$lcssa = $$017$i;$c$018$i$lcssa = $c$018$i;
     break;
    }
    $6 = ((($$0116$i)) + 1|0);
    $7 = HEAP8[$$0116$i>>0]|0;
    $8 = $7&255;
    $c$0$masked$i = $c$018$i & 255;
    $9 = $8 ^ $c$0$masked$i;
    $10 = (141780 + ($9<<2)|0);
    $11 = HEAP32[$10>>2]|0;
    $12 = $c$018$i >>> 8;
    $13 = $11 ^ $12;
    $14 = (($$017$i) + -1)|0;
    $15 = ($14|0)==(0);
    if ($15) {
     $c$4$i = $13;
     break L4;
    } else {
     $$0116$i = $6;$$017$i = $14;$c$018$i = $13;
    }
   }
   $16 = ($$017$i$lcssa>>>0)>(31);
   if ($16) {
    $17 = (($$017$i$lcssa) + -32)|0;
    $18 = $17 & -32;
    $19 = (($18) + 32)|0;
    $$19$i = $$017$i$lcssa;$buf4$011$i = $$0116$i$lcssa;$c$110$i = $c$018$i$lcssa;
    while(1) {
     $25 = ((($buf4$011$i)) + 4|0);
     $26 = HEAP32[$buf4$011$i>>2]|0;
     $27 = $26 ^ $c$110$i;
     $28 = $27 & 255;
     $29 = (144852 + ($28<<2)|0);
     $30 = HEAP32[$29>>2]|0;
     $31 = $27 >>> 8;
     $32 = $31 & 255;
     $33 = (143828 + ($32<<2)|0);
     $34 = HEAP32[$33>>2]|0;
     $35 = $34 ^ $30;
     $36 = $27 >>> 16;
     $37 = $36 & 255;
     $38 = (142804 + ($37<<2)|0);
     $39 = HEAP32[$38>>2]|0;
     $40 = $35 ^ $39;
     $41 = $27 >>> 24;
     $42 = (141780 + ($41<<2)|0);
     $43 = HEAP32[$42>>2]|0;
     $44 = $40 ^ $43;
     $45 = ((($buf4$011$i)) + 8|0);
     $46 = HEAP32[$25>>2]|0;
     $47 = $44 ^ $46;
     $48 = $47 & 255;
     $49 = (144852 + ($48<<2)|0);
     $50 = HEAP32[$49>>2]|0;
     $51 = $47 >>> 8;
     $52 = $51 & 255;
     $53 = (143828 + ($52<<2)|0);
     $54 = HEAP32[$53>>2]|0;
     $55 = $54 ^ $50;
     $56 = $47 >>> 16;
     $57 = $56 & 255;
     $58 = (142804 + ($57<<2)|0);
     $59 = HEAP32[$58>>2]|0;
     $60 = $55 ^ $59;
     $61 = $47 >>> 24;
     $62 = (141780 + ($61<<2)|0);
     $63 = HEAP32[$62>>2]|0;
     $64 = $60 ^ $63;
     $65 = ((($buf4$011$i)) + 12|0);
     $66 = HEAP32[$45>>2]|0;
     $67 = $64 ^ $66;
     $68 = $67 & 255;
     $69 = (144852 + ($68<<2)|0);
     $70 = HEAP32[$69>>2]|0;
     $71 = $67 >>> 8;
     $72 = $71 & 255;
     $73 = (143828 + ($72<<2)|0);
     $74 = HEAP32[$73>>2]|0;
     $75 = $74 ^ $70;
     $76 = $67 >>> 16;
     $77 = $76 & 255;
     $78 = (142804 + ($77<<2)|0);
     $79 = HEAP32[$78>>2]|0;
     $80 = $75 ^ $79;
     $81 = $67 >>> 24;
     $82 = (141780 + ($81<<2)|0);
     $83 = HEAP32[$82>>2]|0;
     $84 = $80 ^ $83;
     $85 = ((($buf4$011$i)) + 16|0);
     $86 = HEAP32[$65>>2]|0;
     $87 = $84 ^ $86;
     $88 = $87 & 255;
     $89 = (144852 + ($88<<2)|0);
     $90 = HEAP32[$89>>2]|0;
     $91 = $87 >>> 8;
     $92 = $91 & 255;
     $93 = (143828 + ($92<<2)|0);
     $94 = HEAP32[$93>>2]|0;
     $95 = $94 ^ $90;
     $96 = $87 >>> 16;
     $97 = $96 & 255;
     $98 = (142804 + ($97<<2)|0);
     $99 = HEAP32[$98>>2]|0;
     $100 = $95 ^ $99;
     $101 = $87 >>> 24;
     $102 = (141780 + ($101<<2)|0);
     $103 = HEAP32[$102>>2]|0;
     $104 = $100 ^ $103;
     $105 = ((($buf4$011$i)) + 20|0);
     $106 = HEAP32[$85>>2]|0;
     $107 = $104 ^ $106;
     $108 = $107 & 255;
     $109 = (144852 + ($108<<2)|0);
     $110 = HEAP32[$109>>2]|0;
     $111 = $107 >>> 8;
     $112 = $111 & 255;
     $113 = (143828 + ($112<<2)|0);
     $114 = HEAP32[$113>>2]|0;
     $115 = $114 ^ $110;
     $116 = $107 >>> 16;
     $117 = $116 & 255;
     $118 = (142804 + ($117<<2)|0);
     $119 = HEAP32[$118>>2]|0;
     $120 = $115 ^ $119;
     $121 = $107 >>> 24;
     $122 = (141780 + ($121<<2)|0);
     $123 = HEAP32[$122>>2]|0;
     $124 = $120 ^ $123;
     $125 = ((($buf4$011$i)) + 24|0);
     $126 = HEAP32[$105>>2]|0;
     $127 = $124 ^ $126;
     $128 = $127 & 255;
     $129 = (144852 + ($128<<2)|0);
     $130 = HEAP32[$129>>2]|0;
     $131 = $127 >>> 8;
     $132 = $131 & 255;
     $133 = (143828 + ($132<<2)|0);
     $134 = HEAP32[$133>>2]|0;
     $135 = $134 ^ $130;
     $136 = $127 >>> 16;
     $137 = $136 & 255;
     $138 = (142804 + ($137<<2)|0);
     $139 = HEAP32[$138>>2]|0;
     $140 = $135 ^ $139;
     $141 = $127 >>> 24;
     $142 = (141780 + ($141<<2)|0);
     $143 = HEAP32[$142>>2]|0;
     $144 = $140 ^ $143;
     $145 = ((($buf4$011$i)) + 28|0);
     $146 = HEAP32[$125>>2]|0;
     $147 = $144 ^ $146;
     $148 = $147 & 255;
     $149 = (144852 + ($148<<2)|0);
     $150 = HEAP32[$149>>2]|0;
     $151 = $147 >>> 8;
     $152 = $151 & 255;
     $153 = (143828 + ($152<<2)|0);
     $154 = HEAP32[$153>>2]|0;
     $155 = $154 ^ $150;
     $156 = $147 >>> 16;
     $157 = $156 & 255;
     $158 = (142804 + ($157<<2)|0);
     $159 = HEAP32[$158>>2]|0;
     $160 = $155 ^ $159;
     $161 = $147 >>> 24;
     $162 = (141780 + ($161<<2)|0);
     $163 = HEAP32[$162>>2]|0;
     $164 = $160 ^ $163;
     $165 = ((($buf4$011$i)) + 32|0);
     $166 = HEAP32[$145>>2]|0;
     $167 = $164 ^ $166;
     $168 = $167 & 255;
     $169 = (144852 + ($168<<2)|0);
     $170 = HEAP32[$169>>2]|0;
     $171 = $167 >>> 8;
     $172 = $171 & 255;
     $173 = (143828 + ($172<<2)|0);
     $174 = HEAP32[$173>>2]|0;
     $175 = $174 ^ $170;
     $176 = $167 >>> 16;
     $177 = $176 & 255;
     $178 = (142804 + ($177<<2)|0);
     $179 = HEAP32[$178>>2]|0;
     $180 = $175 ^ $179;
     $181 = $167 >>> 24;
     $182 = (141780 + ($181<<2)|0);
     $183 = HEAP32[$182>>2]|0;
     $184 = $180 ^ $183;
     $185 = (($$19$i) + -32)|0;
     $186 = ($185>>>0)>(31);
     if ($186) {
      $$19$i = $185;$buf4$011$i = $165;$c$110$i = $184;
     } else {
      $$lcssa37 = $184;
      break;
     }
    }
    $scevgep34$i = (($$0116$i$lcssa) + ($19)|0);
    $20 = (($17) - ($18))|0;
    $$1$lcssa$i = $20;$buf4$0$lcssa$i = $scevgep34$i;$c$1$lcssa$i = $$lcssa37;
   } else {
    $$1$lcssa$i = $$017$i$lcssa;$buf4$0$lcssa$i = $$0116$i$lcssa;$c$1$lcssa$i = $c$018$i$lcssa;
   }
   $21 = ($$1$lcssa$i>>>0)>(3);
   if ($21) {
    $22 = (($$1$lcssa$i) + -4)|0;
    $23 = $22 >>> 2;
    $24 = $23 << 2;
    $$24$i = $$1$lcssa$i;$buf4$16$i = $buf4$0$lcssa$i;$c$25$i = $c$1$lcssa$i;
    while(1) {
     $187 = ((($buf4$16$i)) + 4|0);
     $188 = HEAP32[$buf4$16$i>>2]|0;
     $189 = $188 ^ $c$25$i;
     $190 = $189 & 255;
     $191 = (144852 + ($190<<2)|0);
     $192 = HEAP32[$191>>2]|0;
     $193 = $189 >>> 8;
     $194 = $193 & 255;
     $195 = (143828 + ($194<<2)|0);
     $196 = HEAP32[$195>>2]|0;
     $197 = $196 ^ $192;
     $198 = $189 >>> 16;
     $199 = $198 & 255;
     $200 = (142804 + ($199<<2)|0);
     $201 = HEAP32[$200>>2]|0;
     $202 = $197 ^ $201;
     $203 = $189 >>> 24;
     $204 = (141780 + ($203<<2)|0);
     $205 = HEAP32[$204>>2]|0;
     $206 = $202 ^ $205;
     $207 = (($$24$i) + -4)|0;
     $208 = ($207>>>0)>(3);
     if ($208) {
      $$24$i = $207;$buf4$16$i = $187;$c$25$i = $206;
     } else {
      $$lcssa36 = $206;
      break;
     }
    }
    $209 = (($23) + 1)|0;
    $210 = (($22) - ($24))|0;
    $scevgep$i = (($buf4$0$lcssa$i) + ($209<<2)|0);
    $$2$lcssa$i = $210;$buf4$1$lcssa$i = $scevgep$i;$c$2$lcssa$i = $$lcssa36;
   } else {
    $$2$lcssa$i = $$1$lcssa$i;$buf4$1$lcssa$i = $buf4$0$lcssa$i;$c$2$lcssa$i = $c$1$lcssa$i;
   }
   $211 = ($$2$lcssa$i|0)==(0);
   if ($211) {
    $c$4$i = $c$2$lcssa$i;
   } else {
    $$12$i = $buf4$1$lcssa$i;$$3$i = $$2$lcssa$i;$c$3$i = $c$2$lcssa$i;
    while(1) {
     $212 = ((($$12$i)) + 1|0);
     $213 = HEAP8[$$12$i>>0]|0;
     $214 = $213&255;
     $c$3$masked$i = $c$3$i & 255;
     $215 = $214 ^ $c$3$masked$i;
     $216 = (141780 + ($215<<2)|0);
     $217 = HEAP32[$216>>2]|0;
     $218 = $c$3$i >>> 8;
     $219 = $217 ^ $218;
     $220 = (($$3$i) + -1)|0;
     $221 = ($220|0)==(0);
     if ($221) {
      $c$4$i = $219;
      break;
     } else {
      $$12$i = $212;$$3$i = $220;$c$3$i = $219;
     }
    }
   }
  }
 } while(0);
 $222 = $c$4$i ^ -1;
 $$1 = $222;
 return ($$1|0);
}
function _inflate_fast($strm,$start) {
 $strm = $strm|0;
 $start = $start|0;
 var $$lcssa = 0, $$lcssa13 = 0, $$lcssa20 = 0, $$lcssa22 = 0, $$lcssa236 = 0, $$lcssa237 = 0, $$lcssa239 = 0, $$lcssa24 = 0, $$lcssa240 = 0, $$lcssa242 = 0, $$lcssa243 = 0, $$lcssa247 = 0, $$lcssa248 = 0, $$lcssa249 = 0, $$lcssa251 = 0, $$lcssa257 = 0, $$lcssa258 = 0, $$lcssa269 = 0, $$lcssa270 = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0;
 var $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bits$0 = 0, $bits$1 = 0, $bits$3 = 0, $bits$4 = 0, $bits$5 = 0;
 var $bits$7 = 0, $bits$8 = 0, $bits$9 = 0, $extract$t = 0, $extract$t72 = 0, $from$0 = 0, $from$1 = 0, $from$2 = 0, $from$3 = 0, $from$4$lcssa = 0, $from$4$ph = 0, $from$466 = 0, $from$5 = 0, $from$5$lcssa = 0, $here$sroa$0$0 = 0, $here$sroa$0$0$in = 0, $here$sroa$0$0$in41 = 0, $here$sroa$0$044 = 0, $here$sroa$0$1 = 0, $here$sroa$0$1$in = 0;
 var $here$sroa$0$1$in50 = 0, $here$sroa$0$153 = 0, $here$sroa$10$0 = 0, $here$sroa$10$0$in = 0, $here$sroa$10$0$in43 = 0, $here$sroa$10$0$lcssa = 0, $here$sroa$10$0$lcssa$off0 = 0, $here$sroa$10$046 = 0, $here$sroa$10$047 = 0, $here$sroa$10$047$lcssa = 0, $here$sroa$10$1 = 0, $here$sroa$10$1$in = 0, $here$sroa$10$1$in52 = 0, $here$sroa$10$1$lcssa = 0, $here$sroa$10$155 = 0, $here$sroa$10$156 = 0, $here$sroa$8$0 = 0, $here$sroa$8$0$in = 0, $here$sroa$8$0$in42 = 0, $here$sroa$8$045 = 0;
 var $here$sroa$8$1 = 0, $here$sroa$8$1$in = 0, $here$sroa$8$1$in51 = 0, $here$sroa$8$154 = 0, $hold$0 = 0, $hold$1 = 0, $hold$3 = 0, $hold$4 = 0, $hold$5 = 0, $hold$7 = 0, $hold$8 = 0, $hold$9 = 0, $in$0 = 0, $in$1 = 0, $in$1$lcssa266 = 0, $in$2 = 0, $in$3 = 0, $in$4 = 0, $in$4$lcssa = 0, $in$5 = 0;
 var $in$5$lcssa = 0, $in$6 = 0, $in$7 = 0, $len$0 = 0, $len$1$lcssa = 0, $len$1$ph = 0, $len$165 = 0, $len$2 = 0, $op$0 = 0, $op$1 = 0, $op$2 = 0, $op$3 = 0, $or$cond = 0, $out$0 = 0, $out$0$lcssa = 0, $out$0$lcssa262 = 0, $out$0$lcssa263 = 0, $out$1 = 0, $out$2 = 0, $out$3 = 0;
 var $out$4 = 0, $out$5$lcssa = 0, $out$5$ph = 0, $out$564 = 0, $out$6 = 0, $out$6$lcssa = 0, $out$7 = 0, $out$8 = 0, $scevgep = 0, $scevgep115 = 0, $scevgep116 = 0, $scevgep117 = 0, $scevgep118 = 0, $scevgep119 = 0, $scevgep120 = 0, $scevgep121 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($strm)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$strm>>2]|0;
 $3 = ((($2)) + -1|0);
 $4 = ((($strm)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($5) + -5)|0;
 $7 = (($3) + ($6)|0);
 $8 = ((($strm)) + 12|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + -1|0);
 $11 = ((($strm)) + 16|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($12) - ($start))|0;
 $14 = (($10) + ($13)|0);
 $15 = (($12) + -257)|0;
 $16 = (($10) + ($15)|0);
 $17 = ((($1)) + 40|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($1)) + 44|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ((($1)) + 48|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = ((($1)) + 52|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($1)) + 56|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = ((($1)) + 60|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ((($1)) + 76|0);
 $30 = HEAP32[$29>>2]|0;
 $31 = ((($1)) + 80|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ((($1)) + 84|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = 1 << $34;
 $36 = (($35) + -1)|0;
 $37 = ((($1)) + 88|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = 1 << $38;
 $40 = (($39) + -1)|0;
 $41 = $14;
 $42 = ((($1)) + 7104|0);
 $43 = ((($24)) + -1|0);
 $44 = ($22|0)==(0);
 $45 = (($22) + ($18))|0;
 $46 = (($41) - ($22))|0;
 $bits$0 = $28;$hold$0 = $26;$in$0 = $3;$out$0 = $10;
 L1: while(1) {
  $47 = ($bits$0>>>0)<(15);
  if ($47) {
   $48 = ((($in$0)) + 1|0);
   $49 = HEAP8[$48>>0]|0;
   $50 = $49&255;
   $51 = $50 << $bits$0;
   $52 = (($51) + ($hold$0))|0;
   $53 = (($bits$0) + 8)|0;
   $54 = ((($in$0)) + 2|0);
   $55 = HEAP8[$54>>0]|0;
   $56 = $55&255;
   $57 = $56 << $53;
   $58 = (($52) + ($57))|0;
   $59 = (($bits$0) + 16)|0;
   $bits$1 = $59;$hold$1 = $58;$in$1 = $54;
  } else {
   $bits$1 = $bits$0;$hold$1 = $hold$0;$in$1 = $in$0;
  }
  $60 = $hold$1 & $36;
  $here$sroa$0$0$in41 = (($30) + ($60<<2)|0);
  $here$sroa$8$0$in42 = (((($30) + ($60<<2)|0)) + 1|0);
  $here$sroa$10$0$in43 = (((($30) + ($60<<2)|0)) + 2|0);
  $here$sroa$0$044 = HEAP8[$here$sroa$0$0$in41>>0]|0;
  $here$sroa$8$045 = HEAP8[$here$sroa$8$0$in42>>0]|0;
  $here$sroa$10$046 = HEAP16[$here$sroa$10$0$in43>>1]|0;
  $61 = $here$sroa$8$045&255;
  $62 = $hold$1 >>> $61;
  $63 = (($bits$1) - ($61))|0;
  $64 = ($here$sroa$0$044<<24>>24)==(0);
  $extract$t = $here$sroa$10$046&255;
  do {
   if ($64) {
    $$lcssa = $62;$$lcssa13 = $63;$here$sroa$10$0$lcssa$off0 = $extract$t;
    label = 7;
   } else {
    $65 = $here$sroa$0$044&255;
    $245 = $62;$250 = $63;$68 = $65;$here$sroa$10$047 = $here$sroa$10$046;
    while(1) {
     $67 = $68 & 16;
     $69 = ($67|0)==(0);
     if (!($69)) {
      $$lcssa236 = $68;$$lcssa239 = $250;$$lcssa242 = $245;$here$sroa$10$047$lcssa = $here$sroa$10$047;
      break;
     }
     $239 = $68 & 64;
     $240 = ($239|0)==(0);
     if (!($240)) {
      $$lcssa237 = $68;$$lcssa240 = $250;$$lcssa243 = $245;$in$1$lcssa266 = $in$1;$out$0$lcssa263 = $out$0;
      label = 55;
      break L1;
     }
     $241 = $here$sroa$10$047&65535;
     $242 = 1 << $68;
     $243 = (($242) + -1)|0;
     $244 = $245 & $243;
     $246 = (($244) + ($241))|0;
     $here$sroa$0$0$in = (($30) + ($246<<2)|0);
     $here$sroa$8$0$in = (((($30) + ($246<<2)|0)) + 1|0);
     $here$sroa$10$0$in = (((($30) + ($246<<2)|0)) + 2|0);
     $here$sroa$0$0 = HEAP8[$here$sroa$0$0$in>>0]|0;
     $here$sroa$8$0 = HEAP8[$here$sroa$8$0$in>>0]|0;
     $here$sroa$10$0 = HEAP16[$here$sroa$10$0$in>>1]|0;
     $247 = $here$sroa$8$0&255;
     $248 = $245 >>> $247;
     $249 = (($250) - ($247))|0;
     $251 = $here$sroa$0$0&255;
     $252 = ($here$sroa$0$0<<24>>24)==(0);
     if ($252) {
      $$lcssa247 = $248;$$lcssa248 = $249;$here$sroa$10$0$lcssa = $here$sroa$10$0;
      label = 6;
      break;
     } else {
      $245 = $248;$250 = $249;$68 = $251;$here$sroa$10$047 = $here$sroa$10$0;
     }
    }
    if ((label|0) == 6) {
     label = 0;
     $extract$t72 = $here$sroa$10$0$lcssa&255;
     $$lcssa = $$lcssa247;$$lcssa13 = $$lcssa248;$here$sroa$10$0$lcssa$off0 = $extract$t72;
     label = 7;
     break;
    }
    $70 = $here$sroa$10$047$lcssa&65535;
    $71 = $$lcssa236 & 15;
    $72 = ($71|0)==(0);
    if ($72) {
     $bits$4 = $$lcssa239;$hold$4 = $$lcssa242;$in$3 = $in$1;$len$0 = $70;
    } else {
     $73 = ($$lcssa239>>>0)<($71>>>0);
     if ($73) {
      $74 = ((($in$1)) + 1|0);
      $75 = HEAP8[$74>>0]|0;
      $76 = $75&255;
      $77 = $76 << $$lcssa239;
      $78 = (($77) + ($$lcssa242))|0;
      $79 = (($$lcssa239) + 8)|0;
      $bits$3 = $79;$hold$3 = $78;$in$2 = $74;
     } else {
      $bits$3 = $$lcssa239;$hold$3 = $$lcssa242;$in$2 = $in$1;
     }
     $80 = 1 << $71;
     $81 = (($80) + -1)|0;
     $82 = $hold$3 & $81;
     $83 = (($82) + ($70))|0;
     $84 = $hold$3 >>> $71;
     $85 = (($bits$3) - ($71))|0;
     $bits$4 = $85;$hold$4 = $84;$in$3 = $in$2;$len$0 = $83;
    }
    $86 = ($bits$4>>>0)<(15);
    if ($86) {
     $87 = ((($in$3)) + 1|0);
     $88 = HEAP8[$87>>0]|0;
     $89 = $88&255;
     $90 = $89 << $bits$4;
     $91 = (($90) + ($hold$4))|0;
     $92 = (($bits$4) + 8)|0;
     $93 = ((($in$3)) + 2|0);
     $94 = HEAP8[$93>>0]|0;
     $95 = $94&255;
     $96 = $95 << $92;
     $97 = (($91) + ($96))|0;
     $98 = (($bits$4) + 16)|0;
     $bits$5 = $98;$hold$5 = $97;$in$4 = $93;
    } else {
     $bits$5 = $bits$4;$hold$5 = $hold$4;$in$4 = $in$3;
    }
    $99 = $hold$5 & $40;
    $here$sroa$0$1$in50 = (($32) + ($99<<2)|0);
    $here$sroa$8$1$in51 = (((($32) + ($99<<2)|0)) + 1|0);
    $here$sroa$10$1$in52 = (((($32) + ($99<<2)|0)) + 2|0);
    $here$sroa$0$153 = HEAP8[$here$sroa$0$1$in50>>0]|0;
    $here$sroa$8$154 = HEAP8[$here$sroa$8$1$in51>>0]|0;
    $here$sroa$10$155 = HEAP16[$here$sroa$10$1$in52>>1]|0;
    $100 = $here$sroa$8$154&255;
    $101 = $hold$5 >>> $100;
    $102 = (($bits$5) - ($100))|0;
    $103 = $here$sroa$0$153&255;
    $104 = $103 & 16;
    $105 = ($104|0)==(0);
    if ($105) {
     $223 = $103;$229 = $101;$234 = $102;$here$sroa$10$156 = $here$sroa$10$155;
     while(1) {
      $222 = $223 & 64;
      $224 = ($222|0)==(0);
      if (!($224)) {
       $$lcssa249 = $234;$$lcssa251 = $229;$in$4$lcssa = $in$4;$out$0$lcssa = $out$0;
       label = 52;
       break L1;
      }
      $225 = $here$sroa$10$156&65535;
      $226 = 1 << $223;
      $227 = (($226) + -1)|0;
      $228 = $229 & $227;
      $230 = (($228) + ($225))|0;
      $here$sroa$0$1$in = (($32) + ($230<<2)|0);
      $here$sroa$8$1$in = (((($32) + ($230<<2)|0)) + 1|0);
      $here$sroa$10$1$in = (((($32) + ($230<<2)|0)) + 2|0);
      $here$sroa$0$1 = HEAP8[$here$sroa$0$1$in>>0]|0;
      $here$sroa$8$1 = HEAP8[$here$sroa$8$1$in>>0]|0;
      $here$sroa$10$1 = HEAP16[$here$sroa$10$1$in>>1]|0;
      $231 = $here$sroa$8$1&255;
      $232 = $229 >>> $231;
      $233 = (($234) - ($231))|0;
      $235 = $here$sroa$0$1&255;
      $236 = $235 & 16;
      $237 = ($236|0)==(0);
      if ($237) {
       $223 = $235;$229 = $232;$234 = $233;$here$sroa$10$156 = $here$sroa$10$1;
      } else {
       $$lcssa20 = $232;$$lcssa22 = $233;$$lcssa24 = $235;$here$sroa$10$1$lcssa = $here$sroa$10$1;
       break;
      }
     }
    } else {
     $$lcssa20 = $101;$$lcssa22 = $102;$$lcssa24 = $103;$here$sroa$10$1$lcssa = $here$sroa$10$155;
    }
    $106 = $here$sroa$10$1$lcssa&65535;
    $107 = $$lcssa24 & 15;
    $108 = ($$lcssa22>>>0)<($107>>>0);
    if ($108) {
     $109 = ((($in$4)) + 1|0);
     $110 = HEAP8[$109>>0]|0;
     $111 = $110&255;
     $112 = $111 << $$lcssa22;
     $113 = (($112) + ($$lcssa20))|0;
     $114 = (($$lcssa22) + 8)|0;
     $115 = ($114>>>0)<($107>>>0);
     if ($115) {
      $116 = ((($in$4)) + 2|0);
      $117 = HEAP8[$116>>0]|0;
      $118 = $117&255;
      $119 = $118 << $114;
      $120 = (($119) + ($113))|0;
      $121 = (($$lcssa22) + 16)|0;
      $bits$7 = $121;$hold$7 = $120;$in$5 = $116;
     } else {
      $bits$7 = $114;$hold$7 = $113;$in$5 = $109;
     }
    } else {
     $bits$7 = $$lcssa22;$hold$7 = $$lcssa20;$in$5 = $in$4;
    }
    $122 = 1 << $107;
    $123 = (($122) + -1)|0;
    $124 = $hold$7 & $123;
    $125 = (($124) + ($106))|0;
    $126 = $hold$7 >>> $107;
    $127 = (($bits$7) - ($107))|0;
    $128 = $out$0;
    $129 = (($128) - ($41))|0;
    $130 = ($125>>>0)>($129>>>0);
    if (!($130)) {
     $201 = (0 - ($125))|0;
     $202 = (($out$0) + ($201)|0);
     $from$5 = $202;$len$2 = $len$0;$out$6 = $out$0;
     while(1) {
      $203 = ((($from$5)) + 1|0);
      $204 = HEAP8[$203>>0]|0;
      $205 = ((($out$6)) + 1|0);
      HEAP8[$205>>0] = $204;
      $206 = ((($from$5)) + 2|0);
      $207 = HEAP8[$206>>0]|0;
      $208 = ((($out$6)) + 2|0);
      HEAP8[$208>>0] = $207;
      $209 = ((($from$5)) + 3|0);
      $210 = HEAP8[$209>>0]|0;
      $211 = ((($out$6)) + 3|0);
      HEAP8[$211>>0] = $210;
      $212 = (($len$2) + -3)|0;
      $213 = ($212>>>0)>(2);
      if ($213) {
       $from$5 = $209;$len$2 = $212;$out$6 = $211;
      } else {
       $$lcssa257 = $211;$$lcssa258 = $212;$from$5$lcssa = $from$5;$out$6$lcssa = $out$6;
       break;
      }
     }
     $214 = ($$lcssa258|0)==(0);
     if ($214) {
      $bits$8 = $127;$hold$8 = $126;$in$6 = $in$5;$out$7 = $$lcssa257;
      break;
     }
     $215 = ((($from$5$lcssa)) + 4|0);
     $216 = HEAP8[$215>>0]|0;
     $217 = ((($out$6$lcssa)) + 4|0);
     HEAP8[$217>>0] = $216;
     $218 = ($$lcssa258>>>0)>(1);
     if (!($218)) {
      $bits$8 = $127;$hold$8 = $126;$in$6 = $in$5;$out$7 = $217;
      break;
     }
     $219 = ((($from$5$lcssa)) + 5|0);
     $220 = HEAP8[$219>>0]|0;
     $221 = ((($out$6$lcssa)) + 5|0);
     HEAP8[$221>>0] = $220;
     $bits$8 = $127;$hold$8 = $126;$in$6 = $in$5;$out$7 = $221;
     break;
    }
    $131 = (($125) - ($129))|0;
    $132 = ($131>>>0)>($20>>>0);
    if ($132) {
     $133 = HEAP32[$42>>2]|0;
     $134 = ($133|0)==(0);
     if (!($134)) {
      $$lcssa269 = $126;$$lcssa270 = $127;$in$5$lcssa = $in$5;$out$0$lcssa262 = $out$0;
      label = 22;
      break L1;
     }
    }
    do {
     if ($44) {
      $136 = (($18) - ($131))|0;
      $137 = (($43) + ($136)|0);
      $138 = ($len$0>>>0)>($131>>>0);
      if ($138) {
       $139 = (($len$0) - ($131))|0;
       $140 = (($125) - ($128))|0;
       $from$0 = $137;$op$0 = $131;$out$1 = $out$0;
       while(1) {
        $141 = ((($from$0)) + 1|0);
        $142 = HEAP8[$141>>0]|0;
        $143 = ((($out$1)) + 1|0);
        HEAP8[$143>>0] = $142;
        $144 = (($op$0) + -1)|0;
        $145 = ($144|0)==(0);
        if ($145) {
         break;
        } else {
         $from$0 = $141;$op$0 = $144;$out$1 = $143;
        }
       }
       $scevgep120 = (($out$0) + ($41)|0);
       $scevgep121 = (($scevgep120) + ($140)|0);
       $146 = (0 - ($125))|0;
       $147 = (($scevgep121) + ($146)|0);
       $from$4$ph = $147;$len$1$ph = $139;$out$5$ph = $scevgep121;
      } else {
       $from$4$ph = $137;$len$1$ph = $len$0;$out$5$ph = $out$0;
      }
     } else {
      $148 = ($22>>>0)<($131>>>0);
      if (!($148)) {
       $169 = (($22) - ($131))|0;
       $170 = (($43) + ($169)|0);
       $171 = ($len$0>>>0)>($131>>>0);
       if (!($171)) {
        $from$4$ph = $170;$len$1$ph = $len$0;$out$5$ph = $out$0;
        break;
       }
       $172 = (($len$0) - ($131))|0;
       $173 = (($125) - ($128))|0;
       $from$3 = $170;$op$3 = $131;$out$4 = $out$0;
       while(1) {
        $174 = ((($from$3)) + 1|0);
        $175 = HEAP8[$174>>0]|0;
        $176 = ((($out$4)) + 1|0);
        HEAP8[$176>>0] = $175;
        $177 = (($op$3) + -1)|0;
        $178 = ($177|0)==(0);
        if ($178) {
         break;
        } else {
         $from$3 = $174;$op$3 = $177;$out$4 = $176;
        }
       }
       $scevgep = (($out$0) + ($41)|0);
       $scevgep115 = (($scevgep) + ($173)|0);
       $179 = (0 - ($125))|0;
       $180 = (($scevgep115) + ($179)|0);
       $from$4$ph = $180;$len$1$ph = $172;$out$5$ph = $scevgep115;
       break;
      }
      $149 = (($45) - ($131))|0;
      $150 = (($43) + ($149)|0);
      $151 = (($131) - ($22))|0;
      $152 = ($len$0>>>0)>($151>>>0);
      if ($152) {
       $153 = (($len$0) - ($151))|0;
       $154 = (($125) - ($128))|0;
       $from$1 = $150;$op$1 = $151;$out$2 = $out$0;
       while(1) {
        $155 = ((($from$1)) + 1|0);
        $156 = HEAP8[$155>>0]|0;
        $157 = ((($out$2)) + 1|0);
        HEAP8[$157>>0] = $156;
        $158 = (($op$1) + -1)|0;
        $159 = ($158|0)==(0);
        if ($159) {
         break;
        } else {
         $from$1 = $155;$op$1 = $158;$out$2 = $157;
        }
       }
       $scevgep116 = (($out$0) + ($46)|0);
       $scevgep117 = (($scevgep116) + ($154)|0);
       $160 = ($153>>>0)>($22>>>0);
       if ($160) {
        $161 = (($153) - ($22))|0;
        $from$2 = $43;$op$2 = $22;$out$3 = $scevgep117;
        while(1) {
         $162 = ((($from$2)) + 1|0);
         $163 = HEAP8[$162>>0]|0;
         $164 = ((($out$3)) + 1|0);
         HEAP8[$164>>0] = $163;
         $165 = (($op$2) + -1)|0;
         $166 = ($165|0)==(0);
         if ($166) {
          break;
         } else {
          $from$2 = $162;$op$2 = $165;$out$3 = $164;
         }
        }
        $scevgep118 = (($out$0) + ($41)|0);
        $scevgep119 = (($scevgep118) + ($154)|0);
        $167 = (0 - ($125))|0;
        $168 = (($scevgep119) + ($167)|0);
        $from$4$ph = $168;$len$1$ph = $161;$out$5$ph = $scevgep119;
       } else {
        $from$4$ph = $43;$len$1$ph = $153;$out$5$ph = $scevgep117;
       }
      } else {
       $from$4$ph = $150;$len$1$ph = $len$0;$out$5$ph = $out$0;
      }
     }
    } while(0);
    $181 = ($len$1$ph>>>0)>(2);
    if ($181) {
     $from$466 = $from$4$ph;$len$165 = $len$1$ph;$out$564 = $out$5$ph;
     while(1) {
      $182 = ((($from$466)) + 1|0);
      $183 = HEAP8[$182>>0]|0;
      $184 = ((($out$564)) + 1|0);
      HEAP8[$184>>0] = $183;
      $185 = ((($from$466)) + 2|0);
      $186 = HEAP8[$185>>0]|0;
      $187 = ((($out$564)) + 2|0);
      HEAP8[$187>>0] = $186;
      $188 = ((($from$466)) + 3|0);
      $189 = HEAP8[$188>>0]|0;
      $190 = ((($out$564)) + 3|0);
      HEAP8[$190>>0] = $189;
      $191 = (($len$165) + -3)|0;
      $192 = ($191>>>0)>(2);
      if ($192) {
       $from$466 = $188;$len$165 = $191;$out$564 = $190;
      } else {
       $from$4$lcssa = $188;$len$1$lcssa = $191;$out$5$lcssa = $190;
       break;
      }
     }
    } else {
     $from$4$lcssa = $from$4$ph;$len$1$lcssa = $len$1$ph;$out$5$lcssa = $out$5$ph;
    }
    $193 = ($len$1$lcssa|0)==(0);
    if ($193) {
     $bits$8 = $127;$hold$8 = $126;$in$6 = $in$5;$out$7 = $out$5$lcssa;
    } else {
     $194 = ((($from$4$lcssa)) + 1|0);
     $195 = HEAP8[$194>>0]|0;
     $196 = ((($out$5$lcssa)) + 1|0);
     HEAP8[$196>>0] = $195;
     $197 = ($len$1$lcssa>>>0)>(1);
     if ($197) {
      $198 = ((($from$4$lcssa)) + 2|0);
      $199 = HEAP8[$198>>0]|0;
      $200 = ((($out$5$lcssa)) + 2|0);
      HEAP8[$200>>0] = $199;
      $bits$8 = $127;$hold$8 = $126;$in$6 = $in$5;$out$7 = $200;
     } else {
      $bits$8 = $127;$hold$8 = $126;$in$6 = $in$5;$out$7 = $196;
     }
    }
   }
  } while(0);
  if ((label|0) == 7) {
   label = 0;
   $66 = ((($out$0)) + 1|0);
   HEAP8[$66>>0] = $here$sroa$10$0$lcssa$off0;
   $bits$8 = $$lcssa13;$hold$8 = $$lcssa;$in$6 = $in$1;$out$7 = $66;
  }
  $256 = ($in$6>>>0)<($7>>>0);
  $257 = ($out$7>>>0)<($16>>>0);
  $or$cond = $257 & $256;
  if ($or$cond) {
   $bits$0 = $bits$8;$hold$0 = $hold$8;$in$0 = $in$6;$out$0 = $out$7;
  } else {
   $bits$9 = $bits$8;$hold$9 = $hold$8;$in$7 = $in$6;$out$8 = $out$7;
   break;
  }
 }
 do {
  if ((label|0) == 22) {
   $135 = ((($strm)) + 24|0);
   HEAP32[$135>>2] = 613699;
   HEAP32[$1>>2] = 29;
   $bits$9 = $$lcssa270;$hold$9 = $$lcssa269;$in$7 = $in$5$lcssa;$out$8 = $out$0$lcssa262;
  }
  else if ((label|0) == 52) {
   $238 = ((($strm)) + 24|0);
   HEAP32[$238>>2] = 613729;
   HEAP32[$1>>2] = 29;
   $bits$9 = $$lcssa249;$hold$9 = $$lcssa251;$in$7 = $in$4$lcssa;$out$8 = $out$0$lcssa;
  }
  else if ((label|0) == 55) {
   $253 = $$lcssa237 & 32;
   $254 = ($253|0)==(0);
   if ($254) {
    $255 = ((($strm)) + 24|0);
    HEAP32[$255>>2] = 613751;
    HEAP32[$1>>2] = 29;
    $bits$9 = $$lcssa240;$hold$9 = $$lcssa243;$in$7 = $in$1$lcssa266;$out$8 = $out$0$lcssa263;
    break;
   } else {
    HEAP32[$1>>2] = 11;
    $bits$9 = $$lcssa240;$hold$9 = $$lcssa243;$in$7 = $in$1$lcssa266;$out$8 = $out$0$lcssa263;
    break;
   }
  }
 } while(0);
 $258 = $bits$9 >>> 3;
 $259 = (0 - ($258))|0;
 $260 = (($in$7) + ($259)|0);
 $261 = $258 << 3;
 $262 = (($bits$9) - ($261))|0;
 $263 = 1 << $262;
 $264 = (($263) + -1)|0;
 $265 = $264 & $hold$9;
 $266 = ((($260)) + 1|0);
 HEAP32[$strm>>2] = $266;
 $267 = ((($out$8)) + 1|0);
 HEAP32[$8>>2] = $267;
 $268 = $7;
 $269 = $260;
 $270 = (($268) + 5)|0;
 $271 = (($270) - ($269))|0;
 HEAP32[$4>>2] = $271;
 $272 = $16;
 $273 = $out$8;
 $274 = (($272) + 257)|0;
 $275 = (($274) - ($273))|0;
 HEAP32[$11>>2] = $275;
 HEAP32[$25>>2] = $265;
 HEAP32[$27>>2] = $262;
 return;
}
function _emscripten_GetProcAddress($name_) {
 $name_ = $name_|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0;
 var $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0;
 var $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0;
 var $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0;
 var $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0;
 var $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0;
 var $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0;
 var $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0;
 var $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0;
 var $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0;
 var $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0;
 var $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0;
 var $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0;
 var $549 = 0, $55 = 0, $550 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $end = 0, $name = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $1 = $name_;
 $2 = $1;
 $3 = (_strlen($2)|0);
 $4 = (($3) + 1)|0;
 $5 = (_malloc($4)|0);
 $name = $5;
 $6 = $name;
 $7 = $1;
 (_strcpy($6,$7)|0);
 $8 = $name;
 $9 = (_strstr($8,613779)|0);
 $end = $9;
 $10 = $end;
 $11 = ($10|0)!=(0|0);
 if ($11) {
  $12 = $end;
  HEAP8[$12>>0] = 0;
 }
 $13 = $name;
 $14 = (_strstr($13,613783)|0);
 $end = $14;
 $15 = $end;
 $16 = ($15|0)!=(0|0);
 if ($16) {
  $17 = $end;
  HEAP8[$17>>0] = 0;
 }
 $18 = $name;
 $19 = (_strstr($18,613787)|0);
 $end = $19;
 $20 = $end;
 $21 = ($20|0)!=(0|0);
 if ($21) {
  $22 = $end;
  HEAP8[$22>>0] = 0;
 }
 $23 = $name;
 $24 = (_strstr($23,613791)|0);
 $end = $24;
 $25 = $end;
 $26 = ($25|0)!=(0|0);
 if ($26) {
  $27 = $end;
  HEAP8[$27>>0] = 0;
 }
 $28 = $name;
 $29 = (_strcmp($28,613797)|0);
 $30 = ($29|0)!=(0);
 do {
  if ($30) {
   $31 = $name;
   $32 = (_strcmp($31,613835)|0);
   $33 = ($32|0)!=(0);
   if (!($33)) {
    $name = 613854;
    break;
   }
   $34 = $name;
   $35 = (_strcmp($34,613867)|0);
   $36 = ($35|0)!=(0);
   if (!($36)) {
    $name = 613888;
    break;
   }
   $37 = $name;
   $38 = (_strcmp($37,613903)|0);
   $39 = ($38|0)!=(0);
   if (!($39)) {
    $name = 613918;
    break;
   }
   $40 = $name;
   $41 = (_strcmp($40,613933)|0);
   $42 = ($41|0)!=(0);
   if (!($42)) {
    $name = 613948;
   }
  } else {
   $name = 613819;
  }
 } while(0);
 $43 = $name;
 $44 = (_strcmp($43,613963)|0);
 $45 = ($44|0)!=(0);
 do {
  if ($45) {
   $46 = $name;
   $47 = (_strcmp($46,613977)|0);
   $48 = ($47|0)!=(0);
   if (!($48)) {
    $0 = 4682;
    break;
   }
   $49 = $name;
   $50 = (_strcmp($49,613989)|0);
   $51 = ($50|0)!=(0);
   if (!($51)) {
    $0 = 4683;
    break;
   }
   $52 = $name;
   $53 = (_strcmp($52,614003)|0);
   $54 = ($53|0)!=(0);
   if (!($54)) {
    $0 = 4684;
    break;
   }
   $55 = $name;
   $56 = (_strcmp($55,614015)|0);
   $57 = ($56|0)!=(0);
   if (!($57)) {
    $0 = 4685;
    break;
   }
   $58 = $name;
   $59 = (_strcmp($58,614029)|0);
   $60 = ($59|0)!=(0);
   if (!($60)) {
    $0 = 4686;
    break;
   }
   $61 = $name;
   $62 = (_strcmp($61,614043)|0);
   $63 = ($62|0)!=(0);
   if (!($63)) {
    $0 = 4687;
    break;
   }
   $64 = $name;
   $65 = (_strcmp($64,614060)|0);
   $66 = ($65|0)!=(0);
   if (!($66)) {
    $0 = 4688;
    break;
   }
   $67 = $name;
   $68 = (_strcmp($67,614083)|0);
   $69 = ($68|0)!=(0);
   if (!($69)) {
    $0 = 4689;
    break;
   }
   $70 = $name;
   $71 = (_strcmp($70,614109)|0);
   $72 = ($71|0)!=(0);
   if (!($72)) {
    $0 = 4690;
    break;
   }
   $73 = $name;
   $74 = (_strcmp($73,614122)|0);
   $75 = ($74|0)!=(0);
   if (!($75)) {
    $0 = 4691;
    break;
   }
   $76 = $name;
   $77 = (_strcmp($76,614138)|0);
   $78 = ($77|0)!=(0);
   if (!($78)) {
    $0 = 4692;
    break;
   }
   $79 = $name;
   $80 = (_strcmp($79,614151)|0);
   $81 = ($80|0)!=(0);
   if (!($81)) {
    $0 = 4693;
    break;
   }
   $82 = $name;
   $83 = (_strcmp($82,614165)|0);
   $84 = ($83|0)!=(0);
   if (!($84)) {
    $0 = 4694;
    break;
   }
   $85 = $name;
   $86 = (_strcmp($85,614185)|0);
   $87 = ($86|0)!=(0);
   if (!($87)) {
    $0 = 4695;
    break;
   }
   $88 = $name;
   $89 = (_strcmp($88,614205)|0);
   $90 = ($89|0)!=(0);
   if (!($90)) {
    $0 = 4696;
    break;
   }
   $91 = $name;
   $92 = (_strcmp($91,614222)|0);
   $93 = ($92|0)!=(0);
   if (!($93)) {
    $0 = 4697;
    break;
   }
   $94 = $name;
   $95 = (_strcmp($94,614239)|0);
   $96 = ($95|0)!=(0);
   if (!($96)) {
    $0 = 4698;
    break;
   }
   $97 = $name;
   $98 = (_strcmp($97,614251)|0);
   $99 = ($98|0)!=(0);
   if (!($99)) {
    $0 = 4699;
    break;
   }
   $100 = $name;
   $101 = (_strcmp($100,614264)|0);
   $102 = ($101|0)!=(0);
   if (!($102)) {
    $0 = 4700;
    break;
   }
   $103 = $name;
   $104 = (_strcmp($103,614280)|0);
   $105 = ($104|0)!=(0);
   if (!($105)) {
    $0 = 4701;
    break;
   }
   $106 = $name;
   $107 = (_strcmp($106,614303)|0);
   $108 = ($107|0)!=(0);
   if (!($108)) {
    $0 = 4702;
    break;
   }
   $109 = $name;
   $110 = (_strcmp($109,614316)|0);
   $111 = ($110|0)!=(0);
   if (!($111)) {
    $0 = 4703;
    break;
   }
   $112 = $name;
   $113 = (_strcmp($112,614332)|0);
   $114 = ($113|0)!=(0);
   if (!($114)) {
    $0 = 4704;
    break;
   }
   $115 = $name;
   $116 = (_strcmp($115,614343)|0);
   $117 = ($116|0)!=(0);
   if (!($117)) {
    $0 = 4705;
    break;
   }
   $118 = $name;
   $119 = (_strcmp($118,614362)|0);
   $120 = ($119|0)!=(0);
   if (!($120)) {
    $0 = 4706;
    break;
   }
   $121 = $name;
   $122 = (_strcmp($121,614384)|0);
   $123 = ($122|0)!=(0);
   if (!($123)) {
    $0 = 4707;
    break;
   }
   $124 = $name;
   $125 = (_strcmp($124,614403)|0);
   $126 = ($125|0)!=(0);
   if (!($126)) {
    $0 = 4708;
    break;
   }
   $127 = $name;
   $128 = (_strcmp($127,614432)|0);
   $129 = ($128|0)!=(0);
   if (!($129)) {
    $0 = 4709;
    break;
   }
   $130 = $name;
   $131 = (_strcmp($130,614449)|0);
   $132 = ($131|0)!=(0);
   if (!($132)) {
    $0 = 4710;
    break;
   }
   $133 = $name;
   $134 = (_strcmp($133,614464)|0);
   $135 = ($134|0)!=(0);
   if (!($135)) {
    $0 = 4711;
    break;
   }
   $136 = $name;
   $137 = (_strcmp($136,614479)|0);
   $138 = ($137|0)!=(0);
   if (!($138)) {
    $0 = 4712;
    break;
   }
   $139 = $name;
   $140 = (_strcmp($139,614500)|0);
   $141 = ($140|0)!=(0);
   if (!($141)) {
    $0 = 4713;
    break;
   }
   $142 = $name;
   $143 = (_strcmp($142,614520)|0);
   $144 = ($143|0)!=(0);
   if (!($144)) {
    $0 = 4714;
    break;
   }
   $145 = $name;
   $146 = (_strcmp($145,614540)|0);
   $147 = ($146|0)!=(0);
   if (!($147)) {
    $0 = 4715;
    break;
   }
   $148 = $name;
   $149 = (_strcmp($148,614566)|0);
   $150 = ($149|0)!=(0);
   if (!($150)) {
    $0 = 4716;
    break;
   }
   $151 = $name;
   $152 = (_strcmp($151,614585)|0);
   $153 = ($152|0)!=(0);
   if (!($153)) {
    $0 = 4717;
    break;
   }
   $154 = $name;
   $155 = (_strcmp($154,614597)|0);
   $156 = ($155|0)!=(0);
   if (!($156)) {
    $0 = 4718;
    break;
   }
   $157 = $name;
   $158 = (_strcmp($157,614609)|0);
   $159 = ($158|0)!=(0);
   if (!($159)) {
    $0 = 4719;
    break;
   }
   $160 = $name;
   $161 = (_strcmp($160,614621)|0);
   $162 = ($161|0)!=(0);
   if (!($162)) {
    $0 = 4720;
    break;
   }
   $163 = $name;
   $164 = (_strcmp($163,614633)|0);
   $165 = ($164|0)!=(0);
   if (!($165)) {
    $0 = 4721;
    break;
   }
   $166 = $name;
   $167 = (_strcmp($166,614645)|0);
   $168 = ($167|0)!=(0);
   if (!($168)) {
    $0 = 4722;
    break;
   }
   $169 = $name;
   $170 = (_strcmp($169,614657)|0);
   $171 = ($170|0)!=(0);
   if (!($171)) {
    $0 = 4723;
    break;
   }
   $172 = $name;
   $173 = (_strcmp($172,614669)|0);
   $174 = ($173|0)!=(0);
   if (!($174)) {
    $0 = 4724;
    break;
   }
   $175 = $name;
   $176 = (_strcmp($175,614681)|0);
   $177 = ($176|0)!=(0);
   if (!($177)) {
    $0 = 4725;
    break;
   }
   $178 = $name;
   $179 = (_strcmp($178,614694)|0);
   $180 = ($179|0)!=(0);
   if (!($180)) {
    $0 = 4726;
    break;
   }
   $181 = $name;
   $182 = (_strcmp($181,614707)|0);
   $183 = ($182|0)!=(0);
   if (!($183)) {
    $0 = 4727;
    break;
   }
   $184 = $name;
   $185 = (_strcmp($184,614720)|0);
   $186 = ($185|0)!=(0);
   if (!($186)) {
    $0 = 4728;
    break;
   }
   $187 = $name;
   $188 = (_strcmp($187,614733)|0);
   $189 = ($188|0)!=(0);
   if (!($189)) {
    $0 = 4729;
    break;
   }
   $190 = $name;
   $191 = (_strcmp($190,614746)|0);
   $192 = ($191|0)!=(0);
   if (!($192)) {
    $0 = 4730;
    break;
   }
   $193 = $name;
   $194 = (_strcmp($193,614759)|0);
   $195 = ($194|0)!=(0);
   if (!($195)) {
    $0 = 4731;
    break;
   }
   $196 = $name;
   $197 = (_strcmp($196,614772)|0);
   $198 = ($197|0)!=(0);
   if (!($198)) {
    $0 = 4732;
    break;
   }
   $199 = $name;
   $200 = (_strcmp($199,614785)|0);
   $201 = ($200|0)!=(0);
   if (!($201)) {
    $0 = 4733;
    break;
   }
   $202 = $name;
   $203 = (_strcmp($202,614804)|0);
   $204 = ($203|0)!=(0);
   if (!($204)) {
    $0 = 4734;
    break;
   }
   $205 = $name;
   $206 = (_strcmp($205,614823)|0);
   $207 = ($206|0)!=(0);
   if (!($207)) {
    $0 = 4735;
    break;
   }
   $208 = $name;
   $209 = (_strcmp($208,614842)|0);
   $210 = ($209|0)!=(0);
   if (!($210)) {
    $0 = 4736;
    break;
   }
   $211 = $name;
   $212 = (_strcmp($211,614855)|0);
   $213 = ($212|0)!=(0);
   if (!($213)) {
    $0 = 4737;
    break;
   }
   $214 = $name;
   $215 = (_strcmp($214,614873)|0);
   $216 = ($215|0)!=(0);
   if (!($216)) {
    $0 = 4738;
    break;
   }
   $217 = $name;
   $218 = (_strcmp($217,614891)|0);
   $219 = ($218|0)!=(0);
   if (!($219)) {
    $0 = 4739;
    break;
   }
   $220 = $name;
   $221 = (_strcmp($220,614909)|0);
   $222 = ($221|0)!=(0);
   if (!($222)) {
    $0 = 4740;
    break;
   }
   $223 = $name;
   $224 = (_strcmp($223,614927)|0);
   $225 = ($224|0)!=(0);
   if (!($225)) {
    $0 = 4741;
    break;
   }
   $226 = $name;
   $227 = (_strcmp($226,614947)|0);
   $228 = ($227|0)!=(0);
   if (!($228)) {
    $0 = 4742;
    break;
   }
   $229 = $name;
   $230 = (_strcmp($229,613888)|0);
   $231 = ($230|0)!=(0);
   if (!($231)) {
    $0 = 4743;
    break;
   }
   $232 = $name;
   $233 = (_strcmp($232,614965)|0);
   $234 = ($233|0)!=(0);
   if (!($234)) {
    $0 = 4744;
    break;
   }
   $235 = $name;
   $236 = (_strcmp($235,614980)|0);
   $237 = ($236|0)!=(0);
   if (!($237)) {
    $0 = 4745;
    break;
   }
   $238 = $name;
   $239 = (_strcmp($238,615001)|0);
   $240 = ($239|0)!=(0);
   if (!($240)) {
    $0 = 4746;
    break;
   }
   $241 = $name;
   $242 = (_strcmp($241,615016)|0);
   $243 = ($242|0)!=(0);
   if (!($243)) {
    $0 = 4747;
    break;
   }
   $244 = $name;
   $245 = (_strcmp($244,615034)|0);
   $246 = ($245|0)!=(0);
   if (!($246)) {
    $0 = 4748;
    break;
   }
   $247 = $name;
   $248 = (_strcmp($247,615050)|0);
   $249 = ($248|0)!=(0);
   if (!($249)) {
    $0 = 4749;
    break;
   }
   $250 = $name;
   $251 = (_strcmp($250,615069)|0);
   $252 = ($251|0)!=(0);
   if (!($252)) {
    $0 = 4750;
    break;
   }
   $253 = $name;
   $254 = (_strcmp($253,615083)|0);
   $255 = ($254|0)!=(0);
   if (!($255)) {
    $0 = 4751;
    break;
   }
   $256 = $name;
   $257 = (_strcmp($256,615098)|0);
   $258 = ($257|0)!=(0);
   if (!($258)) {
    $0 = 4752;
    break;
   }
   $259 = $name;
   $260 = (_strcmp($259,613819)|0);
   $261 = ($260|0)!=(0);
   if (!($261)) {
    $0 = 4753;
    break;
   }
   $262 = $name;
   $263 = (_strcmp($262,615109)|0);
   $264 = ($263|0)!=(0);
   if (!($264)) {
    $0 = 4754;
    break;
   }
   $265 = $name;
   $266 = (_strcmp($265,613918)|0);
   $267 = ($266|0)!=(0);
   if (!($267)) {
    $0 = 4755;
    break;
   }
   $268 = $name;
   $269 = (_strcmp($268,613948)|0);
   $270 = ($269|0)!=(0);
   if (!($270)) {
    $0 = 4756;
    break;
   }
   $271 = $name;
   $272 = (_strcmp($271,615125)|0);
   $273 = ($272|0)!=(0);
   if (!($273)) {
    $0 = 4757;
    break;
   }
   $274 = $name;
   $275 = (_strcmp($274,615152)|0);
   $276 = ($275|0)!=(0);
   if (!($276)) {
    $0 = 4758;
    break;
   }
   $277 = $name;
   $278 = (_strcmp($277,615166)|0);
   $279 = ($278|0)!=(0);
   if (!($279)) {
    $0 = 4759;
    break;
   }
   $280 = $name;
   $281 = (_strcmp($280,613854)|0);
   $282 = ($281|0)!=(0);
   if (!($282)) {
    $0 = 4760;
    break;
   }
   $283 = $name;
   $284 = (_strcmp($283,615186)|0);
   $285 = ($284|0)!=(0);
   if (!($285)) {
    $0 = 4761;
    break;
   }
   $286 = $name;
   $287 = (_strcmp($286,615204)|0);
   $288 = ($287|0)!=(0);
   if (!($288)) {
    $0 = 4762;
    break;
   }
   $289 = $name;
   $290 = (_strcmp($289,615216)|0);
   $291 = ($290|0)!=(0);
   if (!($291)) {
    $0 = 4763;
    break;
   }
   $292 = $name;
   $293 = (_strcmp($292,615237)|0);
   $294 = ($293|0)!=(0);
   if (!($294)) {
    $0 = 4764;
    break;
   }
   $295 = $name;
   $296 = (_strcmp($295,615255)|0);
   $297 = ($296|0)!=(0);
   if (!($297)) {
    $0 = 4765;
    break;
   }
   $298 = $name;
   $299 = (_strcmp($298,615273)|0);
   $300 = ($299|0)!=(0);
   if (!($300)) {
    $0 = 4766;
    break;
   }
   $301 = $name;
   $302 = (_strcmp($301,615294)|0);
   $303 = ($302|0)!=(0);
   if (!($303)) {
    $0 = 4767;
    break;
   }
   $304 = $name;
   $305 = (_strcmp($304,615320)|0);
   $306 = ($305|0)!=(0);
   if (!($306)) {
    $0 = 4768;
    break;
   }
   $307 = $name;
   $308 = (_strcmp($307,615343)|0);
   $309 = ($308|0)!=(0);
   if (!($309)) {
    $0 = 4769;
    break;
   }
   $310 = $name;
   $311 = (_strcmp($310,615381)|0);
   $312 = ($311|0)!=(0);
   if (!($312)) {
    $0 = 4770;
    break;
   }
   $313 = $name;
   $314 = (_strcmp($313,615397)|0);
   $315 = ($314|0)!=(0);
   if (!($315)) {
    $0 = 4771;
    break;
   }
   $316 = $name;
   $317 = (_strcmp($316,615412)|0);
   $318 = ($317|0)!=(0);
   if (!($318)) {
    $0 = 4772;
    break;
   }
   $319 = $name;
   $320 = (_strcmp($319,615435)|0);
   $321 = ($320|0)!=(0);
   if (!($321)) {
    $0 = 4773;
    break;
   }
   $322 = $name;
   $323 = (_strcmp($322,615448)|0);
   $324 = ($323|0)!=(0);
   if (!($324)) {
    $0 = 4774;
    break;
   }
   $325 = $name;
   $326 = (_strcmp($325,615462)|0);
   $327 = ($326|0)!=(0);
   if (!($327)) {
    $0 = 4775;
    break;
   }
   $328 = $name;
   $329 = (_strcmp($328,615476)|0);
   $330 = ($329|0)!=(0);
   if (!($330)) {
    $0 = 4776;
    break;
   }
   $331 = $name;
   $332 = (_strcmp($331,615496)|0);
   $333 = ($332|0)!=(0);
   if (!($333)) {
    $0 = 4777;
    break;
   }
   $334 = $name;
   $335 = (_strcmp($334,615516)|0);
   $336 = ($335|0)!=(0);
   if (!($336)) {
    $0 = 4778;
    break;
   }
   $337 = $name;
   $338 = (_strcmp($337,615532)|0);
   $339 = ($338|0)!=(0);
   if (!($339)) {
    $0 = 4779;
    break;
   }
   $340 = $name;
   $341 = (_strcmp($340,615550)|0);
   $342 = ($341|0)!=(0);
   if (!($342)) {
    $0 = 4780;
    break;
   }
   $343 = $name;
   $344 = (_strcmp($343,615566)|0);
   $345 = ($344|0)!=(0);
   if (!($345)) {
    $0 = 4781;
    break;
   }
   $346 = $name;
   $347 = (_strcmp($346,615581)|0);
   $348 = ($347|0)!=(0);
   if (!($348)) {
    $0 = 4782;
    break;
   }
   $349 = $name;
   $350 = (_strcmp($349,615603)|0);
   $351 = ($350|0)!=(0);
   if (!($351)) {
    $0 = 4783;
    break;
   }
   $352 = $name;
   $353 = (_strcmp($352,615621)|0);
   $354 = ($353|0)!=(0);
   if (!($354)) {
    $0 = 4784;
    break;
   }
   $355 = $name;
   $356 = (_strcmp($355,615642)|0);
   $357 = ($356|0)!=(0);
   if (!($357)) {
    $0 = 4785;
    break;
   }
   $358 = $name;
   $359 = (_strcmp($358,615660)|0);
   $360 = ($359|0)!=(0);
   if (!($360)) {
    $0 = 4786;
    break;
   }
   $361 = $name;
   $362 = (_strcmp($361,615673)|0);
   $363 = ($362|0)!=(0);
   if (!($363)) {
    $0 = 4787;
    break;
   }
   $364 = $name;
   $365 = (_strcmp($364,615688)|0);
   $366 = ($365|0)!=(0);
   if (!($366)) {
    $0 = 4788;
    break;
   }
   $367 = $name;
   $368 = (_strcmp($367,615702)|0);
   $369 = ($368|0)!=(0);
   if (!($369)) {
    $0 = 4789;
    break;
   }
   $370 = $name;
   $371 = (_strcmp($370,615712)|0);
   $372 = ($371|0)!=(0);
   if (!($372)) {
    $0 = 4790;
    break;
   }
   $373 = $name;
   $374 = (_strcmp($373,615722)|0);
   $375 = ($374|0)!=(0);
   if (!($375)) {
    $0 = 4791;
    break;
   }
   $376 = $name;
   $377 = (_strcmp($376,615744)|0);
   $378 = ($377|0)!=(0);
   if (!($378)) {
    $0 = 4792;
    break;
   }
   $379 = $name;
   $380 = (_strcmp($379,615770)|0);
   $381 = ($380|0)!=(0);
   if (!($381)) {
    $0 = 4793;
    break;
   }
   $382 = $name;
   $383 = (_strcmp($382,615797)|0);
   $384 = ($383|0)!=(0);
   if (!($384)) {
    $0 = 4794;
    break;
   }
   $385 = $name;
   $386 = (_strcmp($385,615810)|0);
   $387 = ($386|0)!=(0);
   if (!($387)) {
    $0 = 4795;
    break;
   }
   $388 = $name;
   $389 = (_strcmp($388,615825)|0);
   $390 = ($389|0)!=(0);
   if (!($390)) {
    $0 = 4796;
    break;
   }
   $391 = $name;
   $392 = (_strcmp($391,615840)|0);
   $393 = ($392|0)!=(0);
   if (!($393)) {
    $0 = 4797;
    break;
   }
   $394 = $name;
   $395 = (_strcmp($394,615864)|0);
   $396 = ($395|0)!=(0);
   if (!($396)) {
    $0 = 4798;
    break;
   }
   $397 = $name;
   $398 = (_strcmp($397,615875)|0);
   $399 = ($398|0)!=(0);
   if (!($399)) {
    $0 = 4799;
    break;
   }
   $400 = $name;
   $401 = (_strcmp($400,615897)|0);
   $402 = ($401|0)!=(0);
   if (!($402)) {
    $0 = 4800;
    break;
   }
   $403 = $name;
   $404 = (_strcmp($403,615919)|0);
   $405 = ($404|0)!=(0);
   if (!($405)) {
    $0 = 4801;
    break;
   }
   $406 = $name;
   $407 = (_strcmp($406,615943)|0);
   $408 = ($407|0)!=(0);
   if (!($408)) {
    $0 = 4802;
    break;
   }
   $409 = $name;
   $410 = (_strcmp($409,615952)|0);
   $411 = ($410|0)!=(0);
   if (!($411)) {
    $0 = 4803;
    break;
   }
   $412 = $name;
   $413 = (_strcmp($412,615960)|0);
   $414 = ($413|0)!=(0);
   if (!($414)) {
    $0 = 4804;
    break;
   }
   $415 = $name;
   $416 = (_strcmp($415,615973)|0);
   $417 = ($416|0)!=(0);
   if (!($417)) {
    $0 = 4805;
    break;
   }
   $418 = $name;
   $419 = (_strcmp($418,615987)|0);
   $420 = ($419|0)!=(0);
   if (!($420)) {
    $0 = 4806;
    break;
   }
   $421 = $name;
   $422 = (_strcmp($421,615999)|0);
   $423 = ($422|0)!=(0);
   if (!($423)) {
    $0 = 4807;
    break;
   }
   $424 = $name;
   $425 = (_strcmp($424,616008)|0);
   $426 = ($425|0)!=(0);
   if (!($426)) {
    $0 = 4808;
    break;
   }
   $427 = $name;
   $428 = (_strcmp($427,616018)|0);
   $429 = ($428|0)!=(0);
   if (!($429)) {
    $0 = 4809;
    break;
   }
   $430 = $name;
   $431 = (_strcmp($430,616030)|0);
   $432 = ($431|0)!=(0);
   if (!($432)) {
    $0 = 4810;
    break;
   }
   $433 = $name;
   $434 = (_strcmp($433,616041)|0);
   $435 = ($434|0)!=(0);
   if (!($435)) {
    $0 = 4811;
    break;
   }
   $436 = $name;
   $437 = (_strcmp($436,616049)|0);
   $438 = ($437|0)!=(0);
   if (!($438)) {
    $0 = 4812;
    break;
   }
   $439 = $name;
   $440 = (_strcmp($439,616061)|0);
   $441 = ($440|0)!=(0);
   if (!($441)) {
    $0 = 4813;
    break;
   }
   $442 = $name;
   $443 = (_strcmp($442,616076)|0);
   $444 = ($443|0)!=(0);
   if (!($444)) {
    $0 = 4814;
    break;
   }
   $445 = $name;
   $446 = (_strcmp($445,616088)|0);
   $447 = ($446|0)!=(0);
   if (!($447)) {
    $0 = 4815;
    break;
   }
   $448 = $name;
   $449 = (_strcmp($448,616102)|0);
   $450 = ($449|0)!=(0);
   if (!($450)) {
    $0 = 4816;
    break;
   }
   $451 = $name;
   $452 = (_strcmp($451,616127)|0);
   $453 = ($452|0)!=(0);
   if (!($453)) {
    $0 = 4817;
    break;
   }
   $454 = $name;
   $455 = (_strcmp($454,616144)|0);
   $456 = ($455|0)!=(0);
   if (!($456)) {
    $0 = 4818;
    break;
   }
   $457 = $name;
   $458 = (_strcmp($457,616160)|0);
   $459 = ($458|0)!=(0);
   if (!($459)) {
    $0 = 4819;
    break;
   }
   $460 = $name;
   $461 = (_strcmp($460,616176)|0);
   $462 = ($461|0)!=(0);
   if (!($462)) {
    $0 = 4820;
    break;
   }
   $463 = $name;
   $464 = (_strcmp($463,616188)|0);
   $465 = ($464|0)!=(0);
   if (!($465)) {
    $0 = 4821;
    break;
   }
   $466 = $name;
   $467 = (_strcmp($466,616200)|0);
   $468 = ($467|0)!=(0);
   if (!($468)) {
    $0 = 4822;
    break;
   }
   $469 = $name;
   $470 = (_strcmp($469,616224)|0);
   $471 = ($470|0)!=(0);
   if (!($471)) {
    $0 = 4823;
    break;
   }
   $472 = $name;
   $473 = (_strcmp($472,616237)|0);
   $474 = ($473|0)!=(0);
   if (!($474)) {
    $0 = 4824;
    break;
   }
   $475 = $name;
   $476 = (_strcmp($475,616251)|0);
   $477 = ($476|0)!=(0);
   if (!($477)) {
    $0 = 4825;
    break;
   }
   $478 = $name;
   $479 = (_strcmp($478,616273)|0);
   $480 = ($479|0)!=(0);
   if (!($480)) {
    $0 = 4826;
    break;
   }
   $481 = $name;
   $482 = (_strcmp($481,616280)|0);
   $483 = ($482|0)!=(0);
   if (!($483)) {
    $0 = 4827;
    break;
   }
   $484 = $name;
   $485 = (_strcmp($484,616296)|0);
   $486 = ($485|0)!=(0);
   if (!($486)) {
    $0 = 4828;
    break;
   }
   $487 = $name;
   $488 = (_strcmp($487,616313)|0);
   $489 = ($488|0)!=(0);
   if (!($489)) {
    $0 = 4829;
    break;
   }
   $490 = $name;
   $491 = (_strcmp($490,616330)|0);
   $492 = ($491|0)!=(0);
   if (!($492)) {
    $0 = 4830;
    break;
   }
   $493 = $name;
   $494 = (_strcmp($493,616346)|0);
   $495 = ($494|0)!=(0);
   if (!($495)) {
    $0 = 4831;
    break;
   }
   $496 = $name;
   $497 = (_strcmp($496,616362)|0);
   $498 = ($497|0)!=(0);
   if (!($498)) {
    $0 = 4832;
    break;
   }
   $499 = $name;
   $500 = (_strcmp($499,616379)|0);
   $501 = ($500|0)!=(0);
   if (!($501)) {
    $0 = 4833;
    break;
   }
   $502 = $name;
   $503 = (_strcmp($502,616393)|0);
   $504 = ($503|0)!=(0);
   if (!($504)) {
    $0 = 4834;
    break;
   }
   $505 = $name;
   $506 = (_strcmp($505,616405)|0);
   $507 = ($506|0)!=(0);
   if (!($507)) {
    $0 = 4835;
    break;
   }
   $508 = $name;
   $509 = (_strcmp($508,616416)|0);
   $510 = ($509|0)!=(0);
   if (!($510)) {
    $0 = 4836;
    break;
   }
   $511 = $name;
   $512 = (_strcmp($511,616429)|0);
   $513 = ($512|0)!=(0);
   if (!($513)) {
    $0 = 4837;
    break;
   }
   $514 = $name;
   $515 = (_strcmp($514,616439)|0);
   $516 = ($515|0)!=(0);
   if (!($516)) {
    $0 = 4838;
    break;
   }
   $517 = $name;
   $518 = (_strcmp($517,616456)|0);
   $519 = ($518|0)!=(0);
   if (!($519)) {
    $0 = 4839;
    break;
   }
   $520 = $name;
   $521 = (_strcmp($520,616468)|0);
   $522 = ($521|0)!=(0);
   if (!($522)) {
    $0 = 4840;
    break;
   }
   $523 = $name;
   $524 = (_strcmp($523,616490)|0);
   $525 = ($524|0)!=(0);
   if (!($525)) {
    $0 = 4841;
    break;
   }
   $526 = $name;
   $527 = (_strcmp($526,616510)|0);
   $528 = ($527|0)!=(0);
   if (!($528)) {
    $0 = 4842;
    break;
   }
   $529 = $name;
   $530 = (_strcmp($529,616523)|0);
   $531 = ($530|0)!=(0);
   if (!($531)) {
    $0 = 4843;
    break;
   }
   $532 = $name;
   $533 = (_strcmp($532,616545)|0);
   $534 = ($533|0)!=(0);
   if (!($534)) {
    $0 = 4844;
    break;
   }
   $535 = $name;
   $536 = (_strcmp($535,616565)|0);
   $537 = ($536|0)!=(0);
   if (!($537)) {
    $0 = 4845;
    break;
   }
   $538 = $name;
   $539 = (_strcmp($538,616582)|0);
   $540 = ($539|0)!=(0);
   if (!($540)) {
    $0 = 4846;
    break;
   }
   $541 = $name;
   $542 = (_strcmp($541,616599)|0);
   $543 = ($542|0)!=(0);
   if (!($543)) {
    $0 = 4847;
    break;
   }
   $544 = $name;
   $545 = (_strcmp($544,616619)|0);
   $546 = ($545|0)!=(0);
   if ($546) {
    $547 = $1;
    $548 = $name;
    $549 = _emscripten_asm_const_iii(12, ($547|0), ($548|0))|0;
    $0 = 0;
    break;
   } else {
    $0 = 4848;
    break;
   }
  } else {
   $0 = 4681;
  }
 } while(0);
 $550 = $0;
 STACKTOP = sp;return ($550|0);
}
function ___stdio_close($f) {
 $f = $f|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($f)) + 60|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$vararg_buffer>>2] = $1;
 $2 = (___syscall6(6,($vararg_buffer|0))|0);
 $3 = (___syscall_ret($2)|0);
 STACKTOP = sp;return ($3|0);
}
function ___syscall_ret($r) {
 $r = $r|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($r>>>0)>(4294963200);
 if ($0) {
  $1 = (0 - ($r))|0;
  $2 = (___errno_location()|0);
  HEAP32[$2>>2] = $1;
  $$0 = -1;
 } else {
  $$0 = $r;
 }
 return ($$0|0);
}
function ___errno_location() {
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[301488]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  $$0 = 1206000;
 } else {
  $2 = (_pthread_self()|0);
  $3 = ((($2)) + 64|0);
  $4 = HEAP32[$3>>2]|0;
  $$0 = $4;
 }
 return ($$0|0);
}
function ___stdio_write($f,$buf,$len) {
 $f = $f|0;
 $buf = $buf|0;
 $len = $len|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cnt$0 = 0, $cnt$1 = 0, $iov$0 = 0, $iov$0$lcssa11 = 0, $iov$1 = 0, $iovcnt$0 = 0, $iovcnt$0$lcssa12 = 0;
 var $iovcnt$1 = 0, $iovs = 0, $rem$0 = 0, $vararg_buffer = 0, $vararg_buffer3 = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr6 = 0, $vararg_ptr7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $iovs = sp + 32|0;
 $0 = ((($f)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$iovs>>2] = $1;
 $2 = ((($iovs)) + 4|0);
 $3 = ((($f)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) - ($1))|0;
 HEAP32[$2>>2] = $5;
 $6 = ((($iovs)) + 8|0);
 HEAP32[$6>>2] = $buf;
 $7 = ((($iovs)) + 12|0);
 HEAP32[$7>>2] = $len;
 $8 = (($5) + ($len))|0;
 $9 = ((($f)) + 60|0);
 $10 = ((($f)) + 44|0);
 $iov$0 = $iovs;$iovcnt$0 = 2;$rem$0 = $8;
 while(1) {
  $11 = HEAP32[301488]|0;
  $12 = ($11|0)==(0|0);
  if ($12) {
   $16 = HEAP32[$9>>2]|0;
   HEAP32[$vararg_buffer3>>2] = $16;
   $vararg_ptr6 = ((($vararg_buffer3)) + 4|0);
   HEAP32[$vararg_ptr6>>2] = $iov$0;
   $vararg_ptr7 = ((($vararg_buffer3)) + 8|0);
   HEAP32[$vararg_ptr7>>2] = $iovcnt$0;
   $17 = (___syscall146(146,($vararg_buffer3|0))|0);
   $18 = (___syscall_ret($17)|0);
   $cnt$0 = $18;
  } else {
   _pthread_cleanup_push((4849|0),($f|0));
   $13 = HEAP32[$9>>2]|0;
   HEAP32[$vararg_buffer>>2] = $13;
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
   HEAP32[$vararg_ptr1>>2] = $iov$0;
   $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
   HEAP32[$vararg_ptr2>>2] = $iovcnt$0;
   $14 = (___syscall146(146,($vararg_buffer|0))|0);
   $15 = (___syscall_ret($14)|0);
   _pthread_cleanup_pop(0);
   $cnt$0 = $15;
  }
  $19 = ($rem$0|0)==($cnt$0|0);
  if ($19) {
   label = 6;
   break;
  }
  $26 = ($cnt$0|0)<(0);
  if ($26) {
   $iov$0$lcssa11 = $iov$0;$iovcnt$0$lcssa12 = $iovcnt$0;
   label = 8;
   break;
  }
  $34 = (($rem$0) - ($cnt$0))|0;
  $35 = ((($iov$0)) + 4|0);
  $36 = HEAP32[$35>>2]|0;
  $37 = ($cnt$0>>>0)>($36>>>0);
  if ($37) {
   $38 = HEAP32[$10>>2]|0;
   HEAP32[$0>>2] = $38;
   HEAP32[$3>>2] = $38;
   $39 = (($cnt$0) - ($36))|0;
   $40 = ((($iov$0)) + 8|0);
   $41 = (($iovcnt$0) + -1)|0;
   $$phi$trans$insert = ((($iov$0)) + 12|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0;
   $49 = $$pre;$cnt$1 = $39;$iov$1 = $40;$iovcnt$1 = $41;
  } else {
   $42 = ($iovcnt$0|0)==(2);
   if ($42) {
    $43 = HEAP32[$0>>2]|0;
    $44 = (($43) + ($cnt$0)|0);
    HEAP32[$0>>2] = $44;
    $49 = $36;$cnt$1 = $cnt$0;$iov$1 = $iov$0;$iovcnt$1 = 2;
   } else {
    $49 = $36;$cnt$1 = $cnt$0;$iov$1 = $iov$0;$iovcnt$1 = $iovcnt$0;
   }
  }
  $45 = HEAP32[$iov$1>>2]|0;
  $46 = (($45) + ($cnt$1)|0);
  HEAP32[$iov$1>>2] = $46;
  $47 = ((($iov$1)) + 4|0);
  $48 = (($49) - ($cnt$1))|0;
  HEAP32[$47>>2] = $48;
  $iov$0 = $iov$1;$iovcnt$0 = $iovcnt$1;$rem$0 = $34;
 }
 if ((label|0) == 6) {
  $20 = HEAP32[$10>>2]|0;
  $21 = ((($f)) + 48|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = (($20) + ($22)|0);
  $24 = ((($f)) + 16|0);
  HEAP32[$24>>2] = $23;
  $25 = $20;
  HEAP32[$0>>2] = $25;
  HEAP32[$3>>2] = $25;
  $$0 = $len;
 }
 else if ((label|0) == 8) {
  $27 = ((($f)) + 16|0);
  HEAP32[$27>>2] = 0;
  HEAP32[$0>>2] = 0;
  HEAP32[$3>>2] = 0;
  $28 = HEAP32[$f>>2]|0;
  $29 = $28 | 32;
  HEAP32[$f>>2] = $29;
  $30 = ($iovcnt$0$lcssa12|0)==(2);
  if ($30) {
   $$0 = 0;
  } else {
   $31 = ((($iov$0$lcssa11)) + 4|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = (($len) - ($32))|0;
   $$0 = $33;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _cleanup_387($p) {
 $p = $p|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($p)) + 68|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  ___unlockfile($p);
 }
 return;
}
function ___unlockfile($f) {
 $f = $f|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function ___stdio_seek($f,$off,$whence) {
 $f = $f|0;
 $off = $off|0;
 $whence = $whence|0;
 var $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $ret = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0, $vararg_ptr4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $ret = sp + 20|0;
 $0 = ((($f)) + 60|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$vararg_buffer>>2] = $1;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = 0;
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
 HEAP32[$vararg_ptr2>>2] = $off;
 $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
 HEAP32[$vararg_ptr3>>2] = $ret;
 $vararg_ptr4 = ((($vararg_buffer)) + 16|0);
 HEAP32[$vararg_ptr4>>2] = $whence;
 $2 = (___syscall140(140,($vararg_buffer|0))|0);
 $3 = (___syscall_ret($2)|0);
 $4 = ($3|0)<(0);
 if ($4) {
  HEAP32[$ret>>2] = -1;
  $5 = -1;
 } else {
  $$pre = HEAP32[$ret>>2]|0;
  $5 = $$pre;
 }
 STACKTOP = sp;return ($5|0);
}
function ___stdio_read($f,$buf,$len) {
 $f = $f|0;
 $buf = $buf|0;
 $len = $len|0;
 var $$0 = 0, $$cast = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $cnt$0 = 0, $iov = 0, $vararg_buffer = 0, $vararg_buffer3 = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr6 = 0, $vararg_ptr7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $iov = sp + 32|0;
 HEAP32[$iov>>2] = $buf;
 $0 = ((($iov)) + 4|0);
 $1 = ((($f)) + 48|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)!=(0);
 $4 = $3&1;
 $5 = (($len) - ($4))|0;
 HEAP32[$0>>2] = $5;
 $6 = ((($iov)) + 8|0);
 $7 = ((($f)) + 44|0);
 $8 = HEAP32[$7>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = ((($iov)) + 12|0);
 HEAP32[$9>>2] = $2;
 $10 = HEAP32[301488]|0;
 $11 = ($10|0)==(0|0);
 if ($11) {
  $16 = ((($f)) + 60|0);
  $17 = HEAP32[$16>>2]|0;
  HEAP32[$vararg_buffer3>>2] = $17;
  $vararg_ptr6 = ((($vararg_buffer3)) + 4|0);
  HEAP32[$vararg_ptr6>>2] = $iov;
  $vararg_ptr7 = ((($vararg_buffer3)) + 8|0);
  HEAP32[$vararg_ptr7>>2] = 2;
  $18 = (___syscall145(145,($vararg_buffer3|0))|0);
  $19 = (___syscall_ret($18)|0);
  $cnt$0 = $19;
 } else {
  _pthread_cleanup_push((4850|0),($f|0));
  $12 = ((($f)) + 60|0);
  $13 = HEAP32[$12>>2]|0;
  HEAP32[$vararg_buffer>>2] = $13;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $iov;
  $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
  HEAP32[$vararg_ptr2>>2] = 2;
  $14 = (___syscall145(145,($vararg_buffer|0))|0);
  $15 = (___syscall_ret($14)|0);
  _pthread_cleanup_pop(0);
  $cnt$0 = $15;
 }
 $20 = ($cnt$0|0)<(1);
 if ($20) {
  $21 = $cnt$0 & 48;
  $22 = $21 ^ 16;
  $23 = HEAP32[$f>>2]|0;
  $24 = $23 | $22;
  HEAP32[$f>>2] = $24;
  $25 = ((($f)) + 8|0);
  HEAP32[$25>>2] = 0;
  $26 = ((($f)) + 4|0);
  HEAP32[$26>>2] = 0;
  $$0 = $cnt$0;
 } else {
  $27 = HEAP32[$0>>2]|0;
  $28 = ($cnt$0>>>0)>($27>>>0);
  if ($28) {
   $29 = (($cnt$0) - ($27))|0;
   $30 = HEAP32[$7>>2]|0;
   $31 = ((($f)) + 4|0);
   HEAP32[$31>>2] = $30;
   $$cast = $30;
   $32 = (($$cast) + ($29)|0);
   $33 = ((($f)) + 8|0);
   HEAP32[$33>>2] = $32;
   $34 = HEAP32[$1>>2]|0;
   $35 = ($34|0)==(0);
   if ($35) {
    $$0 = $len;
   } else {
    $36 = ((($$cast)) + 1|0);
    HEAP32[$31>>2] = $36;
    $37 = HEAP8[$$cast>>0]|0;
    $38 = (($len) + -1)|0;
    $39 = (($buf) + ($38)|0);
    HEAP8[$39>>0] = $37;
    $$0 = $len;
   }
  } else {
   $$0 = $cnt$0;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _cleanup_382($p) {
 $p = $p|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($p)) + 68|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  ___unlockfile($p);
 }
 return;
}
function ___stdout_write($f,$buf,$len) {
 $f = $f|0;
 $buf = $buf|0;
 $len = $len|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $tio = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $tio = sp + 12|0;
 $0 = ((($f)) + 36|0);
 HEAP32[$0>>2] = 3003;
 $1 = HEAP32[$f>>2]|0;
 $2 = $1 & 64;
 $3 = ($2|0)==(0);
 if ($3) {
  $4 = ((($f)) + 60|0);
  $5 = HEAP32[$4>>2]|0;
  HEAP32[$vararg_buffer>>2] = $5;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = 21505;
  $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
  HEAP32[$vararg_ptr2>>2] = $tio;
  $6 = (___syscall54(54,($vararg_buffer|0))|0);
  $7 = ($6|0)==(0);
  if (!($7)) {
   $8 = ((($f)) + 75|0);
   HEAP8[$8>>0] = -1;
  }
 }
 $9 = (___stdio_write($f,$buf,$len)|0);
 STACKTOP = sp;return ($9|0);
}
function _tolower($c) {
 $c = $c|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_isupper($c)|0);
 $1 = ($0|0)==(0);
 $2 = $c | 32;
 $$0 = $1 ? $c : $2;
 return ($$0|0);
}
function _isupper($c) {
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($c) + -65)|0;
 $1 = ($0>>>0)<(26);
 $2 = $1&1;
 return ($2|0);
}
function _strtoull_l($s,$p,$base,$loc) {
 $s = $s|0;
 $p = $p|0;
 $base = $base|0;
 $loc = $loc|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strtoull($s,$p,$base)|0);
 $1 = tempRet0;
 tempRet0 = ($1);
 return ($0|0);
}
function _strtoull($s,$p,$base) {
 $s = $s|0;
 $p = $p|0;
 $base = $base|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strtox_519($s,$p,$base,-1,-1)|0);
 $1 = tempRet0;
 tempRet0 = ($1);
 return ($0|0);
}
function _strtox_519($s,$p,$base,$0,$1) {
 $s = $s|0;
 $p = $p|0;
 $base = $base|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$sink = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f = sp;
 HEAP32[$f>>2] = 0;
 $2 = ((($f)) + 4|0);
 HEAP32[$2>>2] = $s;
 $3 = ((($f)) + 44|0);
 HEAP32[$3>>2] = $s;
 $4 = ($s|0)<(0|0);
 $5 = ((($s)) + 2147483647|0);
 $$sink = $4 ? (-1) : $5;
 $6 = ((($f)) + 8|0);
 HEAP32[$6>>2] = $$sink;
 $7 = ((($f)) + 76|0);
 HEAP32[$7>>2] = -1;
 ___shlim($f,0);
 $8 = (___intscan($f,$base,1,$0,$1)|0);
 $9 = tempRet0;
 $10 = ($p|0)==(0|0);
 if (!($10)) {
  $11 = ((($f)) + 108|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = HEAP32[$2>>2]|0;
  $14 = HEAP32[$6>>2]|0;
  $15 = (($13) + ($12))|0;
  $16 = (($15) - ($14))|0;
  $17 = (($s) + ($16)|0);
  HEAP32[$p>>2] = $17;
 }
 tempRet0 = ($9);
 STACKTOP = sp;return ($8|0);
}
function ___shlim($f,$lim) {
 $f = $f|0;
 $lim = $lim|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 104|0);
 HEAP32[$0>>2] = $lim;
 $1 = ((($f)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($f)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2) - ($4))|0;
 $6 = ((($f)) + 108|0);
 HEAP32[$6>>2] = $5;
 $7 = ($lim|0)!=(0);
 $8 = ($5|0)>($lim|0);
 $or$cond = $7 & $8;
 if ($or$cond) {
  $9 = $4;
  $10 = (($9) + ($lim)|0);
  $11 = ((($f)) + 100|0);
  HEAP32[$11>>2] = $10;
 } else {
  $12 = ((($f)) + 100|0);
  HEAP32[$12>>2] = $2;
 }
 return;
}
function ___intscan($f,$base,$pok,$0,$1) {
 $f = $f|0;
 $base = $base|0;
 $pok = $pok|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$1 = 0, $$115 = 0, $$116 = 0, $$base14 = 0, $$lcssa = 0, $$lcssa108 = 0, $$lcssa109 = 0, $$lcssa110 = 0, $$lcssa111 = 0, $$lcssa112 = 0, $$lcssa113 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0;
 var $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0;
 var $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0;
 var $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0;
 var $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0;
 var $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0;
 var $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $c$0 = 0, $c$1 = 0, $c$117 = 0, $c$2$be = 0, $c$2$be$lcssa = 0, $c$2$lcssa = 0, $c$3$be = 0, $c$3$lcssa = 0, $c$359 = 0, $c$4$be = 0, $c$4$be$lcssa = 0, $c$4$lcssa = 0, $c$5$be = 0, $c$6$be = 0, $c$6$be$lcssa = 0;
 var $c$6$lcssa = 0, $c$7$be = 0, $c$742 = 0, $c$8 = 0, $c$9$be = 0, $neg$0 = 0, $neg$0$ = 0, $neg$1 = 0, $or$cond = 0, $or$cond12 = 0, $or$cond31 = 0, $or$cond5 = 0, $or$cond7 = 0, $x$070 = 0, $x$136 = 0, $x$254 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($base>>>0)>(36);
 L1: do {
  if ($2) {
   $5 = (___errno_location()|0);
   HEAP32[$5>>2] = 22;
   $286 = 0;$287 = 0;
  } else {
   $3 = ((($f)) + 4|0);
   $4 = ((($f)) + 100|0);
   while(1) {
    $6 = HEAP32[$3>>2]|0;
    $7 = HEAP32[$4>>2]|0;
    $8 = ($6>>>0)<($7>>>0);
    if ($8) {
     $9 = ((($6)) + 1|0);
     HEAP32[$3>>2] = $9;
     $10 = HEAP8[$6>>0]|0;
     $11 = $10&255;
     $13 = $11;
    } else {
     $12 = (___shgetc($f)|0);
     $13 = $12;
    }
    $14 = (_isspace($13)|0);
    $15 = ($14|0)==(0);
    if ($15) {
     $$lcssa113 = $13;
     break;
    }
   }
   $16 = ($$lcssa113|0)==(45);
   L11: do {
    switch ($$lcssa113|0) {
    case 43: case 45:  {
     $17 = $16 << 31 >> 31;
     $18 = HEAP32[$3>>2]|0;
     $19 = HEAP32[$4>>2]|0;
     $20 = ($18>>>0)<($19>>>0);
     if ($20) {
      $21 = ((($18)) + 1|0);
      HEAP32[$3>>2] = $21;
      $22 = HEAP8[$18>>0]|0;
      $23 = $22&255;
      $c$0 = $23;$neg$0 = $17;
      break L11;
     } else {
      $24 = (___shgetc($f)|0);
      $c$0 = $24;$neg$0 = $17;
      break L11;
     }
     break;
    }
    default: {
     $c$0 = $$lcssa113;$neg$0 = 0;
    }
    }
   } while(0);
   $25 = ($base|0)==(0);
   $26 = $base | 16;
   $27 = ($26|0)==(16);
   $28 = ($c$0|0)==(48);
   $or$cond5 = $27 & $28;
   do {
    if ($or$cond5) {
     $29 = HEAP32[$3>>2]|0;
     $30 = HEAP32[$4>>2]|0;
     $31 = ($29>>>0)<($30>>>0);
     if ($31) {
      $32 = ((($29)) + 1|0);
      HEAP32[$3>>2] = $32;
      $33 = HEAP8[$29>>0]|0;
      $34 = $33&255;
      $37 = $34;
     } else {
      $35 = (___shgetc($f)|0);
      $37 = $35;
     }
     $36 = $37 | 32;
     $38 = ($36|0)==(120);
     if (!($38)) {
      if ($25) {
       $$116 = 8;$c$117 = $37;
       label = 46;
       break;
      } else {
       $$1 = $base;$c$1 = $37;
       label = 32;
       break;
      }
     }
     $39 = HEAP32[$3>>2]|0;
     $40 = HEAP32[$4>>2]|0;
     $41 = ($39>>>0)<($40>>>0);
     if ($41) {
      $42 = ((($39)) + 1|0);
      HEAP32[$3>>2] = $42;
      $43 = HEAP8[$39>>0]|0;
      $44 = $43&255;
      $47 = $44;
     } else {
      $45 = (___shgetc($f)|0);
      $47 = $45;
     }
     $46 = ((616736) + ($47)|0);
     $48 = HEAP8[$46>>0]|0;
     $49 = ($48&255)>(15);
     if ($49) {
      $50 = HEAP32[$4>>2]|0;
      $51 = ($50|0)==(0|0);
      if (!($51)) {
       $52 = HEAP32[$3>>2]|0;
       $53 = ((($52)) + -1|0);
       HEAP32[$3>>2] = $53;
      }
      $54 = ($pok|0)==(0);
      if ($54) {
       ___shlim($f,0);
       $286 = 0;$287 = 0;
       break L1;
      }
      if ($51) {
       $286 = 0;$287 = 0;
       break L1;
      }
      $55 = HEAP32[$3>>2]|0;
      $56 = ((($55)) + -1|0);
      HEAP32[$3>>2] = $56;
      $286 = 0;$287 = 0;
      break L1;
     } else {
      $$116 = 16;$c$117 = $47;
      label = 46;
     }
    } else {
     $$base14 = $25 ? 10 : $base;
     $57 = ((616736) + ($c$0)|0);
     $58 = HEAP8[$57>>0]|0;
     $59 = $58&255;
     $60 = ($59>>>0)<($$base14>>>0);
     if ($60) {
      $$1 = $$base14;$c$1 = $c$0;
      label = 32;
     } else {
      $61 = HEAP32[$4>>2]|0;
      $62 = ($61|0)==(0|0);
      if (!($62)) {
       $63 = HEAP32[$3>>2]|0;
       $64 = ((($63)) + -1|0);
       HEAP32[$3>>2] = $64;
      }
      ___shlim($f,0);
      $65 = (___errno_location()|0);
      HEAP32[$65>>2] = 22;
      $286 = 0;$287 = 0;
      break L1;
     }
    }
   } while(0);
   if ((label|0) == 32) {
    $66 = ($$1|0)==(10);
    if ($66) {
     $67 = (($c$1) + -48)|0;
     $68 = ($67>>>0)<(10);
     if ($68) {
      $71 = $67;$x$070 = 0;
      while(1) {
       $69 = ($x$070*10)|0;
       $70 = (($69) + ($71))|0;
       $72 = HEAP32[$3>>2]|0;
       $73 = HEAP32[$4>>2]|0;
       $74 = ($72>>>0)<($73>>>0);
       if ($74) {
        $75 = ((($72)) + 1|0);
        HEAP32[$3>>2] = $75;
        $76 = HEAP8[$72>>0]|0;
        $77 = $76&255;
        $c$2$be = $77;
       } else {
        $78 = (___shgetc($f)|0);
        $c$2$be = $78;
       }
       $79 = (($c$2$be) + -48)|0;
       $80 = ($79>>>0)<(10);
       $81 = ($70>>>0)<(429496729);
       $82 = $80 & $81;
       if ($82) {
        $71 = $79;$x$070 = $70;
       } else {
        $$lcssa112 = $70;$c$2$be$lcssa = $c$2$be;
        break;
       }
      }
      $288 = $$lcssa112;$289 = 0;$c$2$lcssa = $c$2$be$lcssa;
     } else {
      $288 = 0;$289 = 0;$c$2$lcssa = $c$1;
     }
     $83 = (($c$2$lcssa) + -48)|0;
     $84 = ($83>>>0)<(10);
     if ($84) {
      $85 = $288;$86 = $289;$90 = $83;$c$359 = $c$2$lcssa;
      while(1) {
       $87 = (___muldi3(($85|0),($86|0),10,0)|0);
       $88 = tempRet0;
       $89 = ($90|0)<(0);
       $91 = $89 << 31 >> 31;
       $92 = $90 ^ -1;
       $93 = $91 ^ -1;
       $94 = ($88>>>0)>($93>>>0);
       $95 = ($87>>>0)>($92>>>0);
       $96 = ($88|0)==($93|0);
       $97 = $96 & $95;
       $98 = $94 | $97;
       if ($98) {
        $$lcssa = $90;$290 = $85;$291 = $86;$c$3$lcssa = $c$359;
        break;
       }
       $99 = (_i64Add(($87|0),($88|0),($90|0),($91|0))|0);
       $100 = tempRet0;
       $101 = HEAP32[$3>>2]|0;
       $102 = HEAP32[$4>>2]|0;
       $103 = ($101>>>0)<($102>>>0);
       if ($103) {
        $104 = ((($101)) + 1|0);
        HEAP32[$3>>2] = $104;
        $105 = HEAP8[$101>>0]|0;
        $106 = $105&255;
        $c$3$be = $106;
       } else {
        $107 = (___shgetc($f)|0);
        $c$3$be = $107;
       }
       $108 = (($c$3$be) + -48)|0;
       $109 = ($108>>>0)<(10);
       $110 = ($100>>>0)<(429496729);
       $111 = ($99>>>0)<(2576980378);
       $112 = ($100|0)==(429496729);
       $113 = $112 & $111;
       $114 = $110 | $113;
       $or$cond7 = $109 & $114;
       if ($or$cond7) {
        $85 = $99;$86 = $100;$90 = $108;$c$359 = $c$3$be;
       } else {
        $$lcssa = $108;$290 = $99;$291 = $100;$c$3$lcssa = $c$3$be;
        break;
       }
      }
      $115 = ($$lcssa>>>0)>(9);
      if ($115) {
       $260 = $291;$262 = $290;$neg$1 = $neg$0;
      } else {
       $$115 = 10;$292 = $290;$293 = $291;$c$8 = $c$3$lcssa;
       label = 72;
      }
     } else {
      $260 = $289;$262 = $288;$neg$1 = $neg$0;
     }
    } else {
     $$116 = $$1;$c$117 = $c$1;
     label = 46;
    }
   }
   L63: do {
    if ((label|0) == 46) {
     $116 = (($$116) + -1)|0;
     $117 = $116 & $$116;
     $118 = ($117|0)==(0);
     if ($118) {
      $123 = ($$116*23)|0;
      $124 = $123 >>> 5;
      $125 = $124 & 7;
      $126 = (616992 + ($125)|0);
      $127 = HEAP8[$126>>0]|0;
      $128 = $127 << 24 >> 24;
      $129 = ((616736) + ($c$117)|0);
      $130 = HEAP8[$129>>0]|0;
      $131 = $130&255;
      $132 = ($131>>>0)<($$116>>>0);
      if ($132) {
       $135 = $131;$x$136 = 0;
       while(1) {
        $133 = $x$136 << $128;
        $134 = $135 | $133;
        $136 = HEAP32[$3>>2]|0;
        $137 = HEAP32[$4>>2]|0;
        $138 = ($136>>>0)<($137>>>0);
        if ($138) {
         $139 = ((($136)) + 1|0);
         HEAP32[$3>>2] = $139;
         $140 = HEAP8[$136>>0]|0;
         $141 = $140&255;
         $c$4$be = $141;
        } else {
         $142 = (___shgetc($f)|0);
         $c$4$be = $142;
        }
        $143 = ((616736) + ($c$4$be)|0);
        $144 = HEAP8[$143>>0]|0;
        $145 = $144&255;
        $146 = ($145>>>0)<($$116>>>0);
        $147 = ($134>>>0)<(134217728);
        $148 = $147 & $146;
        if ($148) {
         $135 = $145;$x$136 = $134;
        } else {
         $$lcssa108 = $134;$$lcssa109 = $144;$c$4$be$lcssa = $c$4$be;
         break;
        }
       }
       $152 = $$lcssa109;$155 = 0;$157 = $$lcssa108;$c$4$lcssa = $c$4$be$lcssa;
      } else {
       $152 = $130;$155 = 0;$157 = 0;$c$4$lcssa = $c$117;
      }
      $149 = (_bitshift64Lshr(-1,-1,($128|0))|0);
      $150 = tempRet0;
      $151 = $152&255;
      $153 = ($151>>>0)>=($$116>>>0);
      $154 = ($155>>>0)>($150>>>0);
      $156 = ($157>>>0)>($149>>>0);
      $158 = ($155|0)==($150|0);
      $159 = $158 & $156;
      $160 = $154 | $159;
      $or$cond31 = $153 | $160;
      if ($or$cond31) {
       $$115 = $$116;$292 = $157;$293 = $155;$c$8 = $c$4$lcssa;
       label = 72;
       break;
      } else {
       $161 = $157;$162 = $155;$166 = $152;
      }
      while(1) {
       $163 = (_bitshift64Shl(($161|0),($162|0),($128|0))|0);
       $164 = tempRet0;
       $165 = $166&255;
       $167 = $165 | $163;
       $168 = HEAP32[$3>>2]|0;
       $169 = HEAP32[$4>>2]|0;
       $170 = ($168>>>0)<($169>>>0);
       if ($170) {
        $171 = ((($168)) + 1|0);
        HEAP32[$3>>2] = $171;
        $172 = HEAP8[$168>>0]|0;
        $173 = $172&255;
        $c$5$be = $173;
       } else {
        $174 = (___shgetc($f)|0);
        $c$5$be = $174;
       }
       $175 = ((616736) + ($c$5$be)|0);
       $176 = HEAP8[$175>>0]|0;
       $177 = $176&255;
       $178 = ($177>>>0)>=($$116>>>0);
       $179 = ($164>>>0)>($150>>>0);
       $180 = ($167>>>0)>($149>>>0);
       $181 = ($164|0)==($150|0);
       $182 = $181 & $180;
       $183 = $179 | $182;
       $or$cond = $178 | $183;
       if ($or$cond) {
        $$115 = $$116;$292 = $167;$293 = $164;$c$8 = $c$5$be;
        label = 72;
        break L63;
       } else {
        $161 = $167;$162 = $164;$166 = $176;
       }
      }
     }
     $119 = ((616736) + ($c$117)|0);
     $120 = HEAP8[$119>>0]|0;
     $121 = $120&255;
     $122 = ($121>>>0)<($$116>>>0);
     if ($122) {
      $186 = $121;$x$254 = 0;
      while(1) {
       $184 = Math_imul($x$254, $$116)|0;
       $185 = (($186) + ($184))|0;
       $187 = HEAP32[$3>>2]|0;
       $188 = HEAP32[$4>>2]|0;
       $189 = ($187>>>0)<($188>>>0);
       if ($189) {
        $190 = ((($187)) + 1|0);
        HEAP32[$3>>2] = $190;
        $191 = HEAP8[$187>>0]|0;
        $192 = $191&255;
        $c$6$be = $192;
       } else {
        $193 = (___shgetc($f)|0);
        $c$6$be = $193;
       }
       $194 = ((616736) + ($c$6$be)|0);
       $195 = HEAP8[$194>>0]|0;
       $196 = $195&255;
       $197 = ($196>>>0)<($$116>>>0);
       $198 = ($185>>>0)<(119304647);
       $199 = $198 & $197;
       if ($199) {
        $186 = $196;$x$254 = $185;
       } else {
        $$lcssa110 = $185;$$lcssa111 = $195;$c$6$be$lcssa = $c$6$be;
        break;
       }
      }
      $201 = $$lcssa111;$294 = $$lcssa110;$295 = 0;$c$6$lcssa = $c$6$be$lcssa;
     } else {
      $201 = $120;$294 = 0;$295 = 0;$c$6$lcssa = $c$117;
     }
     $200 = $201&255;
     $202 = ($200>>>0)<($$116>>>0);
     if ($202) {
      $203 = (___udivdi3(-1,-1,($$116|0),0)|0);
      $204 = tempRet0;
      $206 = $295;$208 = $294;$215 = $201;$c$742 = $c$6$lcssa;
      while(1) {
       $205 = ($206>>>0)>($204>>>0);
       $207 = ($208>>>0)>($203>>>0);
       $209 = ($206|0)==($204|0);
       $210 = $209 & $207;
       $211 = $205 | $210;
       if ($211) {
        $$115 = $$116;$292 = $208;$293 = $206;$c$8 = $c$742;
        label = 72;
        break L63;
       }
       $212 = (___muldi3(($208|0),($206|0),($$116|0),0)|0);
       $213 = tempRet0;
       $214 = $215&255;
       $216 = $214 ^ -1;
       $217 = ($213>>>0)>(4294967295);
       $218 = ($212>>>0)>($216>>>0);
       $219 = ($213|0)==(-1);
       $220 = $219 & $218;
       $221 = $217 | $220;
       if ($221) {
        $$115 = $$116;$292 = $208;$293 = $206;$c$8 = $c$742;
        label = 72;
        break L63;
       }
       $222 = (_i64Add(($214|0),0,($212|0),($213|0))|0);
       $223 = tempRet0;
       $224 = HEAP32[$3>>2]|0;
       $225 = HEAP32[$4>>2]|0;
       $226 = ($224>>>0)<($225>>>0);
       if ($226) {
        $227 = ((($224)) + 1|0);
        HEAP32[$3>>2] = $227;
        $228 = HEAP8[$224>>0]|0;
        $229 = $228&255;
        $c$7$be = $229;
       } else {
        $230 = (___shgetc($f)|0);
        $c$7$be = $230;
       }
       $231 = ((616736) + ($c$7$be)|0);
       $232 = HEAP8[$231>>0]|0;
       $233 = $232&255;
       $234 = ($233>>>0)<($$116>>>0);
       if ($234) {
        $206 = $223;$208 = $222;$215 = $232;$c$742 = $c$7$be;
       } else {
        $$115 = $$116;$292 = $222;$293 = $223;$c$8 = $c$7$be;
        label = 72;
        break;
       }
      }
     } else {
      $$115 = $$116;$292 = $294;$293 = $295;$c$8 = $c$6$lcssa;
      label = 72;
     }
    }
   } while(0);
   if ((label|0) == 72) {
    $235 = ((616736) + ($c$8)|0);
    $236 = HEAP8[$235>>0]|0;
    $237 = $236&255;
    $238 = ($237>>>0)<($$115>>>0);
    if ($238) {
     while(1) {
      $239 = HEAP32[$3>>2]|0;
      $240 = HEAP32[$4>>2]|0;
      $241 = ($239>>>0)<($240>>>0);
      if ($241) {
       $242 = ((($239)) + 1|0);
       HEAP32[$3>>2] = $242;
       $243 = HEAP8[$239>>0]|0;
       $244 = $243&255;
       $c$9$be = $244;
      } else {
       $245 = (___shgetc($f)|0);
       $c$9$be = $245;
      }
      $246 = ((616736) + ($c$9$be)|0);
      $247 = HEAP8[$246>>0]|0;
      $248 = $247&255;
      $249 = ($248>>>0)<($$115>>>0);
      if (!($249)) {
       break;
      }
     }
     $250 = (___errno_location()|0);
     HEAP32[$250>>2] = 34;
     $251 = $0 & 1;
     $252 = ($251|0)==(0);
     $253 = (0)==(0);
     $254 = $252 & $253;
     $neg$0$ = $254 ? $neg$0 : 0;
     $260 = $1;$262 = $0;$neg$1 = $neg$0$;
    } else {
     $260 = $293;$262 = $292;$neg$1 = $neg$0;
    }
   }
   $255 = HEAP32[$4>>2]|0;
   $256 = ($255|0)==(0|0);
   if (!($256)) {
    $257 = HEAP32[$3>>2]|0;
    $258 = ((($257)) + -1|0);
    HEAP32[$3>>2] = $258;
   }
   $259 = ($260>>>0)<($1>>>0);
   $261 = ($262>>>0)<($0>>>0);
   $263 = ($260|0)==($1|0);
   $264 = $263 & $261;
   $265 = $259 | $264;
   if (!($265)) {
    $266 = $0 & 1;
    $267 = ($266|0)!=(0);
    $268 = (0)!=(0);
    $269 = $267 | $268;
    $270 = ($neg$1|0)!=(0);
    $or$cond12 = $269 | $270;
    if (!($or$cond12)) {
     $271 = (___errno_location()|0);
     HEAP32[$271>>2] = 34;
     $272 = (_i64Add(($0|0),($1|0),-1,-1)|0);
     $273 = tempRet0;
     $286 = $273;$287 = $272;
     break;
    }
    $274 = ($260>>>0)>($1>>>0);
    $275 = ($262>>>0)>($0>>>0);
    $276 = ($260|0)==($1|0);
    $277 = $276 & $275;
    $278 = $274 | $277;
    if ($278) {
     $279 = (___errno_location()|0);
     HEAP32[$279>>2] = 34;
     $286 = $1;$287 = $0;
     break;
    }
   }
   $280 = ($neg$1|0)<(0);
   $281 = $280 << 31 >> 31;
   $282 = $262 ^ $neg$1;
   $283 = $260 ^ $281;
   $284 = (_i64Subtract(($282|0),($283|0),($neg$1|0),($281|0))|0);
   $285 = tempRet0;
   $286 = $285;$287 = $284;
  }
 } while(0);
 tempRet0 = ($286);
 return ($287|0);
}
function ___shgetc($f) {
 $f = $f|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$phi$trans$insert3 = 0, $$pre = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 104|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  label = 3;
 } else {
  $3 = ((($f)) + 108|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ($4|0)<($1|0);
  if ($5) {
   label = 3;
  } else {
   label = 4;
  }
 }
 if ((label|0) == 3) {
  $6 = (___uflow($f)|0);
  $7 = ($6|0)<(0);
  if ($7) {
   label = 4;
  } else {
   $9 = HEAP32[$0>>2]|0;
   $10 = ($9|0)==(0);
   $$phi$trans$insert = ((($f)) + 8|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0;
   if ($10) {
    $11 = $$pre;
    $41 = $11;
    label = 9;
   } else {
    $12 = ((($f)) + 4|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = $13;
    $15 = (($$pre) - ($14))|0;
    $16 = ((($f)) + 108|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = (($9) - ($17))|0;
    $19 = ($15|0)<($18|0);
    $20 = $$pre;
    if ($19) {
     $41 = $20;
     label = 9;
    } else {
     $21 = (($18) + -1)|0;
     $22 = (($13) + ($21)|0);
     $23 = ((($f)) + 100|0);
     HEAP32[$23>>2] = $22;
     $26 = $20;
    }
   }
   if ((label|0) == 9) {
    $24 = ((($f)) + 100|0);
    HEAP32[$24>>2] = $$pre;
    $26 = $41;
   }
   $25 = ($26|0)==(0|0);
   $$phi$trans$insert3 = ((($f)) + 4|0);
   if ($25) {
    $$pre4 = HEAP32[$$phi$trans$insert3>>2]|0;
    $36 = $$pre4;
   } else {
    $27 = HEAP32[$$phi$trans$insert3>>2]|0;
    $28 = $26;
    $29 = ((($f)) + 108|0);
    $30 = HEAP32[$29>>2]|0;
    $31 = (($28) + 1)|0;
    $32 = (($31) - ($27))|0;
    $33 = (($32) + ($30))|0;
    HEAP32[$29>>2] = $33;
    $34 = $27;
    $36 = $34;
   }
   $35 = ((($36)) + -1|0);
   $37 = HEAP8[$35>>0]|0;
   $38 = $37&255;
   $39 = ($38|0)==($6|0);
   if ($39) {
    $$0 = $6;
   } else {
    $40 = $6&255;
    HEAP8[$35>>0] = $40;
    $$0 = $6;
   }
  }
 }
 if ((label|0) == 4) {
  $8 = ((($f)) + 100|0);
  HEAP32[$8>>2] = 0;
  $$0 = -1;
 }
 return ($$0|0);
}
function ___uflow($f) {
 $f = $f|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $c = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $c = sp;
 $0 = ((($f)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $3 = (___toread($f)|0);
  $4 = ($3|0)==(0);
  if ($4) {
   label = 3;
  } else {
   $$0 = -1;
  }
 } else {
  label = 3;
 }
 if ((label|0) == 3) {
  $5 = ((($f)) + 32|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (FUNCTION_TABLE_iiii[$6 & 8191]($f,$c,1)|0);
  $8 = ($7|0)==(1);
  if ($8) {
   $9 = HEAP8[$c>>0]|0;
   $10 = $9&255;
   $$0 = $10;
  } else {
   $$0 = -1;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function ___toread($f) {
 $f = $f|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 74|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = $1 << 24 >> 24;
 $3 = (($2) + 255)|0;
 $4 = $3 | $2;
 $5 = $4&255;
 HEAP8[$0>>0] = $5;
 $6 = ((($f)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($f)) + 44|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($7>>>0)>($9>>>0);
 if ($10) {
  $11 = ((($f)) + 36|0);
  $12 = HEAP32[$11>>2]|0;
  (FUNCTION_TABLE_iiii[$12 & 8191]($f,0,0)|0);
 }
 $13 = ((($f)) + 16|0);
 HEAP32[$13>>2] = 0;
 $14 = ((($f)) + 28|0);
 HEAP32[$14>>2] = 0;
 HEAP32[$6>>2] = 0;
 $15 = HEAP32[$f>>2]|0;
 $16 = $15 & 20;
 $17 = ($16|0)==(0);
 if ($17) {
  $21 = HEAP32[$8>>2]|0;
  $22 = ((($f)) + 8|0);
  HEAP32[$22>>2] = $21;
  $23 = ((($f)) + 4|0);
  HEAP32[$23>>2] = $21;
  $$0 = 0;
 } else {
  $18 = $15 & 4;
  $19 = ($18|0)==(0);
  if ($19) {
   $$0 = -1;
  } else {
   $20 = $15 | 32;
   HEAP32[$f>>2] = $20;
   $$0 = -1;
  }
 }
 return ($$0|0);
}
function _isspace($c) {
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($c|0)==(32);
 $1 = (($c) + -9)|0;
 $2 = ($1>>>0)<(5);
 $3 = $0 | $2;
 $4 = $3&1;
 return ($4|0);
}
function _strtoll_l($s,$p,$base,$loc) {
 $s = $s|0;
 $p = $p|0;
 $base = $base|0;
 $loc = $loc|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strtoll($s,$p,$base)|0);
 $1 = tempRet0;
 tempRet0 = ($1);
 return ($0|0);
}
function _strtoll($s,$p,$base) {
 $s = $s|0;
 $p = $p|0;
 $base = $base|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strtox_519($s,$p,$base,0,-2147483648)|0);
 $1 = tempRet0;
 tempRet0 = ($1);
 return ($0|0);
}
function _strtoul($s,$p,$base) {
 $s = $s|0;
 $p = $p|0;
 $base = $base|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strtox_519($s,$p,$base,-1,0)|0);
 $1 = tempRet0;
 return ($0|0);
}
function _strtol($s,$p,$base) {
 $s = $s|0;
 $p = $p|0;
 $base = $base|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strtox_519($s,$p,$base,-2147483648,0)|0);
 $1 = tempRet0;
 return ($0|0);
}
function _toupper($c) {
 $c = $c|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_islower($c)|0);
 $1 = ($0|0)==(0);
 $2 = $c & 95;
 $$0 = $1 ? $c : $2;
 return ($$0|0);
}
function _islower($c) {
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($c) + -97)|0;
 $1 = ($0>>>0)<(26);
 $2 = $1&1;
 return ($2|0);
}
function _copysign($x,$y) {
 $x = +$x;
 $y = +$y;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $fabs = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAPF64[tempDoublePtr>>3] = $y;$0 = HEAP32[tempDoublePtr>>2]|0;
 $1 = HEAP32[tempDoublePtr+4>>2]|0;
 $fabs = (+Math_abs((+$x)));
 HEAPF64[tempDoublePtr>>3] = $fabs;$2 = HEAP32[tempDoublePtr>>2]|0;
 $3 = HEAP32[tempDoublePtr+4>>2]|0;
 $4 = $1 & -2147483648;
 $5 = $4 | $3;
 HEAP32[tempDoublePtr>>2] = $2;HEAP32[tempDoublePtr+4>>2] = $5;$6 = +HEAPF64[tempDoublePtr>>3];
 return (+$6);
}
function _cosh($x) {
 $x = +$x;
 var $$0 = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $fabs = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $fabs = (+Math_abs((+$x)));
 HEAPF64[tempDoublePtr>>3] = $fabs;$0 = HEAP32[tempDoublePtr>>2]|0;
 $1 = HEAP32[tempDoublePtr+4>>2]|0;
 $2 = ($1>>>0)<(1072049730);
 do {
  if ($2) {
   $3 = ($1>>>0)<(1045430272);
   if ($3) {
    $$0 = 1.0;
   } else {
    $4 = (+_expm1($fabs));
    $5 = $4 * $4;
    $6 = $4 + 1.0;
    $7 = $6 * 2.0;
    $8 = $5 / $7;
    $9 = $8 + 1.0;
    $$0 = $9;
   }
  } else {
   $10 = ($1>>>0)<(1082535490);
   if ($10) {
    $11 = (+Math_exp((+$fabs)));
    $12 = 1.0 / $11;
    $13 = $11 + $12;
    $14 = $13 * 0.5;
    $$0 = $14;
    break;
   } else {
    $15 = (+___expo2($fabs));
    $$0 = $15;
    break;
   }
  }
 } while(0);
 return (+$$0);
}
function _expm1($x) {
 $x = +$x;
 var $$0 = 0.0, $$02 = 0.0, $$pn = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0, $23 = 0.0;
 var $24 = 0.0, $25 = 0, $26 = 0.0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0;
 var $42 = 0.0, $43 = 0.0, $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0.0, $5 = 0, $50 = 0.0, $51 = 0.0, $52 = 0.0, $53 = 0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0;
 var $60 = 0.0, $61 = 0.0, $62 = 0.0, $63 = 0.0, $64 = 0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $72 = 0, $73 = 0, $74 = 0.0, $75 = 0, $76 = 0.0, $77 = 0.0, $78 = 0;
 var $79 = 0.0, $8 = 0, $80 = 0.0, $81 = 0.0, $82 = 0.0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0.0, $91 = 0.0, $92 = 0.0, $93 = 0.0, $94 = 0.0, $c$0 = 0.0, $fabs = 0.0;
 var $hi$0 = 0.0, $k$0 = 0, $k$1 = 0, $lo$0 = 0.0, $y$0 = 0.0, $y$1 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAPF64[tempDoublePtr>>3] = $x;$0 = HEAP32[tempDoublePtr>>2]|0;
 $1 = HEAP32[tempDoublePtr+4>>2]|0;
 $2 = $1 & 2147483647;
 $3 = (_bitshift64Lshr(($0|0),($1|0),63)|0);
 $4 = tempRet0;
 $5 = ($2>>>0)>(1078159481);
 do {
  if ($5) {
   $fabs = (+Math_abs((+$x)));
   HEAPF64[tempDoublePtr>>3] = $fabs;$6 = HEAP32[tempDoublePtr>>2]|0;
   $7 = HEAP32[tempDoublePtr+4>>2]|0;
   $8 = ($7>>>0)>(2146435072);
   $9 = ($6>>>0)>(0);
   $10 = ($7|0)==(2146435072);
   $11 = $10 & $9;
   $12 = $8 | $11;
   if ($12) {
    $$0 = $x;
   } else {
    $13 = ($3|0)==(0);
    if ($13) {
     $14 = $x > 709.78271289338397;
     if ($14) {
      $15 = $x * 8.9884656743115795E+307;
      $$0 = $15;
     } else {
      label = 11;
     }
    } else {
     $$0 = -1.0;
    }
   }
  } else {
   $16 = ($2>>>0)>(1071001154);
   if (!($16)) {
    $33 = ($2>>>0)<(1016070144);
    if ($33) {
     $$0 = $x;
     break;
    } else {
     $$02 = $x;$c$0 = 0.0;$k$1 = 0;
     label = 14;
     break;
    }
   }
   $17 = ($2>>>0)<(1072734898);
   if ($17) {
    $18 = ($3|0)==(0);
    if ($18) {
     $19 = $x + -0.69314718036912382;
     $hi$0 = $19;$k$0 = 1;$lo$0 = 1.9082149292705877E-10;
     label = 12;
     break;
    } else {
     $20 = $x + 0.69314718036912382;
     $hi$0 = $20;$k$0 = -1;$lo$0 = -1.9082149292705877E-10;
     label = 12;
     break;
    }
   } else {
    label = 11;
   }
  }
 } while(0);
 if ((label|0) == 11) {
  $21 = $x * 1.4426950408889634;
  $22 = ($3|0)!=(0);
  $23 = $22 ? -0.5 : 0.5;
  $24 = $21 + $23;
  $25 = (~~(($24)));
  $26 = (+($25|0));
  $27 = $26 * 0.69314718036912382;
  $28 = $x - $27;
  $29 = $26 * 1.9082149292705877E-10;
  $hi$0 = $28;$k$0 = $25;$lo$0 = $29;
  label = 12;
 }
 if ((label|0) == 12) {
  $30 = $hi$0 - $lo$0;
  $31 = $hi$0 - $30;
  $32 = $31 - $lo$0;
  $$02 = $30;$c$0 = $32;$k$1 = $k$0;
  label = 14;
 }
 L18: do {
  if ((label|0) == 14) {
   $34 = $$02 * 0.5;
   $35 = $$02 * $34;
   $36 = $35 * 2.0109921818362437E-7;
   $37 = 4.0082178273293624E-6 - $36;
   $38 = $35 * $37;
   $39 = $38 + -7.9365075786748794E-5;
   $40 = $35 * $39;
   $41 = $40 + 0.0015873015872548146;
   $42 = $35 * $41;
   $43 = $42 + -0.033333333333333132;
   $44 = $35 * $43;
   $45 = $44 + 1.0;
   $46 = $34 * $45;
   $47 = 3.0 - $46;
   $48 = $45 - $47;
   $49 = $$02 * $47;
   $50 = 6.0 - $49;
   $51 = $48 / $50;
   $52 = $35 * $51;
   $53 = ($k$1|0)==(0);
   if ($53) {
    $54 = $$02 * $52;
    $55 = $54 - $35;
    $56 = $$02 - $55;
    $$0 = $56;
    break;
   }
   $57 = $52 - $c$0;
   $58 = $$02 * $57;
   $59 = $58 - $c$0;
   $60 = $59 - $35;
   switch ($k$1|0) {
   case -1:  {
    $61 = $$02 - $60;
    $62 = $61 * 0.5;
    $63 = $62 + -0.5;
    $$0 = $63;
    break L18;
    break;
   }
   case 1:  {
    $64 = $$02 < -0.25;
    if ($64) {
     $65 = $$02 + 0.5;
     $66 = $60 - $65;
     $67 = $66 * -2.0;
     $$0 = $67;
     break L18;
    } else {
     $68 = $$02 - $60;
     $69 = $68 * 2.0;
     $70 = $69 + 1.0;
     $$0 = $70;
     break L18;
    }
    break;
   }
   default: {
    $71 = (($k$1) + 1023)|0;
    $72 = (_bitshift64Shl(($71|0),0,52)|0);
    $73 = tempRet0;
    HEAP32[tempDoublePtr>>2] = $72;HEAP32[tempDoublePtr+4>>2] = $73;$74 = +HEAPF64[tempDoublePtr>>3];
    $75 = ($k$1>>>0)>(56);
    if ($75) {
     $76 = $$02 - $60;
     $77 = $76 + 1.0;
     $78 = ($k$1|0)==(1024);
     $79 = $77 * 2.0;
     $80 = $79 * 8.9884656743115795E+307;
     $81 = $74 * $77;
     $y$0 = $78 ? $80 : $81;
     $82 = $y$0 + -1.0;
     $$0 = $82;
     break L18;
    }
    $83 = (1023 - ($k$1))|0;
    $84 = (_bitshift64Shl(($83|0),0,52)|0);
    $85 = tempRet0;
    $86 = ($k$1|0)<(20);
    if ($86) {
     $87 = $$02 - $60;
     HEAP32[tempDoublePtr>>2] = $84;HEAP32[tempDoublePtr+4>>2] = $85;$88 = +HEAPF64[tempDoublePtr>>3];
     $89 = 1.0 - $88;
     $90 = $89 + $87;
     $$pn = $90;
    } else {
     HEAP32[tempDoublePtr>>2] = $84;HEAP32[tempDoublePtr+4>>2] = $85;$91 = +HEAPF64[tempDoublePtr>>3];
     $92 = $91 + $60;
     $93 = $$02 - $92;
     $94 = $93 + 1.0;
     $$pn = $94;
    }
    $y$1 = $74 * $$pn;
    $$0 = $y$1;
    break L18;
   }
   }
  }
 } while(0);
 return (+$$0);
}
function ___expo2($x) {
 $x = +$x;
 var $0 = 0.0, $1 = 0.0, $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $x + -1416.0996898839683;
 $1 = (+Math_exp((+$0)));
 $2 = $1 * 2.2471164185778949E+307;
 $3 = $2 * 2.2471164185778949E+307;
 return (+$3);
}
function _sinh($x) {
 $x = +$x;
 var $$ = 0.0, $$0 = 0.0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0.0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0.0, $7 = 0, $8 = 0, $9 = 0.0, $fabs = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAPF64[tempDoublePtr>>3] = $x;$0 = HEAP32[tempDoublePtr>>2]|0;
 $1 = HEAP32[tempDoublePtr+4>>2]|0;
 $2 = ($1|0)<(0);
 $$ = $2 ? -0.5 : 0.5;
 $fabs = (+Math_abs((+$x)));
 HEAPF64[tempDoublePtr>>3] = $fabs;$3 = HEAP32[tempDoublePtr>>2]|0;
 $4 = HEAP32[tempDoublePtr+4>>2]|0;
 $5 = ($4>>>0)<(1082535490);
 do {
  if ($5) {
   $6 = (+_expm1($fabs));
   $7 = ($4>>>0)<(1072693248);
   if (!($7)) {
    $15 = $6 + 1.0;
    $16 = $6 / $15;
    $17 = $6 + $16;
    $18 = $$ * $17;
    $$0 = $18;
    break;
   }
   $8 = ($4>>>0)<(1045430272);
   if ($8) {
    $$0 = $x;
   } else {
    $9 = $6 * 2.0;
    $10 = $6 * $6;
    $11 = $6 + 1.0;
    $12 = $10 / $11;
    $13 = $9 - $12;
    $14 = $$ * $13;
    $$0 = $14;
   }
  } else {
   $19 = $$ * 2.0;
   $20 = (+___expo2($fabs));
   $21 = $19 * $20;
   $$0 = $21;
  }
 } while(0);
 return (+$$0);
}
function _memchr($src,$c,$n) {
 $src = $src|0;
 $c = $c|0;
 $n = $n|0;
 var $$0$lcssa = 0, $$0$lcssa30 = 0, $$019 = 0, $$1$lcssa = 0, $$110 = 0, $$110$lcssa = 0, $$24 = 0, $$3 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond18 = 0, $s$0$lcssa = 0, $s$0$lcssa29 = 0, $s$020 = 0, $s$15 = 0, $s$2 = 0, $w$0$lcssa = 0, $w$011 = 0, $w$011$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $c & 255;
 $1 = $src;
 $2 = $1 & 3;
 $3 = ($2|0)!=(0);
 $4 = ($n|0)!=(0);
 $or$cond18 = $4 & $3;
 L1: do {
  if ($or$cond18) {
   $5 = $c&255;
   $$019 = $n;$s$020 = $src;
   while(1) {
    $6 = HEAP8[$s$020>>0]|0;
    $7 = ($6<<24>>24)==($5<<24>>24);
    if ($7) {
     $$0$lcssa30 = $$019;$s$0$lcssa29 = $s$020;
     label = 6;
     break L1;
    }
    $8 = ((($s$020)) + 1|0);
    $9 = (($$019) + -1)|0;
    $10 = $8;
    $11 = $10 & 3;
    $12 = ($11|0)!=(0);
    $13 = ($9|0)!=(0);
    $or$cond = $13 & $12;
    if ($or$cond) {
     $$019 = $9;$s$020 = $8;
    } else {
     $$0$lcssa = $9;$$lcssa = $13;$s$0$lcssa = $8;
     label = 5;
     break;
    }
   }
  } else {
   $$0$lcssa = $n;$$lcssa = $4;$s$0$lcssa = $src;
   label = 5;
  }
 } while(0);
 if ((label|0) == 5) {
  if ($$lcssa) {
   $$0$lcssa30 = $$0$lcssa;$s$0$lcssa29 = $s$0$lcssa;
   label = 6;
  } else {
   $$3 = 0;$s$2 = $s$0$lcssa;
  }
 }
 L8: do {
  if ((label|0) == 6) {
   $14 = HEAP8[$s$0$lcssa29>>0]|0;
   $15 = $c&255;
   $16 = ($14<<24>>24)==($15<<24>>24);
   if ($16) {
    $$3 = $$0$lcssa30;$s$2 = $s$0$lcssa29;
   } else {
    $17 = Math_imul($0, 16843009)|0;
    $18 = ($$0$lcssa30>>>0)>(3);
    L11: do {
     if ($18) {
      $$110 = $$0$lcssa30;$w$011 = $s$0$lcssa29;
      while(1) {
       $19 = HEAP32[$w$011>>2]|0;
       $20 = $19 ^ $17;
       $21 = (($20) + -16843009)|0;
       $22 = $20 & -2139062144;
       $23 = $22 ^ -2139062144;
       $24 = $23 & $21;
       $25 = ($24|0)==(0);
       if (!($25)) {
        $$110$lcssa = $$110;$w$011$lcssa = $w$011;
        break;
       }
       $26 = ((($w$011)) + 4|0);
       $27 = (($$110) + -4)|0;
       $28 = ($27>>>0)>(3);
       if ($28) {
        $$110 = $27;$w$011 = $26;
       } else {
        $$1$lcssa = $27;$w$0$lcssa = $26;
        label = 11;
        break L11;
       }
      }
      $$24 = $$110$lcssa;$s$15 = $w$011$lcssa;
     } else {
      $$1$lcssa = $$0$lcssa30;$w$0$lcssa = $s$0$lcssa29;
      label = 11;
     }
    } while(0);
    if ((label|0) == 11) {
     $29 = ($$1$lcssa|0)==(0);
     if ($29) {
      $$3 = 0;$s$2 = $w$0$lcssa;
      break;
     } else {
      $$24 = $$1$lcssa;$s$15 = $w$0$lcssa;
     }
    }
    while(1) {
     $30 = HEAP8[$s$15>>0]|0;
     $31 = ($30<<24>>24)==($15<<24>>24);
     if ($31) {
      $$3 = $$24;$s$2 = $s$15;
      break L8;
     }
     $32 = ((($s$15)) + 1|0);
     $33 = (($$24) + -1)|0;
     $34 = ($33|0)==(0);
     if ($34) {
      $$3 = 0;$s$2 = $32;
      break;
     } else {
      $$24 = $33;$s$15 = $32;
     }
    }
   }
  }
 } while(0);
 $35 = ($$3|0)!=(0);
 $36 = $35 ? $s$2 : 0;
 return ($36|0);
}
function _strncmp($_l,$_r,$n) {
 $_l = $_l|0;
 $_r = $_r|0;
 $n = $n|0;
 var $$03 = 0, $$08 = 0, $$08$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $l$06 = 0, $or$cond = 0, $or$cond4 = 0, $r$0$lcssa = 0, $r$07 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)==(0);
 if ($0) {
  $$03 = 0;
 } else {
  $1 = HEAP8[$_l>>0]|0;
  $2 = ($1<<24>>24)==(0);
  L3: do {
   if ($2) {
    $13 = 0;$r$0$lcssa = $_r;
   } else {
    $$08$in = $n;$7 = $1;$l$06 = $_l;$r$07 = $_r;
    while(1) {
     $$08 = (($$08$in) + -1)|0;
     $3 = HEAP8[$r$07>>0]|0;
     $4 = ($3<<24>>24)!=(0);
     $5 = ($$08|0)!=(0);
     $or$cond = $5 & $4;
     $6 = ($7<<24>>24)==($3<<24>>24);
     $or$cond4 = $6 & $or$cond;
     if (!($or$cond4)) {
      $13 = $7;$r$0$lcssa = $r$07;
      break L3;
     }
     $8 = ((($l$06)) + 1|0);
     $9 = ((($r$07)) + 1|0);
     $10 = HEAP8[$8>>0]|0;
     $11 = ($10<<24>>24)==(0);
     if ($11) {
      $13 = 0;$r$0$lcssa = $9;
      break;
     } else {
      $$08$in = $$08;$7 = $10;$l$06 = $8;$r$07 = $9;
     }
    }
   }
  } while(0);
  $12 = $13&255;
  $14 = HEAP8[$r$0$lcssa>>0]|0;
  $15 = $14&255;
  $16 = (($12) - ($15))|0;
  $$03 = $16;
 }
 return ($$03|0);
}
function _memcmp($vl,$vr,$n) {
 $vl = $vl|0;
 $vr = $vr|0;
 $n = $n|0;
 var $$03 = 0, $$lcssa = 0, $$lcssa19 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $l$04 = 0, $r$05 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)==(0);
 L1: do {
  if ($0) {
   $11 = 0;
  } else {
   $$03 = $n;$l$04 = $vl;$r$05 = $vr;
   while(1) {
    $1 = HEAP8[$l$04>>0]|0;
    $2 = HEAP8[$r$05>>0]|0;
    $3 = ($1<<24>>24)==($2<<24>>24);
    if (!($3)) {
     $$lcssa = $1;$$lcssa19 = $2;
     break;
    }
    $4 = (($$03) + -1)|0;
    $5 = ((($l$04)) + 1|0);
    $6 = ((($r$05)) + 1|0);
    $7 = ($4|0)==(0);
    if ($7) {
     $11 = 0;
     break L1;
    } else {
     $$03 = $4;$l$04 = $5;$r$05 = $6;
    }
   }
   $8 = $$lcssa&255;
   $9 = $$lcssa19&255;
   $10 = (($8) - ($9))|0;
   $11 = $10;
  }
 } while(0);
 return ($11|0);
}
function _sprintf($s,$fmt,$varargs) {
 $s = $s|0;
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $0 = 0, $ap = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ap = sp;
 HEAP32[$ap>>2] = $varargs;
 $0 = (_vsprintf($s,$fmt,$ap)|0);
 STACKTOP = sp;return ($0|0);
}
function _vsprintf($s,$fmt,$ap) {
 $s = $s|0;
 $fmt = $fmt|0;
 $ap = $ap|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_vsnprintf($s,2147483647,$fmt,$ap)|0);
 return ($0|0);
}
function _vsnprintf($s,$n,$fmt,$ap) {
 $s = $s|0;
 $n = $n|0;
 $fmt = $fmt|0;
 $ap = $ap|0;
 var $$$02 = 0, $$0 = 0, $$01 = 0, $$02 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $b = 0, $f = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $b = sp + 112|0;
 $f = sp;
 dest=$f; src=150528; stop=dest+112|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 $0 = (($n) + -1)|0;
 $1 = ($0>>>0)>(2147483646);
 if ($1) {
  $2 = ($n|0)==(0);
  if ($2) {
   $$01 = $b;$$02 = 1;
   label = 4;
  } else {
   $3 = (___errno_location()|0);
   HEAP32[$3>>2] = 75;
   $$0 = -1;
  }
 } else {
  $$01 = $s;$$02 = $n;
  label = 4;
 }
 if ((label|0) == 4) {
  $4 = $$01;
  $5 = (-2 - ($4))|0;
  $6 = ($$02>>>0)>($5>>>0);
  $$$02 = $6 ? $5 : $$02;
  $7 = ((($f)) + 48|0);
  HEAP32[$7>>2] = $$$02;
  $8 = ((($f)) + 20|0);
  HEAP32[$8>>2] = $$01;
  $9 = ((($f)) + 44|0);
  HEAP32[$9>>2] = $$01;
  $10 = (($$01) + ($$$02)|0);
  $11 = ((($f)) + 16|0);
  HEAP32[$11>>2] = $10;
  $12 = ((($f)) + 28|0);
  HEAP32[$12>>2] = $10;
  $13 = (_vfprintf($f,$fmt,$ap)|0);
  $14 = ($$$02|0)==(0);
  if ($14) {
   $$0 = $13;
  } else {
   $15 = HEAP32[$8>>2]|0;
   $16 = HEAP32[$11>>2]|0;
   $17 = ($15|0)==($16|0);
   $18 = $17 << 31 >> 31;
   $19 = (($15) + ($18)|0);
   HEAP8[$19>>0] = 0;
   $$0 = $13;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _sn_write($f,$s,$l) {
 $f = $f|0;
 $s = $s|0;
 $l = $l|0;
 var $$cast = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $l$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($f)) + 20|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) - ($3))|0;
 $5 = ($4>>>0)>($l>>>0);
 $l$ = $5 ? $l : $4;
 $$cast = $3;
 _memcpy(($$cast|0),($s|0),($l$|0))|0;
 $6 = HEAP32[$2>>2]|0;
 $7 = (($6) + ($l$)|0);
 HEAP32[$2>>2] = $7;
 return ($l|0);
}
function _vfprintf($f,$fmt,$ap) {
 $f = $f|0;
 $fmt = $fmt|0;
 $ap = $ap|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ap2 = 0, $internal_buf = 0, $nl_arg = 0, $nl_type = 0;
 var $ret$1 = 0, $ret$1$ = 0, $vacopy_currentptr = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ap2 = sp + 120|0;
 $nl_type = sp + 80|0;
 $nl_arg = sp;
 $internal_buf = sp + 136|0;
 dest=$nl_type; stop=dest+40|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $vacopy_currentptr = HEAP32[$ap>>2]|0;
 HEAP32[$ap2>>2] = $vacopy_currentptr;
 $0 = (_printf_core(0,$fmt,$ap2,$nl_arg,$nl_type)|0);
 $1 = ($0|0)<(0);
 if ($1) {
  $$0 = -1;
 } else {
  $2 = ((($f)) + 76|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ($3|0)>(-1);
  if ($4) {
   $5 = (___lockfile($f)|0);
   $33 = $5;
  } else {
   $33 = 0;
  }
  $6 = HEAP32[$f>>2]|0;
  $7 = $6 & 32;
  $8 = ((($f)) + 74|0);
  $9 = HEAP8[$8>>0]|0;
  $10 = ($9<<24>>24)<(1);
  if ($10) {
   $11 = $6 & -33;
   HEAP32[$f>>2] = $11;
  }
  $12 = ((($f)) + 48|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ($13|0)==(0);
  if ($14) {
   $16 = ((($f)) + 44|0);
   $17 = HEAP32[$16>>2]|0;
   HEAP32[$16>>2] = $internal_buf;
   $18 = ((($f)) + 28|0);
   HEAP32[$18>>2] = $internal_buf;
   $19 = ((($f)) + 20|0);
   HEAP32[$19>>2] = $internal_buf;
   HEAP32[$12>>2] = 80;
   $20 = ((($internal_buf)) + 80|0);
   $21 = ((($f)) + 16|0);
   HEAP32[$21>>2] = $20;
   $22 = (_printf_core($f,$fmt,$ap2,$nl_arg,$nl_type)|0);
   $23 = ($17|0)==(0|0);
   if ($23) {
    $ret$1 = $22;
   } else {
    $24 = ((($f)) + 36|0);
    $25 = HEAP32[$24>>2]|0;
    (FUNCTION_TABLE_iiii[$25 & 8191]($f,0,0)|0);
    $26 = HEAP32[$19>>2]|0;
    $27 = ($26|0)==(0|0);
    $$ = $27 ? -1 : $22;
    HEAP32[$16>>2] = $17;
    HEAP32[$12>>2] = 0;
    HEAP32[$21>>2] = 0;
    HEAP32[$18>>2] = 0;
    HEAP32[$19>>2] = 0;
    $ret$1 = $$;
   }
  } else {
   $15 = (_printf_core($f,$fmt,$ap2,$nl_arg,$nl_type)|0);
   $ret$1 = $15;
  }
  $28 = HEAP32[$f>>2]|0;
  $29 = $28 & 32;
  $30 = ($29|0)==(0);
  $ret$1$ = $30 ? $ret$1 : -1;
  $31 = $28 | $7;
  HEAP32[$f>>2] = $31;
  $32 = ($33|0)==(0);
  if (!($32)) {
   ___unlockfile($f);
  }
  $$0 = $ret$1$;
 }
 STACKTOP = sp;return ($$0|0);
}
function _printf_core($f,$fmt,$ap,$nl_arg,$nl_type) {
 $f = $f|0;
 $fmt = $fmt|0;
 $ap = $ap|0;
 $nl_arg = $nl_arg|0;
 $nl_type = $nl_type|0;
 var $$ = 0, $$$i = 0, $$0 = 0, $$0$i = 0, $$0$lcssa$i = 0, $$012$i = 0, $$013$i = 0, $$03$i33 = 0, $$07$i = 0.0, $$1$i = 0.0, $$114$i = 0, $$2$i = 0.0, $$20$i = 0.0, $$210$$24$i = 0, $$210$$26$i = 0, $$210$i = 0, $$23$i = 0, $$25$i = 0, $$3$i = 0.0, $$311$i = 0;
 var $$33$i = 0, $$36$i = 0.0, $$4$i = 0.0, $$412$lcssa$i = 0, $$41278$i = 0, $$43 = 0, $$5$lcssa$i = 0, $$589$i = 0, $$a$3$i = 0, $$a$3191$i = 0, $$a$3192$i = 0, $$fl$4 = 0, $$l10n$0 = 0, $$lcssa = 0, $$lcssa162$i = 0, $$lcssa295 = 0, $$lcssa300 = 0, $$lcssa301 = 0, $$lcssa302 = 0, $$lcssa303 = 0;
 var $$lcssa304 = 0, $$lcssa306 = 0, $$lcssa316 = 0, $$lcssa319 = 0.0, $$lcssa321 = 0, $$neg55$i = 0, $$neg56$i = 0, $$p$$i = 0, $$p$5 = 0, $$p$i = 0, $$pn$i = 0, $$pr$i = 0, $$pr50$i = 0, $$pre = 0, $$pre$i = 0, $$pre$phi190$iZ2D = 0, $$pre170 = 0, $$pre171 = 0, $$pre185$i = 0, $$pre188$i = 0;
 var $$pre189$i = 0, $$z$3$i = 0, $$z$4$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0;
 var $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0;
 var $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0;
 var $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0;
 var $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0;
 var $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0;
 var $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0.0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0.0, $363 = 0, $364 = 0, $365 = 0;
 var $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0;
 var $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0.0, $391 = 0.0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0;
 var $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0.0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0.0, $411 = 0.0, $412 = 0.0, $413 = 0.0, $414 = 0.0, $415 = 0.0, $416 = 0, $417 = 0, $418 = 0, $419 = 0;
 var $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0;
 var $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0.0, $442 = 0.0, $443 = 0.0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0;
 var $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0;
 var $474 = 0.0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0.0, $483 = 0.0, $484 = 0.0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0;
 var $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0;
 var $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0;
 var $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0;
 var $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0;
 var $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0;
 var $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0.0, $594 = 0.0, $595 = 0, $596 = 0.0, $597 = 0, $598 = 0, $599 = 0, $6 = 0;
 var $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0;
 var $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0;
 var $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0;
 var $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0;
 var $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0;
 var $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0;
 var $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0;
 var $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0;
 var $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0;
 var $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0;
 var $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $a$0 = 0, $a$1 = 0, $a$1$lcssa$i = 0, $a$1149$i = 0, $a$2 = 0, $a$2$ph$i = 0, $a$3$lcssa$i = 0, $a$3136$i = 0, $a$5$lcssa$i = 0, $a$5111$i = 0, $a$6$i = 0, $a$8$i = 0, $a$9$ph$i = 0, $arg = 0, $arglist_current = 0, $arglist_current2 = 0, $arglist_next = 0, $arglist_next3 = 0;
 var $argpos$0 = 0, $big$i = 0, $buf = 0, $buf$i = 0, $carry$0142$i = 0, $carry3$0130$i = 0, $cnt$0 = 0, $cnt$1 = 0, $cnt$1$lcssa = 0, $d$0$i = 0, $d$0141$i = 0, $d$0143$i = 0, $d$1129$i = 0, $d$2$lcssa$i = 0, $d$2110$i = 0, $d$4$i = 0, $d$584$i = 0, $d$677$i = 0, $d$788$i = 0, $e$0125$i = 0;
 var $e$1$i = 0, $e$2106$i = 0, $e$4$i = 0, $e$5$ph$i = 0, $e2$i = 0, $ebuf0$i = 0, $estr$0$i = 0, $estr$1$lcssa$i = 0, $estr$195$i = 0, $estr$2$i = 0, $exitcond$i = 0, $expanded = 0, $expanded10 = 0, $expanded11 = 0, $expanded13 = 0, $expanded14 = 0, $expanded15 = 0, $expanded4 = 0, $expanded6 = 0, $expanded7 = 0;
 var $expanded8 = 0, $fl$0100 = 0, $fl$053 = 0, $fl$1 = 0, $fl$1$ = 0, $fl$3 = 0, $fl$4 = 0, $fl$6 = 0, $i$0$lcssa = 0, $i$0$lcssa178 = 0, $i$0105 = 0, $i$0124$i = 0, $i$03$i = 0, $i$03$i25 = 0, $i$1$lcssa$i = 0, $i$1116 = 0, $i$1118$i = 0, $i$2105$i = 0, $i$291 = 0, $i$291$lcssa = 0;
 var $i$3101$i = 0, $i$389 = 0, $isdigit = 0, $isdigit$i = 0, $isdigit$i27 = 0, $isdigit10 = 0, $isdigit12 = 0, $isdigit2$i = 0, $isdigit2$i23 = 0, $isdigittmp = 0, $isdigittmp$ = 0, $isdigittmp$i = 0, $isdigittmp$i26 = 0, $isdigittmp1$i = 0, $isdigittmp1$i22 = 0, $isdigittmp11 = 0, $isdigittmp4$i = 0, $isdigittmp4$i24 = 0, $isdigittmp9 = 0, $j$0$i = 0;
 var $j$0117$i = 0, $j$0119$i = 0, $j$1102$i = 0, $j$2$i = 0, $l$0 = 0, $l$0$i = 0, $l$1$i = 0, $l$1104 = 0, $l$2 = 0, $l10n$0 = 0, $l10n$0$lcssa = 0, $l10n$0$phi = 0, $l10n$1 = 0, $l10n$2 = 0, $l10n$3 = 0, $mb = 0, $notlhs$i = 0, $notrhs$i = 0, $or$cond = 0, $or$cond$i = 0;
 var $or$cond122 = 0, $or$cond15 = 0, $or$cond17 = 0, $or$cond18$i = 0, $or$cond20 = 0, $or$cond22$i = 0, $or$cond3$not$i = 0, $or$cond31$i = 0, $or$cond6$i = 0, $p$0 = 0, $p$0$ = 0, $p$1 = 0, $p$2 = 0, $p$2$ = 0, $p$3 = 0, $p$4176 = 0, $p$5 = 0, $pl$0 = 0, $pl$0$i = 0, $pl$1 = 0;
 var $pl$1$i = 0, $pl$2 = 0, $prefix$0 = 0, $prefix$0$$i = 0, $prefix$0$i = 0, $prefix$1 = 0, $prefix$2 = 0, $r$0$a$9$i = 0, $re$171$i = 0, $round$070$i = 0.0, $round6$1$i = 0.0, $s$0 = 0, $s$0$i = 0, $s$1 = 0, $s$1$i = 0, $s$1$i$lcssa = 0, $s$2$lcssa = 0, $s$292 = 0, $s$4 = 0, $s$6 = 0;
 var $s$7 = 0, $s$7$lcssa298 = 0, $s1$0$i = 0, $s7$081$i = 0, $s7$1$i = 0, $s8$0$lcssa$i = 0, $s8$072$i = 0, $s9$0$i = 0, $s9$185$i = 0, $s9$2$i = 0, $scevgep182$i = 0, $scevgep182183$i = 0, $small$0$i = 0.0, $small$1$i = 0.0, $st$0 = 0, $st$0$lcssa299 = 0, $storemerge = 0, $storemerge13 = 0, $storemerge851 = 0, $storemerge899 = 0;
 var $sum = 0, $t$0 = 0, $t$1 = 0, $w$$i = 0, $w$0 = 0, $w$1 = 0, $w$2 = 0, $w$32$i = 0, $wc = 0, $ws$0106 = 0, $ws$1117 = 0, $z$0$i = 0, $z$0$lcssa = 0, $z$093 = 0, $z$1 = 0, $z$1$lcssa$i = 0, $z$1148$i = 0, $z$2 = 0, $z$2$i = 0, $z$2$i$lcssa = 0;
 var $z$3$lcssa$i = 0, $z$3135$i = 0, $z$4$i = 0, $z$7$$i = 0, $z$7$i = 0, $z$7$i$lcssa = 0, $z$7$ph$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 624|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $big$i = sp + 24|0;
 $e2$i = sp + 16|0;
 $buf$i = sp + 588|0;
 $ebuf0$i = sp + 576|0;
 $arg = sp;
 $buf = sp + 536|0;
 $wc = sp + 8|0;
 $mb = sp + 528|0;
 $0 = ($f|0)!=(0|0);
 $1 = ((($buf)) + 40|0);
 $2 = $1;
 $3 = ((($buf)) + 39|0);
 $4 = ((($wc)) + 4|0);
 $5 = $buf$i;
 $6 = (0 - ($5))|0;
 $7 = ((($ebuf0$i)) + 12|0);
 $8 = ((($ebuf0$i)) + 11|0);
 $9 = $7;
 $10 = (($9) - ($5))|0;
 $11 = (-2 - ($5))|0;
 $12 = (($9) + 2)|0;
 $13 = ((($big$i)) + 288|0);
 $14 = ((($buf$i)) + 9|0);
 $15 = $14;
 $16 = ((($buf$i)) + 8|0);
 $cnt$0 = 0;$l$0 = 0;$l10n$0 = 0;$s$0 = $fmt;
 L1: while(1) {
  $17 = ($cnt$0|0)>(-1);
  do {
   if ($17) {
    $18 = (2147483647 - ($cnt$0))|0;
    $19 = ($l$0|0)>($18|0);
    if ($19) {
     $20 = (___errno_location()|0);
     HEAP32[$20>>2] = 75;
     $cnt$1 = -1;
     break;
    } else {
     $21 = (($l$0) + ($cnt$0))|0;
     $cnt$1 = $21;
     break;
    }
   } else {
    $cnt$1 = $cnt$0;
   }
  } while(0);
  $22 = HEAP8[$s$0>>0]|0;
  $23 = ($22<<24>>24)==(0);
  if ($23) {
   $cnt$1$lcssa = $cnt$1;$l10n$0$lcssa = $l10n$0;
   label = 244;
   break;
  } else {
   $24 = $22;$s$1 = $s$0;
  }
  L9: while(1) {
   switch ($24<<24>>24) {
   case 37:  {
    $s$292 = $s$1;$z$093 = $s$1;
    label = 9;
    break L9;
    break;
   }
   case 0:  {
    $s$2$lcssa = $s$1;$z$0$lcssa = $s$1;
    break L9;
    break;
   }
   default: {
   }
   }
   $25 = ((($s$1)) + 1|0);
   $$pre = HEAP8[$25>>0]|0;
   $24 = $$pre;$s$1 = $25;
  }
  L12: do {
   if ((label|0) == 9) {
    while(1) {
     label = 0;
     $26 = ((($s$292)) + 1|0);
     $27 = HEAP8[$26>>0]|0;
     $28 = ($27<<24>>24)==(37);
     if (!($28)) {
      $s$2$lcssa = $s$292;$z$0$lcssa = $z$093;
      break L12;
     }
     $29 = ((($z$093)) + 1|0);
     $30 = ((($s$292)) + 2|0);
     $31 = HEAP8[$30>>0]|0;
     $32 = ($31<<24>>24)==(37);
     if ($32) {
      $s$292 = $30;$z$093 = $29;
      label = 9;
     } else {
      $s$2$lcssa = $30;$z$0$lcssa = $29;
      break;
     }
    }
   }
  } while(0);
  $33 = $z$0$lcssa;
  $34 = $s$0;
  $35 = (($33) - ($34))|0;
  if ($0) {
   $36 = HEAP32[$f>>2]|0;
   $37 = $36 & 32;
   $38 = ($37|0)==(0);
   if ($38) {
    (___fwritex($s$0,$35,$f)|0);
   }
  }
  $39 = ($z$0$lcssa|0)==($s$0|0);
  if (!($39)) {
   $l10n$0$phi = $l10n$0;$cnt$0 = $cnt$1;$l$0 = $35;$s$0 = $s$2$lcssa;$l10n$0 = $l10n$0$phi;
   continue;
  }
  $40 = ((($s$2$lcssa)) + 1|0);
  $41 = HEAP8[$40>>0]|0;
  $42 = $41 << 24 >> 24;
  $isdigittmp = (($42) + -48)|0;
  $isdigit = ($isdigittmp>>>0)<(10);
  if ($isdigit) {
   $43 = ((($s$2$lcssa)) + 2|0);
   $44 = HEAP8[$43>>0]|0;
   $45 = ($44<<24>>24)==(36);
   $46 = ((($s$2$lcssa)) + 3|0);
   $$43 = $45 ? $46 : $40;
   $$l10n$0 = $45 ? 1 : $l10n$0;
   $isdigittmp$ = $45 ? $isdigittmp : -1;
   $$pre170 = HEAP8[$$43>>0]|0;
   $48 = $$pre170;$argpos$0 = $isdigittmp$;$l10n$1 = $$l10n$0;$storemerge = $$43;
  } else {
   $48 = $41;$argpos$0 = -1;$l10n$1 = $l10n$0;$storemerge = $40;
  }
  $47 = $48 << 24 >> 24;
  $49 = $47 & -32;
  $50 = ($49|0)==(32);
  L25: do {
   if ($50) {
    $52 = $47;$57 = $48;$fl$0100 = 0;$storemerge899 = $storemerge;
    while(1) {
     $51 = (($52) + -32)|0;
     $53 = 1 << $51;
     $54 = $53 & 75913;
     $55 = ($54|0)==(0);
     if ($55) {
      $67 = $57;$fl$053 = $fl$0100;$storemerge851 = $storemerge899;
      break L25;
     }
     $56 = $57 << 24 >> 24;
     $58 = (($56) + -32)|0;
     $59 = 1 << $58;
     $60 = $59 | $fl$0100;
     $61 = ((($storemerge899)) + 1|0);
     $62 = HEAP8[$61>>0]|0;
     $63 = $62 << 24 >> 24;
     $64 = $63 & -32;
     $65 = ($64|0)==(32);
     if ($65) {
      $52 = $63;$57 = $62;$fl$0100 = $60;$storemerge899 = $61;
     } else {
      $67 = $62;$fl$053 = $60;$storemerge851 = $61;
      break;
     }
    }
   } else {
    $67 = $48;$fl$053 = 0;$storemerge851 = $storemerge;
   }
  } while(0);
  $66 = ($67<<24>>24)==(42);
  do {
   if ($66) {
    $68 = ((($storemerge851)) + 1|0);
    $69 = HEAP8[$68>>0]|0;
    $70 = $69 << 24 >> 24;
    $isdigittmp11 = (($70) + -48)|0;
    $isdigit12 = ($isdigittmp11>>>0)<(10);
    if ($isdigit12) {
     $71 = ((($storemerge851)) + 2|0);
     $72 = HEAP8[$71>>0]|0;
     $73 = ($72<<24>>24)==(36);
     if ($73) {
      $74 = (($nl_type) + ($isdigittmp11<<2)|0);
      HEAP32[$74>>2] = 10;
      $75 = HEAP8[$68>>0]|0;
      $76 = $75 << 24 >> 24;
      $77 = (($76) + -48)|0;
      $78 = (($nl_arg) + ($77<<3)|0);
      $79 = $78;
      $80 = $79;
      $81 = HEAP32[$80>>2]|0;
      $82 = (($79) + 4)|0;
      $83 = $82;
      $84 = HEAP32[$83>>2]|0;
      $85 = ((($storemerge851)) + 3|0);
      $l10n$2 = 1;$storemerge13 = $85;$w$0 = $81;
     } else {
      label = 24;
     }
    } else {
     label = 24;
    }
    if ((label|0) == 24) {
     label = 0;
     $86 = ($l10n$1|0)==(0);
     if (!($86)) {
      $$0 = -1;
      break L1;
     }
     if (!($0)) {
      $fl$1 = $fl$053;$l10n$3 = 0;$s$4 = $68;$w$1 = 0;
      break;
     }
     $arglist_current = HEAP32[$ap>>2]|0;
     $87 = $arglist_current;
     $88 = ((0) + 4|0);
     $expanded4 = $88;
     $expanded = (($expanded4) - 1)|0;
     $89 = (($87) + ($expanded))|0;
     $90 = ((0) + 4|0);
     $expanded8 = $90;
     $expanded7 = (($expanded8) - 1)|0;
     $expanded6 = $expanded7 ^ -1;
     $91 = $89 & $expanded6;
     $92 = $91;
     $93 = HEAP32[$92>>2]|0;
     $arglist_next = ((($92)) + 4|0);
     HEAP32[$ap>>2] = $arglist_next;
     $l10n$2 = 0;$storemerge13 = $68;$w$0 = $93;
    }
    $94 = ($w$0|0)<(0);
    if ($94) {
     $95 = $fl$053 | 8192;
     $96 = (0 - ($w$0))|0;
     $fl$1 = $95;$l10n$3 = $l10n$2;$s$4 = $storemerge13;$w$1 = $96;
    } else {
     $fl$1 = $fl$053;$l10n$3 = $l10n$2;$s$4 = $storemerge13;$w$1 = $w$0;
    }
   } else {
    $97 = $67 << 24 >> 24;
    $isdigittmp1$i = (($97) + -48)|0;
    $isdigit2$i = ($isdigittmp1$i>>>0)<(10);
    if ($isdigit2$i) {
     $101 = $storemerge851;$i$03$i = 0;$isdigittmp4$i = $isdigittmp1$i;
     while(1) {
      $98 = ($i$03$i*10)|0;
      $99 = (($98) + ($isdigittmp4$i))|0;
      $100 = ((($101)) + 1|0);
      $102 = HEAP8[$100>>0]|0;
      $103 = $102 << 24 >> 24;
      $isdigittmp$i = (($103) + -48)|0;
      $isdigit$i = ($isdigittmp$i>>>0)<(10);
      if ($isdigit$i) {
       $101 = $100;$i$03$i = $99;$isdigittmp4$i = $isdigittmp$i;
      } else {
       $$lcssa = $99;$$lcssa295 = $100;
       break;
      }
     }
     $104 = ($$lcssa|0)<(0);
     if ($104) {
      $$0 = -1;
      break L1;
     } else {
      $fl$1 = $fl$053;$l10n$3 = $l10n$1;$s$4 = $$lcssa295;$w$1 = $$lcssa;
     }
    } else {
     $fl$1 = $fl$053;$l10n$3 = $l10n$1;$s$4 = $storemerge851;$w$1 = 0;
    }
   }
  } while(0);
  $105 = HEAP8[$s$4>>0]|0;
  $106 = ($105<<24>>24)==(46);
  L46: do {
   if ($106) {
    $107 = ((($s$4)) + 1|0);
    $108 = HEAP8[$107>>0]|0;
    $109 = ($108<<24>>24)==(42);
    if (!($109)) {
     $136 = $108 << 24 >> 24;
     $isdigittmp1$i22 = (($136) + -48)|0;
     $isdigit2$i23 = ($isdigittmp1$i22>>>0)<(10);
     if ($isdigit2$i23) {
      $140 = $107;$i$03$i25 = 0;$isdigittmp4$i24 = $isdigittmp1$i22;
     } else {
      $p$0 = 0;$s$6 = $107;
      break;
     }
     while(1) {
      $137 = ($i$03$i25*10)|0;
      $138 = (($137) + ($isdigittmp4$i24))|0;
      $139 = ((($140)) + 1|0);
      $141 = HEAP8[$139>>0]|0;
      $142 = $141 << 24 >> 24;
      $isdigittmp$i26 = (($142) + -48)|0;
      $isdigit$i27 = ($isdigittmp$i26>>>0)<(10);
      if ($isdigit$i27) {
       $140 = $139;$i$03$i25 = $138;$isdigittmp4$i24 = $isdigittmp$i26;
      } else {
       $p$0 = $138;$s$6 = $139;
       break L46;
      }
     }
    }
    $110 = ((($s$4)) + 2|0);
    $111 = HEAP8[$110>>0]|0;
    $112 = $111 << 24 >> 24;
    $isdigittmp9 = (($112) + -48)|0;
    $isdigit10 = ($isdigittmp9>>>0)<(10);
    if ($isdigit10) {
     $113 = ((($s$4)) + 3|0);
     $114 = HEAP8[$113>>0]|0;
     $115 = ($114<<24>>24)==(36);
     if ($115) {
      $116 = (($nl_type) + ($isdigittmp9<<2)|0);
      HEAP32[$116>>2] = 10;
      $117 = HEAP8[$110>>0]|0;
      $118 = $117 << 24 >> 24;
      $119 = (($118) + -48)|0;
      $120 = (($nl_arg) + ($119<<3)|0);
      $121 = $120;
      $122 = $121;
      $123 = HEAP32[$122>>2]|0;
      $124 = (($121) + 4)|0;
      $125 = $124;
      $126 = HEAP32[$125>>2]|0;
      $127 = ((($s$4)) + 4|0);
      $p$0 = $123;$s$6 = $127;
      break;
     }
    }
    $128 = ($l10n$3|0)==(0);
    if (!($128)) {
     $$0 = -1;
     break L1;
    }
    if ($0) {
     $arglist_current2 = HEAP32[$ap>>2]|0;
     $129 = $arglist_current2;
     $130 = ((0) + 4|0);
     $expanded11 = $130;
     $expanded10 = (($expanded11) - 1)|0;
     $131 = (($129) + ($expanded10))|0;
     $132 = ((0) + 4|0);
     $expanded15 = $132;
     $expanded14 = (($expanded15) - 1)|0;
     $expanded13 = $expanded14 ^ -1;
     $133 = $131 & $expanded13;
     $134 = $133;
     $135 = HEAP32[$134>>2]|0;
     $arglist_next3 = ((($134)) + 4|0);
     HEAP32[$ap>>2] = $arglist_next3;
     $p$0 = $135;$s$6 = $110;
    } else {
     $p$0 = 0;$s$6 = $110;
    }
   } else {
    $p$0 = -1;$s$6 = $s$4;
   }
  } while(0);
  $s$7 = $s$6;$st$0 = 0;
  while(1) {
   $143 = HEAP8[$s$7>>0]|0;
   $144 = $143 << 24 >> 24;
   $145 = (($144) + -65)|0;
   $146 = ($145>>>0)>(57);
   if ($146) {
    $$0 = -1;
    break L1;
   }
   $147 = ((($s$7)) + 1|0);
   $148 = ((617001 + (($st$0*58)|0)|0) + ($145)|0);
   $149 = HEAP8[$148>>0]|0;
   $150 = $149&255;
   $151 = (($150) + -1)|0;
   $152 = ($151>>>0)<(8);
   if ($152) {
    $s$7 = $147;$st$0 = $150;
   } else {
    $$lcssa300 = $147;$$lcssa301 = $149;$$lcssa302 = $150;$s$7$lcssa298 = $s$7;$st$0$lcssa299 = $st$0;
    break;
   }
  }
  $153 = ($$lcssa301<<24>>24)==(0);
  if ($153) {
   $$0 = -1;
   break;
  }
  $154 = ($$lcssa301<<24>>24)==(19);
  $155 = ($argpos$0|0)>(-1);
  do {
   if ($154) {
    if ($155) {
     $$0 = -1;
     break L1;
    } else {
     label = 52;
    }
   } else {
    if ($155) {
     $156 = (($nl_type) + ($argpos$0<<2)|0);
     HEAP32[$156>>2] = $$lcssa302;
     $157 = (($nl_arg) + ($argpos$0<<3)|0);
     $158 = $157;
     $159 = $158;
     $160 = HEAP32[$159>>2]|0;
     $161 = (($158) + 4)|0;
     $162 = $161;
     $163 = HEAP32[$162>>2]|0;
     $164 = $arg;
     $165 = $164;
     HEAP32[$165>>2] = $160;
     $166 = (($164) + 4)|0;
     $167 = $166;
     HEAP32[$167>>2] = $163;
     label = 52;
     break;
    }
    if (!($0)) {
     $$0 = 0;
     break L1;
    }
    _pop_arg($arg,$$lcssa302,$ap);
   }
  } while(0);
  if ((label|0) == 52) {
   label = 0;
   if (!($0)) {
    $cnt$0 = $cnt$1;$l$0 = $35;$l10n$0 = $l10n$3;$s$0 = $$lcssa300;
    continue;
   }
  }
  $168 = HEAP8[$s$7$lcssa298>>0]|0;
  $169 = $168 << 24 >> 24;
  $170 = ($st$0$lcssa299|0)!=(0);
  $171 = $169 & 15;
  $172 = ($171|0)==(3);
  $or$cond15 = $170 & $172;
  $173 = $169 & -33;
  $t$0 = $or$cond15 ? $173 : $169;
  $174 = $fl$1 & 8192;
  $175 = ($174|0)==(0);
  $176 = $fl$1 & -65537;
  $fl$1$ = $175 ? $fl$1 : $176;
  L75: do {
   switch ($t$0|0) {
   case 110:  {
    switch ($st$0$lcssa299|0) {
    case 0:  {
     $183 = HEAP32[$arg>>2]|0;
     HEAP32[$183>>2] = $cnt$1;
     $cnt$0 = $cnt$1;$l$0 = $35;$l10n$0 = $l10n$3;$s$0 = $$lcssa300;
     continue L1;
     break;
    }
    case 1:  {
     $184 = HEAP32[$arg>>2]|0;
     HEAP32[$184>>2] = $cnt$1;
     $cnt$0 = $cnt$1;$l$0 = $35;$l10n$0 = $l10n$3;$s$0 = $$lcssa300;
     continue L1;
     break;
    }
    case 2:  {
     $185 = ($cnt$1|0)<(0);
     $186 = $185 << 31 >> 31;
     $187 = HEAP32[$arg>>2]|0;
     $188 = $187;
     $189 = $188;
     HEAP32[$189>>2] = $cnt$1;
     $190 = (($188) + 4)|0;
     $191 = $190;
     HEAP32[$191>>2] = $186;
     $cnt$0 = $cnt$1;$l$0 = $35;$l10n$0 = $l10n$3;$s$0 = $$lcssa300;
     continue L1;
     break;
    }
    case 3:  {
     $192 = $cnt$1&65535;
     $193 = HEAP32[$arg>>2]|0;
     HEAP16[$193>>1] = $192;
     $cnt$0 = $cnt$1;$l$0 = $35;$l10n$0 = $l10n$3;$s$0 = $$lcssa300;
     continue L1;
     break;
    }
    case 4:  {
     $194 = $cnt$1&255;
     $195 = HEAP32[$arg>>2]|0;
     HEAP8[$195>>0] = $194;
     $cnt$0 = $cnt$1;$l$0 = $35;$l10n$0 = $l10n$3;$s$0 = $$lcssa300;
     continue L1;
     break;
    }
    case 6:  {
     $196 = HEAP32[$arg>>2]|0;
     HEAP32[$196>>2] = $cnt$1;
     $cnt$0 = $cnt$1;$l$0 = $35;$l10n$0 = $l10n$3;$s$0 = $$lcssa300;
     continue L1;
     break;
    }
    case 7:  {
     $197 = ($cnt$1|0)<(0);
     $198 = $197 << 31 >> 31;
     $199 = HEAP32[$arg>>2]|0;
     $200 = $199;
     $201 = $200;
     HEAP32[$201>>2] = $cnt$1;
     $202 = (($200) + 4)|0;
     $203 = $202;
     HEAP32[$203>>2] = $198;
     $cnt$0 = $cnt$1;$l$0 = $35;$l10n$0 = $l10n$3;$s$0 = $$lcssa300;
     continue L1;
     break;
    }
    default: {
     $cnt$0 = $cnt$1;$l$0 = $35;$l10n$0 = $l10n$3;$s$0 = $$lcssa300;
     continue L1;
    }
    }
    break;
   }
   case 112:  {
    $204 = ($p$0>>>0)>(8);
    $205 = $204 ? $p$0 : 8;
    $206 = $fl$1$ | 8;
    $fl$3 = $206;$p$1 = $205;$t$1 = 120;
    label = 64;
    break;
   }
   case 88: case 120:  {
    $fl$3 = $fl$1$;$p$1 = $p$0;$t$1 = $t$0;
    label = 64;
    break;
   }
   case 111:  {
    $244 = $arg;
    $245 = $244;
    $246 = HEAP32[$245>>2]|0;
    $247 = (($244) + 4)|0;
    $248 = $247;
    $249 = HEAP32[$248>>2]|0;
    $250 = ($246|0)==(0);
    $251 = ($249|0)==(0);
    $252 = $250 & $251;
    if ($252) {
     $$0$lcssa$i = $1;
    } else {
     $$03$i33 = $1;$254 = $246;$258 = $249;
     while(1) {
      $253 = $254 & 7;
      $255 = $253 | 48;
      $256 = $255&255;
      $257 = ((($$03$i33)) + -1|0);
      HEAP8[$257>>0] = $256;
      $259 = (_bitshift64Lshr(($254|0),($258|0),3)|0);
      $260 = tempRet0;
      $261 = ($259|0)==(0);
      $262 = ($260|0)==(0);
      $263 = $261 & $262;
      if ($263) {
       $$0$lcssa$i = $257;
       break;
      } else {
       $$03$i33 = $257;$254 = $259;$258 = $260;
      }
     }
    }
    $264 = $fl$1$ & 8;
    $265 = ($264|0)==(0);
    if ($265) {
     $a$0 = $$0$lcssa$i;$fl$4 = $fl$1$;$p$2 = $p$0;$pl$1 = 0;$prefix$1 = 617481;
     label = 77;
    } else {
     $266 = $$0$lcssa$i;
     $267 = (($2) - ($266))|0;
     $268 = ($p$0|0)>($267|0);
     $269 = (($267) + 1)|0;
     $p$0$ = $268 ? $p$0 : $269;
     $a$0 = $$0$lcssa$i;$fl$4 = $fl$1$;$p$2 = $p$0$;$pl$1 = 0;$prefix$1 = 617481;
     label = 77;
    }
    break;
   }
   case 105: case 100:  {
    $270 = $arg;
    $271 = $270;
    $272 = HEAP32[$271>>2]|0;
    $273 = (($270) + 4)|0;
    $274 = $273;
    $275 = HEAP32[$274>>2]|0;
    $276 = ($275|0)<(0);
    if ($276) {
     $277 = (_i64Subtract(0,0,($272|0),($275|0))|0);
     $278 = tempRet0;
     $279 = $arg;
     $280 = $279;
     HEAP32[$280>>2] = $277;
     $281 = (($279) + 4)|0;
     $282 = $281;
     HEAP32[$282>>2] = $278;
     $287 = $277;$288 = $278;$pl$0 = 1;$prefix$0 = 617481;
     label = 76;
     break L75;
    }
    $283 = $fl$1$ & 2048;
    $284 = ($283|0)==(0);
    if ($284) {
     $285 = $fl$1$ & 1;
     $286 = ($285|0)==(0);
     $$ = $286 ? 617481 : (617483);
     $287 = $272;$288 = $275;$pl$0 = $285;$prefix$0 = $$;
     label = 76;
    } else {
     $287 = $272;$288 = $275;$pl$0 = 1;$prefix$0 = (617482);
     label = 76;
    }
    break;
   }
   case 117:  {
    $177 = $arg;
    $178 = $177;
    $179 = HEAP32[$178>>2]|0;
    $180 = (($177) + 4)|0;
    $181 = $180;
    $182 = HEAP32[$181>>2]|0;
    $287 = $179;$288 = $182;$pl$0 = 0;$prefix$0 = 617481;
    label = 76;
    break;
   }
   case 99:  {
    $308 = $arg;
    $309 = $308;
    $310 = HEAP32[$309>>2]|0;
    $311 = (($308) + 4)|0;
    $312 = $311;
    $313 = HEAP32[$312>>2]|0;
    $314 = $310&255;
    HEAP8[$3>>0] = $314;
    $a$2 = $3;$fl$6 = $176;$p$5 = 1;$pl$2 = 0;$prefix$2 = 617481;$z$2 = $1;
    break;
   }
   case 109:  {
    $315 = (___errno_location()|0);
    $316 = HEAP32[$315>>2]|0;
    $317 = (_strerror($316)|0);
    $a$1 = $317;
    label = 82;
    break;
   }
   case 115:  {
    $318 = HEAP32[$arg>>2]|0;
    $319 = ($318|0)!=(0|0);
    $320 = $319 ? $318 : 619383;
    $a$1 = $320;
    label = 82;
    break;
   }
   case 67:  {
    $327 = $arg;
    $328 = $327;
    $329 = HEAP32[$328>>2]|0;
    $330 = (($327) + 4)|0;
    $331 = $330;
    $332 = HEAP32[$331>>2]|0;
    HEAP32[$wc>>2] = $329;
    HEAP32[$4>>2] = 0;
    HEAP32[$arg>>2] = $wc;
    $798 = $wc;$p$4176 = -1;
    label = 86;
    break;
   }
   case 83:  {
    $$pre171 = HEAP32[$arg>>2]|0;
    $333 = ($p$0|0)==(0);
    if ($333) {
     _pad($f,32,$w$1,0,$fl$1$);
     $i$0$lcssa178 = 0;
     label = 97;
    } else {
     $798 = $$pre171;$p$4176 = $p$0;
     label = 86;
    }
    break;
   }
   case 65: case 71: case 70: case 69: case 97: case 103: case 102: case 101:  {
    $358 = +HEAPF64[$arg>>3];
    HEAP32[$e2$i>>2] = 0;
    HEAPF64[tempDoublePtr>>3] = $358;$359 = HEAP32[tempDoublePtr>>2]|0;
    $360 = HEAP32[tempDoublePtr+4>>2]|0;
    $361 = ($360|0)<(0);
    if ($361) {
     $362 = -$358;
     $$07$i = $362;$pl$0$i = 1;$prefix$0$i = 619390;
    } else {
     $363 = $fl$1$ & 2048;
     $364 = ($363|0)==(0);
     if ($364) {
      $365 = $fl$1$ & 1;
      $366 = ($365|0)==(0);
      $$$i = $366 ? (619391) : (619396);
      $$07$i = $358;$pl$0$i = $365;$prefix$0$i = $$$i;
     } else {
      $$07$i = $358;$pl$0$i = 1;$prefix$0$i = (619393);
     }
    }
    HEAPF64[tempDoublePtr>>3] = $$07$i;$367 = HEAP32[tempDoublePtr>>2]|0;
    $368 = HEAP32[tempDoublePtr+4>>2]|0;
    $369 = $368 & 2146435072;
    $370 = ($369>>>0)<(2146435072);
    $371 = (0)<(0);
    $372 = ($369|0)==(2146435072);
    $373 = $372 & $371;
    $374 = $370 | $373;
    do {
     if ($374) {
      $390 = (+_frexpl($$07$i,$e2$i));
      $391 = $390 * 2.0;
      $392 = $391 != 0.0;
      if ($392) {
       $393 = HEAP32[$e2$i>>2]|0;
       $394 = (($393) + -1)|0;
       HEAP32[$e2$i>>2] = $394;
      }
      $395 = $t$0 | 32;
      $396 = ($395|0)==(97);
      if ($396) {
       $397 = $t$0 & 32;
       $398 = ($397|0)==(0);
       $399 = ((($prefix$0$i)) + 9|0);
       $prefix$0$$i = $398 ? $prefix$0$i : $399;
       $400 = $pl$0$i | 2;
       $401 = ($p$0>>>0)>(11);
       $402 = (12 - ($p$0))|0;
       $403 = ($402|0)==(0);
       $404 = $401 | $403;
       do {
        if ($404) {
         $$1$i = $391;
        } else {
         $re$171$i = $402;$round$070$i = 8.0;
         while(1) {
          $405 = (($re$171$i) + -1)|0;
          $406 = $round$070$i * 16.0;
          $407 = ($405|0)==(0);
          if ($407) {
           $$lcssa319 = $406;
           break;
          } else {
           $re$171$i = $405;$round$070$i = $406;
          }
         }
         $408 = HEAP8[$prefix$0$$i>>0]|0;
         $409 = ($408<<24>>24)==(45);
         if ($409) {
          $410 = -$391;
          $411 = $410 - $$lcssa319;
          $412 = $$lcssa319 + $411;
          $413 = -$412;
          $$1$i = $413;
          break;
         } else {
          $414 = $391 + $$lcssa319;
          $415 = $414 - $$lcssa319;
          $$1$i = $415;
          break;
         }
        }
       } while(0);
       $416 = HEAP32[$e2$i>>2]|0;
       $417 = ($416|0)<(0);
       $418 = (0 - ($416))|0;
       $419 = $417 ? $418 : $416;
       $420 = ($419|0)<(0);
       $421 = $420 << 31 >> 31;
       $422 = (_fmt_u($419,$421,$7)|0);
       $423 = ($422|0)==($7|0);
       if ($423) {
        HEAP8[$8>>0] = 48;
        $estr$0$i = $8;
       } else {
        $estr$0$i = $422;
       }
       $424 = $416 >> 31;
       $425 = $424 & 2;
       $426 = (($425) + 43)|0;
       $427 = $426&255;
       $428 = ((($estr$0$i)) + -1|0);
       HEAP8[$428>>0] = $427;
       $429 = (($t$0) + 15)|0;
       $430 = $429&255;
       $431 = ((($estr$0$i)) + -2|0);
       HEAP8[$431>>0] = $430;
       $notrhs$i = ($p$0|0)<(1);
       $432 = $fl$1$ & 8;
       $433 = ($432|0)==(0);
       $$2$i = $$1$i;$s$0$i = $buf$i;
       while(1) {
        $434 = (~~(($$2$i)));
        $435 = (617465 + ($434)|0);
        $436 = HEAP8[$435>>0]|0;
        $437 = $436&255;
        $438 = $437 | $397;
        $439 = $438&255;
        $440 = ((($s$0$i)) + 1|0);
        HEAP8[$s$0$i>>0] = $439;
        $441 = (+($434|0));
        $442 = $$2$i - $441;
        $443 = $442 * 16.0;
        $444 = $440;
        $445 = (($444) - ($5))|0;
        $446 = ($445|0)==(1);
        do {
         if ($446) {
          $notlhs$i = $443 == 0.0;
          $or$cond3$not$i = $notrhs$i & $notlhs$i;
          $or$cond$i = $433 & $or$cond3$not$i;
          if ($or$cond$i) {
           $s$1$i = $440;
           break;
          }
          $447 = ((($s$0$i)) + 2|0);
          HEAP8[$440>>0] = 46;
          $s$1$i = $447;
         } else {
          $s$1$i = $440;
         }
        } while(0);
        $448 = $443 != 0.0;
        if ($448) {
         $$2$i = $443;$s$0$i = $s$1$i;
        } else {
         $s$1$i$lcssa = $s$1$i;
         break;
        }
       }
       $449 = ($p$0|0)!=(0);
       $$pre188$i = $s$1$i$lcssa;
       $450 = (($11) + ($$pre188$i))|0;
       $451 = ($450|0)<($p$0|0);
       $or$cond122 = $449 & $451;
       $452 = $431;
       $453 = (($12) + ($p$0))|0;
       $454 = (($453) - ($452))|0;
       $455 = (($10) - ($452))|0;
       $456 = (($455) + ($$pre188$i))|0;
       $l$0$i = $or$cond122 ? $454 : $456;
       $457 = (($l$0$i) + ($400))|0;
       _pad($f,32,$w$1,$457,$fl$1$);
       $458 = HEAP32[$f>>2]|0;
       $459 = $458 & 32;
       $460 = ($459|0)==(0);
       if ($460) {
        (___fwritex($prefix$0$$i,$400,$f)|0);
       }
       $461 = $fl$1$ ^ 65536;
       _pad($f,48,$w$1,$457,$461);
       $462 = (($$pre188$i) - ($5))|0;
       $463 = HEAP32[$f>>2]|0;
       $464 = $463 & 32;
       $465 = ($464|0)==(0);
       if ($465) {
        (___fwritex($buf$i,$462,$f)|0);
       }
       $466 = (($9) - ($452))|0;
       $sum = (($462) + ($466))|0;
       $467 = (($l$0$i) - ($sum))|0;
       _pad($f,48,$467,0,0);
       $468 = HEAP32[$f>>2]|0;
       $469 = $468 & 32;
       $470 = ($469|0)==(0);
       if ($470) {
        (___fwritex($431,$466,$f)|0);
       }
       $471 = $fl$1$ ^ 8192;
       _pad($f,32,$w$1,$457,$471);
       $472 = ($457|0)<($w$1|0);
       $w$$i = $472 ? $w$1 : $457;
       $$0$i = $w$$i;
       break;
      }
      $473 = ($p$0|0)<(0);
      $$p$i = $473 ? 6 : $p$0;
      if ($392) {
       $474 = $391 * 268435456.0;
       $475 = HEAP32[$e2$i>>2]|0;
       $476 = (($475) + -28)|0;
       HEAP32[$e2$i>>2] = $476;
       $$3$i = $474;$478 = $476;
      } else {
       $$pre185$i = HEAP32[$e2$i>>2]|0;
       $$3$i = $391;$478 = $$pre185$i;
      }
      $477 = ($478|0)<(0);
      $$33$i = $477 ? $big$i : $13;
      $479 = $$33$i;
      $$4$i = $$3$i;$z$0$i = $$33$i;
      while(1) {
       $480 = (~~(($$4$i))>>>0);
       HEAP32[$z$0$i>>2] = $480;
       $481 = ((($z$0$i)) + 4|0);
       $482 = (+($480>>>0));
       $483 = $$4$i - $482;
       $484 = $483 * 1.0E+9;
       $485 = $484 != 0.0;
       if ($485) {
        $$4$i = $484;$z$0$i = $481;
       } else {
        $$lcssa303 = $481;
        break;
       }
      }
      $$pr$i = HEAP32[$e2$i>>2]|0;
      $486 = ($$pr$i|0)>(0);
      if ($486) {
       $488 = $$pr$i;$a$1149$i = $$33$i;$z$1148$i = $$lcssa303;
       while(1) {
        $487 = ($488|0)>(29);
        $489 = $487 ? 29 : $488;
        $d$0141$i = ((($z$1148$i)) + -4|0);
        $490 = ($d$0141$i>>>0)<($a$1149$i>>>0);
        do {
         if ($490) {
          $a$2$ph$i = $a$1149$i;
         } else {
          $carry$0142$i = 0;$d$0143$i = $d$0141$i;
          while(1) {
           $491 = HEAP32[$d$0143$i>>2]|0;
           $492 = (_bitshift64Shl(($491|0),0,($489|0))|0);
           $493 = tempRet0;
           $494 = (_i64Add(($492|0),($493|0),($carry$0142$i|0),0)|0);
           $495 = tempRet0;
           $496 = (___uremdi3(($494|0),($495|0),1000000000,0)|0);
           $497 = tempRet0;
           HEAP32[$d$0143$i>>2] = $496;
           $498 = (___udivdi3(($494|0),($495|0),1000000000,0)|0);
           $499 = tempRet0;
           $d$0$i = ((($d$0143$i)) + -4|0);
           $500 = ($d$0$i>>>0)<($a$1149$i>>>0);
           if ($500) {
            $$lcssa304 = $498;
            break;
           } else {
            $carry$0142$i = $498;$d$0143$i = $d$0$i;
           }
          }
          $501 = ($$lcssa304|0)==(0);
          if ($501) {
           $a$2$ph$i = $a$1149$i;
           break;
          }
          $502 = ((($a$1149$i)) + -4|0);
          HEAP32[$502>>2] = $$lcssa304;
          $a$2$ph$i = $502;
         }
        } while(0);
        $z$2$i = $z$1148$i;
        while(1) {
         $503 = ($z$2$i>>>0)>($a$2$ph$i>>>0);
         if (!($503)) {
          $z$2$i$lcssa = $z$2$i;
          break;
         }
         $504 = ((($z$2$i)) + -4|0);
         $505 = HEAP32[$504>>2]|0;
         $506 = ($505|0)==(0);
         if ($506) {
          $z$2$i = $504;
         } else {
          $z$2$i$lcssa = $z$2$i;
          break;
         }
        }
        $507 = HEAP32[$e2$i>>2]|0;
        $508 = (($507) - ($489))|0;
        HEAP32[$e2$i>>2] = $508;
        $509 = ($508|0)>(0);
        if ($509) {
         $488 = $508;$a$1149$i = $a$2$ph$i;$z$1148$i = $z$2$i$lcssa;
        } else {
         $$pr50$i = $508;$a$1$lcssa$i = $a$2$ph$i;$z$1$lcssa$i = $z$2$i$lcssa;
         break;
        }
       }
      } else {
       $$pr50$i = $$pr$i;$a$1$lcssa$i = $$33$i;$z$1$lcssa$i = $$lcssa303;
      }
      $510 = ($$pr50$i|0)<(0);
      if ($510) {
       $511 = (($$p$i) + 25)|0;
       $512 = (($511|0) / 9)&-1;
       $513 = (($512) + 1)|0;
       $514 = ($395|0)==(102);
       $516 = $$pr50$i;$a$3136$i = $a$1$lcssa$i;$z$3135$i = $z$1$lcssa$i;
       while(1) {
        $515 = (0 - ($516))|0;
        $517 = ($515|0)>(9);
        $518 = $517 ? 9 : $515;
        $519 = ($a$3136$i>>>0)<($z$3135$i>>>0);
        do {
         if ($519) {
          $523 = 1 << $518;
          $524 = (($523) + -1)|0;
          $525 = 1000000000 >>> $518;
          $carry3$0130$i = 0;$d$1129$i = $a$3136$i;
          while(1) {
           $526 = HEAP32[$d$1129$i>>2]|0;
           $527 = $526 & $524;
           $528 = $526 >>> $518;
           $529 = (($528) + ($carry3$0130$i))|0;
           HEAP32[$d$1129$i>>2] = $529;
           $530 = Math_imul($527, $525)|0;
           $531 = ((($d$1129$i)) + 4|0);
           $532 = ($531>>>0)<($z$3135$i>>>0);
           if ($532) {
            $carry3$0130$i = $530;$d$1129$i = $531;
           } else {
            $$lcssa306 = $530;
            break;
           }
          }
          $533 = HEAP32[$a$3136$i>>2]|0;
          $534 = ($533|0)==(0);
          $535 = ((($a$3136$i)) + 4|0);
          $$a$3$i = $534 ? $535 : $a$3136$i;
          $536 = ($$lcssa306|0)==(0);
          if ($536) {
           $$a$3192$i = $$a$3$i;$z$4$i = $z$3135$i;
           break;
          }
          $537 = ((($z$3135$i)) + 4|0);
          HEAP32[$z$3135$i>>2] = $$lcssa306;
          $$a$3192$i = $$a$3$i;$z$4$i = $537;
         } else {
          $520 = HEAP32[$a$3136$i>>2]|0;
          $521 = ($520|0)==(0);
          $522 = ((($a$3136$i)) + 4|0);
          $$a$3191$i = $521 ? $522 : $a$3136$i;
          $$a$3192$i = $$a$3191$i;$z$4$i = $z$3135$i;
         }
        } while(0);
        $538 = $514 ? $$33$i : $$a$3192$i;
        $539 = $z$4$i;
        $540 = $538;
        $541 = (($539) - ($540))|0;
        $542 = $541 >> 2;
        $543 = ($542|0)>($513|0);
        $544 = (($538) + ($513<<2)|0);
        $$z$4$i = $543 ? $544 : $z$4$i;
        $545 = HEAP32[$e2$i>>2]|0;
        $546 = (($545) + ($518))|0;
        HEAP32[$e2$i>>2] = $546;
        $547 = ($546|0)<(0);
        if ($547) {
         $516 = $546;$a$3136$i = $$a$3192$i;$z$3135$i = $$z$4$i;
        } else {
         $a$3$lcssa$i = $$a$3192$i;$z$3$lcssa$i = $$z$4$i;
         break;
        }
       }
      } else {
       $a$3$lcssa$i = $a$1$lcssa$i;$z$3$lcssa$i = $z$1$lcssa$i;
      }
      $548 = ($a$3$lcssa$i>>>0)<($z$3$lcssa$i>>>0);
      do {
       if ($548) {
        $549 = $a$3$lcssa$i;
        $550 = (($479) - ($549))|0;
        $551 = $550 >> 2;
        $552 = ($551*9)|0;
        $553 = HEAP32[$a$3$lcssa$i>>2]|0;
        $554 = ($553>>>0)<(10);
        if ($554) {
         $e$1$i = $552;
         break;
        } else {
         $e$0125$i = $552;$i$0124$i = 10;
        }
        while(1) {
         $555 = ($i$0124$i*10)|0;
         $556 = (($e$0125$i) + 1)|0;
         $557 = ($553>>>0)<($555>>>0);
         if ($557) {
          $e$1$i = $556;
          break;
         } else {
          $e$0125$i = $556;$i$0124$i = $555;
         }
        }
       } else {
        $e$1$i = 0;
       }
      } while(0);
      $558 = ($395|0)!=(102);
      $559 = $558 ? $e$1$i : 0;
      $560 = (($$p$i) - ($559))|0;
      $561 = ($395|0)==(103);
      $562 = ($$p$i|0)!=(0);
      $563 = $562 & $561;
      $$neg55$i = $563 << 31 >> 31;
      $564 = (($560) + ($$neg55$i))|0;
      $565 = $z$3$lcssa$i;
      $566 = (($565) - ($479))|0;
      $567 = $566 >> 2;
      $568 = ($567*9)|0;
      $569 = (($568) + -9)|0;
      $570 = ($564|0)<($569|0);
      if ($570) {
       $571 = ((($$33$i)) + 4|0);
       $572 = (($564) + 9216)|0;
       $573 = (($572|0) / 9)&-1;
       $574 = (($573) + -1024)|0;
       $575 = (($571) + ($574<<2)|0);
       $576 = (($572|0) % 9)&-1;
       $j$0117$i = (($576) + 1)|0;
       $577 = ($j$0117$i|0)<(9);
       if ($577) {
        $i$1118$i = 10;$j$0119$i = $j$0117$i;
        while(1) {
         $578 = ($i$1118$i*10)|0;
         $j$0$i = (($j$0119$i) + 1)|0;
         $exitcond$i = ($j$0$i|0)==(9);
         if ($exitcond$i) {
          $i$1$lcssa$i = $578;
          break;
         } else {
          $i$1118$i = $578;$j$0119$i = $j$0$i;
         }
        }
       } else {
        $i$1$lcssa$i = 10;
       }
       $579 = HEAP32[$575>>2]|0;
       $580 = (($579>>>0) % ($i$1$lcssa$i>>>0))&-1;
       $581 = ($580|0)==(0);
       $582 = ((($575)) + 4|0);
       $583 = ($582|0)==($z$3$lcssa$i|0);
       $or$cond18$i = $583 & $581;
       do {
        if ($or$cond18$i) {
         $a$8$i = $a$3$lcssa$i;$d$4$i = $575;$e$4$i = $e$1$i;
        } else {
         $584 = (($579>>>0) / ($i$1$lcssa$i>>>0))&-1;
         $585 = $584 & 1;
         $586 = ($585|0)==(0);
         $$20$i = $586 ? 9007199254740992.0 : 9007199254740994.0;
         $587 = (($i$1$lcssa$i|0) / 2)&-1;
         $588 = ($580>>>0)<($587>>>0);
         if ($588) {
          $small$0$i = 0.5;
         } else {
          $589 = ($580|0)==($587|0);
          $or$cond22$i = $583 & $589;
          $$36$i = $or$cond22$i ? 1.0 : 1.5;
          $small$0$i = $$36$i;
         }
         $590 = ($pl$0$i|0)==(0);
         do {
          if ($590) {
           $round6$1$i = $$20$i;$small$1$i = $small$0$i;
          } else {
           $591 = HEAP8[$prefix$0$i>>0]|0;
           $592 = ($591<<24>>24)==(45);
           if (!($592)) {
            $round6$1$i = $$20$i;$small$1$i = $small$0$i;
            break;
           }
           $593 = -$$20$i;
           $594 = -$small$0$i;
           $round6$1$i = $593;$small$1$i = $594;
          }
         } while(0);
         $595 = (($579) - ($580))|0;
         HEAP32[$575>>2] = $595;
         $596 = $round6$1$i + $small$1$i;
         $597 = $596 != $round6$1$i;
         if (!($597)) {
          $a$8$i = $a$3$lcssa$i;$d$4$i = $575;$e$4$i = $e$1$i;
          break;
         }
         $598 = (($595) + ($i$1$lcssa$i))|0;
         HEAP32[$575>>2] = $598;
         $599 = ($598>>>0)>(999999999);
         if ($599) {
          $a$5111$i = $a$3$lcssa$i;$d$2110$i = $575;
          while(1) {
           $600 = ((($d$2110$i)) + -4|0);
           HEAP32[$d$2110$i>>2] = 0;
           $601 = ($600>>>0)<($a$5111$i>>>0);
           if ($601) {
            $602 = ((($a$5111$i)) + -4|0);
            HEAP32[$602>>2] = 0;
            $a$6$i = $602;
           } else {
            $a$6$i = $a$5111$i;
           }
           $603 = HEAP32[$600>>2]|0;
           $604 = (($603) + 1)|0;
           HEAP32[$600>>2] = $604;
           $605 = ($604>>>0)>(999999999);
           if ($605) {
            $a$5111$i = $a$6$i;$d$2110$i = $600;
           } else {
            $a$5$lcssa$i = $a$6$i;$d$2$lcssa$i = $600;
            break;
           }
          }
         } else {
          $a$5$lcssa$i = $a$3$lcssa$i;$d$2$lcssa$i = $575;
         }
         $606 = $a$5$lcssa$i;
         $607 = (($479) - ($606))|0;
         $608 = $607 >> 2;
         $609 = ($608*9)|0;
         $610 = HEAP32[$a$5$lcssa$i>>2]|0;
         $611 = ($610>>>0)<(10);
         if ($611) {
          $a$8$i = $a$5$lcssa$i;$d$4$i = $d$2$lcssa$i;$e$4$i = $609;
          break;
         } else {
          $e$2106$i = $609;$i$2105$i = 10;
         }
         while(1) {
          $612 = ($i$2105$i*10)|0;
          $613 = (($e$2106$i) + 1)|0;
          $614 = ($610>>>0)<($612>>>0);
          if ($614) {
           $a$8$i = $a$5$lcssa$i;$d$4$i = $d$2$lcssa$i;$e$4$i = $613;
           break;
          } else {
           $e$2106$i = $613;$i$2105$i = $612;
          }
         }
        }
       } while(0);
       $615 = ((($d$4$i)) + 4|0);
       $616 = ($z$3$lcssa$i>>>0)>($615>>>0);
       $$z$3$i = $616 ? $615 : $z$3$lcssa$i;
       $a$9$ph$i = $a$8$i;$e$5$ph$i = $e$4$i;$z$7$ph$i = $$z$3$i;
      } else {
       $a$9$ph$i = $a$3$lcssa$i;$e$5$ph$i = $e$1$i;$z$7$ph$i = $z$3$lcssa$i;
      }
      $617 = (0 - ($e$5$ph$i))|0;
      $z$7$i = $z$7$ph$i;
      while(1) {
       $618 = ($z$7$i>>>0)>($a$9$ph$i>>>0);
       if (!($618)) {
        $$lcssa162$i = 0;$z$7$i$lcssa = $z$7$i;
        break;
       }
       $619 = ((($z$7$i)) + -4|0);
       $620 = HEAP32[$619>>2]|0;
       $621 = ($620|0)==(0);
       if ($621) {
        $z$7$i = $619;
       } else {
        $$lcssa162$i = 1;$z$7$i$lcssa = $z$7$i;
        break;
       }
      }
      do {
       if ($561) {
        $622 = $562&1;
        $623 = $622 ^ 1;
        $$p$$i = (($623) + ($$p$i))|0;
        $624 = ($$p$$i|0)>($e$5$ph$i|0);
        $625 = ($e$5$ph$i|0)>(-5);
        $or$cond6$i = $624 & $625;
        if ($or$cond6$i) {
         $626 = (($t$0) + -1)|0;
         $$neg56$i = (($$p$$i) + -1)|0;
         $627 = (($$neg56$i) - ($e$5$ph$i))|0;
         $$013$i = $626;$$210$i = $627;
        } else {
         $628 = (($t$0) + -2)|0;
         $629 = (($$p$$i) + -1)|0;
         $$013$i = $628;$$210$i = $629;
        }
        $630 = $fl$1$ & 8;
        $631 = ($630|0)==(0);
        if (!($631)) {
         $$114$i = $$013$i;$$311$i = $$210$i;$$pre$phi190$iZ2D = $630;
         break;
        }
        do {
         if ($$lcssa162$i) {
          $632 = ((($z$7$i$lcssa)) + -4|0);
          $633 = HEAP32[$632>>2]|0;
          $634 = ($633|0)==(0);
          if ($634) {
           $j$2$i = 9;
           break;
          }
          $635 = (($633>>>0) % 10)&-1;
          $636 = ($635|0)==(0);
          if ($636) {
           $i$3101$i = 10;$j$1102$i = 0;
          } else {
           $j$2$i = 0;
           break;
          }
          while(1) {
           $637 = ($i$3101$i*10)|0;
           $638 = (($j$1102$i) + 1)|0;
           $639 = (($633>>>0) % ($637>>>0))&-1;
           $640 = ($639|0)==(0);
           if ($640) {
            $i$3101$i = $637;$j$1102$i = $638;
           } else {
            $j$2$i = $638;
            break;
           }
          }
         } else {
          $j$2$i = 9;
         }
        } while(0);
        $641 = $$013$i | 32;
        $642 = ($641|0)==(102);
        $643 = $z$7$i$lcssa;
        $644 = (($643) - ($479))|0;
        $645 = $644 >> 2;
        $646 = ($645*9)|0;
        $647 = (($646) + -9)|0;
        if ($642) {
         $648 = (($647) - ($j$2$i))|0;
         $649 = ($648|0)<(0);
         $$23$i = $649 ? 0 : $648;
         $650 = ($$210$i|0)<($$23$i|0);
         $$210$$24$i = $650 ? $$210$i : $$23$i;
         $$114$i = $$013$i;$$311$i = $$210$$24$i;$$pre$phi190$iZ2D = 0;
         break;
        } else {
         $651 = (($647) + ($e$5$ph$i))|0;
         $652 = (($651) - ($j$2$i))|0;
         $653 = ($652|0)<(0);
         $$25$i = $653 ? 0 : $652;
         $654 = ($$210$i|0)<($$25$i|0);
         $$210$$26$i = $654 ? $$210$i : $$25$i;
         $$114$i = $$013$i;$$311$i = $$210$$26$i;$$pre$phi190$iZ2D = 0;
         break;
        }
       } else {
        $$pre189$i = $fl$1$ & 8;
        $$114$i = $t$0;$$311$i = $$p$i;$$pre$phi190$iZ2D = $$pre189$i;
       }
      } while(0);
      $655 = $$311$i | $$pre$phi190$iZ2D;
      $656 = ($655|0)!=(0);
      $657 = $656&1;
      $658 = $$114$i | 32;
      $659 = ($658|0)==(102);
      if ($659) {
       $660 = ($e$5$ph$i|0)>(0);
       $661 = $660 ? $e$5$ph$i : 0;
       $$pn$i = $661;$estr$2$i = 0;
      } else {
       $662 = ($e$5$ph$i|0)<(0);
       $663 = $662 ? $617 : $e$5$ph$i;
       $664 = ($663|0)<(0);
       $665 = $664 << 31 >> 31;
       $666 = (_fmt_u($663,$665,$7)|0);
       $667 = $666;
       $668 = (($9) - ($667))|0;
       $669 = ($668|0)<(2);
       if ($669) {
        $estr$195$i = $666;
        while(1) {
         $670 = ((($estr$195$i)) + -1|0);
         HEAP8[$670>>0] = 48;
         $671 = $670;
         $672 = (($9) - ($671))|0;
         $673 = ($672|0)<(2);
         if ($673) {
          $estr$195$i = $670;
         } else {
          $estr$1$lcssa$i = $670;
          break;
         }
        }
       } else {
        $estr$1$lcssa$i = $666;
       }
       $674 = $e$5$ph$i >> 31;
       $675 = $674 & 2;
       $676 = (($675) + 43)|0;
       $677 = $676&255;
       $678 = ((($estr$1$lcssa$i)) + -1|0);
       HEAP8[$678>>0] = $677;
       $679 = $$114$i&255;
       $680 = ((($estr$1$lcssa$i)) + -2|0);
       HEAP8[$680>>0] = $679;
       $681 = $680;
       $682 = (($9) - ($681))|0;
       $$pn$i = $682;$estr$2$i = $680;
      }
      $683 = (($pl$0$i) + 1)|0;
      $684 = (($683) + ($$311$i))|0;
      $l$1$i = (($684) + ($657))|0;
      $685 = (($l$1$i) + ($$pn$i))|0;
      _pad($f,32,$w$1,$685,$fl$1$);
      $686 = HEAP32[$f>>2]|0;
      $687 = $686 & 32;
      $688 = ($687|0)==(0);
      if ($688) {
       (___fwritex($prefix$0$i,$pl$0$i,$f)|0);
      }
      $689 = $fl$1$ ^ 65536;
      _pad($f,48,$w$1,$685,$689);
      do {
       if ($659) {
        $690 = ($a$9$ph$i>>>0)>($$33$i>>>0);
        $r$0$a$9$i = $690 ? $$33$i : $a$9$ph$i;
        $d$584$i = $r$0$a$9$i;
        while(1) {
         $691 = HEAP32[$d$584$i>>2]|0;
         $692 = (_fmt_u($691,0,$14)|0);
         $693 = ($d$584$i|0)==($r$0$a$9$i|0);
         do {
          if ($693) {
           $699 = ($692|0)==($14|0);
           if (!($699)) {
            $s7$1$i = $692;
            break;
           }
           HEAP8[$16>>0] = 48;
           $s7$1$i = $16;
          } else {
           $694 = ($692>>>0)>($buf$i>>>0);
           if (!($694)) {
            $s7$1$i = $692;
            break;
           }
           $695 = $692;
           $696 = (($695) - ($5))|0;
           _memset(($buf$i|0),48,($696|0))|0;
           $s7$081$i = $692;
           while(1) {
            $697 = ((($s7$081$i)) + -1|0);
            $698 = ($697>>>0)>($buf$i>>>0);
            if ($698) {
             $s7$081$i = $697;
            } else {
             $s7$1$i = $697;
             break;
            }
           }
          }
         } while(0);
         $700 = HEAP32[$f>>2]|0;
         $701 = $700 & 32;
         $702 = ($701|0)==(0);
         if ($702) {
          $703 = $s7$1$i;
          $704 = (($15) - ($703))|0;
          (___fwritex($s7$1$i,$704,$f)|0);
         }
         $705 = ((($d$584$i)) + 4|0);
         $706 = ($705>>>0)>($$33$i>>>0);
         if ($706) {
          $$lcssa316 = $705;
          break;
         } else {
          $d$584$i = $705;
         }
        }
        $707 = ($655|0)==(0);
        do {
         if (!($707)) {
          $708 = HEAP32[$f>>2]|0;
          $709 = $708 & 32;
          $710 = ($709|0)==(0);
          if (!($710)) {
           break;
          }
          (___fwritex(619449,1,$f)|0);
         }
        } while(0);
        $711 = ($$lcssa316>>>0)<($z$7$i$lcssa>>>0);
        $712 = ($$311$i|0)>(0);
        $713 = $712 & $711;
        if ($713) {
         $$41278$i = $$311$i;$d$677$i = $$lcssa316;
         while(1) {
          $714 = HEAP32[$d$677$i>>2]|0;
          $715 = (_fmt_u($714,0,$14)|0);
          $716 = ($715>>>0)>($buf$i>>>0);
          if ($716) {
           $717 = $715;
           $718 = (($717) - ($5))|0;
           _memset(($buf$i|0),48,($718|0))|0;
           $s8$072$i = $715;
           while(1) {
            $719 = ((($s8$072$i)) + -1|0);
            $720 = ($719>>>0)>($buf$i>>>0);
            if ($720) {
             $s8$072$i = $719;
            } else {
             $s8$0$lcssa$i = $719;
             break;
            }
           }
          } else {
           $s8$0$lcssa$i = $715;
          }
          $721 = HEAP32[$f>>2]|0;
          $722 = $721 & 32;
          $723 = ($722|0)==(0);
          if ($723) {
           $724 = ($$41278$i|0)>(9);
           $725 = $724 ? 9 : $$41278$i;
           (___fwritex($s8$0$lcssa$i,$725,$f)|0);
          }
          $726 = ((($d$677$i)) + 4|0);
          $727 = (($$41278$i) + -9)|0;
          $728 = ($726>>>0)<($z$7$i$lcssa>>>0);
          $729 = ($$41278$i|0)>(9);
          $730 = $729 & $728;
          if ($730) {
           $$41278$i = $727;$d$677$i = $726;
          } else {
           $$412$lcssa$i = $727;
           break;
          }
         }
        } else {
         $$412$lcssa$i = $$311$i;
        }
        $731 = (($$412$lcssa$i) + 9)|0;
        _pad($f,48,$731,9,0);
       } else {
        $732 = ((($a$9$ph$i)) + 4|0);
        $z$7$$i = $$lcssa162$i ? $z$7$i$lcssa : $732;
        $733 = ($$311$i|0)>(-1);
        if ($733) {
         $734 = ($$pre$phi190$iZ2D|0)==(0);
         $$589$i = $$311$i;$d$788$i = $a$9$ph$i;
         while(1) {
          $735 = HEAP32[$d$788$i>>2]|0;
          $736 = (_fmt_u($735,0,$14)|0);
          $737 = ($736|0)==($14|0);
          if ($737) {
           HEAP8[$16>>0] = 48;
           $s9$0$i = $16;
          } else {
           $s9$0$i = $736;
          }
          $738 = ($d$788$i|0)==($a$9$ph$i|0);
          do {
           if ($738) {
            $742 = ((($s9$0$i)) + 1|0);
            $743 = HEAP32[$f>>2]|0;
            $744 = $743 & 32;
            $745 = ($744|0)==(0);
            if ($745) {
             (___fwritex($s9$0$i,1,$f)|0);
            }
            $746 = ($$589$i|0)<(1);
            $or$cond31$i = $734 & $746;
            if ($or$cond31$i) {
             $s9$2$i = $742;
             break;
            }
            $747 = HEAP32[$f>>2]|0;
            $748 = $747 & 32;
            $749 = ($748|0)==(0);
            if (!($749)) {
             $s9$2$i = $742;
             break;
            }
            (___fwritex(619449,1,$f)|0);
            $s9$2$i = $742;
           } else {
            $739 = ($s9$0$i>>>0)>($buf$i>>>0);
            if (!($739)) {
             $s9$2$i = $s9$0$i;
             break;
            }
            $scevgep182$i = (($s9$0$i) + ($6)|0);
            $scevgep182183$i = $scevgep182$i;
            _memset(($buf$i|0),48,($scevgep182183$i|0))|0;
            $s9$185$i = $s9$0$i;
            while(1) {
             $740 = ((($s9$185$i)) + -1|0);
             $741 = ($740>>>0)>($buf$i>>>0);
             if ($741) {
              $s9$185$i = $740;
             } else {
              $s9$2$i = $740;
              break;
             }
            }
           }
          } while(0);
          $750 = $s9$2$i;
          $751 = (($15) - ($750))|0;
          $752 = HEAP32[$f>>2]|0;
          $753 = $752 & 32;
          $754 = ($753|0)==(0);
          if ($754) {
           $755 = ($$589$i|0)>($751|0);
           $756 = $755 ? $751 : $$589$i;
           (___fwritex($s9$2$i,$756,$f)|0);
          }
          $757 = (($$589$i) - ($751))|0;
          $758 = ((($d$788$i)) + 4|0);
          $759 = ($758>>>0)<($z$7$$i>>>0);
          $760 = ($757|0)>(-1);
          $761 = $759 & $760;
          if ($761) {
           $$589$i = $757;$d$788$i = $758;
          } else {
           $$5$lcssa$i = $757;
           break;
          }
         }
        } else {
         $$5$lcssa$i = $$311$i;
        }
        $762 = (($$5$lcssa$i) + 18)|0;
        _pad($f,48,$762,18,0);
        $763 = HEAP32[$f>>2]|0;
        $764 = $763 & 32;
        $765 = ($764|0)==(0);
        if (!($765)) {
         break;
        }
        $766 = $estr$2$i;
        $767 = (($9) - ($766))|0;
        (___fwritex($estr$2$i,$767,$f)|0);
       }
      } while(0);
      $768 = $fl$1$ ^ 8192;
      _pad($f,32,$w$1,$685,$768);
      $769 = ($685|0)<($w$1|0);
      $w$32$i = $769 ? $w$1 : $685;
      $$0$i = $w$32$i;
     } else {
      $375 = $t$0 & 32;
      $376 = ($375|0)!=(0);
      $377 = $376 ? 619409 : 619413;
      $378 = ($$07$i != $$07$i) | (0.0 != 0.0);
      $379 = $376 ? 619445 : 619417;
      $pl$1$i = $378 ? 0 : $pl$0$i;
      $s1$0$i = $378 ? $379 : $377;
      $380 = (($pl$1$i) + 3)|0;
      _pad($f,32,$w$1,$380,$176);
      $381 = HEAP32[$f>>2]|0;
      $382 = $381 & 32;
      $383 = ($382|0)==(0);
      if ($383) {
       (___fwritex($prefix$0$i,$pl$1$i,$f)|0);
       $$pre$i = HEAP32[$f>>2]|0;
       $385 = $$pre$i;
      } else {
       $385 = $381;
      }
      $384 = $385 & 32;
      $386 = ($384|0)==(0);
      if ($386) {
       (___fwritex($s1$0$i,3,$f)|0);
      }
      $387 = $fl$1$ ^ 8192;
      _pad($f,32,$w$1,$380,$387);
      $388 = ($380|0)<($w$1|0);
      $389 = $388 ? $w$1 : $380;
      $$0$i = $389;
     }
    } while(0);
    $cnt$0 = $cnt$1;$l$0 = $$0$i;$l10n$0 = $l10n$3;$s$0 = $$lcssa300;
    continue L1;
    break;
   }
   default: {
    $a$2 = $s$0;$fl$6 = $fl$1$;$p$5 = $p$0;$pl$2 = 0;$prefix$2 = 617481;$z$2 = $1;
   }
   }
  } while(0);
  L311: do {
   if ((label|0) == 64) {
    label = 0;
    $207 = $arg;
    $208 = $207;
    $209 = HEAP32[$208>>2]|0;
    $210 = (($207) + 4)|0;
    $211 = $210;
    $212 = HEAP32[$211>>2]|0;
    $213 = $t$1 & 32;
    $214 = ($209|0)==(0);
    $215 = ($212|0)==(0);
    $216 = $214 & $215;
    if ($216) {
     $a$0 = $1;$fl$4 = $fl$3;$p$2 = $p$1;$pl$1 = 0;$prefix$1 = 617481;
     label = 77;
    } else {
     $$012$i = $1;$218 = $209;$225 = $212;
     while(1) {
      $217 = $218 & 15;
      $219 = (617465 + ($217)|0);
      $220 = HEAP8[$219>>0]|0;
      $221 = $220&255;
      $222 = $221 | $213;
      $223 = $222&255;
      $224 = ((($$012$i)) + -1|0);
      HEAP8[$224>>0] = $223;
      $226 = (_bitshift64Lshr(($218|0),($225|0),4)|0);
      $227 = tempRet0;
      $228 = ($226|0)==(0);
      $229 = ($227|0)==(0);
      $230 = $228 & $229;
      if ($230) {
       $$lcssa321 = $224;
       break;
      } else {
       $$012$i = $224;$218 = $226;$225 = $227;
      }
     }
     $231 = $arg;
     $232 = $231;
     $233 = HEAP32[$232>>2]|0;
     $234 = (($231) + 4)|0;
     $235 = $234;
     $236 = HEAP32[$235>>2]|0;
     $237 = ($233|0)==(0);
     $238 = ($236|0)==(0);
     $239 = $237 & $238;
     $240 = $fl$3 & 8;
     $241 = ($240|0)==(0);
     $or$cond17 = $241 | $239;
     if ($or$cond17) {
      $a$0 = $$lcssa321;$fl$4 = $fl$3;$p$2 = $p$1;$pl$1 = 0;$prefix$1 = 617481;
      label = 77;
     } else {
      $242 = $t$1 >> 4;
      $243 = (617481 + ($242)|0);
      $a$0 = $$lcssa321;$fl$4 = $fl$3;$p$2 = $p$1;$pl$1 = 2;$prefix$1 = $243;
      label = 77;
     }
    }
   }
   else if ((label|0) == 76) {
    label = 0;
    $289 = (_fmt_u($287,$288,$1)|0);
    $a$0 = $289;$fl$4 = $fl$1$;$p$2 = $p$0;$pl$1 = $pl$0;$prefix$1 = $prefix$0;
    label = 77;
   }
   else if ((label|0) == 82) {
    label = 0;
    $321 = (_memchr($a$1,0,$p$0)|0);
    $322 = ($321|0)==(0|0);
    $323 = $321;
    $324 = $a$1;
    $325 = (($323) - ($324))|0;
    $326 = (($a$1) + ($p$0)|0);
    $z$1 = $322 ? $326 : $321;
    $p$3 = $322 ? $p$0 : $325;
    $a$2 = $a$1;$fl$6 = $176;$p$5 = $p$3;$pl$2 = 0;$prefix$2 = 617481;$z$2 = $z$1;
   }
   else if ((label|0) == 86) {
    label = 0;
    $i$0105 = 0;$l$1104 = 0;$ws$0106 = $798;
    while(1) {
     $334 = HEAP32[$ws$0106>>2]|0;
     $335 = ($334|0)==(0);
     if ($335) {
      $i$0$lcssa = $i$0105;$l$2 = $l$1104;
      break;
     }
     $336 = (_wctomb($mb,$334)|0);
     $337 = ($336|0)<(0);
     $338 = (($p$4176) - ($i$0105))|0;
     $339 = ($336>>>0)>($338>>>0);
     $or$cond20 = $337 | $339;
     if ($or$cond20) {
      $i$0$lcssa = $i$0105;$l$2 = $336;
      break;
     }
     $340 = ((($ws$0106)) + 4|0);
     $341 = (($336) + ($i$0105))|0;
     $342 = ($p$4176>>>0)>($341>>>0);
     if ($342) {
      $i$0105 = $341;$l$1104 = $336;$ws$0106 = $340;
     } else {
      $i$0$lcssa = $341;$l$2 = $336;
      break;
     }
    }
    $343 = ($l$2|0)<(0);
    if ($343) {
     $$0 = -1;
     break L1;
    }
    _pad($f,32,$w$1,$i$0$lcssa,$fl$1$);
    $344 = ($i$0$lcssa|0)==(0);
    if ($344) {
     $i$0$lcssa178 = 0;
     label = 97;
    } else {
     $i$1116 = 0;$ws$1117 = $798;
     while(1) {
      $345 = HEAP32[$ws$1117>>2]|0;
      $346 = ($345|0)==(0);
      if ($346) {
       $i$0$lcssa178 = $i$0$lcssa;
       label = 97;
       break L311;
      }
      $347 = ((($ws$1117)) + 4|0);
      $348 = (_wctomb($mb,$345)|0);
      $349 = (($348) + ($i$1116))|0;
      $350 = ($349|0)>($i$0$lcssa|0);
      if ($350) {
       $i$0$lcssa178 = $i$0$lcssa;
       label = 97;
       break L311;
      }
      $351 = HEAP32[$f>>2]|0;
      $352 = $351 & 32;
      $353 = ($352|0)==(0);
      if ($353) {
       (___fwritex($mb,$348,$f)|0);
      }
      $354 = ($349>>>0)<($i$0$lcssa>>>0);
      if ($354) {
       $i$1116 = $349;$ws$1117 = $347;
      } else {
       $i$0$lcssa178 = $i$0$lcssa;
       label = 97;
       break;
      }
     }
    }
   }
  } while(0);
  if ((label|0) == 97) {
   label = 0;
   $355 = $fl$1$ ^ 8192;
   _pad($f,32,$w$1,$i$0$lcssa178,$355);
   $356 = ($w$1|0)>($i$0$lcssa178|0);
   $357 = $356 ? $w$1 : $i$0$lcssa178;
   $cnt$0 = $cnt$1;$l$0 = $357;$l10n$0 = $l10n$3;$s$0 = $$lcssa300;
   continue;
  }
  if ((label|0) == 77) {
   label = 0;
   $290 = ($p$2|0)>(-1);
   $291 = $fl$4 & -65537;
   $$fl$4 = $290 ? $291 : $fl$4;
   $292 = $arg;
   $293 = $292;
   $294 = HEAP32[$293>>2]|0;
   $295 = (($292) + 4)|0;
   $296 = $295;
   $297 = HEAP32[$296>>2]|0;
   $298 = ($294|0)!=(0);
   $299 = ($297|0)!=(0);
   $300 = $298 | $299;
   $301 = ($p$2|0)!=(0);
   $or$cond = $301 | $300;
   if ($or$cond) {
    $302 = $a$0;
    $303 = (($2) - ($302))|0;
    $304 = $300&1;
    $305 = $304 ^ 1;
    $306 = (($305) + ($303))|0;
    $307 = ($p$2|0)>($306|0);
    $p$2$ = $307 ? $p$2 : $306;
    $a$2 = $a$0;$fl$6 = $$fl$4;$p$5 = $p$2$;$pl$2 = $pl$1;$prefix$2 = $prefix$1;$z$2 = $1;
   } else {
    $a$2 = $1;$fl$6 = $$fl$4;$p$5 = 0;$pl$2 = $pl$1;$prefix$2 = $prefix$1;$z$2 = $1;
   }
  }
  $770 = $z$2;
  $771 = $a$2;
  $772 = (($770) - ($771))|0;
  $773 = ($p$5|0)<($772|0);
  $$p$5 = $773 ? $772 : $p$5;
  $774 = (($pl$2) + ($$p$5))|0;
  $775 = ($w$1|0)<($774|0);
  $w$2 = $775 ? $774 : $w$1;
  _pad($f,32,$w$2,$774,$fl$6);
  $776 = HEAP32[$f>>2]|0;
  $777 = $776 & 32;
  $778 = ($777|0)==(0);
  if ($778) {
   (___fwritex($prefix$2,$pl$2,$f)|0);
  }
  $779 = $fl$6 ^ 65536;
  _pad($f,48,$w$2,$774,$779);
  _pad($f,48,$$p$5,$772,0);
  $780 = HEAP32[$f>>2]|0;
  $781 = $780 & 32;
  $782 = ($781|0)==(0);
  if ($782) {
   (___fwritex($a$2,$772,$f)|0);
  }
  $783 = $fl$6 ^ 8192;
  _pad($f,32,$w$2,$774,$783);
  $cnt$0 = $cnt$1;$l$0 = $w$2;$l10n$0 = $l10n$3;$s$0 = $$lcssa300;
 }
 L345: do {
  if ((label|0) == 244) {
   $784 = ($f|0)==(0|0);
   if ($784) {
    $785 = ($l10n$0$lcssa|0)==(0);
    if ($785) {
     $$0 = 0;
    } else {
     $i$291 = 1;
     while(1) {
      $786 = (($nl_type) + ($i$291<<2)|0);
      $787 = HEAP32[$786>>2]|0;
      $788 = ($787|0)==(0);
      if ($788) {
       $i$291$lcssa = $i$291;
       break;
      }
      $790 = (($nl_arg) + ($i$291<<3)|0);
      _pop_arg($790,$787,$ap);
      $791 = (($i$291) + 1)|0;
      $792 = ($791|0)<(10);
      if ($792) {
       $i$291 = $791;
      } else {
       $$0 = 1;
       break L345;
      }
     }
     $789 = ($i$291$lcssa|0)<(10);
     if ($789) {
      $i$389 = $i$291$lcssa;
      while(1) {
       $795 = (($nl_type) + ($i$389<<2)|0);
       $796 = HEAP32[$795>>2]|0;
       $797 = ($796|0)==(0);
       $794 = (($i$389) + 1)|0;
       if (!($797)) {
        $$0 = -1;
        break L345;
       }
       $793 = ($794|0)<(10);
       if ($793) {
        $i$389 = $794;
       } else {
        $$0 = 1;
        break;
       }
      }
     } else {
      $$0 = 1;
     }
    }
   } else {
    $$0 = $cnt$1$lcssa;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function ___fwritex($s,$l,$f) {
 $s = $s|0;
 $l = $l|0;
 $f = $f|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $$pre = 0, $$pre6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0, $i$0$lcssa12 = 0;
 var $i$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $4 = (___towrite($f)|0);
  $5 = ($4|0)==(0);
  if ($5) {
   $$pre = HEAP32[$0>>2]|0;
   $9 = $$pre;
   label = 5;
  } else {
   $$0 = 0;
  }
 } else {
  $3 = $1;
  $9 = $3;
  label = 5;
 }
 L5: do {
  if ((label|0) == 5) {
   $6 = ((($f)) + 20|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = (($9) - ($7))|0;
   $10 = ($8>>>0)<($l>>>0);
   $11 = $7;
   if ($10) {
    $12 = ((($f)) + 36|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = (FUNCTION_TABLE_iiii[$13 & 8191]($f,$s,$l)|0);
    $$0 = $14;
    break;
   }
   $15 = ((($f)) + 75|0);
   $16 = HEAP8[$15>>0]|0;
   $17 = ($16<<24>>24)>(-1);
   L10: do {
    if ($17) {
     $i$0 = $l;
     while(1) {
      $18 = ($i$0|0)==(0);
      if ($18) {
       $$01 = $l;$$02 = $s;$29 = $11;$i$1 = 0;
       break L10;
      }
      $19 = (($i$0) + -1)|0;
      $20 = (($s) + ($19)|0);
      $21 = HEAP8[$20>>0]|0;
      $22 = ($21<<24>>24)==(10);
      if ($22) {
       $i$0$lcssa12 = $i$0;
       break;
      } else {
       $i$0 = $19;
      }
     }
     $23 = ((($f)) + 36|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = (FUNCTION_TABLE_iiii[$24 & 8191]($f,$s,$i$0$lcssa12)|0);
     $26 = ($25>>>0)<($i$0$lcssa12>>>0);
     if ($26) {
      $$0 = $i$0$lcssa12;
      break L5;
     }
     $27 = (($s) + ($i$0$lcssa12)|0);
     $28 = (($l) - ($i$0$lcssa12))|0;
     $$pre6 = HEAP32[$6>>2]|0;
     $$01 = $28;$$02 = $27;$29 = $$pre6;$i$1 = $i$0$lcssa12;
    } else {
     $$01 = $l;$$02 = $s;$29 = $11;$i$1 = 0;
    }
   } while(0);
   _memcpy(($29|0),($$02|0),($$01|0))|0;
   $30 = HEAP32[$6>>2]|0;
   $31 = (($30) + ($$01)|0);
   HEAP32[$6>>2] = $31;
   $32 = (($i$1) + ($$01))|0;
   $$0 = $32;
  }
 } while(0);
 return ($$0|0);
}
function ___towrite($f) {
 $f = $f|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 74|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = $1 << 24 >> 24;
 $3 = (($2) + 255)|0;
 $4 = $3 | $2;
 $5 = $4&255;
 HEAP8[$0>>0] = $5;
 $6 = HEAP32[$f>>2]|0;
 $7 = $6 & 8;
 $8 = ($7|0)==(0);
 if ($8) {
  $10 = ((($f)) + 8|0);
  HEAP32[$10>>2] = 0;
  $11 = ((($f)) + 4|0);
  HEAP32[$11>>2] = 0;
  $12 = ((($f)) + 44|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ((($f)) + 28|0);
  HEAP32[$14>>2] = $13;
  $15 = ((($f)) + 20|0);
  HEAP32[$15>>2] = $13;
  $16 = $13;
  $17 = ((($f)) + 48|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = (($16) + ($18)|0);
  $20 = ((($f)) + 16|0);
  HEAP32[$20>>2] = $19;
  $$0 = 0;
 } else {
  $9 = $6 | 32;
  HEAP32[$f>>2] = $9;
  $$0 = -1;
 }
 return ($$0|0);
}
function _pop_arg($arg,$type,$ap) {
 $arg = $arg|0;
 $type = $type|0;
 $ap = $ap|0;
 var $$mask = 0, $$mask1 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0.0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0.0;
 var $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $arglist_current = 0, $arglist_current11 = 0, $arglist_current14 = 0, $arglist_current17 = 0;
 var $arglist_current2 = 0, $arglist_current20 = 0, $arglist_current23 = 0, $arglist_current26 = 0, $arglist_current5 = 0, $arglist_current8 = 0, $arglist_next = 0, $arglist_next12 = 0, $arglist_next15 = 0, $arglist_next18 = 0, $arglist_next21 = 0, $arglist_next24 = 0, $arglist_next27 = 0, $arglist_next3 = 0, $arglist_next6 = 0, $arglist_next9 = 0, $expanded = 0, $expanded28 = 0, $expanded30 = 0, $expanded31 = 0;
 var $expanded32 = 0, $expanded34 = 0, $expanded35 = 0, $expanded37 = 0, $expanded38 = 0, $expanded39 = 0, $expanded41 = 0, $expanded42 = 0, $expanded44 = 0, $expanded45 = 0, $expanded46 = 0, $expanded48 = 0, $expanded49 = 0, $expanded51 = 0, $expanded52 = 0, $expanded53 = 0, $expanded55 = 0, $expanded56 = 0, $expanded58 = 0, $expanded59 = 0;
 var $expanded60 = 0, $expanded62 = 0, $expanded63 = 0, $expanded65 = 0, $expanded66 = 0, $expanded67 = 0, $expanded69 = 0, $expanded70 = 0, $expanded72 = 0, $expanded73 = 0, $expanded74 = 0, $expanded76 = 0, $expanded77 = 0, $expanded79 = 0, $expanded80 = 0, $expanded81 = 0, $expanded83 = 0, $expanded84 = 0, $expanded86 = 0, $expanded87 = 0;
 var $expanded88 = 0, $expanded90 = 0, $expanded91 = 0, $expanded93 = 0, $expanded94 = 0, $expanded95 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($type>>>0)>(20);
 L1: do {
  if (!($0)) {
   do {
    switch ($type|0) {
    case 9:  {
     $arglist_current = HEAP32[$ap>>2]|0;
     $1 = $arglist_current;
     $2 = ((0) + 4|0);
     $expanded28 = $2;
     $expanded = (($expanded28) - 1)|0;
     $3 = (($1) + ($expanded))|0;
     $4 = ((0) + 4|0);
     $expanded32 = $4;
     $expanded31 = (($expanded32) - 1)|0;
     $expanded30 = $expanded31 ^ -1;
     $5 = $3 & $expanded30;
     $6 = $5;
     $7 = HEAP32[$6>>2]|0;
     $arglist_next = ((($6)) + 4|0);
     HEAP32[$ap>>2] = $arglist_next;
     HEAP32[$arg>>2] = $7;
     break L1;
     break;
    }
    case 10:  {
     $arglist_current2 = HEAP32[$ap>>2]|0;
     $8 = $arglist_current2;
     $9 = ((0) + 4|0);
     $expanded35 = $9;
     $expanded34 = (($expanded35) - 1)|0;
     $10 = (($8) + ($expanded34))|0;
     $11 = ((0) + 4|0);
     $expanded39 = $11;
     $expanded38 = (($expanded39) - 1)|0;
     $expanded37 = $expanded38 ^ -1;
     $12 = $10 & $expanded37;
     $13 = $12;
     $14 = HEAP32[$13>>2]|0;
     $arglist_next3 = ((($13)) + 4|0);
     HEAP32[$ap>>2] = $arglist_next3;
     $15 = ($14|0)<(0);
     $16 = $15 << 31 >> 31;
     $17 = $arg;
     $18 = $17;
     HEAP32[$18>>2] = $14;
     $19 = (($17) + 4)|0;
     $20 = $19;
     HEAP32[$20>>2] = $16;
     break L1;
     break;
    }
    case 11:  {
     $arglist_current5 = HEAP32[$ap>>2]|0;
     $21 = $arglist_current5;
     $22 = ((0) + 4|0);
     $expanded42 = $22;
     $expanded41 = (($expanded42) - 1)|0;
     $23 = (($21) + ($expanded41))|0;
     $24 = ((0) + 4|0);
     $expanded46 = $24;
     $expanded45 = (($expanded46) - 1)|0;
     $expanded44 = $expanded45 ^ -1;
     $25 = $23 & $expanded44;
     $26 = $25;
     $27 = HEAP32[$26>>2]|0;
     $arglist_next6 = ((($26)) + 4|0);
     HEAP32[$ap>>2] = $arglist_next6;
     $28 = $arg;
     $29 = $28;
     HEAP32[$29>>2] = $27;
     $30 = (($28) + 4)|0;
     $31 = $30;
     HEAP32[$31>>2] = 0;
     break L1;
     break;
    }
    case 12:  {
     $arglist_current8 = HEAP32[$ap>>2]|0;
     $32 = $arglist_current8;
     $33 = ((0) + 8|0);
     $expanded49 = $33;
     $expanded48 = (($expanded49) - 1)|0;
     $34 = (($32) + ($expanded48))|0;
     $35 = ((0) + 8|0);
     $expanded53 = $35;
     $expanded52 = (($expanded53) - 1)|0;
     $expanded51 = $expanded52 ^ -1;
     $36 = $34 & $expanded51;
     $37 = $36;
     $38 = $37;
     $39 = $38;
     $40 = HEAP32[$39>>2]|0;
     $41 = (($38) + 4)|0;
     $42 = $41;
     $43 = HEAP32[$42>>2]|0;
     $arglist_next9 = ((($37)) + 8|0);
     HEAP32[$ap>>2] = $arglist_next9;
     $44 = $arg;
     $45 = $44;
     HEAP32[$45>>2] = $40;
     $46 = (($44) + 4)|0;
     $47 = $46;
     HEAP32[$47>>2] = $43;
     break L1;
     break;
    }
    case 13:  {
     $arglist_current11 = HEAP32[$ap>>2]|0;
     $48 = $arglist_current11;
     $49 = ((0) + 4|0);
     $expanded56 = $49;
     $expanded55 = (($expanded56) - 1)|0;
     $50 = (($48) + ($expanded55))|0;
     $51 = ((0) + 4|0);
     $expanded60 = $51;
     $expanded59 = (($expanded60) - 1)|0;
     $expanded58 = $expanded59 ^ -1;
     $52 = $50 & $expanded58;
     $53 = $52;
     $54 = HEAP32[$53>>2]|0;
     $arglist_next12 = ((($53)) + 4|0);
     HEAP32[$ap>>2] = $arglist_next12;
     $55 = $54&65535;
     $56 = $55 << 16 >> 16;
     $57 = ($56|0)<(0);
     $58 = $57 << 31 >> 31;
     $59 = $arg;
     $60 = $59;
     HEAP32[$60>>2] = $56;
     $61 = (($59) + 4)|0;
     $62 = $61;
     HEAP32[$62>>2] = $58;
     break L1;
     break;
    }
    case 14:  {
     $arglist_current14 = HEAP32[$ap>>2]|0;
     $63 = $arglist_current14;
     $64 = ((0) + 4|0);
     $expanded63 = $64;
     $expanded62 = (($expanded63) - 1)|0;
     $65 = (($63) + ($expanded62))|0;
     $66 = ((0) + 4|0);
     $expanded67 = $66;
     $expanded66 = (($expanded67) - 1)|0;
     $expanded65 = $expanded66 ^ -1;
     $67 = $65 & $expanded65;
     $68 = $67;
     $69 = HEAP32[$68>>2]|0;
     $arglist_next15 = ((($68)) + 4|0);
     HEAP32[$ap>>2] = $arglist_next15;
     $$mask1 = $69 & 65535;
     $70 = $arg;
     $71 = $70;
     HEAP32[$71>>2] = $$mask1;
     $72 = (($70) + 4)|0;
     $73 = $72;
     HEAP32[$73>>2] = 0;
     break L1;
     break;
    }
    case 15:  {
     $arglist_current17 = HEAP32[$ap>>2]|0;
     $74 = $arglist_current17;
     $75 = ((0) + 4|0);
     $expanded70 = $75;
     $expanded69 = (($expanded70) - 1)|0;
     $76 = (($74) + ($expanded69))|0;
     $77 = ((0) + 4|0);
     $expanded74 = $77;
     $expanded73 = (($expanded74) - 1)|0;
     $expanded72 = $expanded73 ^ -1;
     $78 = $76 & $expanded72;
     $79 = $78;
     $80 = HEAP32[$79>>2]|0;
     $arglist_next18 = ((($79)) + 4|0);
     HEAP32[$ap>>2] = $arglist_next18;
     $81 = $80&255;
     $82 = $81 << 24 >> 24;
     $83 = ($82|0)<(0);
     $84 = $83 << 31 >> 31;
     $85 = $arg;
     $86 = $85;
     HEAP32[$86>>2] = $82;
     $87 = (($85) + 4)|0;
     $88 = $87;
     HEAP32[$88>>2] = $84;
     break L1;
     break;
    }
    case 16:  {
     $arglist_current20 = HEAP32[$ap>>2]|0;
     $89 = $arglist_current20;
     $90 = ((0) + 4|0);
     $expanded77 = $90;
     $expanded76 = (($expanded77) - 1)|0;
     $91 = (($89) + ($expanded76))|0;
     $92 = ((0) + 4|0);
     $expanded81 = $92;
     $expanded80 = (($expanded81) - 1)|0;
     $expanded79 = $expanded80 ^ -1;
     $93 = $91 & $expanded79;
     $94 = $93;
     $95 = HEAP32[$94>>2]|0;
     $arglist_next21 = ((($94)) + 4|0);
     HEAP32[$ap>>2] = $arglist_next21;
     $$mask = $95 & 255;
     $96 = $arg;
     $97 = $96;
     HEAP32[$97>>2] = $$mask;
     $98 = (($96) + 4)|0;
     $99 = $98;
     HEAP32[$99>>2] = 0;
     break L1;
     break;
    }
    case 17:  {
     $arglist_current23 = HEAP32[$ap>>2]|0;
     $100 = $arglist_current23;
     $101 = ((0) + 8|0);
     $expanded84 = $101;
     $expanded83 = (($expanded84) - 1)|0;
     $102 = (($100) + ($expanded83))|0;
     $103 = ((0) + 8|0);
     $expanded88 = $103;
     $expanded87 = (($expanded88) - 1)|0;
     $expanded86 = $expanded87 ^ -1;
     $104 = $102 & $expanded86;
     $105 = $104;
     $106 = +HEAPF64[$105>>3];
     $arglist_next24 = ((($105)) + 8|0);
     HEAP32[$ap>>2] = $arglist_next24;
     HEAPF64[$arg>>3] = $106;
     break L1;
     break;
    }
    case 18:  {
     $arglist_current26 = HEAP32[$ap>>2]|0;
     $107 = $arglist_current26;
     $108 = ((0) + 8|0);
     $expanded91 = $108;
     $expanded90 = (($expanded91) - 1)|0;
     $109 = (($107) + ($expanded90))|0;
     $110 = ((0) + 8|0);
     $expanded95 = $110;
     $expanded94 = (($expanded95) - 1)|0;
     $expanded93 = $expanded94 ^ -1;
     $111 = $109 & $expanded93;
     $112 = $111;
     $113 = +HEAPF64[$112>>3];
     $arglist_next27 = ((($112)) + 8|0);
     HEAP32[$ap>>2] = $arglist_next27;
     HEAPF64[$arg>>3] = $113;
     break L1;
     break;
    }
    default: {
     break L1;
    }
    }
   } while(0);
  }
 } while(0);
 return;
}
function _fmt_u($0,$1,$s) {
 $0 = $0|0;
 $1 = $1|0;
 $s = $s|0;
 var $$0$lcssa = 0, $$01$lcssa$off0 = 0, $$05 = 0, $$1$lcssa = 0, $$12 = 0, $$lcssa19 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $y$03 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($1>>>0)>(0);
 $3 = ($0>>>0)>(4294967295);
 $4 = ($1|0)==(0);
 $5 = $4 & $3;
 $6 = $2 | $5;
 if ($6) {
  $$05 = $s;$7 = $0;$8 = $1;
  while(1) {
   $9 = (___uremdi3(($7|0),($8|0),10,0)|0);
   $10 = tempRet0;
   $11 = $9 | 48;
   $12 = $11&255;
   $13 = ((($$05)) + -1|0);
   HEAP8[$13>>0] = $12;
   $14 = (___udivdi3(($7|0),($8|0),10,0)|0);
   $15 = tempRet0;
   $16 = ($8>>>0)>(9);
   $17 = ($7>>>0)>(4294967295);
   $18 = ($8|0)==(9);
   $19 = $18 & $17;
   $20 = $16 | $19;
   if ($20) {
    $$05 = $13;$7 = $14;$8 = $15;
   } else {
    $$lcssa19 = $13;$28 = $14;$29 = $15;
    break;
   }
  }
  $$0$lcssa = $$lcssa19;$$01$lcssa$off0 = $28;
 } else {
  $$0$lcssa = $s;$$01$lcssa$off0 = $0;
 }
 $21 = ($$01$lcssa$off0|0)==(0);
 if ($21) {
  $$1$lcssa = $$0$lcssa;
 } else {
  $$12 = $$0$lcssa;$y$03 = $$01$lcssa$off0;
  while(1) {
   $22 = (($y$03>>>0) % 10)&-1;
   $23 = $22 | 48;
   $24 = $23&255;
   $25 = ((($$12)) + -1|0);
   HEAP8[$25>>0] = $24;
   $26 = (($y$03>>>0) / 10)&-1;
   $27 = ($y$03>>>0)<(10);
   if ($27) {
    $$1$lcssa = $25;
    break;
   } else {
    $$12 = $25;$y$03 = $26;
   }
  }
 }
 return ($$1$lcssa|0);
}
function _strerror($e) {
 $e = $e|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$03 = 0, $i$03$lcssa = 0, $i$12 = 0, $s$0$lcssa = 0, $s$01 = 0, $s$1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $i$03 = 0;
 while(1) {
  $1 = (617491 + ($i$03)|0);
  $2 = HEAP8[$1>>0]|0;
  $3 = $2&255;
  $4 = ($3|0)==($e|0);
  if ($4) {
   $i$03$lcssa = $i$03;
   label = 2;
   break;
  }
  $5 = (($i$03) + 1)|0;
  $6 = ($5|0)==(87);
  if ($6) {
   $i$12 = 87;$s$01 = 617579;
   label = 5;
   break;
  } else {
   $i$03 = $5;
  }
 }
 if ((label|0) == 2) {
  $0 = ($i$03$lcssa|0)==(0);
  if ($0) {
   $s$0$lcssa = 617579;
  } else {
   $i$12 = $i$03$lcssa;$s$01 = 617579;
   label = 5;
  }
 }
 if ((label|0) == 5) {
  while(1) {
   label = 0;
   $s$1 = $s$01;
   while(1) {
    $7 = HEAP8[$s$1>>0]|0;
    $8 = ($7<<24>>24)==(0);
    $9 = ((($s$1)) + 1|0);
    if ($8) {
     $$lcssa = $9;
     break;
    } else {
     $s$1 = $9;
    }
   }
   $10 = (($i$12) + -1)|0;
   $11 = ($10|0)==(0);
   if ($11) {
    $s$0$lcssa = $$lcssa;
    break;
   } else {
    $i$12 = $10;$s$01 = $$lcssa;
    label = 5;
   }
  }
 }
 return ($s$0$lcssa|0);
}
function _pad($f,$c,$w,$l,$fl) {
 $f = $f|0;
 $c = $c|0;
 $w = $w|0;
 $l = $l|0;
 $fl = $fl|0;
 var $$0$lcssa6 = 0, $$02 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $or$cond = 0, $pad = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $pad = sp;
 $0 = $fl & 73728;
 $1 = ($0|0)==(0);
 $2 = ($w|0)>($l|0);
 $or$cond = $2 & $1;
 do {
  if ($or$cond) {
   $3 = (($w) - ($l))|0;
   $4 = ($3>>>0)>(256);
   $5 = $4 ? 256 : $3;
   _memset(($pad|0),($c|0),($5|0))|0;
   $6 = ($3>>>0)>(255);
   $7 = HEAP32[$f>>2]|0;
   $8 = $7 & 32;
   $9 = ($8|0)==(0);
   if ($6) {
    $10 = (($w) - ($l))|0;
    $$02 = $3;$17 = $7;$18 = $9;
    while(1) {
     if ($18) {
      (___fwritex($pad,256,$f)|0);
      $$pre = HEAP32[$f>>2]|0;
      $14 = $$pre;
     } else {
      $14 = $17;
     }
     $11 = (($$02) + -256)|0;
     $12 = ($11>>>0)>(255);
     $13 = $14 & 32;
     $15 = ($13|0)==(0);
     if ($12) {
      $$02 = $11;$17 = $14;$18 = $15;
     } else {
      break;
     }
    }
    $16 = $10 & 255;
    if ($15) {
     $$0$lcssa6 = $16;
    } else {
     break;
    }
   } else {
    if ($9) {
     $$0$lcssa6 = $3;
    } else {
     break;
    }
   }
   (___fwritex($pad,$$0$lcssa6,$f)|0);
  }
 } while(0);
 STACKTOP = sp;return;
}
function _wctomb($s,$wc) {
 $s = $s|0;
 $wc = $wc|0;
 var $$0 = 0, $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($s|0)==(0|0);
 if ($0) {
  $$0 = 0;
 } else {
  $1 = (_wcrtomb($s,$wc,0)|0);
  $$0 = $1;
 }
 return ($$0|0);
}
function _wcrtomb($s,$wc,$st) {
 $s = $s|0;
 $wc = $wc|0;
 $st = $st|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($s|0)==(0|0);
 do {
  if ($0) {
   $$0 = 1;
  } else {
   $1 = ($wc>>>0)<(128);
   if ($1) {
    $2 = $wc&255;
    HEAP8[$s>>0] = $2;
    $$0 = 1;
    break;
   }
   $3 = ($wc>>>0)<(2048);
   if ($3) {
    $4 = $wc >>> 6;
    $5 = $4 | 192;
    $6 = $5&255;
    $7 = ((($s)) + 1|0);
    HEAP8[$s>>0] = $6;
    $8 = $wc & 63;
    $9 = $8 | 128;
    $10 = $9&255;
    HEAP8[$7>>0] = $10;
    $$0 = 2;
    break;
   }
   $11 = ($wc>>>0)<(55296);
   $12 = $wc & -8192;
   $13 = ($12|0)==(57344);
   $or$cond = $11 | $13;
   if ($or$cond) {
    $14 = $wc >>> 12;
    $15 = $14 | 224;
    $16 = $15&255;
    $17 = ((($s)) + 1|0);
    HEAP8[$s>>0] = $16;
    $18 = $wc >>> 6;
    $19 = $18 & 63;
    $20 = $19 | 128;
    $21 = $20&255;
    $22 = ((($s)) + 2|0);
    HEAP8[$17>>0] = $21;
    $23 = $wc & 63;
    $24 = $23 | 128;
    $25 = $24&255;
    HEAP8[$22>>0] = $25;
    $$0 = 3;
    break;
   }
   $26 = (($wc) + -65536)|0;
   $27 = ($26>>>0)<(1048576);
   if ($27) {
    $28 = $wc >>> 18;
    $29 = $28 | 240;
    $30 = $29&255;
    $31 = ((($s)) + 1|0);
    HEAP8[$s>>0] = $30;
    $32 = $wc >>> 12;
    $33 = $32 & 63;
    $34 = $33 | 128;
    $35 = $34&255;
    $36 = ((($s)) + 2|0);
    HEAP8[$31>>0] = $35;
    $37 = $wc >>> 6;
    $38 = $37 & 63;
    $39 = $38 | 128;
    $40 = $39&255;
    $41 = ((($s)) + 3|0);
    HEAP8[$36>>0] = $40;
    $42 = $wc & 63;
    $43 = $42 | 128;
    $44 = $43&255;
    HEAP8[$41>>0] = $44;
    $$0 = 4;
    break;
   } else {
    $45 = (___errno_location()|0);
    HEAP32[$45>>2] = 84;
    $$0 = -1;
    break;
   }
  }
 } while(0);
 return ($$0|0);
}
function _frexpl($x,$e) {
 $x = +$x;
 $e = $e|0;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_frexp($x,$e));
 return (+$0);
}
function _frexp($x,$e) {
 $x = +$x;
 $e = $e|0;
 var $$0 = 0.0, $$01 = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0, $9 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAPF64[tempDoublePtr>>3] = $x;$0 = HEAP32[tempDoublePtr>>2]|0;
 $1 = HEAP32[tempDoublePtr+4>>2]|0;
 $2 = (_bitshift64Lshr(($0|0),($1|0),52)|0);
 $3 = tempRet0;
 $4 = $2 & 2047;
 switch ($4|0) {
 case 0:  {
  $5 = $x != 0.0;
  if ($5) {
   $6 = $x * 1.8446744073709552E+19;
   $7 = (+_frexp($6,$e));
   $8 = HEAP32[$e>>2]|0;
   $9 = (($8) + -64)|0;
   $$01 = $7;$storemerge = $9;
  } else {
   $$01 = $x;$storemerge = 0;
  }
  HEAP32[$e>>2] = $storemerge;
  $$0 = $$01;
  break;
 }
 case 2047:  {
  $$0 = $x;
  break;
 }
 default: {
  $10 = (($4) + -1022)|0;
  HEAP32[$e>>2] = $10;
  $11 = $1 & -2146435073;
  $12 = $11 | 1071644672;
  HEAP32[tempDoublePtr>>2] = $0;HEAP32[tempDoublePtr+4>>2] = $12;$13 = +HEAPF64[tempDoublePtr>>3];
  $$0 = $13;
 }
 }
 return (+$$0);
}
function ___lockfile($f) {
 $f = $f|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _strcmp($l,$r) {
 $l = $l|0;
 $r = $r|0;
 var $$014 = 0, $$05 = 0, $$lcssa = 0, $$lcssa2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond3 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$l>>0]|0;
 $1 = HEAP8[$r>>0]|0;
 $2 = ($0<<24>>24)!=($1<<24>>24);
 $3 = ($0<<24>>24)==(0);
 $or$cond3 = $3 | $2;
 if ($or$cond3) {
  $$lcssa = $0;$$lcssa2 = $1;
 } else {
  $$014 = $l;$$05 = $r;
  while(1) {
   $4 = ((($$014)) + 1|0);
   $5 = ((($$05)) + 1|0);
   $6 = HEAP8[$4>>0]|0;
   $7 = HEAP8[$5>>0]|0;
   $8 = ($6<<24>>24)!=($7<<24>>24);
   $9 = ($6<<24>>24)==(0);
   $or$cond = $9 | $8;
   if ($or$cond) {
    $$lcssa = $6;$$lcssa2 = $7;
    break;
   } else {
    $$014 = $4;$$05 = $5;
   }
  }
 }
 $10 = $$lcssa&255;
 $11 = $$lcssa2&255;
 $12 = (($10) - ($11))|0;
 return ($12|0);
}
function ___ctype_b_loc() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (150640|0);
}
function ___ctype_tolower_loc() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (150644|0);
}
function ___ctype_toupper_loc() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (152184|0);
}
function _isalnum($c) {
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, $isdigit = 0, $isdigittmp = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_isalpha($c)|0);
 $isdigittmp = (($c) + -48)|0;
 $isdigit = ($isdigittmp>>>0)<(10);
 $not$ = ($0|0)!=(0);
 $1 = $isdigit | $not$;
 $2 = $1&1;
 return ($2|0);
}
function _isalpha($c) {
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $c | 32;
 $1 = (($0) + -97)|0;
 $2 = ($1>>>0)<(26);
 $3 = $2&1;
 return ($3|0);
}
function _isblank($c) {
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($c|0)==(32);
 $1 = ($c|0)==(9);
 $2 = $0 | $1;
 $3 = $2&1;
 return ($3|0);
}
function _iscntrl($c) {
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($c>>>0)<(32);
 $1 = ($c|0)==(127);
 $2 = $0 | $1;
 $3 = $2&1;
 return ($3|0);
}
function _isgraph($c) {
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($c) + -33)|0;
 $1 = ($0>>>0)<(94);
 $2 = $1&1;
 return ($2|0);
}
function _ispunct($c) {
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_isgraph($c)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  $5 = 0;
 } else {
  $2 = (_isalnum($c)|0);
  $3 = ($2|0)==(0);
  $5 = $3;
 }
 $4 = $5&1;
 return ($4|0);
}
function _wcslen($s) {
 $s = $s|0;
 var $$0 = 0, $$0$lcssa = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $s;
 while(1) {
  $0 = HEAP32[$$0>>2]|0;
  $1 = ($0|0)==(0);
  $2 = ((($$0)) + 4|0);
  if ($1) {
   $$0$lcssa = $$0;
   break;
  } else {
   $$0 = $2;
  }
 }
 $3 = $$0$lcssa;
 $4 = $s;
 $5 = (($3) - ($4))|0;
 $6 = $5 >> 2;
 return ($6|0);
}
function _isxdigit($c) {
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $isdigit = 0, $isdigittmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $isdigittmp = (($c) + -48)|0;
 $isdigit = ($isdigittmp>>>0)<(10);
 if ($isdigit) {
  $4 = 1;
 } else {
  $0 = $c | 32;
  $1 = (($0) + -97)|0;
  $2 = ($1>>>0)<(6);
  $4 = $2;
 }
 $3 = $4&1;
 return ($3|0);
}
function _strcoll($l,$r) {
 $l = $l|0;
 $r = $r|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strcmp($l,$r)|0);
 return ($0|0);
}
function _closedir($dir) {
 $dir = $dir|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$dir>>2]|0;
 $1 = (_close($0)|0);
 _free($dir);
 return ($1|0);
}
function _close($fd) {
 $fd = $fd|0;
 var $$ = 0, $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 HEAP32[$vararg_buffer>>2] = $fd;
 $0 = (___syscall6(6,($vararg_buffer|0))|0);
 $1 = ($0|0)==(-4);
 $$ = $1 ? -115 : $0;
 $2 = (___syscall_ret($$)|0);
 STACKTOP = sp;return ($2|0);
}
function _fstat($fd,$st) {
 $fd = $fd|0;
 $st = $st|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $buf = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_buffer6 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, $vararg_ptr9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer6 = sp + 16|0;
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $buf = sp + 24|0;
 HEAP32[$vararg_buffer>>2] = $fd;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $st;
 $0 = (___syscall197(197,($vararg_buffer|0))|0);
 $1 = ($0|0)==(-9);
 if ($1) {
  HEAP32[$vararg_buffer2>>2] = $fd;
  $vararg_ptr5 = ((($vararg_buffer2)) + 4|0);
  HEAP32[$vararg_ptr5>>2] = 1;
  $2 = (___syscall221(221,($vararg_buffer2|0))|0);
  $3 = ($2|0)<(0);
  if ($3) {
   label = 3;
  } else {
   ___procfdname($buf,$fd);
   HEAP32[$vararg_buffer6>>2] = $buf;
   $vararg_ptr9 = ((($vararg_buffer6)) + 4|0);
   HEAP32[$vararg_ptr9>>2] = $st;
   $5 = (___syscall195(195,($vararg_buffer6|0))|0);
   $6 = (___syscall_ret($5)|0);
   $$0 = $6;
  }
 } else {
  label = 3;
 }
 if ((label|0) == 3) {
  $4 = (___syscall_ret($0)|0);
  $$0 = $4;
 }
 STACKTOP = sp;return ($$0|0);
}
function ___procfdname($buf,$fd) {
 $buf = $buf|0;
 $fd = $fd|0;
 var $$01 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$13 = 0, $i$22 = 0, $j$04 = 0, $scevgep = 0, dest = 0;
 var label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 dest=$buf; src=619421; stop=dest+15|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
 $0 = ($fd|0)==(0);
 if ($0) {
  $scevgep = ((($buf)) + 14|0);
  HEAP8[$scevgep>>0] = 48;
  $1 = ((($buf)) + 15|0);
  HEAP8[$1>>0] = 0;
 } else {
  $i$13 = 14;$j$04 = $fd;
  while(1) {
   $2 = (($j$04>>>0) / 10)&-1;
   $3 = (($i$13) + 1)|0;
   $4 = ($j$04>>>0)<(10);
   if ($4) {
    $$lcssa = $3;
    break;
   } else {
    $i$13 = $3;$j$04 = $2;
   }
  }
  $5 = (($buf) + ($$lcssa)|0);
  HEAP8[$5>>0] = 0;
  $$01 = $fd;$i$22 = $$lcssa;
  while(1) {
   $6 = (($$01>>>0) % 10)&-1;
   $7 = $6 | 48;
   $8 = $7&255;
   $9 = (($i$22) + -1)|0;
   $10 = (($buf) + ($9)|0);
   HEAP8[$10>>0] = $8;
   $11 = (($$01>>>0) / 10)&-1;
   $12 = ($$01>>>0)<(10);
   if ($12) {
    break;
   } else {
    $$01 = $11;$i$22 = $9;
   }
  }
 }
 return;
}
function _fcntl($fd,$cmd,$varargs) {
 $fd = $fd|0;
 $cmd = $cmd|0;
 $varargs = $varargs|0;
 var $$ = 0, $$0 = 0, $$2 = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $ap = 0, $arglist_current = 0, $arglist_next = 0, $ex = 0, $expanded = 0, $expanded2 = 0, $expanded4 = 0, $expanded5 = 0, $expanded6 = 0, $switch = 0, $vararg_buffer = 0, $vararg_buffer13 = 0, $vararg_buffer18 = 0, $vararg_buffer23 = 0, $vararg_buffer28 = 0, $vararg_buffer3 = 0, $vararg_buffer31 = 0, $vararg_buffer36 = 0;
 var $vararg_buffer41 = 0, $vararg_buffer46 = 0, $vararg_buffer8 = 0, $vararg_ptr1 = 0, $vararg_ptr11 = 0, $vararg_ptr12 = 0, $vararg_ptr16 = 0, $vararg_ptr17 = 0, $vararg_ptr2 = 0, $vararg_ptr21 = 0, $vararg_ptr22 = 0, $vararg_ptr26 = 0, $vararg_ptr27 = 0, $vararg_ptr34 = 0, $vararg_ptr35 = 0, $vararg_ptr39 = 0, $vararg_ptr40 = 0, $vararg_ptr44 = 0, $vararg_ptr45 = 0, $vararg_ptr49 = 0;
 var $vararg_ptr50 = 0, $vararg_ptr6 = 0, $vararg_ptr7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer46 = sp + 152|0;
 $vararg_buffer41 = sp + 136|0;
 $vararg_buffer36 = sp + 120|0;
 $vararg_buffer31 = sp + 104|0;
 $vararg_buffer28 = sp + 96|0;
 $vararg_buffer23 = sp + 80|0;
 $vararg_buffer18 = sp + 64|0;
 $vararg_buffer13 = sp + 48|0;
 $vararg_buffer8 = sp + 32|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $ap = sp + 176|0;
 $ex = sp + 168|0;
 HEAP32[$ap>>2] = $varargs;
 $arglist_current = HEAP32[$ap>>2]|0;
 $0 = $arglist_current;
 $1 = ((0) + 4|0);
 $expanded2 = $1;
 $expanded = (($expanded2) - 1)|0;
 $2 = (($0) + ($expanded))|0;
 $3 = ((0) + 4|0);
 $expanded6 = $3;
 $expanded5 = (($expanded6) - 1)|0;
 $expanded4 = $expanded5 ^ -1;
 $4 = $2 & $expanded4;
 $5 = $4;
 $6 = HEAP32[$5>>2]|0;
 $arglist_next = ((($5)) + 4|0);
 HEAP32[$ap>>2] = $arglist_next;
 $7 = ($cmd|0)==(4);
 $8 = $6 | 32768;
 $$ = $7 ? $8 : $6;
 L1: do {
  switch ($cmd|0) {
  case 14:  {
   $9 = $$;
   HEAP32[$vararg_buffer>>2] = $fd;
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
   HEAP32[$vararg_ptr1>>2] = 14;
   $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
   HEAP32[$vararg_ptr2>>2] = $9;
   $10 = (___syscall221(221,($vararg_buffer|0))|0);
   $11 = (___syscall_ret($10)|0);
   $$2 = $11;
   break;
  }
  case 9:  {
   HEAP32[$vararg_buffer3>>2] = $fd;
   $vararg_ptr6 = ((($vararg_buffer3)) + 4|0);
   HEAP32[$vararg_ptr6>>2] = 16;
   $vararg_ptr7 = ((($vararg_buffer3)) + 8|0);
   HEAP32[$vararg_ptr7>>2] = $ex;
   $12 = (___syscall221(221,($vararg_buffer3|0))|0);
   switch ($12|0) {
   case -22:  {
    $13 = $$;
    HEAP32[$vararg_buffer8>>2] = $fd;
    $vararg_ptr11 = ((($vararg_buffer8)) + 4|0);
    HEAP32[$vararg_ptr11>>2] = 9;
    $vararg_ptr12 = ((($vararg_buffer8)) + 8|0);
    HEAP32[$vararg_ptr12>>2] = $13;
    $14 = (___syscall221(221,($vararg_buffer8|0))|0);
    $$0 = $14;
    break;
   }
   case 0:  {
    $16 = HEAP32[$ex>>2]|0;
    $17 = ($16|0)==(2);
    $18 = ((($ex)) + 4|0);
    $19 = HEAP32[$18>>2]|0;
    $20 = (0 - ($19))|0;
    $21 = $17 ? $20 : $19;
    $$0 = $21;
    break;
   }
   default: {
    $15 = (___syscall_ret($12)|0);
    $$0 = $15;
   }
   }
   $$2 = $$0;
   break;
  }
  case 1030:  {
   HEAP32[$vararg_buffer13>>2] = $fd;
   $vararg_ptr16 = ((($vararg_buffer13)) + 4|0);
   HEAP32[$vararg_ptr16>>2] = 1030;
   $vararg_ptr17 = ((($vararg_buffer13)) + 8|0);
   HEAP32[$vararg_ptr17>>2] = $$;
   $22 = (___syscall221(221,($vararg_buffer13|0))|0);
   $23 = ($22|0)==(-22);
   if (!($23)) {
    $24 = ($22|0)>(-1);
    if ($24) {
     HEAP32[$vararg_buffer18>>2] = $22;
     $vararg_ptr21 = ((($vararg_buffer18)) + 4|0);
     HEAP32[$vararg_ptr21>>2] = 2;
     $vararg_ptr22 = ((($vararg_buffer18)) + 8|0);
     HEAP32[$vararg_ptr22>>2] = 1;
     (___syscall221(221,($vararg_buffer18|0))|0);
    }
    $25 = (___syscall_ret($22)|0);
    $$2 = $25;
    break L1;
   }
   HEAP32[$vararg_buffer23>>2] = $fd;
   $vararg_ptr26 = ((($vararg_buffer23)) + 4|0);
   HEAP32[$vararg_ptr26>>2] = 1030;
   $vararg_ptr27 = ((($vararg_buffer23)) + 8|0);
   HEAP32[$vararg_ptr27>>2] = 0;
   $26 = (___syscall221(221,($vararg_buffer23|0))|0);
   $27 = ($26|0)==(-22);
   if ($27) {
    HEAP32[$vararg_buffer31>>2] = $fd;
    $vararg_ptr34 = ((($vararg_buffer31)) + 4|0);
    HEAP32[$vararg_ptr34>>2] = 0;
    $vararg_ptr35 = ((($vararg_buffer31)) + 8|0);
    HEAP32[$vararg_ptr35>>2] = $$;
    $30 = (___syscall221(221,($vararg_buffer31|0))|0);
    $31 = ($30|0)>(-1);
    if ($31) {
     HEAP32[$vararg_buffer36>>2] = $30;
     $vararg_ptr39 = ((($vararg_buffer36)) + 4|0);
     HEAP32[$vararg_ptr39>>2] = 2;
     $vararg_ptr40 = ((($vararg_buffer36)) + 8|0);
     HEAP32[$vararg_ptr40>>2] = 1;
     (___syscall221(221,($vararg_buffer36|0))|0);
    }
    $32 = (___syscall_ret($30)|0);
    $$2 = $32;
    break L1;
   } else {
    $28 = ($26|0)>(-1);
    if ($28) {
     HEAP32[$vararg_buffer28>>2] = $26;
     (___syscall6(6,($vararg_buffer28|0))|0);
    }
    $29 = (___syscall_ret(-22)|0);
    $$2 = $29;
    break L1;
   }
   break;
  }
  default: {
   $$off = (($cmd) + -12)|0;
   $switch = ($$off>>>0)<(5);
   if ($switch) {
    $33 = $$;
    HEAP32[$vararg_buffer41>>2] = $fd;
    $vararg_ptr44 = ((($vararg_buffer41)) + 4|0);
    HEAP32[$vararg_ptr44>>2] = $cmd;
    $vararg_ptr45 = ((($vararg_buffer41)) + 8|0);
    HEAP32[$vararg_ptr45>>2] = $33;
    $34 = (___syscall221(221,($vararg_buffer41|0))|0);
    $35 = (___syscall_ret($34)|0);
    $$2 = $35;
    break L1;
   } else {
    HEAP32[$vararg_buffer46>>2] = $fd;
    $vararg_ptr49 = ((($vararg_buffer46)) + 4|0);
    HEAP32[$vararg_ptr49>>2] = $cmd;
    $vararg_ptr50 = ((($vararg_buffer46)) + 8|0);
    HEAP32[$vararg_ptr50>>2] = $$;
    $36 = (___syscall221(221,($vararg_buffer46|0))|0);
    $37 = (___syscall_ret($36)|0);
    $$2 = $37;
    break L1;
   }
  }
  }
 } while(0);
 STACKTOP = sp;return ($$2|0);
}
function _opendir($name) {
 $name = $name|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = (_open($name,589824,$vararg_buffer)|0);
 $1 = ($0|0)<(0);
 do {
  if ($1) {
   $$0 = 0;
  } else {
   $2 = (_calloc(1,2072)|0);
   $3 = ($2|0)==(0|0);
   if ($3) {
    HEAP32[$vararg_buffer1>>2] = $0;
    (___syscall6(6,($vararg_buffer1|0))|0);
    $$0 = 0;
    break;
   } else {
    HEAP32[$2>>2] = $0;
    $$0 = $2;
    break;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _open($filename,$flags,$varargs) {
 $filename = $filename|0;
 $flags = $flags|0;
 $varargs = $varargs|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ap = 0, $arglist_current = 0, $arglist_next = 0, $expanded = 0, $expanded2 = 0, $expanded4 = 0, $expanded5 = 0, $expanded6 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0;
 var $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $ap = sp + 16|0;
 HEAP32[$ap>>2] = $varargs;
 $arglist_current = HEAP32[$ap>>2]|0;
 $0 = $arglist_current;
 $1 = ((0) + 4|0);
 $expanded2 = $1;
 $expanded = (($expanded2) - 1)|0;
 $2 = (($0) + ($expanded))|0;
 $3 = ((0) + 4|0);
 $expanded6 = $3;
 $expanded5 = (($expanded6) - 1)|0;
 $expanded4 = $expanded5 ^ -1;
 $4 = $2 & $expanded4;
 $5 = $4;
 $6 = HEAP32[$5>>2]|0;
 $arglist_next = ((($5)) + 4|0);
 HEAP32[$ap>>2] = $arglist_next;
 $7 = $flags | 32768;
 HEAP32[$vararg_buffer>>2] = $filename;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $7;
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
 HEAP32[$vararg_ptr2>>2] = $6;
 $8 = (___syscall5(5,($vararg_buffer|0))|0);
 $9 = (___syscall_ret($8)|0);
 STACKTOP = sp;return ($9|0);
}
function _readdir($dir) {
 $dir = $dir|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = ((($dir)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($dir)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)<($3|0);
 do {
  if ($4) {
   $14 = $1;
   label = 6;
  } else {
   $5 = HEAP32[$dir>>2]|0;
   $6 = ((($dir)) + 24|0);
   HEAP32[$vararg_buffer>>2] = $5;
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
   HEAP32[$vararg_ptr1>>2] = $6;
   $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
   HEAP32[$vararg_ptr2>>2] = 2048;
   $7 = (___syscall220(220,($vararg_buffer|0))|0);
   $8 = ($7|0)<(1);
   if (!($8)) {
    HEAP32[$2>>2] = $7;
    HEAP32[$0>>2] = 0;
    $14 = 0;
    label = 6;
    break;
   }
   $9 = ($7|0)<(0);
   $10 = ($7|0)!=(-2);
   $or$cond = $9 & $10;
   if ($or$cond) {
    $11 = (0 - ($7))|0;
    $12 = (___errno_location()|0);
    HEAP32[$12>>2] = $11;
    $$1 = 0;
   } else {
    $$1 = 0;
   }
  }
 } while(0);
 if ((label|0) == 6) {
  $13 = (((($dir)) + 24|0) + ($14)|0);
  $15 = ((($13)) + 8|0);
  $16 = HEAP16[$15>>1]|0;
  $17 = $16&65535;
  $18 = (($17) + ($14))|0;
  HEAP32[$0>>2] = $18;
  $19 = ((($13)) + 4|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = ((($dir)) + 4|0);
  HEAP32[$21>>2] = $20;
  $$1 = $13;
 }
 STACKTOP = sp;return ($$1|0);
}
function _lseek($fd,$offset,$whence) {
 $fd = $fd|0;
 $offset = $offset|0;
 $whence = $whence|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $result = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0, $vararg_ptr4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $result = sp + 20|0;
 HEAP32[$vararg_buffer>>2] = $fd;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = 0;
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
 HEAP32[$vararg_ptr2>>2] = $offset;
 $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
 HEAP32[$vararg_ptr3>>2] = $result;
 $vararg_ptr4 = ((($vararg_buffer)) + 16|0);
 HEAP32[$vararg_ptr4>>2] = $whence;
 $0 = (___syscall140(140,($vararg_buffer|0))|0);
 $1 = (___syscall_ret($0)|0);
 $2 = ($1|0)!=(0);
 $3 = HEAP32[$result>>2]|0;
 $4 = $2 ? -1 : $3;
 STACKTOP = sp;return ($4|0);
}
function _qsort($base,$nel,$width,$cmp) {
 $base = $base|0;
 $nel = $nel|0;
 $width = $width|0;
 $cmp = $cmp|0;
 var $$0$i = 0, $$0$i29 = 0, $$01$i5$i$ph = 0, $$02$i$i = 0, $$02$i3$i = 0, $$phi$trans$insert$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i10 = 0, $$pre$i4 = 0, $$pre$i7 = 0, $$pre1$i = 0, $$pre1$i11 = 0, $$pre1$i5 = 0, $$pre1$i8 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $9$phi = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $head$0$lcssa = 0, $head$033 = 0, $head$1$be = 0, $head$149 = 0, $i$0 = 0, $lp = 0, $nTrailingZeros$03$i$i = 0;
 var $nTrailingZeros$03$i2$i = 0, $or$cond = 0, $or$cond38 = 0, $or$cond3848 = 0, $or$cond47 = 0, $p = 0, $pshift$0$lcssa = 0, $pshift$034 = 0, $pshift$1 = 0, $pshift$2$be = 0, $pshift$250 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 208|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $lp = sp + 8|0;
 $p = sp;
 $0 = Math_imul($width, $nel)|0;
 $1 = $p;
 $2 = $1;
 HEAP32[$2>>2] = 1;
 $3 = (($1) + 4)|0;
 $4 = $3;
 HEAP32[$4>>2] = 0;
 $5 = ($0|0)==(0);
 if (!($5)) {
  $6 = (0 - ($width))|0;
  $7 = ((($lp)) + 4|0);
  HEAP32[$7>>2] = $width;
  HEAP32[$lp>>2] = $width;
  $11 = $width;$9 = $width;$i$0 = 2;
  while(1) {
   $8 = (($9) + ($width))|0;
   $10 = (($8) + ($11))|0;
   $12 = (($lp) + ($i$0<<2)|0);
   HEAP32[$12>>2] = $10;
   $13 = ($10>>>0)<($0>>>0);
   $14 = (($i$0) + 1)|0;
   if ($13) {
    $9$phi = $11;$11 = $10;$i$0 = $14;$9 = $9$phi;
   } else {
    break;
   }
  }
  $15 = (($base) + ($0)|0);
  $16 = (($15) + ($6)|0);
  $17 = ($16>>>0)>($base>>>0);
  $$phi$trans$insert$i = ((($p)) + 4|0);
  if ($17) {
   $18 = $16;
   $20 = 1;$head$033 = $base;$pshift$034 = 1;
   while(1) {
    $19 = $20 & 3;
    $21 = ($19|0)==(3);
    do {
     if ($21) {
      _sift($head$033,$width,$cmp,$pshift$034,$lp);
      $$pre$i = HEAP32[$p>>2]|0;
      $$pre1$i = HEAP32[$$phi$trans$insert$i>>2]|0;
      $22 = $$pre$i >>> 2;
      $23 = $$pre1$i << 30;
      $24 = $23 | $22;
      HEAP32[$p>>2] = $24;
      $25 = $$pre1$i >>> 2;
      HEAP32[$$phi$trans$insert$i>>2] = $25;
      $26 = (($pshift$034) + 2)|0;
      $49 = $24;$pshift$1 = $26;
     } else {
      $27 = (($pshift$034) + -1)|0;
      $28 = (($lp) + ($27<<2)|0);
      $29 = HEAP32[$28>>2]|0;
      $30 = $head$033;
      $31 = (($18) - ($30))|0;
      $32 = ($29>>>0)<($31>>>0);
      if ($32) {
       _sift($head$033,$width,$cmp,$pshift$034,$lp);
      } else {
       _trinkle($head$033,$width,$cmp,$p,$pshift$034,0,$lp);
      }
      $33 = ($pshift$034|0)==(1);
      if ($33) {
       $$pre$i4 = HEAP32[$$phi$trans$insert$i>>2]|0;
       $$pre1$i5 = HEAP32[$p>>2]|0;
       $34 = $$pre$i4 << 1;
       $35 = $$pre1$i5 >>> 31;
       $36 = $35 | $34;
       HEAP32[$$phi$trans$insert$i>>2] = $36;
       $37 = $$pre1$i5 << 1;
       HEAP32[$p>>2] = $37;
       $49 = $37;$pshift$1 = 0;
       break;
      }
      $38 = ($27>>>0)>(31);
      if ($38) {
       $39 = (($pshift$034) + -33)|0;
       $40 = HEAP32[$p>>2]|0;
       HEAP32[$$phi$trans$insert$i>>2] = $40;
       HEAP32[$p>>2] = 0;
       $$0$i = $39;$42 = $40;$45 = 0;
      } else {
       $$pre$i10 = HEAP32[$$phi$trans$insert$i>>2]|0;
       $$pre1$i11 = HEAP32[$p>>2]|0;
       $$0$i = $27;$42 = $$pre$i10;$45 = $$pre1$i11;
      }
      $41 = $42 << $$0$i;
      $43 = (32 - ($$0$i))|0;
      $44 = $45 >>> $43;
      $46 = $44 | $41;
      HEAP32[$$phi$trans$insert$i>>2] = $46;
      $47 = $45 << $$0$i;
      HEAP32[$p>>2] = $47;
      $49 = $47;$pshift$1 = 1;
     }
    } while(0);
    $48 = $49 | 1;
    HEAP32[$p>>2] = $48;
    $50 = (($head$033) + ($width)|0);
    $51 = ($50>>>0)<($16>>>0);
    if ($51) {
     $20 = $48;$head$033 = $50;$pshift$034 = $pshift$1;
    } else {
     $head$0$lcssa = $50;$pshift$0$lcssa = $pshift$1;
     break;
    }
   }
  } else {
   $head$0$lcssa = $base;$pshift$0$lcssa = 1;
  }
  _trinkle($head$0$lcssa,$width,$cmp,$p,$pshift$0$lcssa,0,$lp);
  $52 = ((($p)) + 4|0);
  $53 = ($pshift$0$lcssa|0)==(1);
  $54 = HEAP32[$p>>2]|0;
  $55 = ($54|0)==(1);
  $or$cond47 = $53 & $55;
  $56 = HEAP32[$52>>2]|0;
  $57 = ($56|0)==(0);
  $or$cond3848 = $or$cond47 & $57;
  if (!($or$cond3848)) {
   $116 = $57;$60 = $54;$69 = $56;$head$149 = $head$0$lcssa;$pshift$250 = $pshift$0$lcssa;
   while(1) {
    $58 = ($pshift$250|0)<(2);
    if ($58) {
     $59 = (($60) + -1)|0;
     $61 = ($59|0)==(0);
     do {
      if ($61) {
       $79 = 32;
       label = 28;
      } else {
       $62 = $59 & 1;
       $63 = ($62|0)==(0);
       if ($63) {
        $$02$i$i = $59;$nTrailingZeros$03$i$i = 0;
        while(1) {
         $64 = (($nTrailingZeros$03$i$i) + 1)|0;
         $65 = $$02$i$i >>> 1;
         $66 = $65 & 1;
         $67 = ($66|0)==(0);
         if ($67) {
          $$02$i$i = $65;$nTrailingZeros$03$i$i = $64;
         } else {
          $77 = $64;
          break;
         }
        }
       } else {
        if ($116) {
         $$01$i5$i$ph = 32;
        } else {
         $68 = $69 & 1;
         $70 = ($68|0)==(0);
         if ($70) {
          $$02$i3$i = $69;$nTrailingZeros$03$i2$i = 0;
         } else {
          $$0$i29 = 0;$81 = $60;$84 = $69;$88 = 0;
          break;
         }
         while(1) {
          $71 = (($nTrailingZeros$03$i2$i) + 1)|0;
          $72 = $$02$i3$i >>> 1;
          $73 = $72 & 1;
          $74 = ($73|0)==(0);
          if ($74) {
           $$02$i3$i = $72;$nTrailingZeros$03$i2$i = $71;
          } else {
           $$01$i5$i$ph = $71;
           break;
          }
         }
        }
        $75 = (($$01$i5$i$ph) + 32)|0;
        $77 = $75;
       }
       $76 = ($77>>>0)>(31);
       if ($76) {
        $79 = $77;
        label = 28;
       } else {
        $$0$i29 = $77;$81 = $60;$84 = $69;$88 = $77;
       }
      }
     } while(0);
     if ((label|0) == 28) {
      label = 0;
      $78 = (($79) + -32)|0;
      HEAP32[$p>>2] = $69;
      HEAP32[$$phi$trans$insert$i>>2] = 0;
      $$0$i29 = $78;$81 = $69;$84 = 0;$88 = $79;
     }
     $80 = $81 >>> $$0$i29;
     $82 = (32 - ($$0$i29))|0;
     $83 = $84 << $82;
     $85 = $83 | $80;
     HEAP32[$p>>2] = $85;
     $86 = $84 >>> $$0$i29;
     HEAP32[$$phi$trans$insert$i>>2] = $86;
     $87 = (($88) + ($pshift$250))|0;
     $$pre = (($head$149) + ($6)|0);
     $head$1$be = $$pre;$pshift$2$be = $87;
    } else {
     $89 = $69 << 2;
     $90 = $60 >>> 30;
     $91 = $90 | $89;
     $92 = (($pshift$250) + -2)|0;
     $93 = $60 << 1;
     $94 = $93 & 2147483646;
     $95 = $90 << 31;
     $96 = $94 | $95;
     $97 = $96 ^ 3;
     HEAP32[$p>>2] = $97;
     $98 = $91 >>> 1;
     HEAP32[$$phi$trans$insert$i>>2] = $98;
     $99 = (($lp) + ($92<<2)|0);
     $100 = HEAP32[$99>>2]|0;
     $101 = (0 - ($100))|0;
     $102 = (($head$149) + ($101)|0);
     $103 = (($102) + ($6)|0);
     $104 = (($pshift$250) + -1)|0;
     _trinkle($103,$width,$cmp,$p,$104,1,$lp);
     $$pre$i7 = HEAP32[$$phi$trans$insert$i>>2]|0;
     $$pre1$i8 = HEAP32[$p>>2]|0;
     $105 = $$pre$i7 << 1;
     $106 = $$pre1$i8 >>> 31;
     $107 = $106 | $105;
     HEAP32[$$phi$trans$insert$i>>2] = $107;
     $108 = $$pre1$i8 << 1;
     $109 = $108 | 1;
     HEAP32[$p>>2] = $109;
     $110 = (($head$149) + ($6)|0);
     _trinkle($110,$width,$cmp,$p,$92,1,$lp);
     $head$1$be = $110;$pshift$2$be = $92;
    }
    $111 = ($pshift$2$be|0)==(1);
    $112 = HEAP32[$p>>2]|0;
    $113 = ($112|0)==(1);
    $or$cond = $111 & $113;
    $114 = HEAP32[$52>>2]|0;
    $115 = ($114|0)==(0);
    $or$cond38 = $or$cond & $115;
    if ($or$cond38) {
     break;
    } else {
     $116 = $115;$60 = $112;$69 = $114;$head$149 = $head$1$be;$pshift$250 = $pshift$2$be;
    }
   }
  }
 }
 STACKTOP = sp;return;
}
function _sift($head,$width,$cmp,$pshift,$lp) {
 $head = $head|0;
 $width = $width|0;
 $cmp = $cmp|0;
 $pshift = $pshift|0;
 $lp = $lp|0;
 var $$0$be = 0, $$01$be = 0, $$012 = 0, $$03 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ar = 0, $i$0$lcssa = 0, $i$04 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ar = sp;
 HEAP32[$ar>>2] = $head;
 $0 = ($pshift|0)>(1);
 L1: do {
  if ($0) {
   $1 = (0 - ($width))|0;
   $$012 = $pshift;$$03 = $head;$8 = $head;$i$04 = 1;
   while(1) {
    $2 = (($$03) + ($1)|0);
    $3 = (($$012) + -2)|0;
    $4 = (($lp) + ($3<<2)|0);
    $5 = HEAP32[$4>>2]|0;
    $6 = (0 - ($5))|0;
    $7 = (($2) + ($6)|0);
    $9 = (FUNCTION_TABLE_iii[$cmp & 8191]($8,$7)|0);
    $10 = ($9|0)>(-1);
    if ($10) {
     $11 = (FUNCTION_TABLE_iii[$cmp & 8191]($8,$2)|0);
     $12 = ($11|0)>(-1);
     if ($12) {
      $i$0$lcssa = $i$04;
      break L1;
     }
    }
    $13 = (FUNCTION_TABLE_iii[$cmp & 8191]($7,$2)|0);
    $14 = ($13|0)>(-1);
    $15 = (($i$04) + 1)|0;
    $16 = (($ar) + ($i$04<<2)|0);
    if ($14) {
     HEAP32[$16>>2] = $7;
     $17 = (($$012) + -1)|0;
     $$0$be = $7;$$01$be = $17;
    } else {
     HEAP32[$16>>2] = $2;
     $$0$be = $2;$$01$be = $3;
    }
    $18 = ($$01$be|0)>(1);
    if (!($18)) {
     $i$0$lcssa = $15;
     break L1;
    }
    $$pre = HEAP32[$ar>>2]|0;
    $$012 = $$01$be;$$03 = $$0$be;$8 = $$pre;$i$04 = $15;
   }
  } else {
   $i$0$lcssa = 1;
  }
 } while(0);
 _cycle($width,$ar,$i$0$lcssa);
 STACKTOP = sp;return;
}
function _cycle($width,$ar,$n) {
 $width = $width|0;
 $ar = $ar|0;
 $n = $n|0;
 var $$02 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$01 = 0;
 var $tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tmp = sp;
 $0 = ($n|0)<(2);
 L1: do {
  if (!($0)) {
   $1 = (($ar) + ($n<<2)|0);
   HEAP32[$1>>2] = $tmp;
   $2 = ($width|0)==(0);
   if (!($2)) {
    $$02 = $width;$6 = $tmp;
    while(1) {
     $3 = ($$02>>>0)>(256);
     $4 = $3 ? 256 : $$02;
     $5 = HEAP32[$ar>>2]|0;
     _memcpy(($6|0),($5|0),($4|0))|0;
     $i$01 = 0;
     while(1) {
      $7 = (($ar) + ($i$01<<2)|0);
      $8 = HEAP32[$7>>2]|0;
      $9 = (($i$01) + 1)|0;
      $10 = (($ar) + ($9<<2)|0);
      $11 = HEAP32[$10>>2]|0;
      _memcpy(($8|0),($11|0),($4|0))|0;
      $12 = HEAP32[$7>>2]|0;
      $13 = (($12) + ($4)|0);
      HEAP32[$7>>2] = $13;
      $exitcond = ($9|0)==($n|0);
      if ($exitcond) {
       break;
      } else {
       $i$01 = $9;
      }
     }
     $14 = ($$02|0)==($4|0);
     if ($14) {
      break L1;
     }
     $15 = (($$02) - ($4))|0;
     $$pre = HEAP32[$1>>2]|0;
     $$02 = $15;$6 = $$pre;
    }
   }
  }
 } while(0);
 STACKTOP = sp;return;
}
function _trinkle($head,$width,$cmp,$pp,$pshift,$trusty,$lp) {
 $head = $head|0;
 $width = $width|0;
 $cmp = $cmp|0;
 $pp = $pp|0;
 $pshift = $pshift|0;
 $trusty = $trusty|0;
 $lp = $lp|0;
 var $$0$i = 0, $$0$lcssa = 0, $$0$lcssa32 = 0, $$01$i5$i$ph = 0, $$01245 = 0, $$01245$phi = 0, $$02$i$i = 0, $$02$i3$i = 0, $$02$lcssa = 0, $$02$lcssa34 = 0, $$021047 = 0, $$03$lcssa = 0, $$03948 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $ar = 0, $i$0$lcssa = 0, $i$0$lcssa33 = 0, $i$01146 = 0, $nTrailingZeros$03$i$i = 0, $nTrailingZeros$03$i2$i = 0, $or$cond = 0, $p$sroa$0$0849 = 0, $p$sroa$8$0650 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ar = sp;
 $0 = HEAP32[$pp>>2]|0;
 $1 = ((($pp)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 HEAP32[$ar>>2] = $head;
 $3 = (0 - ($width))|0;
 $4 = ($0|0)!=(1);
 $5 = ($2|0)!=(0);
 $6 = $5 | $4;
 L1: do {
  if ($6) {
   $7 = (($lp) + ($pshift<<2)|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = (0 - ($8))|0;
   $10 = (($head) + ($9)|0);
   $11 = (FUNCTION_TABLE_iii[$cmp & 8191]($10,$head)|0);
   $12 = ($11|0)<(1);
   if ($12) {
    $$0$lcssa = $head;$$02$lcssa = $pshift;$$03$lcssa = $trusty;$i$0$lcssa = 1;
    label = 18;
   } else {
    $phitmp = ($trusty|0)==(0);
    $$01245 = $head;$$021047 = $pshift;$$03948 = $phitmp;$18 = $10;$i$01146 = 1;$p$sroa$0$0849 = $0;$p$sroa$8$0650 = $2;
    while(1) {
     $13 = ($$021047|0)>(1);
     $or$cond = $$03948 & $13;
     if ($or$cond) {
      $14 = (($$01245) + ($3)|0);
      $15 = (($$021047) + -2)|0;
      $16 = (($lp) + ($15<<2)|0);
      $17 = HEAP32[$16>>2]|0;
      $19 = (FUNCTION_TABLE_iii[$cmp & 8191]($14,$18)|0);
      $20 = ($19|0)>(-1);
      if ($20) {
       $$0$lcssa32 = $$01245;$$02$lcssa34 = $$021047;$i$0$lcssa33 = $i$01146;
       label = 19;
       break L1;
      }
      $21 = (0 - ($17))|0;
      $22 = (($14) + ($21)|0);
      $23 = (FUNCTION_TABLE_iii[$cmp & 8191]($22,$18)|0);
      $24 = ($23|0)>(-1);
      if ($24) {
       $$0$lcssa32 = $$01245;$$02$lcssa34 = $$021047;$i$0$lcssa33 = $i$01146;
       label = 19;
       break L1;
      }
     }
     $25 = (($i$01146) + 1)|0;
     $26 = (($ar) + ($i$01146<<2)|0);
     HEAP32[$26>>2] = $18;
     $27 = (($p$sroa$0$0849) + -1)|0;
     $28 = ($27|0)==(0);
     do {
      if ($28) {
       $46 = 32;
       label = 15;
      } else {
       $29 = $27 & 1;
       $30 = ($29|0)==(0);
       if ($30) {
        $$02$i$i = $27;$nTrailingZeros$03$i$i = 0;
        while(1) {
         $31 = (($nTrailingZeros$03$i$i) + 1)|0;
         $32 = $$02$i$i >>> 1;
         $33 = $32 & 1;
         $34 = ($33|0)==(0);
         if ($34) {
          $$02$i$i = $32;$nTrailingZeros$03$i$i = $31;
         } else {
          $44 = $31;
          break;
         }
        }
       } else {
        $35 = ($p$sroa$8$0650|0)==(0);
        if ($35) {
         $$01$i5$i$ph = 32;
        } else {
         $36 = $p$sroa$8$0650 & 1;
         $37 = ($36|0)==(0);
         if ($37) {
          $$02$i3$i = $p$sroa$8$0650;$nTrailingZeros$03$i2$i = 0;
         } else {
          $$0$i = 0;$48 = $p$sroa$0$0849;$51 = $p$sroa$8$0650;$55 = 0;
          break;
         }
         while(1) {
          $38 = (($nTrailingZeros$03$i2$i) + 1)|0;
          $39 = $$02$i3$i >>> 1;
          $40 = $39 & 1;
          $41 = ($40|0)==(0);
          if ($41) {
           $$02$i3$i = $39;$nTrailingZeros$03$i2$i = $38;
          } else {
           $$01$i5$i$ph = $38;
           break;
          }
         }
        }
        $42 = (($$01$i5$i$ph) + 32)|0;
        $44 = $42;
       }
       $43 = ($44>>>0)>(31);
       if ($43) {
        $46 = $44;
        label = 15;
       } else {
        $$0$i = $44;$48 = $p$sroa$0$0849;$51 = $p$sroa$8$0650;$55 = $44;
       }
      }
     } while(0);
     if ((label|0) == 15) {
      label = 0;
      $45 = (($46) + -32)|0;
      $$0$i = $45;$48 = $p$sroa$8$0650;$51 = 0;$55 = $46;
     }
     $47 = $48 >>> $$0$i;
     $49 = (32 - ($$0$i))|0;
     $50 = $51 << $49;
     $52 = $50 | $47;
     $53 = $51 >>> $$0$i;
     $54 = (($55) + ($$021047))|0;
     $56 = ($52|0)!=(1);
     $57 = ($53|0)!=(0);
     $58 = $57 | $56;
     if (!($58)) {
      $$0$lcssa32 = $18;$$02$lcssa34 = $54;$i$0$lcssa33 = $25;
      label = 19;
      break L1;
     }
     $$pre = HEAP32[$ar>>2]|0;
     $59 = (($lp) + ($54<<2)|0);
     $60 = HEAP32[$59>>2]|0;
     $61 = (0 - ($60))|0;
     $62 = (($18) + ($61)|0);
     $63 = (FUNCTION_TABLE_iii[$cmp & 8191]($62,$$pre)|0);
     $64 = ($63|0)<(1);
     if ($64) {
      $$0$lcssa = $18;$$02$lcssa = $54;$$03$lcssa = 0;$i$0$lcssa = $25;
      label = 18;
      break;
     } else {
      $$01245$phi = $18;$$021047 = $54;$$03948 = 1;$18 = $62;$i$01146 = $25;$p$sroa$0$0849 = $52;$p$sroa$8$0650 = $53;$$01245 = $$01245$phi;
     }
    }
   }
  } else {
   $$0$lcssa = $head;$$02$lcssa = $pshift;$$03$lcssa = $trusty;$i$0$lcssa = 1;
   label = 18;
  }
 } while(0);
 if ((label|0) == 18) {
  $65 = ($$03$lcssa|0)==(0);
  if ($65) {
   $$0$lcssa32 = $$0$lcssa;$$02$lcssa34 = $$02$lcssa;$i$0$lcssa33 = $i$0$lcssa;
   label = 19;
  }
 }
 if ((label|0) == 19) {
  _cycle($width,$ar,$i$0$lcssa33);
  _sift($$0$lcssa32,$width,$cmp,$$02$lcssa34,$lp);
 }
 STACKTOP = sp;return;
}
function ___floatscan($f,$prec,$pok) {
 $f = $f|0;
 $prec = $prec|0;
 $pok = $pok|0;
 var $$$i = 0, $$0 = 0.0, $$010$i = 0, $$012$i = 0, $$07$i = 0, $$0710$i = 0, $$0711$i = 0, $$1$i = 0.0, $$111$be$i = 0, $$111$ph$i = 0, $$16$i = 0, $$2$i = 0, $$24$i = 0, $$3$be$i = 0, $$3$lcssa$i = 0, $$3112$i = 0, $$in = 0, $$k$0$i = 0, $$lcssa = 0, $$lcssa258 = 0;
 var $$lcssa258$lcssa = 0, $$lcssa259 = 0, $$lcssa259$lcssa = 0, $$lcssa265 = 0, $$lcssa266 = 0, $$lcssa267 = 0, $$lcssa277 = 0, $$lnz$0$i = 0, $$neg$i = 0, $$neg40$i = 0, $$not$i = 0, $$old8 = 0, $$pn$i = 0.0, $$pre$i = 0, $$pre$i17 = 0, $$pre$phi43$iZ2D = 0.0, $$pre42$i = 0.0, $$promoted$i = 0, $$sink$off0$i = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0.0, $183 = 0.0, $184 = 0.0, $185 = 0.0, $186 = 0, $187 = 0, $188 = 0.0, $189 = 0.0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0.0, $259 = 0.0, $26 = 0, $260 = 0;
 var $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0.0, $268 = 0.0, $269 = 0.0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0;
 var $28 = 0, $280 = 0.0, $281 = 0.0, $282 = 0.0, $283 = 0, $284 = 0, $285 = 0.0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0;
 var $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0.0, $31 = 0, $310 = 0.0, $311 = 0.0, $312 = 0, $313 = 0, $314 = 0;
 var $315 = 0, $316 = 0, $317 = 0.0, $318 = 0.0, $319 = 0.0, $32 = 0, $320 = 0.0, $321 = 0.0, $322 = 0.0, $323 = 0, $324 = 0, $325 = 0.0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0;
 var $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0;
 var $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0;
 var $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0;
 var $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0;
 var $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0;
 var $423 = 0, $424 = 0.0, $425 = 0.0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0;
 var $441 = 0.0, $442 = 0.0, $443 = 0.0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0.0, $454 = 0.0, $455 = 0.0, $456 = 0, $457 = 0, $458 = 0, $459 = 0;
 var $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0.0, $466 = 0.0, $467 = 0.0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0;
 var $478 = 0, $479 = 0.0, $48 = 0, $480 = 0, $481 = 0.0, $482 = 0.0, $483 = 0, $484 = 0.0, $485 = 0, $486 = 0.0, $487 = 0.0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0.0, $492 = 0.0, $493 = 0, $494 = 0, $495 = 0;
 var $496 = 0, $497 = 0.0, $498 = 0.0, $499 = 0.0, $5 = 0, $50 = 0.0, $500 = 0, $501 = 0, $502 = 0, $503 = 0.0, $504 = 0.0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0.0, $510 = 0, $511 = 0, $512 = 0;
 var $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0.0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0;
 var $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0;
 var $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0;
 var $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0;
 var $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0;
 var $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0.0, $619 = 0, $62 = 0, $620 = 0;
 var $621 = 0, $622 = 0, $623 = 0.0, $624 = 0.0, $625 = 0.0, $626 = 0, $627 = 0.0, $628 = 0.0, $629 = 0.0, $63 = 0, $630 = 0.0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0;
 var $64 = 0, $640 = 0, $641 = 0.0, $642 = 0.0, $643 = 0.0, $644 = 0, $645 = 0.0, $646 = 0.0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0.0, $651 = 0.0, $652 = 0.0, $653 = 0.0, $654 = 0, $655 = 0, $656 = 0.0, $657 = 0;
 var $658 = 0.0, $659 = 0.0, $66 = 0, $660 = 0.0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0.0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0.0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0;
 var $676 = 0, $677 = 0.0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0.0, $683 = 0, $684 = 0, $685 = 0.0, $686 = 0.0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0;
 var $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0;
 var $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $a$0$lcssa161$i = 0, $a$093$i = 0, $a$1$i = 0, $a$1$i$lcssa = 0, $a$2$ph46$i = 0, $a$4$i = 0, $a$4$i$lcssa250 = 0, $a$4$i251 = 0, $a$4$ph$i = 0, $a$4$ph167$i = 0, $a$586$i = 0, $a$6$i = 0, $a$6$i$lcssa = 0, $a$6$i$lcssa$lcssa = 0, $bias$0$i = 0.0, $bias$0$i25 = 0.0, $bits$0$ph = 0, $brmerge$i28 = 0;
 var $c$0 = 0, $c$0$i = 0, $c$1$lcssa = 0, $c$1$ph$i = 0, $c$179 = 0, $c$2 = 0, $c$2$i = 0, $c$2$lcssa$i = 0, $c$377 = 0, $c$4 = 0, $c$5 = 0, $c$6 = 0, $carry$095$i = 0, $carry1$0$i = 0, $carry1$1$i = 0, $carry1$1$i$lcssa = 0, $carry1$1$i$lcssa$lcssa = 0, $carry4$089$i = 0, $cond$i = 0, $d$0$i = 0;
 var $denormal$0$i = 0, $denormal$1$i = 0, $denormal$2$i = 0, $e2$0$i19 = 0, $e2$0$ph$i = 0, $e2$1$i = 0, $e2$1$i248 = 0, $e2$1$ph$i = 0, $e2$1$ph166$i = 0, $e2$3$i = 0, $e2$4$i = 0, $emin$0$ph = 0, $exitcond151$i = 0, $frac$0$i = 0.0, $frac$1$i = 0.0, $frac$3$i = 0.0, $gotdig$0$i = 0, $gotdig$0$i$lcssa244 = 0, $gotdig$0$i12 = 0, $gotdig$0$i12$lcssa275 = 0;
 var $gotdig$2$i = 0, $gotdig$2$i$lcssa = 0, $gotdig$2$i13 = 0, $gotdig$3$i = 0, $gotdig$3$lcssa$i = 0, $gotdig$3108$i = 0, $gotdig$3108$i$lcssa = 0, $gotdig$4$i = 0, $gotrad$0$i = 0, $gotrad$0$i$lcssa = 0, $gotrad$0$i14 = 0, $gotrad$1$i = 0, $gotrad$1$lcssa$i = 0, $gotrad$1109$i = 0, $gotrad$2$i = 0, $gottail$0$i = 0, $gottail$1$i = 0, $gottail$2$i = 0, $i$0$lcssa = 0, $i$078 = 0;
 var $i$1 = 0, $i$276 = 0, $i$3 = 0, $i$4 = 0, $i$4$lcssa = 0, $j$0$lcssa$i = 0, $j$0111$i = 0, $j$0111$i$lcssa = 0, $j$075$i = 0, $j$076$i = 0, $j$077$i = 0, $j$2$i = 0, $j$3102$i = 0, $k$0$lcssa$i = 0, $k$0110$i = 0, $k$0110$i$lcssa = 0, $k$071$i = 0, $k$072$i = 0, $k$073$i = 0, $k$2$i = 0;
 var $k$3$i = 0, $k$494$i = 0, $k$5$i = 0, $k$5$in$i = 0, $k$5$z$2$i = 0, $k$687$i = 0, $lnz$0$lcssa$i = 0, $lnz$0107$i = 0, $lnz$0107$i$lcssa = 0, $lnz$065$i = 0, $lnz$066$i = 0, $lnz$067$i = 0, $lnz$2$i = 0, $or$cond = 0, $or$cond$i = 0, $or$cond$i16 = 0, $or$cond18$i = 0, $or$cond192$i = 0, $or$cond20$i = 0, $or$cond21$i = 0;
 var $or$cond22$i = 0, $or$cond25$i = 0, $or$cond26$i = 0, $or$cond3$i = 0, $or$cond4$i = 0, $or$cond5 = 0, $or$cond6$i = 0, $or$cond7 = 0, $or$cond9 = 0, $or$cond9$i = 0, $or$cond9$i27 = 0, $rp$0$lcssa162$i = 0, $rp$092$i = 0, $rp$1$i18 = 0, $rp$1$i18$lcssa = 0, $rp$2$ph44$i = 0, $rp$4$ph$i = 0, $rp$4$ph42$i = 0, $rp$585$i = 0, $rp$6$i = 0;
 var $rp$6$i$lcssa = 0, $rp$6$i$lcssa$lcssa = 0, $scale$0$i = 0.0, $scale$1$i = 0.0, $scale$2$i = 0.0, $sign$0 = 0, $storemerge$i = 0, $sum$i = 0, $x$0$i = 0, $x$0$i$lcssa = 0, $x$1$i = 0, $x$2$i = 0, $x$3$lcssa$i = 0, $x$324$i = 0, $x$4$lcssa$i = 0, $x$419$i = 0, $x$5$i = 0, $x$6$i = 0, $x$i = 0, $y$0$i = 0.0;
 var $y$0$i$lcssa = 0.0, $y$1$i = 0.0, $y$1$i24 = 0.0, $y$2$i = 0.0, $y$2$i26 = 0.0, $y$3$i = 0.0, $y$3$lcssa$i = 0.0, $y$320$i = 0.0, $y$4$i = 0.0, $y$5$i = 0.0, $z$0$i = 0, $z$1$i = 0, $z$1$ph45$i = 0, $z$10$1$i = 0, $z$10$i = 0, $z$2$i = 0, $z$3$i = 0, $z$3$i$lcssa = 0, $z$3$i$lcssa$lcssa = 0, $z$4$i = 0;
 var $z$6$ph$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 512|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $x$i = sp;
 switch ($prec|0) {
 case 0:  {
  $bits$0$ph = 24;$emin$0$ph = -149;
  label = 4;
  break;
 }
 case 1:  {
  $bits$0$ph = 53;$emin$0$ph = -1074;
  label = 4;
  break;
 }
 case 2:  {
  $bits$0$ph = 53;$emin$0$ph = -1074;
  label = 4;
  break;
 }
 default: {
  $$0 = 0.0;
 }
 }
 L4: do {
  if ((label|0) == 4) {
   $0 = ((($f)) + 4|0);
   $1 = ((($f)) + 100|0);
   while(1) {
    $2 = HEAP32[$0>>2]|0;
    $3 = HEAP32[$1>>2]|0;
    $4 = ($2>>>0)<($3>>>0);
    if ($4) {
     $5 = ((($2)) + 1|0);
     HEAP32[$0>>2] = $5;
     $6 = HEAP8[$2>>0]|0;
     $7 = $6&255;
     $9 = $7;
    } else {
     $8 = (___shgetc($f)|0);
     $9 = $8;
    }
    $10 = (_isspace($9)|0);
    $11 = ($10|0)==(0);
    if ($11) {
     $$lcssa277 = $9;
     break;
    }
   }
   $12 = ($$lcssa277|0)==(45);
   L13: do {
    switch ($$lcssa277|0) {
    case 43: case 45:  {
     $13 = $12&1;
     $14 = $13 << 1;
     $15 = (1 - ($14))|0;
     $16 = HEAP32[$0>>2]|0;
     $17 = HEAP32[$1>>2]|0;
     $18 = ($16>>>0)<($17>>>0);
     if ($18) {
      $19 = ((($16)) + 1|0);
      HEAP32[$0>>2] = $19;
      $20 = HEAP8[$16>>0]|0;
      $21 = $20&255;
      $c$0 = $21;$sign$0 = $15;
      break L13;
     } else {
      $22 = (___shgetc($f)|0);
      $c$0 = $22;$sign$0 = $15;
      break L13;
     }
     break;
    }
    default: {
     $c$0 = $$lcssa277;$sign$0 = 1;
    }
    }
   } while(0);
   $c$179 = $c$0;$i$078 = 0;
   while(1) {
    $23 = $c$179 | 32;
    $24 = (619436 + ($i$078)|0);
    $25 = HEAP8[$24>>0]|0;
    $26 = $25 << 24 >> 24;
    $27 = ($23|0)==($26|0);
    if (!($27)) {
     $c$1$lcssa = $c$179;$i$0$lcssa = $i$078;
     break;
    }
    $28 = ($i$078>>>0)<(7);
    do {
     if ($28) {
      $29 = HEAP32[$0>>2]|0;
      $30 = HEAP32[$1>>2]|0;
      $31 = ($29>>>0)<($30>>>0);
      if ($31) {
       $32 = ((($29)) + 1|0);
       HEAP32[$0>>2] = $32;
       $33 = HEAP8[$29>>0]|0;
       $34 = $33&255;
       $c$2 = $34;
       break;
      } else {
       $35 = (___shgetc($f)|0);
       $c$2 = $35;
       break;
      }
     } else {
      $c$2 = $c$179;
     }
    } while(0);
    $36 = (($i$078) + 1)|0;
    $37 = ($36>>>0)<(8);
    if ($37) {
     $c$179 = $c$2;$i$078 = $36;
    } else {
     $c$1$lcssa = $c$2;$i$0$lcssa = $36;
     break;
    }
   }
   L29: do {
    switch ($i$0$lcssa|0) {
    case 8:  {
     break;
    }
    case 3:  {
     label = 23;
     break;
    }
    default: {
     $38 = ($i$0$lcssa>>>0)>(3);
     $39 = ($pok|0)!=(0);
     $or$cond5 = $39 & $38;
     if ($or$cond5) {
      $40 = ($i$0$lcssa|0)==(8);
      if ($40) {
       break L29;
      } else {
       label = 23;
       break L29;
      }
     }
     $53 = ($i$0$lcssa|0)==(0);
     L34: do {
      if ($53) {
       $c$377 = $c$1$lcssa;$i$276 = 0;
       while(1) {
        $54 = $c$377 | 32;
        $55 = (619445 + ($i$276)|0);
        $56 = HEAP8[$55>>0]|0;
        $57 = $56 << 24 >> 24;
        $58 = ($54|0)==($57|0);
        if (!($58)) {
         $c$5 = $c$377;$i$3 = $i$276;
         break L34;
        }
        $59 = ($i$276>>>0)<(2);
        do {
         if ($59) {
          $60 = HEAP32[$0>>2]|0;
          $61 = HEAP32[$1>>2]|0;
          $62 = ($60>>>0)<($61>>>0);
          if ($62) {
           $63 = ((($60)) + 1|0);
           HEAP32[$0>>2] = $63;
           $64 = HEAP8[$60>>0]|0;
           $65 = $64&255;
           $c$4 = $65;
           break;
          } else {
           $66 = (___shgetc($f)|0);
           $c$4 = $66;
           break;
          }
         } else {
          $c$4 = $c$377;
         }
        } while(0);
        $67 = (($i$276) + 1)|0;
        $68 = ($67>>>0)<(3);
        if ($68) {
         $c$377 = $c$4;$i$276 = $67;
        } else {
         $c$5 = $c$4;$i$3 = $67;
         break;
        }
       }
      } else {
       $c$5 = $c$1$lcssa;$i$3 = $i$0$lcssa;
      }
     } while(0);
     switch ($i$3|0) {
     case 3:  {
      $69 = HEAP32[$0>>2]|0;
      $70 = HEAP32[$1>>2]|0;
      $71 = ($69>>>0)<($70>>>0);
      if ($71) {
       $72 = ((($69)) + 1|0);
       HEAP32[$0>>2] = $72;
       $73 = HEAP8[$69>>0]|0;
       $74 = $73&255;
       $77 = $74;
      } else {
       $75 = (___shgetc($f)|0);
       $77 = $75;
      }
      $76 = ($77|0)==(40);
      if ($76) {
       $i$4 = 1;
      } else {
       $78 = HEAP32[$1>>2]|0;
       $79 = ($78|0)==(0|0);
       if ($79) {
        $$0 = nan;
        break L4;
       }
       $80 = HEAP32[$0>>2]|0;
       $81 = ((($80)) + -1|0);
       HEAP32[$0>>2] = $81;
       $$0 = nan;
       break L4;
      }
      while(1) {
       $82 = HEAP32[$0>>2]|0;
       $83 = HEAP32[$1>>2]|0;
       $84 = ($82>>>0)<($83>>>0);
       if ($84) {
        $85 = ((($82)) + 1|0);
        HEAP32[$0>>2] = $85;
        $86 = HEAP8[$82>>0]|0;
        $87 = $86&255;
        $90 = $87;
       } else {
        $88 = (___shgetc($f)|0);
        $90 = $88;
       }
       $89 = (($90) + -48)|0;
       $91 = ($89>>>0)<(10);
       $92 = (($90) + -65)|0;
       $93 = ($92>>>0)<(26);
       $or$cond = $91 | $93;
       if (!($or$cond)) {
        $94 = (($90) + -97)|0;
        $95 = ($94>>>0)<(26);
        $96 = ($90|0)==(95);
        $or$cond7 = $96 | $95;
        if (!($or$cond7)) {
         $$lcssa = $90;$i$4$lcssa = $i$4;
         break;
        }
       }
       $108 = (($i$4) + 1)|0;
       $i$4 = $108;
      }
      $97 = ($$lcssa|0)==(41);
      if ($97) {
       $$0 = nan;
       break L4;
      }
      $98 = HEAP32[$1>>2]|0;
      $99 = ($98|0)==(0|0);
      if (!($99)) {
       $100 = HEAP32[$0>>2]|0;
       $101 = ((($100)) + -1|0);
       HEAP32[$0>>2] = $101;
      }
      if (!($39)) {
       $103 = (___errno_location()|0);
       HEAP32[$103>>2] = 22;
       ___shlim($f,0);
       $$0 = 0.0;
       break L4;
      }
      $102 = ($i$4$lcssa|0)==(0);
      if ($102) {
       $$0 = nan;
       break L4;
      } else {
       $$in = $i$4$lcssa;
      }
      while(1) {
       $104 = (($$in) + -1)|0;
       if (!($99)) {
        $105 = HEAP32[$0>>2]|0;
        $106 = ((($105)) + -1|0);
        HEAP32[$0>>2] = $106;
       }
       $107 = ($104|0)==(0);
       if ($107) {
        $$0 = nan;
        break L4;
       } else {
        $$in = $104;
       }
      }
      break;
     }
     case 0:  {
      $114 = ($c$5|0)==(48);
      do {
       if ($114) {
        $115 = HEAP32[$0>>2]|0;
        $116 = HEAP32[$1>>2]|0;
        $117 = ($115>>>0)<($116>>>0);
        if ($117) {
         $118 = ((($115)) + 1|0);
         HEAP32[$0>>2] = $118;
         $119 = HEAP8[$115>>0]|0;
         $120 = $119&255;
         $123 = $120;
        } else {
         $121 = (___shgetc($f)|0);
         $123 = $121;
        }
        $122 = $123 | 32;
        $124 = ($122|0)==(120);
        if (!($124)) {
         $326 = HEAP32[$1>>2]|0;
         $327 = ($326|0)==(0|0);
         if ($327) {
          $c$6 = 48;
          break;
         }
         $328 = HEAP32[$0>>2]|0;
         $329 = ((($328)) + -1|0);
         HEAP32[$0>>2] = $329;
         $c$6 = 48;
         break;
        }
        $125 = HEAP32[$0>>2]|0;
        $126 = HEAP32[$1>>2]|0;
        $127 = ($125>>>0)<($126>>>0);
        if ($127) {
         $128 = ((($125)) + 1|0);
         HEAP32[$0>>2] = $128;
         $129 = HEAP8[$125>>0]|0;
         $130 = $129&255;
         $c$0$i = $130;$gotdig$0$i = 0;
        } else {
         $131 = (___shgetc($f)|0);
         $c$0$i = $131;$gotdig$0$i = 0;
        }
        L94: while(1) {
         switch ($c$0$i|0) {
         case 46:  {
          $gotdig$0$i$lcssa244 = $gotdig$0$i;
          label = 74;
          break L94;
          break;
         }
         case 48:  {
          break;
         }
         default: {
          $169 = 0;$171 = 0;$693 = 0;$694 = 0;$c$2$i = $c$0$i;$gotdig$2$i = $gotdig$0$i;$gotrad$0$i = 0;$gottail$0$i = 0;$scale$0$i = 1.0;$x$0$i = 0;$y$0$i = 0.0;
          break L94;
         }
         }
         $132 = HEAP32[$0>>2]|0;
         $133 = HEAP32[$1>>2]|0;
         $134 = ($132>>>0)<($133>>>0);
         if ($134) {
          $135 = ((($132)) + 1|0);
          HEAP32[$0>>2] = $135;
          $136 = HEAP8[$132>>0]|0;
          $137 = $136&255;
          $c$0$i = $137;$gotdig$0$i = 1;
          continue;
         } else {
          $138 = (___shgetc($f)|0);
          $c$0$i = $138;$gotdig$0$i = 1;
          continue;
         }
        }
        if ((label|0) == 74) {
         $139 = HEAP32[$0>>2]|0;
         $140 = HEAP32[$1>>2]|0;
         $141 = ($139>>>0)<($140>>>0);
         if ($141) {
          $142 = ((($139)) + 1|0);
          HEAP32[$0>>2] = $142;
          $143 = HEAP8[$139>>0]|0;
          $144 = $143&255;
          $c$1$ph$i = $144;
         } else {
          $145 = (___shgetc($f)|0);
          $c$1$ph$i = $145;
         }
         $146 = ($c$1$ph$i|0)==(48);
         if ($146) {
          $154 = 0;$155 = 0;
          while(1) {
           $147 = HEAP32[$0>>2]|0;
           $148 = HEAP32[$1>>2]|0;
           $149 = ($147>>>0)<($148>>>0);
           if ($149) {
            $150 = ((($147)) + 1|0);
            HEAP32[$0>>2] = $150;
            $151 = HEAP8[$147>>0]|0;
            $152 = $151&255;
            $159 = $152;
           } else {
            $153 = (___shgetc($f)|0);
            $159 = $153;
           }
           $156 = (_i64Add(($154|0),($155|0),-1,-1)|0);
           $157 = tempRet0;
           $158 = ($159|0)==(48);
           if ($158) {
            $154 = $156;$155 = $157;
           } else {
            $169 = 0;$171 = 0;$693 = $156;$694 = $157;$c$2$i = $159;$gotdig$2$i = 1;$gotrad$0$i = 1;$gottail$0$i = 0;$scale$0$i = 1.0;$x$0$i = 0;$y$0$i = 0.0;
            break;
           }
          }
         } else {
          $169 = 0;$171 = 0;$693 = 0;$694 = 0;$c$2$i = $c$1$ph$i;$gotdig$2$i = $gotdig$0$i$lcssa244;$gotrad$0$i = 1;$gottail$0$i = 0;$scale$0$i = 1.0;$x$0$i = 0;$y$0$i = 0.0;
         }
        }
        while(1) {
         $160 = (($c$2$i) + -48)|0;
         $161 = ($160>>>0)<(10);
         $$pre$i = $c$2$i | 32;
         if ($161) {
          label = 86;
         } else {
          $162 = (($$pre$i) + -97)|0;
          $163 = ($162>>>0)<(6);
          $164 = ($c$2$i|0)==(46);
          $or$cond6$i = $164 | $163;
          if (!($or$cond6$i)) {
           $213 = $171;$214 = $693;$216 = $169;$217 = $694;$c$2$lcssa$i = $c$2$i;$gotdig$2$i$lcssa = $gotdig$2$i;$gotrad$0$i$lcssa = $gotrad$0$i;$x$0$i$lcssa = $x$0$i;$y$0$i$lcssa = $y$0$i;
           break;
          }
          if ($164) {
           $165 = ($gotrad$0$i|0)==(0);
           if ($165) {
            $695 = $171;$696 = $169;$697 = $171;$698 = $169;$gotdig$3$i = $gotdig$2$i;$gotrad$1$i = 1;$gottail$2$i = $gottail$0$i;$scale$2$i = $scale$0$i;$x$2$i = $x$0$i;$y$2$i = $y$0$i;
           } else {
            $213 = $171;$214 = $693;$216 = $169;$217 = $694;$c$2$lcssa$i = 46;$gotdig$2$i$lcssa = $gotdig$2$i;$gotrad$0$i$lcssa = $gotrad$0$i;$x$0$i$lcssa = $x$0$i;$y$0$i$lcssa = $y$0$i;
            break;
           }
          } else {
           label = 86;
          }
         }
         if ((label|0) == 86) {
          label = 0;
          $166 = ($c$2$i|0)>(57);
          $167 = (($$pre$i) + -87)|0;
          $d$0$i = $166 ? $167 : $160;
          $168 = ($169|0)<(0);
          $170 = ($171>>>0)<(8);
          $172 = ($169|0)==(0);
          $173 = $172 & $170;
          $174 = $168 | $173;
          do {
           if ($174) {
            $175 = $x$0$i << 4;
            $176 = (($d$0$i) + ($175))|0;
            $gottail$1$i = $gottail$0$i;$scale$1$i = $scale$0$i;$x$1$i = $176;$y$1$i = $y$0$i;
           } else {
            $177 = ($169|0)<(0);
            $178 = ($171>>>0)<(14);
            $179 = ($169|0)==(0);
            $180 = $179 & $178;
            $181 = $177 | $180;
            if ($181) {
             $182 = (+($d$0$i|0));
             $183 = $scale$0$i * 0.0625;
             $184 = $183 * $182;
             $185 = $y$0$i + $184;
             $gottail$1$i = $gottail$0$i;$scale$1$i = $183;$x$1$i = $x$0$i;$y$1$i = $185;
             break;
            }
            $186 = ($d$0$i|0)==(0);
            $187 = ($gottail$0$i|0)!=(0);
            $or$cond$i = $187 | $186;
            if ($or$cond$i) {
             $gottail$1$i = $gottail$0$i;$scale$1$i = $scale$0$i;$x$1$i = $x$0$i;$y$1$i = $y$0$i;
            } else {
             $188 = $scale$0$i * 0.5;
             $189 = $y$0$i + $188;
             $gottail$1$i = 1;$scale$1$i = $scale$0$i;$x$1$i = $x$0$i;$y$1$i = $189;
            }
           }
          } while(0);
          $190 = (_i64Add(($171|0),($169|0),1,0)|0);
          $191 = tempRet0;
          $695 = $693;$696 = $694;$697 = $190;$698 = $191;$gotdig$3$i = 1;$gotrad$1$i = $gotrad$0$i;$gottail$2$i = $gottail$1$i;$scale$2$i = $scale$1$i;$x$2$i = $x$1$i;$y$2$i = $y$1$i;
         }
         $192 = HEAP32[$0>>2]|0;
         $193 = HEAP32[$1>>2]|0;
         $194 = ($192>>>0)<($193>>>0);
         if ($194) {
          $195 = ((($192)) + 1|0);
          HEAP32[$0>>2] = $195;
          $196 = HEAP8[$192>>0]|0;
          $197 = $196&255;
          $169 = $698;$171 = $697;$693 = $695;$694 = $696;$c$2$i = $197;$gotdig$2$i = $gotdig$3$i;$gotrad$0$i = $gotrad$1$i;$gottail$0$i = $gottail$2$i;$scale$0$i = $scale$2$i;$x$0$i = $x$2$i;$y$0$i = $y$2$i;
          continue;
         } else {
          $198 = (___shgetc($f)|0);
          $169 = $698;$171 = $697;$693 = $695;$694 = $696;$c$2$i = $198;$gotdig$2$i = $gotdig$3$i;$gotrad$0$i = $gotrad$1$i;$gottail$0$i = $gottail$2$i;$scale$0$i = $scale$2$i;$x$0$i = $x$2$i;$y$0$i = $y$2$i;
          continue;
         }
        }
        $199 = ($gotdig$2$i$lcssa|0)==(0);
        if ($199) {
         $200 = HEAP32[$1>>2]|0;
         $201 = ($200|0)==(0|0);
         if (!($201)) {
          $202 = HEAP32[$0>>2]|0;
          $203 = ((($202)) + -1|0);
          HEAP32[$0>>2] = $203;
         }
         $204 = ($pok|0)==(0);
         if ($204) {
          ___shlim($f,0);
         } else {
          if (!($201)) {
           $205 = HEAP32[$0>>2]|0;
           $206 = ((($205)) + -1|0);
           HEAP32[$0>>2] = $206;
           $207 = ($gotrad$0$i$lcssa|0)==(0);
           if (!($207)) {
            $208 = ((($205)) + -2|0);
            HEAP32[$0>>2] = $208;
           }
          }
         }
         $209 = (+($sign$0|0));
         $210 = $209 * 0.0;
         $$0 = $210;
         break L4;
        }
        $211 = ($gotrad$0$i$lcssa|0)==(0);
        $212 = $211 ? $213 : $214;
        $215 = $211 ? $216 : $217;
        $218 = ($216|0)<(0);
        $219 = ($213>>>0)<(8);
        $220 = ($216|0)==(0);
        $221 = $220 & $219;
        $222 = $218 | $221;
        if ($222) {
         $224 = $213;$225 = $216;$x$324$i = $x$0$i$lcssa;
         while(1) {
          $223 = $x$324$i << 4;
          $226 = (_i64Add(($224|0),($225|0),1,0)|0);
          $227 = tempRet0;
          $228 = ($227|0)<(0);
          $229 = ($226>>>0)<(8);
          $230 = ($227|0)==(0);
          $231 = $230 & $229;
          $232 = $228 | $231;
          if ($232) {
           $224 = $226;$225 = $227;$x$324$i = $223;
          } else {
           $x$3$lcssa$i = $223;
           break;
          }
         }
        } else {
         $x$3$lcssa$i = $x$0$i$lcssa;
        }
        $233 = $c$2$lcssa$i | 32;
        $234 = ($233|0)==(112);
        if ($234) {
         $235 = (_scanexp($f,$pok)|0);
         $236 = tempRet0;
         $237 = ($235|0)==(0);
         $238 = ($236|0)==(-2147483648);
         $239 = $237 & $238;
         if ($239) {
          $240 = ($pok|0)==(0);
          if ($240) {
           ___shlim($f,0);
           $$0 = 0.0;
           break L4;
          }
          $241 = HEAP32[$1>>2]|0;
          $242 = ($241|0)==(0|0);
          if ($242) {
           $253 = 0;$254 = 0;
          } else {
           $243 = HEAP32[$0>>2]|0;
           $244 = ((($243)) + -1|0);
           HEAP32[$0>>2] = $244;
           $253 = 0;$254 = 0;
          }
         } else {
          $253 = $235;$254 = $236;
         }
        } else {
         $245 = HEAP32[$1>>2]|0;
         $246 = ($245|0)==(0|0);
         if ($246) {
          $253 = 0;$254 = 0;
         } else {
          $247 = HEAP32[$0>>2]|0;
          $248 = ((($247)) + -1|0);
          HEAP32[$0>>2] = $248;
          $253 = 0;$254 = 0;
         }
        }
        $249 = (_bitshift64Shl(($212|0),($215|0),2)|0);
        $250 = tempRet0;
        $251 = (_i64Add(($249|0),($250|0),-32,-1)|0);
        $252 = tempRet0;
        $255 = (_i64Add(($251|0),($252|0),($253|0),($254|0))|0);
        $256 = tempRet0;
        $257 = ($x$3$lcssa$i|0)==(0);
        if ($257) {
         $258 = (+($sign$0|0));
         $259 = $258 * 0.0;
         $$0 = $259;
         break L4;
        }
        $260 = (0 - ($emin$0$ph))|0;
        $261 = ($256|0)>(0);
        $262 = ($255>>>0)>($260>>>0);
        $263 = ($256|0)==(0);
        $264 = $263 & $262;
        $265 = $261 | $264;
        if ($265) {
         $266 = (___errno_location()|0);
         HEAP32[$266>>2] = 34;
         $267 = (+($sign$0|0));
         $268 = $267 * 1.7976931348623157E+308;
         $269 = $268 * 1.7976931348623157E+308;
         $$0 = $269;
         break L4;
        }
        $270 = (($emin$0$ph) + -106)|0;
        $271 = ($270|0)<(0);
        $272 = $271 << 31 >> 31;
        $273 = ($256|0)<($272|0);
        $274 = ($255>>>0)<($270>>>0);
        $275 = ($256|0)==($272|0);
        $276 = $275 & $274;
        $277 = $273 | $276;
        if ($277) {
         $279 = (___errno_location()|0);
         HEAP32[$279>>2] = 34;
         $280 = (+($sign$0|0));
         $281 = $280 * 2.2250738585072014E-308;
         $282 = $281 * 2.2250738585072014E-308;
         $$0 = $282;
         break L4;
        }
        $278 = ($x$3$lcssa$i|0)>(-1);
        if ($278) {
         $288 = $255;$289 = $256;$x$419$i = $x$3$lcssa$i;$y$320$i = $y$0$i$lcssa;
         while(1) {
          $283 = !($y$320$i >= 0.5);
          $284 = $x$419$i << 1;
          $285 = $y$320$i + -1.0;
          $286 = $283&1;
          $287 = $286 | $284;
          $x$5$i = $287 ^ 1;
          $$pn$i = $283 ? $y$320$i : $285;
          $y$4$i = $y$320$i + $$pn$i;
          $290 = (_i64Add(($288|0),($289|0),-1,-1)|0);
          $291 = tempRet0;
          $292 = ($287|0)>(-1);
          if ($292) {
           $288 = $290;$289 = $291;$x$419$i = $x$5$i;$y$320$i = $y$4$i;
          } else {
           $297 = $290;$298 = $291;$x$4$lcssa$i = $x$5$i;$y$3$lcssa$i = $y$4$i;
           break;
          }
         }
        } else {
         $297 = $255;$298 = $256;$x$4$lcssa$i = $x$3$lcssa$i;$y$3$lcssa$i = $y$0$i$lcssa;
        }
        $293 = ($emin$0$ph|0)<(0);
        $294 = $293 << 31 >> 31;
        $295 = (_i64Subtract(32,0,($emin$0$ph|0),($294|0))|0);
        $296 = tempRet0;
        $299 = (_i64Add(($297|0),($298|0),($295|0),($296|0))|0);
        $300 = tempRet0;
        $301 = (0)>($300|0);
        $302 = ($bits$0$ph>>>0)>($299>>>0);
        $303 = (0)==($300|0);
        $304 = $303 & $302;
        $305 = $301 | $304;
        if ($305) {
         $306 = ($299|0)<(0);
         if ($306) {
          $$0710$i = 0;
          label = 127;
         } else {
          $$07$i = $299;
          label = 125;
         }
        } else {
         $$07$i = $bits$0$ph;
         label = 125;
        }
        if ((label|0) == 125) {
         $307 = ($$07$i|0)<(53);
         if ($307) {
          $$0710$i = $$07$i;
          label = 127;
         } else {
          $$pre42$i = (+($sign$0|0));
          $$0711$i = $$07$i;$$pre$phi43$iZ2D = $$pre42$i;$bias$0$i = 0.0;
         }
        }
        if ((label|0) == 127) {
         $308 = (84 - ($$0710$i))|0;
         $309 = (+_scalbn(1.0,$308));
         $310 = (+($sign$0|0));
         $311 = (+_copysignl($309,$310));
         $$0711$i = $$0710$i;$$pre$phi43$iZ2D = $310;$bias$0$i = $311;
        }
        $312 = ($$0711$i|0)<(32);
        $313 = $y$3$lcssa$i != 0.0;
        $or$cond4$i = $313 & $312;
        $314 = $x$4$lcssa$i & 1;
        $315 = ($314|0)==(0);
        $or$cond9$i = $315 & $or$cond4$i;
        $316 = $or$cond9$i&1;
        $x$6$i = (($316) + ($x$4$lcssa$i))|0;
        $y$5$i = $or$cond9$i ? 0.0 : $y$3$lcssa$i;
        $317 = (+($x$6$i>>>0));
        $318 = $$pre$phi43$iZ2D * $317;
        $319 = $bias$0$i + $318;
        $320 = $$pre$phi43$iZ2D * $y$5$i;
        $321 = $320 + $319;
        $322 = $321 - $bias$0$i;
        $323 = $322 != 0.0;
        if (!($323)) {
         $324 = (___errno_location()|0);
         HEAP32[$324>>2] = 34;
        }
        $325 = (+_scalbnl($322,$297));
        $$0 = $325;
        break L4;
       } else {
        $c$6 = $c$5;
       }
      } while(0);
      $sum$i = (($emin$0$ph) + ($bits$0$ph))|0;
      $330 = (0 - ($sum$i))|0;
      $$010$i = $c$6;$gotdig$0$i12 = 0;
      L184: while(1) {
       switch ($$010$i|0) {
       case 46:  {
        $gotdig$0$i12$lcssa275 = $gotdig$0$i12;
        label = 138;
        break L184;
        break;
       }
       case 48:  {
        break;
       }
       default: {
        $$2$i = $$010$i;$699 = 0;$700 = 0;$gotdig$2$i13 = $gotdig$0$i12;$gotrad$0$i14 = 0;
        break L184;
       }
       }
       $331 = HEAP32[$0>>2]|0;
       $332 = HEAP32[$1>>2]|0;
       $333 = ($331>>>0)<($332>>>0);
       if ($333) {
        $334 = ((($331)) + 1|0);
        HEAP32[$0>>2] = $334;
        $335 = HEAP8[$331>>0]|0;
        $336 = $335&255;
        $$010$i = $336;$gotdig$0$i12 = 1;
        continue;
       } else {
        $337 = (___shgetc($f)|0);
        $$010$i = $337;$gotdig$0$i12 = 1;
        continue;
       }
      }
      if ((label|0) == 138) {
       $338 = HEAP32[$0>>2]|0;
       $339 = HEAP32[$1>>2]|0;
       $340 = ($338>>>0)<($339>>>0);
       if ($340) {
        $341 = ((($338)) + 1|0);
        HEAP32[$0>>2] = $341;
        $342 = HEAP8[$338>>0]|0;
        $343 = $342&255;
        $$111$ph$i = $343;
       } else {
        $344 = (___shgetc($f)|0);
        $$111$ph$i = $344;
       }
       $345 = ($$111$ph$i|0)==(48);
       if ($345) {
        $346 = 0;$347 = 0;
        while(1) {
         $348 = (_i64Add(($346|0),($347|0),-1,-1)|0);
         $349 = tempRet0;
         $350 = HEAP32[$0>>2]|0;
         $351 = HEAP32[$1>>2]|0;
         $352 = ($350>>>0)<($351>>>0);
         if ($352) {
          $353 = ((($350)) + 1|0);
          HEAP32[$0>>2] = $353;
          $354 = HEAP8[$350>>0]|0;
          $355 = $354&255;
          $$111$be$i = $355;
         } else {
          $356 = (___shgetc($f)|0);
          $$111$be$i = $356;
         }
         $357 = ($$111$be$i|0)==(48);
         if ($357) {
          $346 = $348;$347 = $349;
         } else {
          $$2$i = $$111$be$i;$699 = $348;$700 = $349;$gotdig$2$i13 = 1;$gotrad$0$i14 = 1;
          break;
         }
        }
       } else {
        $$2$i = $$111$ph$i;$699 = 0;$700 = 0;$gotdig$2$i13 = $gotdig$0$i12$lcssa275;$gotrad$0$i14 = 1;
       }
      }
      HEAP32[$x$i>>2] = 0;
      $358 = (($$2$i) + -48)|0;
      $359 = ($358>>>0)<(10);
      $360 = ($$2$i|0)==(46);
      $361 = $360 | $359;
      L203: do {
       if ($361) {
        $362 = ((($x$i)) + 496|0);
        $$3112$i = $$2$i;$365 = 0;$366 = 0;$701 = $360;$702 = $358;$703 = $699;$704 = $700;$gotdig$3108$i = $gotdig$2$i13;$gotrad$1109$i = $gotrad$0$i14;$j$0111$i = 0;$k$0110$i = 0;$lnz$0107$i = 0;
        L205: while(1) {
         do {
          if ($701) {
           $cond$i = ($gotrad$1109$i|0)==(0);
           if ($cond$i) {
            $705 = $365;$706 = $366;$707 = $365;$708 = $366;$gotdig$4$i = $gotdig$3108$i;$gotrad$2$i = 1;$j$2$i = $j$0111$i;$k$2$i = $k$0110$i;$lnz$2$i = $lnz$0107$i;
           } else {
            $709 = $703;$710 = $704;$711 = $365;$712 = $366;$gotdig$3108$i$lcssa = $gotdig$3108$i;$j$0111$i$lcssa = $j$0111$i;$k$0110$i$lcssa = $k$0110$i;$lnz$0107$i$lcssa = $lnz$0107$i;
            break L205;
           }
          } else {
           $364 = ($k$0110$i|0)<(125);
           $367 = (_i64Add(($365|0),($366|0),1,0)|0);
           $368 = tempRet0;
           $369 = ($$3112$i|0)!=(48);
           if (!($364)) {
            if (!($369)) {
             $705 = $703;$706 = $704;$707 = $367;$708 = $368;$gotdig$4$i = $gotdig$3108$i;$gotrad$2$i = $gotrad$1109$i;$j$2$i = $j$0111$i;$k$2$i = $k$0110$i;$lnz$2$i = $lnz$0107$i;
             break;
            }
            $379 = HEAP32[$362>>2]|0;
            $380 = $379 | 1;
            HEAP32[$362>>2] = $380;
            $705 = $703;$706 = $704;$707 = $367;$708 = $368;$gotdig$4$i = $gotdig$3108$i;$gotrad$2$i = $gotrad$1109$i;$j$2$i = $j$0111$i;$k$2$i = $k$0110$i;$lnz$2$i = $lnz$0107$i;
            break;
           }
           $$lnz$0$i = $369 ? $367 : $lnz$0107$i;
           $370 = ($j$0111$i|0)==(0);
           $371 = (($x$i) + ($k$0110$i<<2)|0);
           if ($370) {
            $storemerge$i = $702;
           } else {
            $372 = HEAP32[$371>>2]|0;
            $373 = ($372*10)|0;
            $374 = (($$3112$i) + -48)|0;
            $375 = (($374) + ($373))|0;
            $storemerge$i = $375;
           }
           HEAP32[$371>>2] = $storemerge$i;
           $376 = (($j$0111$i) + 1)|0;
           $377 = ($376|0)==(9);
           $378 = $377&1;
           $$k$0$i = (($378) + ($k$0110$i))|0;
           $$16$i = $377 ? 0 : $376;
           $705 = $703;$706 = $704;$707 = $367;$708 = $368;$gotdig$4$i = 1;$gotrad$2$i = $gotrad$1109$i;$j$2$i = $$16$i;$k$2$i = $$k$0$i;$lnz$2$i = $$lnz$0$i;
          }
         } while(0);
         $381 = HEAP32[$0>>2]|0;
         $382 = HEAP32[$1>>2]|0;
         $383 = ($381>>>0)<($382>>>0);
         if ($383) {
          $384 = ((($381)) + 1|0);
          HEAP32[$0>>2] = $384;
          $385 = HEAP8[$381>>0]|0;
          $386 = $385&255;
          $$3$be$i = $386;
         } else {
          $387 = (___shgetc($f)|0);
          $$3$be$i = $387;
         }
         $388 = (($$3$be$i) + -48)|0;
         $389 = ($388>>>0)<(10);
         $390 = ($$3$be$i|0)==(46);
         $391 = $390 | $389;
         if ($391) {
          $$3112$i = $$3$be$i;$365 = $707;$366 = $708;$701 = $390;$702 = $388;$703 = $705;$704 = $706;$gotdig$3108$i = $gotdig$4$i;$gotrad$1109$i = $gotrad$2$i;$j$0111$i = $j$2$i;$k$0110$i = $k$2$i;$lnz$0107$i = $lnz$2$i;
         } else {
          $$3$lcssa$i = $$3$be$i;$394 = $707;$395 = $705;$397 = $708;$398 = $706;$gotdig$3$lcssa$i = $gotdig$4$i;$gotrad$1$lcssa$i = $gotrad$2$i;$j$0$lcssa$i = $j$2$i;$k$0$lcssa$i = $k$2$i;$lnz$0$lcssa$i = $lnz$2$i;
          label = 161;
          break L203;
         }
        }
        $363 = ($gotdig$3108$i$lcssa|0)!=(0);
        $713 = $711;$714 = $712;$715 = $709;$716 = $710;$717 = $363;$j$077$i = $j$0111$i$lcssa;$k$073$i = $k$0110$i$lcssa;$lnz$067$i = $lnz$0107$i$lcssa;
        label = 169;
       } else {
        $$3$lcssa$i = $$2$i;$394 = 0;$395 = $699;$397 = 0;$398 = $700;$gotdig$3$lcssa$i = $gotdig$2$i13;$gotrad$1$lcssa$i = $gotrad$0$i14;$j$0$lcssa$i = 0;$k$0$lcssa$i = 0;$lnz$0$lcssa$i = 0;
        label = 161;
       }
      } while(0);
      do {
       if ((label|0) == 161) {
        $392 = ($gotrad$1$lcssa$i|0)==(0);
        $393 = $392 ? $394 : $395;
        $396 = $392 ? $397 : $398;
        $399 = ($gotdig$3$lcssa$i|0)!=(0);
        $400 = $$3$lcssa$i | 32;
        $401 = ($400|0)==(101);
        $or$cond18$i = $401 & $399;
        if (!($or$cond18$i)) {
         $416 = ($$3$lcssa$i|0)>(-1);
         if ($416) {
          $713 = $394;$714 = $397;$715 = $393;$716 = $396;$717 = $399;$j$077$i = $j$0$lcssa$i;$k$073$i = $k$0$lcssa$i;$lnz$067$i = $lnz$0$lcssa$i;
          label = 169;
          break;
         } else {
          $718 = $394;$719 = $397;$720 = $399;$721 = $393;$722 = $396;$j$076$i = $j$0$lcssa$i;$k$072$i = $k$0$lcssa$i;$lnz$066$i = $lnz$0$lcssa$i;
          label = 171;
          break;
         }
        }
        $402 = (_scanexp($f,$pok)|0);
        $403 = tempRet0;
        $404 = ($402|0)==(0);
        $405 = ($403|0)==(-2147483648);
        $406 = $404 & $405;
        if ($406) {
         $407 = ($pok|0)==(0);
         if ($407) {
          ___shlim($f,0);
          $$1$i = 0.0;
          break;
         }
         $408 = HEAP32[$1>>2]|0;
         $409 = ($408|0)==(0|0);
         if ($409) {
          $412 = 0;$413 = 0;
         } else {
          $410 = HEAP32[$0>>2]|0;
          $411 = ((($410)) + -1|0);
          HEAP32[$0>>2] = $411;
          $412 = 0;$413 = 0;
         }
        } else {
         $412 = $402;$413 = $403;
        }
        $414 = (_i64Add(($412|0),($413|0),($393|0),($396|0))|0);
        $415 = tempRet0;
        $427 = $414;$428 = $394;$430 = $415;$431 = $397;$j$075$i = $j$0$lcssa$i;$k$071$i = $k$0$lcssa$i;$lnz$065$i = $lnz$0$lcssa$i;
        label = 173;
       }
      } while(0);
      if ((label|0) == 169) {
       $417 = HEAP32[$1>>2]|0;
       $418 = ($417|0)==(0|0);
       if ($418) {
        $718 = $713;$719 = $714;$720 = $717;$721 = $715;$722 = $716;$j$076$i = $j$077$i;$k$072$i = $k$073$i;$lnz$066$i = $lnz$067$i;
        label = 171;
       } else {
        $419 = HEAP32[$0>>2]|0;
        $420 = ((($419)) + -1|0);
        HEAP32[$0>>2] = $420;
        if ($717) {
         $427 = $715;$428 = $713;$430 = $716;$431 = $714;$j$075$i = $j$077$i;$k$071$i = $k$073$i;$lnz$065$i = $lnz$067$i;
         label = 173;
        } else {
         label = 172;
        }
       }
      }
      if ((label|0) == 171) {
       if ($720) {
        $427 = $721;$428 = $718;$430 = $722;$431 = $719;$j$075$i = $j$076$i;$k$071$i = $k$072$i;$lnz$065$i = $lnz$066$i;
        label = 173;
       } else {
        label = 172;
       }
      }
      do {
       if ((label|0) == 172) {
        $421 = (___errno_location()|0);
        HEAP32[$421>>2] = 22;
        ___shlim($f,0);
        $$1$i = 0.0;
       }
       else if ((label|0) == 173) {
        $422 = HEAP32[$x$i>>2]|0;
        $423 = ($422|0)==(0);
        if ($423) {
         $424 = (+($sign$0|0));
         $425 = $424 * 0.0;
         $$1$i = $425;
         break;
        }
        $426 = ($427|0)==($428|0);
        $429 = ($430|0)==($431|0);
        $432 = $426 & $429;
        $433 = ($431|0)<(0);
        $434 = ($428>>>0)<(10);
        $435 = ($431|0)==(0);
        $436 = $435 & $434;
        $437 = $433 | $436;
        $or$cond$i16 = $437 & $432;
        if ($or$cond$i16) {
         $438 = ($bits$0$ph>>>0)>(30);
         $439 = $422 >>> $bits$0$ph;
         $440 = ($439|0)==(0);
         $or$cond20$i = $438 | $440;
         if ($or$cond20$i) {
          $441 = (+($sign$0|0));
          $442 = (+($422>>>0));
          $443 = $441 * $442;
          $$1$i = $443;
          break;
         }
        }
        $444 = (($emin$0$ph|0) / -2)&-1;
        $445 = ($444|0)<(0);
        $446 = $445 << 31 >> 31;
        $447 = ($430|0)>($446|0);
        $448 = ($427>>>0)>($444>>>0);
        $449 = ($430|0)==($446|0);
        $450 = $449 & $448;
        $451 = $447 | $450;
        if ($451) {
         $452 = (___errno_location()|0);
         HEAP32[$452>>2] = 34;
         $453 = (+($sign$0|0));
         $454 = $453 * 1.7976931348623157E+308;
         $455 = $454 * 1.7976931348623157E+308;
         $$1$i = $455;
         break;
        }
        $456 = (($emin$0$ph) + -106)|0;
        $457 = ($456|0)<(0);
        $458 = $457 << 31 >> 31;
        $459 = ($430|0)<($458|0);
        $460 = ($427>>>0)<($456>>>0);
        $461 = ($430|0)==($458|0);
        $462 = $461 & $460;
        $463 = $459 | $462;
        if ($463) {
         $464 = (___errno_location()|0);
         HEAP32[$464>>2] = 34;
         $465 = (+($sign$0|0));
         $466 = $465 * 2.2250738585072014E-308;
         $467 = $466 * 2.2250738585072014E-308;
         $$1$i = $467;
         break;
        }
        $468 = ($j$075$i|0)==(0);
        if ($468) {
         $k$3$i = $k$071$i;
        } else {
         $469 = ($j$075$i|0)<(9);
         if ($469) {
          $470 = (($x$i) + ($k$071$i<<2)|0);
          $$promoted$i = HEAP32[$470>>2]|0;
          $472 = $$promoted$i;$j$3102$i = $j$075$i;
          while(1) {
           $471 = ($472*10)|0;
           $473 = (($j$3102$i) + 1)|0;
           $exitcond151$i = ($473|0)==(9);
           if ($exitcond151$i) {
            $$lcssa267 = $471;
            break;
           } else {
            $472 = $471;$j$3102$i = $473;
           }
          }
          HEAP32[$470>>2] = $$lcssa267;
         }
         $474 = (($k$071$i) + 1)|0;
         $k$3$i = $474;
        }
        $475 = ($lnz$065$i|0)<(9);
        if ($475) {
         $476 = ($lnz$065$i|0)<=($427|0);
         $477 = ($427|0)<(18);
         $or$cond3$i = $476 & $477;
         if ($or$cond3$i) {
          $478 = ($427|0)==(9);
          if ($478) {
           $479 = (+($sign$0|0));
           $480 = HEAP32[$x$i>>2]|0;
           $481 = (+($480>>>0));
           $482 = $479 * $481;
           $$1$i = $482;
           break;
          }
          $483 = ($427|0)<(9);
          if ($483) {
           $484 = (+($sign$0|0));
           $485 = HEAP32[$x$i>>2]|0;
           $486 = (+($485>>>0));
           $487 = $484 * $486;
           $488 = (8 - ($427))|0;
           $489 = (153724 + ($488<<2)|0);
           $490 = HEAP32[$489>>2]|0;
           $491 = (+($490|0));
           $492 = $487 / $491;
           $$1$i = $492;
           break;
          }
          $$neg$i = Math_imul($427, -3)|0;
          $$neg40$i = (($bits$0$ph) + 27)|0;
          $493 = (($$neg40$i) + ($$neg$i))|0;
          $494 = ($493|0)>(30);
          $$pre$i17 = HEAP32[$x$i>>2]|0;
          $495 = $$pre$i17 >>> $493;
          $496 = ($495|0)==(0);
          $or$cond192$i = $494 | $496;
          if ($or$cond192$i) {
           $497 = (+($sign$0|0));
           $498 = (+($$pre$i17>>>0));
           $499 = $497 * $498;
           $500 = (($427) + -10)|0;
           $501 = (153724 + ($500<<2)|0);
           $502 = HEAP32[$501>>2]|0;
           $503 = (+($502|0));
           $504 = $499 * $503;
           $$1$i = $504;
           break;
          }
         }
        }
        $505 = (($427|0) % 9)&-1;
        $506 = ($505|0)==(0);
        if ($506) {
         $a$2$ph46$i = 0;$e2$0$ph$i = 0;$rp$2$ph44$i = $427;$z$1$ph45$i = $k$3$i;
        } else {
         $507 = ($427|0)>(-1);
         $508 = (($505) + 9)|0;
         $509 = $507 ? $505 : $508;
         $510 = (8 - ($509))|0;
         $511 = (153724 + ($510<<2)|0);
         $512 = HEAP32[$511>>2]|0;
         $513 = ($k$3$i|0)==(0);
         if ($513) {
          $a$0$lcssa161$i = 0;$rp$0$lcssa162$i = $427;$z$0$i = 0;
         } else {
          $514 = (1000000000 / ($512|0))&-1;
          $a$093$i = 0;$carry$095$i = 0;$k$494$i = 0;$rp$092$i = $427;
          while(1) {
           $515 = (($x$i) + ($k$494$i<<2)|0);
           $516 = HEAP32[$515>>2]|0;
           $517 = (($516>>>0) % ($512>>>0))&-1;
           $518 = (($516>>>0) / ($512>>>0))&-1;
           $519 = (($518) + ($carry$095$i))|0;
           HEAP32[$515>>2] = $519;
           $520 = Math_imul($517, $514)|0;
           $521 = ($k$494$i|0)==($a$093$i|0);
           $522 = ($519|0)==(0);
           $or$cond21$i = $521 & $522;
           $523 = (($k$494$i) + 1)|0;
           $524 = $523 & 127;
           $525 = (($rp$092$i) + -9)|0;
           $rp$1$i18 = $or$cond21$i ? $525 : $rp$092$i;
           $a$1$i = $or$cond21$i ? $524 : $a$093$i;
           $526 = ($523|0)==($k$3$i|0);
           if ($526) {
            $$lcssa266 = $520;$a$1$i$lcssa = $a$1$i;$rp$1$i18$lcssa = $rp$1$i18;
            break;
           } else {
            $a$093$i = $a$1$i;$carry$095$i = $520;$k$494$i = $523;$rp$092$i = $rp$1$i18;
           }
          }
          $527 = ($$lcssa266|0)==(0);
          if ($527) {
           $a$0$lcssa161$i = $a$1$i$lcssa;$rp$0$lcssa162$i = $rp$1$i18$lcssa;$z$0$i = $k$3$i;
          } else {
           $528 = (($k$3$i) + 1)|0;
           $529 = (($x$i) + ($k$3$i<<2)|0);
           HEAP32[$529>>2] = $$lcssa266;
           $a$0$lcssa161$i = $a$1$i$lcssa;$rp$0$lcssa162$i = $rp$1$i18$lcssa;$z$0$i = $528;
          }
         }
         $530 = (9 - ($509))|0;
         $531 = (($530) + ($rp$0$lcssa162$i))|0;
         $a$2$ph46$i = $a$0$lcssa161$i;$e2$0$ph$i = 0;$rp$2$ph44$i = $531;$z$1$ph45$i = $z$0$i;
        }
        L284: while(1) {
         $532 = ($rp$2$ph44$i|0)<(18);
         $533 = ($rp$2$ph44$i|0)==(18);
         $534 = (($x$i) + ($a$2$ph46$i<<2)|0);
         $e2$0$i19 = $e2$0$ph$i;$z$1$i = $z$1$ph45$i;
         while(1) {
          if (!($532)) {
           if (!($533)) {
            $a$4$ph$i = $a$2$ph46$i;$e2$1$ph$i = $e2$0$i19;$rp$4$ph42$i = $rp$2$ph44$i;$z$6$ph$i = $z$1$i;
            break L284;
           }
           $535 = HEAP32[$534>>2]|0;
           $536 = ($535>>>0)<(9007199);
           if (!($536)) {
            $a$4$ph$i = $a$2$ph46$i;$e2$1$ph$i = $e2$0$i19;$rp$4$ph42$i = 18;$z$6$ph$i = $z$1$i;
            break L284;
           }
          }
          $537 = (($z$1$i) + 127)|0;
          $carry1$0$i = 0;$k$5$in$i = $537;$z$2$i = $z$1$i;
          while(1) {
           $k$5$i = $k$5$in$i & 127;
           $538 = (($x$i) + ($k$5$i<<2)|0);
           $539 = HEAP32[$538>>2]|0;
           $540 = (_bitshift64Shl(($539|0),0,29)|0);
           $541 = tempRet0;
           $542 = (_i64Add(($540|0),($541|0),($carry1$0$i|0),0)|0);
           $543 = tempRet0;
           $544 = ($543>>>0)>(0);
           $545 = ($542>>>0)>(1000000000);
           $546 = ($543|0)==(0);
           $547 = $546 & $545;
           $548 = $544 | $547;
           if ($548) {
            $549 = (___udivdi3(($542|0),($543|0),1000000000,0)|0);
            $550 = tempRet0;
            $551 = (___uremdi3(($542|0),($543|0),1000000000,0)|0);
            $552 = tempRet0;
            $$sink$off0$i = $551;$carry1$1$i = $549;
           } else {
            $$sink$off0$i = $542;$carry1$1$i = 0;
           }
           HEAP32[$538>>2] = $$sink$off0$i;
           $553 = (($z$2$i) + 127)|0;
           $554 = $553 & 127;
           $555 = ($k$5$i|0)!=($554|0);
           $556 = ($k$5$i|0)==($a$2$ph46$i|0);
           $or$cond22$i = $555 | $556;
           $557 = ($$sink$off0$i|0)==(0);
           $k$5$z$2$i = $557 ? $k$5$i : $z$2$i;
           $z$3$i = $or$cond22$i ? $z$2$i : $k$5$z$2$i;
           $558 = (($k$5$i) + -1)|0;
           if ($556) {
            $carry1$1$i$lcssa = $carry1$1$i;$z$3$i$lcssa = $z$3$i;
            break;
           } else {
            $carry1$0$i = $carry1$1$i;$k$5$in$i = $558;$z$2$i = $z$3$i;
           }
          }
          $559 = (($e2$0$i19) + -29)|0;
          $560 = ($carry1$1$i$lcssa|0)==(0);
          if ($560) {
           $e2$0$i19 = $559;$z$1$i = $z$3$i$lcssa;
          } else {
           $$lcssa265 = $559;$carry1$1$i$lcssa$lcssa = $carry1$1$i$lcssa;$z$3$i$lcssa$lcssa = $z$3$i$lcssa;
           break;
          }
         }
         $561 = (($rp$2$ph44$i) + 9)|0;
         $562 = (($a$2$ph46$i) + 127)|0;
         $563 = $562 & 127;
         $564 = ($563|0)==($z$3$i$lcssa$lcssa|0);
         if ($564) {
          $565 = (($z$3$i$lcssa$lcssa) + 127)|0;
          $566 = $565 & 127;
          $567 = (($x$i) + ($566<<2)|0);
          $568 = HEAP32[$567>>2]|0;
          $569 = (($z$3$i$lcssa$lcssa) + 126)|0;
          $570 = $569 & 127;
          $571 = (($x$i) + ($570<<2)|0);
          $572 = HEAP32[$571>>2]|0;
          $573 = $572 | $568;
          HEAP32[$571>>2] = $573;
          $z$4$i = $566;
         } else {
          $z$4$i = $z$3$i$lcssa$lcssa;
         }
         $574 = (($x$i) + ($563<<2)|0);
         HEAP32[$574>>2] = $carry1$1$i$lcssa$lcssa;
         $a$2$ph46$i = $563;$e2$0$ph$i = $$lcssa265;$rp$2$ph44$i = $561;$z$1$ph45$i = $z$4$i;
        }
        L302: while(1) {
         $605 = (($z$6$ph$i) + 1)|0;
         $603 = $605 & 127;
         $606 = (($z$6$ph$i) + 127)|0;
         $607 = $606 & 127;
         $608 = (($x$i) + ($607<<2)|0);
         $a$4$ph167$i = $a$4$ph$i;$e2$1$ph166$i = $e2$1$ph$i;$rp$4$ph$i = $rp$4$ph42$i;
         while(1) {
          $609 = ($rp$4$ph$i|0)==(18);
          $610 = ($rp$4$ph$i|0)>(27);
          $$24$i = $610 ? 9 : 1;
          $$not$i = $609 ^ 1;
          $a$4$i = $a$4$ph167$i;$e2$1$i = $e2$1$ph166$i;
          while(1) {
           $575 = $a$4$i & 127;
           $576 = ($575|0)==($z$6$ph$i|0);
           do {
            if ($576) {
             label = 219;
            } else {
             $577 = (($x$i) + ($575<<2)|0);
             $578 = HEAP32[$577>>2]|0;
             $579 = ($578>>>0)<(9007199);
             if ($579) {
              label = 219;
              break;
             }
             $580 = ($578>>>0)>(9007199);
             if ($580) {
              break;
             }
             $581 = (($a$4$i) + 1)|0;
             $582 = $581 & 127;
             $583 = ($582|0)==($z$6$ph$i|0);
             if ($583) {
              label = 219;
              break;
             }
             $689 = (($x$i) + ($582<<2)|0);
             $690 = HEAP32[$689>>2]|0;
             $691 = ($690>>>0)<(254740991);
             if ($691) {
              label = 219;
              break;
             }
             $692 = ($690>>>0)>(254740991);
             $brmerge$i28 = $692 | $$not$i;
             if (!($brmerge$i28)) {
              $616 = $575;$a$4$i251 = $a$4$i;$e2$1$i248 = $e2$1$i;$z$10$i = $z$6$ph$i;
              break L302;
             }
            }
           } while(0);
           if ((label|0) == 219) {
            label = 0;
            if ($609) {
             label = 220;
             break L302;
            }
           }
           $584 = (($e2$1$i) + ($$24$i))|0;
           $585 = ($a$4$i|0)==($z$6$ph$i|0);
           if ($585) {
            $a$4$i = $z$6$ph$i;$e2$1$i = $584;
           } else {
            $$lcssa258 = $584;$a$4$i$lcssa250 = $a$4$i;
            break;
           }
          }
          $586 = 1 << $$24$i;
          $587 = (($586) + -1)|0;
          $588 = 1000000000 >>> $$24$i;
          $a$586$i = $a$4$i$lcssa250;$carry4$089$i = 0;$k$687$i = $a$4$i$lcssa250;$rp$585$i = $rp$4$ph$i;
          while(1) {
           $589 = (($x$i) + ($k$687$i<<2)|0);
           $590 = HEAP32[$589>>2]|0;
           $591 = $590 & $587;
           $592 = $590 >>> $$24$i;
           $593 = (($592) + ($carry4$089$i))|0;
           HEAP32[$589>>2] = $593;
           $594 = Math_imul($591, $588)|0;
           $595 = ($k$687$i|0)==($a$586$i|0);
           $596 = ($593|0)==(0);
           $or$cond25$i = $595 & $596;
           $597 = (($k$687$i) + 1)|0;
           $598 = $597 & 127;
           $599 = (($rp$585$i) + -9)|0;
           $rp$6$i = $or$cond25$i ? $599 : $rp$585$i;
           $a$6$i = $or$cond25$i ? $598 : $a$586$i;
           $600 = ($598|0)==($z$6$ph$i|0);
           if ($600) {
            $$lcssa259 = $594;$a$6$i$lcssa = $a$6$i;$rp$6$i$lcssa = $rp$6$i;
            break;
           } else {
            $a$586$i = $a$6$i;$carry4$089$i = $594;$k$687$i = $598;$rp$585$i = $rp$6$i;
           }
          }
          $601 = ($$lcssa259|0)==(0);
          if ($601) {
           $a$4$ph167$i = $a$6$i$lcssa;$e2$1$ph166$i = $$lcssa258;$rp$4$ph$i = $rp$6$i$lcssa;
           continue;
          }
          $602 = ($603|0)==($a$6$i$lcssa|0);
          if (!($602)) {
           $$lcssa258$lcssa = $$lcssa258;$$lcssa259$lcssa = $$lcssa259;$a$6$i$lcssa$lcssa = $a$6$i$lcssa;$rp$6$i$lcssa$lcssa = $rp$6$i$lcssa;
           break;
          }
          $611 = HEAP32[$608>>2]|0;
          $612 = $611 | 1;
          HEAP32[$608>>2] = $612;
          $a$4$ph167$i = $a$6$i$lcssa;$e2$1$ph166$i = $$lcssa258;$rp$4$ph$i = $rp$6$i$lcssa;
         }
         $604 = (($x$i) + ($z$6$ph$i<<2)|0);
         HEAP32[$604>>2] = $$lcssa259$lcssa;
         $a$4$ph$i = $a$6$i$lcssa$lcssa;$e2$1$ph$i = $$lcssa258$lcssa;$rp$4$ph42$i = $rp$6$i$lcssa$lcssa;$z$6$ph$i = $603;
        }
        if ((label|0) == 220) {
         if ($576) {
          $613 = (($603) + -1)|0;
          $614 = (($x$i) + ($613<<2)|0);
          HEAP32[$614>>2] = 0;
          $616 = $z$6$ph$i;$a$4$i251 = $a$4$i;$e2$1$i248 = $e2$1$i;$z$10$i = $603;
         } else {
          $616 = $575;$a$4$i251 = $a$4$i;$e2$1$i248 = $e2$1$i;$z$10$i = $z$6$ph$i;
         }
        }
        $615 = (($x$i) + ($616<<2)|0);
        $617 = HEAP32[$615>>2]|0;
        $618 = (+($617>>>0));
        $619 = (($a$4$i251) + 1)|0;
        $620 = $619 & 127;
        $621 = ($620|0)==($z$10$i|0);
        if ($621) {
         $678 = (($a$4$i251) + 2)|0;
         $679 = $678 & 127;
         $680 = (($679) + -1)|0;
         $681 = (($x$i) + ($680<<2)|0);
         HEAP32[$681>>2] = 0;
         $z$10$1$i = $679;
        } else {
         $z$10$1$i = $z$10$i;
        }
        $682 = $618 * 1.0E+9;
        $683 = (($x$i) + ($620<<2)|0);
        $684 = HEAP32[$683>>2]|0;
        $685 = (+($684>>>0));
        $686 = $682 + $685;
        $642 = (+($sign$0|0));
        $624 = $642 * $686;
        $662 = (($e2$1$i248) + 53)|0;
        $668 = (($662) - ($emin$0$ph))|0;
        $669 = ($668|0)<($bits$0$ph|0);
        $687 = ($668|0)<(0);
        $$$i = $687 ? 0 : $668;
        $denormal$0$i = $669&1;
        $$012$i = $669 ? $$$i : $bits$0$ph;
        $688 = ($$012$i|0)<(53);
        if ($688) {
         $622 = (105 - ($$012$i))|0;
         $623 = (+_scalbn(1.0,$622));
         $625 = (+_copysignl($623,$624));
         $626 = (53 - ($$012$i))|0;
         $627 = (+_scalbn(1.0,$626));
         $628 = (+_fmodl($624,$627));
         $629 = $624 - $628;
         $630 = $625 + $629;
         $bias$0$i25 = $625;$frac$0$i = $628;$y$1$i24 = $630;
        } else {
         $bias$0$i25 = 0.0;$frac$0$i = 0.0;$y$1$i24 = $624;
        }
        $631 = (($a$4$i251) + 2)|0;
        $632 = $631 & 127;
        $633 = ($632|0)==($z$10$1$i|0);
        do {
         if ($633) {
          $frac$3$i = $frac$0$i;
         } else {
          $634 = (($x$i) + ($632<<2)|0);
          $635 = HEAP32[$634>>2]|0;
          $636 = ($635>>>0)<(500000000);
          do {
           if ($636) {
            $637 = ($635|0)==(0);
            if ($637) {
             $638 = (($a$4$i251) + 3)|0;
             $639 = $638 & 127;
             $640 = ($639|0)==($z$10$1$i|0);
             if ($640) {
              $frac$1$i = $frac$0$i;
              break;
             }
            }
            $641 = $642 * 0.25;
            $643 = $641 + $frac$0$i;
            $frac$1$i = $643;
           } else {
            $644 = ($635>>>0)>(500000000);
            if ($644) {
             $645 = $642 * 0.75;
             $646 = $645 + $frac$0$i;
             $frac$1$i = $646;
             break;
            }
            $647 = (($a$4$i251) + 3)|0;
            $648 = $647 & 127;
            $649 = ($648|0)==($z$10$1$i|0);
            if ($649) {
             $650 = $642 * 0.5;
             $651 = $650 + $frac$0$i;
             $frac$1$i = $651;
             break;
            } else {
             $652 = $642 * 0.75;
             $653 = $652 + $frac$0$i;
             $frac$1$i = $653;
             break;
            }
           }
          } while(0);
          $654 = (53 - ($$012$i))|0;
          $655 = ($654|0)>(1);
          if (!($655)) {
           $frac$3$i = $frac$1$i;
           break;
          }
          $656 = (+_fmodl($frac$1$i,1.0));
          $657 = $656 != 0.0;
          if ($657) {
           $frac$3$i = $frac$1$i;
           break;
          }
          $658 = $frac$1$i + 1.0;
          $frac$3$i = $658;
         }
        } while(0);
        $659 = $y$1$i24 + $frac$3$i;
        $660 = $659 - $bias$0$i25;
        $661 = $662 & 2147483647;
        $663 = (-2 - ($sum$i))|0;
        $664 = ($661|0)>($663|0);
        do {
         if ($664) {
          $665 = (+Math_abs((+$660)));
          $666 = !($665 >= 9007199254740992.0);
          if ($666) {
           $denormal$2$i = $denormal$0$i;$e2$3$i = $e2$1$i248;$y$2$i26 = $660;
          } else {
           $667 = ($$012$i|0)==($668|0);
           $or$cond26$i = $669 & $667;
           $denormal$1$i = $or$cond26$i ? 0 : $denormal$0$i;
           $670 = $660 * 0.5;
           $671 = (($e2$1$i248) + 1)|0;
           $denormal$2$i = $denormal$1$i;$e2$3$i = $671;$y$2$i26 = $670;
          }
          $672 = (($e2$3$i) + 50)|0;
          $673 = ($672|0)>($330|0);
          if (!($673)) {
           $674 = ($denormal$2$i|0)!=(0);
           $675 = $frac$3$i != 0.0;
           $or$cond9$i27 = $675 & $674;
           if (!($or$cond9$i27)) {
            $e2$4$i = $e2$3$i;$y$3$i = $y$2$i26;
            break;
           }
          }
          $676 = (___errno_location()|0);
          HEAP32[$676>>2] = 34;
          $e2$4$i = $e2$3$i;$y$3$i = $y$2$i26;
         } else {
          $e2$4$i = $e2$1$i248;$y$3$i = $660;
         }
        } while(0);
        $677 = (+_scalbnl($y$3$i,$e2$4$i));
        $$1$i = $677;
       }
      } while(0);
      $$0 = $$1$i;
      break L4;
      break;
     }
     default: {
      $109 = HEAP32[$1>>2]|0;
      $110 = ($109|0)==(0|0);
      if (!($110)) {
       $111 = HEAP32[$0>>2]|0;
       $112 = ((($111)) + -1|0);
       HEAP32[$0>>2] = $112;
      }
      $113 = (___errno_location()|0);
      HEAP32[$113>>2] = 22;
      ___shlim($f,0);
      $$0 = 0.0;
      break L4;
     }
     }
    }
    }
   } while(0);
   if ((label|0) == 23) {
    $41 = HEAP32[$1>>2]|0;
    $42 = ($41|0)==(0|0);
    if (!($42)) {
     $43 = HEAP32[$0>>2]|0;
     $44 = ((($43)) + -1|0);
     HEAP32[$0>>2] = $44;
    }
    $45 = ($pok|0)!=(0);
    $46 = ($i$0$lcssa>>>0)>(3);
    $or$cond9 = $45 & $46;
    if ($or$cond9) {
     $i$1 = $i$0$lcssa;
     while(1) {
      if (!($42)) {
       $47 = HEAP32[$0>>2]|0;
       $48 = ((($47)) + -1|0);
       HEAP32[$0>>2] = $48;
      }
      $49 = (($i$1) + -1)|0;
      $$old8 = ($49>>>0)>(3);
      if ($$old8) {
       $i$1 = $49;
      } else {
       break;
      }
     }
    }
   }
   $50 = (+($sign$0|0));
   $51 = $50 * inf;
   $52 = $51;
   $$0 = $52;
  }
 } while(0);
 STACKTOP = sp;return (+$$0);
}
function _scanexp($f,$pok) {
 $f = $f|0;
 $pok = $pok|0;
 var $$lcssa22 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $c$0 = 0, $c$1$be = 0, $c$1$be$lcssa = 0, $c$112 = 0, $c$2$be = 0, $c$2$lcssa = 0, $c$27 = 0, $c$3$be = 0, $neg$0 = 0, $or$cond3 = 0, $x$013 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($f)) + 100|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1>>>0)<($3>>>0);
 if ($4) {
  $5 = ((($1)) + 1|0);
  HEAP32[$0>>2] = $5;
  $6 = HEAP8[$1>>0]|0;
  $7 = $6&255;
  $10 = $7;
 } else {
  $8 = (___shgetc($f)|0);
  $10 = $8;
 }
 $9 = ($10|0)==(45);
 switch ($10|0) {
 case 43: case 45:  {
  $11 = $9&1;
  $12 = HEAP32[$0>>2]|0;
  $13 = HEAP32[$2>>2]|0;
  $14 = ($12>>>0)<($13>>>0);
  if ($14) {
   $15 = ((($12)) + 1|0);
   HEAP32[$0>>2] = $15;
   $16 = HEAP8[$12>>0]|0;
   $17 = $16&255;
   $20 = $17;
  } else {
   $18 = (___shgetc($f)|0);
   $20 = $18;
  }
  $19 = (($20) + -48)|0;
  $21 = ($19>>>0)>(9);
  $22 = ($pok|0)!=(0);
  $or$cond3 = $22 & $21;
  if ($or$cond3) {
   $23 = HEAP32[$2>>2]|0;
   $24 = ($23|0)==(0|0);
   if ($24) {
    $c$0 = $20;$neg$0 = $11;
   } else {
    $25 = HEAP32[$0>>2]|0;
    $26 = ((($25)) + -1|0);
    HEAP32[$0>>2] = $26;
    $c$0 = $20;$neg$0 = $11;
   }
  } else {
   $c$0 = $20;$neg$0 = $11;
  }
  break;
 }
 default: {
  $c$0 = $10;$neg$0 = 0;
 }
 }
 $27 = (($c$0) + -48)|0;
 $28 = ($27>>>0)>(9);
 if ($28) {
  $29 = HEAP32[$2>>2]|0;
  $30 = ($29|0)==(0|0);
  if ($30) {
   $98 = -2147483648;$99 = 0;
  } else {
   $31 = HEAP32[$0>>2]|0;
   $32 = ((($31)) + -1|0);
   HEAP32[$0>>2] = $32;
   $98 = -2147483648;$99 = 0;
  }
 } else {
  $c$112 = $c$0;$x$013 = 0;
  while(1) {
   $33 = ($x$013*10)|0;
   $34 = (($c$112) + -48)|0;
   $35 = (($34) + ($33))|0;
   $36 = HEAP32[$0>>2]|0;
   $37 = HEAP32[$2>>2]|0;
   $38 = ($36>>>0)<($37>>>0);
   if ($38) {
    $39 = ((($36)) + 1|0);
    HEAP32[$0>>2] = $39;
    $40 = HEAP8[$36>>0]|0;
    $41 = $40&255;
    $c$1$be = $41;
   } else {
    $42 = (___shgetc($f)|0);
    $c$1$be = $42;
   }
   $43 = (($c$1$be) + -48)|0;
   $44 = ($43>>>0)<(10);
   $45 = ($35|0)<(214748364);
   $46 = $44 & $45;
   if ($46) {
    $c$112 = $c$1$be;$x$013 = $35;
   } else {
    $$lcssa22 = $35;$c$1$be$lcssa = $c$1$be;
    break;
   }
  }
  $47 = ($$lcssa22|0)<(0);
  $48 = $47 << 31 >> 31;
  $49 = (($c$1$be$lcssa) + -48)|0;
  $50 = ($49>>>0)<(10);
  if ($50) {
   $53 = $$lcssa22;$54 = $48;$c$27 = $c$1$be$lcssa;
   while(1) {
    $55 = (___muldi3(($53|0),($54|0),10,0)|0);
    $56 = tempRet0;
    $57 = ($c$27|0)<(0);
    $58 = $57 << 31 >> 31;
    $59 = (_i64Add(($c$27|0),($58|0),-48,-1)|0);
    $60 = tempRet0;
    $61 = (_i64Add(($59|0),($60|0),($55|0),($56|0))|0);
    $62 = tempRet0;
    $63 = HEAP32[$0>>2]|0;
    $64 = HEAP32[$2>>2]|0;
    $65 = ($63>>>0)<($64>>>0);
    if ($65) {
     $66 = ((($63)) + 1|0);
     HEAP32[$0>>2] = $66;
     $67 = HEAP8[$63>>0]|0;
     $68 = $67&255;
     $c$2$be = $68;
    } else {
     $69 = (___shgetc($f)|0);
     $c$2$be = $69;
    }
    $70 = (($c$2$be) + -48)|0;
    $71 = ($70>>>0)<(10);
    $72 = ($62|0)<(21474836);
    $73 = ($61>>>0)<(2061584302);
    $74 = ($62|0)==(21474836);
    $75 = $74 & $73;
    $76 = $72 | $75;
    $77 = $71 & $76;
    if ($77) {
     $53 = $61;$54 = $62;$c$27 = $c$2$be;
    } else {
     $92 = $61;$93 = $62;$c$2$lcssa = $c$2$be;
     break;
    }
   }
  } else {
   $92 = $$lcssa22;$93 = $48;$c$2$lcssa = $c$1$be$lcssa;
  }
  $51 = (($c$2$lcssa) + -48)|0;
  $52 = ($51>>>0)<(10);
  if ($52) {
   while(1) {
    $78 = HEAP32[$0>>2]|0;
    $79 = HEAP32[$2>>2]|0;
    $80 = ($78>>>0)<($79>>>0);
    if ($80) {
     $81 = ((($78)) + 1|0);
     HEAP32[$0>>2] = $81;
     $82 = HEAP8[$78>>0]|0;
     $83 = $82&255;
     $c$3$be = $83;
    } else {
     $84 = (___shgetc($f)|0);
     $c$3$be = $84;
    }
    $85 = (($c$3$be) + -48)|0;
    $86 = ($85>>>0)<(10);
    if (!($86)) {
     break;
    }
   }
  }
  $87 = HEAP32[$2>>2]|0;
  $88 = ($87|0)==(0|0);
  if (!($88)) {
   $89 = HEAP32[$0>>2]|0;
   $90 = ((($89)) + -1|0);
   HEAP32[$0>>2] = $90;
  }
  $91 = ($neg$0|0)!=(0);
  $94 = (_i64Subtract(0,0,($92|0),($93|0))|0);
  $95 = tempRet0;
  $96 = $91 ? $94 : $92;
  $97 = $91 ? $95 : $93;
  $98 = $97;$99 = $96;
 }
 tempRet0 = ($98);
 return ($99|0);
}
function _scalbn($x,$n) {
 $x = +$x;
 $n = $n|0;
 var $$ = 0, $$0 = 0, $$1 = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0.0, $18 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0.0, $9 = 0, $y$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)>(1023);
 if ($0) {
  $1 = $x * 8.9884656743115795E+307;
  $2 = (($n) + -1023)|0;
  $3 = ($2|0)>(1023);
  if ($3) {
   $4 = $1 * 8.9884656743115795E+307;
   $5 = (($n) + -2046)|0;
   $6 = ($5|0)>(1023);
   $$ = $6 ? 1023 : $5;
   $$0 = $$;$y$0 = $4;
  } else {
   $$0 = $2;$y$0 = $1;
  }
 } else {
  $7 = ($n|0)<(-1022);
  if ($7) {
   $8 = $x * 2.2250738585072014E-308;
   $9 = (($n) + 1022)|0;
   $10 = ($9|0)<(-1022);
   if ($10) {
    $11 = $8 * 2.2250738585072014E-308;
    $12 = (($n) + 2044)|0;
    $13 = ($12|0)<(-1022);
    $$1 = $13 ? -1022 : $12;
    $$0 = $$1;$y$0 = $11;
   } else {
    $$0 = $9;$y$0 = $8;
   }
  } else {
   $$0 = $n;$y$0 = $x;
  }
 }
 $14 = (($$0) + 1023)|0;
 $15 = (_bitshift64Shl(($14|0),0,52)|0);
 $16 = tempRet0;
 HEAP32[tempDoublePtr>>2] = $15;HEAP32[tempDoublePtr+4>>2] = $16;$17 = +HEAPF64[tempDoublePtr>>3];
 $18 = $y$0 * $17;
 return (+$18);
}
function _copysignl($x,$y) {
 $x = +$x;
 $y = +$y;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_copysign($x,$y));
 return (+$0);
}
function _scalbnl($x,$n) {
 $x = +$x;
 $n = $n|0;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_scalbn($x,$n));
 return (+$0);
}
function _fmodl($x,$y) {
 $x = +$x;
 $y = +$y;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_fmod($x,$y));
 return (+$0);
}
function _fmod($x,$y) {
 $x = +$x;
 $y = +$y;
 var $$0 = 0.0, $$lcssa7 = 0, $$x = 0.0, $0 = 0, $1 = 0, $10 = 0, $100 = 0.0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0.0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0.0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0.0, $25 = 0.0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0.0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $ex$0$lcssa = 0, $ex$026 = 0, $ex$1 = 0, $ex$2$lcssa = 0, $ex$212 = 0, $ex$3$lcssa = 0, $ex$39 = 0, $ey$0$lcssa = 0, $ey$020 = 0, $ey$1$ph = 0, $fabs = 0.0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAPF64[tempDoublePtr>>3] = $x;$0 = HEAP32[tempDoublePtr>>2]|0;
 $1 = HEAP32[tempDoublePtr+4>>2]|0;
 HEAPF64[tempDoublePtr>>3] = $y;$2 = HEAP32[tempDoublePtr>>2]|0;
 $3 = HEAP32[tempDoublePtr+4>>2]|0;
 $4 = (_bitshift64Lshr(($0|0),($1|0),52)|0);
 $5 = tempRet0;
 $6 = $4 & 2047;
 $7 = (_bitshift64Lshr(($2|0),($3|0),52)|0);
 $8 = tempRet0;
 $9 = $7 & 2047;
 $10 = $1 & -2147483648;
 $11 = (_bitshift64Shl(($2|0),($3|0),1)|0);
 $12 = tempRet0;
 $13 = ($11|0)==(0);
 $14 = ($12|0)==(0);
 $15 = $13 & $14;
 L1: do {
  if ($15) {
   label = 3;
  } else {
   $fabs = (+Math_abs((+$y)));
   HEAPF64[tempDoublePtr>>3] = $fabs;$16 = HEAP32[tempDoublePtr>>2]|0;
   $17 = HEAP32[tempDoublePtr+4>>2]|0;
   $18 = ($17>>>0)>(2146435072);
   $19 = ($16>>>0)>(0);
   $20 = ($17|0)==(2146435072);
   $21 = $20 & $19;
   $22 = $18 | $21;
   $23 = ($6|0)==(2047);
   $or$cond = $22 | $23;
   if ($or$cond) {
    label = 3;
   } else {
    $26 = (_bitshift64Shl(($0|0),($1|0),1)|0);
    $27 = tempRet0;
    $28 = ($27>>>0)>($12>>>0);
    $29 = ($26>>>0)>($11>>>0);
    $30 = ($27|0)==($12|0);
    $31 = $30 & $29;
    $32 = $28 | $31;
    if (!($32)) {
     $33 = ($26|0)==($11|0);
     $34 = ($27|0)==($12|0);
     $35 = $33 & $34;
     $36 = $x * 0.0;
     $$x = $35 ? $36 : $x;
     return (+$$x);
    }
    $37 = ($6|0)==(0);
    if ($37) {
     $38 = (_bitshift64Shl(($0|0),($1|0),12)|0);
     $39 = tempRet0;
     $40 = ($39|0)>(-1);
     $41 = ($38>>>0)>(4294967295);
     $42 = ($39|0)==(-1);
     $43 = $42 & $41;
     $44 = $40 | $43;
     if ($44) {
      $46 = $38;$47 = $39;$ex$026 = 0;
      while(1) {
       $45 = (($ex$026) + -1)|0;
       $48 = (_bitshift64Shl(($46|0),($47|0),1)|0);
       $49 = tempRet0;
       $50 = ($49|0)>(-1);
       $51 = ($48>>>0)>(4294967295);
       $52 = ($49|0)==(-1);
       $53 = $52 & $51;
       $54 = $50 | $53;
       if ($54) {
        $46 = $48;$47 = $49;$ex$026 = $45;
       } else {
        $ex$0$lcssa = $45;
        break;
       }
      }
     } else {
      $ex$0$lcssa = 0;
     }
     $55 = (1 - ($ex$0$lcssa))|0;
     $56 = (_bitshift64Shl(($0|0),($1|0),($55|0))|0);
     $57 = tempRet0;
     $84 = $56;$85 = $57;$ex$1 = $ex$0$lcssa;
    } else {
     $58 = $1 & 1048575;
     $59 = $58 | 1048576;
     $84 = $0;$85 = $59;$ex$1 = $6;
    }
    $60 = ($9|0)==(0);
    if ($60) {
     $61 = (_bitshift64Shl(($2|0),($3|0),12)|0);
     $62 = tempRet0;
     $63 = ($62|0)>(-1);
     $64 = ($61>>>0)>(4294967295);
     $65 = ($62|0)==(-1);
     $66 = $65 & $64;
     $67 = $63 | $66;
     if ($67) {
      $69 = $61;$70 = $62;$ey$020 = 0;
      while(1) {
       $68 = (($ey$020) + -1)|0;
       $71 = (_bitshift64Shl(($69|0),($70|0),1)|0);
       $72 = tempRet0;
       $73 = ($72|0)>(-1);
       $74 = ($71>>>0)>(4294967295);
       $75 = ($72|0)==(-1);
       $76 = $75 & $74;
       $77 = $73 | $76;
       if ($77) {
        $69 = $71;$70 = $72;$ey$020 = $68;
       } else {
        $ey$0$lcssa = $68;
        break;
       }
      }
     } else {
      $ey$0$lcssa = 0;
     }
     $78 = (1 - ($ey$0$lcssa))|0;
     $79 = (_bitshift64Shl(($2|0),($3|0),($78|0))|0);
     $80 = tempRet0;
     $86 = $79;$87 = $80;$ey$1$ph = $ey$0$lcssa;
    } else {
     $81 = $3 & 1048575;
     $82 = $81 | 1048576;
     $86 = $2;$87 = $82;$ey$1$ph = $9;
    }
    $83 = ($ex$1|0)>($ey$1$ph|0);
    $88 = (_i64Subtract(($84|0),($85|0),($86|0),($87|0))|0);
    $89 = tempRet0;
    $90 = ($89|0)>(-1);
    $91 = ($88>>>0)>(4294967295);
    $92 = ($89|0)==(-1);
    $93 = $92 & $91;
    $94 = $90 | $93;
    L23: do {
     if ($83) {
      $153 = $94;$154 = $88;$155 = $89;$96 = $84;$98 = $85;$ex$212 = $ex$1;
      while(1) {
       if ($153) {
        $95 = ($96|0)==($86|0);
        $97 = ($98|0)==($87|0);
        $99 = $95 & $97;
        if ($99) {
         break;
        } else {
         $101 = $154;$102 = $155;
        }
       } else {
        $101 = $96;$102 = $98;
       }
       $103 = (_bitshift64Shl(($101|0),($102|0),1)|0);
       $104 = tempRet0;
       $105 = (($ex$212) + -1)|0;
       $106 = ($105|0)>($ey$1$ph|0);
       $107 = (_i64Subtract(($103|0),($104|0),($86|0),($87|0))|0);
       $108 = tempRet0;
       $109 = ($108|0)>(-1);
       $110 = ($107>>>0)>(4294967295);
       $111 = ($108|0)==(-1);
       $112 = $111 & $110;
       $113 = $109 | $112;
       if ($106) {
        $153 = $113;$154 = $107;$155 = $108;$96 = $103;$98 = $104;$ex$212 = $105;
       } else {
        $$lcssa7 = $113;$115 = $103;$117 = $104;$156 = $107;$157 = $108;$ex$2$lcssa = $105;
        break L23;
       }
      }
      $100 = $x * 0.0;
      $$0 = $100;
      break L1;
     } else {
      $$lcssa7 = $94;$115 = $84;$117 = $85;$156 = $88;$157 = $89;$ex$2$lcssa = $ex$1;
     }
    } while(0);
    if ($$lcssa7) {
     $114 = ($115|0)==($86|0);
     $116 = ($117|0)==($87|0);
     $118 = $114 & $116;
     if ($118) {
      $126 = $x * 0.0;
      $$0 = $126;
      break;
     } else {
      $120 = $157;$122 = $156;
     }
    } else {
     $120 = $117;$122 = $115;
    }
    $119 = ($120>>>0)<(1048576);
    $121 = ($122>>>0)<(0);
    $123 = ($120|0)==(1048576);
    $124 = $123 & $121;
    $125 = $119 | $124;
    if ($125) {
     $127 = $122;$128 = $120;$ex$39 = $ex$2$lcssa;
     while(1) {
      $129 = (_bitshift64Shl(($127|0),($128|0),1)|0);
      $130 = tempRet0;
      $131 = (($ex$39) + -1)|0;
      $132 = ($130>>>0)<(1048576);
      $133 = ($129>>>0)<(0);
      $134 = ($130|0)==(1048576);
      $135 = $134 & $133;
      $136 = $132 | $135;
      if ($136) {
       $127 = $129;$128 = $130;$ex$39 = $131;
      } else {
       $138 = $129;$139 = $130;$ex$3$lcssa = $131;
       break;
      }
     }
    } else {
     $138 = $122;$139 = $120;$ex$3$lcssa = $ex$2$lcssa;
    }
    $137 = ($ex$3$lcssa|0)>(0);
    if ($137) {
     $140 = (_i64Add(($138|0),($139|0),0,-1048576)|0);
     $141 = tempRet0;
     $142 = (_bitshift64Shl(($ex$3$lcssa|0),0,52)|0);
     $143 = tempRet0;
     $144 = $140 | $142;
     $145 = $141 | $143;
     $150 = $145;$151 = $144;
    } else {
     $146 = (1 - ($ex$3$lcssa))|0;
     $147 = (_bitshift64Lshr(($138|0),($139|0),($146|0))|0);
     $148 = tempRet0;
     $150 = $148;$151 = $147;
    }
    $149 = $150 | $10;
    HEAP32[tempDoublePtr>>2] = $151;HEAP32[tempDoublePtr+4>>2] = $149;$152 = +HEAPF64[tempDoublePtr>>3];
    $$0 = $152;
   }
  }
 } while(0);
 if ((label|0) == 3) {
  $24 = $x * $y;
  $25 = $24 / $24;
  $$0 = $25;
 }
 return (+$$0);
}
function _catclose($catd) {
 $catd = $catd|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _catgets($catd,$set_id,$msg_id,$s) {
 $catd = $catd|0;
 $set_id = $set_id|0;
 $msg_id = $msg_id|0;
 $s = $s|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ($s|0);
}
function _catopen($name,$oflag) {
 $name = $name|0;
 $oflag = $oflag|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ((-1)|0);
}
function _freelocale($l) {
 $l = $l|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _free($l);
 return;
}
function _strlen($s) {
 $s = $s|0;
 var $$0 = 0, $$01$lcssa = 0, $$014 = 0, $$1$lcssa = 0, $$lcssa20 = 0, $$pn = 0, $$pn15 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $w$0 = 0, $w$0$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $s;
 $1 = $0 & 3;
 $2 = ($1|0)==(0);
 L1: do {
  if ($2) {
   $$01$lcssa = $s;
   label = 4;
  } else {
   $$014 = $s;$21 = $0;
   while(1) {
    $3 = HEAP8[$$014>>0]|0;
    $4 = ($3<<24>>24)==(0);
    if ($4) {
     $$pn = $21;
     break L1;
    }
    $5 = ((($$014)) + 1|0);
    $6 = $5;
    $7 = $6 & 3;
    $8 = ($7|0)==(0);
    if ($8) {
     $$01$lcssa = $5;
     label = 4;
     break;
    } else {
     $$014 = $5;$21 = $6;
    }
   }
  }
 } while(0);
 if ((label|0) == 4) {
  $w$0 = $$01$lcssa;
  while(1) {
   $9 = HEAP32[$w$0>>2]|0;
   $10 = (($9) + -16843009)|0;
   $11 = $9 & -2139062144;
   $12 = $11 ^ -2139062144;
   $13 = $12 & $10;
   $14 = ($13|0)==(0);
   $15 = ((($w$0)) + 4|0);
   if ($14) {
    $w$0 = $15;
   } else {
    $$lcssa20 = $9;$w$0$lcssa = $w$0;
    break;
   }
  }
  $16 = $$lcssa20&255;
  $17 = ($16<<24>>24)==(0);
  if ($17) {
   $$1$lcssa = $w$0$lcssa;
  } else {
   $$pn15 = $w$0$lcssa;
   while(1) {
    $18 = ((($$pn15)) + 1|0);
    $$pre = HEAP8[$18>>0]|0;
    $19 = ($$pre<<24>>24)==(0);
    if ($19) {
     $$1$lcssa = $18;
     break;
    } else {
     $$pn15 = $18;
    }
   }
  }
  $20 = $$1$lcssa;
  $$pn = $20;
 }
 $$0 = (($$pn) - ($0))|0;
 return ($$0|0);
}
function _mbrtowc($wc,$src,$n,$st) {
 $wc = $wc|0;
 $src = $src|0;
 $n = $n|0;
 $st = $st|0;
 var $$0 = 0, $$024 = 0, $$1 = 0, $$lcssa = 0, $$lcssa35 = 0, $$st = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $c$05 = 0, $c$1 = 0, $c$2 = 0, $dummy = 0, $dummy$wc = 0, $s$06 = 0, $s$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $dummy = sp;
 $0 = ($st|0)==(0|0);
 $$st = $0 ? 1206004 : $st;
 $1 = HEAP32[$$st>>2]|0;
 $2 = ($src|0)==(0|0);
 L1: do {
  if ($2) {
   $3 = ($1|0)==(0);
   if ($3) {
    $$0 = 0;
   } else {
    label = 15;
   }
  } else {
   $4 = ($wc|0)==(0|0);
   $dummy$wc = $4 ? $dummy : $wc;
   $5 = ($n|0)==(0);
   if ($5) {
    $$0 = -2;
   } else {
    $6 = ($1|0)==(0);
    if ($6) {
     $7 = HEAP8[$src>>0]|0;
     $8 = $7&255;
     $9 = ($7<<24>>24)>(-1);
     if ($9) {
      HEAP32[$dummy$wc>>2] = $8;
      $10 = ($7<<24>>24)!=(0);
      $11 = $10&1;
      $$0 = $11;
      break;
     }
     $12 = (($8) + -194)|0;
     $13 = ($12>>>0)>(50);
     if ($13) {
      label = 15;
      break;
     }
     $14 = ((($src)) + 1|0);
     $15 = (149972 + ($12<<2)|0);
     $16 = HEAP32[$15>>2]|0;
     $17 = (($n) + -1)|0;
     $18 = ($17|0)==(0);
     if ($18) {
      $c$2 = $16;
     } else {
      $$024 = $17;$c$05 = $16;$s$06 = $14;
      label = 9;
     }
    } else {
     $$024 = $n;$c$05 = $1;$s$06 = $src;
     label = 9;
    }
    L11: do {
     if ((label|0) == 9) {
      $19 = HEAP8[$s$06>>0]|0;
      $20 = $19&255;
      $21 = $20 >>> 3;
      $22 = (($21) + -16)|0;
      $23 = $c$05 >> 26;
      $24 = (($21) + ($23))|0;
      $25 = $22 | $24;
      $26 = ($25>>>0)>(7);
      if ($26) {
       label = 15;
       break L1;
      } else {
       $$1 = $$024;$30 = $19;$c$1 = $c$05;$s$1 = $s$06;
      }
      while(1) {
       $27 = $c$1 << 6;
       $28 = ((($s$1)) + 1|0);
       $29 = $30&255;
       $31 = (($29) + -128)|0;
       $32 = $31 | $27;
       $33 = (($$1) + -1)|0;
       $34 = ($32|0)<(0);
       if (!($34)) {
        $$lcssa = $32;$$lcssa35 = $33;
        break;
       }
       $36 = ($33|0)==(0);
       if ($36) {
        $c$2 = $32;
        break L11;
       }
       $37 = HEAP8[$28>>0]|0;
       $38 = $37 & -64;
       $39 = ($38<<24>>24)==(-128);
       if ($39) {
        $$1 = $33;$30 = $37;$c$1 = $32;$s$1 = $28;
       } else {
        label = 15;
        break L1;
       }
      }
      HEAP32[$$st>>2] = 0;
      HEAP32[$dummy$wc>>2] = $$lcssa;
      $35 = (($n) - ($$lcssa35))|0;
      $$0 = $35;
      break L1;
     }
    } while(0);
    HEAP32[$$st>>2] = $c$2;
    $$0 = -2;
   }
  }
 } while(0);
 if ((label|0) == 15) {
  HEAP32[$$st>>2] = 0;
  $40 = (___errno_location()|0);
  HEAP32[$40>>2] = 84;
  $$0 = -1;
 }
 STACKTOP = sp;return ($$0|0);
}
function _strcasecmp($_l,$_r) {
 $_l = $_l|0;
 $_r = $_r|0;
 var $$pre$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $l$03 = 0, $l$03$lcssa20 = 0, $r$0$lcssa = 0, $r$04 = 0, $r$04$lcssa19 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$_l>>0]|0;
 $1 = ($0<<24>>24)==(0);
 L1: do {
  if ($1) {
   $19 = 0;$r$0$lcssa = $_r;
  } else {
   $2 = $0&255;
   $6 = $0;$7 = $2;$l$03 = $_l;$r$04 = $_r;
   while(1) {
    $3 = HEAP8[$r$04>>0]|0;
    $4 = ($3<<24>>24)==(0);
    if ($4) {
     $19 = $6;$r$0$lcssa = $r$04;
     break L1;
    }
    $5 = ($6<<24>>24)==($3<<24>>24);
    if (!($5)) {
     $8 = (_tolower($7)|0);
     $9 = HEAP8[$r$04>>0]|0;
     $10 = $9&255;
     $11 = (_tolower($10)|0);
     $12 = ($8|0)==($11|0);
     if (!($12)) {
      $l$03$lcssa20 = $l$03;$r$04$lcssa19 = $r$04;
      break;
     }
    }
    $13 = ((($l$03)) + 1|0);
    $14 = ((($r$04)) + 1|0);
    $15 = HEAP8[$13>>0]|0;
    $16 = $15&255;
    $17 = ($15<<24>>24)==(0);
    if ($17) {
     $19 = 0;$r$0$lcssa = $14;
     break L1;
    } else {
     $6 = $15;$7 = $16;$l$03 = $13;$r$04 = $14;
    }
   }
   $$pre$pre = HEAP8[$l$03$lcssa20>>0]|0;
   $19 = $$pre$pre;$r$0$lcssa = $r$04$lcssa19;
  }
 } while(0);
 $18 = $19&255;
 $20 = (_tolower($18)|0);
 $21 = HEAP8[$r$0$lcssa>>0]|0;
 $22 = $21&255;
 $23 = (_tolower($22)|0);
 $24 = (($20) - ($23))|0;
 return ($24|0);
}
function _isdigit_l($c,$l) {
 $c = $c|0;
 $l = $l|0;
 var $0 = 0, $isdigit = 0, $isdigittmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $isdigittmp = (($c) + -48)|0;
 $isdigit = ($isdigittmp>>>0)<(10);
 $0 = $isdigit&1;
 return ($0|0);
}
function _isxdigit_l($c,$l) {
 $c = $c|0;
 $l = $l|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_isxdigit($c)|0);
 return ($0|0);
}
function _localeconv() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (153756|0);
}
function _newlocale($mask,$name,$base) {
 $mask = $mask|0;
 $name = $name|0;
 $base = $base|0;
 var $$01 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$name>>0]|0;
 $1 = ($0<<24>>24)==(0);
 if ($1) {
  label = 4;
 } else {
  $2 = (_strcmp($name,620519)|0);
  $3 = ($2|0)==(0);
  if ($3) {
   label = 4;
  } else {
   $4 = (_strcmp($name,619451)|0);
   $5 = ($4|0)==(0);
   if ($5) {
    label = 4;
   } else {
    $$01 = 0;
   }
  }
 }
 if ((label|0) == 4) {
  $6 = ($base|0)==(0|0);
  if ($6) {
   $7 = (_calloc(1,4)|0);
   $$01 = $7;
  } else {
   $$01 = $base;
  }
 }
 return ($$01|0);
}
function _setlocale($category,$locale) {
 $category = $category|0;
 $locale = $locale|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (619457|0);
}
function _snprintf($s,$n,$fmt,$varargs) {
 $s = $s|0;
 $n = $n|0;
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $0 = 0, $ap = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ap = sp;
 HEAP32[$ap>>2] = $varargs;
 $0 = (_vsnprintf($s,$n,$fmt,$ap)|0);
 STACKTOP = sp;return ($0|0);
}
function _strncasecmp($_l,$_r,$n) {
 $_l = $_l|0;
 $_r = $_r|0;
 $n = $n|0;
 var $$04 = 0, $$08 = 0, $$08$in = 0, $$pre$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $l$06 = 0, $l$06$lcssa24 = 0, $or$cond = 0, $r$0$lcssa = 0, $r$07 = 0, $r$07$lcssa23 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)==(0);
 if ($0) {
  $$04 = 0;
 } else {
  $1 = HEAP8[$_l>>0]|0;
  $2 = ($1<<24>>24)==(0);
  L3: do {
   if ($2) {
    $21 = 0;$r$0$lcssa = $_r;
   } else {
    $3 = $1&255;
    $$08$in = $n;$8 = $1;$9 = $3;$l$06 = $_l;$r$07 = $_r;
    while(1) {
     $$08 = (($$08$in) + -1)|0;
     $4 = HEAP8[$r$07>>0]|0;
     $5 = ($4<<24>>24)!=(0);
     $6 = ($$08|0)!=(0);
     $or$cond = $6 & $5;
     if (!($or$cond)) {
      $21 = $8;$r$0$lcssa = $r$07;
      break L3;
     }
     $7 = ($8<<24>>24)==($4<<24>>24);
     if (!($7)) {
      $10 = (_tolower($9)|0);
      $11 = HEAP8[$r$07>>0]|0;
      $12 = $11&255;
      $13 = (_tolower($12)|0);
      $14 = ($10|0)==($13|0);
      if (!($14)) {
       $l$06$lcssa24 = $l$06;$r$07$lcssa23 = $r$07;
       break;
      }
     }
     $15 = ((($l$06)) + 1|0);
     $16 = ((($r$07)) + 1|0);
     $17 = HEAP8[$15>>0]|0;
     $18 = $17&255;
     $19 = ($17<<24>>24)==(0);
     if ($19) {
      $21 = 0;$r$0$lcssa = $16;
      break L3;
     } else {
      $$08$in = $$08;$8 = $17;$9 = $18;$l$06 = $15;$r$07 = $16;
     }
    }
    $$pre$pre = HEAP8[$l$06$lcssa24>>0]|0;
    $21 = $$pre$pre;$r$0$lcssa = $r$07$lcssa23;
   }
  } while(0);
  $20 = $21&255;
  $22 = (_tolower($20)|0);
  $23 = HEAP8[$r$0$lcssa>>0]|0;
  $24 = $23&255;
  $25 = (_tolower($24)|0);
  $26 = (($22) - ($25))|0;
  $$04 = $26;
 }
 return ($$04|0);
}
function _strcpy($dest,$src) {
 $dest = $dest|0;
 $src = $src|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 (___stpcpy($dest,$src)|0);
 return ($dest|0);
}
function ___stpcpy($d,$s) {
 $d = $d|0;
 $s = $s|0;
 var $$0$lcssa = 0, $$01$lcssa = 0, $$0115 = 0, $$016 = 0, $$03 = 0, $$1$ph = 0, $$12$ph = 0, $$128 = 0, $$19 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $wd$0$lcssa = 0, $wd$010 = 0, $ws$0$lcssa = 0, $ws$011 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $s;
 $1 = $d;
 $2 = $0 ^ $1;
 $3 = $2 & 3;
 $4 = ($3|0)==(0);
 L1: do {
  if ($4) {
   $5 = $0 & 3;
   $6 = ($5|0)==(0);
   if ($6) {
    $$0$lcssa = $s;$$01$lcssa = $d;
   } else {
    $$0115 = $d;$$016 = $s;
    while(1) {
     $7 = HEAP8[$$016>>0]|0;
     HEAP8[$$0115>>0] = $7;
     $8 = ($7<<24>>24)==(0);
     if ($8) {
      $$03 = $$0115;
      break L1;
     }
     $9 = ((($$016)) + 1|0);
     $10 = ((($$0115)) + 1|0);
     $11 = $9;
     $12 = $11 & 3;
     $13 = ($12|0)==(0);
     if ($13) {
      $$0$lcssa = $9;$$01$lcssa = $10;
      break;
     } else {
      $$0115 = $10;$$016 = $9;
     }
    }
   }
   $14 = HEAP32[$$0$lcssa>>2]|0;
   $15 = (($14) + -16843009)|0;
   $16 = $14 & -2139062144;
   $17 = $16 ^ -2139062144;
   $18 = $17 & $15;
   $19 = ($18|0)==(0);
   if ($19) {
    $22 = $14;$wd$010 = $$01$lcssa;$ws$011 = $$0$lcssa;
    while(1) {
     $20 = ((($ws$011)) + 4|0);
     $21 = ((($wd$010)) + 4|0);
     HEAP32[$wd$010>>2] = $22;
     $23 = HEAP32[$20>>2]|0;
     $24 = (($23) + -16843009)|0;
     $25 = $23 & -2139062144;
     $26 = $25 ^ -2139062144;
     $27 = $26 & $24;
     $28 = ($27|0)==(0);
     if ($28) {
      $22 = $23;$wd$010 = $21;$ws$011 = $20;
     } else {
      $wd$0$lcssa = $21;$ws$0$lcssa = $20;
      break;
     }
    }
   } else {
    $wd$0$lcssa = $$01$lcssa;$ws$0$lcssa = $$0$lcssa;
   }
   $$1$ph = $ws$0$lcssa;$$12$ph = $wd$0$lcssa;
   label = 8;
  } else {
   $$1$ph = $s;$$12$ph = $d;
   label = 8;
  }
 } while(0);
 if ((label|0) == 8) {
  $29 = HEAP8[$$1$ph>>0]|0;
  HEAP8[$$12$ph>>0] = $29;
  $30 = ($29<<24>>24)==(0);
  if ($30) {
   $$03 = $$12$ph;
  } else {
   $$128 = $$12$ph;$$19 = $$1$ph;
   while(1) {
    $31 = ((($$19)) + 1|0);
    $32 = ((($$128)) + 1|0);
    $33 = HEAP8[$31>>0]|0;
    HEAP8[$32>>0] = $33;
    $34 = ($33<<24>>24)==(0);
    if ($34) {
     $$03 = $32;
     break;
    } else {
     $$128 = $32;$$19 = $31;
    }
   }
  }
 }
 return ($$03|0);
}
function _uselocale($l) {
 $l = $l|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_pthread_self()|0);
 $1 = ((($0)) + 180|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($l|0)==(0|0);
 if (!($3)) {
  HEAP32[$1>>2] = $l;
 }
 return ($2|0);
}
function _wmemcpy($d,$s,$n) {
 $d = $d|0;
 $s = $s|0;
 $n = $n|0;
 var $$014 = 0, $$023 = 0, $$05 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)==(0);
 if (!($0)) {
  $$014 = $n;$$023 = $s;$$05 = $d;
  while(1) {
   $1 = (($$014) + -1)|0;
   $2 = ((($$023)) + 4|0);
   $3 = HEAP32[$$023>>2]|0;
   $4 = ((($$05)) + 4|0);
   HEAP32[$$05>>2] = $3;
   $5 = ($1|0)==(0);
   if ($5) {
    break;
   } else {
    $$014 = $1;$$023 = $2;$$05 = $4;
   }
  }
 }
 return ($d|0);
}
function _modf($x,$iptr) {
 $x = +$x;
 $iptr = $iptr|0;
 var $$0 = 0.0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0.0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0.0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAPF64[tempDoublePtr>>3] = $x;$0 = HEAP32[tempDoublePtr>>2]|0;
 $1 = HEAP32[tempDoublePtr+4>>2]|0;
 $2 = (_bitshift64Lshr(($0|0),($1|0),52)|0);
 $3 = tempRet0;
 $4 = $2 & 2047;
 $5 = (($4) + -1023)|0;
 $6 = ($5|0)>(51);
 do {
  if ($6) {
   HEAPF64[$iptr>>3] = $x;
   $7 = ($5|0)!=(1024);
   $8 = $1 & 1048575;
   $9 = ($0|0)==(0);
   $10 = ($8|0)==(0);
   $11 = $9 & $10;
   $or$cond = $11 | $7;
   if ($or$cond) {
    $12 = $1 & -2147483648;
    HEAP32[tempDoublePtr>>2] = 0;HEAP32[tempDoublePtr+4>>2] = $12;$13 = +HEAPF64[tempDoublePtr>>3];
    $$0 = $13;
   } else {
    $$0 = $x;
   }
  } else {
   $14 = ($4>>>0)<(1023);
   if ($14) {
    $15 = $1 & -2147483648;
    $16 = $iptr;
    $17 = $16;
    HEAP32[$17>>2] = 0;
    $18 = (($16) + 4)|0;
    $19 = $18;
    HEAP32[$19>>2] = $15;
    $$0 = $x;
    break;
   }
   $20 = (_bitshift64Lshr(-1,1048575,($5|0))|0);
   $21 = tempRet0;
   $22 = $20 & $0;
   $23 = $21 & $1;
   $24 = ($22|0)==(0);
   $25 = ($23|0)==(0);
   $26 = $24 & $25;
   if ($26) {
    HEAPF64[$iptr>>3] = $x;
    $27 = $1 & -2147483648;
    HEAP32[tempDoublePtr>>2] = 0;HEAP32[tempDoublePtr+4>>2] = $27;$28 = +HEAPF64[tempDoublePtr>>3];
    $$0 = $28;
    break;
   } else {
    $29 = $20 ^ -1;
    $30 = $21 ^ -1;
    $31 = $0 & $29;
    $32 = $1 & $30;
    HEAP32[tempDoublePtr>>2] = $31;HEAP32[tempDoublePtr+4>>2] = $32;$33 = +HEAPF64[tempDoublePtr>>3];
    $34 = $iptr;
    $35 = $34;
    HEAP32[$35>>2] = $31;
    $36 = (($34) + 4)|0;
    $37 = $36;
    HEAP32[$37>>2] = $32;
    $38 = $x - $33;
    $$0 = $38;
    break;
   }
  }
 } while(0);
 return (+$$0);
}
function _exp2($x) {
 $x = +$x;
 var $$0 = 0.0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0.0, $13 = 0, $14 = 0, $15 = 0.0, $16 = 0.0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0.0, $23 = 0.0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0.0, $28 = 0, $29 = 0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAPF64[tempDoublePtr>>3] = $x;$0 = HEAP32[tempDoublePtr>>2]|0;
 $1 = HEAP32[tempDoublePtr+4>>2]|0;
 $2 = $1 & 2147483647;
 $3 = ($2>>>0)>(1083174911);
 do {
  if ($3) {
   $4 = ($2>>>0)>(1083179007);
   $5 = ($1|0)>(-1);
   $6 = ($0>>>0)>(4294967295);
   $7 = ($1|0)==(-1);
   $8 = $7 & $6;
   $9 = $5 | $8;
   $or$cond = $9 & $4;
   if ($or$cond) {
    $10 = $x * 8.9884656743115795E+307;
    $$0 = $10;
    break;
   }
   $11 = ($2>>>0)>(2146435071);
   if ($11) {
    $12 = -1.0 / $x;
    $$0 = $12;
    break;
   } else {
    $13 = !($x <= -1075.0);
    $or$cond1 = $13 | $9;
    if ($or$cond1) {
     label = 9;
     break;
    } else {
     $$0 = 0.0;
     break;
    }
   }
  } else {
   $14 = ($2>>>0)<(1016070144);
   if ($14) {
    $15 = $x + 1.0;
    $$0 = $15;
   } else {
    label = 9;
   }
  }
 } while(0);
 if ((label|0) == 9) {
  $16 = $x + 26388279066624.0;
  HEAPF64[tempDoublePtr>>3] = $16;$17 = HEAP32[tempDoublePtr>>2]|0;
  $18 = HEAP32[tempDoublePtr+4>>2]|0;
  $19 = (($17) + 128)|0;
  $20 = $19 & -256;
  $21 = (($20|0) / 256)&-1;
  $22 = $16 + -26388279066624.0;
  $23 = $x - $22;
  $24 = $19 << 1;
  $25 = $24 & 510;
  $26 = (3448 + ($25<<3)|0);
  $27 = +HEAPF64[$26>>3];
  $28 = $25 | 1;
  $29 = (3448 + ($28<<3)|0);
  $30 = +HEAPF64[$29>>3];
  $31 = $23 - $30;
  $32 = $27 * $31;
  $33 = $31 * 0.0013333559164630223;
  $34 = $33 + 0.0096181298421260663;
  $35 = $31 * $34;
  $36 = $35 + 0.055504108664821403;
  $37 = $31 * $36;
  $38 = $37 + 0.2402265069591;
  $39 = $31 * $38;
  $40 = $39 + 0.69314718055994529;
  $41 = $32 * $40;
  $42 = $27 + $41;
  $43 = (+_scalbn($42,$21));
  $$0 = $43;
 }
 return (+$$0);
}
function _ldexp($x,$n) {
 $x = +$x;
 $n = $n|0;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_scalbn($x,$n));
 return (+$0);
}
function _rint($x) {
 $x = +$x;
 var $$0 = 0.0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, $y$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAPF64[tempDoublePtr>>3] = $x;$0 = HEAP32[tempDoublePtr>>2]|0;
 $1 = HEAP32[tempDoublePtr+4>>2]|0;
 $2 = $1 & 2146435072;
 $3 = ($2>>>0)>(1126170624);
 $4 = (0)>(0);
 $5 = ($2|0)==(1126170624);
 $6 = $5 & $4;
 $7 = $3 | $6;
 if ($7) {
  $$0 = $x;
 } else {
  $8 = ($1|0)<(0);
  $9 = $x + -4503599627370496.0;
  $10 = $9 + 4503599627370496.0;
  $11 = $x + 4503599627370496.0;
  $12 = $11 + -4503599627370496.0;
  $y$0 = $8 ? $10 : $12;
  $13 = $y$0 == 0.0;
  if ($13) {
   $14 = $8 ? -0.0 : 0.0;
   $$0 = $14;
  } else {
   $$0 = $y$0;
  }
 }
 return (+$$0);
}
function _rintf($x) {
 $x = +$x;
 var $$0 = 0.0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0.0, $6 = 0.0, $7 = 0.0, $8 = 0, $9 = 0.0, $y$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (HEAPF32[tempDoublePtr>>2]=$x,HEAP32[tempDoublePtr>>2]|0);
 $1 = $0 & 2130706432;
 $2 = ($1>>>0)>(1249902592);
 if ($2) {
  $$0 = $x;
 } else {
  $3 = ($0|0)<(0);
  $4 = $x + -8388608.0;
  $5 = $4 + 8388608.0;
  $6 = $x + 8388608.0;
  $7 = $6 + -8388608.0;
  $y$0 = $3 ? $5 : $7;
  $8 = $y$0 == 0.0;
  if ($8) {
   $9 = $3 ? -0.0 : 0.0;
   $$0 = $9;
  } else {
   $$0 = $y$0;
  }
 }
 return (+$$0);
}
function _log10($x) {
 $x = +$x;
 var $$0 = 0.0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0;
 var $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0;
 var $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0.0, $71 = 0, $8 = 0, $9 = 0.0, $fabs = 0.0, $hx$0 = 0, $k$0 = 0, $or$cond = 0, $or$cond4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAPF64[tempDoublePtr>>3] = $x;$0 = HEAP32[tempDoublePtr>>2]|0;
 $1 = HEAP32[tempDoublePtr+4>>2]|0;
 $2 = ($1>>>0)<(1048576);
 $3 = ($1|0)<(0);
 $or$cond = $3 | $2;
 do {
  if ($or$cond) {
   $fabs = (+Math_abs((+$x)));
   HEAPF64[tempDoublePtr>>3] = $fabs;$4 = HEAP32[tempDoublePtr>>2]|0;
   $5 = HEAP32[tempDoublePtr+4>>2]|0;
   $6 = ($4|0)==(0);
   $7 = ($5|0)==(0);
   $8 = $6 & $7;
   if ($8) {
    $9 = $x * $x;
    $10 = -1.0 / $9;
    $$0 = $10;
    break;
   }
   if ($3) {
    $11 = $x - $x;
    $12 = $11 / 0.0;
    $$0 = $12;
    break;
   } else {
    $13 = $x * 18014398509481984.0;
    HEAPF64[tempDoublePtr>>3] = $13;$14 = HEAP32[tempDoublePtr>>2]|0;
    $15 = HEAP32[tempDoublePtr+4>>2]|0;
    $26 = $14;$71 = $15;$hx$0 = $15;$k$0 = -1077;
    label = 9;
    break;
   }
  } else {
   $16 = ($1>>>0)>(2146435071);
   if ($16) {
    $$0 = $x;
   } else {
    $17 = ($1|0)==(1072693248);
    $18 = ($0|0)==(0);
    $19 = (0)==(0);
    $20 = $18 & $19;
    $or$cond4 = $20 & $17;
    if ($or$cond4) {
     $$0 = 0.0;
    } else {
     $26 = $0;$71 = $1;$hx$0 = $1;$k$0 = -1023;
     label = 9;
    }
   }
  }
 } while(0);
 if ((label|0) == 9) {
  $21 = (($hx$0) + 614242)|0;
  $22 = $21 >>> 20;
  $23 = (($k$0) + ($22))|0;
  $24 = $21 & 1048575;
  $25 = (($24) + 1072079006)|0;
  HEAP32[tempDoublePtr>>2] = $26;HEAP32[tempDoublePtr+4>>2] = $25;$27 = +HEAPF64[tempDoublePtr>>3];
  $28 = $27 + -1.0;
  $29 = $28 * 0.5;
  $30 = $28 * $29;
  $31 = $28 + 2.0;
  $32 = $28 / $31;
  $33 = $32 * $32;
  $34 = $33 * $33;
  $35 = $34 * 0.15313837699209373;
  $36 = $35 + 0.22222198432149784;
  $37 = $34 * $36;
  $38 = $37 + 0.39999999999409419;
  $39 = $34 * $38;
  $40 = $34 * 0.14798198605116586;
  $41 = $40 + 0.1818357216161805;
  $42 = $34 * $41;
  $43 = $42 + 0.28571428743662391;
  $44 = $34 * $43;
  $45 = $44 + 0.66666666666667351;
  $46 = $33 * $45;
  $47 = $39 + $46;
  $48 = $28 - $30;
  HEAPF64[tempDoublePtr>>3] = $48;$49 = HEAP32[tempDoublePtr>>2]|0;
  $50 = HEAP32[tempDoublePtr+4>>2]|0;
  HEAP32[tempDoublePtr>>2] = 0;HEAP32[tempDoublePtr+4>>2] = $50;$51 = +HEAPF64[tempDoublePtr>>3];
  $52 = $28 - $51;
  $53 = $52 - $30;
  $54 = $30 + $47;
  $55 = $32 * $54;
  $56 = $53 + $55;
  $57 = $51 * 0.43429448187816888;
  $58 = (+($23|0));
  $59 = $58 * 0.30102999566361177;
  $60 = $58 * 3.6942390771589308E-13;
  $61 = $51 + $56;
  $62 = $61 * 2.5082946711645275E-11;
  $63 = $60 + $62;
  $64 = $56 * 0.43429448187816888;
  $65 = $64 + $63;
  $66 = $59 + $57;
  $67 = $59 - $66;
  $68 = $57 + $67;
  $69 = $68 + $65;
  $70 = $66 + $69;
  $$0 = $70;
 }
 return (+$$0);
}
function _log2($x) {
 $x = +$x;
 var $$0 = 0.0, $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0.0, $28 = 0.0, $29 = 0.0, $3 = 0, $30 = 0.0, $31 = 0.0, $32 = 0.0, $33 = 0.0, $34 = 0.0, $35 = 0.0, $36 = 0.0, $37 = 0.0, $38 = 0.0, $39 = 0.0, $4 = 0, $40 = 0.0, $41 = 0.0, $42 = 0.0, $43 = 0.0;
 var $44 = 0.0, $45 = 0.0, $46 = 0.0, $47 = 0.0, $48 = 0.0, $49 = 0, $5 = 0, $50 = 0, $51 = 0.0, $52 = 0.0, $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0.0;
 var $62 = 0.0, $63 = 0.0, $64 = 0.0, $65 = 0.0, $66 = 0.0, $67 = 0.0, $68 = 0, $7 = 0, $8 = 0, $9 = 0.0, $fabs = 0.0, $hx$0 = 0, $k$0 = 0, $or$cond = 0, $or$cond4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAPF64[tempDoublePtr>>3] = $x;$0 = HEAP32[tempDoublePtr>>2]|0;
 $1 = HEAP32[tempDoublePtr+4>>2]|0;
 $2 = ($1>>>0)<(1048576);
 $3 = ($1|0)<(0);
 $or$cond = $3 | $2;
 do {
  if ($or$cond) {
   $fabs = (+Math_abs((+$x)));
   HEAPF64[tempDoublePtr>>3] = $fabs;$4 = HEAP32[tempDoublePtr>>2]|0;
   $5 = HEAP32[tempDoublePtr+4>>2]|0;
   $6 = ($4|0)==(0);
   $7 = ($5|0)==(0);
   $8 = $6 & $7;
   if ($8) {
    $9 = $x * $x;
    $10 = -1.0 / $9;
    $$0 = $10;
    break;
   }
   if ($3) {
    $11 = $x - $x;
    $12 = $11 / 0.0;
    $$0 = $12;
    break;
   } else {
    $13 = $x * 18014398509481984.0;
    HEAPF64[tempDoublePtr>>3] = $13;$14 = HEAP32[tempDoublePtr>>2]|0;
    $15 = HEAP32[tempDoublePtr+4>>2]|0;
    $26 = $14;$68 = $15;$hx$0 = $15;$k$0 = -1077;
    label = 9;
    break;
   }
  } else {
   $16 = ($1>>>0)>(2146435071);
   if ($16) {
    $$0 = $x;
   } else {
    $17 = ($1|0)==(1072693248);
    $18 = ($0|0)==(0);
    $19 = (0)==(0);
    $20 = $18 & $19;
    $or$cond4 = $20 & $17;
    if ($or$cond4) {
     $$0 = 0.0;
    } else {
     $26 = $0;$68 = $1;$hx$0 = $1;$k$0 = -1023;
     label = 9;
    }
   }
  }
 } while(0);
 if ((label|0) == 9) {
  $21 = (($hx$0) + 614242)|0;
  $22 = $21 >>> 20;
  $23 = (($k$0) + ($22))|0;
  $24 = $21 & 1048575;
  $25 = (($24) + 1072079006)|0;
  HEAP32[tempDoublePtr>>2] = $26;HEAP32[tempDoublePtr+4>>2] = $25;$27 = +HEAPF64[tempDoublePtr>>3];
  $28 = $27 + -1.0;
  $29 = $28 * 0.5;
  $30 = $28 * $29;
  $31 = $28 + 2.0;
  $32 = $28 / $31;
  $33 = $32 * $32;
  $34 = $33 * $33;
  $35 = $34 * 0.15313837699209373;
  $36 = $35 + 0.22222198432149784;
  $37 = $34 * $36;
  $38 = $37 + 0.39999999999409419;
  $39 = $34 * $38;
  $40 = $34 * 0.14798198605116586;
  $41 = $40 + 0.1818357216161805;
  $42 = $34 * $41;
  $43 = $42 + 0.28571428743662391;
  $44 = $34 * $43;
  $45 = $44 + 0.66666666666667351;
  $46 = $33 * $45;
  $47 = $39 + $46;
  $48 = $28 - $30;
  HEAPF64[tempDoublePtr>>3] = $48;$49 = HEAP32[tempDoublePtr>>2]|0;
  $50 = HEAP32[tempDoublePtr+4>>2]|0;
  HEAP32[tempDoublePtr>>2] = 0;HEAP32[tempDoublePtr+4>>2] = $50;$51 = +HEAPF64[tempDoublePtr>>3];
  $52 = $28 - $51;
  $53 = $52 - $30;
  $54 = $30 + $47;
  $55 = $32 * $54;
  $56 = $53 + $55;
  $57 = $51 * 1.4426950407214463;
  $58 = $51 + $56;
  $59 = $58 * 1.6751713164886512E-10;
  $60 = $56 * 1.4426950407214463;
  $61 = $60 + $59;
  $62 = (+($23|0));
  $63 = $62 + $57;
  $64 = $62 - $63;
  $65 = $57 + $64;
  $66 = $65 + $61;
  $67 = $63 + $66;
  $$0 = $67;
 }
 return (+$$0);
}
function _tanh($x) {
 $x = +$x;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0.0, $12 = 0.0, $13 = 0, $14 = 0.0, $15 = 0.0, $16 = 0.0, $17 = 0.0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0, $21 = 0.0, $22 = 0.0, $23 = 0.0, $24 = 0, $25 = 0.0, $26 = 0.0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0.0, $8 = 0.0, $9 = 0.0, $fabs = 0.0, $t$0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAPF64[tempDoublePtr>>3] = $x;$0 = HEAP32[tempDoublePtr>>2]|0;
 $1 = HEAP32[tempDoublePtr+4>>2]|0;
 $fabs = (+Math_abs((+$x)));
 HEAPF64[tempDoublePtr>>3] = $fabs;$2 = HEAP32[tempDoublePtr>>2]|0;
 $3 = HEAP32[tempDoublePtr+4>>2]|0;
 $4 = ($3>>>0)>(1071748074);
 do {
  if ($4) {
   $5 = ($3>>>0)>(1077149696);
   if ($5) {
    $6 = 0.0 / $fabs;
    $7 = 1.0 - $6;
    $t$0 = $7;
    break;
   } else {
    $8 = $fabs * 2.0;
    $9 = (+_expm1($8));
    $10 = $9 + 2.0;
    $11 = 2.0 / $10;
    $12 = 1.0 - $11;
    $t$0 = $12;
    break;
   }
  } else {
   $13 = ($3>>>0)>(1070618798);
   if ($13) {
    $14 = $fabs * 2.0;
    $15 = (+_expm1($14));
    $16 = $15 + 2.0;
    $17 = $15 / $16;
    $t$0 = $17;
    break;
   }
   $18 = ($3>>>0)>(1048575);
   if ($18) {
    $19 = $fabs * -2.0;
    $20 = (+_expm1($19));
    $21 = -$20;
    $22 = $20 + 2.0;
    $23 = $21 / $22;
    $t$0 = $23;
   } else {
    $t$0 = $fabs;
   }
  }
 } while(0);
 $24 = ($1|0)<(0);
 $25 = -$t$0;
 $26 = $24 ? $25 : $t$0;
 return (+$26);
}
function _ioctl($fd,$req,$varargs) {
 $fd = $fd|0;
 $req = $req|0;
 $varargs = $varargs|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $ap = 0, $arglist_current = 0, $arglist_next = 0, $expanded = 0, $expanded2 = 0, $expanded4 = 0, $expanded5 = 0, $expanded6 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $ap = sp + 16|0;
 HEAP32[$ap>>2] = $varargs;
 $arglist_current = HEAP32[$ap>>2]|0;
 $0 = $arglist_current;
 $1 = ((0) + 4|0);
 $expanded2 = $1;
 $expanded = (($expanded2) - 1)|0;
 $2 = (($0) + ($expanded))|0;
 $3 = ((0) + 4|0);
 $expanded6 = $3;
 $expanded5 = (($expanded6) - 1)|0;
 $expanded4 = $expanded5 ^ -1;
 $4 = $2 & $expanded4;
 $5 = $4;
 $6 = HEAP32[$5>>2]|0;
 $arglist_next = ((($5)) + 4|0);
 HEAP32[$ap>>2] = $arglist_next;
 HEAP32[$vararg_buffer>>2] = $fd;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $req;
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
 HEAP32[$vararg_ptr2>>2] = $6;
 $7 = (___syscall54(54,($vararg_buffer|0))|0);
 $8 = (___syscall_ret($7)|0);
 STACKTOP = sp;return ($8|0);
}
function _stat($path,$buf) {
 $path = $path|0;
 $buf = $buf|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 HEAP32[$vararg_buffer>>2] = $path;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $buf;
 $0 = (___syscall195(195,($vararg_buffer|0))|0);
 $1 = (___syscall_ret($0)|0);
 STACKTOP = sp;return ($1|0);
}
function ___strdup($s) {
 $s = $s|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strlen($s)|0);
 $1 = (($0) + 1)|0;
 $2 = (_malloc($1)|0);
 $3 = ($2|0)==(0|0);
 if ($3) {
  $$0 = 0;
 } else {
  _memcpy(($2|0),($s|0),($1|0))|0;
  $$0 = $2;
 }
 return ($$0|0);
}
function _getcwd($buf,$size) {
 $buf = $buf|0;
 $size = $size|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $tmp = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 4112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $tmp = sp + 8|0;
 $0 = ($buf|0)==(0|0);
 if ($0) {
  $$0 = 4096;$$01 = $tmp;
  label = 4;
 } else {
  $1 = ($size|0)==(0);
  if ($1) {
   $2 = (___errno_location()|0);
   HEAP32[$2>>2] = 22;
   $$02 = 0;
  } else {
   $$0 = $size;$$01 = $buf;
   label = 4;
  }
 }
 if ((label|0) == 4) {
  HEAP32[$vararg_buffer>>2] = $$01;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $$0;
  $3 = (___syscall183(183,($vararg_buffer|0))|0);
  $4 = (___syscall_ret($3)|0);
  $5 = ($4|0)<(0);
  if ($5) {
   $$02 = 0;
  } else {
   $6 = ($$01|0)==($tmp|0);
   if ($6) {
    $7 = (___strdup($tmp)|0);
    $$02 = $7;
   } else {
    $$02 = $$01;
   }
  }
 }
 STACKTOP = sp;return ($$02|0);
}
function _uname($uts) {
 $uts = $uts|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 HEAP32[$vararg_buffer>>2] = $uts;
 $0 = (___syscall122(122,($vararg_buffer|0))|0);
 $1 = (___syscall_ret($0)|0);
 STACKTOP = sp;return ($1|0);
}
function _mbtowc($wc,$src,$n) {
 $wc = $wc|0;
 $src = $src|0;
 $n = $n|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dummy = 0, $dummy$wc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $dummy = sp;
 $0 = ($src|0)==(0|0);
 L1: do {
  if ($0) {
   $$0 = 0;
  } else {
   $1 = ($n|0)==(0);
   do {
    if (!($1)) {
     $2 = ($wc|0)==(0|0);
     $dummy$wc = $2 ? $dummy : $wc;
     $3 = HEAP8[$src>>0]|0;
     $4 = $3&255;
     $5 = ($3<<24>>24)>(-1);
     if ($5) {
      HEAP32[$dummy$wc>>2] = $4;
      $6 = ($3<<24>>24)!=(0);
      $7 = $6&1;
      $$0 = $7;
      break L1;
     }
     $8 = (($4) + -194)|0;
     $9 = ($8>>>0)>(50);
     if (!($9)) {
      $10 = ((($src)) + 1|0);
      $11 = (149972 + ($8<<2)|0);
      $12 = HEAP32[$11>>2]|0;
      $13 = ($n>>>0)<(4);
      if ($13) {
       $14 = ($n*6)|0;
       $15 = (($14) + -6)|0;
       $16 = -2147483648 >>> $15;
       $17 = $12 & $16;
       $18 = ($17|0)==(0);
       if (!($18)) {
        break;
       }
      }
      $19 = HEAP8[$10>>0]|0;
      $20 = $19&255;
      $21 = $20 >>> 3;
      $22 = (($21) + -16)|0;
      $23 = $12 >> 26;
      $24 = (($21) + ($23))|0;
      $25 = $22 | $24;
      $26 = ($25>>>0)>(7);
      if (!($26)) {
       $27 = $12 << 6;
       $28 = (($20) + -128)|0;
       $29 = $28 | $27;
       $30 = ($29|0)<(0);
       if (!($30)) {
        HEAP32[$dummy$wc>>2] = $29;
        $$0 = 2;
        break L1;
       }
       $31 = ((($src)) + 2|0);
       $32 = HEAP8[$31>>0]|0;
       $33 = $32&255;
       $34 = $33 & 192;
       $35 = ($34|0)==(128);
       if ($35) {
        $36 = $29 << 6;
        $37 = (($33) + -128)|0;
        $38 = $37 | $36;
        $39 = ($38|0)<(0);
        if (!($39)) {
         HEAP32[$dummy$wc>>2] = $38;
         $$0 = 3;
         break L1;
        }
        $40 = ((($src)) + 3|0);
        $41 = HEAP8[$40>>0]|0;
        $42 = $41&255;
        $43 = $42 & 192;
        $44 = ($43|0)==(128);
        if ($44) {
         $45 = $38 << 6;
         $46 = (($42) + -128)|0;
         $47 = $46 | $45;
         HEAP32[$dummy$wc>>2] = $47;
         $$0 = 4;
         break L1;
        }
       }
      }
     }
    }
   } while(0);
   $48 = (___errno_location()|0);
   HEAP32[$48>>2] = 84;
   $$0 = -1;
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _write($fd,$buf,$count) {
 $fd = $fd|0;
 $buf = $buf|0;
 $count = $count|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 HEAP32[$vararg_buffer>>2] = $fd;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $buf;
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
 HEAP32[$vararg_ptr2>>2] = $count;
 $0 = (___syscall4(4,($vararg_buffer|0))|0);
 $1 = (___syscall_ret($0)|0);
 STACKTOP = sp;return ($1|0);
}
function _strchr($s,$c) {
 $s = $s|0;
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (___strchrnul($s,$c)|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = $c&255;
 $3 = ($1<<24>>24)==($2<<24>>24);
 $4 = $3 ? $0 : 0;
 return ($4|0);
}
function ___strchrnul($s,$c) {
 $s = $s|0;
 $c = $c|0;
 var $$0 = 0, $$02$lcssa = 0, $$0211 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond5 = 0, $w$0$lcssa = 0, $w$08 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $c & 255;
 $1 = ($0|0)==(0);
 L1: do {
  if ($1) {
   $6 = (_strlen($s)|0);
   $7 = (($s) + ($6)|0);
   $$0 = $7;
  } else {
   $2 = $s;
   $3 = $2 & 3;
   $4 = ($3|0)==(0);
   if ($4) {
    $$02$lcssa = $s;
   } else {
    $5 = $c&255;
    $$0211 = $s;
    while(1) {
     $8 = HEAP8[$$0211>>0]|0;
     $9 = ($8<<24>>24)==(0);
     $10 = ($8<<24>>24)==($5<<24>>24);
     $or$cond = $9 | $10;
     if ($or$cond) {
      $$0 = $$0211;
      break L1;
     }
     $11 = ((($$0211)) + 1|0);
     $12 = $11;
     $13 = $12 & 3;
     $14 = ($13|0)==(0);
     if ($14) {
      $$02$lcssa = $11;
      break;
     } else {
      $$0211 = $11;
     }
    }
   }
   $15 = Math_imul($0, 16843009)|0;
   $16 = HEAP32[$$02$lcssa>>2]|0;
   $17 = (($16) + -16843009)|0;
   $18 = $16 & -2139062144;
   $19 = $18 ^ -2139062144;
   $20 = $19 & $17;
   $21 = ($20|0)==(0);
   L10: do {
    if ($21) {
     $23 = $16;$w$08 = $$02$lcssa;
     while(1) {
      $22 = $23 ^ $15;
      $24 = (($22) + -16843009)|0;
      $25 = $22 & -2139062144;
      $26 = $25 ^ -2139062144;
      $27 = $26 & $24;
      $28 = ($27|0)==(0);
      if (!($28)) {
       $w$0$lcssa = $w$08;
       break L10;
      }
      $29 = ((($w$08)) + 4|0);
      $30 = HEAP32[$29>>2]|0;
      $31 = (($30) + -16843009)|0;
      $32 = $30 & -2139062144;
      $33 = $32 ^ -2139062144;
      $34 = $33 & $31;
      $35 = ($34|0)==(0);
      if ($35) {
       $23 = $30;$w$08 = $29;
      } else {
       $w$0$lcssa = $29;
       break;
      }
     }
    } else {
     $w$0$lcssa = $$02$lcssa;
    }
   } while(0);
   $36 = $c&255;
   $$1 = $w$0$lcssa;
   while(1) {
    $37 = HEAP8[$$1>>0]|0;
    $38 = ($37<<24>>24)==(0);
    $39 = ($37<<24>>24)==($36<<24>>24);
    $or$cond5 = $38 | $39;
    $40 = ((($$1)) + 1|0);
    if ($or$cond5) {
     $$0 = $$1;
     break;
    } else {
     $$1 = $40;
    }
   }
  }
 } while(0);
 return ($$0|0);
}
function _getpid() {
 var $0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (___syscall20(20,($vararg_buffer|0))|0);
 STACKTOP = sp;return ($0|0);
}
function _fopen($filename,$mode) {
 $filename = $filename|0;
 $mode = $mode|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $memchr = 0, $vararg_buffer = 0, $vararg_buffer3 = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $0 = HEAP8[$mode>>0]|0;
 $1 = $0 << 24 >> 24;
 $memchr = (_memchr(619465,$1,4)|0);
 $2 = ($memchr|0)==(0|0);
 if ($2) {
  $3 = (___errno_location()|0);
  HEAP32[$3>>2] = 22;
  $$0 = 0;
 } else {
  $4 = (___fmodeflags($mode)|0);
  $5 = $4 | 32768;
  HEAP32[$vararg_buffer>>2] = $filename;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $5;
  $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
  HEAP32[$vararg_ptr2>>2] = 438;
  $6 = (___syscall5(5,($vararg_buffer|0))|0);
  $7 = (___syscall_ret($6)|0);
  $8 = ($7|0)<(0);
  if ($8) {
   $$0 = 0;
  } else {
   $9 = (___fdopen($7,$mode)|0);
   $10 = ($9|0)==(0|0);
   if ($10) {
    HEAP32[$vararg_buffer3>>2] = $7;
    (___syscall6(6,($vararg_buffer3|0))|0);
    $$0 = 0;
   } else {
    $$0 = $9;
   }
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function ___fmodeflags($mode) {
 $mode = $mode|0;
 var $$ = 0, $$flags$4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $flags$0 = 0, $flags$0$ = 0, $flags$2 = 0;
 var $flags$2$ = 0, $flags$4 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strchr($mode,43)|0);
 $1 = ($0|0)==(0|0);
 $2 = HEAP8[$mode>>0]|0;
 $not$ = ($2<<24>>24)!=(114);
 $$ = $not$&1;
 $flags$0 = $1 ? $$ : 2;
 $3 = (_strchr($mode,120)|0);
 $4 = ($3|0)==(0|0);
 $5 = $flags$0 | 128;
 $flags$0$ = $4 ? $flags$0 : $5;
 $6 = (_strchr($mode,101)|0);
 $7 = ($6|0)==(0|0);
 $8 = $flags$0$ | 524288;
 $flags$2 = $7 ? $flags$0$ : $8;
 $9 = ($2<<24>>24)==(114);
 $10 = $flags$2 | 64;
 $flags$2$ = $9 ? $flags$2 : $10;
 $11 = ($2<<24>>24)==(119);
 $12 = $flags$2$ | 512;
 $flags$4 = $11 ? $12 : $flags$2$;
 $13 = ($2<<24>>24)==(97);
 $14 = $flags$4 | 1024;
 $$flags$4 = $13 ? $14 : $flags$4;
 return ($$flags$4|0);
}
function ___fdopen($fd,$mode) {
 $fd = $fd|0;
 $mode = $mode|0;
 var $$0 = 0, $$cast = 0, $$pre = 0, $$pre1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $memchr = 0, $tio = 0, $vararg_buffer = 0, $vararg_buffer12 = 0, $vararg_buffer3 = 0, $vararg_buffer7 = 0, $vararg_ptr1 = 0, $vararg_ptr10 = 0, $vararg_ptr11 = 0, $vararg_ptr15 = 0, $vararg_ptr16 = 0, $vararg_ptr2 = 0, $vararg_ptr6 = 0, dest = 0, label = 0;
 var sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer12 = sp + 40|0;
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $tio = sp + 52|0;
 $0 = HEAP8[$mode>>0]|0;
 $1 = $0 << 24 >> 24;
 $memchr = (_memchr(619465,$1,4)|0);
 $2 = ($memchr|0)==(0|0);
 if ($2) {
  $3 = (___errno_location()|0);
  HEAP32[$3>>2] = 22;
  $$0 = 0;
 } else {
  $4 = (_malloc(1144)|0);
  $5 = ($4|0)==(0|0);
  if ($5) {
   $$0 = 0;
  } else {
   dest=$4; stop=dest+112|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
   $6 = (_strchr($mode,43)|0);
   $7 = ($6|0)==(0|0);
   if ($7) {
    $8 = ($0<<24>>24)==(114);
    $9 = $8 ? 8 : 4;
    HEAP32[$4>>2] = $9;
   }
   $10 = (_strchr($mode,101)|0);
   $11 = ($10|0)==(0|0);
   if ($11) {
    $13 = $0;
   } else {
    HEAP32[$vararg_buffer>>2] = $fd;
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
    HEAP32[$vararg_ptr1>>2] = 2;
    $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
    HEAP32[$vararg_ptr2>>2] = 1;
    (___syscall221(221,($vararg_buffer|0))|0);
    $$pre = HEAP8[$mode>>0]|0;
    $13 = $$pre;
   }
   $12 = ($13<<24>>24)==(97);
   if ($12) {
    HEAP32[$vararg_buffer3>>2] = $fd;
    $vararg_ptr6 = ((($vararg_buffer3)) + 4|0);
    HEAP32[$vararg_ptr6>>2] = 3;
    $14 = (___syscall221(221,($vararg_buffer3|0))|0);
    $15 = $14 & 1024;
    $16 = ($15|0)==(0);
    if ($16) {
     $17 = $14 | 1024;
     HEAP32[$vararg_buffer7>>2] = $fd;
     $vararg_ptr10 = ((($vararg_buffer7)) + 4|0);
     HEAP32[$vararg_ptr10>>2] = 4;
     $vararg_ptr11 = ((($vararg_buffer7)) + 8|0);
     HEAP32[$vararg_ptr11>>2] = $17;
     (___syscall221(221,($vararg_buffer7|0))|0);
    }
    $18 = HEAP32[$4>>2]|0;
    $19 = $18 | 128;
    HEAP32[$4>>2] = $19;
    $26 = $19;
   } else {
    $$pre1 = HEAP32[$4>>2]|0;
    $26 = $$pre1;
   }
   $20 = ((($4)) + 60|0);
   HEAP32[$20>>2] = $fd;
   $21 = ((($4)) + 120|0);
   $22 = ((($4)) + 44|0);
   HEAP32[$22>>2] = $21;
   $23 = ((($4)) + 48|0);
   HEAP32[$23>>2] = 1024;
   $24 = ((($4)) + 75|0);
   HEAP8[$24>>0] = -1;
   $25 = $26 & 8;
   $27 = ($25|0)==(0);
   if ($27) {
    HEAP32[$vararg_buffer12>>2] = $fd;
    $vararg_ptr15 = ((($vararg_buffer12)) + 4|0);
    HEAP32[$vararg_ptr15>>2] = 21505;
    $vararg_ptr16 = ((($vararg_buffer12)) + 8|0);
    HEAP32[$vararg_ptr16>>2] = $tio;
    $28 = (___syscall54(54,($vararg_buffer12|0))|0);
    $29 = ($28|0)==(0);
    if ($29) {
     HEAP8[$24>>0] = 10;
    }
   }
   $30 = ((($4)) + 32|0);
   HEAP32[$30>>2] = 3005;
   $31 = ((($4)) + 36|0);
   HEAP32[$31>>2] = 3003;
   $32 = ((($4)) + 40|0);
   HEAP32[$32>>2] = 3004;
   $33 = ((($4)) + 12|0);
   HEAP32[$33>>2] = 3002;
   $34 = HEAP32[(1205956)>>2]|0;
   $35 = ($34|0)==(0);
   if ($35) {
    $36 = ((($4)) + 76|0);
    HEAP32[$36>>2] = -1;
   }
   ___lock(((1205980)|0));
   $37 = HEAP32[(1205976)>>2]|0;
   $38 = ((($4)) + 56|0);
   HEAP32[$38>>2] = $37;
   $39 = ($37|0)==(0);
   if (!($39)) {
    $$cast = $37;
    $40 = ((($$cast)) + 52|0);
    HEAP32[$40>>2] = $4;
   }
   HEAP32[(1205976)>>2] = $4;
   ___unlock(((1205980)|0));
   $$0 = $4;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _fclose($f) {
 $f = $f|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 76|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(-1);
 if ($2) {
 }
 $3 = HEAP32[$f>>2]|0;
 $4 = $3 & 1;
 $5 = ($4|0)!=(0);
 if (!($5)) {
  ___lock(((1205980)|0));
  $6 = ((($f)) + 52|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7|0)==(0|0);
  $9 = $7;
  $$pre = ((($f)) + 56|0);
  if (!($8)) {
   $10 = HEAP32[$$pre>>2]|0;
   $11 = ((($7)) + 56|0);
   HEAP32[$11>>2] = $10;
  }
  $12 = HEAP32[$$pre>>2]|0;
  $13 = ($12|0)==(0|0);
  $14 = $12;
  if (!($13)) {
   $15 = ((($12)) + 52|0);
   HEAP32[$15>>2] = $9;
  }
  $16 = HEAP32[(1205976)>>2]|0;
  $17 = ($16|0)==($f|0);
  if ($17) {
   HEAP32[(1205976)>>2] = $14;
  }
  ___unlock(((1205980)|0));
 }
 $18 = (_fflush($f)|0);
 $19 = ((($f)) + 12|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = (FUNCTION_TABLE_ii[$20 & 8191]($f)|0);
 $22 = $21 | $18;
 $23 = ((($f)) + 92|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ($24|0)==(0|0);
 if (!($25)) {
  _free($24);
 }
 if (!($5)) {
  _free($f);
 }
 return ($22|0);
}
function _fflush($f) {
 $f = $f|0;
 var $$0 = 0, $$01 = 0, $$012 = 0, $$014 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, $r$0$lcssa = 0, $r$03 = 0, $r$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($f|0)==(0|0);
 do {
  if ($0) {
   $7 = HEAP32[37631]|0;
   $8 = ($7|0)==(0|0);
   if ($8) {
    $27 = 0;
   } else {
    $9 = HEAP32[37631]|0;
    $10 = (_fflush($9)|0);
    $27 = $10;
   }
   ___lock(((1205980)|0));
   $$012 = HEAP32[(1205976)>>2]|0;
   $11 = ($$012|0)==(0|0);
   if ($11) {
    $r$0$lcssa = $27;
   } else {
    $$014 = $$012;$r$03 = $27;
    while(1) {
     $12 = ((($$014)) + 76|0);
     $13 = HEAP32[$12>>2]|0;
     $14 = ($13|0)>(-1);
     if ($14) {
      $15 = (___lockfile($$014)|0);
      $24 = $15;
     } else {
      $24 = 0;
     }
     $16 = ((($$014)) + 20|0);
     $17 = HEAP32[$16>>2]|0;
     $18 = ((($$014)) + 28|0);
     $19 = HEAP32[$18>>2]|0;
     $20 = ($17>>>0)>($19>>>0);
     if ($20) {
      $21 = (___fflush_unlocked($$014)|0);
      $22 = $21 | $r$03;
      $r$1 = $22;
     } else {
      $r$1 = $r$03;
     }
     $23 = ($24|0)==(0);
     if (!($23)) {
      ___unlockfile($$014);
     }
     $25 = ((($$014)) + 56|0);
     $$01 = HEAP32[$25>>2]|0;
     $26 = ($$01|0)==(0|0);
     if ($26) {
      $r$0$lcssa = $r$1;
      break;
     } else {
      $$014 = $$01;$r$03 = $r$1;
     }
    }
   }
   ___unlock(((1205980)|0));
   $$0 = $r$0$lcssa;
  } else {
   $1 = ((($f)) + 76|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = ($2|0)>(-1);
   if (!($3)) {
    $4 = (___fflush_unlocked($f)|0);
    $$0 = $4;
    break;
   }
   $5 = (___lockfile($f)|0);
   $phitmp = ($5|0)==(0);
   $6 = (___fflush_unlocked($f)|0);
   if ($phitmp) {
    $$0 = $6;
   } else {
    ___unlockfile($f);
    $$0 = $6;
   }
  }
 } while(0);
 return ($$0|0);
}
function ___fflush_unlocked($f) {
 $f = $f|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($f)) + 28|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1>>>0)>($3>>>0);
 if ($4) {
  $5 = ((($f)) + 36|0);
  $6 = HEAP32[$5>>2]|0;
  (FUNCTION_TABLE_iiii[$6 & 8191]($f,0,0)|0);
  $7 = HEAP32[$0>>2]|0;
  $8 = ($7|0)==(0|0);
  if ($8) {
   $$0 = -1;
  } else {
   label = 3;
  }
 } else {
  label = 3;
 }
 if ((label|0) == 3) {
  $9 = ((($f)) + 4|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ((($f)) + 8|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ($10>>>0)<($12>>>0);
  if ($13) {
   $14 = ((($f)) + 40|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = $10;
   $17 = $12;
   $18 = (($16) - ($17))|0;
   (FUNCTION_TABLE_iiii[$15 & 8191]($f,$18,1)|0);
  }
  $19 = ((($f)) + 16|0);
  HEAP32[$19>>2] = 0;
  HEAP32[$2>>2] = 0;
  HEAP32[$0>>2] = 0;
  HEAP32[$11>>2] = 0;
  HEAP32[$9>>2] = 0;
  $$0 = 0;
 }
 return ($$0|0);
}
function _fgets($s,$n,$f) {
 $s = $s|0;
 $n = $n|0;
 $f = $f|0;
 var $$0 = 0, $$047 = 0, $$05 = 0, $$lcssa13 = 0, $$old2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond3 = 0, $p$0 = 0;
 var $p$1 = 0, $sext$mask = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 76|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(-1);
 if ($2) {
  $3 = (___lockfile($f)|0);
  $13 = $3;
 } else {
  $13 = 0;
 }
 $4 = (($n) + -1)|0;
 $5 = ($n|0)<(2);
 if ($5) {
  $6 = ((($f)) + 74|0);
  $7 = HEAP8[$6>>0]|0;
  $8 = $7 << 24 >> 24;
  $9 = (($8) + 255)|0;
  $10 = $9 | $8;
  $11 = $10&255;
  HEAP8[$6>>0] = $11;
  $12 = ($13|0)==(0);
  if (!($12)) {
   ___unlockfile($f);
  }
  $14 = ($4|0)==(0);
  if ($14) {
   HEAP8[$s>>0] = 0;
   $$0 = $s;
  } else {
   $$0 = 0;
  }
 } else {
  $$old2 = ($4|0)==(0);
  L11: do {
   if ($$old2) {
    $p$1 = $s;
    label = 17;
   } else {
    $15 = ((($f)) + 4|0);
    $16 = ((($f)) + 8|0);
    $$05 = $4;$p$0 = $s;
    while(1) {
     $17 = HEAP32[$15>>2]|0;
     $18 = HEAP32[$16>>2]|0;
     $19 = $17;
     $20 = (($18) - ($19))|0;
     $21 = (_memchr($17,10,$20)|0);
     $22 = ($21|0)==(0|0);
     $23 = $21;
     $24 = (1 - ($19))|0;
     $25 = (($24) + ($23))|0;
     $26 = $22 ? $20 : $25;
     $27 = ($26>>>0)<($$05>>>0);
     $28 = $27 ? $26 : $$05;
     _memcpy(($p$0|0),($17|0),($28|0))|0;
     $29 = HEAP32[$15>>2]|0;
     $30 = (($29) + ($28)|0);
     HEAP32[$15>>2] = $30;
     $31 = (($p$0) + ($28)|0);
     $32 = (($$05) - ($28))|0;
     $or$cond = $22 & $27;
     if (!($or$cond)) {
      $p$1 = $31;
      label = 17;
      break L11;
     }
     $33 = HEAP32[$16>>2]|0;
     $34 = ($30>>>0)<($33>>>0);
     if ($34) {
      $35 = ((($30)) + 1|0);
      HEAP32[$15>>2] = $35;
      $36 = HEAP8[$30>>0]|0;
      $37 = $36&255;
      $46 = $37;
     } else {
      $38 = (___uflow($f)|0);
      $39 = ($38|0)<(0);
      if ($39) {
       $$lcssa13 = $31;
       break;
      } else {
       $46 = $38;
      }
     }
     $44 = (($32) + -1)|0;
     $45 = $46&255;
     $47 = ((($31)) + 1|0);
     HEAP8[$31>>0] = $45;
     $sext$mask = $46 & 255;
     $48 = ($sext$mask|0)!=(10);
     $49 = ($44|0)!=(0);
     $or$cond3 = $49 & $48;
     if ($or$cond3) {
      $$05 = $44;$p$0 = $47;
     } else {
      $p$1 = $47;
      label = 17;
      break L11;
     }
    }
    $40 = ($$lcssa13|0)==($s|0);
    if ($40) {
     $$047 = 0;
    } else {
     $41 = HEAP32[$f>>2]|0;
     $42 = $41 & 16;
     $43 = ($42|0)==(0);
     if ($43) {
      $$047 = 0;
     } else {
      $p$1 = $$lcssa13;
      label = 17;
     }
    }
   }
  } while(0);
  if ((label|0) == 17) {
   $50 = ($s|0)==(0|0);
   if ($50) {
    $$047 = 0;
   } else {
    HEAP8[$p$1>>0] = 0;
    $$047 = $s;
   }
  }
  $51 = ($13|0)==(0);
  if ($51) {
   $$0 = $$047;
  } else {
   ___unlockfile($f);
   $$0 = $$047;
  }
 }
 return ($$0|0);
}
function _feof($f) {
 $f = $f|0;
 var $$lobit = 0, $$lobit1 = 0, $$lobit2 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 76|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(-1);
 if ($2) {
  $5 = (___lockfile($f)|0);
  $phitmp = ($5|0)==(0);
  $6 = HEAP32[$f>>2]|0;
  $7 = $6 >>> 4;
  $$lobit = $7 & 1;
  if ($phitmp) {
   $$lobit2 = $$lobit;
  } else {
   $$lobit2 = $$lobit;
  }
 } else {
  $3 = HEAP32[$f>>2]|0;
  $4 = $3 >>> 4;
  $$lobit1 = $4 & 1;
  $$lobit2 = $$lobit1;
 }
 return ($$lobit2|0);
}
function _ferror($f) {
 $f = $f|0;
 var $$lobit = 0, $$lobit1 = 0, $$lobit2 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 76|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(-1);
 if ($2) {
  $5 = (___lockfile($f)|0);
  $phitmp = ($5|0)==(0);
  $6 = HEAP32[$f>>2]|0;
  $7 = $6 >>> 5;
  $$lobit = $7 & 1;
  if ($phitmp) {
   $$lobit2 = $$lobit;
  } else {
   $$lobit2 = $$lobit;
  }
 } else {
  $3 = HEAP32[$f>>2]|0;
  $4 = $3 >>> 5;
  $$lobit1 = $4 & 1;
  $$lobit2 = $$lobit1;
 }
 return ($$lobit2|0);
}
function _fscanf($f,$fmt,$varargs) {
 $f = $f|0;
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $0 = 0, $ap = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ap = sp;
 HEAP32[$ap>>2] = $varargs;
 $0 = (_vfscanf($f,$fmt,$ap)|0);
 STACKTOP = sp;return ($0|0);
}
function _vfscanf($f,$fmt,$ap) {
 $f = $f|0;
 $fmt = $fmt|0;
 $ap = $ap|0;
 var $$ = 0, $$11 = 0, $$12 = 0, $$13 = 0, $$14 = 0, $$lcssa = 0, $$lcssa386 = 0, $$lcssa40 = 0, $$not = 0, $$old4 = 0, $$pre = 0, $$pre$phi184Z2D = 0, $$pre173 = 0, $$pre175 = 0, $$pre177 = 0, $$pre179 = 0, $$pre180 = 0, $$pre181 = 0, $$pre182 = 0, $$pre183 = 0;
 var $$size$0 = 0, $$width$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0;
 var $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0;
 var $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0;
 var $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0.0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0;
 var $312 = 0, $313 = 0, $314 = 0.0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $alloc$0 = 0, $alloc$0402 = 0, $alloc$1 = 0, $alloc$2 = 0, $ap2$i = 0, $arglist_current = 0, $arglist_current2 = 0, $arglist_next = 0, $arglist_next3 = 0, $base$0 = 0;
 var $c$0102 = 0, $dest$0 = 0, $expanded = 0, $expanded10 = 0, $expanded11 = 0, $expanded13 = 0, $expanded14 = 0, $expanded15 = 0, $expanded4 = 0, $expanded6 = 0, $expanded7 = 0, $expanded8 = 0, $factor = 0, $factor18 = 0, $i$0$i = 0, $i$0$ph = 0, $i$0$ph$phi = 0, $i$0$ph22 = 0, $i$0$ph22$lcssa = 0, $i$1 = 0;
 var $i$2 = 0, $i$2$ph = 0, $i$2$ph$phi = 0, $i$3 = 0, $i$4 = 0, $invert$0 = 0, $isdigit = 0, $isdigit8 = 0, $isdigit897 = 0, $isdigittmp = 0, $isdigittmp7 = 0, $isdigittmp796 = 0, $k$0$ph = 0, $k$1$ph = 0, $matches$0$ = 0, $matches$0107 = 0, $matches$0107$lcssa = 0, $matches$0107371 = 0, $matches$1 = 0, $matches$2 = 0;
 var $matches$3 = 0, $not$ = 0, $or$cond = 0, $or$cond10 = 0, $or$cond3 = 0, $or$cond5 = 0, $p$0110 = 0, $p$1 = 0, $p$1$lcssa = 0, $p$10 = 0, $p$11 = 0, $p$2 = 0, $p$3$lcssa = 0, $p$398 = 0, $p$4 = 0, $p$5 = 0, $p$6 = 0, $p$7 = 0, $p$7$ph = 0, $p$8 = 0;
 var $p$9 = 0, $pos$0111 = 0, $pos$1 = 0, $pos$2 = 0, $s$0105 = 0, $s$0105$lcssa = 0, $s$1 = 0, $s$2$ph = 0, $s$4 = 0, $s$5 = 0, $s$6 = 0, $s$7 = 0, $s$8 = 0, $s$9 = 0, $scanset = 0, $size$0 = 0, $st = 0, $vacopy_currentptr = 0, $wc = 0, $wcs$0106 = 0;
 var $wcs$0106$lcssa = 0, $wcs$1 = 0, $wcs$10 = 0, $wcs$2 = 0, $wcs$3$ph = 0, $wcs$3$ph$lcssa = 0, $wcs$5 = 0, $wcs$6 = 0, $wcs$7 = 0, $wcs$8 = 0, $wcs$9 = 0, $width$0$lcssa = 0, $width$099 = 0, $width$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ap2$i = sp + 16|0;
 $st = sp + 8|0;
 $scanset = sp + 33|0;
 $wc = sp;
 $0 = sp + 32|0;
 $1 = ((($f)) + 76|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)>(-1);
 if ($3) {
  $4 = (___lockfile($f)|0);
  $327 = $4;
 } else {
  $327 = 0;
 }
 $5 = HEAP8[$fmt>>0]|0;
 $6 = ($5<<24>>24)==(0);
 L4: do {
  if ($6) {
   $matches$3 = 0;
  } else {
   $7 = ((($f)) + 4|0);
   $8 = ((($f)) + 100|0);
   $9 = ((($f)) + 108|0);
   $10 = ((($f)) + 8|0);
   $11 = ((($scanset)) + 10|0);
   $12 = ((($scanset)) + 33|0);
   $13 = ((($st)) + 4|0);
   $14 = ((($scanset)) + 46|0);
   $15 = ((($scanset)) + 94|0);
   $17 = $5;$matches$0107 = 0;$p$0110 = $fmt;$pos$0111 = 0;$s$0105 = 0;$wcs$0106 = 0;
   L6: while(1) {
    $16 = $17&255;
    $18 = (_isspace($16)|0);
    $19 = ($18|0)==(0);
    L8: do {
     if ($19) {
      $46 = HEAP8[$p$0110>>0]|0;
      $47 = ($46<<24>>24)==(37);
      L10: do {
       if ($47) {
        $48 = ((($p$0110)) + 1|0);
        $49 = HEAP8[$48>>0]|0;
        L12: do {
         switch ($49<<24>>24) {
         case 37:  {
          break L10;
          break;
         }
         case 42:  {
          $70 = ((($p$0110)) + 2|0);
          $dest$0 = 0;$p$2 = $70;
          break;
         }
         default: {
          $71 = $49&255;
          $isdigittmp = (($71) + -48)|0;
          $isdigit = ($isdigittmp>>>0)<(10);
          if ($isdigit) {
           $72 = ((($p$0110)) + 2|0);
           $73 = HEAP8[$72>>0]|0;
           $74 = ($73<<24>>24)==(36);
           if ($74) {
            $vacopy_currentptr = HEAP32[$ap>>2]|0;
            HEAP32[$ap2$i>>2] = $vacopy_currentptr;
            $i$0$i = $isdigittmp;
            while(1) {
             $75 = ($i$0$i>>>0)>(1);
             $arglist_current = HEAP32[$ap2$i>>2]|0;
             $76 = $arglist_current;
             $77 = ((0) + 4|0);
             $expanded4 = $77;
             $expanded = (($expanded4) - 1)|0;
             $78 = (($76) + ($expanded))|0;
             $79 = ((0) + 4|0);
             $expanded8 = $79;
             $expanded7 = (($expanded8) - 1)|0;
             $expanded6 = $expanded7 ^ -1;
             $80 = $78 & $expanded6;
             $81 = $80;
             $82 = HEAP32[$81>>2]|0;
             $arglist_next = ((($81)) + 4|0);
             HEAP32[$ap2$i>>2] = $arglist_next;
             $83 = (($i$0$i) + -1)|0;
             if ($75) {
              $i$0$i = $83;
             } else {
              $$lcssa = $82;
              break;
             }
            }
            $84 = ((($p$0110)) + 3|0);
            $dest$0 = $$lcssa;$p$2 = $84;
            break L12;
           }
          }
          $arglist_current2 = HEAP32[$ap>>2]|0;
          $85 = $arglist_current2;
          $86 = ((0) + 4|0);
          $expanded11 = $86;
          $expanded10 = (($expanded11) - 1)|0;
          $87 = (($85) + ($expanded10))|0;
          $88 = ((0) + 4|0);
          $expanded15 = $88;
          $expanded14 = (($expanded15) - 1)|0;
          $expanded13 = $expanded14 ^ -1;
          $89 = $87 & $expanded13;
          $90 = $89;
          $91 = HEAP32[$90>>2]|0;
          $arglist_next3 = ((($90)) + 4|0);
          HEAP32[$ap>>2] = $arglist_next3;
          $dest$0 = $91;$p$2 = $48;
         }
         }
        } while(0);
        $92 = HEAP8[$p$2>>0]|0;
        $93 = $92&255;
        $isdigittmp796 = (($93) + -48)|0;
        $isdigit897 = ($isdigittmp796>>>0)<(10);
        if ($isdigit897) {
         $97 = $93;$p$398 = $p$2;$width$099 = 0;
         while(1) {
          $94 = ($width$099*10)|0;
          $95 = (($94) + -48)|0;
          $96 = (($95) + ($97))|0;
          $98 = ((($p$398)) + 1|0);
          $99 = HEAP8[$98>>0]|0;
          $100 = $99&255;
          $isdigittmp7 = (($100) + -48)|0;
          $isdigit8 = ($isdigittmp7>>>0)<(10);
          if ($isdigit8) {
           $97 = $100;$p$398 = $98;$width$099 = $96;
          } else {
           $$lcssa40 = $99;$p$3$lcssa = $98;$width$0$lcssa = $96;
           break;
          }
         }
        } else {
         $$lcssa40 = $92;$p$3$lcssa = $p$2;$width$0$lcssa = 0;
        }
        $101 = ($$lcssa40<<24>>24)==(109);
        if ($101) {
         $102 = ($dest$0|0)!=(0|0);
         $103 = $102&1;
         $104 = ((($p$3$lcssa)) + 1|0);
         $$pre173 = HEAP8[$104>>0]|0;
         $107 = $$pre173;$alloc$0 = $103;$p$4 = $104;$s$1 = 0;$wcs$1 = 0;
        } else {
         $107 = $$lcssa40;$alloc$0 = 0;$p$4 = $p$3$lcssa;$s$1 = $s$0105;$wcs$1 = $wcs$0106;
        }
        $105 = ((($p$4)) + 1|0);
        $106 = $107&255;
        switch ($106|0) {
        case 104:  {
         $108 = HEAP8[$105>>0]|0;
         $109 = ($108<<24>>24)==(104);
         $110 = ((($p$4)) + 2|0);
         $$11 = $109 ? $110 : $105;
         $$12 = $109 ? -2 : -1;
         $p$5 = $$11;$size$0 = $$12;
         break;
        }
        case 108:  {
         $111 = HEAP8[$105>>0]|0;
         $112 = ($111<<24>>24)==(108);
         $113 = ((($p$4)) + 2|0);
         $$13 = $112 ? $113 : $105;
         $$14 = $112 ? 3 : 1;
         $p$5 = $$13;$size$0 = $$14;
         break;
        }
        case 106:  {
         $p$5 = $105;$size$0 = 3;
         break;
        }
        case 116: case 122:  {
         $p$5 = $105;$size$0 = 1;
         break;
        }
        case 76:  {
         $p$5 = $105;$size$0 = 2;
         break;
        }
        case 110: case 112: case 67: case 83: case 91: case 99: case 115: case 88: case 71: case 70: case 69: case 65: case 103: case 102: case 101: case 97: case 120: case 117: case 111: case 105: case 100:  {
         $p$5 = $p$4;$size$0 = 0;
         break;
        }
        default: {
         $alloc$0402 = $alloc$0;$matches$0107371 = $matches$0107;$s$7 = $s$1;$wcs$8 = $wcs$1;
         label = 154;
         break L6;
        }
        }
        $114 = HEAP8[$p$5>>0]|0;
        $115 = $114&255;
        $116 = $115 & 47;
        $117 = ($116|0)==(3);
        $118 = $115 | 32;
        $$ = $117 ? $118 : $115;
        $$size$0 = $117 ? 1 : $size$0;
        switch ($$|0) {
        case 99:  {
         $119 = ($width$0$lcssa|0)<(1);
         $$width$0 = $119 ? 1 : $width$0$lcssa;
         $pos$1 = $pos$0111;$width$1 = $$width$0;
         break;
        }
        case 91:  {
         $pos$1 = $pos$0111;$width$1 = $width$0$lcssa;
         break;
        }
        case 110:  {
         $120 = ($pos$0111|0)<(0);
         $121 = $120 << 31 >> 31;
         $122 = ($dest$0|0)==(0|0);
         if ($122) {
          $matches$1 = $matches$0107;$p$11 = $p$5;$pos$2 = $pos$0111;$s$6 = $s$1;$wcs$7 = $wcs$1;
          break L8;
         }
         switch ($$size$0|0) {
         case -2:  {
          $123 = $pos$0111&255;
          HEAP8[$dest$0>>0] = $123;
          $matches$1 = $matches$0107;$p$11 = $p$5;$pos$2 = $pos$0111;$s$6 = $s$1;$wcs$7 = $wcs$1;
          break L8;
          break;
         }
         case -1:  {
          $124 = $pos$0111&65535;
          HEAP16[$dest$0>>1] = $124;
          $matches$1 = $matches$0107;$p$11 = $p$5;$pos$2 = $pos$0111;$s$6 = $s$1;$wcs$7 = $wcs$1;
          break L8;
          break;
         }
         case 0:  {
          HEAP32[$dest$0>>2] = $pos$0111;
          $matches$1 = $matches$0107;$p$11 = $p$5;$pos$2 = $pos$0111;$s$6 = $s$1;$wcs$7 = $wcs$1;
          break L8;
          break;
         }
         case 1:  {
          HEAP32[$dest$0>>2] = $pos$0111;
          $matches$1 = $matches$0107;$p$11 = $p$5;$pos$2 = $pos$0111;$s$6 = $s$1;$wcs$7 = $wcs$1;
          break L8;
          break;
         }
         case 3:  {
          $125 = $dest$0;
          $126 = $125;
          HEAP32[$126>>2] = $pos$0111;
          $127 = (($125) + 4)|0;
          $128 = $127;
          HEAP32[$128>>2] = $121;
          $matches$1 = $matches$0107;$p$11 = $p$5;$pos$2 = $pos$0111;$s$6 = $s$1;$wcs$7 = $wcs$1;
          break L8;
          break;
         }
         default: {
          $matches$1 = $matches$0107;$p$11 = $p$5;$pos$2 = $pos$0111;$s$6 = $s$1;$wcs$7 = $wcs$1;
          break L8;
         }
         }
         break;
        }
        default: {
         ___shlim($f,0);
         while(1) {
          $129 = HEAP32[$7>>2]|0;
          $130 = HEAP32[$8>>2]|0;
          $131 = ($129>>>0)<($130>>>0);
          if ($131) {
           $132 = ((($129)) + 1|0);
           HEAP32[$7>>2] = $132;
           $133 = HEAP8[$129>>0]|0;
           $134 = $133&255;
           $136 = $134;
          } else {
           $135 = (___shgetc($f)|0);
           $136 = $135;
          }
          $137 = (_isspace($136)|0);
          $138 = ($137|0)==(0);
          if ($138) {
           break;
          }
         }
         $139 = HEAP32[$8>>2]|0;
         $140 = ($139|0)==(0|0);
         if ($140) {
          $$pre175 = HEAP32[$7>>2]|0;
          $148 = $$pre175;
         } else {
          $141 = HEAP32[$7>>2]|0;
          $142 = ((($141)) + -1|0);
          HEAP32[$7>>2] = $142;
          $143 = $142;
          $148 = $143;
         }
         $144 = HEAP32[$9>>2]|0;
         $145 = HEAP32[$10>>2]|0;
         $146 = (($144) + ($pos$0111))|0;
         $147 = (($146) + ($148))|0;
         $149 = (($147) - ($145))|0;
         $pos$1 = $149;$width$1 = $width$0$lcssa;
        }
        }
        ___shlim($f,$width$1);
        $150 = HEAP32[$7>>2]|0;
        $151 = HEAP32[$8>>2]|0;
        $152 = ($150>>>0)<($151>>>0);
        if ($152) {
         $153 = ((($150)) + 1|0);
         HEAP32[$7>>2] = $153;
         $157 = $151;
        } else {
         $154 = (___shgetc($f)|0);
         $155 = ($154|0)<(0);
         if ($155) {
          $alloc$0402 = $alloc$0;$matches$0107371 = $matches$0107;$s$7 = $s$1;$wcs$8 = $wcs$1;
          label = 154;
          break L6;
         }
         $$pre177 = HEAP32[$8>>2]|0;
         $157 = $$pre177;
        }
        $156 = ($157|0)==(0|0);
        if (!($156)) {
         $158 = HEAP32[$7>>2]|0;
         $159 = ((($158)) + -1|0);
         HEAP32[$7>>2] = $159;
        }
        L68: do {
         switch ($$|0) {
         case 91: case 99: case 115:  {
          $160 = ($$|0)==(99);
          $161 = $$ | 16;
          $162 = ($161|0)==(115);
          L70: do {
           if ($162) {
            $163 = ($$|0)==(115);
            _memset(($scanset|0),-1,257)|0;
            HEAP8[$scanset>>0] = 0;
            if ($163) {
             HEAP8[$12>>0] = 0;
             ;HEAP8[$11>>0]=0|0;HEAP8[$11+1>>0]=0|0;HEAP8[$11+2>>0]=0|0;HEAP8[$11+3>>0]=0|0;HEAP8[$11+4>>0]=0|0;
             $p$9 = $p$5;
            } else {
             $p$9 = $p$5;
            }
           } else {
            $164 = ((($p$5)) + 1|0);
            $165 = HEAP8[$164>>0]|0;
            $166 = ($165<<24>>24)==(94);
            $167 = ((($p$5)) + 2|0);
            $invert$0 = $166&1;
            $p$6 = $166 ? $167 : $164;
            $168 = $166&1;
            _memset(($scanset|0),($168|0),257)|0;
            HEAP8[$scanset>>0] = 0;
            $169 = HEAP8[$p$6>>0]|0;
            switch ($169<<24>>24) {
            case 45:  {
             $170 = ((($p$6)) + 1|0);
             $171 = $invert$0 ^ 1;
             $172 = $171&255;
             HEAP8[$14>>0] = $172;
             $$pre$phi184Z2D = $172;$p$7$ph = $170;
             break;
            }
            case 93:  {
             $173 = ((($p$6)) + 1|0);
             $174 = $invert$0 ^ 1;
             $175 = $174&255;
             HEAP8[$15>>0] = $175;
             $$pre$phi184Z2D = $175;$p$7$ph = $173;
             break;
            }
            default: {
             $$pre182 = $invert$0 ^ 1;
             $$pre183 = $$pre182&255;
             $$pre$phi184Z2D = $$pre183;$p$7$ph = $p$6;
            }
            }
            $p$7 = $p$7$ph;
            while(1) {
             $176 = HEAP8[$p$7>>0]|0;
             L81: do {
              switch ($176<<24>>24) {
              case 0:  {
               $alloc$0402 = $alloc$0;$matches$0107371 = $matches$0107;$s$7 = $s$1;$wcs$8 = $wcs$1;
               label = 154;
               break L6;
               break;
              }
              case 93:  {
               $p$9 = $p$7;
               break L70;
               break;
              }
              case 45:  {
               $177 = ((($p$7)) + 1|0);
               $178 = HEAP8[$177>>0]|0;
               switch ($178<<24>>24) {
               case 93: case 0:  {
                $189 = 45;$p$8 = $p$7;
                break L81;
                break;
               }
               default: {
               }
               }
               $179 = ((($p$7)) + -1|0);
               $180 = HEAP8[$179>>0]|0;
               $181 = ($180&255)<($178&255);
               if ($181) {
                $182 = $180&255;
                $c$0102 = $182;
                while(1) {
                 $183 = (($c$0102) + 1)|0;
                 $184 = (($scanset) + ($183)|0);
                 HEAP8[$184>>0] = $$pre$phi184Z2D;
                 $185 = HEAP8[$177>>0]|0;
                 $186 = $185&255;
                 $187 = ($183|0)<($186|0);
                 if ($187) {
                  $c$0102 = $183;
                 } else {
                  $189 = $185;$p$8 = $177;
                  break;
                 }
                }
               } else {
                $189 = $178;$p$8 = $177;
               }
               break;
              }
              default: {
               $189 = $176;$p$8 = $p$7;
              }
              }
             } while(0);
             $188 = $189&255;
             $190 = (($188) + 1)|0;
             $191 = (($scanset) + ($190)|0);
             HEAP8[$191>>0] = $$pre$phi184Z2D;
             $192 = ((($p$8)) + 1|0);
             $p$7 = $192;
            }
           }
          } while(0);
          $193 = (($width$1) + 1)|0;
          $194 = $160 ? $193 : 31;
          $195 = ($$size$0|0)==(1);
          $196 = ($alloc$0|0)!=(0);
          L89: do {
           if ($195) {
            if ($196) {
             $197 = $194 << 2;
             $198 = (_malloc($197)|0);
             $199 = ($198|0)==(0|0);
             if ($199) {
              $alloc$0402 = $alloc$0;$matches$0107371 = $matches$0107;$s$7 = 0;$wcs$8 = $198;
              label = 154;
              break L6;
             } else {
              $wcs$2 = $198;
             }
            } else {
             $wcs$2 = $dest$0;
            }
            HEAP32[$st>>2] = 0;
            HEAP32[$13>>2] = 0;
            $i$0$ph = 0;$k$0$ph = $194;$wcs$3$ph = $wcs$2;
            L95: while(1) {
             $200 = ($wcs$3$ph|0)==(0|0);
             $i$0$ph22 = $i$0$ph;
             while(1) {
              L99: while(1) {
               $201 = HEAP32[$7>>2]|0;
               $202 = HEAP32[$8>>2]|0;
               $203 = ($201>>>0)<($202>>>0);
               if ($203) {
                $204 = ((($201)) + 1|0);
                HEAP32[$7>>2] = $204;
                $205 = HEAP8[$201>>0]|0;
                $206 = $205&255;
                $209 = $206;
               } else {
                $207 = (___shgetc($f)|0);
                $209 = $207;
               }
               $208 = (($209) + 1)|0;
               $210 = (($scanset) + ($208)|0);
               $211 = HEAP8[$210>>0]|0;
               $212 = ($211<<24>>24)==(0);
               if ($212) {
                $i$0$ph22$lcssa = $i$0$ph22;$wcs$3$ph$lcssa = $wcs$3$ph;
                break L95;
               }
               $213 = $209&255;
               HEAP8[$0>>0] = $213;
               $214 = (_mbrtowc($wc,$0,1,$st)|0);
               switch ($214|0) {
               case -1:  {
                $alloc$0402 = $alloc$0;$matches$0107371 = $matches$0107;$s$7 = 0;$wcs$8 = $wcs$3$ph;
                label = 154;
                break L6;
                break;
               }
               case -2:  {
                break;
               }
               default: {
                break L99;
               }
               }
              }
              if ($200) {
               $i$1 = $i$0$ph22;
              } else {
               $215 = HEAP32[$wc>>2]|0;
               $216 = (($i$0$ph22) + 1)|0;
               $217 = (($wcs$3$ph) + ($i$0$ph22<<2)|0);
               HEAP32[$217>>2] = $215;
               $i$1 = $216;
              }
              $218 = ($i$1|0)==($k$0$ph|0);
              $or$cond = $196 & $218;
              if ($or$cond) {
               break;
              } else {
               $i$0$ph22 = $i$1;
              }
             }
             $factor = $k$0$ph << 1;
             $219 = $factor | 1;
             $220 = $219 << 2;
             $221 = (_realloc($wcs$3$ph,$220)|0);
             $222 = ($221|0)==(0|0);
             if ($222) {
              $alloc$0402 = $alloc$0;$matches$0107371 = $matches$0107;$s$7 = 0;$wcs$8 = $wcs$3$ph;
              label = 154;
              break L6;
             } else {
              $i$0$ph$phi = $k$0$ph;$k$0$ph = $219;$wcs$3$ph = $221;$i$0$ph = $i$0$ph$phi;
             }
            }
            $223 = (_mbsinit($st)|0);
            $224 = ($223|0)==(0);
            if ($224) {
             $alloc$0402 = $alloc$0;$matches$0107371 = $matches$0107;$s$7 = 0;$wcs$8 = $wcs$3$ph$lcssa;
             label = 154;
             break L6;
            } else {
             $i$4 = $i$0$ph22$lcssa;$s$4 = 0;$wcs$5 = $wcs$3$ph$lcssa;
            }
           } else {
            if ($196) {
             $225 = (_malloc($194)|0);
             $226 = ($225|0)==(0|0);
             if ($226) {
              $alloc$0402 = $alloc$0;$matches$0107371 = $matches$0107;$s$7 = 0;$wcs$8 = 0;
              label = 154;
              break L6;
             } else {
              $i$2$ph = 0;$k$1$ph = $194;$s$2$ph = $225;
             }
             while(1) {
              $i$2 = $i$2$ph;
              while(1) {
               $227 = HEAP32[$7>>2]|0;
               $228 = HEAP32[$8>>2]|0;
               $229 = ($227>>>0)<($228>>>0);
               if ($229) {
                $230 = ((($227)) + 1|0);
                HEAP32[$7>>2] = $230;
                $231 = HEAP8[$227>>0]|0;
                $232 = $231&255;
                $235 = $232;
               } else {
                $233 = (___shgetc($f)|0);
                $235 = $233;
               }
               $234 = (($235) + 1)|0;
               $236 = (($scanset) + ($234)|0);
               $237 = HEAP8[$236>>0]|0;
               $238 = ($237<<24>>24)==(0);
               if ($238) {
                $i$4 = $i$2;$s$4 = $s$2$ph;$wcs$5 = 0;
                break L89;
               }
               $239 = $235&255;
               $240 = (($i$2) + 1)|0;
               $241 = (($s$2$ph) + ($i$2)|0);
               HEAP8[$241>>0] = $239;
               $242 = ($240|0)==($k$1$ph|0);
               if ($242) {
                break;
               } else {
                $i$2 = $240;
               }
              }
              $factor18 = $k$1$ph << 1;
              $243 = $factor18 | 1;
              $244 = (_realloc($s$2$ph,$243)|0);
              $245 = ($244|0)==(0|0);
              if ($245) {
               $alloc$0402 = $alloc$0;$matches$0107371 = $matches$0107;$s$7 = $s$2$ph;$wcs$8 = 0;
               label = 154;
               break L6;
              } else {
               $i$2$ph$phi = $k$1$ph;$k$1$ph = $243;$s$2$ph = $244;$i$2$ph = $i$2$ph$phi;
              }
             }
            }
            $246 = ($dest$0|0)==(0|0);
            if ($246) {
             $264 = $157;
             while(1) {
              $262 = HEAP32[$7>>2]|0;
              $263 = ($262>>>0)<($264>>>0);
              if ($263) {
               $265 = ((($262)) + 1|0);
               HEAP32[$7>>2] = $265;
               $266 = HEAP8[$262>>0]|0;
               $267 = $266&255;
               $270 = $267;
              } else {
               $268 = (___shgetc($f)|0);
               $270 = $268;
              }
              $269 = (($270) + 1)|0;
              $271 = (($scanset) + ($269)|0);
              $272 = HEAP8[$271>>0]|0;
              $273 = ($272<<24>>24)==(0);
              if ($273) {
               $i$4 = 0;$s$4 = 0;$wcs$5 = 0;
               break L89;
              }
              $$pre180 = HEAP32[$8>>2]|0;
              $264 = $$pre180;
             }
            } else {
             $249 = $157;$i$3 = 0;
             while(1) {
              $247 = HEAP32[$7>>2]|0;
              $248 = ($247>>>0)<($249>>>0);
              if ($248) {
               $250 = ((($247)) + 1|0);
               HEAP32[$7>>2] = $250;
               $251 = HEAP8[$247>>0]|0;
               $252 = $251&255;
               $255 = $252;
              } else {
               $253 = (___shgetc($f)|0);
               $255 = $253;
              }
              $254 = (($255) + 1)|0;
              $256 = (($scanset) + ($254)|0);
              $257 = HEAP8[$256>>0]|0;
              $258 = ($257<<24>>24)==(0);
              if ($258) {
               $i$4 = $i$3;$s$4 = $dest$0;$wcs$5 = 0;
               break L89;
              }
              $259 = $255&255;
              $260 = (($i$3) + 1)|0;
              $261 = (($dest$0) + ($i$3)|0);
              HEAP8[$261>>0] = $259;
              $$pre179 = HEAP32[$8>>2]|0;
              $249 = $$pre179;$i$3 = $260;
             }
            }
           }
          } while(0);
          $274 = HEAP32[$8>>2]|0;
          $275 = ($274|0)==(0|0);
          if ($275) {
           $$pre181 = HEAP32[$7>>2]|0;
           $282 = $$pre181;
          } else {
           $276 = HEAP32[$7>>2]|0;
           $277 = ((($276)) + -1|0);
           HEAP32[$7>>2] = $277;
           $278 = $277;
           $282 = $278;
          }
          $279 = HEAP32[$9>>2]|0;
          $280 = HEAP32[$10>>2]|0;
          $281 = (($282) - ($280))|0;
          $283 = (($281) + ($279))|0;
          $284 = ($283|0)==(0);
          if ($284) {
           $alloc$2 = $alloc$0;$matches$2 = $matches$0107;$s$9 = $s$4;$wcs$10 = $wcs$5;
           break L6;
          }
          $$not = $160 ^ 1;
          $285 = ($283|0)==($width$1|0);
          $or$cond10 = $285 | $$not;
          if (!($or$cond10)) {
           $alloc$2 = $alloc$0;$matches$2 = $matches$0107;$s$9 = $s$4;$wcs$10 = $wcs$5;
           break L6;
          }
          do {
           if ($196) {
            if ($195) {
             HEAP32[$dest$0>>2] = $wcs$5;
             break;
            } else {
             HEAP32[$dest$0>>2] = $s$4;
             break;
            }
           }
          } while(0);
          if ($160) {
           $p$10 = $p$9;$s$5 = $s$4;$wcs$6 = $wcs$5;
          } else {
           $286 = ($wcs$5|0)==(0|0);
           if (!($286)) {
            $287 = (($wcs$5) + ($i$4<<2)|0);
            HEAP32[$287>>2] = 0;
           }
           $288 = ($s$4|0)==(0|0);
           if ($288) {
            $p$10 = $p$9;$s$5 = 0;$wcs$6 = $wcs$5;
            break L68;
           }
           $289 = (($s$4) + ($i$4)|0);
           HEAP8[$289>>0] = 0;
           $p$10 = $p$9;$s$5 = $s$4;$wcs$6 = $wcs$5;
          }
          break;
         }
         case 120: case 88: case 112:  {
          $base$0 = 16;
          label = 136;
          break;
         }
         case 111:  {
          $base$0 = 8;
          label = 136;
          break;
         }
         case 117: case 100:  {
          $base$0 = 10;
          label = 136;
          break;
         }
         case 105:  {
          $base$0 = 0;
          label = 136;
          break;
         }
         case 71: case 103: case 70: case 102: case 69: case 101: case 65: case 97:  {
          $307 = (+___floatscan($f,$$size$0,0));
          $308 = HEAP32[$9>>2]|0;
          $309 = HEAP32[$7>>2]|0;
          $310 = HEAP32[$10>>2]|0;
          $311 = (($310) - ($309))|0;
          $312 = ($308|0)==($311|0);
          if ($312) {
           $alloc$2 = $alloc$0;$matches$2 = $matches$0107;$s$9 = $s$1;$wcs$10 = $wcs$1;
           break L6;
          }
          $313 = ($dest$0|0)==(0|0);
          if ($313) {
           $p$10 = $p$5;$s$5 = $s$1;$wcs$6 = $wcs$1;
          } else {
           switch ($$size$0|0) {
           case 0:  {
            $314 = $307;
            HEAPF32[$dest$0>>2] = $314;
            $p$10 = $p$5;$s$5 = $s$1;$wcs$6 = $wcs$1;
            break L68;
            break;
           }
           case 1:  {
            HEAPF64[$dest$0>>3] = $307;
            $p$10 = $p$5;$s$5 = $s$1;$wcs$6 = $wcs$1;
            break L68;
            break;
           }
           case 2:  {
            HEAPF64[$dest$0>>3] = $307;
            $p$10 = $p$5;$s$5 = $s$1;$wcs$6 = $wcs$1;
            break L68;
            break;
           }
           default: {
            $p$10 = $p$5;$s$5 = $s$1;$wcs$6 = $wcs$1;
            break L68;
           }
           }
          }
          break;
         }
         default: {
          $p$10 = $p$5;$s$5 = $s$1;$wcs$6 = $wcs$1;
         }
         }
        } while(0);
        L169: do {
         if ((label|0) == 136) {
          label = 0;
          $290 = (___intscan($f,$base$0,0,-1,-1)|0);
          $291 = tempRet0;
          $292 = HEAP32[$9>>2]|0;
          $293 = HEAP32[$7>>2]|0;
          $294 = HEAP32[$10>>2]|0;
          $295 = (($294) - ($293))|0;
          $296 = ($292|0)==($295|0);
          if ($296) {
           $alloc$2 = $alloc$0;$matches$2 = $matches$0107;$s$9 = $s$1;$wcs$10 = $wcs$1;
           break L6;
          }
          $297 = ($$|0)==(112);
          $298 = ($dest$0|0)!=(0|0);
          $or$cond3 = $298 & $297;
          if ($or$cond3) {
           $299 = $290;
           HEAP32[$dest$0>>2] = $299;
           $p$10 = $p$5;$s$5 = $s$1;$wcs$6 = $wcs$1;
           break;
          }
          $300 = ($dest$0|0)==(0|0);
          if ($300) {
           $p$10 = $p$5;$s$5 = $s$1;$wcs$6 = $wcs$1;
          } else {
           switch ($$size$0|0) {
           case -2:  {
            $301 = $290&255;
            HEAP8[$dest$0>>0] = $301;
            $p$10 = $p$5;$s$5 = $s$1;$wcs$6 = $wcs$1;
            break L169;
            break;
           }
           case -1:  {
            $302 = $290&65535;
            HEAP16[$dest$0>>1] = $302;
            $p$10 = $p$5;$s$5 = $s$1;$wcs$6 = $wcs$1;
            break L169;
            break;
           }
           case 0:  {
            HEAP32[$dest$0>>2] = $290;
            $p$10 = $p$5;$s$5 = $s$1;$wcs$6 = $wcs$1;
            break L169;
            break;
           }
           case 1:  {
            HEAP32[$dest$0>>2] = $290;
            $p$10 = $p$5;$s$5 = $s$1;$wcs$6 = $wcs$1;
            break L169;
            break;
           }
           case 3:  {
            $303 = $dest$0;
            $304 = $303;
            HEAP32[$304>>2] = $290;
            $305 = (($303) + 4)|0;
            $306 = $305;
            HEAP32[$306>>2] = $291;
            $p$10 = $p$5;$s$5 = $s$1;$wcs$6 = $wcs$1;
            break L169;
            break;
           }
           default: {
            $p$10 = $p$5;$s$5 = $s$1;$wcs$6 = $wcs$1;
            break L169;
           }
           }
          }
         }
        } while(0);
        $315 = HEAP32[$9>>2]|0;
        $316 = HEAP32[$7>>2]|0;
        $317 = HEAP32[$10>>2]|0;
        $318 = (($315) + ($pos$1))|0;
        $319 = (($318) + ($316))|0;
        $320 = (($319) - ($317))|0;
        $not$ = ($dest$0|0)!=(0|0);
        $321 = $not$&1;
        $matches$0$ = (($321) + ($matches$0107))|0;
        $matches$1 = $matches$0$;$p$11 = $p$10;$pos$2 = $320;$s$6 = $s$5;$wcs$7 = $wcs$6;
        break L8;
       }
      } while(0);
      $50 = $47&1;
      $51 = (($p$0110) + ($50)|0);
      ___shlim($f,0);
      $52 = HEAP32[$7>>2]|0;
      $53 = HEAP32[$8>>2]|0;
      $54 = ($52>>>0)<($53>>>0);
      if ($54) {
       $55 = ((($52)) + 1|0);
       HEAP32[$7>>2] = $55;
       $56 = HEAP8[$52>>0]|0;
       $57 = $56&255;
       $62 = $57;
      } else {
       $58 = (___shgetc($f)|0);
       $62 = $58;
      }
      $59 = HEAP8[$51>>0]|0;
      $60 = $59&255;
      $61 = ($62|0)==($60|0);
      if (!($61)) {
       $$lcssa386 = $62;$matches$0107$lcssa = $matches$0107;$s$0105$lcssa = $s$0105;$wcs$0106$lcssa = $wcs$0106;
       label = 22;
       break L6;
      }
      $69 = (($pos$0111) + 1)|0;
      $matches$1 = $matches$0107;$p$11 = $51;$pos$2 = $69;$s$6 = $s$0105;$wcs$7 = $wcs$0106;
     } else {
      $p$1 = $p$0110;
      while(1) {
       $20 = ((($p$1)) + 1|0);
       $21 = HEAP8[$20>>0]|0;
       $22 = $21&255;
       $23 = (_isspace($22)|0);
       $24 = ($23|0)==(0);
       if ($24) {
        $p$1$lcssa = $p$1;
        break;
       } else {
        $p$1 = $20;
       }
      }
      ___shlim($f,0);
      while(1) {
       $25 = HEAP32[$7>>2]|0;
       $26 = HEAP32[$8>>2]|0;
       $27 = ($25>>>0)<($26>>>0);
       if ($27) {
        $28 = ((($25)) + 1|0);
        HEAP32[$7>>2] = $28;
        $29 = HEAP8[$25>>0]|0;
        $30 = $29&255;
        $32 = $30;
       } else {
        $31 = (___shgetc($f)|0);
        $32 = $31;
       }
       $33 = (_isspace($32)|0);
       $34 = ($33|0)==(0);
       if ($34) {
        break;
       }
      }
      $35 = HEAP32[$8>>2]|0;
      $36 = ($35|0)==(0|0);
      if ($36) {
       $$pre = HEAP32[$7>>2]|0;
       $44 = $$pre;
      } else {
       $37 = HEAP32[$7>>2]|0;
       $38 = ((($37)) + -1|0);
       HEAP32[$7>>2] = $38;
       $39 = $38;
       $44 = $39;
      }
      $40 = HEAP32[$9>>2]|0;
      $41 = HEAP32[$10>>2]|0;
      $42 = (($40) + ($pos$0111))|0;
      $43 = (($42) + ($44))|0;
      $45 = (($43) - ($41))|0;
      $matches$1 = $matches$0107;$p$11 = $p$1$lcssa;$pos$2 = $45;$s$6 = $s$0105;$wcs$7 = $wcs$0106;
     }
    } while(0);
    $322 = ((($p$11)) + 1|0);
    $323 = HEAP8[$322>>0]|0;
    $324 = ($323<<24>>24)==(0);
    if ($324) {
     $matches$3 = $matches$1;
     break L4;
    } else {
     $17 = $323;$matches$0107 = $matches$1;$p$0110 = $322;$pos$0111 = $pos$2;$s$0105 = $s$6;$wcs$0106 = $wcs$7;
    }
   }
   if ((label|0) == 22) {
    $63 = HEAP32[$8>>2]|0;
    $64 = ($63|0)==(0|0);
    if (!($64)) {
     $65 = HEAP32[$7>>2]|0;
     $66 = ((($65)) + -1|0);
     HEAP32[$7>>2] = $66;
    }
    $67 = ($$lcssa386|0)>(-1);
    $68 = ($matches$0107$lcssa|0)!=(0);
    $or$cond5 = $68 | $67;
    if ($or$cond5) {
     $matches$3 = $matches$0107$lcssa;
     break;
    } else {
     $alloc$1 = 0;$s$8 = $s$0105$lcssa;$wcs$9 = $wcs$0106$lcssa;
     label = 155;
    }
   }
   else if ((label|0) == 154) {
    $$old4 = ($matches$0107371|0)==(0);
    if ($$old4) {
     $alloc$1 = $alloc$0402;$s$8 = $s$7;$wcs$9 = $wcs$8;
     label = 155;
    } else {
     $alloc$2 = $alloc$0402;$matches$2 = $matches$0107371;$s$9 = $s$7;$wcs$10 = $wcs$8;
    }
   }
   if ((label|0) == 155) {
    $alloc$2 = $alloc$1;$matches$2 = -1;$s$9 = $s$8;$wcs$10 = $wcs$9;
   }
   $325 = ($alloc$2|0)==(0);
   if ($325) {
    $matches$3 = $matches$2;
   } else {
    _free($s$9);
    _free($wcs$10);
    $matches$3 = $matches$2;
   }
  }
 } while(0);
 $326 = ($327|0)==(0);
 if (!($326)) {
  ___unlockfile($f);
 }
 STACKTOP = sp;return ($matches$3|0);
}
function _mbsinit($st) {
 $st = $st|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($st|0)==(0|0);
 if ($0) {
  $4 = 1;
 } else {
  $1 = HEAP32[$st>>2]|0;
  $2 = ($1|0)==(0);
  $4 = $2;
 }
 $3 = $4&1;
 return ($3|0);
}
function _sscanf($s,$fmt,$varargs) {
 $s = $s|0;
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $0 = 0, $ap = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ap = sp;
 HEAP32[$ap>>2] = $varargs;
 $0 = (_vsscanf($s,$fmt,$ap)|0);
 STACKTOP = sp;return ($0|0);
}
function _vsscanf($s,$fmt,$ap) {
 $s = $s|0;
 $fmt = $fmt|0;
 $ap = $ap|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $f = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f = sp;
 dest=$f; stop=dest+112|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $0 = ((($f)) + 32|0);
 HEAP32[$0>>2] = 4851;
 $1 = ((($f)) + 44|0);
 HEAP32[$1>>2] = $s;
 $2 = ((($f)) + 76|0);
 HEAP32[$2>>2] = -1;
 $3 = ((($f)) + 84|0);
 HEAP32[$3>>2] = $s;
 $4 = (_vfscanf($f,$fmt,$ap)|0);
 STACKTOP = sp;return ($4|0);
}
function _do_read($f,$buf,$len) {
 $f = $f|0;
 $buf = $buf|0;
 $len = $len|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (___string_read($f,$buf,$len)|0);
 return ($0|0);
}
function ___string_read($f,$buf,$len) {
 $f = $f|0;
 $buf = $buf|0;
 $len = $len|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $k$0 = 0, $k$0$len = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 84|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($len) + 256)|0;
 $3 = (_memchr($1,0,$2)|0);
 $4 = ($3|0)==(0|0);
 $5 = $3;
 $6 = $1;
 $7 = (($5) - ($6))|0;
 $k$0 = $4 ? $2 : $7;
 $8 = ($k$0>>>0)<($len>>>0);
 $k$0$len = $8 ? $k$0 : $len;
 _memcpy(($buf|0),($1|0),($k$0$len|0))|0;
 $9 = (($1) + ($k$0$len)|0);
 $10 = ((($f)) + 4|0);
 HEAP32[$10>>2] = $9;
 $11 = (($1) + ($k$0)|0);
 $12 = ((($f)) + 8|0);
 HEAP32[$12>>2] = $11;
 HEAP32[$0>>2] = $11;
 return ($k$0$len|0);
}
function _fseek($f,$off,$whence) {
 $f = $f|0;
 $off = $off|0;
 $whence = $whence|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (___fseeko($f,$off,$whence)|0);
 return ($0|0);
}
function ___fseeko($f,$off,$whence) {
 $f = $f|0;
 $off = $off|0;
 $whence = $whence|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 76|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(-1);
 if ($2) {
  $4 = (___lockfile($f)|0);
  $phitmp = ($4|0)==(0);
  $5 = (___fseeko_unlocked($f,$off,$whence)|0);
  if ($phitmp) {
   $6 = $5;
  } else {
   ___unlockfile($f);
   $6 = $5;
  }
 } else {
  $3 = (___fseeko_unlocked($f,$off,$whence)|0);
  $6 = $3;
 }
 return ($6|0);
}
function ___fseeko_unlocked($f,$off,$whence) {
 $f = $f|0;
 $off = $off|0;
 $whence = $whence|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($whence|0)==(1);
 if ($0) {
  $1 = ((($f)) + 8|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ((($f)) + 4|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = (($off) - ($2))|0;
  $6 = (($5) + ($4))|0;
  $$01 = $6;
 } else {
  $$01 = $off;
 }
 $7 = ((($f)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($f)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($8>>>0)>($10>>>0);
 if ($11) {
  $12 = ((($f)) + 36|0);
  $13 = HEAP32[$12>>2]|0;
  (FUNCTION_TABLE_iiii[$13 & 8191]($f,0,0)|0);
  $14 = HEAP32[$7>>2]|0;
  $15 = ($14|0)==(0|0);
  if ($15) {
   $$0 = -1;
  } else {
   label = 5;
  }
 } else {
  label = 5;
 }
 if ((label|0) == 5) {
  $16 = ((($f)) + 16|0);
  HEAP32[$16>>2] = 0;
  HEAP32[$9>>2] = 0;
  HEAP32[$7>>2] = 0;
  $17 = ((($f)) + 40|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = (FUNCTION_TABLE_iiii[$18 & 8191]($f,$$01,$whence)|0);
  $20 = ($19|0)<(0);
  if ($20) {
   $$0 = -1;
  } else {
   $21 = ((($f)) + 8|0);
   HEAP32[$21>>2] = 0;
   $22 = ((($f)) + 4|0);
   HEAP32[$22>>2] = 0;
   $23 = HEAP32[$f>>2]|0;
   $24 = $23 & -17;
   HEAP32[$f>>2] = $24;
   $$0 = 0;
  }
 }
 return ($$0|0);
}
function _fprintf($f,$fmt,$varargs) {
 $f = $f|0;
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $0 = 0, $ap = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ap = sp;
 HEAP32[$ap>>2] = $varargs;
 $0 = (_vfprintf($f,$fmt,$ap)|0);
 STACKTOP = sp;return ($0|0);
}
function _strstr($h,$n) {
 $h = $h|0;
 $n = $n|0;
 var $$0 = 0, $$0$lcssa$i = 0, $$0$lcssa$i12 = 0, $$0$lcssa$i4 = 0, $$01$i = 0, $$02$i = 0, $$02$i8 = 0, $$03$i = 0, $$3$i = 0, $$lcssa$i = 0, $$lcssa$i11 = 0, $$lcssa291 = 0, $$lcssa294 = 0, $$lcssa297 = 0, $$lcssa311 = 0, $$lcssa314 = 0, $$lcssa317 = 0, $$lcssa332 = 0, $$pr$i = 0, $0 = 0;
 var $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $byteset$i = 0;
 var $div$i = 0, $div5$i = 0, $hw$0$in2$i = 0, $hw$03$i = 0, $hw$03$i7 = 0, $ip$0$ph$lcssa$i = 0, $ip$0$ph$lcssa145$i = 0, $ip$0$ph78$i = 0, $ip$1$ip$0$$i = 0, $ip$1$ip$0$i = 0, $ip$1$ph$lcssa$i = 0, $ip$1$ph57$i = 0, $jp$0$ph15$ph72$i = 0, $jp$0$ph1567$i = 0, $jp$0$ph1567$i$lcssa = 0, $jp$0$ph1567$i$lcssa$lcssa = 0, $jp$0$ph79$i = 0, $jp$1$ph11$ph51$i = 0, $jp$1$ph1146$i = 0, $jp$1$ph1146$i$lcssa = 0;
 var $jp$1$ph1146$i$lcssa$lcssa = 0, $jp$1$ph58$i = 0, $k$061$i = 0, $k$141$i = 0, $k$2$i = 0, $k$340$i = 0, $k$340$i$lcssa = 0, $k$4$i = 0, $l$082$i = 0, $l$082$i$lcssa331 = 0, $mem$0$i = 0, $mem0$0$i = 0, $or$cond$i = 0, $or$cond$i2 = 0, $or$cond$i9 = 0, $or$cond6$i = 0, $p$0$ph$ph$lcssa34$i = 0, $p$0$ph$ph$lcssa34146$i = 0, $p$0$ph$ph73$i = 0, $p$1$p$0$i = 0;
 var $p$1$ph$ph$lcssa25$i = 0, $p$1$ph$ph52$i = 0, $p$3$i = 0, $shift$i = 0, $z$0$i = 0, $z$3$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1056|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $byteset$i = sp + 1024|0;
 $shift$i = sp;
 $0 = HEAP8[$n>>0]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $$0 = $h;
  } else {
   $2 = $0 << 24 >> 24;
   $3 = (_strchr($h,$2)|0);
   $4 = ($3|0)==(0|0);
   if ($4) {
    $$0 = 0;
   } else {
    $5 = ((($n)) + 1|0);
    $6 = HEAP8[$5>>0]|0;
    $7 = ($6<<24>>24)==(0);
    if ($7) {
     $$0 = $3;
    } else {
     $8 = ((($3)) + 1|0);
     $9 = HEAP8[$8>>0]|0;
     $10 = ($9<<24>>24)==(0);
     if ($10) {
      $$0 = 0;
     } else {
      $11 = ((($n)) + 2|0);
      $12 = HEAP8[$11>>0]|0;
      $13 = ($12<<24>>24)==(0);
      if ($13) {
       $14 = $0&255;
       $15 = $14 << 8;
       $16 = $6&255;
       $17 = $16 | $15;
       $18 = HEAP8[$3>>0]|0;
       $19 = $18&255;
       $20 = $19 << 8;
       $21 = $9&255;
       $22 = $20 | $21;
       $$01$i = $8;$233 = $9;$hw$0$in2$i = $22;
       while(1) {
        $23 = $hw$0$in2$i & 65535;
        $24 = ($23|0)==($17|0);
        if ($24) {
         $$0$lcssa$i = $$01$i;$32 = $233;
         break;
        }
        $25 = $23 << 8;
        $26 = ((($$01$i)) + 1|0);
        $27 = HEAP8[$26>>0]|0;
        $28 = $27&255;
        $29 = $28 | $25;
        $30 = ($27<<24>>24)==(0);
        if ($30) {
         $$0$lcssa$i = $26;$32 = 0;
         break;
        } else {
         $$01$i = $26;$233 = $27;$hw$0$in2$i = $29;
        }
       }
       $31 = ($32<<24>>24)!=(0);
       $33 = ((($$0$lcssa$i)) + -1|0);
       $34 = $31 ? $33 : 0;
       $$0 = $34;
       break;
      }
      $35 = ((($3)) + 2|0);
      $36 = HEAP8[$35>>0]|0;
      $37 = ($36<<24>>24)==(0);
      if ($37) {
       $$0 = 0;
      } else {
       $38 = ((($n)) + 3|0);
       $39 = HEAP8[$38>>0]|0;
       $40 = ($39<<24>>24)==(0);
       if ($40) {
        $41 = $0&255;
        $42 = $41 << 24;
        $43 = $6&255;
        $44 = $43 << 16;
        $45 = $44 | $42;
        $46 = $12&255;
        $47 = $46 << 8;
        $48 = $45 | $47;
        $49 = HEAP8[$3>>0]|0;
        $50 = $49&255;
        $51 = $50 << 24;
        $52 = $9&255;
        $53 = $52 << 16;
        $54 = $36&255;
        $55 = $54 << 8;
        $56 = $55 | $53;
        $57 = $56 | $51;
        $58 = ($57|0)==($48|0);
        if ($58) {
         $$0$lcssa$i4 = $35;$$lcssa$i = $36;
        } else {
         $$02$i = $35;$hw$03$i = $57;
         while(1) {
          $59 = ((($$02$i)) + 1|0);
          $60 = HEAP8[$59>>0]|0;
          $61 = $60&255;
          $62 = $61 | $hw$03$i;
          $63 = $62 << 8;
          $64 = ($60<<24>>24)==(0);
          $65 = ($63|0)==($48|0);
          $or$cond$i2 = $64 | $65;
          if ($or$cond$i2) {
           $$0$lcssa$i4 = $59;$$lcssa$i = $60;
           break;
          } else {
           $$02$i = $59;$hw$03$i = $63;
          }
         }
        }
        $66 = ($$lcssa$i<<24>>24)!=(0);
        $67 = ((($$0$lcssa$i4)) + -2|0);
        $68 = $66 ? $67 : 0;
        $$0 = $68;
        break;
       }
       $69 = ((($3)) + 3|0);
       $70 = HEAP8[$69>>0]|0;
       $71 = ($70<<24>>24)==(0);
       if ($71) {
        $$0 = 0;
       } else {
        $72 = ((($n)) + 4|0);
        $73 = HEAP8[$72>>0]|0;
        $74 = ($73<<24>>24)==(0);
        if ($74) {
         $75 = $0&255;
         $76 = $75 << 24;
         $77 = $6&255;
         $78 = $77 << 16;
         $79 = $78 | $76;
         $80 = $12&255;
         $81 = $80 << 8;
         $82 = $79 | $81;
         $83 = $39&255;
         $84 = $82 | $83;
         $85 = HEAP8[$3>>0]|0;
         $86 = $85&255;
         $87 = $86 << 24;
         $88 = $9&255;
         $89 = $88 << 16;
         $90 = $36&255;
         $91 = $90 << 8;
         $92 = $70&255;
         $93 = $91 | $89;
         $94 = $93 | $92;
         $95 = $94 | $87;
         $96 = ($95|0)==($84|0);
         if ($96) {
          $$0$lcssa$i12 = $69;$$lcssa$i11 = $70;
         } else {
          $$02$i8 = $69;$hw$03$i7 = $95;
          while(1) {
           $97 = $hw$03$i7 << 8;
           $98 = ((($$02$i8)) + 1|0);
           $99 = HEAP8[$98>>0]|0;
           $100 = $99&255;
           $101 = $100 | $97;
           $102 = ($99<<24>>24)==(0);
           $103 = ($101|0)==($84|0);
           $or$cond$i9 = $102 | $103;
           if ($or$cond$i9) {
            $$0$lcssa$i12 = $98;$$lcssa$i11 = $99;
            break;
           } else {
            $$02$i8 = $98;$hw$03$i7 = $101;
           }
          }
         }
         $104 = ($$lcssa$i11<<24>>24)!=(0);
         $105 = ((($$0$lcssa$i12)) + -3|0);
         $106 = $104 ? $105 : 0;
         $$0 = $106;
         break;
        }
        ;HEAP32[$byteset$i>>2]=0|0;HEAP32[$byteset$i+4>>2]=0|0;HEAP32[$byteset$i+8>>2]=0|0;HEAP32[$byteset$i+12>>2]=0|0;HEAP32[$byteset$i+16>>2]=0|0;HEAP32[$byteset$i+20>>2]=0|0;HEAP32[$byteset$i+24>>2]=0|0;HEAP32[$byteset$i+28>>2]=0|0;
        $111 = $0;$l$082$i = 0;
        while(1) {
         $107 = (($3) + ($l$082$i)|0);
         $108 = HEAP8[$107>>0]|0;
         $109 = ($108<<24>>24)==(0);
         if ($109) {
          $$3$i = 0;
          break;
         }
         $110 = $111 & 31;
         $112 = $110&255;
         $113 = 1 << $112;
         $div5$i = ($111&255) >>> 5;
         $114 = $div5$i&255;
         $115 = (($byteset$i) + ($114<<2)|0);
         $116 = HEAP32[$115>>2]|0;
         $117 = $116 | $113;
         HEAP32[$115>>2] = $117;
         $118 = (($l$082$i) + 1)|0;
         $119 = $111&255;
         $120 = (($shift$i) + ($119<<2)|0);
         HEAP32[$120>>2] = $118;
         $121 = (($n) + ($118)|0);
         $122 = HEAP8[$121>>0]|0;
         $123 = ($122<<24>>24)==(0);
         if ($123) {
          $$lcssa332 = $118;$l$082$i$lcssa331 = $l$082$i;
          label = 23;
          break;
         } else {
          $111 = $122;$l$082$i = $118;
         }
        }
        L32: do {
         if ((label|0) == 23) {
          $124 = ($$lcssa332>>>0)>(1);
          L34: do {
           if ($124) {
            $234 = 1;$ip$0$ph78$i = -1;$jp$0$ph79$i = 0;
            L35: while(1) {
             $235 = $234;$jp$0$ph15$ph72$i = $jp$0$ph79$i;$p$0$ph$ph73$i = 1;
             while(1) {
              $236 = $235;$jp$0$ph1567$i = $jp$0$ph15$ph72$i;
              L39: while(1) {
               $134 = $236;$k$061$i = 1;
               while(1) {
                $130 = (($k$061$i) + ($ip$0$ph78$i))|0;
                $131 = (($n) + ($130)|0);
                $132 = HEAP8[$131>>0]|0;
                $133 = (($n) + ($134)|0);
                $135 = HEAP8[$133>>0]|0;
                $136 = ($132<<24>>24)==($135<<24>>24);
                if (!($136)) {
                 $$lcssa311 = $134;$$lcssa314 = $132;$$lcssa317 = $135;$jp$0$ph1567$i$lcssa = $jp$0$ph1567$i;
                 break L39;
                }
                $137 = ($k$061$i|0)==($p$0$ph$ph73$i|0);
                $128 = (($k$061$i) + 1)|0;
                if ($137) {
                 break;
                }
                $127 = (($128) + ($jp$0$ph1567$i))|0;
                $129 = ($127>>>0)<($$lcssa332>>>0);
                if ($129) {
                 $134 = $127;$k$061$i = $128;
                } else {
                 $ip$0$ph$lcssa$i = $ip$0$ph78$i;$p$0$ph$ph$lcssa34$i = $p$0$ph$ph73$i;
                 break L35;
                }
               }
               $138 = (($jp$0$ph1567$i) + ($p$0$ph$ph73$i))|0;
               $139 = (($138) + 1)|0;
               $140 = ($139>>>0)<($$lcssa332>>>0);
               if ($140) {
                $236 = $139;$jp$0$ph1567$i = $138;
               } else {
                $ip$0$ph$lcssa$i = $ip$0$ph78$i;$p$0$ph$ph$lcssa34$i = $p$0$ph$ph73$i;
                break L35;
               }
              }
              $141 = ($$lcssa314&255)>($$lcssa317&255);
              $142 = (($$lcssa311) - ($ip$0$ph78$i))|0;
              if (!($141)) {
               $jp$0$ph1567$i$lcssa$lcssa = $jp$0$ph1567$i$lcssa;
               break;
              }
              $125 = (($$lcssa311) + 1)|0;
              $126 = ($125>>>0)<($$lcssa332>>>0);
              if ($126) {
               $235 = $125;$jp$0$ph15$ph72$i = $$lcssa311;$p$0$ph$ph73$i = $142;
              } else {
               $ip$0$ph$lcssa$i = $ip$0$ph78$i;$p$0$ph$ph$lcssa34$i = $142;
               break L35;
              }
             }
             $143 = (($jp$0$ph1567$i$lcssa$lcssa) + 1)|0;
             $144 = (($jp$0$ph1567$i$lcssa$lcssa) + 2)|0;
             $145 = ($144>>>0)<($$lcssa332>>>0);
             if ($145) {
              $234 = $144;$ip$0$ph78$i = $jp$0$ph1567$i$lcssa$lcssa;$jp$0$ph79$i = $143;
             } else {
              $ip$0$ph$lcssa$i = $jp$0$ph1567$i$lcssa$lcssa;$p$0$ph$ph$lcssa34$i = 1;
              break;
             }
            }
            $237 = 1;$ip$1$ph57$i = -1;$jp$1$ph58$i = 0;
            while(1) {
             $239 = $237;$jp$1$ph11$ph51$i = $jp$1$ph58$i;$p$1$ph$ph52$i = 1;
             while(1) {
              $238 = $239;$jp$1$ph1146$i = $jp$1$ph11$ph51$i;
              L54: while(1) {
               $153 = $238;$k$141$i = 1;
               while(1) {
                $149 = (($k$141$i) + ($ip$1$ph57$i))|0;
                $150 = (($n) + ($149)|0);
                $151 = HEAP8[$150>>0]|0;
                $152 = (($n) + ($153)|0);
                $154 = HEAP8[$152>>0]|0;
                $155 = ($151<<24>>24)==($154<<24>>24);
                if (!($155)) {
                 $$lcssa291 = $153;$$lcssa294 = $151;$$lcssa297 = $154;$jp$1$ph1146$i$lcssa = $jp$1$ph1146$i;
                 break L54;
                }
                $156 = ($k$141$i|0)==($p$1$ph$ph52$i|0);
                $147 = (($k$141$i) + 1)|0;
                if ($156) {
                 break;
                }
                $146 = (($147) + ($jp$1$ph1146$i))|0;
                $148 = ($146>>>0)<($$lcssa332>>>0);
                if ($148) {
                 $153 = $146;$k$141$i = $147;
                } else {
                 $ip$0$ph$lcssa145$i = $ip$0$ph$lcssa$i;$ip$1$ph$lcssa$i = $ip$1$ph57$i;$p$0$ph$ph$lcssa34146$i = $p$0$ph$ph$lcssa34$i;$p$1$ph$ph$lcssa25$i = $p$1$ph$ph52$i;
                 break L34;
                }
               }
               $157 = (($jp$1$ph1146$i) + ($p$1$ph$ph52$i))|0;
               $158 = (($157) + 1)|0;
               $159 = ($158>>>0)<($$lcssa332>>>0);
               if ($159) {
                $238 = $158;$jp$1$ph1146$i = $157;
               } else {
                $ip$0$ph$lcssa145$i = $ip$0$ph$lcssa$i;$ip$1$ph$lcssa$i = $ip$1$ph57$i;$p$0$ph$ph$lcssa34146$i = $p$0$ph$ph$lcssa34$i;$p$1$ph$ph$lcssa25$i = $p$1$ph$ph52$i;
                break L34;
               }
              }
              $160 = ($$lcssa294&255)<($$lcssa297&255);
              $161 = (($$lcssa291) - ($ip$1$ph57$i))|0;
              if (!($160)) {
               $jp$1$ph1146$i$lcssa$lcssa = $jp$1$ph1146$i$lcssa;
               break;
              }
              $165 = (($$lcssa291) + 1)|0;
              $166 = ($165>>>0)<($$lcssa332>>>0);
              if ($166) {
               $239 = $165;$jp$1$ph11$ph51$i = $$lcssa291;$p$1$ph$ph52$i = $161;
              } else {
               $ip$0$ph$lcssa145$i = $ip$0$ph$lcssa$i;$ip$1$ph$lcssa$i = $ip$1$ph57$i;$p$0$ph$ph$lcssa34146$i = $p$0$ph$ph$lcssa34$i;$p$1$ph$ph$lcssa25$i = $161;
               break L34;
              }
             }
             $162 = (($jp$1$ph1146$i$lcssa$lcssa) + 1)|0;
             $163 = (($jp$1$ph1146$i$lcssa$lcssa) + 2)|0;
             $164 = ($163>>>0)<($$lcssa332>>>0);
             if ($164) {
              $237 = $163;$ip$1$ph57$i = $jp$1$ph1146$i$lcssa$lcssa;$jp$1$ph58$i = $162;
             } else {
              $ip$0$ph$lcssa145$i = $ip$0$ph$lcssa$i;$ip$1$ph$lcssa$i = $jp$1$ph1146$i$lcssa$lcssa;$p$0$ph$ph$lcssa34146$i = $p$0$ph$ph$lcssa34$i;$p$1$ph$ph$lcssa25$i = 1;
              break;
             }
            }
           } else {
            $ip$0$ph$lcssa145$i = -1;$ip$1$ph$lcssa$i = -1;$p$0$ph$ph$lcssa34146$i = 1;$p$1$ph$ph$lcssa25$i = 1;
           }
          } while(0);
          $167 = (($ip$1$ph$lcssa$i) + 1)|0;
          $168 = (($ip$0$ph$lcssa145$i) + 1)|0;
          $169 = ($167>>>0)>($168>>>0);
          $p$1$p$0$i = $169 ? $p$1$ph$ph$lcssa25$i : $p$0$ph$ph$lcssa34146$i;
          $ip$1$ip$0$i = $169 ? $ip$1$ph$lcssa$i : $ip$0$ph$lcssa145$i;
          $170 = (($n) + ($p$1$p$0$i)|0);
          $171 = (($ip$1$ip$0$i) + 1)|0;
          $172 = (_memcmp($n,$170,$171)|0);
          $173 = ($172|0)==(0);
          if ($173) {
           $178 = (($$lcssa332) - ($p$1$p$0$i))|0;
           $mem0$0$i = $178;$p$3$i = $p$1$p$0$i;
          } else {
           $174 = (($$lcssa332) - ($ip$1$ip$0$i))|0;
           $175 = (($174) + -1)|0;
           $176 = ($ip$1$ip$0$i>>>0)>($175>>>0);
           $ip$1$ip$0$$i = $176 ? $ip$1$ip$0$i : $175;
           $177 = (($ip$1$ip$0$$i) + 1)|0;
           $mem0$0$i = 0;$p$3$i = $177;
          }
          $179 = $$lcssa332 | 63;
          $180 = ($mem0$0$i|0)!=(0);
          $181 = (($$lcssa332) - ($p$3$i))|0;
          $$03$i = $3;$mem$0$i = 0;$z$0$i = $3;
          L69: while(1) {
           $182 = $z$0$i;
           $183 = $$03$i;
           $184 = (($182) - ($183))|0;
           $185 = ($184>>>0)<($$lcssa332>>>0);
           do {
            if ($185) {
             $186 = (_memchr($z$0$i,0,$179)|0);
             $187 = ($186|0)==(0|0);
             if ($187) {
              $191 = (($z$0$i) + ($179)|0);
              $z$3$i = $191;
              break;
             } else {
              $188 = $186;
              $189 = (($188) - ($183))|0;
              $190 = ($189>>>0)<($$lcssa332>>>0);
              if ($190) {
               $$3$i = 0;
               break L32;
              } else {
               $z$3$i = $186;
               break;
              }
             }
            } else {
             $z$3$i = $z$0$i;
            }
           } while(0);
           $192 = (($$03$i) + ($l$082$i$lcssa331)|0);
           $193 = HEAP8[$192>>0]|0;
           $div$i = ($193&255) >>> 5;
           $194 = $div$i&255;
           $195 = (($byteset$i) + ($194<<2)|0);
           $196 = HEAP32[$195>>2]|0;
           $197 = $193 & 31;
           $198 = $197&255;
           $199 = 1 << $198;
           $200 = $199 & $196;
           $201 = ($200|0)==(0);
           if ($201) {
            $210 = (($$03$i) + ($$lcssa332)|0);
            $$03$i = $210;$mem$0$i = 0;$z$0$i = $z$3$i;
            continue;
           }
           $202 = $193&255;
           $203 = (($shift$i) + ($202<<2)|0);
           $204 = HEAP32[$203>>2]|0;
           $205 = (($$lcssa332) - ($204))|0;
           $206 = ($$lcssa332|0)==($204|0);
           if (!($206)) {
            $207 = ($mem$0$i|0)!=(0);
            $or$cond$i = $180 & $207;
            $208 = ($205>>>0)<($p$3$i>>>0);
            $or$cond6$i = $or$cond$i & $208;
            $k$2$i = $or$cond6$i ? $181 : $205;
            $209 = (($$03$i) + ($k$2$i)|0);
            $$03$i = $209;$mem$0$i = 0;$z$0$i = $z$3$i;
            continue;
           }
           $211 = ($171>>>0)>($mem$0$i>>>0);
           $212 = $211 ? $171 : $mem$0$i;
           $213 = (($n) + ($212)|0);
           $214 = HEAP8[$213>>0]|0;
           $215 = ($214<<24>>24)==(0);
           L83: do {
            if ($215) {
             $k$4$i = $171;
            } else {
             $$pr$i = $214;$k$340$i = $212;
             while(1) {
              $216 = (($$03$i) + ($k$340$i)|0);
              $217 = HEAP8[$216>>0]|0;
              $218 = ($$pr$i<<24>>24)==($217<<24>>24);
              if (!($218)) {
               $k$340$i$lcssa = $k$340$i;
               break;
              }
              $219 = (($k$340$i) + 1)|0;
              $220 = (($n) + ($219)|0);
              $221 = HEAP8[$220>>0]|0;
              $222 = ($221<<24>>24)==(0);
              if ($222) {
               $k$4$i = $171;
               break L83;
              } else {
               $$pr$i = $221;$k$340$i = $219;
              }
             }
             $223 = (($k$340$i$lcssa) - ($ip$1$ip$0$i))|0;
             $224 = (($$03$i) + ($223)|0);
             $$03$i = $224;$mem$0$i = 0;$z$0$i = $z$3$i;
             continue L69;
            }
           } while(0);
           while(1) {
            $225 = ($k$4$i>>>0)>($mem$0$i>>>0);
            if (!($225)) {
             $$3$i = $$03$i;
             break L32;
            }
            $226 = (($k$4$i) + -1)|0;
            $227 = (($n) + ($226)|0);
            $228 = HEAP8[$227>>0]|0;
            $229 = (($$03$i) + ($226)|0);
            $230 = HEAP8[$229>>0]|0;
            $231 = ($228<<24>>24)==($230<<24>>24);
            if ($231) {
             $k$4$i = $226;
            } else {
             break;
            }
           }
           $232 = (($$03$i) + ($p$3$i)|0);
           $$03$i = $232;$mem$0$i = $mem0$0$i;$z$0$i = $z$3$i;
          }
         }
        } while(0);
        $$0 = $$3$i;
       }
      }
     }
    }
   }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _lstat($path,$buf) {
 $path = $path|0;
 $buf = $buf|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 HEAP32[$vararg_buffer>>2] = $path;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $buf;
 $0 = (___syscall196(196,($vararg_buffer|0))|0);
 $1 = (___syscall_ret($0)|0);
 STACKTOP = sp;return ($1|0);
}
function _access($filename,$amode) {
 $filename = $filename|0;
 $amode = $amode|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 HEAP32[$vararg_buffer>>2] = $filename;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $amode;
 $0 = (___syscall33(33,($vararg_buffer|0))|0);
 $1 = (___syscall_ret($0)|0);
 STACKTOP = sp;return ($1|0);
}
function _strrchr($s,$c) {
 $s = $s|0;
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strlen($s)|0);
 $1 = (($0) + 1)|0;
 $2 = (___memrchr($s,$c,$1)|0);
 return ($2|0);
}
function ___memrchr($m,$c,$n) {
 $m = $m|0;
 $c = $c|0;
 $n = $n|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $c&255;
 $$01 = $n;
 while(1) {
  $1 = (($$01) + -1)|0;
  $2 = ($$01|0)==(0);
  if ($2) {
   $$0 = 0;
   break;
  }
  $3 = (($m) + ($1)|0);
  $4 = HEAP8[$3>>0]|0;
  $5 = ($4<<24>>24)==($0<<24>>24);
  if ($5) {
   $$0 = $3;
   break;
  } else {
   $$01 = $1;
  }
 }
 return ($$0|0);
}
function _realpath($filename,$resolved) {
 $filename = $filename|0;
 $resolved = $resolved|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buf = 0, $st1 = 0, $st2 = 0, $tmp = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_buffer5 = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 4304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 16|0;
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $st1 = sp + 96|0;
 $st2 = sp + 20|0;
 $buf = sp + 4272|0;
 $tmp = sp + 176|0;
 $0 = ($filename|0)==(0|0);
 do {
  if ($0) {
   $1 = (___errno_location()|0);
   HEAP32[$1>>2] = 22;
   $$0 = 0;
  } else {
   HEAP32[$vararg_buffer>>2] = $filename;
   $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
   HEAP32[$vararg_ptr1>>2] = 2656256;
   $2 = (___syscall5(5,($vararg_buffer|0))|0);
   $3 = (___syscall_ret($2)|0);
   $4 = ($3|0)<(0);
   if ($4) {
    $$0 = 0;
   } else {
    ___procfdname($buf,$3);
    $5 = (_readlink($buf,$tmp,4095)|0);
    $6 = ($5|0)<(0);
    if (!($6)) {
     $7 = (($tmp) + ($5)|0);
     HEAP8[$7>>0] = 0;
     (_fstat($3,$st1)|0);
     $8 = (_stat($tmp,$st2)|0);
     $9 = ($8|0)<(0);
     if (!($9)) {
      $10 = HEAP32[$st1>>2]|0;
      $11 = HEAP32[$st2>>2]|0;
      $12 = ($10|0)==($11|0);
      if ($12) {
       $13 = ((($st1)) + 72|0);
       $14 = HEAP32[$13>>2]|0;
       $15 = ((($st2)) + 72|0);
       $16 = HEAP32[$15>>2]|0;
       $17 = ($14|0)==($16|0);
       if ($17) {
        HEAP32[$vararg_buffer2>>2] = $3;
        (___syscall6(6,($vararg_buffer2|0))|0);
        $20 = ($resolved|0)==(0|0);
        if ($20) {
         $22 = (___strdup($tmp)|0);
         $$0 = $22;
         break;
        } else {
         $21 = (_strcpy($resolved,$tmp)|0);
         $$0 = $21;
         break;
        }
       }
      }
      $18 = ($8|0)==(0);
      if ($18) {
       $19 = (___errno_location()|0);
       HEAP32[$19>>2] = 40;
      }
     }
    }
    HEAP32[$vararg_buffer5>>2] = $3;
    (___syscall6(6,($vararg_buffer5|0))|0);
    $$0 = 0;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _readlink($path,$buf,$bufsize) {
 $path = $path|0;
 $buf = $buf|0;
 $bufsize = $bufsize|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 HEAP32[$vararg_buffer>>2] = $path;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $buf;
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
 HEAP32[$vararg_ptr2>>2] = $bufsize;
 $0 = (___syscall85(85,($vararg_buffer|0))|0);
 $1 = (___syscall_ret($0)|0);
 STACKTOP = sp;return ($1|0);
}
function _socket($domain,$type,$protocol) {
 $domain = $domain|0;
 $type = $type|0;
 $protocol = $protocol|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer11 = 0, $vararg_buffer2 = 0, $vararg_buffer6 = 0, $vararg_buffer9 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer11 = sp + 32|0;
 $vararg_buffer9 = sp + 24|0;
 $vararg_buffer6 = sp + 16|0;
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 64|0;
 $1 = sp + 40|0;
 HEAP32[$0>>2] = $domain;
 $2 = ((($0)) + 4|0);
 HEAP32[$2>>2] = $type;
 $3 = ((($0)) + 8|0);
 HEAP32[$3>>2] = $protocol;
 $4 = ((($0)) + 12|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($0)) + 16|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($0)) + 20|0);
 HEAP32[$6>>2] = 0;
 HEAP32[$vararg_buffer>>2] = 1;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $0;
 $7 = (___syscall102(102,($vararg_buffer|0))|0);
 $8 = (___syscall_ret($7)|0);
 $9 = ($8|0)<(0);
 L1: do {
  if ($9) {
   $10 = (___errno_location()|0);
   $11 = HEAP32[$10>>2]|0;
   switch ($11|0) {
   case 93: case 22:  {
    break;
   }
   default: {
    $$0 = $8;
    break L1;
   }
   }
   $12 = $type & 526336;
   $13 = ($12|0)==(0);
   if ($13) {
    $$0 = $8;
   } else {
    HEAP32[$1>>2] = $domain;
    $14 = ((($1)) + 4|0);
    $15 = $type & -526337;
    HEAP32[$14>>2] = $15;
    $16 = ((($1)) + 8|0);
    HEAP32[$16>>2] = $protocol;
    $17 = ((($1)) + 12|0);
    HEAP32[$17>>2] = 0;
    $18 = ((($1)) + 16|0);
    HEAP32[$18>>2] = 0;
    $19 = ((($1)) + 20|0);
    HEAP32[$19>>2] = 0;
    HEAP32[$vararg_buffer2>>2] = 1;
    $vararg_ptr5 = ((($vararg_buffer2)) + 4|0);
    HEAP32[$vararg_ptr5>>2] = $1;
    $20 = (___syscall102(102,($vararg_buffer2|0))|0);
    $21 = (___syscall_ret($20)|0);
    $22 = ($21|0)<(0);
    if ($22) {
     $$0 = $21;
    } else {
     $23 = $type & 524288;
     $24 = ($23|0)==(0);
     if (!($24)) {
      HEAP32[$vararg_buffer6>>2] = 1;
      (_fcntl($21,2,$vararg_buffer6)|0);
     }
     $25 = $type & 2048;
     $26 = ($25|0)==(0);
     if ($26) {
      $$0 = $21;
     } else {
      $27 = (_fcntl($21,3,$vararg_buffer9)|0);
      $28 = $27 | 2048;
      HEAP32[$vararg_buffer11>>2] = $28;
      (_fcntl($21,4,$vararg_buffer11)|0);
      $$0 = $21;
     }
    }
   }
  } else {
   $$0 = $8;
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _connect($fd,$addr,$len) {
 $fd = $fd|0;
 $addr = $addr|0;
 $len = $len|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 HEAP32[$0>>2] = $fd;
 $1 = ((($0)) + 4|0);
 $2 = $addr;
 HEAP32[$1>>2] = $2;
 $3 = ((($0)) + 8|0);
 HEAP32[$3>>2] = $len;
 $4 = ((($0)) + 12|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($0)) + 16|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($0)) + 20|0);
 HEAP32[$6>>2] = 0;
 HEAP32[$vararg_buffer>>2] = 3;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $0;
 $7 = (___syscall102(102,($vararg_buffer|0))|0);
 $8 = (___syscall_ret($7)|0);
 STACKTOP = sp;return ($8|0);
}
function _send($fd,$buf,$len,$flags) {
 $fd = $fd|0;
 $buf = $buf|0;
 $len = $len|0;
 $flags = $flags|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_sendto($fd,$buf,$len,$flags,0,0)|0);
 return ($0|0);
}
function _sendto($fd,$buf,$len,$flags,$addr,$alen) {
 $fd = $fd|0;
 $buf = $buf|0;
 $len = $len|0;
 $flags = $flags|0;
 $addr = $addr|0;
 $alen = $alen|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 HEAP32[$0>>2] = $fd;
 $1 = ((($0)) + 4|0);
 $2 = $buf;
 HEAP32[$1>>2] = $2;
 $3 = ((($0)) + 8|0);
 HEAP32[$3>>2] = $len;
 $4 = ((($0)) + 12|0);
 HEAP32[$4>>2] = $flags;
 $5 = ((($0)) + 16|0);
 $6 = $addr;
 HEAP32[$5>>2] = $6;
 $7 = ((($0)) + 20|0);
 HEAP32[$7>>2] = $alen;
 HEAP32[$vararg_buffer>>2] = 11;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $0;
 $8 = (___syscall102(102,($vararg_buffer|0))|0);
 $9 = (___syscall_ret($8)|0);
 STACKTOP = sp;return ($9|0);
}
function _mbrlen($s,$n,$st) {
 $s = $s|0;
 $n = $n|0;
 $st = $st|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($st|0)!=(0|0);
 $1 = $0 ? $st : 1206008;
 $2 = (_mbrtowc(0,$s,$n,$1)|0);
 return ($2|0);
}
function _mbsnrtowcs($wcs,$src,$n,$wn,$st) {
 $wcs = $wcs|0;
 $src = $src|0;
 $n = $n|0;
 $wn = $wn|0;
 $st = $st|0;
 var $$019 = 0, $$02$ = 0, $$0218 = 0, $$1 = 0, $$13 = 0, $$214 = 0, $$24 = 0, $$313 = 0, $$cast = 0, $$lcssa = 0, $$lcssa57 = 0, $$pr$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $cnt$020 = 0, $cnt$1 = 0, $cnt$215 = 0, $cnt$215$lcssa = 0, $cnt$3 = 0, $or$cond = 0, $or$cond17 = 0, $s = 0, $wbuf = 0, $wcs$ = 0;
 var $wn$ = 0, $ws$010 = 0, $ws$021 = 0, $ws$021$lcssa56 = 0, $ws$1 = 0, $ws$216 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1040|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $wbuf = sp + 8|0;
 $s = sp;
 $0 = HEAP32[$src>>2]|0;
 HEAP32[$s>>2] = $0;
 $1 = ($wcs|0)!=(0|0);
 $wn$ = $1 ? $wn : 256;
 $wcs$ = $1 ? $wcs : $wbuf;
 $$cast = $0;
 $2 = ($0|0)!=(0);
 $3 = ($wn$|0)!=(0);
 $or$cond17 = $3 & $2;
 L1: do {
  if ($or$cond17) {
   $$019 = $n;$$0218 = $wn$;$37 = $$cast;$cnt$020 = 0;$ws$021 = $wcs$;
   while(1) {
    $4 = $$019 >>> 2;
    $5 = ($4>>>0)>=($$0218>>>0);
    $6 = ($$019>>>0)>(131);
    $7 = $6 | $5;
    if (!($7)) {
     $$1 = $$019;$$24 = $$0218;$19 = $37;$cnt$1 = $cnt$020;$ws$010 = $ws$021;
     break L1;
    }
    $$02$ = $5 ? $$0218 : $4;
    $8 = (($$019) - ($$02$))|0;
    $9 = (_mbsrtowcs($ws$021,$s,$$02$,$st)|0);
    $10 = ($9|0)==(-1);
    if ($10) {
     $$lcssa57 = $8;$ws$021$lcssa56 = $ws$021;
     break;
    }
    $11 = ($ws$021|0)==($wbuf|0);
    $12 = (($ws$021) + ($9<<2)|0);
    $13 = $11 ? 0 : $9;
    $$13 = (($$0218) - ($13))|0;
    $ws$1 = $11 ? $ws$021 : $12;
    $14 = (($9) + ($cnt$020))|0;
    $15 = HEAP32[$s>>2]|0;
    $16 = ($15|0)!=(0|0);
    $17 = ($$0218|0)!=($13|0);
    $or$cond = $17 & $16;
    if ($or$cond) {
     $$019 = $8;$$0218 = $$13;$37 = $15;$cnt$020 = $14;$ws$021 = $ws$1;
    } else {
     $$1 = $8;$$24 = $$13;$19 = $15;$cnt$1 = $14;$ws$010 = $ws$1;
     break L1;
    }
   }
   $$pr$pre = HEAP32[$s>>2]|0;
   $$1 = $$lcssa57;$$24 = 0;$19 = $$pr$pre;$cnt$1 = -1;$ws$010 = $ws$021$lcssa56;
  } else {
   $$1 = $n;$$24 = $wn$;$19 = $$cast;$cnt$1 = 0;$ws$010 = $wcs$;
  }
 } while(0);
 $18 = ($19|0)==(0|0);
 L8: do {
  if ($18) {
   $cnt$3 = $cnt$1;
  } else {
   $20 = ($$24|0)!=(0);
   $21 = ($$1|0)!=(0);
   $22 = $20 & $21;
   if ($22) {
    $$214 = $$1;$$313 = $$24;$23 = $19;$cnt$215 = $cnt$1;$ws$216 = $ws$010;
    while(1) {
     $24 = (_mbrtowc($ws$216,$23,$$214,$st)|0);
     $25 = (($24) + 2)|0;
     $26 = ($25>>>0)<(3);
     if ($26) {
      $$lcssa = $24;$cnt$215$lcssa = $cnt$215;
      break;
     }
     $27 = HEAP32[$s>>2]|0;
     $28 = (($27) + ($24)|0);
     HEAP32[$s>>2] = $28;
     $29 = (($$214) - ($24))|0;
     $30 = ((($ws$216)) + 4|0);
     $31 = (($$313) + -1)|0;
     $32 = (($cnt$215) + 1)|0;
     $33 = ($31|0)!=(0);
     $34 = ($$214|0)!=($24|0);
     $35 = $33 & $34;
     if ($35) {
      $$214 = $29;$$313 = $31;$23 = $28;$cnt$215 = $32;$ws$216 = $30;
     } else {
      $cnt$3 = $32;
      break L8;
     }
    }
    switch ($$lcssa|0) {
    case -1:  {
     $cnt$3 = -1;
     break L8;
     break;
    }
    case 0:  {
     HEAP32[$s>>2] = 0;
     $cnt$3 = $cnt$215$lcssa;
     break L8;
     break;
    }
    default: {
     HEAP32[$st>>2] = 0;
     $cnt$3 = $cnt$215$lcssa;
     break L8;
    }
    }
   } else {
    $cnt$3 = $cnt$1;
   }
  }
 } while(0);
 if ($1) {
  $36 = HEAP32[$s>>2]|0;
  HEAP32[$src>>2] = $36;
 }
 STACKTOP = sp;return ($cnt$3|0);
}
function _mbsrtowcs($ws,$src,$wn,$st) {
 $ws = $ws|0;
 $src = $src|0;
 $wn = $wn|0;
 $st = $st|0;
 var $$0 = 0, $$02$ph = 0, $$0219 = 0, $$03 = 0, $$1 = 0, $$1$lcssa = 0, $$1415 = 0, $$2 = 0, $$2$lcssa = 0, $$25 = 0, $$25$lcssa = 0, $$3 = 0, $$36 = 0, $$4 = 0, $$434 = 0, $$47$ph = 0, $$4718 = 0, $$48 = 0, $$5 = 0, $$5$lcssa = 0;
 var $$6 = 0, $$6$lcssa = 0, $$7 = 0, $$8 = 0, $$833 = 0, $$lcssa = 0, $$lcssa102 = 0, $$lcssa88 = 0, $$lcssa89 = 0, $$lcssa90 = 0, $$lcssa94 = 0, $$lcssa96 = 0, $$lcssa97 = 0, $$lcssa98 = 0, $$old1 = 0, $$pre$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $c$2 = 0, $c$4 = 0, $c$5 = 0, $c$6 = 0, $or$cond = 0, $s$0 = 0, $s$10 = 0, $s$1035 = 0, $s$109 = 0, $s$116 = 0, $s$2 = 0, $s$2$lcssa = 0, $s$3 = 0, $s$4 = 0, $s$5$lcssa = 0;
 var $s$5$ph = 0, $s$520 = 0, $s$6 = 0, $s$6$lcssa = 0, $s$7 = 0, $s$7$lcssa = 0, $s$8 = 0, $s$9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$src>>2]|0;
 $1 = ($st|0)==(0|0);
 if ($1) {
  label = 5;
 } else {
  $2 = HEAP32[$st>>2]|0;
  $3 = ($2|0)==(0);
  if ($3) {
   label = 5;
  } else {
   $4 = ($ws|0)==(0|0);
   if ($4) {
    $$36 = $wn;$c$2 = $2;$s$3 = $0;
    label = 16;
   } else {
    HEAP32[$st>>2] = 0;
    $$3 = $ws;$$7 = $wn;$c$4 = $2;$s$8 = $0;
    label = 37;
   }
  }
 }
 if ((label|0) == 5) {
  $5 = ($ws|0)==(0|0);
  if ($5) {
   $$03 = $wn;$s$0 = $0;
   label = 7;
  } else {
   $$02$ph = $ws;$$47$ph = $wn;$s$5$ph = $0;
   label = 6;
  }
 }
 L7: while(1) {
  if ((label|0) == 6) {
   label = 0;
   $6 = ($$47$ph|0)==(0);
   if ($6) {
    $s$5$lcssa = $s$5$ph;
    label = 26;
    break;
   } else {
    $$0219 = $$02$ph;$$4718 = $$47$ph;$s$520 = $s$5$ph;
   }
   while(1) {
    $64 = HEAP8[$s$520>>0]|0;
    $65 = $64&255;
    $66 = (($65) + -1)|0;
    $67 = ($66>>>0)<(127);
    do {
     if ($67) {
      $68 = $s$520;
      $69 = $68 & 3;
      $70 = ($69|0)==(0);
      $71 = ($$4718>>>0)>(4);
      $or$cond = $71 & $70;
      if ($or$cond) {
       $$1 = $$0219;$$5 = $$4718;$s$6 = $s$520;
       while(1) {
        $72 = HEAP32[$s$6>>2]|0;
        $73 = (($72) + -16843009)|0;
        $74 = $73 | $72;
        $75 = $74 & -2139062144;
        $76 = ($75|0)==(0);
        if (!($76)) {
         $$1$lcssa = $$1;$$5$lcssa = $$5;$$lcssa94 = $72;$s$6$lcssa = $s$6;
         label = 32;
         break;
        }
        $77 = ((($s$6)) + 1|0);
        $78 = $72 & 255;
        $79 = ((($$1)) + 4|0);
        HEAP32[$$1>>2] = $78;
        $80 = ((($s$6)) + 2|0);
        $81 = HEAP8[$77>>0]|0;
        $82 = $81&255;
        $83 = ((($$1)) + 8|0);
        HEAP32[$79>>2] = $82;
        $84 = ((($s$6)) + 3|0);
        $85 = HEAP8[$80>>0]|0;
        $86 = $85&255;
        $87 = ((($$1)) + 12|0);
        HEAP32[$83>>2] = $86;
        $88 = ((($s$6)) + 4|0);
        $89 = HEAP8[$84>>0]|0;
        $90 = $89&255;
        $91 = ((($$1)) + 16|0);
        HEAP32[$87>>2] = $90;
        $92 = (($$5) + -4)|0;
        $$old1 = ($92>>>0)>(4);
        if ($$old1) {
         $$1 = $91;$$5 = $92;$s$6 = $88;
        } else {
         $$lcssa96 = $88;$$lcssa97 = $91;$$lcssa98 = $92;
         label = 31;
         break;
        }
       }
       if ((label|0) == 31) {
        label = 0;
        $$pre$pre = HEAP8[$$lcssa96>>0]|0;
        $$2 = $$lcssa97;$$6 = $$lcssa98;$95 = $$pre$pre;$s$7 = $$lcssa96;
        break;
       }
       else if ((label|0) == 32) {
        label = 0;
        $93 = $$lcssa94&255;
        $$2 = $$1$lcssa;$$6 = $$5$lcssa;$95 = $93;$s$7 = $s$6$lcssa;
        break;
       }
      } else {
       $$2 = $$0219;$$6 = $$4718;$95 = $64;$s$7 = $s$520;
      }
     } else {
      $$2 = $$0219;$$6 = $$4718;$95 = $64;$s$7 = $s$520;
     }
    } while(0);
    $94 = $95&255;
    $96 = (($94) + -1)|0;
    $97 = ($96>>>0)<(127);
    if (!($97)) {
     $$2$lcssa = $$2;$$6$lcssa = $$6;$$lcssa102 = $94;$s$7$lcssa = $s$7;
     break;
    }
    $98 = ((($s$7)) + 1|0);
    $99 = ((($$2)) + 4|0);
    HEAP32[$$2>>2] = $94;
    $100 = (($$6) + -1)|0;
    $101 = ($100|0)==(0);
    if ($101) {
     $s$5$lcssa = $98;
     label = 26;
     break L7;
    } else {
     $$0219 = $99;$$4718 = $100;$s$520 = $98;
    }
   }
   $102 = (($$lcssa102) + -194)|0;
   $103 = ($102>>>0)>(50);
   if ($103) {
    $$434 = $$2$lcssa;$$833 = $$6$lcssa;$s$1035 = $s$7$lcssa;
    label = 48;
    break;
   }
   $104 = ((($s$7$lcssa)) + 1|0);
   $105 = (149972 + ($102<<2)|0);
   $106 = HEAP32[$105>>2]|0;
   $$3 = $$2$lcssa;$$7 = $$6$lcssa;$c$4 = $106;$s$8 = $104;
   label = 37;
   continue;
  }
  else if ((label|0) == 7) {
   label = 0;
   $7 = HEAP8[$s$0>>0]|0;
   $8 = $7&255;
   $9 = (($8) + -1)|0;
   $10 = ($9>>>0)<(127);
   if ($10) {
    $11 = $s$0;
    $12 = $11 & 3;
    $13 = ($12|0)==(0);
    if ($13) {
     $14 = HEAP32[$s$0>>2]|0;
     $15 = (($14) + -16843009)|0;
     $16 = $15 | $14;
     $17 = $16 & -2139062144;
     $18 = ($17|0)==(0);
     $19 = $14&255;
     if ($18) {
      $$1415 = $$03;$s$116 = $s$0;
      while(1) {
       $20 = ((($s$116)) + 4|0);
       $21 = (($$1415) + -4)|0;
       $22 = HEAP32[$20>>2]|0;
       $23 = (($22) + -16843009)|0;
       $24 = $23 | $22;
       $25 = $24 & -2139062144;
       $26 = ($25|0)==(0);
       if ($26) {
        $$1415 = $21;$s$116 = $20;
       } else {
        $$lcssa = $20;$$lcssa88 = $21;$$lcssa89 = $22;
        break;
       }
      }
      $27 = $$lcssa89&255;
      $$25 = $$lcssa88;$29 = $27;$s$2 = $$lcssa;
     } else {
      $$25 = $$03;$29 = $19;$s$2 = $s$0;
     }
    } else {
     $$25 = $$03;$29 = $7;$s$2 = $s$0;
    }
   } else {
    $$25 = $$03;$29 = $7;$s$2 = $s$0;
   }
   $28 = $29&255;
   $30 = (($28) + -1)|0;
   $31 = ($30>>>0)<(127);
   if ($31) {
    $32 = ((($s$2)) + 1|0);
    $33 = (($$25) + -1)|0;
    $$03 = $33;$s$0 = $32;
    label = 7;
    continue;
   } else {
    $$25$lcssa = $$25;$$lcssa90 = $28;$s$2$lcssa = $s$2;
   }
   $34 = (($$lcssa90) + -194)|0;
   $35 = ($34>>>0)>(50);
   if ($35) {
    $$434 = $ws;$$833 = $$25$lcssa;$s$1035 = $s$2$lcssa;
    label = 48;
    break;
   }
   $36 = ((($s$2$lcssa)) + 1|0);
   $37 = (149972 + ($34<<2)|0);
   $38 = HEAP32[$37>>2]|0;
   $$36 = $$25$lcssa;$c$2 = $38;$s$3 = $36;
   label = 16;
   continue;
  }
  else if ((label|0) == 16) {
   label = 0;
   $39 = HEAP8[$s$3>>0]|0;
   $40 = $39&255;
   $41 = $40 >>> 3;
   $42 = (($41) + -16)|0;
   $43 = $c$2 >> 26;
   $44 = (($41) + ($43))|0;
   $45 = $42 | $44;
   $46 = ($45>>>0)>(7);
   if ($46) {
    label = 17;
    break;
   }
   $48 = ((($s$3)) + 1|0);
   $49 = $c$2 & 33554432;
   $50 = ($49|0)==(0);
   if ($50) {
    $s$4 = $48;
   } else {
    $51 = HEAP8[$48>>0]|0;
    $52 = $51 & -64;
    $53 = ($52<<24>>24)==(-128);
    if (!($53)) {
     label = 20;
     break;
    }
    $55 = ((($s$3)) + 2|0);
    $56 = $c$2 & 524288;
    $57 = ($56|0)==(0);
    if ($57) {
     $s$4 = $55;
    } else {
     $58 = HEAP8[$55>>0]|0;
     $59 = $58 & -64;
     $60 = ($59<<24>>24)==(-128);
     if (!($60)) {
      label = 23;
      break;
     }
     $62 = ((($s$3)) + 3|0);
     $s$4 = $62;
    }
   }
   $63 = (($$36) + -1)|0;
   $$03 = $63;$s$0 = $s$4;
   label = 7;
   continue;
  }
  else if ((label|0) == 37) {
   label = 0;
   $107 = HEAP8[$s$8>>0]|0;
   $108 = $107&255;
   $109 = $108 >>> 3;
   $110 = (($109) + -16)|0;
   $111 = $c$4 >> 26;
   $112 = (($109) + ($111))|0;
   $113 = $110 | $112;
   $114 = ($113>>>0)>(7);
   if ($114) {
    label = 38;
    break;
   }
   $116 = $c$4 << 6;
   $117 = ((($s$8)) + 1|0);
   $118 = (($108) + -128)|0;
   $119 = $118 | $116;
   $120 = ($119|0)<(0);
   if ($120) {
    $121 = HEAP8[$117>>0]|0;
    $122 = $121&255;
    $123 = $122 & 192;
    $124 = ($123|0)==(128);
    if (!($124)) {
     label = 41;
     break;
    }
    $126 = $119 << 6;
    $127 = ((($s$8)) + 2|0);
    $128 = (($122) + -128)|0;
    $129 = $128 | $126;
    $130 = ($129|0)<(0);
    if ($130) {
     $131 = HEAP8[$127>>0]|0;
     $132 = $131&255;
     $133 = $132 & 192;
     $134 = ($133|0)==(128);
     if (!($134)) {
      label = 44;
      break;
     }
     $136 = $129 << 6;
     $137 = ((($s$8)) + 3|0);
     $138 = (($132) + -128)|0;
     $139 = $138 | $136;
     $c$5 = $139;$s$9 = $137;
    } else {
     $c$5 = $129;$s$9 = $127;
    }
   } else {
    $c$5 = $119;$s$9 = $117;
   }
   $140 = ((($$3)) + 4|0);
   HEAP32[$$3>>2] = $c$5;
   $141 = (($$7) + -1)|0;
   $$02$ph = $140;$$47$ph = $141;$s$5$ph = $s$9;
   label = 6;
   continue;
  }
 }
 if ((label|0) == 17) {
  $47 = ((($s$3)) + -1|0);
  $$4 = $ws;$$8 = $$36;$c$6 = $c$2;$s$10 = $47;
  label = 47;
 }
 else if ((label|0) == 20) {
  $54 = ((($s$3)) + -1|0);
  $$4 = $ws;$$8 = $$36;$c$6 = $c$2;$s$10 = $54;
  label = 47;
 }
 else if ((label|0) == 23) {
  $61 = ((($s$3)) + -1|0);
  $$4 = $ws;$$8 = $$36;$c$6 = $c$2;$s$10 = $61;
  label = 47;
 }
 else if ((label|0) == 26) {
  HEAP32[$src>>2] = $s$5$lcssa;
  $$0 = $wn;
 }
 else if ((label|0) == 38) {
  $115 = ((($s$8)) + -1|0);
  $$4 = $$3;$$8 = $$7;$c$6 = $c$4;$s$10 = $115;
  label = 47;
 }
 else if ((label|0) == 41) {
  $125 = ((($s$8)) + -1|0);
  $$48 = $$3;$s$109 = $125;
  label = 52;
 }
 else if ((label|0) == 44) {
  $135 = ((($s$8)) + -1|0);
  $$48 = $$3;$s$109 = $135;
  label = 52;
 }
 if ((label|0) == 47) {
  $142 = ($c$6|0)==(0);
  if ($142) {
   $$434 = $$4;$$833 = $$8;$s$1035 = $s$10;
   label = 48;
  } else {
   $$48 = $$4;$s$109 = $s$10;
   label = 52;
  }
 }
 if ((label|0) == 48) {
  $143 = HEAP8[$s$1035>>0]|0;
  $144 = ($143<<24>>24)==(0);
  if ($144) {
   $145 = ($$434|0)==(0|0);
   if (!($145)) {
    HEAP32[$$434>>2] = 0;
    HEAP32[$src>>2] = 0;
   }
   $146 = (($wn) - ($$833))|0;
   $$0 = $146;
  } else {
   $$48 = $$434;$s$109 = $s$1035;
   label = 52;
  }
 }
 if ((label|0) == 52) {
  $147 = (___errno_location()|0);
  HEAP32[$147>>2] = 84;
  $148 = ($$48|0)==(0|0);
  if ($148) {
   $$0 = -1;
  } else {
   HEAP32[$src>>2] = $s$109;
   $$0 = -1;
  }
 }
 return ($$0|0);
}
function _wcsnrtombs($dst,$wcs,$wn,$n,$st) {
 $dst = $dst|0;
 $wcs = $wcs|0;
 $wn = $wn|0;
 $n = $n|0;
 $st = $st|0;
 var $$019 = 0, $$02$$0 = 0, $$0218 = 0, $$1 = 0, $$13 = 0, $$214 = 0, $$24 = 0, $$313 = 0, $$cast = 0, $$lcssa = 0, $$lcssa57 = 0, $$pr$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buf = 0, $cnt$020 = 0, $cnt$1 = 0, $cnt$215 = 0, $cnt$215$lcssa = 0, $cnt$3 = 0, $dst$ = 0, $n$ = 0, $or$cond = 0;
 var $or$cond17 = 0, $s$010 = 0, $s$021 = 0, $s$021$lcssa56 = 0, $s$1 = 0, $s$216 = 0, $ws = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 272|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $buf = sp + 8|0;
 $ws = sp;
 $0 = HEAP32[$wcs>>2]|0;
 HEAP32[$ws>>2] = $0;
 $1 = ($dst|0)!=(0|0);
 $n$ = $1 ? $n : 256;
 $dst$ = $1 ? $dst : $buf;
 $$cast = $0;
 $2 = ($0|0)!=(0);
 $3 = ($n$|0)!=(0);
 $or$cond17 = $3 & $2;
 L1: do {
  if ($or$cond17) {
   $$019 = $wn;$$0218 = $n$;$38 = $$cast;$cnt$020 = 0;$s$021 = $dst$;
   while(1) {
    $4 = ($$019>>>0)>=($$0218>>>0);
    $5 = ($$019>>>0)>(32);
    $6 = $4 | $5;
    if (!($6)) {
     $$1 = $$019;$$24 = $$0218;$18 = $38;$cnt$1 = $cnt$020;$s$010 = $s$021;
     break L1;
    }
    $$02$$0 = $4 ? $$0218 : $$019;
    $7 = (($$019) - ($$02$$0))|0;
    $8 = (_wcsrtombs($s$021,$ws,$$02$$0,0)|0);
    $9 = ($8|0)==(-1);
    if ($9) {
     $$lcssa57 = $7;$s$021$lcssa56 = $s$021;
     break;
    }
    $10 = ($s$021|0)==($buf|0);
    $11 = (($s$021) + ($8)|0);
    $12 = $10 ? 0 : $8;
    $$13 = (($$0218) - ($12))|0;
    $s$1 = $10 ? $s$021 : $11;
    $13 = (($8) + ($cnt$020))|0;
    $14 = HEAP32[$ws>>2]|0;
    $15 = ($14|0)!=(0|0);
    $16 = ($$0218|0)!=($12|0);
    $or$cond = $16 & $15;
    if ($or$cond) {
     $$019 = $7;$$0218 = $$13;$38 = $14;$cnt$020 = $13;$s$021 = $s$1;
    } else {
     $$1 = $7;$$24 = $$13;$18 = $14;$cnt$1 = $13;$s$010 = $s$1;
     break L1;
    }
   }
   $$pr$pre = HEAP32[$ws>>2]|0;
   $$1 = $$lcssa57;$$24 = 0;$18 = $$pr$pre;$cnt$1 = -1;$s$010 = $s$021$lcssa56;
  } else {
   $$1 = $wn;$$24 = $n$;$18 = $$cast;$cnt$1 = 0;$s$010 = $dst$;
  }
 } while(0);
 $17 = ($18|0)==(0|0);
 L8: do {
  if ($17) {
   $cnt$3 = $cnt$1;
  } else {
   $19 = ($$24|0)!=(0);
   $20 = ($$1|0)!=(0);
   $21 = $19 & $20;
   if ($21) {
    $$214 = $$1;$$313 = $$24;$23 = $18;$cnt$215 = $cnt$1;$s$216 = $s$010;
    while(1) {
     $22 = HEAP32[$23>>2]|0;
     $24 = (_wcrtomb($s$216,$22,0)|0);
     $25 = (($24) + 1)|0;
     $26 = ($25>>>0)<(2);
     if ($26) {
      $$lcssa = $24;$cnt$215$lcssa = $cnt$215;
      break;
     }
     $28 = HEAP32[$ws>>2]|0;
     $29 = ((($28)) + 4|0);
     HEAP32[$ws>>2] = $29;
     $30 = (($$214) + -1)|0;
     $31 = (($s$216) + ($24)|0);
     $32 = (($$313) - ($24))|0;
     $33 = (($cnt$215) + 1)|0;
     $34 = ($$313|0)!=($24|0);
     $35 = ($30|0)!=(0);
     $36 = $34 & $35;
     if ($36) {
      $$214 = $30;$$313 = $32;$23 = $29;$cnt$215 = $33;$s$216 = $31;
     } else {
      $cnt$3 = $33;
      break L8;
     }
    }
    $27 = ($$lcssa|0)==(0);
    if ($27) {
     HEAP32[$ws>>2] = 0;
     $cnt$3 = $cnt$215$lcssa;
    } else {
     $cnt$3 = -1;
    }
   } else {
    $cnt$3 = $cnt$1;
   }
  }
 } while(0);
 if ($1) {
  $37 = HEAP32[$ws>>2]|0;
  HEAP32[$wcs>>2] = $37;
 }
 STACKTOP = sp;return ($cnt$3|0);
}
function _wcsrtombs($s,$ws,$n,$st) {
 $s = $s|0;
 $ws = $ws|0;
 $n = $n|0;
 $st = $st|0;
 var $$0 = 0, $$01$lcssa = 0, $$0129 = 0, $$0129$lcssa = 0, $$0218 = 0, $$1 = 0, $$13 = 0, $$223 = 0, $$223$lcssa = 0, $$24$lcssa = 0, $$2428 = 0, $$2428$lcssa = 0, $$3 = 0, $$35 = 0, $$422 = 0, $$422$lcssa = 0, $$422$lcssa83 = 0, $$5 = 0, $$pn = 0, $$pre = 0;
 var $$pre44 = 0, $$pre45 = 0, $$pre46 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buf = 0, $ws2$019 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $buf = sp;
 $0 = ($s|0)==(0|0);
 L1: do {
  if ($0) {
   $2 = HEAP32[$ws>>2]|0;
   $3 = HEAP32[$2>>2]|0;
   $4 = ($3|0)==(0);
   if ($4) {
    $$0 = 0;
   } else {
    $$0218 = 0;$6 = $3;$ws2$019 = $2;
    while(1) {
     $5 = ($6>>>0)>(127);
     if ($5) {
      $7 = (_wcrtomb($buf,$6,0)|0);
      $8 = ($7|0)==(-1);
      if ($8) {
       $$0 = -1;
       break L1;
      } else {
       $$pn = $7;
      }
     } else {
      $$pn = 1;
     }
     $$13 = (($$pn) + ($$0218))|0;
     $9 = ((($ws2$019)) + 4|0);
     $10 = HEAP32[$9>>2]|0;
     $11 = ($10|0)==(0);
     if ($11) {
      $$0 = $$13;
      break;
     } else {
      $$0218 = $$13;$6 = $10;$ws2$019 = $9;
     }
    }
   }
  } else {
   $1 = ($n>>>0)>(3);
   L9: do {
    if ($1) {
     $$pre = HEAP32[$ws>>2]|0;
     $$0129 = $s;$$2428 = $n;$14 = $$pre;
     while(1) {
      $13 = HEAP32[$14>>2]|0;
      $15 = (($13) + -1)|0;
      $16 = ($15>>>0)>(126);
      if ($16) {
       $17 = ($13|0)==(0);
       if ($17) {
        $$0129$lcssa = $$0129;$$2428$lcssa = $$2428;
        break;
       }
       $19 = (_wcrtomb($$0129,$13,0)|0);
       $20 = ($19|0)==(-1);
       if ($20) {
        $$0 = -1;
        break L1;
       }
       $21 = (($$0129) + ($19)|0);
       $22 = (($$2428) - ($19))|0;
       $$1 = $21;$$35 = $22;$27 = $14;
      } else {
       $23 = $13&255;
       $24 = ((($$0129)) + 1|0);
       HEAP8[$$0129>>0] = $23;
       $25 = (($$2428) + -1)|0;
       $$pre44 = HEAP32[$ws>>2]|0;
       $$1 = $24;$$35 = $25;$27 = $$pre44;
      }
      $26 = ((($27)) + 4|0);
      HEAP32[$ws>>2] = $26;
      $28 = ($$35>>>0)>(3);
      if ($28) {
       $$0129 = $$1;$$2428 = $$35;$14 = $26;
      } else {
       $$01$lcssa = $$1;$$24$lcssa = $$35;
       break L9;
      }
     }
     HEAP8[$$0129$lcssa>>0] = 0;
     HEAP32[$ws>>2] = 0;
     $18 = (($n) - ($$2428$lcssa))|0;
     $$0 = $18;
     break L1;
    } else {
     $$01$lcssa = $s;$$24$lcssa = $n;
    }
   } while(0);
   $12 = ($$24$lcssa|0)==(0);
   if ($12) {
    $$0 = $n;
   } else {
    $$pre45 = HEAP32[$ws>>2]|0;
    $$223 = $$01$lcssa;$$422 = $$24$lcssa;$30 = $$pre45;
    while(1) {
     $29 = HEAP32[$30>>2]|0;
     $31 = (($29) + -1)|0;
     $32 = ($31>>>0)>(126);
     if ($32) {
      $33 = ($29|0)==(0);
      if ($33) {
       $$223$lcssa = $$223;$$422$lcssa = $$422;
       label = 19;
       break;
      }
      $35 = (_wcrtomb($buf,$29,0)|0);
      $36 = ($35|0)==(-1);
      if ($36) {
       $$0 = -1;
       break L1;
      }
      $37 = ($$422>>>0)<($35>>>0);
      if ($37) {
       $$422$lcssa83 = $$422;
       label = 22;
       break;
      }
      $39 = HEAP32[$30>>2]|0;
      (_wcrtomb($$223,$39,0)|0);
      $40 = (($$223) + ($35)|0);
      $41 = (($$422) - ($35))|0;
      $$3 = $40;$$5 = $41;$46 = $30;
     } else {
      $42 = $29&255;
      $43 = ((($$223)) + 1|0);
      HEAP8[$$223>>0] = $42;
      $44 = (($$422) + -1)|0;
      $$pre46 = HEAP32[$ws>>2]|0;
      $$3 = $43;$$5 = $44;$46 = $$pre46;
     }
     $45 = ((($46)) + 4|0);
     HEAP32[$ws>>2] = $45;
     $47 = ($$5|0)==(0);
     if ($47) {
      $$0 = $n;
      break L1;
     } else {
      $$223 = $$3;$$422 = $$5;$30 = $45;
     }
    }
    if ((label|0) == 19) {
     HEAP8[$$223$lcssa>>0] = 0;
     HEAP32[$ws>>2] = 0;
     $34 = (($n) - ($$422$lcssa))|0;
     $$0 = $34;
     break;
    }
    else if ((label|0) == 22) {
     $38 = (($n) - ($$422$lcssa83))|0;
     $$0 = $38;
     break;
    }
   }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function ___ipparse($dest,$family,$s0) {
 $dest = $dest|0;
 $family = $family|0;
 $s0 = $s0|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$phi$trans$insert6 = 0, $$pre = 0, $$pre$phi8Z2D = 0, $$pre$phiZ2D = 0, $$pre7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $a = 0, $i$0$lcssa = 0, $i$04 = 0, $isdigit = 0, $isdigittmp = 0;
 var $s$03 = 0, $z = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $a = sp + 8|0;
 $z = sp;
 dest=$a; stop=dest+64|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $0 = ($family|0)==(10);
 L1: do {
  if ($0) {
   label = 16;
  } else {
   $i$04 = 0;$s$03 = $s0;
   L2: while(1) {
    $1 = (_strtoul($s$03,$z,0)|0);
    $2 = (($a) + ($i$04<<2)|0);
    HEAP32[$2>>2] = $1;
    $3 = HEAP32[$z>>2]|0;
    $4 = ($3|0)==($s$03|0);
    if ($4) {
     label = 5;
     break;
    }
    $5 = HEAP8[$3>>0]|0;
    switch ($5<<24>>24) {
    case 46: case 0:  {
     break;
    }
    default: {
     label = 5;
     break L2;
    }
    }
    $6 = HEAP8[$s$03>>0]|0;
    $7 = $6 << 24 >> 24;
    $isdigittmp = (($7) + -48)|0;
    $isdigit = ($isdigittmp>>>0)<(10);
    if (!($isdigit)) {
     label = 5;
     break;
    }
    $9 = ($5<<24>>24)==(0);
    if ($9) {
     $i$0$lcssa = $i$04;
     break;
    }
    $10 = ((($3)) + 1|0);
    $11 = (($i$04) + 1)|0;
    $12 = ($11|0)<(4);
    if ($12) {
     $i$04 = $11;$s$03 = $10;
    } else {
     $i$0$lcssa = $11;
     break;
    }
   }
   if ((label|0) == 5) {
    $8 = ($family|0)==(2);
    if ($8) {
     $$0 = -1;
     break;
    } else {
     label = 16;
     break;
    }
   }
   switch ($i$0$lcssa|0) {
   case 4:  {
    $$0 = -1;
    break L1;
    break;
   }
   case 0:  {
    $13 = HEAP32[$a>>2]|0;
    $14 = $13 & 16777215;
    $15 = ((($a)) + 4|0);
    HEAP32[$15>>2] = $14;
    $16 = $13 >>> 24;
    HEAP32[$a>>2] = $16;
    $$pre$phiZ2D = $15;$18 = $14;
    label = 12;
    break;
   }
   case 1:  {
    $$phi$trans$insert = ((($a)) + 4|0);
    $$pre = HEAP32[$$phi$trans$insert>>2]|0;
    $$pre$phiZ2D = $$phi$trans$insert;$18 = $$pre;
    label = 12;
    break;
   }
   case 2:  {
    $$phi$trans$insert6 = ((($a)) + 8|0);
    $$pre7 = HEAP32[$$phi$trans$insert6>>2]|0;
    $$pre$phi8Z2D = $$phi$trans$insert6;$22 = $$pre7;
    label = 13;
    break;
   }
   default: {
   }
   }
   if ((label|0) == 12) {
    $17 = $18 & 65535;
    $19 = ((($a)) + 8|0);
    HEAP32[$19>>2] = $17;
    $20 = $18 >>> 16;
    HEAP32[$$pre$phiZ2D>>2] = $20;
    $$pre$phi8Z2D = $19;$22 = $17;
    label = 13;
   }
   if ((label|0) == 13) {
    $21 = $22 & 255;
    $23 = ((($a)) + 12|0);
    HEAP32[$23>>2] = $21;
    $24 = $22 >>> 8;
    HEAP32[$$pre$phi8Z2D>>2] = $24;
   }
   HEAP16[$dest>>1] = 2;
   $25 = HEAP32[$a>>2]|0;
   $26 = ($25>>>0)>(255);
   if ($26) {
    $$0 = -1;
   } else {
    $27 = ((($dest)) + 4|0);
    $28 = $25&255;
    HEAP8[$27>>0] = $28;
    $29 = ((($a)) + 4|0);
    $30 = HEAP32[$29>>2]|0;
    $31 = ($30>>>0)>(255);
    if ($31) {
     $$0 = -1;
    } else {
     $36 = $30&255;
     $37 = ((($dest)) + 5|0);
     HEAP8[$37>>0] = $36;
     $38 = ((($a)) + 8|0);
     $39 = HEAP32[$38>>2]|0;
     $40 = ($39>>>0)>(255);
     if ($40) {
      $$0 = -1;
     } else {
      $41 = $39&255;
      $42 = ((($dest)) + 6|0);
      HEAP8[$42>>0] = $41;
      $43 = ((($a)) + 12|0);
      $44 = HEAP32[$43>>2]|0;
      $45 = ($44>>>0)>(255);
      if ($45) {
       $$0 = -1;
      } else {
       $46 = $44&255;
       $47 = ((($dest)) + 7|0);
       HEAP8[$47>>0] = $46;
       $$0 = 0;
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 16) {
  HEAP16[$dest>>1] = 10;
  $32 = ((($dest)) + 8|0);
  $33 = (_inet_pton(10,$s0,$32)|0);
  $34 = ($33|0)<(1);
  $35 = $34 << 31 >> 31;
  $$0 = $35;
 }
 STACKTOP = sp;return ($$0|0);
}
function _inet_pton($af,$s,$a0) {
 $af = $af|0;
 $s = $s|0;
 $a0 = $a0|0;
 var $$0 = 0, $$01544 = 0, $$2 = 0, $$2$lcssa121 = 0, $$3 = 0, $$4 = 0, $$lcssa128 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$pre = 0, $$pre73 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $a$047 = 0, $brk$0 = 0, $brk$0$lcssa125 = 0, $brk$1 = 0, $brk$2 = 0, $exitcond = 0, $i$046 = 0, $i$1 = 0, $i$1$lcssa123 = 0, $i$2 = 0, $ip = 0, $isdigit = 0, $isdigit$1 = 0;
 var $isdigit$2 = 0, $isdigittmp = 0, $isdigittmp$1 = 0, $isdigittmp$2 = 0, $j$025 = 0, $j$026 = 0, $j$131 = 0, $j$150 = 0, $j$150$lcssa = 0, $j$348 = 0, $need_v4$0 = 0, $or$cond = 0, $or$cond11 = 0, $or$cond14 = 0, $or$cond17 = 0, $or$cond4 = 0, $or$cond7 = 0, $v$028 = 0, $v$029 = 0, $v$133 = 0;
 var $v$151 = 0, $v$151$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ip = sp;
 L1: do {
  switch ($af|0) {
  case 2:  {
   $$01544 = $s;$i$046 = 0;
   while(1) {
    $0 = HEAP8[$$01544>>0]|0;
    $1 = $0 << 24 >> 24;
    $isdigittmp = (($1) + -48)|0;
    $isdigit = ($isdigittmp>>>0)<(10);
    if (!($isdigit)) {
     $$0 = 0;
     break L1;
    }
    $2 = ((($$01544)) + 1|0);
    $3 = HEAP8[$2>>0]|0;
    $4 = $3 << 24 >> 24;
    $isdigittmp$1 = (($4) + -48)|0;
    $isdigit$1 = ($isdigittmp$1>>>0)<(10);
    if ($isdigit$1) {
     $87 = ($isdigittmp*10)|0;
     $88 = (($87) + -48)|0;
     $89 = (($88) + ($4))|0;
     $90 = ((($$01544)) + 2|0);
     $91 = HEAP8[$90>>0]|0;
     $92 = $91 << 24 >> 24;
     $isdigittmp$2 = (($92) + -48)|0;
     $isdigit$2 = ($isdigittmp$2>>>0)<(10);
     if ($isdigit$2) {
      $93 = ($89*10)|0;
      $94 = (($93) + -48)|0;
      $95 = (($94) + ($92))|0;
      $j$026 = 3;$v$029 = $95;
     } else {
      $j$026 = 2;$v$029 = $89;
     }
     $5 = ($0<<24>>24)==(48);
     $6 = ($v$029|0)>(255);
     $or$cond = $6 | $5;
     if ($or$cond) {
      $$0 = 0;
      break L1;
     } else {
      $j$025 = $j$026;$v$028 = $v$029;
     }
    } else {
     $j$025 = 1;$v$028 = $isdigittmp;
    }
    $7 = $v$028&255;
    $8 = (($a0) + ($i$046)|0);
    HEAP8[$8>>0] = $7;
    $9 = (($$01544) + ($j$025)|0);
    $10 = HEAP8[$9>>0]|0;
    $11 = ($10<<24>>24)==(0);
    $12 = ($i$046|0)==(3);
    $or$cond4 = $12 & $11;
    if ($or$cond4) {
     $$0 = 1;
     break L1;
    }
    $13 = ($10<<24>>24)==(46);
    if (!($13)) {
     $$0 = 0;
     break L1;
    }
    $14 = (($j$025) + 1)|0;
    $15 = (($$01544) + ($14)|0);
    $16 = (($i$046) + 1)|0;
    $17 = ($16|0)<(4);
    if ($17) {
     $$01544 = $15;$i$046 = $16;
    } else {
     $$0 = 0;
     break;
    }
   }
   break;
  }
  case 10:  {
   $19 = HEAP8[$s>>0]|0;
   $20 = ($19<<24>>24)==(58);
   if ($20) {
    $21 = ((($s)) + 1|0);
    $22 = HEAP8[$21>>0]|0;
    $23 = ($22<<24>>24)==(58);
    if ($23) {
     $$2 = $21;$25 = 58;$brk$0 = -1;$i$1 = 0;
    } else {
     $$0 = 0;
     break L1;
    }
   } else {
    $$2 = $s;$25 = $19;$brk$0 = -1;$i$1 = 0;
   }
   L17: while(1) {
    $24 = ($25<<24>>24)==(58);
    $26 = ($brk$0|0)<(0);
    $or$cond7 = $26 & $24;
    if ($or$cond7) {
     $27 = $i$1 & 7;
     $28 = (($ip) + ($27<<1)|0);
     HEAP16[$28>>1] = 0;
     $29 = ((($$2)) + 1|0);
     $30 = HEAP8[$29>>0]|0;
     $31 = ($30<<24>>24)==(0);
     if ($31) {
      $$4 = $29;$brk$2 = $i$1;$i$2 = $i$1;$need_v4$0 = 0;
      break;
     }
     $32 = ($i$1|0)==(7);
     if ($32) {
      $$0 = 0;
      break L1;
     } else {
      $$3 = $29;$$pre = $30;$brk$1 = $i$1;
     }
    } else {
     $34 = $25;$j$150 = 0;$v$151 = 0;
     while(1) {
      $33 = $34 << 24 >> 24;
      $35 = (($33) + -48)|0;
      $36 = ($35>>>0)<(10);
      if ($36) {
       $44 = $35;
      } else {
       $37 = $33 | 32;
       $38 = (($37) + -97)|0;
       $39 = ($38>>>0)<(6);
       $40 = (($37) + -87)|0;
       $41 = ($40|0)>(-1);
       $or$cond17 = $39 & $41;
       if ($or$cond17) {
        $44 = $40;
       } else {
        $j$150$lcssa = $j$150;$v$151$lcssa = $v$151;
        label = 18;
        break;
       }
      }
      $42 = $v$151 << 4;
      $43 = (($44) + ($42))|0;
      $45 = (($j$150) + 1)|0;
      $46 = ($45|0)<(4);
      if (!($46)) {
       $j$131 = $45;$v$133 = $43;
       break;
      }
      $$phi$trans$insert = (($$2) + ($45)|0);
      $$pre73 = HEAP8[$$phi$trans$insert>>0]|0;
      $34 = $$pre73;$j$150 = $45;$v$151 = $43;
     }
     if ((label|0) == 18) {
      label = 0;
      $47 = ($j$150$lcssa|0)==(0);
      if ($47) {
       $$0 = 0;
       break L1;
      } else {
       $j$131 = $j$150$lcssa;$v$133 = $v$151$lcssa;
      }
     }
     $48 = $v$133&65535;
     $49 = $i$1 & 7;
     $50 = (($ip) + ($49<<1)|0);
     HEAP16[$50>>1] = $48;
     $51 = (($$2) + ($j$131)|0);
     $52 = HEAP8[$51>>0]|0;
     $53 = ($52<<24>>24)==(0);
     if ($53) {
      $54 = ($brk$0|0)>(-1);
      $55 = ($i$1|0)==(7);
      $or$cond11 = $55 | $54;
      if ($or$cond11) {
       $$4 = $$2;$brk$2 = $brk$0;$i$2 = $i$1;$need_v4$0 = 0;
       break;
      }
     }
     $56 = ($i$1|0)==(7);
     if ($56) {
      $$0 = 0;
      break L1;
     }
     switch ($52<<24>>24) {
     case 46:  {
      $$2$lcssa121 = $$2;$$lcssa128 = $26;$brk$0$lcssa125 = $brk$0;$i$1$lcssa123 = $i$1;
      label = 23;
      break L17;
      break;
     }
     case 58:  {
      break;
     }
     default: {
      $$0 = 0;
      break L1;
     }
     }
     $59 = (($j$131) + 1)|0;
     $60 = (($$2) + ($59)|0);
     $$pre$pre = HEAP8[$60>>0]|0;
     $$3 = $60;$$pre = $$pre$pre;$brk$1 = $brk$0;
    }
    $61 = (($i$1) + 1)|0;
    $$2 = $$3;$25 = $$pre;$brk$0 = $brk$1;$i$1 = $61;
   }
   if ((label|0) == 23) {
    $57 = ($i$1$lcssa123|0)<(6);
    $or$cond14 = $57 & $$lcssa128;
    if ($or$cond14) {
     $$0 = 0;
     break L1;
    }
    $58 = (($i$1$lcssa123) + 1)|0;
    $$4 = $$2$lcssa121;$brk$2 = $brk$0$lcssa125;$i$2 = $58;$need_v4$0 = 1;
   }
   $62 = ($brk$2|0)>(-1);
   if ($62) {
    $63 = (($ip) + ($brk$2<<1)|0);
    $64 = ((($63)) + 14|0);
    $65 = (0 - ($i$2))|0;
    $66 = (($64) + ($65<<1)|0);
    $67 = (($i$2) + 1)|0;
    $68 = (($67) - ($brk$2))|0;
    $69 = $68 << 1;
    _memmove(($66|0),($63|0),($69|0))|0;
    $70 = (7 - ($i$2))|0;
    $71 = ($70|0)>(0);
    if ($71) {
     $72 = $i$2 << 1;
     $73 = (14 - ($72))|0;
     _memset(($63|0),0,($73|0))|0;
     $a$047 = $a0;$j$348 = 0;
    } else {
     $a$047 = $a0;$j$348 = 0;
    }
   } else {
    $a$047 = $a0;$j$348 = 0;
   }
   while(1) {
    $74 = (($ip) + ($j$348<<1)|0);
    $75 = HEAP16[$74>>1]|0;
    $76 = ($75&65535) >>> 8;
    $77 = $76&255;
    $78 = ((($a$047)) + 1|0);
    HEAP8[$a$047>>0] = $77;
    $79 = HEAP16[$74>>1]|0;
    $80 = $79&255;
    $81 = ((($a$047)) + 2|0);
    HEAP8[$78>>0] = $80;
    $82 = (($j$348) + 1)|0;
    $exitcond = ($82|0)==(8);
    if ($exitcond) {
     break;
    } else {
     $a$047 = $81;$j$348 = $82;
    }
   }
   $83 = ($need_v4$0|0)==(0);
   if (!($83)) {
    $84 = ((($a0)) + 12|0);
    $85 = (_inet_pton(2,$$4,$84)|0);
    $86 = ($85|0)<(1);
    if ($86) {
     $$0 = 0;
     break L1;
    }
   }
   $$0 = 1;
   break;
  }
  default: {
   $18 = (___errno_location()|0);
   HEAP32[$18>>2] = 97;
   $$0 = -1;
  }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _htons($n) {
 $n = $n|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $n&65535;
 $1 = $0 << 8;
 $2 = $0 >>> 8;
 $3 = $1 | $2;
 $4 = $3&65535;
 return ($4|0);
}
function _htonl($n) {
 $n = $n|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_llvm_bswap_i32(($n|0))|0);
 return ($0|0);
}
function _setsockopt($fd,$level,$optname,$optval,$optlen) {
 $fd = $fd|0;
 $level = $level|0;
 $optname = $optname|0;
 $optval = $optval|0;
 $optlen = $optlen|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 HEAP32[$0>>2] = $fd;
 $1 = ((($0)) + 4|0);
 HEAP32[$1>>2] = $level;
 $2 = ((($0)) + 8|0);
 HEAP32[$2>>2] = $optname;
 $3 = ((($0)) + 12|0);
 $4 = $optval;
 HEAP32[$3>>2] = $4;
 $5 = ((($0)) + 16|0);
 HEAP32[$5>>2] = $optlen;
 $6 = ((($0)) + 20|0);
 HEAP32[$6>>2] = 0;
 HEAP32[$vararg_buffer>>2] = 14;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $0;
 $7 = (___syscall102(102,($vararg_buffer|0))|0);
 $8 = (___syscall_ret($7)|0);
 STACKTOP = sp;return ($8|0);
}
function _bind($fd,$addr,$len) {
 $fd = $fd|0;
 $addr = $addr|0;
 $len = $len|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 HEAP32[$0>>2] = $fd;
 $1 = ((($0)) + 4|0);
 $2 = $addr;
 HEAP32[$1>>2] = $2;
 $3 = ((($0)) + 8|0);
 HEAP32[$3>>2] = $len;
 $4 = ((($0)) + 12|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($0)) + 16|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($0)) + 20|0);
 HEAP32[$6>>2] = 0;
 HEAP32[$vararg_buffer>>2] = 2;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $0;
 $7 = (___syscall102(102,($vararg_buffer|0))|0);
 $8 = (___syscall_ret($7)|0);
 STACKTOP = sp;return ($8|0);
}
function _recvfrom($fd,$buf,$len,$flags,$addr,$alen) {
 $fd = $fd|0;
 $buf = $buf|0;
 $len = $len|0;
 $flags = $flags|0;
 $addr = $addr|0;
 $alen = $alen|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 HEAP32[$0>>2] = $fd;
 $1 = ((($0)) + 4|0);
 $2 = $buf;
 HEAP32[$1>>2] = $2;
 $3 = ((($0)) + 8|0);
 HEAP32[$3>>2] = $len;
 $4 = ((($0)) + 12|0);
 HEAP32[$4>>2] = $flags;
 $5 = ((($0)) + 16|0);
 $6 = $addr;
 HEAP32[$5>>2] = $6;
 $7 = ((($0)) + 20|0);
 $8 = $alen;
 HEAP32[$7>>2] = $8;
 HEAP32[$vararg_buffer>>2] = 12;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $0;
 $9 = (___syscall102(102,($vararg_buffer|0))|0);
 $10 = (___syscall_ret($9)|0);
 STACKTOP = sp;return ($10|0);
}
function _accept($fd,$addr,$len) {
 $fd = $fd|0;
 $addr = $addr|0;
 $len = $len|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 HEAP32[$0>>2] = $fd;
 $1 = ((($0)) + 4|0);
 $2 = $addr;
 HEAP32[$1>>2] = $2;
 $3 = ((($0)) + 8|0);
 $4 = $len;
 HEAP32[$3>>2] = $4;
 $5 = ((($0)) + 12|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($0)) + 16|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($0)) + 20|0);
 HEAP32[$7>>2] = 0;
 HEAP32[$vararg_buffer>>2] = 5;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $0;
 $8 = (___syscall102(102,($vararg_buffer|0))|0);
 $9 = (___syscall_ret($8)|0);
 STACKTOP = sp;return ($9|0);
}
function _getpeername($fd,$addr,$len) {
 $fd = $fd|0;
 $addr = $addr|0;
 $len = $len|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 HEAP32[$0>>2] = $fd;
 $1 = ((($0)) + 4|0);
 $2 = $addr;
 HEAP32[$1>>2] = $2;
 $3 = ((($0)) + 8|0);
 $4 = $len;
 HEAP32[$3>>2] = $4;
 $5 = ((($0)) + 12|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($0)) + 16|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($0)) + 20|0);
 HEAP32[$7>>2] = 0;
 HEAP32[$vararg_buffer>>2] = 7;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $0;
 $8 = (___syscall102(102,($vararg_buffer|0))|0);
 $9 = (___syscall_ret($8)|0);
 STACKTOP = sp;return ($9|0);
}
function _getsockname($fd,$addr,$len) {
 $fd = $fd|0;
 $addr = $addr|0;
 $len = $len|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 HEAP32[$0>>2] = $fd;
 $1 = ((($0)) + 4|0);
 $2 = $addr;
 HEAP32[$1>>2] = $2;
 $3 = ((($0)) + 8|0);
 $4 = $len;
 HEAP32[$3>>2] = $4;
 $5 = ((($0)) + 12|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($0)) + 16|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($0)) + 20|0);
 HEAP32[$7>>2] = 0;
 HEAP32[$vararg_buffer>>2] = 6;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $0;
 $8 = (___syscall102(102,($vararg_buffer|0))|0);
 $9 = (___syscall_ret($8)|0);
 STACKTOP = sp;return ($9|0);
}
function ___h_errno_location() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (1205996|0);
}
function _hstrerror($ecode) {
 $ecode = $ecode|0;
 var $$0 = 0, $$01 = 0, $$03 = 0, $$lcssa = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $s$0$lcssa = 0, $s$02 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $$01 = (($ecode) + -1)|0;
 $0 = ($$01|0)==(0);
 L1: do {
  if ($0) {
   $s$0$lcssa = 619469;
  } else {
   $$03 = $$01;$s$02 = 619469;
   while(1) {
    $2 = HEAP8[$s$02>>0]|0;
    $3 = ($2<<24>>24)==(0);
    if ($3) {
     $s$0$lcssa = $s$02;
     break L1;
    }
    $4 = ((($s$02)) + 1|0);
    $5 = $4;
    while(1) {
     $$pre = HEAP8[$5>>0]|0;
     $6 = ($$pre<<24>>24)==(0);
     $7 = ((($5)) + 1|0);
     if ($6) {
      $$lcssa = $7;
      break;
     } else {
      $5 = $7;
     }
    }
    $$0 = (($$03) + -1)|0;
    $1 = ($$0|0)==(0);
    if ($1) {
     $s$0$lcssa = $$lcssa;
     break;
    } else {
     $$03 = $$0;$s$02 = $$lcssa;
    }
   }
  }
 } while(0);
 $8 = HEAP8[$s$0$lcssa>>0]|0;
 $9 = ($8<<24>>24)!=(0);
 $10 = ((($s$0$lcssa)) + 1|0);
 $11 = $9 ? $s$0$lcssa : $10;
 return ($11|0);
}
function _strncpy($d,$s,$n) {
 $d = $d|0;
 $s = $s|0;
 $n = $n|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 (___stpncpy($d,$s,$n)|0);
 return ($d|0);
}
function ___stpncpy($d,$s,$n) {
 $d = $d|0;
 $s = $s|0;
 $n = $n|0;
 var $$0$lcssa = 0, $$0$lcssa42 = 0, $$031 = 0, $$04$lcssa = 0, $$04$lcssa43 = 0, $$0430 = 0, $$06$lcssa = 0, $$06$lcssa44 = 0, $$0629 = 0, $$1$lcssa = 0, $$117 = 0, $$15$ph = 0, $$1511 = 0, $$17$ph = 0, $$1710 = 0, $$2$ph = 0, $$212 = 0, $$28 = 0, $$3 = 0, $$lcssa = 0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond28 = 0, $wd$0$lcssa = 0;
 var $wd$018 = 0, $ws$0$lcssa = 0, $ws$019 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $s;
 $1 = $d;
 $2 = $0 ^ $1;
 $3 = $2 & 3;
 $4 = ($3|0)==(0);
 do {
  if ($4) {
   $5 = $0 & 3;
   $6 = ($5|0)!=(0);
   $7 = ($n|0)!=(0);
   $or$cond28 = $7 & $6;
   L3: do {
    if ($or$cond28) {
     $$031 = $n;$$0430 = $s;$$0629 = $d;
     while(1) {
      $8 = HEAP8[$$0430>>0]|0;
      HEAP8[$$0629>>0] = $8;
      $9 = ($8<<24>>24)==(0);
      if ($9) {
       $$0$lcssa42 = $$031;$$04$lcssa43 = $$0430;$$06$lcssa44 = $$0629;
       break L3;
      }
      $10 = (($$031) + -1)|0;
      $11 = ((($$0430)) + 1|0);
      $12 = ((($$0629)) + 1|0);
      $13 = $11;
      $14 = $13 & 3;
      $15 = ($14|0)!=(0);
      $16 = ($10|0)!=(0);
      $or$cond = $16 & $15;
      if ($or$cond) {
       $$031 = $10;$$0430 = $11;$$0629 = $12;
      } else {
       $$0$lcssa = $10;$$04$lcssa = $11;$$06$lcssa = $12;$$lcssa = $16;
       label = 5;
       break;
      }
     }
    } else {
     $$0$lcssa = $n;$$04$lcssa = $s;$$06$lcssa = $d;$$lcssa = $7;
     label = 5;
    }
   } while(0);
   if ((label|0) == 5) {
    if ($$lcssa) {
     $$0$lcssa42 = $$0$lcssa;$$04$lcssa43 = $$04$lcssa;$$06$lcssa44 = $$06$lcssa;
    } else {
     $$28 = $$06$lcssa;$$3 = 0;
     break;
    }
   }
   $17 = HEAP8[$$04$lcssa43>>0]|0;
   $18 = ($17<<24>>24)==(0);
   if ($18) {
    $$28 = $$06$lcssa44;$$3 = $$0$lcssa42;
   } else {
    $19 = ($$0$lcssa42>>>0)>(3);
    L11: do {
     if ($19) {
      $$117 = $$0$lcssa42;$wd$018 = $$06$lcssa44;$ws$019 = $$04$lcssa43;
      while(1) {
       $20 = HEAP32[$ws$019>>2]|0;
       $21 = (($20) + -16843009)|0;
       $22 = $20 & -2139062144;
       $23 = $22 ^ -2139062144;
       $24 = $23 & $21;
       $25 = ($24|0)==(0);
       if (!($25)) {
        $$1$lcssa = $$117;$wd$0$lcssa = $wd$018;$ws$0$lcssa = $ws$019;
        break L11;
       }
       HEAP32[$wd$018>>2] = $20;
       $26 = (($$117) + -4)|0;
       $27 = ((($ws$019)) + 4|0);
       $28 = ((($wd$018)) + 4|0);
       $29 = ($26>>>0)>(3);
       if ($29) {
        $$117 = $26;$wd$018 = $28;$ws$019 = $27;
       } else {
        $$1$lcssa = $26;$wd$0$lcssa = $28;$ws$0$lcssa = $27;
        break;
       }
      }
     } else {
      $$1$lcssa = $$0$lcssa42;$wd$0$lcssa = $$06$lcssa44;$ws$0$lcssa = $$04$lcssa43;
     }
    } while(0);
    $$15$ph = $ws$0$lcssa;$$17$ph = $wd$0$lcssa;$$2$ph = $$1$lcssa;
    label = 11;
   }
  } else {
   $$15$ph = $s;$$17$ph = $d;$$2$ph = $n;
   label = 11;
  }
 } while(0);
 L16: do {
  if ((label|0) == 11) {
   $30 = ($$2$ph|0)==(0);
   if ($30) {
    $$28 = $$17$ph;$$3 = 0;
   } else {
    $$1511 = $$15$ph;$$1710 = $$17$ph;$$212 = $$2$ph;
    while(1) {
     $31 = HEAP8[$$1511>>0]|0;
     HEAP8[$$1710>>0] = $31;
     $32 = ($31<<24>>24)==(0);
     if ($32) {
      $$28 = $$1710;$$3 = $$212;
      break L16;
     }
     $33 = (($$212) + -1)|0;
     $34 = ((($$1511)) + 1|0);
     $35 = ((($$1710)) + 1|0);
     $36 = ($33|0)==(0);
     if ($36) {
      $$28 = $35;$$3 = 0;
      break;
     } else {
      $$1511 = $34;$$1710 = $35;$$212 = $33;
     }
    }
   }
  }
 } while(0);
 _memset(($$28|0),0,($$3|0))|0;
 return ($$28|0);
}
function _ntohl($n) {
 $n = $n|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_llvm_bswap_i32(($n|0))|0);
 return ($0|0);
}
function _inet_aton($cp,$inp) {
 $cp = $cp|0;
 $inp = $inp|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $sin = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $sin = sp;
 $0 = (___ipparse($sin,2,$cp)|0);
 $1 = ($0|0)<(0);
 if ($1) {
  $$0 = 0;
 } else {
  $2 = ((($sin)) + 4|0);
  $3 = HEAP32[$2>>2]|0;
  HEAP32[$inp>>2] = $3;
  $$0 = 1;
 }
 STACKTOP = sp;return ($$0|0);
}
function _inet_ntoa($in) {
 $in = $in|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = HEAP32[$in>>2]|0;
 $1 = $0 & 255;
 $2 = $0 >>> 8;
 $3 = $2 & 255;
 $4 = $0 >>> 16;
 $5 = $4 & 255;
 $6 = $0 >>> 24;
 HEAP32[$vararg_buffer>>2] = $1;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $3;
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
 HEAP32[$vararg_ptr2>>2] = $5;
 $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
 HEAP32[$vararg_ptr3>>2] = $6;
 (_snprintf(1282429,16,619553,$vararg_buffer)|0);
 STACKTOP = sp;return (1282429|0);
}
function _listen($fd,$backlog) {
 $fd = $fd|0;
 $backlog = $backlog|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 HEAP32[$0>>2] = $fd;
 $1 = ((($0)) + 4|0);
 HEAP32[$1>>2] = $backlog;
 $2 = ((($0)) + 8|0);
 ;HEAP32[$2>>2]=0|0;HEAP32[$2+4>>2]=0|0;HEAP32[$2+8>>2]=0|0;HEAP32[$2+12>>2]=0|0;
 HEAP32[$vararg_buffer>>2] = 4;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $0;
 $3 = (___syscall102(102,($vararg_buffer|0))|0);
 $4 = (___syscall_ret($3)|0);
 STACKTOP = sp;return ($4|0);
}
function _ntohs($n) {
 $n = $n|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $n&65535;
 $1 = $0 << 8;
 $2 = $0 >>> 8;
 $3 = $1 | $2;
 $4 = $3&65535;
 return ($4|0);
}
function _recv($fd,$buf,$len,$flags) {
 $fd = $fd|0;
 $buf = $buf|0;
 $len = $len|0;
 $flags = $flags|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_recvfrom($fd,$buf,$len,$flags,0,0)|0);
 return ($0|0);
}
function _recvmsg($fd,$msg,$flags) {
 $fd = $fd|0;
 $msg = $msg|0;
 $flags = $flags|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 HEAP32[$0>>2] = $fd;
 $1 = ((($0)) + 4|0);
 $2 = $msg;
 HEAP32[$1>>2] = $2;
 $3 = ((($0)) + 8|0);
 HEAP32[$3>>2] = $flags;
 $4 = ((($0)) + 12|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($0)) + 16|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($0)) + 20|0);
 HEAP32[$6>>2] = 0;
 HEAP32[$vararg_buffer>>2] = 17;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $0;
 $7 = (___syscall102(102,($vararg_buffer|0))|0);
 $8 = (___syscall_ret($7)|0);
 STACKTOP = sp;return ($8|0);
}
function _sendmsg($fd,$msg,$flags) {
 $fd = $fd|0;
 $msg = $msg|0;
 $flags = $flags|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 HEAP32[$0>>2] = $fd;
 $1 = ((($0)) + 4|0);
 $2 = $msg;
 HEAP32[$1>>2] = $2;
 $3 = ((($0)) + 8|0);
 HEAP32[$3>>2] = $flags;
 $4 = ((($0)) + 12|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($0)) + 16|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($0)) + 20|0);
 HEAP32[$6>>2] = 0;
 HEAP32[$vararg_buffer>>2] = 16;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $0;
 $7 = (___syscall102(102,($vararg_buffer|0))|0);
 $8 = (___syscall_ret($7)|0);
 STACKTOP = sp;return ($8|0);
}
function _shutdown($fd,$how) {
 $fd = $fd|0;
 $how = $how|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 HEAP32[$0>>2] = $fd;
 $1 = ((($0)) + 4|0);
 HEAP32[$1>>2] = $how;
 $2 = ((($0)) + 8|0);
 ;HEAP32[$2>>2]=0|0;HEAP32[$2+4>>2]=0|0;HEAP32[$2+8>>2]=0|0;HEAP32[$2+12>>2]=0|0;
 HEAP32[$vararg_buffer>>2] = 13;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $0;
 $3 = (___syscall102(102,($vararg_buffer|0))|0);
 $4 = (___syscall_ret($3)|0);
 STACKTOP = sp;return ($4|0);
}
function _srand($s) {
 $s = $s|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($s) + -1)|0;
 $1 = 624936;
 $2 = $1;
 HEAP32[$2>>2] = $0;
 $3 = (($1) + 4)|0;
 $4 = $3;
 HEAP32[$4>>2] = 0;
 return;
}
function _rand() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = 624936;
 $1 = $0;
 $2 = HEAP32[$1>>2]|0;
 $3 = (($0) + 4)|0;
 $4 = $3;
 $5 = HEAP32[$4>>2]|0;
 $6 = (___muldi3(($2|0),($5|0),1284865837,1481765933)|0);
 $7 = tempRet0;
 $8 = (_i64Add(($6|0),($7|0),1,0)|0);
 $9 = tempRet0;
 $10 = 624936;
 $11 = $10;
 HEAP32[$11>>2] = $8;
 $12 = (($10) + 4)|0;
 $13 = $12;
 HEAP32[$13>>2] = $9;
 $14 = (_bitshift64Lshr(($8|0),($9|0),33)|0);
 $15 = tempRet0;
 return ($14|0);
}
function _srandom($seed) {
 $seed = $seed|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 ___lock((1206012|0));
 ___srandom($seed);
 ___unlock((1206012|0));
 return;
}
function ___srandom($seed) {
 $seed = $seed|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $k$01 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[38453]|0;
 $1 = ($0|0)==(0);
 if ($1) {
  $2 = HEAP32[38454]|0;
  HEAP32[$2>>2] = $seed;
 } else {
  $3 = ($0|0)==(31);
  $4 = ($0|0)==(7);
  $5 = $3 | $4;
  $6 = $5 ? 3 : 1;
  HEAP32[38487] = $6;
  HEAP32[301505] = 0;
  $7 = ($0|0)>(0);
  if ($7) {
   $8 = HEAP32[38454]|0;
   $10 = 0;$9 = $seed;$k$01 = 0;
   while(1) {
    $11 = (___muldi3(($9|0),($10|0),1284865837,1481765933)|0);
    $12 = tempRet0;
    $13 = (_i64Add(($11|0),($12|0),1,0)|0);
    $14 = tempRet0;
    $15 = (($8) + ($k$01<<2)|0);
    HEAP32[$15>>2] = $14;
    $16 = (($k$01) + 1)|0;
    $17 = ($16|0)<($0|0);
    if ($17) {
     $10 = $14;$9 = $13;$k$01 = $16;
    } else {
     $19 = $8;
     break;
    }
   }
  } else {
   $$pre = HEAP32[38454]|0;
   $19 = $$pre;
  }
  $18 = HEAP32[$19>>2]|0;
  $20 = $18 | 1;
  HEAP32[$19>>2] = $20;
 }
 return;
}
function _random() {
 var $$ = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $k$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ___lock((1206012|0));
 $0 = HEAP32[38453]|0;
 $1 = ($0|0)==(0);
 if ($1) {
  $2 = HEAP32[38454]|0;
  $3 = HEAP32[$2>>2]|0;
  $4 = Math_imul($3, 1103515245)|0;
  $5 = (($4) + 12345)|0;
  $6 = $5 & 2147483647;
  HEAP32[$2>>2] = $6;
  $k$0 = $6;
 } else {
  $7 = HEAP32[301505]|0;
  $8 = HEAP32[38454]|0;
  $9 = (($8) + ($7<<2)|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = HEAP32[38487]|0;
  $12 = (($8) + ($11<<2)|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = (($13) + ($10))|0;
  HEAP32[$12>>2] = $14;
  $15 = $14 >>> 1;
  $16 = (($11) + 1)|0;
  $17 = ($16|0)==($0|0);
  $$ = $17 ? 0 : $16;
  HEAP32[38487] = $$;
  $18 = (($7) + 1)|0;
  $19 = ($18|0)==($0|0);
  $$1 = $19 ? 0 : $18;
  HEAP32[301505] = $$1;
  $k$0 = $15;
 }
 ___unlock((1206012|0));
 return ($k$0|0);
}
function _select($n,$rfds,$wfds,$efds,$tv) {
 $n = $n|0;
 $rfds = $rfds|0;
 $wfds = $wfds|0;
 $efds = $efds|0;
 $tv = $tv|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0, $vararg_ptr4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 HEAP32[$vararg_buffer>>2] = $n;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $rfds;
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
 HEAP32[$vararg_ptr2>>2] = $wfds;
 $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
 HEAP32[$vararg_ptr3>>2] = $efds;
 $vararg_ptr4 = ((($vararg_buffer)) + 16|0);
 HEAP32[$vararg_ptr4>>2] = $tv;
 $0 = (___syscall142(142,($vararg_buffer|0))|0);
 $1 = (___syscall_ret($0)|0);
 STACKTOP = sp;return ($1|0);
}
function _mkdir($path,$mode) {
 $path = $path|0;
 $mode = $mode|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 HEAP32[$vararg_buffer>>2] = $path;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $mode;
 $0 = (___syscall39(39,($vararg_buffer|0))|0);
 $1 = (___syscall_ret($0)|0);
 STACKTOP = sp;return ($1|0);
}
function ___overflow($f,$_c) {
 $f = $f|0;
 $_c = $_c|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $c = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $c = sp;
 $0 = $_c&255;
 HEAP8[$c>>0] = $0;
 $1 = ((($f)) + 16|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if ($3) {
  $4 = (___towrite($f)|0);
  $5 = ($4|0)==(0);
  if ($5) {
   $$pre = HEAP32[$1>>2]|0;
   $9 = $$pre;
   label = 4;
  } else {
   $$0 = -1;
  }
 } else {
  $9 = $2;
  label = 4;
 }
 do {
  if ((label|0) == 4) {
   $6 = ((($f)) + 20|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7>>>0)<($9>>>0);
   if ($8) {
    $10 = $_c & 255;
    $11 = ((($f)) + 75|0);
    $12 = HEAP8[$11>>0]|0;
    $13 = $12 << 24 >> 24;
    $14 = ($10|0)==($13|0);
    if (!($14)) {
     $15 = ((($7)) + 1|0);
     HEAP32[$6>>2] = $15;
     HEAP8[$7>>0] = $0;
     $$0 = $10;
     break;
    }
   }
   $16 = ((($f)) + 36|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = (FUNCTION_TABLE_iiii[$17 & 8191]($f,$c,1)|0);
   $19 = ($18|0)==(1);
   if ($19) {
    $20 = HEAP8[$c>>0]|0;
    $21 = $20&255;
    $$0 = $21;
   } else {
    $$0 = -1;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _vasprintf($s,$fmt,$ap) {
 $s = $s|0;
 $fmt = $fmt|0;
 $ap = $ap|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ap2 = 0, $vacopy_currentptr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ap2 = sp;
 $0 = (_malloc(240)|0);
 $1 = ($0|0)==(0|0);
 do {
  if ($1) {
   $$0 = -1;
  } else {
   $vacopy_currentptr = HEAP32[$ap>>2]|0;
   HEAP32[$ap2>>2] = $vacopy_currentptr;
   $2 = (_vsnprintf($0,240,$fmt,$ap2)|0);
   $3 = ($2>>>0)<(240);
   if ($3) {
    $4 = (($2) + 1)|0;
    $5 = (_realloc($0,$4)|0);
    $6 = ($5|0)!=(0|0);
    $7 = $6 ? $5 : $0;
    HEAP32[$s>>2] = $7;
    $$0 = $2;
    break;
   }
   _free($0);
   $8 = ($2|0)<(0);
   if ($8) {
    $$0 = -1;
   } else {
    $9 = (($2) + 1)|0;
    $10 = (_malloc($9)|0);
    HEAP32[$s>>2] = $10;
    $11 = ($10|0)==(0|0);
    if ($11) {
     $$0 = -1;
    } else {
     $12 = (_vsnprintf($10,$9,$fmt,$ap)|0);
     $$0 = $12;
    }
   }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _clearerr($f) {
 $f = $f|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 76|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(-1);
 if ($2) {
  $3 = (___lockfile($f)|0);
  $phitmp = ($3|0)==(0);
  $4 = HEAP32[$f>>2]|0;
  $5 = $4 & -49;
  HEAP32[$f>>2] = $5;
  if (!($phitmp)) {
   ___unlockfile($f);
  }
 } else {
  $6 = HEAP32[$f>>2]|0;
  $7 = $6 & -49;
  HEAP32[$f>>2] = $7;
 }
 return;
}
function _ungetc($c,$f) {
 $c = $c|0;
 $f = $f|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($c|0)==(-1);
 do {
  if ($0) {
   $$1 = -1;
  } else {
   $1 = ((($f)) + 76|0);
   $2 = HEAP32[$1>>2]|0;
   $3 = ($2|0)>(-1);
   if ($3) {
    $4 = (___lockfile($f)|0);
    $17 = $4;
   } else {
    $17 = 0;
   }
   $5 = ((($f)) + 8|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($6|0)==(0|0);
   if ($7) {
    $8 = (___toread($f)|0);
    $9 = ($8|0)==(0);
    if ($9) {
     label = 6;
    }
   } else {
    label = 6;
   }
   if ((label|0) == 6) {
    $10 = ((($f)) + 4|0);
    $11 = HEAP32[$10>>2]|0;
    $12 = ((($f)) + 44|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($13)) + -8|0);
    $15 = ($11>>>0)>($14>>>0);
    if ($15) {
     $18 = $c&255;
     $19 = ((($11)) + -1|0);
     HEAP32[$10>>2] = $19;
     HEAP8[$19>>0] = $18;
     $20 = HEAP32[$f>>2]|0;
     $21 = $20 & -17;
     HEAP32[$f>>2] = $21;
     $22 = ($17|0)==(0);
     if ($22) {
      $$1 = $c;
      break;
     }
     ___unlockfile($f);
     $$1 = $c;
     break;
    }
   }
   $16 = ($17|0)==(0);
   if ($16) {
    $$1 = -1;
   } else {
    ___unlockfile($f);
    $$1 = -1;
   }
  }
 } while(0);
 return ($$1|0);
}
function ___ftello($f) {
 $f = $f|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 76|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(-1);
 if ($2) {
  $4 = (___lockfile($f)|0);
  $phitmp = ($4|0)==(0);
  $5 = (___ftello_unlocked($f)|0);
  if ($phitmp) {
   $6 = $5;
  } else {
   $6 = $5;
  }
 } else {
  $3 = (___ftello_unlocked($f)|0);
  $6 = $3;
 }
 return ($6|0);
}
function ___ftello_unlocked($f) {
 $f = $f|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 40|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$f>>2]|0;
 $3 = $2 & 128;
 $4 = ($3|0)==(0);
 if ($4) {
  $10 = 1;
 } else {
  $5 = ((($f)) + 20|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ((($f)) + 28|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($6>>>0)>($8>>>0);
  $phitmp = $9 ? 2 : 1;
  $10 = $phitmp;
 }
 $11 = (FUNCTION_TABLE_iiii[$1 & 8191]($f,0,$10)|0);
 $12 = ($11|0)<(0);
 if ($12) {
  $$0 = $11;
 } else {
  $13 = ((($f)) + 8|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ((($f)) + 4|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = ((($f)) + 20|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = ((($f)) + 28|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = (($11) - ($14))|0;
  $22 = (($21) + ($16))|0;
  $23 = (($22) + ($18))|0;
  $24 = (($23) - ($20))|0;
  $$0 = $24;
 }
 return ($$0|0);
}
function _fputc($c,$f) {
 $c = $c|0;
 $f = $f|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 76|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)<(0);
 if ($2) {
  label = 3;
 } else {
  $3 = (___lockfile($f)|0);
  $4 = ($3|0)==(0);
  if ($4) {
   label = 3;
  } else {
   $18 = ((($f)) + 75|0);
   $19 = HEAP8[$18>>0]|0;
   $20 = $19 << 24 >> 24;
   $21 = ($20|0)==($c|0);
   if ($21) {
    label = 10;
   } else {
    $22 = ((($f)) + 20|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = ((($f)) + 16|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = ($23>>>0)<($25>>>0);
    if ($26) {
     $27 = $c&255;
     $28 = ((($23)) + 1|0);
     HEAP32[$22>>2] = $28;
     HEAP8[$23>>0] = $27;
     $29 = $c & 255;
     $31 = $29;
    } else {
     label = 10;
    }
   }
   if ((label|0) == 10) {
    $30 = (___overflow($f,$c)|0);
    $31 = $30;
   }
   ___unlockfile($f);
   $$0 = $31;
  }
 }
 do {
  if ((label|0) == 3) {
   $5 = ((($f)) + 75|0);
   $6 = HEAP8[$5>>0]|0;
   $7 = $6 << 24 >> 24;
   $8 = ($7|0)==($c|0);
   if (!($8)) {
    $9 = ((($f)) + 20|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = ((($f)) + 16|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = ($10>>>0)<($12>>>0);
    if ($13) {
     $14 = $c&255;
     $15 = ((($10)) + 1|0);
     HEAP32[$9>>2] = $15;
     HEAP8[$10>>0] = $14;
     $16 = $c & 255;
     $$0 = $16;
     break;
    }
   }
   $17 = (___overflow($f,$c)|0);
   $$0 = $17;
  }
 } while(0);
 return ($$0|0);
}
function _fputs($s,$f) {
 $s = $s|0;
 $f = $f|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strlen($s)|0);
 $1 = (_fwrite($s,$0,1,$f)|0);
 $2 = (($1) + -1)|0;
 return ($2|0);
}
function _fwrite($src,$size,$nmemb,$f) {
 $src = $src|0;
 $size = $size|0;
 $nmemb = $nmemb|0;
 $f = $f|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = Math_imul($nmemb, $size)|0;
 $1 = ((($f)) + 76|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)>(-1);
 if ($3) {
  $5 = (___lockfile($f)|0);
  $phitmp = ($5|0)==(0);
  $6 = (___fwritex($src,$0,$f)|0);
  if ($phitmp) {
   $8 = $6;
  } else {
   ___unlockfile($f);
   $8 = $6;
  }
 } else {
  $4 = (___fwritex($src,$0,$f)|0);
  $8 = $4;
 }
 $7 = ($8|0)==($0|0);
 if ($7) {
  $10 = $nmemb;
 } else {
  $9 = (($8>>>0) / ($size>>>0))&-1;
  $10 = $9;
 }
 return ($10|0);
}
function _fread($destv,$size,$nmemb,$f) {
 $destv = $destv|0;
 $size = $size|0;
 $nmemb = $nmemb|0;
 $f = $f|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $dest$0$ph = 0, $dest$02 = 0, $l$0$ph = 0, $l$03 = 0, $l$03$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = Math_imul($nmemb, $size)|0;
 $1 = ((($f)) + 76|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)>(-1);
 if ($3) {
  $4 = (___lockfile($f)|0);
  $31 = $4;
 } else {
  $31 = 0;
 }
 $5 = ((($f)) + 74|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = $6 << 24 >> 24;
 $8 = (($7) + 255)|0;
 $9 = $8 | $7;
 $10 = $9&255;
 HEAP8[$5>>0] = $10;
 $11 = ((($f)) + 8|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($f)) + 4|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($12) - ($14))|0;
 $16 = ($15|0)>(0);
 $17 = $14;
 if ($16) {
  $18 = ($15>>>0)<($0>>>0);
  $$ = $18 ? $15 : $0;
  _memcpy(($destv|0),($17|0),($$|0))|0;
  $19 = (($17) + ($$)|0);
  HEAP32[$13>>2] = $19;
  $20 = (($destv) + ($$)|0);
  $21 = (($0) - ($$))|0;
  $dest$0$ph = $20;$l$0$ph = $21;
 } else {
  $dest$0$ph = $destv;$l$0$ph = $0;
 }
 $22 = ($l$0$ph|0)==(0);
 L7: do {
  if ($22) {
   label = 13;
  } else {
   $23 = ((($f)) + 32|0);
   $dest$02 = $dest$0$ph;$l$03 = $l$0$ph;
   while(1) {
    $24 = (___toread($f)|0);
    $25 = ($24|0)==(0);
    if (!($25)) {
     $l$03$lcssa = $l$03;
     break;
    }
    $26 = HEAP32[$23>>2]|0;
    $27 = (FUNCTION_TABLE_iiii[$26 & 8191]($f,$dest$02,$l$03)|0);
    $28 = (($27) + 1)|0;
    $29 = ($28>>>0)<(2);
    if ($29) {
     $l$03$lcssa = $l$03;
     break;
    }
    $34 = (($l$03) - ($27))|0;
    $35 = (($dest$02) + ($27)|0);
    $36 = ($l$03|0)==($27|0);
    if ($36) {
     label = 13;
     break L7;
    } else {
     $dest$02 = $35;$l$03 = $34;
    }
   }
   $30 = ($31|0)==(0);
   if (!($30)) {
    ___unlockfile($f);
   }
   $32 = (($0) - ($l$03$lcssa))|0;
   $33 = (($32>>>0) / ($size>>>0))&-1;
   $$0 = $33;
  }
 } while(0);
 if ((label|0) == 13) {
  $37 = ($31|0)==(0);
  if ($37) {
   $$0 = $nmemb;
  } else {
   ___unlockfile($f);
   $$0 = $nmemb;
  }
 }
 return ($$0|0);
}
function _freopen($filename,$mode,$f) {
 $filename = $filename|0;
 $mode = $mode|0;
 $f = $f|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer3 = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr6 = 0, $vararg_ptr7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $0 = (___fmodeflags($mode)|0);
 $1 = ((($f)) + 76|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)>(-1);
 if ($3) {
  $4 = (___lockfile($f)|0);
  $41 = $4;
 } else {
  $41 = 0;
 }
 (_fflush($f)|0);
 $5 = ($filename|0)==(0|0);
 do {
  if ($5) {
   $6 = $0 & 524288;
   $7 = ($6|0)==(0);
   $$pre = ((($f)) + 60|0);
   if (!($7)) {
    $8 = HEAP32[$$pre>>2]|0;
    HEAP32[$vararg_buffer>>2] = $8;
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
    HEAP32[$vararg_ptr1>>2] = 2;
    $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
    HEAP32[$vararg_ptr2>>2] = 1;
    (___syscall221(221,($vararg_buffer|0))|0);
   }
   $9 = $0 & -524481;
   $10 = HEAP32[$$pre>>2]|0;
   HEAP32[$vararg_buffer3>>2] = $10;
   $vararg_ptr6 = ((($vararg_buffer3)) + 4|0);
   HEAP32[$vararg_ptr6>>2] = 4;
   $vararg_ptr7 = ((($vararg_buffer3)) + 8|0);
   HEAP32[$vararg_ptr7>>2] = $9;
   $11 = (___syscall221(221,($vararg_buffer3|0))|0);
   $12 = (___syscall_ret($11)|0);
   $13 = ($12|0)<(0);
   if ($13) {
    label = 15;
   } else {
    label = 12;
   }
  } else {
   $14 = (_fopen($filename,$mode)|0);
   $15 = ($14|0)==(0|0);
   if ($15) {
    label = 15;
   } else {
    $16 = ((($14)) + 60|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = ((($f)) + 60|0);
    $19 = HEAP32[$18>>2]|0;
    $20 = ($17|0)==($19|0);
    if ($20) {
     HEAP32[$16>>2] = -1;
    } else {
     $21 = $0 & 524288;
     $22 = (___dup3($17,$19,$21)|0);
     $23 = ($22|0)<(0);
     if ($23) {
      (_fclose($14)|0);
      label = 15;
      break;
     }
    }
    $24 = HEAP32[$f>>2]|0;
    $25 = $24 & 1;
    $26 = HEAP32[$14>>2]|0;
    $27 = $25 | $26;
    HEAP32[$f>>2] = $27;
    $28 = ((($14)) + 32|0);
    $29 = HEAP32[$28>>2]|0;
    $30 = ((($f)) + 32|0);
    HEAP32[$30>>2] = $29;
    $31 = ((($14)) + 36|0);
    $32 = HEAP32[$31>>2]|0;
    $33 = ((($f)) + 36|0);
    HEAP32[$33>>2] = $32;
    $34 = ((($14)) + 40|0);
    $35 = HEAP32[$34>>2]|0;
    $36 = ((($f)) + 40|0);
    HEAP32[$36>>2] = $35;
    $37 = ((($14)) + 12|0);
    $38 = HEAP32[$37>>2]|0;
    $39 = ((($f)) + 12|0);
    HEAP32[$39>>2] = $38;
    (_fclose($14)|0);
    label = 12;
   }
  }
 } while(0);
 if ((label|0) == 12) {
  $40 = ($41|0)==(0);
  if ($40) {
   $$0 = $f;
  } else {
   ___unlockfile($f);
   $$0 = $f;
  }
 }
 else if ((label|0) == 15) {
  (_fclose($f)|0);
  $$0 = 0;
 }
 STACKTOP = sp;return ($$0|0);
}
function ___dup3($old,$new,$flags) {
 $old = $old|0;
 $new = $new|0;
 $flags = $flags|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa12 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $vararg_buffer = 0, $vararg_buffer3 = 0, $vararg_buffer7 = 0, $vararg_ptr1 = 0, $vararg_ptr10 = 0, $vararg_ptr11 = 0, $vararg_ptr2 = 0, $vararg_ptr6 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $0 = ($old|0)==($new|0);
 L1: do {
  if ($0) {
   $1 = (___syscall_ret(-22)|0);
   $$0 = $1;
  } else {
   $2 = $flags & 524288;
   $3 = ($2|0)!=(0);
   L4: do {
    if ($3) {
     L5: while(1) {
      HEAP32[$vararg_buffer>>2] = $old;
      $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
      HEAP32[$vararg_ptr1>>2] = $new;
      $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
      HEAP32[$vararg_ptr2>>2] = $flags;
      $4 = (___syscall330(330,($vararg_buffer|0))|0);
      switch ($4|0) {
      case -38:  {
       break L4;
       break;
      }
      case -16:  {
       break;
      }
      default: {
       $$lcssa12 = $4;
       break L5;
      }
      }
     }
     $5 = (___syscall_ret($$lcssa12)|0);
     $$0 = $5;
     break L1;
    }
   } while(0);
   while(1) {
    HEAP32[$vararg_buffer3>>2] = $old;
    $vararg_ptr6 = ((($vararg_buffer3)) + 4|0);
    HEAP32[$vararg_ptr6>>2] = $new;
    $6 = (___syscall63(63,($vararg_buffer3|0))|0);
    $7 = ($6|0)==(-16);
    if (!($7)) {
     $$lcssa = $6;
     break;
    }
   }
   if ($3) {
    HEAP32[$vararg_buffer7>>2] = $new;
    $vararg_ptr10 = ((($vararg_buffer7)) + 4|0);
    HEAP32[$vararg_ptr10>>2] = 2;
    $vararg_ptr11 = ((($vararg_buffer7)) + 8|0);
    HEAP32[$vararg_ptr11>>2] = 1;
    (___syscall221(221,($vararg_buffer7|0))|0);
   }
   $8 = (___syscall_ret($$lcssa)|0);
   $$0 = $8;
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _ftell($f) {
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (___ftello($f)|0);
 return ($0|0);
}
function _getc($f) {
 $f = $f|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 76|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)<(0);
 if ($2) {
  label = 3;
 } else {
  $3 = (___lockfile($f)|0);
  $4 = ($3|0)==(0);
  if ($4) {
   label = 3;
  } else {
   $14 = ((($f)) + 4|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ((($f)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ($15>>>0)<($17>>>0);
   if ($18) {
    $19 = ((($15)) + 1|0);
    HEAP32[$14>>2] = $19;
    $20 = HEAP8[$15>>0]|0;
    $21 = $20&255;
    $23 = $21;
   } else {
    $22 = (___uflow($f)|0);
    $23 = $22;
   }
   $$0 = $23;
  }
 }
 do {
  if ((label|0) == 3) {
   $5 = ((($f)) + 4|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ((($f)) + 8|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ($6>>>0)<($8>>>0);
   if ($9) {
    $10 = ((($6)) + 1|0);
    HEAP32[$5>>2] = $10;
    $11 = HEAP8[$6>>0]|0;
    $12 = $11&255;
    $$0 = $12;
    break;
   } else {
    $13 = (___uflow($f)|0);
    $$0 = $13;
    break;
   }
  }
 } while(0);
 return ($$0|0);
}
function _printf($fmt,$varargs) {
 $fmt = $fmt|0;
 $varargs = $varargs|0;
 var $0 = 0, $1 = 0, $ap = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $ap = sp;
 HEAP32[$ap>>2] = $varargs;
 $0 = HEAP32[37602]|0;
 $1 = (_vfprintf($0,$fmt,$ap)|0);
 STACKTOP = sp;return ($1|0);
}
function _putchar($c) {
 $c = $c|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[37602]|0;
 $1 = (_fputc($c,$0)|0);
 return ($1|0);
}
function _puts($s) {
 $s = $s|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[37602]|0;
 $1 = ((($0)) + 76|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)>(-1);
 if ($3) {
  $4 = (___lockfile($0)|0);
  $20 = $4;
 } else {
  $20 = 0;
 }
 $5 = (_fputs($s,$0)|0);
 $6 = ($5|0)<(0);
 do {
  if ($6) {
   $18 = 1;
  } else {
   $7 = ((($0)) + 75|0);
   $8 = HEAP8[$7>>0]|0;
   $9 = ($8<<24>>24)==(10);
   if (!($9)) {
    $10 = ((($0)) + 20|0);
    $11 = HEAP32[$10>>2]|0;
    $12 = ((($0)) + 16|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ($11>>>0)<($13>>>0);
    if ($14) {
     $15 = ((($11)) + 1|0);
     HEAP32[$10>>2] = $15;
     HEAP8[$11>>0] = 10;
     $18 = 0;
     break;
    }
   }
   $16 = (___overflow($0,10)|0);
   $phitmp = ($16|0)<(0);
   $18 = $phitmp;
  }
 } while(0);
 $17 = $18 << 31 >> 31;
 $19 = ($20|0)==(0);
 if (!($19)) {
  ___unlockfile($0);
 }
 return ($17|0);
}
function _remove($path) {
 $path = $path|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 HEAP32[$vararg_buffer>>2] = $path;
 $0 = (___syscall10(10,($vararg_buffer|0))|0);
 $1 = (___syscall_ret($0)|0);
 $2 = ($1|0)==(0);
 if ($2) {
  $8 = 0;
 } else {
  $3 = (___errno_location()|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ($4|0)==(21);
  if ($5) {
   HEAP32[$vararg_buffer1>>2] = $path;
   $6 = (___syscall40(40,($vararg_buffer1|0))|0);
   $7 = (___syscall_ret($6)|0);
   $8 = $7;
  } else {
   $8 = $1;
  }
 }
 STACKTOP = sp;return ($8|0);
}
function _rename($old,$new) {
 $old = $old|0;
 $new = $new|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 HEAP32[$vararg_buffer>>2] = $old;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $new;
 $0 = (___syscall38(38,($vararg_buffer|0))|0);
 $1 = (___syscall_ret($0)|0);
 STACKTOP = sp;return ($1|0);
}
function _setvbuf($f,$buf,$type,$size) {
 $f = $f|0;
 $buf = $buf|0;
 $type = $type|0;
 $size = $size|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($f)) + 75|0);
 HEAP8[$0>>0] = -1;
 switch ($type|0) {
 case 2:  {
  $1 = ((($f)) + 48|0);
  HEAP32[$1>>2] = 0;
  break;
 }
 case 1:  {
  HEAP8[$0>>0] = 10;
  break;
 }
 default: {
 }
 }
 $2 = HEAP32[$f>>2]|0;
 $3 = $2 | 64;
 HEAP32[$f>>2] = $3;
 return 0;
}
function _tmpfile() {
 var $$0 = 0, $$lcssa = 0, $$lcssa7 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $buf = 0, $try$02 = 0, $vararg_buffer = 0, $vararg_buffer3 = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $buf = sp + 20|0;
 $try$02 = 0;
 while(1) {
  $2 = (_tmpnam($buf)|0);
  $3 = ($2|0)==(0|0);
  if ($3) {
   $$0 = 0;
   break;
  }
  HEAP32[$vararg_buffer>>2] = $2;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = 32962;
  $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
  HEAP32[$vararg_ptr2>>2] = 384;
  $4 = (___syscall5(5,($vararg_buffer|0))|0);
  $5 = (___syscall_ret($4)|0);
  $6 = ($5|0)>(-1);
  $1 = (($try$02) + 1)|0;
  if ($6) {
   $$lcssa = $5;$$lcssa7 = $2;
   label = 5;
   break;
  }
  $0 = ($1|0)<(100);
  if ($0) {
   $try$02 = $1;
  } else {
   $$0 = 0;
   break;
  }
 }
 if ((label|0) == 5) {
  $7 = (___fdopen($$lcssa,619582)|0);
  HEAP32[$vararg_buffer3>>2] = $$lcssa7;
  (___syscall10(10,($vararg_buffer3|0))|0);
  $$0 = $7;
 }
 STACKTOP = sp;return ($$0|0);
}
function _tmpnam($s) {
 $s = $s|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $try$0 = 0, $try$1 = 0, $ts = 0, $vararg_buffer = 0, $vararg_buffer11 = 0, $vararg_buffer2 = 0, $vararg_buffer7 = 0, $vararg_ptr1 = 0, $vararg_ptr10 = 0, $vararg_ptr14 = 0, $vararg_ptr5 = 0, $vararg_ptr6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer11 = sp + 32|0;
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 48|0;
 $ts = sp + 40|0;
 HEAP32[$0>>2] = $s;
 $1 = ($s|0)==(0|0);
 if ($1) {
  HEAP32[$0>>2] = 1282445;
 }
 HEAP32[$vararg_buffer>>2] = 619565;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = 7;
 $2 = (___syscall33(33,($vararg_buffer|0))|0);
 $3 = ($2|0)==(0);
 if ($3) {
  $4 = ((($ts)) + 4|0);
  $5 = $0;
  $try$0 = 0;
  while(1) {
   HEAP32[$vararg_buffer2>>2] = 0;
   $vararg_ptr5 = ((($vararg_buffer2)) + 4|0);
   HEAP32[$vararg_ptr5>>2] = $ts;
   $vararg_ptr6 = ((($vararg_buffer2)) + 8|0);
   HEAP32[$vararg_ptr6>>2] = 0;
   (___syscall265(265,($vararg_buffer2|0))|0);
   $6 = HEAP32[$4>>2]|0;
   $7 = $6 ^ $5;
   $8 = HEAP32[$0>>2]|0;
   $9 = $8;
   $10 = $7 ^ $9;
   $11 = HEAP32[301506]|0;HEAP32[301506] = (($11+1)|0);
   HEAP32[$vararg_buffer7>>2] = $11;
   $vararg_ptr10 = ((($vararg_buffer7)) + 4|0);
   HEAP32[$vararg_ptr10>>2] = $10;
   (_snprintf($8,20,619570,$vararg_buffer7)|0);
   $12 = HEAP32[$0>>2]|0;
   HEAP32[$vararg_buffer11>>2] = $12;
   $vararg_ptr14 = ((($vararg_buffer11)) + 4|0);
   HEAP32[$vararg_ptr14>>2] = 0;
   $13 = (___syscall33(33,($vararg_buffer11|0))|0);
   $14 = ($13|0)==(0);
   if (!($14)) {
    $try$1 = $try$0;
    break;
   }
   $15 = (($try$0) + 1)|0;
   $16 = ($try$0|0)<(100);
   if ($16) {
    $try$0 = $15;
   } else {
    $try$1 = $15;
    break;
   }
  }
  $17 = ($try$1|0)>(99);
  $18 = HEAP32[$0>>2]|0;
  $19 = $17 ? 0 : $18;
  $$0 = $19;
 } else {
  $$0 = 0;
 }
 STACKTOP = sp;return ($$0|0);
}
function _vprintf($fmt,$ap) {
 $fmt = $fmt|0;
 $ap = $ap|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[37602]|0;
 $1 = (_vfprintf($0,$fmt,$ap)|0);
 return ($1|0);
}
function _atof($s) {
 $s = $s|0;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_strtod($s,0));
 return (+$0);
}
function _strtod($s,$p) {
 $s = $s|0;
 $p = $p|0;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_strtox($s,$p,1));
 return (+$0);
}
function _strtox($s,$p,$prec) {
 $s = $s|0;
 $p = $p|0;
 $prec = $prec|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $f = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $f = sp;
 dest=$f; stop=dest+112|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $0 = ((($f)) + 4|0);
 HEAP32[$0>>2] = $s;
 $1 = ((($f)) + 8|0);
 HEAP32[$1>>2] = (-1);
 $2 = ((($f)) + 44|0);
 HEAP32[$2>>2] = $s;
 $3 = ((($f)) + 76|0);
 HEAP32[$3>>2] = -1;
 ___shlim($f,0);
 $4 = (+___floatscan($f,$prec,1));
 $5 = ((($f)) + 108|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = HEAP32[$0>>2]|0;
 $8 = HEAP32[$1>>2]|0;
 $9 = (($7) - ($8))|0;
 $10 = (($9) + ($6))|0;
 $11 = ($p|0)==(0|0);
 if (!($11)) {
  $12 = ($10|0)!=(0);
  $13 = (($s) + ($10)|0);
  $14 = $12 ? $13 : $s;
  HEAP32[$p>>2] = $14;
 }
 STACKTOP = sp;return (+$4);
}
function _atoi($s) {
 $s = $s|0;
 var $$0 = 0, $$0$lcssa = 0, $$1$ph = 0, $$13 = 0, $$lcssa9 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $isdigit = 0, $isdigit2 = 0, $isdigittmp = 0, $isdigittmp1 = 0, $isdigittmp5 = 0, $n$0$lcssa = 0, $n$04 = 0, $neg$0 = 0, $neg$1$ph = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $s;
 while(1) {
  $0 = HEAP8[$$0>>0]|0;
  $1 = $0 << 24 >> 24;
  $2 = (_isspace($1)|0);
  $3 = ($2|0)==(0);
  $4 = ((($$0)) + 1|0);
  if ($3) {
   $$0$lcssa = $$0;$$lcssa9 = $4;
   break;
  } else {
   $$0 = $4;
  }
 }
 $5 = HEAP8[$$0$lcssa>>0]|0;
 $6 = $5 << 24 >> 24;
 switch ($6|0) {
 case 45:  {
  $neg$0 = 1;
  label = 5;
  break;
 }
 case 43:  {
  $neg$0 = 0;
  label = 5;
  break;
 }
 default: {
  $$1$ph = $$0$lcssa;$8 = $5;$neg$1$ph = 0;
 }
 }
 if ((label|0) == 5) {
  $$pre = HEAP8[$$lcssa9>>0]|0;
  $$1$ph = $$lcssa9;$8 = $$pre;$neg$1$ph = $neg$0;
 }
 $7 = $8 << 24 >> 24;
 $isdigittmp1 = (($7) + -48)|0;
 $isdigit2 = ($isdigittmp1>>>0)<(10);
 if ($isdigit2) {
  $$13 = $$1$ph;$isdigittmp5 = $isdigittmp1;$n$04 = 0;
  while(1) {
   $9 = ($n$04*10)|0;
   $10 = ((($$13)) + 1|0);
   $11 = (($9) - ($isdigittmp5))|0;
   $12 = HEAP8[$10>>0]|0;
   $13 = $12 << 24 >> 24;
   $isdigittmp = (($13) + -48)|0;
   $isdigit = ($isdigittmp>>>0)<(10);
   if ($isdigit) {
    $$13 = $10;$isdigittmp5 = $isdigittmp;$n$04 = $11;
   } else {
    $n$0$lcssa = $11;
    break;
   }
  }
 } else {
  $n$0$lcssa = 0;
 }
 $14 = ($neg$1$ph|0)!=(0);
 $15 = (0 - ($n$0$lcssa))|0;
 $16 = $14 ? $n$0$lcssa : $15;
 return ($16|0);
}
function _atol($s) {
 $s = $s|0;
 var $$0 = 0, $$0$lcssa = 0, $$1$ph = 0, $$13 = 0, $$lcssa9 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $isdigit = 0, $isdigit2 = 0, $isdigittmp = 0, $isdigittmp1 = 0, $isdigittmp5 = 0, $n$0$lcssa = 0, $n$04 = 0, $neg$0 = 0, $neg$1$ph = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $s;
 while(1) {
  $0 = HEAP8[$$0>>0]|0;
  $1 = $0 << 24 >> 24;
  $2 = (_isspace($1)|0);
  $3 = ($2|0)==(0);
  $4 = ((($$0)) + 1|0);
  if ($3) {
   $$0$lcssa = $$0;$$lcssa9 = $4;
   break;
  } else {
   $$0 = $4;
  }
 }
 $5 = HEAP8[$$0$lcssa>>0]|0;
 $6 = $5 << 24 >> 24;
 switch ($6|0) {
 case 45:  {
  $neg$0 = 1;
  label = 5;
  break;
 }
 case 43:  {
  $neg$0 = 0;
  label = 5;
  break;
 }
 default: {
  $$1$ph = $$0$lcssa;$8 = $5;$neg$1$ph = 0;
 }
 }
 if ((label|0) == 5) {
  $$pre = HEAP8[$$lcssa9>>0]|0;
  $$1$ph = $$lcssa9;$8 = $$pre;$neg$1$ph = $neg$0;
 }
 $7 = $8 << 24 >> 24;
 $isdigittmp1 = (($7) + -48)|0;
 $isdigit2 = ($isdigittmp1>>>0)<(10);
 if ($isdigit2) {
  $$13 = $$1$ph;$isdigittmp5 = $isdigittmp1;$n$04 = 0;
  while(1) {
   $9 = ($n$04*10)|0;
   $10 = ((($$13)) + 1|0);
   $11 = (($9) - ($isdigittmp5))|0;
   $12 = HEAP8[$10>>0]|0;
   $13 = $12 << 24 >> 24;
   $isdigittmp = (($13) + -48)|0;
   $isdigit = ($isdigittmp>>>0)<(10);
   if ($isdigit) {
    $$13 = $10;$isdigittmp5 = $isdigittmp;$n$04 = $11;
   } else {
    $n$0$lcssa = $11;
    break;
   }
  }
 } else {
  $n$0$lcssa = 0;
 }
 $14 = ($neg$1$ph|0)!=(0);
 $15 = (0 - ($n$0$lcssa))|0;
 $16 = $14 ? $n$0$lcssa : $15;
 return ($16|0);
}
function _strcspn($s,$c) {
 $s = $s|0;
 $c = $c|0;
 var $$0 = 0, $$027 = 0, $$03$lcssa = 0, $$035 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $byteset = 0, $div = 0, $div4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $byteset = sp;
 $0 = HEAP8[$c>>0]|0;
 $1 = ($0<<24>>24)==(0);
 if ($1) {
  label = 3;
 } else {
  $2 = ((($c)) + 1|0);
  $3 = HEAP8[$2>>0]|0;
  $4 = ($3<<24>>24)==(0);
  if ($4) {
   label = 3;
  } else {
   ;HEAP32[$byteset>>2]=0|0;HEAP32[$byteset+4>>2]=0|0;HEAP32[$byteset+8>>2]=0|0;HEAP32[$byteset+12>>2]=0|0;HEAP32[$byteset+16>>2]=0|0;HEAP32[$byteset+20>>2]=0|0;HEAP32[$byteset+24>>2]=0|0;HEAP32[$byteset+28>>2]=0|0;
   $$027 = $c;$13 = $0;
   while(1) {
    $12 = $13 & 31;
    $14 = $12&255;
    $15 = 1 << $14;
    $div4 = ($13&255) >>> 5;
    $16 = $div4&255;
    $17 = (($byteset) + ($16<<2)|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = $18 | $15;
    HEAP32[$17>>2] = $19;
    $20 = ((($$027)) + 1|0);
    $21 = HEAP8[$20>>0]|0;
    $22 = ($21<<24>>24)==(0);
    if ($22) {
     break;
    } else {
     $$027 = $20;$13 = $21;
    }
   }
   $10 = HEAP8[$s>>0]|0;
   $11 = ($10<<24>>24)==(0);
   L7: do {
    if ($11) {
     $$03$lcssa = $s;
    } else {
     $$035 = $s;$23 = $10;
     while(1) {
      $div = ($23&255) >>> 5;
      $24 = $div&255;
      $25 = (($byteset) + ($24<<2)|0);
      $26 = HEAP32[$25>>2]|0;
      $27 = $23 & 31;
      $28 = $27&255;
      $29 = 1 << $28;
      $30 = $26 & $29;
      $31 = ($30|0)==(0);
      if (!($31)) {
       $$03$lcssa = $$035;
       break L7;
      }
      $32 = ((($$035)) + 1|0);
      $33 = HEAP8[$32>>0]|0;
      $34 = ($33<<24>>24)==(0);
      if ($34) {
       $$03$lcssa = $32;
       break;
      } else {
       $$035 = $32;$23 = $33;
      }
     }
    }
   } while(0);
   $35 = $$03$lcssa;
   $36 = $s;
   $37 = (($35) - ($36))|0;
   $$0 = $37;
  }
 }
 if ((label|0) == 3) {
  $5 = $0 << 24 >> 24;
  $6 = (___strchrnul($s,$5)|0);
  $7 = $6;
  $8 = $s;
  $9 = (($7) - ($8))|0;
  $$0 = $9;
 }
 STACKTOP = sp;return ($$0|0);
}
function _strtold_l($s,$p,$l) {
 $s = $s|0;
 $p = $p|0;
 $l = $l|0;
 var $0 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (+_strtox($s,$p,2));
 return (+$0);
}
function _strcat($dest,$src) {
 $dest = $dest|0;
 $src = $src|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strlen($dest)|0);
 $1 = (($dest) + ($0)|0);
 (_strcpy($1,$src)|0);
 return ($dest|0);
}
function _strncat($d,$s,$n) {
 $d = $d|0;
 $s = $s|0;
 $n = $n|0;
 var $$0$lcssa = 0, $$014 = 0, $$023 = 0, $$05 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strlen($d)|0);
 $1 = (($d) + ($0)|0);
 $2 = ($n|0)==(0);
 L1: do {
  if ($2) {
   $$0$lcssa = $1;
  } else {
   $$014 = $n;$$023 = $s;$$05 = $1;
   while(1) {
    $3 = HEAP8[$$023>>0]|0;
    $4 = ($3<<24>>24)==(0);
    if ($4) {
     $$0$lcssa = $$05;
     break L1;
    }
    $5 = (($$014) + -1)|0;
    $6 = ((($$023)) + 1|0);
    $7 = ((($$05)) + 1|0);
    HEAP8[$$05>>0] = $3;
    $8 = ($5|0)==(0);
    if ($8) {
     $$0$lcssa = $7;
     break;
    } else {
     $$014 = $5;$$023 = $6;$$05 = $7;
    }
   }
  }
 } while(0);
 HEAP8[$$0$lcssa>>0] = 0;
 return ($d|0);
}
function _strpbrk($s,$b) {
 $s = $s|0;
 $b = $b|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strcspn($s,$b)|0);
 $1 = (($s) + ($0)|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = ($2<<24>>24)!=(0);
 $4 = $3 ? $1 : 0;
 return ($4|0);
}
function _wmemset($d,$c,$n) {
 $d = $d|0;
 $c = $c|0;
 $n = $n|0;
 var $$012 = 0, $$03 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)==(0);
 if (!($0)) {
  $$012 = $n;$$03 = $d;
  while(1) {
   $1 = (($$012) + -1)|0;
   $2 = ((($$03)) + 4|0);
   HEAP32[$$03>>2] = $c;
   $3 = ($1|0)==(0);
   if ($3) {
    break;
   } else {
    $$012 = $1;$$03 = $2;
   }
  }
 }
 return ($d|0);
}
function _wmemmove($d,$s,$n) {
 $d = $d|0;
 $s = $s|0;
 $n = $n|0;
 var $$025 = 0, $$07 = 0, $$16 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $d;
 $1 = $s;
 $2 = (($0) - ($1))|0;
 $3 = $2 >> 2;
 $4 = ($3>>>0)<($n>>>0);
 $5 = ($n|0)==(0);
 if ($4) {
  if (!($5)) {
   $$in = $n;
   while(1) {
    $6 = (($$in) + -1)|0;
    $7 = (($s) + ($6<<2)|0);
    $8 = HEAP32[$7>>2]|0;
    $9 = (($d) + ($6<<2)|0);
    HEAP32[$9>>2] = $8;
    $10 = ($6|0)==(0);
    if ($10) {
     break;
    } else {
     $$in = $6;
    }
   }
  }
 } else {
  if (!($5)) {
   $$025 = $s;$$07 = $d;$$16 = $n;
   while(1) {
    $11 = (($$16) + -1)|0;
    $12 = ((($$025)) + 4|0);
    $13 = HEAP32[$$025>>2]|0;
    $14 = ((($$07)) + 4|0);
    HEAP32[$$07>>2] = $13;
    $15 = ($11|0)==(0);
    if ($15) {
     break;
    } else {
     $$025 = $12;$$07 = $14;$$16 = $11;
    }
   }
  }
 }
 return ($d|0);
}
function _getuid() {
 var $0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (___syscall199(199,($vararg_buffer|0))|0);
 STACKTOP = sp;return ($0|0);
}
function _fsync($fd) {
 $fd = $fd|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 HEAP32[$vararg_buffer>>2] = $fd;
 $0 = (___syscall118(118,($vararg_buffer|0))|0);
 $1 = (___syscall_ret($0)|0);
 STACKTOP = sp;return ($1|0);
}
function _gethostname($name,$len) {
 $name = $name|0;
 $len = $len|0;
 var $$01 = 0, $$len = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0$lcssa = 0, $i$02 = 0, $uts = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 400|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $uts = sp;
 $0 = (_uname($uts)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  $2 = ($len>>>0)>(65);
  $$len = $2 ? 65 : $len;
  $3 = ($$len|0)==(0);
  L3: do {
   if ($3) {
    $i$0$lcssa = 0;
   } else {
    $i$02 = 0;
    while(1) {
     $4 = (((($uts)) + 65|0) + ($i$02)|0);
     $5 = HEAP8[$4>>0]|0;
     $6 = (($name) + ($i$02)|0);
     HEAP8[$6>>0] = $5;
     $7 = ($5<<24>>24)==(0);
     if ($7) {
      $i$0$lcssa = $i$02;
      break L3;
     }
     $8 = (($i$02) + 1)|0;
     $9 = ($8>>>0)<($$len>>>0);
     if ($9) {
      $i$02 = $8;
     } else {
      $i$0$lcssa = $8;
      break;
     }
    }
   }
  } while(0);
  $10 = ($i$0$lcssa|0)==($$len|0);
  if ($10) {
   $11 = (($$len) + -1)|0;
   $12 = (($name) + ($11)|0);
   HEAP8[$12>>0] = 0;
   $$01 = 0;
  } else {
   $$01 = 0;
  }
 } else {
  $$01 = -1;
 }
 STACKTOP = sp;return ($$01|0);
}
function _read($fd,$buf,$count) {
 $fd = $fd|0;
 $buf = $buf|0;
 $count = $count|0;
 var $0 = 0, $1 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 HEAP32[$vararg_buffer>>2] = $fd;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $buf;
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
 HEAP32[$vararg_ptr2>>2] = $count;
 $0 = (___syscall3(3,($vararg_buffer|0))|0);
 $1 = (___syscall_ret($0)|0);
 STACKTOP = sp;return ($1|0);
}
function _sleep($seconds) {
 $seconds = $seconds|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $tv = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tv = sp;
 HEAP32[$tv>>2] = $seconds;
 $0 = ((($tv)) + 4|0);
 HEAP32[$0>>2] = 0;
 $1 = (_nanosleep(($tv|0),($tv|0))|0);
 $2 = ($1|0)==(0);
 $3 = HEAP32[$tv>>2]|0;
 $$0 = $2 ? 0 : $3;
 STACKTOP = sp;return ($$0|0);
}
function _malloc($bytes) {
 $bytes = $bytes|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa141 = 0, $$lcssa142 = 0, $$lcssa144 = 0, $$lcssa147 = 0, $$lcssa149 = 0, $$lcssa151 = 0, $$lcssa153 = 0, $$lcssa155 = 0, $$lcssa157 = 0, $$not$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i13 = 0, $$pre$i16$i = 0, $$pre$phi$i$iZ2D = 0, $$pre$phi$i14Z2D = 0, $$pre$phi$i17$iZ2D = 0;
 var $$pre$phi$iZ2D = 0, $$pre$phi10$i$iZ2D = 0, $$pre$phiZ2D = 0, $$pre71 = 0, $$pre9$i$i = 0, $$rsize$0$i = 0, $$rsize$4$i = 0, $$v$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0;
 var $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0;
 var $1026 = 0, $1027 = 0, $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0;
 var $1044 = 0, $1045 = 0, $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0;
 var $1062 = 0, $1063 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0;
 var $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0;
 var $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0;
 var $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0;
 var $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0;
 var $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0;
 var $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0;
 var $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0;
 var $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0;
 var $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0;
 var $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0;
 var $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0;
 var $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0;
 var $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0;
 var $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0;
 var $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0;
 var $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0;
 var $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0;
 var $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0;
 var $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0;
 var $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0;
 var $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0;
 var $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0;
 var $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0;
 var $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0;
 var $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0;
 var $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0;
 var $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0;
 var $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0;
 var $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0;
 var $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0;
 var $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0;
 var $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0;
 var $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0;
 var $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0;
 var $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0;
 var $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0;
 var $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0;
 var $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0;
 var $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0;
 var $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0;
 var $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0;
 var $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0;
 var $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0;
 var $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0;
 var $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $F$0$i$i = 0, $F1$0$i = 0, $F4$0 = 0, $F4$0$i$i = 0, $F5$0$i = 0, $I1$0$i$i = 0, $I7$0$i = 0, $I7$0$i$i = 0;
 var $K12$0$i = 0, $K2$0$i$i = 0, $K8$0$i$i = 0, $R$1$i = 0, $R$1$i$i = 0, $R$1$i$i$lcssa = 0, $R$1$i$lcssa = 0, $R$1$i9 = 0, $R$1$i9$lcssa = 0, $R$3$i = 0, $R$3$i$i = 0, $R$3$i11 = 0, $RP$1$i = 0, $RP$1$i$i = 0, $RP$1$i$i$lcssa = 0, $RP$1$i$lcssa = 0, $RP$1$i8 = 0, $RP$1$i8$lcssa = 0, $T$0$i = 0, $T$0$i$i = 0;
 var $T$0$i$i$lcssa = 0, $T$0$i$i$lcssa140 = 0, $T$0$i$lcssa = 0, $T$0$i$lcssa156 = 0, $T$0$i18$i = 0, $T$0$i18$i$lcssa = 0, $T$0$i18$i$lcssa139 = 0, $br$2$ph$i = 0, $cond$i = 0, $cond$i$i = 0, $cond$i12 = 0, $exitcond$i$i = 0, $i$01$i$i = 0, $idx$0$i = 0, $nb$0 = 0, $not$$i$i = 0, $not$$i20$i = 0, $not$7$i = 0, $oldfirst$0$i$i = 0, $or$cond$i = 0;
 var $or$cond$i17 = 0, $or$cond1$i = 0, $or$cond1$i16 = 0, $or$cond10$i = 0, $or$cond11$i = 0, $or$cond2$i = 0, $or$cond48$i = 0, $or$cond5$i = 0, $or$cond7$i = 0, $or$cond8$i = 0, $p$0$i$i = 0, $qsize$0$i$i = 0, $rsize$0$i = 0, $rsize$0$i$lcssa = 0, $rsize$0$i5 = 0, $rsize$1$i = 0, $rsize$3$i = 0, $rsize$4$lcssa$i = 0, $rsize$412$i = 0, $rst$0$i = 0;
 var $rst$1$i = 0, $sizebits$0$$i = 0, $sizebits$0$i = 0, $sp$0$i$i = 0, $sp$0$i$i$i = 0, $sp$068$i = 0, $sp$068$i$lcssa = 0, $sp$167$i = 0, $sp$167$i$lcssa = 0, $ssize$0$i = 0, $ssize$2$ph$i = 0, $ssize$5$i = 0, $t$0$i = 0, $t$0$i4 = 0, $t$2$i = 0, $t$4$ph$i = 0, $t$4$v$4$i = 0, $t$411$i = 0, $tbase$746$i = 0, $tsize$745$i = 0;
 var $v$0$i = 0, $v$0$i$lcssa = 0, $v$0$i6 = 0, $v$1$i = 0, $v$3$i = 0, $v$4$lcssa$i = 0, $v$413$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($bytes>>>0)<(245);
 do {
  if ($0) {
   $1 = ($bytes>>>0)<(11);
   $2 = (($bytes) + 11)|0;
   $3 = $2 & -8;
   $4 = $1 ? 16 : $3;
   $5 = $4 >>> 3;
   $6 = HEAP32[301507]|0;
   $7 = $6 >>> $5;
   $8 = $7 & 3;
   $9 = ($8|0)==(0);
   if (!($9)) {
    $10 = $7 & 1;
    $11 = $10 ^ 1;
    $12 = (($11) + ($5))|0;
    $13 = $12 << 1;
    $14 = (1206068 + ($13<<2)|0);
    $15 = ((($14)) + 8|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($16)) + 8|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = ($14|0)==($18|0);
    do {
     if ($19) {
      $20 = 1 << $12;
      $21 = $20 ^ -1;
      $22 = $6 & $21;
      HEAP32[301507] = $22;
     } else {
      $23 = HEAP32[(1206044)>>2]|0;
      $24 = ($18>>>0)<($23>>>0);
      if ($24) {
       _abort();
       // unreachable;
      }
      $25 = ((($18)) + 12|0);
      $26 = HEAP32[$25>>2]|0;
      $27 = ($26|0)==($16|0);
      if ($27) {
       HEAP32[$25>>2] = $14;
       HEAP32[$15>>2] = $18;
       break;
      } else {
       _abort();
       // unreachable;
      }
     }
    } while(0);
    $28 = $12 << 3;
    $29 = $28 | 3;
    $30 = ((($16)) + 4|0);
    HEAP32[$30>>2] = $29;
    $31 = (($16) + ($28)|0);
    $32 = ((($31)) + 4|0);
    $33 = HEAP32[$32>>2]|0;
    $34 = $33 | 1;
    HEAP32[$32>>2] = $34;
    $$0 = $17;
    return ($$0|0);
   }
   $35 = HEAP32[(1206036)>>2]|0;
   $36 = ($4>>>0)>($35>>>0);
   if ($36) {
    $37 = ($7|0)==(0);
    if (!($37)) {
     $38 = $7 << $5;
     $39 = 2 << $5;
     $40 = (0 - ($39))|0;
     $41 = $39 | $40;
     $42 = $38 & $41;
     $43 = (0 - ($42))|0;
     $44 = $42 & $43;
     $45 = (($44) + -1)|0;
     $46 = $45 >>> 12;
     $47 = $46 & 16;
     $48 = $45 >>> $47;
     $49 = $48 >>> 5;
     $50 = $49 & 8;
     $51 = $50 | $47;
     $52 = $48 >>> $50;
     $53 = $52 >>> 2;
     $54 = $53 & 4;
     $55 = $51 | $54;
     $56 = $52 >>> $54;
     $57 = $56 >>> 1;
     $58 = $57 & 2;
     $59 = $55 | $58;
     $60 = $56 >>> $58;
     $61 = $60 >>> 1;
     $62 = $61 & 1;
     $63 = $59 | $62;
     $64 = $60 >>> $62;
     $65 = (($63) + ($64))|0;
     $66 = $65 << 1;
     $67 = (1206068 + ($66<<2)|0);
     $68 = ((($67)) + 8|0);
     $69 = HEAP32[$68>>2]|0;
     $70 = ((($69)) + 8|0);
     $71 = HEAP32[$70>>2]|0;
     $72 = ($67|0)==($71|0);
     do {
      if ($72) {
       $73 = 1 << $65;
       $74 = $73 ^ -1;
       $75 = $6 & $74;
       HEAP32[301507] = $75;
       $90 = $35;
      } else {
       $76 = HEAP32[(1206044)>>2]|0;
       $77 = ($71>>>0)<($76>>>0);
       if ($77) {
        _abort();
        // unreachable;
       }
       $78 = ((($71)) + 12|0);
       $79 = HEAP32[$78>>2]|0;
       $80 = ($79|0)==($69|0);
       if ($80) {
        HEAP32[$78>>2] = $67;
        HEAP32[$68>>2] = $71;
        $$pre = HEAP32[(1206036)>>2]|0;
        $90 = $$pre;
        break;
       } else {
        _abort();
        // unreachable;
       }
      }
     } while(0);
     $81 = $65 << 3;
     $82 = (($81) - ($4))|0;
     $83 = $4 | 3;
     $84 = ((($69)) + 4|0);
     HEAP32[$84>>2] = $83;
     $85 = (($69) + ($4)|0);
     $86 = $82 | 1;
     $87 = ((($85)) + 4|0);
     HEAP32[$87>>2] = $86;
     $88 = (($85) + ($82)|0);
     HEAP32[$88>>2] = $82;
     $89 = ($90|0)==(0);
     if (!($89)) {
      $91 = HEAP32[(1206048)>>2]|0;
      $92 = $90 >>> 3;
      $93 = $92 << 1;
      $94 = (1206068 + ($93<<2)|0);
      $95 = HEAP32[301507]|0;
      $96 = 1 << $92;
      $97 = $95 & $96;
      $98 = ($97|0)==(0);
      if ($98) {
       $99 = $95 | $96;
       HEAP32[301507] = $99;
       $$pre71 = ((($94)) + 8|0);
       $$pre$phiZ2D = $$pre71;$F4$0 = $94;
      } else {
       $100 = ((($94)) + 8|0);
       $101 = HEAP32[$100>>2]|0;
       $102 = HEAP32[(1206044)>>2]|0;
       $103 = ($101>>>0)<($102>>>0);
       if ($103) {
        _abort();
        // unreachable;
       } else {
        $$pre$phiZ2D = $100;$F4$0 = $101;
       }
      }
      HEAP32[$$pre$phiZ2D>>2] = $91;
      $104 = ((($F4$0)) + 12|0);
      HEAP32[$104>>2] = $91;
      $105 = ((($91)) + 8|0);
      HEAP32[$105>>2] = $F4$0;
      $106 = ((($91)) + 12|0);
      HEAP32[$106>>2] = $94;
     }
     HEAP32[(1206036)>>2] = $82;
     HEAP32[(1206048)>>2] = $85;
     $$0 = $70;
     return ($$0|0);
    }
    $107 = HEAP32[(1206032)>>2]|0;
    $108 = ($107|0)==(0);
    if ($108) {
     $nb$0 = $4;
    } else {
     $109 = (0 - ($107))|0;
     $110 = $107 & $109;
     $111 = (($110) + -1)|0;
     $112 = $111 >>> 12;
     $113 = $112 & 16;
     $114 = $111 >>> $113;
     $115 = $114 >>> 5;
     $116 = $115 & 8;
     $117 = $116 | $113;
     $118 = $114 >>> $116;
     $119 = $118 >>> 2;
     $120 = $119 & 4;
     $121 = $117 | $120;
     $122 = $118 >>> $120;
     $123 = $122 >>> 1;
     $124 = $123 & 2;
     $125 = $121 | $124;
     $126 = $122 >>> $124;
     $127 = $126 >>> 1;
     $128 = $127 & 1;
     $129 = $125 | $128;
     $130 = $126 >>> $128;
     $131 = (($129) + ($130))|0;
     $132 = (1206332 + ($131<<2)|0);
     $133 = HEAP32[$132>>2]|0;
     $134 = ((($133)) + 4|0);
     $135 = HEAP32[$134>>2]|0;
     $136 = $135 & -8;
     $137 = (($136) - ($4))|0;
     $rsize$0$i = $137;$t$0$i = $133;$v$0$i = $133;
     while(1) {
      $138 = ((($t$0$i)) + 16|0);
      $139 = HEAP32[$138>>2]|0;
      $140 = ($139|0)==(0|0);
      if ($140) {
       $141 = ((($t$0$i)) + 20|0);
       $142 = HEAP32[$141>>2]|0;
       $143 = ($142|0)==(0|0);
       if ($143) {
        $rsize$0$i$lcssa = $rsize$0$i;$v$0$i$lcssa = $v$0$i;
        break;
       } else {
        $145 = $142;
       }
      } else {
       $145 = $139;
      }
      $144 = ((($145)) + 4|0);
      $146 = HEAP32[$144>>2]|0;
      $147 = $146 & -8;
      $148 = (($147) - ($4))|0;
      $149 = ($148>>>0)<($rsize$0$i>>>0);
      $$rsize$0$i = $149 ? $148 : $rsize$0$i;
      $$v$0$i = $149 ? $145 : $v$0$i;
      $rsize$0$i = $$rsize$0$i;$t$0$i = $145;$v$0$i = $$v$0$i;
     }
     $150 = HEAP32[(1206044)>>2]|0;
     $151 = ($v$0$i$lcssa>>>0)<($150>>>0);
     if ($151) {
      _abort();
      // unreachable;
     }
     $152 = (($v$0$i$lcssa) + ($4)|0);
     $153 = ($v$0$i$lcssa>>>0)<($152>>>0);
     if (!($153)) {
      _abort();
      // unreachable;
     }
     $154 = ((($v$0$i$lcssa)) + 24|0);
     $155 = HEAP32[$154>>2]|0;
     $156 = ((($v$0$i$lcssa)) + 12|0);
     $157 = HEAP32[$156>>2]|0;
     $158 = ($157|0)==($v$0$i$lcssa|0);
     do {
      if ($158) {
       $168 = ((($v$0$i$lcssa)) + 20|0);
       $169 = HEAP32[$168>>2]|0;
       $170 = ($169|0)==(0|0);
       if ($170) {
        $171 = ((($v$0$i$lcssa)) + 16|0);
        $172 = HEAP32[$171>>2]|0;
        $173 = ($172|0)==(0|0);
        if ($173) {
         $R$3$i = 0;
         break;
        } else {
         $R$1$i = $172;$RP$1$i = $171;
        }
       } else {
        $R$1$i = $169;$RP$1$i = $168;
       }
       while(1) {
        $174 = ((($R$1$i)) + 20|0);
        $175 = HEAP32[$174>>2]|0;
        $176 = ($175|0)==(0|0);
        if (!($176)) {
         $R$1$i = $175;$RP$1$i = $174;
         continue;
        }
        $177 = ((($R$1$i)) + 16|0);
        $178 = HEAP32[$177>>2]|0;
        $179 = ($178|0)==(0|0);
        if ($179) {
         $R$1$i$lcssa = $R$1$i;$RP$1$i$lcssa = $RP$1$i;
         break;
        } else {
         $R$1$i = $178;$RP$1$i = $177;
        }
       }
       $180 = ($RP$1$i$lcssa>>>0)<($150>>>0);
       if ($180) {
        _abort();
        // unreachable;
       } else {
        HEAP32[$RP$1$i$lcssa>>2] = 0;
        $R$3$i = $R$1$i$lcssa;
        break;
       }
      } else {
       $159 = ((($v$0$i$lcssa)) + 8|0);
       $160 = HEAP32[$159>>2]|0;
       $161 = ($160>>>0)<($150>>>0);
       if ($161) {
        _abort();
        // unreachable;
       }
       $162 = ((($160)) + 12|0);
       $163 = HEAP32[$162>>2]|0;
       $164 = ($163|0)==($v$0$i$lcssa|0);
       if (!($164)) {
        _abort();
        // unreachable;
       }
       $165 = ((($157)) + 8|0);
       $166 = HEAP32[$165>>2]|0;
       $167 = ($166|0)==($v$0$i$lcssa|0);
       if ($167) {
        HEAP32[$162>>2] = $157;
        HEAP32[$165>>2] = $160;
        $R$3$i = $157;
        break;
       } else {
        _abort();
        // unreachable;
       }
      }
     } while(0);
     $181 = ($155|0)==(0|0);
     do {
      if (!($181)) {
       $182 = ((($v$0$i$lcssa)) + 28|0);
       $183 = HEAP32[$182>>2]|0;
       $184 = (1206332 + ($183<<2)|0);
       $185 = HEAP32[$184>>2]|0;
       $186 = ($v$0$i$lcssa|0)==($185|0);
       if ($186) {
        HEAP32[$184>>2] = $R$3$i;
        $cond$i = ($R$3$i|0)==(0|0);
        if ($cond$i) {
         $187 = 1 << $183;
         $188 = $187 ^ -1;
         $189 = HEAP32[(1206032)>>2]|0;
         $190 = $189 & $188;
         HEAP32[(1206032)>>2] = $190;
         break;
        }
       } else {
        $191 = HEAP32[(1206044)>>2]|0;
        $192 = ($155>>>0)<($191>>>0);
        if ($192) {
         _abort();
         // unreachable;
        }
        $193 = ((($155)) + 16|0);
        $194 = HEAP32[$193>>2]|0;
        $195 = ($194|0)==($v$0$i$lcssa|0);
        if ($195) {
         HEAP32[$193>>2] = $R$3$i;
        } else {
         $196 = ((($155)) + 20|0);
         HEAP32[$196>>2] = $R$3$i;
        }
        $197 = ($R$3$i|0)==(0|0);
        if ($197) {
         break;
        }
       }
       $198 = HEAP32[(1206044)>>2]|0;
       $199 = ($R$3$i>>>0)<($198>>>0);
       if ($199) {
        _abort();
        // unreachable;
       }
       $200 = ((($R$3$i)) + 24|0);
       HEAP32[$200>>2] = $155;
       $201 = ((($v$0$i$lcssa)) + 16|0);
       $202 = HEAP32[$201>>2]|0;
       $203 = ($202|0)==(0|0);
       do {
        if (!($203)) {
         $204 = ($202>>>0)<($198>>>0);
         if ($204) {
          _abort();
          // unreachable;
         } else {
          $205 = ((($R$3$i)) + 16|0);
          HEAP32[$205>>2] = $202;
          $206 = ((($202)) + 24|0);
          HEAP32[$206>>2] = $R$3$i;
          break;
         }
        }
       } while(0);
       $207 = ((($v$0$i$lcssa)) + 20|0);
       $208 = HEAP32[$207>>2]|0;
       $209 = ($208|0)==(0|0);
       if (!($209)) {
        $210 = HEAP32[(1206044)>>2]|0;
        $211 = ($208>>>0)<($210>>>0);
        if ($211) {
         _abort();
         // unreachable;
        } else {
         $212 = ((($R$3$i)) + 20|0);
         HEAP32[$212>>2] = $208;
         $213 = ((($208)) + 24|0);
         HEAP32[$213>>2] = $R$3$i;
         break;
        }
       }
      }
     } while(0);
     $214 = ($rsize$0$i$lcssa>>>0)<(16);
     if ($214) {
      $215 = (($rsize$0$i$lcssa) + ($4))|0;
      $216 = $215 | 3;
      $217 = ((($v$0$i$lcssa)) + 4|0);
      HEAP32[$217>>2] = $216;
      $218 = (($v$0$i$lcssa) + ($215)|0);
      $219 = ((($218)) + 4|0);
      $220 = HEAP32[$219>>2]|0;
      $221 = $220 | 1;
      HEAP32[$219>>2] = $221;
     } else {
      $222 = $4 | 3;
      $223 = ((($v$0$i$lcssa)) + 4|0);
      HEAP32[$223>>2] = $222;
      $224 = $rsize$0$i$lcssa | 1;
      $225 = ((($152)) + 4|0);
      HEAP32[$225>>2] = $224;
      $226 = (($152) + ($rsize$0$i$lcssa)|0);
      HEAP32[$226>>2] = $rsize$0$i$lcssa;
      $227 = HEAP32[(1206036)>>2]|0;
      $228 = ($227|0)==(0);
      if (!($228)) {
       $229 = HEAP32[(1206048)>>2]|0;
       $230 = $227 >>> 3;
       $231 = $230 << 1;
       $232 = (1206068 + ($231<<2)|0);
       $233 = HEAP32[301507]|0;
       $234 = 1 << $230;
       $235 = $233 & $234;
       $236 = ($235|0)==(0);
       if ($236) {
        $237 = $233 | $234;
        HEAP32[301507] = $237;
        $$pre$i = ((($232)) + 8|0);
        $$pre$phi$iZ2D = $$pre$i;$F1$0$i = $232;
       } else {
        $238 = ((($232)) + 8|0);
        $239 = HEAP32[$238>>2]|0;
        $240 = HEAP32[(1206044)>>2]|0;
        $241 = ($239>>>0)<($240>>>0);
        if ($241) {
         _abort();
         // unreachable;
        } else {
         $$pre$phi$iZ2D = $238;$F1$0$i = $239;
        }
       }
       HEAP32[$$pre$phi$iZ2D>>2] = $229;
       $242 = ((($F1$0$i)) + 12|0);
       HEAP32[$242>>2] = $229;
       $243 = ((($229)) + 8|0);
       HEAP32[$243>>2] = $F1$0$i;
       $244 = ((($229)) + 12|0);
       HEAP32[$244>>2] = $232;
      }
      HEAP32[(1206036)>>2] = $rsize$0$i$lcssa;
      HEAP32[(1206048)>>2] = $152;
     }
     $245 = ((($v$0$i$lcssa)) + 8|0);
     $$0 = $245;
     return ($$0|0);
    }
   } else {
    $nb$0 = $4;
   }
  } else {
   $246 = ($bytes>>>0)>(4294967231);
   if ($246) {
    $nb$0 = -1;
   } else {
    $247 = (($bytes) + 11)|0;
    $248 = $247 & -8;
    $249 = HEAP32[(1206032)>>2]|0;
    $250 = ($249|0)==(0);
    if ($250) {
     $nb$0 = $248;
    } else {
     $251 = (0 - ($248))|0;
     $252 = $247 >>> 8;
     $253 = ($252|0)==(0);
     if ($253) {
      $idx$0$i = 0;
     } else {
      $254 = ($248>>>0)>(16777215);
      if ($254) {
       $idx$0$i = 31;
      } else {
       $255 = (($252) + 1048320)|0;
       $256 = $255 >>> 16;
       $257 = $256 & 8;
       $258 = $252 << $257;
       $259 = (($258) + 520192)|0;
       $260 = $259 >>> 16;
       $261 = $260 & 4;
       $262 = $261 | $257;
       $263 = $258 << $261;
       $264 = (($263) + 245760)|0;
       $265 = $264 >>> 16;
       $266 = $265 & 2;
       $267 = $262 | $266;
       $268 = (14 - ($267))|0;
       $269 = $263 << $266;
       $270 = $269 >>> 15;
       $271 = (($268) + ($270))|0;
       $272 = $271 << 1;
       $273 = (($271) + 7)|0;
       $274 = $248 >>> $273;
       $275 = $274 & 1;
       $276 = $275 | $272;
       $idx$0$i = $276;
      }
     }
     $277 = (1206332 + ($idx$0$i<<2)|0);
     $278 = HEAP32[$277>>2]|0;
     $279 = ($278|0)==(0|0);
     L123: do {
      if ($279) {
       $rsize$3$i = $251;$t$2$i = 0;$v$3$i = 0;
       label = 86;
      } else {
       $280 = ($idx$0$i|0)==(31);
       $281 = $idx$0$i >>> 1;
       $282 = (25 - ($281))|0;
       $283 = $280 ? 0 : $282;
       $284 = $248 << $283;
       $rsize$0$i5 = $251;$rst$0$i = 0;$sizebits$0$i = $284;$t$0$i4 = $278;$v$0$i6 = 0;
       while(1) {
        $285 = ((($t$0$i4)) + 4|0);
        $286 = HEAP32[$285>>2]|0;
        $287 = $286 & -8;
        $288 = (($287) - ($248))|0;
        $289 = ($288>>>0)<($rsize$0$i5>>>0);
        if ($289) {
         $290 = ($287|0)==($248|0);
         if ($290) {
          $rsize$412$i = $288;$t$411$i = $t$0$i4;$v$413$i = $t$0$i4;
          label = 90;
          break L123;
         } else {
          $rsize$1$i = $288;$v$1$i = $t$0$i4;
         }
        } else {
         $rsize$1$i = $rsize$0$i5;$v$1$i = $v$0$i6;
        }
        $291 = ((($t$0$i4)) + 20|0);
        $292 = HEAP32[$291>>2]|0;
        $293 = $sizebits$0$i >>> 31;
        $294 = (((($t$0$i4)) + 16|0) + ($293<<2)|0);
        $295 = HEAP32[$294>>2]|0;
        $296 = ($292|0)==(0|0);
        $297 = ($292|0)==($295|0);
        $or$cond1$i = $296 | $297;
        $rst$1$i = $or$cond1$i ? $rst$0$i : $292;
        $298 = ($295|0)==(0|0);
        $299 = $298&1;
        $300 = $299 ^ 1;
        $sizebits$0$$i = $sizebits$0$i << $300;
        if ($298) {
         $rsize$3$i = $rsize$1$i;$t$2$i = $rst$1$i;$v$3$i = $v$1$i;
         label = 86;
         break;
        } else {
         $rsize$0$i5 = $rsize$1$i;$rst$0$i = $rst$1$i;$sizebits$0$i = $sizebits$0$$i;$t$0$i4 = $295;$v$0$i6 = $v$1$i;
        }
       }
      }
     } while(0);
     if ((label|0) == 86) {
      $301 = ($t$2$i|0)==(0|0);
      $302 = ($v$3$i|0)==(0|0);
      $or$cond$i = $301 & $302;
      if ($or$cond$i) {
       $303 = 2 << $idx$0$i;
       $304 = (0 - ($303))|0;
       $305 = $303 | $304;
       $306 = $249 & $305;
       $307 = ($306|0)==(0);
       if ($307) {
        $nb$0 = $248;
        break;
       }
       $308 = (0 - ($306))|0;
       $309 = $306 & $308;
       $310 = (($309) + -1)|0;
       $311 = $310 >>> 12;
       $312 = $311 & 16;
       $313 = $310 >>> $312;
       $314 = $313 >>> 5;
       $315 = $314 & 8;
       $316 = $315 | $312;
       $317 = $313 >>> $315;
       $318 = $317 >>> 2;
       $319 = $318 & 4;
       $320 = $316 | $319;
       $321 = $317 >>> $319;
       $322 = $321 >>> 1;
       $323 = $322 & 2;
       $324 = $320 | $323;
       $325 = $321 >>> $323;
       $326 = $325 >>> 1;
       $327 = $326 & 1;
       $328 = $324 | $327;
       $329 = $325 >>> $327;
       $330 = (($328) + ($329))|0;
       $331 = (1206332 + ($330<<2)|0);
       $332 = HEAP32[$331>>2]|0;
       $t$4$ph$i = $332;
      } else {
       $t$4$ph$i = $t$2$i;
      }
      $333 = ($t$4$ph$i|0)==(0|0);
      if ($333) {
       $rsize$4$lcssa$i = $rsize$3$i;$v$4$lcssa$i = $v$3$i;
      } else {
       $rsize$412$i = $rsize$3$i;$t$411$i = $t$4$ph$i;$v$413$i = $v$3$i;
       label = 90;
      }
     }
     if ((label|0) == 90) {
      while(1) {
       label = 0;
       $334 = ((($t$411$i)) + 4|0);
       $335 = HEAP32[$334>>2]|0;
       $336 = $335 & -8;
       $337 = (($336) - ($248))|0;
       $338 = ($337>>>0)<($rsize$412$i>>>0);
       $$rsize$4$i = $338 ? $337 : $rsize$412$i;
       $t$4$v$4$i = $338 ? $t$411$i : $v$413$i;
       $339 = ((($t$411$i)) + 16|0);
       $340 = HEAP32[$339>>2]|0;
       $341 = ($340|0)==(0|0);
       if (!($341)) {
        $rsize$412$i = $$rsize$4$i;$t$411$i = $340;$v$413$i = $t$4$v$4$i;
        label = 90;
        continue;
       }
       $342 = ((($t$411$i)) + 20|0);
       $343 = HEAP32[$342>>2]|0;
       $344 = ($343|0)==(0|0);
       if ($344) {
        $rsize$4$lcssa$i = $$rsize$4$i;$v$4$lcssa$i = $t$4$v$4$i;
        break;
       } else {
        $rsize$412$i = $$rsize$4$i;$t$411$i = $343;$v$413$i = $t$4$v$4$i;
        label = 90;
       }
      }
     }
     $345 = ($v$4$lcssa$i|0)==(0|0);
     if ($345) {
      $nb$0 = $248;
     } else {
      $346 = HEAP32[(1206036)>>2]|0;
      $347 = (($346) - ($248))|0;
      $348 = ($rsize$4$lcssa$i>>>0)<($347>>>0);
      if ($348) {
       $349 = HEAP32[(1206044)>>2]|0;
       $350 = ($v$4$lcssa$i>>>0)<($349>>>0);
       if ($350) {
        _abort();
        // unreachable;
       }
       $351 = (($v$4$lcssa$i) + ($248)|0);
       $352 = ($v$4$lcssa$i>>>0)<($351>>>0);
       if (!($352)) {
        _abort();
        // unreachable;
       }
       $353 = ((($v$4$lcssa$i)) + 24|0);
       $354 = HEAP32[$353>>2]|0;
       $355 = ((($v$4$lcssa$i)) + 12|0);
       $356 = HEAP32[$355>>2]|0;
       $357 = ($356|0)==($v$4$lcssa$i|0);
       do {
        if ($357) {
         $367 = ((($v$4$lcssa$i)) + 20|0);
         $368 = HEAP32[$367>>2]|0;
         $369 = ($368|0)==(0|0);
         if ($369) {
          $370 = ((($v$4$lcssa$i)) + 16|0);
          $371 = HEAP32[$370>>2]|0;
          $372 = ($371|0)==(0|0);
          if ($372) {
           $R$3$i11 = 0;
           break;
          } else {
           $R$1$i9 = $371;$RP$1$i8 = $370;
          }
         } else {
          $R$1$i9 = $368;$RP$1$i8 = $367;
         }
         while(1) {
          $373 = ((($R$1$i9)) + 20|0);
          $374 = HEAP32[$373>>2]|0;
          $375 = ($374|0)==(0|0);
          if (!($375)) {
           $R$1$i9 = $374;$RP$1$i8 = $373;
           continue;
          }
          $376 = ((($R$1$i9)) + 16|0);
          $377 = HEAP32[$376>>2]|0;
          $378 = ($377|0)==(0|0);
          if ($378) {
           $R$1$i9$lcssa = $R$1$i9;$RP$1$i8$lcssa = $RP$1$i8;
           break;
          } else {
           $R$1$i9 = $377;$RP$1$i8 = $376;
          }
         }
         $379 = ($RP$1$i8$lcssa>>>0)<($349>>>0);
         if ($379) {
          _abort();
          // unreachable;
         } else {
          HEAP32[$RP$1$i8$lcssa>>2] = 0;
          $R$3$i11 = $R$1$i9$lcssa;
          break;
         }
        } else {
         $358 = ((($v$4$lcssa$i)) + 8|0);
         $359 = HEAP32[$358>>2]|0;
         $360 = ($359>>>0)<($349>>>0);
         if ($360) {
          _abort();
          // unreachable;
         }
         $361 = ((($359)) + 12|0);
         $362 = HEAP32[$361>>2]|0;
         $363 = ($362|0)==($v$4$lcssa$i|0);
         if (!($363)) {
          _abort();
          // unreachable;
         }
         $364 = ((($356)) + 8|0);
         $365 = HEAP32[$364>>2]|0;
         $366 = ($365|0)==($v$4$lcssa$i|0);
         if ($366) {
          HEAP32[$361>>2] = $356;
          HEAP32[$364>>2] = $359;
          $R$3$i11 = $356;
          break;
         } else {
          _abort();
          // unreachable;
         }
        }
       } while(0);
       $380 = ($354|0)==(0|0);
       do {
        if (!($380)) {
         $381 = ((($v$4$lcssa$i)) + 28|0);
         $382 = HEAP32[$381>>2]|0;
         $383 = (1206332 + ($382<<2)|0);
         $384 = HEAP32[$383>>2]|0;
         $385 = ($v$4$lcssa$i|0)==($384|0);
         if ($385) {
          HEAP32[$383>>2] = $R$3$i11;
          $cond$i12 = ($R$3$i11|0)==(0|0);
          if ($cond$i12) {
           $386 = 1 << $382;
           $387 = $386 ^ -1;
           $388 = HEAP32[(1206032)>>2]|0;
           $389 = $388 & $387;
           HEAP32[(1206032)>>2] = $389;
           break;
          }
         } else {
          $390 = HEAP32[(1206044)>>2]|0;
          $391 = ($354>>>0)<($390>>>0);
          if ($391) {
           _abort();
           // unreachable;
          }
          $392 = ((($354)) + 16|0);
          $393 = HEAP32[$392>>2]|0;
          $394 = ($393|0)==($v$4$lcssa$i|0);
          if ($394) {
           HEAP32[$392>>2] = $R$3$i11;
          } else {
           $395 = ((($354)) + 20|0);
           HEAP32[$395>>2] = $R$3$i11;
          }
          $396 = ($R$3$i11|0)==(0|0);
          if ($396) {
           break;
          }
         }
         $397 = HEAP32[(1206044)>>2]|0;
         $398 = ($R$3$i11>>>0)<($397>>>0);
         if ($398) {
          _abort();
          // unreachable;
         }
         $399 = ((($R$3$i11)) + 24|0);
         HEAP32[$399>>2] = $354;
         $400 = ((($v$4$lcssa$i)) + 16|0);
         $401 = HEAP32[$400>>2]|0;
         $402 = ($401|0)==(0|0);
         do {
          if (!($402)) {
           $403 = ($401>>>0)<($397>>>0);
           if ($403) {
            _abort();
            // unreachable;
           } else {
            $404 = ((($R$3$i11)) + 16|0);
            HEAP32[$404>>2] = $401;
            $405 = ((($401)) + 24|0);
            HEAP32[$405>>2] = $R$3$i11;
            break;
           }
          }
         } while(0);
         $406 = ((($v$4$lcssa$i)) + 20|0);
         $407 = HEAP32[$406>>2]|0;
         $408 = ($407|0)==(0|0);
         if (!($408)) {
          $409 = HEAP32[(1206044)>>2]|0;
          $410 = ($407>>>0)<($409>>>0);
          if ($410) {
           _abort();
           // unreachable;
          } else {
           $411 = ((($R$3$i11)) + 20|0);
           HEAP32[$411>>2] = $407;
           $412 = ((($407)) + 24|0);
           HEAP32[$412>>2] = $R$3$i11;
           break;
          }
         }
        }
       } while(0);
       $413 = ($rsize$4$lcssa$i>>>0)<(16);
       do {
        if ($413) {
         $414 = (($rsize$4$lcssa$i) + ($248))|0;
         $415 = $414 | 3;
         $416 = ((($v$4$lcssa$i)) + 4|0);
         HEAP32[$416>>2] = $415;
         $417 = (($v$4$lcssa$i) + ($414)|0);
         $418 = ((($417)) + 4|0);
         $419 = HEAP32[$418>>2]|0;
         $420 = $419 | 1;
         HEAP32[$418>>2] = $420;
        } else {
         $421 = $248 | 3;
         $422 = ((($v$4$lcssa$i)) + 4|0);
         HEAP32[$422>>2] = $421;
         $423 = $rsize$4$lcssa$i | 1;
         $424 = ((($351)) + 4|0);
         HEAP32[$424>>2] = $423;
         $425 = (($351) + ($rsize$4$lcssa$i)|0);
         HEAP32[$425>>2] = $rsize$4$lcssa$i;
         $426 = $rsize$4$lcssa$i >>> 3;
         $427 = ($rsize$4$lcssa$i>>>0)<(256);
         if ($427) {
          $428 = $426 << 1;
          $429 = (1206068 + ($428<<2)|0);
          $430 = HEAP32[301507]|0;
          $431 = 1 << $426;
          $432 = $430 & $431;
          $433 = ($432|0)==(0);
          if ($433) {
           $434 = $430 | $431;
           HEAP32[301507] = $434;
           $$pre$i13 = ((($429)) + 8|0);
           $$pre$phi$i14Z2D = $$pre$i13;$F5$0$i = $429;
          } else {
           $435 = ((($429)) + 8|0);
           $436 = HEAP32[$435>>2]|0;
           $437 = HEAP32[(1206044)>>2]|0;
           $438 = ($436>>>0)<($437>>>0);
           if ($438) {
            _abort();
            // unreachable;
           } else {
            $$pre$phi$i14Z2D = $435;$F5$0$i = $436;
           }
          }
          HEAP32[$$pre$phi$i14Z2D>>2] = $351;
          $439 = ((($F5$0$i)) + 12|0);
          HEAP32[$439>>2] = $351;
          $440 = ((($351)) + 8|0);
          HEAP32[$440>>2] = $F5$0$i;
          $441 = ((($351)) + 12|0);
          HEAP32[$441>>2] = $429;
          break;
         }
         $442 = $rsize$4$lcssa$i >>> 8;
         $443 = ($442|0)==(0);
         if ($443) {
          $I7$0$i = 0;
         } else {
          $444 = ($rsize$4$lcssa$i>>>0)>(16777215);
          if ($444) {
           $I7$0$i = 31;
          } else {
           $445 = (($442) + 1048320)|0;
           $446 = $445 >>> 16;
           $447 = $446 & 8;
           $448 = $442 << $447;
           $449 = (($448) + 520192)|0;
           $450 = $449 >>> 16;
           $451 = $450 & 4;
           $452 = $451 | $447;
           $453 = $448 << $451;
           $454 = (($453) + 245760)|0;
           $455 = $454 >>> 16;
           $456 = $455 & 2;
           $457 = $452 | $456;
           $458 = (14 - ($457))|0;
           $459 = $453 << $456;
           $460 = $459 >>> 15;
           $461 = (($458) + ($460))|0;
           $462 = $461 << 1;
           $463 = (($461) + 7)|0;
           $464 = $rsize$4$lcssa$i >>> $463;
           $465 = $464 & 1;
           $466 = $465 | $462;
           $I7$0$i = $466;
          }
         }
         $467 = (1206332 + ($I7$0$i<<2)|0);
         $468 = ((($351)) + 28|0);
         HEAP32[$468>>2] = $I7$0$i;
         $469 = ((($351)) + 16|0);
         $470 = ((($469)) + 4|0);
         HEAP32[$470>>2] = 0;
         HEAP32[$469>>2] = 0;
         $471 = HEAP32[(1206032)>>2]|0;
         $472 = 1 << $I7$0$i;
         $473 = $471 & $472;
         $474 = ($473|0)==(0);
         if ($474) {
          $475 = $471 | $472;
          HEAP32[(1206032)>>2] = $475;
          HEAP32[$467>>2] = $351;
          $476 = ((($351)) + 24|0);
          HEAP32[$476>>2] = $467;
          $477 = ((($351)) + 12|0);
          HEAP32[$477>>2] = $351;
          $478 = ((($351)) + 8|0);
          HEAP32[$478>>2] = $351;
          break;
         }
         $479 = HEAP32[$467>>2]|0;
         $480 = ($I7$0$i|0)==(31);
         $481 = $I7$0$i >>> 1;
         $482 = (25 - ($481))|0;
         $483 = $480 ? 0 : $482;
         $484 = $rsize$4$lcssa$i << $483;
         $K12$0$i = $484;$T$0$i = $479;
         while(1) {
          $485 = ((($T$0$i)) + 4|0);
          $486 = HEAP32[$485>>2]|0;
          $487 = $486 & -8;
          $488 = ($487|0)==($rsize$4$lcssa$i|0);
          if ($488) {
           $T$0$i$lcssa = $T$0$i;
           label = 148;
           break;
          }
          $489 = $K12$0$i >>> 31;
          $490 = (((($T$0$i)) + 16|0) + ($489<<2)|0);
          $491 = $K12$0$i << 1;
          $492 = HEAP32[$490>>2]|0;
          $493 = ($492|0)==(0|0);
          if ($493) {
           $$lcssa157 = $490;$T$0$i$lcssa156 = $T$0$i;
           label = 145;
           break;
          } else {
           $K12$0$i = $491;$T$0$i = $492;
          }
         }
         if ((label|0) == 145) {
          $494 = HEAP32[(1206044)>>2]|0;
          $495 = ($$lcssa157>>>0)<($494>>>0);
          if ($495) {
           _abort();
           // unreachable;
          } else {
           HEAP32[$$lcssa157>>2] = $351;
           $496 = ((($351)) + 24|0);
           HEAP32[$496>>2] = $T$0$i$lcssa156;
           $497 = ((($351)) + 12|0);
           HEAP32[$497>>2] = $351;
           $498 = ((($351)) + 8|0);
           HEAP32[$498>>2] = $351;
           break;
          }
         }
         else if ((label|0) == 148) {
          $499 = ((($T$0$i$lcssa)) + 8|0);
          $500 = HEAP32[$499>>2]|0;
          $501 = HEAP32[(1206044)>>2]|0;
          $502 = ($500>>>0)>=($501>>>0);
          $not$7$i = ($T$0$i$lcssa>>>0)>=($501>>>0);
          $503 = $502 & $not$7$i;
          if ($503) {
           $504 = ((($500)) + 12|0);
           HEAP32[$504>>2] = $351;
           HEAP32[$499>>2] = $351;
           $505 = ((($351)) + 8|0);
           HEAP32[$505>>2] = $500;
           $506 = ((($351)) + 12|0);
           HEAP32[$506>>2] = $T$0$i$lcssa;
           $507 = ((($351)) + 24|0);
           HEAP32[$507>>2] = 0;
           break;
          } else {
           _abort();
           // unreachable;
          }
         }
        }
       } while(0);
       $508 = ((($v$4$lcssa$i)) + 8|0);
       $$0 = $508;
       return ($$0|0);
      } else {
       $nb$0 = $248;
      }
     }
    }
   }
  }
 } while(0);
 $509 = HEAP32[(1206036)>>2]|0;
 $510 = ($509>>>0)<($nb$0>>>0);
 if (!($510)) {
  $511 = (($509) - ($nb$0))|0;
  $512 = HEAP32[(1206048)>>2]|0;
  $513 = ($511>>>0)>(15);
  if ($513) {
   $514 = (($512) + ($nb$0)|0);
   HEAP32[(1206048)>>2] = $514;
   HEAP32[(1206036)>>2] = $511;
   $515 = $511 | 1;
   $516 = ((($514)) + 4|0);
   HEAP32[$516>>2] = $515;
   $517 = (($514) + ($511)|0);
   HEAP32[$517>>2] = $511;
   $518 = $nb$0 | 3;
   $519 = ((($512)) + 4|0);
   HEAP32[$519>>2] = $518;
  } else {
   HEAP32[(1206036)>>2] = 0;
   HEAP32[(1206048)>>2] = 0;
   $520 = $509 | 3;
   $521 = ((($512)) + 4|0);
   HEAP32[$521>>2] = $520;
   $522 = (($512) + ($509)|0);
   $523 = ((($522)) + 4|0);
   $524 = HEAP32[$523>>2]|0;
   $525 = $524 | 1;
   HEAP32[$523>>2] = $525;
  }
  $526 = ((($512)) + 8|0);
  $$0 = $526;
  return ($$0|0);
 }
 $527 = HEAP32[(1206040)>>2]|0;
 $528 = ($527>>>0)>($nb$0>>>0);
 if ($528) {
  $529 = (($527) - ($nb$0))|0;
  HEAP32[(1206040)>>2] = $529;
  $530 = HEAP32[(1206052)>>2]|0;
  $531 = (($530) + ($nb$0)|0);
  HEAP32[(1206052)>>2] = $531;
  $532 = $529 | 1;
  $533 = ((($531)) + 4|0);
  HEAP32[$533>>2] = $532;
  $534 = $nb$0 | 3;
  $535 = ((($530)) + 4|0);
  HEAP32[$535>>2] = $534;
  $536 = ((($530)) + 8|0);
  $$0 = $536;
  return ($$0|0);
 }
 $537 = HEAP32[301625]|0;
 $538 = ($537|0)==(0);
 do {
  if ($538) {
   $539 = (_sysconf(30)|0);
   $540 = (($539) + -1)|0;
   $541 = $540 & $539;
   $542 = ($541|0)==(0);
   if ($542) {
    HEAP32[(1206508)>>2] = $539;
    HEAP32[(1206504)>>2] = $539;
    HEAP32[(1206512)>>2] = -1;
    HEAP32[(1206516)>>2] = -1;
    HEAP32[(1206520)>>2] = 0;
    HEAP32[(1206472)>>2] = 0;
    $543 = (_time((0|0))|0);
    $544 = $543 & -16;
    $545 = $544 ^ 1431655768;
    HEAP32[301625] = $545;
    break;
   } else {
    _abort();
    // unreachable;
   }
  }
 } while(0);
 $546 = (($nb$0) + 48)|0;
 $547 = HEAP32[(1206508)>>2]|0;
 $548 = (($nb$0) + 47)|0;
 $549 = (($547) + ($548))|0;
 $550 = (0 - ($547))|0;
 $551 = $549 & $550;
 $552 = ($551>>>0)>($nb$0>>>0);
 if (!($552)) {
  $$0 = 0;
  return ($$0|0);
 }
 $553 = HEAP32[(1206468)>>2]|0;
 $554 = ($553|0)==(0);
 if (!($554)) {
  $555 = HEAP32[(1206460)>>2]|0;
  $556 = (($555) + ($551))|0;
  $557 = ($556>>>0)<=($555>>>0);
  $558 = ($556>>>0)>($553>>>0);
  $or$cond1$i16 = $557 | $558;
  if ($or$cond1$i16) {
   $$0 = 0;
   return ($$0|0);
  }
 }
 $559 = HEAP32[(1206472)>>2]|0;
 $560 = $559 & 4;
 $561 = ($560|0)==(0);
 L257: do {
  if ($561) {
   $562 = HEAP32[(1206052)>>2]|0;
   $563 = ($562|0)==(0|0);
   L259: do {
    if ($563) {
     label = 173;
    } else {
     $sp$0$i$i = (1206476);
     while(1) {
      $564 = HEAP32[$sp$0$i$i>>2]|0;
      $565 = ($564>>>0)>($562>>>0);
      if (!($565)) {
       $566 = ((($sp$0$i$i)) + 4|0);
       $567 = HEAP32[$566>>2]|0;
       $568 = (($564) + ($567)|0);
       $569 = ($568>>>0)>($562>>>0);
       if ($569) {
        $$lcssa153 = $sp$0$i$i;$$lcssa155 = $566;
        break;
       }
      }
      $570 = ((($sp$0$i$i)) + 8|0);
      $571 = HEAP32[$570>>2]|0;
      $572 = ($571|0)==(0|0);
      if ($572) {
       label = 173;
       break L259;
      } else {
       $sp$0$i$i = $571;
      }
     }
     $595 = HEAP32[(1206040)>>2]|0;
     $596 = (($549) - ($595))|0;
     $597 = $596 & $550;
     $598 = ($597>>>0)<(2147483647);
     if ($598) {
      $599 = (_sbrk(($597|0))|0);
      $600 = HEAP32[$$lcssa153>>2]|0;
      $601 = HEAP32[$$lcssa155>>2]|0;
      $602 = (($600) + ($601)|0);
      $603 = ($599|0)==($602|0);
      if ($603) {
       $604 = ($599|0)==((-1)|0);
       if (!($604)) {
        $tbase$746$i = $599;$tsize$745$i = $597;
        label = 193;
        break L257;
       }
      } else {
       $br$2$ph$i = $599;$ssize$2$ph$i = $597;
       label = 183;
      }
     }
    }
   } while(0);
   do {
    if ((label|0) == 173) {
     $573 = (_sbrk(0)|0);
     $574 = ($573|0)==((-1)|0);
     if (!($574)) {
      $575 = $573;
      $576 = HEAP32[(1206504)>>2]|0;
      $577 = (($576) + -1)|0;
      $578 = $577 & $575;
      $579 = ($578|0)==(0);
      if ($579) {
       $ssize$0$i = $551;
      } else {
       $580 = (($577) + ($575))|0;
       $581 = (0 - ($576))|0;
       $582 = $580 & $581;
       $583 = (($551) - ($575))|0;
       $584 = (($583) + ($582))|0;
       $ssize$0$i = $584;
      }
      $585 = HEAP32[(1206460)>>2]|0;
      $586 = (($585) + ($ssize$0$i))|0;
      $587 = ($ssize$0$i>>>0)>($nb$0>>>0);
      $588 = ($ssize$0$i>>>0)<(2147483647);
      $or$cond$i17 = $587 & $588;
      if ($or$cond$i17) {
       $589 = HEAP32[(1206468)>>2]|0;
       $590 = ($589|0)==(0);
       if (!($590)) {
        $591 = ($586>>>0)<=($585>>>0);
        $592 = ($586>>>0)>($589>>>0);
        $or$cond2$i = $591 | $592;
        if ($or$cond2$i) {
         break;
        }
       }
       $593 = (_sbrk(($ssize$0$i|0))|0);
       $594 = ($593|0)==($573|0);
       if ($594) {
        $tbase$746$i = $573;$tsize$745$i = $ssize$0$i;
        label = 193;
        break L257;
       } else {
        $br$2$ph$i = $593;$ssize$2$ph$i = $ssize$0$i;
        label = 183;
       }
      }
     }
    }
   } while(0);
   L279: do {
    if ((label|0) == 183) {
     $605 = (0 - ($ssize$2$ph$i))|0;
     $606 = ($br$2$ph$i|0)!=((-1)|0);
     $607 = ($ssize$2$ph$i>>>0)<(2147483647);
     $or$cond7$i = $607 & $606;
     $608 = ($546>>>0)>($ssize$2$ph$i>>>0);
     $or$cond8$i = $608 & $or$cond7$i;
     do {
      if ($or$cond8$i) {
       $609 = HEAP32[(1206508)>>2]|0;
       $610 = (($548) - ($ssize$2$ph$i))|0;
       $611 = (($610) + ($609))|0;
       $612 = (0 - ($609))|0;
       $613 = $611 & $612;
       $614 = ($613>>>0)<(2147483647);
       if ($614) {
        $615 = (_sbrk(($613|0))|0);
        $616 = ($615|0)==((-1)|0);
        if ($616) {
         (_sbrk(($605|0))|0);
         break L279;
        } else {
         $617 = (($613) + ($ssize$2$ph$i))|0;
         $ssize$5$i = $617;
         break;
        }
       } else {
        $ssize$5$i = $ssize$2$ph$i;
       }
      } else {
       $ssize$5$i = $ssize$2$ph$i;
      }
     } while(0);
     $618 = ($br$2$ph$i|0)==((-1)|0);
     if (!($618)) {
      $tbase$746$i = $br$2$ph$i;$tsize$745$i = $ssize$5$i;
      label = 193;
      break L257;
     }
    }
   } while(0);
   $619 = HEAP32[(1206472)>>2]|0;
   $620 = $619 | 4;
   HEAP32[(1206472)>>2] = $620;
   label = 190;
  } else {
   label = 190;
  }
 } while(0);
 if ((label|0) == 190) {
  $621 = ($551>>>0)<(2147483647);
  if ($621) {
   $622 = (_sbrk(($551|0))|0);
   $623 = (_sbrk(0)|0);
   $624 = ($622|0)!=((-1)|0);
   $625 = ($623|0)!=((-1)|0);
   $or$cond5$i = $624 & $625;
   $626 = ($622>>>0)<($623>>>0);
   $or$cond10$i = $626 & $or$cond5$i;
   if ($or$cond10$i) {
    $627 = $623;
    $628 = $622;
    $629 = (($627) - ($628))|0;
    $630 = (($nb$0) + 40)|0;
    $$not$i = ($629>>>0)>($630>>>0);
    if ($$not$i) {
     $tbase$746$i = $622;$tsize$745$i = $629;
     label = 193;
    }
   }
  }
 }
 if ((label|0) == 193) {
  $631 = HEAP32[(1206460)>>2]|0;
  $632 = (($631) + ($tsize$745$i))|0;
  HEAP32[(1206460)>>2] = $632;
  $633 = HEAP32[(1206464)>>2]|0;
  $634 = ($632>>>0)>($633>>>0);
  if ($634) {
   HEAP32[(1206464)>>2] = $632;
  }
  $635 = HEAP32[(1206052)>>2]|0;
  $636 = ($635|0)==(0|0);
  do {
   if ($636) {
    $637 = HEAP32[(1206044)>>2]|0;
    $638 = ($637|0)==(0|0);
    $639 = ($tbase$746$i>>>0)<($637>>>0);
    $or$cond11$i = $638 | $639;
    if ($or$cond11$i) {
     HEAP32[(1206044)>>2] = $tbase$746$i;
    }
    HEAP32[(1206476)>>2] = $tbase$746$i;
    HEAP32[(1206480)>>2] = $tsize$745$i;
    HEAP32[(1206488)>>2] = 0;
    $640 = HEAP32[301625]|0;
    HEAP32[(1206064)>>2] = $640;
    HEAP32[(1206060)>>2] = -1;
    $i$01$i$i = 0;
    while(1) {
     $641 = $i$01$i$i << 1;
     $642 = (1206068 + ($641<<2)|0);
     $643 = ((($642)) + 12|0);
     HEAP32[$643>>2] = $642;
     $644 = ((($642)) + 8|0);
     HEAP32[$644>>2] = $642;
     $645 = (($i$01$i$i) + 1)|0;
     $exitcond$i$i = ($645|0)==(32);
     if ($exitcond$i$i) {
      break;
     } else {
      $i$01$i$i = $645;
     }
    }
    $646 = (($tsize$745$i) + -40)|0;
    $647 = ((($tbase$746$i)) + 8|0);
    $648 = $647;
    $649 = $648 & 7;
    $650 = ($649|0)==(0);
    $651 = (0 - ($648))|0;
    $652 = $651 & 7;
    $653 = $650 ? 0 : $652;
    $654 = (($tbase$746$i) + ($653)|0);
    $655 = (($646) - ($653))|0;
    HEAP32[(1206052)>>2] = $654;
    HEAP32[(1206040)>>2] = $655;
    $656 = $655 | 1;
    $657 = ((($654)) + 4|0);
    HEAP32[$657>>2] = $656;
    $658 = (($654) + ($655)|0);
    $659 = ((($658)) + 4|0);
    HEAP32[$659>>2] = 40;
    $660 = HEAP32[(1206516)>>2]|0;
    HEAP32[(1206056)>>2] = $660;
   } else {
    $sp$068$i = (1206476);
    while(1) {
     $661 = HEAP32[$sp$068$i>>2]|0;
     $662 = ((($sp$068$i)) + 4|0);
     $663 = HEAP32[$662>>2]|0;
     $664 = (($661) + ($663)|0);
     $665 = ($tbase$746$i|0)==($664|0);
     if ($665) {
      $$lcssa147 = $661;$$lcssa149 = $662;$$lcssa151 = $663;$sp$068$i$lcssa = $sp$068$i;
      label = 203;
      break;
     }
     $666 = ((($sp$068$i)) + 8|0);
     $667 = HEAP32[$666>>2]|0;
     $668 = ($667|0)==(0|0);
     if ($668) {
      break;
     } else {
      $sp$068$i = $667;
     }
    }
    if ((label|0) == 203) {
     $669 = ((($sp$068$i$lcssa)) + 12|0);
     $670 = HEAP32[$669>>2]|0;
     $671 = $670 & 8;
     $672 = ($671|0)==(0);
     if ($672) {
      $673 = ($635>>>0)>=($$lcssa147>>>0);
      $674 = ($635>>>0)<($tbase$746$i>>>0);
      $or$cond48$i = $674 & $673;
      if ($or$cond48$i) {
       $675 = (($$lcssa151) + ($tsize$745$i))|0;
       HEAP32[$$lcssa149>>2] = $675;
       $676 = HEAP32[(1206040)>>2]|0;
       $677 = ((($635)) + 8|0);
       $678 = $677;
       $679 = $678 & 7;
       $680 = ($679|0)==(0);
       $681 = (0 - ($678))|0;
       $682 = $681 & 7;
       $683 = $680 ? 0 : $682;
       $684 = (($635) + ($683)|0);
       $685 = (($tsize$745$i) - ($683))|0;
       $686 = (($685) + ($676))|0;
       HEAP32[(1206052)>>2] = $684;
       HEAP32[(1206040)>>2] = $686;
       $687 = $686 | 1;
       $688 = ((($684)) + 4|0);
       HEAP32[$688>>2] = $687;
       $689 = (($684) + ($686)|0);
       $690 = ((($689)) + 4|0);
       HEAP32[$690>>2] = 40;
       $691 = HEAP32[(1206516)>>2]|0;
       HEAP32[(1206056)>>2] = $691;
       break;
      }
     }
    }
    $692 = HEAP32[(1206044)>>2]|0;
    $693 = ($tbase$746$i>>>0)<($692>>>0);
    if ($693) {
     HEAP32[(1206044)>>2] = $tbase$746$i;
     $757 = $tbase$746$i;
    } else {
     $757 = $692;
    }
    $694 = (($tbase$746$i) + ($tsize$745$i)|0);
    $sp$167$i = (1206476);
    while(1) {
     $695 = HEAP32[$sp$167$i>>2]|0;
     $696 = ($695|0)==($694|0);
     if ($696) {
      $$lcssa144 = $sp$167$i;$sp$167$i$lcssa = $sp$167$i;
      label = 211;
      break;
     }
     $697 = ((($sp$167$i)) + 8|0);
     $698 = HEAP32[$697>>2]|0;
     $699 = ($698|0)==(0|0);
     if ($699) {
      $sp$0$i$i$i = (1206476);
      break;
     } else {
      $sp$167$i = $698;
     }
    }
    if ((label|0) == 211) {
     $700 = ((($sp$167$i$lcssa)) + 12|0);
     $701 = HEAP32[$700>>2]|0;
     $702 = $701 & 8;
     $703 = ($702|0)==(0);
     if ($703) {
      HEAP32[$$lcssa144>>2] = $tbase$746$i;
      $704 = ((($sp$167$i$lcssa)) + 4|0);
      $705 = HEAP32[$704>>2]|0;
      $706 = (($705) + ($tsize$745$i))|0;
      HEAP32[$704>>2] = $706;
      $707 = ((($tbase$746$i)) + 8|0);
      $708 = $707;
      $709 = $708 & 7;
      $710 = ($709|0)==(0);
      $711 = (0 - ($708))|0;
      $712 = $711 & 7;
      $713 = $710 ? 0 : $712;
      $714 = (($tbase$746$i) + ($713)|0);
      $715 = ((($694)) + 8|0);
      $716 = $715;
      $717 = $716 & 7;
      $718 = ($717|0)==(0);
      $719 = (0 - ($716))|0;
      $720 = $719 & 7;
      $721 = $718 ? 0 : $720;
      $722 = (($694) + ($721)|0);
      $723 = $722;
      $724 = $714;
      $725 = (($723) - ($724))|0;
      $726 = (($714) + ($nb$0)|0);
      $727 = (($725) - ($nb$0))|0;
      $728 = $nb$0 | 3;
      $729 = ((($714)) + 4|0);
      HEAP32[$729>>2] = $728;
      $730 = ($722|0)==($635|0);
      do {
       if ($730) {
        $731 = HEAP32[(1206040)>>2]|0;
        $732 = (($731) + ($727))|0;
        HEAP32[(1206040)>>2] = $732;
        HEAP32[(1206052)>>2] = $726;
        $733 = $732 | 1;
        $734 = ((($726)) + 4|0);
        HEAP32[$734>>2] = $733;
       } else {
        $735 = HEAP32[(1206048)>>2]|0;
        $736 = ($722|0)==($735|0);
        if ($736) {
         $737 = HEAP32[(1206036)>>2]|0;
         $738 = (($737) + ($727))|0;
         HEAP32[(1206036)>>2] = $738;
         HEAP32[(1206048)>>2] = $726;
         $739 = $738 | 1;
         $740 = ((($726)) + 4|0);
         HEAP32[$740>>2] = $739;
         $741 = (($726) + ($738)|0);
         HEAP32[$741>>2] = $738;
         break;
        }
        $742 = ((($722)) + 4|0);
        $743 = HEAP32[$742>>2]|0;
        $744 = $743 & 3;
        $745 = ($744|0)==(1);
        if ($745) {
         $746 = $743 & -8;
         $747 = $743 >>> 3;
         $748 = ($743>>>0)<(256);
         L331: do {
          if ($748) {
           $749 = ((($722)) + 8|0);
           $750 = HEAP32[$749>>2]|0;
           $751 = ((($722)) + 12|0);
           $752 = HEAP32[$751>>2]|0;
           $753 = $747 << 1;
           $754 = (1206068 + ($753<<2)|0);
           $755 = ($750|0)==($754|0);
           do {
            if (!($755)) {
             $756 = ($750>>>0)<($757>>>0);
             if ($756) {
              _abort();
              // unreachable;
             }
             $758 = ((($750)) + 12|0);
             $759 = HEAP32[$758>>2]|0;
             $760 = ($759|0)==($722|0);
             if ($760) {
              break;
             }
             _abort();
             // unreachable;
            }
           } while(0);
           $761 = ($752|0)==($750|0);
           if ($761) {
            $762 = 1 << $747;
            $763 = $762 ^ -1;
            $764 = HEAP32[301507]|0;
            $765 = $764 & $763;
            HEAP32[301507] = $765;
            break;
           }
           $766 = ($752|0)==($754|0);
           do {
            if ($766) {
             $$pre9$i$i = ((($752)) + 8|0);
             $$pre$phi10$i$iZ2D = $$pre9$i$i;
            } else {
             $767 = ($752>>>0)<($757>>>0);
             if ($767) {
              _abort();
              // unreachable;
             }
             $768 = ((($752)) + 8|0);
             $769 = HEAP32[$768>>2]|0;
             $770 = ($769|0)==($722|0);
             if ($770) {
              $$pre$phi10$i$iZ2D = $768;
              break;
             }
             _abort();
             // unreachable;
            }
           } while(0);
           $771 = ((($750)) + 12|0);
           HEAP32[$771>>2] = $752;
           HEAP32[$$pre$phi10$i$iZ2D>>2] = $750;
          } else {
           $772 = ((($722)) + 24|0);
           $773 = HEAP32[$772>>2]|0;
           $774 = ((($722)) + 12|0);
           $775 = HEAP32[$774>>2]|0;
           $776 = ($775|0)==($722|0);
           do {
            if ($776) {
             $786 = ((($722)) + 16|0);
             $787 = ((($786)) + 4|0);
             $788 = HEAP32[$787>>2]|0;
             $789 = ($788|0)==(0|0);
             if ($789) {
              $790 = HEAP32[$786>>2]|0;
              $791 = ($790|0)==(0|0);
              if ($791) {
               $R$3$i$i = 0;
               break;
              } else {
               $R$1$i$i = $790;$RP$1$i$i = $786;
              }
             } else {
              $R$1$i$i = $788;$RP$1$i$i = $787;
             }
             while(1) {
              $792 = ((($R$1$i$i)) + 20|0);
              $793 = HEAP32[$792>>2]|0;
              $794 = ($793|0)==(0|0);
              if (!($794)) {
               $R$1$i$i = $793;$RP$1$i$i = $792;
               continue;
              }
              $795 = ((($R$1$i$i)) + 16|0);
              $796 = HEAP32[$795>>2]|0;
              $797 = ($796|0)==(0|0);
              if ($797) {
               $R$1$i$i$lcssa = $R$1$i$i;$RP$1$i$i$lcssa = $RP$1$i$i;
               break;
              } else {
               $R$1$i$i = $796;$RP$1$i$i = $795;
              }
             }
             $798 = ($RP$1$i$i$lcssa>>>0)<($757>>>0);
             if ($798) {
              _abort();
              // unreachable;
             } else {
              HEAP32[$RP$1$i$i$lcssa>>2] = 0;
              $R$3$i$i = $R$1$i$i$lcssa;
              break;
             }
            } else {
             $777 = ((($722)) + 8|0);
             $778 = HEAP32[$777>>2]|0;
             $779 = ($778>>>0)<($757>>>0);
             if ($779) {
              _abort();
              // unreachable;
             }
             $780 = ((($778)) + 12|0);
             $781 = HEAP32[$780>>2]|0;
             $782 = ($781|0)==($722|0);
             if (!($782)) {
              _abort();
              // unreachable;
             }
             $783 = ((($775)) + 8|0);
             $784 = HEAP32[$783>>2]|0;
             $785 = ($784|0)==($722|0);
             if ($785) {
              HEAP32[$780>>2] = $775;
              HEAP32[$783>>2] = $778;
              $R$3$i$i = $775;
              break;
             } else {
              _abort();
              // unreachable;
             }
            }
           } while(0);
           $799 = ($773|0)==(0|0);
           if ($799) {
            break;
           }
           $800 = ((($722)) + 28|0);
           $801 = HEAP32[$800>>2]|0;
           $802 = (1206332 + ($801<<2)|0);
           $803 = HEAP32[$802>>2]|0;
           $804 = ($722|0)==($803|0);
           do {
            if ($804) {
             HEAP32[$802>>2] = $R$3$i$i;
             $cond$i$i = ($R$3$i$i|0)==(0|0);
             if (!($cond$i$i)) {
              break;
             }
             $805 = 1 << $801;
             $806 = $805 ^ -1;
             $807 = HEAP32[(1206032)>>2]|0;
             $808 = $807 & $806;
             HEAP32[(1206032)>>2] = $808;
             break L331;
            } else {
             $809 = HEAP32[(1206044)>>2]|0;
             $810 = ($773>>>0)<($809>>>0);
             if ($810) {
              _abort();
              // unreachable;
             }
             $811 = ((($773)) + 16|0);
             $812 = HEAP32[$811>>2]|0;
             $813 = ($812|0)==($722|0);
             if ($813) {
              HEAP32[$811>>2] = $R$3$i$i;
             } else {
              $814 = ((($773)) + 20|0);
              HEAP32[$814>>2] = $R$3$i$i;
             }
             $815 = ($R$3$i$i|0)==(0|0);
             if ($815) {
              break L331;
             }
            }
           } while(0);
           $816 = HEAP32[(1206044)>>2]|0;
           $817 = ($R$3$i$i>>>0)<($816>>>0);
           if ($817) {
            _abort();
            // unreachable;
           }
           $818 = ((($R$3$i$i)) + 24|0);
           HEAP32[$818>>2] = $773;
           $819 = ((($722)) + 16|0);
           $820 = HEAP32[$819>>2]|0;
           $821 = ($820|0)==(0|0);
           do {
            if (!($821)) {
             $822 = ($820>>>0)<($816>>>0);
             if ($822) {
              _abort();
              // unreachable;
             } else {
              $823 = ((($R$3$i$i)) + 16|0);
              HEAP32[$823>>2] = $820;
              $824 = ((($820)) + 24|0);
              HEAP32[$824>>2] = $R$3$i$i;
              break;
             }
            }
           } while(0);
           $825 = ((($819)) + 4|0);
           $826 = HEAP32[$825>>2]|0;
           $827 = ($826|0)==(0|0);
           if ($827) {
            break;
           }
           $828 = HEAP32[(1206044)>>2]|0;
           $829 = ($826>>>0)<($828>>>0);
           if ($829) {
            _abort();
            // unreachable;
           } else {
            $830 = ((($R$3$i$i)) + 20|0);
            HEAP32[$830>>2] = $826;
            $831 = ((($826)) + 24|0);
            HEAP32[$831>>2] = $R$3$i$i;
            break;
           }
          }
         } while(0);
         $832 = (($722) + ($746)|0);
         $833 = (($746) + ($727))|0;
         $oldfirst$0$i$i = $832;$qsize$0$i$i = $833;
        } else {
         $oldfirst$0$i$i = $722;$qsize$0$i$i = $727;
        }
        $834 = ((($oldfirst$0$i$i)) + 4|0);
        $835 = HEAP32[$834>>2]|0;
        $836 = $835 & -2;
        HEAP32[$834>>2] = $836;
        $837 = $qsize$0$i$i | 1;
        $838 = ((($726)) + 4|0);
        HEAP32[$838>>2] = $837;
        $839 = (($726) + ($qsize$0$i$i)|0);
        HEAP32[$839>>2] = $qsize$0$i$i;
        $840 = $qsize$0$i$i >>> 3;
        $841 = ($qsize$0$i$i>>>0)<(256);
        if ($841) {
         $842 = $840 << 1;
         $843 = (1206068 + ($842<<2)|0);
         $844 = HEAP32[301507]|0;
         $845 = 1 << $840;
         $846 = $844 & $845;
         $847 = ($846|0)==(0);
         do {
          if ($847) {
           $848 = $844 | $845;
           HEAP32[301507] = $848;
           $$pre$i16$i = ((($843)) + 8|0);
           $$pre$phi$i17$iZ2D = $$pre$i16$i;$F4$0$i$i = $843;
          } else {
           $849 = ((($843)) + 8|0);
           $850 = HEAP32[$849>>2]|0;
           $851 = HEAP32[(1206044)>>2]|0;
           $852 = ($850>>>0)<($851>>>0);
           if (!($852)) {
            $$pre$phi$i17$iZ2D = $849;$F4$0$i$i = $850;
            break;
           }
           _abort();
           // unreachable;
          }
         } while(0);
         HEAP32[$$pre$phi$i17$iZ2D>>2] = $726;
         $853 = ((($F4$0$i$i)) + 12|0);
         HEAP32[$853>>2] = $726;
         $854 = ((($726)) + 8|0);
         HEAP32[$854>>2] = $F4$0$i$i;
         $855 = ((($726)) + 12|0);
         HEAP32[$855>>2] = $843;
         break;
        }
        $856 = $qsize$0$i$i >>> 8;
        $857 = ($856|0)==(0);
        do {
         if ($857) {
          $I7$0$i$i = 0;
         } else {
          $858 = ($qsize$0$i$i>>>0)>(16777215);
          if ($858) {
           $I7$0$i$i = 31;
           break;
          }
          $859 = (($856) + 1048320)|0;
          $860 = $859 >>> 16;
          $861 = $860 & 8;
          $862 = $856 << $861;
          $863 = (($862) + 520192)|0;
          $864 = $863 >>> 16;
          $865 = $864 & 4;
          $866 = $865 | $861;
          $867 = $862 << $865;
          $868 = (($867) + 245760)|0;
          $869 = $868 >>> 16;
          $870 = $869 & 2;
          $871 = $866 | $870;
          $872 = (14 - ($871))|0;
          $873 = $867 << $870;
          $874 = $873 >>> 15;
          $875 = (($872) + ($874))|0;
          $876 = $875 << 1;
          $877 = (($875) + 7)|0;
          $878 = $qsize$0$i$i >>> $877;
          $879 = $878 & 1;
          $880 = $879 | $876;
          $I7$0$i$i = $880;
         }
        } while(0);
        $881 = (1206332 + ($I7$0$i$i<<2)|0);
        $882 = ((($726)) + 28|0);
        HEAP32[$882>>2] = $I7$0$i$i;
        $883 = ((($726)) + 16|0);
        $884 = ((($883)) + 4|0);
        HEAP32[$884>>2] = 0;
        HEAP32[$883>>2] = 0;
        $885 = HEAP32[(1206032)>>2]|0;
        $886 = 1 << $I7$0$i$i;
        $887 = $885 & $886;
        $888 = ($887|0)==(0);
        if ($888) {
         $889 = $885 | $886;
         HEAP32[(1206032)>>2] = $889;
         HEAP32[$881>>2] = $726;
         $890 = ((($726)) + 24|0);
         HEAP32[$890>>2] = $881;
         $891 = ((($726)) + 12|0);
         HEAP32[$891>>2] = $726;
         $892 = ((($726)) + 8|0);
         HEAP32[$892>>2] = $726;
         break;
        }
        $893 = HEAP32[$881>>2]|0;
        $894 = ($I7$0$i$i|0)==(31);
        $895 = $I7$0$i$i >>> 1;
        $896 = (25 - ($895))|0;
        $897 = $894 ? 0 : $896;
        $898 = $qsize$0$i$i << $897;
        $K8$0$i$i = $898;$T$0$i18$i = $893;
        while(1) {
         $899 = ((($T$0$i18$i)) + 4|0);
         $900 = HEAP32[$899>>2]|0;
         $901 = $900 & -8;
         $902 = ($901|0)==($qsize$0$i$i|0);
         if ($902) {
          $T$0$i18$i$lcssa = $T$0$i18$i;
          label = 281;
          break;
         }
         $903 = $K8$0$i$i >>> 31;
         $904 = (((($T$0$i18$i)) + 16|0) + ($903<<2)|0);
         $905 = $K8$0$i$i << 1;
         $906 = HEAP32[$904>>2]|0;
         $907 = ($906|0)==(0|0);
         if ($907) {
          $$lcssa = $904;$T$0$i18$i$lcssa139 = $T$0$i18$i;
          label = 278;
          break;
         } else {
          $K8$0$i$i = $905;$T$0$i18$i = $906;
         }
        }
        if ((label|0) == 278) {
         $908 = HEAP32[(1206044)>>2]|0;
         $909 = ($$lcssa>>>0)<($908>>>0);
         if ($909) {
          _abort();
          // unreachable;
         } else {
          HEAP32[$$lcssa>>2] = $726;
          $910 = ((($726)) + 24|0);
          HEAP32[$910>>2] = $T$0$i18$i$lcssa139;
          $911 = ((($726)) + 12|0);
          HEAP32[$911>>2] = $726;
          $912 = ((($726)) + 8|0);
          HEAP32[$912>>2] = $726;
          break;
         }
        }
        else if ((label|0) == 281) {
         $913 = ((($T$0$i18$i$lcssa)) + 8|0);
         $914 = HEAP32[$913>>2]|0;
         $915 = HEAP32[(1206044)>>2]|0;
         $916 = ($914>>>0)>=($915>>>0);
         $not$$i20$i = ($T$0$i18$i$lcssa>>>0)>=($915>>>0);
         $917 = $916 & $not$$i20$i;
         if ($917) {
          $918 = ((($914)) + 12|0);
          HEAP32[$918>>2] = $726;
          HEAP32[$913>>2] = $726;
          $919 = ((($726)) + 8|0);
          HEAP32[$919>>2] = $914;
          $920 = ((($726)) + 12|0);
          HEAP32[$920>>2] = $T$0$i18$i$lcssa;
          $921 = ((($726)) + 24|0);
          HEAP32[$921>>2] = 0;
          break;
         } else {
          _abort();
          // unreachable;
         }
        }
       }
      } while(0);
      $1052 = ((($714)) + 8|0);
      $$0 = $1052;
      return ($$0|0);
     } else {
      $sp$0$i$i$i = (1206476);
     }
    }
    while(1) {
     $922 = HEAP32[$sp$0$i$i$i>>2]|0;
     $923 = ($922>>>0)>($635>>>0);
     if (!($923)) {
      $924 = ((($sp$0$i$i$i)) + 4|0);
      $925 = HEAP32[$924>>2]|0;
      $926 = (($922) + ($925)|0);
      $927 = ($926>>>0)>($635>>>0);
      if ($927) {
       $$lcssa142 = $926;
       break;
      }
     }
     $928 = ((($sp$0$i$i$i)) + 8|0);
     $929 = HEAP32[$928>>2]|0;
     $sp$0$i$i$i = $929;
    }
    $930 = ((($$lcssa142)) + -47|0);
    $931 = ((($930)) + 8|0);
    $932 = $931;
    $933 = $932 & 7;
    $934 = ($933|0)==(0);
    $935 = (0 - ($932))|0;
    $936 = $935 & 7;
    $937 = $934 ? 0 : $936;
    $938 = (($930) + ($937)|0);
    $939 = ((($635)) + 16|0);
    $940 = ($938>>>0)<($939>>>0);
    $941 = $940 ? $635 : $938;
    $942 = ((($941)) + 8|0);
    $943 = ((($941)) + 24|0);
    $944 = (($tsize$745$i) + -40)|0;
    $945 = ((($tbase$746$i)) + 8|0);
    $946 = $945;
    $947 = $946 & 7;
    $948 = ($947|0)==(0);
    $949 = (0 - ($946))|0;
    $950 = $949 & 7;
    $951 = $948 ? 0 : $950;
    $952 = (($tbase$746$i) + ($951)|0);
    $953 = (($944) - ($951))|0;
    HEAP32[(1206052)>>2] = $952;
    HEAP32[(1206040)>>2] = $953;
    $954 = $953 | 1;
    $955 = ((($952)) + 4|0);
    HEAP32[$955>>2] = $954;
    $956 = (($952) + ($953)|0);
    $957 = ((($956)) + 4|0);
    HEAP32[$957>>2] = 40;
    $958 = HEAP32[(1206516)>>2]|0;
    HEAP32[(1206056)>>2] = $958;
    $959 = ((($941)) + 4|0);
    HEAP32[$959>>2] = 27;
    ;HEAP32[$942>>2]=HEAP32[(1206476)>>2]|0;HEAP32[$942+4>>2]=HEAP32[(1206476)+4>>2]|0;HEAP32[$942+8>>2]=HEAP32[(1206476)+8>>2]|0;HEAP32[$942+12>>2]=HEAP32[(1206476)+12>>2]|0;
    HEAP32[(1206476)>>2] = $tbase$746$i;
    HEAP32[(1206480)>>2] = $tsize$745$i;
    HEAP32[(1206488)>>2] = 0;
    HEAP32[(1206484)>>2] = $942;
    $p$0$i$i = $943;
    while(1) {
     $960 = ((($p$0$i$i)) + 4|0);
     HEAP32[$960>>2] = 7;
     $961 = ((($960)) + 4|0);
     $962 = ($961>>>0)<($$lcssa142>>>0);
     if ($962) {
      $p$0$i$i = $960;
     } else {
      break;
     }
    }
    $963 = ($941|0)==($635|0);
    if (!($963)) {
     $964 = $941;
     $965 = $635;
     $966 = (($964) - ($965))|0;
     $967 = HEAP32[$959>>2]|0;
     $968 = $967 & -2;
     HEAP32[$959>>2] = $968;
     $969 = $966 | 1;
     $970 = ((($635)) + 4|0);
     HEAP32[$970>>2] = $969;
     HEAP32[$941>>2] = $966;
     $971 = $966 >>> 3;
     $972 = ($966>>>0)<(256);
     if ($972) {
      $973 = $971 << 1;
      $974 = (1206068 + ($973<<2)|0);
      $975 = HEAP32[301507]|0;
      $976 = 1 << $971;
      $977 = $975 & $976;
      $978 = ($977|0)==(0);
      if ($978) {
       $979 = $975 | $976;
       HEAP32[301507] = $979;
       $$pre$i$i = ((($974)) + 8|0);
       $$pre$phi$i$iZ2D = $$pre$i$i;$F$0$i$i = $974;
      } else {
       $980 = ((($974)) + 8|0);
       $981 = HEAP32[$980>>2]|0;
       $982 = HEAP32[(1206044)>>2]|0;
       $983 = ($981>>>0)<($982>>>0);
       if ($983) {
        _abort();
        // unreachable;
       } else {
        $$pre$phi$i$iZ2D = $980;$F$0$i$i = $981;
       }
      }
      HEAP32[$$pre$phi$i$iZ2D>>2] = $635;
      $984 = ((($F$0$i$i)) + 12|0);
      HEAP32[$984>>2] = $635;
      $985 = ((($635)) + 8|0);
      HEAP32[$985>>2] = $F$0$i$i;
      $986 = ((($635)) + 12|0);
      HEAP32[$986>>2] = $974;
      break;
     }
     $987 = $966 >>> 8;
     $988 = ($987|0)==(0);
     if ($988) {
      $I1$0$i$i = 0;
     } else {
      $989 = ($966>>>0)>(16777215);
      if ($989) {
       $I1$0$i$i = 31;
      } else {
       $990 = (($987) + 1048320)|0;
       $991 = $990 >>> 16;
       $992 = $991 & 8;
       $993 = $987 << $992;
       $994 = (($993) + 520192)|0;
       $995 = $994 >>> 16;
       $996 = $995 & 4;
       $997 = $996 | $992;
       $998 = $993 << $996;
       $999 = (($998) + 245760)|0;
       $1000 = $999 >>> 16;
       $1001 = $1000 & 2;
       $1002 = $997 | $1001;
       $1003 = (14 - ($1002))|0;
       $1004 = $998 << $1001;
       $1005 = $1004 >>> 15;
       $1006 = (($1003) + ($1005))|0;
       $1007 = $1006 << 1;
       $1008 = (($1006) + 7)|0;
       $1009 = $966 >>> $1008;
       $1010 = $1009 & 1;
       $1011 = $1010 | $1007;
       $I1$0$i$i = $1011;
      }
     }
     $1012 = (1206332 + ($I1$0$i$i<<2)|0);
     $1013 = ((($635)) + 28|0);
     HEAP32[$1013>>2] = $I1$0$i$i;
     $1014 = ((($635)) + 20|0);
     HEAP32[$1014>>2] = 0;
     HEAP32[$939>>2] = 0;
     $1015 = HEAP32[(1206032)>>2]|0;
     $1016 = 1 << $I1$0$i$i;
     $1017 = $1015 & $1016;
     $1018 = ($1017|0)==(0);
     if ($1018) {
      $1019 = $1015 | $1016;
      HEAP32[(1206032)>>2] = $1019;
      HEAP32[$1012>>2] = $635;
      $1020 = ((($635)) + 24|0);
      HEAP32[$1020>>2] = $1012;
      $1021 = ((($635)) + 12|0);
      HEAP32[$1021>>2] = $635;
      $1022 = ((($635)) + 8|0);
      HEAP32[$1022>>2] = $635;
      break;
     }
     $1023 = HEAP32[$1012>>2]|0;
     $1024 = ($I1$0$i$i|0)==(31);
     $1025 = $I1$0$i$i >>> 1;
     $1026 = (25 - ($1025))|0;
     $1027 = $1024 ? 0 : $1026;
     $1028 = $966 << $1027;
     $K2$0$i$i = $1028;$T$0$i$i = $1023;
     while(1) {
      $1029 = ((($T$0$i$i)) + 4|0);
      $1030 = HEAP32[$1029>>2]|0;
      $1031 = $1030 & -8;
      $1032 = ($1031|0)==($966|0);
      if ($1032) {
       $T$0$i$i$lcssa = $T$0$i$i;
       label = 307;
       break;
      }
      $1033 = $K2$0$i$i >>> 31;
      $1034 = (((($T$0$i$i)) + 16|0) + ($1033<<2)|0);
      $1035 = $K2$0$i$i << 1;
      $1036 = HEAP32[$1034>>2]|0;
      $1037 = ($1036|0)==(0|0);
      if ($1037) {
       $$lcssa141 = $1034;$T$0$i$i$lcssa140 = $T$0$i$i;
       label = 304;
       break;
      } else {
       $K2$0$i$i = $1035;$T$0$i$i = $1036;
      }
     }
     if ((label|0) == 304) {
      $1038 = HEAP32[(1206044)>>2]|0;
      $1039 = ($$lcssa141>>>0)<($1038>>>0);
      if ($1039) {
       _abort();
       // unreachable;
      } else {
       HEAP32[$$lcssa141>>2] = $635;
       $1040 = ((($635)) + 24|0);
       HEAP32[$1040>>2] = $T$0$i$i$lcssa140;
       $1041 = ((($635)) + 12|0);
       HEAP32[$1041>>2] = $635;
       $1042 = ((($635)) + 8|0);
       HEAP32[$1042>>2] = $635;
       break;
      }
     }
     else if ((label|0) == 307) {
      $1043 = ((($T$0$i$i$lcssa)) + 8|0);
      $1044 = HEAP32[$1043>>2]|0;
      $1045 = HEAP32[(1206044)>>2]|0;
      $1046 = ($1044>>>0)>=($1045>>>0);
      $not$$i$i = ($T$0$i$i$lcssa>>>0)>=($1045>>>0);
      $1047 = $1046 & $not$$i$i;
      if ($1047) {
       $1048 = ((($1044)) + 12|0);
       HEAP32[$1048>>2] = $635;
       HEAP32[$1043>>2] = $635;
       $1049 = ((($635)) + 8|0);
       HEAP32[$1049>>2] = $1044;
       $1050 = ((($635)) + 12|0);
       HEAP32[$1050>>2] = $T$0$i$i$lcssa;
       $1051 = ((($635)) + 24|0);
       HEAP32[$1051>>2] = 0;
       break;
      } else {
       _abort();
       // unreachable;
      }
     }
    }
   }
  } while(0);
  $1053 = HEAP32[(1206040)>>2]|0;
  $1054 = ($1053>>>0)>($nb$0>>>0);
  if ($1054) {
   $1055 = (($1053) - ($nb$0))|0;
   HEAP32[(1206040)>>2] = $1055;
   $1056 = HEAP32[(1206052)>>2]|0;
   $1057 = (($1056) + ($nb$0)|0);
   HEAP32[(1206052)>>2] = $1057;
   $1058 = $1055 | 1;
   $1059 = ((($1057)) + 4|0);
   HEAP32[$1059>>2] = $1058;
   $1060 = $nb$0 | 3;
   $1061 = ((($1056)) + 4|0);
   HEAP32[$1061>>2] = $1060;
   $1062 = ((($1056)) + 8|0);
   $$0 = $1062;
   return ($$0|0);
  }
 }
 $1063 = (___errno_location()|0);
 HEAP32[$1063>>2] = 12;
 $$0 = 0;
 return ($$0|0);
}
function _free($mem) {
 $mem = $mem|0;
 var $$lcssa = 0, $$pre = 0, $$pre$phi41Z2D = 0, $$pre$phi43Z2D = 0, $$pre$phiZ2D = 0, $$pre40 = 0, $$pre42 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0;
 var $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0;
 var $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0;
 var $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0;
 var $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0;
 var $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0;
 var $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0;
 var $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0;
 var $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $F18$0 = 0, $I20$0 = 0, $K21$0 = 0, $R$1 = 0, $R$1$lcssa = 0, $R$3 = 0, $R8$1 = 0, $R8$1$lcssa = 0, $R8$3 = 0, $RP$1 = 0, $RP$1$lcssa = 0, $RP10$1 = 0, $RP10$1$lcssa = 0;
 var $T$0 = 0, $T$0$lcssa = 0, $T$0$lcssa48 = 0, $cond20 = 0, $cond21 = 0, $not$ = 0, $p$1 = 0, $psize$1 = 0, $psize$2 = 0, $sp$0$i = 0, $sp$0$in$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mem|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($mem)) + -8|0);
 $2 = HEAP32[(1206044)>>2]|0;
 $3 = ($1>>>0)<($2>>>0);
 if ($3) {
  _abort();
  // unreachable;
 }
 $4 = ((($mem)) + -4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $5 & 3;
 $7 = ($6|0)==(1);
 if ($7) {
  _abort();
  // unreachable;
 }
 $8 = $5 & -8;
 $9 = (($1) + ($8)|0);
 $10 = $5 & 1;
 $11 = ($10|0)==(0);
 do {
  if ($11) {
   $12 = HEAP32[$1>>2]|0;
   $13 = ($6|0)==(0);
   if ($13) {
    return;
   }
   $14 = (0 - ($12))|0;
   $15 = (($1) + ($14)|0);
   $16 = (($12) + ($8))|0;
   $17 = ($15>>>0)<($2>>>0);
   if ($17) {
    _abort();
    // unreachable;
   }
   $18 = HEAP32[(1206048)>>2]|0;
   $19 = ($15|0)==($18|0);
   if ($19) {
    $104 = ((($9)) + 4|0);
    $105 = HEAP32[$104>>2]|0;
    $106 = $105 & 3;
    $107 = ($106|0)==(3);
    if (!($107)) {
     $p$1 = $15;$psize$1 = $16;
     break;
    }
    HEAP32[(1206036)>>2] = $16;
    $108 = $105 & -2;
    HEAP32[$104>>2] = $108;
    $109 = $16 | 1;
    $110 = ((($15)) + 4|0);
    HEAP32[$110>>2] = $109;
    $111 = (($15) + ($16)|0);
    HEAP32[$111>>2] = $16;
    return;
   }
   $20 = $12 >>> 3;
   $21 = ($12>>>0)<(256);
   if ($21) {
    $22 = ((($15)) + 8|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = ((($15)) + 12|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = $20 << 1;
    $27 = (1206068 + ($26<<2)|0);
    $28 = ($23|0)==($27|0);
    if (!($28)) {
     $29 = ($23>>>0)<($2>>>0);
     if ($29) {
      _abort();
      // unreachable;
     }
     $30 = ((($23)) + 12|0);
     $31 = HEAP32[$30>>2]|0;
     $32 = ($31|0)==($15|0);
     if (!($32)) {
      _abort();
      // unreachable;
     }
    }
    $33 = ($25|0)==($23|0);
    if ($33) {
     $34 = 1 << $20;
     $35 = $34 ^ -1;
     $36 = HEAP32[301507]|0;
     $37 = $36 & $35;
     HEAP32[301507] = $37;
     $p$1 = $15;$psize$1 = $16;
     break;
    }
    $38 = ($25|0)==($27|0);
    if ($38) {
     $$pre42 = ((($25)) + 8|0);
     $$pre$phi43Z2D = $$pre42;
    } else {
     $39 = ($25>>>0)<($2>>>0);
     if ($39) {
      _abort();
      // unreachable;
     }
     $40 = ((($25)) + 8|0);
     $41 = HEAP32[$40>>2]|0;
     $42 = ($41|0)==($15|0);
     if ($42) {
      $$pre$phi43Z2D = $40;
     } else {
      _abort();
      // unreachable;
     }
    }
    $43 = ((($23)) + 12|0);
    HEAP32[$43>>2] = $25;
    HEAP32[$$pre$phi43Z2D>>2] = $23;
    $p$1 = $15;$psize$1 = $16;
    break;
   }
   $44 = ((($15)) + 24|0);
   $45 = HEAP32[$44>>2]|0;
   $46 = ((($15)) + 12|0);
   $47 = HEAP32[$46>>2]|0;
   $48 = ($47|0)==($15|0);
   do {
    if ($48) {
     $58 = ((($15)) + 16|0);
     $59 = ((($58)) + 4|0);
     $60 = HEAP32[$59>>2]|0;
     $61 = ($60|0)==(0|0);
     if ($61) {
      $62 = HEAP32[$58>>2]|0;
      $63 = ($62|0)==(0|0);
      if ($63) {
       $R$3 = 0;
       break;
      } else {
       $R$1 = $62;$RP$1 = $58;
      }
     } else {
      $R$1 = $60;$RP$1 = $59;
     }
     while(1) {
      $64 = ((($R$1)) + 20|0);
      $65 = HEAP32[$64>>2]|0;
      $66 = ($65|0)==(0|0);
      if (!($66)) {
       $R$1 = $65;$RP$1 = $64;
       continue;
      }
      $67 = ((($R$1)) + 16|0);
      $68 = HEAP32[$67>>2]|0;
      $69 = ($68|0)==(0|0);
      if ($69) {
       $R$1$lcssa = $R$1;$RP$1$lcssa = $RP$1;
       break;
      } else {
       $R$1 = $68;$RP$1 = $67;
      }
     }
     $70 = ($RP$1$lcssa>>>0)<($2>>>0);
     if ($70) {
      _abort();
      // unreachable;
     } else {
      HEAP32[$RP$1$lcssa>>2] = 0;
      $R$3 = $R$1$lcssa;
      break;
     }
    } else {
     $49 = ((($15)) + 8|0);
     $50 = HEAP32[$49>>2]|0;
     $51 = ($50>>>0)<($2>>>0);
     if ($51) {
      _abort();
      // unreachable;
     }
     $52 = ((($50)) + 12|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ($53|0)==($15|0);
     if (!($54)) {
      _abort();
      // unreachable;
     }
     $55 = ((($47)) + 8|0);
     $56 = HEAP32[$55>>2]|0;
     $57 = ($56|0)==($15|0);
     if ($57) {
      HEAP32[$52>>2] = $47;
      HEAP32[$55>>2] = $50;
      $R$3 = $47;
      break;
     } else {
      _abort();
      // unreachable;
     }
    }
   } while(0);
   $71 = ($45|0)==(0|0);
   if ($71) {
    $p$1 = $15;$psize$1 = $16;
   } else {
    $72 = ((($15)) + 28|0);
    $73 = HEAP32[$72>>2]|0;
    $74 = (1206332 + ($73<<2)|0);
    $75 = HEAP32[$74>>2]|0;
    $76 = ($15|0)==($75|0);
    if ($76) {
     HEAP32[$74>>2] = $R$3;
     $cond20 = ($R$3|0)==(0|0);
     if ($cond20) {
      $77 = 1 << $73;
      $78 = $77 ^ -1;
      $79 = HEAP32[(1206032)>>2]|0;
      $80 = $79 & $78;
      HEAP32[(1206032)>>2] = $80;
      $p$1 = $15;$psize$1 = $16;
      break;
     }
    } else {
     $81 = HEAP32[(1206044)>>2]|0;
     $82 = ($45>>>0)<($81>>>0);
     if ($82) {
      _abort();
      // unreachable;
     }
     $83 = ((($45)) + 16|0);
     $84 = HEAP32[$83>>2]|0;
     $85 = ($84|0)==($15|0);
     if ($85) {
      HEAP32[$83>>2] = $R$3;
     } else {
      $86 = ((($45)) + 20|0);
      HEAP32[$86>>2] = $R$3;
     }
     $87 = ($R$3|0)==(0|0);
     if ($87) {
      $p$1 = $15;$psize$1 = $16;
      break;
     }
    }
    $88 = HEAP32[(1206044)>>2]|0;
    $89 = ($R$3>>>0)<($88>>>0);
    if ($89) {
     _abort();
     // unreachable;
    }
    $90 = ((($R$3)) + 24|0);
    HEAP32[$90>>2] = $45;
    $91 = ((($15)) + 16|0);
    $92 = HEAP32[$91>>2]|0;
    $93 = ($92|0)==(0|0);
    do {
     if (!($93)) {
      $94 = ($92>>>0)<($88>>>0);
      if ($94) {
       _abort();
       // unreachable;
      } else {
       $95 = ((($R$3)) + 16|0);
       HEAP32[$95>>2] = $92;
       $96 = ((($92)) + 24|0);
       HEAP32[$96>>2] = $R$3;
       break;
      }
     }
    } while(0);
    $97 = ((($91)) + 4|0);
    $98 = HEAP32[$97>>2]|0;
    $99 = ($98|0)==(0|0);
    if ($99) {
     $p$1 = $15;$psize$1 = $16;
    } else {
     $100 = HEAP32[(1206044)>>2]|0;
     $101 = ($98>>>0)<($100>>>0);
     if ($101) {
      _abort();
      // unreachable;
     } else {
      $102 = ((($R$3)) + 20|0);
      HEAP32[$102>>2] = $98;
      $103 = ((($98)) + 24|0);
      HEAP32[$103>>2] = $R$3;
      $p$1 = $15;$psize$1 = $16;
      break;
     }
    }
   }
  } else {
   $p$1 = $1;$psize$1 = $8;
  }
 } while(0);
 $112 = ($p$1>>>0)<($9>>>0);
 if (!($112)) {
  _abort();
  // unreachable;
 }
 $113 = ((($9)) + 4|0);
 $114 = HEAP32[$113>>2]|0;
 $115 = $114 & 1;
 $116 = ($115|0)==(0);
 if ($116) {
  _abort();
  // unreachable;
 }
 $117 = $114 & 2;
 $118 = ($117|0)==(0);
 if ($118) {
  $119 = HEAP32[(1206052)>>2]|0;
  $120 = ($9|0)==($119|0);
  if ($120) {
   $121 = HEAP32[(1206040)>>2]|0;
   $122 = (($121) + ($psize$1))|0;
   HEAP32[(1206040)>>2] = $122;
   HEAP32[(1206052)>>2] = $p$1;
   $123 = $122 | 1;
   $124 = ((($p$1)) + 4|0);
   HEAP32[$124>>2] = $123;
   $125 = HEAP32[(1206048)>>2]|0;
   $126 = ($p$1|0)==($125|0);
   if (!($126)) {
    return;
   }
   HEAP32[(1206048)>>2] = 0;
   HEAP32[(1206036)>>2] = 0;
   return;
  }
  $127 = HEAP32[(1206048)>>2]|0;
  $128 = ($9|0)==($127|0);
  if ($128) {
   $129 = HEAP32[(1206036)>>2]|0;
   $130 = (($129) + ($psize$1))|0;
   HEAP32[(1206036)>>2] = $130;
   HEAP32[(1206048)>>2] = $p$1;
   $131 = $130 | 1;
   $132 = ((($p$1)) + 4|0);
   HEAP32[$132>>2] = $131;
   $133 = (($p$1) + ($130)|0);
   HEAP32[$133>>2] = $130;
   return;
  }
  $134 = $114 & -8;
  $135 = (($134) + ($psize$1))|0;
  $136 = $114 >>> 3;
  $137 = ($114>>>0)<(256);
  do {
   if ($137) {
    $138 = ((($9)) + 8|0);
    $139 = HEAP32[$138>>2]|0;
    $140 = ((($9)) + 12|0);
    $141 = HEAP32[$140>>2]|0;
    $142 = $136 << 1;
    $143 = (1206068 + ($142<<2)|0);
    $144 = ($139|0)==($143|0);
    if (!($144)) {
     $145 = HEAP32[(1206044)>>2]|0;
     $146 = ($139>>>0)<($145>>>0);
     if ($146) {
      _abort();
      // unreachable;
     }
     $147 = ((($139)) + 12|0);
     $148 = HEAP32[$147>>2]|0;
     $149 = ($148|0)==($9|0);
     if (!($149)) {
      _abort();
      // unreachable;
     }
    }
    $150 = ($141|0)==($139|0);
    if ($150) {
     $151 = 1 << $136;
     $152 = $151 ^ -1;
     $153 = HEAP32[301507]|0;
     $154 = $153 & $152;
     HEAP32[301507] = $154;
     break;
    }
    $155 = ($141|0)==($143|0);
    if ($155) {
     $$pre40 = ((($141)) + 8|0);
     $$pre$phi41Z2D = $$pre40;
    } else {
     $156 = HEAP32[(1206044)>>2]|0;
     $157 = ($141>>>0)<($156>>>0);
     if ($157) {
      _abort();
      // unreachable;
     }
     $158 = ((($141)) + 8|0);
     $159 = HEAP32[$158>>2]|0;
     $160 = ($159|0)==($9|0);
     if ($160) {
      $$pre$phi41Z2D = $158;
     } else {
      _abort();
      // unreachable;
     }
    }
    $161 = ((($139)) + 12|0);
    HEAP32[$161>>2] = $141;
    HEAP32[$$pre$phi41Z2D>>2] = $139;
   } else {
    $162 = ((($9)) + 24|0);
    $163 = HEAP32[$162>>2]|0;
    $164 = ((($9)) + 12|0);
    $165 = HEAP32[$164>>2]|0;
    $166 = ($165|0)==($9|0);
    do {
     if ($166) {
      $177 = ((($9)) + 16|0);
      $178 = ((($177)) + 4|0);
      $179 = HEAP32[$178>>2]|0;
      $180 = ($179|0)==(0|0);
      if ($180) {
       $181 = HEAP32[$177>>2]|0;
       $182 = ($181|0)==(0|0);
       if ($182) {
        $R8$3 = 0;
        break;
       } else {
        $R8$1 = $181;$RP10$1 = $177;
       }
      } else {
       $R8$1 = $179;$RP10$1 = $178;
      }
      while(1) {
       $183 = ((($R8$1)) + 20|0);
       $184 = HEAP32[$183>>2]|0;
       $185 = ($184|0)==(0|0);
       if (!($185)) {
        $R8$1 = $184;$RP10$1 = $183;
        continue;
       }
       $186 = ((($R8$1)) + 16|0);
       $187 = HEAP32[$186>>2]|0;
       $188 = ($187|0)==(0|0);
       if ($188) {
        $R8$1$lcssa = $R8$1;$RP10$1$lcssa = $RP10$1;
        break;
       } else {
        $R8$1 = $187;$RP10$1 = $186;
       }
      }
      $189 = HEAP32[(1206044)>>2]|0;
      $190 = ($RP10$1$lcssa>>>0)<($189>>>0);
      if ($190) {
       _abort();
       // unreachable;
      } else {
       HEAP32[$RP10$1$lcssa>>2] = 0;
       $R8$3 = $R8$1$lcssa;
       break;
      }
     } else {
      $167 = ((($9)) + 8|0);
      $168 = HEAP32[$167>>2]|0;
      $169 = HEAP32[(1206044)>>2]|0;
      $170 = ($168>>>0)<($169>>>0);
      if ($170) {
       _abort();
       // unreachable;
      }
      $171 = ((($168)) + 12|0);
      $172 = HEAP32[$171>>2]|0;
      $173 = ($172|0)==($9|0);
      if (!($173)) {
       _abort();
       // unreachable;
      }
      $174 = ((($165)) + 8|0);
      $175 = HEAP32[$174>>2]|0;
      $176 = ($175|0)==($9|0);
      if ($176) {
       HEAP32[$171>>2] = $165;
       HEAP32[$174>>2] = $168;
       $R8$3 = $165;
       break;
      } else {
       _abort();
       // unreachable;
      }
     }
    } while(0);
    $191 = ($163|0)==(0|0);
    if (!($191)) {
     $192 = ((($9)) + 28|0);
     $193 = HEAP32[$192>>2]|0;
     $194 = (1206332 + ($193<<2)|0);
     $195 = HEAP32[$194>>2]|0;
     $196 = ($9|0)==($195|0);
     if ($196) {
      HEAP32[$194>>2] = $R8$3;
      $cond21 = ($R8$3|0)==(0|0);
      if ($cond21) {
       $197 = 1 << $193;
       $198 = $197 ^ -1;
       $199 = HEAP32[(1206032)>>2]|0;
       $200 = $199 & $198;
       HEAP32[(1206032)>>2] = $200;
       break;
      }
     } else {
      $201 = HEAP32[(1206044)>>2]|0;
      $202 = ($163>>>0)<($201>>>0);
      if ($202) {
       _abort();
       // unreachable;
      }
      $203 = ((($163)) + 16|0);
      $204 = HEAP32[$203>>2]|0;
      $205 = ($204|0)==($9|0);
      if ($205) {
       HEAP32[$203>>2] = $R8$3;
      } else {
       $206 = ((($163)) + 20|0);
       HEAP32[$206>>2] = $R8$3;
      }
      $207 = ($R8$3|0)==(0|0);
      if ($207) {
       break;
      }
     }
     $208 = HEAP32[(1206044)>>2]|0;
     $209 = ($R8$3>>>0)<($208>>>0);
     if ($209) {
      _abort();
      // unreachable;
     }
     $210 = ((($R8$3)) + 24|0);
     HEAP32[$210>>2] = $163;
     $211 = ((($9)) + 16|0);
     $212 = HEAP32[$211>>2]|0;
     $213 = ($212|0)==(0|0);
     do {
      if (!($213)) {
       $214 = ($212>>>0)<($208>>>0);
       if ($214) {
        _abort();
        // unreachable;
       } else {
        $215 = ((($R8$3)) + 16|0);
        HEAP32[$215>>2] = $212;
        $216 = ((($212)) + 24|0);
        HEAP32[$216>>2] = $R8$3;
        break;
       }
      }
     } while(0);
     $217 = ((($211)) + 4|0);
     $218 = HEAP32[$217>>2]|0;
     $219 = ($218|0)==(0|0);
     if (!($219)) {
      $220 = HEAP32[(1206044)>>2]|0;
      $221 = ($218>>>0)<($220>>>0);
      if ($221) {
       _abort();
       // unreachable;
      } else {
       $222 = ((($R8$3)) + 20|0);
       HEAP32[$222>>2] = $218;
       $223 = ((($218)) + 24|0);
       HEAP32[$223>>2] = $R8$3;
       break;
      }
     }
    }
   }
  } while(0);
  $224 = $135 | 1;
  $225 = ((($p$1)) + 4|0);
  HEAP32[$225>>2] = $224;
  $226 = (($p$1) + ($135)|0);
  HEAP32[$226>>2] = $135;
  $227 = HEAP32[(1206048)>>2]|0;
  $228 = ($p$1|0)==($227|0);
  if ($228) {
   HEAP32[(1206036)>>2] = $135;
   return;
  } else {
   $psize$2 = $135;
  }
 } else {
  $229 = $114 & -2;
  HEAP32[$113>>2] = $229;
  $230 = $psize$1 | 1;
  $231 = ((($p$1)) + 4|0);
  HEAP32[$231>>2] = $230;
  $232 = (($p$1) + ($psize$1)|0);
  HEAP32[$232>>2] = $psize$1;
  $psize$2 = $psize$1;
 }
 $233 = $psize$2 >>> 3;
 $234 = ($psize$2>>>0)<(256);
 if ($234) {
  $235 = $233 << 1;
  $236 = (1206068 + ($235<<2)|0);
  $237 = HEAP32[301507]|0;
  $238 = 1 << $233;
  $239 = $237 & $238;
  $240 = ($239|0)==(0);
  if ($240) {
   $241 = $237 | $238;
   HEAP32[301507] = $241;
   $$pre = ((($236)) + 8|0);
   $$pre$phiZ2D = $$pre;$F18$0 = $236;
  } else {
   $242 = ((($236)) + 8|0);
   $243 = HEAP32[$242>>2]|0;
   $244 = HEAP32[(1206044)>>2]|0;
   $245 = ($243>>>0)<($244>>>0);
   if ($245) {
    _abort();
    // unreachable;
   } else {
    $$pre$phiZ2D = $242;$F18$0 = $243;
   }
  }
  HEAP32[$$pre$phiZ2D>>2] = $p$1;
  $246 = ((($F18$0)) + 12|0);
  HEAP32[$246>>2] = $p$1;
  $247 = ((($p$1)) + 8|0);
  HEAP32[$247>>2] = $F18$0;
  $248 = ((($p$1)) + 12|0);
  HEAP32[$248>>2] = $236;
  return;
 }
 $249 = $psize$2 >>> 8;
 $250 = ($249|0)==(0);
 if ($250) {
  $I20$0 = 0;
 } else {
  $251 = ($psize$2>>>0)>(16777215);
  if ($251) {
   $I20$0 = 31;
  } else {
   $252 = (($249) + 1048320)|0;
   $253 = $252 >>> 16;
   $254 = $253 & 8;
   $255 = $249 << $254;
   $256 = (($255) + 520192)|0;
   $257 = $256 >>> 16;
   $258 = $257 & 4;
   $259 = $258 | $254;
   $260 = $255 << $258;
   $261 = (($260) + 245760)|0;
   $262 = $261 >>> 16;
   $263 = $262 & 2;
   $264 = $259 | $263;
   $265 = (14 - ($264))|0;
   $266 = $260 << $263;
   $267 = $266 >>> 15;
   $268 = (($265) + ($267))|0;
   $269 = $268 << 1;
   $270 = (($268) + 7)|0;
   $271 = $psize$2 >>> $270;
   $272 = $271 & 1;
   $273 = $272 | $269;
   $I20$0 = $273;
  }
 }
 $274 = (1206332 + ($I20$0<<2)|0);
 $275 = ((($p$1)) + 28|0);
 HEAP32[$275>>2] = $I20$0;
 $276 = ((($p$1)) + 16|0);
 $277 = ((($p$1)) + 20|0);
 HEAP32[$277>>2] = 0;
 HEAP32[$276>>2] = 0;
 $278 = HEAP32[(1206032)>>2]|0;
 $279 = 1 << $I20$0;
 $280 = $278 & $279;
 $281 = ($280|0)==(0);
 do {
  if ($281) {
   $282 = $278 | $279;
   HEAP32[(1206032)>>2] = $282;
   HEAP32[$274>>2] = $p$1;
   $283 = ((($p$1)) + 24|0);
   HEAP32[$283>>2] = $274;
   $284 = ((($p$1)) + 12|0);
   HEAP32[$284>>2] = $p$1;
   $285 = ((($p$1)) + 8|0);
   HEAP32[$285>>2] = $p$1;
  } else {
   $286 = HEAP32[$274>>2]|0;
   $287 = ($I20$0|0)==(31);
   $288 = $I20$0 >>> 1;
   $289 = (25 - ($288))|0;
   $290 = $287 ? 0 : $289;
   $291 = $psize$2 << $290;
   $K21$0 = $291;$T$0 = $286;
   while(1) {
    $292 = ((($T$0)) + 4|0);
    $293 = HEAP32[$292>>2]|0;
    $294 = $293 & -8;
    $295 = ($294|0)==($psize$2|0);
    if ($295) {
     $T$0$lcssa = $T$0;
     label = 130;
     break;
    }
    $296 = $K21$0 >>> 31;
    $297 = (((($T$0)) + 16|0) + ($296<<2)|0);
    $298 = $K21$0 << 1;
    $299 = HEAP32[$297>>2]|0;
    $300 = ($299|0)==(0|0);
    if ($300) {
     $$lcssa = $297;$T$0$lcssa48 = $T$0;
     label = 127;
     break;
    } else {
     $K21$0 = $298;$T$0 = $299;
    }
   }
   if ((label|0) == 127) {
    $301 = HEAP32[(1206044)>>2]|0;
    $302 = ($$lcssa>>>0)<($301>>>0);
    if ($302) {
     _abort();
     // unreachable;
    } else {
     HEAP32[$$lcssa>>2] = $p$1;
     $303 = ((($p$1)) + 24|0);
     HEAP32[$303>>2] = $T$0$lcssa48;
     $304 = ((($p$1)) + 12|0);
     HEAP32[$304>>2] = $p$1;
     $305 = ((($p$1)) + 8|0);
     HEAP32[$305>>2] = $p$1;
     break;
    }
   }
   else if ((label|0) == 130) {
    $306 = ((($T$0$lcssa)) + 8|0);
    $307 = HEAP32[$306>>2]|0;
    $308 = HEAP32[(1206044)>>2]|0;
    $309 = ($307>>>0)>=($308>>>0);
    $not$ = ($T$0$lcssa>>>0)>=($308>>>0);
    $310 = $309 & $not$;
    if ($310) {
     $311 = ((($307)) + 12|0);
     HEAP32[$311>>2] = $p$1;
     HEAP32[$306>>2] = $p$1;
     $312 = ((($p$1)) + 8|0);
     HEAP32[$312>>2] = $307;
     $313 = ((($p$1)) + 12|0);
     HEAP32[$313>>2] = $T$0$lcssa;
     $314 = ((($p$1)) + 24|0);
     HEAP32[$314>>2] = 0;
     break;
    } else {
     _abort();
     // unreachable;
    }
   }
  }
 } while(0);
 $315 = HEAP32[(1206060)>>2]|0;
 $316 = (($315) + -1)|0;
 HEAP32[(1206060)>>2] = $316;
 $317 = ($316|0)==(0);
 if ($317) {
  $sp$0$in$i = (1206484);
 } else {
  return;
 }
 while(1) {
  $sp$0$i = HEAP32[$sp$0$in$i>>2]|0;
  $318 = ($sp$0$i|0)==(0|0);
  $319 = ((($sp$0$i)) + 8|0);
  if ($318) {
   break;
  } else {
   $sp$0$in$i = $319;
  }
 }
 HEAP32[(1206060)>>2] = -1;
 return;
}
function _calloc($n_elements,$elem_size) {
 $n_elements = $n_elements|0;
 $elem_size = $elem_size|0;
 var $$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $req$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n_elements|0)==(0);
 if ($0) {
  $req$0 = 0;
 } else {
  $1 = Math_imul($elem_size, $n_elements)|0;
  $2 = $elem_size | $n_elements;
  $3 = ($2>>>0)>(65535);
  if ($3) {
   $4 = (($1>>>0) / ($n_elements>>>0))&-1;
   $5 = ($4|0)==($elem_size|0);
   $$ = $5 ? $1 : -1;
   $req$0 = $$;
  } else {
   $req$0 = $1;
  }
 }
 $6 = (_malloc($req$0)|0);
 $7 = ($6|0)==(0|0);
 if ($7) {
  return ($6|0);
 }
 $8 = ((($6)) + -4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 & 3;
 $11 = ($10|0)==(0);
 if ($11) {
  return ($6|0);
 }
 _memset(($6|0),0,($req$0|0))|0;
 return ($6|0);
}
function _realloc($oldmem,$bytes) {
 $oldmem = $oldmem|0;
 $bytes = $bytes|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $mem$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($oldmem|0)==(0|0);
 if ($0) {
  $1 = (_malloc($bytes)|0);
  $mem$1 = $1;
  return ($mem$1|0);
 }
 $2 = ($bytes>>>0)>(4294967231);
 if ($2) {
  $3 = (___errno_location()|0);
  HEAP32[$3>>2] = 12;
  $mem$1 = 0;
  return ($mem$1|0);
 }
 $4 = ($bytes>>>0)<(11);
 $5 = (($bytes) + 11)|0;
 $6 = $5 & -8;
 $7 = $4 ? 16 : $6;
 $8 = ((($oldmem)) + -8|0);
 $9 = (_try_realloc_chunk($8,$7)|0);
 $10 = ($9|0)==(0|0);
 if (!($10)) {
  $11 = ((($9)) + 8|0);
  $mem$1 = $11;
  return ($mem$1|0);
 }
 $12 = (_malloc($bytes)|0);
 $13 = ($12|0)==(0|0);
 if ($13) {
  $mem$1 = 0;
  return ($mem$1|0);
 }
 $14 = ((($oldmem)) + -4|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = $15 & -8;
 $17 = $15 & 3;
 $18 = ($17|0)==(0);
 $19 = $18 ? 8 : 4;
 $20 = (($16) - ($19))|0;
 $21 = ($20>>>0)<($bytes>>>0);
 $22 = $21 ? $20 : $bytes;
 _memcpy(($12|0),($oldmem|0),($22|0))|0;
 _free($oldmem);
 $mem$1 = $12;
 return ($mem$1|0);
}
function _try_realloc_chunk($p,$nb) {
 $p = $p|0;
 $nb = $nb|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $R$1 = 0, $R$1$lcssa = 0;
 var $R$3 = 0, $RP$1 = 0, $RP$1$lcssa = 0, $cond = 0, $newp$2 = 0, $notlhs = 0, $notrhs = 0, $or$cond$not = 0, $or$cond3 = 0, $storemerge = 0, $storemerge1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($p)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & -8;
 $3 = (($p) + ($2)|0);
 $4 = HEAP32[(1206044)>>2]|0;
 $5 = $1 & 3;
 $notlhs = ($p>>>0)>=($4>>>0);
 $notrhs = ($5|0)!=(1);
 $or$cond$not = $notrhs & $notlhs;
 $6 = ($p>>>0)<($3>>>0);
 $or$cond3 = $or$cond$not & $6;
 if (!($or$cond3)) {
  _abort();
  // unreachable;
 }
 $7 = ((($3)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $8 & 1;
 $10 = ($9|0)==(0);
 if ($10) {
  _abort();
  // unreachable;
 }
 $11 = ($5|0)==(0);
 if ($11) {
  $12 = ($nb>>>0)<(256);
  if ($12) {
   $newp$2 = 0;
   return ($newp$2|0);
  }
  $13 = (($nb) + 4)|0;
  $14 = ($2>>>0)<($13>>>0);
  if (!($14)) {
   $15 = (($2) - ($nb))|0;
   $16 = HEAP32[(1206508)>>2]|0;
   $17 = $16 << 1;
   $18 = ($15>>>0)>($17>>>0);
   if (!($18)) {
    $newp$2 = $p;
    return ($newp$2|0);
   }
  }
  $newp$2 = 0;
  return ($newp$2|0);
 }
 $19 = ($2>>>0)<($nb>>>0);
 if (!($19)) {
  $20 = (($2) - ($nb))|0;
  $21 = ($20>>>0)>(15);
  if (!($21)) {
   $newp$2 = $p;
   return ($newp$2|0);
  }
  $22 = (($p) + ($nb)|0);
  $23 = $1 & 1;
  $24 = $23 | $nb;
  $25 = $24 | 2;
  HEAP32[$0>>2] = $25;
  $26 = ((($22)) + 4|0);
  $27 = $20 | 3;
  HEAP32[$26>>2] = $27;
  $28 = (($22) + ($20)|0);
  $29 = ((($28)) + 4|0);
  $30 = HEAP32[$29>>2]|0;
  $31 = $30 | 1;
  HEAP32[$29>>2] = $31;
  _dispose_chunk($22,$20);
  $newp$2 = $p;
  return ($newp$2|0);
 }
 $32 = HEAP32[(1206052)>>2]|0;
 $33 = ($3|0)==($32|0);
 if ($33) {
  $34 = HEAP32[(1206040)>>2]|0;
  $35 = (($34) + ($2))|0;
  $36 = ($35>>>0)>($nb>>>0);
  if (!($36)) {
   $newp$2 = 0;
   return ($newp$2|0);
  }
  $37 = (($35) - ($nb))|0;
  $38 = (($p) + ($nb)|0);
  $39 = $1 & 1;
  $40 = $39 | $nb;
  $41 = $40 | 2;
  HEAP32[$0>>2] = $41;
  $42 = ((($38)) + 4|0);
  $43 = $37 | 1;
  HEAP32[$42>>2] = $43;
  HEAP32[(1206052)>>2] = $38;
  HEAP32[(1206040)>>2] = $37;
  $newp$2 = $p;
  return ($newp$2|0);
 }
 $44 = HEAP32[(1206048)>>2]|0;
 $45 = ($3|0)==($44|0);
 if ($45) {
  $46 = HEAP32[(1206036)>>2]|0;
  $47 = (($46) + ($2))|0;
  $48 = ($47>>>0)<($nb>>>0);
  if ($48) {
   $newp$2 = 0;
   return ($newp$2|0);
  }
  $49 = (($47) - ($nb))|0;
  $50 = ($49>>>0)>(15);
  if ($50) {
   $51 = (($p) + ($nb)|0);
   $52 = (($51) + ($49)|0);
   $53 = $1 & 1;
   $54 = $53 | $nb;
   $55 = $54 | 2;
   HEAP32[$0>>2] = $55;
   $56 = ((($51)) + 4|0);
   $57 = $49 | 1;
   HEAP32[$56>>2] = $57;
   HEAP32[$52>>2] = $49;
   $58 = ((($52)) + 4|0);
   $59 = HEAP32[$58>>2]|0;
   $60 = $59 & -2;
   HEAP32[$58>>2] = $60;
   $storemerge = $51;$storemerge1 = $49;
  } else {
   $61 = $1 & 1;
   $62 = $61 | $47;
   $63 = $62 | 2;
   HEAP32[$0>>2] = $63;
   $64 = (($p) + ($47)|0);
   $65 = ((($64)) + 4|0);
   $66 = HEAP32[$65>>2]|0;
   $67 = $66 | 1;
   HEAP32[$65>>2] = $67;
   $storemerge = 0;$storemerge1 = 0;
  }
  HEAP32[(1206036)>>2] = $storemerge1;
  HEAP32[(1206048)>>2] = $storemerge;
  $newp$2 = $p;
  return ($newp$2|0);
 }
 $68 = $8 & 2;
 $69 = ($68|0)==(0);
 if (!($69)) {
  $newp$2 = 0;
  return ($newp$2|0);
 }
 $70 = $8 & -8;
 $71 = (($70) + ($2))|0;
 $72 = ($71>>>0)<($nb>>>0);
 if ($72) {
  $newp$2 = 0;
  return ($newp$2|0);
 }
 $73 = (($71) - ($nb))|0;
 $74 = $8 >>> 3;
 $75 = ($8>>>0)<(256);
 do {
  if ($75) {
   $76 = ((($3)) + 8|0);
   $77 = HEAP32[$76>>2]|0;
   $78 = ((($3)) + 12|0);
   $79 = HEAP32[$78>>2]|0;
   $80 = $74 << 1;
   $81 = (1206068 + ($80<<2)|0);
   $82 = ($77|0)==($81|0);
   if (!($82)) {
    $83 = ($77>>>0)<($4>>>0);
    if ($83) {
     _abort();
     // unreachable;
    }
    $84 = ((($77)) + 12|0);
    $85 = HEAP32[$84>>2]|0;
    $86 = ($85|0)==($3|0);
    if (!($86)) {
     _abort();
     // unreachable;
    }
   }
   $87 = ($79|0)==($77|0);
   if ($87) {
    $88 = 1 << $74;
    $89 = $88 ^ -1;
    $90 = HEAP32[301507]|0;
    $91 = $90 & $89;
    HEAP32[301507] = $91;
    break;
   }
   $92 = ($79|0)==($81|0);
   if ($92) {
    $$pre = ((($79)) + 8|0);
    $$pre$phiZ2D = $$pre;
   } else {
    $93 = ($79>>>0)<($4>>>0);
    if ($93) {
     _abort();
     // unreachable;
    }
    $94 = ((($79)) + 8|0);
    $95 = HEAP32[$94>>2]|0;
    $96 = ($95|0)==($3|0);
    if ($96) {
     $$pre$phiZ2D = $94;
    } else {
     _abort();
     // unreachable;
    }
   }
   $97 = ((($77)) + 12|0);
   HEAP32[$97>>2] = $79;
   HEAP32[$$pre$phiZ2D>>2] = $77;
  } else {
   $98 = ((($3)) + 24|0);
   $99 = HEAP32[$98>>2]|0;
   $100 = ((($3)) + 12|0);
   $101 = HEAP32[$100>>2]|0;
   $102 = ($101|0)==($3|0);
   do {
    if ($102) {
     $112 = ((($3)) + 16|0);
     $113 = ((($112)) + 4|0);
     $114 = HEAP32[$113>>2]|0;
     $115 = ($114|0)==(0|0);
     if ($115) {
      $116 = HEAP32[$112>>2]|0;
      $117 = ($116|0)==(0|0);
      if ($117) {
       $R$3 = 0;
       break;
      } else {
       $R$1 = $116;$RP$1 = $112;
      }
     } else {
      $R$1 = $114;$RP$1 = $113;
     }
     while(1) {
      $118 = ((($R$1)) + 20|0);
      $119 = HEAP32[$118>>2]|0;
      $120 = ($119|0)==(0|0);
      if (!($120)) {
       $R$1 = $119;$RP$1 = $118;
       continue;
      }
      $121 = ((($R$1)) + 16|0);
      $122 = HEAP32[$121>>2]|0;
      $123 = ($122|0)==(0|0);
      if ($123) {
       $R$1$lcssa = $R$1;$RP$1$lcssa = $RP$1;
       break;
      } else {
       $R$1 = $122;$RP$1 = $121;
      }
     }
     $124 = ($RP$1$lcssa>>>0)<($4>>>0);
     if ($124) {
      _abort();
      // unreachable;
     } else {
      HEAP32[$RP$1$lcssa>>2] = 0;
      $R$3 = $R$1$lcssa;
      break;
     }
    } else {
     $103 = ((($3)) + 8|0);
     $104 = HEAP32[$103>>2]|0;
     $105 = ($104>>>0)<($4>>>0);
     if ($105) {
      _abort();
      // unreachable;
     }
     $106 = ((($104)) + 12|0);
     $107 = HEAP32[$106>>2]|0;
     $108 = ($107|0)==($3|0);
     if (!($108)) {
      _abort();
      // unreachable;
     }
     $109 = ((($101)) + 8|0);
     $110 = HEAP32[$109>>2]|0;
     $111 = ($110|0)==($3|0);
     if ($111) {
      HEAP32[$106>>2] = $101;
      HEAP32[$109>>2] = $104;
      $R$3 = $101;
      break;
     } else {
      _abort();
      // unreachable;
     }
    }
   } while(0);
   $125 = ($99|0)==(0|0);
   if (!($125)) {
    $126 = ((($3)) + 28|0);
    $127 = HEAP32[$126>>2]|0;
    $128 = (1206332 + ($127<<2)|0);
    $129 = HEAP32[$128>>2]|0;
    $130 = ($3|0)==($129|0);
    if ($130) {
     HEAP32[$128>>2] = $R$3;
     $cond = ($R$3|0)==(0|0);
     if ($cond) {
      $131 = 1 << $127;
      $132 = $131 ^ -1;
      $133 = HEAP32[(1206032)>>2]|0;
      $134 = $133 & $132;
      HEAP32[(1206032)>>2] = $134;
      break;
     }
    } else {
     $135 = HEAP32[(1206044)>>2]|0;
     $136 = ($99>>>0)<($135>>>0);
     if ($136) {
      _abort();
      // unreachable;
     }
     $137 = ((($99)) + 16|0);
     $138 = HEAP32[$137>>2]|0;
     $139 = ($138|0)==($3|0);
     if ($139) {
      HEAP32[$137>>2] = $R$3;
     } else {
      $140 = ((($99)) + 20|0);
      HEAP32[$140>>2] = $R$3;
     }
     $141 = ($R$3|0)==(0|0);
     if ($141) {
      break;
     }
    }
    $142 = HEAP32[(1206044)>>2]|0;
    $143 = ($R$3>>>0)<($142>>>0);
    if ($143) {
     _abort();
     // unreachable;
    }
    $144 = ((($R$3)) + 24|0);
    HEAP32[$144>>2] = $99;
    $145 = ((($3)) + 16|0);
    $146 = HEAP32[$145>>2]|0;
    $147 = ($146|0)==(0|0);
    do {
     if (!($147)) {
      $148 = ($146>>>0)<($142>>>0);
      if ($148) {
       _abort();
       // unreachable;
      } else {
       $149 = ((($R$3)) + 16|0);
       HEAP32[$149>>2] = $146;
       $150 = ((($146)) + 24|0);
       HEAP32[$150>>2] = $R$3;
       break;
      }
     }
    } while(0);
    $151 = ((($145)) + 4|0);
    $152 = HEAP32[$151>>2]|0;
    $153 = ($152|0)==(0|0);
    if (!($153)) {
     $154 = HEAP32[(1206044)>>2]|0;
     $155 = ($152>>>0)<($154>>>0);
     if ($155) {
      _abort();
      // unreachable;
     } else {
      $156 = ((($R$3)) + 20|0);
      HEAP32[$156>>2] = $152;
      $157 = ((($152)) + 24|0);
      HEAP32[$157>>2] = $R$3;
      break;
     }
    }
   }
  }
 } while(0);
 $158 = ($73>>>0)<(16);
 if ($158) {
  $159 = $1 & 1;
  $160 = $71 | $159;
  $161 = $160 | 2;
  HEAP32[$0>>2] = $161;
  $162 = (($p) + ($71)|0);
  $163 = ((($162)) + 4|0);
  $164 = HEAP32[$163>>2]|0;
  $165 = $164 | 1;
  HEAP32[$163>>2] = $165;
  $newp$2 = $p;
  return ($newp$2|0);
 } else {
  $166 = (($p) + ($nb)|0);
  $167 = $1 & 1;
  $168 = $167 | $nb;
  $169 = $168 | 2;
  HEAP32[$0>>2] = $169;
  $170 = ((($166)) + 4|0);
  $171 = $73 | 3;
  HEAP32[$170>>2] = $171;
  $172 = (($166) + ($73)|0);
  $173 = ((($172)) + 4|0);
  $174 = HEAP32[$173>>2]|0;
  $175 = $174 | 1;
  HEAP32[$173>>2] = $175;
  _dispose_chunk($166,$73);
  $newp$2 = $p;
  return ($newp$2|0);
 }
 return (0)|0;
}
function _dispose_chunk($p,$psize) {
 $p = $p|0;
 $psize = $psize|0;
 var $$1 = 0, $$14 = 0, $$2 = 0, $$lcssa = 0, $$pre = 0, $$pre$phi22Z2D = 0, $$pre$phi24Z2D = 0, $$pre$phiZ2D = 0, $$pre21 = 0, $$pre23 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0;
 var $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0;
 var $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0;
 var $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0;
 var $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0;
 var $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0;
 var $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $F17$0 = 0, $I20$0 = 0, $K21$0 = 0, $R$1 = 0, $R$1$lcssa = 0;
 var $R$3 = 0, $R7$1 = 0, $R7$1$lcssa = 0, $R7$3 = 0, $RP$1 = 0, $RP$1$lcssa = 0, $RP9$1 = 0, $RP9$1$lcssa = 0, $T$0 = 0, $T$0$lcssa = 0, $T$0$lcssa30 = 0, $cond = 0, $cond16 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($p) + ($psize)|0);
 $1 = ((($p)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2 & 1;
 $4 = ($3|0)==(0);
 do {
  if ($4) {
   $5 = HEAP32[$p>>2]|0;
   $6 = $2 & 3;
   $7 = ($6|0)==(0);
   if ($7) {
    return;
   }
   $8 = (0 - ($5))|0;
   $9 = (($p) + ($8)|0);
   $10 = (($5) + ($psize))|0;
   $11 = HEAP32[(1206044)>>2]|0;
   $12 = ($9>>>0)<($11>>>0);
   if ($12) {
    _abort();
    // unreachable;
   }
   $13 = HEAP32[(1206048)>>2]|0;
   $14 = ($9|0)==($13|0);
   if ($14) {
    $99 = ((($0)) + 4|0);
    $100 = HEAP32[$99>>2]|0;
    $101 = $100 & 3;
    $102 = ($101|0)==(3);
    if (!($102)) {
     $$1 = $9;$$14 = $10;
     break;
    }
    HEAP32[(1206036)>>2] = $10;
    $103 = $100 & -2;
    HEAP32[$99>>2] = $103;
    $104 = $10 | 1;
    $105 = ((($9)) + 4|0);
    HEAP32[$105>>2] = $104;
    $106 = (($9) + ($10)|0);
    HEAP32[$106>>2] = $10;
    return;
   }
   $15 = $5 >>> 3;
   $16 = ($5>>>0)<(256);
   if ($16) {
    $17 = ((($9)) + 8|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = ((($9)) + 12|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = $15 << 1;
    $22 = (1206068 + ($21<<2)|0);
    $23 = ($18|0)==($22|0);
    if (!($23)) {
     $24 = ($18>>>0)<($11>>>0);
     if ($24) {
      _abort();
      // unreachable;
     }
     $25 = ((($18)) + 12|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = ($26|0)==($9|0);
     if (!($27)) {
      _abort();
      // unreachable;
     }
    }
    $28 = ($20|0)==($18|0);
    if ($28) {
     $29 = 1 << $15;
     $30 = $29 ^ -1;
     $31 = HEAP32[301507]|0;
     $32 = $31 & $30;
     HEAP32[301507] = $32;
     $$1 = $9;$$14 = $10;
     break;
    }
    $33 = ($20|0)==($22|0);
    if ($33) {
     $$pre23 = ((($20)) + 8|0);
     $$pre$phi24Z2D = $$pre23;
    } else {
     $34 = ($20>>>0)<($11>>>0);
     if ($34) {
      _abort();
      // unreachable;
     }
     $35 = ((($20)) + 8|0);
     $36 = HEAP32[$35>>2]|0;
     $37 = ($36|0)==($9|0);
     if ($37) {
      $$pre$phi24Z2D = $35;
     } else {
      _abort();
      // unreachable;
     }
    }
    $38 = ((($18)) + 12|0);
    HEAP32[$38>>2] = $20;
    HEAP32[$$pre$phi24Z2D>>2] = $18;
    $$1 = $9;$$14 = $10;
    break;
   }
   $39 = ((($9)) + 24|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = ((($9)) + 12|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = ($42|0)==($9|0);
   do {
    if ($43) {
     $53 = ((($9)) + 16|0);
     $54 = ((($53)) + 4|0);
     $55 = HEAP32[$54>>2]|0;
     $56 = ($55|0)==(0|0);
     if ($56) {
      $57 = HEAP32[$53>>2]|0;
      $58 = ($57|0)==(0|0);
      if ($58) {
       $R$3 = 0;
       break;
      } else {
       $R$1 = $57;$RP$1 = $53;
      }
     } else {
      $R$1 = $55;$RP$1 = $54;
     }
     while(1) {
      $59 = ((($R$1)) + 20|0);
      $60 = HEAP32[$59>>2]|0;
      $61 = ($60|0)==(0|0);
      if (!($61)) {
       $R$1 = $60;$RP$1 = $59;
       continue;
      }
      $62 = ((($R$1)) + 16|0);
      $63 = HEAP32[$62>>2]|0;
      $64 = ($63|0)==(0|0);
      if ($64) {
       $R$1$lcssa = $R$1;$RP$1$lcssa = $RP$1;
       break;
      } else {
       $R$1 = $63;$RP$1 = $62;
      }
     }
     $65 = ($RP$1$lcssa>>>0)<($11>>>0);
     if ($65) {
      _abort();
      // unreachable;
     } else {
      HEAP32[$RP$1$lcssa>>2] = 0;
      $R$3 = $R$1$lcssa;
      break;
     }
    } else {
     $44 = ((($9)) + 8|0);
     $45 = HEAP32[$44>>2]|0;
     $46 = ($45>>>0)<($11>>>0);
     if ($46) {
      _abort();
      // unreachable;
     }
     $47 = ((($45)) + 12|0);
     $48 = HEAP32[$47>>2]|0;
     $49 = ($48|0)==($9|0);
     if (!($49)) {
      _abort();
      // unreachable;
     }
     $50 = ((($42)) + 8|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = ($51|0)==($9|0);
     if ($52) {
      HEAP32[$47>>2] = $42;
      HEAP32[$50>>2] = $45;
      $R$3 = $42;
      break;
     } else {
      _abort();
      // unreachable;
     }
    }
   } while(0);
   $66 = ($40|0)==(0|0);
   if ($66) {
    $$1 = $9;$$14 = $10;
   } else {
    $67 = ((($9)) + 28|0);
    $68 = HEAP32[$67>>2]|0;
    $69 = (1206332 + ($68<<2)|0);
    $70 = HEAP32[$69>>2]|0;
    $71 = ($9|0)==($70|0);
    if ($71) {
     HEAP32[$69>>2] = $R$3;
     $cond = ($R$3|0)==(0|0);
     if ($cond) {
      $72 = 1 << $68;
      $73 = $72 ^ -1;
      $74 = HEAP32[(1206032)>>2]|0;
      $75 = $74 & $73;
      HEAP32[(1206032)>>2] = $75;
      $$1 = $9;$$14 = $10;
      break;
     }
    } else {
     $76 = HEAP32[(1206044)>>2]|0;
     $77 = ($40>>>0)<($76>>>0);
     if ($77) {
      _abort();
      // unreachable;
     }
     $78 = ((($40)) + 16|0);
     $79 = HEAP32[$78>>2]|0;
     $80 = ($79|0)==($9|0);
     if ($80) {
      HEAP32[$78>>2] = $R$3;
     } else {
      $81 = ((($40)) + 20|0);
      HEAP32[$81>>2] = $R$3;
     }
     $82 = ($R$3|0)==(0|0);
     if ($82) {
      $$1 = $9;$$14 = $10;
      break;
     }
    }
    $83 = HEAP32[(1206044)>>2]|0;
    $84 = ($R$3>>>0)<($83>>>0);
    if ($84) {
     _abort();
     // unreachable;
    }
    $85 = ((($R$3)) + 24|0);
    HEAP32[$85>>2] = $40;
    $86 = ((($9)) + 16|0);
    $87 = HEAP32[$86>>2]|0;
    $88 = ($87|0)==(0|0);
    do {
     if (!($88)) {
      $89 = ($87>>>0)<($83>>>0);
      if ($89) {
       _abort();
       // unreachable;
      } else {
       $90 = ((($R$3)) + 16|0);
       HEAP32[$90>>2] = $87;
       $91 = ((($87)) + 24|0);
       HEAP32[$91>>2] = $R$3;
       break;
      }
     }
    } while(0);
    $92 = ((($86)) + 4|0);
    $93 = HEAP32[$92>>2]|0;
    $94 = ($93|0)==(0|0);
    if ($94) {
     $$1 = $9;$$14 = $10;
    } else {
     $95 = HEAP32[(1206044)>>2]|0;
     $96 = ($93>>>0)<($95>>>0);
     if ($96) {
      _abort();
      // unreachable;
     } else {
      $97 = ((($R$3)) + 20|0);
      HEAP32[$97>>2] = $93;
      $98 = ((($93)) + 24|0);
      HEAP32[$98>>2] = $R$3;
      $$1 = $9;$$14 = $10;
      break;
     }
    }
   }
  } else {
   $$1 = $p;$$14 = $psize;
  }
 } while(0);
 $107 = HEAP32[(1206044)>>2]|0;
 $108 = ($0>>>0)<($107>>>0);
 if ($108) {
  _abort();
  // unreachable;
 }
 $109 = ((($0)) + 4|0);
 $110 = HEAP32[$109>>2]|0;
 $111 = $110 & 2;
 $112 = ($111|0)==(0);
 if ($112) {
  $113 = HEAP32[(1206052)>>2]|0;
  $114 = ($0|0)==($113|0);
  if ($114) {
   $115 = HEAP32[(1206040)>>2]|0;
   $116 = (($115) + ($$14))|0;
   HEAP32[(1206040)>>2] = $116;
   HEAP32[(1206052)>>2] = $$1;
   $117 = $116 | 1;
   $118 = ((($$1)) + 4|0);
   HEAP32[$118>>2] = $117;
   $119 = HEAP32[(1206048)>>2]|0;
   $120 = ($$1|0)==($119|0);
   if (!($120)) {
    return;
   }
   HEAP32[(1206048)>>2] = 0;
   HEAP32[(1206036)>>2] = 0;
   return;
  }
  $121 = HEAP32[(1206048)>>2]|0;
  $122 = ($0|0)==($121|0);
  if ($122) {
   $123 = HEAP32[(1206036)>>2]|0;
   $124 = (($123) + ($$14))|0;
   HEAP32[(1206036)>>2] = $124;
   HEAP32[(1206048)>>2] = $$1;
   $125 = $124 | 1;
   $126 = ((($$1)) + 4|0);
   HEAP32[$126>>2] = $125;
   $127 = (($$1) + ($124)|0);
   HEAP32[$127>>2] = $124;
   return;
  }
  $128 = $110 & -8;
  $129 = (($128) + ($$14))|0;
  $130 = $110 >>> 3;
  $131 = ($110>>>0)<(256);
  do {
   if ($131) {
    $132 = ((($0)) + 8|0);
    $133 = HEAP32[$132>>2]|0;
    $134 = ((($0)) + 12|0);
    $135 = HEAP32[$134>>2]|0;
    $136 = $130 << 1;
    $137 = (1206068 + ($136<<2)|0);
    $138 = ($133|0)==($137|0);
    if (!($138)) {
     $139 = ($133>>>0)<($107>>>0);
     if ($139) {
      _abort();
      // unreachable;
     }
     $140 = ((($133)) + 12|0);
     $141 = HEAP32[$140>>2]|0;
     $142 = ($141|0)==($0|0);
     if (!($142)) {
      _abort();
      // unreachable;
     }
    }
    $143 = ($135|0)==($133|0);
    if ($143) {
     $144 = 1 << $130;
     $145 = $144 ^ -1;
     $146 = HEAP32[301507]|0;
     $147 = $146 & $145;
     HEAP32[301507] = $147;
     break;
    }
    $148 = ($135|0)==($137|0);
    if ($148) {
     $$pre21 = ((($135)) + 8|0);
     $$pre$phi22Z2D = $$pre21;
    } else {
     $149 = ($135>>>0)<($107>>>0);
     if ($149) {
      _abort();
      // unreachable;
     }
     $150 = ((($135)) + 8|0);
     $151 = HEAP32[$150>>2]|0;
     $152 = ($151|0)==($0|0);
     if ($152) {
      $$pre$phi22Z2D = $150;
     } else {
      _abort();
      // unreachable;
     }
    }
    $153 = ((($133)) + 12|0);
    HEAP32[$153>>2] = $135;
    HEAP32[$$pre$phi22Z2D>>2] = $133;
   } else {
    $154 = ((($0)) + 24|0);
    $155 = HEAP32[$154>>2]|0;
    $156 = ((($0)) + 12|0);
    $157 = HEAP32[$156>>2]|0;
    $158 = ($157|0)==($0|0);
    do {
     if ($158) {
      $168 = ((($0)) + 16|0);
      $169 = ((($168)) + 4|0);
      $170 = HEAP32[$169>>2]|0;
      $171 = ($170|0)==(0|0);
      if ($171) {
       $172 = HEAP32[$168>>2]|0;
       $173 = ($172|0)==(0|0);
       if ($173) {
        $R7$3 = 0;
        break;
       } else {
        $R7$1 = $172;$RP9$1 = $168;
       }
      } else {
       $R7$1 = $170;$RP9$1 = $169;
      }
      while(1) {
       $174 = ((($R7$1)) + 20|0);
       $175 = HEAP32[$174>>2]|0;
       $176 = ($175|0)==(0|0);
       if (!($176)) {
        $R7$1 = $175;$RP9$1 = $174;
        continue;
       }
       $177 = ((($R7$1)) + 16|0);
       $178 = HEAP32[$177>>2]|0;
       $179 = ($178|0)==(0|0);
       if ($179) {
        $R7$1$lcssa = $R7$1;$RP9$1$lcssa = $RP9$1;
        break;
       } else {
        $R7$1 = $178;$RP9$1 = $177;
       }
      }
      $180 = ($RP9$1$lcssa>>>0)<($107>>>0);
      if ($180) {
       _abort();
       // unreachable;
      } else {
       HEAP32[$RP9$1$lcssa>>2] = 0;
       $R7$3 = $R7$1$lcssa;
       break;
      }
     } else {
      $159 = ((($0)) + 8|0);
      $160 = HEAP32[$159>>2]|0;
      $161 = ($160>>>0)<($107>>>0);
      if ($161) {
       _abort();
       // unreachable;
      }
      $162 = ((($160)) + 12|0);
      $163 = HEAP32[$162>>2]|0;
      $164 = ($163|0)==($0|0);
      if (!($164)) {
       _abort();
       // unreachable;
      }
      $165 = ((($157)) + 8|0);
      $166 = HEAP32[$165>>2]|0;
      $167 = ($166|0)==($0|0);
      if ($167) {
       HEAP32[$162>>2] = $157;
       HEAP32[$165>>2] = $160;
       $R7$3 = $157;
       break;
      } else {
       _abort();
       // unreachable;
      }
     }
    } while(0);
    $181 = ($155|0)==(0|0);
    if (!($181)) {
     $182 = ((($0)) + 28|0);
     $183 = HEAP32[$182>>2]|0;
     $184 = (1206332 + ($183<<2)|0);
     $185 = HEAP32[$184>>2]|0;
     $186 = ($0|0)==($185|0);
     if ($186) {
      HEAP32[$184>>2] = $R7$3;
      $cond16 = ($R7$3|0)==(0|0);
      if ($cond16) {
       $187 = 1 << $183;
       $188 = $187 ^ -1;
       $189 = HEAP32[(1206032)>>2]|0;
       $190 = $189 & $188;
       HEAP32[(1206032)>>2] = $190;
       break;
      }
     } else {
      $191 = HEAP32[(1206044)>>2]|0;
      $192 = ($155>>>0)<($191>>>0);
      if ($192) {
       _abort();
       // unreachable;
      }
      $193 = ((($155)) + 16|0);
      $194 = HEAP32[$193>>2]|0;
      $195 = ($194|0)==($0|0);
      if ($195) {
       HEAP32[$193>>2] = $R7$3;
      } else {
       $196 = ((($155)) + 20|0);
       HEAP32[$196>>2] = $R7$3;
      }
      $197 = ($R7$3|0)==(0|0);
      if ($197) {
       break;
      }
     }
     $198 = HEAP32[(1206044)>>2]|0;
     $199 = ($R7$3>>>0)<($198>>>0);
     if ($199) {
      _abort();
      // unreachable;
     }
     $200 = ((($R7$3)) + 24|0);
     HEAP32[$200>>2] = $155;
     $201 = ((($0)) + 16|0);
     $202 = HEAP32[$201>>2]|0;
     $203 = ($202|0)==(0|0);
     do {
      if (!($203)) {
       $204 = ($202>>>0)<($198>>>0);
       if ($204) {
        _abort();
        // unreachable;
       } else {
        $205 = ((($R7$3)) + 16|0);
        HEAP32[$205>>2] = $202;
        $206 = ((($202)) + 24|0);
        HEAP32[$206>>2] = $R7$3;
        break;
       }
      }
     } while(0);
     $207 = ((($201)) + 4|0);
     $208 = HEAP32[$207>>2]|0;
     $209 = ($208|0)==(0|0);
     if (!($209)) {
      $210 = HEAP32[(1206044)>>2]|0;
      $211 = ($208>>>0)<($210>>>0);
      if ($211) {
       _abort();
       // unreachable;
      } else {
       $212 = ((($R7$3)) + 20|0);
       HEAP32[$212>>2] = $208;
       $213 = ((($208)) + 24|0);
       HEAP32[$213>>2] = $R7$3;
       break;
      }
     }
    }
   }
  } while(0);
  $214 = $129 | 1;
  $215 = ((($$1)) + 4|0);
  HEAP32[$215>>2] = $214;
  $216 = (($$1) + ($129)|0);
  HEAP32[$216>>2] = $129;
  $217 = HEAP32[(1206048)>>2]|0;
  $218 = ($$1|0)==($217|0);
  if ($218) {
   HEAP32[(1206036)>>2] = $129;
   return;
  } else {
   $$2 = $129;
  }
 } else {
  $219 = $110 & -2;
  HEAP32[$109>>2] = $219;
  $220 = $$14 | 1;
  $221 = ((($$1)) + 4|0);
  HEAP32[$221>>2] = $220;
  $222 = (($$1) + ($$14)|0);
  HEAP32[$222>>2] = $$14;
  $$2 = $$14;
 }
 $223 = $$2 >>> 3;
 $224 = ($$2>>>0)<(256);
 if ($224) {
  $225 = $223 << 1;
  $226 = (1206068 + ($225<<2)|0);
  $227 = HEAP32[301507]|0;
  $228 = 1 << $223;
  $229 = $227 & $228;
  $230 = ($229|0)==(0);
  if ($230) {
   $231 = $227 | $228;
   HEAP32[301507] = $231;
   $$pre = ((($226)) + 8|0);
   $$pre$phiZ2D = $$pre;$F17$0 = $226;
  } else {
   $232 = ((($226)) + 8|0);
   $233 = HEAP32[$232>>2]|0;
   $234 = HEAP32[(1206044)>>2]|0;
   $235 = ($233>>>0)<($234>>>0);
   if ($235) {
    _abort();
    // unreachable;
   } else {
    $$pre$phiZ2D = $232;$F17$0 = $233;
   }
  }
  HEAP32[$$pre$phiZ2D>>2] = $$1;
  $236 = ((($F17$0)) + 12|0);
  HEAP32[$236>>2] = $$1;
  $237 = ((($$1)) + 8|0);
  HEAP32[$237>>2] = $F17$0;
  $238 = ((($$1)) + 12|0);
  HEAP32[$238>>2] = $226;
  return;
 }
 $239 = $$2 >>> 8;
 $240 = ($239|0)==(0);
 if ($240) {
  $I20$0 = 0;
 } else {
  $241 = ($$2>>>0)>(16777215);
  if ($241) {
   $I20$0 = 31;
  } else {
   $242 = (($239) + 1048320)|0;
   $243 = $242 >>> 16;
   $244 = $243 & 8;
   $245 = $239 << $244;
   $246 = (($245) + 520192)|0;
   $247 = $246 >>> 16;
   $248 = $247 & 4;
   $249 = $248 | $244;
   $250 = $245 << $248;
   $251 = (($250) + 245760)|0;
   $252 = $251 >>> 16;
   $253 = $252 & 2;
   $254 = $249 | $253;
   $255 = (14 - ($254))|0;
   $256 = $250 << $253;
   $257 = $256 >>> 15;
   $258 = (($255) + ($257))|0;
   $259 = $258 << 1;
   $260 = (($258) + 7)|0;
   $261 = $$2 >>> $260;
   $262 = $261 & 1;
   $263 = $262 | $259;
   $I20$0 = $263;
  }
 }
 $264 = (1206332 + ($I20$0<<2)|0);
 $265 = ((($$1)) + 28|0);
 HEAP32[$265>>2] = $I20$0;
 $266 = ((($$1)) + 16|0);
 $267 = ((($$1)) + 20|0);
 HEAP32[$267>>2] = 0;
 HEAP32[$266>>2] = 0;
 $268 = HEAP32[(1206032)>>2]|0;
 $269 = 1 << $I20$0;
 $270 = $268 & $269;
 $271 = ($270|0)==(0);
 if ($271) {
  $272 = $268 | $269;
  HEAP32[(1206032)>>2] = $272;
  HEAP32[$264>>2] = $$1;
  $273 = ((($$1)) + 24|0);
  HEAP32[$273>>2] = $264;
  $274 = ((($$1)) + 12|0);
  HEAP32[$274>>2] = $$1;
  $275 = ((($$1)) + 8|0);
  HEAP32[$275>>2] = $$1;
  return;
 }
 $276 = HEAP32[$264>>2]|0;
 $277 = ($I20$0|0)==(31);
 $278 = $I20$0 >>> 1;
 $279 = (25 - ($278))|0;
 $280 = $277 ? 0 : $279;
 $281 = $$2 << $280;
 $K21$0 = $281;$T$0 = $276;
 while(1) {
  $282 = ((($T$0)) + 4|0);
  $283 = HEAP32[$282>>2]|0;
  $284 = $283 & -8;
  $285 = ($284|0)==($$2|0);
  if ($285) {
   $T$0$lcssa = $T$0;
   label = 127;
   break;
  }
  $286 = $K21$0 >>> 31;
  $287 = (((($T$0)) + 16|0) + ($286<<2)|0);
  $288 = $K21$0 << 1;
  $289 = HEAP32[$287>>2]|0;
  $290 = ($289|0)==(0|0);
  if ($290) {
   $$lcssa = $287;$T$0$lcssa30 = $T$0;
   label = 124;
   break;
  } else {
   $K21$0 = $288;$T$0 = $289;
  }
 }
 if ((label|0) == 124) {
  $291 = HEAP32[(1206044)>>2]|0;
  $292 = ($$lcssa>>>0)<($291>>>0);
  if ($292) {
   _abort();
   // unreachable;
  }
  HEAP32[$$lcssa>>2] = $$1;
  $293 = ((($$1)) + 24|0);
  HEAP32[$293>>2] = $T$0$lcssa30;
  $294 = ((($$1)) + 12|0);
  HEAP32[$294>>2] = $$1;
  $295 = ((($$1)) + 8|0);
  HEAP32[$295>>2] = $$1;
  return;
 }
 else if ((label|0) == 127) {
  $296 = ((($T$0$lcssa)) + 8|0);
  $297 = HEAP32[$296>>2]|0;
  $298 = HEAP32[(1206044)>>2]|0;
  $299 = ($297>>>0)>=($298>>>0);
  $not$ = ($T$0$lcssa>>>0)>=($298>>>0);
  $300 = $299 & $not$;
  if (!($300)) {
   _abort();
   // unreachable;
  }
  $301 = ((($297)) + 12|0);
  HEAP32[$301>>2] = $$1;
  HEAP32[$296>>2] = $$1;
  $302 = ((($$1)) + 8|0);
  HEAP32[$302>>2] = $297;
  $303 = ((($$1)) + 12|0);
  HEAP32[$303>>2] = $T$0$lcssa;
  $304 = ((($$1)) + 24|0);
  HEAP32[$304>>2] = 0;
  return;
 }
}
function __GLOBAL__I_000101() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__18ios_base4InitC2Ev(0);
 (___cxa_atexit((4852|0),(1282465|0),(___dso_handle|0))|0);
 return;
}
function __ZNSt3__18ios_base4InitC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[37573]|0;
 __ZNSt3__110__stdinbufIcEC2EP8_IO_FILEP11__mbstate_t(1207204,$0,1207260);
 HEAP32[301631] = (154748);
 HEAP32[(1206532)>>2] = (154768);
 HEAP32[(1206528)>>2] = 0;
 __THREW__ = 0;
 invoke_vii(3362,((1206532)|0),(1207204|0));
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev((1206532));
  ___resumeException($3|0);
  // unreachable;
 }
 HEAP32[(1206604)>>2] = 0;
 HEAP32[(1206608)>>2] = -1;
 $5 = HEAP32[37602]|0;
 __ZNSt3__111__stdoutbufIcEC2EP8_IO_FILEP11__mbstate_t(1207308,$5,(1207268));
 HEAP32[301653] = (154828);
 HEAP32[(1206616)>>2] = (154848);
 __THREW__ = 0;
 invoke_vii(3362,((1206616)|0),(1207308|0));
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $8 = ___cxa_find_matching_catch_2()|0;
  $9 = tempRet0;
  __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev((1206616));
  ___resumeException($8|0);
  // unreachable;
 }
 HEAP32[(1206688)>>2] = 0;
 HEAP32[(1206692)>>2] = -1;
 $10 = HEAP32[37544]|0;
 __ZNSt3__111__stdoutbufIcEC2EP8_IO_FILEP11__mbstate_t(1207356,$10,(1207276));
 HEAP32[301674] = (154828);
 HEAP32[(1206700)>>2] = (154848);
 __THREW__ = 0;
 invoke_vii(3362,((1206700)|0),(1207356|0));
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev((1206700));
  ___resumeException($13|0);
  // unreachable;
 }
 HEAP32[(1206772)>>2] = 0;
 HEAP32[(1206776)>>2] = -1;
 $15 = HEAP32[301674]|0;
 $16 = ((($15)) + -12|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = (1206696 + ($17)|0);
 $19 = ((($18)) + 24|0);
 $20 = HEAP32[$19>>2]|0;
 HEAP32[301695] = (154828);
 HEAP32[(1206784)>>2] = (154848);
 __THREW__ = 0;
 invoke_vii(3362,((1206784)|0),($20|0));
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $23 = ___cxa_find_matching_catch_2()|0;
  $24 = tempRet0;
  __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev((1206784));
  ___resumeException($23|0);
  // unreachable;
 }
 HEAP32[(1206856)>>2] = 0;
 HEAP32[(1206860)>>2] = -1;
 $25 = HEAP32[301631]|0;
 $26 = ((($25)) + -12|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = (1206524 + ($27)|0);
 $29 = ((($28)) + 72|0);
 HEAP32[$29>>2] = 1206612;
 $30 = HEAP32[301674]|0;
 $31 = ((($30)) + -12|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = (1206696 + ($32)|0);
 $34 = ((($33)) + 4|0);
 $35 = HEAP32[$34>>2]|0;
 $36 = $35 | 8192;
 HEAP32[$34>>2] = $36;
 $37 = HEAP32[301674]|0;
 $38 = ((($37)) + -12|0);
 $39 = HEAP32[$38>>2]|0;
 $40 = (1206696 + ($39)|0);
 $41 = ((($40)) + 72|0);
 HEAP32[$41>>2] = 1206612;
 __ZNSt3__110__stdinbufIwEC2EP8_IO_FILEP11__mbstate_t(1207404,$0,(1207284));
 HEAP32[301716] = (154788);
 HEAP32[(1206872)>>2] = (154808);
 HEAP32[(1206868)>>2] = 0;
 __THREW__ = 0;
 invoke_vii(3362,((1206872)|0),(1207404|0));
 $42 = __THREW__; __THREW__ = 0;
 $43 = $42&1;
 if ($43) {
  $44 = ___cxa_find_matching_catch_2()|0;
  $45 = tempRet0;
  __ZNSt3__19basic_iosIwNS_11char_traitsIwEEED2Ev((1206872));
  ___resumeException($44|0);
  // unreachable;
 }
 HEAP32[(1206944)>>2] = 0;
 HEAP32[(1206948)>>2] = -1;
 __ZNSt3__111__stdoutbufIwEC2EP8_IO_FILEP11__mbstate_t(1207460,$5,(1207292));
 HEAP32[301738] = (154868);
 HEAP32[(1206956)>>2] = (154888);
 __THREW__ = 0;
 invoke_vii(3362,((1206956)|0),(1207460|0));
 $46 = __THREW__; __THREW__ = 0;
 $47 = $46&1;
 if ($47) {
  $48 = ___cxa_find_matching_catch_2()|0;
  $49 = tempRet0;
  __ZNSt3__19basic_iosIwNS_11char_traitsIwEEED2Ev((1206956));
  ___resumeException($48|0);
  // unreachable;
 }
 HEAP32[(1207028)>>2] = 0;
 HEAP32[(1207032)>>2] = -1;
 __ZNSt3__111__stdoutbufIwEC2EP8_IO_FILEP11__mbstate_t(1207508,$10,(1207300));
 HEAP32[301759] = (154868);
 HEAP32[(1207040)>>2] = (154888);
 __THREW__ = 0;
 invoke_vii(3362,((1207040)|0),(1207508|0));
 $50 = __THREW__; __THREW__ = 0;
 $51 = $50&1;
 if ($51) {
  $52 = ___cxa_find_matching_catch_2()|0;
  $53 = tempRet0;
  __ZNSt3__19basic_iosIwNS_11char_traitsIwEEED2Ev((1207040));
  ___resumeException($52|0);
  // unreachable;
 }
 HEAP32[(1207112)>>2] = 0;
 HEAP32[(1207116)>>2] = -1;
 $54 = HEAP32[301759]|0;
 $55 = ((($54)) + -12|0);
 $56 = HEAP32[$55>>2]|0;
 $57 = (1207036 + ($56)|0);
 $58 = ((($57)) + 24|0);
 $59 = HEAP32[$58>>2]|0;
 HEAP32[301780] = (154868);
 HEAP32[(1207124)>>2] = (154888);
 __THREW__ = 0;
 invoke_vii(3362,((1207124)|0),($59|0));
 $60 = __THREW__; __THREW__ = 0;
 $61 = $60&1;
 if ($61) {
  $62 = ___cxa_find_matching_catch_2()|0;
  $63 = tempRet0;
  __ZNSt3__19basic_iosIwNS_11char_traitsIwEEED2Ev((1207124));
  ___resumeException($62|0);
  // unreachable;
 } else {
  HEAP32[(1207196)>>2] = 0;
  HEAP32[(1207200)>>2] = -1;
  $64 = HEAP32[301716]|0;
  $65 = ((($64)) + -12|0);
  $66 = HEAP32[$65>>2]|0;
  $67 = (1206864 + ($66)|0);
  $68 = ((($67)) + 72|0);
  HEAP32[$68>>2] = 1206952;
  $69 = HEAP32[301759]|0;
  $70 = ((($69)) + -12|0);
  $71 = HEAP32[$70>>2]|0;
  $72 = (1207036 + ($71)|0);
  $73 = ((($72)) + 4|0);
  $74 = HEAP32[$73>>2]|0;
  $75 = $74 | 8192;
  HEAP32[$73>>2] = $75;
  $76 = HEAP32[301759]|0;
  $77 = ((($76)) + -12|0);
  $78 = HEAP32[$77>>2]|0;
  $79 = (1207036 + ($78)|0);
  $80 = ((($79)) + 72|0);
  HEAP32[$80>>2] = 1206952;
  return;
 }
}
function __ZNSt3__110__stdinbufIcEC2EP8_IO_FILEP11__mbstate_t($this,$__fp,$__st) {
 $this = $this|0;
 $__fp = $__fp|0;
 $__st = $__st|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 4|0;
 $1 = sp;
 __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEEC2Ev($this);
 HEAP32[$this>>2] = (153960);
 $2 = ((($this)) + 32|0);
 HEAP32[$2>>2] = $__fp;
 $3 = ((($this)) + 40|0);
 HEAP32[$3>>2] = $__st;
 $4 = ((($this)) + 48|0);
 HEAP32[$4>>2] = -1;
 $5 = ((($this)) + 52|0);
 HEAP8[$5>>0] = 0;
 $6 = ((($this)) + 4|0);
 __ZNSt3__16localeC2ERKS0_($0,$6);
 $7 = HEAP32[$0>>2]|0;
 HEAP32[$1>>2] = $7;
 __THREW__ = 0;
 invoke_vii(3010,($this|0),($1|0));
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  __ZNSt3__16localeD2Ev($1);
  __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($this);
  ___resumeException($10|0);
  // unreachable;
 } else {
  __ZNSt3__16localeD2Ev($1);
  STACKTOP = sp;return;
 }
}
function __ZNSt3__110__stdinbufIcED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNSt3__110__stdinbufIcE5imbueERKNS_6localeE($this,$__loc) {
 $this = $this|0;
 $__loc = $__loc|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1209368)|0);
 $1 = ((($this)) + 36|0);
 HEAP32[$1>>2] = $0;
 $2 = HEAP32[$0>>2]|0;
 $3 = ((($2)) + 24|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($0)|0);
 $6 = ((($this)) + 44|0);
 HEAP32[$6>>2] = $5;
 $7 = HEAP32[$1>>2]|0;
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($8)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (FUNCTION_TABLE_ii[$10 & 8191]($7)|0);
 $12 = ((($this)) + 53|0);
 $13 = $11&1;
 HEAP8[$12>>0] = $13;
 $14 = HEAP32[$6>>2]|0;
 $15 = ($14|0)>(8);
 if ($15) {
  __ZNSt3__121__throw_runtime_errorEPKc(619609);
 }
 return;
}
function __ZNSt3__110__stdinbufIcE9underflowEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__110__stdinbufIcE9__getcharEb($this,0)|0);
 return ($0|0);
}
function __ZNSt3__110__stdinbufIcE9__getcharEb($this,$__consume) {
 $this = $this|0;
 $__consume = $__consume|0;
 var $$10 = 0, $$11 = 0, $$9 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $7 = 0, $8 = 0, $9 = 0, $__1buf = 0, $__enxt = 0, $__extbuf = 0, $__i$0 = 0, $__i6$0 = 0, $__inxt = 0, $__nread$0 = 0, $__nread$0$lcssa42 = 0, $__nread$044 = 0, $__nread$3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__extbuf = sp + 16|0;
 $__1buf = sp + 8|0;
 $__enxt = sp + 4|0;
 $__inxt = sp;
 $0 = ((($this)) + 52|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = ($1<<24>>24)==(0);
 if ($2) {
  $5 = ((($this)) + 44|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6|0)>(1);
  $8 = $7 ? $6 : 1;
  $9 = ((($this)) + 32|0);
  $__i$0 = 0;
  while(1) {
   $10 = ($__i$0|0)<($8|0);
   if (!($10)) {
    label = 8;
    break;
   }
   $11 = HEAP32[$9>>2]|0;
   $12 = (_getc($11)|0);
   $13 = ($12|0)==(-1);
   if ($13) {
    $$10 = -1;
    break;
   }
   $14 = $12&255;
   $15 = (($__extbuf) + ($__i$0)|0);
   HEAP8[$15>>0] = $14;
   $16 = (($__i$0) + 1)|0;
   $__i$0 = $16;
  }
  if ((label|0) == 8) {
   $17 = ((($this)) + 53|0);
   $18 = HEAP8[$17>>0]|0;
   $19 = ($18<<24>>24)==(0);
   do {
    if ($19) {
     $21 = ((($this)) + 40|0);
     $22 = ((($this)) + 36|0);
     $23 = ((($__1buf)) + 1|0);
     $__nread$0 = $8;
     L11: while(1) {
      $24 = HEAP32[$21>>2]|0;
      $25 = $24;
      $26 = $25;
      $27 = HEAP32[$26>>2]|0;
      $28 = (($25) + 4)|0;
      $29 = $28;
      $30 = HEAP32[$29>>2]|0;
      $31 = HEAP32[$22>>2]|0;
      $32 = (($__extbuf) + ($__nread$0)|0);
      $33 = HEAP32[$31>>2]|0;
      $34 = ((($33)) + 16|0);
      $35 = HEAP32[$34>>2]|0;
      $36 = (FUNCTION_TABLE_iiiiiiiii[$35 & 8191]($31,$24,$__extbuf,$32,$__enxt,$__1buf,$23,$__inxt)|0);
      switch ($36|0) {
      case 3:  {
       $__nread$0$lcssa42 = $__nread$0;
       label = 15;
       break L11;
       break;
      }
      case 2:  {
       label = 17;
       break L11;
       break;
      }
      case 1:  {
       break;
      }
      default: {
       $__nread$044 = $__nread$0;
       break L11;
      }
      }
      $37 = HEAP32[$21>>2]|0;
      $38 = $37;
      $39 = $38;
      HEAP32[$39>>2] = $27;
      $40 = (($38) + 4)|0;
      $41 = $40;
      HEAP32[$41>>2] = $30;
      $42 = ($__nread$0|0)==(8);
      if ($42) {
       label = 17;
       break;
      }
      $43 = HEAP32[$9>>2]|0;
      $44 = (_getc($43)|0);
      $45 = ($44|0)==(-1);
      if ($45) {
       label = 17;
       break;
      }
      $46 = $44&255;
      HEAP8[$32>>0] = $46;
      $47 = (($__nread$0) + 1)|0;
      $__nread$0 = $47;
     }
     if ((label|0) == 15) {
      $48 = HEAP8[$__extbuf>>0]|0;
      HEAP8[$__1buf>>0] = $48;
      $__nread$044 = $__nread$0$lcssa42;
     }
     else if ((label|0) == 17) {
      $$9 = -1;
      break;
     }
     $__nread$3 = $__nread$044;
     label = 18;
    } else {
     $20 = HEAP8[$__extbuf>>0]|0;
     HEAP8[$__1buf>>0] = $20;
     $__nread$3 = $8;
     label = 18;
    }
   } while(0);
   L21: do {
    if ((label|0) == 18) {
     if ($__consume) {
      $57 = HEAP8[$__1buf>>0]|0;
      $58 = $57&255;
      $59 = ((($this)) + 48|0);
      HEAP32[$59>>2] = $58;
      $61 = $57;
     } else {
      $__i6$0 = $__nread$3;
      while(1) {
       $49 = ($__i6$0|0)>(0);
       if (!($49)) {
        break;
       }
       $50 = (($__i6$0) + -1)|0;
       $51 = (($__extbuf) + ($50)|0);
       $52 = HEAP8[$51>>0]|0;
       $53 = $52&255;
       $54 = HEAP32[$9>>2]|0;
       $55 = (_ungetc($53,$54)|0);
       $56 = ($55|0)==(-1);
       if ($56) {
        $$9 = -1;
        break L21;
       } else {
        $__i6$0 = $50;
       }
      }
      $$pre = HEAP8[$__1buf>>0]|0;
      $61 = $$pre;
     }
     $60 = $61&255;
     $$9 = $60;
    }
   } while(0);
   $$10 = $$9;
  }
  $$11 = $$10;
 } else {
  $3 = ((($this)) + 48|0);
  $4 = HEAP32[$3>>2]|0;
  if ($__consume) {
   HEAP32[$3>>2] = -1;
   HEAP8[$0>>0] = 0;
   $$11 = $4;
  } else {
   $$11 = $4;
  }
 }
 STACKTOP = sp;return ($$11|0);
}
function __ZNSt3__110__stdinbufIcE5uflowEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__110__stdinbufIcE9__getcharEb($this,1)|0);
 return ($0|0);
}
function __ZNSt3__110__stdinbufIcE9pbackfailEi($this,$__c) {
 $this = $this|0;
 $__c = $__c|0;
 var $$1 = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__ci = 0, $__enxt = 0, $__extbuf = 0, $__inxt = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__extbuf = sp + 16|0;
 $__enxt = sp + 4|0;
 $__ci = sp + 8|0;
 $__inxt = sp;
 $0 = ($__c|0)==(-1);
 $1 = ((($this)) + 52|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = ($2<<24>>24)!=(0);
 L1: do {
  if ($0) {
   if ($3) {
    $$1 = -1;
   } else {
    $4 = ((($this)) + 48|0);
    $5 = HEAP32[$4>>2]|0;
    $6 = ($5|0)!=(-1);
    $7 = $6&1;
    HEAP8[$1>>0] = $7;
    $$1 = $5;
   }
  } else {
   do {
    if ($3) {
     $8 = ((($this)) + 48|0);
     $9 = HEAP32[$8>>2]|0;
     $10 = $9&255;
     HEAP8[$__ci>>0] = $10;
     $11 = ((($this)) + 36|0);
     $12 = HEAP32[$11>>2]|0;
     $13 = ((($this)) + 40|0);
     $14 = HEAP32[$13>>2]|0;
     $15 = ((($__ci)) + 1|0);
     $16 = ((($__extbuf)) + 8|0);
     $17 = HEAP32[$12>>2]|0;
     $18 = ((($17)) + 12|0);
     $19 = HEAP32[$18>>2]|0;
     $20 = (FUNCTION_TABLE_iiiiiiiii[$19 & 8191]($12,$14,$__ci,$15,$__inxt,$__extbuf,$16,$__enxt)|0);
     switch ($20|0) {
     case 1: case 2:  {
      $$1 = -1;
      break L1;
      break;
     }
     case 3:  {
      $21 = HEAP32[$8>>2]|0;
      $22 = $21&255;
      HEAP8[$__extbuf>>0] = $22;
      $23 = ((($__extbuf)) + 1|0);
      HEAP32[$__enxt>>2] = $23;
      break;
     }
     default: {
     }
     }
     $24 = ((($this)) + 32|0);
     while(1) {
      $25 = HEAP32[$__enxt>>2]|0;
      $26 = ($25>>>0)>($__extbuf>>>0);
      if (!($26)) {
       label = 13;
       break;
      }
      $27 = ((($25)) + -1|0);
      HEAP32[$__enxt>>2] = $27;
      $28 = HEAP8[$27>>0]|0;
      $29 = $28 << 24 >> 24;
      $30 = HEAP32[$24>>2]|0;
      $31 = (_ungetc($29,$30)|0);
      $32 = ($31|0)==(-1);
      if ($32) {
       label = 12;
       break;
      }
     }
     if ((label|0) == 12) {
      $$1 = -1;
      break L1;
     }
     else if ((label|0) == 13) {
      $$pre$phiZ2D = $8;
      break;
     }
    } else {
     $$pre = ((($this)) + 48|0);
     $$pre$phiZ2D = $$pre;
    }
   } while(0);
   HEAP32[$$pre$phiZ2D>>2] = $__c;
   HEAP8[$1>>0] = 1;
   $$1 = $__c;
  }
 } while(0);
 STACKTOP = sp;return ($$1|0);
}
function __ZNSt3__111__stdoutbufIcEC2EP8_IO_FILEP11__mbstate_t($this,$__fp,$__st) {
 $this = $this|0;
 $__fp = $__fp|0;
 $__st = $__st|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 4|0;
 $1 = sp;
 __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEEC2Ev($this);
 HEAP32[$this>>2] = (154024);
 $2 = ((($this)) + 32|0);
 HEAP32[$2>>2] = $__fp;
 $3 = ((($this)) + 4|0);
 __ZNSt3__16localeC2ERKS0_($0,$3);
 $4 = HEAP32[$0>>2]|0;
 HEAP32[$1>>2] = $4;
 __THREW__ = 0;
 $5 = (invoke_iii(3374,($1|0),(1209368|0))|0);
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  __ZNSt3__16localeD2Ev($1);
  __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($this);
  ___resumeException($16|0);
  // unreachable;
 } else {
  $8 = ((($this)) + 36|0);
  __ZNSt3__16localeD2Ev($1);
  HEAP32[$8>>2] = $5;
  $9 = ((($this)) + 40|0);
  HEAP32[$9>>2] = $__st;
  $10 = ((($this)) + 44|0);
  $11 = HEAP32[$5>>2]|0;
  $12 = ((($11)) + 28|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = (FUNCTION_TABLE_ii[$13 & 8191]($5)|0);
  $15 = $14&1;
  HEAP8[$10>>0] = $15;
  STACKTOP = sp;return;
 }
}
function __ZNSt3__111__stdoutbufIcED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNSt3__111__stdoutbufIcE5imbueERKNS_6localeE($this,$__loc) {
 $this = $this|0;
 $__loc = $__loc|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + 24|0);
 $2 = HEAP32[$1>>2]|0;
 (FUNCTION_TABLE_ii[$2 & 8191]($this)|0);
 $3 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1209368)|0);
 $4 = ((($this)) + 36|0);
 HEAP32[$4>>2] = $3;
 $5 = HEAP32[$3>>2]|0;
 $6 = ((($5)) + 28|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (FUNCTION_TABLE_ii[$7 & 8191]($3)|0);
 $9 = ((($this)) + 44|0);
 $10 = $8&1;
 HEAP8[$9>>0] = $10;
 return;
}
function __ZNSt3__111__stdoutbufIcE4syncEv($this) {
 $this = $this|0;
 var $$2 = 0, $$3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $__extbe = 0, $__extbuf = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__extbuf = sp + 8|0;
 $__extbe = sp;
 $0 = ((($this)) + 36|0);
 $1 = ((($this)) + 40|0);
 $2 = ((($__extbuf)) + 8|0);
 $3 = $__extbuf;
 $4 = ((($this)) + 32|0);
 L1: while(1) {
  $5 = HEAP32[$0>>2]|0;
  $6 = HEAP32[$1>>2]|0;
  $7 = HEAP32[$5>>2]|0;
  $8 = ((($7)) + 20|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = (FUNCTION_TABLE_iiiiii[$9 & 8191]($5,$6,$__extbuf,$2,$__extbe)|0);
  $11 = HEAP32[$__extbe>>2]|0;
  $12 = (($11) - ($3))|0;
  $13 = HEAP32[$4>>2]|0;
  $14 = (_fwrite($__extbuf,1,$12,$13)|0);
  $15 = ($14|0)==($12|0);
  if (!($15)) {
   $$2 = -1;
   break;
  }
  switch ($10|0) {
  case 1:  {
   break;
  }
  case 2:  {
   $$2 = -1;
   break L1;
   break;
  }
  default: {
   label = 4;
   break L1;
  }
  }
 }
 if ((label|0) == 4) {
  $16 = HEAP32[$4>>2]|0;
  $17 = (_fflush($16)|0);
  $not$ = ($17|0)!=(0);
  $$3 = $not$ << 31 >> 31;
  $$2 = $$3;
 }
 STACKTOP = sp;return ($$2|0);
}
function __ZNSt3__111__stdoutbufIcE6xsputnEPKci($this,$__s,$__n) {
 $this = $this|0;
 $__s = $__s|0;
 $__n = $__n|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__i$0 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 44|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = ($1<<24>>24)==(0);
 L1: do {
  if ($2) {
   $$01 = $__s;$__i$0 = 0;
   while(1) {
    $6 = ($__i$0|0)<($__n|0);
    if (!($6)) {
     $$0 = $__i$0;
     break L1;
    }
    $7 = HEAP32[$this>>2]|0;
    $8 = ((($7)) + 52|0);
    $9 = HEAP32[$8>>2]|0;
    $10 = HEAP8[$$01>>0]|0;
    $11 = $10&255;
    $12 = (FUNCTION_TABLE_iii[$9 & 8191]($this,$11)|0);
    $13 = ($12|0)==(-1);
    if ($13) {
     $$0 = $__i$0;
     break L1;
    }
    $14 = (($__i$0) + 1)|0;
    $15 = ((($$01)) + 1|0);
    $$01 = $15;$__i$0 = $14;
   }
  } else {
   $3 = ((($this)) + 32|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = (_fwrite($__s,1,$__n,$4)|0);
   $$0 = $5;
  }
 } while(0);
 return ($$0|0);
}
function __ZNSt3__111__stdoutbufIcE8overflowEi($this,$__c) {
 $this = $this|0;
 $__c = $__c|0;
 var $$4 = 0, $$__c$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__1buf = 0, $__e = 0;
 var $__extbe = 0, $__extbuf = 0, $pbase$0 = 0, $pbase$0$lcssa25 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__extbuf = sp + 16|0;
 $__1buf = sp + 8|0;
 $__extbe = sp + 4|0;
 $__e = sp;
 $0 = ($__c|0)==(-1);
 do {
  if ($0) {
   label = 14;
  } else {
   $1 = $__c&255;
   HEAP8[$__1buf>>0] = $1;
   $2 = ((($this)) + 44|0);
   $3 = HEAP8[$2>>0]|0;
   $4 = ($3<<24>>24)==(0);
   if (!($4)) {
    $5 = ((($this)) + 32|0);
    $6 = HEAP32[$5>>2]|0;
    $7 = (_fwrite($__1buf,1,1,$6)|0);
    $8 = ($7|0)==(1);
    if ($8) {
     label = 14;
     break;
    } else {
     $$4 = -1;
     break;
    }
   }
   HEAP32[$__extbe>>2] = $__extbuf;
   $9 = ((($__1buf)) + 1|0);
   $10 = ((($this)) + 36|0);
   $11 = ((($this)) + 40|0);
   $12 = ((($__extbuf)) + 8|0);
   $13 = $__extbuf;
   $14 = ((($this)) + 32|0);
   $pbase$0 = $__1buf;
   while(1) {
    $15 = HEAP32[$10>>2]|0;
    $16 = HEAP32[$11>>2]|0;
    $17 = HEAP32[$15>>2]|0;
    $18 = ((($17)) + 12|0);
    $19 = HEAP32[$18>>2]|0;
    $20 = (FUNCTION_TABLE_iiiiiiiii[$19 & 8191]($15,$16,$pbase$0,$9,$__e,$__extbuf,$12,$__extbe)|0);
    $21 = HEAP32[$__e>>2]|0;
    $22 = ($21|0)==($pbase$0|0);
    if ($22) {
     label = 13;
     break;
    }
    $23 = ($20|0)==(3);
    if ($23) {
     $pbase$0$lcssa25 = $pbase$0;
     label = 7;
     break;
    }
    $27 = ($20|0)==(1);
    $28 = $20 | 1;
    $29 = ($28|0)==(1);
    if (!($29)) {
     label = 13;
     break;
    }
    $30 = HEAP32[$__extbe>>2]|0;
    $31 = (($30) - ($13))|0;
    $32 = HEAP32[$14>>2]|0;
    $33 = (_fwrite($__extbuf,1,$31,$32)|0);
    $34 = ($33|0)==($31|0);
    if (!($34)) {
     label = 13;
     break;
    }
    $35 = HEAP32[$__e>>2]|0;
    if ($27) {
     $pbase$0 = $35;
    } else {
     label = 12;
     break;
    }
   }
   if ((label|0) == 7) {
    $24 = HEAP32[$14>>2]|0;
    $25 = (_fwrite($pbase$0$lcssa25,1,1,$24)|0);
    $26 = ($25|0)==(1);
    if ($26) {
     label = 12;
    } else {
     label = 13;
    }
   }
   if ((label|0) == 12) {
    label = 14;
    break;
   }
   else if ((label|0) == 13) {
    $$4 = -1;
    break;
   }
  }
 } while(0);
 if ((label|0) == 14) {
  $$__c$i = $0 ? 0 : $__c;
  $$4 = $$__c$i;
 }
 STACKTOP = sp;return ($$4|0);
}
function __ZNSt3__110__stdinbufIwEC2EP8_IO_FILEP11__mbstate_t($this,$__fp,$__st) {
 $this = $this|0;
 $__fp = $__fp|0;
 $__st = $__st|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 4|0;
 $1 = sp;
 __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEEC2Ev($this);
 HEAP32[$this>>2] = (154088);
 $2 = ((($this)) + 32|0);
 HEAP32[$2>>2] = $__fp;
 $3 = ((($this)) + 40|0);
 HEAP32[$3>>2] = $__st;
 $4 = ((($this)) + 48|0);
 HEAP32[$4>>2] = -1;
 $5 = ((($this)) + 52|0);
 HEAP8[$5>>0] = 0;
 $6 = ((($this)) + 4|0);
 __ZNSt3__16localeC2ERKS0_($0,$6);
 $7 = HEAP32[$0>>2]|0;
 HEAP32[$1>>2] = $7;
 __THREW__ = 0;
 invoke_vii(3026,($this|0),($1|0));
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  __ZNSt3__16localeD2Ev($1);
  __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEED2Ev($this);
  ___resumeException($10|0);
  // unreachable;
 } else {
  __ZNSt3__16localeD2Ev($1);
  STACKTOP = sp;return;
 }
}
function __ZNSt3__110__stdinbufIwED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEED2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNSt3__110__stdinbufIwE5imbueERKNS_6localeE($this,$__loc) {
 $this = $this|0;
 $__loc = $__loc|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1209376)|0);
 $1 = ((($this)) + 36|0);
 HEAP32[$1>>2] = $0;
 $2 = HEAP32[$0>>2]|0;
 $3 = ((($2)) + 24|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($0)|0);
 $6 = ((($this)) + 44|0);
 HEAP32[$6>>2] = $5;
 $7 = HEAP32[$1>>2]|0;
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($8)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (FUNCTION_TABLE_ii[$10 & 8191]($7)|0);
 $12 = ((($this)) + 53|0);
 $13 = $11&1;
 HEAP8[$12>>0] = $13;
 $14 = HEAP32[$6>>2]|0;
 $15 = ($14|0)>(8);
 if ($15) {
  __ZNSt3__121__throw_runtime_errorEPKc(619609);
 }
 return;
}
function __ZNSt3__110__stdinbufIwE9underflowEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__110__stdinbufIwE9__getcharEb($this,0)|0);
 return ($0|0);
}
function __ZNSt3__110__stdinbufIwE9__getcharEb($this,$__consume) {
 $this = $this|0;
 $__consume = $__consume|0;
 var $$10 = 0, $$11 = 0, $$9 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $7 = 0, $8 = 0, $9 = 0, $__1buf = 0, $__enxt = 0, $__extbuf = 0, $__i$0 = 0, $__i6$0 = 0, $__inxt = 0, $__nread$0 = 0, $__nread$0$lcssa42 = 0, $__nread$044 = 0, $__nread$3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__extbuf = sp + 16|0;
 $__1buf = sp + 8|0;
 $__enxt = sp + 4|0;
 $__inxt = sp;
 $0 = ((($this)) + 52|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = ($1<<24>>24)==(0);
 if ($2) {
  $5 = ((($this)) + 44|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6|0)>(1);
  $8 = $7 ? $6 : 1;
  $9 = ((($this)) + 32|0);
  $__i$0 = 0;
  while(1) {
   $10 = ($__i$0|0)<($8|0);
   if (!($10)) {
    label = 8;
    break;
   }
   $11 = HEAP32[$9>>2]|0;
   $12 = (_getc($11)|0);
   $13 = ($12|0)==(-1);
   if ($13) {
    $$10 = -1;
    break;
   }
   $14 = $12&255;
   $15 = (($__extbuf) + ($__i$0)|0);
   HEAP8[$15>>0] = $14;
   $16 = (($__i$0) + 1)|0;
   $__i$0 = $16;
  }
  if ((label|0) == 8) {
   $17 = ((($this)) + 53|0);
   $18 = HEAP8[$17>>0]|0;
   $19 = ($18<<24>>24)==(0);
   do {
    if ($19) {
     $22 = ((($this)) + 40|0);
     $23 = ((($this)) + 36|0);
     $24 = ((($__1buf)) + 4|0);
     $__nread$0 = $8;
     L11: while(1) {
      $25 = HEAP32[$22>>2]|0;
      $26 = $25;
      $27 = $26;
      $28 = HEAP32[$27>>2]|0;
      $29 = (($26) + 4)|0;
      $30 = $29;
      $31 = HEAP32[$30>>2]|0;
      $32 = HEAP32[$23>>2]|0;
      $33 = (($__extbuf) + ($__nread$0)|0);
      $34 = HEAP32[$32>>2]|0;
      $35 = ((($34)) + 16|0);
      $36 = HEAP32[$35>>2]|0;
      $37 = (FUNCTION_TABLE_iiiiiiiii[$36 & 8191]($32,$25,$__extbuf,$33,$__enxt,$__1buf,$24,$__inxt)|0);
      switch ($37|0) {
      case 3:  {
       $__nread$0$lcssa42 = $__nread$0;
       label = 15;
       break L11;
       break;
      }
      case 2:  {
       label = 17;
       break L11;
       break;
      }
      case 1:  {
       break;
      }
      default: {
       $__nread$044 = $__nread$0;
       break L11;
      }
      }
      $38 = HEAP32[$22>>2]|0;
      $39 = $38;
      $40 = $39;
      HEAP32[$40>>2] = $28;
      $41 = (($39) + 4)|0;
      $42 = $41;
      HEAP32[$42>>2] = $31;
      $43 = ($__nread$0|0)==(8);
      if ($43) {
       label = 17;
       break;
      }
      $44 = HEAP32[$9>>2]|0;
      $45 = (_getc($44)|0);
      $46 = ($45|0)==(-1);
      if ($46) {
       label = 17;
       break;
      }
      $47 = $45&255;
      HEAP8[$33>>0] = $47;
      $48 = (($__nread$0) + 1)|0;
      $__nread$0 = $48;
     }
     if ((label|0) == 15) {
      $49 = HEAP8[$__extbuf>>0]|0;
      $50 = $49 << 24 >> 24;
      HEAP32[$__1buf>>2] = $50;
      $__nread$044 = $__nread$0$lcssa42;
     }
     else if ((label|0) == 17) {
      $$9 = -1;
      break;
     }
     $__nread$3 = $__nread$044;
     label = 18;
    } else {
     $20 = HEAP8[$__extbuf>>0]|0;
     $21 = $20 << 24 >> 24;
     HEAP32[$__1buf>>2] = $21;
     $__nread$3 = $8;
     label = 18;
    }
   } while(0);
   L21: do {
    if ((label|0) == 18) {
     if ($__consume) {
      $59 = HEAP32[$__1buf>>2]|0;
      $60 = ((($this)) + 48|0);
      HEAP32[$60>>2] = $59;
      $$9 = $59;
      break;
     } else {
      $__i6$0 = $__nread$3;
     }
     while(1) {
      $51 = ($__i6$0|0)>(0);
      if (!($51)) {
       break;
      }
      $52 = (($__i6$0) + -1)|0;
      $53 = (($__extbuf) + ($52)|0);
      $54 = HEAP8[$53>>0]|0;
      $55 = $54 << 24 >> 24;
      $56 = HEAP32[$9>>2]|0;
      $57 = (_ungetc($55,$56)|0);
      $58 = ($57|0)==(-1);
      if ($58) {
       $$9 = -1;
       break L21;
      } else {
       $__i6$0 = $52;
      }
     }
     $$pre = HEAP32[$__1buf>>2]|0;
     $$9 = $$pre;
    }
   } while(0);
   $$10 = $$9;
  }
  $$11 = $$10;
 } else {
  $3 = ((($this)) + 48|0);
  $4 = HEAP32[$3>>2]|0;
  if ($__consume) {
   HEAP32[$3>>2] = -1;
   HEAP8[$0>>0] = 0;
   $$11 = $4;
  } else {
   $$11 = $4;
  }
 }
 STACKTOP = sp;return ($$11|0);
}
function __ZNSt3__110__stdinbufIwE5uflowEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__110__stdinbufIwE9__getcharEb($this,1)|0);
 return ($0|0);
}
function __ZNSt3__110__stdinbufIwE9pbackfailEj($this,$__c) {
 $this = $this|0;
 $__c = $__c|0;
 var $$1 = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__ci = 0, $__enxt = 0, $__extbuf = 0, $__inxt = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__extbuf = sp + 16|0;
 $__enxt = sp + 8|0;
 $__ci = sp + 4|0;
 $__inxt = sp;
 $0 = ($__c|0)==(-1);
 $1 = ((($this)) + 52|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = ($2<<24>>24)!=(0);
 L1: do {
  if ($0) {
   if ($3) {
    $$1 = -1;
   } else {
    $4 = ((($this)) + 48|0);
    $5 = HEAP32[$4>>2]|0;
    $6 = ($5|0)!=(-1);
    $7 = $6&1;
    HEAP8[$1>>0] = $7;
    $$1 = $5;
   }
  } else {
   do {
    if ($3) {
     $8 = ((($this)) + 48|0);
     $9 = HEAP32[$8>>2]|0;
     HEAP32[$__ci>>2] = $9;
     $10 = ((($this)) + 36|0);
     $11 = HEAP32[$10>>2]|0;
     $12 = ((($this)) + 40|0);
     $13 = HEAP32[$12>>2]|0;
     $14 = ((($__ci)) + 4|0);
     $15 = ((($__extbuf)) + 8|0);
     $16 = HEAP32[$11>>2]|0;
     $17 = ((($16)) + 12|0);
     $18 = HEAP32[$17>>2]|0;
     $19 = (FUNCTION_TABLE_iiiiiiiii[$18 & 8191]($11,$13,$__ci,$14,$__inxt,$__extbuf,$15,$__enxt)|0);
     switch ($19|0) {
     case 1: case 2:  {
      $$1 = -1;
      break L1;
      break;
     }
     case 3:  {
      $20 = HEAP32[$8>>2]|0;
      $21 = $20&255;
      HEAP8[$__extbuf>>0] = $21;
      $22 = ((($__extbuf)) + 1|0);
      HEAP32[$__enxt>>2] = $22;
      break;
     }
     default: {
     }
     }
     $23 = ((($this)) + 32|0);
     while(1) {
      $24 = HEAP32[$__enxt>>2]|0;
      $25 = ($24>>>0)>($__extbuf>>>0);
      if (!($25)) {
       label = 13;
       break;
      }
      $26 = ((($24)) + -1|0);
      HEAP32[$__enxt>>2] = $26;
      $27 = HEAP8[$26>>0]|0;
      $28 = $27 << 24 >> 24;
      $29 = HEAP32[$23>>2]|0;
      $30 = (_ungetc($28,$29)|0);
      $31 = ($30|0)==(-1);
      if ($31) {
       label = 12;
       break;
      }
     }
     if ((label|0) == 12) {
      $$1 = -1;
      break L1;
     }
     else if ((label|0) == 13) {
      $$pre$phiZ2D = $8;
      break;
     }
    } else {
     $$pre = ((($this)) + 48|0);
     $$pre$phiZ2D = $$pre;
    }
   } while(0);
   HEAP32[$$pre$phiZ2D>>2] = $__c;
   HEAP8[$1>>0] = 1;
   $$1 = $__c;
  }
 } while(0);
 STACKTOP = sp;return ($$1|0);
}
function __ZNSt3__111__stdoutbufIwEC2EP8_IO_FILEP11__mbstate_t($this,$__fp,$__st) {
 $this = $this|0;
 $__fp = $__fp|0;
 $__st = $__st|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 4|0;
 $1 = sp;
 __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEEC2Ev($this);
 HEAP32[$this>>2] = (154152);
 $2 = ((($this)) + 32|0);
 HEAP32[$2>>2] = $__fp;
 $3 = ((($this)) + 4|0);
 __ZNSt3__16localeC2ERKS0_($0,$3);
 $4 = HEAP32[$0>>2]|0;
 HEAP32[$1>>2] = $4;
 __THREW__ = 0;
 $5 = (invoke_iii(3374,($1|0),(1209376|0))|0);
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  __ZNSt3__16localeD2Ev($1);
  __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEED2Ev($this);
  ___resumeException($16|0);
  // unreachable;
 } else {
  $8 = ((($this)) + 36|0);
  __ZNSt3__16localeD2Ev($1);
  HEAP32[$8>>2] = $5;
  $9 = ((($this)) + 40|0);
  HEAP32[$9>>2] = $__st;
  $10 = ((($this)) + 44|0);
  $11 = HEAP32[$5>>2]|0;
  $12 = ((($11)) + 28|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = (FUNCTION_TABLE_ii[$13 & 8191]($5)|0);
  $15 = $14&1;
  HEAP8[$10>>0] = $15;
  STACKTOP = sp;return;
 }
}
function __ZNSt3__111__stdoutbufIwED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEED2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNSt3__111__stdoutbufIwE5imbueERKNS_6localeE($this,$__loc) {
 $this = $this|0;
 $__loc = $__loc|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + 24|0);
 $2 = HEAP32[$1>>2]|0;
 (FUNCTION_TABLE_ii[$2 & 8191]($this)|0);
 $3 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1209376)|0);
 $4 = ((($this)) + 36|0);
 HEAP32[$4>>2] = $3;
 $5 = HEAP32[$3>>2]|0;
 $6 = ((($5)) + 28|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (FUNCTION_TABLE_ii[$7 & 8191]($3)|0);
 $9 = ((($this)) + 44|0);
 $10 = $8&1;
 HEAP8[$9>>0] = $10;
 return;
}
function __ZNSt3__111__stdoutbufIwE4syncEv($this) {
 $this = $this|0;
 var $$2 = 0, $$3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $__extbe = 0, $__extbuf = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__extbuf = sp + 8|0;
 $__extbe = sp;
 $0 = ((($this)) + 36|0);
 $1 = ((($this)) + 40|0);
 $2 = ((($__extbuf)) + 8|0);
 $3 = $__extbuf;
 $4 = ((($this)) + 32|0);
 L1: while(1) {
  $5 = HEAP32[$0>>2]|0;
  $6 = HEAP32[$1>>2]|0;
  $7 = HEAP32[$5>>2]|0;
  $8 = ((($7)) + 20|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = (FUNCTION_TABLE_iiiiii[$9 & 8191]($5,$6,$__extbuf,$2,$__extbe)|0);
  $11 = HEAP32[$__extbe>>2]|0;
  $12 = (($11) - ($3))|0;
  $13 = HEAP32[$4>>2]|0;
  $14 = (_fwrite($__extbuf,1,$12,$13)|0);
  $15 = ($14|0)==($12|0);
  if (!($15)) {
   $$2 = -1;
   break;
  }
  switch ($10|0) {
  case 1:  {
   break;
  }
  case 2:  {
   $$2 = -1;
   break L1;
   break;
  }
  default: {
   label = 4;
   break L1;
  }
  }
 }
 if ((label|0) == 4) {
  $16 = HEAP32[$4>>2]|0;
  $17 = (_fflush($16)|0);
  $not$ = ($17|0)!=(0);
  $$3 = $not$ << 31 >> 31;
  $$2 = $$3;
 }
 STACKTOP = sp;return ($$2|0);
}
function __ZNSt3__111__stdoutbufIwE6xsputnEPKwi($this,$__s,$__n) {
 $this = $this|0;
 $__s = $__s|0;
 $__n = $__n|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__i$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 44|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = ($1<<24>>24)==(0);
 L1: do {
  if ($2) {
   $$01 = $__s;$__i$0 = 0;
   while(1) {
    $6 = ($__i$0|0)<($__n|0);
    if (!($6)) {
     $$0 = $__i$0;
     break L1;
    }
    $7 = HEAP32[$this>>2]|0;
    $8 = ((($7)) + 52|0);
    $9 = HEAP32[$8>>2]|0;
    $10 = HEAP32[$$01>>2]|0;
    $11 = (FUNCTION_TABLE_iii[$9 & 8191]($this,$10)|0);
    $12 = ($11|0)==(-1);
    if ($12) {
     $$0 = $__i$0;
     break L1;
    }
    $13 = (($__i$0) + 1)|0;
    $14 = ((($$01)) + 4|0);
    $$01 = $14;$__i$0 = $13;
   }
  } else {
   $3 = ((($this)) + 32|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = (_fwrite($__s,4,$__n,$4)|0);
   $$0 = $5;
  }
 } while(0);
 return ($$0|0);
}
function __ZNSt3__111__stdoutbufIwE8overflowEj($this,$__c) {
 $this = $this|0;
 $__c = $__c|0;
 var $$4 = 0, $$__c$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__1buf = 0, $__e = 0, $__extbe = 0;
 var $__extbuf = 0, $pbase$0 = 0, $pbase$0$lcssa25 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__extbuf = sp + 16|0;
 $__1buf = sp + 8|0;
 $__extbe = sp + 4|0;
 $__e = sp;
 $0 = ($__c|0)==(-1);
 do {
  if ($0) {
   label = 14;
  } else {
   HEAP32[$__1buf>>2] = $__c;
   $1 = ((($this)) + 44|0);
   $2 = HEAP8[$1>>0]|0;
   $3 = ($2<<24>>24)==(0);
   if (!($3)) {
    $4 = ((($this)) + 32|0);
    $5 = HEAP32[$4>>2]|0;
    $6 = (_fwrite($__1buf,4,1,$5)|0);
    $7 = ($6|0)==(1);
    if ($7) {
     label = 14;
     break;
    } else {
     $$4 = -1;
     break;
    }
   }
   HEAP32[$__extbe>>2] = $__extbuf;
   $8 = ((($__1buf)) + 4|0);
   $9 = ((($this)) + 36|0);
   $10 = ((($this)) + 40|0);
   $11 = ((($__extbuf)) + 8|0);
   $12 = $__extbuf;
   $13 = ((($this)) + 32|0);
   $pbase$0 = $__1buf;
   while(1) {
    $14 = HEAP32[$9>>2]|0;
    $15 = HEAP32[$10>>2]|0;
    $16 = HEAP32[$14>>2]|0;
    $17 = ((($16)) + 12|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = (FUNCTION_TABLE_iiiiiiiii[$18 & 8191]($14,$15,$pbase$0,$8,$__e,$__extbuf,$11,$__extbe)|0);
    $20 = HEAP32[$__e>>2]|0;
    $21 = ($20|0)==($pbase$0|0);
    if ($21) {
     label = 13;
     break;
    }
    $22 = ($19|0)==(3);
    if ($22) {
     $pbase$0$lcssa25 = $pbase$0;
     label = 7;
     break;
    }
    $26 = ($19|0)==(1);
    $27 = $19 | 1;
    $28 = ($27|0)==(1);
    if (!($28)) {
     label = 13;
     break;
    }
    $29 = HEAP32[$__extbe>>2]|0;
    $30 = (($29) - ($12))|0;
    $31 = HEAP32[$13>>2]|0;
    $32 = (_fwrite($__extbuf,1,$30,$31)|0);
    $33 = ($32|0)==($30|0);
    if (!($33)) {
     label = 13;
     break;
    }
    $34 = HEAP32[$__e>>2]|0;
    if ($26) {
     $pbase$0 = $34;
    } else {
     label = 12;
     break;
    }
   }
   if ((label|0) == 7) {
    $23 = HEAP32[$13>>2]|0;
    $24 = (_fwrite($pbase$0$lcssa25,1,1,$23)|0);
    $25 = ($24|0)==(1);
    if ($25) {
     label = 12;
    } else {
     label = 13;
    }
   }
   if ((label|0) == 12) {
    label = 14;
    break;
   }
   else if ((label|0) == 13) {
    $$4 = -1;
    break;
   }
  }
 } while(0);
 if ((label|0) == 14) {
  $$__c$i = $0 ? 0 : $__c;
  $$4 = $$__c$i;
 }
 STACKTOP = sp;return ($$4|0);
}
function __ZNSt3__18ios_base4InitD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __THREW__ = 0;
 (invoke_ii(3502,(1206612|0))|0);
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 if (!($1)) {
  __THREW__ = 0;
  (invoke_ii(3502,(1206780|0))|0);
  $2 = __THREW__; __THREW__ = 0;
  $3 = $2&1;
  if (!($3)) {
   __THREW__ = 0;
   (invoke_ii(4853,(1206952|0))|0);
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   if (!($5)) {
    __THREW__ = 0;
    (invoke_ii(4853,(1207120|0))|0);
    $6 = __THREW__; __THREW__ = 0;
    $7 = $6&1;
    if (!($7)) {
     return;
    }
   }
  }
 }
 $8 = ___cxa_find_matching_catch_3(0|0)|0;
 $9 = tempRet0;
 ___clang_call_terminate($8);
 // unreachable;
}
function __GLOBAL__sub_I_iostream_cpp() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__118__libcpp_refstringC2EPKc($this,$msg) {
 $this = $this|0;
 $msg = $msg|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strlen($msg)|0);
 $1 = (($0) + 13)|0;
 $2 = (__Znwj($1)|0);
 HEAP32[$2>>2] = $0;
 $3 = ((($2)) + 4|0);
 HEAP32[$3>>2] = $0;
 $4 = ((($2)) + 8|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($2)) + 12|0);
 $6 = (($0) + 1)|0;
 _memcpy(($5|0),($msg|0),($6|0))|0;
 HEAP32[$this>>2] = $5;
 return;
}
function __ZNSt11logic_errorC2EPKc($this,$msg) {
 $this = $this|0;
 $msg = $msg|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (157852);
 $0 = ((($this)) + 4|0);
 __THREW__ = 0;
 invoke_vii(4854,($0|0),($msg|0));
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  ___resumeException($3|0);
  // unreachable;
 } else {
  return;
 }
}
function __ZNSt13runtime_errorC2ERKNSt3__112basic_stringIcNS0_11char_traitsIcEENS0_9allocatorIcEEEE($this,$msg) {
 $this = $this|0;
 $msg = $msg|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (157872);
 $0 = ((($this)) + 4|0);
 $1 = HEAP8[$msg>>0]|0;
 $2 = $1 & 1;
 $3 = ($2<<24>>24)==(0);
 $4 = ((($msg)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($msg)) + 1|0);
 $7 = $3 ? $6 : $5;
 __THREW__ = 0;
 invoke_vii(4854,($0|0),($7|0));
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  ___resumeException($10|0);
  // unreachable;
 } else {
  return;
 }
}
function __ZNSt13runtime_errorC2EPKc($this,$msg) {
 $this = $this|0;
 $msg = $msg|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (157872);
 $0 = ((($this)) + 4|0);
 __THREW__ = 0;
 invoke_vii(4854,($0|0),($msg|0));
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  ___resumeException($3|0);
  // unreachable;
 } else {
  return;
 }
}
function __ZNSt3__112__next_primeEj($n) {
 $n = $n|0;
 var $$0$ph$i = 0, $$0$ph$i$lcssa = 0, $$0$ph$i13 = 0, $$0$ph$i13$lcssa = 0, $$8 = 0, $$9 = 0, $$k0$0 = 0, $$lcssa = 0, $$lcssa73 = 0, $$lcssa76 = 0, $$lcssa77 = 0, $$neg1$i = 0, $$neg1$i16 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0;
 var $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0;
 var $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0;
 var $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0;
 var $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0;
 var $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0;
 var $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0;
 var $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0;
 var $285 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__len$0$i = 0;
 var $__len$0$i$lcssa72 = 0, $__len$0$i15 = 0, $__len$0$i15$lcssa75 = 0, $__len$0$ph$i = 0, $__len$0$ph$i12 = 0, $i$0 = 0, $in$0 = 0, $j$0 = 0, $k0$0 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n>>>0)<(212);
 L1: do {
  if ($0) {
   $$0$ph$i = 154208;$__len$0$ph$i = 48;
   L2: while(1) {
    $__len$0$i = $__len$0$ph$i;
    while(1) {
     $1 = ($__len$0$i|0)==(0);
     if ($1) {
      $$0$ph$i$lcssa = $$0$ph$i;
      break L2;
     }
     $2 = (($__len$0$i|0) / 2)&-1;
     $3 = (($$0$ph$i) + ($2<<2)|0);
     $4 = HEAP32[$3>>2]|0;
     $5 = ($4>>>0)<($n>>>0);
     if ($5) {
      $$lcssa = $2;$$lcssa73 = $3;$__len$0$i$lcssa72 = $__len$0$i;
      break;
     } else {
      $__len$0$i = $2;
     }
    }
    $6 = ((($$lcssa73)) + 4|0);
    $$neg1$i = (($__len$0$i$lcssa72) + -1)|0;
    $7 = (($$neg1$i) - ($$lcssa))|0;
    $$0$ph$i = $6;$__len$0$ph$i = $7;
   }
   $8 = HEAP32[$$0$ph$i$lcssa>>2]|0;
   $$8 = $8;
  } else {
   $9 = ($n>>>0)>(4294967291);
   if ($9) {
    $10 = (___cxa_allocate_exception(8)|0);
    __THREW__ = 0;
    invoke_vii(4855,($10|0),(619721|0));
    $11 = __THREW__; __THREW__ = 0;
    $12 = $11&1;
    if ($12) {
     $13 = ___cxa_find_matching_catch_2()|0;
     $14 = tempRet0;
     ___cxa_free_exception(($10|0));
     ___resumeException($13|0);
     // unreachable;
    } else {
     HEAP32[$10>>2] = (157932);
     ___cxa_throw(($10|0),(9136|0),(3331|0));
     // unreachable;
    }
   }
   $15 = (($n>>>0) / 210)&-1;
   $16 = ($15*210)|0;
   $17 = (($n) - ($16))|0;
   $$0$ph$i13 = 154400;$__len$0$ph$i12 = 48;
   L16: while(1) {
    $__len$0$i15 = $__len$0$ph$i12;
    while(1) {
     $18 = ($__len$0$i15|0)==(0);
     if ($18) {
      $$0$ph$i13$lcssa = $$0$ph$i13;
      break L16;
     }
     $19 = (($__len$0$i15|0) / 2)&-1;
     $20 = (($$0$ph$i13) + ($19<<2)|0);
     $21 = HEAP32[$20>>2]|0;
     $22 = ($21>>>0)<($17>>>0);
     if ($22) {
      $$lcssa76 = $19;$$lcssa77 = $20;$__len$0$i15$lcssa75 = $__len$0$i15;
      break;
     } else {
      $__len$0$i15 = $19;
     }
    }
    $23 = ((($$lcssa77)) + 4|0);
    $$neg1$i16 = (($__len$0$i15$lcssa75) + -1)|0;
    $24 = (($$neg1$i16) - ($$lcssa76))|0;
    $$0$ph$i13 = $23;$__len$0$ph$i12 = $24;
   }
   $25 = $$0$ph$i13$lcssa;
   $26 = (($25) - (154400))|0;
   $27 = $26 >> 2;
   $28 = (154400 + ($27<<2)|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = (($29) + ($16))|0;
   $in$0 = $27;$k0$0 = $15;$storemerge = $30;
   while(1) {
    $j$0 = 5;
    while(1) {
     $31 = ($j$0>>>0)<(47);
     if (!($31)) {
      $i$0 = 211;
      label = 21;
      break;
     }
     $32 = (154208 + ($j$0<<2)|0);
     $33 = HEAP32[$32>>2]|0;
     $34 = (($storemerge>>>0) / ($33>>>0))&-1;
     $35 = ($34>>>0)<($33>>>0);
     if ($35) {
      $$8 = $storemerge;
      break L1;
     }
     $36 = Math_imul($34, $33)|0;
     $37 = ($storemerge|0)==($36|0);
     $38 = (($j$0) + 1)|0;
     if ($37) {
      break;
     } else {
      $j$0 = $38;
     }
    }
    L29: do {
     if ((label|0) == 21) {
      while(1) {
       label = 0;
       $39 = (($storemerge>>>0) / ($i$0>>>0))&-1;
       $40 = ($39>>>0)<($i$0>>>0);
       if ($40) {
        $$8 = $storemerge;
        break L1;
       }
       $41 = Math_imul($39, $i$0)|0;
       $42 = ($storemerge|0)==($41|0);
       if ($42) {
        break L29;
       }
       $43 = (($i$0) + 10)|0;
       $44 = (($storemerge>>>0) / ($43>>>0))&-1;
       $45 = ($44>>>0)<($43>>>0);
       if ($45) {
        $$8 = $storemerge;
        break L1;
       }
       $46 = Math_imul($44, $43)|0;
       $47 = ($storemerge|0)==($46|0);
       if ($47) {
        break L29;
       }
       $48 = (($i$0) + 12)|0;
       $49 = (($storemerge>>>0) / ($48>>>0))&-1;
       $50 = ($49>>>0)<($48>>>0);
       if ($50) {
        $$8 = $storemerge;
        break L1;
       }
       $51 = Math_imul($49, $48)|0;
       $52 = ($storemerge|0)==($51|0);
       if ($52) {
        break L29;
       }
       $53 = (($i$0) + 16)|0;
       $54 = (($storemerge>>>0) / ($53>>>0))&-1;
       $55 = ($54>>>0)<($53>>>0);
       if ($55) {
        $$8 = $storemerge;
        break L1;
       }
       $56 = Math_imul($54, $53)|0;
       $57 = ($storemerge|0)==($56|0);
       if ($57) {
        break L29;
       }
       $58 = (($i$0) + 18)|0;
       $59 = (($storemerge>>>0) / ($58>>>0))&-1;
       $60 = ($59>>>0)<($58>>>0);
       if ($60) {
        $$8 = $storemerge;
        break L1;
       }
       $61 = Math_imul($59, $58)|0;
       $62 = ($storemerge|0)==($61|0);
       if ($62) {
        break L29;
       }
       $63 = (($i$0) + 22)|0;
       $64 = (($storemerge>>>0) / ($63>>>0))&-1;
       $65 = ($64>>>0)<($63>>>0);
       if ($65) {
        $$8 = $storemerge;
        break L1;
       }
       $66 = Math_imul($64, $63)|0;
       $67 = ($storemerge|0)==($66|0);
       if ($67) {
        break L29;
       }
       $68 = (($i$0) + 28)|0;
       $69 = (($storemerge>>>0) / ($68>>>0))&-1;
       $70 = ($69>>>0)<($68>>>0);
       if ($70) {
        $$8 = $storemerge;
        break L1;
       }
       $71 = Math_imul($69, $68)|0;
       $72 = ($storemerge|0)==($71|0);
       if ($72) {
        break L29;
       }
       $73 = (($i$0) + 30)|0;
       $74 = (($storemerge>>>0) / ($73>>>0))&-1;
       $75 = ($74>>>0)<($73>>>0);
       if ($75) {
        $$8 = $storemerge;
        break L1;
       }
       $76 = Math_imul($74, $73)|0;
       $77 = ($storemerge|0)==($76|0);
       if ($77) {
        break L29;
       }
       $78 = (($i$0) + 36)|0;
       $79 = (($storemerge>>>0) / ($78>>>0))&-1;
       $80 = ($79>>>0)<($78>>>0);
       if ($80) {
        $$8 = $storemerge;
        break L1;
       }
       $81 = Math_imul($79, $78)|0;
       $82 = ($storemerge|0)==($81|0);
       if ($82) {
        break L29;
       }
       $83 = (($i$0) + 40)|0;
       $84 = (($storemerge>>>0) / ($83>>>0))&-1;
       $85 = ($84>>>0)<($83>>>0);
       if ($85) {
        $$8 = $storemerge;
        break L1;
       }
       $86 = Math_imul($84, $83)|0;
       $87 = ($storemerge|0)==($86|0);
       if ($87) {
        break L29;
       }
       $88 = (($i$0) + 42)|0;
       $89 = (($storemerge>>>0) / ($88>>>0))&-1;
       $90 = ($89>>>0)<($88>>>0);
       if ($90) {
        $$8 = $storemerge;
        break L1;
       }
       $91 = Math_imul($89, $88)|0;
       $92 = ($storemerge|0)==($91|0);
       if ($92) {
        break L29;
       }
       $93 = (($i$0) + 46)|0;
       $94 = (($storemerge>>>0) / ($93>>>0))&-1;
       $95 = ($94>>>0)<($93>>>0);
       if ($95) {
        $$8 = $storemerge;
        break L1;
       }
       $96 = Math_imul($94, $93)|0;
       $97 = ($storemerge|0)==($96|0);
       if ($97) {
        break L29;
       }
       $98 = (($i$0) + 52)|0;
       $99 = (($storemerge>>>0) / ($98>>>0))&-1;
       $100 = ($99>>>0)<($98>>>0);
       if ($100) {
        $$8 = $storemerge;
        break L1;
       }
       $101 = Math_imul($99, $98)|0;
       $102 = ($storemerge|0)==($101|0);
       if ($102) {
        break L29;
       }
       $103 = (($i$0) + 58)|0;
       $104 = (($storemerge>>>0) / ($103>>>0))&-1;
       $105 = ($104>>>0)<($103>>>0);
       if ($105) {
        $$8 = $storemerge;
        break L1;
       }
       $106 = Math_imul($104, $103)|0;
       $107 = ($storemerge|0)==($106|0);
       if ($107) {
        break L29;
       }
       $108 = (($i$0) + 60)|0;
       $109 = (($storemerge>>>0) / ($108>>>0))&-1;
       $110 = ($109>>>0)<($108>>>0);
       if ($110) {
        $$8 = $storemerge;
        break L1;
       }
       $111 = Math_imul($109, $108)|0;
       $112 = ($storemerge|0)==($111|0);
       if ($112) {
        break L29;
       }
       $113 = (($i$0) + 66)|0;
       $114 = (($storemerge>>>0) / ($113>>>0))&-1;
       $115 = ($114>>>0)<($113>>>0);
       if ($115) {
        $$8 = $storemerge;
        break L1;
       }
       $116 = Math_imul($114, $113)|0;
       $117 = ($storemerge|0)==($116|0);
       if ($117) {
        break L29;
       }
       $118 = (($i$0) + 70)|0;
       $119 = (($storemerge>>>0) / ($118>>>0))&-1;
       $120 = ($119>>>0)<($118>>>0);
       if ($120) {
        $$8 = $storemerge;
        break L1;
       }
       $121 = Math_imul($119, $118)|0;
       $122 = ($storemerge|0)==($121|0);
       if ($122) {
        break L29;
       }
       $123 = (($i$0) + 72)|0;
       $124 = (($storemerge>>>0) / ($123>>>0))&-1;
       $125 = ($124>>>0)<($123>>>0);
       if ($125) {
        $$8 = $storemerge;
        break L1;
       }
       $126 = Math_imul($124, $123)|0;
       $127 = ($storemerge|0)==($126|0);
       if ($127) {
        break L29;
       }
       $128 = (($i$0) + 78)|0;
       $129 = (($storemerge>>>0) / ($128>>>0))&-1;
       $130 = ($129>>>0)<($128>>>0);
       if ($130) {
        $$8 = $storemerge;
        break L1;
       }
       $131 = Math_imul($129, $128)|0;
       $132 = ($storemerge|0)==($131|0);
       if ($132) {
        break L29;
       }
       $133 = (($i$0) + 82)|0;
       $134 = (($storemerge>>>0) / ($133>>>0))&-1;
       $135 = ($134>>>0)<($133>>>0);
       if ($135) {
        $$8 = $storemerge;
        break L1;
       }
       $136 = Math_imul($134, $133)|0;
       $137 = ($storemerge|0)==($136|0);
       if ($137) {
        break L29;
       }
       $138 = (($i$0) + 88)|0;
       $139 = (($storemerge>>>0) / ($138>>>0))&-1;
       $140 = ($139>>>0)<($138>>>0);
       if ($140) {
        $$8 = $storemerge;
        break L1;
       }
       $141 = Math_imul($139, $138)|0;
       $142 = ($storemerge|0)==($141|0);
       if ($142) {
        break L29;
       }
       $143 = (($i$0) + 96)|0;
       $144 = (($storemerge>>>0) / ($143>>>0))&-1;
       $145 = ($144>>>0)<($143>>>0);
       if ($145) {
        $$8 = $storemerge;
        break L1;
       }
       $146 = Math_imul($144, $143)|0;
       $147 = ($storemerge|0)==($146|0);
       if ($147) {
        break L29;
       }
       $148 = (($i$0) + 100)|0;
       $149 = (($storemerge>>>0) / ($148>>>0))&-1;
       $150 = ($149>>>0)<($148>>>0);
       if ($150) {
        $$8 = $storemerge;
        break L1;
       }
       $151 = Math_imul($149, $148)|0;
       $152 = ($storemerge|0)==($151|0);
       if ($152) {
        break L29;
       }
       $153 = (($i$0) + 102)|0;
       $154 = (($storemerge>>>0) / ($153>>>0))&-1;
       $155 = ($154>>>0)<($153>>>0);
       if ($155) {
        $$8 = $storemerge;
        break L1;
       }
       $156 = Math_imul($154, $153)|0;
       $157 = ($storemerge|0)==($156|0);
       if ($157) {
        break L29;
       }
       $158 = (($i$0) + 106)|0;
       $159 = (($storemerge>>>0) / ($158>>>0))&-1;
       $160 = ($159>>>0)<($158>>>0);
       if ($160) {
        $$8 = $storemerge;
        break L1;
       }
       $161 = Math_imul($159, $158)|0;
       $162 = ($storemerge|0)==($161|0);
       if ($162) {
        break L29;
       }
       $163 = (($i$0) + 108)|0;
       $164 = (($storemerge>>>0) / ($163>>>0))&-1;
       $165 = ($164>>>0)<($163>>>0);
       if ($165) {
        $$8 = $storemerge;
        break L1;
       }
       $166 = Math_imul($164, $163)|0;
       $167 = ($storemerge|0)==($166|0);
       if ($167) {
        break L29;
       }
       $168 = (($i$0) + 112)|0;
       $169 = (($storemerge>>>0) / ($168>>>0))&-1;
       $170 = ($169>>>0)<($168>>>0);
       if ($170) {
        $$8 = $storemerge;
        break L1;
       }
       $171 = Math_imul($169, $168)|0;
       $172 = ($storemerge|0)==($171|0);
       if ($172) {
        break L29;
       }
       $173 = (($i$0) + 120)|0;
       $174 = (($storemerge>>>0) / ($173>>>0))&-1;
       $175 = ($174>>>0)<($173>>>0);
       if ($175) {
        $$8 = $storemerge;
        break L1;
       }
       $176 = Math_imul($174, $173)|0;
       $177 = ($storemerge|0)==($176|0);
       if ($177) {
        break L29;
       }
       $178 = (($i$0) + 126)|0;
       $179 = (($storemerge>>>0) / ($178>>>0))&-1;
       $180 = ($179>>>0)<($178>>>0);
       if ($180) {
        $$8 = $storemerge;
        break L1;
       }
       $181 = Math_imul($179, $178)|0;
       $182 = ($storemerge|0)==($181|0);
       if ($182) {
        break L29;
       }
       $183 = (($i$0) + 130)|0;
       $184 = (($storemerge>>>0) / ($183>>>0))&-1;
       $185 = ($184>>>0)<($183>>>0);
       if ($185) {
        $$8 = $storemerge;
        break L1;
       }
       $186 = Math_imul($184, $183)|0;
       $187 = ($storemerge|0)==($186|0);
       if ($187) {
        break L29;
       }
       $188 = (($i$0) + 136)|0;
       $189 = (($storemerge>>>0) / ($188>>>0))&-1;
       $190 = ($189>>>0)<($188>>>0);
       if ($190) {
        $$8 = $storemerge;
        break L1;
       }
       $191 = Math_imul($189, $188)|0;
       $192 = ($storemerge|0)==($191|0);
       if ($192) {
        break L29;
       }
       $193 = (($i$0) + 138)|0;
       $194 = (($storemerge>>>0) / ($193>>>0))&-1;
       $195 = ($194>>>0)<($193>>>0);
       if ($195) {
        $$8 = $storemerge;
        break L1;
       }
       $196 = Math_imul($194, $193)|0;
       $197 = ($storemerge|0)==($196|0);
       if ($197) {
        break L29;
       }
       $198 = (($i$0) + 142)|0;
       $199 = (($storemerge>>>0) / ($198>>>0))&-1;
       $200 = ($199>>>0)<($198>>>0);
       if ($200) {
        $$8 = $storemerge;
        break L1;
       }
       $201 = Math_imul($199, $198)|0;
       $202 = ($storemerge|0)==($201|0);
       if ($202) {
        break L29;
       }
       $203 = (($i$0) + 148)|0;
       $204 = (($storemerge>>>0) / ($203>>>0))&-1;
       $205 = ($204>>>0)<($203>>>0);
       if ($205) {
        $$8 = $storemerge;
        break L1;
       }
       $206 = Math_imul($204, $203)|0;
       $207 = ($storemerge|0)==($206|0);
       if ($207) {
        break L29;
       }
       $208 = (($i$0) + 150)|0;
       $209 = (($storemerge>>>0) / ($208>>>0))&-1;
       $210 = ($209>>>0)<($208>>>0);
       if ($210) {
        $$8 = $storemerge;
        break L1;
       }
       $211 = Math_imul($209, $208)|0;
       $212 = ($storemerge|0)==($211|0);
       if ($212) {
        break L29;
       }
       $213 = (($i$0) + 156)|0;
       $214 = (($storemerge>>>0) / ($213>>>0))&-1;
       $215 = ($214>>>0)<($213>>>0);
       if ($215) {
        $$8 = $storemerge;
        break L1;
       }
       $216 = Math_imul($214, $213)|0;
       $217 = ($storemerge|0)==($216|0);
       if ($217) {
        break L29;
       }
       $218 = (($i$0) + 162)|0;
       $219 = (($storemerge>>>0) / ($218>>>0))&-1;
       $220 = ($219>>>0)<($218>>>0);
       if ($220) {
        $$8 = $storemerge;
        break L1;
       }
       $221 = Math_imul($219, $218)|0;
       $222 = ($storemerge|0)==($221|0);
       if ($222) {
        break L29;
       }
       $223 = (($i$0) + 166)|0;
       $224 = (($storemerge>>>0) / ($223>>>0))&-1;
       $225 = ($224>>>0)<($223>>>0);
       if ($225) {
        $$8 = $storemerge;
        break L1;
       }
       $226 = Math_imul($224, $223)|0;
       $227 = ($storemerge|0)==($226|0);
       if ($227) {
        break L29;
       }
       $228 = (($i$0) + 168)|0;
       $229 = (($storemerge>>>0) / ($228>>>0))&-1;
       $230 = ($229>>>0)<($228>>>0);
       if ($230) {
        $$8 = $storemerge;
        break L1;
       }
       $231 = Math_imul($229, $228)|0;
       $232 = ($storemerge|0)==($231|0);
       if ($232) {
        break L29;
       }
       $233 = (($i$0) + 172)|0;
       $234 = (($storemerge>>>0) / ($233>>>0))&-1;
       $235 = ($234>>>0)<($233>>>0);
       if ($235) {
        $$8 = $storemerge;
        break L1;
       }
       $236 = Math_imul($234, $233)|0;
       $237 = ($storemerge|0)==($236|0);
       if ($237) {
        break L29;
       }
       $238 = (($i$0) + 178)|0;
       $239 = (($storemerge>>>0) / ($238>>>0))&-1;
       $240 = ($239>>>0)<($238>>>0);
       if ($240) {
        $$8 = $storemerge;
        break L1;
       }
       $241 = Math_imul($239, $238)|0;
       $242 = ($storemerge|0)==($241|0);
       if ($242) {
        break L29;
       }
       $243 = (($i$0) + 180)|0;
       $244 = (($storemerge>>>0) / ($243>>>0))&-1;
       $245 = ($244>>>0)<($243>>>0);
       if ($245) {
        $$8 = $storemerge;
        break L1;
       }
       $246 = Math_imul($244, $243)|0;
       $247 = ($storemerge|0)==($246|0);
       if ($247) {
        break L29;
       }
       $248 = (($i$0) + 186)|0;
       $249 = (($storemerge>>>0) / ($248>>>0))&-1;
       $250 = ($249>>>0)<($248>>>0);
       if ($250) {
        $$8 = $storemerge;
        break L1;
       }
       $251 = Math_imul($249, $248)|0;
       $252 = ($storemerge|0)==($251|0);
       if ($252) {
        break L29;
       }
       $253 = (($i$0) + 190)|0;
       $254 = (($storemerge>>>0) / ($253>>>0))&-1;
       $255 = ($254>>>0)<($253>>>0);
       if ($255) {
        $$8 = $storemerge;
        break L1;
       }
       $256 = Math_imul($254, $253)|0;
       $257 = ($storemerge|0)==($256|0);
       if ($257) {
        break L29;
       }
       $258 = (($i$0) + 192)|0;
       $259 = (($storemerge>>>0) / ($258>>>0))&-1;
       $260 = ($259>>>0)<($258>>>0);
       if ($260) {
        $$8 = $storemerge;
        break L1;
       }
       $261 = Math_imul($259, $258)|0;
       $262 = ($storemerge|0)==($261|0);
       if ($262) {
        break L29;
       }
       $263 = (($i$0) + 196)|0;
       $264 = (($storemerge>>>0) / ($263>>>0))&-1;
       $265 = ($264>>>0)<($263>>>0);
       if ($265) {
        $$8 = $storemerge;
        break L1;
       }
       $266 = Math_imul($264, $263)|0;
       $267 = ($storemerge|0)==($266|0);
       if ($267) {
        break L29;
       }
       $268 = (($i$0) + 198)|0;
       $269 = (($storemerge>>>0) / ($268>>>0))&-1;
       $270 = ($269>>>0)<($268>>>0);
       if ($270) {
        $$8 = $storemerge;
        break L1;
       }
       $271 = Math_imul($269, $268)|0;
       $272 = ($storemerge|0)==($271|0);
       if ($272) {
        break L29;
       }
       $273 = (($i$0) + 208)|0;
       $274 = (($storemerge>>>0) / ($273>>>0))&-1;
       $275 = ($274>>>0)<($273>>>0);
       if ($275) {
        $$8 = $storemerge;
        break L1;
       }
       $276 = Math_imul($274, $273)|0;
       $277 = ($storemerge|0)==($276|0);
       $278 = (($i$0) + 210)|0;
       if ($277) {
        break;
       } else {
        $i$0 = $278;
        label = 21;
       }
      }
     }
    } while(0);
    $279 = (($in$0) + 1)|0;
    $280 = ($279|0)==(48);
    $281 = $280&1;
    $$k0$0 = (($281) + ($k0$0))|0;
    $$9 = $280 ? 0 : $279;
    $282 = ($$k0$0*210)|0;
    $283 = (154400 + ($$9<<2)|0);
    $284 = HEAP32[$283>>2]|0;
    $285 = (($284) + ($282))|0;
    $in$0 = $$9;$k0$0 = $$k0$0;$storemerge = $285;
   }
  }
 } while(0);
 return ($$8|0);
}
function __ZNKSt3__121__basic_string_commonILb1EE20__throw_length_errorEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (___cxa_allocate_exception(8)|0);
 __THREW__ = 0;
 invoke_vii(3996,($0|0),(619743|0));
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  ___cxa_free_exception(($0|0));
  ___resumeException($3|0);
  // unreachable;
 } else {
  HEAP32[$0>>2] = (157892);
  ___cxa_throw(($0|0),(9104|0),(3328|0));
  // unreachable;
 }
}
function __ZNKSt3__121__basic_string_commonILb1EE20__throw_out_of_rangeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (___cxa_allocate_exception(8)|0);
 __THREW__ = 0;
 invoke_vii(3996,($0|0),(619743|0));
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  ___cxa_free_exception(($0|0));
  ___resumeException($3|0);
  // unreachable;
 } else {
  HEAP32[$0>>2] = (157912);
  ___cxa_throw(($0|0),(9120|0),(3328|0));
  // unreachable;
 }
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_($this,$__str) {
 $this = $this|0;
 $__str = $__str|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$__str>>0]|0;
 $1 = $0 & 1;
 $2 = ($1<<24>>24)==(0);
 if ($2) {
  ;HEAP32[$this>>2]=HEAP32[$__str>>2]|0;HEAP32[$this+4>>2]=HEAP32[$__str+4>>2]|0;HEAP32[$this+8>>2]=HEAP32[$__str+8>>2]|0;
 } else {
  $3 = ((($__str)) + 8|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ((($__str)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($this,$4,$6);
 }
 return;
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($this,$__s,$__sz) {
 $this = $this|0;
 $__s = $__s|0;
 $__sz = $__sz|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__p$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__sz>>>0)>(4294967279);
 if ($0) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_length_errorEv($this);
  // unreachable;
 }
 $1 = ($__sz>>>0)<(11);
 if ($1) {
  $2 = $__sz << 1;
  $3 = $2&255;
  HEAP8[$this>>0] = $3;
  $4 = ((($this)) + 1|0);
  $__p$0 = $4;
 } else {
  $5 = (($__sz) + 16)|0;
  $6 = $5 & -16;
  $7 = (__Znwj($6)|0);
  $8 = ((($this)) + 8|0);
  HEAP32[$8>>2] = $7;
  $9 = $6 | 1;
  HEAP32[$this>>2] = $9;
  $10 = ((($this)) + 4|0);
  HEAP32[$10>>2] = $__sz;
  $__p$0 = $7;
 }
 _memcpy(($__p$0|0),($__s|0),($__sz|0))|0;
 $11 = (($__p$0) + ($__sz)|0);
 HEAP8[$11>>0] = 0;
 return;
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEjc($this,$__n,$__c) {
 $this = $this|0;
 $__n = $__n|0;
 $__c = $__c|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__p$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__n>>>0)>(4294967279);
 if ($0) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_length_errorEv($this);
  // unreachable;
 }
 $1 = ($__n>>>0)<(11);
 if ($1) {
  $2 = $__n << 1;
  $3 = $2&255;
  HEAP8[$this>>0] = $3;
  $4 = ((($this)) + 1|0);
  $__p$0 = $4;
 } else {
  $5 = (($__n) + 16)|0;
  $6 = $5 & -16;
  $7 = (__Znwj($6)|0);
  $8 = ((($this)) + 8|0);
  HEAP32[$8>>2] = $7;
  $9 = $6 | 1;
  HEAP32[$this>>2] = $9;
  $10 = ((($this)) + 4|0);
  HEAP32[$10>>2] = $__n;
  $__p$0 = $7;
 }
 _memset(($__p$0|0),($__c|0),($__n|0))|0;
 $11 = (($__p$0) + ($__n)|0);
 HEAP8[$11>>0] = 0;
 return;
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_jjRKS4_($this,$__str,$__pos,$__n,$__a) {
 $this = $this|0;
 $__str = $__str|0;
 $__pos = $__pos|0;
 $__n = $__n|0;
 $__a = $__a|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$__str>>0]|0;
 $1 = $0 & 1;
 $2 = ($1<<24>>24)==(0);
 $3 = ((($__str)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $0&255;
 $6 = $5 >>> 1;
 $7 = $2 ? $6 : $4;
 $8 = ($7>>>0)<($__pos>>>0);
 if ($8) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_out_of_rangeEv($this);
  // unreachable;
 } else {
  $9 = ((($__str)) + 8|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ((($__str)) + 1|0);
  $12 = $2 ? $11 : $10;
  $13 = (($12) + ($__pos)|0);
  $14 = (($7) - ($__pos))|0;
  $15 = ($14>>>0)<($__n>>>0);
  $16 = $15 ? $14 : $__n;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($this,$13,$16);
  return;
 }
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initIPKcEENS_9enable_ifIXsr21__is_forward_iteratorIT_EE5valueEvE4typeESA_SA_($this,$__first,$__last) {
 $this = $this|0;
 $__first = $__first|0;
 $__last = $__last|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $__p$0 = 0, $__p$0$ph = 0, $scevgep2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $__first;
 $1 = $__last;
 $2 = (($1) - ($0))|0;
 $3 = ($2>>>0)>(4294967279);
 if ($3) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_length_errorEv($this);
  // unreachable;
 }
 $4 = ($2>>>0)<(11);
 if ($4) {
  $5 = $2 << 1;
  $6 = $5&255;
  HEAP8[$this>>0] = $6;
  $7 = ((($this)) + 1|0);
  $__p$0$ph = $7;
 } else {
  $8 = (($2) + 16)|0;
  $9 = $8 & -16;
  $10 = (__Znwj($9)|0);
  $11 = ((($this)) + 8|0);
  HEAP32[$11>>2] = $10;
  $12 = $9 | 1;
  HEAP32[$this>>2] = $12;
  $13 = ((($this)) + 4|0);
  HEAP32[$13>>2] = $2;
  $__p$0$ph = $10;
 }
 $14 = $__last;
 $15 = (($14) - ($0))|0;
 $$0 = $__first;$__p$0 = $__p$0$ph;
 while(1) {
  $16 = ($$0|0)==($__last|0);
  if ($16) {
   break;
  }
  $17 = HEAP8[$$0>>0]|0;
  HEAP8[$__p$0>>0] = $17;
  $18 = ((($$0)) + 1|0);
  $19 = ((($__p$0)) + 1|0);
  $$0 = $18;$__p$0 = $19;
 }
 $scevgep2 = (($__p$0$ph) + ($15)|0);
 HEAP8[$scevgep2>>0] = 0;
 return;
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$this>>0]|0;
 $1 = $0 & 1;
 $2 = ($1<<24>>24)==(0);
 if (!($2)) {
  $3 = ((($this)) + 8|0);
  $4 = HEAP32[$3>>2]|0;
  __ZdlPv($4);
 }
 return;
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEaSERKS5_($this,$__str) {
 $this = $this|0;
 $__str = $__str|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($this|0)==($__str|0);
 if (!($0)) {
  $1 = HEAP8[$__str>>0]|0;
  $2 = $1 & 1;
  $3 = ($2<<24>>24)==(0);
  $4 = ((($__str)) + 8|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ((($__str)) + 1|0);
  $7 = $3 ? $6 : $5;
  $8 = ((($__str)) + 4|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = $1&255;
  $11 = $10 >>> 1;
  $12 = $3 ? $11 : $9;
  (__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKcj($this,$7,$12)|0);
 }
 return ($this|0);
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKcj($this,$__s,$__n) {
 $this = $this|0;
 $__s = $__s|0;
 $__n = $__n|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$this>>0]|0;
 $1 = $0 & 1;
 $2 = ($1<<24>>24)==(0);
 if ($2) {
  $7 = 10;$9 = $0;
 } else {
  $3 = HEAP32[$this>>2]|0;
  $4 = $3 & -2;
  $phitmp$i = (($4) + -1)|0;
  $5 = $3&255;
  $7 = $phitmp$i;$9 = $5;
 }
 $6 = ($7>>>0)<($__n>>>0);
 $8 = $9 & 1;
 $10 = ($8<<24>>24)==(0);
 do {
  if ($6) {
   if ($10) {
    $24 = $9&255;
    $25 = $24 >>> 1;
    $27 = $25;
   } else {
    $22 = ((($this)) + 4|0);
    $23 = HEAP32[$22>>2]|0;
    $27 = $23;
   }
   $26 = (($__n) - ($7))|0;
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE21__grow_by_and_replaceEjjjjjjPKc($this,$7,$26,$27,0,$27,$__n,$__s);
  } else {
   if ($10) {
    $13 = ((($this)) + 1|0);
    $14 = $13;
   } else {
    $11 = ((($this)) + 8|0);
    $12 = HEAP32[$11>>2]|0;
    $14 = $12;
   }
   _memmove(($14|0),($__s|0),($__n|0))|0;
   $15 = (($14) + ($__n)|0);
   HEAP8[$15>>0] = 0;
   $16 = HEAP8[$this>>0]|0;
   $17 = $16 & 1;
   $18 = ($17<<24>>24)==(0);
   if ($18) {
    $20 = $__n << 1;
    $21 = $20&255;
    HEAP8[$this>>0] = $21;
    break;
   } else {
    $19 = ((($this)) + 4|0);
    HEAP32[$19>>2] = $__n;
    break;
   }
  }
 } while(0);
 return ($this|0);
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE21__grow_by_and_replaceEjjjjjjPKc($this,$__old_cap,$__delta_cap,$__old_sz,$__n_copy,$__n_del,$__n_add,$__p_new_stuff) {
 $this = $this|0;
 $__old_cap = $__old_cap|0;
 $__delta_cap = $__delta_cap|0;
 $__old_sz = $__old_sz|0;
 $__n_copy = $__n_copy|0;
 $__n_del = $__n_del|0;
 $__n_add = $__n_add|0;
 $__p_new_stuff = $__p_new_stuff|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (-18 - ($__old_cap))|0;
 $1 = ($0>>>0)<($__delta_cap>>>0);
 if ($1) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_length_errorEv($this);
  // unreachable;
 }
 $2 = HEAP8[$this>>0]|0;
 $3 = $2 & 1;
 $4 = ($3<<24>>24)==(0);
 if ($4) {
  $7 = ((($this)) + 1|0);
  $20 = $7;
 } else {
  $5 = ((($this)) + 8|0);
  $6 = HEAP32[$5>>2]|0;
  $20 = $6;
 }
 $8 = ($__old_cap>>>0)<(2147483623);
 if ($8) {
  $9 = (($__delta_cap) + ($__old_cap))|0;
  $10 = $__old_cap << 1;
  $11 = ($9>>>0)<($10>>>0);
  $12 = $11 ? $10 : $9;
  $13 = ($12>>>0)<(11);
  $14 = (($12) + 16)|0;
  $15 = $14 & -16;
  $16 = $13 ? 11 : $15;
  $17 = $16;
 } else {
  $17 = -17;
 }
 $18 = (__Znwj($17)|0);
 $19 = ($__n_copy|0)==(0);
 if (!($19)) {
  _memcpy(($18|0),($20|0),($__n_copy|0))|0;
 }
 $21 = ($__n_add|0)==(0);
 if (!($21)) {
  $22 = (($18) + ($__n_copy)|0);
  _memcpy(($22|0),($__p_new_stuff|0),($__n_add|0))|0;
 }
 $23 = (($__old_sz) - ($__n_del))|0;
 $24 = ($23|0)==($__n_copy|0);
 if (!($24)) {
  $25 = (($23) - ($__n_copy))|0;
  $26 = (($18) + ($__n_copy)|0);
  $27 = (($26) + ($__n_add)|0);
  $28 = (($20) + ($__n_copy)|0);
  $29 = (($28) + ($__n_del)|0);
  _memcpy(($27|0),($29|0),($25|0))|0;
 }
 $30 = ($__old_cap|0)==(10);
 if (!($30)) {
  __ZdlPv($20);
 }
 $31 = ((($this)) + 8|0);
 HEAP32[$31>>2] = $18;
 $32 = $17 | 1;
 HEAP32[$this>>2] = $32;
 $33 = (($23) + ($__n_add))|0;
 $34 = ((($this)) + 4|0);
 HEAP32[$34>>2] = $33;
 $35 = (($18) + ($33)|0);
 HEAP8[$35>>0] = 0;
 return;
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7reserveEj($this,$__res_arg) {
 $this = $this|0;
 $__res_arg = $__res_arg|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__new_data$0 = 0, $__new_data$1 = 0, $__now_long$0$off0 = 0, $__p$0 = 0, $__was_long$0$off0 = 0, $phitmp$i = 0;
 var $phitmp$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__res_arg>>>0)>(4294967279);
 if ($0) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_length_errorEv($this);
  // unreachable;
 }
 $1 = HEAP8[$this>>0]|0;
 $2 = $1 & 1;
 $3 = ($2<<24>>24)==(0);
 if ($3) {
  $22 = 10;$8 = $1;
 } else {
  $4 = HEAP32[$this>>2]|0;
  $5 = $4 & -2;
  $phitmp$i = (($5) + -1)|0;
  $6 = $4&255;
  $22 = $phitmp$i;$8 = $6;
 }
 $7 = $8 & 1;
 $9 = ($7<<24>>24)==(0);
 if ($9) {
  $12 = $8&255;
  $13 = $12 >>> 1;
  $15 = $13;
 } else {
  $10 = ((($this)) + 4|0);
  $11 = HEAP32[$10>>2]|0;
  $15 = $11;
 }
 $14 = ($15>>>0)>($__res_arg>>>0);
 $16 = $14 ? $15 : $__res_arg;
 $17 = ($16>>>0)<(11);
 if ($17) {
  $21 = 10;
 } else {
  $18 = (($16) + 16)|0;
  $19 = $18 & -16;
  $phitmp$i2 = (($19) + -1)|0;
  $21 = $phitmp$i2;
 }
 $20 = ($21|0)==($22|0);
 L14: do {
  if (!($20)) {
   $23 = ($21|0)==(10);
   do {
    if ($23) {
     $24 = ((($this)) + 1|0);
     $25 = ((($this)) + 8|0);
     $26 = HEAP32[$25>>2]|0;
     $__new_data$1 = $24;$__now_long$0$off0 = 0;$__p$0 = $26;$__was_long$0$off0 = 1;
    } else {
     $27 = ($21>>>0)>($22>>>0);
     $28 = (($21) + 1)|0;
     if ($27) {
      $29 = (__Znwj($28)|0);
      $__new_data$0 = $29;
     } else {
      __THREW__ = 0;
      $30 = (invoke_ii(3348,($28|0))|0);
      $31 = __THREW__; __THREW__ = 0;
      $32 = $31&1;
      if ($32) {
       $33 = ___cxa_find_matching_catch_3(0|0)|0;
       $34 = tempRet0;
       (___cxa_begin_catch(($33|0))|0);
       ___cxa_end_catch();
       break L14;
      } else {
       $__new_data$0 = $30;
      }
     }
     $35 = $8 & 1;
     $36 = ($35<<24>>24)==(0);
     if ($36) {
      $39 = ((($this)) + 1|0);
      $__new_data$1 = $__new_data$0;$__now_long$0$off0 = 1;$__p$0 = $39;$__was_long$0$off0 = 0;
      break;
     } else {
      $37 = ((($this)) + 8|0);
      $38 = HEAP32[$37>>2]|0;
      $__new_data$1 = $__new_data$0;$__now_long$0$off0 = 1;$__p$0 = $38;$__was_long$0$off0 = 1;
      break;
     }
    }
   } while(0);
   $40 = $8 & 1;
   $41 = ($40<<24>>24)==(0);
   if ($41) {
    $44 = $8&255;
    $45 = $44 >>> 1;
    $47 = $45;
   } else {
    $42 = ((($this)) + 4|0);
    $43 = HEAP32[$42>>2]|0;
    $47 = $43;
   }
   $46 = (($47) + 1)|0;
   _memcpy(($__new_data$1|0),($__p$0|0),($46|0))|0;
   if ($__was_long$0$off0) {
    __ZdlPv($__p$0);
   }
   if ($__now_long$0$off0) {
    $48 = (($21) + 1)|0;
    $49 = $48 | 1;
    HEAP32[$this>>2] = $49;
    $50 = ((($this)) + 4|0);
    HEAP32[$50>>2] = $15;
    $51 = ((($this)) + 8|0);
    HEAP32[$51>>2] = $__new_data$1;
    break;
   } else {
    $52 = $15 << 1;
    $53 = $52&255;
    HEAP8[$this>>0] = $53;
    break;
   }
  }
 } while(0);
 return;
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc($this,$__s) {
 $this = $this|0;
 $__s = $__s|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strlen($__s)|0);
 $1 = (__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKcj($this,$__s,$0)|0);
 return ($1|0);
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6resizeEjc($this,$__n,$__c) {
 $this = $this|0;
 $__n = $__n|0;
 $__c = $__c|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$this>>0]|0;
 $1 = $0 & 1;
 $2 = ($1<<24>>24)==(0);
 if ($2) {
  $5 = $0&255;
  $6 = $5 >>> 1;
  $8 = $6;
 } else {
  $3 = ((($this)) + 4|0);
  $4 = HEAP32[$3>>2]|0;
  $8 = $4;
 }
 $7 = ($8>>>0)<($__n>>>0);
 do {
  if ($7) {
   $9 = (($__n) - ($8))|0;
   (__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6appendEjc($this,$9,$__c)|0);
  } else {
   if ($2) {
    $14 = (((($this)) + 1|0) + ($__n)|0);
    HEAP8[$14>>0] = 0;
    $15 = $__n << 1;
    $16 = $15&255;
    HEAP8[$this>>0] = $16;
    break;
   } else {
    $10 = ((($this)) + 8|0);
    $11 = HEAP32[$10>>2]|0;
    $12 = (($11) + ($__n)|0);
    HEAP8[$12>>0] = 0;
    $13 = ((($this)) + 4|0);
    HEAP32[$13>>2] = $__n;
    break;
   }
  }
 } while(0);
 return;
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6appendEjc($this,$__n,$__c) {
 $this = $this|0;
 $__n = $__n|0;
 $__c = $__c|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__n|0)==(0);
 if (!($0)) {
  $1 = HEAP8[$this>>0]|0;
  $2 = $1 & 1;
  $3 = ($2<<24>>24)==(0);
  if ($3) {
   $15 = 10;$8 = $1;
  } else {
   $4 = HEAP32[$this>>2]|0;
   $5 = $4 & -2;
   $phitmp$i = (($5) + -1)|0;
   $6 = $4&255;
   $15 = $phitmp$i;$8 = $6;
  }
  $7 = $8 & 1;
  $9 = ($7<<24>>24)==(0);
  if ($9) {
   $12 = $8&255;
   $13 = $12 >>> 1;
   $16 = $13;
  } else {
   $10 = ((($this)) + 4|0);
   $11 = HEAP32[$10>>2]|0;
   $16 = $11;
  }
  $14 = (($15) - ($16))|0;
  $17 = ($14>>>0)<($__n>>>0);
  if ($17) {
   $18 = (($__n) - ($15))|0;
   $19 = (($18) + ($16))|0;
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE9__grow_byEjjjjjj($this,$15,$19,$16,$16,0,0);
   $$pre = HEAP8[$this>>0]|0;
   $21 = $$pre;
  } else {
   $21 = $8;
  }
  $20 = $21 & 1;
  $22 = ($20<<24>>24)==(0);
  if ($22) {
   $25 = ((($this)) + 1|0);
   $27 = $25;
  } else {
   $23 = ((($this)) + 8|0);
   $24 = HEAP32[$23>>2]|0;
   $27 = $24;
  }
  $26 = (($27) + ($16)|0);
  _memset(($26|0),($__c|0),($__n|0))|0;
  $28 = (($16) + ($__n))|0;
  $29 = HEAP8[$this>>0]|0;
  $30 = $29 & 1;
  $31 = ($30<<24>>24)==(0);
  if ($31) {
   $33 = $28 << 1;
   $34 = $33&255;
   HEAP8[$this>>0] = $34;
  } else {
   $32 = ((($this)) + 4|0);
   HEAP32[$32>>2] = $28;
  }
  $35 = (($27) + ($28)|0);
  HEAP8[$35>>0] = 0;
 }
 return ($this|0);
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE9__grow_byEjjjjjj($this,$__old_cap,$__delta_cap,$__old_sz,$__n_copy,$__n_del,$__n_add) {
 $this = $this|0;
 $__old_cap = $__old_cap|0;
 $__delta_cap = $__delta_cap|0;
 $__old_sz = $__old_sz|0;
 $__n_copy = $__n_copy|0;
 $__n_del = $__n_del|0;
 $__n_add = $__n_add|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (-17 - ($__old_cap))|0;
 $1 = ($0>>>0)<($__delta_cap>>>0);
 if ($1) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_length_errorEv($this);
  // unreachable;
 }
 $2 = HEAP8[$this>>0]|0;
 $3 = $2 & 1;
 $4 = ($3<<24>>24)==(0);
 if ($4) {
  $7 = ((($this)) + 1|0);
  $20 = $7;
 } else {
  $5 = ((($this)) + 8|0);
  $6 = HEAP32[$5>>2]|0;
  $20 = $6;
 }
 $8 = ($__old_cap>>>0)<(2147483623);
 if ($8) {
  $9 = (($__delta_cap) + ($__old_cap))|0;
  $10 = $__old_cap << 1;
  $11 = ($9>>>0)<($10>>>0);
  $12 = $11 ? $10 : $9;
  $13 = ($12>>>0)<(11);
  $14 = (($12) + 16)|0;
  $15 = $14 & -16;
  $16 = $13 ? 11 : $15;
  $17 = $16;
 } else {
  $17 = -17;
 }
 $18 = (__Znwj($17)|0);
 $19 = ($__n_copy|0)==(0);
 if (!($19)) {
  _memcpy(($18|0),($20|0),($__n_copy|0))|0;
 }
 $21 = (($__old_sz) - ($__n_del))|0;
 $22 = ($21|0)==($__n_copy|0);
 if (!($22)) {
  $23 = (($21) - ($__n_copy))|0;
  $24 = (($18) + ($__n_copy)|0);
  $25 = (($24) + ($__n_add)|0);
  $26 = (($20) + ($__n_copy)|0);
  $27 = (($26) + ($__n_del)|0);
  _memcpy(($25|0),($27|0),($23|0))|0;
 }
 $28 = ($__old_cap|0)==(10);
 if (!($28)) {
  __ZdlPv($20);
 }
 $29 = ((($this)) + 8|0);
 HEAP32[$29>>2] = $18;
 $30 = $17 | 1;
 HEAP32[$this>>2] = $30;
 return;
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6appendEPKcj($this,$__s,$__n) {
 $this = $this|0;
 $__s = $__s|0;
 $__n = $__n|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$this>>0]|0;
 $1 = $0 & 1;
 $2 = ($1<<24>>24)==(0);
 if ($2) {
  $14 = 10;$7 = $0;
 } else {
  $3 = HEAP32[$this>>2]|0;
  $4 = $3 & -2;
  $phitmp$i = (($4) + -1)|0;
  $5 = $3&255;
  $14 = $phitmp$i;$7 = $5;
 }
 $6 = $7 & 1;
 $8 = ($6<<24>>24)==(0);
 if ($8) {
  $11 = $7&255;
  $12 = $11 >>> 1;
  $15 = $12;
 } else {
  $9 = ((($this)) + 4|0);
  $10 = HEAP32[$9>>2]|0;
  $15 = $10;
 }
 $13 = (($14) - ($15))|0;
 $16 = ($13>>>0)<($__n>>>0);
 if ($16) {
  $33 = (($__n) - ($14))|0;
  $34 = (($33) + ($15))|0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE21__grow_by_and_replaceEjjjjjjPKc($this,$14,$34,$15,$15,0,$__n,$__s);
 } else {
  $17 = ($__n|0)==(0);
  if (!($17)) {
   $18 = $7 & 1;
   $19 = ($18<<24>>24)==(0);
   if ($19) {
    $22 = ((($this)) + 1|0);
    $24 = $22;
   } else {
    $20 = ((($this)) + 8|0);
    $21 = HEAP32[$20>>2]|0;
    $24 = $21;
   }
   $23 = (($24) + ($15)|0);
   _memcpy(($23|0),($__s|0),($__n|0))|0;
   $25 = (($15) + ($__n))|0;
   $26 = HEAP8[$this>>0]|0;
   $27 = $26 & 1;
   $28 = ($27<<24>>24)==(0);
   if ($28) {
    $30 = $25 << 1;
    $31 = $30&255;
    HEAP8[$this>>0] = $31;
   } else {
    $29 = ((($this)) + 4|0);
    HEAP32[$29>>2] = $25;
   }
   $32 = (($24) + ($25)|0);
   HEAP8[$32>>0] = 0;
  }
 }
 return ($this|0);
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6appendEPKc($this,$__s) {
 $this = $this|0;
 $__s = $__s|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strlen($__s)|0);
 $1 = (__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6appendEPKcj($this,$__s,$0)|0);
 return ($1|0);
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE9push_backEc($this,$__c) {
 $this = $this|0;
 $__c = $__c|0;
 var $$pn = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__cap$0 = 0, $__p$0 = 0, $__sz$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$this>>0]|0;
 $1 = $0 & 1;
 $2 = ($1<<24>>24)!=(0);
 if ($2) {
  $5 = HEAP32[$this>>2]|0;
  $6 = $5 & -2;
  $7 = (($6) + -1)|0;
  $8 = ((($this)) + 4|0);
  $9 = HEAP32[$8>>2]|0;
  $__cap$0 = $7;$__sz$0 = $9;
 } else {
  $3 = $0&255;
  $4 = $3 >>> 1;
  $__cap$0 = 10;$__sz$0 = $4;
 }
 $10 = ($__sz$0|0)==($__cap$0|0);
 if ($10) {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE9__grow_byEjjjjjj($this,$__cap$0,1,$__cap$0,$__cap$0,0,0);
  $11 = HEAP8[$this>>0]|0;
  $12 = $11 & 1;
  $13 = ($12<<24>>24)==(0);
  if ($13) {
   label = 7;
  } else {
   label = 8;
  }
 } else {
  if ($2) {
   label = 8;
  } else {
   label = 7;
  }
 }
 if ((label|0) == 7) {
  $14 = ((($this)) + 1|0);
  $15 = $__sz$0 << 1;
  $16 = (($15) + 2)|0;
  $17 = $16&255;
  HEAP8[$this>>0] = $17;
  $$pn = $14;
 }
 else if ((label|0) == 8) {
  $18 = ((($this)) + 8|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = (($__sz$0) + 1)|0;
  $21 = ((($this)) + 4|0);
  HEAP32[$21>>2] = $20;
  $$pn = $19;
 }
 $__p$0 = (($$pn) + ($__sz$0)|0);
 HEAP8[$__p$0>>0] = $__c;
 $22 = ((($__p$0)) + 1|0);
 HEAP8[$22>>0] = 0;
 return;
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6insertEjPKcj($this,$__pos,$__s,$__n) {
 $this = $this|0;
 $__pos = $__pos|0;
 $__s = $__s|0;
 $__n = $__n|0;
 var $$0 = 0, $$1 = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$this>>0]|0;
 $1 = $0 & 1;
 $2 = ($1<<24>>24)==(0);
 if ($2) {
  $5 = $0&255;
  $6 = $5 >>> 1;
  $8 = $6;
 } else {
  $3 = ((($this)) + 4|0);
  $4 = HEAP32[$3>>2]|0;
  $8 = $4;
 }
 $7 = ($8>>>0)<($__pos>>>0);
 if ($7) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_out_of_rangeEv($this);
  // unreachable;
 }
 if ($2) {
  $13 = 10;$17 = $0;
 } else {
  $9 = HEAP32[$this>>2]|0;
  $10 = $9 & -2;
  $phitmp$i = (($10) + -1)|0;
  $11 = $9&255;
  $13 = $phitmp$i;$17 = $11;
 }
 $12 = (($13) - ($8))|0;
 $14 = ($12>>>0)<($__n>>>0);
 if ($14) {
  $39 = (($8) + ($__n))|0;
  $40 = (($39) - ($13))|0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE21__grow_by_and_replaceEjjjjjjPKc($this,$13,$40,$8,$__pos,0,$__n,$__s);
 } else {
  $15 = ($__n|0)==(0);
  if (!($15)) {
   $16 = $17 & 1;
   $18 = ($16<<24>>24)==(0);
   if ($18) {
    $21 = ((($this)) + 1|0);
    $23 = $21;
   } else {
    $19 = ((($this)) + 8|0);
    $20 = HEAP32[$19>>2]|0;
    $23 = $20;
   }
   $22 = ($8|0)==($__pos|0);
   if ($22) {
    $$pre = (($23) + ($__pos)|0);
    $$1 = $__s;$$pre$phiZ2D = $$pre;
   } else {
    $24 = (($8) - ($__pos))|0;
    $25 = (($23) + ($__pos)|0);
    $26 = ($25>>>0)<=($__s>>>0);
    $27 = (($23) + ($8)|0);
    $28 = ($27>>>0)>($__s>>>0);
    $or$cond = $26 & $28;
    $29 = (($__s) + ($__n)|0);
    $$0 = $or$cond ? $29 : $__s;
    $30 = (($25) + ($__n)|0);
    _memmove(($30|0),($25|0),($24|0))|0;
    $$1 = $$0;$$pre$phiZ2D = $25;
   }
   _memmove(($$pre$phiZ2D|0),($$1|0),($__n|0))|0;
   $31 = (($8) + ($__n))|0;
   $32 = HEAP8[$this>>0]|0;
   $33 = $32 & 1;
   $34 = ($33<<24>>24)==(0);
   if ($34) {
    $36 = $31 << 1;
    $37 = $36&255;
    HEAP8[$this>>0] = $37;
   } else {
    $35 = ((($this)) + 4|0);
    HEAP32[$35>>2] = $31;
   }
   $38 = (($23) + ($31)|0);
   HEAP8[$38>>0] = 0;
  }
 }
 return ($this|0);
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6insertEjPKc($this,$__pos,$__s) {
 $this = $this|0;
 $__pos = $__pos|0;
 $__s = $__s|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strlen($__s)|0);
 $1 = (__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6insertEjPKcj($this,$__pos,$__s,$0)|0);
 return ($1|0);
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE5eraseEjj($this,$__pos,$__n) {
 $this = $this|0;
 $__pos = $__pos|0;
 $__n = $__n|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$this>>0]|0;
 $1 = $0 & 1;
 $2 = ($1<<24>>24)==(0);
 if ($2) {
  $5 = $0&255;
  $6 = $5 >>> 1;
  $8 = $6;
 } else {
  $3 = ((($this)) + 4|0);
  $4 = HEAP32[$3>>2]|0;
  $8 = $4;
 }
 $7 = ($8>>>0)<($__pos>>>0);
 if ($7) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_out_of_rangeEv($this);
  // unreachable;
 }
 $9 = ($__n|0)==(0);
 if (!($9)) {
  if ($2) {
   $12 = ((($this)) + 1|0);
   $19 = $12;
  } else {
   $10 = ((($this)) + 8|0);
   $11 = HEAP32[$10>>2]|0;
   $19 = $11;
  }
  $13 = (($8) - ($__pos))|0;
  $14 = ($13>>>0)<($__n>>>0);
  $15 = $14 ? $13 : $__n;
  $16 = ($13|0)==($15|0);
  if ($16) {
   $23 = $0;
  } else {
   $17 = (($13) - ($15))|0;
   $18 = (($19) + ($__pos)|0);
   $20 = (($18) + ($15)|0);
   _memmove(($18|0),($20|0),($17|0))|0;
   $$pre = HEAP8[$this>>0]|0;
   $23 = $$pre;
  }
  $21 = (($8) - ($15))|0;
  $22 = $23 & 1;
  $24 = ($22<<24>>24)==(0);
  if ($24) {
   $26 = $21 << 1;
   $27 = $26&255;
   HEAP8[$this>>0] = $27;
  } else {
   $25 = ((($this)) + 4|0);
   HEAP32[$25>>2] = $21;
  }
  $28 = (($19) + ($21)|0);
  HEAP8[$28>>0] = 0;
 }
 return ($this|0);
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7replaceEjjPKcj($this,$__pos,$__n1,$__s,$__n2) {
 $this = $this|0;
 $__pos = $__pos|0;
 $__n1 = $__n1|0;
 $__s = $__s|0;
 $__n2 = $__n2|0;
 var $$01 = 0, $$011 = 0, $$02 = 0, $$06 = 0, $$212 = 0, $$3 = 0, $$313 = 0, $$35 = 0, $$39 = 0, $$4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$this>>0]|0;
 $1 = $0 & 1;
 $2 = ($1<<24>>24)==(0);
 if ($2) {
  $5 = $0&255;
  $6 = $5 >>> 1;
  $8 = $6;
 } else {
  $3 = ((($this)) + 4|0);
  $4 = HEAP32[$3>>2]|0;
  $8 = $4;
 }
 $7 = ($8>>>0)<($__pos>>>0);
 if ($7) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_out_of_rangeEv($this);
  // unreachable;
 }
 $9 = (($8) - ($__pos))|0;
 $10 = ($9>>>0)<($__n1>>>0);
 $11 = $10 ? $9 : $__n1;
 if ($2) {
  $17 = 10;$20 = $0;
 } else {
  $12 = HEAP32[$this>>2]|0;
  $13 = $12 & -2;
  $phitmp$i = (($13) + -1)|0;
  $14 = $12&255;
  $17 = $phitmp$i;$20 = $14;
 }
 $15 = (($11) - ($8))|0;
 $16 = (($15) + ($17))|0;
 $18 = ($16>>>0)<($__n2>>>0);
 if ($18) {
  $56 = (($8) + ($__n2))|0;
  $57 = (($56) - ($11))|0;
  $58 = (($57) - ($17))|0;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE21__grow_by_and_replaceEjjjjjjPKc($this,$17,$58,$8,$__pos,$11,$__n2,$__s);
 } else {
  $19 = $20 & 1;
  $21 = ($19<<24>>24)==(0);
  if ($21) {
   $24 = ((($this)) + 1|0);
   $30 = $24;
  } else {
   $22 = ((($this)) + 8|0);
   $23 = HEAP32[$22>>2]|0;
   $30 = $23;
  }
  $25 = ($11|0)==($__n2|0);
  do {
   if ($25) {
    $$212 = $__n2;$$3 = $__pos;$$35 = $__s;$$39 = $__n2;
    label = 21;
   } else {
    $26 = (($9) - ($11))|0;
    $27 = ($9|0)==($11|0);
    if ($27) {
     $$212 = $9;$$3 = $__pos;$$35 = $__s;$$39 = $__n2;
     label = 21;
    } else {
     $28 = ($11>>>0)>($__n2>>>0);
     $29 = (($30) + ($__pos)|0);
     if ($28) {
      _memmove(($29|0),($__s|0),($__n2|0))|0;
      $44 = (($29) + ($__n2)|0);
      $45 = (($29) + ($11)|0);
      _memmove(($44|0),($45|0),($26|0))|0;
      $$313 = $11;$$4 = $__n2;
      break;
     }
     $31 = ($29>>>0)<($__s>>>0);
     $32 = (($30) + ($8)|0);
     $33 = ($32>>>0)>($__s>>>0);
     $or$cond = $31 & $33;
     do {
      if ($or$cond) {
       $34 = (($29) + ($11)|0);
       $35 = ($34>>>0)>($__s>>>0);
       if ($35) {
        _memmove(($29|0),($__s|0),($11|0))|0;
        $38 = (($11) + ($__pos))|0;
        $39 = (($__s) + ($__n2)|0);
        $40 = (($__n2) - ($11))|0;
        $$01 = $38;$$011 = 0;$$02 = $39;$$06 = $40;
        break;
       } else {
        $36 = (($__n2) - ($11))|0;
        $37 = (($__s) + ($36)|0);
        $$01 = $__pos;$$011 = $11;$$02 = $37;$$06 = $__n2;
        break;
       }
      } else {
       $$01 = $__pos;$$011 = $11;$$02 = $__s;$$06 = $__n2;
      }
     } while(0);
     $41 = (($30) + ($$01)|0);
     $42 = (($41) + ($$06)|0);
     $43 = (($41) + ($$011)|0);
     _memmove(($42|0),($43|0),($26|0))|0;
     $$212 = $$011;$$3 = $$01;$$35 = $$02;$$39 = $$06;
     label = 21;
    }
   }
  } while(0);
  if ((label|0) == 21) {
   $46 = (($30) + ($$3)|0);
   _memmove(($46|0),($$35|0),($$39|0))|0;
   $$313 = $$212;$$4 = $$39;
  }
  $47 = (($$4) - ($$313))|0;
  $48 = (($47) + ($8))|0;
  $49 = HEAP8[$this>>0]|0;
  $50 = $49 & 1;
  $51 = ($50<<24>>24)==(0);
  if ($51) {
   $53 = $48 << 1;
   $54 = $53&255;
   HEAP8[$this>>0] = $54;
  } else {
   $52 = ((($this)) + 4|0);
   HEAP32[$52>>2] = $48;
  }
  $55 = (($30) + ($48)|0);
  HEAP8[$55>>0] = 0;
 }
 return ($this|0);
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7replaceEjjPKc($this,$__pos,$__n1,$__s) {
 $this = $this|0;
 $__pos = $__pos|0;
 $__n1 = $__n1|0;
 $__s = $__s|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strlen($__s)|0);
 $1 = (__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7replaceEjjPKcj($this,$__pos,$__n1,$__s,$0)|0);
 return ($1|0);
}
function __ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE4findEcj($this,$__c,$__pos) {
 $this = $this|0;
 $__c = $__c|0;
 $__pos = $__pos|0;
 var $$0$i = 0, $$1$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$this>>0]|0;
 $1 = $0 & 1;
 $2 = ($1<<24>>24)==(0);
 if ($2) {
  $7 = ((($this)) + 1|0);
  $8 = $0&255;
  $9 = $8 >>> 1;
  $11 = $9;$13 = $7;
 } else {
  $3 = ((($this)) + 8|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ((($this)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  $11 = $6;$13 = $4;
 }
 $10 = ($11>>>0)>($__pos>>>0);
 if ($10) {
  $12 = (($13) + ($__pos)|0);
  $14 = (($11) - ($__pos))|0;
  $15 = $__c&255;
  $16 = (_memchr($12,$15,$14)|0);
  $17 = ($16|0)==(0|0);
  $18 = $16;
  $19 = $13;
  $20 = (($18) - ($19))|0;
  $$0$i = $17 ? -1 : $20;
  $$1$i = $$0$i;
 } else {
  $$1$i = -1;
 }
 return ($$1$i|0);
}
function __ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE5rfindEcj($this,$__c,$__pos) {
 $this = $this|0;
 $__c = $__c|0;
 $__pos = $__pos|0;
 var $$01$i = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__ps$0$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$this>>0]|0;
 $1 = $0 & 1;
 $2 = ($1<<24>>24)==(0);
 if ($2) {
  $7 = ((($this)) + 1|0);
  $8 = $0&255;
  $9 = $8 >>> 1;
  $11 = $9;$15 = $7;
 } else {
  $3 = ((($this)) + 8|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ((($this)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  $11 = $6;$15 = $4;
 }
 $10 = ($11|0)==(0);
 L5: do {
  if ($10) {
   $23 = -1;
  } else {
   $12 = ($11>>>0)>($__pos>>>0);
   $13 = (($__pos) + 1)|0;
   $$01$i = $12 ? $13 : $11;
   $14 = (($15) + ($$01$i)|0);
   $__ps$0$i = $14;
   while(1) {
    $16 = ($__ps$0$i|0)==($15|0);
    if ($16) {
     $23 = -1;
     break L5;
    }
    $17 = ((($__ps$0$i)) + -1|0);
    $18 = HEAP8[$17>>0]|0;
    $19 = ($18<<24>>24)==($__c<<24>>24);
    if ($19) {
     $$lcssa = $17;
     break;
    } else {
     $__ps$0$i = $17;
    }
   }
   $20 = $$lcssa;
   $21 = $15;
   $22 = (($20) - ($21))|0;
   $23 = $22;
  }
 } while(0);
 return ($23|0);
}
function __ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7compareEjjPKcj($this,$__pos1,$__n1,$__s,$__n2) {
 $this = $this|0;
 $__pos1 = $__pos1|0;
 $__n1 = $__n1|0;
 $__s = $__s|0;
 $__n2 = $__n2|0;
 var $$ = 0, $$$ = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$this>>0]|0;
 $1 = $0 & 1;
 $2 = ($1<<24>>24)==(0);
 if ($2) {
  $5 = $0&255;
  $6 = $5 >>> 1;
  $8 = $6;
 } else {
  $3 = ((($this)) + 4|0);
  $4 = HEAP32[$3>>2]|0;
  $8 = $4;
 }
 $7 = ($8>>>0)<($__pos1>>>0);
 $9 = ($__n2|0)==(-1);
 $or$cond = $9 | $7;
 if ($or$cond) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_out_of_rangeEv($this);
  // unreachable;
 }
 $10 = (($8) - ($__pos1))|0;
 $11 = ($10>>>0)<($__n1>>>0);
 $12 = $11 ? $10 : $__n1;
 if ($2) {
  $15 = ((($this)) + 1|0);
  $17 = $15;
 } else {
  $13 = ((($this)) + 8|0);
  $14 = HEAP32[$13>>2]|0;
  $17 = $14;
 }
 $16 = (($17) + ($__pos1)|0);
 $18 = ($12>>>0)>($__n2>>>0);
 $19 = $18 ? $__n2 : $12;
 $20 = (_memcmp($16,$__s,$19)|0);
 $21 = ($20|0)==(0);
 if ($21) {
  $22 = ($12>>>0)<($__n2>>>0);
  $$ = $18&1;
  $$$ = $22 ? -1 : $$;
  return ($$$|0);
 } else {
  return ($20|0);
 }
 return (0)|0;
}
function __ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7compareEPKc($this,$__s) {
 $this = $this|0;
 $__s = $__s|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strlen($__s)|0);
 __THREW__ = 0;
 $1 = (invoke_iiiiii(4856,($this|0),0,-1,($__s|0),($0|0))|0);
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  ___clang_call_terminate($4);
  // unreachable;
 } else {
  return ($1|0);
 }
 return (0)|0;
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcjj($this,$__s,$__sz,$__reserve) {
 $this = $this|0;
 $__s = $__s|0;
 $__sz = $__sz|0;
 $__reserve = $__reserve|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__p$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__reserve>>>0)>(4294967279);
 if ($0) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_length_errorEv($this);
  // unreachable;
 }
 $1 = ($__reserve>>>0)<(11);
 if ($1) {
  $2 = $__sz << 1;
  $3 = $2&255;
  HEAP8[$this>>0] = $3;
  $4 = ((($this)) + 1|0);
  $__p$0 = $4;
 } else {
  $5 = (($__reserve) + 16)|0;
  $6 = $5 & -16;
  $7 = (__Znwj($6)|0);
  $8 = ((($this)) + 8|0);
  HEAP32[$8>>2] = $7;
  $9 = $6 | 1;
  HEAP32[$this>>2] = $9;
  $10 = ((($this)) + 4|0);
  HEAP32[$10>>2] = $__sz;
  $__p$0 = $7;
 }
 _memcpy(($__p$0|0),($__s|0),($__sz|0))|0;
 $11 = (($__p$0) + ($__sz)|0);
 HEAP8[$11>>0] = 0;
 return;
}
function __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6__initEPKwj($this,$__s,$__sz) {
 $this = $this|0;
 $__s = $__s|0;
 $__sz = $__sz|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__p$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__sz>>>0)>(1073741807);
 if ($0) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_length_errorEv($this);
  // unreachable;
 }
 $1 = ($__sz>>>0)<(2);
 if ($1) {
  $2 = $__sz << 1;
  $3 = $2&255;
  HEAP8[$this>>0] = $3;
  $4 = ((($this)) + 4|0);
  $__p$0 = $4;
 } else {
  $5 = (($__sz) + 4)|0;
  $6 = $5 & -4;
  $7 = $6 << 2;
  $8 = (__Znwj($7)|0);
  $9 = ((($this)) + 8|0);
  HEAP32[$9>>2] = $8;
  $10 = $6 | 1;
  HEAP32[$this>>2] = $10;
  $11 = ((($this)) + 4|0);
  HEAP32[$11>>2] = $__sz;
  $__p$0 = $8;
 }
 (_wmemcpy($__p$0,$__s,$__sz)|0);
 $12 = (($__p$0) + ($__sz<<2)|0);
 HEAP32[$12>>2] = 0;
 return;
}
function __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6__initEjw($this,$__n,$__c) {
 $this = $this|0;
 $__n = $__n|0;
 $__c = $__c|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__p$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__n>>>0)>(1073741807);
 if ($0) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_length_errorEv($this);
  // unreachable;
 }
 $1 = ($__n>>>0)<(2);
 if ($1) {
  $2 = $__n << 1;
  $3 = $2&255;
  HEAP8[$this>>0] = $3;
  $4 = ((($this)) + 4|0);
  $__p$0 = $4;
 } else {
  $5 = (($__n) + 4)|0;
  $6 = $5 & -4;
  $7 = $6 << 2;
  $8 = (__Znwj($7)|0);
  $9 = ((($this)) + 8|0);
  HEAP32[$9>>2] = $8;
  $10 = $6 | 1;
  HEAP32[$this>>2] = $10;
  $11 = ((($this)) + 4|0);
  HEAP32[$11>>2] = $__n;
  $__p$0 = $8;
 }
 (_wmemset($__p$0,$__c,$__n)|0);
 $12 = (($__p$0) + ($__n<<2)|0);
 HEAP32[$12>>2] = 0;
 return;
}
function __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6__initIPKwEENS_9enable_ifIXsr21__is_forward_iteratorIT_EE5valueEvE4typeESA_SA_($this,$__first,$__last) {
 $this = $this|0;
 $__first = $__first|0;
 $__last = $__last|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__p$0 = 0, $__p$0$ph = 0, $scevgep = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $__first;
 $1 = $__last;
 $2 = (($1) - ($0))|0;
 $3 = $2 >> 2;
 $4 = ($3>>>0)>(1073741807);
 if ($4) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_length_errorEv($this);
  // unreachable;
 }
 $5 = ($3>>>0)<(2);
 if ($5) {
  $6 = $2 >>> 1;
  $7 = $6&255;
  HEAP8[$this>>0] = $7;
  $8 = ((($this)) + 4|0);
  $__p$0$ph = $8;
 } else {
  $9 = (($3) + 4)|0;
  $10 = $9 & -4;
  $11 = $10 << 2;
  $12 = (__Znwj($11)|0);
  $13 = ((($this)) + 8|0);
  HEAP32[$13>>2] = $12;
  $14 = $10 | 1;
  HEAP32[$this>>2] = $14;
  $15 = ((($this)) + 4|0);
  HEAP32[$15>>2] = $3;
  $__p$0$ph = $12;
 }
 $16 = $__last;
 $17 = (($16) - ($0))|0;
 $18 = $17 >>> 2;
 $$0 = $__first;$__p$0 = $__p$0$ph;
 while(1) {
  $19 = ($$0|0)==($__last|0);
  if ($19) {
   break;
  }
  $20 = HEAP32[$$0>>2]|0;
  HEAP32[$__p$0>>2] = $20;
  $21 = ((($$0)) + 4|0);
  $22 = ((($__p$0)) + 4|0);
  $$0 = $21;$__p$0 = $22;
 }
 $scevgep = (($__p$0$ph) + ($18<<2)|0);
 HEAP32[$scevgep>>2] = 0;
 return;
}
function __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$this>>0]|0;
 $1 = $0 & 1;
 $2 = ($1<<24>>24)==(0);
 if (!($2)) {
  $3 = ((($this)) + 8|0);
  $4 = HEAP32[$3>>2]|0;
  __ZdlPv($4);
 }
 return;
}
function __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKwj($this,$__s,$__n) {
 $this = $this|0;
 $__s = $__s|0;
 $__n = $__n|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$this>>0]|0;
 $1 = $0 & 1;
 $2 = ($1<<24>>24)==(0);
 if ($2) {
  $7 = 1;$9 = $0;
 } else {
  $3 = HEAP32[$this>>2]|0;
  $4 = $3 & -2;
  $phitmp$i = (($4) + -1)|0;
  $5 = $3&255;
  $7 = $phitmp$i;$9 = $5;
 }
 $6 = ($7>>>0)<($__n>>>0);
 $8 = $9 & 1;
 $10 = ($8<<24>>24)==(0);
 do {
  if ($6) {
   if ($10) {
    $24 = $9&255;
    $25 = $24 >>> 1;
    $27 = $25;
   } else {
    $22 = ((($this)) + 4|0);
    $23 = HEAP32[$22>>2]|0;
    $27 = $23;
   }
   $26 = (($__n) - ($7))|0;
   __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE21__grow_by_and_replaceEjjjjjjPKw($this,$7,$26,$27,0,$27,$__n,$__s);
  } else {
   if ($10) {
    $13 = ((($this)) + 4|0);
    $14 = $13;
   } else {
    $11 = ((($this)) + 8|0);
    $12 = HEAP32[$11>>2]|0;
    $14 = $12;
   }
   (_wmemmove($14,$__s,$__n)|0);
   $15 = (($14) + ($__n<<2)|0);
   HEAP32[$15>>2] = 0;
   $16 = HEAP8[$this>>0]|0;
   $17 = $16 & 1;
   $18 = ($17<<24>>24)==(0);
   if ($18) {
    $20 = $__n << 1;
    $21 = $20&255;
    HEAP8[$this>>0] = $21;
    break;
   } else {
    $19 = ((($this)) + 4|0);
    HEAP32[$19>>2] = $__n;
    break;
   }
  }
 } while(0);
 return ($this|0);
}
function __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE21__grow_by_and_replaceEjjjjjjPKw($this,$__old_cap,$__delta_cap,$__old_sz,$__n_copy,$__n_del,$__n_add,$__p_new_stuff) {
 $this = $this|0;
 $__old_cap = $__old_cap|0;
 $__delta_cap = $__delta_cap|0;
 $__old_sz = $__old_sz|0;
 $__n_copy = $__n_copy|0;
 $__n_del = $__n_del|0;
 $__n_add = $__n_add|0;
 $__p_new_stuff = $__p_new_stuff|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (1073741806 - ($__old_cap))|0;
 $1 = ($0>>>0)<($__delta_cap>>>0);
 if ($1) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_length_errorEv($this);
  // unreachable;
 }
 $2 = HEAP8[$this>>0]|0;
 $3 = $2 & 1;
 $4 = ($3<<24>>24)==(0);
 if ($4) {
  $7 = ((($this)) + 4|0);
  $21 = $7;
 } else {
  $5 = ((($this)) + 8|0);
  $6 = HEAP32[$5>>2]|0;
  $21 = $6;
 }
 $8 = ($__old_cap>>>0)<(536870887);
 if ($8) {
  $9 = (($__delta_cap) + ($__old_cap))|0;
  $10 = $__old_cap << 1;
  $11 = ($9>>>0)<($10>>>0);
  $12 = $11 ? $10 : $9;
  $13 = ($12>>>0)<(2);
  $14 = (($12) + 4)|0;
  $15 = $14 & -4;
  $16 = $13 ? 2 : $15;
  $18 = $16;
 } else {
  $18 = 1073741807;
 }
 $17 = $18 << 2;
 $19 = (__Znwj($17)|0);
 $20 = ($__n_copy|0)==(0);
 if (!($20)) {
  (_wmemcpy($19,$21,$__n_copy)|0);
 }
 $22 = ($__n_add|0)==(0);
 if (!($22)) {
  $23 = (($19) + ($__n_copy<<2)|0);
  (_wmemcpy($23,$__p_new_stuff,$__n_add)|0);
 }
 $24 = (($__old_sz) - ($__n_del))|0;
 $25 = ($24|0)==($__n_copy|0);
 if (!($25)) {
  $26 = (($24) - ($__n_copy))|0;
  $27 = (($19) + ($__n_copy<<2)|0);
  $28 = (($27) + ($__n_add<<2)|0);
  $29 = (($21) + ($__n_copy<<2)|0);
  $30 = (($29) + ($__n_del<<2)|0);
  (_wmemcpy($28,$30,$26)|0);
 }
 $31 = ($__old_cap|0)==(1);
 if (!($31)) {
  __ZdlPv($21);
 }
 $32 = ((($this)) + 8|0);
 HEAP32[$32>>2] = $19;
 $33 = $18 | 1;
 HEAP32[$this>>2] = $33;
 $34 = (($24) + ($__n_add))|0;
 $35 = ((($this)) + 4|0);
 HEAP32[$35>>2] = $34;
 $36 = (($19) + ($34<<2)|0);
 HEAP32[$36>>2] = 0;
 return;
}
function __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE7reserveEj($this,$__res_arg) {
 $this = $this|0;
 $__res_arg = $__res_arg|0;
 var $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__new_data$1 = 0, $__now_long$0$off0 = 0, $__p$0 = 0, $__was_long$0$off0 = 0;
 var $phitmp$i = 0, $phitmp$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__res_arg>>>0)>(1073741807);
 if ($0) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_length_errorEv($this);
  // unreachable;
 }
 $1 = HEAP8[$this>>0]|0;
 $2 = $1 & 1;
 $3 = ($2<<24>>24)==(0);
 if ($3) {
  $22 = 1;$8 = $1;
 } else {
  $4 = HEAP32[$this>>2]|0;
  $5 = $4 & -2;
  $phitmp$i = (($5) + -1)|0;
  $6 = $4&255;
  $22 = $phitmp$i;$8 = $6;
 }
 $7 = $8 & 1;
 $9 = ($7<<24>>24)==(0);
 if ($9) {
  $12 = $8&255;
  $13 = $12 >>> 1;
  $15 = $13;
 } else {
  $10 = ((($this)) + 4|0);
  $11 = HEAP32[$10>>2]|0;
  $15 = $11;
 }
 $14 = ($15>>>0)>($__res_arg>>>0);
 $16 = $14 ? $15 : $__res_arg;
 $17 = ($16>>>0)<(2);
 if ($17) {
  $21 = 1;
 } else {
  $18 = (($16) + 4)|0;
  $19 = $18 & -4;
  $phitmp$i2 = (($19) + -1)|0;
  $21 = $phitmp$i2;
 }
 $20 = ($21|0)==($22|0);
 L14: do {
  if (!($20)) {
   $23 = ($21|0)==(1);
   do {
    if ($23) {
     $24 = ((($this)) + 4|0);
     $25 = ((($this)) + 8|0);
     $26 = HEAP32[$25>>2]|0;
     $__new_data$1 = $24;$__now_long$0$off0 = 0;$__p$0 = $26;$__was_long$0$off0 = 1;
    } else {
     $27 = ($21>>>0)>($22>>>0);
     $28 = $21 << 2;
     $29 = (($28) + 4)|0;
     if ($27) {
      $30 = (__Znwj($29)|0);
      $$sink = $30;
     } else {
      __THREW__ = 0;
      $31 = (invoke_ii(3348,($29|0))|0);
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if ($33) {
       $34 = ___cxa_find_matching_catch_3(0|0)|0;
       $35 = tempRet0;
       (___cxa_begin_catch(($34|0))|0);
       ___cxa_end_catch();
       break L14;
      } else {
       $$sink = $31;
      }
     }
     $36 = $8 & 1;
     $37 = ($36<<24>>24)==(0);
     if ($37) {
      $40 = ((($this)) + 4|0);
      $__new_data$1 = $$sink;$__now_long$0$off0 = 1;$__p$0 = $40;$__was_long$0$off0 = 0;
      break;
     } else {
      $38 = ((($this)) + 8|0);
      $39 = HEAP32[$38>>2]|0;
      $__new_data$1 = $$sink;$__now_long$0$off0 = 1;$__p$0 = $39;$__was_long$0$off0 = 1;
      break;
     }
    }
   } while(0);
   $41 = $8 & 1;
   $42 = ($41<<24>>24)==(0);
   if ($42) {
    $45 = $8&255;
    $46 = $45 >>> 1;
    $48 = $46;
   } else {
    $43 = ((($this)) + 4|0);
    $44 = HEAP32[$43>>2]|0;
    $48 = $44;
   }
   $47 = (($48) + 1)|0;
   (_wmemcpy($__new_data$1,$__p$0,$47)|0);
   if ($__was_long$0$off0) {
    __ZdlPv($__p$0);
   }
   if ($__now_long$0$off0) {
    $49 = (($21) + 1)|0;
    $50 = $49 | 1;
    HEAP32[$this>>2] = $50;
    $51 = ((($this)) + 4|0);
    HEAP32[$51>>2] = $15;
    $52 = ((($this)) + 8|0);
    HEAP32[$52>>2] = $__new_data$1;
    break;
   } else {
    $53 = $15 << 1;
    $54 = $53&255;
    HEAP8[$this>>0] = $54;
    break;
   }
  }
 } while(0);
 return;
}
function __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw($this,$__s) {
 $this = $this|0;
 $__s = $__s|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_wcslen($__s)|0);
 $1 = (__ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKwj($this,$__s,$0)|0);
 return ($1|0);
}
function __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE9__grow_byEjjjjjj($this,$__old_cap,$__delta_cap,$__old_sz,$__n_copy,$__n_del,$__n_add) {
 $this = $this|0;
 $__old_cap = $__old_cap|0;
 $__delta_cap = $__delta_cap|0;
 $__old_sz = $__old_sz|0;
 $__n_copy = $__n_copy|0;
 $__n_del = $__n_del|0;
 $__n_add = $__n_add|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (1073741807 - ($__old_cap))|0;
 $1 = ($0>>>0)<($__delta_cap>>>0);
 if ($1) {
  __ZNKSt3__121__basic_string_commonILb1EE20__throw_length_errorEv($this);
  // unreachable;
 }
 $2 = HEAP8[$this>>0]|0;
 $3 = $2 & 1;
 $4 = ($3<<24>>24)==(0);
 if ($4) {
  $7 = ((($this)) + 4|0);
  $21 = $7;
 } else {
  $5 = ((($this)) + 8|0);
  $6 = HEAP32[$5>>2]|0;
  $21 = $6;
 }
 $8 = ($__old_cap>>>0)<(536870887);
 if ($8) {
  $9 = (($__delta_cap) + ($__old_cap))|0;
  $10 = $__old_cap << 1;
  $11 = ($9>>>0)<($10>>>0);
  $12 = $11 ? $10 : $9;
  $13 = ($12>>>0)<(2);
  $14 = (($12) + 4)|0;
  $15 = $14 & -4;
  $16 = $13 ? 2 : $15;
  $18 = $16;
 } else {
  $18 = 1073741807;
 }
 $17 = $18 << 2;
 $19 = (__Znwj($17)|0);
 $20 = ($__n_copy|0)==(0);
 if (!($20)) {
  (_wmemcpy($19,$21,$__n_copy)|0);
 }
 $22 = (($__old_sz) - ($__n_del))|0;
 $23 = ($22|0)==($__n_copy|0);
 if (!($23)) {
  $24 = (($22) - ($__n_copy))|0;
  $25 = (($19) + ($__n_copy<<2)|0);
  $26 = (($25) + ($__n_add<<2)|0);
  $27 = (($21) + ($__n_copy<<2)|0);
  $28 = (($27) + ($__n_del<<2)|0);
  (_wmemcpy($26,$28,$24)|0);
 }
 $29 = ($__old_cap|0)==(1);
 if (!($29)) {
  __ZdlPv($21);
 }
 $30 = ((($this)) + 8|0);
 HEAP32[$30>>2] = $19;
 $31 = $18 | 1;
 HEAP32[$this>>2] = $31;
 return;
}
function __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE9push_backEw($this,$__c) {
 $this = $this|0;
 $__c = $__c|0;
 var $$pn = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__cap$0 = 0, $__p$0 = 0, $__sz$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$this>>0]|0;
 $1 = $0 & 1;
 $2 = ($1<<24>>24)!=(0);
 if ($2) {
  $5 = HEAP32[$this>>2]|0;
  $6 = $5 & -2;
  $7 = (($6) + -1)|0;
  $8 = ((($this)) + 4|0);
  $9 = HEAP32[$8>>2]|0;
  $__cap$0 = $7;$__sz$0 = $9;
 } else {
  $3 = $0&255;
  $4 = $3 >>> 1;
  $__cap$0 = 1;$__sz$0 = $4;
 }
 $10 = ($__sz$0|0)==($__cap$0|0);
 if ($10) {
  __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE9__grow_byEjjjjjj($this,$__cap$0,1,$__cap$0,$__cap$0,0,0);
  $11 = HEAP8[$this>>0]|0;
  $12 = $11 & 1;
  $13 = ($12<<24>>24)==(0);
  if ($13) {
   label = 7;
  } else {
   label = 8;
  }
 } else {
  if ($2) {
   label = 8;
  } else {
   label = 7;
  }
 }
 if ((label|0) == 7) {
  $14 = ((($this)) + 4|0);
  $15 = $__sz$0 << 1;
  $16 = (($15) + 2)|0;
  $17 = $16&255;
  HEAP8[$this>>0] = $17;
  $$pn = $14;
 }
 else if ((label|0) == 8) {
  $18 = ((($this)) + 8|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = (($__sz$0) + 1)|0;
  $21 = ((($this)) + 4|0);
  HEAP32[$21>>2] = $20;
  $$pn = $19;
 }
 $__p$0 = (($$pn) + ($__sz$0<<2)|0);
 HEAP32[$__p$0>>2] = $__c;
 $22 = ((($__p$0)) + 4|0);
 HEAP32[$22>>2] = 0;
 return;
}
function __ZNSt3__19basic_iosIcNS_11char_traitsIcEEED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__18ios_baseD2Ev($this);
 return;
}
function __ZNSt3__18ios_baseD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (154600);
 __THREW__ = 0;
 invoke_vii(4857,($this|0),0);
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 if ($1) {
  $11 = ___cxa_find_matching_catch_3(0|0)|0;
  $12 = tempRet0;
  ___clang_call_terminate($11);
  // unreachable;
 } else {
  $2 = ((($this)) + 28|0);
  __ZNSt3__16localeD2Ev($2);
  $3 = ((($this)) + 32|0);
  $4 = HEAP32[$3>>2]|0;
  _free($4);
  $5 = ((($this)) + 36|0);
  $6 = HEAP32[$5>>2]|0;
  _free($6);
  $7 = ((($this)) + 48|0);
  $8 = HEAP32[$7>>2]|0;
  _free($8);
  $9 = ((($this)) + 60|0);
  $10 = HEAP32[$9>>2]|0;
  _free($10);
  return;
 }
}
function __ZNSt3__18ios_baseD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__18ios_baseD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNSt3__18ios_base16__call_callbacksENS0_5eventE($this,$ev) {
 $this = $this|0;
 $ev = $ev|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($this)) + 32|0);
 $3 = ((($this)) + 36|0);
 $i$0 = $1;
 while(1) {
  $4 = ($i$0|0)==(0);
  if ($4) {
   break;
  }
  $5 = (($i$0) + -1)|0;
  $6 = HEAP32[$2>>2]|0;
  $7 = (($6) + ($5<<2)|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = HEAP32[$3>>2]|0;
  $10 = (($9) + ($5<<2)|0);
  $11 = HEAP32[$10>>2]|0;
  FUNCTION_TABLE_viii[$8 & 8191]($ev,$this,$11);
  $i$0 = $5;
 }
 return;
}
function __ZNSt3__19basic_iosIwNS_11char_traitsIwEEED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__18ios_baseD2Ev($this);
 return;
}
function __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (154616);
 $0 = ((($this)) + 4|0);
 __ZNSt3__16localeD2Ev($0);
 return;
}
function __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (154616);
 $0 = ((($this)) + 4|0);
 __ZNSt3__16localeD2Ev($0);
 __ZdlPv($this);
 return;
}
function __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE5imbueERKNS_6localeE($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE6setbufEPci($this,$0,$1) {
 $this = $this|0;
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ($this|0);
}
function __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE7seekoffExNS_8ios_base7seekdirEj($agg$result,$this,$0,$1,$2,$3) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = $agg$result;
 $5 = $4;
 HEAP32[$5>>2] = 0;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = ((($agg$result)) + 8|0);
 $9 = $8;
 $10 = $9;
 HEAP32[$10>>2] = -1;
 $11 = (($9) + 4)|0;
 $12 = $11;
 HEAP32[$12>>2] = -1;
 return;
}
function __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE7seekposENS_4fposI11__mbstate_tEEj($agg$result,$this,$0,$1) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = $agg$result;
 $3 = $2;
 HEAP32[$3>>2] = 0;
 $4 = (($2) + 4)|0;
 $5 = $4;
 HEAP32[$5>>2] = 0;
 $6 = ((($agg$result)) + 8|0);
 $7 = $6;
 $8 = $7;
 HEAP32[$8>>2] = -1;
 $9 = (($7) + 4)|0;
 $10 = $9;
 HEAP32[$10>>2] = -1;
 return;
}
function __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE4syncEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE9showmanycEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE6xsgetnEPci($this,$__s,$__n) {
 $this = $this|0;
 $__s = $__s|0;
 $__n = $__n|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__i$0 = 0, $__i$0$lcssa = 0, $storemerge = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 $1 = ((($this)) + 16|0);
 $$0 = $__s;$__i$0 = 0;
 while(1) {
  $2 = ($__i$0|0)<($__n|0);
  if (!($2)) {
   $__i$0$lcssa = $__i$0;
   break;
  }
  $3 = HEAP32[$0>>2]|0;
  $4 = HEAP32[$1>>2]|0;
  $5 = ($3>>>0)<($4>>>0);
  if ($5) {
   $6 = ((($3)) + 1|0);
   HEAP32[$0>>2] = $6;
   $7 = HEAP8[$3>>0]|0;
   $storemerge = $7;
  } else {
   $8 = HEAP32[$this>>2]|0;
   $9 = ((($8)) + 40|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = (FUNCTION_TABLE_ii[$10 & 8191]($this)|0);
   $12 = ($11|0)==(-1);
   if ($12) {
    $__i$0$lcssa = $__i$0;
    break;
   }
   $13 = $11&255;
   $storemerge = $13;
  }
  HEAP8[$$0>>0] = $storemerge;
  $14 = (($__i$0) + 1)|0;
  $15 = ((($$0)) + 1|0);
  $$0 = $15;$__i$0 = $14;
 }
 return ($__i$0$lcssa|0);
}
function __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE9underflowEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return -1;
}
function __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE5uflowEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + 36|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0);
 $4 = ($3|0)==(-1);
 if ($4) {
  $$0 = -1;
 } else {
  $5 = ((($this)) + 12|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ((($6)) + 1|0);
  HEAP32[$5>>2] = $7;
  $8 = HEAP8[$6>>0]|0;
  $9 = $8&255;
  $$0 = $9;
 }
 return ($$0|0);
}
function __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE9pbackfailEi($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return -1;
}
function __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE6xsputnEPKci($this,$__s,$__n) {
 $this = $this|0;
 $__s = $__s|0;
 $__n = $__n|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__i$0 = 0, $__i$0$lcssa = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0);
 $1 = ((($this)) + 28|0);
 $$0 = $__s;$__i$0 = 0;
 while(1) {
  $2 = ($__i$0|0)<($__n|0);
  if (!($2)) {
   $__i$0$lcssa = $__i$0;
   break;
  }
  $3 = HEAP32[$0>>2]|0;
  $4 = HEAP32[$1>>2]|0;
  $5 = ($3>>>0)<($4>>>0);
  if ($5) {
   $6 = HEAP8[$$0>>0]|0;
   $7 = ((($3)) + 1|0);
   HEAP32[$0>>2] = $7;
   HEAP8[$3>>0] = $6;
  } else {
   $8 = HEAP32[$this>>2]|0;
   $9 = ((($8)) + 52|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = HEAP8[$$0>>0]|0;
   $12 = $11&255;
   $13 = (FUNCTION_TABLE_iii[$10 & 8191]($this,$12)|0);
   $14 = ($13|0)==(-1);
   if ($14) {
    $__i$0$lcssa = $__i$0;
    break;
   }
  }
  $15 = ((($$0)) + 1|0);
  $16 = (($__i$0) + 1)|0;
  $$0 = $15;$__i$0 = $16;
 }
 return ($__i$0$lcssa|0);
}
function __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE8overflowEi($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return -1;
}
function __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEED2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (154680);
 $0 = ((($this)) + 4|0);
 __ZNSt3__16localeD2Ev($0);
 return;
}
function __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (154680);
 $0 = ((($this)) + 4|0);
 __ZNSt3__16localeD2Ev($0);
 __ZdlPv($this);
 return;
}
function __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE5imbueERKNS_6localeE($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE6setbufEPwi($this,$0,$1) {
 $this = $this|0;
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ($this|0);
}
function __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE7seekoffExNS_8ios_base7seekdirEj($agg$result,$this,$0,$1,$2,$3) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = $agg$result;
 $5 = $4;
 HEAP32[$5>>2] = 0;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = ((($agg$result)) + 8|0);
 $9 = $8;
 $10 = $9;
 HEAP32[$10>>2] = -1;
 $11 = (($9) + 4)|0;
 $12 = $11;
 HEAP32[$12>>2] = -1;
 return;
}
function __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE7seekposENS_4fposI11__mbstate_tEEj($agg$result,$this,$0,$1) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = $agg$result;
 $3 = $2;
 HEAP32[$3>>2] = 0;
 $4 = (($2) + 4)|0;
 $5 = $4;
 HEAP32[$5>>2] = 0;
 $6 = ((($agg$result)) + 8|0);
 $7 = $6;
 $8 = $7;
 HEAP32[$8>>2] = -1;
 $9 = (($7) + 4)|0;
 $10 = $9;
 HEAP32[$10>>2] = -1;
 return;
}
function __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE4syncEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE9showmanycEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE6xsgetnEPwi($this,$__s,$__n) {
 $this = $this|0;
 $__s = $__s|0;
 $__n = $__n|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__i$0 = 0, $__i$0$lcssa = 0, $storemerge = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 $1 = ((($this)) + 16|0);
 $$0 = $__s;$__i$0 = 0;
 while(1) {
  $2 = ($__i$0|0)<($__n|0);
  if (!($2)) {
   $__i$0$lcssa = $__i$0;
   break;
  }
  $3 = HEAP32[$0>>2]|0;
  $4 = HEAP32[$1>>2]|0;
  $5 = ($3>>>0)<($4>>>0);
  if ($5) {
   $6 = ((($3)) + 4|0);
   HEAP32[$0>>2] = $6;
   $7 = HEAP32[$3>>2]|0;
   $storemerge = $7;
  } else {
   $8 = HEAP32[$this>>2]|0;
   $9 = ((($8)) + 40|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = (FUNCTION_TABLE_ii[$10 & 8191]($this)|0);
   $12 = ($11|0)==(-1);
   if ($12) {
    $__i$0$lcssa = $__i$0;
    break;
   } else {
    $storemerge = $11;
   }
  }
  HEAP32[$$0>>2] = $storemerge;
  $13 = (($__i$0) + 1)|0;
  $14 = ((($$0)) + 4|0);
  $$0 = $14;$__i$0 = $13;
 }
 return ($__i$0$lcssa|0);
}
function __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE9underflowEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return -1;
}
function __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE5uflowEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + 36|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (FUNCTION_TABLE_ii[$2 & 8191]($this)|0);
 $4 = ($3|0)==(-1);
 if ($4) {
  $$0 = -1;
 } else {
  $5 = ((($this)) + 12|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ((($6)) + 4|0);
  HEAP32[$5>>2] = $7;
  $8 = HEAP32[$6>>2]|0;
  $$0 = $8;
 }
 return ($$0|0);
}
function __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE9pbackfailEj($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return -1;
}
function __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE6xsputnEPKwi($this,$__s,$__n) {
 $this = $this|0;
 $__s = $__s|0;
 $__n = $__n|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__i$0 = 0, $__i$0$lcssa = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0);
 $1 = ((($this)) + 28|0);
 $$0 = $__s;$__i$0 = 0;
 while(1) {
  $2 = ($__i$0|0)<($__n|0);
  if (!($2)) {
   $__i$0$lcssa = $__i$0;
   break;
  }
  $3 = HEAP32[$0>>2]|0;
  $4 = HEAP32[$1>>2]|0;
  $5 = ($3>>>0)<($4>>>0);
  if ($5) {
   $6 = HEAP32[$$0>>2]|0;
   $7 = ((($3)) + 4|0);
   HEAP32[$0>>2] = $7;
   HEAP32[$3>>2] = $6;
  } else {
   $8 = HEAP32[$this>>2]|0;
   $9 = ((($8)) + 52|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = HEAP32[$$0>>2]|0;
   $12 = (FUNCTION_TABLE_iii[$10 & 8191]($this,$11)|0);
   $13 = ($12|0)==(-1);
   if ($13) {
    $__i$0$lcssa = $__i$0;
    break;
   }
  }
  $14 = ((($$0)) + 4|0);
  $15 = (($__i$0) + 1)|0;
  $$0 = $14;$__i$0 = $15;
 }
 return ($__i$0$lcssa|0);
}
function __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE8overflowEj($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return -1;
}
function __ZNSt3__113basic_istreamIcNS_11char_traitsIcEEED1Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 __ZNSt3__18ios_baseD2Ev($0);
 return;
}
function __ZNSt3__113basic_istreamIcNS_11char_traitsIcEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 __ZNSt3__18ios_baseD2Ev($0);
 __ZdlPv($this);
 return;
}
function __ZTv0_n12_NSt3__113basic_istreamIcNS_11char_traitsIcEEED1Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + -12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($this) + ($2)|0);
 $4 = ((($3)) + 8|0);
 __ZNSt3__18ios_baseD2Ev($4);
 return;
}
function __ZTv0_n12_NSt3__113basic_istreamIcNS_11char_traitsIcEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + -12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($this) + ($2)|0);
 __ZNSt3__113basic_istreamIcNS_11char_traitsIcEEED0Ev($3);
 return;
}
function __ZNSt3__113basic_istreamIwNS_11char_traitsIwEEED1Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 __ZNSt3__18ios_baseD2Ev($0);
 return;
}
function __ZNSt3__113basic_istreamIwNS_11char_traitsIwEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 __ZNSt3__18ios_baseD2Ev($0);
 __ZdlPv($this);
 return;
}
function __ZTv0_n12_NSt3__113basic_istreamIwNS_11char_traitsIwEEED1Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + -12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($this) + ($2)|0);
 $4 = ((($3)) + 8|0);
 __ZNSt3__18ios_baseD2Ev($4);
 return;
}
function __ZTv0_n12_NSt3__113basic_istreamIwNS_11char_traitsIwEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + -12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($this) + ($2)|0);
 __ZNSt3__113basic_istreamIwNS_11char_traitsIwEEED0Ev($3);
 return;
}
function __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEED1Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 __ZNSt3__18ios_baseD2Ev($0);
 return;
}
function __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 __ZNSt3__18ios_baseD2Ev($0);
 __ZdlPv($this);
 return;
}
function __ZTv0_n12_NSt3__113basic_ostreamIcNS_11char_traitsIcEEED1Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + -12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($this) + ($2)|0);
 $4 = ((($3)) + 4|0);
 __ZNSt3__18ios_baseD2Ev($4);
 return;
}
function __ZTv0_n12_NSt3__113basic_ostreamIcNS_11char_traitsIcEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + -12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($this) + ($2)|0);
 __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEED0Ev($3);
 return;
}
function __ZNSt3__113basic_ostreamIwNS_11char_traitsIwEEED1Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 __ZNSt3__18ios_baseD2Ev($0);
 return;
}
function __ZNSt3__113basic_ostreamIwNS_11char_traitsIwEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 __ZNSt3__18ios_baseD2Ev($0);
 __ZdlPv($this);
 return;
}
function __ZTv0_n12_NSt3__113basic_ostreamIwNS_11char_traitsIwEEED1Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + -12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($this) + ($2)|0);
 $4 = ((($3)) + 4|0);
 __ZNSt3__18ios_baseD2Ev($4);
 return;
}
function __ZTv0_n12_NSt3__113basic_ostreamIwNS_11char_traitsIwEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + -12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($this) + ($2)|0);
 __ZNSt3__113basic_ostreamIwNS_11char_traitsIwEEED0Ev($3);
 return;
}
function __ZNSt3__114basic_iostreamIcNS_11char_traitsIcEEED1Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 __ZNSt3__18ios_baseD2Ev($0);
 return;
}
function __ZNSt3__114basic_iostreamIcNS_11char_traitsIcEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 __ZNSt3__18ios_baseD2Ev($0);
 __ZdlPv($this);
 return;
}
function __ZThn8_NSt3__114basic_iostreamIcNS_11char_traitsIcEEED1Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0);
 $1 = ((($0)) + 12|0);
 __ZNSt3__18ios_baseD2Ev($1);
 return;
}
function __ZThn8_NSt3__114basic_iostreamIcNS_11char_traitsIcEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + -8|0);
 __ZNSt3__114basic_iostreamIcNS_11char_traitsIcEEED0Ev($0);
 return;
}
function __ZTv0_n12_NSt3__114basic_iostreamIcNS_11char_traitsIcEEED1Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + -12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($this) + ($2)|0);
 $4 = ((($3)) + 12|0);
 __ZNSt3__18ios_baseD2Ev($4);
 return;
}
function __ZTv0_n12_NSt3__114basic_iostreamIcNS_11char_traitsIcEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + -12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($this) + ($2)|0);
 __ZNSt3__114basic_iostreamIcNS_11char_traitsIcEEED0Ev($3);
 return;
}
function __ZNSt3__18ios_base7failureD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112system_errorD2Ev($this);
 return;
}
function __ZNSt3__18ios_base7failureD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112system_errorD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNSt3__119__iostream_categoryD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__119__iostream_category4nameEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (620248|0);
}
function __ZNKSt3__119__iostream_category7messageEi($agg$result,$this,$ev) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $ev = $ev|0;
 var $0 = 0, $1 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($ev|0)!=(1);
 $1 = ($ev|0)<(257);
 $or$cond = $0 & $1;
 if ($or$cond) {
  __ZNKSt3__112__do_message7messageEi($agg$result,$this,$ev);
 } else {
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($agg$result,620257,35);
 }
 return;
}
function __ZNSt3__18ios_base5clearEj($this,$state) {
 $this = $this|0;
 $state = $state|0;
 var $$state = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = ((($this)) + 24|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 $4 = $3&1;
 $$state = $4 | $state;
 $5 = ((($this)) + 16|0);
 HEAP32[$5>>2] = $$state;
 $6 = $3&1;
 $7 = $6 | $state;
 $8 = ((($this)) + 20|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $7 & $9;
 $11 = ($10|0)==(0);
 if ($11) {
  STACKTOP = sp;return;
 }
 $12 = (___cxa_allocate_exception(16)|0);
 (__ZNSt3__117iostream_categoryEv()|0);
 HEAP32[$0>>2] = 1;
 $13 = ((($0)) + 4|0);
 HEAP32[$13>>2] = (1207556);
 __THREW__ = 0;
 invoke_viii(4858,($12|0),(620293|0),($0|0));
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if ($15) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  ___cxa_free_exception(($12|0));
  ___resumeException($16|0);
  // unreachable;
 } else {
  ___cxa_throw(($12|0),(7792|0),(3059|0));
  // unreachable;
 }
}
function __ZNSt3__117iostream_categoryEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[624944]|0;
 $1 = ($0<<24>>24)==(0);
 if ($1) {
  $2 = (___cxa_guard_acquire((624944|0))|0);
  $3 = ($2|0)==(0);
  if (!($3)) {
   HEAP32[301889] = (154924);
   (___cxa_atexit((3062|0),(1207556|0),(___dso_handle|0))|0);
   ___cxa_guard_release((624944|0));
  }
 }
 return (1207556|0);
}
function __ZNSt3__18ios_base7failureC2EPKcRKNS_10error_codeE($this,$msg,$ec) {
 $this = $this|0;
 $msg = $msg|0;
 $ec = $ec|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $tmpcast$byval_copy = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tmpcast$byval_copy = sp + 8|0;
 $0 = sp;
 $1 = $ec;
 $2 = $1;
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) + 4)|0;
 $5 = $4;
 $6 = HEAP32[$5>>2]|0;
 $7 = $0;
 $8 = $7;
 HEAP32[$8>>2] = $3;
 $9 = (($7) + 4)|0;
 $10 = $9;
 HEAP32[$10>>2] = $6;
 ;HEAP32[$tmpcast$byval_copy>>2]=HEAP32[$0>>2]|0;HEAP32[$tmpcast$byval_copy+4>>2]=HEAP32[$0+4>>2]|0;
 __ZNSt3__112system_errorC2ENS_10error_codeEPKc($this,$tmpcast$byval_copy,$msg);
 HEAP32[$this>>2] = (154904);
 STACKTOP = sp;return;
}
function __ZNSt3__18ios_base4initEPv($this,$sb) {
 $this = $this|0;
 $sb = $sb|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $not$ = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0);
 HEAP32[$0>>2] = $sb;
 $not$ = ($sb|0)==(0|0);
 $1 = $not$&1;
 $2 = ((($this)) + 16|0);
 HEAP32[$2>>2] = $1;
 $3 = ((($this)) + 20|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($this)) + 4|0);
 HEAP32[$4>>2] = 4098;
 $5 = ((($this)) + 12|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($this)) + 8|0);
 HEAP32[$6>>2] = 6;
 $7 = ((($this)) + 32|0);
 $8 = ((($this)) + 28|0);
 dest=$7; stop=dest+40|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 __ZNSt3__16localeC2Ev($8);
 return;
}
function __ZNKSt3__18ios_base6getlocEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = ((($this)) + 28|0);
 __ZNSt3__16localeC2ERKS0_($0,$1);
 $2 = HEAP32[$0>>2]|0;
 STACKTOP = sp;return ($2|0);
}
function __ZNSt3__115basic_streambufIcNS_11char_traitsIcEEEC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (154616);
 $0 = ((($this)) + 4|0);
 __ZNSt3__16localeC2Ev($0);
 $1 = ((($this)) + 8|0);
 ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0;HEAP32[$1+12>>2]=0|0;HEAP32[$1+16>>2]=0|0;HEAP32[$1+20>>2]=0|0;
 return;
}
function __ZNSt3__115basic_streambufIwNS_11char_traitsIwEEEC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (154680);
 $0 = ((($this)) + 4|0);
 __ZNSt3__16localeC2Ev($0);
 $1 = ((($this)) + 8|0);
 ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0;HEAP32[$1+12>>2]=0|0;HEAP32[$1+16>>2]=0|0;HEAP32[$1+20>>2]=0|0;
 return;
}
function __ZNSt3__113basic_istreamIcNS_11char_traitsIcEEED2Ev($this,$vtt) {
 $this = $this|0;
 $vtt = $vtt|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__113basic_istreamIcNS_11char_traitsIcEEE6sentryC2ERS3_b($this,$__is,$__noskipws) {
 $this = $this|0;
 $__is = $__is|0;
 $__noskipws = $__noskipws|0;
 var $$0$i$i = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $__i$sroa$0$113 = 0, $isascii1$i = 0, $sext = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 HEAP8[$this>>0] = 0;
 $1 = HEAP32[$__is>>2]|0;
 $2 = ((($1)) + -12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($__is) + ($3)|0);
 $5 = ((($4)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if ($7) {
  $8 = ((($4)) + 72|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)==(0|0);
  if (!($10)) {
   (__ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE5flushEv($9)|0);
  }
  do {
   if (!($__noskipws)) {
    $11 = HEAP32[$__is>>2]|0;
    $12 = ((($11)) + -12|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = (($__is) + ($13)|0);
    $15 = ((($14)) + 4|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = $16 & 4096;
    $18 = ($17|0)==(0);
    if (!($18)) {
     $19 = (__ZNKSt3__18ios_base6getlocEv($14)|0);
     HEAP32[$0>>2] = $19;
     __THREW__ = 0;
     $20 = (invoke_iii(3374,($0|0),(1207576|0))|0);
     $21 = __THREW__; __THREW__ = 0;
     $22 = $21&1;
     if ($22) {
      $64 = ___cxa_find_matching_catch_2()|0;
      $65 = tempRet0;
      __ZNSt3__16localeD2Ev($0);
      ___resumeException($64|0);
      // unreachable;
     }
     __ZNSt3__16localeD2Ev($0);
     $23 = HEAP32[$__is>>2]|0;
     $24 = ((($23)) + -12|0);
     $25 = HEAP32[$24>>2]|0;
     $26 = (($__is) + ($25)|0);
     $27 = ((($26)) + 24|0);
     $28 = HEAP32[$27>>2]|0;
     $29 = ((($20)) + 8|0);
     $30 = $28;
     $31 = ($28|0)==(0);
     $32 = ((($30)) + 12|0);
     $33 = ((($30)) + 16|0);
     $34 = $28;
     $35 = $28;
     $36 = ((($35)) + 12|0);
     $37 = ((($35)) + 16|0);
     $38 = $28;
     $39 = $28;
     while(1) {
      if ($31) {
       $$pre$phiZ2D = 0;$__i$sroa$0$113 = 0;
       break;
      }
      $40 = HEAP32[$32>>2]|0;
      $41 = HEAP32[$33>>2]|0;
      $42 = ($40|0)==($41|0);
      if ($42) {
       $43 = HEAP32[$34>>2]|0;
       $44 = ((($43)) + 36|0);
       $45 = HEAP32[$44>>2]|0;
       $46 = (FUNCTION_TABLE_ii[$45 & 8191]($30)|0);
       $47 = ($46|0)==(-1);
       if ($47) {
        $$pre$phiZ2D = 0;$__i$sroa$0$113 = 0;
        break;
       }
      }
      $48 = HEAP32[$36>>2]|0;
      $49 = HEAP32[$37>>2]|0;
      $50 = ($48|0)==($49|0);
      if ($50) {
       $51 = HEAP32[$38>>2]|0;
       $52 = ((($51)) + 36|0);
       $53 = HEAP32[$52>>2]|0;
       $54 = (FUNCTION_TABLE_ii[$53 & 8191]($35)|0);
       $$0$i$i = $54;
      } else {
       $55 = HEAP8[$48>>0]|0;
       $56 = $55&255;
       $$0$i$i = $56;
      }
      $57 = $$0$i$i&255;
      $isascii1$i = ($57<<24>>24)>(-1);
      if (!($isascii1$i)) {
       $$pre$phiZ2D = $35;$__i$sroa$0$113 = $28;
       break;
      }
      $sext = $$0$i$i << 24;
      $58 = $sext >> 24;
      $59 = HEAP32[$29>>2]|0;
      $60 = (($59) + ($58<<1)|0);
      $61 = HEAP16[$60>>1]|0;
      $62 = $61 & 8192;
      $63 = ($62<<16>>16)==(0);
      if ($63) {
       $$pre$phiZ2D = $35;$__i$sroa$0$113 = $28;
       break;
      }
      $66 = HEAP32[$36>>2]|0;
      $67 = HEAP32[$37>>2]|0;
      $68 = ($66|0)==($67|0);
      if ($68) {
       $69 = HEAP32[$39>>2]|0;
       $70 = ((($69)) + 40|0);
       $71 = HEAP32[$70>>2]|0;
       (FUNCTION_TABLE_ii[$71 & 8191]($35)|0);
       continue;
      } else {
       $72 = ((($66)) + 1|0);
       HEAP32[$36>>2] = $72;
       continue;
      }
     }
     $73 = ($__i$sroa$0$113|0)==(0);
     if (!($73)) {
      $74 = ((($$pre$phiZ2D)) + 12|0);
      $75 = HEAP32[$74>>2]|0;
      $76 = ((($$pre$phiZ2D)) + 16|0);
      $77 = HEAP32[$76>>2]|0;
      $78 = ($75|0)==($77|0);
      if (!($78)) {
       break;
      }
      $79 = $__i$sroa$0$113;
      $80 = HEAP32[$79>>2]|0;
      $81 = ((($80)) + 36|0);
      $82 = HEAP32[$81>>2]|0;
      $83 = (FUNCTION_TABLE_ii[$82 & 8191]($$pre$phiZ2D)|0);
      $84 = ($83|0)==(-1);
      if (!($84)) {
       break;
      }
     }
     $85 = HEAP32[$__is>>2]|0;
     $86 = ((($85)) + -12|0);
     $87 = HEAP32[$86>>2]|0;
     $88 = (($__is) + ($87)|0);
     $89 = ((($88)) + 16|0);
     $90 = HEAP32[$89>>2]|0;
     $91 = $90 | 6;
     __ZNSt3__18ios_base5clearEj($88,$91);
    }
   }
  } while(0);
  $92 = HEAP32[$__is>>2]|0;
  $93 = ((($92)) + -12|0);
  $94 = HEAP32[$93>>2]|0;
  $95 = (($__is) + ($94)|0);
  $96 = ((($95)) + 16|0);
  $97 = HEAP32[$96>>2]|0;
  $98 = ($97|0)==(0);
  $99 = $98&1;
  HEAP8[$this>>0] = $99;
 } else {
  $100 = $6 | 4;
  __ZNSt3__18ios_base5clearEj($4,$100);
 }
 STACKTOP = sp;return;
}
function __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE5flushEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__s = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__s = sp;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + -12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($this) + ($2)|0);
 $4 = ((($3)) + 24|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 L1: do {
  if (!($6)) {
   __THREW__ = 0;
   invoke_vii(3372,($__s|0),($this|0));
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   L3: do {
    if ($8) {
     $33 = ___cxa_find_matching_catch_3(0|0)|0;
     $34 = tempRet0;
     $$0 = $33;
    } else {
     $9 = HEAP8[$__s>>0]|0;
     $10 = ($9<<24>>24)==(0);
     do {
      if (!($10)) {
       $11 = HEAP32[$this>>2]|0;
       $12 = ((($11)) + -12|0);
       $13 = HEAP32[$12>>2]|0;
       $14 = (($this) + ($13)|0);
       $15 = ((($14)) + 24|0);
       $16 = HEAP32[$15>>2]|0;
       $17 = HEAP32[$16>>2]|0;
       $18 = ((($17)) + 24|0);
       $19 = HEAP32[$18>>2]|0;
       __THREW__ = 0;
       $20 = (invoke_ii($19|0,($16|0))|0);
       $21 = __THREW__; __THREW__ = 0;
       $22 = $21&1;
       if (!($22)) {
        $23 = ($20|0)==(-1);
        if (!($23)) {
         break;
        }
        $24 = HEAP32[$this>>2]|0;
        $25 = ((($24)) + -12|0);
        $26 = HEAP32[$25>>2]|0;
        $27 = (($this) + ($26)|0);
        $28 = ((($27)) + 16|0);
        $29 = HEAP32[$28>>2]|0;
        $30 = $29 | 1;
        __THREW__ = 0;
        invoke_vii(3376,($27|0),($30|0));
        $31 = __THREW__; __THREW__ = 0;
        $32 = $31&1;
        if (!($32)) {
         break;
        }
       }
       $35 = ___cxa_find_matching_catch_3(0|0)|0;
       $36 = tempRet0;
       __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE6sentryD2Ev($__s);
       $$0 = $35;
       break L3;
      }
     } while(0);
     __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE6sentryD2Ev($__s);
     break L1;
    }
   } while(0);
   (___cxa_begin_catch(($$0|0))|0);
   $37 = HEAP32[$this>>2]|0;
   $38 = ((($37)) + -12|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = (($this) + ($39)|0);
   __THREW__ = 0;
   invoke_vi(3377,($40|0));
   $41 = __THREW__; __THREW__ = 0;
   $42 = $41&1;
   if (!($42)) {
    ___cxa_end_catch();
    break;
   }
   $43 = ___cxa_find_matching_catch_2()|0;
   $44 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378);
   $45 = __THREW__; __THREW__ = 0;
   $46 = $45&1;
   if ($46) {
    $47 = ___cxa_find_matching_catch_3(0|0)|0;
    $48 = tempRet0;
    ___clang_call_terminate($47);
    // unreachable;
   } else {
    ___resumeException($43|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return ($this|0);
}
function __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE6sentryC2ERS3_($this,$__os) {
 $this = $this|0;
 $__os = $__os|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$this>>0] = 0;
 $0 = ((($this)) + 4|0);
 HEAP32[$0>>2] = $__os;
 $1 = HEAP32[$__os>>2]|0;
 $2 = ((($1)) + -12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($__os) + ($3)|0);
 $5 = ((($4)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if ($7) {
  $8 = ((($4)) + 72|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)==(0|0);
  if (!($10)) {
   (__ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE5flushEv($9)|0);
  }
  HEAP8[$this>>0] = 1;
 }
 return;
}
function __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE6sentryD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + -12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($1) + ($4)|0);
 $6 = ((($5)) + 24|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 do {
  if (!($8)) {
   $9 = ((($5)) + 16|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($10|0)==(0);
   if ($11) {
    $12 = ((($5)) + 4|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = $13 & 8192;
    $15 = ($14|0)==(0);
    if (!($15)) {
     $16 = (__ZSt18uncaught_exceptionv()|0);
     if (!($16)) {
      $17 = HEAP32[$0>>2]|0;
      $18 = HEAP32[$17>>2]|0;
      $19 = ((($18)) + -12|0);
      $20 = HEAP32[$19>>2]|0;
      $21 = (($17) + ($20)|0);
      $22 = ((($21)) + 24|0);
      $23 = HEAP32[$22>>2]|0;
      $24 = HEAP32[$23>>2]|0;
      $25 = ((($24)) + 24|0);
      $26 = HEAP32[$25>>2]|0;
      __THREW__ = 0;
      $27 = (invoke_ii($26|0,($23|0))|0);
      $28 = __THREW__; __THREW__ = 0;
      $29 = $28&1;
      if (!($29)) {
       $30 = ($27|0)==(-1);
       if (!($30)) {
        break;
       }
       $31 = HEAP32[$0>>2]|0;
       $32 = HEAP32[$31>>2]|0;
       $33 = ((($32)) + -12|0);
       $34 = HEAP32[$33>>2]|0;
       $35 = (($31) + ($34)|0);
       $36 = ((($35)) + 16|0);
       $37 = HEAP32[$36>>2]|0;
       $38 = $37 | 1;
       __THREW__ = 0;
       invoke_vii(3376,($35|0),($38|0));
       $39 = __THREW__; __THREW__ = 0;
       $40 = $39&1;
       if (!($40)) {
        break;
       }
      }
      $41 = ___cxa_find_matching_catch_3(0|0)|0;
      $42 = tempRet0;
      (___cxa_begin_catch(($41|0))|0);
      __THREW__ = 0;
      invoke_v(3378);
      $43 = __THREW__; __THREW__ = 0;
      $44 = $43&1;
      if ($44) {
       $45 = ___cxa_find_matching_catch_3(0|0)|0;
       $46 = tempRet0;
       ___clang_call_terminate($45);
       // unreachable;
      }
     }
    }
   }
  }
 } while(0);
 return;
}
function __ZNSt3__18ios_base33__set_badbit_and_consider_rethrowEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 | 1;
 HEAP32[$0>>2] = $2;
 $3 = ((($this)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 & 1;
 $6 = ($5|0)==(0);
 if ($6) {
  return;
 } else {
  ___cxa_rethrow();
  // unreachable;
 }
}
function __ZNSt3__113basic_ostreamIwNS_11char_traitsIwEEE5flushEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__s = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__s = sp;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + -12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($this) + ($2)|0);
 $4 = ((($3)) + 24|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 L1: do {
  if (!($6)) {
   __THREW__ = 0;
   invoke_vii(4859,($__s|0),($this|0));
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   L3: do {
    if ($8) {
     $33 = ___cxa_find_matching_catch_3(0|0)|0;
     $34 = tempRet0;
     $$0 = $33;
    } else {
     $9 = HEAP8[$__s>>0]|0;
     $10 = ($9<<24>>24)==(0);
     do {
      if (!($10)) {
       $11 = HEAP32[$this>>2]|0;
       $12 = ((($11)) + -12|0);
       $13 = HEAP32[$12>>2]|0;
       $14 = (($this) + ($13)|0);
       $15 = ((($14)) + 24|0);
       $16 = HEAP32[$15>>2]|0;
       $17 = HEAP32[$16>>2]|0;
       $18 = ((($17)) + 24|0);
       $19 = HEAP32[$18>>2]|0;
       __THREW__ = 0;
       $20 = (invoke_ii($19|0,($16|0))|0);
       $21 = __THREW__; __THREW__ = 0;
       $22 = $21&1;
       if (!($22)) {
        $23 = ($20|0)==(-1);
        if (!($23)) {
         break;
        }
        $24 = HEAP32[$this>>2]|0;
        $25 = ((($24)) + -12|0);
        $26 = HEAP32[$25>>2]|0;
        $27 = (($this) + ($26)|0);
        $28 = ((($27)) + 16|0);
        $29 = HEAP32[$28>>2]|0;
        $30 = $29 | 1;
        __THREW__ = 0;
        invoke_vii(3376,($27|0),($30|0));
        $31 = __THREW__; __THREW__ = 0;
        $32 = $31&1;
        if (!($32)) {
         break;
        }
       }
       $35 = ___cxa_find_matching_catch_3(0|0)|0;
       $36 = tempRet0;
       __ZNSt3__113basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($__s);
       $$0 = $35;
       break L3;
      }
     } while(0);
     __ZNSt3__113basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($__s);
     break L1;
    }
   } while(0);
   (___cxa_begin_catch(($$0|0))|0);
   $37 = HEAP32[$this>>2]|0;
   $38 = ((($37)) + -12|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = (($this) + ($39)|0);
   __THREW__ = 0;
   invoke_vi(3377,($40|0));
   $41 = __THREW__; __THREW__ = 0;
   $42 = $41&1;
   if (!($42)) {
    ___cxa_end_catch();
    break;
   }
   $43 = ___cxa_find_matching_catch_2()|0;
   $44 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378);
   $45 = __THREW__; __THREW__ = 0;
   $46 = $45&1;
   if ($46) {
    $47 = ___cxa_find_matching_catch_3(0|0)|0;
    $48 = tempRet0;
    ___clang_call_terminate($47);
    // unreachable;
   } else {
    ___resumeException($43|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return ($this|0);
}
function __ZNSt3__113basic_ostreamIwNS_11char_traitsIwEEE6sentryC2ERS3_($this,$__os) {
 $this = $this|0;
 $__os = $__os|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$this>>0] = 0;
 $0 = ((($this)) + 4|0);
 HEAP32[$0>>2] = $__os;
 $1 = HEAP32[$__os>>2]|0;
 $2 = ((($1)) + -12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($__os) + ($3)|0);
 $5 = ((($4)) + 16|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if ($7) {
  $8 = ((($4)) + 72|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)==(0|0);
  if (!($10)) {
   (__ZNSt3__113basic_ostreamIwNS_11char_traitsIwEEE5flushEv($9)|0);
  }
  HEAP8[$this>>0] = 1;
 }
 return;
}
function __ZNSt3__113basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + -12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($1) + ($4)|0);
 $6 = ((($5)) + 24|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 do {
  if (!($8)) {
   $9 = ((($5)) + 16|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($10|0)==(0);
   if ($11) {
    $12 = ((($5)) + 4|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = $13 & 8192;
    $15 = ($14|0)==(0);
    if (!($15)) {
     $16 = (__ZSt18uncaught_exceptionv()|0);
     if (!($16)) {
      $17 = HEAP32[$0>>2]|0;
      $18 = HEAP32[$17>>2]|0;
      $19 = ((($18)) + -12|0);
      $20 = HEAP32[$19>>2]|0;
      $21 = (($17) + ($20)|0);
      $22 = ((($21)) + 24|0);
      $23 = HEAP32[$22>>2]|0;
      $24 = HEAP32[$23>>2]|0;
      $25 = ((($24)) + 24|0);
      $26 = HEAP32[$25>>2]|0;
      __THREW__ = 0;
      $27 = (invoke_ii($26|0,($23|0))|0);
      $28 = __THREW__; __THREW__ = 0;
      $29 = $28&1;
      if (!($29)) {
       $30 = ($27|0)==(-1);
       if (!($30)) {
        break;
       }
       $31 = HEAP32[$0>>2]|0;
       $32 = HEAP32[$31>>2]|0;
       $33 = ((($32)) + -12|0);
       $34 = HEAP32[$33>>2]|0;
       $35 = (($31) + ($34)|0);
       $36 = ((($35)) + 16|0);
       $37 = HEAP32[$36>>2]|0;
       $38 = $37 | 1;
       __THREW__ = 0;
       invoke_vii(3376,($35|0),($38|0));
       $39 = __THREW__; __THREW__ = 0;
       $40 = $39&1;
       if (!($40)) {
        break;
       }
      }
      $41 = ___cxa_find_matching_catch_3(0|0)|0;
      $42 = tempRet0;
      (___cxa_begin_catch(($41|0))|0);
      __THREW__ = 0;
      invoke_v(3378);
      $43 = __THREW__; __THREW__ = 0;
      $44 = $43&1;
      if ($44) {
       $45 = ___cxa_find_matching_catch_3(0|0)|0;
       $46 = tempRet0;
       ___clang_call_terminate($45);
       // unreachable;
      }
     }
    }
   }
  }
 } while(0);
 return;
}
function __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEElsEi($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$0 = 0, $$1 = 0, $$byval_copy = 0, $$pre$phi5Z2D = 0, $$pre$phi7Z2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0, $__s = 0;
 var $eh$lpad$body$indexZ2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 24|0;
 $0 = sp + 20|0;
 $1 = sp + 16|0;
 $__s = sp + 8|0;
 $2 = sp;
 __THREW__ = 0;
 invoke_vii(3372,($__s|0),($this|0));
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 L1: do {
  if ($4) {
   $55 = ___cxa_find_matching_catch_3(0|0)|0;
   $56 = tempRet0;
   $$1 = $55;$$pre$phi5Z2D = $this;$$pre$phiZ2D = $this;
   label = 19;
  } else {
   $5 = HEAP8[$__s>>0]|0;
   $6 = ($5<<24>>24)==(0);
   do {
    if (!($6)) {
     $7 = HEAP32[$this>>2]|0;
     $8 = ((($7)) + -12|0);
     $9 = HEAP32[$8>>2]|0;
     $10 = (($this) + ($9)|0);
     $11 = (__ZNKSt3__18ios_base6getlocEv($10)|0);
     HEAP32[$2>>2] = $11;
     __THREW__ = 0;
     $12 = (invoke_iii(3374,($2|0),(1207632|0))|0);
     $13 = __THREW__; __THREW__ = 0;
     $14 = $13&1;
     if ($14) {
      $59 = ___cxa_find_matching_catch_3(0|0)|0;
      $60 = tempRet0;
      __ZNSt3__16localeD2Ev($2);
      $$0 = $59;
     } else {
      __ZNSt3__16localeD2Ev($2);
      $15 = HEAP32[$this>>2]|0;
      $16 = ((($15)) + -12|0);
      $17 = HEAP32[$16>>2]|0;
      $18 = (($this) + ($17)|0);
      $19 = ((($18)) + 24|0);
      $20 = HEAP32[$19>>2]|0;
      $21 = ((($18)) + 76|0);
      $22 = HEAP32[$21>>2]|0;
      $23 = ($22|0)==(-1);
      do {
       if ($23) {
        $24 = (__ZNKSt3__18ios_base6getlocEv($18)|0);
        HEAP32[$1>>2] = $24;
        __THREW__ = 0;
        $25 = (invoke_iii(3374,($1|0),(1207576|0))|0);
        $26 = __THREW__; __THREW__ = 0;
        $27 = $26&1;
        if (!($27)) {
         $28 = HEAP32[$25>>2]|0;
         $29 = ((($28)) + 28|0);
         $30 = HEAP32[$29>>2]|0;
         __THREW__ = 0;
         $31 = (invoke_iii($30|0,($25|0),32)|0);
         $32 = __THREW__; __THREW__ = 0;
         $33 = $32&1;
         if (!($33)) {
          __ZNSt3__16localeD2Ev($1);
          $36 = $31 << 24 >> 24;
          HEAP32[$21>>2] = $36;
          $$pre$phi7Z2D = $18;$38 = $36;
          label = 10;
          break;
         }
        }
        $34 = ___cxa_find_matching_catch_3(0|0)|0;
        $35 = tempRet0;
        __ZNSt3__16localeD2Ev($1);
        $eh$lpad$body$indexZ2D = $34;
       } else {
        $$pre$phi7Z2D = $18;$38 = $22;
        label = 10;
       }
      } while(0);
      if ((label|0) == 10) {
       $37 = $38&255;
       $39 = HEAP32[$12>>2]|0;
       $40 = ((($39)) + 16|0);
       $41 = HEAP32[$40>>2]|0;
       HEAP32[$0>>2] = $20;
       __THREW__ = 0;
       ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
       $42 = (invoke_iiiiii($41|0,($12|0),($$byval_copy|0),($$pre$phi7Z2D|0),($37|0),($__n|0))|0);
       $43 = __THREW__; __THREW__ = 0;
       $44 = $43&1;
       if (!($44)) {
        $45 = ($42|0)==(0|0);
        if (!($45)) {
         break;
        }
        $46 = HEAP32[$this>>2]|0;
        $47 = ((($46)) + -12|0);
        $48 = HEAP32[$47>>2]|0;
        $49 = (($this) + ($48)|0);
        $50 = ((($49)) + 16|0);
        $51 = HEAP32[$50>>2]|0;
        $52 = $51 | 5;
        __THREW__ = 0;
        invoke_vii(3376,($49|0),($52|0));
        $53 = __THREW__; __THREW__ = 0;
        $54 = $53&1;
        if (!($54)) {
         break;
        }
       }
       $57 = ___cxa_find_matching_catch_3(0|0)|0;
       $58 = tempRet0;
       $eh$lpad$body$indexZ2D = $57;
      }
      $$0 = $eh$lpad$body$indexZ2D;
     }
     __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE6sentryD2Ev($__s);
     $$1 = $$0;$$pre$phi5Z2D = $this;$$pre$phiZ2D = $this;
     label = 19;
     break L1;
    }
   } while(0);
   __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE6sentryD2Ev($__s);
  }
 } while(0);
 do {
  if ((label|0) == 19) {
   (___cxa_begin_catch(($$1|0))|0);
   $61 = HEAP32[$$pre$phiZ2D>>2]|0;
   $62 = ((($61)) + -12|0);
   $63 = HEAP32[$62>>2]|0;
   $64 = (($$pre$phi5Z2D) + ($63)|0);
   __THREW__ = 0;
   invoke_vi(3377,($64|0));
   $65 = __THREW__; __THREW__ = 0;
   $66 = $65&1;
   if (!($66)) {
    ___cxa_end_catch();
    break;
   }
   $67 = ___cxa_find_matching_catch_2()|0;
   $68 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378);
   $69 = __THREW__; __THREW__ = 0;
   $70 = $69&1;
   if ($70) {
    $71 = ___cxa_find_matching_catch_3(0|0)|0;
    $72 = tempRet0;
    ___clang_call_terminate($71);
    // unreachable;
   } else {
    ___resumeException($67|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return ($this|0);
}
function __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEElsEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$0 = 0, $$1 = 0, $$byval_copy = 0, $$pre$phi4Z2D = 0, $$pre$phi6Z2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0, $__s = 0;
 var $eh$lpad$body$indexZ2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 24|0;
 $0 = sp + 20|0;
 $1 = sp + 16|0;
 $__s = sp + 8|0;
 $2 = sp;
 __THREW__ = 0;
 invoke_vii(3372,($__s|0),($this|0));
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 L1: do {
  if ($4) {
   $55 = ___cxa_find_matching_catch_3(0|0)|0;
   $56 = tempRet0;
   $$1 = $55;$$pre$phi4Z2D = $this;$$pre$phiZ2D = $this;
   label = 19;
  } else {
   $5 = HEAP8[$__s>>0]|0;
   $6 = ($5<<24>>24)==(0);
   do {
    if (!($6)) {
     $7 = HEAP32[$this>>2]|0;
     $8 = ((($7)) + -12|0);
     $9 = HEAP32[$8>>2]|0;
     $10 = (($this) + ($9)|0);
     $11 = (__ZNKSt3__18ios_base6getlocEv($10)|0);
     HEAP32[$2>>2] = $11;
     __THREW__ = 0;
     $12 = (invoke_iii(3374,($2|0),(1207632|0))|0);
     $13 = __THREW__; __THREW__ = 0;
     $14 = $13&1;
     if ($14) {
      $59 = ___cxa_find_matching_catch_3(0|0)|0;
      $60 = tempRet0;
      __ZNSt3__16localeD2Ev($2);
      $$0 = $59;
     } else {
      __ZNSt3__16localeD2Ev($2);
      $15 = HEAP32[$this>>2]|0;
      $16 = ((($15)) + -12|0);
      $17 = HEAP32[$16>>2]|0;
      $18 = (($this) + ($17)|0);
      $19 = ((($18)) + 24|0);
      $20 = HEAP32[$19>>2]|0;
      $21 = ((($18)) + 76|0);
      $22 = HEAP32[$21>>2]|0;
      $23 = ($22|0)==(-1);
      do {
       if ($23) {
        $24 = (__ZNKSt3__18ios_base6getlocEv($18)|0);
        HEAP32[$1>>2] = $24;
        __THREW__ = 0;
        $25 = (invoke_iii(3374,($1|0),(1207576|0))|0);
        $26 = __THREW__; __THREW__ = 0;
        $27 = $26&1;
        if (!($27)) {
         $28 = HEAP32[$25>>2]|0;
         $29 = ((($28)) + 28|0);
         $30 = HEAP32[$29>>2]|0;
         __THREW__ = 0;
         $31 = (invoke_iii($30|0,($25|0),32)|0);
         $32 = __THREW__; __THREW__ = 0;
         $33 = $32&1;
         if (!($33)) {
          __ZNSt3__16localeD2Ev($1);
          $36 = $31 << 24 >> 24;
          HEAP32[$21>>2] = $36;
          $$pre$phi6Z2D = $18;$38 = $36;
          label = 10;
          break;
         }
        }
        $34 = ___cxa_find_matching_catch_3(0|0)|0;
        $35 = tempRet0;
        __ZNSt3__16localeD2Ev($1);
        $eh$lpad$body$indexZ2D = $34;
       } else {
        $$pre$phi6Z2D = $18;$38 = $22;
        label = 10;
       }
      } while(0);
      if ((label|0) == 10) {
       $37 = $38&255;
       $39 = HEAP32[$12>>2]|0;
       $40 = ((($39)) + 24|0);
       $41 = HEAP32[$40>>2]|0;
       HEAP32[$0>>2] = $20;
       __THREW__ = 0;
       ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
       $42 = (invoke_iiiiii($41|0,($12|0),($$byval_copy|0),($$pre$phi6Z2D|0),($37|0),($__n|0))|0);
       $43 = __THREW__; __THREW__ = 0;
       $44 = $43&1;
       if (!($44)) {
        $45 = ($42|0)==(0|0);
        if (!($45)) {
         break;
        }
        $46 = HEAP32[$this>>2]|0;
        $47 = ((($46)) + -12|0);
        $48 = HEAP32[$47>>2]|0;
        $49 = (($this) + ($48)|0);
        $50 = ((($49)) + 16|0);
        $51 = HEAP32[$50>>2]|0;
        $52 = $51 | 5;
        __THREW__ = 0;
        invoke_vii(3376,($49|0),($52|0));
        $53 = __THREW__; __THREW__ = 0;
        $54 = $53&1;
        if (!($54)) {
         break;
        }
       }
       $57 = ___cxa_find_matching_catch_3(0|0)|0;
       $58 = tempRet0;
       $eh$lpad$body$indexZ2D = $57;
      }
      $$0 = $eh$lpad$body$indexZ2D;
     }
     __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE6sentryD2Ev($__s);
     $$1 = $$0;$$pre$phi4Z2D = $this;$$pre$phiZ2D = $this;
     label = 19;
     break L1;
    }
   } while(0);
   __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE6sentryD2Ev($__s);
  }
 } while(0);
 do {
  if ((label|0) == 19) {
   (___cxa_begin_catch(($$1|0))|0);
   $61 = HEAP32[$$pre$phiZ2D>>2]|0;
   $62 = ((($61)) + -12|0);
   $63 = HEAP32[$62>>2]|0;
   $64 = (($$pre$phi4Z2D) + ($63)|0);
   __THREW__ = 0;
   invoke_vi(3377,($64|0));
   $65 = __THREW__; __THREW__ = 0;
   $66 = $65&1;
   if (!($66)) {
    ___cxa_end_catch();
    break;
   }
   $67 = ___cxa_find_matching_catch_2()|0;
   $68 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378);
   $69 = __THREW__; __THREW__ = 0;
   $70 = $69&1;
   if ($70) {
    $71 = ___cxa_find_matching_catch_3(0|0)|0;
    $72 = tempRet0;
    ___clang_call_terminate($71);
    // unreachable;
   } else {
    ___resumeException($67|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return ($this|0);
}
function __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEElsEy($this,$0,$1) {
 $this = $this|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$1 = 0, $$byval_copy = 0, $$pre$phi4Z2D = 0, $$pre$phi6Z2D = 0, $$pre$phiZ2D = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $8 = 0, $9 = 0, $__s = 0;
 var $eh$lpad$body$indexZ2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 24|0;
 $2 = sp + 20|0;
 $3 = sp + 16|0;
 $__s = sp + 8|0;
 $4 = sp;
 __THREW__ = 0;
 invoke_vii(3372,($__s|0),($this|0));
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 L1: do {
  if ($6) {
   $57 = ___cxa_find_matching_catch_3(0|0)|0;
   $58 = tempRet0;
   $$1 = $57;$$pre$phi4Z2D = $this;$$pre$phiZ2D = $this;
   label = 19;
  } else {
   $7 = HEAP8[$__s>>0]|0;
   $8 = ($7<<24>>24)==(0);
   do {
    if (!($8)) {
     $9 = HEAP32[$this>>2]|0;
     $10 = ((($9)) + -12|0);
     $11 = HEAP32[$10>>2]|0;
     $12 = (($this) + ($11)|0);
     $13 = (__ZNKSt3__18ios_base6getlocEv($12)|0);
     HEAP32[$4>>2] = $13;
     __THREW__ = 0;
     $14 = (invoke_iii(3374,($4|0),(1207632|0))|0);
     $15 = __THREW__; __THREW__ = 0;
     $16 = $15&1;
     if ($16) {
      $61 = ___cxa_find_matching_catch_3(0|0)|0;
      $62 = tempRet0;
      __ZNSt3__16localeD2Ev($4);
      $$0 = $61;
     } else {
      __ZNSt3__16localeD2Ev($4);
      $17 = HEAP32[$this>>2]|0;
      $18 = ((($17)) + -12|0);
      $19 = HEAP32[$18>>2]|0;
      $20 = (($this) + ($19)|0);
      $21 = ((($20)) + 24|0);
      $22 = HEAP32[$21>>2]|0;
      $23 = ((($20)) + 76|0);
      $24 = HEAP32[$23>>2]|0;
      $25 = ($24|0)==(-1);
      do {
       if ($25) {
        $26 = (__ZNKSt3__18ios_base6getlocEv($20)|0);
        HEAP32[$3>>2] = $26;
        __THREW__ = 0;
        $27 = (invoke_iii(3374,($3|0),(1207576|0))|0);
        $28 = __THREW__; __THREW__ = 0;
        $29 = $28&1;
        if (!($29)) {
         $30 = HEAP32[$27>>2]|0;
         $31 = ((($30)) + 28|0);
         $32 = HEAP32[$31>>2]|0;
         __THREW__ = 0;
         $33 = (invoke_iii($32|0,($27|0),32)|0);
         $34 = __THREW__; __THREW__ = 0;
         $35 = $34&1;
         if (!($35)) {
          __ZNSt3__16localeD2Ev($3);
          $38 = $33 << 24 >> 24;
          HEAP32[$23>>2] = $38;
          $$pre$phi6Z2D = $20;$40 = $38;
          label = 10;
          break;
         }
        }
        $36 = ___cxa_find_matching_catch_3(0|0)|0;
        $37 = tempRet0;
        __ZNSt3__16localeD2Ev($3);
        $eh$lpad$body$indexZ2D = $36;
       } else {
        $$pre$phi6Z2D = $20;$40 = $24;
        label = 10;
       }
      } while(0);
      if ((label|0) == 10) {
       $39 = $40&255;
       $41 = HEAP32[$14>>2]|0;
       $42 = ((($41)) + 28|0);
       $43 = HEAP32[$42>>2]|0;
       HEAP32[$2>>2] = $22;
       __THREW__ = 0;
       ;HEAP32[$$byval_copy>>2]=HEAP32[$2>>2]|0;
       $44 = (invoke_iiiiiii($43|0,($14|0),($$byval_copy|0),($$pre$phi6Z2D|0),($39|0),($0|0),($1|0))|0);
       $45 = __THREW__; __THREW__ = 0;
       $46 = $45&1;
       if (!($46)) {
        $47 = ($44|0)==(0|0);
        if (!($47)) {
         break;
        }
        $48 = HEAP32[$this>>2]|0;
        $49 = ((($48)) + -12|0);
        $50 = HEAP32[$49>>2]|0;
        $51 = (($this) + ($50)|0);
        $52 = ((($51)) + 16|0);
        $53 = HEAP32[$52>>2]|0;
        $54 = $53 | 5;
        __THREW__ = 0;
        invoke_vii(3376,($51|0),($54|0));
        $55 = __THREW__; __THREW__ = 0;
        $56 = $55&1;
        if (!($56)) {
         break;
        }
       }
       $59 = ___cxa_find_matching_catch_3(0|0)|0;
       $60 = tempRet0;
       $eh$lpad$body$indexZ2D = $59;
      }
      $$0 = $eh$lpad$body$indexZ2D;
     }
     __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE6sentryD2Ev($__s);
     $$1 = $$0;$$pre$phi4Z2D = $this;$$pre$phiZ2D = $this;
     label = 19;
     break L1;
    }
   } while(0);
   __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE6sentryD2Ev($__s);
  }
 } while(0);
 do {
  if ((label|0) == 19) {
   (___cxa_begin_catch(($$1|0))|0);
   $63 = HEAP32[$$pre$phiZ2D>>2]|0;
   $64 = ((($63)) + -12|0);
   $65 = HEAP32[$64>>2]|0;
   $66 = (($$pre$phi4Z2D) + ($65)|0);
   __THREW__ = 0;
   invoke_vi(3377,($66|0));
   $67 = __THREW__; __THREW__ = 0;
   $68 = $67&1;
   if (!($68)) {
    ___cxa_end_catch();
    break;
   }
   $69 = ___cxa_find_matching_catch_2()|0;
   $70 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378);
   $71 = __THREW__; __THREW__ = 0;
   $72 = $71&1;
   if ($72) {
    $73 = ___cxa_find_matching_catch_3(0|0)|0;
    $74 = tempRet0;
    ___clang_call_terminate($73);
    // unreachable;
   } else {
    ___resumeException($69|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return ($this|0);
}
function __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE3putEc($this,$__c) {
 $this = $this|0;
 $__c = $__c|0;
 var $$0 = 0, $$pre$phi3Z2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__s = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__s = sp;
 __THREW__ = 0;
 invoke_vii(3372,($__s|0),($this|0));
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 L1: do {
  if ($1) {
   $36 = ___cxa_find_matching_catch_3(0|0)|0;
   $37 = tempRet0;
   $$0 = $36;$$pre$phi3Z2D = $this;$$pre$phiZ2D = $this;
   label = 11;
  } else {
   $2 = HEAP8[$__s>>0]|0;
   $3 = ($2<<24>>24)==(0);
   do {
    if (!($3)) {
     $4 = HEAP32[$this>>2]|0;
     $5 = ((($4)) + -12|0);
     $6 = HEAP32[$5>>2]|0;
     $7 = (($this) + ($6)|0);
     $8 = ((($7)) + 24|0);
     $9 = HEAP32[$8>>2]|0;
     $10 = $9;
     $11 = ($9|0)==(0);
     if ($11) {
      label = 8;
     } else {
      $12 = ((($10)) + 24|0);
      $13 = HEAP32[$12>>2]|0;
      $14 = ((($10)) + 28|0);
      $15 = HEAP32[$14>>2]|0;
      $16 = ($13|0)==($15|0);
      if (!($16)) {
       $17 = ((($13)) + 1|0);
       HEAP32[$12>>2] = $17;
       HEAP8[$13>>0] = $__c;
       break;
      }
      $18 = $9;
      $19 = HEAP32[$18>>2]|0;
      $20 = ((($19)) + 52|0);
      $21 = HEAP32[$20>>2]|0;
      $22 = $__c&255;
      __THREW__ = 0;
      $23 = (invoke_iii($21|0,($10|0),($22|0))|0);
      $24 = __THREW__; __THREW__ = 0;
      $25 = $24&1;
      if (!($25)) {
       $26 = ($23|0)==(-1);
       if ($26) {
        label = 8;
       } else {
        break;
       }
      }
     }
     if ((label|0) == 8) {
      $27 = HEAP32[$this>>2]|0;
      $28 = ((($27)) + -12|0);
      $29 = HEAP32[$28>>2]|0;
      $30 = (($this) + ($29)|0);
      $31 = ((($30)) + 16|0);
      $32 = HEAP32[$31>>2]|0;
      $33 = $32 | 1;
      __THREW__ = 0;
      invoke_vii(3376,($30|0),($33|0));
      $34 = __THREW__; __THREW__ = 0;
      $35 = $34&1;
      if (!($35)) {
       break;
      }
     }
     $38 = ___cxa_find_matching_catch_3(0|0)|0;
     $39 = tempRet0;
     __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE6sentryD2Ev($__s);
     $$0 = $38;$$pre$phi3Z2D = $this;$$pre$phiZ2D = $this;
     label = 11;
     break L1;
    }
   } while(0);
   __ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE6sentryD2Ev($__s);
  }
 } while(0);
 do {
  if ((label|0) == 11) {
   (___cxa_begin_catch(($$0|0))|0);
   $40 = HEAP32[$$pre$phiZ2D>>2]|0;
   $41 = ((($40)) + -12|0);
   $42 = HEAP32[$41>>2]|0;
   $43 = (($$pre$phi3Z2D) + ($42)|0);
   __THREW__ = 0;
   invoke_vi(3377,($43|0));
   $44 = __THREW__; __THREW__ = 0;
   $45 = $44&1;
   if (!($45)) {
    ___cxa_end_catch();
    break;
   }
   $46 = ___cxa_find_matching_catch_2()|0;
   $47 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378);
   $48 = __THREW__; __THREW__ = 0;
   $49 = $48&1;
   if ($49) {
    $50 = ___cxa_find_matching_catch_3(0|0)|0;
    $51 = tempRet0;
    ___clang_call_terminate($50);
    // unreachable;
   } else {
    ___resumeException($46|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return ($this|0);
}
function __ZNSt3__114basic_iostreamIcNS_11char_traitsIcEEED2Ev($this,$vtt) {
 $this = $this|0;
 $vtt = $vtt|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__17collateIcED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__17collateIcED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNSt3__16locale5facet16__on_zero_sharedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($this|0)==(0|0);
 if (!($0)) {
  $1 = HEAP32[$this>>2]|0;
  $2 = ((($1)) + 4|0);
  $3 = HEAP32[$2>>2]|0;
  FUNCTION_TABLE_vi[$3 & 8191]($this);
 }
 return;
}
function __ZNKSt3__17collateIcE10do_compareEPKcS3_S3_S3_($this,$__lo1,$__hi1,$__lo2,$__hi2) {
 $this = $this|0;
 $__lo1 = $__lo1|0;
 $__hi1 = $__hi1|0;
 $__lo2 = $__lo2|0;
 $__hi2 = $__hi2|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $$02$lcssa = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$01 = $__lo2;$$02 = $__lo1;
 while(1) {
  $0 = ($$01|0)==($__hi2|0);
  if ($0) {
   $$02$lcssa = $$02;
   label = 7;
   break;
  }
  $1 = ($$02|0)==($__hi1|0);
  if ($1) {
   $$0 = -1;
   break;
  }
  $2 = HEAP8[$$02>>0]|0;
  $3 = HEAP8[$$01>>0]|0;
  $4 = ($2<<24>>24)<($3<<24>>24);
  if ($4) {
   $$0 = -1;
   break;
  }
  $5 = ($3<<24>>24)<($2<<24>>24);
  if ($5) {
   $$0 = 1;
   break;
  }
  $6 = ((($$02)) + 1|0);
  $7 = ((($$01)) + 1|0);
  $$01 = $7;$$02 = $6;
 }
 if ((label|0) == 7) {
  $8 = ($$02$lcssa|0)!=($__hi1|0);
  $9 = $8&1;
  $$0 = $9;
 }
 return ($$0|0);
}
function __ZNKSt3__17collateIcE12do_transformEPKcS3_($agg$result,$this,$__lo,$__hi) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $__lo = $__lo|0;
 $__hi = $__hi|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initIPKcEENS_9enable_ifIXsr21__is_forward_iteratorIT_EE5valueEvE4typeESA_SA_($agg$result,$__lo,$__hi);
 return;
}
function __ZNKSt3__17collateIcE7do_hashEPKcS3_($this,$__lo,$__hi) {
 $this = $this|0;
 $__lo = $__lo|0;
 $__hi = $__hi|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__h$0 = 0, $__h$0$lcssa = 0, $__p$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $__h$0 = 0;$__p$0 = $__lo;
 while(1) {
  $0 = ($__p$0|0)==($__hi|0);
  if ($0) {
   $__h$0$lcssa = $__h$0;
   break;
  }
  $1 = $__h$0 << 4;
  $2 = HEAP8[$__p$0>>0]|0;
  $3 = $2 << 24 >> 24;
  $4 = (($3) + ($1))|0;
  $5 = $4 & -268435456;
  $6 = $5 >>> 24;
  $7 = $6 | $5;
  $8 = $7 ^ $4;
  $9 = ((($__p$0)) + 1|0);
  $__h$0 = $8;$__p$0 = $9;
 }
 return ($__h$0$lcssa|0);
}
function __ZNSt3__17collateIwED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__17collateIwED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__17collateIwE10do_compareEPKwS3_S3_S3_($this,$__lo1,$__hi1,$__lo2,$__hi2) {
 $this = $this|0;
 $__lo1 = $__lo1|0;
 $__hi1 = $__hi1|0;
 $__lo2 = $__lo2|0;
 $__hi2 = $__hi2|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $$02$lcssa = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$01 = $__lo2;$$02 = $__lo1;
 while(1) {
  $0 = ($$01|0)==($__hi2|0);
  if ($0) {
   $$02$lcssa = $$02;
   label = 7;
   break;
  }
  $1 = ($$02|0)==($__hi1|0);
  if ($1) {
   $$0 = -1;
   break;
  }
  $2 = HEAP32[$$02>>2]|0;
  $3 = HEAP32[$$01>>2]|0;
  $4 = ($2|0)<($3|0);
  if ($4) {
   $$0 = -1;
   break;
  }
  $5 = ($3|0)<($2|0);
  if ($5) {
   $$0 = 1;
   break;
  }
  $6 = ((($$02)) + 4|0);
  $7 = ((($$01)) + 4|0);
  $$01 = $7;$$02 = $6;
 }
 if ((label|0) == 7) {
  $8 = ($$02$lcssa|0)!=($__hi1|0);
  $9 = $8&1;
  $$0 = $9;
 }
 return ($$0|0);
}
function __ZNKSt3__17collateIwE12do_transformEPKwS3_($agg$result,$this,$__lo,$__hi) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $__lo = $__lo|0;
 $__hi = $__hi|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6__initIPKwEENS_9enable_ifIXsr21__is_forward_iteratorIT_EE5valueEvE4typeESA_SA_($agg$result,$__lo,$__hi);
 return;
}
function __ZNKSt3__17collateIwE7do_hashEPKwS3_($this,$__lo,$__hi) {
 $this = $this|0;
 $__lo = $__lo|0;
 $__hi = $__hi|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $__h$0 = 0, $__h$0$lcssa = 0, $__p$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $__h$0 = 0;$__p$0 = $__lo;
 while(1) {
  $0 = ($__p$0|0)==($__hi|0);
  if ($0) {
   $__h$0$lcssa = $__h$0;
   break;
  }
  $1 = $__h$0 << 4;
  $2 = HEAP32[$__p$0>>2]|0;
  $3 = (($2) + ($1))|0;
  $4 = $3 & -268435456;
  $5 = $4 >>> 24;
  $6 = $5 | $4;
  $7 = $6 ^ $3;
  $8 = ((($__p$0)) + 4|0);
  $__h$0 = $7;$__p$0 = $8;
 }
 return ($__h$0$lcssa|0);
}
function __ZNSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRb($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0 = 0, $$02 = 0, $$03 = 0, $$byval_copy = 0, $$byval_copy1 = 0, $$byval_copy2 = 0, $$sroa$01$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__lv = 0, $__names = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy2 = sp + 60|0;
 $$byval_copy1 = sp + 56|0;
 $$byval_copy = sp + 52|0;
 $__lv = sp + 48|0;
 $0 = sp + 44|0;
 $1 = sp + 40|0;
 $2 = sp + 36|0;
 $3 = sp + 32|0;
 $__names = sp + 8|0;
 $4 = sp;
 $5 = ((($__iob)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $6 & 1;
 $8 = ($7|0)==(0);
 L1: do {
  if ($8) {
   HEAP32[$__lv>>2] = -1;
   $9 = HEAP32[$this>>2]|0;
   $10 = ((($9)) + 16|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = HEAP32[$__b>>2]|0;
   HEAP32[$0>>2] = $12;
   $13 = HEAP32[$__e>>2]|0;
   HEAP32[$1>>2] = $13;
   ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
   ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
   $14 = (FUNCTION_TABLE_iiiiiii[$11 & 8191]($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__lv)|0);
   HEAP32[$__b>>2] = $14;
   $15 = HEAP32[$__lv>>2]|0;
   switch ($15|0) {
   case 0:  {
    HEAP8[$__v>>0] = 0;
    break;
   }
   case 1:  {
    HEAP8[$__v>>0] = 1;
    break;
   }
   default: {
    HEAP8[$__v>>0] = 1;
    HEAP32[$__err>>2] = 4;
   }
   }
   $$sroa$01$0 = $14;
  } else {
   $16 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
   HEAP32[$2>>2] = $16;
   __THREW__ = 0;
   $17 = (invoke_iii(3374,($2|0),(1207576|0))|0);
   $18 = __THREW__; __THREW__ = 0;
   $19 = $18&1;
   do {
    if ($19) {
     $44 = ___cxa_find_matching_catch_2()|0;
     $45 = tempRet0;
     (__ZNSt3__114__shared_count16__release_sharedEv($16)|0);
     $$02 = $44;$$03 = $45;
    } else {
     (__ZNSt3__114__shared_count16__release_sharedEv($16)|0);
     $20 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
     HEAP32[$3>>2] = $20;
     __THREW__ = 0;
     $21 = (invoke_iii(3374,($3|0),(1207588|0))|0);
     $22 = __THREW__; __THREW__ = 0;
     $23 = $22&1;
     if ($23) {
      $46 = ___cxa_find_matching_catch_2()|0;
      $47 = tempRet0;
      (__ZNSt3__114__shared_count16__release_sharedEv($20)|0);
      $$02 = $46;$$03 = $47;
      break;
     }
     (__ZNSt3__114__shared_count16__release_sharedEv($20)|0);
     $24 = HEAP32[$21>>2]|0;
     $25 = ((($24)) + 24|0);
     $26 = HEAP32[$25>>2]|0;
     __THREW__ = 0;
     invoke_vii($26|0,($__names|0),($21|0));
     $27 = __THREW__; __THREW__ = 0;
     $28 = $27&1;
     if ($28) {
      $$0 = $__names;
     } else {
      $29 = ((($__names)) + 12|0);
      $30 = HEAP32[$21>>2]|0;
      $31 = ((($30)) + 28|0);
      $32 = HEAP32[$31>>2]|0;
      __THREW__ = 0;
      invoke_vii($32|0,($29|0),($21|0));
      $33 = __THREW__; __THREW__ = 0;
      $34 = $33&1;
      if ($34) {
       $$0 = $29;
      } else {
       $35 = HEAP32[$__e>>2]|0;
       HEAP32[$4>>2] = $35;
       $36 = ((($__names)) + 24|0);
       __THREW__ = 0;
       ;HEAP32[$$byval_copy2>>2]=HEAP32[$4>>2]|0;
       $37 = (invoke_iiiiiiii(4860,($__b|0),($$byval_copy2|0),($__names|0),($36|0),($17|0),($__err|0),1)|0);
       $38 = __THREW__; __THREW__ = 0;
       $39 = $38&1;
       if ($39) {
        $54 = ___cxa_find_matching_catch_2()|0;
        $55 = tempRet0;
        $56 = ((($__names)) + 12|0);
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($56);
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__names);
        $$02 = $54;$$03 = $55;
        break;
       } else {
        $40 = ($37|0)==($__names|0);
        $41 = $40&1;
        HEAP8[$__v>>0] = $41;
        $42 = HEAP32[$__b>>2]|0;
        $43 = ((($__names)) + 12|0);
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($43);
        __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__names);
        $$sroa$01$0 = $42;
        break L1;
       }
      }
     }
     $48 = ___cxa_find_matching_catch_2()|0;
     $49 = tempRet0;
     $50 = ($__names|0)==($$0|0);
     if ($50) {
      $$02 = $48;$$03 = $49;
     } else {
      $52 = $$0;
      while(1) {
       $51 = ((($52)) + -12|0);
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($51);
       $53 = ($51|0)==($__names|0);
       if ($53) {
        $$02 = $48;$$03 = $49;
        break;
       } else {
        $52 = $51;
       }
      }
     }
    }
   } while(0);
   ___resumeException($$02|0);
   // unreachable;
  }
 } while(0);
 STACKTOP = sp;return ($$sroa$01$0|0);
}
function __ZNKSt3__16locale9use_facetERNS0_2idE($this,$x) {
 $this = $this|0;
 $x = $x|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = (__ZNSt3__16locale2id5__getEv($x)|0);
 $2 = (__ZNKSt3__16locale5__imp9use_facetEl($0,$1)|0);
 return ($2|0);
}
function __ZNSt3__16locale2id5__getEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $__p$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__p$i = sp;
 $0 = HEAP32[$this>>2]|0;
 $1 = ($0|0)==(-1);
 if (!($1)) {
  HEAP32[$__p$i>>2] = $this;
  $2 = ((($__p$i)) + 4|0);
  HEAP32[$2>>2] = (4861);
  $3 = ((($__p$i)) + 8|0);
  HEAP32[$3>>2] = 0;
  __ZNSt3__111__call_onceERVmPvPFvS2_E($this,$__p$i,4862);
 }
 $4 = ((($this)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($5) + -1)|0;
 STACKTOP = sp;return ($6|0);
}
function __ZNSt3__16locale2id6__initEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[301896]|0;HEAP32[301896] = (($0+1)|0);
 $1 = (($0) + 1)|0;
 $2 = ((($this)) + 4|0);
 HEAP32[$2>>2] = $1;
 return;
}
function __ZNSt3__117__call_once_proxyINS_5tupleIJNS_12_GLOBAL__N_111__fake_bindEEEEEEvPv($__vp) {
 $__vp = $__vp|0;
 var $$elt$i$i$i$i = 0, $$elt1$i$i$i$i = 0, $$unpack$i$i$i$i = 0, $$unpack2$i$i$i$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$elt$i$i$i$i = ((($__vp)) + 4|0);
 $$unpack$i$i$i$i = HEAP32[$$elt$i$i$i$i>>2]|0;
 $$elt1$i$i$i$i = ((($__vp)) + 8|0);
 $$unpack2$i$i$i$i = HEAP32[$$elt1$i$i$i$i>>2]|0;
 $0 = HEAP32[$__vp>>2]|0;
 $1 = $$unpack2$i$i$i$i >> 1;
 $2 = (($0) + ($1)|0);
 $3 = $$unpack2$i$i$i$i & 1;
 $4 = ($3|0)==(0);
 if ($4) {
  $8 = $$unpack$i$i$i$i;
  $9 = $8;
 } else {
  $5 = HEAP32[$2>>2]|0;
  $6 = (($5) + ($$unpack$i$i$i$i)|0);
  $7 = HEAP32[$6>>2]|0;
  $9 = $7;
 }
 FUNCTION_TABLE_vi[$9 & 8191]($2);
 return;
}
function __ZNKSt3__16locale5__imp9use_facetEl($this,$id) {
 $this = $this|0;
 $id = $id|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNKSt3__16locale5__imp9has_facetEl($this,$id)|0);
 if ($0) {
  $2 = ((($this)) + 8|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = (($3) + ($id<<2)|0);
  $5 = HEAP32[$4>>2]|0;
  return ($5|0);
 } else {
  $1 = (___cxa_allocate_exception(4)|0);
  __ZNSt8bad_castC2Ev($1);
  ___cxa_throw(($1|0),(9152|0),(3336|0));
  // unreachable;
 }
 return (0)|0;
}
function __ZNKSt3__16locale5__imp9has_facetEl($this,$id) {
 $this = $this|0;
 $id = $id|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 $1 = ((($this)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = HEAP32[$0>>2]|0;
 $4 = (($2) - ($3))|0;
 $5 = $4 >> 2;
 $6 = ($5>>>0)>($id>>>0);
 if ($6) {
  $7 = $3;
  $8 = (($7) + ($id<<2)|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9|0)!=(0|0);
  $11 = $10;
 } else {
  $11 = 0;
 }
 return ($11|0);
}
function __ZNSt3__114__scan_keywordINS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEPKNS_12basic_stringIcS3_NS_9allocatorIcEEEENS_5ctypeIcEEEET0_RT_SE_SD_SD_RKT1_Rjb($__b,$__e,$__kb,$__ke,$__ct,$__err,$__case_sensitive) {
 $__b = $__b|0;
 $__e = $__e|0;
 $__kb = $__kb|0;
 $__ke = $__ke|0;
 $__ct = $__ct|0;
 $__err = $__err|0;
 $__case_sensitive = $__case_sensitive|0;
 var $$0 = 0, $$0$i$i = 0, $$041 = 0, $$lcssa = 0, $$lcssa99 = 0, $$pre$i$i$i = 0, $$pre$i$i$i7 = 0, $$pre$i2$i$i11 = 0, $$pre$i2$i$i11$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0;
 var $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0;
 var $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0;
 var $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__c$0 = 0, $__consume$0$off0 = 0, $__consume$2$off0 = 0, $__indx$0 = 0;
 var $__kc$0 = 0, $__ky$0 = 0, $__ky1$0 = 0, $__ky2$0 = 0, $__n_does_match$0 = 0, $__n_does_match$0$lcssa = 0, $__n_does_match$1 = 0, $__n_does_match$2 = 0, $__n_does_match$3 = 0, $__n_does_match$5 = 0, $__n_does_match$6 = 0, $__n_does_match$7 = 0, $__n_might_match$0 = 0, $__n_might_match$0$lcssa = 0, $__n_might_match$1 = 0, $__n_might_match$2 = 0, $__n_might_match$3 = 0, $__n_might_match$5 = 0, $__st$0 = 0, $__st$1 = 0;
 var $__st$2 = 0, $__st$3 = 0, $__stat_hold$sroa$0$0 = 0, $__stat_hold$sroa$0$0$ph$ph = 0, $__stat_hold$sroa$0$1 = 0, $__statbuf = 0, $__status$0 = 0, $lpad$phi$index = 0, $lpad$phi$index3 = 0, $tmp$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__statbuf = sp;
 $0 = $__ke;
 $1 = $__kb;
 $2 = (($0) - ($1))|0;
 $3 = (($2|0) / 12)&-1;
 $4 = ($3>>>0)>(100);
 do {
  if ($4) {
   $5 = (_malloc($3)|0);
   $6 = ($5|0)==(0|0);
   if ($6) {
    __THREW__ = 0;
    invoke_v(4863);
    $7 = __THREW__; __THREW__ = 0;
    $8 = $7&1;
    if ($8) {
     $__stat_hold$sroa$0$0$ph$ph = 0;
     label = 6;
     break;
    }
   }
   $17 = $5;
   $__stat_hold$sroa$0$1 = $17;$__status$0 = $5;
   label = 11;
  } else {
   $__stat_hold$sroa$0$1 = 0;$__status$0 = $__statbuf;
   label = 11;
  }
 } while(0);
 L6: do {
  if ((label|0) == 11) {
   $__ky$0 = $__kb;$__n_does_match$0 = 0;$__n_might_match$0 = $3;$__st$0 = $__status$0;
   while(1) {
    $18 = ($__ky$0|0)==($__ke|0);
    if ($18) {
     $__n_does_match$0$lcssa = $__n_does_match$0;$__n_might_match$0$lcssa = $__n_might_match$0;
     break;
    }
    $19 = HEAP8[$__ky$0>>0]|0;
    $20 = $19 & 1;
    $21 = ($20<<24>>24)==(0);
    if ($21) {
     $24 = $19&255;
     $25 = $24 >>> 1;
     $27 = $25;
    } else {
     $22 = ((($__ky$0)) + 4|0);
     $23 = HEAP32[$22>>2]|0;
     $27 = $23;
    }
    $26 = ($27|0)==(0);
    if ($26) {
     HEAP8[$__st$0>>0] = 2;
     $28 = (($__n_might_match$0) + -1)|0;
     $29 = (($__n_does_match$0) + 1)|0;
     $__n_does_match$1 = $29;$__n_might_match$1 = $28;
    } else {
     HEAP8[$__st$0>>0] = 1;
     $__n_does_match$1 = $__n_does_match$0;$__n_might_match$1 = $__n_might_match$0;
    }
    $30 = ((($__ky$0)) + 12|0);
    $31 = ((($__st$0)) + 1|0);
    $__ky$0 = $30;$__n_does_match$0 = $__n_does_match$1;$__n_might_match$0 = $__n_might_match$1;$__st$0 = $31;
   }
   $__indx$0 = 0;$__n_does_match$2 = $__n_does_match$0$lcssa;$__n_might_match$2 = $__n_might_match$0$lcssa;
   L20: while(1) {
    $32 = HEAP32[$__b>>2]|0;
    $33 = ($32|0)==(0|0);
    do {
     if ($33) {
      $47 = 0;
     } else {
      $34 = ((($32)) + 12|0);
      $35 = HEAP32[$34>>2]|0;
      $36 = ((($32)) + 16|0);
      $37 = HEAP32[$36>>2]|0;
      $38 = ($35|0)==($37|0);
      if ($38) {
       $39 = HEAP32[$32>>2]|0;
       $40 = ((($39)) + 36|0);
       $41 = HEAP32[$40>>2]|0;
       __THREW__ = 0;
       $42 = (invoke_ii($41|0,($32|0))|0);
       $43 = __THREW__; __THREW__ = 0;
       $44 = $43&1;
       if ($44) {
        label = 5;
        break L20;
       }
       $45 = ($42|0)==(-1);
       if ($45) {
        HEAP32[$__b>>2] = 0;
        $47 = 0;
        break;
       } else {
        $$pre$i$i$i = HEAP32[$__b>>2]|0;
        $47 = $$pre$i$i$i;
        break;
       }
      } else {
       $47 = $32;
      }
     }
    } while(0);
    $46 = ($47|0)==(0|0);
    $48 = HEAP32[$__e>>2]|0;
    $49 = ($48|0)==(0|0);
    if ($49) {
     $$pre$i2$i$i11 = 0;
    } else {
     $50 = ((($48)) + 12|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = ((($48)) + 16|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ($51|0)==($53|0);
     if ($54) {
      $55 = HEAP32[$48>>2]|0;
      $56 = ((($55)) + 36|0);
      $57 = HEAP32[$56>>2]|0;
      __THREW__ = 0;
      $58 = (invoke_ii($57|0,($48|0))|0);
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if ($60) {
       label = 5;
       break;
      }
      $61 = ($58|0)==(-1);
      if ($61) {
       HEAP32[$__e>>2] = 0;
       $$pre$i2$i$i11 = 0;
      } else {
       $$pre$i2$i$i11 = $48;
      }
     } else {
      $$pre$i2$i$i11 = $48;
     }
    }
    $62 = ($$pre$i2$i$i11|0)==(0|0);
    $tmp$i$i = $46 ^ $62;
    $63 = ($__n_might_match$2|0)!=(0);
    $64 = $63 & $tmp$i$i;
    $65 = HEAP32[$__b>>2]|0;
    if (!($64)) {
     $$lcssa = $62;$$lcssa99 = $65;$$pre$i2$i$i11$lcssa = $$pre$i2$i$i11;
     label = 33;
     break;
    }
    $93 = ((($65)) + 12|0);
    $94 = HEAP32[$93>>2]|0;
    $95 = ((($65)) + 16|0);
    $96 = HEAP32[$95>>2]|0;
    $97 = ($94|0)==($96|0);
    if ($97) {
     $98 = HEAP32[$65>>2]|0;
     $99 = ((($98)) + 36|0);
     $100 = HEAP32[$99>>2]|0;
     __THREW__ = 0;
     $101 = (invoke_ii($100|0,($65|0))|0);
     $102 = __THREW__; __THREW__ = 0;
     $103 = $102&1;
     if ($103) {
      label = 5;
      break;
     } else {
      $$0$i$i = $101;
     }
    } else {
     $104 = HEAP8[$94>>0]|0;
     $105 = $104&255;
     $$0$i$i = $105;
    }
    $106 = $$0$i$i&255;
    if ($__case_sensitive) {
     $__c$0 = $106;
    } else {
     $107 = HEAP32[$__ct>>2]|0;
     $108 = ((($107)) + 12|0);
     $109 = HEAP32[$108>>2]|0;
     __THREW__ = 0;
     $110 = (invoke_iii($109|0,($__ct|0),($106|0))|0);
     $111 = __THREW__; __THREW__ = 0;
     $112 = $111&1;
     if ($112) {
      label = 5;
      break;
     } else {
      $__c$0 = $110;
     }
    }
    $113 = (($__indx$0) + 1)|0;
    $__consume$0$off0 = 0;$__ky1$0 = $__kb;$__n_does_match$3 = $__n_does_match$2;$__n_might_match$3 = $__n_might_match$2;$__st$1 = $__status$0;
    while(1) {
     $114 = ($__ky1$0|0)==($__ke|0);
     if ($114) {
      break;
     }
     $115 = HEAP8[$__st$1>>0]|0;
     $116 = ($115<<24>>24)==(1);
     do {
      if ($116) {
       $117 = HEAP8[$__ky1$0>>0]|0;
       $118 = $117 & 1;
       $119 = ($118<<24>>24)==(0);
       if ($119) {
        $122 = ((($__ky1$0)) + 1|0);
        $124 = $122;
       } else {
        $120 = ((($__ky1$0)) + 8|0);
        $121 = HEAP32[$120>>2]|0;
        $124 = $121;
       }
       $123 = (($124) + ($__indx$0)|0);
       $125 = HEAP8[$123>>0]|0;
       if ($__case_sensitive) {
        $__kc$0 = $125;
       } else {
        $126 = HEAP32[$__ct>>2]|0;
        $127 = ((($126)) + 12|0);
        $128 = HEAP32[$127>>2]|0;
        __THREW__ = 0;
        $129 = (invoke_iii($128|0,($__ct|0),($125|0))|0);
        $130 = __THREW__; __THREW__ = 0;
        $131 = $130&1;
        if ($131) {
         label = 4;
         break L20;
        } else {
         $__kc$0 = $129;
        }
       }
       $132 = ($__c$0<<24>>24)==($__kc$0<<24>>24);
       if (!($132)) {
        HEAP8[$__st$1>>0] = 0;
        $144 = (($__n_might_match$3) + -1)|0;
        $__consume$2$off0 = $__consume$0$off0;$__n_does_match$5 = $__n_does_match$3;$__n_might_match$5 = $144;
        break;
       }
       $133 = HEAP8[$__ky1$0>>0]|0;
       $134 = $133 & 1;
       $135 = ($134<<24>>24)==(0);
       if ($135) {
        $138 = $133&255;
        $139 = $138 >>> 1;
        $141 = $139;
       } else {
        $136 = ((($__ky1$0)) + 4|0);
        $137 = HEAP32[$136>>2]|0;
        $141 = $137;
       }
       $140 = ($141|0)==($113|0);
       if ($140) {
        HEAP8[$__st$1>>0] = 2;
        $142 = (($__n_might_match$3) + -1)|0;
        $143 = (($__n_does_match$3) + 1)|0;
        $__consume$2$off0 = 1;$__n_does_match$5 = $143;$__n_might_match$5 = $142;
       } else {
        $__consume$2$off0 = 1;$__n_does_match$5 = $__n_does_match$3;$__n_might_match$5 = $__n_might_match$3;
       }
      } else {
       $__consume$2$off0 = $__consume$0$off0;$__n_does_match$5 = $__n_does_match$3;$__n_might_match$5 = $__n_might_match$3;
      }
     } while(0);
     $145 = ((($__ky1$0)) + 12|0);
     $146 = ((($__st$1)) + 1|0);
     $__consume$0$off0 = $__consume$2$off0;$__ky1$0 = $145;$__n_does_match$3 = $__n_does_match$5;$__n_might_match$3 = $__n_might_match$5;$__st$1 = $146;
    }
    if (!($__consume$0$off0)) {
     $__indx$0 = $113;$__n_does_match$2 = $__n_does_match$3;$__n_might_match$2 = $__n_might_match$3;
     continue;
    }
    $147 = HEAP32[$__b>>2]|0;
    $148 = ((($147)) + 12|0);
    $149 = HEAP32[$148>>2]|0;
    $150 = ((($147)) + 16|0);
    $151 = HEAP32[$150>>2]|0;
    $152 = ($149|0)==($151|0);
    if ($152) {
     $153 = HEAP32[$147>>2]|0;
     $154 = ((($153)) + 40|0);
     $155 = HEAP32[$154>>2]|0;
     __THREW__ = 0;
     (invoke_ii($155|0,($147|0))|0);
     $156 = __THREW__; __THREW__ = 0;
     $157 = $156&1;
     if ($157) {
      label = 5;
      break;
     }
    } else {
     $158 = ((($149)) + 1|0);
     HEAP32[$148>>2] = $158;
    }
    $159 = (($__n_does_match$3) + ($__n_might_match$3))|0;
    $160 = ($159>>>0)>(1);
    if ($160) {
     $__ky2$0 = $__kb;$__n_does_match$6 = $__n_does_match$3;$__st$2 = $__status$0;
    } else {
     $__indx$0 = $113;$__n_does_match$2 = $__n_does_match$3;$__n_might_match$2 = $__n_might_match$3;
     continue;
    }
    while(1) {
     $161 = ($__ky2$0|0)==($__ke|0);
     if ($161) {
      $__indx$0 = $113;$__n_does_match$2 = $__n_does_match$6;$__n_might_match$2 = $__n_might_match$3;
      continue L20;
     }
     $162 = HEAP8[$__st$2>>0]|0;
     $163 = ($162<<24>>24)==(2);
     if ($163) {
      $164 = HEAP8[$__ky2$0>>0]|0;
      $165 = $164 & 1;
      $166 = ($165<<24>>24)==(0);
      if ($166) {
       $169 = $164&255;
       $170 = $169 >>> 1;
       $172 = $170;
      } else {
       $167 = ((($__ky2$0)) + 4|0);
       $168 = HEAP32[$167>>2]|0;
       $172 = $168;
      }
      $171 = ($172|0)==($113|0);
      if ($171) {
       $__n_does_match$7 = $__n_does_match$6;
      } else {
       HEAP8[$__st$2>>0] = 0;
       $173 = (($__n_does_match$6) + -1)|0;
       $__n_does_match$7 = $173;
      }
     } else {
      $__n_does_match$7 = $__n_does_match$6;
     }
     $174 = ((($__ky2$0)) + 12|0);
     $175 = ((($__st$2)) + 1|0);
     $__ky2$0 = $174;$__n_does_match$6 = $__n_does_match$7;$__st$2 = $175;
    }
   }
   if ((label|0) == 4) {
    $9 = ___cxa_find_matching_catch_2()|0;
    $10 = tempRet0;
    $__stat_hold$sroa$0$0 = $__stat_hold$sroa$0$1;$lpad$phi$index = $9;$lpad$phi$index3 = $10;
    break;
   }
   else if ((label|0) == 5) {
    $11 = ___cxa_find_matching_catch_2()|0;
    $12 = tempRet0;
    $__stat_hold$sroa$0$0 = $__stat_hold$sroa$0$1;$lpad$phi$index = $11;$lpad$phi$index3 = $12;
    break;
   }
   else if ((label|0) == 33) {
    $66 = ($$lcssa99|0)==(0|0);
    do {
     if ($66) {
      $80 = 0;
     } else {
      $67 = ((($$lcssa99)) + 12|0);
      $68 = HEAP32[$67>>2]|0;
      $69 = ((($$lcssa99)) + 16|0);
      $70 = HEAP32[$69>>2]|0;
      $71 = ($68|0)==($70|0);
      if ($71) {
       $72 = HEAP32[$$lcssa99>>2]|0;
       $73 = ((($72)) + 36|0);
       $74 = HEAP32[$73>>2]|0;
       __THREW__ = 0;
       $75 = (invoke_ii($74|0,($$lcssa99|0))|0);
       $76 = __THREW__; __THREW__ = 0;
       $77 = $76&1;
       if ($77) {
        $__stat_hold$sroa$0$0$ph$ph = $__stat_hold$sroa$0$1;
        label = 6;
        break L6;
       }
       $78 = ($75|0)==(-1);
       if ($78) {
        HEAP32[$__b>>2] = 0;
        $80 = 0;
        break;
       } else {
        $$pre$i$i$i7 = HEAP32[$__b>>2]|0;
        $80 = $$pre$i$i$i7;
        break;
       }
      } else {
       $80 = $$lcssa99;
      }
     }
    } while(0);
    $79 = ($80|0)==(0|0);
    do {
     if ($$lcssa) {
      label = 79;
     } else {
      $81 = ((($$pre$i2$i$i11$lcssa)) + 12|0);
      $82 = HEAP32[$81>>2]|0;
      $83 = ((($$pre$i2$i$i11$lcssa)) + 16|0);
      $84 = HEAP32[$83>>2]|0;
      $85 = ($82|0)==($84|0);
      if ($85) {
       $86 = HEAP32[$$pre$i2$i$i11$lcssa>>2]|0;
       $87 = ((($86)) + 36|0);
       $88 = HEAP32[$87>>2]|0;
       __THREW__ = 0;
       $89 = (invoke_ii($88|0,($$pre$i2$i$i11$lcssa|0))|0);
       $90 = __THREW__; __THREW__ = 0;
       $91 = $90&1;
       if ($91) {
        $__stat_hold$sroa$0$0$ph$ph = $__stat_hold$sroa$0$1;
        label = 6;
        break L6;
       }
       $92 = ($89|0)==(-1);
       if ($92) {
        HEAP32[$__e>>2] = 0;
        label = 79;
        break;
       }
      }
      if ($79) {
       $$0 = $__kb;$__st$3 = $__status$0;
      } else {
       label = 80;
      }
     }
    } while(0);
    if ((label|0) == 79) {
     if ($79) {
      label = 80;
     } else {
      $$0 = $__kb;$__st$3 = $__status$0;
     }
    }
    if ((label|0) == 80) {
     $176 = HEAP32[$__err>>2]|0;
     $177 = $176 | 2;
     HEAP32[$__err>>2] = $177;
     $$0 = $__kb;$__st$3 = $__status$0;
    }
    while(1) {
     $178 = ($$0|0)==($__ke|0);
     if ($178) {
      label = 84;
      break;
     }
     $179 = HEAP8[$__st$3>>0]|0;
     $180 = ($179<<24>>24)==(2);
     if ($180) {
      $$041 = $$0;
      break;
     }
     $181 = ((($$0)) + 12|0);
     $182 = ((($__st$3)) + 1|0);
     $$0 = $181;$__st$3 = $182;
    }
    if ((label|0) == 84) {
     $183 = HEAP32[$__err>>2]|0;
     $184 = $183 | 4;
     HEAP32[$__err>>2] = $184;
     $$041 = $__ke;
    }
    $185 = ($__stat_hold$sroa$0$1|0)==(0);
    if (!($185)) {
     $186 = $__stat_hold$sroa$0$1;
     _free($186);
    }
    STACKTOP = sp;return ($$041|0);
   }
  }
 } while(0);
 if ((label|0) == 6) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  $__stat_hold$sroa$0$0 = $__stat_hold$sroa$0$0$ph$ph;$lpad$phi$index = $13;$lpad$phi$index3 = $14;
 }
 $15 = ($__stat_hold$sroa$0$0|0)==(0);
 if (!($15)) {
  $16 = $__stat_hold$sroa$0$0;
  _free($16);
 }
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRl($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE15__do_get_signedIlEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE15__do_get_signedIlEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i = 0, $$0$i$i = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i4 = 0, $$pre$i2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__g = 0, $__g_end = 0, $__grouping = 0, $__i$0$i$i = 0, $__thousands_sep = 0, $exitcond$i$i = 0, $lpad$phi$index = 0;
 var $lpad$phi$index2 = 0, $phitmp$i = 0, $phitmp$i19 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 198|0;
 $__thousands_sep = sp + 196|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $0 = ((($__iob)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 74;
 switch ($2|0) {
 case 64:  {
  $$0$i = 8;
  break;
 }
 case 8:  {
  $$0$i = 16;
  break;
 }
 case 0:  {
  $$0$i = 0;
  break;
 }
 default: {
  $$0$i = 10;
 }
 }
 __ZNSt3__19__num_getIcE17__stage2_int_prepERNS_8ios_baseEPcRc($__grouping,$__iob,$__atoms,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $3 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$3>>2] = 0;
  $4 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $4;
 }
 $5 = HEAP8[$__buf>>0]|0;
 $6 = $5 & 1;
 $7 = ($6<<24>>24)==(0);
 if ($7) {
  $10 = 10;
 } else {
  $8 = HEAP32[$__buf>>2]|0;
  $9 = $8 & -2;
  $phitmp$i = (($9) + -1)|0;
  $10 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($10|0),0);
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 L13: do {
  if ($12) {
   label = 32;
  } else {
   $13 = HEAP8[$__buf>>0]|0;
   $14 = $13 & 1;
   $15 = ($14<<24>>24)==(0);
   $16 = ((($__buf)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($__buf)) + 1|0);
   $19 = $15 ? $18 : $17;
   HEAP32[$__a_end>>2] = $19;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   $20 = ((($__buf)) + 4|0);
   $21 = HEAP8[$__thousands_sep>>0]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i4 = $$pre;$__a$0 = $19;
   L15: while(1) {
    $22 = ($$pre$i$i$i4|0)==(0|0);
    if ($22) {
     $$pre$i$i$i = 0;
    } else {
     $23 = ((($$pre$i$i$i4)) + 12|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ((($$pre$i$i$i4)) + 16|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = ($24|0)==($26|0);
     if ($27) {
      $28 = HEAP32[$$pre$i$i$i4>>2]|0;
      $29 = ((($28)) + 36|0);
      $30 = HEAP32[$29>>2]|0;
      __THREW__ = 0;
      $31 = (invoke_ii($30|0,($$pre$i$i$i4|0))|0);
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if ($33) {
       label = 31;
       break;
      }
      $34 = ($31|0)==(-1);
      if ($34) {
       HEAP32[$__b>>2] = 0;
       $$pre$i$i$i = 0;
      } else {
       $$pre$i$i$i = $$pre$i$i$i4;
      }
     } else {
      $$pre$i$i$i = $$pre$i$i$i4;
     }
    }
    $35 = ($$pre$i$i$i|0)==(0|0);
    $36 = HEAP32[$__e>>2]|0;
    $37 = ($36|0)==(0|0);
    do {
     if ($37) {
      label = 23;
     } else {
      $38 = ((($36)) + 12|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ((($36)) + 16|0);
      $41 = HEAP32[$40>>2]|0;
      $42 = ($39|0)==($41|0);
      if (!($42)) {
       if ($35) {
        $158 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
      $43 = HEAP32[$36>>2]|0;
      $44 = ((($43)) + 36|0);
      $45 = HEAP32[$44>>2]|0;
      __THREW__ = 0;
      $46 = (invoke_ii($45|0,($36|0))|0);
      $47 = __THREW__; __THREW__ = 0;
      $48 = $47&1;
      if ($48) {
       label = 31;
       break L15;
      }
      $49 = ($46|0)==(-1);
      if ($49) {
       HEAP32[$__e>>2] = 0;
       label = 23;
       break;
      } else {
       if ($35) {
        $158 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
     }
    } while(0);
    if ((label|0) == 23) {
     label = 0;
     if ($35) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $158 = 0;
     }
    }
    $50 = HEAP32[$__a_end>>2]|0;
    $51 = HEAP8[$__buf>>0]|0;
    $52 = $51 & 1;
    $53 = ($52<<24>>24)==(0);
    $54 = HEAP32[$20>>2]|0;
    $55 = $51&255;
    $56 = $55 >>> 1;
    $57 = $53 ? $56 : $54;
    $58 = (($__a$0) + ($57)|0);
    $59 = ($50|0)==($58|0);
    if ($59) {
     $60 = $57 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($60|0),0);
     $61 = __THREW__; __THREW__ = 0;
     $62 = $61&1;
     if ($62) {
      label = 31;
      break;
     }
     $63 = HEAP8[$__buf>>0]|0;
     $64 = $63 & 1;
     $65 = ($64<<24>>24)==(0);
     if ($65) {
      $68 = 10;
     } else {
      $66 = HEAP32[$__buf>>2]|0;
      $67 = $66 & -2;
      $phitmp$i19 = (($67) + -1)|0;
      $68 = $phitmp$i19;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($68|0),0);
     $69 = __THREW__; __THREW__ = 0;
     $70 = $69&1;
     if ($70) {
      label = 31;
      break;
     }
     $71 = HEAP8[$__buf>>0]|0;
     $72 = $71 & 1;
     $73 = ($72<<24>>24)==(0);
     $74 = HEAP32[$16>>2]|0;
     $75 = $73 ? $18 : $74;
     $76 = (($75) + ($57)|0);
     HEAP32[$__a_end>>2] = $76;
     $__a$1 = $75;
    } else {
     $__a$1 = $__a$0;
    }
    $81 = ((($$pre$i$i$i)) + 12|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = ((($$pre$i$i$i)) + 16|0);
    $84 = HEAP32[$83>>2]|0;
    $85 = ($82|0)==($84|0);
    if ($85) {
     $86 = HEAP32[$$pre$i$i$i>>2]|0;
     $87 = ((($86)) + 36|0);
     $88 = HEAP32[$87>>2]|0;
     __THREW__ = 0;
     $89 = (invoke_ii($88|0,($$pre$i$i$i|0))|0);
     $90 = __THREW__; __THREW__ = 0;
     $91 = $90&1;
     if ($91) {
      label = 31;
      break;
     } else {
      $$0$i$i = $89;
     }
    } else {
     $92 = HEAP8[$82>>0]|0;
     $93 = $92&255;
     $$0$i$i = $93;
    }
    $94 = $$0$i$i&255;
    $95 = (__ZNSt3__19__num_getIcE17__stage2_int_loopEciPcRS2_RjcRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_S2_($94,$$0$i,$__a$1,$__a_end,$__dc,$21,$__grouping,$__g,$__g_end,$__atoms)|0);
    $96 = ($95|0)==(0);
    if (!($96)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $158;$__a$2 = $__a$1;
     break;
    }
    $97 = HEAP32[$81>>2]|0;
    $98 = HEAP32[$83>>2]|0;
    $99 = ($97|0)==($98|0);
    if ($99) {
     $100 = HEAP32[$$pre$i$i$i>>2]|0;
     $101 = ((($100)) + 40|0);
     $102 = HEAP32[$101>>2]|0;
     __THREW__ = 0;
     (invoke_ii($102|0,($$pre$i$i$i|0))|0);
     $103 = __THREW__; __THREW__ = 0;
     $104 = $103&1;
     if ($104) {
      label = 31;
      break;
     } else {
      $$pre$i$i$i4 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $105 = ((($97)) + 1|0);
     HEAP32[$81>>2] = $105;
     $$pre$i$i$i4 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 31) {
    $77 = ___cxa_find_matching_catch_2()|0;
    $78 = tempRet0;
    $lpad$phi$index = $77;$lpad$phi$index2 = $78;
    break;
   }
   $106 = HEAP8[$__grouping>>0]|0;
   $107 = $106 & 1;
   $108 = ($107<<24>>24)==(0);
   $109 = ((($__grouping)) + 4|0);
   $110 = HEAP32[$109>>2]|0;
   $111 = $106&255;
   $112 = $111 >>> 1;
   $113 = $108 ? $112 : $110;
   $114 = ($113|0)==(0);
   if (!($114)) {
    $115 = HEAP32[$__g_end>>2]|0;
    $116 = $115;
    $117 = $__g;
    $118 = (($116) - ($117))|0;
    $119 = ($118|0)<(160);
    if ($119) {
     $120 = HEAP32[$__dc>>2]|0;
     $121 = ((($115)) + 4|0);
     HEAP32[$__g_end>>2] = $121;
     HEAP32[$115>>2] = $120;
    }
   }
   $122 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $123 = (invoke_iiiii(4864,($__a$2|0),($122|0),($__err|0),($$0$i|0))|0);
   $124 = __THREW__; __THREW__ = 0;
   $125 = $124&1;
   if ($125) {
    label = 32;
   } else {
    HEAP32[$__v>>2] = $123;
    $126 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$126,$__err);
    $127 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($127) {
     $141 = 0;
    } else {
     $128 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $129 = HEAP32[$128>>2]|0;
     $130 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $131 = HEAP32[$130>>2]|0;
     $132 = ($129|0)==($131|0);
     if ($132) {
      $133 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $134 = ((($133)) + 36|0);
      $135 = HEAP32[$134>>2]|0;
      __THREW__ = 0;
      $136 = (invoke_ii($135|0,($$pre$i$i$i$lcssa|0))|0);
      $137 = __THREW__; __THREW__ = 0;
      $138 = $137&1;
      if ($138) {
       label = 32;
       break;
      }
      $139 = ($136|0)==(-1);
      if ($139) {
       HEAP32[$__b>>2] = 0;
       $141 = 0;
      } else {
       $141 = $$pre$i$i$i$lcssa;
      }
     } else {
      $141 = $$pre$i$i$i$lcssa;
     }
    }
    $140 = ($141|0)==(0|0);
    $142 = ($$pre$i2$i$i|0)==(0|0);
    do {
     if ($142) {
      label = 56;
     } else {
      $143 = ((($$pre$i2$i$i)) + 12|0);
      $144 = HEAP32[$143>>2]|0;
      $145 = ((($$pre$i2$i$i)) + 16|0);
      $146 = HEAP32[$145>>2]|0;
      $147 = ($144|0)==($146|0);
      if ($147) {
       $148 = HEAP32[$$pre$i2$i$i>>2]|0;
       $149 = ((($148)) + 36|0);
       $150 = HEAP32[$149>>2]|0;
       __THREW__ = 0;
       $151 = (invoke_ii($150|0,($$pre$i2$i$i|0))|0);
       $152 = __THREW__; __THREW__ = 0;
       $153 = $152&1;
       if ($153) {
        label = 32;
        break L13;
       }
       $154 = ($151|0)==(-1);
       if ($154) {
        HEAP32[$__e>>2] = 0;
        label = 56;
        break;
       }
      }
      if (!($140)) {
       label = 57;
      }
     }
    } while(0);
    if ((label|0) == 56) {
     if ($140) {
      label = 57;
     }
    }
    if ((label|0) == 57) {
     $155 = HEAP32[$__err>>2]|0;
     $156 = $155 | 2;
     HEAP32[$__err>>2] = $156;
    }
    $157 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($157|0);
   }
  }
 } while(0);
 if ((label|0) == 32) {
  $79 = ___cxa_find_matching_catch_2()|0;
  $80 = tempRet0;
  $lpad$phi$index = $79;$lpad$phi$index2 = $80;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__19__num_getIcE17__stage2_int_prepERNS_8ios_baseEPcRc($agg$result,$__iob,$__atoms,$__thousands_sep) {
 $agg$result = $agg$result|0;
 $__iob = $__iob|0;
 $__atoms = $__atoms|0;
 $__thousands_sep = $__thousands_sep|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__loc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__loc = sp;
 $0 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$__loc>>2] = $0;
 __THREW__ = 0;
 $1 = (invoke_iii(3374,($__loc|0),(1207576|0))|0);
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  $4 = HEAP32[$1>>2]|0;
  $5 = ((($4)) + 32|0);
  $6 = HEAP32[$5>>2]|0;
  __THREW__ = 0;
  (invoke_iiiii($6|0,($1|0),(620486|0),((620512)|0),($__atoms|0))|0);
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if (!($8)) {
   __THREW__ = 0;
   $9 = (invoke_iii(3374,($__loc|0),(1207588|0))|0);
   $10 = __THREW__; __THREW__ = 0;
   $11 = $10&1;
   if (!($11)) {
    $12 = HEAP32[$9>>2]|0;
    $13 = ((($12)) + 16|0);
    $14 = HEAP32[$13>>2]|0;
    __THREW__ = 0;
    $15 = (invoke_ii($14|0,($9|0))|0);
    $16 = __THREW__; __THREW__ = 0;
    $17 = $16&1;
    if (!($17)) {
     HEAP8[$__thousands_sep>>0] = $15;
     $18 = HEAP32[$9>>2]|0;
     $19 = ((($18)) + 20|0);
     $20 = HEAP32[$19>>2]|0;
     __THREW__ = 0;
     invoke_vii($20|0,($agg$result|0),($9|0));
     $21 = __THREW__; __THREW__ = 0;
     $22 = $21&1;
     if (!($22)) {
      (__ZNSt3__114__shared_count16__release_sharedEv($0)|0);
      STACKTOP = sp;return;
     }
    }
   }
  }
 }
 $23 = ___cxa_find_matching_catch_2()|0;
 $24 = tempRet0;
 (__ZNSt3__114__shared_count16__release_sharedEv($0)|0);
 ___resumeException($23|0);
 // unreachable;
}
function __ZNSt3__19__num_getIcE17__stage2_int_loopEciPcRS2_RjcRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_S2_($__ct,$__base,$__a,$__a_end,$__dc,$__thousands_sep,$__grouping,$__g,$__g_end,$__atoms) {
 $__ct = $__ct|0;
 $__base = $__base|0;
 $__a = $__a|0;
 $__a_end = $__a_end|0;
 $__dc = $__dc|0;
 $__thousands_sep = $__thousands_sep|0;
 $__grouping = $__grouping|0;
 $__g = $__g|0;
 $__g_end = $__g_end|0;
 $__atoms = $__atoms|0;
 var $$0$i = 0, $$0$lcssa$i = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$__a_end>>2]|0;
 $1 = ($0|0)==($__a|0);
 do {
  if ($1) {
   $2 = ((($__atoms)) + 24|0);
   $3 = HEAP8[$2>>0]|0;
   $4 = ($3<<24>>24)==($__ct<<24>>24);
   if (!($4)) {
    $5 = ((($__atoms)) + 25|0);
    $6 = HEAP8[$5>>0]|0;
    $7 = ($6<<24>>24)==($__ct<<24>>24);
    if (!($7)) {
     label = 5;
     break;
    }
   }
   $8 = $4 ? 43 : 45;
   $9 = ((($__a)) + 1|0);
   HEAP32[$__a_end>>2] = $9;
   HEAP8[$__a>>0] = $8;
   HEAP32[$__dc>>2] = 0;
   $$1 = 0;
  } else {
   label = 5;
  }
 } while(0);
 L6: do {
  if ((label|0) == 5) {
   $10 = HEAP8[$__grouping>>0]|0;
   $11 = $10 & 1;
   $12 = ($11<<24>>24)==(0);
   $13 = ((($__grouping)) + 4|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = $10&255;
   $16 = $15 >>> 1;
   $17 = $12 ? $16 : $14;
   $18 = ($17|0)!=(0);
   $19 = ($__ct<<24>>24)==($__thousands_sep<<24>>24);
   $or$cond = $19 & $18;
   if ($or$cond) {
    $20 = HEAP32[$__g_end>>2]|0;
    $21 = $20;
    $22 = $__g;
    $23 = (($21) - ($22))|0;
    $24 = ($23|0)<(160);
    if (!($24)) {
     $$1 = 0;
     break;
    }
    $25 = HEAP32[$__dc>>2]|0;
    $26 = ((($20)) + 4|0);
    HEAP32[$__g_end>>2] = $26;
    HEAP32[$20>>2] = $25;
    HEAP32[$__dc>>2] = 0;
    $$1 = 0;
    break;
   }
   $27 = ((($__atoms)) + 26|0);
   $$0$i = $__atoms;
   while(1) {
    $28 = ($$0$i|0)==($27|0);
    if ($28) {
     $$0$lcssa$i = $27;
     break;
    }
    $29 = HEAP8[$$0$i>>0]|0;
    $30 = ($29<<24>>24)==($__ct<<24>>24);
    if ($30) {
     $$0$lcssa$i = $$0$i;
     break;
    }
    $31 = ((($$0$i)) + 1|0);
    $$0$i = $31;
   }
   $32 = $$0$lcssa$i;
   $33 = $__atoms;
   $34 = (($32) - ($33))|0;
   $35 = ($34|0)>(23);
   if ($35) {
    $$1 = -1;
   } else {
    switch ($__base|0) {
    case 10: case 8:  {
     $36 = ($34|0)<($__base|0);
     if (!($36)) {
      $$1 = -1;
      break L6;
     }
     break;
    }
    case 16:  {
     $37 = ($34|0)<(22);
     if (!($37)) {
      if ($1) {
       $$1 = -1;
       break L6;
      }
      $38 = $0;
      $39 = $__a;
      $40 = (($38) - ($39))|0;
      $41 = ($40|0)<(3);
      if (!($41)) {
       $$1 = -1;
       break L6;
      }
      $42 = ((($0)) + -1|0);
      $43 = HEAP8[$42>>0]|0;
      $44 = ($43<<24>>24)==(48);
      if (!($44)) {
       $$1 = -1;
       break L6;
      }
      HEAP32[$__dc>>2] = 0;
      $45 = (620486 + ($34)|0);
      $46 = HEAP8[$45>>0]|0;
      $47 = ((($0)) + 1|0);
      HEAP32[$__a_end>>2] = $47;
      HEAP8[$0>>0] = $46;
      $$1 = 0;
      break L6;
     }
     break;
    }
    default: {
    }
    }
    $48 = (620486 + ($34)|0);
    $49 = HEAP8[$48>>0]|0;
    $50 = ((($0)) + 1|0);
    HEAP32[$__a_end>>2] = $50;
    HEAP8[$0>>0] = $49;
    $51 = HEAP32[$__dc>>2]|0;
    $52 = (($51) + 1)|0;
    HEAP32[$__dc>>2] = $52;
    $$1 = 0;
   }
  }
 } while(0);
 return ($$1|0);
}
function __ZNSt3__125__num_get_signed_integralIlEET_PKcS3_Rji($__a,$__a_end,$__err,$__base) {
 $__a = $__a|0;
 $__a_end = $__a_end|0;
 $__err = $__err|0;
 $__base = $__base|0;
 var $$0 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__p2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__p2 = sp;
 $0 = ($__a|0)==($__a_end|0);
 if ($0) {
  HEAP32[$__err>>2] = 4;
  $$1 = 0;
 } else {
  $1 = (___errno_location()|0);
  $2 = HEAP32[$1>>2]|0;
  HEAP32[$1>>2] = 0;
  $3 = (__ZNSt3__16__clocEv()|0);
  $4 = (_strtoll_l($__a,$__p2,$__base,$3)|0);
  $5 = tempRet0;
  $6 = HEAP32[$1>>2]|0;
  $7 = ($6|0)==(0);
  if ($7) {
   HEAP32[$1>>2] = $2;
  }
  $8 = HEAP32[$__p2>>2]|0;
  $9 = ($8|0)==($__a_end|0);
  L7: do {
   if ($9) {
    $10 = ($6|0)==(34);
    do {
     if ($10) {
      HEAP32[$__err>>2] = 4;
      $21 = ($5|0)>(0);
      $22 = ($4>>>0)>(0);
      $23 = ($5|0)==(0);
      $24 = $23 & $22;
      $25 = $21 | $24;
      if ($25) {
       $$0 = 2147483647;
       break L7;
      }
     } else {
      $11 = ($5|0)<(-1);
      $12 = ($4>>>0)<(2147483648);
      $13 = ($5|0)==(-1);
      $14 = $13 & $12;
      $15 = $11 | $14;
      if ($15) {
       HEAP32[$__err>>2] = 4;
       break;
      }
      $16 = ($5|0)>(0);
      $17 = ($4>>>0)>(2147483647);
      $18 = ($5|0)==(0);
      $19 = $18 & $17;
      $20 = $16 | $19;
      if ($20) {
       HEAP32[$__err>>2] = 4;
       $$0 = 2147483647;
       break L7;
      } else {
       $$0 = $4;
       break L7;
      }
     }
    } while(0);
    $$0 = -2147483648;
   } else {
    HEAP32[$__err>>2] = 4;
    $$0 = 0;
   }
  } while(0);
  $$1 = $$0;
 }
 STACKTOP = sp;return ($$1|0);
}
function __ZNSt3__16__clocEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[624952]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((624952|0))|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    __THREW__ = 0;
    $4 = (invoke_iiii(4865,2147483647,(620519|0),(0|0))|0);
    $5 = __THREW__; __THREW__ = 0;
    $6 = $5&1;
    if ($6) {
     $8 = ___cxa_find_matching_catch_2()|0;
     $9 = tempRet0;
     ___cxa_guard_abort((624952|0));
     ___resumeException($8|0);
     // unreachable;
    } else {
     HEAP32[301899] = $4;
     ___cxa_guard_release((624952|0));
     break;
    }
   }
  }
 } while(0);
 $7 = HEAP32[301899]|0;
 return ($7|0);
}
function __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$__g_end,$__err) {
 $__grouping = $__grouping|0;
 $__g = $__g|0;
 $__g_end = $__g_end|0;
 $__err = $__err|0;
 var $$0$i$i = 0, $$01$i$i = 0, $$__ig$0 = 0, $$not = 0, $$pre = 0, $$pre11 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__ig$0 = 0, $__r$0 = 0, $or$cond2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$__grouping>>0]|0;
 $1 = $0 & 1;
 $2 = ($1<<24>>24)==(0);
 $3 = ((($__grouping)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $0&255;
 $6 = $5 >>> 1;
 $7 = $2 ? $6 : $4;
 $8 = ($7|0)==(0);
 do {
  if (!($8)) {
   $9 = ($__g|0)==($__g_end|0);
   if ($9) {
    $16 = $0;$25 = $4;
   } else {
    $$0$i$i = $__g_end;$$01$i$i = $__g;
    while(1) {
     $10 = ((($$0$i$i)) + -4|0);
     $11 = ($$01$i$i>>>0)<($10>>>0);
     if (!($11)) {
      break;
     }
     $12 = HEAP32[$$01$i$i>>2]|0;
     $13 = HEAP32[$10>>2]|0;
     HEAP32[$$01$i$i>>2] = $13;
     HEAP32[$10>>2] = $12;
     $14 = ((($$01$i$i)) + 4|0);
     $$0$i$i = $10;$$01$i$i = $14;
    }
    $$pre = HEAP8[$__grouping>>0]|0;
    $$pre11 = HEAP32[$3>>2]|0;
    $16 = $$pre;$25 = $$pre11;
   }
   $15 = $16 & 1;
   $17 = ($15<<24>>24)==(0);
   $18 = ((($__grouping)) + 8|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ((($__grouping)) + 1|0);
   $21 = $17 ? $20 : $19;
   $22 = $16&255;
   $23 = $22 >>> 1;
   $24 = $17 ? $23 : $25;
   $26 = (($21) + ($24)|0);
   $27 = ((($__g_end)) + -4|0);
   $28 = $26;
   $__ig$0 = $21;$__r$0 = $__g;
   while(1) {
    $29 = ($__r$0>>>0)<($27>>>0);
    $30 = HEAP8[$__ig$0>>0]|0;
    $$not = ($30<<24>>24)<(1);
    $31 = ($30<<24>>24)==(127);
    $or$cond2 = $$not | $31;
    if (!($29)) {
     break;
    }
    if (!($or$cond2)) {
     $32 = $30 << 24 >> 24;
     $33 = HEAP32[$__r$0>>2]|0;
     $34 = ($32|0)==($33|0);
     if (!($34)) {
      label = 10;
      break;
     }
    }
    $35 = $__ig$0;
    $36 = (($28) - ($35))|0;
    $37 = ($36|0)>(1);
    $38 = ((($__ig$0)) + 1|0);
    $$__ig$0 = $37 ? $38 : $__ig$0;
    $39 = ((($__r$0)) + 4|0);
    $__ig$0 = $$__ig$0;$__r$0 = $39;
   }
   if ((label|0) == 10) {
    HEAP32[$__err>>2] = 4;
    break;
   }
   if (!($or$cond2)) {
    $40 = $30 << 24 >> 24;
    $41 = HEAP32[$27>>2]|0;
    $42 = (($41) + -1)|0;
    $43 = ($42>>>0)<($40>>>0);
    if (!($43)) {
     HEAP32[$__err>>2] = 4;
    }
   }
  }
 } while(0);
 return;
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRx($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE15__do_get_signedIxEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE15__do_get_signedIxEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i = 0, $$0$i$i = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i4 = 0, $$pre$i2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__g = 0, $__g_end = 0;
 var $__grouping = 0, $__i$0$i$i = 0, $__thousands_sep = 0, $exitcond$i$i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $phitmp$i = 0, $phitmp$i19 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 198|0;
 $__thousands_sep = sp + 196|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $0 = ((($__iob)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 74;
 switch ($2|0) {
 case 64:  {
  $$0$i = 8;
  break;
 }
 case 8:  {
  $$0$i = 16;
  break;
 }
 case 0:  {
  $$0$i = 0;
  break;
 }
 default: {
  $$0$i = 10;
 }
 }
 __ZNSt3__19__num_getIcE17__stage2_int_prepERNS_8ios_baseEPcRc($__grouping,$__iob,$__atoms,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $3 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$3>>2] = 0;
  $4 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $4;
 }
 $5 = HEAP8[$__buf>>0]|0;
 $6 = $5 & 1;
 $7 = ($6<<24>>24)==(0);
 if ($7) {
  $10 = 10;
 } else {
  $8 = HEAP32[$__buf>>2]|0;
  $9 = $8 & -2;
  $phitmp$i = (($9) + -1)|0;
  $10 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($10|0),0);
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 L13: do {
  if ($12) {
   label = 32;
  } else {
   $13 = HEAP8[$__buf>>0]|0;
   $14 = $13 & 1;
   $15 = ($14<<24>>24)==(0);
   $16 = ((($__buf)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($__buf)) + 1|0);
   $19 = $15 ? $18 : $17;
   HEAP32[$__a_end>>2] = $19;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   $20 = ((($__buf)) + 4|0);
   $21 = HEAP8[$__thousands_sep>>0]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i4 = $$pre;$__a$0 = $19;
   L15: while(1) {
    $22 = ($$pre$i$i$i4|0)==(0|0);
    if ($22) {
     $$pre$i$i$i = 0;
    } else {
     $23 = ((($$pre$i$i$i4)) + 12|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ((($$pre$i$i$i4)) + 16|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = ($24|0)==($26|0);
     if ($27) {
      $28 = HEAP32[$$pre$i$i$i4>>2]|0;
      $29 = ((($28)) + 36|0);
      $30 = HEAP32[$29>>2]|0;
      __THREW__ = 0;
      $31 = (invoke_ii($30|0,($$pre$i$i$i4|0))|0);
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if ($33) {
       label = 31;
       break;
      }
      $34 = ($31|0)==(-1);
      if ($34) {
       HEAP32[$__b>>2] = 0;
       $$pre$i$i$i = 0;
      } else {
       $$pre$i$i$i = $$pre$i$i$i4;
      }
     } else {
      $$pre$i$i$i = $$pre$i$i$i4;
     }
    }
    $35 = ($$pre$i$i$i|0)==(0|0);
    $36 = HEAP32[$__e>>2]|0;
    $37 = ($36|0)==(0|0);
    do {
     if ($37) {
      label = 23;
     } else {
      $38 = ((($36)) + 12|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ((($36)) + 16|0);
      $41 = HEAP32[$40>>2]|0;
      $42 = ($39|0)==($41|0);
      if (!($42)) {
       if ($35) {
        $163 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
      $43 = HEAP32[$36>>2]|0;
      $44 = ((($43)) + 36|0);
      $45 = HEAP32[$44>>2]|0;
      __THREW__ = 0;
      $46 = (invoke_ii($45|0,($36|0))|0);
      $47 = __THREW__; __THREW__ = 0;
      $48 = $47&1;
      if ($48) {
       label = 31;
       break L15;
      }
      $49 = ($46|0)==(-1);
      if ($49) {
       HEAP32[$__e>>2] = 0;
       label = 23;
       break;
      } else {
       if ($35) {
        $163 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
     }
    } while(0);
    if ((label|0) == 23) {
     label = 0;
     if ($35) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $163 = 0;
     }
    }
    $50 = HEAP32[$__a_end>>2]|0;
    $51 = HEAP8[$__buf>>0]|0;
    $52 = $51 & 1;
    $53 = ($52<<24>>24)==(0);
    $54 = HEAP32[$20>>2]|0;
    $55 = $51&255;
    $56 = $55 >>> 1;
    $57 = $53 ? $56 : $54;
    $58 = (($__a$0) + ($57)|0);
    $59 = ($50|0)==($58|0);
    if ($59) {
     $60 = $57 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($60|0),0);
     $61 = __THREW__; __THREW__ = 0;
     $62 = $61&1;
     if ($62) {
      label = 31;
      break;
     }
     $63 = HEAP8[$__buf>>0]|0;
     $64 = $63 & 1;
     $65 = ($64<<24>>24)==(0);
     if ($65) {
      $68 = 10;
     } else {
      $66 = HEAP32[$__buf>>2]|0;
      $67 = $66 & -2;
      $phitmp$i19 = (($67) + -1)|0;
      $68 = $phitmp$i19;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($68|0),0);
     $69 = __THREW__; __THREW__ = 0;
     $70 = $69&1;
     if ($70) {
      label = 31;
      break;
     }
     $71 = HEAP8[$__buf>>0]|0;
     $72 = $71 & 1;
     $73 = ($72<<24>>24)==(0);
     $74 = HEAP32[$16>>2]|0;
     $75 = $73 ? $18 : $74;
     $76 = (($75) + ($57)|0);
     HEAP32[$__a_end>>2] = $76;
     $__a$1 = $75;
    } else {
     $__a$1 = $__a$0;
    }
    $81 = ((($$pre$i$i$i)) + 12|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = ((($$pre$i$i$i)) + 16|0);
    $84 = HEAP32[$83>>2]|0;
    $85 = ($82|0)==($84|0);
    if ($85) {
     $86 = HEAP32[$$pre$i$i$i>>2]|0;
     $87 = ((($86)) + 36|0);
     $88 = HEAP32[$87>>2]|0;
     __THREW__ = 0;
     $89 = (invoke_ii($88|0,($$pre$i$i$i|0))|0);
     $90 = __THREW__; __THREW__ = 0;
     $91 = $90&1;
     if ($91) {
      label = 31;
      break;
     } else {
      $$0$i$i = $89;
     }
    } else {
     $92 = HEAP8[$82>>0]|0;
     $93 = $92&255;
     $$0$i$i = $93;
    }
    $94 = $$0$i$i&255;
    $95 = (__ZNSt3__19__num_getIcE17__stage2_int_loopEciPcRS2_RjcRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_S2_($94,$$0$i,$__a$1,$__a_end,$__dc,$21,$__grouping,$__g,$__g_end,$__atoms)|0);
    $96 = ($95|0)==(0);
    if (!($96)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $163;$__a$2 = $__a$1;
     break;
    }
    $97 = HEAP32[$81>>2]|0;
    $98 = HEAP32[$83>>2]|0;
    $99 = ($97|0)==($98|0);
    if ($99) {
     $100 = HEAP32[$$pre$i$i$i>>2]|0;
     $101 = ((($100)) + 40|0);
     $102 = HEAP32[$101>>2]|0;
     __THREW__ = 0;
     (invoke_ii($102|0,($$pre$i$i$i|0))|0);
     $103 = __THREW__; __THREW__ = 0;
     $104 = $103&1;
     if ($104) {
      label = 31;
      break;
     } else {
      $$pre$i$i$i4 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $105 = ((($97)) + 1|0);
     HEAP32[$81>>2] = $105;
     $$pre$i$i$i4 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 31) {
    $77 = ___cxa_find_matching_catch_2()|0;
    $78 = tempRet0;
    $lpad$phi$index = $77;$lpad$phi$index2 = $78;
    break;
   }
   $106 = HEAP8[$__grouping>>0]|0;
   $107 = $106 & 1;
   $108 = ($107<<24>>24)==(0);
   $109 = ((($__grouping)) + 4|0);
   $110 = HEAP32[$109>>2]|0;
   $111 = $106&255;
   $112 = $111 >>> 1;
   $113 = $108 ? $112 : $110;
   $114 = ($113|0)==(0);
   if (!($114)) {
    $115 = HEAP32[$__g_end>>2]|0;
    $116 = $115;
    $117 = $__g;
    $118 = (($116) - ($117))|0;
    $119 = ($118|0)<(160);
    if ($119) {
     $120 = HEAP32[$__dc>>2]|0;
     $121 = ((($115)) + 4|0);
     HEAP32[$__g_end>>2] = $121;
     HEAP32[$115>>2] = $120;
    }
   }
   $122 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $123 = (invoke_iiiii(4866,($__a$2|0),($122|0),($__err|0),($$0$i|0))|0);
   $124 = tempRet0;
   $125 = __THREW__; __THREW__ = 0;
   $126 = $125&1;
   if ($126) {
    label = 32;
   } else {
    $127 = $__v;
    $128 = $127;
    HEAP32[$128>>2] = $123;
    $129 = (($127) + 4)|0;
    $130 = $129;
    HEAP32[$130>>2] = $124;
    $131 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$131,$__err);
    $132 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($132) {
     $146 = 0;
    } else {
     $133 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $134 = HEAP32[$133>>2]|0;
     $135 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $136 = HEAP32[$135>>2]|0;
     $137 = ($134|0)==($136|0);
     if ($137) {
      $138 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $139 = ((($138)) + 36|0);
      $140 = HEAP32[$139>>2]|0;
      __THREW__ = 0;
      $141 = (invoke_ii($140|0,($$pre$i$i$i$lcssa|0))|0);
      $142 = __THREW__; __THREW__ = 0;
      $143 = $142&1;
      if ($143) {
       label = 32;
       break;
      }
      $144 = ($141|0)==(-1);
      if ($144) {
       HEAP32[$__b>>2] = 0;
       $146 = 0;
      } else {
       $146 = $$pre$i$i$i$lcssa;
      }
     } else {
      $146 = $$pre$i$i$i$lcssa;
     }
    }
    $145 = ($146|0)==(0|0);
    $147 = ($$pre$i2$i$i|0)==(0|0);
    do {
     if ($147) {
      label = 56;
     } else {
      $148 = ((($$pre$i2$i$i)) + 12|0);
      $149 = HEAP32[$148>>2]|0;
      $150 = ((($$pre$i2$i$i)) + 16|0);
      $151 = HEAP32[$150>>2]|0;
      $152 = ($149|0)==($151|0);
      if ($152) {
       $153 = HEAP32[$$pre$i2$i$i>>2]|0;
       $154 = ((($153)) + 36|0);
       $155 = HEAP32[$154>>2]|0;
       __THREW__ = 0;
       $156 = (invoke_ii($155|0,($$pre$i2$i$i|0))|0);
       $157 = __THREW__; __THREW__ = 0;
       $158 = $157&1;
       if ($158) {
        label = 32;
        break L13;
       }
       $159 = ($156|0)==(-1);
       if ($159) {
        HEAP32[$__e>>2] = 0;
        label = 56;
        break;
       }
      }
      if (!($145)) {
       label = 57;
      }
     }
    } while(0);
    if ((label|0) == 56) {
     if ($145) {
      label = 57;
     }
    }
    if ((label|0) == 57) {
     $160 = HEAP32[$__err>>2]|0;
     $161 = $160 | 2;
     HEAP32[$__err>>2] = $161;
    }
    $162 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($162|0);
   }
  }
 } while(0);
 if ((label|0) == 32) {
  $79 = ___cxa_find_matching_catch_2()|0;
  $80 = tempRet0;
  $lpad$phi$index = $79;$lpad$phi$index2 = $80;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__125__num_get_signed_integralIxEET_PKcS3_Rji($__a,$__a_end,$__err,$__base) {
 $__a = $__a|0;
 $__a_end = $__a_end|0;
 $__err = $__err|0;
 $__base = $__base|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $__p2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__p2 = sp;
 $0 = ($__a|0)==($__a_end|0);
 if ($0) {
  HEAP32[$__err>>2] = 4;
  $18 = 0;$19 = 0;
 } else {
  $1 = (___errno_location()|0);
  $2 = HEAP32[$1>>2]|0;
  HEAP32[$1>>2] = 0;
  $3 = (__ZNSt3__16__clocEv()|0);
  $4 = (_strtoll_l($__a,$__p2,$__base,$3)|0);
  $5 = tempRet0;
  $6 = HEAP32[$1>>2]|0;
  $7 = ($6|0)==(0);
  if ($7) {
   HEAP32[$1>>2] = $2;
  }
  $8 = HEAP32[$__p2>>2]|0;
  $9 = ($8|0)==($__a_end|0);
  if ($9) {
   $10 = ($6|0)==(34);
   if ($10) {
    HEAP32[$__err>>2] = 4;
    $11 = ($5|0)>(0);
    $12 = ($4>>>0)>(0);
    $13 = ($5|0)==(0);
    $14 = $13 & $12;
    $15 = $11 | $14;
    $16 = $15 ? -1 : 0;
    $17 = $15 ? 2147483647 : -2147483648;
    $20 = $16;$21 = $17;
   } else {
    $20 = $4;$21 = $5;
   }
  } else {
   HEAP32[$__err>>2] = 4;
   $20 = 0;$21 = 0;
  }
  $18 = $21;$19 = $20;
 }
 tempRet0 = ($18);
 STACKTOP = sp;return ($19|0);
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRt($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__do_get_unsignedItEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__do_get_unsignedItEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i = 0, $$0$i$i = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i4 = 0, $$pre$i2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__g = 0, $__g_end = 0, $__grouping = 0, $__i$0$i$i = 0, $__thousands_sep = 0, $exitcond$i$i = 0, $lpad$phi$index = 0;
 var $lpad$phi$index2 = 0, $phitmp$i = 0, $phitmp$i19 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 198|0;
 $__thousands_sep = sp + 196|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $0 = ((($__iob)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 74;
 switch ($2|0) {
 case 64:  {
  $$0$i = 8;
  break;
 }
 case 8:  {
  $$0$i = 16;
  break;
 }
 case 0:  {
  $$0$i = 0;
  break;
 }
 default: {
  $$0$i = 10;
 }
 }
 __ZNSt3__19__num_getIcE17__stage2_int_prepERNS_8ios_baseEPcRc($__grouping,$__iob,$__atoms,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $3 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$3>>2] = 0;
  $4 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $4;
 }
 $5 = HEAP8[$__buf>>0]|0;
 $6 = $5 & 1;
 $7 = ($6<<24>>24)==(0);
 if ($7) {
  $10 = 10;
 } else {
  $8 = HEAP32[$__buf>>2]|0;
  $9 = $8 & -2;
  $phitmp$i = (($9) + -1)|0;
  $10 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($10|0),0);
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 L13: do {
  if ($12) {
   label = 32;
  } else {
   $13 = HEAP8[$__buf>>0]|0;
   $14 = $13 & 1;
   $15 = ($14<<24>>24)==(0);
   $16 = ((($__buf)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($__buf)) + 1|0);
   $19 = $15 ? $18 : $17;
   HEAP32[$__a_end>>2] = $19;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   $20 = ((($__buf)) + 4|0);
   $21 = HEAP8[$__thousands_sep>>0]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i4 = $$pre;$__a$0 = $19;
   L15: while(1) {
    $22 = ($$pre$i$i$i4|0)==(0|0);
    if ($22) {
     $$pre$i$i$i = 0;
    } else {
     $23 = ((($$pre$i$i$i4)) + 12|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ((($$pre$i$i$i4)) + 16|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = ($24|0)==($26|0);
     if ($27) {
      $28 = HEAP32[$$pre$i$i$i4>>2]|0;
      $29 = ((($28)) + 36|0);
      $30 = HEAP32[$29>>2]|0;
      __THREW__ = 0;
      $31 = (invoke_ii($30|0,($$pre$i$i$i4|0))|0);
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if ($33) {
       label = 31;
       break;
      }
      $34 = ($31|0)==(-1);
      if ($34) {
       HEAP32[$__b>>2] = 0;
       $$pre$i$i$i = 0;
      } else {
       $$pre$i$i$i = $$pre$i$i$i4;
      }
     } else {
      $$pre$i$i$i = $$pre$i$i$i4;
     }
    }
    $35 = ($$pre$i$i$i|0)==(0|0);
    $36 = HEAP32[$__e>>2]|0;
    $37 = ($36|0)==(0|0);
    do {
     if ($37) {
      label = 23;
     } else {
      $38 = ((($36)) + 12|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ((($36)) + 16|0);
      $41 = HEAP32[$40>>2]|0;
      $42 = ($39|0)==($41|0);
      if (!($42)) {
       if ($35) {
        $158 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
      $43 = HEAP32[$36>>2]|0;
      $44 = ((($43)) + 36|0);
      $45 = HEAP32[$44>>2]|0;
      __THREW__ = 0;
      $46 = (invoke_ii($45|0,($36|0))|0);
      $47 = __THREW__; __THREW__ = 0;
      $48 = $47&1;
      if ($48) {
       label = 31;
       break L15;
      }
      $49 = ($46|0)==(-1);
      if ($49) {
       HEAP32[$__e>>2] = 0;
       label = 23;
       break;
      } else {
       if ($35) {
        $158 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
     }
    } while(0);
    if ((label|0) == 23) {
     label = 0;
     if ($35) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $158 = 0;
     }
    }
    $50 = HEAP32[$__a_end>>2]|0;
    $51 = HEAP8[$__buf>>0]|0;
    $52 = $51 & 1;
    $53 = ($52<<24>>24)==(0);
    $54 = HEAP32[$20>>2]|0;
    $55 = $51&255;
    $56 = $55 >>> 1;
    $57 = $53 ? $56 : $54;
    $58 = (($__a$0) + ($57)|0);
    $59 = ($50|0)==($58|0);
    if ($59) {
     $60 = $57 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($60|0),0);
     $61 = __THREW__; __THREW__ = 0;
     $62 = $61&1;
     if ($62) {
      label = 31;
      break;
     }
     $63 = HEAP8[$__buf>>0]|0;
     $64 = $63 & 1;
     $65 = ($64<<24>>24)==(0);
     if ($65) {
      $68 = 10;
     } else {
      $66 = HEAP32[$__buf>>2]|0;
      $67 = $66 & -2;
      $phitmp$i19 = (($67) + -1)|0;
      $68 = $phitmp$i19;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($68|0),0);
     $69 = __THREW__; __THREW__ = 0;
     $70 = $69&1;
     if ($70) {
      label = 31;
      break;
     }
     $71 = HEAP8[$__buf>>0]|0;
     $72 = $71 & 1;
     $73 = ($72<<24>>24)==(0);
     $74 = HEAP32[$16>>2]|0;
     $75 = $73 ? $18 : $74;
     $76 = (($75) + ($57)|0);
     HEAP32[$__a_end>>2] = $76;
     $__a$1 = $75;
    } else {
     $__a$1 = $__a$0;
    }
    $81 = ((($$pre$i$i$i)) + 12|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = ((($$pre$i$i$i)) + 16|0);
    $84 = HEAP32[$83>>2]|0;
    $85 = ($82|0)==($84|0);
    if ($85) {
     $86 = HEAP32[$$pre$i$i$i>>2]|0;
     $87 = ((($86)) + 36|0);
     $88 = HEAP32[$87>>2]|0;
     __THREW__ = 0;
     $89 = (invoke_ii($88|0,($$pre$i$i$i|0))|0);
     $90 = __THREW__; __THREW__ = 0;
     $91 = $90&1;
     if ($91) {
      label = 31;
      break;
     } else {
      $$0$i$i = $89;
     }
    } else {
     $92 = HEAP8[$82>>0]|0;
     $93 = $92&255;
     $$0$i$i = $93;
    }
    $94 = $$0$i$i&255;
    $95 = (__ZNSt3__19__num_getIcE17__stage2_int_loopEciPcRS2_RjcRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_S2_($94,$$0$i,$__a$1,$__a_end,$__dc,$21,$__grouping,$__g,$__g_end,$__atoms)|0);
    $96 = ($95|0)==(0);
    if (!($96)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $158;$__a$2 = $__a$1;
     break;
    }
    $97 = HEAP32[$81>>2]|0;
    $98 = HEAP32[$83>>2]|0;
    $99 = ($97|0)==($98|0);
    if ($99) {
     $100 = HEAP32[$$pre$i$i$i>>2]|0;
     $101 = ((($100)) + 40|0);
     $102 = HEAP32[$101>>2]|0;
     __THREW__ = 0;
     (invoke_ii($102|0,($$pre$i$i$i|0))|0);
     $103 = __THREW__; __THREW__ = 0;
     $104 = $103&1;
     if ($104) {
      label = 31;
      break;
     } else {
      $$pre$i$i$i4 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $105 = ((($97)) + 1|0);
     HEAP32[$81>>2] = $105;
     $$pre$i$i$i4 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 31) {
    $77 = ___cxa_find_matching_catch_2()|0;
    $78 = tempRet0;
    $lpad$phi$index = $77;$lpad$phi$index2 = $78;
    break;
   }
   $106 = HEAP8[$__grouping>>0]|0;
   $107 = $106 & 1;
   $108 = ($107<<24>>24)==(0);
   $109 = ((($__grouping)) + 4|0);
   $110 = HEAP32[$109>>2]|0;
   $111 = $106&255;
   $112 = $111 >>> 1;
   $113 = $108 ? $112 : $110;
   $114 = ($113|0)==(0);
   if (!($114)) {
    $115 = HEAP32[$__g_end>>2]|0;
    $116 = $115;
    $117 = $__g;
    $118 = (($116) - ($117))|0;
    $119 = ($118|0)<(160);
    if ($119) {
     $120 = HEAP32[$__dc>>2]|0;
     $121 = ((($115)) + 4|0);
     HEAP32[$__g_end>>2] = $121;
     HEAP32[$115>>2] = $120;
    }
   }
   $122 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $123 = (invoke_iiiii(4867,($__a$2|0),($122|0),($__err|0),($$0$i|0))|0);
   $124 = __THREW__; __THREW__ = 0;
   $125 = $124&1;
   if ($125) {
    label = 32;
   } else {
    HEAP16[$__v>>1] = $123;
    $126 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$126,$__err);
    $127 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($127) {
     $141 = 0;
    } else {
     $128 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $129 = HEAP32[$128>>2]|0;
     $130 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $131 = HEAP32[$130>>2]|0;
     $132 = ($129|0)==($131|0);
     if ($132) {
      $133 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $134 = ((($133)) + 36|0);
      $135 = HEAP32[$134>>2]|0;
      __THREW__ = 0;
      $136 = (invoke_ii($135|0,($$pre$i$i$i$lcssa|0))|0);
      $137 = __THREW__; __THREW__ = 0;
      $138 = $137&1;
      if ($138) {
       label = 32;
       break;
      }
      $139 = ($136|0)==(-1);
      if ($139) {
       HEAP32[$__b>>2] = 0;
       $141 = 0;
      } else {
       $141 = $$pre$i$i$i$lcssa;
      }
     } else {
      $141 = $$pre$i$i$i$lcssa;
     }
    }
    $140 = ($141|0)==(0|0);
    $142 = ($$pre$i2$i$i|0)==(0|0);
    do {
     if ($142) {
      label = 56;
     } else {
      $143 = ((($$pre$i2$i$i)) + 12|0);
      $144 = HEAP32[$143>>2]|0;
      $145 = ((($$pre$i2$i$i)) + 16|0);
      $146 = HEAP32[$145>>2]|0;
      $147 = ($144|0)==($146|0);
      if ($147) {
       $148 = HEAP32[$$pre$i2$i$i>>2]|0;
       $149 = ((($148)) + 36|0);
       $150 = HEAP32[$149>>2]|0;
       __THREW__ = 0;
       $151 = (invoke_ii($150|0,($$pre$i2$i$i|0))|0);
       $152 = __THREW__; __THREW__ = 0;
       $153 = $152&1;
       if ($153) {
        label = 32;
        break L13;
       }
       $154 = ($151|0)==(-1);
       if ($154) {
        HEAP32[$__e>>2] = 0;
        label = 56;
        break;
       }
      }
      if (!($140)) {
       label = 57;
      }
     }
    } while(0);
    if ((label|0) == 56) {
     if ($140) {
      label = 57;
     }
    }
    if ((label|0) == 57) {
     $155 = HEAP32[$__err>>2]|0;
     $156 = $155 | 2;
     HEAP32[$__err>>2] = $156;
    }
    $157 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($157|0);
   }
  }
 } while(0);
 if ((label|0) == 32) {
  $79 = ___cxa_find_matching_catch_2()|0;
  $80 = tempRet0;
  $lpad$phi$index = $79;$lpad$phi$index2 = $80;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__127__num_get_unsigned_integralItEET_PKcS3_Rji($__a,$__a_end,$__err,$__base) {
 $__a = $__a|0;
 $__a_end = $__a_end|0;
 $__err = $__err|0;
 $__base = $__base|0;
 var $$0 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $__p2 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__p2 = sp;
 $0 = ($__a|0)==($__a_end|0);
 do {
  if ($0) {
   HEAP32[$__err>>2] = 4;
   $$1 = 0;
  } else {
   $1 = HEAP8[$__a>>0]|0;
   $2 = ($1<<24>>24)==(45);
   if ($2) {
    HEAP32[$__err>>2] = 4;
    $$1 = 0;
    break;
   }
   $3 = (___errno_location()|0);
   $4 = HEAP32[$3>>2]|0;
   HEAP32[$3>>2] = 0;
   $5 = (__ZNSt3__16__clocEv()|0);
   $6 = (_strtoull_l($__a,$__p2,$__base,$5)|0);
   $7 = tempRet0;
   $8 = HEAP32[$3>>2]|0;
   $9 = ($8|0)==(0);
   if ($9) {
    HEAP32[$3>>2] = $4;
   }
   $10 = HEAP32[$__p2>>2]|0;
   $11 = ($10|0)==($__a_end|0);
   do {
    if ($11) {
     $12 = ($8|0)==(34);
     $13 = ($7>>>0)>(0);
     $14 = ($6>>>0)>(65535);
     $15 = ($7|0)==(0);
     $16 = $15 & $14;
     $17 = $13 | $16;
     $or$cond = $17 | $12;
     if ($or$cond) {
      HEAP32[$__err>>2] = 4;
      $$0 = -1;
      break;
     } else {
      $18 = $6&65535;
      $$0 = $18;
      break;
     }
    } else {
     HEAP32[$__err>>2] = 4;
     $$0 = 0;
    }
   } while(0);
   $$1 = $$0;
  }
 } while(0);
 STACKTOP = sp;return ($$1|0);
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjS8_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__do_get_unsignedIjEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__do_get_unsignedIjEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i = 0, $$0$i$i = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i4 = 0, $$pre$i2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__g = 0, $__g_end = 0, $__grouping = 0, $__i$0$i$i = 0, $__thousands_sep = 0, $exitcond$i$i = 0, $lpad$phi$index = 0;
 var $lpad$phi$index2 = 0, $phitmp$i = 0, $phitmp$i19 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 198|0;
 $__thousands_sep = sp + 196|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $0 = ((($__iob)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 74;
 switch ($2|0) {
 case 64:  {
  $$0$i = 8;
  break;
 }
 case 8:  {
  $$0$i = 16;
  break;
 }
 case 0:  {
  $$0$i = 0;
  break;
 }
 default: {
  $$0$i = 10;
 }
 }
 __ZNSt3__19__num_getIcE17__stage2_int_prepERNS_8ios_baseEPcRc($__grouping,$__iob,$__atoms,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $3 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$3>>2] = 0;
  $4 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $4;
 }
 $5 = HEAP8[$__buf>>0]|0;
 $6 = $5 & 1;
 $7 = ($6<<24>>24)==(0);
 if ($7) {
  $10 = 10;
 } else {
  $8 = HEAP32[$__buf>>2]|0;
  $9 = $8 & -2;
  $phitmp$i = (($9) + -1)|0;
  $10 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($10|0),0);
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 L13: do {
  if ($12) {
   label = 32;
  } else {
   $13 = HEAP8[$__buf>>0]|0;
   $14 = $13 & 1;
   $15 = ($14<<24>>24)==(0);
   $16 = ((($__buf)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($__buf)) + 1|0);
   $19 = $15 ? $18 : $17;
   HEAP32[$__a_end>>2] = $19;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   $20 = ((($__buf)) + 4|0);
   $21 = HEAP8[$__thousands_sep>>0]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i4 = $$pre;$__a$0 = $19;
   L15: while(1) {
    $22 = ($$pre$i$i$i4|0)==(0|0);
    if ($22) {
     $$pre$i$i$i = 0;
    } else {
     $23 = ((($$pre$i$i$i4)) + 12|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ((($$pre$i$i$i4)) + 16|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = ($24|0)==($26|0);
     if ($27) {
      $28 = HEAP32[$$pre$i$i$i4>>2]|0;
      $29 = ((($28)) + 36|0);
      $30 = HEAP32[$29>>2]|0;
      __THREW__ = 0;
      $31 = (invoke_ii($30|0,($$pre$i$i$i4|0))|0);
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if ($33) {
       label = 31;
       break;
      }
      $34 = ($31|0)==(-1);
      if ($34) {
       HEAP32[$__b>>2] = 0;
       $$pre$i$i$i = 0;
      } else {
       $$pre$i$i$i = $$pre$i$i$i4;
      }
     } else {
      $$pre$i$i$i = $$pre$i$i$i4;
     }
    }
    $35 = ($$pre$i$i$i|0)==(0|0);
    $36 = HEAP32[$__e>>2]|0;
    $37 = ($36|0)==(0|0);
    do {
     if ($37) {
      label = 23;
     } else {
      $38 = ((($36)) + 12|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ((($36)) + 16|0);
      $41 = HEAP32[$40>>2]|0;
      $42 = ($39|0)==($41|0);
      if (!($42)) {
       if ($35) {
        $158 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
      $43 = HEAP32[$36>>2]|0;
      $44 = ((($43)) + 36|0);
      $45 = HEAP32[$44>>2]|0;
      __THREW__ = 0;
      $46 = (invoke_ii($45|0,($36|0))|0);
      $47 = __THREW__; __THREW__ = 0;
      $48 = $47&1;
      if ($48) {
       label = 31;
       break L15;
      }
      $49 = ($46|0)==(-1);
      if ($49) {
       HEAP32[$__e>>2] = 0;
       label = 23;
       break;
      } else {
       if ($35) {
        $158 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
     }
    } while(0);
    if ((label|0) == 23) {
     label = 0;
     if ($35) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $158 = 0;
     }
    }
    $50 = HEAP32[$__a_end>>2]|0;
    $51 = HEAP8[$__buf>>0]|0;
    $52 = $51 & 1;
    $53 = ($52<<24>>24)==(0);
    $54 = HEAP32[$20>>2]|0;
    $55 = $51&255;
    $56 = $55 >>> 1;
    $57 = $53 ? $56 : $54;
    $58 = (($__a$0) + ($57)|0);
    $59 = ($50|0)==($58|0);
    if ($59) {
     $60 = $57 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($60|0),0);
     $61 = __THREW__; __THREW__ = 0;
     $62 = $61&1;
     if ($62) {
      label = 31;
      break;
     }
     $63 = HEAP8[$__buf>>0]|0;
     $64 = $63 & 1;
     $65 = ($64<<24>>24)==(0);
     if ($65) {
      $68 = 10;
     } else {
      $66 = HEAP32[$__buf>>2]|0;
      $67 = $66 & -2;
      $phitmp$i19 = (($67) + -1)|0;
      $68 = $phitmp$i19;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($68|0),0);
     $69 = __THREW__; __THREW__ = 0;
     $70 = $69&1;
     if ($70) {
      label = 31;
      break;
     }
     $71 = HEAP8[$__buf>>0]|0;
     $72 = $71 & 1;
     $73 = ($72<<24>>24)==(0);
     $74 = HEAP32[$16>>2]|0;
     $75 = $73 ? $18 : $74;
     $76 = (($75) + ($57)|0);
     HEAP32[$__a_end>>2] = $76;
     $__a$1 = $75;
    } else {
     $__a$1 = $__a$0;
    }
    $81 = ((($$pre$i$i$i)) + 12|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = ((($$pre$i$i$i)) + 16|0);
    $84 = HEAP32[$83>>2]|0;
    $85 = ($82|0)==($84|0);
    if ($85) {
     $86 = HEAP32[$$pre$i$i$i>>2]|0;
     $87 = ((($86)) + 36|0);
     $88 = HEAP32[$87>>2]|0;
     __THREW__ = 0;
     $89 = (invoke_ii($88|0,($$pre$i$i$i|0))|0);
     $90 = __THREW__; __THREW__ = 0;
     $91 = $90&1;
     if ($91) {
      label = 31;
      break;
     } else {
      $$0$i$i = $89;
     }
    } else {
     $92 = HEAP8[$82>>0]|0;
     $93 = $92&255;
     $$0$i$i = $93;
    }
    $94 = $$0$i$i&255;
    $95 = (__ZNSt3__19__num_getIcE17__stage2_int_loopEciPcRS2_RjcRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_S2_($94,$$0$i,$__a$1,$__a_end,$__dc,$21,$__grouping,$__g,$__g_end,$__atoms)|0);
    $96 = ($95|0)==(0);
    if (!($96)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $158;$__a$2 = $__a$1;
     break;
    }
    $97 = HEAP32[$81>>2]|0;
    $98 = HEAP32[$83>>2]|0;
    $99 = ($97|0)==($98|0);
    if ($99) {
     $100 = HEAP32[$$pre$i$i$i>>2]|0;
     $101 = ((($100)) + 40|0);
     $102 = HEAP32[$101>>2]|0;
     __THREW__ = 0;
     (invoke_ii($102|0,($$pre$i$i$i|0))|0);
     $103 = __THREW__; __THREW__ = 0;
     $104 = $103&1;
     if ($104) {
      label = 31;
      break;
     } else {
      $$pre$i$i$i4 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $105 = ((($97)) + 1|0);
     HEAP32[$81>>2] = $105;
     $$pre$i$i$i4 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 31) {
    $77 = ___cxa_find_matching_catch_2()|0;
    $78 = tempRet0;
    $lpad$phi$index = $77;$lpad$phi$index2 = $78;
    break;
   }
   $106 = HEAP8[$__grouping>>0]|0;
   $107 = $106 & 1;
   $108 = ($107<<24>>24)==(0);
   $109 = ((($__grouping)) + 4|0);
   $110 = HEAP32[$109>>2]|0;
   $111 = $106&255;
   $112 = $111 >>> 1;
   $113 = $108 ? $112 : $110;
   $114 = ($113|0)==(0);
   if (!($114)) {
    $115 = HEAP32[$__g_end>>2]|0;
    $116 = $115;
    $117 = $__g;
    $118 = (($116) - ($117))|0;
    $119 = ($118|0)<(160);
    if ($119) {
     $120 = HEAP32[$__dc>>2]|0;
     $121 = ((($115)) + 4|0);
     HEAP32[$__g_end>>2] = $121;
     HEAP32[$115>>2] = $120;
    }
   }
   $122 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $123 = (invoke_iiiii(4868,($__a$2|0),($122|0),($__err|0),($$0$i|0))|0);
   $124 = __THREW__; __THREW__ = 0;
   $125 = $124&1;
   if ($125) {
    label = 32;
   } else {
    HEAP32[$__v>>2] = $123;
    $126 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$126,$__err);
    $127 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($127) {
     $141 = 0;
    } else {
     $128 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $129 = HEAP32[$128>>2]|0;
     $130 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $131 = HEAP32[$130>>2]|0;
     $132 = ($129|0)==($131|0);
     if ($132) {
      $133 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $134 = ((($133)) + 36|0);
      $135 = HEAP32[$134>>2]|0;
      __THREW__ = 0;
      $136 = (invoke_ii($135|0,($$pre$i$i$i$lcssa|0))|0);
      $137 = __THREW__; __THREW__ = 0;
      $138 = $137&1;
      if ($138) {
       label = 32;
       break;
      }
      $139 = ($136|0)==(-1);
      if ($139) {
       HEAP32[$__b>>2] = 0;
       $141 = 0;
      } else {
       $141 = $$pre$i$i$i$lcssa;
      }
     } else {
      $141 = $$pre$i$i$i$lcssa;
     }
    }
    $140 = ($141|0)==(0|0);
    $142 = ($$pre$i2$i$i|0)==(0|0);
    do {
     if ($142) {
      label = 56;
     } else {
      $143 = ((($$pre$i2$i$i)) + 12|0);
      $144 = HEAP32[$143>>2]|0;
      $145 = ((($$pre$i2$i$i)) + 16|0);
      $146 = HEAP32[$145>>2]|0;
      $147 = ($144|0)==($146|0);
      if ($147) {
       $148 = HEAP32[$$pre$i2$i$i>>2]|0;
       $149 = ((($148)) + 36|0);
       $150 = HEAP32[$149>>2]|0;
       __THREW__ = 0;
       $151 = (invoke_ii($150|0,($$pre$i2$i$i|0))|0);
       $152 = __THREW__; __THREW__ = 0;
       $153 = $152&1;
       if ($153) {
        label = 32;
        break L13;
       }
       $154 = ($151|0)==(-1);
       if ($154) {
        HEAP32[$__e>>2] = 0;
        label = 56;
        break;
       }
      }
      if (!($140)) {
       label = 57;
      }
     }
    } while(0);
    if ((label|0) == 56) {
     if ($140) {
      label = 57;
     }
    }
    if ((label|0) == 57) {
     $155 = HEAP32[$__err>>2]|0;
     $156 = $155 | 2;
     HEAP32[$__err>>2] = $156;
    }
    $157 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($157|0);
   }
  }
 } while(0);
 if ((label|0) == 32) {
  $79 = ___cxa_find_matching_catch_2()|0;
  $80 = tempRet0;
  $lpad$phi$index = $79;$lpad$phi$index2 = $80;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__127__num_get_unsigned_integralIjEET_PKcS3_Rji($__a,$__a_end,$__err,$__base) {
 $__a = $__a|0;
 $__a_end = $__a_end|0;
 $__err = $__err|0;
 $__base = $__base|0;
 var $$0 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $__p2 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__p2 = sp;
 $0 = ($__a|0)==($__a_end|0);
 do {
  if ($0) {
   HEAP32[$__err>>2] = 4;
   $$1 = 0;
  } else {
   $1 = HEAP8[$__a>>0]|0;
   $2 = ($1<<24>>24)==(45);
   if ($2) {
    HEAP32[$__err>>2] = 4;
    $$1 = 0;
    break;
   }
   $3 = (___errno_location()|0);
   $4 = HEAP32[$3>>2]|0;
   HEAP32[$3>>2] = 0;
   $5 = (__ZNSt3__16__clocEv()|0);
   $6 = (_strtoull_l($__a,$__p2,$__base,$5)|0);
   $7 = tempRet0;
   $8 = HEAP32[$3>>2]|0;
   $9 = ($8|0)==(0);
   if ($9) {
    HEAP32[$3>>2] = $4;
   }
   $10 = HEAP32[$__p2>>2]|0;
   $11 = ($10|0)==($__a_end|0);
   do {
    if ($11) {
     $12 = ($8|0)==(34);
     $13 = ($7>>>0)>(0);
     $14 = ($6>>>0)>(4294967295);
     $15 = ($7|0)==(0);
     $16 = $15 & $14;
     $17 = $13 | $16;
     $or$cond = $17 | $12;
     if ($or$cond) {
      HEAP32[$__err>>2] = 4;
      $$0 = -1;
      break;
     } else {
      $$0 = $6;
      break;
     }
    } else {
     HEAP32[$__err>>2] = 4;
     $$0 = 0;
    }
   } while(0);
   $$1 = $$0;
  }
 } while(0);
 STACKTOP = sp;return ($$1|0);
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRm($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__do_get_unsignedImEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__do_get_unsignedImEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i = 0, $$0$i$i = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i4 = 0, $$pre$i2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__g = 0, $__g_end = 0, $__grouping = 0, $__i$0$i$i = 0, $__thousands_sep = 0, $exitcond$i$i = 0, $lpad$phi$index = 0;
 var $lpad$phi$index2 = 0, $phitmp$i = 0, $phitmp$i19 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 198|0;
 $__thousands_sep = sp + 196|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $0 = ((($__iob)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 74;
 switch ($2|0) {
 case 64:  {
  $$0$i = 8;
  break;
 }
 case 8:  {
  $$0$i = 16;
  break;
 }
 case 0:  {
  $$0$i = 0;
  break;
 }
 default: {
  $$0$i = 10;
 }
 }
 __ZNSt3__19__num_getIcE17__stage2_int_prepERNS_8ios_baseEPcRc($__grouping,$__iob,$__atoms,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $3 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$3>>2] = 0;
  $4 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $4;
 }
 $5 = HEAP8[$__buf>>0]|0;
 $6 = $5 & 1;
 $7 = ($6<<24>>24)==(0);
 if ($7) {
  $10 = 10;
 } else {
  $8 = HEAP32[$__buf>>2]|0;
  $9 = $8 & -2;
  $phitmp$i = (($9) + -1)|0;
  $10 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($10|0),0);
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 L13: do {
  if ($12) {
   label = 32;
  } else {
   $13 = HEAP8[$__buf>>0]|0;
   $14 = $13 & 1;
   $15 = ($14<<24>>24)==(0);
   $16 = ((($__buf)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($__buf)) + 1|0);
   $19 = $15 ? $18 : $17;
   HEAP32[$__a_end>>2] = $19;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   $20 = ((($__buf)) + 4|0);
   $21 = HEAP8[$__thousands_sep>>0]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i4 = $$pre;$__a$0 = $19;
   L15: while(1) {
    $22 = ($$pre$i$i$i4|0)==(0|0);
    if ($22) {
     $$pre$i$i$i = 0;
    } else {
     $23 = ((($$pre$i$i$i4)) + 12|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ((($$pre$i$i$i4)) + 16|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = ($24|0)==($26|0);
     if ($27) {
      $28 = HEAP32[$$pre$i$i$i4>>2]|0;
      $29 = ((($28)) + 36|0);
      $30 = HEAP32[$29>>2]|0;
      __THREW__ = 0;
      $31 = (invoke_ii($30|0,($$pre$i$i$i4|0))|0);
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if ($33) {
       label = 31;
       break;
      }
      $34 = ($31|0)==(-1);
      if ($34) {
       HEAP32[$__b>>2] = 0;
       $$pre$i$i$i = 0;
      } else {
       $$pre$i$i$i = $$pre$i$i$i4;
      }
     } else {
      $$pre$i$i$i = $$pre$i$i$i4;
     }
    }
    $35 = ($$pre$i$i$i|0)==(0|0);
    $36 = HEAP32[$__e>>2]|0;
    $37 = ($36|0)==(0|0);
    do {
     if ($37) {
      label = 23;
     } else {
      $38 = ((($36)) + 12|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ((($36)) + 16|0);
      $41 = HEAP32[$40>>2]|0;
      $42 = ($39|0)==($41|0);
      if (!($42)) {
       if ($35) {
        $158 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
      $43 = HEAP32[$36>>2]|0;
      $44 = ((($43)) + 36|0);
      $45 = HEAP32[$44>>2]|0;
      __THREW__ = 0;
      $46 = (invoke_ii($45|0,($36|0))|0);
      $47 = __THREW__; __THREW__ = 0;
      $48 = $47&1;
      if ($48) {
       label = 31;
       break L15;
      }
      $49 = ($46|0)==(-1);
      if ($49) {
       HEAP32[$__e>>2] = 0;
       label = 23;
       break;
      } else {
       if ($35) {
        $158 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
     }
    } while(0);
    if ((label|0) == 23) {
     label = 0;
     if ($35) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $158 = 0;
     }
    }
    $50 = HEAP32[$__a_end>>2]|0;
    $51 = HEAP8[$__buf>>0]|0;
    $52 = $51 & 1;
    $53 = ($52<<24>>24)==(0);
    $54 = HEAP32[$20>>2]|0;
    $55 = $51&255;
    $56 = $55 >>> 1;
    $57 = $53 ? $56 : $54;
    $58 = (($__a$0) + ($57)|0);
    $59 = ($50|0)==($58|0);
    if ($59) {
     $60 = $57 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($60|0),0);
     $61 = __THREW__; __THREW__ = 0;
     $62 = $61&1;
     if ($62) {
      label = 31;
      break;
     }
     $63 = HEAP8[$__buf>>0]|0;
     $64 = $63 & 1;
     $65 = ($64<<24>>24)==(0);
     if ($65) {
      $68 = 10;
     } else {
      $66 = HEAP32[$__buf>>2]|0;
      $67 = $66 & -2;
      $phitmp$i19 = (($67) + -1)|0;
      $68 = $phitmp$i19;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($68|0),0);
     $69 = __THREW__; __THREW__ = 0;
     $70 = $69&1;
     if ($70) {
      label = 31;
      break;
     }
     $71 = HEAP8[$__buf>>0]|0;
     $72 = $71 & 1;
     $73 = ($72<<24>>24)==(0);
     $74 = HEAP32[$16>>2]|0;
     $75 = $73 ? $18 : $74;
     $76 = (($75) + ($57)|0);
     HEAP32[$__a_end>>2] = $76;
     $__a$1 = $75;
    } else {
     $__a$1 = $__a$0;
    }
    $81 = ((($$pre$i$i$i)) + 12|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = ((($$pre$i$i$i)) + 16|0);
    $84 = HEAP32[$83>>2]|0;
    $85 = ($82|0)==($84|0);
    if ($85) {
     $86 = HEAP32[$$pre$i$i$i>>2]|0;
     $87 = ((($86)) + 36|0);
     $88 = HEAP32[$87>>2]|0;
     __THREW__ = 0;
     $89 = (invoke_ii($88|0,($$pre$i$i$i|0))|0);
     $90 = __THREW__; __THREW__ = 0;
     $91 = $90&1;
     if ($91) {
      label = 31;
      break;
     } else {
      $$0$i$i = $89;
     }
    } else {
     $92 = HEAP8[$82>>0]|0;
     $93 = $92&255;
     $$0$i$i = $93;
    }
    $94 = $$0$i$i&255;
    $95 = (__ZNSt3__19__num_getIcE17__stage2_int_loopEciPcRS2_RjcRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_S2_($94,$$0$i,$__a$1,$__a_end,$__dc,$21,$__grouping,$__g,$__g_end,$__atoms)|0);
    $96 = ($95|0)==(0);
    if (!($96)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $158;$__a$2 = $__a$1;
     break;
    }
    $97 = HEAP32[$81>>2]|0;
    $98 = HEAP32[$83>>2]|0;
    $99 = ($97|0)==($98|0);
    if ($99) {
     $100 = HEAP32[$$pre$i$i$i>>2]|0;
     $101 = ((($100)) + 40|0);
     $102 = HEAP32[$101>>2]|0;
     __THREW__ = 0;
     (invoke_ii($102|0,($$pre$i$i$i|0))|0);
     $103 = __THREW__; __THREW__ = 0;
     $104 = $103&1;
     if ($104) {
      label = 31;
      break;
     } else {
      $$pre$i$i$i4 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $105 = ((($97)) + 1|0);
     HEAP32[$81>>2] = $105;
     $$pre$i$i$i4 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 31) {
    $77 = ___cxa_find_matching_catch_2()|0;
    $78 = tempRet0;
    $lpad$phi$index = $77;$lpad$phi$index2 = $78;
    break;
   }
   $106 = HEAP8[$__grouping>>0]|0;
   $107 = $106 & 1;
   $108 = ($107<<24>>24)==(0);
   $109 = ((($__grouping)) + 4|0);
   $110 = HEAP32[$109>>2]|0;
   $111 = $106&255;
   $112 = $111 >>> 1;
   $113 = $108 ? $112 : $110;
   $114 = ($113|0)==(0);
   if (!($114)) {
    $115 = HEAP32[$__g_end>>2]|0;
    $116 = $115;
    $117 = $__g;
    $118 = (($116) - ($117))|0;
    $119 = ($118|0)<(160);
    if ($119) {
     $120 = HEAP32[$__dc>>2]|0;
     $121 = ((($115)) + 4|0);
     HEAP32[$__g_end>>2] = $121;
     HEAP32[$115>>2] = $120;
    }
   }
   $122 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $123 = (invoke_iiiii(4869,($__a$2|0),($122|0),($__err|0),($$0$i|0))|0);
   $124 = __THREW__; __THREW__ = 0;
   $125 = $124&1;
   if ($125) {
    label = 32;
   } else {
    HEAP32[$__v>>2] = $123;
    $126 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$126,$__err);
    $127 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($127) {
     $141 = 0;
    } else {
     $128 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $129 = HEAP32[$128>>2]|0;
     $130 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $131 = HEAP32[$130>>2]|0;
     $132 = ($129|0)==($131|0);
     if ($132) {
      $133 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $134 = ((($133)) + 36|0);
      $135 = HEAP32[$134>>2]|0;
      __THREW__ = 0;
      $136 = (invoke_ii($135|0,($$pre$i$i$i$lcssa|0))|0);
      $137 = __THREW__; __THREW__ = 0;
      $138 = $137&1;
      if ($138) {
       label = 32;
       break;
      }
      $139 = ($136|0)==(-1);
      if ($139) {
       HEAP32[$__b>>2] = 0;
       $141 = 0;
      } else {
       $141 = $$pre$i$i$i$lcssa;
      }
     } else {
      $141 = $$pre$i$i$i$lcssa;
     }
    }
    $140 = ($141|0)==(0|0);
    $142 = ($$pre$i2$i$i|0)==(0|0);
    do {
     if ($142) {
      label = 56;
     } else {
      $143 = ((($$pre$i2$i$i)) + 12|0);
      $144 = HEAP32[$143>>2]|0;
      $145 = ((($$pre$i2$i$i)) + 16|0);
      $146 = HEAP32[$145>>2]|0;
      $147 = ($144|0)==($146|0);
      if ($147) {
       $148 = HEAP32[$$pre$i2$i$i>>2]|0;
       $149 = ((($148)) + 36|0);
       $150 = HEAP32[$149>>2]|0;
       __THREW__ = 0;
       $151 = (invoke_ii($150|0,($$pre$i2$i$i|0))|0);
       $152 = __THREW__; __THREW__ = 0;
       $153 = $152&1;
       if ($153) {
        label = 32;
        break L13;
       }
       $154 = ($151|0)==(-1);
       if ($154) {
        HEAP32[$__e>>2] = 0;
        label = 56;
        break;
       }
      }
      if (!($140)) {
       label = 57;
      }
     }
    } while(0);
    if ((label|0) == 56) {
     if ($140) {
      label = 57;
     }
    }
    if ((label|0) == 57) {
     $155 = HEAP32[$__err>>2]|0;
     $156 = $155 | 2;
     HEAP32[$__err>>2] = $156;
    }
    $157 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($157|0);
   }
  }
 } while(0);
 if ((label|0) == 32) {
  $79 = ___cxa_find_matching_catch_2()|0;
  $80 = tempRet0;
  $lpad$phi$index = $79;$lpad$phi$index2 = $80;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__127__num_get_unsigned_integralImEET_PKcS3_Rji($__a,$__a_end,$__err,$__base) {
 $__a = $__a|0;
 $__a_end = $__a_end|0;
 $__err = $__err|0;
 $__base = $__base|0;
 var $$0 = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $__p2 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__p2 = sp;
 $0 = ($__a|0)==($__a_end|0);
 do {
  if ($0) {
   HEAP32[$__err>>2] = 4;
   $$1 = 0;
  } else {
   $1 = HEAP8[$__a>>0]|0;
   $2 = ($1<<24>>24)==(45);
   if ($2) {
    HEAP32[$__err>>2] = 4;
    $$1 = 0;
    break;
   }
   $3 = (___errno_location()|0);
   $4 = HEAP32[$3>>2]|0;
   HEAP32[$3>>2] = 0;
   $5 = (__ZNSt3__16__clocEv()|0);
   $6 = (_strtoull_l($__a,$__p2,$__base,$5)|0);
   $7 = tempRet0;
   $8 = HEAP32[$3>>2]|0;
   $9 = ($8|0)==(0);
   if ($9) {
    HEAP32[$3>>2] = $4;
   }
   $10 = HEAP32[$__p2>>2]|0;
   $11 = ($10|0)==($__a_end|0);
   do {
    if ($11) {
     $12 = ($8|0)==(34);
     $13 = ($7>>>0)>(0);
     $14 = ($6>>>0)>(4294967295);
     $15 = ($7|0)==(0);
     $16 = $15 & $14;
     $17 = $13 | $16;
     $or$cond = $17 | $12;
     if ($or$cond) {
      HEAP32[$__err>>2] = 4;
      $$0 = -1;
      break;
     } else {
      $$0 = $6;
      break;
     }
    } else {
     HEAP32[$__err>>2] = 4;
     $$0 = 0;
    }
   } while(0);
   $$1 = $$0;
  }
 } while(0);
 STACKTOP = sp;return ($$1|0);
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRy($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__do_get_unsignedIyEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__do_get_unsignedIyEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i = 0, $$0$i$i = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i4 = 0, $$pre$i2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0;
 var $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__g = 0, $__g_end = 0;
 var $__grouping = 0, $__i$0$i$i = 0, $__thousands_sep = 0, $exitcond$i$i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $phitmp$i = 0, $phitmp$i19 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 198|0;
 $__thousands_sep = sp + 196|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $0 = ((($__iob)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 74;
 switch ($2|0) {
 case 64:  {
  $$0$i = 8;
  break;
 }
 case 8:  {
  $$0$i = 16;
  break;
 }
 case 0:  {
  $$0$i = 0;
  break;
 }
 default: {
  $$0$i = 10;
 }
 }
 __ZNSt3__19__num_getIcE17__stage2_int_prepERNS_8ios_baseEPcRc($__grouping,$__iob,$__atoms,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $3 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$3>>2] = 0;
  $4 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $4;
 }
 $5 = HEAP8[$__buf>>0]|0;
 $6 = $5 & 1;
 $7 = ($6<<24>>24)==(0);
 if ($7) {
  $10 = 10;
 } else {
  $8 = HEAP32[$__buf>>2]|0;
  $9 = $8 & -2;
  $phitmp$i = (($9) + -1)|0;
  $10 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($10|0),0);
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 L13: do {
  if ($12) {
   label = 32;
  } else {
   $13 = HEAP8[$__buf>>0]|0;
   $14 = $13 & 1;
   $15 = ($14<<24>>24)==(0);
   $16 = ((($__buf)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($__buf)) + 1|0);
   $19 = $15 ? $18 : $17;
   HEAP32[$__a_end>>2] = $19;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   $20 = ((($__buf)) + 4|0);
   $21 = HEAP8[$__thousands_sep>>0]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i4 = $$pre;$__a$0 = $19;
   L15: while(1) {
    $22 = ($$pre$i$i$i4|0)==(0|0);
    if ($22) {
     $$pre$i$i$i = 0;
    } else {
     $23 = ((($$pre$i$i$i4)) + 12|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ((($$pre$i$i$i4)) + 16|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = ($24|0)==($26|0);
     if ($27) {
      $28 = HEAP32[$$pre$i$i$i4>>2]|0;
      $29 = ((($28)) + 36|0);
      $30 = HEAP32[$29>>2]|0;
      __THREW__ = 0;
      $31 = (invoke_ii($30|0,($$pre$i$i$i4|0))|0);
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if ($33) {
       label = 31;
       break;
      }
      $34 = ($31|0)==(-1);
      if ($34) {
       HEAP32[$__b>>2] = 0;
       $$pre$i$i$i = 0;
      } else {
       $$pre$i$i$i = $$pre$i$i$i4;
      }
     } else {
      $$pre$i$i$i = $$pre$i$i$i4;
     }
    }
    $35 = ($$pre$i$i$i|0)==(0|0);
    $36 = HEAP32[$__e>>2]|0;
    $37 = ($36|0)==(0|0);
    do {
     if ($37) {
      label = 23;
     } else {
      $38 = ((($36)) + 12|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ((($36)) + 16|0);
      $41 = HEAP32[$40>>2]|0;
      $42 = ($39|0)==($41|0);
      if (!($42)) {
       if ($35) {
        $163 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
      $43 = HEAP32[$36>>2]|0;
      $44 = ((($43)) + 36|0);
      $45 = HEAP32[$44>>2]|0;
      __THREW__ = 0;
      $46 = (invoke_ii($45|0,($36|0))|0);
      $47 = __THREW__; __THREW__ = 0;
      $48 = $47&1;
      if ($48) {
       label = 31;
       break L15;
      }
      $49 = ($46|0)==(-1);
      if ($49) {
       HEAP32[$__e>>2] = 0;
       label = 23;
       break;
      } else {
       if ($35) {
        $163 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
     }
    } while(0);
    if ((label|0) == 23) {
     label = 0;
     if ($35) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $163 = 0;
     }
    }
    $50 = HEAP32[$__a_end>>2]|0;
    $51 = HEAP8[$__buf>>0]|0;
    $52 = $51 & 1;
    $53 = ($52<<24>>24)==(0);
    $54 = HEAP32[$20>>2]|0;
    $55 = $51&255;
    $56 = $55 >>> 1;
    $57 = $53 ? $56 : $54;
    $58 = (($__a$0) + ($57)|0);
    $59 = ($50|0)==($58|0);
    if ($59) {
     $60 = $57 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($60|0),0);
     $61 = __THREW__; __THREW__ = 0;
     $62 = $61&1;
     if ($62) {
      label = 31;
      break;
     }
     $63 = HEAP8[$__buf>>0]|0;
     $64 = $63 & 1;
     $65 = ($64<<24>>24)==(0);
     if ($65) {
      $68 = 10;
     } else {
      $66 = HEAP32[$__buf>>2]|0;
      $67 = $66 & -2;
      $phitmp$i19 = (($67) + -1)|0;
      $68 = $phitmp$i19;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($68|0),0);
     $69 = __THREW__; __THREW__ = 0;
     $70 = $69&1;
     if ($70) {
      label = 31;
      break;
     }
     $71 = HEAP8[$__buf>>0]|0;
     $72 = $71 & 1;
     $73 = ($72<<24>>24)==(0);
     $74 = HEAP32[$16>>2]|0;
     $75 = $73 ? $18 : $74;
     $76 = (($75) + ($57)|0);
     HEAP32[$__a_end>>2] = $76;
     $__a$1 = $75;
    } else {
     $__a$1 = $__a$0;
    }
    $81 = ((($$pre$i$i$i)) + 12|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = ((($$pre$i$i$i)) + 16|0);
    $84 = HEAP32[$83>>2]|0;
    $85 = ($82|0)==($84|0);
    if ($85) {
     $86 = HEAP32[$$pre$i$i$i>>2]|0;
     $87 = ((($86)) + 36|0);
     $88 = HEAP32[$87>>2]|0;
     __THREW__ = 0;
     $89 = (invoke_ii($88|0,($$pre$i$i$i|0))|0);
     $90 = __THREW__; __THREW__ = 0;
     $91 = $90&1;
     if ($91) {
      label = 31;
      break;
     } else {
      $$0$i$i = $89;
     }
    } else {
     $92 = HEAP8[$82>>0]|0;
     $93 = $92&255;
     $$0$i$i = $93;
    }
    $94 = $$0$i$i&255;
    $95 = (__ZNSt3__19__num_getIcE17__stage2_int_loopEciPcRS2_RjcRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_S2_($94,$$0$i,$__a$1,$__a_end,$__dc,$21,$__grouping,$__g,$__g_end,$__atoms)|0);
    $96 = ($95|0)==(0);
    if (!($96)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $163;$__a$2 = $__a$1;
     break;
    }
    $97 = HEAP32[$81>>2]|0;
    $98 = HEAP32[$83>>2]|0;
    $99 = ($97|0)==($98|0);
    if ($99) {
     $100 = HEAP32[$$pre$i$i$i>>2]|0;
     $101 = ((($100)) + 40|0);
     $102 = HEAP32[$101>>2]|0;
     __THREW__ = 0;
     (invoke_ii($102|0,($$pre$i$i$i|0))|0);
     $103 = __THREW__; __THREW__ = 0;
     $104 = $103&1;
     if ($104) {
      label = 31;
      break;
     } else {
      $$pre$i$i$i4 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $105 = ((($97)) + 1|0);
     HEAP32[$81>>2] = $105;
     $$pre$i$i$i4 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 31) {
    $77 = ___cxa_find_matching_catch_2()|0;
    $78 = tempRet0;
    $lpad$phi$index = $77;$lpad$phi$index2 = $78;
    break;
   }
   $106 = HEAP8[$__grouping>>0]|0;
   $107 = $106 & 1;
   $108 = ($107<<24>>24)==(0);
   $109 = ((($__grouping)) + 4|0);
   $110 = HEAP32[$109>>2]|0;
   $111 = $106&255;
   $112 = $111 >>> 1;
   $113 = $108 ? $112 : $110;
   $114 = ($113|0)==(0);
   if (!($114)) {
    $115 = HEAP32[$__g_end>>2]|0;
    $116 = $115;
    $117 = $__g;
    $118 = (($116) - ($117))|0;
    $119 = ($118|0)<(160);
    if ($119) {
     $120 = HEAP32[$__dc>>2]|0;
     $121 = ((($115)) + 4|0);
     HEAP32[$__g_end>>2] = $121;
     HEAP32[$115>>2] = $120;
    }
   }
   $122 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $123 = (invoke_iiiii(4870,($__a$2|0),($122|0),($__err|0),($$0$i|0))|0);
   $124 = tempRet0;
   $125 = __THREW__; __THREW__ = 0;
   $126 = $125&1;
   if ($126) {
    label = 32;
   } else {
    $127 = $__v;
    $128 = $127;
    HEAP32[$128>>2] = $123;
    $129 = (($127) + 4)|0;
    $130 = $129;
    HEAP32[$130>>2] = $124;
    $131 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$131,$__err);
    $132 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($132) {
     $146 = 0;
    } else {
     $133 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $134 = HEAP32[$133>>2]|0;
     $135 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $136 = HEAP32[$135>>2]|0;
     $137 = ($134|0)==($136|0);
     if ($137) {
      $138 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $139 = ((($138)) + 36|0);
      $140 = HEAP32[$139>>2]|0;
      __THREW__ = 0;
      $141 = (invoke_ii($140|0,($$pre$i$i$i$lcssa|0))|0);
      $142 = __THREW__; __THREW__ = 0;
      $143 = $142&1;
      if ($143) {
       label = 32;
       break;
      }
      $144 = ($141|0)==(-1);
      if ($144) {
       HEAP32[$__b>>2] = 0;
       $146 = 0;
      } else {
       $146 = $$pre$i$i$i$lcssa;
      }
     } else {
      $146 = $$pre$i$i$i$lcssa;
     }
    }
    $145 = ($146|0)==(0|0);
    $147 = ($$pre$i2$i$i|0)==(0|0);
    do {
     if ($147) {
      label = 56;
     } else {
      $148 = ((($$pre$i2$i$i)) + 12|0);
      $149 = HEAP32[$148>>2]|0;
      $150 = ((($$pre$i2$i$i)) + 16|0);
      $151 = HEAP32[$150>>2]|0;
      $152 = ($149|0)==($151|0);
      if ($152) {
       $153 = HEAP32[$$pre$i2$i$i>>2]|0;
       $154 = ((($153)) + 36|0);
       $155 = HEAP32[$154>>2]|0;
       __THREW__ = 0;
       $156 = (invoke_ii($155|0,($$pre$i2$i$i|0))|0);
       $157 = __THREW__; __THREW__ = 0;
       $158 = $157&1;
       if ($158) {
        label = 32;
        break L13;
       }
       $159 = ($156|0)==(-1);
       if ($159) {
        HEAP32[$__e>>2] = 0;
        label = 56;
        break;
       }
      }
      if (!($145)) {
       label = 57;
      }
     }
    } while(0);
    if ((label|0) == 56) {
     if ($145) {
      label = 57;
     }
    }
    if ((label|0) == 57) {
     $160 = HEAP32[$__err>>2]|0;
     $161 = $160 | 2;
     HEAP32[$__err>>2] = $161;
    }
    $162 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($162|0);
   }
  }
 } while(0);
 if ((label|0) == 32) {
  $79 = ___cxa_find_matching_catch_2()|0;
  $80 = tempRet0;
  $lpad$phi$index = $79;$lpad$phi$index2 = $80;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__127__num_get_unsigned_integralIyEET_PKcS3_Rji($__a,$__a_end,$__err,$__base) {
 $__a = $__a|0;
 $__a_end = $__a_end|0;
 $__err = $__err|0;
 $__base = $__base|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__p2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__p2 = sp;
 $0 = ($__a|0)==($__a_end|0);
 do {
  if ($0) {
   HEAP32[$__err>>2] = 4;
   $13 = 0;$14 = 0;
  } else {
   $1 = HEAP8[$__a>>0]|0;
   $2 = ($1<<24>>24)==(45);
   if ($2) {
    HEAP32[$__err>>2] = 4;
    $13 = 0;$14 = 0;
    break;
   }
   $3 = (___errno_location()|0);
   $4 = HEAP32[$3>>2]|0;
   HEAP32[$3>>2] = 0;
   $5 = (__ZNSt3__16__clocEv()|0);
   $6 = (_strtoull_l($__a,$__p2,$__base,$5)|0);
   $7 = tempRet0;
   $8 = HEAP32[$3>>2]|0;
   $9 = ($8|0)==(0);
   if ($9) {
    HEAP32[$3>>2] = $4;
   }
   $10 = HEAP32[$__p2>>2]|0;
   $11 = ($10|0)==($__a_end|0);
   if ($11) {
    $12 = ($8|0)==(34);
    if ($12) {
     HEAP32[$__err>>2] = 4;
     $15 = -1;$16 = -1;
    } else {
     $15 = $6;$16 = $7;
    }
   } else {
    HEAP32[$__err>>2] = 4;
    $15 = 0;$16 = 0;
   }
   $13 = $16;$14 = $15;
  }
 } while(0);
 tempRet0 = ($13);
 STACKTOP = sp;return ($14|0);
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRf($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE23__do_get_floating_pointIfEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE23__do_get_floating_pointIfEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i$i = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i5 = 0, $$pre$i2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0.0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__decimal_point = 0, $__exp = 0, $__g = 0, $__g_end = 0, $__grouping = 0, $__i$0$i$i = 0, $__in_units = 0, $__thousands_sep = 0;
 var $exitcond$i$i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $or$cond = 0, $phitmp$i = 0, $phitmp$i20 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 200|0;
 $__decimal_point = sp + 199|0;
 $__thousands_sep = sp + 198|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $__in_units = sp + 197|0;
 $__exp = sp + 196|0;
 __ZNSt3__19__num_getIcE19__stage2_float_prepERNS_8ios_baseEPcRcS5_($__grouping,$__iob,$__atoms,$__decimal_point,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 $2 = HEAP8[$__buf>>0]|0;
 $3 = $2 & 1;
 $4 = ($3<<24>>24)==(0);
 if ($4) {
  $7 = 10;
 } else {
  $5 = HEAP32[$__buf>>2]|0;
  $6 = $5 & -2;
  $phitmp$i = (($6) + -1)|0;
  $7 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($7|0),0);
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 L8: do {
  if ($9) {
   label = 28;
  } else {
   $10 = HEAP8[$__buf>>0]|0;
   $11 = $10 & 1;
   $12 = ($11<<24>>24)==(0);
   $13 = ((($__buf)) + 8|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ((($__buf)) + 1|0);
   $16 = $12 ? $15 : $14;
   HEAP32[$__a_end>>2] = $16;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   HEAP8[$__in_units>>0] = 1;
   HEAP8[$__exp>>0] = 69;
   $17 = ((($__buf)) + 4|0);
   $18 = HEAP8[$__decimal_point>>0]|0;
   $19 = HEAP8[$__thousands_sep>>0]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i5 = $$pre;$__a$0 = $16;
   L10: while(1) {
    $20 = ($$pre$i$i$i5|0)==(0|0);
    if ($20) {
     $$pre$i$i$i = 0;
    } else {
     $21 = ((($$pre$i$i$i5)) + 12|0);
     $22 = HEAP32[$21>>2]|0;
     $23 = ((($$pre$i$i$i5)) + 16|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ($22|0)==($24|0);
     if ($25) {
      $26 = HEAP32[$$pre$i$i$i5>>2]|0;
      $27 = ((($26)) + 36|0);
      $28 = HEAP32[$27>>2]|0;
      __THREW__ = 0;
      $29 = (invoke_ii($28|0,($$pre$i$i$i5|0))|0);
      $30 = __THREW__; __THREW__ = 0;
      $31 = $30&1;
      if ($31) {
       label = 27;
       break;
      }
      $32 = ($29|0)==(-1);
      if ($32) {
       HEAP32[$__b>>2] = 0;
       $$pre$i$i$i = 0;
      } else {
       $$pre$i$i$i = $$pre$i$i$i5;
      }
     } else {
      $$pre$i$i$i = $$pre$i$i$i5;
     }
    }
    $33 = ($$pre$i$i$i|0)==(0|0);
    $34 = HEAP32[$__e>>2]|0;
    $35 = ($34|0)==(0|0);
    do {
     if ($35) {
      label = 19;
     } else {
      $36 = ((($34)) + 12|0);
      $37 = HEAP32[$36>>2]|0;
      $38 = ((($34)) + 16|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ($37|0)==($39|0);
      if (!($40)) {
       if ($33) {
        $158 = $34;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $34;$__a$2 = $__a$0;
        break L10;
       }
      }
      $41 = HEAP32[$34>>2]|0;
      $42 = ((($41)) + 36|0);
      $43 = HEAP32[$42>>2]|0;
      __THREW__ = 0;
      $44 = (invoke_ii($43|0,($34|0))|0);
      $45 = __THREW__; __THREW__ = 0;
      $46 = $45&1;
      if ($46) {
       label = 27;
       break L10;
      }
      $47 = ($44|0)==(-1);
      if ($47) {
       HEAP32[$__e>>2] = 0;
       label = 19;
       break;
      } else {
       if ($33) {
        $158 = $34;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $34;$__a$2 = $__a$0;
        break L10;
       }
      }
     }
    } while(0);
    if ((label|0) == 19) {
     label = 0;
     if ($33) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $158 = 0;
     }
    }
    $48 = HEAP32[$__a_end>>2]|0;
    $49 = HEAP8[$__buf>>0]|0;
    $50 = $49 & 1;
    $51 = ($50<<24>>24)==(0);
    $52 = HEAP32[$17>>2]|0;
    $53 = $49&255;
    $54 = $53 >>> 1;
    $55 = $51 ? $54 : $52;
    $56 = (($__a$0) + ($55)|0);
    $57 = ($48|0)==($56|0);
    if ($57) {
     $58 = $55 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($58|0),0);
     $59 = __THREW__; __THREW__ = 0;
     $60 = $59&1;
     if ($60) {
      label = 27;
      break;
     }
     $61 = HEAP8[$__buf>>0]|0;
     $62 = $61 & 1;
     $63 = ($62<<24>>24)==(0);
     if ($63) {
      $66 = 10;
     } else {
      $64 = HEAP32[$__buf>>2]|0;
      $65 = $64 & -2;
      $phitmp$i20 = (($65) + -1)|0;
      $66 = $phitmp$i20;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($66|0),0);
     $67 = __THREW__; __THREW__ = 0;
     $68 = $67&1;
     if ($68) {
      label = 27;
      break;
     }
     $69 = HEAP8[$__buf>>0]|0;
     $70 = $69 & 1;
     $71 = ($70<<24>>24)==(0);
     $72 = HEAP32[$13>>2]|0;
     $73 = $71 ? $15 : $72;
     $74 = (($73) + ($55)|0);
     HEAP32[$__a_end>>2] = $74;
     $__a$1 = $73;
    } else {
     $__a$1 = $__a$0;
    }
    $79 = ((($$pre$i$i$i)) + 12|0);
    $80 = HEAP32[$79>>2]|0;
    $81 = ((($$pre$i$i$i)) + 16|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = ($80|0)==($82|0);
    if ($83) {
     $84 = HEAP32[$$pre$i$i$i>>2]|0;
     $85 = ((($84)) + 36|0);
     $86 = HEAP32[$85>>2]|0;
     __THREW__ = 0;
     $87 = (invoke_ii($86|0,($$pre$i$i$i|0))|0);
     $88 = __THREW__; __THREW__ = 0;
     $89 = $88&1;
     if ($89) {
      label = 27;
      break;
     } else {
      $$0$i$i = $87;
     }
    } else {
     $90 = HEAP8[$80>>0]|0;
     $91 = $90&255;
     $$0$i$i = $91;
    }
    $92 = $$0$i$i&255;
    $93 = (__ZNSt3__19__num_getIcE19__stage2_float_loopEcRbRcPcRS4_ccRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSE_RjS4_($92,$__in_units,$__exp,$__a$1,$__a_end,$18,$19,$__grouping,$__g,$__g_end,$__dc,$__atoms)|0);
    $94 = ($93|0)==(0);
    if (!($94)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $158;$__a$2 = $__a$1;
     break;
    }
    $95 = HEAP32[$79>>2]|0;
    $96 = HEAP32[$81>>2]|0;
    $97 = ($95|0)==($96|0);
    if ($97) {
     $98 = HEAP32[$$pre$i$i$i>>2]|0;
     $99 = ((($98)) + 40|0);
     $100 = HEAP32[$99>>2]|0;
     __THREW__ = 0;
     (invoke_ii($100|0,($$pre$i$i$i|0))|0);
     $101 = __THREW__; __THREW__ = 0;
     $102 = $101&1;
     if ($102) {
      label = 27;
      break;
     } else {
      $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $103 = ((($95)) + 1|0);
     HEAP32[$79>>2] = $103;
     $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 27) {
    $75 = ___cxa_find_matching_catch_2()|0;
    $76 = tempRet0;
    $lpad$phi$index = $75;$lpad$phi$index2 = $76;
    break;
   }
   $104 = HEAP8[$__grouping>>0]|0;
   $105 = $104 & 1;
   $106 = ($105<<24>>24)==(0);
   $107 = ((($__grouping)) + 4|0);
   $108 = HEAP32[$107>>2]|0;
   $109 = $104&255;
   $110 = $109 >>> 1;
   $111 = $106 ? $110 : $108;
   $112 = ($111|0)==(0);
   $113 = HEAP8[$__in_units>>0]|0;
   $114 = ($113<<24>>24)==(0);
   $or$cond = $114 | $112;
   if (!($or$cond)) {
    $115 = HEAP32[$__g_end>>2]|0;
    $116 = $115;
    $117 = $__g;
    $118 = (($116) - ($117))|0;
    $119 = ($118|0)<(160);
    if ($119) {
     $120 = HEAP32[$__dc>>2]|0;
     $121 = ((($115)) + 4|0);
     HEAP32[$__g_end>>2] = $121;
     HEAP32[$115>>2] = $120;
    }
   }
   $122 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $123 = (+invoke_diii(4871,($__a$2|0),($122|0),($__err|0)));
   $124 = __THREW__; __THREW__ = 0;
   $125 = $124&1;
   if ($125) {
    label = 28;
   } else {
    HEAPF32[$__v>>2] = $123;
    $126 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$126,$__err);
    $127 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($127) {
     $141 = 0;
    } else {
     $128 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $129 = HEAP32[$128>>2]|0;
     $130 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $131 = HEAP32[$130>>2]|0;
     $132 = ($129|0)==($131|0);
     if ($132) {
      $133 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $134 = ((($133)) + 36|0);
      $135 = HEAP32[$134>>2]|0;
      __THREW__ = 0;
      $136 = (invoke_ii($135|0,($$pre$i$i$i$lcssa|0))|0);
      $137 = __THREW__; __THREW__ = 0;
      $138 = $137&1;
      if ($138) {
       label = 28;
       break;
      }
      $139 = ($136|0)==(-1);
      if ($139) {
       HEAP32[$__b>>2] = 0;
       $141 = 0;
      } else {
       $141 = $$pre$i$i$i$lcssa;
      }
     } else {
      $141 = $$pre$i$i$i$lcssa;
     }
    }
    $140 = ($141|0)==(0|0);
    $142 = ($$pre$i2$i$i|0)==(0|0);
    do {
     if ($142) {
      label = 52;
     } else {
      $143 = ((($$pre$i2$i$i)) + 12|0);
      $144 = HEAP32[$143>>2]|0;
      $145 = ((($$pre$i2$i$i)) + 16|0);
      $146 = HEAP32[$145>>2]|0;
      $147 = ($144|0)==($146|0);
      if ($147) {
       $148 = HEAP32[$$pre$i2$i$i>>2]|0;
       $149 = ((($148)) + 36|0);
       $150 = HEAP32[$149>>2]|0;
       __THREW__ = 0;
       $151 = (invoke_ii($150|0,($$pre$i2$i$i|0))|0);
       $152 = __THREW__; __THREW__ = 0;
       $153 = $152&1;
       if ($153) {
        label = 28;
        break L8;
       }
       $154 = ($151|0)==(-1);
       if ($154) {
        HEAP32[$__e>>2] = 0;
        label = 52;
        break;
       }
      }
      if (!($140)) {
       label = 53;
      }
     }
    } while(0);
    if ((label|0) == 52) {
     if ($140) {
      label = 53;
     }
    }
    if ((label|0) == 53) {
     $155 = HEAP32[$__err>>2]|0;
     $156 = $155 | 2;
     HEAP32[$__err>>2] = $156;
    }
    $157 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($157|0);
   }
  }
 } while(0);
 if ((label|0) == 28) {
  $77 = ___cxa_find_matching_catch_2()|0;
  $78 = tempRet0;
  $lpad$phi$index = $77;$lpad$phi$index2 = $78;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__19__num_getIcE19__stage2_float_prepERNS_8ios_baseEPcRcS5_($agg$result,$__iob,$__atoms,$__decimal_point,$__thousands_sep) {
 $agg$result = $agg$result|0;
 $__iob = $__iob|0;
 $__atoms = $__atoms|0;
 $__decimal_point = $__decimal_point|0;
 $__thousands_sep = $__thousands_sep|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__loc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__loc = sp;
 $0 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$__loc>>2] = $0;
 __THREW__ = 0;
 $1 = (invoke_iii(3374,($__loc|0),(1207576|0))|0);
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  $4 = HEAP32[$1>>2]|0;
  $5 = ((($4)) + 32|0);
  $6 = HEAP32[$5>>2]|0;
  __THREW__ = 0;
  (invoke_iiiii($6|0,($1|0),(620486|0),((620518)|0),($__atoms|0))|0);
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if (!($8)) {
   __THREW__ = 0;
   $9 = (invoke_iii(3374,($__loc|0),(1207588|0))|0);
   $10 = __THREW__; __THREW__ = 0;
   $11 = $10&1;
   if (!($11)) {
    $12 = HEAP32[$9>>2]|0;
    $13 = ((($12)) + 12|0);
    $14 = HEAP32[$13>>2]|0;
    __THREW__ = 0;
    $15 = (invoke_ii($14|0,($9|0))|0);
    $16 = __THREW__; __THREW__ = 0;
    $17 = $16&1;
    if (!($17)) {
     HEAP8[$__decimal_point>>0] = $15;
     $18 = HEAP32[$9>>2]|0;
     $19 = ((($18)) + 16|0);
     $20 = HEAP32[$19>>2]|0;
     __THREW__ = 0;
     $21 = (invoke_ii($20|0,($9|0))|0);
     $22 = __THREW__; __THREW__ = 0;
     $23 = $22&1;
     if (!($23)) {
      HEAP8[$__thousands_sep>>0] = $21;
      $24 = HEAP32[$9>>2]|0;
      $25 = ((($24)) + 20|0);
      $26 = HEAP32[$25>>2]|0;
      __THREW__ = 0;
      invoke_vii($26|0,($agg$result|0),($9|0));
      $27 = __THREW__; __THREW__ = 0;
      $28 = $27&1;
      if (!($28)) {
       (__ZNSt3__114__shared_count16__release_sharedEv($0)|0);
       STACKTOP = sp;return;
      }
     }
    }
   }
  }
 }
 $29 = ___cxa_find_matching_catch_2()|0;
 $30 = tempRet0;
 (__ZNSt3__114__shared_count16__release_sharedEv($0)|0);
 ___resumeException($29|0);
 // unreachable;
}
function __ZNSt3__19__num_getIcE19__stage2_float_loopEcRbRcPcRS4_ccRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSE_RjS4_($__ct,$__in_units,$__exp,$__a,$__a_end,$__decimal_point,$__thousands_sep,$__grouping,$__g,$__g_end,$__dc,$__atoms) {
 $__ct = $__ct|0;
 $__in_units = $__in_units|0;
 $__exp = $__exp|0;
 $__a = $__a|0;
 $__a_end = $__a_end|0;
 $__decimal_point = $__decimal_point|0;
 $__thousands_sep = $__thousands_sep|0;
 $__grouping = $__grouping|0;
 $__g = $__g|0;
 $__g_end = $__g_end|0;
 $__dc = $__dc|0;
 $__atoms = $__atoms|0;
 var $$0$i = 0, $$0$lcssa$i = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__ct<<24>>24)==($__decimal_point<<24>>24);
 L1: do {
  if ($0) {
   $1 = HEAP8[$__in_units>>0]|0;
   $2 = ($1<<24>>24)==(0);
   if ($2) {
    $$2 = -1;
   } else {
    HEAP8[$__in_units>>0] = 0;
    $3 = HEAP32[$__a_end>>2]|0;
    $4 = ((($3)) + 1|0);
    HEAP32[$__a_end>>2] = $4;
    HEAP8[$3>>0] = 46;
    $5 = HEAP8[$__grouping>>0]|0;
    $6 = $5 & 1;
    $7 = ($6<<24>>24)==(0);
    $8 = ((($__grouping)) + 4|0);
    $9 = HEAP32[$8>>2]|0;
    $10 = $5&255;
    $11 = $10 >>> 1;
    $12 = $7 ? $11 : $9;
    $13 = ($12|0)==(0);
    if ($13) {
     $$2 = 0;
    } else {
     $14 = HEAP32[$__g_end>>2]|0;
     $15 = $14;
     $16 = $__g;
     $17 = (($15) - ($16))|0;
     $18 = ($17|0)<(160);
     if ($18) {
      $19 = HEAP32[$__dc>>2]|0;
      $20 = ((($14)) + 4|0);
      HEAP32[$__g_end>>2] = $20;
      HEAP32[$14>>2] = $19;
      $$2 = 0;
     } else {
      $$2 = 0;
     }
    }
   }
  } else {
   $21 = ($__ct<<24>>24)==($__thousands_sep<<24>>24);
   if ($21) {
    $22 = HEAP8[$__grouping>>0]|0;
    $23 = $22 & 1;
    $24 = ($23<<24>>24)==(0);
    $25 = ((($__grouping)) + 4|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = $22&255;
    $28 = $27 >>> 1;
    $29 = $24 ? $28 : $26;
    $30 = ($29|0)==(0);
    if (!($30)) {
     $31 = HEAP8[$__in_units>>0]|0;
     $32 = ($31<<24>>24)==(0);
     if ($32) {
      $$2 = -1;
      break;
     }
     $33 = HEAP32[$__g_end>>2]|0;
     $34 = $33;
     $35 = $__g;
     $36 = (($34) - ($35))|0;
     $37 = ($36|0)<(160);
     if (!($37)) {
      $$2 = 0;
      break;
     }
     $38 = HEAP32[$__dc>>2]|0;
     $39 = ((($33)) + 4|0);
     HEAP32[$__g_end>>2] = $39;
     HEAP32[$33>>2] = $38;
     HEAP32[$__dc>>2] = 0;
     $$2 = 0;
     break;
    }
   }
   $40 = ((($__atoms)) + 32|0);
   $$0$i = $__atoms;
   while(1) {
    $41 = ($$0$i|0)==($40|0);
    if ($41) {
     $$0$lcssa$i = $40;
     break;
    }
    $42 = HEAP8[$$0$i>>0]|0;
    $43 = ($42<<24>>24)==($__ct<<24>>24);
    if ($43) {
     $$0$lcssa$i = $$0$i;
     break;
    }
    $44 = ((($$0$i)) + 1|0);
    $$0$i = $44;
   }
   $45 = $$0$lcssa$i;
   $46 = $__atoms;
   $47 = (($45) - ($46))|0;
   $48 = ($47|0)>(31);
   if ($48) {
    $$2 = -1;
   } else {
    $49 = (620486 + ($47)|0);
    $50 = HEAP8[$49>>0]|0;
    switch ($47|0) {
    case 24: case 25:  {
     $51 = HEAP32[$__a_end>>2]|0;
     $52 = ($51|0)==($__a|0);
     if (!($52)) {
      $53 = ((($51)) + -1|0);
      $54 = HEAP8[$53>>0]|0;
      $55 = $54&255;
      $56 = $55 & 95;
      $57 = HEAP8[$__exp>>0]|0;
      $58 = $57&255;
      $59 = $58 & 127;
      $60 = ($56|0)==($59|0);
      if (!($60)) {
       $$2 = -1;
       break L1;
      }
     }
     $61 = ((($51)) + 1|0);
     HEAP32[$__a_end>>2] = $61;
     HEAP8[$51>>0] = $50;
     $$2 = 0;
     break L1;
     break;
    }
    case 23: case 22:  {
     HEAP8[$__exp>>0] = 80;
     $62 = HEAP32[$__a_end>>2]|0;
     $63 = ((($62)) + 1|0);
     HEAP32[$__a_end>>2] = $63;
     HEAP8[$62>>0] = $50;
     $$2 = 0;
     break L1;
     break;
    }
    default: {
     $64 = $50&255;
     $65 = $64 & 95;
     $66 = HEAP8[$__exp>>0]|0;
     $67 = $66 << 24 >> 24;
     $68 = ($65|0)==($67|0);
     if ($68) {
      $69 = $65 | 128;
      $70 = $69&255;
      HEAP8[$__exp>>0] = $70;
      $71 = HEAP8[$__in_units>>0]|0;
      $72 = ($71<<24>>24)==(0);
      if (!($72)) {
       HEAP8[$__in_units>>0] = 0;
       $73 = HEAP8[$__grouping>>0]|0;
       $74 = $73 & 1;
       $75 = ($74<<24>>24)==(0);
       $76 = ((($__grouping)) + 4|0);
       $77 = HEAP32[$76>>2]|0;
       $78 = $73&255;
       $79 = $78 >>> 1;
       $80 = $75 ? $79 : $77;
       $81 = ($80|0)==(0);
       if (!($81)) {
        $82 = HEAP32[$__g_end>>2]|0;
        $83 = $82;
        $84 = $__g;
        $85 = (($83) - ($84))|0;
        $86 = ($85|0)<(160);
        if ($86) {
         $87 = HEAP32[$__dc>>2]|0;
         $88 = ((($82)) + 4|0);
         HEAP32[$__g_end>>2] = $88;
         HEAP32[$82>>2] = $87;
        }
       }
      }
     }
     $89 = HEAP32[$__a_end>>2]|0;
     $90 = ((($89)) + 1|0);
     HEAP32[$__a_end>>2] = $90;
     HEAP8[$89>>0] = $50;
     $91 = ($47|0)>(21);
     if ($91) {
      $$2 = 0;
      break L1;
     }
     $92 = HEAP32[$__dc>>2]|0;
     $93 = (($92) + 1)|0;
     HEAP32[$__dc>>2] = $93;
     $$2 = 0;
     break L1;
    }
    }
   }
  }
 } while(0);
 return ($$2|0);
}
function __ZNSt3__115__num_get_floatIfEET_PKcS3_Rj($__a,$__a_end,$__err) {
 $__a = $__a|0;
 $__a_end = $__a_end|0;
 $__err = $__err|0;
 var $$0 = 0.0, $$1 = 0.0, $0 = 0, $1 = 0, $10 = 0.0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__p2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__p2 = sp;
 $0 = ($__a|0)==($__a_end|0);
 if ($0) {
  HEAP32[$__err>>2] = 4;
  $$1 = 0.0;
 } else {
  $1 = (___errno_location()|0);
  $2 = HEAP32[$1>>2]|0;
  HEAP32[$1>>2] = 0;
  $3 = (__ZNSt3__16__clocEv()|0);
  $4 = (+_strtold_l($__a,$__p2,$3));
  $5 = HEAP32[$1>>2]|0;
  $6 = ($5|0)==(0);
  if ($6) {
   HEAP32[$1>>2] = $2;
  }
  $7 = HEAP32[$__p2>>2]|0;
  $8 = ($7|0)==($__a_end|0);
  if ($8) {
   $9 = ($5|0)==(34);
   if ($9) {
    HEAP32[$__err>>2] = 4;
   }
   $10 = $4;
   $$0 = $10;
  } else {
   HEAP32[$__err>>2] = 4;
   $$0 = 0.0;
  }
  $$1 = $$0;
 }
 STACKTOP = sp;return (+$$1);
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRd($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE23__do_get_floating_pointIdEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE23__do_get_floating_pointIdEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i$i = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i5 = 0, $$pre$i2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0.0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__decimal_point = 0, $__exp = 0, $__g = 0, $__g_end = 0, $__grouping = 0, $__i$0$i$i = 0, $__in_units = 0, $__thousands_sep = 0;
 var $exitcond$i$i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $or$cond = 0, $phitmp$i = 0, $phitmp$i20 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 200|0;
 $__decimal_point = sp + 199|0;
 $__thousands_sep = sp + 198|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $__in_units = sp + 197|0;
 $__exp = sp + 196|0;
 __ZNSt3__19__num_getIcE19__stage2_float_prepERNS_8ios_baseEPcRcS5_($__grouping,$__iob,$__atoms,$__decimal_point,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 $2 = HEAP8[$__buf>>0]|0;
 $3 = $2 & 1;
 $4 = ($3<<24>>24)==(0);
 if ($4) {
  $7 = 10;
 } else {
  $5 = HEAP32[$__buf>>2]|0;
  $6 = $5 & -2;
  $phitmp$i = (($6) + -1)|0;
  $7 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($7|0),0);
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 L8: do {
  if ($9) {
   label = 28;
  } else {
   $10 = HEAP8[$__buf>>0]|0;
   $11 = $10 & 1;
   $12 = ($11<<24>>24)==(0);
   $13 = ((($__buf)) + 8|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ((($__buf)) + 1|0);
   $16 = $12 ? $15 : $14;
   HEAP32[$__a_end>>2] = $16;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   HEAP8[$__in_units>>0] = 1;
   HEAP8[$__exp>>0] = 69;
   $17 = ((($__buf)) + 4|0);
   $18 = HEAP8[$__decimal_point>>0]|0;
   $19 = HEAP8[$__thousands_sep>>0]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i5 = $$pre;$__a$0 = $16;
   L10: while(1) {
    $20 = ($$pre$i$i$i5|0)==(0|0);
    if ($20) {
     $$pre$i$i$i = 0;
    } else {
     $21 = ((($$pre$i$i$i5)) + 12|0);
     $22 = HEAP32[$21>>2]|0;
     $23 = ((($$pre$i$i$i5)) + 16|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ($22|0)==($24|0);
     if ($25) {
      $26 = HEAP32[$$pre$i$i$i5>>2]|0;
      $27 = ((($26)) + 36|0);
      $28 = HEAP32[$27>>2]|0;
      __THREW__ = 0;
      $29 = (invoke_ii($28|0,($$pre$i$i$i5|0))|0);
      $30 = __THREW__; __THREW__ = 0;
      $31 = $30&1;
      if ($31) {
       label = 27;
       break;
      }
      $32 = ($29|0)==(-1);
      if ($32) {
       HEAP32[$__b>>2] = 0;
       $$pre$i$i$i = 0;
      } else {
       $$pre$i$i$i = $$pre$i$i$i5;
      }
     } else {
      $$pre$i$i$i = $$pre$i$i$i5;
     }
    }
    $33 = ($$pre$i$i$i|0)==(0|0);
    $34 = HEAP32[$__e>>2]|0;
    $35 = ($34|0)==(0|0);
    do {
     if ($35) {
      label = 19;
     } else {
      $36 = ((($34)) + 12|0);
      $37 = HEAP32[$36>>2]|0;
      $38 = ((($34)) + 16|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ($37|0)==($39|0);
      if (!($40)) {
       if ($33) {
        $158 = $34;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $34;$__a$2 = $__a$0;
        break L10;
       }
      }
      $41 = HEAP32[$34>>2]|0;
      $42 = ((($41)) + 36|0);
      $43 = HEAP32[$42>>2]|0;
      __THREW__ = 0;
      $44 = (invoke_ii($43|0,($34|0))|0);
      $45 = __THREW__; __THREW__ = 0;
      $46 = $45&1;
      if ($46) {
       label = 27;
       break L10;
      }
      $47 = ($44|0)==(-1);
      if ($47) {
       HEAP32[$__e>>2] = 0;
       label = 19;
       break;
      } else {
       if ($33) {
        $158 = $34;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $34;$__a$2 = $__a$0;
        break L10;
       }
      }
     }
    } while(0);
    if ((label|0) == 19) {
     label = 0;
     if ($33) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $158 = 0;
     }
    }
    $48 = HEAP32[$__a_end>>2]|0;
    $49 = HEAP8[$__buf>>0]|0;
    $50 = $49 & 1;
    $51 = ($50<<24>>24)==(0);
    $52 = HEAP32[$17>>2]|0;
    $53 = $49&255;
    $54 = $53 >>> 1;
    $55 = $51 ? $54 : $52;
    $56 = (($__a$0) + ($55)|0);
    $57 = ($48|0)==($56|0);
    if ($57) {
     $58 = $55 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($58|0),0);
     $59 = __THREW__; __THREW__ = 0;
     $60 = $59&1;
     if ($60) {
      label = 27;
      break;
     }
     $61 = HEAP8[$__buf>>0]|0;
     $62 = $61 & 1;
     $63 = ($62<<24>>24)==(0);
     if ($63) {
      $66 = 10;
     } else {
      $64 = HEAP32[$__buf>>2]|0;
      $65 = $64 & -2;
      $phitmp$i20 = (($65) + -1)|0;
      $66 = $phitmp$i20;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($66|0),0);
     $67 = __THREW__; __THREW__ = 0;
     $68 = $67&1;
     if ($68) {
      label = 27;
      break;
     }
     $69 = HEAP8[$__buf>>0]|0;
     $70 = $69 & 1;
     $71 = ($70<<24>>24)==(0);
     $72 = HEAP32[$13>>2]|0;
     $73 = $71 ? $15 : $72;
     $74 = (($73) + ($55)|0);
     HEAP32[$__a_end>>2] = $74;
     $__a$1 = $73;
    } else {
     $__a$1 = $__a$0;
    }
    $79 = ((($$pre$i$i$i)) + 12|0);
    $80 = HEAP32[$79>>2]|0;
    $81 = ((($$pre$i$i$i)) + 16|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = ($80|0)==($82|0);
    if ($83) {
     $84 = HEAP32[$$pre$i$i$i>>2]|0;
     $85 = ((($84)) + 36|0);
     $86 = HEAP32[$85>>2]|0;
     __THREW__ = 0;
     $87 = (invoke_ii($86|0,($$pre$i$i$i|0))|0);
     $88 = __THREW__; __THREW__ = 0;
     $89 = $88&1;
     if ($89) {
      label = 27;
      break;
     } else {
      $$0$i$i = $87;
     }
    } else {
     $90 = HEAP8[$80>>0]|0;
     $91 = $90&255;
     $$0$i$i = $91;
    }
    $92 = $$0$i$i&255;
    $93 = (__ZNSt3__19__num_getIcE19__stage2_float_loopEcRbRcPcRS4_ccRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSE_RjS4_($92,$__in_units,$__exp,$__a$1,$__a_end,$18,$19,$__grouping,$__g,$__g_end,$__dc,$__atoms)|0);
    $94 = ($93|0)==(0);
    if (!($94)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $158;$__a$2 = $__a$1;
     break;
    }
    $95 = HEAP32[$79>>2]|0;
    $96 = HEAP32[$81>>2]|0;
    $97 = ($95|0)==($96|0);
    if ($97) {
     $98 = HEAP32[$$pre$i$i$i>>2]|0;
     $99 = ((($98)) + 40|0);
     $100 = HEAP32[$99>>2]|0;
     __THREW__ = 0;
     (invoke_ii($100|0,($$pre$i$i$i|0))|0);
     $101 = __THREW__; __THREW__ = 0;
     $102 = $101&1;
     if ($102) {
      label = 27;
      break;
     } else {
      $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $103 = ((($95)) + 1|0);
     HEAP32[$79>>2] = $103;
     $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 27) {
    $75 = ___cxa_find_matching_catch_2()|0;
    $76 = tempRet0;
    $lpad$phi$index = $75;$lpad$phi$index2 = $76;
    break;
   }
   $104 = HEAP8[$__grouping>>0]|0;
   $105 = $104 & 1;
   $106 = ($105<<24>>24)==(0);
   $107 = ((($__grouping)) + 4|0);
   $108 = HEAP32[$107>>2]|0;
   $109 = $104&255;
   $110 = $109 >>> 1;
   $111 = $106 ? $110 : $108;
   $112 = ($111|0)==(0);
   $113 = HEAP8[$__in_units>>0]|0;
   $114 = ($113<<24>>24)==(0);
   $or$cond = $114 | $112;
   if (!($or$cond)) {
    $115 = HEAP32[$__g_end>>2]|0;
    $116 = $115;
    $117 = $__g;
    $118 = (($116) - ($117))|0;
    $119 = ($118|0)<(160);
    if ($119) {
     $120 = HEAP32[$__dc>>2]|0;
     $121 = ((($115)) + 4|0);
     HEAP32[$__g_end>>2] = $121;
     HEAP32[$115>>2] = $120;
    }
   }
   $122 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $123 = (+invoke_diii(4872,($__a$2|0),($122|0),($__err|0)));
   $124 = __THREW__; __THREW__ = 0;
   $125 = $124&1;
   if ($125) {
    label = 28;
   } else {
    HEAPF64[$__v>>3] = $123;
    $126 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$126,$__err);
    $127 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($127) {
     $141 = 0;
    } else {
     $128 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $129 = HEAP32[$128>>2]|0;
     $130 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $131 = HEAP32[$130>>2]|0;
     $132 = ($129|0)==($131|0);
     if ($132) {
      $133 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $134 = ((($133)) + 36|0);
      $135 = HEAP32[$134>>2]|0;
      __THREW__ = 0;
      $136 = (invoke_ii($135|0,($$pre$i$i$i$lcssa|0))|0);
      $137 = __THREW__; __THREW__ = 0;
      $138 = $137&1;
      if ($138) {
       label = 28;
       break;
      }
      $139 = ($136|0)==(-1);
      if ($139) {
       HEAP32[$__b>>2] = 0;
       $141 = 0;
      } else {
       $141 = $$pre$i$i$i$lcssa;
      }
     } else {
      $141 = $$pre$i$i$i$lcssa;
     }
    }
    $140 = ($141|0)==(0|0);
    $142 = ($$pre$i2$i$i|0)==(0|0);
    do {
     if ($142) {
      label = 52;
     } else {
      $143 = ((($$pre$i2$i$i)) + 12|0);
      $144 = HEAP32[$143>>2]|0;
      $145 = ((($$pre$i2$i$i)) + 16|0);
      $146 = HEAP32[$145>>2]|0;
      $147 = ($144|0)==($146|0);
      if ($147) {
       $148 = HEAP32[$$pre$i2$i$i>>2]|0;
       $149 = ((($148)) + 36|0);
       $150 = HEAP32[$149>>2]|0;
       __THREW__ = 0;
       $151 = (invoke_ii($150|0,($$pre$i2$i$i|0))|0);
       $152 = __THREW__; __THREW__ = 0;
       $153 = $152&1;
       if ($153) {
        label = 28;
        break L8;
       }
       $154 = ($151|0)==(-1);
       if ($154) {
        HEAP32[$__e>>2] = 0;
        label = 52;
        break;
       }
      }
      if (!($140)) {
       label = 53;
      }
     }
    } while(0);
    if ((label|0) == 52) {
     if ($140) {
      label = 53;
     }
    }
    if ((label|0) == 53) {
     $155 = HEAP32[$__err>>2]|0;
     $156 = $155 | 2;
     HEAP32[$__err>>2] = $156;
    }
    $157 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($157|0);
   }
  }
 } while(0);
 if ((label|0) == 28) {
  $77 = ___cxa_find_matching_catch_2()|0;
  $78 = tempRet0;
  $lpad$phi$index = $77;$lpad$phi$index2 = $78;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__115__num_get_floatIdEET_PKcS3_Rj($__a,$__a_end,$__err) {
 $__a = $__a|0;
 $__a_end = $__a_end|0;
 $__err = $__err|0;
 var $$0 = 0.0, $$1 = 0.0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__p2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__p2 = sp;
 $0 = ($__a|0)==($__a_end|0);
 if ($0) {
  HEAP32[$__err>>2] = 4;
  $$1 = 0.0;
 } else {
  $1 = (___errno_location()|0);
  $2 = HEAP32[$1>>2]|0;
  HEAP32[$1>>2] = 0;
  $3 = (__ZNSt3__16__clocEv()|0);
  $4 = (+_strtold_l($__a,$__p2,$3));
  $5 = HEAP32[$1>>2]|0;
  $6 = ($5|0)==(0);
  if ($6) {
   HEAP32[$1>>2] = $2;
  }
  $7 = HEAP32[$__p2>>2]|0;
  $8 = ($7|0)==($__a_end|0);
  if ($8) {
   $9 = ($5|0)==(34);
   if ($9) {
    HEAP32[$__err>>2] = 4;
    $$0 = $4;
   } else {
    $$0 = $4;
   }
  } else {
   HEAP32[$__err>>2] = 4;
   $$0 = 0.0;
  }
  $$1 = $$0;
 }
 STACKTOP = sp;return (+$$1);
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRe($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE23__do_get_floating_pointIeEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE23__do_get_floating_pointIeEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i$i = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i5 = 0, $$pre$i2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0.0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__decimal_point = 0, $__exp = 0, $__g = 0, $__g_end = 0, $__grouping = 0, $__i$0$i$i = 0, $__in_units = 0, $__thousands_sep = 0;
 var $exitcond$i$i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $or$cond = 0, $phitmp$i = 0, $phitmp$i20 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 200|0;
 $__decimal_point = sp + 199|0;
 $__thousands_sep = sp + 198|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $__in_units = sp + 197|0;
 $__exp = sp + 196|0;
 __ZNSt3__19__num_getIcE19__stage2_float_prepERNS_8ios_baseEPcRcS5_($__grouping,$__iob,$__atoms,$__decimal_point,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 $2 = HEAP8[$__buf>>0]|0;
 $3 = $2 & 1;
 $4 = ($3<<24>>24)==(0);
 if ($4) {
  $7 = 10;
 } else {
  $5 = HEAP32[$__buf>>2]|0;
  $6 = $5 & -2;
  $phitmp$i = (($6) + -1)|0;
  $7 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($7|0),0);
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 L8: do {
  if ($9) {
   label = 28;
  } else {
   $10 = HEAP8[$__buf>>0]|0;
   $11 = $10 & 1;
   $12 = ($11<<24>>24)==(0);
   $13 = ((($__buf)) + 8|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ((($__buf)) + 1|0);
   $16 = $12 ? $15 : $14;
   HEAP32[$__a_end>>2] = $16;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   HEAP8[$__in_units>>0] = 1;
   HEAP8[$__exp>>0] = 69;
   $17 = ((($__buf)) + 4|0);
   $18 = HEAP8[$__decimal_point>>0]|0;
   $19 = HEAP8[$__thousands_sep>>0]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i5 = $$pre;$__a$0 = $16;
   L10: while(1) {
    $20 = ($$pre$i$i$i5|0)==(0|0);
    if ($20) {
     $$pre$i$i$i = 0;
    } else {
     $21 = ((($$pre$i$i$i5)) + 12|0);
     $22 = HEAP32[$21>>2]|0;
     $23 = ((($$pre$i$i$i5)) + 16|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ($22|0)==($24|0);
     if ($25) {
      $26 = HEAP32[$$pre$i$i$i5>>2]|0;
      $27 = ((($26)) + 36|0);
      $28 = HEAP32[$27>>2]|0;
      __THREW__ = 0;
      $29 = (invoke_ii($28|0,($$pre$i$i$i5|0))|0);
      $30 = __THREW__; __THREW__ = 0;
      $31 = $30&1;
      if ($31) {
       label = 27;
       break;
      }
      $32 = ($29|0)==(-1);
      if ($32) {
       HEAP32[$__b>>2] = 0;
       $$pre$i$i$i = 0;
      } else {
       $$pre$i$i$i = $$pre$i$i$i5;
      }
     } else {
      $$pre$i$i$i = $$pre$i$i$i5;
     }
    }
    $33 = ($$pre$i$i$i|0)==(0|0);
    $34 = HEAP32[$__e>>2]|0;
    $35 = ($34|0)==(0|0);
    do {
     if ($35) {
      label = 19;
     } else {
      $36 = ((($34)) + 12|0);
      $37 = HEAP32[$36>>2]|0;
      $38 = ((($34)) + 16|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ($37|0)==($39|0);
      if (!($40)) {
       if ($33) {
        $158 = $34;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $34;$__a$2 = $__a$0;
        break L10;
       }
      }
      $41 = HEAP32[$34>>2]|0;
      $42 = ((($41)) + 36|0);
      $43 = HEAP32[$42>>2]|0;
      __THREW__ = 0;
      $44 = (invoke_ii($43|0,($34|0))|0);
      $45 = __THREW__; __THREW__ = 0;
      $46 = $45&1;
      if ($46) {
       label = 27;
       break L10;
      }
      $47 = ($44|0)==(-1);
      if ($47) {
       HEAP32[$__e>>2] = 0;
       label = 19;
       break;
      } else {
       if ($33) {
        $158 = $34;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $34;$__a$2 = $__a$0;
        break L10;
       }
      }
     }
    } while(0);
    if ((label|0) == 19) {
     label = 0;
     if ($33) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $158 = 0;
     }
    }
    $48 = HEAP32[$__a_end>>2]|0;
    $49 = HEAP8[$__buf>>0]|0;
    $50 = $49 & 1;
    $51 = ($50<<24>>24)==(0);
    $52 = HEAP32[$17>>2]|0;
    $53 = $49&255;
    $54 = $53 >>> 1;
    $55 = $51 ? $54 : $52;
    $56 = (($__a$0) + ($55)|0);
    $57 = ($48|0)==($56|0);
    if ($57) {
     $58 = $55 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($58|0),0);
     $59 = __THREW__; __THREW__ = 0;
     $60 = $59&1;
     if ($60) {
      label = 27;
      break;
     }
     $61 = HEAP8[$__buf>>0]|0;
     $62 = $61 & 1;
     $63 = ($62<<24>>24)==(0);
     if ($63) {
      $66 = 10;
     } else {
      $64 = HEAP32[$__buf>>2]|0;
      $65 = $64 & -2;
      $phitmp$i20 = (($65) + -1)|0;
      $66 = $phitmp$i20;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($66|0),0);
     $67 = __THREW__; __THREW__ = 0;
     $68 = $67&1;
     if ($68) {
      label = 27;
      break;
     }
     $69 = HEAP8[$__buf>>0]|0;
     $70 = $69 & 1;
     $71 = ($70<<24>>24)==(0);
     $72 = HEAP32[$13>>2]|0;
     $73 = $71 ? $15 : $72;
     $74 = (($73) + ($55)|0);
     HEAP32[$__a_end>>2] = $74;
     $__a$1 = $73;
    } else {
     $__a$1 = $__a$0;
    }
    $79 = ((($$pre$i$i$i)) + 12|0);
    $80 = HEAP32[$79>>2]|0;
    $81 = ((($$pre$i$i$i)) + 16|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = ($80|0)==($82|0);
    if ($83) {
     $84 = HEAP32[$$pre$i$i$i>>2]|0;
     $85 = ((($84)) + 36|0);
     $86 = HEAP32[$85>>2]|0;
     __THREW__ = 0;
     $87 = (invoke_ii($86|0,($$pre$i$i$i|0))|0);
     $88 = __THREW__; __THREW__ = 0;
     $89 = $88&1;
     if ($89) {
      label = 27;
      break;
     } else {
      $$0$i$i = $87;
     }
    } else {
     $90 = HEAP8[$80>>0]|0;
     $91 = $90&255;
     $$0$i$i = $91;
    }
    $92 = $$0$i$i&255;
    $93 = (__ZNSt3__19__num_getIcE19__stage2_float_loopEcRbRcPcRS4_ccRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSE_RjS4_($92,$__in_units,$__exp,$__a$1,$__a_end,$18,$19,$__grouping,$__g,$__g_end,$__dc,$__atoms)|0);
    $94 = ($93|0)==(0);
    if (!($94)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $158;$__a$2 = $__a$1;
     break;
    }
    $95 = HEAP32[$79>>2]|0;
    $96 = HEAP32[$81>>2]|0;
    $97 = ($95|0)==($96|0);
    if ($97) {
     $98 = HEAP32[$$pre$i$i$i>>2]|0;
     $99 = ((($98)) + 40|0);
     $100 = HEAP32[$99>>2]|0;
     __THREW__ = 0;
     (invoke_ii($100|0,($$pre$i$i$i|0))|0);
     $101 = __THREW__; __THREW__ = 0;
     $102 = $101&1;
     if ($102) {
      label = 27;
      break;
     } else {
      $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $103 = ((($95)) + 1|0);
     HEAP32[$79>>2] = $103;
     $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 27) {
    $75 = ___cxa_find_matching_catch_2()|0;
    $76 = tempRet0;
    $lpad$phi$index = $75;$lpad$phi$index2 = $76;
    break;
   }
   $104 = HEAP8[$__grouping>>0]|0;
   $105 = $104 & 1;
   $106 = ($105<<24>>24)==(0);
   $107 = ((($__grouping)) + 4|0);
   $108 = HEAP32[$107>>2]|0;
   $109 = $104&255;
   $110 = $109 >>> 1;
   $111 = $106 ? $110 : $108;
   $112 = ($111|0)==(0);
   $113 = HEAP8[$__in_units>>0]|0;
   $114 = ($113<<24>>24)==(0);
   $or$cond = $114 | $112;
   if (!($or$cond)) {
    $115 = HEAP32[$__g_end>>2]|0;
    $116 = $115;
    $117 = $__g;
    $118 = (($116) - ($117))|0;
    $119 = ($118|0)<(160);
    if ($119) {
     $120 = HEAP32[$__dc>>2]|0;
     $121 = ((($115)) + 4|0);
     HEAP32[$__g_end>>2] = $121;
     HEAP32[$115>>2] = $120;
    }
   }
   $122 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $123 = (+invoke_diii(4873,($__a$2|0),($122|0),($__err|0)));
   $124 = __THREW__; __THREW__ = 0;
   $125 = $124&1;
   if ($125) {
    label = 28;
   } else {
    HEAPF64[$__v>>3] = $123;
    $126 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$126,$__err);
    $127 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($127) {
     $141 = 0;
    } else {
     $128 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $129 = HEAP32[$128>>2]|0;
     $130 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $131 = HEAP32[$130>>2]|0;
     $132 = ($129|0)==($131|0);
     if ($132) {
      $133 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $134 = ((($133)) + 36|0);
      $135 = HEAP32[$134>>2]|0;
      __THREW__ = 0;
      $136 = (invoke_ii($135|0,($$pre$i$i$i$lcssa|0))|0);
      $137 = __THREW__; __THREW__ = 0;
      $138 = $137&1;
      if ($138) {
       label = 28;
       break;
      }
      $139 = ($136|0)==(-1);
      if ($139) {
       HEAP32[$__b>>2] = 0;
       $141 = 0;
      } else {
       $141 = $$pre$i$i$i$lcssa;
      }
     } else {
      $141 = $$pre$i$i$i$lcssa;
     }
    }
    $140 = ($141|0)==(0|0);
    $142 = ($$pre$i2$i$i|0)==(0|0);
    do {
     if ($142) {
      label = 52;
     } else {
      $143 = ((($$pre$i2$i$i)) + 12|0);
      $144 = HEAP32[$143>>2]|0;
      $145 = ((($$pre$i2$i$i)) + 16|0);
      $146 = HEAP32[$145>>2]|0;
      $147 = ($144|0)==($146|0);
      if ($147) {
       $148 = HEAP32[$$pre$i2$i$i>>2]|0;
       $149 = ((($148)) + 36|0);
       $150 = HEAP32[$149>>2]|0;
       __THREW__ = 0;
       $151 = (invoke_ii($150|0,($$pre$i2$i$i|0))|0);
       $152 = __THREW__; __THREW__ = 0;
       $153 = $152&1;
       if ($153) {
        label = 28;
        break L8;
       }
       $154 = ($151|0)==(-1);
       if ($154) {
        HEAP32[$__e>>2] = 0;
        label = 52;
        break;
       }
      }
      if (!($140)) {
       label = 53;
      }
     }
    } while(0);
    if ((label|0) == 52) {
     if ($140) {
      label = 53;
     }
    }
    if ((label|0) == 53) {
     $155 = HEAP32[$__err>>2]|0;
     $156 = $155 | 2;
     HEAP32[$__err>>2] = $156;
    }
    $157 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($157|0);
   }
  }
 } while(0);
 if ((label|0) == 28) {
  $77 = ___cxa_find_matching_catch_2()|0;
  $78 = tempRet0;
  $lpad$phi$index = $77;$lpad$phi$index2 = $78;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__115__num_get_floatIeEET_PKcS3_Rj($__a,$__a_end,$__err) {
 $__a = $__a|0;
 $__a_end = $__a_end|0;
 $__err = $__err|0;
 var $$0 = 0.0, $$1 = 0.0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__p2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__p2 = sp;
 $0 = ($__a|0)==($__a_end|0);
 if ($0) {
  HEAP32[$__err>>2] = 4;
  $$1 = 0.0;
 } else {
  $1 = (___errno_location()|0);
  $2 = HEAP32[$1>>2]|0;
  HEAP32[$1>>2] = 0;
  $3 = (__ZNSt3__16__clocEv()|0);
  $4 = (+_strtold_l($__a,$__p2,$3));
  $5 = HEAP32[$1>>2]|0;
  $6 = ($5|0)==(0);
  if ($6) {
   HEAP32[$1>>2] = $2;
  }
  $7 = HEAP32[$__p2>>2]|0;
  $8 = ($7|0)==($__a_end|0);
  if ($8) {
   $9 = ($5|0)==(34);
   if ($9) {
    HEAP32[$__err>>2] = 4;
    $$0 = $4;
   } else {
    $$0 = $4;
   }
  } else {
   HEAP32[$__err>>2] = 4;
   $$0 = 0.0;
  }
  $$1 = $$0;
 }
 STACKTOP = sp;return (+$$1);
}
function __ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRPv($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0 = 0, $$0$i$i = 0, $$01 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i10 = 0, $$pre$i2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0;
 var $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__g = 0, $__g_end = 0, $__grouping = 0, $__i$0$i$i = 0, $__i$0$i$i5 = 0, $exitcond$i$i = 0, $exitcond$i$i6 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $phitmp$i = 0, $phitmp$i26 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $__atoms = sp + 208|0;
 $__grouping = sp + 196|0;
 $0 = sp + 192|0;
 $__buf = sp + 180|0;
 $__a_end = sp + 176|0;
 $__g = sp + 16|0;
 $__g_end = sp + 8|0;
 $__dc = sp + 4|0;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $1 = (($__grouping) + ($__i$0$i$i<<2)|0);
  HEAP32[$1>>2] = 0;
  $2 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $2;
 }
 __THREW__ = 0;
 $3 = (invoke_ii(3373,($__iob|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 do {
  if ($5) {
   $87 = ___cxa_find_matching_catch_2()|0;
   $88 = tempRet0;
   $$0 = $87;$$01 = $88;
  } else {
   HEAP32[$0>>2] = $3;
   __THREW__ = 0;
   $6 = (invoke_iii(3374,($0|0),(1207576|0))|0);
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if (!($8)) {
    $9 = HEAP32[$6>>2]|0;
    $10 = ((($9)) + 32|0);
    $11 = HEAP32[$10>>2]|0;
    __THREW__ = 0;
    (invoke_iiiii($11|0,($6|0),(620486|0),((620512)|0),($__atoms|0))|0);
    $12 = __THREW__; __THREW__ = 0;
    $13 = $12&1;
    if (!($13)) {
     (__ZNSt3__114__shared_count16__release_sharedEv($3)|0);
     $__i$0$i$i5 = 0;
     while(1) {
      $exitcond$i$i6 = ($__i$0$i$i5|0)==(3);
      if ($exitcond$i$i6) {
       break;
      }
      $14 = (($__buf) + ($__i$0$i$i5<<2)|0);
      HEAP32[$14>>2] = 0;
      $15 = (($__i$0$i$i5) + 1)|0;
      $__i$0$i$i5 = $15;
     }
     $16 = HEAP8[$__buf>>0]|0;
     $17 = $16 & 1;
     $18 = ($17<<24>>24)==(0);
     if ($18) {
      $21 = 10;
     } else {
      $19 = HEAP32[$__buf>>2]|0;
      $20 = $19 & -2;
      $phitmp$i = (($20) + -1)|0;
      $21 = $phitmp$i;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($21|0),0);
     $22 = __THREW__; __THREW__ = 0;
     $23 = $22&1;
     L18: do {
      if ($23) {
       label = 36;
      } else {
       $24 = HEAP8[$__buf>>0]|0;
       $25 = $24 & 1;
       $26 = ($25<<24>>24)==(0);
       $27 = ((($__buf)) + 8|0);
       $28 = HEAP32[$27>>2]|0;
       $29 = ((($__buf)) + 1|0);
       $30 = $26 ? $29 : $28;
       HEAP32[$__a_end>>2] = $30;
       HEAP32[$__g_end>>2] = $__g;
       HEAP32[$__dc>>2] = 0;
       $31 = ((($__buf)) + 4|0);
       $$pre = HEAP32[$__b>>2]|0;
       $$pre$i$i$i10 = $$pre;$__a$0 = $30;
       L20: while(1) {
        $32 = ($$pre$i$i$i10|0)==(0|0);
        if ($32) {
         $$pre$i$i$i = 0;
        } else {
         $33 = ((($$pre$i$i$i10)) + 12|0);
         $34 = HEAP32[$33>>2]|0;
         $35 = ((($$pre$i$i$i10)) + 16|0);
         $36 = HEAP32[$35>>2]|0;
         $37 = ($34|0)==($36|0);
         if ($37) {
          $38 = HEAP32[$$pre$i$i$i10>>2]|0;
          $39 = ((($38)) + 36|0);
          $40 = HEAP32[$39>>2]|0;
          __THREW__ = 0;
          $41 = (invoke_ii($40|0,($$pre$i$i$i10|0))|0);
          $42 = __THREW__; __THREW__ = 0;
          $43 = $42&1;
          if ($43) {
           label = 35;
           break;
          }
          $44 = ($41|0)==(-1);
          if ($44) {
           HEAP32[$__b>>2] = 0;
           $$pre$i$i$i = 0;
          } else {
           $$pre$i$i$i = $$pre$i$i$i10;
          }
         } else {
          $$pre$i$i$i = $$pre$i$i$i10;
         }
        }
        $45 = ($$pre$i$i$i|0)==(0|0);
        $46 = HEAP32[$__e>>2]|0;
        $47 = ($46|0)==(0|0);
        do {
         if ($47) {
          label = 25;
         } else {
          $48 = ((($46)) + 12|0);
          $49 = HEAP32[$48>>2]|0;
          $50 = ((($46)) + 16|0);
          $51 = HEAP32[$50>>2]|0;
          $52 = ($49|0)==($51|0);
          if (!($52)) {
           if ($45) {
            $168 = $46;
            break;
           } else {
            $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $46;$__a$2 = $__a$0;
            break L20;
           }
          }
          $53 = HEAP32[$46>>2]|0;
          $54 = ((($53)) + 36|0);
          $55 = HEAP32[$54>>2]|0;
          __THREW__ = 0;
          $56 = (invoke_ii($55|0,($46|0))|0);
          $57 = __THREW__; __THREW__ = 0;
          $58 = $57&1;
          if ($58) {
           label = 35;
           break L20;
          }
          $59 = ($56|0)==(-1);
          if ($59) {
           HEAP32[$__e>>2] = 0;
           label = 25;
           break;
          } else {
           if ($45) {
            $168 = $46;
            break;
           } else {
            $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $46;$__a$2 = $__a$0;
            break L20;
           }
          }
         }
        } while(0);
        if ((label|0) == 25) {
         label = 0;
         if ($45) {
          $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = 0;$__a$2 = $__a$0;
          break;
         } else {
          $168 = 0;
         }
        }
        $60 = HEAP32[$__a_end>>2]|0;
        $61 = HEAP8[$__buf>>0]|0;
        $62 = $61 & 1;
        $63 = ($62<<24>>24)==(0);
        $64 = HEAP32[$31>>2]|0;
        $65 = $61&255;
        $66 = $65 >>> 1;
        $67 = $63 ? $66 : $64;
        $68 = (($__a$0) + ($67)|0);
        $69 = ($60|0)==($68|0);
        if ($69) {
         $70 = $67 << 1;
         __THREW__ = 0;
         invoke_viii(3371,($__buf|0),($70|0),0);
         $71 = __THREW__; __THREW__ = 0;
         $72 = $71&1;
         if ($72) {
          label = 35;
          break;
         }
         $73 = HEAP8[$__buf>>0]|0;
         $74 = $73 & 1;
         $75 = ($74<<24>>24)==(0);
         if ($75) {
          $78 = 10;
         } else {
          $76 = HEAP32[$__buf>>2]|0;
          $77 = $76 & -2;
          $phitmp$i26 = (($77) + -1)|0;
          $78 = $phitmp$i26;
         }
         __THREW__ = 0;
         invoke_viii(3371,($__buf|0),($78|0),0);
         $79 = __THREW__; __THREW__ = 0;
         $80 = $79&1;
         if ($80) {
          label = 35;
          break;
         }
         $81 = HEAP8[$__buf>>0]|0;
         $82 = $81 & 1;
         $83 = ($82<<24>>24)==(0);
         $84 = HEAP32[$27>>2]|0;
         $85 = $83 ? $29 : $84;
         $86 = (($85) + ($67)|0);
         HEAP32[$__a_end>>2] = $86;
         $__a$1 = $85;
        } else {
         $__a$1 = $__a$0;
        }
        $95 = ((($$pre$i$i$i)) + 12|0);
        $96 = HEAP32[$95>>2]|0;
        $97 = ((($$pre$i$i$i)) + 16|0);
        $98 = HEAP32[$97>>2]|0;
        $99 = ($96|0)==($98|0);
        if ($99) {
         $100 = HEAP32[$$pre$i$i$i>>2]|0;
         $101 = ((($100)) + 36|0);
         $102 = HEAP32[$101>>2]|0;
         __THREW__ = 0;
         $103 = (invoke_ii($102|0,($$pre$i$i$i|0))|0);
         $104 = __THREW__; __THREW__ = 0;
         $105 = $104&1;
         if ($105) {
          label = 35;
          break;
         } else {
          $$0$i$i = $103;
         }
        } else {
         $106 = HEAP8[$96>>0]|0;
         $107 = $106&255;
         $$0$i$i = $107;
        }
        $108 = $$0$i$i&255;
        $109 = (__ZNSt3__19__num_getIcE17__stage2_int_loopEciPcRS2_RjcRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_S2_($108,16,$__a$1,$__a_end,$__dc,0,$__grouping,$__g,$__g_end,$__atoms)|0);
        $110 = ($109|0)==(0);
        if (!($110)) {
         $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i2$i$i = $168;$__a$2 = $__a$1;
         break;
        }
        $111 = HEAP32[$95>>2]|0;
        $112 = HEAP32[$97>>2]|0;
        $113 = ($111|0)==($112|0);
        if ($113) {
         $114 = HEAP32[$$pre$i$i$i>>2]|0;
         $115 = ((($114)) + 40|0);
         $116 = HEAP32[$115>>2]|0;
         __THREW__ = 0;
         (invoke_ii($116|0,($$pre$i$i$i|0))|0);
         $117 = __THREW__; __THREW__ = 0;
         $118 = $117&1;
         if ($118) {
          label = 35;
          break;
         } else {
          $$pre$i$i$i10 = $$pre$i$i$i;$__a$0 = $__a$1;
          continue;
         }
        } else {
         $119 = ((($111)) + 1|0);
         HEAP32[$95>>2] = $119;
         $$pre$i$i$i10 = $$pre$i$i$i;$__a$0 = $__a$1;
         continue;
        }
       }
       if ((label|0) == 35) {
        $91 = ___cxa_find_matching_catch_2()|0;
        $92 = tempRet0;
        $lpad$phi$index = $91;$lpad$phi$index2 = $92;
        break;
       }
       $120 = HEAP32[$__a_end>>2]|0;
       $121 = $__a$2;
       $122 = (($120) - ($121))|0;
       __THREW__ = 0;
       invoke_viii(3371,($__buf|0),($122|0),0);
       $123 = __THREW__; __THREW__ = 0;
       $124 = $123&1;
       if ($124) {
        label = 36;
       } else {
        $125 = HEAP8[$__buf>>0]|0;
        $126 = HEAP32[$27>>2]|0;
        __THREW__ = 0;
        $127 = (invoke_i(4874)|0);
        $128 = __THREW__; __THREW__ = 0;
        $129 = $128&1;
        if ($129) {
         label = 36;
        } else {
         $130 = $125 & 1;
         $131 = ($130<<24>>24)==(0);
         $132 = $131 ? $29 : $126;
         __THREW__ = 0;
         HEAP32[$vararg_buffer>>2] = $__v;
         $133 = (invoke_iiiii(4875,($132|0),($127|0),(620521|0),($vararg_buffer|0))|0);
         $134 = __THREW__; __THREW__ = 0;
         $135 = $134&1;
         if ($135) {
          label = 36;
         } else {
          $136 = ($133|0)==(1);
          if (!($136)) {
           HEAP32[$__err>>2] = 4;
          }
          $137 = ($$pre$i$i$i$lcssa|0)==(0|0);
          if ($137) {
           $151 = 0;
          } else {
           $138 = ((($$pre$i$i$i$lcssa)) + 12|0);
           $139 = HEAP32[$138>>2]|0;
           $140 = ((($$pre$i$i$i$lcssa)) + 16|0);
           $141 = HEAP32[$140>>2]|0;
           $142 = ($139|0)==($141|0);
           if ($142) {
            $143 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
            $144 = ((($143)) + 36|0);
            $145 = HEAP32[$144>>2]|0;
            __THREW__ = 0;
            $146 = (invoke_ii($145|0,($$pre$i$i$i$lcssa|0))|0);
            $147 = __THREW__; __THREW__ = 0;
            $148 = $147&1;
            if ($148) {
             label = 36;
             break;
            }
            $149 = ($146|0)==(-1);
            if ($149) {
             HEAP32[$__b>>2] = 0;
             $151 = 0;
            } else {
             $151 = $$pre$i$i$i$lcssa;
            }
           } else {
            $151 = $$pre$i$i$i$lcssa;
           }
          }
          $150 = ($151|0)==(0|0);
          $152 = ($$pre$i2$i$i|0)==(0|0);
          do {
           if ($152) {
            label = 61;
           } else {
            $153 = ((($$pre$i2$i$i)) + 12|0);
            $154 = HEAP32[$153>>2]|0;
            $155 = ((($$pre$i2$i$i)) + 16|0);
            $156 = HEAP32[$155>>2]|0;
            $157 = ($154|0)==($156|0);
            if ($157) {
             $158 = HEAP32[$$pre$i2$i$i>>2]|0;
             $159 = ((($158)) + 36|0);
             $160 = HEAP32[$159>>2]|0;
             __THREW__ = 0;
             $161 = (invoke_ii($160|0,($$pre$i2$i$i|0))|0);
             $162 = __THREW__; __THREW__ = 0;
             $163 = $162&1;
             if ($163) {
              label = 36;
              break L18;
             }
             $164 = ($161|0)==(-1);
             if ($164) {
              HEAP32[$__e>>2] = 0;
              label = 61;
              break;
             }
            }
            if (!($150)) {
             label = 62;
            }
           }
          } while(0);
          if ((label|0) == 61) {
           if ($150) {
            label = 62;
           }
          }
          if ((label|0) == 62) {
           $165 = HEAP32[$__err>>2]|0;
           $166 = $165 | 2;
           HEAP32[$__err>>2] = $166;
          }
          $167 = HEAP32[$__b>>2]|0;
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
          STACKTOP = sp;return ($167|0);
         }
        }
       }
      }
     } while(0);
     if ((label|0) == 36) {
      $93 = ___cxa_find_matching_catch_2()|0;
      $94 = tempRet0;
      $lpad$phi$index = $93;$lpad$phi$index2 = $94;
     }
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
     $$0 = $lpad$phi$index;$$01 = $lpad$phi$index2;
     break;
    }
   }
   $89 = ___cxa_find_matching_catch_2()|0;
   $90 = tempRet0;
   (__ZNSt3__114__shared_count16__release_sharedEv($3)|0);
   $$0 = $89;$$01 = $90;
  }
 } while(0);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($$0|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__110__sscanf_lEPKcP15__locale_structS1_z($__s,$__l,$__format,$varargs) {
 $__s = $__s|0;
 $__l = $__l|0;
 $__format = $__format|0;
 $varargs = $varargs|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $__va = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__va = sp;
 HEAP32[$__va>>2] = $varargs;
 $0 = (_uselocale($__l)|0);
 $1 = (_vsscanf($__s,$__format,$__va)|0);
 $2 = ($0|0)==(0|0);
 if (!($2)) {
  __THREW__ = 0;
  (invoke_ii(4876,($0|0))|0);
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if ($4) {
   $5 = ___cxa_find_matching_catch_3(0|0)|0;
   $6 = tempRet0;
   ___clang_call_terminate($5);
   // unreachable;
  }
 }
 STACKTOP = sp;return ($1|0);
}
function __ZNSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRb($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0 = 0, $$02 = 0, $$03 = 0, $$byval_copy = 0, $$byval_copy1 = 0, $$byval_copy2 = 0, $$sroa$01$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__lv = 0, $__names = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy2 = sp + 60|0;
 $$byval_copy1 = sp + 56|0;
 $$byval_copy = sp + 52|0;
 $__lv = sp + 48|0;
 $0 = sp + 44|0;
 $1 = sp + 40|0;
 $2 = sp + 36|0;
 $3 = sp + 32|0;
 $__names = sp + 8|0;
 $4 = sp;
 $5 = ((($__iob)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $6 & 1;
 $8 = ($7|0)==(0);
 L1: do {
  if ($8) {
   HEAP32[$__lv>>2] = -1;
   $9 = HEAP32[$this>>2]|0;
   $10 = ((($9)) + 16|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = HEAP32[$__b>>2]|0;
   HEAP32[$0>>2] = $12;
   $13 = HEAP32[$__e>>2]|0;
   HEAP32[$1>>2] = $13;
   ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
   ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
   $14 = (FUNCTION_TABLE_iiiiiii[$11 & 8191]($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__lv)|0);
   HEAP32[$__b>>2] = $14;
   $15 = HEAP32[$__lv>>2]|0;
   switch ($15|0) {
   case 0:  {
    HEAP8[$__v>>0] = 0;
    break;
   }
   case 1:  {
    HEAP8[$__v>>0] = 1;
    break;
   }
   default: {
    HEAP8[$__v>>0] = 1;
    HEAP32[$__err>>2] = 4;
   }
   }
   $$sroa$01$0 = $14;
  } else {
   $16 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
   HEAP32[$2>>2] = $16;
   __THREW__ = 0;
   $17 = (invoke_iii(3374,($2|0),(1207608|0))|0);
   $18 = __THREW__; __THREW__ = 0;
   $19 = $18&1;
   do {
    if ($19) {
     $44 = ___cxa_find_matching_catch_2()|0;
     $45 = tempRet0;
     (__ZNSt3__114__shared_count16__release_sharedEv($16)|0);
     $$02 = $44;$$03 = $45;
    } else {
     (__ZNSt3__114__shared_count16__release_sharedEv($16)|0);
     $20 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
     HEAP32[$3>>2] = $20;
     __THREW__ = 0;
     $21 = (invoke_iii(3374,($3|0),(1207616|0))|0);
     $22 = __THREW__; __THREW__ = 0;
     $23 = $22&1;
     if ($23) {
      $46 = ___cxa_find_matching_catch_2()|0;
      $47 = tempRet0;
      (__ZNSt3__114__shared_count16__release_sharedEv($20)|0);
      $$02 = $46;$$03 = $47;
      break;
     }
     (__ZNSt3__114__shared_count16__release_sharedEv($20)|0);
     $24 = HEAP32[$21>>2]|0;
     $25 = ((($24)) + 24|0);
     $26 = HEAP32[$25>>2]|0;
     __THREW__ = 0;
     invoke_vii($26|0,($__names|0),($21|0));
     $27 = __THREW__; __THREW__ = 0;
     $28 = $27&1;
     if ($28) {
      $$0 = $__names;
     } else {
      $29 = ((($__names)) + 12|0);
      $30 = HEAP32[$21>>2]|0;
      $31 = ((($30)) + 28|0);
      $32 = HEAP32[$31>>2]|0;
      __THREW__ = 0;
      invoke_vii($32|0,($29|0),($21|0));
      $33 = __THREW__; __THREW__ = 0;
      $34 = $33&1;
      if ($34) {
       $$0 = $29;
      } else {
       $35 = HEAP32[$__e>>2]|0;
       HEAP32[$4>>2] = $35;
       $36 = ((($__names)) + 24|0);
       __THREW__ = 0;
       ;HEAP32[$$byval_copy2>>2]=HEAP32[$4>>2]|0;
       $37 = (invoke_iiiiiiii(4877,($__b|0),($$byval_copy2|0),($__names|0),($36|0),($17|0),($__err|0),1)|0);
       $38 = __THREW__; __THREW__ = 0;
       $39 = $38&1;
       if ($39) {
        $54 = ___cxa_find_matching_catch_2()|0;
        $55 = tempRet0;
        $56 = ((($__names)) + 12|0);
        __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($56);
        __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__names);
        $$02 = $54;$$03 = $55;
        break;
       } else {
        $40 = ($37|0)==($__names|0);
        $41 = $40&1;
        HEAP8[$__v>>0] = $41;
        $42 = HEAP32[$__b>>2]|0;
        $43 = ((($__names)) + 12|0);
        __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($43);
        __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__names);
        $$sroa$01$0 = $42;
        break L1;
       }
      }
     }
     $48 = ___cxa_find_matching_catch_2()|0;
     $49 = tempRet0;
     $50 = ($__names|0)==($$0|0);
     if ($50) {
      $$02 = $48;$$03 = $49;
     } else {
      $52 = $$0;
      while(1) {
       $51 = ((($52)) + -12|0);
       __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($51);
       $53 = ($51|0)==($__names|0);
       if ($53) {
        $$02 = $48;$$03 = $49;
        break;
       } else {
        $52 = $51;
       }
      }
     }
    }
   } while(0);
   ___resumeException($$02|0);
   // unreachable;
  }
 } while(0);
 STACKTOP = sp;return ($$sroa$01$0|0);
}
function __ZNSt3__114__scan_keywordINS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEPKNS_12basic_stringIwS3_NS_9allocatorIwEEEENS_5ctypeIwEEEET0_RT_SE_SD_SD_RKT1_Rjb($__b,$__e,$__kb,$__ke,$__ct,$__err,$__case_sensitive) {
 $__b = $__b|0;
 $__e = $__e|0;
 $__kb = $__kb|0;
 $__ke = $__ke|0;
 $__ct = $__ct|0;
 $__err = $__err|0;
 $__case_sensitive = $__case_sensitive|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i6 = 0, $$0$i$i1$i$i = 0, $$0$i$i1$i$i12 = 0, $$045 = 0, $$lcssa = 0, $$pre$i$i$i = 0, $$pre$i$i$i8 = 0, $$pre$i3$i$i14 = 0, $$pre$i3$i$i14$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $__c$0 = 0, $__consume$0$off0 = 0, $__consume$2$off0 = 0, $__indx$0 = 0, $__kc$0 = 0, $__ky$0 = 0, $__ky1$0 = 0, $__ky2$0 = 0, $__n_does_match$0 = 0, $__n_does_match$0$lcssa = 0, $__n_does_match$1 = 0, $__n_does_match$2 = 0, $__n_does_match$3 = 0, $__n_does_match$5 = 0, $__n_does_match$6 = 0, $__n_does_match$7 = 0, $__n_might_match$0 = 0, $__n_might_match$0$lcssa = 0, $__n_might_match$1 = 0, $__n_might_match$2 = 0;
 var $__n_might_match$3 = 0, $__n_might_match$5 = 0, $__st$0 = 0, $__st$1 = 0, $__st$2 = 0, $__st$3 = 0, $__stat_hold$sroa$0$0 = 0, $__stat_hold$sroa$0$0$ph$ph = 0, $__stat_hold$sroa$0$1 = 0, $__statbuf = 0, $__status$0 = 0, $lpad$phi$index = 0, $lpad$phi$index3 = 0, $phitmp$i$i$i = 0, $phitmp$i$i$i9 = 0, $tmp$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__statbuf = sp;
 $0 = $__ke;
 $1 = $__kb;
 $2 = (($0) - ($1))|0;
 $3 = (($2|0) / 12)&-1;
 $4 = ($3>>>0)>(100);
 do {
  if ($4) {
   $5 = (_malloc($3)|0);
   $6 = ($5|0)==(0|0);
   if ($6) {
    __THREW__ = 0;
    invoke_v(4863);
    $7 = __THREW__; __THREW__ = 0;
    $8 = $7&1;
    if ($8) {
     $__stat_hold$sroa$0$0$ph$ph = 0;
     label = 6;
     break;
    }
   }
   $17 = $5;
   $__stat_hold$sroa$0$1 = $17;$__status$0 = $5;
   label = 11;
  } else {
   $__stat_hold$sroa$0$1 = 0;$__status$0 = $__statbuf;
   label = 11;
  }
 } while(0);
 L6: do {
  if ((label|0) == 11) {
   $__ky$0 = $__kb;$__n_does_match$0 = 0;$__n_might_match$0 = $3;$__st$0 = $__status$0;
   while(1) {
    $18 = ($__ky$0|0)==($__ke|0);
    if ($18) {
     $__n_does_match$0$lcssa = $__n_does_match$0;$__n_might_match$0$lcssa = $__n_might_match$0;
     break;
    }
    $19 = HEAP8[$__ky$0>>0]|0;
    $20 = $19 & 1;
    $21 = ($20<<24>>24)==(0);
    if ($21) {
     $24 = $19&255;
     $25 = $24 >>> 1;
     $27 = $25;
    } else {
     $22 = ((($__ky$0)) + 4|0);
     $23 = HEAP32[$22>>2]|0;
     $27 = $23;
    }
    $26 = ($27|0)==(0);
    if ($26) {
     HEAP8[$__st$0>>0] = 2;
     $28 = (($__n_might_match$0) + -1)|0;
     $29 = (($__n_does_match$0) + 1)|0;
     $__n_does_match$1 = $29;$__n_might_match$1 = $28;
    } else {
     HEAP8[$__st$0>>0] = 1;
     $__n_does_match$1 = $__n_does_match$0;$__n_might_match$1 = $__n_might_match$0;
    }
    $30 = ((($__ky$0)) + 12|0);
    $31 = ((($__st$0)) + 1|0);
    $__ky$0 = $30;$__n_does_match$0 = $__n_does_match$1;$__n_might_match$0 = $__n_might_match$1;$__st$0 = $31;
   }
   $__indx$0 = 0;$__n_does_match$2 = $__n_does_match$0$lcssa;$__n_might_match$2 = $__n_might_match$0$lcssa;
   L20: while(1) {
    $32 = HEAP32[$__b>>2]|0;
    $33 = ($32|0)==(0|0);
    do {
     if ($33) {
      $62 = 1;
     } else {
      $34 = ((($32)) + 12|0);
      $35 = HEAP32[$34>>2]|0;
      $36 = ((($32)) + 16|0);
      $37 = HEAP32[$36>>2]|0;
      $38 = ($35|0)==($37|0);
      if ($38) {
       $39 = HEAP32[$32>>2]|0;
       $40 = ((($39)) + 36|0);
       $41 = HEAP32[$40>>2]|0;
       __THREW__ = 0;
       $42 = (invoke_ii($41|0,($32|0))|0);
       $43 = __THREW__; __THREW__ = 0;
       $44 = $43&1;
       if ($44) {
        label = 5;
        break L20;
       } else {
        $$0$i$i$i$i = $42;
       }
      } else {
       $45 = HEAP32[$35>>2]|0;
       $$0$i$i$i$i = $45;
      }
      $46 = ($$0$i$i$i$i|0)==(-1);
      if ($46) {
       HEAP32[$__b>>2] = 0;
       $62 = 1;
       break;
      } else {
       $$pre$i$i$i = HEAP32[$__b>>2]|0;
       $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
       $62 = $phitmp$i$i$i;
       break;
      }
     }
    } while(0);
    $47 = HEAP32[$__e>>2]|0;
    $48 = ($47|0)==(0|0);
    if ($48) {
     $$pre$i3$i$i14 = 0;$63 = 1;
    } else {
     $49 = ((($47)) + 12|0);
     $50 = HEAP32[$49>>2]|0;
     $51 = ((($47)) + 16|0);
     $52 = HEAP32[$51>>2]|0;
     $53 = ($50|0)==($52|0);
     if ($53) {
      $54 = HEAP32[$47>>2]|0;
      $55 = ((($54)) + 36|0);
      $56 = HEAP32[$55>>2]|0;
      __THREW__ = 0;
      $57 = (invoke_ii($56|0,($47|0))|0);
      $58 = __THREW__; __THREW__ = 0;
      $59 = $58&1;
      if ($59) {
       label = 5;
       break;
      } else {
       $$0$i$i1$i$i = $57;
      }
     } else {
      $60 = HEAP32[$50>>2]|0;
      $$0$i$i1$i$i = $60;
     }
     $61 = ($$0$i$i1$i$i|0)==(-1);
     if ($61) {
      HEAP32[$__e>>2] = 0;
      $$pre$i3$i$i14 = 0;$63 = 1;
     } else {
      $$pre$i3$i$i14 = $47;$63 = 0;
     }
    }
    $tmp$i$i = $62 ^ $63;
    $64 = ($__n_might_match$2|0)!=(0);
    $65 = $64 & $tmp$i$i;
    $66 = HEAP32[$__b>>2]|0;
    if (!($65)) {
     $$lcssa = $66;$$pre$i3$i$i14$lcssa = $$pre$i3$i$i14;
     label = 35;
     break;
    }
    $95 = ((($66)) + 12|0);
    $96 = HEAP32[$95>>2]|0;
    $97 = ((($66)) + 16|0);
    $98 = HEAP32[$97>>2]|0;
    $99 = ($96|0)==($98|0);
    if ($99) {
     $100 = HEAP32[$66>>2]|0;
     $101 = ((($100)) + 36|0);
     $102 = HEAP32[$101>>2]|0;
     __THREW__ = 0;
     $103 = (invoke_ii($102|0,($66|0))|0);
     $104 = __THREW__; __THREW__ = 0;
     $105 = $104&1;
     if ($105) {
      label = 5;
      break;
     } else {
      $$0$i$i = $103;
     }
    } else {
     $106 = HEAP32[$96>>2]|0;
     $$0$i$i = $106;
    }
    if ($__case_sensitive) {
     $__c$0 = $$0$i$i;
    } else {
     $107 = HEAP32[$__ct>>2]|0;
     $108 = ((($107)) + 28|0);
     $109 = HEAP32[$108>>2]|0;
     __THREW__ = 0;
     $110 = (invoke_iii($109|0,($__ct|0),($$0$i$i|0))|0);
     $111 = __THREW__; __THREW__ = 0;
     $112 = $111&1;
     if ($112) {
      label = 5;
      break;
     } else {
      $__c$0 = $110;
     }
    }
    $113 = (($__indx$0) + 1)|0;
    $__consume$0$off0 = 0;$__ky1$0 = $__kb;$__n_does_match$3 = $__n_does_match$2;$__n_might_match$3 = $__n_might_match$2;$__st$1 = $__status$0;
    while(1) {
     $114 = ($__ky1$0|0)==($__ke|0);
     if ($114) {
      break;
     }
     $115 = HEAP8[$__st$1>>0]|0;
     $116 = ($115<<24>>24)==(1);
     do {
      if ($116) {
       $117 = HEAP8[$__ky1$0>>0]|0;
       $118 = $117 & 1;
       $119 = ($118<<24>>24)==(0);
       if ($119) {
        $122 = ((($__ky1$0)) + 4|0);
        $124 = $122;
       } else {
        $120 = ((($__ky1$0)) + 8|0);
        $121 = HEAP32[$120>>2]|0;
        $124 = $121;
       }
       $123 = (($124) + ($__indx$0<<2)|0);
       $125 = HEAP32[$123>>2]|0;
       if ($__case_sensitive) {
        $__kc$0 = $125;
       } else {
        $126 = HEAP32[$__ct>>2]|0;
        $127 = ((($126)) + 28|0);
        $128 = HEAP32[$127>>2]|0;
        __THREW__ = 0;
        $129 = (invoke_iii($128|0,($__ct|0),($125|0))|0);
        $130 = __THREW__; __THREW__ = 0;
        $131 = $130&1;
        if ($131) {
         label = 4;
         break L20;
        } else {
         $__kc$0 = $129;
        }
       }
       $132 = ($__c$0|0)==($__kc$0|0);
       if (!($132)) {
        HEAP8[$__st$1>>0] = 0;
        $144 = (($__n_might_match$3) + -1)|0;
        $__consume$2$off0 = $__consume$0$off0;$__n_does_match$5 = $__n_does_match$3;$__n_might_match$5 = $144;
        break;
       }
       $133 = HEAP8[$__ky1$0>>0]|0;
       $134 = $133 & 1;
       $135 = ($134<<24>>24)==(0);
       if ($135) {
        $138 = $133&255;
        $139 = $138 >>> 1;
        $141 = $139;
       } else {
        $136 = ((($__ky1$0)) + 4|0);
        $137 = HEAP32[$136>>2]|0;
        $141 = $137;
       }
       $140 = ($141|0)==($113|0);
       if ($140) {
        HEAP8[$__st$1>>0] = 2;
        $142 = (($__n_might_match$3) + -1)|0;
        $143 = (($__n_does_match$3) + 1)|0;
        $__consume$2$off0 = 1;$__n_does_match$5 = $143;$__n_might_match$5 = $142;
       } else {
        $__consume$2$off0 = 1;$__n_does_match$5 = $__n_does_match$3;$__n_might_match$5 = $__n_might_match$3;
       }
      } else {
       $__consume$2$off0 = $__consume$0$off0;$__n_does_match$5 = $__n_does_match$3;$__n_might_match$5 = $__n_might_match$3;
      }
     } while(0);
     $145 = ((($__ky1$0)) + 12|0);
     $146 = ((($__st$1)) + 1|0);
     $__consume$0$off0 = $__consume$2$off0;$__ky1$0 = $145;$__n_does_match$3 = $__n_does_match$5;$__n_might_match$3 = $__n_might_match$5;$__st$1 = $146;
    }
    if (!($__consume$0$off0)) {
     $__indx$0 = $113;$__n_does_match$2 = $__n_does_match$3;$__n_might_match$2 = $__n_might_match$3;
     continue;
    }
    $147 = HEAP32[$__b>>2]|0;
    $148 = ((($147)) + 12|0);
    $149 = HEAP32[$148>>2]|0;
    $150 = ((($147)) + 16|0);
    $151 = HEAP32[$150>>2]|0;
    $152 = ($149|0)==($151|0);
    if ($152) {
     $153 = HEAP32[$147>>2]|0;
     $154 = ((($153)) + 40|0);
     $155 = HEAP32[$154>>2]|0;
     __THREW__ = 0;
     (invoke_ii($155|0,($147|0))|0);
     $156 = __THREW__; __THREW__ = 0;
     $157 = $156&1;
     if ($157) {
      label = 5;
      break;
     }
    } else {
     $158 = ((($149)) + 4|0);
     HEAP32[$148>>2] = $158;
    }
    $159 = (($__n_does_match$3) + ($__n_might_match$3))|0;
    $160 = ($159>>>0)>(1);
    if ($160) {
     $__ky2$0 = $__kb;$__n_does_match$6 = $__n_does_match$3;$__st$2 = $__status$0;
    } else {
     $__indx$0 = $113;$__n_does_match$2 = $__n_does_match$3;$__n_might_match$2 = $__n_might_match$3;
     continue;
    }
    while(1) {
     $161 = ($__ky2$0|0)==($__ke|0);
     if ($161) {
      $__indx$0 = $113;$__n_does_match$2 = $__n_does_match$6;$__n_might_match$2 = $__n_might_match$3;
      continue L20;
     }
     $162 = HEAP8[$__st$2>>0]|0;
     $163 = ($162<<24>>24)==(2);
     if ($163) {
      $164 = HEAP8[$__ky2$0>>0]|0;
      $165 = $164 & 1;
      $166 = ($165<<24>>24)==(0);
      if ($166) {
       $169 = $164&255;
       $170 = $169 >>> 1;
       $172 = $170;
      } else {
       $167 = ((($__ky2$0)) + 4|0);
       $168 = HEAP32[$167>>2]|0;
       $172 = $168;
      }
      $171 = ($172|0)==($113|0);
      if ($171) {
       $__n_does_match$7 = $__n_does_match$6;
      } else {
       HEAP8[$__st$2>>0] = 0;
       $173 = (($__n_does_match$6) + -1)|0;
       $__n_does_match$7 = $173;
      }
     } else {
      $__n_does_match$7 = $__n_does_match$6;
     }
     $174 = ((($__ky2$0)) + 12|0);
     $175 = ((($__st$2)) + 1|0);
     $__ky2$0 = $174;$__n_does_match$6 = $__n_does_match$7;$__st$2 = $175;
    }
   }
   if ((label|0) == 4) {
    $9 = ___cxa_find_matching_catch_2()|0;
    $10 = tempRet0;
    $__stat_hold$sroa$0$0 = $__stat_hold$sroa$0$1;$lpad$phi$index = $9;$lpad$phi$index3 = $10;
    break;
   }
   else if ((label|0) == 5) {
    $11 = ___cxa_find_matching_catch_2()|0;
    $12 = tempRet0;
    $__stat_hold$sroa$0$0 = $__stat_hold$sroa$0$1;$lpad$phi$index = $11;$lpad$phi$index3 = $12;
    break;
   }
   else if ((label|0) == 35) {
    $67 = ($$lcssa|0)==(0|0);
    do {
     if ($67) {
      $187 = 1;
     } else {
      $68 = ((($$lcssa)) + 12|0);
      $69 = HEAP32[$68>>2]|0;
      $70 = ((($$lcssa)) + 16|0);
      $71 = HEAP32[$70>>2]|0;
      $72 = ($69|0)==($71|0);
      if ($72) {
       $73 = HEAP32[$$lcssa>>2]|0;
       $74 = ((($73)) + 36|0);
       $75 = HEAP32[$74>>2]|0;
       __THREW__ = 0;
       $76 = (invoke_ii($75|0,($$lcssa|0))|0);
       $77 = __THREW__; __THREW__ = 0;
       $78 = $77&1;
       if ($78) {
        $__stat_hold$sroa$0$0$ph$ph = $__stat_hold$sroa$0$1;
        label = 6;
        break L6;
       } else {
        $$0$i$i$i$i6 = $76;
       }
      } else {
       $79 = HEAP32[$69>>2]|0;
       $$0$i$i$i$i6 = $79;
      }
      $80 = ($$0$i$i$i$i6|0)==(-1);
      if ($80) {
       HEAP32[$__b>>2] = 0;
       $187 = 1;
       break;
      } else {
       $$pre$i$i$i8 = HEAP32[$__b>>2]|0;
       $phitmp$i$i$i9 = ($$pre$i$i$i8|0)==(0|0);
       $187 = $phitmp$i$i$i9;
       break;
      }
     }
    } while(0);
    $81 = ($$pre$i3$i$i14$lcssa|0)==(0|0);
    do {
     if ($81) {
      label = 82;
     } else {
      $82 = ((($$pre$i3$i$i14$lcssa)) + 12|0);
      $83 = HEAP32[$82>>2]|0;
      $84 = ((($$pre$i3$i$i14$lcssa)) + 16|0);
      $85 = HEAP32[$84>>2]|0;
      $86 = ($83|0)==($85|0);
      if ($86) {
       $87 = HEAP32[$$pre$i3$i$i14$lcssa>>2]|0;
       $88 = ((($87)) + 36|0);
       $89 = HEAP32[$88>>2]|0;
       __THREW__ = 0;
       $90 = (invoke_ii($89|0,($$pre$i3$i$i14$lcssa|0))|0);
       $91 = __THREW__; __THREW__ = 0;
       $92 = $91&1;
       if ($92) {
        $__stat_hold$sroa$0$0$ph$ph = $__stat_hold$sroa$0$1;
        label = 6;
        break L6;
       } else {
        $$0$i$i1$i$i12 = $90;
       }
      } else {
       $93 = HEAP32[$83>>2]|0;
       $$0$i$i1$i$i12 = $93;
      }
      $94 = ($$0$i$i1$i$i12|0)==(-1);
      if ($94) {
       HEAP32[$__e>>2] = 0;
       label = 82;
       break;
      } else {
       if ($187) {
        $$0 = $__kb;$__st$3 = $__status$0;
        break;
       } else {
        label = 84;
        break;
       }
      }
     }
    } while(0);
    if ((label|0) == 82) {
     if ($187) {
      label = 84;
     } else {
      $$0 = $__kb;$__st$3 = $__status$0;
     }
    }
    if ((label|0) == 84) {
     $176 = HEAP32[$__err>>2]|0;
     $177 = $176 | 2;
     HEAP32[$__err>>2] = $177;
     $$0 = $__kb;$__st$3 = $__status$0;
    }
    while(1) {
     $178 = ($$0|0)==($__ke|0);
     if ($178) {
      label = 88;
      break;
     }
     $179 = HEAP8[$__st$3>>0]|0;
     $180 = ($179<<24>>24)==(2);
     if ($180) {
      $$045 = $$0;
      break;
     }
     $181 = ((($$0)) + 12|0);
     $182 = ((($__st$3)) + 1|0);
     $$0 = $181;$__st$3 = $182;
    }
    if ((label|0) == 88) {
     $183 = HEAP32[$__err>>2]|0;
     $184 = $183 | 4;
     HEAP32[$__err>>2] = $184;
     $$045 = $__ke;
    }
    $185 = ($__stat_hold$sroa$0$1|0)==(0);
    if (!($185)) {
     $186 = $__stat_hold$sroa$0$1;
     _free($186);
    }
    STACKTOP = sp;return ($$045|0);
   }
  }
 } while(0);
 if ((label|0) == 6) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  $__stat_hold$sroa$0$0 = $__stat_hold$sroa$0$0$ph$ph;$lpad$phi$index = $13;$lpad$phi$index3 = $14;
 }
 $15 = ($__stat_hold$sroa$0$0|0)==(0);
 if (!($15)) {
  $16 = $__stat_hold$sroa$0$0;
  _free($16);
 }
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRl($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE15__do_get_signedIlEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE15__do_get_signedIlEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i3 = 0, $$0$i$i1$i$i = 0, $$0$i$i1$i$i9 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i5 = 0, $$pre$i3$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__g = 0, $__g_end = 0;
 var $__grouping = 0, $__i$0$i$i = 0, $__thousands_sep = 0, $exitcond$i$i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $phitmp$i = 0, $phitmp$i23 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 200|0;
 $__thousands_sep = sp + 196|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $0 = ((($__iob)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 74;
 switch ($2|0) {
 case 64:  {
  $$0$i = 8;
  break;
 }
 case 8:  {
  $$0$i = 16;
  break;
 }
 case 0:  {
  $$0$i = 0;
  break;
 }
 default: {
  $$0$i = 10;
 }
 }
 __ZNSt3__19__num_getIwE17__stage2_int_prepERNS_8ios_baseEPwRw($__grouping,$__iob,$__atoms,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $3 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$3>>2] = 0;
  $4 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $4;
 }
 $5 = HEAP8[$__buf>>0]|0;
 $6 = $5 & 1;
 $7 = ($6<<24>>24)==(0);
 if ($7) {
  $10 = 10;
 } else {
  $8 = HEAP32[$__buf>>2]|0;
  $9 = $8 & -2;
  $phitmp$i = (($9) + -1)|0;
  $10 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($10|0),0);
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 L13: do {
  if ($12) {
   label = 33;
  } else {
   $13 = HEAP8[$__buf>>0]|0;
   $14 = $13 & 1;
   $15 = ($14<<24>>24)==(0);
   $16 = ((($__buf)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($__buf)) + 1|0);
   $19 = $15 ? $18 : $17;
   HEAP32[$__a_end>>2] = $19;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   $20 = ((($__buf)) + 4|0);
   $21 = HEAP32[$__thousands_sep>>2]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i5 = $$pre;$__a$0 = $19;
   L15: while(1) {
    $22 = ($$pre$i$i$i5|0)==(0|0);
    if ($22) {
     $$pre$i$i$i = 0;$157 = 1;
    } else {
     $23 = ((($$pre$i$i$i5)) + 12|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ((($$pre$i$i$i5)) + 16|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = ($24|0)==($26|0);
     if ($27) {
      $28 = HEAP32[$$pre$i$i$i5>>2]|0;
      $29 = ((($28)) + 36|0);
      $30 = HEAP32[$29>>2]|0;
      __THREW__ = 0;
      $31 = (invoke_ii($30|0,($$pre$i$i$i5|0))|0);
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if ($33) {
       label = 32;
       break;
      } else {
       $$0$i$i$i$i3 = $31;
      }
     } else {
      $34 = HEAP32[$24>>2]|0;
      $$0$i$i$i$i3 = $34;
     }
     $35 = ($$0$i$i$i$i3|0)==(-1);
     if ($35) {
      HEAP32[$__b>>2] = 0;
      $$pre$i$i$i = 0;$157 = 1;
     } else {
      $$pre$i$i$i = $$pre$i$i$i5;$157 = 0;
     }
    }
    $36 = HEAP32[$__e>>2]|0;
    $37 = ($36|0)==(0|0);
    do {
     if ($37) {
      label = 24;
     } else {
      $38 = ((($36)) + 12|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ((($36)) + 16|0);
      $41 = HEAP32[$40>>2]|0;
      $42 = ($39|0)==($41|0);
      if ($42) {
       $43 = HEAP32[$36>>2]|0;
       $44 = ((($43)) + 36|0);
       $45 = HEAP32[$44>>2]|0;
       __THREW__ = 0;
       $46 = (invoke_ii($45|0,($36|0))|0);
       $47 = __THREW__; __THREW__ = 0;
       $48 = $47&1;
       if ($48) {
        label = 32;
        break L15;
       } else {
        $$0$i$i1$i$i9 = $46;
       }
      } else {
       $49 = HEAP32[$39>>2]|0;
       $$0$i$i1$i$i9 = $49;
      }
      $50 = ($$0$i$i1$i$i9|0)==(-1);
      if ($50) {
       HEAP32[$__e>>2] = 0;
       label = 24;
       break;
      } else {
       if ($157) {
        $158 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
     }
    } while(0);
    if ((label|0) == 24) {
     label = 0;
     if ($157) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $158 = 0;
     }
    }
    $51 = HEAP32[$__a_end>>2]|0;
    $52 = HEAP8[$__buf>>0]|0;
    $53 = $52 & 1;
    $54 = ($53<<24>>24)==(0);
    $55 = HEAP32[$20>>2]|0;
    $56 = $52&255;
    $57 = $56 >>> 1;
    $58 = $54 ? $57 : $55;
    $59 = (($__a$0) + ($58)|0);
    $60 = ($51|0)==($59|0);
    if ($60) {
     $61 = $58 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($61|0),0);
     $62 = __THREW__; __THREW__ = 0;
     $63 = $62&1;
     if ($63) {
      label = 32;
      break;
     }
     $64 = HEAP8[$__buf>>0]|0;
     $65 = $64 & 1;
     $66 = ($65<<24>>24)==(0);
     if ($66) {
      $69 = 10;
     } else {
      $67 = HEAP32[$__buf>>2]|0;
      $68 = $67 & -2;
      $phitmp$i23 = (($68) + -1)|0;
      $69 = $phitmp$i23;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($69|0),0);
     $70 = __THREW__; __THREW__ = 0;
     $71 = $70&1;
     if ($71) {
      label = 32;
      break;
     }
     $72 = HEAP8[$__buf>>0]|0;
     $73 = $72 & 1;
     $74 = ($73<<24>>24)==(0);
     $75 = HEAP32[$16>>2]|0;
     $76 = $74 ? $18 : $75;
     $77 = (($76) + ($58)|0);
     HEAP32[$__a_end>>2] = $77;
     $__a$1 = $76;
    } else {
     $__a$1 = $__a$0;
    }
    $82 = ((($$pre$i$i$i)) + 12|0);
    $83 = HEAP32[$82>>2]|0;
    $84 = ((($$pre$i$i$i)) + 16|0);
    $85 = HEAP32[$84>>2]|0;
    $86 = ($83|0)==($85|0);
    if ($86) {
     $87 = HEAP32[$$pre$i$i$i>>2]|0;
     $88 = ((($87)) + 36|0);
     $89 = HEAP32[$88>>2]|0;
     __THREW__ = 0;
     $90 = (invoke_ii($89|0,($$pre$i$i$i|0))|0);
     $91 = __THREW__; __THREW__ = 0;
     $92 = $91&1;
     if ($92) {
      label = 32;
      break;
     } else {
      $$0$i$i = $90;
     }
    } else {
     $93 = HEAP32[$83>>2]|0;
     $$0$i$i = $93;
    }
    $94 = (__ZNSt3__19__num_getIwE17__stage2_int_loopEwiPcRS2_RjwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_Pw($$0$i$i,$$0$i,$__a$1,$__a_end,$__dc,$21,$__grouping,$__g,$__g_end,$__atoms)|0);
    $95 = ($94|0)==(0);
    if (!($95)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $158;$__a$2 = $__a$1;
     break;
    }
    $96 = HEAP32[$82>>2]|0;
    $97 = HEAP32[$84>>2]|0;
    $98 = ($96|0)==($97|0);
    if ($98) {
     $99 = HEAP32[$$pre$i$i$i>>2]|0;
     $100 = ((($99)) + 40|0);
     $101 = HEAP32[$100>>2]|0;
     __THREW__ = 0;
     (invoke_ii($101|0,($$pre$i$i$i|0))|0);
     $102 = __THREW__; __THREW__ = 0;
     $103 = $102&1;
     if ($103) {
      label = 32;
      break;
     } else {
      $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $104 = ((($96)) + 4|0);
     HEAP32[$82>>2] = $104;
     $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 32) {
    $78 = ___cxa_find_matching_catch_2()|0;
    $79 = tempRet0;
    $lpad$phi$index = $78;$lpad$phi$index2 = $79;
    break;
   }
   $105 = HEAP8[$__grouping>>0]|0;
   $106 = $105 & 1;
   $107 = ($106<<24>>24)==(0);
   $108 = ((($__grouping)) + 4|0);
   $109 = HEAP32[$108>>2]|0;
   $110 = $105&255;
   $111 = $110 >>> 1;
   $112 = $107 ? $111 : $109;
   $113 = ($112|0)==(0);
   if (!($113)) {
    $114 = HEAP32[$__g_end>>2]|0;
    $115 = $114;
    $116 = $__g;
    $117 = (($115) - ($116))|0;
    $118 = ($117|0)<(160);
    if ($118) {
     $119 = HEAP32[$__dc>>2]|0;
     $120 = ((($114)) + 4|0);
     HEAP32[$__g_end>>2] = $120;
     HEAP32[$114>>2] = $119;
    }
   }
   $121 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $122 = (invoke_iiiii(4864,($__a$2|0),($121|0),($__err|0),($$0$i|0))|0);
   $123 = __THREW__; __THREW__ = 0;
   $124 = $123&1;
   if ($124) {
    label = 33;
   } else {
    HEAP32[$__v>>2] = $122;
    $125 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$125,$__err);
    $126 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($126) {
     $159 = 1;
    } else {
     $127 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $128 = HEAP32[$127>>2]|0;
     $129 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $130 = HEAP32[$129>>2]|0;
     $131 = ($128|0)==($130|0);
     if ($131) {
      $132 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $133 = ((($132)) + 36|0);
      $134 = HEAP32[$133>>2]|0;
      __THREW__ = 0;
      $135 = (invoke_ii($134|0,($$pre$i$i$i$lcssa|0))|0);
      $136 = __THREW__; __THREW__ = 0;
      $137 = $136&1;
      if ($137) {
       label = 33;
       break;
      } else {
       $$0$i$i$i$i = $135;
      }
     } else {
      $138 = HEAP32[$128>>2]|0;
      $$0$i$i$i$i = $138;
     }
     $139 = ($$0$i$i$i$i|0)==(-1);
     if ($139) {
      HEAP32[$__b>>2] = 0;
      $159 = 1;
     } else {
      $159 = 0;
     }
    }
    $140 = ($$pre$i3$i$i|0)==(0|0);
    do {
     if ($140) {
      label = 58;
     } else {
      $141 = ((($$pre$i3$i$i)) + 12|0);
      $142 = HEAP32[$141>>2]|0;
      $143 = ((($$pre$i3$i$i)) + 16|0);
      $144 = HEAP32[$143>>2]|0;
      $145 = ($142|0)==($144|0);
      if ($145) {
       $146 = HEAP32[$$pre$i3$i$i>>2]|0;
       $147 = ((($146)) + 36|0);
       $148 = HEAP32[$147>>2]|0;
       __THREW__ = 0;
       $149 = (invoke_ii($148|0,($$pre$i3$i$i|0))|0);
       $150 = __THREW__; __THREW__ = 0;
       $151 = $150&1;
       if ($151) {
        label = 33;
        break L13;
       } else {
        $$0$i$i1$i$i = $149;
       }
      } else {
       $152 = HEAP32[$142>>2]|0;
       $$0$i$i1$i$i = $152;
      }
      $153 = ($$0$i$i1$i$i|0)==(-1);
      if ($153) {
       HEAP32[$__e>>2] = 0;
       label = 58;
       break;
      } else {
       if ($159) {
        break;
       } else {
        label = 60;
        break;
       }
      }
     }
    } while(0);
    if ((label|0) == 58) {
     if ($159) {
      label = 60;
     }
    }
    if ((label|0) == 60) {
     $154 = HEAP32[$__err>>2]|0;
     $155 = $154 | 2;
     HEAP32[$__err>>2] = $155;
    }
    $156 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($156|0);
   }
  }
 } while(0);
 if ((label|0) == 33) {
  $80 = ___cxa_find_matching_catch_2()|0;
  $81 = tempRet0;
  $lpad$phi$index = $80;$lpad$phi$index2 = $81;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__19__num_getIwE17__stage2_int_prepERNS_8ios_baseEPwRw($agg$result,$__iob,$__atoms,$__thousands_sep) {
 $agg$result = $agg$result|0;
 $__iob = $__iob|0;
 $__atoms = $__atoms|0;
 $__thousands_sep = $__thousands_sep|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__loc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__loc = sp;
 $0 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$__loc>>2] = $0;
 __THREW__ = 0;
 $1 = (invoke_iii(3374,($__loc|0),(1207608|0))|0);
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  $4 = HEAP32[$1>>2]|0;
  $5 = ((($4)) + 48|0);
  $6 = HEAP32[$5>>2]|0;
  __THREW__ = 0;
  (invoke_iiiii($6|0,($1|0),(620486|0),((620512)|0),($__atoms|0))|0);
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if (!($8)) {
   __THREW__ = 0;
   $9 = (invoke_iii(3374,($__loc|0),(1207616|0))|0);
   $10 = __THREW__; __THREW__ = 0;
   $11 = $10&1;
   if (!($11)) {
    $12 = HEAP32[$9>>2]|0;
    $13 = ((($12)) + 16|0);
    $14 = HEAP32[$13>>2]|0;
    __THREW__ = 0;
    $15 = (invoke_ii($14|0,($9|0))|0);
    $16 = __THREW__; __THREW__ = 0;
    $17 = $16&1;
    if (!($17)) {
     HEAP32[$__thousands_sep>>2] = $15;
     $18 = HEAP32[$9>>2]|0;
     $19 = ((($18)) + 20|0);
     $20 = HEAP32[$19>>2]|0;
     __THREW__ = 0;
     invoke_vii($20|0,($agg$result|0),($9|0));
     $21 = __THREW__; __THREW__ = 0;
     $22 = $21&1;
     if (!($22)) {
      (__ZNSt3__114__shared_count16__release_sharedEv($0)|0);
      STACKTOP = sp;return;
     }
    }
   }
  }
 }
 $23 = ___cxa_find_matching_catch_2()|0;
 $24 = tempRet0;
 (__ZNSt3__114__shared_count16__release_sharedEv($0)|0);
 ___resumeException($23|0);
 // unreachable;
}
function __ZNSt3__19__num_getIwE17__stage2_int_loopEwiPcRS2_RjwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_Pw($__ct,$__base,$__a,$__a_end,$__dc,$__thousands_sep,$__grouping,$__g,$__g_end,$__atoms) {
 $__ct = $__ct|0;
 $__base = $__base|0;
 $__a = $__a|0;
 $__a_end = $__a_end|0;
 $__dc = $__dc|0;
 $__thousands_sep = $__thousands_sep|0;
 $__grouping = $__grouping|0;
 $__g = $__g|0;
 $__g_end = $__g_end|0;
 $__atoms = $__atoms|0;
 var $$0$i = 0, $$0$lcssa$i = 0, $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$__a_end>>2]|0;
 $1 = ($0|0)==($__a|0);
 do {
  if ($1) {
   $2 = ((($__atoms)) + 96|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = ($3|0)==($__ct|0);
   if (!($4)) {
    $5 = ((($__atoms)) + 100|0);
    $6 = HEAP32[$5>>2]|0;
    $7 = ($6|0)==($__ct|0);
    if (!($7)) {
     label = 5;
     break;
    }
   }
   $8 = $4 ? 43 : 45;
   $9 = ((($__a)) + 1|0);
   HEAP32[$__a_end>>2] = $9;
   HEAP8[$__a>>0] = $8;
   HEAP32[$__dc>>2] = 0;
   $$1 = 0;
  } else {
   label = 5;
  }
 } while(0);
 L6: do {
  if ((label|0) == 5) {
   $10 = HEAP8[$__grouping>>0]|0;
   $11 = $10 & 1;
   $12 = ($11<<24>>24)==(0);
   $13 = ((($__grouping)) + 4|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = $10&255;
   $16 = $15 >>> 1;
   $17 = $12 ? $16 : $14;
   $18 = ($17|0)!=(0);
   $19 = ($__ct|0)==($__thousands_sep|0);
   $or$cond = $19 & $18;
   if ($or$cond) {
    $20 = HEAP32[$__g_end>>2]|0;
    $21 = $20;
    $22 = $__g;
    $23 = (($21) - ($22))|0;
    $24 = ($23|0)<(160);
    if (!($24)) {
     $$1 = 0;
     break;
    }
    $25 = HEAP32[$__dc>>2]|0;
    $26 = ((($20)) + 4|0);
    HEAP32[$__g_end>>2] = $26;
    HEAP32[$20>>2] = $25;
    HEAP32[$__dc>>2] = 0;
    $$1 = 0;
    break;
   }
   $27 = ((($__atoms)) + 104|0);
   $$0$i = $__atoms;
   while(1) {
    $28 = ($$0$i|0)==($27|0);
    if ($28) {
     $$0$lcssa$i = $27;
     break;
    }
    $29 = HEAP32[$$0$i>>2]|0;
    $30 = ($29|0)==($__ct|0);
    if ($30) {
     $$0$lcssa$i = $$0$i;
     break;
    }
    $31 = ((($$0$i)) + 4|0);
    $$0$i = $31;
   }
   $32 = $$0$lcssa$i;
   $33 = $__atoms;
   $34 = (($32) - ($33))|0;
   $35 = $34 >> 2;
   $36 = ($34|0)>(92);
   if ($36) {
    $$1 = -1;
   } else {
    switch ($__base|0) {
    case 10: case 8:  {
     $37 = ($35|0)<($__base|0);
     if (!($37)) {
      $$1 = -1;
      break L6;
     }
     break;
    }
    case 16:  {
     $38 = ($34|0)<(88);
     if (!($38)) {
      if ($1) {
       $$1 = -1;
       break L6;
      }
      $39 = $0;
      $40 = $__a;
      $41 = (($39) - ($40))|0;
      $42 = ($41|0)<(3);
      if (!($42)) {
       $$1 = -1;
       break L6;
      }
      $43 = ((($0)) + -1|0);
      $44 = HEAP8[$43>>0]|0;
      $45 = ($44<<24>>24)==(48);
      if (!($45)) {
       $$1 = -1;
       break L6;
      }
      HEAP32[$__dc>>2] = 0;
      $46 = (620486 + ($35)|0);
      $47 = HEAP8[$46>>0]|0;
      $48 = ((($0)) + 1|0);
      HEAP32[$__a_end>>2] = $48;
      HEAP8[$0>>0] = $47;
      $$1 = 0;
      break L6;
     }
     break;
    }
    default: {
    }
    }
    $49 = (620486 + ($35)|0);
    $50 = HEAP8[$49>>0]|0;
    $51 = ((($0)) + 1|0);
    HEAP32[$__a_end>>2] = $51;
    HEAP8[$0>>0] = $50;
    $52 = HEAP32[$__dc>>2]|0;
    $53 = (($52) + 1)|0;
    HEAP32[$__dc>>2] = $53;
    $$1 = 0;
   }
  }
 } while(0);
 return ($$1|0);
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRx($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE15__do_get_signedIxEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE15__do_get_signedIxEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i3 = 0, $$0$i$i1$i$i = 0, $$0$i$i1$i$i9 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i5 = 0, $$pre$i3$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0;
 var $__atoms = 0, $__buf = 0, $__dc = 0, $__g = 0, $__g_end = 0, $__grouping = 0, $__i$0$i$i = 0, $__thousands_sep = 0, $exitcond$i$i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $phitmp$i = 0, $phitmp$i23 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 200|0;
 $__thousands_sep = sp + 196|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $0 = ((($__iob)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 74;
 switch ($2|0) {
 case 64:  {
  $$0$i = 8;
  break;
 }
 case 8:  {
  $$0$i = 16;
  break;
 }
 case 0:  {
  $$0$i = 0;
  break;
 }
 default: {
  $$0$i = 10;
 }
 }
 __ZNSt3__19__num_getIwE17__stage2_int_prepERNS_8ios_baseEPwRw($__grouping,$__iob,$__atoms,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $3 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$3>>2] = 0;
  $4 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $4;
 }
 $5 = HEAP8[$__buf>>0]|0;
 $6 = $5 & 1;
 $7 = ($6<<24>>24)==(0);
 if ($7) {
  $10 = 10;
 } else {
  $8 = HEAP32[$__buf>>2]|0;
  $9 = $8 & -2;
  $phitmp$i = (($9) + -1)|0;
  $10 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($10|0),0);
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 L13: do {
  if ($12) {
   label = 33;
  } else {
   $13 = HEAP8[$__buf>>0]|0;
   $14 = $13 & 1;
   $15 = ($14<<24>>24)==(0);
   $16 = ((($__buf)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($__buf)) + 1|0);
   $19 = $15 ? $18 : $17;
   HEAP32[$__a_end>>2] = $19;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   $20 = ((($__buf)) + 4|0);
   $21 = HEAP32[$__thousands_sep>>2]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i5 = $$pre;$__a$0 = $19;
   L15: while(1) {
    $22 = ($$pre$i$i$i5|0)==(0|0);
    if ($22) {
     $$pre$i$i$i = 0;$162 = 1;
    } else {
     $23 = ((($$pre$i$i$i5)) + 12|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ((($$pre$i$i$i5)) + 16|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = ($24|0)==($26|0);
     if ($27) {
      $28 = HEAP32[$$pre$i$i$i5>>2]|0;
      $29 = ((($28)) + 36|0);
      $30 = HEAP32[$29>>2]|0;
      __THREW__ = 0;
      $31 = (invoke_ii($30|0,($$pre$i$i$i5|0))|0);
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if ($33) {
       label = 32;
       break;
      } else {
       $$0$i$i$i$i3 = $31;
      }
     } else {
      $34 = HEAP32[$24>>2]|0;
      $$0$i$i$i$i3 = $34;
     }
     $35 = ($$0$i$i$i$i3|0)==(-1);
     if ($35) {
      HEAP32[$__b>>2] = 0;
      $$pre$i$i$i = 0;$162 = 1;
     } else {
      $$pre$i$i$i = $$pre$i$i$i5;$162 = 0;
     }
    }
    $36 = HEAP32[$__e>>2]|0;
    $37 = ($36|0)==(0|0);
    do {
     if ($37) {
      label = 24;
     } else {
      $38 = ((($36)) + 12|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ((($36)) + 16|0);
      $41 = HEAP32[$40>>2]|0;
      $42 = ($39|0)==($41|0);
      if ($42) {
       $43 = HEAP32[$36>>2]|0;
       $44 = ((($43)) + 36|0);
       $45 = HEAP32[$44>>2]|0;
       __THREW__ = 0;
       $46 = (invoke_ii($45|0,($36|0))|0);
       $47 = __THREW__; __THREW__ = 0;
       $48 = $47&1;
       if ($48) {
        label = 32;
        break L15;
       } else {
        $$0$i$i1$i$i9 = $46;
       }
      } else {
       $49 = HEAP32[$39>>2]|0;
       $$0$i$i1$i$i9 = $49;
      }
      $50 = ($$0$i$i1$i$i9|0)==(-1);
      if ($50) {
       HEAP32[$__e>>2] = 0;
       label = 24;
       break;
      } else {
       if ($162) {
        $163 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
     }
    } while(0);
    if ((label|0) == 24) {
     label = 0;
     if ($162) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $163 = 0;
     }
    }
    $51 = HEAP32[$__a_end>>2]|0;
    $52 = HEAP8[$__buf>>0]|0;
    $53 = $52 & 1;
    $54 = ($53<<24>>24)==(0);
    $55 = HEAP32[$20>>2]|0;
    $56 = $52&255;
    $57 = $56 >>> 1;
    $58 = $54 ? $57 : $55;
    $59 = (($__a$0) + ($58)|0);
    $60 = ($51|0)==($59|0);
    if ($60) {
     $61 = $58 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($61|0),0);
     $62 = __THREW__; __THREW__ = 0;
     $63 = $62&1;
     if ($63) {
      label = 32;
      break;
     }
     $64 = HEAP8[$__buf>>0]|0;
     $65 = $64 & 1;
     $66 = ($65<<24>>24)==(0);
     if ($66) {
      $69 = 10;
     } else {
      $67 = HEAP32[$__buf>>2]|0;
      $68 = $67 & -2;
      $phitmp$i23 = (($68) + -1)|0;
      $69 = $phitmp$i23;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($69|0),0);
     $70 = __THREW__; __THREW__ = 0;
     $71 = $70&1;
     if ($71) {
      label = 32;
      break;
     }
     $72 = HEAP8[$__buf>>0]|0;
     $73 = $72 & 1;
     $74 = ($73<<24>>24)==(0);
     $75 = HEAP32[$16>>2]|0;
     $76 = $74 ? $18 : $75;
     $77 = (($76) + ($58)|0);
     HEAP32[$__a_end>>2] = $77;
     $__a$1 = $76;
    } else {
     $__a$1 = $__a$0;
    }
    $82 = ((($$pre$i$i$i)) + 12|0);
    $83 = HEAP32[$82>>2]|0;
    $84 = ((($$pre$i$i$i)) + 16|0);
    $85 = HEAP32[$84>>2]|0;
    $86 = ($83|0)==($85|0);
    if ($86) {
     $87 = HEAP32[$$pre$i$i$i>>2]|0;
     $88 = ((($87)) + 36|0);
     $89 = HEAP32[$88>>2]|0;
     __THREW__ = 0;
     $90 = (invoke_ii($89|0,($$pre$i$i$i|0))|0);
     $91 = __THREW__; __THREW__ = 0;
     $92 = $91&1;
     if ($92) {
      label = 32;
      break;
     } else {
      $$0$i$i = $90;
     }
    } else {
     $93 = HEAP32[$83>>2]|0;
     $$0$i$i = $93;
    }
    $94 = (__ZNSt3__19__num_getIwE17__stage2_int_loopEwiPcRS2_RjwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_Pw($$0$i$i,$$0$i,$__a$1,$__a_end,$__dc,$21,$__grouping,$__g,$__g_end,$__atoms)|0);
    $95 = ($94|0)==(0);
    if (!($95)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $163;$__a$2 = $__a$1;
     break;
    }
    $96 = HEAP32[$82>>2]|0;
    $97 = HEAP32[$84>>2]|0;
    $98 = ($96|0)==($97|0);
    if ($98) {
     $99 = HEAP32[$$pre$i$i$i>>2]|0;
     $100 = ((($99)) + 40|0);
     $101 = HEAP32[$100>>2]|0;
     __THREW__ = 0;
     (invoke_ii($101|0,($$pre$i$i$i|0))|0);
     $102 = __THREW__; __THREW__ = 0;
     $103 = $102&1;
     if ($103) {
      label = 32;
      break;
     } else {
      $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $104 = ((($96)) + 4|0);
     HEAP32[$82>>2] = $104;
     $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 32) {
    $78 = ___cxa_find_matching_catch_2()|0;
    $79 = tempRet0;
    $lpad$phi$index = $78;$lpad$phi$index2 = $79;
    break;
   }
   $105 = HEAP8[$__grouping>>0]|0;
   $106 = $105 & 1;
   $107 = ($106<<24>>24)==(0);
   $108 = ((($__grouping)) + 4|0);
   $109 = HEAP32[$108>>2]|0;
   $110 = $105&255;
   $111 = $110 >>> 1;
   $112 = $107 ? $111 : $109;
   $113 = ($112|0)==(0);
   if (!($113)) {
    $114 = HEAP32[$__g_end>>2]|0;
    $115 = $114;
    $116 = $__g;
    $117 = (($115) - ($116))|0;
    $118 = ($117|0)<(160);
    if ($118) {
     $119 = HEAP32[$__dc>>2]|0;
     $120 = ((($114)) + 4|0);
     HEAP32[$__g_end>>2] = $120;
     HEAP32[$114>>2] = $119;
    }
   }
   $121 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $122 = (invoke_iiiii(4866,($__a$2|0),($121|0),($__err|0),($$0$i|0))|0);
   $123 = tempRet0;
   $124 = __THREW__; __THREW__ = 0;
   $125 = $124&1;
   if ($125) {
    label = 33;
   } else {
    $126 = $__v;
    $127 = $126;
    HEAP32[$127>>2] = $122;
    $128 = (($126) + 4)|0;
    $129 = $128;
    HEAP32[$129>>2] = $123;
    $130 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$130,$__err);
    $131 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($131) {
     $164 = 1;
    } else {
     $132 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $133 = HEAP32[$132>>2]|0;
     $134 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $135 = HEAP32[$134>>2]|0;
     $136 = ($133|0)==($135|0);
     if ($136) {
      $137 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $138 = ((($137)) + 36|0);
      $139 = HEAP32[$138>>2]|0;
      __THREW__ = 0;
      $140 = (invoke_ii($139|0,($$pre$i$i$i$lcssa|0))|0);
      $141 = __THREW__; __THREW__ = 0;
      $142 = $141&1;
      if ($142) {
       label = 33;
       break;
      } else {
       $$0$i$i$i$i = $140;
      }
     } else {
      $143 = HEAP32[$133>>2]|0;
      $$0$i$i$i$i = $143;
     }
     $144 = ($$0$i$i$i$i|0)==(-1);
     if ($144) {
      HEAP32[$__b>>2] = 0;
      $164 = 1;
     } else {
      $164 = 0;
     }
    }
    $145 = ($$pre$i3$i$i|0)==(0|0);
    do {
     if ($145) {
      label = 58;
     } else {
      $146 = ((($$pre$i3$i$i)) + 12|0);
      $147 = HEAP32[$146>>2]|0;
      $148 = ((($$pre$i3$i$i)) + 16|0);
      $149 = HEAP32[$148>>2]|0;
      $150 = ($147|0)==($149|0);
      if ($150) {
       $151 = HEAP32[$$pre$i3$i$i>>2]|0;
       $152 = ((($151)) + 36|0);
       $153 = HEAP32[$152>>2]|0;
       __THREW__ = 0;
       $154 = (invoke_ii($153|0,($$pre$i3$i$i|0))|0);
       $155 = __THREW__; __THREW__ = 0;
       $156 = $155&1;
       if ($156) {
        label = 33;
        break L13;
       } else {
        $$0$i$i1$i$i = $154;
       }
      } else {
       $157 = HEAP32[$147>>2]|0;
       $$0$i$i1$i$i = $157;
      }
      $158 = ($$0$i$i1$i$i|0)==(-1);
      if ($158) {
       HEAP32[$__e>>2] = 0;
       label = 58;
       break;
      } else {
       if ($164) {
        break;
       } else {
        label = 60;
        break;
       }
      }
     }
    } while(0);
    if ((label|0) == 58) {
     if ($164) {
      label = 60;
     }
    }
    if ((label|0) == 60) {
     $159 = HEAP32[$__err>>2]|0;
     $160 = $159 | 2;
     HEAP32[$__err>>2] = $160;
    }
    $161 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($161|0);
   }
  }
 } while(0);
 if ((label|0) == 33) {
  $80 = ___cxa_find_matching_catch_2()|0;
  $81 = tempRet0;
  $lpad$phi$index = $80;$lpad$phi$index2 = $81;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRt($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__do_get_unsignedItEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__do_get_unsignedItEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i3 = 0, $$0$i$i1$i$i = 0, $$0$i$i1$i$i9 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i5 = 0, $$pre$i3$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__g = 0, $__g_end = 0;
 var $__grouping = 0, $__i$0$i$i = 0, $__thousands_sep = 0, $exitcond$i$i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $phitmp$i = 0, $phitmp$i23 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 200|0;
 $__thousands_sep = sp + 196|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $0 = ((($__iob)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 74;
 switch ($2|0) {
 case 64:  {
  $$0$i = 8;
  break;
 }
 case 8:  {
  $$0$i = 16;
  break;
 }
 case 0:  {
  $$0$i = 0;
  break;
 }
 default: {
  $$0$i = 10;
 }
 }
 __ZNSt3__19__num_getIwE17__stage2_int_prepERNS_8ios_baseEPwRw($__grouping,$__iob,$__atoms,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $3 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$3>>2] = 0;
  $4 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $4;
 }
 $5 = HEAP8[$__buf>>0]|0;
 $6 = $5 & 1;
 $7 = ($6<<24>>24)==(0);
 if ($7) {
  $10 = 10;
 } else {
  $8 = HEAP32[$__buf>>2]|0;
  $9 = $8 & -2;
  $phitmp$i = (($9) + -1)|0;
  $10 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($10|0),0);
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 L13: do {
  if ($12) {
   label = 33;
  } else {
   $13 = HEAP8[$__buf>>0]|0;
   $14 = $13 & 1;
   $15 = ($14<<24>>24)==(0);
   $16 = ((($__buf)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($__buf)) + 1|0);
   $19 = $15 ? $18 : $17;
   HEAP32[$__a_end>>2] = $19;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   $20 = ((($__buf)) + 4|0);
   $21 = HEAP32[$__thousands_sep>>2]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i5 = $$pre;$__a$0 = $19;
   L15: while(1) {
    $22 = ($$pre$i$i$i5|0)==(0|0);
    if ($22) {
     $$pre$i$i$i = 0;$157 = 1;
    } else {
     $23 = ((($$pre$i$i$i5)) + 12|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ((($$pre$i$i$i5)) + 16|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = ($24|0)==($26|0);
     if ($27) {
      $28 = HEAP32[$$pre$i$i$i5>>2]|0;
      $29 = ((($28)) + 36|0);
      $30 = HEAP32[$29>>2]|0;
      __THREW__ = 0;
      $31 = (invoke_ii($30|0,($$pre$i$i$i5|0))|0);
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if ($33) {
       label = 32;
       break;
      } else {
       $$0$i$i$i$i3 = $31;
      }
     } else {
      $34 = HEAP32[$24>>2]|0;
      $$0$i$i$i$i3 = $34;
     }
     $35 = ($$0$i$i$i$i3|0)==(-1);
     if ($35) {
      HEAP32[$__b>>2] = 0;
      $$pre$i$i$i = 0;$157 = 1;
     } else {
      $$pre$i$i$i = $$pre$i$i$i5;$157 = 0;
     }
    }
    $36 = HEAP32[$__e>>2]|0;
    $37 = ($36|0)==(0|0);
    do {
     if ($37) {
      label = 24;
     } else {
      $38 = ((($36)) + 12|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ((($36)) + 16|0);
      $41 = HEAP32[$40>>2]|0;
      $42 = ($39|0)==($41|0);
      if ($42) {
       $43 = HEAP32[$36>>2]|0;
       $44 = ((($43)) + 36|0);
       $45 = HEAP32[$44>>2]|0;
       __THREW__ = 0;
       $46 = (invoke_ii($45|0,($36|0))|0);
       $47 = __THREW__; __THREW__ = 0;
       $48 = $47&1;
       if ($48) {
        label = 32;
        break L15;
       } else {
        $$0$i$i1$i$i9 = $46;
       }
      } else {
       $49 = HEAP32[$39>>2]|0;
       $$0$i$i1$i$i9 = $49;
      }
      $50 = ($$0$i$i1$i$i9|0)==(-1);
      if ($50) {
       HEAP32[$__e>>2] = 0;
       label = 24;
       break;
      } else {
       if ($157) {
        $158 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
     }
    } while(0);
    if ((label|0) == 24) {
     label = 0;
     if ($157) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $158 = 0;
     }
    }
    $51 = HEAP32[$__a_end>>2]|0;
    $52 = HEAP8[$__buf>>0]|0;
    $53 = $52 & 1;
    $54 = ($53<<24>>24)==(0);
    $55 = HEAP32[$20>>2]|0;
    $56 = $52&255;
    $57 = $56 >>> 1;
    $58 = $54 ? $57 : $55;
    $59 = (($__a$0) + ($58)|0);
    $60 = ($51|0)==($59|0);
    if ($60) {
     $61 = $58 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($61|0),0);
     $62 = __THREW__; __THREW__ = 0;
     $63 = $62&1;
     if ($63) {
      label = 32;
      break;
     }
     $64 = HEAP8[$__buf>>0]|0;
     $65 = $64 & 1;
     $66 = ($65<<24>>24)==(0);
     if ($66) {
      $69 = 10;
     } else {
      $67 = HEAP32[$__buf>>2]|0;
      $68 = $67 & -2;
      $phitmp$i23 = (($68) + -1)|0;
      $69 = $phitmp$i23;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($69|0),0);
     $70 = __THREW__; __THREW__ = 0;
     $71 = $70&1;
     if ($71) {
      label = 32;
      break;
     }
     $72 = HEAP8[$__buf>>0]|0;
     $73 = $72 & 1;
     $74 = ($73<<24>>24)==(0);
     $75 = HEAP32[$16>>2]|0;
     $76 = $74 ? $18 : $75;
     $77 = (($76) + ($58)|0);
     HEAP32[$__a_end>>2] = $77;
     $__a$1 = $76;
    } else {
     $__a$1 = $__a$0;
    }
    $82 = ((($$pre$i$i$i)) + 12|0);
    $83 = HEAP32[$82>>2]|0;
    $84 = ((($$pre$i$i$i)) + 16|0);
    $85 = HEAP32[$84>>2]|0;
    $86 = ($83|0)==($85|0);
    if ($86) {
     $87 = HEAP32[$$pre$i$i$i>>2]|0;
     $88 = ((($87)) + 36|0);
     $89 = HEAP32[$88>>2]|0;
     __THREW__ = 0;
     $90 = (invoke_ii($89|0,($$pre$i$i$i|0))|0);
     $91 = __THREW__; __THREW__ = 0;
     $92 = $91&1;
     if ($92) {
      label = 32;
      break;
     } else {
      $$0$i$i = $90;
     }
    } else {
     $93 = HEAP32[$83>>2]|0;
     $$0$i$i = $93;
    }
    $94 = (__ZNSt3__19__num_getIwE17__stage2_int_loopEwiPcRS2_RjwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_Pw($$0$i$i,$$0$i,$__a$1,$__a_end,$__dc,$21,$__grouping,$__g,$__g_end,$__atoms)|0);
    $95 = ($94|0)==(0);
    if (!($95)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $158;$__a$2 = $__a$1;
     break;
    }
    $96 = HEAP32[$82>>2]|0;
    $97 = HEAP32[$84>>2]|0;
    $98 = ($96|0)==($97|0);
    if ($98) {
     $99 = HEAP32[$$pre$i$i$i>>2]|0;
     $100 = ((($99)) + 40|0);
     $101 = HEAP32[$100>>2]|0;
     __THREW__ = 0;
     (invoke_ii($101|0,($$pre$i$i$i|0))|0);
     $102 = __THREW__; __THREW__ = 0;
     $103 = $102&1;
     if ($103) {
      label = 32;
      break;
     } else {
      $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $104 = ((($96)) + 4|0);
     HEAP32[$82>>2] = $104;
     $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 32) {
    $78 = ___cxa_find_matching_catch_2()|0;
    $79 = tempRet0;
    $lpad$phi$index = $78;$lpad$phi$index2 = $79;
    break;
   }
   $105 = HEAP8[$__grouping>>0]|0;
   $106 = $105 & 1;
   $107 = ($106<<24>>24)==(0);
   $108 = ((($__grouping)) + 4|0);
   $109 = HEAP32[$108>>2]|0;
   $110 = $105&255;
   $111 = $110 >>> 1;
   $112 = $107 ? $111 : $109;
   $113 = ($112|0)==(0);
   if (!($113)) {
    $114 = HEAP32[$__g_end>>2]|0;
    $115 = $114;
    $116 = $__g;
    $117 = (($115) - ($116))|0;
    $118 = ($117|0)<(160);
    if ($118) {
     $119 = HEAP32[$__dc>>2]|0;
     $120 = ((($114)) + 4|0);
     HEAP32[$__g_end>>2] = $120;
     HEAP32[$114>>2] = $119;
    }
   }
   $121 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $122 = (invoke_iiiii(4867,($__a$2|0),($121|0),($__err|0),($$0$i|0))|0);
   $123 = __THREW__; __THREW__ = 0;
   $124 = $123&1;
   if ($124) {
    label = 33;
   } else {
    HEAP16[$__v>>1] = $122;
    $125 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$125,$__err);
    $126 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($126) {
     $159 = 1;
    } else {
     $127 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $128 = HEAP32[$127>>2]|0;
     $129 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $130 = HEAP32[$129>>2]|0;
     $131 = ($128|0)==($130|0);
     if ($131) {
      $132 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $133 = ((($132)) + 36|0);
      $134 = HEAP32[$133>>2]|0;
      __THREW__ = 0;
      $135 = (invoke_ii($134|0,($$pre$i$i$i$lcssa|0))|0);
      $136 = __THREW__; __THREW__ = 0;
      $137 = $136&1;
      if ($137) {
       label = 33;
       break;
      } else {
       $$0$i$i$i$i = $135;
      }
     } else {
      $138 = HEAP32[$128>>2]|0;
      $$0$i$i$i$i = $138;
     }
     $139 = ($$0$i$i$i$i|0)==(-1);
     if ($139) {
      HEAP32[$__b>>2] = 0;
      $159 = 1;
     } else {
      $159 = 0;
     }
    }
    $140 = ($$pre$i3$i$i|0)==(0|0);
    do {
     if ($140) {
      label = 58;
     } else {
      $141 = ((($$pre$i3$i$i)) + 12|0);
      $142 = HEAP32[$141>>2]|0;
      $143 = ((($$pre$i3$i$i)) + 16|0);
      $144 = HEAP32[$143>>2]|0;
      $145 = ($142|0)==($144|0);
      if ($145) {
       $146 = HEAP32[$$pre$i3$i$i>>2]|0;
       $147 = ((($146)) + 36|0);
       $148 = HEAP32[$147>>2]|0;
       __THREW__ = 0;
       $149 = (invoke_ii($148|0,($$pre$i3$i$i|0))|0);
       $150 = __THREW__; __THREW__ = 0;
       $151 = $150&1;
       if ($151) {
        label = 33;
        break L13;
       } else {
        $$0$i$i1$i$i = $149;
       }
      } else {
       $152 = HEAP32[$142>>2]|0;
       $$0$i$i1$i$i = $152;
      }
      $153 = ($$0$i$i1$i$i|0)==(-1);
      if ($153) {
       HEAP32[$__e>>2] = 0;
       label = 58;
       break;
      } else {
       if ($159) {
        break;
       } else {
        label = 60;
        break;
       }
      }
     }
    } while(0);
    if ((label|0) == 58) {
     if ($159) {
      label = 60;
     }
    }
    if ((label|0) == 60) {
     $154 = HEAP32[$__err>>2]|0;
     $155 = $154 | 2;
     HEAP32[$__err>>2] = $155;
    }
    $156 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($156|0);
   }
  }
 } while(0);
 if ((label|0) == 33) {
  $80 = ___cxa_find_matching_catch_2()|0;
  $81 = tempRet0;
  $lpad$phi$index = $80;$lpad$phi$index2 = $81;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjS8_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__do_get_unsignedIjEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__do_get_unsignedIjEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i3 = 0, $$0$i$i1$i$i = 0, $$0$i$i1$i$i9 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i5 = 0, $$pre$i3$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__g = 0, $__g_end = 0;
 var $__grouping = 0, $__i$0$i$i = 0, $__thousands_sep = 0, $exitcond$i$i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $phitmp$i = 0, $phitmp$i23 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 200|0;
 $__thousands_sep = sp + 196|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $0 = ((($__iob)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 74;
 switch ($2|0) {
 case 64:  {
  $$0$i = 8;
  break;
 }
 case 8:  {
  $$0$i = 16;
  break;
 }
 case 0:  {
  $$0$i = 0;
  break;
 }
 default: {
  $$0$i = 10;
 }
 }
 __ZNSt3__19__num_getIwE17__stage2_int_prepERNS_8ios_baseEPwRw($__grouping,$__iob,$__atoms,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $3 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$3>>2] = 0;
  $4 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $4;
 }
 $5 = HEAP8[$__buf>>0]|0;
 $6 = $5 & 1;
 $7 = ($6<<24>>24)==(0);
 if ($7) {
  $10 = 10;
 } else {
  $8 = HEAP32[$__buf>>2]|0;
  $9 = $8 & -2;
  $phitmp$i = (($9) + -1)|0;
  $10 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($10|0),0);
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 L13: do {
  if ($12) {
   label = 33;
  } else {
   $13 = HEAP8[$__buf>>0]|0;
   $14 = $13 & 1;
   $15 = ($14<<24>>24)==(0);
   $16 = ((($__buf)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($__buf)) + 1|0);
   $19 = $15 ? $18 : $17;
   HEAP32[$__a_end>>2] = $19;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   $20 = ((($__buf)) + 4|0);
   $21 = HEAP32[$__thousands_sep>>2]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i5 = $$pre;$__a$0 = $19;
   L15: while(1) {
    $22 = ($$pre$i$i$i5|0)==(0|0);
    if ($22) {
     $$pre$i$i$i = 0;$157 = 1;
    } else {
     $23 = ((($$pre$i$i$i5)) + 12|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ((($$pre$i$i$i5)) + 16|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = ($24|0)==($26|0);
     if ($27) {
      $28 = HEAP32[$$pre$i$i$i5>>2]|0;
      $29 = ((($28)) + 36|0);
      $30 = HEAP32[$29>>2]|0;
      __THREW__ = 0;
      $31 = (invoke_ii($30|0,($$pre$i$i$i5|0))|0);
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if ($33) {
       label = 32;
       break;
      } else {
       $$0$i$i$i$i3 = $31;
      }
     } else {
      $34 = HEAP32[$24>>2]|0;
      $$0$i$i$i$i3 = $34;
     }
     $35 = ($$0$i$i$i$i3|0)==(-1);
     if ($35) {
      HEAP32[$__b>>2] = 0;
      $$pre$i$i$i = 0;$157 = 1;
     } else {
      $$pre$i$i$i = $$pre$i$i$i5;$157 = 0;
     }
    }
    $36 = HEAP32[$__e>>2]|0;
    $37 = ($36|0)==(0|0);
    do {
     if ($37) {
      label = 24;
     } else {
      $38 = ((($36)) + 12|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ((($36)) + 16|0);
      $41 = HEAP32[$40>>2]|0;
      $42 = ($39|0)==($41|0);
      if ($42) {
       $43 = HEAP32[$36>>2]|0;
       $44 = ((($43)) + 36|0);
       $45 = HEAP32[$44>>2]|0;
       __THREW__ = 0;
       $46 = (invoke_ii($45|0,($36|0))|0);
       $47 = __THREW__; __THREW__ = 0;
       $48 = $47&1;
       if ($48) {
        label = 32;
        break L15;
       } else {
        $$0$i$i1$i$i9 = $46;
       }
      } else {
       $49 = HEAP32[$39>>2]|0;
       $$0$i$i1$i$i9 = $49;
      }
      $50 = ($$0$i$i1$i$i9|0)==(-1);
      if ($50) {
       HEAP32[$__e>>2] = 0;
       label = 24;
       break;
      } else {
       if ($157) {
        $158 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
     }
    } while(0);
    if ((label|0) == 24) {
     label = 0;
     if ($157) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $158 = 0;
     }
    }
    $51 = HEAP32[$__a_end>>2]|0;
    $52 = HEAP8[$__buf>>0]|0;
    $53 = $52 & 1;
    $54 = ($53<<24>>24)==(0);
    $55 = HEAP32[$20>>2]|0;
    $56 = $52&255;
    $57 = $56 >>> 1;
    $58 = $54 ? $57 : $55;
    $59 = (($__a$0) + ($58)|0);
    $60 = ($51|0)==($59|0);
    if ($60) {
     $61 = $58 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($61|0),0);
     $62 = __THREW__; __THREW__ = 0;
     $63 = $62&1;
     if ($63) {
      label = 32;
      break;
     }
     $64 = HEAP8[$__buf>>0]|0;
     $65 = $64 & 1;
     $66 = ($65<<24>>24)==(0);
     if ($66) {
      $69 = 10;
     } else {
      $67 = HEAP32[$__buf>>2]|0;
      $68 = $67 & -2;
      $phitmp$i23 = (($68) + -1)|0;
      $69 = $phitmp$i23;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($69|0),0);
     $70 = __THREW__; __THREW__ = 0;
     $71 = $70&1;
     if ($71) {
      label = 32;
      break;
     }
     $72 = HEAP8[$__buf>>0]|0;
     $73 = $72 & 1;
     $74 = ($73<<24>>24)==(0);
     $75 = HEAP32[$16>>2]|0;
     $76 = $74 ? $18 : $75;
     $77 = (($76) + ($58)|0);
     HEAP32[$__a_end>>2] = $77;
     $__a$1 = $76;
    } else {
     $__a$1 = $__a$0;
    }
    $82 = ((($$pre$i$i$i)) + 12|0);
    $83 = HEAP32[$82>>2]|0;
    $84 = ((($$pre$i$i$i)) + 16|0);
    $85 = HEAP32[$84>>2]|0;
    $86 = ($83|0)==($85|0);
    if ($86) {
     $87 = HEAP32[$$pre$i$i$i>>2]|0;
     $88 = ((($87)) + 36|0);
     $89 = HEAP32[$88>>2]|0;
     __THREW__ = 0;
     $90 = (invoke_ii($89|0,($$pre$i$i$i|0))|0);
     $91 = __THREW__; __THREW__ = 0;
     $92 = $91&1;
     if ($92) {
      label = 32;
      break;
     } else {
      $$0$i$i = $90;
     }
    } else {
     $93 = HEAP32[$83>>2]|0;
     $$0$i$i = $93;
    }
    $94 = (__ZNSt3__19__num_getIwE17__stage2_int_loopEwiPcRS2_RjwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_Pw($$0$i$i,$$0$i,$__a$1,$__a_end,$__dc,$21,$__grouping,$__g,$__g_end,$__atoms)|0);
    $95 = ($94|0)==(0);
    if (!($95)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $158;$__a$2 = $__a$1;
     break;
    }
    $96 = HEAP32[$82>>2]|0;
    $97 = HEAP32[$84>>2]|0;
    $98 = ($96|0)==($97|0);
    if ($98) {
     $99 = HEAP32[$$pre$i$i$i>>2]|0;
     $100 = ((($99)) + 40|0);
     $101 = HEAP32[$100>>2]|0;
     __THREW__ = 0;
     (invoke_ii($101|0,($$pre$i$i$i|0))|0);
     $102 = __THREW__; __THREW__ = 0;
     $103 = $102&1;
     if ($103) {
      label = 32;
      break;
     } else {
      $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $104 = ((($96)) + 4|0);
     HEAP32[$82>>2] = $104;
     $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 32) {
    $78 = ___cxa_find_matching_catch_2()|0;
    $79 = tempRet0;
    $lpad$phi$index = $78;$lpad$phi$index2 = $79;
    break;
   }
   $105 = HEAP8[$__grouping>>0]|0;
   $106 = $105 & 1;
   $107 = ($106<<24>>24)==(0);
   $108 = ((($__grouping)) + 4|0);
   $109 = HEAP32[$108>>2]|0;
   $110 = $105&255;
   $111 = $110 >>> 1;
   $112 = $107 ? $111 : $109;
   $113 = ($112|0)==(0);
   if (!($113)) {
    $114 = HEAP32[$__g_end>>2]|0;
    $115 = $114;
    $116 = $__g;
    $117 = (($115) - ($116))|0;
    $118 = ($117|0)<(160);
    if ($118) {
     $119 = HEAP32[$__dc>>2]|0;
     $120 = ((($114)) + 4|0);
     HEAP32[$__g_end>>2] = $120;
     HEAP32[$114>>2] = $119;
    }
   }
   $121 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $122 = (invoke_iiiii(4868,($__a$2|0),($121|0),($__err|0),($$0$i|0))|0);
   $123 = __THREW__; __THREW__ = 0;
   $124 = $123&1;
   if ($124) {
    label = 33;
   } else {
    HEAP32[$__v>>2] = $122;
    $125 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$125,$__err);
    $126 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($126) {
     $159 = 1;
    } else {
     $127 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $128 = HEAP32[$127>>2]|0;
     $129 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $130 = HEAP32[$129>>2]|0;
     $131 = ($128|0)==($130|0);
     if ($131) {
      $132 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $133 = ((($132)) + 36|0);
      $134 = HEAP32[$133>>2]|0;
      __THREW__ = 0;
      $135 = (invoke_ii($134|0,($$pre$i$i$i$lcssa|0))|0);
      $136 = __THREW__; __THREW__ = 0;
      $137 = $136&1;
      if ($137) {
       label = 33;
       break;
      } else {
       $$0$i$i$i$i = $135;
      }
     } else {
      $138 = HEAP32[$128>>2]|0;
      $$0$i$i$i$i = $138;
     }
     $139 = ($$0$i$i$i$i|0)==(-1);
     if ($139) {
      HEAP32[$__b>>2] = 0;
      $159 = 1;
     } else {
      $159 = 0;
     }
    }
    $140 = ($$pre$i3$i$i|0)==(0|0);
    do {
     if ($140) {
      label = 58;
     } else {
      $141 = ((($$pre$i3$i$i)) + 12|0);
      $142 = HEAP32[$141>>2]|0;
      $143 = ((($$pre$i3$i$i)) + 16|0);
      $144 = HEAP32[$143>>2]|0;
      $145 = ($142|0)==($144|0);
      if ($145) {
       $146 = HEAP32[$$pre$i3$i$i>>2]|0;
       $147 = ((($146)) + 36|0);
       $148 = HEAP32[$147>>2]|0;
       __THREW__ = 0;
       $149 = (invoke_ii($148|0,($$pre$i3$i$i|0))|0);
       $150 = __THREW__; __THREW__ = 0;
       $151 = $150&1;
       if ($151) {
        label = 33;
        break L13;
       } else {
        $$0$i$i1$i$i = $149;
       }
      } else {
       $152 = HEAP32[$142>>2]|0;
       $$0$i$i1$i$i = $152;
      }
      $153 = ($$0$i$i1$i$i|0)==(-1);
      if ($153) {
       HEAP32[$__e>>2] = 0;
       label = 58;
       break;
      } else {
       if ($159) {
        break;
       } else {
        label = 60;
        break;
       }
      }
     }
    } while(0);
    if ((label|0) == 58) {
     if ($159) {
      label = 60;
     }
    }
    if ((label|0) == 60) {
     $154 = HEAP32[$__err>>2]|0;
     $155 = $154 | 2;
     HEAP32[$__err>>2] = $155;
    }
    $156 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($156|0);
   }
  }
 } while(0);
 if ((label|0) == 33) {
  $80 = ___cxa_find_matching_catch_2()|0;
  $81 = tempRet0;
  $lpad$phi$index = $80;$lpad$phi$index2 = $81;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRm($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__do_get_unsignedImEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__do_get_unsignedImEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i3 = 0, $$0$i$i1$i$i = 0, $$0$i$i1$i$i9 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i5 = 0, $$pre$i3$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__g = 0, $__g_end = 0;
 var $__grouping = 0, $__i$0$i$i = 0, $__thousands_sep = 0, $exitcond$i$i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $phitmp$i = 0, $phitmp$i23 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 200|0;
 $__thousands_sep = sp + 196|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $0 = ((($__iob)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 74;
 switch ($2|0) {
 case 64:  {
  $$0$i = 8;
  break;
 }
 case 8:  {
  $$0$i = 16;
  break;
 }
 case 0:  {
  $$0$i = 0;
  break;
 }
 default: {
  $$0$i = 10;
 }
 }
 __ZNSt3__19__num_getIwE17__stage2_int_prepERNS_8ios_baseEPwRw($__grouping,$__iob,$__atoms,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $3 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$3>>2] = 0;
  $4 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $4;
 }
 $5 = HEAP8[$__buf>>0]|0;
 $6 = $5 & 1;
 $7 = ($6<<24>>24)==(0);
 if ($7) {
  $10 = 10;
 } else {
  $8 = HEAP32[$__buf>>2]|0;
  $9 = $8 & -2;
  $phitmp$i = (($9) + -1)|0;
  $10 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($10|0),0);
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 L13: do {
  if ($12) {
   label = 33;
  } else {
   $13 = HEAP8[$__buf>>0]|0;
   $14 = $13 & 1;
   $15 = ($14<<24>>24)==(0);
   $16 = ((($__buf)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($__buf)) + 1|0);
   $19 = $15 ? $18 : $17;
   HEAP32[$__a_end>>2] = $19;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   $20 = ((($__buf)) + 4|0);
   $21 = HEAP32[$__thousands_sep>>2]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i5 = $$pre;$__a$0 = $19;
   L15: while(1) {
    $22 = ($$pre$i$i$i5|0)==(0|0);
    if ($22) {
     $$pre$i$i$i = 0;$157 = 1;
    } else {
     $23 = ((($$pre$i$i$i5)) + 12|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ((($$pre$i$i$i5)) + 16|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = ($24|0)==($26|0);
     if ($27) {
      $28 = HEAP32[$$pre$i$i$i5>>2]|0;
      $29 = ((($28)) + 36|0);
      $30 = HEAP32[$29>>2]|0;
      __THREW__ = 0;
      $31 = (invoke_ii($30|0,($$pre$i$i$i5|0))|0);
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if ($33) {
       label = 32;
       break;
      } else {
       $$0$i$i$i$i3 = $31;
      }
     } else {
      $34 = HEAP32[$24>>2]|0;
      $$0$i$i$i$i3 = $34;
     }
     $35 = ($$0$i$i$i$i3|0)==(-1);
     if ($35) {
      HEAP32[$__b>>2] = 0;
      $$pre$i$i$i = 0;$157 = 1;
     } else {
      $$pre$i$i$i = $$pre$i$i$i5;$157 = 0;
     }
    }
    $36 = HEAP32[$__e>>2]|0;
    $37 = ($36|0)==(0|0);
    do {
     if ($37) {
      label = 24;
     } else {
      $38 = ((($36)) + 12|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ((($36)) + 16|0);
      $41 = HEAP32[$40>>2]|0;
      $42 = ($39|0)==($41|0);
      if ($42) {
       $43 = HEAP32[$36>>2]|0;
       $44 = ((($43)) + 36|0);
       $45 = HEAP32[$44>>2]|0;
       __THREW__ = 0;
       $46 = (invoke_ii($45|0,($36|0))|0);
       $47 = __THREW__; __THREW__ = 0;
       $48 = $47&1;
       if ($48) {
        label = 32;
        break L15;
       } else {
        $$0$i$i1$i$i9 = $46;
       }
      } else {
       $49 = HEAP32[$39>>2]|0;
       $$0$i$i1$i$i9 = $49;
      }
      $50 = ($$0$i$i1$i$i9|0)==(-1);
      if ($50) {
       HEAP32[$__e>>2] = 0;
       label = 24;
       break;
      } else {
       if ($157) {
        $158 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
     }
    } while(0);
    if ((label|0) == 24) {
     label = 0;
     if ($157) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $158 = 0;
     }
    }
    $51 = HEAP32[$__a_end>>2]|0;
    $52 = HEAP8[$__buf>>0]|0;
    $53 = $52 & 1;
    $54 = ($53<<24>>24)==(0);
    $55 = HEAP32[$20>>2]|0;
    $56 = $52&255;
    $57 = $56 >>> 1;
    $58 = $54 ? $57 : $55;
    $59 = (($__a$0) + ($58)|0);
    $60 = ($51|0)==($59|0);
    if ($60) {
     $61 = $58 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($61|0),0);
     $62 = __THREW__; __THREW__ = 0;
     $63 = $62&1;
     if ($63) {
      label = 32;
      break;
     }
     $64 = HEAP8[$__buf>>0]|0;
     $65 = $64 & 1;
     $66 = ($65<<24>>24)==(0);
     if ($66) {
      $69 = 10;
     } else {
      $67 = HEAP32[$__buf>>2]|0;
      $68 = $67 & -2;
      $phitmp$i23 = (($68) + -1)|0;
      $69 = $phitmp$i23;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($69|0),0);
     $70 = __THREW__; __THREW__ = 0;
     $71 = $70&1;
     if ($71) {
      label = 32;
      break;
     }
     $72 = HEAP8[$__buf>>0]|0;
     $73 = $72 & 1;
     $74 = ($73<<24>>24)==(0);
     $75 = HEAP32[$16>>2]|0;
     $76 = $74 ? $18 : $75;
     $77 = (($76) + ($58)|0);
     HEAP32[$__a_end>>2] = $77;
     $__a$1 = $76;
    } else {
     $__a$1 = $__a$0;
    }
    $82 = ((($$pre$i$i$i)) + 12|0);
    $83 = HEAP32[$82>>2]|0;
    $84 = ((($$pre$i$i$i)) + 16|0);
    $85 = HEAP32[$84>>2]|0;
    $86 = ($83|0)==($85|0);
    if ($86) {
     $87 = HEAP32[$$pre$i$i$i>>2]|0;
     $88 = ((($87)) + 36|0);
     $89 = HEAP32[$88>>2]|0;
     __THREW__ = 0;
     $90 = (invoke_ii($89|0,($$pre$i$i$i|0))|0);
     $91 = __THREW__; __THREW__ = 0;
     $92 = $91&1;
     if ($92) {
      label = 32;
      break;
     } else {
      $$0$i$i = $90;
     }
    } else {
     $93 = HEAP32[$83>>2]|0;
     $$0$i$i = $93;
    }
    $94 = (__ZNSt3__19__num_getIwE17__stage2_int_loopEwiPcRS2_RjwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_Pw($$0$i$i,$$0$i,$__a$1,$__a_end,$__dc,$21,$__grouping,$__g,$__g_end,$__atoms)|0);
    $95 = ($94|0)==(0);
    if (!($95)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $158;$__a$2 = $__a$1;
     break;
    }
    $96 = HEAP32[$82>>2]|0;
    $97 = HEAP32[$84>>2]|0;
    $98 = ($96|0)==($97|0);
    if ($98) {
     $99 = HEAP32[$$pre$i$i$i>>2]|0;
     $100 = ((($99)) + 40|0);
     $101 = HEAP32[$100>>2]|0;
     __THREW__ = 0;
     (invoke_ii($101|0,($$pre$i$i$i|0))|0);
     $102 = __THREW__; __THREW__ = 0;
     $103 = $102&1;
     if ($103) {
      label = 32;
      break;
     } else {
      $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $104 = ((($96)) + 4|0);
     HEAP32[$82>>2] = $104;
     $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 32) {
    $78 = ___cxa_find_matching_catch_2()|0;
    $79 = tempRet0;
    $lpad$phi$index = $78;$lpad$phi$index2 = $79;
    break;
   }
   $105 = HEAP8[$__grouping>>0]|0;
   $106 = $105 & 1;
   $107 = ($106<<24>>24)==(0);
   $108 = ((($__grouping)) + 4|0);
   $109 = HEAP32[$108>>2]|0;
   $110 = $105&255;
   $111 = $110 >>> 1;
   $112 = $107 ? $111 : $109;
   $113 = ($112|0)==(0);
   if (!($113)) {
    $114 = HEAP32[$__g_end>>2]|0;
    $115 = $114;
    $116 = $__g;
    $117 = (($115) - ($116))|0;
    $118 = ($117|0)<(160);
    if ($118) {
     $119 = HEAP32[$__dc>>2]|0;
     $120 = ((($114)) + 4|0);
     HEAP32[$__g_end>>2] = $120;
     HEAP32[$114>>2] = $119;
    }
   }
   $121 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $122 = (invoke_iiiii(4869,($__a$2|0),($121|0),($__err|0),($$0$i|0))|0);
   $123 = __THREW__; __THREW__ = 0;
   $124 = $123&1;
   if ($124) {
    label = 33;
   } else {
    HEAP32[$__v>>2] = $122;
    $125 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$125,$__err);
    $126 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($126) {
     $159 = 1;
    } else {
     $127 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $128 = HEAP32[$127>>2]|0;
     $129 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $130 = HEAP32[$129>>2]|0;
     $131 = ($128|0)==($130|0);
     if ($131) {
      $132 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $133 = ((($132)) + 36|0);
      $134 = HEAP32[$133>>2]|0;
      __THREW__ = 0;
      $135 = (invoke_ii($134|0,($$pre$i$i$i$lcssa|0))|0);
      $136 = __THREW__; __THREW__ = 0;
      $137 = $136&1;
      if ($137) {
       label = 33;
       break;
      } else {
       $$0$i$i$i$i = $135;
      }
     } else {
      $138 = HEAP32[$128>>2]|0;
      $$0$i$i$i$i = $138;
     }
     $139 = ($$0$i$i$i$i|0)==(-1);
     if ($139) {
      HEAP32[$__b>>2] = 0;
      $159 = 1;
     } else {
      $159 = 0;
     }
    }
    $140 = ($$pre$i3$i$i|0)==(0|0);
    do {
     if ($140) {
      label = 58;
     } else {
      $141 = ((($$pre$i3$i$i)) + 12|0);
      $142 = HEAP32[$141>>2]|0;
      $143 = ((($$pre$i3$i$i)) + 16|0);
      $144 = HEAP32[$143>>2]|0;
      $145 = ($142|0)==($144|0);
      if ($145) {
       $146 = HEAP32[$$pre$i3$i$i>>2]|0;
       $147 = ((($146)) + 36|0);
       $148 = HEAP32[$147>>2]|0;
       __THREW__ = 0;
       $149 = (invoke_ii($148|0,($$pre$i3$i$i|0))|0);
       $150 = __THREW__; __THREW__ = 0;
       $151 = $150&1;
       if ($151) {
        label = 33;
        break L13;
       } else {
        $$0$i$i1$i$i = $149;
       }
      } else {
       $152 = HEAP32[$142>>2]|0;
       $$0$i$i1$i$i = $152;
      }
      $153 = ($$0$i$i1$i$i|0)==(-1);
      if ($153) {
       HEAP32[$__e>>2] = 0;
       label = 58;
       break;
      } else {
       if ($159) {
        break;
       } else {
        label = 60;
        break;
       }
      }
     }
    } while(0);
    if ((label|0) == 58) {
     if ($159) {
      label = 60;
     }
    }
    if ((label|0) == 60) {
     $154 = HEAP32[$__err>>2]|0;
     $155 = $154 | 2;
     HEAP32[$__err>>2] = $155;
    }
    $156 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($156|0);
   }
  }
 } while(0);
 if ((label|0) == 33) {
  $80 = ___cxa_find_matching_catch_2()|0;
  $81 = tempRet0;
  $lpad$phi$index = $80;$lpad$phi$index2 = $81;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRy($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__do_get_unsignedIyEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__do_get_unsignedIyEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i3 = 0, $$0$i$i1$i$i = 0, $$0$i$i1$i$i9 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i5 = 0, $$pre$i3$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0;
 var $__atoms = 0, $__buf = 0, $__dc = 0, $__g = 0, $__g_end = 0, $__grouping = 0, $__i$0$i$i = 0, $__thousands_sep = 0, $exitcond$i$i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $phitmp$i = 0, $phitmp$i23 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 200|0;
 $__thousands_sep = sp + 196|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $0 = ((($__iob)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 74;
 switch ($2|0) {
 case 64:  {
  $$0$i = 8;
  break;
 }
 case 8:  {
  $$0$i = 16;
  break;
 }
 case 0:  {
  $$0$i = 0;
  break;
 }
 default: {
  $$0$i = 10;
 }
 }
 __ZNSt3__19__num_getIwE17__stage2_int_prepERNS_8ios_baseEPwRw($__grouping,$__iob,$__atoms,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $3 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$3>>2] = 0;
  $4 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $4;
 }
 $5 = HEAP8[$__buf>>0]|0;
 $6 = $5 & 1;
 $7 = ($6<<24>>24)==(0);
 if ($7) {
  $10 = 10;
 } else {
  $8 = HEAP32[$__buf>>2]|0;
  $9 = $8 & -2;
  $phitmp$i = (($9) + -1)|0;
  $10 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($10|0),0);
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 L13: do {
  if ($12) {
   label = 33;
  } else {
   $13 = HEAP8[$__buf>>0]|0;
   $14 = $13 & 1;
   $15 = ($14<<24>>24)==(0);
   $16 = ((($__buf)) + 8|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($__buf)) + 1|0);
   $19 = $15 ? $18 : $17;
   HEAP32[$__a_end>>2] = $19;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   $20 = ((($__buf)) + 4|0);
   $21 = HEAP32[$__thousands_sep>>2]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i5 = $$pre;$__a$0 = $19;
   L15: while(1) {
    $22 = ($$pre$i$i$i5|0)==(0|0);
    if ($22) {
     $$pre$i$i$i = 0;$162 = 1;
    } else {
     $23 = ((($$pre$i$i$i5)) + 12|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ((($$pre$i$i$i5)) + 16|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = ($24|0)==($26|0);
     if ($27) {
      $28 = HEAP32[$$pre$i$i$i5>>2]|0;
      $29 = ((($28)) + 36|0);
      $30 = HEAP32[$29>>2]|0;
      __THREW__ = 0;
      $31 = (invoke_ii($30|0,($$pre$i$i$i5|0))|0);
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if ($33) {
       label = 32;
       break;
      } else {
       $$0$i$i$i$i3 = $31;
      }
     } else {
      $34 = HEAP32[$24>>2]|0;
      $$0$i$i$i$i3 = $34;
     }
     $35 = ($$0$i$i$i$i3|0)==(-1);
     if ($35) {
      HEAP32[$__b>>2] = 0;
      $$pre$i$i$i = 0;$162 = 1;
     } else {
      $$pre$i$i$i = $$pre$i$i$i5;$162 = 0;
     }
    }
    $36 = HEAP32[$__e>>2]|0;
    $37 = ($36|0)==(0|0);
    do {
     if ($37) {
      label = 24;
     } else {
      $38 = ((($36)) + 12|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ((($36)) + 16|0);
      $41 = HEAP32[$40>>2]|0;
      $42 = ($39|0)==($41|0);
      if ($42) {
       $43 = HEAP32[$36>>2]|0;
       $44 = ((($43)) + 36|0);
       $45 = HEAP32[$44>>2]|0;
       __THREW__ = 0;
       $46 = (invoke_ii($45|0,($36|0))|0);
       $47 = __THREW__; __THREW__ = 0;
       $48 = $47&1;
       if ($48) {
        label = 32;
        break L15;
       } else {
        $$0$i$i1$i$i9 = $46;
       }
      } else {
       $49 = HEAP32[$39>>2]|0;
       $$0$i$i1$i$i9 = $49;
      }
      $50 = ($$0$i$i1$i$i9|0)==(-1);
      if ($50) {
       HEAP32[$__e>>2] = 0;
       label = 24;
       break;
      } else {
       if ($162) {
        $163 = $36;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $36;$__a$2 = $__a$0;
        break L15;
       }
      }
     }
    } while(0);
    if ((label|0) == 24) {
     label = 0;
     if ($162) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $163 = 0;
     }
    }
    $51 = HEAP32[$__a_end>>2]|0;
    $52 = HEAP8[$__buf>>0]|0;
    $53 = $52 & 1;
    $54 = ($53<<24>>24)==(0);
    $55 = HEAP32[$20>>2]|0;
    $56 = $52&255;
    $57 = $56 >>> 1;
    $58 = $54 ? $57 : $55;
    $59 = (($__a$0) + ($58)|0);
    $60 = ($51|0)==($59|0);
    if ($60) {
     $61 = $58 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($61|0),0);
     $62 = __THREW__; __THREW__ = 0;
     $63 = $62&1;
     if ($63) {
      label = 32;
      break;
     }
     $64 = HEAP8[$__buf>>0]|0;
     $65 = $64 & 1;
     $66 = ($65<<24>>24)==(0);
     if ($66) {
      $69 = 10;
     } else {
      $67 = HEAP32[$__buf>>2]|0;
      $68 = $67 & -2;
      $phitmp$i23 = (($68) + -1)|0;
      $69 = $phitmp$i23;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($69|0),0);
     $70 = __THREW__; __THREW__ = 0;
     $71 = $70&1;
     if ($71) {
      label = 32;
      break;
     }
     $72 = HEAP8[$__buf>>0]|0;
     $73 = $72 & 1;
     $74 = ($73<<24>>24)==(0);
     $75 = HEAP32[$16>>2]|0;
     $76 = $74 ? $18 : $75;
     $77 = (($76) + ($58)|0);
     HEAP32[$__a_end>>2] = $77;
     $__a$1 = $76;
    } else {
     $__a$1 = $__a$0;
    }
    $82 = ((($$pre$i$i$i)) + 12|0);
    $83 = HEAP32[$82>>2]|0;
    $84 = ((($$pre$i$i$i)) + 16|0);
    $85 = HEAP32[$84>>2]|0;
    $86 = ($83|0)==($85|0);
    if ($86) {
     $87 = HEAP32[$$pre$i$i$i>>2]|0;
     $88 = ((($87)) + 36|0);
     $89 = HEAP32[$88>>2]|0;
     __THREW__ = 0;
     $90 = (invoke_ii($89|0,($$pre$i$i$i|0))|0);
     $91 = __THREW__; __THREW__ = 0;
     $92 = $91&1;
     if ($92) {
      label = 32;
      break;
     } else {
      $$0$i$i = $90;
     }
    } else {
     $93 = HEAP32[$83>>2]|0;
     $$0$i$i = $93;
    }
    $94 = (__ZNSt3__19__num_getIwE17__stage2_int_loopEwiPcRS2_RjwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_Pw($$0$i$i,$$0$i,$__a$1,$__a_end,$__dc,$21,$__grouping,$__g,$__g_end,$__atoms)|0);
    $95 = ($94|0)==(0);
    if (!($95)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $163;$__a$2 = $__a$1;
     break;
    }
    $96 = HEAP32[$82>>2]|0;
    $97 = HEAP32[$84>>2]|0;
    $98 = ($96|0)==($97|0);
    if ($98) {
     $99 = HEAP32[$$pre$i$i$i>>2]|0;
     $100 = ((($99)) + 40|0);
     $101 = HEAP32[$100>>2]|0;
     __THREW__ = 0;
     (invoke_ii($101|0,($$pre$i$i$i|0))|0);
     $102 = __THREW__; __THREW__ = 0;
     $103 = $102&1;
     if ($103) {
      label = 32;
      break;
     } else {
      $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $104 = ((($96)) + 4|0);
     HEAP32[$82>>2] = $104;
     $$pre$i$i$i5 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 32) {
    $78 = ___cxa_find_matching_catch_2()|0;
    $79 = tempRet0;
    $lpad$phi$index = $78;$lpad$phi$index2 = $79;
    break;
   }
   $105 = HEAP8[$__grouping>>0]|0;
   $106 = $105 & 1;
   $107 = ($106<<24>>24)==(0);
   $108 = ((($__grouping)) + 4|0);
   $109 = HEAP32[$108>>2]|0;
   $110 = $105&255;
   $111 = $110 >>> 1;
   $112 = $107 ? $111 : $109;
   $113 = ($112|0)==(0);
   if (!($113)) {
    $114 = HEAP32[$__g_end>>2]|0;
    $115 = $114;
    $116 = $__g;
    $117 = (($115) - ($116))|0;
    $118 = ($117|0)<(160);
    if ($118) {
     $119 = HEAP32[$__dc>>2]|0;
     $120 = ((($114)) + 4|0);
     HEAP32[$__g_end>>2] = $120;
     HEAP32[$114>>2] = $119;
    }
   }
   $121 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $122 = (invoke_iiiii(4870,($__a$2|0),($121|0),($__err|0),($$0$i|0))|0);
   $123 = tempRet0;
   $124 = __THREW__; __THREW__ = 0;
   $125 = $124&1;
   if ($125) {
    label = 33;
   } else {
    $126 = $__v;
    $127 = $126;
    HEAP32[$127>>2] = $122;
    $128 = (($126) + 4)|0;
    $129 = $128;
    HEAP32[$129>>2] = $123;
    $130 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$130,$__err);
    $131 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($131) {
     $164 = 1;
    } else {
     $132 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $133 = HEAP32[$132>>2]|0;
     $134 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $135 = HEAP32[$134>>2]|0;
     $136 = ($133|0)==($135|0);
     if ($136) {
      $137 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $138 = ((($137)) + 36|0);
      $139 = HEAP32[$138>>2]|0;
      __THREW__ = 0;
      $140 = (invoke_ii($139|0,($$pre$i$i$i$lcssa|0))|0);
      $141 = __THREW__; __THREW__ = 0;
      $142 = $141&1;
      if ($142) {
       label = 33;
       break;
      } else {
       $$0$i$i$i$i = $140;
      }
     } else {
      $143 = HEAP32[$133>>2]|0;
      $$0$i$i$i$i = $143;
     }
     $144 = ($$0$i$i$i$i|0)==(-1);
     if ($144) {
      HEAP32[$__b>>2] = 0;
      $164 = 1;
     } else {
      $164 = 0;
     }
    }
    $145 = ($$pre$i3$i$i|0)==(0|0);
    do {
     if ($145) {
      label = 58;
     } else {
      $146 = ((($$pre$i3$i$i)) + 12|0);
      $147 = HEAP32[$146>>2]|0;
      $148 = ((($$pre$i3$i$i)) + 16|0);
      $149 = HEAP32[$148>>2]|0;
      $150 = ($147|0)==($149|0);
      if ($150) {
       $151 = HEAP32[$$pre$i3$i$i>>2]|0;
       $152 = ((($151)) + 36|0);
       $153 = HEAP32[$152>>2]|0;
       __THREW__ = 0;
       $154 = (invoke_ii($153|0,($$pre$i3$i$i|0))|0);
       $155 = __THREW__; __THREW__ = 0;
       $156 = $155&1;
       if ($156) {
        label = 33;
        break L13;
       } else {
        $$0$i$i1$i$i = $154;
       }
      } else {
       $157 = HEAP32[$147>>2]|0;
       $$0$i$i1$i$i = $157;
      }
      $158 = ($$0$i$i1$i$i|0)==(-1);
      if ($158) {
       HEAP32[$__e>>2] = 0;
       label = 58;
       break;
      } else {
       if ($164) {
        break;
       } else {
        label = 60;
        break;
       }
      }
     }
    } while(0);
    if ((label|0) == 58) {
     if ($164) {
      label = 60;
     }
    }
    if ((label|0) == 60) {
     $159 = HEAP32[$__err>>2]|0;
     $160 = $159 | 2;
     HEAP32[$__err>>2] = $160;
    }
    $161 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($161|0);
   }
  }
 } while(0);
 if ((label|0) == 33) {
  $80 = ___cxa_find_matching_catch_2()|0;
  $81 = tempRet0;
  $lpad$phi$index = $80;$lpad$phi$index2 = $81;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRf($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE23__do_get_floating_pointIfEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE23__do_get_floating_pointIfEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i4 = 0, $$0$i$i1$i$i = 0, $$0$i$i1$i$i10 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i6 = 0, $$pre$i3$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0.0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__decimal_point = 0, $__exp = 0, $__g = 0;
 var $__g_end = 0, $__grouping = 0, $__i$0$i$i = 0, $__in_units = 0, $__thousands_sep = 0, $exitcond$i$i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $or$cond = 0, $phitmp$i = 0, $phitmp$i24 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 352|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 208|0;
 $__decimal_point = sp + 200|0;
 $__thousands_sep = sp + 196|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $__in_units = sp + 337|0;
 $__exp = sp + 336|0;
 __ZNSt3__19__num_getIwE19__stage2_float_prepERNS_8ios_baseEPwRwS5_($__grouping,$__iob,$__atoms,$__decimal_point,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 $2 = HEAP8[$__buf>>0]|0;
 $3 = $2 & 1;
 $4 = ($3<<24>>24)==(0);
 if ($4) {
  $7 = 10;
 } else {
  $5 = HEAP32[$__buf>>2]|0;
  $6 = $5 & -2;
  $phitmp$i = (($6) + -1)|0;
  $7 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($7|0),0);
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 L8: do {
  if ($9) {
   label = 29;
  } else {
   $10 = HEAP8[$__buf>>0]|0;
   $11 = $10 & 1;
   $12 = ($11<<24>>24)==(0);
   $13 = ((($__buf)) + 8|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ((($__buf)) + 1|0);
   $16 = $12 ? $15 : $14;
   HEAP32[$__a_end>>2] = $16;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   HEAP8[$__in_units>>0] = 1;
   HEAP8[$__exp>>0] = 69;
   $17 = ((($__buf)) + 4|0);
   $18 = HEAP32[$__decimal_point>>2]|0;
   $19 = HEAP32[$__thousands_sep>>2]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i6 = $$pre;$__a$0 = $16;
   L10: while(1) {
    $20 = ($$pre$i$i$i6|0)==(0|0);
    if ($20) {
     $$pre$i$i$i = 0;$157 = 1;
    } else {
     $21 = ((($$pre$i$i$i6)) + 12|0);
     $22 = HEAP32[$21>>2]|0;
     $23 = ((($$pre$i$i$i6)) + 16|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ($22|0)==($24|0);
     if ($25) {
      $26 = HEAP32[$$pre$i$i$i6>>2]|0;
      $27 = ((($26)) + 36|0);
      $28 = HEAP32[$27>>2]|0;
      __THREW__ = 0;
      $29 = (invoke_ii($28|0,($$pre$i$i$i6|0))|0);
      $30 = __THREW__; __THREW__ = 0;
      $31 = $30&1;
      if ($31) {
       label = 28;
       break;
      } else {
       $$0$i$i$i$i4 = $29;
      }
     } else {
      $32 = HEAP32[$22>>2]|0;
      $$0$i$i$i$i4 = $32;
     }
     $33 = ($$0$i$i$i$i4|0)==(-1);
     if ($33) {
      HEAP32[$__b>>2] = 0;
      $$pre$i$i$i = 0;$157 = 1;
     } else {
      $$pre$i$i$i = $$pre$i$i$i6;$157 = 0;
     }
    }
    $34 = HEAP32[$__e>>2]|0;
    $35 = ($34|0)==(0|0);
    do {
     if ($35) {
      label = 20;
     } else {
      $36 = ((($34)) + 12|0);
      $37 = HEAP32[$36>>2]|0;
      $38 = ((($34)) + 16|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ($37|0)==($39|0);
      if ($40) {
       $41 = HEAP32[$34>>2]|0;
       $42 = ((($41)) + 36|0);
       $43 = HEAP32[$42>>2]|0;
       __THREW__ = 0;
       $44 = (invoke_ii($43|0,($34|0))|0);
       $45 = __THREW__; __THREW__ = 0;
       $46 = $45&1;
       if ($46) {
        label = 28;
        break L10;
       } else {
        $$0$i$i1$i$i10 = $44;
       }
      } else {
       $47 = HEAP32[$37>>2]|0;
       $$0$i$i1$i$i10 = $47;
      }
      $48 = ($$0$i$i1$i$i10|0)==(-1);
      if ($48) {
       HEAP32[$__e>>2] = 0;
       label = 20;
       break;
      } else {
       if ($157) {
        $158 = $34;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $34;$__a$2 = $__a$0;
        break L10;
       }
      }
     }
    } while(0);
    if ((label|0) == 20) {
     label = 0;
     if ($157) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $158 = 0;
     }
    }
    $49 = HEAP32[$__a_end>>2]|0;
    $50 = HEAP8[$__buf>>0]|0;
    $51 = $50 & 1;
    $52 = ($51<<24>>24)==(0);
    $53 = HEAP32[$17>>2]|0;
    $54 = $50&255;
    $55 = $54 >>> 1;
    $56 = $52 ? $55 : $53;
    $57 = (($__a$0) + ($56)|0);
    $58 = ($49|0)==($57|0);
    if ($58) {
     $59 = $56 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($59|0),0);
     $60 = __THREW__; __THREW__ = 0;
     $61 = $60&1;
     if ($61) {
      label = 28;
      break;
     }
     $62 = HEAP8[$__buf>>0]|0;
     $63 = $62 & 1;
     $64 = ($63<<24>>24)==(0);
     if ($64) {
      $67 = 10;
     } else {
      $65 = HEAP32[$__buf>>2]|0;
      $66 = $65 & -2;
      $phitmp$i24 = (($66) + -1)|0;
      $67 = $phitmp$i24;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($67|0),0);
     $68 = __THREW__; __THREW__ = 0;
     $69 = $68&1;
     if ($69) {
      label = 28;
      break;
     }
     $70 = HEAP8[$__buf>>0]|0;
     $71 = $70 & 1;
     $72 = ($71<<24>>24)==(0);
     $73 = HEAP32[$13>>2]|0;
     $74 = $72 ? $15 : $73;
     $75 = (($74) + ($56)|0);
     HEAP32[$__a_end>>2] = $75;
     $__a$1 = $74;
    } else {
     $__a$1 = $__a$0;
    }
    $80 = ((($$pre$i$i$i)) + 12|0);
    $81 = HEAP32[$80>>2]|0;
    $82 = ((($$pre$i$i$i)) + 16|0);
    $83 = HEAP32[$82>>2]|0;
    $84 = ($81|0)==($83|0);
    if ($84) {
     $85 = HEAP32[$$pre$i$i$i>>2]|0;
     $86 = ((($85)) + 36|0);
     $87 = HEAP32[$86>>2]|0;
     __THREW__ = 0;
     $88 = (invoke_ii($87|0,($$pre$i$i$i|0))|0);
     $89 = __THREW__; __THREW__ = 0;
     $90 = $89&1;
     if ($90) {
      label = 28;
      break;
     } else {
      $$0$i$i = $88;
     }
    } else {
     $91 = HEAP32[$81>>2]|0;
     $$0$i$i = $91;
    }
    $92 = (__ZNSt3__19__num_getIwE19__stage2_float_loopEwRbRcPcRS4_wwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSE_RjPw($$0$i$i,$__in_units,$__exp,$__a$1,$__a_end,$18,$19,$__grouping,$__g,$__g_end,$__dc,$__atoms)|0);
    $93 = ($92|0)==(0);
    if (!($93)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $158;$__a$2 = $__a$1;
     break;
    }
    $94 = HEAP32[$80>>2]|0;
    $95 = HEAP32[$82>>2]|0;
    $96 = ($94|0)==($95|0);
    if ($96) {
     $97 = HEAP32[$$pre$i$i$i>>2]|0;
     $98 = ((($97)) + 40|0);
     $99 = HEAP32[$98>>2]|0;
     __THREW__ = 0;
     (invoke_ii($99|0,($$pre$i$i$i|0))|0);
     $100 = __THREW__; __THREW__ = 0;
     $101 = $100&1;
     if ($101) {
      label = 28;
      break;
     } else {
      $$pre$i$i$i6 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $102 = ((($94)) + 4|0);
     HEAP32[$80>>2] = $102;
     $$pre$i$i$i6 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 28) {
    $76 = ___cxa_find_matching_catch_2()|0;
    $77 = tempRet0;
    $lpad$phi$index = $76;$lpad$phi$index2 = $77;
    break;
   }
   $103 = HEAP8[$__grouping>>0]|0;
   $104 = $103 & 1;
   $105 = ($104<<24>>24)==(0);
   $106 = ((($__grouping)) + 4|0);
   $107 = HEAP32[$106>>2]|0;
   $108 = $103&255;
   $109 = $108 >>> 1;
   $110 = $105 ? $109 : $107;
   $111 = ($110|0)==(0);
   $112 = HEAP8[$__in_units>>0]|0;
   $113 = ($112<<24>>24)==(0);
   $or$cond = $113 | $111;
   if (!($or$cond)) {
    $114 = HEAP32[$__g_end>>2]|0;
    $115 = $114;
    $116 = $__g;
    $117 = (($115) - ($116))|0;
    $118 = ($117|0)<(160);
    if ($118) {
     $119 = HEAP32[$__dc>>2]|0;
     $120 = ((($114)) + 4|0);
     HEAP32[$__g_end>>2] = $120;
     HEAP32[$114>>2] = $119;
    }
   }
   $121 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $122 = (+invoke_diii(4871,($__a$2|0),($121|0),($__err|0)));
   $123 = __THREW__; __THREW__ = 0;
   $124 = $123&1;
   if ($124) {
    label = 29;
   } else {
    HEAPF32[$__v>>2] = $122;
    $125 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$125,$__err);
    $126 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($126) {
     $159 = 1;
    } else {
     $127 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $128 = HEAP32[$127>>2]|0;
     $129 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $130 = HEAP32[$129>>2]|0;
     $131 = ($128|0)==($130|0);
     if ($131) {
      $132 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $133 = ((($132)) + 36|0);
      $134 = HEAP32[$133>>2]|0;
      __THREW__ = 0;
      $135 = (invoke_ii($134|0,($$pre$i$i$i$lcssa|0))|0);
      $136 = __THREW__; __THREW__ = 0;
      $137 = $136&1;
      if ($137) {
       label = 29;
       break;
      } else {
       $$0$i$i$i$i = $135;
      }
     } else {
      $138 = HEAP32[$128>>2]|0;
      $$0$i$i$i$i = $138;
     }
     $139 = ($$0$i$i$i$i|0)==(-1);
     if ($139) {
      HEAP32[$__b>>2] = 0;
      $159 = 1;
     } else {
      $159 = 0;
     }
    }
    $140 = ($$pre$i3$i$i|0)==(0|0);
    do {
     if ($140) {
      label = 54;
     } else {
      $141 = ((($$pre$i3$i$i)) + 12|0);
      $142 = HEAP32[$141>>2]|0;
      $143 = ((($$pre$i3$i$i)) + 16|0);
      $144 = HEAP32[$143>>2]|0;
      $145 = ($142|0)==($144|0);
      if ($145) {
       $146 = HEAP32[$$pre$i3$i$i>>2]|0;
       $147 = ((($146)) + 36|0);
       $148 = HEAP32[$147>>2]|0;
       __THREW__ = 0;
       $149 = (invoke_ii($148|0,($$pre$i3$i$i|0))|0);
       $150 = __THREW__; __THREW__ = 0;
       $151 = $150&1;
       if ($151) {
        label = 29;
        break L8;
       } else {
        $$0$i$i1$i$i = $149;
       }
      } else {
       $152 = HEAP32[$142>>2]|0;
       $$0$i$i1$i$i = $152;
      }
      $153 = ($$0$i$i1$i$i|0)==(-1);
      if ($153) {
       HEAP32[$__e>>2] = 0;
       label = 54;
       break;
      } else {
       if ($159) {
        break;
       } else {
        label = 56;
        break;
       }
      }
     }
    } while(0);
    if ((label|0) == 54) {
     if ($159) {
      label = 56;
     }
    }
    if ((label|0) == 56) {
     $154 = HEAP32[$__err>>2]|0;
     $155 = $154 | 2;
     HEAP32[$__err>>2] = $155;
    }
    $156 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($156|0);
   }
  }
 } while(0);
 if ((label|0) == 29) {
  $78 = ___cxa_find_matching_catch_2()|0;
  $79 = tempRet0;
  $lpad$phi$index = $78;$lpad$phi$index2 = $79;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__19__num_getIwE19__stage2_float_prepERNS_8ios_baseEPwRwS5_($agg$result,$__iob,$__atoms,$__decimal_point,$__thousands_sep) {
 $agg$result = $agg$result|0;
 $__iob = $__iob|0;
 $__atoms = $__atoms|0;
 $__decimal_point = $__decimal_point|0;
 $__thousands_sep = $__thousands_sep|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__loc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__loc = sp;
 $0 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$__loc>>2] = $0;
 __THREW__ = 0;
 $1 = (invoke_iii(3374,($__loc|0),(1207608|0))|0);
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if (!($3)) {
  $4 = HEAP32[$1>>2]|0;
  $5 = ((($4)) + 48|0);
  $6 = HEAP32[$5>>2]|0;
  __THREW__ = 0;
  (invoke_iiiii($6|0,($1|0),(620486|0),((620518)|0),($__atoms|0))|0);
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if (!($8)) {
   __THREW__ = 0;
   $9 = (invoke_iii(3374,($__loc|0),(1207616|0))|0);
   $10 = __THREW__; __THREW__ = 0;
   $11 = $10&1;
   if (!($11)) {
    $12 = HEAP32[$9>>2]|0;
    $13 = ((($12)) + 12|0);
    $14 = HEAP32[$13>>2]|0;
    __THREW__ = 0;
    $15 = (invoke_ii($14|0,($9|0))|0);
    $16 = __THREW__; __THREW__ = 0;
    $17 = $16&1;
    if (!($17)) {
     HEAP32[$__decimal_point>>2] = $15;
     $18 = HEAP32[$9>>2]|0;
     $19 = ((($18)) + 16|0);
     $20 = HEAP32[$19>>2]|0;
     __THREW__ = 0;
     $21 = (invoke_ii($20|0,($9|0))|0);
     $22 = __THREW__; __THREW__ = 0;
     $23 = $22&1;
     if (!($23)) {
      HEAP32[$__thousands_sep>>2] = $21;
      $24 = HEAP32[$9>>2]|0;
      $25 = ((($24)) + 20|0);
      $26 = HEAP32[$25>>2]|0;
      __THREW__ = 0;
      invoke_vii($26|0,($agg$result|0),($9|0));
      $27 = __THREW__; __THREW__ = 0;
      $28 = $27&1;
      if (!($28)) {
       (__ZNSt3__114__shared_count16__release_sharedEv($0)|0);
       STACKTOP = sp;return;
      }
     }
    }
   }
  }
 }
 $29 = ___cxa_find_matching_catch_2()|0;
 $30 = tempRet0;
 (__ZNSt3__114__shared_count16__release_sharedEv($0)|0);
 ___resumeException($29|0);
 // unreachable;
}
function __ZNSt3__19__num_getIwE19__stage2_float_loopEwRbRcPcRS4_wwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSE_RjPw($__ct,$__in_units,$__exp,$__a,$__a_end,$__decimal_point,$__thousands_sep,$__grouping,$__g,$__g_end,$__dc,$__atoms) {
 $__ct = $__ct|0;
 $__in_units = $__in_units|0;
 $__exp = $__exp|0;
 $__a = $__a|0;
 $__a_end = $__a_end|0;
 $__decimal_point = $__decimal_point|0;
 $__thousands_sep = $__thousands_sep|0;
 $__grouping = $__grouping|0;
 $__g = $__g|0;
 $__g_end = $__g_end|0;
 $__dc = $__dc|0;
 $__atoms = $__atoms|0;
 var $$0$i = 0, $$0$lcssa$i = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__ct|0)==($__decimal_point|0);
 L1: do {
  if ($0) {
   $1 = HEAP8[$__in_units>>0]|0;
   $2 = ($1<<24>>24)==(0);
   if ($2) {
    $$2 = -1;
   } else {
    HEAP8[$__in_units>>0] = 0;
    $3 = HEAP32[$__a_end>>2]|0;
    $4 = ((($3)) + 1|0);
    HEAP32[$__a_end>>2] = $4;
    HEAP8[$3>>0] = 46;
    $5 = HEAP8[$__grouping>>0]|0;
    $6 = $5 & 1;
    $7 = ($6<<24>>24)==(0);
    $8 = ((($__grouping)) + 4|0);
    $9 = HEAP32[$8>>2]|0;
    $10 = $5&255;
    $11 = $10 >>> 1;
    $12 = $7 ? $11 : $9;
    $13 = ($12|0)==(0);
    if ($13) {
     $$2 = 0;
    } else {
     $14 = HEAP32[$__g_end>>2]|0;
     $15 = $14;
     $16 = $__g;
     $17 = (($15) - ($16))|0;
     $18 = ($17|0)<(160);
     if ($18) {
      $19 = HEAP32[$__dc>>2]|0;
      $20 = ((($14)) + 4|0);
      HEAP32[$__g_end>>2] = $20;
      HEAP32[$14>>2] = $19;
      $$2 = 0;
     } else {
      $$2 = 0;
     }
    }
   }
  } else {
   $21 = ($__ct|0)==($__thousands_sep|0);
   if ($21) {
    $22 = HEAP8[$__grouping>>0]|0;
    $23 = $22 & 1;
    $24 = ($23<<24>>24)==(0);
    $25 = ((($__grouping)) + 4|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = $22&255;
    $28 = $27 >>> 1;
    $29 = $24 ? $28 : $26;
    $30 = ($29|0)==(0);
    if (!($30)) {
     $31 = HEAP8[$__in_units>>0]|0;
     $32 = ($31<<24>>24)==(0);
     if ($32) {
      $$2 = -1;
      break;
     }
     $33 = HEAP32[$__g_end>>2]|0;
     $34 = $33;
     $35 = $__g;
     $36 = (($34) - ($35))|0;
     $37 = ($36|0)<(160);
     if (!($37)) {
      $$2 = 0;
      break;
     }
     $38 = HEAP32[$__dc>>2]|0;
     $39 = ((($33)) + 4|0);
     HEAP32[$__g_end>>2] = $39;
     HEAP32[$33>>2] = $38;
     HEAP32[$__dc>>2] = 0;
     $$2 = 0;
     break;
    }
   }
   $40 = ((($__atoms)) + 128|0);
   $$0$i = $__atoms;
   while(1) {
    $41 = ($$0$i|0)==($40|0);
    if ($41) {
     $$0$lcssa$i = $40;
     break;
    }
    $42 = HEAP32[$$0$i>>2]|0;
    $43 = ($42|0)==($__ct|0);
    if ($43) {
     $$0$lcssa$i = $$0$i;
     break;
    }
    $44 = ((($$0$i)) + 4|0);
    $$0$i = $44;
   }
   $45 = $$0$lcssa$i;
   $46 = $__atoms;
   $47 = (($45) - ($46))|0;
   $48 = $47 >> 2;
   $49 = ($47|0)>(124);
   if ($49) {
    $$2 = -1;
   } else {
    $50 = (620486 + ($48)|0);
    $51 = HEAP8[$50>>0]|0;
    switch ($48|0) {
    case 24: case 25:  {
     $52 = HEAP32[$__a_end>>2]|0;
     $53 = ($52|0)==($__a|0);
     if (!($53)) {
      $54 = ((($52)) + -1|0);
      $55 = HEAP8[$54>>0]|0;
      $56 = $55&255;
      $57 = $56 & 95;
      $58 = HEAP8[$__exp>>0]|0;
      $59 = $58&255;
      $60 = $59 & 127;
      $61 = ($57|0)==($60|0);
      if (!($61)) {
       $$2 = -1;
       break L1;
      }
     }
     $62 = ((($52)) + 1|0);
     HEAP32[$__a_end>>2] = $62;
     HEAP8[$52>>0] = $51;
     $$2 = 0;
     break L1;
     break;
    }
    case 23: case 22:  {
     HEAP8[$__exp>>0] = 80;
     break;
    }
    default: {
     $63 = $51&255;
     $64 = $63 & 95;
     $65 = HEAP8[$__exp>>0]|0;
     $66 = $65 << 24 >> 24;
     $67 = ($64|0)==($66|0);
     if ($67) {
      $68 = $64 | 128;
      $69 = $68&255;
      HEAP8[$__exp>>0] = $69;
      $70 = HEAP8[$__in_units>>0]|0;
      $71 = ($70<<24>>24)==(0);
      if (!($71)) {
       HEAP8[$__in_units>>0] = 0;
       $72 = HEAP8[$__grouping>>0]|0;
       $73 = $72 & 1;
       $74 = ($73<<24>>24)==(0);
       $75 = ((($__grouping)) + 4|0);
       $76 = HEAP32[$75>>2]|0;
       $77 = $72&255;
       $78 = $77 >>> 1;
       $79 = $74 ? $78 : $76;
       $80 = ($79|0)==(0);
       if (!($80)) {
        $81 = HEAP32[$__g_end>>2]|0;
        $82 = $81;
        $83 = $__g;
        $84 = (($82) - ($83))|0;
        $85 = ($84|0)<(160);
        if ($85) {
         $86 = HEAP32[$__dc>>2]|0;
         $87 = ((($81)) + 4|0);
         HEAP32[$__g_end>>2] = $87;
         HEAP32[$81>>2] = $86;
        }
       }
      }
     }
    }
    }
    $88 = HEAP32[$__a_end>>2]|0;
    $89 = ((($88)) + 1|0);
    HEAP32[$__a_end>>2] = $89;
    HEAP8[$88>>0] = $51;
    $90 = ($47|0)>(84);
    if ($90) {
     $$2 = 0;
    } else {
     $91 = HEAP32[$__dc>>2]|0;
     $92 = (($91) + 1)|0;
     HEAP32[$__dc>>2] = $92;
     $$2 = 0;
    }
   }
  }
 } while(0);
 return ($$2|0);
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRd($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE23__do_get_floating_pointIdEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE23__do_get_floating_pointIdEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i4 = 0, $$0$i$i1$i$i = 0, $$0$i$i1$i$i10 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i6 = 0, $$pre$i3$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0.0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__decimal_point = 0, $__exp = 0, $__g = 0;
 var $__g_end = 0, $__grouping = 0, $__i$0$i$i = 0, $__in_units = 0, $__thousands_sep = 0, $exitcond$i$i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $or$cond = 0, $phitmp$i = 0, $phitmp$i24 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 352|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 208|0;
 $__decimal_point = sp + 200|0;
 $__thousands_sep = sp + 196|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $__in_units = sp + 337|0;
 $__exp = sp + 336|0;
 __ZNSt3__19__num_getIwE19__stage2_float_prepERNS_8ios_baseEPwRwS5_($__grouping,$__iob,$__atoms,$__decimal_point,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 $2 = HEAP8[$__buf>>0]|0;
 $3 = $2 & 1;
 $4 = ($3<<24>>24)==(0);
 if ($4) {
  $7 = 10;
 } else {
  $5 = HEAP32[$__buf>>2]|0;
  $6 = $5 & -2;
  $phitmp$i = (($6) + -1)|0;
  $7 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($7|0),0);
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 L8: do {
  if ($9) {
   label = 29;
  } else {
   $10 = HEAP8[$__buf>>0]|0;
   $11 = $10 & 1;
   $12 = ($11<<24>>24)==(0);
   $13 = ((($__buf)) + 8|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ((($__buf)) + 1|0);
   $16 = $12 ? $15 : $14;
   HEAP32[$__a_end>>2] = $16;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   HEAP8[$__in_units>>0] = 1;
   HEAP8[$__exp>>0] = 69;
   $17 = ((($__buf)) + 4|0);
   $18 = HEAP32[$__decimal_point>>2]|0;
   $19 = HEAP32[$__thousands_sep>>2]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i6 = $$pre;$__a$0 = $16;
   L10: while(1) {
    $20 = ($$pre$i$i$i6|0)==(0|0);
    if ($20) {
     $$pre$i$i$i = 0;$157 = 1;
    } else {
     $21 = ((($$pre$i$i$i6)) + 12|0);
     $22 = HEAP32[$21>>2]|0;
     $23 = ((($$pre$i$i$i6)) + 16|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ($22|0)==($24|0);
     if ($25) {
      $26 = HEAP32[$$pre$i$i$i6>>2]|0;
      $27 = ((($26)) + 36|0);
      $28 = HEAP32[$27>>2]|0;
      __THREW__ = 0;
      $29 = (invoke_ii($28|0,($$pre$i$i$i6|0))|0);
      $30 = __THREW__; __THREW__ = 0;
      $31 = $30&1;
      if ($31) {
       label = 28;
       break;
      } else {
       $$0$i$i$i$i4 = $29;
      }
     } else {
      $32 = HEAP32[$22>>2]|0;
      $$0$i$i$i$i4 = $32;
     }
     $33 = ($$0$i$i$i$i4|0)==(-1);
     if ($33) {
      HEAP32[$__b>>2] = 0;
      $$pre$i$i$i = 0;$157 = 1;
     } else {
      $$pre$i$i$i = $$pre$i$i$i6;$157 = 0;
     }
    }
    $34 = HEAP32[$__e>>2]|0;
    $35 = ($34|0)==(0|0);
    do {
     if ($35) {
      label = 20;
     } else {
      $36 = ((($34)) + 12|0);
      $37 = HEAP32[$36>>2]|0;
      $38 = ((($34)) + 16|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ($37|0)==($39|0);
      if ($40) {
       $41 = HEAP32[$34>>2]|0;
       $42 = ((($41)) + 36|0);
       $43 = HEAP32[$42>>2]|0;
       __THREW__ = 0;
       $44 = (invoke_ii($43|0,($34|0))|0);
       $45 = __THREW__; __THREW__ = 0;
       $46 = $45&1;
       if ($46) {
        label = 28;
        break L10;
       } else {
        $$0$i$i1$i$i10 = $44;
       }
      } else {
       $47 = HEAP32[$37>>2]|0;
       $$0$i$i1$i$i10 = $47;
      }
      $48 = ($$0$i$i1$i$i10|0)==(-1);
      if ($48) {
       HEAP32[$__e>>2] = 0;
       label = 20;
       break;
      } else {
       if ($157) {
        $158 = $34;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $34;$__a$2 = $__a$0;
        break L10;
       }
      }
     }
    } while(0);
    if ((label|0) == 20) {
     label = 0;
     if ($157) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $158 = 0;
     }
    }
    $49 = HEAP32[$__a_end>>2]|0;
    $50 = HEAP8[$__buf>>0]|0;
    $51 = $50 & 1;
    $52 = ($51<<24>>24)==(0);
    $53 = HEAP32[$17>>2]|0;
    $54 = $50&255;
    $55 = $54 >>> 1;
    $56 = $52 ? $55 : $53;
    $57 = (($__a$0) + ($56)|0);
    $58 = ($49|0)==($57|0);
    if ($58) {
     $59 = $56 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($59|0),0);
     $60 = __THREW__; __THREW__ = 0;
     $61 = $60&1;
     if ($61) {
      label = 28;
      break;
     }
     $62 = HEAP8[$__buf>>0]|0;
     $63 = $62 & 1;
     $64 = ($63<<24>>24)==(0);
     if ($64) {
      $67 = 10;
     } else {
      $65 = HEAP32[$__buf>>2]|0;
      $66 = $65 & -2;
      $phitmp$i24 = (($66) + -1)|0;
      $67 = $phitmp$i24;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($67|0),0);
     $68 = __THREW__; __THREW__ = 0;
     $69 = $68&1;
     if ($69) {
      label = 28;
      break;
     }
     $70 = HEAP8[$__buf>>0]|0;
     $71 = $70 & 1;
     $72 = ($71<<24>>24)==(0);
     $73 = HEAP32[$13>>2]|0;
     $74 = $72 ? $15 : $73;
     $75 = (($74) + ($56)|0);
     HEAP32[$__a_end>>2] = $75;
     $__a$1 = $74;
    } else {
     $__a$1 = $__a$0;
    }
    $80 = ((($$pre$i$i$i)) + 12|0);
    $81 = HEAP32[$80>>2]|0;
    $82 = ((($$pre$i$i$i)) + 16|0);
    $83 = HEAP32[$82>>2]|0;
    $84 = ($81|0)==($83|0);
    if ($84) {
     $85 = HEAP32[$$pre$i$i$i>>2]|0;
     $86 = ((($85)) + 36|0);
     $87 = HEAP32[$86>>2]|0;
     __THREW__ = 0;
     $88 = (invoke_ii($87|0,($$pre$i$i$i|0))|0);
     $89 = __THREW__; __THREW__ = 0;
     $90 = $89&1;
     if ($90) {
      label = 28;
      break;
     } else {
      $$0$i$i = $88;
     }
    } else {
     $91 = HEAP32[$81>>2]|0;
     $$0$i$i = $91;
    }
    $92 = (__ZNSt3__19__num_getIwE19__stage2_float_loopEwRbRcPcRS4_wwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSE_RjPw($$0$i$i,$__in_units,$__exp,$__a$1,$__a_end,$18,$19,$__grouping,$__g,$__g_end,$__dc,$__atoms)|0);
    $93 = ($92|0)==(0);
    if (!($93)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $158;$__a$2 = $__a$1;
     break;
    }
    $94 = HEAP32[$80>>2]|0;
    $95 = HEAP32[$82>>2]|0;
    $96 = ($94|0)==($95|0);
    if ($96) {
     $97 = HEAP32[$$pre$i$i$i>>2]|0;
     $98 = ((($97)) + 40|0);
     $99 = HEAP32[$98>>2]|0;
     __THREW__ = 0;
     (invoke_ii($99|0,($$pre$i$i$i|0))|0);
     $100 = __THREW__; __THREW__ = 0;
     $101 = $100&1;
     if ($101) {
      label = 28;
      break;
     } else {
      $$pre$i$i$i6 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $102 = ((($94)) + 4|0);
     HEAP32[$80>>2] = $102;
     $$pre$i$i$i6 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 28) {
    $76 = ___cxa_find_matching_catch_2()|0;
    $77 = tempRet0;
    $lpad$phi$index = $76;$lpad$phi$index2 = $77;
    break;
   }
   $103 = HEAP8[$__grouping>>0]|0;
   $104 = $103 & 1;
   $105 = ($104<<24>>24)==(0);
   $106 = ((($__grouping)) + 4|0);
   $107 = HEAP32[$106>>2]|0;
   $108 = $103&255;
   $109 = $108 >>> 1;
   $110 = $105 ? $109 : $107;
   $111 = ($110|0)==(0);
   $112 = HEAP8[$__in_units>>0]|0;
   $113 = ($112<<24>>24)==(0);
   $or$cond = $113 | $111;
   if (!($or$cond)) {
    $114 = HEAP32[$__g_end>>2]|0;
    $115 = $114;
    $116 = $__g;
    $117 = (($115) - ($116))|0;
    $118 = ($117|0)<(160);
    if ($118) {
     $119 = HEAP32[$__dc>>2]|0;
     $120 = ((($114)) + 4|0);
     HEAP32[$__g_end>>2] = $120;
     HEAP32[$114>>2] = $119;
    }
   }
   $121 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $122 = (+invoke_diii(4872,($__a$2|0),($121|0),($__err|0)));
   $123 = __THREW__; __THREW__ = 0;
   $124 = $123&1;
   if ($124) {
    label = 29;
   } else {
    HEAPF64[$__v>>3] = $122;
    $125 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$125,$__err);
    $126 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($126) {
     $159 = 1;
    } else {
     $127 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $128 = HEAP32[$127>>2]|0;
     $129 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $130 = HEAP32[$129>>2]|0;
     $131 = ($128|0)==($130|0);
     if ($131) {
      $132 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $133 = ((($132)) + 36|0);
      $134 = HEAP32[$133>>2]|0;
      __THREW__ = 0;
      $135 = (invoke_ii($134|0,($$pre$i$i$i$lcssa|0))|0);
      $136 = __THREW__; __THREW__ = 0;
      $137 = $136&1;
      if ($137) {
       label = 29;
       break;
      } else {
       $$0$i$i$i$i = $135;
      }
     } else {
      $138 = HEAP32[$128>>2]|0;
      $$0$i$i$i$i = $138;
     }
     $139 = ($$0$i$i$i$i|0)==(-1);
     if ($139) {
      HEAP32[$__b>>2] = 0;
      $159 = 1;
     } else {
      $159 = 0;
     }
    }
    $140 = ($$pre$i3$i$i|0)==(0|0);
    do {
     if ($140) {
      label = 54;
     } else {
      $141 = ((($$pre$i3$i$i)) + 12|0);
      $142 = HEAP32[$141>>2]|0;
      $143 = ((($$pre$i3$i$i)) + 16|0);
      $144 = HEAP32[$143>>2]|0;
      $145 = ($142|0)==($144|0);
      if ($145) {
       $146 = HEAP32[$$pre$i3$i$i>>2]|0;
       $147 = ((($146)) + 36|0);
       $148 = HEAP32[$147>>2]|0;
       __THREW__ = 0;
       $149 = (invoke_ii($148|0,($$pre$i3$i$i|0))|0);
       $150 = __THREW__; __THREW__ = 0;
       $151 = $150&1;
       if ($151) {
        label = 29;
        break L8;
       } else {
        $$0$i$i1$i$i = $149;
       }
      } else {
       $152 = HEAP32[$142>>2]|0;
       $$0$i$i1$i$i = $152;
      }
      $153 = ($$0$i$i1$i$i|0)==(-1);
      if ($153) {
       HEAP32[$__e>>2] = 0;
       label = 54;
       break;
      } else {
       if ($159) {
        break;
       } else {
        label = 56;
        break;
       }
      }
     }
    } while(0);
    if ((label|0) == 54) {
     if ($159) {
      label = 56;
     }
    }
    if ((label|0) == 56) {
     $154 = HEAP32[$__err>>2]|0;
     $155 = $154 | 2;
     HEAP32[$__err>>2] = $155;
    }
    $156 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($156|0);
   }
  }
 } while(0);
 if ((label|0) == 29) {
  $78 = ___cxa_find_matching_catch_2()|0;
  $79 = tempRet0;
  $lpad$phi$index = $78;$lpad$phi$index2 = $79;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRe($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE23__do_get_floating_pointIeEES4_S4_S4_RNS_8ios_baseERjRT_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__v)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE23__do_get_floating_pointIeEES4_S4_S4_RNS_8ios_baseERjRT_($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i4 = 0, $$0$i$i1$i$i = 0, $$0$i$i1$i$i10 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i6 = 0, $$pre$i3$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0.0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__decimal_point = 0, $__exp = 0, $__g = 0;
 var $__g_end = 0, $__grouping = 0, $__i$0$i$i = 0, $__in_units = 0, $__thousands_sep = 0, $exitcond$i$i = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $or$cond = 0, $phitmp$i = 0, $phitmp$i24 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 352|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__atoms = sp + 208|0;
 $__decimal_point = sp + 200|0;
 $__thousands_sep = sp + 196|0;
 $__grouping = sp + 184|0;
 $__buf = sp + 172|0;
 $__a_end = sp + 168|0;
 $__g = sp + 8|0;
 $__g_end = sp + 4|0;
 $__dc = sp;
 $__in_units = sp + 337|0;
 $__exp = sp + 336|0;
 __ZNSt3__19__num_getIwE19__stage2_float_prepERNS_8ios_baseEPwRwS5_($__grouping,$__iob,$__atoms,$__decimal_point,$__thousands_sep);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($__buf) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 $2 = HEAP8[$__buf>>0]|0;
 $3 = $2 & 1;
 $4 = ($3<<24>>24)==(0);
 if ($4) {
  $7 = 10;
 } else {
  $5 = HEAP32[$__buf>>2]|0;
  $6 = $5 & -2;
  $phitmp$i = (($6) + -1)|0;
  $7 = $phitmp$i;
 }
 __THREW__ = 0;
 invoke_viii(3371,($__buf|0),($7|0),0);
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 L8: do {
  if ($9) {
   label = 29;
  } else {
   $10 = HEAP8[$__buf>>0]|0;
   $11 = $10 & 1;
   $12 = ($11<<24>>24)==(0);
   $13 = ((($__buf)) + 8|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ((($__buf)) + 1|0);
   $16 = $12 ? $15 : $14;
   HEAP32[$__a_end>>2] = $16;
   HEAP32[$__g_end>>2] = $__g;
   HEAP32[$__dc>>2] = 0;
   HEAP8[$__in_units>>0] = 1;
   HEAP8[$__exp>>0] = 69;
   $17 = ((($__buf)) + 4|0);
   $18 = HEAP32[$__decimal_point>>2]|0;
   $19 = HEAP32[$__thousands_sep>>2]|0;
   $$pre = HEAP32[$__b>>2]|0;
   $$pre$i$i$i6 = $$pre;$__a$0 = $16;
   L10: while(1) {
    $20 = ($$pre$i$i$i6|0)==(0|0);
    if ($20) {
     $$pre$i$i$i = 0;$157 = 1;
    } else {
     $21 = ((($$pre$i$i$i6)) + 12|0);
     $22 = HEAP32[$21>>2]|0;
     $23 = ((($$pre$i$i$i6)) + 16|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ($22|0)==($24|0);
     if ($25) {
      $26 = HEAP32[$$pre$i$i$i6>>2]|0;
      $27 = ((($26)) + 36|0);
      $28 = HEAP32[$27>>2]|0;
      __THREW__ = 0;
      $29 = (invoke_ii($28|0,($$pre$i$i$i6|0))|0);
      $30 = __THREW__; __THREW__ = 0;
      $31 = $30&1;
      if ($31) {
       label = 28;
       break;
      } else {
       $$0$i$i$i$i4 = $29;
      }
     } else {
      $32 = HEAP32[$22>>2]|0;
      $$0$i$i$i$i4 = $32;
     }
     $33 = ($$0$i$i$i$i4|0)==(-1);
     if ($33) {
      HEAP32[$__b>>2] = 0;
      $$pre$i$i$i = 0;$157 = 1;
     } else {
      $$pre$i$i$i = $$pre$i$i$i6;$157 = 0;
     }
    }
    $34 = HEAP32[$__e>>2]|0;
    $35 = ($34|0)==(0|0);
    do {
     if ($35) {
      label = 20;
     } else {
      $36 = ((($34)) + 12|0);
      $37 = HEAP32[$36>>2]|0;
      $38 = ((($34)) + 16|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ($37|0)==($39|0);
      if ($40) {
       $41 = HEAP32[$34>>2]|0;
       $42 = ((($41)) + 36|0);
       $43 = HEAP32[$42>>2]|0;
       __THREW__ = 0;
       $44 = (invoke_ii($43|0,($34|0))|0);
       $45 = __THREW__; __THREW__ = 0;
       $46 = $45&1;
       if ($46) {
        label = 28;
        break L10;
       } else {
        $$0$i$i1$i$i10 = $44;
       }
      } else {
       $47 = HEAP32[$37>>2]|0;
       $$0$i$i1$i$i10 = $47;
      }
      $48 = ($$0$i$i1$i$i10|0)==(-1);
      if ($48) {
       HEAP32[$__e>>2] = 0;
       label = 20;
       break;
      } else {
       if ($157) {
        $158 = $34;
        break;
       } else {
        $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $34;$__a$2 = $__a$0;
        break L10;
       }
      }
     }
    } while(0);
    if ((label|0) == 20) {
     label = 0;
     if ($157) {
      $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = 0;$__a$2 = $__a$0;
      break;
     } else {
      $158 = 0;
     }
    }
    $49 = HEAP32[$__a_end>>2]|0;
    $50 = HEAP8[$__buf>>0]|0;
    $51 = $50 & 1;
    $52 = ($51<<24>>24)==(0);
    $53 = HEAP32[$17>>2]|0;
    $54 = $50&255;
    $55 = $54 >>> 1;
    $56 = $52 ? $55 : $53;
    $57 = (($__a$0) + ($56)|0);
    $58 = ($49|0)==($57|0);
    if ($58) {
     $59 = $56 << 1;
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($59|0),0);
     $60 = __THREW__; __THREW__ = 0;
     $61 = $60&1;
     if ($61) {
      label = 28;
      break;
     }
     $62 = HEAP8[$__buf>>0]|0;
     $63 = $62 & 1;
     $64 = ($63<<24>>24)==(0);
     if ($64) {
      $67 = 10;
     } else {
      $65 = HEAP32[$__buf>>2]|0;
      $66 = $65 & -2;
      $phitmp$i24 = (($66) + -1)|0;
      $67 = $phitmp$i24;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($67|0),0);
     $68 = __THREW__; __THREW__ = 0;
     $69 = $68&1;
     if ($69) {
      label = 28;
      break;
     }
     $70 = HEAP8[$__buf>>0]|0;
     $71 = $70 & 1;
     $72 = ($71<<24>>24)==(0);
     $73 = HEAP32[$13>>2]|0;
     $74 = $72 ? $15 : $73;
     $75 = (($74) + ($56)|0);
     HEAP32[$__a_end>>2] = $75;
     $__a$1 = $74;
    } else {
     $__a$1 = $__a$0;
    }
    $80 = ((($$pre$i$i$i)) + 12|0);
    $81 = HEAP32[$80>>2]|0;
    $82 = ((($$pre$i$i$i)) + 16|0);
    $83 = HEAP32[$82>>2]|0;
    $84 = ($81|0)==($83|0);
    if ($84) {
     $85 = HEAP32[$$pre$i$i$i>>2]|0;
     $86 = ((($85)) + 36|0);
     $87 = HEAP32[$86>>2]|0;
     __THREW__ = 0;
     $88 = (invoke_ii($87|0,($$pre$i$i$i|0))|0);
     $89 = __THREW__; __THREW__ = 0;
     $90 = $89&1;
     if ($90) {
      label = 28;
      break;
     } else {
      $$0$i$i = $88;
     }
    } else {
     $91 = HEAP32[$81>>2]|0;
     $$0$i$i = $91;
    }
    $92 = (__ZNSt3__19__num_getIwE19__stage2_float_loopEwRbRcPcRS4_wwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSE_RjPw($$0$i$i,$__in_units,$__exp,$__a$1,$__a_end,$18,$19,$__grouping,$__g,$__g_end,$__dc,$__atoms)|0);
    $93 = ($92|0)==(0);
    if (!($93)) {
     $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $158;$__a$2 = $__a$1;
     break;
    }
    $94 = HEAP32[$80>>2]|0;
    $95 = HEAP32[$82>>2]|0;
    $96 = ($94|0)==($95|0);
    if ($96) {
     $97 = HEAP32[$$pre$i$i$i>>2]|0;
     $98 = ((($97)) + 40|0);
     $99 = HEAP32[$98>>2]|0;
     __THREW__ = 0;
     (invoke_ii($99|0,($$pre$i$i$i|0))|0);
     $100 = __THREW__; __THREW__ = 0;
     $101 = $100&1;
     if ($101) {
      label = 28;
      break;
     } else {
      $$pre$i$i$i6 = $$pre$i$i$i;$__a$0 = $__a$1;
      continue;
     }
    } else {
     $102 = ((($94)) + 4|0);
     HEAP32[$80>>2] = $102;
     $$pre$i$i$i6 = $$pre$i$i$i;$__a$0 = $__a$1;
     continue;
    }
   }
   if ((label|0) == 28) {
    $76 = ___cxa_find_matching_catch_2()|0;
    $77 = tempRet0;
    $lpad$phi$index = $76;$lpad$phi$index2 = $77;
    break;
   }
   $103 = HEAP8[$__grouping>>0]|0;
   $104 = $103 & 1;
   $105 = ($104<<24>>24)==(0);
   $106 = ((($__grouping)) + 4|0);
   $107 = HEAP32[$106>>2]|0;
   $108 = $103&255;
   $109 = $108 >>> 1;
   $110 = $105 ? $109 : $107;
   $111 = ($110|0)==(0);
   $112 = HEAP8[$__in_units>>0]|0;
   $113 = ($112<<24>>24)==(0);
   $or$cond = $113 | $111;
   if (!($or$cond)) {
    $114 = HEAP32[$__g_end>>2]|0;
    $115 = $114;
    $116 = $__g;
    $117 = (($115) - ($116))|0;
    $118 = ($117|0)<(160);
    if ($118) {
     $119 = HEAP32[$__dc>>2]|0;
     $120 = ((($114)) + 4|0);
     HEAP32[$__g_end>>2] = $120;
     HEAP32[$114>>2] = $119;
    }
   }
   $121 = HEAP32[$__a_end>>2]|0;
   __THREW__ = 0;
   $122 = (+invoke_diii(4873,($__a$2|0),($121|0),($__err|0)));
   $123 = __THREW__; __THREW__ = 0;
   $124 = $123&1;
   if ($124) {
    label = 29;
   } else {
    HEAPF64[$__v>>3] = $122;
    $125 = HEAP32[$__g_end>>2]|0;
    __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grouping,$__g,$125,$__err);
    $126 = ($$pre$i$i$i$lcssa|0)==(0|0);
    if ($126) {
     $159 = 1;
    } else {
     $127 = ((($$pre$i$i$i$lcssa)) + 12|0);
     $128 = HEAP32[$127>>2]|0;
     $129 = ((($$pre$i$i$i$lcssa)) + 16|0);
     $130 = HEAP32[$129>>2]|0;
     $131 = ($128|0)==($130|0);
     if ($131) {
      $132 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
      $133 = ((($132)) + 36|0);
      $134 = HEAP32[$133>>2]|0;
      __THREW__ = 0;
      $135 = (invoke_ii($134|0,($$pre$i$i$i$lcssa|0))|0);
      $136 = __THREW__; __THREW__ = 0;
      $137 = $136&1;
      if ($137) {
       label = 29;
       break;
      } else {
       $$0$i$i$i$i = $135;
      }
     } else {
      $138 = HEAP32[$128>>2]|0;
      $$0$i$i$i$i = $138;
     }
     $139 = ($$0$i$i$i$i|0)==(-1);
     if ($139) {
      HEAP32[$__b>>2] = 0;
      $159 = 1;
     } else {
      $159 = 0;
     }
    }
    $140 = ($$pre$i3$i$i|0)==(0|0);
    do {
     if ($140) {
      label = 54;
     } else {
      $141 = ((($$pre$i3$i$i)) + 12|0);
      $142 = HEAP32[$141>>2]|0;
      $143 = ((($$pre$i3$i$i)) + 16|0);
      $144 = HEAP32[$143>>2]|0;
      $145 = ($142|0)==($144|0);
      if ($145) {
       $146 = HEAP32[$$pre$i3$i$i>>2]|0;
       $147 = ((($146)) + 36|0);
       $148 = HEAP32[$147>>2]|0;
       __THREW__ = 0;
       $149 = (invoke_ii($148|0,($$pre$i3$i$i|0))|0);
       $150 = __THREW__; __THREW__ = 0;
       $151 = $150&1;
       if ($151) {
        label = 29;
        break L8;
       } else {
        $$0$i$i1$i$i = $149;
       }
      } else {
       $152 = HEAP32[$142>>2]|0;
       $$0$i$i1$i$i = $152;
      }
      $153 = ($$0$i$i1$i$i|0)==(-1);
      if ($153) {
       HEAP32[$__e>>2] = 0;
       label = 54;
       break;
      } else {
       if ($159) {
        break;
       } else {
        label = 56;
        break;
       }
      }
     }
    } while(0);
    if ((label|0) == 54) {
     if ($159) {
      label = 56;
     }
    }
    if ((label|0) == 56) {
     $154 = HEAP32[$__err>>2]|0;
     $155 = $154 | 2;
     HEAP32[$__err>>2] = $155;
    }
    $156 = HEAP32[$__b>>2]|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return ($156|0);
   }
  }
 } while(0);
 if ((label|0) == 29) {
  $78 = ___cxa_find_matching_catch_2()|0;
  $79 = tempRet0;
  $lpad$phi$index = $78;$lpad$phi$index2 = $79;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRPv($this,$__b,$__e,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i9 = 0, $$0$i$i1$i$i = 0, $$0$i$i1$i$i15 = 0, $$01 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i$lcssa = 0, $$pre$i$i$i11 = 0, $$pre$i3$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $__a$0 = 0, $__a$1 = 0, $__a$2 = 0, $__a_end = 0, $__atoms = 0, $__buf = 0, $__dc = 0, $__g = 0, $__g_end = 0, $__grouping = 0, $__i$0$i$i = 0, $__i$0$i$i5 = 0, $exitcond$i$i = 0, $exitcond$i$i6 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $phitmp$i = 0, $phitmp$i30 = 0;
 var $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 320|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $__atoms = sp + 208|0;
 $__grouping = sp + 196|0;
 $0 = sp + 192|0;
 $__buf = sp + 180|0;
 $__a_end = sp + 176|0;
 $__g = sp + 16|0;
 $__g_end = sp + 8|0;
 $__dc = sp + 4|0;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $1 = (($__grouping) + ($__i$0$i$i<<2)|0);
  HEAP32[$1>>2] = 0;
  $2 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $2;
 }
 __THREW__ = 0;
 $3 = (invoke_ii(3373,($__iob|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 do {
  if ($5) {
   $88 = ___cxa_find_matching_catch_2()|0;
   $89 = tempRet0;
   $$0 = $88;$$01 = $89;
  } else {
   HEAP32[$0>>2] = $3;
   __THREW__ = 0;
   $6 = (invoke_iii(3374,($0|0),(1207608|0))|0);
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if (!($8)) {
    $9 = HEAP32[$6>>2]|0;
    $10 = ((($9)) + 48|0);
    $11 = HEAP32[$10>>2]|0;
    __THREW__ = 0;
    (invoke_iiiii($11|0,($6|0),(620486|0),((620512)|0),($__atoms|0))|0);
    $12 = __THREW__; __THREW__ = 0;
    $13 = $12&1;
    if (!($13)) {
     (__ZNSt3__114__shared_count16__release_sharedEv($3)|0);
     $__i$0$i$i5 = 0;
     while(1) {
      $exitcond$i$i6 = ($__i$0$i$i5|0)==(3);
      if ($exitcond$i$i6) {
       break;
      }
      $14 = (($__buf) + ($__i$0$i$i5<<2)|0);
      HEAP32[$14>>2] = 0;
      $15 = (($__i$0$i$i5) + 1)|0;
      $__i$0$i$i5 = $15;
     }
     $16 = HEAP8[$__buf>>0]|0;
     $17 = $16 & 1;
     $18 = ($17<<24>>24)==(0);
     if ($18) {
      $21 = 10;
     } else {
      $19 = HEAP32[$__buf>>2]|0;
      $20 = $19 & -2;
      $phitmp$i = (($20) + -1)|0;
      $21 = $phitmp$i;
     }
     __THREW__ = 0;
     invoke_viii(3371,($__buf|0),($21|0),0);
     $22 = __THREW__; __THREW__ = 0;
     $23 = $22&1;
     L18: do {
      if ($23) {
       label = 37;
      } else {
       $24 = HEAP8[$__buf>>0]|0;
       $25 = $24 & 1;
       $26 = ($25<<24>>24)==(0);
       $27 = ((($__buf)) + 8|0);
       $28 = HEAP32[$27>>2]|0;
       $29 = ((($__buf)) + 1|0);
       $30 = $26 ? $29 : $28;
       HEAP32[$__a_end>>2] = $30;
       HEAP32[$__g_end>>2] = $__g;
       HEAP32[$__dc>>2] = 0;
       $31 = ((($__buf)) + 4|0);
       $$pre = HEAP32[$__b>>2]|0;
       $$pre$i$i$i11 = $$pre;$__a$0 = $30;
       L20: while(1) {
        $32 = ($$pre$i$i$i11|0)==(0|0);
        if ($32) {
         $$pre$i$i$i = 0;$167 = 1;
        } else {
         $33 = ((($$pre$i$i$i11)) + 12|0);
         $34 = HEAP32[$33>>2]|0;
         $35 = ((($$pre$i$i$i11)) + 16|0);
         $36 = HEAP32[$35>>2]|0;
         $37 = ($34|0)==($36|0);
         if ($37) {
          $38 = HEAP32[$$pre$i$i$i11>>2]|0;
          $39 = ((($38)) + 36|0);
          $40 = HEAP32[$39>>2]|0;
          __THREW__ = 0;
          $41 = (invoke_ii($40|0,($$pre$i$i$i11|0))|0);
          $42 = __THREW__; __THREW__ = 0;
          $43 = $42&1;
          if ($43) {
           label = 36;
           break;
          } else {
           $$0$i$i$i$i9 = $41;
          }
         } else {
          $44 = HEAP32[$34>>2]|0;
          $$0$i$i$i$i9 = $44;
         }
         $45 = ($$0$i$i$i$i9|0)==(-1);
         if ($45) {
          HEAP32[$__b>>2] = 0;
          $$pre$i$i$i = 0;$167 = 1;
         } else {
          $$pre$i$i$i = $$pre$i$i$i11;$167 = 0;
         }
        }
        $46 = HEAP32[$__e>>2]|0;
        $47 = ($46|0)==(0|0);
        do {
         if ($47) {
          label = 26;
         } else {
          $48 = ((($46)) + 12|0);
          $49 = HEAP32[$48>>2]|0;
          $50 = ((($46)) + 16|0);
          $51 = HEAP32[$50>>2]|0;
          $52 = ($49|0)==($51|0);
          if ($52) {
           $53 = HEAP32[$46>>2]|0;
           $54 = ((($53)) + 36|0);
           $55 = HEAP32[$54>>2]|0;
           __THREW__ = 0;
           $56 = (invoke_ii($55|0,($46|0))|0);
           $57 = __THREW__; __THREW__ = 0;
           $58 = $57&1;
           if ($58) {
            label = 36;
            break L20;
           } else {
            $$0$i$i1$i$i15 = $56;
           }
          } else {
           $59 = HEAP32[$49>>2]|0;
           $$0$i$i1$i$i15 = $59;
          }
          $60 = ($$0$i$i1$i$i15|0)==(-1);
          if ($60) {
           HEAP32[$__e>>2] = 0;
           label = 26;
           break;
          } else {
           if ($167) {
            $168 = $46;
            break;
           } else {
            $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $46;$__a$2 = $__a$0;
            break L20;
           }
          }
         }
        } while(0);
        if ((label|0) == 26) {
         label = 0;
         if ($167) {
          $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = 0;$__a$2 = $__a$0;
          break;
         } else {
          $168 = 0;
         }
        }
        $61 = HEAP32[$__a_end>>2]|0;
        $62 = HEAP8[$__buf>>0]|0;
        $63 = $62 & 1;
        $64 = ($63<<24>>24)==(0);
        $65 = HEAP32[$31>>2]|0;
        $66 = $62&255;
        $67 = $66 >>> 1;
        $68 = $64 ? $67 : $65;
        $69 = (($__a$0) + ($68)|0);
        $70 = ($61|0)==($69|0);
        if ($70) {
         $71 = $68 << 1;
         __THREW__ = 0;
         invoke_viii(3371,($__buf|0),($71|0),0);
         $72 = __THREW__; __THREW__ = 0;
         $73 = $72&1;
         if ($73) {
          label = 36;
          break;
         }
         $74 = HEAP8[$__buf>>0]|0;
         $75 = $74 & 1;
         $76 = ($75<<24>>24)==(0);
         if ($76) {
          $79 = 10;
         } else {
          $77 = HEAP32[$__buf>>2]|0;
          $78 = $77 & -2;
          $phitmp$i30 = (($78) + -1)|0;
          $79 = $phitmp$i30;
         }
         __THREW__ = 0;
         invoke_viii(3371,($__buf|0),($79|0),0);
         $80 = __THREW__; __THREW__ = 0;
         $81 = $80&1;
         if ($81) {
          label = 36;
          break;
         }
         $82 = HEAP8[$__buf>>0]|0;
         $83 = $82 & 1;
         $84 = ($83<<24>>24)==(0);
         $85 = HEAP32[$27>>2]|0;
         $86 = $84 ? $29 : $85;
         $87 = (($86) + ($68)|0);
         HEAP32[$__a_end>>2] = $87;
         $__a$1 = $86;
        } else {
         $__a$1 = $__a$0;
        }
        $96 = ((($$pre$i$i$i)) + 12|0);
        $97 = HEAP32[$96>>2]|0;
        $98 = ((($$pre$i$i$i)) + 16|0);
        $99 = HEAP32[$98>>2]|0;
        $100 = ($97|0)==($99|0);
        if ($100) {
         $101 = HEAP32[$$pre$i$i$i>>2]|0;
         $102 = ((($101)) + 36|0);
         $103 = HEAP32[$102>>2]|0;
         __THREW__ = 0;
         $104 = (invoke_ii($103|0,($$pre$i$i$i|0))|0);
         $105 = __THREW__; __THREW__ = 0;
         $106 = $105&1;
         if ($106) {
          label = 36;
          break;
         } else {
          $$0$i$i = $104;
         }
        } else {
         $107 = HEAP32[$97>>2]|0;
         $$0$i$i = $107;
        }
        $108 = (__ZNSt3__19__num_getIwE17__stage2_int_loopEwiPcRS2_RjwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_Pw($$0$i$i,16,$__a$1,$__a_end,$__dc,0,$__grouping,$__g,$__g_end,$__atoms)|0);
        $109 = ($108|0)==(0);
        if (!($109)) {
         $$pre$i$i$i$lcssa = $$pre$i$i$i;$$pre$i3$i$i = $168;$__a$2 = $__a$1;
         break;
        }
        $110 = HEAP32[$96>>2]|0;
        $111 = HEAP32[$98>>2]|0;
        $112 = ($110|0)==($111|0);
        if ($112) {
         $113 = HEAP32[$$pre$i$i$i>>2]|0;
         $114 = ((($113)) + 40|0);
         $115 = HEAP32[$114>>2]|0;
         __THREW__ = 0;
         (invoke_ii($115|0,($$pre$i$i$i|0))|0);
         $116 = __THREW__; __THREW__ = 0;
         $117 = $116&1;
         if ($117) {
          label = 36;
          break;
         } else {
          $$pre$i$i$i11 = $$pre$i$i$i;$__a$0 = $__a$1;
          continue;
         }
        } else {
         $118 = ((($110)) + 4|0);
         HEAP32[$96>>2] = $118;
         $$pre$i$i$i11 = $$pre$i$i$i;$__a$0 = $__a$1;
         continue;
        }
       }
       if ((label|0) == 36) {
        $92 = ___cxa_find_matching_catch_2()|0;
        $93 = tempRet0;
        $lpad$phi$index = $92;$lpad$phi$index2 = $93;
        break;
       }
       $119 = HEAP32[$__a_end>>2]|0;
       $120 = $__a$2;
       $121 = (($119) - ($120))|0;
       __THREW__ = 0;
       invoke_viii(3371,($__buf|0),($121|0),0);
       $122 = __THREW__; __THREW__ = 0;
       $123 = $122&1;
       if ($123) {
        label = 37;
       } else {
        $124 = HEAP8[$__buf>>0]|0;
        $125 = HEAP32[$27>>2]|0;
        __THREW__ = 0;
        $126 = (invoke_i(4874)|0);
        $127 = __THREW__; __THREW__ = 0;
        $128 = $127&1;
        if ($128) {
         label = 37;
        } else {
         $129 = $124 & 1;
         $130 = ($129<<24>>24)==(0);
         $131 = $130 ? $29 : $125;
         __THREW__ = 0;
         HEAP32[$vararg_buffer>>2] = $__v;
         $132 = (invoke_iiiii(4875,($131|0),($126|0),(620521|0),($vararg_buffer|0))|0);
         $133 = __THREW__; __THREW__ = 0;
         $134 = $133&1;
         if ($134) {
          label = 37;
         } else {
          $135 = ($132|0)==(1);
          if (!($135)) {
           HEAP32[$__err>>2] = 4;
          }
          $136 = ($$pre$i$i$i$lcssa|0)==(0|0);
          if ($136) {
           $169 = 1;
          } else {
           $137 = ((($$pre$i$i$i$lcssa)) + 12|0);
           $138 = HEAP32[$137>>2]|0;
           $139 = ((($$pre$i$i$i$lcssa)) + 16|0);
           $140 = HEAP32[$139>>2]|0;
           $141 = ($138|0)==($140|0);
           if ($141) {
            $142 = HEAP32[$$pre$i$i$i$lcssa>>2]|0;
            $143 = ((($142)) + 36|0);
            $144 = HEAP32[$143>>2]|0;
            __THREW__ = 0;
            $145 = (invoke_ii($144|0,($$pre$i$i$i$lcssa|0))|0);
            $146 = __THREW__; __THREW__ = 0;
            $147 = $146&1;
            if ($147) {
             label = 37;
             break;
            } else {
             $$0$i$i$i$i = $145;
            }
           } else {
            $148 = HEAP32[$138>>2]|0;
            $$0$i$i$i$i = $148;
           }
           $149 = ($$0$i$i$i$i|0)==(-1);
           if ($149) {
            HEAP32[$__b>>2] = 0;
            $169 = 1;
           } else {
            $169 = 0;
           }
          }
          $150 = ($$pre$i3$i$i|0)==(0|0);
          do {
           if ($150) {
            label = 63;
           } else {
            $151 = ((($$pre$i3$i$i)) + 12|0);
            $152 = HEAP32[$151>>2]|0;
            $153 = ((($$pre$i3$i$i)) + 16|0);
            $154 = HEAP32[$153>>2]|0;
            $155 = ($152|0)==($154|0);
            if ($155) {
             $156 = HEAP32[$$pre$i3$i$i>>2]|0;
             $157 = ((($156)) + 36|0);
             $158 = HEAP32[$157>>2]|0;
             __THREW__ = 0;
             $159 = (invoke_ii($158|0,($$pre$i3$i$i|0))|0);
             $160 = __THREW__; __THREW__ = 0;
             $161 = $160&1;
             if ($161) {
              label = 37;
              break L18;
             } else {
              $$0$i$i1$i$i = $159;
             }
            } else {
             $162 = HEAP32[$152>>2]|0;
             $$0$i$i1$i$i = $162;
            }
            $163 = ($$0$i$i1$i$i|0)==(-1);
            if ($163) {
             HEAP32[$__e>>2] = 0;
             label = 63;
             break;
            } else {
             if ($169) {
              break;
             } else {
              label = 65;
              break;
             }
            }
           }
          } while(0);
          if ((label|0) == 63) {
           if ($169) {
            label = 65;
           }
          }
          if ((label|0) == 65) {
           $164 = HEAP32[$__err>>2]|0;
           $165 = $164 | 2;
           HEAP32[$__err>>2] = $165;
          }
          $166 = HEAP32[$__b>>2]|0;
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
          __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
          STACKTOP = sp;return ($166|0);
         }
        }
       }
      }
     } while(0);
     if ((label|0) == 37) {
      $94 = ___cxa_find_matching_catch_2()|0;
      $95 = tempRet0;
      $lpad$phi$index = $94;$lpad$phi$index2 = $95;
     }
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__buf);
     $$0 = $lpad$phi$index;$$01 = $lpad$phi$index2;
     break;
    }
   }
   $90 = ___cxa_find_matching_catch_2()|0;
   $91 = tempRet0;
   (__ZNSt3__114__shared_count16__release_sharedEv($3)|0);
   $$0 = $90;$$01 = $91;
  }
 } while(0);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($$0|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcb($this,$__s,$__iob,$__fl,$__v) {
 $this = $this|0;
 $__s = $__s|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $__v = $__v|0;
 var $$0 = 0, $$01 = 0, $$byval_copy = 0, $$pre = 0, $$pre3 = 0, $$sroa$0$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $7 = 0, $8 = 0, $9 = 0, $__i$sroa$0$0 = 0, $__i$sroa$0$0$ph = 0, $__nm = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 20|0;
 $0 = sp + 16|0;
 $1 = sp + 12|0;
 $__nm = sp;
 $2 = ((($__iob)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 1;
 $5 = ($4|0)==(0);
 L1: do {
  if ($5) {
   $6 = HEAP32[$this>>2]|0;
   $7 = ((($6)) + 24|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = HEAP32[$__s>>2]|0;
   HEAP32[$0>>2] = $9;
   $10 = $__v&1;
   ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
   $11 = (FUNCTION_TABLE_iiiiii[$8 & 8191]($this,$$byval_copy,$__iob,$__fl,$10)|0);
   $$sroa$0$0 = $11;
  } else {
   $12 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
   HEAP32[$1>>2] = $12;
   __THREW__ = 0;
   $13 = (invoke_iii(3374,($1|0),(1207588|0))|0);
   $14 = __THREW__; __THREW__ = 0;
   $15 = $14&1;
   do {
    if ($15) {
     $40 = ___cxa_find_matching_catch_2()|0;
     $41 = tempRet0;
     (__ZNSt3__114__shared_count16__release_sharedEv($12)|0);
     $$0 = $41;$$01 = $40;
    } else {
     (__ZNSt3__114__shared_count16__release_sharedEv($12)|0);
     $16 = HEAP32[$13>>2]|0;
     if ($__v) {
      $17 = ((($16)) + 24|0);
      $18 = HEAP32[$17>>2]|0;
      FUNCTION_TABLE_vii[$18 & 8191]($__nm,$13);
     } else {
      $19 = ((($16)) + 28|0);
      $20 = HEAP32[$19>>2]|0;
      FUNCTION_TABLE_vii[$20 & 8191]($__nm,$13);
     }
     $21 = HEAP8[$__nm>>0]|0;
     $22 = $21 & 1;
     $23 = ($22<<24>>24)==(0);
     $24 = ((($__nm)) + 1|0);
     $25 = ((($__nm)) + 8|0);
     $26 = HEAP32[$25>>2]|0;
     $__i$sroa$0$0$ph = $23 ? $24 : $26;
     $27 = ((($__nm)) + 4|0);
     $29 = $21;$35 = $26;$__i$sroa$0$0 = $__i$sroa$0$0$ph;
     L11: while(1) {
      $28 = $29 & 1;
      $30 = ($28<<24>>24)==(0);
      $31 = HEAP32[$27>>2]|0;
      $32 = $29&255;
      $33 = $32 >>> 1;
      $34 = $30 ? $24 : $35;
      $36 = $30 ? $33 : $31;
      $37 = (($34) + ($36)|0);
      $38 = ($__i$sroa$0$0|0)==($37|0);
      if ($38) {
       label = 9;
       break;
      }
      $42 = HEAP8[$__i$sroa$0$0>>0]|0;
      $43 = HEAP32[$__s>>2]|0;
      $44 = ($43|0)==(0|0);
      do {
       if (!($44)) {
        $45 = ((($43)) + 24|0);
        $46 = HEAP32[$45>>2]|0;
        $47 = ((($43)) + 28|0);
        $48 = HEAP32[$47>>2]|0;
        $49 = ($46|0)==($48|0);
        if (!($49)) {
         $50 = ((($46)) + 1|0);
         HEAP32[$45>>2] = $50;
         HEAP8[$46>>0] = $42;
         break;
        }
        $51 = HEAP32[$43>>2]|0;
        $52 = ((($51)) + 52|0);
        $53 = HEAP32[$52>>2]|0;
        $54 = $42&255;
        __THREW__ = 0;
        $55 = (invoke_iii($53|0,($43|0),($54|0))|0);
        $56 = __THREW__; __THREW__ = 0;
        $57 = $56&1;
        if ($57) {
         label = 18;
         break L11;
        }
        $58 = ($55|0)==(-1);
        if ($58) {
         HEAP32[$__s>>2] = 0;
        }
       }
      } while(0);
      $59 = ((($__i$sroa$0$0)) + 1|0);
      $$pre = HEAP8[$__nm>>0]|0;
      $$pre3 = HEAP32[$25>>2]|0;
      $29 = $$pre;$35 = $$pre3;$__i$sroa$0$0 = $59;
     }
     if ((label|0) == 9) {
      $39 = HEAP32[$__s>>2]|0;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__nm);
      $$sroa$0$0 = $39;
      break L1;
     }
     else if ((label|0) == 18) {
      $60 = ___cxa_find_matching_catch_2()|0;
      $61 = tempRet0;
      __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__nm);
      $$0 = $61;$$01 = $60;
      break;
     }
    }
   } while(0);
   ___resumeException($$01|0);
   // unreachable;
  }
 } while(0);
 STACKTOP = sp;return ($$sroa$0$0|0);
}
function __ZNKSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcl($this,$__s,$__iob,$__fl,$__v) {
 $this = $this|0;
 $__s = $__s|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__fmt = 0;
 var $__nar = 0, $__o = 0, $__oe = 0, $__op = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 20|0;
 $vararg_buffer = sp;
 $__fmt = sp + 60|0;
 $__nar = sp + 48|0;
 $__o = sp + 24|0;
 $__op = sp + 16|0;
 $__oe = sp + 12|0;
 $0 = sp + 8|0;
 $1 = sp + 4|0;
 ;HEAP8[$__fmt>>0]=HEAP8[620729>>0]|0;HEAP8[$__fmt+1>>0]=HEAP8[620729+1>>0]|0;HEAP8[$__fmt+2>>0]=HEAP8[620729+2>>0]|0;HEAP8[$__fmt+3>>0]=HEAP8[620729+3>>0]|0;HEAP8[$__fmt+4>>0]=HEAP8[620729+4>>0]|0;HEAP8[$__fmt+5>>0]=HEAP8[620729+5>>0]|0;
 $2 = ((($__fmt)) + 1|0);
 $3 = ((($__iob)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 __ZNSt3__114__num_put_base12__format_intEPcPKcbj($2,620735,1,$4);
 $5 = (__ZNSt3__16__clocEv()|0);
 HEAP32[$vararg_buffer>>2] = $__v;
 $6 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,12,$5,$__fmt,$vararg_buffer)|0);
 $7 = (($__nar) + ($6)|0);
 $8 = (__ZNSt3__114__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($__nar,$7,$__iob)|0);
 $9 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$0>>2] = $9;
 __THREW__ = 0;
 invoke_viiiiiii(4878,($__nar|0),($8|0),($7|0),($__o|0),($__op|0),($__oe|0),($0|0));
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($9)|0);
  ___resumeException($16|0);
  // unreachable;
 } else {
  (__ZNSt3__114__shared_count16__release_sharedEv($9)|0);
  $12 = HEAP32[$__s>>2]|0;
  HEAP32[$1>>2] = $12;
  $13 = HEAP32[$__op>>2]|0;
  $14 = HEAP32[$__oe>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
  $15 = (__ZNSt3__116__pad_and_outputIcNS_11char_traitsIcEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$__o,$13,$14,$__iob,$__fl)|0);
  STACKTOP = sp;return ($15|0);
 }
 return (0)|0;
}
function __ZNSt3__114__num_put_base12__format_intEPcPKcbj($__fmtp,$__len,$__signd,$__flags) {
 $__fmtp = $__fmtp|0;
 $__len = $__len|0;
 $__signd = $__signd|0;
 $__flags = $__flags|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$1$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $__flags & 2048;
 $1 = ($0|0)==(0);
 if ($1) {
  $$0 = $__fmtp;
 } else {
  $2 = ((($__fmtp)) + 1|0);
  HEAP8[$__fmtp>>0] = 43;
  $$0 = $2;
 }
 $3 = $__flags & 512;
 $4 = ($3|0)==(0);
 if ($4) {
  $$01 = $__len;$$1 = $$0;
 } else {
  $5 = ((($$0)) + 1|0);
  HEAP8[$$0>>0] = 35;
  $$01 = $__len;$$1 = $5;
 }
 while(1) {
  $6 = HEAP8[$$01>>0]|0;
  $7 = ($6<<24>>24)==(0);
  if ($7) {
   $$1$lcssa = $$1;
   break;
  }
  $8 = ((($$01)) + 1|0);
  $9 = ((($$1)) + 1|0);
  HEAP8[$$1>>0] = $6;
  $$01 = $8;$$1 = $9;
 }
 $10 = $__flags & 74;
 L10: do {
  switch ($10|0) {
  case 64:  {
   HEAP8[$$1$lcssa>>0] = 111;
   break;
  }
  case 8:  {
   $11 = $__flags & 16384;
   $12 = ($11|0)==(0);
   if ($12) {
    HEAP8[$$1$lcssa>>0] = 120;
    break L10;
   } else {
    HEAP8[$$1$lcssa>>0] = 88;
    break L10;
   }
   break;
  }
  default: {
   if ($__signd) {
    HEAP8[$$1$lcssa>>0] = 100;
    break L10;
   } else {
    HEAP8[$$1$lcssa>>0] = 117;
    break L10;
   }
  }
  }
 } while(0);
 return;
}
function __ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__s,$__n,$__l,$__format,$varargs) {
 $__s = $__s|0;
 $__n = $__n|0;
 $__l = $__l|0;
 $__format = $__format|0;
 $varargs = $varargs|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $__va = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__va = sp;
 HEAP32[$__va>>2] = $varargs;
 $0 = (_uselocale($__l)|0);
 $1 = (_vsnprintf($__s,$__n,$__format,$__va)|0);
 $2 = ($0|0)==(0|0);
 if (!($2)) {
  __THREW__ = 0;
  (invoke_ii(4876,($0|0))|0);
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if ($4) {
   $5 = ___cxa_find_matching_catch_3(0|0)|0;
   $6 = tempRet0;
   ___clang_call_terminate($5);
   // unreachable;
  }
 }
 STACKTOP = sp;return ($1|0);
}
function __ZNSt3__114__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($__nb,$__ne,$__iob) {
 $__nb = $__nb|0;
 $__ne = $__ne|0;
 $__iob = $__iob|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($__iob)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 176;
 L1: do {
  switch ($2|0) {
  case 16:  {
   $3 = HEAP8[$__nb>>0]|0;
   switch ($3<<24>>24) {
   case 43: case 45:  {
    $4 = ((($__nb)) + 1|0);
    $$0 = $4;
    break L1;
    break;
   }
   default: {
   }
   }
   $5 = $__ne;
   $6 = $__nb;
   $7 = (($5) - ($6))|0;
   $8 = ($7|0)>(1);
   $9 = ($3<<24>>24)==(48);
   $or$cond = $8 & $9;
   if ($or$cond) {
    $10 = ((($__nb)) + 1|0);
    $11 = HEAP8[$10>>0]|0;
    switch ($11<<24>>24) {
    case 88: case 120:  {
     break;
    }
    default: {
     label = 7;
     break L1;
    }
    }
    $12 = ((($__nb)) + 2|0);
    $$0 = $12;
   } else {
    label = 7;
   }
   break;
  }
  case 32:  {
   $$0 = $__ne;
   break;
  }
  default: {
   label = 7;
  }
  }
 } while(0);
 if ((label|0) == 7) {
  $$0 = $__nb;
 }
 return ($$0|0);
}
function __ZNSt3__19__num_putIcE21__widen_and_group_intEPcS2_S2_S2_RS2_S3_RKNS_6localeE($__nb,$__np,$__ne,$__ob,$__op,$__oe,$__loc) {
 $__nb = $__nb|0;
 $__np = $__np|0;
 $__ne = $__ne|0;
 $__ob = $__ob|0;
 $__op = $__op|0;
 $__oe = $__oe|0;
 $__loc = $__loc|0;
 var $$0$i$i = 0, $$0$i$i6 = 0, $$01$i$i = 0, $$01$i$i5 = 0, $$__dg$0 = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__dc$0 = 0, $__dc$1 = 0, $__dg$0 = 0, $__dg$1 = 0, $__grouping = 0, $__nf$0 = 0;
 var $__nf$1 = 0, $__p$0 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $or$cond = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__grouping = sp;
 $0 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1207576)|0);
 $1 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1207588)|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 FUNCTION_TABLE_vii[$4 & 8191]($__grouping,$1);
 $5 = HEAP8[$__grouping>>0]|0;
 $6 = $5 & 1;
 $7 = ($6<<24>>24)==(0);
 $8 = ((($__grouping)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $5&255;
 $11 = $10 >>> 1;
 $12 = $7 ? $11 : $9;
 $13 = ($12|0)==(0);
 L1: do {
  if ($13) {
   $14 = HEAP32[$0>>2]|0;
   $15 = ((($14)) + 32|0);
   $16 = HEAP32[$15>>2]|0;
   __THREW__ = 0;
   (invoke_iiiii($16|0,($0|0),($__nb|0),($__ne|0),($__ob|0))|0);
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if ($18) {
    label = 5;
   } else {
    $19 = $__ne;
    $20 = $__nb;
    $21 = (($19) - ($20))|0;
    $22 = (($__ob) + ($21)|0);
    HEAP32[$__oe>>2] = $22;
    $$pre$phiZ2D = $20;$126 = $22;
    label = 30;
   }
  } else {
   HEAP32[$__oe>>2] = $__ob;
   $27 = HEAP8[$__nb>>0]|0;
   switch ($27<<24>>24) {
   case 43: case 45:  {
    $28 = HEAP32[$0>>2]|0;
    $29 = ((($28)) + 28|0);
    $30 = HEAP32[$29>>2]|0;
    __THREW__ = 0;
    $31 = (invoke_iii($30|0,($0|0),($27|0))|0);
    $32 = __THREW__; __THREW__ = 0;
    $33 = $32&1;
    if ($33) {
     label = 5;
     break L1;
    }
    $34 = ((($__nb)) + 1|0);
    $35 = HEAP32[$__oe>>2]|0;
    $36 = ((($35)) + 1|0);
    HEAP32[$__oe>>2] = $36;
    HEAP8[$35>>0] = $31;
    $__nf$0 = $34;
    break;
   }
   default: {
    $__nf$0 = $__nb;
   }
   }
   $37 = $__ne;
   $38 = $__nf$0;
   $39 = (($37) - ($38))|0;
   $40 = ($39|0)>(1);
   L9: do {
    if ($40) {
     $41 = HEAP8[$__nf$0>>0]|0;
     $42 = ($41<<24>>24)==(48);
     if ($42) {
      $43 = ((($__nf$0)) + 1|0);
      $44 = HEAP8[$43>>0]|0;
      switch ($44<<24>>24) {
      case 88: case 120:  {
       break;
      }
      default: {
       $__nf$1 = $__nf$0;
       break L9;
      }
      }
      $45 = HEAP32[$0>>2]|0;
      $46 = ((($45)) + 28|0);
      $47 = HEAP32[$46>>2]|0;
      __THREW__ = 0;
      $48 = (invoke_iii($47|0,($0|0),48)|0);
      $49 = __THREW__; __THREW__ = 0;
      $50 = $49&1;
      if ($50) {
       label = 5;
       break L1;
      }
      $51 = HEAP32[$__oe>>2]|0;
      $52 = ((($51)) + 1|0);
      HEAP32[$__oe>>2] = $52;
      HEAP8[$51>>0] = $48;
      $53 = HEAP8[$43>>0]|0;
      $54 = HEAP32[$0>>2]|0;
      $55 = ((($54)) + 28|0);
      $56 = HEAP32[$55>>2]|0;
      __THREW__ = 0;
      $57 = (invoke_iii($56|0,($0|0),($53|0))|0);
      $58 = __THREW__; __THREW__ = 0;
      $59 = $58&1;
      if ($59) {
       label = 5;
       break L1;
      }
      $60 = ((($__nf$0)) + 2|0);
      $61 = HEAP32[$__oe>>2]|0;
      $62 = ((($61)) + 1|0);
      HEAP32[$__oe>>2] = $62;
      HEAP8[$61>>0] = $57;
      $__nf$1 = $60;
     } else {
      $__nf$1 = $__nf$0;
     }
    } else {
     $__nf$1 = $__nf$0;
    }
   } while(0);
   $63 = ($__nf$1|0)==($__ne|0);
   L16: do {
    if (!($63)) {
     $$0$i$i = $__ne;$$01$i$i = $__nf$1;
     while(1) {
      $64 = ((($$0$i$i)) + -1|0);
      $65 = ($$01$i$i>>>0)<($64>>>0);
      if (!($65)) {
       break L16;
      }
      $66 = HEAP8[$$01$i$i>>0]|0;
      $67 = HEAP8[$64>>0]|0;
      HEAP8[$$01$i$i>>0] = $67;
      HEAP8[$64>>0] = $66;
      $68 = ((($$01$i$i)) + 1|0);
      $$0$i$i = $64;$$01$i$i = $68;
     }
    }
   } while(0);
   $69 = HEAP32[$1>>2]|0;
   $70 = ((($69)) + 16|0);
   $71 = HEAP32[$70>>2]|0;
   __THREW__ = 0;
   $72 = (invoke_ii($71|0,($1|0))|0);
   $73 = __THREW__; __THREW__ = 0;
   $74 = $73&1;
   if ($74) {
    label = 5;
   } else {
    $75 = ((($__grouping)) + 8|0);
    $76 = ((($__grouping)) + 1|0);
    $__dc$0 = 0;$__dg$0 = 0;$__p$0 = $__nf$1;
    while(1) {
     $77 = ($__p$0>>>0)<($__ne>>>0);
     if (!($77)) {
      break;
     }
     $89 = HEAP8[$__grouping>>0]|0;
     $90 = $89 & 1;
     $91 = ($90<<24>>24)==(0);
     $92 = HEAP32[$75>>2]|0;
     $93 = $91 ? $76 : $92;
     $94 = (($93) + ($__dg$0)|0);
     $95 = HEAP8[$94>>0]|0;
     $96 = ($95<<24>>24)!=(0);
     $97 = $95 << 24 >> 24;
     $98 = ($__dc$0|0)==($97|0);
     $or$cond = $96 & $98;
     if ($or$cond) {
      $99 = HEAP32[$__oe>>2]|0;
      $100 = ((($99)) + 1|0);
      HEAP32[$__oe>>2] = $100;
      HEAP8[$99>>0] = $72;
      $101 = HEAP8[$__grouping>>0]|0;
      $102 = $101 & 1;
      $103 = ($102<<24>>24)==(0);
      $104 = HEAP32[$8>>2]|0;
      $105 = $101&255;
      $106 = $105 >>> 1;
      $107 = $103 ? $106 : $104;
      $108 = (($107) + -1)|0;
      $109 = ($__dg$0>>>0)<($108>>>0);
      $110 = $109&1;
      $$__dg$0 = (($110) + ($__dg$0))|0;
      $__dc$1 = 0;$__dg$1 = $$__dg$0;
     } else {
      $__dc$1 = $__dc$0;$__dg$1 = $__dg$0;
     }
     $111 = HEAP8[$__p$0>>0]|0;
     $112 = HEAP32[$0>>2]|0;
     $113 = ((($112)) + 28|0);
     $114 = HEAP32[$113>>2]|0;
     __THREW__ = 0;
     $115 = (invoke_iii($114|0,($0|0),($111|0))|0);
     $116 = __THREW__; __THREW__ = 0;
     $117 = $116&1;
     if ($117) {
      label = 4;
      break;
     }
     $118 = HEAP32[$__oe>>2]|0;
     $119 = ((($118)) + 1|0);
     HEAP32[$__oe>>2] = $119;
     HEAP8[$118>>0] = $115;
     $120 = (($__dc$1) + 1)|0;
     $121 = ((($__p$0)) + 1|0);
     $__dc$0 = $120;$__dg$0 = $__dg$1;$__p$0 = $121;
    }
    if ((label|0) == 4) {
     $23 = ___cxa_find_matching_catch_2()|0;
     $24 = tempRet0;
     $lpad$phi$index = $23;$lpad$phi$index2 = $24;
     break;
    }
    $78 = $__nf$1;
    $79 = $__nb;
    $80 = (($78) - ($79))|0;
    $81 = (($__ob) + ($80)|0);
    $82 = HEAP32[$__oe>>2]|0;
    $83 = ($81|0)==($82|0);
    if ($83) {
     $$pre$phiZ2D = $79;$126 = $81;
     label = 30;
    } else {
     $$0$i$i6 = $82;$$01$i$i5 = $81;
     while(1) {
      $84 = ((($$0$i$i6)) + -1|0);
      $85 = ($$01$i$i5>>>0)<($84>>>0);
      if (!($85)) {
       break;
      }
      $86 = HEAP8[$$01$i$i5>>0]|0;
      $87 = HEAP8[$84>>0]|0;
      HEAP8[$$01$i$i5>>0] = $87;
      HEAP8[$84>>0] = $86;
      $88 = ((($$01$i$i5)) + 1|0);
      $$0$i$i6 = $84;$$01$i$i5 = $88;
     }
     $$pre = HEAP32[$__oe>>2]|0;
     $$pre$phiZ2D = $79;$126 = $$pre;
     label = 30;
    }
   }
  }
 } while(0);
 if ((label|0) == 5) {
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  $lpad$phi$index = $25;$lpad$phi$index2 = $26;
 }
 else if ((label|0) == 30) {
  $122 = ($__np|0)==($__ne|0);
  $123 = $__np;
  $124 = (($123) - ($$pre$phiZ2D))|0;
  $125 = (($__ob) + ($124)|0);
  $storemerge = $122 ? $126 : $125;
  HEAP32[$__op>>2] = $storemerge;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
  STACKTOP = sp;return;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZNKSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcx($this,$__s,$__iob,$__fl,$0,$1) {
 $this = $this|0;
 $__s = $__s|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__fmt = 0, $__nar = 0, $__o = 0, $__oe = 0, $__op = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 32|0;
 $vararg_buffer = sp + 8|0;
 $__fmt = sp;
 $__nar = sp + 78|0;
 $__o = sp + 36|0;
 $__op = sp + 28|0;
 $__oe = sp + 24|0;
 $2 = sp + 20|0;
 $3 = sp + 16|0;
 $4 = $__fmt;
 $5 = $4;
 HEAP32[$5>>2] = 37;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = ((($__fmt)) + 1|0);
 $9 = ((($__iob)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 __ZNSt3__114__num_put_base12__format_intEPcPKcbj($8,620737,1,$10);
 $11 = (__ZNSt3__16__clocEv()|0);
 $12 = $vararg_buffer;
 $13 = $12;
 HEAP32[$13>>2] = $0;
 $14 = (($12) + 4)|0;
 $15 = $14;
 HEAP32[$15>>2] = $1;
 $16 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,22,$11,$__fmt,$vararg_buffer)|0);
 $17 = (($__nar) + ($16)|0);
 $18 = (__ZNSt3__114__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($__nar,$17,$__iob)|0);
 $19 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$2>>2] = $19;
 __THREW__ = 0;
 invoke_viiiiiii(4878,($__nar|0),($18|0),($17|0),($__o|0),($__op|0),($__oe|0),($2|0));
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($19)|0);
  ___resumeException($26|0);
  // unreachable;
 } else {
  (__ZNSt3__114__shared_count16__release_sharedEv($19)|0);
  $22 = HEAP32[$__s>>2]|0;
  HEAP32[$3>>2] = $22;
  $23 = HEAP32[$__op>>2]|0;
  $24 = HEAP32[$__oe>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$3>>2]|0;
  $25 = (__ZNSt3__116__pad_and_outputIcNS_11char_traitsIcEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$__o,$23,$24,$__iob,$__fl)|0);
  STACKTOP = sp;return ($25|0);
 }
 return (0)|0;
}
function __ZNKSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcm($this,$__s,$__iob,$__fl,$__v) {
 $this = $this|0;
 $__s = $__s|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__fmt = 0;
 var $__nar = 0, $__o = 0, $__oe = 0, $__op = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 20|0;
 $vararg_buffer = sp;
 $__fmt = sp + 60|0;
 $__nar = sp + 48|0;
 $__o = sp + 24|0;
 $__op = sp + 16|0;
 $__oe = sp + 12|0;
 $0 = sp + 8|0;
 $1 = sp + 4|0;
 ;HEAP8[$__fmt>>0]=HEAP8[620729>>0]|0;HEAP8[$__fmt+1>>0]=HEAP8[620729+1>>0]|0;HEAP8[$__fmt+2>>0]=HEAP8[620729+2>>0]|0;HEAP8[$__fmt+3>>0]=HEAP8[620729+3>>0]|0;HEAP8[$__fmt+4>>0]=HEAP8[620729+4>>0]|0;HEAP8[$__fmt+5>>0]=HEAP8[620729+5>>0]|0;
 $2 = ((($__fmt)) + 1|0);
 $3 = ((($__iob)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 __ZNSt3__114__num_put_base12__format_intEPcPKcbj($2,620735,0,$4);
 $5 = (__ZNSt3__16__clocEv()|0);
 HEAP32[$vararg_buffer>>2] = $__v;
 $6 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,12,$5,$__fmt,$vararg_buffer)|0);
 $7 = (($__nar) + ($6)|0);
 $8 = (__ZNSt3__114__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($__nar,$7,$__iob)|0);
 $9 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$0>>2] = $9;
 __THREW__ = 0;
 invoke_viiiiiii(4878,($__nar|0),($8|0),($7|0),($__o|0),($__op|0),($__oe|0),($0|0));
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($9)|0);
  ___resumeException($16|0);
  // unreachable;
 } else {
  (__ZNSt3__114__shared_count16__release_sharedEv($9)|0);
  $12 = HEAP32[$__s>>2]|0;
  HEAP32[$1>>2] = $12;
  $13 = HEAP32[$__op>>2]|0;
  $14 = HEAP32[$__oe>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
  $15 = (__ZNSt3__116__pad_and_outputIcNS_11char_traitsIcEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$__o,$13,$14,$__iob,$__fl)|0);
  STACKTOP = sp;return ($15|0);
 }
 return (0)|0;
}
function __ZNKSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcy($this,$__s,$__iob,$__fl,$0,$1) {
 $this = $this|0;
 $__s = $__s|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__fmt = 0, $__nar = 0, $__o = 0, $__oe = 0, $__op = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 32|0;
 $vararg_buffer = sp + 8|0;
 $__fmt = sp;
 $__nar = sp + 79|0;
 $__o = sp + 36|0;
 $__op = sp + 28|0;
 $__oe = sp + 24|0;
 $2 = sp + 20|0;
 $3 = sp + 16|0;
 $4 = $__fmt;
 $5 = $4;
 HEAP32[$5>>2] = 37;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = ((($__fmt)) + 1|0);
 $9 = ((($__iob)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 __ZNSt3__114__num_put_base12__format_intEPcPKcbj($8,620737,0,$10);
 $11 = (__ZNSt3__16__clocEv()|0);
 $12 = $vararg_buffer;
 $13 = $12;
 HEAP32[$13>>2] = $0;
 $14 = (($12) + 4)|0;
 $15 = $14;
 HEAP32[$15>>2] = $1;
 $16 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,23,$11,$__fmt,$vararg_buffer)|0);
 $17 = (($__nar) + ($16)|0);
 $18 = (__ZNSt3__114__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($__nar,$17,$__iob)|0);
 $19 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$2>>2] = $19;
 __THREW__ = 0;
 invoke_viiiiiii(4878,($__nar|0),($18|0),($17|0),($__o|0),($__op|0),($__oe|0),($2|0));
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($19)|0);
  ___resumeException($26|0);
  // unreachable;
 } else {
  (__ZNSt3__114__shared_count16__release_sharedEv($19)|0);
  $22 = HEAP32[$__s>>2]|0;
  HEAP32[$3>>2] = $22;
  $23 = HEAP32[$__op>>2]|0;
  $24 = HEAP32[$__oe>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$3>>2]|0;
  $25 = (__ZNSt3__116__pad_and_outputIcNS_11char_traitsIcEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$__o,$23,$24,$__iob,$__fl)|0);
  STACKTOP = sp;return ($25|0);
 }
 return (0)|0;
}
function __ZNKSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcd($this,$__s,$__iob,$__fl,$__v) {
 $this = $this|0;
 $__s = $__s|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $__v = +$__v;
 var $$0 = 0, $$02 = 0, $$117 = 0, $$1316 = 0, $$byval_copy = 0, $$pre = 0, $$pre10 = 0, $$pre10$pre = 0, $$pre1012 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $8 = 0, $9 = 0, $__fmt = 0, $__nar = 0, $__nb = 0, $__nbh$sroa$0$0 = 0, $__nc$0 = 0, $__nc$1 = 0, $__nc$2 = 0, $__o = 0, $__ob$0 = 0, $__obh$sroa$0$0 = 0, $__obh$sroa$0$1 = 0, $__obh$sroa$0$2 = 0, $__oe = 0, $__op = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_buffer5 = 0;
 var $vararg_buffer9 = 0, $vararg_ptr1 = 0, $vararg_ptr12 = 0, $vararg_ptr8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 84|0;
 $vararg_buffer9 = sp + 48|0;
 $vararg_buffer5 = sp + 32|0;
 $vararg_buffer2 = sp + 24|0;
 $vararg_buffer = sp + 8|0;
 $__fmt = sp;
 $__nar = sp + 146|0;
 $__nb = sp + 80|0;
 $__o = sp + 88|0;
 $__op = sp + 76|0;
 $__oe = sp + 72|0;
 $0 = sp + 68|0;
 $1 = sp + 64|0;
 $2 = $__fmt;
 $3 = $2;
 HEAP32[$3>>2] = 37;
 $4 = (($2) + 4)|0;
 $5 = $4;
 HEAP32[$5>>2] = 0;
 $6 = ((($__fmt)) + 1|0);
 $7 = ((($__iob)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (__ZNSt3__114__num_put_base14__format_floatEPcPKcj($6,1282466,$8)|0);
 HEAP32[$__nb>>2] = $__nar;
 $10 = (__ZNSt3__16__clocEv()|0);
 if ($9) {
  $11 = ((($__iob)) + 8|0);
  $12 = HEAP32[$11>>2]|0;
  HEAP32[$vararg_buffer>>2] = $12;
  $vararg_ptr1 = ((($vararg_buffer)) + 8|0);
  HEAPF64[$vararg_ptr1>>3] = $__v;
  $13 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,30,$10,$__fmt,$vararg_buffer)|0);
  $__nc$0 = $13;
 } else {
  HEAPF64[$vararg_buffer2>>3] = $__v;
  $14 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,30,$10,$__fmt,$vararg_buffer2)|0);
  $__nc$0 = $14;
 }
 $15 = ($__nc$0|0)>(29);
 do {
  if ($15) {
   if ($9) {
    __THREW__ = 0;
    $16 = (invoke_i(4874)|0);
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if (!($18)) {
     $19 = ((($__iob)) + 8|0);
     $20 = HEAP32[$19>>2]|0;
     __THREW__ = 0;
     HEAP32[$vararg_buffer5>>2] = $20;
     $vararg_ptr8 = ((($vararg_buffer5)) + 8|0);
     HEAPF64[$vararg_ptr8>>3] = $__v;
     $21 = (invoke_iiiii(4879,($__nb|0),($16|0),($__fmt|0),($vararg_buffer5|0))|0);
     $22 = __THREW__; __THREW__ = 0;
     $23 = $22&1;
     if (!($23)) {
      $__nc$1 = $21;
      label = 12;
     }
    }
   } else {
    __THREW__ = 0;
    $26 = (invoke_i(4874)|0);
    $27 = __THREW__; __THREW__ = 0;
    $28 = $27&1;
    if (!($28)) {
     $29 = ((($__iob)) + 8|0);
     $30 = HEAP32[$29>>2]|0;
     __THREW__ = 0;
     HEAP32[$vararg_buffer9>>2] = $30;
     $vararg_ptr12 = ((($vararg_buffer9)) + 8|0);
     HEAPF64[$vararg_ptr12>>3] = $__v;
     $31 = (invoke_iiiii(4879,($__nb|0),($26|0),($__fmt|0),($vararg_buffer9|0))|0);
     $32 = __THREW__; __THREW__ = 0;
     $33 = $32&1;
     if (!($33)) {
      $__nc$1 = $31;
      label = 12;
     }
    }
   }
   if ((label|0) == 12) {
    $34 = HEAP32[$__nb>>2]|0;
    $35 = ($34|0)==(0|0);
    if (!($35)) {
     $36 = $34;
     $$pre1012 = $34;$__nbh$sroa$0$0 = $36;$__nc$2 = $__nc$1;
     label = 16;
     break;
    }
    __THREW__ = 0;
    invoke_v(4863);
    $37 = __THREW__; __THREW__ = 0;
    $38 = $37&1;
    if (!($38)) {
     $$pre = HEAP32[$__nb>>2]|0;
     $39 = $$pre;
     $$pre1012 = $39;$__nbh$sroa$0$0 = $$pre;$__nc$2 = $__nc$1;
     label = 16;
     break;
    }
   }
   $24 = ___cxa_find_matching_catch_2()|0;
   $25 = tempRet0;
   $$117 = $24;$$1316 = $25;
  } else {
   $$pre9 = HEAP32[$__nb>>2]|0;
   $$pre1012 = $$pre9;$__nbh$sroa$0$0 = 0;$__nc$2 = $__nc$0;
   label = 16;
  }
 } while(0);
 if ((label|0) == 16) {
  $40 = (($$pre1012) + ($__nc$2)|0);
  $41 = (__ZNSt3__114__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($$pre1012,$40,$__iob)|0);
  $42 = ($$pre1012|0)==($__nar|0);
  do {
   if ($42) {
    $54 = $__nar;$__ob$0 = $__o;$__obh$sroa$0$1 = 0;
    label = 22;
   } else {
    $43 = $__nc$2 << 1;
    $44 = (_malloc($43)|0);
    $45 = ($44|0)==(0|0);
    if ($45) {
     __THREW__ = 0;
     invoke_v(4863);
     $46 = __THREW__; __THREW__ = 0;
     $47 = $46&1;
     if ($47) {
      $__obh$sroa$0$0 = 0;
      label = 20;
      break;
     }
     $$pre10$pre = HEAP32[$__nb>>2]|0;
     $$pre10 = $$pre10$pre;
    } else {
     $$pre10 = $$pre1012;
    }
    $50 = $44;
    $54 = $$pre10;$__ob$0 = $44;$__obh$sroa$0$1 = $50;
    label = 22;
   }
  } while(0);
  do {
   if ((label|0) == 22) {
    __THREW__ = 0;
    $51 = (invoke_ii(3373,($__iob|0))|0);
    $52 = __THREW__; __THREW__ = 0;
    $53 = $52&1;
    if ($53) {
     $__obh$sroa$0$0 = $__obh$sroa$0$1;
     label = 20;
    } else {
     HEAP32[$0>>2] = $51;
     __THREW__ = 0;
     invoke_viiiiiii(4880,($54|0),($41|0),($40|0),($__ob$0|0),($__op|0),($__oe|0),($0|0));
     $55 = __THREW__; __THREW__ = 0;
     $56 = $55&1;
     if ($56) {
      $67 = ___cxa_find_matching_catch_2()|0;
      $68 = tempRet0;
      (__ZNSt3__114__shared_count16__release_sharedEv($51)|0);
      $$0 = $67;$$02 = $68;$__obh$sroa$0$2 = $__obh$sroa$0$1;
      break;
     }
     (__ZNSt3__114__shared_count16__release_sharedEv($51)|0);
     $57 = HEAP32[$__s>>2]|0;
     HEAP32[$1>>2] = $57;
     $58 = HEAP32[$__op>>2]|0;
     $59 = HEAP32[$__oe>>2]|0;
     __THREW__ = 0;
     ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
     $60 = (invoke_iiiiiii(3375,($$byval_copy|0),($__ob$0|0),($58|0),($59|0),($__iob|0),($__fl|0))|0);
     $61 = __THREW__; __THREW__ = 0;
     $62 = $61&1;
     if ($62) {
      $__obh$sroa$0$0 = $__obh$sroa$0$1;
      label = 20;
     } else {
      HEAP32[$__s>>2] = $60;
      $63 = ($__obh$sroa$0$1|0)==(0);
      if (!($63)) {
       $64 = $__obh$sroa$0$1;
       _free($64);
      }
      $65 = ($__nbh$sroa$0$0|0)==(0);
      if (!($65)) {
       $66 = $__nbh$sroa$0$0;
       _free($66);
      }
      STACKTOP = sp;return ($60|0);
     }
    }
   }
  } while(0);
  if ((label|0) == 20) {
   $48 = ___cxa_find_matching_catch_2()|0;
   $49 = tempRet0;
   $$0 = $48;$$02 = $49;$__obh$sroa$0$2 = $__obh$sroa$0$0;
  }
  $69 = ($__obh$sroa$0$2|0)==(0);
  if (!($69)) {
   $70 = $__obh$sroa$0$2;
   _free($70);
  }
  $71 = ($__nbh$sroa$0$0|0)==(0);
  if ($71) {
   $$117 = $$0;$$1316 = $$02;
  } else {
   $72 = $__nbh$sroa$0$0;
   _free($72);
   $$117 = $$0;$$1316 = $$02;
  }
 }
 ___resumeException($$117|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__114__num_put_base14__format_floatEPcPKcj($__fmtp,$__len,$__flags) {
 $__fmtp = $__fmtp|0;
 $__len = $__len|0;
 $__flags = $__flags|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$2 = 0, $$2$lcssa = 0, $$2$ph = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $specify_precision$0$off0$ph = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $__flags & 2048;
 $1 = ($0|0)==(0);
 if ($1) {
  $$0 = $__fmtp;
 } else {
  $2 = ((($__fmtp)) + 1|0);
  HEAP8[$__fmtp>>0] = 43;
  $$0 = $2;
 }
 $3 = $__flags & 1024;
 $4 = ($3|0)==(0);
 if ($4) {
  $$1 = $$0;
 } else {
  $5 = ((($$0)) + 1|0);
  HEAP8[$$0>>0] = 35;
  $$1 = $5;
 }
 $6 = $__flags & 260;
 $7 = $__flags >>> 14;
 $8 = ($6|0)==(260);
 if ($8) {
  $$2$ph = $$1;$specify_precision$0$off0$ph = 0;
 } else {
  $9 = ((($$1)) + 1|0);
  HEAP8[$$1>>0] = 46;
  $10 = ((($$1)) + 2|0);
  HEAP8[$9>>0] = 42;
  $$2$ph = $10;$specify_precision$0$off0$ph = 1;
 }
 $$01 = $__len;$$2 = $$2$ph;
 while(1) {
  $11 = HEAP8[$$01>>0]|0;
  $12 = ($11<<24>>24)==(0);
  if ($12) {
   $$2$lcssa = $$2;
   break;
  }
  $13 = ((($$01)) + 1|0);
  $14 = ((($$2)) + 1|0);
  HEAP8[$$2>>0] = $11;
  $$01 = $13;$$2 = $14;
 }
 L14: do {
  switch ($6|0) {
  case 4:  {
   $15 = $7 & 1;
   $16 = ($15|0)==(0);
   if ($16) {
    HEAP8[$$2$lcssa>>0] = 102;
    break L14;
   } else {
    HEAP8[$$2$lcssa>>0] = 70;
    break L14;
   }
   break;
  }
  case 256:  {
   $17 = $7 & 1;
   $18 = ($17|0)==(0);
   if ($18) {
    HEAP8[$$2$lcssa>>0] = 101;
    break L14;
   } else {
    HEAP8[$$2$lcssa>>0] = 69;
    break L14;
   }
   break;
  }
  default: {
   $19 = $7 & 1;
   $20 = ($19|0)!=(0);
   if ($8) {
    if ($20) {
     HEAP8[$$2$lcssa>>0] = 65;
     break L14;
    } else {
     HEAP8[$$2$lcssa>>0] = 97;
     break L14;
    }
   } else {
    if ($20) {
     HEAP8[$$2$lcssa>>0] = 71;
     break L14;
    } else {
     HEAP8[$$2$lcssa>>0] = 103;
     break L14;
    }
   }
  }
  }
 } while(0);
 return ($specify_precision$0$off0$ph|0);
}
function __ZNSt3__112__asprintf_lEPPcP15__locale_structPKcz($__s,$__l,$__format,$varargs) {
 $__s = $__s|0;
 $__l = $__l|0;
 $__format = $__format|0;
 $varargs = $varargs|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__va = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__va = sp;
 HEAP32[$__va>>2] = $varargs;
 $0 = (_uselocale($__l)|0);
 __THREW__ = 0;
 $1 = (invoke_iiii(4881,($__s|0),($__format|0),($__va|0))|0);
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  $11 = ($0|0)==(0|0);
  if (!($11)) {
   __THREW__ = 0;
   (invoke_ii(4876,($0|0))|0);
   $12 = __THREW__; __THREW__ = 0;
   $13 = $12&1;
   if ($13) {
    $14 = ___cxa_find_matching_catch_3(0|0)|0;
    $15 = tempRet0;
    ___clang_call_terminate($14);
    // unreachable;
   }
  }
  ___resumeException($9|0);
  // unreachable;
 } else {
  $4 = ($0|0)==(0|0);
  if (!($4)) {
   __THREW__ = 0;
   (invoke_ii(4876,($0|0))|0);
   $5 = __THREW__; __THREW__ = 0;
   $6 = $5&1;
   if ($6) {
    $7 = ___cxa_find_matching_catch_3(0|0)|0;
    $8 = tempRet0;
    ___clang_call_terminate($7);
    // unreachable;
   }
  }
  STACKTOP = sp;return ($1|0);
 }
 return (0)|0;
}
function __ZNSt3__19__num_putIcE23__widen_and_group_floatEPcS2_S2_S2_RS2_S3_RKNS_6localeE($__nb,$__np,$__ne,$__ob,$__op,$__oe,$__loc) {
 $__nb = $__nb|0;
 $__np = $__np|0;
 $__ne = $__ne|0;
 $__ob = $__ob|0;
 $__op = $__op|0;
 $__oe = $__oe|0;
 $__loc = $__loc|0;
 var $$0$i$i = 0, $$0$i$i6 = 0, $$01$i$i = 0, $$01$i$i5 = 0, $$__dg$0 = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__dc$0 = 0, $__dc$1 = 0, $__dg$0 = 0, $__dg$1 = 0, $__grouping = 0, $__nf$0 = 0;
 var $__nf$1 = 0, $__nf$2 = 0, $__nf$2$lcssa46 = 0, $__nf$3 = 0, $__ns$0 = 0, $__ns$1 = 0, $__ns$2 = 0, $__p$0 = 0, $lpad$phi$index = 0, $lpad$phi$index5 = 0, $or$cond = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__grouping = sp;
 $0 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1207576)|0);
 $1 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1207588)|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 FUNCTION_TABLE_vii[$4 & 8191]($__grouping,$1);
 HEAP32[$__oe>>2] = $__ob;
 $5 = HEAP8[$__nb>>0]|0;
 switch ($5<<24>>24) {
 case 43: case 45:  {
  $6 = HEAP32[$0>>2]|0;
  $7 = ((($6)) + 28|0);
  $8 = HEAP32[$7>>2]|0;
  __THREW__ = 0;
  $9 = (invoke_iii($8|0,($0|0),($5|0))|0);
  $10 = __THREW__; __THREW__ = 0;
  $11 = $10&1;
  if ($11) {
   label = 8;
  } else {
   $12 = ((($__nb)) + 1|0);
   $13 = HEAP32[$__oe>>2]|0;
   $14 = ((($13)) + 1|0);
   HEAP32[$__oe>>2] = $14;
   HEAP8[$13>>0] = $9;
   $__nf$0 = $12;
   label = 10;
  }
  break;
 }
 default: {
  $__nf$0 = $__nb;
  label = 10;
 }
 }
 L4: do {
  if ((label|0) == 10) {
   $25 = $__ne;
   $26 = $__nf$0;
   $27 = (($25) - ($26))|0;
   $28 = ($27|0)>(1);
   L6: do {
    if ($28) {
     $29 = HEAP8[$__nf$0>>0]|0;
     $30 = ($29<<24>>24)==(48);
     if ($30) {
      $31 = ((($__nf$0)) + 1|0);
      $32 = HEAP8[$31>>0]|0;
      switch ($32<<24>>24) {
      case 88: case 120:  {
       break;
      }
      default: {
       $__ns$1 = $__nf$0;
       label = 21;
       break L6;
      }
      }
      $33 = HEAP32[$0>>2]|0;
      $34 = ((($33)) + 28|0);
      $35 = HEAP32[$34>>2]|0;
      __THREW__ = 0;
      $36 = (invoke_iii($35|0,($0|0),48)|0);
      $37 = __THREW__; __THREW__ = 0;
      $38 = $37&1;
      if ($38) {
       label = 8;
       break L4;
      }
      $39 = HEAP32[$__oe>>2]|0;
      $40 = ((($39)) + 1|0);
      HEAP32[$__oe>>2] = $40;
      HEAP8[$39>>0] = $36;
      $41 = ((($__nf$0)) + 2|0);
      $42 = HEAP8[$31>>0]|0;
      $43 = HEAP32[$0>>2]|0;
      $44 = ((($43)) + 28|0);
      $45 = HEAP32[$44>>2]|0;
      __THREW__ = 0;
      $46 = (invoke_iii($45|0,($0|0),($42|0))|0);
      $47 = __THREW__; __THREW__ = 0;
      $48 = $47&1;
      if ($48) {
       label = 8;
       break L4;
      }
      $49 = HEAP32[$__oe>>2]|0;
      $50 = ((($49)) + 1|0);
      HEAP32[$__oe>>2] = $50;
      HEAP8[$49>>0] = $46;
      $__ns$0 = $41;
      while(1) {
       $51 = ($__ns$0>>>0)<($__ne>>>0);
       if (!($51)) {
        $__nf$1 = $41;$__ns$2 = $__ns$0;
        break L6;
       }
       $52 = HEAP8[$__ns$0>>0]|0;
       __THREW__ = 0;
       $53 = (invoke_i(4874)|0);
       $54 = __THREW__; __THREW__ = 0;
       $55 = $54&1;
       if ($55) {
        break;
       }
       $56 = $52 << 24 >> 24;
       __THREW__ = 0;
       $57 = (invoke_iii(4882,($56|0),($53|0))|0);
       $58 = __THREW__; __THREW__ = 0;
       $59 = $58&1;
       if ($59) {
        break;
       }
       $60 = ($57|0)==(0);
       if ($60) {
        $__nf$1 = $41;$__ns$2 = $__ns$0;
        break L6;
       }
       $61 = ((($__ns$0)) + 1|0);
       $__ns$0 = $61;
      }
      $21 = ___cxa_find_matching_catch_2()|0;
      $22 = tempRet0;
      $lpad$phi$index = $21;$lpad$phi$index5 = $22;
      break L4;
     } else {
      $__ns$1 = $__nf$0;
      label = 21;
     }
    } else {
     $__ns$1 = $__nf$0;
     label = 21;
    }
   } while(0);
   L19: do {
    if ((label|0) == 21) {
     while(1) {
      label = 0;
      $62 = ($__ns$1>>>0)<($__ne>>>0);
      if (!($62)) {
       $__nf$1 = $__nf$0;$__ns$2 = $__ns$1;
       break L19;
      }
      $63 = HEAP8[$__ns$1>>0]|0;
      __THREW__ = 0;
      $64 = (invoke_i(4874)|0);
      $65 = __THREW__; __THREW__ = 0;
      $66 = $65&1;
      if ($66) {
       break;
      }
      $67 = $63 << 24 >> 24;
      __THREW__ = 0;
      $68 = (invoke_iii(4883,($67|0),($64|0))|0);
      $69 = __THREW__; __THREW__ = 0;
      $70 = $69&1;
      if ($70) {
       break;
      }
      $71 = ($68|0)==(0);
      if ($71) {
       $__nf$1 = $__nf$0;$__ns$2 = $__ns$1;
       break L19;
      }
      $72 = ((($__ns$1)) + 1|0);
      $__ns$1 = $72;
      label = 21;
     }
     $19 = ___cxa_find_matching_catch_2()|0;
     $20 = tempRet0;
     $lpad$phi$index = $19;$lpad$phi$index5 = $20;
     break L4;
    }
   } while(0);
   $73 = HEAP8[$__grouping>>0]|0;
   $74 = $73 & 1;
   $75 = ($74<<24>>24)==(0);
   $76 = ((($__grouping)) + 4|0);
   $77 = HEAP32[$76>>2]|0;
   $78 = $73&255;
   $79 = $78 >>> 1;
   $80 = $75 ? $79 : $77;
   $81 = ($80|0)==(0);
   L28: do {
    if ($81) {
     $82 = HEAP32[$__oe>>2]|0;
     $83 = HEAP32[$0>>2]|0;
     $84 = ((($83)) + 32|0);
     $85 = HEAP32[$84>>2]|0;
     __THREW__ = 0;
     (invoke_iiiii($85|0,($0|0),($__nf$1|0),($__ns$2|0),($82|0))|0);
     $86 = __THREW__; __THREW__ = 0;
     $87 = $86&1;
     if ($87) {
      label = 8;
      break L4;
     }
     $88 = $__ns$2;
     $89 = $__nf$1;
     $90 = (($88) - ($89))|0;
     $91 = HEAP32[$__oe>>2]|0;
     $92 = (($91) + ($90)|0);
     HEAP32[$__oe>>2] = $92;
     $$pre$phiZ2D = $0;
    } else {
     $93 = ($__nf$1|0)==($__ns$2|0);
     L32: do {
      if (!($93)) {
       $$0$i$i = $__ns$2;$$01$i$i = $__nf$1;
       while(1) {
        $94 = ((($$0$i$i)) + -1|0);
        $95 = ($$01$i$i>>>0)<($94>>>0);
        if (!($95)) {
         break L32;
        }
        $96 = HEAP8[$$01$i$i>>0]|0;
        $97 = HEAP8[$94>>0]|0;
        HEAP8[$$01$i$i>>0] = $97;
        HEAP8[$94>>0] = $96;
        $98 = ((($$01$i$i)) + 1|0);
        $$0$i$i = $94;$$01$i$i = $98;
       }
      }
     } while(0);
     $99 = HEAP32[$1>>2]|0;
     $100 = ((($99)) + 16|0);
     $101 = HEAP32[$100>>2]|0;
     __THREW__ = 0;
     $102 = (invoke_ii($101|0,($1|0))|0);
     $103 = __THREW__; __THREW__ = 0;
     $104 = $103&1;
     if ($104) {
      label = 8;
      break L4;
     }
     $105 = ((($__grouping)) + 8|0);
     $106 = ((($__grouping)) + 1|0);
     $__dc$0 = 0;$__dg$0 = 0;$__p$0 = $__nf$1;
     while(1) {
      $107 = ($__p$0>>>0)<($__ns$2>>>0);
      if (!($107)) {
       break;
      }
      $119 = HEAP8[$__grouping>>0]|0;
      $120 = $119 & 1;
      $121 = ($120<<24>>24)==(0);
      $122 = HEAP32[$105>>2]|0;
      $123 = $121 ? $106 : $122;
      $124 = (($123) + ($__dg$0)|0);
      $125 = HEAP8[$124>>0]|0;
      $126 = ($125<<24>>24)>(0);
      $127 = $125 << 24 >> 24;
      $128 = ($__dc$0|0)==($127|0);
      $or$cond = $126 & $128;
      if ($or$cond) {
       $129 = HEAP32[$__oe>>2]|0;
       $130 = ((($129)) + 1|0);
       HEAP32[$__oe>>2] = $130;
       HEAP8[$129>>0] = $102;
       $131 = HEAP8[$__grouping>>0]|0;
       $132 = $131 & 1;
       $133 = ($132<<24>>24)==(0);
       $134 = HEAP32[$76>>2]|0;
       $135 = $131&255;
       $136 = $135 >>> 1;
       $137 = $133 ? $136 : $134;
       $138 = (($137) + -1)|0;
       $139 = ($__dg$0>>>0)<($138>>>0);
       $140 = $139&1;
       $$__dg$0 = (($140) + ($__dg$0))|0;
       $__dc$1 = 0;$__dg$1 = $$__dg$0;
      } else {
       $__dc$1 = $__dc$0;$__dg$1 = $__dg$0;
      }
      $141 = HEAP8[$__p$0>>0]|0;
      $142 = HEAP32[$0>>2]|0;
      $143 = ((($142)) + 28|0);
      $144 = HEAP32[$143>>2]|0;
      __THREW__ = 0;
      $145 = (invoke_iii($144|0,($0|0),($141|0))|0);
      $146 = __THREW__; __THREW__ = 0;
      $147 = $146&1;
      if ($147) {
       label = 5;
       break;
      }
      $148 = HEAP32[$__oe>>2]|0;
      $149 = ((($148)) + 1|0);
      HEAP32[$__oe>>2] = $149;
      HEAP8[$148>>0] = $145;
      $150 = (($__dc$1) + 1)|0;
      $151 = ((($__p$0)) + 1|0);
      $__dc$0 = $150;$__dg$0 = $__dg$1;$__p$0 = $151;
     }
     if ((label|0) == 5) {
      $17 = ___cxa_find_matching_catch_2()|0;
      $18 = tempRet0;
      $lpad$phi$index = $17;$lpad$phi$index5 = $18;
      break L4;
     }
     $108 = $__nf$1;
     $109 = $__nb;
     $110 = (($108) - ($109))|0;
     $111 = (($__ob) + ($110)|0);
     $112 = HEAP32[$__oe>>2]|0;
     $113 = ($111|0)==($112|0);
     if ($113) {
      $$pre$phiZ2D = $0;
     } else {
      $$0$i$i6 = $112;$$01$i$i5 = $111;
      while(1) {
       $114 = ((($$0$i$i6)) + -1|0);
       $115 = ($$01$i$i5>>>0)<($114>>>0);
       if (!($115)) {
        $$pre$phiZ2D = $0;
        break L28;
       }
       $116 = HEAP8[$$01$i$i5>>0]|0;
       $117 = HEAP8[$114>>0]|0;
       HEAP8[$$01$i$i5>>0] = $117;
       HEAP8[$114>>0] = $116;
       $118 = ((($$01$i$i5)) + 1|0);
       $$0$i$i6 = $114;$$01$i$i5 = $118;
      }
     }
    }
   } while(0);
   $__nf$2 = $__ns$2;
   while(1) {
    $152 = ($__nf$2>>>0)<($__ne>>>0);
    if (!($152)) {
     $__nf$3 = $__nf$2;
     break;
    }
    $153 = HEAP8[$__nf$2>>0]|0;
    $154 = ($153<<24>>24)==(46);
    if ($154) {
     $__nf$2$lcssa46 = $__nf$2;
     label = 45;
     break;
    }
    $164 = HEAP32[$$pre$phiZ2D>>2]|0;
    $165 = ((($164)) + 28|0);
    $166 = HEAP32[$165>>2]|0;
    __THREW__ = 0;
    $167 = (invoke_iii($166|0,($0|0),($153|0))|0);
    $168 = __THREW__; __THREW__ = 0;
    $169 = $168&1;
    if ($169) {
     label = 4;
     break;
    }
    $170 = HEAP32[$__oe>>2]|0;
    $171 = ((($170)) + 1|0);
    HEAP32[$__oe>>2] = $171;
    HEAP8[$170>>0] = $167;
    $172 = ((($__nf$2)) + 1|0);
    $__nf$2 = $172;
   }
   if ((label|0) == 4) {
    $15 = ___cxa_find_matching_catch_2()|0;
    $16 = tempRet0;
    $lpad$phi$index = $15;$lpad$phi$index5 = $16;
    break;
   }
   if ((label|0) == 45) {
    $155 = HEAP32[$1>>2]|0;
    $156 = ((($155)) + 12|0);
    $157 = HEAP32[$156>>2]|0;
    __THREW__ = 0;
    $158 = (invoke_ii($157|0,($1|0))|0);
    $159 = __THREW__; __THREW__ = 0;
    $160 = $159&1;
    if ($160) {
     label = 8;
     break;
    }
    $161 = HEAP32[$__oe>>2]|0;
    $162 = ((($161)) + 1|0);
    HEAP32[$__oe>>2] = $162;
    HEAP8[$161>>0] = $158;
    $163 = ((($__nf$2$lcssa46)) + 1|0);
    $__nf$3 = $163;
   }
   $173 = HEAP32[$__oe>>2]|0;
   $174 = HEAP32[$0>>2]|0;
   $175 = ((($174)) + 32|0);
   $176 = HEAP32[$175>>2]|0;
   __THREW__ = 0;
   (invoke_iiiii($176|0,($0|0),($__nf$3|0),($__ne|0),($173|0))|0);
   $177 = __THREW__; __THREW__ = 0;
   $178 = $177&1;
   if ($178) {
    label = 8;
   } else {
    $179 = $__nf$3;
    $180 = (($25) - ($179))|0;
    $181 = HEAP32[$__oe>>2]|0;
    $182 = (($181) + ($180)|0);
    HEAP32[$__oe>>2] = $182;
    $183 = ($__np|0)==($__ne|0);
    $184 = $__np;
    $185 = $__nb;
    $186 = (($184) - ($185))|0;
    $187 = (($__ob) + ($186)|0);
    $storemerge = $183 ? $182 : $187;
    HEAP32[$__op>>2] = $storemerge;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return;
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $23 = ___cxa_find_matching_catch_2()|0;
  $24 = tempRet0;
  $lpad$phi$index = $23;$lpad$phi$index5 = $24;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZNKSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEce($this,$__s,$__iob,$__fl,$__v) {
 $this = $this|0;
 $__s = $__s|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $__v = +$__v;
 var $$0 = 0, $$02 = 0, $$117 = 0, $$1316 = 0, $$byval_copy = 0, $$pre = 0, $$pre10 = 0, $$pre10$pre = 0, $$pre1012 = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $8 = 0;
 var $9 = 0, $__fmt = 0, $__nar = 0, $__nb = 0, $__nbh$sroa$0$0 = 0, $__nc$0 = 0, $__nc$1 = 0, $__nc$2 = 0, $__o = 0, $__ob$0 = 0, $__obh$sroa$0$0 = 0, $__obh$sroa$0$1 = 0, $__obh$sroa$0$2 = 0, $__oe = 0, $__op = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_buffer5 = 0, $vararg_buffer9 = 0, $vararg_ptr1 = 0;
 var $vararg_ptr8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 76|0;
 $vararg_buffer9 = sp + 48|0;
 $vararg_buffer5 = sp + 32|0;
 $vararg_buffer2 = sp + 24|0;
 $vararg_buffer = sp + 8|0;
 $__fmt = sp;
 $__nar = sp + 138|0;
 $__nb = sp + 72|0;
 $__o = sp + 80|0;
 $__op = sp + 68|0;
 $__oe = sp + 64|0;
 $0 = sp + 60|0;
 $1 = sp + 56|0;
 $2 = $__fmt;
 $3 = $2;
 HEAP32[$3>>2] = 37;
 $4 = (($2) + 4)|0;
 $5 = $4;
 HEAP32[$5>>2] = 0;
 $6 = ((($__fmt)) + 1|0);
 $7 = ((($__iob)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (__ZNSt3__114__num_put_base14__format_floatEPcPKcj($6,620740,$8)|0);
 HEAP32[$__nb>>2] = $__nar;
 $10 = (__ZNSt3__16__clocEv()|0);
 if ($9) {
  $11 = ((($__iob)) + 8|0);
  $12 = HEAP32[$11>>2]|0;
  HEAP32[$vararg_buffer>>2] = $12;
  $vararg_ptr1 = ((($vararg_buffer)) + 8|0);
  HEAPF64[$vararg_ptr1>>3] = $__v;
  $13 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,30,$10,$__fmt,$vararg_buffer)|0);
  $__nc$0 = $13;
 } else {
  HEAPF64[$vararg_buffer2>>3] = $__v;
  $14 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,30,$10,$__fmt,$vararg_buffer2)|0);
  $__nc$0 = $14;
 }
 $15 = ($__nc$0|0)>(29);
 do {
  if ($15) {
   if ($9) {
    __THREW__ = 0;
    $16 = (invoke_i(4874)|0);
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if (!($18)) {
     $19 = ((($__iob)) + 8|0);
     $20 = HEAP32[$19>>2]|0;
     __THREW__ = 0;
     HEAP32[$vararg_buffer5>>2] = $20;
     $vararg_ptr8 = ((($vararg_buffer5)) + 8|0);
     HEAPF64[$vararg_ptr8>>3] = $__v;
     $21 = (invoke_iiiii(4879,($__nb|0),($16|0),($__fmt|0),($vararg_buffer5|0))|0);
     $22 = __THREW__; __THREW__ = 0;
     $23 = $22&1;
     if (!($23)) {
      $__nc$1 = $21;
      label = 12;
     }
    }
   } else {
    __THREW__ = 0;
    $26 = (invoke_i(4874)|0);
    $27 = __THREW__; __THREW__ = 0;
    $28 = $27&1;
    if (!($28)) {
     __THREW__ = 0;
     HEAPF64[$vararg_buffer9>>3] = $__v;
     $29 = (invoke_iiiii(4879,($__nb|0),($26|0),($__fmt|0),($vararg_buffer9|0))|0);
     $30 = __THREW__; __THREW__ = 0;
     $31 = $30&1;
     if (!($31)) {
      $__nc$1 = $29;
      label = 12;
     }
    }
   }
   if ((label|0) == 12) {
    $32 = HEAP32[$__nb>>2]|0;
    $33 = ($32|0)==(0|0);
    if (!($33)) {
     $34 = $32;
     $$pre1012 = $32;$__nbh$sroa$0$0 = $34;$__nc$2 = $__nc$1;
     label = 16;
     break;
    }
    __THREW__ = 0;
    invoke_v(4863);
    $35 = __THREW__; __THREW__ = 0;
    $36 = $35&1;
    if (!($36)) {
     $$pre = HEAP32[$__nb>>2]|0;
     $37 = $$pre;
     $$pre1012 = $37;$__nbh$sroa$0$0 = $$pre;$__nc$2 = $__nc$1;
     label = 16;
     break;
    }
   }
   $24 = ___cxa_find_matching_catch_2()|0;
   $25 = tempRet0;
   $$117 = $24;$$1316 = $25;
  } else {
   $$pre9 = HEAP32[$__nb>>2]|0;
   $$pre1012 = $$pre9;$__nbh$sroa$0$0 = 0;$__nc$2 = $__nc$0;
   label = 16;
  }
 } while(0);
 if ((label|0) == 16) {
  $38 = (($$pre1012) + ($__nc$2)|0);
  $39 = (__ZNSt3__114__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($$pre1012,$38,$__iob)|0);
  $40 = ($$pre1012|0)==($__nar|0);
  do {
   if ($40) {
    $52 = $__nar;$__ob$0 = $__o;$__obh$sroa$0$1 = 0;
    label = 22;
   } else {
    $41 = $__nc$2 << 1;
    $42 = (_malloc($41)|0);
    $43 = ($42|0)==(0|0);
    if ($43) {
     __THREW__ = 0;
     invoke_v(4863);
     $44 = __THREW__; __THREW__ = 0;
     $45 = $44&1;
     if ($45) {
      $__obh$sroa$0$0 = 0;
      label = 20;
      break;
     }
     $$pre10$pre = HEAP32[$__nb>>2]|0;
     $$pre10 = $$pre10$pre;
    } else {
     $$pre10 = $$pre1012;
    }
    $48 = $42;
    $52 = $$pre10;$__ob$0 = $42;$__obh$sroa$0$1 = $48;
    label = 22;
   }
  } while(0);
  do {
   if ((label|0) == 22) {
    __THREW__ = 0;
    $49 = (invoke_ii(3373,($__iob|0))|0);
    $50 = __THREW__; __THREW__ = 0;
    $51 = $50&1;
    if ($51) {
     $__obh$sroa$0$0 = $__obh$sroa$0$1;
     label = 20;
    } else {
     HEAP32[$0>>2] = $49;
     __THREW__ = 0;
     invoke_viiiiiii(4880,($52|0),($39|0),($38|0),($__ob$0|0),($__op|0),($__oe|0),($0|0));
     $53 = __THREW__; __THREW__ = 0;
     $54 = $53&1;
     if ($54) {
      $65 = ___cxa_find_matching_catch_2()|0;
      $66 = tempRet0;
      (__ZNSt3__114__shared_count16__release_sharedEv($49)|0);
      $$0 = $65;$$02 = $66;$__obh$sroa$0$2 = $__obh$sroa$0$1;
      break;
     }
     (__ZNSt3__114__shared_count16__release_sharedEv($49)|0);
     $55 = HEAP32[$__s>>2]|0;
     HEAP32[$1>>2] = $55;
     $56 = HEAP32[$__op>>2]|0;
     $57 = HEAP32[$__oe>>2]|0;
     __THREW__ = 0;
     ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
     $58 = (invoke_iiiiiii(3375,($$byval_copy|0),($__ob$0|0),($56|0),($57|0),($__iob|0),($__fl|0))|0);
     $59 = __THREW__; __THREW__ = 0;
     $60 = $59&1;
     if ($60) {
      $__obh$sroa$0$0 = $__obh$sroa$0$1;
      label = 20;
     } else {
      HEAP32[$__s>>2] = $58;
      $61 = ($__obh$sroa$0$1|0)==(0);
      if (!($61)) {
       $62 = $__obh$sroa$0$1;
       _free($62);
      }
      $63 = ($__nbh$sroa$0$0|0)==(0);
      if (!($63)) {
       $64 = $__nbh$sroa$0$0;
       _free($64);
      }
      STACKTOP = sp;return ($58|0);
     }
    }
   }
  } while(0);
  if ((label|0) == 20) {
   $46 = ___cxa_find_matching_catch_2()|0;
   $47 = tempRet0;
   $$0 = $46;$$02 = $47;$__obh$sroa$0$2 = $__obh$sroa$0$0;
  }
  $67 = ($__obh$sroa$0$2|0)==(0);
  if (!($67)) {
   $68 = $__obh$sroa$0$2;
   _free($68);
  }
  $69 = ($__nbh$sroa$0$0|0)==(0);
  if ($69) {
   $$117 = $$0;$$1316 = $$02;
  } else {
   $70 = $__nbh$sroa$0$0;
   _free($70);
   $$117 = $$0;$$1316 = $$02;
  }
 }
 ___resumeException($$117|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcPKv($this,$__s,$__iob,$__fl,$__v) {
 $this = $this|0;
 $__s = $__s|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__fmt = 0, $__nar = 0, $__o = 0, $__op$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 12|0;
 $vararg_buffer = sp;
 $__fmt = sp + 76|0;
 $__nar = sp + 56|0;
 $__o = sp + 16|0;
 $0 = sp + 8|0;
 $1 = sp + 4|0;
 ;HEAP8[$__fmt>>0]=HEAP8[620742>>0]|0;HEAP8[$__fmt+1>>0]=HEAP8[620742+1>>0]|0;HEAP8[$__fmt+2>>0]=HEAP8[620742+2>>0]|0;HEAP8[$__fmt+3>>0]=HEAP8[620742+3>>0]|0;HEAP8[$__fmt+4>>0]=HEAP8[620742+4>>0]|0;HEAP8[$__fmt+5>>0]=HEAP8[620742+5>>0]|0;
 $2 = (__ZNSt3__16__clocEv()|0);
 HEAP32[$vararg_buffer>>2] = $__v;
 $3 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,20,$2,$__fmt,$vararg_buffer)|0);
 $4 = (($__nar) + ($3)|0);
 $5 = (__ZNSt3__114__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($__nar,$4,$__iob)|0);
 $6 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$0>>2] = $6;
 __THREW__ = 0;
 $7 = (invoke_iii(3374,($0|0),(1207576|0))|0);
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($6)|0);
  ___resumeException($21|0);
  // unreachable;
 } else {
  (__ZNSt3__114__shared_count16__release_sharedEv($6)|0);
  $10 = HEAP32[$7>>2]|0;
  $11 = ((($10)) + 32|0);
  $12 = HEAP32[$11>>2]|0;
  (FUNCTION_TABLE_iiiii[$12 & 8191]($7,$__nar,$4,$__o)|0);
  $13 = (($__o) + ($3)|0);
  $14 = ($5|0)==($4|0);
  $15 = $__nar;
  $16 = $5;
  $17 = (($16) - ($15))|0;
  $18 = (($__o) + ($17)|0);
  $__op$0 = $14 ? $13 : $18;
  $19 = HEAP32[$__s>>2]|0;
  HEAP32[$1>>2] = $19;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
  $20 = (__ZNSt3__116__pad_and_outputIcNS_11char_traitsIcEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$__o,$__op$0,$13,$__iob,$__fl)|0);
  STACKTOP = sp;return ($20|0);
 }
 return (0)|0;
}
function __ZNSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwb($this,$__s,$__iob,$__fl,$__v) {
 $this = $this|0;
 $__s = $__s|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $__v = $__v|0;
 var $$0 = 0, $$0$i$i = 0, $$01 = 0, $$byval_copy = 0, $$pre = 0, $$pre3 = 0, $$sroa$0$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__i$sroa$0$0 = 0, $__i$sroa$0$0$ph = 0, $__nm = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 20|0;
 $0 = sp + 16|0;
 $1 = sp + 12|0;
 $__nm = sp;
 $2 = ((($__iob)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 1;
 $5 = ($4|0)==(0);
 L1: do {
  if ($5) {
   $6 = HEAP32[$this>>2]|0;
   $7 = ((($6)) + 24|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = HEAP32[$__s>>2]|0;
   HEAP32[$0>>2] = $9;
   $10 = $__v&1;
   ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
   $11 = (FUNCTION_TABLE_iiiiii[$8 & 8191]($this,$$byval_copy,$__iob,$__fl,$10)|0);
   $$sroa$0$0 = $11;
  } else {
   $12 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
   HEAP32[$1>>2] = $12;
   __THREW__ = 0;
   $13 = (invoke_iii(3374,($1|0),(1207616|0))|0);
   $14 = __THREW__; __THREW__ = 0;
   $15 = $14&1;
   do {
    if ($15) {
     $39 = ___cxa_find_matching_catch_2()|0;
     $40 = tempRet0;
     (__ZNSt3__114__shared_count16__release_sharedEv($12)|0);
     $$0 = $40;$$01 = $39;
    } else {
     (__ZNSt3__114__shared_count16__release_sharedEv($12)|0);
     $16 = HEAP32[$13>>2]|0;
     if ($__v) {
      $17 = ((($16)) + 24|0);
      $18 = HEAP32[$17>>2]|0;
      FUNCTION_TABLE_vii[$18 & 8191]($__nm,$13);
     } else {
      $19 = ((($16)) + 28|0);
      $20 = HEAP32[$19>>2]|0;
      FUNCTION_TABLE_vii[$20 & 8191]($__nm,$13);
     }
     $21 = HEAP8[$__nm>>0]|0;
     $22 = $21 & 1;
     $23 = ($22<<24>>24)==(0);
     $24 = ((($__nm)) + 4|0);
     $25 = ((($__nm)) + 8|0);
     $26 = HEAP32[$25>>2]|0;
     $__i$sroa$0$0$ph = $23 ? $24 : $26;
     $28 = $21;$34 = $26;$__i$sroa$0$0 = $__i$sroa$0$0$ph;
     while(1) {
      $27 = $28 & 1;
      $29 = ($27<<24>>24)==(0);
      $30 = HEAP32[$24>>2]|0;
      $31 = $28&255;
      $32 = $31 >>> 1;
      $33 = $29 ? $24 : $34;
      $35 = $29 ? $32 : $30;
      $36 = (($33) + ($35<<2)|0);
      $37 = ($__i$sroa$0$0|0)==($36|0);
      if ($37) {
       label = 9;
       break;
      }
      $41 = HEAP32[$__i$sroa$0$0>>2]|0;
      $42 = HEAP32[$__s>>2]|0;
      $43 = ($42|0)==(0|0);
      if (!($43)) {
       $44 = ((($42)) + 24|0);
       $45 = HEAP32[$44>>2]|0;
       $46 = ((($42)) + 28|0);
       $47 = HEAP32[$46>>2]|0;
       $48 = ($45|0)==($47|0);
       if ($48) {
        $49 = HEAP32[$42>>2]|0;
        $50 = ((($49)) + 52|0);
        $51 = HEAP32[$50>>2]|0;
        __THREW__ = 0;
        $52 = (invoke_iii($51|0,($42|0),($41|0))|0);
        $53 = __THREW__; __THREW__ = 0;
        $54 = $53&1;
        if ($54) {
         label = 18;
         break;
        } else {
         $$0$i$i = $52;
        }
       } else {
        $55 = ((($45)) + 4|0);
        HEAP32[$44>>2] = $55;
        HEAP32[$45>>2] = $41;
        $$0$i$i = $41;
       }
       $56 = ($$0$i$i|0)==(-1);
       if ($56) {
        HEAP32[$__s>>2] = 0;
       }
      }
      $57 = ((($__i$sroa$0$0)) + 4|0);
      $$pre = HEAP8[$__nm>>0]|0;
      $$pre3 = HEAP32[$25>>2]|0;
      $28 = $$pre;$34 = $$pre3;$__i$sroa$0$0 = $57;
     }
     if ((label|0) == 9) {
      $38 = HEAP32[$__s>>2]|0;
      __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__nm);
      $$sroa$0$0 = $38;
      break L1;
     }
     else if ((label|0) == 18) {
      $58 = ___cxa_find_matching_catch_2()|0;
      $59 = tempRet0;
      __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__nm);
      $$0 = $59;$$01 = $58;
      break;
     }
    }
   } while(0);
   ___resumeException($$01|0);
   // unreachable;
  }
 } while(0);
 STACKTOP = sp;return ($$sroa$0$0|0);
}
function __ZNKSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwl($this,$__s,$__iob,$__fl,$__v) {
 $this = $this|0;
 $__s = $__s|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__fmt = 0;
 var $__nar = 0, $__o = 0, $__oe = 0, $__op = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 104|0;
 $vararg_buffer = sp;
 $__fmt = sp + 120|0;
 $__nar = sp + 108|0;
 $__o = sp + 20|0;
 $__op = sp + 16|0;
 $__oe = sp + 12|0;
 $0 = sp + 8|0;
 $1 = sp + 4|0;
 ;HEAP8[$__fmt>>0]=HEAP8[620729>>0]|0;HEAP8[$__fmt+1>>0]=HEAP8[620729+1>>0]|0;HEAP8[$__fmt+2>>0]=HEAP8[620729+2>>0]|0;HEAP8[$__fmt+3>>0]=HEAP8[620729+3>>0]|0;HEAP8[$__fmt+4>>0]=HEAP8[620729+4>>0]|0;HEAP8[$__fmt+5>>0]=HEAP8[620729+5>>0]|0;
 $2 = ((($__fmt)) + 1|0);
 $3 = ((($__iob)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 __ZNSt3__114__num_put_base12__format_intEPcPKcbj($2,620735,1,$4);
 $5 = (__ZNSt3__16__clocEv()|0);
 HEAP32[$vararg_buffer>>2] = $__v;
 $6 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,12,$5,$__fmt,$vararg_buffer)|0);
 $7 = (($__nar) + ($6)|0);
 $8 = (__ZNSt3__114__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($__nar,$7,$__iob)|0);
 $9 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$0>>2] = $9;
 __THREW__ = 0;
 invoke_viiiiiii(4884,($__nar|0),($8|0),($7|0),($__o|0),($__op|0),($__oe|0),($0|0));
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($9)|0);
  ___resumeException($16|0);
  // unreachable;
 } else {
  (__ZNSt3__114__shared_count16__release_sharedEv($9)|0);
  $12 = HEAP32[$__s>>2]|0;
  HEAP32[$1>>2] = $12;
  $13 = HEAP32[$__op>>2]|0;
  $14 = HEAP32[$__oe>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
  $15 = (__ZNSt3__116__pad_and_outputIwNS_11char_traitsIwEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$__o,$13,$14,$__iob,$__fl)|0);
  STACKTOP = sp;return ($15|0);
 }
 return (0)|0;
}
function __ZNSt3__19__num_putIwE21__widen_and_group_intEPcS2_S2_PwRS3_S4_RKNS_6localeE($__nb,$__np,$__ne,$__ob,$__op,$__oe,$__loc) {
 $__nb = $__nb|0;
 $__np = $__np|0;
 $__ne = $__ne|0;
 $__ob = $__ob|0;
 $__op = $__op|0;
 $__oe = $__oe|0;
 $__loc = $__loc|0;
 var $$0$i$i = 0, $$0$i$i6 = 0, $$01$i$i = 0, $$01$i$i5 = 0, $$__dg$0 = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__dc$0 = 0, $__dc$1 = 0, $__dg$0 = 0, $__dg$1 = 0, $__grouping = 0, $__nf$0 = 0, $__nf$1 = 0;
 var $__p$0 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0, $or$cond = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__grouping = sp;
 $0 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1207608)|0);
 $1 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1207616)|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 FUNCTION_TABLE_vii[$4 & 8191]($__grouping,$1);
 $5 = HEAP8[$__grouping>>0]|0;
 $6 = $5 & 1;
 $7 = ($6<<24>>24)==(0);
 $8 = ((($__grouping)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $5&255;
 $11 = $10 >>> 1;
 $12 = $7 ? $11 : $9;
 $13 = ($12|0)==(0);
 L1: do {
  if ($13) {
   $14 = HEAP32[$0>>2]|0;
   $15 = ((($14)) + 48|0);
   $16 = HEAP32[$15>>2]|0;
   __THREW__ = 0;
   (invoke_iiiii($16|0,($0|0),($__nb|0),($__ne|0),($__ob|0))|0);
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if ($18) {
    label = 5;
   } else {
    $19 = $__ne;
    $20 = $__nb;
    $21 = (($19) - ($20))|0;
    $22 = (($__ob) + ($21<<2)|0);
    HEAP32[$__oe>>2] = $22;
    $$pre$phiZ2D = $20;$126 = $22;
    label = 29;
   }
  } else {
   HEAP32[$__oe>>2] = $__ob;
   $27 = HEAP8[$__nb>>0]|0;
   switch ($27<<24>>24) {
   case 43: case 45:  {
    $28 = HEAP32[$0>>2]|0;
    $29 = ((($28)) + 44|0);
    $30 = HEAP32[$29>>2]|0;
    __THREW__ = 0;
    $31 = (invoke_iii($30|0,($0|0),($27|0))|0);
    $32 = __THREW__; __THREW__ = 0;
    $33 = $32&1;
    if ($33) {
     label = 5;
     break L1;
    }
    $34 = ((($__nb)) + 1|0);
    $35 = HEAP32[$__oe>>2]|0;
    $36 = ((($35)) + 4|0);
    HEAP32[$__oe>>2] = $36;
    HEAP32[$35>>2] = $31;
    $__nf$0 = $34;
    break;
   }
   default: {
    $__nf$0 = $__nb;
   }
   }
   $37 = $__ne;
   $38 = $__nf$0;
   $39 = (($37) - ($38))|0;
   $40 = ($39|0)>(1);
   L9: do {
    if ($40) {
     $41 = HEAP8[$__nf$0>>0]|0;
     $42 = ($41<<24>>24)==(48);
     if ($42) {
      $43 = ((($__nf$0)) + 1|0);
      $44 = HEAP8[$43>>0]|0;
      switch ($44<<24>>24) {
      case 88: case 120:  {
       break;
      }
      default: {
       $__nf$1 = $__nf$0;
       break L9;
      }
      }
      $45 = HEAP32[$0>>2]|0;
      $46 = ((($45)) + 44|0);
      $47 = HEAP32[$46>>2]|0;
      __THREW__ = 0;
      $48 = (invoke_iii($47|0,($0|0),48)|0);
      $49 = __THREW__; __THREW__ = 0;
      $50 = $49&1;
      if ($50) {
       label = 5;
       break L1;
      }
      $51 = HEAP32[$__oe>>2]|0;
      $52 = ((($51)) + 4|0);
      HEAP32[$__oe>>2] = $52;
      HEAP32[$51>>2] = $48;
      $53 = HEAP8[$43>>0]|0;
      $54 = HEAP32[$0>>2]|0;
      $55 = ((($54)) + 44|0);
      $56 = HEAP32[$55>>2]|0;
      __THREW__ = 0;
      $57 = (invoke_iii($56|0,($0|0),($53|0))|0);
      $58 = __THREW__; __THREW__ = 0;
      $59 = $58&1;
      if ($59) {
       label = 5;
       break L1;
      }
      $60 = ((($__nf$0)) + 2|0);
      $61 = HEAP32[$__oe>>2]|0;
      $62 = ((($61)) + 4|0);
      HEAP32[$__oe>>2] = $62;
      HEAP32[$61>>2] = $57;
      $__nf$1 = $60;
     } else {
      $__nf$1 = $__nf$0;
     }
    } else {
     $__nf$1 = $__nf$0;
    }
   } while(0);
   $63 = ($__nf$1|0)==($__ne|0);
   L16: do {
    if (!($63)) {
     $$0$i$i = $__ne;$$01$i$i = $__nf$1;
     while(1) {
      $64 = ((($$0$i$i)) + -1|0);
      $65 = ($$01$i$i>>>0)<($64>>>0);
      if (!($65)) {
       break L16;
      }
      $66 = HEAP8[$$01$i$i>>0]|0;
      $67 = HEAP8[$64>>0]|0;
      HEAP8[$$01$i$i>>0] = $67;
      HEAP8[$64>>0] = $66;
      $68 = ((($$01$i$i)) + 1|0);
      $$0$i$i = $64;$$01$i$i = $68;
     }
    }
   } while(0);
   $69 = HEAP32[$1>>2]|0;
   $70 = ((($69)) + 16|0);
   $71 = HEAP32[$70>>2]|0;
   __THREW__ = 0;
   $72 = (invoke_ii($71|0,($1|0))|0);
   $73 = __THREW__; __THREW__ = 0;
   $74 = $73&1;
   if ($74) {
    label = 5;
   } else {
    $75 = ((($__grouping)) + 8|0);
    $76 = ((($__grouping)) + 1|0);
    $__dc$0 = 0;$__dg$0 = 0;$__p$0 = $__nf$1;
    while(1) {
     $77 = ($__p$0>>>0)<($__ne>>>0);
     if (!($77)) {
      break;
     }
     $89 = HEAP8[$__grouping>>0]|0;
     $90 = $89 & 1;
     $91 = ($90<<24>>24)==(0);
     $92 = HEAP32[$75>>2]|0;
     $93 = $91 ? $76 : $92;
     $94 = (($93) + ($__dg$0)|0);
     $95 = HEAP8[$94>>0]|0;
     $96 = ($95<<24>>24)!=(0);
     $97 = $95 << 24 >> 24;
     $98 = ($__dc$0|0)==($97|0);
     $or$cond = $96 & $98;
     if ($or$cond) {
      $99 = HEAP32[$__oe>>2]|0;
      $100 = ((($99)) + 4|0);
      HEAP32[$__oe>>2] = $100;
      HEAP32[$99>>2] = $72;
      $101 = HEAP8[$__grouping>>0]|0;
      $102 = $101 & 1;
      $103 = ($102<<24>>24)==(0);
      $104 = HEAP32[$8>>2]|0;
      $105 = $101&255;
      $106 = $105 >>> 1;
      $107 = $103 ? $106 : $104;
      $108 = (($107) + -1)|0;
      $109 = ($__dg$0>>>0)<($108>>>0);
      $110 = $109&1;
      $$__dg$0 = (($110) + ($__dg$0))|0;
      $__dc$1 = 0;$__dg$1 = $$__dg$0;
     } else {
      $__dc$1 = $__dc$0;$__dg$1 = $__dg$0;
     }
     $111 = HEAP8[$__p$0>>0]|0;
     $112 = HEAP32[$0>>2]|0;
     $113 = ((($112)) + 44|0);
     $114 = HEAP32[$113>>2]|0;
     __THREW__ = 0;
     $115 = (invoke_iii($114|0,($0|0),($111|0))|0);
     $116 = __THREW__; __THREW__ = 0;
     $117 = $116&1;
     if ($117) {
      label = 4;
      break;
     }
     $118 = HEAP32[$__oe>>2]|0;
     $119 = ((($118)) + 4|0);
     HEAP32[$__oe>>2] = $119;
     HEAP32[$118>>2] = $115;
     $120 = (($__dc$1) + 1)|0;
     $121 = ((($__p$0)) + 1|0);
     $__dc$0 = $120;$__dg$0 = $__dg$1;$__p$0 = $121;
    }
    if ((label|0) == 4) {
     $23 = ___cxa_find_matching_catch_2()|0;
     $24 = tempRet0;
     $lpad$phi$index = $23;$lpad$phi$index2 = $24;
     break;
    }
    $78 = $__nf$1;
    $79 = $__nb;
    $80 = (($78) - ($79))|0;
    $81 = (($__ob) + ($80<<2)|0);
    $82 = HEAP32[$__oe>>2]|0;
    $83 = ($81|0)==($82|0);
    if ($83) {
     $$pre$phiZ2D = $79;$126 = $81;
     label = 29;
    } else {
     $$0$i$i6 = $82;$$01$i$i5 = $81;
     while(1) {
      $84 = ((($$0$i$i6)) + -4|0);
      $85 = ($$01$i$i5>>>0)<($84>>>0);
      if (!($85)) {
       $$pre$phiZ2D = $79;$126 = $82;
       label = 29;
       break L1;
      }
      $86 = HEAP32[$$01$i$i5>>2]|0;
      $87 = HEAP32[$84>>2]|0;
      HEAP32[$$01$i$i5>>2] = $87;
      HEAP32[$84>>2] = $86;
      $88 = ((($$01$i$i5)) + 4|0);
      $$0$i$i6 = $84;$$01$i$i5 = $88;
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 5) {
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  $lpad$phi$index = $25;$lpad$phi$index2 = $26;
 }
 else if ((label|0) == 29) {
  $122 = ($__np|0)==($__ne|0);
  $123 = $__np;
  $124 = (($123) - ($$pre$phiZ2D))|0;
  $125 = (($__ob) + ($124<<2)|0);
  $storemerge = $122 ? $126 : $125;
  HEAP32[$__op>>2] = $storemerge;
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
  STACKTOP = sp;return;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZNSt3__116__pad_and_outputIwNS_11char_traitsIwEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($__s,$__ob,$__op,$__oe,$__iob,$__fl) {
 $__s = $__s|0;
 $__ob = $__ob|0;
 $__op = $__op|0;
 $__oe = $__oe|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 var $$sroa$0$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__ns$0 = 0, $__sp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__sp = sp;
 $0 = HEAP32[$__s>>2]|0;
 $1 = ($0|0)==(0|0);
 L1: do {
  if ($1) {
   $$sroa$0$2 = 0;
  } else {
   $2 = $__oe;
   $3 = $__ob;
   $4 = (($2) - ($3))|0;
   $5 = $4 >> 2;
   $6 = ((($__iob)) + 12|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)>($5|0);
   $9 = (($7) - ($5))|0;
   $__ns$0 = $8 ? $9 : 0;
   $10 = $__op;
   $11 = (($10) - ($3))|0;
   $12 = $11 >> 2;
   $13 = ($11|0)>(0);
   if ($13) {
    $14 = HEAP32[$0>>2]|0;
    $15 = ((($14)) + 48|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = (FUNCTION_TABLE_iiii[$16 & 8191]($0,$__ob,$12)|0);
    $18 = ($17|0)==($12|0);
    if (!($18)) {
     HEAP32[$__s>>2] = 0;
     $$sroa$0$2 = 0;
     break;
    }
   }
   $19 = ($__ns$0|0)>(0);
   do {
    if ($19) {
     __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6__initEjw($__sp,$__ns$0,$__fl);
     $20 = HEAP8[$__sp>>0]|0;
     $21 = $20 & 1;
     $22 = ($21<<24>>24)==(0);
     $23 = ((($__sp)) + 8|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ((($__sp)) + 4|0);
     $26 = $22 ? $25 : $24;
     $27 = HEAP32[$0>>2]|0;
     $28 = ((($27)) + 48|0);
     $29 = HEAP32[$28>>2]|0;
     __THREW__ = 0;
     $30 = (invoke_iiii($29|0,($0|0),($26|0),($__ns$0|0))|0);
     $31 = __THREW__; __THREW__ = 0;
     $32 = $31&1;
     if ($32) {
      $34 = ___cxa_find_matching_catch_2()|0;
      $35 = tempRet0;
      __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__sp);
      ___resumeException($34|0);
      // unreachable;
     }
     $33 = ($30|0)==($__ns$0|0);
     if ($33) {
      __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__sp);
      break;
     } else {
      HEAP32[$__s>>2] = 0;
      __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__sp);
      $$sroa$0$2 = 0;
      break L1;
     }
    }
   } while(0);
   $36 = (($2) - ($10))|0;
   $37 = $36 >> 2;
   $38 = ($36|0)>(0);
   if ($38) {
    $39 = HEAP32[$0>>2]|0;
    $40 = ((($39)) + 48|0);
    $41 = HEAP32[$40>>2]|0;
    $42 = (FUNCTION_TABLE_iiii[$41 & 8191]($0,$__op,$37)|0);
    $43 = ($42|0)==($37|0);
    if (!($43)) {
     HEAP32[$__s>>2] = 0;
     $$sroa$0$2 = 0;
     break;
    }
   }
   HEAP32[$6>>2] = 0;
   $$sroa$0$2 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$sroa$0$2|0);
}
function __ZNKSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwx($this,$__s,$__iob,$__fl,$0,$1) {
 $this = $this|0;
 $__s = $__s|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__fmt = 0, $__nar = 0, $__o = 0, $__oe = 0, $__op = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 196|0;
 $vararg_buffer = sp + 8|0;
 $__fmt = sp;
 $__nar = sp + 200|0;
 $__o = sp + 32|0;
 $__op = sp + 28|0;
 $__oe = sp + 24|0;
 $2 = sp + 20|0;
 $3 = sp + 16|0;
 $4 = $__fmt;
 $5 = $4;
 HEAP32[$5>>2] = 37;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = ((($__fmt)) + 1|0);
 $9 = ((($__iob)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 __ZNSt3__114__num_put_base12__format_intEPcPKcbj($8,620737,1,$10);
 $11 = (__ZNSt3__16__clocEv()|0);
 $12 = $vararg_buffer;
 $13 = $12;
 HEAP32[$13>>2] = $0;
 $14 = (($12) + 4)|0;
 $15 = $14;
 HEAP32[$15>>2] = $1;
 $16 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,22,$11,$__fmt,$vararg_buffer)|0);
 $17 = (($__nar) + ($16)|0);
 $18 = (__ZNSt3__114__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($__nar,$17,$__iob)|0);
 $19 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$2>>2] = $19;
 __THREW__ = 0;
 invoke_viiiiiii(4884,($__nar|0),($18|0),($17|0),($__o|0),($__op|0),($__oe|0),($2|0));
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($19)|0);
  ___resumeException($26|0);
  // unreachable;
 } else {
  (__ZNSt3__114__shared_count16__release_sharedEv($19)|0);
  $22 = HEAP32[$__s>>2]|0;
  HEAP32[$3>>2] = $22;
  $23 = HEAP32[$__op>>2]|0;
  $24 = HEAP32[$__oe>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$3>>2]|0;
  $25 = (__ZNSt3__116__pad_and_outputIwNS_11char_traitsIwEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$__o,$23,$24,$__iob,$__fl)|0);
  STACKTOP = sp;return ($25|0);
 }
 return (0)|0;
}
function __ZNKSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwm($this,$__s,$__iob,$__fl,$__v) {
 $this = $this|0;
 $__s = $__s|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__fmt = 0;
 var $__nar = 0, $__o = 0, $__oe = 0, $__op = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 104|0;
 $vararg_buffer = sp;
 $__fmt = sp + 120|0;
 $__nar = sp + 108|0;
 $__o = sp + 20|0;
 $__op = sp + 16|0;
 $__oe = sp + 12|0;
 $0 = sp + 8|0;
 $1 = sp + 4|0;
 ;HEAP8[$__fmt>>0]=HEAP8[620729>>0]|0;HEAP8[$__fmt+1>>0]=HEAP8[620729+1>>0]|0;HEAP8[$__fmt+2>>0]=HEAP8[620729+2>>0]|0;HEAP8[$__fmt+3>>0]=HEAP8[620729+3>>0]|0;HEAP8[$__fmt+4>>0]=HEAP8[620729+4>>0]|0;HEAP8[$__fmt+5>>0]=HEAP8[620729+5>>0]|0;
 $2 = ((($__fmt)) + 1|0);
 $3 = ((($__iob)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 __ZNSt3__114__num_put_base12__format_intEPcPKcbj($2,620735,0,$4);
 $5 = (__ZNSt3__16__clocEv()|0);
 HEAP32[$vararg_buffer>>2] = $__v;
 $6 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,12,$5,$__fmt,$vararg_buffer)|0);
 $7 = (($__nar) + ($6)|0);
 $8 = (__ZNSt3__114__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($__nar,$7,$__iob)|0);
 $9 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$0>>2] = $9;
 __THREW__ = 0;
 invoke_viiiiiii(4884,($__nar|0),($8|0),($7|0),($__o|0),($__op|0),($__oe|0),($0|0));
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($9)|0);
  ___resumeException($16|0);
  // unreachable;
 } else {
  (__ZNSt3__114__shared_count16__release_sharedEv($9)|0);
  $12 = HEAP32[$__s>>2]|0;
  HEAP32[$1>>2] = $12;
  $13 = HEAP32[$__op>>2]|0;
  $14 = HEAP32[$__oe>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
  $15 = (__ZNSt3__116__pad_and_outputIwNS_11char_traitsIwEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$__o,$13,$14,$__iob,$__fl)|0);
  STACKTOP = sp;return ($15|0);
 }
 return (0)|0;
}
function __ZNKSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwy($this,$__s,$__iob,$__fl,$0,$1) {
 $this = $this|0;
 $__s = $__s|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $0 = $0|0;
 $1 = $1|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__fmt = 0, $__nar = 0, $__o = 0, $__oe = 0, $__op = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 204|0;
 $vararg_buffer = sp + 8|0;
 $__fmt = sp;
 $__nar = sp + 208|0;
 $__o = sp + 32|0;
 $__op = sp + 28|0;
 $__oe = sp + 24|0;
 $2 = sp + 20|0;
 $3 = sp + 16|0;
 $4 = $__fmt;
 $5 = $4;
 HEAP32[$5>>2] = 37;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = ((($__fmt)) + 1|0);
 $9 = ((($__iob)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 __ZNSt3__114__num_put_base12__format_intEPcPKcbj($8,620737,0,$10);
 $11 = (__ZNSt3__16__clocEv()|0);
 $12 = $vararg_buffer;
 $13 = $12;
 HEAP32[$13>>2] = $0;
 $14 = (($12) + 4)|0;
 $15 = $14;
 HEAP32[$15>>2] = $1;
 $16 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,23,$11,$__fmt,$vararg_buffer)|0);
 $17 = (($__nar) + ($16)|0);
 $18 = (__ZNSt3__114__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($__nar,$17,$__iob)|0);
 $19 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$2>>2] = $19;
 __THREW__ = 0;
 invoke_viiiiiii(4884,($__nar|0),($18|0),($17|0),($__o|0),($__op|0),($__oe|0),($2|0));
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($19)|0);
  ___resumeException($26|0);
  // unreachable;
 } else {
  (__ZNSt3__114__shared_count16__release_sharedEv($19)|0);
  $22 = HEAP32[$__s>>2]|0;
  HEAP32[$3>>2] = $22;
  $23 = HEAP32[$__op>>2]|0;
  $24 = HEAP32[$__oe>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$3>>2]|0;
  $25 = (__ZNSt3__116__pad_and_outputIwNS_11char_traitsIwEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$__o,$23,$24,$__iob,$__fl)|0);
  STACKTOP = sp;return ($25|0);
 }
 return (0)|0;
}
function __ZNKSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwd($this,$__s,$__iob,$__fl,$__v) {
 $this = $this|0;
 $__s = $__s|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $__v = +$__v;
 var $$0 = 0, $$02 = 0, $$115 = 0, $$1314 = 0, $$byval_copy = 0, $$pre = 0, $$pre7 = 0, $$pre8 = 0, $$pre8$pre = 0, $$pre810 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $8 = 0, $9 = 0, $__fmt = 0, $__nar = 0, $__nb = 0, $__nbh$sroa$0$0 = 0, $__nc$0 = 0, $__nc$1 = 0, $__nc$2 = 0, $__o = 0, $__ob$0 = 0, $__obh$sroa$0$0 = 0, $__obh$sroa$0$1 = 0, $__obh$sroa$0$2 = 0, $__oe = 0, $__op = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_buffer5 = 0;
 var $vararg_buffer9 = 0, $vararg_ptr1 = 0, $vararg_ptr12 = 0, $vararg_ptr8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 352|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 312|0;
 $vararg_buffer9 = sp + 48|0;
 $vararg_buffer5 = sp + 32|0;
 $vararg_buffer2 = sp + 24|0;
 $vararg_buffer = sp + 8|0;
 $__fmt = sp;
 $__nar = sp + 316|0;
 $__nb = sp + 308|0;
 $__o = sp + 80|0;
 $__op = sp + 76|0;
 $__oe = sp + 72|0;
 $0 = sp + 68|0;
 $1 = sp + 64|0;
 $2 = $__fmt;
 $3 = $2;
 HEAP32[$3>>2] = 37;
 $4 = (($2) + 4)|0;
 $5 = $4;
 HEAP32[$5>>2] = 0;
 $6 = ((($__fmt)) + 1|0);
 $7 = ((($__iob)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (__ZNSt3__114__num_put_base14__format_floatEPcPKcj($6,1282466,$8)|0);
 HEAP32[$__nb>>2] = $__nar;
 $10 = (__ZNSt3__16__clocEv()|0);
 if ($9) {
  $11 = ((($__iob)) + 8|0);
  $12 = HEAP32[$11>>2]|0;
  HEAP32[$vararg_buffer>>2] = $12;
  $vararg_ptr1 = ((($vararg_buffer)) + 8|0);
  HEAPF64[$vararg_ptr1>>3] = $__v;
  $13 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,30,$10,$__fmt,$vararg_buffer)|0);
  $__nc$0 = $13;
 } else {
  HEAPF64[$vararg_buffer2>>3] = $__v;
  $14 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,30,$10,$__fmt,$vararg_buffer2)|0);
  $__nc$0 = $14;
 }
 $15 = ($__nc$0|0)>(29);
 do {
  if ($15) {
   if ($9) {
    __THREW__ = 0;
    $16 = (invoke_i(4874)|0);
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if (!($18)) {
     $19 = ((($__iob)) + 8|0);
     $20 = HEAP32[$19>>2]|0;
     __THREW__ = 0;
     HEAP32[$vararg_buffer5>>2] = $20;
     $vararg_ptr8 = ((($vararg_buffer5)) + 8|0);
     HEAPF64[$vararg_ptr8>>3] = $__v;
     $21 = (invoke_iiiii(4879,($__nb|0),($16|0),($__fmt|0),($vararg_buffer5|0))|0);
     $22 = __THREW__; __THREW__ = 0;
     $23 = $22&1;
     if (!($23)) {
      $__nc$1 = $21;
      label = 12;
     }
    }
   } else {
    __THREW__ = 0;
    $26 = (invoke_i(4874)|0);
    $27 = __THREW__; __THREW__ = 0;
    $28 = $27&1;
    if (!($28)) {
     $29 = ((($__iob)) + 8|0);
     $30 = HEAP32[$29>>2]|0;
     __THREW__ = 0;
     HEAP32[$vararg_buffer9>>2] = $30;
     $vararg_ptr12 = ((($vararg_buffer9)) + 8|0);
     HEAPF64[$vararg_ptr12>>3] = $__v;
     $31 = (invoke_iiiii(4879,($__nb|0),($26|0),($__fmt|0),($vararg_buffer9|0))|0);
     $32 = __THREW__; __THREW__ = 0;
     $33 = $32&1;
     if (!($33)) {
      $__nc$1 = $31;
      label = 12;
     }
    }
   }
   if ((label|0) == 12) {
    $34 = HEAP32[$__nb>>2]|0;
    $35 = ($34|0)==(0|0);
    if (!($35)) {
     $36 = $34;
     $$pre810 = $34;$__nbh$sroa$0$0 = $36;$__nc$2 = $__nc$1;
     label = 16;
     break;
    }
    __THREW__ = 0;
    invoke_v(4863);
    $37 = __THREW__; __THREW__ = 0;
    $38 = $37&1;
    if (!($38)) {
     $$pre = HEAP32[$__nb>>2]|0;
     $39 = $$pre;
     $$pre810 = $39;$__nbh$sroa$0$0 = $$pre;$__nc$2 = $__nc$1;
     label = 16;
     break;
    }
   }
   $24 = ___cxa_find_matching_catch_2()|0;
   $25 = tempRet0;
   $$115 = $24;$$1314 = $25;
  } else {
   $$pre7 = HEAP32[$__nb>>2]|0;
   $$pre810 = $$pre7;$__nbh$sroa$0$0 = 0;$__nc$2 = $__nc$0;
   label = 16;
  }
 } while(0);
 if ((label|0) == 16) {
  $40 = (($$pre810) + ($__nc$2)|0);
  $41 = (__ZNSt3__114__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($$pre810,$40,$__iob)|0);
  $42 = ($$pre810|0)==($__nar|0);
  do {
   if ($42) {
    $54 = $__nar;$__ob$0 = $__o;$__obh$sroa$0$1 = 0;
    label = 22;
   } else {
    $43 = $__nc$2 << 3;
    $44 = (_malloc($43)|0);
    $45 = ($44|0)==(0|0);
    if ($45) {
     __THREW__ = 0;
     invoke_v(4863);
     $46 = __THREW__; __THREW__ = 0;
     $47 = $46&1;
     if ($47) {
      $__obh$sroa$0$0 = 0;
      label = 20;
      break;
     }
     $$pre8$pre = HEAP32[$__nb>>2]|0;
     $$pre8 = $$pre8$pre;
    } else {
     $$pre8 = $$pre810;
    }
    $50 = $44;
    $54 = $$pre8;$__ob$0 = $44;$__obh$sroa$0$1 = $50;
    label = 22;
   }
  } while(0);
  do {
   if ((label|0) == 22) {
    __THREW__ = 0;
    $51 = (invoke_ii(3373,($__iob|0))|0);
    $52 = __THREW__; __THREW__ = 0;
    $53 = $52&1;
    if ($53) {
     $__obh$sroa$0$0 = $__obh$sroa$0$1;
     label = 20;
    } else {
     HEAP32[$0>>2] = $51;
     __THREW__ = 0;
     invoke_viiiiiii(4885,($54|0),($41|0),($40|0),($__ob$0|0),($__op|0),($__oe|0),($0|0));
     $55 = __THREW__; __THREW__ = 0;
     $56 = $55&1;
     if ($56) {
      $67 = ___cxa_find_matching_catch_2()|0;
      $68 = tempRet0;
      (__ZNSt3__114__shared_count16__release_sharedEv($51)|0);
      $$0 = $67;$$02 = $68;$__obh$sroa$0$2 = $__obh$sroa$0$1;
      break;
     }
     (__ZNSt3__114__shared_count16__release_sharedEv($51)|0);
     $57 = HEAP32[$__s>>2]|0;
     HEAP32[$1>>2] = $57;
     $58 = HEAP32[$__op>>2]|0;
     $59 = HEAP32[$__oe>>2]|0;
     __THREW__ = 0;
     ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
     $60 = (invoke_iiiiiii(4886,($$byval_copy|0),($__ob$0|0),($58|0),($59|0),($__iob|0),($__fl|0))|0);
     $61 = __THREW__; __THREW__ = 0;
     $62 = $61&1;
     if ($62) {
      $__obh$sroa$0$0 = $__obh$sroa$0$1;
      label = 20;
     } else {
      HEAP32[$__s>>2] = $60;
      $63 = ($__obh$sroa$0$1|0)==(0);
      if (!($63)) {
       $64 = $__obh$sroa$0$1;
       _free($64);
      }
      $65 = ($__nbh$sroa$0$0|0)==(0);
      if (!($65)) {
       $66 = $__nbh$sroa$0$0;
       _free($66);
      }
      STACKTOP = sp;return ($60|0);
     }
    }
   }
  } while(0);
  if ((label|0) == 20) {
   $48 = ___cxa_find_matching_catch_2()|0;
   $49 = tempRet0;
   $$0 = $48;$$02 = $49;$__obh$sroa$0$2 = $__obh$sroa$0$0;
  }
  $69 = ($__obh$sroa$0$2|0)==(0);
  if (!($69)) {
   $70 = $__obh$sroa$0$2;
   _free($70);
  }
  $71 = ($__nbh$sroa$0$0|0)==(0);
  if ($71) {
   $$115 = $$0;$$1314 = $$02;
  } else {
   $72 = $__nbh$sroa$0$0;
   _free($72);
   $$115 = $$0;$$1314 = $$02;
  }
 }
 ___resumeException($$115|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__19__num_putIwE23__widen_and_group_floatEPcS2_S2_PwRS3_S4_RKNS_6localeE($__nb,$__np,$__ne,$__ob,$__op,$__oe,$__loc) {
 $__nb = $__nb|0;
 $__np = $__np|0;
 $__ne = $__ne|0;
 $__ob = $__ob|0;
 $__op = $__op|0;
 $__oe = $__oe|0;
 $__loc = $__loc|0;
 var $$0$i$i = 0, $$0$i$i6 = 0, $$01$i$i = 0, $$01$i$i5 = 0, $$__dg$0 = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__dc$0 = 0, $__dc$1 = 0, $__dg$0 = 0, $__dg$1 = 0;
 var $__grouping = 0, $__nf$0 = 0, $__nf$1 = 0, $__nf$2 = 0, $__nf$2$lcssa53 = 0, $__nf$3 = 0, $__ns$0 = 0, $__ns$1 = 0, $__ns$2 = 0, $__p$0 = 0, $lpad$phi$index = 0, $lpad$phi$index5 = 0, $or$cond = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__grouping = sp;
 $0 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1207608)|0);
 $1 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1207616)|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 FUNCTION_TABLE_vii[$4 & 8191]($__grouping,$1);
 HEAP32[$__oe>>2] = $__ob;
 $5 = HEAP8[$__nb>>0]|0;
 switch ($5<<24>>24) {
 case 43: case 45:  {
  $6 = HEAP32[$0>>2]|0;
  $7 = ((($6)) + 44|0);
  $8 = HEAP32[$7>>2]|0;
  __THREW__ = 0;
  $9 = (invoke_iii($8|0,($0|0),($5|0))|0);
  $10 = __THREW__; __THREW__ = 0;
  $11 = $10&1;
  if ($11) {
   label = 8;
  } else {
   $12 = ((($__nb)) + 1|0);
   $13 = HEAP32[$__oe>>2]|0;
   $14 = ((($13)) + 4|0);
   HEAP32[$__oe>>2] = $14;
   HEAP32[$13>>2] = $9;
   $__nf$0 = $12;
   label = 10;
  }
  break;
 }
 default: {
  $__nf$0 = $__nb;
  label = 10;
 }
 }
 L4: do {
  if ((label|0) == 10) {
   $25 = $__ne;
   $26 = $__nf$0;
   $27 = (($25) - ($26))|0;
   $28 = ($27|0)>(1);
   L6: do {
    if ($28) {
     $29 = HEAP8[$__nf$0>>0]|0;
     $30 = ($29<<24>>24)==(48);
     if ($30) {
      $31 = ((($__nf$0)) + 1|0);
      $32 = HEAP8[$31>>0]|0;
      switch ($32<<24>>24) {
      case 88: case 120:  {
       break;
      }
      default: {
       $__ns$1 = $__nf$0;
       label = 21;
       break L6;
      }
      }
      $33 = HEAP32[$0>>2]|0;
      $34 = ((($33)) + 44|0);
      $35 = HEAP32[$34>>2]|0;
      __THREW__ = 0;
      $36 = (invoke_iii($35|0,($0|0),48)|0);
      $37 = __THREW__; __THREW__ = 0;
      $38 = $37&1;
      if ($38) {
       label = 8;
       break L4;
      }
      $39 = HEAP32[$__oe>>2]|0;
      $40 = ((($39)) + 4|0);
      HEAP32[$__oe>>2] = $40;
      HEAP32[$39>>2] = $36;
      $41 = ((($__nf$0)) + 2|0);
      $42 = HEAP8[$31>>0]|0;
      $43 = HEAP32[$0>>2]|0;
      $44 = ((($43)) + 44|0);
      $45 = HEAP32[$44>>2]|0;
      __THREW__ = 0;
      $46 = (invoke_iii($45|0,($0|0),($42|0))|0);
      $47 = __THREW__; __THREW__ = 0;
      $48 = $47&1;
      if ($48) {
       label = 8;
       break L4;
      }
      $49 = HEAP32[$__oe>>2]|0;
      $50 = ((($49)) + 4|0);
      HEAP32[$__oe>>2] = $50;
      HEAP32[$49>>2] = $46;
      $__ns$0 = $41;
      while(1) {
       $51 = ($__ns$0>>>0)<($__ne>>>0);
       if (!($51)) {
        $__nf$1 = $41;$__ns$2 = $__ns$0;
        break L6;
       }
       $52 = HEAP8[$__ns$0>>0]|0;
       __THREW__ = 0;
       $53 = (invoke_i(4874)|0);
       $54 = __THREW__; __THREW__ = 0;
       $55 = $54&1;
       if ($55) {
        break;
       }
       $56 = $52 << 24 >> 24;
       __THREW__ = 0;
       $57 = (invoke_iii(4882,($56|0),($53|0))|0);
       $58 = __THREW__; __THREW__ = 0;
       $59 = $58&1;
       if ($59) {
        break;
       }
       $60 = ($57|0)==(0);
       if ($60) {
        $__nf$1 = $41;$__ns$2 = $__ns$0;
        break L6;
       }
       $61 = ((($__ns$0)) + 1|0);
       $__ns$0 = $61;
      }
      $21 = ___cxa_find_matching_catch_2()|0;
      $22 = tempRet0;
      $lpad$phi$index = $21;$lpad$phi$index5 = $22;
      break L4;
     } else {
      $__ns$1 = $__nf$0;
      label = 21;
     }
    } else {
     $__ns$1 = $__nf$0;
     label = 21;
    }
   } while(0);
   L19: do {
    if ((label|0) == 21) {
     while(1) {
      label = 0;
      $62 = ($__ns$1>>>0)<($__ne>>>0);
      if (!($62)) {
       $__nf$1 = $__nf$0;$__ns$2 = $__ns$1;
       break L19;
      }
      $63 = HEAP8[$__ns$1>>0]|0;
      __THREW__ = 0;
      $64 = (invoke_i(4874)|0);
      $65 = __THREW__; __THREW__ = 0;
      $66 = $65&1;
      if ($66) {
       break;
      }
      $67 = $63 << 24 >> 24;
      __THREW__ = 0;
      $68 = (invoke_iii(4883,($67|0),($64|0))|0);
      $69 = __THREW__; __THREW__ = 0;
      $70 = $69&1;
      if ($70) {
       break;
      }
      $71 = ($68|0)==(0);
      if ($71) {
       $__nf$1 = $__nf$0;$__ns$2 = $__ns$1;
       break L19;
      }
      $72 = ((($__ns$1)) + 1|0);
      $__ns$1 = $72;
      label = 21;
     }
     $19 = ___cxa_find_matching_catch_2()|0;
     $20 = tempRet0;
     $lpad$phi$index = $19;$lpad$phi$index5 = $20;
     break L4;
    }
   } while(0);
   $73 = HEAP8[$__grouping>>0]|0;
   $74 = $73 & 1;
   $75 = ($74<<24>>24)==(0);
   $76 = ((($__grouping)) + 4|0);
   $77 = HEAP32[$76>>2]|0;
   $78 = $73&255;
   $79 = $78 >>> 1;
   $80 = $75 ? $79 : $77;
   $81 = ($80|0)==(0);
   L28: do {
    if ($81) {
     $82 = HEAP32[$__oe>>2]|0;
     $83 = HEAP32[$0>>2]|0;
     $84 = ((($83)) + 48|0);
     $85 = HEAP32[$84>>2]|0;
     __THREW__ = 0;
     (invoke_iiiii($85|0,($0|0),($__nf$1|0),($__ns$2|0),($82|0))|0);
     $86 = __THREW__; __THREW__ = 0;
     $87 = $86&1;
     if ($87) {
      label = 8;
      break L4;
     }
     $88 = $__ns$2;
     $89 = $__nf$1;
     $90 = (($88) - ($89))|0;
     $91 = HEAP32[$__oe>>2]|0;
     $92 = (($91) + ($90<<2)|0);
     HEAP32[$__oe>>2] = $92;
     $$pre$phiZ2D = $0;$188 = $92;
    } else {
     $93 = ($__nf$1|0)==($__ns$2|0);
     L32: do {
      if (!($93)) {
       $$0$i$i = $__ns$2;$$01$i$i = $__nf$1;
       while(1) {
        $94 = ((($$0$i$i)) + -1|0);
        $95 = ($$01$i$i>>>0)<($94>>>0);
        if (!($95)) {
         break L32;
        }
        $96 = HEAP8[$$01$i$i>>0]|0;
        $97 = HEAP8[$94>>0]|0;
        HEAP8[$$01$i$i>>0] = $97;
        HEAP8[$94>>0] = $96;
        $98 = ((($$01$i$i)) + 1|0);
        $$0$i$i = $94;$$01$i$i = $98;
       }
      }
     } while(0);
     $99 = HEAP32[$1>>2]|0;
     $100 = ((($99)) + 16|0);
     $101 = HEAP32[$100>>2]|0;
     __THREW__ = 0;
     $102 = (invoke_ii($101|0,($1|0))|0);
     $103 = __THREW__; __THREW__ = 0;
     $104 = $103&1;
     if ($104) {
      label = 8;
      break L4;
     }
     $105 = ((($__grouping)) + 8|0);
     $106 = ((($__grouping)) + 1|0);
     $__dc$0 = 0;$__dg$0 = 0;$__p$0 = $__nf$1;
     while(1) {
      $107 = ($__p$0>>>0)<($__ns$2>>>0);
      if (!($107)) {
       break;
      }
      $119 = HEAP8[$__grouping>>0]|0;
      $120 = $119 & 1;
      $121 = ($120<<24>>24)==(0);
      $122 = HEAP32[$105>>2]|0;
      $123 = $121 ? $106 : $122;
      $124 = (($123) + ($__dg$0)|0);
      $125 = HEAP8[$124>>0]|0;
      $126 = ($125<<24>>24)>(0);
      $127 = $125 << 24 >> 24;
      $128 = ($__dc$0|0)==($127|0);
      $or$cond = $126 & $128;
      if ($or$cond) {
       $129 = HEAP32[$__oe>>2]|0;
       $130 = ((($129)) + 4|0);
       HEAP32[$__oe>>2] = $130;
       HEAP32[$129>>2] = $102;
       $131 = HEAP8[$__grouping>>0]|0;
       $132 = $131 & 1;
       $133 = ($132<<24>>24)==(0);
       $134 = HEAP32[$76>>2]|0;
       $135 = $131&255;
       $136 = $135 >>> 1;
       $137 = $133 ? $136 : $134;
       $138 = (($137) + -1)|0;
       $139 = ($__dg$0>>>0)<($138>>>0);
       $140 = $139&1;
       $$__dg$0 = (($140) + ($__dg$0))|0;
       $__dc$1 = 0;$__dg$1 = $$__dg$0;
      } else {
       $__dc$1 = $__dc$0;$__dg$1 = $__dg$0;
      }
      $141 = HEAP8[$__p$0>>0]|0;
      $142 = HEAP32[$0>>2]|0;
      $143 = ((($142)) + 44|0);
      $144 = HEAP32[$143>>2]|0;
      __THREW__ = 0;
      $145 = (invoke_iii($144|0,($0|0),($141|0))|0);
      $146 = __THREW__; __THREW__ = 0;
      $147 = $146&1;
      if ($147) {
       label = 5;
       break;
      }
      $148 = HEAP32[$__oe>>2]|0;
      $149 = ((($148)) + 4|0);
      HEAP32[$__oe>>2] = $149;
      HEAP32[$148>>2] = $145;
      $150 = (($__dc$1) + 1)|0;
      $151 = ((($__p$0)) + 1|0);
      $__dc$0 = $150;$__dg$0 = $__dg$1;$__p$0 = $151;
     }
     if ((label|0) == 5) {
      $17 = ___cxa_find_matching_catch_2()|0;
      $18 = tempRet0;
      $lpad$phi$index = $17;$lpad$phi$index5 = $18;
      break L4;
     }
     $108 = $__nf$1;
     $109 = $__nb;
     $110 = (($108) - ($109))|0;
     $111 = (($__ob) + ($110<<2)|0);
     $112 = HEAP32[$__oe>>2]|0;
     $113 = ($111|0)==($112|0);
     if ($113) {
      $$pre$phiZ2D = $0;$188 = $111;
     } else {
      $$0$i$i6 = $112;$$01$i$i5 = $111;
      while(1) {
       $114 = ((($$0$i$i6)) + -4|0);
       $115 = ($$01$i$i5>>>0)<($114>>>0);
       if (!($115)) {
        $$pre$phiZ2D = $0;$188 = $112;
        break L28;
       }
       $116 = HEAP32[$$01$i$i5>>2]|0;
       $117 = HEAP32[$114>>2]|0;
       HEAP32[$$01$i$i5>>2] = $117;
       HEAP32[$114>>2] = $116;
       $118 = ((($$01$i$i5)) + 4|0);
       $$0$i$i6 = $114;$$01$i$i5 = $118;
      }
     }
    }
   } while(0);
   $189 = $188;$__nf$2 = $__ns$2;
   while(1) {
    $152 = ($__nf$2>>>0)<($__ne>>>0);
    if (!($152)) {
     $176 = $189;$__nf$3 = $__nf$2;
     break;
    }
    $153 = HEAP8[$__nf$2>>0]|0;
    $154 = ($153<<24>>24)==(46);
    if ($154) {
     $__nf$2$lcssa53 = $__nf$2;
     label = 45;
     break;
    }
    $164 = HEAP32[$$pre$phiZ2D>>2]|0;
    $165 = ((($164)) + 44|0);
    $166 = HEAP32[$165>>2]|0;
    __THREW__ = 0;
    $167 = (invoke_iii($166|0,($0|0),($153|0))|0);
    $168 = __THREW__; __THREW__ = 0;
    $169 = $168&1;
    if ($169) {
     label = 4;
     break;
    }
    $170 = HEAP32[$__oe>>2]|0;
    $171 = ((($170)) + 4|0);
    HEAP32[$__oe>>2] = $171;
    HEAP32[$170>>2] = $167;
    $172 = ((($__nf$2)) + 1|0);
    $189 = $171;$__nf$2 = $172;
   }
   if ((label|0) == 4) {
    $15 = ___cxa_find_matching_catch_2()|0;
    $16 = tempRet0;
    $lpad$phi$index = $15;$lpad$phi$index5 = $16;
    break;
   }
   if ((label|0) == 45) {
    $155 = HEAP32[$1>>2]|0;
    $156 = ((($155)) + 12|0);
    $157 = HEAP32[$156>>2]|0;
    __THREW__ = 0;
    $158 = (invoke_ii($157|0,($1|0))|0);
    $159 = __THREW__; __THREW__ = 0;
    $160 = $159&1;
    if ($160) {
     label = 8;
     break;
    }
    $161 = HEAP32[$__oe>>2]|0;
    $162 = ((($161)) + 4|0);
    HEAP32[$__oe>>2] = $162;
    HEAP32[$161>>2] = $158;
    $163 = ((($__nf$2$lcssa53)) + 1|0);
    $176 = $162;$__nf$3 = $163;
   }
   $173 = HEAP32[$0>>2]|0;
   $174 = ((($173)) + 48|0);
   $175 = HEAP32[$174>>2]|0;
   __THREW__ = 0;
   (invoke_iiiii($175|0,($0|0),($__nf$3|0),($__ne|0),($176|0))|0);
   $177 = __THREW__; __THREW__ = 0;
   $178 = $177&1;
   if ($178) {
    label = 8;
   } else {
    $179 = $__nf$3;
    $180 = (($25) - ($179))|0;
    $181 = HEAP32[$__oe>>2]|0;
    $182 = (($181) + ($180<<2)|0);
    HEAP32[$__oe>>2] = $182;
    $183 = ($__np|0)==($__ne|0);
    $184 = $__np;
    $185 = $__nb;
    $186 = (($184) - ($185))|0;
    $187 = (($__ob) + ($186<<2)|0);
    $storemerge = $183 ? $182 : $187;
    HEAP32[$__op>>2] = $storemerge;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
    STACKTOP = sp;return;
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $23 = ___cxa_find_matching_catch_2()|0;
  $24 = tempRet0;
  $lpad$phi$index = $23;$lpad$phi$index5 = $24;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grouping);
 ___resumeException($lpad$phi$index|0);
 // unreachable;
}
function __ZNKSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwe($this,$__s,$__iob,$__fl,$__v) {
 $this = $this|0;
 $__s = $__s|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $__v = +$__v;
 var $$0 = 0, $$02 = 0, $$115 = 0, $$1314 = 0, $$byval_copy = 0, $$pre = 0, $$pre7 = 0, $$pre8 = 0, $$pre8$pre = 0, $$pre810 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $8 = 0;
 var $9 = 0, $__fmt = 0, $__nar = 0, $__nb = 0, $__nbh$sroa$0$0 = 0, $__nc$0 = 0, $__nc$1 = 0, $__nc$2 = 0, $__o = 0, $__ob$0 = 0, $__obh$sroa$0$0 = 0, $__obh$sroa$0$1 = 0, $__obh$sroa$0$2 = 0, $__oe = 0, $__op = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_buffer5 = 0, $vararg_buffer9 = 0, $vararg_ptr1 = 0;
 var $vararg_ptr8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 352|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 304|0;
 $vararg_buffer9 = sp + 48|0;
 $vararg_buffer5 = sp + 32|0;
 $vararg_buffer2 = sp + 24|0;
 $vararg_buffer = sp + 8|0;
 $__fmt = sp;
 $__nar = sp + 308|0;
 $__nb = sp + 300|0;
 $__o = sp + 72|0;
 $__op = sp + 68|0;
 $__oe = sp + 64|0;
 $0 = sp + 60|0;
 $1 = sp + 56|0;
 $2 = $__fmt;
 $3 = $2;
 HEAP32[$3>>2] = 37;
 $4 = (($2) + 4)|0;
 $5 = $4;
 HEAP32[$5>>2] = 0;
 $6 = ((($__fmt)) + 1|0);
 $7 = ((($__iob)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (__ZNSt3__114__num_put_base14__format_floatEPcPKcj($6,620740,$8)|0);
 HEAP32[$__nb>>2] = $__nar;
 $10 = (__ZNSt3__16__clocEv()|0);
 if ($9) {
  $11 = ((($__iob)) + 8|0);
  $12 = HEAP32[$11>>2]|0;
  HEAP32[$vararg_buffer>>2] = $12;
  $vararg_ptr1 = ((($vararg_buffer)) + 8|0);
  HEAPF64[$vararg_ptr1>>3] = $__v;
  $13 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,30,$10,$__fmt,$vararg_buffer)|0);
  $__nc$0 = $13;
 } else {
  HEAPF64[$vararg_buffer2>>3] = $__v;
  $14 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,30,$10,$__fmt,$vararg_buffer2)|0);
  $__nc$0 = $14;
 }
 $15 = ($__nc$0|0)>(29);
 do {
  if ($15) {
   if ($9) {
    __THREW__ = 0;
    $16 = (invoke_i(4874)|0);
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if (!($18)) {
     $19 = ((($__iob)) + 8|0);
     $20 = HEAP32[$19>>2]|0;
     __THREW__ = 0;
     HEAP32[$vararg_buffer5>>2] = $20;
     $vararg_ptr8 = ((($vararg_buffer5)) + 8|0);
     HEAPF64[$vararg_ptr8>>3] = $__v;
     $21 = (invoke_iiiii(4879,($__nb|0),($16|0),($__fmt|0),($vararg_buffer5|0))|0);
     $22 = __THREW__; __THREW__ = 0;
     $23 = $22&1;
     if (!($23)) {
      $__nc$1 = $21;
      label = 12;
     }
    }
   } else {
    __THREW__ = 0;
    $26 = (invoke_i(4874)|0);
    $27 = __THREW__; __THREW__ = 0;
    $28 = $27&1;
    if (!($28)) {
     __THREW__ = 0;
     HEAPF64[$vararg_buffer9>>3] = $__v;
     $29 = (invoke_iiiii(4879,($__nb|0),($26|0),($__fmt|0),($vararg_buffer9|0))|0);
     $30 = __THREW__; __THREW__ = 0;
     $31 = $30&1;
     if (!($31)) {
      $__nc$1 = $29;
      label = 12;
     }
    }
   }
   if ((label|0) == 12) {
    $32 = HEAP32[$__nb>>2]|0;
    $33 = ($32|0)==(0|0);
    if (!($33)) {
     $34 = $32;
     $$pre810 = $32;$__nbh$sroa$0$0 = $34;$__nc$2 = $__nc$1;
     label = 16;
     break;
    }
    __THREW__ = 0;
    invoke_v(4863);
    $35 = __THREW__; __THREW__ = 0;
    $36 = $35&1;
    if (!($36)) {
     $$pre = HEAP32[$__nb>>2]|0;
     $37 = $$pre;
     $$pre810 = $37;$__nbh$sroa$0$0 = $$pre;$__nc$2 = $__nc$1;
     label = 16;
     break;
    }
   }
   $24 = ___cxa_find_matching_catch_2()|0;
   $25 = tempRet0;
   $$115 = $24;$$1314 = $25;
  } else {
   $$pre7 = HEAP32[$__nb>>2]|0;
   $$pre810 = $$pre7;$__nbh$sroa$0$0 = 0;$__nc$2 = $__nc$0;
   label = 16;
  }
 } while(0);
 if ((label|0) == 16) {
  $38 = (($$pre810) + ($__nc$2)|0);
  $39 = (__ZNSt3__114__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($$pre810,$38,$__iob)|0);
  $40 = ($$pre810|0)==($__nar|0);
  do {
   if ($40) {
    $52 = $__nar;$__ob$0 = $__o;$__obh$sroa$0$1 = 0;
    label = 22;
   } else {
    $41 = $__nc$2 << 3;
    $42 = (_malloc($41)|0);
    $43 = ($42|0)==(0|0);
    if ($43) {
     __THREW__ = 0;
     invoke_v(4863);
     $44 = __THREW__; __THREW__ = 0;
     $45 = $44&1;
     if ($45) {
      $__obh$sroa$0$0 = 0;
      label = 20;
      break;
     }
     $$pre8$pre = HEAP32[$__nb>>2]|0;
     $$pre8 = $$pre8$pre;
    } else {
     $$pre8 = $$pre810;
    }
    $48 = $42;
    $52 = $$pre8;$__ob$0 = $42;$__obh$sroa$0$1 = $48;
    label = 22;
   }
  } while(0);
  do {
   if ((label|0) == 22) {
    __THREW__ = 0;
    $49 = (invoke_ii(3373,($__iob|0))|0);
    $50 = __THREW__; __THREW__ = 0;
    $51 = $50&1;
    if ($51) {
     $__obh$sroa$0$0 = $__obh$sroa$0$1;
     label = 20;
    } else {
     HEAP32[$0>>2] = $49;
     __THREW__ = 0;
     invoke_viiiiiii(4885,($52|0),($39|0),($38|0),($__ob$0|0),($__op|0),($__oe|0),($0|0));
     $53 = __THREW__; __THREW__ = 0;
     $54 = $53&1;
     if ($54) {
      $65 = ___cxa_find_matching_catch_2()|0;
      $66 = tempRet0;
      (__ZNSt3__114__shared_count16__release_sharedEv($49)|0);
      $$0 = $65;$$02 = $66;$__obh$sroa$0$2 = $__obh$sroa$0$1;
      break;
     }
     (__ZNSt3__114__shared_count16__release_sharedEv($49)|0);
     $55 = HEAP32[$__s>>2]|0;
     HEAP32[$1>>2] = $55;
     $56 = HEAP32[$__op>>2]|0;
     $57 = HEAP32[$__oe>>2]|0;
     __THREW__ = 0;
     ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
     $58 = (invoke_iiiiiii(4886,($$byval_copy|0),($__ob$0|0),($56|0),($57|0),($__iob|0),($__fl|0))|0);
     $59 = __THREW__; __THREW__ = 0;
     $60 = $59&1;
     if ($60) {
      $__obh$sroa$0$0 = $__obh$sroa$0$1;
      label = 20;
     } else {
      HEAP32[$__s>>2] = $58;
      $61 = ($__obh$sroa$0$1|0)==(0);
      if (!($61)) {
       $62 = $__obh$sroa$0$1;
       _free($62);
      }
      $63 = ($__nbh$sroa$0$0|0)==(0);
      if (!($63)) {
       $64 = $__nbh$sroa$0$0;
       _free($64);
      }
      STACKTOP = sp;return ($58|0);
     }
    }
   }
  } while(0);
  if ((label|0) == 20) {
   $46 = ___cxa_find_matching_catch_2()|0;
   $47 = tempRet0;
   $$0 = $46;$$02 = $47;$__obh$sroa$0$2 = $__obh$sroa$0$0;
  }
  $67 = ($__obh$sroa$0$2|0)==(0);
  if (!($67)) {
   $68 = $__obh$sroa$0$2;
   _free($68);
  }
  $69 = ($__nbh$sroa$0$0|0)==(0);
  if ($69) {
   $$115 = $$0;$$1314 = $$02;
  } else {
   $70 = $__nbh$sroa$0$0;
   _free($70);
   $$115 = $$0;$$1314 = $$02;
  }
 }
 ___resumeException($$115|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwPKv($this,$__s,$__iob,$__fl,$__v) {
 $this = $this|0;
 $__s = $__s|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $__v = $__v|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__fmt = 0, $__nar = 0, $__o = 0, $__op$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 160|0;
 $vararg_buffer = sp;
 $__fmt = sp + 184|0;
 $__nar = sp + 164|0;
 $__o = sp + 12|0;
 $0 = sp + 8|0;
 $1 = sp + 4|0;
 ;HEAP8[$__fmt>>0]=HEAP8[620742>>0]|0;HEAP8[$__fmt+1>>0]=HEAP8[620742+1>>0]|0;HEAP8[$__fmt+2>>0]=HEAP8[620742+2>>0]|0;HEAP8[$__fmt+3>>0]=HEAP8[620742+3>>0]|0;HEAP8[$__fmt+4>>0]=HEAP8[620742+4>>0]|0;HEAP8[$__fmt+5>>0]=HEAP8[620742+5>>0]|0;
 $2 = (__ZNSt3__16__clocEv()|0);
 HEAP32[$vararg_buffer>>2] = $__v;
 $3 = (__ZNSt3__112__snprintf_lEPcjP15__locale_structPKcz($__nar,20,$2,$__fmt,$vararg_buffer)|0);
 $4 = (($__nar) + ($3)|0);
 $5 = (__ZNSt3__114__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($__nar,$4,$__iob)|0);
 $6 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$0>>2] = $6;
 __THREW__ = 0;
 $7 = (invoke_iii(3374,($0|0),(1207608|0))|0);
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($6)|0);
  ___resumeException($21|0);
  // unreachable;
 } else {
  (__ZNSt3__114__shared_count16__release_sharedEv($6)|0);
  $10 = HEAP32[$7>>2]|0;
  $11 = ((($10)) + 48|0);
  $12 = HEAP32[$11>>2]|0;
  (FUNCTION_TABLE_iiiii[$12 & 8191]($7,$__nar,$4,$__o)|0);
  $13 = (($__o) + ($3<<2)|0);
  $14 = ($5|0)==($4|0);
  $15 = $__nar;
  $16 = $5;
  $17 = (($16) - ($15))|0;
  $18 = (($__o) + ($17<<2)|0);
  $__op$0 = $14 ? $13 : $18;
  $19 = HEAP32[$__s>>2]|0;
  HEAP32[$1>>2] = $19;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
  $20 = (__ZNSt3__116__pad_and_outputIwNS_11char_traitsIwEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$__o,$__op$0,$13,$__iob,$__fl)|0);
  STACKTOP = sp;return ($20|0);
 }
 return (0)|0;
}
function __ZNSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE13do_date_orderEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 2;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11do_get_timeES4_S4_RNS_8ios_baseERjP2tm($this,$__b,$__e,$__iob,$__err,$__tm) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__tm = $__tm|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__tm,620960,(620968))|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($this,$__b,$__e,$__iob,$__err,$__tm,$__fmtb,$__fmte) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__tm = $__tm|0;
 $__fmtb = $__fmtb|0;
 $__fmte = $__fmte|0;
 var $$0$i$i = 0, $$0$i$i27 = 0, $$0$ph = 0, $$0$ph$be = 0, $$0$pn = 0, $$3 = 0, $$3$lcssa = 0, $$byval_copy = 0, $$byval_copy1 = 0, $$lcssa112 = 0, $$lcssa115 = 0, $$lcssa116 = 0, $$lcssa116$lcssa = 0, $$lcssa116$lcssa119 = 0, $$lcssa117 = 0, $$lcssa118 = 0, $$lcssa121 = 0, $$lcssa123 = 0, $$lcssa126 = 0, $$lcssa127 = 0;
 var $$lcssa128 = 0, $$pre = 0, $$pre$i$i$i18 = 0, $$pre$i$i$i9 = 0, $$pre$i2$i$i22 = 0, $$pre$i2$i$i2240 = 0, $$pre$i2$i$i2241 = 0, $$pre$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $__cmd$0 = 0, $__opt$0 = 0, $isascii1$i = 0, $isascii1$i25 = 0, $isascii1$i6 = 0, $or$cond = 0, $sext = 0, $tmp$i$i24 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 16|0;
 $$byval_copy = sp + 12|0;
 $0 = sp + 8|0;
 $1 = sp + 4|0;
 $2 = sp;
 $3 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$0>>2] = $3;
 __THREW__ = 0;
 $4 = (invoke_iii(3374,($0|0),(1207576|0))|0);
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $39 = ___cxa_find_matching_catch_2()|0;
  $40 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($3)|0);
  ___resumeException($39|0);
  // unreachable;
 }
 (__ZNSt3__114__shared_count16__release_sharedEv($3)|0);
 HEAP32[$__err>>2] = 0;
 $7 = ((($4)) + 8|0);
 $$0$ph = $__fmtb;$$pre = 0;
 L4: while(1) {
  $8 = ($$0$ph|0)!=($__fmte|0);
  $10 = $$pre;
  while(1) {
   $9 = ($10|0)==(0);
   $or$cond = $8 & $9;
   $11 = HEAP32[$__b>>2]|0;
   if (!($or$cond)) {
    $$pre$i$i$i9 = $11;
    break L4;
   }
   $12 = ($11|0)==(0|0);
   $13 = $11;
   if ($12) {
    $189 = $13;$25 = 0;
   } else {
    $14 = ((($11)) + 12|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = ((($11)) + 16|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = ($15|0)==($17|0);
    if ($18) {
     $19 = HEAP32[$11>>2]|0;
     $20 = ((($19)) + 36|0);
     $21 = HEAP32[$20>>2]|0;
     $22 = (FUNCTION_TABLE_ii[$21 & 8191]($11)|0);
     $23 = ($22|0)==(-1);
     if ($23) {
      HEAP32[$__b>>2] = 0;
      $189 = 0;$25 = 0;
     } else {
      $189 = $13;$25 = $11;
     }
    } else {
     $189 = $13;$25 = $11;
    }
   }
   $24 = ($25|0)==(0|0);
   $26 = HEAP32[$__e>>2]|0;
   $27 = ($26|0)==(0|0);
   $28 = $26;
   do {
    if ($27) {
     $190 = $28;
     label = 14;
    } else {
     $29 = ((($26)) + 12|0);
     $30 = HEAP32[$29>>2]|0;
     $31 = ((($26)) + 16|0);
     $32 = HEAP32[$31>>2]|0;
     $33 = ($30|0)==($32|0);
     if ($33) {
      $34 = HEAP32[$26>>2]|0;
      $35 = ((($34)) + 36|0);
      $36 = HEAP32[$35>>2]|0;
      $37 = (FUNCTION_TABLE_ii[$36 & 8191]($26)|0);
      $38 = ($37|0)==(-1);
      if ($38) {
       HEAP32[$__e>>2] = 0;
       $190 = 0;
       label = 14;
       break;
      }
     }
     if ($24) {
      $191 = $26;$192 = $28;
     } else {
      $$lcssa115 = $25;
      label = 15;
      break L4;
     }
    }
   } while(0);
   if ((label|0) == 14) {
    label = 0;
    if ($24) {
     $$lcssa115 = $25;
     label = 15;
     break L4;
    } else {
     $191 = 0;$192 = $190;
    }
   }
   $41 = HEAP8[$$0$ph>>0]|0;
   $42 = HEAP32[$4>>2]|0;
   $43 = ((($42)) + 36|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = (FUNCTION_TABLE_iiii[$44 & 8191]($4,$41,0)|0);
   $46 = ($45<<24>>24)==(37);
   if ($46) {
    $$lcssa112 = $189;$$lcssa116 = $25;$$lcssa123 = $192;
    label = 18;
    break;
   }
   $67 = HEAP8[$$0$ph>>0]|0;
   $isascii1$i = ($67<<24>>24)>(-1);
   if ($isascii1$i) {
    $68 = $67 << 24 >> 24;
    $69 = HEAP32[$7>>2]|0;
    $70 = (($69) + ($68<<1)|0);
    $71 = HEAP16[$70>>1]|0;
    $72 = $71 & 8192;
    $73 = ($72<<16>>16)==(0);
    if (!($73)) {
     $$lcssa117 = $25;$$lcssa121 = $191;$$lcssa126 = $69;
     label = 27;
     break;
    }
   }
   $132 = ((($25)) + 12|0);
   $133 = HEAP32[$132>>2]|0;
   $134 = ((($25)) + 16|0);
   $135 = HEAP32[$134>>2]|0;
   $136 = ($133|0)==($135|0);
   if ($136) {
    $137 = HEAP32[$25>>2]|0;
    $138 = ((($137)) + 36|0);
    $139 = HEAP32[$138>>2]|0;
    $140 = (FUNCTION_TABLE_ii[$139 & 8191]($25)|0);
    $$0$i$i27 = $140;
   } else {
    $141 = HEAP8[$133>>0]|0;
    $142 = $141&255;
    $$0$i$i27 = $142;
   }
   $143 = $$0$i$i27&255;
   $144 = HEAP32[$4>>2]|0;
   $145 = ((($144)) + 12|0);
   $146 = HEAP32[$145>>2]|0;
   $147 = (FUNCTION_TABLE_iii[$146 & 8191]($4,$143)|0);
   $148 = HEAP8[$$0$ph>>0]|0;
   $149 = HEAP32[$4>>2]|0;
   $150 = ((($149)) + 12|0);
   $151 = HEAP32[$150>>2]|0;
   $152 = (FUNCTION_TABLE_iii[$151 & 8191]($4,$148)|0);
   $153 = ($147<<24>>24)==($152<<24>>24);
   if ($153) {
    $$lcssa118 = $25;$$lcssa127 = $132;$$lcssa128 = $134;
    label = 55;
    break;
   }
   HEAP32[$__err>>2] = 4;
   $10 = 4;
  }
  L32: do {
   if ((label|0) == 18) {
    label = 0;
    $47 = ((($$0$ph)) + 1|0);
    $48 = ($47|0)==($__fmte|0);
    if ($48) {
     $$lcssa116$lcssa = $$lcssa116;
     label = 19;
     break L4;
    }
    $49 = HEAP8[$47>>0]|0;
    $50 = HEAP32[$4>>2]|0;
    $51 = ((($50)) + 36|0);
    $52 = HEAP32[$51>>2]|0;
    $53 = (FUNCTION_TABLE_iiii[$52 & 8191]($4,$49,0)|0);
    switch ($53<<24>>24) {
    case 48: case 69:  {
     $54 = ((($$0$ph)) + 2|0);
     $55 = ($54|0)==($__fmte|0);
     if ($55) {
      $$lcssa116$lcssa119 = $$lcssa116;
      label = 24;
      break L4;
     }
     $56 = HEAP8[$54>>0]|0;
     $57 = HEAP32[$4>>2]|0;
     $58 = ((($57)) + 36|0);
     $59 = HEAP32[$58>>2]|0;
     $60 = (FUNCTION_TABLE_iiii[$59 & 8191]($4,$56,0)|0);
     $66 = $47;$__cmd$0 = $60;$__opt$0 = $53;
     break;
    }
    default: {
     $66 = $$0$ph;$__cmd$0 = $53;$__opt$0 = 0;
    }
    }
    $61 = HEAP32[$this>>2]|0;
    $62 = ((($61)) + 36|0);
    $63 = HEAP32[$62>>2]|0;
    HEAP32[$1>>2] = $$lcssa112;
    HEAP32[$2>>2] = $$lcssa123;
    ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
    ;HEAP32[$$byval_copy1>>2]=HEAP32[$2>>2]|0;
    $64 = (FUNCTION_TABLE_iiiiiiiii[$63 & 8191]($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__tm,$__cmd$0,$__opt$0)|0);
    HEAP32[$__b>>2] = $64;
    $65 = ((($66)) + 2|0);
    $$0$ph$be = $65;
   }
   else if ((label|0) == 27) {
    label = 0;
    $$0$pn = $$0$ph;
    while(1) {
     $$3 = ((($$0$pn)) + 1|0);
     $74 = ($$3|0)==($__fmte|0);
     if ($74) {
      $$3$lcssa = $__fmte;
      break;
     }
     $75 = HEAP8[$$3>>0]|0;
     $isascii1$i6 = ($75<<24>>24)>(-1);
     if (!($isascii1$i6)) {
      $$3$lcssa = $$3;
      break;
     }
     $76 = $75 << 24 >> 24;
     $77 = (($$lcssa126) + ($76<<1)|0);
     $78 = HEAP16[$77>>1]|0;
     $79 = $78 & 8192;
     $80 = ($79<<16>>16)==(0);
     if ($80) {
      $$3$lcssa = $$3;
      break;
     } else {
      $$0$pn = $$3;
     }
    }
    $$pre$i$i$i18 = $$lcssa117;$$pre$i2$i$i22 = $$lcssa121;$95 = $$lcssa121;
    while(1) {
     $81 = ($$pre$i$i$i18|0)==(0|0);
     if ($81) {
      $93 = 0;
     } else {
      $82 = ((($$pre$i$i$i18)) + 12|0);
      $83 = HEAP32[$82>>2]|0;
      $84 = ((($$pre$i$i$i18)) + 16|0);
      $85 = HEAP32[$84>>2]|0;
      $86 = ($83|0)==($85|0);
      if ($86) {
       $87 = HEAP32[$$pre$i$i$i18>>2]|0;
       $88 = ((($87)) + 36|0);
       $89 = HEAP32[$88>>2]|0;
       $90 = (FUNCTION_TABLE_ii[$89 & 8191]($$pre$i$i$i18)|0);
       $91 = ($90|0)==(-1);
       if ($91) {
        HEAP32[$__b>>2] = 0;
        $93 = 0;
       } else {
        $93 = $$pre$i$i$i18;
       }
      } else {
       $93 = $$pre$i$i$i18;
      }
     }
     $92 = ($93|0)==(0|0);
     $94 = ($95|0)==(0|0);
     do {
      if ($94) {
       $$pre$i2$i$i2241 = $$pre$i2$i$i22;
       label = 41;
      } else {
       $96 = ((($95)) + 12|0);
       $97 = HEAP32[$96>>2]|0;
       $98 = ((($95)) + 16|0);
       $99 = HEAP32[$98>>2]|0;
       $100 = ($97|0)==($99|0);
       if (!($100)) {
        if ($92) {
         $$pre$i2$i$i2240 = $$pre$i2$i$i22;$193 = $95;
         break;
        } else {
         $$0$ph$be = $$3$lcssa;
         break L32;
        }
       }
       $101 = HEAP32[$95>>2]|0;
       $102 = ((($101)) + 36|0);
       $103 = HEAP32[$102>>2]|0;
       $104 = (FUNCTION_TABLE_ii[$103 & 8191]($95)|0);
       $105 = ($104|0)==(-1);
       if ($105) {
        HEAP32[$__e>>2] = 0;
        $$pre$i2$i$i2241 = 0;
        label = 41;
        break;
       } else {
        $106 = ($$pre$i2$i$i22|0)==(0|0);
        $tmp$i$i24 = $92 ^ $106;
        if ($tmp$i$i24) {
         $$pre$i2$i$i2240 = $$pre$i2$i$i22;$193 = $$pre$i2$i$i22;
         break;
        } else {
         $$0$ph$be = $$3$lcssa;
         break L32;
        }
       }
      }
     } while(0);
     if ((label|0) == 41) {
      label = 0;
      if ($92) {
       $$0$ph$be = $$3$lcssa;
       break L32;
      } else {
       $$pre$i2$i$i2240 = $$pre$i2$i$i2241;$193 = 0;
      }
     }
     $107 = ((($93)) + 12|0);
     $108 = HEAP32[$107>>2]|0;
     $109 = ((($93)) + 16|0);
     $110 = HEAP32[$109>>2]|0;
     $111 = ($108|0)==($110|0);
     if ($111) {
      $112 = HEAP32[$93>>2]|0;
      $113 = ((($112)) + 36|0);
      $114 = HEAP32[$113>>2]|0;
      $115 = (FUNCTION_TABLE_ii[$114 & 8191]($93)|0);
      $$0$i$i = $115;
     } else {
      $116 = HEAP8[$108>>0]|0;
      $117 = $116&255;
      $$0$i$i = $117;
     }
     $118 = $$0$i$i&255;
     $isascii1$i25 = ($118<<24>>24)>(-1);
     if (!($isascii1$i25)) {
      $$0$ph$be = $$3$lcssa;
      break L32;
     }
     $sext = $$0$i$i << 24;
     $119 = $sext >> 24;
     $120 = HEAP32[$7>>2]|0;
     $121 = (($120) + ($119<<1)|0);
     $122 = HEAP16[$121>>1]|0;
     $123 = $122 & 8192;
     $124 = ($123<<16>>16)==(0);
     if ($124) {
      $$0$ph$be = $$3$lcssa;
      break L32;
     }
     $125 = HEAP32[$107>>2]|0;
     $126 = HEAP32[$109>>2]|0;
     $127 = ($125|0)==($126|0);
     if ($127) {
      $128 = HEAP32[$93>>2]|0;
      $129 = ((($128)) + 40|0);
      $130 = HEAP32[$129>>2]|0;
      (FUNCTION_TABLE_ii[$130 & 8191]($93)|0);
      $$pre$i$i$i18 = $93;$$pre$i2$i$i22 = $$pre$i2$i$i2240;$95 = $193;
      continue;
     } else {
      $131 = ((($125)) + 1|0);
      HEAP32[$107>>2] = $131;
      $$pre$i$i$i18 = $93;$$pre$i2$i$i22 = $$pre$i2$i$i2240;$95 = $193;
      continue;
     }
    }
   }
   else if ((label|0) == 55) {
    label = 0;
    $154 = HEAP32[$$lcssa127>>2]|0;
    $155 = HEAP32[$$lcssa128>>2]|0;
    $156 = ($154|0)==($155|0);
    if ($156) {
     $157 = HEAP32[$$lcssa118>>2]|0;
     $158 = ((($157)) + 40|0);
     $159 = HEAP32[$158>>2]|0;
     (FUNCTION_TABLE_ii[$159 & 8191]($$lcssa118)|0);
    } else {
     $160 = ((($154)) + 1|0);
     HEAP32[$$lcssa127>>2] = $160;
    }
    $161 = ((($$0$ph)) + 1|0);
    $$0$ph$be = $161;
   }
  } while(0);
  $$pre$pre = HEAP32[$__err>>2]|0;
  $$0$ph = $$0$ph$be;$$pre = $$pre$pre;
 }
 if ((label|0) == 15) {
  HEAP32[$__err>>2] = 4;
  $$pre$i$i$i9 = $$lcssa115;
 }
 else if ((label|0) == 19) {
  HEAP32[$__err>>2] = 4;
  $$pre$i$i$i9 = $$lcssa116$lcssa;
 }
 else if ((label|0) == 24) {
  HEAP32[$__err>>2] = 4;
  $$pre$i$i$i9 = $$lcssa116$lcssa119;
 }
 $162 = ($$pre$i$i$i9|0)==(0|0);
 if ($162) {
  $174 = 0;
 } else {
  $163 = ((($$pre$i$i$i9)) + 12|0);
  $164 = HEAP32[$163>>2]|0;
  $165 = ((($$pre$i$i$i9)) + 16|0);
  $166 = HEAP32[$165>>2]|0;
  $167 = ($164|0)==($166|0);
  if ($167) {
   $168 = HEAP32[$$pre$i$i$i9>>2]|0;
   $169 = ((($168)) + 36|0);
   $170 = HEAP32[$169>>2]|0;
   $171 = (FUNCTION_TABLE_ii[$170 & 8191]($$pre$i$i$i9)|0);
   $172 = ($171|0)==(-1);
   if ($172) {
    HEAP32[$__b>>2] = 0;
    $174 = 0;
   } else {
    $174 = $$pre$i$i$i9;
   }
  } else {
   $174 = $$pre$i$i$i9;
  }
 }
 $173 = ($174|0)==(0|0);
 $175 = HEAP32[$__e>>2]|0;
 $176 = ($175|0)==(0|0);
 do {
  if ($176) {
   label = 70;
  } else {
   $177 = ((($175)) + 12|0);
   $178 = HEAP32[$177>>2]|0;
   $179 = ((($175)) + 16|0);
   $180 = HEAP32[$179>>2]|0;
   $181 = ($178|0)==($180|0);
   if ($181) {
    $182 = HEAP32[$175>>2]|0;
    $183 = ((($182)) + 36|0);
    $184 = HEAP32[$183>>2]|0;
    $185 = (FUNCTION_TABLE_ii[$184 & 8191]($175)|0);
    $186 = ($185|0)==(-1);
    if ($186) {
     HEAP32[$__e>>2] = 0;
     label = 70;
     break;
    }
   }
   if (!($173)) {
    label = 71;
   }
  }
 } while(0);
 if ((label|0) == 70) {
  if ($173) {
   label = 71;
  }
 }
 if ((label|0) == 71) {
  $187 = HEAP32[$__err>>2]|0;
  $188 = $187 | 2;
  HEAP32[$__err>>2] = $188;
 }
 STACKTOP = sp;return ($174|0);
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11do_get_dateES4_S4_RNS_8ios_baseERjP2tm($this,$__b,$__e,$__iob,$__err,$__tm) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__tm = $__tm|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = ((($this)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 20|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_ii[$5 & 8191]($2)|0);
 $7 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $7;
 $8 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $8;
 $9 = HEAP8[$6>>0]|0;
 $10 = $9 & 1;
 $11 = ($10<<24>>24)==(0);
 $12 = ((($6)) + 8|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($6)) + 1|0);
 $15 = $11 ? $14 : $13;
 $16 = ((($6)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $9&255;
 $19 = $18 >>> 1;
 $20 = $11 ? $19 : $17;
 $21 = (($15) + ($20)|0);
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $22 = (__ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__tm,$15,$21)|0);
 STACKTOP = sp;return ($22|0);
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE14do_get_weekdayES4_S4_RNS_8ios_baseERjP2tm($this,$__b,$__e,$__iob,$__err,$__tm) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__tm = $__tm|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$0>>2] = $2;
 __THREW__ = 0;
 $3 = (invoke_iii(3374,($0|0),(1207576|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($2)|0);
  ___resumeException($9|0);
  // unreachable;
 } else {
  (__ZNSt3__114__shared_count16__release_sharedEv($2)|0);
  $6 = ((($__tm)) + 24|0);
  $7 = HEAP32[$__e>>2]|0;
  HEAP32[$1>>2] = $7;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
  __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__get_weekdaynameERiRS4_S4_RjRKNS_5ctypeIcEE($this,$6,$__b,$$byval_copy,$__err,$3);
  $8 = HEAP32[$__b>>2]|0;
  STACKTOP = sp;return ($8|0);
 }
 return (0)|0;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__get_weekdaynameERiRS4_S4_RjRKNS_5ctypeIcEE($this,$__w,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__w = $__w|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = ((($this)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($1)|0);
 $5 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $5;
 $6 = ((($4)) + 168|0);
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $7 = (__ZNSt3__114__scan_keywordINS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEPKNS_12basic_stringIcS3_NS_9allocatorIcEEEENS_5ctypeIcEEEET0_RT_SE_SD_SD_RKT1_Rjb($__b,$$byval_copy,$4,$6,$__ct,$__err,0)|0);
 $8 = $7;
 $9 = $4;
 $10 = (($8) - ($9))|0;
 $11 = ($10|0)<(168);
 if ($11) {
  $12 = (($10|0) / 12)&-1;
  $13 = (($12|0) % 7)&-1;
  HEAP32[$__w>>2] = $13;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE16do_get_monthnameES4_S4_RNS_8ios_baseERjP2tm($this,$__b,$__e,$__iob,$__err,$__tm) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__tm = $__tm|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$0>>2] = $2;
 __THREW__ = 0;
 $3 = (invoke_iii(3374,($0|0),(1207576|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($2)|0);
  ___resumeException($9|0);
  // unreachable;
 } else {
  (__ZNSt3__114__shared_count16__release_sharedEv($2)|0);
  $6 = ((($__tm)) + 16|0);
  $7 = HEAP32[$__e>>2]|0;
  HEAP32[$1>>2] = $7;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
  __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE15__get_monthnameERiRS4_S4_RjRKNS_5ctypeIcEE($this,$6,$__b,$$byval_copy,$__err,$3);
  $8 = HEAP32[$__b>>2]|0;
  STACKTOP = sp;return ($8|0);
 }
 return (0)|0;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE15__get_monthnameERiRS4_S4_RjRKNS_5ctypeIcEE($this,$__m,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__m = $__m|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = ((($this)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($1)|0);
 $6 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $6;
 $7 = ((($5)) + 288|0);
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $8 = (__ZNSt3__114__scan_keywordINS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEPKNS_12basic_stringIcS3_NS_9allocatorIcEEEENS_5ctypeIcEEEET0_RT_SE_SD_SD_RKT1_Rjb($__b,$$byval_copy,$5,$7,$__ct,$__err,0)|0);
 $9 = $8;
 $10 = $5;
 $11 = (($9) - ($10))|0;
 $12 = ($11|0)<(288);
 if ($12) {
  $13 = (($11|0) / 12)&-1;
  $14 = (($13|0) % 12)&-1;
  HEAP32[$__m>>2] = $14;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11do_get_yearES4_S4_RNS_8ios_baseERjP2tm($this,$__b,$__e,$__iob,$__err,$__tm) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__tm = $__tm|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$0>>2] = $2;
 __THREW__ = 0;
 $3 = (invoke_iii(3374,($0|0),(1207576|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($2)|0);
  ___resumeException($9|0);
  // unreachable;
 } else {
  (__ZNSt3__114__shared_count16__release_sharedEv($2)|0);
  $6 = ((($__tm)) + 20|0);
  $7 = HEAP32[$__e>>2]|0;
  HEAP32[$1>>2] = $7;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
  __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE10__get_yearERiRS4_S4_RjRKNS_5ctypeIcEE($this,$6,$__b,$$byval_copy,$__err,$3);
  $8 = HEAP32[$__b>>2]|0;
  STACKTOP = sp;return ($8|0);
 }
 return (0)|0;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE10__get_yearERiRS4_S4_RjRKNS_5ctypeIcEE($this,$__y,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__y = $__y|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$ = 0, $$byval_copy = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__t$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,4)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 if ($5) {
  $6 = ($2|0)<(69);
  if ($6) {
   $7 = (($2) + 2000)|0;
   $__t$0 = $7;
  } else {
   $$off = (($2) + -69)|0;
   $8 = ($$off>>>0)<(31);
   $9 = (($2) + 1900)|0;
   $$ = $8 ? $9 : $2;
   $__t$0 = $$;
  }
  $10 = (($__t$0) + -1900)|0;
  HEAP32[$__y>>2] = $10;
 }
 STACKTOP = sp;return;
}
function __ZNSt3__120__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$__e,$__err,$__ct,$__n) {
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 $__n = $__n|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i10 = 0, $$0$in = 0, $$2 = 0, $$lcssa = 0, $$pre$i$i$i = 0, $$pre$i$i$i16 = 0, $$pre$i$i$i3 = 0, $$pre$i2$i$i20 = 0, $$pre$i2$i$i20$lcssa = 0, $$pre$i2$i$i7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__r$0 = 0, $__r$0$in = 0, $__r$0$lcssa = 0, $isascii1$i = 0, $isascii1$i12 = 0, $sext = 0, $sext24 = 0, $tmp$i$i9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$__b>>2]|0;
 $1 = ($0|0)==(0|0);
 do {
  if ($1) {
   $13 = 0;
  } else {
   $2 = ((($0)) + 12|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = ((($0)) + 16|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = ($3|0)==($5|0);
   if ($6) {
    $7 = HEAP32[$0>>2]|0;
    $8 = ((($7)) + 36|0);
    $9 = HEAP32[$8>>2]|0;
    $10 = (FUNCTION_TABLE_ii[$9 & 8191]($0)|0);
    $11 = ($10|0)==(-1);
    if ($11) {
     HEAP32[$__b>>2] = 0;
     $13 = 0;
     break;
    } else {
     $$pre$i$i$i = HEAP32[$__b>>2]|0;
     $13 = $$pre$i$i$i;
     break;
    }
   } else {
    $13 = $0;
   }
  }
 } while(0);
 $12 = ($13|0)==(0|0);
 $14 = HEAP32[$__e>>2]|0;
 $15 = ($14|0)==(0|0);
 do {
  if ($15) {
   label = 11;
  } else {
   $16 = ((($14)) + 12|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($14)) + 16|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ($17|0)==($19|0);
   if ($20) {
    $21 = HEAP32[$14>>2]|0;
    $22 = ((($21)) + 36|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = (FUNCTION_TABLE_ii[$23 & 8191]($14)|0);
    $25 = ($24|0)==(-1);
    if ($25) {
     HEAP32[$__e>>2] = 0;
     label = 11;
     break;
    }
   }
   if ($12) {
    $157 = $14;
    label = 13;
   } else {
    label = 12;
   }
  }
 } while(0);
 if ((label|0) == 11) {
  if ($12) {
   label = 12;
  } else {
   $157 = 0;
   label = 13;
  }
 }
 L16: do {
  if ((label|0) == 12) {
   $26 = HEAP32[$__err>>2]|0;
   $27 = $26 | 6;
   HEAP32[$__err>>2] = $27;
   $$2 = 0;
  }
  else if ((label|0) == 13) {
   $28 = HEAP32[$__b>>2]|0;
   $29 = ((($28)) + 12|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = ((($28)) + 16|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = ($30|0)==($32|0);
   if ($33) {
    $34 = HEAP32[$28>>2]|0;
    $35 = ((($34)) + 36|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = (FUNCTION_TABLE_ii[$36 & 8191]($28)|0);
    $$0$i$i = $37;
   } else {
    $38 = HEAP8[$30>>0]|0;
    $39 = $38&255;
    $$0$i$i = $39;
   }
   $40 = $$0$i$i&255;
   $isascii1$i = ($40<<24>>24)>(-1);
   if ($isascii1$i) {
    $sext = $$0$i$i << 24;
    $41 = $sext >> 24;
    $42 = ((($__ct)) + 8|0);
    $43 = HEAP32[$42>>2]|0;
    $44 = (($43) + ($41<<1)|0);
    $45 = HEAP16[$44>>1]|0;
    $46 = $45 & 2048;
    $47 = ($46<<16>>16)==(0);
    if (!($47)) {
     $50 = HEAP32[$__ct>>2]|0;
     $51 = ((($50)) + 36|0);
     $52 = HEAP32[$51>>2]|0;
     $53 = (FUNCTION_TABLE_iiii[$52 & 8191]($__ct,$40,0)|0);
     $54 = $53 << 24 >> 24;
     $55 = HEAP32[$__b>>2]|0;
     $56 = ((($55)) + 12|0);
     $57 = HEAP32[$56>>2]|0;
     $58 = ((($55)) + 16|0);
     $59 = HEAP32[$58>>2]|0;
     $60 = ($57|0)==($59|0);
     if ($60) {
      $61 = HEAP32[$55>>2]|0;
      $62 = ((($61)) + 40|0);
      $63 = HEAP32[$62>>2]|0;
      (FUNCTION_TABLE_ii[$63 & 8191]($55)|0);
      $$0$in = $__n;$$pre$i2$i$i7 = $157;$80 = $157;$__r$0$in = $54;
     } else {
      $64 = ((($57)) + 1|0);
      HEAP32[$56>>2] = $64;
      $$0$in = $__n;$$pre$i2$i$i7 = $157;$80 = $157;$__r$0$in = $54;
     }
     while(1) {
      $__r$0 = (($__r$0$in) + -48)|0;
      $$0 = (($$0$in) + -1)|0;
      $65 = HEAP32[$__b>>2]|0;
      $66 = ($65|0)==(0|0);
      do {
       if ($66) {
        $78 = 0;
       } else {
        $67 = ((($65)) + 12|0);
        $68 = HEAP32[$67>>2]|0;
        $69 = ((($65)) + 16|0);
        $70 = HEAP32[$69>>2]|0;
        $71 = ($68|0)==($70|0);
        if ($71) {
         $72 = HEAP32[$65>>2]|0;
         $73 = ((($72)) + 36|0);
         $74 = HEAP32[$73>>2]|0;
         $75 = (FUNCTION_TABLE_ii[$74 & 8191]($65)|0);
         $76 = ($75|0)==(-1);
         if ($76) {
          HEAP32[$__b>>2] = 0;
          $78 = 0;
          break;
         } else {
          $$pre$i$i$i3 = HEAP32[$__b>>2]|0;
          $78 = $$pre$i$i$i3;
          break;
         }
        } else {
         $78 = $65;
        }
       }
      } while(0);
      $77 = ($78|0)==(0|0);
      $79 = ($80|0)==(0|0);
      if ($79) {
       $$pre$i2$i$i20 = $$pre$i2$i$i7;$92 = 0;
      } else {
       $81 = ((($80)) + 12|0);
       $82 = HEAP32[$81>>2]|0;
       $83 = ((($80)) + 16|0);
       $84 = HEAP32[$83>>2]|0;
       $85 = ($82|0)==($84|0);
       if ($85) {
        $86 = HEAP32[$80>>2]|0;
        $87 = ((($86)) + 36|0);
        $88 = HEAP32[$87>>2]|0;
        $89 = (FUNCTION_TABLE_ii[$88 & 8191]($80)|0);
        $90 = ($89|0)==(-1);
        if ($90) {
         HEAP32[$__e>>2] = 0;
         $$pre$i2$i$i20 = 0;$92 = 0;
        } else {
         $$pre$i2$i$i20 = $$pre$i2$i$i7;$92 = $$pre$i2$i$i7;
        }
       } else {
        $$pre$i2$i$i20 = $$pre$i2$i$i7;$92 = $80;
       }
      }
      $91 = ($92|0)==(0|0);
      $tmp$i$i9 = $77 ^ $91;
      $93 = ($$0$in|0)>(1);
      $94 = $93 & $tmp$i$i9;
      $95 = HEAP32[$__b>>2]|0;
      if (!($94)) {
       $$lcssa = $95;$$pre$i2$i$i20$lcssa = $$pre$i2$i$i20;$__r$0$lcssa = $__r$0;
       break;
      }
      $96 = ((($95)) + 12|0);
      $97 = HEAP32[$96>>2]|0;
      $98 = ((($95)) + 16|0);
      $99 = HEAP32[$98>>2]|0;
      $100 = ($97|0)==($99|0);
      if ($100) {
       $101 = HEAP32[$95>>2]|0;
       $102 = ((($101)) + 36|0);
       $103 = HEAP32[$102>>2]|0;
       $104 = (FUNCTION_TABLE_ii[$103 & 8191]($95)|0);
       $$0$i$i10 = $104;
      } else {
       $105 = HEAP8[$97>>0]|0;
       $106 = $105&255;
       $$0$i$i10 = $106;
      }
      $107 = $$0$i$i10&255;
      $isascii1$i12 = ($107<<24>>24)>(-1);
      if (!($isascii1$i12)) {
       $$2 = $__r$0;
       break L16;
      }
      $sext24 = $$0$i$i10 << 24;
      $108 = $sext24 >> 24;
      $109 = HEAP32[$42>>2]|0;
      $110 = (($109) + ($108<<1)|0);
      $111 = HEAP16[$110>>1]|0;
      $112 = $111 & 2048;
      $113 = ($112<<16>>16)==(0);
      if ($113) {
       $$2 = $__r$0;
       break L16;
      }
      $114 = ($__r$0*10)|0;
      $115 = HEAP32[$__ct>>2]|0;
      $116 = ((($115)) + 36|0);
      $117 = HEAP32[$116>>2]|0;
      $118 = (FUNCTION_TABLE_iiii[$117 & 8191]($__ct,$107,0)|0);
      $119 = $118 << 24 >> 24;
      $120 = (($119) + ($114))|0;
      $121 = HEAP32[$__b>>2]|0;
      $122 = ((($121)) + 12|0);
      $123 = HEAP32[$122>>2]|0;
      $124 = ((($121)) + 16|0);
      $125 = HEAP32[$124>>2]|0;
      $126 = ($123|0)==($125|0);
      if ($126) {
       $127 = HEAP32[$121>>2]|0;
       $128 = ((($127)) + 40|0);
       $129 = HEAP32[$128>>2]|0;
       (FUNCTION_TABLE_ii[$129 & 8191]($121)|0);
       $$0$in = $$0;$$pre$i2$i$i7 = $$pre$i2$i$i20;$80 = $92;$__r$0$in = $120;
       continue;
      } else {
       $130 = ((($123)) + 1|0);
       HEAP32[$122>>2] = $130;
       $$0$in = $$0;$$pre$i2$i$i7 = $$pre$i2$i$i20;$80 = $92;$__r$0$in = $120;
       continue;
      }
     }
     $131 = ($$lcssa|0)==(0|0);
     do {
      if ($131) {
       $143 = 0;
      } else {
       $132 = ((($$lcssa)) + 12|0);
       $133 = HEAP32[$132>>2]|0;
       $134 = ((($$lcssa)) + 16|0);
       $135 = HEAP32[$134>>2]|0;
       $136 = ($133|0)==($135|0);
       if ($136) {
        $137 = HEAP32[$$lcssa>>2]|0;
        $138 = ((($137)) + 36|0);
        $139 = HEAP32[$138>>2]|0;
        $140 = (FUNCTION_TABLE_ii[$139 & 8191]($$lcssa)|0);
        $141 = ($140|0)==(-1);
        if ($141) {
         HEAP32[$__b>>2] = 0;
         $143 = 0;
         break;
        } else {
         $$pre$i$i$i16 = HEAP32[$__b>>2]|0;
         $143 = $$pre$i$i$i16;
         break;
        }
       } else {
        $143 = $$lcssa;
       }
      }
     } while(0);
     $142 = ($143|0)==(0|0);
     $144 = ($$pre$i2$i$i20$lcssa|0)==(0|0);
     do {
      if ($144) {
       label = 50;
      } else {
       $145 = ((($$pre$i2$i$i20$lcssa)) + 12|0);
       $146 = HEAP32[$145>>2]|0;
       $147 = ((($$pre$i2$i$i20$lcssa)) + 16|0);
       $148 = HEAP32[$147>>2]|0;
       $149 = ($146|0)==($148|0);
       if ($149) {
        $150 = HEAP32[$$pre$i2$i$i20$lcssa>>2]|0;
        $151 = ((($150)) + 36|0);
        $152 = HEAP32[$151>>2]|0;
        $153 = (FUNCTION_TABLE_ii[$152 & 8191]($$pre$i2$i$i20$lcssa)|0);
        $154 = ($153|0)==(-1);
        if ($154) {
         HEAP32[$__e>>2] = 0;
         label = 50;
         break;
        }
       }
       if ($142) {
        $$2 = $__r$0$lcssa;
        break L16;
       }
      }
     } while(0);
     if ((label|0) == 50) {
      if (!($142)) {
       $$2 = $__r$0$lcssa;
       break;
      }
     }
     $155 = HEAP32[$__err>>2]|0;
     $156 = $155 | 2;
     HEAP32[$__err>>2] = $156;
     $$2 = $__r$0$lcssa;
     break;
    }
   }
   $48 = HEAP32[$__err>>2]|0;
   $49 = $48 | 4;
   HEAP32[$__err>>2] = $49;
   $$2 = 0;
  }
 } while(0);
 return ($$2|0);
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjP2tmcc($this,$__b,$__e,$__iob,$__err,$__tm,$__fmt,$0) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__tm = $__tm|0;
 $__fmt = $__fmt|0;
 $0 = $0|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $$byval_copy10 = 0, $$byval_copy11 = 0, $$byval_copy12 = 0, $$byval_copy13 = 0, $$byval_copy14 = 0, $$byval_copy15 = 0, $$byval_copy16 = 0, $$byval_copy17 = 0, $$byval_copy18 = 0, $$byval_copy19 = 0, $$byval_copy2 = 0, $$byval_copy20 = 0, $$byval_copy21 = 0, $$byval_copy22 = 0, $$byval_copy23 = 0, $$byval_copy24 = 0, $$byval_copy25 = 0, $$byval_copy26 = 0;
 var $$byval_copy27 = 0, $$byval_copy28 = 0, $$byval_copy29 = 0, $$byval_copy3 = 0, $$byval_copy30 = 0, $$byval_copy4 = 0, $$byval_copy5 = 0, $$byval_copy6 = 0, $$byval_copy7 = 0, $$byval_copy8 = 0, $$byval_copy9 = 0, $$sroa$07$0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy30 = sp + 248|0;
 $$byval_copy29 = sp + 244|0;
 $$byval_copy28 = sp + 240|0;
 $$byval_copy27 = sp + 236|0;
 $$byval_copy26 = sp + 232|0;
 $$byval_copy25 = sp + 228|0;
 $$byval_copy24 = sp + 224|0;
 $$byval_copy23 = sp + 220|0;
 $$byval_copy22 = sp + 216|0;
 $$byval_copy21 = sp + 212|0;
 $$byval_copy20 = sp + 208|0;
 $$byval_copy19 = sp + 204|0;
 $$byval_copy18 = sp + 200|0;
 $$byval_copy17 = sp + 196|0;
 $$byval_copy16 = sp + 192|0;
 $$byval_copy15 = sp + 188|0;
 $$byval_copy14 = sp + 184|0;
 $$byval_copy13 = sp + 180|0;
 $$byval_copy12 = sp + 176|0;
 $$byval_copy11 = sp + 172|0;
 $$byval_copy10 = sp + 168|0;
 $$byval_copy9 = sp + 164|0;
 $$byval_copy8 = sp + 160|0;
 $$byval_copy7 = sp + 156|0;
 $$byval_copy6 = sp + 152|0;
 $$byval_copy5 = sp + 148|0;
 $$byval_copy4 = sp + 144|0;
 $$byval_copy3 = sp + 140|0;
 $$byval_copy2 = sp + 136|0;
 $$byval_copy1 = sp + 132|0;
 $$byval_copy = sp + 128|0;
 $1 = sp + 124|0;
 $2 = sp + 120|0;
 $3 = sp + 116|0;
 $4 = sp + 112|0;
 $5 = sp + 108|0;
 $6 = sp + 104|0;
 $7 = sp + 100|0;
 $8 = sp + 96|0;
 $9 = sp + 92|0;
 $10 = sp + 88|0;
 $11 = sp + 84|0;
 $12 = sp + 80|0;
 $13 = sp + 76|0;
 $14 = sp + 72|0;
 $15 = sp + 68|0;
 $16 = sp + 64|0;
 $17 = sp + 60|0;
 $18 = sp + 56|0;
 $19 = sp + 52|0;
 $20 = sp + 48|0;
 $21 = sp + 44|0;
 $22 = sp + 40|0;
 $23 = sp + 36|0;
 $24 = sp + 32|0;
 $25 = sp + 28|0;
 $26 = sp + 24|0;
 $27 = sp + 20|0;
 $28 = sp + 16|0;
 $29 = sp + 12|0;
 $30 = sp + 8|0;
 $31 = sp + 4|0;
 $32 = sp;
 HEAP32[$__err>>2] = 0;
 $33 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$1>>2] = $33;
 __THREW__ = 0;
 $34 = (invoke_iii(3374,($1|0),(1207576|0))|0);
 $35 = __THREW__; __THREW__ = 0;
 $36 = $35&1;
 if ($36) {
  $38 = ___cxa_find_matching_catch_2()|0;
  $39 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($33)|0);
  ___resumeException($38|0);
  // unreachable;
 }
 (__ZNSt3__114__shared_count16__release_sharedEv($33)|0);
 $37 = $__fmt << 24 >> 24;
 do {
  switch ($37|0) {
  case 65: case 97:  {
   $40 = ((($__tm)) + 24|0);
   $41 = HEAP32[$__e>>2]|0;
   HEAP32[$2>>2] = $41;
   ;HEAP32[$$byval_copy>>2]=HEAP32[$2>>2]|0;
   __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__get_weekdaynameERiRS4_S4_RjRKNS_5ctypeIcEE($this,$40,$__b,$$byval_copy,$__err,$34);
   label = 28;
   break;
  }
  case 104: case 66: case 98:  {
   $42 = ((($__tm)) + 16|0);
   $43 = HEAP32[$__e>>2]|0;
   HEAP32[$3>>2] = $43;
   ;HEAP32[$$byval_copy1>>2]=HEAP32[$3>>2]|0;
   __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE15__get_monthnameERiRS4_S4_RjRKNS_5ctypeIcEE($this,$42,$__b,$$byval_copy1,$__err,$34);
   label = 28;
   break;
  }
  case 99:  {
   $44 = ((($this)) + 8|0);
   $45 = HEAP32[$44>>2]|0;
   $46 = ((($45)) + 12|0);
   $47 = HEAP32[$46>>2]|0;
   $48 = (FUNCTION_TABLE_ii[$47 & 8191]($44)|0);
   $49 = HEAP32[$__b>>2]|0;
   HEAP32[$4>>2] = $49;
   $50 = HEAP32[$__e>>2]|0;
   HEAP32[$5>>2] = $50;
   $51 = HEAP8[$48>>0]|0;
   $52 = $51 & 1;
   $53 = ($52<<24>>24)==(0);
   $54 = ((($48)) + 8|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = ((($48)) + 1|0);
   $57 = $53 ? $56 : $55;
   $58 = ((($48)) + 4|0);
   $59 = HEAP32[$58>>2]|0;
   $60 = $51&255;
   $61 = $60 >>> 1;
   $62 = $53 ? $61 : $59;
   $63 = (($57) + ($62)|0);
   ;HEAP32[$$byval_copy2>>2]=HEAP32[$4>>2]|0;
   ;HEAP32[$$byval_copy3>>2]=HEAP32[$5>>2]|0;
   $64 = (__ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($this,$$byval_copy2,$$byval_copy3,$__iob,$__err,$__tm,$57,$63)|0);
   HEAP32[$__b>>2] = $64;
   label = 28;
   break;
  }
  case 101: case 100:  {
   $65 = ((($__tm)) + 12|0);
   $66 = HEAP32[$__e>>2]|0;
   HEAP32[$6>>2] = $66;
   ;HEAP32[$$byval_copy4>>2]=HEAP32[$6>>2]|0;
   __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE9__get_dayERiRS4_S4_RjRKNS_5ctypeIcEE($this,$65,$__b,$$byval_copy4,$__err,$34);
   label = 28;
   break;
  }
  case 68:  {
   $67 = HEAP32[$__b>>2]|0;
   HEAP32[$7>>2] = $67;
   $68 = HEAP32[$__e>>2]|0;
   HEAP32[$8>>2] = $68;
   ;HEAP32[$$byval_copy5>>2]=HEAP32[$7>>2]|0;
   ;HEAP32[$$byval_copy6>>2]=HEAP32[$8>>2]|0;
   $69 = (__ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($this,$$byval_copy5,$$byval_copy6,$__iob,$__err,$__tm,620968,(620976))|0);
   HEAP32[$__b>>2] = $69;
   label = 28;
   break;
  }
  case 70:  {
   $70 = HEAP32[$__b>>2]|0;
   HEAP32[$9>>2] = $70;
   $71 = HEAP32[$__e>>2]|0;
   HEAP32[$10>>2] = $71;
   ;HEAP32[$$byval_copy7>>2]=HEAP32[$9>>2]|0;
   ;HEAP32[$$byval_copy8>>2]=HEAP32[$10>>2]|0;
   $72 = (__ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($this,$$byval_copy7,$$byval_copy8,$__iob,$__err,$__tm,620976,(620984))|0);
   HEAP32[$__b>>2] = $72;
   label = 28;
   break;
  }
  case 72:  {
   $73 = ((($__tm)) + 8|0);
   $74 = HEAP32[$__e>>2]|0;
   HEAP32[$11>>2] = $74;
   ;HEAP32[$$byval_copy9>>2]=HEAP32[$11>>2]|0;
   __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE10__get_hourERiRS4_S4_RjRKNS_5ctypeIcEE($this,$73,$__b,$$byval_copy9,$__err,$34);
   label = 28;
   break;
  }
  case 73:  {
   $75 = ((($__tm)) + 8|0);
   $76 = HEAP32[$__e>>2]|0;
   HEAP32[$12>>2] = $76;
   ;HEAP32[$$byval_copy10>>2]=HEAP32[$12>>2]|0;
   __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE13__get_12_hourERiRS4_S4_RjRKNS_5ctypeIcEE($this,$75,$__b,$$byval_copy10,$__err,$34);
   label = 28;
   break;
  }
  case 106:  {
   $77 = ((($__tm)) + 28|0);
   $78 = HEAP32[$__e>>2]|0;
   HEAP32[$13>>2] = $78;
   ;HEAP32[$$byval_copy11>>2]=HEAP32[$13>>2]|0;
   __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE18__get_day_year_numERiRS4_S4_RjRKNS_5ctypeIcEE($this,$77,$__b,$$byval_copy11,$__err,$34);
   label = 28;
   break;
  }
  case 109:  {
   $79 = ((($__tm)) + 16|0);
   $80 = HEAP32[$__e>>2]|0;
   HEAP32[$14>>2] = $80;
   ;HEAP32[$$byval_copy12>>2]=HEAP32[$14>>2]|0;
   __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11__get_monthERiRS4_S4_RjRKNS_5ctypeIcEE($this,$79,$__b,$$byval_copy12,$__err,$34);
   label = 28;
   break;
  }
  case 77:  {
   $81 = ((($__tm)) + 4|0);
   $82 = HEAP32[$__e>>2]|0;
   HEAP32[$15>>2] = $82;
   ;HEAP32[$$byval_copy13>>2]=HEAP32[$15>>2]|0;
   __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE12__get_minuteERiRS4_S4_RjRKNS_5ctypeIcEE($this,$81,$__b,$$byval_copy13,$__err,$34);
   label = 28;
   break;
  }
  case 116: case 110:  {
   $83 = HEAP32[$__e>>2]|0;
   HEAP32[$16>>2] = $83;
   ;HEAP32[$$byval_copy14>>2]=HEAP32[$16>>2]|0;
   __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__get_white_spaceERS4_S4_RjRKNS_5ctypeIcEE($this,$__b,$$byval_copy14,$__err,$34);
   label = 28;
   break;
  }
  case 112:  {
   $84 = ((($__tm)) + 8|0);
   $85 = HEAP32[$__e>>2]|0;
   HEAP32[$17>>2] = $85;
   ;HEAP32[$$byval_copy15>>2]=HEAP32[$17>>2]|0;
   __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11__get_am_pmERiRS4_S4_RjRKNS_5ctypeIcEE($this,$84,$__b,$$byval_copy15,$__err,$34);
   label = 28;
   break;
  }
  case 114:  {
   $86 = HEAP32[$__b>>2]|0;
   HEAP32[$18>>2] = $86;
   $87 = HEAP32[$__e>>2]|0;
   HEAP32[$19>>2] = $87;
   ;HEAP32[$$byval_copy16>>2]=HEAP32[$18>>2]|0;
   ;HEAP32[$$byval_copy17>>2]=HEAP32[$19>>2]|0;
   $88 = (__ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($this,$$byval_copy16,$$byval_copy17,$__iob,$__err,$__tm,620984,(620995))|0);
   HEAP32[$__b>>2] = $88;
   label = 28;
   break;
  }
  case 82:  {
   $89 = HEAP32[$__b>>2]|0;
   HEAP32[$20>>2] = $89;
   $90 = HEAP32[$__e>>2]|0;
   HEAP32[$21>>2] = $90;
   ;HEAP32[$$byval_copy18>>2]=HEAP32[$20>>2]|0;
   ;HEAP32[$$byval_copy19>>2]=HEAP32[$21>>2]|0;
   $91 = (__ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($this,$$byval_copy18,$$byval_copy19,$__iob,$__err,$__tm,620995,(621000))|0);
   HEAP32[$__b>>2] = $91;
   label = 28;
   break;
  }
  case 83:  {
   $92 = HEAP32[$__e>>2]|0;
   HEAP32[$22>>2] = $92;
   ;HEAP32[$$byval_copy20>>2]=HEAP32[$22>>2]|0;
   __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE12__get_secondERiRS4_S4_RjRKNS_5ctypeIcEE($this,$__tm,$__b,$$byval_copy20,$__err,$34);
   label = 28;
   break;
  }
  case 84:  {
   $93 = HEAP32[$__b>>2]|0;
   HEAP32[$23>>2] = $93;
   $94 = HEAP32[$__e>>2]|0;
   HEAP32[$24>>2] = $94;
   ;HEAP32[$$byval_copy21>>2]=HEAP32[$23>>2]|0;
   ;HEAP32[$$byval_copy22>>2]=HEAP32[$24>>2]|0;
   $95 = (__ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($this,$$byval_copy21,$$byval_copy22,$__iob,$__err,$__tm,621000,(621008))|0);
   HEAP32[$__b>>2] = $95;
   label = 28;
   break;
  }
  case 119:  {
   $96 = ((($__tm)) + 24|0);
   $97 = HEAP32[$__e>>2]|0;
   HEAP32[$25>>2] = $97;
   ;HEAP32[$$byval_copy23>>2]=HEAP32[$25>>2]|0;
   __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE13__get_weekdayERiRS4_S4_RjRKNS_5ctypeIcEE($this,$96,$__b,$$byval_copy23,$__err,$34);
   label = 28;
   break;
  }
  case 120:  {
   $98 = HEAP32[$this>>2]|0;
   $99 = ((($98)) + 20|0);
   $100 = HEAP32[$99>>2]|0;
   $101 = HEAP32[$__b>>2]|0;
   HEAP32[$26>>2] = $101;
   $102 = HEAP32[$__e>>2]|0;
   HEAP32[$27>>2] = $102;
   ;HEAP32[$$byval_copy24>>2]=HEAP32[$26>>2]|0;
   ;HEAP32[$$byval_copy25>>2]=HEAP32[$27>>2]|0;
   $103 = (FUNCTION_TABLE_iiiiiii[$100 & 8191]($this,$$byval_copy24,$$byval_copy25,$__iob,$__err,$__tm)|0);
   $$sroa$07$0 = $103;
   break;
  }
  case 88:  {
   $104 = ((($this)) + 8|0);
   $105 = HEAP32[$104>>2]|0;
   $106 = ((($105)) + 24|0);
   $107 = HEAP32[$106>>2]|0;
   $108 = (FUNCTION_TABLE_ii[$107 & 8191]($104)|0);
   $109 = HEAP32[$__b>>2]|0;
   HEAP32[$28>>2] = $109;
   $110 = HEAP32[$__e>>2]|0;
   HEAP32[$29>>2] = $110;
   $111 = HEAP8[$108>>0]|0;
   $112 = $111 & 1;
   $113 = ($112<<24>>24)==(0);
   $114 = ((($108)) + 8|0);
   $115 = HEAP32[$114>>2]|0;
   $116 = ((($108)) + 1|0);
   $117 = $113 ? $116 : $115;
   $118 = ((($108)) + 4|0);
   $119 = HEAP32[$118>>2]|0;
   $120 = $111&255;
   $121 = $120 >>> 1;
   $122 = $113 ? $121 : $119;
   $123 = (($117) + ($122)|0);
   ;HEAP32[$$byval_copy26>>2]=HEAP32[$28>>2]|0;
   ;HEAP32[$$byval_copy27>>2]=HEAP32[$29>>2]|0;
   $124 = (__ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($this,$$byval_copy26,$$byval_copy27,$__iob,$__err,$__tm,$117,$123)|0);
   HEAP32[$__b>>2] = $124;
   label = 28;
   break;
  }
  case 121:  {
   $125 = ((($__tm)) + 20|0);
   $126 = HEAP32[$__e>>2]|0;
   HEAP32[$30>>2] = $126;
   ;HEAP32[$$byval_copy28>>2]=HEAP32[$30>>2]|0;
   __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE10__get_yearERiRS4_S4_RjRKNS_5ctypeIcEE($this,$125,$__b,$$byval_copy28,$__err,$34);
   label = 28;
   break;
  }
  case 89:  {
   $127 = ((($__tm)) + 20|0);
   $128 = HEAP32[$__e>>2]|0;
   HEAP32[$31>>2] = $128;
   ;HEAP32[$$byval_copy29>>2]=HEAP32[$31>>2]|0;
   __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11__get_year4ERiRS4_S4_RjRKNS_5ctypeIcEE($this,$127,$__b,$$byval_copy29,$__err,$34);
   label = 28;
   break;
  }
  case 37:  {
   $129 = HEAP32[$__e>>2]|0;
   HEAP32[$32>>2] = $129;
   ;HEAP32[$$byval_copy30>>2]=HEAP32[$32>>2]|0;
   __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE13__get_percentERS4_S4_RjRKNS_5ctypeIcEE($this,$__b,$$byval_copy30,$__err,$34);
   label = 28;
   break;
  }
  default: {
   $130 = HEAP32[$__err>>2]|0;
   $131 = $130 | 4;
   HEAP32[$__err>>2] = $131;
   label = 28;
  }
  }
 } while(0);
 if ((label|0) == 28) {
  $132 = HEAP32[$__b>>2]|0;
  $$sroa$07$0 = $132;
 }
 STACKTOP = sp;return ($$sroa$07$0|0);
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE9__get_dayERiRS4_S4_RjRKNS_5ctypeIcEE($this,$__d,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__d = $__d|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $$off = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,2)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 $$off = (($2) + -1)|0;
 $6 = ($$off>>>0)<(31);
 $7 = $6 & $5;
 if ($7) {
  HEAP32[$__d>>2] = $2;
 } else {
  $8 = $3 | 4;
  HEAP32[$__err>>2] = $8;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE10__get_hourERiRS4_S4_RjRKNS_5ctypeIcEE($this,$__h,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__h = $__h|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,2)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 $6 = ($2|0)<(24);
 $or$cond = $6 & $5;
 if ($or$cond) {
  HEAP32[$__h>>2] = $2;
 } else {
  $7 = $3 | 4;
  HEAP32[$__err>>2] = $7;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE13__get_12_hourERiRS4_S4_RjRKNS_5ctypeIcEE($this,$__h,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__h = $__h|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $$off = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,2)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 $$off = (($2) + -1)|0;
 $6 = ($$off>>>0)<(12);
 $7 = $6 & $5;
 if ($7) {
  HEAP32[$__h>>2] = $2;
 } else {
  $8 = $3 | 4;
  HEAP32[$__err>>2] = $8;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE18__get_day_year_numERiRS4_S4_RjRKNS_5ctypeIcEE($this,$__d,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__d = $__d|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,3)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 $6 = ($2|0)<(366);
 $or$cond = $6 & $5;
 if ($or$cond) {
  HEAP32[$__d>>2] = $2;
 } else {
  $7 = $3 | 4;
  HEAP32[$__err>>2] = $7;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11__get_monthERiRS4_S4_RjRKNS_5ctypeIcEE($this,$__m,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__m = $__m|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,2)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 $6 = ($2|0)<(13);
 $or$cond = $6 & $5;
 if ($or$cond) {
  $7 = (($2) + -1)|0;
  HEAP32[$__m>>2] = $7;
 } else {
  $8 = $3 | 4;
  HEAP32[$__err>>2] = $8;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE12__get_minuteERiRS4_S4_RjRKNS_5ctypeIcEE($this,$__m,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__m = $__m|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,2)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 $6 = ($2|0)<(60);
 $or$cond = $6 & $5;
 if ($or$cond) {
  HEAP32[$__m>>2] = $2;
 } else {
  $7 = $3 | 4;
  HEAP32[$__err>>2] = $7;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__get_white_spaceERS4_S4_RjRKNS_5ctypeIcEE($this,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$0$i$i = 0, $$pre$i$i$i = 0, $$pre$i$i$i2 = 0, $$pre$i2$i$i6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $9 = 0, $isascii1$i = 0, $sext = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($__ct)) + 8|0);
 L1: while(1) {
  $1 = HEAP32[$__b>>2]|0;
  $2 = ($1|0)==(0|0);
  do {
   if ($2) {
    $14 = 0;
   } else {
    $3 = ((($1)) + 12|0);
    $4 = HEAP32[$3>>2]|0;
    $5 = ((($1)) + 16|0);
    $6 = HEAP32[$5>>2]|0;
    $7 = ($4|0)==($6|0);
    if ($7) {
     $8 = HEAP32[$1>>2]|0;
     $9 = ((($8)) + 36|0);
     $10 = HEAP32[$9>>2]|0;
     $11 = (FUNCTION_TABLE_ii[$10 & 8191]($1)|0);
     $12 = ($11|0)==(-1);
     if ($12) {
      HEAP32[$__b>>2] = 0;
      $14 = 0;
      break;
     } else {
      $$pre$i$i$i = HEAP32[$__b>>2]|0;
      $14 = $$pre$i$i$i;
      break;
     }
    } else {
     $14 = $1;
    }
   }
  } while(0);
  $13 = ($14|0)==(0|0);
  $15 = HEAP32[$__e>>2]|0;
  $16 = ($15|0)==(0|0);
  do {
   if ($16) {
    label = 12;
   } else {
    $17 = ((($15)) + 12|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = ((($15)) + 16|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = ($18|0)==($20|0);
    if (!($21)) {
     if ($13) {
      $83 = $15;
      break;
     } else {
      $$pre$i2$i$i6 = $15;
      break L1;
     }
    }
    $22 = HEAP32[$15>>2]|0;
    $23 = ((($22)) + 36|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = (FUNCTION_TABLE_ii[$24 & 8191]($15)|0);
    $26 = ($25|0)==(-1);
    if ($26) {
     HEAP32[$__e>>2] = 0;
     label = 12;
     break;
    } else {
     if ($13) {
      $83 = $15;
      break;
     } else {
      $$pre$i2$i$i6 = $15;
      break L1;
     }
    }
   }
  } while(0);
  if ((label|0) == 12) {
   label = 0;
   if ($13) {
    $$pre$i2$i$i6 = 0;
    break;
   } else {
    $83 = 0;
   }
  }
  $27 = HEAP32[$__b>>2]|0;
  $28 = ((($27)) + 12|0);
  $29 = HEAP32[$28>>2]|0;
  $30 = ((($27)) + 16|0);
  $31 = HEAP32[$30>>2]|0;
  $32 = ($29|0)==($31|0);
  if ($32) {
   $33 = HEAP32[$27>>2]|0;
   $34 = ((($33)) + 36|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = (FUNCTION_TABLE_ii[$35 & 8191]($27)|0);
   $$0$i$i = $36;
  } else {
   $37 = HEAP8[$29>>0]|0;
   $38 = $37&255;
   $$0$i$i = $38;
  }
  $39 = $$0$i$i&255;
  $isascii1$i = ($39<<24>>24)>(-1);
  if (!($isascii1$i)) {
   $$pre$i2$i$i6 = $83;
   break;
  }
  $sext = $$0$i$i << 24;
  $40 = $sext >> 24;
  $41 = HEAP32[$0>>2]|0;
  $42 = (($41) + ($40<<1)|0);
  $43 = HEAP16[$42>>1]|0;
  $44 = $43 & 8192;
  $45 = ($44<<16>>16)==(0);
  if ($45) {
   $$pre$i2$i$i6 = $83;
   break;
  }
  $46 = HEAP32[$__b>>2]|0;
  $47 = ((($46)) + 12|0);
  $48 = HEAP32[$47>>2]|0;
  $49 = ((($46)) + 16|0);
  $50 = HEAP32[$49>>2]|0;
  $51 = ($48|0)==($50|0);
  if ($51) {
   $52 = HEAP32[$46>>2]|0;
   $53 = ((($52)) + 40|0);
   $54 = HEAP32[$53>>2]|0;
   (FUNCTION_TABLE_ii[$54 & 8191]($46)|0);
   continue;
  } else {
   $55 = ((($48)) + 1|0);
   HEAP32[$47>>2] = $55;
   continue;
  }
 }
 $56 = HEAP32[$__b>>2]|0;
 $57 = ($56|0)==(0|0);
 do {
  if ($57) {
   $69 = 0;
  } else {
   $58 = ((($56)) + 12|0);
   $59 = HEAP32[$58>>2]|0;
   $60 = ((($56)) + 16|0);
   $61 = HEAP32[$60>>2]|0;
   $62 = ($59|0)==($61|0);
   if ($62) {
    $63 = HEAP32[$56>>2]|0;
    $64 = ((($63)) + 36|0);
    $65 = HEAP32[$64>>2]|0;
    $66 = (FUNCTION_TABLE_ii[$65 & 8191]($56)|0);
    $67 = ($66|0)==(-1);
    if ($67) {
     HEAP32[$__b>>2] = 0;
     $69 = 0;
     break;
    } else {
     $$pre$i$i$i2 = HEAP32[$__b>>2]|0;
     $69 = $$pre$i$i$i2;
     break;
    }
   } else {
    $69 = $56;
   }
  }
 } while(0);
 $68 = ($69|0)==(0|0);
 $70 = ($$pre$i2$i$i6|0)==(0|0);
 do {
  if ($70) {
   label = 32;
  } else {
   $71 = ((($$pre$i2$i$i6)) + 12|0);
   $72 = HEAP32[$71>>2]|0;
   $73 = ((($$pre$i2$i$i6)) + 16|0);
   $74 = HEAP32[$73>>2]|0;
   $75 = ($72|0)==($74|0);
   if ($75) {
    $76 = HEAP32[$$pre$i2$i$i6>>2]|0;
    $77 = ((($76)) + 36|0);
    $78 = HEAP32[$77>>2]|0;
    $79 = (FUNCTION_TABLE_ii[$78 & 8191]($$pre$i2$i$i6)|0);
    $80 = ($79|0)==(-1);
    if ($80) {
     HEAP32[$__e>>2] = 0;
     label = 32;
     break;
    }
   }
   if (!($68)) {
    label = 33;
   }
  }
 } while(0);
 if ((label|0) == 32) {
  if ($68) {
   label = 33;
  }
 }
 if ((label|0) == 33) {
  $81 = HEAP32[$__err>>2]|0;
  $82 = $81 | 2;
  HEAP32[$__err>>2] = $82;
 }
 return;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11__get_am_pmERiRS4_S4_RjRKNS_5ctypeIcEE($this,$__h,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__h = $__h|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $or$cond = 0, $or$cond3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = ((($this)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($1)|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = $6 & 1;
 $8 = ($7<<24>>24)==(0);
 if ($8) {
  $11 = $6&255;
  $12 = $11 >>> 1;
  $24 = $12;
 } else {
  $9 = ((($5)) + 4|0);
  $10 = HEAP32[$9>>2]|0;
  $24 = $10;
 }
 $13 = ((($5)) + 12|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = $14 & 1;
 $16 = ($15<<24>>24)==(0);
 if ($16) {
  $19 = $14&255;
  $20 = $19 >>> 1;
  $22 = $20;
 } else {
  $17 = ((($5)) + 16|0);
  $18 = HEAP32[$17>>2]|0;
  $22 = $18;
 }
 $21 = (0 - ($22))|0;
 $23 = ($24|0)==($21|0);
 do {
  if ($23) {
   $25 = HEAP32[$__err>>2]|0;
   $26 = $25 | 4;
   HEAP32[$__err>>2] = $26;
  } else {
   $27 = HEAP32[$__e>>2]|0;
   HEAP32[$0>>2] = $27;
   $28 = ((($5)) + 24|0);
   ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
   $29 = (__ZNSt3__114__scan_keywordINS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEPKNS_12basic_stringIcS3_NS_9allocatorIcEEEENS_5ctypeIcEEEET0_RT_SE_SD_SD_RKT1_Rjb($__b,$$byval_copy,$5,$28,$__ct,$__err,0)|0);
   $30 = ($29|0)==($5|0);
   $31 = HEAP32[$__h>>2]|0;
   $32 = ($31|0)==(12);
   $or$cond = $30 & $32;
   if ($or$cond) {
    HEAP32[$__h>>2] = 0;
    break;
   }
   $33 = $29;
   $34 = $5;
   $35 = (($33) - ($34))|0;
   $36 = ($35|0)==(12);
   $37 = ($31|0)<(12);
   $or$cond3 = $37 & $36;
   if ($or$cond3) {
    $38 = (($31) + 12)|0;
    HEAP32[$__h>>2] = $38;
   }
  }
 } while(0);
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE12__get_secondERiRS4_S4_RjRKNS_5ctypeIcEE($this,$__s,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__s = $__s|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,2)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 $6 = ($2|0)<(61);
 $or$cond = $6 & $5;
 if ($or$cond) {
  HEAP32[$__s>>2] = $2;
 } else {
  $7 = $3 | 4;
  HEAP32[$__err>>2] = $7;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE13__get_weekdayERiRS4_S4_RjRKNS_5ctypeIcEE($this,$__w,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__w = $__w|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,1)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 $6 = ($2|0)<(7);
 $or$cond = $6 & $5;
 if ($or$cond) {
  HEAP32[$__w>>2] = $2;
 } else {
  $7 = $3 | 4;
  HEAP32[$__err>>2] = $7;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11__get_year4ERiRS4_S4_RjRKNS_5ctypeIcEE($this,$__y,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__y = $__y|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,4)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 if ($5) {
  $6 = (($2) + -1900)|0;
  HEAP32[$__y>>2] = $6;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE13__get_percentERS4_S4_RjRKNS_5ctypeIcEE($this,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$0$i$i = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i2 = 0, $$pre$i2$i$i611 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$__b>>2]|0;
 $1 = ($0|0)==(0|0);
 do {
  if ($1) {
   $13 = 0;
  } else {
   $2 = ((($0)) + 12|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = ((($0)) + 16|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = ($3|0)==($5|0);
   if ($6) {
    $7 = HEAP32[$0>>2]|0;
    $8 = ((($7)) + 36|0);
    $9 = HEAP32[$8>>2]|0;
    $10 = (FUNCTION_TABLE_ii[$9 & 8191]($0)|0);
    $11 = ($10|0)==(-1);
    if ($11) {
     HEAP32[$__b>>2] = 0;
     $13 = 0;
     break;
    } else {
     $$pre$i$i$i = HEAP32[$__b>>2]|0;
     $13 = $$pre$i$i$i;
     break;
    }
   } else {
    $13 = $0;
   }
  }
 } while(0);
 $12 = ($13|0)==(0|0);
 $14 = HEAP32[$__e>>2]|0;
 $15 = ($14|0)==(0|0);
 do {
  if ($15) {
   label = 11;
  } else {
   $16 = ((($14)) + 12|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($14)) + 16|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ($17|0)==($19|0);
   if ($20) {
    $21 = HEAP32[$14>>2]|0;
    $22 = ((($21)) + 36|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = (FUNCTION_TABLE_ii[$23 & 8191]($14)|0);
    $25 = ($24|0)==(-1);
    if ($25) {
     HEAP32[$__e>>2] = 0;
     label = 11;
     break;
    }
   }
   if ($12) {
    $$pre$i2$i$i611 = $14;
    label = 13;
   } else {
    label = 12;
   }
  }
 } while(0);
 if ((label|0) == 11) {
  if ($12) {
   label = 12;
  } else {
   $$pre$i2$i$i611 = 0;
   label = 13;
  }
 }
 L16: do {
  if ((label|0) == 12) {
   $26 = HEAP32[$__err>>2]|0;
   $27 = $26 | 6;
   HEAP32[$__err>>2] = $27;
  }
  else if ((label|0) == 13) {
   $28 = HEAP32[$__b>>2]|0;
   $29 = ((($28)) + 12|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = ((($28)) + 16|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = ($30|0)==($32|0);
   if ($33) {
    $34 = HEAP32[$28>>2]|0;
    $35 = ((($34)) + 36|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = (FUNCTION_TABLE_ii[$36 & 8191]($28)|0);
    $$0$i$i = $37;
   } else {
    $38 = HEAP8[$30>>0]|0;
    $39 = $38&255;
    $$0$i$i = $39;
   }
   $40 = $$0$i$i&255;
   $41 = HEAP32[$__ct>>2]|0;
   $42 = ((($41)) + 36|0);
   $43 = HEAP32[$42>>2]|0;
   $44 = (FUNCTION_TABLE_iiii[$43 & 8191]($__ct,$40,0)|0);
   $45 = ($44<<24>>24)==(37);
   if (!($45)) {
    $46 = HEAP32[$__err>>2]|0;
    $47 = $46 | 4;
    HEAP32[$__err>>2] = $47;
    break;
   }
   $48 = HEAP32[$__b>>2]|0;
   $49 = ((($48)) + 12|0);
   $50 = HEAP32[$49>>2]|0;
   $51 = ((($48)) + 16|0);
   $52 = HEAP32[$51>>2]|0;
   $53 = ($50|0)==($52|0);
   if ($53) {
    $55 = HEAP32[$48>>2]|0;
    $56 = ((($55)) + 40|0);
    $57 = HEAP32[$56>>2]|0;
    (FUNCTION_TABLE_ii[$57 & 8191]($48)|0);
    $$pre = HEAP32[$__b>>2]|0;
    $58 = ($$pre|0)==(0|0);
    if ($58) {
     $71 = 0;
    } else {
     $60 = $$pre;
     label = 21;
    }
   } else {
    $54 = ((($50)) + 1|0);
    HEAP32[$49>>2] = $54;
    $60 = $48;
    label = 21;
   }
   do {
    if ((label|0) == 21) {
     $59 = ((($60)) + 12|0);
     $61 = HEAP32[$59>>2]|0;
     $62 = ((($60)) + 16|0);
     $63 = HEAP32[$62>>2]|0;
     $64 = ($61|0)==($63|0);
     if ($64) {
      $65 = HEAP32[$60>>2]|0;
      $66 = ((($65)) + 36|0);
      $67 = HEAP32[$66>>2]|0;
      $68 = (FUNCTION_TABLE_ii[$67 & 8191]($60)|0);
      $69 = ($68|0)==(-1);
      if ($69) {
       HEAP32[$__b>>2] = 0;
       $71 = 0;
       break;
      } else {
       $$pre$i$i$i2 = HEAP32[$__b>>2]|0;
       $71 = $$pre$i$i$i2;
       break;
      }
     } else {
      $71 = $60;
     }
    }
   } while(0);
   $70 = ($71|0)==(0|0);
   $72 = ($$pre$i2$i$i611|0)==(0|0);
   do {
    if ($72) {
     label = 30;
    } else {
     $73 = ((($$pre$i2$i$i611)) + 12|0);
     $74 = HEAP32[$73>>2]|0;
     $75 = ((($$pre$i2$i$i611)) + 16|0);
     $76 = HEAP32[$75>>2]|0;
     $77 = ($74|0)==($76|0);
     if ($77) {
      $78 = HEAP32[$$pre$i2$i$i611>>2]|0;
      $79 = ((($78)) + 36|0);
      $80 = HEAP32[$79>>2]|0;
      $81 = (FUNCTION_TABLE_ii[$80 & 8191]($$pre$i2$i$i611)|0);
      $82 = ($81|0)==(-1);
      if ($82) {
       HEAP32[$__e>>2] = 0;
       label = 30;
       break;
      }
     }
     if ($70) {
      break L16;
     }
    }
   } while(0);
   if ((label|0) == 30) {
    if (!($70)) {
     break;
    }
   }
   $83 = HEAP32[$__err>>2]|0;
   $84 = $83 | 2;
   HEAP32[$__err>>2] = $84;
  }
 } while(0);
 return;
}
function __ZNKSt3__120__time_get_c_storageIcE7__weeksEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__i$0$i$i$i = 0, $exitcond$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[624960]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((624960|0))|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    $4 = HEAP8[624968]|0;
    $5 = ($4<<24>>24)==(0);
    if ($5) {
     $6 = (___cxa_guard_acquire((624968|0))|0);
     $7 = ($6|0)==(0);
     if (!($7)) {
      $9 = 1207648;
      while(1) {
       $__i$0$i$i$i = 0;
       while(1) {
        $exitcond$i$i$i = ($__i$0$i$i$i|0)==(3);
        if ($exitcond$i$i$i) {
         break;
        }
        $8 = (($9) + ($__i$0$i$i$i<<2)|0);
        HEAP32[$8>>2] = 0;
        $10 = (($__i$0$i$i$i) + 1)|0;
        $__i$0$i$i$i = $10;
       }
       $11 = ((($9)) + 12|0);
       $12 = ($11|0)==((1207816)|0);
       if ($12) {
        break;
       } else {
        $9 = $11;
       }
      }
      (___cxa_atexit((4887|0),(0|0),(___dso_handle|0))|0);
      ___cxa_guard_release((624968|0));
     }
    }
    __THREW__ = 0;
    (invoke_iii(3360,(1207648|0),(621008|0))|0);
    $13 = __THREW__; __THREW__ = 0;
    $14 = $13&1;
    if (!($14)) {
     __THREW__ = 0;
     (invoke_iii(3360,((1207660)|0),(621015|0))|0);
     $15 = __THREW__; __THREW__ = 0;
     $16 = $15&1;
     if (!($16)) {
      __THREW__ = 0;
      (invoke_iii(3360,((1207672)|0),(621022|0))|0);
      $17 = __THREW__; __THREW__ = 0;
      $18 = $17&1;
      if (!($18)) {
       __THREW__ = 0;
       (invoke_iii(3360,((1207684)|0),(621030|0))|0);
       $19 = __THREW__; __THREW__ = 0;
       $20 = $19&1;
       if (!($20)) {
        __THREW__ = 0;
        (invoke_iii(3360,((1207696)|0),(621040|0))|0);
        $21 = __THREW__; __THREW__ = 0;
        $22 = $21&1;
        if (!($22)) {
         __THREW__ = 0;
         (invoke_iii(3360,((1207708)|0),(621049|0))|0);
         $23 = __THREW__; __THREW__ = 0;
         $24 = $23&1;
         if (!($24)) {
          __THREW__ = 0;
          (invoke_iii(3360,((1207720)|0),(621056|0))|0);
          $25 = __THREW__; __THREW__ = 0;
          $26 = $25&1;
          if (!($26)) {
           __THREW__ = 0;
           (invoke_iii(3360,((1207732)|0),(621065|0))|0);
           $27 = __THREW__; __THREW__ = 0;
           $28 = $27&1;
           if (!($28)) {
            __THREW__ = 0;
            (invoke_iii(3360,((1207744)|0),(621069|0))|0);
            $29 = __THREW__; __THREW__ = 0;
            $30 = $29&1;
            if (!($30)) {
             __THREW__ = 0;
             (invoke_iii(3360,((1207756)|0),(621073|0))|0);
             $31 = __THREW__; __THREW__ = 0;
             $32 = $31&1;
             if (!($32)) {
              __THREW__ = 0;
              (invoke_iii(3360,((1207768)|0),(621077|0))|0);
              $33 = __THREW__; __THREW__ = 0;
              $34 = $33&1;
              if (!($34)) {
               __THREW__ = 0;
               (invoke_iii(3360,((1207780)|0),(621081|0))|0);
               $35 = __THREW__; __THREW__ = 0;
               $36 = $35&1;
               if (!($36)) {
                __THREW__ = 0;
                (invoke_iii(3360,((1207792)|0),(621085|0))|0);
                $37 = __THREW__; __THREW__ = 0;
                $38 = $37&1;
                if (!($38)) {
                 __THREW__ = 0;
                 (invoke_iii(3360,((1207804)|0),(621089|0))|0);
                 $39 = __THREW__; __THREW__ = 0;
                 $40 = $39&1;
                 if (!($40)) {
                  HEAP32[301954] = 1207648;
                  ___cxa_guard_release((624960|0));
                  break;
                 }
                }
               }
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    }
    $42 = ___cxa_find_matching_catch_2()|0;
    $43 = tempRet0;
    ___cxa_guard_abort((624960|0));
    ___resumeException($42|0);
    // unreachable;
   }
  }
 } while(0);
 $41 = HEAP32[301954]|0;
 return ($41|0);
}
function ___cxx_global_array_dtor_10066($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207804));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207792));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207780));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207768));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207756));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207744));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207732));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207720));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207708));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207696));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207684));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207672));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207660));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev(1207648);
 return;
}
function __ZNKSt3__120__time_get_c_storageIcE8__monthsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $7 = 0, $8 = 0, $9 = 0, $__i$0$i$i$i = 0, $exitcond$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[624976]|0;
 $1 = ($0<<24>>24)==(0);
 L1: do {
  if ($1) {
   $2 = (___cxa_guard_acquire((624976|0))|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    $4 = HEAP8[624984]|0;
    $5 = ($4<<24>>24)==(0);
    if ($5) {
     $6 = (___cxa_guard_acquire((624984|0))|0);
     $7 = ($6|0)==(0);
     if (!($7)) {
      $9 = 1207820;
      while(1) {
       $__i$0$i$i$i = 0;
       while(1) {
        $exitcond$i$i$i = ($__i$0$i$i$i|0)==(3);
        if ($exitcond$i$i$i) {
         break;
        }
        $8 = (($9) + ($__i$0$i$i$i<<2)|0);
        HEAP32[$8>>2] = 0;
        $10 = (($__i$0$i$i$i) + 1)|0;
        $__i$0$i$i$i = $10;
       }
       $11 = ((($9)) + 12|0);
       $12 = ($11|0)==((1208108)|0);
       if ($12) {
        break;
       } else {
        $9 = $11;
       }
      }
      (___cxa_atexit((4888|0),(0|0),(___dso_handle|0))|0);
      ___cxa_guard_release((624984|0));
     }
    }
    __THREW__ = 0;
    (invoke_iii(3360,(1207820|0),(621093|0))|0);
    $13 = __THREW__; __THREW__ = 0;
    $14 = $13&1;
    do {
     if (!($14)) {
      __THREW__ = 0;
      (invoke_iii(3360,((1207832)|0),(621101|0))|0);
      $15 = __THREW__; __THREW__ = 0;
      $16 = $15&1;
      if (!($16)) {
       __THREW__ = 0;
       (invoke_iii(3360,((1207844)|0),(621110|0))|0);
       $17 = __THREW__; __THREW__ = 0;
       $18 = $17&1;
       if (!($18)) {
        __THREW__ = 0;
        (invoke_iii(3360,((1207856)|0),(621116|0))|0);
        $19 = __THREW__; __THREW__ = 0;
        $20 = $19&1;
        if (!($20)) {
         __THREW__ = 0;
         (invoke_iii(3360,((1207868)|0),(621122|0))|0);
         $21 = __THREW__; __THREW__ = 0;
         $22 = $21&1;
         if (!($22)) {
          __THREW__ = 0;
          (invoke_iii(3360,((1207880)|0),(621126|0))|0);
          $23 = __THREW__; __THREW__ = 0;
          $24 = $23&1;
          if (!($24)) {
           __THREW__ = 0;
           (invoke_iii(3360,((1207892)|0),(621131|0))|0);
           $25 = __THREW__; __THREW__ = 0;
           $26 = $25&1;
           if (!($26)) {
            __THREW__ = 0;
            (invoke_iii(3360,((1207904)|0),(621136|0))|0);
            $27 = __THREW__; __THREW__ = 0;
            $28 = $27&1;
            if (!($28)) {
             __THREW__ = 0;
             (invoke_iii(3360,((1207916)|0),(621143|0))|0);
             $29 = __THREW__; __THREW__ = 0;
             $30 = $29&1;
             if (!($30)) {
              __THREW__ = 0;
              (invoke_iii(3360,((1207928)|0),(621153|0))|0);
              $31 = __THREW__; __THREW__ = 0;
              $32 = $31&1;
              if (!($32)) {
               __THREW__ = 0;
               (invoke_iii(3360,((1207940)|0),(621161|0))|0);
               $33 = __THREW__; __THREW__ = 0;
               $34 = $33&1;
               if (!($34)) {
                __THREW__ = 0;
                (invoke_iii(3360,((1207952)|0),(621170|0))|0);
                $35 = __THREW__; __THREW__ = 0;
                $36 = $35&1;
                if (!($36)) {
                 __THREW__ = 0;
                 (invoke_iii(3360,((1207964)|0),(621179|0))|0);
                 $37 = __THREW__; __THREW__ = 0;
                 $38 = $37&1;
                 if (!($38)) {
                  __THREW__ = 0;
                  (invoke_iii(3360,((1207976)|0),(621183|0))|0);
                  $39 = __THREW__; __THREW__ = 0;
                  $40 = $39&1;
                  if (!($40)) {
                   __THREW__ = 0;
                   (invoke_iii(3360,((1207988)|0),(621187|0))|0);
                   $41 = __THREW__; __THREW__ = 0;
                   $42 = $41&1;
                   if (!($42)) {
                    __THREW__ = 0;
                    (invoke_iii(3360,((1208000)|0),(621191|0))|0);
                    $43 = __THREW__; __THREW__ = 0;
                    $44 = $43&1;
                    if ($44) {
                     break;
                    }
                    __THREW__ = 0;
                    (invoke_iii(3360,((1208012)|0),(621122|0))|0);
                    $45 = __THREW__; __THREW__ = 0;
                    $46 = $45&1;
                    if ($46) {
                     break;
                    }
                    __THREW__ = 0;
                    (invoke_iii(3360,((1208024)|0),(621195|0))|0);
                    $47 = __THREW__; __THREW__ = 0;
                    $48 = $47&1;
                    if ($48) {
                     break;
                    }
                    __THREW__ = 0;
                    (invoke_iii(3360,((1208036)|0),(621199|0))|0);
                    $49 = __THREW__; __THREW__ = 0;
                    $50 = $49&1;
                    if ($50) {
                     break;
                    }
                    __THREW__ = 0;
                    (invoke_iii(3360,((1208048)|0),(621203|0))|0);
                    $51 = __THREW__; __THREW__ = 0;
                    $52 = $51&1;
                    if ($52) {
                     break;
                    }
                    __THREW__ = 0;
                    (invoke_iii(3360,((1208060)|0),(621207|0))|0);
                    $53 = __THREW__; __THREW__ = 0;
                    $54 = $53&1;
                    if ($54) {
                     break;
                    }
                    __THREW__ = 0;
                    (invoke_iii(3360,((1208072)|0),(621211|0))|0);
                    $55 = __THREW__; __THREW__ = 0;
                    $56 = $55&1;
                    if ($56) {
                     break;
                    }
                    __THREW__ = 0;
                    (invoke_iii(3360,((1208084)|0),(621215|0))|0);
                    $57 = __THREW__; __THREW__ = 0;
                    $58 = $57&1;
                    if ($58) {
                     break;
                    }
                    __THREW__ = 0;
                    (invoke_iii(3360,((1208096)|0),(621219|0))|0);
                    $59 = __THREW__; __THREW__ = 0;
                    $60 = $59&1;
                    if ($60) {
                     break;
                    }
                    HEAP32[302027] = 1207820;
                    ___cxa_guard_release((624976|0));
                    break L1;
                   }
                  }
                 }
                }
               }
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    } while(0);
    $62 = ___cxa_find_matching_catch_2()|0;
    $63 = tempRet0;
    ___cxa_guard_abort((624976|0));
    ___resumeException($62|0);
    // unreachable;
   }
  }
 } while(0);
 $61 = HEAP32[302027]|0;
 return ($61|0);
}
function ___cxx_global_array_dtor_69($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208096));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208084));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208072));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208060));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208048));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208036));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208024));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208012));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208000));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207988));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207976));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207964));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207952));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207940));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207928));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207916));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207904));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207892));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207880));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207868));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207856));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207844));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1207832));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev(1207820);
 return;
}
function __ZNKSt3__120__time_get_c_storageIcE7__am_pmEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $__i$0$i$i$i = 0, $exitcond$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[624992]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((624992|0))|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    $4 = HEAP8[625000]|0;
    $5 = ($4<<24>>24)==(0);
    if ($5) {
     $6 = (___cxa_guard_acquire((625000|0))|0);
     $7 = ($6|0)==(0);
     if (!($7)) {
      $9 = 1208112;
      while(1) {
       $__i$0$i$i$i = 0;
       while(1) {
        $exitcond$i$i$i = ($__i$0$i$i$i|0)==(3);
        if ($exitcond$i$i$i) {
         break;
        }
        $8 = (($9) + ($__i$0$i$i$i<<2)|0);
        HEAP32[$8>>2] = 0;
        $10 = (($__i$0$i$i$i) + 1)|0;
        $__i$0$i$i$i = $10;
       }
       $11 = ((($9)) + 12|0);
       $12 = ($11|0)==((1208400)|0);
       if ($12) {
        break;
       } else {
        $9 = $11;
       }
      }
      (___cxa_atexit((4889|0),(0|0),(___dso_handle|0))|0);
      ___cxa_guard_release((625000|0));
     }
    }
    __THREW__ = 0;
    (invoke_iii(3360,(1208112|0),(621223|0))|0);
    $13 = __THREW__; __THREW__ = 0;
    $14 = $13&1;
    if (!($14)) {
     __THREW__ = 0;
     (invoke_iii(3360,((1208124)|0),(621226|0))|0);
     $15 = __THREW__; __THREW__ = 0;
     $16 = $15&1;
     if (!($16)) {
      HEAP32[302100] = 1208112;
      ___cxa_guard_release((624992|0));
      break;
     }
    }
    $18 = ___cxa_find_matching_catch_2()|0;
    $19 = tempRet0;
    ___cxa_guard_abort((624992|0));
    ___resumeException($18|0);
    // unreachable;
   }
  }
 } while(0);
 $17 = HEAP32[302100]|0;
 return ($17|0);
}
function ___cxx_global_array_dtor_117($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208388));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208376));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208364));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208352));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208340));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208328));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208316));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208304));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208292));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208280));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208268));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208256));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208244));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208232));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208220));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208208));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208196));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208184));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208172));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208160));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208148));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208136));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev((1208124));
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev(1208112);
 return;
}
function __ZNKSt3__120__time_get_c_storageIcE3__cEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[625008]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((625008|0))|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    __THREW__ = 0;
    invoke_viii(3460,(1208404|0),(621229|0),20);
    $4 = __THREW__; __THREW__ = 0;
    $5 = $4&1;
    if ($5) {
     $6 = ___cxa_find_matching_catch_2()|0;
     $7 = tempRet0;
     ___cxa_guard_abort((625008|0));
     ___resumeException($6|0);
     // unreachable;
    } else {
     (___cxa_atexit((3974|0),(1208404|0),(___dso_handle|0))|0);
     ___cxa_guard_release((625008|0));
     break;
    }
   }
  }
 } while(0);
 return (1208404|0);
}
function __ZNKSt3__120__time_get_c_storageIcE3__rEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[625016]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((625016|0))|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    __THREW__ = 0;
    invoke_viii(3460,(1208416|0),(621250|0),11);
    $4 = __THREW__; __THREW__ = 0;
    $5 = $4&1;
    if ($5) {
     $6 = ___cxa_find_matching_catch_2()|0;
     $7 = tempRet0;
     ___cxa_guard_abort((625016|0));
     ___resumeException($6|0);
     // unreachable;
    } else {
     (___cxa_atexit((3974|0),(1208416|0),(___dso_handle|0))|0);
     ___cxa_guard_release((625016|0));
     break;
    }
   }
  }
 } while(0);
 return (1208416|0);
}
function __ZNKSt3__120__time_get_c_storageIcE3__xEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[625024]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((625024|0))|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    __THREW__ = 0;
    invoke_viii(3460,(1208428|0),(621262|0),8);
    $4 = __THREW__; __THREW__ = 0;
    $5 = $4&1;
    if ($5) {
     $6 = ___cxa_find_matching_catch_2()|0;
     $7 = tempRet0;
     ___cxa_guard_abort((625024|0));
     ___resumeException($6|0);
     // unreachable;
    } else {
     (___cxa_atexit((3974|0),(1208428|0),(___dso_handle|0))|0);
     ___cxa_guard_release((625024|0));
     break;
    }
   }
  }
 } while(0);
 return (1208428|0);
}
function __ZNKSt3__120__time_get_c_storageIcE3__XEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[625032]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((625032|0))|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    __THREW__ = 0;
    invoke_viii(3460,(1208440|0),(621271|0),8);
    $4 = __THREW__; __THREW__ = 0;
    $5 = $4&1;
    if ($5) {
     $6 = ___cxa_find_matching_catch_2()|0;
     $7 = tempRet0;
     ___cxa_guard_abort((625032|0));
     ___resumeException($6|0);
     // unreachable;
    } else {
     (___cxa_atexit((3974|0),(1208440|0),(___dso_handle|0))|0);
     ___cxa_guard_release((625032|0));
     break;
    }
   }
  }
 } while(0);
 return (1208440|0);
}
function __ZNSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE13do_date_orderEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 2;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11do_get_timeES4_S4_RNS_8ios_baseERjP2tm($this,$__b,$__e,$__iob,$__err,$__tm) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__tm = $__tm|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $2;
 $3 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $3;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $4 = (__ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__tm,155416,(155448))|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($this,$__b,$__e,$__iob,$__err,$__tm,$__fmtb,$__fmte) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__tm = $__tm|0;
 $__fmtb = $__fmtb|0;
 $__fmte = $__fmte|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i21 = 0, $$0$i$i$i$i6 = 0, $$0$i$i1$i$i = 0, $$0$i$i1$i$i12 = 0, $$0$i$i1$i$i27 = 0, $$0$i$i18 = 0, $$0$ph = 0, $$0$ph$be = 0, $$0$pn = 0, $$3 = 0, $$3$lcssa = 0, $$byval_copy = 0, $$byval_copy1 = 0, $$lcssa114 = 0, $$lcssa117 = 0, $$lcssa118 = 0, $$lcssa118$lcssa = 0, $$lcssa118$lcssa121 = 0;
 var $$lcssa119 = 0, $$lcssa120 = 0, $$lcssa123 = 0, $$lcssa125 = 0, $$lcssa128 = 0, $$lcssa129 = 0, $$ph = 0, $$pre = 0, $$pre$i$i$i23 = 0, $$pre$i$i$i8 = 0, $$pre$i3$i$i14 = 0, $$pre$i3$i$i1445 = 0, $$pre$i3$i$i1446 = 0, $$pre$pre = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $__cmd$0 = 0, $__opt$0 = 0, $or$cond = 0, $phitmp$i4$i$i15 = 0, $tmp$i$i17 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 16|0;
 $$byval_copy = sp + 12|0;
 $0 = sp + 8|0;
 $1 = sp + 4|0;
 $2 = sp;
 $3 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$0>>2] = $3;
 __THREW__ = 0;
 $4 = (invoke_iii(3374,($0|0),(1207608|0))|0);
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $38 = ___cxa_find_matching_catch_2()|0;
  $39 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($3)|0);
  ___resumeException($38|0);
  // unreachable;
 }
 (__ZNSt3__114__shared_count16__release_sharedEv($3)|0);
 HEAP32[$__err>>2] = 0;
 $$0$ph = $__fmtb;$$pre = 0;
 L4: while(1) {
  $7 = ($$0$ph|0)!=($__fmte|0);
  $9 = $$pre;
  while(1) {
   $8 = ($9|0)==(0);
   $or$cond = $7 & $8;
   $10 = HEAP32[$__b>>2]|0;
   if (!($or$cond)) {
    $$pre$i$i$i23 = $10;
    break L4;
   }
   $11 = ($10|0)==(0|0);
   $12 = $10;
   if ($11) {
    $126 = 0;$182 = $12;$183 = 1;
   } else {
    $13 = ((($10)) + 12|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = ((($10)) + 16|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ($14|0)==($16|0);
    if ($17) {
     $18 = HEAP32[$10>>2]|0;
     $19 = ((($18)) + 36|0);
     $20 = HEAP32[$19>>2]|0;
     $21 = (FUNCTION_TABLE_ii[$20 & 8191]($10)|0);
     $$0$i$i$i$i = $21;
    } else {
     $22 = HEAP32[$14>>2]|0;
     $$0$i$i$i$i = $22;
    }
    $23 = ($$0$i$i$i$i|0)==(-1);
    if ($23) {
     HEAP32[$__b>>2] = 0;
     $126 = 0;$182 = 0;$183 = 1;
    } else {
     $126 = $10;$182 = $12;$183 = 0;
    }
   }
   $24 = HEAP32[$__e>>2]|0;
   $25 = ($24|0)==(0|0);
   $26 = $24;
   do {
    if ($25) {
     $$ph = $26;
     label = 17;
    } else {
     $27 = ((($24)) + 12|0);
     $28 = HEAP32[$27>>2]|0;
     $29 = ((($24)) + 16|0);
     $30 = HEAP32[$29>>2]|0;
     $31 = ($28|0)==($30|0);
     if ($31) {
      $32 = HEAP32[$24>>2]|0;
      $33 = ((($32)) + 36|0);
      $34 = HEAP32[$33>>2]|0;
      $35 = (FUNCTION_TABLE_ii[$34 & 8191]($24)|0);
      $$0$i$i1$i$i = $35;
     } else {
      $36 = HEAP32[$28>>2]|0;
      $$0$i$i1$i$i = $36;
     }
     $37 = ($$0$i$i1$i$i|0)==(-1);
     if ($37) {
      HEAP32[$__e>>2] = 0;
      $$ph = 0;
      label = 17;
      break;
     } else {
      if ($183) {
       $184 = $24;$185 = $26;
       break;
      } else {
       $$lcssa117 = $126;
       label = 19;
       break L4;
      }
     }
    }
   } while(0);
   if ((label|0) == 17) {
    label = 0;
    if ($183) {
     $$lcssa117 = $126;
     label = 19;
     break L4;
    } else {
     $184 = 0;$185 = $$ph;
    }
   }
   $40 = HEAP32[$$0$ph>>2]|0;
   $41 = HEAP32[$4>>2]|0;
   $42 = ((($41)) + 52|0);
   $43 = HEAP32[$42>>2]|0;
   $44 = (FUNCTION_TABLE_iiii[$43 & 8191]($4,$40,0)|0);
   $45 = ($44<<24>>24)==(37);
   if ($45) {
    $$lcssa114 = $182;$$lcssa118 = $126;$$lcssa125 = $185;
    label = 22;
    break;
   }
   $66 = HEAP32[$$0$ph>>2]|0;
   $67 = HEAP32[$4>>2]|0;
   $68 = ((($67)) + 12|0);
   $69 = HEAP32[$68>>2]|0;
   $70 = (FUNCTION_TABLE_iiii[$69 & 8191]($4,8192,$66)|0);
   if ($70) {
    $$lcssa119 = $126;$$lcssa123 = $184;
    label = 30;
    break;
   }
   $125 = ((($126)) + 12|0);
   $127 = HEAP32[$125>>2]|0;
   $128 = ((($126)) + 16|0);
   $129 = HEAP32[$128>>2]|0;
   $130 = ($127|0)==($129|0);
   if ($130) {
    $131 = HEAP32[$126>>2]|0;
    $132 = ((($131)) + 36|0);
    $133 = HEAP32[$132>>2]|0;
    $134 = (FUNCTION_TABLE_ii[$133 & 8191]($126)|0);
    $$0$i$i18 = $134;
   } else {
    $135 = HEAP32[$127>>2]|0;
    $$0$i$i18 = $135;
   }
   $136 = HEAP32[$4>>2]|0;
   $137 = ((($136)) + 28|0);
   $138 = HEAP32[$137>>2]|0;
   $139 = (FUNCTION_TABLE_iii[$138 & 8191]($4,$$0$i$i18)|0);
   $140 = HEAP32[$$0$ph>>2]|0;
   $141 = HEAP32[$4>>2]|0;
   $142 = ((($141)) + 28|0);
   $143 = HEAP32[$142>>2]|0;
   $144 = (FUNCTION_TABLE_iii[$143 & 8191]($4,$140)|0);
   $145 = ($139|0)==($144|0);
   if ($145) {
    $$lcssa120 = $126;$$lcssa128 = $125;$$lcssa129 = $128;
    label = 59;
    break;
   }
   HEAP32[$__err>>2] = 4;
   $9 = 4;
  }
  L36: do {
   if ((label|0) == 22) {
    label = 0;
    $46 = ((($$0$ph)) + 4|0);
    $47 = ($46|0)==($__fmte|0);
    if ($47) {
     $$lcssa118$lcssa = $$lcssa118;
     label = 23;
     break L4;
    }
    $48 = HEAP32[$46>>2]|0;
    $49 = HEAP32[$4>>2]|0;
    $50 = ((($49)) + 52|0);
    $51 = HEAP32[$50>>2]|0;
    $52 = (FUNCTION_TABLE_iiii[$51 & 8191]($4,$48,0)|0);
    switch ($52<<24>>24) {
    case 48: case 69:  {
     $53 = ((($$0$ph)) + 8|0);
     $54 = ($53|0)==($__fmte|0);
     if ($54) {
      $$lcssa118$lcssa121 = $$lcssa118;
      label = 28;
      break L4;
     }
     $55 = HEAP32[$53>>2]|0;
     $56 = HEAP32[$4>>2]|0;
     $57 = ((($56)) + 52|0);
     $58 = HEAP32[$57>>2]|0;
     $59 = (FUNCTION_TABLE_iiii[$58 & 8191]($4,$55,0)|0);
     $65 = $46;$__cmd$0 = $59;$__opt$0 = $52;
     break;
    }
    default: {
     $65 = $$0$ph;$__cmd$0 = $52;$__opt$0 = 0;
    }
    }
    $60 = HEAP32[$this>>2]|0;
    $61 = ((($60)) + 36|0);
    $62 = HEAP32[$61>>2]|0;
    HEAP32[$1>>2] = $$lcssa114;
    HEAP32[$2>>2] = $$lcssa125;
    ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
    ;HEAP32[$$byval_copy1>>2]=HEAP32[$2>>2]|0;
    $63 = (FUNCTION_TABLE_iiiiiiiii[$62 & 8191]($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__tm,$__cmd$0,$__opt$0)|0);
    HEAP32[$__b>>2] = $63;
    $64 = ((($65)) + 8|0);
    $$0$ph$be = $64;
   }
   else if ((label|0) == 30) {
    label = 0;
    $$0$pn = $$0$ph;
    while(1) {
     $$3 = ((($$0$pn)) + 4|0);
     $71 = ($$3|0)==($__fmte|0);
     if ($71) {
      $$3$lcssa = $__fmte;
      break;
     }
     $72 = HEAP32[$$3>>2]|0;
     $73 = HEAP32[$4>>2]|0;
     $74 = ((($73)) + 12|0);
     $75 = HEAP32[$74>>2]|0;
     $76 = (FUNCTION_TABLE_iiii[$75 & 8191]($4,8192,$72)|0);
     if ($76) {
      $$0$pn = $$3;
     } else {
      $$3$lcssa = $$3;
      break;
     }
    }
    $$pre$i$i$i8 = $$lcssa119;$$pre$i3$i$i14 = $$lcssa123;$90 = $$lcssa123;
    while(1) {
     $77 = ($$pre$i$i$i8|0)==(0|0);
     if ($77) {
      $102 = 1;$104 = 0;
     } else {
      $78 = ((($$pre$i$i$i8)) + 12|0);
      $79 = HEAP32[$78>>2]|0;
      $80 = ((($$pre$i$i$i8)) + 16|0);
      $81 = HEAP32[$80>>2]|0;
      $82 = ($79|0)==($81|0);
      if ($82) {
       $83 = HEAP32[$$pre$i$i$i8>>2]|0;
       $84 = ((($83)) + 36|0);
       $85 = HEAP32[$84>>2]|0;
       $86 = (FUNCTION_TABLE_ii[$85 & 8191]($$pre$i$i$i8)|0);
       $$0$i$i$i$i6 = $86;
      } else {
       $87 = HEAP32[$79>>2]|0;
       $$0$i$i$i$i6 = $87;
      }
      $88 = ($$0$i$i$i$i6|0)==(-1);
      if ($88) {
       HEAP32[$__b>>2] = 0;
       $102 = 1;$104 = 0;
      } else {
       $102 = 0;$104 = $$pre$i$i$i8;
      }
     }
     $89 = ($90|0)==(0|0);
     do {
      if ($89) {
       $$pre$i3$i$i1446 = $$pre$i3$i$i14;
       label = 46;
      } else {
       $91 = ((($90)) + 12|0);
       $92 = HEAP32[$91>>2]|0;
       $93 = ((($90)) + 16|0);
       $94 = HEAP32[$93>>2]|0;
       $95 = ($92|0)==($94|0);
       if ($95) {
        $96 = HEAP32[$90>>2]|0;
        $97 = ((($96)) + 36|0);
        $98 = HEAP32[$97>>2]|0;
        $99 = (FUNCTION_TABLE_ii[$98 & 8191]($90)|0);
        $$0$i$i1$i$i12 = $99;
       } else {
        $100 = HEAP32[$92>>2]|0;
        $$0$i$i1$i$i12 = $100;
       }
       $101 = ($$0$i$i1$i$i12|0)==(-1);
       if ($101) {
        HEAP32[$__e>>2] = 0;
        $$pre$i3$i$i1446 = 0;
        label = 46;
        break;
       } else {
        $phitmp$i4$i$i15 = ($$pre$i3$i$i14|0)==(0|0);
        $tmp$i$i17 = $102 ^ $phitmp$i4$i$i15;
        if ($tmp$i$i17) {
         $$pre$i3$i$i1445 = $$pre$i3$i$i14;$186 = $$pre$i3$i$i14;
         break;
        } else {
         $$0$ph$be = $$3$lcssa;
         break L36;
        }
       }
      }
     } while(0);
     if ((label|0) == 46) {
      label = 0;
      if ($102) {
       $$0$ph$be = $$3$lcssa;
       break L36;
      } else {
       $$pre$i3$i$i1445 = $$pre$i3$i$i1446;$186 = 0;
      }
     }
     $103 = ((($104)) + 12|0);
     $105 = HEAP32[$103>>2]|0;
     $106 = ((($104)) + 16|0);
     $107 = HEAP32[$106>>2]|0;
     $108 = ($105|0)==($107|0);
     if ($108) {
      $109 = HEAP32[$104>>2]|0;
      $110 = ((($109)) + 36|0);
      $111 = HEAP32[$110>>2]|0;
      $112 = (FUNCTION_TABLE_ii[$111 & 8191]($104)|0);
      $$0$i$i = $112;
     } else {
      $113 = HEAP32[$105>>2]|0;
      $$0$i$i = $113;
     }
     $114 = HEAP32[$4>>2]|0;
     $115 = ((($114)) + 12|0);
     $116 = HEAP32[$115>>2]|0;
     $117 = (FUNCTION_TABLE_iiii[$116 & 8191]($4,8192,$$0$i$i)|0);
     if (!($117)) {
      $$0$ph$be = $$3$lcssa;
      break L36;
     }
     $118 = HEAP32[$103>>2]|0;
     $119 = HEAP32[$106>>2]|0;
     $120 = ($118|0)==($119|0);
     if ($120) {
      $121 = HEAP32[$104>>2]|0;
      $122 = ((($121)) + 40|0);
      $123 = HEAP32[$122>>2]|0;
      (FUNCTION_TABLE_ii[$123 & 8191]($104)|0);
      $$pre$i$i$i8 = $104;$$pre$i3$i$i14 = $$pre$i3$i$i1445;$90 = $186;
      continue;
     } else {
      $124 = ((($118)) + 4|0);
      HEAP32[$103>>2] = $124;
      $$pre$i$i$i8 = $104;$$pre$i3$i$i14 = $$pre$i3$i$i1445;$90 = $186;
      continue;
     }
    }
   }
   else if ((label|0) == 59) {
    label = 0;
    $146 = HEAP32[$$lcssa128>>2]|0;
    $147 = HEAP32[$$lcssa129>>2]|0;
    $148 = ($146|0)==($147|0);
    if ($148) {
     $149 = HEAP32[$$lcssa120>>2]|0;
     $150 = ((($149)) + 40|0);
     $151 = HEAP32[$150>>2]|0;
     (FUNCTION_TABLE_ii[$151 & 8191]($$lcssa120)|0);
    } else {
     $152 = ((($146)) + 4|0);
     HEAP32[$$lcssa128>>2] = $152;
    }
    $153 = ((($$0$ph)) + 4|0);
    $$0$ph$be = $153;
   }
  } while(0);
  $$pre$pre = HEAP32[$__err>>2]|0;
  $$0$ph = $$0$ph$be;$$pre = $$pre$pre;
 }
 if ((label|0) == 19) {
  HEAP32[$__err>>2] = 4;
  $$pre$i$i$i23 = $$lcssa117;
 }
 else if ((label|0) == 23) {
  HEAP32[$__err>>2] = 4;
  $$pre$i$i$i23 = $$lcssa118$lcssa;
 }
 else if ((label|0) == 28) {
  HEAP32[$__err>>2] = 4;
  $$pre$i$i$i23 = $$lcssa118$lcssa121;
 }
 $154 = ($$pre$i$i$i23|0)==(0|0);
 if ($154) {
  $181 = 0;$187 = 1;
 } else {
  $155 = ((($$pre$i$i$i23)) + 12|0);
  $156 = HEAP32[$155>>2]|0;
  $157 = ((($$pre$i$i$i23)) + 16|0);
  $158 = HEAP32[$157>>2]|0;
  $159 = ($156|0)==($158|0);
  if ($159) {
   $160 = HEAP32[$$pre$i$i$i23>>2]|0;
   $161 = ((($160)) + 36|0);
   $162 = HEAP32[$161>>2]|0;
   $163 = (FUNCTION_TABLE_ii[$162 & 8191]($$pre$i$i$i23)|0);
   $$0$i$i$i$i21 = $163;
  } else {
   $164 = HEAP32[$156>>2]|0;
   $$0$i$i$i$i21 = $164;
  }
  $165 = ($$0$i$i$i$i21|0)==(-1);
  if ($165) {
   HEAP32[$__b>>2] = 0;
   $181 = 0;$187 = 1;
  } else {
   $181 = $$pre$i$i$i23;$187 = 0;
  }
 }
 $166 = HEAP32[$__e>>2]|0;
 $167 = ($166|0)==(0|0);
 do {
  if ($167) {
   label = 77;
  } else {
   $168 = ((($166)) + 12|0);
   $169 = HEAP32[$168>>2]|0;
   $170 = ((($166)) + 16|0);
   $171 = HEAP32[$170>>2]|0;
   $172 = ($169|0)==($171|0);
   if ($172) {
    $173 = HEAP32[$166>>2]|0;
    $174 = ((($173)) + 36|0);
    $175 = HEAP32[$174>>2]|0;
    $176 = (FUNCTION_TABLE_ii[$175 & 8191]($166)|0);
    $$0$i$i1$i$i27 = $176;
   } else {
    $177 = HEAP32[$169>>2]|0;
    $$0$i$i1$i$i27 = $177;
   }
   $178 = ($$0$i$i1$i$i27|0)==(-1);
   if ($178) {
    HEAP32[$__e>>2] = 0;
    label = 77;
    break;
   } else {
    if ($187) {
     break;
    } else {
     label = 79;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 77) {
  if ($187) {
   label = 79;
  }
 }
 if ((label|0) == 79) {
  $179 = HEAP32[$__err>>2]|0;
  $180 = $179 | 2;
  HEAP32[$__err>>2] = $180;
 }
 STACKTOP = sp;return ($181|0);
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11do_get_dateES4_S4_RNS_8ios_baseERjP2tm($this,$__b,$__e,$__iob,$__err,$__tm) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__tm = $__tm|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = ((($this)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 20|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_ii[$5 & 8191]($2)|0);
 $7 = HEAP32[$__b>>2]|0;
 HEAP32[$0>>2] = $7;
 $8 = HEAP32[$__e>>2]|0;
 HEAP32[$1>>2] = $8;
 $9 = HEAP8[$6>>0]|0;
 $10 = $9 & 1;
 $11 = ($10<<24>>24)==(0);
 $12 = ((($6)) + 8|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($6)) + 4|0);
 $15 = $11 ? $14 : $13;
 $16 = HEAP32[$14>>2]|0;
 $17 = $9&255;
 $18 = $17 >>> 1;
 $19 = $11 ? $18 : $16;
 $20 = (($15) + ($19<<2)|0);
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$1>>2]|0;
 $21 = (__ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($this,$$byval_copy,$$byval_copy1,$__iob,$__err,$__tm,$15,$20)|0);
 STACKTOP = sp;return ($21|0);
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE14do_get_weekdayES4_S4_RNS_8ios_baseERjP2tm($this,$__b,$__e,$__iob,$__err,$__tm) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__tm = $__tm|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$0>>2] = $2;
 __THREW__ = 0;
 $3 = (invoke_iii(3374,($0|0),(1207608|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($2)|0);
  ___resumeException($9|0);
  // unreachable;
 } else {
  (__ZNSt3__114__shared_count16__release_sharedEv($2)|0);
  $6 = ((($__tm)) + 24|0);
  $7 = HEAP32[$__e>>2]|0;
  HEAP32[$1>>2] = $7;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
  __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__get_weekdaynameERiRS4_S4_RjRKNS_5ctypeIwEE($this,$6,$__b,$$byval_copy,$__err,$3);
  $8 = HEAP32[$__b>>2]|0;
  STACKTOP = sp;return ($8|0);
 }
 return (0)|0;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__get_weekdaynameERiRS4_S4_RjRKNS_5ctypeIwEE($this,$__w,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__w = $__w|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = ((($this)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_ii[$3 & 8191]($1)|0);
 $5 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $5;
 $6 = ((($4)) + 168|0);
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $7 = (__ZNSt3__114__scan_keywordINS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEPKNS_12basic_stringIwS3_NS_9allocatorIwEEEENS_5ctypeIwEEEET0_RT_SE_SD_SD_RKT1_Rjb($__b,$$byval_copy,$4,$6,$__ct,$__err,0)|0);
 $8 = $7;
 $9 = $4;
 $10 = (($8) - ($9))|0;
 $11 = ($10|0)<(168);
 if ($11) {
  $12 = (($10|0) / 12)&-1;
  $13 = (($12|0) % 7)&-1;
  HEAP32[$__w>>2] = $13;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE16do_get_monthnameES4_S4_RNS_8ios_baseERjP2tm($this,$__b,$__e,$__iob,$__err,$__tm) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__tm = $__tm|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$0>>2] = $2;
 __THREW__ = 0;
 $3 = (invoke_iii(3374,($0|0),(1207608|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($2)|0);
  ___resumeException($9|0);
  // unreachable;
 } else {
  (__ZNSt3__114__shared_count16__release_sharedEv($2)|0);
  $6 = ((($__tm)) + 16|0);
  $7 = HEAP32[$__e>>2]|0;
  HEAP32[$1>>2] = $7;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
  __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE15__get_monthnameERiRS4_S4_RjRKNS_5ctypeIwEE($this,$6,$__b,$$byval_copy,$__err,$3);
  $8 = HEAP32[$__b>>2]|0;
  STACKTOP = sp;return ($8|0);
 }
 return (0)|0;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE15__get_monthnameERiRS4_S4_RjRKNS_5ctypeIwEE($this,$__m,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__m = $__m|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = ((($this)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($1)|0);
 $6 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $6;
 $7 = ((($5)) + 288|0);
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $8 = (__ZNSt3__114__scan_keywordINS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEPKNS_12basic_stringIwS3_NS_9allocatorIwEEEENS_5ctypeIwEEEET0_RT_SE_SD_SD_RKT1_Rjb($__b,$$byval_copy,$5,$7,$__ct,$__err,0)|0);
 $9 = $8;
 $10 = $5;
 $11 = (($9) - ($10))|0;
 $12 = ($11|0)<(288);
 if ($12) {
  $13 = (($11|0) / 12)&-1;
  $14 = (($13|0) % 12)&-1;
  HEAP32[$__m>>2] = $14;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11do_get_yearES4_S4_RNS_8ios_baseERjP2tm($this,$__b,$__e,$__iob,$__err,$__tm) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__tm = $__tm|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 8|0;
 $0 = sp + 4|0;
 $1 = sp;
 $2 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$0>>2] = $2;
 __THREW__ = 0;
 $3 = (invoke_iii(3374,($0|0),(1207608|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($2)|0);
  ___resumeException($9|0);
  // unreachable;
 } else {
  (__ZNSt3__114__shared_count16__release_sharedEv($2)|0);
  $6 = ((($__tm)) + 20|0);
  $7 = HEAP32[$__e>>2]|0;
  HEAP32[$1>>2] = $7;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$1>>2]|0;
  __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE10__get_yearERiRS4_S4_RjRKNS_5ctypeIwEE($this,$6,$__b,$$byval_copy,$__err,$3);
  $8 = HEAP32[$__b>>2]|0;
  STACKTOP = sp;return ($8|0);
 }
 return (0)|0;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE10__get_yearERiRS4_S4_RjRKNS_5ctypeIwEE($this,$__y,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__y = $__y|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$ = 0, $$byval_copy = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__t$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,4)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 if ($5) {
  $6 = ($2|0)<(69);
  if ($6) {
   $7 = (($2) + 2000)|0;
   $__t$0 = $7;
  } else {
   $$off = (($2) + -69)|0;
   $8 = ($$off>>>0)<(31);
   $9 = (($2) + 1900)|0;
   $$ = $8 ? $9 : $2;
   $__t$0 = $$;
  }
  $10 = (($__t$0) + -1900)|0;
  HEAP32[$__y>>2] = $10;
 }
 STACKTOP = sp;return;
}
function __ZNSt3__120__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$__e,$__err,$__ct,$__n) {
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 $__n = $__n|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i17 = 0, $$0$i$i$i$i2 = 0, $$0$i$i1$i$i = 0, $$0$i$i1$i$i23 = 0, $$0$i$i1$i$i8 = 0, $$0$i$i14 = 0, $$0$in = 0, $$2 = 0, $$lcssa = 0, $$pre$i$i$i = 0, $$pre$i$i$i19 = 0, $$pre$i$i$i4 = 0, $$pre$i3$i$i10 = 0, $$pre$i3$i$i25 = 0, $$pre$i3$i$i25$lcssa = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__r$0 = 0, $__r$0$in = 0, $__r$0$lcssa = 0, $phitmp$i$i$i = 0, $phitmp$i$i$i20 = 0, $phitmp$i$i$i5 = 0, $phitmp$i4$i$i11 = 0, $tmp$i$i13 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$__b>>2]|0;
 $1 = ($0|0)==(0|0);
 do {
  if ($1) {
   $148 = 1;
  } else {
   $2 = ((($0)) + 12|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = ((($0)) + 16|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = ($3|0)==($5|0);
   if ($6) {
    $7 = HEAP32[$0>>2]|0;
    $8 = ((($7)) + 36|0);
    $9 = HEAP32[$8>>2]|0;
    $10 = (FUNCTION_TABLE_ii[$9 & 8191]($0)|0);
    $$0$i$i$i$i = $10;
   } else {
    $11 = HEAP32[$3>>2]|0;
    $$0$i$i$i$i = $11;
   }
   $12 = ($$0$i$i$i$i|0)==(-1);
   if ($12) {
    HEAP32[$__b>>2] = 0;
    $148 = 1;
    break;
   } else {
    $$pre$i$i$i = HEAP32[$__b>>2]|0;
    $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
    $148 = $phitmp$i$i$i;
    break;
   }
  }
 } while(0);
 $13 = HEAP32[$__e>>2]|0;
 $14 = ($13|0)==(0|0);
 do {
  if ($14) {
   label = 14;
  } else {
   $15 = ((($13)) + 12|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = ((($13)) + 16|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = ($16|0)==($18|0);
   if ($19) {
    $20 = HEAP32[$13>>2]|0;
    $21 = ((($20)) + 36|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = (FUNCTION_TABLE_ii[$22 & 8191]($13)|0);
    $$0$i$i1$i$i = $23;
   } else {
    $24 = HEAP32[$16>>2]|0;
    $$0$i$i1$i$i = $24;
   }
   $25 = ($$0$i$i1$i$i|0)==(-1);
   if ($25) {
    HEAP32[$__e>>2] = 0;
    label = 14;
    break;
   } else {
    if ($148) {
     $149 = $13;
     label = 17;
     break;
    } else {
     label = 16;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 14) {
  if ($148) {
   label = 16;
  } else {
   $149 = 0;
   label = 17;
  }
 }
 L22: do {
  if ((label|0) == 16) {
   $26 = HEAP32[$__err>>2]|0;
   $27 = $26 | 6;
   HEAP32[$__err>>2] = $27;
   $$2 = 0;
  }
  else if ((label|0) == 17) {
   $28 = HEAP32[$__b>>2]|0;
   $29 = ((($28)) + 12|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = ((($28)) + 16|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = ($30|0)==($32|0);
   if ($33) {
    $34 = HEAP32[$28>>2]|0;
    $35 = ((($34)) + 36|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = (FUNCTION_TABLE_ii[$36 & 8191]($28)|0);
    $$0$i$i = $37;
   } else {
    $38 = HEAP32[$30>>2]|0;
    $$0$i$i = $38;
   }
   $39 = HEAP32[$__ct>>2]|0;
   $40 = ((($39)) + 12|0);
   $41 = HEAP32[$40>>2]|0;
   $42 = (FUNCTION_TABLE_iiii[$41 & 8191]($__ct,2048,$$0$i$i)|0);
   if (!($42)) {
    $43 = HEAP32[$__err>>2]|0;
    $44 = $43 | 4;
    HEAP32[$__err>>2] = $44;
    $$2 = 0;
    break;
   }
   $45 = HEAP32[$__ct>>2]|0;
   $46 = ((($45)) + 52|0);
   $47 = HEAP32[$46>>2]|0;
   $48 = (FUNCTION_TABLE_iiii[$47 & 8191]($__ct,$$0$i$i,0)|0);
   $49 = $48 << 24 >> 24;
   $50 = HEAP32[$__b>>2]|0;
   $51 = ((($50)) + 12|0);
   $52 = HEAP32[$51>>2]|0;
   $53 = ((($50)) + 16|0);
   $54 = HEAP32[$53>>2]|0;
   $55 = ($52|0)==($54|0);
   if ($55) {
    $56 = HEAP32[$50>>2]|0;
    $57 = ((($56)) + 40|0);
    $58 = HEAP32[$57>>2]|0;
    (FUNCTION_TABLE_ii[$58 & 8191]($50)|0);
    $$0$in = $__n;$$pre$i3$i$i10 = $149;$74 = $149;$__r$0$in = $49;
   } else {
    $59 = ((($52)) + 4|0);
    HEAP32[$51>>2] = $59;
    $$0$in = $__n;$$pre$i3$i$i10 = $149;$74 = $149;$__r$0$in = $49;
   }
   while(1) {
    $__r$0 = (($__r$0$in) + -48)|0;
    $$0 = (($$0$in) + -1)|0;
    $60 = HEAP32[$__b>>2]|0;
    $61 = ($60|0)==(0|0);
    do {
     if ($61) {
      $86 = 1;
     } else {
      $62 = ((($60)) + 12|0);
      $63 = HEAP32[$62>>2]|0;
      $64 = ((($60)) + 16|0);
      $65 = HEAP32[$64>>2]|0;
      $66 = ($63|0)==($65|0);
      if ($66) {
       $67 = HEAP32[$60>>2]|0;
       $68 = ((($67)) + 36|0);
       $69 = HEAP32[$68>>2]|0;
       $70 = (FUNCTION_TABLE_ii[$69 & 8191]($60)|0);
       $$0$i$i$i$i2 = $70;
      } else {
       $71 = HEAP32[$63>>2]|0;
       $$0$i$i$i$i2 = $71;
      }
      $72 = ($$0$i$i$i$i2|0)==(-1);
      if ($72) {
       HEAP32[$__b>>2] = 0;
       $86 = 1;
       break;
      } else {
       $$pre$i$i$i4 = HEAP32[$__b>>2]|0;
       $phitmp$i$i$i5 = ($$pre$i$i$i4|0)==(0|0);
       $86 = $phitmp$i$i$i5;
       break;
      }
     }
    } while(0);
    $73 = ($74|0)==(0|0);
    do {
     if ($73) {
      $$pre$i3$i$i25 = $$pre$i3$i$i10;$150 = 0;$87 = 1;
     } else {
      $75 = ((($74)) + 12|0);
      $76 = HEAP32[$75>>2]|0;
      $77 = ((($74)) + 16|0);
      $78 = HEAP32[$77>>2]|0;
      $79 = ($76|0)==($78|0);
      if ($79) {
       $80 = HEAP32[$74>>2]|0;
       $81 = ((($80)) + 36|0);
       $82 = HEAP32[$81>>2]|0;
       $83 = (FUNCTION_TABLE_ii[$82 & 8191]($74)|0);
       $$0$i$i1$i$i8 = $83;
      } else {
       $84 = HEAP32[$76>>2]|0;
       $$0$i$i1$i$i8 = $84;
      }
      $85 = ($$0$i$i1$i$i8|0)==(-1);
      if ($85) {
       HEAP32[$__e>>2] = 0;
       $$pre$i3$i$i25 = 0;$150 = 0;$87 = 1;
       break;
      } else {
       $phitmp$i4$i$i11 = ($$pre$i3$i$i10|0)==(0|0);
       $$pre$i3$i$i25 = $$pre$i3$i$i10;$150 = $$pre$i3$i$i10;$87 = $phitmp$i4$i$i11;
       break;
      }
     }
    } while(0);
    $tmp$i$i13 = $86 ^ $87;
    $88 = ($$0$in|0)>(1);
    $89 = $88 & $tmp$i$i13;
    $90 = HEAP32[$__b>>2]|0;
    if (!($89)) {
     $$lcssa = $90;$$pre$i3$i$i25$lcssa = $$pre$i3$i$i25;$__r$0$lcssa = $__r$0;
     break;
    }
    $91 = ((($90)) + 12|0);
    $92 = HEAP32[$91>>2]|0;
    $93 = ((($90)) + 16|0);
    $94 = HEAP32[$93>>2]|0;
    $95 = ($92|0)==($94|0);
    if ($95) {
     $96 = HEAP32[$90>>2]|0;
     $97 = ((($96)) + 36|0);
     $98 = HEAP32[$97>>2]|0;
     $99 = (FUNCTION_TABLE_ii[$98 & 8191]($90)|0);
     $$0$i$i14 = $99;
    } else {
     $100 = HEAP32[$92>>2]|0;
     $$0$i$i14 = $100;
    }
    $101 = HEAP32[$__ct>>2]|0;
    $102 = ((($101)) + 12|0);
    $103 = HEAP32[$102>>2]|0;
    $104 = (FUNCTION_TABLE_iiii[$103 & 8191]($__ct,2048,$$0$i$i14)|0);
    if (!($104)) {
     $$2 = $__r$0;
     break L22;
    }
    $105 = ($__r$0*10)|0;
    $106 = HEAP32[$__ct>>2]|0;
    $107 = ((($106)) + 52|0);
    $108 = HEAP32[$107>>2]|0;
    $109 = (FUNCTION_TABLE_iiii[$108 & 8191]($__ct,$$0$i$i14,0)|0);
    $110 = $109 << 24 >> 24;
    $111 = (($110) + ($105))|0;
    $112 = HEAP32[$__b>>2]|0;
    $113 = ((($112)) + 12|0);
    $114 = HEAP32[$113>>2]|0;
    $115 = ((($112)) + 16|0);
    $116 = HEAP32[$115>>2]|0;
    $117 = ($114|0)==($116|0);
    if ($117) {
     $118 = HEAP32[$112>>2]|0;
     $119 = ((($118)) + 40|0);
     $120 = HEAP32[$119>>2]|0;
     (FUNCTION_TABLE_ii[$120 & 8191]($112)|0);
     $$0$in = $$0;$$pre$i3$i$i10 = $$pre$i3$i$i25;$74 = $150;$__r$0$in = $111;
     continue;
    } else {
     $121 = ((($114)) + 4|0);
     HEAP32[$113>>2] = $121;
     $$0$in = $$0;$$pre$i3$i$i10 = $$pre$i3$i$i25;$74 = $150;$__r$0$in = $111;
     continue;
    }
   }
   $122 = ($$lcssa|0)==(0|0);
   do {
    if ($122) {
     $151 = 1;
    } else {
     $123 = ((($$lcssa)) + 12|0);
     $124 = HEAP32[$123>>2]|0;
     $125 = ((($$lcssa)) + 16|0);
     $126 = HEAP32[$125>>2]|0;
     $127 = ($124|0)==($126|0);
     if ($127) {
      $128 = HEAP32[$$lcssa>>2]|0;
      $129 = ((($128)) + 36|0);
      $130 = HEAP32[$129>>2]|0;
      $131 = (FUNCTION_TABLE_ii[$130 & 8191]($$lcssa)|0);
      $$0$i$i$i$i17 = $131;
     } else {
      $132 = HEAP32[$124>>2]|0;
      $$0$i$i$i$i17 = $132;
     }
     $133 = ($$0$i$i$i$i17|0)==(-1);
     if ($133) {
      HEAP32[$__b>>2] = 0;
      $151 = 1;
      break;
     } else {
      $$pre$i$i$i19 = HEAP32[$__b>>2]|0;
      $phitmp$i$i$i20 = ($$pre$i$i$i19|0)==(0|0);
      $151 = $phitmp$i$i$i20;
      break;
     }
    }
   } while(0);
   $134 = ($$pre$i3$i$i25$lcssa|0)==(0|0);
   do {
    if ($134) {
     label = 60;
    } else {
     $135 = ((($$pre$i3$i$i25$lcssa)) + 12|0);
     $136 = HEAP32[$135>>2]|0;
     $137 = ((($$pre$i3$i$i25$lcssa)) + 16|0);
     $138 = HEAP32[$137>>2]|0;
     $139 = ($136|0)==($138|0);
     if ($139) {
      $140 = HEAP32[$$pre$i3$i$i25$lcssa>>2]|0;
      $141 = ((($140)) + 36|0);
      $142 = HEAP32[$141>>2]|0;
      $143 = (FUNCTION_TABLE_ii[$142 & 8191]($$pre$i3$i$i25$lcssa)|0);
      $$0$i$i1$i$i23 = $143;
     } else {
      $144 = HEAP32[$136>>2]|0;
      $$0$i$i1$i$i23 = $144;
     }
     $145 = ($$0$i$i1$i$i23|0)==(-1);
     if ($145) {
      HEAP32[$__e>>2] = 0;
      label = 60;
      break;
     } else {
      if ($151) {
       $$2 = $__r$0$lcssa;
       break L22;
      } else {
       break;
      }
     }
    }
   } while(0);
   if ((label|0) == 60) {
    if (!($151)) {
     $$2 = $__r$0$lcssa;
     break;
    }
   }
   $146 = HEAP32[$__err>>2]|0;
   $147 = $146 | 2;
   HEAP32[$__err>>2] = $147;
   $$2 = $__r$0$lcssa;
  }
 } while(0);
 return ($$2|0);
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjP2tmcc($this,$__b,$__e,$__iob,$__err,$__tm,$__fmt,$0) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__tm = $__tm|0;
 $__fmt = $__fmt|0;
 $0 = $0|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $$byval_copy10 = 0, $$byval_copy11 = 0, $$byval_copy12 = 0, $$byval_copy13 = 0, $$byval_copy14 = 0, $$byval_copy15 = 0, $$byval_copy16 = 0, $$byval_copy17 = 0, $$byval_copy18 = 0, $$byval_copy19 = 0, $$byval_copy2 = 0, $$byval_copy20 = 0, $$byval_copy21 = 0, $$byval_copy22 = 0, $$byval_copy23 = 0, $$byval_copy24 = 0, $$byval_copy25 = 0, $$byval_copy26 = 0;
 var $$byval_copy27 = 0, $$byval_copy28 = 0, $$byval_copy29 = 0, $$byval_copy3 = 0, $$byval_copy30 = 0, $$byval_copy4 = 0, $$byval_copy5 = 0, $$byval_copy6 = 0, $$byval_copy7 = 0, $$byval_copy8 = 0, $$byval_copy9 = 0, $$sroa$07$0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy30 = sp + 248|0;
 $$byval_copy29 = sp + 244|0;
 $$byval_copy28 = sp + 240|0;
 $$byval_copy27 = sp + 236|0;
 $$byval_copy26 = sp + 232|0;
 $$byval_copy25 = sp + 228|0;
 $$byval_copy24 = sp + 224|0;
 $$byval_copy23 = sp + 220|0;
 $$byval_copy22 = sp + 216|0;
 $$byval_copy21 = sp + 212|0;
 $$byval_copy20 = sp + 208|0;
 $$byval_copy19 = sp + 204|0;
 $$byval_copy18 = sp + 200|0;
 $$byval_copy17 = sp + 196|0;
 $$byval_copy16 = sp + 192|0;
 $$byval_copy15 = sp + 188|0;
 $$byval_copy14 = sp + 184|0;
 $$byval_copy13 = sp + 180|0;
 $$byval_copy12 = sp + 176|0;
 $$byval_copy11 = sp + 172|0;
 $$byval_copy10 = sp + 168|0;
 $$byval_copy9 = sp + 164|0;
 $$byval_copy8 = sp + 160|0;
 $$byval_copy7 = sp + 156|0;
 $$byval_copy6 = sp + 152|0;
 $$byval_copy5 = sp + 148|0;
 $$byval_copy4 = sp + 144|0;
 $$byval_copy3 = sp + 140|0;
 $$byval_copy2 = sp + 136|0;
 $$byval_copy1 = sp + 132|0;
 $$byval_copy = sp + 128|0;
 $1 = sp + 124|0;
 $2 = sp + 120|0;
 $3 = sp + 116|0;
 $4 = sp + 112|0;
 $5 = sp + 108|0;
 $6 = sp + 104|0;
 $7 = sp + 100|0;
 $8 = sp + 96|0;
 $9 = sp + 92|0;
 $10 = sp + 88|0;
 $11 = sp + 84|0;
 $12 = sp + 80|0;
 $13 = sp + 76|0;
 $14 = sp + 72|0;
 $15 = sp + 68|0;
 $16 = sp + 64|0;
 $17 = sp + 60|0;
 $18 = sp + 56|0;
 $19 = sp + 52|0;
 $20 = sp + 48|0;
 $21 = sp + 44|0;
 $22 = sp + 40|0;
 $23 = sp + 36|0;
 $24 = sp + 32|0;
 $25 = sp + 28|0;
 $26 = sp + 24|0;
 $27 = sp + 20|0;
 $28 = sp + 16|0;
 $29 = sp + 12|0;
 $30 = sp + 8|0;
 $31 = sp + 4|0;
 $32 = sp;
 HEAP32[$__err>>2] = 0;
 $33 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$1>>2] = $33;
 __THREW__ = 0;
 $34 = (invoke_iii(3374,($1|0),(1207608|0))|0);
 $35 = __THREW__; __THREW__ = 0;
 $36 = $35&1;
 if ($36) {
  $38 = ___cxa_find_matching_catch_2()|0;
  $39 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($33)|0);
  ___resumeException($38|0);
  // unreachable;
 }
 (__ZNSt3__114__shared_count16__release_sharedEv($33)|0);
 $37 = $__fmt << 24 >> 24;
 do {
  switch ($37|0) {
  case 65: case 97:  {
   $40 = ((($__tm)) + 24|0);
   $41 = HEAP32[$__e>>2]|0;
   HEAP32[$2>>2] = $41;
   ;HEAP32[$$byval_copy>>2]=HEAP32[$2>>2]|0;
   __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__get_weekdaynameERiRS4_S4_RjRKNS_5ctypeIwEE($this,$40,$__b,$$byval_copy,$__err,$34);
   label = 28;
   break;
  }
  case 104: case 66: case 98:  {
   $42 = ((($__tm)) + 16|0);
   $43 = HEAP32[$__e>>2]|0;
   HEAP32[$3>>2] = $43;
   ;HEAP32[$$byval_copy1>>2]=HEAP32[$3>>2]|0;
   __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE15__get_monthnameERiRS4_S4_RjRKNS_5ctypeIwEE($this,$42,$__b,$$byval_copy1,$__err,$34);
   label = 28;
   break;
  }
  case 99:  {
   $44 = ((($this)) + 8|0);
   $45 = HEAP32[$44>>2]|0;
   $46 = ((($45)) + 12|0);
   $47 = HEAP32[$46>>2]|0;
   $48 = (FUNCTION_TABLE_ii[$47 & 8191]($44)|0);
   $49 = HEAP32[$__b>>2]|0;
   HEAP32[$4>>2] = $49;
   $50 = HEAP32[$__e>>2]|0;
   HEAP32[$5>>2] = $50;
   $51 = HEAP8[$48>>0]|0;
   $52 = $51 & 1;
   $53 = ($52<<24>>24)==(0);
   $54 = ((($48)) + 8|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = ((($48)) + 4|0);
   $57 = $53 ? $56 : $55;
   $58 = HEAP32[$56>>2]|0;
   $59 = $51&255;
   $60 = $59 >>> 1;
   $61 = $53 ? $60 : $58;
   $62 = (($57) + ($61<<2)|0);
   ;HEAP32[$$byval_copy2>>2]=HEAP32[$4>>2]|0;
   ;HEAP32[$$byval_copy3>>2]=HEAP32[$5>>2]|0;
   $63 = (__ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($this,$$byval_copy2,$$byval_copy3,$__iob,$__err,$__tm,$57,$62)|0);
   HEAP32[$__b>>2] = $63;
   label = 28;
   break;
  }
  case 101: case 100:  {
   $64 = ((($__tm)) + 12|0);
   $65 = HEAP32[$__e>>2]|0;
   HEAP32[$6>>2] = $65;
   ;HEAP32[$$byval_copy4>>2]=HEAP32[$6>>2]|0;
   __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE9__get_dayERiRS4_S4_RjRKNS_5ctypeIwEE($this,$64,$__b,$$byval_copy4,$__err,$34);
   label = 28;
   break;
  }
  case 68:  {
   $66 = HEAP32[$__b>>2]|0;
   HEAP32[$7>>2] = $66;
   $67 = HEAP32[$__e>>2]|0;
   HEAP32[$8>>2] = $67;
   ;HEAP32[$$byval_copy5>>2]=HEAP32[$7>>2]|0;
   ;HEAP32[$$byval_copy6>>2]=HEAP32[$8>>2]|0;
   $68 = (__ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($this,$$byval_copy5,$$byval_copy6,$__iob,$__err,$__tm,155448,(155480))|0);
   HEAP32[$__b>>2] = $68;
   label = 28;
   break;
  }
  case 70:  {
   $69 = HEAP32[$__b>>2]|0;
   HEAP32[$9>>2] = $69;
   $70 = HEAP32[$__e>>2]|0;
   HEAP32[$10>>2] = $70;
   ;HEAP32[$$byval_copy7>>2]=HEAP32[$9>>2]|0;
   ;HEAP32[$$byval_copy8>>2]=HEAP32[$10>>2]|0;
   $71 = (__ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($this,$$byval_copy7,$$byval_copy8,$__iob,$__err,$__tm,155480,(155512))|0);
   HEAP32[$__b>>2] = $71;
   label = 28;
   break;
  }
  case 72:  {
   $72 = ((($__tm)) + 8|0);
   $73 = HEAP32[$__e>>2]|0;
   HEAP32[$11>>2] = $73;
   ;HEAP32[$$byval_copy9>>2]=HEAP32[$11>>2]|0;
   __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE10__get_hourERiRS4_S4_RjRKNS_5ctypeIwEE($this,$72,$__b,$$byval_copy9,$__err,$34);
   label = 28;
   break;
  }
  case 73:  {
   $74 = ((($__tm)) + 8|0);
   $75 = HEAP32[$__e>>2]|0;
   HEAP32[$12>>2] = $75;
   ;HEAP32[$$byval_copy10>>2]=HEAP32[$12>>2]|0;
   __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE13__get_12_hourERiRS4_S4_RjRKNS_5ctypeIwEE($this,$74,$__b,$$byval_copy10,$__err,$34);
   label = 28;
   break;
  }
  case 106:  {
   $76 = ((($__tm)) + 28|0);
   $77 = HEAP32[$__e>>2]|0;
   HEAP32[$13>>2] = $77;
   ;HEAP32[$$byval_copy11>>2]=HEAP32[$13>>2]|0;
   __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE18__get_day_year_numERiRS4_S4_RjRKNS_5ctypeIwEE($this,$76,$__b,$$byval_copy11,$__err,$34);
   label = 28;
   break;
  }
  case 109:  {
   $78 = ((($__tm)) + 16|0);
   $79 = HEAP32[$__e>>2]|0;
   HEAP32[$14>>2] = $79;
   ;HEAP32[$$byval_copy12>>2]=HEAP32[$14>>2]|0;
   __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11__get_monthERiRS4_S4_RjRKNS_5ctypeIwEE($this,$78,$__b,$$byval_copy12,$__err,$34);
   label = 28;
   break;
  }
  case 77:  {
   $80 = ((($__tm)) + 4|0);
   $81 = HEAP32[$__e>>2]|0;
   HEAP32[$15>>2] = $81;
   ;HEAP32[$$byval_copy13>>2]=HEAP32[$15>>2]|0;
   __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE12__get_minuteERiRS4_S4_RjRKNS_5ctypeIwEE($this,$80,$__b,$$byval_copy13,$__err,$34);
   label = 28;
   break;
  }
  case 116: case 110:  {
   $82 = HEAP32[$__e>>2]|0;
   HEAP32[$16>>2] = $82;
   ;HEAP32[$$byval_copy14>>2]=HEAP32[$16>>2]|0;
   __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__get_white_spaceERS4_S4_RjRKNS_5ctypeIwEE($this,$__b,$$byval_copy14,$__err,$34);
   label = 28;
   break;
  }
  case 112:  {
   $83 = ((($__tm)) + 8|0);
   $84 = HEAP32[$__e>>2]|0;
   HEAP32[$17>>2] = $84;
   ;HEAP32[$$byval_copy15>>2]=HEAP32[$17>>2]|0;
   __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11__get_am_pmERiRS4_S4_RjRKNS_5ctypeIwEE($this,$83,$__b,$$byval_copy15,$__err,$34);
   label = 28;
   break;
  }
  case 114:  {
   $85 = HEAP32[$__b>>2]|0;
   HEAP32[$18>>2] = $85;
   $86 = HEAP32[$__e>>2]|0;
   HEAP32[$19>>2] = $86;
   ;HEAP32[$$byval_copy16>>2]=HEAP32[$18>>2]|0;
   ;HEAP32[$$byval_copy17>>2]=HEAP32[$19>>2]|0;
   $87 = (__ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($this,$$byval_copy16,$$byval_copy17,$__iob,$__err,$__tm,155512,(155556))|0);
   HEAP32[$__b>>2] = $87;
   label = 28;
   break;
  }
  case 82:  {
   $88 = HEAP32[$__b>>2]|0;
   HEAP32[$20>>2] = $88;
   $89 = HEAP32[$__e>>2]|0;
   HEAP32[$21>>2] = $89;
   ;HEAP32[$$byval_copy18>>2]=HEAP32[$20>>2]|0;
   ;HEAP32[$$byval_copy19>>2]=HEAP32[$21>>2]|0;
   $90 = (__ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($this,$$byval_copy18,$$byval_copy19,$__iob,$__err,$__tm,155556,(155576))|0);
   HEAP32[$__b>>2] = $90;
   label = 28;
   break;
  }
  case 83:  {
   $91 = HEAP32[$__e>>2]|0;
   HEAP32[$22>>2] = $91;
   ;HEAP32[$$byval_copy20>>2]=HEAP32[$22>>2]|0;
   __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE12__get_secondERiRS4_S4_RjRKNS_5ctypeIwEE($this,$__tm,$__b,$$byval_copy20,$__err,$34);
   label = 28;
   break;
  }
  case 84:  {
   $92 = HEAP32[$__b>>2]|0;
   HEAP32[$23>>2] = $92;
   $93 = HEAP32[$__e>>2]|0;
   HEAP32[$24>>2] = $93;
   ;HEAP32[$$byval_copy21>>2]=HEAP32[$23>>2]|0;
   ;HEAP32[$$byval_copy22>>2]=HEAP32[$24>>2]|0;
   $94 = (__ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($this,$$byval_copy21,$$byval_copy22,$__iob,$__err,$__tm,155576,(155608))|0);
   HEAP32[$__b>>2] = $94;
   label = 28;
   break;
  }
  case 119:  {
   $95 = ((($__tm)) + 24|0);
   $96 = HEAP32[$__e>>2]|0;
   HEAP32[$25>>2] = $96;
   ;HEAP32[$$byval_copy23>>2]=HEAP32[$25>>2]|0;
   __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE13__get_weekdayERiRS4_S4_RjRKNS_5ctypeIwEE($this,$95,$__b,$$byval_copy23,$__err,$34);
   label = 28;
   break;
  }
  case 120:  {
   $97 = HEAP32[$this>>2]|0;
   $98 = ((($97)) + 20|0);
   $99 = HEAP32[$98>>2]|0;
   $100 = HEAP32[$__b>>2]|0;
   HEAP32[$26>>2] = $100;
   $101 = HEAP32[$__e>>2]|0;
   HEAP32[$27>>2] = $101;
   ;HEAP32[$$byval_copy24>>2]=HEAP32[$26>>2]|0;
   ;HEAP32[$$byval_copy25>>2]=HEAP32[$27>>2]|0;
   $102 = (FUNCTION_TABLE_iiiiiii[$99 & 8191]($this,$$byval_copy24,$$byval_copy25,$__iob,$__err,$__tm)|0);
   $$sroa$07$0 = $102;
   break;
  }
  case 88:  {
   $103 = ((($this)) + 8|0);
   $104 = HEAP32[$103>>2]|0;
   $105 = ((($104)) + 24|0);
   $106 = HEAP32[$105>>2]|0;
   $107 = (FUNCTION_TABLE_ii[$106 & 8191]($103)|0);
   $108 = HEAP32[$__b>>2]|0;
   HEAP32[$28>>2] = $108;
   $109 = HEAP32[$__e>>2]|0;
   HEAP32[$29>>2] = $109;
   $110 = HEAP8[$107>>0]|0;
   $111 = $110 & 1;
   $112 = ($111<<24>>24)==(0);
   $113 = ((($107)) + 8|0);
   $114 = HEAP32[$113>>2]|0;
   $115 = ((($107)) + 4|0);
   $116 = $112 ? $115 : $114;
   $117 = HEAP32[$115>>2]|0;
   $118 = $110&255;
   $119 = $118 >>> 1;
   $120 = $112 ? $119 : $117;
   $121 = (($116) + ($120<<2)|0);
   ;HEAP32[$$byval_copy26>>2]=HEAP32[$28>>2]|0;
   ;HEAP32[$$byval_copy27>>2]=HEAP32[$29>>2]|0;
   $122 = (__ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($this,$$byval_copy26,$$byval_copy27,$__iob,$__err,$__tm,$116,$121)|0);
   HEAP32[$__b>>2] = $122;
   label = 28;
   break;
  }
  case 121:  {
   $123 = ((($__tm)) + 20|0);
   $124 = HEAP32[$__e>>2]|0;
   HEAP32[$30>>2] = $124;
   ;HEAP32[$$byval_copy28>>2]=HEAP32[$30>>2]|0;
   __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE10__get_yearERiRS4_S4_RjRKNS_5ctypeIwEE($this,$123,$__b,$$byval_copy28,$__err,$34);
   label = 28;
   break;
  }
  case 89:  {
   $125 = ((($__tm)) + 20|0);
   $126 = HEAP32[$__e>>2]|0;
   HEAP32[$31>>2] = $126;
   ;HEAP32[$$byval_copy29>>2]=HEAP32[$31>>2]|0;
   __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11__get_year4ERiRS4_S4_RjRKNS_5ctypeIwEE($this,$125,$__b,$$byval_copy29,$__err,$34);
   label = 28;
   break;
  }
  case 37:  {
   $127 = HEAP32[$__e>>2]|0;
   HEAP32[$32>>2] = $127;
   ;HEAP32[$$byval_copy30>>2]=HEAP32[$32>>2]|0;
   __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE13__get_percentERS4_S4_RjRKNS_5ctypeIwEE($this,$__b,$$byval_copy30,$__err,$34);
   label = 28;
   break;
  }
  default: {
   $128 = HEAP32[$__err>>2]|0;
   $129 = $128 | 4;
   HEAP32[$__err>>2] = $129;
   label = 28;
  }
  }
 } while(0);
 if ((label|0) == 28) {
  $130 = HEAP32[$__b>>2]|0;
  $$sroa$07$0 = $130;
 }
 STACKTOP = sp;return ($$sroa$07$0|0);
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE9__get_dayERiRS4_S4_RjRKNS_5ctypeIwEE($this,$__d,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__d = $__d|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $$off = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,2)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 $$off = (($2) + -1)|0;
 $6 = ($$off>>>0)<(31);
 $7 = $6 & $5;
 if ($7) {
  HEAP32[$__d>>2] = $2;
 } else {
  $8 = $3 | 4;
  HEAP32[$__err>>2] = $8;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE10__get_hourERiRS4_S4_RjRKNS_5ctypeIwEE($this,$__h,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__h = $__h|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,2)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 $6 = ($2|0)<(24);
 $or$cond = $6 & $5;
 if ($or$cond) {
  HEAP32[$__h>>2] = $2;
 } else {
  $7 = $3 | 4;
  HEAP32[$__err>>2] = $7;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE13__get_12_hourERiRS4_S4_RjRKNS_5ctypeIwEE($this,$__h,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__h = $__h|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $$off = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,2)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 $$off = (($2) + -1)|0;
 $6 = ($$off>>>0)<(12);
 $7 = $6 & $5;
 if ($7) {
  HEAP32[$__h>>2] = $2;
 } else {
  $8 = $3 | 4;
  HEAP32[$__err>>2] = $8;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE18__get_day_year_numERiRS4_S4_RjRKNS_5ctypeIwEE($this,$__d,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__d = $__d|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,3)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 $6 = ($2|0)<(366);
 $or$cond = $6 & $5;
 if ($or$cond) {
  HEAP32[$__d>>2] = $2;
 } else {
  $7 = $3 | 4;
  HEAP32[$__err>>2] = $7;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11__get_monthERiRS4_S4_RjRKNS_5ctypeIwEE($this,$__m,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__m = $__m|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,2)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 $6 = ($2|0)<(13);
 $or$cond = $6 & $5;
 if ($or$cond) {
  $7 = (($2) + -1)|0;
  HEAP32[$__m>>2] = $7;
 } else {
  $8 = $3 | 4;
  HEAP32[$__err>>2] = $8;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE12__get_minuteERiRS4_S4_RjRKNS_5ctypeIwEE($this,$__m,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__m = $__m|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,2)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 $6 = ($2|0)<(60);
 $or$cond = $6 & $5;
 if ($or$cond) {
  HEAP32[$__m>>2] = $2;
 } else {
  $7 = $3 | 4;
  HEAP32[$__err>>2] = $7;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__get_white_spaceERS4_S4_RjRKNS_5ctypeIwEE($this,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i1 = 0, $$0$i$i1$i$i = 0, $$0$i$i1$i$i7 = 0, $$pre$i$i$i = 0, $$pre$i$i$i3 = 0, $$pre$i3$i$i9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $9 = 0, $phitmp$i$i$i = 0, $phitmp$i$i$i4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 L1: while(1) {
  $0 = HEAP32[$__b>>2]|0;
  $1 = ($0|0)==(0|0);
  do {
   if ($1) {
    $78 = 1;
   } else {
    $2 = ((($0)) + 12|0);
    $3 = HEAP32[$2>>2]|0;
    $4 = ((($0)) + 16|0);
    $5 = HEAP32[$4>>2]|0;
    $6 = ($3|0)==($5|0);
    if ($6) {
     $7 = HEAP32[$0>>2]|0;
     $8 = ((($7)) + 36|0);
     $9 = HEAP32[$8>>2]|0;
     $10 = (FUNCTION_TABLE_ii[$9 & 8191]($0)|0);
     $$0$i$i$i$i = $10;
    } else {
     $11 = HEAP32[$3>>2]|0;
     $$0$i$i$i$i = $11;
    }
    $12 = ($$0$i$i$i$i|0)==(-1);
    if ($12) {
     HEAP32[$__b>>2] = 0;
     $78 = 1;
     break;
    } else {
     $$pre$i$i$i = HEAP32[$__b>>2]|0;
     $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
     $78 = $phitmp$i$i$i;
     break;
    }
   }
  } while(0);
  $13 = HEAP32[$__e>>2]|0;
  $14 = ($13|0)==(0|0);
  do {
   if ($14) {
    label = 15;
   } else {
    $15 = ((($13)) + 12|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ((($13)) + 16|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = ($16|0)==($18|0);
    if ($19) {
     $20 = HEAP32[$13>>2]|0;
     $21 = ((($20)) + 36|0);
     $22 = HEAP32[$21>>2]|0;
     $23 = (FUNCTION_TABLE_ii[$22 & 8191]($13)|0);
     $$0$i$i1$i$i = $23;
    } else {
     $24 = HEAP32[$16>>2]|0;
     $$0$i$i1$i$i = $24;
    }
    $25 = ($$0$i$i1$i$i|0)==(-1);
    if ($25) {
     HEAP32[$__e>>2] = 0;
     label = 15;
     break;
    } else {
     if ($78) {
      $79 = $13;
      break;
     } else {
      $$pre$i3$i$i9 = $13;
      break L1;
     }
    }
   }
  } while(0);
  if ((label|0) == 15) {
   label = 0;
   if ($78) {
    $$pre$i3$i$i9 = 0;
    break;
   } else {
    $79 = 0;
   }
  }
  $26 = HEAP32[$__b>>2]|0;
  $27 = ((($26)) + 12|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = ((($26)) + 16|0);
  $30 = HEAP32[$29>>2]|0;
  $31 = ($28|0)==($30|0);
  if ($31) {
   $32 = HEAP32[$26>>2]|0;
   $33 = ((($32)) + 36|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = (FUNCTION_TABLE_ii[$34 & 8191]($26)|0);
   $$0$i$i = $35;
  } else {
   $36 = HEAP32[$28>>2]|0;
   $$0$i$i = $36;
  }
  $37 = HEAP32[$__ct>>2]|0;
  $38 = ((($37)) + 12|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = (FUNCTION_TABLE_iiii[$39 & 8191]($__ct,8192,$$0$i$i)|0);
  if (!($40)) {
   $$pre$i3$i$i9 = $79;
   break;
  }
  $41 = HEAP32[$__b>>2]|0;
  $42 = ((($41)) + 12|0);
  $43 = HEAP32[$42>>2]|0;
  $44 = ((($41)) + 16|0);
  $45 = HEAP32[$44>>2]|0;
  $46 = ($43|0)==($45|0);
  if ($46) {
   $47 = HEAP32[$41>>2]|0;
   $48 = ((($47)) + 40|0);
   $49 = HEAP32[$48>>2]|0;
   (FUNCTION_TABLE_ii[$49 & 8191]($41)|0);
   continue;
  } else {
   $50 = ((($43)) + 4|0);
   HEAP32[$42>>2] = $50;
   continue;
  }
 }
 $51 = HEAP32[$__b>>2]|0;
 $52 = ($51|0)==(0|0);
 do {
  if ($52) {
   $80 = 1;
  } else {
   $53 = ((($51)) + 12|0);
   $54 = HEAP32[$53>>2]|0;
   $55 = ((($51)) + 16|0);
   $56 = HEAP32[$55>>2]|0;
   $57 = ($54|0)==($56|0);
   if ($57) {
    $58 = HEAP32[$51>>2]|0;
    $59 = ((($58)) + 36|0);
    $60 = HEAP32[$59>>2]|0;
    $61 = (FUNCTION_TABLE_ii[$60 & 8191]($51)|0);
    $$0$i$i$i$i1 = $61;
   } else {
    $62 = HEAP32[$54>>2]|0;
    $$0$i$i$i$i1 = $62;
   }
   $63 = ($$0$i$i$i$i1|0)==(-1);
   if ($63) {
    HEAP32[$__b>>2] = 0;
    $80 = 1;
    break;
   } else {
    $$pre$i$i$i3 = HEAP32[$__b>>2]|0;
    $phitmp$i$i$i4 = ($$pre$i$i$i3|0)==(0|0);
    $80 = $phitmp$i$i$i4;
    break;
   }
  }
 } while(0);
 $64 = ($$pre$i3$i$i9|0)==(0|0);
 do {
  if ($64) {
   label = 37;
  } else {
   $65 = ((($$pre$i3$i$i9)) + 12|0);
   $66 = HEAP32[$65>>2]|0;
   $67 = ((($$pre$i3$i$i9)) + 16|0);
   $68 = HEAP32[$67>>2]|0;
   $69 = ($66|0)==($68|0);
   if ($69) {
    $70 = HEAP32[$$pre$i3$i$i9>>2]|0;
    $71 = ((($70)) + 36|0);
    $72 = HEAP32[$71>>2]|0;
    $73 = (FUNCTION_TABLE_ii[$72 & 8191]($$pre$i3$i$i9)|0);
    $$0$i$i1$i$i7 = $73;
   } else {
    $74 = HEAP32[$66>>2]|0;
    $$0$i$i1$i$i7 = $74;
   }
   $75 = ($$0$i$i1$i$i7|0)==(-1);
   if ($75) {
    HEAP32[$__e>>2] = 0;
    label = 37;
    break;
   } else {
    if ($80) {
     break;
    } else {
     label = 39;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 37) {
  if ($80) {
   label = 39;
  }
 }
 if ((label|0) == 39) {
  $76 = HEAP32[$__err>>2]|0;
  $77 = $76 | 2;
  HEAP32[$__err>>2] = $77;
 }
 return;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11__get_am_pmERiRS4_S4_RjRKNS_5ctypeIwEE($this,$__h,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__h = $__h|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $or$cond = 0, $or$cond3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = ((($this)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (FUNCTION_TABLE_ii[$4 & 8191]($1)|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = $6 & 1;
 $8 = ($7<<24>>24)==(0);
 if ($8) {
  $11 = $6&255;
  $12 = $11 >>> 1;
  $24 = $12;
 } else {
  $9 = ((($5)) + 4|0);
  $10 = HEAP32[$9>>2]|0;
  $24 = $10;
 }
 $13 = ((($5)) + 12|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = $14 & 1;
 $16 = ($15<<24>>24)==(0);
 if ($16) {
  $19 = $14&255;
  $20 = $19 >>> 1;
  $22 = $20;
 } else {
  $17 = ((($5)) + 16|0);
  $18 = HEAP32[$17>>2]|0;
  $22 = $18;
 }
 $21 = (0 - ($22))|0;
 $23 = ($24|0)==($21|0);
 do {
  if ($23) {
   $25 = HEAP32[$__err>>2]|0;
   $26 = $25 | 4;
   HEAP32[$__err>>2] = $26;
  } else {
   $27 = HEAP32[$__e>>2]|0;
   HEAP32[$0>>2] = $27;
   $28 = ((($5)) + 24|0);
   ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
   $29 = (__ZNSt3__114__scan_keywordINS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEPKNS_12basic_stringIwS3_NS_9allocatorIwEEEENS_5ctypeIwEEEET0_RT_SE_SD_SD_RKT1_Rjb($__b,$$byval_copy,$5,$28,$__ct,$__err,0)|0);
   $30 = ($29|0)==($5|0);
   $31 = HEAP32[$__h>>2]|0;
   $32 = ($31|0)==(12);
   $or$cond = $30 & $32;
   if ($or$cond) {
    HEAP32[$__h>>2] = 0;
    break;
   }
   $33 = $29;
   $34 = $5;
   $35 = (($33) - ($34))|0;
   $36 = ($35|0)==(12);
   $37 = ($31|0)<(12);
   $or$cond3 = $37 & $36;
   if ($or$cond3) {
    $38 = (($31) + 12)|0;
    HEAP32[$__h>>2] = $38;
   }
  }
 } while(0);
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE12__get_secondERiRS4_S4_RjRKNS_5ctypeIwEE($this,$__s,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__s = $__s|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,2)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 $6 = ($2|0)<(61);
 $or$cond = $6 & $5;
 if ($or$cond) {
  HEAP32[$__s>>2] = $2;
 } else {
  $7 = $3 | 4;
  HEAP32[$__err>>2] = $7;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE13__get_weekdayERiRS4_S4_RjRKNS_5ctypeIwEE($this,$__w,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__w = $__w|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,1)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 $6 = ($2|0)<(7);
 $or$cond = $6 & $5;
 if ($or$cond) {
  HEAP32[$__w>>2] = $2;
 } else {
  $7 = $3 | 4;
  HEAP32[$__err>>2] = $7;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11__get_year4ERiRS4_S4_RjRKNS_5ctypeIwEE($this,$__y,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__y = $__y|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$byval_copy = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 4|0;
 $0 = sp;
 $1 = HEAP32[$__e>>2]|0;
 HEAP32[$0>>2] = $1;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
 $2 = (__ZNSt3__120__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($__b,$$byval_copy,$__err,$__ct,4)|0);
 $3 = HEAP32[$__err>>2]|0;
 $4 = $3 & 4;
 $5 = ($4|0)==(0);
 if ($5) {
  $6 = (($2) + -1900)|0;
  HEAP32[$__y>>2] = $6;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE13__get_percentERS4_S4_RjRKNS_5ctypeIwEE($this,$__b,$__e,$__err,$__ct) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__err = $__err|0;
 $__ct = $__ct|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i1 = 0, $$0$i$i1$i$i = 0, $$0$i$i1$i$i7 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i3 = 0, $$pre$i3$i$i915 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $9 = 0, $phitmp$i$i$i = 0, $phitmp$i$i$i4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$__b>>2]|0;
 $1 = ($0|0)==(0|0);
 do {
  if ($1) {
   $83 = 1;
  } else {
   $2 = ((($0)) + 12|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = ((($0)) + 16|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = ($3|0)==($5|0);
   if ($6) {
    $7 = HEAP32[$0>>2]|0;
    $8 = ((($7)) + 36|0);
    $9 = HEAP32[$8>>2]|0;
    $10 = (FUNCTION_TABLE_ii[$9 & 8191]($0)|0);
    $$0$i$i$i$i = $10;
   } else {
    $11 = HEAP32[$3>>2]|0;
    $$0$i$i$i$i = $11;
   }
   $12 = ($$0$i$i$i$i|0)==(-1);
   if ($12) {
    HEAP32[$__b>>2] = 0;
    $83 = 1;
    break;
   } else {
    $$pre$i$i$i = HEAP32[$__b>>2]|0;
    $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
    $83 = $phitmp$i$i$i;
    break;
   }
  }
 } while(0);
 $13 = HEAP32[$__e>>2]|0;
 $14 = ($13|0)==(0|0);
 do {
  if ($14) {
   label = 14;
  } else {
   $15 = ((($13)) + 12|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = ((($13)) + 16|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = ($16|0)==($18|0);
   if ($19) {
    $20 = HEAP32[$13>>2]|0;
    $21 = ((($20)) + 36|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = (FUNCTION_TABLE_ii[$22 & 8191]($13)|0);
    $$0$i$i1$i$i = $23;
   } else {
    $24 = HEAP32[$16>>2]|0;
    $$0$i$i1$i$i = $24;
   }
   $25 = ($$0$i$i1$i$i|0)==(-1);
   if ($25) {
    HEAP32[$__e>>2] = 0;
    label = 14;
    break;
   } else {
    if ($83) {
     $$pre$i3$i$i915 = $13;
     label = 17;
     break;
    } else {
     label = 16;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 14) {
  if ($83) {
   label = 16;
  } else {
   $$pre$i3$i$i915 = 0;
   label = 17;
  }
 }
 L22: do {
  if ((label|0) == 16) {
   $26 = HEAP32[$__err>>2]|0;
   $27 = $26 | 6;
   HEAP32[$__err>>2] = $27;
  }
  else if ((label|0) == 17) {
   $28 = HEAP32[$__b>>2]|0;
   $29 = ((($28)) + 12|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = ((($28)) + 16|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = ($30|0)==($32|0);
   if ($33) {
    $34 = HEAP32[$28>>2]|0;
    $35 = ((($34)) + 36|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = (FUNCTION_TABLE_ii[$36 & 8191]($28)|0);
    $$0$i$i = $37;
   } else {
    $38 = HEAP32[$30>>2]|0;
    $$0$i$i = $38;
   }
   $39 = HEAP32[$__ct>>2]|0;
   $40 = ((($39)) + 52|0);
   $41 = HEAP32[$40>>2]|0;
   $42 = (FUNCTION_TABLE_iiii[$41 & 8191]($__ct,$$0$i$i,0)|0);
   $43 = ($42<<24>>24)==(37);
   if (!($43)) {
    $44 = HEAP32[$__err>>2]|0;
    $45 = $44 | 4;
    HEAP32[$__err>>2] = $45;
    break;
   }
   $46 = HEAP32[$__b>>2]|0;
   $47 = ((($46)) + 12|0);
   $48 = HEAP32[$47>>2]|0;
   $49 = ((($46)) + 16|0);
   $50 = HEAP32[$49>>2]|0;
   $51 = ($48|0)==($50|0);
   if ($51) {
    $53 = HEAP32[$46>>2]|0;
    $54 = ((($53)) + 40|0);
    $55 = HEAP32[$54>>2]|0;
    (FUNCTION_TABLE_ii[$55 & 8191]($46)|0);
    $$pre = HEAP32[$__b>>2]|0;
    $56 = ($$pre|0)==(0|0);
    if ($56) {
     $84 = 1;
    } else {
     $58 = $$pre;
     label = 25;
    }
   } else {
    $52 = ((($48)) + 4|0);
    HEAP32[$47>>2] = $52;
    $58 = $46;
    label = 25;
   }
   do {
    if ((label|0) == 25) {
     $57 = ((($58)) + 12|0);
     $59 = HEAP32[$57>>2]|0;
     $60 = ((($58)) + 16|0);
     $61 = HEAP32[$60>>2]|0;
     $62 = ($59|0)==($61|0);
     if ($62) {
      $63 = HEAP32[$58>>2]|0;
      $64 = ((($63)) + 36|0);
      $65 = HEAP32[$64>>2]|0;
      $66 = (FUNCTION_TABLE_ii[$65 & 8191]($58)|0);
      $$0$i$i$i$i1 = $66;
     } else {
      $67 = HEAP32[$59>>2]|0;
      $$0$i$i$i$i1 = $67;
     }
     $68 = ($$0$i$i$i$i1|0)==(-1);
     if ($68) {
      HEAP32[$__b>>2] = 0;
      $84 = 1;
      break;
     } else {
      $$pre$i$i$i3 = HEAP32[$__b>>2]|0;
      $phitmp$i$i$i4 = ($$pre$i$i$i3|0)==(0|0);
      $84 = $phitmp$i$i$i4;
      break;
     }
    }
   } while(0);
   $69 = ($$pre$i3$i$i915|0)==(0|0);
   do {
    if ($69) {
     label = 37;
    } else {
     $70 = ((($$pre$i3$i$i915)) + 12|0);
     $71 = HEAP32[$70>>2]|0;
     $72 = ((($$pre$i3$i$i915)) + 16|0);
     $73 = HEAP32[$72>>2]|0;
     $74 = ($71|0)==($73|0);
     if ($74) {
      $75 = HEAP32[$$pre$i3$i$i915>>2]|0;
      $76 = ((($75)) + 36|0);
      $77 = HEAP32[$76>>2]|0;
      $78 = (FUNCTION_TABLE_ii[$77 & 8191]($$pre$i3$i$i915)|0);
      $$0$i$i1$i$i7 = $78;
     } else {
      $79 = HEAP32[$71>>2]|0;
      $$0$i$i1$i$i7 = $79;
     }
     $80 = ($$0$i$i1$i$i7|0)==(-1);
     if ($80) {
      HEAP32[$__e>>2] = 0;
      label = 37;
      break;
     } else {
      if ($84) {
       break L22;
      } else {
       break;
      }
     }
    }
   } while(0);
   if ((label|0) == 37) {
    if (!($84)) {
     break;
    }
   }
   $81 = HEAP32[$__err>>2]|0;
   $82 = $81 | 2;
   HEAP32[$__err>>2] = $82;
  }
 } while(0);
 return;
}
function __ZNKSt3__120__time_get_c_storageIwE7__weeksEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__i$0$i$i$i = 0, $exitcond$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[625040]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((625040|0))|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    $4 = HEAP8[625048]|0;
    $5 = ($4<<24>>24)==(0);
    if ($5) {
     $6 = (___cxa_guard_acquire((625048|0))|0);
     $7 = ($6|0)==(0);
     if (!($7)) {
      $9 = 1208460;
      while(1) {
       $__i$0$i$i$i = 0;
       while(1) {
        $exitcond$i$i$i = ($__i$0$i$i$i|0)==(3);
        if ($exitcond$i$i$i) {
         break;
        }
        $8 = (($9) + ($__i$0$i$i$i<<2)|0);
        HEAP32[$8>>2] = 0;
        $10 = (($__i$0$i$i$i) + 1)|0;
        $__i$0$i$i$i = $10;
       }
       $11 = ((($9)) + 12|0);
       $12 = ($11|0)==((1208628)|0);
       if ($12) {
        break;
       } else {
        $9 = $11;
       }
      }
      (___cxa_atexit((4890|0),(0|0),(___dso_handle|0))|0);
      ___cxa_guard_release((625048|0));
     }
    }
    __THREW__ = 0;
    (invoke_iii(4891,(1208460|0),(155608|0))|0);
    $13 = __THREW__; __THREW__ = 0;
    $14 = $13&1;
    if (!($14)) {
     __THREW__ = 0;
     (invoke_iii(4891,((1208472)|0),(155636|0))|0);
     $15 = __THREW__; __THREW__ = 0;
     $16 = $15&1;
     if (!($16)) {
      __THREW__ = 0;
      (invoke_iii(4891,((1208484)|0),(155664|0))|0);
      $17 = __THREW__; __THREW__ = 0;
      $18 = $17&1;
      if (!($18)) {
       __THREW__ = 0;
       (invoke_iii(4891,((1208496)|0),(155696|0))|0);
       $19 = __THREW__; __THREW__ = 0;
       $20 = $19&1;
       if (!($20)) {
        __THREW__ = 0;
        (invoke_iii(4891,((1208508)|0),(155736|0))|0);
        $21 = __THREW__; __THREW__ = 0;
        $22 = $21&1;
        if (!($22)) {
         __THREW__ = 0;
         (invoke_iii(4891,((1208520)|0),(155772|0))|0);
         $23 = __THREW__; __THREW__ = 0;
         $24 = $23&1;
         if (!($24)) {
          __THREW__ = 0;
          (invoke_iii(4891,((1208532)|0),(155800|0))|0);
          $25 = __THREW__; __THREW__ = 0;
          $26 = $25&1;
          if (!($26)) {
           __THREW__ = 0;
           (invoke_iii(4891,((1208544)|0),(155836|0))|0);
           $27 = __THREW__; __THREW__ = 0;
           $28 = $27&1;
           if (!($28)) {
            __THREW__ = 0;
            (invoke_iii(4891,((1208556)|0),(155852|0))|0);
            $29 = __THREW__; __THREW__ = 0;
            $30 = $29&1;
            if (!($30)) {
             __THREW__ = 0;
             (invoke_iii(4891,((1208568)|0),(155868|0))|0);
             $31 = __THREW__; __THREW__ = 0;
             $32 = $31&1;
             if (!($32)) {
              __THREW__ = 0;
              (invoke_iii(4891,((1208580)|0),(155884|0))|0);
              $33 = __THREW__; __THREW__ = 0;
              $34 = $33&1;
              if (!($34)) {
               __THREW__ = 0;
               (invoke_iii(4891,((1208592)|0),(155900|0))|0);
               $35 = __THREW__; __THREW__ = 0;
               $36 = $35&1;
               if (!($36)) {
                __THREW__ = 0;
                (invoke_iii(4891,((1208604)|0),(155916|0))|0);
                $37 = __THREW__; __THREW__ = 0;
                $38 = $37&1;
                if (!($38)) {
                 __THREW__ = 0;
                 (invoke_iii(4891,((1208616)|0),(155932|0))|0);
                 $39 = __THREW__; __THREW__ = 0;
                 $40 = $39&1;
                 if (!($40)) {
                  HEAP32[302157] = 1208460;
                  ___cxa_guard_release((625040|0));
                  break;
                 }
                }
               }
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    }
    $42 = ___cxa_find_matching_catch_2()|0;
    $43 = tempRet0;
    ___cxa_guard_abort((625040|0));
    ___resumeException($42|0);
    // unreachable;
   }
  }
 } while(0);
 $41 = HEAP32[302157]|0;
 return ($41|0);
}
function ___cxx_global_array_dtor_54($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208616));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208604));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208592));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208580));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208568));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208556));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208544));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208532));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208520));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208508));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208496));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208484));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208472));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev(1208460);
 return;
}
function __ZNKSt3__120__time_get_c_storageIwE8__monthsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $7 = 0, $8 = 0, $9 = 0, $__i$0$i$i$i = 0, $exitcond$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[625056]|0;
 $1 = ($0<<24>>24)==(0);
 L1: do {
  if ($1) {
   $2 = (___cxa_guard_acquire((625056|0))|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    $4 = HEAP8[625064]|0;
    $5 = ($4<<24>>24)==(0);
    if ($5) {
     $6 = (___cxa_guard_acquire((625064|0))|0);
     $7 = ($6|0)==(0);
     if (!($7)) {
      $9 = 1208632;
      while(1) {
       $__i$0$i$i$i = 0;
       while(1) {
        $exitcond$i$i$i = ($__i$0$i$i$i|0)==(3);
        if ($exitcond$i$i$i) {
         break;
        }
        $8 = (($9) + ($__i$0$i$i$i<<2)|0);
        HEAP32[$8>>2] = 0;
        $10 = (($__i$0$i$i$i) + 1)|0;
        $__i$0$i$i$i = $10;
       }
       $11 = ((($9)) + 12|0);
       $12 = ($11|0)==((1208920)|0);
       if ($12) {
        break;
       } else {
        $9 = $11;
       }
      }
      (___cxa_atexit((4892|0),(0|0),(___dso_handle|0))|0);
      ___cxa_guard_release((625064|0));
     }
    }
    __THREW__ = 0;
    (invoke_iii(4891,(1208632|0),(155948|0))|0);
    $13 = __THREW__; __THREW__ = 0;
    $14 = $13&1;
    do {
     if (!($14)) {
      __THREW__ = 0;
      (invoke_iii(4891,((1208644)|0),(155980|0))|0);
      $15 = __THREW__; __THREW__ = 0;
      $16 = $15&1;
      if (!($16)) {
       __THREW__ = 0;
       (invoke_iii(4891,((1208656)|0),(156016|0))|0);
       $17 = __THREW__; __THREW__ = 0;
       $18 = $17&1;
       if (!($18)) {
        __THREW__ = 0;
        (invoke_iii(4891,((1208668)|0),(156040|0))|0);
        $19 = __THREW__; __THREW__ = 0;
        $20 = $19&1;
        if (!($20)) {
         __THREW__ = 0;
         (invoke_iii(4891,((1208680)|0),(156064|0))|0);
         $21 = __THREW__; __THREW__ = 0;
         $22 = $21&1;
         if (!($22)) {
          __THREW__ = 0;
          (invoke_iii(4891,((1208692)|0),(156080|0))|0);
          $23 = __THREW__; __THREW__ = 0;
          $24 = $23&1;
          if (!($24)) {
           __THREW__ = 0;
           (invoke_iii(4891,((1208704)|0),(156100|0))|0);
           $25 = __THREW__; __THREW__ = 0;
           $26 = $25&1;
           if (!($26)) {
            __THREW__ = 0;
            (invoke_iii(4891,((1208716)|0),(156120|0))|0);
            $27 = __THREW__; __THREW__ = 0;
            $28 = $27&1;
            if (!($28)) {
             __THREW__ = 0;
             (invoke_iii(4891,((1208728)|0),(156148|0))|0);
             $29 = __THREW__; __THREW__ = 0;
             $30 = $29&1;
             if (!($30)) {
              __THREW__ = 0;
              (invoke_iii(4891,((1208740)|0),(156188|0))|0);
              $31 = __THREW__; __THREW__ = 0;
              $32 = $31&1;
              if (!($32)) {
               __THREW__ = 0;
               (invoke_iii(4891,((1208752)|0),(156220|0))|0);
               $33 = __THREW__; __THREW__ = 0;
               $34 = $33&1;
               if (!($34)) {
                __THREW__ = 0;
                (invoke_iii(4891,((1208764)|0),(156256|0))|0);
                $35 = __THREW__; __THREW__ = 0;
                $36 = $35&1;
                if (!($36)) {
                 __THREW__ = 0;
                 (invoke_iii(4891,((1208776)|0),(156292|0))|0);
                 $37 = __THREW__; __THREW__ = 0;
                 $38 = $37&1;
                 if (!($38)) {
                  __THREW__ = 0;
                  (invoke_iii(4891,((1208788)|0),(156308|0))|0);
                  $39 = __THREW__; __THREW__ = 0;
                  $40 = $39&1;
                  if (!($40)) {
                   __THREW__ = 0;
                   (invoke_iii(4891,((1208800)|0),(156324|0))|0);
                   $41 = __THREW__; __THREW__ = 0;
                   $42 = $41&1;
                   if (!($42)) {
                    __THREW__ = 0;
                    (invoke_iii(4891,((1208812)|0),(156340|0))|0);
                    $43 = __THREW__; __THREW__ = 0;
                    $44 = $43&1;
                    if ($44) {
                     break;
                    }
                    __THREW__ = 0;
                    (invoke_iii(4891,((1208824)|0),(156064|0))|0);
                    $45 = __THREW__; __THREW__ = 0;
                    $46 = $45&1;
                    if ($46) {
                     break;
                    }
                    __THREW__ = 0;
                    (invoke_iii(4891,((1208836)|0),(156356|0))|0);
                    $47 = __THREW__; __THREW__ = 0;
                    $48 = $47&1;
                    if ($48) {
                     break;
                    }
                    __THREW__ = 0;
                    (invoke_iii(4891,((1208848)|0),(156372|0))|0);
                    $49 = __THREW__; __THREW__ = 0;
                    $50 = $49&1;
                    if ($50) {
                     break;
                    }
                    __THREW__ = 0;
                    (invoke_iii(4891,((1208860)|0),(156388|0))|0);
                    $51 = __THREW__; __THREW__ = 0;
                    $52 = $51&1;
                    if ($52) {
                     break;
                    }
                    __THREW__ = 0;
                    (invoke_iii(4891,((1208872)|0),(156404|0))|0);
                    $53 = __THREW__; __THREW__ = 0;
                    $54 = $53&1;
                    if ($54) {
                     break;
                    }
                    __THREW__ = 0;
                    (invoke_iii(4891,((1208884)|0),(156420|0))|0);
                    $55 = __THREW__; __THREW__ = 0;
                    $56 = $55&1;
                    if ($56) {
                     break;
                    }
                    __THREW__ = 0;
                    (invoke_iii(4891,((1208896)|0),(156436|0))|0);
                    $57 = __THREW__; __THREW__ = 0;
                    $58 = $57&1;
                    if ($58) {
                     break;
                    }
                    __THREW__ = 0;
                    (invoke_iii(4891,((1208908)|0),(156452|0))|0);
                    $59 = __THREW__; __THREW__ = 0;
                    $60 = $59&1;
                    if ($60) {
                     break;
                    }
                    HEAP32[302230] = 1208632;
                    ___cxa_guard_release((625056|0));
                    break L1;
                   }
                  }
                 }
                }
               }
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    } while(0);
    $62 = ___cxa_find_matching_catch_2()|0;
    $63 = tempRet0;
    ___cxa_guard_abort((625056|0));
    ___resumeException($62|0);
    // unreachable;
   }
  }
 } while(0);
 $61 = HEAP32[302230]|0;
 return ($61|0);
}
function ___cxx_global_array_dtor_93($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208908));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208896));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208884));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208872));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208860));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208848));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208836));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208824));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208812));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208800));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208788));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208776));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208764));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208752));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208740));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208728));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208716));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208704));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208692));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208680));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208668));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208656));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208644));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev(1208632);
 return;
}
function __ZNKSt3__120__time_get_c_storageIwE7__am_pmEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $__i$0$i$i$i = 0, $exitcond$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[625072]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((625072|0))|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    $4 = HEAP8[625080]|0;
    $5 = ($4<<24>>24)==(0);
    if ($5) {
     $6 = (___cxa_guard_acquire((625080|0))|0);
     $7 = ($6|0)==(0);
     if (!($7)) {
      $9 = 1208924;
      while(1) {
       $__i$0$i$i$i = 0;
       while(1) {
        $exitcond$i$i$i = ($__i$0$i$i$i|0)==(3);
        if ($exitcond$i$i$i) {
         break;
        }
        $8 = (($9) + ($__i$0$i$i$i<<2)|0);
        HEAP32[$8>>2] = 0;
        $10 = (($__i$0$i$i$i) + 1)|0;
        $__i$0$i$i$i = $10;
       }
       $11 = ((($9)) + 12|0);
       $12 = ($11|0)==((1209212)|0);
       if ($12) {
        break;
       } else {
        $9 = $11;
       }
      }
      (___cxa_atexit((4893|0),(0|0),(___dso_handle|0))|0);
      ___cxa_guard_release((625080|0));
     }
    }
    __THREW__ = 0;
    (invoke_iii(4891,(1208924|0),(156468|0))|0);
    $13 = __THREW__; __THREW__ = 0;
    $14 = $13&1;
    if (!($14)) {
     __THREW__ = 0;
     (invoke_iii(4891,((1208936)|0),(156480|0))|0);
     $15 = __THREW__; __THREW__ = 0;
     $16 = $15&1;
     if (!($16)) {
      HEAP32[302303] = 1208924;
      ___cxa_guard_release((625072|0));
      break;
     }
    }
    $18 = ___cxa_find_matching_catch_2()|0;
    $19 = tempRet0;
    ___cxa_guard_abort((625072|0));
    ___resumeException($18|0);
    // unreachable;
   }
  }
 } while(0);
 $17 = HEAP32[302303]|0;
 return ($17|0);
}
function ___cxx_global_array_dtor_120($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1209200));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1209188));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1209176));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1209164));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1209152));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1209140));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1209128));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1209116));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1209104));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1209092));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1209080));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1209068));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1209056));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1209044));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1209032));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1209020));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1209008));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208996));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208984));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208972));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208960));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208948));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev((1208936));
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev(1208924);
 return;
}
function __ZNKSt3__120__time_get_c_storageIwE3__cEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[625088]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((625088|0))|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    __THREW__ = 0;
    $4 = (invoke_ii(4894,(156492|0))|0);
    $5 = __THREW__; __THREW__ = 0;
    $6 = $5&1;
    if (!($6)) {
     __THREW__ = 0;
     invoke_viii(4895,(1209216|0),(156492|0),($4|0));
     $7 = __THREW__; __THREW__ = 0;
     $8 = $7&1;
     if (!($8)) {
      (___cxa_atexit((4896|0),(1209216|0),(___dso_handle|0))|0);
      ___cxa_guard_release((625088|0));
      break;
     }
    }
    $9 = ___cxa_find_matching_catch_2()|0;
    $10 = tempRet0;
    ___cxa_guard_abort((625088|0));
    ___resumeException($9|0);
    // unreachable;
   }
  }
 } while(0);
 return (1209216|0);
}
function __ZNKSt3__120__time_get_c_storageIwE3__rEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[625096]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((625096|0))|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    __THREW__ = 0;
    $4 = (invoke_ii(4894,(156576|0))|0);
    $5 = __THREW__; __THREW__ = 0;
    $6 = $5&1;
    if (!($6)) {
     __THREW__ = 0;
     invoke_viii(4895,(1209228|0),(156576|0),($4|0));
     $7 = __THREW__; __THREW__ = 0;
     $8 = $7&1;
     if (!($8)) {
      (___cxa_atexit((4896|0),(1209228|0),(___dso_handle|0))|0);
      ___cxa_guard_release((625096|0));
      break;
     }
    }
    $9 = ___cxa_find_matching_catch_2()|0;
    $10 = tempRet0;
    ___cxa_guard_abort((625096|0));
    ___resumeException($9|0);
    // unreachable;
   }
  }
 } while(0);
 return (1209228|0);
}
function __ZNKSt3__120__time_get_c_storageIwE3__xEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[625104]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((625104|0))|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    __THREW__ = 0;
    $4 = (invoke_ii(4894,(156624|0))|0);
    $5 = __THREW__; __THREW__ = 0;
    $6 = $5&1;
    if (!($6)) {
     __THREW__ = 0;
     invoke_viii(4895,(1209240|0),(156624|0),($4|0));
     $7 = __THREW__; __THREW__ = 0;
     $8 = $7&1;
     if (!($8)) {
      (___cxa_atexit((4896|0),(1209240|0),(___dso_handle|0))|0);
      ___cxa_guard_release((625104|0));
      break;
     }
    }
    $9 = ___cxa_find_matching_catch_2()|0;
    $10 = tempRet0;
    ___cxa_guard_abort((625104|0));
    ___resumeException($9|0);
    // unreachable;
   }
  }
 } while(0);
 return (1209240|0);
}
function __ZNKSt3__120__time_get_c_storageIwE3__XEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[625112]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((625112|0))|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    __THREW__ = 0;
    $4 = (invoke_ii(4894,(156660|0))|0);
    $5 = __THREW__; __THREW__ = 0;
    $6 = $5&1;
    if (!($6)) {
     __THREW__ = 0;
     invoke_viii(4895,(1209252|0),(156660|0),($4|0));
     $7 = __THREW__; __THREW__ = 0;
     $8 = $7&1;
     if (!($8)) {
      (___cxa_atexit((4896|0),(1209252|0),(___dso_handle|0))|0);
      ___cxa_guard_release((625112|0));
      break;
     }
    }
    $9 = ___cxa_find_matching_catch_2()|0;
    $10 = tempRet0;
    ___cxa_guard_abort((625112|0));
    ___resumeException($9|0);
    // unreachable;
   }
  }
 } while(0);
 return (1209252|0);
}
function __ZNSt3__18time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 __ZNSt3__110__time_putD2Ev($0);
 return;
}
function __ZNSt3__110__time_putD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 __THREW__ = 0;
 $1 = (invoke_i(4874)|0);
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 do {
  if (!($3)) {
   $4 = ($0|0)==($1|0);
   if (!($4)) {
    $5 = HEAP32[$this>>2]|0;
    __THREW__ = 0;
    invoke_vi(4897,($5|0));
    $6 = __THREW__; __THREW__ = 0;
    $7 = $6&1;
    if ($7) {
     break;
    }
   }
   return;
  }
 } while(0);
 $8 = ___cxa_find_matching_catch_3(0|0)|0;
 $9 = tempRet0;
 ___clang_call_terminate($8);
 // unreachable;
}
function __ZNSt3__18time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 __ZNSt3__110__time_putD2Ev($0);
 __ZdlPv($this);
 return;
}
function __ZNKSt3__18time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcPK2tmcc($this,$__s,$0,$1,$__tm,$__fmt,$__mod) {
 $this = $this|0;
 $__s = $__s|0;
 $0 = $0|0;
 $1 = $1|0;
 $__tm = $__tm|0;
 $__fmt = $__fmt|0;
 $__mod = $__mod|0;
 var $$$sroa$02$0$i = 0, $$0$i$i = 0, $$sroa$02$0$i = 0, $$sroa$02$0$i$lcssa = 0, $$sroa$02$1$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__nar = 0, $__ne = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__nar = sp + 4|0;
 $__ne = sp;
 $2 = ((($__nar)) + 100|0);
 HEAP32[$__ne>>2] = $2;
 $3 = ((($this)) + 8|0);
 __ZNKSt3__110__time_put8__do_putEPcRS1_PK2tmcc($3,$__nar,$__ne,$__tm,$__fmt,$__mod);
 $4 = HEAP32[$__ne>>2]|0;
 $5 = HEAP32[$__s>>2]|0;
 $$0$i$i = $__nar;$$sroa$02$0$i = $5;
 while(1) {
  $6 = ($$0$i$i|0)==($4|0);
  if ($6) {
   $$sroa$02$0$i$lcssa = $$sroa$02$0$i;
   break;
  }
  $7 = HEAP8[$$0$i$i>>0]|0;
  $8 = ($$sroa$02$0$i|0)==(0|0);
  do {
   if ($8) {
    $$sroa$02$1$i = 0;
   } else {
    $9 = ((($$sroa$02$0$i)) + 24|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = ((($$sroa$02$0$i)) + 28|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = ($10|0)==($12|0);
    if ($13) {
     $15 = HEAP32[$$sroa$02$0$i>>2]|0;
     $16 = ((($15)) + 52|0);
     $17 = HEAP32[$16>>2]|0;
     $18 = $7&255;
     $19 = (FUNCTION_TABLE_iii[$17 & 8191]($$sroa$02$0$i,$18)|0);
     $20 = ($19|0)==(-1);
     $$$sroa$02$0$i = $20 ? 0 : $$sroa$02$0$i;
     $$sroa$02$1$i = $$$sroa$02$0$i;
     break;
    } else {
     $14 = ((($10)) + 1|0);
     HEAP32[$9>>2] = $14;
     HEAP8[$10>>0] = $7;
     $$sroa$02$1$i = $$sroa$02$0$i;
     break;
    }
   }
  } while(0);
  $21 = ((($$0$i$i)) + 1|0);
  $$0$i$i = $21;$$sroa$02$0$i = $$sroa$02$1$i;
 }
 STACKTOP = sp;return ($$sroa$02$0$i$lcssa|0);
}
function __ZNKSt3__110__time_put8__do_putEPcRS1_PK2tmcc($this,$__nb,$__ne,$__tm,$__fmt,$__mod) {
 $this = $this|0;
 $__nb = $__nb|0;
 $__ne = $__ne|0;
 $__tm = $__tm|0;
 $__fmt = $__fmt|0;
 $__mod = $__mod|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $fmt = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $fmt = sp;
 HEAP8[$fmt>>0] = 37;
 $0 = ((($fmt)) + 1|0);
 HEAP8[$0>>0] = $__fmt;
 $1 = ((($fmt)) + 2|0);
 HEAP8[$1>>0] = $__mod;
 $2 = ((($fmt)) + 3|0);
 HEAP8[$2>>0] = 0;
 $3 = ($__mod<<24>>24)==(0);
 if (!($3)) {
  HEAP8[$0>>0] = $__mod;
  HEAP8[$1>>0] = $__fmt;
 }
 $4 = HEAP32[$__ne>>2]|0;
 $5 = $__nb;
 $6 = (($4) - ($5))|0;
 $7 = HEAP32[$this>>2]|0;
 $8 = (_strftime_l(($__nb|0),($6|0),($fmt|0),($__tm|0),($7|0))|0);
 $9 = (($__nb) + ($8)|0);
 HEAP32[$__ne>>2] = $9;
 STACKTOP = sp;return;
}
function __ZNSt3__18time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 __ZNSt3__110__time_putD2Ev($0);
 return;
}
function __ZNSt3__18time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 __ZNSt3__110__time_putD2Ev($0);
 __ZdlPv($this);
 return;
}
function __ZNKSt3__18time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwPK2tmcc($this,$__s,$0,$1,$__tm,$__fmt,$__mod) {
 $this = $this|0;
 $__s = $__s|0;
 $0 = $0|0;
 $1 = $1|0;
 $__tm = $__tm|0;
 $__fmt = $__fmt|0;
 $__mod = $__mod|0;
 var $$$sroa$02$0$i = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$sroa$02$0$i = 0, $$sroa$02$0$i$lcssa = 0, $$sroa$02$1$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__nar = 0, $__ne = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 416|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__nar = sp + 8|0;
 $__ne = sp;
 $2 = ((($__nar)) + 400|0);
 HEAP32[$__ne>>2] = $2;
 $3 = ((($this)) + 8|0);
 __ZNKSt3__110__time_put8__do_putEPwRS1_PK2tmcc($3,$__nar,$__ne,$__tm,$__fmt,$__mod);
 $4 = HEAP32[$__ne>>2]|0;
 $5 = HEAP32[$__s>>2]|0;
 $$0$i$i = $__nar;$$sroa$02$0$i = $5;
 while(1) {
  $6 = ($$0$i$i|0)==($4|0);
  if ($6) {
   $$sroa$02$0$i$lcssa = $$sroa$02$0$i;
   break;
  }
  $7 = HEAP32[$$0$i$i>>2]|0;
  $8 = ($$sroa$02$0$i|0)==(0|0);
  if ($8) {
   $$sroa$02$1$i = 0;
  } else {
   $9 = ((($$sroa$02$0$i)) + 24|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ((($$sroa$02$0$i)) + 28|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($10|0)==($12|0);
   if ($13) {
    $14 = HEAP32[$$sroa$02$0$i>>2]|0;
    $15 = ((($14)) + 52|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = (FUNCTION_TABLE_iii[$16 & 8191]($$sroa$02$0$i,$7)|0);
    $$0$i$i$i$i = $17;
   } else {
    $18 = ((($10)) + 4|0);
    HEAP32[$9>>2] = $18;
    HEAP32[$10>>2] = $7;
    $$0$i$i$i$i = $7;
   }
   $19 = ($$0$i$i$i$i|0)==(-1);
   $$$sroa$02$0$i = $19 ? 0 : $$sroa$02$0$i;
   $$sroa$02$1$i = $$$sroa$02$0$i;
  }
  $20 = ((($$0$i$i)) + 4|0);
  $$0$i$i = $20;$$sroa$02$0$i = $$sroa$02$1$i;
 }
 STACKTOP = sp;return ($$sroa$02$0$i$lcssa|0);
}
function __ZNKSt3__110__time_put8__do_putEPwRS1_PK2tmcc($this,$__wb,$__we,$__tm,$__fmt,$__mod) {
 $this = $this|0;
 $__wb = $__wb|0;
 $__we = $__we|0;
 $__tm = $__tm|0;
 $__fmt = $__fmt|0;
 $__mod = $__mod|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__nar = 0, $__nb = 0, $__ne = 0, $mb = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__nar = sp + 16|0;
 $__ne = sp + 12|0;
 $mb = sp;
 $__nb = sp + 8|0;
 $0 = ((($__nar)) + 100|0);
 HEAP32[$__ne>>2] = $0;
 __ZNKSt3__110__time_put8__do_putEPcRS1_PK2tmcc($this,$__nar,$__ne,$__tm,$__fmt,$__mod);
 $1 = $mb;
 $2 = $1;
 HEAP32[$2>>2] = 0;
 $3 = (($1) + 4)|0;
 $4 = $3;
 HEAP32[$4>>2] = 0;
 HEAP32[$__nb>>2] = $__nar;
 $5 = HEAP32[$__we>>2]|0;
 $6 = $__wb;
 $7 = (($5) - ($6))|0;
 $8 = $7 >> 2;
 $9 = HEAP32[$this>>2]|0;
 $10 = (_uselocale($9)|0);
 __THREW__ = 0;
 $11 = (invoke_iiiii(4898,($__wb|0),($__nb|0),($8|0),($mb|0))|0);
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if ($13) {
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  $21 = ($10|0)==(0|0);
  if (!($21)) {
   __THREW__ = 0;
   (invoke_ii(4876,($10|0))|0);
   $22 = __THREW__; __THREW__ = 0;
   $23 = $22&1;
   if ($23) {
    $24 = ___cxa_find_matching_catch_3(0|0)|0;
    $25 = tempRet0;
    ___clang_call_terminate($24);
    // unreachable;
   }
  }
  ___resumeException($19|0);
  // unreachable;
 }
 $14 = ($10|0)==(0|0);
 if (!($14)) {
  __THREW__ = 0;
  (invoke_ii(4876,($10|0))|0);
  $15 = __THREW__; __THREW__ = 0;
  $16 = $15&1;
  if ($16) {
   $17 = ___cxa_find_matching_catch_3(0|0)|0;
   $18 = tempRet0;
   ___clang_call_terminate($17);
   // unreachable;
  }
 }
 $26 = ($11|0)==(-1);
 if ($26) {
  __ZNSt3__121__throw_runtime_errorEPKc(621542);
  // unreachable;
 } else {
  $27 = (($__wb) + ($11<<2)|0);
  HEAP32[$__we>>2] = $27;
  STACKTOP = sp;return;
 }
}
function __ZNSt3__121__throw_runtime_errorEPKc($msg) {
 $msg = $msg|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (___cxa_allocate_exception(8)|0);
 __THREW__ = 0;
 invoke_vii(4855,($0|0),($msg|0));
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  ___cxa_free_exception(($0|0));
  ___resumeException($3|0);
  // unreachable;
 } else {
  ___cxa_throw(($0|0),(9088|0),(3331|0));
  // unreachable;
 }
}
function __ZNSt3__110moneypunctIcLb0EED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__110moneypunctIcLb0EED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__110moneypunctIcLb0EE16do_decimal_pointEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 127;
}
function __ZNKSt3__110moneypunctIcLb0EE16do_thousands_sepEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 127;
}
function __ZNKSt3__110moneypunctIcLb0EE11do_groupingEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $__i$0$i$i = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($agg$result) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 return;
}
function __ZNKSt3__110moneypunctIcLb0EE14do_curr_symbolEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $__i$0$i$i = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($agg$result) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 return;
}
function __ZNKSt3__110moneypunctIcLb0EE16do_positive_signEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $__i$0$i$i = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($agg$result) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 return;
}
function __ZNKSt3__110moneypunctIcLb0EE16do_negative_signEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEjc($agg$result,1,45);
 return;
}
function __ZNKSt3__110moneypunctIcLb0EE14do_frac_digitsEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__110moneypunctIcLb0EE13do_pos_formatEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$agg$result>>0]=67109634&255;HEAP8[$agg$result+1>>0]=(67109634>>8)&255;HEAP8[$agg$result+2>>0]=(67109634>>16)&255;HEAP8[$agg$result+3>>0]=67109634>>24;
 return;
}
function __ZNKSt3__110moneypunctIcLb0EE13do_neg_formatEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$agg$result>>0]=67109634&255;HEAP8[$agg$result+1>>0]=(67109634>>8)&255;HEAP8[$agg$result+2>>0]=(67109634>>16)&255;HEAP8[$agg$result+3>>0]=67109634>>24;
 return;
}
function __ZNSt3__110moneypunctIcLb1EED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__110moneypunctIcLb1EED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__110moneypunctIcLb1EE16do_decimal_pointEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 127;
}
function __ZNKSt3__110moneypunctIcLb1EE16do_thousands_sepEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 127;
}
function __ZNKSt3__110moneypunctIcLb1EE11do_groupingEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $__i$0$i$i = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($agg$result) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 return;
}
function __ZNKSt3__110moneypunctIcLb1EE14do_curr_symbolEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $__i$0$i$i = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($agg$result) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 return;
}
function __ZNKSt3__110moneypunctIcLb1EE16do_positive_signEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $__i$0$i$i = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($agg$result) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 return;
}
function __ZNKSt3__110moneypunctIcLb1EE16do_negative_signEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEjc($agg$result,1,45);
 return;
}
function __ZNKSt3__110moneypunctIcLb1EE14do_frac_digitsEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__110moneypunctIcLb1EE13do_pos_formatEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$agg$result>>0]=67109634&255;HEAP8[$agg$result+1>>0]=(67109634>>8)&255;HEAP8[$agg$result+2>>0]=(67109634>>16)&255;HEAP8[$agg$result+3>>0]=67109634>>24;
 return;
}
function __ZNKSt3__110moneypunctIcLb1EE13do_neg_formatEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$agg$result>>0]=67109634&255;HEAP8[$agg$result+1>>0]=(67109634>>8)&255;HEAP8[$agg$result+2>>0]=(67109634>>16)&255;HEAP8[$agg$result+3>>0]=67109634>>24;
 return;
}
function __ZNSt3__110moneypunctIwLb0EED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__110moneypunctIwLb0EED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__110moneypunctIwLb0EE16do_decimal_pointEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 2147483647;
}
function __ZNKSt3__110moneypunctIwLb0EE16do_thousands_sepEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 2147483647;
}
function __ZNKSt3__110moneypunctIwLb0EE11do_groupingEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $__i$0$i$i = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($agg$result) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 return;
}
function __ZNKSt3__110moneypunctIwLb0EE14do_curr_symbolEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $__i$0$i$i = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($agg$result) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 return;
}
function __ZNKSt3__110moneypunctIwLb0EE16do_positive_signEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $__i$0$i$i = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($agg$result) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 return;
}
function __ZNKSt3__110moneypunctIwLb0EE16do_negative_signEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6__initEjw($agg$result,1,45);
 return;
}
function __ZNKSt3__110moneypunctIwLb0EE14do_frac_digitsEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__110moneypunctIwLb0EE13do_pos_formatEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$agg$result>>0]=67109634&255;HEAP8[$agg$result+1>>0]=(67109634>>8)&255;HEAP8[$agg$result+2>>0]=(67109634>>16)&255;HEAP8[$agg$result+3>>0]=67109634>>24;
 return;
}
function __ZNKSt3__110moneypunctIwLb0EE13do_neg_formatEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$agg$result>>0]=67109634&255;HEAP8[$agg$result+1>>0]=(67109634>>8)&255;HEAP8[$agg$result+2>>0]=(67109634>>16)&255;HEAP8[$agg$result+3>>0]=67109634>>24;
 return;
}
function __ZNSt3__110moneypunctIwLb1EED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__110moneypunctIwLb1EED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__110moneypunctIwLb1EE16do_decimal_pointEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 2147483647;
}
function __ZNKSt3__110moneypunctIwLb1EE16do_thousands_sepEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 2147483647;
}
function __ZNKSt3__110moneypunctIwLb1EE11do_groupingEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $__i$0$i$i = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($agg$result) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 return;
}
function __ZNKSt3__110moneypunctIwLb1EE14do_curr_symbolEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $__i$0$i$i = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($agg$result) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 return;
}
function __ZNKSt3__110moneypunctIwLb1EE16do_positive_signEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $__i$0$i$i = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($agg$result) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 return;
}
function __ZNKSt3__110moneypunctIwLb1EE16do_negative_signEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6__initEjw($agg$result,1,45);
 return;
}
function __ZNKSt3__110moneypunctIwLb1EE14do_frac_digitsEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__110moneypunctIwLb1EE13do_pos_formatEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$agg$result>>0]=67109634&255;HEAP8[$agg$result+1>>0]=(67109634>>8)&255;HEAP8[$agg$result+2>>0]=(67109634>>16)&255;HEAP8[$agg$result+3>>0]=67109634>>24;
 return;
}
function __ZNKSt3__110moneypunctIwLb1EE13do_neg_formatEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$agg$result>>0]=67109634&255;HEAP8[$agg$result+1>>0]=(67109634>>8)&255;HEAP8[$agg$result+2>>0]=(67109634>>16)&255;HEAP8[$agg$result+3>>0]=67109634>>24;
 return;
}
function __ZNSt3__19money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__19money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__19money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_bRNS_8ios_baseERjRe($this,$__b,$__e,$__intl,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__intl = $__intl|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0 = 0, $$0$i = 0, $$0$lcssa$i = 0, $$01 = 0, $$114 = 0, $$115 = 0, $$1212 = 0, $$1213 = 0, $$byval_copy = 0, $$phi$trans$insert$pre$phiZZZZ2D = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre16 = 0, $$pre8 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__atoms = 0, $__h$sroa$0$0 = 0, $__h$sroa$0$1 = 0, $__loc = 0, $__nbuf = 0, $__nc$0 = 0, $__nc$1 = 0, $__nc$2 = 0, $__nc$2$lcssa = 0, $__neg = 0, $__w$0 = 0, $__wb = 0;
 var $__wbuf = 0, $__wn = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 24|0;
 $vararg_buffer = sp;
 $__wbuf = sp + 140|0;
 $__wb = sp + 16|0;
 $__wn = sp + 12|0;
 $__loc = sp + 8|0;
 $__neg = sp + 138|0;
 $0 = sp + 4|0;
 $__atoms = sp + 128|0;
 $__nbuf = sp + 28|0;
 $1 = $__wbuf;
 HEAP32[$__wb>>2] = $1;
 $2 = ((($__wb)) + 4|0);
 HEAP32[$2>>2] = (4899);
 $3 = ((($__wbuf)) + 100|0);
 __THREW__ = 0;
 $4 = (invoke_ii(3373,($__iob|0))|0);
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $33 = ___cxa_find_matching_catch_2()|0;
  $34 = tempRet0;
  HEAP32[$__wb>>2] = 0;
  $$114 = $33;$$1212 = $34;$108 = 4899;$109 = $__wbuf;
  label = 47;
 } else {
  HEAP32[$__loc>>2] = $4;
  __THREW__ = 0;
  $7 = (invoke_iii(3374,($__loc|0),(1207576|0))|0);
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  L4: do {
   if ($9) {
    label = 10;
   } else {
    HEAP8[$__neg>>0] = 0;
    $10 = HEAP32[$__e>>2]|0;
    HEAP32[$0>>2] = $10;
    $11 = ((($__iob)) + 4|0);
    $12 = HEAP32[$11>>2]|0;
    __THREW__ = 0;
    ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
    $13 = (invoke_iiiiiiiiiiii(4900,($__b|0),($$byval_copy|0),($__intl|0),($__loc|0),($12|0),($__err|0),($__neg|0),($7|0),($__wb|0),($__wn|0),($3|0))|0);
    $14 = __THREW__; __THREW__ = 0;
    $15 = $14&1;
    if ($15) {
     label = 10;
    } else {
     L7: do {
      if ($13) {
       $16 = HEAP32[$7>>2]|0;
       $17 = ((($16)) + 32|0);
       $18 = HEAP32[$17>>2]|0;
       __THREW__ = 0;
       (invoke_iiiii($18|0,($7|0),(621791|0),((621801)|0),($__atoms|0))|0);
       $19 = __THREW__; __THREW__ = 0;
       $20 = $19&1;
       if ($20) {
        label = 10;
        break L4;
       }
       $21 = HEAP32[$__wn>>2]|0;
       $22 = HEAP32[$__wb>>2]|0;
       $23 = (($21) - ($22))|0;
       $24 = ($23|0)>(98);
       $25 = $22;
       $26 = $21;
       if ($24) {
        $27 = (($23) + 2)|0;
        $28 = (_malloc($27)|0);
        $29 = $28;
        $30 = ($28|0)==(0|0);
        if ($30) {
         __THREW__ = 0;
         invoke_v(4863);
         $31 = __THREW__; __THREW__ = 0;
         $32 = $31&1;
         if ($32) {
          $__h$sroa$0$0 = $29;
         } else {
          $__h$sroa$0$1 = $29;$__nc$0 = 0;
          label = 13;
         }
        } else {
         $__h$sroa$0$1 = $29;$__nc$0 = $28;
         label = 13;
        }
       } else {
        $__h$sroa$0$1 = 0;$__nc$0 = $__nbuf;
        label = 13;
       }
       do {
        if ((label|0) == 13) {
         $41 = HEAP8[$__neg>>0]|0;
         $42 = ($41<<24>>24)==(0);
         if ($42) {
          $__nc$1 = $__nc$0;
         } else {
          $43 = ((($__nc$0)) + 1|0);
          HEAP8[$__nc$0>>0] = 45;
          $__nc$1 = $43;
         }
         $44 = ((($__atoms)) + 10|0);
         $45 = $__atoms;
         $47 = $26;$__nc$2 = $__nc$1;$__w$0 = $25;
         while(1) {
          $46 = ($__w$0>>>0)<($47>>>0);
          if (!($46)) {
           $__nc$2$lcssa = $__nc$2;
           break;
          }
          $50 = HEAP8[$__w$0>>0]|0;
          $$0$i = $__atoms;
          while(1) {
           $51 = ($$0$i|0)==($44|0);
           if ($51) {
            $$0$lcssa$i = $44;
            break;
           }
           $52 = HEAP8[$$0$i>>0]|0;
           $53 = ($52<<24>>24)==($50<<24>>24);
           if ($53) {
            $$0$lcssa$i = $$0$i;
            break;
           }
           $54 = ((($$0$i)) + 1|0);
           $$0$i = $54;
          }
          $55 = $$0$lcssa$i;
          $56 = (($55) - ($45))|0;
          $57 = (621791 + ($56)|0);
          $58 = HEAP8[$57>>0]|0;
          HEAP8[$__nc$2>>0] = $58;
          $59 = ((($__w$0)) + 1|0);
          $60 = ((($__nc$2)) + 1|0);
          $$pre = HEAP32[$__wn>>2]|0;
          $47 = $$pre;$__nc$2 = $60;$__w$0 = $59;
         }
         HEAP8[$__nc$2$lcssa>>0] = 0;
         HEAP32[$vararg_buffer>>2] = $__v;
         $48 = (_sscanf($__nbuf,621802,$vararg_buffer)|0);
         $49 = ($48|0)==(1);
         if (!($49)) {
          __THREW__ = 0;
          invoke_vi(4901,(621806|0));
          $61 = __THREW__; __THREW__ = 0;
          $62 = $61&1;
          if ($62) {
           $__h$sroa$0$0 = $__h$sroa$0$1;
           break;
          }
         }
         $63 = ($__h$sroa$0$1|0)==(0);
         if (!($63)) {
          $64 = $__h$sroa$0$1;
          _free($64);
         }
         break L7;
        }
       } while(0);
       $37 = ___cxa_find_matching_catch_2()|0;
       $38 = tempRet0;
       $39 = ($__h$sroa$0$0|0)==(0);
       if ($39) {
        $$0 = $37;$$01 = $38;$$phi$trans$insert$pre$phiZZZZ2D = $__wb;
        break L4;
       }
       $40 = $__h$sroa$0$0;
       _free($40);
       $$0 = $37;$$01 = $38;$$phi$trans$insert$pre$phiZZZZ2D = $__wb;
       break L4;
      }
     } while(0);
     $65 = HEAP32[$__b>>2]|0;
     $66 = ($65|0)==(0|0);
     do {
      if ($66) {
       $80 = 0;
      } else {
       $67 = ((($65)) + 12|0);
       $68 = HEAP32[$67>>2]|0;
       $69 = ((($65)) + 16|0);
       $70 = HEAP32[$69>>2]|0;
       $71 = ($68|0)==($70|0);
       if ($71) {
        $72 = HEAP32[$65>>2]|0;
        $73 = ((($72)) + 36|0);
        $74 = HEAP32[$73>>2]|0;
        __THREW__ = 0;
        $75 = (invoke_ii($74|0,($65|0))|0);
        $76 = __THREW__; __THREW__ = 0;
        $77 = $76&1;
        if ($77) {
         label = 10;
         break L4;
        }
        $78 = ($75|0)==(-1);
        if ($78) {
         HEAP32[$__b>>2] = 0;
         $80 = 0;
         break;
        } else {
         $$pre$i$i$i = HEAP32[$__b>>2]|0;
         $80 = $$pre$i$i$i;
         break;
        }
       } else {
        $80 = $65;
       }
      }
     } while(0);
     $79 = ($80|0)==(0|0);
     $81 = HEAP32[$__e>>2]|0;
     $82 = ($81|0)==(0|0);
     do {
      if ($82) {
       label = 39;
      } else {
       $83 = ((($81)) + 12|0);
       $84 = HEAP32[$83>>2]|0;
       $85 = ((($81)) + 16|0);
       $86 = HEAP32[$85>>2]|0;
       $87 = ($84|0)==($86|0);
       if ($87) {
        $88 = HEAP32[$81>>2]|0;
        $89 = ((($88)) + 36|0);
        $90 = HEAP32[$89>>2]|0;
        __THREW__ = 0;
        $91 = (invoke_ii($90|0,($81|0))|0);
        $92 = __THREW__; __THREW__ = 0;
        $93 = $92&1;
        if ($93) {
         label = 10;
         break L4;
        }
        $94 = ($91|0)==(-1);
        if ($94) {
         HEAP32[$__e>>2] = 0;
         label = 39;
         break;
        }
       }
       if (!($79)) {
        label = 40;
       }
      }
     } while(0);
     if ((label|0) == 39) {
      if ($79) {
       label = 40;
      }
     }
     if ((label|0) == 40) {
      $95 = HEAP32[$__err>>2]|0;
      $96 = $95 | 2;
      HEAP32[$__err>>2] = $96;
     }
     $97 = HEAP32[$__b>>2]|0;
     $98 = HEAP32[$__loc>>2]|0;
     (__ZNSt3__114__shared_count16__release_sharedEv($98)|0);
     $99 = HEAP32[$__wb>>2]|0;
     HEAP32[$__wb>>2] = 0;
     $100 = ($99|0)==(0|0);
     if (!($100)) {
      $101 = HEAP32[$2>>2]|0;
      __THREW__ = 0;
      invoke_vi($101|0,($99|0));
      $102 = __THREW__; __THREW__ = 0;
      $103 = $102&1;
      if ($103) {
       $104 = ___cxa_find_matching_catch_3(0|0)|0;
       $105 = tempRet0;
       ___clang_call_terminate($104);
       // unreachable;
      }
     }
     STACKTOP = sp;return ($97|0);
    }
   }
  } while(0);
  if ((label|0) == 10) {
   $35 = ___cxa_find_matching_catch_2()|0;
   $36 = tempRet0;
   $$0 = $35;$$01 = $36;$$phi$trans$insert$pre$phiZZZZ2D = $__wb;
  }
  $106 = HEAP32[$__loc>>2]|0;
  (__ZNSt3__114__shared_count16__release_sharedEv($106)|0);
  $$pre8 = HEAP32[$$phi$trans$insert$pre$phiZZZZ2D>>2]|0;
  HEAP32[$$phi$trans$insert$pre$phiZZZZ2D>>2] = 0;
  $107 = ($$pre8|0)==(0|0);
  if ($107) {
   $$115 = $$0;$$1213 = $$01;
  } else {
   $$pre16 = HEAP32[$2>>2]|0;
   $$114 = $$0;$$1212 = $$01;$108 = $$pre16;$109 = $$pre8;
   label = 47;
  }
 }
 if ((label|0) == 47) {
  __THREW__ = 0;
  invoke_vi($108|0,($109|0));
  $110 = __THREW__; __THREW__ = 0;
  $111 = $110&1;
  if ($111) {
   $112 = ___cxa_find_matching_catch_3(0|0)|0;
   $113 = tempRet0;
   ___clang_call_terminate($112);
   // unreachable;
  } else {
   $$115 = $$114;$$1213 = $$1212;
  }
 }
 ___resumeException($$115|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__112__do_nothingEPv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__19money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE8__do_getERS4_S4_bRKNS_6localeEjRjRbRKNS_5ctypeIcEERNS_10unique_ptrIcPFvPvEEERPcSM_($__b,$__e,$__intl,$__loc,$__flags,$__err,$__neg,$__ct,$__wb,$__wn,$__we) {
 $__b = $__b|0;
 $__e = $__e|0;
 $__intl = $__intl|0;
 $__loc = $__loc|0;
 $__flags = $__flags|0;
 $__err = $__err|0;
 $__neg = $__neg|0;
 $__ct = $__ct|0;
 $__wb = $__wb|0;
 $__wn = $__wn|0;
 $__we = $__we|0;
 var $$0$i$i = 0, $$0$i$i109 = 0, $$0$i$i112 = 0, $$0$i$i129 = 0, $$0$i$i136 = 0, $$0$i$i145 = 0, $$0$i$i155 = 0, $$0$i$i172 = 0, $$0$i$i57 = 0, $$0$i$i60 = 0, $$0$i$i73 = 0, $$0$i$i80 = 0, $$0$i$i83 = 0, $$0$i$i90 = 0, $$10 = 0, $$lcssa = 0, $$lcssa392 = 0, $$lcssa394 = 0, $$phi$trans$insert = 0, $$phi$trans$insert301 = 0;
 var $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i119 = 0, $$pre$i$i$i159 = 0, $$pre$i$i$i176 = 0, $$pre$i$i$i47 = 0, $$pre$i$i$i63 = 0, $$pre$i$i$i95 = 0, $$pre$i2$i$i123 = 0, $$pre$i2$i$i163 = 0, $$pre$i2$i$i163283 = 0, $$pre$i2$i$i163285 = 0, $$pre$i2$i$i180 = 0, $$pre$i2$i$i180295 = 0, $$pre$i2$i$i180296 = 0, $$pre$i2$i$i67 = 0, $$pre$i2$i$i67305 = 0, $$pre$i2$i$i67307 = 0, $$pre$i2$i$i99 = 0, $$pre$i2$i$i99291$ph = 0;
 var $$pre$i2$i$i99291310 = 0, $$pre288 = 0, $$pre290 = 0, $$pre293 = 0, $$pre294 = 0, $$pre298 = 0, $$pre299 = 0, $$pre300 = 0, $$pre302 = 0, $$pre303 = 0, $$pre304 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0;
 var $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0;
 var $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0;
 var $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0;
 var $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0;
 var $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0;
 var $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0;
 var $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0;
 var $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0;
 var $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0;
 var $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0;
 var $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0;
 var $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0;
 var $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0;
 var $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0;
 var $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0;
 var $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0;
 var $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0;
 var $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0;
 var $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0;
 var $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0;
 var $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0;
 var $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0;
 var $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0;
 var $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0;
 var $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0;
 var $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0;
 var $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0;
 var $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0;
 var $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0;
 var $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0;
 var $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0;
 var $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0;
 var $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0;
 var $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0;
 var $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0;
 var $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0;
 var $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0;
 var $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__dp = 0, $__et = 0, $__fd = 0, $__gb = 0, $__gbuf = 0, $__ge = 0, $__gn = 0, $__grp = 0, $__i$0 = 0, $__i$0$i$i = 0, $__i$0$i$i29 = 0;
 var $__i$0$i$i32 = 0, $__i$0$i$i36 = 0, $__i$0$i$i41 = 0, $__ng$0 = 0, $__ng$0$lcssa393 = 0, $__ng$2$ph = 0, $__nsn = 0, $__nsn$__trailing_sign$0 = 0, $__nsn$__trailing_sign$023 = 0, $__p$0 = 0, $__pat = 0, $__psn = 0, $__psn$__trailing_sign$0 = 0, $__psn$__trailing_sign$022 = 0, $__spaces = 0, $__sym = 0, $__sym_curr_char$sroa$0$0 = 0, $__sym_space_end$sroa$0$0 = 0, $__sym_space_end$sroa$0$0$lcssa = 0, $__trailing_sign$0 = 0;
 var $__trailing_sign$0$lcssa = 0, $__trailing_sign$1 = 0, $__ts = 0, $brmerge = 0, $cond28 = 0, $exitcond$i$i = 0, $exitcond$i$i30 = 0, $exitcond$i$i33 = 0, $exitcond$i$i37 = 0, $exitcond$i$i42 = 0, $isascii1$i = 0, $isascii1$i153 = 0, $isascii1$i186 = 0, $isascii1$i76 = 0, $isascii1$i88 = 0, $lpad$phi$index = 0, $lpad$phi$index7 = 0, $or$cond = 0, $or$cond12 = 0, $or$cond14 = 0;
 var $or$cond25 = 0, $sext = 0, $sext239 = 0, $sext240 = 0, $sext241 = 0, $switch = 0, $tmp$i$i101 = 0, $tmp$i$i165 = 0, $tmp$i$i182 = 0, $tmp$i$i69 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 512|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 488|0;
 $__gbuf = sp + 88|0;
 $__gb = sp + 80|0;
 $__gn = sp + 72|0;
 $__ge = sp + 68|0;
 $__pat = sp + 496|0;
 $__dp = sp + 493|0;
 $__ts = sp + 492|0;
 $__grp = sp + 56|0;
 $__sym = sp + 44|0;
 $__psn = sp + 32|0;
 $__nsn = sp + 20|0;
 $__spaces = sp + 8|0;
 $__fd = sp + 4|0;
 $__et = sp;
 HEAP32[$0>>2] = $__we;
 $1 = $__gbuf;
 HEAP32[$__gb>>2] = $1;
 $2 = ((($__gb)) + 4|0);
 HEAP32[$2>>2] = (4899);
 HEAP32[$__gn>>2] = $__gbuf;
 $3 = ((($__gbuf)) + 400|0);
 HEAP32[$__ge>>2] = $3;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $4 = (($__grp) + ($__i$0$i$i<<2)|0);
  HEAP32[$4>>2] = 0;
  $5 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $5;
 }
 $__i$0$i$i29 = 0;
 while(1) {
  $exitcond$i$i30 = ($__i$0$i$i29|0)==(3);
  if ($exitcond$i$i30) {
   break;
  }
  $6 = (($__sym) + ($__i$0$i$i29<<2)|0);
  HEAP32[$6>>2] = 0;
  $7 = (($__i$0$i$i29) + 1)|0;
  $__i$0$i$i29 = $7;
 }
 $__i$0$i$i32 = 0;
 while(1) {
  $exitcond$i$i33 = ($__i$0$i$i32|0)==(3);
  if ($exitcond$i$i33) {
   break;
  }
  $8 = (($__psn) + ($__i$0$i$i32<<2)|0);
  HEAP32[$8>>2] = 0;
  $9 = (($__i$0$i$i32) + 1)|0;
  $__i$0$i$i32 = $9;
 }
 $__i$0$i$i36 = 0;
 while(1) {
  $exitcond$i$i37 = ($__i$0$i$i36|0)==(3);
  if ($exitcond$i$i37) {
   break;
  }
  $10 = (($__nsn) + ($__i$0$i$i36<<2)|0);
  HEAP32[$10>>2] = 0;
  $11 = (($__i$0$i$i36) + 1)|0;
  $__i$0$i$i36 = $11;
 }
 $__i$0$i$i41 = 0;
 while(1) {
  $exitcond$i$i42 = ($__i$0$i$i41|0)==(3);
  if ($exitcond$i$i42) {
   break;
  }
  $12 = (($__spaces) + ($__i$0$i$i41<<2)|0);
  HEAP32[$12>>2] = 0;
  $13 = (($__i$0$i$i41) + 1)|0;
  $__i$0$i$i41 = $13;
 }
 __THREW__ = 0;
 invoke_viiiiiiiiii(4902,($__intl|0),($__loc|0),($__pat|0),($__dp|0),($__ts|0),($__grp|0),($__sym|0),($__psn|0),($__nsn|0),($__fd|0));
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 L21: do {
  if ($15) {
   $77 = ___cxa_find_matching_catch_2()|0;
   $78 = tempRet0;
   $lpad$phi$index = $77;$lpad$phi$index7 = $78;
  } else {
   $16 = HEAP32[$__wb>>2]|0;
   HEAP32[$__wn>>2] = $16;
   $17 = ((($__ct)) + 8|0);
   $18 = ((($__psn)) + 4|0);
   $19 = ((($__nsn)) + 4|0);
   $20 = ((($__nsn)) + 8|0);
   $21 = ((($__nsn)) + 1|0);
   $22 = ((($__psn)) + 8|0);
   $23 = ((($__psn)) + 1|0);
   $24 = $__flags & 512;
   $25 = ($24|0)!=(0);
   $26 = ((($__sym)) + 8|0);
   $27 = ((($__sym)) + 1|0);
   $28 = ((($__sym)) + 4|0);
   $29 = ((($__spaces)) + 4|0);
   $30 = ((($__spaces)) + 8|0);
   $31 = ((($__spaces)) + 1|0);
   $32 = ((($__pat)) + 3|0);
   $33 = ((($__grp)) + 4|0);
   $867 = $__gbuf;$__p$0 = 0;$__trailing_sign$0 = 0;
   L24: while(1) {
    $34 = ($__p$0>>>0)<(4);
    if (!($34)) {
     $$lcssa394 = $867;$__trailing_sign$0$lcssa = $__trailing_sign$0;
     label = 251;
     break;
    }
    $35 = HEAP32[$__b>>2]|0;
    $36 = ($35|0)==(0|0);
    do {
     if ($36) {
      $50 = 0;
     } else {
      $37 = ((($35)) + 12|0);
      $38 = HEAP32[$37>>2]|0;
      $39 = ((($35)) + 16|0);
      $40 = HEAP32[$39>>2]|0;
      $41 = ($38|0)==($40|0);
      if ($41) {
       $42 = HEAP32[$35>>2]|0;
       $43 = ((($42)) + 36|0);
       $44 = HEAP32[$43>>2]|0;
       __THREW__ = 0;
       $45 = (invoke_ii($44|0,($35|0))|0);
       $46 = __THREW__; __THREW__ = 0;
       $47 = $46&1;
       if ($47) {
        label = 38;
        break L24;
       }
       $48 = ($45|0)==(-1);
       if ($48) {
        HEAP32[$__b>>2] = 0;
        $50 = 0;
        break;
       } else {
        $$pre$i$i$i47 = HEAP32[$__b>>2]|0;
        $50 = $$pre$i$i$i47;
        break;
       }
      } else {
       $50 = $35;
      }
     }
    } while(0);
    $49 = ($50|0)==(0|0);
    $51 = HEAP32[$__e>>2]|0;
    $52 = ($51|0)==(0|0);
    do {
     if ($52) {
      label = 31;
     } else {
      $53 = ((($51)) + 12|0);
      $54 = HEAP32[$53>>2]|0;
      $55 = ((($51)) + 16|0);
      $56 = HEAP32[$55>>2]|0;
      $57 = ($54|0)==($56|0);
      if (!($57)) {
       if ($49) {
        $868 = $51;
        break;
       } else {
        $$lcssa394 = $867;$__trailing_sign$0$lcssa = $__trailing_sign$0;
        label = 251;
        break L24;
       }
      }
      $58 = HEAP32[$51>>2]|0;
      $59 = ((($58)) + 36|0);
      $60 = HEAP32[$59>>2]|0;
      __THREW__ = 0;
      $61 = (invoke_ii($60|0,($51|0))|0);
      $62 = __THREW__; __THREW__ = 0;
      $63 = $62&1;
      if ($63) {
       label = 38;
       break L24;
      }
      $64 = ($61|0)==(-1);
      if ($64) {
       HEAP32[$__e>>2] = 0;
       label = 31;
       break;
      } else {
       if ($49) {
        $868 = $51;
        break;
       } else {
        $$lcssa394 = $867;$__trailing_sign$0$lcssa = $__trailing_sign$0;
        label = 251;
        break L24;
       }
      }
     }
    } while(0);
    if ((label|0) == 31) {
     label = 0;
     if ($49) {
      $$lcssa394 = $867;$__trailing_sign$0$lcssa = $__trailing_sign$0;
      label = 251;
      break;
     } else {
      $868 = 0;
     }
    }
    $86 = (($__pat) + ($__p$0)|0);
    $87 = HEAP8[$86>>0]|0;
    $88 = $87 << 24 >> 24;
    L47: do {
     switch ($88|0) {
     case 1:  {
      $cond28 = ($__p$0|0)==(3);
      if ($cond28) {
       $869 = $867;$__trailing_sign$1 = $__trailing_sign$0;
      } else {
       $90 = HEAP32[$__b>>2]|0;
       $91 = ((($90)) + 12|0);
       $92 = HEAP32[$91>>2]|0;
       $93 = ((($90)) + 16|0);
       $94 = HEAP32[$93>>2]|0;
       $95 = ($92|0)==($94|0);
       if ($95) {
        $96 = HEAP32[$90>>2]|0;
        $97 = ((($96)) + 36|0);
        $98 = HEAP32[$97>>2]|0;
        __THREW__ = 0;
        $99 = (invoke_ii($98|0,($90|0))|0);
        $100 = __THREW__; __THREW__ = 0;
        $101 = $100&1;
        if ($101) {
         label = 38;
         break L24;
        } else {
         $$0$i$i57 = $99;
        }
       } else {
        $102 = HEAP8[$92>>0]|0;
        $103 = $102&255;
        $$0$i$i57 = $103;
       }
       $104 = $$0$i$i57&255;
       $isascii1$i = ($104<<24>>24)>(-1);
       if (!($isascii1$i)) {
        label = 56;
        break L24;
       }
       $sext241 = $$0$i$i57 << 24;
       $105 = $sext241 >> 24;
       $106 = HEAP32[$17>>2]|0;
       $107 = (($106) + ($105<<1)|0);
       $108 = HEAP16[$107>>1]|0;
       $109 = $108 & 8192;
       $110 = ($109<<16>>16)==(0);
       if ($110) {
        label = 56;
        break L24;
       }
       $111 = HEAP32[$__b>>2]|0;
       $112 = ((($111)) + 12|0);
       $113 = HEAP32[$112>>2]|0;
       $114 = ((($111)) + 16|0);
       $115 = HEAP32[$114>>2]|0;
       $116 = ($113|0)==($115|0);
       if ($116) {
        $117 = HEAP32[$111>>2]|0;
        $118 = ((($117)) + 40|0);
        $119 = HEAP32[$118>>2]|0;
        __THREW__ = 0;
        $120 = (invoke_ii($119|0,($111|0))|0);
        $121 = __THREW__; __THREW__ = 0;
        $122 = $121&1;
        if ($122) {
         label = 38;
         break L24;
        } else {
         $$0$i$i60 = $120;
        }
       } else {
        $123 = ((($113)) + 1|0);
        HEAP32[$112>>2] = $123;
        $124 = HEAP8[$113>>0]|0;
        $125 = $124&255;
        $$0$i$i60 = $125;
       }
       $126 = $$0$i$i60&255;
       __THREW__ = 0;
       invoke_vii(3370,($__spaces|0),($126|0));
       $127 = __THREW__; __THREW__ = 0;
       $128 = $127&1;
       if ($128) {
        label = 38;
        break L24;
       } else {
        $$pre$i2$i$i67 = $868;$149 = $868;
        label = 58;
       }
      }
      break;
     }
     case 0:  {
      $131 = ($__p$0|0)==(3);
      if ($131) {
       $869 = $867;$__trailing_sign$1 = $__trailing_sign$0;
      } else {
       $$pre$i2$i$i67 = $868;$149 = $868;
       label = 58;
      }
      break;
     }
     case 3:  {
      $202 = HEAP8[$__psn>>0]|0;
      $203 = $202 & 1;
      $204 = ($203<<24>>24)==(0);
      $205 = HEAP32[$18>>2]|0;
      $206 = $202&255;
      $207 = $206 >>> 1;
      $208 = $204 ? $207 : $205;
      $209 = HEAP8[$__nsn>>0]|0;
      $210 = $209 & 1;
      $211 = ($210<<24>>24)==(0);
      $212 = HEAP32[$19>>2]|0;
      $213 = $209&255;
      $214 = $213 >>> 1;
      $215 = $211 ? $214 : $212;
      $216 = (0 - ($215))|0;
      $217 = ($208|0)==($216|0);
      if ($217) {
       $869 = $867;$__trailing_sign$1 = $__trailing_sign$0;
      } else {
       $218 = ($208|0)==(0);
       if ($218) {
        $262 = HEAP32[$__b>>2]|0;
        $263 = ((($262)) + 12|0);
        $264 = HEAP32[$263>>2]|0;
        $265 = ((($262)) + 16|0);
        $266 = HEAP32[$265>>2]|0;
        $267 = ($264|0)==($266|0);
        if ($267) {
         $268 = HEAP32[$262>>2]|0;
         $269 = ((($268)) + 36|0);
         $270 = HEAP32[$269>>2]|0;
         __THREW__ = 0;
         $271 = (invoke_ii($270|0,($262|0))|0);
         $272 = __THREW__; __THREW__ = 0;
         $273 = $272&1;
         if ($273) {
          label = 38;
          break L24;
         }
         $$pre304 = HEAP8[$__nsn>>0]|0;
         $$0$i$i129 = $271;$278 = $$pre304;
        } else {
         $274 = HEAP8[$264>>0]|0;
         $275 = $274&255;
         $$0$i$i129 = $275;$278 = $209;
        }
        $276 = $$0$i$i129&255;
        $277 = $278 & 1;
        $279 = ($277<<24>>24)==(0);
        $280 = HEAP32[$20>>2]|0;
        $281 = $279 ? $21 : $280;
        $282 = HEAP8[$281>>0]|0;
        $283 = ($276<<24>>24)==($282<<24>>24);
        if (!($283)) {
         $869 = $867;$__trailing_sign$1 = $__trailing_sign$0;
         break L47;
        }
        $284 = HEAP32[$__b>>2]|0;
        $285 = ((($284)) + 12|0);
        $286 = HEAP32[$285>>2]|0;
        $287 = ((($284)) + 16|0);
        $288 = HEAP32[$287>>2]|0;
        $289 = ($286|0)==($288|0);
        if ($289) {
         $290 = HEAP32[$284>>2]|0;
         $291 = ((($290)) + 40|0);
         $292 = HEAP32[$291>>2]|0;
         __THREW__ = 0;
         (invoke_ii($292|0,($284|0))|0);
         $293 = __THREW__; __THREW__ = 0;
         $294 = $293&1;
         if ($294) {
          label = 38;
          break L24;
         }
        } else {
         $295 = ((($286)) + 1|0);
         HEAP32[$285>>2] = $295;
        }
        HEAP8[$__neg>>0] = 1;
        $296 = HEAP8[$__nsn>>0]|0;
        $297 = $296 & 1;
        $298 = ($297<<24>>24)==(0);
        $299 = HEAP32[$19>>2]|0;
        $300 = $296&255;
        $301 = $300 >>> 1;
        $302 = $298 ? $301 : $299;
        $303 = ($302>>>0)>(1);
        $__nsn$__trailing_sign$0 = $303 ? $__nsn : $__trailing_sign$0;
        $869 = $867;$__trailing_sign$1 = $__nsn$__trailing_sign$0;
        break L47;
       }
       $219 = ($215|0)==(0);
       $220 = HEAP32[$__b>>2]|0;
       $221 = ((($220)) + 12|0);
       $222 = HEAP32[$221>>2]|0;
       $223 = ((($220)) + 16|0);
       $224 = HEAP32[$223>>2]|0;
       $225 = ($222|0)==($224|0);
       if ($219) {
        if ($225) {
         $226 = HEAP32[$220>>2]|0;
         $227 = ((($226)) + 36|0);
         $228 = HEAP32[$227>>2]|0;
         __THREW__ = 0;
         $229 = (invoke_ii($228|0,($220|0))|0);
         $230 = __THREW__; __THREW__ = 0;
         $231 = $230&1;
         if ($231) {
          label = 38;
          break L24;
         }
         $$pre303 = HEAP8[$__psn>>0]|0;
         $$0$i$i109 = $229;$236 = $$pre303;
        } else {
         $232 = HEAP8[$222>>0]|0;
         $233 = $232&255;
         $$0$i$i109 = $233;$236 = $202;
        }
        $234 = $$0$i$i109&255;
        $235 = $236 & 1;
        $237 = ($235<<24>>24)==(0);
        $238 = HEAP32[$22>>2]|0;
        $239 = $237 ? $23 : $238;
        $240 = HEAP8[$239>>0]|0;
        $241 = ($234<<24>>24)==($240<<24>>24);
        if (!($241)) {
         HEAP8[$__neg>>0] = 1;
         $869 = $867;$__trailing_sign$1 = $__trailing_sign$0;
         break L47;
        }
        $242 = HEAP32[$__b>>2]|0;
        $243 = ((($242)) + 12|0);
        $244 = HEAP32[$243>>2]|0;
        $245 = ((($242)) + 16|0);
        $246 = HEAP32[$245>>2]|0;
        $247 = ($244|0)==($246|0);
        if ($247) {
         $248 = HEAP32[$242>>2]|0;
         $249 = ((($248)) + 40|0);
         $250 = HEAP32[$249>>2]|0;
         __THREW__ = 0;
         (invoke_ii($250|0,($242|0))|0);
         $251 = __THREW__; __THREW__ = 0;
         $252 = $251&1;
         if ($252) {
          label = 38;
          break L24;
         }
        } else {
         $253 = ((($244)) + 1|0);
         HEAP32[$243>>2] = $253;
        }
        $254 = HEAP8[$__psn>>0]|0;
        $255 = $254 & 1;
        $256 = ($255<<24>>24)==(0);
        $257 = HEAP32[$18>>2]|0;
        $258 = $254&255;
        $259 = $258 >>> 1;
        $260 = $256 ? $259 : $257;
        $261 = ($260>>>0)>(1);
        $__psn$__trailing_sign$0 = $261 ? $__psn : $__trailing_sign$0;
        $869 = $867;$__trailing_sign$1 = $__psn$__trailing_sign$0;
        break L47;
       }
       if ($225) {
        $304 = HEAP32[$220>>2]|0;
        $305 = ((($304)) + 36|0);
        $306 = HEAP32[$305>>2]|0;
        __THREW__ = 0;
        $307 = (invoke_ii($306|0,($220|0))|0);
        $308 = __THREW__; __THREW__ = 0;
        $309 = $308&1;
        if ($309) {
         label = 38;
         break L24;
        }
        $$pre298 = HEAP8[$__psn>>0]|0;
        $$pre299 = HEAP32[$__b>>2]|0;
        $$phi$trans$insert = ((($$pre299)) + 12|0);
        $$pre300 = HEAP32[$$phi$trans$insert>>2]|0;
        $$phi$trans$insert301 = ((($$pre299)) + 16|0);
        $$pre302 = HEAP32[$$phi$trans$insert301>>2]|0;
        $$0$i$i136 = $307;$314 = $$pre298;$321 = $$pre299;$323 = $$pre300;$324 = $$pre302;
       } else {
        $310 = HEAP8[$222>>0]|0;
        $311 = $310&255;
        $$0$i$i136 = $311;$314 = $202;$321 = $220;$323 = $222;$324 = $224;
       }
       $312 = $$0$i$i136&255;
       $313 = $314 & 1;
       $315 = ($313<<24>>24)==(0);
       $316 = HEAP32[$22>>2]|0;
       $317 = $315 ? $23 : $316;
       $318 = HEAP8[$317>>0]|0;
       $319 = ($312<<24>>24)==($318<<24>>24);
       $320 = ((($321)) + 12|0);
       $322 = ($323|0)==($324|0);
       if ($319) {
        if ($322) {
         $325 = HEAP32[$321>>2]|0;
         $326 = ((($325)) + 40|0);
         $327 = HEAP32[$326>>2]|0;
         __THREW__ = 0;
         (invoke_ii($327|0,($321|0))|0);
         $328 = __THREW__; __THREW__ = 0;
         $329 = $328&1;
         if ($329) {
          label = 38;
          break L24;
         }
        } else {
         $330 = ((($323)) + 1|0);
         HEAP32[$320>>2] = $330;
        }
        $331 = HEAP8[$__psn>>0]|0;
        $332 = $331 & 1;
        $333 = ($332<<24>>24)==(0);
        $334 = HEAP32[$18>>2]|0;
        $335 = $331&255;
        $336 = $335 >>> 1;
        $337 = $333 ? $336 : $334;
        $338 = ($337>>>0)>(1);
        $__psn$__trailing_sign$022 = $338 ? $__psn : $__trailing_sign$0;
        $869 = $867;$__trailing_sign$1 = $__psn$__trailing_sign$022;
        break L47;
       }
       if ($322) {
        $339 = HEAP32[$321>>2]|0;
        $340 = ((($339)) + 36|0);
        $341 = HEAP32[$340>>2]|0;
        __THREW__ = 0;
        $342 = (invoke_ii($341|0,($321|0))|0);
        $343 = __THREW__; __THREW__ = 0;
        $344 = $343&1;
        if ($344) {
         label = 38;
         break L24;
        } else {
         $$0$i$i145 = $342;
        }
       } else {
        $345 = HEAP8[$323>>0]|0;
        $346 = $345&255;
        $$0$i$i145 = $346;
       }
       $347 = $$0$i$i145&255;
       $348 = HEAP8[$__nsn>>0]|0;
       $349 = $348 & 1;
       $350 = ($349<<24>>24)==(0);
       $351 = HEAP32[$20>>2]|0;
       $352 = $350 ? $21 : $351;
       $353 = HEAP8[$352>>0]|0;
       $354 = ($347<<24>>24)==($353<<24>>24);
       if (!($354)) {
        label = 120;
        break L24;
       }
       $355 = HEAP32[$__b>>2]|0;
       $356 = ((($355)) + 12|0);
       $357 = HEAP32[$356>>2]|0;
       $358 = ((($355)) + 16|0);
       $359 = HEAP32[$358>>2]|0;
       $360 = ($357|0)==($359|0);
       if ($360) {
        $361 = HEAP32[$355>>2]|0;
        $362 = ((($361)) + 40|0);
        $363 = HEAP32[$362>>2]|0;
        __THREW__ = 0;
        (invoke_ii($363|0,($355|0))|0);
        $364 = __THREW__; __THREW__ = 0;
        $365 = $364&1;
        if ($365) {
         label = 38;
         break L24;
        }
       } else {
        $366 = ((($357)) + 1|0);
        HEAP32[$356>>2] = $366;
       }
       HEAP8[$__neg>>0] = 1;
       $367 = HEAP8[$__nsn>>0]|0;
       $368 = $367 & 1;
       $369 = ($368<<24>>24)==(0);
       $370 = HEAP32[$19>>2]|0;
       $371 = $367&255;
       $372 = $371 >>> 1;
       $373 = $369 ? $372 : $370;
       $374 = ($373>>>0)>(1);
       $__nsn$__trailing_sign$023 = $374 ? $__nsn : $__trailing_sign$0;
       $869 = $867;$__trailing_sign$1 = $__nsn$__trailing_sign$023;
      }
      break;
     }
     case 2:  {
      $377 = ($__trailing_sign$0|0)!=(0|0);
      $378 = ($__p$0>>>0)<(2);
      $or$cond = $378 | $377;
      if (!($or$cond)) {
       $379 = ($__p$0|0)==(2);
       $380 = HEAP8[$32>>0]|0;
       $381 = ($380<<24>>24)!=(0);
       $382 = $379 & $381;
       $brmerge = $25 | $382;
       if (!($brmerge)) {
        $869 = $867;$__trailing_sign$1 = 0;
        break L47;
       }
      }
      $383 = HEAP8[$__sym>>0]|0;
      $384 = $383 & 1;
      $385 = ($384<<24>>24)==(0);
      $386 = HEAP32[$26>>2]|0;
      $387 = $385 ? $27 : $386;
      $388 = $387;
      $389 = ($__p$0|0)==(0);
      L116: do {
       if ($389) {
        $$pre$i2$i$i180 = $868;$431 = $383;$437 = $386;$459 = $868;$__sym_curr_char$sroa$0$0 = $388;
       } else {
        $390 = (($__p$0) + -1)|0;
        $391 = (($__pat) + ($390)|0);
        $392 = HEAP8[$391>>0]|0;
        $switch = ($392&255)<(2);
        if ($switch) {
         $393 = HEAP32[$28>>2]|0;
         $394 = $383&255;
         $395 = $394 >>> 1;
         $396 = $385 ? $395 : $393;
         $397 = (($387) + ($396)|0);
         $__sym_space_end$sroa$0$0 = $388;
         while(1) {
          $398 = $__sym_space_end$sroa$0$0;
          $399 = ($398|0)==($397|0);
          if ($399) {
           $__sym_space_end$sroa$0$0$lcssa = $__sym_space_end$sroa$0$0;
           break;
          }
          $400 = HEAP8[$398>>0]|0;
          $isascii1$i186 = ($400<<24>>24)>(-1);
          if (!($isascii1$i186)) {
           $__sym_space_end$sroa$0$0$lcssa = $__sym_space_end$sroa$0$0;
           break;
          }
          $401 = $400 << 24 >> 24;
          $402 = HEAP32[$17>>2]|0;
          $403 = (($402) + ($401<<1)|0);
          $404 = HEAP16[$403>>1]|0;
          $405 = $404 & 8192;
          $406 = ($405<<16>>16)==(0);
          if ($406) {
           $__sym_space_end$sroa$0$0$lcssa = $__sym_space_end$sroa$0$0;
           break;
          }
          $407 = ((($398)) + 1|0);
          $408 = $407;
          $__sym_space_end$sroa$0$0 = $408;
         }
         $409 = (($__sym_space_end$sroa$0$0$lcssa) - ($388))|0;
         $410 = HEAP8[$__spaces>>0]|0;
         $411 = $410 & 1;
         $412 = ($411<<24>>24)==(0);
         $413 = HEAP32[$29>>2]|0;
         $414 = $410&255;
         $415 = $414 >>> 1;
         $416 = $412 ? $415 : $413;
         $417 = ($409>>>0)>($416>>>0);
         if ($417) {
          $$pre$i2$i$i180 = $868;$431 = $383;$437 = $386;$459 = $868;$__sym_curr_char$sroa$0$0 = $388;
         } else {
          $418 = HEAP32[$30>>2]|0;
          $419 = $412 ? $31 : $418;
          $420 = (($419) + ($416)|0);
          $421 = (0 - ($409))|0;
          $422 = (($420) + ($421)|0);
          $$sink = $387;$424 = $422;
          while(1) {
           $423 = ($424|0)==($420|0);
           if ($423) {
            $$pre$i2$i$i180 = $868;$431 = $383;$437 = $386;$459 = $868;$__sym_curr_char$sroa$0$0 = $__sym_space_end$sroa$0$0$lcssa;
            break L116;
           }
           $425 = HEAP8[$424>>0]|0;
           $426 = HEAP8[$$sink>>0]|0;
           $427 = ($425<<24>>24)==($426<<24>>24);
           if (!($427)) {
            $$pre$i2$i$i180 = $868;$431 = $383;$437 = $386;$459 = $868;$__sym_curr_char$sroa$0$0 = $388;
            break L116;
           }
           $428 = ((($424)) + 1|0);
           $429 = ((($$sink)) + 1|0);
           $$sink = $429;$424 = $428;
          }
         }
        } else {
         $$pre$i2$i$i180 = $868;$431 = $383;$437 = $386;$459 = $868;$__sym_curr_char$sroa$0$0 = $388;
        }
       }
      } while(0);
      L130: while(1) {
       $430 = $431 & 1;
       $432 = ($430<<24>>24)==(0);
       $433 = HEAP32[$28>>2]|0;
       $434 = $431&255;
       $435 = $434 >>> 1;
       $436 = $432 ? $27 : $437;
       $438 = $432 ? $435 : $433;
       $439 = (($436) + ($438)|0);
       $440 = $__sym_curr_char$sroa$0$0;
       $441 = ($440|0)==($439|0);
       if ($441) {
        $$lcssa = $439;
        break;
       }
       $442 = HEAP32[$__b>>2]|0;
       $443 = ($442|0)==(0|0);
       do {
        if ($443) {
         $457 = 0;
        } else {
         $444 = ((($442)) + 12|0);
         $445 = HEAP32[$444>>2]|0;
         $446 = ((($442)) + 16|0);
         $447 = HEAP32[$446>>2]|0;
         $448 = ($445|0)==($447|0);
         if ($448) {
          $449 = HEAP32[$442>>2]|0;
          $450 = ((($449)) + 36|0);
          $451 = HEAP32[$450>>2]|0;
          __THREW__ = 0;
          $452 = (invoke_ii($451|0,($442|0))|0);
          $453 = __THREW__; __THREW__ = 0;
          $454 = $453&1;
          if ($454) {
           label = 35;
           break L24;
          }
          $455 = ($452|0)==(-1);
          if ($455) {
           HEAP32[$__b>>2] = 0;
           $457 = 0;
           break;
          } else {
           $$pre$i$i$i176 = HEAP32[$__b>>2]|0;
           $457 = $$pre$i$i$i176;
           break;
          }
         } else {
          $457 = $442;
         }
        }
       } while(0);
       $456 = ($457|0)==(0|0);
       $458 = ($459|0)==(0|0);
       do {
        if ($458) {
         $$pre$i2$i$i180296 = $$pre$i2$i$i180;
         label = 148;
        } else {
         $460 = ((($459)) + 12|0);
         $461 = HEAP32[$460>>2]|0;
         $462 = ((($459)) + 16|0);
         $463 = HEAP32[$462>>2]|0;
         $464 = ($461|0)==($463|0);
         if (!($464)) {
          if ($456) {
           $$pre$i2$i$i180295 = $$pre$i2$i$i180;$871 = $459;
           break;
          } else {
           $$lcssa = $440;
           break L130;
          }
         }
         $465 = HEAP32[$459>>2]|0;
         $466 = ((($465)) + 36|0);
         $467 = HEAP32[$466>>2]|0;
         __THREW__ = 0;
         $468 = (invoke_ii($467|0,($459|0))|0);
         $469 = __THREW__; __THREW__ = 0;
         $470 = $469&1;
         if ($470) {
          label = 35;
          break L24;
         }
         $471 = ($468|0)==(-1);
         if ($471) {
          HEAP32[$__e>>2] = 0;
          $$pre$i2$i$i180296 = 0;
          label = 148;
          break;
         } else {
          $472 = ($$pre$i2$i$i180|0)==(0|0);
          $tmp$i$i182 = $456 ^ $472;
          if ($tmp$i$i182) {
           $$pre$i2$i$i180295 = $$pre$i2$i$i180;$871 = $$pre$i2$i$i180;
           break;
          } else {
           $$lcssa = $440;
           break L130;
          }
         }
        }
       } while(0);
       if ((label|0) == 148) {
        label = 0;
        if ($456) {
         $$lcssa = $440;
         break;
        } else {
         $$pre$i2$i$i180295 = $$pre$i2$i$i180296;$871 = 0;
        }
       }
       $473 = HEAP32[$__b>>2]|0;
       $474 = ((($473)) + 12|0);
       $475 = HEAP32[$474>>2]|0;
       $476 = ((($473)) + 16|0);
       $477 = HEAP32[$476>>2]|0;
       $478 = ($475|0)==($477|0);
       if ($478) {
        $479 = HEAP32[$473>>2]|0;
        $480 = ((($479)) + 36|0);
        $481 = HEAP32[$480>>2]|0;
        __THREW__ = 0;
        $482 = (invoke_ii($481|0,($473|0))|0);
        $483 = __THREW__; __THREW__ = 0;
        $484 = $483&1;
        if ($484) {
         label = 35;
         break L24;
        } else {
         $$0$i$i172 = $482;
        }
       } else {
        $485 = HEAP8[$475>>0]|0;
        $486 = $485&255;
        $$0$i$i172 = $486;
       }
       $487 = $$0$i$i172&255;
       $488 = HEAP8[$440>>0]|0;
       $489 = ($487<<24>>24)==($488<<24>>24);
       if (!($489)) {
        $$lcssa = $440;
        break;
       }
       $490 = HEAP32[$__b>>2]|0;
       $491 = ((($490)) + 12|0);
       $492 = HEAP32[$491>>2]|0;
       $493 = ((($490)) + 16|0);
       $494 = HEAP32[$493>>2]|0;
       $495 = ($492|0)==($494|0);
       if ($495) {
        $496 = HEAP32[$490>>2]|0;
        $497 = ((($496)) + 40|0);
        $498 = HEAP32[$497>>2]|0;
        __THREW__ = 0;
        (invoke_ii($498|0,($490|0))|0);
        $499 = __THREW__; __THREW__ = 0;
        $500 = $499&1;
        if ($500) {
         label = 35;
         break L24;
        }
       } else {
        $501 = ((($492)) + 1|0);
        HEAP32[$491>>2] = $501;
       }
       $502 = ((($440)) + 1|0);
       $503 = $502;
       $$pre293 = HEAP8[$__sym>>0]|0;
       $$pre294 = HEAP32[$26>>2]|0;
       $$pre$i2$i$i180 = $$pre$i2$i$i180295;$431 = $$pre293;$437 = $$pre294;$459 = $871;$__sym_curr_char$sroa$0$0 = $503;
      }
      if ($25) {
       $504 = HEAP8[$__sym>>0]|0;
       $505 = $504 & 1;
       $506 = ($505<<24>>24)==(0);
       $507 = HEAP32[$26>>2]|0;
       $508 = HEAP32[$28>>2]|0;
       $509 = $504&255;
       $510 = $509 >>> 1;
       $511 = $506 ? $27 : $507;
       $512 = $506 ? $510 : $508;
       $513 = (($511) + ($512)|0);
       $514 = ($$lcssa|0)==($513|0);
       if ($514) {
        $869 = $867;$__trailing_sign$1 = $__trailing_sign$0;
       } else {
        label = 160;
        break L24;
       }
      } else {
       $869 = $867;$__trailing_sign$1 = $__trailing_sign$0;
      }
      break;
     }
     case 4:  {
      $89 = HEAP8[$__ts>>0]|0;
      $$pre$i2$i$i163 = $868;$534 = $868;$589 = $867;$__ng$0 = 0;
      L165: while(1) {
       $517 = HEAP32[$__b>>2]|0;
       $518 = ($517|0)==(0|0);
       do {
        if ($518) {
         $532 = 0;
        } else {
         $519 = ((($517)) + 12|0);
         $520 = HEAP32[$519>>2]|0;
         $521 = ((($517)) + 16|0);
         $522 = HEAP32[$521>>2]|0;
         $523 = ($520|0)==($522|0);
         if ($523) {
          $524 = HEAP32[$517>>2]|0;
          $525 = ((($524)) + 36|0);
          $526 = HEAP32[$525>>2]|0;
          __THREW__ = 0;
          $527 = (invoke_ii($526|0,($517|0))|0);
          $528 = __THREW__; __THREW__ = 0;
          $529 = $528&1;
          if ($529) {
           label = 37;
           break L24;
          }
          $530 = ($527|0)==(-1);
          if ($530) {
           HEAP32[$__b>>2] = 0;
           $532 = 0;
           break;
          } else {
           $$pre$i$i$i159 = HEAP32[$__b>>2]|0;
           $532 = $$pre$i$i$i159;
           break;
          }
         } else {
          $532 = $517;
         }
        }
       } while(0);
       $531 = ($532|0)==(0|0);
       $533 = ($534|0)==(0|0);
       do {
        if ($533) {
         $$pre$i2$i$i163283 = $$pre$i2$i$i163;
         label = 173;
        } else {
         $535 = ((($534)) + 12|0);
         $536 = HEAP32[$535>>2]|0;
         $537 = ((($534)) + 16|0);
         $538 = HEAP32[$537>>2]|0;
         $539 = ($536|0)==($538|0);
         if (!($539)) {
          if ($531) {
           $$pre$i2$i$i163285 = $$pre$i2$i$i163;$872 = $534;
           break;
          } else {
           $$lcssa392 = $589;$$pre$i2$i$i123 = $$pre$i2$i$i163;$__ng$0$lcssa393 = $__ng$0;
           break L165;
          }
         }
         $540 = HEAP32[$534>>2]|0;
         $541 = ((($540)) + 36|0);
         $542 = HEAP32[$541>>2]|0;
         __THREW__ = 0;
         $543 = (invoke_ii($542|0,($534|0))|0);
         $544 = __THREW__; __THREW__ = 0;
         $545 = $544&1;
         if ($545) {
          label = 37;
          break L24;
         }
         $546 = ($543|0)==(-1);
         if ($546) {
          HEAP32[$__e>>2] = 0;
          $$pre$i2$i$i163283 = 0;
          label = 173;
          break;
         } else {
          $547 = ($$pre$i2$i$i163|0)==(0|0);
          $tmp$i$i165 = $531 ^ $547;
          if ($tmp$i$i165) {
           $$pre$i2$i$i163285 = $$pre$i2$i$i163;$872 = $$pre$i2$i$i163;
           break;
          } else {
           $$lcssa392 = $589;$$pre$i2$i$i123 = $$pre$i2$i$i163;$__ng$0$lcssa393 = $__ng$0;
           break L165;
          }
         }
        }
       } while(0);
       if ((label|0) == 173) {
        label = 0;
        if ($531) {
         $$lcssa392 = $589;$$pre$i2$i$i123 = $$pre$i2$i$i163283;$__ng$0$lcssa393 = $__ng$0;
         break;
        } else {
         $$pre$i2$i$i163285 = $$pre$i2$i$i163283;$872 = 0;
        }
       }
       $548 = HEAP32[$__b>>2]|0;
       $549 = ((($548)) + 12|0);
       $550 = HEAP32[$549>>2]|0;
       $551 = ((($548)) + 16|0);
       $552 = HEAP32[$551>>2]|0;
       $553 = ($550|0)==($552|0);
       if ($553) {
        $554 = HEAP32[$548>>2]|0;
        $555 = ((($554)) + 36|0);
        $556 = HEAP32[$555>>2]|0;
        __THREW__ = 0;
        $557 = (invoke_ii($556|0,($548|0))|0);
        $558 = __THREW__; __THREW__ = 0;
        $559 = $558&1;
        if ($559) {
         label = 37;
         break L24;
        } else {
         $$0$i$i155 = $557;
        }
       } else {
        $560 = HEAP8[$550>>0]|0;
        $561 = $560&255;
        $$0$i$i155 = $561;
       }
       $562 = $$0$i$i155&255;
       $isascii1$i153 = ($562<<24>>24)>(-1);
       if ($isascii1$i153) {
        $sext239 = $$0$i$i155 << 24;
        $563 = $sext239 >> 24;
        $564 = HEAP32[$17>>2]|0;
        $565 = (($564) + ($563<<1)|0);
        $566 = HEAP16[$565>>1]|0;
        $567 = $566 & 2048;
        $568 = ($567<<16>>16)==(0);
        if ($568) {
         label = 184;
        } else {
         $569 = HEAP32[$__wn>>2]|0;
         $570 = HEAP32[$0>>2]|0;
         $571 = ($569|0)==($570|0);
         if ($571) {
          __THREW__ = 0;
          invoke_viii(4903,($__wb|0),($__wn|0),($0|0));
          $572 = __THREW__; __THREW__ = 0;
          $573 = $572&1;
          if ($573) {
           label = 37;
           break L24;
          }
          $$pre = HEAP32[$__wn>>2]|0;
          $575 = $$pre;
         } else {
          $575 = $569;
         }
         $574 = ((($575)) + 1|0);
         HEAP32[$__wn>>2] = $574;
         HEAP8[$575>>0] = $562;
         $576 = (($__ng$0) + 1)|0;
         $873 = $589;$__ng$2$ph = $576;
        }
       } else {
        label = 184;
       }
       if ((label|0) == 184) {
        label = 0;
        $577 = HEAP8[$__grp>>0]|0;
        $578 = $577 & 1;
        $579 = ($578<<24>>24)==(0);
        $580 = HEAP32[$33>>2]|0;
        $581 = $577&255;
        $582 = $581 >>> 1;
        $583 = $579 ? $582 : $580;
        $584 = ($583|0)!=(0);
        $585 = ($__ng$0|0)!=(0);
        $or$cond12 = $585 & $584;
        $586 = ($562<<24>>24)==($89<<24>>24);
        $or$cond25 = $586 & $or$cond12;
        if (!($or$cond25)) {
         $$lcssa392 = $589;$$pre$i2$i$i123 = $$pre$i2$i$i163285;$__ng$0$lcssa393 = $__ng$0;
         break;
        }
        $587 = HEAP32[$__ge>>2]|0;
        $588 = ($589|0)==($587|0);
        if ($588) {
         __THREW__ = 0;
         invoke_viii(4904,($__gb|0),($__gn|0),($__ge|0));
         $590 = __THREW__; __THREW__ = 0;
         $591 = $590&1;
         if ($591) {
          label = 37;
          break L24;
         }
         $$pre288 = HEAP32[$__gn>>2]|0;
         $593 = $$pre288;
        } else {
         $593 = $589;
        }
        $592 = ((($593)) + 4|0);
        HEAP32[$__gn>>2] = $592;
        HEAP32[$593>>2] = $__ng$0;
        $873 = $592;$__ng$2$ph = 0;
       }
       $594 = HEAP32[$__b>>2]|0;
       $595 = ((($594)) + 12|0);
       $596 = HEAP32[$595>>2]|0;
       $597 = ((($594)) + 16|0);
       $598 = HEAP32[$597>>2]|0;
       $599 = ($596|0)==($598|0);
       if ($599) {
        $600 = HEAP32[$594>>2]|0;
        $601 = ((($600)) + 40|0);
        $602 = HEAP32[$601>>2]|0;
        __THREW__ = 0;
        (invoke_ii($602|0,($594|0))|0);
        $603 = __THREW__; __THREW__ = 0;
        $604 = $603&1;
        if ($604) {
         label = 37;
         break L24;
        } else {
         $$pre$i2$i$i163 = $$pre$i2$i$i163285;$534 = $872;$589 = $873;$__ng$0 = $__ng$2$ph;
         continue;
        }
       } else {
        $605 = ((($596)) + 1|0);
        HEAP32[$595>>2] = $605;
        $$pre$i2$i$i163 = $$pre$i2$i$i163285;$534 = $872;$589 = $873;$__ng$0 = $__ng$2$ph;
        continue;
       }
      }
      $606 = HEAP32[$__gb>>2]|0;
      $607 = ($606|0)!=($$lcssa392|0);
      $608 = ($__ng$0$lcssa393|0)!=(0);
      $or$cond14 = $608 & $607;
      if ($or$cond14) {
       $609 = HEAP32[$__ge>>2]|0;
       $610 = ($$lcssa392|0)==($609|0);
       if ($610) {
        __THREW__ = 0;
        invoke_viii(4904,($__gb|0),($__gn|0),($__ge|0));
        $611 = __THREW__; __THREW__ = 0;
        $612 = $611&1;
        if ($612) {
         label = 38;
         break L24;
        }
        $$pre290 = HEAP32[$__gn>>2]|0;
        $614 = $$pre290;
       } else {
        $614 = $$lcssa392;
       }
       $613 = ((($614)) + 4|0);
       HEAP32[$__gn>>2] = $613;
       HEAP32[$614>>2] = $__ng$0$lcssa393;
       $874 = $613;
      } else {
       $874 = $$lcssa392;
      }
      $615 = HEAP32[$__fd>>2]|0;
      $616 = ($615|0)>(0);
      L217: do {
       if ($616) {
        $617 = HEAP32[$__b>>2]|0;
        $618 = ($617|0)==(0|0);
        do {
         if ($618) {
          $632 = 0;
         } else {
          $619 = ((($617)) + 12|0);
          $620 = HEAP32[$619>>2]|0;
          $621 = ((($617)) + 16|0);
          $622 = HEAP32[$621>>2]|0;
          $623 = ($620|0)==($622|0);
          if ($623) {
           $624 = HEAP32[$617>>2]|0;
           $625 = ((($624)) + 36|0);
           $626 = HEAP32[$625>>2]|0;
           __THREW__ = 0;
           $627 = (invoke_ii($626|0,($617|0))|0);
           $628 = __THREW__; __THREW__ = 0;
           $629 = $628&1;
           if ($629) {
            label = 38;
            break L24;
           }
           $630 = ($627|0)==(-1);
           if ($630) {
            HEAP32[$__b>>2] = 0;
            $632 = 0;
            break;
           } else {
            $$pre$i$i$i119 = HEAP32[$__b>>2]|0;
            $632 = $$pre$i$i$i119;
            break;
           }
          } else {
           $632 = $617;
          }
         }
        } while(0);
        $631 = ($632|0)==(0|0);
        $633 = ($$pre$i2$i$i123|0)==(0|0);
        do {
         if ($633) {
          label = 210;
         } else {
          $634 = ((($$pre$i2$i$i123)) + 12|0);
          $635 = HEAP32[$634>>2]|0;
          $636 = ((($$pre$i2$i$i123)) + 16|0);
          $637 = HEAP32[$636>>2]|0;
          $638 = ($635|0)==($637|0);
          if ($638) {
           $639 = HEAP32[$$pre$i2$i$i123>>2]|0;
           $640 = ((($639)) + 36|0);
           $641 = HEAP32[$640>>2]|0;
           __THREW__ = 0;
           $642 = (invoke_ii($641|0,($$pre$i2$i$i123|0))|0);
           $643 = __THREW__; __THREW__ = 0;
           $644 = $643&1;
           if ($644) {
            label = 38;
            break L24;
           }
           $645 = ($642|0)==(-1);
           if ($645) {
            HEAP32[$__e>>2] = 0;
            label = 210;
            break;
           }
          }
          if ($631) {
           $875 = $$pre$i2$i$i123;
          } else {
           label = 215;
           break L24;
          }
         }
        } while(0);
        if ((label|0) == 210) {
         label = 0;
         if ($631) {
          label = 215;
          break L24;
         } else {
          $875 = 0;
         }
        }
        $646 = HEAP32[$__b>>2]|0;
        $647 = ((($646)) + 12|0);
        $648 = HEAP32[$647>>2]|0;
        $649 = ((($646)) + 16|0);
        $650 = HEAP32[$649>>2]|0;
        $651 = ($648|0)==($650|0);
        if ($651) {
         $652 = HEAP32[$646>>2]|0;
         $653 = ((($652)) + 36|0);
         $654 = HEAP32[$653>>2]|0;
         __THREW__ = 0;
         $655 = (invoke_ii($654|0,($646|0))|0);
         $656 = __THREW__; __THREW__ = 0;
         $657 = $656&1;
         if ($657) {
          label = 38;
          break L24;
         } else {
          $$0$i$i112 = $655;
         }
        } else {
         $658 = HEAP8[$648>>0]|0;
         $659 = $658&255;
         $$0$i$i112 = $659;
        }
        $660 = $$0$i$i112&255;
        $661 = HEAP8[$__dp>>0]|0;
        $662 = ($660<<24>>24)==($661<<24>>24);
        if (!($662)) {
         label = 215;
         break L24;
        }
        $665 = HEAP32[$__b>>2]|0;
        $666 = ((($665)) + 12|0);
        $667 = HEAP32[$666>>2]|0;
        $668 = ((($665)) + 16|0);
        $669 = HEAP32[$668>>2]|0;
        $670 = ($667|0)==($669|0);
        if ($670) {
         $671 = HEAP32[$665>>2]|0;
         $672 = ((($671)) + 40|0);
         $673 = HEAP32[$672>>2]|0;
         __THREW__ = 0;
         (invoke_ii($673|0,($665|0))|0);
         $674 = __THREW__; __THREW__ = 0;
         $675 = $674&1;
         if ($675) {
          label = 38;
          break L24;
         } else {
          $$pre$i2$i$i99 = $875;$678 = $615;$696 = $875;
         }
        } else {
         $676 = ((($667)) + 1|0);
         HEAP32[$666>>2] = $676;
         $$pre$i2$i$i99 = $875;$678 = $615;$696 = $875;
        }
        while(1) {
         $677 = ($678|0)>(0);
         if (!($677)) {
          break L217;
         }
         $679 = HEAP32[$__b>>2]|0;
         $680 = ($679|0)==(0|0);
         do {
          if ($680) {
           $694 = 0;
          } else {
           $681 = ((($679)) + 12|0);
           $682 = HEAP32[$681>>2]|0;
           $683 = ((($679)) + 16|0);
           $684 = HEAP32[$683>>2]|0;
           $685 = ($682|0)==($684|0);
           if ($685) {
            $686 = HEAP32[$679>>2]|0;
            $687 = ((($686)) + 36|0);
            $688 = HEAP32[$687>>2]|0;
            __THREW__ = 0;
            $689 = (invoke_ii($688|0,($679|0))|0);
            $690 = __THREW__; __THREW__ = 0;
            $691 = $690&1;
            if ($691) {
             label = 36;
             break L24;
            }
            $692 = ($689|0)==(-1);
            if ($692) {
             HEAP32[$__b>>2] = 0;
             $694 = 0;
             break;
            } else {
             $$pre$i$i$i95 = HEAP32[$__b>>2]|0;
             $694 = $$pre$i$i$i95;
             break;
            }
           } else {
            $694 = $679;
           }
          }
         } while(0);
         $693 = ($694|0)==(0|0);
         $695 = ($696|0)==(0|0);
         do {
          if ($695) {
           $$pre$i2$i$i99291$ph = $$pre$i2$i$i99;
           label = 232;
          } else {
           $697 = ((($696)) + 12|0);
           $698 = HEAP32[$697>>2]|0;
           $699 = ((($696)) + 16|0);
           $700 = HEAP32[$699>>2]|0;
           $701 = ($698|0)==($700|0);
           if (!($701)) {
            if ($693) {
             $$pre$i2$i$i99291310 = $$pre$i2$i$i99;$876 = $696;
             break;
            } else {
             label = 239;
             break L24;
            }
           }
           $702 = HEAP32[$696>>2]|0;
           $703 = ((($702)) + 36|0);
           $704 = HEAP32[$703>>2]|0;
           __THREW__ = 0;
           $705 = (invoke_ii($704|0,($696|0))|0);
           $706 = __THREW__; __THREW__ = 0;
           $707 = $706&1;
           if ($707) {
            label = 36;
            break L24;
           }
           $708 = ($705|0)==(-1);
           if ($708) {
            HEAP32[$__e>>2] = 0;
            $$pre$i2$i$i99291$ph = 0;
            label = 232;
            break;
           } else {
            $709 = ($$pre$i2$i$i99|0)==(0|0);
            $tmp$i$i101 = $693 ^ $709;
            if ($tmp$i$i101) {
             $$pre$i2$i$i99291310 = $$pre$i2$i$i99;$876 = $$pre$i2$i$i99;
             break;
            } else {
             label = 239;
             break L24;
            }
           }
          }
         } while(0);
         if ((label|0) == 232) {
          label = 0;
          if ($693) {
           label = 239;
           break L24;
          } else {
           $$pre$i2$i$i99291310 = $$pre$i2$i$i99291$ph;$876 = 0;
          }
         }
         $710 = HEAP32[$__b>>2]|0;
         $711 = ((($710)) + 12|0);
         $712 = HEAP32[$711>>2]|0;
         $713 = ((($710)) + 16|0);
         $714 = HEAP32[$713>>2]|0;
         $715 = ($712|0)==($714|0);
         if ($715) {
          $716 = HEAP32[$710>>2]|0;
          $717 = ((($716)) + 36|0);
          $718 = HEAP32[$717>>2]|0;
          __THREW__ = 0;
          $719 = (invoke_ii($718|0,($710|0))|0);
          $720 = __THREW__; __THREW__ = 0;
          $721 = $720&1;
          if ($721) {
           label = 36;
           break L24;
          } else {
           $$0$i$i90 = $719;
          }
         } else {
          $722 = HEAP8[$712>>0]|0;
          $723 = $722&255;
          $$0$i$i90 = $723;
         }
         $724 = $$0$i$i90&255;
         $isascii1$i88 = ($724<<24>>24)>(-1);
         if (!($isascii1$i88)) {
          label = 239;
          break L24;
         }
         $sext = $$0$i$i90 << 24;
         $725 = $sext >> 24;
         $726 = HEAP32[$17>>2]|0;
         $727 = (($726) + ($725<<1)|0);
         $728 = HEAP16[$727>>1]|0;
         $729 = $728 & 2048;
         $730 = ($729<<16>>16)==(0);
         if ($730) {
          label = 239;
          break L24;
         }
         $733 = HEAP32[$__wn>>2]|0;
         $734 = HEAP32[$0>>2]|0;
         $735 = ($733|0)==($734|0);
         if ($735) {
          __THREW__ = 0;
          invoke_viii(4903,($__wb|0),($__wn|0),($0|0));
          $736 = __THREW__; __THREW__ = 0;
          $737 = $736&1;
          if ($737) {
           label = 36;
           break L24;
          }
         }
         $738 = HEAP32[$__b>>2]|0;
         $739 = ((($738)) + 12|0);
         $740 = HEAP32[$739>>2]|0;
         $741 = ((($738)) + 16|0);
         $742 = HEAP32[$741>>2]|0;
         $743 = ($740|0)==($742|0);
         if ($743) {
          $744 = HEAP32[$738>>2]|0;
          $745 = ((($744)) + 36|0);
          $746 = HEAP32[$745>>2]|0;
          __THREW__ = 0;
          $747 = (invoke_ii($746|0,($738|0))|0);
          $748 = __THREW__; __THREW__ = 0;
          $749 = $748&1;
          if ($749) {
           label = 36;
           break L24;
          } else {
           $$0$i$i83 = $747;
          }
         } else {
          $750 = HEAP8[$740>>0]|0;
          $751 = $750&255;
          $$0$i$i83 = $751;
         }
         $752 = $$0$i$i83&255;
         $753 = HEAP32[$__wn>>2]|0;
         $754 = ((($753)) + 1|0);
         HEAP32[$__wn>>2] = $754;
         HEAP8[$753>>0] = $752;
         $755 = (($678) + -1)|0;
         HEAP32[$__fd>>2] = $755;
         $756 = HEAP32[$__b>>2]|0;
         $757 = ((($756)) + 12|0);
         $758 = HEAP32[$757>>2]|0;
         $759 = ((($756)) + 16|0);
         $760 = HEAP32[$759>>2]|0;
         $761 = ($758|0)==($760|0);
         if ($761) {
          $762 = HEAP32[$756>>2]|0;
          $763 = ((($762)) + 40|0);
          $764 = HEAP32[$763>>2]|0;
          __THREW__ = 0;
          (invoke_ii($764|0,($756|0))|0);
          $765 = __THREW__; __THREW__ = 0;
          $766 = $765&1;
          if ($766) {
           label = 36;
           break L24;
          } else {
           $$pre$i2$i$i99 = $$pre$i2$i$i99291310;$678 = $755;$696 = $876;
           continue;
          }
         } else {
          $767 = ((($758)) + 1|0);
          HEAP32[$757>>2] = $767;
          $$pre$i2$i$i99 = $$pre$i2$i$i99291310;$678 = $755;$696 = $876;
          continue;
         }
        }
       }
      } while(0);
      $768 = HEAP32[$__wn>>2]|0;
      $769 = HEAP32[$__wb>>2]|0;
      $770 = ($768|0)==($769|0);
      if ($770) {
       label = 249;
       break L24;
      } else {
       $869 = $874;$__trailing_sign$1 = $__trailing_sign$0;
      }
      break;
     }
     default: {
      $869 = $867;$__trailing_sign$1 = $__trailing_sign$0;
     }
     }
    } while(0);
    L285: do {
     if ((label|0) == 58) {
      while(1) {
       label = 0;
       $132 = HEAP32[$__b>>2]|0;
       $133 = ($132|0)==(0|0);
       do {
        if ($133) {
         $147 = 0;
        } else {
         $134 = ((($132)) + 12|0);
         $135 = HEAP32[$134>>2]|0;
         $136 = ((($132)) + 16|0);
         $137 = HEAP32[$136>>2]|0;
         $138 = ($135|0)==($137|0);
         if ($138) {
          $139 = HEAP32[$132>>2]|0;
          $140 = ((($139)) + 36|0);
          $141 = HEAP32[$140>>2]|0;
          __THREW__ = 0;
          $142 = (invoke_ii($141|0,($132|0))|0);
          $143 = __THREW__; __THREW__ = 0;
          $144 = $143&1;
          if ($144) {
           label = 34;
           break L24;
          }
          $145 = ($142|0)==(-1);
          if ($145) {
           HEAP32[$__b>>2] = 0;
           $147 = 0;
           break;
          } else {
           $$pre$i$i$i63 = HEAP32[$__b>>2]|0;
           $147 = $$pre$i$i$i63;
           break;
          }
         } else {
          $147 = $132;
         }
        }
       } while(0);
       $146 = ($147|0)==(0|0);
       $148 = ($149|0)==(0|0);
       do {
        if ($148) {
         $$pre$i2$i$i67305 = $$pre$i2$i$i67;
         label = 70;
        } else {
         $150 = ((($149)) + 12|0);
         $151 = HEAP32[$150>>2]|0;
         $152 = ((($149)) + 16|0);
         $153 = HEAP32[$152>>2]|0;
         $154 = ($151|0)==($153|0);
         if (!($154)) {
          if ($146) {
           $$pre$i2$i$i67307 = $$pre$i2$i$i67;$870 = $149;
           break;
          } else {
           $869 = $867;$__trailing_sign$1 = $__trailing_sign$0;
           break L285;
          }
         }
         $155 = HEAP32[$149>>2]|0;
         $156 = ((($155)) + 36|0);
         $157 = HEAP32[$156>>2]|0;
         __THREW__ = 0;
         $158 = (invoke_ii($157|0,($149|0))|0);
         $159 = __THREW__; __THREW__ = 0;
         $160 = $159&1;
         if ($160) {
          label = 34;
          break L24;
         }
         $161 = ($158|0)==(-1);
         if ($161) {
          HEAP32[$__e>>2] = 0;
          $$pre$i2$i$i67305 = 0;
          label = 70;
          break;
         } else {
          $162 = ($$pre$i2$i$i67|0)==(0|0);
          $tmp$i$i69 = $146 ^ $162;
          if ($tmp$i$i69) {
           $$pre$i2$i$i67307 = $$pre$i2$i$i67;$870 = $$pre$i2$i$i67;
           break;
          } else {
           $869 = $867;$__trailing_sign$1 = $__trailing_sign$0;
           break L285;
          }
         }
        }
       } while(0);
       if ((label|0) == 70) {
        label = 0;
        if ($146) {
         $869 = $867;$__trailing_sign$1 = $__trailing_sign$0;
         break L285;
        } else {
         $$pre$i2$i$i67307 = $$pre$i2$i$i67305;$870 = 0;
        }
       }
       $163 = HEAP32[$__b>>2]|0;
       $164 = ((($163)) + 12|0);
       $165 = HEAP32[$164>>2]|0;
       $166 = ((($163)) + 16|0);
       $167 = HEAP32[$166>>2]|0;
       $168 = ($165|0)==($167|0);
       if ($168) {
        $169 = HEAP32[$163>>2]|0;
        $170 = ((($169)) + 36|0);
        $171 = HEAP32[$170>>2]|0;
        __THREW__ = 0;
        $172 = (invoke_ii($171|0,($163|0))|0);
        $173 = __THREW__; __THREW__ = 0;
        $174 = $173&1;
        if ($174) {
         label = 34;
         break L24;
        } else {
         $$0$i$i73 = $172;
        }
       } else {
        $175 = HEAP8[$165>>0]|0;
        $176 = $175&255;
        $$0$i$i73 = $176;
       }
       $177 = $$0$i$i73&255;
       $isascii1$i76 = ($177<<24>>24)>(-1);
       if (!($isascii1$i76)) {
        $869 = $867;$__trailing_sign$1 = $__trailing_sign$0;
        break L285;
       }
       $sext240 = $$0$i$i73 << 24;
       $178 = $sext240 >> 24;
       $179 = HEAP32[$17>>2]|0;
       $180 = (($179) + ($178<<1)|0);
       $181 = HEAP16[$180>>1]|0;
       $182 = $181 & 8192;
       $183 = ($182<<16>>16)==(0);
       if ($183) {
        $869 = $867;$__trailing_sign$1 = $__trailing_sign$0;
        break L285;
       }
       $184 = HEAP32[$__b>>2]|0;
       $185 = ((($184)) + 12|0);
       $186 = HEAP32[$185>>2]|0;
       $187 = ((($184)) + 16|0);
       $188 = HEAP32[$187>>2]|0;
       $189 = ($186|0)==($188|0);
       if ($189) {
        $190 = HEAP32[$184>>2]|0;
        $191 = ((($190)) + 40|0);
        $192 = HEAP32[$191>>2]|0;
        __THREW__ = 0;
        $193 = (invoke_ii($192|0,($184|0))|0);
        $194 = __THREW__; __THREW__ = 0;
        $195 = $194&1;
        if ($195) {
         label = 34;
         break L24;
        } else {
         $$0$i$i80 = $193;
        }
       } else {
        $196 = ((($186)) + 1|0);
        HEAP32[$185>>2] = $196;
        $197 = HEAP8[$186>>0]|0;
        $198 = $197&255;
        $$0$i$i80 = $198;
       }
       $199 = $$0$i$i80&255;
       __THREW__ = 0;
       invoke_vii(3370,($__spaces|0),($199|0));
       $200 = __THREW__; __THREW__ = 0;
       $201 = $200&1;
       if ($201) {
        label = 34;
        break L24;
       } else {
        $$pre$i2$i$i67 = $$pre$i2$i$i67307;$149 = $870;
        label = 58;
       }
      }
     }
    } while(0);
    $773 = (($__p$0) + 1)|0;
    $867 = $869;$__p$0 = $773;$__trailing_sign$0 = $__trailing_sign$1;
   }
   L319:    switch (label|0) {
    case 34: {
     $67 = ___cxa_find_matching_catch_2()|0;
     $68 = tempRet0;
     $lpad$phi$index = $67;$lpad$phi$index7 = $68;
     break L21;
     break;
    }
    case 35: {
     $69 = ___cxa_find_matching_catch_2()|0;
     $70 = tempRet0;
     $lpad$phi$index = $69;$lpad$phi$index7 = $70;
     break L21;
     break;
    }
    case 36: {
     $71 = ___cxa_find_matching_catch_2()|0;
     $72 = tempRet0;
     $lpad$phi$index = $71;$lpad$phi$index7 = $72;
     break L21;
     break;
    }
    case 37: {
     $73 = ___cxa_find_matching_catch_2()|0;
     $74 = tempRet0;
     $lpad$phi$index = $73;$lpad$phi$index7 = $74;
     break L21;
     break;
    }
    case 38: {
     $75 = ___cxa_find_matching_catch_2()|0;
     $76 = tempRet0;
     $lpad$phi$index = $75;$lpad$phi$index7 = $76;
     break L21;
     break;
    }
    case 56: {
     $129 = HEAP32[$__err>>2]|0;
     $130 = $129 | 4;
     HEAP32[$__err>>2] = $130;
     $$10 = 0;
     break;
    }
    case 120: {
     $375 = HEAP32[$__err>>2]|0;
     $376 = $375 | 4;
     HEAP32[$__err>>2] = $376;
     $$10 = 0;
     break;
    }
    case 160: {
     $515 = HEAP32[$__err>>2]|0;
     $516 = $515 | 4;
     HEAP32[$__err>>2] = $516;
     $$10 = 0;
     break;
    }
    case 215: {
     $663 = HEAP32[$__err>>2]|0;
     $664 = $663 | 4;
     HEAP32[$__err>>2] = $664;
     $$10 = 0;
     break;
    }
    case 239: {
     $731 = HEAP32[$__err>>2]|0;
     $732 = $731 | 4;
     HEAP32[$__err>>2] = $732;
     $$10 = 0;
     break;
    }
    case 249: {
     $771 = HEAP32[$__err>>2]|0;
     $772 = $771 | 4;
     HEAP32[$__err>>2] = $772;
     $$10 = 0;
     break;
    }
    case 251: {
     $774 = ($__trailing_sign$0$lcssa|0)==(0|0);
     L321: do {
      if (!($774)) {
       $775 = ((($__trailing_sign$0$lcssa)) + 1|0);
       $776 = ((($__trailing_sign$0$lcssa)) + 8|0);
       $777 = ((($__trailing_sign$0$lcssa)) + 4|0);
       $__i$0 = 1;
       L323: while(1) {
        $778 = HEAP8[$__trailing_sign$0$lcssa>>0]|0;
        $779 = $778 & 1;
        $780 = ($779<<24>>24)==(0);
        if ($780) {
         $782 = $778&255;
         $783 = $782 >>> 1;
         $785 = $783;
        } else {
         $781 = HEAP32[$777>>2]|0;
         $785 = $781;
        }
        $784 = ($__i$0>>>0)<($785>>>0);
        if (!($784)) {
         break L321;
        }
        $786 = HEAP32[$__b>>2]|0;
        $787 = ($786|0)==(0|0);
        do {
         if ($787) {
          $801 = 0;
         } else {
          $788 = ((($786)) + 12|0);
          $789 = HEAP32[$788>>2]|0;
          $790 = ((($786)) + 16|0);
          $791 = HEAP32[$790>>2]|0;
          $792 = ($789|0)==($791|0);
          if ($792) {
           $793 = HEAP32[$786>>2]|0;
           $794 = ((($793)) + 36|0);
           $795 = HEAP32[$794>>2]|0;
           __THREW__ = 0;
           $796 = (invoke_ii($795|0,($786|0))|0);
           $797 = __THREW__; __THREW__ = 0;
           $798 = $797&1;
           if ($798) {
            label = 33;
            break L323;
           }
           $799 = ($796|0)==(-1);
           if ($799) {
            HEAP32[$__b>>2] = 0;
            $801 = 0;
            break;
           } else {
            $$pre$i$i$i = HEAP32[$__b>>2]|0;
            $801 = $$pre$i$i$i;
            break;
           }
          } else {
           $801 = $786;
          }
         }
        } while(0);
        $800 = ($801|0)==(0|0);
        $802 = HEAP32[$__e>>2]|0;
        $803 = ($802|0)==(0|0);
        do {
         if ($803) {
          label = 269;
         } else {
          $804 = ((($802)) + 12|0);
          $805 = HEAP32[$804>>2]|0;
          $806 = ((($802)) + 16|0);
          $807 = HEAP32[$806>>2]|0;
          $808 = ($805|0)==($807|0);
          if ($808) {
           $809 = HEAP32[$802>>2]|0;
           $810 = ((($809)) + 36|0);
           $811 = HEAP32[$810>>2]|0;
           __THREW__ = 0;
           $812 = (invoke_ii($811|0,($802|0))|0);
           $813 = __THREW__; __THREW__ = 0;
           $814 = $813&1;
           if ($814) {
            label = 33;
            break L323;
           }
           $815 = ($812|0)==(-1);
           if ($815) {
            HEAP32[$__e>>2] = 0;
            label = 269;
            break;
           }
          }
          if (!($800)) {
           label = 276;
           break L323;
          }
         }
        } while(0);
        if ((label|0) == 269) {
         label = 0;
         if ($800) {
          label = 276;
          break;
         }
        }
        $816 = HEAP32[$__b>>2]|0;
        $817 = ((($816)) + 12|0);
        $818 = HEAP32[$817>>2]|0;
        $819 = ((($816)) + 16|0);
        $820 = HEAP32[$819>>2]|0;
        $821 = ($818|0)==($820|0);
        if ($821) {
         $822 = HEAP32[$816>>2]|0;
         $823 = ((($822)) + 36|0);
         $824 = HEAP32[$823>>2]|0;
         __THREW__ = 0;
         $825 = (invoke_ii($824|0,($816|0))|0);
         $826 = __THREW__; __THREW__ = 0;
         $827 = $826&1;
         if ($827) {
          label = 33;
          break;
         } else {
          $$0$i$i = $825;
         }
        } else {
         $828 = HEAP8[$818>>0]|0;
         $829 = $828&255;
         $$0$i$i = $829;
        }
        $830 = $$0$i$i&255;
        $831 = HEAP8[$__trailing_sign$0$lcssa>>0]|0;
        $832 = $831 & 1;
        $833 = ($832<<24>>24)==(0);
        if ($833) {
         $836 = $775;
        } else {
         $834 = HEAP32[$776>>2]|0;
         $836 = $834;
        }
        $835 = (($836) + ($__i$0)|0);
        $837 = HEAP8[$835>>0]|0;
        $838 = ($830<<24>>24)==($837<<24>>24);
        if (!($838)) {
         label = 276;
         break;
        }
        $841 = (($__i$0) + 1)|0;
        $842 = HEAP32[$__b>>2]|0;
        $843 = ((($842)) + 12|0);
        $844 = HEAP32[$843>>2]|0;
        $845 = ((($842)) + 16|0);
        $846 = HEAP32[$845>>2]|0;
        $847 = ($844|0)==($846|0);
        if ($847) {
         $848 = HEAP32[$842>>2]|0;
         $849 = ((($848)) + 40|0);
         $850 = HEAP32[$849>>2]|0;
         __THREW__ = 0;
         (invoke_ii($850|0,($842|0))|0);
         $851 = __THREW__; __THREW__ = 0;
         $852 = $851&1;
         if ($852) {
          label = 33;
          break;
         } else {
          $__i$0 = $841;
          continue;
         }
        } else {
         $853 = ((($844)) + 1|0);
         HEAP32[$843>>2] = $853;
         $__i$0 = $841;
         continue;
        }
       }
       if ((label|0) == 33) {
        $65 = ___cxa_find_matching_catch_2()|0;
        $66 = tempRet0;
        $lpad$phi$index = $65;$lpad$phi$index7 = $66;
        break L21;
       }
       else if ((label|0) == 276) {
        $839 = HEAP32[$__err>>2]|0;
        $840 = $839 | 4;
        HEAP32[$__err>>2] = $840;
        $$10 = 0;
        break L319;
       }
      }
     } while(0);
     $854 = HEAP32[$__gb>>2]|0;
     $855 = ($854|0)==($$lcssa394|0);
     if ($855) {
      $$10 = 1;
     } else {
      HEAP32[$__et>>2] = 0;
      __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grp,$854,$$lcssa394,$__et);
      $856 = HEAP32[$__et>>2]|0;
      $857 = ($856|0)==(0);
      if ($857) {
       $$10 = 1;
       break;
      } else {
       $858 = HEAP32[$__err>>2]|0;
       $859 = $858 | 4;
       HEAP32[$__err>>2] = $859;
       $$10 = 0;
       break;
      }
     }
     break;
    }
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__spaces);
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__nsn);
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__psn);
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__sym);
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grp);
   $860 = HEAP32[$__gb>>2]|0;
   HEAP32[$__gb>>2] = 0;
   $861 = ($860|0)==(0|0);
   if (!($861)) {
    $862 = HEAP32[$2>>2]|0;
    __THREW__ = 0;
    invoke_vi($862|0,($860|0));
    $863 = __THREW__; __THREW__ = 0;
    $864 = $863&1;
    if ($864) {
     $865 = ___cxa_find_matching_catch_3(0|0)|0;
     $866 = tempRet0;
     ___clang_call_terminate($865);
     // unreachable;
    }
   }
   STACKTOP = sp;return ($$10|0);
  }
 } while(0);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__spaces);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__nsn);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__psn);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__sym);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grp);
 $79 = HEAP32[$__gb>>2]|0;
 HEAP32[$__gb>>2] = 0;
 $80 = ($79|0)==(0|0);
 if (!($80)) {
  $81 = HEAP32[$2>>2]|0;
  __THREW__ = 0;
  invoke_vi($81|0,($79|0));
  $82 = __THREW__; __THREW__ = 0;
  $83 = $82&1;
  if ($83) {
   $84 = ___cxa_find_matching_catch_3(0|0)|0;
   $85 = tempRet0;
   ___clang_call_terminate($84);
   // unreachable;
  }
 }
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__111__money_getIcE13__gather_infoEbRKNS_6localeERNS_10money_base7patternERcS8_RNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEESF_SF_SF_Ri($__intl,$__loc,$__pat,$__dp,$__ts,$__grp,$__sym,$__psn,$__nsn,$__fd) {
 $__intl = $__intl|0;
 $__loc = $__loc|0;
 $__pat = $__pat|0;
 $__dp = $__dp|0;
 $__ts = $__ts|0;
 $__grp = $__grp|0;
 $__sym = $__sym|0;
 $__psn = $__psn|0;
 $__nsn = $__nsn|0;
 $__fd = $__fd|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__i$0$i$i$i = 0, $__i$0$i$i$i14 = 0, $__i$0$i$i$i19 = 0, $__i$0$i$i$i24 = 0, $__i$0$i$i$i29 = 0, $__i$0$i$i$i34 = 0, $__i$0$i$i$i4 = 0, $__i$0$i$i$i9 = 0;
 var $exitcond$i$i$i = 0, $exitcond$i$i$i10 = 0, $exitcond$i$i$i15 = 0, $exitcond$i$i$i20 = 0, $exitcond$i$i$i25 = 0, $exitcond$i$i$i30 = 0, $exitcond$i$i$i35 = 0, $exitcond$i$i$i5 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 100|0;
 $1 = sp + 88|0;
 $2 = sp + 76|0;
 $3 = sp + 64|0;
 $4 = sp + 52|0;
 $5 = sp + 48|0;
 $6 = sp + 36|0;
 $7 = sp + 24|0;
 $8 = sp + 12|0;
 $9 = sp;
 if ($__intl) {
  $10 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1209296)|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ((($11)) + 44|0);
  $13 = HEAP32[$12>>2]|0;
  FUNCTION_TABLE_vii[$13 & 8191]($0,$10);
  $14 = HEAP32[$0>>2]|0;
  HEAP8[$__pat>>0]=$14&255;HEAP8[$__pat+1>>0]=($14>>8)&255;HEAP8[$__pat+2>>0]=($14>>16)&255;HEAP8[$__pat+3>>0]=$14>>24;
  $15 = HEAP32[$10>>2]|0;
  $16 = ((($15)) + 32|0);
  $17 = HEAP32[$16>>2]|0;
  FUNCTION_TABLE_vii[$17 & 8191]($1,$10);
  $18 = HEAP8[$__nsn>>0]|0;
  $19 = $18 & 1;
  $20 = ($19<<24>>24)==(0);
  if ($20) {
   $24 = ((($__nsn)) + 1|0);
   HEAP8[$24>>0] = 0;
   HEAP8[$__nsn>>0] = 0;
  } else {
   $21 = ((($__nsn)) + 8|0);
   $22 = HEAP32[$21>>2]|0;
   HEAP8[$22>>0] = 0;
   $23 = ((($__nsn)) + 4|0);
   HEAP32[$23>>2] = 0;
  }
  __THREW__ = 0;
  invoke_vii(3368,($__nsn|0),0);
  $25 = __THREW__; __THREW__ = 0;
  $26 = $25&1;
  if ($26) {
   $27 = ___cxa_find_matching_catch_3(0|0)|0;
   $28 = tempRet0;
   ___clang_call_terminate($27);
   // unreachable;
  }
  ;HEAP32[$__nsn>>2]=HEAP32[$1>>2]|0;HEAP32[$__nsn+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$__nsn+8>>2]=HEAP32[$1+8>>2]|0;
  $__i$0$i$i$i = 0;
  while(1) {
   $exitcond$i$i$i = ($__i$0$i$i$i|0)==(3);
   if ($exitcond$i$i$i) {
    break;
   }
   $29 = (($1) + ($__i$0$i$i$i<<2)|0);
   HEAP32[$29>>2] = 0;
   $30 = (($__i$0$i$i$i) + 1)|0;
   $__i$0$i$i$i = $30;
  }
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1);
  $31 = HEAP32[$10>>2]|0;
  $32 = ((($31)) + 28|0);
  $33 = HEAP32[$32>>2]|0;
  FUNCTION_TABLE_vii[$33 & 8191]($2,$10);
  $34 = HEAP8[$__psn>>0]|0;
  $35 = $34 & 1;
  $36 = ($35<<24>>24)==(0);
  if ($36) {
   $40 = ((($__psn)) + 1|0);
   HEAP8[$40>>0] = 0;
   HEAP8[$__psn>>0] = 0;
  } else {
   $37 = ((($__psn)) + 8|0);
   $38 = HEAP32[$37>>2]|0;
   HEAP8[$38>>0] = 0;
   $39 = ((($__psn)) + 4|0);
   HEAP32[$39>>2] = 0;
  }
  __THREW__ = 0;
  invoke_vii(3368,($__psn|0),0);
  $41 = __THREW__; __THREW__ = 0;
  $42 = $41&1;
  if ($42) {
   $43 = ___cxa_find_matching_catch_3(0|0)|0;
   $44 = tempRet0;
   ___clang_call_terminate($43);
   // unreachable;
  }
  ;HEAP32[$__psn>>2]=HEAP32[$2>>2]|0;HEAP32[$__psn+4>>2]=HEAP32[$2+4>>2]|0;HEAP32[$__psn+8>>2]=HEAP32[$2+8>>2]|0;
  $__i$0$i$i$i4 = 0;
  while(1) {
   $exitcond$i$i$i5 = ($__i$0$i$i$i4|0)==(3);
   if ($exitcond$i$i$i5) {
    break;
   }
   $45 = (($2) + ($__i$0$i$i$i4<<2)|0);
   HEAP32[$45>>2] = 0;
   $46 = (($__i$0$i$i$i4) + 1)|0;
   $__i$0$i$i$i4 = $46;
  }
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($2);
  $47 = HEAP32[$10>>2]|0;
  $48 = ((($47)) + 12|0);
  $49 = HEAP32[$48>>2]|0;
  $50 = (FUNCTION_TABLE_ii[$49 & 8191]($10)|0);
  HEAP8[$__dp>>0] = $50;
  $51 = HEAP32[$10>>2]|0;
  $52 = ((($51)) + 16|0);
  $53 = HEAP32[$52>>2]|0;
  $54 = (FUNCTION_TABLE_ii[$53 & 8191]($10)|0);
  HEAP8[$__ts>>0] = $54;
  $55 = HEAP32[$10>>2]|0;
  $56 = ((($55)) + 20|0);
  $57 = HEAP32[$56>>2]|0;
  FUNCTION_TABLE_vii[$57 & 8191]($3,$10);
  $58 = HEAP8[$__grp>>0]|0;
  $59 = $58 & 1;
  $60 = ($59<<24>>24)==(0);
  if ($60) {
   $64 = ((($__grp)) + 1|0);
   HEAP8[$64>>0] = 0;
   HEAP8[$__grp>>0] = 0;
  } else {
   $61 = ((($__grp)) + 8|0);
   $62 = HEAP32[$61>>2]|0;
   HEAP8[$62>>0] = 0;
   $63 = ((($__grp)) + 4|0);
   HEAP32[$63>>2] = 0;
  }
  __THREW__ = 0;
  invoke_vii(3368,($__grp|0),0);
  $65 = __THREW__; __THREW__ = 0;
  $66 = $65&1;
  if ($66) {
   $67 = ___cxa_find_matching_catch_3(0|0)|0;
   $68 = tempRet0;
   ___clang_call_terminate($67);
   // unreachable;
  }
  ;HEAP32[$__grp>>2]=HEAP32[$3>>2]|0;HEAP32[$__grp+4>>2]=HEAP32[$3+4>>2]|0;HEAP32[$__grp+8>>2]=HEAP32[$3+8>>2]|0;
  $__i$0$i$i$i9 = 0;
  while(1) {
   $exitcond$i$i$i10 = ($__i$0$i$i$i9|0)==(3);
   if ($exitcond$i$i$i10) {
    break;
   }
   $69 = (($3) + ($__i$0$i$i$i9<<2)|0);
   HEAP32[$69>>2] = 0;
   $70 = (($__i$0$i$i$i9) + 1)|0;
   $__i$0$i$i$i9 = $70;
  }
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3);
  $71 = HEAP32[$10>>2]|0;
  $72 = ((($71)) + 24|0);
  $73 = HEAP32[$72>>2]|0;
  FUNCTION_TABLE_vii[$73 & 8191]($4,$10);
  $74 = HEAP8[$__sym>>0]|0;
  $75 = $74 & 1;
  $76 = ($75<<24>>24)==(0);
  if ($76) {
   $80 = ((($__sym)) + 1|0);
   HEAP8[$80>>0] = 0;
   HEAP8[$__sym>>0] = 0;
  } else {
   $77 = ((($__sym)) + 8|0);
   $78 = HEAP32[$77>>2]|0;
   HEAP8[$78>>0] = 0;
   $79 = ((($__sym)) + 4|0);
   HEAP32[$79>>2] = 0;
  }
  __THREW__ = 0;
  invoke_vii(3368,($__sym|0),0);
  $81 = __THREW__; __THREW__ = 0;
  $82 = $81&1;
  if ($82) {
   $83 = ___cxa_find_matching_catch_3(0|0)|0;
   $84 = tempRet0;
   ___clang_call_terminate($83);
   // unreachable;
  }
  ;HEAP32[$__sym>>2]=HEAP32[$4>>2]|0;HEAP32[$__sym+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$__sym+8>>2]=HEAP32[$4+8>>2]|0;
  $__i$0$i$i$i14 = 0;
  while(1) {
   $exitcond$i$i$i15 = ($__i$0$i$i$i14|0)==(3);
   if ($exitcond$i$i$i15) {
    break;
   }
   $85 = (($4) + ($__i$0$i$i$i14<<2)|0);
   HEAP32[$85>>2] = 0;
   $86 = (($__i$0$i$i$i14) + 1)|0;
   $__i$0$i$i$i14 = $86;
  }
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4);
  $87 = HEAP32[$10>>2]|0;
  $88 = ((($87)) + 36|0);
  $89 = HEAP32[$88>>2]|0;
  $90 = (FUNCTION_TABLE_ii[$89 & 8191]($10)|0);
  $storemerge = $90;
 } else {
  $91 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1209288)|0);
  $92 = HEAP32[$91>>2]|0;
  $93 = ((($92)) + 44|0);
  $94 = HEAP32[$93>>2]|0;
  FUNCTION_TABLE_vii[$94 & 8191]($5,$91);
  $95 = HEAP32[$5>>2]|0;
  HEAP8[$__pat>>0]=$95&255;HEAP8[$__pat+1>>0]=($95>>8)&255;HEAP8[$__pat+2>>0]=($95>>16)&255;HEAP8[$__pat+3>>0]=$95>>24;
  $96 = HEAP32[$91>>2]|0;
  $97 = ((($96)) + 32|0);
  $98 = HEAP32[$97>>2]|0;
  FUNCTION_TABLE_vii[$98 & 8191]($6,$91);
  $99 = HEAP8[$__nsn>>0]|0;
  $100 = $99 & 1;
  $101 = ($100<<24>>24)==(0);
  if ($101) {
   $105 = ((($__nsn)) + 1|0);
   HEAP8[$105>>0] = 0;
   HEAP8[$__nsn>>0] = 0;
  } else {
   $102 = ((($__nsn)) + 8|0);
   $103 = HEAP32[$102>>2]|0;
   HEAP8[$103>>0] = 0;
   $104 = ((($__nsn)) + 4|0);
   HEAP32[$104>>2] = 0;
  }
  __THREW__ = 0;
  invoke_vii(3368,($__nsn|0),0);
  $106 = __THREW__; __THREW__ = 0;
  $107 = $106&1;
  if ($107) {
   $108 = ___cxa_find_matching_catch_3(0|0)|0;
   $109 = tempRet0;
   ___clang_call_terminate($108);
   // unreachable;
  }
  ;HEAP32[$__nsn>>2]=HEAP32[$6>>2]|0;HEAP32[$__nsn+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$__nsn+8>>2]=HEAP32[$6+8>>2]|0;
  $__i$0$i$i$i19 = 0;
  while(1) {
   $exitcond$i$i$i20 = ($__i$0$i$i$i19|0)==(3);
   if ($exitcond$i$i$i20) {
    break;
   }
   $110 = (($6) + ($__i$0$i$i$i19<<2)|0);
   HEAP32[$110>>2] = 0;
   $111 = (($__i$0$i$i$i19) + 1)|0;
   $__i$0$i$i$i19 = $111;
  }
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6);
  $112 = HEAP32[$91>>2]|0;
  $113 = ((($112)) + 28|0);
  $114 = HEAP32[$113>>2]|0;
  FUNCTION_TABLE_vii[$114 & 8191]($7,$91);
  $115 = HEAP8[$__psn>>0]|0;
  $116 = $115 & 1;
  $117 = ($116<<24>>24)==(0);
  if ($117) {
   $121 = ((($__psn)) + 1|0);
   HEAP8[$121>>0] = 0;
   HEAP8[$__psn>>0] = 0;
  } else {
   $118 = ((($__psn)) + 8|0);
   $119 = HEAP32[$118>>2]|0;
   HEAP8[$119>>0] = 0;
   $120 = ((($__psn)) + 4|0);
   HEAP32[$120>>2] = 0;
  }
  __THREW__ = 0;
  invoke_vii(3368,($__psn|0),0);
  $122 = __THREW__; __THREW__ = 0;
  $123 = $122&1;
  if ($123) {
   $124 = ___cxa_find_matching_catch_3(0|0)|0;
   $125 = tempRet0;
   ___clang_call_terminate($124);
   // unreachable;
  }
  ;HEAP32[$__psn>>2]=HEAP32[$7>>2]|0;HEAP32[$__psn+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$__psn+8>>2]=HEAP32[$7+8>>2]|0;
  $__i$0$i$i$i24 = 0;
  while(1) {
   $exitcond$i$i$i25 = ($__i$0$i$i$i24|0)==(3);
   if ($exitcond$i$i$i25) {
    break;
   }
   $126 = (($7) + ($__i$0$i$i$i24<<2)|0);
   HEAP32[$126>>2] = 0;
   $127 = (($__i$0$i$i$i24) + 1)|0;
   $__i$0$i$i$i24 = $127;
  }
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
  $128 = HEAP32[$91>>2]|0;
  $129 = ((($128)) + 12|0);
  $130 = HEAP32[$129>>2]|0;
  $131 = (FUNCTION_TABLE_ii[$130 & 8191]($91)|0);
  HEAP8[$__dp>>0] = $131;
  $132 = HEAP32[$91>>2]|0;
  $133 = ((($132)) + 16|0);
  $134 = HEAP32[$133>>2]|0;
  $135 = (FUNCTION_TABLE_ii[$134 & 8191]($91)|0);
  HEAP8[$__ts>>0] = $135;
  $136 = HEAP32[$91>>2]|0;
  $137 = ((($136)) + 20|0);
  $138 = HEAP32[$137>>2]|0;
  FUNCTION_TABLE_vii[$138 & 8191]($8,$91);
  $139 = HEAP8[$__grp>>0]|0;
  $140 = $139 & 1;
  $141 = ($140<<24>>24)==(0);
  if ($141) {
   $145 = ((($__grp)) + 1|0);
   HEAP8[$145>>0] = 0;
   HEAP8[$__grp>>0] = 0;
  } else {
   $142 = ((($__grp)) + 8|0);
   $143 = HEAP32[$142>>2]|0;
   HEAP8[$143>>0] = 0;
   $144 = ((($__grp)) + 4|0);
   HEAP32[$144>>2] = 0;
  }
  __THREW__ = 0;
  invoke_vii(3368,($__grp|0),0);
  $146 = __THREW__; __THREW__ = 0;
  $147 = $146&1;
  if ($147) {
   $148 = ___cxa_find_matching_catch_3(0|0)|0;
   $149 = tempRet0;
   ___clang_call_terminate($148);
   // unreachable;
  }
  ;HEAP32[$__grp>>2]=HEAP32[$8>>2]|0;HEAP32[$__grp+4>>2]=HEAP32[$8+4>>2]|0;HEAP32[$__grp+8>>2]=HEAP32[$8+8>>2]|0;
  $__i$0$i$i$i29 = 0;
  while(1) {
   $exitcond$i$i$i30 = ($__i$0$i$i$i29|0)==(3);
   if ($exitcond$i$i$i30) {
    break;
   }
   $150 = (($8) + ($__i$0$i$i$i29<<2)|0);
   HEAP32[$150>>2] = 0;
   $151 = (($__i$0$i$i$i29) + 1)|0;
   $__i$0$i$i$i29 = $151;
  }
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
  $152 = HEAP32[$91>>2]|0;
  $153 = ((($152)) + 24|0);
  $154 = HEAP32[$153>>2]|0;
  FUNCTION_TABLE_vii[$154 & 8191]($9,$91);
  $155 = HEAP8[$__sym>>0]|0;
  $156 = $155 & 1;
  $157 = ($156<<24>>24)==(0);
  if ($157) {
   $161 = ((($__sym)) + 1|0);
   HEAP8[$161>>0] = 0;
   HEAP8[$__sym>>0] = 0;
  } else {
   $158 = ((($__sym)) + 8|0);
   $159 = HEAP32[$158>>2]|0;
   HEAP8[$159>>0] = 0;
   $160 = ((($__sym)) + 4|0);
   HEAP32[$160>>2] = 0;
  }
  __THREW__ = 0;
  invoke_vii(3368,($__sym|0),0);
  $162 = __THREW__; __THREW__ = 0;
  $163 = $162&1;
  if ($163) {
   $164 = ___cxa_find_matching_catch_3(0|0)|0;
   $165 = tempRet0;
   ___clang_call_terminate($164);
   // unreachable;
  }
  ;HEAP32[$__sym>>2]=HEAP32[$9>>2]|0;HEAP32[$__sym+4>>2]=HEAP32[$9+4>>2]|0;HEAP32[$__sym+8>>2]=HEAP32[$9+8>>2]|0;
  $__i$0$i$i$i34 = 0;
  while(1) {
   $exitcond$i$i$i35 = ($__i$0$i$i$i34|0)==(3);
   if ($exitcond$i$i$i35) {
    break;
   }
   $166 = (($9) + ($__i$0$i$i$i34<<2)|0);
   HEAP32[$166>>2] = 0;
   $167 = (($__i$0$i$i$i34) + 1)|0;
   $__i$0$i$i$i34 = $167;
  }
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
  $168 = HEAP32[$91>>2]|0;
  $169 = ((($168)) + 36|0);
  $170 = HEAP32[$169>>2]|0;
  $171 = (FUNCTION_TABLE_ii[$170 & 8191]($91)|0);
  $storemerge = $171;
 }
 HEAP32[$__fd>>2] = $storemerge;
 STACKTOP = sp;return;
}
function __ZNSt3__119__double_or_nothingIcEEvRNS_10unique_ptrIT_PFvPvEEERPS2_S9_($__b,$__n,$__e) {
 $__b = $__b|0;
 $__n = $__n|0;
 $__e = $__e|0;
 var $$ = 0, $$pr = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($__b)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)!=(4899|0);
 $3 = HEAP32[$__e>>2]|0;
 $4 = HEAP32[$__b>>2]|0;
 $5 = (($3) - ($4))|0;
 $6 = ($5>>>0)<(2147483647);
 $7 = $5 << 1;
 $$ = $6 ? $7 : -1;
 $8 = HEAP32[$__n>>2]|0;
 $9 = $4;
 $10 = (($8) - ($4))|0;
 $11 = $2 ? $9 : 0;
 $12 = (_realloc($11,$$)|0);
 $13 = ($12|0)==(0|0);
 if ($13) {
  __ZSt17__throw_bad_allocv();
 }
 do {
  if ($2) {
   $14 = $12;
   HEAP32[$__b>>2] = $14;
   $23 = $12;
  } else {
   $$pr = HEAP32[$__b>>2]|0;
   $15 = $12;
   HEAP32[$__b>>2] = $15;
   $16 = ($$pr|0)==(0|0);
   if ($16) {
    $23 = $12;
   } else {
    $17 = HEAP32[$0>>2]|0;
    __THREW__ = 0;
    invoke_vi($17|0,($$pr|0));
    $18 = __THREW__; __THREW__ = 0;
    $19 = $18&1;
    if ($19) {
     $20 = ___cxa_find_matching_catch_3(0|0)|0;
     $21 = tempRet0;
     ___clang_call_terminate($20);
     // unreachable;
    } else {
     $$pre = HEAP32[$__b>>2]|0;
     $23 = $$pre;
     break;
    }
   }
  }
 } while(0);
 HEAP32[$0>>2] = (337);
 $22 = (($23) + ($10)|0);
 HEAP32[$__n>>2] = $22;
 $24 = HEAP32[$__b>>2]|0;
 $25 = (($24) + ($$)|0);
 HEAP32[$__e>>2] = $25;
 return;
}
function __ZNSt3__119__double_or_nothingIjEEvRNS_10unique_ptrIT_PFvPvEEERPS2_S9_($__b,$__n,$__e) {
 $__b = $__b|0;
 $__n = $__n|0;
 $__e = $__e|0;
 var $$ = 0, $$pr = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($__b)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)!=(4899|0);
 $3 = HEAP32[$__e>>2]|0;
 $4 = HEAP32[$__b>>2]|0;
 $5 = (($3) - ($4))|0;
 $6 = ($5>>>0)<(2147483647);
 $7 = $5 << 1;
 $$ = $6 ? $7 : -1;
 $8 = HEAP32[$__n>>2]|0;
 $9 = (($8) - ($4))|0;
 $10 = $9 >> 2;
 $phitmp = $4;
 $11 = $2 ? $phitmp : 0;
 $12 = (_realloc($11,$$)|0);
 $13 = ($12|0)==(0|0);
 if ($13) {
  __ZSt17__throw_bad_allocv();
 }
 do {
  if ($2) {
   $14 = $12;
   HEAP32[$__b>>2] = $14;
   $24 = $12;
  } else {
   $$pr = HEAP32[$__b>>2]|0;
   $15 = $12;
   HEAP32[$__b>>2] = $15;
   $16 = ($$pr|0)==(0|0);
   if ($16) {
    $24 = $12;
   } else {
    $17 = HEAP32[$0>>2]|0;
    __THREW__ = 0;
    invoke_vi($17|0,($$pr|0));
    $18 = __THREW__; __THREW__ = 0;
    $19 = $18&1;
    if ($19) {
     $20 = ___cxa_find_matching_catch_3(0|0)|0;
     $21 = tempRet0;
     ___clang_call_terminate($20);
     // unreachable;
    } else {
     $$pre = HEAP32[$__b>>2]|0;
     $24 = $$pre;
     break;
    }
   }
  }
 } while(0);
 HEAP32[$0>>2] = (337);
 $22 = $$ >>> 2;
 $23 = (($24) + ($10<<2)|0);
 HEAP32[$__n>>2] = $23;
 $25 = HEAP32[$__b>>2]|0;
 $26 = (($25) + ($22<<2)|0);
 HEAP32[$__e>>2] = $26;
 return;
}
function __ZNKSt3__19money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_bRNS_8ios_baseERjRNS_12basic_stringIcS3_NS_9allocatorIcEEEE($this,$__b,$__e,$__intl,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__intl = $__intl|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$011 = 0, $$0110 = 0, $$012 = 0, $$019 = 0, $$byval_copy = 0, $$pre = 0, $$pre$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__loc = 0, $__neg = 0, $__w$0 = 0, $__w$0$lcssa = 0, $__wb = 0, $__wbuf = 0, $__wn = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 24|0;
 $__wbuf = sp + 32|0;
 $__wb = sp + 16|0;
 $__wn = sp + 8|0;
 $__loc = sp + 4|0;
 $__neg = sp + 28|0;
 $0 = sp;
 $1 = $__wbuf;
 HEAP32[$__wb>>2] = $1;
 $2 = ((($__wb)) + 4|0);
 HEAP32[$2>>2] = (4899);
 $3 = ((($__wbuf)) + 100|0);
 __THREW__ = 0;
 $4 = (invoke_ii(3373,($__iob|0))|0);
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $34 = ___cxa_find_matching_catch_2()|0;
  $35 = tempRet0;
  HEAP32[$__wb>>2] = 0;
  $$011 = $34;$$019 = $35;$95 = $__wbuf;
  label = 37;
 } else {
  HEAP32[$__loc>>2] = $4;
  __THREW__ = 0;
  $7 = (invoke_iii(3374,($__loc|0),(1207576|0))|0);
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  L4: do {
   if (!($9)) {
    HEAP8[$__neg>>0] = 0;
    $10 = HEAP32[$__e>>2]|0;
    HEAP32[$0>>2] = $10;
    $11 = ((($__iob)) + 4|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = $10;
    __THREW__ = 0;
    ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
    $14 = (invoke_iiiiiiiiiiii(4900,($__b|0),($$byval_copy|0),($__intl|0),($__loc|0),($12|0),($__err|0),($__neg|0),($7|0),($__wb|0),($__wn|0),($3|0))|0);
    $15 = __THREW__; __THREW__ = 0;
    $16 = $15&1;
    if (!($16)) {
     if ($14) {
      $17 = HEAP8[$__v>>0]|0;
      $18 = $17 & 1;
      $19 = ($18<<24>>24)==(0);
      if ($19) {
       $23 = ((($__v)) + 1|0);
       HEAP8[$23>>0] = 0;
       HEAP8[$__v>>0] = 0;
      } else {
       $20 = ((($__v)) + 8|0);
       $21 = HEAP32[$20>>2]|0;
       HEAP8[$21>>0] = 0;
       $22 = ((($__v)) + 4|0);
       HEAP32[$22>>2] = 0;
      }
      $24 = HEAP8[$__neg>>0]|0;
      $25 = ($24<<24>>24)==(0);
      if (!($25)) {
       $26 = HEAP32[$7>>2]|0;
       $27 = ((($26)) + 28|0);
       $28 = HEAP32[$27>>2]|0;
       __THREW__ = 0;
       $29 = (invoke_iii($28|0,($7|0),45)|0);
       $30 = __THREW__; __THREW__ = 0;
       $31 = $30&1;
       if ($31) {
        break;
       }
       __THREW__ = 0;
       invoke_vii(3370,($__v|0),($29|0));
       $32 = __THREW__; __THREW__ = 0;
       $33 = $32&1;
       if ($33) {
        break;
       }
      }
      $36 = HEAP32[$7>>2]|0;
      $37 = ((($36)) + 28|0);
      $38 = HEAP32[$37>>2]|0;
      __THREW__ = 0;
      $39 = (invoke_iii($38|0,($7|0),48)|0);
      $40 = __THREW__; __THREW__ = 0;
      $41 = $40&1;
      if ($41) {
       break;
      }
      $42 = HEAP32[$__wb>>2]|0;
      $43 = HEAP32[$__wn>>2]|0;
      $44 = ((($43)) + -1|0);
      $__w$0 = $42;
      while(1) {
       $45 = ($__w$0>>>0)<($44>>>0);
       if (!($45)) {
        $__w$0$lcssa = $__w$0;
        break;
       }
       $46 = HEAP8[$__w$0>>0]|0;
       $47 = ($46<<24>>24)==($39<<24>>24);
       if (!($47)) {
        $__w$0$lcssa = $__w$0;
        break;
       }
       $48 = ((($__w$0)) + 1|0);
       $__w$0 = $48;
      }
      __THREW__ = 0;
      (invoke_iiii(4905,($__v|0),($__w$0$lcssa|0),($43|0))|0);
      $49 = __THREW__; __THREW__ = 0;
      $50 = $49&1;
      if ($50) {
       break;
      }
     }
     $51 = HEAP32[$__b>>2]|0;
     $52 = ($51|0)==(0|0);
     do {
      if ($52) {
       $66 = 0;
      } else {
       $53 = ((($51)) + 12|0);
       $54 = HEAP32[$53>>2]|0;
       $55 = ((($51)) + 16|0);
       $56 = HEAP32[$55>>2]|0;
       $57 = ($54|0)==($56|0);
       if ($57) {
        $58 = HEAP32[$51>>2]|0;
        $59 = ((($58)) + 36|0);
        $60 = HEAP32[$59>>2]|0;
        __THREW__ = 0;
        $61 = (invoke_ii($60|0,($51|0))|0);
        $62 = __THREW__; __THREW__ = 0;
        $63 = $62&1;
        if ($63) {
         break L4;
        }
        $64 = ($61|0)==(-1);
        if ($64) {
         HEAP32[$__b>>2] = 0;
         $66 = 0;
         break;
        } else {
         $$pre$i$i$i = HEAP32[$__b>>2]|0;
         $66 = $$pre$i$i$i;
         break;
        }
       } else {
        $66 = $51;
       }
      }
     } while(0);
     $65 = ($66|0)==(0|0);
     $67 = ($10|0)==(0);
     do {
      if ($67) {
       label = 30;
      } else {
       $68 = ((($13)) + 12|0);
       $69 = HEAP32[$68>>2]|0;
       $70 = ((($13)) + 16|0);
       $71 = HEAP32[$70>>2]|0;
       $72 = ($69|0)==($71|0);
       if ($72) {
        $73 = $10;
        $74 = HEAP32[$73>>2]|0;
        $75 = ((($74)) + 36|0);
        $76 = HEAP32[$75>>2]|0;
        __THREW__ = 0;
        $77 = (invoke_ii($76|0,($13|0))|0);
        $78 = __THREW__; __THREW__ = 0;
        $79 = $78&1;
        if ($79) {
         break L4;
        }
        $80 = ($77|0)==(-1);
        if ($80) {
         HEAP32[$__e>>2] = 0;
         label = 30;
         break;
        }
       }
       if (!($65)) {
        label = 31;
       }
      }
     } while(0);
     if ((label|0) == 30) {
      if ($65) {
       label = 31;
      }
     }
     if ((label|0) == 31) {
      $81 = HEAP32[$__err>>2]|0;
      $82 = $81 | 2;
      HEAP32[$__err>>2] = $82;
     }
     $83 = HEAP32[$__b>>2]|0;
     (__ZNSt3__114__shared_count16__release_sharedEv($4)|0);
     $84 = HEAP32[$__wb>>2]|0;
     HEAP32[$__wb>>2] = 0;
     $85 = ($84|0)==(0|0);
     if (!($85)) {
      $86 = HEAP32[$2>>2]|0;
      __THREW__ = 0;
      invoke_vi($86|0,($84|0));
      $87 = __THREW__; __THREW__ = 0;
      $88 = $87&1;
      if ($88) {
       $89 = ___cxa_find_matching_catch_3(0|0)|0;
       $90 = tempRet0;
       ___clang_call_terminate($89);
       // unreachable;
      }
     }
     STACKTOP = sp;return ($83|0);
    }
   }
  } while(0);
  $91 = ___cxa_find_matching_catch_2()|0;
  $92 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($4)|0);
  $$pre = HEAP32[$__wb>>2]|0;
  HEAP32[$__wb>>2] = 0;
  $93 = ($$pre|0)==(0|0);
  if ($93) {
   $$0110 = $92;$$012 = $91;
  } else {
   $$011 = $91;$$019 = $92;$95 = $$pre;
   label = 37;
  }
 }
 if ((label|0) == 37) {
  $94 = HEAP32[$2>>2]|0;
  __THREW__ = 0;
  invoke_vi($94|0,($95|0));
  $96 = __THREW__; __THREW__ = 0;
  $97 = $96&1;
  if ($97) {
   $98 = ___cxa_find_matching_catch_3(0|0)|0;
   $99 = tempRet0;
   ___clang_call_terminate($98);
   // unreachable;
  } else {
   $$0110 = $$019;$$012 = $$011;
  }
 }
 ___resumeException($$012|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6appendIPcEENS_9enable_ifIXsr21__is_forward_iteratorIT_EE5valueERS5_E4typeES9_S9_($this,$__first,$__last) {
 $this = $this|0;
 $__first = $__first|0;
 $__last = $__last|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $__p$0 = 0, $phitmp$i = 0, $scevgep = 0, $scevgep1 = 0, $scevgep2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $__first;
 $1 = HEAP8[$this>>0]|0;
 $2 = $1 & 1;
 $3 = ($2<<24>>24)==(0);
 if ($3) {
  $4 = $1&255;
  $5 = $4 >>> 1;
  $15 = 10;$16 = $5;$40 = $1;
 } else {
  $6 = ((($this)) + 4|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = HEAP32[$this>>2]|0;
  $9 = $8 & -2;
  $phitmp$i = (($9) + -1)|0;
  $10 = $8&255;
  $15 = $phitmp$i;$16 = $7;$40 = $10;
 }
 $11 = $__last;
 $12 = (($11) - ($0))|0;
 $13 = ($__last|0)==($__first|0);
 do {
  if (!($13)) {
   $14 = (($15) - ($16))|0;
   $17 = ($14>>>0)<($12>>>0);
   if ($17) {
    $18 = (($16) + ($12))|0;
    $19 = (($18) - ($15))|0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE9__grow_byEjjjjjj($this,$15,$19,$16,$16,0,0);
    $$pre = HEAP8[$this>>0]|0;
    $21 = $$pre;
   } else {
    $21 = $40;
   }
   $20 = $21 & 1;
   $22 = ($20<<24>>24)==(0);
   if ($22) {
    $25 = ((($this)) + 1|0);
    $27 = $25;
   } else {
    $23 = ((($this)) + 8|0);
    $24 = HEAP32[$23>>2]|0;
    $27 = $24;
   }
   $26 = (($27) + ($16)|0);
   $28 = (($16) - ($0))|0;
   $scevgep = (($__last) + ($28)|0);
   $scevgep1 = $scevgep;
   $$0 = $__first;$__p$0 = $26;
   while(1) {
    $29 = ($$0|0)==($__last|0);
    if ($29) {
     break;
    }
    $30 = HEAP8[$$0>>0]|0;
    HEAP8[$__p$0>>0] = $30;
    $31 = ((($__p$0)) + 1|0);
    $32 = ((($$0)) + 1|0);
    $$0 = $32;$__p$0 = $31;
   }
   $scevgep2 = (($27) + ($scevgep1)|0);
   HEAP8[$scevgep2>>0] = 0;
   $33 = (($16) + ($12))|0;
   $34 = HEAP8[$this>>0]|0;
   $35 = $34 & 1;
   $36 = ($35<<24>>24)==(0);
   if ($36) {
    $38 = $33 << 1;
    $39 = $38&255;
    HEAP8[$this>>0] = $39;
    break;
   } else {
    $37 = ((($this)) + 4|0);
    HEAP32[$37>>2] = $33;
    break;
   }
  }
 } while(0);
 return ($this|0);
}
function __ZNSt3__19money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__19money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__19money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_bRNS_8ios_baseERjRe($this,$__b,$__e,$__intl,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__intl = $__intl|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0 = 0, $$0$i = 0, $$0$i$i$i$i = 0, $$0$i$i1$i$i = 0, $$0$lcssa$i = 0, $$01 = 0, $$113 = 0, $$114 = 0, $$1211 = 0, $$1212 = 0, $$byval_copy = 0, $$phi$trans$insert$pre$phiZZZZ2D = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre15 = 0, $$pre7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $12 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__atoms = 0, $__h$sroa$0$0 = 0, $__h$sroa$0$1 = 0, $__loc = 0, $__nbuf = 0, $__nc$0 = 0, $__nc$1 = 0;
 var $__nc$2 = 0, $__nc$2$lcssa = 0, $__neg = 0, $__w$0 = 0, $__wb = 0, $__wbuf = 0, $__wn = 0, $phitmp$i$i$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 592|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 472|0;
 $vararg_buffer = sp;
 $__wbuf = sp + 72|0;
 $__wb = sp + 64|0;
 $__wn = sp + 56|0;
 $__loc = sp + 52|0;
 $__neg = sp + 576|0;
 $0 = sp + 48|0;
 $__atoms = sp + 8|0;
 $__nbuf = sp + 476|0;
 $1 = $__wbuf;
 HEAP32[$__wb>>2] = $1;
 $2 = ((($__wb)) + 4|0);
 HEAP32[$2>>2] = (4899);
 $3 = ((($__wbuf)) + 400|0);
 __THREW__ = 0;
 $4 = (invoke_ii(3373,($__iob|0))|0);
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $34 = ___cxa_find_matching_catch_2()|0;
  $35 = tempRet0;
  HEAP32[$__wb>>2] = 0;
  $$113 = $34;$$1211 = $35;$110 = 4899;$111 = $__wbuf;
  label = 49;
 } else {
  HEAP32[$__loc>>2] = $4;
  __THREW__ = 0;
  $7 = (invoke_iii(3374,($__loc|0),(1207608|0))|0);
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  L4: do {
   if ($9) {
    label = 10;
   } else {
    HEAP8[$__neg>>0] = 0;
    $10 = HEAP32[$__e>>2]|0;
    HEAP32[$0>>2] = $10;
    $11 = ((($__iob)) + 4|0);
    $12 = HEAP32[$11>>2]|0;
    __THREW__ = 0;
    ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
    $13 = (invoke_iiiiiiiiiiii(4906,($__b|0),($$byval_copy|0),($__intl|0),($__loc|0),($12|0),($__err|0),($__neg|0),($7|0),($__wb|0),($__wn|0),($3|0))|0);
    $14 = __THREW__; __THREW__ = 0;
    $15 = $14&1;
    if ($15) {
     label = 10;
    } else {
     L7: do {
      if ($13) {
       $16 = HEAP32[$7>>2]|0;
       $17 = ((($16)) + 48|0);
       $18 = HEAP32[$17>>2]|0;
       __THREW__ = 0;
       (invoke_iiiii($18|0,($7|0),(621917|0),((621927)|0),($__atoms|0))|0);
       $19 = __THREW__; __THREW__ = 0;
       $20 = $19&1;
       if ($20) {
        label = 10;
        break L4;
       }
       $21 = HEAP32[$__wn>>2]|0;
       $22 = HEAP32[$__wb>>2]|0;
       $23 = (($21) - ($22))|0;
       $24 = ($23|0)>(392);
       $25 = $22;
       $26 = $21;
       if ($24) {
        $27 = $23 >> 2;
        $28 = (($27) + 2)|0;
        $29 = (_malloc($28)|0);
        $30 = $29;
        $31 = ($29|0)==(0|0);
        if ($31) {
         __THREW__ = 0;
         invoke_v(4863);
         $32 = __THREW__; __THREW__ = 0;
         $33 = $32&1;
         if ($33) {
          $__h$sroa$0$0 = $30;
         } else {
          $__h$sroa$0$1 = $30;$__nc$0 = 0;
          label = 13;
         }
        } else {
         $__h$sroa$0$1 = $30;$__nc$0 = $29;
         label = 13;
        }
       } else {
        $__h$sroa$0$1 = 0;$__nc$0 = $__nbuf;
        label = 13;
       }
       do {
        if ((label|0) == 13) {
         $42 = HEAP8[$__neg>>0]|0;
         $43 = ($42<<24>>24)==(0);
         if ($43) {
          $__nc$1 = $__nc$0;
         } else {
          $44 = ((($__nc$0)) + 1|0);
          HEAP8[$__nc$0>>0] = 45;
          $__nc$1 = $44;
         }
         $45 = ((($__atoms)) + 40|0);
         $46 = $__atoms;
         $48 = $26;$__nc$2 = $__nc$1;$__w$0 = $25;
         while(1) {
          $47 = ($__w$0>>>0)<($48>>>0);
          if (!($47)) {
           $__nc$2$lcssa = $__nc$2;
           break;
          }
          $51 = HEAP32[$__w$0>>2]|0;
          $$0$i = $__atoms;
          while(1) {
           $52 = ($$0$i|0)==($45|0);
           if ($52) {
            $$0$lcssa$i = $45;
            break;
           }
           $53 = HEAP32[$$0$i>>2]|0;
           $54 = ($53|0)==($51|0);
           if ($54) {
            $$0$lcssa$i = $$0$i;
            break;
           }
           $55 = ((($$0$i)) + 4|0);
           $$0$i = $55;
          }
          $56 = $$0$lcssa$i;
          $57 = (($56) - ($46))|0;
          $58 = $57 >> 2;
          $59 = (621917 + ($58)|0);
          $60 = HEAP8[$59>>0]|0;
          HEAP8[$__nc$2>>0] = $60;
          $61 = ((($__w$0)) + 4|0);
          $62 = ((($__nc$2)) + 1|0);
          $$pre = HEAP32[$__wn>>2]|0;
          $48 = $$pre;$__nc$2 = $62;$__w$0 = $61;
         }
         HEAP8[$__nc$2$lcssa>>0] = 0;
         HEAP32[$vararg_buffer>>2] = $__v;
         $49 = (_sscanf($__nbuf,621802,$vararg_buffer)|0);
         $50 = ($49|0)==(1);
         if (!($50)) {
          __THREW__ = 0;
          invoke_vi(4901,(621806|0));
          $63 = __THREW__; __THREW__ = 0;
          $64 = $63&1;
          if ($64) {
           $__h$sroa$0$0 = $__h$sroa$0$1;
           break;
          }
         }
         $65 = ($__h$sroa$0$1|0)==(0);
         if (!($65)) {
          $66 = $__h$sroa$0$1;
          _free($66);
         }
         break L7;
        }
       } while(0);
       $38 = ___cxa_find_matching_catch_2()|0;
       $39 = tempRet0;
       $40 = ($__h$sroa$0$0|0)==(0);
       if ($40) {
        $$0 = $38;$$01 = $39;$$phi$trans$insert$pre$phiZZZZ2D = $__wb;
        break L4;
       }
       $41 = $__h$sroa$0$0;
       _free($41);
       $$0 = $38;$$01 = $39;$$phi$trans$insert$pre$phiZZZZ2D = $__wb;
       break L4;
      }
     } while(0);
     $67 = HEAP32[$__b>>2]|0;
     $68 = ($67|0)==(0|0);
     do {
      if ($68) {
       $116 = 1;
      } else {
       $69 = ((($67)) + 12|0);
       $70 = HEAP32[$69>>2]|0;
       $71 = ((($67)) + 16|0);
       $72 = HEAP32[$71>>2]|0;
       $73 = ($70|0)==($72|0);
       if ($73) {
        $74 = HEAP32[$67>>2]|0;
        $75 = ((($74)) + 36|0);
        $76 = HEAP32[$75>>2]|0;
        __THREW__ = 0;
        $77 = (invoke_ii($76|0,($67|0))|0);
        $78 = __THREW__; __THREW__ = 0;
        $79 = $78&1;
        if ($79) {
         label = 10;
         break L4;
        } else {
         $$0$i$i$i$i = $77;
        }
       } else {
        $80 = HEAP32[$70>>2]|0;
        $$0$i$i$i$i = $80;
       }
       $81 = ($$0$i$i$i$i|0)==(-1);
       if ($81) {
        HEAP32[$__b>>2] = 0;
        $116 = 1;
        break;
       } else {
        $$pre$i$i$i = HEAP32[$__b>>2]|0;
        $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
        $116 = $phitmp$i$i$i;
        break;
       }
      }
     } while(0);
     $82 = HEAP32[$__e>>2]|0;
     $83 = ($82|0)==(0|0);
     do {
      if ($83) {
       label = 40;
      } else {
       $84 = ((($82)) + 12|0);
       $85 = HEAP32[$84>>2]|0;
       $86 = ((($82)) + 16|0);
       $87 = HEAP32[$86>>2]|0;
       $88 = ($85|0)==($87|0);
       if ($88) {
        $89 = HEAP32[$82>>2]|0;
        $90 = ((($89)) + 36|0);
        $91 = HEAP32[$90>>2]|0;
        __THREW__ = 0;
        $92 = (invoke_ii($91|0,($82|0))|0);
        $93 = __THREW__; __THREW__ = 0;
        $94 = $93&1;
        if ($94) {
         label = 10;
         break L4;
        } else {
         $$0$i$i1$i$i = $92;
        }
       } else {
        $95 = HEAP32[$85>>2]|0;
        $$0$i$i1$i$i = $95;
       }
       $96 = ($$0$i$i1$i$i|0)==(-1);
       if ($96) {
        HEAP32[$__e>>2] = 0;
        label = 40;
        break;
       } else {
        if ($116) {
         break;
        } else {
         label = 42;
         break;
        }
       }
      }
     } while(0);
     if ((label|0) == 40) {
      if ($116) {
       label = 42;
      }
     }
     if ((label|0) == 42) {
      $97 = HEAP32[$__err>>2]|0;
      $98 = $97 | 2;
      HEAP32[$__err>>2] = $98;
     }
     $99 = HEAP32[$__b>>2]|0;
     $100 = HEAP32[$__loc>>2]|0;
     (__ZNSt3__114__shared_count16__release_sharedEv($100)|0);
     $101 = HEAP32[$__wb>>2]|0;
     HEAP32[$__wb>>2] = 0;
     $102 = ($101|0)==(0|0);
     if (!($102)) {
      $103 = HEAP32[$2>>2]|0;
      __THREW__ = 0;
      invoke_vi($103|0,($101|0));
      $104 = __THREW__; __THREW__ = 0;
      $105 = $104&1;
      if ($105) {
       $106 = ___cxa_find_matching_catch_3(0|0)|0;
       $107 = tempRet0;
       ___clang_call_terminate($106);
       // unreachable;
      }
     }
     STACKTOP = sp;return ($99|0);
    }
   }
  } while(0);
  if ((label|0) == 10) {
   $36 = ___cxa_find_matching_catch_2()|0;
   $37 = tempRet0;
   $$0 = $36;$$01 = $37;$$phi$trans$insert$pre$phiZZZZ2D = $__wb;
  }
  $108 = HEAP32[$__loc>>2]|0;
  (__ZNSt3__114__shared_count16__release_sharedEv($108)|0);
  $$pre7 = HEAP32[$$phi$trans$insert$pre$phiZZZZ2D>>2]|0;
  HEAP32[$$phi$trans$insert$pre$phiZZZZ2D>>2] = 0;
  $109 = ($$pre7|0)==(0|0);
  if ($109) {
   $$114 = $$0;$$1212 = $$01;
  } else {
   $$pre15 = HEAP32[$2>>2]|0;
   $$113 = $$0;$$1211 = $$01;$110 = $$pre15;$111 = $$pre7;
   label = 49;
  }
 }
 if ((label|0) == 49) {
  __THREW__ = 0;
  invoke_vi($110|0,($111|0));
  $112 = __THREW__; __THREW__ = 0;
  $113 = $112&1;
  if ($113) {
   $114 = ___cxa_find_matching_catch_3(0|0)|0;
   $115 = tempRet0;
   ___clang_call_terminate($114);
   // unreachable;
  } else {
   $$114 = $$113;$$1212 = $$1211;
  }
 }
 ___resumeException($$114|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__19money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE8__do_getERS4_S4_bRKNS_6localeEjRjRbRKNS_5ctypeIwEERNS_10unique_ptrIwPFvPvEEERPwSM_($__b,$__e,$__intl,$__loc,$__flags,$__err,$__neg,$__ct,$__wb,$__wn,$__we) {
 $__b = $__b|0;
 $__e = $__e|0;
 $__intl = $__intl|0;
 $__loc = $__loc|0;
 $__flags = $__flags|0;
 $__err = $__err|0;
 $__neg = $__neg|0;
 $__ct = $__ct|0;
 $__wb = $__wb|0;
 $__wn = $__wn|0;
 $__we = $__we|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i107 = 0, $$0$i$i$i$i141 = 0, $$0$i$i$i$i173 = 0, $$0$i$i$i$i198 = 0, $$0$i$i$i$i45 = 0, $$0$i$i$i$i66 = 0, $$0$i$i1$i$i = 0, $$0$i$i1$i$i113 = 0, $$0$i$i1$i$i147 = 0, $$0$i$i1$i$i179 = 0, $$0$i$i1$i$i204 = 0, $$0$i$i1$i$i51 = 0, $$0$i$i1$i$i72 = 0, $$0$i$i101 = 0, $$0$i$i104 = 0, $$0$i$i125 = 0, $$0$i$i134 = 0, $$0$i$i137 = 0;
 var $$0$i$i159 = 0, $$0$i$i170 = 0, $$0$i$i194 = 0, $$0$i$i60 = 0, $$0$i$i64 = 0, $$0$i$i81 = 0, $$0$i$i88 = 0, $$0$i$i93 = 0, $$10 = 0, $$lcssa = 0, $$lcssa450 = 0, $$lcssa461 = 0, $$phi$trans$insert = 0, $$phi$trans$insert323 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i109 = 0, $$pre$i$i$i143 = 0, $$pre$i$i$i175 = 0, $$pre$i$i$i200 = 0;
 var $$pre$i$i$i47 = 0, $$pre$i$i$i68 = 0, $$pre$i3$i$i115 = 0, $$pre$i3$i$i115309$ph = 0, $$pre$i3$i$i115309333 = 0, $$pre$i3$i$i149 = 0, $$pre$i3$i$i181 = 0, $$pre$i3$i$i181302 = 0, $$pre$i3$i$i181304 = 0, $$pre$i3$i$i206 = 0, $$pre$i3$i$i206317 = 0, $$pre$i3$i$i206318 = 0, $$pre$i3$i$i74 = 0, $$pre$i3$i$i74327 = 0, $$pre$i3$i$i74329 = 0, $$pre306 = 0, $$pre308 = 0, $$pre311 = 0, $$pre312 = 0, $$pre313 = 0;
 var $$pre314 = 0, $$pre315 = 0, $$pre316 = 0, $$pre320 = 0, $$pre321 = 0, $$pre322 = 0, $$pre324 = 0, $$pre325 = 0, $$pre326 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0;
 var $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0;
 var $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0;
 var $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0;
 var $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0;
 var $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0;
 var $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0;
 var $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0;
 var $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0;
 var $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0;
 var $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0;
 var $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0;
 var $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0;
 var $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0;
 var $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0;
 var $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0;
 var $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0;
 var $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0;
 var $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0;
 var $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0;
 var $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0;
 var $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0;
 var $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0;
 var $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0;
 var $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0;
 var $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0;
 var $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0;
 var $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0;
 var $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0;
 var $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0;
 var $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0;
 var $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0;
 var $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0;
 var $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0;
 var $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0;
 var $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0;
 var $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0;
 var $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0;
 var $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0;
 var $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__dp = 0, $__et = 0, $__fd = 0, $__gb = 0, $__gbuf = 0, $__ge = 0, $__gn = 0, $__grp = 0, $__i$0 = 0, $__i$0$i$i = 0;
 var $__i$0$i$i29 = 0, $__i$0$i$i31 = 0, $__i$0$i$i35 = 0, $__i$0$i$i40 = 0, $__ng$0 = 0, $__ng$0$lcssa451 = 0, $__ng$2$ph = 0, $__nsn = 0, $__nsn$__trailing_sign$0 = 0, $__nsn$__trailing_sign$023 = 0, $__p$0 = 0, $__pat = 0, $__psn = 0, $__psn$__trailing_sign$0 = 0, $__psn$__trailing_sign$022 = 0, $__spaces = 0, $__sym = 0, $__sym_curr_char$sroa$0$0 = 0, $__sym_space_end$sroa$0$0 = 0, $__sym_space_end$sroa$0$0$lcssa459 = 0;
 var $__sym_space_end$sroa$0$0460 = 0, $__trailing_sign$0 = 0, $__trailing_sign$0$lcssa = 0, $__trailing_sign$1 = 0, $__ts = 0, $brmerge = 0, $cond28 = 0, $exitcond$i$i = 0, $exitcond$i$i30 = 0, $exitcond$i$i32 = 0, $exitcond$i$i36 = 0, $exitcond$i$i41 = 0, $lpad$phi$index = 0, $lpad$phi$index8 = 0, $or$cond = 0, $or$cond12 = 0, $or$cond14 = 0, $or$cond25 = 0, $phitmp$i$i$i = 0, $phitmp$i$i$i110 = 0;
 var $phitmp$i$i$i144 = 0, $phitmp$i$i$i176 = 0, $phitmp$i$i$i201 = 0, $phitmp$i$i$i48 = 0, $phitmp$i$i$i69 = 0, $phitmp$i4$i$i116 = 0, $phitmp$i4$i$i182 = 0, $phitmp$i4$i$i207 = 0, $phitmp$i4$i$i75 = 0, $switch = 0, $tmp$i$i118 = 0, $tmp$i$i184 = 0, $tmp$i$i209 = 0, $tmp$i$i77 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 512|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 496|0;
 $__gbuf = sp + 96|0;
 $__gb = sp + 88|0;
 $__gn = sp + 80|0;
 $__ge = sp + 76|0;
 $__pat = sp + 500|0;
 $__dp = sp + 72|0;
 $__ts = sp + 68|0;
 $__grp = sp + 56|0;
 $__sym = sp + 44|0;
 $__psn = sp + 32|0;
 $__nsn = sp + 20|0;
 $__spaces = sp + 8|0;
 $__fd = sp + 4|0;
 $__et = sp;
 HEAP32[$0>>2] = $__we;
 $1 = $__gbuf;
 HEAP32[$__gb>>2] = $1;
 $2 = ((($__gb)) + 4|0);
 HEAP32[$2>>2] = (4899);
 HEAP32[$__gn>>2] = $__gbuf;
 $3 = ((($__gbuf)) + 400|0);
 HEAP32[$__ge>>2] = $3;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $4 = (($__grp) + ($__i$0$i$i<<2)|0);
  HEAP32[$4>>2] = 0;
  $5 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $5;
 }
 $__i$0$i$i29 = 0;
 while(1) {
  $exitcond$i$i30 = ($__i$0$i$i29|0)==(3);
  if ($exitcond$i$i30) {
   break;
  }
  $6 = (($__sym) + ($__i$0$i$i29<<2)|0);
  HEAP32[$6>>2] = 0;
  $7 = (($__i$0$i$i29) + 1)|0;
  $__i$0$i$i29 = $7;
 }
 $__i$0$i$i31 = 0;
 while(1) {
  $exitcond$i$i32 = ($__i$0$i$i31|0)==(3);
  if ($exitcond$i$i32) {
   break;
  }
  $8 = (($__psn) + ($__i$0$i$i31<<2)|0);
  HEAP32[$8>>2] = 0;
  $9 = (($__i$0$i$i31) + 1)|0;
  $__i$0$i$i31 = $9;
 }
 $__i$0$i$i35 = 0;
 while(1) {
  $exitcond$i$i36 = ($__i$0$i$i35|0)==(3);
  if ($exitcond$i$i36) {
   break;
  }
  $10 = (($__nsn) + ($__i$0$i$i35<<2)|0);
  HEAP32[$10>>2] = 0;
  $11 = (($__i$0$i$i35) + 1)|0;
  $__i$0$i$i35 = $11;
 }
 $__i$0$i$i40 = 0;
 while(1) {
  $exitcond$i$i41 = ($__i$0$i$i40|0)==(3);
  if ($exitcond$i$i41) {
   break;
  }
  $12 = (($__spaces) + ($__i$0$i$i40<<2)|0);
  HEAP32[$12>>2] = 0;
  $13 = (($__i$0$i$i40) + 1)|0;
  $__i$0$i$i40 = $13;
 }
 __THREW__ = 0;
 invoke_viiiiiiiiii(4907,($__intl|0),($__loc|0),($__pat|0),($__dp|0),($__ts|0),($__grp|0),($__sym|0),($__psn|0),($__nsn|0),($__fd|0));
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 L21: do {
  if ($15) {
   $74 = ___cxa_find_matching_catch_2()|0;
   $75 = tempRet0;
   $lpad$phi$index = $74;$lpad$phi$index8 = $75;
  } else {
   $16 = HEAP32[$__wb>>2]|0;
   HEAP32[$__wn>>2] = $16;
   $17 = ((($__psn)) + 4|0);
   $18 = ((($__nsn)) + 4|0);
   $19 = ((($__nsn)) + 8|0);
   $20 = ((($__psn)) + 8|0);
   $21 = $__flags & 512;
   $22 = ($21|0)!=(0);
   $23 = ((($__sym)) + 8|0);
   $24 = ((($__sym)) + 4|0);
   $25 = ((($__spaces)) + 4|0);
   $26 = ((($__spaces)) + 8|0);
   $27 = ((($__pat)) + 3|0);
   $28 = ((($__grp)) + 4|0);
   $847 = $__gbuf;$__p$0 = 0;$__trailing_sign$0 = 0;
   L24: while(1) {
    $29 = ($__p$0>>>0)<(4);
    if (!($29)) {
     $$lcssa461 = $847;$__trailing_sign$0$lcssa = $__trailing_sign$0;
     label = 259;
     break;
    }
    $30 = HEAP32[$__b>>2]|0;
    $31 = ($30|0)==(0|0);
    do {
     if ($31) {
      $848 = 1;
     } else {
      $32 = ((($30)) + 12|0);
      $33 = HEAP32[$32>>2]|0;
      $34 = ((($30)) + 16|0);
      $35 = HEAP32[$34>>2]|0;
      $36 = ($33|0)==($35|0);
      if ($36) {
       $37 = HEAP32[$30>>2]|0;
       $38 = ((($37)) + 36|0);
       $39 = HEAP32[$38>>2]|0;
       __THREW__ = 0;
       $40 = (invoke_ii($39|0,($30|0))|0);
       $41 = __THREW__; __THREW__ = 0;
       $42 = $41&1;
       if ($42) {
        label = 40;
        break L24;
       } else {
        $$0$i$i$i$i45 = $40;
       }
      } else {
       $43 = HEAP32[$33>>2]|0;
       $$0$i$i$i$i45 = $43;
      }
      $44 = ($$0$i$i$i$i45|0)==(-1);
      if ($44) {
       HEAP32[$__b>>2] = 0;
       $848 = 1;
       break;
      } else {
       $$pre$i$i$i47 = HEAP32[$__b>>2]|0;
       $phitmp$i$i$i48 = ($$pre$i$i$i47|0)==(0|0);
       $848 = $phitmp$i$i$i48;
       break;
      }
     }
    } while(0);
    $45 = HEAP32[$__e>>2]|0;
    $46 = ($45|0)==(0|0);
    do {
     if ($46) {
      label = 32;
     } else {
      $47 = ((($45)) + 12|0);
      $48 = HEAP32[$47>>2]|0;
      $49 = ((($45)) + 16|0);
      $50 = HEAP32[$49>>2]|0;
      $51 = ($48|0)==($50|0);
      if ($51) {
       $52 = HEAP32[$45>>2]|0;
       $53 = ((($52)) + 36|0);
       $54 = HEAP32[$53>>2]|0;
       __THREW__ = 0;
       $55 = (invoke_ii($54|0,($45|0))|0);
       $56 = __THREW__; __THREW__ = 0;
       $57 = $56&1;
       if ($57) {
        label = 40;
        break L24;
       } else {
        $$0$i$i1$i$i51 = $55;
       }
      } else {
       $58 = HEAP32[$48>>2]|0;
       $$0$i$i1$i$i51 = $58;
      }
      $59 = ($$0$i$i1$i$i51|0)==(-1);
      if ($59) {
       HEAP32[$__e>>2] = 0;
       label = 32;
       break;
      } else {
       if ($848) {
        $849 = $45;
        break;
       } else {
        $$lcssa461 = $847;$__trailing_sign$0$lcssa = $__trailing_sign$0;
        label = 259;
        break L24;
       }
      }
     }
    } while(0);
    if ((label|0) == 32) {
     label = 0;
     if ($848) {
      $$lcssa461 = $847;$__trailing_sign$0$lcssa = $__trailing_sign$0;
      label = 259;
      break;
     } else {
      $849 = 0;
     }
    }
    $83 = (($__pat) + ($__p$0)|0);
    $84 = HEAP8[$83>>0]|0;
    $85 = $84 << 24 >> 24;
    L49: do {
     switch ($85|0) {
     case 1:  {
      $cond28 = ($__p$0|0)==(3);
      if ($cond28) {
       $850 = $847;$__trailing_sign$1 = $__trailing_sign$0;
      } else {
       $87 = HEAP32[$__b>>2]|0;
       $88 = ((($87)) + 12|0);
       $89 = HEAP32[$88>>2]|0;
       $90 = ((($87)) + 16|0);
       $91 = HEAP32[$90>>2]|0;
       $92 = ($89|0)==($91|0);
       if ($92) {
        $93 = HEAP32[$87>>2]|0;
        $94 = ((($93)) + 36|0);
        $95 = HEAP32[$94>>2]|0;
        __THREW__ = 0;
        $96 = (invoke_ii($95|0,($87|0))|0);
        $97 = __THREW__; __THREW__ = 0;
        $98 = $97&1;
        if ($98) {
         label = 40;
         break L24;
        } else {
         $$0$i$i60 = $96;
        }
       } else {
        $99 = HEAP32[$89>>2]|0;
        $$0$i$i60 = $99;
       }
       $100 = HEAP32[$__ct>>2]|0;
       $101 = ((($100)) + 12|0);
       $102 = HEAP32[$101>>2]|0;
       __THREW__ = 0;
       $103 = (invoke_iiii($102|0,($__ct|0),8192,($$0$i$i60|0))|0);
       $104 = __THREW__; __THREW__ = 0;
       $105 = $104&1;
       if ($105) {
        label = 40;
        break L24;
       }
       if (!($103)) {
        label = 58;
        break L24;
       }
       $106 = HEAP32[$__b>>2]|0;
       $107 = ((($106)) + 12|0);
       $108 = HEAP32[$107>>2]|0;
       $109 = ((($106)) + 16|0);
       $110 = HEAP32[$109>>2]|0;
       $111 = ($108|0)==($110|0);
       if ($111) {
        $112 = HEAP32[$106>>2]|0;
        $113 = ((($112)) + 40|0);
        $114 = HEAP32[$113>>2]|0;
        __THREW__ = 0;
        $115 = (invoke_ii($114|0,($106|0))|0);
        $116 = __THREW__; __THREW__ = 0;
        $117 = $116&1;
        if ($117) {
         label = 40;
         break L24;
        } else {
         $$0$i$i64 = $115;
        }
       } else {
        $118 = ((($108)) + 4|0);
        HEAP32[$107>>2] = $118;
        $119 = HEAP32[$108>>2]|0;
        $$0$i$i64 = $119;
       }
       __THREW__ = 0;
       invoke_vii(4908,($__spaces|0),($$0$i$i64|0));
       $120 = __THREW__; __THREW__ = 0;
       $121 = $120&1;
       if ($121) {
        label = 40;
        break L24;
       } else {
        $$pre$i3$i$i74 = $849;$141 = $849;
        label = 60;
       }
      }
      break;
     }
     case 0:  {
      $124 = ($__p$0|0)==(3);
      if ($124) {
       $850 = $847;$__trailing_sign$1 = $__trailing_sign$0;
      } else {
       $$pre$i3$i$i74 = $849;$141 = $849;
       label = 60;
      }
      break;
     }
     case 3:  {
      $191 = HEAP8[$__psn>>0]|0;
      $192 = $191 & 1;
      $193 = ($192<<24>>24)==(0);
      $194 = HEAP32[$17>>2]|0;
      $195 = $191&255;
      $196 = $195 >>> 1;
      $197 = $193 ? $196 : $194;
      $198 = HEAP8[$__nsn>>0]|0;
      $199 = $198 & 1;
      $200 = ($199<<24>>24)==(0);
      $201 = HEAP32[$18>>2]|0;
      $202 = $198&255;
      $203 = $202 >>> 1;
      $204 = $200 ? $203 : $201;
      $205 = (0 - ($204))|0;
      $206 = ($197|0)==($205|0);
      if ($206) {
       $850 = $847;$__trailing_sign$1 = $__trailing_sign$0;
      } else {
       $207 = ($197|0)==(0);
       if ($207) {
        $249 = HEAP32[$__b>>2]|0;
        $250 = ((($249)) + 12|0);
        $251 = HEAP32[$250>>2]|0;
        $252 = ((($249)) + 16|0);
        $253 = HEAP32[$252>>2]|0;
        $254 = ($251|0)==($253|0);
        if ($254) {
         $255 = HEAP32[$249>>2]|0;
         $256 = ((($255)) + 36|0);
         $257 = HEAP32[$256>>2]|0;
         __THREW__ = 0;
         $258 = (invoke_ii($257|0,($249|0))|0);
         $259 = __THREW__; __THREW__ = 0;
         $260 = $259&1;
         if ($260) {
          label = 40;
          break L24;
         }
         $$pre326 = HEAP8[$__nsn>>0]|0;
         $$0$i$i125 = $258;$263 = $$pre326;
        } else {
         $261 = HEAP32[$251>>2]|0;
         $$0$i$i125 = $261;$263 = $198;
        }
        $262 = $263 & 1;
        $264 = ($262<<24>>24)==(0);
        $265 = HEAP32[$19>>2]|0;
        $266 = $264 ? $18 : $265;
        $267 = HEAP32[$266>>2]|0;
        $268 = ($$0$i$i125|0)==($267|0);
        if (!($268)) {
         $850 = $847;$__trailing_sign$1 = $__trailing_sign$0;
         break L49;
        }
        $269 = HEAP32[$__b>>2]|0;
        $270 = ((($269)) + 12|0);
        $271 = HEAP32[$270>>2]|0;
        $272 = ((($269)) + 16|0);
        $273 = HEAP32[$272>>2]|0;
        $274 = ($271|0)==($273|0);
        if ($274) {
         $275 = HEAP32[$269>>2]|0;
         $276 = ((($275)) + 40|0);
         $277 = HEAP32[$276>>2]|0;
         __THREW__ = 0;
         (invoke_ii($277|0,($269|0))|0);
         $278 = __THREW__; __THREW__ = 0;
         $279 = $278&1;
         if ($279) {
          label = 40;
          break L24;
         }
        } else {
         $280 = ((($271)) + 4|0);
         HEAP32[$270>>2] = $280;
        }
        HEAP8[$__neg>>0] = 1;
        $281 = HEAP8[$__nsn>>0]|0;
        $282 = $281 & 1;
        $283 = ($282<<24>>24)==(0);
        $284 = HEAP32[$18>>2]|0;
        $285 = $281&255;
        $286 = $285 >>> 1;
        $287 = $283 ? $286 : $284;
        $288 = ($287>>>0)>(1);
        $__nsn$__trailing_sign$0 = $288 ? $__nsn : $__trailing_sign$0;
        $850 = $847;$__trailing_sign$1 = $__nsn$__trailing_sign$0;
        break L49;
       }
       $208 = ($204|0)==(0);
       $209 = HEAP32[$__b>>2]|0;
       $210 = ((($209)) + 12|0);
       $211 = HEAP32[$210>>2]|0;
       $212 = ((($209)) + 16|0);
       $213 = HEAP32[$212>>2]|0;
       $214 = ($211|0)==($213|0);
       if ($208) {
        if ($214) {
         $215 = HEAP32[$209>>2]|0;
         $216 = ((($215)) + 36|0);
         $217 = HEAP32[$216>>2]|0;
         __THREW__ = 0;
         $218 = (invoke_ii($217|0,($209|0))|0);
         $219 = __THREW__; __THREW__ = 0;
         $220 = $219&1;
         if ($220) {
          label = 40;
          break L24;
         }
         $$pre325 = HEAP8[$__psn>>0]|0;
         $$0$i$i104 = $218;$223 = $$pre325;
        } else {
         $221 = HEAP32[$211>>2]|0;
         $$0$i$i104 = $221;$223 = $191;
        }
        $222 = $223 & 1;
        $224 = ($222<<24>>24)==(0);
        $225 = HEAP32[$20>>2]|0;
        $226 = $224 ? $17 : $225;
        $227 = HEAP32[$226>>2]|0;
        $228 = ($$0$i$i104|0)==($227|0);
        if (!($228)) {
         HEAP8[$__neg>>0] = 1;
         $850 = $847;$__trailing_sign$1 = $__trailing_sign$0;
         break L49;
        }
        $229 = HEAP32[$__b>>2]|0;
        $230 = ((($229)) + 12|0);
        $231 = HEAP32[$230>>2]|0;
        $232 = ((($229)) + 16|0);
        $233 = HEAP32[$232>>2]|0;
        $234 = ($231|0)==($233|0);
        if ($234) {
         $235 = HEAP32[$229>>2]|0;
         $236 = ((($235)) + 40|0);
         $237 = HEAP32[$236>>2]|0;
         __THREW__ = 0;
         (invoke_ii($237|0,($229|0))|0);
         $238 = __THREW__; __THREW__ = 0;
         $239 = $238&1;
         if ($239) {
          label = 40;
          break L24;
         }
        } else {
         $240 = ((($231)) + 4|0);
         HEAP32[$230>>2] = $240;
        }
        $241 = HEAP8[$__psn>>0]|0;
        $242 = $241 & 1;
        $243 = ($242<<24>>24)==(0);
        $244 = HEAP32[$17>>2]|0;
        $245 = $241&255;
        $246 = $245 >>> 1;
        $247 = $243 ? $246 : $244;
        $248 = ($247>>>0)>(1);
        $__psn$__trailing_sign$0 = $248 ? $__psn : $__trailing_sign$0;
        $850 = $847;$__trailing_sign$1 = $__psn$__trailing_sign$0;
        break L49;
       }
       if ($214) {
        $289 = HEAP32[$209>>2]|0;
        $290 = ((($289)) + 36|0);
        $291 = HEAP32[$290>>2]|0;
        __THREW__ = 0;
        $292 = (invoke_ii($291|0,($209|0))|0);
        $293 = __THREW__; __THREW__ = 0;
        $294 = $293&1;
        if ($294) {
         label = 40;
         break L24;
        }
        $$pre320 = HEAP8[$__psn>>0]|0;
        $$pre321 = HEAP32[$__b>>2]|0;
        $$phi$trans$insert = ((($$pre321)) + 12|0);
        $$pre322 = HEAP32[$$phi$trans$insert>>2]|0;
        $$phi$trans$insert323 = ((($$pre321)) + 16|0);
        $$pre324 = HEAP32[$$phi$trans$insert323>>2]|0;
        $$0$i$i137 = $292;$297 = $$pre320;$304 = $$pre321;$306 = $$pre322;$307 = $$pre324;
       } else {
        $295 = HEAP32[$211>>2]|0;
        $$0$i$i137 = $295;$297 = $191;$304 = $209;$306 = $211;$307 = $213;
       }
       $296 = $297 & 1;
       $298 = ($296<<24>>24)==(0);
       $299 = HEAP32[$20>>2]|0;
       $300 = $298 ? $17 : $299;
       $301 = HEAP32[$300>>2]|0;
       $302 = ($$0$i$i137|0)==($301|0);
       $303 = ((($304)) + 12|0);
       $305 = ($306|0)==($307|0);
       if ($302) {
        if ($305) {
         $308 = HEAP32[$304>>2]|0;
         $309 = ((($308)) + 40|0);
         $310 = HEAP32[$309>>2]|0;
         __THREW__ = 0;
         (invoke_ii($310|0,($304|0))|0);
         $311 = __THREW__; __THREW__ = 0;
         $312 = $311&1;
         if ($312) {
          label = 40;
          break L24;
         }
        } else {
         $313 = ((($306)) + 4|0);
         HEAP32[$303>>2] = $313;
        }
        $314 = HEAP8[$__psn>>0]|0;
        $315 = $314 & 1;
        $316 = ($315<<24>>24)==(0);
        $317 = HEAP32[$17>>2]|0;
        $318 = $314&255;
        $319 = $318 >>> 1;
        $320 = $316 ? $319 : $317;
        $321 = ($320>>>0)>(1);
        $__psn$__trailing_sign$022 = $321 ? $__psn : $__trailing_sign$0;
        $850 = $847;$__trailing_sign$1 = $__psn$__trailing_sign$022;
        break L49;
       }
       if ($305) {
        $322 = HEAP32[$304>>2]|0;
        $323 = ((($322)) + 36|0);
        $324 = HEAP32[$323>>2]|0;
        __THREW__ = 0;
        $325 = (invoke_ii($324|0,($304|0))|0);
        $326 = __THREW__; __THREW__ = 0;
        $327 = $326&1;
        if ($327) {
         label = 40;
         break L24;
        } else {
         $$0$i$i159 = $325;
        }
       } else {
        $328 = HEAP32[$306>>2]|0;
        $$0$i$i159 = $328;
       }
       $329 = HEAP8[$__nsn>>0]|0;
       $330 = $329 & 1;
       $331 = ($330<<24>>24)==(0);
       $332 = HEAP32[$19>>2]|0;
       $333 = $331 ? $18 : $332;
       $334 = HEAP32[$333>>2]|0;
       $335 = ($$0$i$i159|0)==($334|0);
       if (!($335)) {
        label = 123;
        break L24;
       }
       $336 = HEAP32[$__b>>2]|0;
       $337 = ((($336)) + 12|0);
       $338 = HEAP32[$337>>2]|0;
       $339 = ((($336)) + 16|0);
       $340 = HEAP32[$339>>2]|0;
       $341 = ($338|0)==($340|0);
       if ($341) {
        $342 = HEAP32[$336>>2]|0;
        $343 = ((($342)) + 40|0);
        $344 = HEAP32[$343>>2]|0;
        __THREW__ = 0;
        (invoke_ii($344|0,($336|0))|0);
        $345 = __THREW__; __THREW__ = 0;
        $346 = $345&1;
        if ($346) {
         label = 40;
         break L24;
        }
       } else {
        $347 = ((($338)) + 4|0);
        HEAP32[$337>>2] = $347;
       }
       HEAP8[$__neg>>0] = 1;
       $348 = HEAP8[$__nsn>>0]|0;
       $349 = $348 & 1;
       $350 = ($349<<24>>24)==(0);
       $351 = HEAP32[$18>>2]|0;
       $352 = $348&255;
       $353 = $352 >>> 1;
       $354 = $350 ? $353 : $351;
       $355 = ($354>>>0)>(1);
       $__nsn$__trailing_sign$023 = $355 ? $__nsn : $__trailing_sign$0;
       $850 = $847;$__trailing_sign$1 = $__nsn$__trailing_sign$023;
      }
      break;
     }
     case 2:  {
      $358 = ($__trailing_sign$0|0)!=(0|0);
      $359 = ($__p$0>>>0)<(2);
      $or$cond = $359 | $358;
      if (!($or$cond)) {
       $360 = ($__p$0|0)==(2);
       $361 = HEAP8[$27>>0]|0;
       $362 = ($361<<24>>24)!=(0);
       $363 = $360 & $362;
       $brmerge = $22 | $363;
       if (!($brmerge)) {
        $850 = $847;$__trailing_sign$1 = 0;
        break L49;
       }
      }
      $364 = HEAP8[$__sym>>0]|0;
      $365 = $364 & 1;
      $366 = ($365<<24>>24)==(0);
      $367 = HEAP32[$23>>2]|0;
      $368 = $366 ? $24 : $367;
      $369 = $368;
      $370 = ($__p$0|0)==(0);
      L118: do {
       if ($370) {
        $$pre$i3$i$i206 = $849;$424 = $364;$430 = $367;$451 = $849;$__sym_curr_char$sroa$0$0 = $369;
       } else {
        $371 = (($__p$0) + -1)|0;
        $372 = (($__pat) + ($371)|0);
        $373 = HEAP8[$372>>0]|0;
        $switch = ($373&255)<(2);
        if ($switch) {
         $375 = $364;$381 = $367;$__sym_space_end$sroa$0$0 = $369;
         while(1) {
          $374 = $375 & 1;
          $376 = ($374<<24>>24)==(0);
          $377 = HEAP32[$24>>2]|0;
          $378 = $375&255;
          $379 = $378 >>> 1;
          $380 = $376 ? $24 : $381;
          $382 = $376 ? $379 : $377;
          $383 = (($380) + ($382<<2)|0);
          $384 = $__sym_space_end$sroa$0$0;
          $385 = ($384|0)==($383|0);
          if ($385) {
           $396 = $375;$399 = $381;$__sym_space_end$sroa$0$0460 = $__sym_space_end$sroa$0$0;
           break;
          }
          $386 = HEAP32[$384>>2]|0;
          $387 = HEAP32[$__ct>>2]|0;
          $388 = ((($387)) + 12|0);
          $389 = HEAP32[$388>>2]|0;
          __THREW__ = 0;
          $390 = (invoke_iiii($389|0,($__ct|0),8192,($386|0))|0);
          $391 = __THREW__; __THREW__ = 0;
          $392 = $391&1;
          if ($392) {
           label = 37;
           break L24;
          }
          if (!($390)) {
           $__sym_space_end$sroa$0$0$lcssa459 = $__sym_space_end$sroa$0$0;
           label = 131;
           break;
          }
          $393 = ((($384)) + 4|0);
          $394 = $393;
          $$pre311 = HEAP8[$__sym>>0]|0;
          $$pre312 = HEAP32[$23>>2]|0;
          $375 = $$pre311;$381 = $$pre312;$__sym_space_end$sroa$0$0 = $394;
         }
         if ((label|0) == 131) {
          label = 0;
          $$pre313 = HEAP8[$__sym>>0]|0;
          $$pre314 = HEAP32[$23>>2]|0;
          $396 = $$pre313;$399 = $$pre314;$__sym_space_end$sroa$0$0460 = $__sym_space_end$sroa$0$0$lcssa459;
         }
         $395 = $396 & 1;
         $397 = ($395<<24>>24)==(0);
         $398 = $397 ? $24 : $399;
         $400 = $398;
         $401 = (($__sym_space_end$sroa$0$0460) - ($400))|0;
         $402 = $401 >> 2;
         $403 = HEAP8[$__spaces>>0]|0;
         $404 = $403 & 1;
         $405 = ($404<<24>>24)==(0);
         $406 = HEAP32[$25>>2]|0;
         $407 = $403&255;
         $408 = $407 >>> 1;
         $409 = $405 ? $408 : $406;
         $410 = ($402>>>0)>($409>>>0);
         if ($410) {
          $$pre$i3$i$i206 = $849;$424 = $396;$430 = $399;$451 = $849;$__sym_curr_char$sroa$0$0 = $400;
         } else {
          $411 = HEAP32[$26>>2]|0;
          $412 = $405 ? $25 : $411;
          $413 = (($412) + ($409<<2)|0);
          $414 = (0 - ($402))|0;
          $415 = (($413) + ($414<<2)|0);
          $$sink = $398;$417 = $415;
          while(1) {
           $416 = ($417|0)==($413|0);
           if ($416) {
            $$pre$i3$i$i206 = $849;$424 = $396;$430 = $399;$451 = $849;$__sym_curr_char$sroa$0$0 = $__sym_space_end$sroa$0$0460;
            break L118;
           }
           $418 = HEAP32[$417>>2]|0;
           $419 = HEAP32[$$sink>>2]|0;
           $420 = ($418|0)==($419|0);
           if (!($420)) {
            $$pre$i3$i$i206 = $849;$424 = $396;$430 = $399;$451 = $849;$__sym_curr_char$sroa$0$0 = $400;
            break L118;
           }
           $421 = ((($417)) + 4|0);
           $422 = ((($$sink)) + 4|0);
           $$sink = $422;$417 = $421;
          }
         }
        } else {
         $$pre$i3$i$i206 = $849;$424 = $364;$430 = $367;$451 = $849;$__sym_curr_char$sroa$0$0 = $369;
        }
       }
      } while(0);
      L133: while(1) {
       $423 = $424 & 1;
       $425 = ($423<<24>>24)==(0);
       $426 = HEAP32[$24>>2]|0;
       $427 = $424&255;
       $428 = $427 >>> 1;
       $429 = $425 ? $24 : $430;
       $431 = $425 ? $428 : $426;
       $432 = (($429) + ($431<<2)|0);
       $433 = $__sym_curr_char$sroa$0$0;
       $434 = ($433|0)==($432|0);
       if ($434) {
        $$lcssa = $432;
        break;
       }
       $435 = HEAP32[$__b>>2]|0;
       $436 = ($435|0)==(0|0);
       do {
        if ($436) {
         $465 = 1;
        } else {
         $437 = ((($435)) + 12|0);
         $438 = HEAP32[$437>>2]|0;
         $439 = ((($435)) + 16|0);
         $440 = HEAP32[$439>>2]|0;
         $441 = ($438|0)==($440|0);
         if ($441) {
          $442 = HEAP32[$435>>2]|0;
          $443 = ((($442)) + 36|0);
          $444 = HEAP32[$443>>2]|0;
          __THREW__ = 0;
          $445 = (invoke_ii($444|0,($435|0))|0);
          $446 = __THREW__; __THREW__ = 0;
          $447 = $446&1;
          if ($447) {
           label = 36;
           break L24;
          } else {
           $$0$i$i$i$i198 = $445;
          }
         } else {
          $448 = HEAP32[$438>>2]|0;
          $$0$i$i$i$i198 = $448;
         }
         $449 = ($$0$i$i$i$i198|0)==(-1);
         if ($449) {
          HEAP32[$__b>>2] = 0;
          $465 = 1;
          break;
         } else {
          $$pre$i$i$i200 = HEAP32[$__b>>2]|0;
          $phitmp$i$i$i201 = ($$pre$i$i$i200|0)==(0|0);
          $465 = $phitmp$i$i$i201;
          break;
         }
        }
       } while(0);
       $450 = ($451|0)==(0|0);
       do {
        if ($450) {
         $$pre$i3$i$i206318 = $$pre$i3$i$i206;
         label = 152;
        } else {
         $452 = ((($451)) + 12|0);
         $453 = HEAP32[$452>>2]|0;
         $454 = ((($451)) + 16|0);
         $455 = HEAP32[$454>>2]|0;
         $456 = ($453|0)==($455|0);
         if ($456) {
          $457 = HEAP32[$451>>2]|0;
          $458 = ((($457)) + 36|0);
          $459 = HEAP32[$458>>2]|0;
          __THREW__ = 0;
          $460 = (invoke_ii($459|0,($451|0))|0);
          $461 = __THREW__; __THREW__ = 0;
          $462 = $461&1;
          if ($462) {
           label = 36;
           break L24;
          } else {
           $$0$i$i1$i$i204 = $460;
          }
         } else {
          $463 = HEAP32[$453>>2]|0;
          $$0$i$i1$i$i204 = $463;
         }
         $464 = ($$0$i$i1$i$i204|0)==(-1);
         if ($464) {
          HEAP32[$__e>>2] = 0;
          $$pre$i3$i$i206318 = 0;
          label = 152;
          break;
         } else {
          $phitmp$i4$i$i207 = ($$pre$i3$i$i206|0)==(0|0);
          $tmp$i$i209 = $465 ^ $phitmp$i4$i$i207;
          if ($tmp$i$i209) {
           $$pre$i3$i$i206317 = $$pre$i3$i$i206;$852 = $$pre$i3$i$i206;
           break;
          } else {
           $$lcssa = $433;
           break L133;
          }
         }
        }
       } while(0);
       if ((label|0) == 152) {
        label = 0;
        if ($465) {
         $$lcssa = $433;
         break;
        } else {
         $$pre$i3$i$i206317 = $$pre$i3$i$i206318;$852 = 0;
        }
       }
       $466 = HEAP32[$__b>>2]|0;
       $467 = ((($466)) + 12|0);
       $468 = HEAP32[$467>>2]|0;
       $469 = ((($466)) + 16|0);
       $470 = HEAP32[$469>>2]|0;
       $471 = ($468|0)==($470|0);
       if ($471) {
        $472 = HEAP32[$466>>2]|0;
        $473 = ((($472)) + 36|0);
        $474 = HEAP32[$473>>2]|0;
        __THREW__ = 0;
        $475 = (invoke_ii($474|0,($466|0))|0);
        $476 = __THREW__; __THREW__ = 0;
        $477 = $476&1;
        if ($477) {
         label = 36;
         break L24;
        } else {
         $$0$i$i194 = $475;
        }
       } else {
        $478 = HEAP32[$468>>2]|0;
        $$0$i$i194 = $478;
       }
       $479 = HEAP32[$433>>2]|0;
       $480 = ($$0$i$i194|0)==($479|0);
       if (!($480)) {
        $$lcssa = $433;
        break;
       }
       $481 = HEAP32[$__b>>2]|0;
       $482 = ((($481)) + 12|0);
       $483 = HEAP32[$482>>2]|0;
       $484 = ((($481)) + 16|0);
       $485 = HEAP32[$484>>2]|0;
       $486 = ($483|0)==($485|0);
       if ($486) {
        $487 = HEAP32[$481>>2]|0;
        $488 = ((($487)) + 40|0);
        $489 = HEAP32[$488>>2]|0;
        __THREW__ = 0;
        (invoke_ii($489|0,($481|0))|0);
        $490 = __THREW__; __THREW__ = 0;
        $491 = $490&1;
        if ($491) {
         label = 36;
         break L24;
        }
       } else {
        $492 = ((($483)) + 4|0);
        HEAP32[$482>>2] = $492;
       }
       $493 = ((($433)) + 4|0);
       $494 = $493;
       $$pre315 = HEAP8[$__sym>>0]|0;
       $$pre316 = HEAP32[$23>>2]|0;
       $$pre$i3$i$i206 = $$pre$i3$i$i206317;$424 = $$pre315;$430 = $$pre316;$451 = $852;$__sym_curr_char$sroa$0$0 = $494;
      }
      if ($22) {
       $495 = HEAP8[$__sym>>0]|0;
       $496 = $495 & 1;
       $497 = ($496<<24>>24)==(0);
       $498 = HEAP32[$23>>2]|0;
       $499 = HEAP32[$24>>2]|0;
       $500 = $495&255;
       $501 = $500 >>> 1;
       $502 = $497 ? $24 : $498;
       $503 = $497 ? $501 : $499;
       $504 = (($502) + ($503<<2)|0);
       $505 = ($$lcssa|0)==($504|0);
       if ($505) {
        $850 = $847;$__trailing_sign$1 = $__trailing_sign$0;
       } else {
        label = 164;
        break L24;
       }
      } else {
       $850 = $847;$__trailing_sign$1 = $__trailing_sign$0;
      }
      break;
     }
     case 4:  {
      $86 = HEAP32[$__ts>>2]|0;
      $$pre$i3$i$i181 = $849;$524 = $849;$578 = $847;$__ng$0 = 0;
      L170: while(1) {
       $508 = HEAP32[$__b>>2]|0;
       $509 = ($508|0)==(0|0);
       do {
        if ($509) {
         $538 = 1;
        } else {
         $510 = ((($508)) + 12|0);
         $511 = HEAP32[$510>>2]|0;
         $512 = ((($508)) + 16|0);
         $513 = HEAP32[$512>>2]|0;
         $514 = ($511|0)==($513|0);
         if ($514) {
          $515 = HEAP32[$508>>2]|0;
          $516 = ((($515)) + 36|0);
          $517 = HEAP32[$516>>2]|0;
          __THREW__ = 0;
          $518 = (invoke_ii($517|0,($508|0))|0);
          $519 = __THREW__; __THREW__ = 0;
          $520 = $519&1;
          if ($520) {
           label = 39;
           break L24;
          } else {
           $$0$i$i$i$i173 = $518;
          }
         } else {
          $521 = HEAP32[$511>>2]|0;
          $$0$i$i$i$i173 = $521;
         }
         $522 = ($$0$i$i$i$i173|0)==(-1);
         if ($522) {
          HEAP32[$__b>>2] = 0;
          $538 = 1;
          break;
         } else {
          $$pre$i$i$i175 = HEAP32[$__b>>2]|0;
          $phitmp$i$i$i176 = ($$pre$i$i$i175|0)==(0|0);
          $538 = $phitmp$i$i$i176;
          break;
         }
        }
       } while(0);
       $523 = ($524|0)==(0|0);
       do {
        if ($523) {
         $$pre$i3$i$i181302 = $$pre$i3$i$i181;
         label = 178;
        } else {
         $525 = ((($524)) + 12|0);
         $526 = HEAP32[$525>>2]|0;
         $527 = ((($524)) + 16|0);
         $528 = HEAP32[$527>>2]|0;
         $529 = ($526|0)==($528|0);
         if ($529) {
          $530 = HEAP32[$524>>2]|0;
          $531 = ((($530)) + 36|0);
          $532 = HEAP32[$531>>2]|0;
          __THREW__ = 0;
          $533 = (invoke_ii($532|0,($524|0))|0);
          $534 = __THREW__; __THREW__ = 0;
          $535 = $534&1;
          if ($535) {
           label = 39;
           break L24;
          } else {
           $$0$i$i1$i$i179 = $533;
          }
         } else {
          $536 = HEAP32[$526>>2]|0;
          $$0$i$i1$i$i179 = $536;
         }
         $537 = ($$0$i$i1$i$i179|0)==(-1);
         if ($537) {
          HEAP32[$__e>>2] = 0;
          $$pre$i3$i$i181302 = 0;
          label = 178;
          break;
         } else {
          $phitmp$i4$i$i182 = ($$pre$i3$i$i181|0)==(0|0);
          $tmp$i$i184 = $538 ^ $phitmp$i4$i$i182;
          if ($tmp$i$i184) {
           $$pre$i3$i$i181304 = $$pre$i3$i$i181;$853 = $$pre$i3$i$i181;
           break;
          } else {
           $$lcssa450 = $578;$$pre$i3$i$i149 = $$pre$i3$i$i181;$__ng$0$lcssa451 = $__ng$0;
           break L170;
          }
         }
        }
       } while(0);
       if ((label|0) == 178) {
        label = 0;
        if ($538) {
         $$lcssa450 = $578;$$pre$i3$i$i149 = $$pre$i3$i$i181302;$__ng$0$lcssa451 = $__ng$0;
         break;
        } else {
         $$pre$i3$i$i181304 = $$pre$i3$i$i181302;$853 = 0;
        }
       }
       $539 = HEAP32[$__b>>2]|0;
       $540 = ((($539)) + 12|0);
       $541 = HEAP32[$540>>2]|0;
       $542 = ((($539)) + 16|0);
       $543 = HEAP32[$542>>2]|0;
       $544 = ($541|0)==($543|0);
       if ($544) {
        $545 = HEAP32[$539>>2]|0;
        $546 = ((($545)) + 36|0);
        $547 = HEAP32[$546>>2]|0;
        __THREW__ = 0;
        $548 = (invoke_ii($547|0,($539|0))|0);
        $549 = __THREW__; __THREW__ = 0;
        $550 = $549&1;
        if ($550) {
         label = 39;
         break L24;
        } else {
         $$0$i$i170 = $548;
        }
       } else {
        $551 = HEAP32[$541>>2]|0;
        $$0$i$i170 = $551;
       }
       $552 = HEAP32[$__ct>>2]|0;
       $553 = ((($552)) + 12|0);
       $554 = HEAP32[$553>>2]|0;
       __THREW__ = 0;
       $555 = (invoke_iiii($554|0,($__ct|0),2048,($$0$i$i170|0))|0);
       $556 = __THREW__; __THREW__ = 0;
       $557 = $556&1;
       if ($557) {
        label = 39;
        break L24;
       }
       if ($555) {
        $558 = HEAP32[$__wn>>2]|0;
        $559 = HEAP32[$0>>2]|0;
        $560 = ($558|0)==($559|0);
        if ($560) {
         __THREW__ = 0;
         invoke_viii(4909,($__wb|0),($__wn|0),($0|0));
         $561 = __THREW__; __THREW__ = 0;
         $562 = $561&1;
         if ($562) {
          label = 39;
          break L24;
         }
         $$pre306 = HEAP32[$__wn>>2]|0;
         $564 = $$pre306;
        } else {
         $564 = $558;
        }
        $563 = ((($564)) + 4|0);
        HEAP32[$__wn>>2] = $563;
        HEAP32[$564>>2] = $$0$i$i170;
        $565 = (($__ng$0) + 1)|0;
        $854 = $578;$__ng$2$ph = $565;
       } else {
        $566 = HEAP8[$__grp>>0]|0;
        $567 = $566 & 1;
        $568 = ($567<<24>>24)==(0);
        $569 = HEAP32[$28>>2]|0;
        $570 = $566&255;
        $571 = $570 >>> 1;
        $572 = $568 ? $571 : $569;
        $573 = ($572|0)!=(0);
        $574 = ($__ng$0|0)!=(0);
        $or$cond12 = $574 & $573;
        $575 = ($$0$i$i170|0)==($86|0);
        $or$cond25 = $575 & $or$cond12;
        if (!($or$cond25)) {
         $$lcssa450 = $578;$$pre$i3$i$i149 = $$pre$i3$i$i181304;$__ng$0$lcssa451 = $__ng$0;
         break;
        }
        $576 = HEAP32[$__ge>>2]|0;
        $577 = ($578|0)==($576|0);
        if ($577) {
         __THREW__ = 0;
         invoke_viii(4904,($__gb|0),($__gn|0),($__ge|0));
         $579 = __THREW__; __THREW__ = 0;
         $580 = $579&1;
         if ($580) {
          label = 39;
          break L24;
         }
         $$pre = HEAP32[$__gn>>2]|0;
         $582 = $$pre;
        } else {
         $582 = $578;
        }
        $581 = ((($582)) + 4|0);
        HEAP32[$__gn>>2] = $581;
        HEAP32[$582>>2] = $__ng$0;
        $854 = $581;$__ng$2$ph = 0;
       }
       $583 = HEAP32[$__b>>2]|0;
       $584 = ((($583)) + 12|0);
       $585 = HEAP32[$584>>2]|0;
       $586 = ((($583)) + 16|0);
       $587 = HEAP32[$586>>2]|0;
       $588 = ($585|0)==($587|0);
       if ($588) {
        $589 = HEAP32[$583>>2]|0;
        $590 = ((($589)) + 40|0);
        $591 = HEAP32[$590>>2]|0;
        __THREW__ = 0;
        (invoke_ii($591|0,($583|0))|0);
        $592 = __THREW__; __THREW__ = 0;
        $593 = $592&1;
        if ($593) {
         label = 39;
         break L24;
        } else {
         $$pre$i3$i$i181 = $$pre$i3$i$i181304;$524 = $853;$578 = $854;$__ng$0 = $__ng$2$ph;
         continue;
        }
       } else {
        $594 = ((($585)) + 4|0);
        HEAP32[$584>>2] = $594;
        $$pre$i3$i$i181 = $$pre$i3$i$i181304;$524 = $853;$578 = $854;$__ng$0 = $__ng$2$ph;
        continue;
       }
      }
      $595 = HEAP32[$__gb>>2]|0;
      $596 = ($595|0)!=($$lcssa450|0);
      $597 = ($__ng$0$lcssa451|0)!=(0);
      $or$cond14 = $597 & $596;
      if ($or$cond14) {
       $598 = HEAP32[$__ge>>2]|0;
       $599 = ($$lcssa450|0)==($598|0);
       if ($599) {
        __THREW__ = 0;
        invoke_viii(4904,($__gb|0),($__gn|0),($__ge|0));
        $600 = __THREW__; __THREW__ = 0;
        $601 = $600&1;
        if ($601) {
         label = 40;
         break L24;
        }
        $$pre308 = HEAP32[$__gn>>2]|0;
        $603 = $$pre308;
       } else {
        $603 = $$lcssa450;
       }
       $602 = ((($603)) + 4|0);
       HEAP32[$__gn>>2] = $602;
       HEAP32[$603>>2] = $__ng$0$lcssa451;
       $855 = $602;
      } else {
       $855 = $$lcssa450;
      }
      $604 = HEAP32[$__fd>>2]|0;
      $605 = ($604|0)>(0);
      L223: do {
       if ($605) {
        $606 = HEAP32[$__b>>2]|0;
        $607 = ($606|0)==(0|0);
        do {
         if ($607) {
          $856 = 1;
         } else {
          $608 = ((($606)) + 12|0);
          $609 = HEAP32[$608>>2]|0;
          $610 = ((($606)) + 16|0);
          $611 = HEAP32[$610>>2]|0;
          $612 = ($609|0)==($611|0);
          if ($612) {
           $613 = HEAP32[$606>>2]|0;
           $614 = ((($613)) + 36|0);
           $615 = HEAP32[$614>>2]|0;
           __THREW__ = 0;
           $616 = (invoke_ii($615|0,($606|0))|0);
           $617 = __THREW__; __THREW__ = 0;
           $618 = $617&1;
           if ($618) {
            label = 40;
            break L24;
           } else {
            $$0$i$i$i$i141 = $616;
           }
          } else {
           $619 = HEAP32[$609>>2]|0;
           $$0$i$i$i$i141 = $619;
          }
          $620 = ($$0$i$i$i$i141|0)==(-1);
          if ($620) {
           HEAP32[$__b>>2] = 0;
           $856 = 1;
           break;
          } else {
           $$pre$i$i$i143 = HEAP32[$__b>>2]|0;
           $phitmp$i$i$i144 = ($$pre$i$i$i143|0)==(0|0);
           $856 = $phitmp$i$i$i144;
           break;
          }
         }
        } while(0);
        $621 = ($$pre$i3$i$i149|0)==(0|0);
        do {
         if ($621) {
          label = 216;
         } else {
          $622 = ((($$pre$i3$i$i149)) + 12|0);
          $623 = HEAP32[$622>>2]|0;
          $624 = ((($$pre$i3$i$i149)) + 16|0);
          $625 = HEAP32[$624>>2]|0;
          $626 = ($623|0)==($625|0);
          if ($626) {
           $627 = HEAP32[$$pre$i3$i$i149>>2]|0;
           $628 = ((($627)) + 36|0);
           $629 = HEAP32[$628>>2]|0;
           __THREW__ = 0;
           $630 = (invoke_ii($629|0,($$pre$i3$i$i149|0))|0);
           $631 = __THREW__; __THREW__ = 0;
           $632 = $631&1;
           if ($632) {
            label = 40;
            break L24;
           } else {
            $$0$i$i1$i$i147 = $630;
           }
          } else {
           $633 = HEAP32[$623>>2]|0;
           $$0$i$i1$i$i147 = $633;
          }
          $634 = ($$0$i$i1$i$i147|0)==(-1);
          if ($634) {
           HEAP32[$__e>>2] = 0;
           label = 216;
           break;
          } else {
           if ($856) {
            $857 = $$pre$i3$i$i149;
            break;
           } else {
            label = 222;
            break L24;
           }
          }
         }
        } while(0);
        if ((label|0) == 216) {
         label = 0;
         if ($856) {
          label = 222;
          break L24;
         } else {
          $857 = 0;
         }
        }
        $635 = HEAP32[$__b>>2]|0;
        $636 = ((($635)) + 12|0);
        $637 = HEAP32[$636>>2]|0;
        $638 = ((($635)) + 16|0);
        $639 = HEAP32[$638>>2]|0;
        $640 = ($637|0)==($639|0);
        if ($640) {
         $641 = HEAP32[$635>>2]|0;
         $642 = ((($641)) + 36|0);
         $643 = HEAP32[$642>>2]|0;
         __THREW__ = 0;
         $644 = (invoke_ii($643|0,($635|0))|0);
         $645 = __THREW__; __THREW__ = 0;
         $646 = $645&1;
         if ($646) {
          label = 40;
          break L24;
         } else {
          $$0$i$i134 = $644;
         }
        } else {
         $647 = HEAP32[$637>>2]|0;
         $$0$i$i134 = $647;
        }
        $648 = HEAP32[$__dp>>2]|0;
        $649 = ($$0$i$i134|0)==($648|0);
        if (!($649)) {
         label = 222;
         break L24;
        }
        $652 = HEAP32[$__b>>2]|0;
        $653 = ((($652)) + 12|0);
        $654 = HEAP32[$653>>2]|0;
        $655 = ((($652)) + 16|0);
        $656 = HEAP32[$655>>2]|0;
        $657 = ($654|0)==($656|0);
        if ($657) {
         $658 = HEAP32[$652>>2]|0;
         $659 = ((($658)) + 40|0);
         $660 = HEAP32[$659>>2]|0;
         __THREW__ = 0;
         (invoke_ii($660|0,($652|0))|0);
         $661 = __THREW__; __THREW__ = 0;
         $662 = $661&1;
         if ($662) {
          label = 40;
          break L24;
         } else {
          $$pre$i3$i$i115 = $857;$665 = $604;$682 = $857;
         }
        } else {
         $663 = ((($654)) + 4|0);
         HEAP32[$653>>2] = $663;
         $$pre$i3$i$i115 = $857;$665 = $604;$682 = $857;
        }
        while(1) {
         $664 = ($665|0)>(0);
         if (!($664)) {
          break L223;
         }
         $666 = HEAP32[$__b>>2]|0;
         $667 = ($666|0)==(0|0);
         do {
          if ($667) {
           $696 = 1;
          } else {
           $668 = ((($666)) + 12|0);
           $669 = HEAP32[$668>>2]|0;
           $670 = ((($666)) + 16|0);
           $671 = HEAP32[$670>>2]|0;
           $672 = ($669|0)==($671|0);
           if ($672) {
            $673 = HEAP32[$666>>2]|0;
            $674 = ((($673)) + 36|0);
            $675 = HEAP32[$674>>2]|0;
            __THREW__ = 0;
            $676 = (invoke_ii($675|0,($666|0))|0);
            $677 = __THREW__; __THREW__ = 0;
            $678 = $677&1;
            if ($678) {
             label = 38;
             break L24;
            } else {
             $$0$i$i$i$i107 = $676;
            }
           } else {
            $679 = HEAP32[$669>>2]|0;
            $$0$i$i$i$i107 = $679;
           }
           $680 = ($$0$i$i$i$i107|0)==(-1);
           if ($680) {
            HEAP32[$__b>>2] = 0;
            $696 = 1;
            break;
           } else {
            $$pre$i$i$i109 = HEAP32[$__b>>2]|0;
            $phitmp$i$i$i110 = ($$pre$i$i$i109|0)==(0|0);
            $696 = $phitmp$i$i$i110;
            break;
           }
          }
         } while(0);
         $681 = ($682|0)==(0|0);
         do {
          if ($681) {
           $$pre$i3$i$i115309$ph = $$pre$i3$i$i115;
           label = 240;
          } else {
           $683 = ((($682)) + 12|0);
           $684 = HEAP32[$683>>2]|0;
           $685 = ((($682)) + 16|0);
           $686 = HEAP32[$685>>2]|0;
           $687 = ($684|0)==($686|0);
           if ($687) {
            $688 = HEAP32[$682>>2]|0;
            $689 = ((($688)) + 36|0);
            $690 = HEAP32[$689>>2]|0;
            __THREW__ = 0;
            $691 = (invoke_ii($690|0,($682|0))|0);
            $692 = __THREW__; __THREW__ = 0;
            $693 = $692&1;
            if ($693) {
             label = 38;
             break L24;
            } else {
             $$0$i$i1$i$i113 = $691;
            }
           } else {
            $694 = HEAP32[$684>>2]|0;
            $$0$i$i1$i$i113 = $694;
           }
           $695 = ($$0$i$i1$i$i113|0)==(-1);
           if ($695) {
            HEAP32[$__e>>2] = 0;
            $$pre$i3$i$i115309$ph = 0;
            label = 240;
            break;
           } else {
            $phitmp$i4$i$i116 = ($$pre$i3$i$i115|0)==(0|0);
            $tmp$i$i118 = $696 ^ $phitmp$i4$i$i116;
            if ($tmp$i$i118) {
             $$pre$i3$i$i115309333 = $$pre$i3$i$i115;$858 = $$pre$i3$i$i115;
             break;
            } else {
             label = 247;
             break L24;
            }
           }
          }
         } while(0);
         if ((label|0) == 240) {
          label = 0;
          if ($696) {
           label = 247;
           break L24;
          } else {
           $$pre$i3$i$i115309333 = $$pre$i3$i$i115309$ph;$858 = 0;
          }
         }
         $697 = HEAP32[$__b>>2]|0;
         $698 = ((($697)) + 12|0);
         $699 = HEAP32[$698>>2]|0;
         $700 = ((($697)) + 16|0);
         $701 = HEAP32[$700>>2]|0;
         $702 = ($699|0)==($701|0);
         if ($702) {
          $703 = HEAP32[$697>>2]|0;
          $704 = ((($703)) + 36|0);
          $705 = HEAP32[$704>>2]|0;
          __THREW__ = 0;
          $706 = (invoke_ii($705|0,($697|0))|0);
          $707 = __THREW__; __THREW__ = 0;
          $708 = $707&1;
          if ($708) {
           label = 38;
           break L24;
          } else {
           $$0$i$i101 = $706;
          }
         } else {
          $709 = HEAP32[$699>>2]|0;
          $$0$i$i101 = $709;
         }
         $710 = HEAP32[$__ct>>2]|0;
         $711 = ((($710)) + 12|0);
         $712 = HEAP32[$711>>2]|0;
         __THREW__ = 0;
         $713 = (invoke_iiii($712|0,($__ct|0),2048,($$0$i$i101|0))|0);
         $714 = __THREW__; __THREW__ = 0;
         $715 = $714&1;
         if ($715) {
          label = 38;
          break L24;
         }
         if (!($713)) {
          label = 247;
          break L24;
         }
         $718 = HEAP32[$__wn>>2]|0;
         $719 = HEAP32[$0>>2]|0;
         $720 = ($718|0)==($719|0);
         if ($720) {
          __THREW__ = 0;
          invoke_viii(4909,($__wb|0),($__wn|0),($0|0));
          $721 = __THREW__; __THREW__ = 0;
          $722 = $721&1;
          if ($722) {
           label = 38;
           break L24;
          }
         }
         $723 = HEAP32[$__b>>2]|0;
         $724 = ((($723)) + 12|0);
         $725 = HEAP32[$724>>2]|0;
         $726 = ((($723)) + 16|0);
         $727 = HEAP32[$726>>2]|0;
         $728 = ($725|0)==($727|0);
         if ($728) {
          $729 = HEAP32[$723>>2]|0;
          $730 = ((($729)) + 36|0);
          $731 = HEAP32[$730>>2]|0;
          __THREW__ = 0;
          $732 = (invoke_ii($731|0,($723|0))|0);
          $733 = __THREW__; __THREW__ = 0;
          $734 = $733&1;
          if ($734) {
           label = 38;
           break L24;
          } else {
           $$0$i$i93 = $732;
          }
         } else {
          $735 = HEAP32[$725>>2]|0;
          $$0$i$i93 = $735;
         }
         $736 = HEAP32[$__wn>>2]|0;
         $737 = ((($736)) + 4|0);
         HEAP32[$__wn>>2] = $737;
         HEAP32[$736>>2] = $$0$i$i93;
         $738 = (($665) + -1)|0;
         HEAP32[$__fd>>2] = $738;
         $739 = HEAP32[$__b>>2]|0;
         $740 = ((($739)) + 12|0);
         $741 = HEAP32[$740>>2]|0;
         $742 = ((($739)) + 16|0);
         $743 = HEAP32[$742>>2]|0;
         $744 = ($741|0)==($743|0);
         if ($744) {
          $745 = HEAP32[$739>>2]|0;
          $746 = ((($745)) + 40|0);
          $747 = HEAP32[$746>>2]|0;
          __THREW__ = 0;
          (invoke_ii($747|0,($739|0))|0);
          $748 = __THREW__; __THREW__ = 0;
          $749 = $748&1;
          if ($749) {
           label = 38;
           break L24;
          } else {
           $$pre$i3$i$i115 = $$pre$i3$i$i115309333;$665 = $738;$682 = $858;
           continue;
          }
         } else {
          $750 = ((($741)) + 4|0);
          HEAP32[$740>>2] = $750;
          $$pre$i3$i$i115 = $$pre$i3$i$i115309333;$665 = $738;$682 = $858;
          continue;
         }
        }
       }
      } while(0);
      $751 = HEAP32[$__wn>>2]|0;
      $752 = HEAP32[$__wb>>2]|0;
      $753 = ($751|0)==($752|0);
      if ($753) {
       label = 257;
       break L24;
      } else {
       $850 = $855;$__trailing_sign$1 = $__trailing_sign$0;
      }
      break;
     }
     default: {
      $850 = $847;$__trailing_sign$1 = $__trailing_sign$0;
     }
     }
    } while(0);
    L297: do {
     if ((label|0) == 60) {
      while(1) {
       label = 0;
       $125 = HEAP32[$__b>>2]|0;
       $126 = ($125|0)==(0|0);
       do {
        if ($126) {
         $155 = 1;
        } else {
         $127 = ((($125)) + 12|0);
         $128 = HEAP32[$127>>2]|0;
         $129 = ((($125)) + 16|0);
         $130 = HEAP32[$129>>2]|0;
         $131 = ($128|0)==($130|0);
         if ($131) {
          $132 = HEAP32[$125>>2]|0;
          $133 = ((($132)) + 36|0);
          $134 = HEAP32[$133>>2]|0;
          __THREW__ = 0;
          $135 = (invoke_ii($134|0,($125|0))|0);
          $136 = __THREW__; __THREW__ = 0;
          $137 = $136&1;
          if ($137) {
           label = 35;
           break L24;
          } else {
           $$0$i$i$i$i66 = $135;
          }
         } else {
          $138 = HEAP32[$128>>2]|0;
          $$0$i$i$i$i66 = $138;
         }
         $139 = ($$0$i$i$i$i66|0)==(-1);
         if ($139) {
          HEAP32[$__b>>2] = 0;
          $155 = 1;
          break;
         } else {
          $$pre$i$i$i68 = HEAP32[$__b>>2]|0;
          $phitmp$i$i$i69 = ($$pre$i$i$i68|0)==(0|0);
          $155 = $phitmp$i$i$i69;
          break;
         }
        }
       } while(0);
       $140 = ($141|0)==(0|0);
       do {
        if ($140) {
         $$pre$i3$i$i74327 = $$pre$i3$i$i74;
         label = 73;
        } else {
         $142 = ((($141)) + 12|0);
         $143 = HEAP32[$142>>2]|0;
         $144 = ((($141)) + 16|0);
         $145 = HEAP32[$144>>2]|0;
         $146 = ($143|0)==($145|0);
         if ($146) {
          $147 = HEAP32[$141>>2]|0;
          $148 = ((($147)) + 36|0);
          $149 = HEAP32[$148>>2]|0;
          __THREW__ = 0;
          $150 = (invoke_ii($149|0,($141|0))|0);
          $151 = __THREW__; __THREW__ = 0;
          $152 = $151&1;
          if ($152) {
           label = 35;
           break L24;
          } else {
           $$0$i$i1$i$i72 = $150;
          }
         } else {
          $153 = HEAP32[$143>>2]|0;
          $$0$i$i1$i$i72 = $153;
         }
         $154 = ($$0$i$i1$i$i72|0)==(-1);
         if ($154) {
          HEAP32[$__e>>2] = 0;
          $$pre$i3$i$i74327 = 0;
          label = 73;
          break;
         } else {
          $phitmp$i4$i$i75 = ($$pre$i3$i$i74|0)==(0|0);
          $tmp$i$i77 = $155 ^ $phitmp$i4$i$i75;
          if ($tmp$i$i77) {
           $$pre$i3$i$i74329 = $$pre$i3$i$i74;$851 = $$pre$i3$i$i74;
           break;
          } else {
           $850 = $847;$__trailing_sign$1 = $__trailing_sign$0;
           break L297;
          }
         }
        }
       } while(0);
       if ((label|0) == 73) {
        label = 0;
        if ($155) {
         $850 = $847;$__trailing_sign$1 = $__trailing_sign$0;
         break L297;
        } else {
         $$pre$i3$i$i74329 = $$pre$i3$i$i74327;$851 = 0;
        }
       }
       $156 = HEAP32[$__b>>2]|0;
       $157 = ((($156)) + 12|0);
       $158 = HEAP32[$157>>2]|0;
       $159 = ((($156)) + 16|0);
       $160 = HEAP32[$159>>2]|0;
       $161 = ($158|0)==($160|0);
       if ($161) {
        $162 = HEAP32[$156>>2]|0;
        $163 = ((($162)) + 36|0);
        $164 = HEAP32[$163>>2]|0;
        __THREW__ = 0;
        $165 = (invoke_ii($164|0,($156|0))|0);
        $166 = __THREW__; __THREW__ = 0;
        $167 = $166&1;
        if ($167) {
         label = 35;
         break L24;
        } else {
         $$0$i$i81 = $165;
        }
       } else {
        $168 = HEAP32[$158>>2]|0;
        $$0$i$i81 = $168;
       }
       $169 = HEAP32[$__ct>>2]|0;
       $170 = ((($169)) + 12|0);
       $171 = HEAP32[$170>>2]|0;
       __THREW__ = 0;
       $172 = (invoke_iiii($171|0,($__ct|0),8192,($$0$i$i81|0))|0);
       $173 = __THREW__; __THREW__ = 0;
       $174 = $173&1;
       if ($174) {
        label = 35;
        break L24;
       }
       if (!($172)) {
        $850 = $847;$__trailing_sign$1 = $__trailing_sign$0;
        break L297;
       }
       $175 = HEAP32[$__b>>2]|0;
       $176 = ((($175)) + 12|0);
       $177 = HEAP32[$176>>2]|0;
       $178 = ((($175)) + 16|0);
       $179 = HEAP32[$178>>2]|0;
       $180 = ($177|0)==($179|0);
       if ($180) {
        $181 = HEAP32[$175>>2]|0;
        $182 = ((($181)) + 40|0);
        $183 = HEAP32[$182>>2]|0;
        __THREW__ = 0;
        $184 = (invoke_ii($183|0,($175|0))|0);
        $185 = __THREW__; __THREW__ = 0;
        $186 = $185&1;
        if ($186) {
         label = 35;
         break L24;
        } else {
         $$0$i$i88 = $184;
        }
       } else {
        $187 = ((($177)) + 4|0);
        HEAP32[$176>>2] = $187;
        $188 = HEAP32[$177>>2]|0;
        $$0$i$i88 = $188;
       }
       __THREW__ = 0;
       invoke_vii(4908,($__spaces|0),($$0$i$i88|0));
       $189 = __THREW__; __THREW__ = 0;
       $190 = $189&1;
       if ($190) {
        label = 35;
        break L24;
       } else {
        $$pre$i3$i$i74 = $$pre$i3$i$i74329;$141 = $851;
        label = 60;
       }
      }
     }
    } while(0);
    $756 = (($__p$0) + 1)|0;
    $847 = $850;$__p$0 = $756;$__trailing_sign$0 = $__trailing_sign$1;
   }
   L333:    switch (label|0) {
    case 35: {
     $62 = ___cxa_find_matching_catch_2()|0;
     $63 = tempRet0;
     $lpad$phi$index = $62;$lpad$phi$index8 = $63;
     break L21;
     break;
    }
    case 36: {
     $64 = ___cxa_find_matching_catch_2()|0;
     $65 = tempRet0;
     $lpad$phi$index = $64;$lpad$phi$index8 = $65;
     break L21;
     break;
    }
    case 37: {
     $66 = ___cxa_find_matching_catch_2()|0;
     $67 = tempRet0;
     $lpad$phi$index = $66;$lpad$phi$index8 = $67;
     break L21;
     break;
    }
    case 38: {
     $68 = ___cxa_find_matching_catch_2()|0;
     $69 = tempRet0;
     $lpad$phi$index = $68;$lpad$phi$index8 = $69;
     break L21;
     break;
    }
    case 39: {
     $70 = ___cxa_find_matching_catch_2()|0;
     $71 = tempRet0;
     $lpad$phi$index = $70;$lpad$phi$index8 = $71;
     break L21;
     break;
    }
    case 40: {
     $72 = ___cxa_find_matching_catch_2()|0;
     $73 = tempRet0;
     $lpad$phi$index = $72;$lpad$phi$index8 = $73;
     break L21;
     break;
    }
    case 58: {
     $122 = HEAP32[$__err>>2]|0;
     $123 = $122 | 4;
     HEAP32[$__err>>2] = $123;
     $$10 = 0;
     break;
    }
    case 123: {
     $356 = HEAP32[$__err>>2]|0;
     $357 = $356 | 4;
     HEAP32[$__err>>2] = $357;
     $$10 = 0;
     break;
    }
    case 164: {
     $506 = HEAP32[$__err>>2]|0;
     $507 = $506 | 4;
     HEAP32[$__err>>2] = $507;
     $$10 = 0;
     break;
    }
    case 222: {
     $650 = HEAP32[$__err>>2]|0;
     $651 = $650 | 4;
     HEAP32[$__err>>2] = $651;
     $$10 = 0;
     break;
    }
    case 247: {
     $716 = HEAP32[$__err>>2]|0;
     $717 = $716 | 4;
     HEAP32[$__err>>2] = $717;
     $$10 = 0;
     break;
    }
    case 257: {
     $754 = HEAP32[$__err>>2]|0;
     $755 = $754 | 4;
     HEAP32[$__err>>2] = $755;
     $$10 = 0;
     break;
    }
    case 259: {
     $757 = ($__trailing_sign$0$lcssa|0)==(0|0);
     L335: do {
      if (!($757)) {
       $758 = ((($__trailing_sign$0$lcssa)) + 4|0);
       $759 = ((($__trailing_sign$0$lcssa)) + 8|0);
       $__i$0 = 1;
       L337: while(1) {
        $760 = HEAP8[$__trailing_sign$0$lcssa>>0]|0;
        $761 = $760 & 1;
        $762 = ($761<<24>>24)==(0);
        if ($762) {
         $764 = $760&255;
         $765 = $764 >>> 1;
         $767 = $765;
        } else {
         $763 = HEAP32[$758>>2]|0;
         $767 = $763;
        }
        $766 = ($__i$0>>>0)<($767>>>0);
        if (!($766)) {
         break L335;
        }
        $768 = HEAP32[$__b>>2]|0;
        $769 = ($768|0)==(0|0);
        do {
         if ($769) {
          $859 = 1;
         } else {
          $770 = ((($768)) + 12|0);
          $771 = HEAP32[$770>>2]|0;
          $772 = ((($768)) + 16|0);
          $773 = HEAP32[$772>>2]|0;
          $774 = ($771|0)==($773|0);
          if ($774) {
           $775 = HEAP32[$768>>2]|0;
           $776 = ((($775)) + 36|0);
           $777 = HEAP32[$776>>2]|0;
           __THREW__ = 0;
           $778 = (invoke_ii($777|0,($768|0))|0);
           $779 = __THREW__; __THREW__ = 0;
           $780 = $779&1;
           if ($780) {
            label = 34;
            break L337;
           } else {
            $$0$i$i$i$i = $778;
           }
          } else {
           $781 = HEAP32[$771>>2]|0;
           $$0$i$i$i$i = $781;
          }
          $782 = ($$0$i$i$i$i|0)==(-1);
          if ($782) {
           HEAP32[$__b>>2] = 0;
           $859 = 1;
           break;
          } else {
           $$pre$i$i$i = HEAP32[$__b>>2]|0;
           $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
           $859 = $phitmp$i$i$i;
           break;
          }
         }
        } while(0);
        $783 = HEAP32[$__e>>2]|0;
        $784 = ($783|0)==(0|0);
        do {
         if ($784) {
          label = 278;
         } else {
          $785 = ((($783)) + 12|0);
          $786 = HEAP32[$785>>2]|0;
          $787 = ((($783)) + 16|0);
          $788 = HEAP32[$787>>2]|0;
          $789 = ($786|0)==($788|0);
          if ($789) {
           $790 = HEAP32[$783>>2]|0;
           $791 = ((($790)) + 36|0);
           $792 = HEAP32[$791>>2]|0;
           __THREW__ = 0;
           $793 = (invoke_ii($792|0,($783|0))|0);
           $794 = __THREW__; __THREW__ = 0;
           $795 = $794&1;
           if ($795) {
            label = 34;
            break L337;
           } else {
            $$0$i$i1$i$i = $793;
           }
          } else {
           $796 = HEAP32[$786>>2]|0;
           $$0$i$i1$i$i = $796;
          }
          $797 = ($$0$i$i1$i$i|0)==(-1);
          if ($797) {
           HEAP32[$__e>>2] = 0;
           label = 278;
           break;
          } else {
           if ($859) {
            break;
           } else {
            label = 286;
            break L337;
           }
          }
         }
        } while(0);
        if ((label|0) == 278) {
         label = 0;
         if ($859) {
          label = 286;
          break;
         }
        }
        $798 = HEAP32[$__b>>2]|0;
        $799 = ((($798)) + 12|0);
        $800 = HEAP32[$799>>2]|0;
        $801 = ((($798)) + 16|0);
        $802 = HEAP32[$801>>2]|0;
        $803 = ($800|0)==($802|0);
        if ($803) {
         $804 = HEAP32[$798>>2]|0;
         $805 = ((($804)) + 36|0);
         $806 = HEAP32[$805>>2]|0;
         __THREW__ = 0;
         $807 = (invoke_ii($806|0,($798|0))|0);
         $808 = __THREW__; __THREW__ = 0;
         $809 = $808&1;
         if ($809) {
          label = 34;
          break;
         } else {
          $$0$i$i = $807;
         }
        } else {
         $810 = HEAP32[$800>>2]|0;
         $$0$i$i = $810;
        }
        $811 = HEAP8[$__trailing_sign$0$lcssa>>0]|0;
        $812 = $811 & 1;
        $813 = ($812<<24>>24)==(0);
        if ($813) {
         $816 = $758;
        } else {
         $814 = HEAP32[$759>>2]|0;
         $816 = $814;
        }
        $815 = (($816) + ($__i$0<<2)|0);
        $817 = HEAP32[$815>>2]|0;
        $818 = ($$0$i$i|0)==($817|0);
        if (!($818)) {
         label = 286;
         break;
        }
        $821 = (($__i$0) + 1)|0;
        $822 = HEAP32[$__b>>2]|0;
        $823 = ((($822)) + 12|0);
        $824 = HEAP32[$823>>2]|0;
        $825 = ((($822)) + 16|0);
        $826 = HEAP32[$825>>2]|0;
        $827 = ($824|0)==($826|0);
        if ($827) {
         $828 = HEAP32[$822>>2]|0;
         $829 = ((($828)) + 40|0);
         $830 = HEAP32[$829>>2]|0;
         __THREW__ = 0;
         (invoke_ii($830|0,($822|0))|0);
         $831 = __THREW__; __THREW__ = 0;
         $832 = $831&1;
         if ($832) {
          label = 34;
          break;
         } else {
          $__i$0 = $821;
          continue;
         }
        } else {
         $833 = ((($824)) + 4|0);
         HEAP32[$823>>2] = $833;
         $__i$0 = $821;
         continue;
        }
       }
       if ((label|0) == 34) {
        $60 = ___cxa_find_matching_catch_2()|0;
        $61 = tempRet0;
        $lpad$phi$index = $60;$lpad$phi$index8 = $61;
        break L21;
       }
       else if ((label|0) == 286) {
        $819 = HEAP32[$__err>>2]|0;
        $820 = $819 | 4;
        HEAP32[$__err>>2] = $820;
        $$10 = 0;
        break L333;
       }
      }
     } while(0);
     $834 = HEAP32[$__gb>>2]|0;
     $835 = ($834|0)==($$lcssa461|0);
     if ($835) {
      $$10 = 1;
     } else {
      HEAP32[$__et>>2] = 0;
      __ZNSt3__116__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($__grp,$834,$$lcssa461,$__et);
      $836 = HEAP32[$__et>>2]|0;
      $837 = ($836|0)==(0);
      if ($837) {
       $$10 = 1;
       break;
      } else {
       $838 = HEAP32[$__err>>2]|0;
       $839 = $838 | 4;
       HEAP32[$__err>>2] = $839;
       $$10 = 0;
       break;
      }
     }
     break;
    }
   }
   __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__spaces);
   __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__nsn);
   __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__psn);
   __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__sym);
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grp);
   $840 = HEAP32[$__gb>>2]|0;
   HEAP32[$__gb>>2] = 0;
   $841 = ($840|0)==(0|0);
   if (!($841)) {
    $842 = HEAP32[$2>>2]|0;
    __THREW__ = 0;
    invoke_vi($842|0,($840|0));
    $843 = __THREW__; __THREW__ = 0;
    $844 = $843&1;
    if ($844) {
     $845 = ___cxa_find_matching_catch_3(0|0)|0;
     $846 = tempRet0;
     ___clang_call_terminate($845);
     // unreachable;
    }
   }
   STACKTOP = sp;return ($$10|0);
  }
 } while(0);
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__spaces);
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__nsn);
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__psn);
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__sym);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grp);
 $76 = HEAP32[$__gb>>2]|0;
 HEAP32[$__gb>>2] = 0;
 $77 = ($76|0)==(0|0);
 if (!($77)) {
  $78 = HEAP32[$2>>2]|0;
  __THREW__ = 0;
  invoke_vi($78|0,($76|0));
  $79 = __THREW__; __THREW__ = 0;
  $80 = $79&1;
  if ($80) {
   $81 = ___cxa_find_matching_catch_3(0|0)|0;
   $82 = tempRet0;
   ___clang_call_terminate($81);
   // unreachable;
  }
 }
 ___resumeException($lpad$phi$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__111__money_getIwE13__gather_infoEbRKNS_6localeERNS_10money_base7patternERwS8_RNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERNS9_IwNSA_IwEENSC_IwEEEESJ_SJ_Ri($__intl,$__loc,$__pat,$__dp,$__ts,$__grp,$__sym,$__psn,$__nsn,$__fd) {
 $__intl = $__intl|0;
 $__loc = $__loc|0;
 $__pat = $__pat|0;
 $__dp = $__dp|0;
 $__ts = $__ts|0;
 $__grp = $__grp|0;
 $__sym = $__sym|0;
 $__psn = $__psn|0;
 $__nsn = $__nsn|0;
 $__fd = $__fd|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__i$0$i$i$i = 0, $__i$0$i$i$i11 = 0, $__i$0$i$i$i16 = 0, $__i$0$i$i$i21 = 0, $__i$0$i$i$i26 = 0, $__i$0$i$i$i31 = 0, $__i$0$i$i$i4 = 0, $__i$0$i$i$i7 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i12 = 0, $exitcond$i$i$i17 = 0, $exitcond$i$i$i22 = 0, $exitcond$i$i$i27 = 0, $exitcond$i$i$i32 = 0;
 var $exitcond$i$i$i5 = 0, $exitcond$i$i$i8 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 100|0;
 $1 = sp + 88|0;
 $2 = sp + 76|0;
 $3 = sp + 64|0;
 $4 = sp + 52|0;
 $5 = sp + 48|0;
 $6 = sp + 36|0;
 $7 = sp + 24|0;
 $8 = sp + 12|0;
 $9 = sp;
 if ($__intl) {
  $10 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1209312)|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ((($11)) + 44|0);
  $13 = HEAP32[$12>>2]|0;
  FUNCTION_TABLE_vii[$13 & 8191]($0,$10);
  $14 = HEAP32[$0>>2]|0;
  HEAP8[$__pat>>0]=$14&255;HEAP8[$__pat+1>>0]=($14>>8)&255;HEAP8[$__pat+2>>0]=($14>>16)&255;HEAP8[$__pat+3>>0]=$14>>24;
  $15 = HEAP32[$10>>2]|0;
  $16 = ((($15)) + 32|0);
  $17 = HEAP32[$16>>2]|0;
  FUNCTION_TABLE_vii[$17 & 8191]($1,$10);
  $18 = HEAP8[$__nsn>>0]|0;
  $19 = $18 & 1;
  $20 = ($19<<24>>24)==(0);
  if ($20) {
   HEAP8[$__nsn>>0] = 0;
  } else {
   $21 = ((($__nsn)) + 8|0);
   $22 = HEAP32[$21>>2]|0;
   HEAP32[$22>>2] = 0;
  }
  $23 = ((($__nsn)) + 4|0);
  HEAP32[$23>>2] = 0;
  __THREW__ = 0;
  invoke_vii(4910,($__nsn|0),0);
  $24 = __THREW__; __THREW__ = 0;
  $25 = $24&1;
  if ($25) {
   $26 = ___cxa_find_matching_catch_3(0|0)|0;
   $27 = tempRet0;
   ___clang_call_terminate($26);
   // unreachable;
  }
  ;HEAP32[$__nsn>>2]=HEAP32[$1>>2]|0;HEAP32[$__nsn+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$__nsn+8>>2]=HEAP32[$1+8>>2]|0;
  $__i$0$i$i$i = 0;
  while(1) {
   $exitcond$i$i$i = ($__i$0$i$i$i|0)==(3);
   if ($exitcond$i$i$i) {
    break;
   }
   $28 = (($1) + ($__i$0$i$i$i<<2)|0);
   HEAP32[$28>>2] = 0;
   $29 = (($__i$0$i$i$i) + 1)|0;
   $__i$0$i$i$i = $29;
  }
  __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($1);
  $30 = HEAP32[$10>>2]|0;
  $31 = ((($30)) + 28|0);
  $32 = HEAP32[$31>>2]|0;
  FUNCTION_TABLE_vii[$32 & 8191]($2,$10);
  $33 = HEAP8[$__psn>>0]|0;
  $34 = $33 & 1;
  $35 = ($34<<24>>24)==(0);
  if ($35) {
   HEAP8[$__psn>>0] = 0;
  } else {
   $36 = ((($__psn)) + 8|0);
   $37 = HEAP32[$36>>2]|0;
   HEAP32[$37>>2] = 0;
  }
  $38 = ((($__psn)) + 4|0);
  HEAP32[$38>>2] = 0;
  __THREW__ = 0;
  invoke_vii(4910,($__psn|0),0);
  $39 = __THREW__; __THREW__ = 0;
  $40 = $39&1;
  if ($40) {
   $41 = ___cxa_find_matching_catch_3(0|0)|0;
   $42 = tempRet0;
   ___clang_call_terminate($41);
   // unreachable;
  }
  ;HEAP32[$__psn>>2]=HEAP32[$2>>2]|0;HEAP32[$__psn+4>>2]=HEAP32[$2+4>>2]|0;HEAP32[$__psn+8>>2]=HEAP32[$2+8>>2]|0;
  $__i$0$i$i$i4 = 0;
  while(1) {
   $exitcond$i$i$i5 = ($__i$0$i$i$i4|0)==(3);
   if ($exitcond$i$i$i5) {
    break;
   }
   $43 = (($2) + ($__i$0$i$i$i4<<2)|0);
   HEAP32[$43>>2] = 0;
   $44 = (($__i$0$i$i$i4) + 1)|0;
   $__i$0$i$i$i4 = $44;
  }
  __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($2);
  $45 = HEAP32[$10>>2]|0;
  $46 = ((($45)) + 12|0);
  $47 = HEAP32[$46>>2]|0;
  $48 = (FUNCTION_TABLE_ii[$47 & 8191]($10)|0);
  HEAP32[$__dp>>2] = $48;
  $49 = HEAP32[$10>>2]|0;
  $50 = ((($49)) + 16|0);
  $51 = HEAP32[$50>>2]|0;
  $52 = (FUNCTION_TABLE_ii[$51 & 8191]($10)|0);
  HEAP32[$__ts>>2] = $52;
  $53 = HEAP32[$10>>2]|0;
  $54 = ((($53)) + 20|0);
  $55 = HEAP32[$54>>2]|0;
  FUNCTION_TABLE_vii[$55 & 8191]($3,$10);
  $56 = HEAP8[$__grp>>0]|0;
  $57 = $56 & 1;
  $58 = ($57<<24>>24)==(0);
  if ($58) {
   $62 = ((($__grp)) + 1|0);
   HEAP8[$62>>0] = 0;
   HEAP8[$__grp>>0] = 0;
  } else {
   $59 = ((($__grp)) + 8|0);
   $60 = HEAP32[$59>>2]|0;
   HEAP8[$60>>0] = 0;
   $61 = ((($__grp)) + 4|0);
   HEAP32[$61>>2] = 0;
  }
  __THREW__ = 0;
  invoke_vii(3368,($__grp|0),0);
  $63 = __THREW__; __THREW__ = 0;
  $64 = $63&1;
  if ($64) {
   $65 = ___cxa_find_matching_catch_3(0|0)|0;
   $66 = tempRet0;
   ___clang_call_terminate($65);
   // unreachable;
  }
  ;HEAP32[$__grp>>2]=HEAP32[$3>>2]|0;HEAP32[$__grp+4>>2]=HEAP32[$3+4>>2]|0;HEAP32[$__grp+8>>2]=HEAP32[$3+8>>2]|0;
  $__i$0$i$i$i7 = 0;
  while(1) {
   $exitcond$i$i$i8 = ($__i$0$i$i$i7|0)==(3);
   if ($exitcond$i$i$i8) {
    break;
   }
   $67 = (($3) + ($__i$0$i$i$i7<<2)|0);
   HEAP32[$67>>2] = 0;
   $68 = (($__i$0$i$i$i7) + 1)|0;
   $__i$0$i$i$i7 = $68;
  }
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3);
  $69 = HEAP32[$10>>2]|0;
  $70 = ((($69)) + 24|0);
  $71 = HEAP32[$70>>2]|0;
  FUNCTION_TABLE_vii[$71 & 8191]($4,$10);
  $72 = HEAP8[$__sym>>0]|0;
  $73 = $72 & 1;
  $74 = ($73<<24>>24)==(0);
  if ($74) {
   HEAP8[$__sym>>0] = 0;
  } else {
   $75 = ((($__sym)) + 8|0);
   $76 = HEAP32[$75>>2]|0;
   HEAP32[$76>>2] = 0;
  }
  $77 = ((($__sym)) + 4|0);
  HEAP32[$77>>2] = 0;
  __THREW__ = 0;
  invoke_vii(4910,($__sym|0),0);
  $78 = __THREW__; __THREW__ = 0;
  $79 = $78&1;
  if ($79) {
   $80 = ___cxa_find_matching_catch_3(0|0)|0;
   $81 = tempRet0;
   ___clang_call_terminate($80);
   // unreachable;
  }
  ;HEAP32[$__sym>>2]=HEAP32[$4>>2]|0;HEAP32[$__sym+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$__sym+8>>2]=HEAP32[$4+8>>2]|0;
  $__i$0$i$i$i11 = 0;
  while(1) {
   $exitcond$i$i$i12 = ($__i$0$i$i$i11|0)==(3);
   if ($exitcond$i$i$i12) {
    break;
   }
   $82 = (($4) + ($__i$0$i$i$i11<<2)|0);
   HEAP32[$82>>2] = 0;
   $83 = (($__i$0$i$i$i11) + 1)|0;
   $__i$0$i$i$i11 = $83;
  }
  __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($4);
  $84 = HEAP32[$10>>2]|0;
  $85 = ((($84)) + 36|0);
  $86 = HEAP32[$85>>2]|0;
  $87 = (FUNCTION_TABLE_ii[$86 & 8191]($10)|0);
  $storemerge = $87;
 } else {
  $88 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1209304)|0);
  $89 = HEAP32[$88>>2]|0;
  $90 = ((($89)) + 44|0);
  $91 = HEAP32[$90>>2]|0;
  FUNCTION_TABLE_vii[$91 & 8191]($5,$88);
  $92 = HEAP32[$5>>2]|0;
  HEAP8[$__pat>>0]=$92&255;HEAP8[$__pat+1>>0]=($92>>8)&255;HEAP8[$__pat+2>>0]=($92>>16)&255;HEAP8[$__pat+3>>0]=$92>>24;
  $93 = HEAP32[$88>>2]|0;
  $94 = ((($93)) + 32|0);
  $95 = HEAP32[$94>>2]|0;
  FUNCTION_TABLE_vii[$95 & 8191]($6,$88);
  $96 = HEAP8[$__nsn>>0]|0;
  $97 = $96 & 1;
  $98 = ($97<<24>>24)==(0);
  if ($98) {
   HEAP8[$__nsn>>0] = 0;
  } else {
   $99 = ((($__nsn)) + 8|0);
   $100 = HEAP32[$99>>2]|0;
   HEAP32[$100>>2] = 0;
  }
  $101 = ((($__nsn)) + 4|0);
  HEAP32[$101>>2] = 0;
  __THREW__ = 0;
  invoke_vii(4910,($__nsn|0),0);
  $102 = __THREW__; __THREW__ = 0;
  $103 = $102&1;
  if ($103) {
   $104 = ___cxa_find_matching_catch_3(0|0)|0;
   $105 = tempRet0;
   ___clang_call_terminate($104);
   // unreachable;
  }
  ;HEAP32[$__nsn>>2]=HEAP32[$6>>2]|0;HEAP32[$__nsn+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$__nsn+8>>2]=HEAP32[$6+8>>2]|0;
  $__i$0$i$i$i16 = 0;
  while(1) {
   $exitcond$i$i$i17 = ($__i$0$i$i$i16|0)==(3);
   if ($exitcond$i$i$i17) {
    break;
   }
   $106 = (($6) + ($__i$0$i$i$i16<<2)|0);
   HEAP32[$106>>2] = 0;
   $107 = (($__i$0$i$i$i16) + 1)|0;
   $__i$0$i$i$i16 = $107;
  }
  __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($6);
  $108 = HEAP32[$88>>2]|0;
  $109 = ((($108)) + 28|0);
  $110 = HEAP32[$109>>2]|0;
  FUNCTION_TABLE_vii[$110 & 8191]($7,$88);
  $111 = HEAP8[$__psn>>0]|0;
  $112 = $111 & 1;
  $113 = ($112<<24>>24)==(0);
  if ($113) {
   HEAP8[$__psn>>0] = 0;
  } else {
   $114 = ((($__psn)) + 8|0);
   $115 = HEAP32[$114>>2]|0;
   HEAP32[$115>>2] = 0;
  }
  $116 = ((($__psn)) + 4|0);
  HEAP32[$116>>2] = 0;
  __THREW__ = 0;
  invoke_vii(4910,($__psn|0),0);
  $117 = __THREW__; __THREW__ = 0;
  $118 = $117&1;
  if ($118) {
   $119 = ___cxa_find_matching_catch_3(0|0)|0;
   $120 = tempRet0;
   ___clang_call_terminate($119);
   // unreachable;
  }
  ;HEAP32[$__psn>>2]=HEAP32[$7>>2]|0;HEAP32[$__psn+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$__psn+8>>2]=HEAP32[$7+8>>2]|0;
  $__i$0$i$i$i21 = 0;
  while(1) {
   $exitcond$i$i$i22 = ($__i$0$i$i$i21|0)==(3);
   if ($exitcond$i$i$i22) {
    break;
   }
   $121 = (($7) + ($__i$0$i$i$i21<<2)|0);
   HEAP32[$121>>2] = 0;
   $122 = (($__i$0$i$i$i21) + 1)|0;
   $__i$0$i$i$i21 = $122;
  }
  __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($7);
  $123 = HEAP32[$88>>2]|0;
  $124 = ((($123)) + 12|0);
  $125 = HEAP32[$124>>2]|0;
  $126 = (FUNCTION_TABLE_ii[$125 & 8191]($88)|0);
  HEAP32[$__dp>>2] = $126;
  $127 = HEAP32[$88>>2]|0;
  $128 = ((($127)) + 16|0);
  $129 = HEAP32[$128>>2]|0;
  $130 = (FUNCTION_TABLE_ii[$129 & 8191]($88)|0);
  HEAP32[$__ts>>2] = $130;
  $131 = HEAP32[$88>>2]|0;
  $132 = ((($131)) + 20|0);
  $133 = HEAP32[$132>>2]|0;
  FUNCTION_TABLE_vii[$133 & 8191]($8,$88);
  $134 = HEAP8[$__grp>>0]|0;
  $135 = $134 & 1;
  $136 = ($135<<24>>24)==(0);
  if ($136) {
   $140 = ((($__grp)) + 1|0);
   HEAP8[$140>>0] = 0;
   HEAP8[$__grp>>0] = 0;
  } else {
   $137 = ((($__grp)) + 8|0);
   $138 = HEAP32[$137>>2]|0;
   HEAP8[$138>>0] = 0;
   $139 = ((($__grp)) + 4|0);
   HEAP32[$139>>2] = 0;
  }
  __THREW__ = 0;
  invoke_vii(3368,($__grp|0),0);
  $141 = __THREW__; __THREW__ = 0;
  $142 = $141&1;
  if ($142) {
   $143 = ___cxa_find_matching_catch_3(0|0)|0;
   $144 = tempRet0;
   ___clang_call_terminate($143);
   // unreachable;
  }
  ;HEAP32[$__grp>>2]=HEAP32[$8>>2]|0;HEAP32[$__grp+4>>2]=HEAP32[$8+4>>2]|0;HEAP32[$__grp+8>>2]=HEAP32[$8+8>>2]|0;
  $__i$0$i$i$i26 = 0;
  while(1) {
   $exitcond$i$i$i27 = ($__i$0$i$i$i26|0)==(3);
   if ($exitcond$i$i$i27) {
    break;
   }
   $145 = (($8) + ($__i$0$i$i$i26<<2)|0);
   HEAP32[$145>>2] = 0;
   $146 = (($__i$0$i$i$i26) + 1)|0;
   $__i$0$i$i$i26 = $146;
  }
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
  $147 = HEAP32[$88>>2]|0;
  $148 = ((($147)) + 24|0);
  $149 = HEAP32[$148>>2]|0;
  FUNCTION_TABLE_vii[$149 & 8191]($9,$88);
  $150 = HEAP8[$__sym>>0]|0;
  $151 = $150 & 1;
  $152 = ($151<<24>>24)==(0);
  if ($152) {
   HEAP8[$__sym>>0] = 0;
  } else {
   $153 = ((($__sym)) + 8|0);
   $154 = HEAP32[$153>>2]|0;
   HEAP32[$154>>2] = 0;
  }
  $155 = ((($__sym)) + 4|0);
  HEAP32[$155>>2] = 0;
  __THREW__ = 0;
  invoke_vii(4910,($__sym|0),0);
  $156 = __THREW__; __THREW__ = 0;
  $157 = $156&1;
  if ($157) {
   $158 = ___cxa_find_matching_catch_3(0|0)|0;
   $159 = tempRet0;
   ___clang_call_terminate($158);
   // unreachable;
  }
  ;HEAP32[$__sym>>2]=HEAP32[$9>>2]|0;HEAP32[$__sym+4>>2]=HEAP32[$9+4>>2]|0;HEAP32[$__sym+8>>2]=HEAP32[$9+8>>2]|0;
  $__i$0$i$i$i31 = 0;
  while(1) {
   $exitcond$i$i$i32 = ($__i$0$i$i$i31|0)==(3);
   if ($exitcond$i$i$i32) {
    break;
   }
   $160 = (($9) + ($__i$0$i$i$i31<<2)|0);
   HEAP32[$160>>2] = 0;
   $161 = (($__i$0$i$i$i31) + 1)|0;
   $__i$0$i$i$i31 = $161;
  }
  __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($9);
  $162 = HEAP32[$88>>2]|0;
  $163 = ((($162)) + 36|0);
  $164 = HEAP32[$163>>2]|0;
  $165 = (FUNCTION_TABLE_ii[$164 & 8191]($88)|0);
  $storemerge = $165;
 }
 HEAP32[$__fd>>2] = $storemerge;
 STACKTOP = sp;return;
}
function __ZNSt3__119__double_or_nothingIwEEvRNS_10unique_ptrIT_PFvPvEEERPS2_S9_($__b,$__n,$__e) {
 $__b = $__b|0;
 $__n = $__n|0;
 $__e = $__e|0;
 var $$ = 0, $$pr = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($__b)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)!=(4899|0);
 $3 = HEAP32[$__e>>2]|0;
 $4 = HEAP32[$__b>>2]|0;
 $5 = (($3) - ($4))|0;
 $6 = ($5>>>0)<(2147483647);
 $7 = $5 << 1;
 $$ = $6 ? $7 : -1;
 $8 = HEAP32[$__n>>2]|0;
 $9 = (($8) - ($4))|0;
 $10 = $9 >> 2;
 $phitmp = $4;
 $11 = $2 ? $phitmp : 0;
 $12 = (_realloc($11,$$)|0);
 $13 = ($12|0)==(0|0);
 if ($13) {
  __ZSt17__throw_bad_allocv();
 }
 do {
  if ($2) {
   $14 = $12;
   HEAP32[$__b>>2] = $14;
   $24 = $12;
  } else {
   $$pr = HEAP32[$__b>>2]|0;
   $15 = $12;
   HEAP32[$__b>>2] = $15;
   $16 = ($$pr|0)==(0|0);
   if ($16) {
    $24 = $12;
   } else {
    $17 = HEAP32[$0>>2]|0;
    __THREW__ = 0;
    invoke_vi($17|0,($$pr|0));
    $18 = __THREW__; __THREW__ = 0;
    $19 = $18&1;
    if ($19) {
     $20 = ___cxa_find_matching_catch_3(0|0)|0;
     $21 = tempRet0;
     ___clang_call_terminate($20);
     // unreachable;
    } else {
     $$pre = HEAP32[$__b>>2]|0;
     $24 = $$pre;
     break;
    }
   }
  }
 } while(0);
 HEAP32[$0>>2] = (337);
 $22 = $$ >>> 2;
 $23 = (($24) + ($10<<2)|0);
 HEAP32[$__n>>2] = $23;
 $25 = HEAP32[$__b>>2]|0;
 $26 = (($25) + ($22<<2)|0);
 HEAP32[$__e>>2] = $26;
 return;
}
function __ZNKSt3__19money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_bRNS_8ios_baseERjRNS_12basic_stringIwS3_NS_9allocatorIwEEEE($this,$__b,$__e,$__intl,$__iob,$__err,$__v) {
 $this = $this|0;
 $__b = $__b|0;
 $__e = $__e|0;
 $__intl = $__intl|0;
 $__iob = $__iob|0;
 $__err = $__err|0;
 $__v = $__v|0;
 var $$0$i$i$i$i = 0, $$0$i$i1$i$i = 0, $$011 = 0, $$0110 = 0, $$012 = 0, $$019 = 0, $$byval_copy = 0, $$pre = 0, $$pre$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__loc = 0, $__neg = 0, $__w$0 = 0, $__w$0$lcssa = 0, $__wb = 0, $__wbuf = 0, $__wn = 0, $phitmp$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 432|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 424|0;
 $__wbuf = sp + 24|0;
 $__wb = sp + 16|0;
 $__wn = sp + 8|0;
 $__loc = sp + 4|0;
 $__neg = sp + 428|0;
 $0 = sp;
 $1 = $__wbuf;
 HEAP32[$__wb>>2] = $1;
 $2 = ((($__wb)) + 4|0);
 HEAP32[$2>>2] = (4899);
 $3 = ((($__wbuf)) + 400|0);
 __THREW__ = 0;
 $4 = (invoke_ii(3373,($__iob|0))|0);
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $33 = ___cxa_find_matching_catch_2()|0;
  $34 = tempRet0;
  HEAP32[$__wb>>2] = 0;
  $$011 = $33;$$019 = $34;$94 = $__wbuf;
  label = 39;
 } else {
  HEAP32[$__loc>>2] = $4;
  __THREW__ = 0;
  $7 = (invoke_iii(3374,($__loc|0),(1207608|0))|0);
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  L4: do {
   if (!($9)) {
    HEAP8[$__neg>>0] = 0;
    $10 = HEAP32[$__e>>2]|0;
    HEAP32[$0>>2] = $10;
    $11 = ((($__iob)) + 4|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = $10;
    __THREW__ = 0;
    ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
    $14 = (invoke_iiiiiiiiiiii(4906,($__b|0),($$byval_copy|0),($__intl|0),($__loc|0),($12|0),($__err|0),($__neg|0),($7|0),($__wb|0),($__wn|0),($3|0))|0);
    $15 = __THREW__; __THREW__ = 0;
    $16 = $15&1;
    if (!($16)) {
     if ($14) {
      $17 = HEAP8[$__v>>0]|0;
      $18 = $17 & 1;
      $19 = ($18<<24>>24)==(0);
      if ($19) {
       HEAP8[$__v>>0] = 0;
      } else {
       $20 = ((($__v)) + 8|0);
       $21 = HEAP32[$20>>2]|0;
       HEAP32[$21>>2] = 0;
      }
      $22 = ((($__v)) + 4|0);
      HEAP32[$22>>2] = 0;
      $23 = HEAP8[$__neg>>0]|0;
      $24 = ($23<<24>>24)==(0);
      if (!($24)) {
       $25 = HEAP32[$7>>2]|0;
       $26 = ((($25)) + 44|0);
       $27 = HEAP32[$26>>2]|0;
       __THREW__ = 0;
       $28 = (invoke_iii($27|0,($7|0),45)|0);
       $29 = __THREW__; __THREW__ = 0;
       $30 = $29&1;
       if ($30) {
        break;
       }
       __THREW__ = 0;
       invoke_vii(4908,($__v|0),($28|0));
       $31 = __THREW__; __THREW__ = 0;
       $32 = $31&1;
       if ($32) {
        break;
       }
      }
      $35 = HEAP32[$7>>2]|0;
      $36 = ((($35)) + 44|0);
      $37 = HEAP32[$36>>2]|0;
      __THREW__ = 0;
      $38 = (invoke_iii($37|0,($7|0),48)|0);
      $39 = __THREW__; __THREW__ = 0;
      $40 = $39&1;
      if ($40) {
       break;
      }
      $41 = HEAP32[$__wb>>2]|0;
      $42 = HEAP32[$__wn>>2]|0;
      $43 = ((($42)) + -4|0);
      $__w$0 = $41;
      while(1) {
       $44 = ($__w$0>>>0)<($43>>>0);
       if (!($44)) {
        $__w$0$lcssa = $__w$0;
        break;
       }
       $45 = HEAP32[$__w$0>>2]|0;
       $46 = ($45|0)==($38|0);
       if (!($46)) {
        $__w$0$lcssa = $__w$0;
        break;
       }
       $47 = ((($__w$0)) + 4|0);
       $__w$0 = $47;
      }
      __THREW__ = 0;
      (invoke_iiii(4911,($__v|0),($__w$0$lcssa|0),($42|0))|0);
      $48 = __THREW__; __THREW__ = 0;
      $49 = $48&1;
      if ($49) {
       break;
      }
     }
     $50 = HEAP32[$__b>>2]|0;
     $51 = ($50|0)==(0|0);
     do {
      if ($51) {
       $99 = 1;
      } else {
       $52 = ((($50)) + 12|0);
       $53 = HEAP32[$52>>2]|0;
       $54 = ((($50)) + 16|0);
       $55 = HEAP32[$54>>2]|0;
       $56 = ($53|0)==($55|0);
       if ($56) {
        $57 = HEAP32[$50>>2]|0;
        $58 = ((($57)) + 36|0);
        $59 = HEAP32[$58>>2]|0;
        __THREW__ = 0;
        $60 = (invoke_ii($59|0,($50|0))|0);
        $61 = __THREW__; __THREW__ = 0;
        $62 = $61&1;
        if ($62) {
         break L4;
        } else {
         $$0$i$i$i$i = $60;
        }
       } else {
        $63 = HEAP32[$53>>2]|0;
        $$0$i$i$i$i = $63;
       }
       $64 = ($$0$i$i$i$i|0)==(-1);
       if ($64) {
        HEAP32[$__b>>2] = 0;
        $99 = 1;
        break;
       } else {
        $$pre$i$i$i = HEAP32[$__b>>2]|0;
        $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
        $99 = $phitmp$i$i$i;
        break;
       }
      }
     } while(0);
     $65 = ($10|0)==(0);
     do {
      if ($65) {
       label = 31;
      } else {
       $66 = ((($13)) + 12|0);
       $67 = HEAP32[$66>>2]|0;
       $68 = ((($13)) + 16|0);
       $69 = HEAP32[$68>>2]|0;
       $70 = ($67|0)==($69|0);
       if ($70) {
        $71 = $10;
        $72 = HEAP32[$71>>2]|0;
        $73 = ((($72)) + 36|0);
        $74 = HEAP32[$73>>2]|0;
        __THREW__ = 0;
        $75 = (invoke_ii($74|0,($13|0))|0);
        $76 = __THREW__; __THREW__ = 0;
        $77 = $76&1;
        if ($77) {
         break L4;
        } else {
         $$0$i$i1$i$i = $75;
        }
       } else {
        $78 = HEAP32[$67>>2]|0;
        $$0$i$i1$i$i = $78;
       }
       $79 = ($$0$i$i1$i$i|0)==(-1);
       if ($79) {
        HEAP32[$__e>>2] = 0;
        label = 31;
        break;
       } else {
        if ($99) {
         break;
        } else {
         label = 33;
         break;
        }
       }
      }
     } while(0);
     if ((label|0) == 31) {
      if ($99) {
       label = 33;
      }
     }
     if ((label|0) == 33) {
      $80 = HEAP32[$__err>>2]|0;
      $81 = $80 | 2;
      HEAP32[$__err>>2] = $81;
     }
     $82 = HEAP32[$__b>>2]|0;
     (__ZNSt3__114__shared_count16__release_sharedEv($4)|0);
     $83 = HEAP32[$__wb>>2]|0;
     HEAP32[$__wb>>2] = 0;
     $84 = ($83|0)==(0|0);
     if (!($84)) {
      $85 = HEAP32[$2>>2]|0;
      __THREW__ = 0;
      invoke_vi($85|0,($83|0));
      $86 = __THREW__; __THREW__ = 0;
      $87 = $86&1;
      if ($87) {
       $88 = ___cxa_find_matching_catch_3(0|0)|0;
       $89 = tempRet0;
       ___clang_call_terminate($88);
       // unreachable;
      }
     }
     STACKTOP = sp;return ($82|0);
    }
   }
  } while(0);
  $90 = ___cxa_find_matching_catch_2()|0;
  $91 = tempRet0;
  (__ZNSt3__114__shared_count16__release_sharedEv($4)|0);
  $$pre = HEAP32[$__wb>>2]|0;
  HEAP32[$__wb>>2] = 0;
  $92 = ($$pre|0)==(0|0);
  if ($92) {
   $$0110 = $91;$$012 = $90;
  } else {
   $$011 = $90;$$019 = $91;$94 = $$pre;
   label = 39;
  }
 }
 if ((label|0) == 39) {
  $93 = HEAP32[$2>>2]|0;
  __THREW__ = 0;
  invoke_vi($93|0,($94|0));
  $95 = __THREW__; __THREW__ = 0;
  $96 = $95&1;
  if ($96) {
   $97 = ___cxa_find_matching_catch_3(0|0)|0;
   $98 = tempRet0;
   ___clang_call_terminate($97);
   // unreachable;
  } else {
   $$0110 = $$019;$$012 = $$011;
  }
 }
 ___resumeException($$012|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6appendIPwEENS_9enable_ifIXsr21__is_forward_iteratorIT_EE5valueERS5_E4typeES9_S9_($this,$__first,$__last) {
 $this = $this|0;
 $__first = $__first|0;
 $__last = $__last|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__p$0 = 0, $phitmp$i = 0, $scevgep = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $__first;
 $1 = HEAP8[$this>>0]|0;
 $2 = $1 & 1;
 $3 = ($2<<24>>24)==(0);
 if ($3) {
  $4 = $1&255;
  $5 = $4 >>> 1;
  $16 = 1;$17 = $5;$44 = $1;
 } else {
  $6 = ((($this)) + 4|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = HEAP32[$this>>2]|0;
  $9 = $8 & -2;
  $phitmp$i = (($9) + -1)|0;
  $10 = $8&255;
  $16 = $phitmp$i;$17 = $7;$44 = $10;
 }
 $11 = $__last;
 $12 = (($11) - ($0))|0;
 $13 = $12 >> 2;
 $14 = ($13|0)==(0);
 do {
  if (!($14)) {
   $15 = (($16) - ($17))|0;
   $18 = ($15>>>0)<($13>>>0);
   if ($18) {
    $19 = (($17) + ($13))|0;
    $20 = (($19) - ($16))|0;
    __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE9__grow_byEjjjjjj($this,$16,$20,$17,$17,0,0);
    $$pre = HEAP8[$this>>0]|0;
    $22 = $$pre;
   } else {
    $22 = $44;
   }
   $21 = $22 & 1;
   $23 = ($21<<24>>24)==(0);
   if ($23) {
    $26 = ((($this)) + 4|0);
    $28 = $26;
   } else {
    $24 = ((($this)) + 8|0);
    $25 = HEAP32[$24>>2]|0;
    $28 = $25;
   }
   $27 = (($28) + ($17<<2)|0);
   $29 = $__last;
   $30 = (($29) - ($0))|0;
   $31 = $30 >>> 2;
   $32 = (($17) + ($31))|0;
   $$0 = $__first;$__p$0 = $27;
   while(1) {
    $33 = ($$0|0)==($__last|0);
    if ($33) {
     break;
    }
    $34 = HEAP32[$$0>>2]|0;
    HEAP32[$__p$0>>2] = $34;
    $35 = ((($__p$0)) + 4|0);
    $36 = ((($$0)) + 4|0);
    $$0 = $36;$__p$0 = $35;
   }
   $scevgep = (($28) + ($32<<2)|0);
   HEAP32[$scevgep>>2] = 0;
   $37 = (($17) + ($13))|0;
   $38 = HEAP8[$this>>0]|0;
   $39 = $38 & 1;
   $40 = ($39<<24>>24)==(0);
   if ($40) {
    $42 = $37 << 1;
    $43 = $42&255;
    HEAP8[$this>>0] = $43;
    break;
   } else {
    $41 = ((($this)) + 4|0);
    HEAP32[$41>>2] = $37;
    break;
   }
  }
 } while(0);
 return ($this|0);
}
function __ZNSt3__19money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__19money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__19money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_bRNS_8ios_baseEce($this,$__s,$__intl,$__iob,$__fl,$__units) {
 $this = $this|0;
 $__s = $__s|0;
 $__intl = $__intl|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $__units = +$__units;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$byval_copy = 0, $$pre = 0, $$pre23 = 0, $$pre23$pre = 0, $$pre25 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $__bb = 0, $__buf = 0, $__db$0 = 0, $__digits = 0, $__dp = 0, $__fd = 0, $__grp = 0, $__hd$sroa$0$0 = 0, $__hd$sroa$0$1 = 0, $__hd$sroa$0$2 = 0, $__hn$sroa$0$0 = 0, $__hn$sroa$0$1 = 0, $__hn$sroa$0$2 = 0, $__hw$sroa$0$0 = 0, $__hw$sroa$0$1 = 0, $__i$0$i$i = 0, $__i$0$i$i6 = 0, $__i$0$i$i9 = 0;
 var $__loc = 0, $__mb$0 = 0, $__mbuf = 0, $__me = 0, $__mi = 0, $__n$0 = 0, $__pat = 0, $__sn = 0, $__sym = 0, $__ts = 0, $exitcond$i$i = 0, $exitcond$i$i10 = 0, $exitcond$i$i7 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 400|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 76|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $__buf = sp + 288|0;
 $__bb = sp + 72|0;
 $__digits = sp + 188|0;
 $__loc = sp + 68|0;
 $__pat = sp + 184|0;
 $__dp = sp + 181|0;
 $__ts = sp + 180|0;
 $__grp = sp + 56|0;
 $__sym = sp + 44|0;
 $__sn = sp + 32|0;
 $__fd = sp + 28|0;
 $__mbuf = sp + 80|0;
 $__mi = sp + 24|0;
 $__me = sp + 20|0;
 $0 = sp + 16|0;
 HEAP32[$__bb>>2] = $__buf;
 HEAPF64[$vararg_buffer>>3] = $__units;
 $1 = (_snprintf($__buf,100,622023,$vararg_buffer)|0);
 $2 = ($1>>>0)>(99);
 do {
  if ($2) {
   __THREW__ = 0;
   $3 = (invoke_i(4874)|0);
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   if ($5) {
    $__hd$sroa$0$0 = 0;$__hn$sroa$0$0 = 0;
    label = 8;
   } else {
    __THREW__ = 0;
    HEAPF64[$vararg_buffer1>>3] = $__units;
    $6 = (invoke_iiiii(4879,($__bb|0),($3|0),(622023|0),($vararg_buffer1|0))|0);
    $7 = __THREW__; __THREW__ = 0;
    $8 = $7&1;
    if ($8) {
     $__hd$sroa$0$0 = 0;$__hn$sroa$0$0 = 0;
     label = 8;
    } else {
     $9 = HEAP32[$__bb>>2]|0;
     $10 = ($9|0)==(0|0);
     if ($10) {
      __THREW__ = 0;
      invoke_v(4863);
      $12 = __THREW__; __THREW__ = 0;
      $13 = $12&1;
      if ($13) {
       $__hd$sroa$0$0 = 0;$__hn$sroa$0$0 = 0;
       label = 8;
       break;
      }
      $$pre = HEAP32[$__bb>>2]|0;
      $130 = $$pre;
     } else {
      $11 = $9;
      $130 = $11;
     }
     $16 = (_malloc($6)|0);
     $17 = $16;
     $18 = ($16|0)==(0|0);
     if ($18) {
      __THREW__ = 0;
      invoke_v(4863);
      $19 = __THREW__; __THREW__ = 0;
      $20 = $19&1;
      if ($20) {
       $__hd$sroa$0$0 = $17;$__hn$sroa$0$0 = $130;
       label = 8;
      } else {
       $__db$0 = 0;$__hd$sroa$0$1 = $17;$__hn$sroa$0$1 = $130;$__n$0 = $6;
       label = 11;
      }
     } else {
      $__db$0 = $16;$__hd$sroa$0$1 = $17;$__hn$sroa$0$1 = $130;$__n$0 = $6;
      label = 11;
     }
    }
   }
  } else {
   $__db$0 = $__digits;$__hd$sroa$0$1 = 0;$__hn$sroa$0$1 = 0;$__n$0 = $1;
   label = 11;
  }
 } while(0);
 if ((label|0) == 11) {
  __THREW__ = 0;
  $21 = (invoke_ii(3373,($__iob|0))|0);
  $22 = __THREW__; __THREW__ = 0;
  $23 = $22&1;
  if ($23) {
   $__hd$sroa$0$0 = $__hd$sroa$0$1;$__hn$sroa$0$0 = $__hn$sroa$0$1;
   label = 8;
  } else {
   HEAP32[$__loc>>2] = $21;
   __THREW__ = 0;
   $24 = (invoke_iii(3374,($__loc|0),(1207576|0))|0);
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    label = 32;
   } else {
    $27 = HEAP32[$__bb>>2]|0;
    $28 = (($27) + ($__n$0)|0);
    $29 = HEAP32[$24>>2]|0;
    $30 = ((($29)) + 32|0);
    $31 = HEAP32[$30>>2]|0;
    __THREW__ = 0;
    (invoke_iiiii($31|0,($24|0),($27|0),($28|0),($__db$0|0))|0);
    $32 = __THREW__; __THREW__ = 0;
    $33 = $32&1;
    if ($33) {
     label = 32;
    } else {
     $34 = ($__n$0|0)==(0);
     if ($34) {
      $44 = 0;
     } else {
      $35 = HEAP32[$__bb>>2]|0;
      $36 = HEAP8[$35>>0]|0;
      $37 = ($36<<24>>24)==(45);
      $44 = $37;
     }
     $__i$0$i$i = 0;
     while(1) {
      $exitcond$i$i = ($__i$0$i$i|0)==(3);
      if ($exitcond$i$i) {
       break;
      }
      $38 = (($__grp) + ($__i$0$i$i<<2)|0);
      HEAP32[$38>>2] = 0;
      $39 = (($__i$0$i$i) + 1)|0;
      $__i$0$i$i = $39;
     }
     $__i$0$i$i6 = 0;
     while(1) {
      $exitcond$i$i7 = ($__i$0$i$i6|0)==(3);
      if ($exitcond$i$i7) {
       break;
      }
      $40 = (($__sym) + ($__i$0$i$i6<<2)|0);
      HEAP32[$40>>2] = 0;
      $41 = (($__i$0$i$i6) + 1)|0;
      $__i$0$i$i6 = $41;
     }
     $__i$0$i$i9 = 0;
     while(1) {
      $exitcond$i$i10 = ($__i$0$i$i9|0)==(3);
      if ($exitcond$i$i10) {
       break;
      }
      $42 = (($__sn) + ($__i$0$i$i9<<2)|0);
      HEAP32[$42>>2] = 0;
      $43 = (($__i$0$i$i9) + 1)|0;
      $__i$0$i$i9 = $43;
     }
     __THREW__ = 0;
     invoke_viiiiiiiiii(4912,($__intl|0),($44|0),($__loc|0),($__pat|0),($__dp|0),($__ts|0),($__grp|0),($__sym|0),($__sn|0),($__fd|0));
     $45 = __THREW__; __THREW__ = 0;
     $46 = $45&1;
     if ($46) {
      $99 = ___cxa_find_matching_catch_2()|0;
      $100 = tempRet0;
      $$0 = $99;$$01 = $100;$$pre23 = $21;
     } else {
      $47 = HEAP32[$__fd>>2]|0;
      $48 = ($__n$0|0)>($47|0);
      if ($48) {
       $49 = (($__n$0) - ($47))|0;
       $50 = $49 << 1;
       $51 = HEAP8[$__sn>>0]|0;
       $52 = $51 & 1;
       $53 = ($52<<24>>24)==(0);
       $54 = ((($__sn)) + 4|0);
       $55 = HEAP32[$54>>2]|0;
       $56 = $51&255;
       $57 = $56 >>> 1;
       $58 = $53 ? $57 : $55;
       $59 = HEAP8[$__sym>>0]|0;
       $60 = $59 & 1;
       $61 = ($60<<24>>24)==(0);
       $62 = ((($__sym)) + 4|0);
       $63 = HEAP32[$62>>2]|0;
       $64 = $59&255;
       $65 = $64 >>> 1;
       $66 = $61 ? $65 : $63;
       $67 = $50 | 1;
       $68 = (($67) + ($47))|0;
       $69 = (($68) + ($58))|0;
       $70 = (($69) + ($66))|0;
       $91 = $70;
      } else {
       $71 = HEAP8[$__sn>>0]|0;
       $72 = $71 & 1;
       $73 = ($72<<24>>24)==(0);
       $74 = ((($__sn)) + 4|0);
       $75 = HEAP32[$74>>2]|0;
       $76 = $71&255;
       $77 = $76 >>> 1;
       $78 = $73 ? $77 : $75;
       $79 = HEAP8[$__sym>>0]|0;
       $80 = $79 & 1;
       $81 = ($80<<24>>24)==(0);
       $82 = ((($__sym)) + 4|0);
       $83 = HEAP32[$82>>2]|0;
       $84 = $79&255;
       $85 = $84 >>> 1;
       $86 = $81 ? $85 : $83;
       $87 = (($47) + 2)|0;
       $88 = (($87) + ($78))|0;
       $89 = (($88) + ($86))|0;
       $91 = $89;
      }
      $90 = ($91>>>0)>(100);
      if ($90) {
       $92 = (_malloc($91)|0);
       $93 = $92;
       $94 = ($92|0)==(0|0);
       if ($94) {
        __THREW__ = 0;
        invoke_v(4863);
        $95 = __THREW__; __THREW__ = 0;
        $96 = $95&1;
        if ($96) {
         $__hw$sroa$0$0 = $93;
        } else {
         $__hw$sroa$0$1 = $93;$__mb$0 = 0;
         label = 36;
        }
       } else {
        $__hw$sroa$0$1 = $93;$__mb$0 = $92;
        label = 36;
       }
      } else {
       $__hw$sroa$0$1 = 0;$__mb$0 = $__mbuf;
       label = 36;
      }
      if ((label|0) == 36) {
       $105 = ((($__iob)) + 4|0);
       $106 = HEAP32[$105>>2]|0;
       $107 = (($__db$0) + ($__n$0)|0);
       $108 = HEAP8[$__dp>>0]|0;
       $109 = HEAP8[$__ts>>0]|0;
       __THREW__ = 0;
       invoke_viiiiiiiiiiiiiii(4913,($__mb$0|0),($__mi|0),($__me|0),($106|0),($__db$0|0),($107|0),($24|0),($44|0),($__pat|0),($108|0),($109|0),($__grp|0),($__sym|0),($__sn|0),($47|0));
       $110 = __THREW__; __THREW__ = 0;
       $111 = $110&1;
       if ($111) {
        $__hw$sroa$0$0 = $__hw$sroa$0$1;
       } else {
        $112 = HEAP32[$__s>>2]|0;
        HEAP32[$0>>2] = $112;
        $113 = HEAP32[$__mi>>2]|0;
        $114 = HEAP32[$__me>>2]|0;
        __THREW__ = 0;
        ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
        $115 = (invoke_iiiiiii(3375,($$byval_copy|0),($__mb$0|0),($113|0),($114|0),($__iob|0),($__fl|0))|0);
        $116 = __THREW__; __THREW__ = 0;
        $117 = $116&1;
        if ($117) {
         $__hw$sroa$0$0 = $__hw$sroa$0$1;
        } else {
         $118 = ($__hw$sroa$0$1|0)==(0);
         if ($118) {
          $120 = $21;
         } else {
          $119 = $__hw$sroa$0$1;
          _free($119);
          $$pre25 = HEAP32[$__loc>>2]|0;
          $120 = $$pre25;
         }
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__sn);
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__sym);
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grp);
         (__ZNSt3__114__shared_count16__release_sharedEv($120)|0);
         $121 = ($__hd$sroa$0$1|0)==(0);
         if (!($121)) {
          $122 = $__hd$sroa$0$1;
          _free($122);
         }
         $123 = ($__hn$sroa$0$1|0)==(0);
         if (!($123)) {
          $124 = $__hn$sroa$0$1;
          _free($124);
         }
         STACKTOP = sp;return ($115|0);
        }
       }
      }
      $101 = ___cxa_find_matching_catch_2()|0;
      $102 = tempRet0;
      $103 = ($__hw$sroa$0$0|0)==(0);
      if ($103) {
       $$0 = $101;$$01 = $102;$$pre23 = $21;
      } else {
       $104 = $__hw$sroa$0$0;
       _free($104);
       $$pre23$pre = HEAP32[$__loc>>2]|0;
       $$0 = $101;$$01 = $102;$$pre23 = $$pre23$pre;
      }
     }
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__sn);
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__sym);
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grp);
     $$1 = $$0;$$12 = $$01;$125 = $$pre23;
    }
   }
   if ((label|0) == 32) {
    $97 = ___cxa_find_matching_catch_2()|0;
    $98 = tempRet0;
    $$1 = $97;$$12 = $98;$125 = $21;
   }
   (__ZNSt3__114__shared_count16__release_sharedEv($125)|0);
   $$2 = $$1;$$23 = $$12;$__hd$sroa$0$2 = $__hd$sroa$0$1;$__hn$sroa$0$2 = $__hn$sroa$0$1;
  }
 }
 if ((label|0) == 8) {
  $14 = ___cxa_find_matching_catch_2()|0;
  $15 = tempRet0;
  $$2 = $14;$$23 = $15;$__hd$sroa$0$2 = $__hd$sroa$0$0;$__hn$sroa$0$2 = $__hn$sroa$0$0;
 }
 $126 = ($__hd$sroa$0$2|0)==(0);
 if (!($126)) {
  $127 = $__hd$sroa$0$2;
  _free($127);
 }
 $128 = ($__hn$sroa$0$2|0)==(0);
 if (!($128)) {
  $129 = $__hn$sroa$0$2;
  _free($129);
 }
 ___resumeException($$2|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__111__money_putIcE13__gather_infoEbbRKNS_6localeERNS_10money_base7patternERcS8_RNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEESF_SF_Ri($__intl,$__neg,$__loc,$__pat,$__dp,$__ts,$__grp,$__sym,$__sn,$__fd) {
 $__intl = $__intl|0;
 $__neg = $__neg|0;
 $__loc = $__loc|0;
 $__pat = $__pat|0;
 $__dp = $__dp|0;
 $__ts = $__ts|0;
 $__grp = $__grp|0;
 $__sym = $__sym|0;
 $__sn = $__sn|0;
 $__fd = $__fd|0;
 var $$pre$phi39Z2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $__i$0$i$i$i = 0, $__i$0$i$i$i11 = 0, $__i$0$i$i$i16 = 0, $__i$0$i$i$i21 = 0, $__i$0$i$i$i26 = 0, $__i$0$i$i$i31 = 0, $__i$0$i$i$i36 = 0, $__i$0$i$i$i6 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i12 = 0, $exitcond$i$i$i17 = 0, $exitcond$i$i$i22 = 0, $exitcond$i$i$i27 = 0, $exitcond$i$i$i32 = 0, $exitcond$i$i$i37 = 0, $exitcond$i$i$i7 = 0, $storemerge = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 108|0;
 $1 = sp + 96|0;
 $2 = sp + 92|0;
 $3 = sp + 80|0;
 $4 = sp + 68|0;
 $5 = sp + 56|0;
 $6 = sp + 52|0;
 $7 = sp + 40|0;
 $8 = sp + 36|0;
 $9 = sp + 24|0;
 $10 = sp + 12|0;
 $11 = sp;
 if ($__intl) {
  $12 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1209296)|0);
  $13 = HEAP32[$12>>2]|0;
  if ($__neg) {
   $14 = ((($13)) + 44|0);
   $15 = HEAP32[$14>>2]|0;
   FUNCTION_TABLE_vii[$15 & 8191]($0,$12);
   $16 = HEAP32[$0>>2]|0;
   HEAP8[$__pat>>0]=$16&255;HEAP8[$__pat+1>>0]=($16>>8)&255;HEAP8[$__pat+2>>0]=($16>>16)&255;HEAP8[$__pat+3>>0]=$16>>24;
   $17 = HEAP32[$12>>2]|0;
   $18 = ((($17)) + 32|0);
   $19 = HEAP32[$18>>2]|0;
   FUNCTION_TABLE_vii[$19 & 8191]($1,$12);
   $20 = HEAP8[$__sn>>0]|0;
   $21 = $20 & 1;
   $22 = ($21<<24>>24)==(0);
   if ($22) {
    $26 = ((($__sn)) + 1|0);
    HEAP8[$26>>0] = 0;
    HEAP8[$__sn>>0] = 0;
   } else {
    $23 = ((($__sn)) + 8|0);
    $24 = HEAP32[$23>>2]|0;
    HEAP8[$24>>0] = 0;
    $25 = ((($__sn)) + 4|0);
    HEAP32[$25>>2] = 0;
   }
   __THREW__ = 0;
   invoke_vii(3368,($__sn|0),0);
   $27 = __THREW__; __THREW__ = 0;
   $28 = $27&1;
   if ($28) {
    $29 = ___cxa_find_matching_catch_3(0|0)|0;
    $30 = tempRet0;
    ___clang_call_terminate($29);
    // unreachable;
   }
   ;HEAP32[$__sn>>2]=HEAP32[$1>>2]|0;HEAP32[$__sn+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$__sn+8>>2]=HEAP32[$1+8>>2]|0;
   $__i$0$i$i$i = 0;
   while(1) {
    $exitcond$i$i$i = ($__i$0$i$i$i|0)==(3);
    if ($exitcond$i$i$i) {
     break;
    }
    $31 = (($1) + ($__i$0$i$i$i<<2)|0);
    HEAP32[$31>>2] = 0;
    $32 = (($__i$0$i$i$i) + 1)|0;
    $__i$0$i$i$i = $32;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1);
   $$pre$phiZ2D = $12;
  } else {
   $33 = ((($13)) + 40|0);
   $34 = HEAP32[$33>>2]|0;
   FUNCTION_TABLE_vii[$34 & 8191]($2,$12);
   $35 = HEAP32[$2>>2]|0;
   HEAP8[$__pat>>0]=$35&255;HEAP8[$__pat+1>>0]=($35>>8)&255;HEAP8[$__pat+2>>0]=($35>>16)&255;HEAP8[$__pat+3>>0]=$35>>24;
   $36 = HEAP32[$12>>2]|0;
   $37 = ((($36)) + 28|0);
   $38 = HEAP32[$37>>2]|0;
   FUNCTION_TABLE_vii[$38 & 8191]($3,$12);
   $39 = HEAP8[$__sn>>0]|0;
   $40 = $39 & 1;
   $41 = ($40<<24>>24)==(0);
   if ($41) {
    $45 = ((($__sn)) + 1|0);
    HEAP8[$45>>0] = 0;
    HEAP8[$__sn>>0] = 0;
   } else {
    $42 = ((($__sn)) + 8|0);
    $43 = HEAP32[$42>>2]|0;
    HEAP8[$43>>0] = 0;
    $44 = ((($__sn)) + 4|0);
    HEAP32[$44>>2] = 0;
   }
   __THREW__ = 0;
   invoke_vii(3368,($__sn|0),0);
   $46 = __THREW__; __THREW__ = 0;
   $47 = $46&1;
   if ($47) {
    $48 = ___cxa_find_matching_catch_3(0|0)|0;
    $49 = tempRet0;
    ___clang_call_terminate($48);
    // unreachable;
   }
   ;HEAP32[$__sn>>2]=HEAP32[$3>>2]|0;HEAP32[$__sn+4>>2]=HEAP32[$3+4>>2]|0;HEAP32[$__sn+8>>2]=HEAP32[$3+8>>2]|0;
   $__i$0$i$i$i6 = 0;
   while(1) {
    $exitcond$i$i$i7 = ($__i$0$i$i$i6|0)==(3);
    if ($exitcond$i$i$i7) {
     break;
    }
    $50 = (($3) + ($__i$0$i$i$i6<<2)|0);
    HEAP32[$50>>2] = 0;
    $51 = (($__i$0$i$i$i6) + 1)|0;
    $__i$0$i$i$i6 = $51;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3);
   $$pre$phiZ2D = $12;
  }
  $52 = HEAP32[$12>>2]|0;
  $53 = ((($52)) + 12|0);
  $54 = HEAP32[$53>>2]|0;
  $55 = (FUNCTION_TABLE_ii[$54 & 8191]($12)|0);
  HEAP8[$__dp>>0] = $55;
  $56 = HEAP32[$12>>2]|0;
  $57 = ((($56)) + 16|0);
  $58 = HEAP32[$57>>2]|0;
  $59 = (FUNCTION_TABLE_ii[$58 & 8191]($12)|0);
  HEAP8[$__ts>>0] = $59;
  $60 = HEAP32[$$pre$phiZ2D>>2]|0;
  $61 = ((($60)) + 20|0);
  $62 = HEAP32[$61>>2]|0;
  FUNCTION_TABLE_vii[$62 & 8191]($4,$12);
  $63 = HEAP8[$__grp>>0]|0;
  $64 = $63 & 1;
  $65 = ($64<<24>>24)==(0);
  if ($65) {
   $69 = ((($__grp)) + 1|0);
   HEAP8[$69>>0] = 0;
   HEAP8[$__grp>>0] = 0;
  } else {
   $66 = ((($__grp)) + 8|0);
   $67 = HEAP32[$66>>2]|0;
   HEAP8[$67>>0] = 0;
   $68 = ((($__grp)) + 4|0);
   HEAP32[$68>>2] = 0;
  }
  __THREW__ = 0;
  invoke_vii(3368,($__grp|0),0);
  $70 = __THREW__; __THREW__ = 0;
  $71 = $70&1;
  if ($71) {
   $72 = ___cxa_find_matching_catch_3(0|0)|0;
   $73 = tempRet0;
   ___clang_call_terminate($72);
   // unreachable;
  }
  ;HEAP32[$__grp>>2]=HEAP32[$4>>2]|0;HEAP32[$__grp+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$__grp+8>>2]=HEAP32[$4+8>>2]|0;
  $__i$0$i$i$i11 = 0;
  while(1) {
   $exitcond$i$i$i12 = ($__i$0$i$i$i11|0)==(3);
   if ($exitcond$i$i$i12) {
    break;
   }
   $74 = (($4) + ($__i$0$i$i$i11<<2)|0);
   HEAP32[$74>>2] = 0;
   $75 = (($__i$0$i$i$i11) + 1)|0;
   $__i$0$i$i$i11 = $75;
  }
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4);
  $76 = HEAP32[$$pre$phiZ2D>>2]|0;
  $77 = ((($76)) + 24|0);
  $78 = HEAP32[$77>>2]|0;
  FUNCTION_TABLE_vii[$78 & 8191]($5,$12);
  $79 = HEAP8[$__sym>>0]|0;
  $80 = $79 & 1;
  $81 = ($80<<24>>24)==(0);
  if ($81) {
   $85 = ((($__sym)) + 1|0);
   HEAP8[$85>>0] = 0;
   HEAP8[$__sym>>0] = 0;
  } else {
   $82 = ((($__sym)) + 8|0);
   $83 = HEAP32[$82>>2]|0;
   HEAP8[$83>>0] = 0;
   $84 = ((($__sym)) + 4|0);
   HEAP32[$84>>2] = 0;
  }
  __THREW__ = 0;
  invoke_vii(3368,($__sym|0),0);
  $86 = __THREW__; __THREW__ = 0;
  $87 = $86&1;
  if ($87) {
   $88 = ___cxa_find_matching_catch_3(0|0)|0;
   $89 = tempRet0;
   ___clang_call_terminate($88);
   // unreachable;
  }
  ;HEAP32[$__sym>>2]=HEAP32[$5>>2]|0;HEAP32[$__sym+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$__sym+8>>2]=HEAP32[$5+8>>2]|0;
  $__i$0$i$i$i16 = 0;
  while(1) {
   $exitcond$i$i$i17 = ($__i$0$i$i$i16|0)==(3);
   if ($exitcond$i$i$i17) {
    break;
   }
   $90 = (($5) + ($__i$0$i$i$i16<<2)|0);
   HEAP32[$90>>2] = 0;
   $91 = (($__i$0$i$i$i16) + 1)|0;
   $__i$0$i$i$i16 = $91;
  }
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5);
  $92 = HEAP32[$12>>2]|0;
  $93 = ((($92)) + 36|0);
  $94 = HEAP32[$93>>2]|0;
  $95 = (FUNCTION_TABLE_ii[$94 & 8191]($12)|0);
  $storemerge = $95;
 } else {
  $96 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1209288)|0);
  $97 = HEAP32[$96>>2]|0;
  if ($__neg) {
   $98 = ((($97)) + 44|0);
   $99 = HEAP32[$98>>2]|0;
   FUNCTION_TABLE_vii[$99 & 8191]($6,$96);
   $100 = HEAP32[$6>>2]|0;
   HEAP8[$__pat>>0]=$100&255;HEAP8[$__pat+1>>0]=($100>>8)&255;HEAP8[$__pat+2>>0]=($100>>16)&255;HEAP8[$__pat+3>>0]=$100>>24;
   $101 = HEAP32[$96>>2]|0;
   $102 = ((($101)) + 32|0);
   $103 = HEAP32[$102>>2]|0;
   FUNCTION_TABLE_vii[$103 & 8191]($7,$96);
   $104 = HEAP8[$__sn>>0]|0;
   $105 = $104 & 1;
   $106 = ($105<<24>>24)==(0);
   if ($106) {
    $110 = ((($__sn)) + 1|0);
    HEAP8[$110>>0] = 0;
    HEAP8[$__sn>>0] = 0;
   } else {
    $107 = ((($__sn)) + 8|0);
    $108 = HEAP32[$107>>2]|0;
    HEAP8[$108>>0] = 0;
    $109 = ((($__sn)) + 4|0);
    HEAP32[$109>>2] = 0;
   }
   __THREW__ = 0;
   invoke_vii(3368,($__sn|0),0);
   $111 = __THREW__; __THREW__ = 0;
   $112 = $111&1;
   if ($112) {
    $113 = ___cxa_find_matching_catch_3(0|0)|0;
    $114 = tempRet0;
    ___clang_call_terminate($113);
    // unreachable;
   }
   ;HEAP32[$__sn>>2]=HEAP32[$7>>2]|0;HEAP32[$__sn+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$__sn+8>>2]=HEAP32[$7+8>>2]|0;
   $__i$0$i$i$i21 = 0;
   while(1) {
    $exitcond$i$i$i22 = ($__i$0$i$i$i21|0)==(3);
    if ($exitcond$i$i$i22) {
     break;
    }
    $115 = (($7) + ($__i$0$i$i$i21<<2)|0);
    HEAP32[$115>>2] = 0;
    $116 = (($__i$0$i$i$i21) + 1)|0;
    $__i$0$i$i$i21 = $116;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
   $$pre$phi39Z2D = $96;
  } else {
   $117 = ((($97)) + 40|0);
   $118 = HEAP32[$117>>2]|0;
   FUNCTION_TABLE_vii[$118 & 8191]($8,$96);
   $119 = HEAP32[$8>>2]|0;
   HEAP8[$__pat>>0]=$119&255;HEAP8[$__pat+1>>0]=($119>>8)&255;HEAP8[$__pat+2>>0]=($119>>16)&255;HEAP8[$__pat+3>>0]=$119>>24;
   $120 = HEAP32[$96>>2]|0;
   $121 = ((($120)) + 28|0);
   $122 = HEAP32[$121>>2]|0;
   FUNCTION_TABLE_vii[$122 & 8191]($9,$96);
   $123 = HEAP8[$__sn>>0]|0;
   $124 = $123 & 1;
   $125 = ($124<<24>>24)==(0);
   if ($125) {
    $129 = ((($__sn)) + 1|0);
    HEAP8[$129>>0] = 0;
    HEAP8[$__sn>>0] = 0;
   } else {
    $126 = ((($__sn)) + 8|0);
    $127 = HEAP32[$126>>2]|0;
    HEAP8[$127>>0] = 0;
    $128 = ((($__sn)) + 4|0);
    HEAP32[$128>>2] = 0;
   }
   __THREW__ = 0;
   invoke_vii(3368,($__sn|0),0);
   $130 = __THREW__; __THREW__ = 0;
   $131 = $130&1;
   if ($131) {
    $132 = ___cxa_find_matching_catch_3(0|0)|0;
    $133 = tempRet0;
    ___clang_call_terminate($132);
    // unreachable;
   }
   ;HEAP32[$__sn>>2]=HEAP32[$9>>2]|0;HEAP32[$__sn+4>>2]=HEAP32[$9+4>>2]|0;HEAP32[$__sn+8>>2]=HEAP32[$9+8>>2]|0;
   $__i$0$i$i$i26 = 0;
   while(1) {
    $exitcond$i$i$i27 = ($__i$0$i$i$i26|0)==(3);
    if ($exitcond$i$i$i27) {
     break;
    }
    $134 = (($9) + ($__i$0$i$i$i26<<2)|0);
    HEAP32[$134>>2] = 0;
    $135 = (($__i$0$i$i$i26) + 1)|0;
    $__i$0$i$i$i26 = $135;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
   $$pre$phi39Z2D = $96;
  }
  $136 = HEAP32[$96>>2]|0;
  $137 = ((($136)) + 12|0);
  $138 = HEAP32[$137>>2]|0;
  $139 = (FUNCTION_TABLE_ii[$138 & 8191]($96)|0);
  HEAP8[$__dp>>0] = $139;
  $140 = HEAP32[$96>>2]|0;
  $141 = ((($140)) + 16|0);
  $142 = HEAP32[$141>>2]|0;
  $143 = (FUNCTION_TABLE_ii[$142 & 8191]($96)|0);
  HEAP8[$__ts>>0] = $143;
  $144 = HEAP32[$$pre$phi39Z2D>>2]|0;
  $145 = ((($144)) + 20|0);
  $146 = HEAP32[$145>>2]|0;
  FUNCTION_TABLE_vii[$146 & 8191]($10,$96);
  $147 = HEAP8[$__grp>>0]|0;
  $148 = $147 & 1;
  $149 = ($148<<24>>24)==(0);
  if ($149) {
   $153 = ((($__grp)) + 1|0);
   HEAP8[$153>>0] = 0;
   HEAP8[$__grp>>0] = 0;
  } else {
   $150 = ((($__grp)) + 8|0);
   $151 = HEAP32[$150>>2]|0;
   HEAP8[$151>>0] = 0;
   $152 = ((($__grp)) + 4|0);
   HEAP32[$152>>2] = 0;
  }
  __THREW__ = 0;
  invoke_vii(3368,($__grp|0),0);
  $154 = __THREW__; __THREW__ = 0;
  $155 = $154&1;
  if ($155) {
   $156 = ___cxa_find_matching_catch_3(0|0)|0;
   $157 = tempRet0;
   ___clang_call_terminate($156);
   // unreachable;
  }
  ;HEAP32[$__grp>>2]=HEAP32[$10>>2]|0;HEAP32[$__grp+4>>2]=HEAP32[$10+4>>2]|0;HEAP32[$__grp+8>>2]=HEAP32[$10+8>>2]|0;
  $__i$0$i$i$i31 = 0;
  while(1) {
   $exitcond$i$i$i32 = ($__i$0$i$i$i31|0)==(3);
   if ($exitcond$i$i$i32) {
    break;
   }
   $158 = (($10) + ($__i$0$i$i$i31<<2)|0);
   HEAP32[$158>>2] = 0;
   $159 = (($__i$0$i$i$i31) + 1)|0;
   $__i$0$i$i$i31 = $159;
  }
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
  $160 = HEAP32[$$pre$phi39Z2D>>2]|0;
  $161 = ((($160)) + 24|0);
  $162 = HEAP32[$161>>2]|0;
  FUNCTION_TABLE_vii[$162 & 8191]($11,$96);
  $163 = HEAP8[$__sym>>0]|0;
  $164 = $163 & 1;
  $165 = ($164<<24>>24)==(0);
  if ($165) {
   $169 = ((($__sym)) + 1|0);
   HEAP8[$169>>0] = 0;
   HEAP8[$__sym>>0] = 0;
  } else {
   $166 = ((($__sym)) + 8|0);
   $167 = HEAP32[$166>>2]|0;
   HEAP8[$167>>0] = 0;
   $168 = ((($__sym)) + 4|0);
   HEAP32[$168>>2] = 0;
  }
  __THREW__ = 0;
  invoke_vii(3368,($__sym|0),0);
  $170 = __THREW__; __THREW__ = 0;
  $171 = $170&1;
  if ($171) {
   $172 = ___cxa_find_matching_catch_3(0|0)|0;
   $173 = tempRet0;
   ___clang_call_terminate($172);
   // unreachable;
  }
  ;HEAP32[$__sym>>2]=HEAP32[$11>>2]|0;HEAP32[$__sym+4>>2]=HEAP32[$11+4>>2]|0;HEAP32[$__sym+8>>2]=HEAP32[$11+8>>2]|0;
  $__i$0$i$i$i36 = 0;
  while(1) {
   $exitcond$i$i$i37 = ($__i$0$i$i$i36|0)==(3);
   if ($exitcond$i$i$i37) {
    break;
   }
   $174 = (($11) + ($__i$0$i$i$i36<<2)|0);
   HEAP32[$174>>2] = 0;
   $175 = (($__i$0$i$i$i36) + 1)|0;
   $__i$0$i$i$i36 = $175;
  }
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($11);
  $176 = HEAP32[$96>>2]|0;
  $177 = ((($176)) + 36|0);
  $178 = HEAP32[$177>>2]|0;
  $179 = (FUNCTION_TABLE_ii[$178 & 8191]($96)|0);
  $storemerge = $179;
 }
 HEAP32[$__fd>>2] = $storemerge;
 STACKTOP = sp;return;
}
function __ZNSt3__111__money_putIcE8__formatEPcRS2_S3_jPKcS5_RKNS_5ctypeIcEEbRKNS_10money_base7patternEccRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEESL_SL_i($__mb,$__mi,$__me,$__flags,$__db,$__de,$__ct,$__neg,$__pat,$__dp,$__ts,$__grp,$__sym,$__sn,$__fd) {
 $__mb = $__mb|0;
 $__mi = $__mi|0;
 $__me = $__me|0;
 $__flags = $__flags|0;
 $__db = $__db|0;
 $__de = $__de|0;
 $__ct = $__ct|0;
 $__neg = $__neg|0;
 $__pat = $__pat|0;
 $__dp = $__dp|0;
 $__ts = $__ts|0;
 $__grp = $__grp|0;
 $__sym = $__sym|0;
 $__sn = $__sn|0;
 $__fd = $__fd|0;
 var $$ = 0, $$$0 = 0, $$0 = 0, $$0$i$i = 0, $$0$i$i$lcssa = 0, $$0$i$i2 = 0, $$0$i$i9 = 0, $$0$i$i9$lcssa = 0, $$01$i$i = 0, $$2 = 0, $$lcssa36 = 0, $$pn = 0, $$sroa$04$0$i = 0, $$sroa$04$0$i8 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $__d$0 = 0, $__d$0$lcssa = 0, $__d$1 = 0, $__d$2 = 0, $__d$3 = 0, $__f$0 = 0, $__f$1 = 0, $__gl$0 = 0, $__gl$1 = 0, $__ig$0 = 0, $__ig$1 = 0, $__ng$0 = 0, $__ng$1 = 0, $__p$0 = 0, $exitcond = 0, $isascii1$i = 0, $or$cond = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$__me>>2] = $__mb;
 $0 = ((($__sn)) + 4|0);
 $1 = ((($__sn)) + 8|0);
 $2 = ((($__sn)) + 1|0);
 $3 = ((($__sym)) + 4|0);
 $4 = $__flags & 512;
 $5 = ($4|0)==(0);
 $6 = ((($__sym)) + 8|0);
 $7 = ((($__sym)) + 1|0);
 $8 = ((($__ct)) + 8|0);
 $9 = ($__fd|0)>(0);
 $10 = ((($__grp)) + 4|0);
 $11 = ((($__grp)) + 8|0);
 $12 = ((($__grp)) + 1|0);
 $$0 = $__db;$__p$0 = 0;
 while(1) {
  $exitcond = ($__p$0|0)==(4);
  if ($exitcond) {
   break;
  }
  $21 = (($__pat) + ($__p$0)|0);
  $22 = HEAP8[$21>>0]|0;
  $23 = $22 << 24 >> 24;
  L4: do {
   switch ($23|0) {
   case 0:  {
    $24 = HEAP32[$__me>>2]|0;
    HEAP32[$__mi>>2] = $24;
    $$2 = $$0;
    break;
   }
   case 1:  {
    $25 = HEAP32[$__me>>2]|0;
    HEAP32[$__mi>>2] = $25;
    $26 = HEAP32[$__ct>>2]|0;
    $27 = ((($26)) + 28|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = (FUNCTION_TABLE_iii[$28 & 8191]($__ct,32)|0);
    $30 = HEAP32[$__me>>2]|0;
    $31 = ((($30)) + 1|0);
    HEAP32[$__me>>2] = $31;
    HEAP8[$30>>0] = $29;
    $$2 = $$0;
    break;
   }
   case 3:  {
    $32 = HEAP8[$__sn>>0]|0;
    $33 = $32 & 1;
    $34 = ($33<<24>>24)==(0);
    $35 = HEAP32[$0>>2]|0;
    $36 = $32&255;
    $37 = $36 >>> 1;
    $38 = $34 ? $37 : $35;
    $39 = ($38|0)==(0);
    if ($39) {
     $$2 = $$0;
    } else {
     $40 = HEAP32[$1>>2]|0;
     $41 = $34 ? $2 : $40;
     $42 = HEAP8[$41>>0]|0;
     $43 = HEAP32[$__me>>2]|0;
     $44 = ((($43)) + 1|0);
     HEAP32[$__me>>2] = $44;
     HEAP8[$43>>0] = $42;
     $$2 = $$0;
    }
    break;
   }
   case 2:  {
    $45 = HEAP8[$__sym>>0]|0;
    $46 = $45 & 1;
    $47 = ($46<<24>>24)==(0);
    $48 = HEAP32[$3>>2]|0;
    $49 = $45&255;
    $50 = $49 >>> 1;
    $51 = $47 ? $50 : $48;
    $52 = ($51|0)==(0);
    $or$cond = $5 | $52;
    if ($or$cond) {
     $$2 = $$0;
    } else {
     $53 = HEAP32[$6>>2]|0;
     $54 = $47 ? $7 : $53;
     $55 = (($54) + ($51)|0);
     $56 = HEAP32[$__me>>2]|0;
     $$0$i$i9 = $56;$$sroa$04$0$i8 = $54;
     while(1) {
      $57 = ($$sroa$04$0$i8|0)==($55|0);
      if ($57) {
       $$0$i$i9$lcssa = $$0$i$i9;
       break;
      }
      $58 = HEAP8[$$sroa$04$0$i8>>0]|0;
      HEAP8[$$0$i$i9>>0] = $58;
      $59 = ((($$sroa$04$0$i8)) + 1|0);
      $60 = ((($$0$i$i9)) + 1|0);
      $$0$i$i9 = $60;$$sroa$04$0$i8 = $59;
     }
     HEAP32[$__me>>2] = $$0$i$i9$lcssa;
     $$2 = $$0;
    }
    break;
   }
   case 4:  {
    $61 = HEAP32[$__me>>2]|0;
    $62 = ((($$0)) + 1|0);
    $$$0 = $__neg ? $62 : $$0;
    $__d$0 = $$$0;
    while(1) {
     $63 = ($__d$0>>>0)<($__de>>>0);
     if (!($63)) {
      $__d$0$lcssa = $__d$0;
      break;
     }
     $64 = HEAP8[$__d$0>>0]|0;
     $isascii1$i = ($64<<24>>24)>(-1);
     if (!($isascii1$i)) {
      $__d$0$lcssa = $__d$0;
      break;
     }
     $65 = $64 << 24 >> 24;
     $66 = HEAP32[$8>>2]|0;
     $67 = (($66) + ($65<<1)|0);
     $68 = HEAP16[$67>>1]|0;
     $69 = $68 & 2048;
     $70 = ($69<<16>>16)==(0);
     if ($70) {
      $__d$0$lcssa = $__d$0;
      break;
     }
     $71 = ((($__d$0)) + 1|0);
     $__d$0 = $71;
    }
    if ($9) {
     $__d$1 = $__d$0$lcssa;$__f$0 = $__fd;
     while(1) {
      $72 = ($__d$1>>>0)>($$$0>>>0);
      $73 = ($__f$0|0)>(0);
      $74 = $72 & $73;
      if (!($74)) {
       break;
      }
      $75 = ((($__d$1)) + -1|0);
      $76 = HEAP8[$75>>0]|0;
      $77 = HEAP32[$__me>>2]|0;
      $78 = ((($77)) + 1|0);
      HEAP32[$__me>>2] = $78;
      HEAP8[$77>>0] = $76;
      $79 = (($__f$0) + -1)|0;
      $__d$1 = $75;$__f$0 = $79;
     }
     if ($73) {
      $80 = HEAP32[$__ct>>2]|0;
      $81 = ((($80)) + 28|0);
      $82 = HEAP32[$81>>2]|0;
      $83 = (FUNCTION_TABLE_iii[$82 & 8191]($__ct,48)|0);
      $87 = $83;
     } else {
      $87 = 0;
     }
     $__f$1 = $__f$0;
     while(1) {
      $84 = ($__f$1|0)>(0);
      $85 = HEAP32[$__me>>2]|0;
      $86 = ((($85)) + 1|0);
      HEAP32[$__me>>2] = $86;
      if (!($84)) {
       $$lcssa36 = $85;
       break;
      }
      HEAP8[$85>>0] = $87;
      $88 = (($__f$1) + -1)|0;
      $__f$1 = $88;
     }
     HEAP8[$$lcssa36>>0] = $__dp;
     $__d$2 = $__d$1;
    } else {
     $__d$2 = $__d$0$lcssa;
    }
    $89 = ($__d$2|0)==($$$0|0);
    L35: do {
     if ($89) {
      $90 = HEAP32[$__ct>>2]|0;
      $91 = ((($90)) + 28|0);
      $92 = HEAP32[$91>>2]|0;
      $93 = (FUNCTION_TABLE_iii[$92 & 8191]($__ct,48)|0);
      $94 = HEAP32[$__me>>2]|0;
      $95 = ((($94)) + 1|0);
      HEAP32[$__me>>2] = $95;
      HEAP8[$94>>0] = $93;
     } else {
      $96 = HEAP8[$__grp>>0]|0;
      $97 = $96 & 1;
      $98 = ($97<<24>>24)==(0);
      $99 = HEAP32[$10>>2]|0;
      $100 = $96&255;
      $101 = $100 >>> 1;
      $102 = $98 ? $101 : $99;
      $103 = ($102|0)==(0);
      if ($103) {
       $__d$3 = $__d$2;$__gl$0 = -1;$__ig$0 = 0;$__ng$0 = 0;
      } else {
       $104 = HEAP32[$11>>2]|0;
       $105 = $98 ? $12 : $104;
       $106 = HEAP8[$105>>0]|0;
       $107 = $106 << 24 >> 24;
       $__d$3 = $__d$2;$__gl$0 = $107;$__ig$0 = 0;$__ng$0 = 0;
      }
      while(1) {
       $108 = ($__d$3|0)==($$$0|0);
       if ($108) {
        break L35;
       }
       $109 = ($__ng$0|0)==($__gl$0|0);
       if ($109) {
        $110 = HEAP32[$__me>>2]|0;
        $111 = ((($110)) + 1|0);
        HEAP32[$__me>>2] = $111;
        HEAP8[$110>>0] = $__ts;
        $112 = (($__ig$0) + 1)|0;
        $113 = HEAP8[$__grp>>0]|0;
        $114 = $113 & 1;
        $115 = ($114<<24>>24)==(0);
        $116 = HEAP32[$10>>2]|0;
        $117 = $113&255;
        $118 = $117 >>> 1;
        $119 = $115 ? $118 : $116;
        $120 = ($112>>>0)<($119>>>0);
        if ($120) {
         $121 = HEAP32[$11>>2]|0;
         $122 = $115 ? $12 : $121;
         $123 = (($122) + ($112)|0);
         $124 = HEAP8[$123>>0]|0;
         $125 = ($124<<24>>24)==(127);
         $126 = $124 << 24 >> 24;
         $$ = $125 ? -1 : $126;
         $__gl$1 = $$;$__ig$1 = $112;$__ng$1 = 0;
        } else {
         $__gl$1 = $__ng$0;$__ig$1 = $112;$__ng$1 = 0;
        }
       } else {
        $__gl$1 = $__gl$0;$__ig$1 = $__ig$0;$__ng$1 = $__ng$0;
       }
       $127 = ((($__d$3)) + -1|0);
       $128 = HEAP8[$127>>0]|0;
       $129 = HEAP32[$__me>>2]|0;
       $130 = ((($129)) + 1|0);
       HEAP32[$__me>>2] = $130;
       HEAP8[$129>>0] = $128;
       $131 = (($__ng$1) + 1)|0;
       $__d$3 = $127;$__gl$0 = $__gl$1;$__ig$0 = $__ig$1;$__ng$0 = $131;
      }
     }
    } while(0);
    $132 = HEAP32[$__me>>2]|0;
    $133 = ($61|0)==($132|0);
    if ($133) {
     $$2 = $$$0;
    } else {
     $$0$i$i2 = $132;$$01$i$i = $61;
     while(1) {
      $134 = ((($$0$i$i2)) + -1|0);
      $135 = ($$01$i$i>>>0)<($134>>>0);
      if (!($135)) {
       $$2 = $$$0;
       break L4;
      }
      $136 = HEAP8[$$01$i$i>>0]|0;
      $137 = HEAP8[$134>>0]|0;
      HEAP8[$$01$i$i>>0] = $137;
      HEAP8[$134>>0] = $136;
      $138 = ((($$01$i$i)) + 1|0);
      $$0$i$i2 = $134;$$01$i$i = $138;
     }
    }
    break;
   }
   default: {
    $$2 = $$0;
   }
   }
  } while(0);
  $139 = (($__p$0) + 1)|0;
  $$0 = $$2;$__p$0 = $139;
 }
 $13 = HEAP8[$__sn>>0]|0;
 $14 = $13 & 1;
 $15 = ($14<<24>>24)==(0);
 $16 = HEAP32[$0>>2]|0;
 $17 = $13&255;
 $18 = $17 >>> 1;
 $19 = $15 ? $18 : $16;
 $20 = ($19>>>0)>(1);
 if ($20) {
  $140 = HEAP32[$1>>2]|0;
  $141 = $15 ? $2 : $140;
  $142 = (($141) + ($19)|0);
  $143 = HEAP32[$__me>>2]|0;
  $$0$i$i = $143;$$pn = $141;
  while(1) {
   $$sroa$04$0$i = ((($$pn)) + 1|0);
   $144 = ($$sroa$04$0$i|0)==($142|0);
   if ($144) {
    $$0$i$i$lcssa = $$0$i$i;
    break;
   }
   $145 = HEAP8[$$sroa$04$0$i>>0]|0;
   HEAP8[$$0$i$i>>0] = $145;
   $146 = ((($$0$i$i)) + 1|0);
   $$0$i$i = $146;$$pn = $$sroa$04$0$i;
  }
  HEAP32[$__me>>2] = $$0$i$i$lcssa;
 }
 $147 = $__flags & 176;
 switch ($147|0) {
 case 32:  {
  $148 = HEAP32[$__me>>2]|0;
  HEAP32[$__mi>>2] = $148;
  break;
 }
 case 16:  {
  break;
 }
 default: {
  HEAP32[$__mi>>2] = $__mb;
 }
 }
 return;
}
function __ZNKSt3__19money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_bRNS_8ios_baseEcRKNS_12basic_stringIcS3_NS_9allocatorIcEEEE($this,$__s,$__intl,$__iob,$__fl,$__digits) {
 $this = $this|0;
 $__s = $__s|0;
 $__intl = $__intl|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $__digits = $__digits|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$byval_copy = 0, $$pre = 0, $$pre18 = 0, $$pre19 = 0, $$pre19$pre = 0, $$pre21 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $127 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__dp = 0, $__fd = 0;
 var $__grp = 0, $__h$sroa$0$0 = 0, $__h$sroa$0$1 = 0, $__i$0$i$i = 0, $__i$0$i$i3 = 0, $__i$0$i$i6 = 0, $__loc = 0, $__mb$0 = 0, $__mbuf = 0, $__me = 0, $__mi = 0, $__pat = 0, $__sn = 0, $__sym = 0, $__ts = 0, $exitcond$i$i = 0, $exitcond$i$i4 = 0, $exitcond$i$i7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 56|0;
 $__loc = sp + 52|0;
 $__pat = sp + 164|0;
 $__dp = sp + 161|0;
 $__ts = sp + 160|0;
 $__grp = sp + 40|0;
 $__sym = sp + 28|0;
 $__sn = sp + 16|0;
 $__fd = sp + 12|0;
 $__mbuf = sp + 60|0;
 $__mi = sp + 8|0;
 $__me = sp + 4|0;
 $0 = sp;
 $1 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$__loc>>2] = $1;
 __THREW__ = 0;
 $2 = (invoke_iii(3374,($__loc|0),(1207576|0))|0);
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 do {
  if ($4) {
   label = 22;
  } else {
   $5 = HEAP8[$__digits>>0]|0;
   $6 = $5 & 1;
   $7 = ($6<<24>>24)==(0);
   $8 = ((($__digits)) + 4|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = $5&255;
   $11 = $10 >>> 1;
   $12 = $7 ? $11 : $9;
   $13 = ($12|0)==(0);
   if ($13) {
    $32 = 0;
   } else {
    $14 = ((($__digits)) + 8|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = ((($__digits)) + 1|0);
    $17 = $7 ? $16 : $15;
    $18 = HEAP8[$17>>0]|0;
    $19 = HEAP32[$2>>2]|0;
    $20 = ((($19)) + 28|0);
    $21 = HEAP32[$20>>2]|0;
    __THREW__ = 0;
    $22 = (invoke_iii($21|0,($2|0),45)|0);
    $23 = __THREW__; __THREW__ = 0;
    $24 = $23&1;
    if ($24) {
     label = 22;
     break;
    }
    $25 = ($18<<24>>24)==($22<<24>>24);
    $32 = $25;
   }
   $__i$0$i$i = 0;
   while(1) {
    $exitcond$i$i = ($__i$0$i$i|0)==(3);
    if ($exitcond$i$i) {
     break;
    }
    $26 = (($__grp) + ($__i$0$i$i<<2)|0);
    HEAP32[$26>>2] = 0;
    $27 = (($__i$0$i$i) + 1)|0;
    $__i$0$i$i = $27;
   }
   $__i$0$i$i3 = 0;
   while(1) {
    $exitcond$i$i4 = ($__i$0$i$i3|0)==(3);
    if ($exitcond$i$i4) {
     break;
    }
    $28 = (($__sym) + ($__i$0$i$i3<<2)|0);
    HEAP32[$28>>2] = 0;
    $29 = (($__i$0$i$i3) + 1)|0;
    $__i$0$i$i3 = $29;
   }
   $__i$0$i$i6 = 0;
   while(1) {
    $exitcond$i$i7 = ($__i$0$i$i6|0)==(3);
    if ($exitcond$i$i7) {
     break;
    }
    $30 = (($__sn) + ($__i$0$i$i6<<2)|0);
    HEAP32[$30>>2] = 0;
    $31 = (($__i$0$i$i6) + 1)|0;
    $__i$0$i$i6 = $31;
   }
   __THREW__ = 0;
   invoke_viiiiiiiiii(4912,($__intl|0),($32|0),($__loc|0),($__pat|0),($__dp|0),($__ts|0),($__grp|0),($__sym|0),($__sn|0),($__fd|0));
   $33 = __THREW__; __THREW__ = 0;
   $34 = $33&1;
   if ($34) {
    $94 = ___cxa_find_matching_catch_2()|0;
    $95 = tempRet0;
    $$0 = $94;$$01 = $95;$$pre19 = $1;
   } else {
    $35 = HEAP8[$__digits>>0]|0;
    $36 = $35 & 1;
    $37 = ($36<<24>>24)==(0);
    $38 = HEAP32[$8>>2]|0;
    $39 = $35&255;
    $40 = $39 >>> 1;
    $41 = $37 ? $40 : $38;
    $42 = HEAP32[$__fd>>2]|0;
    $43 = ($41|0)>($42|0);
    if ($43) {
     $44 = (($41) - ($42))|0;
     $45 = $44 << 1;
     $46 = HEAP8[$__sn>>0]|0;
     $47 = $46 & 1;
     $48 = ($47<<24>>24)==(0);
     $49 = ((($__sn)) + 4|0);
     $50 = HEAP32[$49>>2]|0;
     $51 = $46&255;
     $52 = $51 >>> 1;
     $53 = $48 ? $52 : $50;
     $54 = HEAP8[$__sym>>0]|0;
     $55 = $54 & 1;
     $56 = ($55<<24>>24)==(0);
     $57 = ((($__sym)) + 4|0);
     $58 = HEAP32[$57>>2]|0;
     $59 = $54&255;
     $60 = $59 >>> 1;
     $61 = $56 ? $60 : $58;
     $62 = $45 | 1;
     $63 = (($62) + ($42))|0;
     $64 = (($63) + ($53))|0;
     $65 = (($64) + ($61))|0;
     $86 = $65;
    } else {
     $66 = HEAP8[$__sn>>0]|0;
     $67 = $66 & 1;
     $68 = ($67<<24>>24)==(0);
     $69 = ((($__sn)) + 4|0);
     $70 = HEAP32[$69>>2]|0;
     $71 = $66&255;
     $72 = $71 >>> 1;
     $73 = $68 ? $72 : $70;
     $74 = HEAP8[$__sym>>0]|0;
     $75 = $74 & 1;
     $76 = ($75<<24>>24)==(0);
     $77 = ((($__sym)) + 4|0);
     $78 = HEAP32[$77>>2]|0;
     $79 = $74&255;
     $80 = $79 >>> 1;
     $81 = $76 ? $80 : $78;
     $82 = (($42) + 2)|0;
     $83 = (($82) + ($73))|0;
     $84 = (($83) + ($81))|0;
     $86 = $84;
    }
    $85 = ($86>>>0)>(100);
    if ($85) {
     $87 = (_malloc($86)|0);
     $88 = $87;
     $89 = ($87|0)==(0|0);
     if ($89) {
      __THREW__ = 0;
      invoke_v(4863);
      $90 = __THREW__; __THREW__ = 0;
      $91 = $90&1;
      if ($91) {
       $__h$sroa$0$0 = $88;
      } else {
       $$pre = HEAP8[$__digits>>0]|0;
       $$pre18 = HEAP32[$8>>2]|0;
       $103 = $$pre;$112 = $$pre18;$__h$sroa$0$1 = $88;$__mb$0 = 0;
       label = 26;
      }
     } else {
      $103 = $35;$112 = $38;$__h$sroa$0$1 = $88;$__mb$0 = $87;
      label = 26;
     }
    } else {
     $103 = $35;$112 = $38;$__h$sroa$0$1 = 0;$__mb$0 = $__mbuf;
     label = 26;
    }
    if ((label|0) == 26) {
     $100 = ((($__iob)) + 4|0);
     $101 = HEAP32[$100>>2]|0;
     $102 = $103 & 1;
     $104 = ($102<<24>>24)==(0);
     $105 = ((($__digits)) + 8|0);
     $106 = HEAP32[$105>>2]|0;
     $107 = ((($__digits)) + 1|0);
     $108 = $104 ? $107 : $106;
     $109 = $103&255;
     $110 = $109 >>> 1;
     $111 = $104 ? $110 : $112;
     $113 = (($108) + ($111)|0);
     $114 = HEAP8[$__dp>>0]|0;
     $115 = HEAP8[$__ts>>0]|0;
     __THREW__ = 0;
     invoke_viiiiiiiiiiiiiii(4913,($__mb$0|0),($__mi|0),($__me|0),($101|0),($108|0),($113|0),($2|0),($32|0),($__pat|0),($114|0),($115|0),($__grp|0),($__sym|0),($__sn|0),($42|0));
     $116 = __THREW__; __THREW__ = 0;
     $117 = $116&1;
     if ($117) {
      $__h$sroa$0$0 = $__h$sroa$0$1;
     } else {
      $118 = HEAP32[$__s>>2]|0;
      HEAP32[$0>>2] = $118;
      $119 = HEAP32[$__mi>>2]|0;
      $120 = HEAP32[$__me>>2]|0;
      __THREW__ = 0;
      ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
      $121 = (invoke_iiiiiii(3375,($$byval_copy|0),($__mb$0|0),($119|0),($120|0),($__iob|0),($__fl|0))|0);
      $122 = __THREW__; __THREW__ = 0;
      $123 = $122&1;
      if ($123) {
       $__h$sroa$0$0 = $__h$sroa$0$1;
      } else {
       $124 = ($__h$sroa$0$1|0)==(0);
       if ($124) {
        $126 = $1;
       } else {
        $125 = $__h$sroa$0$1;
        _free($125);
        $$pre21 = HEAP32[$__loc>>2]|0;
        $126 = $$pre21;
       }
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__sn);
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__sym);
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grp);
       (__ZNSt3__114__shared_count16__release_sharedEv($126)|0);
       STACKTOP = sp;return ($121|0);
      }
     }
    }
    $96 = ___cxa_find_matching_catch_2()|0;
    $97 = tempRet0;
    $98 = ($__h$sroa$0$0|0)==(0);
    if ($98) {
     $$0 = $96;$$01 = $97;$$pre19 = $1;
    } else {
     $99 = $__h$sroa$0$0;
     _free($99);
     $$pre19$pre = HEAP32[$__loc>>2]|0;
     $$0 = $96;$$01 = $97;$$pre19 = $$pre19$pre;
    }
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__sn);
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__sym);
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grp);
   $$1 = $$0;$$12 = $$01;$127 = $$pre19;
  }
 } while(0);
 if ((label|0) == 22) {
  $92 = ___cxa_find_matching_catch_2()|0;
  $93 = tempRet0;
  $$1 = $92;$$12 = $93;$127 = $1;
 }
 (__ZNSt3__114__shared_count16__release_sharedEv($127)|0);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__19money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__19money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__19money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_bRNS_8ios_baseEwe($this,$__s,$__intl,$__iob,$__fl,$__units) {
 $this = $this|0;
 $__s = $__s|0;
 $__intl = $__intl|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $__units = +$__units;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$byval_copy = 0, $$pre = 0, $$pre20 = 0, $$pre20$pre = 0, $$pre22 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__bb = 0, $__buf = 0, $__db$0 = 0, $__digits = 0, $__dp = 0, $__fd = 0, $__grp = 0, $__hd$sroa$0$0 = 0, $__hd$sroa$0$1 = 0, $__hd$sroa$0$2 = 0, $__hn$sroa$0$0 = 0, $__hn$sroa$0$1 = 0, $__hn$sroa$0$2 = 0, $__hw$sroa$0$0 = 0, $__hw$sroa$0$1 = 0, $__i$0$i$i = 0;
 var $__i$0$i$i4 = 0, $__i$0$i$i6 = 0, $__loc = 0, $__mb$0 = 0, $__mbuf = 0, $__me = 0, $__mi = 0, $__n$0 = 0, $__pat = 0, $__sn = 0, $__sym = 0, $__ts = 0, $exitcond$i$i = 0, $exitcond$i$i5 = 0, $exitcond$i$i7 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1008|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 892|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $__buf = sp + 900|0;
 $__bb = sp + 888|0;
 $__digits = sp + 488|0;
 $__loc = sp + 480|0;
 $__pat = sp + 896|0;
 $__dp = sp + 476|0;
 $__ts = sp + 472|0;
 $__grp = sp + 460|0;
 $__sym = sp + 448|0;
 $__sn = sp + 436|0;
 $__fd = sp + 432|0;
 $__mbuf = sp + 32|0;
 $__mi = sp + 24|0;
 $__me = sp + 20|0;
 $0 = sp + 16|0;
 HEAP32[$__bb>>2] = $__buf;
 HEAPF64[$vararg_buffer>>3] = $__units;
 $1 = (_snprintf($__buf,100,622023,$vararg_buffer)|0);
 $2 = ($1>>>0)>(99);
 do {
  if ($2) {
   __THREW__ = 0;
   $3 = (invoke_i(4874)|0);
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   if ($5) {
    $__hd$sroa$0$0 = 0;$__hn$sroa$0$0 = 0;
    label = 8;
   } else {
    __THREW__ = 0;
    HEAPF64[$vararg_buffer1>>3] = $__units;
    $6 = (invoke_iiiii(4879,($__bb|0),($3|0),(622023|0),($vararg_buffer1|0))|0);
    $7 = __THREW__; __THREW__ = 0;
    $8 = $7&1;
    if ($8) {
     $__hd$sroa$0$0 = 0;$__hn$sroa$0$0 = 0;
     label = 8;
    } else {
     $9 = HEAP32[$__bb>>2]|0;
     $10 = ($9|0)==(0|0);
     if ($10) {
      __THREW__ = 0;
      invoke_v(4863);
      $12 = __THREW__; __THREW__ = 0;
      $13 = $12&1;
      if ($13) {
       $__hd$sroa$0$0 = 0;$__hn$sroa$0$0 = 0;
       label = 8;
       break;
      }
      $$pre = HEAP32[$__bb>>2]|0;
      $132 = $$pre;
     } else {
      $11 = $9;
      $132 = $11;
     }
     $16 = $6 << 2;
     $17 = (_malloc($16)|0);
     $18 = $17;
     $19 = ($17|0)==(0|0);
     if ($19) {
      __THREW__ = 0;
      invoke_v(4863);
      $20 = __THREW__; __THREW__ = 0;
      $21 = $20&1;
      if ($21) {
       $__hd$sroa$0$0 = $18;$__hn$sroa$0$0 = $132;
       label = 8;
      } else {
       $__db$0 = 0;$__hd$sroa$0$1 = $18;$__hn$sroa$0$1 = $132;$__n$0 = $6;
       label = 11;
      }
     } else {
      $__db$0 = $17;$__hd$sroa$0$1 = $18;$__hn$sroa$0$1 = $132;$__n$0 = $6;
      label = 11;
     }
    }
   }
  } else {
   $__db$0 = $__digits;$__hd$sroa$0$1 = 0;$__hn$sroa$0$1 = 0;$__n$0 = $1;
   label = 11;
  }
 } while(0);
 if ((label|0) == 11) {
  __THREW__ = 0;
  $22 = (invoke_ii(3373,($__iob|0))|0);
  $23 = __THREW__; __THREW__ = 0;
  $24 = $23&1;
  if ($24) {
   $__hd$sroa$0$0 = $__hd$sroa$0$1;$__hn$sroa$0$0 = $__hn$sroa$0$1;
   label = 8;
  } else {
   HEAP32[$__loc>>2] = $22;
   __THREW__ = 0;
   $25 = (invoke_iii(3374,($__loc|0),(1207608|0))|0);
   $26 = __THREW__; __THREW__ = 0;
   $27 = $26&1;
   if ($27) {
    label = 32;
   } else {
    $28 = HEAP32[$__bb>>2]|0;
    $29 = (($28) + ($__n$0)|0);
    $30 = HEAP32[$25>>2]|0;
    $31 = ((($30)) + 48|0);
    $32 = HEAP32[$31>>2]|0;
    __THREW__ = 0;
    (invoke_iiiii($32|0,($25|0),($28|0),($29|0),($__db$0|0))|0);
    $33 = __THREW__; __THREW__ = 0;
    $34 = $33&1;
    if ($34) {
     label = 32;
    } else {
     $35 = ($__n$0|0)==(0);
     if ($35) {
      $45 = 0;
     } else {
      $36 = HEAP32[$__bb>>2]|0;
      $37 = HEAP8[$36>>0]|0;
      $38 = ($37<<24>>24)==(45);
      $45 = $38;
     }
     $__i$0$i$i = 0;
     while(1) {
      $exitcond$i$i = ($__i$0$i$i|0)==(3);
      if ($exitcond$i$i) {
       break;
      }
      $39 = (($__grp) + ($__i$0$i$i<<2)|0);
      HEAP32[$39>>2] = 0;
      $40 = (($__i$0$i$i) + 1)|0;
      $__i$0$i$i = $40;
     }
     $__i$0$i$i4 = 0;
     while(1) {
      $exitcond$i$i5 = ($__i$0$i$i4|0)==(3);
      if ($exitcond$i$i5) {
       break;
      }
      $41 = (($__sym) + ($__i$0$i$i4<<2)|0);
      HEAP32[$41>>2] = 0;
      $42 = (($__i$0$i$i4) + 1)|0;
      $__i$0$i$i4 = $42;
     }
     $__i$0$i$i6 = 0;
     while(1) {
      $exitcond$i$i7 = ($__i$0$i$i6|0)==(3);
      if ($exitcond$i$i7) {
       break;
      }
      $43 = (($__sn) + ($__i$0$i$i6<<2)|0);
      HEAP32[$43>>2] = 0;
      $44 = (($__i$0$i$i6) + 1)|0;
      $__i$0$i$i6 = $44;
     }
     __THREW__ = 0;
     invoke_viiiiiiiiii(4914,($__intl|0),($45|0),($__loc|0),($__pat|0),($__dp|0),($__ts|0),($__grp|0),($__sym|0),($__sn|0),($__fd|0));
     $46 = __THREW__; __THREW__ = 0;
     $47 = $46&1;
     if ($47) {
      $101 = ___cxa_find_matching_catch_2()|0;
      $102 = tempRet0;
      $$0 = $101;$$01 = $102;$$pre20 = $22;
     } else {
      $48 = HEAP32[$__fd>>2]|0;
      $49 = ($__n$0|0)>($48|0);
      if ($49) {
       $50 = (($__n$0) - ($48))|0;
       $51 = $50 << 1;
       $52 = HEAP8[$__sn>>0]|0;
       $53 = $52 & 1;
       $54 = ($53<<24>>24)==(0);
       $55 = ((($__sn)) + 4|0);
       $56 = HEAP32[$55>>2]|0;
       $57 = $52&255;
       $58 = $57 >>> 1;
       $59 = $54 ? $58 : $56;
       $60 = HEAP8[$__sym>>0]|0;
       $61 = $60 & 1;
       $62 = ($61<<24>>24)==(0);
       $63 = ((($__sym)) + 4|0);
       $64 = HEAP32[$63>>2]|0;
       $65 = $60&255;
       $66 = $65 >>> 1;
       $67 = $62 ? $66 : $64;
       $68 = $51 | 1;
       $69 = (($68) + ($48))|0;
       $70 = (($69) + ($59))|0;
       $71 = (($70) + ($67))|0;
       $92 = $71;
      } else {
       $72 = HEAP8[$__sn>>0]|0;
       $73 = $72 & 1;
       $74 = ($73<<24>>24)==(0);
       $75 = ((($__sn)) + 4|0);
       $76 = HEAP32[$75>>2]|0;
       $77 = $72&255;
       $78 = $77 >>> 1;
       $79 = $74 ? $78 : $76;
       $80 = HEAP8[$__sym>>0]|0;
       $81 = $80 & 1;
       $82 = ($81<<24>>24)==(0);
       $83 = ((($__sym)) + 4|0);
       $84 = HEAP32[$83>>2]|0;
       $85 = $80&255;
       $86 = $85 >>> 1;
       $87 = $82 ? $86 : $84;
       $88 = (($48) + 2)|0;
       $89 = (($88) + ($79))|0;
       $90 = (($89) + ($87))|0;
       $92 = $90;
      }
      $91 = ($92>>>0)>(100);
      if ($91) {
       $93 = $92 << 2;
       $94 = (_malloc($93)|0);
       $95 = $94;
       $96 = ($94|0)==(0|0);
       if ($96) {
        __THREW__ = 0;
        invoke_v(4863);
        $97 = __THREW__; __THREW__ = 0;
        $98 = $97&1;
        if ($98) {
         $__hw$sroa$0$0 = $95;
        } else {
         $__hw$sroa$0$1 = $95;$__mb$0 = 0;
         label = 36;
        }
       } else {
        $__hw$sroa$0$1 = $95;$__mb$0 = $94;
        label = 36;
       }
      } else {
       $__hw$sroa$0$1 = 0;$__mb$0 = $__mbuf;
       label = 36;
      }
      if ((label|0) == 36) {
       $107 = ((($__iob)) + 4|0);
       $108 = HEAP32[$107>>2]|0;
       $109 = (($__db$0) + ($__n$0<<2)|0);
       $110 = HEAP32[$__dp>>2]|0;
       $111 = HEAP32[$__ts>>2]|0;
       __THREW__ = 0;
       invoke_viiiiiiiiiiiiiii(4915,($__mb$0|0),($__mi|0),($__me|0),($108|0),($__db$0|0),($109|0),($25|0),($45|0),($__pat|0),($110|0),($111|0),($__grp|0),($__sym|0),($__sn|0),($48|0));
       $112 = __THREW__; __THREW__ = 0;
       $113 = $112&1;
       if ($113) {
        $__hw$sroa$0$0 = $__hw$sroa$0$1;
       } else {
        $114 = HEAP32[$__s>>2]|0;
        HEAP32[$0>>2] = $114;
        $115 = HEAP32[$__mi>>2]|0;
        $116 = HEAP32[$__me>>2]|0;
        __THREW__ = 0;
        ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
        $117 = (invoke_iiiiiii(4886,($$byval_copy|0),($__mb$0|0),($115|0),($116|0),($__iob|0),($__fl|0))|0);
        $118 = __THREW__; __THREW__ = 0;
        $119 = $118&1;
        if ($119) {
         $__hw$sroa$0$0 = $__hw$sroa$0$1;
        } else {
         $120 = ($__hw$sroa$0$1|0)==(0);
         if ($120) {
          $122 = $22;
         } else {
          $121 = $__hw$sroa$0$1;
          _free($121);
          $$pre22 = HEAP32[$__loc>>2]|0;
          $122 = $$pre22;
         }
         __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__sn);
         __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__sym);
         __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grp);
         (__ZNSt3__114__shared_count16__release_sharedEv($122)|0);
         $123 = ($__hd$sroa$0$1|0)==(0);
         if (!($123)) {
          $124 = $__hd$sroa$0$1;
          _free($124);
         }
         $125 = ($__hn$sroa$0$1|0)==(0);
         if (!($125)) {
          $126 = $__hn$sroa$0$1;
          _free($126);
         }
         STACKTOP = sp;return ($117|0);
        }
       }
      }
      $103 = ___cxa_find_matching_catch_2()|0;
      $104 = tempRet0;
      $105 = ($__hw$sroa$0$0|0)==(0);
      if ($105) {
       $$0 = $103;$$01 = $104;$$pre20 = $22;
      } else {
       $106 = $__hw$sroa$0$0;
       _free($106);
       $$pre20$pre = HEAP32[$__loc>>2]|0;
       $$0 = $103;$$01 = $104;$$pre20 = $$pre20$pre;
      }
     }
     __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__sn);
     __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__sym);
     __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grp);
     $$1 = $$0;$$12 = $$01;$127 = $$pre20;
    }
   }
   if ((label|0) == 32) {
    $99 = ___cxa_find_matching_catch_2()|0;
    $100 = tempRet0;
    $$1 = $99;$$12 = $100;$127 = $22;
   }
   (__ZNSt3__114__shared_count16__release_sharedEv($127)|0);
   $$2 = $$1;$$23 = $$12;$__hd$sroa$0$2 = $__hd$sroa$0$1;$__hn$sroa$0$2 = $__hn$sroa$0$1;
  }
 }
 if ((label|0) == 8) {
  $14 = ___cxa_find_matching_catch_2()|0;
  $15 = tempRet0;
  $$2 = $14;$$23 = $15;$__hd$sroa$0$2 = $__hd$sroa$0$0;$__hn$sroa$0$2 = $__hn$sroa$0$0;
 }
 $128 = ($__hd$sroa$0$2|0)==(0);
 if (!($128)) {
  $129 = $__hd$sroa$0$2;
  _free($129);
 }
 $130 = ($__hn$sroa$0$2|0)==(0);
 if (!($130)) {
  $131 = $__hn$sroa$0$2;
  _free($131);
 }
 ___resumeException($$2|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__111__money_putIwE13__gather_infoEbbRKNS_6localeERNS_10money_base7patternERwS8_RNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERNS9_IwNSA_IwEENSC_IwEEEESJ_Ri($__intl,$__neg,$__loc,$__pat,$__dp,$__ts,$__grp,$__sym,$__sn,$__fd) {
 $__intl = $__intl|0;
 $__neg = $__neg|0;
 $__loc = $__loc|0;
 $__pat = $__pat|0;
 $__dp = $__dp|0;
 $__ts = $__ts|0;
 $__grp = $__grp|0;
 $__sym = $__sym|0;
 $__sn = $__sn|0;
 $__fd = $__fd|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__i$0$i$i$i = 0, $__i$0$i$i$i13 = 0, $__i$0$i$i$i18 = 0, $__i$0$i$i$i23 = 0, $__i$0$i$i$i28 = 0, $__i$0$i$i$i33 = 0;
 var $__i$0$i$i$i6 = 0, $__i$0$i$i$i9 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i10 = 0, $exitcond$i$i$i14 = 0, $exitcond$i$i$i19 = 0, $exitcond$i$i$i24 = 0, $exitcond$i$i$i29 = 0, $exitcond$i$i$i34 = 0, $exitcond$i$i$i7 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 108|0;
 $1 = sp + 96|0;
 $2 = sp + 92|0;
 $3 = sp + 80|0;
 $4 = sp + 68|0;
 $5 = sp + 56|0;
 $6 = sp + 52|0;
 $7 = sp + 40|0;
 $8 = sp + 36|0;
 $9 = sp + 24|0;
 $10 = sp + 12|0;
 $11 = sp;
 if ($__intl) {
  $12 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1209312)|0);
  $13 = HEAP32[$12>>2]|0;
  if ($__neg) {
   $14 = ((($13)) + 44|0);
   $15 = HEAP32[$14>>2]|0;
   FUNCTION_TABLE_vii[$15 & 8191]($0,$12);
   $16 = HEAP32[$0>>2]|0;
   HEAP8[$__pat>>0]=$16&255;HEAP8[$__pat+1>>0]=($16>>8)&255;HEAP8[$__pat+2>>0]=($16>>16)&255;HEAP8[$__pat+3>>0]=$16>>24;
   $17 = HEAP32[$12>>2]|0;
   $18 = ((($17)) + 32|0);
   $19 = HEAP32[$18>>2]|0;
   FUNCTION_TABLE_vii[$19 & 8191]($1,$12);
   $20 = HEAP8[$__sn>>0]|0;
   $21 = $20 & 1;
   $22 = ($21<<24>>24)==(0);
   if ($22) {
    HEAP8[$__sn>>0] = 0;
   } else {
    $23 = ((($__sn)) + 8|0);
    $24 = HEAP32[$23>>2]|0;
    HEAP32[$24>>2] = 0;
   }
   $25 = ((($__sn)) + 4|0);
   HEAP32[$25>>2] = 0;
   __THREW__ = 0;
   invoke_vii(4910,($__sn|0),0);
   $26 = __THREW__; __THREW__ = 0;
   $27 = $26&1;
   if ($27) {
    $28 = ___cxa_find_matching_catch_3(0|0)|0;
    $29 = tempRet0;
    ___clang_call_terminate($28);
    // unreachable;
   }
   ;HEAP32[$__sn>>2]=HEAP32[$1>>2]|0;HEAP32[$__sn+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$__sn+8>>2]=HEAP32[$1+8>>2]|0;
   $__i$0$i$i$i = 0;
   while(1) {
    $exitcond$i$i$i = ($__i$0$i$i$i|0)==(3);
    if ($exitcond$i$i$i) {
     break;
    }
    $30 = (($1) + ($__i$0$i$i$i<<2)|0);
    HEAP32[$30>>2] = 0;
    $31 = (($__i$0$i$i$i) + 1)|0;
    $__i$0$i$i$i = $31;
   }
   __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($1);
  } else {
   $32 = ((($13)) + 40|0);
   $33 = HEAP32[$32>>2]|0;
   FUNCTION_TABLE_vii[$33 & 8191]($2,$12);
   $34 = HEAP32[$2>>2]|0;
   HEAP8[$__pat>>0]=$34&255;HEAP8[$__pat+1>>0]=($34>>8)&255;HEAP8[$__pat+2>>0]=($34>>16)&255;HEAP8[$__pat+3>>0]=$34>>24;
   $35 = HEAP32[$12>>2]|0;
   $36 = ((($35)) + 28|0);
   $37 = HEAP32[$36>>2]|0;
   FUNCTION_TABLE_vii[$37 & 8191]($3,$12);
   $38 = HEAP8[$__sn>>0]|0;
   $39 = $38 & 1;
   $40 = ($39<<24>>24)==(0);
   if ($40) {
    HEAP8[$__sn>>0] = 0;
   } else {
    $41 = ((($__sn)) + 8|0);
    $42 = HEAP32[$41>>2]|0;
    HEAP32[$42>>2] = 0;
   }
   $43 = ((($__sn)) + 4|0);
   HEAP32[$43>>2] = 0;
   __THREW__ = 0;
   invoke_vii(4910,($__sn|0),0);
   $44 = __THREW__; __THREW__ = 0;
   $45 = $44&1;
   if ($45) {
    $46 = ___cxa_find_matching_catch_3(0|0)|0;
    $47 = tempRet0;
    ___clang_call_terminate($46);
    // unreachable;
   }
   ;HEAP32[$__sn>>2]=HEAP32[$3>>2]|0;HEAP32[$__sn+4>>2]=HEAP32[$3+4>>2]|0;HEAP32[$__sn+8>>2]=HEAP32[$3+8>>2]|0;
   $__i$0$i$i$i6 = 0;
   while(1) {
    $exitcond$i$i$i7 = ($__i$0$i$i$i6|0)==(3);
    if ($exitcond$i$i$i7) {
     break;
    }
    $48 = (($3) + ($__i$0$i$i$i6<<2)|0);
    HEAP32[$48>>2] = 0;
    $49 = (($__i$0$i$i$i6) + 1)|0;
    $__i$0$i$i$i6 = $49;
   }
   __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($3);
  }
  $50 = HEAP32[$12>>2]|0;
  $51 = ((($50)) + 12|0);
  $52 = HEAP32[$51>>2]|0;
  $53 = (FUNCTION_TABLE_ii[$52 & 8191]($12)|0);
  HEAP32[$__dp>>2] = $53;
  $54 = HEAP32[$12>>2]|0;
  $55 = ((($54)) + 16|0);
  $56 = HEAP32[$55>>2]|0;
  $57 = (FUNCTION_TABLE_ii[$56 & 8191]($12)|0);
  HEAP32[$__ts>>2] = $57;
  $58 = HEAP32[$12>>2]|0;
  $59 = ((($58)) + 20|0);
  $60 = HEAP32[$59>>2]|0;
  FUNCTION_TABLE_vii[$60 & 8191]($4,$12);
  $61 = HEAP8[$__grp>>0]|0;
  $62 = $61 & 1;
  $63 = ($62<<24>>24)==(0);
  if ($63) {
   $67 = ((($__grp)) + 1|0);
   HEAP8[$67>>0] = 0;
   HEAP8[$__grp>>0] = 0;
  } else {
   $64 = ((($__grp)) + 8|0);
   $65 = HEAP32[$64>>2]|0;
   HEAP8[$65>>0] = 0;
   $66 = ((($__grp)) + 4|0);
   HEAP32[$66>>2] = 0;
  }
  __THREW__ = 0;
  invoke_vii(3368,($__grp|0),0);
  $68 = __THREW__; __THREW__ = 0;
  $69 = $68&1;
  if ($69) {
   $70 = ___cxa_find_matching_catch_3(0|0)|0;
   $71 = tempRet0;
   ___clang_call_terminate($70);
   // unreachable;
  }
  ;HEAP32[$__grp>>2]=HEAP32[$4>>2]|0;HEAP32[$__grp+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$__grp+8>>2]=HEAP32[$4+8>>2]|0;
  $__i$0$i$i$i9 = 0;
  while(1) {
   $exitcond$i$i$i10 = ($__i$0$i$i$i9|0)==(3);
   if ($exitcond$i$i$i10) {
    break;
   }
   $72 = (($4) + ($__i$0$i$i$i9<<2)|0);
   HEAP32[$72>>2] = 0;
   $73 = (($__i$0$i$i$i9) + 1)|0;
   $__i$0$i$i$i9 = $73;
  }
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4);
  $74 = HEAP32[$12>>2]|0;
  $75 = ((($74)) + 24|0);
  $76 = HEAP32[$75>>2]|0;
  FUNCTION_TABLE_vii[$76 & 8191]($5,$12);
  $77 = HEAP8[$__sym>>0]|0;
  $78 = $77 & 1;
  $79 = ($78<<24>>24)==(0);
  if ($79) {
   HEAP8[$__sym>>0] = 0;
  } else {
   $80 = ((($__sym)) + 8|0);
   $81 = HEAP32[$80>>2]|0;
   HEAP32[$81>>2] = 0;
  }
  $82 = ((($__sym)) + 4|0);
  HEAP32[$82>>2] = 0;
  __THREW__ = 0;
  invoke_vii(4910,($__sym|0),0);
  $83 = __THREW__; __THREW__ = 0;
  $84 = $83&1;
  if ($84) {
   $85 = ___cxa_find_matching_catch_3(0|0)|0;
   $86 = tempRet0;
   ___clang_call_terminate($85);
   // unreachable;
  }
  ;HEAP32[$__sym>>2]=HEAP32[$5>>2]|0;HEAP32[$__sym+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$__sym+8>>2]=HEAP32[$5+8>>2]|0;
  $__i$0$i$i$i13 = 0;
  while(1) {
   $exitcond$i$i$i14 = ($__i$0$i$i$i13|0)==(3);
   if ($exitcond$i$i$i14) {
    break;
   }
   $87 = (($5) + ($__i$0$i$i$i13<<2)|0);
   HEAP32[$87>>2] = 0;
   $88 = (($__i$0$i$i$i13) + 1)|0;
   $__i$0$i$i$i13 = $88;
  }
  __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($5);
  $89 = HEAP32[$12>>2]|0;
  $90 = ((($89)) + 36|0);
  $91 = HEAP32[$90>>2]|0;
  $92 = (FUNCTION_TABLE_ii[$91 & 8191]($12)|0);
  $storemerge = $92;
 } else {
  $93 = (__ZNKSt3__16locale9use_facetERNS0_2idE($__loc,1209304)|0);
  $94 = HEAP32[$93>>2]|0;
  if ($__neg) {
   $95 = ((($94)) + 44|0);
   $96 = HEAP32[$95>>2]|0;
   FUNCTION_TABLE_vii[$96 & 8191]($6,$93);
   $97 = HEAP32[$6>>2]|0;
   HEAP8[$__pat>>0]=$97&255;HEAP8[$__pat+1>>0]=($97>>8)&255;HEAP8[$__pat+2>>0]=($97>>16)&255;HEAP8[$__pat+3>>0]=$97>>24;
   $98 = HEAP32[$93>>2]|0;
   $99 = ((($98)) + 32|0);
   $100 = HEAP32[$99>>2]|0;
   FUNCTION_TABLE_vii[$100 & 8191]($7,$93);
   $101 = HEAP8[$__sn>>0]|0;
   $102 = $101 & 1;
   $103 = ($102<<24>>24)==(0);
   if ($103) {
    HEAP8[$__sn>>0] = 0;
   } else {
    $104 = ((($__sn)) + 8|0);
    $105 = HEAP32[$104>>2]|0;
    HEAP32[$105>>2] = 0;
   }
   $106 = ((($__sn)) + 4|0);
   HEAP32[$106>>2] = 0;
   __THREW__ = 0;
   invoke_vii(4910,($__sn|0),0);
   $107 = __THREW__; __THREW__ = 0;
   $108 = $107&1;
   if ($108) {
    $109 = ___cxa_find_matching_catch_3(0|0)|0;
    $110 = tempRet0;
    ___clang_call_terminate($109);
    // unreachable;
   }
   ;HEAP32[$__sn>>2]=HEAP32[$7>>2]|0;HEAP32[$__sn+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$__sn+8>>2]=HEAP32[$7+8>>2]|0;
   $__i$0$i$i$i18 = 0;
   while(1) {
    $exitcond$i$i$i19 = ($__i$0$i$i$i18|0)==(3);
    if ($exitcond$i$i$i19) {
     break;
    }
    $111 = (($7) + ($__i$0$i$i$i18<<2)|0);
    HEAP32[$111>>2] = 0;
    $112 = (($__i$0$i$i$i18) + 1)|0;
    $__i$0$i$i$i18 = $112;
   }
   __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($7);
  } else {
   $113 = ((($94)) + 40|0);
   $114 = HEAP32[$113>>2]|0;
   FUNCTION_TABLE_vii[$114 & 8191]($8,$93);
   $115 = HEAP32[$8>>2]|0;
   HEAP8[$__pat>>0]=$115&255;HEAP8[$__pat+1>>0]=($115>>8)&255;HEAP8[$__pat+2>>0]=($115>>16)&255;HEAP8[$__pat+3>>0]=$115>>24;
   $116 = HEAP32[$93>>2]|0;
   $117 = ((($116)) + 28|0);
   $118 = HEAP32[$117>>2]|0;
   FUNCTION_TABLE_vii[$118 & 8191]($9,$93);
   $119 = HEAP8[$__sn>>0]|0;
   $120 = $119 & 1;
   $121 = ($120<<24>>24)==(0);
   if ($121) {
    HEAP8[$__sn>>0] = 0;
   } else {
    $122 = ((($__sn)) + 8|0);
    $123 = HEAP32[$122>>2]|0;
    HEAP32[$123>>2] = 0;
   }
   $124 = ((($__sn)) + 4|0);
   HEAP32[$124>>2] = 0;
   __THREW__ = 0;
   invoke_vii(4910,($__sn|0),0);
   $125 = __THREW__; __THREW__ = 0;
   $126 = $125&1;
   if ($126) {
    $127 = ___cxa_find_matching_catch_3(0|0)|0;
    $128 = tempRet0;
    ___clang_call_terminate($127);
    // unreachable;
   }
   ;HEAP32[$__sn>>2]=HEAP32[$9>>2]|0;HEAP32[$__sn+4>>2]=HEAP32[$9+4>>2]|0;HEAP32[$__sn+8>>2]=HEAP32[$9+8>>2]|0;
   $__i$0$i$i$i23 = 0;
   while(1) {
    $exitcond$i$i$i24 = ($__i$0$i$i$i23|0)==(3);
    if ($exitcond$i$i$i24) {
     break;
    }
    $129 = (($9) + ($__i$0$i$i$i23<<2)|0);
    HEAP32[$129>>2] = 0;
    $130 = (($__i$0$i$i$i23) + 1)|0;
    $__i$0$i$i$i23 = $130;
   }
   __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($9);
  }
  $131 = HEAP32[$93>>2]|0;
  $132 = ((($131)) + 12|0);
  $133 = HEAP32[$132>>2]|0;
  $134 = (FUNCTION_TABLE_ii[$133 & 8191]($93)|0);
  HEAP32[$__dp>>2] = $134;
  $135 = HEAP32[$93>>2]|0;
  $136 = ((($135)) + 16|0);
  $137 = HEAP32[$136>>2]|0;
  $138 = (FUNCTION_TABLE_ii[$137 & 8191]($93)|0);
  HEAP32[$__ts>>2] = $138;
  $139 = HEAP32[$93>>2]|0;
  $140 = ((($139)) + 20|0);
  $141 = HEAP32[$140>>2]|0;
  FUNCTION_TABLE_vii[$141 & 8191]($10,$93);
  $142 = HEAP8[$__grp>>0]|0;
  $143 = $142 & 1;
  $144 = ($143<<24>>24)==(0);
  if ($144) {
   $148 = ((($__grp)) + 1|0);
   HEAP8[$148>>0] = 0;
   HEAP8[$__grp>>0] = 0;
  } else {
   $145 = ((($__grp)) + 8|0);
   $146 = HEAP32[$145>>2]|0;
   HEAP8[$146>>0] = 0;
   $147 = ((($__grp)) + 4|0);
   HEAP32[$147>>2] = 0;
  }
  __THREW__ = 0;
  invoke_vii(3368,($__grp|0),0);
  $149 = __THREW__; __THREW__ = 0;
  $150 = $149&1;
  if ($150) {
   $151 = ___cxa_find_matching_catch_3(0|0)|0;
   $152 = tempRet0;
   ___clang_call_terminate($151);
   // unreachable;
  }
  ;HEAP32[$__grp>>2]=HEAP32[$10>>2]|0;HEAP32[$__grp+4>>2]=HEAP32[$10+4>>2]|0;HEAP32[$__grp+8>>2]=HEAP32[$10+8>>2]|0;
  $__i$0$i$i$i28 = 0;
  while(1) {
   $exitcond$i$i$i29 = ($__i$0$i$i$i28|0)==(3);
   if ($exitcond$i$i$i29) {
    break;
   }
   $153 = (($10) + ($__i$0$i$i$i28<<2)|0);
   HEAP32[$153>>2] = 0;
   $154 = (($__i$0$i$i$i28) + 1)|0;
   $__i$0$i$i$i28 = $154;
  }
  __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
  $155 = HEAP32[$93>>2]|0;
  $156 = ((($155)) + 24|0);
  $157 = HEAP32[$156>>2]|0;
  FUNCTION_TABLE_vii[$157 & 8191]($11,$93);
  $158 = HEAP8[$__sym>>0]|0;
  $159 = $158 & 1;
  $160 = ($159<<24>>24)==(0);
  if ($160) {
   HEAP8[$__sym>>0] = 0;
  } else {
   $161 = ((($__sym)) + 8|0);
   $162 = HEAP32[$161>>2]|0;
   HEAP32[$162>>2] = 0;
  }
  $163 = ((($__sym)) + 4|0);
  HEAP32[$163>>2] = 0;
  __THREW__ = 0;
  invoke_vii(4910,($__sym|0),0);
  $164 = __THREW__; __THREW__ = 0;
  $165 = $164&1;
  if ($165) {
   $166 = ___cxa_find_matching_catch_3(0|0)|0;
   $167 = tempRet0;
   ___clang_call_terminate($166);
   // unreachable;
  }
  ;HEAP32[$__sym>>2]=HEAP32[$11>>2]|0;HEAP32[$__sym+4>>2]=HEAP32[$11+4>>2]|0;HEAP32[$__sym+8>>2]=HEAP32[$11+8>>2]|0;
  $__i$0$i$i$i33 = 0;
  while(1) {
   $exitcond$i$i$i34 = ($__i$0$i$i$i33|0)==(3);
   if ($exitcond$i$i$i34) {
    break;
   }
   $168 = (($11) + ($__i$0$i$i$i33<<2)|0);
   HEAP32[$168>>2] = 0;
   $169 = (($__i$0$i$i$i33) + 1)|0;
   $__i$0$i$i$i33 = $169;
  }
  __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($11);
  $170 = HEAP32[$93>>2]|0;
  $171 = ((($170)) + 36|0);
  $172 = HEAP32[$171>>2]|0;
  $173 = (FUNCTION_TABLE_ii[$172 & 8191]($93)|0);
  $storemerge = $173;
 }
 HEAP32[$__fd>>2] = $storemerge;
 STACKTOP = sp;return;
}
function __ZNSt3__111__money_putIwE8__formatEPwRS2_S3_jPKwS5_RKNS_5ctypeIwEEbRKNS_10money_base7patternEwwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERKNSE_IwNSF_IwEENSH_IwEEEESQ_i($__mb,$__mi,$__me,$__flags,$__db,$__de,$__ct,$__neg,$__pat,$__dp,$__ts,$__grp,$__sym,$__sn,$__fd) {
 $__mb = $__mb|0;
 $__mi = $__mi|0;
 $__me = $__me|0;
 $__flags = $__flags|0;
 $__db = $__db|0;
 $__de = $__de|0;
 $__ct = $__ct|0;
 $__neg = $__neg|0;
 $__pat = $__pat|0;
 $__dp = $__dp|0;
 $__ts = $__ts|0;
 $__grp = $__grp|0;
 $__sym = $__sym|0;
 $__sn = $__sn|0;
 $__fd = $__fd|0;
 var $$ = 0, $$$0 = 0, $$0 = 0, $$0$i$i = 0, $$0$i$i2 = 0, $$0$i$i6 = 0, $$01$i$i = 0, $$2 = 0, $$lcssa34 = 0, $$lcssa35 = 0, $$pre27 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $__d$0 = 0, $__d$0$lcssa = 0, $__d$1 = 0, $__d$2 = 0, $__d$3 = 0, $__f$0 = 0, $__f$1 = 0, $__gl$0 = 0, $__gl$1 = 0, $__ig$0 = 0, $__ig$1 = 0, $__me$promoted = 0, $__ng$0 = 0, $__ng$1 = 0, $__p$0 = 0, $exitcond = 0, $or$cond = 0, $scevgep$i = 0;
 var $scevgep$i7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$__me>>2] = $__mb;
 $0 = ((($__sn)) + 4|0);
 $1 = ((($__sn)) + 8|0);
 $2 = ((($__sym)) + 4|0);
 $3 = $__flags & 512;
 $4 = ($3|0)==(0);
 $5 = ((($__sym)) + 8|0);
 $6 = ($__fd|0)>(0);
 $7 = ((($__grp)) + 4|0);
 $8 = ((($__grp)) + 8|0);
 $9 = ((($__grp)) + 1|0);
 $$0 = $__db;$__p$0 = 0;
 while(1) {
  $exitcond = ($__p$0|0)==(4);
  if ($exitcond) {
   break;
  }
  $18 = (($__pat) + ($__p$0)|0);
  $19 = HEAP8[$18>>0]|0;
  $20 = $19 << 24 >> 24;
  L4: do {
   switch ($20|0) {
   case 0:  {
    $21 = HEAP32[$__me>>2]|0;
    HEAP32[$__mi>>2] = $21;
    $$2 = $$0;
    break;
   }
   case 1:  {
    $22 = HEAP32[$__me>>2]|0;
    HEAP32[$__mi>>2] = $22;
    $23 = HEAP32[$__ct>>2]|0;
    $24 = ((($23)) + 44|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = (FUNCTION_TABLE_iii[$25 & 8191]($__ct,32)|0);
    $27 = HEAP32[$__me>>2]|0;
    $28 = ((($27)) + 4|0);
    HEAP32[$__me>>2] = $28;
    HEAP32[$27>>2] = $26;
    $$2 = $$0;
    break;
   }
   case 3:  {
    $29 = HEAP8[$__sn>>0]|0;
    $30 = $29 & 1;
    $31 = ($30<<24>>24)==(0);
    $32 = HEAP32[$0>>2]|0;
    $33 = $29&255;
    $34 = $33 >>> 1;
    $35 = $31 ? $34 : $32;
    $36 = ($35|0)==(0);
    if ($36) {
     $$2 = $$0;
    } else {
     $37 = HEAP32[$1>>2]|0;
     $38 = $31 ? $0 : $37;
     $39 = HEAP32[$38>>2]|0;
     $40 = HEAP32[$__me>>2]|0;
     $41 = ((($40)) + 4|0);
     HEAP32[$__me>>2] = $41;
     HEAP32[$40>>2] = $39;
     $$2 = $$0;
    }
    break;
   }
   case 2:  {
    $42 = HEAP8[$__sym>>0]|0;
    $43 = $42 & 1;
    $44 = ($43<<24>>24)==(0);
    $45 = HEAP32[$2>>2]|0;
    $46 = $42&255;
    $47 = $46 >>> 1;
    $48 = $44 ? $47 : $45;
    $49 = ($48|0)==(0);
    $or$cond = $4 | $49;
    if ($or$cond) {
     $$2 = $$0;
    } else {
     $50 = HEAP32[$5>>2]|0;
     $51 = $44 ? $2 : $50;
     $52 = (($51) + ($48<<2)|0);
     $53 = HEAP32[$__me>>2]|0;
     $$0$i$i6 = $53;$55 = $51;
     while(1) {
      $54 = ($55|0)==($52|0);
      if ($54) {
       break;
      }
      $56 = HEAP32[$55>>2]|0;
      HEAP32[$$0$i$i6>>2] = $56;
      $57 = ((($55)) + 4|0);
      $58 = ((($$0$i$i6)) + 4|0);
      $$0$i$i6 = $58;$55 = $57;
     }
     $scevgep$i7 = (($53) + ($48<<2)|0);
     HEAP32[$__me>>2] = $scevgep$i7;
     $$2 = $$0;
    }
    break;
   }
   case 4:  {
    $59 = HEAP32[$__me>>2]|0;
    $60 = ((($$0)) + 4|0);
    $$$0 = $__neg ? $60 : $$0;
    $__d$0 = $$$0;
    while(1) {
     $61 = ($__d$0>>>0)<($__de>>>0);
     if (!($61)) {
      $__d$0$lcssa = $__d$0;
      break;
     }
     $62 = HEAP32[$__d$0>>2]|0;
     $63 = HEAP32[$__ct>>2]|0;
     $64 = ((($63)) + 12|0);
     $65 = HEAP32[$64>>2]|0;
     $66 = (FUNCTION_TABLE_iiii[$65 & 8191]($__ct,2048,$62)|0);
     if (!($66)) {
      $__d$0$lcssa = $__d$0;
      break;
     }
     $67 = ((($__d$0)) + 4|0);
     $__d$0 = $67;
    }
    if ($6) {
     $__d$1 = $__d$0$lcssa;$__f$0 = $__fd;
     while(1) {
      $68 = ($__d$1>>>0)>($$$0>>>0);
      $69 = ($__f$0|0)>(0);
      $70 = $68 & $69;
      if (!($70)) {
       break;
      }
      $71 = ((($__d$1)) + -4|0);
      $72 = HEAP32[$71>>2]|0;
      $73 = HEAP32[$__me>>2]|0;
      $74 = ((($73)) + 4|0);
      HEAP32[$__me>>2] = $74;
      HEAP32[$73>>2] = $72;
      $75 = (($__f$0) + -1)|0;
      $__d$1 = $71;$__f$0 = $75;
     }
     if ($69) {
      $76 = HEAP32[$__ct>>2]|0;
      $77 = ((($76)) + 44|0);
      $78 = HEAP32[$77>>2]|0;
      $79 = (FUNCTION_TABLE_iii[$78 & 8191]($__ct,48)|0);
      $83 = $79;
     } else {
      $83 = 0;
     }
     $__me$promoted = HEAP32[$__me>>2]|0;
     $82 = $__me$promoted;$__f$1 = $__f$0;
     while(1) {
      $80 = ($__f$1|0)>(0);
      $81 = ((($82)) + 4|0);
      if (!($80)) {
       $$lcssa34 = $82;$$lcssa35 = $81;
       break;
      }
      HEAP32[$82>>2] = $83;
      $84 = (($__f$1) + -1)|0;
      $82 = $81;$__f$1 = $84;
     }
     HEAP32[$__me>>2] = $$lcssa35;
     HEAP32[$$lcssa34>>2] = $__dp;
     $__d$2 = $__d$1;
    } else {
     $__d$2 = $__d$0$lcssa;
    }
    $85 = ($__d$2|0)==($$$0|0);
    if ($85) {
     $86 = HEAP32[$__ct>>2]|0;
     $87 = ((($86)) + 44|0);
     $88 = HEAP32[$87>>2]|0;
     $89 = (FUNCTION_TABLE_iii[$88 & 8191]($__ct,48)|0);
     $90 = HEAP32[$__me>>2]|0;
     $91 = ((($90)) + 4|0);
     HEAP32[$__me>>2] = $91;
     HEAP32[$90>>2] = $89;
     $128 = $91;
    } else {
     $92 = HEAP8[$__grp>>0]|0;
     $93 = $92 & 1;
     $94 = ($93<<24>>24)==(0);
     $95 = HEAP32[$7>>2]|0;
     $96 = $92&255;
     $97 = $96 >>> 1;
     $98 = $94 ? $97 : $95;
     $99 = ($98|0)==(0);
     if ($99) {
      $__d$3 = $__d$2;$__gl$0 = -1;$__ig$0 = 0;$__ng$0 = 0;
     } else {
      $100 = HEAP32[$8>>2]|0;
      $101 = $94 ? $9 : $100;
      $102 = HEAP8[$101>>0]|0;
      $103 = $102 << 24 >> 24;
      $__d$3 = $__d$2;$__gl$0 = $103;$__ig$0 = 0;$__ng$0 = 0;
     }
     while(1) {
      $104 = ($__d$3|0)==($$$0|0);
      if ($104) {
       break;
      }
      $105 = ($__ng$0|0)==($__gl$0|0);
      $106 = HEAP32[$__me>>2]|0;
      if ($105) {
       $107 = ((($106)) + 4|0);
       HEAP32[$__me>>2] = $107;
       HEAP32[$106>>2] = $__ts;
       $108 = (($__ig$0) + 1)|0;
       $109 = HEAP8[$__grp>>0]|0;
       $110 = $109 & 1;
       $111 = ($110<<24>>24)==(0);
       $112 = $109&255;
       $113 = $112 >>> 1;
       $114 = $111 ? $113 : $95;
       $115 = ($108>>>0)<($114>>>0);
       if ($115) {
        $116 = HEAP32[$8>>2]|0;
        $117 = $111 ? $9 : $116;
        $118 = (($117) + ($108)|0);
        $119 = HEAP8[$118>>0]|0;
        $120 = ($119<<24>>24)==(127);
        $121 = $119 << 24 >> 24;
        $$ = $120 ? -1 : $121;
        $125 = $107;$__gl$1 = $$;$__ig$1 = $108;$__ng$1 = 0;
       } else {
        $125 = $107;$__gl$1 = $__ng$0;$__ig$1 = $108;$__ng$1 = 0;
       }
      } else {
       $125 = $106;$__gl$1 = $__gl$0;$__ig$1 = $__ig$0;$__ng$1 = $__ng$0;
      }
      $122 = ((($__d$3)) + -4|0);
      $123 = HEAP32[$122>>2]|0;
      $124 = ((($125)) + 4|0);
      HEAP32[$__me>>2] = $124;
      HEAP32[$125>>2] = $123;
      $126 = (($__ng$1) + 1)|0;
      $__d$3 = $122;$__gl$0 = $__gl$1;$__ig$0 = $__ig$1;$__ng$0 = $126;
     }
     $$pre27 = HEAP32[$__me>>2]|0;
     $128 = $$pre27;
    }
    $127 = ($59|0)==($128|0);
    if ($127) {
     $$2 = $$$0;
    } else {
     $$0$i$i2 = $128;$$01$i$i = $59;
     while(1) {
      $129 = ((($$0$i$i2)) + -4|0);
      $130 = ($$01$i$i>>>0)<($129>>>0);
      if (!($130)) {
       $$2 = $$$0;
       break L4;
      }
      $131 = HEAP32[$$01$i$i>>2]|0;
      $132 = HEAP32[$129>>2]|0;
      HEAP32[$$01$i$i>>2] = $132;
      HEAP32[$129>>2] = $131;
      $133 = ((($$01$i$i)) + 4|0);
      $$0$i$i2 = $129;$$01$i$i = $133;
     }
    }
    break;
   }
   default: {
    $$2 = $$0;
   }
   }
  } while(0);
  $134 = (($__p$0) + 1)|0;
  $$0 = $$2;$__p$0 = $134;
 }
 $10 = HEAP8[$__sn>>0]|0;
 $11 = $10 & 1;
 $12 = ($11<<24>>24)==(0);
 $13 = HEAP32[$0>>2]|0;
 $14 = $10&255;
 $15 = $14 >>> 1;
 $16 = $12 ? $15 : $13;
 $17 = ($16>>>0)>(1);
 if ($17) {
  $135 = HEAP32[$1>>2]|0;
  $136 = $12 ? $0 : $135;
  $137 = ((($136)) + 4|0);
  $138 = (($136) + ($16<<2)|0);
  $139 = HEAP32[$__me>>2]|0;
  $140 = $137;
  $141 = $138;
  $142 = (($141) - ($140))|0;
  $$0$i$i = $139;$144 = $137;
  while(1) {
   $143 = ($144|0)==($138|0);
   if ($143) {
    break;
   }
   $145 = HEAP32[$144>>2]|0;
   HEAP32[$$0$i$i>>2] = $145;
   $146 = ((($144)) + 4|0);
   $147 = ((($$0$i$i)) + 4|0);
   $$0$i$i = $147;$144 = $146;
  }
  $148 = $142 >>> 2;
  $scevgep$i = (($139) + ($148<<2)|0);
  HEAP32[$__me>>2] = $scevgep$i;
 }
 $149 = $__flags & 176;
 switch ($149|0) {
 case 32:  {
  $150 = HEAP32[$__me>>2]|0;
  HEAP32[$__mi>>2] = $150;
  break;
 }
 case 16:  {
  break;
 }
 default: {
  HEAP32[$__mi>>2] = $__mb;
 }
 }
 return;
}
function __ZNKSt3__19money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_bRNS_8ios_baseEwRKNS_12basic_stringIwS3_NS_9allocatorIwEEEE($this,$__s,$__intl,$__iob,$__fl,$__digits) {
 $this = $this|0;
 $__s = $__s|0;
 $__intl = $__intl|0;
 $__iob = $__iob|0;
 $__fl = $__fl|0;
 $__digits = $__digits|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$byval_copy = 0, $$pre = 0, $$pre17 = 0, $$pre18 = 0, $$pre18$pre = 0, $$pre20 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0;
 var $125 = 0, $126 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__dp = 0, $__fd = 0, $__grp = 0;
 var $__h$sroa$0$0 = 0, $__h$sroa$0$1 = 0, $__i$0$i$i = 0, $__i$0$i$i3 = 0, $__i$0$i$i5 = 0, $__loc = 0, $__mb$0 = 0, $__mbuf = 0, $__me = 0, $__mi = 0, $__pat = 0, $__sn = 0, $__sym = 0, $__ts = 0, $exitcond$i$i = 0, $exitcond$i$i4 = 0, $exitcond$i$i6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 480|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $$byval_copy = sp + 468|0;
 $__loc = sp + 464|0;
 $__pat = sp + 472|0;
 $__dp = sp + 460|0;
 $__ts = sp + 456|0;
 $__grp = sp + 444|0;
 $__sym = sp + 432|0;
 $__sn = sp + 420|0;
 $__fd = sp + 416|0;
 $__mbuf = sp + 16|0;
 $__mi = sp + 8|0;
 $__me = sp + 4|0;
 $0 = sp;
 $1 = (__ZNKSt3__18ios_base6getlocEv($__iob)|0);
 HEAP32[$__loc>>2] = $1;
 __THREW__ = 0;
 $2 = (invoke_iii(3374,($__loc|0),(1207608|0))|0);
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 do {
  if ($4) {
   label = 22;
  } else {
   $5 = HEAP8[$__digits>>0]|0;
   $6 = $5 & 1;
   $7 = ($6<<24>>24)==(0);
   $8 = ((($__digits)) + 4|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = $5&255;
   $11 = $10 >>> 1;
   $12 = $7 ? $11 : $9;
   $13 = ($12|0)==(0);
   if ($13) {
    $31 = 0;
   } else {
    $14 = ((($__digits)) + 8|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = $7 ? $8 : $15;
    $17 = HEAP32[$16>>2]|0;
    $18 = HEAP32[$2>>2]|0;
    $19 = ((($18)) + 44|0);
    $20 = HEAP32[$19>>2]|0;
    __THREW__ = 0;
    $21 = (invoke_iii($20|0,($2|0),45)|0);
    $22 = __THREW__; __THREW__ = 0;
    $23 = $22&1;
    if ($23) {
     label = 22;
     break;
    }
    $24 = ($17|0)==($21|0);
    $31 = $24;
   }
   $__i$0$i$i = 0;
   while(1) {
    $exitcond$i$i = ($__i$0$i$i|0)==(3);
    if ($exitcond$i$i) {
     break;
    }
    $25 = (($__grp) + ($__i$0$i$i<<2)|0);
    HEAP32[$25>>2] = 0;
    $26 = (($__i$0$i$i) + 1)|0;
    $__i$0$i$i = $26;
   }
   $__i$0$i$i3 = 0;
   while(1) {
    $exitcond$i$i4 = ($__i$0$i$i3|0)==(3);
    if ($exitcond$i$i4) {
     break;
    }
    $27 = (($__sym) + ($__i$0$i$i3<<2)|0);
    HEAP32[$27>>2] = 0;
    $28 = (($__i$0$i$i3) + 1)|0;
    $__i$0$i$i3 = $28;
   }
   $__i$0$i$i5 = 0;
   while(1) {
    $exitcond$i$i6 = ($__i$0$i$i5|0)==(3);
    if ($exitcond$i$i6) {
     break;
    }
    $29 = (($__sn) + ($__i$0$i$i5<<2)|0);
    HEAP32[$29>>2] = 0;
    $30 = (($__i$0$i$i5) + 1)|0;
    $__i$0$i$i5 = $30;
   }
   __THREW__ = 0;
   invoke_viiiiiiiiii(4914,($__intl|0),($31|0),($__loc|0),($__pat|0),($__dp|0),($__ts|0),($__grp|0),($__sym|0),($__sn|0),($__fd|0));
   $32 = __THREW__; __THREW__ = 0;
   $33 = $32&1;
   if ($33) {
    $94 = ___cxa_find_matching_catch_2()|0;
    $95 = tempRet0;
    $$0 = $94;$$01 = $95;$$pre18 = $1;
   } else {
    $34 = HEAP8[$__digits>>0]|0;
    $35 = $34 & 1;
    $36 = ($35<<24>>24)==(0);
    $37 = HEAP32[$8>>2]|0;
    $38 = $34&255;
    $39 = $38 >>> 1;
    $40 = $36 ? $39 : $37;
    $41 = HEAP32[$__fd>>2]|0;
    $42 = ($40|0)>($41|0);
    if ($42) {
     $43 = (($40) - ($41))|0;
     $44 = $43 << 1;
     $45 = HEAP8[$__sn>>0]|0;
     $46 = $45 & 1;
     $47 = ($46<<24>>24)==(0);
     $48 = ((($__sn)) + 4|0);
     $49 = HEAP32[$48>>2]|0;
     $50 = $45&255;
     $51 = $50 >>> 1;
     $52 = $47 ? $51 : $49;
     $53 = HEAP8[$__sym>>0]|0;
     $54 = $53 & 1;
     $55 = ($54<<24>>24)==(0);
     $56 = ((($__sym)) + 4|0);
     $57 = HEAP32[$56>>2]|0;
     $58 = $53&255;
     $59 = $58 >>> 1;
     $60 = $55 ? $59 : $57;
     $61 = $44 | 1;
     $62 = (($61) + ($41))|0;
     $63 = (($62) + ($52))|0;
     $64 = (($63) + ($60))|0;
     $85 = $64;
    } else {
     $65 = HEAP8[$__sn>>0]|0;
     $66 = $65 & 1;
     $67 = ($66<<24>>24)==(0);
     $68 = ((($__sn)) + 4|0);
     $69 = HEAP32[$68>>2]|0;
     $70 = $65&255;
     $71 = $70 >>> 1;
     $72 = $67 ? $71 : $69;
     $73 = HEAP8[$__sym>>0]|0;
     $74 = $73 & 1;
     $75 = ($74<<24>>24)==(0);
     $76 = ((($__sym)) + 4|0);
     $77 = HEAP32[$76>>2]|0;
     $78 = $73&255;
     $79 = $78 >>> 1;
     $80 = $75 ? $79 : $77;
     $81 = (($41) + 2)|0;
     $82 = (($81) + ($72))|0;
     $83 = (($82) + ($80))|0;
     $85 = $83;
    }
    $84 = ($85>>>0)>(100);
    if ($84) {
     $86 = $85 << 2;
     $87 = (_malloc($86)|0);
     $88 = $87;
     $89 = ($87|0)==(0|0);
     if ($89) {
      __THREW__ = 0;
      invoke_v(4863);
      $90 = __THREW__; __THREW__ = 0;
      $91 = $90&1;
      if ($91) {
       $__h$sroa$0$0 = $88;
      } else {
       $$pre = HEAP8[$__digits>>0]|0;
       $$pre17 = HEAP32[$8>>2]|0;
       $103 = $$pre;$111 = $$pre17;$__h$sroa$0$1 = $88;$__mb$0 = 0;
       label = 26;
      }
     } else {
      $103 = $34;$111 = $37;$__h$sroa$0$1 = $88;$__mb$0 = $87;
      label = 26;
     }
    } else {
     $103 = $34;$111 = $37;$__h$sroa$0$1 = 0;$__mb$0 = $__mbuf;
     label = 26;
    }
    if ((label|0) == 26) {
     $100 = ((($__iob)) + 4|0);
     $101 = HEAP32[$100>>2]|0;
     $102 = $103 & 1;
     $104 = ($102<<24>>24)==(0);
     $105 = ((($__digits)) + 8|0);
     $106 = HEAP32[$105>>2]|0;
     $107 = $104 ? $8 : $106;
     $108 = $103&255;
     $109 = $108 >>> 1;
     $110 = $104 ? $109 : $111;
     $112 = (($107) + ($110<<2)|0);
     $113 = HEAP32[$__dp>>2]|0;
     $114 = HEAP32[$__ts>>2]|0;
     __THREW__ = 0;
     invoke_viiiiiiiiiiiiiii(4915,($__mb$0|0),($__mi|0),($__me|0),($101|0),($107|0),($112|0),($2|0),($31|0),($__pat|0),($113|0),($114|0),($__grp|0),($__sym|0),($__sn|0),($41|0));
     $115 = __THREW__; __THREW__ = 0;
     $116 = $115&1;
     if ($116) {
      $__h$sroa$0$0 = $__h$sroa$0$1;
     } else {
      $117 = HEAP32[$__s>>2]|0;
      HEAP32[$0>>2] = $117;
      $118 = HEAP32[$__mi>>2]|0;
      $119 = HEAP32[$__me>>2]|0;
      __THREW__ = 0;
      ;HEAP32[$$byval_copy>>2]=HEAP32[$0>>2]|0;
      $120 = (invoke_iiiiiii(4886,($$byval_copy|0),($__mb$0|0),($118|0),($119|0),($__iob|0),($__fl|0))|0);
      $121 = __THREW__; __THREW__ = 0;
      $122 = $121&1;
      if ($122) {
       $__h$sroa$0$0 = $__h$sroa$0$1;
      } else {
       $123 = ($__h$sroa$0$1|0)==(0);
       if ($123) {
        $125 = $1;
       } else {
        $124 = $__h$sroa$0$1;
        _free($124);
        $$pre20 = HEAP32[$__loc>>2]|0;
        $125 = $$pre20;
       }
       __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__sn);
       __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__sym);
       __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grp);
       (__ZNSt3__114__shared_count16__release_sharedEv($125)|0);
       STACKTOP = sp;return ($120|0);
      }
     }
    }
    $96 = ___cxa_find_matching_catch_2()|0;
    $97 = tempRet0;
    $98 = ($__h$sroa$0$0|0)==(0);
    if ($98) {
     $$0 = $96;$$01 = $97;$$pre18 = $1;
    } else {
     $99 = $__h$sroa$0$0;
     _free($99);
     $$pre18$pre = HEAP32[$__loc>>2]|0;
     $$0 = $96;$$01 = $97;$$pre18 = $$pre18$pre;
    }
   }
   __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__sn);
   __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($__sym);
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__grp);
   $$1 = $$0;$$12 = $$01;$126 = $$pre18;
  }
 } while(0);
 if ((label|0) == 22) {
  $92 = ___cxa_find_matching_catch_2()|0;
  $93 = tempRet0;
  $$1 = $92;$$12 = $93;$126 = $1;
 }
 (__ZNSt3__114__shared_count16__release_sharedEv($126)|0);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__18messagesIcED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__18messagesIcED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__18messagesIcE7do_openERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERKNS_6localeE($this,$__nm,$0) {
 $this = $this|0;
 $__nm = $__nm|0;
 $0 = $0|0;
 var $$ = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP8[$__nm>>0]|0;
 $2 = $1 & 1;
 $3 = ($2<<24>>24)==(0);
 $4 = ((($__nm)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($__nm)) + 1|0);
 $7 = $3 ? $6 : $5;
 $8 = (_catopen($7,1)|0);
 $9 = $8;
 $not$ = ($8|0)!=((-1)|0);
 $10 = $not$&1;
 $$ = $9 >>> $10;
 return ($$|0);
}
function __ZNKSt3__18messagesIcE6do_getEiiiRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEE($agg$result,$this,$__c,$__set,$__msgid,$__dflt) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $__c = $__c|0;
 $__set = $__set|0;
 $__msgid = $__msgid|0;
 $__dflt = $__dflt|0;
 var $$0$i = 0, $$0$i8 = 0, $$02 = 0, $$03 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__i$0$i$i = 0, $__i$0$i$i4 = 0, $__ndflt = 0, $exitcond$i$i = 0, $exitcond$i$i5 = 0, $lpad$phi$index = 0, $lpad$phi$index2 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__ndflt = sp;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $0 = (($__ndflt) + ($__i$0$i$i<<2)|0);
  HEAP32[$0>>2] = 0;
  $1 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $1;
 }
 $2 = HEAP8[$__dflt>>0]|0;
 $3 = $2 & 1;
 $4 = ($3<<24>>24)==(0);
 $5 = ((($__dflt)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($__dflt)) + 1|0);
 $8 = $4 ? $7 : $6;
 $9 = ((($__dflt)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $2&255;
 $12 = $11 >>> 1;
 $13 = $4 ? $12 : $10;
 $14 = (($8) + ($13)|0);
 $$0$i8 = $8;
 while(1) {
  $15 = ($$0$i8>>>0)<($14>>>0);
  if (!($15)) {
   label = 8;
   break;
  }
  $16 = HEAP8[$$0$i8>>0]|0;
  __THREW__ = 0;
  invoke_vii(3370,($__ndflt|0),($16|0));
  $17 = __THREW__; __THREW__ = 0;
  $18 = $17&1;
  if ($18) {
   label = 9;
   break;
  }
  $19 = ((($$0$i8)) + 1|0);
  $$0$i8 = $19;
 }
 do {
  if ((label|0) == 8) {
   $20 = ($__c|0)==(-1);
   $21 = $__c << 1;
   $22 = $21;
   $23 = $20 ? (-1) : $22;
   $24 = HEAP8[$__ndflt>>0]|0;
   $25 = $24 & 1;
   $26 = ($25<<24>>24)==(0);
   $27 = ((($__ndflt)) + 8|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = ((($__ndflt)) + 1|0);
   $30 = $26 ? $29 : $28;
   __THREW__ = 0;
   $31 = (invoke_iiiii(4916,($23|0),($__set|0),($__msgid|0),($30|0))|0);
   $32 = __THREW__; __THREW__ = 0;
   $33 = $32&1;
   if ($33) {
    $36 = ___cxa_find_matching_catch_2()|0;
    $37 = tempRet0;
    $lpad$phi$index = $36;$lpad$phi$index2 = $37;
    label = 11;
    break;
   }
   $__i$0$i$i4 = 0;
   while(1) {
    $exitcond$i$i5 = ($__i$0$i$i4|0)==(3);
    if ($exitcond$i$i5) {
     break;
    }
    $38 = (($agg$result) + ($__i$0$i$i4<<2)|0);
    HEAP32[$38>>2] = 0;
    $39 = (($__i$0$i$i4) + 1)|0;
    $__i$0$i$i4 = $39;
   }
   $40 = (_strlen($31)|0);
   $41 = (($31) + ($40)|0);
   $$0$i = $31;
   while(1) {
    $42 = ($$0$i>>>0)<($41>>>0);
    if (!($42)) {
     label = 20;
     break;
    }
    $43 = HEAP8[$$0$i>>0]|0;
    __THREW__ = 0;
    invoke_vii(3370,($agg$result|0),($43|0));
    $44 = __THREW__; __THREW__ = 0;
    $45 = $44&1;
    if ($45) {
     label = 19;
     break;
    }
    $46 = ((($$0$i)) + 1|0);
    $$0$i = $46;
   }
   if ((label|0) == 19) {
    $47 = ___cxa_find_matching_catch_2()|0;
    $48 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($agg$result);
    $$02 = $47;$$03 = $48;
    break;
   }
   else if ((label|0) == 20) {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__ndflt);
    STACKTOP = sp;return;
   }
  }
  else if ((label|0) == 9) {
   $34 = ___cxa_find_matching_catch_2()|0;
   $35 = tempRet0;
   $lpad$phi$index = $34;$lpad$phi$index2 = $35;
   label = 11;
  }
 } while(0);
 if ((label|0) == 11) {
  $$02 = $lpad$phi$index;$$03 = $lpad$phi$index2;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__ndflt);
 ___resumeException($$02|0);
 // unreachable;
}
function __ZNKSt3__18messagesIcE8do_closeEi($this,$__c) {
 $this = $this|0;
 $__c = $__c|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__18messagesIwED2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__18messagesIwED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__18messagesIwE7do_openERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERKNS_6localeE($this,$__nm,$0) {
 $this = $this|0;
 $__nm = $__nm|0;
 $0 = $0|0;
 var $$ = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP8[$__nm>>0]|0;
 $2 = $1 & 1;
 $3 = ($2<<24>>24)==(0);
 $4 = ((($__nm)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($__nm)) + 1|0);
 $7 = $3 ? $6 : $5;
 $8 = (_catopen($7,1)|0);
 $9 = $8;
 $not$ = ($8|0)!=((-1)|0);
 $10 = $not$&1;
 $$ = $9 >>> $10;
 return ($$|0);
}
function __ZNKSt3__18messagesIwE6do_getEiiiRKNS_12basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEEE($agg$result,$this,$__c,$__set,$__msgid,$__dflt) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $__c = $__c|0;
 $__set = $__set|0;
 $__msgid = $__msgid|0;
 $__dflt = $__dflt|0;
 var $$0$i = 0, $$0$i14 = 0, $$1 = 0, $$14 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $__bn$i = 0, $__bn$i12 = 0, $__buf$i = 0, $__buf$i11 = 0, $__i$0$i$i = 0, $__i$0$i$i8 = 0, $__mb$i = 0, $__mb$i10 = 0, $__ndflt = 0, $__nn$i = 0, $__p$0$i = 0, $__p$0$i16 = 0, $__r$0$i = 0, $__r$0$i13 = 0, $__wn$i = 0, $exitcond$i$i = 0, $exitcond$i$i9 = 0, $lpad$phi$index = 0;
 var $lpad$phi$index10 = 0, $lpad$phi30$index = 0, $lpad$phi30$index3 = 0, $or$cond$i = 0, $or$cond$i15 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__mb$i10 = sp + 184|0;
 $__buf$i11 = sp + 192|0;
 $__bn$i12 = sp + 180|0;
 $__wn$i = sp + 176|0;
 $__mb$i = sp + 168|0;
 $__buf$i = sp + 40|0;
 $__bn$i = sp + 32|0;
 $__nn$i = sp + 28|0;
 $__ndflt = sp + 16|0;
 $0 = sp + 8|0;
 $1 = sp;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $2 = (($__ndflt) + ($__i$0$i$i<<2)|0);
  HEAP32[$2>>2] = 0;
  $3 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $3;
 }
 $4 = ((($0)) + 4|0);
 HEAP32[$4>>2] = 0;
 HEAP32[$0>>2] = (157152);
 $5 = HEAP8[$__dflt>>0]|0;
 $6 = $5 & 1;
 $7 = ($6<<24>>24)==(0);
 $8 = ((($__dflt)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($__dflt)) + 4|0);
 $11 = $7 ? $10 : $9;
 $12 = HEAP32[$10>>2]|0;
 $13 = $5&255;
 $14 = $13 >>> 1;
 $15 = $7 ? $14 : $12;
 $16 = (($11) + ($15<<2)|0);
 $17 = ((($__buf$i11)) + 32|0);
 $$0$i14 = $11;$__r$0$i13 = 0;
 L5: while(1) {
  $18 = ($$0$i14>>>0)<($16>>>0);
  $19 = ($__r$0$i13|0)!=(2);
  $20 = $19 & $18;
  if (!($20)) {
   label = 13;
   break;
  }
  HEAP32[$__wn$i>>2] = $$0$i14;
  $21 = HEAP32[$0>>2]|0;
  $22 = ((($21)) + 12|0);
  $23 = HEAP32[$22>>2]|0;
  __THREW__ = 0;
  $24 = (invoke_iiiiiiiii($23|0,($0|0),($__mb$i10|0),($$0$i14|0),($16|0),($__wn$i|0),($__buf$i11|0),($17|0),($__bn$i12|0))|0);
  $25 = __THREW__; __THREW__ = 0;
  $26 = $25&1;
  if ($26) {
   label = 16;
   break;
  }
  $27 = ($24|0)==(2);
  $28 = HEAP32[$__wn$i>>2]|0;
  $29 = ($28|0)==($$0$i14|0);
  $or$cond$i15 = $27 | $29;
  if ($or$cond$i15) {
   label = 8;
   break;
  } else {
   $__p$0$i16 = $__buf$i11;
  }
  while(1) {
   $33 = HEAP32[$__bn$i12>>2]|0;
   $34 = ($__p$0$i16>>>0)<($33>>>0);
   if (!($34)) {
    break;
   }
   $36 = HEAP8[$__p$0$i16>>0]|0;
   __THREW__ = 0;
   invoke_vii(3370,($__ndflt|0),($36|0));
   $37 = __THREW__; __THREW__ = 0;
   $38 = $37&1;
   if ($38) {
    label = 15;
    break L5;
   }
   $39 = ((($__p$0$i16)) + 1|0);
   $__p$0$i16 = $39;
  }
  $35 = HEAP32[$__wn$i>>2]|0;
  $$0$i14 = $35;$__r$0$i13 = $24;
 }
 do {
  if ((label|0) == 8) {
   __THREW__ = 0;
   invoke_vi(4901,(621542|0));
   $30 = __THREW__; __THREW__ = 0;
   $31 = ___cxa_find_matching_catch_2()|0;
   $32 = tempRet0;
   $lpad$phi30$index = $31;$lpad$phi30$index3 = $32;
   label = 17;
  }
  else if ((label|0) == 13) {
   $40 = ($__c|0)==(-1);
   $41 = $__c << 1;
   $42 = $41;
   $43 = $40 ? (-1) : $42;
   $44 = HEAP8[$__ndflt>>0]|0;
   $45 = $44 & 1;
   $46 = ($45<<24>>24)==(0);
   $47 = ((($__ndflt)) + 8|0);
   $48 = HEAP32[$47>>2]|0;
   $49 = ((($__ndflt)) + 1|0);
   $50 = $46 ? $49 : $48;
   __THREW__ = 0;
   $51 = (invoke_iiiii(4916,($43|0),($__set|0),($__msgid|0),($50|0))|0);
   $52 = __THREW__; __THREW__ = 0;
   $53 = $52&1;
   if ($53) {
    $54 = ___cxa_find_matching_catch_2()|0;
    $55 = tempRet0;
    $$1 = $54;$$14 = $55;
    break;
   }
   $__i$0$i$i8 = 0;
   while(1) {
    $exitcond$i$i9 = ($__i$0$i$i8|0)==(3);
    if ($exitcond$i$i9) {
     break;
    }
    $60 = (($agg$result) + ($__i$0$i$i8<<2)|0);
    HEAP32[$60>>2] = 0;
    $61 = (($__i$0$i$i8) + 1)|0;
    $__i$0$i$i8 = $61;
   }
   $62 = ((($1)) + 4|0);
   HEAP32[$62>>2] = 0;
   HEAP32[$1>>2] = (157200);
   $63 = (_strlen($51)|0);
   $64 = (($51) + ($63)|0);
   $65 = $64;
   $66 = ((($__buf$i)) + 128|0);
   $$0$i = $51;$__r$0$i = 0;
   L23: while(1) {
    $67 = ($$0$i>>>0)<($64>>>0);
    $68 = ($__r$0$i|0)!=(2);
    $69 = $68 & $67;
    if (!($69)) {
     label = 30;
     break;
    }
    HEAP32[$__nn$i>>2] = $$0$i;
    $70 = HEAP32[$1>>2]|0;
    $71 = ((($70)) + 16|0);
    $72 = HEAP32[$71>>2]|0;
    $73 = $$0$i;
    $74 = (($65) - ($73))|0;
    $75 = ($74|0)>(32);
    $76 = ((($$0$i)) + 32|0);
    $77 = $75 ? $76 : $64;
    __THREW__ = 0;
    $78 = (invoke_iiiiiiiii($72|0,($1|0),($__mb$i|0),($$0$i|0),($77|0),($__nn$i|0),($__buf$i|0),($66|0),($__bn$i|0))|0);
    $79 = __THREW__; __THREW__ = 0;
    $80 = $79&1;
    if ($80) {
     label = 32;
     break;
    }
    $81 = ($78|0)==(2);
    $82 = HEAP32[$__nn$i>>2]|0;
    $83 = ($82|0)==($$0$i|0);
    $or$cond$i = $81 | $83;
    if ($or$cond$i) {
     label = 25;
     break;
    } else {
     $__p$0$i = $__buf$i;
    }
    while(1) {
     $87 = HEAP32[$__bn$i>>2]|0;
     $88 = ($__p$0$i>>>0)<($87>>>0);
     if (!($88)) {
      break;
     }
     $90 = HEAP32[$__p$0$i>>2]|0;
     __THREW__ = 0;
     invoke_vii(4908,($agg$result|0),($90|0));
     $91 = __THREW__; __THREW__ = 0;
     $92 = $91&1;
     if ($92) {
      label = 31;
      break L23;
     }
     $93 = ((($__p$0$i)) + 4|0);
     $__p$0$i = $93;
    }
    $89 = HEAP32[$__nn$i>>2]|0;
    $$0$i = $89;$__r$0$i = $78;
   }
   if ((label|0) == 25) {
    __THREW__ = 0;
    invoke_vi(4901,(621542|0));
    $84 = __THREW__; __THREW__ = 0;
    $85 = ___cxa_find_matching_catch_2()|0;
    $86 = tempRet0;
    $lpad$phi$index = $85;$lpad$phi$index10 = $86;
   }
   else if ((label|0) == 30) {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__ndflt);
    STACKTOP = sp;return;
   }
   else if ((label|0) == 31) {
    $94 = ___cxa_find_matching_catch_2()|0;
    $95 = tempRet0;
    $lpad$phi$index = $94;$lpad$phi$index10 = $95;
   }
   else if ((label|0) == 32) {
    $96 = ___cxa_find_matching_catch_2()|0;
    $97 = tempRet0;
    $lpad$phi$index = $96;$lpad$phi$index10 = $97;
   }
   __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($agg$result);
   $$1 = $lpad$phi$index;$$14 = $lpad$phi$index10;
  }
  else if ((label|0) == 15) {
   $56 = ___cxa_find_matching_catch_2()|0;
   $57 = tempRet0;
   $lpad$phi30$index = $56;$lpad$phi30$index3 = $57;
   label = 17;
  }
  else if ((label|0) == 16) {
   $58 = ___cxa_find_matching_catch_2()|0;
   $59 = tempRet0;
   $lpad$phi30$index = $58;$lpad$phi30$index3 = $59;
   label = 17;
  }
 } while(0);
 if ((label|0) == 17) {
  $$1 = $lpad$phi30$index;$$14 = $lpad$phi30$index3;
 }
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($__ndflt);
 ___resumeException($$1|0);
 // unreachable;
}
function __ZNSt3__16locale5facetD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__116__narrow_to_utf8ILj32EED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__17codecvtIDic11__mbstate_tE6do_outERS1_PKDiS5_RS5_PcS7_RS7_($this,$0,$frm,$frm_end,$frm_nxt,$to,$to_end,$to_nxt) {
 $this = $this|0;
 $0 = $0|0;
 $frm = $frm|0;
 $frm_end = $frm_end|0;
 $frm_nxt = $frm_nxt|0;
 $to = $to|0;
 $to_end = $to_end|0;
 $to_nxt = $to_nxt|0;
 var $1 = 0, $2 = 0, $3 = 0, $_frm_nxt = 0, $_to_nxt = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $_frm_nxt = sp + 4|0;
 $_to_nxt = sp;
 HEAP32[$_frm_nxt>>2] = $frm;
 HEAP32[$_to_nxt>>2] = $to;
 $1 = (__ZNSt3__1L12ucs4_to_utf8EPKjS1_RS1_PhS3_RS3_mNS_12codecvt_modeE($frm,$frm_end,$_frm_nxt,$to,$to_end,$_to_nxt,1114111,0)|0);
 $2 = HEAP32[$_frm_nxt>>2]|0;
 HEAP32[$frm_nxt>>2] = $2;
 $3 = HEAP32[$_to_nxt>>2]|0;
 HEAP32[$to_nxt>>2] = $3;
 STACKTOP = sp;return ($1|0);
}
function __ZNSt3__1L12ucs4_to_utf8EPKjS1_RS1_PhS3_RS3_mNS_12codecvt_modeE($frm,$frm_end,$frm_nxt,$to,$to_end,$to_nxt,$Maxcode,$mode) {
 $frm = $frm|0;
 $frm_end = $frm_end|0;
 $frm_nxt = $frm_nxt|0;
 $to = $to|0;
 $to_end = $to_end|0;
 $to_nxt = $to_nxt|0;
 $Maxcode = $Maxcode|0;
 $mode = $mode|0;
 var $$2 = 0, $$pre = 0, $$pre2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$frm_nxt>>2] = $frm;
 HEAP32[$to_nxt>>2] = $to;
 $0 = $mode & 2;
 $1 = ($0|0)==(0);
 $$pre2 = $to_end;
 if ($1) {
  label = 4;
 } else {
  $2 = $to;
  $3 = (($$pre2) - ($2))|0;
  $4 = ($3|0)<(3);
  if ($4) {
   $$2 = 1;
  } else {
   $5 = ((($to)) + 1|0);
   HEAP32[$to_nxt>>2] = $5;
   HEAP8[$to>>0] = -17;
   $6 = HEAP32[$to_nxt>>2]|0;
   $7 = ((($6)) + 1|0);
   HEAP32[$to_nxt>>2] = $7;
   HEAP8[$6>>0] = -69;
   $8 = HEAP32[$to_nxt>>2]|0;
   $9 = ((($8)) + 1|0);
   HEAP32[$to_nxt>>2] = $9;
   HEAP8[$8>>0] = -65;
   label = 4;
  }
 }
 L4: do {
  if ((label|0) == 4) {
   $$pre = HEAP32[$frm_nxt>>2]|0;
   $11 = $$pre;
   while(1) {
    $10 = ($11>>>0)<($frm_end>>>0);
    if (!($10)) {
     $$2 = 0;
     break L4;
    }
    $12 = HEAP32[$11>>2]|0;
    $13 = $12 & -2048;
    $14 = ($13|0)==(55296);
    $15 = ($12>>>0)>($Maxcode>>>0);
    $or$cond = $15 | $14;
    if ($or$cond) {
     $$2 = 2;
     break L4;
    }
    $16 = ($12>>>0)<(128);
    do {
     if ($16) {
      $17 = HEAP32[$to_nxt>>2]|0;
      $18 = $17;
      $19 = (($$pre2) - ($18))|0;
      $20 = ($19|0)<(1);
      if ($20) {
       $$2 = 1;
       break L4;
      }
      $21 = $12&255;
      $22 = ((($17)) + 1|0);
      HEAP32[$to_nxt>>2] = $22;
      HEAP8[$17>>0] = $21;
     } else {
      $23 = ($12>>>0)<(2048);
      if ($23) {
       $24 = HEAP32[$to_nxt>>2]|0;
       $25 = $24;
       $26 = (($$pre2) - ($25))|0;
       $27 = ($26|0)<(2);
       if ($27) {
        $$2 = 1;
        break L4;
       }
       $28 = $12 >>> 6;
       $29 = $28 | 192;
       $30 = $29&255;
       $31 = ((($24)) + 1|0);
       HEAP32[$to_nxt>>2] = $31;
       HEAP8[$24>>0] = $30;
       $32 = $12 & 63;
       $33 = $32 | 128;
       $34 = $33&255;
       $35 = HEAP32[$to_nxt>>2]|0;
       $36 = ((($35)) + 1|0);
       HEAP32[$to_nxt>>2] = $36;
       HEAP8[$35>>0] = $34;
       break;
      }
      $37 = ($12>>>0)<(65536);
      $38 = HEAP32[$to_nxt>>2]|0;
      $39 = $38;
      $40 = (($$pre2) - ($39))|0;
      if ($37) {
       $41 = ($40|0)<(3);
       if ($41) {
        $$2 = 1;
        break L4;
       }
       $42 = $12 >>> 12;
       $43 = $42 | 224;
       $44 = $43&255;
       $45 = ((($38)) + 1|0);
       HEAP32[$to_nxt>>2] = $45;
       HEAP8[$38>>0] = $44;
       $46 = $12 >>> 6;
       $47 = $46 & 63;
       $48 = $47 | 128;
       $49 = $48&255;
       $50 = HEAP32[$to_nxt>>2]|0;
       $51 = ((($50)) + 1|0);
       HEAP32[$to_nxt>>2] = $51;
       HEAP8[$50>>0] = $49;
       $52 = $12 & 63;
       $53 = $52 | 128;
       $54 = $53&255;
       $55 = HEAP32[$to_nxt>>2]|0;
       $56 = ((($55)) + 1|0);
       HEAP32[$to_nxt>>2] = $56;
       HEAP8[$55>>0] = $54;
       break;
      } else {
       $57 = ($40|0)<(4);
       if ($57) {
        $$2 = 1;
        break L4;
       }
       $58 = $12 >>> 18;
       $59 = $58 | 240;
       $60 = $59&255;
       $61 = ((($38)) + 1|0);
       HEAP32[$to_nxt>>2] = $61;
       HEAP8[$38>>0] = $60;
       $62 = $12 >>> 12;
       $63 = $62 & 63;
       $64 = $63 | 128;
       $65 = $64&255;
       $66 = HEAP32[$to_nxt>>2]|0;
       $67 = ((($66)) + 1|0);
       HEAP32[$to_nxt>>2] = $67;
       HEAP8[$66>>0] = $65;
       $68 = $12 >>> 6;
       $69 = $68 & 63;
       $70 = $69 | 128;
       $71 = $70&255;
       $72 = HEAP32[$to_nxt>>2]|0;
       $73 = ((($72)) + 1|0);
       HEAP32[$to_nxt>>2] = $73;
       HEAP8[$72>>0] = $71;
       $74 = $12 & 63;
       $75 = $74 | 128;
       $76 = $75&255;
       $77 = HEAP32[$to_nxt>>2]|0;
       $78 = ((($77)) + 1|0);
       HEAP32[$to_nxt>>2] = $78;
       HEAP8[$77>>0] = $76;
       break;
      }
     }
    } while(0);
    $79 = HEAP32[$frm_nxt>>2]|0;
    $80 = ((($79)) + 4|0);
    HEAP32[$frm_nxt>>2] = $80;
    $11 = $80;
   }
  }
 } while(0);
 return ($$2|0);
}
function __ZNKSt3__17codecvtIDic11__mbstate_tE5do_inERS1_PKcS5_RS5_PDiS7_RS7_($this,$0,$frm,$frm_end,$frm_nxt,$to,$to_end,$to_nxt) {
 $this = $this|0;
 $0 = $0|0;
 $frm = $frm|0;
 $frm_end = $frm_end|0;
 $frm_nxt = $frm_nxt|0;
 $to = $to|0;
 $to_end = $to_end|0;
 $to_nxt = $to_nxt|0;
 var $1 = 0, $2 = 0, $3 = 0, $_frm_nxt = 0, $_to_nxt = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $_frm_nxt = sp + 4|0;
 $_to_nxt = sp;
 HEAP32[$_frm_nxt>>2] = $frm;
 HEAP32[$_to_nxt>>2] = $to;
 $1 = (__ZNSt3__1L12utf8_to_ucs4EPKhS1_RS1_PjS3_RS3_mNS_12codecvt_modeE($frm,$frm_end,$_frm_nxt,$to,$to_end,$_to_nxt,1114111,0)|0);
 $2 = HEAP32[$_frm_nxt>>2]|0;
 HEAP32[$frm_nxt>>2] = $2;
 $3 = HEAP32[$_to_nxt>>2]|0;
 HEAP32[$to_nxt>>2] = $3;
 STACKTOP = sp;return ($1|0);
}
function __ZNSt3__1L12utf8_to_ucs4EPKhS1_RS1_PjS3_RS3_mNS_12codecvt_modeE($frm,$frm_end,$frm_nxt,$to,$to_end,$to_nxt,$Maxcode,$mode) {
 $frm = $frm|0;
 $frm_end = $frm_end|0;
 $frm_nxt = $frm_nxt|0;
 $to = $to|0;
 $to_end = $to_end|0;
 $to_nxt = $to_nxt|0;
 $Maxcode = $Maxcode|0;
 $mode = $mode|0;
 var $$9 = 0, $$lcssa = 0, $$off = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$frm_nxt>>2] = $frm;
 HEAP32[$to_nxt>>2] = $to;
 $0 = $mode & 4;
 $1 = ($0|0)==(0);
 if ($1) {
  $$pre = $frm_end;
  $$pre$phiZ2D = $$pre;
 } else {
  $2 = HEAP32[$frm_nxt>>2]|0;
  $3 = $frm_end;
  $4 = $2;
  $5 = (($3) - ($4))|0;
  $6 = ($5|0)>(2);
  if ($6) {
   $7 = HEAP8[$2>>0]|0;
   $8 = ($7<<24>>24)==(-17);
   if ($8) {
    $9 = ((($2)) + 1|0);
    $10 = HEAP8[$9>>0]|0;
    $11 = ($10<<24>>24)==(-69);
    if ($11) {
     $12 = ((($2)) + 2|0);
     $13 = HEAP8[$12>>0]|0;
     $14 = ($13<<24>>24)==(-65);
     if ($14) {
      $15 = ((($2)) + 3|0);
      HEAP32[$frm_nxt>>2] = $15;
      $$pre$phiZ2D = $3;
     } else {
      $$pre$phiZ2D = $3;
     }
    } else {
     $$pre$phiZ2D = $3;
    }
   } else {
    $$pre$phiZ2D = $3;
   }
  } else {
   $$pre$phiZ2D = $3;
  }
 }
 L9: while(1) {
  $16 = HEAP32[$frm_nxt>>2]|0;
  $17 = ($16>>>0)<($frm_end>>>0);
  if (!($17)) {
   $$lcssa = $17;
   label = 40;
   break;
  }
  $18 = HEAP32[$to_nxt>>2]|0;
  $19 = ($18>>>0)<($to_end>>>0);
  if (!($19)) {
   $$lcssa = $17;
   label = 40;
   break;
  }
  $20 = HEAP8[$16>>0]|0;
  $21 = $20&255;
  $22 = ($20<<24>>24)>(-1);
  do {
   if ($22) {
    $23 = ($21>>>0)>($Maxcode>>>0);
    if ($23) {
     $$9 = 2;
     break L9;
    }
    HEAP32[$18>>2] = $21;
    $24 = ((($16)) + 1|0);
    HEAP32[$frm_nxt>>2] = $24;
   } else {
    $25 = ($20&255)<(194);
    if ($25) {
     $$9 = 2;
     break L9;
    }
    $26 = ($20&255)<(224);
    if ($26) {
     $27 = $16;
     $28 = (($$pre$phiZ2D) - ($27))|0;
     $29 = ($28|0)<(2);
     if ($29) {
      $$9 = 1;
      break L9;
     }
     $30 = ((($16)) + 1|0);
     $31 = HEAP8[$30>>0]|0;
     $32 = $31&255;
     $33 = $32 & 192;
     $34 = ($33|0)==(128);
     if (!($34)) {
      $$9 = 2;
      break L9;
     }
     $35 = $21 << 6;
     $36 = $35 & 1984;
     $37 = $32 & 63;
     $38 = $37 | $36;
     $39 = ($38>>>0)>($Maxcode>>>0);
     if ($39) {
      $$9 = 2;
      break L9;
     }
     HEAP32[$18>>2] = $38;
     $40 = ((($16)) + 2|0);
     HEAP32[$frm_nxt>>2] = $40;
     break;
    }
    $41 = ($20&255)<(240);
    if ($41) {
     $42 = $16;
     $43 = (($$pre$phiZ2D) - ($42))|0;
     $44 = ($43|0)<(3);
     if ($44) {
      $$9 = 1;
      break L9;
     }
     $45 = ((($16)) + 1|0);
     $46 = HEAP8[$45>>0]|0;
     $47 = ((($16)) + 2|0);
     $48 = HEAP8[$47>>0]|0;
     switch ($21|0) {
     case 224:  {
      $49 = $46 & -32;
      $50 = ($49<<24>>24)==(-96);
      if (!($50)) {
       $$9 = 2;
       break L9;
      }
      break;
     }
     case 237:  {
      $51 = $46 & -32;
      $52 = ($51<<24>>24)==(-128);
      if (!($52)) {
       $$9 = 2;
       break L9;
      }
      break;
     }
     default: {
      $53 = $46 & -64;
      $54 = ($53<<24>>24)==(-128);
      if (!($54)) {
       $$9 = 2;
       break L9;
      }
     }
     }
     $55 = $48&255;
     $56 = $55 & 192;
     $57 = ($56|0)==(128);
     if (!($57)) {
      $$9 = 2;
      break L9;
     }
     $58 = $21 << 12;
     $59 = $58 & 61440;
     $60 = $46&255;
     $61 = $60 << 6;
     $62 = $61 & 4032;
     $63 = $62 | $59;
     $64 = $55 & 63;
     $65 = $63 | $64;
     $66 = ($65>>>0)>($Maxcode>>>0);
     if ($66) {
      $$9 = 2;
      break L9;
     }
     HEAP32[$18>>2] = $65;
     $67 = ((($16)) + 3|0);
     HEAP32[$frm_nxt>>2] = $67;
     break;
    }
    $68 = ($20&255)<(245);
    if (!($68)) {
     $$9 = 2;
     break L9;
    }
    $69 = $16;
    $70 = (($$pre$phiZ2D) - ($69))|0;
    $71 = ($70|0)<(4);
    if ($71) {
     $$9 = 1;
     break L9;
    }
    $72 = ((($16)) + 1|0);
    $73 = HEAP8[$72>>0]|0;
    $74 = ((($16)) + 2|0);
    $75 = HEAP8[$74>>0]|0;
    $76 = ((($16)) + 3|0);
    $77 = HEAP8[$76>>0]|0;
    switch ($21|0) {
    case 240:  {
     $$off = (($73) + 112)<<24>>24;
     $78 = ($$off&255)<(48);
     if (!($78)) {
      $$9 = 2;
      break L9;
     }
     break;
    }
    case 244:  {
     $79 = $73 & -16;
     $80 = ($79<<24>>24)==(-128);
     if (!($80)) {
      $$9 = 2;
      break L9;
     }
     break;
    }
    default: {
     $81 = $73 & -64;
     $82 = ($81<<24>>24)==(-128);
     if (!($82)) {
      $$9 = 2;
      break L9;
     }
    }
    }
    $83 = $75&255;
    $84 = $83 & 192;
    $85 = ($84|0)==(128);
    if (!($85)) {
     $$9 = 2;
     break L9;
    }
    $86 = $77&255;
    $87 = $86 & 192;
    $88 = ($87|0)==(128);
    if (!($88)) {
     $$9 = 2;
     break L9;
    }
    $89 = $21 << 18;
    $90 = $89 & 1835008;
    $91 = $73&255;
    $92 = $91 << 12;
    $93 = $92 & 258048;
    $94 = $93 | $90;
    $95 = $83 << 6;
    $96 = $95 & 4032;
    $97 = $94 | $96;
    $98 = $86 & 63;
    $99 = $97 | $98;
    $100 = ($99>>>0)>($Maxcode>>>0);
    if ($100) {
     $$9 = 2;
     break L9;
    }
    HEAP32[$18>>2] = $99;
    $101 = ((($16)) + 4|0);
    HEAP32[$frm_nxt>>2] = $101;
   }
  } while(0);
  $102 = HEAP32[$to_nxt>>2]|0;
  $103 = ((($102)) + 4|0);
  HEAP32[$to_nxt>>2] = $103;
 }
 if ((label|0) == 40) {
  $104 = $$lcssa&1;
  $$9 = $104;
 }
 return ($$9|0);
}
function __ZNKSt3__17codecvtIDic11__mbstate_tE10do_unshiftERS1_PcS4_RS4_($this,$0,$to,$1,$to_nxt) {
 $this = $this|0;
 $0 = $0|0;
 $to = $to|0;
 $1 = $1|0;
 $to_nxt = $to_nxt|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$to_nxt>>2] = $to;
 return 3;
}
function __ZNKSt3__17codecvtIDic11__mbstate_tE11do_encodingEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__17codecvtIDic11__mbstate_tE16do_always_noconvEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__17codecvtIDic11__mbstate_tE9do_lengthERS1_PKcS5_j($this,$0,$frm,$frm_end,$mx) {
 $this = $this|0;
 $0 = $0|0;
 $frm = $frm|0;
 $frm_end = $frm_end|0;
 $mx = $mx|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (__ZNSt3__1L19utf8_to_ucs4_lengthEPKhS1_jmNS_12codecvt_modeE($frm,$frm_end,$mx,1114111,0)|0);
 return ($1|0);
}
function __ZNSt3__1L19utf8_to_ucs4_lengthEPKhS1_jmNS_12codecvt_modeE($frm,$frm_end,$mx,$Maxcode,$mode) {
 $frm = $frm|0;
 $frm_end = $frm_end|0;
 $mx = $mx|0;
 $Maxcode = $Maxcode|0;
 $mode = $mode|0;
 var $$37 = 0, $$frm = 0, $$lcssa = 0, $$lcssa88 = 0, $$lcssa89 = 0, $$lcssa90 = 0, $$lcssa91 = 0, $$lcssa92 = 0, $$off = 0, $$pre = 0, $$pre43 = 0, $$pre45 = 0, $$pre47 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $frm_nxt$1 = 0, $frm_nxt$1$lcssa = 0, $frm_nxt$5 = 0, $nchar32_t$0 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $mode & 4;
 $1 = ($0|0)==(0);
 $$pre47 = $frm_end;
 if ($1) {
  $frm_nxt$1 = $frm;$nchar32_t$0 = 0;
 } else {
  $2 = $frm;
  $3 = (($$pre47) - ($2))|0;
  $4 = ($3|0)>(2);
  if ($4) {
   $5 = HEAP8[$frm>>0]|0;
   $6 = ($5<<24>>24)==(-17);
   if ($6) {
    $7 = ((($frm)) + 1|0);
    $8 = HEAP8[$7>>0]|0;
    $9 = ($8<<24>>24)==(-69);
    if ($9) {
     $10 = ((($frm)) + 2|0);
     $11 = HEAP8[$10>>0]|0;
     $12 = ($11<<24>>24)==(-65);
     $13 = ((($frm)) + 3|0);
     $$frm = $12 ? $13 : $frm;
     $frm_nxt$1 = $$frm;$nchar32_t$0 = 0;
    } else {
     $frm_nxt$1 = $frm;$nchar32_t$0 = 0;
    }
   } else {
    $frm_nxt$1 = $frm;$nchar32_t$0 = 0;
   }
  } else {
   $frm_nxt$1 = $frm;$nchar32_t$0 = 0;
  }
 }
 L6: while(1) {
  $14 = ($frm_nxt$1>>>0)<($frm_end>>>0);
  $15 = ($nchar32_t$0>>>0)<($mx>>>0);
  $or$cond = $15 & $14;
  if (!($or$cond)) {
   $frm_nxt$1$lcssa = $frm_nxt$1;
   label = 40;
   break;
  }
  $16 = HEAP8[$frm_nxt$1>>0]|0;
  $17 = $16&255;
  $18 = ($16<<24>>24)>(-1);
  do {
   if ($18) {
    $19 = ($17>>>0)>($Maxcode>>>0);
    if ($19) {
     $frm_nxt$1$lcssa = $frm_nxt$1;
     label = 40;
     break L6;
    }
    $20 = ((($frm_nxt$1)) + 1|0);
    $frm_nxt$5 = $20;
   } else {
    $21 = ($16&255)<(194);
    if ($21) {
     $frm_nxt$1$lcssa = $frm_nxt$1;
     label = 40;
     break L6;
    }
    $22 = ($16&255)<(224);
    if ($22) {
     $23 = $frm_nxt$1;
     $24 = (($$pre47) - ($23))|0;
     $25 = ($24|0)<(2);
     if ($25) {
      $frm_nxt$1$lcssa = $frm_nxt$1;
      label = 40;
      break L6;
     }
     $26 = ((($frm_nxt$1)) + 1|0);
     $27 = HEAP8[$26>>0]|0;
     $28 = $27&255;
     $29 = $28 & 192;
     $30 = ($29|0)==(128);
     if (!($30)) {
      $frm_nxt$1$lcssa = $frm_nxt$1;
      label = 40;
      break L6;
     }
     $31 = $17 << 6;
     $32 = $31 & 1984;
     $33 = $28 & 63;
     $34 = $33 | $32;
     $35 = ($34>>>0)>($Maxcode>>>0);
     if ($35) {
      $frm_nxt$1$lcssa = $frm_nxt$1;
      label = 40;
      break L6;
     }
     $36 = ((($frm_nxt$1)) + 2|0);
     $frm_nxt$5 = $36;
     break;
    }
    $37 = ($16&255)<(240);
    if ($37) {
     $38 = $frm_nxt$1;
     $39 = (($$pre47) - ($38))|0;
     $40 = ($39|0)<(3);
     if ($40) {
      $frm_nxt$1$lcssa = $frm_nxt$1;
      label = 40;
      break L6;
     }
     $41 = ((($frm_nxt$1)) + 1|0);
     $42 = HEAP8[$41>>0]|0;
     $43 = ((($frm_nxt$1)) + 2|0);
     $44 = HEAP8[$43>>0]|0;
     switch ($17|0) {
     case 224:  {
      $45 = $42 & -32;
      $46 = ($45<<24>>24)==(-96);
      if (!($46)) {
       $$lcssa91 = $38;
       label = 20;
       break L6;
      }
      break;
     }
     case 237:  {
      $49 = $42 & -32;
      $50 = ($49<<24>>24)==(-128);
      if (!($50)) {
       $$lcssa90 = $38;
       label = 22;
       break L6;
      }
      break;
     }
     default: {
      $53 = $42 & -64;
      $54 = ($53<<24>>24)==(-128);
      if (!($54)) {
       $$lcssa92 = $38;
       label = 24;
       break L6;
      }
     }
     }
     $57 = $44&255;
     $58 = $57 & 192;
     $59 = ($58|0)==(128);
     if (!($59)) {
      $frm_nxt$1$lcssa = $frm_nxt$1;
      label = 40;
      break L6;
     }
     $60 = $17 << 12;
     $61 = $60 & 61440;
     $62 = $42&255;
     $63 = $62 << 6;
     $64 = $63 & 4032;
     $65 = $64 | $61;
     $66 = $57 & 63;
     $67 = $65 | $66;
     $68 = ($67>>>0)>($Maxcode>>>0);
     $69 = ((($frm_nxt$1)) + 3|0);
     if ($68) {
      $frm_nxt$1$lcssa = $frm_nxt$1;
      label = 40;
      break L6;
     } else {
      $frm_nxt$5 = $69;
      break;
     }
    }
    $70 = ($16&255)<(245);
    if (!($70)) {
     $frm_nxt$1$lcssa = $frm_nxt$1;
     label = 40;
     break L6;
    }
    $71 = $frm_nxt$1;
    $72 = (($$pre47) - ($71))|0;
    $73 = ($72|0)<(4);
    if ($73) {
     $frm_nxt$1$lcssa = $frm_nxt$1;
     label = 40;
     break L6;
    }
    $74 = ((($frm_nxt$1)) + 1|0);
    $75 = HEAP8[$74>>0]|0;
    $76 = ((($frm_nxt$1)) + 2|0);
    $77 = HEAP8[$76>>0]|0;
    $78 = ((($frm_nxt$1)) + 3|0);
    $79 = HEAP8[$78>>0]|0;
    switch ($17|0) {
    case 240:  {
     $$off = (($75) + 112)<<24>>24;
     $80 = ($$off&255)<(48);
     if (!($80)) {
      $$lcssa88 = $71;
      label = 31;
      break L6;
     }
     break;
    }
    case 244:  {
     $83 = $75 & -16;
     $84 = ($83<<24>>24)==(-128);
     if (!($84)) {
      $$lcssa = $71;
      label = 33;
      break L6;
     }
     break;
    }
    default: {
     $87 = $75 & -64;
     $88 = ($87<<24>>24)==(-128);
     if (!($88)) {
      $$lcssa89 = $71;
      label = 35;
      break L6;
     }
    }
    }
    $91 = $77&255;
    $92 = $91 & 192;
    $93 = ($92|0)==(128);
    if (!($93)) {
     $frm_nxt$1$lcssa = $frm_nxt$1;
     label = 40;
     break L6;
    }
    $94 = $79&255;
    $95 = $94 & 192;
    $96 = ($95|0)==(128);
    if (!($96)) {
     $frm_nxt$1$lcssa = $frm_nxt$1;
     label = 40;
     break L6;
    }
    $97 = $17 << 18;
    $98 = $97 & 1835008;
    $99 = $75&255;
    $100 = $99 << 12;
    $101 = $100 & 258048;
    $102 = $101 | $98;
    $103 = $91 << 6;
    $104 = $103 & 4032;
    $105 = $102 | $104;
    $106 = $94 & 63;
    $107 = $105 | $106;
    $108 = ($107>>>0)>($Maxcode>>>0);
    $109 = ((($frm_nxt$1)) + 4|0);
    if ($108) {
     $frm_nxt$1$lcssa = $frm_nxt$1;
     label = 40;
     break L6;
    } else {
     $frm_nxt$5 = $109;
    }
   }
  } while(0);
  $110 = (($nchar32_t$0) + 1)|0;
  $frm_nxt$1 = $frm_nxt$5;$nchar32_t$0 = $110;
 }
 if ((label|0) == 20) {
  $47 = $frm;
  $48 = (($$lcssa91) - ($47))|0;
  $$37 = $48;
 }
 else if ((label|0) == 22) {
  $51 = $frm;
  $52 = (($$lcssa90) - ($51))|0;
  $$37 = $52;
 }
 else if ((label|0) == 24) {
  $55 = $frm;
  $56 = (($$lcssa92) - ($55))|0;
  $$37 = $56;
 }
 else if ((label|0) == 31) {
  $81 = $frm;
  $82 = (($$lcssa88) - ($81))|0;
  $$37 = $82;
 }
 else if ((label|0) == 33) {
  $85 = $frm;
  $86 = (($$lcssa) - ($85))|0;
  $$37 = $86;
 }
 else if ((label|0) == 35) {
  $89 = $frm;
  $90 = (($$lcssa89) - ($89))|0;
  $$37 = $90;
 }
 else if ((label|0) == 40) {
  $$pre = $frm_nxt$1$lcssa;
  $$pre43 = $frm;
  $$pre45 = (($$pre) - ($$pre43))|0;
  $$37 = $$pre45;
 }
 return ($$37|0);
}
function __ZNKSt3__17codecvtIDic11__mbstate_tE13do_max_lengthEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 4;
}
function __ZNSt3__117__widen_from_utf8ILj32EED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__18messagesIwE8do_closeEi($this,$__c) {
 $this = $this|0;
 $__c = $__c|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNKSt3__17codecvtIcc11__mbstate_tE6do_outERS1_PKcS5_RS5_PcS7_RS7_($this,$0,$frm,$1,$frm_nxt,$to,$2,$to_nxt) {
 $this = $this|0;
 $0 = $0|0;
 $frm = $frm|0;
 $1 = $1|0;
 $frm_nxt = $frm_nxt|0;
 $to = $to|0;
 $2 = $2|0;
 $to_nxt = $to_nxt|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$frm_nxt>>2] = $frm;
 HEAP32[$to_nxt>>2] = $to;
 return 3;
}
function __ZNKSt3__17codecvtIcc11__mbstate_tE5do_inERS1_PKcS5_RS5_PcS7_RS7_($this,$0,$frm,$1,$frm_nxt,$to,$2,$to_nxt) {
 $this = $this|0;
 $0 = $0|0;
 $frm = $frm|0;
 $1 = $1|0;
 $frm_nxt = $frm_nxt|0;
 $to = $to|0;
 $2 = $2|0;
 $to_nxt = $to_nxt|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$frm_nxt>>2] = $frm;
 HEAP32[$to_nxt>>2] = $to;
 return 3;
}
function __ZNKSt3__17codecvtIcc11__mbstate_tE10do_unshiftERS1_PcS4_RS4_($this,$0,$to,$1,$to_nxt) {
 $this = $this|0;
 $0 = $0|0;
 $to = $to|0;
 $1 = $1|0;
 $to_nxt = $to_nxt|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$to_nxt>>2] = $to;
 return 3;
}
function __ZNKSt3__17codecvtIcc11__mbstate_tE11do_encodingEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1;
}
function __ZNKSt3__17codecvtIcc11__mbstate_tE16do_always_noconvEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1;
}
function __ZNKSt3__17codecvtIcc11__mbstate_tE9do_lengthERS1_PKcS5_j($this,$0,$frm,$end,$mx) {
 $this = $this|0;
 $0 = $0|0;
 $frm = $frm|0;
 $end = $end|0;
 $mx = $mx|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = $end;
 $2 = $frm;
 $3 = (($1) - ($2))|0;
 $4 = ($3>>>0)<($mx>>>0);
 $5 = $4 ? $3 : $mx;
 return ($5|0);
}
function __ZNKSt3__17codecvtIcc11__mbstate_tE13do_max_lengthEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1;
}
function __ZNSt3__17codecvtIwc11__mbstate_tED2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (157248);
 $0 = ((($this)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 __THREW__ = 0;
 $2 = (invoke_i(4874)|0);
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 do {
  if (!($4)) {
   $5 = ($1|0)==($2|0);
   if (!($5)) {
    $6 = HEAP32[$0>>2]|0;
    __THREW__ = 0;
    invoke_vi(4897,($6|0));
    $7 = __THREW__; __THREW__ = 0;
    $8 = $7&1;
    if ($8) {
     break;
    }
   }
   return;
  }
 } while(0);
 $9 = ___cxa_find_matching_catch_3(0|0)|0;
 $10 = tempRet0;
 ___clang_call_terminate($9);
 // unreachable;
}
function __ZNSt3__17codecvtIwc11__mbstate_tED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__17codecvtIwc11__mbstate_tED2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNKSt3__17codecvtIwc11__mbstate_tE6do_outERS1_PKwS5_RS5_PcS7_RS7_($this,$st,$frm,$frm_end,$frm_nxt,$to,$to_end,$to_nxt) {
 $this = $this|0;
 $st = $st|0;
 $frm = $frm|0;
 $frm_end = $frm_end|0;
 $frm_nxt = $frm_nxt|0;
 $to = $to|0;
 $to_end = $to_end|0;
 $to_nxt = $to_nxt|0;
 var $$03 = 0, $$03$lcssa201 = 0, $$05 = 0, $$05$lcssa193 = 0, $$12 = 0, $$14 = 0, $$14$lcssa = 0, $$3$ph = 0, $$4 = 0, $$lcssa = 0, $$lcssa207 = 0, $$lcssa215 = 0, $$pre = 0, $$pre96 = 0, $$pre97 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $fend$0 = 0, $fend$0$lcssa = 0, $fend$1 = 0, $fend$2 = 0, $fend$2$lcssa = 0, $fend$5$ph13 = 0, $n$0 = 0, $not$ = 0, $or$cond = 0, $p$0 = 0, $save_state = 0, $tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $save_state = sp;
 $tmp = sp + 8|0;
 $fend$0 = $frm;
 while(1) {
  $0 = ($fend$0|0)==($frm_end|0);
  if ($0) {
   $fend$0$lcssa = $frm_end;
   break;
  }
  $1 = HEAP32[$fend$0>>2]|0;
  $2 = ($1|0)==(0);
  if ($2) {
   $fend$0$lcssa = $fend$0;
   break;
  }
  $3 = ((($fend$0)) + 4|0);
  $fend$0 = $3;
 }
 HEAP32[$to_nxt>>2] = $to;
 HEAP32[$frm_nxt>>2] = $frm;
 $4 = $to_end;
 $5 = ((($this)) + 8|0);
 $$03 = $frm;$$05 = $to;$fend$1 = $fend$0$lcssa;
 L6: while(1) {
  $6 = ($$03|0)==($frm_end|0);
  $7 = ($$05|0)==($to_end|0);
  $or$cond = $7 | $6;
  if ($or$cond) {
   $103 = $$03;
   label = 53;
   break;
  }
  $8 = $st;
  $9 = $8;
  $10 = HEAP32[$9>>2]|0;
  $11 = (($8) + 4)|0;
  $12 = $11;
  $13 = HEAP32[$12>>2]|0;
  $14 = $save_state;
  $15 = $14;
  HEAP32[$15>>2] = $10;
  $16 = (($14) + 4)|0;
  $17 = $16;
  HEAP32[$17>>2] = $13;
  $18 = $fend$1;
  $19 = $$03;
  $20 = (($18) - ($19))|0;
  $21 = $20 >> 2;
  $22 = $$05;
  $23 = (($4) - ($22))|0;
  $24 = HEAP32[$5>>2]|0;
  $25 = (_uselocale($24)|0);
  __THREW__ = 0;
  $26 = (invoke_iiiiii(4917,($$05|0),($frm_nxt|0),($21|0),($23|0),($st|0))|0);
  $27 = __THREW__; __THREW__ = 0;
  $28 = $27&1;
  if ($28) {
   $$lcssa207 = $25;
   label = 11;
   break;
  }
  $29 = ($25|0)==(0|0);
  if (!($29)) {
   __THREW__ = 0;
   (invoke_ii(4876,($25|0))|0);
   $30 = __THREW__; __THREW__ = 0;
   $31 = $30&1;
   if ($31) {
    label = 10;
    break;
   }
  }
  switch ($26|0) {
  case -1:  {
   $$03$lcssa201 = $$03;$$05$lcssa193 = $$05;
   label = 16;
   break L6;
   break;
  }
  case 0:  {
   $$3$ph = 1;
   label = 50;
   break L6;
   break;
  }
  default: {
  }
  }
  $66 = HEAP32[$to_nxt>>2]|0;
  $67 = (($66) + ($26)|0);
  HEAP32[$to_nxt>>2] = $67;
  $68 = ($67|0)==($to_end|0);
  if ($68) {
   label = 51;
   break;
  }
  $69 = ($fend$1|0)==($frm_end|0);
  if ($69) {
   $$pre = HEAP32[$frm_nxt>>2]|0;
   $105 = $67;$106 = $$pre;$fend$5$ph13 = $frm_end;
  } else {
   $70 = HEAP32[$5>>2]|0;
   $71 = (_uselocale($70)|0);
   __THREW__ = 0;
   $72 = (invoke_iiii(4918,($tmp|0),0,($st|0))|0);
   $73 = __THREW__; __THREW__ = 0;
   $74 = $73&1;
   if ($74) {
    $$lcssa215 = $71;
    label = 36;
    break;
   }
   $75 = ($71|0)==(0|0);
   if (!($75)) {
    __THREW__ = 0;
    (invoke_ii(4876,($71|0))|0);
    $76 = __THREW__; __THREW__ = 0;
    $77 = $76&1;
    if ($77) {
     label = 35;
     break;
    }
   }
   $87 = ($72|0)==(-1);
   if ($87) {
    $$12 = 2;
    label = 49;
    break;
   }
   $88 = HEAP32[$to_nxt>>2]|0;
   $89 = (($4) - ($88))|0;
   $90 = ($72>>>0)>($89>>>0);
   if ($90) {
    $$12 = 1;
    label = 49;
    break;
   } else {
    $n$0 = $72;$p$0 = $tmp;
   }
   while(1) {
    $91 = ($n$0|0)==(0);
    if ($91) {
     break;
    }
    $94 = ((($p$0)) + 1|0);
    $95 = HEAP8[$p$0>>0]|0;
    $96 = HEAP32[$to_nxt>>2]|0;
    $97 = ((($96)) + 1|0);
    HEAP32[$to_nxt>>2] = $97;
    HEAP8[$96>>0] = $95;
    $98 = (($n$0) + -1)|0;
    $n$0 = $98;$p$0 = $94;
   }
   $92 = HEAP32[$frm_nxt>>2]|0;
   $93 = ((($92)) + 4|0);
   HEAP32[$frm_nxt>>2] = $93;
   $fend$2 = $93;
   while(1) {
    $99 = ($fend$2|0)==($frm_end|0);
    if ($99) {
     $fend$2$lcssa = $frm_end;
     break;
    }
    $100 = HEAP32[$fend$2>>2]|0;
    $101 = ($100|0)==(0);
    if ($101) {
     $fend$2$lcssa = $fend$2;
     break;
    }
    $102 = ((($fend$2)) + 4|0);
    $fend$2 = $102;
   }
   $$pre96 = HEAP32[$to_nxt>>2]|0;
   $105 = $$pre96;$106 = $93;$fend$5$ph13 = $fend$2$lcssa;
  }
  $$03 = $106;$$05 = $105;$fend$1 = $fend$5$ph13;
 }
 do {
  if ((label|0) == 10) {
   $32 = ___cxa_find_matching_catch_3(0|0)|0;
   $33 = tempRet0;
   ___clang_call_terminate($32);
   // unreachable;
  }
  else if ((label|0) == 11) {
   $34 = ___cxa_find_matching_catch_2()|0;
   $35 = tempRet0;
   $36 = ($$lcssa207|0)==(0|0);
   if (!($36)) {
    __THREW__ = 0;
    (invoke_ii(4876,($$lcssa207|0))|0);
    $37 = __THREW__; __THREW__ = 0;
    $38 = $37&1;
    if ($38) {
     $39 = ___cxa_find_matching_catch_3(0|0)|0;
     $40 = tempRet0;
     ___clang_call_terminate($39);
     // unreachable;
    }
   }
   ___resumeException($34|0);
   // unreachable;
  }
  else if ((label|0) == 16) {
   HEAP32[$to_nxt>>2] = $$05$lcssa193;
   $$14 = $$03$lcssa201;$46 = $$05$lcssa193;
   while(1) {
    $41 = HEAP32[$frm_nxt>>2]|0;
    $42 = ($$14|0)==($41|0);
    if ($42) {
     $$14$lcssa = $$14;
     label = 28;
     break;
    }
    $43 = HEAP32[$$14>>2]|0;
    $44 = HEAP32[$5>>2]|0;
    $45 = (_uselocale($44)|0);
    __THREW__ = 0;
    $47 = (invoke_iiii(4918,($46|0),($43|0),($save_state|0))|0);
    $48 = __THREW__; __THREW__ = 0;
    $49 = $48&1;
    if ($49) {
     $$lcssa = $45;
     label = 22;
     break;
    }
    $50 = ($45|0)==(0|0);
    if (!($50)) {
     __THREW__ = 0;
     (invoke_ii(4876,($45|0))|0);
     $51 = __THREW__; __THREW__ = 0;
     $52 = $51&1;
     if ($52) {
      label = 21;
      break;
     }
    }
    $62 = ($47|0)==(-1);
    if ($62) {
     $$14$lcssa = $$14;
     label = 28;
     break;
    }
    $63 = HEAP32[$to_nxt>>2]|0;
    $64 = (($63) + ($47)|0);
    HEAP32[$to_nxt>>2] = $64;
    $65 = ((($$14)) + 4|0);
    $$14 = $65;$46 = $64;
   }
   if ((label|0) == 21) {
    $53 = ___cxa_find_matching_catch_3(0|0)|0;
    $54 = tempRet0;
    ___clang_call_terminate($53);
    // unreachable;
   }
   else if ((label|0) == 22) {
    $55 = ___cxa_find_matching_catch_2()|0;
    $56 = tempRet0;
    $57 = ($$lcssa|0)==(0|0);
    if (!($57)) {
     __THREW__ = 0;
     (invoke_ii(4876,($$lcssa|0))|0);
     $58 = __THREW__; __THREW__ = 0;
     $59 = $58&1;
     if ($59) {
      $60 = ___cxa_find_matching_catch_3(0|0)|0;
      $61 = tempRet0;
      ___clang_call_terminate($60);
      // unreachable;
     }
    }
    ___resumeException($55|0);
    // unreachable;
   }
   else if ((label|0) == 28) {
    HEAP32[$frm_nxt>>2] = $$14$lcssa;
    $$3$ph = 2;
    label = 50;
    break;
   }
  }
  else if ((label|0) == 35) {
   $78 = ___cxa_find_matching_catch_3(0|0)|0;
   $79 = tempRet0;
   ___clang_call_terminate($78);
   // unreachable;
  }
  else if ((label|0) == 36) {
   $80 = ___cxa_find_matching_catch_2()|0;
   $81 = tempRet0;
   $82 = ($$lcssa215|0)==(0|0);
   if (!($82)) {
    __THREW__ = 0;
    (invoke_ii(4876,($$lcssa215|0))|0);
    $83 = __THREW__; __THREW__ = 0;
    $84 = $83&1;
    if ($84) {
     $85 = ___cxa_find_matching_catch_3(0|0)|0;
     $86 = tempRet0;
     ___clang_call_terminate($85);
     // unreachable;
    }
   }
   ___resumeException($80|0);
   // unreachable;
  }
  else if ((label|0) == 49) {
   $$3$ph = $$12;
   label = 50;
  }
  else if ((label|0) == 51) {
   $$pre97 = HEAP32[$frm_nxt>>2]|0;
   $103 = $$pre97;
   label = 53;
  }
 } while(0);
 if ((label|0) == 50) {
  $$4 = $$3$ph;
 }
 else if ((label|0) == 53) {
  $not$ = ($103|0)!=($frm_end|0);
  $104 = $not$&1;
  $$4 = $104;
 }
 STACKTOP = sp;return ($$4|0);
}
function __ZNKSt3__17codecvtIwc11__mbstate_tE5do_inERS1_PKcS5_RS5_PwS7_RS7_($this,$st,$frm,$frm_end,$frm_nxt,$to,$to_end,$to_nxt) {
 $this = $this|0;
 $st = $st|0;
 $frm = $frm|0;
 $frm_end = $frm_end|0;
 $frm_nxt = $frm_nxt|0;
 $to = $to|0;
 $to_end = $to_end|0;
 $to_nxt = $to_nxt|0;
 var $$02 = 0, $$02$lcssa231 = 0, $$05 = 0, $$05$lcssa224 = 0, $$1$ph = 0, $$13 = 0, $$13$lcssa = 0, $$13$lcssa217 = 0, $$13$lcssa218 = 0, $$2 = 0, $$24 = 0, $$lcssa = 0, $$lcssa238 = 0, $$lcssa243 = 0, $$lcssa250 = 0, $$pre = 0, $$pre102 = 0, $$pre103 = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $fend$0 = 0;
 var $fend$0$lcssa = 0, $fend$1 = 0, $fend$2 = 0, $fend$2$lcssa = 0, $fend$4$ph = 0, $not$ = 0, $not$6 = 0, $or$cond = 0, $save_state = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $save_state = sp;
 $fend$0 = $frm;
 while(1) {
  $0 = ($fend$0|0)==($frm_end|0);
  if ($0) {
   $fend$0$lcssa = $frm_end;
   break;
  }
  $1 = HEAP8[$fend$0>>0]|0;
  $2 = ($1<<24>>24)==(0);
  if ($2) {
   $fend$0$lcssa = $fend$0;
   break;
  }
  $3 = ((($fend$0)) + 1|0);
  $fend$0 = $3;
 }
 HEAP32[$to_nxt>>2] = $to;
 HEAP32[$frm_nxt>>2] = $frm;
 $4 = $to_end;
 $5 = ((($this)) + 8|0);
 $$02 = $frm;$$05 = $to;$fend$1 = $fend$0$lcssa;
 L6: while(1) {
  $6 = ($$02|0)==($frm_end|0);
  $7 = ($$05|0)==($to_end|0);
  $or$cond = $7 | $6;
  if ($or$cond) {
   $97 = $$02;
   label = 52;
   break;
  }
  $8 = $st;
  $9 = $8;
  $10 = HEAP32[$9>>2]|0;
  $11 = (($8) + 4)|0;
  $12 = $11;
  $13 = HEAP32[$12>>2]|0;
  $14 = $save_state;
  $15 = $14;
  HEAP32[$15>>2] = $10;
  $16 = (($14) + 4)|0;
  $17 = $16;
  HEAP32[$17>>2] = $13;
  $18 = $fend$1;
  $19 = $$02;
  $20 = (($18) - ($19))|0;
  $21 = $$05;
  $22 = (($4) - ($21))|0;
  $23 = $22 >> 2;
  $24 = HEAP32[$5>>2]|0;
  $25 = (_uselocale($24)|0);
  __THREW__ = 0;
  $26 = (invoke_iiiiii(4919,($$05|0),($frm_nxt|0),($20|0),($23|0),($st|0))|0);
  $27 = __THREW__; __THREW__ = 0;
  $28 = $27&1;
  if ($28) {
   $$lcssa243 = $25;
   label = 11;
   break;
  }
  $29 = ($25|0)==(0|0);
  if (!($29)) {
   __THREW__ = 0;
   (invoke_ii(4876,($25|0))|0);
   $30 = __THREW__; __THREW__ = 0;
   $31 = $30&1;
   if ($31) {
    label = 10;
    break;
   }
  }
  switch ($26|0) {
  case -1:  {
   $$02$lcssa231 = $$02;$$05$lcssa224 = $$05;$$lcssa238 = $18;
   label = 16;
   break L6;
   break;
  }
  case 0:  {
   $$1$ph = 2;
   label = 48;
   break L6;
   break;
  }
  default: {
  }
  }
  $67 = HEAP32[$to_nxt>>2]|0;
  $68 = (($67) + ($26<<2)|0);
  HEAP32[$to_nxt>>2] = $68;
  $69 = ($68|0)==($to_end|0);
  if ($69) {
   label = 49;
   break;
  }
  $70 = ($fend$1|0)==($frm_end|0);
  $$pre = HEAP32[$frm_nxt>>2]|0;
  if ($70) {
   $100 = $$pre;$99 = $68;$fend$4$ph = $frm_end;
  } else {
   $71 = HEAP32[$5>>2]|0;
   $72 = (_uselocale($71)|0);
   __THREW__ = 0;
   $73 = (invoke_iiiii(4920,($68|0),($$pre|0),1,($st|0))|0);
   $74 = __THREW__; __THREW__ = 0;
   $75 = $74&1;
   if ($75) {
    $$lcssa250 = $72;
    label = 39;
    break;
   }
   $76 = ($72|0)==(0|0);
   if (!($76)) {
    __THREW__ = 0;
    (invoke_ii(4876,($72|0))|0);
    $77 = __THREW__; __THREW__ = 0;
    $78 = $77&1;
    if ($78) {
     label = 38;
     break;
    }
   }
   $88 = ($73|0)==(0);
   if (!($88)) {
    $$1$ph = 2;
    label = 48;
    break;
   }
   $89 = HEAP32[$to_nxt>>2]|0;
   $90 = ((($89)) + 4|0);
   HEAP32[$to_nxt>>2] = $90;
   $91 = HEAP32[$frm_nxt>>2]|0;
   $92 = ((($91)) + 1|0);
   HEAP32[$frm_nxt>>2] = $92;
   $fend$2 = $92;
   while(1) {
    $93 = ($fend$2|0)==($frm_end|0);
    if ($93) {
     $fend$2$lcssa = $frm_end;
     break;
    }
    $94 = HEAP8[$fend$2>>0]|0;
    $95 = ($94<<24>>24)==(0);
    if ($95) {
     $fend$2$lcssa = $fend$2;
     break;
    }
    $96 = ((($fend$2)) + 1|0);
    $fend$2 = $96;
   }
   $$pre102 = HEAP32[$to_nxt>>2]|0;
   $100 = $92;$99 = $$pre102;$fend$4$ph = $fend$2$lcssa;
  }
  $$02 = $100;$$05 = $99;$fend$1 = $fend$4$ph;
 }
 do {
  if ((label|0) == 10) {
   $32 = ___cxa_find_matching_catch_3(0|0)|0;
   $33 = tempRet0;
   ___clang_call_terminate($32);
   // unreachable;
  }
  else if ((label|0) == 11) {
   $34 = ___cxa_find_matching_catch_2()|0;
   $35 = tempRet0;
   $36 = ($$lcssa243|0)==(0|0);
   if (!($36)) {
    __THREW__ = 0;
    (invoke_ii(4876,($$lcssa243|0))|0);
    $37 = __THREW__; __THREW__ = 0;
    $38 = $37&1;
    if ($38) {
     $39 = ___cxa_find_matching_catch_3(0|0)|0;
     $40 = tempRet0;
     ___clang_call_terminate($39);
     // unreachable;
    }
   }
   ___resumeException($34|0);
   // unreachable;
  }
  else if ((label|0) == 16) {
   $$13 = $$02$lcssa231;$storemerge = $$05$lcssa224;
   L31: while(1) {
    HEAP32[$to_nxt>>2] = $storemerge;
    $41 = HEAP32[$frm_nxt>>2]|0;
    $42 = ($$13|0)==($41|0);
    if ($42) {
     $$13$lcssa = $$13;
     label = 32;
     break;
    }
    $43 = $$13;
    $44 = (($$lcssa238) - ($43))|0;
    $45 = HEAP32[$5>>2]|0;
    $46 = (_uselocale($45)|0);
    __THREW__ = 0;
    $47 = (invoke_iiiii(4920,($storemerge|0),($$13|0),($44|0),($save_state|0))|0);
    $48 = __THREW__; __THREW__ = 0;
    $49 = $48&1;
    if ($49) {
     $$lcssa = $46;
     label = 22;
     break;
    }
    $50 = ($46|0)==(0|0);
    if (!($50)) {
     __THREW__ = 0;
     (invoke_ii(4876,($46|0))|0);
     $51 = __THREW__; __THREW__ = 0;
     $52 = $51&1;
     if ($52) {
      label = 21;
      break;
     }
    }
    switch ($47|0) {
    case -1:  {
     $$13$lcssa217 = $$13;
     label = 28;
     break L31;
     break;
    }
    case -2:  {
     $$13$lcssa218 = $$13;
     label = 29;
     break L31;
     break;
    }
    case 0:  {
     $62 = ((($$13)) + 1|0);
     $$24 = $62;
     break;
    }
    default: {
     $63 = (($$13) + ($47)|0);
     $$24 = $63;
    }
    }
    $64 = HEAP32[$to_nxt>>2]|0;
    $65 = ((($64)) + 4|0);
    $$13 = $$24;$storemerge = $65;
   }
   if ((label|0) == 21) {
    $53 = ___cxa_find_matching_catch_3(0|0)|0;
    $54 = tempRet0;
    ___clang_call_terminate($53);
    // unreachable;
   }
   else if ((label|0) == 22) {
    $55 = ___cxa_find_matching_catch_2()|0;
    $56 = tempRet0;
    $57 = ($$lcssa|0)==(0|0);
    if (!($57)) {
     __THREW__ = 0;
     (invoke_ii(4876,($$lcssa|0))|0);
     $58 = __THREW__; __THREW__ = 0;
     $59 = $58&1;
     if ($59) {
      $60 = ___cxa_find_matching_catch_3(0|0)|0;
      $61 = tempRet0;
      ___clang_call_terminate($60);
      // unreachable;
     }
    }
    ___resumeException($55|0);
    // unreachable;
   }
   else if ((label|0) == 28) {
    HEAP32[$frm_nxt>>2] = $$13$lcssa217;
    $$1$ph = 2;
    label = 48;
    break;
   }
   else if ((label|0) == 29) {
    HEAP32[$frm_nxt>>2] = $$13$lcssa218;
    $$1$ph = 1;
    label = 48;
    break;
   }
   else if ((label|0) == 32) {
    HEAP32[$frm_nxt>>2] = $$13$lcssa;
    $not$6 = ($$13$lcssa|0)!=($frm_end|0);
    $66 = $not$6&1;
    $$1$ph = $66;
    label = 48;
    break;
   }
  }
  else if ((label|0) == 38) {
   $79 = ___cxa_find_matching_catch_3(0|0)|0;
   $80 = tempRet0;
   ___clang_call_terminate($79);
   // unreachable;
  }
  else if ((label|0) == 39) {
   $81 = ___cxa_find_matching_catch_2()|0;
   $82 = tempRet0;
   $83 = ($$lcssa250|0)==(0|0);
   if (!($83)) {
    __THREW__ = 0;
    (invoke_ii(4876,($$lcssa250|0))|0);
    $84 = __THREW__; __THREW__ = 0;
    $85 = $84&1;
    if ($85) {
     $86 = ___cxa_find_matching_catch_3(0|0)|0;
     $87 = tempRet0;
     ___clang_call_terminate($86);
     // unreachable;
    }
   }
   ___resumeException($81|0);
   // unreachable;
  }
  else if ((label|0) == 49) {
   $$pre103 = HEAP32[$frm_nxt>>2]|0;
   $97 = $$pre103;
   label = 52;
  }
 } while(0);
 if ((label|0) == 48) {
  $$2 = $$1$ph;
 }
 else if ((label|0) == 52) {
  $not$ = ($97|0)!=($frm_end|0);
  $98 = $not$&1;
  $$2 = $98;
 }
 STACKTOP = sp;return ($$2|0);
}
function __ZNKSt3__17codecvtIwc11__mbstate_tE10do_unshiftERS1_PcS4_RS4_($this,$st,$to,$to_end,$to_nxt) {
 $this = $this|0;
 $st = $st|0;
 $to = $to|0;
 $to_end = $to_end|0;
 $to_nxt = $to_nxt|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $n$0 = 0, $p$0 = 0, $tmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $tmp = sp;
 HEAP32[$to_nxt>>2] = $to;
 $0 = ((($this)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_uselocale($1)|0);
 __THREW__ = 0;
 $3 = (invoke_iiii(4918,($tmp|0),0,($st|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $11 = ___cxa_find_matching_catch_2()|0;
  $12 = tempRet0;
  $13 = ($2|0)==(0|0);
  if (!($13)) {
   __THREW__ = 0;
   (invoke_ii(4876,($2|0))|0);
   $14 = __THREW__; __THREW__ = 0;
   $15 = $14&1;
   if ($15) {
    $16 = ___cxa_find_matching_catch_3(0|0)|0;
    $17 = tempRet0;
    ___clang_call_terminate($16);
    // unreachable;
   }
  }
  ___resumeException($11|0);
  // unreachable;
 }
 $6 = ($2|0)==(0|0);
 if (!($6)) {
  __THREW__ = 0;
  (invoke_ii(4876,($2|0))|0);
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if ($8) {
   $9 = ___cxa_find_matching_catch_3(0|0)|0;
   $10 = tempRet0;
   ___clang_call_terminate($9);
   // unreachable;
  }
 }
 L12: do {
  switch ($3|0) {
  case 0: case -1:  {
   $$0 = 2;
   break;
  }
  default: {
   $18 = (($3) + -1)|0;
   $19 = HEAP32[$to_nxt>>2]|0;
   $20 = $to_end;
   $21 = (($20) - ($19))|0;
   $22 = ($18>>>0)>($21>>>0);
   if ($22) {
    $$0 = 1;
   } else {
    $n$0 = $18;$p$0 = $tmp;
    while(1) {
     $23 = ($n$0|0)==(0);
     if ($23) {
      $$0 = 0;
      break L12;
     }
     $24 = ((($p$0)) + 1|0);
     $25 = HEAP8[$p$0>>0]|0;
     $26 = HEAP32[$to_nxt>>2]|0;
     $27 = ((($26)) + 1|0);
     HEAP32[$to_nxt>>2] = $27;
     HEAP8[$26>>0] = $25;
     $28 = (($n$0) + -1)|0;
     $n$0 = $28;$p$0 = $24;
    }
   }
  }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function __ZNKSt3__17codecvtIwc11__mbstate_tE11do_encodingEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$indexZ2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 __THREW__ = 0;
 $2 = (invoke_ii(4876,($1|0))|0);
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 do {
  if ($4) {
   label = 16;
  } else {
   __THREW__ = 0;
   $5 = (invoke_iiii(4921,(0|0),(0|0),4)|0);
   $6 = __THREW__; __THREW__ = 0;
   $7 = $6&1;
   if ($7) {
    $13 = ___cxa_find_matching_catch_3(0|0)|0;
    $14 = tempRet0;
    $15 = ($2|0)==(0|0);
    if ($15) {
     $eh$lpad$body$indexZ2D = $13;
     break;
    }
    __THREW__ = 0;
    (invoke_ii(4876,($2|0))|0);
    $16 = __THREW__; __THREW__ = 0;
    $17 = $16&1;
    if (!($17)) {
     $eh$lpad$body$indexZ2D = $13;
     break;
    }
    $18 = ___cxa_find_matching_catch_3(0|0)|0;
    $19 = tempRet0;
    ___clang_call_terminate($18);
    // unreachable;
   }
   $8 = ($2|0)==(0|0);
   if (!($8)) {
    __THREW__ = 0;
    (invoke_ii(4876,($2|0))|0);
    $9 = __THREW__; __THREW__ = 0;
    $10 = $9&1;
    if ($10) {
     $11 = ___cxa_find_matching_catch_3(0|0)|0;
     $12 = tempRet0;
     ___clang_call_terminate($11);
     // unreachable;
    }
   }
   $20 = ($5|0)==(0);
   if ($20) {
    $21 = HEAP32[$0>>2]|0;
    $22 = ($21|0)==(0|0);
    if ($22) {
     $$0 = 1;
    } else {
     __THREW__ = 0;
     $23 = (invoke_ii(4876,($21|0))|0);
     $24 = __THREW__; __THREW__ = 0;
     $25 = $24&1;
     if ($25) {
      label = 16;
      break;
     }
     $26 = ($23|0)==(0|0);
     if ($26) {
      $$0 = 0;
     } else {
      __THREW__ = 0;
      (invoke_ii(4876,($23|0))|0);
      $27 = __THREW__; __THREW__ = 0;
      $28 = $27&1;
      if ($28) {
       $29 = ___cxa_find_matching_catch_3(0|0)|0;
       $30 = tempRet0;
       ___clang_call_terminate($29);
       // unreachable;
      } else {
       $$0 = 0;
      }
     }
    }
   } else {
    $$0 = -1;
   }
   return ($$0|0);
  }
 } while(0);
 if ((label|0) == 16) {
  $31 = ___cxa_find_matching_catch_3(0|0)|0;
  $32 = tempRet0;
  $eh$lpad$body$indexZ2D = $31;
 }
 ___clang_call_terminate($eh$lpad$body$indexZ2D);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__17codecvtIwc11__mbstate_tE16do_always_noconvEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__17codecvtIwc11__mbstate_tE9do_lengthERS1_PKcS5_j($this,$st,$frm,$frm_end,$mx) {
 $this = $this|0;
 $st = $st|0;
 $frm = $frm|0;
 $frm_end = $frm_end|0;
 $mx = $mx|0;
 var $$04 = 0, $$15 = 0, $$lcssa = 0, $$pn = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $nbytes$0 = 0, $nbytes$0$lcssa = 0, $nbytes$1 = 0, $nwchar_t$0 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $frm_end;
 $1 = ((($this)) + 8|0);
 $$04 = $frm;$nbytes$0 = 0;$nwchar_t$0 = 0;
 L1: while(1) {
  $2 = ($nwchar_t$0>>>0)>=($mx>>>0);
  $3 = ($$04|0)==($frm_end|0);
  $or$cond = $3 | $2;
  if ($or$cond) {
   $nbytes$0$lcssa = $nbytes$0;
   label = 15;
   break;
  }
  $4 = $$04;
  $5 = (($0) - ($4))|0;
  $6 = HEAP32[$1>>2]|0;
  $7 = (_uselocale($6)|0);
  __THREW__ = 0;
  $8 = (invoke_iiii(4922,($$04|0),($5|0),($st|0))|0);
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if ($10) {
   $$lcssa = $7;
   label = 7;
   break;
  }
  $11 = ($7|0)==(0|0);
  if (!($11)) {
   __THREW__ = 0;
   (invoke_ii(4876,($7|0))|0);
   $12 = __THREW__; __THREW__ = 0;
   $13 = $12&1;
   if ($13) {
    label = 6;
    break;
   }
  }
  switch ($8|0) {
  case -2: case -1:  {
   $nbytes$0$lcssa = $nbytes$0;
   label = 15;
   break L1;
   break;
  }
  case 0:  {
   $23 = ((($$04)) + 1|0);
   $$15 = $23;$$pn = 1;
   break;
  }
  default: {
   $24 = (($$04) + ($8)|0);
   $$15 = $24;$$pn = $8;
  }
  }
  $nbytes$1 = (($$pn) + ($nbytes$0))|0;
  $25 = (($nwchar_t$0) + 1)|0;
  $$04 = $$15;$nbytes$0 = $nbytes$1;$nwchar_t$0 = $25;
 }
 if ((label|0) == 6) {
  $14 = ___cxa_find_matching_catch_3(0|0)|0;
  $15 = tempRet0;
  ___clang_call_terminate($14);
  // unreachable;
 }
 else if ((label|0) == 7) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  $18 = ($$lcssa|0)==(0|0);
  if (!($18)) {
   __THREW__ = 0;
   (invoke_ii(4876,($$lcssa|0))|0);
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   if ($20) {
    $21 = ___cxa_find_matching_catch_3(0|0)|0;
    $22 = tempRet0;
    ___clang_call_terminate($21);
    // unreachable;
   }
  }
  ___resumeException($16|0);
  // unreachable;
 }
 else if ((label|0) == 15) {
  return ($nbytes$0$lcssa|0);
 }
 return (0)|0;
}
function __ZNKSt3__17codecvtIwc11__mbstate_tE13do_max_lengthEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $11 = 1;
 } else {
  __THREW__ = 0;
  $3 = (invoke_ii(4876,($1|0))|0);
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $12 = ___cxa_find_matching_catch_3(0|0)|0;
   $13 = tempRet0;
   ___clang_call_terminate($12);
   // unreachable;
  }
  $6 = ($3|0)==(0|0);
  if ($6) {
   $11 = 4;
  } else {
   __THREW__ = 0;
   (invoke_ii(4876,($3|0))|0);
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if ($8) {
    $9 = ___cxa_find_matching_catch_3(0|0)|0;
    $10 = tempRet0;
    ___clang_call_terminate($9);
    // unreachable;
   } else {
    $11 = 4;
   }
  }
 }
 return ($11|0);
}
function __ZNKSt3__17codecvtIDsc11__mbstate_tE6do_outERS1_PKDsS5_RS5_PcS7_RS7_($this,$0,$frm,$frm_end,$frm_nxt,$to,$to_end,$to_nxt) {
 $this = $this|0;
 $0 = $0|0;
 $frm = $frm|0;
 $frm_end = $frm_end|0;
 $frm_nxt = $frm_nxt|0;
 $to = $to|0;
 $to_end = $to_end|0;
 $to_nxt = $to_nxt|0;
 var $1 = 0, $2 = 0, $3 = 0, $_frm_nxt = 0, $_to_nxt = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $_frm_nxt = sp + 4|0;
 $_to_nxt = sp;
 HEAP32[$_frm_nxt>>2] = $frm;
 HEAP32[$_to_nxt>>2] = $to;
 $1 = (__ZNSt3__1L13utf16_to_utf8EPKtS1_RS1_PhS3_RS3_mNS_12codecvt_modeE($frm,$frm_end,$_frm_nxt,$to,$to_end,$_to_nxt,1114111,0)|0);
 $2 = HEAP32[$_frm_nxt>>2]|0;
 HEAP32[$frm_nxt>>2] = $2;
 $3 = HEAP32[$_to_nxt>>2]|0;
 HEAP32[$to_nxt>>2] = $3;
 STACKTOP = sp;return ($1|0);
}
function __ZNSt3__1L13utf16_to_utf8EPKtS1_RS1_PhS3_RS3_mNS_12codecvt_modeE($frm,$frm_end,$frm_nxt,$to,$to_end,$to_nxt,$Maxcode,$mode) {
 $frm = $frm|0;
 $frm_end = $frm_end|0;
 $frm_nxt = $frm_nxt|0;
 $to = $to|0;
 $to_end = $to_end|0;
 $to_nxt = $to_nxt|0;
 $Maxcode = $Maxcode|0;
 $mode = $mode|0;
 var $$4 = 0, $$pre = 0, $$pre7 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$frm_nxt>>2] = $frm;
 HEAP32[$to_nxt>>2] = $to;
 $0 = $mode & 2;
 $1 = ($0|0)==(0);
 $$pre7 = $to_end;
 if ($1) {
  label = 4;
 } else {
  $2 = $to;
  $3 = (($$pre7) - ($2))|0;
  $4 = ($3|0)<(3);
  if ($4) {
   $$4 = 1;
  } else {
   $5 = ((($to)) + 1|0);
   HEAP32[$to_nxt>>2] = $5;
   HEAP8[$to>>0] = -17;
   $6 = HEAP32[$to_nxt>>2]|0;
   $7 = ((($6)) + 1|0);
   HEAP32[$to_nxt>>2] = $7;
   HEAP8[$6>>0] = -69;
   $8 = HEAP32[$to_nxt>>2]|0;
   $9 = ((($8)) + 1|0);
   HEAP32[$to_nxt>>2] = $9;
   HEAP8[$8>>0] = -65;
   label = 4;
  }
 }
 L4: do {
  if ((label|0) == 4) {
   $10 = $frm_end;
   $$pre = HEAP32[$frm_nxt>>2]|0;
   $12 = $$pre;
   while(1) {
    $11 = ($12>>>0)<($frm_end>>>0);
    if (!($11)) {
     $$4 = 0;
     break L4;
    }
    $13 = HEAP16[$12>>1]|0;
    $14 = $13&65535;
    $15 = ($14>>>0)>($Maxcode>>>0);
    if ($15) {
     $$4 = 2;
     break L4;
    }
    $16 = ($13&65535)<(128);
    do {
     if ($16) {
      $17 = HEAP32[$to_nxt>>2]|0;
      $18 = $17;
      $19 = (($$pre7) - ($18))|0;
      $20 = ($19|0)<(1);
      if ($20) {
       $$4 = 1;
       break L4;
      }
      $21 = $13&255;
      $22 = ((($17)) + 1|0);
      HEAP32[$to_nxt>>2] = $22;
      HEAP8[$17>>0] = $21;
     } else {
      $23 = ($13&65535)<(2048);
      if ($23) {
       $24 = HEAP32[$to_nxt>>2]|0;
       $25 = $24;
       $26 = (($$pre7) - ($25))|0;
       $27 = ($26|0)<(2);
       if ($27) {
        $$4 = 1;
        break L4;
       }
       $28 = $14 >>> 6;
       $29 = $28 | 192;
       $30 = $29&255;
       $31 = ((($24)) + 1|0);
       HEAP32[$to_nxt>>2] = $31;
       HEAP8[$24>>0] = $30;
       $32 = $14 & 63;
       $33 = $32 | 128;
       $34 = $33&255;
       $35 = HEAP32[$to_nxt>>2]|0;
       $36 = ((($35)) + 1|0);
       HEAP32[$to_nxt>>2] = $36;
       HEAP8[$35>>0] = $34;
       break;
      }
      $37 = ($13&65535)<(55296);
      if ($37) {
       $38 = HEAP32[$to_nxt>>2]|0;
       $39 = $38;
       $40 = (($$pre7) - ($39))|0;
       $41 = ($40|0)<(3);
       if ($41) {
        $$4 = 1;
        break L4;
       }
       $42 = $14 >>> 12;
       $43 = $42 | 224;
       $44 = $43&255;
       $45 = ((($38)) + 1|0);
       HEAP32[$to_nxt>>2] = $45;
       HEAP8[$38>>0] = $44;
       $46 = $14 >>> 6;
       $47 = $46 & 63;
       $48 = $47 | 128;
       $49 = $48&255;
       $50 = HEAP32[$to_nxt>>2]|0;
       $51 = ((($50)) + 1|0);
       HEAP32[$to_nxt>>2] = $51;
       HEAP8[$50>>0] = $49;
       $52 = $14 & 63;
       $53 = $52 | 128;
       $54 = $53&255;
       $55 = HEAP32[$to_nxt>>2]|0;
       $56 = ((($55)) + 1|0);
       HEAP32[$to_nxt>>2] = $56;
       HEAP8[$55>>0] = $54;
       break;
      }
      $57 = ($13&65535)<(56320);
      if (!($57)) {
       $108 = ($13&65535)<(57344);
       if ($108) {
        $$4 = 2;
        break L4;
       }
       $109 = HEAP32[$to_nxt>>2]|0;
       $110 = $109;
       $111 = (($$pre7) - ($110))|0;
       $112 = ($111|0)<(3);
       if ($112) {
        $$4 = 1;
        break L4;
       }
       $113 = $14 >>> 12;
       $114 = $113 | 224;
       $115 = $114&255;
       $116 = ((($109)) + 1|0);
       HEAP32[$to_nxt>>2] = $116;
       HEAP8[$109>>0] = $115;
       $117 = $14 >>> 6;
       $118 = $117 & 63;
       $119 = $118 | 128;
       $120 = $119&255;
       $121 = HEAP32[$to_nxt>>2]|0;
       $122 = ((($121)) + 1|0);
       HEAP32[$to_nxt>>2] = $122;
       HEAP8[$121>>0] = $120;
       $123 = $14 & 63;
       $124 = $123 | 128;
       $125 = $124&255;
       $126 = HEAP32[$to_nxt>>2]|0;
       $127 = ((($126)) + 1|0);
       HEAP32[$to_nxt>>2] = $127;
       HEAP8[$126>>0] = $125;
       break;
      }
      $58 = $12;
      $59 = (($10) - ($58))|0;
      $60 = ($59|0)<(4);
      if ($60) {
       $$4 = 1;
       break L4;
      }
      $61 = ((($12)) + 2|0);
      $62 = HEAP16[$61>>1]|0;
      $63 = $62&65535;
      $64 = $63 & 64512;
      $65 = ($64|0)==(56320);
      if (!($65)) {
       $$4 = 2;
       break L4;
      }
      $66 = HEAP32[$to_nxt>>2]|0;
      $67 = (($$pre7) - ($66))|0;
      $68 = ($67|0)<(4);
      if ($68) {
       $$4 = 1;
       break L4;
      }
      $69 = $14 & 960;
      $70 = $69 << 10;
      $71 = (($70) + 65536)|0;
      $72 = $14 << 10;
      $73 = $72 & 64512;
      $74 = $71 | $73;
      $75 = $63 & 1023;
      $76 = $74 | $75;
      $77 = ($76>>>0)>($Maxcode>>>0);
      if ($77) {
       $$4 = 2;
       break L4;
      }
      HEAP32[$frm_nxt>>2] = $61;
      $78 = $69 >>> 6;
      $79 = (($78) + 1)|0;
      $80 = $79 >>> 2;
      $81 = $80 | 240;
      $82 = $81&255;
      $83 = HEAP32[$to_nxt>>2]|0;
      $84 = ((($83)) + 1|0);
      HEAP32[$to_nxt>>2] = $84;
      HEAP8[$83>>0] = $82;
      $85 = $79 << 4;
      $86 = $85 & 48;
      $87 = $14 >>> 2;
      $88 = $87 & 15;
      $89 = $88 | $86;
      $90 = $89 | 128;
      $91 = $90&255;
      $92 = HEAP32[$to_nxt>>2]|0;
      $93 = ((($92)) + 1|0);
      HEAP32[$to_nxt>>2] = $93;
      HEAP8[$92>>0] = $91;
      $94 = $14 << 4;
      $95 = $94 & 48;
      $96 = $63 >>> 6;
      $97 = $96 & 15;
      $98 = $95 | $97;
      $99 = $98 | 128;
      $100 = $99&255;
      $101 = HEAP32[$to_nxt>>2]|0;
      $102 = ((($101)) + 1|0);
      HEAP32[$to_nxt>>2] = $102;
      HEAP8[$101>>0] = $100;
      $103 = $63 & 63;
      $104 = $103 | 128;
      $105 = $104&255;
      $106 = HEAP32[$to_nxt>>2]|0;
      $107 = ((($106)) + 1|0);
      HEAP32[$to_nxt>>2] = $107;
      HEAP8[$106>>0] = $105;
     }
    } while(0);
    $128 = HEAP32[$frm_nxt>>2]|0;
    $129 = ((($128)) + 2|0);
    HEAP32[$frm_nxt>>2] = $129;
    $12 = $129;
   }
  }
 } while(0);
 return ($$4|0);
}
function __ZNKSt3__17codecvtIDsc11__mbstate_tE5do_inERS1_PKcS5_RS5_PDsS7_RS7_($this,$0,$frm,$frm_end,$frm_nxt,$to,$to_end,$to_nxt) {
 $this = $this|0;
 $0 = $0|0;
 $frm = $frm|0;
 $frm_end = $frm_end|0;
 $frm_nxt = $frm_nxt|0;
 $to = $to|0;
 $to_end = $to_end|0;
 $to_nxt = $to_nxt|0;
 var $1 = 0, $2 = 0, $3 = 0, $_frm_nxt = 0, $_to_nxt = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $_frm_nxt = sp + 4|0;
 $_to_nxt = sp;
 HEAP32[$_frm_nxt>>2] = $frm;
 HEAP32[$_to_nxt>>2] = $to;
 $1 = (__ZNSt3__1L13utf8_to_utf16EPKhS1_RS1_PtS3_RS3_mNS_12codecvt_modeE($frm,$frm_end,$_frm_nxt,$to,$to_end,$_to_nxt,1114111,0)|0);
 $2 = HEAP32[$_frm_nxt>>2]|0;
 HEAP32[$frm_nxt>>2] = $2;
 $3 = HEAP32[$_to_nxt>>2]|0;
 HEAP32[$to_nxt>>2] = $3;
 STACKTOP = sp;return ($1|0);
}
function __ZNSt3__1L13utf8_to_utf16EPKhS1_RS1_PtS3_RS3_mNS_12codecvt_modeE($frm,$frm_end,$frm_nxt,$to,$to_end,$to_nxt,$Maxcode,$mode) {
 $frm = $frm|0;
 $frm_end = $frm_end|0;
 $frm_nxt = $frm_nxt|0;
 $to = $to|0;
 $to_end = $to_end|0;
 $to_nxt = $to_nxt|0;
 $Maxcode = $Maxcode|0;
 $mode = $mode|0;
 var $$8 = 0, $$lcssa = 0, $$off = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$frm_nxt>>2] = $frm;
 HEAP32[$to_nxt>>2] = $to;
 $0 = $mode & 4;
 $1 = ($0|0)==(0);
 if ($1) {
  $$pre = $frm_end;
  $$pre$phiZ2D = $$pre;
 } else {
  $2 = HEAP32[$frm_nxt>>2]|0;
  $3 = $frm_end;
  $4 = $2;
  $5 = (($3) - ($4))|0;
  $6 = ($5|0)>(2);
  if ($6) {
   $7 = HEAP8[$2>>0]|0;
   $8 = ($7<<24>>24)==(-17);
   if ($8) {
    $9 = ((($2)) + 1|0);
    $10 = HEAP8[$9>>0]|0;
    $11 = ($10<<24>>24)==(-69);
    if ($11) {
     $12 = ((($2)) + 2|0);
     $13 = HEAP8[$12>>0]|0;
     $14 = ($13<<24>>24)==(-65);
     if ($14) {
      $15 = ((($2)) + 3|0);
      HEAP32[$frm_nxt>>2] = $15;
      $$pre$phiZ2D = $3;
     } else {
      $$pre$phiZ2D = $3;
     }
    } else {
     $$pre$phiZ2D = $3;
    }
   } else {
    $$pre$phiZ2D = $3;
   }
  } else {
   $$pre$phiZ2D = $3;
  }
 }
 $16 = $to_end;
 L9: while(1) {
  $17 = HEAP32[$frm_nxt>>2]|0;
  $18 = ($17>>>0)<($frm_end>>>0);
  if (!($18)) {
   $$lcssa = $18;
   label = 41;
   break;
  }
  $19 = HEAP32[$to_nxt>>2]|0;
  $20 = ($19>>>0)<($to_end>>>0);
  if (!($20)) {
   $$lcssa = $18;
   label = 41;
   break;
  }
  $21 = HEAP8[$17>>0]|0;
  $22 = $21&255;
  $23 = ($22>>>0)>($Maxcode>>>0);
  if ($23) {
   $$8 = 2;
   break;
  }
  $24 = ($21<<24>>24)>(-1);
  do {
   if ($24) {
    $25 = $21&255;
    HEAP16[$19>>1] = $25;
    $26 = ((($17)) + 1|0);
    HEAP32[$frm_nxt>>2] = $26;
   } else {
    $27 = ($21&255)<(194);
    if ($27) {
     $$8 = 2;
     break L9;
    }
    $28 = ($21&255)<(224);
    if ($28) {
     $29 = $17;
     $30 = (($$pre$phiZ2D) - ($29))|0;
     $31 = ($30|0)<(2);
     if ($31) {
      $$8 = 1;
      break L9;
     }
     $32 = ((($17)) + 1|0);
     $33 = HEAP8[$32>>0]|0;
     $34 = $33&255;
     $35 = $34 & 192;
     $36 = ($35|0)==(128);
     if (!($36)) {
      $$8 = 2;
      break L9;
     }
     $37 = $22 << 6;
     $38 = $37 & 1984;
     $39 = $34 & 63;
     $40 = $39 | $38;
     $41 = ($40>>>0)>($Maxcode>>>0);
     if ($41) {
      $$8 = 2;
      break L9;
     }
     $42 = $40&65535;
     HEAP16[$19>>1] = $42;
     $43 = ((($17)) + 2|0);
     HEAP32[$frm_nxt>>2] = $43;
     break;
    }
    $44 = ($21&255)<(240);
    if ($44) {
     $45 = $17;
     $46 = (($$pre$phiZ2D) - ($45))|0;
     $47 = ($46|0)<(3);
     if ($47) {
      $$8 = 1;
      break L9;
     }
     $48 = ((($17)) + 1|0);
     $49 = HEAP8[$48>>0]|0;
     $50 = ((($17)) + 2|0);
     $51 = HEAP8[$50>>0]|0;
     switch ($22|0) {
     case 224:  {
      $52 = $49 & -32;
      $53 = ($52<<24>>24)==(-96);
      if (!($53)) {
       $$8 = 2;
       break L9;
      }
      break;
     }
     case 237:  {
      $54 = $49 & -32;
      $55 = ($54<<24>>24)==(-128);
      if (!($55)) {
       $$8 = 2;
       break L9;
      }
      break;
     }
     default: {
      $56 = $49 & -64;
      $57 = ($56<<24>>24)==(-128);
      if (!($57)) {
       $$8 = 2;
       break L9;
      }
     }
     }
     $58 = $51&255;
     $59 = $58 & 192;
     $60 = ($59|0)==(128);
     if (!($60)) {
      $$8 = 2;
      break L9;
     }
     $61 = $22 << 12;
     $62 = $49&255;
     $63 = $62 << 6;
     $64 = $63 & 4032;
     $65 = $64 | $61;
     $66 = $58 & 63;
     $67 = $65 | $66;
     $68 = $67 & 65535;
     $69 = ($68>>>0)>($Maxcode>>>0);
     if ($69) {
      $$8 = 2;
      break L9;
     }
     $70 = $67&65535;
     HEAP16[$19>>1] = $70;
     $71 = ((($17)) + 3|0);
     HEAP32[$frm_nxt>>2] = $71;
     break;
    }
    $72 = ($21&255)<(245);
    if (!($72)) {
     $$8 = 2;
     break L9;
    }
    $73 = $17;
    $74 = (($$pre$phiZ2D) - ($73))|0;
    $75 = ($74|0)<(4);
    if ($75) {
     $$8 = 1;
     break L9;
    }
    $76 = ((($17)) + 1|0);
    $77 = HEAP8[$76>>0]|0;
    $78 = ((($17)) + 2|0);
    $79 = HEAP8[$78>>0]|0;
    $80 = ((($17)) + 3|0);
    $81 = HEAP8[$80>>0]|0;
    switch ($22|0) {
    case 240:  {
     $$off = (($77) + 112)<<24>>24;
     $82 = ($$off&255)<(48);
     if (!($82)) {
      $$8 = 2;
      break L9;
     }
     break;
    }
    case 244:  {
     $83 = $77 & -16;
     $84 = ($83<<24>>24)==(-128);
     if (!($84)) {
      $$8 = 2;
      break L9;
     }
     break;
    }
    default: {
     $85 = $77 & -64;
     $86 = ($85<<24>>24)==(-128);
     if (!($86)) {
      $$8 = 2;
      break L9;
     }
    }
    }
    $87 = $79&255;
    $88 = $87 & 192;
    $89 = ($88|0)==(128);
    if (!($89)) {
     $$8 = 2;
     break L9;
    }
    $90 = $81&255;
    $91 = $90 & 192;
    $92 = ($91|0)==(128);
    if (!($92)) {
     $$8 = 2;
     break L9;
    }
    $93 = $19;
    $94 = (($16) - ($93))|0;
    $95 = ($94|0)<(4);
    if ($95) {
     $$8 = 1;
     break L9;
    }
    $96 = $22 & 7;
    $97 = $96 << 18;
    $98 = $77&255;
    $99 = $98 << 12;
    $100 = $99 & 258048;
    $101 = $100 | $97;
    $102 = $87 << 6;
    $103 = $102 & 4032;
    $104 = $101 | $103;
    $105 = $90 & 63;
    $106 = $104 | $105;
    $107 = ($106>>>0)>($Maxcode>>>0);
    if ($107) {
     $$8 = 2;
     break L9;
    }
    $108 = $96 << 2;
    $109 = $98 >>> 4;
    $110 = $109 & 3;
    $111 = $110 | $108;
    $112 = $111 << 6;
    $113 = (($112) + 16320)|0;
    $114 = $98 << 2;
    $115 = $114 & 60;
    $116 = $87 >>> 4;
    $117 = $116 & 3;
    $118 = $115 | $117;
    $119 = $118 | $113;
    $120 = $119 | 55296;
    $121 = $120&65535;
    HEAP16[$19>>1] = $121;
    $122 = $102 & 960;
    $123 = $105 | $122;
    $124 = $123 | 56320;
    $125 = $124&65535;
    $126 = ((($19)) + 2|0);
    HEAP32[$to_nxt>>2] = $126;
    HEAP16[$126>>1] = $125;
    $127 = HEAP32[$frm_nxt>>2]|0;
    $128 = ((($127)) + 4|0);
    HEAP32[$frm_nxt>>2] = $128;
   }
  } while(0);
  $129 = HEAP32[$to_nxt>>2]|0;
  $130 = ((($129)) + 2|0);
  HEAP32[$to_nxt>>2] = $130;
 }
 if ((label|0) == 41) {
  $131 = $$lcssa&1;
  $$8 = $131;
 }
 return ($$8|0);
}
function __ZNKSt3__17codecvtIDsc11__mbstate_tE10do_unshiftERS1_PcS4_RS4_($this,$0,$to,$1,$to_nxt) {
 $this = $this|0;
 $0 = $0|0;
 $to = $to|0;
 $1 = $1|0;
 $to_nxt = $to_nxt|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$to_nxt>>2] = $to;
 return 3;
}
function __ZNKSt3__17codecvtIDsc11__mbstate_tE11do_encodingEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__17codecvtIDsc11__mbstate_tE16do_always_noconvEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__17codecvtIDsc11__mbstate_tE9do_lengthERS1_PKcS5_j($this,$0,$frm,$frm_end,$mx) {
 $this = $this|0;
 $0 = $0|0;
 $frm = $frm|0;
 $frm_end = $frm_end|0;
 $mx = $mx|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (__ZNSt3__1L20utf8_to_utf16_lengthEPKhS1_jmNS_12codecvt_modeE($frm,$frm_end,$mx,1114111,0)|0);
 return ($1|0);
}
function __ZNSt3__1L20utf8_to_utf16_lengthEPKhS1_jmNS_12codecvt_modeE($frm,$frm_end,$mx,$Maxcode,$mode) {
 $frm = $frm|0;
 $frm_end = $frm_end|0;
 $mx = $mx|0;
 $Maxcode = $Maxcode|0;
 $mode = $mode|0;
 var $$frm = 0, $$lcssa = 0, $$lcssa83 = 0, $$lcssa84 = 0, $$lcssa85 = 0, $$lcssa86 = 0, $$lcssa87 = 0, $$off = 0, $$pre = 0, $$pre42 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0;
 var $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $frm_nxt$1 = 0, $frm_nxt$1$lcssa = 0, $frm_nxt$5 = 0, $nchar16_t$0 = 0, $nchar16_t$2 = 0, $or$cond = 0, $or$cond11 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $mode & 4;
 $1 = ($0|0)==(0);
 $$pre42 = $frm_end;
 if ($1) {
  $frm_nxt$1 = $frm;$nchar16_t$0 = 0;
 } else {
  $2 = $frm;
  $3 = (($$pre42) - ($2))|0;
  $4 = ($3|0)>(2);
  if ($4) {
   $5 = HEAP8[$frm>>0]|0;
   $6 = ($5<<24>>24)==(-17);
   if ($6) {
    $7 = ((($frm)) + 1|0);
    $8 = HEAP8[$7>>0]|0;
    $9 = ($8<<24>>24)==(-69);
    if ($9) {
     $10 = ((($frm)) + 2|0);
     $11 = HEAP8[$10>>0]|0;
     $12 = ($11<<24>>24)==(-65);
     $13 = ((($frm)) + 3|0);
     $$frm = $12 ? $13 : $frm;
     $frm_nxt$1 = $$frm;$nchar16_t$0 = 0;
    } else {
     $frm_nxt$1 = $frm;$nchar16_t$0 = 0;
    }
   } else {
    $frm_nxt$1 = $frm;$nchar16_t$0 = 0;
   }
  } else {
   $frm_nxt$1 = $frm;$nchar16_t$0 = 0;
  }
 }
 L6: while(1) {
  $14 = ($frm_nxt$1>>>0)<($frm_end>>>0);
  $15 = ($nchar16_t$0>>>0)<($mx>>>0);
  $or$cond = $15 & $14;
  if (!($or$cond)) {
   $frm_nxt$1$lcssa = $frm_nxt$1;
   label = 40;
   break;
  }
  $16 = HEAP8[$frm_nxt$1>>0]|0;
  $17 = $16&255;
  $18 = ($17>>>0)>($Maxcode>>>0);
  if ($18) {
   $frm_nxt$1$lcssa = $frm_nxt$1;
   label = 40;
   break;
  }
  $19 = ($16<<24>>24)>(-1);
  do {
   if ($19) {
    $20 = ((($frm_nxt$1)) + 1|0);
    $frm_nxt$5 = $20;$nchar16_t$2 = $nchar16_t$0;
   } else {
    $21 = ($16&255)<(194);
    if ($21) {
     $frm_nxt$1$lcssa = $frm_nxt$1;
     label = 40;
     break L6;
    }
    $22 = ($16&255)<(224);
    if ($22) {
     $23 = $frm_nxt$1;
     $24 = (($$pre42) - ($23))|0;
     $25 = ($24|0)<(2);
     if ($25) {
      $frm_nxt$1$lcssa = $frm_nxt$1;
      label = 40;
      break L6;
     }
     $26 = ((($frm_nxt$1)) + 1|0);
     $27 = HEAP8[$26>>0]|0;
     $28 = $27&255;
     $29 = $28 & 192;
     $30 = ($29|0)==(128);
     if (!($30)) {
      $frm_nxt$1$lcssa = $frm_nxt$1;
      label = 40;
      break L6;
     }
     $31 = $17 << 6;
     $32 = $31 & 1984;
     $33 = $28 & 63;
     $34 = $33 | $32;
     $35 = ($34>>>0)>($Maxcode>>>0);
     $36 = ((($frm_nxt$1)) + 2|0);
     if ($35) {
      $frm_nxt$1$lcssa = $frm_nxt$1;
      label = 40;
      break L6;
     } else {
      $frm_nxt$5 = $36;$nchar16_t$2 = $nchar16_t$0;
      break;
     }
    }
    $37 = ($16&255)<(240);
    if ($37) {
     $38 = $frm_nxt$1;
     $39 = (($$pre42) - ($38))|0;
     $40 = ($39|0)<(3);
     if ($40) {
      $frm_nxt$1$lcssa = $frm_nxt$1;
      label = 40;
      break L6;
     }
     $41 = ((($frm_nxt$1)) + 1|0);
     $42 = HEAP8[$41>>0]|0;
     $43 = ((($frm_nxt$1)) + 2|0);
     $44 = HEAP8[$43>>0]|0;
     switch ($17|0) {
     case 224:  {
      $45 = $42 & -32;
      $46 = ($45<<24>>24)==(-96);
      if (!($46)) {
       $$lcssa86 = $38;
       label = 19;
       break L6;
      }
      break;
     }
     case 237:  {
      $49 = $42 & -32;
      $50 = ($49<<24>>24)==(-128);
      if (!($50)) {
       $$lcssa85 = $38;
       label = 21;
       break L6;
      }
      break;
     }
     default: {
      $53 = $42 & -64;
      $54 = ($53<<24>>24)==(-128);
      if (!($54)) {
       $$lcssa87 = $38;
       label = 23;
       break L6;
      }
     }
     }
     $57 = $44&255;
     $58 = $57 & 192;
     $59 = ($58|0)==(128);
     if (!($59)) {
      $frm_nxt$1$lcssa = $frm_nxt$1;
      label = 40;
      break L6;
     }
     $60 = $17 << 12;
     $61 = $60 & 61440;
     $62 = $42&255;
     $63 = $62 << 6;
     $64 = $63 & 4032;
     $65 = $64 | $61;
     $66 = $57 & 63;
     $67 = $65 | $66;
     $68 = ($67>>>0)>($Maxcode>>>0);
     $69 = ((($frm_nxt$1)) + 3|0);
     if ($68) {
      $frm_nxt$1$lcssa = $frm_nxt$1;
      label = 40;
      break L6;
     } else {
      $frm_nxt$5 = $69;$nchar16_t$2 = $nchar16_t$0;
      break;
     }
    }
    $70 = ($16&255)<(245);
    if (!($70)) {
     $frm_nxt$1$lcssa = $frm_nxt$1;
     label = 40;
     break L6;
    }
    $71 = $frm_nxt$1;
    $72 = (($$pre42) - ($71))|0;
    $73 = ($72|0)<(4);
    $74 = (($mx) - ($nchar16_t$0))|0;
    $75 = ($74>>>0)<(2);
    $or$cond11 = $75 | $73;
    if ($or$cond11) {
     $frm_nxt$1$lcssa = $frm_nxt$1;
     label = 40;
     break L6;
    }
    $76 = ((($frm_nxt$1)) + 1|0);
    $77 = HEAP8[$76>>0]|0;
    $78 = ((($frm_nxt$1)) + 2|0);
    $79 = HEAP8[$78>>0]|0;
    $80 = ((($frm_nxt$1)) + 3|0);
    $81 = HEAP8[$80>>0]|0;
    switch ($17|0) {
    case 240:  {
     $$off = (($77) + 112)<<24>>24;
     $82 = ($$off&255)<(48);
     if (!($82)) {
      $$lcssa83 = $71;
      label = 30;
      break L6;
     }
     break;
    }
    case 244:  {
     $85 = $77 & -16;
     $86 = ($85<<24>>24)==(-128);
     if (!($86)) {
      $$lcssa = $71;
      label = 32;
      break L6;
     }
     break;
    }
    default: {
     $89 = $77 & -64;
     $90 = ($89<<24>>24)==(-128);
     if (!($90)) {
      $$lcssa84 = $71;
      label = 34;
      break L6;
     }
    }
    }
    $93 = $79&255;
    $94 = $93 & 192;
    $95 = ($94|0)==(128);
    if (!($95)) {
     $frm_nxt$1$lcssa = $frm_nxt$1;
     label = 40;
     break L6;
    }
    $96 = $81&255;
    $97 = $96 & 192;
    $98 = ($97|0)==(128);
    if (!($98)) {
     $frm_nxt$1$lcssa = $frm_nxt$1;
     label = 40;
     break L6;
    }
    $99 = $17 << 18;
    $100 = $99 & 1835008;
    $101 = $77&255;
    $102 = $101 << 12;
    $103 = $102 & 258048;
    $104 = $103 | $100;
    $105 = $93 << 6;
    $106 = $105 & 4032;
    $107 = $104 | $106;
    $108 = $96 & 63;
    $109 = $107 | $108;
    $110 = ($109>>>0)>($Maxcode>>>0);
    if ($110) {
     $frm_nxt$1$lcssa = $frm_nxt$1;
     label = 40;
     break L6;
    }
    $111 = (($nchar16_t$0) + 1)|0;
    $112 = ((($frm_nxt$1)) + 4|0);
    $frm_nxt$5 = $112;$nchar16_t$2 = $111;
   }
  } while(0);
  $113 = (($nchar16_t$2) + 1)|0;
  $frm_nxt$1 = $frm_nxt$5;$nchar16_t$0 = $113;
 }
 if ((label|0) == 19) {
  $47 = $frm;
  $48 = (($$lcssa86) - ($47))|0;
  $116 = $48;
 }
 else if ((label|0) == 21) {
  $51 = $frm;
  $52 = (($$lcssa85) - ($51))|0;
  $116 = $52;
 }
 else if ((label|0) == 23) {
  $55 = $frm;
  $56 = (($$lcssa87) - ($55))|0;
  $116 = $56;
 }
 else if ((label|0) == 30) {
  $83 = $frm;
  $84 = (($$lcssa83) - ($83))|0;
  $116 = $84;
 }
 else if ((label|0) == 32) {
  $87 = $frm;
  $88 = (($$lcssa) - ($87))|0;
  $116 = $88;
 }
 else if ((label|0) == 34) {
  $91 = $frm;
  $92 = (($$lcssa84) - ($91))|0;
  $116 = $92;
 }
 else if ((label|0) == 40) {
  $$pre = $frm;
  $114 = $frm_nxt$1$lcssa;
  $115 = (($114) - ($$pre))|0;
  $116 = $115;
 }
 return ($116|0);
}
function __ZNKSt3__17codecvtIDsc11__mbstate_tE13do_max_lengthEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 4;
}
function __ZNSt3__16locale5__impD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (157296);
 $0 = ((($this)) + 8|0);
 $1 = ((($this)) + 12|0);
 $i$0 = 0;
 while(1) {
  $2 = HEAP32[$1>>2]|0;
  $3 = HEAP32[$0>>2]|0;
  $4 = (($2) - ($3))|0;
  $5 = $4 >> 2;
  $6 = ($i$0>>>0)<($5>>>0);
  if (!($6)) {
   break;
  }
  $8 = $3;
  $9 = (($8) + ($i$0<<2)|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ($10|0)==(0|0);
  if (!($11)) {
   (__ZNSt3__114__shared_count16__release_sharedEv($10)|0);
  }
  $12 = (($i$0) + 1)|0;
  $i$0 = $12;
 }
 $7 = ((($this)) + 144|0);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
 __ZNSt3__113__vector_baseIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEED2Ev($0);
 return;
}
function __ZNSt3__113__vector_baseIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEED2Ev($this) {
 $this = $this|0;
 var $$pre$i$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ($0|0)==(0|0);
 do {
  if (!($1)) {
   $2 = ((($this)) + 4|0);
   $$pre$i$i = HEAP32[$2>>2]|0;
   $4 = $$pre$i$i;
   while(1) {
    $3 = ($4|0)==($0|0);
    if ($3) {
     break;
    }
    $5 = ((($4)) + -4|0);
    HEAP32[$2>>2] = $5;
    $4 = $5;
   }
   $6 = ((($this)) + 16|0);
   $7 = ($6|0)==($0|0);
   if ($7) {
    $8 = ((($this)) + 128|0);
    HEAP8[$8>>0] = 0;
    break;
   } else {
    __ZdlPv($0);
    break;
   }
  }
 } while(0);
 return;
}
function __ZNSt3__16locale5__impD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__16locale5__impD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNSt3__15ctypeIcED2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (157316);
 $0 = ((($this)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = ((($this)) + 12|0);
  $4 = HEAP8[$3>>0]|0;
  $5 = ($4<<24>>24)==(0);
  if (!($5)) {
   __ZdaPv($1);
  }
 }
 return;
}
function __ZNSt3__15ctypeIcED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__15ctypeIcED2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNKSt3__15ctypeIcE10do_toupperEc($this,$c) {
 $this = $this|0;
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $isascii1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $isascii1 = ($c<<24>>24)>(-1);
 if ($isascii1) {
  $0 = $c&255;
  $1 = (__ZNSt3__15ctypeIcE21__classic_upper_tableEv()|0);
  $2 = (($1) + ($0<<2)|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = $3&255;
  $5 = $4;
 } else {
  $5 = $c;
 }
 return ($5|0);
}
function __ZNSt3__15ctypeIcE21__classic_upper_tableEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __THREW__ = 0;
 $0 = (invoke_i(4923)|0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  ___clang_call_terminate($4);
  // unreachable;
 } else {
  $3 = HEAP32[$0>>2]|0;
  return ($3|0);
 }
 return (0)|0;
}
function __ZNKSt3__15ctypeIcE10do_toupperEPcPKc($this,$low,$high) {
 $this = $this|0;
 $low = $low|0;
 $high = $high|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $isascii1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $low;
 while(1) {
  $0 = ($$0|0)==($high|0);
  if ($0) {
   break;
  }
  $1 = HEAP8[$$0>>0]|0;
  $isascii1 = ($1<<24>>24)>(-1);
  if ($isascii1) {
   $2 = $1 << 24 >> 24;
   $3 = (__ZNSt3__15ctypeIcE21__classic_upper_tableEv()|0);
   $4 = (($3) + ($2<<2)|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = $5&255;
   $7 = $6;
  } else {
   $7 = $1;
  }
  HEAP8[$$0>>0] = $7;
  $8 = ((($$0)) + 1|0);
  $$0 = $8;
 }
 return ($high|0);
}
function __ZNKSt3__15ctypeIcE10do_tolowerEc($this,$c) {
 $this = $this|0;
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $isascii1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $isascii1 = ($c<<24>>24)>(-1);
 if ($isascii1) {
  $0 = $c << 24 >> 24;
  $1 = (__ZNSt3__15ctypeIcE21__classic_lower_tableEv()|0);
  $2 = (($1) + ($0<<2)|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = $3&255;
  $5 = $4;
 } else {
  $5 = $c;
 }
 return ($5|0);
}
function __ZNSt3__15ctypeIcE21__classic_lower_tableEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __THREW__ = 0;
 $0 = (invoke_i(4924)|0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  ___clang_call_terminate($4);
  // unreachable;
 } else {
  $3 = HEAP32[$0>>2]|0;
  return ($3|0);
 }
 return (0)|0;
}
function __ZNKSt3__15ctypeIcE10do_tolowerEPcPKc($this,$low,$high) {
 $this = $this|0;
 $low = $low|0;
 $high = $high|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $isascii1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $low;
 while(1) {
  $0 = ($$0|0)==($high|0);
  if ($0) {
   break;
  }
  $1 = HEAP8[$$0>>0]|0;
  $isascii1 = ($1<<24>>24)>(-1);
  if ($isascii1) {
   $2 = $1 << 24 >> 24;
   $3 = (__ZNSt3__15ctypeIcE21__classic_lower_tableEv()|0);
   $4 = (($3) + ($2<<2)|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = $5&255;
   $7 = $6;
  } else {
   $7 = $1;
  }
  HEAP8[$$0>>0] = $7;
  $8 = ((($$0)) + 1|0);
  $$0 = $8;
 }
 return ($high|0);
}
function __ZNKSt3__15ctypeIcE8do_widenEc($this,$c) {
 $this = $this|0;
 $c = $c|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ($c|0);
}
function __ZNKSt3__15ctypeIcE8do_widenEPKcS3_Pc($this,$low,$high,$dest) {
 $this = $this|0;
 $low = $low|0;
 $high = $high|0;
 $dest = $dest|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $dest;$$01 = $low;
 while(1) {
  $0 = ($$01|0)==($high|0);
  if ($0) {
   break;
  }
  $1 = HEAP8[$$01>>0]|0;
  HEAP8[$$0>>0] = $1;
  $2 = ((($$01)) + 1|0);
  $3 = ((($$0)) + 1|0);
  $$0 = $3;$$01 = $2;
 }
 return ($high|0);
}
function __ZNKSt3__15ctypeIcE9do_narrowEcc($this,$c,$dfault) {
 $this = $this|0;
 $c = $c|0;
 $dfault = $dfault|0;
 var $c$dfault = 0, $isascii1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $isascii1 = ($c<<24>>24)>(-1);
 $c$dfault = $isascii1 ? $c : $dfault;
 return ($c$dfault|0);
}
function __ZNKSt3__15ctypeIcE9do_narrowEPKcS3_cPc($this,$low,$high,$dfault,$dest) {
 $this = $this|0;
 $low = $low|0;
 $high = $high|0;
 $dfault = $dfault|0;
 $dest = $dest|0;
 var $$0 = 0, $$01 = 0, $$dfault = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $isascii2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $dest;$$01 = $low;
 while(1) {
  $0 = ($$01|0)==($high|0);
  if ($0) {
   break;
  }
  $1 = HEAP8[$$01>>0]|0;
  $isascii2 = ($1<<24>>24)>(-1);
  $$dfault = $isascii2 ? $1 : $dfault;
  HEAP8[$$0>>0] = $$dfault;
  $2 = ((($$01)) + 1|0);
  $3 = ((($$0)) + 1|0);
  $$0 = $3;$$01 = $2;
 }
 return ($high|0);
}
function __ZNSt3__15ctypeIcE13classic_tableEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __THREW__ = 0;
 $0 = (invoke_i(4925)|0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  ___clang_call_terminate($4);
  // unreachable;
 } else {
  $3 = HEAP32[$0>>2]|0;
  return ($3|0);
 }
 return (0)|0;
}
function __ZNSt3__18numpunctIcED2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (157368);
 $0 = ((($this)) + 12|0);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0);
 return;
}
function __ZNSt3__18numpunctIcED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__18numpunctIcED2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNKSt3__18numpunctIcE16do_decimal_pointEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 $1 = HEAP8[$0>>0]|0;
 return ($1|0);
}
function __ZNKSt3__18numpunctIcE16do_thousands_sepEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 9|0);
 $1 = HEAP8[$0>>0]|0;
 return ($1|0);
}
function __ZNKSt3__18numpunctIcE11do_groupingEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_($agg$result,$0);
 return;
}
function __ZNKSt3__18numpunctIcE11do_truenameEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($agg$result,622520,4);
 return;
}
function __ZNKSt3__18numpunctIcE12do_falsenameEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($agg$result,622525,5);
 return;
}
function __ZNSt3__18numpunctIwED2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (157408);
 $0 = ((($this)) + 16|0);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0);
 return;
}
function __ZNSt3__18numpunctIwED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__18numpunctIwED2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNKSt3__18numpunctIwE16do_decimal_pointEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNKSt3__18numpunctIwE16do_thousands_sepEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNKSt3__18numpunctIwE11do_groupingEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_($agg$result,$0);
 return;
}
function __ZNKSt3__18numpunctIwE11do_truenameEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_wcslen(157440)|0);
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6__initEPKwj($agg$result,157440,$0);
 return;
}
function __ZNKSt3__18numpunctIwE12do_falsenameEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_wcslen(157460)|0);
 __ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6__initEPKwj($agg$result,157460,$0);
 return;
}
function __ZNSt3__16locale5facetD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNSt3__15ctypeIwED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__15ctypeIwE5do_isEtw($this,$m,$c) {
 $this = $this|0;
 $m = $m|0;
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $isascii = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $isascii = ($c>>>0)<(128);
 if ($isascii) {
  $0 = (__ZNSt3__15ctypeIcE13classic_tableEv()|0);
  $1 = (($0) + ($c<<1)|0);
  $2 = HEAP16[$1>>1]|0;
  $3 = $2 & $m;
  $4 = ($3<<16>>16)!=(0);
  $5 = $4;
 } else {
  $5 = 0;
 }
 return ($5|0);
}
function __ZNKSt3__15ctypeIwE5do_isEPKwS3_Pt($this,$low,$high,$vec) {
 $this = $this|0;
 $low = $low|0;
 $high = $high|0;
 $vec = $vec|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $isascii = 0, $low3 = 0, $scevgep = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $low3 = $low;
 $0 = $high;
 $1 = (($0) - ($low3))|0;
 $2 = $1 >>> 2;
 $$0 = $vec;$$01 = $low;
 while(1) {
  $3 = ($$01|0)==($high|0);
  if ($3) {
   break;
  }
  $4 = HEAP32[$$01>>2]|0;
  $isascii = ($4>>>0)<(128);
  if ($isascii) {
   $5 = (__ZNSt3__15ctypeIcE13classic_tableEv()|0);
   $6 = (($5) + ($4<<1)|0);
   $7 = HEAP16[$6>>1]|0;
   $8 = $7&65535;
   $10 = $8;
  } else {
   $10 = 0;
  }
  $9 = $10&65535;
  HEAP16[$$0>>1] = $9;
  $11 = ((($$01)) + 4|0);
  $12 = ((($$0)) + 2|0);
  $$0 = $12;$$01 = $11;
 }
 $scevgep = (($low) + ($2<<2)|0);
 return ($scevgep|0);
}
function __ZNKSt3__15ctypeIwE10do_scan_isEtPKwS3_($this,$m,$low,$high) {
 $this = $this|0;
 $m = $m|0;
 $low = $low|0;
 $high = $high|0;
 var $$0 = 0, $$0$lcssa = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $isascii = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $low;
 while(1) {
  $0 = ($$0|0)==($high|0);
  if ($0) {
   $$0$lcssa = $high;
   break;
  }
  $1 = HEAP32[$$0>>2]|0;
  $isascii = ($1>>>0)<(128);
  if ($isascii) {
   $2 = (__ZNSt3__15ctypeIcE13classic_tableEv()|0);
   $3 = (($2) + ($1<<1)|0);
   $4 = HEAP16[$3>>1]|0;
   $5 = $4 & $m;
   $6 = ($5<<16>>16)==(0);
   if (!($6)) {
    $$0$lcssa = $$0;
    break;
   }
  }
  $7 = ((($$0)) + 4|0);
  $$0 = $7;
 }
 return ($$0$lcssa|0);
}
function __ZNKSt3__15ctypeIwE11do_scan_notEtPKwS3_($this,$m,$low,$high) {
 $this = $this|0;
 $m = $m|0;
 $low = $low|0;
 $high = $high|0;
 var $$0 = 0, $$0$lcssa = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $isascii = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $low;
 while(1) {
  $0 = ($$0|0)==($high|0);
  if ($0) {
   $$0$lcssa = $high;
   break;
  }
  $1 = HEAP32[$$0>>2]|0;
  $isascii = ($1>>>0)<(128);
  if (!($isascii)) {
   $$0$lcssa = $$0;
   break;
  }
  $2 = (__ZNSt3__15ctypeIcE13classic_tableEv()|0);
  $3 = (($2) + ($1<<1)|0);
  $4 = HEAP16[$3>>1]|0;
  $5 = $4 & $m;
  $6 = ($5<<16>>16)==(0);
  if ($6) {
   $$0$lcssa = $$0;
   break;
  }
  $7 = ((($$0)) + 4|0);
  $$0 = $7;
 }
 return ($$0$lcssa|0);
}
function __ZNKSt3__15ctypeIwE10do_toupperEw($this,$c) {
 $this = $this|0;
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $isascii = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $isascii = ($c>>>0)<(128);
 if ($isascii) {
  $0 = (__ZNSt3__15ctypeIcE21__classic_upper_tableEv()|0);
  $1 = (($0) + ($c<<2)|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = $2;
 } else {
  $3 = $c;
 }
 return ($3|0);
}
function __ZNKSt3__15ctypeIwE10do_toupperEPwPKw($this,$low,$high) {
 $this = $this|0;
 $low = $low|0;
 $high = $high|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $isascii = 0, $low2 = 0, $scevgep = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $low2 = $low;
 $0 = $high;
 $1 = (($0) - ($low2))|0;
 $2 = $1 >>> 2;
 $$0 = $low;
 while(1) {
  $3 = ($$0|0)==($high|0);
  if ($3) {
   break;
  }
  $4 = HEAP32[$$0>>2]|0;
  $isascii = ($4>>>0)<(128);
  if ($isascii) {
   $5 = (__ZNSt3__15ctypeIcE21__classic_upper_tableEv()|0);
   $6 = (($5) + ($4<<2)|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = $7;
  } else {
   $8 = $4;
  }
  HEAP32[$$0>>2] = $8;
  $9 = ((($$0)) + 4|0);
  $$0 = $9;
 }
 $scevgep = (($low) + ($2<<2)|0);
 return ($scevgep|0);
}
function __ZNKSt3__15ctypeIwE10do_tolowerEw($this,$c) {
 $this = $this|0;
 $c = $c|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $isascii = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $isascii = ($c>>>0)<(128);
 if ($isascii) {
  $0 = (__ZNSt3__15ctypeIcE21__classic_lower_tableEv()|0);
  $1 = (($0) + ($c<<2)|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = $2;
 } else {
  $3 = $c;
 }
 return ($3|0);
}
function __ZNKSt3__15ctypeIwE10do_tolowerEPwPKw($this,$low,$high) {
 $this = $this|0;
 $low = $low|0;
 $high = $high|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $isascii = 0, $low2 = 0, $scevgep = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $low2 = $low;
 $0 = $high;
 $1 = (($0) - ($low2))|0;
 $2 = $1 >>> 2;
 $$0 = $low;
 while(1) {
  $3 = ($$0|0)==($high|0);
  if ($3) {
   break;
  }
  $4 = HEAP32[$$0>>2]|0;
  $isascii = ($4>>>0)<(128);
  if ($isascii) {
   $5 = (__ZNSt3__15ctypeIcE21__classic_lower_tableEv()|0);
   $6 = (($5) + ($4<<2)|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = $7;
  } else {
   $8 = $4;
  }
  HEAP32[$$0>>2] = $8;
  $9 = ((($$0)) + 4|0);
  $$0 = $9;
 }
 $scevgep = (($low) + ($2<<2)|0);
 return ($scevgep|0);
}
function __ZNKSt3__15ctypeIwE8do_widenEc($this,$c) {
 $this = $this|0;
 $c = $c|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $c << 24 >> 24;
 return ($0|0);
}
function __ZNKSt3__15ctypeIwE8do_widenEPKcS3_Pw($this,$low,$high,$dest) {
 $this = $this|0;
 $low = $low|0;
 $high = $high|0;
 $dest = $dest|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $dest;$$01 = $low;
 while(1) {
  $0 = ($$01|0)==($high|0);
  if ($0) {
   break;
  }
  $1 = HEAP8[$$01>>0]|0;
  $2 = $1 << 24 >> 24;
  HEAP32[$$0>>2] = $2;
  $3 = ((($$01)) + 1|0);
  $4 = ((($$0)) + 4|0);
  $$0 = $4;$$01 = $3;
 }
 return ($high|0);
}
function __ZNKSt3__15ctypeIwE9do_narrowEwc($this,$c,$dfault) {
 $this = $this|0;
 $c = $c|0;
 $dfault = $dfault|0;
 var $$0 = 0, $0 = 0, $isascii = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $isascii = ($c>>>0)<(128);
 $0 = $c&255;
 $$0 = $isascii ? $0 : $dfault;
 return ($$0|0);
}
function __ZNKSt3__15ctypeIwE9do_narrowEPKwS3_cPc($this,$low,$high,$dfault,$dest) {
 $this = $this|0;
 $low = $low|0;
 $high = $high|0;
 $dfault = $dfault|0;
 $dest = $dest|0;
 var $$0 = 0, $$01 = 0, $$dfault = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $isascii = 0, $low3 = 0, $scevgep = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $low3 = $low;
 $0 = $high;
 $1 = (($0) - ($low3))|0;
 $2 = $1 >>> 2;
 $$0 = $dest;$$01 = $low;
 while(1) {
  $3 = ($$01|0)==($high|0);
  if ($3) {
   break;
  }
  $4 = HEAP32[$$01>>2]|0;
  $isascii = ($4>>>0)<(128);
  $5 = $4&255;
  $$dfault = $isascii ? $5 : $dfault;
  HEAP8[$$0>>0] = $$dfault;
  $6 = ((($$01)) + 4|0);
  $7 = ((($$0)) + 1|0);
  $$0 = $7;$$01 = $6;
 }
 $scevgep = (($low) + ($2<<2)|0);
 return ($scevgep|0);
}
function __ZNSt3__17codecvtIcc11__mbstate_tED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNSt3__17codecvtIDsc11__mbstate_tED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNSt3__17codecvtIDic11__mbstate_tED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (___cxa_allocate_exception(8)|0);
 __THREW__ = 0;
 invoke_vii(3996,($0|0),(622552|0));
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  ___cxa_free_exception(($0|0));
  ___resumeException($3|0);
  // unreachable;
 } else {
  HEAP32[$0>>2] = (157892);
  ___cxa_throw(($0|0),(9104|0),(3328|0));
  // unreachable;
 }
}
function __ZNSt3__16locale5__impC2Ej($this,$refs) {
 $this = $this|0;
 $refs = $refs|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$pre$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $9 = 0, $eh$lpad$body$index3Z2D = 0, $eh$lpad$body$indexZ2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($refs) + -1)|0;
 $1 = ((($this)) + 4|0);
 HEAP32[$1>>2] = $0;
 HEAP32[$this>>2] = (157296);
 $2 = ((($this)) + 8|0);
 __THREW__ = 0;
 invoke_vii(4926,($2|0),28);
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $82 = ___cxa_find_matching_catch_2()|0;
  $83 = tempRet0;
  $$1 = $83;$$12 = $82;
 } else {
  $5 = ((($this)) + 144|0);
  __THREW__ = 0;
  invoke_viii(3460,($5|0),(620519|0),1);
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if ($7) {
   $84 = ___cxa_find_matching_catch_2()|0;
   $85 = tempRet0;
   $$0 = $85;$$01 = $84;
  } else {
   $8 = HEAP32[$2>>2]|0;
   $9 = ((($this)) + 12|0);
   $$pre$i$i$i = HEAP32[$9>>2]|0;
   $11 = $$pre$i$i$i;
   while(1) {
    $10 = ($11|0)==($8|0);
    if ($10) {
     break;
    }
    $12 = ((($11)) + -4|0);
    HEAP32[$9>>2] = $12;
    $11 = $12;
   }
   HEAP32[(625124)>>2] = 0;
   HEAP32[156280] = (154960);
   __THREW__ = 0;
   invoke_vii(4927,($this|0),(625120|0));
   $13 = __THREW__; __THREW__ = 0;
   $14 = $13&1;
   do {
    if ($14) {
     label = 42;
    } else {
     HEAP32[(625132)>>2] = 0;
     HEAP32[156282] = (154992);
     __THREW__ = 0;
     invoke_vii(4928,($this|0),(625128|0));
     $15 = __THREW__; __THREW__ = 0;
     $16 = $15&1;
     if ($16) {
      label = 42;
     } else {
      HEAP32[(625140)>>2] = 0;
      HEAP32[156284] = (157316);
      HEAP32[(625144)>>2] = 0;
      HEAP8[(625148)>>0] = 0;
      $17 = (__ZNSt3__15ctypeIcE13classic_tableEv()|0);
      HEAP32[(625144)>>2] = $17;
      __THREW__ = 0;
      invoke_vii(4929,($this|0),(625136|0));
      $18 = __THREW__; __THREW__ = 0;
      $19 = $18&1;
      if ($19) {
       label = 42;
      } else {
       HEAP32[(625156)>>2] = 0;
       HEAP32[156288] = (157512);
       __THREW__ = 0;
       invoke_vii(4930,($this|0),(625152|0));
       $20 = __THREW__; __THREW__ = 0;
       $21 = $20&1;
       if ($21) {
        label = 42;
       } else {
        HEAP32[(625164)>>2] = 0;
        HEAP32[156290] = (157580);
        __THREW__ = 0;
        invoke_vii(4931,($this|0),(625160|0));
        $22 = __THREW__; __THREW__ = 0;
        $23 = $22&1;
        if ($23) {
         label = 42;
        } else {
         __THREW__ = 0;
         invoke_vii(4932,(625168|0),1);
         $24 = __THREW__; __THREW__ = 0;
         $25 = $24&1;
         if ($25) {
          label = 42;
         } else {
          __THREW__ = 0;
          invoke_vii(4933,($this|0),(625168|0));
          $26 = __THREW__; __THREW__ = 0;
          $27 = $26&1;
          if ($27) {
           label = 42;
          } else {
           HEAP32[(625188)>>2] = 0;
           HEAP32[156296] = (157628);
           __THREW__ = 0;
           invoke_vii(4934,($this|0),(625184|0));
           $28 = __THREW__; __THREW__ = 0;
           $29 = $28&1;
           if ($29) {
            label = 42;
           } else {
            HEAP32[(625196)>>2] = 0;
            HEAP32[156298] = (157676);
            __THREW__ = 0;
            invoke_vii(4935,($this|0),(625192|0));
            $30 = __THREW__; __THREW__ = 0;
            $31 = $30&1;
            if ($31) {
             label = 42;
            } else {
             __ZNSt3__18numpunctIcEC2Ej(625200,1);
             __THREW__ = 0;
             invoke_vii(4936,($this|0),(625200|0));
             $32 = __THREW__; __THREW__ = 0;
             $33 = $32&1;
             if ($33) {
              label = 42;
             } else {
              __ZNSt3__18numpunctIwEC2Ej(625224,1);
              __THREW__ = 0;
              invoke_vii(4937,($this|0),(625224|0));
              $34 = __THREW__; __THREW__ = 0;
              $35 = $34&1;
              if ($35) {
               label = 42;
              } else {
               HEAP32[(625260)>>2] = 0;
               HEAP32[156314] = (155024);
               __THREW__ = 0;
               invoke_vii(4938,($this|0),(625256|0));
               $36 = __THREW__; __THREW__ = 0;
               $37 = $36&1;
               if ($37) {
                label = 42;
               } else {
                HEAP32[(625268)>>2] = 0;
                HEAP32[156316] = (155088);
                __THREW__ = 0;
                invoke_vii(4939,($this|0),(625264|0));
                $38 = __THREW__; __THREW__ = 0;
                $39 = $38&1;
                if ($39) {
                 label = 42;
                } else {
                 HEAP32[(625276)>>2] = 0;
                 HEAP32[156318] = (155152);
                 __THREW__ = 0;
                 invoke_vii(4940,($this|0),(625272|0));
                 $40 = __THREW__; __THREW__ = 0;
                 $41 = $40&1;
                 if ($41) {
                  label = 42;
                 } else {
                  HEAP32[(625284)>>2] = 0;
                  HEAP32[156320] = (155204);
                  __THREW__ = 0;
                  invoke_vii(4941,($this|0),(625280|0));
                  $42 = __THREW__; __THREW__ = 0;
                  $43 = $42&1;
                  if ($43) {
                   label = 42;
                  } else {
                   HEAP32[(625292)>>2] = 0;
                   HEAP32[156322] = (156752);
                   __THREW__ = 0;
                   invoke_vii(4942,($this|0),(625288|0));
                   $44 = __THREW__; __THREW__ = 0;
                   $45 = $44&1;
                   if ($45) {
                    label = 42;
                    break;
                   }
                   HEAP32[(625300)>>2] = 0;
                   HEAP32[156324] = (156808);
                   __THREW__ = 0;
                   invoke_vii(4943,($this|0),(625296|0));
                   $46 = __THREW__; __THREW__ = 0;
                   $47 = $46&1;
                   if ($47) {
                    label = 42;
                    break;
                   }
                   HEAP32[(625308)>>2] = 0;
                   HEAP32[156326] = (156864);
                   __THREW__ = 0;
                   invoke_vii(4944,($this|0),(625304|0));
                   $48 = __THREW__; __THREW__ = 0;
                   $49 = $48&1;
                   if ($49) {
                    label = 42;
                    break;
                   }
                   HEAP32[(625316)>>2] = 0;
                   HEAP32[156328] = (156920);
                   __THREW__ = 0;
                   invoke_vii(4945,($this|0),(625312|0));
                   $50 = __THREW__; __THREW__ = 0;
                   $51 = $50&1;
                   if ($51) {
                    label = 42;
                    break;
                   }
                   HEAP32[(625324)>>2] = 0;
                   HEAP32[156330] = (156976);
                   __THREW__ = 0;
                   invoke_vii(4946,($this|0),(625320|0));
                   $52 = __THREW__; __THREW__ = 0;
                   $53 = $52&1;
                   if ($53) {
                    label = 42;
                    break;
                   }
                   HEAP32[(625332)>>2] = 0;
                   HEAP32[156332] = (157004);
                   __THREW__ = 0;
                   invoke_vii(4947,($this|0),(625328|0));
                   $54 = __THREW__; __THREW__ = 0;
                   $55 = $54&1;
                   if ($55) {
                    label = 42;
                    break;
                   }
                   HEAP32[(625340)>>2] = 0;
                   HEAP32[156334] = (157032);
                   __THREW__ = 0;
                   invoke_vii(4948,($this|0),(625336|0));
                   $56 = __THREW__; __THREW__ = 0;
                   $57 = $56&1;
                   if ($57) {
                    label = 42;
                    break;
                   }
                   HEAP32[(625348)>>2] = 0;
                   HEAP32[156336] = (157060);
                   __THREW__ = 0;
                   invoke_vii(4949,($this|0),(625344|0));
                   $58 = __THREW__; __THREW__ = 0;
                   $59 = $58&1;
                   if ($59) {
                    label = 42;
                    break;
                   }
                   HEAP32[(625356)>>2] = 0;
                   HEAP32[156338] = (155256);
                   HEAP32[(625360)>>2] = (155304);
                   __THREW__ = 0;
                   invoke_vii(4950,($this|0),(625352|0));
                   $60 = __THREW__; __THREW__ = 0;
                   $61 = $60&1;
                   if ($61) {
                    label = 42;
                    break;
                   }
                   HEAP32[(625372)>>2] = 0;
                   HEAP32[156342] = (155340);
                   HEAP32[(625376)>>2] = (155388);
                   __THREW__ = 0;
                   invoke_vii(4951,($this|0),(625368|0));
                   $62 = __THREW__; __THREW__ = 0;
                   $63 = $62&1;
                   if ($63) {
                    label = 42;
                    break;
                   }
                   HEAP32[(625388)>>2] = 0;
                   HEAP32[156346] = (157492);
                   __THREW__ = 0;
                   $64 = (invoke_i(4874)|0);
                   $65 = __THREW__; __THREW__ = 0;
                   $66 = $65&1;
                   if ($66) {
                    $67 = ___cxa_find_matching_catch_2()|0;
                    $68 = tempRet0;
                    $eh$lpad$body$index3Z2D = $68;$eh$lpad$body$indexZ2D = $67;
                    break;
                   }
                   HEAP32[(625392)>>2] = $64;
                   HEAP32[156346] = (156704);
                   __THREW__ = 0;
                   invoke_vii(4952,($this|0),(625384|0));
                   $69 = __THREW__; __THREW__ = 0;
                   $70 = $69&1;
                   if ($70) {
                    label = 42;
                    break;
                   }
                   HEAP32[(625404)>>2] = 0;
                   HEAP32[156350] = (157492);
                   __THREW__ = 0;
                   $71 = (invoke_i(4874)|0);
                   $72 = __THREW__; __THREW__ = 0;
                   $73 = $72&1;
                   if ($73) {
                    $74 = ___cxa_find_matching_catch_2()|0;
                    $75 = tempRet0;
                    $eh$lpad$body$index3Z2D = $75;$eh$lpad$body$indexZ2D = $74;
                    break;
                   }
                   HEAP32[(625408)>>2] = $71;
                   HEAP32[156350] = (156728);
                   __THREW__ = 0;
                   invoke_vii(4953,($this|0),(625400|0));
                   $76 = __THREW__; __THREW__ = 0;
                   $77 = $76&1;
                   if ($77) {
                    label = 42;
                    break;
                   }
                   HEAP32[(625420)>>2] = 0;
                   HEAP32[156354] = (157088);
                   __THREW__ = 0;
                   invoke_vii(4954,($this|0),(625416|0));
                   $78 = __THREW__; __THREW__ = 0;
                   $79 = $78&1;
                   if ($79) {
                    label = 42;
                    break;
                   }
                   HEAP32[(625428)>>2] = 0;
                   HEAP32[156356] = (157120);
                   __THREW__ = 0;
                   invoke_vii(4955,($this|0),(625424|0));
                   $80 = __THREW__; __THREW__ = 0;
                   $81 = $80&1;
                   if ($81) {
                    label = 42;
                    break;
                   }
                   return;
                  }
                 }
                }
               }
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 42) {
    $86 = ___cxa_find_matching_catch_2()|0;
    $87 = tempRet0;
    $eh$lpad$body$index3Z2D = $87;$eh$lpad$body$indexZ2D = $86;
   }
   __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5);
   $$0 = $eh$lpad$body$index3Z2D;$$01 = $eh$lpad$body$indexZ2D;
  }
  __ZNSt3__113__vector_baseIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEED2Ev($2);
  $$1 = $$0;$$12 = $$01;
 }
 ___resumeException($$12|0);
 // unreachable;
}
function __ZNSt3__16vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEEC2Ej($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$0$i = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = 0;
 $0 = ((($this)) + 4|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($this)) + 8|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($this)) + 128|0);
 HEAP8[$2>>0] = 0;
 $3 = ($__n|0)==(0);
 if (!($3)) {
  __THREW__ = 0;
  invoke_vii(4956,($this|0),($__n|0));
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if ($5) {
   $11 = ___cxa_find_matching_catch_2()|0;
   $12 = tempRet0;
   __ZNSt3__113__vector_baseIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEED2Ev($this);
   ___resumeException($11|0);
   // unreachable;
  }
  $$pre$i = HEAP32[$0>>2]|0;
  $$0$i = $__n;$6 = $$pre$i;
  while(1) {
   HEAP32[$6>>2] = 0;
   $7 = HEAP32[$0>>2]|0;
   $8 = ((($7)) + 4|0);
   HEAP32[$0>>2] = $8;
   $9 = (($$0$i) + -1)|0;
   $10 = ($9|0)==(0);
   if ($10) {
    break;
   } else {
    $$0$i = $9;$6 = $8;
   }
  }
 }
 return;
}
function __ZNSt3__16vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE8allocateEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$0$i$i = 0, $$not$i$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($__n>>>0)>(1073741823);
 if ($0) {
  __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this);
  // unreachable;
 }
 $1 = ((($this)) + 128|0);
 $2 = HEAP8[$1>>0]|0;
 $$not$i$i = ($2<<24>>24)==(0);
 $3 = ($__n>>>0)<(29);
 $or$cond$i$i = $3 & $$not$i$i;
 if ($or$cond$i$i) {
  $4 = ((($this)) + 16|0);
  HEAP8[$1>>0] = 1;
  $$0$i$i = $4;
 } else {
  $5 = $__n << 2;
  $6 = (__Znwj($5)|0);
  $$0$i$i = $6;
 }
 $7 = ((($this)) + 4|0);
 HEAP32[$7>>2] = $$0$i$i;
 HEAP32[$this>>2] = $$0$i$i;
 $8 = (($$0$i$i) + ($__n<<2)|0);
 $9 = ((($this)) + 8|0);
 HEAP32[$9>>2] = $8;
 return;
}
function __ZNSt3__16locale5__imp7installINS_7collateIcEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1207560)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$id) {
 $this = $this|0;
 $f = $f|0;
 $id = $id|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $$pre2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__114__shared_count12__add_sharedEv($f);
 $0 = ((($this)) + 8|0);
 $1 = ((($this)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = HEAP32[$0>>2]|0;
 $4 = (($2) - ($3))|0;
 $5 = $4 >> 2;
 $6 = ($5>>>0)>($id>>>0);
 do {
  if ($6) {
   $7 = $3;
   $$pre$phiZ2D = $0;$15 = $7;
  } else {
   $8 = (($id) + 1)|0;
   __THREW__ = 0;
   invoke_vii(4957,($0|0),($8|0));
   $9 = __THREW__; __THREW__ = 0;
   $10 = $9&1;
   if (!($10)) {
    $$pre = HEAP32[$0>>2]|0;
    $$pre$phiZ2D = $0;$15 = $$pre;
    break;
   }
   $11 = ___cxa_find_matching_catch_2()|0;
   $12 = tempRet0;
   $13 = ($f|0)==(0|0);
   if (!($13)) {
    (__ZNSt3__114__shared_count16__release_sharedEv($f)|0);
   }
   ___resumeException($11|0);
   // unreachable;
  }
 } while(0);
 $14 = (($15) + ($id<<2)|0);
 $16 = HEAP32[$14>>2]|0;
 $17 = ($16|0)==(0|0);
 if ($17) {
  $19 = $15;
 } else {
  (__ZNSt3__114__shared_count16__release_sharedEv($16)|0);
  $$pre2 = HEAP32[$$pre$phiZ2D>>2]|0;
  $19 = $$pre2;
 }
 $18 = (($19) + ($id<<2)|0);
 HEAP32[$18>>2] = $f;
 return;
}
function __ZNSt3__16vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE6resizeEj($this,$__sz) {
 $this = $this|0;
 $__sz = $__sz|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$this>>2]|0;
 $3 = (($1) - ($2))|0;
 $4 = $3 >> 2;
 $5 = ($4>>>0)<($__sz>>>0);
 $6 = $2;
 $7 = $1;
 L1: do {
  if ($5) {
   $8 = (($__sz) - ($4))|0;
   __ZNSt3__16vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE8__appendEj($this,$8);
  } else {
   $9 = ($4>>>0)>($__sz>>>0);
   if ($9) {
    $10 = (($6) + ($__sz<<2)|0);
    $12 = $7;
    while(1) {
     $11 = ($12|0)==($10|0);
     if ($11) {
      break L1;
     }
     $13 = ((($12)) + -4|0);
     HEAP32[$0>>2] = $13;
     $12 = $13;
    }
   }
  }
 } while(0);
 return;
}
function __ZNSt3__16vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE8__appendEj($this,$__n) {
 $this = $this|0;
 $__n = $__n|0;
 var $$0$i = 0, $$0$i3 = 0, $$pre$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__v = 0, $scevgep = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $__v = sp;
 $0 = ((($this)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($this)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) - ($3))|0;
 $5 = $4 >> 2;
 $6 = ($5>>>0)<($__n>>>0);
 if ($6) {
  $13 = ((($this)) + 16|0);
  $14 = HEAP32[$this>>2]|0;
  $15 = (($3) - ($14))|0;
  $16 = $15 >> 2;
  $17 = (($16) + ($__n))|0;
  $18 = ($17>>>0)>(1073741823);
  if ($18) {
   __ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv($this);
   // unreachable;
  }
  $19 = (($1) - ($14))|0;
  $20 = $19 >> 2;
  $21 = ($20>>>0)<(536870911);
  if ($21) {
   $22 = $19 >> 1;
   $23 = ($22>>>0)<($17>>>0);
   $24 = $23 ? $17 : $22;
   $$0$i3 = $24;
  } else {
   $$0$i3 = 1073741823;
  }
  __ZNSt3__114__split_bufferIPNS_6locale5facetERNS_15__sso_allocatorIS3_Lj28EEEEC2EjjS6_($__v,$$0$i3,$16,$13);
  $25 = ((($__v)) + 8|0);
  $$pre$i1 = HEAP32[$25>>2]|0;
  $26 = $__n << 2;
  _memset(($$pre$i1|0),0,($26|0))|0;
  $scevgep = (($$pre$i1) + ($__n<<2)|0);
  HEAP32[$25>>2] = $scevgep;
  __ZNSt3__16vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE26__swap_out_circular_bufferERNS_14__split_bufferIS3_RS5_EE($this,$__v);
  __ZNSt3__114__split_bufferIPNS_6locale5facetERNS_15__sso_allocatorIS3_Lj28EEEED2Ev($__v);
 } else {
  $7 = $3;
  $$0$i = $__n;$8 = $7;
  while(1) {
   HEAP32[$8>>2] = 0;
   $9 = HEAP32[$2>>2]|0;
   $10 = ((($9)) + 4|0);
   HEAP32[$2>>2] = $10;
   $11 = (($$0$i) + -1)|0;
   $12 = ($11|0)==(0);
   if ($12) {
    break;
   } else {
    $$0$i = $11;$8 = $10;
   }
  }
 }
 STACKTOP = sp;return;
}
function __ZNSt3__114__split_bufferIPNS_6locale5facetERNS_15__sso_allocatorIS3_Lj28EEEEC2EjjS6_($this,$__cap,$__start,$__a) {
 $this = $this|0;
 $__cap = $__cap|0;
 $__start = $__start|0;
 $__a = $__a|0;
 var $$not$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($this)) + 16|0);
 HEAP32[$1>>2] = $__a;
 $2 = ($__cap|0)==(0);
 do {
  if ($2) {
   $8 = 0;
  } else {
   $3 = ((($__a)) + 112|0);
   $4 = HEAP8[$3>>0]|0;
   $$not$i$i = ($4<<24>>24)==(0);
   $5 = ($__cap>>>0)<(29);
   $or$cond$i$i = $5 & $$not$i$i;
   if ($or$cond$i$i) {
    HEAP8[$3>>0] = 1;
    $8 = $__a;
    break;
   } else {
    $6 = $__cap << 2;
    $7 = (__Znwj($6)|0);
    $8 = $7;
    break;
   }
  }
 } while(0);
 HEAP32[$this>>2] = $8;
 $9 = (($8) + ($__start<<2)|0);
 $10 = ((($this)) + 8|0);
 HEAP32[$10>>2] = $9;
 $11 = ((($this)) + 4|0);
 HEAP32[$11>>2] = $9;
 $12 = (($8) + ($__cap<<2)|0);
 $13 = ((($this)) + 12|0);
 HEAP32[$13>>2] = $12;
 return;
}
function __ZNSt3__16vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE26__swap_out_circular_bufferERNS_14__split_bufferIS3_RS5_EE($this,$__v) {
 $this = $this|0;
 $__v = $__v|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($this)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($__v)) + 4|0);
 $4 = $0;
 $5 = (($2) - ($4))|0;
 $6 = $5 >> 2;
 $7 = HEAP32[$3>>2]|0;
 $8 = (0 - ($6))|0;
 $9 = (($7) + ($8<<2)|0);
 HEAP32[$3>>2] = $9;
 _memcpy(($9|0),($0|0),($5|0))|0;
 $10 = HEAP32[$this>>2]|0;
 $11 = HEAP32[$3>>2]|0;
 HEAP32[$this>>2] = $11;
 HEAP32[$3>>2] = $10;
 $12 = ((($__v)) + 8|0);
 $13 = HEAP32[$1>>2]|0;
 $14 = HEAP32[$12>>2]|0;
 HEAP32[$1>>2] = $14;
 HEAP32[$12>>2] = $13;
 $15 = ((($this)) + 8|0);
 $16 = ((($__v)) + 12|0);
 $17 = HEAP32[$15>>2]|0;
 $18 = HEAP32[$16>>2]|0;
 HEAP32[$15>>2] = $18;
 HEAP32[$16>>2] = $17;
 $19 = HEAP32[$3>>2]|0;
 HEAP32[$__v>>2] = $19;
 return;
}
function __ZNSt3__114__split_bufferIPNS_6locale5facetERNS_15__sso_allocatorIS3_Lj28EEEED2Ev($this) {
 $this = $this|0;
 var $$pre$i$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($this)) + 8|0);
 $$pre$i$i$i = HEAP32[$2>>2]|0;
 $4 = $$pre$i$i$i;
 while(1) {
  $3 = ($4|0)==($1|0);
  if ($3) {
   break;
  }
  $5 = ((($4)) + -4|0);
  HEAP32[$2>>2] = $5;
  $4 = $5;
 }
 $6 = HEAP32[$this>>2]|0;
 $7 = ($6|0)==(0|0);
 do {
  if (!($7)) {
   $8 = ((($this)) + 16|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = ($9|0)==($6|0);
   if ($10) {
    $11 = ((($9)) + 112|0);
    HEAP8[$11>>0] = 0;
    break;
   } else {
    __ZdlPv($6);
    break;
   }
  }
 } while(0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_7collateIwEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1207568)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_5ctypeIcEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1207576)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_5ctypeIwEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1207608)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_7codecvtIcc11__mbstate_tEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1209368)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__17codecvtIwc11__mbstate_tEC2Ej($this,$refs) {
 $this = $this|0;
 $refs = $refs|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($refs) + -1)|0;
 $1 = ((($this)) + 4|0);
 HEAP32[$1>>2] = $0;
 HEAP32[$this>>2] = (157248);
 __THREW__ = 0;
 $2 = (invoke_i(4874)|0);
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  ___resumeException($6|0);
  // unreachable;
 } else {
  $5 = ((($this)) + 8|0);
  HEAP32[$5>>2] = $2;
  return;
 }
}
function __ZNSt3__16locale5__imp7installINS_7codecvtIwc11__mbstate_tEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1209376)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_7codecvtIDsc11__mbstate_tEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1209384)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_7codecvtIDic11__mbstate_tEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1209392)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__18numpunctIcEC2Ej($this,$refs) {
 $this = $this|0;
 $refs = $refs|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $__i$0$i$i = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($refs) + -1)|0;
 $1 = ((($this)) + 4|0);
 HEAP32[$1>>2] = $0;
 HEAP32[$this>>2] = (157368);
 $2 = ((($this)) + 8|0);
 HEAP8[$2>>0] = 46;
 $3 = ((($this)) + 9|0);
 HEAP8[$3>>0] = 44;
 $4 = ((($this)) + 12|0);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $5 = (($4) + ($__i$0$i$i<<2)|0);
  HEAP32[$5>>2] = 0;
  $6 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $6;
 }
 return;
}
function __ZNSt3__16locale5__imp7installINS_8numpunctIcEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1207588)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__18numpunctIwEC2Ej($this,$refs) {
 $this = $this|0;
 $refs = $refs|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $__i$0$i$i = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($refs) + -1)|0;
 $1 = ((($this)) + 4|0);
 HEAP32[$1>>2] = $0;
 HEAP32[$this>>2] = (157408);
 $2 = ((($this)) + 8|0);
 HEAP32[$2>>2] = 46;
 $3 = ((($this)) + 12|0);
 HEAP32[$3>>2] = 44;
 $4 = ((($this)) + 16|0);
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $5 = (($4) + ($__i$0$i$i<<2)|0);
  HEAP32[$5>>2] = 0;
  $6 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $6;
 }
 return;
}
function __ZNSt3__16locale5__imp7installINS_8numpunctIwEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1207616)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_7num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1207600)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_7num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1207624)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_7num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1207632)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_7num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1207640)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_10moneypunctIcLb0EEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1209288)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_10moneypunctIcLb1EEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1209296)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_10moneypunctIwLb0EEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1209304)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_10moneypunctIwLb1EEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1209312)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_9money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1209320)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_9money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1209328)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_9money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1209336)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_9money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1209344)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_8time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1208452)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_8time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1209264)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_8time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1209272)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_8time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1209280)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_8messagesIcEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1209352)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale5__imp7installINS_8messagesIwEEEEvPT_($this,$f) {
 $this = $this|0;
 $f = $f|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale2id5__getEv(1209360)|0);
 __ZNSt3__16locale5__imp7installEPNS0_5facetEl($this,$f,$0);
 return;
}
function __ZNSt3__16locale7classicEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[625432]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((625432|0))|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    __THREW__ = 0;
    (invoke_i(4958)|0);
    $4 = __THREW__; __THREW__ = 0;
    $5 = $4&1;
    if ($5) {
     $7 = ___cxa_find_matching_catch_2()|0;
     $8 = tempRet0;
     ___cxa_guard_abort((625432|0));
     ___resumeException($7|0);
     // unreachable;
    } else {
     HEAP32[302351] = 1209400;
     ___cxa_guard_release((625432|0));
     break;
    }
   }
  }
 } while(0);
 $6 = HEAP32[302351]|0;
 return ($6|0);
}
function __ZNSt3__16locale5__imp12make_classicEv() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__16locale5__impC2Ej(625440,1);
 HEAP32[302350] = 625440;
 return (1209400|0);
}
function __ZNSt3__16locale5__imp11make_globalEv() {
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__16locale7classicEv()|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[302352] = $1;
 __ZNSt3__114__shared_count12__add_sharedEv($1);
 return (1209408|0);
}
function __ZNSt3__16locale8__globalEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[625600]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire((625600|0))|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    __THREW__ = 0;
    (invoke_i(4959)|0);
    $4 = __THREW__; __THREW__ = 0;
    $5 = $4&1;
    if ($5) {
     $7 = ___cxa_find_matching_catch_2()|0;
     $8 = tempRet0;
     ___cxa_guard_abort((625600|0));
     ___resumeException($7|0);
     // unreachable;
    } else {
     HEAP32[302353] = 1209408;
     ___cxa_guard_release((625600|0));
     break;
    }
   }
  }
 } while(0);
 $6 = HEAP32[302353]|0;
 return ($6|0);
}
function __ZNSt3__16localeC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __THREW__ = 0;
 $0 = (invoke_i(4960)|0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  ___clang_call_terminate($4);
  // unreachable;
 } else {
  $3 = HEAP32[$0>>2]|0;
  HEAP32[$this>>2] = $3;
  __ZNSt3__114__shared_count12__add_sharedEv($3);
  return;
 }
}
function __ZNSt3__16localeC2ERKS0_($this,$l) {
 $this = $this|0;
 $l = $l|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$l>>2]|0;
 HEAP32[$this>>2] = $0;
 __ZNSt3__114__shared_count12__add_sharedEv($0);
 return;
}
function __ZNSt3__16localeD2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 (__ZNSt3__114__shared_count16__release_sharedEv($0)|0);
 return;
}
function __ZNSt3__114__shared_countD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__114__shared_count12__add_sharedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;HEAP32[$0>>2] = (($1+1)|0);
 return;
}
function __ZNSt3__114__shared_count16__release_sharedEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;HEAP32[$0>>2] = (($1+-1)|0);
 $2 = ($1|0)==(0);
 if ($2) {
  $3 = HEAP32[$this>>2]|0;
  $4 = ((($3)) + 8|0);
  $5 = HEAP32[$4>>2]|0;
  FUNCTION_TABLE_vi[$5 & 8191]($this);
  $$0 = 1;
 } else {
  $$0 = 0;
 }
 return ($$0|0);
}
function __ZNSt3__114error_categoryD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNKSt3__114error_category23default_error_conditionEi($agg$result,$this,$ev) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $ev = $ev|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$agg$result>>2] = $ev;
 $0 = ((($agg$result)) + 4|0);
 HEAP32[$0>>2] = $this;
 return;
}
function __ZNKSt3__114error_category10equivalentEiRKNS_15error_conditionE($this,$code,$condition) {
 $this = $this|0;
 $code = $code|0;
 $condition = $condition|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = HEAP32[$this>>2]|0;
 $2 = ((($1)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 FUNCTION_TABLE_viii[$3 & 8191]($0,$this,$code);
 $4 = ((($0)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($condition)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($5|0)==($7|0);
 if ($8) {
  $9 = HEAP32[$0>>2]|0;
  $10 = HEAP32[$condition>>2]|0;
  $11 = ($9|0)==($10|0);
  $12 = $11;
 } else {
  $12 = 0;
 }
 STACKTOP = sp;return ($12|0);
}
function __ZNKSt3__114error_category10equivalentERKNS_10error_codeEi($this,$code,$condition) {
 $this = $this|0;
 $code = $code|0;
 $condition = $condition|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($code)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==($this|0);
 $3 = HEAP32[$code>>2]|0;
 $4 = ($3|0)==($condition|0);
 $5 = $2 & $4;
 return ($5|0);
}
function __ZNSt3__112system_errorD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt13runtime_errorD2Ev($this);
 return;
}
function __ZNSt3__112system_errorD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt13runtime_errorD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNKSt3__112__do_message7messageEi($agg$result,$this,$ev) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $ev = $ev|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_strerror($ev)|0);
 $1 = (_strlen($0)|0);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($agg$result,$0,$1);
 return;
}
function __ZNSt3__112system_error6__initERKNS_10error_codeENS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEE($agg$result,$ec,$what_arg) {
 $agg$result = $agg$result|0;
 $ec = $ec|0;
 $what_arg = $what_arg|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $__i$0$i$i = 0, $exitcond$i$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp;
 $1 = HEAP32[$ec>>2]|0;
 $2 = ($1|0)==(0);
 do {
  if (!($2)) {
   $3 = HEAP8[$what_arg>>0]|0;
   $4 = $3 & 1;
   $5 = ($4<<24>>24)==(0);
   if ($5) {
    $8 = $3&255;
    $9 = $8 >>> 1;
    $11 = $9;
   } else {
    $6 = ((($what_arg)) + 4|0);
    $7 = HEAP32[$6>>2]|0;
    $11 = $7;
   }
   $10 = ($11|0)==(0);
   if ($10) {
    $17 = $1;
   } else {
    (__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6appendEPKc($what_arg,622655)|0);
    $$pre = HEAP32[$ec>>2]|0;
    $17 = $$pre;
   }
   $12 = ((($ec)) + 4|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = HEAP32[$13>>2]|0;
   $15 = ((($14)) + 24|0);
   $16 = HEAP32[$15>>2]|0;
   FUNCTION_TABLE_viii[$16 & 8191]($0,$13,$17);
   $18 = HEAP8[$0>>0]|0;
   $19 = $18 & 1;
   $20 = ($19<<24>>24)==(0);
   $21 = ((($0)) + 8|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = ((($0)) + 1|0);
   $24 = $20 ? $23 : $22;
   $25 = ((($0)) + 4|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = $18&255;
   $28 = $27 >>> 1;
   $29 = $20 ? $28 : $26;
   __THREW__ = 0;
   (invoke_iiii(3462,($what_arg|0),($24|0),($29|0))|0);
   $30 = __THREW__; __THREW__ = 0;
   $31 = $30&1;
   if ($31) {
    $32 = ___cxa_find_matching_catch_2()|0;
    $33 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0);
    ___resumeException($32|0);
    // unreachable;
   } else {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0);
    break;
   }
  }
 } while(0);
 ;HEAP32[$agg$result>>2]=HEAP32[$what_arg>>2]|0;HEAP32[$agg$result+4>>2]=HEAP32[$what_arg+4>>2]|0;HEAP32[$agg$result+8>>2]=HEAP32[$what_arg+8>>2]|0;
 $__i$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($__i$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $34 = (($what_arg) + ($__i$0$i$i<<2)|0);
  HEAP32[$34>>2] = 0;
  $35 = (($__i$0$i$i) + 1)|0;
  $__i$0$i$i = $35;
 }
 STACKTOP = sp;return;
}
function __ZNSt3__112system_errorC2ENS_10error_codeEPKc($this,$ec,$what_arg) {
 $this = $this|0;
 $ec = $ec|0;
 $what_arg = $what_arg|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $0 = sp + 12|0;
 $1 = sp;
 $2 = (_strlen($what_arg)|0);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj($1,$what_arg,$2);
 __THREW__ = 0;
 invoke_viii(4961,($0|0),($ec|0),($1|0));
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 do {
  if ($4) {
   $18 = ___cxa_find_matching_catch_2()|0;
   $19 = tempRet0;
   $$0 = $19;$$01 = $18;
  } else {
   __THREW__ = 0;
   invoke_vii(4962,($this|0),($0|0));
   $5 = __THREW__; __THREW__ = 0;
   $6 = $5&1;
   if ($6) {
    $20 = ___cxa_find_matching_catch_2()|0;
    $21 = tempRet0;
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0);
    $$0 = $21;$$01 = $20;
    break;
   } else {
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0);
    __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1);
    HEAP32[$this>>2] = (157724);
    $7 = ((($this)) + 8|0);
    $8 = $ec;
    $9 = $8;
    $10 = HEAP32[$9>>2]|0;
    $11 = (($8) + 4)|0;
    $12 = $11;
    $13 = HEAP32[$12>>2]|0;
    $14 = $7;
    $15 = $14;
    HEAP32[$15>>2] = $10;
    $16 = (($14) + 4)|0;
    $17 = $16;
    HEAP32[$17>>2] = $13;
    STACKTOP = sp;return;
   }
  }
 } while(0);
 __ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1);
 ___resumeException($$01|0);
 // unreachable;
}
function __ZNSt3__111__call_onceERVmPvPFvS2_E($flag,$arg,$func) {
 $flag = $flag|0;
 $arg = $arg|0;
 $func = $func|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 (_pthread_mutex_lock((1209416|0))|0);
 while(1) {
  $0 = HEAP32[$flag>>2]|0;
  $1 = ($0|0)==(1);
  if (!($1)) {
   break;
  }
  (_pthread_cond_wait((1209444|0),(1209416|0))|0);
 }
 $2 = HEAP32[$flag>>2]|0;
 $3 = ($2|0)==(0);
 do {
  if ($3) {
   HEAP32[$flag>>2] = 1;
   __THREW__ = 0;
   (invoke_ii(4963,(1209416|0))|0);
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   if (!($5)) {
    __THREW__ = 0;
    invoke_vi($func|0,($arg|0));
    $6 = __THREW__; __THREW__ = 0;
    $7 = $6&1;
    if (!($7)) {
     __THREW__ = 0;
     (invoke_ii(4964,(1209416|0))|0);
     $8 = __THREW__; __THREW__ = 0;
     $9 = $8&1;
     if (!($9)) {
      HEAP32[$flag>>2] = -1;
      __THREW__ = 0;
      (invoke_ii(4963,(1209416|0))|0);
      $10 = __THREW__; __THREW__ = 0;
      $11 = $10&1;
      if (!($11)) {
       __THREW__ = 0;
       (invoke_ii(4965,(1209444|0))|0);
       $12 = __THREW__; __THREW__ = 0;
       $13 = $12&1;
       if (!($13)) {
        break;
       }
      }
     }
    }
   }
   $14 = ___cxa_find_matching_catch_3(0|0)|0;
   $15 = tempRet0;
   (___cxa_begin_catch(($14|0))|0);
   __THREW__ = 0;
   (invoke_ii(4964,(1209416|0))|0);
   $16 = __THREW__; __THREW__ = 0;
   $17 = $16&1;
   if (!($17)) {
    HEAP32[$flag>>2] = 0;
    __THREW__ = 0;
    (invoke_ii(4963,(1209416|0))|0);
    $18 = __THREW__; __THREW__ = 0;
    $19 = $18&1;
    if (!($19)) {
     __THREW__ = 0;
     (invoke_ii(4965,(1209444|0))|0);
     $20 = __THREW__; __THREW__ = 0;
     $21 = $20&1;
     if (!($21)) {
      __THREW__ = 0;
      invoke_v(3656);
      $22 = __THREW__; __THREW__ = 0;
     }
    }
   }
   $23 = ___cxa_find_matching_catch_2()|0;
   $24 = tempRet0;
   __THREW__ = 0;
   invoke_v(3378);
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    $27 = ___cxa_find_matching_catch_3(0|0)|0;
    $28 = tempRet0;
    ___clang_call_terminate($27);
    // unreachable;
   } else {
    ___resumeException($23|0);
    // unreachable;
   }
  } else {
   (_pthread_mutex_unlock((1209416|0))|0);
  }
 } while(0);
 return;
}
function __ZL25default_terminate_handlerv() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $thrown_object = 0, $vararg_buffer = 0, $vararg_buffer10 = 0;
 var $vararg_buffer3 = 0, $vararg_buffer7 = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer10 = sp + 32|0;
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $thrown_object = sp + 36|0;
 $0 = (___cxa_get_globals_fast()|0);
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  $2 = HEAP32[$0>>2]|0;
  $3 = ($2|0)==(0|0);
  if (!($3)) {
   $4 = ((($2)) + 80|0);
   $5 = ((($2)) + 48|0);
   $6 = $5;
   $7 = $6;
   $8 = HEAP32[$7>>2]|0;
   $9 = (($6) + 4)|0;
   $10 = $9;
   $11 = HEAP32[$10>>2]|0;
   $12 = $8 & -256;
   $13 = ($12|0)==(1126902528);
   $14 = ($11|0)==(1129074247);
   $15 = $13 & $14;
   if (!($15)) {
    $36 = HEAP32[39455]|0;
    HEAP32[$vararg_buffer7>>2] = $36;
    _abort_message(623035,$vararg_buffer7);
    // unreachable;
   }
   $16 = ($8|0)==(1126902529);
   $17 = ($11|0)==(1129074247);
   $18 = $16 & $17;
   if ($18) {
    $19 = ((($2)) + 44|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = $20;
   } else {
    $21 = $4;
   }
   HEAP32[$thrown_object>>2] = $21;
   $22 = HEAP32[$2>>2]|0;
   $23 = ((($22)) + 4|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = HEAP32[2248]|0;
   $26 = ((($25)) + 16|0);
   $27 = HEAP32[$26>>2]|0;
   $28 = (FUNCTION_TABLE_iiii[$27 & 8191](8992,$22,$thrown_object)|0);
   if ($28) {
    $29 = HEAP32[$thrown_object>>2]|0;
    $30 = HEAP32[39455]|0;
    $31 = HEAP32[$29>>2]|0;
    $32 = ((($31)) + 8|0);
    $33 = HEAP32[$32>>2]|0;
    $34 = (FUNCTION_TABLE_ii[$33 & 8191]($29)|0);
    HEAP32[$vararg_buffer>>2] = $30;
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
    HEAP32[$vararg_ptr1>>2] = $24;
    $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
    HEAP32[$vararg_ptr2>>2] = $34;
    _abort_message(622949,$vararg_buffer);
    // unreachable;
   } else {
    $35 = HEAP32[39455]|0;
    HEAP32[$vararg_buffer3>>2] = $35;
    $vararg_ptr6 = ((($vararg_buffer3)) + 4|0);
    HEAP32[$vararg_ptr6>>2] = $24;
    _abort_message(622994,$vararg_buffer3);
    // unreachable;
   }
  }
 }
 _abort_message(623073,$vararg_buffer10);
 // unreachable;
}
function ___cxa_get_globals_fast() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_pthread_once((1209492|0),(4966|0))|0);
 $1 = ($0|0)==(0);
 if ($1) {
  $2 = HEAP32[302374]|0;
  $3 = (_pthread_getspecific(($2|0))|0);
  STACKTOP = sp;return ($3|0);
 } else {
  _abort_message(622761,$vararg_buffer);
  // unreachable;
 }
 return (0)|0;
}
function __ZN10__cxxabiv112_GLOBAL__N_110construct_Ev() {
 var $0 = 0, $1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 $0 = (_pthread_key_create((1209496|0),(4967|0))|0);
 $1 = ($0|0)==(0);
 if ($1) {
  STACKTOP = sp;return;
 } else {
  _abort_message(622711,$vararg_buffer);
  // unreachable;
 }
}
function __ZN10__cxxabiv112_GLOBAL__N_19destruct_EPv($p) {
 $p = $p|0;
 var $0 = 0, $1 = 0, $2 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer = sp;
 _free($p);
 $0 = HEAP32[302374]|0;
 $1 = (_pthread_setspecific(($0|0),(0|0))|0);
 $2 = ($1|0)==(0);
 if ($2) {
  STACKTOP = sp;return;
 } else {
  _abort_message(622658,$vararg_buffer);
  // unreachable;
 }
}
function _abort_message($format,$varargs) {
 $format = $format|0;
 $varargs = $varargs|0;
 var $0 = 0, $list = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $list = sp;
 HEAP32[$list>>2] = $varargs;
 $0 = HEAP32[37544]|0;
 (_vfprintf($0,$format,$list)|0);
 (_fputc(10,$0)|0);
 _abort();
 // unreachable;
}
function __ZN10__cxxabiv116__shim_type_infoD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt9type_infoD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZN10__cxxabiv120__si_class_type_infoD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZdlPv($ptr) {
 $ptr = $ptr|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _free($ptr);
 return;
}
function __ZNK10__cxxabiv116__shim_type_info5noop1Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNK10__cxxabiv116__shim_type_info5noop2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNK10__cxxabiv117__class_type_info9can_catchEPKNS_16__shim_type_infoERPv($this,$thrown_type,$adjustedPtr) {
 $this = $this|0;
 $thrown_type = $thrown_type|0;
 $adjustedPtr = $adjustedPtr|0;
 var $$0 = 0, $$2 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $info = 0, dest = 0;
 var label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $info = sp;
 $0 = ($this|0)==($thrown_type|0);
 if ($0) {
  $$2 = 1;
 } else {
  $1 = ($thrown_type|0)==(0|0);
  if ($1) {
   $$2 = 0;
  } else {
   $2 = (___dynamic_cast($thrown_type,9032,9000,0)|0);
   $3 = ($2|0)==(0|0);
   if ($3) {
    $$2 = 0;
   } else {
    dest=$info; stop=dest+56|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
    HEAP32[$info>>2] = $2;
    $4 = ((($info)) + 8|0);
    HEAP32[$4>>2] = $this;
    $5 = ((($info)) + 12|0);
    HEAP32[$5>>2] = -1;
    $6 = ((($info)) + 48|0);
    HEAP32[$6>>2] = 1;
    $7 = HEAP32[$2>>2]|0;
    $8 = ((($7)) + 28|0);
    $9 = HEAP32[$8>>2]|0;
    $10 = HEAP32[$adjustedPtr>>2]|0;
    FUNCTION_TABLE_viiii[$9 & 8191]($2,$info,$10,1);
    $11 = ((($info)) + 24|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = ($12|0)==(1);
    if ($13) {
     $14 = ((($info)) + 16|0);
     $15 = HEAP32[$14>>2]|0;
     HEAP32[$adjustedPtr>>2] = $15;
     $$0 = 1;
    } else {
     $$0 = 0;
    }
    $$2 = $$0;
   }
  }
 }
 STACKTOP = sp;return ($$2|0);
}
function ___dynamic_cast($static_ptr,$static_type,$dst_type,$src2dst_offset) {
 $static_ptr = $static_ptr|0;
 $static_type = $static_type|0;
 $dst_type = $dst_type|0;
 $src2dst_offset = $src2dst_offset|0;
 var $$ = 0, $$8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $dst_ptr$0 = 0, $info = 0, $or$cond = 0, $or$cond3 = 0, $or$cond5 = 0, $or$cond7 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $info = sp;
 $0 = HEAP32[$static_ptr>>2]|0;
 $1 = ((($0)) + -8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($static_ptr) + ($2)|0);
 $4 = ((($0)) + -4|0);
 $5 = HEAP32[$4>>2]|0;
 HEAP32[$info>>2] = $dst_type;
 $6 = ((($info)) + 4|0);
 HEAP32[$6>>2] = $static_ptr;
 $7 = ((($info)) + 8|0);
 HEAP32[$7>>2] = $static_type;
 $8 = ((($info)) + 12|0);
 HEAP32[$8>>2] = $src2dst_offset;
 $9 = ((($info)) + 16|0);
 $10 = ((($info)) + 20|0);
 $11 = ((($info)) + 24|0);
 $12 = ((($info)) + 28|0);
 $13 = ((($info)) + 32|0);
 $14 = ((($info)) + 40|0);
 $15 = ($5|0)==($dst_type|0);
 dest=$9; stop=dest+36|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));HEAP16[$9+36>>1]=0|0;HEAP8[$9+38>>0]=0|0;
 L1: do {
  if ($15) {
   $16 = ((($info)) + 48|0);
   HEAP32[$16>>2] = 1;
   $17 = HEAP32[$dst_type>>2]|0;
   $18 = ((($17)) + 20|0);
   $19 = HEAP32[$18>>2]|0;
   FUNCTION_TABLE_viiiiii[$19 & 8191]($dst_type,$info,$3,$3,1,0);
   $20 = HEAP32[$11>>2]|0;
   $21 = ($20|0)==(1);
   $$ = $21 ? $3 : 0;
   $dst_ptr$0 = $$;
  } else {
   $22 = ((($info)) + 36|0);
   $23 = HEAP32[$5>>2]|0;
   $24 = ((($23)) + 24|0);
   $25 = HEAP32[$24>>2]|0;
   FUNCTION_TABLE_viiiii[$25 & 8191]($5,$info,$3,1,0);
   $26 = HEAP32[$22>>2]|0;
   switch ($26|0) {
   case 0:  {
    $27 = HEAP32[$14>>2]|0;
    $28 = ($27|0)==(1);
    $29 = HEAP32[$12>>2]|0;
    $30 = ($29|0)==(1);
    $or$cond = $28 & $30;
    $31 = HEAP32[$13>>2]|0;
    $32 = ($31|0)==(1);
    $or$cond3 = $or$cond & $32;
    $33 = HEAP32[$10>>2]|0;
    $$8 = $or$cond3 ? $33 : 0;
    $dst_ptr$0 = $$8;
    break L1;
    break;
   }
   case 1:  {
    break;
   }
   default: {
    $dst_ptr$0 = 0;
    break L1;
   }
   }
   $34 = HEAP32[$11>>2]|0;
   $35 = ($34|0)==(1);
   if (!($35)) {
    $36 = HEAP32[$14>>2]|0;
    $37 = ($36|0)==(0);
    $38 = HEAP32[$12>>2]|0;
    $39 = ($38|0)==(1);
    $or$cond5 = $37 & $39;
    $40 = HEAP32[$13>>2]|0;
    $41 = ($40|0)==(1);
    $or$cond7 = $or$cond5 & $41;
    if (!($or$cond7)) {
     $dst_ptr$0 = 0;
     break;
    }
   }
   $42 = HEAP32[$9>>2]|0;
   $dst_ptr$0 = $42;
  }
 } while(0);
 STACKTOP = sp;return ($dst_ptr$0|0);
}
function __ZNK10__cxxabiv120__si_class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib($this,$info,$dst_ptr,$current_ptr,$path_below,$use_strcmp) {
 $this = $this|0;
 $info = $info|0;
 $dst_ptr = $dst_ptr|0;
 $current_ptr = $current_ptr|0;
 $path_below = $path_below|0;
 $use_strcmp = $use_strcmp|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($this|0)==($1|0);
 if ($2) {
  __ZNK10__cxxabiv117__class_type_info29process_static_type_above_dstEPNS_19__dynamic_cast_infoEPKvS4_i(0,$info,$dst_ptr,$current_ptr,$path_below);
 } else {
  $3 = ((($this)) + 8|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = HEAP32[$4>>2]|0;
  $6 = ((($5)) + 20|0);
  $7 = HEAP32[$6>>2]|0;
  FUNCTION_TABLE_viiiiii[$7 & 8191]($4,$info,$dst_ptr,$current_ptr,$path_below,$use_strcmp);
 }
 return;
}
function __ZNK10__cxxabiv117__class_type_info29process_static_type_above_dstEPNS_19__dynamic_cast_infoEPKvS4_i($this,$info,$dst_ptr,$current_ptr,$path_below) {
 $this = $this|0;
 $info = $info|0;
 $dst_ptr = $dst_ptr|0;
 $current_ptr = $current_ptr|0;
 $path_below = $path_below|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 53|0);
 HEAP8[$0>>0] = 1;
 $1 = ((($info)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==($current_ptr|0);
 do {
  if ($3) {
   $4 = ((($info)) + 52|0);
   HEAP8[$4>>0] = 1;
   $5 = ((($info)) + 16|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($6|0)==(0|0);
   if ($7) {
    HEAP32[$5>>2] = $dst_ptr;
    $8 = ((($info)) + 24|0);
    HEAP32[$8>>2] = $path_below;
    $9 = ((($info)) + 36|0);
    HEAP32[$9>>2] = 1;
    $10 = ((($info)) + 48|0);
    $11 = HEAP32[$10>>2]|0;
    $12 = ($11|0)==(1);
    $13 = ($path_below|0)==(1);
    $or$cond = $12 & $13;
    if (!($or$cond)) {
     break;
    }
    $14 = ((($info)) + 54|0);
    HEAP8[$14>>0] = 1;
    break;
   }
   $15 = ($6|0)==($dst_ptr|0);
   if (!($15)) {
    $25 = ((($info)) + 36|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = (($26) + 1)|0;
    HEAP32[$25>>2] = $27;
    $28 = ((($info)) + 54|0);
    HEAP8[$28>>0] = 1;
    break;
   }
   $16 = ((($info)) + 24|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ($17|0)==(2);
   if ($18) {
    HEAP32[$16>>2] = $path_below;
    $23 = $path_below;
   } else {
    $23 = $17;
   }
   $19 = ((($info)) + 48|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = ($20|0)==(1);
   $22 = ($23|0)==(1);
   $or$cond1 = $21 & $22;
   if ($or$cond1) {
    $24 = ((($info)) + 54|0);
    HEAP8[$24>>0] = 1;
   }
  }
 } while(0);
 return;
}
function __ZNK10__cxxabiv120__si_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib($this,$info,$current_ptr,$path_below,$use_strcmp) {
 $this = $this|0;
 $info = $info|0;
 $current_ptr = $current_ptr|0;
 $path_below = $path_below|0;
 $use_strcmp = $use_strcmp|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $is_dst_type_derived_from_static_type$0$off02 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($this|0)==($1|0);
 do {
  if ($2) {
   $3 = ((($info)) + 4|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==($current_ptr|0);
   if ($5) {
    $6 = ((($info)) + 28|0);
    $7 = HEAP32[$6>>2]|0;
    $8 = ($7|0)==(1);
    if (!($8)) {
     HEAP32[$6>>2] = $path_below;
    }
   }
  } else {
   $9 = HEAP32[$info>>2]|0;
   $10 = ($this|0)==($9|0);
   if (!($10)) {
    $44 = ((($this)) + 8|0);
    $45 = HEAP32[$44>>2]|0;
    $46 = HEAP32[$45>>2]|0;
    $47 = ((($46)) + 24|0);
    $48 = HEAP32[$47>>2]|0;
    FUNCTION_TABLE_viiiii[$48 & 8191]($45,$info,$current_ptr,$path_below,$use_strcmp);
    break;
   }
   $11 = ((($info)) + 16|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($12|0)==($current_ptr|0);
   if (!($13)) {
    $14 = ((($info)) + 20|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = ($15|0)==($current_ptr|0);
    if (!($16)) {
     $19 = ((($info)) + 32|0);
     HEAP32[$19>>2] = $path_below;
     $20 = ((($info)) + 44|0);
     $21 = HEAP32[$20>>2]|0;
     $22 = ($21|0)==(4);
     if ($22) {
      break;
     }
     $23 = ((($info)) + 52|0);
     HEAP8[$23>>0] = 0;
     $24 = ((($info)) + 53|0);
     HEAP8[$24>>0] = 0;
     $25 = ((($this)) + 8|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = HEAP32[$26>>2]|0;
     $28 = ((($27)) + 20|0);
     $29 = HEAP32[$28>>2]|0;
     FUNCTION_TABLE_viiiiii[$29 & 8191]($26,$info,$current_ptr,$current_ptr,1,$use_strcmp);
     $30 = HEAP8[$24>>0]|0;
     $31 = ($30<<24>>24)==(0);
     if ($31) {
      $is_dst_type_derived_from_static_type$0$off02 = 0;
      label = 13;
     } else {
      $32 = HEAP8[$23>>0]|0;
      $not$ = ($32<<24>>24)==(0);
      if ($not$) {
       $is_dst_type_derived_from_static_type$0$off02 = 1;
       label = 13;
      } else {
       label = 17;
      }
     }
     do {
      if ((label|0) == 13) {
       HEAP32[$14>>2] = $current_ptr;
       $33 = ((($info)) + 40|0);
       $34 = HEAP32[$33>>2]|0;
       $35 = (($34) + 1)|0;
       HEAP32[$33>>2] = $35;
       $36 = ((($info)) + 36|0);
       $37 = HEAP32[$36>>2]|0;
       $38 = ($37|0)==(1);
       if ($38) {
        $39 = ((($info)) + 24|0);
        $40 = HEAP32[$39>>2]|0;
        $41 = ($40|0)==(2);
        if ($41) {
         $42 = ((($info)) + 54|0);
         HEAP8[$42>>0] = 1;
         if ($is_dst_type_derived_from_static_type$0$off02) {
          label = 17;
          break;
         } else {
          $43 = 4;
          break;
         }
        }
       }
       if ($is_dst_type_derived_from_static_type$0$off02) {
        label = 17;
       } else {
        $43 = 4;
       }
      }
     } while(0);
     if ((label|0) == 17) {
      $43 = 3;
     }
     HEAP32[$20>>2] = $43;
     break;
    }
   }
   $17 = ($path_below|0)==(1);
   if ($17) {
    $18 = ((($info)) + 32|0);
    HEAP32[$18>>2] = 1;
   }
  }
 } while(0);
 return;
}
function __ZNK10__cxxabiv120__si_class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi($this,$info,$adjustedPtr,$path_below) {
 $this = $this|0;
 $info = $info|0;
 $adjustedPtr = $adjustedPtr|0;
 $path_below = $path_below|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($this|0)==($1|0);
 if ($2) {
  __ZNK10__cxxabiv117__class_type_info24process_found_base_classEPNS_19__dynamic_cast_infoEPvi(0,$info,$adjustedPtr,$path_below);
 } else {
  $3 = ((($this)) + 8|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = HEAP32[$4>>2]|0;
  $6 = ((($5)) + 28|0);
  $7 = HEAP32[$6>>2]|0;
  FUNCTION_TABLE_viiii[$7 & 8191]($4,$info,$adjustedPtr,$path_below);
 }
 return;
}
function __ZNK10__cxxabiv117__class_type_info24process_found_base_classEPNS_19__dynamic_cast_infoEPvi($this,$info,$adjustedPtr,$path_below) {
 $this = $this|0;
 $info = $info|0;
 $adjustedPtr = $adjustedPtr|0;
 $path_below = $path_below|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 do {
  if ($2) {
   HEAP32[$0>>2] = $adjustedPtr;
   $3 = ((($info)) + 24|0);
   HEAP32[$3>>2] = $path_below;
   $4 = ((($info)) + 36|0);
   HEAP32[$4>>2] = 1;
  } else {
   $5 = ($1|0)==($adjustedPtr|0);
   if (!($5)) {
    $9 = ((($info)) + 36|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = (($10) + 1)|0;
    HEAP32[$9>>2] = $11;
    $12 = ((($info)) + 24|0);
    HEAP32[$12>>2] = 2;
    $13 = ((($info)) + 54|0);
    HEAP8[$13>>0] = 1;
    break;
   }
   $6 = ((($info)) + 24|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(2);
   if ($8) {
    HEAP32[$6>>2] = $path_below;
   }
  }
 } while(0);
 return;
}
function __ZN10__cxxabiv117__class_type_infoD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNK10__cxxabiv117__class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib($this,$info,$dst_ptr,$current_ptr,$path_below,$use_strcmp) {
 $this = $this|0;
 $info = $info|0;
 $dst_ptr = $dst_ptr|0;
 $current_ptr = $current_ptr|0;
 $path_below = $path_below|0;
 $use_strcmp = $use_strcmp|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($this|0)==($1|0);
 if ($2) {
  __ZNK10__cxxabiv117__class_type_info29process_static_type_above_dstEPNS_19__dynamic_cast_infoEPKvS4_i(0,$info,$dst_ptr,$current_ptr,$path_below);
 }
 return;
}
function __ZNK10__cxxabiv117__class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib($this,$info,$current_ptr,$path_below,$use_strcmp) {
 $this = $this|0;
 $info = $info|0;
 $current_ptr = $current_ptr|0;
 $path_below = $path_below|0;
 $use_strcmp = $use_strcmp|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($this|0)==($1|0);
 do {
  if ($2) {
   $3 = ((($info)) + 4|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==($current_ptr|0);
   if ($5) {
    $6 = ((($info)) + 28|0);
    $7 = HEAP32[$6>>2]|0;
    $8 = ($7|0)==(1);
    if (!($8)) {
     HEAP32[$6>>2] = $path_below;
    }
   }
  } else {
   $9 = HEAP32[$info>>2]|0;
   $10 = ($this|0)==($9|0);
   if ($10) {
    $11 = ((($info)) + 16|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = ($12|0)==($current_ptr|0);
    if (!($13)) {
     $14 = ((($info)) + 20|0);
     $15 = HEAP32[$14>>2]|0;
     $16 = ($15|0)==($current_ptr|0);
     if (!($16)) {
      $19 = ((($info)) + 32|0);
      HEAP32[$19>>2] = $path_below;
      HEAP32[$14>>2] = $current_ptr;
      $20 = ((($info)) + 40|0);
      $21 = HEAP32[$20>>2]|0;
      $22 = (($21) + 1)|0;
      HEAP32[$20>>2] = $22;
      $23 = ((($info)) + 36|0);
      $24 = HEAP32[$23>>2]|0;
      $25 = ($24|0)==(1);
      if ($25) {
       $26 = ((($info)) + 24|0);
       $27 = HEAP32[$26>>2]|0;
       $28 = ($27|0)==(2);
       if ($28) {
        $29 = ((($info)) + 54|0);
        HEAP8[$29>>0] = 1;
       }
      }
      $30 = ((($info)) + 44|0);
      HEAP32[$30>>2] = 4;
      break;
     }
    }
    $17 = ($path_below|0)==(1);
    if ($17) {
     $18 = ((($info)) + 32|0);
     HEAP32[$18>>2] = 1;
    }
   }
  }
 } while(0);
 return;
}
function __ZNK10__cxxabiv117__class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi($this,$info,$adjustedPtr,$path_below) {
 $this = $this|0;
 $info = $info|0;
 $adjustedPtr = $adjustedPtr|0;
 $path_below = $path_below|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($this|0)==($1|0);
 if ($2) {
  __ZNK10__cxxabiv117__class_type_info24process_found_base_classEPNS_19__dynamic_cast_infoEPvi(0,$info,$adjustedPtr,$path_below);
 }
 return;
}
function __ZSt9terminatev() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 __THREW__ = 0;
 $0 = (invoke_i(4968)|0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $45 = ___cxa_find_matching_catch_3(0|0)|0;
  $46 = tempRet0;
  ___clang_call_terminate($45);
  // unreachable;
 }
 $3 = ($0|0)==(0|0);
 if (!($3)) {
  $4 = HEAP32[$0>>2]|0;
  $5 = ($4|0)==(0|0);
  if (!($5)) {
   $6 = ((($4)) + 48|0);
   $7 = $6;
   $8 = $7;
   $9 = HEAP32[$8>>2]|0;
   $10 = (($7) + 4)|0;
   $11 = $10;
   $12 = HEAP32[$11>>2]|0;
   $13 = $9 & -256;
   $14 = ($13|0)==(1126902528);
   $15 = ($12|0)==(1129074247);
   $16 = $14 & $15;
   if ($16) {
    $17 = ((($4)) + 12|0);
    $18 = HEAP32[$17>>2]|0;
    __THREW__ = 0;
    invoke_v($18|0);
    $19 = __THREW__; __THREW__ = 0;
    $20 = $19&1;
    if (!($20)) {
     __THREW__ = 0;
     invoke_vii(4969,(623085|0),($vararg_buffer|0));
     $21 = __THREW__; __THREW__ = 0;
    }
    $22 = ___cxa_find_matching_catch_3(0|0)|0;
    $23 = tempRet0;
    (___cxa_begin_catch(($22|0))|0);
    __THREW__ = 0;
    invoke_vii(4969,(623125|0),($vararg_buffer1|0));
    $24 = __THREW__; __THREW__ = 0;
    $25 = ___cxa_find_matching_catch_3(0|0)|0;
    $26 = tempRet0;
    __THREW__ = 0;
    invoke_v(3378);
    $27 = __THREW__; __THREW__ = 0;
    $28 = $27&1;
    if ($28) {
     $29 = ___cxa_find_matching_catch_3(0|0)|0;
     $30 = tempRet0;
     ___clang_call_terminate($29);
     // unreachable;
    } else {
     ___clang_call_terminate($25);
     // unreachable;
    }
   }
  }
 }
 $31 = HEAP32[39434]|0;HEAP32[39434] = (($31+0)|0);
 $32 = $31;
 __THREW__ = 0;
 invoke_v($32|0);
 $33 = __THREW__; __THREW__ = 0;
 $34 = $33&1;
 if (!($34)) {
  __THREW__ = 0;
  invoke_vii(4969,(623085|0),($vararg_buffer3|0));
  $35 = __THREW__; __THREW__ = 0;
 }
 $36 = ___cxa_find_matching_catch_3(0|0)|0;
 $37 = tempRet0;
 (___cxa_begin_catch(($36|0))|0);
 __THREW__ = 0;
 invoke_vii(4969,(623125|0),($vararg_buffer5|0));
 $38 = __THREW__; __THREW__ = 0;
 $39 = ___cxa_find_matching_catch_3(0|0)|0;
 $40 = tempRet0;
 __THREW__ = 0;
 invoke_v(3378);
 $41 = __THREW__; __THREW__ = 0;
 $42 = $41&1;
 if ($42) {
  $43 = ___cxa_find_matching_catch_3(0|0)|0;
  $44 = tempRet0;
  ___clang_call_terminate($43);
  // unreachable;
 } else {
  ___clang_call_terminate($39);
  // unreachable;
 }
}
function __ZNSt9bad_allocD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt9exceptionD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt9bad_allocD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt9bad_alloc4whatEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (623188|0);
}
function __ZNSt11logic_errorD2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (157852);
 $0 = ((($this)) + 4|0);
 __ZN12_GLOBAL__N_114__libcpp_nmstrD2Ev($0);
 return;
}
function __ZN12_GLOBAL__N_114__libcpp_nmstrD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + -4|0);
 $2 = HEAP32[$1>>2]|0;HEAP32[$1>>2] = (($2+-1)|0);
 $3 = (($2) + -1)|0;
 $4 = ($3|0)<(0);
 if ($4) {
  $5 = HEAP32[$this>>2]|0;
  $6 = ((($5)) + -12|0);
  __ZdlPv($6);
 }
 return;
}
function __ZNSt11logic_errorD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt11logic_errorD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNKSt11logic_error4whatEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNSt13runtime_errorD2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (157872);
 $0 = ((($this)) + 4|0);
 __ZN12_GLOBAL__N_114__libcpp_nmstrD2Ev($0);
 return;
}
function __ZNSt13runtime_errorD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt13runtime_errorD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNKSt13runtime_error4whatEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNSt12length_errorD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt11logic_errorD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNSt12out_of_rangeD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt11logic_errorD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNSt14overflow_errorD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt13runtime_errorD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNSt8bad_castD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt8bad_castD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt8bad_cast4whatEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (623302|0);
}
function __ZNSt10bad_typeidD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt10bad_typeidD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNKSt10bad_typeid4whatEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (623331|0);
}
function __ZN10__cxxabiv121__vmi_class_type_infoD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZNK10__cxxabiv121__vmi_class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib($this,$info,$dst_ptr,$current_ptr,$path_below,$use_strcmp) {
 $this = $this|0;
 $info = $info|0;
 $dst_ptr = $dst_ptr|0;
 $current_ptr = $current_ptr|0;
 $path_below = $path_below|0;
 $use_strcmp = $use_strcmp|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($this|0)==($1|0);
 if ($2) {
  __ZNK10__cxxabiv117__class_type_info29process_static_type_above_dstEPNS_19__dynamic_cast_infoEPKvS4_i(0,$info,$dst_ptr,$current_ptr,$path_below);
 } else {
  $3 = ((($info)) + 52|0);
  $4 = HEAP16[$3>>1]|0;
  $5 = $4&255;
  $6 = ((($info)) + 53|0);
  $7 = ($4&65535) >>> 8;
  $8 = $7&255;
  $9 = ((($this)) + 16|0);
  $10 = ((($this)) + 12|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (((($this)) + 16|0) + ($11<<3)|0);
  HEAP8[$3>>0] = 0;
  HEAP8[$6>>0] = 0;
  __ZNK10__cxxabiv122__base_class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib($9,$info,$dst_ptr,$current_ptr,$path_below,$use_strcmp);
  $13 = ($11|0)>(1);
  L4: do {
   if ($13) {
    $14 = ((($this)) + 24|0);
    $15 = ((($info)) + 24|0);
    $16 = ((($this)) + 8|0);
    $17 = ((($info)) + 54|0);
    $p$0 = $14;
    while(1) {
     $18 = HEAP8[$17>>0]|0;
     $19 = ($18<<24>>24)==(0);
     if (!($19)) {
      break L4;
     }
     $20 = HEAP16[$3>>1]|0;
     $21 = $20&255;
     $22 = ($21<<24>>24)==(0);
     if ($22) {
      $28 = ($20&65535)<(256);
      if (!($28)) {
       $29 = HEAP32[$16>>2]|0;
       $30 = $29 & 1;
       $31 = ($30|0)==(0);
       if ($31) {
        break L4;
       }
      }
     } else {
      $23 = HEAP32[$15>>2]|0;
      $24 = ($23|0)==(1);
      if ($24) {
       break L4;
      }
      $25 = HEAP32[$16>>2]|0;
      $26 = $25 & 2;
      $27 = ($26|0)==(0);
      if ($27) {
       break L4;
      }
     }
     HEAP8[$3>>0] = 0;
     HEAP8[$6>>0] = 0;
     __ZNK10__cxxabiv122__base_class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib($p$0,$info,$dst_ptr,$current_ptr,$path_below,$use_strcmp);
     $32 = ((($p$0)) + 8|0);
     $33 = ($32>>>0)<($12>>>0);
     if ($33) {
      $p$0 = $32;
     } else {
      break;
     }
    }
   }
  } while(0);
  HEAP8[$3>>0] = $5;
  HEAP8[$6>>0] = $8;
 }
 return;
}
function __ZNK10__cxxabiv122__base_class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib($this,$info,$dst_ptr,$current_ptr,$path_below,$use_strcmp) {
 $this = $this|0;
 $info = $info|0;
 $dst_ptr = $dst_ptr|0;
 $current_ptr = $current_ptr|0;
 $path_below = $path_below|0;
 $use_strcmp = $use_strcmp|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $offset_to_base$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 >> 8;
 $3 = $1 & 1;
 $4 = ($3|0)==(0);
 if ($4) {
  $offset_to_base$0 = $2;
 } else {
  $5 = HEAP32[$current_ptr>>2]|0;
  $6 = (($5) + ($2)|0);
  $7 = HEAP32[$6>>2]|0;
  $offset_to_base$0 = $7;
 }
 $8 = HEAP32[$this>>2]|0;
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 20|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($current_ptr) + ($offset_to_base$0)|0);
 $13 = $1 & 2;
 $14 = ($13|0)!=(0);
 $15 = $14 ? $path_below : 2;
 FUNCTION_TABLE_viiiiii[$11 & 8191]($8,$info,$dst_ptr,$12,$15,$use_strcmp);
 return;
}
function __ZNK10__cxxabiv121__vmi_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib($this,$info,$current_ptr,$path_below,$use_strcmp) {
 $this = $this|0;
 $info = $info|0;
 $current_ptr = $current_ptr|0;
 $path_below = $path_below|0;
 $use_strcmp = $use_strcmp|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $does_dst_type_point_to_our_static_type$0$off0 = 0, $does_dst_type_point_to_our_static_type$0$off0$lcssa = 0, $does_dst_type_point_to_our_static_type$1$off0 = 0, $is_dst_type_derived_from_static_type$0$off0 = 0, $is_dst_type_derived_from_static_type$1$off0 = 0, $is_dst_type_derived_from_static_type$2$off0 = 0;
 var $p$0 = 0, $p2$0 = 0, $p2$1 = 0, $p2$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($this|0)==($1|0);
 L1: do {
  if ($2) {
   $3 = ((($info)) + 4|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==($current_ptr|0);
   if ($5) {
    $6 = ((($info)) + 28|0);
    $7 = HEAP32[$6>>2]|0;
    $8 = ($7|0)==(1);
    if (!($8)) {
     HEAP32[$6>>2] = $path_below;
    }
   }
  } else {
   $9 = HEAP32[$info>>2]|0;
   $10 = ($this|0)==($9|0);
   if (!($10)) {
    $57 = ((($this)) + 16|0);
    $58 = ((($this)) + 12|0);
    $59 = HEAP32[$58>>2]|0;
    $60 = (((($this)) + 16|0) + ($59<<3)|0);
    __ZNK10__cxxabiv122__base_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib($57,$info,$current_ptr,$path_below,$use_strcmp);
    $61 = ((($this)) + 24|0);
    $62 = ($59|0)>(1);
    if (!($62)) {
     break;
    }
    $63 = ((($this)) + 8|0);
    $64 = HEAP32[$63>>2]|0;
    $65 = $64 & 2;
    $66 = ($65|0)==(0);
    if ($66) {
     $67 = ((($info)) + 36|0);
     $68 = HEAP32[$67>>2]|0;
     $69 = ($68|0)==(1);
     if (!($69)) {
      $75 = $64 & 1;
      $76 = ($75|0)==(0);
      if ($76) {
       $79 = ((($info)) + 54|0);
       $p2$2 = $61;
       while(1) {
        $88 = HEAP8[$79>>0]|0;
        $89 = ($88<<24>>24)==(0);
        if (!($89)) {
         break L1;
        }
        $90 = HEAP32[$67>>2]|0;
        $91 = ($90|0)==(1);
        if ($91) {
         break L1;
        }
        __ZNK10__cxxabiv122__base_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib($p2$2,$info,$current_ptr,$path_below,$use_strcmp);
        $92 = ((($p2$2)) + 8|0);
        $93 = ($92>>>0)<($60>>>0);
        if ($93) {
         $p2$2 = $92;
        } else {
         break L1;
        }
       }
      }
      $77 = ((($info)) + 24|0);
      $78 = ((($info)) + 54|0);
      $p2$1 = $61;
      while(1) {
       $80 = HEAP8[$78>>0]|0;
       $81 = ($80<<24>>24)==(0);
       if (!($81)) {
        break L1;
       }
       $82 = HEAP32[$67>>2]|0;
       $83 = ($82|0)==(1);
       if ($83) {
        $84 = HEAP32[$77>>2]|0;
        $85 = ($84|0)==(1);
        if ($85) {
         break L1;
        }
       }
       __ZNK10__cxxabiv122__base_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib($p2$1,$info,$current_ptr,$path_below,$use_strcmp);
       $86 = ((($p2$1)) + 8|0);
       $87 = ($86>>>0)<($60>>>0);
       if ($87) {
        $p2$1 = $86;
       } else {
        break L1;
       }
      }
     }
    }
    $70 = ((($info)) + 54|0);
    $p2$0 = $61;
    while(1) {
     $71 = HEAP8[$70>>0]|0;
     $72 = ($71<<24>>24)==(0);
     if (!($72)) {
      break L1;
     }
     __ZNK10__cxxabiv122__base_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib($p2$0,$info,$current_ptr,$path_below,$use_strcmp);
     $73 = ((($p2$0)) + 8|0);
     $74 = ($73>>>0)<($60>>>0);
     if ($74) {
      $p2$0 = $73;
     } else {
      break L1;
     }
    }
   }
   $11 = ((($info)) + 16|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($12|0)==($current_ptr|0);
   if (!($13)) {
    $14 = ((($info)) + 20|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = ($15|0)==($current_ptr|0);
    if (!($16)) {
     $19 = ((($info)) + 32|0);
     HEAP32[$19>>2] = $path_below;
     $20 = ((($info)) + 44|0);
     $21 = HEAP32[$20>>2]|0;
     $22 = ($21|0)==(4);
     if ($22) {
      break;
     }
     $23 = ((($this)) + 16|0);
     $24 = ((($this)) + 12|0);
     $25 = HEAP32[$24>>2]|0;
     $26 = (((($this)) + 16|0) + ($25<<3)|0);
     $27 = ((($info)) + 52|0);
     $28 = ((($info)) + 53|0);
     $29 = ((($info)) + 54|0);
     $30 = ((($this)) + 8|0);
     $31 = ((($info)) + 24|0);
     $does_dst_type_point_to_our_static_type$0$off0 = 0;$is_dst_type_derived_from_static_type$0$off0 = 0;$p$0 = $23;
     L34: while(1) {
      $32 = ($p$0>>>0)<($26>>>0);
      if (!($32)) {
       $does_dst_type_point_to_our_static_type$0$off0$lcssa = $does_dst_type_point_to_our_static_type$0$off0;$is_dst_type_derived_from_static_type$2$off0 = $is_dst_type_derived_from_static_type$0$off0;
       label = 20;
       break;
      }
      HEAP8[$27>>0] = 0;
      HEAP8[$28>>0] = 0;
      __ZNK10__cxxabiv122__base_class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib($p$0,$info,$current_ptr,$current_ptr,1,$use_strcmp);
      $33 = HEAP8[$29>>0]|0;
      $34 = ($33<<24>>24)==(0);
      if (!($34)) {
       $does_dst_type_point_to_our_static_type$0$off0$lcssa = $does_dst_type_point_to_our_static_type$0$off0;$is_dst_type_derived_from_static_type$2$off0 = $is_dst_type_derived_from_static_type$0$off0;
       label = 20;
       break;
      }
      $35 = HEAP8[$28>>0]|0;
      $36 = ($35<<24>>24)==(0);
      do {
       if ($36) {
        $does_dst_type_point_to_our_static_type$1$off0 = $does_dst_type_point_to_our_static_type$0$off0;$is_dst_type_derived_from_static_type$1$off0 = $is_dst_type_derived_from_static_type$0$off0;
       } else {
        $37 = HEAP8[$27>>0]|0;
        $38 = ($37<<24>>24)==(0);
        if ($38) {
         $44 = HEAP32[$30>>2]|0;
         $45 = $44 & 1;
         $46 = ($45|0)==(0);
         if ($46) {
          $does_dst_type_point_to_our_static_type$0$off0$lcssa = $does_dst_type_point_to_our_static_type$0$off0;$is_dst_type_derived_from_static_type$2$off0 = 1;
          label = 20;
          break L34;
         } else {
          $does_dst_type_point_to_our_static_type$1$off0 = $does_dst_type_point_to_our_static_type$0$off0;$is_dst_type_derived_from_static_type$1$off0 = 1;
          break;
         }
        }
        $39 = HEAP32[$31>>2]|0;
        $40 = ($39|0)==(1);
        if ($40) {
         label = 25;
         break L34;
        }
        $41 = HEAP32[$30>>2]|0;
        $42 = $41 & 2;
        $43 = ($42|0)==(0);
        if ($43) {
         label = 25;
         break L34;
        } else {
         $does_dst_type_point_to_our_static_type$1$off0 = 1;$is_dst_type_derived_from_static_type$1$off0 = 1;
        }
       }
      } while(0);
      $47 = ((($p$0)) + 8|0);
      $does_dst_type_point_to_our_static_type$0$off0 = $does_dst_type_point_to_our_static_type$1$off0;$is_dst_type_derived_from_static_type$0$off0 = $is_dst_type_derived_from_static_type$1$off0;$p$0 = $47;
     }
     do {
      if ((label|0) == 20) {
       if (!($does_dst_type_point_to_our_static_type$0$off0$lcssa)) {
        HEAP32[$14>>2] = $current_ptr;
        $48 = ((($info)) + 40|0);
        $49 = HEAP32[$48>>2]|0;
        $50 = (($49) + 1)|0;
        HEAP32[$48>>2] = $50;
        $51 = ((($info)) + 36|0);
        $52 = HEAP32[$51>>2]|0;
        $53 = ($52|0)==(1);
        if ($53) {
         $54 = HEAP32[$31>>2]|0;
         $55 = ($54|0)==(2);
         if ($55) {
          HEAP8[$29>>0] = 1;
          if ($is_dst_type_derived_from_static_type$2$off0) {
           label = 25;
           break;
          } else {
           $56 = 4;
           break;
          }
         }
        }
       }
       if ($is_dst_type_derived_from_static_type$2$off0) {
        label = 25;
       } else {
        $56 = 4;
       }
      }
     } while(0);
     if ((label|0) == 25) {
      $56 = 3;
     }
     HEAP32[$20>>2] = $56;
     break;
    }
   }
   $17 = ($path_below|0)==(1);
   if ($17) {
    $18 = ((($info)) + 32|0);
    HEAP32[$18>>2] = 1;
   }
  }
 } while(0);
 return;
}
function __ZNK10__cxxabiv122__base_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib($this,$info,$current_ptr,$path_below,$use_strcmp) {
 $this = $this|0;
 $info = $info|0;
 $current_ptr = $current_ptr|0;
 $path_below = $path_below|0;
 $use_strcmp = $use_strcmp|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $offset_to_base$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 >> 8;
 $3 = $1 & 1;
 $4 = ($3|0)==(0);
 if ($4) {
  $offset_to_base$0 = $2;
 } else {
  $5 = HEAP32[$current_ptr>>2]|0;
  $6 = (($5) + ($2)|0);
  $7 = HEAP32[$6>>2]|0;
  $offset_to_base$0 = $7;
 }
 $8 = HEAP32[$this>>2]|0;
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 24|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($current_ptr) + ($offset_to_base$0)|0);
 $13 = $1 & 2;
 $14 = ($13|0)!=(0);
 $15 = $14 ? $path_below : 2;
 FUNCTION_TABLE_viiiii[$11 & 8191]($8,$info,$12,$15,$use_strcmp);
 return;
}
function __ZNK10__cxxabiv121__vmi_class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi($this,$info,$adjustedPtr,$path_below) {
 $this = $this|0;
 $info = $info|0;
 $adjustedPtr = $adjustedPtr|0;
 $path_below = $path_below|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($info)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($this|0)==($1|0);
 L1: do {
  if ($2) {
   __ZNK10__cxxabiv117__class_type_info24process_found_base_classEPNS_19__dynamic_cast_infoEPvi(0,$info,$adjustedPtr,$path_below);
  } else {
   $3 = ((($this)) + 16|0);
   $4 = ((($this)) + 12|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = (((($this)) + 16|0) + ($5<<3)|0);
   __ZNK10__cxxabiv122__base_class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi($3,$info,$adjustedPtr,$path_below);
   $7 = ($5|0)>(1);
   if ($7) {
    $8 = ((($this)) + 24|0);
    $9 = ((($info)) + 54|0);
    $p$0 = $8;
    while(1) {
     __ZNK10__cxxabiv122__base_class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi($p$0,$info,$adjustedPtr,$path_below);
     $10 = HEAP8[$9>>0]|0;
     $11 = ($10<<24>>24)==(0);
     if (!($11)) {
      break L1;
     }
     $12 = ((($p$0)) + 8|0);
     $13 = ($12>>>0)<($6>>>0);
     if ($13) {
      $p$0 = $12;
     } else {
      break;
     }
    }
   }
  }
 } while(0);
 return;
}
function __ZNK10__cxxabiv122__base_class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi($this,$info,$adjustedPtr,$path_below) {
 $this = $this|0;
 $info = $info|0;
 $adjustedPtr = $adjustedPtr|0;
 $path_below = $path_below|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $offset_to_base$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 >> 8;
 $3 = $1 & 1;
 $4 = ($3|0)==(0);
 if ($4) {
  $offset_to_base$0 = $2;
 } else {
  $5 = HEAP32[$adjustedPtr>>2]|0;
  $6 = (($5) + ($2)|0);
  $7 = HEAP32[$6>>2]|0;
  $offset_to_base$0 = $7;
 }
 $8 = HEAP32[$this>>2]|0;
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 28|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($adjustedPtr) + ($offset_to_base$0)|0);
 $13 = $1 & 2;
 $14 = ($13|0)!=(0);
 $15 = $14 ? $path_below : 2;
 FUNCTION_TABLE_viiii[$11 & 8191]($8,$info,$12,$15);
 return;
}
function __ZNSt8bad_castC2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (157952);
 return;
}
function ___cxa_bad_typeid() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (___cxa_allocate_exception(4)|0);
 __ZNSt10bad_typeidC2Ev($0);
 ___cxa_throw(($0|0),(9168|0),(3339|0));
 // unreachable;
}
function __ZNSt10bad_typeidC2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (157972);
 return;
}
function __Znwj($size) {
 $size = $size|0;
 var $$lcssa = 0, $$size = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($size|0)==(0);
 $$size = $0 ? 1 : $size;
 while(1) {
  $1 = (_malloc($$size)|0);
  $2 = ($1|0)==(0|0);
  if (!($2)) {
   $$lcssa = $1;
   label = 6;
   break;
  }
  $3 = (__ZSt15get_new_handlerv()|0);
  $4 = ($3|0)==(0|0);
  if ($4) {
   label = 5;
   break;
  }
  FUNCTION_TABLE_v[$3 & 8191]();
 }
 if ((label|0) == 5) {
  $5 = (___cxa_allocate_exception(4)|0);
  HEAP32[$5>>2] = (157832);
  ___cxa_throw(($5|0),(9056|0),(3325|0));
  // unreachable;
 }
 else if ((label|0) == 6) {
  return ($$lcssa|0);
 }
 return (0)|0;
}
function __ZSt15get_new_handlerv() {
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[302375]|0;HEAP32[302375] = (($0+0)|0);
 $1 = $0;
 return ($1|0);
}
function __Znaj($size) {
 $size = $size|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znwj($size)|0);
 return ($0|0);
}
function __ZnajRKSt9nothrow_t($size,$0) {
 $size = $size|0;
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __THREW__ = 0;
 $1 = (invoke_ii(3492,($size|0))|0);
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  (___cxa_begin_catch(($4|0))|0);
  __THREW__ = 0;
  invoke_v(3378);
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if ($7) {
   $8 = ___cxa_find_matching_catch_3(0|0)|0;
   $9 = tempRet0;
   ___clang_call_terminate($8);
   // unreachable;
  } else {
   $p$0 = 0;
  }
 } else {
  $p$0 = $1;
 }
 return ($p$0|0);
}
function __ZdaPv($ptr) {
 $ptr = $ptr|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($ptr);
 return;
}
function __ZNSt9bad_allocC2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (157832);
 return;
}
function ___cxa_can_catch($catchType,$excpType,$thrown) {
 $catchType = $catchType|0;
 $excpType = $excpType|0;
 $thrown = $thrown|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $temp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abort();
 $temp = sp;
 $0 = HEAP32[$thrown>>2]|0;
 HEAP32[$temp>>2] = $0;
 $1 = HEAP32[$catchType>>2]|0;
 $2 = ((($1)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_iiii[$3 & 8191]($catchType,$excpType,$temp)|0);
 $5 = $4&1;
 if ($4) {
  $6 = HEAP32[$temp>>2]|0;
  HEAP32[$thrown>>2] = $6;
 }
 STACKTOP = sp;return ($5|0);
}
function ___cxa_is_pointer_type($type) {
 $type = $type|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($type|0)==(0|0);
 if ($0) {
  $3 = 0;
 } else {
  $1 = (___dynamic_cast($type,9032,9200,0)|0);
  $phitmp = ($1|0)!=(0|0);
  $3 = $phitmp;
 }
 $2 = $3&1;
 return ($2|0);
}
function __ZSt17__throw_bad_allocv() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (___cxa_allocate_exception(4)|0);
 __ZNSt9bad_allocC2Ev($0);
 ___cxa_throw(($0|0),(9056|0),(3325|0));
 // unreachable;
}
function runPostSets() {
}
function _llvm_bswap_i32(x) {
    x = x|0;
    return (((x&0xff)<<24) | (((x>>8)&0xff)<<16) | (((x>>16)&0xff)<<8) | (x>>>24))|0;
}
function _memset(ptr, value, num) {
    ptr = ptr|0; value = value|0; num = num|0;
    var stop = 0, value4 = 0, stop4 = 0, unaligned = 0;
    stop = (ptr + num)|0;
    if ((num|0) >= 20) {
      // This is unaligned, but quite large, so work hard to get to aligned settings
      value = value & 0xff;
      unaligned = ptr & 3;
      value4 = value | (value << 8) | (value << 16) | (value << 24);
      stop4 = stop & ~3;
      if (unaligned) {
        unaligned = (ptr + 4 - unaligned)|0;
        while ((ptr|0) < (unaligned|0)) { // no need to check for stop, since we have large num
          HEAP8[((ptr)>>0)]=value;
          ptr = (ptr+1)|0;
        }
      }
      while ((ptr|0) < (stop4|0)) {
        HEAP32[((ptr)>>2)]=value4;
        ptr = (ptr+4)|0;
      }
    }
    while ((ptr|0) < (stop|0)) {
      HEAP8[((ptr)>>0)]=value;
      ptr = (ptr+1)|0;
    }
    return (ptr-num)|0;
}
function _llvm_cttz_i32(x) {
    x = x|0;
    var ret = 0;
    ret = ((HEAP8[(((cttz_i8)+(x & 0xff))>>0)])|0);
    if ((ret|0) < 8) return ret|0;
    ret = ((HEAP8[(((cttz_i8)+((x >> 8)&0xff))>>0)])|0);
    if ((ret|0) < 8) return (ret + 8)|0;
    ret = ((HEAP8[(((cttz_i8)+((x >> 16)&0xff))>>0)])|0);
    if ((ret|0) < 8) return (ret + 16)|0;
    return (((HEAP8[(((cttz_i8)+(x >>> 24))>>0)])|0) + 24)|0;
}
function _i64Add(a, b, c, d) {
    /*
      x = a + b*2^32
      y = c + d*2^32
      result = l + h*2^32
    */
    a = a|0; b = b|0; c = c|0; d = d|0;
    var l = 0, h = 0;
    l = (a + c)>>>0;
    h = (b + d + (((l>>>0) < (a>>>0))|0))>>>0; // Add carry from low word to high word on overflow.
    return ((tempRet0 = h,l|0)|0);
}
function _saveSetjmp(env, label, table, size) {
    // Not particularly fast: slow table lookup of setjmpId to label. But setjmp
    // prevents relooping anyhow, so slowness is to be expected. And typical case
    // is 1 setjmp per invocation, or less.
    env = env|0;
    label = label|0;
    table = table|0;
    size = size|0;
    var i = 0;
    setjmpId = (setjmpId+1)|0;
    HEAP32[((env)>>2)]=setjmpId;
    while ((i|0) < (size|0)) {
      if (((HEAP32[(((table)+((i<<3)))>>2)])|0) == 0) {
        HEAP32[(((table)+((i<<3)))>>2)]=setjmpId;
        HEAP32[(((table)+((i<<3)+4))>>2)]=label;
        // prepare next slot
        HEAP32[(((table)+((i<<3)+8))>>2)]=0;
        tempRet0 = size;
        return table | 0;
      }
      i = i+1|0;
    }
    // grow the table
    size = (size*2)|0;
    table = _realloc(table|0, 8*(size+1|0)|0) | 0;
    table = _saveSetjmp(env|0, label|0, table|0, size|0) | 0;
    tempRet0 = size;
    return table | 0;
}
function _testSetjmp(id, table, size) {
    id = id|0;
    table = table|0;
    size = size|0;
    var i = 0, curr = 0;
    while ((i|0) < (size|0)) {
      curr = ((HEAP32[(((table)+((i<<3)))>>2)])|0);
      if ((curr|0) == 0) break;
      if ((curr|0) == (id|0)) {
        return ((HEAP32[(((table)+((i<<3)+4))>>2)])|0);
      }
      i = i+1|0;
    }
    return 0;
}
function _memcpy(dest, src, num) {
    dest = dest|0; src = src|0; num = num|0;
    var ret = 0;
    if ((num|0) >= 4096) return _emscripten_memcpy_big(dest|0, src|0, num|0)|0;
    ret = dest|0;
    if ((dest&3) == (src&3)) {
      while (dest & 3) {
        if ((num|0) == 0) return ret|0;
        HEAP8[((dest)>>0)]=((HEAP8[((src)>>0)])|0);
        dest = (dest+1)|0;
        src = (src+1)|0;
        num = (num-1)|0;
      }
      while ((num|0) >= 4) {
        HEAP32[((dest)>>2)]=((HEAP32[((src)>>2)])|0);
        dest = (dest+4)|0;
        src = (src+4)|0;
        num = (num-4)|0;
      }
    }
    while ((num|0) > 0) {
      HEAP8[((dest)>>0)]=((HEAP8[((src)>>0)])|0);
      dest = (dest+1)|0;
      src = (src+1)|0;
      num = (num-1)|0;
    }
    return ret|0;
}
function _memmove(dest, src, num) {
    dest = dest|0; src = src|0; num = num|0;
    var ret = 0;
    if (((src|0) < (dest|0)) & ((dest|0) < ((src + num)|0))) {
      // Unlikely case: Copy backwards in a safe manner
      ret = dest;
      src = (src + num)|0;
      dest = (dest + num)|0;
      while ((num|0) > 0) {
        dest = (dest - 1)|0;
        src = (src - 1)|0;
        num = (num - 1)|0;
        HEAP8[((dest)>>0)]=((HEAP8[((src)>>0)])|0);
      }
      dest = ret;
    } else {
      _memcpy(dest, src, num) | 0;
    }
    return dest | 0;
}
function _llvm_ctlz_i64(l, h, isZeroUndef) {
    l = l | 0;
    h = h | 0;
    isZeroUndef = isZeroUndef | 0;
    var ret = 0;
    ret = Math_clz32(h) | 0;
    if ((ret | 0) == 32) ret = ret + (Math_clz32(l) | 0) | 0;
    tempRet0 = 0;
    return ret | 0;
}
function _bitshift64Shl(low, high, bits) {
    low = low|0; high = high|0; bits = bits|0;
    var ander = 0;
    if ((bits|0) < 32) {
      ander = ((1 << bits) - 1)|0;
      tempRet0 = (high << bits) | ((low&(ander << (32 - bits))) >>> (32 - bits));
      return low << bits;
    }
    tempRet0 = low << (bits - 32);
    return 0;
}
function _i64Subtract(a, b, c, d) {
    a = a|0; b = b|0; c = c|0; d = d|0;
    var l = 0, h = 0;
    l = (a - c)>>>0;
    h = (b - d)>>>0;
    h = (b - d - (((c>>>0) > (a>>>0))|0))>>>0; // Borrow one from high word to low word on underflow.
    return ((tempRet0 = h,l|0)|0);
}
function _bitshift64Lshr(low, high, bits) {
    low = low|0; high = high|0; bits = bits|0;
    var ander = 0;
    if ((bits|0) < 32) {
      ander = ((1 << bits) - 1)|0;
      tempRet0 = high >>> bits;
      return (low >>> bits) | ((high&ander) << (32 - bits));
    }
    tempRet0 = 0;
    return (high >>> (bits - 32))|0;
}
function _bitshift64Ashr(low, high, bits) {
    low = low|0; high = high|0; bits = bits|0;
    var ander = 0;
    if ((bits|0) < 32) {
      ander = ((1 << bits) - 1)|0;
      tempRet0 = high >> bits;
      return (low >>> bits) | ((high&ander) << (32 - bits));
    }
    tempRet0 = (high|0) < 0 ? -1 : 0;
    return (high >> (bits - 32))|0;
}
// ======== compiled code from system/lib/compiler-rt , see readme therein
function ___muldsi3($a, $b) {
  $a = $a | 0;
  $b = $b | 0;
  var $1 = 0, $2 = 0, $3 = 0, $6 = 0, $8 = 0, $11 = 0, $12 = 0;
  $1 = $a & 65535;
  $2 = $b & 65535;
  $3 = Math_imul($2, $1) | 0;
  $6 = $a >>> 16;
  $8 = ($3 >>> 16) + (Math_imul($2, $6) | 0) | 0;
  $11 = $b >>> 16;
  $12 = Math_imul($11, $1) | 0;
  return (tempRet0 = (($8 >>> 16) + (Math_imul($11, $6) | 0) | 0) + ((($8 & 65535) + $12 | 0) >>> 16) | 0, 0 | ($8 + $12 << 16 | $3 & 65535)) | 0;
}
function ___divdi3($a$0, $a$1, $b$0, $b$1) {
  $a$0 = $a$0 | 0;
  $a$1 = $a$1 | 0;
  $b$0 = $b$0 | 0;
  $b$1 = $b$1 | 0;
  var $1$0 = 0, $1$1 = 0, $2$0 = 0, $2$1 = 0, $4$0 = 0, $4$1 = 0, $6$0 = 0, $7$0 = 0, $7$1 = 0, $8$0 = 0, $10$0 = 0;
  $1$0 = $a$1 >> 31 | (($a$1 | 0) < 0 ? -1 : 0) << 1;
  $1$1 = (($a$1 | 0) < 0 ? -1 : 0) >> 31 | (($a$1 | 0) < 0 ? -1 : 0) << 1;
  $2$0 = $b$1 >> 31 | (($b$1 | 0) < 0 ? -1 : 0) << 1;
  $2$1 = (($b$1 | 0) < 0 ? -1 : 0) >> 31 | (($b$1 | 0) < 0 ? -1 : 0) << 1;
  $4$0 = _i64Subtract($1$0 ^ $a$0 | 0, $1$1 ^ $a$1 | 0, $1$0 | 0, $1$1 | 0) | 0;
  $4$1 = tempRet0;
  $6$0 = _i64Subtract($2$0 ^ $b$0 | 0, $2$1 ^ $b$1 | 0, $2$0 | 0, $2$1 | 0) | 0;
  $7$0 = $2$0 ^ $1$0;
  $7$1 = $2$1 ^ $1$1;
  $8$0 = ___udivmoddi4($4$0, $4$1, $6$0, tempRet0, 0) | 0;
  $10$0 = _i64Subtract($8$0 ^ $7$0 | 0, tempRet0 ^ $7$1 | 0, $7$0 | 0, $7$1 | 0) | 0;
  return $10$0 | 0;
}
function ___remdi3($a$0, $a$1, $b$0, $b$1) {
  $a$0 = $a$0 | 0;
  $a$1 = $a$1 | 0;
  $b$0 = $b$0 | 0;
  $b$1 = $b$1 | 0;
  var $rem = 0, $1$0 = 0, $1$1 = 0, $2$0 = 0, $2$1 = 0, $4$0 = 0, $4$1 = 0, $6$0 = 0, $10$0 = 0, $10$1 = 0, __stackBase__ = 0;
  __stackBase__ = STACKTOP;
  STACKTOP = STACKTOP + 16 | 0;
  $rem = __stackBase__ | 0;
  $1$0 = $a$1 >> 31 | (($a$1 | 0) < 0 ? -1 : 0) << 1;
  $1$1 = (($a$1 | 0) < 0 ? -1 : 0) >> 31 | (($a$1 | 0) < 0 ? -1 : 0) << 1;
  $2$0 = $b$1 >> 31 | (($b$1 | 0) < 0 ? -1 : 0) << 1;
  $2$1 = (($b$1 | 0) < 0 ? -1 : 0) >> 31 | (($b$1 | 0) < 0 ? -1 : 0) << 1;
  $4$0 = _i64Subtract($1$0 ^ $a$0 | 0, $1$1 ^ $a$1 | 0, $1$0 | 0, $1$1 | 0) | 0;
  $4$1 = tempRet0;
  $6$0 = _i64Subtract($2$0 ^ $b$0 | 0, $2$1 ^ $b$1 | 0, $2$0 | 0, $2$1 | 0) | 0;
  ___udivmoddi4($4$0, $4$1, $6$0, tempRet0, $rem) | 0;
  $10$0 = _i64Subtract(HEAP32[$rem >> 2] ^ $1$0 | 0, HEAP32[$rem + 4 >> 2] ^ $1$1 | 0, $1$0 | 0, $1$1 | 0) | 0;
  $10$1 = tempRet0;
  STACKTOP = __stackBase__;
  return (tempRet0 = $10$1, $10$0) | 0;
}
function ___muldi3($a$0, $a$1, $b$0, $b$1) {
  $a$0 = $a$0 | 0;
  $a$1 = $a$1 | 0;
  $b$0 = $b$0 | 0;
  $b$1 = $b$1 | 0;
  var $x_sroa_0_0_extract_trunc = 0, $y_sroa_0_0_extract_trunc = 0, $1$0 = 0, $1$1 = 0, $2 = 0;
  $x_sroa_0_0_extract_trunc = $a$0;
  $y_sroa_0_0_extract_trunc = $b$0;
  $1$0 = ___muldsi3($x_sroa_0_0_extract_trunc, $y_sroa_0_0_extract_trunc) | 0;
  $1$1 = tempRet0;
  $2 = Math_imul($a$1, $y_sroa_0_0_extract_trunc) | 0;
  return (tempRet0 = ((Math_imul($b$1, $x_sroa_0_0_extract_trunc) | 0) + $2 | 0) + $1$1 | $1$1 & 0, 0 | $1$0 & -1) | 0;
}
function ___udivdi3($a$0, $a$1, $b$0, $b$1) {
  $a$0 = $a$0 | 0;
  $a$1 = $a$1 | 0;
  $b$0 = $b$0 | 0;
  $b$1 = $b$1 | 0;
  var $1$0 = 0;
  $1$0 = ___udivmoddi4($a$0, $a$1, $b$0, $b$1, 0) | 0;
  return $1$0 | 0;
}
function ___uremdi3($a$0, $a$1, $b$0, $b$1) {
  $a$0 = $a$0 | 0;
  $a$1 = $a$1 | 0;
  $b$0 = $b$0 | 0;
  $b$1 = $b$1 | 0;
  var $rem = 0, __stackBase__ = 0;
  __stackBase__ = STACKTOP;
  STACKTOP = STACKTOP + 16 | 0;
  $rem = __stackBase__ | 0;
  ___udivmoddi4($a$0, $a$1, $b$0, $b$1, $rem) | 0;
  STACKTOP = __stackBase__;
  return (tempRet0 = HEAP32[$rem + 4 >> 2] | 0, HEAP32[$rem >> 2] | 0) | 0;
}
function ___udivmoddi4($a$0, $a$1, $b$0, $b$1, $rem) {
  $a$0 = $a$0 | 0;
  $a$1 = $a$1 | 0;
  $b$0 = $b$0 | 0;
  $b$1 = $b$1 | 0;
  $rem = $rem | 0;
  var $n_sroa_0_0_extract_trunc = 0, $n_sroa_1_4_extract_shift$0 = 0, $n_sroa_1_4_extract_trunc = 0, $d_sroa_0_0_extract_trunc = 0, $d_sroa_1_4_extract_shift$0 = 0, $d_sroa_1_4_extract_trunc = 0, $4 = 0, $17 = 0, $37 = 0, $49 = 0, $51 = 0, $57 = 0, $58 = 0, $66 = 0, $78 = 0, $86 = 0, $88 = 0, $89 = 0, $91 = 0, $92 = 0, $95 = 0, $105 = 0, $117 = 0, $119 = 0, $125 = 0, $126 = 0, $130 = 0, $q_sroa_1_1_ph = 0, $q_sroa_0_1_ph = 0, $r_sroa_1_1_ph = 0, $r_sroa_0_1_ph = 0, $sr_1_ph = 0, $d_sroa_0_0_insert_insert99$0 = 0, $d_sroa_0_0_insert_insert99$1 = 0, $137$0 = 0, $137$1 = 0, $carry_0203 = 0, $sr_1202 = 0, $r_sroa_0_1201 = 0, $r_sroa_1_1200 = 0, $q_sroa_0_1199 = 0, $q_sroa_1_1198 = 0, $147 = 0, $149 = 0, $r_sroa_0_0_insert_insert42$0 = 0, $r_sroa_0_0_insert_insert42$1 = 0, $150$1 = 0, $151$0 = 0, $152 = 0, $154$0 = 0, $r_sroa_0_0_extract_trunc = 0, $r_sroa_1_4_extract_trunc = 0, $155 = 0, $carry_0_lcssa$0 = 0, $carry_0_lcssa$1 = 0, $r_sroa_0_1_lcssa = 0, $r_sroa_1_1_lcssa = 0, $q_sroa_0_1_lcssa = 0, $q_sroa_1_1_lcssa = 0, $q_sroa_0_0_insert_ext75$0 = 0, $q_sroa_0_0_insert_ext75$1 = 0, $q_sroa_0_0_insert_insert77$1 = 0, $_0$0 = 0, $_0$1 = 0;
  $n_sroa_0_0_extract_trunc = $a$0;
  $n_sroa_1_4_extract_shift$0 = $a$1;
  $n_sroa_1_4_extract_trunc = $n_sroa_1_4_extract_shift$0;
  $d_sroa_0_0_extract_trunc = $b$0;
  $d_sroa_1_4_extract_shift$0 = $b$1;
  $d_sroa_1_4_extract_trunc = $d_sroa_1_4_extract_shift$0;
  if (($n_sroa_1_4_extract_trunc | 0) == 0) {
    $4 = ($rem | 0) != 0;
    if (($d_sroa_1_4_extract_trunc | 0) == 0) {
      if ($4) {
        HEAP32[$rem >> 2] = ($n_sroa_0_0_extract_trunc >>> 0) % ($d_sroa_0_0_extract_trunc >>> 0);
        HEAP32[$rem + 4 >> 2] = 0;
      }
      $_0$1 = 0;
      $_0$0 = ($n_sroa_0_0_extract_trunc >>> 0) / ($d_sroa_0_0_extract_trunc >>> 0) >>> 0;
      return (tempRet0 = $_0$1, $_0$0) | 0;
    } else {
      if (!$4) {
        $_0$1 = 0;
        $_0$0 = 0;
        return (tempRet0 = $_0$1, $_0$0) | 0;
      }
      HEAP32[$rem >> 2] = $a$0 & -1;
      HEAP32[$rem + 4 >> 2] = $a$1 & 0;
      $_0$1 = 0;
      $_0$0 = 0;
      return (tempRet0 = $_0$1, $_0$0) | 0;
    }
  }
  $17 = ($d_sroa_1_4_extract_trunc | 0) == 0;
  do {
    if (($d_sroa_0_0_extract_trunc | 0) == 0) {
      if ($17) {
        if (($rem | 0) != 0) {
          HEAP32[$rem >> 2] = ($n_sroa_1_4_extract_trunc >>> 0) % ($d_sroa_0_0_extract_trunc >>> 0);
          HEAP32[$rem + 4 >> 2] = 0;
        }
        $_0$1 = 0;
        $_0$0 = ($n_sroa_1_4_extract_trunc >>> 0) / ($d_sroa_0_0_extract_trunc >>> 0) >>> 0;
        return (tempRet0 = $_0$1, $_0$0) | 0;
      }
      if (($n_sroa_0_0_extract_trunc | 0) == 0) {
        if (($rem | 0) != 0) {
          HEAP32[$rem >> 2] = 0;
          HEAP32[$rem + 4 >> 2] = ($n_sroa_1_4_extract_trunc >>> 0) % ($d_sroa_1_4_extract_trunc >>> 0);
        }
        $_0$1 = 0;
        $_0$0 = ($n_sroa_1_4_extract_trunc >>> 0) / ($d_sroa_1_4_extract_trunc >>> 0) >>> 0;
        return (tempRet0 = $_0$1, $_0$0) | 0;
      }
      $37 = $d_sroa_1_4_extract_trunc - 1 | 0;
      if (($37 & $d_sroa_1_4_extract_trunc | 0) == 0) {
        if (($rem | 0) != 0) {
          HEAP32[$rem >> 2] = 0 | $a$0 & -1;
          HEAP32[$rem + 4 >> 2] = $37 & $n_sroa_1_4_extract_trunc | $a$1 & 0;
        }
        $_0$1 = 0;
        $_0$0 = $n_sroa_1_4_extract_trunc >>> ((_llvm_cttz_i32($d_sroa_1_4_extract_trunc | 0) | 0) >>> 0);
        return (tempRet0 = $_0$1, $_0$0) | 0;
      }
      $49 = Math_clz32($d_sroa_1_4_extract_trunc | 0) | 0;
      $51 = $49 - (Math_clz32($n_sroa_1_4_extract_trunc | 0) | 0) | 0;
      if ($51 >>> 0 <= 30) {
        $57 = $51 + 1 | 0;
        $58 = 31 - $51 | 0;
        $sr_1_ph = $57;
        $r_sroa_0_1_ph = $n_sroa_1_4_extract_trunc << $58 | $n_sroa_0_0_extract_trunc >>> ($57 >>> 0);
        $r_sroa_1_1_ph = $n_sroa_1_4_extract_trunc >>> ($57 >>> 0);
        $q_sroa_0_1_ph = 0;
        $q_sroa_1_1_ph = $n_sroa_0_0_extract_trunc << $58;
        break;
      }
      if (($rem | 0) == 0) {
        $_0$1 = 0;
        $_0$0 = 0;
        return (tempRet0 = $_0$1, $_0$0) | 0;
      }
      HEAP32[$rem >> 2] = 0 | $a$0 & -1;
      HEAP32[$rem + 4 >> 2] = $n_sroa_1_4_extract_shift$0 | $a$1 & 0;
      $_0$1 = 0;
      $_0$0 = 0;
      return (tempRet0 = $_0$1, $_0$0) | 0;
    } else {
      if (!$17) {
        $117 = Math_clz32($d_sroa_1_4_extract_trunc | 0) | 0;
        $119 = $117 - (Math_clz32($n_sroa_1_4_extract_trunc | 0) | 0) | 0;
        if ($119 >>> 0 <= 31) {
          $125 = $119 + 1 | 0;
          $126 = 31 - $119 | 0;
          $130 = $119 - 31 >> 31;
          $sr_1_ph = $125;
          $r_sroa_0_1_ph = $n_sroa_0_0_extract_trunc >>> ($125 >>> 0) & $130 | $n_sroa_1_4_extract_trunc << $126;
          $r_sroa_1_1_ph = $n_sroa_1_4_extract_trunc >>> ($125 >>> 0) & $130;
          $q_sroa_0_1_ph = 0;
          $q_sroa_1_1_ph = $n_sroa_0_0_extract_trunc << $126;
          break;
        }
        if (($rem | 0) == 0) {
          $_0$1 = 0;
          $_0$0 = 0;
          return (tempRet0 = $_0$1, $_0$0) | 0;
        }
        HEAP32[$rem >> 2] = 0 | $a$0 & -1;
        HEAP32[$rem + 4 >> 2] = $n_sroa_1_4_extract_shift$0 | $a$1 & 0;
        $_0$1 = 0;
        $_0$0 = 0;
        return (tempRet0 = $_0$1, $_0$0) | 0;
      }
      $66 = $d_sroa_0_0_extract_trunc - 1 | 0;
      if (($66 & $d_sroa_0_0_extract_trunc | 0) != 0) {
        $86 = (Math_clz32($d_sroa_0_0_extract_trunc | 0) | 0) + 33 | 0;
        $88 = $86 - (Math_clz32($n_sroa_1_4_extract_trunc | 0) | 0) | 0;
        $89 = 64 - $88 | 0;
        $91 = 32 - $88 | 0;
        $92 = $91 >> 31;
        $95 = $88 - 32 | 0;
        $105 = $95 >> 31;
        $sr_1_ph = $88;
        $r_sroa_0_1_ph = $91 - 1 >> 31 & $n_sroa_1_4_extract_trunc >>> ($95 >>> 0) | ($n_sroa_1_4_extract_trunc << $91 | $n_sroa_0_0_extract_trunc >>> ($88 >>> 0)) & $105;
        $r_sroa_1_1_ph = $105 & $n_sroa_1_4_extract_trunc >>> ($88 >>> 0);
        $q_sroa_0_1_ph = $n_sroa_0_0_extract_trunc << $89 & $92;
        $q_sroa_1_1_ph = ($n_sroa_1_4_extract_trunc << $89 | $n_sroa_0_0_extract_trunc >>> ($95 >>> 0)) & $92 | $n_sroa_0_0_extract_trunc << $91 & $88 - 33 >> 31;
        break;
      }
      if (($rem | 0) != 0) {
        HEAP32[$rem >> 2] = $66 & $n_sroa_0_0_extract_trunc;
        HEAP32[$rem + 4 >> 2] = 0;
      }
      if (($d_sroa_0_0_extract_trunc | 0) == 1) {
        $_0$1 = $n_sroa_1_4_extract_shift$0 | $a$1 & 0;
        $_0$0 = 0 | $a$0 & -1;
        return (tempRet0 = $_0$1, $_0$0) | 0;
      } else {
        $78 = _llvm_cttz_i32($d_sroa_0_0_extract_trunc | 0) | 0;
        $_0$1 = 0 | $n_sroa_1_4_extract_trunc >>> ($78 >>> 0);
        $_0$0 = $n_sroa_1_4_extract_trunc << 32 - $78 | $n_sroa_0_0_extract_trunc >>> ($78 >>> 0) | 0;
        return (tempRet0 = $_0$1, $_0$0) | 0;
      }
    }
  } while (0);
  if (($sr_1_ph | 0) == 0) {
    $q_sroa_1_1_lcssa = $q_sroa_1_1_ph;
    $q_sroa_0_1_lcssa = $q_sroa_0_1_ph;
    $r_sroa_1_1_lcssa = $r_sroa_1_1_ph;
    $r_sroa_0_1_lcssa = $r_sroa_0_1_ph;
    $carry_0_lcssa$1 = 0;
    $carry_0_lcssa$0 = 0;
  } else {
    $d_sroa_0_0_insert_insert99$0 = 0 | $b$0 & -1;
    $d_sroa_0_0_insert_insert99$1 = $d_sroa_1_4_extract_shift$0 | $b$1 & 0;
    $137$0 = _i64Add($d_sroa_0_0_insert_insert99$0 | 0, $d_sroa_0_0_insert_insert99$1 | 0, -1, -1) | 0;
    $137$1 = tempRet0;
    $q_sroa_1_1198 = $q_sroa_1_1_ph;
    $q_sroa_0_1199 = $q_sroa_0_1_ph;
    $r_sroa_1_1200 = $r_sroa_1_1_ph;
    $r_sroa_0_1201 = $r_sroa_0_1_ph;
    $sr_1202 = $sr_1_ph;
    $carry_0203 = 0;
    while (1) {
      $147 = $q_sroa_0_1199 >>> 31 | $q_sroa_1_1198 << 1;
      $149 = $carry_0203 | $q_sroa_0_1199 << 1;
      $r_sroa_0_0_insert_insert42$0 = 0 | ($r_sroa_0_1201 << 1 | $q_sroa_1_1198 >>> 31);
      $r_sroa_0_0_insert_insert42$1 = $r_sroa_0_1201 >>> 31 | $r_sroa_1_1200 << 1 | 0;
      _i64Subtract($137$0 | 0, $137$1 | 0, $r_sroa_0_0_insert_insert42$0 | 0, $r_sroa_0_0_insert_insert42$1 | 0) | 0;
      $150$1 = tempRet0;
      $151$0 = $150$1 >> 31 | (($150$1 | 0) < 0 ? -1 : 0) << 1;
      $152 = $151$0 & 1;
      $154$0 = _i64Subtract($r_sroa_0_0_insert_insert42$0 | 0, $r_sroa_0_0_insert_insert42$1 | 0, $151$0 & $d_sroa_0_0_insert_insert99$0 | 0, ((($150$1 | 0) < 0 ? -1 : 0) >> 31 | (($150$1 | 0) < 0 ? -1 : 0) << 1) & $d_sroa_0_0_insert_insert99$1 | 0) | 0;
      $r_sroa_0_0_extract_trunc = $154$0;
      $r_sroa_1_4_extract_trunc = tempRet0;
      $155 = $sr_1202 - 1 | 0;
      if (($155 | 0) == 0) {
        break;
      } else {
        $q_sroa_1_1198 = $147;
        $q_sroa_0_1199 = $149;
        $r_sroa_1_1200 = $r_sroa_1_4_extract_trunc;
        $r_sroa_0_1201 = $r_sroa_0_0_extract_trunc;
        $sr_1202 = $155;
        $carry_0203 = $152;
      }
    }
    $q_sroa_1_1_lcssa = $147;
    $q_sroa_0_1_lcssa = $149;
    $r_sroa_1_1_lcssa = $r_sroa_1_4_extract_trunc;
    $r_sroa_0_1_lcssa = $r_sroa_0_0_extract_trunc;
    $carry_0_lcssa$1 = 0;
    $carry_0_lcssa$0 = $152;
  }
  $q_sroa_0_0_insert_ext75$0 = $q_sroa_0_1_lcssa;
  $q_sroa_0_0_insert_ext75$1 = 0;
  $q_sroa_0_0_insert_insert77$1 = $q_sroa_1_1_lcssa | $q_sroa_0_0_insert_ext75$1;
  if (($rem | 0) != 0) {
    HEAP32[$rem >> 2] = 0 | $r_sroa_0_1_lcssa;
    HEAP32[$rem + 4 >> 2] = $r_sroa_1_1_lcssa | 0;
  }
  $_0$1 = (0 | $q_sroa_0_0_insert_ext75$0) >>> 31 | $q_sroa_0_0_insert_insert77$1 << 1 | ($q_sroa_0_0_insert_ext75$1 << 1 | $q_sroa_0_0_insert_ext75$0 >>> 31) & 0 | $carry_0_lcssa$1;
  $_0$0 = ($q_sroa_0_0_insert_ext75$0 << 1 | 0 >>> 31) & -2 | $carry_0_lcssa$0;
  return (tempRet0 = $_0$1, $_0$0) | 0;
}
// =======================================================================



  
function dynCall_iiiiiid(index,a1,a2,a3,a4,a5,a6) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return FUNCTION_TABLE_iiiiiid[index&4095](a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}


function dynCall_viiidiii(index,a1,a2,a3,a4,a5,a6,a7) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=+a4; a5=a5|0; a6=a6|0; a7=a7|0;
  FUNCTION_TABLE_viiidiii[index&4095](a1|0,a2|0,a3|0,+a4,a5|0,a6|0,a7|0);
}


function dynCall_viiiddddddi(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=+a4; a5=+a5; a6=+a6; a7=+a7; a8=+a8; a9=+a9; a10=a10|0;
  FUNCTION_TABLE_viiiddddddi[index&4095](a1|0,a2|0,a3|0,+a4,+a5,+a6,+a7,+a8,+a9,a10|0);
}


function dynCall_iiiiiii(index,a1,a2,a3,a4,a5,a6) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return FUNCTION_TABLE_iiiiiii[index&8191](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}


function dynCall_viiiiddi(index,a1,a2,a3,a4,a5,a6,a7) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5; a6=+a6; a7=a7|0;
  FUNCTION_TABLE_viiiiddi[index&4095](a1|0,a2|0,a3|0,a4|0,+a5,+a6,a7|0);
}


function dynCall_iidddddddddi(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  index = index|0;
  a1=a1|0; a2=+a2; a3=+a3; a4=+a4; a5=+a5; a6=+a6; a7=+a7; a8=+a8; a9=+a9; a10=+a10; a11=a11|0;
  return FUNCTION_TABLE_iidddddddddi[index&4095](a1|0,+a2,+a3,+a4,+a5,+a6,+a7,+a8,+a9,+a10,a11|0)|0;
}


function dynCall_viiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  FUNCTION_TABLE_viiiiiiiiiii[index&8191](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0);
}


function dynCall_iidd(index,a1,a2,a3) {
  index = index|0;
  a1=a1|0; a2=+a2; a3=+a3;
  return FUNCTION_TABLE_iidd[index&4095](a1|0,+a2,+a3)|0;
}


function dynCall_vidi(index,a1,a2,a3) {
  index = index|0;
  a1=a1|0; a2=+a2; a3=a3|0;
  FUNCTION_TABLE_vidi[index&2047](a1|0,+a2,a3|0);
}


function dynCall_vidd(index,a1,a2,a3) {
  index = index|0;
  a1=a1|0; a2=+a2; a3=+a3;
  FUNCTION_TABLE_vidd[index&8191](a1|0,+a2,+a3);
}


function dynCall_iiiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return FUNCTION_TABLE_iiiiiiiiiiii[index&8191](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}


function dynCall_viidddiddddddd(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=+a3; a4=+a4; a5=+a5; a6=a6|0; a7=+a7; a8=+a8; a9=+a9; a10=+a10; a11=+a11; a12=+a12; a13=+a13;
  FUNCTION_TABLE_viidddiddddddd[index&4095](a1|0,a2|0,+a3,+a4,+a5,a6|0,+a7,+a8,+a9,+a10,+a11,+a12,+a13);
}


function dynCall_viiiiiiidi(index,a1,a2,a3,a4,a5,a6,a7,a8,a9) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=+a8; a9=a9|0;
  FUNCTION_TABLE_viiiiiiidi[index&4095](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,+a8,a9|0);
}


function dynCall_di(index,a1) {
  index = index|0;
  a1=a1|0;
  return +FUNCTION_TABLE_di[index&4095](a1|0);
}


function dynCall_iiiiddi(index,a1,a2,a3,a4,a5,a6) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=+a4; a5=+a5; a6=a6|0;
  return FUNCTION_TABLE_iiiiddi[index&4095](a1|0,a2|0,a3|0,+a4,+a5,a6|0)|0;
}


function dynCall_iiddddd(index,a1,a2,a3,a4,a5,a6) {
  index = index|0;
  a1=a1|0; a2=+a2; a3=+a3; a4=+a4; a5=+a5; a6=+a6;
  return FUNCTION_TABLE_iiddddd[index&4095](a1|0,+a2,+a3,+a4,+a5,+a6)|0;
}


function dynCall_d(index) {
  index = index|0;
  
  return +FUNCTION_TABLE_d[index&4095]();
}


function dynCall_dii(index,a1,a2) {
  index = index|0;
  a1=a1|0; a2=a2|0;
  return +FUNCTION_TABLE_dii[index&8191](a1|0,a2|0);
}


function dynCall_viiidddddi(index,a1,a2,a3,a4,a5,a6,a7,a8,a9) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=+a4; a5=+a5; a6=+a6; a7=+a7; a8=+a8; a9=a9|0;
  FUNCTION_TABLE_viiidddddi[index&4095](a1|0,a2|0,a3|0,+a4,+a5,+a6,+a7,+a8,a9|0);
}


function dynCall_did(index,a1,a2) {
  index = index|0;
  a1=a1|0; a2=+a2;
  return +FUNCTION_TABLE_did[index&2047](a1|0,+a2);
}


function dynCall_iiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0;
  return FUNCTION_TABLE_iiiiiiiiii[index&8191](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0)|0;
}


function dynCall_vdddddd(index,a1,a2,a3,a4,a5,a6) {
  index = index|0;
  a1=+a1; a2=+a2; a3=+a3; a4=+a4; a5=+a5; a6=+a6;
  FUNCTION_TABLE_vdddddd[index&8191](+a1,+a2,+a3,+a4,+a5,+a6);
}


function dynCall_viiiddddi(index,a1,a2,a3,a4,a5,a6,a7,a8) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=+a4; a5=+a5; a6=+a6; a7=+a7; a8=a8|0;
  FUNCTION_TABLE_viiiddddi[index&4095](a1|0,a2|0,a3|0,+a4,+a5,+a6,+a7,a8|0);
}


function dynCall_vdd(index,a1,a2) {
  index = index|0;
  a1=+a1; a2=+a2;
  FUNCTION_TABLE_vdd[index&8191](+a1,+a2);
}


function dynCall_vdi(index,a1,a2) {
  index = index|0;
  a1=+a1; a2=a2|0;
  FUNCTION_TABLE_vdi[index&8191](+a1,a2|0);
}


function dynCall_iiiid(index,a1,a2,a3,a4) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=+a4;
  return FUNCTION_TABLE_iiiid[index&4095](a1|0,a2|0,a3|0,+a4)|0;
}


function dynCall_iiiiiiii(index,a1,a2,a3,a4,a5,a6,a7) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return FUNCTION_TABLE_iiiiiiii[index&8191](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}


function dynCall_iiiddi(index,a1,a2,a3,a4,a5) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=+a3; a4=+a4; a5=a5|0;
  return FUNCTION_TABLE_iiiddi[index&4095](a1|0,a2|0,+a3,+a4,a5|0)|0;
}


function dynCall_iiiiddddi(index,a1,a2,a3,a4,a5,a6,a7,a8) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=+a4; a5=+a5; a6=+a6; a7=+a7; a8=a8|0;
  return FUNCTION_TABLE_iiiiddddi[index&4095](a1|0,a2|0,a3|0,+a4,+a5,+a6,+a7,a8|0)|0;
}


function dynCall_iidddd(index,a1,a2,a3,a4,a5) {
  index = index|0;
  a1=a1|0; a2=+a2; a3=+a3; a4=+a4; a5=+a5;
  return FUNCTION_TABLE_iidddd[index&4095](a1|0,+a2,+a3,+a4,+a5)|0;
}


function dynCall_iiiiddddddi(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=+a4; a5=+a5; a6=+a6; a7=+a7; a8=+a8; a9=+a9; a10=a10|0;
  return FUNCTION_TABLE_iiiiddddddi[index&4095](a1|0,a2|0,a3|0,+a4,+a5,+a6,+a7,+a8,+a9,a10|0)|0;
}


function dynCall_iiiiiiiidi(index,a1,a2,a3,a4,a5,a6,a7,a8,a9) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=+a8; a9=a9|0;
  return FUNCTION_TABLE_iiiiiiiidi[index&4095](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,+a8,a9|0)|0;
}


function dynCall_viidddd(index,a1,a2,a3,a4,a5,a6) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=+a3; a4=+a4; a5=+a5; a6=+a6;
  FUNCTION_TABLE_viidddd[index&2047](a1|0,a2|0,+a3,+a4,+a5,+a6);
}


function dynCall_iiiii(index,a1,a2,a3,a4) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return FUNCTION_TABLE_iiiii[index&8191](a1|0,a2|0,a3|0,a4|0)|0;
}


function dynCall_viiiiii(index,a1,a2,a3,a4,a5,a6) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  FUNCTION_TABLE_viiiiii[index&8191](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}


function dynCall_ddd(index,a1,a2) {
  index = index|0;
  a1=+a1; a2=+a2;
  return +FUNCTION_TABLE_ddd[index&127](+a1,+a2);
}


function dynCall_viiidi(index,a1,a2,a3,a4,a5) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=+a4; a5=a5|0;
  FUNCTION_TABLE_viiidi[index&4095](a1|0,a2|0,a3|0,+a4,a5|0);
}


function dynCall_vidddd(index,a1,a2,a3,a4,a5) {
  index = index|0;
  a1=a1|0; a2=+a2; a3=+a3; a4=+a4; a5=+a5;
  FUNCTION_TABLE_vidddd[index&8191](a1|0,+a2,+a3,+a4,+a5);
}


function dynCall_diid(index,a1,a2,a3) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=+a3;
  return +FUNCTION_TABLE_diid[index&4095](a1|0,a2|0,+a3);
}


function dynCall_iiddd(index,a1,a2,a3,a4) {
  index = index|0;
  a1=a1|0; a2=+a2; a3=+a3; a4=+a4;
  return FUNCTION_TABLE_iiddd[index&4095](a1|0,+a2,+a3,+a4)|0;
}


function dynCall_iiddi(index,a1,a2,a3,a4) {
  index = index|0;
  a1=a1|0; a2=+a2; a3=+a3; a4=a4|0;
  return FUNCTION_TABLE_iiddi[index&4095](a1|0,+a2,+a3,a4|0)|0;
}


function dynCall_viiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  FUNCTION_TABLE_viiiiiiiiii[index&8191](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}


function dynCall_diii(index,a1,a2,a3) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +FUNCTION_TABLE_diii[index&8191](a1|0,a2|0,a3|0);
}


function dynCall_iiididdddddddd(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=+a3; a4=a4|0; a5=+a5; a6=+a6; a7=+a7; a8=+a8; a9=+a9; a10=+a10; a11=+a11; a12=+a12; a13=+a13;
  return FUNCTION_TABLE_iiididdddddddd[index&4095](a1|0,a2|0,+a3,a4|0,+a5,+a6,+a7,+a8,+a9,+a10,+a11,+a12,+a13)|0;
}


function dynCall_viiiddi(index,a1,a2,a3,a4,a5,a6) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=+a4; a5=+a5; a6=a6|0;
  FUNCTION_TABLE_viiiddi[index&4095](a1|0,a2|0,a3|0,+a4,+a5,a6|0);
}


function dynCall_iiiiii(index,a1,a2,a3,a4,a5) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return FUNCTION_TABLE_iiiiii[index&8191](a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}


function dynCall_viiid(index,a1,a2,a3,a4) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=+a4;
  FUNCTION_TABLE_viiid[index&4095](a1|0,a2|0,a3|0,+a4);
}


function dynCall_viidii(index,a1,a2,a3,a4,a5) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=+a3; a4=a4|0; a5=a5|0;
  FUNCTION_TABLE_viidii[index&8191](a1|0,a2|0,+a3,a4|0,a5|0);
}


function dynCall_iiiiid(index,a1,a2,a3,a4,a5) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return FUNCTION_TABLE_iiiiid[index&4095](a1|0,a2|0,a3|0,a4|0,+a5)|0;
}


function dynCall_viiii(index,a1,a2,a3,a4) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  FUNCTION_TABLE_viiii[index&8191](a1|0,a2|0,a3|0,a4|0);
}


function dynCall_viiiii(index,a1,a2,a3,a4,a5) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  FUNCTION_TABLE_viiiii[index&8191](a1|0,a2|0,a3|0,a4|0,a5|0);
}


function dynCall_vid(index,a1,a2) {
  index = index|0;
  a1=a1|0; a2=+a2;
  FUNCTION_TABLE_vid[index&8191](a1|0,+a2);
}


function dynCall_iiiidi(index,a1,a2,a3,a4,a5) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=+a4; a5=a5|0;
  return FUNCTION_TABLE_iiiidi[index&4095](a1|0,a2|0,a3|0,+a4,a5|0)|0;
}


function dynCall_iiidd(index,a1,a2,a3,a4) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=+a3; a4=+a4;
  return FUNCTION_TABLE_iiidd[index&4095](a1|0,a2|0,+a3,+a4)|0;
}


function dynCall_vii(index,a1,a2) {
  index = index|0;
  a1=a1|0; a2=a2|0;
  FUNCTION_TABLE_vii[index&8191](a1|0,a2|0);
}


function dynCall_viidd(index,a1,a2,a3,a4) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=+a3; a4=+a4;
  FUNCTION_TABLE_viidd[index&4095](a1|0,a2|0,+a3,+a4);
}


function dynCall_viddii(index,a1,a2,a3,a4,a5) {
  index = index|0;
  a1=a1|0; a2=+a2; a3=+a3; a4=a4|0; a5=a5|0;
  FUNCTION_TABLE_viddii[index&4095](a1|0,+a2,+a3,a4|0,a5|0);
}


function dynCall_viidi(index,a1,a2,a3,a4) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=+a3; a4=a4|0;
  FUNCTION_TABLE_viidi[index&4095](a1|0,a2|0,+a3,a4|0);
}


function dynCall_viddd(index,a1,a2,a3,a4) {
  index = index|0;
  a1=a1|0; a2=+a2; a3=+a3; a4=+a4;
  FUNCTION_TABLE_viddd[index&8191](a1|0,+a2,+a3,+a4);
}


function dynCall_viiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  FUNCTION_TABLE_viiiiiiii[index&8191](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}


function dynCall_viiiiiiid(index,a1,a2,a3,a4,a5,a6,a7,a8) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=+a8;
  FUNCTION_TABLE_viiiiiiid[index&4095](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,+a8);
}


function dynCall_viiiiiiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  FUNCTION_TABLE_viiiiiiiiiiiiiii[index&8191](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}


function dynCall_iid(index,a1,a2) {
  index = index|0;
  a1=a1|0; a2=+a2;
  return FUNCTION_TABLE_iid[index&4095](a1|0,+a2)|0;
}


function dynCall_dddd(index,a1,a2,a3) {
  index = index|0;
  a1=+a1; a2=+a2; a3=+a3;
  return +FUNCTION_TABLE_dddd[index&127](+a1,+a2,+a3);
}


function dynCall_viiiiiii(index,a1,a2,a3,a4,a5,a6,a7) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  FUNCTION_TABLE_viiiiiii[index&8191](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}


function dynCall_viiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0;
  FUNCTION_TABLE_viiiiiiiii[index&8191](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0);
}


function dynCall_iii(index,a1,a2) {
  index = index|0;
  a1=a1|0; a2=a2|0;
  return FUNCTION_TABLE_iii[index&8191](a1|0,a2|0)|0;
}


function dynCall_viiddi(index,a1,a2,a3,a4,a5) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=+a3; a4=+a4; a5=a5|0;
  FUNCTION_TABLE_viiddi[index&4095](a1|0,a2|0,+a3,+a4,a5|0);
}


function dynCall_viiddddddddd(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=+a3; a4=+a4; a5=+a5; a6=+a6; a7=+a7; a8=+a8; a9=+a9; a10=+a10; a11=+a11;
  FUNCTION_TABLE_viiddddddddd[index&4095](a1|0,a2|0,+a3,+a4,+a5,+a6,+a7,+a8,+a9,+a10,+a11);
}


function dynCall_vdddd(index,a1,a2,a3,a4) {
  index = index|0;
  a1=+a1; a2=+a2; a3=+a3; a4=+a4;
  FUNCTION_TABLE_vdddd[index&8191](+a1,+a2,+a3,+a4);
}


function dynCall_viii(index,a1,a2,a3) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0;
  FUNCTION_TABLE_viii[index&8191](a1|0,a2|0,a3|0);
}


function dynCall_v(index) {
  index = index|0;
  
  FUNCTION_TABLE_v[index&8191]();
}


function dynCall_viid(index,a1,a2,a3) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=+a3;
  FUNCTION_TABLE_viid[index&8191](a1|0,a2|0,+a3);
}


function dynCall_vd(index,a1) {
  index = index|0;
  a1=+a1;
  FUNCTION_TABLE_vd[index&8191](+a1);
}


function dynCall_vi(index,a1) {
  index = index|0;
  a1=a1|0;
  FUNCTION_TABLE_vi[index&8191](a1|0);
}


function dynCall_viddddddddd(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  index = index|0;
  a1=a1|0; a2=+a2; a3=+a3; a4=+a4; a5=+a5; a6=+a6; a7=+a7; a8=+a8; a9=+a9; a10=+a10;
  FUNCTION_TABLE_viddddddddd[index&8191](a1|0,+a2,+a3,+a4,+a5,+a6,+a7,+a8,+a9,+a10);
}


function dynCall_ii(index,a1) {
  index = index|0;
  a1=a1|0;
  return FUNCTION_TABLE_ii[index&8191](a1|0)|0;
}


function dynCall_ddddd(index,a1,a2,a3,a4) {
  index = index|0;
  a1=+a1; a2=+a2; a3=+a3; a4=+a4;
  return +FUNCTION_TABLE_ddddd[index&127](+a1,+a2,+a3,+a4);
}


function dynCall_iiiddddddddd(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=+a3; a4=+a4; a5=+a5; a6=+a6; a7=+a7; a8=+a8; a9=+a9; a10=+a10; a11=+a11;
  return FUNCTION_TABLE_iiiddddddddd[index&4095](a1|0,a2|0,+a3,+a4,+a5,+a6,+a7,+a8,+a9,+a10,+a11)|0;
}


function dynCall_iiii(index,a1,a2,a3) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0;
  return FUNCTION_TABLE_iiii[index&8191](a1|0,a2|0,a3|0)|0;
}


function dynCall_dd(index,a1) {
  index = index|0;
  a1=+a1;
  return +FUNCTION_TABLE_dd[index&4095](+a1);
}


function dynCall_viiiidii(index,a1,a2,a3,a4,a5,a6,a7) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5; a6=a6|0; a7=a7|0;
  FUNCTION_TABLE_viiiidii[index&4095](a1|0,a2|0,a3|0,a4|0,+a5,a6|0,a7|0);
}


function dynCall_diiiid(index,a1,a2,a3,a4,a5) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return +FUNCTION_TABLE_diiiid[index&1023](a1|0,a2|0,a3|0,a4|0,+a5);
}


function dynCall_iiidddddddddi(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=+a3; a4=+a4; a5=+a5; a6=+a6; a7=+a7; a8=+a8; a9=+a9; a10=+a10; a11=+a11; a12=a12|0;
  return FUNCTION_TABLE_iiidddddddddi[index&4095](a1|0,a2|0,+a3,+a4,+a5,+a6,+a7,+a8,+a9,+a10,+a11,a12|0)|0;
}


function dynCall_i(index) {
  index = index|0;
  
  return FUNCTION_TABLE_i[index&8191]()|0;
}


function dynCall_iiiiidii(index,a1,a2,a3,a4,a5,a6,a7) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5; a6=a6|0; a7=a7|0;
  return FUNCTION_TABLE_iiiiidii[index&8191](a1|0,a2|0,a3|0,a4|0,+a5,a6|0,a7|0)|0;
}


function dynCall_iiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return FUNCTION_TABLE_iiiiiiiii[index&8191](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}


function dynCall_viiiiiiiid(index,a1,a2,a3,a4,a5,a6,a7,a8,a9) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=+a9;
  FUNCTION_TABLE_viiiiiiiid[index&2047](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,+a9);
}


function dynCall_iiiidddddi(index,a1,a2,a3,a4,a5,a6,a7,a8,a9) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=+a4; a5=+a5; a6=+a6; a7=+a7; a8=+a8; a9=a9|0;
  return FUNCTION_TABLE_iiiidddddi[index&4095](a1|0,a2|0,a3|0,+a4,+a5,+a6,+a7,+a8,a9|0)|0;
}

function b0(p0,p1,p2,p3,p4,p5) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = +p5; nullFunc_iiiiiid(0);return 0;
}
function b1(p0,p1,p2,p3,p4,p5,p6) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = +p3;p4 = p4|0;p5 = p5|0;p6 = p6|0; nullFunc_viiidiii(1);
}
function b2(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = +p3;p4 = +p4;p5 = +p5;p6 = +p6;p7 = +p7;p8 = +p8;p9 = p9|0; nullFunc_viiiddddddi(2);
}
function b3(p0,p1,p2,p3,p4,p5) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0; nullFunc_iiiiiii(3);return 0;
}
function b4(p0,p1,p2,p3,p4,p5,p6) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = +p4;p5 = +p5;p6 = p6|0; nullFunc_viiiiddi(4);
}
function b5(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10) {
 p0 = p0|0;p1 = +p1;p2 = +p2;p3 = +p3;p4 = +p4;p5 = +p5;p6 = +p6;p7 = +p7;p8 = +p8;p9 = +p9;p10 = p10|0; nullFunc_iidddddddddi(5);return 0;
}
function b6(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0;p8 = p8|0;p9 = p9|0;p10 = p10|0; nullFunc_viiiiiiiiiii(6);
}
function b7(p0,p1,p2) {
 p0 = p0|0;p1 = +p1;p2 = +p2; nullFunc_iidd(7);return 0;
}
function b8(p0,p1,p2) {
 p0 = p0|0;p1 = +p1;p2 = p2|0; nullFunc_vidi(8);
}
function b9(p0,p1,p2) {
 p0 = p0|0;p1 = +p1;p2 = +p2; nullFunc_vidd(9);
}
function _emscripten_glUniform2f__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = +p1;p2 = +p2; _emscripten_glUniform2f(p0|0,+p1,+p2);
}
function _emscripten_glVertexAttrib2f__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = +p1;p2 = +p2; _emscripten_glVertexAttrib2f(p0|0,+p1,+p2);
}
function b10(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0;p8 = p8|0;p9 = p9|0;p10 = p10|0; nullFunc_iiiiiiiiiiii(10);return 0;
}
function b11(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12) {
 p0 = p0|0;p1 = p1|0;p2 = +p2;p3 = +p3;p4 = +p4;p5 = p5|0;p6 = +p6;p7 = +p7;p8 = +p8;p9 = +p9;p10 = +p10;p11 = +p11;p12 = +p12; nullFunc_viidddiddddddd(11);
}
function b12(p0,p1,p2,p3,p4,p5,p6,p7,p8) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = +p7;p8 = p8|0; nullFunc_viiiiiiidi(12);
}
function b13(p0) {
 p0 = p0|0; nullFunc_di(13);return +0;
}
function b14(p0,p1,p2,p3,p4,p5) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = +p3;p4 = +p4;p5 = p5|0; nullFunc_iiiiddi(14);return 0;
}
function b15(p0,p1,p2,p3,p4,p5) {
 p0 = p0|0;p1 = +p1;p2 = +p2;p3 = +p3;p4 = +p4;p5 = +p5; nullFunc_iiddddd(15);return 0;
}
function b16() {
 ; nullFunc_d(16);return +0;
}
function _emscripten_get_now__wrapper() {
 ; return +_emscripten_get_now();
}
function b17(p0,p1) {
 p0 = p0|0;p1 = p1|0; nullFunc_dii(17);return +0;
}
function b18(p0,p1,p2,p3,p4,p5,p6,p7,p8) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = +p3;p4 = +p4;p5 = +p5;p6 = +p6;p7 = +p7;p8 = p8|0; nullFunc_viiidddddi(18);
}
function b19(p0,p1) {
 p0 = p0|0;p1 = +p1; nullFunc_did(19);return +0;
}
function b20(p0,p1,p2,p3,p4,p5,p6,p7,p8) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0;p8 = p8|0; nullFunc_iiiiiiiiii(20);return 0;
}
function b21(p0,p1,p2,p3,p4,p5) {
 p0 = +p0;p1 = +p1;p2 = +p2;p3 = +p3;p4 = +p4;p5 = +p5; nullFunc_vdddddd(21);
}
function _emscripten_glFrustum__wrapper(p0,p1,p2,p3,p4,p5) {
 p0 = +p0;p1 = +p1;p2 = +p2;p3 = +p3;p4 = +p4;p5 = +p5; _emscripten_glFrustum(+p0,+p1,+p2,+p3,+p4,+p5);
}
function b22(p0,p1,p2,p3,p4,p5,p6,p7) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = +p3;p4 = +p4;p5 = +p5;p6 = +p6;p7 = p7|0; nullFunc_viiiddddi(22);
}
function b23(p0,p1) {
 p0 = +p0;p1 = +p1; nullFunc_vdd(23);
}
function _emscripten_glDepthRange__wrapper(p0,p1) {
 p0 = +p0;p1 = +p1; _emscripten_glDepthRange(+p0,+p1);
}
function _emscripten_glDepthRangef__wrapper(p0,p1) {
 p0 = +p0;p1 = +p1; _emscripten_glDepthRangef(+p0,+p1);
}
function _emscripten_glPolygonOffset__wrapper(p0,p1) {
 p0 = +p0;p1 = +p1; _emscripten_glPolygonOffset(+p0,+p1);
}
function b24(p0,p1) {
 p0 = +p0;p1 = p1|0; nullFunc_vdi(24);
}
function _emscripten_glSampleCoverage__wrapper(p0,p1) {
 p0 = +p0;p1 = p1|0; _emscripten_glSampleCoverage(+p0,p1|0);
}
function b25(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = +p3; nullFunc_iiiid(25);return 0;
}
function b26(p0,p1,p2,p3,p4,p5,p6) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0; nullFunc_iiiiiiii(26);return 0;
}
function b27(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = +p2;p3 = +p3;p4 = p4|0; nullFunc_iiiddi(27);return 0;
}
function b28(p0,p1,p2,p3,p4,p5,p6,p7) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = +p3;p4 = +p4;p5 = +p5;p6 = +p6;p7 = p7|0; nullFunc_iiiiddddi(28);return 0;
}
function b29(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = +p1;p2 = +p2;p3 = +p3;p4 = +p4; nullFunc_iidddd(29);return 0;
}
function b30(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = +p3;p4 = +p4;p5 = +p5;p6 = +p6;p7 = +p7;p8 = +p8;p9 = p9|0; nullFunc_iiiiddddddi(30);return 0;
}
function b31(p0,p1,p2,p3,p4,p5,p6,p7,p8) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = +p7;p8 = p8|0; nullFunc_iiiiiiiidi(31);return 0;
}
function b32(p0,p1,p2,p3,p4,p5) {
 p0 = p0|0;p1 = p1|0;p2 = +p2;p3 = +p3;p4 = +p4;p5 = +p5; nullFunc_viidddd(32);
}
function b33(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; nullFunc_iiiii(33);return 0;
}
function _eglGetConfigAttrib__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; return _eglGetConfigAttrib(p0|0,p1|0,p2|0,p3|0)|0;
}
function _eglCreateContext__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; return _eglCreateContext(p0|0,p1|0,p2|0,p3|0)|0;
}
function _eglCreateWindowSurface__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; return _eglCreateWindowSurface(p0|0,p1|0,p2|0,p3|0)|0;
}
function _eglMakeCurrent__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; return _eglMakeCurrent(p0|0,p1|0,p2|0,p3|0)|0;
}
function b34(p0,p1,p2,p3,p4,p5) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0; nullFunc_viiiiii(34);
}
function _emscripten_glDrawRangeElements__wrapper(p0,p1,p2,p3,p4,p5) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0; _emscripten_glDrawRangeElements(p0|0,p1|0,p2|0,p3|0,p4|0,p5|0);
}
function _emscripten_glVertexAttribPointer__wrapper(p0,p1,p2,p3,p4,p5) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0; _emscripten_glVertexAttribPointer(p0|0,p1|0,p2|0,p3|0,p4|0,p5|0);
}
function b35(p0,p1) {
 p0 = +p0;p1 = +p1; nullFunc_ddd(35);return +0;
}
function b36(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = +p3;p4 = p4|0; nullFunc_viiidi(36);
}
function b37(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = +p1;p2 = +p2;p3 = +p3;p4 = +p4; nullFunc_vidddd(37);
}
function _emscripten_glUniform4f__wrapper(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = +p1;p2 = +p2;p3 = +p3;p4 = +p4; _emscripten_glUniform4f(p0|0,+p1,+p2,+p3,+p4);
}
function _emscripten_glVertexAttrib4f__wrapper(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = +p1;p2 = +p2;p3 = +p3;p4 = +p4; _emscripten_glVertexAttrib4f(p0|0,+p1,+p2,+p3,+p4);
}
function b38(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = +p2; nullFunc_diid(38);return +0;
}
function b39(p0,p1,p2,p3) {
 p0 = p0|0;p1 = +p1;p2 = +p2;p3 = +p3; nullFunc_iiddd(39);return 0;
}
function b40(p0,p1,p2,p3) {
 p0 = p0|0;p1 = +p1;p2 = +p2;p3 = p3|0; nullFunc_iiddi(40);return 0;
}
function b41(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0;p8 = p8|0;p9 = p9|0; nullFunc_viiiiiiiiii(41);
}
function b42(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; nullFunc_diii(42);return +0;
}
function b43(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12) {
 p0 = p0|0;p1 = p1|0;p2 = +p2;p3 = p3|0;p4 = +p4;p5 = +p5;p6 = +p6;p7 = +p7;p8 = +p8;p9 = +p9;p10 = +p10;p11 = +p11;p12 = +p12; nullFunc_iiididdddddddd(43);return 0;
}
function b44(p0,p1,p2,p3,p4,p5) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = +p3;p4 = +p4;p5 = p5|0; nullFunc_viiiddi(44);
}
function b45(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0; nullFunc_iiiiii(45);return 0;
}
function _eglChooseConfig__wrapper(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0; return _eglChooseConfig(p0|0,p1|0,p2|0,p3|0,p4|0)|0;
}
function b46(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = +p3; nullFunc_viiid(46);
}
function b47(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = +p2;p3 = p3|0;p4 = p4|0; nullFunc_viidii(47);
}
function b48(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = +p4; nullFunc_iiiiid(48);return 0;
}
function b49(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; nullFunc_viiii(49);
}
function _emscripten_glBufferData__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glBufferData(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glBufferSubData__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glBufferSubData(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glUniform3i__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glUniform3i(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glUniformMatrix2fv__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glUniformMatrix2fv(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glUniformMatrix3fv__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glUniformMatrix3fv(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glUniformMatrix4fv__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glUniformMatrix4fv(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glGetAttachedShaders__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glGetAttachedShaders(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glShaderSource__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glShaderSource(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glGetShaderSource__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glGetShaderSource(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glGetShaderInfoLog__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glGetShaderInfoLog(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glGetShaderPrecisionFormat__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glGetShaderPrecisionFormat(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glGetProgramInfoLog__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glGetProgramInfoLog(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glFramebufferRenderbuffer__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glFramebufferRenderbuffer(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glGetFramebufferAttachmentParameteriv__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glGetFramebufferAttachmentParameteriv(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glGetInfoLogARB__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glGetInfoLogARB(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glVertexPointer__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glVertexPointer(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glTexCoordPointer__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glTexCoordPointer(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glColorPointer__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glColorPointer(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glDrawElements__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glDrawElements(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glDrawArraysInstanced__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glDrawArraysInstanced(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glViewport__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glViewport(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glScissor__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glScissor(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glColorMask__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glColorMask(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glRenderbufferStorage__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glRenderbufferStorage(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glBlendFuncSeparate__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glBlendFuncSeparate(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glStencilFuncSeparate__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glStencilFuncSeparate(p0|0,p1|0,p2|0,p3|0);
}
function _emscripten_glStencilOpSeparate__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; _emscripten_glStencilOpSeparate(p0|0,p1|0,p2|0,p3|0);
}
function b50(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0; nullFunc_viiiii(50);
}
function _alBufferData__wrapper(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0; _alBufferData(p0|0,p1|0,p2|0,p3|0,p4|0);
}
function _emscripten_glUniform4i__wrapper(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0; _emscripten_glUniform4i(p0|0,p1|0,p2|0,p3|0,p4|0);
}
function _emscripten_glFramebufferTexture2D__wrapper(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0; _emscripten_glFramebufferTexture2D(p0|0,p1|0,p2|0,p3|0,p4|0);
}
function _emscripten_glShaderBinary__wrapper(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0; _emscripten_glShaderBinary(p0|0,p1|0,p2|0,p3|0,p4|0);
}
function _emscripten_glDrawElementsInstanced__wrapper(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0; _emscripten_glDrawElementsInstanced(p0|0,p1|0,p2|0,p3|0,p4|0);
}
function b51(p0,p1) {
 p0 = p0|0;p1 = +p1; nullFunc_vid(51);
}
function _emscripten_glUniform1f__wrapper(p0,p1) {
 p0 = p0|0;p1 = +p1; _emscripten_glUniform1f(p0|0,+p1);
}
function _emscripten_glVertexAttrib1f__wrapper(p0,p1) {
 p0 = p0|0;p1 = +p1; _emscripten_glVertexAttrib1f(p0|0,+p1);
}
function b52(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = +p3;p4 = p4|0; nullFunc_iiiidi(52);return 0;
}
function b53(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = +p2;p3 = +p3; nullFunc_iiidd(53);return 0;
}
function b54(p0,p1) {
 p0 = p0|0;p1 = p1|0; nullFunc_vii(54);
}
function _alGenSources__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _alGenSources(p0|0,p1|0);
}
function _alDeleteSources__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _alDeleteSources(p0|0,p1|0);
}
function _alDeleteBuffers__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _alDeleteBuffers(p0|0,p1|0);
}
function _alGenBuffers__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _alGenBuffers(p0|0,p1|0);
}
function _emscripten_longjmp__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_longjmp(p0|0,p1|0);
}
function _emscripten_glPixelStorei__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glPixelStorei(p0|0,p1|0);
}
function _emscripten_glGetIntegerv__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glGetIntegerv(p0|0,p1|0);
}
function _emscripten_glGetFloatv__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glGetFloatv(p0|0,p1|0);
}
function _emscripten_glGetBooleanv__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glGetBooleanv(p0|0,p1|0);
}
function _emscripten_glGenTextures__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glGenTextures(p0|0,p1|0);
}
function _emscripten_glDeleteTextures__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glDeleteTextures(p0|0,p1|0);
}
function _emscripten_glBindTexture__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glBindTexture(p0|0,p1|0);
}
function _emscripten_glGenBuffers__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glGenBuffers(p0|0,p1|0);
}
function _emscripten_glDeleteBuffers__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glDeleteBuffers(p0|0,p1|0);
}
function _emscripten_glGenRenderbuffers__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glGenRenderbuffers(p0|0,p1|0);
}
function _emscripten_glDeleteRenderbuffers__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glDeleteRenderbuffers(p0|0,p1|0);
}
function _emscripten_glBindRenderbuffer__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glBindRenderbuffer(p0|0,p1|0);
}
function _emscripten_glUniform1i__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glUniform1i(p0|0,p1|0);
}
function _emscripten_glBindBuffer__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glBindBuffer(p0|0,p1|0);
}
function _emscripten_glVertexAttrib1fv__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glVertexAttrib1fv(p0|0,p1|0);
}
function _emscripten_glVertexAttrib2fv__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glVertexAttrib2fv(p0|0,p1|0);
}
function _emscripten_glVertexAttrib3fv__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glVertexAttrib3fv(p0|0,p1|0);
}
function _emscripten_glVertexAttrib4fv__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glVertexAttrib4fv(p0|0,p1|0);
}
function _emscripten_glAttachShader__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glAttachShader(p0|0,p1|0);
}
function _emscripten_glDetachShader__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glDetachShader(p0|0,p1|0);
}
function _emscripten_glBindFramebuffer__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glBindFramebuffer(p0|0,p1|0);
}
function _emscripten_glGenFramebuffers__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glGenFramebuffers(p0|0,p1|0);
}
function _emscripten_glDeleteFramebuffers__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glDeleteFramebuffers(p0|0,p1|0);
}
function _emscripten_glBindProgramARB__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glBindProgramARB(p0|0,p1|0);
}
function _emscripten_glGetPointerv__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glGetPointerv(p0|0,p1|0);
}
function _emscripten_glGenVertexArrays__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glGenVertexArrays(p0|0,p1|0);
}
function _emscripten_glDeleteVertexArrays__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glDeleteVertexArrays(p0|0,p1|0);
}
function _emscripten_glVertexAttribDivisor__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glVertexAttribDivisor(p0|0,p1|0);
}
function _emscripten_glBlendFunc__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glBlendFunc(p0|0,p1|0);
}
function _emscripten_glBlendEquationSeparate__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glBlendEquationSeparate(p0|0,p1|0);
}
function _emscripten_glStencilMaskSeparate__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glStencilMaskSeparate(p0|0,p1|0);
}
function _emscripten_glHint__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glHint(p0|0,p1|0);
}
function _emscripten_glDrawBuffers__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; _emscripten_glDrawBuffers(p0|0,p1|0);
}
function b55(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = +p2;p3 = +p3; nullFunc_viidd(55);
}
function b56(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = +p1;p2 = +p2;p3 = p3|0;p4 = p4|0; nullFunc_viddii(56);
}
function b57(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = +p2;p3 = p3|0; nullFunc_viidi(57);
}
function b58(p0,p1,p2,p3) {
 p0 = p0|0;p1 = +p1;p2 = +p2;p3 = +p3; nullFunc_viddd(58);
}
function _emscripten_glUniform3f__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = +p1;p2 = +p2;p3 = +p3; _emscripten_glUniform3f(p0|0,+p1,+p2,+p3);
}
function _emscripten_glVertexAttrib3f__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = +p1;p2 = +p2;p3 = +p3; _emscripten_glVertexAttrib3f(p0|0,+p1,+p2,+p3);
}
function b59(p0,p1,p2,p3,p4,p5,p6,p7) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0; nullFunc_viiiiiiii(59);
}
function _emscripten_glCompressedTexImage2D__wrapper(p0,p1,p2,p3,p4,p5,p6,p7) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0; _emscripten_glCompressedTexImage2D(p0|0,p1|0,p2|0,p3|0,p4|0,p5|0,p6|0,p7|0);
}
function _emscripten_glCopyTexImage2D__wrapper(p0,p1,p2,p3,p4,p5,p6,p7) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0; _emscripten_glCopyTexImage2D(p0|0,p1|0,p2|0,p3|0,p4|0,p5|0,p6|0,p7|0);
}
function _emscripten_glCopyTexSubImage2D__wrapper(p0,p1,p2,p3,p4,p5,p6,p7) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0; _emscripten_glCopyTexSubImage2D(p0|0,p1|0,p2|0,p3|0,p4|0,p5|0,p6|0,p7|0);
}
function b60(p0,p1,p2,p3,p4,p5,p6,p7) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = +p7; nullFunc_viiiiiiid(60);
}
function b61(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0;p8 = p8|0;p9 = p9|0;p10 = p10|0;p11 = p11|0;p12 = p12|0;p13 = p13|0;p14 = p14|0; nullFunc_viiiiiiiiiiiiiii(61);
}
function b62(p0,p1) {
 p0 = p0|0;p1 = +p1; nullFunc_iid(62);return 0;
}
function b63(p0,p1,p2) {
 p0 = +p0;p1 = +p1;p2 = +p2; nullFunc_dddd(63);return +0;
}
function b64(p0,p1,p2,p3,p4,p5,p6) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0; nullFunc_viiiiiii(64);
}
function _emscripten_glReadPixels__wrapper(p0,p1,p2,p3,p4,p5,p6) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0; _emscripten_glReadPixels(p0|0,p1|0,p2|0,p3|0,p4|0,p5|0,p6|0);
}
function _emscripten_glGetActiveUniform__wrapper(p0,p1,p2,p3,p4,p5,p6) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0; _emscripten_glGetActiveUniform(p0|0,p1|0,p2|0,p3|0,p4|0,p5|0,p6|0);
}
function _emscripten_glGetActiveAttrib__wrapper(p0,p1,p2,p3,p4,p5,p6) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0; _emscripten_glGetActiveAttrib(p0|0,p1|0,p2|0,p3|0,p4|0,p5|0,p6|0);
}
function b65(p0,p1,p2,p3,p4,p5,p6,p7,p8) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0;p8 = p8|0; nullFunc_viiiiiiiii(65);
}
function _emscripten_glCompressedTexSubImage2D__wrapper(p0,p1,p2,p3,p4,p5,p6,p7,p8) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0;p8 = p8|0; _emscripten_glCompressedTexSubImage2D(p0|0,p1|0,p2|0,p3|0,p4|0,p5|0,p6|0,p7|0,p8|0);
}
function _emscripten_glTexImage2D__wrapper(p0,p1,p2,p3,p4,p5,p6,p7,p8) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0;p8 = p8|0; _emscripten_glTexImage2D(p0|0,p1|0,p2|0,p3|0,p4|0,p5|0,p6|0,p7|0,p8|0);
}
function _emscripten_glTexSubImage2D__wrapper(p0,p1,p2,p3,p4,p5,p6,p7,p8) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0;p8 = p8|0; _emscripten_glTexSubImage2D(p0|0,p1|0,p2|0,p3|0,p4|0,p5|0,p6|0,p7|0,p8|0);
}
function b66(p0,p1) {
 p0 = p0|0;p1 = p1|0; nullFunc_iii(66);return 0;
}
function _alcCreateContext__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; return _alcCreateContext(p0|0,p1|0)|0;
}
function _eglDestroyContext__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; return _eglDestroyContext(p0|0,p1|0)|0;
}
function _eglDestroySurface__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; return _eglDestroySurface(p0|0,p1|0)|0;
}
function _eglSwapBuffers__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; return _eglSwapBuffers(p0|0,p1|0)|0;
}
function _eglSwapInterval__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; return _eglSwapInterval(p0|0,p1|0)|0;
}
function _emscripten_glGetUniformLocation__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; return _emscripten_glGetUniformLocation(p0|0,p1|0)|0;
}
function _emscripten_glGetAttribLocation__wrapper(p0,p1) {
 p0 = p0|0;p1 = p1|0; return _emscripten_glGetAttribLocation(p0|0,p1|0)|0;
}
function b67(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = +p2;p3 = +p3;p4 = p4|0; nullFunc_viiddi(67);
}
function b68(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10) {
 p0 = p0|0;p1 = p1|0;p2 = +p2;p3 = +p3;p4 = +p4;p5 = +p5;p6 = +p6;p7 = +p7;p8 = +p8;p9 = +p9;p10 = +p10; nullFunc_viiddddddddd(68);
}
function b69(p0,p1,p2,p3) {
 p0 = +p0;p1 = +p1;p2 = +p2;p3 = +p3; nullFunc_vdddd(69);
}
function _emscripten_glRotatef__wrapper(p0,p1,p2,p3) {
 p0 = +p0;p1 = +p1;p2 = +p2;p3 = +p3; _emscripten_glRotatef(+p0,+p1,+p2,+p3);
}
function _emscripten_glClearColor__wrapper(p0,p1,p2,p3) {
 p0 = +p0;p1 = +p1;p2 = +p2;p3 = +p3; _emscripten_glClearColor(+p0,+p1,+p2,+p3);
}
function _emscripten_glBlendColor__wrapper(p0,p1,p2,p3) {
 p0 = +p0;p1 = +p1;p2 = +p2;p3 = +p3; _emscripten_glBlendColor(+p0,+p1,+p2,+p3);
}
function b70(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; nullFunc_viii(70);
}
function ___cxa_throw__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; ___cxa_throw(p0|0,p1|0,p2|0);
}
function _emscripten_glGetTexParameterfv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glGetTexParameterfv(p0|0,p1|0,p2|0);
}
function _emscripten_glGetTexParameteriv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glGetTexParameteriv(p0|0,p1|0,p2|0);
}
function _emscripten_glTexParameterfv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glTexParameterfv(p0|0,p1|0,p2|0);
}
function _emscripten_glTexParameteriv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glTexParameteriv(p0|0,p1|0,p2|0);
}
function _emscripten_glGetBufferParameteriv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glGetBufferParameteriv(p0|0,p1|0,p2|0);
}
function _emscripten_glGetRenderbufferParameteriv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glGetRenderbufferParameteriv(p0|0,p1|0,p2|0);
}
function _emscripten_glGetUniformfv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glGetUniformfv(p0|0,p1|0,p2|0);
}
function _emscripten_glGetUniformiv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glGetUniformiv(p0|0,p1|0,p2|0);
}
function _emscripten_glGetVertexAttribfv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glGetVertexAttribfv(p0|0,p1|0,p2|0);
}
function _emscripten_glGetVertexAttribiv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glGetVertexAttribiv(p0|0,p1|0,p2|0);
}
function _emscripten_glGetVertexAttribPointerv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glGetVertexAttribPointerv(p0|0,p1|0,p2|0);
}
function _emscripten_glUniform2i__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glUniform2i(p0|0,p1|0,p2|0);
}
function _emscripten_glUniform1iv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glUniform1iv(p0|0,p1|0,p2|0);
}
function _emscripten_glUniform2iv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glUniform2iv(p0|0,p1|0,p2|0);
}
function _emscripten_glUniform3iv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glUniform3iv(p0|0,p1|0,p2|0);
}
function _emscripten_glUniform4iv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glUniform4iv(p0|0,p1|0,p2|0);
}
function _emscripten_glUniform1fv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glUniform1fv(p0|0,p1|0,p2|0);
}
function _emscripten_glUniform2fv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glUniform2fv(p0|0,p1|0,p2|0);
}
function _emscripten_glUniform3fv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glUniform3fv(p0|0,p1|0,p2|0);
}
function _emscripten_glUniform4fv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glUniform4fv(p0|0,p1|0,p2|0);
}
function _emscripten_glGetShaderiv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glGetShaderiv(p0|0,p1|0,p2|0);
}
function _emscripten_glGetProgramiv__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glGetProgramiv(p0|0,p1|0,p2|0);
}
function _emscripten_glBindAttribLocation__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glBindAttribLocation(p0|0,p1|0,p2|0);
}
function _emscripten_glGetObjectParameterivARB__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glGetObjectParameterivARB(p0|0,p1|0,p2|0);
}
function _emscripten_glNormalPointer__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glNormalPointer(p0|0,p1|0,p2|0);
}
function _emscripten_glDrawArrays__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glDrawArrays(p0|0,p1|0,p2|0);
}
function _emscripten_glTexParameteri__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glTexParameteri(p0|0,p1|0,p2|0);
}
function _emscripten_glStencilFunc__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glStencilFunc(p0|0,p1|0,p2|0);
}
function _emscripten_glStencilOp__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; _emscripten_glStencilOp(p0|0,p1|0,p2|0);
}
function b71() {
 ; nullFunc_v(71);
}
function ___cxa_pure_virtual__wrapper() {
 ; ___cxa_pure_virtual();
}
function _abort__wrapper() {
 ; _abort();
}
function ___cxa_end_catch__wrapper() {
 ; ___cxa_end_catch();
}
function ___cxa_rethrow__wrapper() {
 ; ___cxa_rethrow();
}
function _emscripten_glLoadIdentity__wrapper() {
 ; _emscripten_glLoadIdentity();
}
function _emscripten_glReleaseShaderCompiler__wrapper() {
 ; _emscripten_glReleaseShaderCompiler();
}
function _emscripten_glFinish__wrapper() {
 ; _emscripten_glFinish();
}
function _emscripten_glFlush__wrapper() {
 ; _emscripten_glFlush();
}
function b72(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = +p2; nullFunc_viid(72);
}
function _emscripten_glTexParameterf__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = +p2; _emscripten_glTexParameterf(p0|0,p1|0,+p2);
}
function b73(p0) {
 p0 = +p0; nullFunc_vd(73);
}
function _emscripten_glClearDepth__wrapper(p0) {
 p0 = +p0; _emscripten_glClearDepth(+p0);
}
function _emscripten_glClearDepthf__wrapper(p0) {
 p0 = +p0; _emscripten_glClearDepthf(+p0);
}
function _emscripten_glLineWidth__wrapper(p0) {
 p0 = +p0; _emscripten_glLineWidth(+p0);
}
function b74(p0) {
 p0 = p0|0; nullFunc_vi(74);
}
function _alcDestroyContext__wrapper(p0) {
 p0 = p0|0; _alcDestroyContext(p0|0);
}
function _emscripten_glDeleteShader__wrapper(p0) {
 p0 = p0|0; _emscripten_glDeleteShader(p0|0);
}
function _emscripten_glCompileShader__wrapper(p0) {
 p0 = p0|0; _emscripten_glCompileShader(p0|0);
}
function _emscripten_glDeleteProgram__wrapper(p0) {
 p0 = p0|0; _emscripten_glDeleteProgram(p0|0);
}
function _emscripten_glLinkProgram__wrapper(p0) {
 p0 = p0|0; _emscripten_glLinkProgram(p0|0);
}
function _emscripten_glUseProgram__wrapper(p0) {
 p0 = p0|0; _emscripten_glUseProgram(p0|0);
}
function _emscripten_glValidateProgram__wrapper(p0) {
 p0 = p0|0; _emscripten_glValidateProgram(p0|0);
}
function _emscripten_glDeleteObjectARB__wrapper(p0) {
 p0 = p0|0; _emscripten_glDeleteObjectARB(p0|0);
}
function _emscripten_glEnableClientState__wrapper(p0) {
 p0 = p0|0; _emscripten_glEnableClientState(p0|0);
}
function _emscripten_glClientActiveTexture__wrapper(p0) {
 p0 = p0|0; _emscripten_glClientActiveTexture(p0|0);
}
function _emscripten_glBindVertexArray__wrapper(p0) {
 p0 = p0|0; _emscripten_glBindVertexArray(p0|0);
}
function _emscripten_glMatrixMode__wrapper(p0) {
 p0 = p0|0; _emscripten_glMatrixMode(p0|0);
}
function _emscripten_glLoadMatrixf__wrapper(p0) {
 p0 = p0|0; _emscripten_glLoadMatrixf(p0|0);
}
function _emscripten_glEnableVertexAttribArray__wrapper(p0) {
 p0 = p0|0; _emscripten_glEnableVertexAttribArray(p0|0);
}
function _emscripten_glDisableVertexAttribArray__wrapper(p0) {
 p0 = p0|0; _emscripten_glDisableVertexAttribArray(p0|0);
}
function _emscripten_glDepthFunc__wrapper(p0) {
 p0 = p0|0; _emscripten_glDepthFunc(p0|0);
}
function _emscripten_glEnable__wrapper(p0) {
 p0 = p0|0; _emscripten_glEnable(p0|0);
}
function _emscripten_glDisable__wrapper(p0) {
 p0 = p0|0; _emscripten_glDisable(p0|0);
}
function _emscripten_glFrontFace__wrapper(p0) {
 p0 = p0|0; _emscripten_glFrontFace(p0|0);
}
function _emscripten_glCullFace__wrapper(p0) {
 p0 = p0|0; _emscripten_glCullFace(p0|0);
}
function _emscripten_glClear__wrapper(p0) {
 p0 = p0|0; _emscripten_glClear(p0|0);
}
function _emscripten_glClearStencil__wrapper(p0) {
 p0 = p0|0; _emscripten_glClearStencil(p0|0);
}
function _emscripten_glDepthMask__wrapper(p0) {
 p0 = p0|0; _emscripten_glDepthMask(p0|0);
}
function _emscripten_glStencilMask__wrapper(p0) {
 p0 = p0|0; _emscripten_glStencilMask(p0|0);
}
function _emscripten_glGenerateMipmap__wrapper(p0) {
 p0 = p0|0; _emscripten_glGenerateMipmap(p0|0);
}
function _emscripten_glActiveTexture__wrapper(p0) {
 p0 = p0|0; _emscripten_glActiveTexture(p0|0);
}
function _emscripten_glBlendEquation__wrapper(p0) {
 p0 = p0|0; _emscripten_glBlendEquation(p0|0);
}
function b75(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9) {
 p0 = p0|0;p1 = +p1;p2 = +p2;p3 = +p3;p4 = +p4;p5 = +p5;p6 = +p6;p7 = +p7;p8 = +p8;p9 = +p9; nullFunc_viddddddddd(75);
}
function b76(p0) {
 p0 = p0|0; nullFunc_ii(76);return 0;
}
function _alcMakeContextCurrent__wrapper(p0) {
 p0 = p0|0; return _alcMakeContextCurrent(p0|0)|0;
}
function _alcCloseDevice__wrapper(p0) {
 p0 = p0|0; return _alcCloseDevice(p0|0)|0;
}
function _alcOpenDevice__wrapper(p0) {
 p0 = p0|0; return _alcOpenDevice(p0|0)|0;
}
function _alcGetError__wrapper(p0) {
 p0 = p0|0; return _alcGetError(p0|0)|0;
}
function _eglGetDisplay__wrapper(p0) {
 p0 = p0|0; return _eglGetDisplay(p0|0)|0;
}
function _eglTerminate__wrapper(p0) {
 p0 = p0|0; return _eglTerminate(p0|0)|0;
}
function _eglGetProcAddress__wrapper(p0) {
 p0 = p0|0; return _eglGetProcAddress(p0|0)|0;
}
function _eglWaitNative__wrapper(p0) {
 p0 = p0|0; return _eglWaitNative(p0|0)|0;
}
function _eglBindAPI__wrapper(p0) {
 p0 = p0|0; return _eglBindAPI(p0|0)|0;
}
function _emscripten_glGetString__wrapper(p0) {
 p0 = p0|0; return _emscripten_glGetString(p0|0)|0;
}
function _emscripten_glIsTexture__wrapper(p0) {
 p0 = p0|0; return _emscripten_glIsTexture(p0|0)|0;
}
function _emscripten_glIsBuffer__wrapper(p0) {
 p0 = p0|0; return _emscripten_glIsBuffer(p0|0)|0;
}
function _emscripten_glIsRenderbuffer__wrapper(p0) {
 p0 = p0|0; return _emscripten_glIsRenderbuffer(p0|0)|0;
}
function _emscripten_glCreateShader__wrapper(p0) {
 p0 = p0|0; return _emscripten_glCreateShader(p0|0)|0;
}
function _emscripten_glIsShader__wrapper(p0) {
 p0 = p0|0; return _emscripten_glIsShader(p0|0)|0;
}
function _emscripten_glIsProgram__wrapper(p0) {
 p0 = p0|0; return _emscripten_glIsProgram(p0|0)|0;
}
function _emscripten_glIsFramebuffer__wrapper(p0) {
 p0 = p0|0; return _emscripten_glIsFramebuffer(p0|0)|0;
}
function _emscripten_glCheckFramebufferStatus__wrapper(p0) {
 p0 = p0|0; return _emscripten_glCheckFramebufferStatus(p0|0)|0;
}
function _emscripten_glIsEnabled__wrapper(p0) {
 p0 = p0|0; return _emscripten_glIsEnabled(p0|0)|0;
}
function _pthread_mutex_unlock__wrapper(p0) {
 p0 = p0|0; return _pthread_mutex_unlock(p0|0)|0;
}
function _pthread_mutex_lock__wrapper(p0) {
 p0 = p0|0; return _pthread_mutex_lock(p0|0)|0;
}
function _pthread_cond_broadcast__wrapper(p0) {
 p0 = p0|0; return _pthread_cond_broadcast(p0|0)|0;
}
function b77(p0,p1,p2,p3) {
 p0 = +p0;p1 = +p1;p2 = +p2;p3 = +p3; nullFunc_ddddd(77);return +0;
}
function b78(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10) {
 p0 = p0|0;p1 = p1|0;p2 = +p2;p3 = +p3;p4 = +p4;p5 = +p5;p6 = +p6;p7 = +p7;p8 = +p8;p9 = +p9;p10 = +p10; nullFunc_iiiddddddddd(78);return 0;
}
function b79(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; nullFunc_iiii(79);return 0;
}
function _eglInitialize__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; return _eglInitialize(p0|0,p1|0,p2|0)|0;
}
function b80(p0) {
 p0 = +p0; nullFunc_dd(80);return +0;
}
function b81(p0,p1,p2,p3,p4,p5,p6) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = +p4;p5 = p5|0;p6 = p6|0; nullFunc_viiiidii(81);
}
function b82(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = +p4; nullFunc_diiiid(82);return +0;
}
function b83(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11) {
 p0 = p0|0;p1 = p1|0;p2 = +p2;p3 = +p3;p4 = +p4;p5 = +p5;p6 = +p6;p7 = +p7;p8 = +p8;p9 = +p9;p10 = +p10;p11 = p11|0; nullFunc_iiidddddddddi(83);return 0;
}
function b84() {
 ; nullFunc_i(84);return 0;
}
function _alGetError__wrapper() {
 ; return _alGetError()|0;
}
function _eglWaitGL__wrapper() {
 ; return _eglWaitGL()|0;
}
function _emscripten_glCreateProgram__wrapper() {
 ; return _emscripten_glCreateProgram()|0;
}
function _emscripten_glGetError__wrapper() {
 ; return _emscripten_glGetError()|0;
}
function b85(p0,p1,p2,p3,p4,p5,p6) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = +p4;p5 = p5|0;p6 = p6|0; nullFunc_iiiiidii(85);return 0;
}
function b86(p0,p1,p2,p3,p4,p5,p6,p7) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0; nullFunc_iiiiiiiii(86);return 0;
}
function b87(p0,p1,p2,p3,p4,p5,p6,p7,p8) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0;p8 = +p8; nullFunc_viiiiiiiid(87);
}
function b88(p0,p1,p2,p3,p4,p5,p6,p7,p8) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = +p3;p4 = +p4;p5 = +p5;p6 = +p6;p7 = +p7;p8 = p8|0; nullFunc_iiiidddddi(88);return 0;
}

// EMSCRIPTEN_END_FUNCS
var FUNCTION_TABLE_iiiiiid = [b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,__ZNKSt3__19money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_bRNS_8ios_baseEce,b0,b0,b0,__ZNKSt3__19money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_bRNS_8ios_baseEwe,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0];
var FUNCTION_TABLE_viiidiii = [b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,__ZN4love8graphics6opengl4Font25generateVerticesFormattedERKNS2_17ColoredCodepointsEfNS2_9AlignModeERNSt3__16vectorINS2_11GlyphVertexENS7_9allocatorIS9_EEEEPNS2_8TextInfoE,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1];
var FUNCTION_TABLE_viiiddddddi = [b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,__ZN4love7physics5box2d14PrismaticJointC2EPNS1_4BodyES4_ffffffb,b2,b2,b2,b2,__ZN4love7physics5box2d10WheelJointC2EPNS1_4BodyES4_ffffffb
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2];
var FUNCTION_TABLE_iiiiiii = [b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZN4love4math13LZ4Compressor8compressENS0_10Compressor6FormatEPKcjiRj,b3,b3,b3,__ZN4love4math14zlibCompressor8compressENS0_10Compressor6FormatEPKcjiRj,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZN4love5image6magpie10KTXHandler5parseEPNS_10filesystem8FileDataERNSt3__16vectorINS0_19CompressedImageData8SubImageENS6_9allocatorIS9_EEEERjRNS8_6FormatERb,b3,b3,__ZN4love5image6magpie10PKMHandler5parseEPNS_10filesystem8FileDataERNSt3__16vectorINS0_19CompressedImageData8SubImageENS6_9allocatorIS9_EEEERjRNS8_6FormatERb,b3,b3,b3,b3,b3,b3,b3,b3
,b3,__ZN4love5image6magpie10PVRHandler5parseEPNS_10filesystem8FileDataERNSt3__16vectorINS0_19CompressedImageData8SubImageENS6_9allocatorIS9_EEEERjRNS8_6FormatERb,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZN4love5image6magpie11ASTCHandler5parseEPNS_10filesystem8FileDataERNSt3__16vectorINS0_19CompressedImageData8SubImageENS6_9allocatorIS9_EEEERjRNS8_6FormatERb,b3,b3,b3,b3,__ZN4love5image6magpie10DDSHandler5parseEPNS_10filesystem8FileDataERNSt3__16vectorINS0_19CompressedImageData8SubImageENS6_9allocatorIS9_EEEERjRNS8_6FormatERb
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,_ps_unicodes_init,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRb,__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRl,__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRx,__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRt,__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjS8_,__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRm,__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRy
,__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRf,__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRd,__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRe,__ZNKSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRPv,b3,b3,__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRb,__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRl,__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRx,__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRt,__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjS8_,__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRm,__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRy,__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRf,__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRd,__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRe,__ZNKSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRPv,b3,b3,b3,b3,__ZNKSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcx,b3,__ZNKSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcy,b3,b3,b3,b3,b3,b3
,b3,__ZNKSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwx,b3,__ZNKSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwy,b3,b3,b3,b3,b3,b3,__ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11do_get_timeES4_S4_RNS_8ios_baseERjP2tm,__ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11do_get_dateES4_S4_RNS_8ios_baseERjP2tm,__ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE14do_get_weekdayES4_S4_RNS_8ios_baseERjP2tm,__ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE16do_get_monthnameES4_S4_RNS_8ios_baseERjP2tm,__ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11do_get_yearES4_S4_RNS_8ios_baseERjP2tm,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11do_get_timeES4_S4_RNS_8ios_baseERjP2tm,__ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11do_get_dateES4_S4_RNS_8ios_baseERjP2tm,__ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE14do_get_weekdayES4_S4_RNS_8ios_baseERjP2tm,__ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE16do_get_monthnameES4_S4_RNS_8ios_baseERjP2tm
,__ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11do_get_yearES4_S4_RNS_8ios_baseERjP2tm,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZNKSt3__19money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_bRNS_8ios_baseEcRKNS_12basic_stringIcS3_NS_9allocatorIcEEEE,b3,b3,b3,__ZNKSt3__19money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_bRNS_8ios_baseEwRKNS_12basic_stringIwS3_NS_9allocatorIwEEEE,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__116__pad_and_outputIcNS_11char_traitsIcEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,_enet_range_coder_compress
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZN4love8graphics6opengl8Graphics7newMeshERKNSt3__16vectorINS1_4Mesh12AttribFormatENS3_9allocatorIS6_EEEEPKvjNS5_8DrawModeENS5_5UsageE,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZN4love8graphics6opengl6Canvas12newImageDataEPNS_5image5ImageEiiii,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,__Z14lodepng_decodePPhPjS1_P12LodePNGStatePKhj,b3,b3,__Z14lodepng_encodePPhPjPKhjjP12LodePNGState,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__116__pad_and_outputIwNS_11char_traitsIwEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3];
var FUNCTION_TABLE_viiiiddi = [b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZN4love8graphics6opengl8Polyline6renderEPKfjjffb,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4];
var FUNCTION_TABLE_iidddddddddi = [b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,__ZN4love8graphics6opengl11SpriteBatch3addEfffffffffi,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5];
var FUNCTION_TABLE_viiiiiiiiiii = [b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,_SDL_BlendLine_RGB565,_SDL_BlendLine_RGB2,_SDL_BlendLine_RGB555,_SDL_BlendLine_ARGB8888,_SDL_BlendLine_RGB888,_SDL_BlendLine_RGBA4,_SDL_BlendLine_RGB4,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6];
var FUNCTION_TABLE_iidd = [b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,__ZN4love7physics5box2d7Physics17newRectangleShapeEff,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,__ZNK4love4math11BezierCurve10getSegmentEdd,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7];
var FUNCTION_TABLE_vidi = [b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,__ZN4love5audio4null6Source4seekEfNS0_6Source4UnitE,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,__ZN4love5audio6openal6Source4seekEfNS0_6Source4UnitE,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,__ZN4love5audio6openal6Source10seekAtomicEfPv,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8];
var FUNCTION_TABLE_vidd = [b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,__ZN4love5mouse3sdl5Mouse11setPositionEdd,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,__ZN4love7physics5box2d14PrismaticJoint9setLimitsEff,b9,b9,__ZN4love7physics5box2d13RevoluteJoint9setLimitsEff,b9,b9,__ZN4love7physics5box2d5World15translateOriginEff,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,__ZN4love7physics5box2d4Body11setPositionEff,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,__ZN4love7physics5box2d10ChainShape13setNextVertexEff,__ZN4love7physics5box2d10ChainShape17setPreviousVertexEff,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,_emscripten_glUniform2f__wrapper,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,_emscripten_glVertexAttrib2f__wrapper,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9];
var FUNCTION_TABLE_iiiiiiiiiiii = [b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,__ZNSt3__19money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE8__do_getERS4_S4_bRKNS_6localeEjRjRbRKNS_5ctypeIcEERNS_10unique_ptrIcPFvPvEEERPcSM_,b10,b10,b10,b10,b10,__ZNSt3__19money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE8__do_getERS4_S4_bRKNS_6localeEjRjRbRKNS_5ctypeIwEERNS_10unique_ptrIwPFvPvEEERPwSM_,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10];
var FUNCTION_TABLE_viidddiddddddd = [b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,__ZN4love8graphics6opengl8Graphics6printfERKNSt3__16vectorINS1_4Font13ColoredStringENS3_9allocatorIS6_EEEEfffNS5_9AlignModeEfffffff,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11];
var FUNCTION_TABLE_viiiiiiidi = [b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,__ZN4love7physics5box2d11PulleyJointC2EPNS1_4BodyES4_6b2Vec2S5_S5_S5_fb,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12];
var FUNCTION_TABLE_di = [b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,__ZN4love4math3__28__invokeEPNS_5ProxyE,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,__ZNK4love5mouse3sdl5Mouse4getXEv,__ZNK4love5mouse3sdl5Mouse4getYEv,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,__ZN4love5sound7lullaby13VorbisDecoder11getDurationEv,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,__ZN4love5sound7lullaby11WaveDecoder11getDurationEv,b13,b13,b13,b13,b13,b13,b13,b13,__ZNK4love5sound9SoundData11getDurationEv,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,__ZNK4love5timer5Timer8getDeltaEv,b13,__ZNK4love5timer5Timer15getAverageDeltaEv,__ZNK4love5timer5Timer7getTimeEv,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,__ZNK4love5video11VideoStream4tellEv,b13,b13,b13,b13,b13,__ZNK4love5video11VideoStream9DeltaSync11getPositionEv,b13
,b13,b13,b13,__ZNK4love5video11VideoStream9FrameSync4tellEv,b13,b13,b13,__ZNK4love5video11VideoStream10SourceSync11getPositionEv,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,__ZNK4love6window3sdl6Window13getPixelScaleEv,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,__ZN4love5sound7lullaby13Mpg123Decoder11getDurationEv,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,__ZNK4love5audio4null5Audio9getVolumeEv,b13,b13,b13,b13,b13,b13,b13,__ZNK4love5audio4null5Audio15getDopplerScaleEv,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,__ZNK4love5audio4null6Source8getPitchEv,b13,__ZNK4love5audio4null6Source9getVolumeEv,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,__ZNK4love5audio4null6Source12getMinVolumeEv,b13,__ZNK4love5audio4null6Source12getMaxVolumeEv,b13,__ZNK4love5audio4null6Source20getReferenceDistanceEv,b13,__ZNK4love5audio4null6Source16getRolloffFactorEv,b13,__ZNK4love5audio4null6Source14getMaxDistanceEv,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,__ZNK4love5audio6openal5Audio9getVolumeEv,b13,b13,b13,b13,b13,b13,b13,__ZNK4love5audio6openal5Audio15getDopplerScaleEv,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,__ZNK4love5audio6openal6Source8getPitchEv,b13,__ZNK4love5audio6openal6Source9getVolumeEv,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,__ZNK4love5audio6openal6Source12getMinVolumeEv,b13,__ZNK4love5audio6openal6Source12getMaxVolumeEv,b13,__ZNK4love5audio6openal6Source20getReferenceDistanceEv,b13,__ZNK4love5audio6openal6Source16getRolloffFactorEv,b13,__ZNK4love5audio6openal6Source14getMaxDistanceEv
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,__ZN4love5sound7lullaby14ModPlugDecoder11getDurationEv,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,__ZL9pat_sinusi,__ZL10pat_squarei,__ZL12pat_sawtoothi,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,__ZNK7b2Timer15GetMillisecondsEv,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13];
var FUNCTION_TABLE_iiiiddi = [b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,__ZN4love7physics5box2d7Physics16newRevoluteJointEPNS1_4BodyES4_ffb,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14];
var FUNCTION_TABLE_iiddddd = [b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,__ZN4love7physics5box2d7Physics17newRectangleShapeEfffff,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15];
var FUNCTION_TABLE_d = [b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,_emscripten_get_now__wrapper,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16];
var FUNCTION_TABLE_dii = [b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,__ZN4love5audio4null6Source4tellENS0_6Source4UnitE,__ZN4love5audio4null6Source11getDurationENS0_6Source4UnitE,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,__ZN4love5audio6openal6Source4tellENS0_6Source4UnitE,__ZN4love5audio6openal6Source11getDurationENS0_6Source4UnitE,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,__ZNK4love5audio6openal6Source10tellAtomicEPv,__ZN4love5audio6openal6Source17getDurationAtomicEPv,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,__ZNK4love8joystick3sdl8Joystick7getAxisEi,b17,b17,b17,b17,b17,__ZNK4love8joystick3sdl8Joystick14getGamepadAxisENS0_8Joystick11GamepadAxisE,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,_luaL_checknumber,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,__ZNK4love5sound9SoundData9getSampleEi,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,_lua_tonumber,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,_INT123_init_layer3_gainpow2,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17];
var FUNCTION_TABLE_viiidddddi = [b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,__ZN4love7physics5box2d9RopeJointC2EPNS1_4BodyES4_fffffb,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18];
var FUNCTION_TABLE_did = [b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,__ZNK4love6window3sdl6Window8toPixelsEd,b19,__ZNK4love6window3sdl6Window10fromPixelsEd,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,__ZNK12b2MotorJoint17GetReactionTorqueEf,b19,b19,b19,b19,b19,b19
,b19,b19,__ZNK12b2MouseJoint17GetReactionTorqueEf,b19,b19,b19,b19,b19,b19,b19,b19,b19,__ZNK16b2PrismaticJoint17GetReactionTorqueEf,b19,b19,b19,b19,b19,b19,b19,b19,__ZNK13b2PulleyJoint17GetReactionTorqueEf,b19,b19,b19,b19,b19,b19,b19,b19
,b19,__ZNK15b2RevoluteJoint17GetReactionTorqueEf,b19,b19,b19,b19,b19,b19,b19,b19,__ZNK11b2RopeJoint17GetReactionTorqueEf,b19,b19,b19,b19,b19,b19,b19,b19,__ZNK11b2WeldJoint17GetReactionTorqueEf,b19,b19,b19,b19,b19,b19,b19,b19,__ZNK12b2WheelJoint17GetReactionTorqueEf,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,__ZNK15b2DistanceJoint17GetReactionTorqueEf,b19,b19,b19,b19,b19,b19,b19,b19,__ZNK15b2FrictionJoint17GetReactionTorqueEf,b19,b19,b19,b19,b19,b19,b19,b19,__ZNK11b2GearJoint17GetReactionTorqueEf,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19];
var FUNCTION_TABLE_iiiiiiiiii = [b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,_t1_decoder_init,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,_SDL_CreateRGBSurfaceFrom,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,_GLES2_UpdateTextureYUV,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20];
var FUNCTION_TABLE_vdddddd = [b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,_emscripten_glFrustum__wrapper,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21];
var FUNCTION_TABLE_viiiddddi = [b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,__ZN4love7physics5box2d13DistanceJointC2EPNS1_4BodyES4_ffffb,b22,b22,b22,b22,b22,__ZN4love7physics5box2d13FrictionJointC2EPNS1_4BodyES4_ffffb,__ZN4love7physics5box2d9WeldJointC2EPNS1_4BodyES4_ffffb,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22];
var FUNCTION_TABLE_vdd = [b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,_emscripten_glDepthRange__wrapper,_emscripten_glDepthRangef__wrapper,b23,b23,_emscripten_glPolygonOffset__wrapper,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23];
var FUNCTION_TABLE_vdi = [b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,_emscripten_glSampleCoverage__wrapper,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24];
var FUNCTION_TABLE_iiiid = [b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,__ZN4love7physics5box2d7Physics10newFixtureEPNS1_4BodyEPNS1_5ShapeEf,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,_mpg123_param,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25];
var FUNCTION_TABLE_iiiiiiii = [b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,_tt_face_load_sbit_image,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,__ZNKSt3__18time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcPK2tmcc,b26,b26,__ZNKSt3__18time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwPK2tmcc,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,__ZNKSt3__19money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_bRNS_8ios_baseERjRe,__ZNKSt3__19money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_bRNS_8ios_baseERjRNS_12basic_stringIcS3_NS_9allocatorIcEEEE,b26,b26,__ZNKSt3__19money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_bRNS_8ios_baseERjRe,__ZNKSt3__19money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_bRNS_8ios_baseERjRNS_12basic_stringIwS3_NS_9allocatorIwEEEE,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,_SDL_BlendFillRect_RGB888
,_SDL_BlendFillRect_ARGB8888,_SDL_BlendFillRect_RGB,_SDL_BlendFillRect_RGBA,_SDL_BlendFillRect_RGB555,_SDL_BlendFillRect_RGB565,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,__ZNSt3__114__scan_keywordINS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEPKNS_12basic_stringIcS3_NS_9allocatorIcEEEENS_5ctypeIcEEEET0_RT_SE_SD_SD_RKT1_Rjb,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,__ZNSt3__114__scan_keywordINS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEPKNS_12basic_stringIwS3_NS_9allocatorIwEEEENS_5ctypeIwEEEET0_RT_SE_SD_SD_RKT1_Rjb,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26];
var FUNCTION_TABLE_iiiddi = [b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,__ZN4love7physics5box2d7Physics7newBodyEPNS1_5WorldEffNS0_4Body4TypeE,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27
,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27,b27];
var FUNCTION_TABLE_iiiiddddi = [b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,__ZN4love7physics5box2d7Physics16newDistanceJointEPNS1_4BodyES4_ffffb
,b28,b28,b28,b28,b28,__ZN4love7physics5box2d7Physics16newFrictionJointEPNS1_4BodyES4_ffffb,__ZN4love7physics5box2d7Physics12newWeldJointEPNS1_4BodyES4_ffffb,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28
,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28,b28];
var FUNCTION_TABLE_iidddd = [b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,__ZN4love7physics5box2d7Physics12newEdgeShapeEffff,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29
,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29,b29];
var FUNCTION_TABLE_iiiiddddddi = [b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,__ZN4love7physics5box2d7Physics17newPrismaticJointEPNS1_4BodyES4_ffffffb,b30,b30,b30,b30,__ZN4love7physics5box2d7Physics13newWheelJointEPNS1_4BodyES4_ffffffb,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30
,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30,b30];
var FUNCTION_TABLE_iiiiiiiidi = [b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,__ZN4love7physics5box2d7Physics14newPulleyJointEPNS1_4BodyES4_6b2Vec2S5_S5_S5_fb,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31
,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31,b31];
var FUNCTION_TABLE_viidddd = [b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,__ZN4love8keyboard3sdl8Keyboard12setTextInputEbdddd,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32,b32
,b32,b32,b32,b32,b32,b32,b32,b32,b32];
var FUNCTION_TABLE_iiiii = [b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,__ZN4love5mouse3sdl5Mouse9newCursorEPNS_5image9ImageDataEii,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,__ZN4love6window3sdl6Window9setWindowEiiPNS0_14WindowSettingsE,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,__ZN4love10filesystem6physfs10Filesystem5mountEPKcS4_b,b33,b33,__ZNK4love10filesystem6physfs10Filesystem11newFileDataEPvjPKc,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,__ZNK4love10filesystem6physfs10Filesystem4readEPKcx,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,__ZN4love10filesystem11DroppedFile4readEPvx,__ZN4love10filesystem11DroppedFile5writeEPKvx,__ZN4love10filesystem4File5writeEPKNS_4DataEx,b33,b33,b33,b33,__ZN4love10filesystem11DroppedFile9setBufferENS0_4File10BufferModeEx,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,__ZN4love4font8freetype4Font21newTrueTypeRasterizerEPNS_4DataEiNS0_18TrueTypeRasterizer7HintingE,b33,__ZN4love4font4Font18newImageRasterizerEPNS_5image9ImageDataERKNSt3__112basic_stringIcNS5_11char_traitsIcEENS5_9allocatorIcEEEEi,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,__ZN4love8joystick3sdl14JoystickModule17setGamepadMappingERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEENS0_8Joystick12GamepadInputENSC_13JoystickInputE,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,__ZN4love10filesystem6physfs4File4readEPvx,__ZN4love10filesystem6physfs4File5writeEPKvx,b33,b33,b33,b33,__ZN4love10filesystem6physfs4File9setBufferENS0_4File10BufferModeEx,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_INT123_synth_1to1,_INT123_synth_1to1_8bit,_INT123_synth_1to1_real,_INT123_synth_1to1_s32,_INT123_synth_2to1
,_INT123_synth_2to1_8bit,_INT123_synth_2to1_real,_INT123_synth_2to1_s32,_INT123_synth_4to1,_INT123_synth_4to1_8bit,_INT123_synth_4to1_real,_INT123_synth_4to1_s32,_INT123_synth_ntom,_INT123_synth_ntom_8bit,_INT123_synth_ntom_real,_INT123_synth_ntom_s32,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_DIR_read,_DIR_write,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,_ZIP_read,_ZIP_write,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,_BDF_Glyph_Load,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_cff_glyph_load,_cff_get_kerning,b33,b33
,b33,b33,b33,b33,_cff_get_glyph_name,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_cff_get_ros,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,_cid_slot_load_glyph,b33,b33,b33,b33,_cid_get_ros,b33,b33,b33,b33,b33,b33,b33,b33,_PCF_Glyph_Load,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,_pfr_slot_load,_pfr_get_kerning,b33,_pfr_face_get_kerning,b33,b33,b33,_ps_table_add,b33,b33,b33,b33,b33,b33,b33,b33,b33,_ps_parser_to_fixed_array,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,_afm_parser_init,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_ft_raster1_render,_ft_raster1_transform,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_tt_cmap14_char_var_index,b33,b33
,b33,b33,b33,b33,b33,_tt_face_goto_table,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,_sfnt_get_glyph_name,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_gray_cubic_to,b33,_ft_smooth_render,_ft_smooth_transform,b33,b33,_ft_smooth_render_lcd,_ft_smooth_render_lcd_v,b33,b33,b33
,b33,b33,b33,b33,b33,_tt_glyph_load,_tt_get_kerning,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_T1_Load_Glyph,_Get_Kerning,b33,b33,b33
,b33,_t1_get_glyph_name,b33,b33,b33,b33,b33,b33,_T1_Get_Track_Kerning,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,_T42_GlyphSlot_Load,b33,b33,_t42_get_glyph_name,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_FNT_Load_Glyph,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_floor0_inverse2,b33,b33,b33,b33,b33,b33,_floor1_inverse2
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,__ZNKSt3__15ctypeIcE8do_widenEPKcS3_Pc,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,__ZNKSt3__15ctypeIwE5do_isEPKwS3_Pt,__ZNKSt3__15ctypeIwE10do_scan_isEtPKwS3_,__ZNKSt3__15ctypeIwE11do_scan_notEtPKwS3_,b33,b33,b33,b33,b33,__ZNKSt3__15ctypeIwE8do_widenEPKcS3_Pw,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,__ZN4love4math4Math8compressENS0_10Compressor6FormatEPNS_4DataEi,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,__ZN4love12luax_convobjEP9lua_StateiPKcS3_,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,_luaL_loadbuffer,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,__ZN4love13luax_boolflagEP9lua_StateiPKcb,b33,b33
,b33,b33,b33,b33,b33,b33,b33,__ZN4love8graphics6opengl8Graphics14newSpriteBatchEPNS0_7TextureEiNS1_4Mesh5UsageE,b33,b33,b33,b33,b33,_lua_pcall,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,__ZN4love8graphics6opengl8Graphics7newMeshERKNSt3__16vectorINS_6VertexENS3_9allocatorIS5_EEEENS1_4Mesh8DrawModeENSB_5UsageE,__ZN4love8graphics6opengl8Graphics7newMeshEiNS1_4Mesh8DrawModeENS3_5UsageE,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,__ZN4love8graphics6opengl4Mesh9getVertexEjPvj,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,__ZN4love8graphics6opengl6Shader17getExternVariableERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEERiSC_,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,__ZN4love5sound7lullabyL10vorbisSeekEPvxi,__ZN4love5sound7lullabyL10vorbisReadEPvjjS2_,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_mpg123_replace_reader_handle,b33,b33,b33,b33,_mpg123_getformat,b33,_mpg123_format,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,__ZNSt3__16vectorIN4love8graphics6opengl4Font11DrawCommandENS_9allocatorIS5_EEE6insertINS_11__wrap_iterIPS5_EEEENS_9enable_ifIXaasr21__is_forward_iteratorIT_EE5valuesr16is_constructibleIS5_NS_15iterator_traitsISE_E9referenceEEE5valueESC_E4typeENSA_IPKS5_EESE_SE_,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,_INT123_synth_1to1_8bit_wrap,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,_luaU_undump,_luaY_parser,b33,b33,b33,b33,_l_alloc,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,_writer,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_ft_ansi_stream_io,b33
,b33,_ft_realloc,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_ps_hints_apply,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_FT_CMap_New,b33,b33,b33,b33,b33,b33,_TT_Access_Glyph_Frame,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,_ft_gzip_stream_io,b33,b33,b33,_ft_lzw_stream_io,b33,b33,b33,_vorbis_book_decodevs_add,_vorbis_book_decodev_add,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_GLES2_RenderCopy
,b33,b33,b33,b33,b33,_GLES2_BindTexture,b33,b33,b33,b33,b33,_SDL_RLEBlit,_SDL_RLEAlphaBlit,b33,b33,b33,b33,b33,b33,b33,_SDL_SoftBlit,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,_SW_RenderCopy,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_Emscripten_UpdateWindowFramebuffer,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_eglGetConfigAttrib__wrapper,_eglCreateContext__wrapper,b33,_eglCreateWindowSurface__wrapper,b33,_eglMakeCurrent__wrapper,b33,b33,b33
,b33,b33,_Emscripten_OpenDevice,b33,b33,b33,b33,b33,b33,b33,_SDL_DUMMY_UpdateWindowFramebuffer,b33,b33,b33,_stdio_seek,_stdio_read,_stdio_write,b33,_DISKAUD_OpenDevice,b33,b33,b33,b33,b33,_DUMMYAUD_OpenDevice,b33,b33,b33,_SDL_UpdateWindowTexture,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_SDL_AudioOpenDevice_Default,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,__ZNSt3__125__num_get_signed_integralIlEET_PKcS3_Rji,b33,__ZNSt3__125__num_get_signed_integralIxEET_PKcS3_Rji,__ZNSt3__127__num_get_unsigned_integralItEET_PKcS3_Rji,__ZNSt3__127__num_get_unsigned_integralIjEET_PKcS3_Rji,__ZNSt3__127__num_get_unsigned_integralImEET_PKcS3_Rji,__ZNSt3__127__num_get_unsigned_integralIyEET_PKcS3_Rji,b33,b33,b33,b33,__ZNSt3__110__sscanf_lEPKcP15__locale_structS1_z,b33,b33,b33,__ZNSt3__112__asprintf_lEPPcP15__locale_structPKcz,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,_mbsrtowcs,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,_catgets,b33,b33
,b33,_mbrtowc,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33,b33
,b33,b33,b33];
var FUNCTION_TABLE_viiiiii = [b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,__ZNSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEE7seekoffExNS_8ios_base7seekdirEj,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,_DIR_enumerateFiles,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,_ZIP_enumerateFiles,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,__ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE7seekoffExNS_8ios_base7seekdirEj,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,__ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE7seekoffExNS_8ios_base7seekdirEj
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,__ZNKSt3__18messagesIcE6do_getEiiiRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEE,b34,b34,b34,b34,__ZNKSt3__18messagesIwE6do_getEiiiRKNS_12basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEEE,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,__ZNK10__cxxabiv117__class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib,b34,b34,b34,__ZNK10__cxxabiv120__si_class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,__ZNK10__cxxabiv121__vmi_class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,__ZN4love4math14CompressedDataC2ENS0_10Compressor6FormatEPcjjb,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,__ZN4love5sound9SoundData4loadEiiiiPv,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,_wuff_int8_to_int8,_wuff_int16_to_int8,_wuff_int24_to_int8,_wuff_int32_to_int8,_wuff_float32_to_int8,_wuff_float64_to_int8,_wuff_int8_to_int16,_wuff_int16_to_int16,_wuff_int24_to_int16,_wuff_int32_to_int16,_wuff_float32_to_int16,_wuff_float64_to_int16,_wuff_int8_to_int24,_wuff_int16_to_int24,_wuff_int24_to_int24,_wuff_int32_to_int24,_wuff_float32_to_int24,_wuff_float64_to_int24,_wuff_int8_to_int32,_wuff_int16_to_int32,_wuff_int24_to_int32,_wuff_int32_to_int32,_wuff_float32_to_int32
,_wuff_float64_to_int32,_wuff_int8_to_float32,_wuff_int16_to_float32,_wuff_int24_to_float32,_wuff_int32_to_float32,_wuff_float32_to_float32,_wuff_float64_to_float32,_wuff_int8_to_float64,_wuff_int16_to_float64,_wuff_int24_to_float64,_wuff_int32_to_float64,_wuff_float32_to_float64,_wuff_float64_to_float64,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,__ZN4love5sound9SoundDataC2EPviiii,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,__ZN4love8graphics6opengl8GLBufferC2EjPKvjjj,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,__ZN4love8graphics6opengl4MeshC2ERKNSt3__16vectorINS2_12AttribFormatENS3_9allocatorIS5_EEEEPKvjNS2_8DrawModeENS2_5UsageE,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,__ZN4love5image6magpie9ImageDataC2ENSt3__14listIPNS1_13FormatHandlerENS3_9allocatorIS6_EEEEiiPvb,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,__ZL22stbi__YCbCr_to_RGB_rowPhPKhS1_S1_ii,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,_oc_state_frag_recon_c,_oc_state_frag_copy_list_c,_oc_state_loop_filter_frag_rows_c,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,_Vertical_Sweep_Span,_Vertical_Sweep_Drop,b34,b34,_Horizontal_Sweep_Span,_Horizontal_Sweep_Drop,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,_emscripten_glDrawRangeElements__wrapper,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,_emscripten_glVertexAttribPointer__wrapper,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34,b34
,b34,b34,b34];
var FUNCTION_TABLE_ddd = [b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35
,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35
,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,__ZN4love4math3__68__invokeEff,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35
,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35,b35
,b35,b35,b35,b35,b35,b35,b35,b35,b35];
var FUNCTION_TABLE_viiidi = [b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,__ZN4love7physics5box2d9GearJointC2EPNS1_5JointES4_fb,b36,b36,b36
,b36,b36,__ZN4love7physics5box2d10MotorJointC2EPNS1_4BodyES4_fb,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36
,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36,b36];
var FUNCTION_TABLE_vidddd = [b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,__ZN4love7physics5box2d4Body11setMassDataEffff
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,__ZN4love7Matrix45orthoEffff,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,_emscripten_glUniform4f__wrapper,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,_emscripten_glVertexAttrib4f__wrapper,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37,b37
,b37,b37,b37];
var FUNCTION_TABLE_diid = [b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,_luaL_optnumber,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38
,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38,b38];
var FUNCTION_TABLE_iiddd = [b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,__ZN4love8joystick3sdl8Joystick12setVibrationEfff,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,__ZN4love7physics5box2d7Physics14newCircleShapeEfff,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39
,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39,b39];
var FUNCTION_TABLE_iiddi = [b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,__ZN4love7physics5box2d7Physics8newWorldEffb,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40
,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40,b40];
var FUNCTION_TABLE_viiiiiiiiii = [b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,__ZNSt3__111__money_getIcE13__gather_infoEbRKNS_6localeERNS_10money_base7patternERcS8_RNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEESF_SF_SF_Ri,b41,b41,b41,b41,__ZNSt3__111__money_getIwE13__gather_infoEbRKNS_6localeERNS_10money_base7patternERwS8_RNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERNS9_IwNSA_IwEENSC_IwEEEESJ_SJ_Ri,b41,b41,b41,b41,__ZNSt3__111__money_putIcE13__gather_infoEbbRKNS_6localeERNS_10money_base7patternERcS8_RNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEESF_SF_Ri,b41,__ZNSt3__111__money_putIwE13__gather_infoEbbRKNS_6localeERNS_10money_base7patternERwS8_RNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERNS9_IwNSA_IwEENSC_IwEEEESJ_Ri,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41,b41
,b41,b41,b41];
var FUNCTION_TABLE_diii = [b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,__ZNK4love4font8freetype18TrueTypeRasterizer10getKerningEjj,b42,b42,b42,b42,b42,b42,__ZNK4love4font16BMFontRasterizer10getKerningEjj,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,__ZNK4love4font10Rasterizer10getKerningEjj,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,__ZN4love8graphics6opengl4Font10getKerningEjj,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,__ZNSt3__115__num_get_floatIfEET_PKcS3_Rj,__ZNSt3__115__num_get_floatIdEET_PKcS3_Rj,__ZNSt3__115__num_get_floatIeEET_PKcS3_Rj,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42,b42
,b42,b42,b42];
var FUNCTION_TABLE_iiididdddddddd = [b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,__ZN4love8graphics6opengl4Text4addfERKNSt3__16vectorINS1_4Font13ColoredStringENS3_9allocatorIS6_EEEEfNS5_9AlignModeEfffffffff,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43
,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43,b43];
var FUNCTION_TABLE_viiiddi = [b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,__ZN4love7physics5box2d13RevoluteJointC2EPNS1_4BodyES4_ffb,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44
,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44,b44];
var FUNCTION_TABLE_iiiiii = [b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,__ZN4love4math13LZ4Compressor10decompressENS0_10Compressor6FormatEPKcjRj,b45,b45,b45,__ZN4love4math14zlibCompressor10decompressENS0_10Compressor6FormatEPKcjRj,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,__ZN4love6window3sdl6Window14showMessageBoxERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEESB_NS0_6Window14MessageBoxTypeEb,b45,b45,b45,b45,b45,b45,b45,__ZNK12b2ChainShape7RayCastEP15b2RayCastOutputRK14b2RayCastInputRK11b2Transformi,b45,b45,b45,b45,b45,b45,b45,__ZNK13b2CircleShape7RayCastEP15b2RayCastOutputRK14b2RayCastInputRK11b2Transformi,b45,b45,b45,b45,b45,b45,__ZNK11b2EdgeShape7RayCastEP15b2RayCastOutputRK14b2RayCastInputRK11b2Transformi,b45,b45,b45,b45,b45,b45
,__ZNK14b2PolygonShape7RayCastEP15b2RayCastOutputRK14b2RayCastInputRK11b2Transformi,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,__ZN4love4font4Font18newImageRasterizerEPNS_5image9ImageDataEPjii,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,__ZN4love5image6magpie5Image12newImageDataEiiPvb,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_raccess_guess_apple_double,_raccess_guess_apple_single,_raccess_guess_darwin_ufs_export,_raccess_guess_darwin_newvfs
,_raccess_guess_darwin_hfsplus,_raccess_guess_vfat,_raccess_guess_linux_cap,_raccess_guess_linux_double,_raccess_guess_linux_netatalk,b45,_BDF_Face_Init,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_cff_face_init,b45,b45,b45,b45,b45,b45,b45,_cff_get_advances,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_cid_face_init,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_PCF_Face_Init,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,_pfr_face_init,b45,b45,b45,b45,b45,_pfr_get_metrics,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_ps_parser_to_bytes,b45,b45,b45,b45,_ps_parser_load_field,_ps_parser_load_field_table,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,_sfnt_init_face,_sfnt_load_face,b45,_tt_face_load_any,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,_sfnt_table_info,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,_tt_face_init,b45,b45,b45,b45,b45,b45,_tt_get_advances,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_T1_Face_Init,b45,b45,b45,b45,b45,b45,b45,b45,_T1_Get_Advances,b45
,b45,b45,b45,b45,b45,b45,b45,_t1_ps_get_font_value,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_T42_Face_Init,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_FNT_Face_Init,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_af_autofitter_load_glyph,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,_res0_inverse,b45,_res1_class,b45,_res1_inverse,_res2_class,b45,_res2_inverse,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,__ZNKSt3__17collateIcE10do_compareEPKcS3_S3_S3_,b45,b45,b45,b45,__ZNKSt3__17collateIwE10do_compareEPKwS3_S3_S3_,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,__ZNKSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcb,__ZNKSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcl,b45,__ZNKSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcm,b45,b45,b45,__ZNKSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcPKv,b45,b45,__ZNKSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwb
,__ZNKSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwl,b45,__ZNKSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwm,b45,b45,b45,__ZNKSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwPKv,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,__ZNKSt3__17codecvtIDic11__mbstate_tE10do_unshiftERS1_PcS4_RS4_
,b45,b45,__ZNKSt3__17codecvtIDic11__mbstate_tE9do_lengthERS1_PKcS5_j,b45,b45,b45,b45,b45,b45,__ZNKSt3__17codecvtIwc11__mbstate_tE10do_unshiftERS1_PcS4_RS4_,b45,b45,__ZNKSt3__17codecvtIwc11__mbstate_tE9do_lengthERS1_PKcS5_j,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,__ZNKSt3__15ctypeIcE9do_narrowEPKcS3_cPc,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,__ZNKSt3__15ctypeIwE9do_narrowEPKwS3_cPc,b45,b45,b45,__ZNKSt3__17codecvtIcc11__mbstate_tE10do_unshiftERS1_PcS4_RS4_,b45,b45
,__ZNKSt3__17codecvtIcc11__mbstate_tE9do_lengthERS1_PKcS5_j,b45,b45,b45,b45,__ZNKSt3__17codecvtIDsc11__mbstate_tE10do_unshiftERS1_PcS4_RS4_,b45,b45,__ZNKSt3__17codecvtIDsc11__mbstate_tE9do_lengthERS1_PKcS5_j,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,__ZN4love4math4Math8compressENS0_10Compressor6FormatEPKcji,b45,b45,__ZN4love4math4Math10decompressENS0_10Compressor6FormatEPKcjRj,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,__ZN4love5sound5Sound12newSoundDataEiiii,b45,b45
,b45,b45,b45,b45,b45,b45,__ZN4love12luax_convobjEP9lua_StatePiiPKcS4_,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,_enet_range_coder_decompress,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7replaceEjjPKcj,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,__ZN4love8graphics6opengl8Graphics9newCanvasEiiNS1_6Canvas6FormatEi,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,__ZN4love8graphics6opengl8Graphics7newMeshERKNSt3__16vectorINS1_4Mesh12AttribFormatENS3_9allocatorIS6_EEEEiNS5_8DrawModeENS5_5UsageE,b45,b45,__ZN4love8graphics6opengl23luax_writeAttributeDataEP9lua_StateiNS1_4Mesh8DataTypeEiPc,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,__ZN4love8graphics6opengl5Image7refreshEiiii,b45,b45,b45,b45,__ZN4love8graphics6opengl4Mesh18getVertexAttributeEjiPvj,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_ov_open_callbacks,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,__ZN15b2CircleContact6CreateEP9b2FixtureiS1_iP16b2BlockAllocator,b45,__ZN25b2PolygonAndCircleContact6CreateEP9b2FixtureiS1_iP16b2BlockAllocator,b45,__ZN16b2PolygonContact6CreateEP9b2FixtureiS1_iP16b2BlockAllocator,b45,__ZN22b2EdgeAndCircleContact6CreateEP9b2FixtureiS1_iP16b2BlockAllocator
,b45,__ZN23b2EdgeAndPolygonContact6CreateEP9b2FixtureiS1_iP16b2BlockAllocator,b45,__ZN23b2ChainAndCircleContact6CreateEP9b2FixtureiS1_iP16b2BlockAllocator,b45,__ZN24b2ChainAndPolygonContact6CreateEP9b2FixtureiS1_iP16b2BlockAllocator,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_socket_send,_socket_recv,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_FT_New_Memory_Face,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,__ZNSt3__14listIPN4love5image6magpie13FormatHandlerENS_9allocatorIS5_EEE6insertIPKS5_EENS_15__list_iteratorIS5_PvEENS_21__list_const_iteratorIS5_SD_EET_SH_PNS_9enable_ifIXsr19__is_input_iteratorISH_EE5valueEvE4typeE,__ZNSt3__14listIPN4love5image6magpie23CompressedFormatHandlerENS_9allocatorIS5_EEE6insertIPKS5_EENS_15__list_iteratorIS5_PvEENS_21__list_const_iteratorIS5_SD_EET_SH_PNS_9enable_ifIXsr19__is_input_iteratorISH_EE5valueEvE4typeE,b45,b45
,__Z15lodepng_inspectPjS_P12LodePNGStatePKhj,__ZN4love5image6magpieL14zlibDecompressEPPhPjPKhjPK25LodePNGDecompressSettings,b45,b45,__ZN4love5image6magpieL12zlibCompressEPPhPjPKhjPK23LodePNGCompressSettings,b45,b45,b45,__ZL23stbi__resample_row_hv_2PhS_S_ii,__ZL14resample_row_1PhS_S_ii,__ZL22stbi__resample_row_v_2PhS_S_ii,__ZL22stbi__resample_row_h_2PhS_S_ii,__ZL26stbi__resample_row_genericPhS_S_ii,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,__Z16X86_Convert32To8PvPijS0_S0_,__Z17X86_Convert32To16PvPijS0_S0_,__Z17X86_Convert32To24PvPijS0_S0_,__Z17X86_Convert32To32PvPijS0_S0_,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,__bdf_parse_start,__bdf_parse_properties,__bdf_parse_glyphs,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_tt_sbit_decoder_load_byte_aligned,_tt_sbit_decoder_load_bit_aligned,_tt_sbit_decoder_load_compound,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_GLES2_UpdateTexture,b45,_GLES2_LockTexture,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,_GLES2_RenderReadPixels,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_copy_transl_565,_copy_transl_555,_copy_opaque_16,_copy_32,_uncopy_opaque_16,_uncopy_32,_uncopy_transl_16,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_SW_UpdateTexture,_SW_LockTexture,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,_SW_RenderReadPixels,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_Emscripten_CreateWindowFramebuffer,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_eglChooseConfig__wrapper,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,_SDL_DUMMY_CreateWindowFramebuffer,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_SDL_CreateWindowTexture,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,__ZNKSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7compareEjjPKcj,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,_wcsnrtombs,b45
,_mbsnrtowcs,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45,b45
,b45,b45,b45];
var FUNCTION_TABLE_viiid = [b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,__ZN4love7physics5box2d7FixtureC2EPNS1_4BodyEPNS1_5ShapeEf,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46
,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46,b46];
var FUNCTION_TABLE_viidii = [b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,__ZN4love8graphics6opengl4Font7getWrapERKNS2_17ColoredCodepointsEfRNSt3__16vectorIS3_NS6_9allocatorIS3_EEEEPNS7_IiNS8_IiEEEE,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,__ZN4love8graphics6opengl4Font7getWrapERKNSt3__16vectorINS2_13ColoredStringENS3_9allocatorIS5_EEEEfRNS4_INS3_12basic_stringIcNS3_11char_traitsIcEENS6_IcEEEENS6_ISF_EEEEPNS4_IiNS6_IiEEEE,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47,b47
,b47,b47,b47];
var FUNCTION_TABLE_iiiiid = [b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,__ZNKSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcd,__ZNKSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEce,b48,b48,b48,b48
,b48,b48,b48,b48,__ZNKSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwd,__ZNKSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwe,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48
,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48,b48];
var FUNCTION_TABLE_viiii = [b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZNSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEE7seekposENS_4fposI11__mbstate_tEEj,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZN4love6window3sdl6Window9getWindowERiS3_RNS0_14WindowSettingsE,b49,b49,b49,b49,b49,b49
,b49,__ZNK4love6window3sdl6Window20getDesktopDimensionsEiRiS3_,__ZN4love6window3sdl6Window11setPositionEiii,__ZN4love6window3sdl6Window11getPositionERiS3_S3_,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZNK12b2ChainShape11ComputeAABBEP6b2AABBRK11b2Transformi,b49,b49,b49,b49,b49,b49,b49,__ZNK13b2CircleShape11ComputeAABBEP6b2AABBRK11b2Transformi,b49,b49,b49,b49,b49,b49,__ZNK11b2EdgeShape11ComputeAABBEP6b2AABBRK11b2Transformi,b49,b49,b49,b49,b49
,b49,__ZNK14b2PolygonShape11ComputeAABBEP6b2AABBRK11b2Transformi,b49,b49,b49,b49,b49,b49,__ZN22b2EdgeAndCircleContact8EvaluateEP10b2ManifoldRK11b2TransformS4_,b49,__ZN23b2EdgeAndPolygonContact8EvaluateEP10b2ManifoldRK11b2TransformS4_,b49,__ZN25b2PolygonAndCircleContact8EvaluateEP10b2ManifoldRK11b2TransformS4_,b49,__ZN16b2PolygonContact8EvaluateEP10b2ManifoldRK11b2TransformS4_,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZNK4love5audio4null6Source7getConeERfS3_S3_,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZNK4love5audio6openal6Source7getConeERfS3_S3_,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZN4love5image6magpie10PNGHandler6encodeERKNS1_13FormatHandler12DecodedImageENS0_9ImageData13EncodedFormatE,b49,b49
,b49,b49,b49,b49,b49,b49,__ZN4love5image6magpie10STBHandler6encodeERKNS1_13FormatHandler12DecodedImageENS0_9ImageData13EncodedFormatE,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZN4love8joystick3sdl14JoystickModule17getGamepadMappingERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEENS0_8Joystick12GamepadInputE,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZN23b2ChainAndCircleContact8EvaluateEP10b2ManifoldRK11b2TransformS4_,b49,__ZN24b2ChainAndPolygonContact8EvaluateEP10b2ManifoldRK11b2TransformS4_,b49
,__ZN15b2CircleContact8EvaluateEP10b2ManifoldRK11b2TransformS4_,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,__ZN4love5image6magpie13FormatHandler6encodeERKNS2_12DecodedImageENS0_9ImageData13EncodedFormatE,b49,b49,b49,b49,b49,b49,_convert_latin1,_convert_utf16bom,_convert_utf8,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,_oc_mb_fill_cmapping00,_oc_mb_fill_cmapping01,_oc_mb_fill_cmapping10,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,_ps_parser_init,b49,b49,b49,b49,b49,b49,b49,b49,b49,_ps_parser_to_token_array,b49,b49,b49,b49,b49
,_t1_builder_add_point,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,_SDL_LogOutput,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE7seekposENS_4fposI11__mbstate_tEEj,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,__ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE7seekposENS_4fposI11__mbstate_tEEj,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZNKSt3__17collateIcE12do_transformEPKcS3_,b49,b49,b49,b49,__ZNKSt3__17collateIwE12do_transformEPKwS3_,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZNK10__cxxabiv117__class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi,b49,b49,b49,__ZNK10__cxxabiv120__si_class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZNK10__cxxabiv121__vmi_class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcjj,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZN4love8graphics6opengl4Mesh9setVertexEjPKvj,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZNK4love5image9ImageData8getPixelEii,__ZN4love5image9ImageData8setPixelEiiNS0_5pixelE,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZN4love5mouse3sdl6CursorC2EPNS_5image9ImageDataEii,b49,b49,__ZN4love7physics5box2d4BodyC2EPNS1_5WorldE6b2Vec2NS0_4Body4TypeE,b49,b49,b49,b49,b49,b49,__ZN4love7physics5box2d10ChainShapeC2EP12b2ChainShapebb,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZN18b2RevoluteJointDef10InitializeEP6b2BodyS1_RK6b2Vec2,b49,b49,b49,b49,__ZN14b2WeldJointDef10InitializeEP6b2BodyS1_RK6b2Vec2,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,__ZN4love5sound7lullaby14ModPlugDecoderC2EPNS_4DataERKNSt3__112basic_stringIcNS5_11char_traitsIcEENS5_9allocatorIcEEEEi,b49,__ZN4love5sound7lullaby13Mpg123DecoderC2EPNS_4DataERKNSt3__112basic_stringIcNS5_11char_traitsIcEENS5_9allocatorIcEEEEi,b49,__ZN4love5sound7lullaby13VorbisDecoderC2EPNS_4DataERKNSt3__112basic_stringIcNS5_11char_traitsIcEENS5_9allocatorIcEEEEi,b49,__ZN4love5sound7lullaby11WaveDecoderC2EPNS_4DataERKNSt3__112basic_stringIcNS5_11char_traitsIcEENS5_9allocatorIcEEEEi,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZN8b2Island8SolveTOIERK10b2TimeStepii,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZN4love10filesystem8FileDataC2EyRKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE,b49,b49,b49,b49,b49,b49,b49,b49,__ZN4love4font9GlyphDataC2EjNS0_12GlyphMetricsENS1_6FormatE,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZN4love8graphics6opengl6OpenGL10drawArraysEjii,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZN4love8graphics6opengl4Font6printvERKNS_7Matrix4ERKNSt3__16vectorINS2_11DrawCommandENS6_9allocatorIS8_EEEERKNS7_INS2_11GlyphVertexENS9_ISE_EEEE,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZN4love8graphics6opengl11SpriteBatchC2EPNS0_7TextureEiNS1_4Mesh5UsageE,b49,b49,b49,__ZN4love8graphics6opengl4MeshC2ERKNSt3__16vectorINS_6VertexENS3_9allocatorIS5_EEEENS2_8DrawModeENS2_5UsageE,__ZN4love8graphics6opengl4MeshC2EiNS2_8DrawModeENS2_5UsageE,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,__ZN4love8graphics6opengl6OpenGL17bindTextureToUnitEjib,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZN4love8graphics6opengl11SpriteBatch4addvEPKNS_6VertexERKNS_7Matrix3Ei,__ZN4love8graphics6opengl8GLBuffer4fillEjjPKv
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZN4love5image6magpie9ImageDataC2ENSt3__14listIPNS1_13FormatHandlerENS3_9allocatorIS6_EEEEii,b49,b49,b49,b49,b49,b49,__ZN4love5image6magpie9ImageData6createEiiPv
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,__ZN18b2FrictionJointDef10InitializeEP6b2BodyS1_RK6b2Vec2,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,_oc_frag_recon_inter_c,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,_t2_hints_stems
,_ps_hints_t2mask,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,_ft_validator_init,b49,b49,b49,b49,_gray_render_span,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,_Direct_Move,_Direct_Move_Orig,_Direct_Move_X,_Direct_Move_Orig_X,_Direct_Move_Y,_Direct_Move_Orig_Y,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,_Emscripten_SetWindowFullscreen,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,_Emscripten_GLES_GetDrawableSize,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,_SDL_JoystickAllowBackgroundEventsChanged,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,_emscripten_glBufferData__wrapper,_emscripten_glBufferSubData__wrapper,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,_emscripten_glUniform3i__wrapper,b49,b49,b49,b49,b49,b49,b49,b49,b49,_emscripten_glUniformMatrix2fv__wrapper,_emscripten_glUniformMatrix3fv__wrapper,_emscripten_glUniformMatrix4fv__wrapper,b49,b49,b49
,b49,b49,b49,b49,b49,b49,_emscripten_glGetAttachedShaders__wrapper,_emscripten_glShaderSource__wrapper,_emscripten_glGetShaderSource__wrapper,b49,_emscripten_glGetShaderInfoLog__wrapper,b49,b49,b49,b49,b49,b49,b49,_emscripten_glGetShaderPrecisionFormat__wrapper,b49,_emscripten_glGetProgramInfoLog__wrapper,b49,b49,b49,b49,b49,b49,b49,_emscripten_glFramebufferRenderbuffer__wrapper,b49
,_emscripten_glGetFramebufferAttachmentParameteriv__wrapper,b49,b49,b49,_emscripten_glGetInfoLogARB__wrapper,b49,b49,b49,b49,_emscripten_glVertexPointer__wrapper,_emscripten_glTexCoordPointer__wrapper,b49,_emscripten_glColorPointer__wrapper,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,_emscripten_glDrawElements__wrapper,b49,b49,b49
,b49,_emscripten_glDrawArraysInstanced__wrapper,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,_emscripten_glViewport__wrapper,b49,_emscripten_glScissor__wrapper,b49,_emscripten_glColorMask__wrapper,_emscripten_glRenderbufferStorage__wrapper,_emscripten_glBlendFuncSeparate__wrapper,b49,_emscripten_glStencilFuncSeparate__wrapper,_emscripten_glStencilOpSeparate__wrapper,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49,b49
,b49,b49,b49];
var FUNCTION_TABLE_viiiii = [b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,__ZNK4love10filesystem6physfs10Filesystem5writeEPKcPKvx,__ZNK4love10filesystem6physfs10Filesystem6appendEPKcPKvx,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,_t1_builder_init,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,_tt_face_get_metrics,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,__ZNK10__cxxabiv117__class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib,b50,b50,b50,__ZNK10__cxxabiv120__si_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,__ZNK10__cxxabiv121__vmi_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_jjRKS4_,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,__ZN4love8graphics6opengl4Mesh18setVertexAttributeEjiPKvj,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,__ZN4love8graphics6opengl6Shader7sendIntERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEEiPKii,__ZN4love8graphics6opengl6Shader9sendFloatERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEEiPKfi,__ZN4love8graphics6opengl6Shader10sendMatrixERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEEiPKfi,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,__ZN19b2PrismaticJointDef10InitializeEP6b2BodyS1_RK6b2Vec2S4_,b50,b50,b50,b50,b50,b50,b50,b50,__ZN15b2WheelJointDef10InitializeEP6b2BodyS1_RK6b2Vec2S4_,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,__ZN4love5sound9SoundDataC2Eiiii,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,__ZN8b2Island5SolveEP9b2ProfileRK10b2TimeStepRK6b2Vec2b,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,_alBufferData__wrapper,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,__ZN4love4font8freetype18TrueTypeRasterizerC2EP14FT_LibraryRec_PNS_4DataEiNS0_18TrueTypeRasterizer7HintingE,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,__ZN4love4font15ImageRasterizerC2EPNS_5image9ImageDataEPjii,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,__ZN4love8graphics6opengl6CanvasC2EiiNS2_6FormatEi,b50,b50,b50,__ZN4love8graphics6opengl4MeshC2ERKNSt3__16vectorINS2_12AttribFormatENS3_9allocatorIS5_EEEEiNS2_8DrawModeENS2_5UsageE,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,__ZN4love8graphics6opengl6OpenGL12drawElementsEjijPKv,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,__ZNK4love8graphics6opengl6Shader20checkSetUniformErrorERKNS2_7UniformEiiNS2_11UniformTypeE,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,__ZN18b2DistanceJointDef10InitializeEP6b2BodyS1_RK6b2Vec2S4_,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,_oc_frag_recon_inter2_c,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,_psh_globals_set_scale,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,_emscripten_glUniform4i__wrapper,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,_emscripten_glFramebufferTexture2D__wrapper,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,_emscripten_glShaderBinary__wrapper,b50,b50
,b50,b50,_emscripten_glDrawElementsInstanced__wrapper,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50,b50
,b50,b50,b50];
var FUNCTION_TABLE_vid = [b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,__ZN4love5mouse3sdl5Mouse4setXEd,__ZN4love5mouse3sdl5Mouse4setYEd,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,__ZNK4love6system6System7vibrateEd,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,__ZNK4love5timer3sdl5Timer5sleepEd,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,__ZN4love5video11VideoStream4seekEd,b51,b51,b51,b51,b51,b51,b51,__ZN4love5video11VideoStream9DeltaSync6updateEd
,b51,b51,__ZN4love5video11VideoStream9DeltaSync4seekEd,b51,b51,b51,b51,b51,__ZN4love5video11VideoStream9FrameSync6updateEd,b51,b51,__ZN4love5video11VideoStream10SourceSync4seekEd,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,__ZN4love5audio4null5Audio9setVolumeEf,b51,b51,b51,b51,b51,b51,b51,__ZN4love5audio4null5Audio15setDopplerScaleEf,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,__ZN4love5audio4null6Source8setPitchEf,b51,__ZN4love5audio4null6Source9setVolumeEf,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,__ZN4love5audio4null6Source12setMinVolumeEf
,b51,__ZN4love5audio4null6Source12setMaxVolumeEf,b51,__ZN4love5audio4null6Source20setReferenceDistanceEf,b51,__ZN4love5audio4null6Source16setRolloffFactorEf,b51,__ZN4love5audio4null6Source14setMaxDistanceEf,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,__ZN4love5audio6openal5Audio9setVolumeEf,b51,b51,b51,b51,b51,b51,b51,__ZN4love5audio6openal5Audio15setDopplerScaleEf,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,__ZN4love5audio6openal6Source8setPitchEf,b51,__ZN4love5audio6openal6Source9setVolumeEf,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,__ZN4love5audio6openal6Source12setMinVolumeEf,b51,__ZN4love5audio6openal6Source12setMaxVolumeEf,b51,__ZN4love5audio6openal6Source20setReferenceDistanceEf,b51,__ZN4love5audio6openal6Source16setRolloffFactorEf,b51,__ZN4love5audio6openal6Source14setMaxDistanceEf,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,_lua_pushnumber,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,__ZN4love7physics5box2d14PrismaticJoint13setUpperLimitEf,__ZN4love7physics5box2d14PrismaticJoint13setLowerLimitEf,b51,__ZN4love7physics5box2d13RevoluteJoint13setUpperLimitEf,__ZN4love7physics5box2d13RevoluteJoint13setLowerLimitEf,b51,b51,__ZN4love7physics5box2d5World6updateEf,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,__ZN4love8graphics14ParticleSystem15setEmissionRateEf,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,__ZN4love7physics5box2d4Body4setXEf,__ZN4love7physics5box2d4Body4setYEf,__ZN4love7physics5box2d4Body8setAngleEf,b51,b51,b51
,__ZN4love7physics5box2d4Body7setMassEf,__ZN4love7physics5box2d4Body10setInertiaEf,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,__ZN4love7physics5box2d7Fixture10setDensityEf,b51,b51,__ZN4love7physics5box2d13FrictionJoint11setMaxForceEf,__ZN4love7physics5box2d13FrictionJoint12setMaxTorqueEf,__ZN4love7physics5box2d9GearJoint8setRatioEf,b51,b51,b51,b51,b51,__ZN4love7physics5box2d10MotorJoint11setMaxForceEf,__ZN4love7physics5box2d10MotorJoint12setMaxTorqueEf,__ZN4love7physics5box2d10MotorJoint19setCorrectionFactorEf,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,__ZN4love5video6theora11VideoStream22threadedFillBackBufferEd,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,__ZN4love7VariantC2Ed,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,_emscripten_glUniform1f__wrapper,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,_emscripten_glVertexAttrib1f__wrapper,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51,b51
,b51,b51,b51];
var FUNCTION_TABLE_iiiidi = [b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,__ZN4love7physics5box2d7Physics12newGearJointEPNS1_5JointES4_fb,b52,b52,b52,b52,__ZN4love7physics5box2d7Physics13newMotorJointEPNS1_4BodyES4_fb,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52
,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52,b52];
var FUNCTION_TABLE_iiidd = [b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,__ZN4love7physics5box2d7Physics13newMouseJointEPNS1_4BodyEff,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53
,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53,b53];
var FUNCTION_TABLE_vii = [b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE5imbueERKNS_6localeE,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love8graphics14ParticleSystem13setBufferSizeEj,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love5mouse3sdl5Mouse9setCursorEPNS0_6CursorE,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love5mouse3sdl5Mouse10setVisibleEb
,b54,b54,__ZN4love5mouse3sdl5Mouse10setGrabbedEb,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,__ZN4love7physics5box2d5World12BeginContactEP9b2Contact,__ZN4love7physics5box2d5World10EndContactEP9b2Contact,b54,b54,b54,__ZN4love7physics5box2d5World10SayGoodbyeEP9b2Fixture,__ZN4love7physics5box2d5World10SayGoodbyeEP7b2Joint,b54,b54,__ZThn8_N4love7physics5box2d5World12BeginContactEP9b2Contact,__ZThn8_N4love7physics5box2d5World10EndContactEP9b2Contact,b54,b54,b54,b54,b54,b54,b54,__ZThn16_N4love7physics5box2d5World10SayGoodbyeEP7b2Joint,__ZThn16_N4love7physics5box2d5World10SayGoodbyeEP9b2Fixture,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZNK4love6system3sdl6System16setClipboardTextERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE,__ZNK4love6system3sdl6System16getClipboardTextEv,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love5video6theora11VideoStream7setSyncEPNS0_11VideoStream9FrameSyncE,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love5video11VideoStream7setSyncEPNS1_9FrameSyncE,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,__ZN4love6window3sdl6Window14setWindowTitleERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE,b54,b54,b54,__ZN4love6window3sdl6Window22setDisplaySleepEnabledEb,b54,b54,b54,b54,b54,b54,b54,__ZN4love6window3sdl6Window15setMouseVisibleEb,b54,__ZN4love6window3sdl6Window12setMouseGrabEb,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,__ZN4love6window3sdl6Window16requestAttentionEb,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN7b2Joint11ShiftOriginERK6b2Vec2,b54,b54,__ZNK12b2MotorJoint10GetAnchorAEv,__ZNK12b2MotorJoint10GetAnchorBEv,b54,b54,b54,b54,__ZN12b2MotorJoint23InitVelocityConstraintsERK12b2SolverData,__ZN12b2MotorJoint24SolveVelocityConstraintsERK12b2SolverData,b54,__ZNK12b2MouseJoint10GetAnchorAEv
,__ZNK12b2MouseJoint10GetAnchorBEv,b54,b54,b54,__ZN12b2MouseJoint11ShiftOriginERK6b2Vec2,b54,__ZN12b2MouseJoint23InitVelocityConstraintsERK12b2SolverData,__ZN12b2MouseJoint24SolveVelocityConstraintsERK12b2SolverData,b54,__ZNK16b2PrismaticJoint10GetAnchorAEv,__ZNK16b2PrismaticJoint10GetAnchorBEv,b54,b54,b54,b54,__ZN16b2PrismaticJoint23InitVelocityConstraintsERK12b2SolverData,__ZN16b2PrismaticJoint24SolveVelocityConstraintsERK12b2SolverData,b54,__ZNK13b2PulleyJoint10GetAnchorAEv,__ZNK13b2PulleyJoint10GetAnchorBEv,b54,b54,b54,__ZN13b2PulleyJoint11ShiftOriginERK6b2Vec2,b54,__ZN13b2PulleyJoint23InitVelocityConstraintsERK12b2SolverData,__ZN13b2PulleyJoint24SolveVelocityConstraintsERK12b2SolverData,b54,__ZNK15b2RevoluteJoint10GetAnchorAEv,__ZNK15b2RevoluteJoint10GetAnchorBEv
,b54,b54,b54,b54,__ZN15b2RevoluteJoint23InitVelocityConstraintsERK12b2SolverData,__ZN15b2RevoluteJoint24SolveVelocityConstraintsERK12b2SolverData,b54,__ZNK11b2RopeJoint10GetAnchorAEv,__ZNK11b2RopeJoint10GetAnchorBEv,b54,b54,b54,b54,__ZN11b2RopeJoint23InitVelocityConstraintsERK12b2SolverData,__ZN11b2RopeJoint24SolveVelocityConstraintsERK12b2SolverData,b54,__ZNK11b2WeldJoint10GetAnchorAEv,__ZNK11b2WeldJoint10GetAnchorBEv,b54,b54,b54,b54,__ZN11b2WeldJoint23InitVelocityConstraintsERK12b2SolverData,__ZN11b2WeldJoint24SolveVelocityConstraintsERK12b2SolverData,b54,__ZNK12b2WheelJoint10GetAnchorAEv,__ZNK12b2WheelJoint10GetAnchorBEv,b54,b54,b54
,b54,__ZN12b2WheelJoint23InitVelocityConstraintsERK12b2SolverData,__ZN12b2WheelJoint24SolveVelocityConstraintsERK12b2SolverData,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,__ZN4love5audio4null5Audio4stopEPNS0_6SourceE,b54,__ZN4love5audio4null5Audio5pauseEPNS0_6SourceE,b54,__ZN4love5audio4null5Audio6resumeEPNS0_6SourceE,b54,__ZN4love5audio4null5Audio6rewindEPNS0_6SourceE,b54,b54,b54,__ZNK4love5audio4null5Audio11getPositionEPf,__ZN4love5audio4null5Audio11setPositionEPf,__ZNK4love5audio4null5Audio14getOrientationEPf,__ZN4love5audio4null5Audio14setOrientationEPf,__ZNK4love5audio4null5Audio11getVelocityEPf,__ZN4love5audio4null5Audio11setVelocityEPf,b54,b54,b54,b54,b54,b54,b54,__ZN4love5audio4null5Audio16setDistanceModelENS0_5Audio13DistanceModelE,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love5audio4null6Source11setPositionEPf,__ZNK4love5audio4null6Source11getPositionEPf,__ZN4love5audio4null6Source11setVelocityEPf,__ZNK4love5audio4null6Source11getVelocityEPf,__ZN4love5audio4null6Source12setDirectionEPf,__ZNK4love5audio4null6Source12getDirectionEPf,b54,b54,__ZN4love5audio4null6Source11setRelativeEb,b54,__ZN4love5audio4null6Source10setLoopingEb,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love5audio6openal5Audio4stopEPNS0_6SourceE,b54,__ZN4love5audio6openal5Audio5pauseEPNS0_6SourceE,b54,__ZN4love5audio6openal5Audio6resumeEPNS0_6SourceE,b54,__ZN4love5audio6openal5Audio6rewindEPNS0_6SourceE
,b54,b54,b54,__ZNK4love5audio6openal5Audio11getPositionEPf,__ZN4love5audio6openal5Audio11setPositionEPf,__ZNK4love5audio6openal5Audio14getOrientationEPf,__ZN4love5audio6openal5Audio14setOrientationEPf,__ZNK4love5audio6openal5Audio11getVelocityEPf,__ZN4love5audio6openal5Audio11setVelocityEPf,b54,b54,b54,b54,b54,b54,b54,__ZN4love5audio6openal5Audio16setDistanceModelENS0_5Audio13DistanceModelE,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love5audio6openal6Source11setPositionEPf,__ZNK4love5audio6openal6Source11getPositionEPf,__ZN4love5audio6openal6Source11setVelocityEPf,__ZNK4love5audio6openal6Source11getVelocityEPf,__ZN4love5audio6openal6Source12setDirectionEPf,__ZNK4love5audio6openal6Source12getDirectionEPf,b54,b54,__ZN4love5audio6openal6Source11setRelativeEb,b54,__ZN4love5audio6openal6Source10setLoopingEb,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love10filesystem6physfs10Filesystem4initEPKc,__ZN4love10filesystem6physfs10Filesystem8setFusedEb,b54,b54,__ZN4love10filesystem10Filesystem22setAndroidSaveExternalEb,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love10filesystem6physfs10Filesystem16getUserDirectoryEv,__ZN4love10filesystem6physfs10Filesystem19getAppdataDirectoryEv,b54,__ZNK4love10filesystem6physfs10Filesystem22getSourceBaseDirectoryEv,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love10filesystem6physfs10Filesystem18setSymlinksEnabledEb,b54,b54
,__ZN4love10filesystem6physfs10Filesystem20allowMountingForPathERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE,b54,__ZNK4love10filesystem10Filesystem17getExecutablePathEv,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZNK4love10filesystem4File12getExtensionEv,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love8graphics6opengl6Canvas9setFilterERKNS0_7Texture6FilterE,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love8graphics6opengl8Graphics9setActiveEb,b54,b54,b54,b54,b54,__ZN4love8graphics6opengl5Image9setFilterERKNS0_7Texture6FilterE,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,__ZN4love8graphics6opengl14ParticleSystem13setBufferSizeEj,b54,b54,__ZN4love8graphics6opengl8Polyline26calc_overdraw_vertex_countEb,b54,__ZN4love8graphics6opengl8Polyline16fill_color_arrayEPNS0_6ColorTIhEE,b54,__ZN4love8graphics6opengl16NoneJoinPolyline26calc_overdraw_vertex_countEb,b54,__ZN4love8graphics6opengl16NoneJoinPolyline16fill_color_arrayEPNS0_6ColorTIhEE,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love5image6magpie10PNGHandler4freeEPh,b54
,b54,b54,b54,b54,b54,b54,b54,__ZN4love5image6magpie10STBHandler4freeEPh,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZNK4love8joystick3sdl8Joystick7getAxesEv,b54,b54,b54,b54,b54,b54,b54,__ZNK4love8joystick3sdl8Joystick7getGUIDEv,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love8joystick3sdl14JoystickModule14removeJoystickEPNS0_8JoystickE,b54,b54,b54,b54,b54,b54,__ZN4love8joystick3sdl14JoystickModule19loadGamepadMappingsERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE,__ZN4love8joystick3sdl14JoystickModule19saveGamepadMappingsEv,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love8keyboard3sdl8Keyboard12setKeyRepeatEb,b54,b54,b54,b54,b54,__ZN4love8keyboard3sdl8Keyboard12setTextInputEb,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN17b2ContactListener12BeginContactEP9b2Contact,__ZN17b2ContactListener10EndContactEP9b2Contact,b54,b54,b54,b54,b54,b54
,b54,b54,__ZNK15b2DistanceJoint10GetAnchorAEv,__ZNK15b2DistanceJoint10GetAnchorBEv,b54,b54,b54,b54,__ZN15b2DistanceJoint23InitVelocityConstraintsERK12b2SolverData,__ZN15b2DistanceJoint24SolveVelocityConstraintsERK12b2SolverData,b54,__ZNK15b2FrictionJoint10GetAnchorAEv,__ZNK15b2FrictionJoint10GetAnchorBEv,b54,b54,b54,b54,__ZN15b2FrictionJoint23InitVelocityConstraintsERK12b2SolverData,__ZN15b2FrictionJoint24SolveVelocityConstraintsERK12b2SolverData,b54,__ZNK11b2GearJoint10GetAnchorAEv,__ZNK11b2GearJoint10GetAnchorBEv,b54,b54,b54,b54,__ZN11b2GearJoint23InitVelocityConstraintsERK12b2SolverData,__ZN11b2GearJoint24SolveVelocityConstraintsERK12b2SolverData,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,__ZN4love5image6magpie13FormatHandler4freeEPh,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,_oc_set_chroma_mvs00,_oc_set_chroma_mvs01,_oc_set_chroma_mvs10,_oc_set_chroma_mvs11,b54,b54,b54,_oc_mb_fill_cmapping11,b54,b54,b54,_ft_bitmap_glyph_bbox,b54,b54,b54,b54,_ft_outline_glyph_bbox,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_parse_fd_array,_cid_parse_font_matrix,_parse_expansion_factor,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_ps_parser_to_token,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_t1_parse_font_matrix,_parse_encoding,_parse_subrs,_parse_charstrings,_parse_private,_parse_blend_design_positions,_parse_blend_design_map,_parse_blend_axis_types,_parse_weight_vector,_parse_buildchar,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_t42_parse_font_matrix,_t42_parse_encoding,_t42_parse_charstrings,_t42_parse_sfnts,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,_af_latin_metrics_scale,b54,b54,b54,_af_cjk_metrics_scale,b54,b54,b54,_af_indic_metrics_scale,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_floor1_pack,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_res0_pack,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZNSt3__110__stdinbufIcE5imbueERKNS_6localeE,b54,b54,b54,b54,b54,b54,b54,__ZNSt3__111__stdoutbufIcE5imbueERKNS_6localeE,b54,b54,b54,b54,b54,b54,b54,__ZNSt3__110__stdinbufIwE5imbueERKNS_6localeE,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZNSt3__111__stdoutbufIwE5imbueERKNS_6localeE,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE5imbueERKNS_6localeE,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZNKSt3__110moneypunctIcLb0EE11do_groupingEv,__ZNKSt3__110moneypunctIcLb0EE14do_curr_symbolEv,__ZNKSt3__110moneypunctIcLb0EE16do_positive_signEv,__ZNKSt3__110moneypunctIcLb0EE16do_negative_signEv,b54,__ZNKSt3__110moneypunctIcLb0EE13do_pos_formatEv,__ZNKSt3__110moneypunctIcLb0EE13do_neg_formatEv,b54,b54,b54,b54
,__ZNKSt3__110moneypunctIcLb1EE11do_groupingEv,__ZNKSt3__110moneypunctIcLb1EE14do_curr_symbolEv,__ZNKSt3__110moneypunctIcLb1EE16do_positive_signEv,__ZNKSt3__110moneypunctIcLb1EE16do_negative_signEv,b54,__ZNKSt3__110moneypunctIcLb1EE13do_pos_formatEv,__ZNKSt3__110moneypunctIcLb1EE13do_neg_formatEv,b54,b54,b54,b54,__ZNKSt3__110moneypunctIwLb0EE11do_groupingEv,__ZNKSt3__110moneypunctIwLb0EE14do_curr_symbolEv,__ZNKSt3__110moneypunctIwLb0EE16do_positive_signEv,__ZNKSt3__110moneypunctIwLb0EE16do_negative_signEv,b54,__ZNKSt3__110moneypunctIwLb0EE13do_pos_formatEv,__ZNKSt3__110moneypunctIwLb0EE13do_neg_formatEv,b54,b54,b54,b54,__ZNKSt3__110moneypunctIwLb1EE11do_groupingEv,__ZNKSt3__110moneypunctIwLb1EE14do_curr_symbolEv,__ZNKSt3__110moneypunctIwLb1EE16do_positive_signEv,__ZNKSt3__110moneypunctIwLb1EE16do_negative_signEv,b54,__ZNKSt3__110moneypunctIwLb1EE13do_pos_formatEv,__ZNKSt3__110moneypunctIwLb1EE13do_neg_formatEv,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZNKSt3__18messagesIcE8do_closeEi,b54,b54,b54,b54,__ZNKSt3__18messagesIwE8do_closeEi,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,__ZNKSt3__18numpunctIcE11do_groupingEv,__ZNKSt3__18numpunctIcE11do_truenameEv,__ZNKSt3__18numpunctIcE12do_falsenameEv,b54,b54,b54,b54,__ZNKSt3__18numpunctIwE11do_groupingEv,__ZNKSt3__18numpunctIwE11do_truenameEv,__ZNKSt3__18numpunctIwE12do_falsenameEv,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,__ZNSt3__18ios_base4initEPv,b54,__ZNSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEE3strERKNS_12basic_stringIcS2_S4_EE,b54,b54,__ZNKSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEE3strEv,__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7reserveEj,_lua_pushboolean,__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE9push_backEc,b54,__ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE6sentryC2ERS3_,b54,b54,b54,__ZNSt3__18ios_base5clearEj,b54,b54,__ZN4love4math15RandomGenerator7setSeedENS1_4SeedE,_lua_pushstring,b54,b54,__ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_,_lua_settop,b54,b54,b54,__ZNSt3__16vectorIN4love6VertexENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_
,b54,__ZNSt3__16vectorIN4love8TriangleENS_9allocatorIS2_EEE21__push_back_slow_pathIS2_EEvOT_,b54,b54,b54,b54,b54,b54,__ZN4love16luax_pushbooleanEP9lua_Stateb,b54,b54,b54,b54,b54,b54,__ZN4love4math15RandomGenerator8setStateERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE,__ZN4love15luax_pushstringEP9lua_StateRKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE,__ZNSt3__16vectorIiNS_9allocatorIiEEE21__push_back_slow_pathIiEEvOT_,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,_lua_pushvalue,b54,_lua_replace,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love5video11VideoStream10SourceSyncC2EPNS_5audio6SourceE,b54,_lua_pushinteger,b54,_lua_settable,b54,__ZNSt3__16vectorINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEENS4_IS6_EEE21__push_back_slow_pathIS6_EEvOT_,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love9ReferenceC2EP9lua_State,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZNSt3__16vectorIN4love9StrongRefINS1_7VariantEEENS_9allocatorIS4_EEE21__push_back_slow_pathIS4_EEvOT_,b54
,__ZN4love5event5Event4pushEPNS0_7MessageE,b54,b54,b54,b54,b54,b54,b54,__ZNSt3__16vectorINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEENS4_IS6_EEE21__push_back_slow_pathIRKS6_EEvOT_,b54,b54,b54,b54,b54,b54,__ZNSt3__16vectorIPN4love5image9ImageDataENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_,b54,b54,__ZNK4love4font9GlyphData14getGlyphStringEv,b54,__ZN4love8graphics6opengl8Graphics5clearERKNSt3__16vectorINS2_14OptionalColorfENS3_9allocatorIS5_EEEE,__ZN4love8graphics6opengl8Graphics5clearENS0_6ColorTIfEE,__ZNSt3__16vectorIbNS_9allocatorIbEEE7reserveEj,__ZNK4love8graphics6opengl8Graphics9getCanvasEv,b54,b54,b54,b54,b54,__ZNSt3__16vectorIPN4love5image19CompressedImageDataENS_9allocatorIS4_EEE21__push_back_slow_pathIS4_EEvOT_
,__ZNSt3__16vectorIPN4love5image9ImageDataENS_9allocatorIS4_EEE21__push_back_slow_pathIS4_EEvOT_,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZNSt3__16vectorIN4love8graphics6opengl4Mesh12AttribFormatENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,__ZNSt3__16vectorIPN4love8graphics6opengl6CanvasENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_,__ZN4love8graphics6opengl8Graphics9setCanvasERKNSt3__16vectorIPNS1_6CanvasENS3_9allocatorIS6_EEEE,b54,b54,b54,__ZNK4love8graphics6opengl8Graphics15getRendererInfoEv,b54,b54,b54,__ZN4love8graphics6opengl8Graphics4pushENS0_8Graphics9StackTypeE,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZNSt3__16vectorIjNS_9allocatorIjEEE21__push_back_slow_pathIjEEvOT_,__ZN4love8graphics6opengl4Mesh12setVertexMapERKNSt3__16vectorIjNS3_9allocatorIjEEEE
,b54,b54,b54,__ZN4love8graphics14ParticleSystem8setSizesERKNSt3__16vectorIfNS2_9allocatorIfEEEE,__ZN4love8graphics14ParticleSystem8setColorERKNSt3__16vectorINS0_6ColorTIfEENS2_9allocatorIS5_EEEE,b54,__ZNSt3__16vectorIPN4love8graphics4QuadENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_,__ZN4love8graphics14ParticleSystem8setQuadsERKNSt3__16vectorIPNS0_4QuadENS2_9allocatorIS5_EEEE,b54,b54,b54,b54,b54,b54,b54,__ZN4love8graphics6opengl11SpriteBatch13setBufferSizeEi,b54,__ZNSt3__16vectorIN4love8graphics6opengl4Font13ColoredStringENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_,b54,__ZN4love8graphics6opengl4Text3setERKNSt3__16vectorINS1_4Font13ColoredStringENS3_9allocatorIS6_EEEE,b54,b54,b54,b54,__ZN4love8graphics6opengl4Text7setFontEPNS1_4FontE,__ZN4love8graphics6opengl5Video9setFilterERKNS0_7Texture6FilterE,__ZNSt3__16vectorIfNS_9allocatorIfEEE21__push_back_slow_pathIfEEvOT_,__ZNSt3__16vectorIN4love8graphics6ColorTIfEENS_9allocatorIS4_EEE21__push_back_slow_pathIS4_EEvOT_,__ZNSt3__16vectorIfNS_9allocatorIfEEEC2ERKS3_,__ZNSt3__16vectorIN4love8graphics6ColorTIfEENS_9allocatorIS4_EEEC2ERKS7_
,__ZNSt3__16vectorIN4love9StrongRefINS1_8graphics4QuadEEENS_9allocatorIS5_EEEC2ERKS8_,__ZNSt3__16vectorIN4love9StrongRefINS1_8graphics4QuadEEENS_9allocatorIS5_EEE7reserveEj,b54,__ZNSt3__16vectorIN4love9StrongRefINS1_8graphics4QuadEEENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_,__ZNK4love8graphics4Quad11getViewportEv,__ZNSt3__16vectorIPN4love8graphics4QuadENS_9allocatorIS4_EEE21__push_back_slow_pathIS4_EEvOT_,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZNSt3__16vectorIN4love8keyboard8Keyboard3KeyENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_,b54,__ZNSt3__16vectorIN4love8keyboard8Keyboard8ScancodeENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_,b54,b54,__ZN4love4math11BezierCurveC2ERKNSt3__16vectorINS_6VectorENS2_9allocatorIS4_EEEE,b54,b54,b54,__ZNK4love4math11BezierCurve13getDerivativeEv
,b54,b54,b54,__ZN4love4math11BezierCurve18removeControlPointEi,b54,b54,b54,b54,b54,b54,b54,__ZN4love5mouse3sdl6CursorC2ENS0_6Cursor12SystemCursorE,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,__ZN4love7physics5box2d7Physics9scaleDownERK6b2Vec2,b54,b54,b54,b54,b54,b54,b54,__ZN4love8Memoizer3addEPvS1_,b54,b54,__ZN4love7physics5box2d7ContactC2EP9b2Contact,__ZN7b2WorldC2ERK6b2Vec2,__ZN7b2World16SetAllowSleepingEb,__ZN7b2World18SetContactListenerEP17b2ContactListener,__ZN7b2World16SetContactFilterEP15b2ContactFilter,__ZN7b2World22SetDestructionListenerEP21b2DestructionListener,b54,b54,b54,b54,b54,b54,b54
,b54,b54,__ZN4love7physics5box2d4Body7setTypeENS0_4Body4TypeE,__ZN4love7physics5box2d4Body9setActiveEb,__ZN4love7physics5box2d4Body16setFixedRotationEb,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love7physics5box2d7Fixture7destroyEb,b54,b54,b54,b54,b54,b54,b54,__ZN4love7physics5box2d5Joint12destroyJointEb,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love5sound9SoundDataC2EPNS0_7DecoderE,b54,b54,b54,b54,b54,b54,b54
,b54,__ZN4love5video6theora11VideoStreamC2EPNS_10filesystem4FileE,__ZNSt3__16vectorIN4love9StrongRefINS1_5video6theora11VideoStreamEEENS_9allocatorIS6_EEE21__push_back_slow_pathIS6_EEvOT_,b54,__ZN4love6thread4LockC2EPNS0_5MutexE,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,__ZNSt3__16vectorIN4love6window6Window10WindowSizeENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_,b54,b54,__ZNSt3__16vectorI24SDL_MessageBoxButtonDataNS_9allocatorIS1_EEE21__push_back_slow_pathIRKS1_EEvOT_,b54,b54,b54,b54,b54,b54,__ZN9b2Fixture7DestroyEP16b2BlockAllocator,b54,__ZN16b2StackAllocator4FreeEPv,__Z10loveAssertbPKc,b54,b54,b54,b54,b54,__Z14b2TimeOfImpactP11b2TOIOutputPK10b2TOIInput,__ZN9b2Contact6UpdateEP17b2ContactListener,b54,b54,b54,__ZN15b2CircleContact7DestroyEP9b2ContactP16b2BlockAllocator,b54,__ZN25b2PolygonAndCircleContact7DestroyEP9b2ContactP16b2BlockAllocator,b54,__ZN16b2PolygonContact7DestroyEP9b2ContactP16b2BlockAllocator,b54
,__ZN22b2EdgeAndCircleContact7DestroyEP9b2ContactP16b2BlockAllocator,b54,__ZN23b2EdgeAndPolygonContact7DestroyEP9b2ContactP16b2BlockAllocator,b54,__ZN23b2ChainAndCircleContact7DestroyEP9b2ContactP16b2BlockAllocator,b54,__ZN24b2ChainAndPolygonContact7DestroyEP9b2ContactP16b2BlockAllocator,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_alGenSources__wrapper,b54,b54,_alDeleteSources__wrapper,b54,b54,b54,b54,b54,b54
,_alDeleteBuffers__wrapper,__ZN4love5audio6openal4Pool4stopEPNS1_6SourceE,__ZN4love5audio6openal6SourceC2ERKS2_,_alGenBuffers__wrapper,b54,_SDL_DelEventWatch,b54,__ZNSt3__16vectorIN4love9StrongRefINS1_7VariantEEENS_9allocatorIS4_EEE7reserveEj,b54,b54,b54,b54,__ZN4love7VariantC2Eb,b54,b54,b54,__ZN4love7VariantC2EPv,b54,b54,__ZN4love10filesystem11DroppedFileC2ERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,_SDL_AddEventWatch,__ZNSt3__16vectorIN4love9StrongRefINS1_7VariantEEENS_9allocatorIS4_EEEC2ERKS7_,b54,b54,b54,b54,b54,b54,__ZN12_GLOBAL__N_19normalizeERKNSt3__112basic_stringIcNS0_11char_traitsIcEENS0_9allocatorIcEEEE,b54,b54,b54,__ZN4love10filesystem6physfs4FileC2ERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love4font16BMFontRasterizer11parseConfigERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE,b54,b54
,b54,b54,b54,b54,b54,b54,b54,__ZNSt11logic_errorC2EPKc,b54,b54,b54,b54,__ZN4love8graphics6opengl6Canvas8stopGrabEb,b54,b54,__ZN4love7Matrix4mLERKS0_,__ZN4love8graphics6opengl6OpenGL11bindTextureEj,__ZN4love8graphics6opengl6OpenGL21useVertexAttribArraysEj,b54,b54,b54,b54,b54,__ZNSt3__16vectorIN4love7Matrix4ENS_9allocatorIS2_EEE21__push_back_slow_pathIS2_EEvOT_,b54,__ZN4love8graphics6opengl6OpenGL13deleteTextureEj,b54,__ZNSt3__16vectorIjNS_9allocatorIjEEE21__push_back_slow_pathIRKjEEvOT_,b54,b54
,b54,__ZN4love8graphics6opengl11QuadIndicesC2Ej,b54,b54,b54,__ZNSt3__16vectorIN4love8graphics6opengl4Font11GlyphVertexENS_9allocatorIS5_EEE8__appendEj,b54,__ZNSt3__16vectorIN4love8graphics6opengl4Font11GlyphVertexENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_,__ZNSt3__16vectorIN4love8graphics6opengl4Font11DrawCommandENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_,b54,b54,b54,b54,__ZNSt3__16vectorIjNS_9allocatorIjEEEC2ERKS3_,__ZNSt3__16vectorIN4love8graphics6opengl4Font17ColoredCodepointsENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_,__ZNSt3__16vectorIN4love8graphics6opengl4Font12IndexedColorENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_,__ZNSt3__16vectorIN4love8graphics6opengl4Font12IndexedColorENS_9allocatorIS5_EEEC2ERKS8_,b54,b54,__ZN4love8graphics6opengl4Font23getCodepointsFromStringERKNSt3__16vectorINS2_13ColoredStringENS3_9allocatorIS5_EEEERNS2_17ColoredCodepointsE,b54,b54,b54,b54,__ZN4love8graphics6opengl6OpenGL11setViewportERKNS2_8ViewportE,__ZNK4love8graphics6opengl6OpenGL11getViewportEv,b54,__ZN4love8graphics6opengl8Graphics9setCanvasERKNSt3__16vectorINS_9StrongRefINS1_6CanvasEEENS3_9allocatorIS7_EEEE,__ZN4love8graphics6opengl6Canvas9startGrabERKNSt3__16vectorIPS2_NS3_9allocatorIS5_EEEE,__ZNSt3__16vectorIN4love9StrongRefINS1_8graphics6opengl6CanvasEEENS_9allocatorIS6_EEE7reserveEj
,__ZNSt3__16vectorIN4love9StrongRefINS1_8graphics6opengl6CanvasEEENS_9allocatorIS6_EEE21__push_back_slow_pathIS6_EEvOT_,__ZN4love8graphics6opengl6ShaderC2ERKNS2_12ShaderSourceE,b54,b54,__ZNSt3__16vectorIN4love8graphics6opengl8Graphics12DisplayStateENS_9allocatorIS5_EEE26__swap_out_circular_bufferERNS_14__split_bufferIS5_RS7_EE,__ZNSt3__16vectorIN4love8graphics6opengl8Graphics12DisplayStateENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_,__ZNSt3__16vectorIN4love9StrongRefINS1_8graphics6opengl6CanvasEEENS_9allocatorIS6_EEEC2ERKS9_,b54,b54,b54,__ZN4love8graphics6opengl8Graphics12restoreStateERKNS2_12DisplayStateE,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love8graphics6opengl5VideoC2EPNS_5video11VideoStreamE,b54,b54,b54,b54
,b54,b54,b54,__ZNSt3__16vectorIN4love9StrongRefINS1_5image9ImageDataEEENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_,b54,__ZNSt3__16vectorIN4love9StrongRefINS1_5image19CompressedImageDataEEENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_,b54,b54,b54,b54,b54,b54,b54,__ZNSt3__16vectorIN4love8graphics6opengl4Mesh12AttribFormatENS_9allocatorIS5_EEEC2ERKS8_,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love8graphics14ParticleSystemC2ERKS1_,__ZN4love8graphics6opengl11QuadIndicesC2ERKS2_
,__ZN4love8graphics6opengl14ParticleSystem14createVerticesEj,b54,b54,__ZN4love8graphics6opengl6OpenGL14setTextureUnitEi,b54,__ZNK4love8graphics6opengl6Shader18getProgramWarningsEv,b54,b54,b54,b54,b54,__ZN4love7Matrix3C2ERKNS_7Matrix4E,__ZNK4love7Matrix317transposedInverseEv,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,__ZN4love8graphics6opengl4Text11addTextDataERKNS2_8TextDataE,b54,__ZNSt3__16vectorIN4love8graphics6opengl4Text8TextDataENS_9allocatorIS5_EEE21__push_back_slow_pathIRKS5_EEvOT_,b54,b54,__ZN4love8graphics6opengl6Shader6attachEb,__ZN4love8graphics6opengl8Graphics9setCanvasEPNS1_6CanvasE,b54,b54,b54,__ZN4love8graphics6opengl4Font9setFilterERKNS0_7Texture6FilterE,b54,b54,__ZN4love8graphics6opengl4Font12setFallbacksERKNSt3__16vectorIPS2_NS3_9allocatorIS5_EEEE,b54,__ZNSt3__16vectorIPN4love8graphics6opengl4FontENS_9allocatorIS5_EEE21__push_back_slow_pathIS5_EEvOT_,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love5image6magpie9ImageData6decodeEPNS_10filesystem8FileDataE,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZN4love8joystick3sdl8JoystickC2Ei,__ZNSt3__16vectorIPN4love8joystick8JoystickENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_,b54,b54,b54,b54,b54,b54,__ZNK4love8joystick3sdl14JoystickModule13checkGamepadsERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE,b54,b54,_SDL_JoystickGetGUIDFromString
,b54,b54,b54,__ZNSt3__16vectorIN4love8joystick8Joystick13GamepadButtonENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_,__ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEEC2ERKS5_,__ZN12_GLOBAL__N_19subdivideERNSt3__16vectorIN4love6VectorENS0_9allocatorIS3_EEEEi,__ZNSt3__16vectorIN4love6VectorENS_9allocatorIS2_EEE8__appendEj,b54,b54,b54,b54,__ZN6b2Body7SetTypeE10b2BodyType,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_ModPlug_SetMasterVolume,b54,b54,b54,b54,b54,__ZN4love6thread3sdl6ThreadC2EPNS0_10ThreadableE
,__ZN4love6thread9EmptyLock7setLockEPNS0_5MutexE,b54,__ZN4love6thread7ChannelC2ERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZNSt3__16vectorIN4love5image19CompressedImageData8SubImageENS_9allocatorIS4_EEE21__push_back_slow_pathIRKS4_EEvOT_,__ZN4love6thread4LockC2ERNS0_5MutexE,_SDL_WaitThread,b54
,b54,b54,b54,__ZNSt3__16vectorIN3dds5ImageENS_9allocatorIS2_EEE21__push_back_slow_pathIRKS2_EEvOT_,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_zlibPhysfsFree
,_f_call,_resume,_f_parser,b54,b54,_f_luaopen,_callallgcTM,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_hookf,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_oc_idct8x8_c,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,_ft_free,b54,b54,b54,b54,b54,b54,_cf2_builder_moveTo,_cf2_builder_lineTo,_cf2_builder_cubeTo,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_ps_hints_t1reset,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_emscripten_longjmp__wrapper,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,_ft_gzip_free,b54,b54,b54,_zcfree_9181,b54,b54,b54,b54,b54,b54,b54,b54,_GLES2_WindowEvent,b54,b54,b54,b54,b54,_GLES2_UnlockTexture,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,_GLES2_DestroyTexture,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_SW_WindowEvent,b54,b54,b54,b54,b54,b54,b54,_SW_UnlockTexture
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_SW_DestroyTexture,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_Emscripten_SetWindowTitle,_Emscripten_SetWindowSize,_Emscripten_DestroyWindow,b54,b54,b54,_Emscripten_DestroyWindowFramebuffer,b54,b54,b54,b54,b54,b54,b54,_Emscripten_GLES_SwapWindow,_Emscripten_GLES_DeleteContext,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_SDL_DUMMY_DestroyWindowFramebuffer,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_SDL_DestroyWindowTexture
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,_zcfree,_emscripten_glPixelStorei__wrapper,b54,_emscripten_glGetIntegerv__wrapper,_emscripten_glGetFloatv__wrapper,_emscripten_glGetBooleanv__wrapper,_emscripten_glGenTextures__wrapper,_emscripten_glDeleteTextures__wrapper,b54,b54,b54,b54,b54,_emscripten_glBindTexture__wrapper,b54,b54,b54,b54,b54,_emscripten_glGenBuffers__wrapper,_emscripten_glDeleteBuffers__wrapper,b54,b54,b54,b54,_emscripten_glGenRenderbuffers__wrapper,_emscripten_glDeleteRenderbuffers__wrapper,_emscripten_glBindRenderbuffer__wrapper,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_emscripten_glUniform1i__wrapper,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_emscripten_glBindBuffer__wrapper,_emscripten_glVertexAttrib1fv__wrapper,_emscripten_glVertexAttrib2fv__wrapper,_emscripten_glVertexAttrib3fv__wrapper,_emscripten_glVertexAttrib4fv__wrapper,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_emscripten_glAttachShader__wrapper,_emscripten_glDetachShader__wrapper,b54,b54,b54,b54,b54,b54,b54,_emscripten_glBindFramebuffer__wrapper,_emscripten_glGenFramebuffers__wrapper,_emscripten_glDeleteFramebuffers__wrapper,b54,b54
,b54,b54,b54,b54,b54,_emscripten_glBindProgramARB__wrapper,_emscripten_glGetPointerv__wrapper,b54,b54,b54,b54,b54,b54,b54,_emscripten_glGenVertexArrays__wrapper,_emscripten_glDeleteVertexArrays__wrapper,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,_emscripten_glVertexAttribDivisor__wrapper,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_emscripten_glBlendFunc__wrapper,_emscripten_glBlendEquationSeparate__wrapper,b54,b54,_emscripten_glStencilMaskSeparate__wrapper,_emscripten_glHint__wrapper,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,_emscripten_glDrawBuffers__wrapper,b54,b54,b54,b54,b54,__ZNSt3__118__libcpp_refstringC2EPKc,__ZNSt13runtime_errorC2EPKc,b54,__ZNSt3__18ios_base16__call_callbacksENS0_5eventE,b54
,__ZNSt3__113basic_ostreamIwNS_11char_traitsIwEEE6sentryC2ERS3_,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,__ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE9push_backEw,b54,__ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE7reserveEj,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,__ZNSt3__16vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEEC2Ej,__ZNSt3__16locale5__imp7installINS_7collateIcEEEEvPT_,__ZNSt3__16locale5__imp7installINS_7collateIwEEEEvPT_,__ZNSt3__16locale5__imp7installINS_5ctypeIcEEEEvPT_,__ZNSt3__16locale5__imp7installINS_5ctypeIwEEEEvPT_,__ZNSt3__16locale5__imp7installINS_7codecvtIcc11__mbstate_tEEEEvPT_,__ZNSt3__17codecvtIwc11__mbstate_tEC2Ej,__ZNSt3__16locale5__imp7installINS_7codecvtIwc11__mbstate_tEEEEvPT_,__ZNSt3__16locale5__imp7installINS_7codecvtIDsc11__mbstate_tEEEEvPT_,__ZNSt3__16locale5__imp7installINS_7codecvtIDic11__mbstate_tEEEEvPT_,__ZNSt3__16locale5__imp7installINS_8numpunctIcEEEEvPT_,__ZNSt3__16locale5__imp7installINS_8numpunctIwEEEEvPT_,__ZNSt3__16locale5__imp7installINS_7num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_,__ZNSt3__16locale5__imp7installINS_7num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_,__ZNSt3__16locale5__imp7installINS_7num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_,__ZNSt3__16locale5__imp7installINS_7num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_,__ZNSt3__16locale5__imp7installINS_10moneypunctIcLb0EEEEEvPT_,__ZNSt3__16locale5__imp7installINS_10moneypunctIcLb1EEEEEvPT_,__ZNSt3__16locale5__imp7installINS_10moneypunctIwLb0EEEEEvPT_,__ZNSt3__16locale5__imp7installINS_10moneypunctIwLb1EEEEEvPT_,__ZNSt3__16locale5__imp7installINS_9money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_,__ZNSt3__16locale5__imp7installINS_9money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_,__ZNSt3__16locale5__imp7installINS_9money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_
,__ZNSt3__16locale5__imp7installINS_9money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_,__ZNSt3__16locale5__imp7installINS_8time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_,__ZNSt3__16locale5__imp7installINS_8time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_,__ZNSt3__16locale5__imp7installINS_8time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_,__ZNSt3__16locale5__imp7installINS_8time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_,__ZNSt3__16locale5__imp7installINS_8messagesIcEEEEvPT_,__ZNSt3__16locale5__imp7installINS_8messagesIwEEEEvPT_,__ZNSt3__16vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE8allocateEj,__ZNSt3__16vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE6resizeEj,b54,b54,b54,b54,__ZNSt13runtime_errorC2ERKNSt3__112basic_stringIcNS0_11char_traitsIcEENS0_9allocatorIcEEEE,b54,b54,b54,b54,b54,b54,_abort_message,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54,b54
,b54,b54,b54];
var FUNCTION_TABLE_viidd = [b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,__ZN4love7physics5box2d10MouseJointC2EPNS1_4BodyEff,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,__ZN4love8graphics4QuadC2ERKNS1_8ViewportEdd,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55
,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55,b55];
var FUNCTION_TABLE_viddii = [b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,__ZNK4love6window3sdl6Window8toPixelsEddRdS3_,b56,__ZNK4love6window3sdl6Window10fromPixelsEddRdS3_,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,__ZN4love7physics5box2d4Body13getLocalPointEffRfS3_,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56
,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56,b56];
var FUNCTION_TABLE_viidi = [b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,__ZN4love8graphics6opengl8Polyline15render_overdrawERKNSt3__16vectorINS_6VectorENS3_9allocatorIS5_EEEEfb,b57,b57,b57,__ZN4love8graphics6opengl16NoneJoinPolyline15render_overdrawERKNSt3__16vectorINS_6VectorENS3_9allocatorIS5_EEEEfb,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,__ZN4love8graphics6opengl4Text3setERKNSt3__16vectorINS1_4Font13ColoredStringENS3_9allocatorIS6_EEEEfNS5_9AlignModeE,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57
,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57,b57];
var FUNCTION_TABLE_viddd = [b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,__ZN4love5audio4null6Source7setConeEfff,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,__ZN4love5audio6openal6Source7setConeEfff,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,_emscripten_glUniform3f__wrapper,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,_emscripten_glVertexAttrib3f__wrapper,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58,b58
,b58,b58,b58];
var FUNCTION_TABLE_viiiiiiii = [b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,_emscripten_glCompressedTexImage2D__wrapper,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,_emscripten_glCopyTexImage2D__wrapper,_emscripten_glCopyTexSubImage2D__wrapper,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59,b59
,b59,b59,b59];
var FUNCTION_TABLE_viiiiiiid = [b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,__ZN16b2PulleyJointDef10InitializeEP6b2BodyS1_RK6b2Vec2S4_S4_S4_f,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60
,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60,b60];
var FUNCTION_TABLE_viiiiiiiiiiiiiii = [b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,__ZNSt3__111__money_putIcE8__formatEPcRS2_S3_jPKcS5_RKNS_5ctypeIcEEbRKNS_10money_base7patternEccRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEESL_SL_i,b61,__ZNSt3__111__money_putIwE8__formatEPwRS2_S3_jPKwS5_RKNS_5ctypeIwEEbRKNS_10money_base7patternEwwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERKNSE_IwNSF_IwEENSH_IwEEEESQ_i,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61,b61
,b61,b61,b61];
var FUNCTION_TABLE_iid = [b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,__ZN4love5sound7lullaby13VorbisDecoder4seekEf,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,__ZN4love5sound7lullaby11WaveDecoder4seekEf,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,__ZN4love5sound7lullaby13Mpg123Decoder4seekEf,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,__ZN4love5sound7lullaby14ModPlugDecoder4seekEf,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,__ZN4love7physics5box2d7Physics14newCircleShapeEf,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62
,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62,b62];
var FUNCTION_TABLE_dddd = [b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63
,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63
,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,__ZN4love4math3__78__invokeEfff,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63
,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63,b63
,b63,b63,b63,b63,b63,b63,b63,b63,b63];
var FUNCTION_TABLE_viiiiiii = [b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,__ZN4love8graphics6openglL7debugCBEjjjjiPKcPKv,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,_SDL_DrawLine2,_SDL_DrawLine1,_SDL_DrawLine4,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,_emscripten_glReadPixels__wrapper,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,_emscripten_glGetActiveUniform__wrapper,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,_emscripten_glGetActiveAttrib__wrapper,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,__ZNSt3__19__num_putIcE21__widen_and_group_intEPcS2_S2_S2_RS2_S3_RKNS_6localeE,b64,__ZNSt3__19__num_putIcE23__widen_and_group_floatEPcS2_S2_S2_RS2_S3_RKNS_6localeE,b64,b64,b64,__ZNSt3__19__num_putIwE21__widen_and_group_intEPcS2_S2_PwRS3_S4_RKNS_6localeE,__ZNSt3__19__num_putIwE23__widen_and_group_floatEPcS2_S2_PwRS3_S4_RKNS_6localeE,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64,b64
,b64,b64,b64];
var FUNCTION_TABLE_viiiiiiiii = [b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,_Color16DitherYV12Mod1X,_Color16DitherYV12Mod2X,_Color24DitherYV12Mod1X,_Color24DitherYV12Mod2X,_Color32DitherYV12Mod1X,_Color32DitherYV12Mod2X,_Color16DitherYUY2Mod1X,_Color16DitherYUY2Mod2X,_Color24DitherYUY2Mod1X,_Color24DitherYUY2Mod2X,_Color32DitherYUY2Mod1X,_Color32DitherYUY2Mod2X,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,_emscripten_glCompressedTexSubImage2D__wrapper,_emscripten_glTexImage2D__wrapper,_emscripten_glTexSubImage2D__wrapper,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65,b65
,b65,b65,b65];
var FUNCTION_TABLE_iii = [b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZNSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEE9pbackfailEi,b66,__ZNSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEE8overflowEi,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZNK4love4math13LZ4Compressor11isSupportedENS0_10Compressor6FormatE,b66,b66,b66,__ZNK4love4math14zlibCompressor11isSupportedENS0_10Compressor6FormatE,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love5mouse3sdl5Mouse15getSystemCursorENS0_6Cursor12SystemCursorE,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,__ZNK4love5mouse3sdl5Mouse6isDownERKNSt3__16vectorIiNS3_9allocatorIiEEEE,b66,b66,b66,__ZN4love5mouse3sdl5Mouse15setRelativeModeEb,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,__ZN4love7physics5box2d5World13QueryCallback13ReportFixtureEP9b2Fixture,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love5sound7lullabyL13tell_callbackEPvPy,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZNK4love6system6System7openURLERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE,b66,b66,b66,b66,b66,b66,b66
,__ZN4love6thread12ThreadModule10getChannelERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,__ZN4love5video6theora5Video14newVideoStreamEPNS_10filesystem4FileE,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love6window3sdl6Window13setFullscreenEb,b66,b66,__ZNK4love6window3sdl6Window14getDisplayNameEi
,b66,b66,b66,b66,b66,b66,b66,__ZN4love6window3sdl6Window7setIconEPNS_5image9ImageDataE,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,__ZN4love6window3sdl6Window14showMessageBoxERKNS0_6Window14MessageBoxDataE,b66,b66,b66,__ZNK12b2ChainShape5CloneEP16b2BlockAllocator,b66,b66,b66,b66,b66,b66,b66,__ZNK13b2CircleShape5CloneEP16b2BlockAllocator,b66,b66,b66,b66,b66,b66,__ZNK11b2EdgeShape5CloneEP16b2BlockAllocator,b66,b66,b66,b66,b66,b66,__ZNK14b2PolygonShape5CloneEP16b2BlockAllocator,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN12b2MotorJoint24SolvePositionConstraintsERK12b2SolverData,b66
,b66,b66,b66,b66,b66,b66,b66,b66,__ZN12b2MouseJoint24SolvePositionConstraintsERK12b2SolverData,b66,b66,b66,b66,b66,b66,b66,b66,__ZN16b2PrismaticJoint24SolvePositionConstraintsERK12b2SolverData,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN13b2PulleyJoint24SolvePositionConstraintsERK12b2SolverData,b66,b66
,b66,b66,b66,b66,b66,b66,__ZN15b2RevoluteJoint24SolvePositionConstraintsERK12b2SolverData,b66,b66,b66,b66,b66,b66,b66,b66,__ZN11b2RopeJoint24SolvePositionConstraintsERK12b2SolverData,b66,b66,b66,b66,b66,b66,b66,b66,__ZN11b2WeldJoint24SolvePositionConstraintsERK12b2SolverData,b66,b66,b66,b66,b66
,b66,b66,b66,__ZN12b2WheelJoint24SolvePositionConstraintsERK12b2SolverData,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_opt_keepalive,_opt_reuseaddr,_opt_tcp_nodelay,_opt_linger,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,_opt_dontroute,_opt_broadcast,_opt_ip_multicast_ttl,_opt_ip_multicast_loop,_opt_ip_add_membership,_opt_ip_drop_membersip,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love5audio4null5Audio9newSourceEPNS_5sound7DecoderE,__ZN4love5audio4null5Audio9newSourceEPNS_5sound9SoundDataE
,b66,b66,b66,__ZN4love5audio4null5Audio4playEPNS0_6SourceE,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love5audio4null5Audio13stopRecordingEb,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love5audio6openal5Audio9newSourceEPNS_5sound7DecoderE,__ZN4love5audio6openal5Audio9newSourceEPNS_5sound9SoundDataE,b66,b66,b66,__ZN4love5audio6openal5Audio4playEPNS0_6SourceE,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love5audio6openal5Audio13stopRecordingEb,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,__ZN4love10filesystem6physfs10Filesystem9setSourceEPKc,b66,b66,__ZN4love10filesystem6physfs10Filesystem7unmountEPKc,__ZNK4love10filesystem6physfs10Filesystem7newFileEPKc,b66,b66,b66,b66,b66,b66,b66,b66,__ZNK4love10filesystem6physfs10Filesystem6existsEPKc,__ZNK4love10filesystem6physfs10Filesystem11isDirectoryEPKc,__ZNK4love10filesystem6physfs10Filesystem6isFileEPKc,__ZNK4love10filesystem6physfs10Filesystem9isSymlinkEPKc,__ZN4love10filesystem6physfs10Filesystem15createDirectoryEPKc,__ZN4love10filesystem6physfs10Filesystem6removeEPKc,b66,b66,b66,b66,__ZNK4love10filesystem6physfs10Filesystem15getLastModifiedEPKc,__ZNK4love10filesystem6physfs10Filesystem7getSizeEPKc,b66,b66,b66
,b66,__ZNK4love10filesystem10Filesystem15isRealDirectoryERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE,b66,b66,b66,__ZN4love10filesystem11DroppedFile4openENS0_4File4ModeE,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZNK4love10filesystem11DroppedFile9getBufferERx,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love4font8freetype4Font13newRasterizerEPNS_10filesystem8FileDataE,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,__ZNK4love4font8freetype18TrueTypeRasterizer12getGlyphDataEj,__ZNK4love4font10Rasterizer12getGlyphDataERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE,b66,__ZNK4love4font8freetype18TrueTypeRasterizer8hasGlyphEj,__ZNK4love4font10Rasterizer9hasGlyphsERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEE,b66,b66,b66,b66,__ZNK4love4font16BMFontRasterizer12getGlyphDataEj,b66,__ZNK4love4font16BMFontRasterizer8hasGlyphEj,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,__ZNK4love4font15ImageRasterizer12getGlyphDataEj,b66,__ZNK4love4font15ImageRasterizer8hasGlyphEj,b66,b66,b66,b66,b66,b66,__ZN4love8graphics6opengl6Canvas7setWrapERKNS0_7Texture4WrapE,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love8graphics6opengl5Image7setWrapERKNS0_7Texture4WrapE,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,__ZN4love5image6magpie5Image12newImageDataEPNS_10filesystem8FileDataE,b66,b66,__ZN4love5image6magpie5Image17newCompressedDataEPNS_10filesystem8FileDataE,__ZN4love5image6magpie5Image12isCompressedEPNS_10filesystem8FileDataE,b66,b66,b66,b66,b66,b66,__ZN4love5image6magpie10KTXHandler8canParseEPKNS_10filesystem8FileDataE,b66,b66,__ZN4love5image6magpie10PKMHandler8canParseEPKNS_10filesystem8FileDataE,b66,b66,b66,__ZN4love5image6magpie10PNGHandler9canDecodeEPNS_10filesystem8FileDataE,__ZN4love5image6magpie10PNGHandler9canEncodeENS0_9ImageData13EncodedFormatE,b66,b66,b66,b66
,__ZN4love5image6magpie10PVRHandler8canParseEPKNS_10filesystem8FileDataE,b66,b66,__ZN4love5image6magpie10STBHandler9canDecodeEPNS_10filesystem8FileDataE,__ZN4love5image6magpie10STBHandler9canEncodeENS0_9ImageData13EncodedFormatE,b66,b66,b66,b66,b66,b66,b66,__ZN4love8joystick3sdl8Joystick4openEi,b66,b66,b66,b66,b66,b66,b66,b66,__ZNK4love8joystick3sdl8Joystick6getHatEi,__ZNK4love8joystick3sdl8Joystick6isDownERKNSt3__16vectorIiNS3_9allocatorIiEEEE,__ZN4love8joystick3sdl8Joystick11openGamepadEi,b66,b66,__ZNK4love8joystick3sdl8Joystick13isGamepadDownERKNSt3__16vectorINS0_8Joystick13GamepadButtonENS3_9allocatorIS6_EEEE,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love8joystick3sdl14JoystickModule11addJoystickEi,b66,__ZN4love8joystick3sdl14JoystickModule17getJoystickFromIDEi,__ZN4love8joystick3sdl14JoystickModule11getJoystickEi,__ZN4love8joystick3sdl14JoystickModule8getIndexEPKNS0_8JoystickE,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZNK4love8keyboard3sdl8Keyboard6isDownERKNSt3__16vectorINS0_8Keyboard3KeyENS3_9allocatorIS6_EEEE,__ZNK4love8keyboard3sdl8Keyboard14isScancodeDownERKNSt3__16vectorINS0_8Keyboard8ScancodeENS3_9allocatorIS6_EEEE,__ZNK4love8keyboard3sdl8Keyboard18getKeyFromScancodeENS0_8Keyboard8ScancodeE,__ZNK4love8keyboard3sdl8Keyboard18getScancodeFromKeyENS0_8Keyboard3KeyE,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN15b2DistanceJoint24SolvePositionConstraintsERK12b2SolverData,b66,b66,b66,b66,b66,b66,b66,b66,__ZN15b2FrictionJoint24SolvePositionConstraintsERK12b2SolverData,b66,b66,b66,b66,b66,b66,b66,b66,__ZN11b2GearJoint24SolvePositionConstraintsERK12b2SolverData,b66
,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love10filesystem6physfs4File4openENS0_4File4ModeE,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZNK4love10filesystem6physfs4File9getBufferERx,b66,b66,b66,__ZN4love5image6magpie11ASTCHandler8canParseEPKNS_10filesystem8FileDataE,b66,b66,b66,b66,__ZN4love5image6magpie10DDSHandler8canParseEPKNS_10filesystem8FileDataE,b66
,b66,__ZN4love5image6magpie13FormatHandler9canDecodeEPNS_10filesystem8FileDataE,__ZN4love5image6magpie13FormatHandler9canEncodeENS0_9ImageData13EncodedFormatE,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_bad_head_read,_bad_head_shift,_bad_skip_bytes,b66,_bad_back_bytes,_bad_seek_frame,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_INT123_synth_1to1_m2s,_INT123_synth_1to1_8bit_m2s,_INT123_synth_1to1_real_m2s,_INT123_synth_1to1_s32_m2s,_INT123_synth_2to1_m2s,_INT123_synth_2to1_8bit_m2s,_INT123_synth_2to1_real_m2s,_INT123_synth_2to1_s32_m2s,_INT123_synth_4to1_m2s,_INT123_synth_4to1_8bit_m2s,_INT123_synth_4to1_real_m2s,_INT123_synth_4to1_s32_m2s,_INT123_synth_ntom_m2s,_INT123_synth_ntom_8bit_m2s,_INT123_synth_ntom_real_m2s,_INT123_synth_ntom_s32_m2s,_INT123_synth_1to1_mono,_INT123_synth_1to1_8bit_mono
,_INT123_synth_1to1_real_mono,_INT123_synth_1to1_s32_mono,_INT123_synth_2to1_mono,_INT123_synth_2to1_8bit_mono,_INT123_synth_2to1_real_mono,_INT123_synth_2to1_s32_mono,_INT123_synth_4to1_mono,_INT123_synth_4to1_8bit_mono,_INT123_synth_4to1_real_mono,_INT123_synth_4to1_s32_mono,_INT123_synth_ntom_mono,_INT123_synth_ntom_8bit_mono,_INT123_synth_ntom_real_mono,_INT123_synth_ntom_s32_mono,b66,b66,b66,_generic_head_read,_generic_head_shift,_stream_skip_bytes,b66,_stream_back_bytes,_stream_seek_frame,b66,b66,b66,b66,b66,_feed_skip_bytes,_feed_back_bytes
,_feed_seek_frame,b66,b66,_DIR_isArchive,_DIR_openArchive,b66,_DIR_exists,b66,b66,b66,b66,_DIR_openWrite,_DIR_openAppend,_DIR_remove,_DIR_mkdir,b66,b66,b66,b66,b66,b66,b66,b66,_ZIP_isArchive,_ZIP_openArchive,b66,_ZIP_exists,b66,b66,b66
,b66,_ZIP_openWrite,_ZIP_openAppend,_ZIP_remove,_ZIP_mkdir,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_ft_bitmap_glyph_init,b66,_ft_bitmap_glyph_copy,b66,_ft_outline_glyph_init,b66,_ft_outline_glyph_copy,b66,b66,_ft_outline_glyph_prepare,b66,b66,b66,b66
,b66,b66,b66,b66,b66,_bdf_driver_requester,b66,b66,b66,_BDF_Size_Request,_BDF_Size_Select,b66,b66,_bdf_cmap_init,b66,_bdf_cmap_char_index,_bdf_cmap_char_next,b66,b66,_cff_get_interface,b66,b66,b66,b66,b66,b66,b66,b66,b66,_cff_size_request
,_cff_size_select,_cff_ps_get_font_info,b66,b66,b66,_cff_get_name_index,_cff_get_cmap_info,_cff_cmap_encoding_init,b66,_cff_cmap_encoding_char_index,_cff_cmap_encoding_char_next,_cff_cmap_unicode_init,b66,_cff_cmap_unicode_char_index,_cff_cmap_unicode_char_next,b66,_cff_get_is_cid,b66,b66,b66,b66,b66,b66,b66,b66,b66,_cid_get_interface,b66,b66,b66
,b66,b66,b66,b66,_cid_size_request,b66,_cid_ps_get_font_info,_cid_ps_get_font_extra,b66,_cid_get_is_cid,b66,b66,b66,b66,_pcf_driver_requester,b66,b66,b66,_PCF_Size_Request,_PCF_Size_Select,b66,b66,_pcf_cmap_init,b66,_pcf_cmap_char_index,_pcf_cmap_char_next,_pfr_cmap_init,b66,_pfr_cmap_char_index,_pfr_cmap_char_next
,_pfr_get_service,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_ps_parser_to_fixed,b66,b66,b66,b66,b66,b66,b66,b66,b66,_t1_builder_check_points
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_t1_cmap_standard_init,b66,_t1_cmap_std_char_index,_t1_cmap_std_char_next,_t1_cmap_expert_init,_t1_cmap_custom_init,b66,_t1_cmap_custom_char_index,_t1_cmap_custom_char_next,_t1_cmap_unicode_init,b66,_t1_cmap_unicode_char_index,_t1_cmap_unicode_char_next,b66,b66,b66,b66,b66,b66
,_psnames_get_service,b66,b66,_ps_unicodes_char_index,_ps_unicodes_char_next,b66,b66,_ft_black_new,b66,b66,_ft_black_render,b66,b66,b66,b66,b66,b66,_tt_cmap_init,_tt_cmap0_char_index,_tt_cmap0_char_next,_tt_cmap0_validate,_tt_cmap0_get_info,_tt_cmap2_char_index,_tt_cmap2_char_next,_tt_cmap2_validate,_tt_cmap2_get_info,_tt_cmap4_init,_tt_cmap4_char_index,_tt_cmap4_char_next,_tt_cmap4_validate
,_tt_cmap4_get_info,_tt_cmap6_char_index,_tt_cmap6_char_next,_tt_cmap6_validate,_tt_cmap6_get_info,_tt_cmap8_char_index,_tt_cmap8_char_next,_tt_cmap8_validate,_tt_cmap8_get_info,_tt_cmap10_char_index,_tt_cmap10_char_next,_tt_cmap10_validate,_tt_cmap10_get_info,_tt_cmap12_init,_tt_cmap12_char_index,_tt_cmap12_char_next,_tt_cmap12_validate,_tt_cmap12_get_info,_tt_cmap13_init,_tt_cmap13_char_index,_tt_cmap13_char_next,_tt_cmap13_validate,_tt_cmap13_get_info,_tt_cmap14_init,b66,_tt_cmap14_char_index,_tt_cmap14_char_next,b66,b66,_tt_cmap14_variants
,b66,b66,_tt_cmap14_validate,_tt_cmap14_get_info,_sfnt_get_interface,b66,b66,b66,b66,b66,_tt_face_load_head,b66,_tt_face_load_cmap,_tt_face_load_maxp,_tt_face_load_os2,_tt_face_load_post,_tt_face_load_name,b66,_tt_face_load_kern,_tt_face_load_gasp,_tt_face_load_pclt,_tt_face_load_bhed,b66,b66,b66,b66,_tt_face_load_font_dir,b66,_tt_face_load_sbit,b66
,b66,b66,b66,_get_sfnt_table,b66,b66,b66,_sfnt_get_name_index,b66,b66,_tt_get_cmap_info,_gray_raster_new,b66,b66,_gray_raster_render,b66,_gray_move_to,_gray_line_to,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_tt_get_interface
,b66,b66,b66,b66,b66,b66,b66,b66,_tt_size_request,_tt_size_select,b66,_TT_Get_MM_Var,b66,b66,b66,b66,b66,b66,_Get_Interface,b66,b66,b66,b66,b66,b66,b66,b66,_T1_Read_Metrics,b66,_T1_Size_Request
,b66,b66,_t1_get_name_index,_t1_ps_get_font_info,_t1_ps_get_font_extra,b66,_t1_ps_get_font_private,b66,b66,_T1_Get_Multi_Master,b66,b66,_T1_Get_MM_Var,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_T42_Get_Interface,b66,b66,b66
,b66,b66,b66,b66,_T42_Size_Request,_T42_Size_Select,b66,_t42_get_name_index,b66,_t42_ps_get_font_info,_t42_ps_get_font_extra,b66,_t42_ps_get_font_private,b66,b66,b66,b66,_winfnt_get_service,b66,b66,b66,_FNT_Size_Request,_FNT_Size_Select,_winfnt_get_header,_fnt_cmap_init,_fnt_cmap_char_index,_fnt_cmap_char_next,_af_dummy_hints_init,b66,_af_latin_metrics_init
,b66,_af_latin_hints_init,b66,_af_cjk_metrics_init,b66,_af_cjk_hints_init,b66,_af_indic_metrics_init,b66,_af_indic_hints_init,b66,b66,b66,b66,_af_get_interface,b66,b66,_floor0_unpack,_floor0_look,b66,b66,_floor0_inverse1,b66,b66,_floor1_unpack,_floor1_look,b66,b66,_floor1_inverse1,b66
,b66,_mapping0_unpack,b66,b66,_mapping0_inverse,_res0_unpack,_res0_look,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_GLES2_CreateRenderer,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_SW_CreateRenderer,b66,b66,b66,b66,b66,b66,b66
,_deflate_stored,_deflate_fast,_deflate_slow,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZNSt3__110__stdinbufIcE9pbackfailEi,__ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE8overflowEi,b66,b66,b66,b66,__ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE9pbackfailEi,b66,__ZNSt3__111__stdoutbufIcE8overflowEi,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,__ZNSt3__110__stdinbufIwE9pbackfailEj,b66,__ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE8overflowEj,b66,b66,b66,b66,b66,__ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE9pbackfailEj,b66,__ZNSt3__111__stdoutbufIwE8overflowEj,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZNKSt3__15ctypeIcE10do_toupperEc,b66,__ZNKSt3__15ctypeIcE10do_tolowerEc,b66,__ZNKSt3__15ctypeIcE8do_widenEc,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZNKSt3__15ctypeIwE10do_toupperEw,b66,__ZNKSt3__15ctypeIwE10do_tolowerEw,b66,__ZNKSt3__15ctypeIwE8do_widenEc,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_lua_type
,b66,__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc,b66,b66,b66,b66,__ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEElsEi,b66,b66,b66,b66,b66,b66,b66,b66,__ZNKSt3__16locale9use_facetERNS0_2idE,b66,b66,b66,b66,b66,b66,__ZN4love11luax_objlenEP9lua_Statei,b66,b66,b66,b66,__ZN4love4math4Math14newBezierCurveERKNSt3__16vectorINS_6VectorENS2_9allocatorIS4_EEEE,b66,b66
,b66,b66,b66,b66,b66,b66,b66,__ZN4love4math4Math8isConvexERKNSt3__16vectorINS_6VertexENS2_9allocatorIS4_EEEE,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love7physics5box2d7Physics15newPolygonShapeEP9lua_State,__ZN4love7physics5box2d7Physics13newChainShapeEP9lua_State,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZNK4love7physics5box2d5Shape7rayCastEP9lua_State,b66,b66,__ZNK4love7physics5box2d5World11getBodyListEP9lua_State,__ZNK4love7physics5box2d5World12getJointListEP9lua_State,__ZNK4love7physics5box2d5World14getContactListEP9lua_State,__ZN4love7physics5box2d5World7rayCastEP9lua_State,b66,b66,__ZN4love5sound17luax_checkdecoderEP9lua_Statei,__ZN4love5sound5Sound12newSoundDataEPNS0_7DecoderE
,b66,b66,b66,b66,_lua_isstring,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love6window6Window11getConstantEPKcRNS1_14MessageBoxTypeE,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_lua_touserdata,b66,__ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE3putEc,b66,b66,__ZN4love14luax_tobooleanEP9lua_Statei,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love8graphics6opengl5Image11getConstantENS2_8FlagTypeERPKc,b66,__ZN4love10filesystem16luax_getfiledataEP9lua_Statei,b66
,b66,__ZN4love5image29luax_checkcompressedimagedataEP9lua_Statei,__ZN4love5image19luax_checkimagedataEP9lua_Statei,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love8graphics6opengl8Graphics9newShaderERKNS1_6Shader12ShaderSourceE,b66,b66,b66,__ZN4love8graphics6opengl4Mesh11getConstantEPKcRNS2_8DrawModeE,__ZN4love8graphics6opengl4Mesh11getConstantEPKcRNS2_5UsageE,__ZN4love8graphics6opengl4Mesh11getConstantEPKcRNS2_8DataTypeE,b66,_lua_isnumber,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,__ZN4love8graphics6opengl8Graphics8newVideoEPNS_5video11VideoStreamE,b66,b66,b66,__ZN4love8graphics6opengl16luax_checkcanvasEP9lua_Statei,b66,b66,b66,b66,__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEaSERKS5_,b66,b66,__ZN4love8graphics6opengl4Font11getConstantEPKcRNS2_9AlignModeE,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZNK4love8graphics6opengl4Mesh18isAttributeEnabledERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE,b66,b66,_luaL_checkinteger,b66,b66
,__ZNK4love8graphics6opengl4Mesh12getVertexMapERNSt3__16vectorIjNS3_9allocatorIjEEEE,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,__ZNK4love5image19CompressedImageData8getWidthEi,__ZNK4love5image19CompressedImageData9getHeightEi,b66,b66,b66,b66,b66,__ZN4love8joystick18luax_checkjoystickEP9lua_Statei,__ZN4love8joystick8Joystick11getConstantEPKcRNS1_11GamepadAxisE,__ZN4love8joystick8Joystick11getConstantEPKcRNS1_13GamepadButtonE,__ZN4love8joystick8Joystick11getConstantENS1_9InputTypeERPKc,__ZN4love8joystick8Joystick11getConstantENS1_3HatERPKc,b66,__ZN4love8keyboard8Keyboard11getConstantEPKcRNS1_3KeyE,b66,__ZN4love8keyboard8Keyboard11getConstantEPKcRNS1_8ScancodeE,b66,b66,b66,b66,b66,b66,b66,b66
,__ZNK4love4math11BezierCurve15getControlPointEi,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love7physics5box2d5Joint11createJointEP10b2JointDef,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN7b2World10CreateBodyEPK9b2BodyDef,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,__ZNK4love7physics5box2d4Body14getFixtureListEP9lua_State,__ZNK4love7physics5box2d4Body12getJointListEP9lua_State,__ZNK4love7physics5box2d4Body14getContactListEP9lua_State,b66,b66,b66,__ZNK4love7physics5box2d10ChainShape12getChildEdgeEi,b66,b66,b66,__ZNK4love7physics5box2d7Fixture7rayCastEP9lua_State,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_ov_info,_ov_comment,b66,b66,b66,_wuff_stream_info,_wuff_format,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_SDL_GL_SetAttribute,__ZZN4love6window3sdl6Window22createWindowAndContextEiiiijiENK3__0clENS2_14ContextAttribsE,b66,__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6appendEPKc,b66,b66
,b66,b66,_SDL_ShowMessageBox,b66,b66,b66,b66,b66,b66,b66,b66,__ZN16b2StackAllocator8AllocateEi,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_socket_ioerror,b66,b66,_mpg123_new,b66,b66,b66,b66,b66,_mpg123_open_handle,b66,b66,b66,b66,b66
,b66,b66,b66,_lua_isuserdata,b66,__ZN4love11getTypeNameEPKcRNS_4TypeE,b66,b66,b66,b66,b66,b66,b66,b66,b66,_alcCreateContext__wrapper,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love5audio6openal4Pool12removeSourceEPNS1_6SourceE,b66,b66
,b66,b66,b66,b66,b66,b66,__ZN4love5event3sdlL14watchAppEventsEPvP9SDL_Event,b66,__ZN4love8keyboard8Keyboard11getConstantENS1_3KeyERPKc,__ZN4love8keyboard3sdl8Keyboard11getConstantE12SDL_ScancodeRNS0_8Keyboard8ScancodeE,__ZN4love8keyboard8Keyboard11getConstantENS1_8ScancodeERPKc,b66,b66,b66,b66,b66,b66,__ZNK4love5event3sdl5Event20convertJoystickEventERK9SDL_Event,__ZNK4love5event3sdl5Event18convertWindowEventERK9SDL_Event,b66,b66,b66,__ZN4love8joystick3sdl8Joystick11getConstantEhRNS0_8Joystick3HatE,__ZN4love8joystick3sdl8Joystick11getConstantE24SDL_GameControllerButtonRNS0_8Joystick13GamepadButtonE,__ZN4love8joystick8Joystick11getConstantENS1_13GamepadButtonERPKc,__ZN4love8joystick3sdl8Joystick11getConstantE22SDL_GameControllerAxisRNS0_8Joystick11GamepadAxisE,__ZN4love8joystick8Joystick11getConstantENS1_11GamepadAxisERPKc,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZNSt3__113unordered_mapIiN4love9StrongRefINS1_5image9ImageDataEEENS_4hashIiEENS_8equal_toIiEENS_9allocatorINS_4pairIKiS5_EEEEEixERSC_,b66,b66,__ZNSt3__113unordered_mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES6_NS_4hashIS6_EENS_8equal_toIS6_EENS4_INS_4pairIKS6_S6_EEEEEixERSC_
,__ZNSt3__113unordered_mapIjN4love4font16BMFontRasterizer15BMFontCharacterENS_4hashIjEENS_8equal_toIjEENS_9allocatorINS_4pairIKjS4_EEEEEixERSB_,__ZNSt3__113unordered_mapIyiNS_4hashIyEENS_8equal_toIyEENS_9allocatorINS_4pairIKyiEEEEEixERS7_,b66,b66,b66,b66,b66,b66,__ZN4utf84nextINSt3__111__wrap_iterIPKcEEEEjRT_S6_,b66,__ZN4utf86appendIPcEET_jS2_,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love8graphics6opengl4Font8addGlyphEj,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love8graphics6opengl11QuadIndicesaSERKS2_,b66,b66,b66,b66,__ZN4love8graphics6opengl8GLBuffer4loadEb,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEElsEj,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZNK4love8graphics6opengl8GLBuffer10getPointerEj,b66,b66,b66,b66,b66,b66,b66,__ZN4love8graphics6opengl6Shader11getConstantENS1_14VertexAttribIDERPKc,b66,b66,b66,b66,b66,__ZNK4love8graphics6opengl11QuadIndices13getIndexCountEj,__ZNK4love8graphics6opengl11QuadIndices7getTypeEj,__ZNK4love8graphics6opengl11QuadIndices10getIndicesEj,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl6Shader7UniformENS_4lessIS6_EENS4_INS_4pairIKS6_SB_EEEEEixERSF_,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love8graphics6opengl6Shader14getTextureUnitERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE,__ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ObjectENS_4lessIS6_EENS4_INS_4pairIKS6_S9_EEEEEixERSD_,b66,b66,b66,__ZNK4love8graphics6opengl11QuadIndices10getPointerEj,b66,b66,b66
,__ZNK4love8graphics6opengl4Mesh17getAttributeIndexERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE,__ZNSt3__113unordered_mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl11SpriteBatch17AttachedAttributeENS_4hashIS6_EENS_8equal_toIS6_EENS4_INS_4pairIKS6_SB_EEEEEixERSH_,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN4love8graphics6opengl4Font8getWidthERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE,b66,b66,__ZNK4love8graphics6opengl4Font9hasGlyphsERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE,__ZNK4love8graphics6opengl4Font8hasGlyphEj,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_SDL_JoystickGetAxis,b66,b66,b66,b66,__ZN4love8joystick3sdl8Joystick11getConstantENS0_8Joystick3HatERh,b66,b66,b66,__ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEbNS_4lessIS6_EENS4_INS_4pairIKS6_bEEEEEixERSA_,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,__ZN6b2Body13CreateFixtureEPK12b2FixtureDef,b66,b66,b66,b66,b66,b66,b66,_ModPlug_Load,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,__ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6thread7ChannelENS_4lessIS6_EENS4_INS_4pairIKS6_SA_EEEEEixERSE_,b66,b66,b66,__ZN4love6thread7Channel4pushEPNS_7VariantE,b66,__Z14b2PairLessThanRK6b2PairS1_,__ZN12b2BroadPhase13QueryCallbackEi,b66,b66,b66,b66,b66,b66,__ZL12coin_comparePKvS0_,b66,b66,b66,b66,b66,b66,b66,__ZNK3dds6Parser12getImageDataEj,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_INT123_synth_1to1_8bit_wrap_mono,_INT123_synth_1to1_8bit_wrap_m2s,b66,b66,b66,b66,b66,_mallocAllocatorMalloc,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_ft_raccess_sort_ref_by_id,b66,b66
,_ft_alloc,b66,b66,b66,b66,b66,_by_encoding,_cff_sid_to_glyph_name,b66,b66,b66,b66,_cid_load_glyph,b66,b66,b66,b66,_afm_compare_kern_pairs,_psaux_get_glyph_name,b66,b66,b66,b66,_ps_hints_close,b66,b66,b66,b66,b66,b66
,b66,b66,_compare_uni_maps,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_compare_offsets,b66,_tt_name_entry_ascii_from_utf16,_tt_name_entry_ascii_from_other,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,_Read_CVT,_Read_CVT_Stretched,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_T1_Parse_Glyph,b66,_compare_kern_pairs
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_icomp,_apsort,b66,b66,_sort32a,b66,b66,_GLES2_CreateTexture,b66,b66,b66,b66,_GLES2_SetRenderTarget,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,_GLES2_UnbindTexture,_cmpmodes,_RemovePendingResizedEvents,_RemovePendingSizeChangedEvents,_RemovePendingMoveEvents,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_SW_CreateTexture,_SW_SetTextureColorMod,_SW_SetTextureAlphaMod,_SW_SetTextureBlendMode,b66,b66,b66
,_SW_SetRenderTarget,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,_Emscripten_CreateWindow,b66,b66,b66,b66,b66,b66,b66,_Emscripten_GLES_LoadLibrary,_SDL_EGL_GetProcAddress,b66,_Emscripten_GLES_CreateContext,b66,_SDL_EGL_SetSwapInterval,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_eglDestroyContext__wrapper,b66,_eglDestroySurface__wrapper,b66,_eglSwapBuffers__wrapper,_eglSwapInterval__wrapper,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,_SDL_GameControllerEventWatcher,_SDL_RendererEventWatch,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,_emscripten_glGetUniformLocation__wrapper,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,_emscripten_glGetAttribLocation__wrapper,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,_isxdigit_l,_isdigit_l,b66,b66,b66,b66,b66
,b66,b66,__ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66,b66
,b66,b66,b66];
var FUNCTION_TABLE_viiddi = [b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,__ZNK4love4math11BezierCurve13renderSegmentEddj,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67
,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67,b67];
var FUNCTION_TABLE_viiddddddddd = [b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,__ZN4love8graphics6opengl6Canvas5drawqEPNS0_4QuadEfffffffff,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,__ZN4love8graphics6opengl5Image5drawqEPNS0_4QuadEfffffffff,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,__ZN4love8graphics6opengl8Graphics5printERKNSt3__16vectorINS1_4Font13ColoredStringENS3_9allocatorIS6_EEEEfffffffff,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68
,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68,b68];
var FUNCTION_TABLE_vdddd = [b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,_emscripten_glRotatef__wrapper,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,_emscripten_glClearColor__wrapper,b69,b69,b69,b69,b69,_emscripten_glBlendColor__wrapper,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69,b69
,b69,b69,b69];
var FUNCTION_TABLE_viii = [b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZNK4love5mouse3sdl5Mouse11getPositionERdS3_,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,__ZN4love7physics5box2d5World8PreSolveEP9b2ContactPK10b2Manifold,__ZN4love7physics5box2d5World9PostSolveEP9b2ContactPK16b2ContactImpulse,b70,b70,b70,b70,b70,b70,b70,__ZThn8_N4love7physics5box2d5World8PreSolveEP9b2ContactPK10b2Manifold,__ZThn8_N4love7physics5box2d5World9PostSolveEP9b2ContactPK16b2ContactImpulse,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,__ZNK4love6window3sdl6Window18getFullscreenSizesEi,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZNK4love6window3sdl6Window18getPixelDimensionsERiS3_,__ZNK4love6window3sdl6Window19windowToPixelCoordsEPdS3_,__ZNK4love6window3sdl6Window19pixelToWindowCoordsEPdS3_,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZNK4love10filesystem6physfs10Filesystem16getRealDirectoryEPKc,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZN4love10filesystem6physfs10Filesystem17getDirectoryItemsEPKcRNSt3__16vectorINS5_12basic_stringIcNS5_11char_traitsIcEENS5_9allocatorIcEEEENSA_ISC_EEEE,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZN4love8graphics6opengl8Graphics15setViewportSizeEii,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZN4love5image6magpie10PNGHandler6decodeEPNS_10filesystem8FileDataE,b70,b70,b70
,b70,b70,b70,b70,b70,__ZN4love5image6magpie10STBHandler6decodeEPNS_10filesystem8FileDataE,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,__ZN4love8joystick3sdl8Joystick12getVibrationERfS3_,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZN17b2ContactListener8PreSolveEP9b2ContactPK10b2Manifold,__ZN17b2ContactListener9PostSolveEP9b2ContactPK16b2ContactImpulse,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,__ZN4love5image6magpie13FormatHandler6decodeEPNS_10filesystem8FileDataE,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__Z11Mono8BitMixP11_MODCHANNELPiS1_,__Z12Mono16BitMixP11_MODCHANNELPiS1_,__Z13Stereo8BitMixP11_MODCHANNELPiS1_,__Z14Stereo16BitMixP11_MODCHANNELPiS1_,__Z15Mono8BitRampMixP11_MODCHANNELPiS1_,__Z16Mono16BitRampMixP11_MODCHANNELPiS1_,__Z17Stereo8BitRampMixP11_MODCHANNELPiS1_
,__Z18Stereo16BitRampMixP11_MODCHANNELPiS1_,__Z17FilterMono8BitMixP11_MODCHANNELPiS1_,__Z18FilterMono16BitMixP11_MODCHANNELPiS1_,__Z19FilterStereo8BitMixP11_MODCHANNELPiS1_,__Z20FilterStereo16BitMixP11_MODCHANNELPiS1_,__Z21FilterMono8BitRampMixP11_MODCHANNELPiS1_,__Z22FilterMono16BitRampMixP11_MODCHANNELPiS1_,__Z23FilterStereo8BitRampMixP11_MODCHANNELPiS1_,__Z24FilterStereo16BitRampMixP11_MODCHANNELPiS1_,__Z17Mono8BitLinearMixP11_MODCHANNELPiS1_,__Z18Mono16BitLinearMixP11_MODCHANNELPiS1_,__Z19Stereo8BitLinearMixP11_MODCHANNELPiS1_,__Z20Stereo16BitLinearMixP11_MODCHANNELPiS1_,__Z21Mono8BitLinearRampMixP11_MODCHANNELPiS1_,__Z22Mono16BitLinearRampMixP11_MODCHANNELPiS1_,__Z23Stereo8BitLinearRampMixP11_MODCHANNELPiS1_,__Z24Stereo16BitLinearRampMixP11_MODCHANNELPiS1_,__Z23FilterMono8BitLinearMixP11_MODCHANNELPiS1_,__Z24FilterMono16BitLinearMixP11_MODCHANNELPiS1_,__Z25FilterStereo8BitLinearMixP11_MODCHANNELPiS1_,__Z26FilterStereo16BitLinearMixP11_MODCHANNELPiS1_,__Z27FilterMono8BitLinearRampMixP11_MODCHANNELPiS1_,__Z28FilterMono16BitLinearRampMixP11_MODCHANNELPiS1_,__Z29FilterStereo8BitLinearRampMixP11_MODCHANNELPiS1_,__Z30FilterStereo16BitLinearRampMixP11_MODCHANNELPiS1_,__Z17Mono8BitSplineMixP11_MODCHANNELPiS1_,__Z18Mono16BitSplineMixP11_MODCHANNELPiS1_,__Z19Stereo8BitSplineMixP11_MODCHANNELPiS1_,__Z20Stereo16BitSplineMixP11_MODCHANNELPiS1_,__Z21Mono8BitSplineRampMixP11_MODCHANNELPiS1_
,__Z22Mono16BitSplineRampMixP11_MODCHANNELPiS1_,__Z23Stereo8BitSplineRampMixP11_MODCHANNELPiS1_,__Z24Stereo16BitSplineRampMixP11_MODCHANNELPiS1_,__Z23FilterMono8BitSplineMixP11_MODCHANNELPiS1_,__Z24FilterMono16BitSplineMixP11_MODCHANNELPiS1_,__Z25FilterStereo8BitSplineMixP11_MODCHANNELPiS1_,__Z26FilterStereo16BitSplineMixP11_MODCHANNELPiS1_,__Z27FilterMono8BitSplineRampMixP11_MODCHANNELPiS1_,__Z28FilterMono16BitSplineRampMixP11_MODCHANNELPiS1_,__Z29FilterStereo8BitSplineRampMixP11_MODCHANNELPiS1_,__Z30FilterStereo16BitSplineRampMixP11_MODCHANNELPiS1_,__Z20Mono8BitFirFilterMixP11_MODCHANNELPiS1_,__Z21Mono16BitFirFilterMixP11_MODCHANNELPiS1_,__Z22Stereo8BitFirFilterMixP11_MODCHANNELPiS1_,__Z23Stereo16BitFirFilterMixP11_MODCHANNELPiS1_,__Z24Mono8BitFirFilterRampMixP11_MODCHANNELPiS1_,__Z25Mono16BitFirFilterRampMixP11_MODCHANNELPiS1_,__Z26Stereo8BitFirFilterRampMixP11_MODCHANNELPiS1_,__Z27Stereo16BitFirFilterRampMixP11_MODCHANNELPiS1_,__Z26FilterMono8BitFirFilterMixP11_MODCHANNELPiS1_,__Z27FilterMono16BitFirFilterMixP11_MODCHANNELPiS1_,__Z28FilterStereo8BitFirFilterMixP11_MODCHANNELPiS1_,__Z29FilterStereo16BitFirFilterMixP11_MODCHANNELPiS1_,__Z30FilterMono8BitFirFilterRampMixP11_MODCHANNELPiS1_,__Z31FilterMono16BitFirFilterRampMixP11_MODCHANNELPiS1_,__Z32FilterStereo8BitFirFilterRampMixP11_MODCHANNELPiS1_,__Z33FilterStereo16BitFirFilterRampMixP11_MODCHANNELPiS1_,__Z15FastMono8BitMixP11_MODCHANNELPiS1_,__Z16FastMono16BitMixP11_MODCHANNELPiS1_,__Z19FastMono8BitRampMixP11_MODCHANNELPiS1_
,__Z20FastMono16BitRampMixP11_MODCHANNELPiS1_,__Z21FastMono8BitLinearMixP11_MODCHANNELPiS1_,__Z22FastMono16BitLinearMixP11_MODCHANNELPiS1_,__Z25FastMono8BitLinearRampMixP11_MODCHANNELPiS1_,__Z26FastMono16BitLinearRampMixP11_MODCHANNELPiS1_,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,_ft_outline_glyph_transform,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,_t1_decrypt,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,_ft_black_reset,b70,b70,b70,b70,b70,b70,_ft_raster1_get_cbox,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,_gray_raster_reset,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,_ft_smooth_get_cbox,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,_mapping0_pack,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,__ZNKSt3__114error_category23default_error_conditionEi,b70,b70,__ZNKSt3__119__iostream_category7messageEi,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZN4love9ExceptionC2EPKcz,___cxa_throw__wrapper,b70,__ZNSt3__16__treeINS_12__value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6ModuleEEENS_19__map_value_compareIS7_SB_NS_4lessIS7_EELb1EEENS5_ISB_EEE15__insert_uniqueINS_4pairIS7_SA_EEEENSJ_INS_15__tree_iteratorISB_PNS_11__tree_nodeISB_PvEEiEEbEEOT_,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6resizeEjc,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,_lua_rawgeti,b70,b70,b70,b70,__ZN4love13luax_pushtypeEP9lua_StateNS_4TypeEPNS_6ObjectE,b70
,b70,b70,__ZN4love4math4Math11triangulateERKNSt3__16vectorINS_6VertexENS2_9allocatorIS4_EEEE,b70,_lua_createtable,b70,_lua_rawseti,b70,b70,b70,b70,b70,b70,__ZN4love4math20luax_checkrandomseedEP9lua_Statei,__ZN4love16luax_checkstringEP9lua_Statei,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6__initEPKcj,b70,b70,b70,b70,b70,b70,b70,b70,_lua_setfield,b70,_lua_getfield,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZN4love5event7MessageC2ERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEERKNS2_6vectorINS_9StrongRefINS_7VariantEEENS6_ISE_EEEE
,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZNSt3__113basic_istreamIcNS_11char_traitsIcEEE6sentryC2ERS3_b,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZN4love8graphics6opengl8Graphics7discardERKNSt3__16vectorIbNS3_9allocatorIbEEEEb,_luaL_checktype,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZN4love8graphics6opengl23luax_checkcoloredstringEP9lua_StateiRNSt3__16vectorINS1_4Font13ColoredStringENS4_9allocatorIS7_EEEE
,b70,b70,__ZN4love8graphics6opengl8Graphics12setBlendModeENS0_8Graphics9BlendModeENS3_10BlendAlphaE,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZN4love8graphics6opengl4Mesh19setAttributeEnabledERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEEb,b70,__ZN4love8graphics6opengl4Mesh15attachAttributeERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEEPS2_,b70,b70,b70,b70
,b70,__ZN4love8graphics6opengl4Mesh12setDrawRangeEii,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZN4love8graphics6opengl6Shader11sendTextureERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEEPNS0_7TextureE,b70,b70,b70,b70,__ZN4love8graphics6opengl11SpriteBatch15attachAttributeERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEEPNS1_4MeshE,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,__ZNSt3__16vectorIN4love9StrongRefINS1_8graphics4QuadEEENS_9allocatorIS5_EEE6assignIPS5_EENS_9enable_ifIXaasr21__is_forward_iteratorIT_EE5valuesr16is_constructibleIS5_NS_15iterator_traitsISC_E9referenceEEE5valueEvE4typeESC_SC_,b70,b70,b70,b70,b70,b70,b70,b70,b70,_lua_call,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,__ZN4love4math11BezierCurve15setControlPointEiRKNS_6VectorE,__ZN4love4math11BezierCurve18insertControlPointERKNS_6VectorEi,b70,b70,b70,__ZNK4love4math11BezierCurve6renderEi,b70,b70,b70,b70,b70,__ZN4love7physics5box2d5WorldC2E6b2Vec2b,b70,__ZN4love7physics5box2d11CircleShapeC2EP13b2CircleShapeb,__ZN4love7physics5box2d12PolygonShapeC2EP14b2PolygonShapeb,__ZN4love7physics5box2d9EdgeShapeC2EP11b2EdgeShapeb,__ZN14b2PolygonShape3SetEPK6b2Vec2i,__ZN12b2ChainShape10CreateLoopEPK6b2Vec2i,__ZN12b2ChainShape11CreateChainEPK6b2Vec2i,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,__ZN4love7physics5box2d10MotorJointC2EPNS1_4BodyES4_,b70,b70,__ZN15b2DistanceProxy3SetEPK7b2Shapei,__Z10b2DistanceP16b2DistanceOutputP14b2SimplexCachePK15b2DistanceInput,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZNK4love7physics5box2d10ChainShape8getPointEi,__ZN4love7physics5box2d7Contact11getFixturesERPNS1_7FixtureES5_,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZN4love6thread9LuaThreadC2ERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEEPNS_4DataE,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZN4love6window3sdl6Window26setGLFramebufferAttributesEib,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,_luaL_unref,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZN4love5audio6openal6SourceC2EPNS1_4PoolEPNS_5sound7DecoderE,__ZN4love5audio6openal6SourceC2EPNS1_4PoolEPNS_5sound9SoundDataE,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZN4love7VariantC2EPKcj,b70,b70,b70,__ZN4love5touch3sdl5Touch7onEventEjRKNS0_5Touch9TouchInfoE,b70,b70,b70,b70,__ZN4love7VariantC2ENS_4TypeEPv,b70,b70,b70,b70,b70,b70,b70,_SDL_GL_GetDrawableSize,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZNSt3__16vectorINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEENS4_IS6_EEE6assignIPKS6_EENS_9enable_ifIXaasr21__is_forward_iteratorIT_EE5valuesr16is_constructibleIS6_NS_15iterator_traitsISD_E9referenceEEE5valueEvE4typeESD_SD_,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZNSt3__112__hash_tableINS_17__hash_value_typeIiN4love9StrongRefINS2_5image9ImageDataEEEEENS_22__unordered_map_hasherIiS7_NS_4hashIiEELb1EEENS_21__unordered_map_equalIiS7_NS_8equal_toIiEELb1EEENS_9allocatorIS7_EEE20__node_insert_uniqueEPNS_11__hash_nodeIS7_PvEE,b70
,b70,b70,b70,__ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEES7_EENS_22__unordered_map_hasherIS7_S8_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_S8_NS_8equal_toIS7_EELb1EEENS5_IS8_EEE20__node_insert_uniqueEPNS_11__hash_nodeIS8_PvEE,__ZNSt3__112__hash_tableINS_17__hash_value_typeIjN4love4font16BMFontRasterizer15BMFontCharacterEEENS_22__unordered_map_hasherIjS6_NS_4hashIjEELb1EEENS_21__unordered_map_equalIjS6_NS_8equal_toIjEELb1EEENS_9allocatorIS6_EEE20__node_insert_uniqueEPNS_11__hash_nodeIS6_PvEE,__ZNSt3__112__hash_tableINS_17__hash_value_typeIyiEENS_22__unordered_map_hasherIyS2_NS_4hashIyEELb1EEENS_21__unordered_map_equalIyS2_NS_8equal_toIyEELb1EEENS_9allocatorIS2_EEE20__node_insert_uniqueEPNS_11__hash_nodeIS2_PvEE,__ZN4love4font16BMFontRasterizerC2EPNS_10filesystem8FileDataERKNSt3__16vectorIPNS_5image9ImageDataENS5_9allocatorIS9_EEEE,b70,b70,b70,b70,b70,b70,__ZN4love8graphics6opengl6Canvas5drawvERKNS_7Matrix4EPKNS_6VertexE,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZNSt3__16vectorIPN4love8graphics6opengl6CanvasENS_9allocatorIS5_EEE6assignIPS5_EENS_9enable_ifIXaasr21__is_forward_iteratorIT_EE5valuesr16is_constructibleIS5_NS_15iterator_traitsISC_E9referenceEEE5valueEvE4typeESC_SC_,b70,__ZN4love8graphics6opengl6OpenGL23updateTextureMemorySizeEjj,b70,b70,b70
,__ZNSt3__112__hash_tableINS_17__hash_value_typeIjN4love8graphics6opengl4Font5GlyphEEENS_22__unordered_map_hasherIjS7_NS_4hashIjEELb1EEENS_21__unordered_map_equalIjS7_NS_8equal_toIjEELb1EEENS_9allocatorIS7_EEE20__node_insert_uniqueEPNS_11__hash_nodeIS7_PvEE,b70,b70,b70,__ZNSt3__112__hash_tableINS_17__hash_value_typeIyfEENS_22__unordered_map_hasherIyS2_NS_4hashIyEELb1EEENS_21__unordered_map_equalIyS2_NS_8equal_toIyEELb1EEENS_9allocatorIS2_EEE20__node_insert_uniqueEPNS_11__hash_nodeIS2_PvEE,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZN4love8graphics6opengl4Font12drawVerticesERKNSt3__16vectorINS2_11DrawCommandENS3_9allocatorIS5_EEEEb,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,__ZN4love8graphics6opengl4FontC2EPNS_4font10RasterizerERKNS0_7Texture6FilterE,b70,b70,b70,b70,__ZN4love8graphics6opengl5ImageC2ERKNSt3__16vectorIPNS_5image9ImageDataENS3_9allocatorIS7_EEEERKNS2_5FlagsE,__ZN4love8graphics6opengl5ImageC2ERKNSt3__16vectorIPNS_5image19CompressedImageDataENS3_9allocatorIS7_EEEERKNS2_5FlagsE,b70,b70,__ZN4love8graphics6opengl14ParticleSystemC2EPNS0_7TextureEj,b70,b70,b70,b70,b70,b70,__ZN4love8graphics6opengl4TextC2EPNS1_4FontERKNSt3__16vectorINS3_13ColoredStringENS5_9allocatorIS7_EEEE,b70,b70,b70,__ZN4love8graphics6opengl5Image5drawvERKNS_7Matrix4EPKNS_6VertexE,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl4Mesh17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE20__node_insert_uniqueEPNS_11__hash_nodeISD_PvEE,b70,__ZN4love8graphics6opengl8GLBuffer22setMappedRangeModifiedEjj,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZNK4love7Matrix4mlERKS0_,b70,b70,b70,b70,__ZNSt3__16vectorIiNS_9allocatorIiEEE8__appendEjRKi,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,__ZNSt3__112__hash_tableINS_17__hash_value_typeINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEN4love8graphics6opengl11SpriteBatch17AttachedAttributeEEENS_22__unordered_map_hasherIS7_SD_NS_4hashIS7_EELb1EEENS_21__unordered_map_equalIS7_SD_NS_8equal_toIS7_EELb1EEENS5_ISD_EEE20__node_insert_uniqueEPNS_11__hash_nodeISD_PvEE,b70,b70,__ZN4love8graphics6opengl4Text14uploadVerticesERKNSt3__16vectorINS1_4Font11GlyphVertexENS3_9allocatorIS6_EEEEj,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZN4love5image6magpie9ImageDataC2ENSt3__14listIPNS1_13FormatHandlerENS3_9allocatorIS6_EEEEPNS_10filesystem8FileDataE,b70,b70,__ZN4love5image6magpie19CompressedImageDataC2ENSt3__14listIPNS1_23CompressedFormatHandlerENS3_9allocatorIS6_EEEEPNS_10filesystem8FileDataE,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,__ZL16stbi__idct_blockPhiPs,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZNK4love8joystick3sdl14JoystickModule22stringFromGamepadInputENS0_8Joystick12GamepadInputE,__ZNK4love8joystick3sdl14JoystickModule23removeBindFromMapStringERNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEERKS9_,b70,b70,b70,__ZNK4love8joystick3sdl14JoystickModule23JoystickInputFromStringERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZNK12b2ChainShape12GetChildEdgeEP11b2EdgeShapei,b70,b70,b70,__ZN4love7physics5box2d5ShapeC2EP7b2Shapeb,b70,__ZN15b2MotorJointDef10InitializeEP6b2BodyS1_,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZN3dds6ParserC2EPKvj,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,__ZNSt3__16vectorIN3dds5ImageENS_9allocatorIS2_EEE6assignIPS2_EENS_9enable_ifIXaasr21__is_forward_iteratorIT_EE5valuesr16is_constructibleIS2_NS_15iterator_traitsIS9_E9referenceEEE5valueEvE4typeES9_S9_,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,_enumFilesCallback,b70,b70,b70,b70,_zip_entry_swap,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,_oc_frag_copy_c,_oc_frag_recon_intra_c,b70,b70,b70,b70,b70,b70,b70,b70,_destroy_size,_destroy_face,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,_t1_hints_stem,_ps_hints_t1stem3,b70,b70,b70,b70
,b70,_ps_hints_t2counter,b70,_Vertical_Sweep_Init,b70,b70,b70,_Horizontal_Sweep_Init,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,_Write_CVT,_Write_CVT_Stretched,_Move_CVT,_Move_CVT_Stretched,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,_Emscripten_WarpMouse,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,_emscripten_glGetTexParameterfv__wrapper,_emscripten_glGetTexParameteriv__wrapper,_emscripten_glTexParameterfv__wrapper,_emscripten_glTexParameteriv__wrapper,b70,b70,b70,_emscripten_glGetBufferParameteriv__wrapper,b70,b70,b70,b70,b70,b70,_emscripten_glGetRenderbufferParameteriv__wrapper,b70,_emscripten_glGetUniformfv__wrapper,_emscripten_glGetUniformiv__wrapper,b70,_emscripten_glGetVertexAttribfv__wrapper,_emscripten_glGetVertexAttribiv__wrapper,_emscripten_glGetVertexAttribPointerv__wrapper,b70,b70,b70,b70,b70,b70,_emscripten_glUniform2i__wrapper,b70,b70,_emscripten_glUniform1iv__wrapper,_emscripten_glUniform2iv__wrapper,_emscripten_glUniform3iv__wrapper,_emscripten_glUniform4iv__wrapper,_emscripten_glUniform1fv__wrapper,_emscripten_glUniform2fv__wrapper,_emscripten_glUniform3fv__wrapper,_emscripten_glUniform4fv__wrapper,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,_emscripten_glGetShaderiv__wrapper,_emscripten_glGetProgramiv__wrapper,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,_emscripten_glBindAttribLocation__wrapper,b70,b70,b70,b70,b70
,b70,b70,b70,_emscripten_glGetObjectParameterivARB__wrapper,b70,b70,b70,b70,b70,b70,b70,_emscripten_glNormalPointer__wrapper,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,_emscripten_glDrawArrays__wrapper,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,_emscripten_glTexParameteri__wrapper,b70,b70,_emscripten_glStencilFunc__wrapper,_emscripten_glStencilOp__wrapper,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZNSt3__18ios_base7failureC2EPKcRKNS_10error_codeE
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,__ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6__initEPKwj,b70,b70,b70,b70,b70,b70,b70,__ZNSt3__119__double_or_nothingIcEEvRNS_10unique_ptrIT_PFvPvEEERPS2_S9_,__ZNSt3__119__double_or_nothingIjEEvRNS_10unique_ptrIT_PFvPvEEERPS2_S9_,b70,b70,b70,b70,__ZNSt3__119__double_or_nothingIwEEvRNS_10unique_ptrIT_PFvPvEEERPS2_S9_,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,__ZNSt3__112system_error6__initERKNS_10error_codeENS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEE,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70,b70
,b70,b70,b70];
var FUNCTION_TABLE_v = [b71,b71,b71,b71,b71,___cxa_pure_virtual__wrapper,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,_abort__wrapper,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,__ZL25default_terminate_handlerv,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,___cxa_end_catch__wrapper,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,_enet_deinitialize,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,___cxa_rethrow__wrapper,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,__ZN4love5sound7lullaby13Mpg123Decoder4quitEv,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,_oc_restore_fpu_c,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,_DISKAUD_DetectDevices,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,_SDL_AudioDetectDevices_Default,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,_SDL_AudioDeinitialize_Default,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,_emscripten_glLoadIdentity__wrapper,b71,b71,b71,b71,b71,b71,b71,b71,b71,_emscripten_glReleaseShaderCompiler__wrapper,b71
,b71,b71,b71,_emscripten_glFinish__wrapper,_emscripten_glFlush__wrapper,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,__ZSt17__throw_bad_allocv,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,__ZN10__cxxabiv112_GLOBAL__N_110construct_Ev,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71,b71
,b71,b71,b71];
var FUNCTION_TABLE_viid = [b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,__ZNK12b2ChainShape11ComputeMassEP10b2MassDataf,b72,b72,b72,b72,b72,b72,b72,__ZNK13b2CircleShape11ComputeMassEP10b2MassDataf,b72,b72,b72,b72,b72,b72,__ZNK11b2EdgeShape11ComputeMassEP10b2MassDataf,b72,b72,b72,b72
,b72,b72,__ZNK14b2PolygonShape11ComputeMassEP10b2MassDataf,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,__ZNK12b2MotorJoint16GetReactionForceEf,b72,b72,b72,b72,b72,b72,b72
,b72,__ZNK12b2MouseJoint16GetReactionForceEf,b72,b72,b72,b72,b72,b72,b72,b72,b72,__ZNK16b2PrismaticJoint16GetReactionForceEf,b72,b72,b72,b72,b72,b72,b72,b72,__ZNK13b2PulleyJoint16GetReactionForceEf,b72,b72,b72,b72,b72,b72,b72,b72,b72
,__ZNK15b2RevoluteJoint16GetReactionForceEf,b72,b72,b72,b72,b72,b72,b72,b72,__ZNK11b2RopeJoint16GetReactionForceEf,b72,b72,b72,b72,b72,b72,b72,b72,__ZNK11b2WeldJoint16GetReactionForceEf,b72,b72,b72,b72,b72,b72,b72,b72,__ZNK12b2WheelJoint16GetReactionForceEf,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,__ZNK15b2DistanceJoint16GetReactionForceEf,b72,b72,b72,b72,b72,b72,b72,b72,__ZNK15b2FrictionJoint16GetReactionForceEf,b72,b72,b72,b72,b72,b72,b72,b72,__ZNK11b2GearJoint16GetReactionForceEf,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,__ZN4love5sound9SoundData9setSampleEif,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,__ZNK4love4math11BezierCurve8evaluateEd,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,_emscripten_glTexParameterf__wrapper,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72,b72
,b72,b72,b72];
var FUNCTION_TABLE_vd = [b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,_emscripten_glClearDepth__wrapper,_emscripten_glClearDepthf__wrapper,b73,b73,b73,b73,b73,b73,_emscripten_glLineWidth__wrapper,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73,b73
,b73,b73,b73];
var FUNCTION_TABLE_vi = [b74,__ZN4love6ModuleD2Ev,__ZN4love6ModuleD0Ev,__ZN4love6Object6retainEv,__ZN4love6Object7releaseEv,b74,__ZNSt3__118basic_stringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED1Ev,__ZNSt3__118basic_stringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED0Ev,__ZThn8_NSt3__118basic_stringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED1Ev,__ZThn8_NSt3__118basic_stringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED0Ev,__ZTv0_n12_NSt3__118basic_stringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED1Ev,__ZTv0_n12_NSt3__118basic_stringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED0Ev,__ZNSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev,__ZNSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEED0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__114basic_iostreamIcNS_11char_traitsIcEEED1Ev,__ZNSt3__114basic_iostreamIcNS_11char_traitsIcEEED0Ev,__ZThn8_NSt3__114basic_iostreamIcNS_11char_traitsIcEEED1Ev
,__ZThn8_NSt3__114basic_iostreamIcNS_11char_traitsIcEEED0Ev,__ZTv0_n12_NSt3__114basic_iostreamIcNS_11char_traitsIcEEED1Ev,__ZTv0_n12_NSt3__114basic_iostreamIcNS_11char_traitsIcEEED0Ev,__ZNSt3__113basic_istreamIcNS_11char_traitsIcEEED1Ev,__ZNSt3__113basic_istreamIcNS_11char_traitsIcEEED0Ev,__ZTv0_n12_NSt3__113basic_istreamIcNS_11char_traitsIcEEED1Ev,__ZTv0_n12_NSt3__113basic_istreamIcNS_11char_traitsIcEEED0Ev,__ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEED1Ev,__ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEED0Ev,__ZTv0_n12_NSt3__113basic_ostreamIcNS_11char_traitsIcEEED1Ev,__ZTv0_n12_NSt3__113basic_ostreamIcNS_11char_traitsIcEEED0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,_free,b74,b74,b74,b74,b74,b74,__ZN4love9ExceptionD2Ev,__ZN4love9ExceptionD0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love10filesystem8FileDataD2Ev,__ZN4love10filesystem8FileDataD0Ev,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,__ZN4love8graphics14ParticleSystemD2Ev,__ZN4love8graphics14ParticleSystemD0Ev,b74,__ZN4love8graphics4QuadD2Ev,__ZN4love8graphics4QuadD0Ev,__ZN4love8graphics7TextureD2Ev,__ZN4love8graphics7TextureD0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love5image19CompressedImageDataD2Ev,__ZN4love5image19CompressedImageDataD0Ev
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love5image3__38__invokeEPNS_5ProxyE,__ZN4love5image3__48__invokeEPNS_5ProxyE,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love4math14CompressedDataD2Ev,__ZN4love4math14CompressedDataD0Ev,b74,b74,__ZN4love4math10CompressorD2Ev,__ZN4love4math13LZ4CompressorD0Ev,b74,b74,b74,__ZN4love4math14zlibCompressorD0Ev,b74,b74,b74,__ZN4love4math4MathD2Ev,__ZN4love4math4MathD0Ev,b74,b74,__ZN4love6ObjectD2Ev,__ZN4love4math15RandomGeneratorD0Ev,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love5mouse3sdl5MouseD2Ev,__ZN4love5mouse3sdl5MouseD0Ev,b74,b74,b74,b74,b74,__ZN4love5mouse3sdl5Mouse9setCursorEv,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,__ZN4love7physics5box2d5JointD2Ev,__ZN4love7physics5box2d5JointD0Ev,__ZN4love7physics5box2d7PhysicsD0Ev,b74,b74,__ZN4love7physics5box2d12PolygonShapeD2Ev,__ZN4love7physics5box2d12PolygonShapeD0Ev,__ZN4love7physics5box2d14PrismaticJointD2Ev,__ZN4love7physics5box2d14PrismaticJointD0Ev,__ZN4love7physics5box2d11PulleyJointD2Ev,__ZN4love7physics5box2d11PulleyJointD0Ev,__ZN4love7physics5box2d13RevoluteJointD2Ev,__ZN4love7physics5box2d13RevoluteJointD0Ev,__ZN4love7physics5box2d9RopeJointD2Ev,__ZN4love7physics5box2d9RopeJointD0Ev,__ZN4love7physics5box2d5ShapeD2Ev,__ZN4love7physics5box2d5ShapeD0Ev,__ZN4love7physics5box2d9WeldJointD2Ev,__ZN4love7physics5box2d9WeldJointD0Ev,__ZN4love7physics5box2d10WheelJointD2Ev,__ZN4love7physics5box2d10WheelJointD0Ev,__ZN4love7physics5box2d5World13QueryCallbackD2Ev,__ZN4love7physics5box2d5World13QueryCallbackD0Ev
,b74,__ZN4love7physics5box2d5World15RayCastCallbackD2Ev,__ZN4love7physics5box2d5World15RayCastCallbackD0Ev,b74,__ZN4love7physics5box2d5WorldD2Ev,__ZN4love7physics5box2d5WorldD0Ev,b74,b74,b74,b74,b74,b74,b74,__ZThn8_N4love7physics5box2d5WorldD1Ev,__ZThn8_N4love7physics5box2d5WorldD0Ev,b74,b74,b74,b74,__ZThn12_N4love7physics5box2d5WorldD1Ev,__ZThn12_N4love7physics5box2d5WorldD0Ev,b74,__ZThn16_N4love7physics5box2d5WorldD1Ev,__ZThn16_N4love7physics5box2d5WorldD0Ev,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love5sound7lullaby5SoundD2Ev
,__ZN4love5sound7lullaby5SoundD0Ev,b74,b74,b74,__ZN4love5sound7lullaby13VorbisDecoderD2Ev,__ZN4love5sound7lullaby13VorbisDecoderD0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love5sound7lullaby11WaveDecoderD2Ev,__ZN4love5sound7lullaby11WaveDecoderD0Ev,b74,b74,b74,b74,b74,b74,b74
,b74,b74,__ZN4love5sound9SoundDataD2Ev,__ZN4love5sound9SoundDataD0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love6system3sdl6SystemD0Ev,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love6system6SystemD0Ev,__ZN4love6thread12ThreadModuleD0Ev,b74,b74,b74,b74
,b74,__ZN4love6thread10ThreadableD2Ev,__ZN4love6thread10ThreadableD0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love5timer3sdl5TimerD2Ev,__ZN4love5timer3sdl5TimerD0Ev,b74,b74,__ZN4love5timer5Timer4stepEv,b74,b74,b74,b74,b74,__ZN4love5timer5TimerD0Ev,__ZN4love5touch3sdl5TouchD2Ev,__ZN4love5touch3sdl5TouchD0Ev,b74,b74
,b74,b74,__ZN4love5video6theora5VideoD2Ev,__ZN4love5video6theora5VideoD0Ev,b74,b74,b74,__ZN4love5video6theora6WorkerD2Ev,__ZN4love5video6theora6WorkerD0Ev,__ZN4love5video6theora6Worker14threadFunctionEv,__ZN4love5video6theora11VideoStreamD2Ev,__ZN4love5video6theora11VideoStreamD0Ev,__ZN4love5video6theora11VideoStream14fillBackBufferEv,b74,b74,b74,b74,b74,b74,__ZN4love5video11VideoStream4playEv,__ZN4love5video11VideoStream5pauseEv,b74,b74,b74,b74,b74,__ZN4love5video11VideoStream9DeltaSyncD2Ev,__ZN4love5video11VideoStream9DeltaSyncD0Ev,b74,b74
,__ZN4love5video11VideoStream9DeltaSync4playEv,__ZN4love5video11VideoStream9DeltaSync5pauseEv,b74,b74,b74,__ZN4love5video11VideoStream10SourceSyncD2Ev,__ZN4love5video11VideoStream10SourceSyncD0Ev,b74,b74,__ZN4love5video11VideoStream10SourceSync4playEv,__ZN4love5video11VideoStream10SourceSync5pauseEv,b74,b74,__ZN4love5video11VideoStreamD2Ev,__ZN4love5video11VideoStreamD0Ev,__ZN4love6Stream14fillBackBufferEv,b74,b74,__ZN4love6window3sdl6WindowD2Ev,__ZN4love6window3sdl6WindowD0Ev,b74,b74,b74,b74,__ZN4love6window3sdl6Window5closeEv,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love6window3sdl6Window8minimizeEv,__ZN4love6window3sdl6Window8maximizeEv,__ZN4love6window3sdl6Window11swapBuffersEv,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,__ZN12b2ChainShapeD2Ev,__ZN12b2ChainShapeD0Ev,b74,b74,b74,b74,b74,b74,__ZN7b2ShapeD2Ev,__ZN13b2CircleShapeD0Ev,b74,b74,b74,b74,b74,b74,__ZN11b2EdgeShapeD0Ev,b74,b74,b74,b74,b74,b74,__ZN14b2PolygonShapeD0Ev,b74,b74,b74
,b74,b74,b74,__ZN15b2ContactFilterD2Ev,__ZN15b2ContactFilterD0Ev,b74,__ZN9b2ContactD2Ev,__ZN9b2ContactD0Ev,b74,__ZN22b2EdgeAndCircleContactD0Ev,b74,__ZN23b2EdgeAndPolygonContactD0Ev,b74,__ZN25b2PolygonAndCircleContactD0Ev,b74,__ZN16b2PolygonContactD0Ev,__ZN7b2Joint4DumpEv,b74,__ZN7b2JointD2Ev,__ZN7b2JointD0Ev,b74,b74,b74,b74,__ZN12b2MotorJoint4DumpEv,__ZN12b2MotorJointD0Ev,b74,b74,b74,b74
,b74,b74,b74,__ZN12b2MouseJoint4DumpEv,b74,__ZN12b2MouseJointD0Ev,b74,b74,b74,b74,b74,b74,b74,__ZN16b2PrismaticJoint4DumpEv,__ZN16b2PrismaticJointD0Ev,b74,b74,b74,b74,b74,b74,b74,__ZN13b2PulleyJoint4DumpEv,b74,__ZN13b2PulleyJointD0Ev,b74,b74,b74,b74,b74
,b74,b74,__ZN15b2RevoluteJoint4DumpEv,__ZN15b2RevoluteJointD0Ev,b74,b74,b74,b74,b74,b74,b74,__ZN11b2RopeJoint4DumpEv,__ZN11b2RopeJointD0Ev,b74,b74,b74,b74,b74,b74,b74,__ZN11b2WeldJoint4DumpEv,__ZN11b2WeldJointD0Ev,b74,b74,b74,b74,b74,b74,b74,__ZN12b2WheelJoint4DumpEv
,__ZN12b2WheelJointD0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love5sound7lullaby13Mpg123DecoderD2Ev,__ZN4love5sound7lullaby13Mpg123DecoderD0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love9ReferenceD2Ev,__ZN4love9ReferenceD0Ev,__ZN4love7VariantD2Ev,__ZN4love7VariantD0Ev,__ZN4love5audio4null5AudioD2Ev,__ZN4love5audio4null5AudioD0Ev,b74,b74,b74,b74
,b74,b74,__ZN4love5audio4null5Audio6updateEv,b74,b74,__ZN4love5audio4null5Audio4stopEv,b74,__ZN4love5audio4null5Audio5pauseEv,b74,__ZN4love5audio4null5Audio6resumeEv,b74,__ZN4love5audio4null5Audio6rewindEv,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love5audio4null5Audio6recordEv,b74,b74,b74,b74,b74,__ZN4love5audio4null6SourceD2Ev,__ZN4love5audio4null6SourceD0Ev
,b74,b74,__ZN4love5audio4null6Source4stopEv,__ZN4love5audio4null6Source5pauseEv,__ZN4love5audio4null6Source6resumeEv,__ZN4love5audio4null6Source6rewindEv,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love5audio6openal5Audio10PoolThreadD2Ev,__ZN4love5audio6openal5Audio10PoolThreadD0Ev,__ZN4love5audio6openal5Audio10PoolThread14threadFunctionEv,__ZN4love5audio6openal5AudioD2Ev,__ZN4love5audio6openal5AudioD0Ev,b74,b74,b74,b74,b74,__ZN4love5audio6openal5Audio6updateEv,b74,b74,__ZN4love5audio6openal5Audio4stopEv,b74,__ZN4love5audio6openal5Audio5pauseEv,b74,__ZN4love5audio6openal5Audio6resumeEv,b74
,__ZN4love5audio6openal5Audio6rewindEv,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love5audio6openal5Audio6recordEv,b74,b74,b74,b74,b74,__ZN4love5audio6openal16StaticDataBufferD2Ev,__ZN4love5audio6openal16StaticDataBufferD0Ev,__ZN4love5audio6openal6SourceD2Ev,__ZN4love5audio6openal6SourceD0Ev,b74,b74,__ZN4love5audio6openal6Source4stopEv,__ZN4love5audio6openal6Source5pauseEv,__ZN4love5audio6openal6Source6resumeEv,__ZN4love5audio6openal6Source6rewindEv,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,__ZN4love5audio6openal23SpatialSupportExceptionD0Ev,__ZN4love5audio6openal22InvalidFormatExceptionD0Ev,__ZN4love5audio6SourceD2Ev,__ZN4love5audio6SourceD0Ev,__ZN4love5event3sdl5EventD2Ev,__ZN4love5event3sdl5EventD0Ev,b74,b74,__ZN4love5event3sdl5Event5clearEv,__ZN4love5event3sdl5Event4pumpEv,b74,__ZN4love5event7MessageD2Ev,__ZN4love5event7MessageD0Ev,__ZN4love5event5EventD2Ev,__ZN4love5event5EventD0Ev,__ZN4love5event5Event5clearEv,__ZN4love10filesystem6physfs10FilesystemD2Ev,__ZN4love10filesystem6physfs10FilesystemD0Ev,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,__ZN4love10filesystem11DroppedFileD2Ev,__ZN4love10filesystem11DroppedFileD0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love10filesystem10FilesystemD2Ev,__ZN4love10filesystem10FilesystemD0Ev,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love4font8freetype4FontD2Ev,__ZN4love4font8freetype4FontD0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love4font8freetype18TrueTypeRasterizerD2Ev,__ZN4love4font8freetype18TrueTypeRasterizerD0Ev,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,__ZN4love4font16BMFontRasterizerD2Ev,__ZN4love4font16BMFontRasterizerD0Ev,b74,b74,b74,b74,b74,__ZN4love4font15DefaultFontDataD0Ev,b74,b74,__ZNSt9exceptionD2Ev,__ZN4utf815not_enough_roomD0Ev,b74,__ZN4utf812invalid_utf8D0Ev,b74,__ZN4utf818invalid_code_pointD0Ev,b74,__ZN4love4font9GlyphDataD2Ev,__ZN4love4font9GlyphDataD0Ev,b74,b74,b74,b74
,__ZN4love4font15ImageRasterizerD2Ev,__ZN4love4font15ImageRasterizerD0Ev,b74,b74,b74,b74,b74,__ZN4love8graphics6opengl6CanvasD2Ev,__ZN4love8graphics6opengl6CanvasD0Ev,b74,b74,b74,b74,b74,b74,__ZN4love8graphics6opengl6Canvas14unloadVolatileEv,__ZThn120_N4love8graphics6opengl6CanvasD1Ev,__ZThn120_N4love8graphics6opengl6CanvasD0Ev,b74,__ZThn120_N4love8graphics6opengl6Canvas14unloadVolatileEv,__ZN4love8graphics6opengl4FontD2Ev,__ZN4love8graphics6opengl4FontD0Ev,b74,__ZN4love8graphics6opengl4Font14unloadVolatileEv,__ZThn8_N4love8graphics6opengl4FontD1Ev,__ZThn8_N4love8graphics6opengl4FontD0Ev,b74,__ZThn8_N4love8graphics6opengl4Font14unloadVolatileEv,__ZNSt3__119basic_istringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED1Ev,__ZNSt3__119basic_istringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED0Ev
,__ZTv0_n12_NSt3__119basic_istringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED1Ev,__ZTv0_n12_NSt3__119basic_istringstreamIcNS_11char_traitsIcEENS_9allocatorIcEEED0Ev,__ZN4love8graphics6opengl8GLBufferD2Ev,__ZN4love8graphics6opengl8GLBufferD0Ev,b74,__ZN4love8graphics6opengl8GLBuffer14unloadVolatileEv,__ZN4love8graphics6opengl8GraphicsD2Ev,__ZN4love8graphics6opengl8GraphicsD0Ev,b74,b74,b74,b74,__ZN4love8graphics6opengl8Graphics9unSetModeEv,b74,b74,__ZN4love8graphics6opengl5ImageD2Ev,__ZN4love8graphics6opengl5ImageD0Ev,b74,b74,b74,b74,b74,b74,__ZN4love8graphics6opengl5Image14unloadVolatileEv,__ZThn120_N4love8graphics6opengl5ImageD1Ev,__ZThn120_N4love8graphics6opengl5ImageD0Ev,b74,__ZThn120_N4love8graphics6opengl5Image14unloadVolatileEv,__ZN4love8graphics6opengl4MeshD2Ev,__ZN4love8graphics6opengl4MeshD0Ev
,b74,__ZN4love8graphics6opengl14ParticleSystemD2Ev,__ZN4love8graphics6opengl14ParticleSystemD0Ev,b74,b74,b74,__ZN4love8graphics6opengl8PolylineD2Ev,__ZN4love8graphics6opengl8PolylineD0Ev,b74,b74,b74,__ZN4love8graphics6opengl16NoneJoinPolylineD0Ev,b74,b74,b74,b74,__ZN4love8graphics6opengl17MiterJoinPolylineD0Ev,b74,__ZN4love8graphics6opengl17BevelJoinPolylineD0Ev,b74,__ZN4love8graphics6opengl6ShaderD2Ev,__ZN4love8graphics6opengl6ShaderD0Ev,b74,__ZN4love8graphics6opengl6Shader14unloadVolatileEv,__ZThn8_N4love8graphics6opengl6ShaderD1Ev,__ZThn8_N4love8graphics6opengl6ShaderD0Ev,b74,__ZThn8_N4love8graphics6opengl6Shader14unloadVolatileEv,__ZN4love8graphics6opengl11SpriteBatchD2Ev,__ZN4love8graphics6opengl11SpriteBatchD0Ev
,b74,__ZN4love8graphics6opengl4TextD2Ev,__ZN4love8graphics6opengl4TextD0Ev,b74,__ZN4love8graphics6opengl5VideoD2Ev,__ZN4love8graphics6opengl5VideoD0Ev,b74,b74,__ZN4love8graphics6opengl5Video14unloadVolatileEv,__ZThn8_N4love8graphics6opengl5VideoD1Ev,__ZThn8_N4love8graphics6opengl5VideoD0Ev,b74,__ZThn8_N4love8graphics6opengl5Video14unloadVolatileEv,__ZN4love8graphics8DrawableD0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,__ZN4love8graphics8VolatileD2Ev,__ZN4love8graphics8VolatileD0Ev,__ZN4love5image6magpie5ImageD2Ev,__ZN4love5image6magpie5ImageD0Ev,b74,b74,b74,b74,b74,b74,b74,__ZN4love5image6magpie9ImageDataD2Ev,__ZN4love5image6magpie9ImageDataD0Ev,b74,b74,b74,__ZN4love5image6magpie10KTXHandlerD0Ev,b74,b74,__ZN4love5image6magpie10PKMHandlerD0Ev,b74,b74,__ZN4love5image6magpie13FormatHandlerD2Ev,__ZN4love5image6magpie10PNGHandlerD0Ev,b74,b74,b74,b74,b74,__ZN4love5image6magpie10PVRHandlerD0Ev
,b74,b74,__ZN4love5image6magpie10STBHandlerD0Ev,b74,b74,b74,b74,b74,__ZN4love5image9ImageDataD2Ev,__ZN4love5image9ImageDataD0Ev,__ZN4love8joystick3sdl8JoystickD2Ev,__ZN4love8joystick3sdl8JoystickD0Ev,b74,__ZN4love8joystick3sdl8Joystick5closeEv,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,__ZN4love8joystick3sdl14JoystickModuleD2Ev,__ZN4love8joystick3sdl14JoystickModuleD0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,__ZN4love8keyboard3sdl8KeyboardD0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love4math11BezierCurveD2Ev,__ZN4love4math11BezierCurveD0Ev,__ZN4love5mouse3sdl6CursorD2Ev,__ZN4love5mouse3sdl6CursorD0Ev,b74,b74,b74,__ZN4love7physics5box2d4BodyD2Ev,__ZN4love7physics5box2d4BodyD0Ev,__ZN4love7physics5box2d10ChainShapeD2Ev
,__ZN4love7physics5box2d10ChainShapeD0Ev,__ZN4love7physics5box2d11CircleShapeD2Ev,__ZN4love7physics5box2d11CircleShapeD0Ev,__ZN4love7physics5box2d7ContactD2Ev,__ZN4love7physics5box2d7ContactD0Ev,__ZN4love7physics5box2d13DistanceJointD2Ev,__ZN4love7physics5box2d13DistanceJointD0Ev,__ZN4love7physics5box2d9EdgeShapeD2Ev,__ZN4love7physics5box2d9EdgeShapeD0Ev,__ZN4love7physics5box2d7FixtureD2Ev,__ZN4love7physics5box2d7FixtureD0Ev,__ZN4love7physics5box2d13FrictionJointD2Ev,__ZN4love7physics5box2d13FrictionJointD0Ev,__ZN4love7physics5box2d9GearJointD2Ev,__ZN4love7physics5box2d9GearJointD0Ev,__ZN4love7physics5box2d10MotorJointD2Ev,__ZN4love7physics5box2d10MotorJointD0Ev,__ZN4love7physics5box2d10MouseJointD2Ev,__ZN4love7physics5box2d10MouseJointD0Ev,__ZN4love5sound7lullaby7DecoderD2Ev,__ZN4love5sound7lullaby7DecoderD0Ev,__ZN4love5sound7lullaby14ModPlugDecoderD2Ev,__ZN4love5sound7lullaby14ModPlugDecoderD0Ev,b74,b74,b74,b74,b74,b74,b74
,b74,__ZN4love6thread3sdl5MutexD2Ev,__ZN4love6thread3sdl5MutexD0Ev,__ZN4love6thread3sdl5Mutex4lockEv,__ZN4love6thread3sdl5Mutex6unlockEv,__ZN4love6thread3sdl11ConditionalD2Ev,__ZN4love6thread3sdl11ConditionalD0Ev,__ZN4love6thread3sdl11Conditional6signalEv,__ZN4love6thread3sdl11Conditional9broadcastEv,b74,__ZN4love6thread7ChannelD2Ev,__ZN4love6thread7ChannelD0Ev,__ZN4love6thread7Channel6retainEv,__ZN4love6thread7Channel7releaseEv,__ZN4love6thread9LuaThreadD2Ev,__ZN4love6thread9LuaThreadD0Ev,__ZN4love6thread9LuaThread14threadFunctionEv,__ZThn8_N4love6thread9LuaThreadD1Ev,__ZThn8_N4love6thread9LuaThreadD0Ev,__ZThn8_N4love6thread9LuaThread14threadFunctionEv,__ZN17b2ContactListenerD2Ev,__ZN17b2ContactListenerD0Ev,b74,b74,b74,b74,b74,__ZN23b2ChainAndCircleContactD0Ev,b74,__ZN24b2ChainAndPolygonContactD0Ev
,b74,__ZN15b2CircleContactD0Ev,b74,b74,b74,b74,__ZN15b2DistanceJoint4DumpEv,__ZN15b2DistanceJointD0Ev,b74,b74,b74,b74,b74,b74,b74,__ZN15b2FrictionJoint4DumpEv,__ZN15b2FrictionJointD0Ev,b74,b74,b74,b74,b74,b74,b74,__ZN11b2GearJoint4DumpEv,__ZN11b2GearJointD0Ev,b74,b74,b74,__ZN7lodepng5StateD2Ev
,__ZN7lodepng5StateD0Ev,b74,b74,b74,b74,b74,__ZN4love10filesystem6physfs4FileD2Ev,__ZN4love10filesystem6physfs4FileD0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love5image6magpie11ASTCHandlerD0Ev,b74,b74,__ZN4love5image6magpie19CompressedImageDataD2Ev,__ZN4love5image6magpie19CompressedImageDataD0Ev,__ZN4love5image6magpie10DDSHandlerD0Ev,b74,b74
,__ZN4love5image6magpie13FormatHandlerD0Ev,b74,b74,b74,b74,b74,__ZN4love6thread3sdl6ThreadD2Ev,__ZN4love6thread3sdl6ThreadD0Ev,b74,__ZN4love6thread3sdl6Thread4waitEv,b74,b74,b74,b74,b74,_bad_close,b74,b74,b74,b74,b74,b74,b74,b74,_bad_rewind,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_stream_close,b74,b74,b74,b74,b74,b74,b74,b74,_stream_rewind,b74,b74,b74,b74,b74
,b74,_buffered_forget,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_DIR_dirClose,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,_ZIP_dirClose,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_ft_bitmap_glyph_done,b74,b74,b74,_ft_outline_glyph_done,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,_BDF_Face_Done,b74,b74,b74,b74,b74,b74,_bdf_cmap_done,b74,b74,b74,_cff_driver_done,b74,b74,_cff_face_done,b74,_cff_size_done,b74,_cff_slot_done,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,_cff_cmap_encoding_done,b74,b74,b74,_cff_cmap_unicode_done,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_cid_driver_done,b74,b74,_cid_face_done,b74
,_cid_size_done,b74,_cid_slot_done,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_PCF_Face_Done,b74,b74,b74,b74,b74,b74,_pcf_cmap_done,b74,b74,b74,_pfr_cmap_done,b74,b74
,b74,b74,_pfr_face_done,b74,_pfr_slot_done,b74,b74,b74,b74,b74,b74,_ps_table_done,b74,_ps_table_release,b74,_ps_parser_done,_ps_parser_skip_spaces,_ps_parser_skip_PS_token,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_t1_builder_done,b74
,b74,b74,b74,b74,_t1_builder_close_contour,b74,_t1_decoder_done,b74,b74,_afm_parser_done,b74,b74,_t1_cmap_std_done,b74,b74,b74,b74,_t1_cmap_custom_done,b74,b74,b74,_t1_cmap_unicode_done,b74,b74,b74,b74,_ps_hinter_done,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_ft_black_done,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_tt_cmap14_done,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,_sfnt_done_face,b74,b74,b74,b74,b74,b74,b74,b74,_tt_face_free_name,b74,b74,b74,b74,b74,b74,_tt_face_free_ps_names,b74,b74,b74,b74,_tt_face_free_sbit
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_gray_raster_done,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_tt_driver_done,b74
,b74,_tt_face_done,b74,_tt_size_done,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_T1_Driver_Done,b74,b74,_T1_Face_Done,b74,_T1_Size_Done,b74,_T1_GlyphSlot_Done,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_T42_Driver_Done,b74,b74,_T42_Face_Done,b74
,_T42_Size_Done,b74,_T42_GlyphSlot_Done,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_FNT_Face_Done,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_af_autofitter_done,b74,b74,b74,b74,b74,_floor0_free_info,_floor0_free_look,b74,b74,b74,b74,b74,_floor1_free_info,_floor1_free_look,b74,b74
,b74,b74,_mapping0_free_info,b74,b74,b74,b74,_res0_free_info,_res0_free_look,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_BlitBto1,_BlitBto2,_BlitBto3,_BlitBto4,_BlitBto1Key,_BlitBto2Key,_BlitBto3Key
,_BlitBto4Key,_Blit1to1,_Blit1to2,_Blit1to3,_Blit1to4,_Blit1to1Key,_Blit1to2Key,_Blit1to3Key,_Blit1to4Key,_BlitNtoN,_Blit_RGB565_ARGB8888,_Blit_RGB565_ABGR8888,_Blit_RGB565_RGBA8888,_Blit_RGB565_BGRA8888,_Blit_RGB888_RGB565,_Blit_RGB888_RGB555,_SDL_Blit_RGB888_RGB888_Scale,_SDL_Blit_RGB888_RGB888_Blend,_SDL_Blit_RGB888_RGB888_Blend_Scale,_SDL_Blit_RGB888_RGB888_Modulate,_SDL_Blit_RGB888_RGB888_Modulate_Scale,_SDL_Blit_RGB888_RGB888_Modulate_Blend,_SDL_Blit_RGB888_RGB888_Modulate_Blend_Scale,_SDL_Blit_RGB888_BGR888_Scale,_SDL_Blit_RGB888_BGR888_Blend,_SDL_Blit_RGB888_BGR888_Blend_Scale,_SDL_Blit_RGB888_BGR888_Modulate,_SDL_Blit_RGB888_BGR888_Modulate_Scale,_SDL_Blit_RGB888_BGR888_Modulate_Blend,_SDL_Blit_RGB888_BGR888_Modulate_Blend_Scale
,_SDL_Blit_RGB888_ARGB8888_Scale,_SDL_Blit_RGB888_ARGB8888_Blend,_SDL_Blit_RGB888_ARGB8888_Blend_Scale,_SDL_Blit_RGB888_ARGB8888_Modulate,_SDL_Blit_RGB888_ARGB8888_Modulate_Scale,_SDL_Blit_RGB888_ARGB8888_Modulate_Blend,_SDL_Blit_RGB888_ARGB8888_Modulate_Blend_Scale,_SDL_Blit_BGR888_RGB888_Scale,_SDL_Blit_BGR888_RGB888_Blend,_SDL_Blit_BGR888_RGB888_Blend_Scale,_SDL_Blit_BGR888_RGB888_Modulate,_SDL_Blit_BGR888_RGB888_Modulate_Scale,_SDL_Blit_BGR888_RGB888_Modulate_Blend,_SDL_Blit_BGR888_RGB888_Modulate_Blend_Scale,_SDL_Blit_BGR888_BGR888_Scale,_SDL_Blit_BGR888_BGR888_Blend,_SDL_Blit_BGR888_BGR888_Blend_Scale,_SDL_Blit_BGR888_BGR888_Modulate,_SDL_Blit_BGR888_BGR888_Modulate_Scale,_SDL_Blit_BGR888_BGR888_Modulate_Blend,_SDL_Blit_BGR888_BGR888_Modulate_Blend_Scale,_SDL_Blit_BGR888_ARGB8888_Scale,_SDL_Blit_BGR888_ARGB8888_Blend,_SDL_Blit_BGR888_ARGB8888_Blend_Scale,_SDL_Blit_BGR888_ARGB8888_Modulate,_SDL_Blit_BGR888_ARGB8888_Modulate_Scale,_SDL_Blit_BGR888_ARGB8888_Modulate_Blend,_SDL_Blit_BGR888_ARGB8888_Modulate_Blend_Scale,_SDL_Blit_ARGB8888_RGB888_Scale,_SDL_Blit_ARGB8888_RGB888_Blend
,_SDL_Blit_ARGB8888_RGB888_Blend_Scale,_SDL_Blit_ARGB8888_RGB888_Modulate,_SDL_Blit_ARGB8888_RGB888_Modulate_Scale,_SDL_Blit_ARGB8888_RGB888_Modulate_Blend,_SDL_Blit_ARGB8888_RGB888_Modulate_Blend_Scale,_SDL_Blit_ARGB8888_BGR888_Scale,_SDL_Blit_ARGB8888_BGR888_Blend,_SDL_Blit_ARGB8888_BGR888_Blend_Scale,_SDL_Blit_ARGB8888_BGR888_Modulate,_SDL_Blit_ARGB8888_BGR888_Modulate_Scale,_SDL_Blit_ARGB8888_BGR888_Modulate_Blend,_SDL_Blit_ARGB8888_BGR888_Modulate_Blend_Scale,_SDL_Blit_ARGB8888_ARGB8888_Scale,_SDL_Blit_ARGB8888_ARGB8888_Blend,_SDL_Blit_ARGB8888_ARGB8888_Blend_Scale,_SDL_Blit_ARGB8888_ARGB8888_Modulate,_SDL_Blit_ARGB8888_ARGB8888_Modulate_Scale,_SDL_Blit_ARGB8888_ARGB8888_Modulate_Blend,_SDL_Blit_ARGB8888_ARGB8888_Modulate_Blend_Scale,_SDL_Blit_RGBA8888_RGB888_Scale,_SDL_Blit_RGBA8888_RGB888_Blend,_SDL_Blit_RGBA8888_RGB888_Blend_Scale,_SDL_Blit_RGBA8888_RGB888_Modulate,_SDL_Blit_RGBA8888_RGB888_Modulate_Scale,_SDL_Blit_RGBA8888_RGB888_Modulate_Blend,_SDL_Blit_RGBA8888_RGB888_Modulate_Blend_Scale,_SDL_Blit_RGBA8888_BGR888_Scale,_SDL_Blit_RGBA8888_BGR888_Blend,_SDL_Blit_RGBA8888_BGR888_Blend_Scale,_SDL_Blit_RGBA8888_BGR888_Modulate
,_SDL_Blit_RGBA8888_BGR888_Modulate_Scale,_SDL_Blit_RGBA8888_BGR888_Modulate_Blend,_SDL_Blit_RGBA8888_BGR888_Modulate_Blend_Scale,_SDL_Blit_RGBA8888_ARGB8888_Scale,_SDL_Blit_RGBA8888_ARGB8888_Blend,_SDL_Blit_RGBA8888_ARGB8888_Blend_Scale,_SDL_Blit_RGBA8888_ARGB8888_Modulate,_SDL_Blit_RGBA8888_ARGB8888_Modulate_Scale,_SDL_Blit_RGBA8888_ARGB8888_Modulate_Blend,_SDL_Blit_RGBA8888_ARGB8888_Modulate_Blend_Scale,_SDL_Blit_ABGR8888_RGB888_Scale,_SDL_Blit_ABGR8888_RGB888_Blend,_SDL_Blit_ABGR8888_RGB888_Blend_Scale,_SDL_Blit_ABGR8888_RGB888_Modulate,_SDL_Blit_ABGR8888_RGB888_Modulate_Scale,_SDL_Blit_ABGR8888_RGB888_Modulate_Blend,_SDL_Blit_ABGR8888_RGB888_Modulate_Blend_Scale,_SDL_Blit_ABGR8888_BGR888_Scale,_SDL_Blit_ABGR8888_BGR888_Blend,_SDL_Blit_ABGR8888_BGR888_Blend_Scale,_SDL_Blit_ABGR8888_BGR888_Modulate,_SDL_Blit_ABGR8888_BGR888_Modulate_Scale,_SDL_Blit_ABGR8888_BGR888_Modulate_Blend,_SDL_Blit_ABGR8888_BGR888_Modulate_Blend_Scale,_SDL_Blit_ABGR8888_ARGB8888_Scale,_SDL_Blit_ABGR8888_ARGB8888_Blend,_SDL_Blit_ABGR8888_ARGB8888_Blend_Scale,_SDL_Blit_ABGR8888_ARGB8888_Modulate,_SDL_Blit_ABGR8888_ARGB8888_Modulate_Scale,_SDL_Blit_ABGR8888_ARGB8888_Modulate_Blend
,_SDL_Blit_ABGR8888_ARGB8888_Modulate_Blend_Scale,_SDL_Blit_BGRA8888_RGB888_Scale,_SDL_Blit_BGRA8888_RGB888_Blend,_SDL_Blit_BGRA8888_RGB888_Blend_Scale,_SDL_Blit_BGRA8888_RGB888_Modulate,_SDL_Blit_BGRA8888_RGB888_Modulate_Scale,_SDL_Blit_BGRA8888_RGB888_Modulate_Blend,_SDL_Blit_BGRA8888_RGB888_Modulate_Blend_Scale,_SDL_Blit_BGRA8888_BGR888_Scale,_SDL_Blit_BGRA8888_BGR888_Blend,_SDL_Blit_BGRA8888_BGR888_Blend_Scale,_SDL_Blit_BGRA8888_BGR888_Modulate,_SDL_Blit_BGRA8888_BGR888_Modulate_Scale,_SDL_Blit_BGRA8888_BGR888_Modulate_Blend,_SDL_Blit_BGRA8888_BGR888_Modulate_Blend_Scale,_SDL_Blit_BGRA8888_ARGB8888_Scale,_SDL_Blit_BGRA8888_ARGB8888_Blend,_SDL_Blit_BGRA8888_ARGB8888_Blend_Scale,_SDL_Blit_BGRA8888_ARGB8888_Modulate,_SDL_Blit_BGRA8888_ARGB8888_Modulate_Scale,_SDL_Blit_BGRA8888_ARGB8888_Modulate_Blend,_SDL_Blit_BGRA8888_ARGB8888_Modulate_Blend_Scale,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED2Ev,__ZNSt3__110__stdinbufIcED0Ev,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__111__stdoutbufIcED0Ev,b74,b74,b74,b74,b74,b74,__ZNSt3__115basic_streambufIwNS_11char_traitsIwEEED2Ev,__ZNSt3__110__stdinbufIwED0Ev,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__111__stdoutbufIwED0Ev,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__18ios_baseD2Ev,__ZNSt3__18ios_baseD0Ev,__ZNSt3__115basic_streambufIcNS_11char_traitsIcEEED0Ev,__ZNSt3__115basic_streambufIwNS_11char_traitsIwEEED0Ev,b74,__ZNSt3__113basic_istreamIwNS_11char_traitsIwEEED1Ev,__ZNSt3__113basic_istreamIwNS_11char_traitsIwEEED0Ev,__ZTv0_n12_NSt3__113basic_istreamIwNS_11char_traitsIwEEED1Ev,__ZTv0_n12_NSt3__113basic_istreamIwNS_11char_traitsIwEEED0Ev,__ZNSt3__113basic_ostreamIwNS_11char_traitsIwEEED1Ev,__ZNSt3__113basic_ostreamIwNS_11char_traitsIwEEED0Ev,__ZTv0_n12_NSt3__113basic_ostreamIwNS_11char_traitsIwEEED1Ev,__ZTv0_n12_NSt3__113basic_ostreamIwNS_11char_traitsIwEEED0Ev
,__ZNSt3__18ios_base7failureD2Ev,__ZNSt3__18ios_base7failureD0Ev,b74,__ZNSt3__114error_categoryD2Ev,__ZNSt3__119__iostream_categoryD0Ev,b74,b74,b74,b74,b74,__ZNSt3__17collateIcED2Ev,__ZNSt3__17collateIcED0Ev,__ZNSt3__16locale5facet16__on_zero_sharedEv,b74,b74,b74,__ZNSt3__17collateIwED2Ev,__ZNSt3__17collateIwED0Ev,b74,b74,b74,__ZNSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev,__ZNSt3__17num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,__ZNSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev,__ZNSt3__17num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev,__ZNSt3__17num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev,b74,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev,__ZNSt3__17num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev,b74
,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev,__ZNSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev,__ZNSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__18time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev,__ZNSt3__18time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev,b74,__ZNSt3__18time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev,__ZNSt3__18time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev,b74,__ZNSt3__110moneypunctIcLb0EED2Ev,__ZNSt3__110moneypunctIcLb0EED0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__110moneypunctIcLb1EED2Ev,__ZNSt3__110moneypunctIcLb1EED0Ev,b74,b74
,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__110moneypunctIwLb0EED2Ev,__ZNSt3__110moneypunctIwLb0EED0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__110moneypunctIwLb1EED2Ev,__ZNSt3__110moneypunctIwLb1EED0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__19money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev
,__ZNSt3__19money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev,b74,b74,__ZNSt3__19money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev,__ZNSt3__19money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev,b74,b74,__ZNSt3__19money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev,__ZNSt3__19money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev,b74,b74,__ZNSt3__19money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev,__ZNSt3__19money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev,b74,b74,__ZNSt3__18messagesIcED2Ev,__ZNSt3__18messagesIcED0Ev,b74,b74,b74,__ZNSt3__18messagesIwED2Ev,__ZNSt3__18messagesIwED0Ev,b74,b74,b74,__ZNSt3__16locale5facetD2Ev,__ZNSt3__116__narrow_to_utf8ILj32EED0Ev,b74,b74,b74
,b74,b74,b74,b74,__ZNSt3__117__widen_from_utf8ILj32EED0Ev,__ZNSt3__17codecvtIwc11__mbstate_tED2Ev,__ZNSt3__17codecvtIwc11__mbstate_tED0Ev,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__16locale5__impD2Ev,__ZNSt3__16locale5__impD0Ev,__ZNSt3__15ctypeIcED2Ev,__ZNSt3__15ctypeIcED0Ev,b74,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__18numpunctIcED2Ev,__ZNSt3__18numpunctIcED0Ev,b74,b74
,b74,b74,b74,__ZNSt3__18numpunctIwED2Ev,__ZNSt3__18numpunctIwED0Ev,b74,b74,b74,b74,b74,__ZNSt3__16locale5facetD0Ev,__ZNSt3__15ctypeIwED0Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__17codecvtIcc11__mbstate_tED0Ev,b74,b74,b74,b74,b74
,b74,b74,__ZNSt3__17codecvtIDsc11__mbstate_tED0Ev,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__17codecvtIDic11__mbstate_tED0Ev,__ZNSt3__112system_errorD2Ev,__ZNSt3__112system_errorD0Ev,b74,__ZN10__cxxabiv116__shim_type_infoD2Ev,__ZN10__cxxabiv117__class_type_infoD0Ev,__ZNK10__cxxabiv116__shim_type_info5noop1Ev,__ZNK10__cxxabiv116__shim_type_info5noop2Ev,b74,b74,b74,b74,__ZN10__cxxabiv120__si_class_type_infoD0Ev,b74,b74,b74,__ZNSt9bad_allocD2Ev,__ZNSt9bad_allocD0Ev,b74,__ZNSt11logic_errorD2Ev
,__ZNSt11logic_errorD0Ev,b74,__ZNSt13runtime_errorD2Ev,__ZNSt13runtime_errorD0Ev,__ZNSt12length_errorD0Ev,__ZNSt12out_of_rangeD0Ev,__ZNSt14overflow_errorD0Ev,__ZNSt8bad_castD2Ev,__ZNSt8bad_castD0Ev,b74,__ZNSt10bad_typeidD2Ev,__ZNSt10bad_typeidD0Ev,b74,__ZN10__cxxabiv121__vmi_class_type_infoD0Ev,b74,b74,b74,b74,__ZZ4mainEN3__08__invokeEPv,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,__ZNSt3__115basic_streambufIcNS_11char_traitsIcEEEC2Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__18ios_base33__set_badbit_and_consider_rethrowEv,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love5mouse3sdl5MouseC2Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love7physics5box2d7Physics8setMeterEi,__ZN4love6ObjectC2Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love7physics5box2d5World7destroyEv,b74,b74,b74
,__ZN4love5sound7lullaby5SoundC2Ev,b74,b74,__ZN4love6system3sdl6SystemC2Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love5timer3sdl5TimerC2Ev,__ZN4love5video6theora5VideoC2Ev,b74,__ZN4love5video11VideoStream9DeltaSyncC2Ev,b74,b74,b74,b74,b74,b74,b74,__ZN4love6window3sdl6WindowC2Ev,b74,b74,b74
,b74,_enet_range_coder_destroy,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love5audio6openal5AudioC2Ev,b74,b74,__ZN4love5audio4null5AudioC2Ev,b74,b74,b74,b74,b74
,b74,__ZN4love5event3sdl5EventC2Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love10filesystem6physfs10FilesystemC2Ev,b74,b74,b74,b74,__ZN4love4font8freetype4FontC2Ev,b74,__ZNKSt3__120__vector_base_commonILb1EE20__throw_length_errorEv,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love8graphics6opengl4Mesh5flushEv,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,__ZN4love8graphics6opengl8Graphics9setCanvasEv,_lua_pushnil,b74,b74,b74,b74,b74,b74,__ZN4love8graphics6opengl8Graphics3popEv,__ZN4love8graphics6opengl8GraphicsC2Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love8graphics6opengl4Mesh12setVertexMapEv,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love8graphics6opengl4Text3setEv,b74,b74,b74,b74,__ZN4love8graphics6opengl4Text5clearEv,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love5image6magpie5ImageC2Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love8joystick3sdl14JoystickModuleC2Ev,b74,b74,b74,b74,__ZN4love8keyboard3sdl8KeyboardC2Ev,__ZN4love4math15RandomGeneratorC2Ev,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,_SDL_SetCursor,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love8Memoizer6removeEPv,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love7physics5box2d4Body13resetMassDataEv,b74
,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love7physics5box2d4Body7destroyEv,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,___cxx_global_array_dtor,b74,b74,b74,b74,___cxx_global_array_dtor_2219,b74,b74,b74,b74,_SDL_free,b74,__ZN4love6thread7ChannelC2Ev,b74
,b74,b74,b74,b74,b74,b74,b74,__ZN4love5video6theora6WorkerC2Ev,b74,__ZN4love6thread8MutexRefC2Ev,_th_decode_free,_th_info_clear,b74,b74,b74,_th_info_init,__ZN4love5video11VideoStream5FrameC2Ev,__ZN4love5video6theora11VideoStream11parseHeaderEv,_SDL_GL_DeleteContext,_SDL_DestroyWindow,_SDL_FlushEvent,_SDL_QuitSubSystem,b74,b74,b74,b74,__ZN4love8graphics15setGammaCorrectEb,b74,b74,b74
,b74,b74,b74,b74,b74,b74,_SDL_GetVersion,__Z6b2FreePv,__ZN16b2StackAllocatorC2Ev,__ZN16b2ContactManagerC2Ev,b74,b74,b74,b74,b74,__ZN7b2TimerC2Ev,__ZN16b2ContactManager15FindNewContactsEv,__ZN6b2Body19SynchronizeFixturesEv,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_mpg123_delete,b74,b74,b74,b74,b74,b74,__ZN4love5sound7lullabyL16cleanup_callbackEPv,b74,b74,b74,b74,___cxx_global_array_dtor_3008,__ZNSt3__13mapIPvS1_NS_4lessIS1_EENS_9allocatorINS_4pairIKS1_S1_EEEEED2Ev
,b74,b74,b74,b74,b74,b74,__ZN4love5audio4null6SourceC2Ev,__ZN4love6thread10Threadable4waitEv,b74,_alcDestroyContext__wrapper,b74,b74,b74,b74,b74,b74,b74,__ZN4love5audio6openal4PoolC2Ev,__ZN4love6thread10ThreadableC2Ev,b74,b74,__ZNSt3__15dequeIjNS_9allocatorIjEEE19__add_back_capacityEv,__ZN4love5audio6openal4Pool4stopEv,b74,__ZN4love5audio6openal6Source10stopAtomicEv,__ZN4love5audio6openal6Source12rewindAtomicEv,b74,b74,__ZN4love5audio6openal6Source11pauseAtomicEv,__ZN4love5audio6openal6Source12resumeAtomicEv
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__13mapIiN4love8keyboard8Keyboard3KeyENS_4lessIiEENS_9allocatorINS_4pairIKiS4_EEEEED2Ev
,b74,b74,__ZNSt3__15dequeIPN4love5event7MessageENS_9allocatorIS4_EEE19__add_back_capacityEv,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love4font15ImageRasterizer4loadEv,b74,b74,b74,b74,b74,b74,__ZN4love8graphics6opengl6OpenGL11prepareDrawEv,b74,__ZN4love8graphics6opengl6OpenGL12popTransformEv,b74,__ZN4love8graphics8VolatileC2Ev,b74,b74,b74,b74,b74,b74,__ZN4love8graphics6opengl4Font13createTextureEv
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,__ZN4love8graphics6opengl6OpenGLC2Ev,b74,b74,b74,b74,b74,__ZN4love8graphics6opengl8Graphics23stopDrawToStencilBufferEv,b74,__ZN4love7Matrix411setIdentityEv,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN4love8graphics6opengl8Polyline4drawEv,b74,__ZN4love8graphics6opengl5Image22loadFromCompressedDataEv
,__ZN4love8graphics6opengl5Image17loadFromImageDataEv,b74,b74,b74,b74,b74,b74,__ZN4love8graphics6opengl6OpenGL13pushTransformEv,__ZN4love8graphics6opengl8GLBuffer4bindEv,__ZN4love8graphics6opengl8GLBuffer5unmapEv,__ZN4love8graphics6opengl8GLBuffer6unbindEv,b74,b74,b74,__ZN4love8graphics6opengl4Mesh23setupAttachedAttributesEv,__ZN4love8graphics6opengl4Mesh23calculateAttributeSizesEv,b74,b74,b74,b74,__ZN4love8graphics6opengl6OpenGLD2Ev,__ZN4love7Matrix4C2Ev,b74,__ZN4love7Matrix3C2Ev,b74,b74,b74,b74,b74,b74
,b74,__ZNSt3__113__vector_baseIN4love6VectorENS_9allocatorIS2_EEED2Ev,b74,b74,b74,b74,__ZN4love8graphics6opengl6Shader17mapActiveUniformsEv,__ZN4love8graphics6opengl6Shader23checkSetBuiltinUniformsEv,b74,b74,b74,b74,b74,b74,b74,___cxx_global_array_dtor_3998,___cxx_global_array_dtor_2,__ZNSt3__113__vector_baseIiNS_9allocatorIiEEED2Ev,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,__ZN4love8graphics6opengl4Text18regenerateVerticesEv,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZNSt3__110__list_impIPN4love8graphics8VolatileENS_9allocatorIS4_EEED2Ev,b74,b74,b74,b74,__ZN4love5image6magpie13FormatHandlerC2Ev,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_SDL_HapticClose,_SDL_GameControllerClose,_SDL_JoystickClose,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,_SDL_FreeCursor,b74,_SDL_FreeSurface,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_ModPlug_Unload,_ModPlug_SetSettings,b74,b74,___cxx_global_array_dtor_4964,_SDL_DestroyMutex,_SDL_DestroyCond,b74,b74,b74
,b74,__ZNSt3__13mapINS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPN4love6thread7ChannelENS_4lessIS6_EENS4_INS_4pairIKS6_SA_EEEEED2Ev,b74,b74,b74,b74,__ZNSt3__15dequeIPN4love7VariantENS_9allocatorIS3_EEE19__add_back_capacityEv,b74,b74,b74,b74,__ZN15b2ContactSolver29InitializeVelocityConstraintsEv,__ZN15b2ContactSolver9WarmStartEv,__ZN15b2ContactSolver13StoreImpulsesEv,__ZN15b2ContactSolver24SolveVelocityConstraintsEv,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_INT123_make_decode_tables,b74,b74,b74,b74,b74,b74,b74,b74,b74,_mallocAllocatorFree,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_memory_stream_close,b74,b74,b74,b74,_ft_ansi_stream_close
,b74,b74,b74,b74,b74,b74,b74,b74,_cf2_free_instance,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_psh_globals_destroy,_t1_hints_open,b74,b74,b74,b74,b74,_t2_hints_open,b74
,b74,b74,b74,b74,b74,b74,_Vertical_Sweep_Step,b74,b74,b74,_Horizontal_Sweep_Step,_Split_Conic,_Split_Cubic,b74,_sfnt_stream_close,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_TT_Forget_Glyph_Frame
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,_af_face_globals_free,b74,b74,_ft_gzip_stream_close,b74,b74,b74,b74,b74,_ft_lzw_stream_close,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,_GLES2_RenderPresent,b74,_GLES2_DestroyRenderer,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_SDL_BlitCopy,_SDL_Blit_Slow,_BlitBtoNAlpha,_BlitBtoNAlphaKey,_Blit1toNAlpha,_Blit1toNAlphaKey,_BlitARGBto555PixelAlpha,_BlitNtoNPixelAlpha,_BlitARGBto565PixelAlpha
,_Blit555to555SurfaceAlpha,_BlitNtoNSurfaceAlpha,_Blit565to565SurfaceAlpha,_BlitNto1SurfaceAlphaKey,_BlitNtoNSurfaceAlphaKey,_BlitNto1PixelAlpha,_BlitRGBtoRGBPixelAlpha,_BlitNto1SurfaceAlpha,_BlitRGBtoRGBSurfaceAlpha,_Blit4to4CopyAlpha,_BlitNtoNCopyAlpha,_Blit_RGB888_index8,_Blit_RGB101010_index8,_BlitNto1,_Blit2101010toN,_BlitNto2101010,_Blit4to4MaskAlpha,_Blit2to2Key,_BlitNto1Key,_BlitNtoNKeyCopyAlpha,_BlitNtoNKey,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_SW_RenderPresent,b74,_SW_DestroyRenderer,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,_Emscripten_VideoQuit,b74,_Emscripten_PumpEvents,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_SDL_EGL_UnloadLibrary,b74,b74,b74,b74,b74,b74,b74,_Emscripten_DeleteDevice,b74,b74
,_Emscripten_FreeCursor,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,_Emscripten_CloseDevice,_HandleAudioProcess,b74,_DUMMY_VideoQuit,b74,_DUMMY_PumpEvents,b74,b74,b74,_DUMMY_DeleteDevice,b74,b74,b74,b74,b74,b74,_DISKAUD_WaitDevice,_DISKAUD_PlayDevice,b74,_DISKAUD_CloseDevice,b74,b74,_SDL_HandleSIG,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_SDL_AudioThreadInit_Default,_SDL_AudioWaitDevice_Default,_SDL_AudioPlayDevice_Default,b74,b74,_SDL_AudioWaitDone_Default,_SDL_AudioCloseDevice_Default,_SDL_AudioLockDevice_Default,_SDL_AudioUnlockDevice_Default,_SDL_AudioFreeDeviceHandle_Default,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,_emscripten_glDeleteShader__wrapper,b74,b74,b74,_emscripten_glCompileShader__wrapper,b74,b74,b74,b74,b74,_emscripten_glDeleteProgram__wrapper,b74,b74,b74,_emscripten_glLinkProgram__wrapper,b74,_emscripten_glUseProgram__wrapper,_emscripten_glValidateProgram__wrapper,b74,b74,b74,b74,b74,b74,b74
,b74,b74,_emscripten_glDeleteObjectARB__wrapper,b74,b74,b74,b74,b74,_emscripten_glEnableClientState__wrapper,b74,b74,b74,b74,_emscripten_glClientActiveTexture__wrapper,b74,b74,_emscripten_glBindVertexArray__wrapper,_emscripten_glMatrixMode__wrapper,b74,_emscripten_glLoadMatrixf__wrapper,b74,b74,b74,_emscripten_glEnableVertexAttribArray__wrapper,_emscripten_glDisableVertexAttribArray__wrapper,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,_emscripten_glDepthFunc__wrapper,_emscripten_glEnable__wrapper,_emscripten_glDisable__wrapper,_emscripten_glFrontFace__wrapper,_emscripten_glCullFace__wrapper,_emscripten_glClear__wrapper,b74,_emscripten_glClearStencil__wrapper,_emscripten_glDepthMask__wrapper,_emscripten_glStencilMask__wrapper,b74,_emscripten_glGenerateMipmap__wrapper,_emscripten_glActiveTexture__wrapper,_emscripten_glBlendEquation__wrapper,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,_cleanup_387,_cleanup_382,b74,__ZNSt3__18ios_base4InitD2Ev,b74,b74,b74,b74,b74,b74
,b74,b74,__ZNSt3__16locale2id6__initEv,__ZNSt3__117__call_once_proxyINS_5tupleIJNS_12_GLOBAL__N_111__fake_bindEEEEEEvPv,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,___cxx_global_array_dtor_10066,___cxx_global_array_dtor_69
,___cxx_global_array_dtor_117,___cxx_global_array_dtor_54,b74,___cxx_global_array_dtor_93,___cxx_global_array_dtor_120,b74,b74,__ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev,_freelocale,b74,__ZNSt3__112__do_nothingEPv,b74,__ZNSt3__121__throw_runtime_errorEPKc,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,__ZN10__cxxabiv112_GLOBAL__N_19destruct_EPv,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74,b74
,b74,b74,b74];
var FUNCTION_TABLE_viddddddddd = [b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,__ZN4love8graphics6opengl6Canvas4drawEfffffffff,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,__ZN4love8graphics6opengl5Image4drawEfffffffff,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,__ZN4love8graphics6opengl4Mesh4drawEfffffffff,b75,b75,__ZN4love8graphics6opengl14ParticleSystem4drawEfffffffff,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,__ZN4love8graphics6opengl11SpriteBatch4drawEfffffffff,b75,b75,__ZN4love8graphics6opengl4Text4drawEfffffffff,b75,b75,__ZN4love8graphics6opengl5Video4drawEfffffffff,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,__ZN4love7Matrix4C2Efffffffff,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,__ZN4love7Matrix317setTransformationEfffffffff,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,__ZN4love7Matrix3C2Efffffffff,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75,b75
,b75,b75,b75];
var FUNCTION_TABLE_ii = [b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE4syncEv,__ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE9showmanycEv,b76,__ZNSt3__115basic_stringbufIcNS_11char_traitsIcEENS_9allocatorIcEEE9underflowEv,__ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE5uflowEv,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_luaopen_love_audio,_luaopen_love_event,_luaopen_love_filesystem,_luaopen_love_font,_luaopen_love_graphics,_luaopen_love_image,_luaopen_love_joystick,_luaopen_love_keyboard,_luaopen_love_math,_luaopen_love_mouse,_luaopen_love_physics,_luaopen_love_sound,_luaopen_love_system,_luaopen_love_timer,_luaopen_love_thread,_luaopen_love_touch,_luaopen_love_video,_luaopen_love_window,_luaopen_love_nogame
,_luaopen_love_boot,__ZN4love4math21w__getRandomGeneratorEP9lua_State,__ZN4love4math20w_newRandomGeneratorEP9lua_State,__ZN4love4math16w_newBezierCurveEP9lua_State,__ZN4love4math13w_triangulateEP9lua_State,__ZN4love4math10w_isConvexEP9lua_State,__ZN4love4math15w_gammaToLinearEP9lua_State,__ZN4love4math15w_linearToGammaEP9lua_State,__ZN4love4math7w_noiseEP9lua_State,__ZN4love4math10w_compressEP9lua_State,__ZN4love4math12w_decompressEP9lua_State,_luaopen_randomgenerator,_luaopen_beziercurve,_luaopen_compresseddata,b76,b76,b76,b76,b76,b76,__ZN4love4math25w_RandomGenerator__randomEP9lua_State,__ZN4love4math30w_RandomGenerator_randomNormalEP9lua_State,__ZN4love4math25w_RandomGenerator_setSeedEP9lua_State,__ZN4love4math25w_RandomGenerator_getSeedEP9lua_State,__ZN4love4math26w_RandomGenerator_setStateEP9lua_State,__ZN4love4math26w_RandomGenerator_getStateEP9lua_State,b76,__ZN4love5mouse11w_newCursorEP9lua_State,__ZN4love5mouse17w_getSystemCursorEP9lua_State,__ZN4love5mouse11w_setCursorEP9lua_State
,__ZN4love5mouse11w_getCursorEP9lua_State,__ZN4love5mouse11w_hasCursorEP9lua_State,__ZN4love5mouse6w_getXEP9lua_State,__ZN4love5mouse6w_getYEP9lua_State,__ZN4love5mouse6w_setXEP9lua_State,__ZN4love5mouse6w_setYEP9lua_State,__ZN4love5mouse13w_setPositionEP9lua_State,__ZN4love5mouse8w_isDownEP9lua_State,__ZN4love5mouse12w_setVisibleEP9lua_State,__ZN4love5mouse11w_isVisibleEP9lua_State,__ZN4love5mouse13w_getPositionEP9lua_State,__ZN4love5mouse12w_setGrabbedEP9lua_State,__ZN4love5mouse11w_isGrabbedEP9lua_State,__ZN4love5mouse17w_setRelativeModeEP9lua_State,__ZN4love5mouse17w_getRelativeModeEP9lua_State,_luaopen_cursor,__ZN4love7physics5box2d10w_newWorldEP9lua_State,__ZN4love7physics5box2d9w_newBodyEP9lua_State,__ZN4love7physics5box2d12w_newFixtureEP9lua_State,__ZN4love7physics5box2d16w_newCircleShapeEP9lua_State,__ZN4love7physics5box2d19w_newRectangleShapeEP9lua_State,__ZN4love7physics5box2d17w_newPolygonShapeEP9lua_State,__ZN4love7physics5box2d14w_newEdgeShapeEP9lua_State,__ZN4love7physics5box2d15w_newChainShapeEP9lua_State,__ZN4love7physics5box2d18w_newDistanceJointEP9lua_State,__ZN4love7physics5box2d15w_newMouseJointEP9lua_State,__ZN4love7physics5box2d18w_newRevoluteJointEP9lua_State,__ZN4love7physics5box2d19w_newPrismaticJointEP9lua_State,__ZN4love7physics5box2d16w_newPulleyJointEP9lua_State,__ZN4love7physics5box2d14w_newGearJointEP9lua_State
,__ZN4love7physics5box2d18w_newFrictionJointEP9lua_State,__ZN4love7physics5box2d14w_newWeldJointEP9lua_State,__ZN4love7physics5box2d15w_newWheelJointEP9lua_State,__ZN4love7physics5box2d14w_newRopeJointEP9lua_State,__ZN4love7physics5box2d15w_newMotorJointEP9lua_State,__ZN4love7physics5box2d13w_getDistanceEP9lua_State,__ZN4love7physics5box2d10w_getMeterEP9lua_State,__ZN4love7physics5box2d10w_setMeterEP9lua_State,_luaopen_world,_luaopen_contact,_luaopen_body,_luaopen_fixture,_luaopen_shape,_luaopen_circleshape,_luaopen_polygonshape,_luaopen_edgeshape,_luaopen_chainshape,_luaopen_joint,_luaopen_mousejoint,_luaopen_distancejoint,_luaopen_prismaticjoint,_luaopen_revolutejoint,_luaopen_pulleyjoint,_luaopen_gearjoint,_luaopen_frictionjoint,_luaopen_weldjoint,_luaopen_wheeljoint,_luaopen_ropejoint,_luaopen_motorjoint,__ZN4love7physics5box2d24w_PolygonShape_getPointsEP9lua_State
,__ZN4love7physics5box2d23w_PolygonShape_validateEP9lua_State,__ZN4love7physics5box2d36w_PrismaticJoint_getJointTranslationEP9lua_State,__ZN4love7physics5box2d30w_PrismaticJoint_getJointSpeedEP9lua_State,__ZN4love7physics5box2d32w_PrismaticJoint_setMotorEnabledEP9lua_State,__ZN4love7physics5box2d31w_PrismaticJoint_isMotorEnabledEP9lua_State,__ZN4love7physics5box2d33w_PrismaticJoint_setMaxMotorForceEP9lua_State,__ZN4love7physics5box2d30w_PrismaticJoint_setMotorSpeedEP9lua_State,__ZN4love7physics5box2d30w_PrismaticJoint_getMotorSpeedEP9lua_State,__ZN4love7physics5box2d30w_PrismaticJoint_getMotorForceEP9lua_State,__ZN4love7physics5box2d33w_PrismaticJoint_getMaxMotorForceEP9lua_State,__ZN4love7physics5box2d33w_PrismaticJoint_setLimitsEnabledEP9lua_State,__ZN4love7physics5box2d33w_PrismaticJoint_hasLimitsEnabledEP9lua_State,__ZN4love7physics5box2d30w_PrismaticJoint_setUpperLimitEP9lua_State,__ZN4love7physics5box2d30w_PrismaticJoint_setLowerLimitEP9lua_State,__ZN4love7physics5box2d26w_PrismaticJoint_setLimitsEP9lua_State,__ZN4love7physics5box2d30w_PrismaticJoint_getLowerLimitEP9lua_State,__ZN4love7physics5box2d30w_PrismaticJoint_getUpperLimitEP9lua_State,__ZN4love7physics5box2d26w_PrismaticJoint_getLimitsEP9lua_State,__ZN4love7physics5box2d30w_PulleyJoint_getGroundAnchorsEP9lua_State,__ZN4love7physics5box2d24w_PulleyJoint_getLengthAEP9lua_State,__ZN4love7physics5box2d24w_PulleyJoint_getLengthBEP9lua_State,__ZN4love7physics5box2d22w_PulleyJoint_getRatioEP9lua_State,__ZN4love7physics5box2d29w_RevoluteJoint_getJointAngleEP9lua_State,__ZN4love7physics5box2d29w_RevoluteJoint_getJointSpeedEP9lua_State,__ZN4love7physics5box2d31w_RevoluteJoint_setMotorEnabledEP9lua_State,__ZN4love7physics5box2d30w_RevoluteJoint_isMotorEnabledEP9lua_State,__ZN4love7physics5box2d33w_RevoluteJoint_setMaxMotorTorqueEP9lua_State,__ZN4love7physics5box2d29w_RevoluteJoint_setMotorSpeedEP9lua_State,__ZN4love7physics5box2d29w_RevoluteJoint_getMotorSpeedEP9lua_State,__ZN4love7physics5box2d30w_RevoluteJoint_getMotorTorqueEP9lua_State
,__ZN4love7physics5box2d33w_RevoluteJoint_getMaxMotorTorqueEP9lua_State,__ZN4love7physics5box2d32w_RevoluteJoint_setLimitsEnabledEP9lua_State,__ZN4love7physics5box2d32w_RevoluteJoint_hasLimitsEnabledEP9lua_State,__ZN4love7physics5box2d29w_RevoluteJoint_setUpperLimitEP9lua_State,__ZN4love7physics5box2d29w_RevoluteJoint_setLowerLimitEP9lua_State,__ZN4love7physics5box2d25w_RevoluteJoint_setLimitsEP9lua_State,__ZN4love7physics5box2d29w_RevoluteJoint_getLowerLimitEP9lua_State,__ZN4love7physics5box2d29w_RevoluteJoint_getUpperLimitEP9lua_State,__ZN4love7physics5box2d25w_RevoluteJoint_getLimitsEP9lua_State,__ZN4love7physics5box2d24w_RopeJoint_getMaxLengthEP9lua_State,__ZN4love7physics5box2d15w_Shape_getTypeEP9lua_State,__ZN4love7physics5box2d17w_Shape_getRadiusEP9lua_State,__ZN4love7physics5box2d21w_Shape_getChildCountEP9lua_State,__ZN4love7physics5box2d17w_Shape_testPointEP9lua_State,__ZN4love7physics5box2d15w_Shape_rayCastEP9lua_State,__ZN4love7physics5box2d19w_Shape_computeAABBEP9lua_State,__ZN4love7physics5box2d19w_Shape_computeMassEP9lua_State,__ZN4love7physics5box2d24w_WeldJoint_setFrequencyEP9lua_State,__ZN4love7physics5box2d24w_WeldJoint_getFrequencyEP9lua_State,__ZN4love7physics5box2d27w_WeldJoint_setDampingRatioEP9lua_State,__ZN4love7physics5box2d27w_WeldJoint_getDampingRatioEP9lua_State,__ZN4love7physics5box2d32w_WheelJoint_getJointTranslationEP9lua_State,__ZN4love7physics5box2d26w_WheelJoint_getJointSpeedEP9lua_State,__ZN4love7physics5box2d28w_WheelJoint_setMotorEnabledEP9lua_State,__ZN4love7physics5box2d27w_WheelJoint_isMotorEnabledEP9lua_State,__ZN4love7physics5box2d26w_WheelJoint_setMotorSpeedEP9lua_State,__ZN4love7physics5box2d26w_WheelJoint_getMotorSpeedEP9lua_State,__ZN4love7physics5box2d30w_WheelJoint_setMaxMotorTorqueEP9lua_State,__ZN4love7physics5box2d30w_WheelJoint_getMaxMotorTorqueEP9lua_State,__ZN4love7physics5box2d27w_WheelJoint_getMotorTorqueEP9lua_State
,__ZN4love7physics5box2d31w_WheelJoint_setSpringFrequencyEP9lua_State,__ZN4love7physics5box2d31w_WheelJoint_getSpringFrequencyEP9lua_State,__ZN4love7physics5box2d34w_WheelJoint_setSpringDampingRatioEP9lua_State,__ZN4love7physics5box2d34w_WheelJoint_getSpringDampingRatioEP9lua_State,__ZN4love7physics5box2d14w_World_updateEP9lua_State,__ZN4love7physics5box2d20w_World_setCallbacksEP9lua_State,__ZN4love7physics5box2d20w_World_getCallbacksEP9lua_State,__ZN4love7physics5box2d24w_World_setContactFilterEP9lua_State,__ZN4love7physics5box2d24w_World_getContactFilterEP9lua_State,__ZN4love7physics5box2d18w_World_setGravityEP9lua_State,__ZN4love7physics5box2d18w_World_getGravityEP9lua_State,__ZN4love7physics5box2d23w_World_translateOriginEP9lua_State,__ZN4love7physics5box2d26w_World_setSleepingAllowedEP9lua_State,__ZN4love7physics5box2d25w_World_isSleepingAllowedEP9lua_State,__ZN4love7physics5box2d16w_World_isLockedEP9lua_State,__ZN4love7physics5box2d20w_World_getBodyCountEP9lua_State,__ZN4love7physics5box2d21w_World_getJointCountEP9lua_State,__ZN4love7physics5box2d23w_World_getContactCountEP9lua_State,__ZN4love7physics5box2d19w_World_getBodyListEP9lua_State,__ZN4love7physics5box2d20w_World_getJointListEP9lua_State,__ZN4love7physics5box2d22w_World_getContactListEP9lua_State,__ZN4love7physics5box2d24w_World_queryBoundingBoxEP9lua_State,__ZN4love7physics5box2d15w_World_rayCastEP9lua_State,__ZN4love7physics5box2d15w_World_destroyEP9lua_State,__ZN4love7physics5box2d19w_World_isDestroyedEP9lua_State,__ZN4love5sound12w_newDecoderEP9lua_State,__ZN4love5sound14w_newSoundDataEP9lua_State,_luaopen_sounddata,_luaopen_decoder,__ZN4love5sound23w_SoundData_getChannelsEP9lua_State
,__ZN4love5sound23w_SoundData_getBitDepthEP9lua_State,__ZN4love5sound25w_SoundData_getSampleRateEP9lua_State,__ZN4love5sound26w_SoundData_getSampleCountEP9lua_State,__ZN4love5sound23w_SoundData_getDurationEP9lua_State,__ZN4love5sound21w_SoundData_setSampleEP9lua_State,__ZN4love5sound21w_SoundData_getSampleEP9lua_State,__ZN4love6system7w_getOSEP9lua_State,__ZN4love6system19w_getProcessorCountEP9lua_State,__ZN4love6system18w_setClipboardTextEP9lua_State,__ZN4love6system18w_getClipboardTextEP9lua_State,__ZN4love6system14w_getPowerInfoEP9lua_State,__ZN4love6system9w_openURLEP9lua_State,__ZN4love6system9w_vibrateEP9lua_State,__ZN4love6thread11w_newThreadEP9lua_State,__ZN4love6thread12w_newChannelEP9lua_State,__ZN4love6thread12w_getChannelEP9lua_State,_luaopen_thread,_luaopen_channel,__ZN4love5timer6w_stepEP9lua_State,__ZN4love5timer10w_getDeltaEP9lua_State,__ZN4love5timer8w_getFPSEP9lua_State,__ZN4love5timer17w_getAverageDeltaEP9lua_State,__ZN4love5timer7w_sleepEP9lua_State,__ZN4love5timer9w_getTimeEP9lua_State,__ZN4love5touch12w_getTouchesEP9lua_State,__ZN4love5touch13w_getPositionEP9lua_State,__ZN4love5touch13w_getPressureEP9lua_State,__ZN4love5video16w_newVideoStreamEP9lua_State,__ZN4love5video19luaopen_videostreamEP9lua_State,__ZN4love5video21w_VideoStream_setSyncEP9lua_State
,__ZN4love5video25w_VideoStream_getFilenameEP9lua_State,__ZN4love5video18w_VideoStream_playEP9lua_State,__ZN4love5video19w_VideoStream_pauseEP9lua_State,__ZN4love5video18w_VideoStream_seekEP9lua_State,__ZN4love5video20w_VideoStream_rewindEP9lua_State,__ZN4love5video18w_VideoStream_tellEP9lua_State,__ZN4love5video23w_VideoStream_isPlayingEP9lua_State,__ZN4love6window17w_getDisplayCountEP9lua_State,__ZN4love6window16w_getDisplayNameEP9lua_State,__ZN4love6window9w_setModeEP9lua_State,__ZN4love6window9w_getModeEP9lua_State,__ZN4love6window20w_getFullscreenModesEP9lua_State,__ZN4love6window15w_setFullscreenEP9lua_State,__ZN4love6window15w_getFullscreenEP9lua_State,__ZN4love6window8w_isOpenEP9lua_State,__ZN4love6window7w_closeEP9lua_State,__ZN4love6window22w_getDesktopDimensionsEP9lua_State,__ZN4love6window13w_setPositionEP9lua_State,__ZN4love6window13w_getPositionEP9lua_State,__ZN4love6window9w_setIconEP9lua_State,__ZN4love6window9w_getIconEP9lua_State,__ZN4love6window24w_setDisplaySleepEnabledEP9lua_State,__ZN4love6window23w_isDisplaySleepEnabledEP9lua_State,__ZN4love6window10w_setTitleEP9lua_State,__ZN4love6window10w_getTitleEP9lua_State,__ZN4love6window10w_hasFocusEP9lua_State,__ZN4love6window15w_hasMouseFocusEP9lua_State,__ZN4love6window11w_isVisibleEP9lua_State,__ZN4love6window15w_getPixelScaleEP9lua_State,__ZN4love6window10w_toPixelsEP9lua_State
,__ZN4love6window12w_fromPixelsEP9lua_State,__ZN4love6window10w_minimizeEP9lua_State,__ZN4love6window10w_maximizeEP9lua_State,__ZN4love6window16w_showMessageBoxEP9lua_State,__ZN4love6window18w_requestAttentionEP9lua_State,__ZL12host_serviceP9lua_State,__ZL17host_check_eventsP9lua_State,__ZL30host_compress_with_range_coderP9lua_State,__ZL12host_connectP9lua_State,__ZL10host_flushP9lua_State,__ZL14host_broadcastP9lua_State,__ZL18host_channel_limitP9lua_State,__ZL20host_bandwidth_limitP9lua_State,__ZL23host_get_socket_addressP9lua_State,__ZL7host_gcP9lua_State,__ZL20host_total_sent_dataP9lua_State,__ZL24host_total_received_dataP9lua_State,__ZL17host_service_timeP9lua_State,__ZL15host_peer_countP9lua_State,__ZL13host_get_peerP9lua_State,__ZL15peer_disconnectP9lua_State,__ZL19peer_disconnect_nowP9lua_State,__ZL21peer_disconnect_laterP9lua_State,__ZL10peer_resetP9lua_State,__ZL9peer_pingP9lua_State,__ZL12peer_receiveP9lua_State,__ZL9peer_sendP9lua_State,__ZL23peer_throttle_configureP9lua_State,__ZL18peer_ping_intervalP9lua_State,__ZL12peer_timeoutP9lua_State
,__ZL10peer_indexP9lua_State,__ZL10peer_stateP9lua_State,__ZL15peer_connect_idP9lua_State,__ZL20peer_round_trip_timeP9lua_State,__ZL25peer_last_round_trip_timeP9lua_State,__ZL11host_createP9lua_State,__ZL14linked_versionP9lua_State,_malloc,b76,b76,_byteoffset,_codepoint,_utfchar,_utflen,_iter_codes,b76,b76,__ZNK4love9Exception4whatEv,__ZN4love16w_Data_getStringEP9lua_State,__ZN4love17w_Data_getPointerEP9lua_State,__ZN4love14w_Data_getSizeEP9lua_State,__ZN4love5audio16w_getSourceCountEP9lua_State,__ZN4love5audio11w_newSourceEP9lua_State,__ZN4love5audio6w_playEP9lua_State,__ZN4love5audio6w_stopEP9lua_State,__ZN4love5audio7w_pauseEP9lua_State,__ZN4love5audio8w_resumeEP9lua_State,__ZN4love5audio8w_rewindEP9lua_State,__ZN4love5audio11w_setVolumeEP9lua_State,__ZN4love5audio11w_getVolumeEP9lua_State
,__ZN4love5audio13w_setPositionEP9lua_State,__ZN4love5audio13w_getPositionEP9lua_State,__ZN4love5audio16w_setOrientationEP9lua_State,__ZN4love5audio16w_getOrientationEP9lua_State,__ZN4love5audio13w_setVelocityEP9lua_State,__ZN4love5audio13w_getVelocityEP9lua_State,__ZN4love5audio17w_setDopplerScaleEP9lua_State,__ZN4love5audio17w_getDopplerScaleEP9lua_State,__ZN4love5audio18w_setDistanceModelEP9lua_State,__ZN4love5audio18w_getDistanceModelEP9lua_State,_luaopen_source,__ZN4love5audio14w_Source_cloneEP9lua_State,__ZN4love5audio13w_Source_playEP9lua_State,__ZN4love5audio13w_Source_stopEP9lua_State,__ZN4love5audio14w_Source_pauseEP9lua_State,__ZN4love5audio15w_Source_resumeEP9lua_State,__ZN4love5audio15w_Source_rewindEP9lua_State,__ZN4love5audio17w_Source_setPitchEP9lua_State,__ZN4love5audio17w_Source_getPitchEP9lua_State,__ZN4love5audio18w_Source_setVolumeEP9lua_State,__ZN4love5audio18w_Source_getVolumeEP9lua_State,__ZN4love5audio13w_Source_seekEP9lua_State,__ZN4love5audio13w_Source_tellEP9lua_State,__ZN4love5audio20w_Source_getDurationEP9lua_State,__ZN4love5audio20w_Source_setPositionEP9lua_State,__ZN4love5audio20w_Source_getPositionEP9lua_State,__ZN4love5audio20w_Source_setVelocityEP9lua_State,__ZN4love5audio20w_Source_getVelocityEP9lua_State,__ZN4love5audio21w_Source_setDirectionEP9lua_State,__ZN4love5audio21w_Source_getDirectionEP9lua_State
,__ZN4love5audio16w_Source_setConeEP9lua_State,__ZN4love5audio16w_Source_getConeEP9lua_State,__ZN4love5audio20w_Source_setRelativeEP9lua_State,__ZN4love5audio19w_Source_isRelativeEP9lua_State,__ZN4love5audio19w_Source_setLoopingEP9lua_State,__ZN4love5audio18w_Source_isLoopingEP9lua_State,__ZN4love5audio18w_Source_isStoppedEP9lua_State,__ZN4love5audio17w_Source_isPausedEP9lua_State,__ZN4love5audio18w_Source_isPlayingEP9lua_State,__ZN4love5audio24w_Source_setVolumeLimitsEP9lua_State,__ZN4love5audio24w_Source_getVolumeLimitsEP9lua_State,__ZN4love5audio32w_Source_setAttenuationDistancesEP9lua_State,__ZN4love5audio32w_Source_getAttenuationDistancesEP9lua_State,__ZN4love5audio19w_Source_setRolloffEP9lua_State,__ZN4love5audio19w_Source_getRolloffEP9lua_State,__ZN4love5audio20w_Source_getChannelsEP9lua_State,__ZN4love5audio16w_Source_getTypeEP9lua_State,__ZN4love5event6w_pumpEP9lua_State,__ZN4love5event6w_pollEP9lua_State,__ZN4love5event6w_waitEP9lua_State,__ZN4love5event6w_pushEP9lua_State,__ZN4love5event7w_clearEP9lua_State,__ZN4love5event6w_quitEP9lua_State,b76,b76,__ZNK4love10filesystem8FileData7getDataEv,__ZNK4love10filesystem8FileData7getSizeEv,__ZN4love10filesystem6w_initEP9lua_State,__ZN4love10filesystem10w_setFusedEP9lua_State,__ZN4love10filesystem9w_isFusedEP9lua_State
,__ZN4love10filesystem24w_setAndroidSaveExternalEP9lua_State,__ZN4love10filesystem13w_setIdentityEP9lua_State,__ZN4love10filesystem13w_getIdentityEP9lua_State,__ZN4love10filesystem11w_setSourceEP9lua_State,__ZN4love10filesystem11w_getSourceEP9lua_State,__ZN4love10filesystem7w_mountEP9lua_State,__ZN4love10filesystem9w_unmountEP9lua_State,__ZN4love10filesystem9w_newFileEP9lua_State,__ZN4love10filesystem21w_getWorkingDirectoryEP9lua_State,__ZN4love10filesystem18w_getUserDirectoryEP9lua_State,__ZN4love10filesystem21w_getAppdataDirectoryEP9lua_State,__ZN4love10filesystem18w_getSaveDirectoryEP9lua_State,__ZN4love10filesystem24w_getSourceBaseDirectoryEP9lua_State,__ZN4love10filesystem18w_getRealDirectoryEP9lua_State,__ZN4love10filesystem19w_getExecutablePathEP9lua_State,__ZN4love10filesystem8w_existsEP9lua_State,__ZN4love10filesystem13w_isDirectoryEP9lua_State,__ZN4love10filesystem8w_isFileEP9lua_State,__ZN4love10filesystem11w_isSymlinkEP9lua_State,__ZN4love10filesystem17w_createDirectoryEP9lua_State,__ZN4love10filesystem8w_removeEP9lua_State,__ZN4love10filesystem6w_readEP9lua_State,__ZN4love10filesystem7w_writeEP9lua_State,__ZN4love10filesystem8w_appendEP9lua_State,__ZN4love10filesystem19w_getDirectoryItemsEP9lua_State,__ZN4love10filesystem7w_linesEP9lua_State,__ZN4love10filesystem6w_loadEP9lua_State,__ZN4love10filesystem17w_getLastModifiedEP9lua_State,__ZN4love10filesystem9w_getSizeEP9lua_State,__ZN4love10filesystem20w_setSymlinksEnabledEP9lua_State
,__ZN4love10filesystem20w_areSymlinksEnabledEP9lua_State,__ZN4love10filesystem13w_newFileDataEP9lua_State,__ZN4love10filesystem16w_getRequirePathEP9lua_State,__ZN4love10filesystem16w_setRequirePathEP9lua_State,_luaopen_file,_luaopen_droppedfile,_luaopen_filedata,__ZN4love4font15w_newRasterizerEP9lua_State,__ZN4love4font23w_newTrueTypeRasterizerEP9lua_State,__ZN4love4font21w_newBMFontRasterizerEP9lua_State,__ZN4love4font20w_newImageRasterizerEP9lua_State,__ZN4love4font14w_newGlyphDataEP9lua_State,_luaopen_glyphdata,_luaopen_rasterizer,__ZN4love4font20w_GlyphData_getWidthEP9lua_State,__ZN4love4font21w_GlyphData_getHeightEP9lua_State,__ZN4love4font25w_GlyphData_getDimensionsEP9lua_State,__ZN4love4font20w_GlyphData_getGlyphEP9lua_State,__ZN4love4font26w_GlyphData_getGlyphStringEP9lua_State,__ZN4love4font22w_GlyphData_getAdvanceEP9lua_State,__ZN4love4font22w_GlyphData_getBearingEP9lua_State,__ZN4love4font26w_GlyphData_getBoundingBoxEP9lua_State,__ZN4love4font21w_GlyphData_getFormatEP9lua_State,__ZN4love4font22w_Rasterizer_getHeightEP9lua_State,__ZN4love4font23w_Rasterizer_getAdvanceEP9lua_State,__ZN4love4font22w_Rasterizer_getAscentEP9lua_State,__ZN4love4font23w_Rasterizer_getDescentEP9lua_State,__ZN4love4font26w_Rasterizer_getLineHeightEP9lua_State,__ZN4love4font25w_Rasterizer_getGlyphDataEP9lua_State,__ZN4love4font26w_Rasterizer_getGlyphCountEP9lua_State
,__ZN4love4font22w_Rasterizer_hasGlyphsEP9lua_State,__ZN4love8graphics6opengl7w_resetEP9lua_State,__ZN4love8graphics6opengl7w_clearEP9lua_State,__ZN4love8graphics6opengl9w_discardEP9lua_State,__ZN4love8graphics6opengl9w_presentEP9lua_State,__ZN4love8graphics6opengl10w_newImageEP9lua_State,__ZN4love8graphics6opengl9w_newQuadEP9lua_State,__ZN4love8graphics6opengl9w_newFontEP9lua_State,__ZN4love8graphics6opengl14w_newImageFontEP9lua_State,__ZN4love8graphics6opengl16w_newSpriteBatchEP9lua_State,__ZN4love8graphics6opengl19w_newParticleSystemEP9lua_State,__ZN4love8graphics6opengl11w_newCanvasEP9lua_State,__ZN4love8graphics6opengl11w_newShaderEP9lua_State,__ZN4love8graphics6opengl9w_newMeshEP9lua_State,__ZN4love8graphics6opengl9w_newTextEP9lua_State,__ZN4love8graphics6opengl10w_newVideoEP9lua_State,__ZN4love8graphics6opengl10w_setColorEP9lua_State,__ZN4love8graphics6opengl10w_getColorEP9lua_State,__ZN4love8graphics6opengl20w_setBackgroundColorEP9lua_State,__ZN4love8graphics6opengl20w_getBackgroundColorEP9lua_State,__ZN4love8graphics6opengl12w_setNewFontEP9lua_State,__ZN4love8graphics6opengl9w_setFontEP9lua_State,__ZN4love8graphics6opengl9w_getFontEP9lua_State,__ZN4love8graphics6opengl14w_setColorMaskEP9lua_State,__ZN4love8graphics6opengl14w_getColorMaskEP9lua_State,__ZN4love8graphics6opengl14w_setBlendModeEP9lua_State,__ZN4love8graphics6opengl14w_getBlendModeEP9lua_State,__ZN4love8graphics6opengl18w_setDefaultFilterEP9lua_State,__ZN4love8graphics6opengl18w_getDefaultFilterEP9lua_State,__ZN4love8graphics6opengl24w_setDefaultMipmapFilterEP9lua_State
,__ZN4love8graphics6opengl24w_getDefaultMipmapFilterEP9lua_State,__ZN4love8graphics6opengl14w_setLineWidthEP9lua_State,__ZN4love8graphics6opengl14w_setLineStyleEP9lua_State,__ZN4love8graphics6opengl13w_setLineJoinEP9lua_State,__ZN4love8graphics6opengl14w_getLineWidthEP9lua_State,__ZN4love8graphics6opengl14w_getLineStyleEP9lua_State,__ZN4love8graphics6opengl13w_getLineJoinEP9lua_State,__ZN4love8graphics6opengl14w_setPointSizeEP9lua_State,__ZN4love8graphics6opengl14w_getPointSizeEP9lua_State,__ZN4love8graphics6opengl14w_setWireframeEP9lua_State,__ZN4love8graphics6opengl13w_isWireframeEP9lua_State,__ZN4love8graphics6opengl15w_newScreenshotEP9lua_State,__ZN4love8graphics6opengl11w_setCanvasEP9lua_State,__ZN4love8graphics6opengl11w_getCanvasEP9lua_State,__ZN4love8graphics6opengl11w_setShaderEP9lua_State,__ZN4love8graphics6opengl11w_getShaderEP9lua_State,__ZN4love8graphics6opengl22w_setDefaultShaderCodeEP9lua_State,__ZN4love8graphics6opengl14w_getSupportedEP9lua_State,__ZN4love8graphics6opengl18w_getCanvasFormatsEP9lua_State,__ZN4love8graphics6opengl27w_getCompressedImageFormatsEP9lua_State,__ZN4love8graphics6opengl17w_getRendererInfoEP9lua_State,__ZN4love8graphics6opengl17w_getSystemLimitsEP9lua_State,__ZN4love8graphics6opengl10w_getStatsEP9lua_State,__ZN4love8graphics6opengl6w_drawEP9lua_State,__ZN4love8graphics6opengl7w_printEP9lua_State,__ZN4love8graphics6opengl8w_printfEP9lua_State,__ZN4love8graphics6opengl11w_isCreatedEP9lua_State,__ZN4love8graphics6opengl10w_isActiveEP9lua_State,__ZN4love8graphics6opengl16w_isGammaCorrectEP9lua_State,__ZN4love8graphics6opengl10w_getWidthEP9lua_State
,__ZN4love8graphics6opengl11w_getHeightEP9lua_State,__ZN4love8graphics6opengl15w_getDimensionsEP9lua_State,__ZN4love8graphics6opengl12w_setScissorEP9lua_State,__ZN4love8graphics6opengl18w_intersectScissorEP9lua_State,__ZN4love8graphics6opengl12w_getScissorEP9lua_State,__ZN4love8graphics6opengl9w_stencilEP9lua_State,__ZN4love8graphics6opengl16w_setStencilTestEP9lua_State,__ZN4love8graphics6opengl16w_getStencilTestEP9lua_State,__ZN4love8graphics6opengl8w_pointsEP9lua_State,__ZN4love8graphics6opengl6w_lineEP9lua_State,__ZN4love8graphics6opengl11w_rectangleEP9lua_State,__ZN4love8graphics6opengl8w_circleEP9lua_State,__ZN4love8graphics6opengl9w_ellipseEP9lua_State,__ZN4love8graphics6opengl5w_arcEP9lua_State,__ZN4love8graphics6opengl9w_polygonEP9lua_State,__ZN4love8graphics6opengl6w_pushEP9lua_State,__ZN4love8graphics6opengl5w_popEP9lua_State,__ZN4love8graphics6opengl8w_rotateEP9lua_State,__ZN4love8graphics6opengl7w_scaleEP9lua_State,__ZN4love8graphics6opengl11w_translateEP9lua_State,__ZN4love8graphics6opengl7w_shearEP9lua_State,__ZN4love8graphics6opengl8w_originEP9lua_State,__ZN4love8graphics6openglL16luaopen_drawableEP9lua_State,_luaopen_texture,_luaopen_font,_luaopen_image,_luaopen_quad,_luaopen_spritebatch,_luaopen_particlesystem,_luaopen_canvas
,_luaopen_shader,_luaopen_mesh,_luaopen_text,__ZN4love8graphics6opengl13luaopen_videoEP9lua_State,__ZN4love8graphics6opengl23w_Image_setMipmapFilterEP9lua_State,__ZN4love8graphics6opengl23w_Image_getMipmapFilterEP9lua_State,__ZN4love8graphics6opengl20w_Image_isCompressedEP9lua_State,__ZN4love8graphics6opengl15w_Image_refreshEP9lua_State,__ZN4love8graphics6opengl15w_Image_getDataEP9lua_State,__ZN4love8graphics6opengl16w_Image_getFlagsEP9lua_State,__ZN4love8graphics6opengl18w_Mesh_setVerticesEP9lua_State,__ZN4love8graphics6opengl16w_Mesh_setVertexEP9lua_State,__ZN4love8graphics6opengl16w_Mesh_getVertexEP9lua_State,__ZN4love8graphics6opengl25w_Mesh_setVertexAttributeEP9lua_State,__ZN4love8graphics6opengl25w_Mesh_getVertexAttributeEP9lua_State,__ZN4love8graphics6opengl21w_Mesh_getVertexCountEP9lua_State,__ZN4love8graphics6opengl22w_Mesh_getVertexFormatEP9lua_State,__ZN4love8graphics6opengl26w_Mesh_setAttributeEnabledEP9lua_State,__ZN4love8graphics6opengl25w_Mesh_isAttributeEnabledEP9lua_State,__ZN4love8graphics6opengl22w_Mesh_attachAttributeEP9lua_State,__ZN4love8graphics6opengl12w_Mesh_flushEP9lua_State,__ZN4love8graphics6opengl19w_Mesh_setVertexMapEP9lua_State,__ZN4love8graphics6opengl19w_Mesh_getVertexMapEP9lua_State,__ZN4love8graphics6opengl17w_Mesh_setTextureEP9lua_State,__ZN4love8graphics6opengl17w_Mesh_getTextureEP9lua_State,__ZN4love8graphics6opengl18w_Mesh_setDrawModeEP9lua_State,__ZN4love8graphics6opengl18w_Mesh_getDrawModeEP9lua_State,__ZN4love8graphics6opengl19w_Mesh_setDrawRangeEP9lua_State,__ZN4love8graphics6opengl19w_Mesh_getDrawRangeEP9lua_State,__ZN4love8graphics6opengl22w_ParticleSystem_cloneEP9lua_State
,__ZN4love8graphics6opengl27w_ParticleSystem_setTextureEP9lua_State,__ZN4love8graphics6opengl27w_ParticleSystem_getTextureEP9lua_State,__ZN4love8graphics6opengl30w_ParticleSystem_setBufferSizeEP9lua_State,__ZN4love8graphics6opengl30w_ParticleSystem_getBufferSizeEP9lua_State,__ZN4love8graphics6opengl30w_ParticleSystem_setInsertModeEP9lua_State,__ZN4love8graphics6opengl30w_ParticleSystem_getInsertModeEP9lua_State,__ZN4love8graphics6opengl32w_ParticleSystem_setEmissionRateEP9lua_State,__ZN4love8graphics6opengl32w_ParticleSystem_getEmissionRateEP9lua_State,__ZN4love8graphics6opengl35w_ParticleSystem_setEmitterLifetimeEP9lua_State,__ZN4love8graphics6opengl35w_ParticleSystem_getEmitterLifetimeEP9lua_State,__ZN4love8graphics6opengl36w_ParticleSystem_setParticleLifetimeEP9lua_State,__ZN4love8graphics6opengl36w_ParticleSystem_getParticleLifetimeEP9lua_State,__ZN4love8graphics6opengl28w_ParticleSystem_setPositionEP9lua_State,__ZN4love8graphics6opengl28w_ParticleSystem_getPositionEP9lua_State,__ZN4love8graphics6opengl23w_ParticleSystem_moveToEP9lua_State,__ZN4love8graphics6opengl30w_ParticleSystem_setAreaSpreadEP9lua_State,__ZN4love8graphics6opengl30w_ParticleSystem_getAreaSpreadEP9lua_State,__ZN4love8graphics6opengl29w_ParticleSystem_setDirectionEP9lua_State,__ZN4love8graphics6opengl29w_ParticleSystem_getDirectionEP9lua_State,__ZN4love8graphics6opengl26w_ParticleSystem_setSpreadEP9lua_State,__ZN4love8graphics6opengl26w_ParticleSystem_getSpreadEP9lua_State,__ZN4love8graphics6opengl25w_ParticleSystem_setSpeedEP9lua_State,__ZN4love8graphics6opengl25w_ParticleSystem_getSpeedEP9lua_State,__ZN4love8graphics6opengl38w_ParticleSystem_setLinearAccelerationEP9lua_State,__ZN4love8graphics6opengl38w_ParticleSystem_getLinearAccelerationEP9lua_State,__ZN4love8graphics6opengl38w_ParticleSystem_setRadialAccelerationEP9lua_State,__ZN4love8graphics6opengl38w_ParticleSystem_getRadialAccelerationEP9lua_State,__ZN4love8graphics6opengl42w_ParticleSystem_setTangentialAccelerationEP9lua_State,__ZN4love8graphics6opengl42w_ParticleSystem_getTangentialAccelerationEP9lua_State,__ZN4love8graphics6opengl33w_ParticleSystem_setLinearDampingEP9lua_State
,__ZN4love8graphics6opengl33w_ParticleSystem_getLinearDampingEP9lua_State,__ZN4love8graphics6opengl25w_ParticleSystem_setSizesEP9lua_State,__ZN4love8graphics6opengl25w_ParticleSystem_getSizesEP9lua_State,__ZN4love8graphics6opengl33w_ParticleSystem_setSizeVariationEP9lua_State,__ZN4love8graphics6opengl33w_ParticleSystem_getSizeVariationEP9lua_State,__ZN4love8graphics6opengl28w_ParticleSystem_setRotationEP9lua_State,__ZN4love8graphics6opengl28w_ParticleSystem_getRotationEP9lua_State,__ZN4love8graphics6opengl24w_ParticleSystem_setSpinEP9lua_State,__ZN4love8graphics6opengl24w_ParticleSystem_getSpinEP9lua_State,__ZN4love8graphics6opengl33w_ParticleSystem_setSpinVariationEP9lua_State,__ZN4love8graphics6opengl33w_ParticleSystem_getSpinVariationEP9lua_State,__ZN4love8graphics6opengl26w_ParticleSystem_setColorsEP9lua_State,__ZN4love8graphics6opengl26w_ParticleSystem_getColorsEP9lua_State,__ZN4love8graphics6opengl25w_ParticleSystem_setQuadsEP9lua_State,__ZN4love8graphics6opengl25w_ParticleSystem_getQuadsEP9lua_State,__ZN4love8graphics6opengl26w_ParticleSystem_setOffsetEP9lua_State,__ZN4love8graphics6opengl26w_ParticleSystem_getOffsetEP9lua_State,__ZN4love8graphics6opengl36w_ParticleSystem_setRelativeRotationEP9lua_State,__ZN4love8graphics6opengl36w_ParticleSystem_hasRelativeRotationEP9lua_State,__ZN4love8graphics6opengl25w_ParticleSystem_getCountEP9lua_State,__ZN4love8graphics6opengl22w_ParticleSystem_startEP9lua_State,__ZN4love8graphics6opengl21w_ParticleSystem_stopEP9lua_State,__ZN4love8graphics6opengl22w_ParticleSystem_pauseEP9lua_State,__ZN4love8graphics6opengl22w_ParticleSystem_resetEP9lua_State,__ZN4love8graphics6opengl21w_ParticleSystem_emitEP9lua_State,__ZN4love8graphics6opengl25w_ParticleSystem_isActiveEP9lua_State,__ZN4love8graphics6opengl25w_ParticleSystem_isPausedEP9lua_State,__ZN4love8graphics6opengl26w_ParticleSystem_isStoppedEP9lua_State,__ZN4love8graphics6opengl23w_ParticleSystem_updateEP9lua_State,__ZN4love8graphics6opengl20w_Shader_getWarningsEP9lua_State
,__ZN4love8graphics6opengl16w_Shader_sendIntEP9lua_State,__ZN4love8graphics6opengl18w_Shader_sendFloatEP9lua_State,__ZN4love8graphics6opengl18w_Shader_sendColorEP9lua_State,__ZN4love8graphics6opengl19w_Shader_sendMatrixEP9lua_State,__ZN4love8graphics6opengl20w_Shader_sendTextureEP9lua_State,__ZN4love8graphics6opengl13w_Shader_sendEP9lua_State,__ZN4love8graphics6opengl26w_Shader_getExternVariableEP9lua_State,__ZN4love8graphics6opengl17w_SpriteBatch_addEP9lua_State,__ZN4love8graphics6opengl17w_SpriteBatch_setEP9lua_State,__ZN4love8graphics6opengl19w_SpriteBatch_clearEP9lua_State,__ZN4love8graphics6opengl19w_SpriteBatch_flushEP9lua_State,__ZN4love8graphics6opengl24w_SpriteBatch_setTextureEP9lua_State,__ZN4love8graphics6opengl24w_SpriteBatch_getTextureEP9lua_State,__ZN4love8graphics6opengl22w_SpriteBatch_setColorEP9lua_State,__ZN4love8graphics6opengl22w_SpriteBatch_getColorEP9lua_State,__ZN4love8graphics6opengl22w_SpriteBatch_getCountEP9lua_State,__ZN4love8graphics6opengl27w_SpriteBatch_setBufferSizeEP9lua_State,__ZN4love8graphics6opengl27w_SpriteBatch_getBufferSizeEP9lua_State,__ZN4love8graphics6opengl29w_SpriteBatch_attachAttributeEP9lua_State,__ZN4love8graphics6opengl10w_Text_setEP9lua_State,__ZN4love8graphics6opengl11w_Text_setfEP9lua_State,__ZN4love8graphics6opengl10w_Text_addEP9lua_State,__ZN4love8graphics6opengl11w_Text_addfEP9lua_State,__ZN4love8graphics6opengl12w_Text_clearEP9lua_State,__ZN4love8graphics6opengl14w_Text_setFontEP9lua_State,__ZN4love8graphics6opengl14w_Text_getFontEP9lua_State,__ZN4love8graphics6opengl15w_Text_getWidthEP9lua_State,__ZN4love8graphics6opengl16w_Text_getHeightEP9lua_State,__ZN4love8graphics6opengl20w_Text_getDimensionsEP9lua_State,__ZN4love8graphics6opengl17w_Video_getStreamEP9lua_State
,__ZN4love8graphics6opengl17w_Video_getSourceEP9lua_State,__ZN4love8graphics6opengl17w_Video_setSourceEP9lua_State,__ZN4love8graphics6opengl16w_Video_getWidthEP9lua_State,__ZN4love8graphics6opengl17w_Video_getHeightEP9lua_State,__ZN4love8graphics6opengl21w_Video_getDimensionsEP9lua_State,__ZN4love8graphics6opengl17w_Video_setFilterEP9lua_State,__ZN4love8graphics6opengl17w_Video_getFilterEP9lua_State,b76,b76,b76,b76,b76,b76,b76,__ZNK4love8graphics7Texture8getWidthEv,__ZNK4love8graphics7Texture9getHeightEv,__ZNK4love8graphics7Texture9getFilterEv,__ZNK4love8graphics7Texture7getWrapEv,__ZNK4love8graphics7Texture11getVerticesEv,__ZN4love8graphics18w_Quad_setViewportEP9lua_State,__ZN4love8graphics18w_Quad_getViewportEP9lua_State,__ZN4love8graphics18w_Texture_getWidthEP9lua_State,__ZN4love8graphics19w_Texture_getHeightEP9lua_State,__ZN4love8graphics23w_Texture_getDimensionsEP9lua_State,__ZN4love8graphics19w_Texture_setFilterEP9lua_State,__ZN4love8graphics19w_Texture_getFilterEP9lua_State,__ZN4love8graphics17w_Texture_setWrapEP9lua_State,__ZN4love8graphics17w_Texture_getWrapEP9lua_State,b76,b76
,__ZNK4love5image19CompressedImageData7getDataEv,__ZNK4love5image19CompressedImageData7getSizeEv,__ZN4love5image30w_CompressedImageData_getWidthEP9lua_State,__ZN4love5image31w_CompressedImageData_getHeightEP9lua_State,__ZN4love5image35w_CompressedImageData_getDimensionsEP9lua_State,__ZN4love5image36w_CompressedImageData_getMipmapCountEP9lua_State,__ZN4love5image31w_CompressedImageData_getFormatEP9lua_State,__ZN4love5image14w_newImageDataEP9lua_State,__ZN4love5image19w_newCompressedDataEP9lua_State,__ZN4love5image14w_isCompressedEP9lua_State,_luaopen_imagedata,_luaopen_compressedimagedata,__ZN4love5image20w_ImageData_getWidthEP9lua_State,__ZN4love5image21w_ImageData_getHeightEP9lua_State,__ZN4love5image25w_ImageData_getDimensionsEP9lua_State,__ZN4love5image20w_ImageData_getPixelEP9lua_State,__ZN4love5image20w_ImageData_setPixelEP9lua_State,__ZN4love5image17w_ImageData_pasteEP9lua_State,__ZN4love5image18w_ImageData_encodeEP9lua_State,__ZN4love5image27w_ImageData__mapPixelUnsafeEP9lua_State,__ZN4love5image26w_ImageData__performAtomicEP9lua_State,b76,b76,__ZN4love8joystick14w_getJoysticksEP9lua_State,__ZN4love8joystick18w_getJoystickCountEP9lua_State,__ZN4love8joystick19w_setGamepadMappingEP9lua_State,__ZN4love8joystick19w_getGamepadMappingEP9lua_State,__ZN4love8joystick21w_loadGamepadMappingsEP9lua_State,__ZN4love8joystick21w_saveGamepadMappingsEP9lua_State,_luaopen_joystick
,__ZN4love8keyboard14w_setKeyRepeatEP9lua_State,__ZN4love8keyboard14w_hasKeyRepeatEP9lua_State,__ZN4love8keyboard14w_setTextInputEP9lua_State,__ZN4love8keyboard14w_hasTextInputEP9lua_State,__ZN4love8keyboard19w_hasScreenKeyboardEP9lua_State,__ZN4love8keyboard8w_isDownEP9lua_State,__ZN4love8keyboard16w_isScancodeDownEP9lua_State,__ZN4love8keyboard20w_getScancodeFromKeyEP9lua_State,__ZN4love8keyboard20w_getKeyFromScancodeEP9lua_State,b76,b76,__ZNK4love4math14CompressedData7getDataEv,__ZNK4love4math14CompressedData7getSizeEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNK4love4math4Math13getModuleTypeEv,__ZNK4love4math4Math7getNameEv,b76,b76,__ZN4love4math23w_BezierCurve_getDegreeEP9lua_State,__ZN4love4math27w_BezierCurve_getDerivativeEP9lua_State
,__ZN4love4math29w_BezierCurve_getControlPointEP9lua_State,__ZN4love4math29w_BezierCurve_setControlPointEP9lua_State,__ZN4love4math32w_BezierCurve_insertControlPointEP9lua_State,__ZN4love4math32w_BezierCurve_removeControlPointEP9lua_State,__ZN4love4math34w_BezierCurve_getControlPointCountEP9lua_State,__ZN4love4math23w_BezierCurve_translateEP9lua_State,__ZN4love4math20w_BezierCurve_rotateEP9lua_State,__ZN4love4math19w_BezierCurve_scaleEP9lua_State,__ZN4love4math22w_BezierCurve_evaluateEP9lua_State,__ZN4love4math24w_BezierCurve_getSegmentEP9lua_State,__ZN4love4math20w_BezierCurve_renderEP9lua_State,__ZN4love4math27w_BezierCurve_renderSegmentEP9lua_State,__ZN4love4math26w_CompressedData_getFormatEP9lua_State,b76,b76,__ZNK4love5mouse5Mouse13getModuleTypeEv,__ZNK4love5mouse3sdl5Mouse7getNameEv,b76,b76,b76,b76,__ZNK4love5mouse3sdl5Mouse9getCursorEv,__ZNK4love5mouse3sdl5Mouse9hasCursorEv,b76,b76,b76,b76,b76,b76,b76
,b76,__ZNK4love5mouse3sdl5Mouse9isVisibleEv,b76,__ZNK4love5mouse3sdl5Mouse9isGrabbedEv,b76,__ZNK4love5mouse3sdl5Mouse15getRelativeModeEv,__ZN4love5mouse16w_Cursor_getTypeEP9lua_State,b76,b76,b76,__ZNK4love7physics5box2d7Physics13getModuleTypeEv,__ZNK4love7physics5box2d7Physics7getNameEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZN4love7physics5box2d11w_Body_getXEP9lua_State,__ZN4love7physics5box2d11w_Body_getYEP9lua_State,__ZN4love7physics5box2d15w_Body_getAngleEP9lua_State,__ZN4love7physics5box2d18w_Body_getPositionEP9lua_State
,__ZN4love7physics5box2d24w_Body_getLinearVelocityEP9lua_State,__ZN4love7physics5box2d21w_Body_getWorldCenterEP9lua_State,__ZN4love7physics5box2d21w_Body_getLocalCenterEP9lua_State,__ZN4love7physics5box2d25w_Body_getAngularVelocityEP9lua_State,__ZN4love7physics5box2d14w_Body_getMassEP9lua_State,__ZN4love7physics5box2d17w_Body_getInertiaEP9lua_State,__ZN4love7physics5box2d18w_Body_getMassDataEP9lua_State,__ZN4love7physics5box2d24w_Body_getAngularDampingEP9lua_State,__ZN4love7physics5box2d23w_Body_getLinearDampingEP9lua_State,__ZN4love7physics5box2d22w_Body_getGravityScaleEP9lua_State,__ZN4love7physics5box2d14w_Body_getTypeEP9lua_State,__ZN4love7physics5box2d25w_Body_applyLinearImpulseEP9lua_State,__ZN4love7physics5box2d26w_Body_applyAngularImpulseEP9lua_State,__ZN4love7physics5box2d18w_Body_applyTorqueEP9lua_State,__ZN4love7physics5box2d17w_Body_applyForceEP9lua_State,__ZN4love7physics5box2d11w_Body_setXEP9lua_State,__ZN4love7physics5box2d11w_Body_setYEP9lua_State,__ZN4love7physics5box2d24w_Body_setLinearVelocityEP9lua_State,__ZN4love7physics5box2d15w_Body_setAngleEP9lua_State,__ZN4love7physics5box2d25w_Body_setAngularVelocityEP9lua_State,__ZN4love7physics5box2d18w_Body_setPositionEP9lua_State,__ZN4love7physics5box2d20w_Body_resetMassDataEP9lua_State,__ZN4love7physics5box2d18w_Body_setMassDataEP9lua_State,__ZN4love7physics5box2d14w_Body_setMassEP9lua_State,__ZN4love7physics5box2d17w_Body_setInertiaEP9lua_State,__ZN4love7physics5box2d24w_Body_setAngularDampingEP9lua_State,__ZN4love7physics5box2d23w_Body_setLinearDampingEP9lua_State,__ZN4love7physics5box2d22w_Body_setGravityScaleEP9lua_State,__ZN4love7physics5box2d14w_Body_setTypeEP9lua_State,__ZN4love7physics5box2d20w_Body_getWorldPointEP9lua_State
,__ZN4love7physics5box2d21w_Body_getWorldVectorEP9lua_State,__ZN4love7physics5box2d21w_Body_getWorldPointsEP9lua_State,__ZN4love7physics5box2d20w_Body_getLocalPointEP9lua_State,__ZN4love7physics5box2d21w_Body_getLocalVectorEP9lua_State,__ZN4love7physics5box2d38w_Body_getLinearVelocityFromWorldPointEP9lua_State,__ZN4love7physics5box2d38w_Body_getLinearVelocityFromLocalPointEP9lua_State,__ZN4love7physics5box2d15w_Body_isBulletEP9lua_State,__ZN4love7physics5box2d16w_Body_setBulletEP9lua_State,__ZN4love7physics5box2d15w_Body_isActiveEP9lua_State,__ZN4love7physics5box2d14w_Body_isAwakeEP9lua_State,__ZN4love7physics5box2d25w_Body_setSleepingAllowedEP9lua_State,__ZN4love7physics5box2d24w_Body_isSleepingAllowedEP9lua_State,__ZN4love7physics5box2d16w_Body_setActiveEP9lua_State,__ZN4love7physics5box2d15w_Body_setAwakeEP9lua_State,__ZN4love7physics5box2d23w_Body_setFixedRotationEP9lua_State,__ZN4love7physics5box2d22w_Body_isFixedRotationEP9lua_State,__ZN4love7physics5box2d15w_Body_getWorldEP9lua_State,__ZN4love7physics5box2d21w_Body_getFixtureListEP9lua_State,__ZN4love7physics5box2d19w_Body_getJointListEP9lua_State,__ZN4love7physics5box2d21w_Body_getContactListEP9lua_State,__ZN4love7physics5box2d14w_Body_destroyEP9lua_State,__ZN4love7physics5box2d18w_Body_isDestroyedEP9lua_State,__ZN4love7physics5box2d18w_Body_setUserDataEP9lua_State,__ZN4love7physics5box2d18w_Body_getUserDataEP9lua_State,__ZN4love7physics5box2d26w_ChainShape_setNextVertexEP9lua_State,__ZN4love7physics5box2d30w_ChainShape_setPreviousVertexEP9lua_State,__ZN4love7physics5box2d25w_ChainShape_getChildEdgeEP9lua_State,__ZN4love7physics5box2d27w_ChainShape_getVertexCountEP9lua_State,__ZN4love7physics5box2d21w_ChainShape_getPointEP9lua_State,__ZN4love7physics5box2d22w_ChainShape_getPointsEP9lua_State
,__ZN4love7physics5box2d23w_CircleShape_getRadiusEP9lua_State,__ZN4love7physics5box2d23w_CircleShape_setRadiusEP9lua_State,__ZN4love7physics5box2d22w_CircleShape_getPointEP9lua_State,__ZN4love7physics5box2d22w_CircleShape_setPointEP9lua_State,__ZN4love7physics5box2d22w_Contact_getPositionsEP9lua_State,__ZN4love7physics5box2d19w_Contact_getNormalEP9lua_State,__ZN4love7physics5box2d21w_Contact_getFrictionEP9lua_State,__ZN4love7physics5box2d24w_Contact_getRestitutionEP9lua_State,__ZN4love7physics5box2d19w_Contact_isEnabledEP9lua_State,__ZN4love7physics5box2d20w_Contact_isTouchingEP9lua_State,__ZN4love7physics5box2d21w_Contact_setFrictionEP9lua_State,__ZN4love7physics5box2d24w_Contact_setRestitutionEP9lua_State,__ZN4love7physics5box2d20w_Contact_setEnabledEP9lua_State,__ZN4love7physics5box2d23w_Contact_resetFrictionEP9lua_State,__ZN4love7physics5box2d26w_Contact_resetRestitutionEP9lua_State,__ZN4love7physics5box2d25w_Contact_setTangentSpeedEP9lua_State,__ZN4love7physics5box2d25w_Contact_getTangentSpeedEP9lua_State,__ZN4love7physics5box2d21w_Contact_getChildrenEP9lua_State,__ZN4love7physics5box2d21w_Contact_getFixturesEP9lua_State,__ZN4love7physics5box2d21w_Contact_isDestroyedEP9lua_State,__ZN4love7physics5box2d25w_DistanceJoint_setLengthEP9lua_State,__ZN4love7physics5box2d25w_DistanceJoint_getLengthEP9lua_State,__ZN4love7physics5box2d28w_DistanceJoint_setFrequencyEP9lua_State,__ZN4love7physics5box2d28w_DistanceJoint_getFrequencyEP9lua_State,__ZN4love7physics5box2d31w_DistanceJoint_setDampingRatioEP9lua_State,__ZN4love7physics5box2d31w_DistanceJoint_getDampingRatioEP9lua_State,__ZN4love7physics5box2d21w_EdgeShape_getPointsEP9lua_State,__ZN4love7physics5box2d17w_Fixture_getTypeEP9lua_State,__ZN4love7physics5box2d21w_Fixture_setFrictionEP9lua_State,__ZN4love7physics5box2d24w_Fixture_setRestitutionEP9lua_State
,__ZN4love7physics5box2d20w_Fixture_setDensityEP9lua_State,__ZN4love7physics5box2d19w_Fixture_setSensorEP9lua_State,__ZN4love7physics5box2d21w_Fixture_getFrictionEP9lua_State,__ZN4love7physics5box2d24w_Fixture_getRestitutionEP9lua_State,__ZN4love7physics5box2d20w_Fixture_getDensityEP9lua_State,__ZN4love7physics5box2d17w_Fixture_getBodyEP9lua_State,__ZN4love7physics5box2d18w_Fixture_getShapeEP9lua_State,__ZN4love7physics5box2d18w_Fixture_isSensorEP9lua_State,__ZN4love7physics5box2d19w_Fixture_testPointEP9lua_State,__ZN4love7physics5box2d17w_Fixture_rayCastEP9lua_State,__ZN4love7physics5box2d23w_Fixture_setFilterDataEP9lua_State,__ZN4love7physics5box2d23w_Fixture_getFilterDataEP9lua_State,__ZN4love7physics5box2d21w_Fixture_setCategoryEP9lua_State,__ZN4love7physics5box2d21w_Fixture_getCategoryEP9lua_State,__ZN4love7physics5box2d17w_Fixture_setMaskEP9lua_State,__ZN4love7physics5box2d17w_Fixture_getMaskEP9lua_State,__ZN4love7physics5box2d21w_Fixture_setUserDataEP9lua_State,__ZN4love7physics5box2d21w_Fixture_getUserDataEP9lua_State,__ZN4love7physics5box2d24w_Fixture_getBoundingBoxEP9lua_State,__ZN4love7physics5box2d21w_Fixture_getMassDataEP9lua_State,__ZN4love7physics5box2d23w_Fixture_getGroupIndexEP9lua_State,__ZN4love7physics5box2d23w_Fixture_setGroupIndexEP9lua_State,__ZN4love7physics5box2d17w_Fixture_destroyEP9lua_State,__ZN4love7physics5box2d21w_Fixture_isDestroyedEP9lua_State,__ZN4love7physics5box2d27w_FrictionJoint_setMaxForceEP9lua_State,__ZN4love7physics5box2d27w_FrictionJoint_getMaxForceEP9lua_State,__ZN4love7physics5box2d28w_FrictionJoint_setMaxTorqueEP9lua_State,__ZN4love7physics5box2d28w_FrictionJoint_getMaxTorqueEP9lua_State,__ZN4love7physics5box2d20w_GearJoint_setRatioEP9lua_State,__ZN4love7physics5box2d20w_GearJoint_getRatioEP9lua_State
,__ZN4love7physics5box2d21w_GearJoint_getJointsEP9lua_State,__ZN4love7physics5box2d15w_Joint_getTypeEP9lua_State,__ZN4love7physics5box2d17w_Joint_getBodiesEP9lua_State,__ZN4love7physics5box2d18w_Joint_getAnchorsEP9lua_State,__ZN4love7physics5box2d24w_Joint_getReactionForceEP9lua_State,__ZN4love7physics5box2d25w_Joint_getReactionTorqueEP9lua_State,__ZN4love7physics5box2d27w_Joint_getCollideConnectedEP9lua_State,__ZN4love7physics5box2d19w_Joint_setUserDataEP9lua_State,__ZN4love7physics5box2d19w_Joint_getUserDataEP9lua_State,__ZN4love7physics5box2d15w_Joint_destroyEP9lua_State,__ZN4love7physics5box2d19w_Joint_isDestroyedEP9lua_State,__ZN4love7physics5box2d28w_MotorJoint_setLinearOffsetEP9lua_State,__ZN4love7physics5box2d28w_MotorJoint_getLinearOffsetEP9lua_State,__ZN4love7physics5box2d29w_MotorJoint_setAngularOffsetEP9lua_State,__ZN4love7physics5box2d29w_MotorJoint_getAngularOffsetEP9lua_State,__ZN4love7physics5box2d24w_MotorJoint_setMaxForceEP9lua_State,__ZN4love7physics5box2d24w_MotorJoint_getMaxForceEP9lua_State,__ZN4love7physics5box2d25w_MotorJoint_setMaxTorqueEP9lua_State,__ZN4love7physics5box2d25w_MotorJoint_getMaxTorqueEP9lua_State,__ZN4love7physics5box2d32w_MotorJoint_setCorrectionFactorEP9lua_State,__ZN4love7physics5box2d32w_MotorJoint_getCorrectionFactorEP9lua_State,__ZN4love7physics5box2d22w_MouseJoint_setTargetEP9lua_State,__ZN4love7physics5box2d22w_MouseJoint_getTargetEP9lua_State,__ZN4love7physics5box2d24w_MouseJoint_setMaxForceEP9lua_State,__ZN4love7physics5box2d24w_MouseJoint_getMaxForceEP9lua_State,__ZN4love7physics5box2d25w_MouseJoint_setFrequencyEP9lua_State,__ZN4love7physics5box2d25w_MouseJoint_getFrequencyEP9lua_State,__ZN4love7physics5box2d28w_MouseJoint_setDampingRatioEP9lua_State,__ZN4love7physics5box2d28w_MouseJoint_getDampingRatioEP9lua_State,b76
,b76,__ZNK4love5sound5Sound13getModuleTypeEv,__ZNK4love5sound7lullaby5Sound7getNameEv,b76,b76,b76,__ZN4love5sound7lullaby13VorbisDecoder5cloneEv,__ZN4love5sound7lullaby13VorbisDecoder6decodeEv,__ZNK4love5sound7lullaby7Decoder7getSizeEv,__ZNK4love5sound7lullaby7Decoder9getBufferEv,b76,__ZN4love5sound7lullaby13VorbisDecoder6rewindEv,__ZN4love5sound7lullaby13VorbisDecoder10isSeekableEv,__ZN4love5sound7lullaby7Decoder10isFinishedEv,__ZNK4love5sound7lullaby13VorbisDecoder11getChannelsEv,__ZNK4love5sound7lullaby13VorbisDecoder11getBitDepthEv,__ZNK4love5sound7lullaby13VorbisDecoder13getSampleRateEv,b76,b76,b76,b76,b76,b76,__ZN4love5sound7lullaby11WaveDecoder5cloneEv,__ZN4love5sound7lullaby11WaveDecoder6decodeEv,b76,__ZN4love5sound7lullaby11WaveDecoder6rewindEv,__ZN4love5sound7lullaby11WaveDecoder10isSeekableEv,__ZNK4love5sound7lullaby11WaveDecoder11getChannelsEv,__ZNK4love5sound7lullaby11WaveDecoder11getBitDepthEv
,__ZNK4love5sound7lullaby11WaveDecoder13getSampleRateEv,b76,b76,b76,__ZNK4love5sound9SoundData7getDataEv,__ZNK4love5sound9SoundData7getSizeEv,__ZNK4love5sound9SoundData11getChannelsEv,__ZNK4love5sound9SoundData11getBitDepthEv,__ZNK4love5sound9SoundData13getSampleRateEv,__ZNK4love5sound9SoundData14getSampleCountEv,b76,__ZN4love5sound21w_Decoder_getChannelsEP9lua_State,__ZN4love5sound21w_Decoder_getBitDepthEP9lua_State,__ZN4love5sound23w_Decoder_getSampleRateEP9lua_State,__ZN4love5sound21w_Decoder_getDurationEP9lua_State,b76,__ZNK4love6system6System13getModuleTypeEv,__ZNK4love6system3sdl6System7getNameEv,__ZNK4love6system3sdl6System17getProcessorCountEv,b76,b76,b76,b76,b76,b76,b76,__ZNK4love6thread12ThreadModule13getModuleTypeEv,__ZNK4love6thread12ThreadModule7getNameEv,b76,__ZN4love6thread12ThreadModule10newChannelEv
,b76,b76,b76,__ZN4love6thread14w_Channel_pushEP9lua_State,__ZN4love6thread16w_Channel_supplyEP9lua_State,__ZN4love6thread13w_Channel_popEP9lua_State,__ZN4love6thread16w_Channel_demandEP9lua_State,__ZN4love6thread14w_Channel_peekEP9lua_State,__ZN4love6thread18w_Channel_getCountEP9lua_State,__ZN4love6thread15w_Channel_clearEP9lua_State,__ZN4love6thread23w_Channel_performAtomicEP9lua_State,__ZN4love6thread14w_Thread_startEP9lua_State,__ZN4love6thread13w_Thread_waitEP9lua_State,__ZN4love6thread17w_Thread_getErrorEP9lua_State,__ZN4love6thread18w_Thread_isRunningEP9lua_State,b76,b76,__ZNK4love5timer5Timer13getModuleTypeEv,__ZNK4love5timer3sdl5Timer7getNameEv,b76,b76,b76,__ZNK4love5timer5Timer6getFPSEv,b76,b76,b76,b76,b76,__ZNK4love5touch5Touch13getModuleTypeEv,__ZNK4love5touch3sdl5Touch7getNameEv
,__ZNK4love5touch3sdl5Touch10getTouchesEv,b76,b76,b76,__ZNK4love5video5Video13getModuleTypeEv,__ZNK4love5video6theora5Video7getNameEv,b76,b76,b76,b76,b76,b76,b76,__ZNK4love5video6theora11VideoStream14getFrontBufferEv,__ZNK4love5video6theora11VideoStream7getSizeEv,__ZN4love5video6theora11VideoStream11swapBuffersEv,__ZNK4love5video6theora11VideoStream8getWidthEv,__ZNK4love5video6theora11VideoStream9getHeightEv,__ZNK4love5video6theora11VideoStream11getFilenameEv,b76,b76,b76,b76,__ZNK4love5video6theora11VideoStream9isPlayingEv,b76,__ZNK4love5video11VideoStream7getSyncEv,b76,b76,b76,b76
,b76,b76,b76,b76,__ZNK4love5video11VideoStream9DeltaSync9isPlayingEv,b76,b76,b76,b76,b76,b76,b76,__ZNK4love5video11VideoStream10SourceSync9isPlayingEv,b76,b76,b76,__ZNK4love5video11VideoStream9isPlayingEv,b76,b76,b76,__ZNK4love6window6Window13getModuleTypeEv,__ZNK4love6window3sdl6Window7getNameEv,b76,b76,b76,b76,b76,b76,__ZNK4love6window3sdl6Window15getDisplayCountEv,b76
,b76,b76,b76,b76,__ZNK4love6window3sdl6Window6isOpenEv,b76,__ZNK4love6window3sdl6Window14getWindowTitleEv,b76,__ZN4love6window3sdl6Window7getIconEv,b76,__ZNK4love6window3sdl6Window21isDisplaySleepEnabledEv,b76,b76,b76,__ZNK4love6window3sdl6Window8hasFocusEv,__ZNK4love6window3sdl6Window13hasMouseFocusEv,__ZNK4love6window3sdl6Window9isVisibleEv,b76,__ZNK4love6window3sdl6Window15getMouseVisibleEv,b76,__ZNK4love6window3sdl6Window14isMouseGrabbedEv,b76,b76,b76,b76,b76,b76,b76,b76,__ZNK4love6window3sdl6Window9getHandleEv
,b76,b76,b76,b76,b76,b76,__ZNK12b2ChainShape13GetChildCountEv,b76,b76,b76,b76,b76,b76,b76,__ZNK13b2CircleShape13GetChildCountEv,b76,b76,b76,b76,b76,b76,__ZNK11b2EdgeShape13GetChildCountEv,b76,b76,b76,b76,b76,b76,__ZNK14b2PolygonShape13GetChildCountEv,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,_global_skip,_global_unload,_auxiliar_open,_except_open,_timeout_open,_buffer_open,_inet_open,_tcp_open,_udp_open,_select_open,_mime_global_dot,_mime_global_b64,_mime_global_eol,_mime_global_qp,_mime_global_qpwrp,_mime_global_unb64,_mime_global_unqp,_mime_global_wrp,_global_select,_meth_close,_auxiliar_tostring,_meth_accept,_meth_bind,_meth_connect,_meth_dirty,_meth_getfd
,_meth_getpeername,_meth_getsockname,_meth_getstats,_meth_setstats,_meth_listen,_meth_receive,_meth_send,_meth_setfd,_meth_setoption,_meth_settimeout,_meth_shutdown,b76,b76,b76,b76,_global_create,_timeout_lua_gettime,_timeout_lua_sleep,_meth_close_2900,_meth_dirty_2905,_meth_getfd_2907,_meth_getpeername_2909,_meth_getsockname_2911,_meth_receive_2913,_meth_receivefrom,_meth_send_2918,_meth_sendto,_meth_setfd_2922,_meth_setoption_2924,_meth_setpeername
,_meth_setsockname,_meth_settimeout_2937,b76,b76,b76,b76,b76,b76,_global_create_2941,b76,b76,__ZN4love5sound7lullaby13Mpg123Decoder5cloneEv,__ZN4love5sound7lullaby13Mpg123Decoder6decodeEv,b76,__ZN4love5sound7lullaby13Mpg123Decoder6rewindEv,__ZN4love5sound7lullaby13Mpg123Decoder10isSeekableEv,__ZNK4love5sound7lullaby13Mpg123Decoder11getChannelsEv,__ZNK4love5sound7lullaby13Mpg123Decoder11getBitDepthEv,__ZNK4love5sound7lullaby7Decoder13getSampleRateEv,b76,b76,b76,b76,b76,b76,b76,__ZNK4love5audio5Audio13getModuleTypeEv,__ZNK4love5audio4null5Audio7getNameEv,b76,b76
,__ZNK4love5audio4null5Audio14getSourceCountEv,__ZNK4love5audio4null5Audio13getMaxSourcesEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZN4love5audio4null5Audio15getRecordedDataEv,b76,__ZN4love5audio4null5Audio9canRecordEv,__ZNK4love5audio4null5Audio16getDistanceModelEv,b76,b76,b76
,__ZN4love5audio4null6Source5cloneEv,__ZN4love5audio4null6Source4playEv,b76,b76,b76,b76,__ZNK4love5audio4null6Source9isStoppedEv,__ZNK4love5audio4null6Source8isPausedEv,__ZNK4love5audio4null6Source10isFinishedEv,__ZN4love5audio4null6Source6updateEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNK4love5audio4null6Source10isRelativeEv,b76,__ZNK4love5audio4null6Source9isLoopingEv,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNK4love5audio4null6Source11getChannelsEv,__ZNK4love5audio6Source7getTypeEv,b76,b76,b76,b76,b76,__ZNK4love5audio6openal5Audio7getNameEv,b76,b76,__ZNK4love5audio6openal5Audio14getSourceCountEv,__ZNK4love5audio6openal5Audio13getMaxSourcesEv,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZN4love5audio6openal5Audio15getRecordedDataEv,b76,__ZN4love5audio6openal5Audio9canRecordEv,__ZNK4love5audio6openal5Audio16getDistanceModelEv,b76,b76,b76,b76,b76,__ZN4love5audio6openal6Source5cloneEv,__ZN4love5audio6openal6Source4playEv,b76,b76,b76,b76,__ZNK4love5audio6openal6Source9isStoppedEv,__ZNK4love5audio6openal6Source8isPausedEv,__ZNK4love5audio6openal6Source10isFinishedEv
,__ZN4love5audio6openal6Source6updateEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNK4love5audio6openal6Source10isRelativeEv,b76,__ZNK4love5audio6openal6Source9isLoopingEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,__ZNK4love5audio6openal6Source11getChannelsEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNK4love5event5Event13getModuleTypeEv,__ZNK4love5event3sdl5Event7getNameEv,b76,b76,__ZN4love5event3sdl5Event4waitEv,b76,b76,b76,b76,b76,b76,b76,__ZNK4love10filesystem10Filesystem13getModuleTypeEv,__ZNK4love10filesystem6physfs10Filesystem7getNameEv,b76,b76,__ZNK4love10filesystem6physfs10Filesystem7isFusedEv,__ZN4love10filesystem6physfs10Filesystem19setupWriteDirectoryEv,b76,__ZNK4love10filesystem10Filesystem21isAndroidSaveExternalEv
,b76,__ZNK4love10filesystem6physfs10Filesystem11getIdentityEv,b76,__ZNK4love10filesystem6physfs10Filesystem9getSourceEv,b76,b76,b76,b76,b76,__ZN4love10filesystem6physfs10Filesystem19getWorkingDirectoryEv,b76,b76,__ZN4love10filesystem6physfs10Filesystem16getSaveDirectoryEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNK4love10filesystem6physfs10Filesystem18areSymlinksEnabledEv,__ZN4love10filesystem6physfs10Filesystem14getRequirePathEv
,b76,b76,b76,b76,b76,b76,__ZN4love10filesystem11DroppedFile5closeEv,__ZNK4love10filesystem11DroppedFile6isOpenEv,__ZN4love10filesystem11DroppedFile7getSizeEv,b76,b76,b76,b76,__ZN4love10filesystem11DroppedFile5flushEv,__ZN4love10filesystem11DroppedFile5isEOFEv,__ZN4love10filesystem11DroppedFile4tellEv,b76,b76,b76,__ZNK4love10filesystem11DroppedFile7getModeEv,__ZNK4love10filesystem11DroppedFile11getFilenameEv,b76,b76,b76,__ZN4love10filesystem14w_File_getSizeEP9lua_State,__ZN4love10filesystem11w_File_openEP9lua_State,__ZN4love10filesystem12w_File_closeEP9lua_State,__ZN4love10filesystem13w_File_isOpenEP9lua_State,__ZN4love10filesystem11w_File_readEP9lua_State,__ZN4love10filesystem12w_File_writeEP9lua_State
,__ZN4love10filesystem12w_File_flushEP9lua_State,__ZN4love10filesystem12w_File_isEOFEP9lua_State,__ZN4love10filesystem11w_File_tellEP9lua_State,__ZN4love10filesystem11w_File_seekEP9lua_State,__ZN4love10filesystem12w_File_linesEP9lua_State,__ZN4love10filesystem16w_File_setBufferEP9lua_State,__ZN4love10filesystem16w_File_getBufferEP9lua_State,__ZN4love10filesystem14w_File_getModeEP9lua_State,__ZN4love10filesystem18w_File_getFilenameEP9lua_State,__ZN4love10filesystem19w_File_getExtensionEP9lua_State,__ZN4love10filesystem22w_FileData_getFilenameEP9lua_State,__ZN4love10filesystem23w_FileData_getExtensionEP9lua_State,b76,b76,__ZNK4love4font4Font13getModuleTypeEv,__ZNK4love4font8freetype4Font7getNameEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNK4love4font10Rasterizer9getHeightEv,__ZNK4love4font10Rasterizer10getAdvanceEv,__ZNK4love4font10Rasterizer9getAscentEv,__ZNK4love4font10Rasterizer10getDescentEv
,__ZNK4love4font8freetype18TrueTypeRasterizer13getLineHeightEv,b76,b76,__ZNK4love4font8freetype18TrueTypeRasterizer13getGlyphCountEv,b76,b76,b76,b76,b76,__ZNK4love4font16BMFontRasterizer13getLineHeightEv,b76,__ZNK4love4font16BMFontRasterizer13getGlyphCountEv,b76,b76,b76,__ZNK4love4font15DefaultFontData7getDataEv,__ZNK4love4font15DefaultFontData7getSizeEv,b76,b76,__ZNK4utf815not_enough_room4whatEv,b76,__ZNK4utf812invalid_utf84whatEv,b76,__ZNK4utf818invalid_code_point4whatEv,b76,b76,__ZNK4love4font9GlyphData7getDataEv,__ZNK4love4font9GlyphData7getSizeEv,__ZNK4love4font9GlyphData9getHeightEv,__ZNK4love4font9GlyphData8getWidthEv
,b76,b76,__ZNK4love4font15ImageRasterizer13getLineHeightEv,b76,__ZNK4love4font15ImageRasterizer13getGlyphCountEv,b76,b76,b76,b76,b76,b76,b76,b76,__ZNK4love8graphics6opengl6Canvas9getHandleEv,__ZN4love8graphics6opengl6Canvas12loadVolatileEv,b76,b76,b76,__ZThn120_N4love8graphics6opengl6Canvas12loadVolatileEv,b76,b76,b76,__ZN4love8graphics6opengl4Font12loadVolatileEv,b76,b76,b76,__ZThn8_N4love8graphics6opengl4Font12loadVolatileEv,b76,b76,b76
,b76,b76,b76,b76,__ZN4love8graphics6opengl8GLBuffer12loadVolatileEv,b76,b76,b76,__ZNK4love8graphics8Graphics13getModuleTypeEv,__ZNK4love8graphics6opengl8Graphics7getNameEv,b76,b76,b76,b76,__ZNK4love8graphics6opengl8Graphics8isActiveEv,b76,b76,b76,b76,b76,b76,__ZNK4love8graphics6opengl5Image9getHandleEv,__ZN4love8graphics6opengl5Image12loadVolatileEv,b76,b76,b76,__ZThn120_N4love8graphics6opengl5Image12loadVolatileEv,b76,b76,b76
,b76,b76,b76,b76,__ZN4love8graphics6opengl14ParticleSystem5cloneEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZN4love8graphics6opengl6Shader12loadVolatileEv,b76,b76,b76,__ZThn8_N4love8graphics6opengl6Shader12loadVolatileEv,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,__ZN4love8graphics6opengl5Video12loadVolatileEv,b76,b76,b76,__ZThn8_N4love8graphics6opengl5Video12loadVolatileEv,b76,b76,__ZN4love8graphics6opengl17w_Canvas_renderToEP9lua_State,__ZN4love8graphics6opengl21w_Canvas_newImageDataEP9lua_State,__ZN4love8graphics6opengl18w_Canvas_getFormatEP9lua_State,__ZN4love8graphics6opengl16w_Canvas_getMSAAEP9lua_State,__ZN4love8graphics6opengl16w_Font_getHeightEP9lua_State,__ZN4love8graphics6opengl15w_Font_getWidthEP9lua_State,__ZN4love8graphics6opengl14w_Font_getWrapEP9lua_State,__ZN4love8graphics6opengl20w_Font_setLineHeightEP9lua_State,__ZN4love8graphics6opengl20w_Font_getLineHeightEP9lua_State,__ZN4love8graphics6opengl16w_Font_setFilterEP9lua_State,__ZN4love8graphics6opengl16w_Font_getFilterEP9lua_State,__ZN4love8graphics6opengl16w_Font_getAscentEP9lua_State,__ZN4love8graphics6opengl17w_Font_getDescentEP9lua_State,__ZN4love8graphics6opengl18w_Font_getBaselineEP9lua_State,__ZN4love8graphics6opengl16w_Font_hasGlyphsEP9lua_State,__ZN4love8graphics6opengl19w_Font_setFallbacksEP9lua_State
,b76,b76,b76,b76,__ZNK4love5image5Image13getModuleTypeEv,__ZNK4love5image6magpie5Image7getNameEv,b76,b76,b76,b76,b76,b76,b76,__ZNK4love5image9ImageData7getDataEv,__ZNK4love5image9ImageData7getSizeEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNK4love8joystick3sdl8Joystick11isConnectedEv,__ZNK4love8joystick3sdl8Joystick7getNameEv,__ZNK4love8joystick3sdl8Joystick12getAxisCountEv,__ZNK4love8joystick3sdl8Joystick14getButtonCountEv,__ZNK4love8joystick3sdl8Joystick11getHatCountEv,b76,b76,b76,b76,b76,__ZNK4love8joystick3sdl8Joystick9isGamepadEv,b76,b76,__ZNK4love8joystick3sdl8Joystick9getHandleEv,b76,__ZNK4love8joystick3sdl8Joystick13getInstanceIDEv
,__ZNK4love8joystick3sdl8Joystick5getIDEv,__ZN4love8joystick3sdl8Joystick20isVibrationSupportedEv,b76,__ZN4love8joystick3sdl8Joystick12setVibrationEv,b76,b76,b76,__ZNK4love8joystick14JoystickModule13getModuleTypeEv,__ZNK4love8joystick3sdl14JoystickModule7getNameEv,b76,b76,b76,b76,b76,__ZNK4love8joystick3sdl14JoystickModule16getJoystickCountEv,b76,b76,b76,b76,__ZN4love8joystick22w_Joystick_isConnectedEP9lua_State,__ZN4love8joystick18w_Joystick_getNameEP9lua_State,__ZN4love8joystick16w_Joystick_getIDEP9lua_State,__ZN4love8joystick18w_Joystick_getGUIDEP9lua_State,__ZN4love8joystick23w_Joystick_getAxisCountEP9lua_State,__ZN4love8joystick25w_Joystick_getButtonCountEP9lua_State,__ZN4love8joystick22w_Joystick_getHatCountEP9lua_State,__ZN4love8joystick18w_Joystick_getAxisEP9lua_State,__ZN4love8joystick18w_Joystick_getAxesEP9lua_State,__ZN4love8joystick17w_Joystick_getHatEP9lua_State,__ZN4love8joystick17w_Joystick_isDownEP9lua_State
,__ZN4love8joystick20w_Joystick_isGamepadEP9lua_State,__ZN4love8joystick25w_Joystick_getGamepadAxisEP9lua_State,__ZN4love8joystick24w_Joystick_isGamepadDownEP9lua_State,__ZN4love8joystick31w_Joystick_isVibrationSupportedEP9lua_State,__ZN4love8joystick23w_Joystick_setVibrationEP9lua_State,__ZN4love8joystick23w_Joystick_getVibrationEP9lua_State,__ZN4love8joystick10w_getIndexEP9lua_State,b76,__ZNK4love8keyboard8Keyboard13getModuleTypeEv,__ZNK4love8keyboard3sdl8Keyboard7getNameEv,b76,__ZNK4love8keyboard3sdl8Keyboard12hasKeyRepeatEv,b76,b76,b76,b76,b76,b76,__ZNK4love8keyboard3sdl8Keyboard12hasTextInputEv,__ZNK4love8keyboard3sdl8Keyboard17hasScreenKeyboardEv,b76,b76,b76,b76,__ZNK4love5mouse3sdl6Cursor9getHandleEv,__ZNK4love5mouse3sdl6Cursor7getTypeEv,__ZNK4love5mouse3sdl6Cursor13getSystemTypeEv,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZN4love5sound7lullaby14ModPlugDecoder5cloneEv,__ZN4love5sound7lullaby14ModPlugDecoder6decodeEv,b76,__ZN4love5sound7lullaby14ModPlugDecoder6rewindEv,__ZN4love5sound7lullaby14ModPlugDecoder10isSeekableEv,__ZNK4love5sound7lullaby14ModPlugDecoder11getChannelsEv,__ZNK4love5sound7lullaby14ModPlugDecoder11getBitDepthEv
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,_global_newtry,_global_protect,_inet_global_toip,_inet_global_tohostname,_inet_global_gethostname,b76,b76,b76,__ZN4love10filesystem6physfs4File5closeEv,__ZNK4love10filesystem6physfs4File6isOpenEv,__ZN4love10filesystem6physfs4File7getSizeEv,b76,b76,__ZN4love10filesystem6physfs4File5flushEv,__ZN4love10filesystem6physfs4File5isEOFEv,__ZN4love10filesystem6physfs4File4tellEv,b76,b76,b76,__ZNK4love10filesystem6physfs4File7getModeEv,__ZNK4love10filesystem6physfs4File11getFilenameEv,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,__ZN4love6thread3sdl6Thread5startEv,b76,__ZN4love6thread3sdl6Thread9isRunningEv,b76,b76,b76,_bad_init,b76,b76,b76,b76,b76,b76,b76,b76,_bad_tell,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_default_init,b76,b76,b76,b76,b76,b76,b76,b76,_generic_tell,b76,b76,_feed_init,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_DIR_eof,_DIR_tell,b76,_DIR_fileLength,_DIR_fileClose,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,_ZIP_eof,_ZIP_tell,b76,_ZIP_fileLength,_ZIP_fileClose,_luaB_assert,_luaB_collectgarbage,_luaB_dofile,_luaB_error,_luaB_gcinfo,_luaB_getfenv,_luaB_getmetatable,_luaB_loadfile,_luaB_load,_luaB_loadstring,_luaB_next,_luaB_pcall,_luaB_print,_luaB_rawequal,_luaB_rawget,_luaB_rawset,_luaB_select
,_luaB_setfenv,_luaB_setmetatable,_luaB_tonumber,_luaB_tostring,_luaB_type,_luaB_unpack,_luaB_xpcall,_luaB_cocreate,_luaB_coresume,_luaB_corunning,_luaB_costatus,_luaB_cowrap,_luaB_yield,_db_debug,_db_getfenv,_db_gethook,_db_getinfo,_db_getlocal,_db_getregistry,_db_getmetatable,_db_getupvalue,_db_setfenv,_db_sethook,_db_setlocal,_db_setmetatable,_db_setupvalue,_db_errorfb,_io_close,_f_flush,_f_lines
,_f_read,_f_seek,_f_setvbuf,_f_write,_io_gc,_io_tostring,_io_flush,_io_input,_io_lines,_io_open,_io_output,_io_popen,_io_read,_io_tmpfile,_io_type,_io_write,_math_abs,_math_acos,_math_asin,_math_atan2,_math_atan,_math_ceil,_math_cosh,_math_cos,_math_deg,_math_exp,_math_floor,_math_fmod,_math_frexp,_math_ldexp
,_math_log10,_math_log,_math_max,_math_min,_math_modf,_math_pow,_math_rad,_math_random,_math_randomseed,_math_sinh,_math_sin,_math_sqrt,_math_tanh,_math_tan,_os_clock,_os_date,_os_difftime,_os_execute,_os_exit,_os_getenv,_os_remove,_os_rename,_os_setlocale,_os_time,_os_tmpname,_tconcat,_foreach,_foreachi,_getn,_maxn
,_tinsert,_tremove,_setn,_sort,_str_byte,_str_char,_str_dump,_str_find,_str_format,_gfind_nodef,_gmatch,_str_gsub,_str_len,_str_lower,_str_match,_str_rep,_str_reverse,_str_sub,_str_upper,_ll_loadlib,_ll_seeall,_ll_module,_ll_require,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_cff_driver_init,b76,b76,b76,b76,_cff_size_init,b76,_cff_slot_init,b76,b76,b76,b76,b76
,b76,b76,_cff_ps_has_glyph_names,_cff_get_ps_name,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_cff_parse_font_matrix,_cff_parse_font_bbox,_cff_parse_private_dict,_cff_parse_cid_ros,_cid_driver_init,b76,b76,b76,b76,_cid_size_init
,b76,_cid_slot_init,b76,b76,b76,_cid_get_postscript_name,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,_pfr_slot_init,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_ps_parser_to_int,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,_t1_builder_add_contour,b76,b76,b76,b76,b76,b76,b76,_afm_parser_parse,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_ps_hinter_init,b76,_pshinter_get_globals_funcs,_pshinter_get_t1_funcs,_pshinter_get_t2_funcs
,b76,_ps_unicode_value,b76,b76,b76,_ps_get_macintosh_name,_ps_get_standard_strings,b76,b76,b76,b76,b76,_ft_raster1_init,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,_sfnt_get_ps_name,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_ft_smooth_init,b76,b76,b76,b76,b76,b76,_tt_driver_init,b76,b76
,b76,b76,_tt_size_init,b76,_tt_slot_init,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_T1_Driver_Init,b76,b76,b76,b76,_T1_Size_Init,b76,_T1_GlyphSlot_Init,b76,b76,b76,b76,b76,b76
,_t1_get_ps_name,b76,b76,b76,b76,_t1_ps_has_glyph_names,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_T42_Driver_Init,b76,b76,b76,b76,_T42_Size_Init
,b76,_T42_GlyphSlot_Init,b76,b76,b76,b76,b76,b76,_t42_get_ps_font_name,b76,b76,_t42_ps_has_glyph_names,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_af_autofitter_init,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,_mapping0_forward,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_getpix_8,_getpix_16,_getpix_24,_getpix_32,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_Emscripten_CreateDevice,_Emscripten_Init,b76,_DUMMY_CreateDevice,_DISKAUD_Init,_DUMMYAUD_Init
,b76,b76,b76,___stdio_close,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNSt3__110__stdinbufIcE9underflowEv,__ZNSt3__110__stdinbufIcE5uflowEv,b76,b76,b76,b76,__ZNSt3__111__stdoutbufIcE4syncEv,__ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE9underflowEv,b76,b76,b76,b76,b76,b76,b76,b76
,b76,__ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE4syncEv,__ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE9showmanycEv,b76,__ZNSt3__110__stdinbufIwE9underflowEv,__ZNSt3__110__stdinbufIwE5uflowEv,b76,b76,b76,b76,b76,__ZNSt3__111__stdoutbufIwE4syncEv,__ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE9underflowEv,__ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE5uflowEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,__ZNKSt13runtime_error4whatEv,b76,b76,__ZNKSt3__119__iostream_category4nameEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE13do_date_orderEv,b76,b76,b76,b76,b76,b76,__ZNKSt3__120__time_get_c_storageIcE7__weeksEv,__ZNKSt3__120__time_get_c_storageIcE8__monthsEv,__ZNKSt3__120__time_get_c_storageIcE7__am_pmEv,__ZNKSt3__120__time_get_c_storageIcE3__cEv,__ZNKSt3__120__time_get_c_storageIcE3__rEv,__ZNKSt3__120__time_get_c_storageIcE3__xEv,__ZNKSt3__120__time_get_c_storageIcE3__XEv,b76,b76,__ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE13do_date_orderEv,b76,b76,b76,b76
,b76,b76,__ZNKSt3__120__time_get_c_storageIwE7__weeksEv,__ZNKSt3__120__time_get_c_storageIwE8__monthsEv,__ZNKSt3__120__time_get_c_storageIwE7__am_pmEv,__ZNKSt3__120__time_get_c_storageIwE3__cEv,__ZNKSt3__120__time_get_c_storageIwE3__rEv,__ZNKSt3__120__time_get_c_storageIwE3__xEv,__ZNKSt3__120__time_get_c_storageIwE3__XEv,b76,b76,b76,b76,b76,b76,b76,b76,__ZNKSt3__110moneypunctIcLb0EE16do_decimal_pointEv,__ZNKSt3__110moneypunctIcLb0EE16do_thousands_sepEv,b76,b76,b76,b76,__ZNKSt3__110moneypunctIcLb0EE14do_frac_digitsEv,b76,b76,b76,b76,__ZNKSt3__110moneypunctIcLb1EE16do_decimal_pointEv,__ZNKSt3__110moneypunctIcLb1EE16do_thousands_sepEv
,b76,b76,b76,b76,__ZNKSt3__110moneypunctIcLb1EE14do_frac_digitsEv,b76,b76,b76,b76,__ZNKSt3__110moneypunctIwLb0EE16do_decimal_pointEv,__ZNKSt3__110moneypunctIwLb0EE16do_thousands_sepEv,b76,b76,b76,b76,__ZNKSt3__110moneypunctIwLb0EE14do_frac_digitsEv,b76,b76,b76,b76,__ZNKSt3__110moneypunctIwLb1EE16do_decimal_pointEv,__ZNKSt3__110moneypunctIwLb1EE16do_thousands_sepEv,b76,b76,b76,b76,__ZNKSt3__110moneypunctIwLb1EE14do_frac_digitsEv,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,__ZNKSt3__17codecvtIDic11__mbstate_tE11do_encodingEv,__ZNKSt3__17codecvtIDic11__mbstate_tE16do_always_noconvEv,b76,__ZNKSt3__17codecvtIDic11__mbstate_tE13do_max_lengthEv,b76,b76,b76,b76,b76,b76,__ZNKSt3__17codecvtIwc11__mbstate_tE11do_encodingEv,__ZNKSt3__17codecvtIwc11__mbstate_tE16do_always_noconvEv,b76,__ZNKSt3__17codecvtIwc11__mbstate_tE13do_max_lengthEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNKSt3__18numpunctIcE16do_decimal_pointEv,__ZNKSt3__18numpunctIcE16do_thousands_sepEv
,b76,b76,b76,b76,b76,__ZNKSt3__18numpunctIwE16do_decimal_pointEv,__ZNKSt3__18numpunctIwE16do_thousands_sepEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNKSt3__17codecvtIcc11__mbstate_tE11do_encodingEv,__ZNKSt3__17codecvtIcc11__mbstate_tE16do_always_noconvEv
,b76,__ZNKSt3__17codecvtIcc11__mbstate_tE13do_max_lengthEv,b76,b76,b76,b76,__ZNKSt3__17codecvtIDsc11__mbstate_tE11do_encodingEv,__ZNKSt3__17codecvtIDsc11__mbstate_tE16do_always_noconvEv,b76,__ZNKSt3__17codecvtIDsc11__mbstate_tE13do_max_lengthEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNKSt9bad_alloc4whatEv,b76
,b76,__ZNKSt11logic_error4whatEv,b76,b76,b76,b76,b76,b76,b76,__ZNKSt8bad_cast4whatEv,b76,b76,__ZNKSt10bad_typeid4whatEv,b76,b76,b76,b76,_luaopen_love,b76,__Znwj,b76,b76,b76,b76,__ZL18w__setGammaCorrectP9lua_State,__ZL17w_love_getVersionP9lua_State,__ZL26w_love_isVersionCompatibleP9lua_State,__Z12luaopen_enetP9lua_State,_luaopen_luautf8,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNKSt3__18ios_base6getlocEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_lua_gettop,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNK4love10filesystem8FileData11getFilenameEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZL13peer_tostringP9lua_State,b76
,b76,b76,_luaopen_socket_core,_luaopen_mime_core,__ZN4love9luasocket23__open_luasocket_socketEP9lua_State,__ZN4love9luasocket20__open_luasocket_ftpEP9lua_State,__ZN4love9luasocket21__open_luasocket_httpEP9lua_State,__ZN4love9luasocket22__open_luasocket_ltn12EP9lua_State,__ZN4love9luasocket21__open_luasocket_mimeEP9lua_State,__ZN4love9luasocket21__open_luasocket_smtpEP9lua_State,__ZN4love9luasocket19__open_luasocket_tpEP9lua_State,__ZN4love9luasocket20__open_luasocket_urlEP9lua_State,_iter_aux,__Znaj,b76,__ZN4loveL5w__gcEP9lua_State,__ZN4loveL5w__eqEP9lua_State,__ZN4loveL11w__tostringEP9lua_State,__ZN4loveL7w__typeEP9lua_State,__ZN4loveL9w__typeOfEP9lua_State,b76,b76,b76,__ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEE5flushEv,b76,b76,__ZN4love5eventL8w_poll_iEP9lua_State,b76,b76,b76
,b76,b76,b76,b76,__ZN4love10filesystem14w_File_lines_iEP9lua_State,b76,b76,b76,b76,b76,b76,b76,b76,__ZN4love10filesystem6loaderEP9lua_State,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,__ZNK4love8graphics6opengl8Graphics16getDefaultFilterEv,b76,b76,b76,b76,b76,b76,b76,b76,_lua_error,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,__ZN4love8graphics6opengl8Graphics7getFontEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNK4love8graphics6opengl4Mesh15getVertexStrideEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNK4love7physics5box2d9GearJoint9getJointAEv,__ZNK4love7physics5box2d9GearJoint9getJointBEv,__ZNK4love7physics5box2d5Joint8getBodyAEv,__ZNK4love7physics5box2d5Joint8getBodyBEv,b76,b76,b76,b76,b76,_tolower
,__ZN4love5sound7lullaby14ModPlugDecoder7acceptsERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE,b76,__ZN4love5sound7lullaby13Mpg123Decoder7acceptsERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE,b76,__ZN4love5sound7lullaby13VorbisDecoder7acceptsERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE,b76,__ZN4love5sound7lullaby11WaveDecoder7acceptsERKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE,b76,_ov_clear,__ZN4love5sound7lullabyL11vorbisCloseEPv,b76,b76,__ZN4love5sound7lullabyL10vorbisTellEPv,b76,b76,b76,b76,_wuff_close,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZN4love6thread9newThreadEPNS0_10ThreadableE
,b76,b76,b76,__ZNK4love6thread8MutexRefcvPNS0_5MutexEEv,b76,__ZNK4love6Object17getReferenceCountEv,b76,b76,__ZN4love6thread10Threadable5startEv,b76,b76,b76,_ogg_stream_clear,_ogg_sync_clear,_ogg_sync_init,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_SDL_GetNumDisplayModes,b76
,b76,b76,b76,b76,_SDL_InitSubSystem,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__Z7b2Alloci,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_mpg123_format_none,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,_alcMakeContextCurrent__wrapper,b76,_alcCloseDevice__wrapper,b76,b76,b76,_alcOpenDevice__wrapper,b76,_alcGetError__wrapper,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZN4love5audio6openal6Source10playAtomicEv,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZN4love6Module11getInstanceERKNSt3__112basic_stringIcNS1_11char_traitsIcEENS1_9allocatorIcEEEE,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_SDL_GetWindowFromID,b76,b76
,b76,b76,b76,b76,b76,b76,_PHYSFS_setWriteDir,b76,b76,_PHYSFS_removeFromSearchPath,_PHYSFS_getRealDir,_PHYSFS_getMountPoint,b76,b76,b76,b76,b76,b76,_FT_Done_FreeType,b76,_FT_Init_FreeType,_FT_Done_Face,b76,b76,b76,__ZNK4love5image9ImageData8getWidthEv,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZN4love8graphics6opengl6OpenGL12getTransformEv,b76,b76,b76,b76,b76,b76,__ZNK4love8graphics4Quad11getVerticesEv,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,__ZNK4love8graphics6opengl6OpenGL17getMaxTextureSizeEv,__ZNK4love4font9GlyphData9getFormatEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNK4love8graphics6opengl6OpenGL13getDefaultFBOEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,__ZN4love8graphics6opengl6OpenGL11errorStringEj,__ZNK4love5image9ImageData9getHeightEv,b76,__ZNK4love5image19CompressedImageData14getMipmapCountEv,b76,__ZNK4love5image19CompressedImageData9getFormatEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZN4love8graphics6opengl8GLBuffer3mapEv,b76,b76,b76,b76,__ZN4love8graphics6openglL18LOVEGetProcAddressEPKc,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNK4love7Matrix411getElementsEv,b76,b76,__ZNK4love7Matrix311getElementsEv,b76,b76,b76,b76,__ZNK4love8graphics6opengl6OpenGL18getMaxTextureUnitsEv,b76,b76,b76,b76,__ZNK4love8graphics6opengl4Mesh14getVertexCountEv,b76,__ZNK4love8graphics6opengl11QuadIndices9getBufferEv,b76,__ZN4love8graphics6opengl4Mesh16getGLBufferUsageENS2_5UsageE,b76,b76
,b76,b76,b76,__ZNK4love8graphics6opengl4Font17getTextureCacheIDEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,__Z18lodepng_error_textj,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_SDL_WasInit,b76,b76,_SDL_GameControllerMappingForGUID,b76,b76,b76,_SDL_GameControllerAddMapping,b76,b76,b76,b76,b76
,b76,_SDL_JoystickEventState,_SDL_GameControllerEventState,b76,b76,b76,b76,b76,b76,b76,_SDL_CreateSystemCursor,b76,b76,b76,b76,__ZNK4love7physics5box2d5Shape7getTypeEv,b76,b76,b76,__ZNK4love7physics5box2d5World13getGroundBodyEv,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,__ZN4love6thread7Channel3popEv,b76,b76,b76,b76,b76,b76,__ZN15b2ContactSolver24SolvePositionConstraintsEv,b76,b76,_do_nothing,_finalize,_protected_,_PHYSFS_close,b76,__ZNK3dds6Parser9getFormatEv,__ZNK3dds6Parser14getMipmapCountEv,b76,b76,b76,b76,__ZNK4love6thread10Threadable13getThreadNameEv
,b76,__ZN4love6thread3sdl6Thread13thread_runnerEPv,b76,b76,b76,_INT123_do_layer1,_INT123_do_layer2,_INT123_do_layer3,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,___PHYSFS_platformOpenRead,___PHYSFS_platformOpenWrite,___PHYSFS_platformOpenAppend,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_panic,_luaopen_base,_luaopen_package,_luaopen_table,_luaopen_io,_luaopen_os,_luaopen_string,_luaopen_math,_luaopen_debug,_ipairsaux,_luaB_ipairs,_luaB_pairs,_luaB_newproxy,b76,_luaB_auxwrap,b76,_io_fclose,_io_noclose,_io_pclose,_io_readline
,b76,_gmatch_aux,_gctm,_loader_preload,_loader_Lua,_loader_C,_loader_Croot,b76,b76,b76,b76,_oc_clc_mode_unpack,_oc_vlc_mode_unpack,_oc_clc_mv_comp_unpack,_oc_vlc_mv_comp_unpack,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_TT_Load_Glyph_Header,_TT_Load_Simple_Glyph,_TT_Load_Composite_Glyph,b76
,_TT_RunIns,_Current_Ppem,_Current_Ppem_Stretched,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_GLES2_UpdateViewport,_GLES2_UpdateClipRect,_GLES2_RenderClear,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,_SW_UpdateViewport,_SW_UpdateClipRect,_SW_RenderClear,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,_Emscripten_VideoInit,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_SDL_EGL_GetSwapInterval,b76,b76,b76,b76,b76,_Emscripten_ShowCursor
,b76,b76,_Emscripten_CreateSystemCursor,_Emscripten_SetRelativeMouseMode,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_eglGetDisplay__wrapper,b76,_eglTerminate__wrapper,_eglGetProcAddress__wrapper,b76,b76,b76,b76,b76,b76,b76,b76,b76,_eglWaitNative__wrapper,b76,_eglBindAPI__wrapper,b76,b76,b76,_DUMMY_VideoInit,b76,b76,b76,b76,b76,b76,b76,_stdio_size,b76,b76,b76,_stdio_close,b76,b76,b76,_DISKAUD_GetDeviceBuf,b76,b76,b76,b76,_SDL_TimerThread,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_SDL_AudioGetPendingBytes_Default,_SDL_AudioGetDeviceBuf_Default,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,_emscripten_glGetString__wrapper,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_emscripten_glIsTexture__wrapper,b76,b76,b76,b76,b76,_emscripten_glIsBuffer__wrapper,b76,b76,b76,b76
,_emscripten_glIsRenderbuffer__wrapper,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,_emscripten_glCreateShader__wrapper,b76,b76,b76,b76,b76,b76,b76,b76,_emscripten_glIsShader__wrapper,b76,b76,b76,b76,b76,b76,b76,b76,b76,_emscripten_glIsProgram__wrapper,b76,b76,b76,b76,b76,b76
,b76,_emscripten_glIsFramebuffer__wrapper,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_emscripten_glCheckFramebufferStatus__wrapper,b76,b76,b76,_emscripten_glIsEnabled__wrapper,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,__ZNSt3__113basic_ostreamIwNS_11char_traitsIwEEE5flushEv,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_uselocale,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,_wcslen,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,_pthread_mutex_unlock__wrapper,_pthread_mutex_lock__wrapper,_pthread_cond_broadcast__wrapper,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76,b76
,b76,b76,b76];
var FUNCTION_TABLE_ddddd = [b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77
,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77
,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,__ZN4love4math3__88__invokeEffff,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77
,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77,b77
,b77,b77,b77,b77,b77,b77,b77,b77,b77];
var FUNCTION_TABLE_iiiddddddddd = [b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,__ZN4love8graphics6opengl4Text3addERKNSt3__16vectorINS1_4Font13ColoredStringENS3_9allocatorIS6_EEEEfffffffff,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78
,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78,b78];
var FUNCTION_TABLE_iiii = [b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE6setbufEPci,b79,b79,b79,b79,__ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE6xsgetnEPci,b79,b79,b79,__ZNSt3__115basic_streambufIcNS_11char_traitsIcEEE6xsputnEPKci,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZN4love7physics5box2d5World13ShouldCollideEP9b2FixtureS4_,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZThn12_N4love7physics5box2d5World13ShouldCollideEP9b2FixtureS4_,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,__ZN4love5sound7lullaby5Sound10newDecoderEPNS_10filesystem8FileDataEi,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZN4love5sound7lullabyL13read_callbackEPvPhPj,__ZN4love5sound7lullabyL13seek_callbackEPvy,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZNK4love6system3sdl6System12getPowerInfoERiS3_,b79,b79,b79,b79,b79,b79,__ZN4love6thread12ThreadModule9newThreadERKNSt3__112basic_stringIcNS2_11char_traitsIcEENS2_9allocatorIcEEEEPNS_4DataE,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,__ZNK4love5touch3sdl5Touch8getTouchEx,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZN4love6window3sdl6Window13setFullscreenEbNS0_6Window14FullscreenTypeE,b79,__ZN4love6window3sdl6Window13onSizeChangedEii,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,__ZNK12b2ChainShape9TestPointERK11b2TransformRK6b2Vec2,b79,b79,b79,b79,b79,b79,b79,__ZNK13b2CircleShape9TestPointERK11b2TransformRK6b2Vec2,b79,b79,b79,b79,b79,b79,__ZNK11b2EdgeShape9TestPointERK11b2TransformRK6b2Vec2,b79,b79,b79,b79,b79,b79,__ZNK14b2PolygonShape9TestPointERK11b2TransformRK6b2Vec2
,b79,b79,b79,b79,b79,__ZN15b2ContactFilter13ShouldCollideEP9b2FixtureS1_,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,__ZN4love10filesystem6physfs10Filesystem11setIdentityEPKcb,b79,b79,b79,b79,b79,b79,b79,__ZNK4love10filesystem6physfs10Filesystem11newFileDataEPKcS4_,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZN4love10filesystem4File4readEx,b79,b79,b79,b79,b79,b79,__ZN4love10filesystem11DroppedFile4seekEy,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZN4love4font4Font21newTrueTypeRasterizerEiNS0_18TrueTypeRasterizer7HintingE,b79,__ZN4love4font4Font19newBMFontRasterizerEPNS_10filesystem8FileDataERKNSt3__16vectorIPNS_5image9ImageDataENS5_9allocatorIS9_EEEE,b79,b79,__ZN4love4font4Font12newGlyphDataEPNS0_10RasterizerERKNSt3__112basic_stringIcNS4_11char_traitsIcEENS4_9allocatorIcEEEE,__ZN4love4font4Font12newGlyphDataEPNS0_10RasterizerEj,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZN4love8graphics6opengl8Graphics7setModeEii,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,__ZN4love5image6magpie5Image12newImageDataEii,b79,b79,b79,b79,b79,b79,b79,__ZN4love5image6magpie9ImageData6encodeENS0_9ImageData13EncodedFormatEPKc,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZN4love6thread3sdl11Conditional4waitEPNS0_5MutexEi,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZN4love10filesystem6physfs4File4seekEy,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_bad_fullread,b79,b79,b79,_bad_read_frame_body,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_synth_stereo_wrap,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_plain_fullread,b79,b79,b79,_generic_read_frame_body,b79,b79,b79,b79,_icy_fullread,b79,_feed_read,b79,b79
,b79,b79,_buffered_fullread,b79,b79,b79,b79,_DIR_isDirectory,_DIR_isSymLink,_DIR_getLastModTime,_DIR_openRead,b79,b79,b79,b79,b79,b79,b79,b79,b79,_DIR_seek,b79,b79,b79,b79,b79,b79,_ZIP_isDirectory,_ZIP_isSymLink,_ZIP_getLastModTime
,_ZIP_openRead,b79,b79,b79,b79,b79,b79,b79,b79,b79,_ZIP_seek,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_bdf_get_charset_id,_bdf_get_bdf_property,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_cff_get_cid_from_glyph_index,_cff_property_set,_cff_property_get,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_cid_get_cid_from_glyph_index,b79,b79,b79,b79,b79,b79,b79,b79,b79,_pcf_get_charset_id,_pcf_get_bdf_property,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,_pfr_get_advance,_ps_table_new,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_ps_parser_to_coord_array,b79,b79,b79,b79,b79,b79,b79,b79
,b79,_t1_builder_add_point1,b79,_t1_builder_start_point,b79,b79,b79,_t1_decoder_parse_charstrings,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,_ft_black_set_mode,b79,b79,b79,b79,b79,b79,_ft_raster1_set_mode,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_tt_cmap14_char_var_isdefault,b79
,_tt_cmap14_char_variants,_tt_cmap14_variant_chars,b79,b79,b79,b79,b79,b79,b79,b79,b79,_tt_face_load_hhea,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_tt_face_get_ps_name,b79,_tt_face_get_kerning,b79,_tt_face_load_hmtx,b79,b79
,_tt_face_set_sbit_strike,_tt_face_load_strike_metrics,b79,b79,b79,b79,b79,b79,_sfnt_get_charset_id,_tt_face_find_bdf_prop,b79,b79,b79,_gray_raster_set_mode,b79,b79,b79,b79,_gray_conic_to,b79,b79,b79,b79,b79,_ft_smooth_set_mode,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_TT_Set_MM_Blend,b79,_TT_Set_Var_Design,_tt_face_get_location,_tt_property_set,_tt_property_get,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_T1_Set_MM_Design,_T1_Set_MM_Blend,b79,_T1_Set_Var_Design,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_af_dummy_hints_apply,b79
,b79,b79,_af_latin_hints_apply,b79,b79,b79,_af_cjk_hints_apply,b79,b79,b79,_af_indic_hints_apply,b79,b79,b79,b79,_af_property_set,_af_property_get,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,___stdio_write,___stdio_seek,___stdio_read,___stdout_write,_sn_write,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZNSt3__111__stdoutbufIcE6xsputnEPKci,b79,b79,b79,b79,__ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE6setbufEPwi,b79
,b79,b79,b79,__ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE6xsgetnEPwi,b79,b79,b79,__ZNSt3__115basic_streambufIwNS_11char_traitsIwEEE6xsputnEPKwi,b79,b79,b79,b79,b79,b79,b79,__ZNSt3__111__stdoutbufIwE6xsputnEPKwi,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,__ZNKSt3__114error_category10equivalentEiRKNS_15error_conditionE,__ZNKSt3__114error_category10equivalentERKNS_10error_codeEi,b79,b79,b79,b79,b79,b79,__ZNKSt3__17collateIcE7do_hashEPKcS3_,b79,b79,b79,b79,__ZNKSt3__17collateIwE7do_hashEPKwS3_,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZNKSt3__18messagesIcE7do_openERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERKNS_6localeE,b79,b79,b79,b79,__ZNKSt3__18messagesIwE7do_openERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERKNS_6localeE,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZNKSt3__15ctypeIcE10do_toupperEPcPKc,b79,__ZNKSt3__15ctypeIcE10do_tolowerEPcPKc,b79,b79,__ZNKSt3__15ctypeIcE9do_narrowEcc,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZNKSt3__15ctypeIwE5do_isEtw,b79,b79,b79,b79,__ZNKSt3__15ctypeIwE10do_toupperEPwPKw,b79,__ZNKSt3__15ctypeIwE10do_tolowerEPwPKw,b79,b79,__ZNKSt3__15ctypeIwE9do_narrowEwc,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZNK10__cxxabiv117__class_type_info9can_catchEPKNS_16__shim_type_infoERPv,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,_luaL_checklstring,b79,b79,b79,b79,b79,b79,__ZNSt3__124__put_character_sequenceIcNS_11char_traitsIcEEEERNS_13basic_ostreamIT_T0_EES7_PKS4_j,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,_luaL_error,b79,b79,_lua_tolstring,b79,b79,b79,b79,b79,b79,b79,__ZN4love4math4Math10decompressEPNS0_14CompressedDataERj,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZN4love7physics5box2d7Physics13newMotorJointEPNS1_4BodyES4_,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZN4love11luax_istypeEP9lua_StateiNS_4TypeE,__ZN4love14luax_checktypeINS_10filesystem8FileDataEEEPT_P9lua_StateiNS_4TypeE,b79,b79,__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6appendEPKcj,__ZN4love14luax_checktypeINS_4DataEEEPT_P9lua_StateiNS_4TypeE,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZN4love15luax_optbooleanEP9lua_Stateib,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZN4love7Variant7fromLuaEP9lua_Stateib,b79,b79
,b79,b79,b79,__ZN4love10filesystem12luax_ioErrorEP9lua_StatePKcz,b79,b79,b79,__ZNSt3__17getlineIcNS_11char_traitsIcEENS_9allocatorIcEEEERNS_13basic_istreamIT_T0_EES9_RNS_12basic_stringIS6_S7_T1_EES6_,b79,b79,b79,_lua_pushfstring,b79,b79,__ZN4love14luax_checktypeINS_5image9ImageDataEEEPT_P9lua_StateiNS_4TypeE,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,__ZN4love8graphics6opengl8Graphics8newImageERKNSt3__16vectorIPNS_5image19CompressedImageDataENS3_9allocatorIS7_EEEERKNS1_5Image5FlagsE,__ZN4love8graphics6opengl8Graphics8newImageERKNSt3__16vectorIPNS_5image9ImageDataENS3_9allocatorIS7_EEEERKNS1_5Image5FlagsE,b79,__ZN4love8graphics6opengl8Graphics7newFontEPNS_4font10RasterizerERKNS0_7Texture6FilterE,b79,__ZN4love8graphics6opengl8Graphics17newParticleSystemEPNS0_7TextureEi,b79,_luaL_argerror,b79,__ZN4love16luax_getfunctionEP9lua_StatePKcS3_,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_luax_typerror,b79,b79,__ZN4love8graphics6opengl8Graphics7newTextEPNS1_4FontERKNSt3__16vectorINS3_13ColoredStringENS5_9allocatorIS7_EEEE,b79
,b79,b79,b79,__ZN4love8graphics6opengl8Graphics13newScreenshotEPNS_5image5ImageEb,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZNK4love8graphics6opengl4Mesh16getAttributeInfoEiRi,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,__ZN4love14luax_checktypeINS_8graphics4QuadEEEPT_P9lua_StateiNS_4TypeE,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6insertEjPKc,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZNSt3__113basic_ostreamIcNS_11char_traitsIcEEElsEy,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_wuff_open,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_SDL_GetDisplayMode
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZN4love5sound7lullabyL13read_callbackEPvS2_j,__ZN4love5sound7lullabyL13seek_callbackEPvii,b79,b79,b79,b79,b79,b79,b79
,__ZN4love11luax_insistEP9lua_StateiPKc,b79,b79,b79,_luaL_getmetafield,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,_PHYSFS_mount,b79,b79,b79,b79,b79,b79,b79,b79,b79,_luaL_optinteger,b79,b79,b79,b79,b79,b79,_FT_Set_Pixel_Sizes,b79,b79,b79,b79,b79
,b79,b79,__ZNK4love5image9ImageData6insideEii,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,__ZN4love8graphics6opengl6Shader11compileCodeENS2_11ShaderStageERKNSt3__112basic_stringIcNS4_11char_traitsIcEENS4_9allocatorIcEEEE,b79,b79,b79,__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE5eraseEjj,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZN4love8graphics6opengl4Mesh26bindAttributeToShaderInputEiRKNSt3__112basic_stringIcNS3_11char_traitsIcEENS3_9allocatorIcEEEE,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZN4love14luax_checktypeINS_8graphics6opengl4FontEEEPT_P9lua_StateiNS_4TypeE,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,_SDL_CreateColorCursor,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZN15b2ContactSolver27SolveTOIPositionConstraintsEii,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,_SDL_CreateThread,b79,__ZN3dds6Parser9parseDataEPKvj,b79,b79,b79,b79,b79,b79,b79,_INT123_init_layer12_table,b79,b79,b79,_timeout_read,_plain_read,_posix_read,_posix_lseek,_nix_lseek,b79,_mallocAllocatorRealloc,b79,b79,b79,b79,b79,_zip_entry_cmp,b79,_zlibPhysfsAlloc,b79
,b79,b79,b79,b79,b79,b79,b79,_getF,_getS,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_generic_reader,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_pfr_extra_item_load_bitmap_info,_pfr_extra_item_load_font_id,_pfr_extra_item_load_stem_snaps,_pfr_extra_item_load_kerning_pairs,b79,b79,_psh_globals_new,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_FT_Outline_Decompose,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,_Round_None,_Round_To_Grid,_Round_Up_To_Grid,_Round_Down_To_Grid,_Round_To_Half_Grid,_Round_To_Double_Grid,_Round_Super,_Round_Super_45,_Project_x,_Project_y,_Project,_Dual_Project,b79,b79,b79,b79,b79,b79,b79,_t1_get_index,b79
,b79,_ft_gzip_alloc,b79,b79,b79,_zcalloc_9179,b79,_adler32_9183,b79,b79,b79,b79,b79,b79,b79,b79,_GLES2_GetOutputSize,b79,b79,b79,b79,b79,b79,b79,b79,b79,_GLES2_RenderDrawPoints,_GLES2_RenderDrawLines,_GLES2_RenderFillRects,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_SW_GetOutputSize,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,_SW_RenderDrawPoints,_SW_RenderDrawLines,_SW_RenderFillRects,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,_Emscripten_SetDisplayMode,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_Emscripten_GLES_MakeCurrent,b79,b79,b79,b79,b79,b79,_Emscripten_CreateCursor,b79
,b79,b79,b79,b79,_Emscripten_HandleMouseMove,_Emscripten_HandleMouseButton,_Emscripten_HandleMouseFocus,_Emscripten_HandleWheel,_Emscripten_HandleFocus,_Emscripten_HandleTouch,_Emscripten_HandleKey,_Emscripten_HandleKeyPress,_Emscripten_HandleFullscreenChange,_Emscripten_HandleResize,_Emscripten_HandleVisibilityChange,_Emscripten_HandleCanvasResize,b79,_eglInitialize__wrapper,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,_DUMMY_SetDisplayMode,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_Emscripten_JoyStickConnected,_Emscripten_JoyStickDisconnected
,_zcalloc,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_do_read,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,_newlocale,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,_vasprintf,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,__ZNSt3__112basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6appendIPcEENS_9enable_ifIXsr21__is_forward_iteratorIT_EE5valueERS5_E4typeES9_S9_,b79,b79,b79,b79,b79,__ZNSt3__112basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6appendIPwEENS_9enable_ifIXsr21__is_forward_iteratorIT_EE5valueERS5_E4typeES9_S9_,b79,b79,b79,b79,b79,b79,_wcrtomb
,b79,b79,_mbtowc,_mbrlen,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79,b79
,b79,b79,b79];
var FUNCTION_TABLE_dd = [b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,__ZN4love4math3__58__invokeEf,b80,b80,b80,__ZN4love4math3__98__invokeEf,__ZN4love4math4__108__invokeEf,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,__ZN4love7physics5box2d7Physics9scaleDownEf,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,__ZN4love8joystick8Joystick8clampvalEf,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80
,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80,b80];
var FUNCTION_TABLE_viiiidii = [b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,__ZN4love8graphics6opengl4Font16generateVerticesERKNS2_17ColoredCodepointsERNSt3__16vectorINS2_11GlyphVertexENS6_9allocatorIS8_EEEEfNS_6VectorEPNS2_8TextInfoE,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81
,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81,b81];
var FUNCTION_TABLE_diiiid = [b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,__ZN4love7physics5box2d5World15RayCastCallback13ReportFixtureEP9b2FixtureRK6b2Vec2S8_f,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82,b82
,b82,b82,b82,b82,b82];
var FUNCTION_TABLE_iiidddddddddi = [b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,__ZN4love8graphics6opengl11SpriteBatch4addqEPNS0_4QuadEfffffffffi,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83
,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83,b83];
var FUNCTION_TABLE_i = [b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,_Emscripten_Available,b84,b84,_DUMMY_Available,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,_SDL_GetDefaultCursor,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,_SDL_GetClipboardText,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,__ZN4love8graphics14isGammaCorrectEv,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,_SDL_GetError,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,_mpg123_init,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,_alGetError__wrapper,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,_PHYSFS_isInit,_PHYSFS_deinit,_PHYSFS_getLastError,b84,b84,b84,b84,b84,b84,b84,b84,_PHYSFS_getUserDir,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,__ZN4love8graphics7Texture16getDefaultFilterEv,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,_SDL_GetPlatform,b84
,_SDL_NumJoysticks,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,_SDL_CreateMutex,_SDL_CreateCond,b84
,b84,b84,b84,b84,__ZN4love6thread8newMutexEv,__ZN4love6thread14newConditionalEv,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,_eglWaitGL__wrapper,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,_emscripten_glCreateProgram__wrapper,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,_emscripten_glGetError__wrapper,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,__ZNSt3__16__clocEv,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,___ctype_toupper_loc,___ctype_tolower_loc,___ctype_b_loc,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,__ZNSt3__16locale5__imp12make_classicEv,__ZNSt3__16locale5__imp11make_globalEv,__ZNSt3__16locale8__globalEv,b84,b84,b84,b84,b84,b84,b84,___cxa_get_globals_fast,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84,b84
,b84,b84,b84];
var FUNCTION_TABLE_iiiiidii = [b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,_GLES2_RenderCopyEx,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,_SW_RenderCopyEx,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85,b85
,b85,b85,b85];
var FUNCTION_TABLE_iiiiiiiii = [b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,_res1_forward,b86,b86,_res2_forward,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,__ZNKSt3__18time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjP2tmcc,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,__ZNKSt3__18time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjP2tmcc,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,__ZNKSt3__17codecvtIDic11__mbstate_tE6do_outERS1_PKDiS5_RS5_PcS7_RS7_,__ZNKSt3__17codecvtIDic11__mbstate_tE5do_inERS1_PKcS5_RS5_PDiS7_RS7_,b86
,b86,b86,b86,b86,b86,b86,b86,__ZNKSt3__17codecvtIwc11__mbstate_tE6do_outERS1_PKwS5_RS5_PcS7_RS7_,__ZNKSt3__17codecvtIwc11__mbstate_tE5do_inERS1_PKcS5_RS5_PwS7_RS7_,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,__ZNKSt3__17codecvtIcc11__mbstate_tE6do_outERS1_PKcS5_RS5_PcS7_RS7_,__ZNKSt3__17codecvtIcc11__mbstate_tE5do_inERS1_PKcS5_RS5_PcS7_RS7_,b86,b86,b86
,b86,b86,b86,__ZNKSt3__17codecvtIDsc11__mbstate_tE6do_outERS1_PKDsS5_RS5_PcS7_RS7_,__ZNKSt3__17codecvtIDsc11__mbstate_tE5do_inERS1_PKcS5_RS5_PDsS7_RS7_,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,_SDL_BlendPoint_RGB888,_SDL_BlendPoint_ARGB8888,_SDL_BlendPoint_RGB,_SDL_BlendPoint_RGBA,_SDL_BlendPoint_RGB555,_SDL_BlendPoint_RGB565,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86,b86
,b86,b86,b86];
var FUNCTION_TABLE_viiiiiiiid = [b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,__ZN4love8graphics6opengl16NoneJoinPolyline10renderEdgeERNSt3__16vectorINS_6VectorENS3_9allocatorIS5_EEEES9_RS5_RfSA_RKS5_SD_f,b87,__ZN4love8graphics6opengl17MiterJoinPolyline10renderEdgeERNSt3__16vectorINS_6VectorENS3_9allocatorIS5_EEEES9_RS5_RfSA_RKS5_SD_f,b87,__ZN4love8graphics6opengl17BevelJoinPolyline10renderEdgeERNSt3__16vectorINS_6VectorENS3_9allocatorIS5_EEEES9_RS5_RfSA_RKS5_SD_f,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87,b87
,b87,b87,b87,b87,b87,b87,b87,b87,b87];
var FUNCTION_TABLE_iiiidddddi = [b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,__ZN4love7physics5box2d7Physics12newRopeJointEPNS1_4BodyES4_fffffb,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88
,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88,b88];

  return { _main: _main, _llvm_cttz_i32: _llvm_cttz_i32, _bitshift64Lshr: _bitshift64Lshr, _bitshift64Shl: _bitshift64Shl, _fflush: _fflush, ___cxa_is_pointer_type: ___cxa_is_pointer_type, _llvm_ctlz_i64: _llvm_ctlz_i64, _bitshift64Ashr: _bitshift64Ashr, _memset: _memset, _memcpy: _memcpy, _llvm_bswap_i32: _llvm_bswap_i32, _i64Subtract: _i64Subtract, _ntohs: _ntohs, _htonl: _htonl, _realloc: _realloc, _i64Add: _i64Add, _saveSetjmp: _saveSetjmp, _htons: _htons, _emscripten_GetProcAddress: _emscripten_GetProcAddress, ___errno_location: ___errno_location, _testSetjmp: _testSetjmp, ___cxa_can_catch: ___cxa_can_catch, _free: _free, _memmove: _memmove, _strstr: _strstr, _malloc: _malloc, __GLOBAL__I_000101: __GLOBAL__I_000101, __GLOBAL__sub_I_Shape_cpp: __GLOBAL__sub_I_Shape_cpp, __GLOBAL__sub_I_types_cpp: __GLOBAL__sub_I_types_cpp, __GLOBAL__sub_I_FileData_cpp: __GLOBAL__sub_I_FileData_cpp, __GLOBAL__sub_I_Graphics_cpp: __GLOBAL__sub_I_Graphics_cpp, __GLOBAL__sub_I_ParticleSystem_cpp: __GLOBAL__sub_I_ParticleSystem_cpp, __GLOBAL__sub_I_Texture_cpp: __GLOBAL__sub_I_Texture_cpp, __GLOBAL__sub_I_CompressedImageData_cpp: __GLOBAL__sub_I_CompressedImageData_cpp, __GLOBAL__sub_I_Compressor_cpp: __GLOBAL__sub_I_Compressor_cpp, __GLOBAL__sub_I_MathModule_cpp: __GLOBAL__sub_I_MathModule_cpp, __GLOBAL__sub_I_Cursor_cpp: __GLOBAL__sub_I_Cursor_cpp, __GLOBAL__sub_I_Joint_cpp: __GLOBAL__sub_I_Joint_cpp, __GLOBAL__sub_I_System_cpp: __GLOBAL__sub_I_System_cpp, __GLOBAL__sub_I_System_cpp_2273: __GLOBAL__sub_I_System_cpp_2273, __GLOBAL__sub_I_Window_cpp: __GLOBAL__sub_I_Window_cpp, __GLOBAL__sub_I_Memoizer_cpp: __GLOBAL__sub_I_Memoizer_cpp, __GLOBAL__sub_I_Audio_cpp: __GLOBAL__sub_I_Audio_cpp, __GLOBAL__sub_I_Source_cpp: __GLOBAL__sub_I_Source_cpp, __GLOBAL__sub_I_Event_cpp: __GLOBAL__sub_I_Event_cpp, __GLOBAL__sub_I_File_cpp: __GLOBAL__sub_I_File_cpp, __GLOBAL__sub_I_GlyphData_cpp: __GLOBAL__sub_I_GlyphData_cpp, __GLOBAL__sub_I_TrueTypeRasterizer_cpp: __GLOBAL__sub_I_TrueTypeRasterizer_cpp, __GLOBAL__sub_I_Canvas_cpp: __GLOBAL__sub_I_Canvas_cpp, __GLOBAL__sub_I_Font_cpp: __GLOBAL__sub_I_Font_cpp, __GLOBAL__sub_I_Image_cpp: __GLOBAL__sub_I_Image_cpp, __GLOBAL__sub_I_Mesh_cpp: __GLOBAL__sub_I_Mesh_cpp, __GLOBAL__sub_I_OpenGL_cpp: __GLOBAL__sub_I_OpenGL_cpp, __GLOBAL__sub_I_Shader_cpp: __GLOBAL__sub_I_Shader_cpp, __GLOBAL__sub_I_Volatile_cpp: __GLOBAL__sub_I_Volatile_cpp, __GLOBAL__sub_I_ImageData_cpp: __GLOBAL__sub_I_ImageData_cpp, __GLOBAL__sub_I_Joystick_cpp: __GLOBAL__sub_I_Joystick_cpp, __GLOBAL__sub_I_Joystick_cpp_4336: __GLOBAL__sub_I_Joystick_cpp_4336, __GLOBAL__sub_I_Keyboard_cpp: __GLOBAL__sub_I_Keyboard_cpp, __GLOBAL__sub_I_Keyboard_cpp_4545: __GLOBAL__sub_I_Keyboard_cpp_4545, __GLOBAL__sub_I_Cursor_cpp_4586: __GLOBAL__sub_I_Cursor_cpp_4586, __GLOBAL__sub_I_Channel_cpp: __GLOBAL__sub_I_Channel_cpp, __GLOBAL__sub_I_fastmix_cpp: __GLOBAL__sub_I_fastmix_cpp, __GLOBAL__sub_I_Body_cpp: __GLOBAL__sub_I_Body_cpp, __GLOBAL__sub_I_iostream_cpp: __GLOBAL__sub_I_iostream_cpp, runPostSets: runPostSets, _emscripten_replace_memory: _emscripten_replace_memory, stackAlloc: stackAlloc, stackSave: stackSave, stackRestore: stackRestore, establishStackSpace: establishStackSpace, setThrew: setThrew, setTempRet0: setTempRet0, getTempRet0: getTempRet0, dynCall_iiiiiid: dynCall_iiiiiid, dynCall_viiidiii: dynCall_viiidiii, dynCall_viiiddddddi: dynCall_viiiddddddi, dynCall_iiiiiii: dynCall_iiiiiii, dynCall_viiiiddi: dynCall_viiiiddi, dynCall_iidddddddddi: dynCall_iidddddddddi, dynCall_viiiiiiiiiii: dynCall_viiiiiiiiiii, dynCall_iidd: dynCall_iidd, dynCall_vidi: dynCall_vidi, dynCall_vidd: dynCall_vidd, dynCall_iiiiiiiiiiii: dynCall_iiiiiiiiiiii, dynCall_viidddiddddddd: dynCall_viidddiddddddd, dynCall_viiiiiiidi: dynCall_viiiiiiidi, dynCall_di: dynCall_di, dynCall_iiiiddi: dynCall_iiiiddi, dynCall_iiddddd: dynCall_iiddddd, dynCall_d: dynCall_d, dynCall_dii: dynCall_dii, dynCall_viiidddddi: dynCall_viiidddddi, dynCall_did: dynCall_did, dynCall_iiiiiiiiii: dynCall_iiiiiiiiii, dynCall_vdddddd: dynCall_vdddddd, dynCall_viiiddddi: dynCall_viiiddddi, dynCall_vdd: dynCall_vdd, dynCall_vdi: dynCall_vdi, dynCall_iiiid: dynCall_iiiid, dynCall_iiiiiiii: dynCall_iiiiiiii, dynCall_iiiddi: dynCall_iiiddi, dynCall_iiiiddddi: dynCall_iiiiddddi, dynCall_iidddd: dynCall_iidddd, dynCall_iiiiddddddi: dynCall_iiiiddddddi, dynCall_iiiiiiiidi: dynCall_iiiiiiiidi, dynCall_viidddd: dynCall_viidddd, dynCall_iiiii: dynCall_iiiii, dynCall_viiiiii: dynCall_viiiiii, dynCall_ddd: dynCall_ddd, dynCall_viiidi: dynCall_viiidi, dynCall_vidddd: dynCall_vidddd, dynCall_diid: dynCall_diid, dynCall_iiddd: dynCall_iiddd, dynCall_iiddi: dynCall_iiddi, dynCall_viiiiiiiiii: dynCall_viiiiiiiiii, dynCall_diii: dynCall_diii, dynCall_iiididdddddddd: dynCall_iiididdddddddd, dynCall_viiiddi: dynCall_viiiddi, dynCall_iiiiii: dynCall_iiiiii, dynCall_viiid: dynCall_viiid, dynCall_viidii: dynCall_viidii, dynCall_iiiiid: dynCall_iiiiid, dynCall_viiii: dynCall_viiii, dynCall_viiiii: dynCall_viiiii, dynCall_vid: dynCall_vid, dynCall_iiiidi: dynCall_iiiidi, dynCall_iiidd: dynCall_iiidd, dynCall_vii: dynCall_vii, dynCall_viidd: dynCall_viidd, dynCall_viddii: dynCall_viddii, dynCall_viidi: dynCall_viidi, dynCall_viddd: dynCall_viddd, dynCall_viiiiiiii: dynCall_viiiiiiii, dynCall_viiiiiiid: dynCall_viiiiiiid, dynCall_viiiiiiiiiiiiiii: dynCall_viiiiiiiiiiiiiii, dynCall_iid: dynCall_iid, dynCall_dddd: dynCall_dddd, dynCall_viiiiiii: dynCall_viiiiiii, dynCall_viiiiiiiii: dynCall_viiiiiiiii, dynCall_iii: dynCall_iii, dynCall_viiddi: dynCall_viiddi, dynCall_viiddddddddd: dynCall_viiddddddddd, dynCall_vdddd: dynCall_vdddd, dynCall_viii: dynCall_viii, dynCall_v: dynCall_v, dynCall_viid: dynCall_viid, dynCall_vd: dynCall_vd, dynCall_vi: dynCall_vi, dynCall_viddddddddd: dynCall_viddddddddd, dynCall_ii: dynCall_ii, dynCall_ddddd: dynCall_ddddd, dynCall_iiiddddddddd: dynCall_iiiddddddddd, dynCall_iiii: dynCall_iiii, dynCall_dd: dynCall_dd, dynCall_viiiidii: dynCall_viiiidii, dynCall_diiiid: dynCall_diiiid, dynCall_iiidddddddddi: dynCall_iiidddddddddi, dynCall_i: dynCall_i, dynCall_iiiiidii: dynCall_iiiiidii, dynCall_iiiiiiiii: dynCall_iiiiiiiii, dynCall_viiiiiiiid: dynCall_viiiiiiiid, dynCall_iiiidddddi: dynCall_iiiidddddi };
})
// EMSCRIPTEN_END_ASM
(Module.asmGlobalArg, Module.asmLibraryArg, buffer);
var real___GLOBAL__sub_I_Texture_cpp = asm["__GLOBAL__sub_I_Texture_cpp"]; asm["__GLOBAL__sub_I_Texture_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Texture_cpp.apply(null, arguments);
};

var real__main = asm["_main"]; asm["_main"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__main.apply(null, arguments);
};

var real___GLOBAL__sub_I_Source_cpp = asm["__GLOBAL__sub_I_Source_cpp"]; asm["__GLOBAL__sub_I_Source_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Source_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_File_cpp = asm["__GLOBAL__sub_I_File_cpp"]; asm["__GLOBAL__sub_I_File_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_File_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_Keyboard_cpp = asm["__GLOBAL__sub_I_Keyboard_cpp"]; asm["__GLOBAL__sub_I_Keyboard_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Keyboard_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_CompressedImageData_cpp = asm["__GLOBAL__sub_I_CompressedImageData_cpp"]; asm["__GLOBAL__sub_I_CompressedImageData_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_CompressedImageData_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_TrueTypeRasterizer_cpp = asm["__GLOBAL__sub_I_TrueTypeRasterizer_cpp"]; asm["__GLOBAL__sub_I_TrueTypeRasterizer_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_TrueTypeRasterizer_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_GlyphData_cpp = asm["__GLOBAL__sub_I_GlyphData_cpp"]; asm["__GLOBAL__sub_I_GlyphData_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_GlyphData_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_Audio_cpp = asm["__GLOBAL__sub_I_Audio_cpp"]; asm["__GLOBAL__sub_I_Audio_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Audio_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_Keyboard_cpp_4545 = asm["__GLOBAL__sub_I_Keyboard_cpp_4545"]; asm["__GLOBAL__sub_I_Keyboard_cpp_4545"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Keyboard_cpp_4545.apply(null, arguments);
};

var real__bitshift64Lshr = asm["_bitshift64Lshr"]; asm["_bitshift64Lshr"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__bitshift64Lshr.apply(null, arguments);
};

var real___GLOBAL__sub_I_Body_cpp = asm["__GLOBAL__sub_I_Body_cpp"]; asm["__GLOBAL__sub_I_Body_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Body_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_Joint_cpp = asm["__GLOBAL__sub_I_Joint_cpp"]; asm["__GLOBAL__sub_I_Joint_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Joint_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_Joystick_cpp = asm["__GLOBAL__sub_I_Joystick_cpp"]; asm["__GLOBAL__sub_I_Joystick_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Joystick_cpp.apply(null, arguments);
};

var real__bitshift64Shl = asm["_bitshift64Shl"]; asm["_bitshift64Shl"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__bitshift64Shl.apply(null, arguments);
};

var real___GLOBAL__sub_I_Mesh_cpp = asm["__GLOBAL__sub_I_Mesh_cpp"]; asm["__GLOBAL__sub_I_Mesh_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Mesh_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_System_cpp = asm["__GLOBAL__sub_I_System_cpp"]; asm["__GLOBAL__sub_I_System_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_System_cpp.apply(null, arguments);
};

var real__fflush = asm["_fflush"]; asm["_fflush"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__fflush.apply(null, arguments);
};

var real___GLOBAL__sub_I_Cursor_cpp = asm["__GLOBAL__sub_I_Cursor_cpp"]; asm["__GLOBAL__sub_I_Cursor_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Cursor_cpp.apply(null, arguments);
};

var real____cxa_is_pointer_type = asm["___cxa_is_pointer_type"]; asm["___cxa_is_pointer_type"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real____cxa_is_pointer_type.apply(null, arguments);
};

var real__llvm_ctlz_i64 = asm["_llvm_ctlz_i64"]; asm["_llvm_ctlz_i64"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__llvm_ctlz_i64.apply(null, arguments);
};

var real___GLOBAL__sub_I_MathModule_cpp = asm["__GLOBAL__sub_I_MathModule_cpp"]; asm["__GLOBAL__sub_I_MathModule_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_MathModule_cpp.apply(null, arguments);
};

var real__bitshift64Ashr = asm["_bitshift64Ashr"]; asm["_bitshift64Ashr"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__bitshift64Ashr.apply(null, arguments);
};

var real__llvm_cttz_i32 = asm["_llvm_cttz_i32"]; asm["_llvm_cttz_i32"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__llvm_cttz_i32.apply(null, arguments);
};

var real___GLOBAL__sub_I_types_cpp = asm["__GLOBAL__sub_I_types_cpp"]; asm["__GLOBAL__sub_I_types_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_types_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_System_cpp_2273 = asm["__GLOBAL__sub_I_System_cpp_2273"]; asm["__GLOBAL__sub_I_System_cpp_2273"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_System_cpp_2273.apply(null, arguments);
};

var real__llvm_bswap_i32 = asm["_llvm_bswap_i32"]; asm["_llvm_bswap_i32"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__llvm_bswap_i32.apply(null, arguments);
};

var real___GLOBAL__sub_I_Event_cpp = asm["__GLOBAL__sub_I_Event_cpp"]; asm["__GLOBAL__sub_I_Event_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Event_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_Window_cpp = asm["__GLOBAL__sub_I_Window_cpp"]; asm["__GLOBAL__sub_I_Window_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Window_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_Graphics_cpp = asm["__GLOBAL__sub_I_Graphics_cpp"]; asm["__GLOBAL__sub_I_Graphics_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Graphics_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_Cursor_cpp_4586 = asm["__GLOBAL__sub_I_Cursor_cpp_4586"]; asm["__GLOBAL__sub_I_Cursor_cpp_4586"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Cursor_cpp_4586.apply(null, arguments);
};

var real___GLOBAL__sub_I_ParticleSystem_cpp = asm["__GLOBAL__sub_I_ParticleSystem_cpp"]; asm["__GLOBAL__sub_I_ParticleSystem_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_ParticleSystem_cpp.apply(null, arguments);
};

var real__i64Subtract = asm["_i64Subtract"]; asm["_i64Subtract"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__i64Subtract.apply(null, arguments);
};

var real___GLOBAL__sub_I_Channel_cpp = asm["__GLOBAL__sub_I_Channel_cpp"]; asm["__GLOBAL__sub_I_Channel_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Channel_cpp.apply(null, arguments);
};

var real__ntohs = asm["_ntohs"]; asm["_ntohs"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__ntohs.apply(null, arguments);
};

var real__htonl = asm["_htonl"]; asm["_htonl"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__htonl.apply(null, arguments);
};

var real__realloc = asm["_realloc"]; asm["_realloc"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__realloc.apply(null, arguments);
};

var real__i64Add = asm["_i64Add"]; asm["_i64Add"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__i64Add.apply(null, arguments);
};

var real____cxa_can_catch = asm["___cxa_can_catch"]; asm["___cxa_can_catch"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real____cxa_can_catch.apply(null, arguments);
};

var real___GLOBAL__sub_I_Joystick_cpp_4336 = asm["__GLOBAL__sub_I_Joystick_cpp_4336"]; asm["__GLOBAL__sub_I_Joystick_cpp_4336"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Joystick_cpp_4336.apply(null, arguments);
};

var real___GLOBAL__I_000101 = asm["__GLOBAL__I_000101"]; asm["__GLOBAL__I_000101"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__I_000101.apply(null, arguments);
};

var real___GLOBAL__sub_I_Memoizer_cpp = asm["__GLOBAL__sub_I_Memoizer_cpp"]; asm["__GLOBAL__sub_I_Memoizer_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Memoizer_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_Shader_cpp = asm["__GLOBAL__sub_I_Shader_cpp"]; asm["__GLOBAL__sub_I_Shader_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Shader_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_Shape_cpp = asm["__GLOBAL__sub_I_Shape_cpp"]; asm["__GLOBAL__sub_I_Shape_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Shape_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_iostream_cpp = asm["__GLOBAL__sub_I_iostream_cpp"]; asm["__GLOBAL__sub_I_iostream_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_iostream_cpp.apply(null, arguments);
};

var real__htons = asm["_htons"]; asm["_htons"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__htons.apply(null, arguments);
};

var real__emscripten_GetProcAddress = asm["_emscripten_GetProcAddress"]; asm["_emscripten_GetProcAddress"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__emscripten_GetProcAddress.apply(null, arguments);
};

var real___GLOBAL__sub_I_Image_cpp = asm["__GLOBAL__sub_I_Image_cpp"]; asm["__GLOBAL__sub_I_Image_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Image_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_Font_cpp = asm["__GLOBAL__sub_I_Font_cpp"]; asm["__GLOBAL__sub_I_Font_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Font_cpp.apply(null, arguments);
};

var real____errno_location = asm["___errno_location"]; asm["___errno_location"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real____errno_location.apply(null, arguments);
};

var real__testSetjmp = asm["_testSetjmp"]; asm["_testSetjmp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__testSetjmp.apply(null, arguments);
};

var real__saveSetjmp = asm["_saveSetjmp"]; asm["_saveSetjmp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__saveSetjmp.apply(null, arguments);
};

var real__free = asm["_free"]; asm["_free"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__free.apply(null, arguments);
};

var real___GLOBAL__sub_I_OpenGL_cpp = asm["__GLOBAL__sub_I_OpenGL_cpp"]; asm["__GLOBAL__sub_I_OpenGL_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_OpenGL_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_Canvas_cpp = asm["__GLOBAL__sub_I_Canvas_cpp"]; asm["__GLOBAL__sub_I_Canvas_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Canvas_cpp.apply(null, arguments);
};

var real__memmove = asm["_memmove"]; asm["_memmove"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__memmove.apply(null, arguments);
};

var real__strstr = asm["_strstr"]; asm["_strstr"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__strstr.apply(null, arguments);
};

var real___GLOBAL__sub_I_Compressor_cpp = asm["__GLOBAL__sub_I_Compressor_cpp"]; asm["__GLOBAL__sub_I_Compressor_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Compressor_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_FileData_cpp = asm["__GLOBAL__sub_I_FileData_cpp"]; asm["__GLOBAL__sub_I_FileData_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_FileData_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_fastmix_cpp = asm["__GLOBAL__sub_I_fastmix_cpp"]; asm["__GLOBAL__sub_I_fastmix_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_fastmix_cpp.apply(null, arguments);
};

var real__malloc = asm["_malloc"]; asm["_malloc"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real__malloc.apply(null, arguments);
};

var real___GLOBAL__sub_I_Volatile_cpp = asm["__GLOBAL__sub_I_Volatile_cpp"]; asm["__GLOBAL__sub_I_Volatile_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_Volatile_cpp.apply(null, arguments);
};

var real___GLOBAL__sub_I_ImageData_cpp = asm["__GLOBAL__sub_I_ImageData_cpp"]; asm["__GLOBAL__sub_I_ImageData_cpp"] = function() {
assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
return real___GLOBAL__sub_I_ImageData_cpp.apply(null, arguments);
};
var __GLOBAL__sub_I_Texture_cpp = Module["__GLOBAL__sub_I_Texture_cpp"] = asm["__GLOBAL__sub_I_Texture_cpp"];
var _main = Module["_main"] = asm["_main"];
var __GLOBAL__sub_I_Source_cpp = Module["__GLOBAL__sub_I_Source_cpp"] = asm["__GLOBAL__sub_I_Source_cpp"];
var __GLOBAL__sub_I_File_cpp = Module["__GLOBAL__sub_I_File_cpp"] = asm["__GLOBAL__sub_I_File_cpp"];
var __GLOBAL__sub_I_Keyboard_cpp = Module["__GLOBAL__sub_I_Keyboard_cpp"] = asm["__GLOBAL__sub_I_Keyboard_cpp"];
var __GLOBAL__sub_I_CompressedImageData_cpp = Module["__GLOBAL__sub_I_CompressedImageData_cpp"] = asm["__GLOBAL__sub_I_CompressedImageData_cpp"];
var _memset = Module["_memset"] = asm["_memset"];
var __GLOBAL__sub_I_TrueTypeRasterizer_cpp = Module["__GLOBAL__sub_I_TrueTypeRasterizer_cpp"] = asm["__GLOBAL__sub_I_TrueTypeRasterizer_cpp"];
var __GLOBAL__sub_I_GlyphData_cpp = Module["__GLOBAL__sub_I_GlyphData_cpp"] = asm["__GLOBAL__sub_I_GlyphData_cpp"];
var __GLOBAL__sub_I_Audio_cpp = Module["__GLOBAL__sub_I_Audio_cpp"] = asm["__GLOBAL__sub_I_Audio_cpp"];
var __GLOBAL__sub_I_Keyboard_cpp_4545 = Module["__GLOBAL__sub_I_Keyboard_cpp_4545"] = asm["__GLOBAL__sub_I_Keyboard_cpp_4545"];
var _bitshift64Lshr = Module["_bitshift64Lshr"] = asm["_bitshift64Lshr"];
var __GLOBAL__sub_I_Body_cpp = Module["__GLOBAL__sub_I_Body_cpp"] = asm["__GLOBAL__sub_I_Body_cpp"];
var __GLOBAL__sub_I_Joint_cpp = Module["__GLOBAL__sub_I_Joint_cpp"] = asm["__GLOBAL__sub_I_Joint_cpp"];
var __GLOBAL__sub_I_Joystick_cpp = Module["__GLOBAL__sub_I_Joystick_cpp"] = asm["__GLOBAL__sub_I_Joystick_cpp"];
var _bitshift64Shl = Module["_bitshift64Shl"] = asm["_bitshift64Shl"];
var __GLOBAL__sub_I_Mesh_cpp = Module["__GLOBAL__sub_I_Mesh_cpp"] = asm["__GLOBAL__sub_I_Mesh_cpp"];
var __GLOBAL__sub_I_System_cpp = Module["__GLOBAL__sub_I_System_cpp"] = asm["__GLOBAL__sub_I_System_cpp"];
var _fflush = Module["_fflush"] = asm["_fflush"];
var __GLOBAL__sub_I_Cursor_cpp = Module["__GLOBAL__sub_I_Cursor_cpp"] = asm["__GLOBAL__sub_I_Cursor_cpp"];
var ___cxa_is_pointer_type = Module["___cxa_is_pointer_type"] = asm["___cxa_is_pointer_type"];
var _llvm_ctlz_i64 = Module["_llvm_ctlz_i64"] = asm["_llvm_ctlz_i64"];
var __GLOBAL__sub_I_MathModule_cpp = Module["__GLOBAL__sub_I_MathModule_cpp"] = asm["__GLOBAL__sub_I_MathModule_cpp"];
var _bitshift64Ashr = Module["_bitshift64Ashr"] = asm["_bitshift64Ashr"];
var _llvm_cttz_i32 = Module["_llvm_cttz_i32"] = asm["_llvm_cttz_i32"];
var __GLOBAL__sub_I_types_cpp = Module["__GLOBAL__sub_I_types_cpp"] = asm["__GLOBAL__sub_I_types_cpp"];
var __GLOBAL__sub_I_System_cpp_2273 = Module["__GLOBAL__sub_I_System_cpp_2273"] = asm["__GLOBAL__sub_I_System_cpp_2273"];
var _memcpy = Module["_memcpy"] = asm["_memcpy"];
var _llvm_bswap_i32 = Module["_llvm_bswap_i32"] = asm["_llvm_bswap_i32"];
var __GLOBAL__sub_I_Event_cpp = Module["__GLOBAL__sub_I_Event_cpp"] = asm["__GLOBAL__sub_I_Event_cpp"];
var __GLOBAL__sub_I_Window_cpp = Module["__GLOBAL__sub_I_Window_cpp"] = asm["__GLOBAL__sub_I_Window_cpp"];
var __GLOBAL__sub_I_Graphics_cpp = Module["__GLOBAL__sub_I_Graphics_cpp"] = asm["__GLOBAL__sub_I_Graphics_cpp"];
var __GLOBAL__sub_I_Cursor_cpp_4586 = Module["__GLOBAL__sub_I_Cursor_cpp_4586"] = asm["__GLOBAL__sub_I_Cursor_cpp_4586"];
var __GLOBAL__sub_I_ParticleSystem_cpp = Module["__GLOBAL__sub_I_ParticleSystem_cpp"] = asm["__GLOBAL__sub_I_ParticleSystem_cpp"];
var _i64Subtract = Module["_i64Subtract"] = asm["_i64Subtract"];
var __GLOBAL__sub_I_Channel_cpp = Module["__GLOBAL__sub_I_Channel_cpp"] = asm["__GLOBAL__sub_I_Channel_cpp"];
var _ntohs = Module["_ntohs"] = asm["_ntohs"];
var _htonl = Module["_htonl"] = asm["_htonl"];
var _realloc = Module["_realloc"] = asm["_realloc"];
var _i64Add = Module["_i64Add"] = asm["_i64Add"];
var ___cxa_can_catch = Module["___cxa_can_catch"] = asm["___cxa_can_catch"];
var __GLOBAL__sub_I_Joystick_cpp_4336 = Module["__GLOBAL__sub_I_Joystick_cpp_4336"] = asm["__GLOBAL__sub_I_Joystick_cpp_4336"];
var __GLOBAL__I_000101 = Module["__GLOBAL__I_000101"] = asm["__GLOBAL__I_000101"];
var __GLOBAL__sub_I_Memoizer_cpp = Module["__GLOBAL__sub_I_Memoizer_cpp"] = asm["__GLOBAL__sub_I_Memoizer_cpp"];
var __GLOBAL__sub_I_Shader_cpp = Module["__GLOBAL__sub_I_Shader_cpp"] = asm["__GLOBAL__sub_I_Shader_cpp"];
var __GLOBAL__sub_I_Shape_cpp = Module["__GLOBAL__sub_I_Shape_cpp"] = asm["__GLOBAL__sub_I_Shape_cpp"];
var __GLOBAL__sub_I_iostream_cpp = Module["__GLOBAL__sub_I_iostream_cpp"] = asm["__GLOBAL__sub_I_iostream_cpp"];
var _htons = Module["_htons"] = asm["_htons"];
var _emscripten_GetProcAddress = Module["_emscripten_GetProcAddress"] = asm["_emscripten_GetProcAddress"];
var __GLOBAL__sub_I_Image_cpp = Module["__GLOBAL__sub_I_Image_cpp"] = asm["__GLOBAL__sub_I_Image_cpp"];
var __GLOBAL__sub_I_Font_cpp = Module["__GLOBAL__sub_I_Font_cpp"] = asm["__GLOBAL__sub_I_Font_cpp"];
var ___errno_location = Module["___errno_location"] = asm["___errno_location"];
var runPostSets = Module["runPostSets"] = asm["runPostSets"];
var _testSetjmp = Module["_testSetjmp"] = asm["_testSetjmp"];
var _saveSetjmp = Module["_saveSetjmp"] = asm["_saveSetjmp"];
var _free = Module["_free"] = asm["_free"];
var __GLOBAL__sub_I_OpenGL_cpp = Module["__GLOBAL__sub_I_OpenGL_cpp"] = asm["__GLOBAL__sub_I_OpenGL_cpp"];
var __GLOBAL__sub_I_Canvas_cpp = Module["__GLOBAL__sub_I_Canvas_cpp"] = asm["__GLOBAL__sub_I_Canvas_cpp"];
var _memmove = Module["_memmove"] = asm["_memmove"];
var _strstr = Module["_strstr"] = asm["_strstr"];
var __GLOBAL__sub_I_Compressor_cpp = Module["__GLOBAL__sub_I_Compressor_cpp"] = asm["__GLOBAL__sub_I_Compressor_cpp"];
var __GLOBAL__sub_I_FileData_cpp = Module["__GLOBAL__sub_I_FileData_cpp"] = asm["__GLOBAL__sub_I_FileData_cpp"];
var __GLOBAL__sub_I_fastmix_cpp = Module["__GLOBAL__sub_I_fastmix_cpp"] = asm["__GLOBAL__sub_I_fastmix_cpp"];
var _malloc = Module["_malloc"] = asm["_malloc"];
var _emscripten_replace_memory = Module["_emscripten_replace_memory"] = asm["_emscripten_replace_memory"];
var __GLOBAL__sub_I_Volatile_cpp = Module["__GLOBAL__sub_I_Volatile_cpp"] = asm["__GLOBAL__sub_I_Volatile_cpp"];
var __GLOBAL__sub_I_ImageData_cpp = Module["__GLOBAL__sub_I_ImageData_cpp"] = asm["__GLOBAL__sub_I_ImageData_cpp"];
var dynCall_iiiiiid = Module["dynCall_iiiiiid"] = asm["dynCall_iiiiiid"];
var dynCall_viiidiii = Module["dynCall_viiidiii"] = asm["dynCall_viiidiii"];
var dynCall_viiiddddddi = Module["dynCall_viiiddddddi"] = asm["dynCall_viiiddddddi"];
var dynCall_iiiiiii = Module["dynCall_iiiiiii"] = asm["dynCall_iiiiiii"];
var dynCall_viiiiddi = Module["dynCall_viiiiddi"] = asm["dynCall_viiiiddi"];
var dynCall_iidddddddddi = Module["dynCall_iidddddddddi"] = asm["dynCall_iidddddddddi"];
var dynCall_viiiiiiiiiii = Module["dynCall_viiiiiiiiiii"] = asm["dynCall_viiiiiiiiiii"];
var dynCall_iidd = Module["dynCall_iidd"] = asm["dynCall_iidd"];
var dynCall_vidi = Module["dynCall_vidi"] = asm["dynCall_vidi"];
var dynCall_vidd = Module["dynCall_vidd"] = asm["dynCall_vidd"];
var dynCall_iiiiiiiiiiii = Module["dynCall_iiiiiiiiiiii"] = asm["dynCall_iiiiiiiiiiii"];
var dynCall_viidddiddddddd = Module["dynCall_viidddiddddddd"] = asm["dynCall_viidddiddddddd"];
var dynCall_viiiiiiidi = Module["dynCall_viiiiiiidi"] = asm["dynCall_viiiiiiidi"];
var dynCall_di = Module["dynCall_di"] = asm["dynCall_di"];
var dynCall_iiiiddi = Module["dynCall_iiiiddi"] = asm["dynCall_iiiiddi"];
var dynCall_iiddddd = Module["dynCall_iiddddd"] = asm["dynCall_iiddddd"];
var dynCall_d = Module["dynCall_d"] = asm["dynCall_d"];
var dynCall_dii = Module["dynCall_dii"] = asm["dynCall_dii"];
var dynCall_viiidddddi = Module["dynCall_viiidddddi"] = asm["dynCall_viiidddddi"];
var dynCall_did = Module["dynCall_did"] = asm["dynCall_did"];
var dynCall_iiiiiiiiii = Module["dynCall_iiiiiiiiii"] = asm["dynCall_iiiiiiiiii"];
var dynCall_vdddddd = Module["dynCall_vdddddd"] = asm["dynCall_vdddddd"];
var dynCall_viiiddddi = Module["dynCall_viiiddddi"] = asm["dynCall_viiiddddi"];
var dynCall_vdd = Module["dynCall_vdd"] = asm["dynCall_vdd"];
var dynCall_vdi = Module["dynCall_vdi"] = asm["dynCall_vdi"];
var dynCall_iiiid = Module["dynCall_iiiid"] = asm["dynCall_iiiid"];
var dynCall_iiiiiiii = Module["dynCall_iiiiiiii"] = asm["dynCall_iiiiiiii"];
var dynCall_iiiddi = Module["dynCall_iiiddi"] = asm["dynCall_iiiddi"];
var dynCall_iiiiddddi = Module["dynCall_iiiiddddi"] = asm["dynCall_iiiiddddi"];
var dynCall_iidddd = Module["dynCall_iidddd"] = asm["dynCall_iidddd"];
var dynCall_iiiiddddddi = Module["dynCall_iiiiddddddi"] = asm["dynCall_iiiiddddddi"];
var dynCall_iiiiiiiidi = Module["dynCall_iiiiiiiidi"] = asm["dynCall_iiiiiiiidi"];
var dynCall_viidddd = Module["dynCall_viidddd"] = asm["dynCall_viidddd"];
var dynCall_iiiii = Module["dynCall_iiiii"] = asm["dynCall_iiiii"];
var dynCall_viiiiii = Module["dynCall_viiiiii"] = asm["dynCall_viiiiii"];
var dynCall_ddd = Module["dynCall_ddd"] = asm["dynCall_ddd"];
var dynCall_viiidi = Module["dynCall_viiidi"] = asm["dynCall_viiidi"];
var dynCall_vidddd = Module["dynCall_vidddd"] = asm["dynCall_vidddd"];
var dynCall_diid = Module["dynCall_diid"] = asm["dynCall_diid"];
var dynCall_iiddd = Module["dynCall_iiddd"] = asm["dynCall_iiddd"];
var dynCall_iiddi = Module["dynCall_iiddi"] = asm["dynCall_iiddi"];
var dynCall_viiiiiiiiii = Module["dynCall_viiiiiiiiii"] = asm["dynCall_viiiiiiiiii"];
var dynCall_diii = Module["dynCall_diii"] = asm["dynCall_diii"];
var dynCall_iiididdddddddd = Module["dynCall_iiididdddddddd"] = asm["dynCall_iiididdddddddd"];
var dynCall_viiiddi = Module["dynCall_viiiddi"] = asm["dynCall_viiiddi"];
var dynCall_iiiiii = Module["dynCall_iiiiii"] = asm["dynCall_iiiiii"];
var dynCall_viiid = Module["dynCall_viiid"] = asm["dynCall_viiid"];
var dynCall_viidii = Module["dynCall_viidii"] = asm["dynCall_viidii"];
var dynCall_iiiiid = Module["dynCall_iiiiid"] = asm["dynCall_iiiiid"];
var dynCall_viiii = Module["dynCall_viiii"] = asm["dynCall_viiii"];
var dynCall_viiiii = Module["dynCall_viiiii"] = asm["dynCall_viiiii"];
var dynCall_vid = Module["dynCall_vid"] = asm["dynCall_vid"];
var dynCall_iiiidi = Module["dynCall_iiiidi"] = asm["dynCall_iiiidi"];
var dynCall_iiidd = Module["dynCall_iiidd"] = asm["dynCall_iiidd"];
var dynCall_vii = Module["dynCall_vii"] = asm["dynCall_vii"];
var dynCall_viidd = Module["dynCall_viidd"] = asm["dynCall_viidd"];
var dynCall_viddii = Module["dynCall_viddii"] = asm["dynCall_viddii"];
var dynCall_viidi = Module["dynCall_viidi"] = asm["dynCall_viidi"];
var dynCall_viddd = Module["dynCall_viddd"] = asm["dynCall_viddd"];
var dynCall_viiiiiiii = Module["dynCall_viiiiiiii"] = asm["dynCall_viiiiiiii"];
var dynCall_viiiiiiid = Module["dynCall_viiiiiiid"] = asm["dynCall_viiiiiiid"];
var dynCall_viiiiiiiiiiiiiii = Module["dynCall_viiiiiiiiiiiiiii"] = asm["dynCall_viiiiiiiiiiiiiii"];
var dynCall_iid = Module["dynCall_iid"] = asm["dynCall_iid"];
var dynCall_dddd = Module["dynCall_dddd"] = asm["dynCall_dddd"];
var dynCall_viiiiiii = Module["dynCall_viiiiiii"] = asm["dynCall_viiiiiii"];
var dynCall_viiiiiiiii = Module["dynCall_viiiiiiiii"] = asm["dynCall_viiiiiiiii"];
var dynCall_iii = Module["dynCall_iii"] = asm["dynCall_iii"];
var dynCall_viiddi = Module["dynCall_viiddi"] = asm["dynCall_viiddi"];
var dynCall_viiddddddddd = Module["dynCall_viiddddddddd"] = asm["dynCall_viiddddddddd"];
var dynCall_vdddd = Module["dynCall_vdddd"] = asm["dynCall_vdddd"];
var dynCall_viii = Module["dynCall_viii"] = asm["dynCall_viii"];
var dynCall_v = Module["dynCall_v"] = asm["dynCall_v"];
var dynCall_viid = Module["dynCall_viid"] = asm["dynCall_viid"];
var dynCall_vd = Module["dynCall_vd"] = asm["dynCall_vd"];
var dynCall_vi = Module["dynCall_vi"] = asm["dynCall_vi"];
var dynCall_viddddddddd = Module["dynCall_viddddddddd"] = asm["dynCall_viddddddddd"];
var dynCall_ii = Module["dynCall_ii"] = asm["dynCall_ii"];
var dynCall_ddddd = Module["dynCall_ddddd"] = asm["dynCall_ddddd"];
var dynCall_iiiddddddddd = Module["dynCall_iiiddddddddd"] = asm["dynCall_iiiddddddddd"];
var dynCall_iiii = Module["dynCall_iiii"] = asm["dynCall_iiii"];
var dynCall_dd = Module["dynCall_dd"] = asm["dynCall_dd"];
var dynCall_viiiidii = Module["dynCall_viiiidii"] = asm["dynCall_viiiidii"];
var dynCall_diiiid = Module["dynCall_diiiid"] = asm["dynCall_diiiid"];
var dynCall_iiidddddddddi = Module["dynCall_iiidddddddddi"] = asm["dynCall_iiidddddddddi"];
var dynCall_i = Module["dynCall_i"] = asm["dynCall_i"];
var dynCall_iiiiidii = Module["dynCall_iiiiidii"] = asm["dynCall_iiiiidii"];
var dynCall_iiiiiiiii = Module["dynCall_iiiiiiiii"] = asm["dynCall_iiiiiiiii"];
var dynCall_viiiiiiiid = Module["dynCall_viiiiiiiid"] = asm["dynCall_viiiiiiiid"];
var dynCall_iiiidddddi = Module["dynCall_iiiidddddi"] = asm["dynCall_iiiidddddi"];
;

Runtime.stackAlloc = asm['stackAlloc'];
Runtime.stackSave = asm['stackSave'];
Runtime.stackRestore = asm['stackRestore'];
Runtime.establishStackSpace = asm['establishStackSpace'];

Runtime.setTempRet0 = asm['setTempRet0'];
Runtime.getTempRet0 = asm['getTempRet0'];



// === Auto-generated postamble setup entry stuff ===




function ExitStatus(status) {
  this.name = "ExitStatus";
  this.message = "Program terminated with exit(" + status + ")";
  this.status = status;
};
ExitStatus.prototype = new Error();
ExitStatus.prototype.constructor = ExitStatus;

var initialStackTop;
var preloadStartTime = null;
var calledMain = false;

dependenciesFulfilled = function runCaller() {
  // If run has never been called, and we should call run (INVOKE_RUN is true, and Module.noInitialRun is not false)
  if (!Module['calledRun']) run();
  if (!Module['calledRun']) dependenciesFulfilled = runCaller; // try this again later, after new deps are fulfilled
}

Module['callMain'] = Module.callMain = function callMain(args) {
  assert(runDependencies == 0, 'cannot call main when async dependencies remain! (listen on __ATMAIN__)');
  assert(__ATPRERUN__.length == 0, 'cannot call main when preRun functions remain to be called');

  args = args || [];

  ensureInitRuntime();

  var argc = args.length+1;
  function pad() {
    for (var i = 0; i < 4-1; i++) {
      argv.push(0);
    }
  }
  var argv = [allocate(intArrayFromString(Module['thisProgram']), 'i8', ALLOC_NORMAL) ];
  pad();
  for (var i = 0; i < argc-1; i = i + 1) {
    argv.push(allocate(intArrayFromString(args[i]), 'i8', ALLOC_NORMAL));
    pad();
  }
  argv.push(0);
  argv = allocate(argv, 'i32', ALLOC_NORMAL);


  try {

    var ret = Module['_main'](argc, argv, 0);


    // if we're not running an evented main loop, it's time to exit
    exit(ret, /* implicit = */ true);
  }
  catch(e) {
    if (e instanceof ExitStatus) {
      // exit() throws this once it's done to make sure execution
      // has been stopped completely
      return;
    } else if (e == 'SimulateInfiniteLoop') {
      // running an evented main loop, don't immediately exit
      Module['noExitRuntime'] = true;
      return;
    } else {
      if (e && typeof e === 'object' && e.stack) Module.printErr('exception thrown: ' + [e, e.stack]);
      throw e;
    }
  } finally {
    calledMain = true;
  }
}




function run(args) {
  args = args || Module['arguments'];

  if (preloadStartTime === null) preloadStartTime = Date.now();

  if (runDependencies > 0) {
    Module.printErr('run() called, but dependencies remain, so not running');
    return;
  }

  preRun();

  if (runDependencies > 0) return; // a preRun added a dependency, run will be called later
  if (Module['calledRun']) return; // run may have just been called through dependencies being fulfilled just in this very frame

  function doRun() {
    if (Module['calledRun']) return; // run may have just been called while the async setStatus time below was happening
    Module['calledRun'] = true;

    if (ABORT) return; 

    ensureInitRuntime();

    preMain();

    if (ENVIRONMENT_IS_WEB && preloadStartTime !== null) {
      Module.printErr('pre-main prep time: ' + (Date.now() - preloadStartTime) + ' ms');
    }

    if (Module['onRuntimeInitialized']) Module['onRuntimeInitialized']();

    if (Module['_main'] && shouldRunNow) Module['callMain'](args);

    postRun();
  }

  if (Module['setStatus']) {
    Module['setStatus']('Running...');
    setTimeout(function() {
      setTimeout(function() {
        Module['setStatus']('');
      }, 1);
      doRun();
    }, 1);
  } else {
    doRun();
  }
}
Module['run'] = Module.run = run;

function exit(status, implicit) {
  if (implicit && Module['noExitRuntime']) {
    Module.printErr('exit(' + status + ') implicitly called by end of main(), but noExitRuntime, so not exiting the runtime (you can use emscripten_force_exit, if you want to force a true shutdown)');
    return;
  }

  if (Module['noExitRuntime']) {
    Module.printErr('exit(' + status + ') called, but noExitRuntime, so halting execution but not exiting the runtime or preventing further async execution (you can use emscripten_force_exit, if you want to force a true shutdown)');
  } else {

    ABORT = true;
    EXITSTATUS = status;
    STACKTOP = initialStackTop;

    exitRuntime();

    if (Module['onExit']) Module['onExit'](status);
  }

  if (ENVIRONMENT_IS_NODE) {
    process['exit'](status);
  } else if (ENVIRONMENT_IS_SHELL && typeof quit === 'function') {
    quit(status);
  }
  // if we reach here, we must throw an exception to halt the current execution
  throw new ExitStatus(status);
}
Module['exit'] = Module.exit = exit;

var abortDecorators = [];

function abort(what) {
  if (what !== undefined) {
    Module.print(what);
    Module.printErr(what);
    what = JSON.stringify(what)
  } else {
    what = '';
  }

  ABORT = true;
  EXITSTATUS = 1;

  var extra = '';

  var output = 'abort(' + what + ') at ' + stackTrace() + extra;
  if (abortDecorators) {
    abortDecorators.forEach(function(decorator) {
      output = decorator(output, what);
    });
  }
  throw output;
}
Module['abort'] = Module.abort = abort;

// {{PRE_RUN_ADDITIONS}}

if (Module['preInit']) {
  if (typeof Module['preInit'] == 'function') Module['preInit'] = [Module['preInit']];
  while (Module['preInit'].length > 0) {
    Module['preInit'].pop()();
  }
}

// shouldRunNow refers to calling main(), not run().
var shouldRunNow = false;
if (Module['noInitialRun']) {
  shouldRunNow = false;
}


run();

// {{POST_RUN_ADDITIONS}}






// {{MODULE_ADDITIONS}}



  Module.setStatus('Syncing filesystem...')
  FS.mkdir(ENV.HOME + '/love');
  FS.mount(IDBFS, {}, ENV.HOME + '/love');

  FS.syncfs(true, function (err) {
    if (err) {
      Module.printErr(err);
    } else {
      Module.didSyncFS = true;
      Module.setStatus('')
    }
  });

//# sourceMappingURL=love.js.map